C***********************
      Subroutine BD_TIPS_2017(
     I MOL,            ! HITRAN molecule number
     I Temp,            ! temperature in K
     I ISO,            ! isotopomer index
     O gi,            ! state independent degeneracy factor
     O QT)            ! total internal partition sum
C*********************** 
C    Subroutine BD_TIPS_2017 written by R.R. Gamache
C 
c
c...date last changed 27         June, 2017
c
c MIT License
c
c Copyright (c) [2016] [Robert Gamache]
c
c Permission is hereby granted, free of charge, to any person obtaining a copy
c of this software and associated documentation files (the "Software"), to deal
c in the Software without restriction, including without limitation the rights
c to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
c copies of the Software, and to permit persons to whom the Software is
c furnished to do so, subject to the following conditions:
c
c The above copyright notice and this permission notice shall be included in all
c copies or substantial portions of the Software.
c
c THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
c IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
c FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
c AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
c LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
c OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
c SOFTWARE.
c
c
c  --  UPDATES  --
C    17 November, 2016 too many to list - see publication 
C    Gamache et al., Total internal partition sums for molecular isotopologues on the HITRAN database and beyond, JQSRT, 2017. 
c  --  UPDATES  --
C    7 May 2104 addition of 737 and 646 of CO2, 
C    16 April 2104 corrected and greatly improved Q(T)s for O2 including 88, 78, and 77 isotopologues, 
C    18 March 2011 addition of molecules 43-51, 838,837 of CO2, 
C     1222 of C2H2, CH3D, 13CH3D, 13C12CH6,CH3Br
C    19 July 2010 correction of the sign of D in CO  constants
C    13 July 2010 addition of several new species/isotopologues: ^13C^18O2, 
C       ^18O^13C^17O, ^13CH3D,  ^13C12CH2, CF4: note a number of the block 
C       data codes were updated as well
C    15 May 2008 Dijon Q(T) values for CH4
C    18 December 2003 better vibrational fundamentals for PH3
c
C    This program calculates the total internal
c    partition sum (TIPS) for a given molecule, isotopologue, and
c    temperature.  Current limitations are the molecular species on the
c    HITRAN molecular database plus a few additional  molecules and the temperature range 1 - 3500 K.
c
c...This program calculates the TIPS by 4-point LaaGrange interpolation
c
c..  JQSRT - 82, 401-412, 2003
c..  J. Fischer(a) R.R. Gamache(a&), A. Goldman(b), L.S. Rothman(c), and A. Perrin(d)
c..  
c..  (a)  Department of Environmental, Earth, and Atmospheric Sciences, 
c..       University of Massachusetts Lowell, Lowell, MA 01854, U.S.A.
c..  
c..  (b)  Department of Physics, University of Denver, Denver, CO 80208, U.S.A.
c..  
c..  (c)  Harvard-Smithsonian Center for Astrophysics, 60 Garden St, Cambridge, MA 02138 USA
c..  
c..  (d)  Laboratoire de Photophysique Molculaire, Universit Paris Sud, 91405 Orsay, FRANCE
c..  
c..  &  Corresponding author. Email address: Robert_Gamache@uml.edu
c..  Abstract
c..        Total internal partition sums (TIPS) are calculated for all molecular species in 
c..  the 2000 HITRAN database.  In addition, the TIPS for 13 other isotopomers/isotopologues 
c..  of ozone and carbon dioxide are presented.  The calculations address the corrections 
c..  suggested by Goldman et al. (JQSRT 2000;66:55-86).  The calculations consider the 
c..  temperature range 70-3000 K to be applicable to a variety of remote sensing needs.  
c..  The method of calculation for each molecular species is stated and comparisons with 
c..  data from the literature are discussed.  A new method of recall for the partition sums, 
c..  Lagrange 4-point interpolation, is developed.  This method, unlike previous versions of 
c..  the TIPS code, allows all molecular species to be considered.  
c
      implicit DOUBLE PRECISION (a-h,o-z)
 
c++
      INCLUDE 'SPECIES_2016.CMN'
c++
      PARAMETER (NT=4510)
c++
      INCLUDE 'ISOTOPS.CMN'
      INCLUDE 'MOLEC.CMN'
c++:  bd-QT
      COMMON/Temperatures/tdat(NT)
 
      character*30 stopNgo
      dimension Tmax(NMOL,Max_ISO)
c
c...max temperatures for isotopologues
      data (Tmax(1,I),I=1,9)/5000.,5000.,5000.,5000.,5000.,5000.,
     + 6000.,6000.,6000./  ! H2O
      data (Tmax(2,I),I=1,13)/5000.,5000.,3500.,3500.,3500.,3500.,
     + 5000.,3500.,5000.,5000.,3500.,5000.,5000./  ! CO2
      data (Tmax(3,I),I=1,18)/3500.,3500.,3500.,3500.,3500.,3500.,
     + 3500.,3500.,3500.,  ! O3
     +                        3500.,3500.,3500.,3500.,3500.,3500.,
     + 3500.,3500.,3500./
      data (Tmax(4,I),I=1,5)/5000.,3500.,3500.,3500.,3500./  ! N2O
      data (Tmax(5,I),I=1,9)/9000.,9000.,9000.,9000.,9000.,9000.,
     + 9000.,9000.,9000./  ! CO
      data (Tmax(6,I),I=1,4)/3500.,3500.,4500.,3500./ ! CH4
      data (Tmax(7,I),I=1,6)/7500.,7500.,7500.,7500.,7500.,
     + 7500./  ! O2
      data (Tmax(8,I),I=1,3)/3500.,3500.,3500./  ! NO
      data (Tmax(9,I),I=1,2)/3500.,3500./  ! SO2
      data Tmax(10,1)/3500./  ! NO2
      data (Tmax(11,I),I=1,2)/6000.,6000./  ! NH3
      data (Tmax(12,I),I=1,2)/3500.,3500./  ! HNO3
      data (Tmax(13,I),I=1,3)/9000.,3500.,3500./  ! OH
      data (Tmax(14,I),I=1,2)/6000.,6000./  ! HF
      data (Tmax(15,I),I=1,4)/6000.,6000.,6000.,4000./  ! HCl
      data (Tmax(16,I),I=1,4)/6000.,6000.,6000.,6000./  ! HBr
      data (Tmax(17,I),I=1,2)/6000.,6000./  ! HI
      data (Tmax(18,I),I=1,2)/3500.,3500./  ! ClO
      data (Tmax(19,I),I=1,5)/3500.,3500.,3500.,3500.,3500./  ! OCS
      data (Tmax(20,I),I=1,3)/3500.,3500.,3500./  ! H2CO
      data (Tmax(21,I),I=1,2)/3500.,3500./  ! HOCl
      data (Tmax(22,I),I=1,3)/9000.,9000.,9000./  ! N2
      data (Tmax(23,I),I=1,3)/3500.,3500.,3500./  ! HCN
      data (Tmax(24,I),I=1,2)/3500.,3500./  ! CH3Cl
      data Tmax(25,1)/6000./  ! H2O2
      data (Tmax(26,I),I=1,3)/5000.,3500.,3500./  ! C2H2 
      data (Tmax(27,I),I=1,2)/3500.,3500./  ! C2H6
      data Tmax(28,1)/4500./  ! PH3
      data (Tmax(29,I),I=1,2)/3500.,3500./  ! COF2
      data Tmax(30,1)/3500./  ! SF6
      data (Tmax(31,I),I=1,3)/4000.,3500.,3500./  ! H2S
      data Tmax(32,1)/3500./  ! HCOOH
      data Tmax(33,1)/3500./  ! HO2
c     O atom do data
      data Tmax(34,1)/10./  ! HO2
      data (Tmax(35,I),I=1,2)/3500.,3500./  ! ClONO2
      data Tmax(36,1)/3500./  ! NO+
      data (Tmax(37,I),I=1,2)/3500.,3500./  ! HOBr
      data (Tmax(38,I),I=1,2)/3500.,3500./  ! C2H4
      data Tmax(39,1)/3500./  ! CH3OH
      data (Tmax(40,I),I=1,2)/3500.,3500./  ! CH3Br
      data (Tmax(41,I),I=1,4)/3500.,3500.,3500.,3500./  ! CH3CN
      data Tmax(42,1)/3010./  ! CF4
      data Tmax(43,1)/3500./  ! C4H2
      data (Tmax(44,I),I=1,6)/3500.,3500.,3500.,3500.,3500.,
     + 3500./  ! HC3N
      data (Tmax(45,I),I=1,2)/6000.,6000./  ! H2
      data (Tmax(46,I),I=1,4)/3500.,3500.,3500.,3500./  ! CS
      data Tmax(47,1)/3500./  ! SO3
      data (Tmax(48,I),I=1,2)/3500.,3500./  ! C2N2
      data (Tmax(49,I),I=1,2)/5000.,5000./  ! COCl2
      data (Tmax(50,I),I=1,3)/3500.,3500.,3500./  ! SO
      data Tmax(51,1)/3500./  ! C3H4
      data Tmax(52,1)/3500./  ! CH3
      data (Tmax(53,I),I=1,4)/3500.,3500.,3500.,3500./  ! CS2
 
c...fill T array
      I = 0
      do IT = 1,20
        I = I+1
        Tdat(I) = IT*1.
      enddo
      do IT = 22,9000,2
        I = I+1
        Tdat(I) = IT*1.
      enddo
 
c
C
C     GO TO PARTICULAR MOLECULE:  
 
      IF(MOL.EQ.1) THEN 
      CALL QT_H2O(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
C 
      IF(MOL.EQ.2) THEN 
      CALL QT_CO2(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
C 
      IF(MOL.EQ.3) THEN 
      CALL QT_O3(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
C 
      IF(MOL.EQ.4) THEN 
      CALL QT_N2O(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
C 
      IF(MOL.EQ.5) THEN 
      CALL QT_CO(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
C 
      IF(MOL.EQ.6) THEN 
      CALL QT_CH4(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
C 
      IF(MOL.EQ.7) THEN 
      CALL QT_O2(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
C 
      IF(MOL.EQ.8) THEN 
      CALL QT_NO(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
C 
      IF(MOL.EQ.9) THEN 
      CALL QT_SO2(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
C 
      IF(MOL.EQ.10) THEN
      CALL QT_NO2(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
C 
      IF(MOL.EQ.11) THEN
      CALL QT_NH3(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
C 
      IF(MOL.EQ.12) THEN
      CALL QT_HNO3(Temp,ISO,gi,QT)
      go to 100 
      ENDIF 
C 
      IF(MOL.EQ.13) THEN
      CALL QT_OH(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
C 
      IF(MOL.EQ.14) THEN
      CALL QT_HF(Temp,ISO,gi,QT) 
      go to 100
      ENDIF 
C 
      IF(MOL.EQ.15) THEN
      CALL QT_HCL(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
C 
      IF(MOL.EQ.16) THEN
      CALL QT_HBR(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
C 
      IF(MOL.EQ.17) THEN
      CALL QT_HI(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
C 
      IF(MOL.EQ.18) THEN
      CALL QT_CLO(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
C 
      IF(MOL.EQ.19) THEN
      CALL QT_OCS(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
C 
      IF(MOL.EQ.20) THEN
      CALL QT_H2CO(Temp,ISO,gi,QT) 
      go to 100 
      ENDIF 
C 
      IF(MOL.EQ.21) THEN
      CALL QT_HOCL(Temp,ISO,gi,QT) 
      go to 100 
      ENDIF 
C 
      IF(MOL.EQ.22) THEN
      CALL QT_N2(Temp,ISO,gi,QT) 
      go to 100
      ENDIF 
C 
      IF(MOL.EQ.23) THEN
      CALL QT_HCN(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
C 
      IF(MOL.EQ.24) THEN
      CALL QT_CH3CL(Temp,ISO,gi,QT)
      go to 100 
      ENDIF 
C 
      IF(MOL.EQ.25) THEN
      CALL QT_H2O2(Temp,ISO,gi,QT) 
      go to 100 
      ENDIF 
C 
      IF(MOL.EQ.26) THEN
      CALL QT_C2H2(Temp,ISO,gi,QT)
      go to 100 
      ENDIF 
C 
      IF(MOL.EQ.27) THEN
      CALL QT_C2H6(Temp,ISO,gi,QT)
      go to 100 
      ENDIF 
C 
      IF(MOL.EQ.28) THEN
      CALL QT_PH3(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
c
      IF(MOL.EQ.29) THEN 
      CALL QT_COF2(Temp,ISO,gi,QT)
      go to 100 
      ENDIF 
c
      IF(MOL.EQ.30) THEN 
      CALL QT_SF6(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
c
      IF(MOL.EQ.31) THEN
      CALL QT_H2S(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
c
      IF(MOL.EQ.32) THEN
      CALL QT_HCOOH(Temp,ISO,gi,QT)
      go to 100 
      ENDIF 
c
      IF(MOL.EQ.33) THEN 
      CALL QT_HO2(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
c
      IF(MOL.EQ.34) THEN 
c...not applicable to O
      gi=0
      QT = 0.
      go to 100
      ENDIF 
c
      IF(MOL.EQ.35) THEN 
      CALL QT_ClONO2(Temp,ISO,gi,QT)
      go to 100 
      ENDIF 
c
      IF(MOL.EQ.36) THEN 
      CALL QT_NOp(Temp,ISO,gi,QT)
      go to 100
      ENDIF     
C 
      IF(MOL.EQ.37) THEN 
      CALL QT_HOBr(Temp,ISO,gi,QT)
      go to 100
      ENDIF     
C 
      IF(MOL.EQ.38) THEN 
      CALL QT_C2H4(Temp,ISO,gi,QT)
      go to 100
      ENDIF     
C 
      IF(MOL.EQ.39) THEN 
      CALL QT_CH3OH(Temp,ISO,gi,QT)
      go to 100
      ENDIF     
c
      IF(MOL.EQ.40) THEN 
      CALL QT_CH3Br(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
c
      IF(MOL.EQ.41) THEN 
      CALL QT_CH3CN(Temp,ISO,gi,QT)
      go to 100
      ENDIF 
c
      IF(MOL.EQ.42) THEN 
c...no data for CF4
      gi=0
      QT = 0.
      go to 100
      ENDIF     
c
      IF(MOL.EQ.43) THEN 
      CALL QT_C4H2(Temp,ISO,gi,QT)
      go to 100
      ENDIF     
c
      IF(MOL.EQ.44) THEN 
      CALL QT_HC3N(Temp,ISO,gi,QT)
      go to 100
      ENDIF     
c
      IF(MOL.EQ.45) THEN 
      CALL QT_H2(Temp,ISO,gi,QT)
      go to 100
      ENDIF     
c
      IF(MOL.EQ.46) THEN 
      CALL QT_CS(Temp,ISO,gi,QT)
      go to 100
      ENDIF     
c
      IF(MOL.EQ.47) THEN 
      CALL QT_SO3(Temp,ISO,gi,QT)
      go to 100
      ENDIF     
c
      IF(MOL.EQ.48) THEN 
      CALL QT_C2N2(Temp,ISO,gi,QT)
      go to 100
      ENDIF     
c
      IF(MOL.EQ.49) THEN 
      CALL QT_COCL2(Temp,ISO,gi,QT)
      go to 100
      ENDIF     
c
      IF(MOL.EQ.50) THEN 
      CALL QT_SO(Temp,ISO,gi,QT)
      go to 100
      ENDIF     
c
      IF(MOL.EQ.51) THEN 
      CALL QT_C3H4(Temp,ISO,gi,QT)
      go to 100
      ENDIF     
c
      IF(MOL.EQ.52) THEN 
      CALL QT_CH3(Temp,ISO,gi,QT)
      go to 100
      ENDIF     
c
      IF(MOL.EQ.53) THEN 
      CALL QT_CS2(Temp,ISO,gi,QT)
      go to 100
      ENDIF     
C 
 100  RETURN
      END 
  
  
C**************************************
      SUBROUTINE CLEAR
C**************************************
      implicit DOUBLE PRECISION (a-h,o-z)
C
      DO 10 I = 1,24
       WRITE (*,'(1X)')
   10 CONTINUE
      RETURN
      END
  
  
      include 'BD_MOL_2016.FOR'
      include 'BD_ISO_2016.for'
  
  
c
c     *****************
      Subroutine QT_H2O   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 9),QofT( 9,NT),Q(NT)
      data (Tmx(I),I=1,9)/5000.,5000.,5000.,5000.,5000.,5000.,6000.,
     + 6000.,6000./
      data xgj/ 1.,1.,6.,6.,6.,36.,1.,1.,6./
c...      H2O
c...        --       161
      data (QofT( 1,J),J=1,2408)/ 1.0000E+00, 1.0000E+00, 1.0001E+00,
     +  1.0017E+00, 1.0097E+00, 1.0307E+00, 1.0706E+00, 1.1329E+00,
     + 1.2189E+00, 1.3284E+00, 1.4601E+00, 1.6123E+00, 1.7830E+00,
     +  1.9702E+00, 2.1721E+00, 2.3870E+00, 2.6134E+00, 2.8499E+00,
     + 3.0954E+00, 3.3489E+00, 3.8767E+00, 4.4281E+00, 4.9993E+00,
     +  5.5878E+00, 6.1918E+00, 6.8103E+00, 7.4424E+00, 8.0877E+00,
     + 8.7459E+00, 9.4170E+00, 1.0101E+01, 1.0797E+01, 1.1506E+01,
     +  1.2227E+01, 1.2962E+01, 1.3708E+01, 1.4467E+01, 1.5239E+01,
     + 1.6023E+01, 1.6819E+01, 1.7627E+01, 1.8447E+01, 1.9280E+01,
     +  2.0124E+01, 2.0980E+01, 2.1847E+01, 2.2726E+01, 2.3617E+01,
     + 2.4519E+01, 2.5432E+01, 2.6356E+01, 2.7291E+01, 2.8237E+01,
     +  2.9194E+01, 3.0161E+01, 3.1139E+01, 3.2127E+01, 3.3125E+01,
     + 3.4134E+01, 3.5153E+01, 3.6182E+01, 3.7221E+01, 3.8270E+01,
     +  3.9328E+01, 4.0396E+01, 4.1474E+01, 4.2561E+01, 4.3658E+01,
     + 4.4764E+01, 4.5880E+01, 4.7004E+01, 4.8138E+01, 4.9281E+01,
     +  5.0432E+01, 5.1593E+01, 5.2763E+01, 5.3941E+01, 5.5128E+01,
     + 5.6324E+01, 5.7528E+01, 5.8741E+01, 5.9963E+01, 6.1193E+01,
     +  6.2431E+01, 6.3678E+01, 6.4932E+01, 6.6196E+01, 6.7467E+01,
     + 6.8746E+01, 7.0034E+01, 7.1330E+01, 7.2633E+01, 7.3945E+01,
     +  7.5264E+01, 7.6591E+01, 7.7926E+01, 7.9269E+01, 8.0620E+01,
     + 8.1978E+01, 8.3344E+01, 8.4718E+01, 8.6099E+01, 8.7487E+01,
     +  8.8883E+01, 9.0287E+01, 9.1698E+01, 9.3116E+01, 9.4542E+01,
     + 9.5975E+01, 9.7415E+01, 9.8863E+01, 1.0032E+02, 1.0178E+02,
     +  1.0325E+02, 1.0472E+02, 1.0621E+02, 1.0770E+02, 1.0920E+02,
     + 1.1070E+02, 1.1221E+02, 1.1373E+02, 1.1525E+02, 1.1679E+02,
     +  1.1833E+02, 1.1987E+02, 1.2142E+02, 1.2298E+02, 1.2455E+02,
     + 1.2612E+02, 1.2770E+02, 1.2929E+02, 1.3088E+02, 1.3248E+02,
     +  1.3409E+02, 1.3570E+02, 1.3732E+02, 1.3895E+02, 1.4058E+02,
     + 1.4222E+02, 1.4386E+02, 1.4552E+02, 1.4717E+02, 1.4884E+02,
     +  1.5051E+02, 1.5219E+02, 1.5387E+02, 1.5556E+02, 1.5726E+02,
     + 1.5896E+02, 1.6067E+02, 1.6239E+02, 1.6411E+02, 1.6584E+02,
     +  1.6758E+02, 1.6932E+02, 1.7107E+02, 1.7282E+02, 1.7458E+02,
     + 1.7635E+02, 1.7812E+02, 1.7990E+02, 1.8168E+02, 1.8348E+02,
     +  1.8527E+02, 1.8708E+02, 1.8889E+02, 1.9070E+02, 1.9253E+02,
     + 1.9435E+02, 1.9619E+02, 1.9803E+02, 1.9988E+02, 2.0173E+02,
     +  2.0359E+02, 2.0545E+02, 2.0733E+02, 2.0920E+02, 2.1109E+02,
     + 2.1298E+02, 2.1487E+02, 2.1677E+02, 2.1868E+02, 2.2060E+02,
     +  2.2252E+02, 2.2444E+02, 2.2637E+02, 2.2831E+02, 2.3026E+02,
     + 2.3221E+02, 2.3416E+02, 2.3613E+02, 2.3810E+02, 2.4007E+02,
     +  2.4205E+02, 2.4404E+02, 2.4603E+02, 2.4803E+02, 2.5003E+02,
     + 2.5205E+02, 2.5406E+02, 2.5609E+02, 2.5812E+02, 2.6015E+02,
     +  2.6219E+02, 2.6424E+02, 2.6629E+02, 2.6835E+02, 2.7042E+02,
     + 2.7249E+02, 2.7457E+02, 2.7665E+02, 2.7874E+02, 2.8084E+02,
     +  2.8294E+02, 2.8505E+02, 2.8716E+02, 2.8929E+02, 2.9141E+02,
     + 2.9354E+02, 2.9568E+02, 2.9783E+02, 2.9998E+02, 3.0214E+02,
     +  3.0430E+02, 3.0647E+02, 3.0865E+02, 3.1083E+02, 3.1302E+02,
     + 3.1521E+02, 3.1741E+02, 3.1962E+02, 3.2183E+02, 3.2405E+02,
     +  3.2627E+02, 3.2851E+02, 3.3074E+02, 3.3299E+02, 3.3524E+02,
     + 3.3749E+02, 3.3976E+02, 3.4203E+02, 3.4430E+02, 3.4658E+02,
     +  3.4887E+02, 3.5116E+02, 3.5346E+02, 3.5577E+02, 3.5808E+02,
     + 3.6040E+02, 3.6273E+02, 3.6506E+02, 3.6740E+02, 3.6974E+02,
     +  3.7209E+02, 3.7445E+02, 3.7681E+02, 3.7918E+02, 3.8156E+02,
     + 3.8394E+02, 3.8633E+02, 3.8872E+02, 3.9113E+02, 3.9353E+02,
     +  3.9595E+02, 3.9837E+02, 4.0080E+02, 4.0323E+02, 4.0567E+02,
     + 4.0812E+02, 4.1057E+02, 4.1303E+02, 4.1550E+02, 4.1797E+02,
     +  4.2045E+02, 4.2294E+02, 4.2543E+02, 4.2793E+02, 4.3043E+02,
     + 4.3295E+02, 4.3546E+02, 4.3799E+02, 4.4052E+02, 4.4306E+02,
     +  4.4561E+02, 4.4816E+02, 4.5072E+02, 4.5328E+02, 4.5585E+02,
     + 4.5843E+02, 4.6102E+02, 4.6361E+02, 4.6621E+02, 4.6881E+02,
     +  4.7143E+02, 4.7405E+02, 4.7667E+02, 4.7930E+02, 4.8194E+02,
     + 4.8459E+02, 4.8724E+02, 4.8990E+02, 4.9257E+02, 4.9524E+02,
     +  4.9792E+02, 5.0061E+02, 5.0331E+02, 5.0601E+02, 5.0871E+02,
     + 5.1143E+02, 5.1415E+02, 5.1688E+02, 5.1962E+02, 5.2236E+02,
     +  5.2511E+02, 5.2787E+02, 5.3063E+02, 5.3340E+02, 5.3618E+02,
     + 5.3896E+02, 5.4175E+02, 5.4455E+02, 5.4736E+02, 5.5017E+02,
     +  5.5299E+02, 5.5582E+02, 5.5866E+02, 5.6150E+02, 5.6435E+02,
     + 5.6720E+02, 5.7007E+02, 5.7294E+02, 5.7581E+02, 5.7870E+02,
     +  5.8159E+02, 5.8449E+02, 5.8740E+02, 5.9031E+02, 5.9323E+02,
     + 5.9616E+02, 5.9910E+02, 6.0204E+02, 6.0499E+02, 6.0795E+02,
     +  6.1091E+02, 6.1389E+02, 6.1687E+02, 6.1986E+02, 6.2285E+02,
     + 6.2585E+02, 6.2886E+02, 6.3188E+02, 6.3491E+02, 6.3794E+02,
     +  6.4098E+02, 6.4403E+02, 6.4708E+02, 6.5014E+02, 6.5322E+02,
     + 6.5629E+02, 6.5938E+02, 6.6247E+02, 6.6557E+02, 6.6868E+02,
     +  6.7180E+02, 6.7492E+02, 6.7805E+02, 6.8119E+02, 6.8434E+02,
     + 6.8749E+02, 6.9066E+02, 6.9383E+02, 6.9701E+02, 7.0019E+02,
     +  7.0339E+02, 7.0659E+02, 7.0980E+02, 7.1301E+02, 7.1624E+02,
     + 7.1947E+02, 7.2271E+02, 7.2596E+02, 7.2922E+02, 7.3249E+02,
     +  7.3576E+02, 7.3904E+02, 7.4233E+02, 7.4563E+02, 7.4893E+02,
     + 7.5224E+02, 7.5557E+02, 7.5890E+02, 7.6223E+02, 7.6558E+02,
     +  7.6893E+02, 7.7229E+02, 7.7566E+02, 7.7904E+02, 7.8243E+02,
     + 7.8582E+02, 7.8923E+02, 7.9264E+02, 7.9606E+02, 7.9949E+02,
     +  8.0292E+02, 8.0637E+02, 8.0982E+02, 8.1328E+02, 8.1675E+02,
     + 8.2023E+02, 8.2372E+02, 8.2721E+02, 8.3071E+02, 8.3423E+02,
     +  8.3775E+02, 8.4128E+02, 8.4481E+02, 8.4836E+02, 8.5191E+02,
     + 8.5548E+02, 8.5905E+02, 8.6263E+02, 8.6622E+02, 8.6982E+02,
     +  8.7342E+02, 8.7704E+02, 8.8066E+02, 8.8429E+02, 8.8793E+02,
     + 8.9158E+02, 8.9524E+02, 8.9891E+02, 9.0259E+02, 9.0627E+02,
     +  9.0997E+02, 9.1367E+02, 9.1738E+02, 9.2110E+02, 9.2483E+02,
     + 9.2857E+02, 9.3232E+02, 9.3607E+02, 9.3984E+02, 9.4361E+02,
     +  9.4740E+02, 9.5119E+02, 9.5499E+02, 9.5880E+02, 9.6262E+02,
     + 9.6645E+02, 9.7029E+02, 9.7414E+02, 9.7799E+02, 9.8186E+02,
     +  9.8573E+02, 9.8962E+02, 9.9351E+02, 9.9742E+02, 1.0013E+03,
     + 1.0052E+03, 1.0092E+03, 1.0131E+03, 1.0171E+03, 1.0210E+03,
     +  1.0250E+03, 1.0290E+03, 1.0330E+03, 1.0370E+03, 1.0410E+03,
     + 1.0450E+03, 1.0490E+03, 1.0530E+03, 1.0571E+03, 1.0611E+03,
     +  1.0652E+03, 1.0693E+03, 1.0734E+03, 1.0774E+03, 1.0816E+03,
     + 1.0857E+03, 1.0898E+03, 1.0939E+03, 1.0981E+03, 1.1022E+03,
     +  1.1064E+03, 1.1105E+03, 1.1147E+03, 1.1189E+03, 1.1231E+03,
     + 1.1273E+03, 1.1315E+03, 1.1358E+03, 1.1400E+03, 1.1443E+03,
     +  1.1485E+03, 1.1528E+03, 1.1571E+03, 1.1614E+03, 1.1657E+03,
     + 1.1700E+03, 1.1743E+03, 1.1786E+03, 1.1830E+03, 1.1873E+03,
     +  1.1917E+03, 1.1961E+03, 1.2004E+03, 1.2048E+03, 1.2092E+03,
     + 1.2136E+03, 1.2181E+03, 1.2225E+03, 1.2269E+03, 1.2314E+03,
     +  1.2359E+03, 1.2403E+03, 1.2448E+03, 1.2493E+03, 1.2538E+03,
     + 1.2584E+03, 1.2629E+03, 1.2674E+03, 1.2720E+03, 1.2765E+03,
     +  1.2811E+03, 1.2857E+03, 1.2903E+03, 1.2949E+03, 1.2995E+03,
     + 1.3041E+03, 1.3088E+03, 1.3134E+03, 1.3181E+03, 1.3227E+03,
     +  1.3274E+03, 1.3321E+03, 1.3368E+03, 1.3415E+03, 1.3462E+03,
     + 1.3510E+03, 1.3557E+03, 1.3605E+03, 1.3652E+03, 1.3700E+03,
     +  1.3748E+03, 1.3796E+03, 1.3844E+03, 1.3892E+03, 1.3941E+03,
     + 1.3989E+03, 1.4038E+03, 1.4086E+03, 1.4135E+03, 1.4184E+03,
     +  1.4233E+03, 1.4282E+03, 1.4331E+03, 1.4381E+03, 1.4430E+03,
     + 1.4480E+03, 1.4530E+03, 1.4579E+03, 1.4629E+03, 1.4679E+03,
     +  1.4730E+03, 1.4780E+03, 1.4830E+03, 1.4881E+03, 1.4931E+03,
     + 1.4982E+03, 1.5033E+03, 1.5084E+03, 1.5135E+03, 1.5186E+03,
     +  1.5238E+03, 1.5289E+03, 1.5341E+03, 1.5392E+03, 1.5444E+03,
     + 1.5496E+03, 1.5548E+03, 1.5600E+03, 1.5653E+03, 1.5705E+03,
     +  1.5757E+03, 1.5810E+03, 1.5863E+03, 1.5916E+03, 1.5969E+03,
     + 1.6022E+03, 1.6075E+03, 1.6129E+03, 1.6182E+03, 1.6236E+03,
     +  1.6289E+03, 1.6343E+03, 1.6397E+03, 1.6451E+03, 1.6506E+03,
     + 1.6560E+03, 1.6615E+03, 1.6669E+03, 1.6724E+03, 1.6779E+03,
     +  1.6834E+03, 1.6889E+03, 1.6944E+03, 1.7000E+03, 1.7055E+03,
     + 1.7111E+03, 1.7166E+03, 1.7222E+03, 1.7278E+03, 1.7334E+03,
     +  1.7391E+03, 1.7447E+03, 1.7504E+03, 1.7560E+03, 1.7617E+03,
     + 1.7674E+03, 1.7731E+03, 1.7788E+03, 1.7846E+03, 1.7903E+03,
     +  1.7961E+03, 1.8018E+03, 1.8076E+03, 1.8134E+03, 1.8192E+03,
     + 1.8250E+03, 1.8309E+03, 1.8367E+03, 1.8426E+03, 1.8484E+03,
     +  1.8543E+03, 1.8602E+03, 1.8662E+03, 1.8721E+03, 1.8780E+03,
     + 1.8840E+03, 1.8899E+03, 1.8959E+03, 1.9019E+03, 1.9079E+03,
     +  1.9140E+03, 1.9200E+03, 1.9260E+03, 1.9321E+03, 1.9382E+03,
     + 1.9443E+03, 1.9504E+03, 1.9565E+03, 1.9626E+03, 1.9688E+03,
     +  1.9749E+03, 1.9811E+03, 1.9873E+03, 1.9935E+03, 1.9997E+03,
     + 2.0059E+03, 2.0122E+03, 2.0184E+03, 2.0247E+03, 2.0310E+03,
     +  2.0373E+03, 2.0436E+03, 2.0499E+03, 2.0563E+03, 2.0626E+03,
     + 2.0690E+03, 2.0754E+03, 2.0818E+03, 2.0882E+03, 2.0946E+03,
     +  2.1011E+03, 2.1075E+03, 2.1140E+03, 2.1205E+03, 2.1270E+03,
     + 2.1335E+03, 2.1401E+03, 2.1466E+03, 2.1532E+03, 2.1597E+03,
     +  2.1663E+03, 2.1729E+03, 2.1795E+03, 2.1862E+03, 2.1928E+03,
     + 2.1995E+03, 2.2062E+03, 2.2128E+03, 2.2196E+03, 2.2263E+03,
     +  2.2330E+03, 2.2398E+03, 2.2465E+03, 2.2533E+03, 2.2601E+03,
     + 2.2669E+03, 2.2737E+03, 2.2806E+03, 2.2874E+03, 2.2943E+03,
     +  2.3012E+03, 2.3081E+03, 2.3150E+03, 2.3220E+03, 2.3289E+03,
     + 2.3359E+03, 2.3428E+03, 2.3498E+03, 2.3569E+03, 2.3639E+03,
     +  2.3709E+03, 2.3780E+03, 2.3851E+03, 2.3921E+03, 2.3992E+03,
     + 2.4064E+03, 2.4135E+03, 2.4206E+03, 2.4278E+03, 2.4350E+03,
     +  2.4422E+03, 2.4494E+03, 2.4566E+03, 2.4639E+03, 2.4711E+03,
     + 2.4784E+03, 2.4857E+03, 2.4930E+03, 2.5004E+03, 2.5077E+03,
     +  2.5151E+03, 2.5224E+03, 2.5298E+03, 2.5372E+03, 2.5446E+03,
     + 2.5521E+03, 2.5595E+03, 2.5670E+03, 2.5745E+03, 2.5820E+03,
     +  2.5895E+03, 2.5971E+03, 2.6046E+03, 2.6122E+03, 2.6198E+03,
     + 2.6274E+03, 2.6350E+03, 2.6426E+03, 2.6503E+03, 2.6579E+03,
     +  2.6656E+03, 2.6733E+03, 2.6811E+03, 2.6888E+03, 2.6965E+03,
     + 2.7043E+03, 2.7121E+03, 2.7199E+03, 2.7277E+03, 2.7356E+03,
     +  2.7434E+03, 2.7513E+03, 2.7592E+03, 2.7671E+03, 2.7750E+03,
     + 2.7829E+03, 2.7909E+03, 2.7989E+03, 2.8069E+03, 2.8149E+03,
     +  2.8229E+03, 2.8309E+03, 2.8390E+03, 2.8471E+03, 2.8552E+03,
     + 2.8633E+03, 2.8714E+03, 2.8796E+03, 2.8877E+03, 2.8959E+03,
     +  2.9041E+03, 2.9123E+03, 2.9206E+03, 2.9288E+03, 2.9371E+03,
     + 2.9454E+03, 2.9537E+03, 2.9620E+03, 2.9704E+03, 2.9787E+03,
     +  2.9871E+03, 2.9955E+03, 3.0039E+03, 3.0124E+03, 3.0208E+03,
     + 3.0293E+03, 3.0378E+03, 3.0463E+03, 3.0548E+03, 3.0633E+03,
     +  3.0719E+03, 3.0805E+03, 3.0891E+03, 3.0977E+03, 3.1063E+03,
     + 3.1150E+03, 3.1236E+03, 3.1323E+03, 3.1410E+03, 3.1498E+03,
     +  3.1585E+03, 3.1673E+03, 3.1760E+03, 3.1848E+03, 3.1937E+03,
     + 3.2025E+03, 3.2114E+03, 3.2202E+03, 3.2291E+03, 3.2380E+03,
     +  3.2470E+03, 3.2559E+03, 3.2649E+03, 3.2739E+03, 3.2829E+03,
     + 3.2919E+03, 3.3010E+03, 3.3100E+03, 3.3191E+03, 3.3282E+03,
     +  3.3373E+03, 3.3465E+03, 3.3556E+03, 3.3648E+03, 3.3740E+03,
     + 3.3832E+03, 3.3925E+03, 3.4017E+03, 3.4110E+03, 3.4203E+03,
     +  3.4296E+03, 3.4390E+03, 3.4483E+03, 3.4577E+03, 3.4671E+03,
     + 3.4765E+03, 3.4859E+03, 3.4954E+03, 3.5049E+03, 3.5144E+03,
     +  3.5239E+03, 3.5334E+03, 3.5430E+03, 3.5525E+03, 3.5621E+03,
     + 3.5717E+03, 3.5814E+03, 3.5910E+03, 3.6007E+03, 3.6104E+03,
     +  3.6201E+03, 3.6298E+03, 3.6396E+03, 3.6494E+03, 3.6592E+03,
     + 3.6690E+03, 3.6788E+03, 3.6887E+03, 3.6986E+03, 3.7085E+03,
     +  3.7184E+03, 3.7283E+03, 3.7383E+03, 3.7483E+03, 3.7583E+03,
     + 3.7683E+03, 3.7783E+03, 3.7884E+03, 3.7985E+03, 3.8086E+03,
     +  3.8187E+03, 3.8288E+03, 3.8390E+03, 3.8492E+03, 3.8594E+03,
     + 3.8696E+03, 3.8799E+03, 3.8902E+03, 3.9005E+03, 3.9108E+03,
     +  3.9211E+03, 3.9315E+03, 3.9419E+03, 3.9523E+03, 3.9627E+03,
     + 3.9731E+03, 3.9836E+03, 3.9941E+03, 4.0046E+03, 4.0151E+03,
     +  4.0257E+03, 4.0362E+03, 4.0468E+03, 4.0575E+03, 4.0681E+03,
     + 4.0788E+03, 4.0894E+03, 4.1001E+03, 4.1109E+03, 4.1216E+03,
     +  4.1324E+03, 4.1432E+03, 4.1540E+03, 4.1648E+03, 4.1757E+03,
     + 4.1866E+03, 4.1975E+03, 4.2084E+03, 4.2193E+03, 4.2303E+03,
     +  4.2413E+03, 4.2523E+03, 4.2634E+03, 4.2744E+03, 4.2855E+03,
     + 4.2966E+03, 4.3077E+03, 4.3189E+03, 4.3300E+03, 4.3412E+03,
     +  4.3525E+03, 4.3637E+03, 4.3750E+03, 4.3862E+03, 4.3976E+03,
     + 4.4089E+03, 4.4202E+03, 4.4316E+03, 4.4430E+03, 4.4544E+03,
     +  4.4659E+03, 4.4774E+03, 4.4888E+03, 4.5004E+03, 4.5119E+03,
     + 4.5235E+03, 4.5350E+03, 4.5467E+03, 4.5583E+03, 4.5699E+03,
     +  4.5816E+03, 4.5933E+03, 4.6051E+03, 4.6168E+03, 4.6286E+03,
     + 4.6404E+03, 4.6522E+03, 4.6640E+03, 4.6759E+03, 4.6878E+03,
     +  4.6997E+03, 4.7117E+03, 4.7236E+03, 4.7356E+03, 4.7476E+03,
     + 4.7596E+03, 4.7717E+03, 4.7838E+03, 4.7959E+03, 4.8080E+03,
     +  4.8202E+03, 4.8324E+03, 4.8446E+03, 4.8568E+03, 4.8691E+03,
     + 4.8813E+03, 4.8936E+03, 4.9060E+03, 4.9183E+03, 4.9307E+03,
     +  4.9431E+03, 4.9555E+03, 4.9680E+03, 4.9804E+03, 4.9929E+03,
     + 5.0055E+03, 5.0180E+03, 5.0306E+03, 5.0432E+03, 5.0558E+03,
     +  5.0684E+03, 5.0811E+03, 5.0938E+03, 5.1065E+03, 5.1193E+03,
     + 5.1321E+03, 5.1449E+03, 5.1577E+03, 5.1705E+03, 5.1834E+03,
     +  5.1963E+03, 5.2092E+03, 5.2222E+03, 5.2352E+03, 5.2482E+03,
     + 5.2612E+03, 5.2742E+03, 5.2873E+03, 5.3004E+03, 5.3136E+03,
     +  5.3267E+03, 5.3399E+03, 5.3531E+03, 5.3663E+03, 5.3796E+03,
     + 5.3929E+03, 5.4062E+03, 5.4195E+03, 5.4329E+03, 5.4463E+03,
     +  5.4597E+03, 5.4731E+03, 5.4866E+03, 5.5001E+03, 5.5136E+03,
     + 5.5272E+03, 5.5408E+03, 5.5544E+03, 5.5680E+03, 5.5816E+03,
     +  5.5953E+03, 5.6090E+03, 5.6228E+03, 5.6365E+03, 5.6503E+03,
     + 5.6641E+03, 5.6780E+03, 5.6918E+03, 5.7057E+03, 5.7196E+03,
     +  5.7336E+03, 5.7476E+03, 5.7616E+03, 5.7756E+03, 5.7896E+03,
     + 5.8037E+03, 5.8178E+03, 5.8320E+03, 5.8461E+03, 5.8603E+03,
     +  5.8746E+03, 5.8888E+03, 5.9031E+03, 5.9174E+03, 5.9317E+03,
     + 5.9461E+03, 5.9604E+03, 5.9749E+03, 5.9893E+03, 6.0038E+03,
     +  6.0183E+03, 6.0328E+03, 6.0473E+03, 6.0619E+03, 6.0765E+03,
     + 6.0912E+03, 6.1058E+03, 6.1205E+03, 6.1352E+03, 6.1500E+03,
     +  6.1648E+03, 6.1796E+03, 6.1944E+03, 6.2093E+03, 6.2242E+03,
     + 6.2391E+03, 6.2540E+03, 6.2690E+03, 6.2840E+03, 6.2990E+03,
     +  6.3141E+03, 6.3292E+03, 6.3443E+03, 6.3595E+03, 6.3746E+03,
     + 6.3898E+03, 6.4051E+03, 6.4203E+03, 6.4356E+03, 6.4509E+03,
     +  6.4663E+03, 6.4817E+03, 6.4971E+03, 6.5125E+03, 6.5280E+03,
     + 6.5435E+03, 6.5590E+03, 6.5746E+03, 6.5901E+03, 6.6058E+03,
     +  6.6214E+03, 6.6371E+03, 6.6528E+03, 6.6685E+03, 6.6843E+03,
     + 6.7001E+03, 6.7159E+03, 6.7317E+03, 6.7476E+03, 6.7635E+03,
     +  6.7795E+03, 6.7954E+03, 6.8114E+03, 6.8275E+03, 6.8435E+03,
     + 6.8596E+03, 6.8758E+03, 6.8919E+03, 6.9081E+03, 6.9243E+03,
     +  6.9405E+03, 6.9568E+03, 6.9731E+03, 6.9895E+03, 7.0058E+03,
     + 7.0222E+03, 7.0386E+03, 7.0551E+03, 7.0716E+03, 7.0881E+03,
     +  7.1047E+03, 7.1212E+03, 7.1379E+03, 7.1545E+03, 7.1712E+03,
     + 7.1879E+03, 7.2046E+03, 7.2214E+03, 7.2382E+03, 7.2550E+03,
     +  7.2719E+03, 7.2888E+03, 7.3057E+03, 7.3226E+03, 7.3396E+03,
     + 7.3566E+03, 7.3737E+03, 7.3908E+03, 7.4079E+03, 7.4250E+03,
     +  7.4422E+03, 7.4594E+03, 7.4767E+03, 7.4939E+03, 7.5112E+03,
     + 7.5286E+03, 7.5459E+03, 7.5633E+03, 7.5808E+03, 7.5982E+03,
     +  7.6157E+03, 7.6333E+03, 7.6508E+03, 7.6684E+03, 7.6860E+03,
     + 7.7037E+03, 7.7214E+03, 7.7391E+03, 7.7569E+03, 7.7747E+03,
     +  7.7925E+03, 7.8103E+03, 7.8282E+03, 7.8461E+03, 7.8641E+03,
     + 7.8821E+03, 7.9001E+03, 7.9182E+03, 7.9362E+03, 7.9544E+03,
     +  7.9725E+03, 7.9907E+03, 8.0089E+03, 8.0272E+03, 8.0454E+03,
     + 8.0638E+03, 8.0821E+03, 8.1005E+03, 8.1189E+03, 8.1374E+03,
     +  8.1558E+03, 8.1744E+03, 8.1929E+03, 8.2115E+03, 8.2301E+03,
     + 8.2488E+03, 8.2675E+03, 8.2862E+03, 8.3049E+03, 8.3237E+03,
     +  8.3426E+03, 8.3614E+03, 8.3803E+03, 8.3992E+03, 8.4182E+03,
     + 8.4372E+03, 8.4562E+03, 8.4753E+03, 8.4944E+03, 8.5135E+03,
     +  8.5327E+03, 8.5519E+03, 8.5711E+03, 8.5904E+03, 8.6097E+03,
     + 8.6290E+03, 8.6484E+03, 8.6678E+03, 8.6873E+03, 8.7068E+03,
     +  8.7263E+03, 8.7458E+03, 8.7654E+03, 8.7850E+03, 8.8047E+03,
     + 8.8244E+03, 8.8441E+03, 8.8639E+03, 8.8837E+03, 8.9035E+03,
     +  8.9234E+03, 8.9433E+03, 8.9632E+03, 8.9832E+03, 9.0032E+03,
     + 9.0233E+03, 9.0433E+03, 9.0635E+03, 9.0836E+03, 9.1038E+03,
     +  9.1240E+03, 9.1443E+03, 9.1646E+03, 9.1849E+03, 9.2053E+03,
     + 9.2257E+03, 9.2462E+03, 9.2667E+03, 9.2872E+03, 9.3077E+03,
     +  9.3283E+03, 9.3489E+03, 9.3696E+03, 9.3903E+03, 9.4111E+03,
     + 9.4318E+03, 9.4526E+03, 9.4735E+03, 9.4944E+03, 9.5153E+03,
     +  9.5363E+03, 9.5573E+03, 9.5783E+03, 9.5994E+03, 9.6205E+03,
     + 9.6416E+03, 9.6628E+03, 9.6840E+03, 9.7053E+03, 9.7266E+03,
     +  9.7479E+03, 9.7693E+03, 9.7907E+03, 9.8121E+03, 9.8336E+03,
     + 9.8551E+03, 9.8767E+03, 9.8983E+03, 9.9199E+03, 9.9416E+03,
     +  9.9633E+03, 9.9851E+03, 1.0007E+04, 1.0029E+04, 1.0051E+04,
     + 1.0072E+04, 1.0094E+04, 1.0116E+04, 1.0138E+04, 1.0160E+04,
     +  1.0183E+04, 1.0205E+04, 1.0227E+04, 1.0249E+04, 1.0271E+04,
     + 1.0294E+04, 1.0316E+04, 1.0338E+04, 1.0361E+04, 1.0383E+04,
     +  1.0406E+04, 1.0428E+04, 1.0451E+04, 1.0473E+04, 1.0496E+04,
     + 1.0519E+04, 1.0541E+04, 1.0564E+04, 1.0587E+04, 1.0610E+04,
     +  1.0633E+04, 1.0656E+04, 1.0679E+04, 1.0702E+04, 1.0725E+04,
     + 1.0748E+04, 1.0771E+04, 1.0794E+04, 1.0817E+04, 1.0840E+04,
     +  1.0864E+04, 1.0887E+04, 1.0910E+04, 1.0934E+04, 1.0957E+04,
     + 1.0981E+04, 1.1004E+04, 1.1028E+04, 1.1051E+04, 1.1075E+04,
     +  1.1099E+04, 1.1122E+04, 1.1146E+04, 1.1170E+04, 1.1194E+04,
     + 1.1218E+04, 1.1242E+04, 1.1266E+04, 1.1290E+04, 1.1314E+04,
     +  1.1338E+04, 1.1362E+04, 1.1386E+04, 1.1410E+04, 1.1434E+04,
     + 1.1459E+04, 1.1483E+04, 1.1508E+04, 1.1532E+04, 1.1556E+04,
     +  1.1581E+04, 1.1605E+04, 1.1630E+04, 1.1655E+04, 1.1679E+04,
     + 1.1704E+04, 1.1729E+04, 1.1754E+04, 1.1779E+04, 1.1803E+04,
     +  1.1828E+04, 1.1853E+04, 1.1878E+04, 1.1903E+04, 1.1928E+04,
     + 1.1954E+04, 1.1979E+04, 1.2004E+04, 1.2029E+04, 1.2055E+04,
     +  1.2080E+04, 1.2105E+04, 1.2131E+04, 1.2156E+04, 1.2182E+04,
     + 1.2207E+04, 1.2233E+04, 1.2259E+04, 1.2284E+04, 1.2310E+04,
     +  1.2336E+04, 1.2362E+04, 1.2388E+04, 1.2414E+04, 1.2439E+04,
     + 1.2465E+04, 1.2492E+04, 1.2518E+04, 1.2544E+04, 1.2570E+04,
     +  1.2596E+04, 1.2622E+04, 1.2649E+04, 1.2675E+04, 1.2702E+04,
     + 1.2728E+04, 1.2754E+04, 1.2781E+04, 1.2808E+04, 1.2834E+04,
     +  1.2861E+04, 1.2888E+04, 1.2914E+04, 1.2941E+04, 1.2968E+04,
     + 1.2995E+04, 1.3022E+04, 1.3049E+04, 1.3076E+04, 1.3103E+04,
     +  1.3130E+04, 1.3157E+04, 1.3184E+04, 1.3212E+04, 1.3239E+04,
     + 1.3266E+04, 1.3294E+04, 1.3321E+04, 1.3349E+04, 1.3376E+04,
     +  1.3404E+04, 1.3431E+04, 1.3459E+04, 1.3487E+04, 1.3514E+04,
     + 1.3542E+04, 1.3570E+04, 1.3598E+04, 1.3626E+04, 1.3654E+04,
     +  1.3682E+04, 1.3710E+04, 1.3738E+04, 1.3766E+04, 1.3795E+04,
     + 1.3823E+04, 1.3851E+04, 1.3880E+04, 1.3908E+04, 1.3936E+04,
     +  1.3965E+04, 1.3993E+04, 1.4022E+04, 1.4051E+04, 1.4079E+04,
     + 1.4108E+04, 1.4137E+04, 1.4166E+04, 1.4195E+04, 1.4223E+04,
     +  1.4252E+04, 1.4281E+04, 1.4311E+04, 1.4340E+04, 1.4369E+04,
     + 1.4398E+04, 1.4427E+04, 1.4457E+04, 1.4486E+04, 1.4515E+04,
     +  1.4545E+04, 1.4574E+04, 1.4604E+04, 1.4633E+04, 1.4663E+04,
     + 1.4693E+04, 1.4723E+04, 1.4752E+04, 1.4782E+04, 1.4812E+04,
     +  1.4842E+04, 1.4872E+04, 1.4902E+04, 1.4932E+04, 1.4962E+04,
     + 1.4992E+04, 1.5023E+04, 1.5053E+04, 1.5083E+04, 1.5114E+04,
     +  1.5144E+04, 1.5174E+04, 1.5205E+04, 1.5236E+04, 1.5266E+04,
     + 1.5297E+04, 1.5328E+04, 1.5358E+04, 1.5389E+04, 1.5420E+04,
     +  1.5451E+04, 1.5482E+04, 1.5513E+04, 1.5544E+04, 1.5575E+04,
     + 1.5606E+04, 1.5638E+04, 1.5669E+04, 1.5700E+04, 1.5731E+04,
     +  1.5763E+04, 1.5794E+04, 1.5826E+04, 1.5857E+04, 1.5889E+04,
     + 1.5921E+04, 1.5953E+04, 1.5984E+04, 1.6016E+04, 1.6048E+04,
     +  1.6080E+04, 1.6112E+04, 1.6144E+04, 1.6176E+04, 1.6208E+04,
     + 1.6240E+04, 1.6273E+04, 1.6305E+04, 1.6337E+04, 1.6370E+04,
     +  1.6402E+04, 1.6435E+04, 1.6467E+04, 1.6500E+04, 1.6532E+04,
     + 1.6565E+04, 1.6598E+04, 1.6631E+04, 1.6663E+04, 1.6696E+04,
     +  1.6729E+04, 1.6762E+04, 1.6795E+04, 1.6829E+04, 1.6862E+04,
     + 1.6895E+04, 1.6928E+04, 1.6962E+04, 1.6995E+04, 1.7028E+04,
     +  1.7062E+04, 1.7095E+04, 1.7129E+04, 1.7163E+04, 1.7196E+04,
     + 1.7230E+04, 1.7264E+04, 1.7298E+04, 1.7332E+04, 1.7366E+04,
     +  1.7400E+04, 1.7434E+04, 1.7468E+04, 1.7502E+04, 1.7536E+04,
     + 1.7571E+04, 1.7605E+04, 1.7639E+04, 1.7674E+04, 1.7708E+04,
     +  1.7743E+04, 1.7778E+04, 1.7812E+04, 1.7847E+04, 1.7882E+04,
     + 1.7917E+04, 1.7951E+04, 1.7986E+04, 1.8021E+04, 1.8056E+04,
     +  1.8092E+04, 1.8127E+04, 1.8162E+04, 1.8197E+04, 1.8232E+04,
     + 1.8268E+04, 1.8303E+04, 1.8339E+04, 1.8374E+04, 1.8410E+04,
     +  1.8446E+04, 1.8481E+04, 1.8517E+04, 1.8553E+04, 1.8589E+04,
     + 1.8625E+04, 1.8661E+04, 1.8697E+04, 1.8733E+04, 1.8769E+04,
     +  1.8805E+04, 1.8842E+04, 1.8878E+04, 1.8914E+04, 1.8951E+04,
     + 1.8987E+04, 1.9024E+04, 1.9060E+04, 1.9097E+04, 1.9134E+04,
     +  1.9171E+04, 1.9207E+04, 1.9244E+04, 1.9281E+04, 1.9318E+04,
     + 1.9355E+04, 1.9392E+04, 1.9430E+04, 1.9467E+04, 1.9504E+04,
     +  1.9542E+04, 1.9579E+04, 1.9616E+04, 1.9654E+04, 1.9691E+04,
     + 1.9729E+04, 1.9767E+04, 1.9805E+04, 1.9842E+04, 1.9880E+04,
     +  1.9918E+04, 1.9956E+04, 1.9994E+04, 2.0032E+04, 2.0070E+04,
     + 2.0109E+04, 2.0147E+04, 2.0185E+04, 2.0224E+04, 2.0262E+04,
     +  2.0301E+04, 2.0339E+04, 2.0378E+04, 2.0417E+04, 2.0455E+04,
     + 2.0494E+04, 2.0533E+04, 2.0572E+04, 2.0611E+04, 2.0650E+04,
     +  2.0689E+04, 2.0728E+04, 2.0767E+04, 2.0807E+04, 2.0846E+04,
     + 2.0885E+04, 2.0925E+04, 2.0964E+04, 2.1004E+04, 2.1044E+04,
     +  2.1083E+04, 2.1123E+04, 2.1163E+04, 2.1203E+04, 2.1243E+04,
     + 2.1283E+04, 2.1323E+04, 2.1363E+04, 2.1403E+04, 2.1443E+04,
     +  2.1484E+04, 2.1524E+04, 2.1564E+04, 2.1605E+04, 2.1646E+04,
     + 2.1686E+04, 2.1727E+04, 2.1768E+04, 2.1808E+04, 2.1849E+04,
     +  2.1890E+04, 2.1931E+04, 2.1972E+04, 2.2013E+04, 2.2054E+04,
     + 2.2096E+04, 2.2137E+04, 2.2178E+04, 2.2220E+04, 2.2261E+04,
     +  2.2303E+04, 2.2344E+04, 2.2386E+04, 2.2428E+04, 2.2469E+04,
     + 2.2511E+04, 2.2553E+04, 2.2595E+04, 2.2637E+04, 2.2679E+04,
     +  2.2721E+04, 2.2764E+04, 2.2806E+04, 2.2848E+04, 2.2891E+04,
     + 2.2933E+04, 2.2976E+04, 2.3018E+04, 2.3061E+04, 2.3104E+04,
     +  2.3146E+04, 2.3189E+04, 2.3232E+04, 2.3275E+04, 2.3318E+04,
     + 2.3361E+04, 2.3404E+04, 2.3448E+04, 2.3491E+04, 2.3534E+04,
     +  2.3578E+04, 2.3621E+04, 2.3665E+04, 2.3708E+04, 2.3752E+04,
     + 2.3796E+04, 2.3839E+04, 2.3883E+04, 2.3927E+04, 2.3971E+04,
     +  2.4015E+04, 2.4059E+04, 2.4104E+04, 2.4148E+04, 2.4192E+04,
     + 2.4237E+04, 2.4281E+04, 2.4326E+04, 2.4370E+04, 2.4415E+04,
     +  2.4460E+04, 2.4504E+04, 2.4549E+04, 2.4594E+04, 2.4639E+04,
     + 2.4684E+04, 2.4729E+04, 2.4774E+04, 2.4820E+04, 2.4865E+04,
     +  2.4910E+04, 2.4956E+04, 2.5001E+04, 2.5047E+04, 2.5092E+04,
     + 2.5138E+04, 2.5184E+04, 2.5230E+04, 2.5276E+04, 2.5322E+04,
     +  2.5368E+04, 2.5414E+04, 2.5460E+04, 2.5506E+04, 2.5552E+04,
     + 2.5599E+04, 2.5645E+04, 2.5692E+04, 2.5738E+04, 2.5785E+04,
     +  2.5832E+04, 2.5878E+04, 2.5925E+04, 2.5972E+04, 2.6019E+04,
     + 2.6066E+04, 2.6113E+04, 2.6160E+04, 2.6208E+04, 2.6255E+04,
     +  2.6302E+04, 2.6350E+04, 2.6397E+04, 2.6445E+04, 2.6493E+04,
     + 2.6540E+04, 2.6588E+04, 2.6636E+04, 2.6684E+04, 2.6732E+04,
     +  2.6780E+04, 2.6828E+04, 2.6876E+04, 2.6925E+04, 2.6973E+04,
     + 2.7021E+04, 2.7070E+04, 2.7118E+04, 2.7167E+04, 2.7216E+04,
     +  2.7264E+04, 2.7313E+04, 2.7362E+04, 2.7411E+04, 2.7460E+04,
     + 2.7509E+04, 2.7558E+04, 2.7608E+04, 2.7657E+04, 2.7706E+04,
     +  2.7756E+04, 2.7805E+04, 2.7855E+04, 2.7905E+04, 2.7954E+04,
     + 2.8004E+04, 2.8054E+04, 2.8104E+04, 2.8154E+04, 2.8204E+04,
     +  2.8254E+04, 2.8304E+04, 2.8355E+04, 2.8405E+04, 2.8455E+04,
     + 2.8506E+04, 2.8557E+04, 2.8607E+04, 2.8658E+04, 2.8709E+04,
     +  2.8760E+04, 2.8811E+04, 2.8862E+04, 2.8913E+04, 2.8964E+04,
     + 2.9015E+04, 2.9066E+04, 2.9118E+04, 2.9169E+04, 2.9221E+04,
     +  2.9272E+04, 2.9324E+04, 2.9376E+04, 2.9427E+04, 2.9479E+04,
     + 2.9531E+04, 2.9583E+04, 2.9635E+04, 2.9688E+04, 2.9740E+04,
     +  2.9792E+04, 2.9844E+04, 2.9897E+04, 2.9949E+04, 3.0002E+04,
     + 3.0055E+04, 3.0108E+04, 3.0160E+04, 3.0213E+04, 3.0266E+04,
     +  3.0319E+04, 3.0372E+04, 3.0426E+04, 3.0479E+04, 3.0532E+04,
     + 3.0586E+04, 3.0639E+04, 3.0693E+04, 3.0746E+04, 3.0800E+04,
     +  3.0854E+04, 3.0908E+04, 3.0962E+04, 3.1016E+04, 3.1070E+04,
     + 3.1124E+04, 3.1178E+04, 3.1232E+04, 3.1287E+04, 3.1341E+04,
     +  3.1396E+04, 3.1450E+04, 3.1505E+04, 3.1560E+04, 3.1615E+04,
     + 3.1669E+04, 3.1724E+04, 3.1779E+04, 3.1835E+04, 3.1890E+04,
     +  3.1945E+04, 3.2000E+04, 3.2056E+04, 3.2111E+04, 3.2167E+04,
     + 3.2223E+04, 3.2278E+04, 3.2334E+04, 3.2390E+04, 3.2446E+04,
     +  3.2502E+04, 3.2558E+04, 3.2614E+04, 3.2671E+04, 3.2727E+04,
     + 3.2783E+04, 3.2840E+04, 3.2897E+04, 3.2953E+04, 3.3010E+04,
     +  3.3067E+04, 3.3124E+04, 3.3181E+04, 3.3238E+04, 3.3295E+04,
     + 3.3352E+04, 3.3409E+04, 3.3467E+04, 3.3524E+04, 3.3581E+04,
     +  3.3639E+04, 3.3697E+04, 3.3754E+04, 3.3812E+04, 3.3870E+04,
     + 3.3928E+04, 3.3986E+04, 3.4044E+04, 3.4102E+04, 3.4161E+04,
     +  3.4219E+04, 3.4278E+04, 3.4336E+04, 3.4395E+04, 3.4453E+04,
     + 3.4512E+04, 3.4571E+04, 3.4630E+04, 3.4689E+04, 3.4748E+04,
     +  3.4807E+04, 3.4866E+04, 3.4926E+04, 3.4985E+04, 3.5044E+04,
     + 3.5104E+04, 3.5164E+04, 3.5223E+04, 3.5283E+04, 3.5343E+04,
     +  3.5403E+04, 3.5463E+04, 3.5523E+04, 3.5583E+04, 3.5644E+04,
     + 3.5704E+04, 3.5764E+04, 3.5825E+04, 3.5885E+04, 3.5946E+04,
     +  3.6007E+04, 3.6068E+04, 3.6129E+04, 3.6189E+04, 3.6251E+04,
     + 3.6312E+04, 3.6373E+04, 3.6434E+04, 3.6496E+04, 3.6557E+04,
     +  3.6619E+04, 3.6680E+04, 3.6742E+04, 3.6804E+04, 3.6866E+04,
     + 3.6928E+04, 3.6990E+04, 3.7052E+04, 3.7114E+04, 3.7176E+04,
     +  3.7239E+04, 3.7301E+04, 3.7364E+04, 3.7426E+04, 3.7489E+04,
     + 3.7552E+04, 3.7615E+04, 3.7678E+04, 3.7741E+04, 3.7804E+04,
     +  3.7867E+04, 3.7930E+04, 3.7994E+04, 3.8057E+04, 3.8120E+04,
     + 3.8184E+04, 3.8248E+04, 3.8312E+04, 3.8375E+04, 3.8439E+04,
     +  3.8503E+04, 3.8567E+04, 3.8632E+04, 3.8696E+04, 3.8760E+04,
     + 3.8825E+04, 3.8889E+04, 3.8954E+04, 3.9018E+04, 3.9083E+04,
     +  3.9148E+04, 3.9213E+04, 3.9278E+04, 3.9343E+04, 3.9408E+04,
     + 3.9474E+04, 3.9539E+04, 3.9604E+04, 3.9670E+04, 3.9736E+04,
     +  3.9801E+04, 3.9867E+04, 3.9933E+04, 3.9999E+04, 4.0065E+04,
     + 4.0131E+04, 4.0197E+04, 4.0264E+04, 4.0330E+04, 4.0396E+04,
     +  4.0463E+04, 4.0530E+04, 4.0596E+04, 4.0663E+04, 4.0730E+04,
     + 4.0797E+04, 4.0864E+04, 4.0931E+04, 4.0998E+04, 4.1066E+04,
     +  4.1133E+04, 4.1201E+04, 4.1268E+04, 4.1336E+04, 4.1404E+04,
     + 4.1471E+04, 4.1539E+04, 4.1607E+04, 4.1675E+04, 4.1744E+04,
     +  4.1812E+04, 4.1880E+04, 4.1949E+04, 4.2017E+04, 4.2086E+04,
     + 4.2155E+04, 4.2223E+04, 4.2292E+04, 4.2361E+04, 4.2430E+04,
     +  4.2499E+04, 4.2569E+04, 4.2638E+04, 4.2707E+04, 4.2777E+04,
     + 4.2846E+04, 4.2916E+04, 4.2986E+04, 4.3056E+04, 4.3125E+04,
     +  4.3196E+04, 4.3266E+04, 4.3336E+04, 4.3406E+04, 4.3476E+04,
     + 4.3547E+04, 4.3617E+04, 4.3688E+04, 4.3759E+04, 4.3830E+04,
     +  4.3900E+04, 4.3971E+04, 4.4043E+04, 4.4114E+04, 4.4185E+04,
     + 4.4256E+04, 4.4328E+04, 4.4399E+04, 4.4471E+04, 4.4543E+04,
     +  4.4614E+04, 4.4686E+04, 4.4758E+04, 4.4830E+04, 4.4902E+04,
     + 4.4975E+04, 4.5047E+04, 4.5119E+04, 4.5192E+04, 4.5264E+04,
     +  4.5337E+04, 4.5410E+04, 4.5483E+04, 4.5556E+04, 4.5629E+04,
     + 4.5702E+04, 4.5775E+04, 4.5848E+04, 4.5922E+04, 4.5995E+04,
     +  4.6069E+04, 4.6143E+04, 4.6216E+04, 4.6290E+04, 4.6364E+04,
     + 4.6438E+04, 4.6512E+04, 4.6587E+04, 4.6661E+04, 4.6735E+04,
     +  4.6810E+04, 4.6884E+04, 4.6959E+04, 4.7034E+04, 4.7109E+04,
     + 4.7184E+04, 4.7259E+04, 4.7334E+04, 4.7409E+04, 4.7484E+04,
     +  4.7560E+04, 4.7635E+04, 4.7711E+04, 4.7787E+04, 4.7862E+04,
     + 4.7938E+04, 4.8014E+04, 4.8090E+04, 4.8167E+04, 4.8243E+04,
     +  4.8319E+04, 4.8396E+04, 4.8472E+04, 4.8549E+04, 4.8625E+04,
     + 4.8702E+04, 4.8779E+04, 4.8856E+04, 4.8933E+04, 4.9010E+04,
     +  4.9088E+04, 4.9165E+04, 4.9242E+04, 4.9320E+04, 4.9398E+04,
     + 4.9475E+04, 4.9553E+04, 4.9631E+04, 4.9709E+04, 4.9787E+04,
     +  4.9866E+04, 4.9944E+04, 5.0022E+04, 5.0101E+04, 5.0179E+04,
     + 5.0258E+04, 5.0337E+04, 5.0416E+04, 5.0495E+04, 5.0574E+04,
     +  5.0653E+04, 5.0732E+04, 5.0811E+04, 5.0891E+04, 5.0970E+04,
     + 5.1050E+04, 5.1130E+04, 5.1209E+04, 5.1289E+04, 5.1369E+04,
     +  5.1450E+04, 5.1530E+04, 5.1610E+04, 5.1690E+04, 5.1771E+04,
     + 5.1851E+04, 5.1932E+04, 5.2013E+04, 5.2094E+04, 5.2175E+04,
     +  5.2256E+04, 5.2337E+04, 5.2418E+04, 5.2500E+04, 5.2581E+04,
     + 5.2663E+04, 5.2744E+04, 5.2826E+04, 5.2908E+04, 5.2990E+04,
     +  5.3072E+04, 5.3154E+04, 5.3236E+04, 5.3318E+04, 5.3401E+04,
     + 5.3483E+04, 5.3566E+04, 5.3649E+04, 5.3731E+04, 5.3814E+04,
     +  5.3897E+04, 5.3980E+04, 5.4064E+04, 5.4147E+04, 5.4230E+04,
     + 5.4314E+04, 5.4397E+04, 5.4481E+04, 5.4565E+04, 5.4649E+04,
     +  5.4733E+04, 5.4817E+04, 5.4901E+04, 5.4985E+04, 5.5069E+04,
     + 5.5154E+04, 5.5238E+04, 5.5323E+04, 5.5408E+04, 5.5493E+04,
     +  5.5578E+04, 5.5663E+04, 5.5748E+04, 5.5833E+04, 5.5918E+04,
     + 5.6004E+04, 5.6089E+04, 5.6175E+04, 5.6261E+04, 5.6347E+04,
     +  5.6433E+04, 5.6519E+04, 5.6605E+04, 5.6691E+04, 5.6777E+04,
     + 5.6864E+04, 5.6950E+04, 5.7037E+04, 5.7124E+04, 5.7210E+04,
     +  5.7297E+04, 5.7384E+04, 5.7472E+04, 5.7559E+04, 5.7646E+04,
     + 5.7734E+04, 5.7821E+04, 5.7909E+04, 5.7996E+04, 5.8084E+04,
     +  5.8172E+04, 5.8260E+04, 5.8348E+04, 5.8437E+04, 5.8525E+04,
     + 5.8613E+04, 5.8702E+04, 5.8791E+04, 5.8879E+04, 5.8968E+04,
     +  5.9057E+04, 5.9146E+04, 5.9235E+04, 5.9325E+04, 5.9414E+04,
     + 5.9503E+04, 5.9593E+04, 5.9683E+04, 5.9772E+04, 5.9862E+04,
     +  5.9952E+04, 6.0042E+04, 6.0132E+04, 6.0223E+04, 6.0313E+04,
     + 6.0404E+04, 6.0494E+04, 6.0585E+04, 6.0676E+04, 6.0766E+04,
     +  6.0857E+04, 6.0948E+04, 6.1040E+04, 6.1131E+04, 6.1222E+04,
     + 6.1314E+04, 6.1405E+04, 6.1497E+04, 6.1589E+04, 6.1681E+04,
     +  6.1773E+04, 6.1865E+04, 6.1957E+04, 6.2049E+04, 6.2142E+04,
     + 6.2234E+04, 6.2327E+04, 6.2420E+04, 6.2513E+04, 6.2606E+04,
     +  6.2699E+04, 6.2792E+04, 6.2885E+04, 6.2978E+04, 6.3072E+04,
     + 6.3165E+04, 6.3259E+04, 6.3353E+04, 6.3447E+04, 6.3541E+04,
     +  6.3635E+04, 6.3729E+04, 6.3823E+04, 6.3918E+04, 6.4012E+04,
     + 6.4107E+04, 6.4201E+04, 6.4296E+04, 6.4391E+04, 6.4486E+04,
     +  6.4581E+04, 6.4676E+04, 6.4772E+04, 6.4867E+04, 6.4963E+04,
     + 6.5058E+04, 6.5154E+04, 6.5250E+04, 6.5346E+04, 6.5442E+04,
     +  6.5538E+04, 6.5634E+04, 6.5731E+04, 6.5827E+04, 6.5924E+04,
     + 6.6021E+04, 6.6117E+04, 6.6214E+04, 6.6311E+04, 6.6409E+04,
     +  6.6506E+04, 6.6603E+04, 6.6701E+04, 6.6798E+04, 6.6896E+04,
     + 6.6993E+04, 6.7091E+04, 6.7189E+04, 6.7287E+04, 6.7386E+04,
     +  6.7484E+04, 6.7582E+04, 6.7681E+04, 6.7779E+04, 6.7878E+04,
     + 6.7977E+04, 6.8076E+04, 6.8175E+04, 6.8274E+04, 6.8373E+04,
     +  6.8473E+04, 6.8572E+04, 6.8672E+04, 6.8771E+04, 6.8871E+04,
     + 6.8971E+04, 6.9071E+04, 6.9171E+04, 6.9271E+04, 6.9372E+04,
     +  6.9472E+04, 6.9573E+04, 6.9673E+04, 6.9774E+04, 6.9875E+04,
     + 6.9976E+04, 7.0077E+04, 7.0178E+04, 7.0280E+04, 7.0381E+04,
     +  7.0482E+04, 7.0584E+04, 7.0686E+04, 7.0788E+04, 7.0890E+04,
     + 7.0992E+04, 7.1094E+04, 7.1196E+04, 7.1299E+04, 7.1401E+04,
     +  7.1504E+04, 7.1606E+04, 7.1709E+04, 7.1812E+04, 7.1915E+04,
     + 7.2018E+04, 7.2121E+04, 7.2225E+04, 7.2328E+04, 7.2432E+04,
     +  7.2536E+04, 7.2639E+04, 7.2743E+04, 7.2847E+04, 7.2951E+04/
      data (QofT( 1,J),J=2409,4510)/ 7.3056E+04, 7.3160E+04,
     +  7.3265E+04, 7.3369E+04, 7.3474E+04, 7.3579E+04, 7.3683E+04,
     +  7.3788E+04,
     + 7.3894E+04, 7.3999E+04, 7.4104E+04, 7.4210E+04, 7.4315E+04,
     +  7.4421E+04, 7.4527E+04, 7.4633E+04, 7.4739E+04, 7.4845E+04,
     + 7.4951E+04, 7.5057E+04, 7.5164E+04, 7.5270E+04, 7.5377E+04,
     +  7.5484E+04, 7.5591E+04, 7.5698E+04, 7.5805E+04, 7.5912E+04,
     + 7.6019E+04, 7.6127E+04, 7.6234E+04, 7.6342E+04, 7.6450E+04,
     +  7.6558E+04, 7.6666E+04, 7.6774E+04, 7.6882E+04, 7.6990E+04,
     + 7.7099E+04, 7.7207E+04, 7.7316E+04, 7.7425E+04, 7.7534E+04,
     +  7.7643E+04, 7.7752E+04, 7.7861E+04, 7.7970E+04, 7.8080E+04,
     + 7.8189E+04, 7.8299E+04, 7.8409E+04, 7.8519E+04, 7.8629E+04,
     +  7.8739E+04, 7.8849E+04, 7.8959E+04, 7.9070E+04, 7.9180E+04,
     + 7.9291E+04, 7.9402E+04, 7.9513E+04, 7.9624E+04, 7.9735E+04,
     +  7.9846E+04, 7.9958E+04, 8.0069E+04, 8.0181E+04, 8.0293E+04,
     + 8.0404E+04, 8.0516E+04, 8.0628E+04, 8.0740E+04, 8.0853E+04,
     +  8.0965E+04, 8.1078E+04, 8.1190E+04, 8.1303E+04, 8.1416E+04,
     + 8.1529E+04, 8.1642E+04, 8.1755E+04, 8.1868E+04, 8.1982E+04,
     +  8.2095E+04, 8.2209E+04, 8.2323E+04, 8.2437E+04, 8.2551E+04,
     + 8.2665E+04, 8.2779E+04, 8.2893E+04, 8.3008E+04, 8.3122E+04,
     +  8.3237E+04, 8.3352E+04, 8.3467E+04, 8.3582E+04, 8.3697E+04,
     + 8.3812E+04, 8.3927E+04, 8.4043E+04, 8.4159E+04, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       181
      data (QofT( 2,J),J=1,2408)/ 1.0000E+00, 1.0000E+00, 1.0001E+00,
     +  1.0018E+00, 1.0098E+00, 1.0310E+00, 1.0714E+00, 1.1343E+00,
     + 1.2213E+00, 1.3319E+00, 1.4650E+00, 1.6189E+00, 1.7914E+00,
     +  1.9806E+00, 2.1846E+00, 2.4016E+00, 2.6303E+00, 2.8691E+00,
     + 3.1169E+00, 3.3728E+00, 3.9054E+00, 4.4617E+00, 5.0378E+00,
     +  5.6313E+00, 6.2404E+00, 6.8640E+00, 7.5014E+00, 8.1520E+00,
     + 8.8157E+00, 9.4923E+00, 1.0182E+01, 1.0884E+01, 1.1599E+01,
     +  1.2326E+01, 1.3066E+01, 1.3819E+01, 1.4585E+01, 1.5362E+01,
     + 1.6153E+01, 1.6956E+01, 1.7771E+01, 1.8598E+01, 1.9437E+01,
     +  2.0288E+01, 2.1151E+01, 2.2026E+01, 2.2913E+01, 2.3811E+01,
     + 2.4720E+01, 2.5641E+01, 2.6573E+01, 2.7516E+01, 2.8470E+01,
     +  2.9434E+01, 3.0410E+01, 3.1396E+01, 3.2392E+01, 3.3399E+01,
     + 3.4417E+01, 3.5444E+01, 3.6482E+01, 3.7529E+01, 3.8587E+01,
     +  3.9654E+01, 4.0731E+01, 4.1818E+01, 4.2915E+01, 4.4021E+01,
     + 4.5136E+01, 4.6261E+01, 4.7395E+01, 4.8538E+01, 4.9691E+01,
     +  5.0852E+01, 5.2023E+01, 5.3202E+01, 5.4390E+01, 5.5587E+01,
     + 5.6793E+01, 5.8008E+01, 5.9231E+01, 6.0463E+01, 6.1703E+01,
     +  6.2952E+01, 6.4209E+01, 6.5475E+01, 6.6748E+01, 6.8030E+01,
     + 6.9321E+01, 7.0619E+01, 7.1926E+01, 7.3240E+01, 7.4563E+01,
     +  7.5893E+01, 7.7232E+01, 7.8578E+01, 7.9932E+01, 8.1294E+01,
     + 8.2664E+01, 8.4042E+01, 8.5427E+01, 8.6819E+01, 8.8220E+01,
     +  8.9628E+01, 9.1043E+01, 9.2466E+01, 9.3896E+01, 9.5334E+01,
     + 9.6779E+01, 9.8232E+01, 9.9691E+01, 1.0116E+02, 1.0263E+02,
     +  1.0411E+02, 1.0560E+02, 1.0710E+02, 1.0860E+02, 1.1011E+02,
     + 1.1163E+02, 1.1315E+02, 1.1468E+02, 1.1622E+02, 1.1777E+02,
     +  1.1932E+02, 1.2088E+02, 1.2244E+02, 1.2402E+02, 1.2560E+02,
     + 1.2718E+02, 1.2877E+02, 1.3037E+02, 1.3198E+02, 1.3359E+02,
     +  1.3521E+02, 1.3684E+02, 1.3847E+02, 1.4011E+02, 1.4176E+02,
     + 1.4341E+02, 1.4507E+02, 1.4674E+02, 1.4841E+02, 1.5009E+02,
     +  1.5178E+02, 1.5347E+02, 1.5517E+02, 1.5687E+02, 1.5858E+02,
     + 1.6030E+02, 1.6203E+02, 1.6376E+02, 1.6550E+02, 1.6724E+02,
     +  1.6899E+02, 1.7075E+02, 1.7251E+02, 1.7428E+02, 1.7605E+02,
     + 1.7783E+02, 1.7962E+02, 1.8142E+02, 1.8322E+02, 1.8502E+02,
     +  1.8684E+02, 1.8866E+02, 1.9048E+02, 1.9231E+02, 1.9415E+02,
     + 1.9599E+02, 1.9784E+02, 1.9970E+02, 2.0156E+02, 2.0343E+02,
     +  2.0531E+02, 2.0719E+02, 2.0908E+02, 2.1097E+02, 2.1287E+02,
     + 2.1478E+02, 2.1669E+02, 2.1861E+02, 2.2053E+02, 2.2246E+02,
     +  2.2440E+02, 2.2634E+02, 2.2829E+02, 2.3024E+02, 2.3221E+02,
     + 2.3417E+02, 2.3615E+02, 2.3813E+02, 2.4011E+02, 2.4210E+02,
     +  2.4410E+02, 2.4611E+02, 2.4812E+02, 2.5013E+02, 2.5215E+02,
     + 2.5418E+02, 2.5622E+02, 2.5826E+02, 2.6031E+02, 2.6236E+02,
     +  2.6442E+02, 2.6648E+02, 2.6856E+02, 2.7063E+02, 2.7272E+02,
     + 2.7481E+02, 2.7690E+02, 2.7900E+02, 2.8111E+02, 2.8323E+02,
     +  2.8535E+02, 2.8748E+02, 2.8961E+02, 2.9175E+02, 2.9389E+02,
     + 2.9604E+02, 2.9820E+02, 3.0037E+02, 3.0254E+02, 3.0471E+02,
     +  3.0690E+02, 3.0908E+02, 3.1128E+02, 3.1348E+02, 3.1569E+02,
     + 3.1790E+02, 3.2012E+02, 3.2235E+02, 3.2458E+02, 3.2682E+02,
     +  3.2906E+02, 3.3131E+02, 3.3357E+02, 3.3584E+02, 3.3811E+02,
     + 3.4038E+02, 3.4266E+02, 3.4495E+02, 3.4725E+02, 3.4955E+02,
     +  3.5186E+02, 3.5417E+02, 3.5649E+02, 3.5882E+02, 3.6115E+02,
     + 3.6349E+02, 3.6584E+02, 3.6819E+02, 3.7055E+02, 3.7292E+02,
     +  3.7529E+02, 3.7767E+02, 3.8005E+02, 3.8244E+02, 3.8484E+02,
     + 3.8724E+02, 3.8965E+02, 3.9207E+02, 3.9449E+02, 3.9692E+02,
     +  3.9936E+02, 4.0180E+02, 4.0425E+02, 4.0671E+02, 4.0917E+02,
     + 4.1164E+02, 4.1412E+02, 4.1660E+02, 4.1909E+02, 4.2158E+02,
     +  4.2409E+02, 4.2659E+02, 4.2911E+02, 4.3163E+02, 4.3416E+02,
     + 4.3669E+02, 4.3924E+02, 4.4179E+02, 4.4434E+02, 4.4690E+02,
     +  4.4947E+02, 4.5205E+02, 4.5463E+02, 4.5722E+02, 4.5981E+02,
     + 4.6242E+02, 4.6502E+02, 4.6764E+02, 4.7026E+02, 4.7289E+02,
     +  4.7553E+02, 4.7817E+02, 4.8082E+02, 4.8348E+02, 4.8614E+02,
     + 4.8881E+02, 4.9149E+02, 4.9418E+02, 4.9687E+02, 4.9957E+02,
     +  5.0227E+02, 5.0498E+02, 5.0770E+02, 5.1043E+02, 5.1316E+02,
     + 5.1590E+02, 5.1865E+02, 5.2141E+02, 5.2417E+02, 5.2693E+02,
     +  5.2971E+02, 5.3249E+02, 5.3528E+02, 5.3808E+02, 5.4088E+02,
     + 5.4369E+02, 5.4651E+02, 5.4934E+02, 5.5217E+02, 5.5501E+02,
     +  5.5786E+02, 5.6071E+02, 5.6357E+02, 5.6644E+02, 5.6932E+02,
     + 5.7220E+02, 5.7509E+02, 5.7799E+02, 5.8089E+02, 5.8380E+02,
     +  5.8672E+02, 5.8965E+02, 5.9259E+02, 5.9553E+02, 5.9848E+02,
     + 6.0143E+02, 6.0440E+02, 6.0737E+02, 6.1035E+02, 6.1333E+02,
     +  6.1633E+02, 6.1933E+02, 6.2234E+02, 6.2535E+02, 6.2838E+02,
     + 6.3141E+02, 6.3445E+02, 6.3749E+02, 6.4055E+02, 6.4361E+02,
     +  6.4668E+02, 6.4975E+02, 6.5284E+02, 6.5593E+02, 6.5903E+02,
     + 6.6214E+02, 6.6525E+02, 6.6838E+02, 6.7151E+02, 6.7465E+02,
     +  6.7779E+02, 6.8095E+02, 6.8411E+02, 6.8728E+02, 6.9045E+02,
     + 6.9364E+02, 6.9683E+02, 7.0003E+02, 7.0324E+02, 7.0646E+02,
     +  7.0968E+02, 7.1292E+02, 7.1616E+02, 7.1940E+02, 7.2266E+02,
     + 7.2593E+02, 7.2920E+02, 7.3248E+02, 7.3577E+02, 7.3906E+02,
     +  7.4237E+02, 7.4568E+02, 7.4900E+02, 7.5233E+02, 7.5567E+02,
     + 7.5901E+02, 7.6237E+02, 7.6573E+02, 7.6910E+02, 7.7248E+02,
     +  7.7586E+02, 7.7926E+02, 7.8266E+02, 7.8607E+02, 7.8949E+02,
     + 7.9292E+02, 7.9636E+02, 7.9980E+02, 8.0325E+02, 8.0671E+02,
     +  8.1018E+02, 8.1366E+02, 8.1715E+02, 8.2064E+02, 8.2415E+02,
     + 8.2766E+02, 8.3118E+02, 8.3471E+02, 8.3824E+02, 8.4179E+02,
     +  8.4535E+02, 8.4891E+02, 8.5248E+02, 8.5606E+02, 8.5965E+02,
     + 8.6325E+02, 8.6685E+02, 8.7047E+02, 8.7409E+02, 8.7773E+02,
     +  8.8137E+02, 8.8502E+02, 8.8868E+02, 8.9234E+02, 8.9602E+02,
     + 8.9970E+02, 9.0340E+02, 9.0710E+02, 9.1081E+02, 9.1453E+02,
     +  9.1826E+02, 9.2200E+02, 9.2575E+02, 9.2951E+02, 9.3327E+02,
     + 9.3705E+02, 9.4083E+02, 9.4462E+02, 9.4843E+02, 9.5224E+02,
     +  9.5606E+02, 9.5989E+02, 9.6372E+02, 9.6757E+02, 9.7143E+02,
     + 9.7529E+02, 9.7917E+02, 9.8305E+02, 9.8695E+02, 9.9085E+02,
     +  9.9476E+02, 9.9868E+02, 1.0026E+03, 1.0066E+03, 1.0105E+03,
     + 1.0145E+03, 1.0184E+03, 1.0224E+03, 1.0264E+03, 1.0304E+03,
     +  1.0344E+03, 1.0384E+03, 1.0424E+03, 1.0465E+03, 1.0505E+03,
     + 1.0546E+03, 1.0586E+03, 1.0627E+03, 1.0668E+03, 1.0709E+03,
     +  1.0750E+03, 1.0791E+03, 1.0832E+03, 1.0874E+03, 1.0915E+03,
     + 1.0956E+03, 1.0998E+03, 1.1040E+03, 1.1082E+03, 1.1123E+03,
     +  1.1165E+03, 1.1208E+03, 1.1250E+03, 1.1292E+03, 1.1334E+03,
     + 1.1377E+03, 1.1420E+03, 1.1462E+03, 1.1505E+03, 1.1548E+03,
     +  1.1591E+03, 1.1634E+03, 1.1677E+03, 1.1721E+03, 1.1764E+03,
     + 1.1807E+03, 1.1851E+03, 1.1895E+03, 1.1939E+03, 1.1982E+03,
     +  1.2026E+03, 1.2071E+03, 1.2115E+03, 1.2159E+03, 1.2204E+03,
     + 1.2248E+03, 1.2293E+03, 1.2337E+03, 1.2382E+03, 1.2427E+03,
     +  1.2472E+03, 1.2518E+03, 1.2563E+03, 1.2608E+03, 1.2654E+03,
     + 1.2699E+03, 1.2745E+03, 1.2791E+03, 1.2837E+03, 1.2883E+03,
     +  1.2929E+03, 1.2975E+03, 1.3021E+03, 1.3068E+03, 1.3114E+03,
     + 1.3161E+03, 1.3208E+03, 1.3255E+03, 1.3302E+03, 1.3349E+03,
     +  1.3396E+03, 1.3443E+03, 1.3491E+03, 1.3538E+03, 1.3586E+03,
     + 1.3633E+03, 1.3681E+03, 1.3729E+03, 1.3777E+03, 1.3826E+03,
     +  1.3874E+03, 1.3922E+03, 1.3971E+03, 1.4019E+03, 1.4068E+03,
     + 1.4117E+03, 1.4166E+03, 1.4215E+03, 1.4264E+03, 1.4314E+03,
     +  1.4363E+03, 1.4413E+03, 1.4462E+03, 1.4512E+03, 1.4562E+03,
     + 1.4612E+03, 1.4662E+03, 1.4712E+03, 1.4763E+03, 1.4813E+03,
     +  1.4864E+03, 1.4914E+03, 1.4965E+03, 1.5016E+03, 1.5067E+03,
     + 1.5118E+03, 1.5170E+03, 1.5221E+03, 1.5272E+03, 1.5324E+03,
     +  1.5376E+03, 1.5428E+03, 1.5480E+03, 1.5532E+03, 1.5584E+03,
     + 1.5636E+03, 1.5689E+03, 1.5741E+03, 1.5794E+03, 1.5847E+03,
     +  1.5900E+03, 1.5953E+03, 1.6006E+03, 1.6059E+03, 1.6113E+03,
     + 1.6166E+03, 1.6220E+03, 1.6274E+03, 1.6328E+03, 1.6382E+03,
     +  1.6436E+03, 1.6490E+03, 1.6544E+03, 1.6599E+03, 1.6654E+03,
     + 1.6708E+03, 1.6763E+03, 1.6818E+03, 1.6873E+03, 1.6929E+03,
     +  1.6984E+03, 1.7040E+03, 1.7095E+03, 1.7151E+03, 1.7207E+03,
     + 1.7263E+03, 1.7319E+03, 1.7375E+03, 1.7432E+03, 1.7488E+03,
     +  1.7545E+03, 1.7602E+03, 1.7659E+03, 1.7716E+03, 1.7773E+03,
     + 1.7830E+03, 1.7888E+03, 1.7945E+03, 1.8003E+03, 1.8061E+03,
     +  1.8119E+03, 1.8177E+03, 1.8235E+03, 1.8293E+03, 1.8352E+03,
     + 1.8410E+03, 1.8469E+03, 1.8528E+03, 1.8587E+03, 1.8646E+03,
     +  1.8705E+03, 1.8764E+03, 1.8824E+03, 1.8884E+03, 1.8943E+03,
     + 1.9003E+03, 1.9063E+03, 1.9124E+03, 1.9184E+03, 1.9244E+03,
     +  1.9305E+03, 1.9366E+03, 1.9426E+03, 1.9487E+03, 1.9549E+03,
     + 1.9610E+03, 1.9671E+03, 1.9733E+03, 1.9794E+03, 1.9856E+03,
     +  1.9918E+03, 1.9980E+03, 2.0042E+03, 2.0105E+03, 2.0167E+03,
     + 2.0230E+03, 2.0293E+03, 2.0356E+03, 2.0419E+03, 2.0482E+03,
     +  2.0545E+03, 2.0609E+03, 2.0672E+03, 2.0736E+03, 2.0800E+03,
     + 2.0864E+03, 2.0928E+03, 2.0992E+03, 2.1057E+03, 2.1121E+03,
     +  2.1186E+03, 2.1251E+03, 2.1316E+03, 2.1381E+03, 2.1447E+03,
     + 2.1512E+03, 2.1578E+03, 2.1643E+03, 2.1709E+03, 2.1775E+03,
     +  2.1841E+03, 2.1908E+03, 2.1974E+03, 2.2041E+03, 2.2108E+03,
     + 2.2175E+03, 2.2242E+03, 2.2309E+03, 2.2376E+03, 2.2444E+03,
     +  2.2511E+03, 2.2579E+03, 2.2647E+03, 2.2715E+03, 2.2783E+03,
     + 2.2852E+03, 2.2920E+03, 2.2989E+03, 2.3058E+03, 2.3127E+03,
     +  2.3196E+03, 2.3265E+03, 2.3334E+03, 2.3404E+03, 2.3474E+03,
     + 2.3544E+03, 2.3614E+03, 2.3684E+03, 2.3754E+03, 2.3825E+03,
     +  2.3895E+03, 2.3966E+03, 2.4037E+03, 2.4108E+03, 2.4179E+03,
     + 2.4251E+03, 2.4322E+03, 2.4394E+03, 2.4466E+03, 2.4538E+03,
     +  2.4610E+03, 2.4682E+03, 2.4755E+03, 2.4827E+03, 2.4900E+03,
     + 2.4973E+03, 2.5046E+03, 2.5120E+03, 2.5193E+03, 2.5267E+03,
     +  2.5340E+03, 2.5414E+03, 2.5488E+03, 2.5562E+03, 2.5637E+03,
     + 2.5711E+03, 2.5786E+03, 2.5861E+03, 2.5936E+03, 2.6011E+03,
     +  2.6086E+03, 2.6162E+03, 2.6238E+03, 2.6313E+03, 2.6389E+03,
     + 2.6466E+03, 2.6542E+03, 2.6618E+03, 2.6695E+03, 2.6772E+03,
     +  2.6849E+03, 2.6926E+03, 2.7003E+03, 2.7080E+03, 2.7158E+03,
     + 2.7236E+03, 2.7314E+03, 2.7392E+03, 2.7470E+03, 2.7549E+03,
     +  2.7627E+03, 2.7706E+03, 2.7785E+03, 2.7864E+03, 2.7943E+03,
     + 2.8023E+03, 2.8102E+03, 2.8182E+03, 2.8262E+03, 2.8342E+03,
     +  2.8422E+03, 2.8503E+03, 2.8583E+03, 2.8664E+03, 2.8745E+03,
     + 2.8826E+03, 2.8908E+03, 2.8989E+03, 2.9071E+03, 2.9153E+03,
     +  2.9235E+03, 2.9317E+03, 2.9399E+03, 2.9482E+03, 2.9564E+03,
     + 2.9647E+03, 2.9730E+03, 2.9813E+03, 2.9897E+03, 2.9980E+03,
     +  3.0064E+03, 3.0148E+03, 3.0232E+03, 3.0316E+03, 3.0400E+03,
     + 3.0485E+03, 3.0570E+03, 3.0655E+03, 3.0740E+03, 3.0825E+03,
     +  3.0911E+03, 3.0996E+03, 3.1082E+03, 3.1168E+03, 3.1254E+03,
     + 3.1341E+03, 3.1427E+03, 3.1514E+03, 3.1601E+03, 3.1688E+03,
     +  3.1775E+03, 3.1863E+03, 3.1950E+03, 3.2038E+03, 3.2126E+03,
     + 3.2214E+03, 3.2303E+03, 3.2391E+03, 3.2480E+03, 3.2569E+03,
     +  3.2658E+03, 3.2747E+03, 3.2837E+03, 3.2926E+03, 3.3016E+03,
     + 3.3106E+03, 3.3196E+03, 3.3287E+03, 3.3377E+03, 3.3468E+03,
     +  3.3559E+03, 3.3650E+03, 3.3741E+03, 3.3833E+03, 3.3924E+03,
     + 3.4016E+03, 3.4108E+03, 3.4201E+03, 3.4293E+03, 3.4386E+03,
     +  3.4478E+03, 3.4571E+03, 3.4665E+03, 3.4758E+03, 3.4852E+03,
     + 3.4945E+03, 3.5039E+03, 3.5133E+03, 3.5228E+03, 3.5322E+03,
     +  3.5417E+03, 3.5512E+03, 3.5607E+03, 3.5702E+03, 3.5798E+03,
     + 3.5894E+03, 3.5989E+03, 3.6086E+03, 3.6182E+03, 3.6278E+03,
     +  3.6375E+03, 3.6472E+03, 3.6569E+03, 3.6666E+03, 3.6763E+03,
     + 3.6861E+03, 3.6959E+03, 3.7057E+03, 3.7155E+03, 3.7254E+03,
     +  3.7352E+03, 3.7451E+03, 3.7550E+03, 3.7649E+03, 3.7749E+03,
     + 3.7848E+03, 3.7948E+03, 3.8048E+03, 3.8148E+03, 3.8249E+03,
     +  3.8350E+03, 3.8450E+03, 3.8551E+03, 3.8653E+03, 3.8754E+03,
     + 3.8856E+03, 3.8958E+03, 3.9060E+03, 3.9162E+03, 3.9264E+03,
     +  3.9367E+03, 3.9470E+03, 3.9573E+03, 3.9676E+03, 3.9780E+03,
     + 3.9883E+03, 3.9987E+03, 4.0091E+03, 4.0196E+03, 4.0300E+03,
     +  4.0405E+03, 4.0510E+03, 4.0615E+03, 4.0720E+03, 4.0826E+03,
     + 4.0932E+03, 4.1038E+03, 4.1144E+03, 4.1250E+03, 4.1357E+03,
     +  4.1464E+03, 4.1571E+03, 4.1678E+03, 4.1785E+03, 4.1893E+03,
     + 4.2001E+03, 4.2109E+03, 4.2217E+03, 4.2326E+03, 4.2434E+03,
     +  4.2543E+03, 4.2652E+03, 4.2762E+03, 4.2871E+03, 4.2981E+03,
     + 4.3091E+03, 4.3201E+03, 4.3312E+03, 4.3423E+03, 4.3533E+03,
     +  4.3645E+03, 4.3756E+03, 4.3867E+03, 4.3979E+03, 4.4091E+03,
     + 4.4203E+03, 4.4316E+03, 4.4428E+03, 4.4541E+03, 4.4654E+03,
     +  4.4768E+03, 4.4881E+03, 4.4995E+03, 4.5109E+03, 4.5223E+03,
     + 4.5337E+03, 4.5452E+03, 4.5567E+03, 4.5682E+03, 4.5797E+03,
     +  4.5913E+03, 4.6028E+03, 4.6144E+03, 4.6261E+03, 4.6377E+03,
     + 4.6494E+03, 4.6610E+03, 4.6728E+03, 4.6845E+03, 4.6962E+03,
     +  4.7080E+03, 4.7198E+03, 4.7316E+03, 4.7435E+03, 4.7554E+03,
     + 4.7672E+03, 4.7792E+03, 4.7911E+03, 4.8031E+03, 4.8150E+03,
     +  4.8270E+03, 4.8391E+03, 4.8511E+03, 4.8632E+03, 4.8753E+03,
     + 4.8874E+03, 4.8996E+03, 4.9117E+03, 4.9239E+03, 4.9361E+03,
     +  4.9484E+03, 4.9606E+03, 4.9729E+03, 4.9852E+03, 4.9976E+03,
     + 5.0099E+03, 5.0223E+03, 5.0347E+03, 5.0471E+03, 5.0596E+03,
     +  5.0720E+03, 5.0845E+03, 5.0971E+03, 5.1096E+03, 5.1222E+03,
     + 5.1348E+03, 5.1474E+03, 5.1600E+03, 5.1727E+03, 5.1854E+03,
     +  5.1981E+03, 5.2108E+03, 5.2236E+03, 5.2364E+03, 5.2492E+03,
     + 5.2620E+03, 5.2749E+03, 5.2878E+03, 5.3007E+03, 5.3136E+03,
     +  5.3266E+03, 5.3395E+03, 5.3525E+03, 5.3656E+03, 5.3786E+03,
     + 5.3917E+03, 5.4048E+03, 5.4179E+03, 5.4311E+03, 5.4443E+03,
     +  5.4575E+03, 5.4707E+03, 5.4839E+03, 5.4972E+03, 5.5105E+03,
     + 5.5239E+03, 5.5372E+03, 5.5506E+03, 5.5640E+03, 5.5774E+03,
     +  5.5909E+03, 5.6043E+03, 5.6178E+03, 5.6314E+03, 5.6449E+03,
     + 5.6585E+03, 5.6721E+03, 5.6857E+03, 5.6994E+03, 5.7131E+03,
     +  5.7268E+03, 5.7405E+03, 5.7542E+03, 5.7680E+03, 5.7818E+03,
     + 5.7957E+03, 5.8095E+03, 5.8234E+03, 5.8373E+03, 5.8513E+03,
     +  5.8652E+03, 5.8792E+03, 5.8932E+03, 5.9073E+03, 5.9213E+03,
     + 5.9354E+03, 5.9495E+03, 5.9637E+03, 5.9778E+03, 5.9920E+03,
     +  6.0063E+03, 6.0205E+03, 6.0348E+03, 6.0491E+03, 6.0634E+03,
     + 6.0778E+03, 6.0921E+03, 6.1065E+03, 6.1210E+03, 6.1354E+03,
     +  6.1499E+03, 6.1644E+03, 6.1790E+03, 6.1935E+03, 6.2081E+03,
     + 6.2227E+03, 6.2374E+03, 6.2521E+03, 6.2668E+03, 6.2815E+03,
     +  6.2962E+03, 6.3110E+03, 6.3258E+03, 6.3407E+03, 6.3555E+03,
     + 6.3704E+03, 6.3853E+03, 6.4003E+03, 6.4152E+03, 6.4302E+03,
     +  6.4453E+03, 6.4603E+03, 6.4754E+03, 6.4905E+03, 6.5056E+03,
     + 6.5208E+03, 6.5360E+03, 6.5512E+03, 6.5665E+03, 6.5817E+03,
     +  6.5970E+03, 6.6124E+03, 6.6277E+03, 6.6431E+03, 6.6585E+03,
     + 6.6739E+03, 6.6894E+03, 6.7049E+03, 6.7204E+03, 6.7360E+03,
     +  6.7516E+03, 6.7672E+03, 6.7828E+03, 6.7985E+03, 6.8141E+03,
     + 6.8299E+03, 6.8456E+03, 6.8614E+03, 6.8772E+03, 6.8930E+03,
     +  6.9089E+03, 6.9248E+03, 6.9407E+03, 6.9566E+03, 6.9726E+03,
     + 6.9886E+03, 7.0047E+03, 7.0207E+03, 7.0368E+03, 7.0529E+03,
     +  7.0691E+03, 7.0852E+03, 7.1015E+03, 7.1177E+03, 7.1340E+03,
     + 7.1502E+03, 7.1666E+03, 7.1829E+03, 7.1993E+03, 7.2157E+03,
     +  7.2321E+03, 7.2486E+03, 7.2651E+03, 7.2816E+03, 7.2982E+03,
     + 7.3148E+03, 7.3314E+03, 7.3480E+03, 7.3647E+03, 7.3814E+03,
     +  7.3981E+03, 7.4149E+03, 7.4317E+03, 7.4485E+03, 7.4654E+03,
     + 7.4822E+03, 7.4992E+03, 7.5161E+03, 7.5331E+03, 7.5501E+03,
     +  7.5671E+03, 7.5842E+03, 7.6012E+03, 7.6184E+03, 7.6355E+03,
     + 7.6527E+03, 7.6699E+03, 7.6871E+03, 7.7044E+03, 7.7217E+03,
     +  7.7391E+03, 7.7564E+03, 7.7738E+03, 7.7912E+03, 7.8087E+03,
     + 7.8262E+03, 7.8437E+03, 7.8612E+03, 7.8788E+03, 7.8964E+03,
     +  7.9141E+03, 7.9317E+03, 7.9494E+03, 7.9672E+03, 7.9849E+03,
     + 8.0027E+03, 8.0205E+03, 8.0384E+03, 8.0563E+03, 8.0742E+03,
     +  8.0921E+03, 8.1101E+03, 8.1281E+03, 8.1462E+03, 8.1642E+03,
     + 8.1824E+03, 8.2005E+03, 8.2187E+03, 8.2369E+03, 8.2551E+03,
     +  8.2734E+03, 8.2916E+03, 8.3100E+03, 8.3283E+03, 8.3467E+03,
     + 8.3651E+03, 8.3836E+03, 8.4021E+03, 8.4206E+03, 8.4391E+03,
     +  8.4577E+03, 8.4763E+03, 8.4950E+03, 8.5137E+03, 8.5324E+03,
     + 8.5511E+03, 8.5699E+03, 8.5887E+03, 8.6075E+03, 8.6264E+03,
     +  8.6453E+03, 8.6642E+03, 8.6832E+03, 8.7022E+03, 8.7212E+03,
     + 8.7403E+03, 8.7594E+03, 8.7785E+03, 8.7977E+03, 8.8169E+03,
     +  8.8361E+03, 8.8554E+03, 8.8747E+03, 8.8940E+03, 8.9134E+03,
     + 8.9328E+03, 8.9522E+03, 8.9717E+03, 8.9911E+03, 9.0107E+03,
     +  9.0302E+03, 9.0498E+03, 9.0695E+03, 9.0891E+03, 9.1088E+03,
     + 9.1285E+03, 9.1483E+03, 9.1681E+03, 9.1879E+03, 9.2078E+03,
     +  9.2277E+03, 9.2476E+03, 9.2676E+03, 9.2876E+03, 9.3076E+03,
     + 9.3277E+03, 9.3478E+03, 9.3679E+03, 9.3881E+03, 9.4083E+03,
     +  9.4285E+03, 9.4488E+03, 9.4691E+03, 9.4894E+03, 9.5098E+03,
     + 9.5302E+03, 9.5506E+03, 9.5711E+03, 9.5916E+03, 9.6121E+03,
     +  9.6327E+03, 9.6533E+03, 9.6740E+03, 9.6947E+03, 9.7154E+03,
     + 9.7361E+03, 9.7569E+03, 9.7777E+03, 9.7986E+03, 9.8195E+03,
     +  9.8404E+03, 9.8614E+03, 9.8824E+03, 9.9034E+03, 9.9245E+03,
     + 9.9456E+03, 9.9667E+03, 9.9879E+03, 1.0009E+04, 1.0030E+04,
     +  1.0052E+04, 1.0073E+04, 1.0094E+04, 1.0116E+04, 1.0137E+04,
     + 1.0159E+04, 1.0180E+04, 1.0202E+04, 1.0223E+04, 1.0245E+04,
     +  1.0266E+04, 1.0288E+04, 1.0310E+04, 1.0331E+04, 1.0353E+04,
     + 1.0375E+04, 1.0397E+04, 1.0419E+04, 1.0441E+04, 1.0463E+04,
     +  1.0485E+04, 1.0507E+04, 1.0529E+04, 1.0551E+04, 1.0573E+04,
     + 1.0595E+04, 1.0617E+04, 1.0640E+04, 1.0662E+04, 1.0684E+04,
     +  1.0707E+04, 1.0729E+04, 1.0751E+04, 1.0774E+04, 1.0796E+04,
     + 1.0819E+04, 1.0842E+04, 1.0864E+04, 1.0887E+04, 1.0909E+04,
     +  1.0932E+04, 1.0955E+04, 1.0978E+04, 1.1001E+04, 1.1024E+04,
     + 1.1046E+04, 1.1069E+04, 1.1092E+04, 1.1115E+04, 1.1138E+04,
     +  1.1162E+04, 1.1185E+04, 1.1208E+04, 1.1231E+04, 1.1254E+04,
     + 1.1278E+04, 1.1301E+04, 1.1324E+04, 1.1348E+04, 1.1371E+04,
     +  1.1395E+04, 1.1418E+04, 1.1442E+04, 1.1465E+04, 1.1489E+04,
     + 1.1513E+04, 1.1536E+04, 1.1560E+04, 1.1584E+04, 1.1608E+04,
     +  1.1632E+04, 1.1656E+04, 1.1679E+04, 1.1703E+04, 1.1727E+04,
     + 1.1752E+04, 1.1776E+04, 1.1800E+04, 1.1824E+04, 1.1848E+04,
     +  1.1872E+04, 1.1897E+04, 1.1921E+04, 1.1945E+04, 1.1970E+04,
     + 1.1994E+04, 1.2019E+04, 1.2043E+04, 1.2068E+04, 1.2092E+04,
     +  1.2117E+04, 1.2142E+04, 1.2167E+04, 1.2191E+04, 1.2216E+04,
     + 1.2241E+04, 1.2266E+04, 1.2291E+04, 1.2316E+04, 1.2341E+04,
     +  1.2366E+04, 1.2391E+04, 1.2416E+04, 1.2441E+04, 1.2466E+04,
     + 1.2492E+04, 1.2517E+04, 1.2542E+04, 1.2568E+04, 1.2593E+04,
     +  1.2618E+04, 1.2644E+04, 1.2669E+04, 1.2695E+04, 1.2721E+04,
     + 1.2746E+04, 1.2772E+04, 1.2798E+04, 1.2823E+04, 1.2849E+04,
     +  1.2875E+04, 1.2901E+04, 1.2927E+04, 1.2953E+04, 1.2979E+04,
     + 1.3005E+04, 1.3031E+04, 1.3057E+04, 1.3083E+04, 1.3110E+04,
     +  1.3136E+04, 1.3162E+04, 1.3189E+04, 1.3215E+04, 1.3241E+04,
     + 1.3268E+04, 1.3294E+04, 1.3321E+04, 1.3348E+04, 1.3374E+04,
     +  1.3401E+04, 1.3428E+04, 1.3454E+04, 1.3481E+04, 1.3508E+04,
     + 1.3535E+04, 1.3562E+04, 1.3589E+04, 1.3616E+04, 1.3643E+04,
     +  1.3670E+04, 1.3697E+04, 1.3724E+04, 1.3752E+04, 1.3779E+04,
     + 1.3806E+04, 1.3833E+04, 1.3861E+04, 1.3888E+04, 1.3916E+04,
     +  1.3943E+04, 1.3971E+04, 1.3999E+04, 1.4026E+04, 1.4054E+04,
     + 1.4082E+04, 1.4109E+04, 1.4137E+04, 1.4165E+04, 1.4193E+04,
     +  1.4221E+04, 1.4249E+04, 1.4277E+04, 1.4305E+04, 1.4333E+04,
     + 1.4361E+04, 1.4390E+04, 1.4418E+04, 1.4446E+04, 1.4475E+04,
     +  1.4503E+04, 1.4531E+04, 1.4560E+04, 1.4588E+04, 1.4617E+04,
     + 1.4646E+04, 1.4674E+04, 1.4703E+04, 1.4732E+04, 1.4761E+04,
     +  1.4789E+04, 1.4818E+04, 1.4847E+04, 1.4876E+04, 1.4905E+04,
     + 1.4934E+04, 1.4963E+04, 1.4992E+04, 1.5022E+04, 1.5051E+04,
     +  1.5080E+04, 1.5110E+04, 1.5139E+04, 1.5168E+04, 1.5198E+04,
     + 1.5227E+04, 1.5257E+04, 1.5286E+04, 1.5316E+04, 1.5346E+04,
     +  1.5375E+04, 1.5405E+04, 1.5435E+04, 1.5465E+04, 1.5495E+04,
     + 1.5525E+04, 1.5555E+04, 1.5585E+04, 1.5615E+04, 1.5645E+04,
     +  1.5675E+04, 1.5706E+04, 1.5736E+04, 1.5766E+04, 1.5797E+04,
     + 1.5827E+04, 1.5857E+04, 1.5888E+04, 1.5919E+04, 1.5949E+04,
     +  1.5980E+04, 1.6010E+04, 1.6041E+04, 1.6072E+04, 1.6103E+04,
     + 1.6134E+04, 1.6165E+04, 1.6196E+04, 1.6227E+04, 1.6258E+04,
     +  1.6289E+04, 1.6320E+04, 1.6351E+04, 1.6382E+04, 1.6414E+04,
     + 1.6445E+04, 1.6477E+04, 1.6508E+04, 1.6539E+04, 1.6571E+04,
     +  1.6603E+04, 1.6634E+04, 1.6666E+04, 1.6698E+04, 1.6729E+04,
     + 1.6761E+04, 1.6793E+04, 1.6825E+04, 1.6857E+04, 1.6889E+04,
     +  1.6921E+04, 1.6953E+04, 1.6985E+04, 1.7018E+04, 1.7050E+04,
     + 1.7082E+04, 1.7115E+04, 1.7147E+04, 1.7179E+04, 1.7212E+04,
     +  1.7244E+04, 1.7277E+04, 1.7310E+04, 1.7342E+04, 1.7375E+04,
     + 1.7408E+04, 1.7441E+04, 1.7474E+04, 1.7507E+04, 1.7540E+04,
     +  1.7573E+04, 1.7606E+04, 1.7639E+04, 1.7672E+04, 1.7705E+04,
     + 1.7739E+04, 1.7772E+04, 1.7805E+04, 1.7839E+04, 1.7872E+04,
     +  1.7906E+04, 1.7939E+04, 1.7973E+04, 1.8007E+04, 1.8040E+04,
     + 1.8074E+04, 1.8108E+04, 1.8142E+04, 1.8176E+04, 1.8210E+04,
     +  1.8244E+04, 1.8278E+04, 1.8312E+04, 1.8346E+04, 1.8380E+04,
     + 1.8415E+04, 1.8449E+04, 1.8483E+04, 1.8518E+04, 1.8552E+04,
     +  1.8587E+04, 1.8621E+04, 1.8656E+04, 1.8691E+04, 1.8725E+04,
     + 1.8760E+04, 1.8795E+04, 1.8830E+04, 1.8865E+04, 1.8900E+04,
     +  1.8935E+04, 1.8970E+04, 1.9005E+04, 1.9040E+04, 1.9075E+04,
     + 1.9111E+04, 1.9146E+04, 1.9182E+04, 1.9217E+04, 1.9252E+04,
     +  1.9288E+04, 1.9324E+04, 1.9359E+04, 1.9395E+04, 1.9431E+04,
     + 1.9467E+04, 1.9502E+04, 1.9538E+04, 1.9574E+04, 1.9610E+04,
     +  1.9646E+04, 1.9683E+04, 1.9719E+04, 1.9755E+04, 1.9791E+04,
     + 1.9828E+04, 1.9864E+04, 1.9900E+04, 1.9937E+04, 1.9973E+04,
     +  2.0010E+04, 2.0047E+04, 2.0083E+04, 2.0120E+04, 2.0157E+04,
     + 2.0194E+04, 2.0231E+04, 2.0268E+04, 2.0305E+04, 2.0342E+04,
     +  2.0379E+04, 2.0416E+04, 2.0453E+04, 2.0491E+04, 2.0528E+04,
     + 2.0565E+04, 2.0603E+04, 2.0640E+04, 2.0678E+04, 2.0716E+04,
     +  2.0753E+04, 2.0791E+04, 2.0829E+04, 2.0866E+04, 2.0904E+04,
     + 2.0942E+04, 2.0980E+04, 2.1018E+04, 2.1056E+04, 2.1095E+04,
     +  2.1133E+04, 2.1171E+04, 2.1209E+04, 2.1248E+04, 2.1286E+04,
     + 2.1325E+04, 2.1363E+04, 2.1402E+04, 2.1440E+04, 2.1479E+04,
     +  2.1518E+04, 2.1557E+04, 2.1596E+04, 2.1634E+04, 2.1673E+04,
     + 2.1712E+04, 2.1752E+04, 2.1791E+04, 2.1830E+04, 2.1869E+04,
     +  2.1908E+04, 2.1948E+04, 2.1987E+04, 2.2027E+04, 2.2066E+04,
     + 2.2106E+04, 2.2145E+04, 2.2185E+04, 2.2225E+04, 2.2265E+04,
     +  2.2305E+04, 2.2345E+04, 2.2385E+04, 2.2425E+04, 2.2465E+04,
     + 2.2505E+04, 2.2545E+04, 2.2585E+04, 2.2626E+04, 2.2666E+04,
     +  2.2706E+04, 2.2747E+04, 2.2787E+04, 2.2828E+04, 2.2869E+04,
     + 2.2909E+04, 2.2950E+04, 2.2991E+04, 2.3032E+04, 2.3073E+04,
     +  2.3114E+04, 2.3155E+04, 2.3196E+04, 2.3237E+04, 2.3278E+04,
     + 2.3320E+04, 2.3361E+04, 2.3402E+04, 2.3444E+04, 2.3485E+04,
     +  2.3527E+04, 2.3569E+04, 2.3610E+04, 2.3652E+04, 2.3694E+04,
     + 2.3736E+04, 2.3778E+04, 2.3820E+04, 2.3862E+04, 2.3904E+04,
     +  2.3946E+04, 2.3988E+04, 2.4030E+04, 2.4073E+04, 2.4115E+04,
     + 2.4158E+04, 2.4200E+04, 2.4243E+04, 2.4285E+04, 2.4328E+04,
     +  2.4371E+04, 2.4414E+04, 2.4456E+04, 2.4499E+04, 2.4542E+04,
     + 2.4585E+04, 2.4629E+04, 2.4672E+04, 2.4715E+04, 2.4758E+04,
     +  2.4802E+04, 2.4845E+04, 2.4888E+04, 2.4932E+04, 2.4975E+04,
     + 2.5019E+04, 2.5063E+04, 2.5107E+04, 2.5150E+04, 2.5194E+04,
     +  2.5238E+04, 2.5282E+04, 2.5326E+04, 2.5370E+04, 2.5415E+04,
     + 2.5459E+04, 2.5503E+04, 2.5547E+04, 2.5592E+04, 2.5636E+04,
     +  2.5681E+04, 2.5726E+04, 2.5770E+04, 2.5815E+04, 2.5860E+04,
     + 2.5905E+04, 2.5950E+04, 2.5994E+04, 2.6040E+04, 2.6085E+04,
     +  2.6130E+04, 2.6175E+04, 2.6220E+04, 2.6266E+04, 2.6311E+04,
     + 2.6356E+04, 2.6402E+04, 2.6448E+04, 2.6493E+04, 2.6539E+04,
     +  2.6585E+04, 2.6631E+04, 2.6676E+04, 2.6722E+04, 2.6768E+04,
     + 2.6815E+04, 2.6861E+04, 2.6907E+04, 2.6953E+04, 2.6999E+04,
     +  2.7046E+04, 2.7092E+04, 2.7139E+04, 2.7185E+04, 2.7232E+04,
     + 2.7279E+04, 2.7326E+04, 2.7372E+04, 2.7419E+04, 2.7466E+04,
     +  2.7513E+04, 2.7560E+04, 2.7608E+04, 2.7655E+04, 2.7702E+04,
     + 2.7749E+04, 2.7797E+04, 2.7844E+04, 2.7892E+04, 2.7939E+04,
     +  2.7987E+04, 2.8035E+04, 2.8083E+04, 2.8130E+04, 2.8178E+04,
     + 2.8226E+04, 2.8274E+04, 2.8323E+04, 2.8371E+04, 2.8419E+04,
     +  2.8467E+04, 2.8516E+04, 2.8564E+04, 2.8613E+04, 2.8661E+04,
     + 2.8710E+04, 2.8758E+04, 2.8807E+04, 2.8856E+04, 2.8905E+04,
     +  2.8954E+04, 2.9003E+04, 2.9052E+04, 2.9101E+04, 2.9150E+04,
     + 2.9200E+04, 2.9249E+04, 2.9298E+04, 2.9348E+04, 2.9397E+04,
     +  2.9447E+04, 2.9497E+04, 2.9546E+04, 2.9596E+04, 2.9646E+04,
     + 2.9696E+04, 2.9746E+04, 2.9796E+04, 2.9846E+04, 2.9896E+04,
     +  2.9947E+04, 2.9997E+04, 3.0047E+04, 3.0098E+04, 3.0148E+04,
     + 3.0199E+04, 3.0250E+04, 3.0300E+04, 3.0351E+04, 3.0402E+04,
     +  3.0453E+04, 3.0504E+04, 3.0555E+04, 3.0606E+04, 3.0657E+04,
     + 3.0709E+04, 3.0760E+04, 3.0811E+04, 3.0863E+04, 3.0914E+04,
     +  3.0966E+04, 3.1018E+04, 3.1069E+04, 3.1121E+04, 3.1173E+04,
     + 3.1225E+04, 3.1277E+04, 3.1329E+04, 3.1381E+04, 3.1433E+04,
     +  3.1486E+04, 3.1538E+04, 3.1590E+04, 3.1643E+04, 3.1695E+04,
     + 3.1748E+04, 3.1801E+04, 3.1854E+04, 3.1906E+04, 3.1959E+04,
     +  3.2012E+04, 3.2065E+04, 3.2118E+04, 3.2171E+04, 3.2225E+04,
     + 3.2278E+04, 3.2331E+04, 3.2385E+04, 3.2438E+04, 3.2492E+04,
     +  3.2546E+04, 3.2599E+04, 3.2653E+04, 3.2707E+04, 3.2761E+04,
     + 3.2815E+04, 3.2869E+04, 3.2923E+04, 3.2977E+04, 3.3032E+04,
     +  3.3086E+04, 3.3140E+04, 3.3195E+04, 3.3250E+04, 3.3304E+04,
     + 3.3359E+04, 3.3414E+04, 3.3468E+04, 3.3523E+04, 3.3578E+04,
     +  3.3633E+04, 3.3689E+04, 3.3744E+04, 3.3799E+04, 3.3854E+04,
     + 3.3910E+04, 3.3965E+04, 3.4021E+04, 3.4076E+04, 3.4132E+04,
     +  3.4188E+04, 3.4244E+04, 3.4300E+04, 3.4356E+04, 3.4412E+04,
     + 3.4468E+04, 3.4524E+04, 3.4580E+04, 3.4637E+04, 3.4693E+04,
     +  3.4749E+04, 3.4806E+04, 3.4863E+04, 3.4919E+04, 3.4976E+04,
     + 3.5033E+04, 3.5090E+04, 3.5147E+04, 3.5204E+04, 3.5261E+04,
     +  3.5318E+04, 3.5376E+04, 3.5433E+04, 3.5490E+04, 3.5548E+04,
     + 3.5605E+04, 3.5663E+04, 3.5721E+04, 3.5779E+04, 3.5836E+04,
     +  3.5894E+04, 3.5952E+04, 3.6010E+04, 3.6069E+04, 3.6127E+04,
     + 3.6185E+04, 3.6243E+04, 3.6302E+04, 3.6360E+04, 3.6419E+04,
     +  3.6478E+04, 3.6536E+04, 3.6595E+04, 3.6654E+04, 3.6713E+04,
     + 3.6772E+04, 3.6831E+04, 3.6890E+04, 3.6950E+04, 3.7009E+04,
     +  3.7068E+04, 3.7128E+04, 3.7187E+04, 3.7247E+04, 3.7307E+04,
     + 3.7367E+04, 3.7426E+04, 3.7486E+04, 3.7546E+04, 3.7606E+04,
     +  3.7667E+04, 3.7727E+04, 3.7787E+04, 3.7848E+04, 3.7908E+04,
     + 3.7969E+04, 3.8029E+04, 3.8090E+04, 3.8151E+04, 3.8211E+04,
     +  3.8272E+04, 3.8333E+04, 3.8394E+04, 3.8456E+04, 3.8517E+04,
     + 3.8578E+04, 3.8639E+04, 3.8701E+04, 3.8762E+04, 3.8824E+04,
     +  3.8886E+04, 3.8947E+04, 3.9009E+04, 3.9071E+04, 3.9133E+04,
     + 3.9195E+04, 3.9257E+04, 3.9320E+04, 3.9382E+04, 3.9444E+04,
     +  3.9507E+04, 3.9569E+04, 3.9632E+04, 3.9694E+04, 3.9757E+04,
     + 3.9820E+04, 3.9883E+04, 3.9946E+04, 4.0009E+04, 4.0072E+04,
     +  4.0135E+04, 4.0199E+04, 4.0262E+04, 4.0325E+04, 4.0389E+04,
     + 4.0453E+04, 4.0516E+04, 4.0580E+04, 4.0644E+04, 4.0708E+04,
     +  4.0772E+04, 4.0836E+04, 4.0900E+04, 4.0964E+04, 4.1028E+04,
     + 4.1093E+04, 4.1157E+04, 4.1222E+04, 4.1286E+04, 4.1351E+04,
     +  4.1416E+04, 4.1481E+04, 4.1546E+04, 4.1611E+04, 4.1676E+04,
     + 4.1741E+04, 4.1806E+04, 4.1872E+04, 4.1937E+04, 4.2003E+04,
     +  4.2068E+04, 4.2134E+04, 4.2200E+04, 4.2265E+04, 4.2331E+04,
     + 4.2397E+04, 4.2463E+04, 4.2529E+04, 4.2596E+04, 4.2662E+04,
     +  4.2728E+04, 4.2795E+04, 4.2861E+04, 4.2928E+04, 4.2995E+04,
     + 4.3061E+04, 4.3128E+04, 4.3195E+04, 4.3262E+04, 4.3329E+04,
     +  4.3396E+04, 4.3464E+04, 4.3531E+04, 4.3599E+04, 4.3666E+04,
     + 4.3734E+04, 4.3801E+04, 4.3869E+04, 4.3937E+04, 4.4005E+04,
     +  4.4073E+04, 4.4141E+04, 4.4209E+04, 4.4277E+04, 4.4346E+04,
     + 4.4414E+04, 4.4482E+04, 4.4551E+04, 4.4620E+04, 4.4688E+04,
     +  4.4757E+04, 4.4826E+04, 4.4895E+04, 4.4964E+04, 4.5033E+04,
     + 4.5103E+04, 4.5172E+04, 4.5241E+04, 4.5311E+04, 4.5380E+04,
     +  4.5450E+04, 4.5520E+04, 4.5589E+04, 4.5659E+04, 4.5729E+04,
     + 4.5799E+04, 4.5870E+04, 4.5940E+04, 4.6010E+04, 4.6080E+04,
     +  4.6151E+04, 4.6221E+04, 4.6292E+04, 4.6363E+04, 4.6434E+04,
     + 4.6505E+04, 4.6576E+04, 4.6647E+04, 4.6718E+04, 4.6789E+04,
     +  4.6860E+04, 4.6932E+04, 4.7003E+04, 4.7075E+04, 4.7146E+04,
     + 4.7218E+04, 4.7290E+04, 4.7362E+04, 4.7434E+04, 4.7506E+04,
     +  4.7578E+04, 4.7650E+04, 4.7723E+04, 4.7795E+04, 4.7868E+04,
     + 4.7940E+04, 4.8013E+04, 4.8086E+04, 4.8158E+04, 4.8231E+04,
     +  4.8304E+04, 4.8377E+04, 4.8451E+04, 4.8524E+04, 4.8597E+04,
     + 4.8671E+04, 4.8744E+04, 4.8818E+04, 4.8892E+04, 4.8965E+04,
     +  4.9039E+04, 4.9113E+04, 4.9187E+04, 4.9261E+04, 4.9336E+04,
     + 4.9410E+04, 4.9484E+04, 4.9559E+04, 4.9633E+04, 4.9708E+04,
     +  4.9783E+04, 4.9857E+04, 4.9932E+04, 5.0007E+04, 5.0082E+04,
     + 5.0158E+04, 5.0233E+04, 5.0308E+04, 5.0384E+04, 5.0459E+04,
     +  5.0535E+04, 5.0611E+04, 5.0686E+04, 5.0762E+04, 5.0838E+04,
     + 5.0914E+04, 5.0990E+04, 5.1067E+04, 5.1143E+04, 5.1219E+04,
     +  5.1296E+04, 5.1372E+04, 5.1449E+04, 5.1526E+04, 5.1603E+04,
     + 5.1680E+04, 5.1757E+04, 5.1834E+04, 5.1911E+04, 5.1988E+04,
     +  5.2066E+04, 5.2143E+04, 5.2221E+04, 5.2298E+04, 5.2376E+04,
     + 5.2454E+04, 5.2532E+04, 5.2610E+04, 5.2688E+04, 5.2766E+04,
     +  5.2844E+04, 5.2923E+04, 5.3001E+04, 5.3080E+04, 5.3158E+04,
     + 5.3237E+04, 5.3316E+04, 5.3395E+04, 5.3474E+04, 5.3553E+04,
     +  5.3632E+04, 5.3711E+04, 5.3791E+04, 5.3870E+04, 5.3950E+04,
     + 5.4029E+04, 5.4109E+04, 5.4189E+04, 5.4269E+04, 5.4349E+04,
     +  5.4429E+04, 5.4509E+04, 5.4589E+04, 5.4669E+04, 5.4750E+04,
     + 5.4830E+04, 5.4911E+04, 5.4992E+04, 5.5073E+04, 5.5153E+04,
     +  5.5234E+04, 5.5315E+04, 5.5397E+04, 5.5478E+04, 5.5559E+04,
     + 5.5641E+04, 5.5722E+04, 5.5804E+04, 5.5886E+04, 5.5967E+04,
     +  5.6049E+04, 5.6131E+04, 5.6213E+04, 5.6296E+04, 5.6378E+04,
     + 5.6460E+04, 5.6543E+04, 5.6625E+04, 5.6708E+04, 5.6791E+04,
     +  5.6874E+04, 5.6956E+04, 5.7039E+04, 5.7123E+04, 5.7206E+04,
     + 5.7289E+04, 5.7373E+04, 5.7456E+04, 5.7540E+04, 5.7623E+04,
     +  5.7707E+04, 5.7791E+04, 5.7875E+04, 5.7959E+04, 5.8043E+04,
     + 5.8127E+04, 5.8212E+04, 5.8296E+04, 5.8381E+04, 5.8465E+04,
     +  5.8550E+04, 5.8635E+04, 5.8720E+04, 5.8805E+04, 5.8890E+04,
     + 5.8975E+04, 5.9060E+04, 5.9146E+04, 5.9231E+04, 5.9317E+04,
     +  5.9402E+04, 5.9488E+04, 5.9574E+04, 5.9660E+04, 5.9746E+04,
     + 5.9832E+04, 5.9918E+04, 6.0004E+04, 6.0091E+04, 6.0177E+04,
     +  6.0264E+04, 6.0351E+04, 6.0437E+04, 6.0524E+04, 6.0611E+04,
     + 6.0698E+04, 6.0786E+04, 6.0873E+04, 6.0960E+04, 6.1048E+04,
     +  6.1135E+04, 6.1223E+04, 6.1311E+04, 6.1399E+04, 6.1487E+04,
     + 6.1575E+04, 6.1663E+04, 6.1751E+04, 6.1839E+04, 6.1928E+04,
     +  6.2016E+04, 6.2105E+04, 6.2194E+04, 6.2283E+04, 6.2372E+04,
     + 6.2461E+04, 6.2550E+04, 6.2639E+04, 6.2728E+04, 6.2818E+04,
     +  6.2907E+04, 6.2997E+04, 6.3086E+04, 6.3176E+04, 6.3266E+04,
     + 6.3356E+04, 6.3446E+04, 6.3536E+04, 6.3627E+04, 6.3717E+04,
     +  6.3808E+04, 6.3898E+04, 6.3989E+04, 6.4080E+04, 6.4171E+04,
     + 6.4262E+04, 6.4353E+04, 6.4444E+04, 6.4535E+04, 6.4626E+04,
     +  6.4718E+04, 6.4810E+04, 6.4901E+04, 6.4993E+04, 6.5085E+04,
     + 6.5177E+04, 6.5269E+04, 6.5361E+04, 6.5453E+04, 6.5546E+04,
     +  6.5638E+04, 6.5731E+04, 6.5823E+04, 6.5916E+04, 6.6009E+04,
     + 6.6102E+04, 6.6195E+04, 6.6288E+04, 6.6382E+04, 6.6475E+04,
     +  6.6568E+04, 6.6662E+04, 6.6756E+04, 6.6849E+04, 6.6943E+04,
     + 6.7037E+04, 6.7131E+04, 6.7226E+04, 6.7320E+04, 6.7414E+04,
     +  6.7509E+04, 6.7603E+04, 6.7698E+04, 6.7793E+04, 6.7888E+04/
      data (QofT( 2,J),J=2409,4510)/ 6.7983E+04, 6.8078E+04,
     +  6.8173E+04, 6.8268E+04, 6.8364E+04, 6.8459E+04, 6.8555E+04,
     +  6.8650E+04,
     + 6.8746E+04, 6.8842E+04, 6.8938E+04, 6.9034E+04, 6.9130E+04,
     +  6.9227E+04, 6.9323E+04, 6.9420E+04, 6.9516E+04, 6.9613E+04,
     + 6.9710E+04, 6.9807E+04, 6.9904E+04, 7.0001E+04, 7.0098E+04,
     +  7.0195E+04, 7.0293E+04, 7.0390E+04, 7.0488E+04, 7.0586E+04,
     + 7.0684E+04, 7.0782E+04, 7.0880E+04, 7.0978E+04, 7.1076E+04,
     +  7.1174E+04, 7.1273E+04, 7.1371E+04, 7.1470E+04, 7.1569E+04,
     + 7.1668E+04, 7.1767E+04, 7.1866E+04, 7.1965E+04, 7.2064E+04,
     +  7.2164E+04, 7.2263E+04, 7.2363E+04, 7.2463E+04, 7.2563E+04,
     + 7.2662E+04, 7.2762E+04, 7.2863E+04, 7.2963E+04, 7.3063E+04,
     +  7.3164E+04, 7.3264E+04, 7.3365E+04, 7.3466E+04, 7.3567E+04,
     + 7.3668E+04, 7.3769E+04, 7.3870E+04, 7.3971E+04, 7.4073E+04,
     +  7.4174E+04, 7.4276E+04, 7.4377E+04, 7.4479E+04, 7.4581E+04,
     + 7.4683E+04, 7.4785E+04, 7.4888E+04, 7.4990E+04, 7.5092E+04,
     +  7.5195E+04, 7.5298E+04, 7.5400E+04, 7.5503E+04, 7.5606E+04,
     + 7.5709E+04, 7.5813E+04, 7.5916E+04, 7.6019E+04, 7.6123E+04,
     +  7.6226E+04, 7.6330E+04, 7.6434E+04, 7.6538E+04, 7.6642E+04,
     + 7.6746E+04, 7.6851E+04, 7.6955E+04, 7.7059E+04, 7.7164E+04,
     +  7.7269E+04, 7.7374E+04, 7.7478E+04, 7.7583E+04, 7.7689E+04,
     + 7.7794E+04, 7.7899E+04, 7.8005E+04, 7.8110E+04, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       171
      data (QofT( 3,J),J=1,2408)/ 6.0000E+00, 6.0000E+00, 6.0006E+00,
     +  6.0105E+00, 6.0585E+00, 6.1853E+00, 6.4260E+00, 6.8018E+00,
     + 7.3208E+00, 7.9814E+00, 8.7762E+00, 9.6944E+00, 1.0724E+01,
     +  1.1854E+01, 1.3072E+01, 1.4368E+01, 1.5734E+01, 1.7160E+01,
     + 1.8640E+01, 2.0169E+01, 2.3351E+01, 2.6675E+01, 3.0118E+01,
     +  3.3665E+01, 3.7305E+01, 4.1032E+01, 4.4841E+01, 4.8730E+01,
     + 5.2697E+01, 5.6740E+01, 6.0860E+01, 6.5057E+01, 6.9329E+01,
     +  7.3677E+01, 7.8101E+01, 8.2600E+01, 8.7175E+01, 9.1824E+01,
     + 9.6548E+01, 1.0135E+02, 1.0622E+02, 1.1116E+02, 1.1618E+02,
     +  1.2126E+02, 1.2642E+02, 1.3165E+02, 1.3695E+02, 1.4232E+02,
     + 1.4775E+02, 1.5325E+02, 1.5882E+02, 1.6446E+02, 1.7016E+02,
     +  1.7592E+02, 1.8175E+02, 1.8765E+02, 1.9360E+02, 1.9962E+02,
     + 2.0570E+02, 2.1184E+02, 2.1804E+02, 2.2430E+02, 2.3062E+02,
     +  2.3700E+02, 2.4344E+02, 2.4993E+02, 2.5649E+02, 2.6310E+02,
     + 2.6976E+02, 2.7648E+02, 2.8326E+02, 2.9009E+02, 2.9698E+02,
     +  3.0392E+02, 3.1092E+02, 3.1797E+02, 3.2507E+02, 3.3222E+02,
     + 3.3943E+02, 3.4669E+02, 3.5400E+02, 3.6136E+02, 3.6877E+02,
     +  3.7623E+02, 3.8375E+02, 3.9131E+02, 3.9892E+02, 4.0659E+02,
     + 4.1430E+02, 4.2206E+02, 4.2986E+02, 4.3772E+02, 4.4563E+02,
     +  4.5358E+02, 4.6158E+02, 4.6962E+02, 4.7772E+02, 4.8586E+02,
     + 4.9404E+02, 5.0227E+02, 5.1055E+02, 5.1887E+02, 5.2724E+02,
     +  5.3566E+02, 5.4412E+02, 5.5262E+02, 5.6117E+02, 5.6976E+02,
     + 5.7840E+02, 5.8708E+02, 5.9580E+02, 6.0457E+02, 6.1338E+02,
     +  6.2223E+02, 6.3113E+02, 6.4007E+02, 6.4905E+02, 6.5807E+02,
     + 6.6714E+02, 6.7625E+02, 6.8540E+02, 6.9459E+02, 7.0382E+02,
     +  7.1310E+02, 7.2241E+02, 7.3177E+02, 7.4117E+02, 7.5061E+02,
     + 7.6009E+02, 7.6961E+02, 7.7917E+02, 7.8877E+02, 7.9841E+02,
     +  8.0809E+02, 8.1781E+02, 8.2757E+02, 8.3737E+02, 8.4721E+02,
     + 8.5709E+02, 8.6701E+02, 8.7697E+02, 8.8696E+02, 8.9700E+02,
     +  9.0707E+02, 9.1719E+02, 9.2734E+02, 9.3753E+02, 9.4776E+02,
     + 9.5802E+02, 9.6833E+02, 9.7867E+02, 9.8906E+02, 9.9948E+02,
     +  1.0099E+03, 1.0204E+03, 1.0310E+03, 1.0415E+03, 1.0521E+03,
     + 1.0628E+03, 1.0735E+03, 1.0842E+03, 1.0950E+03, 1.1058E+03,
     +  1.1166E+03, 1.1275E+03, 1.1384E+03, 1.1493E+03, 1.1603E+03,
     + 1.1713E+03, 1.1824E+03, 1.1935E+03, 1.2046E+03, 1.2158E+03,
     +  1.2270E+03, 1.2382E+03, 1.2495E+03, 1.2608E+03, 1.2722E+03,
     + 1.2836E+03, 1.2950E+03, 1.3064E+03, 1.3179E+03, 1.3295E+03,
     +  1.3410E+03, 1.3527E+03, 1.3643E+03, 1.3760E+03, 1.3877E+03,
     + 1.3995E+03, 1.4113E+03, 1.4231E+03, 1.4349E+03, 1.4469E+03,
     +  1.4588E+03, 1.4708E+03, 1.4828E+03, 1.4948E+03, 1.5069E+03,
     + 1.5190E+03, 1.5312E+03, 1.5434E+03, 1.5556E+03, 1.5679E+03,
     +  1.5802E+03, 1.5925E+03, 1.6049E+03, 1.6173E+03, 1.6298E+03,
     + 1.6423E+03, 1.6548E+03, 1.6673E+03, 1.6799E+03, 1.6926E+03,
     +  1.7052E+03, 1.7180E+03, 1.7307E+03, 1.7435E+03, 1.7563E+03,
     + 1.7692E+03, 1.7820E+03, 1.7950E+03, 1.8079E+03, 1.8209E+03,
     +  1.8340E+03, 1.8471E+03, 1.8602E+03, 1.8733E+03, 1.8865E+03,
     + 1.8997E+03, 1.9130E+03, 1.9263E+03, 1.9396E+03, 1.9530E+03,
     +  1.9664E+03, 1.9799E+03, 1.9933E+03, 2.0069E+03, 2.0204E+03,
     + 2.0340E+03, 2.0477E+03, 2.0613E+03, 2.0751E+03, 2.0888E+03,
     +  2.1026E+03, 2.1164E+03, 2.1303E+03, 2.1442E+03, 2.1581E+03,
     + 2.1721E+03, 2.1861E+03, 2.2002E+03, 2.2142E+03, 2.2284E+03,
     +  2.2425E+03, 2.2567E+03, 2.2710E+03, 2.2853E+03, 2.2996E+03,
     + 2.3139E+03, 2.3283E+03, 2.3428E+03, 2.3573E+03, 2.3718E+03,
     +  2.3863E+03, 2.4009E+03, 2.4155E+03, 2.4302E+03, 2.4449E+03,
     + 2.4597E+03, 2.4744E+03, 2.4893E+03, 2.5041E+03, 2.5190E+03,
     +  2.5340E+03, 2.5490E+03, 2.5640E+03, 2.5790E+03, 2.5941E+03,
     + 2.6093E+03, 2.6245E+03, 2.6397E+03, 2.6549E+03, 2.6702E+03,
     +  2.6856E+03, 2.7010E+03, 2.7164E+03, 2.7318E+03, 2.7473E+03,
     + 2.7629E+03, 2.7784E+03, 2.7941E+03, 2.8097E+03, 2.8254E+03,
     +  2.8412E+03, 2.8569E+03, 2.8728E+03, 2.8886E+03, 2.9045E+03,
     + 2.9205E+03, 2.9365E+03, 2.9525E+03, 2.9686E+03, 2.9847E+03,
     +  3.0008E+03, 3.0170E+03, 3.0333E+03, 3.0495E+03, 3.0658E+03,
     + 3.0822E+03, 3.0986E+03, 3.1150E+03, 3.1315E+03, 3.1481E+03,
     +  3.1646E+03, 3.1812E+03, 3.1979E+03, 3.2146E+03, 3.2313E+03,
     + 3.2481E+03, 3.2649E+03, 3.2818E+03, 3.2987E+03, 3.3156E+03,
     +  3.3326E+03, 3.3497E+03, 3.3667E+03, 3.3839E+03, 3.4010E+03,
     + 3.4182E+03, 3.4355E+03, 3.4528E+03, 3.4701E+03, 3.4875E+03,
     +  3.5049E+03, 3.5224E+03, 3.5399E+03, 3.5574E+03, 3.5750E+03,
     + 3.5927E+03, 3.6104E+03, 3.6281E+03, 3.6459E+03, 3.6637E+03,
     +  3.6815E+03, 3.6995E+03, 3.7174E+03, 3.7354E+03, 3.7534E+03,
     + 3.7715E+03, 3.7897E+03, 3.8078E+03, 3.8261E+03, 3.8443E+03,
     +  3.8626E+03, 3.8810E+03, 3.8994E+03, 3.9178E+03, 3.9363E+03,
     + 3.9549E+03, 3.9735E+03, 3.9921E+03, 4.0108E+03, 4.0295E+03,
     +  4.0483E+03, 4.0671E+03, 4.0859E+03, 4.1048E+03, 4.1238E+03,
     + 4.1428E+03, 4.1619E+03, 4.1809E+03, 4.2001E+03, 4.2193E+03,
     +  4.2385E+03, 4.2578E+03, 4.2771E+03, 4.2965E+03, 4.3159E+03,
     + 4.3354E+03, 4.3549E+03, 4.3745E+03, 4.3941E+03, 4.4138E+03,
     +  4.4335E+03, 4.4532E+03, 4.4730E+03, 4.4929E+03, 4.5128E+03,
     + 4.5327E+03, 4.5527E+03, 4.5728E+03, 4.5929E+03, 4.6130E+03,
     +  4.6332E+03, 4.6535E+03, 4.6738E+03, 4.6941E+03, 4.7145E+03,
     + 4.7349E+03, 4.7554E+03, 4.7760E+03, 4.7965E+03, 4.8172E+03,
     +  4.8379E+03, 4.8586E+03, 4.8794E+03, 4.9002E+03, 4.9211E+03,
     + 4.9421E+03, 4.9631E+03, 4.9841E+03, 5.0052E+03, 5.0263E+03,
     +  5.0475E+03, 5.0688E+03, 5.0901E+03, 5.1114E+03, 5.1328E+03,
     + 5.1543E+03, 5.1758E+03, 5.1973E+03, 5.2189E+03, 5.2406E+03,
     +  5.2623E+03, 5.2840E+03, 5.3059E+03, 5.3277E+03, 5.3496E+03,
     + 5.3716E+03, 5.3936E+03, 5.4157E+03, 5.4378E+03, 5.4600E+03,
     +  5.4823E+03, 5.5045E+03, 5.5269E+03, 5.5493E+03, 5.5717E+03,
     + 5.5942E+03, 5.6168E+03, 5.6394E+03, 5.6620E+03, 5.6848E+03,
     +  5.7075E+03, 5.7304E+03, 5.7532E+03, 5.7762E+03, 5.7991E+03,
     + 5.8222E+03, 5.8453E+03, 5.8684E+03, 5.8916E+03, 5.9149E+03,
     +  5.9382E+03, 5.9616E+03, 5.9850E+03, 6.0085E+03, 6.0320E+03,
     + 6.0556E+03, 6.0793E+03, 6.1030E+03, 6.1268E+03, 6.1506E+03,
     +  6.1745E+03, 6.1984E+03, 6.2224E+03, 6.2464E+03, 6.2705E+03,
     + 6.2947E+03, 6.3189E+03, 6.3432E+03, 6.3675E+03, 6.3919E+03,
     +  6.4163E+03, 6.4408E+03, 6.4654E+03, 6.4900E+03, 6.5147E+03,
     + 6.5394E+03, 6.5642E+03, 6.5891E+03, 6.6140E+03, 6.6389E+03,
     +  6.6640E+03, 6.6890E+03, 6.7142E+03, 6.7394E+03, 6.7647E+03,
     + 6.7900E+03, 6.8154E+03, 6.8408E+03, 6.8663E+03, 6.8919E+03,
     +  6.9175E+03, 6.9432E+03, 6.9689E+03, 6.9947E+03, 7.0206E+03,
     + 7.0465E+03, 7.0725E+03, 7.0985E+03, 7.1246E+03, 7.1508E+03,
     +  7.1770E+03, 7.2033E+03, 7.2296E+03, 7.2560E+03, 7.2825E+03,
     + 7.3090E+03, 7.3356E+03, 7.3623E+03, 7.3890E+03, 7.4158E+03,
     +  7.4426E+03, 7.4695E+03, 7.4965E+03, 7.5235E+03, 7.5506E+03,
     + 7.5778E+03, 7.6050E+03, 7.6323E+03, 7.6596E+03, 7.6871E+03,
     +  7.7145E+03, 7.7421E+03, 7.7697E+03, 7.7973E+03, 7.8251E+03,
     + 7.8529E+03, 7.8807E+03, 7.9086E+03, 7.9366E+03, 7.9647E+03,
     +  7.9928E+03, 8.0210E+03, 8.0492E+03, 8.0775E+03, 8.1059E+03,
     + 8.1344E+03, 8.1629E+03, 8.1914E+03, 8.2201E+03, 8.2488E+03,
     +  8.2776E+03, 8.3064E+03, 8.3353E+03, 8.3643E+03, 8.3933E+03,
     + 8.4224E+03, 8.4516E+03, 8.4808E+03, 8.5101E+03, 8.5395E+03,
     +  8.5690E+03, 8.5985E+03, 8.6280E+03, 8.6577E+03, 8.6874E+03,
     + 8.7172E+03, 8.7470E+03, 8.7770E+03, 8.8069E+03, 8.8370E+03,
     +  8.8671E+03, 8.8973E+03, 8.9276E+03, 8.9579E+03, 8.9883E+03,
     + 9.0188E+03, 9.0493E+03, 9.0799E+03, 9.1106E+03, 9.1414E+03,
     +  9.1722E+03, 9.2031E+03, 9.2340E+03, 9.2651E+03, 9.2962E+03,
     + 9.3273E+03, 9.3586E+03, 9.3899E+03, 9.4213E+03, 9.4528E+03,
     +  9.4843E+03, 9.5159E+03, 9.5476E+03, 9.5793E+03, 9.6111E+03,
     + 9.6430E+03, 9.6750E+03, 9.7070E+03, 9.7391E+03, 9.7713E+03,
     +  9.8035E+03, 9.8359E+03, 9.8683E+03, 9.9008E+03, 9.9333E+03,
     + 9.9659E+03, 9.9986E+03, 1.0031E+04, 1.0064E+04, 1.0097E+04,
     +  1.0130E+04, 1.0163E+04, 1.0196E+04, 1.0230E+04, 1.0263E+04,
     + 1.0296E+04, 1.0330E+04, 1.0363E+04, 1.0397E+04, 1.0430E+04,
     +  1.0464E+04, 1.0498E+04, 1.0532E+04, 1.0566E+04, 1.0600E+04,
     + 1.0634E+04, 1.0668E+04, 1.0702E+04, 1.0737E+04, 1.0771E+04,
     +  1.0806E+04, 1.0840E+04, 1.0875E+04, 1.0910E+04, 1.0945E+04,
     + 1.0979E+04, 1.1014E+04, 1.1050E+04, 1.1085E+04, 1.1120E+04,
     +  1.1155E+04, 1.1191E+04, 1.1226E+04, 1.1261E+04, 1.1297E+04,
     + 1.1333E+04, 1.1369E+04, 1.1404E+04, 1.1440E+04, 1.1476E+04,
     +  1.1512E+04, 1.1549E+04, 1.1585E+04, 1.1621E+04, 1.1658E+04,
     + 1.1694E+04, 1.1731E+04, 1.1767E+04, 1.1804E+04, 1.1841E+04,
     +  1.1878E+04, 1.1915E+04, 1.1952E+04, 1.1989E+04, 1.2026E+04,
     + 1.2063E+04, 1.2101E+04, 1.2138E+04, 1.2176E+04, 1.2213E+04,
     +  1.2251E+04, 1.2289E+04, 1.2327E+04, 1.2365E+04, 1.2403E+04,
     + 1.2441E+04, 1.2479E+04, 1.2518E+04, 1.2556E+04, 1.2594E+04,
     +  1.2633E+04, 1.2672E+04, 1.2710E+04, 1.2749E+04, 1.2788E+04,
     + 1.2827E+04, 1.2866E+04, 1.2905E+04, 1.2944E+04, 1.2984E+04,
     +  1.3023E+04, 1.3063E+04, 1.3102E+04, 1.3142E+04, 1.3182E+04,
     + 1.3222E+04, 1.3262E+04, 1.3302E+04, 1.3342E+04, 1.3382E+04,
     +  1.3422E+04, 1.3463E+04, 1.3503E+04, 1.3544E+04, 1.3584E+04,
     + 1.3625E+04, 1.3666E+04, 1.3707E+04, 1.3748E+04, 1.3789E+04,
     +  1.3830E+04, 1.3871E+04, 1.3912E+04, 1.3954E+04, 1.3995E+04,
     + 1.4037E+04, 1.4079E+04, 1.4121E+04, 1.4162E+04, 1.4204E+04,
     +  1.4247E+04, 1.4289E+04, 1.4331E+04, 1.4373E+04, 1.4416E+04,
     + 1.4458E+04, 1.4501E+04, 1.4544E+04, 1.4586E+04, 1.4629E+04,
     +  1.4672E+04, 1.4715E+04, 1.4759E+04, 1.4802E+04, 1.4845E+04,
     + 1.4889E+04, 1.4932E+04, 1.4976E+04, 1.5020E+04, 1.5063E+04,
     +  1.5107E+04, 1.5151E+04, 1.5195E+04, 1.5240E+04, 1.5284E+04,
     + 1.5328E+04, 1.5373E+04, 1.5417E+04, 1.5462E+04, 1.5507E+04,
     +  1.5552E+04, 1.5597E+04, 1.5642E+04, 1.5687E+04, 1.5732E+04,
     + 1.5778E+04, 1.5823E+04, 1.5869E+04, 1.5914E+04, 1.5960E+04,
     +  1.6006E+04, 1.6052E+04, 1.6098E+04, 1.6144E+04, 1.6190E+04,
     + 1.6237E+04, 1.6283E+04, 1.6330E+04, 1.6376E+04, 1.6423E+04,
     +  1.6470E+04, 1.6517E+04, 1.6564E+04, 1.6611E+04, 1.6658E+04,
     + 1.6705E+04, 1.6753E+04, 1.6800E+04, 1.6848E+04, 1.6896E+04,
     +  1.6943E+04, 1.6991E+04, 1.7039E+04, 1.7088E+04, 1.7136E+04,
     + 1.7184E+04, 1.7233E+04, 1.7281E+04, 1.7330E+04, 1.7379E+04,
     +  1.7427E+04, 1.7476E+04, 1.7525E+04, 1.7575E+04, 1.7624E+04,
     + 1.7673E+04, 1.7723E+04, 1.7772E+04, 1.7822E+04, 1.7872E+04,
     +  1.7921E+04, 1.7971E+04, 1.8022E+04, 1.8072E+04, 1.8122E+04,
     + 1.8172E+04, 1.8223E+04, 1.8274E+04, 1.8324E+04, 1.8375E+04,
     +  1.8426E+04, 1.8477E+04, 1.8528E+04, 1.8579E+04, 1.8631E+04,
     + 1.8682E+04, 1.8734E+04, 1.8786E+04, 1.8837E+04, 1.8889E+04,
     +  1.8941E+04, 1.8993E+04, 1.9046E+04, 1.9098E+04, 1.9150E+04,
     + 1.9203E+04, 1.9256E+04, 1.9308E+04, 1.9361E+04, 1.9414E+04,
     +  1.9467E+04, 1.9520E+04, 1.9574E+04, 1.9627E+04, 1.9681E+04,
     + 1.9734E+04, 1.9788E+04, 1.9842E+04, 1.9896E+04, 1.9950E+04,
     +  2.0004E+04, 2.0058E+04, 2.0113E+04, 2.0167E+04, 2.0222E+04,
     + 2.0277E+04, 2.0332E+04, 2.0387E+04, 2.0442E+04, 2.0497E+04,
     +  2.0552E+04, 2.0608E+04, 2.0663E+04, 2.0719E+04, 2.0775E+04,
     + 2.0831E+04, 2.0887E+04, 2.0943E+04, 2.0999E+04, 2.1055E+04,
     +  2.1112E+04, 2.1168E+04, 2.1225E+04, 2.1282E+04, 2.1339E+04,
     + 2.1396E+04, 2.1453E+04, 2.1510E+04, 2.1567E+04, 2.1625E+04,
     +  2.1683E+04, 2.1740E+04, 2.1798E+04, 2.1856E+04, 2.1914E+04,
     + 2.1972E+04, 2.2031E+04, 2.2089E+04, 2.2148E+04, 2.2206E+04,
     +  2.2265E+04, 2.2324E+04, 2.2383E+04, 2.2442E+04, 2.2502E+04,
     + 2.2561E+04, 2.2620E+04, 2.2680E+04, 2.2740E+04, 2.2800E+04,
     +  2.2860E+04, 2.2920E+04, 2.2980E+04, 2.3040E+04, 2.3101E+04,
     + 2.3161E+04, 2.3222E+04, 2.3283E+04, 2.3344E+04, 2.3405E+04,
     +  2.3466E+04, 2.3527E+04, 2.3589E+04, 2.3650E+04, 2.3712E+04,
     + 2.3774E+04, 2.3836E+04, 2.3898E+04, 2.3960E+04, 2.4022E+04,
     +  2.4085E+04, 2.4147E+04, 2.4210E+04, 2.4273E+04, 2.4336E+04,
     + 2.4399E+04, 2.4462E+04, 2.4525E+04, 2.4589E+04, 2.4652E+04,
     +  2.4716E+04, 2.4780E+04, 2.4844E+04, 2.4908E+04, 2.4972E+04,
     + 2.5036E+04, 2.5101E+04, 2.5165E+04, 2.5230E+04, 2.5295E+04,
     +  2.5360E+04, 2.5425E+04, 2.5490E+04, 2.5556E+04, 2.5621E+04,
     + 2.5687E+04, 2.5752E+04, 2.5818E+04, 2.5884E+04, 2.5950E+04,
     +  2.6017E+04, 2.6083E+04, 2.6150E+04, 2.6216E+04, 2.6283E+04,
     + 2.6350E+04, 2.6417E+04, 2.6484E+04, 2.6551E+04, 2.6619E+04,
     +  2.6686E+04, 2.6754E+04, 2.6822E+04, 2.6890E+04, 2.6958E+04,
     + 2.7026E+04, 2.7094E+04, 2.7163E+04, 2.7232E+04, 2.7300E+04,
     +  2.7369E+04, 2.7438E+04, 2.7507E+04, 2.7577E+04, 2.7646E+04,
     + 2.7716E+04, 2.7785E+04, 2.7855E+04, 2.7925E+04, 2.7995E+04,
     +  2.8066E+04, 2.8136E+04, 2.8206E+04, 2.8277E+04, 2.8348E+04,
     + 2.8419E+04, 2.8490E+04, 2.8561E+04, 2.8632E+04, 2.8704E+04,
     +  2.8776E+04, 2.8847E+04, 2.8919E+04, 2.8991E+04, 2.9063E+04,
     + 2.9136E+04, 2.9208E+04, 2.9281E+04, 2.9353E+04, 2.9426E+04,
     +  2.9499E+04, 2.9572E+04, 2.9646E+04, 2.9719E+04, 2.9793E+04,
     + 2.9866E+04, 2.9940E+04, 3.0014E+04, 3.0088E+04, 3.0163E+04,
     +  3.0237E+04, 3.0312E+04, 3.0386E+04, 3.0461E+04, 3.0536E+04,
     + 3.0611E+04, 3.0687E+04, 3.0762E+04, 3.0838E+04, 3.0913E+04,
     +  3.0989E+04, 3.1065E+04, 3.1141E+04, 3.1217E+04, 3.1294E+04,
     + 3.1370E+04, 3.1447E+04, 3.1524E+04, 3.1601E+04, 3.1678E+04,
     +  3.1756E+04, 3.1833E+04, 3.1911E+04, 3.1988E+04, 3.2066E+04,
     + 3.2144E+04, 3.2222E+04, 3.2301E+04, 3.2379E+04, 3.2458E+04,
     +  3.2537E+04, 3.2616E+04, 3.2695E+04, 3.2774E+04, 3.2853E+04,
     + 3.2933E+04, 3.3012E+04, 3.3092E+04, 3.3172E+04, 3.3252E+04,
     +  3.3333E+04, 3.3413E+04, 3.3494E+04, 3.3574E+04, 3.3655E+04,
     + 3.3736E+04, 3.3817E+04, 3.3899E+04, 3.3980E+04, 3.4062E+04,
     +  3.4144E+04, 3.4226E+04, 3.4308E+04, 3.4390E+04, 3.4472E+04,
     + 3.4555E+04, 3.4638E+04, 3.4720E+04, 3.4803E+04, 3.4887E+04,
     +  3.4970E+04, 3.5053E+04, 3.5137E+04, 3.5221E+04, 3.5305E+04,
     + 3.5389E+04, 3.5473E+04, 3.5558E+04, 3.5642E+04, 3.5727E+04,
     +  3.5812E+04, 3.5897E+04, 3.5982E+04, 3.6067E+04, 3.6153E+04,
     + 3.6239E+04, 3.6324E+04, 3.6410E+04, 3.6497E+04, 3.6583E+04,
     +  3.6669E+04, 3.6756E+04, 3.6843E+04, 3.6930E+04, 3.7017E+04,
     + 3.7104E+04, 3.7191E+04, 3.7279E+04, 3.7367E+04, 3.7455E+04,
     +  3.7543E+04, 3.7631E+04, 3.7719E+04, 3.7808E+04, 3.7897E+04,
     + 3.7986E+04, 3.8075E+04, 3.8164E+04, 3.8253E+04, 3.8343E+04,
     +  3.8432E+04, 3.8522E+04, 3.8612E+04, 3.8703E+04, 3.8793E+04,
     + 3.8883E+04, 3.8974E+04, 3.9065E+04, 3.9156E+04, 3.9247E+04,
     +  3.9339E+04, 3.9430E+04, 3.9522E+04, 3.9614E+04, 3.9706E+04,
     + 3.9798E+04, 3.9890E+04, 3.9983E+04, 4.0075E+04, 4.0168E+04,
     +  4.0261E+04, 4.0354E+04, 4.0448E+04, 4.0541E+04, 4.0635E+04,
     + 4.0729E+04, 4.0823E+04, 4.0917E+04, 4.1011E+04, 4.1106E+04,
     +  4.1201E+04, 4.1296E+04, 4.1391E+04, 4.1486E+04, 4.1581E+04,
     + 4.1677E+04, 4.1773E+04, 4.1869E+04, 4.1965E+04, 4.2061E+04,
     +  4.2157E+04, 4.2254E+04, 4.2351E+04, 4.2448E+04, 4.2545E+04,
     + 4.2642E+04, 4.2740E+04, 4.2837E+04, 4.2935E+04, 4.3033E+04,
     +  4.3131E+04, 4.3230E+04, 4.3328E+04, 4.3427E+04, 4.3526E+04,
     + 4.3625E+04, 4.3724E+04, 4.3823E+04, 4.3923E+04, 4.4023E+04,
     +  4.4123E+04, 4.4223E+04, 4.4323E+04, 4.4424E+04, 4.4524E+04,
     + 4.4625E+04, 4.4726E+04, 4.4827E+04, 4.4929E+04, 4.5030E+04,
     +  4.5132E+04, 4.5234E+04, 4.5336E+04, 4.5438E+04, 4.5541E+04,
     + 4.5643E+04, 4.5746E+04, 4.5849E+04, 4.5952E+04, 4.6056E+04,
     +  4.6159E+04, 4.6263E+04, 4.6367E+04, 4.6471E+04, 4.6575E+04,
     + 4.6680E+04, 4.6784E+04, 4.6889E+04, 4.6994E+04, 4.7099E+04,
     +  4.7205E+04, 4.7310E+04, 4.7416E+04, 4.7522E+04, 4.7628E+04,
     + 4.7734E+04, 4.7841E+04, 4.7948E+04, 4.8054E+04, 4.8162E+04,
     +  4.8269E+04, 4.8376E+04, 4.8484E+04, 4.8592E+04, 4.8700E+04,
     + 4.8808E+04, 4.8916E+04, 4.9025E+04, 4.9133E+04, 4.9242E+04,
     +  4.9352E+04, 4.9461E+04, 4.9570E+04, 4.9680E+04, 4.9790E+04,
     + 4.9900E+04, 5.0010E+04, 5.0121E+04, 5.0231E+04, 5.0342E+04,
     +  5.0453E+04, 5.0565E+04, 5.0676E+04, 5.0788E+04, 5.0899E+04,
     + 5.1011E+04, 5.1124E+04, 5.1236E+04, 5.1349E+04, 5.1461E+04,
     +  5.1574E+04, 5.1688E+04, 5.1801E+04, 5.1914E+04, 5.2028E+04,
     + 5.2142E+04, 5.2256E+04, 5.2371E+04, 5.2485E+04, 5.2600E+04,
     +  5.2715E+04, 5.2830E+04, 5.2945E+04, 5.3061E+04, 5.3177E+04,
     + 5.3293E+04, 5.3409E+04, 5.3525E+04, 5.3641E+04, 5.3758E+04,
     +  5.3875E+04, 5.3992E+04, 5.4110E+04, 5.4227E+04, 5.4345E+04,
     + 5.4463E+04, 5.4581E+04, 5.4699E+04, 5.4818E+04, 5.4936E+04,
     +  5.5055E+04, 5.5175E+04, 5.5294E+04, 5.5413E+04, 5.5533E+04,
     + 5.5653E+04, 5.5773E+04, 5.5894E+04, 5.6014E+04, 5.6135E+04,
     +  5.6256E+04, 5.6377E+04, 5.6498E+04, 5.6620E+04, 5.6742E+04,
     + 5.6864E+04, 5.6986E+04, 5.7108E+04, 5.7231E+04, 5.7354E+04,
     +  5.7477E+04, 5.7600E+04, 5.7724E+04, 5.7847E+04, 5.7971E+04,
     + 5.8095E+04, 5.8219E+04, 5.8344E+04, 5.8469E+04, 5.8594E+04,
     +  5.8719E+04, 5.8844E+04, 5.8970E+04, 5.9095E+04, 5.9221E+04,
     + 5.9347E+04, 5.9474E+04, 5.9600E+04, 5.9727E+04, 5.9854E+04,
     +  5.9982E+04, 6.0109E+04, 6.0237E+04, 6.0365E+04, 6.0493E+04,
     + 6.0621E+04, 6.0749E+04, 6.0878E+04, 6.1007E+04, 6.1136E+04,
     +  6.1266E+04, 6.1395E+04, 6.1525E+04, 6.1655E+04, 6.1785E+04,
     + 6.1916E+04, 6.2046E+04, 6.2177E+04, 6.2308E+04, 6.2440E+04,
     +  6.2571E+04, 6.2703E+04, 6.2835E+04, 6.2967E+04, 6.3100E+04,
     + 6.3232E+04, 6.3365E+04, 6.3498E+04, 6.3632E+04, 6.3765E+04,
     +  6.3899E+04, 6.4033E+04, 6.4167E+04, 6.4301E+04, 6.4436E+04,
     + 6.4571E+04, 6.4706E+04, 6.4841E+04, 6.4977E+04, 6.5113E+04,
     +  6.5249E+04, 6.5385E+04, 6.5521E+04, 6.5658E+04, 6.5795E+04,
     + 6.5932E+04, 6.6069E+04, 6.6207E+04, 6.6344E+04, 6.6482E+04,
     +  6.6621E+04, 6.6759E+04, 6.6898E+04, 6.7037E+04, 6.7176E+04,
     + 6.7315E+04, 6.7455E+04, 6.7594E+04, 6.7734E+04, 6.7875E+04,
     +  6.8015E+04, 6.8156E+04, 6.8297E+04, 6.8438E+04, 6.8579E+04,
     + 6.8721E+04, 6.8863E+04, 6.9005E+04, 6.9147E+04, 6.9290E+04,
     +  6.9433E+04, 6.9576E+04, 6.9719E+04, 6.9863E+04, 7.0006E+04,
     + 7.0150E+04, 7.0294E+04, 7.0439E+04, 7.0584E+04, 7.0728E+04,
     +  7.0874E+04, 7.1019E+04, 7.1165E+04, 7.1310E+04, 7.1456E+04,
     + 7.1603E+04, 7.1749E+04, 7.1896E+04, 7.2043E+04, 7.2190E+04,
     +  7.2338E+04, 7.2486E+04, 7.2634E+04, 7.2782E+04, 7.2930E+04,
     + 7.3079E+04, 7.3228E+04, 7.3377E+04, 7.3526E+04, 7.3676E+04,
     +  7.3826E+04, 7.3976E+04, 7.4126E+04, 7.4277E+04, 7.4428E+04,
     + 7.4579E+04, 7.4730E+04, 7.4882E+04, 7.5034E+04, 7.5186E+04,
     +  7.5338E+04, 7.5491E+04, 7.5643E+04, 7.5796E+04, 7.5950E+04,
     + 7.6103E+04, 7.6257E+04, 7.6411E+04, 7.6565E+04, 7.6720E+04,
     +  7.6875E+04, 7.7030E+04, 7.7185E+04, 7.7340E+04, 7.7496E+04,
     + 7.7652E+04, 7.7808E+04, 7.7965E+04, 7.8121E+04, 7.8278E+04,
     +  7.8436E+04, 7.8593E+04, 7.8751E+04, 7.8909E+04, 7.9067E+04,
     + 7.9225E+04, 7.9384E+04, 7.9543E+04, 7.9702E+04, 7.9862E+04,
     +  8.0022E+04, 8.0182E+04, 8.0342E+04, 8.0502E+04, 8.0663E+04,
     + 8.0824E+04, 8.0985E+04, 8.1147E+04, 8.1309E+04, 8.1471E+04,
     +  8.1633E+04, 8.1795E+04, 8.1958E+04, 8.2121E+04, 8.2284E+04,
     + 8.2448E+04, 8.2612E+04, 8.2776E+04, 8.2940E+04, 8.3105E+04,
     +  8.3270E+04, 8.3435E+04, 8.3600E+04, 8.3766E+04, 8.3931E+04,
     + 8.4098E+04, 8.4264E+04, 8.4431E+04, 8.4597E+04, 8.4765E+04,
     +  8.4932E+04, 8.5100E+04, 8.5268E+04, 8.5436E+04, 8.5604E+04,
     + 8.5773E+04, 8.5942E+04, 8.6111E+04, 8.6281E+04, 8.6451E+04,
     +  8.6621E+04, 8.6791E+04, 8.6962E+04, 8.7132E+04, 8.7303E+04,
     + 8.7475E+04, 8.7646E+04, 8.7818E+04, 8.7991E+04, 8.8163E+04,
     +  8.8336E+04, 8.8509E+04, 8.8682E+04, 8.8855E+04, 8.9029E+04,
     + 8.9203E+04, 8.9377E+04, 8.9552E+04, 8.9727E+04, 8.9902E+04,
     +  9.0077E+04, 9.0253E+04, 9.0429E+04, 9.0605E+04, 9.0782E+04,
     + 9.0958E+04, 9.1135E+04, 9.1313E+04, 9.1490E+04, 9.1668E+04,
     +  9.1846E+04, 9.2024E+04, 9.2203E+04, 9.2382E+04, 9.2561E+04,
     + 9.2741E+04, 9.2920E+04, 9.3100E+04, 9.3281E+04, 9.3461E+04,
     +  9.3642E+04, 9.3823E+04, 9.4005E+04, 9.4186E+04, 9.4368E+04,
     + 9.4551E+04, 9.4733E+04, 9.4916E+04, 9.5099E+04, 9.5282E+04,
     +  9.5466E+04, 9.5650E+04, 9.5834E+04, 9.6018E+04, 9.6203E+04,
     + 9.6388E+04, 9.6573E+04, 9.6759E+04, 9.6945E+04, 9.7131E+04,
     +  9.7317E+04, 9.7504E+04, 9.7691E+04, 9.7878E+04, 9.8066E+04,
     + 9.8254E+04, 9.8442E+04, 9.8630E+04, 9.8819E+04, 9.9008E+04,
     +  9.9197E+04, 9.9387E+04, 9.9577E+04, 9.9767E+04, 9.9957E+04,
     + 1.0015E+05, 1.0034E+05, 1.0053E+05, 1.0072E+05, 1.0091E+05,
     +  1.0111E+05, 1.0130E+05, 1.0149E+05, 1.0168E+05, 1.0188E+05,
     + 1.0207E+05, 1.0226E+05, 1.0246E+05, 1.0265E+05, 1.0285E+05,
     +  1.0304E+05, 1.0324E+05, 1.0343E+05, 1.0363E+05, 1.0383E+05,
     + 1.0402E+05, 1.0422E+05, 1.0442E+05, 1.0461E+05, 1.0481E+05,
     +  1.0501E+05, 1.0521E+05, 1.0541E+05, 1.0561E+05, 1.0580E+05,
     + 1.0600E+05, 1.0620E+05, 1.0640E+05, 1.0660E+05, 1.0680E+05,
     +  1.0701E+05, 1.0721E+05, 1.0741E+05, 1.0761E+05, 1.0781E+05,
     + 1.0801E+05, 1.0822E+05, 1.0842E+05, 1.0862E+05, 1.0883E+05,
     +  1.0903E+05, 1.0924E+05, 1.0944E+05, 1.0964E+05, 1.0985E+05,
     + 1.1006E+05, 1.1026E+05, 1.1047E+05, 1.1067E+05, 1.1088E+05,
     +  1.1109E+05, 1.1129E+05, 1.1150E+05, 1.1171E+05, 1.1192E+05,
     + 1.1213E+05, 1.1234E+05, 1.1254E+05, 1.1275E+05, 1.1296E+05,
     +  1.1317E+05, 1.1338E+05, 1.1359E+05, 1.1381E+05, 1.1402E+05,
     + 1.1423E+05, 1.1444E+05, 1.1465E+05, 1.1487E+05, 1.1508E+05,
     +  1.1529E+05, 1.1550E+05, 1.1572E+05, 1.1593E+05, 1.1615E+05,
     + 1.1636E+05, 1.1658E+05, 1.1679E+05, 1.1701E+05, 1.1722E+05,
     +  1.1744E+05, 1.1766E+05, 1.1787E+05, 1.1809E+05, 1.1831E+05,
     + 1.1853E+05, 1.1874E+05, 1.1896E+05, 1.1918E+05, 1.1940E+05,
     +  1.1962E+05, 1.1984E+05, 1.2006E+05, 1.2028E+05, 1.2050E+05,
     + 1.2072E+05, 1.2094E+05, 1.2116E+05, 1.2139E+05, 1.2161E+05,
     +  1.2183E+05, 1.2205E+05, 1.2228E+05, 1.2250E+05, 1.2272E+05,
     + 1.2295E+05, 1.2317E+05, 1.2340E+05, 1.2362E+05, 1.2385E+05,
     +  1.2407E+05, 1.2430E+05, 1.2453E+05, 1.2475E+05, 1.2498E+05,
     + 1.2521E+05, 1.2544E+05, 1.2566E+05, 1.2589E+05, 1.2612E+05,
     +  1.2635E+05, 1.2658E+05, 1.2681E+05, 1.2704E+05, 1.2727E+05,
     + 1.2750E+05, 1.2773E+05, 1.2796E+05, 1.2820E+05, 1.2843E+05,
     +  1.2866E+05, 1.2889E+05, 1.2913E+05, 1.2936E+05, 1.2959E+05,
     + 1.2983E+05, 1.3006E+05, 1.3030E+05, 1.3053E+05, 1.3077E+05,
     +  1.3100E+05, 1.3124E+05, 1.3148E+05, 1.3171E+05, 1.3195E+05,
     + 1.3219E+05, 1.3242E+05, 1.3266E+05, 1.3290E+05, 1.3314E+05,
     +  1.3338E+05, 1.3362E+05, 1.3386E+05, 1.3410E+05, 1.3434E+05,
     + 1.3458E+05, 1.3482E+05, 1.3506E+05, 1.3530E+05, 1.3555E+05,
     +  1.3579E+05, 1.3603E+05, 1.3627E+05, 1.3652E+05, 1.3676E+05,
     + 1.3701E+05, 1.3725E+05, 1.3750E+05, 1.3774E+05, 1.3799E+05,
     +  1.3823E+05, 1.3848E+05, 1.3872E+05, 1.3897E+05, 1.3922E+05,
     + 1.3947E+05, 1.3971E+05, 1.3996E+05, 1.4021E+05, 1.4046E+05,
     +  1.4071E+05, 1.4096E+05, 1.4121E+05, 1.4146E+05, 1.4171E+05,
     + 1.4196E+05, 1.4221E+05, 1.4247E+05, 1.4272E+05, 1.4297E+05,
     +  1.4322E+05, 1.4348E+05, 1.4373E+05, 1.4398E+05, 1.4424E+05,
     + 1.4449E+05, 1.4475E+05, 1.4500E+05, 1.4526E+05, 1.4552E+05,
     +  1.4577E+05, 1.4603E+05, 1.4629E+05, 1.4654E+05, 1.4680E+05,
     + 1.4706E+05, 1.4732E+05, 1.4758E+05, 1.4784E+05, 1.4810E+05,
     +  1.4836E+05, 1.4862E+05, 1.4888E+05, 1.4914E+05, 1.4940E+05,
     + 1.4966E+05, 1.4992E+05, 1.5019E+05, 1.5045E+05, 1.5071E+05,
     +  1.5098E+05, 1.5124E+05, 1.5150E+05, 1.5177E+05, 1.5203E+05,
     + 1.5230E+05, 1.5257E+05, 1.5283E+05, 1.5310E+05, 1.5336E+05,
     +  1.5363E+05, 1.5390E+05, 1.5417E+05, 1.5444E+05, 1.5470E+05,
     + 1.5497E+05, 1.5524E+05, 1.5551E+05, 1.5578E+05, 1.5605E+05,
     +  1.5633E+05, 1.5660E+05, 1.5687E+05, 1.5714E+05, 1.5741E+05,
     + 1.5769E+05, 1.5796E+05, 1.5823E+05, 1.5851E+05, 1.5878E+05,
     +  1.5906E+05, 1.5933E+05, 1.5961E+05, 1.5988E+05, 1.6016E+05,
     + 1.6043E+05, 1.6071E+05, 1.6099E+05, 1.6127E+05, 1.6154E+05,
     +  1.6182E+05, 1.6210E+05, 1.6238E+05, 1.6266E+05, 1.6294E+05,
     + 1.6322E+05, 1.6350E+05, 1.6378E+05, 1.6406E+05, 1.6435E+05,
     +  1.6463E+05, 1.6491E+05, 1.6519E+05, 1.6548E+05, 1.6576E+05,
     + 1.6604E+05, 1.6633E+05, 1.6661E+05, 1.6690E+05, 1.6718E+05,
     +  1.6747E+05, 1.6776E+05, 1.6804E+05, 1.6833E+05, 1.6862E+05,
     + 1.6891E+05, 1.6919E+05, 1.6948E+05, 1.6977E+05, 1.7006E+05,
     +  1.7035E+05, 1.7064E+05, 1.7093E+05, 1.7122E+05, 1.7152E+05,
     + 1.7181E+05, 1.7210E+05, 1.7239E+05, 1.7269E+05, 1.7298E+05,
     +  1.7327E+05, 1.7357E+05, 1.7386E+05, 1.7416E+05, 1.7445E+05,
     + 1.7475E+05, 1.7504E+05, 1.7534E+05, 1.7564E+05, 1.7593E+05,
     +  1.7623E+05, 1.7653E+05, 1.7683E+05, 1.7713E+05, 1.7743E+05,
     + 1.7773E+05, 1.7803E+05, 1.7833E+05, 1.7863E+05, 1.7893E+05,
     +  1.7923E+05, 1.7953E+05, 1.7984E+05, 1.8014E+05, 1.8044E+05,
     + 1.8075E+05, 1.8105E+05, 1.8135E+05, 1.8166E+05, 1.8196E+05,
     +  1.8227E+05, 1.8258E+05, 1.8288E+05, 1.8319E+05, 1.8350E+05,
     + 1.8381E+05, 1.8411E+05, 1.8442E+05, 1.8473E+05, 1.8504E+05,
     +  1.8535E+05, 1.8566E+05, 1.8597E+05, 1.8628E+05, 1.8659E+05,
     + 1.8690E+05, 1.8722E+05, 1.8753E+05, 1.8784E+05, 1.8816E+05,
     +  1.8847E+05, 1.8878E+05, 1.8910E+05, 1.8941E+05, 1.8973E+05,
     + 1.9005E+05, 1.9036E+05, 1.9068E+05, 1.9100E+05, 1.9131E+05,
     +  1.9163E+05, 1.9195E+05, 1.9227E+05, 1.9259E+05, 1.9291E+05,
     + 1.9323E+05, 1.9355E+05, 1.9387E+05, 1.9419E+05, 1.9451E+05,
     +  1.9483E+05, 1.9516E+05, 1.9548E+05, 1.9580E+05, 1.9613E+05,
     + 1.9645E+05, 1.9678E+05, 1.9710E+05, 1.9743E+05, 1.9775E+05,
     +  1.9808E+05, 1.9840E+05, 1.9873E+05, 1.9906E+05, 1.9939E+05,
     + 1.9972E+05, 2.0004E+05, 2.0037E+05, 2.0070E+05, 2.0103E+05,
     +  2.0136E+05, 2.0170E+05, 2.0203E+05, 2.0236E+05, 2.0269E+05,
     + 2.0302E+05, 2.0336E+05, 2.0369E+05, 2.0402E+05, 2.0436E+05,
     +  2.0469E+05, 2.0503E+05, 2.0536E+05, 2.0570E+05, 2.0604E+05,
     + 2.0637E+05, 2.0671E+05, 2.0705E+05, 2.0739E+05, 2.0773E+05,
     +  2.0807E+05, 2.0841E+05, 2.0875E+05, 2.0909E+05, 2.0943E+05,
     + 2.0977E+05, 2.1011E+05, 2.1045E+05, 2.1079E+05, 2.1114E+05,
     +  2.1148E+05, 2.1183E+05, 2.1217E+05, 2.1251E+05, 2.1286E+05,
     + 2.1321E+05, 2.1355E+05, 2.1390E+05, 2.1424E+05, 2.1459E+05,
     +  2.1494E+05, 2.1529E+05, 2.1564E+05, 2.1599E+05, 2.1634E+05,
     + 2.1669E+05, 2.1704E+05, 2.1739E+05, 2.1774E+05, 2.1809E+05,
     +  2.1844E+05, 2.1880E+05, 2.1915E+05, 2.1950E+05, 2.1986E+05,
     + 2.2021E+05, 2.2057E+05, 2.2092E+05, 2.2128E+05, 2.2163E+05,
     +  2.2199E+05, 2.2235E+05, 2.2271E+05, 2.2306E+05, 2.2342E+05,
     + 2.2378E+05, 2.2414E+05, 2.2450E+05, 2.2486E+05, 2.2522E+05,
     +  2.2558E+05, 2.2595E+05, 2.2631E+05, 2.2667E+05, 2.2703E+05,
     + 2.2740E+05, 2.2776E+05, 2.2813E+05, 2.2849E+05, 2.2886E+05,
     +  2.2922E+05, 2.2959E+05, 2.2995E+05, 2.3032E+05, 2.3069E+05,
     + 2.3106E+05, 2.3143E+05, 2.3179E+05, 2.3216E+05, 2.3253E+05,
     +  2.3290E+05, 2.3328E+05, 2.3365E+05, 2.3402E+05, 2.3439E+05,
     + 2.3476E+05, 2.3514E+05, 2.3551E+05, 2.3588E+05, 2.3626E+05,
     +  2.3663E+05, 2.3701E+05, 2.3739E+05, 2.3776E+05, 2.3814E+05,
     + 2.3852E+05, 2.3889E+05, 2.3927E+05, 2.3965E+05, 2.4003E+05,
     +  2.4041E+05, 2.4079E+05, 2.4117E+05, 2.4155E+05, 2.4193E+05,
     + 2.4232E+05, 2.4270E+05, 2.4308E+05, 2.4346E+05, 2.4385E+05,
     +  2.4423E+05, 2.4462E+05, 2.4500E+05, 2.4539E+05, 2.4578E+05,
     + 2.4616E+05, 2.4655E+05, 2.4694E+05, 2.4732E+05, 2.4771E+05,
     +  2.4810E+05, 2.4849E+05, 2.4888E+05, 2.4927E+05, 2.4966E+05,
     + 2.5006E+05, 2.5045E+05, 2.5084E+05, 2.5123E+05, 2.5163E+05,
     +  2.5202E+05, 2.5241E+05, 2.5281E+05, 2.5320E+05, 2.5360E+05,
     + 2.5400E+05, 2.5439E+05, 2.5479E+05, 2.5519E+05, 2.5559E+05,
     +  2.5599E+05, 2.5638E+05, 2.5678E+05, 2.5718E+05, 2.5759E+05,
     + 2.5799E+05, 2.5839E+05, 2.5879E+05, 2.5919E+05, 2.5960E+05,
     +  2.6000E+05, 2.6040E+05, 2.6081E+05, 2.6121E+05, 2.6162E+05,
     + 2.6202E+05, 2.6243E+05, 2.6284E+05, 2.6325E+05, 2.6365E+05,
     +  2.6406E+05, 2.6447E+05, 2.6488E+05, 2.6529E+05, 2.6570E+05,
     + 2.6611E+05, 2.6652E+05, 2.6694E+05, 2.6735E+05, 2.6776E+05,
     +  2.6817E+05, 2.6859E+05, 2.6900E+05, 2.6942E+05, 2.6983E+05,
     + 2.7025E+05, 2.7066E+05, 2.7108E+05, 2.7150E+05, 2.7192E+05,
     +  2.7234E+05, 2.7275E+05, 2.7317E+05, 2.7359E+05, 2.7401E+05,
     + 2.7443E+05, 2.7486E+05, 2.7528E+05, 2.7570E+05, 2.7612E+05,
     +  2.7655E+05, 2.7697E+05, 2.7739E+05, 2.7782E+05, 2.7824E+05,
     + 2.7867E+05, 2.7910E+05, 2.7952E+05, 2.7995E+05, 2.8038E+05,
     +  2.8081E+05, 2.8124E+05, 2.8167E+05, 2.8210E+05, 2.8253E+05,
     + 2.8296E+05, 2.8339E+05, 2.8382E+05, 2.8425E+05, 2.8469E+05,
     +  2.8512E+05, 2.8555E+05, 2.8599E+05, 2.8642E+05, 2.8686E+05,
     + 2.8730E+05, 2.8773E+05, 2.8817E+05, 2.8861E+05, 2.8905E+05,
     +  2.8948E+05, 2.8992E+05, 2.9036E+05, 2.9080E+05, 2.9124E+05,
     + 2.9169E+05, 2.9213E+05, 2.9257E+05, 2.9301E+05, 2.9346E+05,
     +  2.9390E+05, 2.9434E+05, 2.9479E+05, 2.9523E+05, 2.9568E+05,
     + 2.9613E+05, 2.9657E+05, 2.9702E+05, 2.9747E+05, 2.9792E+05,
     +  2.9837E+05, 2.9882E+05, 2.9927E+05, 2.9972E+05, 3.0017E+05,
     + 3.0062E+05, 3.0107E+05, 3.0152E+05, 3.0198E+05, 3.0243E+05,
     +  3.0289E+05, 3.0334E+05, 3.0380E+05, 3.0425E+05, 3.0471E+05,
     + 3.0517E+05, 3.0562E+05, 3.0608E+05, 3.0654E+05, 3.0700E+05,
     +  3.0746E+05, 3.0792E+05, 3.0838E+05, 3.0884E+05, 3.0930E+05,
     + 3.0976E+05, 3.1023E+05, 3.1069E+05, 3.1115E+05, 3.1162E+05,
     +  3.1208E+05, 3.1255E+05, 3.1301E+05, 3.1348E+05, 3.1395E+05,
     + 3.1442E+05, 3.1488E+05, 3.1535E+05, 3.1582E+05, 3.1629E+05,
     +  3.1676E+05, 3.1723E+05, 3.1770E+05, 3.1818E+05, 3.1865E+05,
     + 3.1912E+05, 3.1960E+05, 3.2007E+05, 3.2054E+05, 3.2102E+05,
     +  3.2150E+05, 3.2197E+05, 3.2245E+05, 3.2293E+05, 3.2340E+05,
     + 3.2388E+05, 3.2436E+05, 3.2484E+05, 3.2532E+05, 3.2580E+05,
     +  3.2628E+05, 3.2676E+05, 3.2725E+05, 3.2773E+05, 3.2821E+05,
     + 3.2870E+05, 3.2918E+05, 3.2967E+05, 3.3015E+05, 3.3064E+05,
     +  3.3112E+05, 3.3161E+05, 3.3210E+05, 3.3259E+05, 3.3308E+05,
     + 3.3357E+05, 3.3406E+05, 3.3455E+05, 3.3504E+05, 3.3553E+05,
     +  3.3602E+05, 3.3651E+05, 3.3701E+05, 3.3750E+05, 3.3800E+05,
     + 3.3849E+05, 3.3899E+05, 3.3948E+05, 3.3998E+05, 3.4048E+05,
     +  3.4097E+05, 3.4147E+05, 3.4197E+05, 3.4247E+05, 3.4297E+05,
     + 3.4347E+05, 3.4397E+05, 3.4448E+05, 3.4498E+05, 3.4548E+05,
     +  3.4598E+05, 3.4649E+05, 3.4699E+05, 3.4750E+05, 3.4800E+05,
     + 3.4851E+05, 3.4902E+05, 3.4952E+05, 3.5003E+05, 3.5054E+05,
     +  3.5105E+05, 3.5156E+05, 3.5207E+05, 3.5258E+05, 3.5309E+05,
     + 3.5360E+05, 3.5412E+05, 3.5463E+05, 3.5514E+05, 3.5566E+05,
     +  3.5617E+05, 3.5669E+05, 3.5720E+05, 3.5772E+05, 3.5824E+05,
     + 3.5875E+05, 3.5927E+05, 3.5979E+05, 3.6031E+05, 3.6083E+05,
     +  3.6135E+05, 3.6187E+05, 3.6239E+05, 3.6292E+05, 3.6344E+05,
     + 3.6396E+05, 3.6449E+05, 3.6501E+05, 3.6554E+05, 3.6606E+05,
     +  3.6659E+05, 3.6711E+05, 3.6764E+05, 3.6817E+05, 3.6870E+05,
     + 3.6923E+05, 3.6976E+05, 3.7029E+05, 3.7082E+05, 3.7135E+05,
     +  3.7188E+05, 3.7241E+05, 3.7295E+05, 3.7348E+05, 3.7402E+05,
     + 3.7455E+05, 3.7509E+05, 3.7562E+05, 3.7616E+05, 3.7670E+05,
     +  3.7724E+05, 3.7777E+05, 3.7831E+05, 3.7885E+05, 3.7939E+05,
     + 3.7993E+05, 3.8048E+05, 3.8102E+05, 3.8156E+05, 3.8210E+05,
     +  3.8265E+05, 3.8319E+05, 3.8374E+05, 3.8428E+05, 3.8483E+05,
     + 3.8538E+05, 3.8592E+05, 3.8647E+05, 3.8702E+05, 3.8757E+05,
     +  3.8812E+05, 3.8867E+05, 3.8922E+05, 3.8977E+05, 3.9032E+05,
     + 3.9088E+05, 3.9143E+05, 3.9198E+05, 3.9254E+05, 3.9309E+05,
     +  3.9365E+05, 3.9421E+05, 3.9476E+05, 3.9532E+05, 3.9588E+05,
     + 3.9644E+05, 3.9700E+05, 3.9756E+05, 3.9812E+05, 3.9868E+05,
     +  3.9924E+05, 3.9980E+05, 4.0037E+05, 4.0093E+05, 4.0149E+05,
     + 4.0206E+05, 4.0262E+05, 4.0319E+05, 4.0376E+05, 4.0432E+05,
     +  4.0489E+05, 4.0546E+05, 4.0603E+05, 4.0660E+05, 4.0717E+05/
      data (QofT( 3,J),J=2409,4510)/ 4.0774E+05, 4.0831E+05,
     +  4.0888E+05, 4.0946E+05, 4.1003E+05, 4.1060E+05, 4.1118E+05,
     +  4.1175E+05,
     + 4.1233E+05, 4.1291E+05, 4.1348E+05, 4.1406E+05, 4.1464E+05,
     +  4.1522E+05, 4.1580E+05, 4.1638E+05, 4.1696E+05, 4.1754E+05,
     + 4.1812E+05, 4.1870E+05, 4.1929E+05, 4.1987E+05, 4.2046E+05,
     +  4.2104E+05, 4.2163E+05, 4.2221E+05, 4.2280E+05, 4.2339E+05,
     + 4.2398E+05, 4.2456E+05, 4.2515E+05, 4.2574E+05, 4.2633E+05,
     +  4.2693E+05, 4.2752E+05, 4.2811E+05, 4.2870E+05, 4.2930E+05,
     + 4.2989E+05, 4.3049E+05, 4.3108E+05, 4.3168E+05, 4.3228E+05,
     +  4.3287E+05, 4.3347E+05, 4.3407E+05, 4.3467E+05, 4.3527E+05,
     + 4.3587E+05, 4.3647E+05, 4.3707E+05, 4.3768E+05, 4.3828E+05,
     +  4.3888E+05, 4.3949E+05, 4.4009E+05, 4.4070E+05, 4.4130E+05,
     + 4.4191E+05, 4.4252E+05, 4.4313E+05, 4.4374E+05, 4.4435E+05,
     +  4.4496E+05, 4.4557E+05, 4.4618E+05, 4.4679E+05, 4.4740E+05,
     + 4.4802E+05, 4.4863E+05, 4.4924E+05, 4.4986E+05, 4.5048E+05,
     +  4.5109E+05, 4.5171E+05, 4.5233E+05, 4.5295E+05, 4.5356E+05,
     + 4.5418E+05, 4.5481E+05, 4.5543E+05, 4.5605E+05, 4.5667E+05,
     +  4.5729E+05, 4.5792E+05, 4.5854E+05, 4.5917E+05, 4.5979E+05,
     + 4.6042E+05, 4.6104E+05, 4.6167E+05, 4.6230E+05, 4.6293E+05,
     +  4.6356E+05, 4.6419E+05, 4.6482E+05, 4.6545E+05, 4.6608E+05,
     + 4.6671E+05, 4.6735E+05, 4.6798E+05, 4.6862E+05, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       162
      data (QofT( 4,J),J=1,2408)/ 6.0000E+00, 6.0003E+00, 6.0106E+00,
     +  6.0686E+00, 6.2126E+00, 6.4588E+00, 6.8073E+00, 7.2517E+00,
     + 7.7840E+00, 8.3962E+00, 9.0812E+00, 9.8326E+00, 1.0645E+01,
     +  1.1513E+01, 1.2431E+01, 1.3397E+01, 1.4407E+01, 1.5457E+01,
     + 1.6545E+01, 1.7668E+01, 2.0013E+01, 2.2479E+01, 2.5055E+01,
     +  2.7733E+01, 3.0509E+01, 3.3377E+01, 3.6333E+01, 3.9374E+01,
     + 4.2498E+01, 4.5702E+01, 4.8985E+01, 5.2343E+01, 5.5776E+01,
     +  5.9282E+01, 6.2860E+01, 6.6507E+01, 7.0223E+01, 7.4006E+01,
     + 7.7856E+01, 8.1771E+01, 8.5750E+01, 8.9792E+01, 9.3896E+01,
     +  9.8061E+01, 1.0229E+02, 1.0657E+02, 1.1092E+02, 1.1532E+02,
     + 1.1978E+02, 1.2429E+02, 1.2886E+02, 1.3349E+02, 1.3817E+02,
     +  1.4291E+02, 1.4769E+02, 1.5253E+02, 1.5743E+02, 1.6237E+02,
     + 1.6737E+02, 1.7241E+02, 1.7751E+02, 1.8266E+02, 1.8785E+02,
     +  1.9310E+02, 1.9839E+02, 2.0373E+02, 2.0912E+02, 2.1455E+02,
     + 2.2003E+02, 2.2556E+02, 2.3114E+02, 2.3675E+02, 2.4242E+02,
     +  2.4813E+02, 2.5388E+02, 2.5968E+02, 2.6552E+02, 2.7141E+02,
     + 2.7734E+02, 2.8331E+02, 2.8932E+02, 2.9538E+02, 3.0148E+02,
     +  3.0762E+02, 3.1380E+02, 3.2003E+02, 3.2629E+02, 3.3260E+02,
     + 3.3894E+02, 3.4533E+02, 3.5176E+02, 3.5822E+02, 3.6473E+02,
     +  3.7127E+02, 3.7786E+02, 3.8448E+02, 3.9114E+02, 3.9784E+02,
     + 4.0458E+02, 4.1136E+02, 4.1817E+02, 4.2503E+02, 4.3192E+02,
     +  4.3885E+02, 4.4581E+02, 4.5281E+02, 4.5985E+02, 4.6693E+02,
     + 4.7404E+02, 4.8119E+02, 4.8837E+02, 4.9559E+02, 5.0285E+02,
     +  5.1014E+02, 5.1747E+02, 5.2483E+02, 5.3223E+02, 5.3966E+02,
     + 5.4713E+02, 5.5464E+02, 5.6218E+02, 5.6975E+02, 5.7736E+02,
     +  5.8500E+02, 5.9268E+02, 6.0039E+02, 6.0813E+02, 6.1591E+02,
     + 6.2373E+02, 6.3157E+02, 6.3945E+02, 6.4737E+02, 6.5532E+02,
     +  6.6330E+02, 6.7131E+02, 6.7936E+02, 6.8745E+02, 6.9556E+02,
     + 7.0371E+02, 7.1189E+02, 7.2010E+02, 7.2835E+02, 7.3663E+02,
     +  7.4494E+02, 7.5329E+02, 7.6167E+02, 7.7008E+02, 7.7852E+02,
     + 7.8700E+02, 7.9551E+02, 8.0405E+02, 8.1262E+02, 8.2123E+02,
     +  8.2987E+02, 8.3854E+02, 8.4724E+02, 8.5598E+02, 8.6474E+02,
     + 8.7354E+02, 8.8237E+02, 8.9124E+02, 9.0013E+02, 9.0906E+02,
     +  9.1802E+02, 9.2701E+02, 9.3604E+02, 9.4509E+02, 9.5418E+02,
     + 9.6330E+02, 9.7245E+02, 9.8164E+02, 9.9086E+02, 1.0001E+03,
     +  1.0094E+03, 1.0187E+03, 1.0280E+03, 1.0374E+03, 1.0468E+03,
     + 1.0563E+03, 1.0657E+03, 1.0752E+03, 1.0848E+03, 1.0944E+03,
     +  1.1040E+03, 1.1136E+03, 1.1232E+03, 1.1329E+03, 1.1427E+03,
     + 1.1524E+03, 1.1622E+03, 1.1721E+03, 1.1819E+03, 1.1918E+03,
     +  1.2017E+03, 1.2117E+03, 1.2217E+03, 1.2317E+03, 1.2417E+03,
     + 1.2518E+03, 1.2620E+03, 1.2721E+03, 1.2823E+03, 1.2925E+03,
     +  1.3027E+03, 1.3130E+03, 1.3233E+03, 1.3337E+03, 1.3441E+03,
     + 1.3545E+03, 1.3649E+03, 1.3754E+03, 1.3859E+03, 1.3965E+03,
     +  1.4070E+03, 1.4176E+03, 1.4283E+03, 1.4390E+03, 1.4497E+03,
     + 1.4604E+03, 1.4712E+03, 1.4820E+03, 1.4929E+03, 1.5037E+03,
     +  1.5146E+03, 1.5256E+03, 1.5366E+03, 1.5476E+03, 1.5586E+03,
     + 1.5697E+03, 1.5808E+03, 1.5920E+03, 1.6031E+03, 1.6144E+03,
     +  1.6256E+03, 1.6369E+03, 1.6482E+03, 1.6596E+03, 1.6710E+03,
     + 1.6824E+03, 1.6938E+03, 1.7053E+03, 1.7169E+03, 1.7284E+03,
     +  1.7400E+03, 1.7516E+03, 1.7633E+03, 1.7750E+03, 1.7867E+03,
     + 1.7985E+03, 1.8103E+03, 1.8222E+03, 1.8340E+03, 1.8460E+03,
     +  1.8579E+03, 1.8699E+03, 1.8819E+03, 1.8940E+03, 1.9061E+03,
     + 1.9182E+03, 1.9303E+03, 1.9425E+03, 1.9548E+03, 1.9671E+03,
     +  1.9794E+03, 1.9917E+03, 2.0041E+03, 2.0165E+03, 2.0290E+03,
     + 2.0414E+03, 2.0540E+03, 2.0665E+03, 2.0791E+03, 2.0918E+03,
     +  2.1045E+03, 2.1172E+03, 2.1299E+03, 2.1427E+03, 2.1555E+03,
     + 2.1684E+03, 2.1813E+03, 2.1942E+03, 2.2072E+03, 2.2202E+03,
     +  2.2333E+03, 2.2464E+03, 2.2595E+03, 2.2727E+03, 2.2859E+03,
     + 2.2991E+03, 2.3124E+03, 2.3257E+03, 2.3391E+03, 2.3525E+03,
     +  2.3659E+03, 2.3794E+03, 2.3929E+03, 2.4064E+03, 2.4200E+03,
     + 2.4337E+03, 2.4473E+03, 2.4610E+03, 2.4748E+03, 2.4886E+03,
     +  2.5024E+03, 2.5163E+03, 2.5302E+03, 2.5441E+03, 2.5581E+03,
     + 2.5721E+03, 2.5862E+03, 2.6003E+03, 2.6145E+03, 2.6287E+03,
     +  2.6429E+03, 2.6572E+03, 2.6715E+03, 2.6858E+03, 2.7002E+03,
     + 2.7147E+03, 2.7291E+03, 2.7437E+03, 2.7582E+03, 2.7728E+03,
     +  2.7875E+03, 2.8021E+03, 2.8169E+03, 2.8316E+03, 2.8465E+03,
     + 2.8613E+03, 2.8762E+03, 2.8911E+03, 2.9061E+03, 2.9211E+03,
     +  2.9362E+03, 2.9513E+03, 2.9665E+03, 2.9817E+03, 2.9969E+03,
     + 3.0122E+03, 3.0275E+03, 3.0429E+03, 3.0583E+03, 3.0737E+03,
     +  3.0892E+03, 3.1048E+03, 3.1204E+03, 3.1360E+03, 3.1517E+03,
     + 3.1674E+03, 3.1831E+03, 3.1990E+03, 3.2148E+03, 3.2307E+03,
     +  3.2466E+03, 3.2626E+03, 3.2787E+03, 3.2947E+03, 3.3109E+03,
     + 3.3270E+03, 3.3432E+03, 3.3595E+03, 3.3758E+03, 3.3922E+03,
     +  3.4085E+03, 3.4250E+03, 3.4415E+03, 3.4580E+03, 3.4746E+03,
     + 3.4912E+03, 3.5079E+03, 3.5246E+03, 3.5414E+03, 3.5582E+03,
     +  3.5751E+03, 3.5920E+03, 3.6089E+03, 3.6259E+03, 3.6430E+03,
     + 3.6601E+03, 3.6772E+03, 3.6944E+03, 3.7116E+03, 3.7289E+03,
     +  3.7463E+03, 3.7637E+03, 3.7811E+03, 3.7986E+03, 3.8161E+03,
     + 3.8337E+03, 3.8513E+03, 3.8690E+03, 3.8867E+03, 3.9045E+03,
     +  3.9223E+03, 3.9402E+03, 3.9581E+03, 3.9761E+03, 3.9941E+03,
     + 4.0122E+03, 4.0303E+03, 4.0485E+03, 4.0667E+03, 4.0850E+03,
     +  4.1033E+03, 4.1217E+03, 4.1401E+03, 4.1586E+03, 4.1772E+03,
     + 4.1957E+03, 4.2144E+03, 4.2331E+03, 4.2518E+03, 4.2706E+03,
     +  4.2894E+03, 4.3083E+03, 4.3273E+03, 4.3462E+03, 4.3653E+03,
     + 4.3844E+03, 4.4036E+03, 4.4228E+03, 4.4420E+03, 4.4613E+03,
     +  4.4807E+03, 4.5001E+03, 4.5196E+03, 4.5391E+03, 4.5587E+03,
     + 4.5783E+03, 4.5980E+03, 4.6178E+03, 4.6375E+03, 4.6574E+03,
     +  4.6773E+03, 4.6973E+03, 4.7173E+03, 4.7373E+03, 4.7575E+03,
     + 4.7776E+03, 4.7979E+03, 4.8182E+03, 4.8385E+03, 4.8589E+03,
     +  4.8794E+03, 4.8999E+03, 4.9204E+03, 4.9411E+03, 4.9618E+03,
     + 4.9825E+03, 5.0033E+03, 5.0241E+03, 5.0450E+03, 5.0660E+03,
     +  5.0870E+03, 5.1081E+03, 5.1292E+03, 5.1504E+03, 5.1717E+03,
     + 5.1930E+03, 5.2144E+03, 5.2358E+03, 5.2573E+03, 5.2788E+03,
     +  5.3004E+03, 5.3221E+03, 5.3438E+03, 5.3655E+03, 5.3874E+03,
     + 5.4093E+03, 5.4312E+03, 5.4532E+03, 5.4753E+03, 5.4974E+03,
     +  5.5196E+03, 5.5419E+03, 5.5642E+03, 5.5866E+03, 5.6090E+03,
     + 5.6315E+03, 5.6541E+03, 5.6767E+03, 5.6993E+03, 5.7221E+03,
     +  5.7449E+03, 5.7677E+03, 5.7907E+03, 5.8136E+03, 5.8367E+03,
     + 5.8598E+03, 5.8830E+03, 5.9062E+03, 5.9295E+03, 5.9528E+03,
     +  5.9763E+03, 5.9997E+03, 6.0233E+03, 6.0469E+03, 6.0706E+03,
     + 6.0943E+03, 6.1181E+03, 6.1420E+03, 6.1659E+03, 6.1899E+03,
     +  6.2139E+03, 6.2381E+03, 6.2622E+03, 6.2865E+03, 6.3108E+03,
     + 6.3352E+03, 6.3596E+03, 6.3841E+03, 6.4087E+03, 6.4334E+03,
     +  6.4581E+03, 6.4828E+03, 6.5077E+03, 6.5326E+03, 6.5575E+03,
     + 6.5826E+03, 6.6077E+03, 6.6328E+03, 6.6581E+03, 6.6834E+03,
     +  6.7088E+03, 6.7342E+03, 6.7597E+03, 6.7853E+03, 6.8109E+03,
     + 6.8366E+03, 6.8624E+03, 6.8882E+03, 6.9142E+03, 6.9401E+03,
     +  6.9662E+03, 6.9923E+03, 7.0185E+03, 7.0448E+03, 7.0711E+03,
     + 7.0975E+03, 7.1240E+03, 7.1505E+03, 7.1771E+03, 7.2038E+03,
     +  7.2305E+03, 7.2573E+03, 7.2842E+03, 7.3112E+03, 7.3382E+03,
     + 7.3653E+03, 7.3925E+03, 7.4197E+03, 7.4470E+03, 7.4744E+03,
     +  7.5019E+03, 7.5294E+03, 7.5570E+03, 7.5847E+03, 7.6125E+03,
     + 7.6403E+03, 7.6682E+03, 7.6961E+03, 7.7242E+03, 7.7523E+03,
     +  7.7805E+03, 7.8087E+03, 7.8371E+03, 7.8655E+03, 7.8940E+03,
     + 7.9225E+03, 7.9512E+03, 7.9799E+03, 8.0086E+03, 8.0375E+03,
     +  8.0664E+03, 8.0954E+03, 8.1245E+03, 8.1537E+03, 8.1829E+03,
     + 8.2122E+03, 8.2416E+03, 8.2711E+03, 8.3006E+03, 8.3302E+03,
     +  8.3599E+03, 8.3897E+03, 8.4195E+03, 8.4495E+03, 8.4795E+03,
     + 8.5095E+03, 8.5397E+03, 8.5699E+03, 8.6002E+03, 8.6306E+03,
     +  8.6611E+03, 8.6917E+03, 8.7223E+03, 8.7530E+03, 8.7838E+03,
     + 8.8146E+03, 8.8456E+03, 8.8766E+03, 8.9077E+03, 8.9389E+03,
     +  8.9702E+03, 9.0015E+03, 9.0329E+03, 9.0644E+03, 9.0960E+03,
     + 9.1277E+03, 9.1594E+03, 9.1913E+03, 9.2232E+03, 9.2552E+03,
     +  9.2872E+03, 9.3194E+03, 9.3516E+03, 9.3840E+03, 9.4164E+03,
     + 9.4489E+03, 9.4814E+03, 9.5141E+03, 9.5468E+03, 9.5796E+03,
     +  9.6125E+03, 9.6455E+03, 9.6786E+03, 9.7118E+03, 9.7450E+03,
     + 9.7783E+03, 9.8117E+03, 9.8452E+03, 9.8788E+03, 9.9125E+03,
     +  9.9462E+03, 9.9801E+03, 1.0014E+04, 1.0048E+04, 1.0082E+04,
     + 1.0116E+04, 1.0151E+04, 1.0185E+04, 1.0219E+04, 1.0254E+04,
     +  1.0288E+04, 1.0323E+04, 1.0358E+04, 1.0393E+04, 1.0428E+04,
     + 1.0463E+04, 1.0498E+04, 1.0533E+04, 1.0569E+04, 1.0604E+04,
     +  1.0639E+04, 1.0675E+04, 1.0711E+04, 1.0746E+04, 1.0782E+04,
     + 1.0818E+04, 1.0854E+04, 1.0890E+04, 1.0927E+04, 1.0963E+04,
     +  1.0999E+04, 1.1036E+04, 1.1072E+04, 1.1109E+04, 1.1146E+04,
     + 1.1183E+04, 1.1220E+04, 1.1257E+04, 1.1294E+04, 1.1331E+04,
     +  1.1368E+04, 1.1406E+04, 1.1443E+04, 1.1481E+04, 1.1519E+04,
     + 1.1556E+04, 1.1594E+04, 1.1632E+04, 1.1670E+04, 1.1708E+04,
     +  1.1747E+04, 1.1785E+04, 1.1824E+04, 1.1862E+04, 1.1901E+04,
     + 1.1939E+04, 1.1978E+04, 1.2017E+04, 1.2056E+04, 1.2095E+04,
     +  1.2135E+04, 1.2174E+04, 1.2213E+04, 1.2253E+04, 1.2292E+04,
     + 1.2332E+04, 1.2372E+04, 1.2412E+04, 1.2452E+04, 1.2492E+04,
     +  1.2532E+04, 1.2572E+04, 1.2613E+04, 1.2653E+04, 1.2694E+04,
     + 1.2734E+04, 1.2775E+04, 1.2816E+04, 1.2857E+04, 1.2898E+04,
     +  1.2939E+04, 1.2981E+04, 1.3022E+04, 1.3064E+04, 1.3105E+04,
     + 1.3147E+04, 1.3189E+04, 1.3230E+04, 1.3272E+04, 1.3315E+04,
     +  1.3357E+04, 1.3399E+04, 1.3441E+04, 1.3484E+04, 1.3527E+04,
     + 1.3569E+04, 1.3612E+04, 1.3655E+04, 1.3698E+04, 1.3741E+04,
     +  1.3784E+04, 1.3828E+04, 1.3871E+04, 1.3915E+04, 1.3958E+04,
     + 1.4002E+04, 1.4046E+04, 1.4090E+04, 1.4134E+04, 1.4178E+04,
     +  1.4223E+04, 1.4267E+04, 1.4312E+04, 1.4356E+04, 1.4401E+04,
     + 1.4446E+04, 1.4491E+04, 1.4536E+04, 1.4581E+04, 1.4626E+04,
     +  1.4671E+04, 1.4717E+04, 1.4762E+04, 1.4808E+04, 1.4854E+04,
     + 1.4900E+04, 1.4946E+04, 1.4992E+04, 1.5038E+04, 1.5085E+04,
     +  1.5131E+04, 1.5178E+04, 1.5224E+04, 1.5271E+04, 1.5318E+04,
     + 1.5365E+04, 1.5412E+04, 1.5459E+04, 1.5507E+04, 1.5554E+04,
     +  1.5602E+04, 1.5650E+04, 1.5697E+04, 1.5745E+04, 1.5793E+04,
     + 1.5842E+04, 1.5890E+04, 1.5938E+04, 1.5987E+04, 1.6035E+04,
     +  1.6084E+04, 1.6133E+04, 1.6182E+04, 1.6231E+04, 1.6280E+04,
     + 1.6329E+04, 1.6379E+04, 1.6428E+04, 1.6478E+04, 1.6528E+04,
     +  1.6578E+04, 1.6628E+04, 1.6678E+04, 1.6728E+04, 1.6778E+04,
     + 1.6829E+04, 1.6879E+04, 1.6930E+04, 1.6981E+04, 1.7032E+04,
     +  1.7083E+04, 1.7134E+04, 1.7185E+04, 1.7237E+04, 1.7288E+04,
     + 1.7340E+04, 1.7392E+04, 1.7444E+04, 1.7496E+04, 1.7548E+04,
     +  1.7600E+04, 1.7653E+04, 1.7705E+04, 1.7758E+04, 1.7810E+04,
     + 1.7863E+04, 1.7916E+04, 1.7969E+04, 1.8023E+04, 1.8076E+04,
     +  1.8130E+04, 1.8183E+04, 1.8237E+04, 1.8291E+04, 1.8345E+04,
     + 1.8399E+04, 1.8453E+04, 1.8508E+04, 1.8562E+04, 1.8617E+04,
     +  1.8671E+04, 1.8726E+04, 1.8781E+04, 1.8836E+04, 1.8892E+04,
     + 1.8947E+04, 1.9003E+04, 1.9058E+04, 1.9114E+04, 1.9170E+04,
     +  1.9226E+04, 1.9282E+04, 1.9338E+04, 1.9395E+04, 1.9451E+04,
     + 1.9508E+04, 1.9565E+04, 1.9622E+04, 1.9679E+04, 1.9736E+04,
     +  1.9793E+04, 1.9851E+04, 1.9908E+04, 1.9966E+04, 2.0024E+04,
     + 2.0082E+04, 2.0140E+04, 2.0198E+04, 2.0256E+04, 2.0315E+04,
     +  2.0373E+04, 2.0432E+04, 2.0491E+04, 2.0550E+04, 2.0609E+04,
     + 2.0669E+04, 2.0728E+04, 2.0788E+04, 2.0847E+04, 2.0907E+04,
     +  2.0967E+04, 2.1027E+04, 2.1087E+04, 2.1148E+04, 2.1208E+04,
     + 2.1269E+04, 2.1330E+04, 2.1391E+04, 2.1452E+04, 2.1513E+04,
     +  2.1574E+04, 2.1636E+04, 2.1697E+04, 2.1759E+04, 2.1821E+04,
     + 2.1883E+04, 2.1945E+04, 2.2008E+04, 2.2070E+04, 2.2133E+04,
     +  2.2195E+04, 2.2258E+04, 2.2321E+04, 2.2384E+04, 2.2448E+04,
     + 2.2511E+04, 2.2575E+04, 2.2638E+04, 2.2702E+04, 2.2766E+04,
     +  2.2830E+04, 2.2895E+04, 2.2959E+04, 2.3024E+04, 2.3088E+04,
     + 2.3153E+04, 2.3218E+04, 2.3283E+04, 2.3349E+04, 2.3414E+04,
     +  2.3480E+04, 2.3545E+04, 2.3611E+04, 2.3677E+04, 2.3743E+04,
     + 2.3810E+04, 2.3876E+04, 2.3943E+04, 2.4009E+04, 2.4076E+04,
     +  2.4143E+04, 2.4210E+04, 2.4278E+04, 2.4345E+04, 2.4413E+04,
     + 2.4481E+04, 2.4549E+04, 2.4617E+04, 2.4685E+04, 2.4753E+04,
     +  2.4822E+04, 2.4890E+04, 2.4959E+04, 2.5028E+04, 2.5097E+04,
     + 2.5167E+04, 2.5236E+04, 2.5306E+04, 2.5375E+04, 2.5445E+04,
     +  2.5515E+04, 2.5585E+04, 2.5656E+04, 2.5726E+04, 2.5797E+04,
     + 2.5868E+04, 2.5939E+04, 2.6010E+04, 2.6081E+04, 2.6152E+04,
     +  2.6224E+04, 2.6296E+04, 2.6368E+04, 2.6440E+04, 2.6512E+04,
     + 2.6584E+04, 2.6657E+04, 2.6729E+04, 2.6802E+04, 2.6875E+04,
     +  2.6948E+04, 2.7022E+04, 2.7095E+04, 2.7169E+04, 2.7242E+04,
     + 2.7316E+04, 2.7390E+04, 2.7465E+04, 2.7539E+04, 2.7614E+04,
     +  2.7688E+04, 2.7763E+04, 2.7838E+04, 2.7913E+04, 2.7989E+04,
     + 2.8064E+04, 2.8140E+04, 2.8216E+04, 2.8292E+04, 2.8368E+04,
     +  2.8444E+04, 2.8521E+04, 2.8598E+04, 2.8674E+04, 2.8751E+04,
     + 2.8828E+04, 2.8906E+04, 2.8983E+04, 2.9061E+04, 2.9139E+04,
     +  2.9217E+04, 2.9295E+04, 2.9373E+04, 2.9452E+04, 2.9530E+04,
     + 2.9609E+04, 2.9688E+04, 2.9767E+04, 2.9846E+04, 2.9926E+04,
     +  3.0006E+04, 3.0085E+04, 3.0165E+04, 3.0246E+04, 3.0326E+04,
     + 3.0406E+04, 3.0487E+04, 3.0568E+04, 3.0649E+04, 3.0730E+04,
     +  3.0811E+04, 3.0893E+04, 3.0974E+04, 3.1056E+04, 3.1138E+04,
     + 3.1220E+04, 3.1303E+04, 3.1385E+04, 3.1468E+04, 3.1551E+04,
     +  3.1634E+04, 3.1717E+04, 3.1801E+04, 3.1884E+04, 3.1968E+04,
     + 3.2052E+04, 3.2136E+04, 3.2220E+04, 3.2305E+04, 3.2389E+04,
     +  3.2474E+04, 3.2559E+04, 3.2644E+04, 3.2730E+04, 3.2815E+04,
     + 3.2901E+04, 3.2987E+04, 3.3073E+04, 3.3159E+04, 3.3245E+04,
     +  3.3332E+04, 3.3419E+04, 3.3505E+04, 3.3593E+04, 3.3680E+04,
     + 3.3767E+04, 3.3855E+04, 3.3943E+04, 3.4031E+04, 3.4119E+04,
     +  3.4207E+04, 3.4296E+04, 3.4385E+04, 3.4474E+04, 3.4563E+04,
     + 3.4652E+04, 3.4741E+04, 3.4831E+04, 3.4921E+04, 3.5011E+04,
     +  3.5101E+04, 3.5192E+04, 3.5282E+04, 3.5373E+04, 3.5464E+04,
     + 3.5555E+04, 3.5646E+04, 3.5738E+04, 3.5829E+04, 3.5921E+04,
     +  3.6013E+04, 3.6106E+04, 3.6198E+04, 3.6291E+04, 3.6383E+04,
     + 3.6476E+04, 3.6570E+04, 3.6663E+04, 3.6756E+04, 3.6850E+04,
     +  3.6944E+04, 3.7038E+04, 3.7133E+04, 3.7227E+04, 3.7322E+04,
     + 3.7417E+04, 3.7512E+04, 3.7607E+04, 3.7703E+04, 3.7798E+04,
     +  3.7894E+04, 3.7990E+04, 3.8086E+04, 3.8183E+04, 3.8279E+04,
     + 3.8376E+04, 3.8473E+04, 3.8570E+04, 3.8668E+04, 3.8765E+04,
     +  3.8863E+04, 3.8961E+04, 3.9059E+04, 3.9158E+04, 3.9256E+04,
     + 3.9355E+04, 3.9454E+04, 3.9553E+04, 3.9653E+04, 3.9752E+04,
     +  3.9852E+04, 3.9952E+04, 4.0052E+04, 4.0152E+04, 4.0253E+04,
     + 4.0354E+04, 4.0455E+04, 4.0556E+04, 4.0657E+04, 4.0759E+04,
     +  4.0860E+04, 4.0962E+04, 4.1064E+04, 4.1167E+04, 4.1269E+04,
     + 4.1372E+04, 4.1475E+04, 4.1578E+04, 4.1682E+04, 4.1785E+04,
     +  4.1889E+04, 4.1993E+04, 4.2097E+04, 4.2201E+04, 4.2306E+04,
     + 4.2411E+04, 4.2516E+04, 4.2621E+04, 4.2726E+04, 4.2832E+04,
     +  4.2938E+04, 4.3044E+04, 4.3150E+04, 4.3257E+04, 4.3363E+04,
     + 4.3470E+04, 4.3577E+04, 4.3685E+04, 4.3792E+04, 4.3900E+04,
     +  4.4008E+04, 4.4116E+04, 4.4224E+04, 4.4333E+04, 4.4441E+04,
     + 4.4550E+04, 4.4660E+04, 4.4769E+04, 4.4878E+04, 4.4988E+04,
     +  4.5098E+04, 4.5209E+04, 4.5319E+04, 4.5430E+04, 4.5541E+04,
     + 4.5652E+04, 4.5763E+04, 4.5874E+04, 4.5986E+04, 4.6098E+04,
     +  4.6210E+04, 4.6323E+04, 4.6435E+04, 4.6548E+04, 4.6661E+04,
     + 4.6774E+04, 4.6888E+04, 4.7001E+04, 4.7115E+04, 4.7229E+04,
     +  4.7344E+04, 4.7458E+04, 4.7573E+04, 4.7688E+04, 4.7803E+04,
     + 4.7919E+04, 4.8034E+04, 4.8150E+04, 4.8266E+04, 4.8383E+04,
     +  4.8499E+04, 4.8616E+04, 4.8733E+04, 4.8850E+04, 4.8967E+04,
     + 4.9085E+04, 4.9203E+04, 4.9321E+04, 4.9439E+04, 4.9558E+04,
     +  4.9676E+04, 4.9795E+04, 4.9915E+04, 5.0034E+04, 5.0154E+04,
     + 5.0274E+04, 5.0394E+04, 5.0514E+04, 5.0635E+04, 5.0755E+04,
     +  5.0876E+04, 5.0998E+04, 5.1119E+04, 5.1241E+04, 5.1363E+04,
     + 5.1485E+04, 5.1607E+04, 5.1730E+04, 5.1853E+04, 5.1976E+04,
     +  5.2099E+04, 5.2222E+04, 5.2346E+04, 5.2470E+04, 5.2594E+04,
     + 5.2719E+04, 5.2844E+04, 5.2968E+04, 5.3094E+04, 5.3219E+04,
     +  5.3345E+04, 5.3470E+04, 5.3596E+04, 5.3723E+04, 5.3849E+04,
     + 5.3976E+04, 5.4103E+04, 5.4230E+04, 5.4358E+04, 5.4485E+04,
     +  5.4613E+04, 5.4742E+04, 5.4870E+04, 5.4999E+04, 5.5128E+04,
     + 5.5257E+04, 5.5386E+04, 5.5516E+04, 5.5646E+04, 5.5776E+04,
     +  5.5906E+04, 5.6036E+04, 5.6167E+04, 5.6298E+04, 5.6430E+04,
     + 5.6561E+04, 5.6693E+04, 5.6825E+04, 5.6957E+04, 5.7090E+04,
     +  5.7222E+04, 5.7355E+04, 5.7488E+04, 5.7622E+04, 5.7756E+04,
     + 5.7890E+04, 5.8024E+04, 5.8158E+04, 5.8293E+04, 5.8428E+04,
     +  5.8563E+04, 5.8698E+04, 5.8834E+04, 5.8970E+04, 5.9106E+04,
     + 5.9242E+04, 5.9379E+04, 5.9516E+04, 5.9653E+04, 5.9790E+04,
     +  5.9928E+04, 6.0066E+04, 6.0204E+04, 6.0342E+04, 6.0481E+04,
     + 6.0620E+04, 6.0759E+04, 6.0898E+04, 6.1038E+04, 6.1178E+04,
     +  6.1318E+04, 6.1458E+04, 6.1599E+04, 6.1740E+04, 6.1881E+04,
     + 6.2022E+04, 6.2164E+04, 6.2306E+04, 6.2448E+04, 6.2590E+04,
     +  6.2733E+04, 6.2876E+04, 6.3019E+04, 6.3162E+04, 6.3306E+04,
     + 6.3450E+04, 6.3594E+04, 6.3738E+04, 6.3883E+04, 6.4028E+04,
     +  6.4173E+04, 6.4318E+04, 6.4464E+04, 6.4610E+04, 6.4756E+04,
     + 6.4903E+04, 6.5049E+04, 6.5196E+04, 6.5344E+04, 6.5491E+04,
     +  6.5639E+04, 6.5787E+04, 6.5935E+04, 6.6084E+04, 6.6232E+04,
     + 6.6381E+04, 6.6531E+04, 6.6680E+04, 6.6830E+04, 6.6980E+04,
     +  6.7131E+04, 6.7281E+04, 6.7432E+04, 6.7583E+04, 6.7735E+04,
     + 6.7886E+04, 6.8038E+04, 6.8191E+04, 6.8343E+04, 6.8496E+04,
     +  6.8649E+04, 6.8802E+04, 6.8956E+04, 6.9109E+04, 6.9263E+04,
     + 6.9418E+04, 6.9572E+04, 6.9727E+04, 6.9882E+04, 7.0038E+04,
     +  7.0193E+04, 7.0349E+04, 7.0506E+04, 7.0662E+04, 7.0819E+04,
     + 7.0976E+04, 7.1133E+04, 7.1291E+04, 7.1448E+04, 7.1607E+04,
     +  7.1765E+04, 7.1924E+04, 7.2082E+04, 7.2242E+04, 7.2401E+04,
     + 7.2561E+04, 7.2721E+04, 7.2881E+04, 7.3042E+04, 7.3202E+04,
     +  7.3363E+04, 7.3525E+04, 7.3686E+04, 7.3848E+04, 7.4011E+04,
     + 7.4173E+04, 7.4336E+04, 7.4499E+04, 7.4662E+04, 7.4826E+04,
     +  7.4989E+04, 7.5154E+04, 7.5318E+04, 7.5483E+04, 7.5648E+04,
     + 7.5813E+04, 7.5978E+04, 7.6144E+04, 7.6310E+04, 7.6477E+04,
     +  7.6643E+04, 7.6810E+04, 7.6977E+04, 7.7145E+04, 7.7313E+04,
     + 7.7481E+04, 7.7649E+04, 7.7817E+04, 7.7986E+04, 7.8155E+04,
     +  7.8325E+04, 7.8495E+04, 7.8665E+04, 7.8835E+04, 7.9006E+04,
     + 7.9176E+04, 7.9347E+04, 7.9519E+04, 7.9691E+04, 7.9863E+04,
     +  8.0035E+04, 8.0207E+04, 8.0380E+04, 8.0553E+04, 8.0727E+04,
     + 8.0901E+04, 8.1075E+04, 8.1249E+04, 8.1423E+04, 8.1598E+04,
     +  8.1773E+04, 8.1949E+04, 8.2125E+04, 8.2301E+04, 8.2477E+04,
     + 8.2654E+04, 8.2830E+04, 8.3008E+04, 8.3185E+04, 8.3363E+04,
     +  8.3541E+04, 8.3719E+04, 8.3898E+04, 8.4077E+04, 8.4256E+04,
     + 8.4435E+04, 8.4615E+04, 8.4795E+04, 8.4976E+04, 8.5156E+04,
     +  8.5337E+04, 8.5519E+04, 8.5700E+04, 8.5882E+04, 8.6064E+04,
     + 8.6247E+04, 8.6429E+04, 8.6613E+04, 8.6796E+04, 8.6979E+04,
     +  8.7163E+04, 8.7348E+04, 8.7532E+04, 8.7717E+04, 8.7902E+04,
     + 8.8088E+04, 8.8273E+04, 8.8459E+04, 8.8646E+04, 8.8832E+04,
     +  8.9019E+04, 8.9206E+04, 8.9394E+04, 8.9582E+04, 8.9770E+04,
     + 8.9958E+04, 9.0147E+04, 9.0336E+04, 9.0525E+04, 9.0715E+04,
     +  9.0905E+04, 9.1095E+04, 9.1286E+04, 9.1476E+04, 9.1668E+04,
     + 9.1859E+04, 9.2051E+04, 9.2243E+04, 9.2435E+04, 9.2628E+04,
     +  9.2821E+04, 9.3014E+04, 9.3208E+04, 9.3401E+04, 9.3596E+04,
     + 9.3790E+04, 9.3985E+04, 9.4180E+04, 9.4375E+04, 9.4571E+04,
     +  9.4767E+04, 9.4964E+04, 9.5160E+04, 9.5357E+04, 9.5554E+04,
     + 9.5752E+04, 9.5950E+04, 9.6148E+04, 9.6347E+04, 9.6545E+04,
     +  9.6745E+04, 9.6944E+04, 9.7144E+04, 9.7344E+04, 9.7544E+04,
     + 9.7745E+04, 9.7946E+04, 9.8147E+04, 9.8349E+04, 9.8551E+04,
     +  9.8753E+04, 9.8956E+04, 9.9158E+04, 9.9362E+04, 9.9565E+04,
     + 9.9769E+04, 9.9973E+04, 1.0018E+05, 1.0038E+05, 1.0059E+05,
     +  1.0079E+05, 1.0100E+05, 1.0120E+05, 1.0141E+05, 1.0162E+05,
     + 1.0182E+05, 1.0203E+05, 1.0224E+05, 1.0245E+05, 1.0266E+05,
     +  1.0286E+05, 1.0307E+05, 1.0328E+05, 1.0349E+05, 1.0370E+05,
     + 1.0391E+05, 1.0412E+05, 1.0433E+05, 1.0454E+05, 1.0476E+05,
     +  1.0497E+05, 1.0518E+05, 1.0539E+05, 1.0560E+05, 1.0582E+05,
     + 1.0603E+05, 1.0625E+05, 1.0646E+05, 1.0667E+05, 1.0689E+05,
     +  1.0710E+05, 1.0732E+05, 1.0753E+05, 1.0775E+05, 1.0797E+05,
     + 1.0818E+05, 1.0840E+05, 1.0862E+05, 1.0883E+05, 1.0905E+05,
     +  1.0927E+05, 1.0949E+05, 1.0971E+05, 1.0993E+05, 1.1015E+05,
     + 1.1037E+05, 1.1059E+05, 1.1081E+05, 1.1103E+05, 1.1125E+05,
     +  1.1147E+05, 1.1169E+05, 1.1192E+05, 1.1214E+05, 1.1236E+05,
     + 1.1258E+05, 1.1281E+05, 1.1303E+05, 1.1326E+05, 1.1348E+05,
     +  1.1371E+05, 1.1393E+05, 1.1416E+05, 1.1438E+05, 1.1461E+05,
     + 1.1483E+05, 1.1506E+05, 1.1529E+05, 1.1552E+05, 1.1574E+05,
     +  1.1597E+05, 1.1620E+05, 1.1643E+05, 1.1666E+05, 1.1689E+05,
     + 1.1712E+05, 1.1735E+05, 1.1758E+05, 1.1781E+05, 1.1804E+05,
     +  1.1827E+05, 1.1850E+05, 1.1874E+05, 1.1897E+05, 1.1920E+05,
     + 1.1944E+05, 1.1967E+05, 1.1990E+05, 1.2014E+05, 1.2037E+05,
     +  1.2061E+05, 1.2084E+05, 1.2108E+05, 1.2131E+05, 1.2155E+05,
     + 1.2179E+05, 1.2202E+05, 1.2226E+05, 1.2250E+05, 1.2274E+05,
     +  1.2297E+05, 1.2321E+05, 1.2345E+05, 1.2369E+05, 1.2393E+05,
     + 1.2417E+05, 1.2441E+05, 1.2465E+05, 1.2489E+05, 1.2514E+05,
     +  1.2538E+05, 1.2562E+05, 1.2586E+05, 1.2610E+05, 1.2635E+05,
     + 1.2659E+05, 1.2684E+05, 1.2708E+05, 1.2732E+05, 1.2757E+05,
     +  1.2781E+05, 1.2806E+05, 1.2831E+05, 1.2855E+05, 1.2880E+05,
     + 1.2905E+05, 1.2929E+05, 1.2954E+05, 1.2979E+05, 1.3004E+05,
     +  1.3029E+05, 1.3053E+05, 1.3078E+05, 1.3103E+05, 1.3128E+05,
     + 1.3153E+05, 1.3179E+05, 1.3204E+05, 1.3229E+05, 1.3254E+05,
     +  1.3279E+05, 1.3304E+05, 1.3330E+05, 1.3355E+05, 1.3380E+05,
     + 1.3406E+05, 1.3431E+05, 1.3457E+05, 1.3482E+05, 1.3508E+05,
     +  1.3533E+05, 1.3559E+05, 1.3585E+05, 1.3610E+05, 1.3636E+05,
     + 1.3662E+05, 1.3688E+05, 1.3713E+05, 1.3739E+05, 1.3765E+05,
     +  1.3791E+05, 1.3817E+05, 1.3843E+05, 1.3869E+05, 1.3895E+05,
     + 1.3921E+05, 1.3947E+05, 1.3974E+05, 1.4000E+05, 1.4026E+05,
     +  1.4052E+05, 1.4079E+05, 1.4105E+05, 1.4131E+05, 1.4158E+05,
     + 1.4184E+05, 1.4211E+05, 1.4237E+05, 1.4264E+05, 1.4291E+05,
     +  1.4317E+05, 1.4344E+05, 1.4371E+05, 1.4397E+05, 1.4424E+05,
     + 1.4451E+05, 1.4478E+05, 1.4505E+05, 1.4532E+05, 1.4559E+05,
     +  1.4586E+05, 1.4613E+05, 1.4640E+05, 1.4667E+05, 1.4694E+05,
     + 1.4721E+05, 1.4749E+05, 1.4776E+05, 1.4803E+05, 1.4831E+05,
     +  1.4858E+05, 1.4885E+05, 1.4913E+05, 1.4940E+05, 1.4968E+05,
     + 1.4995E+05, 1.5023E+05, 1.5051E+05, 1.5078E+05, 1.5106E+05,
     +  1.5134E+05, 1.5161E+05, 1.5189E+05, 1.5217E+05, 1.5245E+05,
     + 1.5273E+05, 1.5301E+05, 1.5329E+05, 1.5357E+05, 1.5385E+05,
     +  1.5413E+05, 1.5441E+05, 1.5470E+05, 1.5498E+05, 1.5526E+05,
     + 1.5554E+05, 1.5583E+05, 1.5611E+05, 1.5639E+05, 1.5668E+05,
     +  1.5696E+05, 1.5725E+05, 1.5753E+05, 1.5782E+05, 1.5811E+05,
     + 1.5839E+05, 1.5868E+05, 1.5897E+05, 1.5926E+05, 1.5954E+05,
     +  1.5983E+05, 1.6012E+05, 1.6041E+05, 1.6070E+05, 1.6099E+05,
     + 1.6128E+05, 1.6157E+05, 1.6186E+05, 1.6215E+05, 1.6245E+05,
     +  1.6274E+05, 1.6303E+05, 1.6332E+05, 1.6362E+05, 1.6391E+05,
     + 1.6421E+05, 1.6450E+05, 1.6480E+05, 1.6509E+05, 1.6539E+05,
     +  1.6568E+05, 1.6598E+05, 1.6628E+05, 1.6657E+05, 1.6687E+05,
     + 1.6717E+05, 1.6747E+05, 1.6777E+05, 1.6807E+05, 1.6836E+05,
     +  1.6866E+05, 1.6897E+05, 1.6927E+05, 1.6957E+05, 1.6987E+05,
     + 1.7017E+05, 1.7047E+05, 1.7077E+05, 1.7108E+05, 1.7138E+05,
     +  1.7168E+05, 1.7199E+05, 1.7229E+05, 1.7260E+05, 1.7290E+05,
     + 1.7321E+05, 1.7352E+05, 1.7382E+05, 1.7413E+05, 1.7444E+05,
     +  1.7474E+05, 1.7505E+05, 1.7536E+05, 1.7567E+05, 1.7598E+05,
     + 1.7629E+05, 1.7660E+05, 1.7691E+05, 1.7722E+05, 1.7753E+05,
     +  1.7784E+05, 1.7815E+05, 1.7846E+05, 1.7878E+05, 1.7909E+05,
     + 1.7940E+05, 1.7972E+05, 1.8003E+05, 1.8034E+05, 1.8066E+05,
     +  1.8098E+05, 1.8129E+05, 1.8161E+05, 1.8192E+05, 1.8224E+05,
     + 1.8256E+05, 1.8287E+05, 1.8319E+05, 1.8351E+05, 1.8383E+05,
     +  1.8415E+05, 1.8447E+05, 1.8479E+05, 1.8511E+05, 1.8543E+05,
     + 1.8575E+05, 1.8607E+05, 1.8639E+05, 1.8672E+05, 1.8704E+05,
     +  1.8736E+05, 1.8769E+05, 1.8801E+05, 1.8833E+05, 1.8866E+05,
     + 1.8898E+05, 1.8931E+05, 1.8963E+05, 1.8996E+05, 1.9029E+05,
     +  1.9061E+05, 1.9094E+05, 1.9127E+05, 1.9160E+05, 1.9193E+05,
     + 1.9226E+05, 1.9258E+05, 1.9291E+05, 1.9324E+05, 1.9358E+05,
     +  1.9391E+05, 1.9424E+05, 1.9457E+05, 1.9490E+05, 1.9523E+05,
     + 1.9557E+05, 1.9590E+05, 1.9623E+05, 1.9657E+05, 1.9690E+05,
     +  1.9724E+05, 1.9757E+05, 1.9791E+05, 1.9824E+05, 1.9858E+05,
     + 1.9892E+05, 1.9926E+05, 1.9959E+05, 1.9993E+05, 2.0027E+05,
     +  2.0061E+05, 2.0095E+05, 2.0129E+05, 2.0163E+05, 2.0197E+05,
     + 2.0231E+05, 2.0265E+05, 2.0299E+05, 2.0333E+05, 2.0368E+05,
     +  2.0402E+05, 2.0436E+05, 2.0471E+05, 2.0505E+05, 2.0539E+05,
     + 2.0574E+05, 2.0608E+05, 2.0643E+05, 2.0678E+05, 2.0712E+05,
     +  2.0747E+05, 2.0782E+05, 2.0817E+05, 2.0851E+05, 2.0886E+05,
     + 2.0921E+05, 2.0956E+05, 2.0991E+05, 2.1026E+05, 2.1061E+05,
     +  2.1096E+05, 2.1131E+05, 2.1166E+05, 2.1202E+05, 2.1237E+05,
     + 2.1272E+05, 2.1308E+05, 2.1343E+05, 2.1378E+05, 2.1414E+05,
     +  2.1449E+05, 2.1485E+05, 2.1520E+05, 2.1556E+05, 2.1592E+05,
     + 2.1627E+05, 2.1663E+05, 2.1699E+05, 2.1735E+05, 2.1771E+05,
     +  2.1806E+05, 2.1842E+05, 2.1878E+05, 2.1914E+05, 2.1950E+05,
     + 2.1987E+05, 2.2023E+05, 2.2059E+05, 2.2095E+05, 2.2131E+05,
     +  2.2168E+05, 2.2204E+05, 2.2240E+05, 2.2277E+05, 2.2313E+05,
     + 2.2350E+05, 2.2386E+05, 2.2423E+05, 2.2460E+05, 2.2496E+05,
     +  2.2533E+05, 2.2570E+05, 2.2607E+05, 2.2643E+05, 2.2680E+05,
     + 2.2717E+05, 2.2754E+05, 2.2791E+05, 2.2828E+05, 2.2865E+05,
     +  2.2902E+05, 2.2940E+05, 2.2977E+05, 2.3014E+05, 2.3051E+05,
     + 2.3089E+05, 2.3126E+05, 2.3164E+05, 2.3201E+05, 2.3238E+05,
     +  2.3276E+05, 2.3314E+05, 2.3351E+05, 2.3389E+05, 2.3427E+05,
     + 2.3464E+05, 2.3502E+05, 2.3540E+05, 2.3578E+05, 2.3616E+05,
     +  2.3654E+05, 2.3692E+05, 2.3730E+05, 2.3768E+05, 2.3806E+05,
     + 2.3844E+05, 2.3882E+05, 2.3921E+05, 2.3959E+05, 2.3997E+05,
     +  2.4035E+05, 2.4074E+05, 2.4112E+05, 2.4151E+05, 2.4189E+05,
     + 2.4228E+05, 2.4267E+05, 2.4305E+05, 2.4344E+05, 2.4383E+05,
     +  2.4421E+05, 2.4460E+05, 2.4499E+05, 2.4538E+05, 2.4577E+05,
     + 2.4616E+05, 2.4655E+05, 2.4694E+05, 2.4733E+05, 2.4772E+05,
     +  2.4812E+05, 2.4851E+05, 2.4890E+05, 2.4929E+05, 2.4969E+05,
     + 2.5008E+05, 2.5048E+05, 2.5087E+05, 2.5127E+05, 2.5166E+05,
     +  2.5206E+05, 2.5245E+05, 2.5285E+05, 2.5325E+05, 2.5365E+05,
     + 2.5405E+05, 2.5444E+05, 2.5484E+05, 2.5524E+05, 2.5564E+05,
     +  2.5604E+05, 2.5644E+05, 2.5684E+05, 2.5725E+05, 2.5765E+05,
     + 2.5805E+05, 2.5845E+05, 2.5886E+05, 2.5926E+05, 2.5966E+05,
     +  2.6007E+05, 2.6047E+05, 2.6088E+05, 2.6129E+05, 2.6169E+05,
     + 2.6210E+05, 2.6251E+05, 2.6291E+05, 2.6332E+05, 2.6373E+05,
     +  2.6414E+05, 2.6455E+05, 2.6496E+05, 2.6537E+05, 2.6578E+05,
     + 2.6619E+05, 2.6660E+05, 2.6701E+05, 2.6742E+05, 2.6784E+05,
     +  2.6825E+05, 2.6866E+05, 2.6908E+05, 2.6949E+05, 2.6991E+05,
     + 2.7032E+05, 2.7074E+05, 2.7115E+05, 2.7157E+05, 2.7199E+05,
     +  2.7240E+05, 2.7282E+05, 2.7324E+05, 2.7366E+05, 2.7408E+05,
     + 2.7449E+05, 2.7491E+05, 2.7533E+05, 2.7576E+05, 2.7618E+05,
     +  2.7660E+05, 2.7702E+05, 2.7744E+05, 2.7786E+05, 2.7829E+05,
     + 2.7871E+05, 2.7914E+05, 2.7956E+05, 2.7998E+05, 2.8041E+05,
     +  2.8084E+05, 2.8126E+05, 2.8169E+05, 2.8212E+05, 2.8254E+05,
     + 2.8297E+05, 2.8340E+05, 2.8383E+05, 2.8426E+05, 2.8469E+05,
     +  2.8512E+05, 2.8555E+05, 2.8598E+05, 2.8641E+05, 2.8684E+05,
     + 2.8727E+05, 2.8771E+05, 2.8814E+05, 2.8857E+05, 2.8901E+05,
     +  2.8944E+05, 2.8987E+05, 2.9031E+05, 2.9075E+05, 2.9118E+05,
     + 2.9162E+05, 2.9205E+05, 2.9249E+05, 2.9293E+05, 2.9337E+05,
     +  2.9381E+05, 2.9424E+05, 2.9468E+05, 2.9512E+05, 2.9556E+05,
     + 2.9601E+05, 2.9645E+05, 2.9689E+05, 2.9733E+05, 2.9777E+05,
     +  2.9821E+05, 2.9866E+05, 2.9910E+05, 2.9955E+05, 2.9999E+05,
     + 3.0044E+05, 3.0088E+05, 3.0133E+05, 3.0177E+05, 3.0222E+05,
     +  3.0267E+05, 3.0311E+05, 3.0356E+05, 3.0401E+05, 3.0446E+05,
     + 3.0491E+05, 3.0536E+05, 3.0581E+05, 3.0626E+05, 3.0671E+05,
     +  3.0716E+05, 3.0761E+05, 3.0807E+05, 3.0852E+05, 3.0897E+05,
     + 3.0943E+05, 3.0988E+05, 3.1033E+05, 3.1079E+05, 3.1124E+05,
     +  3.1170E+05, 3.1216E+05, 3.1261E+05, 3.1307E+05, 3.1353E+05,
     + 3.1398E+05, 3.1444E+05, 3.1490E+05, 3.1536E+05, 3.1582E+05,
     +  3.1628E+05, 3.1674E+05, 3.1720E+05, 3.1766E+05, 3.1813E+05,
     + 3.1859E+05, 3.1905E+05, 3.1951E+05, 3.1998E+05, 3.2044E+05,
     +  3.2090E+05, 3.2137E+05, 3.2184E+05, 3.2230E+05, 3.2277E+05,
     + 3.2323E+05, 3.2370E+05, 3.2417E+05, 3.2464E+05, 3.2510E+05,
     +  3.2557E+05, 3.2604E+05, 3.2651E+05, 3.2698E+05, 3.2745E+05,
     + 3.2792E+05, 3.2839E+05, 3.2887E+05, 3.2934E+05, 3.2981E+05,
     +  3.3028E+05, 3.3076E+05, 3.3123E+05, 3.3171E+05, 3.3218E+05,
     + 3.3266E+05, 3.3313E+05, 3.3361E+05, 3.3408E+05, 3.3456E+05,
     +  3.3504E+05, 3.3552E+05, 3.3599E+05, 3.3647E+05, 3.3695E+05,
     + 3.3743E+05, 3.3791E+05, 3.3839E+05, 3.3887E+05, 3.3935E+05,
     +  3.3984E+05, 3.4032E+05, 3.4080E+05, 3.4128E+05, 3.4177E+05,
     + 3.4225E+05, 3.4273E+05, 3.4322E+05, 3.4370E+05, 3.4419E+05,
     +  3.4468E+05, 3.4516E+05, 3.4565E+05, 3.4614E+05, 3.4663E+05,
     + 3.4711E+05, 3.4760E+05, 3.4809E+05, 3.4858E+05, 3.4907E+05,
     +  3.4956E+05, 3.5005E+05, 3.5054E+05, 3.5104E+05, 3.5153E+05,
     + 3.5202E+05, 3.5251E+05, 3.5301E+05, 3.5350E+05, 3.5400E+05,
     +  3.5449E+05, 3.5499E+05, 3.5548E+05, 3.5598E+05, 3.5647E+05,
     + 3.5697E+05, 3.5747E+05, 3.5797E+05, 3.5846E+05, 3.5896E+05,
     +  3.5946E+05, 3.5996E+05, 3.6046E+05, 3.6096E+05, 3.6146E+05,
     + 3.6196E+05, 3.6247E+05, 3.6297E+05, 3.6347E+05, 3.6397E+05,
     +  3.6448E+05, 3.6498E+05, 3.6549E+05, 3.6599E+05, 3.6650E+05,
     + 3.6700E+05, 3.6751E+05, 3.6801E+05, 3.6852E+05, 3.6903E+05,
     +  3.6954E+05, 3.7005E+05, 3.7055E+05, 3.7106E+05, 3.7157E+05,
     + 3.7208E+05, 3.7259E+05, 3.7310E+05, 3.7362E+05, 3.7413E+05,
     +  3.7464E+05, 3.7515E+05, 3.7567E+05, 3.7618E+05, 3.7669E+05,
     + 3.7721E+05, 3.7772E+05, 3.7824E+05, 3.7875E+05, 3.7927E+05,
     +  3.7979E+05, 3.8030E+05, 3.8082E+05, 3.8134E+05, 3.8186E+05,
     + 3.8238E+05, 3.8290E+05, 3.8342E+05, 3.8394E+05, 3.8446E+05,
     +  3.8498E+05, 3.8550E+05, 3.8602E+05, 3.8654E+05, 3.8707E+05,
     + 3.8759E+05, 3.8811E+05, 3.8864E+05, 3.8916E+05, 3.8969E+05,
     +  3.9021E+05, 3.9074E+05, 3.9126E+05, 3.9179E+05, 3.9232E+05,
     + 3.9285E+05, 3.9337E+05, 3.9390E+05, 3.9443E+05, 3.9496E+05,
     +  3.9549E+05, 3.9602E+05, 3.9655E+05, 3.9708E+05, 3.9761E+05,
     + 3.9815E+05, 3.9868E+05, 3.9921E+05, 3.9974E+05, 4.0028E+05,
     +  4.0081E+05, 4.0135E+05, 4.0188E+05, 4.0242E+05, 4.0295E+05,
     + 4.0349E+05, 4.0403E+05, 4.0456E+05, 4.0510E+05, 4.0564E+05,
     +  4.0618E+05, 4.0672E+05, 4.0726E+05, 4.0780E+05, 4.0834E+05,
     + 4.0888E+05, 4.0942E+05, 4.0996E+05, 4.1050E+05, 4.1104E+05,
     +  4.1159E+05, 4.1213E+05, 4.1267E+05, 4.1322E+05, 4.1376E+05,
     + 4.1431E+05, 4.1485E+05, 4.1540E+05, 4.1595E+05, 4.1649E+05,
     +  4.1704E+05, 4.1759E+05, 4.1814E+05, 4.1869E+05, 4.1924E+05,
     + 4.1978E+05, 4.2033E+05, 4.2089E+05, 4.2144E+05, 4.2199E+05,
     +  4.2254E+05, 4.2309E+05, 4.2364E+05, 4.2420E+05, 4.2475E+05,
     + 4.2530E+05, 4.2586E+05, 4.2641E+05, 4.2697E+05, 4.2752E+05,
     +  4.2808E+05, 4.2864E+05, 4.2919E+05, 4.2975E+05, 4.3031E+05/
      data (QofT( 4,J),J=2409,4510)/ 4.3087E+05, 4.3143E+05,
     +  4.3199E+05, 4.3254E+05, 4.3311E+05, 4.3367E+05, 4.3423E+05,
     +  4.3479E+05,
     + 4.3535E+05, 4.3591E+05, 4.3647E+05, 4.3704E+05, 4.3760E+05,
     +  4.3817E+05, 4.3873E+05, 4.3929E+05, 4.3986E+05, 4.4043E+05,
     + 4.4099E+05, 4.4156E+05, 4.4213E+05, 4.4269E+05, 4.4326E+05,
     +  4.4383E+05, 4.4440E+05, 4.4497E+05, 4.4554E+05, 4.4611E+05,
     + 4.4668E+05, 4.4725E+05, 4.4782E+05, 4.4839E+05, 4.4896E+05,
     +  4.4954E+05, 4.5011E+05, 4.5068E+05, 4.5126E+05, 4.5183E+05,
     + 4.5241E+05, 4.5298E+05, 4.5356E+05, 4.5414E+05, 4.5471E+05,
     +  4.5529E+05, 4.5587E+05, 4.5645E+05, 4.5702E+05, 4.5760E+05,
     + 4.5818E+05, 4.5876E+05, 4.5934E+05, 4.5992E+05, 4.6050E+05,
     +  4.6109E+05, 4.6167E+05, 4.6225E+05, 4.6283E+05, 4.6342E+05,
     + 4.6400E+05, 4.6458E+05, 4.6517E+05, 4.6575E+05, 4.6634E+05,
     +  4.6692E+05, 4.6751E+05, 4.6810E+05, 4.6868E+05, 4.6927E+05,
     + 4.6986E+05, 4.7045E+05, 4.7104E+05, 4.7163E+05, 4.7222E+05,
     +  4.7281E+05, 4.7340E+05, 4.7399E+05, 4.7458E+05, 4.7517E+05,
     + 4.7577E+05, 4.7636E+05, 4.7695E+05, 4.7755E+05, 4.7814E+05,
     +  4.7873E+05, 4.7933E+05, 4.7993E+05, 4.8052E+05, 4.8112E+05,
     + 4.8171E+05, 4.8231E+05, 4.8291E+05, 4.8351E+05, 4.8411E+05,
     +  4.8471E+05, 4.8531E+05, 4.8591E+05, 4.8651E+05, 4.8711E+05,
     + 4.8771E+05, 4.8831E+05, 4.8891E+05, 4.8951E+05, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       182
      data (QofT( 5,J),J=1,2408)/ 6.0000E+00, 6.0003E+00, 6.0111E+00,
     +  6.0709E+00, 6.2186E+00, 6.4702E+00, 6.8259E+00, 7.2791E+00,
     + 7.8214E+00, 8.4450E+00, 9.1423E+00, 9.9067E+00, 1.0733E+01,
     +  1.1614E+01, 1.2548E+01, 1.3529E+01, 1.4553E+01, 1.5619E+01,
     + 1.6722E+01, 1.7861E+01, 2.0238E+01, 2.2736E+01, 2.5345E+01,
     +  2.8058E+01, 3.0868E+01, 3.3772E+01, 3.6765E+01, 3.9844E+01,
     + 4.3007E+01, 4.6252E+01, 4.9575E+01, 5.2975E+01, 5.6451E+01,
     +  6.0001E+01, 6.3623E+01, 6.7315E+01, 7.1078E+01, 7.4908E+01,
     + 7.8806E+01, 8.2770E+01, 8.6798E+01, 9.0891E+01, 9.5046E+01,
     +  9.9263E+01, 1.0354E+02, 1.0788E+02, 1.1228E+02, 1.1674E+02,
     + 1.2125E+02, 1.2582E+02, 1.3045E+02, 1.3513E+02, 1.3987E+02,
     +  1.4467E+02, 1.4952E+02, 1.5442E+02, 1.5937E+02, 1.6438E+02,
     + 1.6944E+02, 1.7455E+02, 1.7971E+02, 1.8492E+02, 1.9018E+02,
     +  1.9549E+02, 2.0085E+02, 2.0625E+02, 2.1171E+02, 2.1721E+02,
     + 2.2276E+02, 2.2836E+02, 2.3400E+02, 2.3969E+02, 2.4543E+02,
     +  2.5121E+02, 2.5704E+02, 2.6291E+02, 2.6882E+02, 2.7478E+02,
     + 2.8079E+02, 2.8683E+02, 2.9292E+02, 2.9905E+02, 3.0523E+02,
     +  3.1145E+02, 3.1771E+02, 3.2401E+02, 3.3035E+02, 3.3674E+02,
     + 3.4316E+02, 3.4963E+02, 3.5614E+02, 3.6268E+02, 3.6927E+02,
     +  3.7590E+02, 3.8257E+02, 3.8927E+02, 3.9602E+02, 4.0280E+02,
     + 4.0963E+02, 4.1649E+02, 4.2339E+02, 4.3033E+02, 4.3730E+02,
     +  4.4432E+02, 4.5137E+02, 4.5846E+02, 4.6559E+02, 4.7275E+02,
     + 4.7995E+02, 4.8719E+02, 4.9447E+02, 5.0178E+02, 5.0913E+02,
     +  5.1651E+02, 5.2393E+02, 5.3139E+02, 5.3888E+02, 5.4640E+02,
     + 5.5397E+02, 5.6157E+02, 5.6920E+02, 5.7687E+02, 5.8457E+02,
     +  5.9231E+02, 6.0008E+02, 6.0789E+02, 6.1573E+02, 6.2361E+02,
     + 6.3152E+02, 6.3947E+02, 6.4745E+02, 6.5546E+02, 6.6351E+02,
     +  6.7159E+02, 6.7971E+02, 6.8786E+02, 6.9604E+02, 7.0426E+02,
     + 7.1251E+02, 7.2080E+02, 7.2911E+02, 7.3746E+02, 7.4585E+02,
     +  7.5427E+02, 7.6272E+02, 7.7120E+02, 7.7972E+02, 7.8827E+02,
     + 7.9685E+02, 8.0547E+02, 8.1411E+02, 8.2280E+02, 8.3151E+02,
     +  8.4026E+02, 8.4904E+02, 8.5785E+02, 8.6669E+02, 8.7557E+02,
     + 8.8448E+02, 8.9343E+02, 9.0240E+02, 9.1141E+02, 9.2045E+02,
     +  9.2952E+02, 9.3863E+02, 9.4777E+02, 9.5694E+02, 9.6614E+02,
     + 9.7538E+02, 9.8465E+02, 9.9395E+02, 1.0033E+03, 1.0126E+03,
     +  1.0220E+03, 1.0315E+03, 1.0409E+03, 1.0504E+03, 1.0600E+03,
     + 1.0695E+03, 1.0791E+03, 1.0887E+03, 1.0984E+03, 1.1081E+03,
     +  1.1178E+03, 1.1276E+03, 1.1374E+03, 1.1472E+03, 1.1570E+03,
     + 1.1669E+03, 1.1768E+03, 1.1868E+03, 1.1968E+03, 1.2068E+03,
     +  1.2168E+03, 1.2269E+03, 1.2370E+03, 1.2472E+03, 1.2574E+03,
     + 1.2676E+03, 1.2778E+03, 1.2881E+03, 1.2984E+03, 1.3088E+03,
     +  1.3192E+03, 1.3296E+03, 1.3400E+03, 1.3505E+03, 1.3610E+03,
     + 1.3716E+03, 1.3821E+03, 1.3928E+03, 1.4034E+03, 1.4141E+03,
     +  1.4248E+03, 1.4355E+03, 1.4463E+03, 1.4571E+03, 1.4680E+03,
     + 1.4789E+03, 1.4898E+03, 1.5007E+03, 1.5117E+03, 1.5228E+03,
     +  1.5338E+03, 1.5449E+03, 1.5560E+03, 1.5672E+03, 1.5784E+03,
     + 1.5896E+03, 1.6009E+03, 1.6122E+03, 1.6235E+03, 1.6348E+03,
     +  1.6462E+03, 1.6577E+03, 1.6692E+03, 1.6807E+03, 1.6922E+03,
     + 1.7038E+03, 1.7154E+03, 1.7270E+03, 1.7387E+03, 1.7504E+03,
     +  1.7622E+03, 1.7740E+03, 1.7858E+03, 1.7977E+03, 1.8095E+03,
     + 1.8215E+03, 1.8334E+03, 1.8454E+03, 1.8575E+03, 1.8696E+03,
     +  1.8817E+03, 1.8938E+03, 1.9060E+03, 1.9182E+03, 1.9305E+03,
     + 1.9428E+03, 1.9551E+03, 1.9675E+03, 1.9799E+03, 1.9923E+03,
     +  2.0048E+03, 2.0173E+03, 2.0299E+03, 2.0425E+03, 2.0551E+03,
     + 2.0678E+03, 2.0805E+03, 2.0932E+03, 2.1060E+03, 2.1188E+03,
     +  2.1317E+03, 2.1445E+03, 2.1575E+03, 2.1704E+03, 2.1834E+03,
     + 2.1965E+03, 2.2096E+03, 2.2227E+03, 2.2359E+03, 2.2491E+03,
     +  2.2623E+03, 2.2756E+03, 2.2889E+03, 2.3022E+03, 2.3156E+03,
     + 2.3291E+03, 2.3425E+03, 2.3561E+03, 2.3696E+03, 2.3832E+03,
     +  2.3968E+03, 2.4105E+03, 2.4242E+03, 2.4380E+03, 2.4518E+03,
     + 2.4656E+03, 2.4795E+03, 2.4934E+03, 2.5073E+03, 2.5213E+03,
     +  2.5354E+03, 2.5494E+03, 2.5635E+03, 2.5777E+03, 2.5919E+03,
     + 2.6061E+03, 2.6204E+03, 2.6347E+03, 2.6491E+03, 2.6635E+03,
     +  2.6780E+03, 2.6924E+03, 2.7070E+03, 2.7215E+03, 2.7362E+03,
     + 2.7508E+03, 2.7655E+03, 2.7802E+03, 2.7950E+03, 2.8099E+03,
     +  2.8247E+03, 2.8396E+03, 2.8546E+03, 2.8696E+03, 2.8846E+03,
     + 2.8997E+03, 2.9148E+03, 2.9300E+03, 2.9452E+03, 2.9605E+03,
     +  2.9758E+03, 2.9911E+03, 3.0065E+03, 3.0219E+03, 3.0374E+03,
     + 3.0529E+03, 3.0685E+03, 3.0841E+03, 3.0998E+03, 3.1155E+03,
     +  3.1312E+03, 3.1470E+03, 3.1628E+03, 3.1787E+03, 3.1946E+03,
     + 3.2106E+03, 3.2266E+03, 3.2427E+03, 3.2588E+03, 3.2750E+03,
     +  3.2912E+03, 3.3074E+03, 3.3237E+03, 3.3400E+03, 3.3564E+03,
     + 3.3728E+03, 3.3893E+03, 3.4058E+03, 3.4224E+03, 3.4390E+03,
     +  3.4557E+03, 3.4724E+03, 3.4892E+03, 3.5060E+03, 3.5228E+03,
     + 3.5397E+03, 3.5567E+03, 3.5737E+03, 3.5907E+03, 3.6078E+03,
     +  3.6250E+03, 3.6421E+03, 3.6594E+03, 3.6767E+03, 3.6940E+03,
     + 3.7114E+03, 3.7288E+03, 3.7463E+03, 3.7638E+03, 3.7814E+03,
     +  3.7991E+03, 3.8167E+03, 3.8345E+03, 3.8522E+03, 3.8701E+03,
     + 3.8880E+03, 3.9059E+03, 3.9239E+03, 3.9419E+03, 3.9600E+03,
     +  3.9781E+03, 3.9963E+03, 4.0145E+03, 4.0328E+03, 4.0511E+03,
     + 4.0695E+03, 4.0880E+03, 4.1065E+03, 4.1250E+03, 4.1436E+03,
     +  4.1622E+03, 4.1809E+03, 4.1997E+03, 4.2185E+03, 4.2374E+03,
     + 4.2563E+03, 4.2752E+03, 4.2942E+03, 4.3133E+03, 4.3324E+03,
     +  4.3516E+03, 4.3708E+03, 4.3901E+03, 4.4094E+03, 4.4288E+03,
     + 4.4483E+03, 4.4677E+03, 4.4873E+03, 4.5069E+03, 4.5266E+03,
     +  4.5463E+03, 4.5660E+03, 4.5859E+03, 4.6057E+03, 4.6257E+03,
     + 4.6456E+03, 4.6657E+03, 4.6858E+03, 4.7059E+03, 4.7261E+03,
     +  4.7464E+03, 4.7667E+03, 4.7871E+03, 4.8075E+03, 4.8280E+03,
     + 4.8486E+03, 4.8692E+03, 4.8898E+03, 4.9105E+03, 4.9313E+03,
     +  4.9521E+03, 4.9730E+03, 4.9940E+03, 5.0150E+03, 5.0360E+03,
     + 5.0571E+03, 5.0783E+03, 5.0996E+03, 5.1208E+03, 5.1422E+03,
     +  5.1636E+03, 5.1851E+03, 5.2066E+03, 5.2282E+03, 5.2498E+03,
     + 5.2715E+03, 5.2933E+03, 5.3151E+03, 5.3370E+03, 5.3589E+03,
     +  5.3809E+03, 5.4030E+03, 5.4251E+03, 5.4473E+03, 5.4695E+03,
     + 5.4918E+03, 5.5142E+03, 5.5366E+03, 5.5591E+03, 5.5817E+03,
     +  5.6043E+03, 5.6269E+03, 5.6497E+03, 5.6725E+03, 5.6953E+03,
     + 5.7182E+03, 5.7412E+03, 5.7643E+03, 5.7874E+03, 5.8105E+03,
     +  5.8338E+03, 5.8571E+03, 5.8804E+03, 5.9038E+03, 5.9273E+03,
     + 5.9509E+03, 5.9745E+03, 5.9981E+03, 6.0219E+03, 6.0457E+03,
     +  6.0695E+03, 6.0935E+03, 6.1175E+03, 6.1415E+03, 6.1656E+03,
     + 6.1898E+03, 6.2141E+03, 6.2384E+03, 6.2628E+03, 6.2872E+03,
     +  6.3118E+03, 6.3363E+03, 6.3610E+03, 6.3857E+03, 6.4105E+03,
     + 6.4353E+03, 6.4602E+03, 6.4852E+03, 6.5103E+03, 6.5354E+03,
     +  6.5606E+03, 6.5858E+03, 6.6111E+03, 6.6365E+03, 6.6620E+03,
     + 6.6875E+03, 6.7131E+03, 6.7387E+03, 6.7644E+03, 6.7902E+03,
     +  6.8161E+03, 6.8420E+03, 6.8680E+03, 6.8941E+03, 6.9202E+03,
     + 6.9464E+03, 6.9727E+03, 6.9991E+03, 7.0255E+03, 7.0520E+03,
     +  7.0785E+03, 7.1051E+03, 7.1318E+03, 7.1586E+03, 7.1854E+03,
     + 7.2123E+03, 7.2393E+03, 7.2664E+03, 7.2935E+03, 7.3207E+03,
     +  7.3480E+03, 7.3753E+03, 7.4027E+03, 7.4302E+03, 7.4578E+03,
     + 7.4854E+03, 7.5131E+03, 7.5409E+03, 7.5687E+03, 7.5966E+03,
     +  7.6246E+03, 7.6527E+03, 7.6808E+03, 7.7090E+03, 7.7373E+03,
     + 7.7657E+03, 7.7941E+03, 7.8226E+03, 7.8512E+03, 7.8799E+03,
     +  7.9086E+03, 7.9374E+03, 7.9663E+03, 7.9953E+03, 8.0243E+03,
     + 8.0534E+03, 8.0826E+03, 8.1119E+03, 8.1412E+03, 8.1706E+03,
     +  8.2001E+03, 8.2297E+03, 8.2593E+03, 8.2891E+03, 8.3189E+03,
     + 8.3487E+03, 8.3787E+03, 8.4087E+03, 8.4388E+03, 8.4690E+03,
     +  8.4993E+03, 8.5296E+03, 8.5601E+03, 8.5906E+03, 8.6212E+03,
     + 8.6518E+03, 8.6826E+03, 8.7134E+03, 8.7443E+03, 8.7753E+03,
     +  8.8063E+03, 8.8374E+03, 8.8687E+03, 8.9000E+03, 8.9313E+03,
     + 8.9628E+03, 8.9943E+03, 9.0260E+03, 9.0577E+03, 9.0895E+03,
     +  9.1213E+03, 9.1533E+03, 9.1853E+03, 9.2174E+03, 9.2496E+03,
     + 9.2819E+03, 9.3142E+03, 9.3467E+03, 9.3792E+03, 9.4118E+03,
     +  9.4445E+03, 9.4773E+03, 9.5101E+03, 9.5430E+03, 9.5761E+03,
     + 9.6092E+03, 9.6424E+03, 9.6756E+03, 9.7090E+03, 9.7424E+03,
     +  9.7760E+03, 9.8096E+03, 9.8433E+03, 9.8771E+03, 9.9109E+03,
     + 9.9449E+03, 9.9789E+03, 1.0013E+04, 1.0047E+04, 1.0082E+04,
     +  1.0116E+04, 1.0150E+04, 1.0185E+04, 1.0220E+04, 1.0254E+04,
     + 1.0289E+04, 1.0324E+04, 1.0359E+04, 1.0394E+04, 1.0429E+04,
     +  1.0465E+04, 1.0500E+04, 1.0535E+04, 1.0571E+04, 1.0607E+04,
     + 1.0642E+04, 1.0678E+04, 1.0714E+04, 1.0750E+04, 1.0786E+04,
     +  1.0822E+04, 1.0858E+04, 1.0895E+04, 1.0931E+04, 1.0968E+04,
     + 1.1004E+04, 1.1041E+04, 1.1078E+04, 1.1115E+04, 1.1152E+04,
     +  1.1189E+04, 1.1226E+04, 1.1263E+04, 1.1300E+04, 1.1338E+04,
     + 1.1375E+04, 1.1413E+04, 1.1451E+04, 1.1488E+04, 1.1526E+04,
     +  1.1564E+04, 1.1602E+04, 1.1641E+04, 1.1679E+04, 1.1717E+04,
     + 1.1756E+04, 1.1794E+04, 1.1833E+04, 1.1872E+04, 1.1910E+04,
     +  1.1949E+04, 1.1988E+04, 1.2028E+04, 1.2067E+04, 1.2106E+04,
     + 1.2146E+04, 1.2185E+04, 1.2225E+04, 1.2264E+04, 1.2304E+04,
     +  1.2344E+04, 1.2384E+04, 1.2424E+04, 1.2464E+04, 1.2505E+04,
     + 1.2545E+04, 1.2585E+04, 1.2626E+04, 1.2667E+04, 1.2708E+04,
     +  1.2748E+04, 1.2789E+04, 1.2830E+04, 1.2872E+04, 1.2913E+04,
     + 1.2954E+04, 1.2996E+04, 1.3037E+04, 1.3079E+04, 1.3121E+04,
     +  1.3163E+04, 1.3205E+04, 1.3247E+04, 1.3289E+04, 1.3331E+04,
     + 1.3374E+04, 1.3416E+04, 1.3459E+04, 1.3501E+04, 1.3544E+04,
     +  1.3587E+04, 1.3630E+04, 1.3673E+04, 1.3716E+04, 1.3760E+04,
     + 1.3803E+04, 1.3846E+04, 1.3890E+04, 1.3934E+04, 1.3978E+04,
     +  1.4022E+04, 1.4066E+04, 1.4110E+04, 1.4154E+04, 1.4198E+04,
     + 1.4243E+04, 1.4287E+04, 1.4332E+04, 1.4377E+04, 1.4422E+04,
     +  1.4467E+04, 1.4512E+04, 1.4557E+04, 1.4602E+04, 1.4648E+04,
     + 1.4693E+04, 1.4739E+04, 1.4784E+04, 1.4830E+04, 1.4876E+04,
     +  1.4922E+04, 1.4968E+04, 1.5015E+04, 1.5061E+04, 1.5108E+04,
     + 1.5154E+04, 1.5201E+04, 1.5248E+04, 1.5295E+04, 1.5342E+04,
     +  1.5389E+04, 1.5436E+04, 1.5483E+04, 1.5531E+04, 1.5579E+04,
     + 1.5626E+04, 1.5674E+04, 1.5722E+04, 1.5770E+04, 1.5818E+04,
     +  1.5866E+04, 1.5915E+04, 1.5963E+04, 1.6012E+04, 1.6061E+04,
     + 1.6109E+04, 1.6158E+04, 1.6207E+04, 1.6257E+04, 1.6306E+04,
     +  1.6355E+04, 1.6405E+04, 1.6454E+04, 1.6504E+04, 1.6554E+04,
     + 1.6604E+04, 1.6654E+04, 1.6704E+04, 1.6754E+04, 1.6805E+04,
     +  1.6855E+04, 1.6906E+04, 1.6957E+04, 1.7008E+04, 1.7059E+04,
     + 1.7110E+04, 1.7161E+04, 1.7212E+04, 1.7264E+04, 1.7315E+04,
     +  1.7367E+04, 1.7419E+04, 1.7471E+04, 1.7523E+04, 1.7575E+04,
     + 1.7627E+04, 1.7680E+04, 1.7732E+04, 1.7785E+04, 1.7838E+04,
     +  1.7891E+04, 1.7944E+04, 1.7997E+04, 1.8050E+04, 1.8104E+04,
     + 1.8157E+04, 1.8211E+04, 1.8264E+04, 1.8318E+04, 1.8372E+04,
     +  1.8426E+04, 1.8481E+04, 1.8535E+04, 1.8589E+04, 1.8644E+04,
     + 1.8699E+04, 1.8753E+04, 1.8808E+04, 1.8864E+04, 1.8919E+04,
     +  1.8974E+04, 1.9030E+04, 1.9085E+04, 1.9141E+04, 1.9197E+04,
     + 1.9253E+04, 1.9309E+04, 1.9365E+04, 1.9421E+04, 1.9478E+04,
     +  1.9534E+04, 1.9591E+04, 1.9648E+04, 1.9705E+04, 1.9762E+04,
     + 1.9819E+04, 1.9876E+04, 1.9934E+04, 1.9991E+04, 2.0049E+04,
     +  2.0107E+04, 2.0165E+04, 2.0223E+04, 2.0281E+04, 2.0340E+04,
     + 2.0398E+04, 2.0457E+04, 2.0515E+04, 2.0574E+04, 2.0633E+04,
     +  2.0692E+04, 2.0752E+04, 2.0811E+04, 2.0871E+04, 2.0930E+04,
     + 2.0990E+04, 2.1050E+04, 2.1110E+04, 2.1170E+04, 2.1231E+04,
     +  2.1291E+04, 2.1352E+04, 2.1412E+04, 2.1473E+04, 2.1534E+04,
     + 2.1595E+04, 2.1656E+04, 2.1718E+04, 2.1779E+04, 2.1841E+04,
     +  2.1903E+04, 2.1965E+04, 2.2027E+04, 2.2089E+04, 2.2151E+04,
     + 2.2214E+04, 2.2276E+04, 2.2339E+04, 2.2402E+04, 2.2465E+04,
     +  2.2528E+04, 2.2591E+04, 2.2654E+04, 2.2718E+04, 2.2782E+04,
     + 2.2845E+04, 2.2909E+04, 2.2974E+04, 2.3038E+04, 2.3102E+04,
     +  2.3167E+04, 2.3231E+04, 2.3296E+04, 2.3361E+04, 2.3426E+04,
     + 2.3491E+04, 2.3556E+04, 2.3622E+04, 2.3688E+04, 2.3753E+04,
     +  2.3819E+04, 2.3885E+04, 2.3951E+04, 2.4018E+04, 2.4084E+04,
     + 2.4151E+04, 2.4217E+04, 2.4284E+04, 2.4351E+04, 2.4418E+04,
     +  2.4486E+04, 2.4553E+04, 2.4621E+04, 2.4688E+04, 2.4756E+04,
     + 2.4824E+04, 2.4892E+04, 2.4961E+04, 2.5029E+04, 2.5098E+04,
     +  2.5167E+04, 2.5235E+04, 2.5304E+04, 2.5374E+04, 2.5443E+04,
     + 2.5512E+04, 2.5582E+04, 2.5652E+04, 2.5722E+04, 2.5792E+04,
     +  2.5862E+04, 2.5932E+04, 2.6003E+04, 2.6073E+04, 2.6144E+04,
     + 2.6215E+04, 2.6286E+04, 2.6357E+04, 2.6429E+04, 2.6500E+04,
     +  2.6572E+04, 2.6644E+04, 2.6716E+04, 2.6788E+04, 2.6860E+04,
     + 2.6932E+04, 2.7005E+04, 2.7078E+04, 2.7150E+04, 2.7223E+04,
     +  2.7297E+04, 2.7370E+04, 2.7443E+04, 2.7517E+04, 2.7591E+04,
     + 2.7665E+04, 2.7739E+04, 2.7813E+04, 2.7887E+04, 2.7962E+04,
     +  2.8036E+04, 2.8111E+04, 2.8186E+04, 2.8261E+04, 2.8337E+04,
     + 2.8412E+04, 2.8488E+04, 2.8564E+04, 2.8639E+04, 2.8715E+04,
     +  2.8792E+04, 2.8868E+04, 2.8945E+04, 2.9021E+04, 2.9098E+04,
     + 2.9175E+04, 2.9252E+04, 2.9330E+04, 2.9407E+04, 2.9485E+04,
     +  2.9562E+04, 2.9640E+04, 2.9718E+04, 2.9797E+04, 2.9875E+04,
     + 2.9954E+04, 3.0032E+04, 3.0111E+04, 3.0190E+04, 3.0270E+04,
     +  3.0349E+04, 3.0428E+04, 3.0508E+04, 3.0588E+04, 3.0668E+04,
     + 3.0748E+04, 3.0829E+04, 3.0909E+04, 3.0990E+04, 3.1070E+04,
     +  3.1151E+04, 3.1233E+04, 3.1314E+04, 3.1395E+04, 3.1477E+04,
     + 3.1559E+04, 3.1641E+04, 3.1723E+04, 3.1805E+04, 3.1887E+04,
     +  3.1970E+04, 3.2053E+04, 3.2136E+04, 3.2219E+04, 3.2302E+04,
     + 3.2386E+04, 3.2469E+04, 3.2553E+04, 3.2637E+04, 3.2721E+04,
     +  3.2805E+04, 3.2890E+04, 3.2974E+04, 3.3059E+04, 3.3144E+04,
     + 3.3229E+04, 3.3314E+04, 3.3400E+04, 3.3485E+04, 3.3571E+04,
     +  3.3657E+04, 3.3743E+04, 3.3829E+04, 3.3916E+04, 3.4002E+04,
     + 3.4089E+04, 3.4176E+04, 3.4263E+04, 3.4350E+04, 3.4438E+04,
     +  3.4525E+04, 3.4613E+04, 3.4701E+04, 3.4789E+04, 3.4878E+04,
     + 3.4966E+04, 3.5055E+04, 3.5144E+04, 3.5233E+04, 3.5322E+04,
     +  3.5411E+04, 3.5501E+04, 3.5590E+04, 3.5680E+04, 3.5770E+04,
     + 3.5860E+04, 3.5951E+04, 3.6041E+04, 3.6132E+04, 3.6223E+04,
     +  3.6314E+04, 3.6405E+04, 3.6497E+04, 3.6588E+04, 3.6680E+04,
     + 3.6772E+04, 3.6864E+04, 3.6957E+04, 3.7049E+04, 3.7142E+04,
     +  3.7235E+04, 3.7328E+04, 3.7421E+04, 3.7514E+04, 3.7608E+04,
     + 3.7701E+04, 3.7795E+04, 3.7890E+04, 3.7984E+04, 3.8078E+04,
     +  3.8173E+04, 3.8268E+04, 3.8363E+04, 3.8458E+04, 3.8553E+04,
     + 3.8649E+04, 3.8744E+04, 3.8840E+04, 3.8936E+04, 3.9033E+04,
     +  3.9129E+04, 3.9226E+04, 3.9323E+04, 3.9420E+04, 3.9517E+04,
     + 3.9614E+04, 3.9712E+04, 3.9809E+04, 3.9907E+04, 4.0005E+04,
     +  4.0104E+04, 4.0202E+04, 4.0301E+04, 4.0400E+04, 4.0499E+04,
     + 4.0598E+04, 4.0697E+04, 4.0797E+04, 4.0897E+04, 4.0997E+04,
     +  4.1097E+04, 4.1197E+04, 4.1298E+04, 4.1398E+04, 4.1499E+04,
     + 4.1600E+04, 4.1701E+04, 4.1803E+04, 4.1905E+04, 4.2006E+04,
     +  4.2108E+04, 4.2211E+04, 4.2313E+04, 4.2416E+04, 4.2518E+04,
     + 4.2621E+04, 4.2724E+04, 4.2828E+04, 4.2931E+04, 4.3035E+04,
     +  4.3139E+04, 4.3243E+04, 4.3347E+04, 4.3452E+04, 4.3557E+04,
     + 4.3662E+04, 4.3767E+04, 4.3872E+04, 4.3977E+04, 4.4083E+04,
     +  4.4189E+04, 4.4295E+04, 4.4401E+04, 4.4508E+04, 4.4614E+04,
     + 4.4721E+04, 4.4828E+04, 4.4935E+04, 4.5043E+04, 4.5150E+04,
     +  4.5258E+04, 4.5366E+04, 4.5474E+04, 4.5583E+04, 4.5691E+04,
     + 4.5800E+04, 4.5909E+04, 4.6018E+04, 4.6128E+04, 4.6237E+04,
     +  4.6347E+04, 4.6457E+04, 4.6567E+04, 4.6677E+04, 4.6788E+04,
     + 4.6899E+04, 4.7010E+04, 4.7121E+04, 4.7232E+04, 4.7344E+04,
     +  4.7456E+04, 4.7568E+04, 4.7680E+04, 4.7792E+04, 4.7905E+04,
     + 4.8018E+04, 4.8131E+04, 4.8244E+04, 4.8357E+04, 4.8471E+04,
     +  4.8585E+04, 4.8699E+04, 4.8813E+04, 4.8927E+04, 4.9042E+04,
     + 4.9157E+04, 4.9272E+04, 4.9387E+04, 4.9502E+04, 4.9618E+04,
     +  4.9734E+04, 4.9850E+04, 4.9966E+04, 5.0083E+04, 5.0199E+04,
     + 5.0316E+04, 5.0433E+04, 5.0551E+04, 5.0668E+04, 5.0786E+04,
     +  5.0904E+04, 5.1022E+04, 5.1140E+04, 5.1259E+04, 5.1378E+04,
     + 5.1497E+04, 5.1616E+04, 5.1735E+04, 5.1855E+04, 5.1974E+04,
     +  5.2095E+04, 5.2215E+04, 5.2335E+04, 5.2456E+04, 5.2577E+04,
     + 5.2698E+04, 5.2819E+04, 5.2941E+04, 5.3062E+04, 5.3184E+04,
     +  5.3306E+04, 5.3429E+04, 5.3551E+04, 5.3674E+04, 5.3797E+04,
     + 5.3920E+04, 5.4044E+04, 5.4167E+04, 5.4291E+04, 5.4415E+04,
     +  5.4540E+04, 5.4664E+04, 5.4789E+04, 5.4914E+04, 5.5039E+04,
     + 5.5164E+04, 5.5290E+04, 5.5416E+04, 5.5542E+04, 5.5668E+04,
     +  5.5794E+04, 5.5921E+04, 5.6048E+04, 5.6175E+04, 5.6302E+04,
     + 5.6430E+04, 5.6558E+04, 5.6686E+04, 5.6814E+04, 5.6942E+04,
     +  5.7071E+04, 5.7200E+04, 5.7329E+04, 5.7458E+04, 5.7588E+04,
     + 5.7718E+04, 5.7848E+04, 5.7978E+04, 5.8108E+04, 5.8239E+04,
     +  5.8370E+04, 5.8501E+04, 5.8632E+04, 5.8764E+04, 5.8895E+04,
     + 5.9027E+04, 5.9160E+04, 5.9292E+04, 5.9425E+04, 5.9558E+04,
     +  5.9691E+04, 5.9824E+04, 5.9958E+04, 6.0092E+04, 6.0226E+04,
     + 6.0360E+04, 6.0494E+04, 6.0629E+04, 6.0764E+04, 6.0899E+04,
     +  6.1035E+04, 6.1170E+04, 6.1306E+04, 6.1442E+04, 6.1578E+04,
     + 6.1715E+04, 6.1852E+04, 6.1989E+04, 6.2126E+04, 6.2263E+04,
     +  6.2401E+04, 6.2539E+04, 6.2677E+04, 6.2815E+04, 6.2954E+04,
     + 6.3093E+04, 6.3232E+04, 6.3371E+04, 6.3511E+04, 6.3651E+04,
     +  6.3791E+04, 6.3931E+04, 6.4071E+04, 6.4212E+04, 6.4353E+04,
     + 6.4494E+04, 6.4636E+04, 6.4777E+04, 6.4919E+04, 6.5061E+04,
     +  6.5204E+04, 6.5346E+04, 6.5489E+04, 6.5632E+04, 6.5775E+04,
     + 6.5919E+04, 6.6063E+04, 6.6207E+04, 6.6351E+04, 6.6496E+04,
     +  6.6640E+04, 6.6785E+04, 6.6930E+04, 6.7076E+04, 6.7222E+04,
     + 6.7367E+04, 6.7514E+04, 6.7660E+04, 6.7807E+04, 6.7954E+04,
     +  6.8101E+04, 6.8248E+04, 6.8396E+04, 6.8544E+04, 6.8692E+04,
     + 6.8840E+04, 6.8989E+04, 6.9137E+04, 6.9286E+04, 6.9436E+04,
     +  6.9585E+04, 6.9735E+04, 6.9885E+04, 7.0035E+04, 7.0186E+04,
     + 7.0337E+04, 7.0488E+04, 7.0639E+04, 7.0790E+04, 7.0942E+04,
     +  7.1094E+04, 7.1246E+04, 7.1399E+04, 7.1552E+04, 7.1705E+04,
     + 7.1858E+04, 7.2011E+04, 7.2165E+04, 7.2319E+04, 7.2473E+04,
     +  7.2628E+04, 7.2783E+04, 7.2938E+04, 7.3093E+04, 7.3248E+04,
     + 7.3404E+04, 7.3560E+04, 7.3716E+04, 7.3873E+04, 7.4029E+04,
     +  7.4186E+04, 7.4344E+04, 7.4501E+04, 7.4659E+04, 7.4817E+04,
     + 7.4975E+04, 7.5134E+04, 7.5292E+04, 7.5451E+04, 7.5610E+04,
     +  7.5770E+04, 7.5930E+04, 7.6090E+04, 7.6250E+04, 7.6411E+04,
     + 7.6571E+04, 7.6732E+04, 7.6894E+04, 7.7055E+04, 7.7217E+04,
     +  7.7379E+04, 7.7541E+04, 7.7704E+04, 7.7867E+04, 7.8030E+04,
     + 7.8193E+04, 7.8357E+04, 7.8521E+04, 7.8685E+04, 7.8849E+04,
     +  7.9014E+04, 7.9179E+04, 7.9344E+04, 7.9509E+04, 7.9675E+04,
     + 7.9841E+04, 8.0007E+04, 8.0174E+04, 8.0340E+04, 8.0507E+04,
     +  8.0675E+04, 8.0842E+04, 8.1010E+04, 8.1178E+04, 8.1346E+04,
     + 8.1515E+04, 8.1684E+04, 8.1853E+04, 8.2022E+04, 8.2192E+04,
     +  8.2362E+04, 8.2532E+04, 8.2702E+04, 8.2873E+04, 8.3044E+04,
     + 8.3215E+04, 8.3386E+04, 8.3558E+04, 8.3730E+04, 8.3902E+04,
     +  8.4075E+04, 8.4248E+04, 8.4421E+04, 8.4594E+04, 8.4768E+04,
     + 8.4942E+04, 8.5116E+04, 8.5290E+04, 8.5465E+04, 8.5640E+04,
     +  8.5815E+04, 8.5991E+04, 8.6166E+04, 8.6342E+04, 8.6519E+04,
     + 8.6695E+04, 8.6872E+04, 8.7049E+04, 8.7227E+04, 8.7404E+04,
     +  8.7582E+04, 8.7760E+04, 8.7939E+04, 8.8118E+04, 8.8297E+04,
     + 8.8476E+04, 8.8656E+04, 8.8835E+04, 8.9016E+04, 8.9196E+04,
     +  8.9377E+04, 8.9558E+04, 8.9739E+04, 8.9920E+04, 9.0102E+04,
     + 9.0284E+04, 9.0467E+04, 9.0649E+04, 9.0832E+04, 9.1015E+04,
     +  9.1199E+04, 9.1382E+04, 9.1566E+04, 9.1751E+04, 9.1935E+04,
     + 9.2120E+04, 9.2305E+04, 9.2491E+04, 9.2676E+04, 9.2862E+04,
     +  9.3049E+04, 9.3235E+04, 9.3422E+04, 9.3609E+04, 9.3796E+04,
     + 9.3984E+04, 9.4172E+04, 9.4360E+04, 9.4549E+04, 9.4737E+04,
     +  9.4926E+04, 9.5116E+04, 9.5305E+04, 9.5495E+04, 9.5686E+04,
     + 9.5876E+04, 9.6067E+04, 9.6258E+04, 9.6449E+04, 9.6641E+04,
     +  9.6833E+04, 9.7025E+04, 9.7217E+04, 9.7410E+04, 9.7603E+04,
     + 9.7797E+04, 9.7990E+04, 9.8184E+04, 9.8378E+04, 9.8573E+04,
     +  9.8768E+04, 9.8963E+04, 9.9158E+04, 9.9354E+04, 9.9550E+04,
     + 9.9746E+04, 9.9942E+04, 1.0014E+05, 1.0034E+05, 1.0053E+05,
     +  1.0073E+05, 1.0093E+05, 1.0113E+05, 1.0133E+05, 1.0153E+05,
     + 1.0172E+05, 1.0192E+05, 1.0212E+05, 1.0232E+05, 1.0252E+05,
     +  1.0272E+05, 1.0293E+05, 1.0313E+05, 1.0333E+05, 1.0353E+05,
     + 1.0373E+05, 1.0393E+05, 1.0414E+05, 1.0434E+05, 1.0454E+05,
     +  1.0475E+05, 1.0495E+05, 1.0515E+05, 1.0536E+05, 1.0556E+05,
     + 1.0577E+05, 1.0597E+05, 1.0618E+05, 1.0639E+05, 1.0659E+05,
     +  1.0680E+05, 1.0701E+05, 1.0721E+05, 1.0742E+05, 1.0763E+05,
     + 1.0784E+05, 1.0804E+05, 1.0825E+05, 1.0846E+05, 1.0867E+05,
     +  1.0888E+05, 1.0909E+05, 1.0930E+05, 1.0951E+05, 1.0972E+05,
     + 1.0993E+05, 1.1014E+05, 1.1036E+05, 1.1057E+05, 1.1078E+05,
     +  1.1099E+05, 1.1121E+05, 1.1142E+05, 1.1163E+05, 1.1185E+05,
     + 1.1206E+05, 1.1228E+05, 1.1249E+05, 1.1271E+05, 1.1292E+05,
     +  1.1314E+05, 1.1335E+05, 1.1357E+05, 1.1379E+05, 1.1400E+05,
     + 1.1422E+05, 1.1444E+05, 1.1465E+05, 1.1487E+05, 1.1509E+05,
     +  1.1531E+05, 1.1553E+05, 1.1575E+05, 1.1597E+05, 1.1619E+05,
     + 1.1641E+05, 1.1663E+05, 1.1685E+05, 1.1707E+05, 1.1729E+05,
     +  1.1752E+05, 1.1774E+05, 1.1796E+05, 1.1818E+05, 1.1841E+05,
     + 1.1863E+05, 1.1885E+05, 1.1908E+05, 1.1930E+05, 1.1953E+05,
     +  1.1975E+05, 1.1998E+05, 1.2020E+05, 1.2043E+05, 1.2065E+05,
     + 1.2088E+05, 1.2111E+05, 1.2134E+05, 1.2156E+05, 1.2179E+05,
     +  1.2202E+05, 1.2225E+05, 1.2248E+05, 1.2271E+05, 1.2294E+05,
     + 1.2317E+05, 1.2340E+05, 1.2363E+05, 1.2386E+05, 1.2409E+05,
     +  1.2432E+05, 1.2455E+05, 1.2478E+05, 1.2502E+05, 1.2525E+05,
     + 1.2548E+05, 1.2571E+05, 1.2595E+05, 1.2618E+05, 1.2642E+05,
     +  1.2665E+05, 1.2689E+05, 1.2712E+05, 1.2736E+05, 1.2759E+05,
     + 1.2783E+05, 1.2807E+05, 1.2830E+05, 1.2854E+05, 1.2878E+05,
     +  1.2902E+05, 1.2925E+05, 1.2949E+05, 1.2973E+05, 1.2997E+05,
     + 1.3021E+05, 1.3045E+05, 1.3069E+05, 1.3093E+05, 1.3117E+05,
     +  1.3141E+05, 1.3166E+05, 1.3190E+05, 1.3214E+05, 1.3238E+05,
     + 1.3262E+05, 1.3287E+05, 1.3311E+05, 1.3336E+05, 1.3360E+05,
     +  1.3384E+05, 1.3409E+05, 1.3433E+05, 1.3458E+05, 1.3483E+05,
     + 1.3507E+05, 1.3532E+05, 1.3557E+05, 1.3581E+05, 1.3606E+05,
     +  1.3631E+05, 1.3656E+05, 1.3681E+05, 1.3705E+05, 1.3730E+05,
     + 1.3755E+05, 1.3780E+05, 1.3805E+05, 1.3830E+05, 1.3855E+05,
     +  1.3881E+05, 1.3906E+05, 1.3931E+05, 1.3956E+05, 1.3981E+05,
     + 1.4007E+05, 1.4032E+05, 1.4057E+05, 1.4083E+05, 1.4108E+05,
     +  1.4134E+05, 1.4159E+05, 1.4185E+05, 1.4210E+05, 1.4236E+05,
     + 1.4262E+05, 1.4287E+05, 1.4313E+05, 1.4339E+05, 1.4364E+05,
     +  1.4390E+05, 1.4416E+05, 1.4442E+05, 1.4468E+05, 1.4494E+05,
     + 1.4520E+05, 1.4546E+05, 1.4572E+05, 1.4598E+05, 1.4624E+05,
     +  1.4650E+05, 1.4677E+05, 1.4703E+05, 1.4729E+05, 1.4755E+05,
     + 1.4782E+05, 1.4808E+05, 1.4834E+05, 1.4861E+05, 1.4887E+05,
     +  1.4914E+05, 1.4940E+05, 1.4967E+05, 1.4994E+05, 1.5020E+05,
     + 1.5047E+05, 1.5074E+05, 1.5100E+05, 1.5127E+05, 1.5154E+05,
     +  1.5181E+05, 1.5208E+05, 1.5235E+05, 1.5262E+05, 1.5289E+05,
     + 1.5316E+05, 1.5343E+05, 1.5370E+05, 1.5397E+05, 1.5424E+05,
     +  1.5451E+05, 1.5479E+05, 1.5506E+05, 1.5533E+05, 1.5561E+05,
     + 1.5588E+05, 1.5615E+05, 1.5643E+05, 1.5670E+05, 1.5698E+05,
     +  1.5725E+05, 1.5753E+05, 1.5781E+05, 1.5808E+05, 1.5836E+05,
     + 1.5864E+05, 1.5892E+05, 1.5919E+05, 1.5947E+05, 1.5975E+05,
     +  1.6003E+05, 1.6031E+05, 1.6059E+05, 1.6087E+05, 1.6115E+05,
     + 1.6143E+05, 1.6171E+05, 1.6200E+05, 1.6228E+05, 1.6256E+05,
     +  1.6284E+05, 1.6313E+05, 1.6341E+05, 1.6370E+05, 1.6398E+05,
     + 1.6426E+05, 1.6455E+05, 1.6484E+05, 1.6512E+05, 1.6541E+05,
     +  1.6569E+05, 1.6598E+05, 1.6627E+05, 1.6656E+05, 1.6684E+05,
     + 1.6713E+05, 1.6742E+05, 1.6771E+05, 1.6800E+05, 1.6829E+05,
     +  1.6858E+05, 1.6887E+05, 1.6916E+05, 1.6945E+05, 1.6975E+05,
     + 1.7004E+05, 1.7033E+05, 1.7062E+05, 1.7092E+05, 1.7121E+05,
     +  1.7150E+05, 1.7180E+05, 1.7209E+05, 1.7239E+05, 1.7268E+05,
     + 1.7298E+05, 1.7328E+05, 1.7357E+05, 1.7387E+05, 1.7417E+05,
     +  1.7446E+05, 1.7476E+05, 1.7506E+05, 1.7536E+05, 1.7566E+05,
     + 1.7596E+05, 1.7626E+05, 1.7656E+05, 1.7686E+05, 1.7716E+05,
     +  1.7746E+05, 1.7776E+05, 1.7807E+05, 1.7837E+05, 1.7867E+05,
     + 1.7898E+05, 1.7928E+05, 1.7958E+05, 1.7989E+05, 1.8019E+05,
     +  1.8050E+05, 1.8080E+05, 1.8111E+05, 1.8142E+05, 1.8172E+05,
     + 1.8203E+05, 1.8234E+05, 1.8265E+05, 1.8296E+05, 1.8326E+05,
     +  1.8357E+05, 1.8388E+05, 1.8419E+05, 1.8450E+05, 1.8481E+05,
     + 1.8513E+05, 1.8544E+05, 1.8575E+05, 1.8606E+05, 1.8637E+05,
     +  1.8669E+05, 1.8700E+05, 1.8731E+05, 1.8763E+05, 1.8794E+05,
     + 1.8826E+05, 1.8857E+05, 1.8889E+05, 1.8920E+05, 1.8952E+05,
     +  1.8984E+05, 1.9016E+05, 1.9047E+05, 1.9079E+05, 1.9111E+05,
     + 1.9143E+05, 1.9175E+05, 1.9207E+05, 1.9239E+05, 1.9271E+05,
     +  1.9303E+05, 1.9335E+05, 1.9367E+05, 1.9399E+05, 1.9432E+05,
     + 1.9464E+05, 1.9496E+05, 1.9529E+05, 1.9561E+05, 1.9593E+05,
     +  1.9626E+05, 1.9658E+05, 1.9691E+05, 1.9724E+05, 1.9756E+05,
     + 1.9789E+05, 1.9822E+05, 1.9854E+05, 1.9887E+05, 1.9920E+05,
     +  1.9953E+05, 1.9986E+05, 2.0019E+05, 2.0052E+05, 2.0085E+05,
     + 2.0118E+05, 2.0151E+05, 2.0184E+05, 2.0217E+05, 2.0251E+05,
     +  2.0284E+05, 2.0317E+05, 2.0351E+05, 2.0384E+05, 2.0417E+05,
     + 2.0451E+05, 2.0484E+05, 2.0518E+05, 2.0552E+05, 2.0585E+05,
     +  2.0619E+05, 2.0653E+05, 2.0686E+05, 2.0720E+05, 2.0754E+05,
     + 2.0788E+05, 2.0822E+05, 2.0856E+05, 2.0890E+05, 2.0924E+05,
     +  2.0958E+05, 2.0992E+05, 2.1026E+05, 2.1060E+05, 2.1095E+05,
     + 2.1129E+05, 2.1163E+05, 2.1198E+05, 2.1232E+05, 2.1267E+05,
     +  2.1301E+05, 2.1336E+05, 2.1370E+05, 2.1405E+05, 2.1439E+05,
     + 2.1474E+05, 2.1509E+05, 2.1544E+05, 2.1579E+05, 2.1613E+05,
     +  2.1648E+05, 2.1683E+05, 2.1718E+05, 2.1753E+05, 2.1788E+05,
     + 2.1823E+05, 2.1859E+05, 2.1894E+05, 2.1929E+05, 2.1964E+05,
     +  2.2000E+05, 2.2035E+05, 2.2070E+05, 2.2106E+05, 2.2141E+05,
     + 2.2177E+05, 2.2212E+05, 2.2248E+05, 2.2284E+05, 2.2319E+05,
     +  2.2355E+05, 2.2391E+05, 2.2427E+05, 2.2463E+05, 2.2499E+05,
     + 2.2535E+05, 2.2571E+05, 2.2607E+05, 2.2643E+05, 2.2679E+05,
     +  2.2715E+05, 2.2751E+05, 2.2787E+05, 2.2824E+05, 2.2860E+05,
     + 2.2896E+05, 2.2933E+05, 2.2969E+05, 2.3006E+05, 2.3042E+05,
     +  2.3079E+05, 2.3115E+05, 2.3152E+05, 2.3189E+05, 2.3226E+05,
     + 2.3262E+05, 2.3299E+05, 2.3336E+05, 2.3373E+05, 2.3410E+05,
     +  2.3447E+05, 2.3484E+05, 2.3521E+05, 2.3558E+05, 2.3596E+05,
     + 2.3633E+05, 2.3670E+05, 2.3707E+05, 2.3745E+05, 2.3782E+05,
     +  2.3820E+05, 2.3857E+05, 2.3895E+05, 2.3932E+05, 2.3970E+05,
     + 2.4007E+05, 2.4045E+05, 2.4083E+05, 2.4121E+05, 2.4158E+05,
     +  2.4196E+05, 2.4234E+05, 2.4272E+05, 2.4310E+05, 2.4348E+05,
     + 2.4386E+05, 2.4424E+05, 2.4463E+05, 2.4501E+05, 2.4539E+05,
     +  2.4577E+05, 2.4616E+05, 2.4654E+05, 2.4693E+05, 2.4731E+05,
     + 2.4770E+05, 2.4808E+05, 2.4847E+05, 2.4885E+05, 2.4924E+05,
     +  2.4963E+05, 2.5002E+05, 2.5040E+05, 2.5079E+05, 2.5118E+05,
     + 2.5157E+05, 2.5196E+05, 2.5235E+05, 2.5274E+05, 2.5314E+05,
     +  2.5353E+05, 2.5392E+05, 2.5431E+05, 2.5471E+05, 2.5510E+05,
     + 2.5549E+05, 2.5589E+05, 2.5628E+05, 2.5668E+05, 2.5707E+05,
     +  2.5747E+05, 2.5787E+05, 2.5826E+05, 2.5866E+05, 2.5906E+05,
     + 2.5946E+05, 2.5986E+05, 2.6026E+05, 2.6066E+05, 2.6106E+05,
     +  2.6146E+05, 2.6186E+05, 2.6226E+05, 2.6266E+05, 2.6306E+05,
     + 2.6347E+05, 2.6387E+05, 2.6428E+05, 2.6468E+05, 2.6508E+05,
     +  2.6549E+05, 2.6590E+05, 2.6630E+05, 2.6671E+05, 2.6711E+05,
     + 2.6752E+05, 2.6793E+05, 2.6834E+05, 2.6875E+05, 2.6916E+05,
     +  2.6957E+05, 2.6998E+05, 2.7039E+05, 2.7080E+05, 2.7121E+05,
     + 2.7162E+05, 2.7203E+05, 2.7245E+05, 2.7286E+05, 2.7327E+05,
     +  2.7369E+05, 2.7410E+05, 2.7452E+05, 2.7493E+05, 2.7535E+05,
     + 2.7577E+05, 2.7618E+05, 2.7660E+05, 2.7702E+05, 2.7744E+05,
     +  2.7786E+05, 2.7828E+05, 2.7870E+05, 2.7912E+05, 2.7954E+05,
     + 2.7996E+05, 2.8038E+05, 2.8080E+05, 2.8122E+05, 2.8165E+05,
     +  2.8207E+05, 2.8249E+05, 2.8292E+05, 2.8334E+05, 2.8377E+05,
     + 2.8419E+05, 2.8462E+05, 2.8505E+05, 2.8547E+05, 2.8590E+05,
     +  2.8633E+05, 2.8676E+05, 2.8719E+05, 2.8762E+05, 2.8805E+05,
     + 2.8848E+05, 2.8891E+05, 2.8934E+05, 2.8977E+05, 2.9020E+05,
     +  2.9064E+05, 2.9107E+05, 2.9150E+05, 2.9194E+05, 2.9237E+05,
     + 2.9281E+05, 2.9324E+05, 2.9368E+05, 2.9412E+05, 2.9455E+05,
     +  2.9499E+05, 2.9543E+05, 2.9587E+05, 2.9630E+05, 2.9674E+05,
     + 2.9718E+05, 2.9762E+05, 2.9806E+05, 2.9851E+05, 2.9895E+05,
     +  2.9939E+05, 2.9983E+05, 3.0028E+05, 3.0072E+05, 3.0116E+05,
     + 3.0161E+05, 3.0205E+05, 3.0250E+05, 3.0294E+05, 3.0339E+05,
     +  3.0384E+05, 3.0428E+05, 3.0473E+05, 3.0518E+05, 3.0563E+05,
     + 3.0608E+05, 3.0653E+05, 3.0698E+05, 3.0743E+05, 3.0788E+05,
     +  3.0833E+05, 3.0878E+05, 3.0924E+05, 3.0969E+05, 3.1014E+05,
     + 3.1060E+05, 3.1105E+05, 3.1151E+05, 3.1196E+05, 3.1242E+05,
     +  3.1287E+05, 3.1333E+05, 3.1379E+05, 3.1425E+05, 3.1470E+05,
     + 3.1516E+05, 3.1562E+05, 3.1608E+05, 3.1654E+05, 3.1700E+05,
     +  3.1746E+05, 3.1793E+05, 3.1839E+05, 3.1885E+05, 3.1931E+05,
     + 3.1978E+05, 3.2024E+05, 3.2071E+05, 3.2117E+05, 3.2164E+05,
     +  3.2210E+05, 3.2257E+05, 3.2304E+05, 3.2350E+05, 3.2397E+05,
     + 3.2444E+05, 3.2491E+05, 3.2538E+05, 3.2585E+05, 3.2632E+05,
     +  3.2679E+05, 3.2726E+05, 3.2773E+05, 3.2821E+05, 3.2868E+05,
     + 3.2915E+05, 3.2963E+05, 3.3010E+05, 3.3058E+05, 3.3105E+05,
     +  3.3153E+05, 3.3200E+05, 3.3248E+05, 3.3296E+05, 3.3343E+05,
     + 3.3391E+05, 3.3439E+05, 3.3487E+05, 3.3535E+05, 3.3583E+05,
     +  3.3631E+05, 3.3679E+05, 3.3727E+05, 3.3776E+05, 3.3824E+05,
     + 3.3872E+05, 3.3921E+05, 3.3969E+05, 3.4018E+05, 3.4066E+05,
     +  3.4115E+05, 3.4163E+05, 3.4212E+05, 3.4261E+05, 3.4309E+05,
     + 3.4358E+05, 3.4407E+05, 3.4456E+05, 3.4505E+05, 3.4554E+05,
     +  3.4603E+05, 3.4652E+05, 3.4701E+05, 3.4751E+05, 3.4800E+05,
     + 3.4849E+05, 3.4899E+05, 3.4948E+05, 3.4997E+05, 3.5047E+05,
     +  3.5096E+05, 3.5146E+05, 3.5196E+05, 3.5245E+05, 3.5295E+05,
     + 3.5345E+05, 3.5395E+05, 3.5445E+05, 3.5495E+05, 3.5545E+05,
     +  3.5595E+05, 3.5645E+05, 3.5695E+05, 3.5745E+05, 3.5796E+05,
     + 3.5846E+05, 3.5896E+05, 3.5947E+05, 3.5997E+05, 3.6048E+05,
     +  3.6098E+05, 3.6149E+05, 3.6200E+05, 3.6250E+05, 3.6301E+05,
     + 3.6352E+05, 3.6403E+05, 3.6454E+05, 3.6505E+05, 3.6556E+05,
     +  3.6607E+05, 3.6658E+05, 3.6709E+05, 3.6760E+05, 3.6812E+05,
     + 3.6863E+05, 3.6914E+05, 3.6966E+05, 3.7017E+05, 3.7069E+05,
     +  3.7121E+05, 3.7172E+05, 3.7224E+05, 3.7276E+05, 3.7327E+05,
     + 3.7379E+05, 3.7431E+05, 3.7483E+05, 3.7535E+05, 3.7587E+05,
     +  3.7639E+05, 3.7691E+05, 3.7744E+05, 3.7796E+05, 3.7848E+05,
     + 3.7901E+05, 3.7953E+05, 3.8005E+05, 3.8058E+05, 3.8111E+05,
     +  3.8163E+05, 3.8216E+05, 3.8269E+05, 3.8321E+05, 3.8374E+05,
     + 3.8427E+05, 3.8480E+05, 3.8533E+05, 3.8586E+05, 3.8639E+05,
     +  3.8692E+05, 3.8745E+05, 3.8799E+05, 3.8852E+05, 3.8905E+05,
     + 3.8959E+05, 3.9012E+05, 3.9066E+05, 3.9119E+05, 3.9173E+05,
     +  3.9226E+05, 3.9280E+05, 3.9334E+05, 3.9388E+05, 3.9442E+05,
     + 3.9496E+05, 3.9550E+05, 3.9604E+05, 3.9658E+05, 3.9712E+05,
     +  3.9766E+05, 3.9820E+05, 3.9874E+05, 3.9929E+05, 3.9983E+05,
     + 4.0038E+05, 4.0092E+05, 4.0147E+05, 4.0201E+05, 4.0256E+05,
     +  4.0311E+05, 4.0365E+05, 4.0420E+05, 4.0475E+05, 4.0530E+05,
     + 4.0585E+05, 4.0640E+05, 4.0695E+05, 4.0750E+05, 4.0805E+05,
     +  4.0861E+05, 4.0916E+05, 4.0971E+05, 4.1027E+05, 4.1082E+05,
     + 4.1138E+05, 4.1193E+05, 4.1249E+05, 4.1304E+05, 4.1360E+05,
     +  4.1416E+05, 4.1472E+05, 4.1528E+05, 4.1583E+05, 4.1639E+05/
      data (QofT( 5,J),J=2409,4510)/ 4.1695E+05, 4.1752E+05,
     +  4.1808E+05, 4.1864E+05, 4.1920E+05, 4.1976E+05, 4.2033E+05,
     +  4.2089E+05,
     + 4.2146E+05, 4.2202E+05, 4.2259E+05, 4.2315E+05, 4.2372E+05,
     +  4.2429E+05, 4.2486E+05, 4.2542E+05, 4.2599E+05, 4.2656E+05,
     + 4.2713E+05, 4.2770E+05, 4.2827E+05, 4.2884E+05, 4.2942E+05,
     +  4.2999E+05, 4.3056E+05, 4.3114E+05, 4.3171E+05, 4.3229E+05,
     + 4.3286E+05, 4.3344E+05, 4.3401E+05, 4.3459E+05, 4.3517E+05,
     +  4.3575E+05, 4.3632E+05, 4.3690E+05, 4.3748E+05, 4.3806E+05,
     + 4.3864E+05, 4.3922E+05, 4.3981E+05, 4.4039E+05, 4.4097E+05,
     +  4.4156E+05, 4.4214E+05, 4.4272E+05, 4.4331E+05, 4.4389E+05,
     + 4.4448E+05, 4.4507E+05, 4.4565E+05, 4.4624E+05, 4.4683E+05,
     +  4.4742E+05, 4.4801E+05, 4.4860E+05, 4.4919E+05, 4.4978E+05,
     + 4.5037E+05, 4.5096E+05, 4.5156E+05, 4.5215E+05, 4.5274E+05,
     +  4.5334E+05, 4.5393E+05, 4.5453E+05, 4.5512E+05, 4.5572E+05,
     + 4.5632E+05, 4.5692E+05, 4.5751E+05, 4.5811E+05, 4.5871E+05,
     +  4.5931E+05, 4.5991E+05, 4.6051E+05, 4.6111E+05, 4.6172E+05,
     + 4.6232E+05, 4.6292E+05, 4.6353E+05, 4.6413E+05, 4.6474E+05,
     +  4.6534E+05, 4.6595E+05, 4.6655E+05, 4.6716E+05, 4.6777E+05,
     + 4.6838E+05, 4.6898E+05, 4.6959E+05, 4.7020E+05, 4.7081E+05,
     +  4.7142E+05, 4.7204E+05, 4.7265E+05, 4.7326E+05, 4.7387E+05,
     + 4.7449E+05, 4.7510E+05, 4.7572E+05, 4.7633E+05, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       172
      data (QofT( 6,J),J=1,2408)/ 3.6000E+01, 3.6002E+01, 3.6065E+01,
     +  3.6419E+01, 3.7294E+01, 3.8789E+01, 4.0902E+01, 4.3596E+01,
     + 4.6822E+01, 5.0531E+01, 5.4680E+01, 5.9229E+01, 6.4145E+01,
     +  6.9396E+01, 7.4956E+01, 8.0798E+01, 8.6904E+01, 9.3252E+01,
     + 9.9829E+01, 1.0662E+02, 1.2079E+02, 1.3568E+02, 1.5124E+02,
     +  1.6742E+02, 1.8419E+02, 2.0151E+02, 2.1936E+02, 2.3773E+02,
     + 2.5660E+02, 2.7595E+02, 2.9577E+02, 3.1606E+02, 3.3679E+02,
     +  3.5796E+02, 3.7957E+02, 4.0160E+02, 4.2404E+02, 4.4689E+02,
     + 4.7014E+02, 4.9378E+02, 5.1781E+02, 5.4222E+02, 5.6701E+02,
     +  5.9216E+02, 6.1768E+02, 6.4356E+02, 6.6980E+02, 6.9638E+02,
     + 7.2331E+02, 7.5058E+02, 7.7819E+02, 8.0613E+02, 8.3440E+02,
     +  8.6300E+02, 8.9192E+02, 9.2116E+02, 9.5071E+02, 9.8057E+02,
     + 1.0107E+03, 1.0412E+03, 1.0720E+03, 1.1031E+03, 1.1345E+03,
     +  1.1661E+03, 1.1981E+03, 1.2304E+03, 1.2629E+03, 1.2957E+03,
     + 1.3288E+03, 1.3622E+03, 1.3959E+03, 1.4298E+03, 1.4640E+03,
     +  1.4985E+03, 1.5333E+03, 1.5683E+03, 1.6036E+03, 1.6391E+03,
     + 1.6749E+03, 1.7110E+03, 1.7474E+03, 1.7839E+03, 1.8208E+03,
     +  1.8579E+03, 1.8952E+03, 1.9328E+03, 1.9707E+03, 2.0088E+03,
     + 2.0471E+03, 2.0857E+03, 2.1245E+03, 2.1636E+03, 2.2029E+03,
     +  2.2424E+03, 2.2822E+03, 2.3222E+03, 2.3625E+03, 2.4030E+03,
     + 2.4437E+03, 2.4846E+03, 2.5258E+03, 2.5672E+03, 2.6089E+03,
     +  2.6507E+03, 2.6928E+03, 2.7351E+03, 2.7777E+03, 2.8204E+03,
     + 2.8634E+03, 2.9066E+03, 2.9500E+03, 2.9937E+03, 3.0376E+03,
     +  3.0816E+03, 3.1259E+03, 3.1704E+03, 3.2152E+03, 3.2601E+03,
     + 3.3053E+03, 3.3506E+03, 3.3962E+03, 3.4420E+03, 3.4880E+03,
     +  3.5342E+03, 3.5807E+03, 3.6273E+03, 3.6741E+03, 3.7212E+03,
     + 3.7684E+03, 3.8159E+03, 3.8636E+03, 3.9114E+03, 3.9595E+03,
     +  4.0078E+03, 4.0563E+03, 4.1050E+03, 4.1539E+03, 4.2029E+03,
     + 4.2522E+03, 4.3017E+03, 4.3514E+03, 4.4013E+03, 4.4514E+03,
     +  4.5017E+03, 4.5522E+03, 4.6029E+03, 4.6538E+03, 4.7049E+03,
     + 4.7562E+03, 4.8077E+03, 4.8594E+03, 4.9113E+03, 4.9634E+03,
     +  5.0157E+03, 5.0682E+03, 5.1208E+03, 5.1737E+03, 5.2268E+03,
     + 5.2801E+03, 5.3335E+03, 5.3872E+03, 5.4411E+03, 5.4951E+03,
     +  5.5494E+03, 5.6038E+03, 5.6585E+03, 5.7133E+03, 5.7683E+03,
     + 5.8236E+03, 5.8790E+03, 5.9346E+03, 5.9904E+03, 6.0464E+03,
     +  6.1026E+03, 6.1590E+03, 6.2156E+03, 6.2724E+03, 6.3294E+03,
     + 6.3866E+03, 6.4440E+03, 6.5016E+03, 6.5594E+03, 6.6173E+03,
     +  6.6755E+03, 6.7339E+03, 6.7924E+03, 6.8512E+03, 6.9101E+03,
     + 6.9693E+03, 7.0286E+03, 7.0882E+03, 7.1479E+03, 7.2079E+03,
     +  7.2680E+03, 7.3283E+03, 7.3889E+03, 7.4496E+03, 7.5105E+03,
     + 7.5717E+03, 7.6330E+03, 7.6945E+03, 7.7563E+03, 7.8182E+03,
     +  7.8803E+03, 7.9426E+03, 8.0052E+03, 8.0679E+03, 8.1308E+03,
     + 8.1940E+03, 8.2573E+03, 8.3208E+03, 8.3846E+03, 8.4485E+03,
     +  8.5126E+03, 8.5770E+03, 8.6415E+03, 8.7063E+03, 8.7712E+03,
     + 8.8364E+03, 8.9017E+03, 8.9673E+03, 9.0331E+03, 9.0991E+03,
     +  9.1652E+03, 9.2316E+03, 9.2982E+03, 9.3650E+03, 9.4320E+03,
     + 9.4993E+03, 9.5667E+03, 9.6343E+03, 9.7022E+03, 9.7702E+03,
     +  9.8385E+03, 9.9070E+03, 9.9756E+03, 1.0045E+04, 1.0114E+04,
     + 1.0183E+04, 1.0252E+04, 1.0322E+04, 1.0392E+04, 1.0462E+04,
     +  1.0533E+04, 1.0603E+04, 1.0674E+04, 1.0745E+04, 1.0816E+04,
     + 1.0888E+04, 1.0959E+04, 1.1031E+04, 1.1103E+04, 1.1176E+04,
     +  1.1248E+04, 1.1321E+04, 1.1394E+04, 1.1467E+04, 1.1541E+04,
     + 1.1614E+04, 1.1688E+04, 1.1762E+04, 1.1837E+04, 1.1911E+04,
     +  1.1986E+04, 1.2061E+04, 1.2136E+04, 1.2211E+04, 1.2287E+04,
     + 1.2363E+04, 1.2439E+04, 1.2515E+04, 1.2592E+04, 1.2669E+04,
     +  1.2746E+04, 1.2823E+04, 1.2900E+04, 1.2978E+04, 1.3056E+04,
     + 1.3134E+04, 1.3212E+04, 1.3291E+04, 1.3370E+04, 1.3449E+04,
     +  1.3528E+04, 1.3608E+04, 1.3688E+04, 1.3768E+04, 1.3848E+04,
     + 1.3928E+04, 1.4009E+04, 1.4090E+04, 1.4171E+04, 1.4253E+04,
     +  1.4334E+04, 1.4416E+04, 1.4498E+04, 1.4581E+04, 1.4663E+04,
     + 1.4746E+04, 1.4829E+04, 1.4913E+04, 1.4996E+04, 1.5080E+04,
     +  1.5164E+04, 1.5248E+04, 1.5333E+04, 1.5418E+04, 1.5503E+04,
     + 1.5588E+04, 1.5674E+04, 1.5759E+04, 1.5845E+04, 1.5932E+04,
     +  1.6018E+04, 1.6105E+04, 1.6192E+04, 1.6279E+04, 1.6367E+04,
     + 1.6455E+04, 1.6543E+04, 1.6631E+04, 1.6720E+04, 1.6809E+04,
     +  1.6898E+04, 1.6987E+04, 1.7077E+04, 1.7166E+04, 1.7256E+04,
     + 1.7347E+04, 1.7437E+04, 1.7528E+04, 1.7620E+04, 1.7711E+04,
     +  1.7803E+04, 1.7895E+04, 1.7987E+04, 1.8079E+04, 1.8172E+04,
     + 1.8265E+04, 1.8358E+04, 1.8452E+04, 1.8545E+04, 1.8639E+04,
     +  1.8734E+04, 1.8828E+04, 1.8923E+04, 1.9018E+04, 1.9114E+04,
     + 1.9209E+04, 1.9305E+04, 1.9402E+04, 1.9498E+04, 1.9595E+04,
     +  1.9692E+04, 1.9789E+04, 1.9887E+04, 1.9985E+04, 2.0083E+04,
     + 2.0181E+04, 2.0280E+04, 2.0379E+04, 2.0478E+04, 2.0578E+04,
     +  2.0678E+04, 2.0778E+04, 2.0878E+04, 2.0979E+04, 2.1080E+04,
     + 2.1181E+04, 2.1283E+04, 2.1384E+04, 2.1486E+04, 2.1589E+04,
     +  2.1692E+04, 2.1795E+04, 2.1898E+04, 2.2001E+04, 2.2105E+04,
     + 2.2209E+04, 2.2314E+04, 2.2418E+04, 2.2523E+04, 2.2629E+04,
     +  2.2734E+04, 2.2840E+04, 2.2946E+04, 2.3053E+04, 2.3160E+04,
     + 2.3267E+04, 2.3374E+04, 2.3482E+04, 2.3590E+04, 2.3698E+04,
     +  2.3807E+04, 2.3916E+04, 2.4025E+04, 2.4134E+04, 2.4244E+04,
     + 2.4354E+04, 2.4465E+04, 2.4576E+04, 2.4687E+04, 2.4798E+04,
     +  2.4910E+04, 2.5022E+04, 2.5134E+04, 2.5247E+04, 2.5359E+04,
     + 2.5473E+04, 2.5586E+04, 2.5700E+04, 2.5814E+04, 2.5929E+04,
     +  2.6044E+04, 2.6159E+04, 2.6274E+04, 2.6390E+04, 2.6506E+04,
     + 2.6622E+04, 2.6739E+04, 2.6856E+04, 2.6974E+04, 2.7091E+04,
     +  2.7209E+04, 2.7328E+04, 2.7446E+04, 2.7565E+04, 2.7685E+04,
     + 2.7804E+04, 2.7924E+04, 2.8045E+04, 2.8165E+04, 2.8286E+04,
     +  2.8408E+04, 2.8529E+04, 2.8651E+04, 2.8774E+04, 2.8896E+04,
     + 2.9019E+04, 2.9143E+04, 2.9266E+04, 2.9390E+04, 2.9515E+04,
     +  2.9639E+04, 2.9764E+04, 2.9890E+04, 3.0015E+04, 3.0141E+04,
     + 3.0268E+04, 3.0395E+04, 3.0522E+04, 3.0649E+04, 3.0777E+04,
     +  3.0905E+04, 3.1034E+04, 3.1162E+04, 3.1292E+04, 3.1421E+04,
     + 3.1551E+04, 3.1681E+04, 3.1812E+04, 3.1943E+04, 3.2074E+04,
     +  3.2206E+04, 3.2338E+04, 3.2470E+04, 3.2603E+04, 3.2736E+04,
     + 3.2870E+04, 3.3003E+04, 3.3138E+04, 3.3272E+04, 3.3407E+04,
     +  3.3542E+04, 3.3678E+04, 3.3814E+04, 3.3950E+04, 3.4087E+04,
     + 3.4224E+04, 3.4362E+04, 3.4500E+04, 3.4638E+04, 3.4776E+04,
     +  3.4915E+04, 3.5055E+04, 3.5194E+04, 3.5335E+04, 3.5475E+04,
     + 3.5616E+04, 3.5757E+04, 3.5899E+04, 3.6041E+04, 3.6183E+04,
     +  3.6326E+04, 3.6469E+04, 3.6612E+04, 3.6756E+04, 3.6901E+04,
     + 3.7045E+04, 3.7190E+04, 3.7336E+04, 3.7482E+04, 3.7628E+04,
     +  3.7775E+04, 3.7922E+04, 3.8069E+04, 3.8217E+04, 3.8365E+04,
     + 3.8513E+04, 3.8662E+04, 3.8812E+04, 3.8962E+04, 3.9112E+04,
     +  3.9262E+04, 3.9413E+04, 3.9565E+04, 3.9716E+04, 3.9869E+04,
     + 4.0021E+04, 4.0174E+04, 4.0328E+04, 4.0481E+04, 4.0635E+04,
     +  4.0790E+04, 4.0945E+04, 4.1101E+04, 4.1256E+04, 4.1413E+04,
     + 4.1569E+04, 4.1726E+04, 4.1884E+04, 4.2042E+04, 4.2200E+04,
     +  4.2359E+04, 4.2518E+04, 4.2677E+04, 4.2837E+04, 4.2998E+04,
     + 4.3159E+04, 4.3320E+04, 4.3482E+04, 4.3644E+04, 4.3806E+04,
     +  4.3969E+04, 4.4132E+04, 4.4296E+04, 4.4460E+04, 4.4625E+04,
     + 4.4790E+04, 4.4956E+04, 4.5122E+04, 4.5288E+04, 4.5455E+04,
     +  4.5622E+04, 4.5790E+04, 4.5958E+04, 4.6126E+04, 4.6295E+04,
     + 4.6465E+04, 4.6635E+04, 4.6805E+04, 4.6976E+04, 4.7147E+04,
     +  4.7319E+04, 4.7491E+04, 4.7663E+04, 4.7836E+04, 4.8010E+04,
     + 4.8183E+04, 4.8358E+04, 4.8532E+04, 4.8708E+04, 4.8883E+04,
     +  4.9059E+04, 4.9236E+04, 4.9413E+04, 4.9591E+04, 4.9768E+04,
     + 4.9947E+04, 5.0126E+04, 5.0305E+04, 5.0485E+04, 5.0665E+04,
     +  5.0846E+04, 5.1027E+04, 5.1209E+04, 5.1391E+04, 5.1573E+04,
     + 5.1756E+04, 5.1940E+04, 5.2124E+04, 5.2308E+04, 5.2493E+04,
     +  5.2678E+04, 5.2864E+04, 5.3051E+04, 5.3237E+04, 5.3425E+04,
     + 5.3612E+04, 5.3801E+04, 5.3989E+04, 5.4178E+04, 5.4368E+04,
     +  5.4558E+04, 5.4749E+04, 5.4940E+04, 5.5132E+04, 5.5324E+04,
     + 5.5516E+04, 5.5709E+04, 5.5903E+04, 5.6097E+04, 5.6291E+04,
     +  5.6486E+04, 5.6682E+04, 5.6878E+04, 5.7074E+04, 5.7271E+04,
     + 5.7469E+04, 5.7667E+04, 5.7865E+04, 5.8064E+04, 5.8264E+04,
     +  5.8464E+04, 5.8664E+04, 5.8865E+04, 5.9066E+04, 5.9268E+04,
     + 5.9471E+04, 5.9674E+04, 5.9877E+04, 6.0081E+04, 6.0286E+04,
     +  6.0491E+04, 6.0696E+04, 6.0903E+04, 6.1109E+04, 6.1316E+04,
     + 6.1524E+04, 6.1732E+04, 6.1940E+04, 6.2150E+04, 6.2359E+04,
     +  6.2569E+04, 6.2780E+04, 6.2991E+04, 6.3203E+04, 6.3415E+04,
     + 6.3628E+04, 6.3841E+04, 6.4055E+04, 6.4270E+04, 6.4485E+04,
     +  6.4700E+04, 6.4916E+04, 6.5132E+04, 6.5350E+04, 6.5567E+04,
     + 6.5785E+04, 6.6004E+04, 6.6223E+04, 6.6443E+04, 6.6663E+04,
     +  6.6884E+04, 6.7105E+04, 6.7327E+04, 6.7549E+04, 6.7772E+04,
     + 6.7996E+04, 6.8220E+04, 6.8444E+04, 6.8670E+04, 6.8895E+04,
     +  6.9122E+04, 6.9348E+04, 6.9576E+04, 6.9804E+04, 7.0032E+04,
     + 7.0261E+04, 7.0491E+04, 7.0721E+04, 7.0951E+04, 7.1183E+04,
     +  7.1414E+04, 7.1647E+04, 7.1880E+04, 7.2113E+04, 7.2347E+04,
     + 7.2582E+04, 7.2817E+04, 7.3053E+04, 7.3289E+04, 7.3526E+04,
     +  7.3763E+04, 7.4001E+04, 7.4240E+04, 7.4479E+04, 7.4719E+04,
     + 7.4959E+04, 7.5200E+04, 7.5442E+04, 7.5684E+04, 7.5926E+04,
     +  7.6170E+04, 7.6413E+04, 7.6658E+04, 7.6903E+04, 7.7148E+04,
     + 7.7394E+04, 7.7641E+04, 7.7889E+04, 7.8136E+04, 7.8385E+04,
     +  7.8634E+04, 7.8884E+04, 7.9134E+04, 7.9385E+04, 7.9636E+04,
     + 7.9888E+04, 8.0141E+04, 8.0394E+04, 8.0648E+04, 8.0903E+04,
     +  8.1158E+04, 8.1413E+04, 8.1670E+04, 8.1926E+04, 8.2184E+04,
     + 8.2442E+04, 8.2701E+04, 8.2960E+04, 8.3220E+04, 8.3480E+04,
     +  8.3742E+04, 8.4003E+04, 8.4266E+04, 8.4529E+04, 8.4792E+04,
     + 8.5057E+04, 8.5321E+04, 8.5587E+04, 8.5853E+04, 8.6120E+04,
     +  8.6387E+04, 8.6655E+04, 8.6923E+04, 8.7193E+04, 8.7462E+04,
     + 8.7733E+04, 8.8004E+04, 8.8276E+04, 8.8548E+04, 8.8821E+04,
     +  8.9095E+04, 8.9369E+04, 8.9644E+04, 8.9919E+04, 9.0195E+04,
     + 9.0472E+04, 9.0750E+04, 9.1028E+04, 9.1306E+04, 9.1586E+04,
     +  9.1866E+04, 9.2146E+04, 9.2428E+04, 9.2710E+04, 9.2992E+04,
     + 9.3276E+04, 9.3559E+04, 9.3844E+04, 9.4129E+04, 9.4415E+04,
     +  9.4702E+04, 9.4989E+04, 9.5277E+04, 9.5565E+04, 9.5854E+04,
     + 9.6144E+04, 9.6435E+04, 9.6726E+04, 9.7017E+04, 9.7310E+04,
     +  9.7603E+04, 9.7897E+04, 9.8191E+04, 9.8486E+04, 9.8782E+04,
     + 9.9079E+04, 9.9376E+04, 9.9674E+04, 9.9972E+04, 1.0027E+05,
     +  1.0057E+05, 1.0087E+05, 1.0117E+05, 1.0147E+05, 1.0178E+05,
     + 1.0208E+05, 1.0238E+05, 1.0269E+05, 1.0299E+05, 1.0330E+05,
     +  1.0361E+05, 1.0391E+05, 1.0422E+05, 1.0453E+05, 1.0484E+05,
     + 1.0515E+05, 1.0546E+05, 1.0577E+05, 1.0609E+05, 1.0640E+05,
     +  1.0671E+05, 1.0703E+05, 1.0734E+05, 1.0766E+05, 1.0797E+05,
     + 1.0829E+05, 1.0861E+05, 1.0893E+05, 1.0925E+05, 1.0957E+05,
     +  1.0989E+05, 1.1021E+05, 1.1053E+05, 1.1085E+05, 1.1118E+05,
     + 1.1150E+05, 1.1183E+05, 1.1215E+05, 1.1248E+05, 1.1281E+05,
     +  1.1313E+05, 1.1346E+05, 1.1379E+05, 1.1412E+05, 1.1445E+05,
     + 1.1478E+05, 1.1512E+05, 1.1545E+05, 1.1578E+05, 1.1612E+05,
     +  1.1645E+05, 1.1679E+05, 1.1713E+05, 1.1746E+05, 1.1780E+05,
     + 1.1814E+05, 1.1848E+05, 1.1882E+05, 1.1916E+05, 1.1950E+05,
     +  1.1985E+05, 1.2019E+05, 1.2053E+05, 1.2088E+05, 1.2122E+05,
     + 1.2157E+05, 1.2192E+05, 1.2226E+05, 1.2261E+05, 1.2296E+05,
     +  1.2331E+05, 1.2366E+05, 1.2401E+05, 1.2437E+05, 1.2472E+05,
     + 1.2507E+05, 1.2543E+05, 1.2578E+05, 1.2614E+05, 1.2650E+05,
     +  1.2685E+05, 1.2721E+05, 1.2757E+05, 1.2793E+05, 1.2829E+05,
     + 1.2865E+05, 1.2902E+05, 1.2938E+05, 1.2974E+05, 1.3011E+05,
     +  1.3047E+05, 1.3084E+05, 1.3121E+05, 1.3158E+05, 1.3194E+05,
     + 1.3231E+05, 1.3268E+05, 1.3305E+05, 1.3343E+05, 1.3380E+05,
     +  1.3417E+05, 1.3455E+05, 1.3492E+05, 1.3530E+05, 1.3567E+05,
     + 1.3605E+05, 1.3643E+05, 1.3681E+05, 1.3719E+05, 1.3757E+05,
     +  1.3795E+05, 1.3833E+05, 1.3871E+05, 1.3910E+05, 1.3948E+05,
     + 1.3987E+05, 1.4025E+05, 1.4064E+05, 1.4103E+05, 1.4142E+05,
     +  1.4181E+05, 1.4220E+05, 1.4259E+05, 1.4298E+05, 1.4337E+05,
     + 1.4377E+05, 1.4416E+05, 1.4456E+05, 1.4495E+05, 1.4535E+05,
     +  1.4575E+05, 1.4614E+05, 1.4654E+05, 1.4694E+05, 1.4735E+05,
     + 1.4775E+05, 1.4815E+05, 1.4855E+05, 1.4896E+05, 1.4936E+05,
     +  1.4977E+05, 1.5018E+05, 1.5058E+05, 1.5099E+05, 1.5140E+05,
     + 1.5181E+05, 1.5222E+05, 1.5263E+05, 1.5305E+05, 1.5346E+05,
     +  1.5387E+05, 1.5429E+05, 1.5471E+05, 1.5512E+05, 1.5554E+05,
     + 1.5596E+05, 1.5638E+05, 1.5680E+05, 1.5722E+05, 1.5764E+05,
     +  1.5807E+05, 1.5849E+05, 1.5891E+05, 1.5934E+05, 1.5977E+05,
     + 1.6019E+05, 1.6062E+05, 1.6105E+05, 1.6148E+05, 1.6191E+05,
     +  1.6234E+05, 1.6278E+05, 1.6321E+05, 1.6364E+05, 1.6408E+05,
     + 1.6452E+05, 1.6495E+05, 1.6539E+05, 1.6583E+05, 1.6627E+05,
     +  1.6671E+05, 1.6715E+05, 1.6759E+05, 1.6804E+05, 1.6848E+05,
     + 1.6893E+05, 1.6937E+05, 1.6982E+05, 1.7027E+05, 1.7072E+05,
     +  1.7116E+05, 1.7162E+05, 1.7207E+05, 1.7252E+05, 1.7297E+05,
     + 1.7343E+05, 1.7388E+05, 1.7434E+05, 1.7479E+05, 1.7525E+05,
     +  1.7571E+05, 1.7617E+05, 1.7663E+05, 1.7709E+05, 1.7755E+05,
     + 1.7802E+05, 1.7848E+05, 1.7895E+05, 1.7941E+05, 1.7988E+05,
     +  1.8035E+05, 1.8081E+05, 1.8128E+05, 1.8175E+05, 1.8223E+05,
     + 1.8270E+05, 1.8317E+05, 1.8365E+05, 1.8412E+05, 1.8460E+05,
     +  1.8508E+05, 1.8555E+05, 1.8603E+05, 1.8651E+05, 1.8699E+05,
     + 1.8747E+05, 1.8796E+05, 1.8844E+05, 1.8893E+05, 1.8941E+05,
     +  1.8990E+05, 1.9039E+05, 1.9087E+05, 1.9136E+05, 1.9185E+05,
     + 1.9235E+05, 1.9284E+05, 1.9333E+05, 1.9382E+05, 1.9432E+05,
     +  1.9482E+05, 1.9531E+05, 1.9581E+05, 1.9631E+05, 1.9681E+05,
     + 1.9731E+05, 1.9781E+05, 1.9832E+05, 1.9882E+05, 1.9932E+05,
     +  1.9983E+05, 2.0034E+05, 2.0085E+05, 2.0135E+05, 2.0186E+05,
     + 2.0237E+05, 2.0289E+05, 2.0340E+05, 2.0391E+05, 2.0443E+05,
     +  2.0494E+05, 2.0546E+05, 2.0598E+05, 2.0650E+05, 2.0702E+05,
     + 2.0754E+05, 2.0806E+05, 2.0858E+05, 2.0910E+05, 2.0963E+05,
     +  2.1015E+05, 2.1068E+05, 2.1121E+05, 2.1174E+05, 2.1227E+05,
     + 2.1280E+05, 2.1333E+05, 2.1386E+05, 2.1440E+05, 2.1493E+05,
     +  2.1547E+05, 2.1600E+05, 2.1654E+05, 2.1708E+05, 2.1762E+05,
     + 2.1816E+05, 2.1870E+05, 2.1924E+05, 2.1979E+05, 2.2033E+05,
     +  2.2088E+05, 2.2143E+05, 2.2197E+05, 2.2252E+05, 2.2307E+05,
     + 2.2362E+05, 2.2418E+05, 2.2473E+05, 2.2528E+05, 2.2584E+05,
     +  2.2640E+05, 2.2695E+05, 2.2751E+05, 2.2807E+05, 2.2863E+05,
     + 2.2919E+05, 2.2976E+05, 2.3032E+05, 2.3088E+05, 2.3145E+05,
     +  2.3202E+05, 2.3258E+05, 2.3315E+05, 2.3372E+05, 2.3429E+05,
     + 2.3487E+05, 2.3544E+05, 2.3601E+05, 2.3659E+05, 2.3716E+05,
     +  2.3774E+05, 2.3832E+05, 2.3890E+05, 2.3948E+05, 2.4006E+05,
     + 2.4064E+05, 2.4123E+05, 2.4181E+05, 2.4240E+05, 2.4299E+05,
     +  2.4357E+05, 2.4416E+05, 2.4475E+05, 2.4535E+05, 2.4594E+05,
     + 2.4653E+05, 2.4713E+05, 2.4772E+05, 2.4832E+05, 2.4892E+05,
     +  2.4952E+05, 2.5012E+05, 2.5072E+05, 2.5132E+05, 2.5192E+05,
     + 2.5253E+05, 2.5313E+05, 2.5374E+05, 2.5435E+05, 2.5496E+05,
     +  2.5557E+05, 2.5618E+05, 2.5679E+05, 2.5740E+05, 2.5802E+05,
     + 2.5863E+05, 2.5925E+05, 2.5987E+05, 2.6049E+05, 2.6111E+05,
     +  2.6173E+05, 2.6235E+05, 2.6297E+05, 2.6360E+05, 2.6422E+05,
     + 2.6485E+05, 2.6548E+05, 2.6611E+05, 2.6674E+05, 2.6737E+05,
     +  2.6800E+05, 2.6863E+05, 2.6927E+05, 2.6990E+05, 2.7054E+05,
     + 2.7118E+05, 2.7182E+05, 2.7246E+05, 2.7310E+05, 2.7374E+05,
     +  2.7439E+05, 2.7503E+05, 2.7568E+05, 2.7632E+05, 2.7697E+05,
     + 2.7762E+05, 2.7827E+05, 2.7893E+05, 2.7958E+05, 2.8023E+05,
     +  2.8089E+05, 2.8154E+05, 2.8220E+05, 2.8286E+05, 2.8352E+05,
     + 2.8418E+05, 2.8484E+05, 2.8551E+05, 2.8617E+05, 2.8684E+05,
     +  2.8751E+05, 2.8817E+05, 2.8884E+05, 2.8951E+05, 2.9019E+05,
     + 2.9086E+05, 2.9153E+05, 2.9221E+05, 2.9288E+05, 2.9356E+05,
     +  2.9424E+05, 2.9492E+05, 2.9560E+05, 2.9628E+05, 2.9697E+05,
     + 2.9765E+05, 2.9834E+05, 2.9903E+05, 2.9971E+05, 3.0040E+05,
     +  3.0110E+05, 3.0179E+05, 3.0248E+05, 3.0317E+05, 3.0387E+05,
     + 3.0457E+05, 3.0527E+05, 3.0596E+05, 3.0667E+05, 3.0737E+05,
     +  3.0807E+05, 3.0877E+05, 3.0948E+05, 3.1019E+05, 3.1089E+05,
     + 3.1160E+05, 3.1231E+05, 3.1302E+05, 3.1374E+05, 3.1445E+05,
     +  3.1517E+05, 3.1588E+05, 3.1660E+05, 3.1732E+05, 3.1804E+05,
     + 3.1876E+05, 3.1948E+05, 3.2021E+05, 3.2093E+05, 3.2166E+05,
     +  3.2239E+05, 3.2311E+05, 3.2384E+05, 3.2458E+05, 3.2531E+05,
     + 3.2604E+05, 3.2678E+05, 3.2751E+05, 3.2825E+05, 3.2899E+05,
     +  3.2973E+05, 3.3047E+05, 3.3121E+05, 3.3196E+05, 3.3270E+05,
     + 3.3345E+05, 3.3420E+05, 3.3494E+05, 3.3569E+05, 3.3645E+05,
     +  3.3720E+05, 3.3795E+05, 3.3871E+05, 3.3946E+05, 3.4022E+05,
     + 3.4098E+05, 3.4174E+05, 3.4250E+05, 3.4327E+05, 3.4403E+05,
     +  3.4480E+05, 3.4556E+05, 3.4633E+05, 3.4710E+05, 3.4787E+05,
     + 3.4864E+05, 3.4942E+05, 3.5019E+05, 3.5097E+05, 3.5174E+05,
     +  3.5252E+05, 3.5330E+05, 3.5408E+05, 3.5486E+05, 3.5565E+05,
     + 3.5643E+05, 3.5722E+05, 3.5801E+05, 3.5880E+05, 3.5959E+05,
     +  3.6038E+05, 3.6117E+05, 3.6196E+05, 3.6276E+05, 3.6356E+05,
     + 3.6435E+05, 3.6515E+05, 3.6595E+05, 3.6676E+05, 3.6756E+05,
     +  3.6836E+05, 3.6917E+05, 3.6998E+05, 3.7079E+05, 3.7160E+05,
     + 3.7241E+05, 3.7322E+05, 3.7404E+05, 3.7485E+05, 3.7567E+05,
     +  3.7649E+05, 3.7730E+05, 3.7813E+05, 3.7895E+05, 3.7977E+05,
     + 3.8060E+05, 3.8142E+05, 3.8225E+05, 3.8308E+05, 3.8391E+05,
     +  3.8474E+05, 3.8557E+05, 3.8641E+05, 3.8724E+05, 3.8808E+05,
     + 3.8892E+05, 3.8976E+05, 3.9060E+05, 3.9144E+05, 3.9229E+05,
     +  3.9313E+05, 3.9398E+05, 3.9482E+05, 3.9567E+05, 3.9652E+05,
     + 3.9738E+05, 3.9823E+05, 3.9909E+05, 3.9994E+05, 4.0080E+05,
     +  4.0166E+05, 4.0252E+05, 4.0338E+05, 4.0424E+05, 4.0511E+05,
     + 4.0597E+05, 4.0684E+05, 4.0771E+05, 4.0858E+05, 4.0945E+05,
     +  4.1032E+05, 4.1120E+05, 4.1207E+05, 4.1295E+05, 4.1383E+05,
     + 4.1471E+05, 4.1559E+05, 4.1647E+05, 4.1736E+05, 4.1824E+05,
     +  4.1913E+05, 4.2002E+05, 4.2091E+05, 4.2180E+05, 4.2269E+05,
     + 4.2359E+05, 4.2448E+05, 4.2538E+05, 4.2628E+05, 4.2718E+05,
     +  4.2808E+05, 4.2898E+05, 4.2989E+05, 4.3079E+05, 4.3170E+05,
     + 4.3261E+05, 4.3352E+05, 4.3443E+05, 4.3534E+05, 4.3626E+05,
     +  4.3717E+05, 4.3809E+05, 4.3901E+05, 4.3993E+05, 4.4085E+05,
     + 4.4177E+05, 4.4270E+05, 4.4362E+05, 4.4455E+05, 4.4548E+05,
     +  4.4641E+05, 4.4734E+05, 4.4827E+05, 4.4921E+05, 4.5014E+05,
     + 4.5108E+05, 4.5202E+05, 4.5296E+05, 4.5390E+05, 4.5484E+05,
     +  4.5579E+05, 4.5674E+05, 4.5768E+05, 4.5863E+05, 4.5958E+05,
     + 4.6054E+05, 4.6149E+05, 4.6244E+05, 4.6340E+05, 4.6436E+05,
     +  4.6532E+05, 4.6628E+05, 4.6724E+05, 4.6821E+05, 4.6917E+05,
     + 4.7014E+05, 4.7111E+05, 4.7208E+05, 4.7305E+05, 4.7402E+05,
     +  4.7500E+05, 4.7597E+05, 4.7695E+05, 4.7793E+05, 4.7891E+05,
     + 4.7989E+05, 4.8088E+05, 4.8186E+05, 4.8285E+05, 4.8384E+05,
     +  4.8483E+05, 4.8582E+05, 4.8681E+05, 4.8780E+05, 4.8880E+05,
     + 4.8980E+05, 4.9080E+05, 4.9180E+05, 4.9280E+05, 4.9380E+05,
     +  4.9481E+05, 4.9581E+05, 4.9682E+05, 4.9783E+05, 4.9884E+05,
     + 4.9985E+05, 5.0087E+05, 5.0188E+05, 5.0290E+05, 5.0392E+05,
     +  5.0494E+05, 5.0596E+05, 5.0699E+05, 5.0801E+05, 5.0904E+05,
     + 5.1007E+05, 5.1110E+05, 5.1213E+05, 5.1316E+05, 5.1419E+05,
     +  5.1523E+05, 5.1627E+05, 5.1731E+05, 5.1835E+05, 5.1939E+05,
     + 5.2043E+05, 5.2148E+05, 5.2253E+05, 5.2357E+05, 5.2462E+05,
     +  5.2568E+05, 5.2673E+05, 5.2778E+05, 5.2884E+05, 5.2990E+05,
     + 5.3096E+05, 5.3202E+05, 5.3308E+05, 5.3415E+05, 5.3521E+05,
     +  5.3628E+05, 5.3735E+05, 5.3842E+05, 5.3949E+05, 5.4057E+05,
     + 5.4164E+05, 5.4272E+05, 5.4380E+05, 5.4488E+05, 5.4596E+05,
     +  5.4704E+05, 5.4813E+05, 5.4922E+05, 5.5030E+05, 5.5139E+05,
     + 5.5249E+05, 5.5358E+05, 5.5467E+05, 5.5577E+05, 5.5687E+05,
     +  5.5797E+05, 5.5907E+05, 5.6017E+05, 5.6128E+05, 5.6238E+05,
     + 5.6349E+05, 5.6460E+05, 5.6571E+05, 5.6682E+05, 5.6794E+05,
     +  5.6906E+05, 5.7017E+05, 5.7129E+05, 5.7241E+05, 5.7354E+05,
     + 5.7466E+05, 5.7579E+05, 5.7691E+05, 5.7804E+05, 5.7917E+05,
     +  5.8031E+05, 5.8144E+05, 5.8258E+05, 5.8371E+05, 5.8485E+05,
     + 5.8599E+05, 5.8714E+05, 5.8828E+05, 5.8943E+05, 5.9057E+05,
     +  5.9172E+05, 5.9287E+05, 5.9403E+05, 5.9518E+05, 5.9634E+05,
     + 5.9749E+05, 5.9865E+05, 5.9981E+05, 6.0098E+05, 6.0214E+05,
     +  6.0331E+05, 6.0447E+05, 6.0564E+05, 6.0681E+05, 6.0799E+05,
     + 6.0916E+05, 6.1034E+05, 6.1152E+05, 6.1270E+05, 6.1388E+05,
     +  6.1506E+05, 6.1624E+05, 6.1743E+05, 6.1862E+05, 6.1981E+05,
     + 6.2100E+05, 6.2219E+05, 6.2339E+05, 6.2458E+05, 6.2578E+05,
     +  6.2698E+05, 6.2818E+05, 6.2939E+05, 6.3059E+05, 6.3180E+05,
     + 6.3301E+05, 6.3422E+05, 6.3543E+05, 6.3664E+05, 6.3786E+05,
     +  6.3908E+05, 6.4030E+05, 6.4152E+05, 6.4274E+05, 6.4396E+05,
     + 6.4519E+05, 6.4642E+05, 6.4765E+05, 6.4888E+05, 6.5011E+05,
     +  6.5134E+05, 6.5258E+05, 6.5382E+05, 6.5506E+05, 6.5630E+05,
     + 6.5754E+05, 6.5879E+05, 6.6004E+05, 6.6128E+05, 6.6254E+05,
     +  6.6379E+05, 6.6504E+05, 6.6630E+05, 6.6755E+05, 6.6881E+05,
     + 6.7007E+05, 6.7134E+05, 6.7260E+05, 6.7387E+05, 6.7514E+05,
     +  6.7641E+05, 6.7768E+05, 6.7895E+05, 6.8023E+05, 6.8150E+05,
     + 6.8278E+05, 6.8406E+05, 6.8535E+05, 6.8663E+05, 6.8792E+05,
     +  6.8920E+05, 6.9049E+05, 6.9178E+05, 6.9308E+05, 6.9437E+05,
     + 6.9567E+05, 6.9697E+05, 6.9827E+05, 6.9957E+05, 7.0087E+05,
     +  7.0218E+05, 7.0349E+05, 7.0480E+05, 7.0611E+05, 7.0742E+05,
     + 7.0874E+05, 7.1005E+05, 7.1137E+05, 7.1269E+05, 7.1401E+05,
     +  7.1534E+05, 7.1666E+05, 7.1799E+05, 7.1932E+05, 7.2065E+05,
     + 7.2199E+05, 7.2332E+05, 7.2466E+05, 7.2600E+05, 7.2734E+05,
     +  7.2868E+05, 7.3002E+05, 7.3137E+05, 7.3272E+05, 7.3407E+05,
     + 7.3542E+05, 7.3677E+05, 7.3813E+05, 7.3948E+05, 7.4084E+05,
     +  7.4220E+05, 7.4357E+05, 7.4493E+05, 7.4630E+05, 7.4767E+05,
     + 7.4904E+05, 7.5041E+05, 7.5178E+05, 7.5316E+05, 7.5454E+05,
     +  7.5592E+05, 7.5730E+05, 7.5868E+05, 7.6007E+05, 7.6145E+05,
     + 7.6284E+05, 7.6423E+05, 7.6562E+05, 7.6702E+05, 7.6842E+05,
     +  7.6981E+05, 7.7121E+05, 7.7262E+05, 7.7402E+05, 7.7543E+05,
     + 7.7683E+05, 7.7824E+05, 7.7965E+05, 7.8107E+05, 7.8248E+05,
     +  7.8390E+05, 7.8532E+05, 7.8674E+05, 7.8816E+05, 7.8959E+05,
     + 7.9102E+05, 7.9244E+05, 7.9388E+05, 7.9531E+05, 7.9674E+05,
     +  7.9818E+05, 7.9962E+05, 8.0106E+05, 8.0250E+05, 8.0394E+05,
     + 8.0539E+05, 8.0684E+05, 8.0829E+05, 8.0974E+05, 8.1119E+05,
     +  8.1265E+05, 8.1411E+05, 8.1557E+05, 8.1703E+05, 8.1849E+05,
     + 8.1996E+05, 8.2142E+05, 8.2289E+05, 8.2436E+05, 8.2584E+05,
     +  8.2731E+05, 8.2879E+05, 8.3027E+05, 8.3175E+05, 8.3323E+05,
     + 8.3472E+05, 8.3621E+05, 8.3769E+05, 8.3919E+05, 8.4068E+05,
     +  8.4217E+05, 8.4367E+05, 8.4517E+05, 8.4667E+05, 8.4817E+05,
     + 8.4968E+05, 8.5118E+05, 8.5269E+05, 8.5420E+05, 8.5571E+05,
     +  8.5723E+05, 8.5875E+05, 8.6026E+05, 8.6178E+05, 8.6331E+05,
     + 8.6483E+05, 8.6636E+05, 8.6789E+05, 8.6942E+05, 8.7095E+05,
     +  8.7248E+05, 8.7402E+05, 8.7556E+05, 8.7710E+05, 8.7864E+05,
     + 8.8019E+05, 8.8173E+05, 8.8328E+05, 8.8483E+05, 8.8638E+05,
     +  8.8794E+05, 8.8950E+05, 8.9105E+05, 8.9262E+05, 8.9418E+05,
     + 8.9574E+05, 8.9731E+05, 8.9888E+05, 9.0045E+05, 9.0202E+05,
     +  9.0360E+05, 9.0517E+05, 9.0675E+05, 9.0833E+05, 9.0992E+05,
     + 9.1150E+05, 9.1309E+05, 9.1468E+05, 9.1627E+05, 9.1786E+05,
     +  9.1946E+05, 9.2105E+05, 9.2265E+05, 9.2425E+05, 9.2586E+05,
     + 9.2746E+05, 9.2907E+05, 9.3068E+05, 9.3229E+05, 9.3391E+05,
     +  9.3552E+05, 9.3714E+05, 9.3876E+05, 9.4038E+05, 9.4201E+05,
     + 9.4363E+05, 9.4526E+05, 9.4689E+05, 9.4852E+05, 9.5016E+05,
     +  9.5179E+05, 9.5343E+05, 9.5507E+05, 9.5672E+05, 9.5836E+05,
     + 9.6001E+05, 9.6166E+05, 9.6331E+05, 9.6496E+05, 9.6662E+05,
     +  9.6827E+05, 9.6993E+05, 9.7159E+05, 9.7326E+05, 9.7492E+05,
     + 9.7659E+05, 9.7826E+05, 9.7993E+05, 9.8161E+05, 9.8328E+05,
     +  9.8496E+05, 9.8664E+05, 9.8833E+05, 9.9001E+05, 9.9170E+05,
     + 9.9339E+05, 9.9508E+05, 9.9677E+05, 9.9847E+05, 1.0002E+06,
     +  1.0019E+06, 1.0036E+06, 1.0053E+06, 1.0070E+06, 1.0087E+06,
     + 1.0104E+06, 1.0121E+06, 1.0138E+06, 1.0155E+06, 1.0173E+06,
     +  1.0190E+06, 1.0207E+06, 1.0224E+06, 1.0242E+06, 1.0259E+06,
     + 1.0276E+06, 1.0294E+06, 1.0311E+06, 1.0328E+06, 1.0346E+06,
     +  1.0363E+06, 1.0381E+06, 1.0398E+06, 1.0415E+06, 1.0433E+06,
     + 1.0451E+06, 1.0468E+06, 1.0486E+06, 1.0503E+06, 1.0521E+06,
     +  1.0539E+06, 1.0556E+06, 1.0574E+06, 1.0592E+06, 1.0609E+06,
     + 1.0627E+06, 1.0645E+06, 1.0663E+06, 1.0680E+06, 1.0698E+06,
     +  1.0716E+06, 1.0734E+06, 1.0752E+06, 1.0770E+06, 1.0788E+06,
     + 1.0806E+06, 1.0824E+06, 1.0842E+06, 1.0860E+06, 1.0878E+06,
     +  1.0896E+06, 1.0914E+06, 1.0932E+06, 1.0950E+06, 1.0969E+06,
     + 1.0987E+06, 1.1005E+06, 1.1023E+06, 1.1041E+06, 1.1060E+06,
     +  1.1078E+06, 1.1096E+06, 1.1115E+06, 1.1133E+06, 1.1151E+06,
     + 1.1170E+06, 1.1188E+06, 1.1207E+06, 1.1225E+06, 1.1244E+06,
     +  1.1262E+06, 1.1281E+06, 1.1299E+06, 1.1318E+06, 1.1337E+06,
     + 1.1355E+06, 1.1374E+06, 1.1393E+06, 1.1411E+06, 1.1430E+06,
     +  1.1449E+06, 1.1468E+06, 1.1486E+06, 1.1505E+06, 1.1524E+06,
     + 1.1543E+06, 1.1562E+06, 1.1581E+06, 1.1600E+06, 1.1619E+06,
     +  1.1638E+06, 1.1657E+06, 1.1676E+06, 1.1695E+06, 1.1714E+06,
     + 1.1733E+06, 1.1752E+06, 1.1771E+06, 1.1790E+06, 1.1809E+06,
     +  1.1829E+06, 1.1848E+06, 1.1867E+06, 1.1886E+06, 1.1906E+06,
     + 1.1925E+06, 1.1944E+06, 1.1964E+06, 1.1983E+06, 1.2002E+06,
     +  1.2022E+06, 1.2041E+06, 1.2061E+06, 1.2080E+06, 1.2100E+06,
     + 1.2119E+06, 1.2139E+06, 1.2159E+06, 1.2178E+06, 1.2198E+06,
     +  1.2218E+06, 1.2237E+06, 1.2257E+06, 1.2277E+06, 1.2296E+06,
     + 1.2316E+06, 1.2336E+06, 1.2356E+06, 1.2376E+06, 1.2396E+06,
     +  1.2415E+06, 1.2435E+06, 1.2455E+06, 1.2475E+06, 1.2495E+06,
     + 1.2515E+06, 1.2535E+06, 1.2555E+06, 1.2576E+06, 1.2596E+06,
     +  1.2616E+06, 1.2636E+06, 1.2656E+06, 1.2676E+06, 1.2697E+06,
     + 1.2717E+06, 1.2737E+06, 1.2757E+06, 1.2778E+06, 1.2798E+06,
     +  1.2818E+06, 1.2839E+06, 1.2859E+06, 1.2880E+06, 1.2900E+06,
     + 1.2921E+06, 1.2941E+06, 1.2962E+06, 1.2982E+06, 1.3003E+06,
     +  1.3023E+06, 1.3044E+06, 1.3065E+06, 1.3085E+06, 1.3106E+06,
     + 1.3127E+06, 1.3148E+06, 1.3168E+06, 1.3189E+06, 1.3210E+06,
     +  1.3231E+06, 1.3252E+06, 1.3273E+06, 1.3294E+06, 1.3314E+06,
     + 1.3335E+06, 1.3356E+06, 1.3377E+06, 1.3398E+06, 1.3420E+06,
     +  1.3441E+06, 1.3462E+06, 1.3483E+06, 1.3504E+06, 1.3525E+06,
     + 1.3546E+06, 1.3568E+06, 1.3589E+06, 1.3610E+06, 1.3632E+06,
     +  1.3653E+06, 1.3674E+06, 1.3696E+06, 1.3717E+06, 1.3738E+06,
     + 1.3760E+06, 1.3781E+06, 1.3803E+06, 1.3824E+06, 1.3846E+06,
     +  1.3868E+06, 1.3889E+06, 1.3911E+06, 1.3932E+06, 1.3954E+06,
     + 1.3976E+06, 1.3998E+06, 1.4019E+06, 1.4041E+06, 1.4063E+06,
     +  1.4085E+06, 1.4106E+06, 1.4128E+06, 1.4150E+06, 1.4172E+06,
     + 1.4194E+06, 1.4216E+06, 1.4238E+06, 1.4260E+06, 1.4282E+06,
     +  1.4304E+06, 1.4326E+06, 1.4348E+06, 1.4371E+06, 1.4393E+06,
     + 1.4415E+06, 1.4437E+06, 1.4459E+06, 1.4482E+06, 1.4504E+06,
     +  1.4526E+06, 1.4549E+06, 1.4571E+06, 1.4593E+06, 1.4616E+06,
     + 1.4638E+06, 1.4661E+06, 1.4683E+06, 1.4706E+06, 1.4728E+06,
     +  1.4751E+06, 1.4773E+06, 1.4796E+06, 1.4819E+06, 1.4841E+06,
     + 1.4864E+06, 1.4887E+06, 1.4910E+06, 1.4932E+06, 1.4955E+06,
     +  1.4978E+06, 1.5001E+06, 1.5024E+06, 1.5046E+06, 1.5069E+06,
     + 1.5092E+06, 1.5115E+06, 1.5138E+06, 1.5161E+06, 1.5184E+06,
     +  1.5207E+06, 1.5231E+06, 1.5254E+06, 1.5277E+06, 1.5300E+06,
     + 1.5323E+06, 1.5346E+06, 1.5370E+06, 1.5393E+06, 1.5416E+06,
     +  1.5440E+06, 1.5463E+06, 1.5486E+06, 1.5510E+06, 1.5533E+06,
     + 1.5557E+06, 1.5580E+06, 1.5604E+06, 1.5627E+06, 1.5651E+06,
     +  1.5674E+06, 1.5698E+06, 1.5722E+06, 1.5745E+06, 1.5769E+06,
     + 1.5793E+06, 1.5816E+06, 1.5840E+06, 1.5864E+06, 1.5888E+06,
     +  1.5912E+06, 1.5935E+06, 1.5959E+06, 1.5983E+06, 1.6007E+06,
     + 1.6031E+06, 1.6055E+06, 1.6079E+06, 1.6103E+06, 1.6127E+06,
     +  1.6151E+06, 1.6176E+06, 1.6200E+06, 1.6224E+06, 1.6248E+06,
     + 1.6272E+06, 1.6297E+06, 1.6321E+06, 1.6345E+06, 1.6370E+06,
     +  1.6394E+06, 1.6418E+06, 1.6443E+06, 1.6467E+06, 1.6492E+06,
     + 1.6516E+06, 1.6541E+06, 1.6565E+06, 1.6590E+06, 1.6614E+06,
     +  1.6639E+06, 1.6664E+06, 1.6688E+06, 1.6713E+06, 1.6738E+06,
     + 1.6763E+06, 1.6787E+06, 1.6812E+06, 1.6837E+06, 1.6862E+06,
     +  1.6887E+06, 1.6912E+06, 1.6937E+06, 1.6962E+06, 1.6987E+06,
     + 1.7012E+06, 1.7037E+06, 1.7062E+06, 1.7087E+06, 1.7112E+06,
     +  1.7137E+06, 1.7162E+06, 1.7187E+06, 1.7213E+06, 1.7238E+06,
     + 1.7263E+06, 1.7289E+06, 1.7314E+06, 1.7339E+06, 1.7365E+06,
     +  1.7390E+06, 1.7416E+06, 1.7441E+06, 1.7467E+06, 1.7492E+06,
     + 1.7518E+06, 1.7543E+06, 1.7569E+06, 1.7595E+06, 1.7620E+06,
     +  1.7646E+06, 1.7672E+06, 1.7697E+06, 1.7723E+06, 1.7749E+06,
     + 1.7775E+06, 1.7801E+06, 1.7826E+06, 1.7852E+06, 1.7878E+06,
     +  1.7904E+06, 1.7930E+06, 1.7956E+06, 1.7982E+06, 1.8008E+06,
     + 1.8035E+06, 1.8061E+06, 1.8087E+06, 1.8113E+06, 1.8139E+06,
     +  1.8165E+06, 1.8192E+06, 1.8218E+06, 1.8244E+06, 1.8271E+06,
     + 1.8297E+06, 1.8323E+06, 1.8350E+06, 1.8376E+06, 1.8403E+06,
     +  1.8429E+06, 1.8456E+06, 1.8482E+06, 1.8509E+06, 1.8536E+06,
     + 1.8562E+06, 1.8589E+06, 1.8616E+06, 1.8642E+06, 1.8669E+06,
     +  1.8696E+06, 1.8723E+06, 1.8750E+06, 1.8776E+06, 1.8803E+06,
     + 1.8830E+06, 1.8857E+06, 1.8884E+06, 1.8911E+06, 1.8938E+06,
     +  1.8965E+06, 1.8992E+06, 1.9019E+06, 1.9047E+06, 1.9074E+06,
     + 1.9101E+06, 1.9128E+06, 1.9155E+06, 1.9183E+06, 1.9210E+06,
     +  1.9237E+06, 1.9265E+06, 1.9292E+06, 1.9320E+06, 1.9347E+06,
     + 1.9375E+06, 1.9402E+06, 1.9430E+06, 1.9457E+06, 1.9485E+06,
     +  1.9512E+06, 1.9540E+06, 1.9568E+06, 1.9595E+06, 1.9623E+06,
     + 1.9651E+06, 1.9679E+06, 1.9706E+06, 1.9734E+06, 1.9762E+06,
     +  1.9790E+06, 1.9818E+06, 1.9846E+06, 1.9874E+06, 1.9902E+06,
     + 1.9930E+06, 1.9958E+06, 1.9986E+06, 2.0014E+06, 2.0042E+06,
     +  2.0071E+06, 2.0099E+06, 2.0127E+06, 2.0155E+06, 2.0184E+06,
     + 2.0212E+06, 2.0240E+06, 2.0269E+06, 2.0297E+06, 2.0326E+06,
     +  2.0354E+06, 2.0382E+06, 2.0411E+06, 2.0440E+06, 2.0468E+06,
     + 2.0497E+06, 2.0525E+06, 2.0554E+06, 2.0583E+06, 2.0611E+06,
     +  2.0640E+06, 2.0669E+06, 2.0698E+06, 2.0727E+06, 2.0755E+06,
     + 2.0784E+06, 2.0813E+06, 2.0842E+06, 2.0871E+06, 2.0900E+06,
     +  2.0929E+06, 2.0958E+06, 2.0987E+06, 2.1017E+06, 2.1046E+06,
     + 2.1075E+06, 2.1104E+06, 2.1133E+06, 2.1163E+06, 2.1192E+06,
     +  2.1221E+06, 2.1251E+06, 2.1280E+06, 2.1309E+06, 2.1339E+06,
     + 2.1368E+06, 2.1398E+06, 2.1427E+06, 2.1457E+06, 2.1486E+06,
     +  2.1516E+06, 2.1546E+06, 2.1575E+06, 2.1605E+06, 2.1635E+06,
     + 2.1665E+06, 2.1694E+06, 2.1724E+06, 2.1754E+06, 2.1784E+06,
     +  2.1814E+06, 2.1844E+06, 2.1874E+06, 2.1904E+06, 2.1934E+06,
     + 2.1964E+06, 2.1994E+06, 2.2024E+06, 2.2054E+06, 2.2084E+06,
     +  2.2114E+06, 2.2145E+06, 2.2175E+06, 2.2205E+06, 2.2236E+06,
     + 2.2266E+06, 2.2296E+06, 2.2327E+06, 2.2357E+06, 2.2388E+06,
     +  2.2418E+06, 2.2449E+06, 2.2479E+06, 2.2510E+06, 2.2540E+06,
     + 2.2571E+06, 2.2602E+06, 2.2632E+06, 2.2663E+06, 2.2694E+06,
     +  2.2725E+06, 2.2755E+06, 2.2786E+06, 2.2817E+06, 2.2848E+06,
     + 2.2879E+06, 2.2910E+06, 2.2941E+06, 2.2972E+06, 2.3003E+06,
     +  2.3034E+06, 2.3065E+06, 2.3096E+06, 2.3128E+06, 2.3159E+06,
     + 2.3190E+06, 2.3221E+06, 2.3252E+06, 2.3284E+06, 2.3315E+06,
     +  2.3347E+06, 2.3378E+06, 2.3409E+06, 2.3441E+06, 2.3472E+06,
     + 2.3504E+06, 2.3535E+06, 2.3567E+06, 2.3599E+06, 2.3630E+06,
     +  2.3662E+06, 2.3694E+06, 2.3725E+06, 2.3757E+06, 2.3789E+06,
     + 2.3821E+06, 2.3853E+06, 2.3885E+06, 2.3917E+06, 2.3948E+06,
     +  2.3980E+06, 2.4012E+06, 2.4044E+06, 2.4077E+06, 2.4109E+06/
      data (QofT( 6,J),J=2409,4510)/ 2.4141E+06, 2.4173E+06,
     +  2.4205E+06, 2.4237E+06, 2.4270E+06, 2.4302E+06, 2.4334E+06,
     +  2.4367E+06,
     + 2.4399E+06, 2.4431E+06, 2.4464E+06, 2.4496E+06, 2.4529E+06,
     +  2.4561E+06, 2.4594E+06, 2.4626E+06, 2.4659E+06, 2.4692E+06,
     + 2.4724E+06, 2.4757E+06, 2.4790E+06, 2.4823E+06, 2.4855E+06,
     +  2.4888E+06, 2.4921E+06, 2.4954E+06, 2.4987E+06, 2.5020E+06,
     + 2.5053E+06, 2.5086E+06, 2.5119E+06, 2.5152E+06, 2.5185E+06,
     +  2.5218E+06, 2.5251E+06, 2.5285E+06, 2.5318E+06, 2.5351E+06,
     + 2.5384E+06, 2.5418E+06, 2.5451E+06, 2.5484E+06, 2.5518E+06,
     +  2.5551E+06, 2.5585E+06, 2.5618E+06, 2.5652E+06, 2.5685E+06,
     + 2.5719E+06, 2.5752E+06, 2.5786E+06, 2.5820E+06, 2.5854E+06,
     +  2.5887E+06, 2.5921E+06, 2.5955E+06, 2.5989E+06, 2.6023E+06,
     + 2.6057E+06, 2.6090E+06, 2.6124E+06, 2.6158E+06, 2.6192E+06,
     +  2.6226E+06, 2.6261E+06, 2.6295E+06, 2.6329E+06, 2.6363E+06,
     + 2.6397E+06, 2.6431E+06, 2.6466E+06, 2.6500E+06, 2.6534E+06,
     +  2.6569E+06, 2.6603E+06, 2.6638E+06, 2.6672E+06, 2.6707E+06,
     + 2.6741E+06, 2.6776E+06, 2.6810E+06, 2.6845E+06, 2.6880E+06,
     +  2.6914E+06, 2.6949E+06, 2.6984E+06, 2.7018E+06, 2.7053E+06,
     + 2.7088E+06, 2.7123E+06, 2.7158E+06, 2.7193E+06, 2.7228E+06,
     +  2.7263E+06, 2.7298E+06, 2.7333E+06, 2.7368E+06, 2.7403E+06,
     + 2.7438E+06, 2.7473E+06, 2.7509E+06, 2.7544E+06, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       262
      data (QofT( 7,J),J=1,2408)/ 6.0000E+00, 6.0015E+00, 6.0282E+00,
     +  6.1302E+00, 6.3426E+00, 6.6776E+00, 7.1332E+00, 7.7007E+00,
     + 8.3695E+00, 9.1292E+00, 9.9706E+00, 1.0885E+01, 1.1867E+01,
     +  1.2909E+01, 1.4007E+01, 1.5156E+01, 1.6354E+01, 1.7596E+01,
     + 1.8881E+01, 2.0206E+01, 2.2969E+01, 2.5872E+01, 2.8906E+01,
     +  3.2063E+01, 3.5336E+01, 3.8720E+01, 4.2210E+01, 4.5802E+01,
     + 4.9494E+01, 5.3282E+01, 5.7163E+01, 6.1136E+01, 6.5198E+01,
     +  6.9346E+01, 7.3581E+01, 7.7898E+01, 8.2298E+01, 8.6778E+01,
     + 9.1337E+01, 9.5974E+01, 1.0069E+02, 1.0548E+02, 1.1034E+02,
     +  1.1527E+02, 1.2028E+02, 1.2536E+02, 1.3051E+02, 1.3573E+02,
     + 1.4101E+02, 1.4637E+02, 1.5179E+02, 1.5727E+02, 1.6282E+02,
     +  1.6844E+02, 1.7412E+02, 1.7986E+02, 1.8566E+02, 1.9152E+02,
     + 1.9745E+02, 2.0344E+02, 2.0948E+02, 2.1559E+02, 2.2175E+02,
     +  2.2797E+02, 2.3425E+02, 2.4059E+02, 2.4698E+02, 2.5343E+02,
     + 2.5994E+02, 2.6650E+02, 2.7311E+02, 2.7978E+02, 2.8650E+02,
     +  2.9328E+02, 3.0011E+02, 3.0699E+02, 3.1392E+02, 3.2091E+02,
     + 3.2795E+02, 3.3504E+02, 3.4218E+02, 3.4937E+02, 3.5661E+02,
     +  3.6390E+02, 3.7124E+02, 3.7863E+02, 3.8606E+02, 3.9355E+02,
     + 4.0109E+02, 4.0867E+02, 4.1630E+02, 4.2398E+02, 4.3171E+02,
     +  4.3948E+02, 4.4730E+02, 4.5517E+02, 4.6308E+02, 4.7104E+02,
     + 4.7905E+02, 4.8710E+02, 4.9519E+02, 5.0334E+02, 5.1152E+02,
     +  5.1975E+02, 5.2803E+02, 5.3635E+02, 5.4472E+02, 5.5313E+02,
     + 5.6158E+02, 5.7008E+02, 5.7862E+02, 5.8721E+02, 5.9584E+02,
     +  6.0451E+02, 6.1322E+02, 6.2198E+02, 6.3079E+02, 6.3963E+02,
     + 6.4852E+02, 6.5745E+02, 6.6642E+02, 6.7544E+02, 6.8449E+02,
     +  6.9359E+02, 7.0273E+02, 7.1192E+02, 7.2114E+02, 7.3041E+02,
     + 7.3972E+02, 7.4907E+02, 7.5847E+02, 7.6790E+02, 7.7738E+02,
     +  7.8690E+02, 7.9646E+02, 8.0606E+02, 8.1571E+02, 8.2539E+02,
     + 8.3512E+02, 8.4489E+02, 8.5470E+02, 8.6455E+02, 8.7444E+02,
     +  8.8437E+02, 8.9435E+02, 9.0436E+02, 9.1442E+02, 9.2452E+02,
     + 9.3466E+02, 9.4484E+02, 9.5507E+02, 9.6533E+02, 9.7564E+02,
     +  9.8598E+02, 9.9637E+02, 1.0068E+03, 1.0173E+03, 1.0278E+03,
     + 1.0383E+03, 1.0489E+03, 1.0596E+03, 1.0703E+03, 1.0810E+03,
     +  1.0917E+03, 1.1025E+03, 1.1134E+03, 1.1243E+03, 1.1352E+03,
     + 1.1462E+03, 1.1572E+03, 1.1682E+03, 1.1793E+03, 1.1905E+03,
     +  1.2017E+03, 1.2129E+03, 1.2242E+03, 1.2355E+03, 1.2468E+03,
     + 1.2582E+03, 1.2697E+03, 1.2811E+03, 1.2927E+03, 1.3042E+03,
     +  1.3158E+03, 1.3275E+03, 1.3392E+03, 1.3509E+03, 1.3627E+03,
     + 1.3745E+03, 1.3864E+03, 1.3983E+03, 1.4102E+03, 1.4222E+03,
     +  1.4343E+03, 1.4463E+03, 1.4585E+03, 1.4706E+03, 1.4829E+03,
     + 1.4951E+03, 1.5074E+03, 1.5198E+03, 1.5322E+03, 1.5446E+03,
     +  1.5571E+03, 1.5696E+03, 1.5822E+03, 1.5948E+03, 1.6074E+03,
     + 1.6201E+03, 1.6329E+03, 1.6457E+03, 1.6585E+03, 1.6714E+03,
     +  1.6843E+03, 1.6973E+03, 1.7103E+03, 1.7234E+03, 1.7365E+03,
     + 1.7497E+03, 1.7629E+03, 1.7761E+03, 1.7894E+03, 1.8028E+03,
     +  1.8161E+03, 1.8296E+03, 1.8431E+03, 1.8566E+03, 1.8702E+03,
     + 1.8838E+03, 1.8975E+03, 1.9112E+03, 1.9249E+03, 1.9388E+03,
     +  1.9526E+03, 1.9665E+03, 1.9805E+03, 1.9945E+03, 2.0085E+03,
     + 2.0226E+03, 2.0368E+03, 2.0510E+03, 2.0652E+03, 2.0795E+03,
     +  2.0939E+03, 2.1083E+03, 2.1227E+03, 2.1372E+03, 2.1517E+03,
     + 2.1663E+03, 2.1809E+03, 2.1956E+03, 2.2104E+03, 2.2252E+03,
     +  2.2400E+03, 2.2549E+03, 2.2698E+03, 2.2848E+03, 2.2998E+03,
     + 2.3149E+03, 2.3301E+03, 2.3453E+03, 2.3605E+03, 2.3758E+03,
     +  2.3911E+03, 2.4065E+03, 2.4220E+03, 2.4375E+03, 2.4530E+03,
     + 2.4686E+03, 2.4843E+03, 2.5000E+03, 2.5158E+03, 2.5316E+03,
     +  2.5475E+03, 2.5634E+03, 2.5793E+03, 2.5954E+03, 2.6114E+03,
     + 2.6276E+03, 2.6438E+03, 2.6600E+03, 2.6763E+03, 2.6926E+03,
     +  2.7090E+03, 2.7255E+03, 2.7420E+03, 2.7586E+03, 2.7752E+03,
     + 2.7919E+03, 2.8086E+03, 2.8254E+03, 2.8422E+03, 2.8591E+03,
     +  2.8761E+03, 2.8931E+03, 2.9101E+03, 2.9273E+03, 2.9444E+03,
     + 2.9617E+03, 2.9790E+03, 2.9963E+03, 3.0137E+03, 3.0312E+03,
     +  3.0487E+03, 3.0663E+03, 3.0839E+03, 3.1016E+03, 3.1193E+03,
     + 3.1371E+03, 3.1550E+03, 3.1729E+03, 3.1909E+03, 3.2089E+03,
     +  3.2270E+03, 3.2452E+03, 3.2634E+03, 3.2817E+03, 3.3000E+03,
     + 3.3184E+03, 3.3369E+03, 3.3554E+03, 3.3740E+03, 3.3926E+03,
     +  3.4113E+03, 3.4300E+03, 3.4489E+03, 3.4677E+03, 3.4867E+03,
     + 3.5057E+03, 3.5247E+03, 3.5439E+03, 3.5630E+03, 3.5823E+03,
     +  3.6016E+03, 3.6210E+03, 3.6404E+03, 3.6599E+03, 3.6795E+03,
     + 3.6991E+03, 3.7188E+03, 3.7385E+03, 3.7583E+03, 3.7782E+03,
     +  3.7982E+03, 3.8182E+03, 3.8382E+03, 3.8584E+03, 3.8786E+03,
     + 3.8988E+03, 3.9192E+03, 3.9396E+03, 3.9600E+03, 3.9806E+03,
     +  4.0012E+03, 4.0218E+03, 4.0425E+03, 4.0633E+03, 4.0842E+03,
     + 4.1051E+03, 4.1261E+03, 4.1472E+03, 4.1683E+03, 4.1895E+03,
     +  4.2107E+03, 4.2321E+03, 4.2535E+03, 4.2749E+03, 4.2965E+03,
     + 4.3181E+03, 4.3397E+03, 4.3615E+03, 4.3833E+03, 4.4052E+03,
     +  4.4271E+03, 4.4491E+03, 4.4712E+03, 4.4934E+03, 4.5156E+03,
     + 4.5379E+03, 4.5603E+03, 4.5827E+03, 4.6052E+03, 4.6278E+03,
     +  4.6505E+03, 4.6732E+03, 4.6960E+03, 4.7189E+03, 4.7418E+03,
     + 4.7648E+03, 4.7879E+03, 4.8111E+03, 4.8343E+03, 4.8576E+03,
     +  4.8810E+03, 4.9045E+03, 4.9280E+03, 4.9516E+03, 4.9753E+03,
     + 4.9990E+03, 5.0229E+03, 5.0468E+03, 5.0708E+03, 5.0948E+03,
     +  5.1189E+03, 5.1432E+03, 5.1674E+03, 5.1918E+03, 5.2162E+03,
     + 5.2407E+03, 5.2653E+03, 5.2900E+03, 5.3147E+03, 5.3396E+03,
     +  5.3645E+03, 5.3894E+03, 5.4145E+03, 5.4396E+03, 5.4648E+03,
     + 5.4901E+03, 5.5155E+03, 5.5410E+03, 5.5665E+03, 5.5921E+03,
     +  5.6178E+03, 5.6436E+03, 5.6694E+03, 5.6953E+03, 5.7214E+03,
     + 5.7474E+03, 5.7736E+03, 5.7999E+03, 5.8262E+03, 5.8526E+03,
     +  5.8791E+03, 5.9057E+03, 5.9324E+03, 5.9591E+03, 5.9860E+03,
     + 6.0129E+03, 6.0399E+03, 6.0670E+03, 6.0941E+03, 6.1214E+03,
     +  6.1487E+03, 6.1761E+03, 6.2036E+03, 6.2312E+03, 6.2589E+03,
     + 6.2867E+03, 6.3145E+03, 6.3424E+03, 6.3705E+03, 6.3986E+03,
     +  6.4267E+03, 6.4550E+03, 6.4834E+03, 6.5118E+03, 6.5404E+03,
     + 6.5690E+03, 6.5977E+03, 6.6265E+03, 6.6554E+03, 6.6844E+03,
     +  6.7135E+03, 6.7426E+03, 6.7719E+03, 6.8012E+03, 6.8307E+03,
     + 6.8602E+03, 6.8898E+03, 6.9195E+03, 6.9493E+03, 6.9792E+03,
     +  7.0091E+03, 7.0392E+03, 7.0694E+03, 7.0996E+03, 7.1299E+03,
     + 7.1604E+03, 7.1909E+03, 7.2215E+03, 7.2522E+03, 7.2831E+03,
     +  7.3140E+03, 7.3449E+03, 7.3760E+03, 7.4072E+03, 7.4385E+03,
     + 7.4699E+03, 7.5013E+03, 7.5329E+03, 7.5646E+03, 7.5963E+03,
     +  7.6282E+03, 7.6601E+03, 7.6922E+03, 7.7243E+03, 7.7565E+03,
     + 7.7889E+03, 7.8213E+03, 7.8538E+03, 7.8865E+03, 7.9192E+03,
     +  7.9520E+03, 7.9850E+03, 8.0180E+03, 8.0511E+03, 8.0843E+03,
     + 8.1177E+03, 8.1511E+03, 8.1846E+03, 8.2182E+03, 8.2520E+03,
     +  8.2858E+03, 8.3197E+03, 8.3538E+03, 8.3879E+03, 8.4221E+03,
     + 8.4565E+03, 8.4909E+03, 8.5255E+03, 8.5601E+03, 8.5948E+03,
     +  8.6297E+03, 8.6647E+03, 8.6997E+03, 8.7349E+03, 8.7702E+03,
     + 8.8055E+03, 8.8410E+03, 8.8766E+03, 8.9123E+03, 8.9481E+03,
     +  8.9840E+03, 9.0200E+03, 9.0562E+03, 9.0924E+03, 9.1287E+03,
     + 9.1652E+03, 9.2017E+03, 9.2384E+03, 9.2751E+03, 9.3120E+03,
     +  9.3490E+03, 9.3861E+03, 9.4233E+03, 9.4606E+03, 9.4980E+03,
     + 9.5356E+03, 9.5732E+03, 9.6110E+03, 9.6488E+03, 9.6868E+03,
     +  9.7249E+03, 9.7631E+03, 9.8014E+03, 9.8399E+03, 9.8784E+03,
     + 9.9170E+03, 9.9558E+03, 9.9947E+03, 1.0034E+04, 1.0073E+04,
     +  1.0112E+04, 1.0151E+04, 1.0191E+04, 1.0230E+04, 1.0270E+04,
     + 1.0310E+04, 1.0350E+04, 1.0390E+04, 1.0430E+04, 1.0470E+04,
     +  1.0511E+04, 1.0551E+04, 1.0592E+04, 1.0632E+04, 1.0673E+04,
     + 1.0714E+04, 1.0755E+04, 1.0797E+04, 1.0838E+04, 1.0879E+04,
     +  1.0921E+04, 1.0963E+04, 1.1004E+04, 1.1046E+04, 1.1088E+04,
     + 1.1131E+04, 1.1173E+04, 1.1215E+04, 1.1258E+04, 1.1301E+04,
     +  1.1343E+04, 1.1386E+04, 1.1429E+04, 1.1472E+04, 1.1516E+04,
     + 1.1559E+04, 1.1603E+04, 1.1646E+04, 1.1690E+04, 1.1734E+04,
     +  1.1778E+04, 1.1822E+04, 1.1866E+04, 1.1911E+04, 1.1955E+04,
     + 1.2000E+04, 1.2045E+04, 1.2090E+04, 1.2135E+04, 1.2180E+04,
     +  1.2225E+04, 1.2271E+04, 1.2316E+04, 1.2362E+04, 1.2408E+04,
     + 1.2454E+04, 1.2500E+04, 1.2546E+04, 1.2592E+04, 1.2639E+04,
     +  1.2685E+04, 1.2732E+04, 1.2779E+04, 1.2826E+04, 1.2873E+04,
     + 1.2921E+04, 1.2968E+04, 1.3015E+04, 1.3063E+04, 1.3111E+04,
     +  1.3159E+04, 1.3207E+04, 1.3255E+04, 1.3304E+04, 1.3352E+04,
     + 1.3401E+04, 1.3449E+04, 1.3498E+04, 1.3547E+04, 1.3596E+04,
     +  1.3646E+04, 1.3695E+04, 1.3745E+04, 1.3795E+04, 1.3844E+04,
     + 1.3894E+04, 1.3945E+04, 1.3995E+04, 1.4045E+04, 1.4096E+04,
     +  1.4146E+04, 1.4197E+04, 1.4248E+04, 1.4299E+04, 1.4351E+04,
     + 1.4402E+04, 1.4454E+04, 1.4505E+04, 1.4557E+04, 1.4609E+04,
     +  1.4661E+04, 1.4714E+04, 1.4766E+04, 1.4819E+04, 1.4871E+04,
     + 1.4924E+04, 1.4977E+04, 1.5030E+04, 1.5084E+04, 1.5137E+04,
     +  1.5191E+04, 1.5244E+04, 1.5298E+04, 1.5352E+04, 1.5406E+04,
     + 1.5461E+04, 1.5515E+04, 1.5570E+04, 1.5625E+04, 1.5679E+04,
     +  1.5735E+04, 1.5790E+04, 1.5845E+04, 1.5901E+04, 1.5956E+04,
     + 1.6012E+04, 1.6068E+04, 1.6124E+04, 1.6181E+04, 1.6237E+04,
     +  1.6294E+04, 1.6350E+04, 1.6407E+04, 1.6464E+04, 1.6522E+04,
     + 1.6579E+04, 1.6636E+04, 1.6694E+04, 1.6752E+04, 1.6810E+04,
     +  1.6868E+04, 1.6926E+04, 1.6985E+04, 1.7043E+04, 1.7102E+04,
     + 1.7161E+04, 1.7220E+04, 1.7280E+04, 1.7339E+04, 1.7399E+04,
     +  1.7458E+04, 1.7518E+04, 1.7578E+04, 1.7638E+04, 1.7699E+04,
     + 1.7759E+04, 1.7820E+04, 1.7881E+04, 1.7942E+04, 1.8003E+04,
     +  1.8065E+04, 1.8126E+04, 1.8188E+04, 1.8250E+04, 1.8312E+04,
     + 1.8374E+04, 1.8436E+04, 1.8499E+04, 1.8561E+04, 1.8624E+04,
     +  1.8687E+04, 1.8750E+04, 1.8814E+04, 1.8877E+04, 1.8941E+04,
     + 1.9005E+04, 1.9069E+04, 1.9133E+04, 1.9197E+04, 1.9262E+04,
     +  1.9327E+04, 1.9392E+04, 1.9457E+04, 1.9522E+04, 1.9587E+04,
     + 1.9653E+04, 1.9719E+04, 1.9784E+04, 1.9851E+04, 1.9917E+04,
     +  1.9983E+04, 2.0050E+04, 2.0117E+04, 2.0184E+04, 2.0251E+04,
     + 2.0318E+04, 2.0386E+04, 2.0453E+04, 2.0521E+04, 2.0589E+04,
     +  2.0657E+04, 2.0726E+04, 2.0794E+04, 2.0863E+04, 2.0932E+04,
     + 2.1001E+04, 2.1070E+04, 2.1140E+04, 2.1209E+04, 2.1279E+04,
     +  2.1349E+04, 2.1419E+04, 2.1490E+04, 2.1560E+04, 2.1631E+04,
     + 2.1702E+04, 2.1773E+04, 2.1844E+04, 2.1916E+04, 2.1987E+04,
     +  2.2059E+04, 2.2131E+04, 2.2203E+04, 2.2276E+04, 2.2348E+04,
     + 2.2421E+04, 2.2494E+04, 2.2567E+04, 2.2640E+04, 2.2714E+04,
     +  2.2788E+04, 2.2862E+04, 2.2936E+04, 2.3010E+04, 2.3084E+04,
     + 2.3159E+04, 2.3234E+04, 2.3309E+04, 2.3384E+04, 2.3460E+04,
     +  2.3535E+04, 2.3611E+04, 2.3687E+04, 2.3763E+04, 2.3840E+04,
     + 2.3916E+04, 2.3993E+04, 2.4070E+04, 2.4147E+04, 2.4224E+04,
     +  2.4302E+04, 2.4380E+04, 2.4458E+04, 2.4536E+04, 2.4614E+04,
     + 2.4693E+04, 2.4771E+04, 2.4850E+04, 2.4930E+04, 2.5009E+04,
     +  2.5088E+04, 2.5168E+04, 2.5248E+04, 2.5328E+04, 2.5409E+04,
     + 2.5489E+04, 2.5570E+04, 2.5651E+04, 2.5732E+04, 2.5813E+04,
     +  2.5895E+04, 2.5977E+04, 2.6059E+04, 2.6141E+04, 2.6223E+04,
     + 2.6306E+04, 2.6389E+04, 2.6472E+04, 2.6555E+04, 2.6638E+04,
     +  2.6722E+04, 2.6806E+04, 2.6890E+04, 2.6974E+04, 2.7059E+04,
     + 2.7143E+04, 2.7228E+04, 2.7313E+04, 2.7399E+04, 2.7484E+04,
     +  2.7570E+04, 2.7656E+04, 2.7742E+04, 2.7829E+04, 2.7915E+04,
     + 2.8002E+04, 2.8089E+04, 2.8176E+04, 2.8264E+04, 2.8351E+04,
     +  2.8439E+04, 2.8527E+04, 2.8616E+04, 2.8704E+04, 2.8793E+04,
     + 2.8882E+04, 2.8971E+04, 2.9061E+04, 2.9150E+04, 2.9240E+04,
     +  2.9330E+04, 2.9420E+04, 2.9511E+04, 2.9602E+04, 2.9693E+04,
     + 2.9784E+04, 2.9875E+04, 2.9967E+04, 3.0059E+04, 3.0151E+04,
     +  3.0243E+04, 3.0336E+04, 3.0428E+04, 3.0521E+04, 3.0615E+04,
     + 3.0708E+04, 3.0802E+04, 3.0896E+04, 3.0990E+04, 3.1084E+04,
     +  3.1179E+04, 3.1273E+04, 3.1368E+04, 3.1464E+04, 3.1559E+04,
     + 3.1655E+04, 3.1751E+04, 3.1847E+04, 3.1943E+04, 3.2040E+04,
     +  3.2137E+04, 3.2234E+04, 3.2332E+04, 3.2429E+04, 3.2527E+04,
     + 3.2625E+04, 3.2723E+04, 3.2822E+04, 3.2921E+04, 3.3020E+04,
     +  3.3119E+04, 3.3218E+04, 3.3318E+04, 3.3418E+04, 3.3518E+04,
     + 3.3619E+04, 3.3719E+04, 3.3820E+04, 3.3921E+04, 3.4023E+04,
     +  3.4124E+04, 3.4226E+04, 3.4328E+04, 3.4431E+04, 3.4533E+04,
     + 3.4636E+04, 3.4739E+04, 3.4843E+04, 3.4946E+04, 3.5050E+04,
     +  3.5154E+04, 3.5259E+04, 3.5363E+04, 3.5468E+04, 3.5573E+04,
     + 3.5678E+04, 3.5784E+04, 3.5890E+04, 3.5996E+04, 3.6102E+04,
     +  3.6209E+04, 3.6316E+04, 3.6423E+04, 3.6530E+04, 3.6638E+04,
     + 3.6745E+04, 3.6854E+04, 3.6962E+04, 3.7070E+04, 3.7179E+04,
     +  3.7288E+04, 3.7398E+04, 3.7507E+04, 3.7617E+04, 3.7727E+04,
     + 3.7838E+04, 3.7949E+04, 3.8059E+04, 3.8171E+04, 3.8282E+04,
     +  3.8394E+04, 3.8506E+04, 3.8618E+04, 3.8730E+04, 3.8843E+04,
     + 3.8956E+04, 3.9069E+04, 3.9183E+04, 3.9297E+04, 3.9411E+04,
     +  3.9525E+04, 3.9640E+04, 3.9754E+04, 3.9870E+04, 3.9985E+04,
     + 4.0101E+04, 4.0217E+04, 4.0333E+04, 4.0449E+04, 4.0566E+04,
     +  4.0683E+04, 4.0800E+04, 4.0918E+04, 4.1035E+04, 4.1154E+04,
     + 4.1272E+04, 4.1390E+04, 4.1509E+04, 4.1629E+04, 4.1748E+04,
     +  4.1868E+04, 4.1988E+04, 4.2108E+04, 4.2228E+04, 4.2349E+04,
     + 4.2470E+04, 4.2592E+04, 4.2713E+04, 4.2835E+04, 4.2957E+04,
     +  4.3080E+04, 4.3203E+04, 4.3326E+04, 4.3449E+04, 4.3573E+04,
     + 4.3697E+04, 4.3821E+04, 4.3945E+04, 4.4070E+04, 4.4195E+04,
     +  4.4320E+04, 4.4446E+04, 4.4572E+04, 4.4698E+04, 4.4824E+04,
     + 4.4951E+04, 4.5078E+04, 4.5205E+04, 4.5333E+04, 4.5461E+04,
     +  4.5589E+04, 4.5717E+04, 4.5846E+04, 4.5975E+04, 4.6104E+04,
     + 4.6234E+04, 4.6364E+04, 4.6494E+04, 4.6625E+04, 4.6755E+04,
     +  4.6886E+04, 4.7018E+04, 4.7150E+04, 4.7281E+04, 4.7414E+04,
     + 4.7546E+04, 4.7679E+04, 4.7812E+04, 4.7946E+04, 4.8079E+04,
     +  4.8214E+04, 4.8348E+04, 4.8482E+04, 4.8617E+04, 4.8753E+04,
     + 4.8888E+04, 4.9024E+04, 4.9160E+04, 4.9297E+04, 4.9433E+04,
     +  4.9570E+04, 4.9708E+04, 4.9846E+04, 4.9984E+04, 5.0122E+04,
     + 5.0260E+04, 5.0399E+04, 5.0538E+04, 5.0678E+04, 5.0818E+04,
     +  5.0958E+04, 5.1098E+04, 5.1239E+04, 5.1380E+04, 5.1522E+04,
     + 5.1663E+04, 5.1805E+04, 5.1948E+04, 5.2090E+04, 5.2233E+04,
     +  5.2376E+04, 5.2520E+04, 5.2664E+04, 5.2808E+04, 5.2952E+04,
     + 5.3097E+04, 5.3242E+04, 5.3388E+04, 5.3534E+04, 5.3680E+04,
     +  5.3826E+04, 5.3973E+04, 5.4120E+04, 5.4267E+04, 5.4415E+04,
     + 5.4563E+04, 5.4712E+04, 5.4860E+04, 5.5009E+04, 5.5159E+04,
     +  5.5308E+04, 5.5458E+04, 5.5608E+04, 5.5759E+04, 5.5910E+04,
     + 5.6061E+04, 5.6213E+04, 5.6365E+04, 5.6517E+04, 5.6670E+04,
     +  5.6823E+04, 5.6976E+04, 5.7129E+04, 5.7283E+04, 5.7438E+04,
     + 5.7592E+04, 5.7747E+04, 5.7902E+04, 5.8058E+04, 5.8214E+04,
     +  5.8370E+04, 5.8527E+04, 5.8684E+04, 5.8841E+04, 5.8999E+04,
     + 5.9157E+04, 5.9315E+04, 5.9474E+04, 5.9632E+04, 5.9792E+04,
     +  5.9951E+04, 6.0111E+04, 6.0272E+04, 6.0432E+04, 6.0594E+04,
     + 6.0755E+04, 6.0917E+04, 6.1079E+04, 6.1241E+04, 6.1404E+04,
     +  6.1567E+04, 6.1730E+04, 6.1894E+04, 6.2058E+04, 6.2223E+04,
     + 6.2388E+04, 6.2553E+04, 6.2718E+04, 6.2884E+04, 6.3050E+04,
     +  6.3217E+04, 6.3384E+04, 6.3551E+04, 6.3719E+04, 6.3887E+04,
     + 6.4055E+04, 6.4224E+04, 6.4393E+04, 6.4563E+04, 6.4732E+04,
     +  6.4903E+04, 6.5073E+04, 6.5244E+04, 6.5415E+04, 6.5587E+04,
     + 6.5759E+04, 6.5931E+04, 6.6104E+04, 6.6277E+04, 6.6450E+04,
     +  6.6624E+04, 6.6798E+04, 6.6973E+04, 6.7148E+04, 6.7323E+04,
     + 6.7498E+04, 6.7674E+04, 6.7851E+04, 6.8027E+04, 6.8204E+04,
     +  6.8382E+04, 6.8560E+04, 6.8738E+04, 6.8916E+04, 6.9095E+04,
     + 6.9275E+04, 6.9454E+04, 6.9634E+04, 6.9815E+04, 6.9996E+04,
     +  7.0177E+04, 7.0358E+04, 7.0540E+04, 7.0723E+04, 7.0905E+04,
     + 7.1088E+04, 7.1272E+04, 7.1456E+04, 7.1640E+04, 7.1825E+04,
     +  7.2010E+04, 7.2195E+04, 7.2381E+04, 7.2567E+04, 7.2753E+04,
     + 7.2940E+04, 7.3127E+04, 7.3315E+04, 7.3503E+04, 7.3692E+04,
     +  7.3880E+04, 7.4070E+04, 7.4259E+04, 7.4449E+04, 7.4640E+04,
     + 7.4830E+04, 7.5021E+04, 7.5213E+04, 7.5405E+04, 7.5597E+04,
     +  7.5790E+04, 7.5983E+04, 7.6177E+04, 7.6371E+04, 7.6565E+04,
     + 7.6760E+04, 7.6955E+04, 7.7150E+04, 7.7346E+04, 7.7543E+04,
     +  7.7739E+04, 7.7936E+04, 7.8134E+04, 7.8332E+04, 7.8530E+04,
     + 7.8729E+04, 7.8928E+04, 7.9128E+04, 7.9327E+04, 7.9528E+04,
     +  7.9729E+04, 7.9930E+04, 8.0131E+04, 8.0333E+04, 8.0536E+04,
     + 8.0738E+04, 8.0942E+04, 8.1145E+04, 8.1349E+04, 8.1554E+04,
     +  8.1759E+04, 8.1964E+04, 8.2169E+04, 8.2376E+04, 8.2582E+04,
     + 8.2789E+04, 8.2996E+04, 8.3204E+04, 8.3412E+04, 8.3621E+04,
     +  8.3830E+04, 8.4039E+04, 8.4249E+04, 8.4459E+04, 8.4670E+04,
     + 8.4881E+04, 8.5093E+04, 8.5305E+04, 8.5517E+04, 8.5730E+04,
     +  8.5943E+04, 8.6157E+04, 8.6371E+04, 8.6585E+04, 8.6800E+04,
     + 8.7016E+04, 8.7231E+04, 8.7448E+04, 8.7664E+04, 8.7881E+04,
     +  8.8099E+04, 8.8317E+04, 8.8535E+04, 8.8754E+04, 8.8973E+04,
     + 8.9193E+04, 8.9413E+04, 8.9634E+04, 8.9855E+04, 9.0076E+04,
     +  9.0298E+04, 9.0520E+04, 9.0743E+04, 9.0966E+04, 9.1190E+04,
     + 9.1414E+04, 9.1639E+04, 9.1864E+04, 9.2089E+04, 9.2315E+04,
     +  9.2541E+04, 9.2768E+04, 9.2995E+04, 9.3223E+04, 9.3451E+04,
     + 9.3679E+04, 9.3908E+04, 9.4138E+04, 9.4368E+04, 9.4598E+04,
     +  9.4829E+04, 9.5060E+04, 9.5292E+04, 9.5524E+04, 9.5757E+04,
     + 9.5990E+04, 9.6223E+04, 9.6457E+04, 9.6692E+04, 9.6926E+04,
     +  9.7162E+04, 9.7398E+04, 9.7634E+04, 9.7871E+04, 9.8108E+04,
     + 9.8345E+04, 9.8584E+04, 9.8822E+04, 9.9061E+04, 9.9301E+04,
     +  9.9541E+04, 9.9781E+04, 1.0002E+05, 1.0026E+05, 1.0051E+05,
     + 1.0075E+05, 1.0099E+05, 1.0123E+05, 1.0148E+05, 1.0172E+05,
     +  1.0197E+05, 1.0221E+05, 1.0246E+05, 1.0270E+05, 1.0295E+05,
     + 1.0320E+05, 1.0344E+05, 1.0369E+05, 1.0394E+05, 1.0419E+05,
     +  1.0444E+05, 1.0469E+05, 1.0494E+05, 1.0519E+05, 1.0544E+05,
     + 1.0569E+05, 1.0595E+05, 1.0620E+05, 1.0645E+05, 1.0671E+05,
     +  1.0696E+05, 1.0721E+05, 1.0747E+05, 1.0773E+05, 1.0798E+05,
     + 1.0824E+05, 1.0850E+05, 1.0875E+05, 1.0901E+05, 1.0927E+05,
     +  1.0953E+05, 1.0979E+05, 1.1005E+05, 1.1031E+05, 1.1057E+05,
     + 1.1083E+05, 1.1109E+05, 1.1136E+05, 1.1162E+05, 1.1188E+05,
     +  1.1215E+05, 1.1241E+05, 1.1268E+05, 1.1294E+05, 1.1321E+05,
     + 1.1348E+05, 1.1374E+05, 1.1401E+05, 1.1428E+05, 1.1455E+05,
     +  1.1482E+05, 1.1509E+05, 1.1536E+05, 1.1563E+05, 1.1590E+05,
     + 1.1617E+05, 1.1644E+05, 1.1672E+05, 1.1699E+05, 1.1726E+05,
     +  1.1754E+05, 1.1781E+05, 1.1809E+05, 1.1836E+05, 1.1864E+05,
     + 1.1892E+05, 1.1919E+05, 1.1947E+05, 1.1975E+05, 1.2003E+05,
     +  1.2031E+05, 1.2059E+05, 1.2087E+05, 1.2115E+05, 1.2143E+05,
     + 1.2171E+05, 1.2200E+05, 1.2228E+05, 1.2256E+05, 1.2285E+05,
     +  1.2313E+05, 1.2342E+05, 1.2370E+05, 1.2399E+05, 1.2428E+05,
     + 1.2456E+05, 1.2485E+05, 1.2514E+05, 1.2543E+05, 1.2572E+05,
     +  1.2601E+05, 1.2630E+05, 1.2659E+05, 1.2688E+05, 1.2718E+05,
     + 1.2747E+05, 1.2776E+05, 1.2806E+05, 1.2835E+05, 1.2865E+05,
     +  1.2894E+05, 1.2924E+05, 1.2953E+05, 1.2983E+05, 1.3013E+05,
     + 1.3043E+05, 1.3073E+05, 1.3102E+05, 1.3132E+05, 1.3162E+05,
     +  1.3193E+05, 1.3223E+05, 1.3253E+05, 1.3283E+05, 1.3314E+05,
     + 1.3344E+05, 1.3374E+05, 1.3405E+05, 1.3435E+05, 1.3466E+05,
     +  1.3497E+05, 1.3527E+05, 1.3558E+05, 1.3589E+05, 1.3620E+05,
     + 1.3651E+05, 1.3682E+05, 1.3713E+05, 1.3744E+05, 1.3775E+05,
     +  1.3806E+05, 1.3838E+05, 1.3869E+05, 1.3900E+05, 1.3932E+05,
     + 1.3963E+05, 1.3995E+05, 1.4027E+05, 1.4058E+05, 1.4090E+05,
     +  1.4122E+05, 1.4154E+05, 1.4186E+05, 1.4218E+05, 1.4250E+05,
     + 1.4282E+05, 1.4314E+05, 1.4346E+05, 1.4378E+05, 1.4411E+05,
     +  1.4443E+05, 1.4475E+05, 1.4508E+05, 1.4540E+05, 1.4573E+05,
     + 1.4606E+05, 1.4638E+05, 1.4671E+05, 1.4704E+05, 1.4737E+05,
     +  1.4770E+05, 1.4803E+05, 1.4836E+05, 1.4869E+05, 1.4902E+05,
     + 1.4936E+05, 1.4969E+05, 1.5002E+05, 1.5036E+05, 1.5069E+05,
     +  1.5103E+05, 1.5137E+05, 1.5170E+05, 1.5204E+05, 1.5238E+05,
     + 1.5272E+05, 1.5306E+05, 1.5340E+05, 1.5374E+05, 1.5408E+05,
     +  1.5442E+05, 1.5476E+05, 1.5510E+05, 1.5545E+05, 1.5579E+05,
     + 1.5614E+05, 1.5648E+05, 1.5683E+05, 1.5717E+05, 1.5752E+05,
     +  1.5787E+05, 1.5822E+05, 1.5857E+05, 1.5892E+05, 1.5927E+05,
     + 1.5962E+05, 1.5997E+05, 1.6032E+05, 1.6067E+05, 1.6103E+05,
     +  1.6138E+05, 1.6174E+05, 1.6209E+05, 1.6245E+05, 1.6280E+05,
     + 1.6316E+05, 1.6352E+05, 1.6388E+05, 1.6424E+05, 1.6460E+05,
     +  1.6496E+05, 1.6532E+05, 1.6568E+05, 1.6604E+05, 1.6641E+05,
     + 1.6677E+05, 1.6713E+05, 1.6750E+05, 1.6786E+05, 1.6823E+05,
     +  1.6860E+05, 1.6896E+05, 1.6933E+05, 1.6970E+05, 1.7007E+05,
     + 1.7044E+05, 1.7081E+05, 1.7118E+05, 1.7155E+05, 1.7193E+05,
     +  1.7230E+05, 1.7267E+05, 1.7305E+05, 1.7342E+05, 1.7380E+05,
     + 1.7417E+05, 1.7455E+05, 1.7493E+05, 1.7531E+05, 1.7569E+05,
     +  1.7607E+05, 1.7645E+05, 1.7683E+05, 1.7721E+05, 1.7759E+05,
     + 1.7798E+05, 1.7836E+05, 1.7874E+05, 1.7913E+05, 1.7952E+05,
     +  1.7990E+05, 1.8029E+05, 1.8068E+05, 1.8106E+05, 1.8145E+05,
     + 1.8184E+05, 1.8223E+05, 1.8263E+05, 1.8302E+05, 1.8341E+05,
     +  1.8380E+05, 1.8420E+05, 1.8459E+05, 1.8499E+05, 1.8538E+05,
     + 1.8578E+05, 1.8618E+05, 1.8657E+05, 1.8697E+05, 1.8737E+05,
     +  1.8777E+05, 1.8817E+05, 1.8857E+05, 1.8898E+05, 1.8938E+05,
     + 1.8978E+05, 1.9019E+05, 1.9059E+05, 1.9100E+05, 1.9140E+05,
     +  1.9181E+05, 1.9222E+05, 1.9263E+05, 1.9303E+05, 1.9344E+05,
     + 1.9385E+05, 1.9427E+05, 1.9468E+05, 1.9509E+05, 1.9550E+05,
     +  1.9592E+05, 1.9633E+05, 1.9675E+05, 1.9716E+05, 1.9758E+05,
     + 1.9800E+05, 1.9842E+05, 1.9883E+05, 1.9925E+05, 1.9967E+05,
     +  2.0010E+05, 2.0052E+05, 2.0094E+05, 2.0136E+05, 2.0179E+05,
     + 2.0221E+05, 2.0264E+05, 2.0306E+05, 2.0349E+05, 2.0392E+05,
     +  2.0434E+05, 2.0477E+05, 2.0520E+05, 2.0563E+05, 2.0606E+05,
     + 2.0650E+05, 2.0693E+05, 2.0736E+05, 2.0780E+05, 2.0823E+05,
     +  2.0867E+05, 2.0910E+05, 2.0954E+05, 2.0998E+05, 2.1042E+05,
     + 2.1085E+05, 2.1129E+05, 2.1174E+05, 2.1218E+05, 2.1262E+05,
     +  2.1306E+05, 2.1351E+05, 2.1395E+05, 2.1439E+05, 2.1484E+05,
     + 2.1529E+05, 2.1573E+05, 2.1618E+05, 2.1663E+05, 2.1708E+05,
     +  2.1753E+05, 2.1798E+05, 2.1843E+05, 2.1889E+05, 2.1934E+05,
     + 2.1979E+05, 2.2025E+05, 2.2070E+05, 2.2116E+05, 2.2162E+05,
     +  2.2208E+05, 2.2253E+05, 2.2299E+05, 2.2345E+05, 2.2391E+05,
     + 2.2438E+05, 2.2484E+05, 2.2530E+05, 2.2577E+05, 2.2623E+05,
     +  2.2670E+05, 2.2716E+05, 2.2763E+05, 2.2810E+05, 2.2857E+05,
     + 2.2904E+05, 2.2951E+05, 2.2998E+05, 2.3045E+05, 2.3092E+05,
     +  2.3139E+05, 2.3187E+05, 2.3234E+05, 2.3282E+05, 2.3330E+05,
     + 2.3377E+05, 2.3425E+05, 2.3473E+05, 2.3521E+05, 2.3569E+05,
     +  2.3617E+05, 2.3665E+05, 2.3714E+05, 2.3762E+05, 2.3810E+05,
     + 2.3859E+05, 2.3907E+05, 2.3956E+05, 2.4005E+05, 2.4054E+05,
     +  2.4103E+05, 2.4152E+05, 2.4201E+05, 2.4250E+05, 2.4299E+05,
     + 2.4348E+05, 2.4398E+05, 2.4447E+05, 2.4497E+05, 2.4546E+05,
     +  2.4596E+05, 2.4646E+05, 2.4696E+05, 2.4746E+05, 2.4796E+05,
     + 2.4846E+05, 2.4896E+05, 2.4946E+05, 2.4997E+05, 2.5047E+05,
     +  2.5098E+05, 2.5148E+05, 2.5199E+05, 2.5250E+05, 2.5301E+05,
     + 2.5352E+05, 2.5403E+05, 2.5454E+05, 2.5505E+05, 2.5556E+05,
     +  2.5607E+05, 2.5659E+05, 2.5710E+05, 2.5762E+05, 2.5814E+05,
     + 2.5865E+05, 2.5917E+05, 2.5969E+05, 2.6021E+05, 2.6073E+05,
     +  2.6126E+05, 2.6178E+05, 2.6230E+05, 2.6283E+05, 2.6335E+05,
     + 2.6388E+05, 2.6441E+05, 2.6493E+05, 2.6546E+05, 2.6599E+05,
     +  2.6652E+05, 2.6705E+05, 2.6758E+05, 2.6812E+05, 2.6865E+05,
     + 2.6919E+05, 2.6972E+05, 2.7026E+05, 2.7079E+05, 2.7133E+05,
     +  2.7187E+05, 2.7241E+05, 2.7295E+05, 2.7349E+05, 2.7404E+05,
     + 2.7458E+05, 2.7512E+05, 2.7567E+05, 2.7621E+05, 2.7676E+05,
     +  2.7731E+05, 2.7786E+05, 2.7841E+05, 2.7896E+05, 2.7951E+05,
     + 2.8006E+05, 2.8061E+05, 2.8117E+05, 2.8172E+05, 2.8228E+05,
     +  2.8283E+05, 2.8339E+05, 2.8395E+05, 2.8451E+05, 2.8507E+05,
     + 2.8563E+05, 2.8619E+05, 2.8675E+05, 2.8731E+05, 2.8788E+05,
     +  2.8844E+05, 2.8901E+05, 2.8958E+05, 2.9014E+05, 2.9071E+05,
     + 2.9128E+05, 2.9185E+05, 2.9243E+05, 2.9300E+05, 2.9357E+05,
     +  2.9415E+05, 2.9472E+05, 2.9530E+05, 2.9587E+05, 2.9645E+05,
     + 2.9703E+05, 2.9761E+05, 2.9819E+05, 2.9877E+05, 2.9935E+05,
     +  2.9994E+05, 3.0052E+05, 3.0111E+05, 3.0169E+05, 3.0228E+05,
     + 3.0287E+05, 3.0346E+05, 3.0405E+05, 3.0464E+05, 3.0523E+05,
     +  3.0582E+05, 3.0641E+05, 3.0701E+05, 3.0760E+05, 3.0820E+05,
     + 3.0880E+05, 3.0939E+05, 3.0999E+05, 3.1059E+05, 3.1119E+05,
     +  3.1180E+05, 3.1240E+05, 3.1300E+05, 3.1361E+05, 3.1421E+05,
     + 3.1482E+05, 3.1543E+05, 3.1603E+05, 3.1664E+05, 3.1725E+05,
     +  3.1786E+05, 3.1848E+05, 3.1909E+05, 3.1970E+05, 3.2032E+05,
     + 3.2093E+05, 3.2155E+05, 3.2217E+05, 3.2279E+05, 3.2341E+05,
     +  3.2403E+05, 3.2465E+05, 3.2527E+05, 3.2590E+05, 3.2652E+05,
     + 3.2715E+05, 3.2777E+05, 3.2840E+05, 3.2903E+05, 3.2966E+05,
     +  3.3029E+05, 3.3092E+05, 3.3155E+05, 3.3218E+05, 3.3282E+05,
     + 3.3345E+05, 3.3409E+05, 3.3473E+05, 3.3537E+05, 3.3600E+05,
     +  3.3664E+05, 3.3729E+05, 3.3793E+05, 3.3857E+05, 3.3921E+05,
     + 3.3986E+05, 3.4051E+05, 3.4115E+05, 3.4180E+05, 3.4245E+05,
     +  3.4310E+05, 3.4375E+05, 3.4440E+05, 3.4506E+05, 3.4571E+05,
     + 3.4636E+05, 3.4702E+05, 3.4768E+05, 3.4833E+05, 3.4899E+05,
     +  3.4965E+05, 3.5031E+05, 3.5098E+05, 3.5164E+05, 3.5230E+05,
     + 3.5297E+05, 3.5363E+05, 3.5430E+05, 3.5497E+05, 3.5564E+05,
     +  3.5631E+05, 3.5698E+05, 3.5765E+05, 3.5832E+05, 3.5900E+05,
     + 3.5967E+05, 3.6035E+05, 3.6103E+05, 3.6170E+05, 3.6238E+05,
     +  3.6306E+05, 3.6374E+05, 3.6443E+05, 3.6511E+05, 3.6579E+05,
     + 3.6648E+05, 3.6717E+05, 3.6785E+05, 3.6854E+05, 3.6923E+05,
     +  3.6992E+05, 3.7061E+05, 3.7130E+05, 3.7200E+05, 3.7269E+05,
     + 3.7339E+05, 3.7409E+05, 3.7478E+05, 3.7548E+05, 3.7618E+05,
     +  3.7688E+05, 3.7758E+05, 3.7829E+05, 3.7899E+05, 3.7970E+05,
     + 3.8040E+05, 3.8111E+05, 3.8182E+05, 3.8253E+05, 3.8324E+05,
     +  3.8395E+05, 3.8466E+05, 3.8538E+05, 3.8609E+05, 3.8681E+05,
     + 3.8752E+05, 3.8824E+05, 3.8896E+05, 3.8968E+05, 3.9040E+05,
     +  3.9112E+05, 3.9185E+05, 3.9257E+05, 3.9329E+05, 3.9402E+05,
     + 3.9475E+05, 3.9548E+05, 3.9621E+05, 3.9694E+05, 3.9767E+05,
     +  3.9840E+05, 3.9914E+05, 3.9987E+05, 4.0061E+05, 4.0134E+05,
     + 4.0208E+05, 4.0282E+05, 4.0356E+05, 4.0430E+05, 4.0505E+05,
     +  4.0579E+05, 4.0654E+05, 4.0728E+05, 4.0803E+05, 4.0878E+05,
     + 4.0953E+05, 4.1028E+05, 4.1103E+05, 4.1178E+05, 4.1253E+05,
     +  4.1329E+05, 4.1405E+05, 4.1480E+05, 4.1556E+05, 4.1632E+05,
     + 4.1708E+05, 4.1784E+05, 4.1860E+05, 4.1937E+05, 4.2013E+05,
     +  4.2090E+05, 4.2167E+05, 4.2243E+05, 4.2320E+05, 4.2397E+05,
     + 4.2474E+05, 4.2552E+05, 4.2629E+05, 4.2707E+05, 4.2784E+05,
     +  4.2862E+05, 4.2940E+05, 4.3018E+05, 4.3096E+05, 4.3174E+05,
     + 4.3252E+05, 4.3331E+05, 4.3409E+05, 4.3488E+05, 4.3567E+05,
     +  4.3645E+05, 4.3724E+05, 4.3803E+05, 4.3883E+05, 4.3962E+05,
     + 4.4041E+05, 4.4121E+05, 4.4201E+05, 4.4280E+05, 4.4360E+05,
     +  4.4440E+05, 4.4520E+05, 4.4601E+05, 4.4681E+05, 4.4761E+05,
     + 4.4842E+05, 4.4923E+05, 4.5004E+05, 4.5085E+05, 4.5166E+05,
     +  4.5247E+05, 4.5328E+05, 4.5409E+05, 4.5491E+05, 4.5573E+05,
     + 4.5654E+05, 4.5736E+05, 4.5818E+05, 4.5900E+05, 4.5983E+05,
     +  4.6065E+05, 4.6147E+05, 4.6230E+05, 4.6313E+05, 4.6395E+05,
     + 4.6478E+05, 4.6561E+05, 4.6645E+05, 4.6728E+05, 4.6811E+05,
     +  4.6895E+05, 4.6979E+05, 4.7062E+05, 4.7146E+05, 4.7230E+05,
     + 4.7314E+05, 4.7399E+05, 4.7483E+05, 4.7567E+05, 4.7652E+05,
     +  4.7737E+05, 4.7822E+05, 4.7907E+05, 4.7992E+05, 4.8077E+05,
     + 4.8162E+05, 4.8248E+05, 4.8333E+05, 4.8419E+05, 4.8505E+05,
     +  4.8591E+05, 4.8677E+05, 4.8763E+05, 4.8849E+05, 4.8936E+05,
     + 4.9022E+05, 4.9109E+05, 4.9196E+05, 4.9283E+05, 4.9370E+05,
     +  4.9457E+05, 4.9544E+05, 4.9632E+05, 4.9719E+05, 4.9807E+05,
     + 4.9895E+05, 4.9983E+05, 5.0071E+05, 5.0159E+05, 5.0247E+05,
     +  5.0336E+05, 5.0424E+05, 5.0513E+05, 5.0602E+05, 5.0690E+05,
     + 5.0779E+05, 5.0869E+05, 5.0958E+05, 5.1047E+05, 5.1137E+05,
     +  5.1227E+05, 5.1316E+05, 5.1406E+05, 5.1496E+05, 5.1586E+05,
     + 5.1677E+05, 5.1767E+05, 5.1858E+05, 5.1948E+05, 5.2039E+05,
     +  5.2130E+05, 5.2221E+05, 5.2312E+05, 5.2404E+05, 5.2495E+05,
     + 5.2587E+05, 5.2678E+05, 5.2770E+05, 5.2862E+05, 5.2954E+05,
     +  5.3046E+05, 5.3139E+05, 5.3231E+05, 5.3324E+05, 5.3416E+05,
     + 5.3509E+05, 5.3602E+05, 5.3695E+05, 5.3789E+05, 5.3882E+05,
     +  5.3975E+05, 5.4069E+05, 5.4163E+05, 5.4257E+05, 5.4351E+05,
     + 5.4445E+05, 5.4539E+05, 5.4634E+05, 5.4728E+05, 5.4823E+05,
     +  5.4918E+05, 5.5012E+05, 5.5108E+05, 5.5203E+05, 5.5298E+05,
     + 5.5394E+05, 5.5489E+05, 5.5585E+05, 5.5681E+05, 5.5777E+05,
     +  5.5873E+05, 5.5969E+05, 5.6065E+05, 5.6162E+05, 5.6259E+05,
     + 5.6355E+05, 5.6452E+05, 5.6549E+05, 5.6646E+05, 5.6744E+05,
     +  5.6841E+05, 5.6939E+05, 5.7036E+05, 5.7134E+05, 5.7232E+05,
     + 5.7330E+05, 5.7429E+05, 5.7527E+05, 5.7626E+05, 5.7724E+05,
     +  5.7823E+05, 5.7922E+05, 5.8021E+05, 5.8120E+05, 5.8219E+05,
     + 5.8319E+05, 5.8419E+05, 5.8518E+05, 5.8618E+05, 5.8718E+05,
     +  5.8818E+05, 5.8919E+05, 5.9019E+05, 5.9119E+05, 5.9220E+05,
     + 5.9321E+05, 5.9422E+05, 5.9523E+05, 5.9624E+05, 5.9726E+05,
     +  5.9827E+05, 5.9929E+05, 6.0031E+05, 6.0133E+05, 6.0235E+05,
     + 6.0337E+05, 6.0439E+05, 6.0542E+05, 6.0644E+05, 6.0747E+05,
     +  6.0850E+05, 6.0953E+05, 6.1056E+05, 6.1159E+05, 6.1263E+05,
     + 6.1366E+05, 6.1470E+05, 6.1574E+05, 6.1678E+05, 6.1782E+05,
     +  6.1886E+05, 6.1991E+05, 6.2095E+05, 6.2200E+05, 6.2305E+05,
     + 6.2410E+05, 6.2515E+05, 6.2620E+05, 6.2726E+05, 6.2831E+05,
     +  6.2937E+05, 6.3043E+05, 6.3149E+05, 6.3255E+05, 6.3361E+05,
     + 6.3468E+05, 6.3574E+05, 6.3681E+05, 6.3788E+05, 6.3895E+05,
     +  6.4002E+05, 6.4109E+05, 6.4216E+05, 6.4324E+05, 6.4432E+05,
     + 6.4540E+05, 6.4648E+05, 6.4756E+05, 6.4864E+05, 6.4972E+05,
     +  6.5081E+05, 6.5190E+05, 6.5298E+05, 6.5407E+05, 6.5517E+05,
     + 6.5626E+05, 6.5735E+05, 6.5845E+05, 6.5955E+05, 6.6064E+05,
     +  6.6174E+05, 6.6285E+05, 6.6395E+05, 6.6505E+05, 6.6616E+05,
     + 6.6727E+05, 6.6838E+05, 6.6949E+05, 6.7060E+05, 6.7171E+05,
     +  6.7283E+05, 6.7394E+05, 6.7506E+05, 6.7618E+05, 6.7730E+05,
     + 6.7842E+05, 6.7954E+05, 6.8067E+05, 6.8180E+05, 6.8292E+05,
     +  6.8405E+05, 6.8518E+05, 6.8632E+05, 6.8745E+05, 6.8859E+05,
     + 6.8972E+05, 6.9086E+05, 6.9200E+05, 6.9314E+05, 6.9429E+05,
     +  6.9543E+05, 6.9658E+05, 6.9772E+05, 6.9887E+05, 7.0002E+05,
     + 7.0117E+05, 7.0233E+05, 7.0348E+05, 7.0464E+05, 7.0580E+05,
     +  7.0696E+05, 7.0812E+05, 7.0928E+05, 7.1044E+05, 7.1161E+05,
     + 7.1278E+05, 7.1395E+05, 7.1512E+05, 7.1629E+05, 7.1746E+05,
     +  7.1863E+05, 7.1981E+05, 7.2099E+05, 7.2217E+05, 7.2335E+05,
     + 7.2453E+05, 7.2571E+05, 7.2690E+05, 7.2809E+05, 7.2928E+05,
     +  7.3047E+05, 7.3166E+05, 7.3285E+05, 7.3404E+05, 7.3524E+05,
     + 7.3644E+05, 7.3764E+05, 7.3884E+05, 7.4004E+05, 7.4124E+05,
     +  7.4245E+05, 7.4366E+05, 7.4487E+05, 7.4608E+05, 7.4729E+05,
     + 7.4850E+05, 7.4972E+05, 7.5093E+05, 7.5215E+05, 7.5337E+05,
     +  7.5459E+05, 7.5581E+05, 7.5704E+05, 7.5826E+05, 7.5949E+05,
     + 7.6072E+05, 7.6195E+05, 7.6318E+05, 7.6442E+05, 7.6565E+05,
     +  7.6689E+05, 7.6813E+05, 7.6937E+05, 7.7061E+05, 7.7185E+05,
     + 7.7310E+05, 7.7434E+05, 7.7559E+05, 7.7684E+05, 7.7809E+05,
     +  7.7934E+05, 7.8060E+05, 7.8185E+05, 7.8311E+05, 7.8437E+05,
     + 7.8563E+05, 7.8689E+05, 7.8816E+05, 7.8942E+05, 7.9069E+05,
     +  7.9196E+05, 7.9323E+05, 7.9450E+05, 7.9577E+05, 7.9705E+05,
     + 7.9833E+05, 7.9960E+05, 8.0088E+05, 8.0217E+05, 8.0345E+05,
     +  8.0473E+05, 8.0602E+05, 8.0731E+05, 8.0860E+05, 8.0989E+05,
     + 8.1118E+05, 8.1248E+05, 8.1377E+05, 8.1507E+05, 8.1637E+05,
     +  8.1767E+05, 8.1898E+05, 8.2028E+05, 8.2159E+05, 8.2289E+05/
      data (QofT( 7,J),J=2409,4510)/ 8.2420E+05, 8.2552E+05,
     +  8.2683E+05, 8.2814E+05, 8.2946E+05, 8.3078E+05, 8.3210E+05,
     +  8.3342E+05,
     + 8.3474E+05, 8.3606E+05, 8.3739E+05, 8.3872E+05, 8.4005E+05,
     +  8.4138E+05, 8.4271E+05, 8.4404E+05, 8.4538E+05, 8.4672E+05,
     + 8.4806E+05, 8.4940E+05, 8.5074E+05, 8.5209E+05, 8.5343E+05,
     +  8.5478E+05, 8.5613E+05, 8.5748E+05, 8.5883E+05, 8.6019E+05,
     + 8.6154E+05, 8.6290E+05, 8.6426E+05, 8.6562E+05, 8.6698E+05,
     +  8.6835E+05, 8.6971E+05, 8.7108E+05, 8.7245E+05, 8.7382E+05,
     + 8.7520E+05, 8.7657E+05, 8.7795E+05, 8.7933E+05, 8.8071E+05,
     +  8.8209E+05, 8.8347E+05, 8.8486E+05, 8.8624E+05, 8.8763E+05,
     + 8.8902E+05, 8.9041E+05, 8.9181E+05, 8.9320E+05, 8.9460E+05,
     +  8.9600E+05, 8.9740E+05, 8.9880E+05, 9.0020E+05, 9.0161E+05,
     + 9.0302E+05, 9.0443E+05, 9.0584E+05, 9.0725E+05, 9.0866E+05,
     +  9.1008E+05, 9.1150E+05, 9.1292E+05, 9.1434E+05, 9.1576E+05,
     + 9.1719E+05, 9.1861E+05, 9.2004E+05, 9.2147E+05, 9.2290E+05,
     +  9.2434E+05, 9.2577E+05, 9.2721E+05, 9.2865E+05, 9.3009E+05,
     + 9.3153E+05, 9.3297E+05, 9.3442E+05, 9.3587E+05, 9.3732E+05,
     +  9.3877E+05, 9.4022E+05, 9.4167E+05, 9.4313E+05, 9.4459E+05,
     + 9.4605E+05, 9.4751E+05, 9.4897E+05, 9.5044E+05, 9.5191E+05,
     +  9.5337E+05, 9.5484E+05, 9.5632E+05, 9.5779E+05, 9.5927E+05,
     + 9.6074E+05, 9.6222E+05, 9.6371E+05, 9.6519E+05, 9.6667E+05,
     +  9.6816E+05, 9.6965E+05, 9.7114E+05, 9.7263E+05, 9.7412E+05,
     + 9.7562E+05, 9.7712E+05, 9.7862E+05, 9.8012E+05, 9.8162E+05,
     +  9.8313E+05, 9.8463E+05, 9.8614E+05, 9.8765E+05, 9.8916E+05,
     + 9.9068E+05, 9.9219E+05, 9.9371E+05, 9.9523E+05, 9.9675E+05,
     +  9.9827E+05, 9.9980E+05, 1.0013E+06, 1.0029E+06, 1.0044E+06,
     + 1.0059E+06, 1.0074E+06, 1.0090E+06, 1.0105E+06, 1.0121E+06,
     +  1.0136E+06, 1.0151E+06, 1.0167E+06, 1.0182E+06, 1.0198E+06,
     + 1.0213E+06, 1.0229E+06, 1.0244E+06, 1.0260E+06, 1.0276E+06,
     +  1.0291E+06, 1.0307E+06, 1.0322E+06, 1.0338E+06, 1.0354E+06,
     + 1.0369E+06, 1.0385E+06, 1.0401E+06, 1.0417E+06, 1.0432E+06,
     +  1.0448E+06, 1.0464E+06, 1.0480E+06, 1.0496E+06, 1.0511E+06,
     + 1.0527E+06, 1.0543E+06, 1.0559E+06, 1.0575E+06, 1.0591E+06,
     +  1.0607E+06, 1.0623E+06, 1.0639E+06, 1.0655E+06, 1.0671E+06,
     + 1.0687E+06, 1.0703E+06, 1.0719E+06, 1.0735E+06, 1.0752E+06,
     +  1.0768E+06, 1.0784E+06, 1.0800E+06, 1.0816E+06, 1.0833E+06,
     + 1.0849E+06, 1.0865E+06, 1.0881E+06, 1.0898E+06, 1.0914E+06,
     +  1.0930E+06, 1.0947E+06, 1.0963E+06, 1.0979E+06, 1.0996E+06,
     + 1.1012E+06, 1.1029E+06, 1.1045E+06, 1.1062E+06, 1.1078E+06,
     +  1.1095E+06, 1.1111E+06, 1.1128E+06, 1.1145E+06, 1.1161E+06,
     + 1.1178E+06, 1.1195E+06, 1.1211E+06, 1.1228E+06, 1.1245E+06,
     +  1.1261E+06, 1.1278E+06, 1.1295E+06, 1.1312E+06, 1.1328E+06,
     + 1.1345E+06, 1.1362E+06, 1.1379E+06, 1.1396E+06, 1.1413E+06,
     +  1.1430E+06, 1.1447E+06, 1.1464E+06, 1.1481E+06, 1.1498E+06,
     + 1.1515E+06, 1.1532E+06, 1.1549E+06, 1.1566E+06, 1.1583E+06,
     +  1.1600E+06, 1.1617E+06, 1.1634E+06, 1.1652E+06, 1.1669E+06,
     + 1.1686E+06, 1.1703E+06, 1.1721E+06, 1.1738E+06, 1.1755E+06,
     +  1.1772E+06, 1.1790E+06, 1.1807E+06, 1.1825E+06, 1.1842E+06,
     + 1.1859E+06, 1.1877E+06, 1.1894E+06, 1.1912E+06, 1.1929E+06,
     +  1.1947E+06, 1.1964E+06, 1.1982E+06, 1.1999E+06, 1.2017E+06,
     + 1.2035E+06, 1.2052E+06, 1.2070E+06, 1.2088E+06, 1.2105E+06,
     +  1.2123E+06, 1.2141E+06, 1.2159E+06, 1.2176E+06, 1.2194E+06,
     + 1.2212E+06, 1.2230E+06, 1.2248E+06, 1.2266E+06, 1.2284E+06,
     +  1.2301E+06, 1.2319E+06, 1.2337E+06, 1.2355E+06, 1.2373E+06,
     + 1.2391E+06, 1.2409E+06, 1.2427E+06, 1.2446E+06, 1.2464E+06,
     +  1.2482E+06, 1.2500E+06, 1.2518E+06, 1.2536E+06, 1.2555E+06,
     + 1.2573E+06, 1.2591E+06, 1.2609E+06, 1.2628E+06, 1.2646E+06,
     +  1.2664E+06, 1.2683E+06, 1.2701E+06, 1.2719E+06, 1.2738E+06,
     + 1.2756E+06, 1.2775E+06, 1.2793E+06, 1.2812E+06, 1.2830E+06,
     +  1.2849E+06, 1.2867E+06, 1.2886E+06, 1.2904E+06, 1.2923E+06,
     + 1.2942E+06, 1.2960E+06, 1.2979E+06, 1.2998E+06, 1.3016E+06,
     +  1.3035E+06, 1.3054E+06, 1.3073E+06, 1.3092E+06, 1.3110E+06,
     + 1.3129E+06, 1.3148E+06, 1.3167E+06, 1.3186E+06, 1.3205E+06,
     +  1.3224E+06, 1.3243E+06, 1.3262E+06, 1.3281E+06, 1.3300E+06,
     + 1.3319E+06, 1.3338E+06, 1.3357E+06, 1.3376E+06, 1.3395E+06,
     +  1.3415E+06, 1.3434E+06, 1.3453E+06, 1.3472E+06, 1.3492E+06,
     + 1.3511E+06, 1.3530E+06, 1.3549E+06, 1.3569E+06, 1.3588E+06,
     +  1.3607E+06, 1.3627E+06, 1.3646E+06, 1.3666E+06, 1.3685E+06,
     + 1.3705E+06, 1.3724E+06, 1.3744E+06, 1.3763E+06, 1.3783E+06,
     +  1.3802E+06, 1.3822E+06, 1.3842E+06, 1.3861E+06, 1.3881E+06,
     + 1.3901E+06, 1.3920E+06, 1.3940E+06, 1.3960E+06, 1.3980E+06,
     +  1.4000E+06, 1.4019E+06, 1.4039E+06, 1.4059E+06, 1.4079E+06,
     + 1.4099E+06, 1.4119E+06, 1.4139E+06, 1.4159E+06, 1.4179E+06,
     +  1.4199E+06, 1.4219E+06, 1.4239E+06, 1.4259E+06, 1.4279E+06,
     + 1.4299E+06, 1.4320E+06, 1.4340E+06, 1.4360E+06, 1.4380E+06,
     +  1.4400E+06, 1.4421E+06, 1.4441E+06, 1.4461E+06, 1.4482E+06,
     + 1.4502E+06, 1.4522E+06, 1.4543E+06, 1.4563E+06, 1.4584E+06,
     +  1.4604E+06, 1.4625E+06, 1.4645E+06, 1.4666E+06, 1.4686E+06,
     + 1.4707E+06, 1.4727E+06, 1.4748E+06, 1.4769E+06, 1.4789E+06,
     +  1.4810E+06, 1.4831E+06, 1.4851E+06, 1.4872E+06, 1.4893E+06,
     + 1.4914E+06, 1.4934E+06, 1.4955E+06, 1.4976E+06, 1.4997E+06,
     +  1.5018E+06, 1.5039E+06, 1.5060E+06, 1.5081E+06, 1.5102E+06,
     + 1.5123E+06, 1.5144E+06, 1.5165E+06, 1.5186E+06, 1.5207E+06,
     +  1.5228E+06, 1.5249E+06, 1.5271E+06, 1.5292E+06, 1.5313E+06,
     + 1.5334E+06, 1.5356E+06, 1.5377E+06, 1.5398E+06, 1.5420E+06,
     +  1.5441E+06, 1.5462E+06, 1.5484E+06, 1.5505E+06, 1.5527E+06,
     + 1.5548E+06, 1.5570E+06, 1.5591E+06, 1.5613E+06, 1.5634E+06,
     +  1.5656E+06, 1.5677E+06, 1.5699E+06, 1.5721E+06, 1.5742E+06,
     + 1.5764E+06, 1.5786E+06, 1.5807E+06, 1.5829E+06, 1.5851E+06,
     +  1.5873E+06, 1.5895E+06, 1.5917E+06, 1.5938E+06, 1.5960E+06,
     + 1.5982E+06, 1.6004E+06, 1.6026E+06, 1.6048E+06, 1.6070E+06,
     +  1.6092E+06, 1.6114E+06, 1.6136E+06, 1.6159E+06, 1.6181E+06,
     + 1.6203E+06, 1.6225E+06, 1.6247E+06, 1.6269E+06, 1.6292E+06,
     +  1.6314E+06, 1.6336E+06, 1.6359E+06, 1.6381E+06, 1.6403E+06,
     + 1.6426E+06, 1.6448E+06, 1.6471E+06, 1.6493E+06, 1.6515E+06,
     +  1.6538E+06, 1.6561E+06, 1.6583E+06, 1.6606E+06, 1.6628E+06,
     + 1.6651E+06, 1.6674E+06, 1.6696E+06, 1.6719E+06, 1.6742E+06,
     +  1.6764E+06, 1.6787E+06, 1.6810E+06, 1.6833E+06, 1.6856E+06,
     + 1.6878E+06, 1.6901E+06, 1.6924E+06, 1.6947E+06, 1.6970E+06,
     +  1.6993E+06, 1.7016E+06, 1.7039E+06, 1.7062E+06, 1.7085E+06,
     + 1.7108E+06, 1.7131E+06, 1.7155E+06, 1.7178E+06, 1.7201E+06,
     +  1.7224E+06, 1.7247E+06, 1.7271E+06, 1.7294E+06, 1.7317E+06,
     + 1.7341E+06, 1.7364E+06, 1.7387E+06, 1.7411E+06, 1.7434E+06,
     +  1.7458E+06, 1.7481E+06, 1.7505E+06, 1.7528E+06, 1.7552E+06,
     + 1.7575E+06, 1.7599E+06, 1.7622E+06, 1.7646E+06, 1.7670E+06,
     +  1.7694E+06, 1.7717E+06, 1.7741E+06, 1.7765E+06, 1.7789E+06,
     + 1.7812E+06, 1.7836E+06, 1.7860E+06, 1.7884E+06, 1.7908E+06,
     +  1.7932E+06, 1.7956E+06, 1.7980E+06, 1.8004E+06, 1.8028E+06,
     + 1.8052E+06, 1.8076E+06, 1.8100E+06, 1.8124E+06, 1.8148E+06,
     +  1.8172E+06, 1.8197E+06, 1.8221E+06, 1.8245E+06, 1.8269E+06,
     + 1.8294E+06, 1.8318E+06, 1.8342E+06, 1.8367E+06, 1.8391E+06,
     +  1.8416E+06, 1.8440E+06, 1.8464E+06, 1.8489E+06, 1.8514E+06,
     + 1.8538E+06, 1.8563E+06, 1.8587E+06, 1.8612E+06, 1.8636E+06,
     +  1.8661E+06, 1.8686E+06, 1.8711E+06, 1.8735E+06, 1.8760E+06,
     + 1.8785E+06, 1.8810E+06, 1.8835E+06, 1.8859E+06, 1.8884E+06,
     +  1.8909E+06, 1.8934E+06, 1.8959E+06, 1.8984E+06, 1.9009E+06,
     + 1.9034E+06, 1.9059E+06, 1.9084E+06, 1.9109E+06, 1.9135E+06,
     +  1.9160E+06, 1.9185E+06, 1.9210E+06, 1.9235E+06, 1.9261E+06,
     + 1.9286E+06, 1.9311E+06, 1.9337E+06, 1.9362E+06, 1.9387E+06,
     +  1.9413E+06, 1.9438E+06, 1.9464E+06, 1.9489E+06, 1.9515E+06,
     + 1.9540E+06, 1.9566E+06, 1.9591E+06, 1.9617E+06, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       282
      data (QofT( 8,J),J=1,2408)/ 6.0000E+00, 6.0015E+00, 6.0289E+00,
     +  6.1338E+00, 6.3529E+00, 6.6988E+00, 7.1686E+00, 7.7525E+00,
     + 8.4392E+00, 9.2175E+00, 1.0078E+01, 1.1012E+01, 1.2012E+01,
     +  1.3073E+01, 1.4191E+01, 1.5359E+01, 1.6576E+01, 1.7839E+01,
     + 1.9144E+01, 2.0489E+01, 2.3295E+01, 2.6242E+01, 2.9322E+01,
     +  3.2526E+01, 3.5848E+01, 3.9283E+01, 4.2825E+01, 4.6471E+01,
     + 5.0218E+01, 5.4063E+01, 5.8002E+01, 6.2035E+01, 6.6157E+01,
     +  7.0368E+01, 7.4666E+01, 7.9048E+01, 8.3514E+01, 8.8062E+01,
     + 9.2689E+01, 9.7396E+01, 1.0218E+02, 1.0704E+02, 1.1198E+02,
     +  1.1699E+02, 1.2207E+02, 1.2722E+02, 1.3245E+02, 1.3775E+02,
     + 1.4311E+02, 1.4855E+02, 1.5405E+02, 1.5962E+02, 1.6525E+02,
     +  1.7095E+02, 1.7671E+02, 1.8254E+02, 1.8843E+02, 1.9438E+02,
     + 2.0040E+02, 2.0647E+02, 2.1261E+02, 2.1881E+02, 2.2507E+02,
     +  2.3138E+02, 2.3775E+02, 2.4419E+02, 2.5068E+02, 2.5722E+02,
     + 2.6382E+02, 2.7048E+02, 2.7720E+02, 2.8397E+02, 2.9079E+02,
     +  2.9767E+02, 3.0460E+02, 3.1159E+02, 3.1863E+02, 3.2572E+02,
     + 3.3286E+02, 3.4006E+02, 3.4730E+02, 3.5460E+02, 3.6195E+02,
     +  3.6935E+02, 3.7680E+02, 3.8431E+02, 3.9186E+02, 3.9946E+02,
     + 4.0710E+02, 4.1480E+02, 4.2255E+02, 4.3034E+02, 4.3819E+02,
     +  4.4608E+02, 4.5402E+02, 4.6200E+02, 4.7003E+02, 4.7811E+02,
     + 4.8624E+02, 4.9441E+02, 5.0263E+02, 5.1090E+02, 5.1921E+02,
     +  5.2757E+02, 5.3597E+02, 5.4441E+02, 5.5291E+02, 5.6144E+02,
     + 5.7003E+02, 5.7865E+02, 5.8732E+02, 5.9604E+02, 6.0480E+02,
     +  6.1360E+02, 6.2245E+02, 6.3134E+02, 6.4028E+02, 6.4926E+02,
     + 6.5828E+02, 6.6735E+02, 6.7646E+02, 6.8561E+02, 6.9481E+02,
     +  7.0404E+02, 7.1333E+02, 7.2265E+02, 7.3202E+02, 7.4143E+02,
     + 7.5088E+02, 7.6037E+02, 7.6991E+02, 7.7949E+02, 7.8911E+02,
     +  7.9878E+02, 8.0849E+02, 8.1824E+02, 8.2803E+02, 8.3786E+02,
     + 8.4774E+02, 8.5766E+02, 8.6762E+02, 8.7762E+02, 8.8767E+02,
     +  8.9775E+02, 9.0788E+02, 9.1805E+02, 9.2827E+02, 9.3852E+02,
     + 9.4882E+02, 9.5916E+02, 9.6954E+02, 9.7997E+02, 9.9043E+02,
     +  1.0009E+03, 1.0115E+03, 1.0221E+03, 1.0327E+03, 1.0434E+03,
     + 1.0541E+03, 1.0649E+03, 1.0757E+03, 1.0865E+03, 1.0974E+03,
     +  1.1084E+03, 1.1193E+03, 1.1303E+03, 1.1414E+03, 1.1525E+03,
     + 1.1637E+03, 1.1748E+03, 1.1861E+03, 1.1973E+03, 1.2087E+03,
     +  1.2200E+03, 1.2314E+03, 1.2429E+03, 1.2544E+03, 1.2659E+03,
     + 1.2775E+03, 1.2891E+03, 1.3007E+03, 1.3125E+03, 1.3242E+03,
     +  1.3360E+03, 1.3478E+03, 1.3597E+03, 1.3716E+03, 1.3836E+03,
     + 1.3956E+03, 1.4077E+03, 1.4198E+03, 1.4319E+03, 1.4441E+03,
     +  1.4563E+03, 1.4686E+03, 1.4810E+03, 1.4933E+03, 1.5057E+03,
     + 1.5182E+03, 1.5307E+03, 1.5432E+03, 1.5558E+03, 1.5685E+03,
     +  1.5812E+03, 1.5939E+03, 1.6067E+03, 1.6195E+03, 1.6324E+03,
     + 1.6453E+03, 1.6582E+03, 1.6712E+03, 1.6843E+03, 1.6974E+03,
     +  1.7105E+03, 1.7237E+03, 1.7369E+03, 1.7502E+03, 1.7636E+03,
     + 1.7769E+03, 1.7904E+03, 1.8038E+03, 1.8173E+03, 1.8309E+03,
     +  1.8445E+03, 1.8582E+03, 1.8719E+03, 1.8857E+03, 1.8995E+03,
     + 1.9133E+03, 1.9272E+03, 1.9412E+03, 1.9552E+03, 1.9692E+03,
     +  1.9833E+03, 1.9974E+03, 2.0116E+03, 2.0259E+03, 2.0402E+03,
     + 2.0545E+03, 2.0689E+03, 2.0833E+03, 2.0978E+03, 2.1124E+03,
     +  2.1269E+03, 2.1416E+03, 2.1563E+03, 2.1710E+03, 2.1858E+03,
     + 2.2006E+03, 2.2155E+03, 2.2305E+03, 2.2455E+03, 2.2605E+03,
     +  2.2756E+03, 2.2907E+03, 2.3059E+03, 2.3212E+03, 2.3365E+03,
     + 2.3518E+03, 2.3672E+03, 2.3827E+03, 2.3982E+03, 2.4138E+03,
     +  2.4294E+03, 2.4450E+03, 2.4608E+03, 2.4765E+03, 2.4924E+03,
     + 2.5082E+03, 2.5242E+03, 2.5401E+03, 2.5562E+03, 2.5723E+03,
     +  2.5884E+03, 2.6046E+03, 2.6209E+03, 2.6372E+03, 2.6536E+03,
     + 2.6700E+03, 2.6864E+03, 2.7030E+03, 2.7196E+03, 2.7362E+03,
     +  2.7529E+03, 2.7696E+03, 2.7864E+03, 2.8033E+03, 2.8202E+03,
     + 2.8372E+03, 2.8542E+03, 2.8713E+03, 2.8884E+03, 2.9056E+03,
     +  2.9229E+03, 2.9402E+03, 2.9576E+03, 2.9750E+03, 2.9925E+03,
     + 3.0100E+03, 3.0276E+03, 3.0453E+03, 3.0630E+03, 3.0808E+03,
     +  3.0986E+03, 3.1165E+03, 3.1345E+03, 3.1525E+03, 3.1706E+03,
     + 3.1887E+03, 3.2069E+03, 3.2251E+03, 3.2434E+03, 3.2618E+03,
     +  3.2802E+03, 3.2987E+03, 3.3173E+03, 3.3359E+03, 3.3546E+03,
     + 3.3733E+03, 3.3921E+03, 3.4110E+03, 3.4299E+03, 3.4488E+03,
     +  3.4679E+03, 3.4870E+03, 3.5062E+03, 3.5254E+03, 3.5447E+03,
     + 3.5640E+03, 3.5834E+03, 3.6029E+03, 3.6225E+03, 3.6421E+03,
     +  3.6618E+03, 3.6815E+03, 3.7013E+03, 3.7211E+03, 3.7411E+03,
     + 3.7611E+03, 3.7811E+03, 3.8012E+03, 3.8214E+03, 3.8417E+03,
     +  3.8620E+03, 3.8824E+03, 3.9028E+03, 3.9234E+03, 3.9439E+03,
     + 3.9646E+03, 3.9853E+03, 4.0061E+03, 4.0269E+03, 4.0478E+03,
     +  4.0688E+03, 4.0899E+03, 4.1110E+03, 4.1322E+03, 4.1534E+03,
     + 4.1748E+03, 4.1962E+03, 4.2176E+03, 4.2391E+03, 4.2607E+03,
     +  4.2824E+03, 4.3042E+03, 4.3260E+03, 4.3478E+03, 4.3698E+03,
     + 4.3918E+03, 4.4139E+03, 4.4361E+03, 4.4583E+03, 4.4806E+03,
     +  4.5030E+03, 4.5254E+03, 4.5479E+03, 4.5705E+03, 4.5932E+03,
     + 4.6159E+03, 4.6387E+03, 4.6616E+03, 4.6845E+03, 4.7076E+03,
     +  4.7307E+03, 4.7538E+03, 4.7771E+03, 4.8004E+03, 4.8238E+03,
     + 4.8472E+03, 4.8708E+03, 4.8944E+03, 4.9181E+03, 4.9419E+03,
     +  4.9657E+03, 4.9896E+03, 5.0136E+03, 5.0377E+03, 5.0618E+03,
     + 5.0860E+03, 5.1103E+03, 5.1347E+03, 5.1592E+03, 5.1837E+03,
     +  5.2083E+03, 5.2330E+03, 5.2578E+03, 5.2826E+03, 5.3075E+03,
     + 5.3325E+03, 5.3576E+03, 5.3828E+03, 5.4080E+03, 5.4333E+03,
     +  5.4587E+03, 5.4842E+03, 5.5097E+03, 5.5354E+03, 5.5611E+03,
     + 5.5869E+03, 5.6128E+03, 5.6387E+03, 5.6648E+03, 5.6909E+03,
     +  5.7171E+03, 5.7434E+03, 5.7697E+03, 5.7962E+03, 5.8227E+03,
     + 5.8493E+03, 5.8760E+03, 5.9028E+03, 5.9297E+03, 5.9566E+03,
     +  5.9837E+03, 6.0108E+03, 6.0380E+03, 6.0653E+03, 6.0927E+03,
     + 6.1201E+03, 6.1477E+03, 6.1753E+03, 6.2030E+03, 6.2308E+03,
     +  6.2587E+03, 6.2867E+03, 6.3147E+03, 6.3429E+03, 6.3711E+03,
     + 6.3994E+03, 6.4279E+03, 6.4564E+03, 6.4850E+03, 6.5136E+03,
     +  6.5424E+03, 6.5713E+03, 6.6002E+03, 6.6292E+03, 6.6584E+03,
     + 6.6876E+03, 6.7169E+03, 6.7463E+03, 6.7758E+03, 6.8053E+03,
     +  6.8350E+03, 6.8648E+03, 6.8946E+03, 6.9246E+03, 6.9546E+03,
     + 6.9847E+03, 7.0149E+03, 7.0452E+03, 7.0757E+03, 7.1062E+03,
     +  7.1367E+03, 7.1674E+03, 7.1982E+03, 7.2291E+03, 7.2601E+03,
     + 7.2911E+03, 7.3223E+03, 7.3535E+03, 7.3849E+03, 7.4163E+03,
     +  7.4479E+03, 7.4795E+03, 7.5112E+03, 7.5431E+03, 7.5750E+03,
     + 7.6070E+03, 7.6392E+03, 7.6714E+03, 7.7037E+03, 7.7361E+03,
     +  7.7686E+03, 7.8012E+03, 7.8340E+03, 7.8668E+03, 7.8997E+03,
     + 7.9327E+03, 7.9658E+03, 7.9990E+03, 8.0323E+03, 8.0658E+03,
     +  8.0993E+03, 8.1329E+03, 8.1666E+03, 8.2004E+03, 8.2344E+03,
     + 8.2684E+03, 8.3025E+03, 8.3367E+03, 8.3711E+03, 8.4055E+03,
     +  8.4401E+03, 8.4747E+03, 8.5095E+03, 8.5443E+03, 8.5793E+03,
     + 8.6143E+03, 8.6495E+03, 8.6848E+03, 8.7202E+03, 8.7556E+03,
     +  8.7912E+03, 8.8269E+03, 8.8627E+03, 8.8987E+03, 8.9347E+03,
     + 8.9708E+03, 9.0070E+03, 9.0434E+03, 9.0798E+03, 9.1164E+03,
     +  9.1531E+03, 9.1899E+03, 9.2267E+03, 9.2637E+03, 9.3009E+03,
     + 9.3381E+03, 9.3754E+03, 9.4128E+03, 9.4504E+03, 9.4881E+03,
     +  9.5258E+03, 9.5637E+03, 9.6017E+03, 9.6398E+03, 9.6781E+03,
     + 9.7164E+03, 9.7549E+03, 9.7934E+03, 9.8321E+03, 9.8709E+03,
     +  9.9098E+03, 9.9488E+03, 9.9880E+03, 1.0027E+04, 1.0067E+04,
     + 1.0106E+04, 1.0146E+04, 1.0185E+04, 1.0225E+04, 1.0265E+04,
     +  1.0305E+04, 1.0345E+04, 1.0386E+04, 1.0426E+04, 1.0467E+04,
     + 1.0507E+04, 1.0548E+04, 1.0589E+04, 1.0630E+04, 1.0671E+04,
     +  1.0712E+04, 1.0754E+04, 1.0795E+04, 1.0837E+04, 1.0879E+04,
     + 1.0921E+04, 1.0963E+04, 1.1005E+04, 1.1047E+04, 1.1089E+04,
     +  1.1132E+04, 1.1174E+04, 1.1217E+04, 1.1260E+04, 1.1303E+04,
     + 1.1346E+04, 1.1389E+04, 1.1432E+04, 1.1476E+04, 1.1520E+04,
     +  1.1563E+04, 1.1607E+04, 1.1651E+04, 1.1695E+04, 1.1739E+04,
     + 1.1784E+04, 1.1828E+04, 1.1873E+04, 1.1918E+04, 1.1962E+04,
     +  1.2007E+04, 1.2053E+04, 1.2098E+04, 1.2143E+04, 1.2189E+04,
     + 1.2234E+04, 1.2280E+04, 1.2326E+04, 1.2372E+04, 1.2418E+04,
     +  1.2465E+04, 1.2511E+04, 1.2558E+04, 1.2604E+04, 1.2651E+04,
     + 1.2698E+04, 1.2745E+04, 1.2792E+04, 1.2840E+04, 1.2887E+04,
     +  1.2935E+04, 1.2983E+04, 1.3031E+04, 1.3079E+04, 1.3127E+04,
     + 1.3175E+04, 1.3224E+04, 1.3272E+04, 1.3321E+04, 1.3370E+04,
     +  1.3419E+04, 1.3468E+04, 1.3517E+04, 1.3567E+04, 1.3616E+04,
     + 1.3666E+04, 1.3716E+04, 1.3766E+04, 1.3816E+04, 1.3866E+04,
     +  1.3916E+04, 1.3967E+04, 1.4018E+04, 1.4068E+04, 1.4119E+04,
     + 1.4171E+04, 1.4222E+04, 1.4273E+04, 1.4325E+04, 1.4376E+04,
     +  1.4428E+04, 1.4480E+04, 1.4532E+04, 1.4585E+04, 1.4637E+04,
     + 1.4690E+04, 1.4742E+04, 1.4795E+04, 1.4848E+04, 1.4901E+04,
     +  1.4955E+04, 1.5008E+04, 1.5062E+04, 1.5115E+04, 1.5169E+04,
     + 1.5223E+04, 1.5277E+04, 1.5332E+04, 1.5386E+04, 1.5441E+04,
     +  1.5496E+04, 1.5551E+04, 1.5606E+04, 1.5661E+04, 1.5716E+04,
     + 1.5772E+04, 1.5828E+04, 1.5883E+04, 1.5939E+04, 1.5996E+04,
     +  1.6052E+04, 1.6108E+04, 1.6165E+04, 1.6222E+04, 1.6279E+04,
     + 1.6336E+04, 1.6393E+04, 1.6450E+04, 1.6508E+04, 1.6566E+04,
     +  1.6624E+04, 1.6682E+04, 1.6740E+04, 1.6798E+04, 1.6857E+04,
     + 1.6915E+04, 1.6974E+04, 1.7033E+04, 1.7092E+04, 1.7152E+04,
     +  1.7211E+04, 1.7271E+04, 1.7330E+04, 1.7390E+04, 1.7450E+04,
     + 1.7511E+04, 1.7571E+04, 1.7632E+04, 1.7693E+04, 1.7753E+04,
     +  1.7815E+04, 1.7876E+04, 1.7937E+04, 1.7999E+04, 1.8061E+04,
     + 1.8123E+04, 1.8185E+04, 1.8247E+04, 1.8309E+04, 1.8372E+04,
     +  1.8435E+04, 1.8498E+04, 1.8561E+04, 1.8624E+04, 1.8687E+04,
     + 1.8751E+04, 1.8815E+04, 1.8879E+04, 1.8943E+04, 1.9007E+04,
     +  1.9071E+04, 1.9136E+04, 1.9201E+04, 1.9266E+04, 1.9331E+04,
     + 1.9396E+04, 1.9462E+04, 1.9528E+04, 1.9593E+04, 1.9659E+04,
     +  1.9726E+04, 1.9792E+04, 1.9858E+04, 1.9925E+04, 1.9992E+04,
     + 2.0059E+04, 2.0126E+04, 2.0194E+04, 2.0261E+04, 2.0329E+04,
     +  2.0397E+04, 2.0465E+04, 2.0534E+04, 2.0602E+04, 2.0671E+04,
     + 2.0740E+04, 2.0809E+04, 2.0878E+04, 2.0947E+04, 2.1017E+04,
     +  2.1087E+04, 2.1157E+04, 2.1227E+04, 2.1297E+04, 2.1368E+04,
     + 2.1438E+04, 2.1509E+04, 2.1580E+04, 2.1651E+04, 2.1723E+04,
     +  2.1794E+04, 2.1866E+04, 2.1938E+04, 2.2010E+04, 2.2083E+04,
     + 2.2155E+04, 2.2228E+04, 2.2301E+04, 2.2374E+04, 2.2447E+04,
     +  2.2521E+04, 2.2595E+04, 2.2668E+04, 2.2742E+04, 2.2817E+04,
     + 2.2891E+04, 2.2966E+04, 2.3041E+04, 2.3116E+04, 2.3191E+04,
     +  2.3266E+04, 2.3342E+04, 2.3418E+04, 2.3494E+04, 2.3570E+04,
     + 2.3646E+04, 2.3723E+04, 2.3800E+04, 2.3877E+04, 2.3954E+04,
     +  2.4031E+04, 2.4109E+04, 2.4186E+04, 2.4264E+04, 2.4343E+04,
     + 2.4421E+04, 2.4500E+04, 2.4578E+04, 2.4657E+04, 2.4736E+04,
     +  2.4816E+04, 2.4895E+04, 2.4975E+04, 2.5055E+04, 2.5135E+04,
     + 2.5216E+04, 2.5296E+04, 2.5377E+04, 2.5458E+04, 2.5539E+04,
     +  2.5621E+04, 2.5702E+04, 2.5784E+04, 2.5866E+04, 2.5948E+04,
     + 2.6031E+04, 2.6113E+04, 2.6196E+04, 2.6279E+04, 2.6363E+04,
     +  2.6446E+04, 2.6530E+04, 2.6614E+04, 2.6698E+04, 2.6782E+04,
     + 2.6867E+04, 2.6951E+04, 2.7036E+04, 2.7122E+04, 2.7207E+04,
     +  2.7293E+04, 2.7378E+04, 2.7464E+04, 2.7551E+04, 2.7637E+04,
     + 2.7724E+04, 2.7811E+04, 2.7898E+04, 2.7985E+04, 2.8073E+04,
     +  2.8160E+04, 2.8248E+04, 2.8337E+04, 2.8425E+04, 2.8514E+04,
     + 2.8603E+04, 2.8692E+04, 2.8781E+04, 2.8870E+04, 2.8960E+04,
     +  2.9050E+04, 2.9140E+04, 2.9231E+04, 2.9321E+04, 2.9412E+04,
     + 2.9503E+04, 2.9595E+04, 2.9686E+04, 2.9778E+04, 2.9870E+04,
     +  2.9962E+04, 3.0054E+04, 3.0147E+04, 3.0240E+04, 3.0333E+04,
     + 3.0427E+04, 3.0520E+04, 3.0614E+04, 3.0708E+04, 3.0802E+04,
     +  3.0897E+04, 3.0991E+04, 3.1086E+04, 3.1182E+04, 3.1277E+04,
     + 3.1373E+04, 3.1468E+04, 3.1565E+04, 3.1661E+04, 3.1758E+04,
     +  3.1854E+04, 3.1951E+04, 3.2049E+04, 3.2146E+04, 3.2244E+04,
     + 3.2342E+04, 3.2440E+04, 3.2539E+04, 3.2637E+04, 3.2736E+04,
     +  3.2836E+04, 3.2935E+04, 3.3035E+04, 3.3135E+04, 3.3235E+04,
     + 3.3335E+04, 3.3436E+04, 3.3537E+04, 3.3638E+04, 3.3739E+04,
     +  3.3841E+04, 3.3943E+04, 3.4045E+04, 3.4147E+04, 3.4250E+04,
     + 3.4353E+04, 3.4456E+04, 3.4559E+04, 3.4663E+04, 3.4766E+04,
     +  3.4870E+04, 3.4975E+04, 3.5079E+04, 3.5184E+04, 3.5289E+04,
     + 3.5395E+04, 3.5500E+04, 3.5606E+04, 3.5712E+04, 3.5818E+04,
     +  3.5925E+04, 3.6032E+04, 3.6139E+04, 3.6246E+04, 3.6354E+04,
     + 3.6462E+04, 3.6570E+04, 3.6678E+04, 3.6787E+04, 3.6896E+04,
     +  3.7005E+04, 3.7114E+04, 3.7224E+04, 3.7334E+04, 3.7444E+04,
     + 3.7554E+04, 3.7665E+04, 3.7776E+04, 3.7887E+04, 3.7999E+04,
     +  3.8110E+04, 3.8222E+04, 3.8335E+04, 3.8447E+04, 3.8560E+04,
     + 3.8673E+04, 3.8786E+04, 3.8900E+04, 3.9014E+04, 3.9128E+04,
     +  3.9242E+04, 3.9357E+04, 3.9472E+04, 3.9587E+04, 3.9703E+04,
     + 3.9818E+04, 3.9934E+04, 4.0051E+04, 4.0167E+04, 4.0284E+04,
     +  4.0401E+04, 4.0518E+04, 4.0636E+04, 4.0754E+04, 4.0872E+04,
     + 4.0991E+04, 4.1109E+04, 4.1228E+04, 4.1348E+04, 4.1467E+04,
     +  4.1587E+04, 4.1707E+04, 4.1827E+04, 4.1948E+04, 4.2069E+04,
     + 4.2190E+04, 4.2312E+04, 4.2433E+04, 4.2556E+04, 4.2678E+04,
     +  4.2800E+04, 4.2923E+04, 4.3047E+04, 4.3170E+04, 4.3294E+04,
     + 4.3418E+04, 4.3542E+04, 4.3667E+04, 4.3792E+04, 4.3917E+04,
     +  4.4042E+04, 4.4168E+04, 4.4294E+04, 4.4420E+04, 4.4547E+04,
     + 4.4674E+04, 4.4801E+04, 4.4929E+04, 4.5056E+04, 4.5184E+04,
     +  4.5313E+04, 4.5441E+04, 4.5570E+04, 4.5700E+04, 4.5829E+04,
     + 4.5959E+04, 4.6089E+04, 4.6219E+04, 4.6350E+04, 4.6481E+04,
     +  4.6612E+04, 4.6744E+04, 4.6876E+04, 4.7008E+04, 4.7141E+04,
     + 4.7273E+04, 4.7406E+04, 4.7540E+04, 4.7674E+04, 4.7808E+04,
     +  4.7942E+04, 4.8076E+04, 4.8211E+04, 4.8346E+04, 4.8482E+04,
     + 4.8618E+04, 4.8754E+04, 4.8890E+04, 4.9027E+04, 4.9164E+04,
     +  4.9301E+04, 4.9439E+04, 4.9577E+04, 4.9715E+04, 4.9854E+04,
     + 4.9993E+04, 5.0132E+04, 5.0271E+04, 5.0411E+04, 5.0551E+04,
     +  5.0692E+04, 5.0833E+04, 5.0974E+04, 5.1115E+04, 5.1257E+04,
     + 5.1399E+04, 5.1541E+04, 5.1684E+04, 5.1827E+04, 5.1970E+04,
     +  5.2113E+04, 5.2257E+04, 5.2402E+04, 5.2546E+04, 5.2691E+04,
     + 5.2836E+04, 5.2982E+04, 5.3127E+04, 5.3274E+04, 5.3420E+04,
     +  5.3567E+04, 5.3714E+04, 5.3861E+04, 5.4009E+04, 5.4157E+04,
     + 5.4306E+04, 5.4454E+04, 5.4603E+04, 5.4753E+04, 5.4903E+04,
     +  5.5053E+04, 5.5203E+04, 5.5354E+04, 5.5505E+04, 5.5656E+04,
     + 5.5808E+04, 5.5960E+04, 5.6112E+04, 5.6265E+04, 5.6418E+04,
     +  5.6571E+04, 5.6725E+04, 5.6879E+04, 5.7033E+04, 5.7188E+04,
     + 5.7343E+04, 5.7499E+04, 5.7654E+04, 5.7810E+04, 5.7967E+04,
     +  5.8123E+04, 5.8281E+04, 5.8438E+04, 5.8596E+04, 5.8754E+04,
     + 5.8912E+04, 5.9071E+04, 5.9230E+04, 5.9390E+04, 5.9549E+04,
     +  5.9710E+04, 5.9870E+04, 6.0031E+04, 6.0192E+04, 6.0354E+04,
     + 6.0516E+04, 6.0678E+04, 6.0840E+04, 6.1003E+04, 6.1167E+04,
     +  6.1330E+04, 6.1494E+04, 6.1659E+04, 6.1823E+04, 6.1988E+04,
     + 6.2154E+04, 6.2320E+04, 6.2486E+04, 6.2652E+04, 6.2819E+04,
     +  6.2986E+04, 6.3154E+04, 6.3322E+04, 6.3490E+04, 6.3658E+04,
     + 6.3827E+04, 6.3997E+04, 6.4166E+04, 6.4336E+04, 6.4507E+04,
     +  6.4678E+04, 6.4849E+04, 6.5020E+04, 6.5192E+04, 6.5364E+04,
     + 6.5537E+04, 6.5710E+04, 6.5883E+04, 6.6057E+04, 6.6231E+04,
     +  6.6406E+04, 6.6580E+04, 6.6756E+04, 6.6931E+04, 6.7107E+04,
     + 6.7283E+04, 6.7460E+04, 6.7637E+04, 6.7814E+04, 6.7992E+04,
     +  6.8170E+04, 6.8349E+04, 6.8528E+04, 6.8707E+04, 6.8887E+04,
     + 6.9067E+04, 6.9247E+04, 6.9428E+04, 6.9609E+04, 6.9791E+04,
     +  6.9973E+04, 7.0155E+04, 7.0338E+04, 7.0521E+04, 7.0704E+04,
     + 7.0888E+04, 7.1072E+04, 7.1257E+04, 7.1442E+04, 7.1627E+04,
     +  7.1813E+04, 7.1999E+04, 7.2186E+04, 7.2372E+04, 7.2560E+04,
     + 7.2747E+04, 7.2936E+04, 7.3124E+04, 7.3313E+04, 7.3502E+04,
     +  7.3692E+04, 7.3882E+04, 7.4072E+04, 7.4263E+04, 7.4454E+04,
     + 7.4646E+04, 7.4838E+04, 7.5031E+04, 7.5223E+04, 7.5417E+04,
     +  7.5610E+04, 7.5804E+04, 7.5999E+04, 7.6193E+04, 7.6389E+04,
     + 7.6584E+04, 7.6780E+04, 7.6977E+04, 7.7174E+04, 7.7371E+04,
     +  7.7568E+04, 7.7766E+04, 7.7965E+04, 7.8164E+04, 7.8363E+04,
     + 7.8563E+04, 7.8763E+04, 7.8963E+04, 7.9164E+04, 7.9366E+04,
     +  7.9567E+04, 7.9769E+04, 7.9972E+04, 8.0175E+04, 8.0378E+04,
     + 8.0582E+04, 8.0786E+04, 8.0991E+04, 8.1196E+04, 8.1401E+04,
     +  8.1607E+04, 8.1814E+04, 8.2020E+04, 8.2227E+04, 8.2435E+04,
     + 8.2643E+04, 8.2851E+04, 8.3060E+04, 8.3269E+04, 8.3479E+04,
     +  8.3689E+04, 8.3900E+04, 8.4111E+04, 8.4322E+04, 8.4534E+04,
     + 8.4746E+04, 8.4959E+04, 8.5172E+04, 8.5385E+04, 8.5599E+04,
     +  8.5814E+04, 8.6028E+04, 8.6244E+04, 8.6459E+04, 8.6675E+04,
     + 8.6892E+04, 8.7109E+04, 8.7326E+04, 8.7544E+04, 8.7762E+04,
     +  8.7981E+04, 8.8200E+04, 8.8420E+04, 8.8640E+04, 8.8860E+04,
     + 8.9081E+04, 8.9303E+04, 8.9524E+04, 8.9747E+04, 8.9969E+04,
     +  9.0192E+04, 9.0416E+04, 9.0640E+04, 9.0865E+04, 9.1089E+04,
     + 9.1315E+04, 9.1541E+04, 9.1767E+04, 9.1994E+04, 9.2221E+04,
     +  9.2448E+04, 9.2676E+04, 9.2905E+04, 9.3134E+04, 9.3363E+04,
     + 9.3593E+04, 9.3823E+04, 9.4054E+04, 9.4285E+04, 9.4517E+04,
     +  9.4749E+04, 9.4982E+04, 9.5215E+04, 9.5449E+04, 9.5683E+04,
     + 9.5917E+04, 9.6152E+04, 9.6387E+04, 9.6623E+04, 9.6859E+04,
     +  9.7096E+04, 9.7333E+04, 9.7571E+04, 9.7809E+04, 9.8048E+04,
     + 9.8287E+04, 9.8527E+04, 9.8767E+04, 9.9007E+04, 9.9248E+04,
     +  9.9490E+04, 9.9732E+04, 9.9974E+04, 1.0022E+05, 1.0046E+05,
     + 1.0070E+05, 1.0095E+05, 1.0119E+05, 1.0144E+05, 1.0168E+05,
     +  1.0193E+05, 1.0218E+05, 1.0242E+05, 1.0267E+05, 1.0292E+05,
     + 1.0317E+05, 1.0342E+05, 1.0367E+05, 1.0392E+05, 1.0417E+05,
     +  1.0442E+05, 1.0467E+05, 1.0492E+05, 1.0517E+05, 1.0543E+05,
     + 1.0568E+05, 1.0594E+05, 1.0619E+05, 1.0644E+05, 1.0670E+05,
     +  1.0696E+05, 1.0721E+05, 1.0747E+05, 1.0773E+05, 1.0798E+05,
     + 1.0824E+05, 1.0850E+05, 1.0876E+05, 1.0902E+05, 1.0928E+05,
     +  1.0954E+05, 1.0980E+05, 1.1007E+05, 1.1033E+05, 1.1059E+05,
     + 1.1085E+05, 1.1112E+05, 1.1138E+05, 1.1165E+05, 1.1191E+05,
     +  1.1218E+05, 1.1245E+05, 1.1271E+05, 1.1298E+05, 1.1325E+05,
     + 1.1352E+05, 1.1379E+05, 1.1406E+05, 1.1433E+05, 1.1460E+05,
     +  1.1487E+05, 1.1514E+05, 1.1541E+05, 1.1568E+05, 1.1596E+05,
     + 1.1623E+05, 1.1650E+05, 1.1678E+05, 1.1705E+05, 1.1733E+05,
     +  1.1761E+05, 1.1788E+05, 1.1816E+05, 1.1844E+05, 1.1872E+05,
     + 1.1900E+05, 1.1927E+05, 1.1955E+05, 1.1984E+05, 1.2012E+05,
     +  1.2040E+05, 1.2068E+05, 1.2096E+05, 1.2125E+05, 1.2153E+05,
     + 1.2181E+05, 1.2210E+05, 1.2238E+05, 1.2267E+05, 1.2296E+05,
     +  1.2324E+05, 1.2353E+05, 1.2382E+05, 1.2411E+05, 1.2440E+05,
     + 1.2468E+05, 1.2497E+05, 1.2527E+05, 1.2556E+05, 1.2585E+05,
     +  1.2614E+05, 1.2643E+05, 1.2673E+05, 1.2702E+05, 1.2732E+05,
     + 1.2761E+05, 1.2791E+05, 1.2820E+05, 1.2850E+05, 1.2880E+05,
     +  1.2909E+05, 1.2939E+05, 1.2969E+05, 1.2999E+05, 1.3029E+05,
     + 1.3059E+05, 1.3089E+05, 1.3119E+05, 1.3150E+05, 1.3180E+05,
     +  1.3210E+05, 1.3241E+05, 1.3271E+05, 1.3302E+05, 1.3332E+05,
     + 1.3363E+05, 1.3393E+05, 1.3424E+05, 1.3455E+05, 1.3486E+05,
     +  1.3517E+05, 1.3548E+05, 1.3579E+05, 1.3610E+05, 1.3641E+05,
     + 1.3672E+05, 1.3703E+05, 1.3735E+05, 1.3766E+05, 1.3797E+05,
     +  1.3829E+05, 1.3860E+05, 1.3892E+05, 1.3924E+05, 1.3955E+05,
     + 1.3987E+05, 1.4019E+05, 1.4051E+05, 1.4083E+05, 1.4115E+05,
     +  1.4147E+05, 1.4179E+05, 1.4211E+05, 1.4243E+05, 1.4276E+05,
     + 1.4308E+05, 1.4340E+05, 1.4373E+05, 1.4405E+05, 1.4438E+05,
     +  1.4471E+05, 1.4503E+05, 1.4536E+05, 1.4569E+05, 1.4602E+05,
     + 1.4635E+05, 1.4668E+05, 1.4701E+05, 1.4734E+05, 1.4767E+05,
     +  1.4800E+05, 1.4834E+05, 1.4867E+05, 1.4901E+05, 1.4934E+05,
     + 1.4968E+05, 1.5001E+05, 1.5035E+05, 1.5069E+05, 1.5102E+05,
     +  1.5136E+05, 1.5170E+05, 1.5204E+05, 1.5238E+05, 1.5272E+05,
     + 1.5306E+05, 1.5341E+05, 1.5375E+05, 1.5409E+05, 1.5444E+05,
     +  1.5478E+05, 1.5513E+05, 1.5547E+05, 1.5582E+05, 1.5617E+05,
     + 1.5651E+05, 1.5686E+05, 1.5721E+05, 1.5756E+05, 1.5791E+05,
     +  1.5826E+05, 1.5861E+05, 1.5897E+05, 1.5932E+05, 1.5967E+05,
     + 1.6003E+05, 1.6038E+05, 1.6074E+05, 1.6109E+05, 1.6145E+05,
     +  1.6181E+05, 1.6216E+05, 1.6252E+05, 1.6288E+05, 1.6324E+05,
     + 1.6360E+05, 1.6396E+05, 1.6432E+05, 1.6469E+05, 1.6505E+05,
     +  1.6541E+05, 1.6578E+05, 1.6614E+05, 1.6651E+05, 1.6687E+05,
     + 1.6724E+05, 1.6761E+05, 1.6798E+05, 1.6834E+05, 1.6871E+05,
     +  1.6908E+05, 1.6946E+05, 1.6983E+05, 1.7020E+05, 1.7057E+05,
     + 1.7094E+05, 1.7132E+05, 1.7169E+05, 1.7207E+05, 1.7244E+05,
     +  1.7282E+05, 1.7320E+05, 1.7358E+05, 1.7396E+05, 1.7434E+05,
     + 1.7472E+05, 1.7510E+05, 1.7548E+05, 1.7586E+05, 1.7624E+05,
     +  1.7663E+05, 1.7701E+05, 1.7739E+05, 1.7778E+05, 1.7817E+05,
     + 1.7855E+05, 1.7894E+05, 1.7933E+05, 1.7972E+05, 1.8011E+05,
     +  1.8050E+05, 1.8089E+05, 1.8128E+05, 1.8167E+05, 1.8206E+05,
     + 1.8246E+05, 1.8285E+05, 1.8325E+05, 1.8364E+05, 1.8404E+05,
     +  1.8443E+05, 1.8483E+05, 1.8523E+05, 1.8563E+05, 1.8603E+05,
     + 1.8643E+05, 1.8683E+05, 1.8723E+05, 1.8763E+05, 1.8804E+05,
     +  1.8844E+05, 1.8885E+05, 1.8925E+05, 1.8966E+05, 1.9006E+05,
     + 1.9047E+05, 1.9088E+05, 1.9129E+05, 1.9170E+05, 1.9211E+05,
     +  1.9252E+05, 1.9293E+05, 1.9334E+05, 1.9376E+05, 1.9417E+05,
     + 1.9458E+05, 1.9500E+05, 1.9542E+05, 1.9583E+05, 1.9625E+05,
     +  1.9667E+05, 1.9709E+05, 1.9751E+05, 1.9793E+05, 1.9835E+05,
     + 1.9877E+05, 1.9919E+05, 1.9961E+05, 2.0004E+05, 2.0046E+05,
     +  2.0089E+05, 2.0131E+05, 2.0174E+05, 2.0217E+05, 2.0260E+05,
     + 2.0302E+05, 2.0345E+05, 2.0388E+05, 2.0432E+05, 2.0475E+05,
     +  2.0518E+05, 2.0561E+05, 2.0605E+05, 2.0648E+05, 2.0692E+05,
     + 2.0735E+05, 2.0779E+05, 2.0823E+05, 2.0867E+05, 2.0911E+05,
     +  2.0955E+05, 2.0999E+05, 2.1043E+05, 2.1087E+05, 2.1131E+05,
     + 2.1176E+05, 2.1220E+05, 2.1265E+05, 2.1309E+05, 2.1354E+05,
     +  2.1399E+05, 2.1444E+05, 2.1488E+05, 2.1533E+05, 2.1578E+05,
     + 2.1624E+05, 2.1669E+05, 2.1714E+05, 2.1759E+05, 2.1805E+05,
     +  2.1850E+05, 2.1896E+05, 2.1942E+05, 2.1987E+05, 2.2033E+05,
     + 2.2079E+05, 2.2125E+05, 2.2171E+05, 2.2217E+05, 2.2263E+05,
     +  2.2310E+05, 2.2356E+05, 2.2402E+05, 2.2449E+05, 2.2495E+05,
     + 2.2542E+05, 2.2589E+05, 2.2636E+05, 2.2683E+05, 2.2730E+05,
     +  2.2777E+05, 2.2824E+05, 2.2871E+05, 2.2918E+05, 2.2966E+05,
     + 2.3013E+05, 2.3061E+05, 2.3108E+05, 2.3156E+05, 2.3204E+05,
     +  2.3252E+05, 2.3299E+05, 2.3347E+05, 2.3396E+05, 2.3444E+05,
     + 2.3492E+05, 2.3540E+05, 2.3589E+05, 2.3637E+05, 2.3686E+05,
     +  2.3734E+05, 2.3783E+05, 2.3832E+05, 2.3881E+05, 2.3930E+05,
     + 2.3979E+05, 2.4028E+05, 2.4077E+05, 2.4126E+05, 2.4176E+05,
     +  2.4225E+05, 2.4275E+05, 2.4324E+05, 2.4374E+05, 2.4424E+05,
     + 2.4474E+05, 2.4524E+05, 2.4574E+05, 2.4624E+05, 2.4674E+05,
     +  2.4724E+05, 2.4775E+05, 2.4825E+05, 2.4875E+05, 2.4926E+05,
     + 2.4977E+05, 2.5028E+05, 2.5078E+05, 2.5129E+05, 2.5180E+05,
     +  2.5231E+05, 2.5283E+05, 2.5334E+05, 2.5385E+05, 2.5437E+05,
     + 2.5488E+05, 2.5540E+05, 2.5591E+05, 2.5643E+05, 2.5695E+05,
     +  2.5747E+05, 2.5799E+05, 2.5851E+05, 2.5903E+05, 2.5956E+05,
     + 2.6008E+05, 2.6060E+05, 2.6113E+05, 2.6166E+05, 2.6218E+05,
     +  2.6271E+05, 2.6324E+05, 2.6377E+05, 2.6430E+05, 2.6483E+05,
     + 2.6536E+05, 2.6590E+05, 2.6643E+05, 2.6697E+05, 2.6750E+05,
     +  2.6804E+05, 2.6857E+05, 2.6911E+05, 2.6965E+05, 2.7019E+05,
     + 2.7073E+05, 2.7127E+05, 2.7182E+05, 2.7236E+05, 2.7291E+05,
     +  2.7345E+05, 2.7400E+05, 2.7454E+05, 2.7509E+05, 2.7564E+05,
     + 2.7619E+05, 2.7674E+05, 2.7729E+05, 2.7784E+05, 2.7840E+05,
     +  2.7895E+05, 2.7951E+05, 2.8006E+05, 2.8062E+05, 2.8118E+05,
     + 2.8173E+05, 2.8229E+05, 2.8285E+05, 2.8342E+05, 2.8398E+05,
     +  2.8454E+05, 2.8510E+05, 2.8567E+05, 2.8623E+05, 2.8680E+05,
     + 2.8737E+05, 2.8794E+05, 2.8851E+05, 2.8908E+05, 2.8965E+05,
     +  2.9022E+05, 2.9079E+05, 2.9137E+05, 2.9194E+05, 2.9252E+05,
     + 2.9309E+05, 2.9367E+05, 2.9425E+05, 2.9483E+05, 2.9541E+05,
     +  2.9599E+05, 2.9657E+05, 2.9715E+05, 2.9774E+05, 2.9832E+05,
     + 2.9891E+05, 2.9950E+05, 3.0008E+05, 3.0067E+05, 3.0126E+05,
     +  3.0185E+05, 3.0244E+05, 3.0304E+05, 3.0363E+05, 3.0422E+05,
     + 3.0482E+05, 3.0541E+05, 3.0601E+05, 3.0661E+05, 3.0721E+05,
     +  3.0781E+05, 3.0841E+05, 3.0901E+05, 3.0961E+05, 3.1022E+05,
     + 3.1082E+05, 3.1143E+05, 3.1203E+05, 3.1264E+05, 3.1325E+05,
     +  3.1386E+05, 3.1447E+05, 3.1508E+05, 3.1569E+05, 3.1630E+05,
     + 3.1692E+05, 3.1753E+05, 3.1815E+05, 3.1876E+05, 3.1938E+05,
     +  3.2000E+05, 3.2062E+05, 3.2124E+05, 3.2186E+05, 3.2249E+05,
     + 3.2311E+05, 3.2373E+05, 3.2436E+05, 3.2499E+05, 3.2561E+05,
     +  3.2624E+05, 3.2687E+05, 3.2750E+05, 3.2813E+05, 3.2877E+05,
     + 3.2940E+05, 3.3003E+05, 3.3067E+05, 3.3131E+05, 3.3194E+05,
     +  3.3258E+05, 3.3322E+05, 3.3386E+05, 3.3450E+05, 3.3514E+05,
     + 3.3579E+05, 3.3643E+05, 3.3708E+05, 3.3772E+05, 3.3837E+05,
     +  3.3902E+05, 3.3967E+05, 3.4032E+05, 3.4097E+05, 3.4162E+05,
     + 3.4228E+05, 3.4293E+05, 3.4359E+05, 3.4424E+05, 3.4490E+05,
     +  3.4556E+05, 3.4622E+05, 3.4688E+05, 3.4754E+05, 3.4820E+05,
     + 3.4886E+05, 3.4953E+05, 3.5019E+05, 3.5086E+05, 3.5153E+05,
     +  3.5220E+05, 3.5287E+05, 3.5354E+05, 3.5421E+05, 3.5488E+05,
     + 3.5555E+05, 3.5623E+05, 3.5690E+05, 3.5758E+05, 3.5826E+05,
     +  3.5894E+05, 3.5962E+05, 3.6030E+05, 3.6098E+05, 3.6166E+05,
     + 3.6235E+05, 3.6303E+05, 3.6372E+05, 3.6440E+05, 3.6509E+05,
     +  3.6578E+05, 3.6647E+05, 3.6716E+05, 3.6785E+05, 3.6855E+05,
     + 3.6924E+05, 3.6994E+05, 3.7063E+05, 3.7133E+05, 3.7203E+05,
     +  3.7273E+05, 3.7343E+05, 3.7413E+05, 3.7484E+05, 3.7554E+05,
     + 3.7624E+05, 3.7695E+05, 3.7766E+05, 3.7837E+05, 3.7907E+05,
     +  3.7978E+05, 3.8050E+05, 3.8121E+05, 3.8192E+05, 3.8264E+05,
     + 3.8335E+05, 3.8407E+05, 3.8479E+05, 3.8551E+05, 3.8623E+05,
     +  3.8695E+05, 3.8767E+05, 3.8839E+05, 3.8912E+05, 3.8984E+05,
     + 3.9057E+05, 3.9130E+05, 3.9202E+05, 3.9275E+05, 3.9348E+05,
     +  3.9422E+05, 3.9495E+05, 3.9568E+05, 3.9642E+05, 3.9716E+05,
     + 3.9789E+05, 3.9863E+05, 3.9937E+05, 4.0011E+05, 4.0085E+05,
     +  4.0160E+05, 4.0234E+05, 4.0308E+05, 4.0383E+05, 4.0458E+05,
     + 4.0533E+05, 4.0608E+05, 4.0683E+05, 4.0758E+05, 4.0833E+05,
     +  4.0909E+05, 4.0984E+05, 4.1060E+05, 4.1135E+05, 4.1211E+05,
     + 4.1287E+05, 4.1363E+05, 4.1440E+05, 4.1516E+05, 4.1592E+05,
     +  4.1669E+05, 4.1745E+05, 4.1822E+05, 4.1899E+05, 4.1976E+05,
     + 4.2053E+05, 4.2130E+05, 4.2208E+05, 4.2285E+05, 4.2363E+05,
     +  4.2440E+05, 4.2518E+05, 4.2596E+05, 4.2674E+05, 4.2752E+05,
     + 4.2830E+05, 4.2909E+05, 4.2987E+05, 4.3066E+05, 4.3144E+05,
     +  4.3223E+05, 4.3302E+05, 4.3381E+05, 4.3460E+05, 4.3540E+05,
     + 4.3619E+05, 4.3698E+05, 4.3778E+05, 4.3858E+05, 4.3938E+05,
     +  4.4018E+05, 4.4098E+05, 4.4178E+05, 4.4258E+05, 4.4339E+05,
     + 4.4419E+05, 4.4500E+05, 4.4581E+05, 4.4662E+05, 4.4743E+05,
     +  4.4824E+05, 4.4905E+05, 4.4986E+05, 4.5068E+05, 4.5149E+05,
     + 4.5231E+05, 4.5313E+05, 4.5395E+05, 4.5477E+05, 4.5559E+05,
     +  4.5642E+05, 4.5724E+05, 4.5807E+05, 4.5889E+05, 4.5972E+05,
     + 4.6055E+05, 4.6138E+05, 4.6221E+05, 4.6305E+05, 4.6388E+05,
     +  4.6472E+05, 4.6555E+05, 4.6639E+05, 4.6723E+05, 4.6807E+05,
     + 4.6891E+05, 4.6975E+05, 4.7060E+05, 4.7144E+05, 4.7229E+05,
     +  4.7314E+05, 4.7398E+05, 4.7483E+05, 4.7568E+05, 4.7654E+05,
     + 4.7739E+05, 4.7825E+05, 4.7910E+05, 4.7996E+05, 4.8082E+05,
     +  4.8168E+05, 4.8254E+05, 4.8340E+05, 4.8426E+05, 4.8513E+05,
     + 4.8599E+05, 4.8686E+05, 4.8773E+05, 4.8860E+05, 4.8947E+05,
     +  4.9034E+05, 4.9121E+05, 4.9209E+05, 4.9296E+05, 4.9384E+05,
     + 4.9472E+05, 4.9560E+05, 4.9648E+05, 4.9736E+05, 4.9825E+05,
     +  4.9913E+05, 5.0002E+05, 5.0090E+05, 5.0179E+05, 5.0268E+05,
     + 5.0357E+05, 5.0446E+05, 5.0536E+05, 5.0625E+05, 5.0715E+05,
     +  5.0805E+05, 5.0894E+05, 5.0984E+05, 5.1074E+05, 5.1165E+05,
     + 5.1255E+05, 5.1346E+05, 5.1436E+05, 5.1527E+05, 5.1618E+05,
     +  5.1709E+05, 5.1800E+05, 5.1891E+05, 5.1982E+05, 5.2074E+05,
     + 5.2166E+05, 5.2257E+05, 5.2349E+05, 5.2441E+05, 5.2533E+05,
     +  5.2626E+05, 5.2718E+05, 5.2811E+05, 5.2903E+05, 5.2996E+05,
     + 5.3089E+05, 5.3182E+05, 5.3275E+05, 5.3369E+05, 5.3462E+05,
     +  5.3556E+05, 5.3650E+05, 5.3743E+05, 5.3837E+05, 5.3931E+05,
     + 5.4026E+05, 5.4120E+05, 5.4215E+05, 5.4309E+05, 5.4404E+05,
     +  5.4499E+05, 5.4594E+05, 5.4689E+05, 5.4784E+05, 5.4880E+05,
     + 5.4975E+05, 5.5071E+05, 5.5167E+05, 5.5263E+05, 5.5359E+05,
     +  5.5455E+05, 5.5552E+05, 5.5648E+05, 5.5745E+05, 5.5842E+05,
     + 5.5939E+05, 5.6036E+05, 5.6133E+05, 5.6230E+05, 5.6327E+05,
     +  5.6425E+05, 5.6523E+05, 5.6621E+05, 5.6719E+05, 5.6817E+05,
     + 5.6915E+05, 5.7013E+05, 5.7112E+05, 5.7211E+05, 5.7309E+05,
     +  5.7408E+05, 5.7507E+05, 5.7607E+05, 5.7706E+05, 5.7805E+05,
     + 5.7905E+05, 5.8005E+05, 5.8105E+05, 5.8205E+05, 5.8305E+05,
     +  5.8405E+05, 5.8506E+05, 5.8606E+05, 5.8707E+05, 5.8808E+05,
     + 5.8909E+05, 5.9010E+05, 5.9111E+05, 5.9213E+05, 5.9314E+05,
     +  5.9416E+05, 5.9518E+05, 5.9620E+05, 5.9722E+05, 5.9824E+05,
     + 5.9927E+05, 6.0029E+05, 6.0132E+05, 6.0235E+05, 6.0337E+05,
     +  6.0441E+05, 6.0544E+05, 6.0647E+05, 6.0751E+05, 6.0854E+05,
     + 6.0958E+05, 6.1062E+05, 6.1166E+05, 6.1270E+05, 6.1375E+05,
     +  6.1479E+05, 6.1584E+05, 6.1689E+05, 6.1794E+05, 6.1899E+05,
     + 6.2004E+05, 6.2109E+05, 6.2215E+05, 6.2320E+05, 6.2426E+05,
     +  6.2532E+05, 6.2638E+05, 6.2744E+05, 6.2851E+05, 6.2957E+05,
     + 6.3064E+05, 6.3171E+05, 6.3278E+05, 6.3385E+05, 6.3492E+05,
     +  6.3599E+05, 6.3707E+05, 6.3815E+05, 6.3922E+05, 6.4030E+05,
     + 6.4138E+05, 6.4247E+05, 6.4355E+05, 6.4464E+05, 6.4572E+05,
     +  6.4681E+05, 6.4790E+05, 6.4899E+05, 6.5008E+05, 6.5118E+05,
     + 6.5227E+05, 6.5337E+05, 6.5447E+05, 6.5557E+05, 6.5667E+05,
     +  6.5777E+05, 6.5888E+05, 6.5998E+05, 6.6109E+05, 6.6220E+05,
     + 6.6331E+05, 6.6442E+05, 6.6554E+05, 6.6665E+05, 6.6777E+05,
     +  6.6888E+05, 6.7000E+05, 6.7112E+05, 6.7225E+05, 6.7337E+05,
     + 6.7449E+05, 6.7562E+05, 6.7675E+05, 6.7788E+05, 6.7901E+05,
     +  6.8014E+05, 6.8128E+05, 6.8241E+05, 6.8355E+05, 6.8469E+05,
     + 6.8583E+05, 6.8697E+05, 6.8811E+05, 6.8926E+05, 6.9041E+05,
     +  6.9155E+05, 6.9270E+05, 6.9385E+05, 6.9501E+05, 6.9616E+05,
     + 6.9731E+05, 6.9847E+05, 6.9963E+05, 7.0079E+05, 7.0195E+05,
     +  7.0311E+05, 7.0428E+05, 7.0545E+05, 7.0661E+05, 7.0778E+05,
     + 7.0895E+05, 7.1013E+05, 7.1130E+05, 7.1247E+05, 7.1365E+05,
     +  7.1483E+05, 7.1601E+05, 7.1719E+05, 7.1837E+05, 7.1956E+05,
     + 7.2074E+05, 7.2193E+05, 7.2312E+05, 7.2431E+05, 7.2551E+05,
     +  7.2670E+05, 7.2789E+05, 7.2909E+05, 7.3029E+05, 7.3149E+05,
     + 7.3269E+05, 7.3390E+05, 7.3510E+05, 7.3631E+05, 7.3752E+05,
     +  7.3872E+05, 7.3994E+05, 7.4115E+05, 7.4236E+05, 7.4358E+05,
     + 7.4480E+05, 7.4602E+05, 7.4724E+05, 7.4846E+05, 7.4968E+05,
     +  7.5091E+05, 7.5213E+05, 7.5336E+05, 7.5459E+05, 7.5583E+05,
     + 7.5706E+05, 7.5829E+05, 7.5953E+05, 7.6077E+05, 7.6201E+05,
     +  7.6325E+05, 7.6449E+05, 7.6574E+05, 7.6698E+05, 7.6823E+05,
     + 7.6948E+05, 7.7073E+05, 7.7198E+05, 7.7324E+05, 7.7449E+05,
     +  7.7575E+05, 7.7701E+05, 7.7827E+05, 7.7953E+05, 7.8080E+05,
     + 7.8206E+05, 7.8333E+05, 7.8460E+05, 7.8587E+05, 7.8714E+05,
     +  7.8842E+05, 7.8969E+05, 7.9097E+05, 7.9225E+05, 7.9353E+05,
     + 7.9481E+05, 7.9609E+05, 7.9738E+05, 7.9867E+05, 7.9995E+05,
     +  8.0124E+05, 8.0254E+05, 8.0383E+05, 8.0512E+05, 8.0642E+05,
     + 8.0772E+05, 8.0902E+05, 8.1032E+05, 8.1162E+05, 8.1293E+05,
     +  8.1424E+05, 8.1554E+05, 8.1685E+05, 8.1817E+05, 8.1948E+05,
     + 8.2079E+05, 8.2211E+05, 8.2343E+05, 8.2475E+05, 8.2607E+05,
     +  8.2739E+05, 8.2872E+05, 8.3005E+05, 8.3137E+05, 8.3270E+05,
     + 8.3404E+05, 8.3537E+05, 8.3671E+05, 8.3804E+05, 8.3938E+05,
     +  8.4072E+05, 8.4206E+05, 8.4341E+05, 8.4475E+05, 8.4610E+05/
      data (QofT( 8,J),J=2409,4510)/ 8.4745E+05, 8.4880E+05,
     +  8.5015E+05, 8.5150E+05, 8.5286E+05, 8.5422E+05, 8.5557E+05,
     +  8.5694E+05,
     + 8.5830E+05, 8.5966E+05, 8.6103E+05, 8.6239E+05, 8.6376E+05,
     +  8.6513E+05, 8.6651E+05, 8.6788E+05, 8.6926E+05, 8.7064E+05,
     + 8.7201E+05, 8.7340E+05, 8.7478E+05, 8.7616E+05, 8.7755E+05,
     +  8.7894E+05, 8.8033E+05, 8.8172E+05, 8.8311E+05, 8.8451E+05,
     + 8.8590E+05, 8.8730E+05, 8.8870E+05, 8.9011E+05, 8.9151E+05,
     +  8.9291E+05, 8.9432E+05, 8.9573E+05, 8.9714E+05, 8.9855E+05,
     + 8.9997E+05, 9.0138E+05, 9.0280E+05, 9.0422E+05, 9.0564E+05,
     +  9.0707E+05, 9.0849E+05, 9.0992E+05, 9.1135E+05, 9.1278E+05,
     + 9.1421E+05, 9.1564E+05, 9.1708E+05, 9.1851E+05, 9.1995E+05,
     +  9.2139E+05, 9.2284E+05, 9.2428E+05, 9.2573E+05, 9.2717E+05,
     + 9.2862E+05, 9.3008E+05, 9.3153E+05, 9.3298E+05, 9.3444E+05,
     +  9.3590E+05, 9.3736E+05, 9.3882E+05, 9.4029E+05, 9.4175E+05,
     + 9.4322E+05, 9.4469E+05, 9.4616E+05, 9.4763E+05, 9.4911E+05,
     +  9.5058E+05, 9.5206E+05, 9.5354E+05, 9.5502E+05, 9.5651E+05,
     + 9.5799E+05, 9.5948E+05, 9.6097E+05, 9.6246E+05, 9.6395E+05,
     +  9.6545E+05, 9.6695E+05, 9.6844E+05, 9.6994E+05, 9.7145E+05,
     + 9.7295E+05, 9.7446E+05, 9.7596E+05, 9.7747E+05, 9.7898E+05,
     +  9.8050E+05, 9.8201E+05, 9.8353E+05, 9.8505E+05, 9.8657E+05,
     + 9.8809E+05, 9.8961E+05, 9.9114E+05, 9.9267E+05, 9.9420E+05,
     +  9.9573E+05, 9.9726E+05, 9.9880E+05, 1.0003E+06, 1.0019E+06,
     + 1.0034E+06, 1.0050E+06, 1.0065E+06, 1.0080E+06, 1.0096E+06,
     +  1.0111E+06, 1.0127E+06, 1.0142E+06, 1.0158E+06, 1.0174E+06,
     + 1.0189E+06, 1.0205E+06, 1.0220E+06, 1.0236E+06, 1.0252E+06,
     +  1.0267E+06, 1.0283E+06, 1.0299E+06, 1.0315E+06, 1.0330E+06,
     + 1.0346E+06, 1.0362E+06, 1.0378E+06, 1.0394E+06, 1.0409E+06,
     +  1.0425E+06, 1.0441E+06, 1.0457E+06, 1.0473E+06, 1.0489E+06,
     + 1.0505E+06, 1.0521E+06, 1.0537E+06, 1.0553E+06, 1.0569E+06,
     +  1.0585E+06, 1.0601E+06, 1.0617E+06, 1.0633E+06, 1.0650E+06,
     + 1.0666E+06, 1.0682E+06, 1.0698E+06, 1.0714E+06, 1.0731E+06,
     +  1.0747E+06, 1.0763E+06, 1.0779E+06, 1.0796E+06, 1.0812E+06,
     + 1.0828E+06, 1.0845E+06, 1.0861E+06, 1.0878E+06, 1.0894E+06,
     +  1.0911E+06, 1.0927E+06, 1.0944E+06, 1.0960E+06, 1.0977E+06,
     + 1.0993E+06, 1.1010E+06, 1.1026E+06, 1.1043E+06, 1.1059E+06,
     +  1.1076E+06, 1.1093E+06, 1.1109E+06, 1.1126E+06, 1.1143E+06,
     + 1.1160E+06, 1.1176E+06, 1.1193E+06, 1.1210E+06, 1.1227E+06,
     +  1.1244E+06, 1.1261E+06, 1.1277E+06, 1.1294E+06, 1.1311E+06,
     + 1.1328E+06, 1.1345E+06, 1.1362E+06, 1.1379E+06, 1.1396E+06,
     +  1.1413E+06, 1.1430E+06, 1.1447E+06, 1.1464E+06, 1.1482E+06,
     + 1.1499E+06, 1.1516E+06, 1.1533E+06, 1.1550E+06, 1.1567E+06,
     +  1.1585E+06, 1.1602E+06, 1.1619E+06, 1.1637E+06, 1.1654E+06,
     + 1.1671E+06, 1.1689E+06, 1.1706E+06, 1.1723E+06, 1.1741E+06,
     +  1.1758E+06, 1.1776E+06, 1.1793E+06, 1.1811E+06, 1.1828E+06,
     + 1.1846E+06, 1.1863E+06, 1.1881E+06, 1.1898E+06, 1.1916E+06,
     +  1.1934E+06, 1.1951E+06, 1.1969E+06, 1.1987E+06, 1.2005E+06,
     + 1.2022E+06, 1.2040E+06, 1.2058E+06, 1.2076E+06, 1.2093E+06,
     +  1.2111E+06, 1.2129E+06, 1.2147E+06, 1.2165E+06, 1.2183E+06,
     + 1.2201E+06, 1.2219E+06, 1.2237E+06, 1.2255E+06, 1.2273E+06,
     +  1.2291E+06, 1.2309E+06, 1.2327E+06, 1.2345E+06, 1.2363E+06,
     + 1.2381E+06, 1.2400E+06, 1.2418E+06, 1.2436E+06, 1.2454E+06,
     +  1.2473E+06, 1.2491E+06, 1.2509E+06, 1.2527E+06, 1.2546E+06,
     + 1.2564E+06, 1.2583E+06, 1.2601E+06, 1.2619E+06, 1.2638E+06,
     +  1.2656E+06, 1.2675E+06, 1.2693E+06, 1.2712E+06, 1.2730E+06,
     + 1.2749E+06, 1.2768E+06, 1.2786E+06, 1.2805E+06, 1.2823E+06,
     +  1.2842E+06, 1.2861E+06, 1.2880E+06, 1.2898E+06, 1.2917E+06,
     + 1.2936E+06, 1.2955E+06, 1.2973E+06, 1.2992E+06, 1.3011E+06,
     +  1.3030E+06, 1.3049E+06, 1.3068E+06, 1.3087E+06, 1.3106E+06,
     + 1.3125E+06, 1.3144E+06, 1.3163E+06, 1.3182E+06, 1.3201E+06,
     +  1.3220E+06, 1.3239E+06, 1.3258E+06, 1.3278E+06, 1.3297E+06,
     + 1.3316E+06, 1.3335E+06, 1.3354E+06, 1.3374E+06, 1.3393E+06,
     +  1.3412E+06, 1.3432E+06, 1.3451E+06, 1.3470E+06, 1.3490E+06,
     + 1.3509E+06, 1.3529E+06, 1.3548E+06, 1.3568E+06, 1.3587E+06,
     +  1.3607E+06, 1.3626E+06, 1.3646E+06, 1.3665E+06, 1.3685E+06,
     + 1.3705E+06, 1.3724E+06, 1.3744E+06, 1.3764E+06, 1.3783E+06,
     +  1.3803E+06, 1.3823E+06, 1.3843E+06, 1.3863E+06, 1.3882E+06,
     + 1.3902E+06, 1.3922E+06, 1.3942E+06, 1.3962E+06, 1.3982E+06,
     +  1.4002E+06, 1.4022E+06, 1.4042E+06, 1.4062E+06, 1.4082E+06,
     + 1.4102E+06, 1.4122E+06, 1.4142E+06, 1.4163E+06, 1.4183E+06,
     +  1.4203E+06, 1.4223E+06, 1.4243E+06, 1.4264E+06, 1.4284E+06,
     + 1.4304E+06, 1.4324E+06, 1.4345E+06, 1.4365E+06, 1.4386E+06,
     +  1.4406E+06, 1.4426E+06, 1.4447E+06, 1.4467E+06, 1.4488E+06,
     + 1.4508E+06, 1.4529E+06, 1.4550E+06, 1.4570E+06, 1.4591E+06,
     +  1.4611E+06, 1.4632E+06, 1.4653E+06, 1.4673E+06, 1.4694E+06,
     + 1.4715E+06, 1.4736E+06, 1.4756E+06, 1.4777E+06, 1.4798E+06,
     +  1.4819E+06, 1.4840E+06, 1.4861E+06, 1.4882E+06, 1.4903E+06,
     + 1.4924E+06, 1.4945E+06, 1.4966E+06, 1.4987E+06, 1.5008E+06,
     +  1.5029E+06, 1.5050E+06, 1.5071E+06, 1.5092E+06, 1.5113E+06,
     + 1.5135E+06, 1.5156E+06, 1.5177E+06, 1.5198E+06, 1.5220E+06,
     +  1.5241E+06, 1.5262E+06, 1.5284E+06, 1.5305E+06, 1.5326E+06,
     + 1.5348E+06, 1.5369E+06, 1.5391E+06, 1.5412E+06, 1.5434E+06,
     +  1.5455E+06, 1.5477E+06, 1.5499E+06, 1.5520E+06, 1.5542E+06,
     + 1.5563E+06, 1.5585E+06, 1.5607E+06, 1.5629E+06, 1.5650E+06,
     +  1.5672E+06, 1.5694E+06, 1.5716E+06, 1.5738E+06, 1.5759E+06,
     + 1.5781E+06, 1.5803E+06, 1.5825E+06, 1.5847E+06, 1.5869E+06,
     +  1.5891E+06, 1.5913E+06, 1.5935E+06, 1.5957E+06, 1.5979E+06,
     + 1.6002E+06, 1.6024E+06, 1.6046E+06, 1.6068E+06, 1.6090E+06,
     +  1.6112E+06, 1.6135E+06, 1.6157E+06, 1.6179E+06, 1.6202E+06,
     + 1.6224E+06, 1.6246E+06, 1.6269E+06, 1.6291E+06, 1.6314E+06,
     +  1.6336E+06, 1.6359E+06, 1.6381E+06, 1.6404E+06, 1.6426E+06,
     + 1.6449E+06, 1.6472E+06, 1.6494E+06, 1.6517E+06, 1.6540E+06,
     +  1.6562E+06, 1.6585E+06, 1.6608E+06, 1.6631E+06, 1.6653E+06,
     + 1.6676E+06, 1.6699E+06, 1.6722E+06, 1.6745E+06, 1.6768E+06,
     +  1.6791E+06, 1.6814E+06, 1.6837E+06, 1.6860E+06, 1.6883E+06,
     + 1.6906E+06, 1.6929E+06, 1.6952E+06, 1.6975E+06, 1.6998E+06,
     +  1.7022E+06, 1.7045E+06, 1.7068E+06, 1.7091E+06, 1.7115E+06,
     + 1.7138E+06, 1.7161E+06, 1.7185E+06, 1.7208E+06, 1.7231E+06,
     +  1.7255E+06, 1.7278E+06, 1.7302E+06, 1.7325E+06, 1.7349E+06,
     + 1.7372E+06, 1.7396E+06, 1.7420E+06, 1.7443E+06, 1.7467E+06,
     +  1.7491E+06, 1.7514E+06, 1.7538E+06, 1.7562E+06, 1.7586E+06,
     + 1.7609E+06, 1.7633E+06, 1.7657E+06, 1.7681E+06, 1.7705E+06,
     +  1.7729E+06, 1.7753E+06, 1.7777E+06, 1.7801E+06, 1.7825E+06,
     + 1.7849E+06, 1.7873E+06, 1.7897E+06, 1.7921E+06, 1.7945E+06,
     +  1.7969E+06, 1.7993E+06, 1.8018E+06, 1.8042E+06, 1.8066E+06,
     + 1.8090E+06, 1.8115E+06, 1.8139E+06, 1.8164E+06, 1.8188E+06,
     +  1.8212E+06, 1.8237E+06, 1.8261E+06, 1.8286E+06, 1.8310E+06,
     + 1.8335E+06, 1.8359E+06, 1.8384E+06, 1.8409E+06, 1.8433E+06,
     +  1.8458E+06, 1.8483E+06, 1.8507E+06, 1.8532E+06, 1.8557E+06,
     + 1.8582E+06, 1.8606E+06, 1.8631E+06, 1.8656E+06, 1.8681E+06,
     +  1.8706E+06, 1.8731E+06, 1.8756E+06, 1.8781E+06, 1.8806E+06,
     + 1.8831E+06, 1.8856E+06, 1.8881E+06, 1.8906E+06, 1.8931E+06,
     +  1.8956E+06, 1.8982E+06, 1.9007E+06, 1.9032E+06, 1.9057E+06,
     + 1.9083E+06, 1.9108E+06, 1.9133E+06, 1.9159E+06, 1.9184E+06,
     +  1.9209E+06, 1.9235E+06, 1.9260E+06, 1.9286E+06, 1.9311E+06,
     + 1.9337E+06, 1.9362E+06, 1.9388E+06, 1.9414E+06, 1.9439E+06,
     +  1.9465E+06, 1.9491E+06, 1.9516E+06, 1.9542E+06, 1.9568E+06,
     + 1.9594E+06, 1.9620E+06, 1.9645E+06, 1.9671E+06, 1.9697E+06,
     +  1.9723E+06, 1.9749E+06, 1.9775E+06, 1.9801E+06, 1.9827E+06,
     + 1.9853E+06, 1.9879E+06, 1.9905E+06, 1.9932E+06, 1.9958E+06,
     +  1.9984E+06, 2.0010E+06, 2.0036E+06, 2.0063E+06, 2.0089E+06,
     + 2.0115E+06, 2.0142E+06, 2.0168E+06, 2.0194E+06, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       272
      data (QofT( 9,J),J=1,2408)/ 3.6000E+01, 3.6009E+01, 3.6171E+01,
     +  3.6792E+01, 3.8088E+01, 4.0132E+01, 4.2910E+01, 4.6367E+01,
     + 5.0436E+01, 5.5054E+01, 6.0162E+01, 6.5711E+01, 7.1660E+01,
     +  7.7974E+01, 8.4622E+01, 9.1580E+01, 9.8826E+01, 1.0634E+02,
     + 1.1412E+02, 1.2213E+02, 1.3884E+02, 1.5640E+02, 1.7475E+02,
     +  1.9384E+02, 2.1364E+02, 2.3410E+02, 2.5520E+02, 2.7693E+02,
     + 2.9925E+02, 3.2216E+02, 3.4563E+02, 3.6966E+02, 3.9422E+02,
     +  4.1931E+02, 4.4492E+02, 4.7103E+02, 4.9763E+02, 5.2473E+02,
     + 5.5230E+02, 5.8034E+02, 6.0884E+02, 6.3780E+02, 6.6721E+02,
     +  6.9706E+02, 7.2734E+02, 7.5805E+02, 7.8919E+02, 8.2074E+02,
     + 8.5271E+02, 8.8508E+02, 9.1786E+02, 9.5103E+02, 9.8460E+02,
     +  1.0186E+03, 1.0529E+03, 1.0876E+03, 1.1227E+03, 1.1582E+03,
     + 1.1940E+03, 1.2302E+03, 1.2668E+03, 1.3037E+03, 1.3410E+03,
     +  1.3786E+03, 1.4166E+03, 1.4549E+03, 1.4936E+03, 1.5326E+03,
     + 1.5719E+03, 1.6116E+03, 1.6516E+03, 1.6919E+03, 1.7325E+03,
     +  1.7735E+03, 1.8148E+03, 1.8564E+03, 1.8984E+03, 1.9406E+03,
     + 1.9832E+03, 2.0261E+03, 2.0692E+03, 2.1127E+03, 2.1565E+03,
     +  2.2006E+03, 2.2450E+03, 2.2897E+03, 2.3347E+03, 2.3799E+03,
     + 2.4255E+03, 2.4714E+03, 2.5175E+03, 2.5640E+03, 2.6107E+03,
     +  2.6577E+03, 2.7050E+03, 2.7526E+03, 2.8004E+03, 2.8486E+03,
     + 2.8970E+03, 2.9457E+03, 2.9946E+03, 3.0439E+03, 3.0934E+03,
     +  3.1432E+03, 3.1932E+03, 3.2436E+03, 3.2942E+03, 3.3450E+03,
     + 3.3961E+03, 3.4475E+03, 3.4992E+03, 3.5511E+03, 3.6033E+03,
     +  3.6558E+03, 3.7085E+03, 3.7614E+03, 3.8147E+03, 3.8682E+03,
     + 3.9219E+03, 3.9759E+03, 4.0302E+03, 4.0847E+03, 4.1395E+03,
     +  4.1945E+03, 4.2498E+03, 4.3054E+03, 4.3612E+03, 4.4172E+03,
     + 4.4735E+03, 4.5301E+03, 4.5869E+03, 4.6440E+03, 4.7013E+03,
     +  4.7589E+03, 4.8167E+03, 4.8748E+03, 4.9331E+03, 4.9917E+03,
     + 5.0505E+03, 5.1096E+03, 5.1690E+03, 5.2285E+03, 5.2884E+03,
     +  5.3485E+03, 5.4088E+03, 5.4694E+03, 5.5302E+03, 5.5913E+03,
     + 5.6527E+03, 5.7143E+03, 5.7761E+03, 5.8382E+03, 5.9005E+03,
     +  5.9631E+03, 6.0260E+03, 6.0891E+03, 6.1524E+03, 6.2160E+03,
     + 6.2799E+03, 6.3440E+03, 6.4083E+03, 6.4729E+03, 6.5378E+03,
     +  6.6029E+03, 6.6683E+03, 6.7339E+03, 6.7997E+03, 6.8659E+03,
     + 6.9323E+03, 6.9989E+03, 7.0658E+03, 7.1329E+03, 7.2003E+03,
     +  7.2680E+03, 7.3359E+03, 7.4040E+03, 7.4725E+03, 7.5411E+03,
     + 7.6101E+03, 7.6793E+03, 7.7487E+03, 7.8184E+03, 7.8884E+03,
     +  7.9587E+03, 8.0291E+03, 8.0999E+03, 8.1709E+03, 8.2422E+03,
     + 8.3137E+03, 8.3855E+03, 8.4576E+03, 8.5299E+03, 8.6025E+03,
     +  8.6753E+03, 8.7485E+03, 8.8218E+03, 8.8955E+03, 8.9694E+03,
     + 9.0436E+03, 9.1180E+03, 9.1928E+03, 9.2677E+03, 9.3430E+03,
     +  9.4185E+03, 9.4943E+03, 9.5704E+03, 9.6467E+03, 9.7233E+03,
     + 9.8002E+03, 9.8774E+03, 9.9548E+03, 1.0032E+04, 1.0110E+04,
     +  1.0189E+04, 1.0267E+04, 1.0346E+04, 1.0425E+04, 1.0505E+04,
     + 1.0584E+04, 1.0664E+04, 1.0744E+04, 1.0825E+04, 1.0906E+04,
     +  1.0987E+04, 1.1068E+04, 1.1149E+04, 1.1231E+04, 1.1314E+04,
     + 1.1396E+04, 1.1479E+04, 1.1562E+04, 1.1645E+04, 1.1729E+04,
     +  1.1813E+04, 1.1897E+04, 1.1981E+04, 1.2066E+04, 1.2151E+04,
     + 1.2237E+04, 1.2322E+04, 1.2408E+04, 1.2494E+04, 1.2581E+04,
     +  1.2668E+04, 1.2755E+04, 1.2842E+04, 1.2930E+04, 1.3018E+04,
     + 1.3106E+04, 1.3195E+04, 1.3284E+04, 1.3373E+04, 1.3463E+04,
     +  1.3552E+04, 1.3643E+04, 1.3733E+04, 1.3824E+04, 1.3915E+04,
     + 1.4006E+04, 1.4098E+04, 1.4190E+04, 1.4282E+04, 1.4375E+04,
     +  1.4468E+04, 1.4561E+04, 1.4654E+04, 1.4748E+04, 1.4843E+04,
     + 1.4937E+04, 1.5032E+04, 1.5127E+04, 1.5222E+04, 1.5318E+04,
     +  1.5414E+04, 1.5511E+04, 1.5607E+04, 1.5704E+04, 1.5802E+04,
     + 1.5899E+04, 1.5997E+04, 1.6096E+04, 1.6194E+04, 1.6293E+04,
     +  1.6393E+04, 1.6492E+04, 1.6592E+04, 1.6693E+04, 1.6793E+04,
     + 1.6894E+04, 1.6996E+04, 1.7097E+04, 1.7199E+04, 1.7302E+04,
     +  1.7404E+04, 1.7507E+04, 1.7611E+04, 1.7714E+04, 1.7819E+04,
     + 1.7923E+04, 1.8028E+04, 1.8133E+04, 1.8238E+04, 1.8344E+04,
     +  1.8450E+04, 1.8556E+04, 1.8663E+04, 1.8770E+04, 1.8878E+04,
     + 1.8986E+04, 1.9094E+04, 1.9202E+04, 1.9311E+04, 1.9421E+04,
     +  1.9530E+04, 1.9640E+04, 1.9751E+04, 1.9861E+04, 1.9972E+04,
     + 2.0084E+04, 2.0196E+04, 2.0308E+04, 2.0420E+04, 2.0533E+04,
     +  2.0647E+04, 2.0760E+04, 2.0874E+04, 2.0989E+04, 2.1103E+04,
     + 2.1218E+04, 2.1334E+04, 2.1450E+04, 2.1566E+04, 2.1683E+04,
     +  2.1800E+04, 2.1917E+04, 2.2035E+04, 2.2153E+04, 2.2271E+04,
     + 2.2390E+04, 2.2510E+04, 2.2629E+04, 2.2749E+04, 2.2870E+04,
     +  2.2991E+04, 2.3112E+04, 2.3233E+04, 2.3355E+04, 2.3478E+04,
     + 2.3601E+04, 2.3724E+04, 2.3847E+04, 2.3971E+04, 2.4096E+04,
     +  2.4221E+04, 2.4346E+04, 2.4471E+04, 2.4597E+04, 2.4724E+04,
     + 2.4851E+04, 2.4978E+04, 2.5105E+04, 2.5233E+04, 2.5362E+04,
     +  2.5491E+04, 2.5620E+04, 2.5750E+04, 2.5880E+04, 2.6010E+04,
     + 2.6141E+04, 2.6272E+04, 2.6404E+04, 2.6536E+04, 2.6669E+04,
     +  2.6802E+04, 2.6936E+04, 2.7069E+04, 2.7204E+04, 2.7338E+04,
     + 2.7474E+04, 2.7609E+04, 2.7745E+04, 2.7882E+04, 2.8018E+04,
     +  2.8156E+04, 2.8294E+04, 2.8432E+04, 2.8570E+04, 2.8709E+04,
     + 2.8849E+04, 2.8989E+04, 2.9129E+04, 2.9270E+04, 2.9411E+04,
     +  2.9553E+04, 2.9695E+04, 2.9838E+04, 2.9981E+04, 3.0125E+04,
     + 3.0269E+04, 3.0413E+04, 3.0558E+04, 3.0703E+04, 3.0849E+04,
     +  3.0995E+04, 3.1142E+04, 3.1289E+04, 3.1437E+04, 3.1585E+04,
     + 3.1734E+04, 3.1883E+04, 3.2032E+04, 3.2182E+04, 3.2333E+04,
     +  3.2484E+04, 3.2635E+04, 3.2787E+04, 3.2939E+04, 3.3092E+04,
     + 3.3246E+04, 3.3399E+04, 3.3554E+04, 3.3709E+04, 3.3864E+04,
     +  3.4020E+04, 3.4176E+04, 3.4332E+04, 3.4490E+04, 3.4647E+04,
     + 3.4806E+04, 3.4964E+04, 3.5123E+04, 3.5283E+04, 3.5443E+04,
     +  3.5604E+04, 3.5765E+04, 3.5927E+04, 3.6089E+04, 3.6252E+04,
     + 3.6415E+04, 3.6579E+04, 3.6743E+04, 3.6907E+04, 3.7073E+04,
     +  3.7238E+04, 3.7405E+04, 3.7571E+04, 3.7739E+04, 3.7906E+04,
     + 3.8075E+04, 3.8244E+04, 3.8413E+04, 3.8583E+04, 3.8753E+04,
     +  3.8924E+04, 3.9096E+04, 3.9268E+04, 3.9440E+04, 3.9613E+04,
     + 3.9787E+04, 3.9961E+04, 4.0136E+04, 4.0311E+04, 4.0487E+04,
     +  4.0663E+04, 4.0840E+04, 4.1017E+04, 4.1195E+04, 4.1373E+04,
     + 4.1552E+04, 4.1732E+04, 4.1912E+04, 4.2093E+04, 4.2274E+04,
     +  4.2456E+04, 4.2638E+04, 4.2821E+04, 4.3004E+04, 4.3188E+04,
     + 4.3373E+04, 4.3558E+04, 4.3744E+04, 4.3930E+04, 4.4117E+04,
     +  4.4304E+04, 4.4492E+04, 4.4681E+04, 4.4870E+04, 4.5060E+04,
     + 4.5250E+04, 4.5441E+04, 4.5632E+04, 4.5824E+04, 4.6017E+04,
     +  4.6210E+04, 4.6404E+04, 4.6598E+04, 4.6793E+04, 4.6988E+04,
     + 4.7184E+04, 4.7381E+04, 4.7578E+04, 4.7776E+04, 4.7975E+04,
     +  4.8174E+04, 4.8374E+04, 4.8574E+04, 4.8775E+04, 4.8976E+04,
     + 4.9178E+04, 4.9381E+04, 4.9584E+04, 4.9788E+04, 4.9993E+04,
     +  5.0198E+04, 5.0404E+04, 5.0610E+04, 5.0817E+04, 5.1025E+04,
     + 5.1233E+04, 5.1442E+04, 5.1652E+04, 5.1862E+04, 5.2073E+04,
     +  5.2284E+04, 5.2496E+04, 5.2709E+04, 5.2922E+04, 5.3136E+04,
     + 5.3351E+04, 5.3566E+04, 5.3782E+04, 5.3998E+04, 5.4215E+04,
     +  5.4433E+04, 5.4652E+04, 5.4871E+04, 5.5090E+04, 5.5311E+04,
     + 5.5532E+04, 5.5754E+04, 5.5976E+04, 5.6199E+04, 5.6423E+04,
     +  5.6647E+04, 5.6872E+04, 5.7098E+04, 5.7324E+04, 5.7551E+04,
     + 5.7779E+04, 5.8007E+04, 5.8236E+04, 5.8466E+04, 5.8696E+04,
     +  5.8927E+04, 5.9159E+04, 5.9391E+04, 5.9624E+04, 5.9858E+04,
     + 6.0093E+04, 6.0328E+04, 6.0564E+04, 6.0800E+04, 6.1037E+04,
     +  6.1275E+04, 6.1514E+04, 6.1753E+04, 6.1993E+04, 6.2234E+04,
     + 6.2475E+04, 6.2718E+04, 6.2960E+04, 6.3204E+04, 6.3448E+04,
     +  6.3693E+04, 6.3939E+04, 6.4185E+04, 6.4432E+04, 6.4680E+04,
     + 6.4929E+04, 6.5178E+04, 6.5428E+04, 6.5679E+04, 6.5930E+04,
     +  6.6182E+04, 6.6435E+04, 6.6689E+04, 6.6943E+04, 6.7199E+04,
     + 6.7454E+04, 6.7711E+04, 6.7968E+04, 6.8226E+04, 6.8485E+04,
     +  6.8745E+04, 6.9005E+04, 6.9266E+04, 6.9528E+04, 6.9791E+04,
     + 7.0054E+04, 7.0318E+04, 7.0583E+04, 7.0848E+04, 7.1115E+04,
     +  7.1382E+04, 7.1650E+04, 7.1918E+04, 7.2188E+04, 7.2458E+04,
     + 7.2729E+04, 7.3001E+04, 7.3273E+04, 7.3547E+04, 7.3821E+04,
     +  7.4095E+04, 7.4371E+04, 7.4648E+04, 7.4925E+04, 7.5203E+04,
     + 7.5482E+04, 7.5761E+04, 7.6042E+04, 7.6323E+04, 7.6605E+04,
     +  7.6887E+04, 7.7171E+04, 7.7455E+04, 7.7741E+04, 7.8027E+04,
     + 7.8313E+04, 7.8601E+04, 7.8889E+04, 7.9179E+04, 7.9469E+04,
     +  7.9760E+04, 8.0051E+04, 8.0344E+04, 8.0637E+04, 8.0931E+04,
     + 8.1226E+04, 8.1522E+04, 8.1819E+04, 8.2116E+04, 8.2415E+04,
     +  8.2714E+04, 8.3014E+04, 8.3314E+04, 8.3616E+04, 8.3919E+04,
     + 8.4222E+04, 8.4526E+04, 8.4831E+04, 8.5137E+04, 8.5444E+04,
     +  8.5752E+04, 8.6060E+04, 8.6369E+04, 8.6680E+04, 8.6991E+04,
     + 8.7303E+04, 8.7615E+04, 8.7929E+04, 8.8244E+04, 8.8559E+04,
     +  8.8875E+04, 8.9192E+04, 8.9510E+04, 8.9829E+04, 9.0149E+04,
     + 9.0470E+04, 9.0791E+04, 9.1114E+04, 9.1437E+04, 9.1761E+04,
     +  9.2087E+04, 9.2413E+04, 9.2739E+04, 9.3067E+04, 9.3396E+04,
     + 9.3726E+04, 9.4056E+04, 9.4388E+04, 9.4720E+04, 9.5053E+04,
     +  9.5387E+04, 9.5722E+04, 9.6058E+04, 9.6395E+04, 9.6733E+04,
     + 9.7072E+04, 9.7412E+04, 9.7752E+04, 9.8094E+04, 9.8436E+04,
     +  9.8780E+04, 9.9124E+04, 9.9469E+04, 9.9816E+04, 1.0016E+05,
     + 1.0051E+05, 1.0086E+05, 1.0121E+05, 1.0156E+05, 1.0191E+05,
     +  1.0227E+05, 1.0262E+05, 1.0297E+05, 1.0333E+05, 1.0369E+05,
     + 1.0404E+05, 1.0440E+05, 1.0476E+05, 1.0512E+05, 1.0548E+05,
     +  1.0585E+05, 1.0621E+05, 1.0658E+05, 1.0694E+05, 1.0731E+05,
     + 1.0767E+05, 1.0804E+05, 1.0841E+05, 1.0878E+05, 1.0915E+05,
     +  1.0953E+05, 1.0990E+05, 1.1027E+05, 1.1065E+05, 1.1103E+05,
     + 1.1140E+05, 1.1178E+05, 1.1216E+05, 1.1254E+05, 1.1292E+05,
     +  1.1330E+05, 1.1369E+05, 1.1407E+05, 1.1446E+05, 1.1484E+05,
     + 1.1523E+05, 1.1562E+05, 1.1601E+05, 1.1640E+05, 1.1679E+05,
     +  1.1719E+05, 1.1758E+05, 1.1797E+05, 1.1837E+05, 1.1877E+05,
     + 1.1916E+05, 1.1956E+05, 1.1996E+05, 1.2036E+05, 1.2077E+05,
     +  1.2117E+05, 1.2157E+05, 1.2198E+05, 1.2239E+05, 1.2279E+05,
     + 1.2320E+05, 1.2361E+05, 1.2402E+05, 1.2443E+05, 1.2485E+05,
     +  1.2526E+05, 1.2568E+05, 1.2609E+05, 1.2651E+05, 1.2693E+05,
     + 1.2735E+05, 1.2777E+05, 1.2819E+05, 1.2861E+05, 1.2903E+05,
     +  1.2946E+05, 1.2988E+05, 1.3031E+05, 1.3074E+05, 1.3117E+05,
     + 1.3160E+05, 1.3203E+05, 1.3246E+05, 1.3290E+05, 1.3333E+05,
     +  1.3377E+05, 1.3421E+05, 1.3464E+05, 1.3508E+05, 1.3552E+05,
     + 1.3596E+05, 1.3641E+05, 1.3685E+05, 1.3730E+05, 1.3774E+05,
     +  1.3819E+05, 1.3864E+05, 1.3909E+05, 1.3954E+05, 1.3999E+05,
     + 1.4044E+05, 1.4090E+05, 1.4135E+05, 1.4181E+05, 1.4227E+05,
     +  1.4273E+05, 1.4319E+05, 1.4365E+05, 1.4411E+05, 1.4457E+05,
     + 1.4504E+05, 1.4550E+05, 1.4597E+05, 1.4644E+05, 1.4691E+05,
     +  1.4738E+05, 1.4785E+05, 1.4833E+05, 1.4880E+05, 1.4928E+05,
     + 1.4975E+05, 1.5023E+05, 1.5071E+05, 1.5119E+05, 1.5167E+05,
     +  1.5215E+05, 1.5264E+05, 1.5312E+05, 1.5361E+05, 1.5410E+05,
     + 1.5459E+05, 1.5508E+05, 1.5557E+05, 1.5606E+05, 1.5655E+05,
     +  1.5705E+05, 1.5755E+05, 1.5804E+05, 1.5854E+05, 1.5904E+05,
     + 1.5954E+05, 1.6005E+05, 1.6055E+05, 1.6106E+05, 1.6156E+05,
     +  1.6207E+05, 1.6258E+05, 1.6309E+05, 1.6360E+05, 1.6411E+05,
     + 1.6463E+05, 1.6514E+05, 1.6566E+05, 1.6618E+05, 1.6670E+05,
     +  1.6722E+05, 1.6774E+05, 1.6826E+05, 1.6879E+05, 1.6931E+05,
     + 1.6984E+05, 1.7037E+05, 1.7090E+05, 1.7143E+05, 1.7196E+05,
     +  1.7249E+05, 1.7303E+05, 1.7356E+05, 1.7410E+05, 1.7464E+05,
     + 1.7518E+05, 1.7572E+05, 1.7626E+05, 1.7681E+05, 1.7735E+05,
     +  1.7790E+05, 1.7845E+05, 1.7900E+05, 1.7955E+05, 1.8010E+05,
     + 1.8065E+05, 1.8121E+05, 1.8177E+05, 1.8232E+05, 1.8288E+05,
     +  1.8344E+05, 1.8400E+05, 1.8457E+05, 1.8513E+05, 1.8570E+05,
     + 1.8627E+05, 1.8683E+05, 1.8740E+05, 1.8797E+05, 1.8855E+05,
     +  1.8912E+05, 1.8970E+05, 1.9027E+05, 1.9085E+05, 1.9143E+05,
     + 1.9201E+05, 1.9260E+05, 1.9318E+05, 1.9377E+05, 1.9435E+05,
     +  1.9494E+05, 1.9553E+05, 1.9612E+05, 1.9671E+05, 1.9731E+05,
     + 1.9790E+05, 1.9850E+05, 1.9910E+05, 1.9970E+05, 2.0030E+05,
     +  2.0090E+05, 2.0150E+05, 2.0211E+05, 2.0272E+05, 2.0332E+05,
     + 2.0393E+05, 2.0455E+05, 2.0516E+05, 2.0577E+05, 2.0639E+05,
     +  2.0701E+05, 2.0762E+05, 2.0824E+05, 2.0887E+05, 2.0949E+05,
     + 2.1011E+05, 2.1074E+05, 2.1137E+05, 2.1199E+05, 2.1263E+05,
     +  2.1326E+05, 2.1389E+05, 2.1453E+05, 2.1516E+05, 2.1580E+05,
     + 2.1644E+05, 2.1708E+05, 2.1772E+05, 2.1837E+05, 2.1901E+05,
     +  2.1966E+05, 2.2031E+05, 2.2096E+05, 2.2161E+05, 2.2226E+05,
     + 2.2292E+05, 2.2357E+05, 2.2423E+05, 2.2489E+05, 2.2555E+05,
     +  2.2621E+05, 2.2688E+05, 2.2754E+05, 2.2821E+05, 2.2888E+05,
     + 2.2955E+05, 2.3022E+05, 2.3089E+05, 2.3157E+05, 2.3225E+05,
     +  2.3292E+05, 2.3360E+05, 2.3429E+05, 2.3497E+05, 2.3565E+05,
     + 2.3634E+05, 2.3703E+05, 2.3772E+05, 2.3841E+05, 2.3910E+05,
     +  2.3979E+05, 2.4049E+05, 2.4119E+05, 2.4189E+05, 2.4259E+05,
     + 2.4329E+05, 2.4399E+05, 2.4470E+05, 2.4540E+05, 2.4611E+05,
     +  2.4682E+05, 2.4754E+05, 2.4825E+05, 2.4896E+05, 2.4968E+05,
     + 2.5040E+05, 2.5112E+05, 2.5184E+05, 2.5257E+05, 2.5329E+05,
     +  2.5402E+05, 2.5475E+05, 2.5548E+05, 2.5621E+05, 2.5694E+05,
     + 2.5768E+05, 2.5841E+05, 2.5915E+05, 2.5989E+05, 2.6063E+05,
     +  2.6138E+05, 2.6212E+05, 2.6287E+05, 2.6362E+05, 2.6437E+05,
     + 2.6512E+05, 2.6588E+05, 2.6663E+05, 2.6739E+05, 2.6815E+05,
     +  2.6891E+05, 2.6967E+05, 2.7044E+05, 2.7120E+05, 2.7197E+05,
     + 2.7274E+05, 2.7351E+05, 2.7428E+05, 2.7506E+05, 2.7583E+05,
     +  2.7661E+05, 2.7739E+05, 2.7817E+05, 2.7896E+05, 2.7974E+05,
     + 2.8053E+05, 2.8132E+05, 2.8211E+05, 2.8290E+05, 2.8370E+05,
     +  2.8449E+05, 2.8529E+05, 2.8609E+05, 2.8689E+05, 2.8769E+05,
     + 2.8850E+05, 2.8931E+05, 2.9011E+05, 2.9092E+05, 2.9174E+05,
     +  2.9255E+05, 2.9337E+05, 2.9418E+05, 2.9500E+05, 2.9582E+05,
     + 2.9665E+05, 2.9747E+05, 2.9830E+05, 2.9913E+05, 2.9996E+05,
     +  3.0079E+05, 3.0163E+05, 3.0246E+05, 3.0330E+05, 3.0414E+05,
     + 3.0498E+05, 3.0582E+05, 3.0667E+05, 3.0752E+05, 3.0837E+05,
     +  3.0922E+05, 3.1007E+05, 3.1092E+05, 3.1178E+05, 3.1264E+05,
     + 3.1350E+05, 3.1436E+05, 3.1523E+05, 3.1609E+05, 3.1696E+05,
     +  3.1783E+05, 3.1870E+05, 3.1958E+05, 3.2045E+05, 3.2133E+05,
     + 3.2221E+05, 3.2309E+05, 3.2397E+05, 3.2486E+05, 3.2575E+05,
     +  3.2664E+05, 3.2753E+05, 3.2842E+05, 3.2932E+05, 3.3021E+05,
     + 3.3111E+05, 3.3201E+05, 3.3292E+05, 3.3382E+05, 3.3473E+05,
     +  3.3564E+05, 3.3655E+05, 3.3746E+05, 3.3837E+05, 3.3929E+05,
     + 3.4021E+05, 3.4113E+05, 3.4205E+05, 3.4298E+05, 3.4391E+05,
     +  3.4483E+05, 3.4577E+05, 3.4670E+05, 3.4763E+05, 3.4857E+05,
     + 3.4951E+05, 3.5045E+05, 3.5139E+05, 3.5234E+05, 3.5328E+05,
     +  3.5423E+05, 3.5519E+05, 3.5614E+05, 3.5709E+05, 3.5805E+05,
     + 3.5901E+05, 3.5997E+05, 3.6094E+05, 3.6190E+05, 3.6287E+05,
     +  3.6384E+05, 3.6481E+05, 3.6578E+05, 3.6676E+05, 3.6774E+05,
     + 3.6872E+05, 3.6970E+05, 3.7068E+05, 3.7167E+05, 3.7266E+05,
     +  3.7365E+05, 3.7464E+05, 3.7564E+05, 3.7663E+05, 3.7763E+05,
     + 3.7863E+05, 3.7964E+05, 3.8064E+05, 3.8165E+05, 3.8266E+05,
     +  3.8367E+05, 3.8469E+05, 3.8570E+05, 3.8672E+05, 3.8774E+05,
     + 3.8877E+05, 3.8979E+05, 3.9082E+05, 3.9185E+05, 3.9288E+05,
     +  3.9391E+05, 3.9495E+05, 3.9599E+05, 3.9703E+05, 3.9807E+05,
     + 3.9911E+05, 4.0016E+05, 4.0121E+05, 4.0226E+05, 4.0331E+05,
     +  4.0437E+05, 4.0543E+05, 4.0649E+05, 4.0755E+05, 4.0861E+05,
     + 4.0968E+05, 4.1075E+05, 4.1182E+05, 4.1289E+05, 4.1397E+05,
     +  4.1505E+05, 4.1613E+05, 4.1721E+05, 4.1829E+05, 4.1938E+05,
     + 4.2047E+05, 4.2156E+05, 4.2266E+05, 4.2375E+05, 4.2485E+05,
     +  4.2595E+05, 4.2705E+05, 4.2816E+05, 4.2927E+05, 4.3038E+05,
     + 4.3149E+05, 4.3260E+05, 4.3372E+05, 4.3484E+05, 4.3596E+05,
     +  4.3708E+05, 4.3821E+05, 4.3934E+05, 4.4047E+05, 4.4160E+05,
     + 4.4274E+05, 4.4387E+05, 4.4501E+05, 4.4616E+05, 4.4730E+05,
     +  4.4845E+05, 4.4960E+05, 4.5075E+05, 4.5190E+05, 4.5306E+05,
     + 4.5422E+05, 4.5538E+05, 4.5655E+05, 4.5771E+05, 4.5888E+05,
     +  4.6005E+05, 4.6122E+05, 4.6240E+05, 4.6358E+05, 4.6476E+05,
     + 4.6594E+05, 4.6713E+05, 4.6831E+05, 4.6950E+05, 4.7070E+05,
     +  4.7189E+05, 4.7309E+05, 4.7429E+05, 4.7549E+05, 4.7670E+05,
     + 4.7790E+05, 4.7911E+05, 4.8033E+05, 4.8154E+05, 4.8276E+05,
     +  4.8398E+05, 4.8520E+05, 4.8643E+05, 4.8765E+05, 4.8888E+05,
     + 4.9011E+05, 4.9135E+05, 4.9259E+05, 4.9383E+05, 4.9507E+05,
     +  4.9631E+05, 4.9756E+05, 4.9881E+05, 5.0006E+05, 5.0132E+05,
     + 5.0257E+05, 5.0383E+05, 5.0509E+05, 5.0636E+05, 5.0763E+05,
     +  5.0890E+05, 5.1017E+05, 5.1144E+05, 5.1272E+05, 5.1400E+05,
     + 5.1528E+05, 5.1657E+05, 5.1786E+05, 5.1915E+05, 5.2044E+05,
     +  5.2174E+05, 5.2304E+05, 5.2434E+05, 5.2564E+05, 5.2695E+05,
     + 5.2825E+05, 5.2957E+05, 5.3088E+05, 5.3220E+05, 5.3352E+05,
     +  5.3484E+05, 5.3616E+05, 5.3749E+05, 5.3882E+05, 5.4015E+05,
     + 5.4148E+05, 5.4282E+05, 5.4416E+05, 5.4551E+05, 5.4685E+05,
     +  5.4820E+05, 5.4955E+05, 5.5090E+05, 5.5226E+05, 5.5362E+05,
     + 5.5498E+05, 5.5635E+05, 5.5771E+05, 5.5908E+05, 5.6045E+05,
     +  5.6183E+05, 5.6321E+05, 5.6459E+05, 5.6597E+05, 5.6736E+05,
     + 5.6875E+05, 5.7014E+05, 5.7153E+05, 5.7293E+05, 5.7433E+05,
     +  5.7573E+05, 5.7714E+05, 5.7854E+05, 5.7996E+05, 5.8137E+05,
     + 5.8278E+05, 5.8420E+05, 5.8563E+05, 5.8705E+05, 5.8848E+05,
     +  5.8991E+05, 5.9134E+05, 5.9278E+05, 5.9422E+05, 5.9566E+05,
     + 5.9710E+05, 5.9855E+05, 6.0000E+05, 6.0145E+05, 6.0291E+05,
     +  6.0436E+05, 6.0583E+05, 6.0729E+05, 6.0876E+05, 6.1023E+05,
     + 6.1170E+05, 6.1317E+05, 6.1465E+05, 6.1613E+05, 6.1762E+05,
     +  6.1910E+05, 6.2059E+05, 6.2209E+05, 6.2358E+05, 6.2508E+05,
     + 6.2658E+05, 6.2809E+05, 6.2959E+05, 6.3110E+05, 6.3262E+05,
     +  6.3413E+05, 6.3565E+05, 6.3717E+05, 6.3870E+05, 6.4023E+05,
     + 6.4176E+05, 6.4329E+05, 6.4483E+05, 6.4636E+05, 6.4791E+05,
     +  6.4945E+05, 6.5100E+05, 6.5255E+05, 6.5411E+05, 6.5566E+05,
     + 6.5722E+05, 6.5879E+05, 6.6035E+05, 6.6192E+05, 6.6349E+05,
     +  6.6507E+05, 6.6665E+05, 6.6823E+05, 6.6981E+05, 6.7140E+05,
     + 6.7299E+05, 6.7458E+05, 6.7618E+05, 6.7778E+05, 6.7938E+05,
     +  6.8099E+05, 6.8260E+05, 6.8421E+05, 6.8582E+05, 6.8744E+05,
     + 6.8906E+05, 6.9068E+05, 6.9231E+05, 6.9394E+05, 6.9557E+05,
     +  6.9721E+05, 6.9885E+05, 7.0049E+05, 7.0214E+05, 7.0378E+05,
     + 7.0544E+05, 7.0709E+05, 7.0875E+05, 7.1041E+05, 7.1207E+05,
     +  7.1374E+05, 7.1541E+05, 7.1709E+05, 7.1876E+05, 7.2044E+05,
     + 7.2213E+05, 7.2381E+05, 7.2550E+05, 7.2719E+05, 7.2889E+05,
     +  7.3059E+05, 7.3229E+05, 7.3400E+05, 7.3570E+05, 7.3742E+05,
     + 7.3913E+05, 7.4085E+05, 7.4257E+05, 7.4429E+05, 7.4602E+05,
     +  7.4775E+05, 7.4949E+05, 7.5122E+05, 7.5297E+05, 7.5471E+05,
     + 7.5646E+05, 7.5821E+05, 7.5996E+05, 7.6172E+05, 7.6348E+05,
     +  7.6524E+05, 7.6701E+05, 7.6878E+05, 7.7055E+05, 7.7233E+05,
     + 7.7411E+05, 7.7589E+05, 7.7768E+05, 7.7947E+05, 7.8126E+05,
     +  7.8306E+05, 7.8486E+05, 7.8666E+05, 7.8847E+05, 7.9028E+05,
     + 7.9209E+05, 7.9391E+05, 7.9573E+05, 7.9755E+05, 7.9938E+05,
     +  8.0121E+05, 8.0304E+05, 8.0488E+05, 8.0672E+05, 8.0856E+05,
     + 8.1041E+05, 8.1226E+05, 8.1411E+05, 8.1597E+05, 8.1783E+05,
     +  8.1969E+05, 8.2156E+05, 8.2343E+05, 8.2530E+05, 8.2718E+05,
     + 8.2906E+05, 8.3095E+05, 8.3284E+05, 8.3473E+05, 8.3662E+05,
     +  8.3852E+05, 8.4042E+05, 8.4233E+05, 8.4424E+05, 8.4615E+05,
     + 8.4806E+05, 8.4998E+05, 8.5191E+05, 8.5383E+05, 8.5576E+05,
     +  8.5769E+05, 8.5963E+05, 8.6157E+05, 8.6352E+05, 8.6546E+05,
     + 8.6741E+05, 8.6937E+05, 8.7133E+05, 8.7329E+05, 8.7525E+05,
     +  8.7722E+05, 8.7919E+05, 8.8117E+05, 8.8315E+05, 8.8513E+05,
     + 8.8712E+05, 8.8911E+05, 8.9110E+05, 8.9310E+05, 8.9510E+05,
     +  8.9710E+05, 8.9911E+05, 9.0112E+05, 9.0314E+05, 9.0516E+05,
     + 9.0718E+05, 9.0921E+05, 9.1124E+05, 9.1327E+05, 9.1531E+05,
     +  9.1735E+05, 9.1939E+05, 9.2144E+05, 9.2349E+05, 9.2555E+05,
     + 9.2761E+05, 9.2967E+05, 9.3174E+05, 9.3381E+05, 9.3588E+05,
     +  9.3796E+05, 9.4004E+05, 9.4212E+05, 9.4421E+05, 9.4630E+05,
     + 9.4840E+05, 9.5050E+05, 9.5260E+05, 9.5471E+05, 9.5682E+05,
     +  9.5894E+05, 9.6106E+05, 9.6318E+05, 9.6531E+05, 9.6744E+05,
     + 9.6957E+05, 9.7171E+05, 9.7385E+05, 9.7599E+05, 9.7814E+05,
     +  9.8030E+05, 9.8245E+05, 9.8461E+05, 9.8678E+05, 9.8895E+05,
     + 9.9112E+05, 9.9329E+05, 9.9547E+05, 9.9766E+05, 9.9984E+05,
     +  1.0020E+06, 1.0042E+06, 1.0064E+06, 1.0086E+06, 1.0108E+06,
     + 1.0131E+06, 1.0153E+06, 1.0175E+06, 1.0197E+06, 1.0219E+06,
     +  1.0242E+06, 1.0264E+06, 1.0286E+06, 1.0309E+06, 1.0331E+06,
     + 1.0354E+06, 1.0376E+06, 1.0399E+06, 1.0421E+06, 1.0444E+06,
     +  1.0467E+06, 1.0490E+06, 1.0512E+06, 1.0535E+06, 1.0558E+06,
     + 1.0581E+06, 1.0604E+06, 1.0627E+06, 1.0650E+06, 1.0673E+06,
     +  1.0696E+06, 1.0719E+06, 1.0742E+06, 1.0765E+06, 1.0789E+06,
     + 1.0812E+06, 1.0835E+06, 1.0859E+06, 1.0882E+06, 1.0906E+06,
     +  1.0929E+06, 1.0953E+06, 1.0976E+06, 1.1000E+06, 1.1024E+06,
     + 1.1047E+06, 1.1071E+06, 1.1095E+06, 1.1119E+06, 1.1142E+06,
     +  1.1166E+06, 1.1190E+06, 1.1214E+06, 1.1238E+06, 1.1262E+06,
     + 1.1287E+06, 1.1311E+06, 1.1335E+06, 1.1359E+06, 1.1383E+06,
     +  1.1408E+06, 1.1432E+06, 1.1457E+06, 1.1481E+06, 1.1505E+06,
     + 1.1530E+06, 1.1555E+06, 1.1579E+06, 1.1604E+06, 1.1629E+06,
     +  1.1653E+06, 1.1678E+06, 1.1703E+06, 1.1728E+06, 1.1753E+06,
     + 1.1778E+06, 1.1803E+06, 1.1828E+06, 1.1853E+06, 1.1878E+06,
     +  1.1903E+06, 1.1928E+06, 1.1954E+06, 1.1979E+06, 1.2004E+06,
     + 1.2030E+06, 1.2055E+06, 1.2081E+06, 1.2106E+06, 1.2132E+06,
     +  1.2157E+06, 1.2183E+06, 1.2209E+06, 1.2234E+06, 1.2260E+06,
     + 1.2286E+06, 1.2312E+06, 1.2338E+06, 1.2364E+06, 1.2390E+06,
     +  1.2416E+06, 1.2442E+06, 1.2468E+06, 1.2494E+06, 1.2520E+06,
     + 1.2547E+06, 1.2573E+06, 1.2599E+06, 1.2626E+06, 1.2652E+06,
     +  1.2679E+06, 1.2705E+06, 1.2732E+06, 1.2758E+06, 1.2785E+06,
     + 1.2812E+06, 1.2838E+06, 1.2865E+06, 1.2892E+06, 1.2919E+06,
     +  1.2946E+06, 1.2973E+06, 1.3000E+06, 1.3027E+06, 1.3054E+06,
     + 1.3081E+06, 1.3109E+06, 1.3136E+06, 1.3163E+06, 1.3190E+06,
     +  1.3218E+06, 1.3245E+06, 1.3273E+06, 1.3300E+06, 1.3328E+06,
     + 1.3355E+06, 1.3383E+06, 1.3411E+06, 1.3439E+06, 1.3466E+06,
     +  1.3494E+06, 1.3522E+06, 1.3550E+06, 1.3578E+06, 1.3606E+06,
     + 1.3634E+06, 1.3662E+06, 1.3691E+06, 1.3719E+06, 1.3747E+06,
     +  1.3775E+06, 1.3804E+06, 1.3832E+06, 1.3861E+06, 1.3889E+06,
     + 1.3918E+06, 1.3946E+06, 1.3975E+06, 1.4004E+06, 1.4032E+06,
     +  1.4061E+06, 1.4090E+06, 1.4119E+06, 1.4148E+06, 1.4177E+06,
     + 1.4206E+06, 1.4235E+06, 1.4264E+06, 1.4293E+06, 1.4322E+06,
     +  1.4352E+06, 1.4381E+06, 1.4410E+06, 1.4440E+06, 1.4469E+06,
     + 1.4499E+06, 1.4528E+06, 1.4558E+06, 1.4588E+06, 1.4617E+06,
     +  1.4647E+06, 1.4677E+06, 1.4707E+06, 1.4737E+06, 1.4767E+06,
     + 1.4797E+06, 1.4827E+06, 1.4857E+06, 1.4887E+06, 1.4917E+06,
     +  1.4947E+06, 1.4978E+06, 1.5008E+06, 1.5038E+06, 1.5069E+06,
     + 1.5099E+06, 1.5130E+06, 1.5160E+06, 1.5191E+06, 1.5222E+06,
     +  1.5252E+06, 1.5283E+06, 1.5314E+06, 1.5345E+06, 1.5376E+06,
     + 1.5407E+06, 1.5438E+06, 1.5469E+06, 1.5500E+06, 1.5531E+06,
     +  1.5563E+06, 1.5594E+06, 1.5625E+06, 1.5657E+06, 1.5688E+06,
     + 1.5720E+06, 1.5751E+06, 1.5783E+06, 1.5814E+06, 1.5846E+06,
     +  1.5878E+06, 1.5910E+06, 1.5942E+06, 1.5973E+06, 1.6005E+06,
     + 1.6037E+06, 1.6069E+06, 1.6102E+06, 1.6134E+06, 1.6166E+06,
     +  1.6198E+06, 1.6231E+06, 1.6263E+06, 1.6295E+06, 1.6328E+06,
     + 1.6360E+06, 1.6393E+06, 1.6426E+06, 1.6458E+06, 1.6491E+06,
     +  1.6524E+06, 1.6557E+06, 1.6589E+06, 1.6622E+06, 1.6655E+06,
     + 1.6688E+06, 1.6722E+06, 1.6755E+06, 1.6788E+06, 1.6821E+06,
     +  1.6855E+06, 1.6888E+06, 1.6921E+06, 1.6955E+06, 1.6988E+06,
     + 1.7022E+06, 1.7056E+06, 1.7089E+06, 1.7123E+06, 1.7157E+06,
     +  1.7191E+06, 1.7225E+06, 1.7259E+06, 1.7293E+06, 1.7327E+06,
     + 1.7361E+06, 1.7395E+06, 1.7429E+06, 1.7463E+06, 1.7498E+06,
     +  1.7532E+06, 1.7567E+06, 1.7601E+06, 1.7636E+06, 1.7670E+06,
     + 1.7705E+06, 1.7740E+06, 1.7774E+06, 1.7809E+06, 1.7844E+06,
     +  1.7879E+06, 1.7914E+06, 1.7949E+06, 1.7984E+06, 1.8019E+06,
     + 1.8055E+06, 1.8090E+06, 1.8125E+06, 1.8161E+06, 1.8196E+06,
     +  1.8231E+06, 1.8267E+06, 1.8303E+06, 1.8338E+06, 1.8374E+06,
     + 1.8410E+06, 1.8446E+06, 1.8482E+06, 1.8517E+06, 1.8553E+06,
     +  1.8589E+06, 1.8626E+06, 1.8662E+06, 1.8698E+06, 1.8734E+06,
     + 1.8771E+06, 1.8807E+06, 1.8843E+06, 1.8880E+06, 1.8917E+06,
     +  1.8953E+06, 1.8990E+06, 1.9027E+06, 1.9063E+06, 1.9100E+06,
     + 1.9137E+06, 1.9174E+06, 1.9211E+06, 1.9248E+06, 1.9285E+06,
     +  1.9322E+06, 1.9360E+06, 1.9397E+06, 1.9434E+06, 1.9472E+06,
     + 1.9509E+06, 1.9547E+06, 1.9584E+06, 1.9622E+06, 1.9660E+06,
     +  1.9698E+06, 1.9735E+06, 1.9773E+06, 1.9811E+06, 1.9849E+06,
     + 1.9887E+06, 1.9925E+06, 1.9964E+06, 2.0002E+06, 2.0040E+06,
     +  2.0079E+06, 2.0117E+06, 2.0155E+06, 2.0194E+06, 2.0233E+06,
     + 2.0271E+06, 2.0310E+06, 2.0349E+06, 2.0388E+06, 2.0426E+06,
     +  2.0465E+06, 2.0504E+06, 2.0544E+06, 2.0583E+06, 2.0622E+06,
     + 2.0661E+06, 2.0700E+06, 2.0740E+06, 2.0779E+06, 2.0819E+06,
     +  2.0858E+06, 2.0898E+06, 2.0938E+06, 2.0977E+06, 2.1017E+06,
     + 2.1057E+06, 2.1097E+06, 2.1137E+06, 2.1177E+06, 2.1217E+06,
     +  2.1257E+06, 2.1297E+06, 2.1338E+06, 2.1378E+06, 2.1418E+06,
     + 2.1459E+06, 2.1499E+06, 2.1540E+06, 2.1581E+06, 2.1621E+06,
     +  2.1662E+06, 2.1703E+06, 2.1744E+06, 2.1785E+06, 2.1826E+06,
     + 2.1867E+06, 2.1908E+06, 2.1949E+06, 2.1991E+06, 2.2032E+06,
     +  2.2073E+06, 2.2115E+06, 2.2156E+06, 2.2198E+06, 2.2240E+06,
     + 2.2281E+06, 2.2323E+06, 2.2365E+06, 2.2407E+06, 2.2449E+06,
     +  2.2491E+06, 2.2533E+06, 2.2575E+06, 2.2617E+06, 2.2659E+06,
     + 2.2702E+06, 2.2744E+06, 2.2787E+06, 2.2829E+06, 2.2872E+06,
     +  2.2915E+06, 2.2957E+06, 2.3000E+06, 2.3043E+06, 2.3086E+06,
     + 2.3129E+06, 2.3172E+06, 2.3215E+06, 2.3258E+06, 2.3301E+06,
     +  2.3345E+06, 2.3388E+06, 2.3431E+06, 2.3475E+06, 2.3519E+06,
     + 2.3562E+06, 2.3606E+06, 2.3650E+06, 2.3693E+06, 2.3737E+06,
     +  2.3781E+06, 2.3825E+06, 2.3869E+06, 2.3914E+06, 2.3958E+06,
     + 2.4002E+06, 2.4046E+06, 2.4091E+06, 2.4135E+06, 2.4180E+06,
     +  2.4224E+06, 2.4269E+06, 2.4314E+06, 2.4359E+06, 2.4404E+06,
     + 2.4448E+06, 2.4494E+06, 2.4539E+06, 2.4584E+06, 2.4629E+06,
     +  2.4674E+06, 2.4720E+06, 2.4765E+06, 2.4810E+06, 2.4856E+06,
     + 2.4902E+06, 2.4947E+06, 2.4993E+06, 2.5039E+06, 2.5085E+06,
     +  2.5131E+06, 2.5177E+06, 2.5223E+06, 2.5269E+06, 2.5315E+06,
     + 2.5361E+06, 2.5408E+06, 2.5454E+06, 2.5501E+06, 2.5547E+06,
     +  2.5594E+06, 2.5641E+06, 2.5687E+06, 2.5734E+06, 2.5781E+06,
     + 2.5828E+06, 2.5875E+06, 2.5922E+06, 2.5969E+06, 2.6017E+06,
     +  2.6064E+06, 2.6111E+06, 2.6159E+06, 2.6206E+06, 2.6254E+06,
     + 2.6302E+06, 2.6349E+06, 2.6397E+06, 2.6445E+06, 2.6493E+06,
     +  2.6541E+06, 2.6589E+06, 2.6637E+06, 2.6685E+06, 2.6734E+06,
     + 2.6782E+06, 2.6830E+06, 2.6879E+06, 2.6927E+06, 2.6976E+06,
     +  2.7025E+06, 2.7074E+06, 2.7122E+06, 2.7171E+06, 2.7220E+06,
     + 2.7269E+06, 2.7319E+06, 2.7368E+06, 2.7417E+06, 2.7466E+06,
     +  2.7516E+06, 2.7565E+06, 2.7615E+06, 2.7664E+06, 2.7714E+06,
     + 2.7764E+06, 2.7814E+06, 2.7864E+06, 2.7914E+06, 2.7964E+06,
     +  2.8014E+06, 2.8064E+06, 2.8114E+06, 2.8165E+06, 2.8215E+06,
     + 2.8266E+06, 2.8316E+06, 2.8367E+06, 2.8418E+06, 2.8468E+06,
     +  2.8519E+06, 2.8570E+06, 2.8621E+06, 2.8672E+06, 2.8723E+06,
     + 2.8775E+06, 2.8826E+06, 2.8877E+06, 2.8929E+06, 2.8980E+06,
     +  2.9032E+06, 2.9084E+06, 2.9135E+06, 2.9187E+06, 2.9239E+06,
     + 2.9291E+06, 2.9343E+06, 2.9395E+06, 2.9447E+06, 2.9499E+06,
     +  2.9552E+06, 2.9604E+06, 2.9657E+06, 2.9709E+06, 2.9762E+06,
     + 2.9815E+06, 2.9867E+06, 2.9920E+06, 2.9973E+06, 3.0026E+06,
     +  3.0079E+06, 3.0132E+06, 3.0186E+06, 3.0239E+06, 3.0292E+06,
     + 3.0346E+06, 3.0399E+06, 3.0453E+06, 3.0507E+06, 3.0560E+06,
     +  3.0614E+06, 3.0668E+06, 3.0722E+06, 3.0776E+06, 3.0830E+06,
     + 3.0884E+06, 3.0939E+06, 3.0993E+06, 3.1048E+06, 3.1102E+06,
     +  3.1157E+06, 3.1211E+06, 3.1266E+06, 3.1321E+06, 3.1376E+06,
     + 3.1431E+06, 3.1486E+06, 3.1541E+06, 3.1596E+06, 3.1651E+06,
     +  3.1707E+06, 3.1762E+06, 3.1818E+06, 3.1873E+06, 3.1929E+06,
     + 3.1985E+06, 3.2041E+06, 3.2097E+06, 3.2153E+06, 3.2209E+06,
     +  3.2265E+06, 3.2321E+06, 3.2377E+06, 3.2434E+06, 3.2490E+06,
     + 3.2547E+06, 3.2603E+06, 3.2660E+06, 3.2717E+06, 3.2774E+06,
     +  3.2831E+06, 3.2888E+06, 3.2945E+06, 3.3002E+06, 3.3059E+06,
     + 3.3116E+06, 3.3174E+06, 3.3231E+06, 3.3289E+06, 3.3346E+06,
     +  3.3404E+06, 3.3462E+06, 3.3520E+06, 3.3578E+06, 3.3636E+06,
     + 3.3694E+06, 3.3752E+06, 3.3810E+06, 3.3869E+06, 3.3927E+06,
     +  3.3986E+06, 3.4044E+06, 3.4103E+06, 3.4162E+06, 3.4221E+06,
     + 3.4280E+06, 3.4339E+06, 3.4398E+06, 3.4457E+06, 3.4516E+06,
     +  3.4576E+06, 3.4635E+06, 3.4694E+06, 3.4754E+06, 3.4814E+06,
     + 3.4873E+06, 3.4933E+06, 3.4993E+06, 3.5053E+06, 3.5113E+06,
     +  3.5173E+06, 3.5234E+06, 3.5294E+06, 3.5354E+06, 3.5415E+06,
     + 3.5475E+06, 3.5536E+06, 3.5597E+06, 3.5658E+06, 3.5719E+06,
     +  3.5780E+06, 3.5841E+06, 3.5902E+06, 3.5963E+06, 3.6024E+06,
     + 3.6086E+06, 3.6147E+06, 3.6209E+06, 3.6270E+06, 3.6332E+06,
     +  3.6394E+06, 3.6456E+06, 3.6518E+06, 3.6580E+06, 3.6642E+06,
     + 3.6704E+06, 3.6767E+06, 3.6829E+06, 3.6892E+06, 3.6954E+06,
     +  3.7017E+06, 3.7080E+06, 3.7142E+06, 3.7205E+06, 3.7268E+06,
     + 3.7331E+06, 3.7395E+06, 3.7458E+06, 3.7521E+06, 3.7585E+06,
     +  3.7648E+06, 3.7712E+06, 3.7775E+06, 3.7839E+06, 3.7903E+06,
     + 3.7967E+06, 3.8031E+06, 3.8095E+06, 3.8159E+06, 3.8224E+06,
     +  3.8288E+06, 3.8352E+06, 3.8417E+06, 3.8482E+06, 3.8546E+06,
     + 3.8611E+06, 3.8676E+06, 3.8741E+06, 3.8806E+06, 3.8871E+06,
     +  3.8936E+06, 3.9002E+06, 3.9067E+06, 3.9133E+06, 3.9198E+06,
     + 3.9264E+06, 3.9330E+06, 3.9395E+06, 3.9461E+06, 3.9527E+06,
     +  3.9594E+06, 3.9660E+06, 3.9726E+06, 3.9792E+06, 3.9859E+06,
     + 3.9925E+06, 3.9992E+06, 4.0059E+06, 4.0126E+06, 4.0193E+06,
     +  4.0260E+06, 4.0327E+06, 4.0394E+06, 4.0461E+06, 4.0528E+06,
     + 4.0596E+06, 4.0663E+06, 4.0731E+06, 4.0799E+06, 4.0867E+06,
     +  4.0934E+06, 4.1002E+06, 4.1071E+06, 4.1139E+06, 4.1207E+06,
     + 4.1275E+06, 4.1344E+06, 4.1412E+06, 4.1481E+06, 4.1550E+06,
     +  4.1618E+06, 4.1687E+06, 4.1756E+06, 4.1825E+06, 4.1894E+06,
     + 4.1964E+06, 4.2033E+06, 4.2102E+06, 4.2172E+06, 4.2242E+06,
     +  4.2311E+06, 4.2381E+06, 4.2451E+06, 4.2521E+06, 4.2591E+06,
     + 4.2661E+06, 4.2731E+06, 4.2802E+06, 4.2872E+06, 4.2943E+06,
     +  4.3013E+06, 4.3084E+06, 4.3155E+06, 4.3226E+06, 4.3297E+06,
     + 4.3368E+06, 4.3439E+06, 4.3510E+06, 4.3582E+06, 4.3653E+06,
     +  4.3725E+06, 4.3796E+06, 4.3868E+06, 4.3940E+06, 4.4012E+06,
     + 4.4084E+06, 4.4156E+06, 4.4228E+06, 4.4300E+06, 4.4373E+06,
     +  4.4445E+06, 4.4518E+06, 4.4590E+06, 4.4663E+06, 4.4736E+06,
     + 4.4809E+06, 4.4882E+06, 4.4955E+06, 4.5028E+06, 4.5102E+06,
     +  4.5175E+06, 4.5249E+06, 4.5322E+06, 4.5396E+06, 4.5470E+06,
     + 4.5544E+06, 4.5618E+06, 4.5692E+06, 4.5766E+06, 4.5840E+06,
     +  4.5915E+06, 4.5989E+06, 4.6064E+06, 4.6138E+06, 4.6213E+06,
     + 4.6288E+06, 4.6363E+06, 4.6438E+06, 4.6513E+06, 4.6588E+06,
     +  4.6664E+06, 4.6739E+06, 4.6815E+06, 4.6890E+06, 4.6966E+06,
     + 4.7042E+06, 4.7118E+06, 4.7194E+06, 4.7270E+06, 4.7346E+06,
     +  4.7422E+06, 4.7499E+06, 4.7575E+06, 4.7652E+06, 4.7729E+06,
     + 4.7805E+06, 4.7882E+06, 4.7959E+06, 4.8036E+06, 4.8114E+06,
     +  4.8191E+06, 4.8268E+06, 4.8346E+06, 4.8423E+06, 4.8501E+06,
     + 4.8579E+06, 4.8657E+06, 4.8735E+06, 4.8813E+06, 4.8891E+06,
     +  4.8969E+06, 4.9047E+06, 4.9126E+06, 4.9205E+06, 4.9283E+06,
     + 4.9362E+06, 4.9441E+06, 4.9520E+06, 4.9599E+06, 4.9678E+06,
     +  4.9757E+06, 4.9837E+06, 4.9916E+06, 4.9996E+06, 5.0075E+06/
      data (QofT( 9,J),J=2409,4510)/ 5.0155E+06, 5.0235E+06,
     +  5.0315E+06, 5.0395E+06, 5.0475E+06, 5.0555E+06, 5.0636E+06,
     +  5.0716E+06,
     + 5.0797E+06, 5.0878E+06, 5.0958E+06, 5.1039E+06, 5.1120E+06,
     +  5.1201E+06, 5.1282E+06, 5.1364E+06, 5.1445E+06, 5.1527E+06,
     + 5.1608E+06, 5.1690E+06, 5.1772E+06, 5.1854E+06, 5.1936E+06,
     +  5.2018E+06, 5.2100E+06, 5.2182E+06, 5.2265E+06, 5.2347E+06,
     + 5.2430E+06, 5.2512E+06, 5.2595E+06, 5.2678E+06, 5.2761E+06,
     +  5.2844E+06, 5.2928E+06, 5.3011E+06, 5.3094E+06, 5.3178E+06,
     + 5.3261E+06, 5.3345E+06, 5.3429E+06, 5.3513E+06, 5.3597E+06,
     +  5.3681E+06, 5.3766E+06, 5.3850E+06, 5.3934E+06, 5.4019E+06,
     + 5.4104E+06, 5.4188E+06, 5.4273E+06, 5.4358E+06, 5.4443E+06,
     +  5.4529E+06, 5.4614E+06, 5.4699E+06, 5.4785E+06, 5.4870E+06,
     + 5.4956E+06, 5.5042E+06, 5.5128E+06, 5.5214E+06, 5.5300E+06,
     +  5.5386E+06, 5.5473E+06, 5.5559E+06, 5.5646E+06, 5.5733E+06,
     + 5.5819E+06, 5.5906E+06, 5.5993E+06, 5.6080E+06, 5.6168E+06,
     +  5.6255E+06, 5.6342E+06, 5.6430E+06, 5.6518E+06, 5.6605E+06,
     + 5.6693E+06, 5.6781E+06, 5.6869E+06, 5.6957E+06, 5.7046E+06,
     +  5.7134E+06, 5.7223E+06, 5.7311E+06, 5.7400E+06, 5.7489E+06,
     + 5.7578E+06, 5.7667E+06, 5.7756E+06, 5.7845E+06, 5.7935E+06,
     +  5.8024E+06, 5.8114E+06, 5.8203E+06, 5.8293E+06, 5.8383E+06,
     + 5.8473E+06, 5.8563E+06, 5.8653E+06, 5.8744E+06, 5.8834E+06,
     +  5.8925E+06, 5.9016E+06, 5.9106E+06, 5.9197E+06, 5.9288E+06,
     + 5.9379E+06, 5.9471E+06, 5.9562E+06, 5.9653E+06, 5.9745E+06,
     +  5.9837E+06, 5.9928E+06, 6.0020E+06, 6.0112E+06, 6.0204E+06,
     + 6.0297E+06, 6.0389E+06, 6.0481E+06, 6.0574E+06, 6.0667E+06,
     +  6.0759E+06, 6.0852E+06, 6.0945E+06, 6.1038E+06, 6.1132E+06,
     + 6.1225E+06, 6.1318E+06, 6.1412E+06, 6.1506E+06, 6.1599E+06,
     +  6.1693E+06, 6.1787E+06, 6.1881E+06, 6.1976E+06, 6.2070E+06,
     + 6.2164E+06, 6.2259E+06, 6.2354E+06, 6.2448E+06, 6.2543E+06,
     +  6.2638E+06, 6.2734E+06, 6.2829E+06, 6.2924E+06, 6.3020E+06,
     + 6.3115E+06, 6.3211E+06, 6.3307E+06, 6.3403E+06, 6.3499E+06,
     +  6.3595E+06, 6.3691E+06, 6.3787E+06, 6.3884E+06, 6.3981E+06,
     + 6.4077E+06, 6.4174E+06, 6.4271E+06, 6.4368E+06, 6.4465E+06,
     +  6.4563E+06, 6.4660E+06, 6.4758E+06, 6.4855E+06, 6.4953E+06,
     + 6.5051E+06, 6.5149E+06, 6.5247E+06, 6.5345E+06, 6.5443E+06,
     +  6.5542E+06, 6.5640E+06, 6.5739E+06, 6.5838E+06, 6.5937E+06,
     + 6.6036E+06, 6.6135E+06, 6.6234E+06, 6.6334E+06, 6.6433E+06,
     +  6.6533E+06, 6.6632E+06, 6.6732E+06, 6.6832E+06, 6.6932E+06,
     + 6.7033E+06, 6.7133E+06, 6.7233E+06, 6.7334E+06, 6.7434E+06,
     +  6.7535E+06, 6.7636E+06, 6.7737E+06, 6.7838E+06, 6.7940E+06,
     + 6.8041E+06, 6.8142E+06, 6.8244E+06, 6.8346E+06, 6.8448E+06,
     +  6.8550E+06, 6.8652E+06, 6.8754E+06, 6.8856E+06, 6.8959E+06,
     + 6.9061E+06, 6.9164E+06, 6.9267E+06, 6.9369E+06, 6.9472E+06,
     +  6.9576E+06, 6.9679E+06, 6.9782E+06, 6.9886E+06, 6.9989E+06,
     + 7.0093E+06, 7.0197E+06, 7.0301E+06, 7.0405E+06, 7.0509E+06,
     +  7.0614E+06, 7.0718E+06, 7.0823E+06, 7.0927E+06, 7.1032E+06,
     + 7.1137E+06, 7.1242E+06, 7.1347E+06, 7.1453E+06, 7.1558E+06,
     +  7.1664E+06, 7.1769E+06, 7.1875E+06, 7.1981E+06, 7.2087E+06,
     + 7.2193E+06, 7.2300E+06, 7.2406E+06, 7.2512E+06, 7.2619E+06,
     +  7.2726E+06, 7.2833E+06, 7.2940E+06, 7.3047E+06, 7.3154E+06,
     + 7.3262E+06, 7.3369E+06, 7.3477E+06, 7.3584E+06, 7.3692E+06,
     +  7.3800E+06, 7.3908E+06, 7.4017E+06, 7.4125E+06, 7.4233E+06,
     + 7.4342E+06, 7.4451E+06, 7.4560E+06, 7.4669E+06, 7.4778E+06,
     +  7.4887E+06, 7.4996E+06, 7.5106E+06, 7.5215E+06, 7.5325E+06,
     + 7.5435E+06, 7.5545E+06, 7.5655E+06, 7.5765E+06, 7.5875E+06,
     +  7.5986E+06, 7.6096E+06, 7.6207E+06, 7.6318E+06, 7.6429E+06,
     + 7.6540E+06, 7.6651E+06, 7.6763E+06, 7.6874E+06, 7.6986E+06,
     +  7.7097E+06, 7.7209E+06, 7.7321E+06, 7.7433E+06, 7.7545E+06,
     + 7.7658E+06, 7.7770E+06, 7.7883E+06, 7.7996E+06, 7.8108E+06,
     +  7.8221E+06, 7.8334E+06, 7.8448E+06, 7.8561E+06, 7.8675E+06,
     + 7.8788E+06, 7.8902E+06, 7.9016E+06, 7.9130E+06, 7.9244E+06,
     +  7.9358E+06, 7.9472E+06, 7.9587E+06, 7.9702E+06, 7.9816E+06,
     + 7.9931E+06, 8.0046E+06, 8.0161E+06, 8.0277E+06, 8.0392E+06,
     +  8.0507E+06, 8.0623E+06, 8.0739E+06, 8.0855E+06, 8.0971E+06,
     + 8.1087E+06, 8.1203E+06, 8.1320E+06, 8.1436E+06, 8.1553E+06,
     +  8.1670E+06, 8.1787E+06, 8.1904E+06, 8.2021E+06, 8.2138E+06,
     + 8.2256E+06, 8.2373E+06, 8.2491E+06, 8.2609E+06, 8.2727E+06,
     +  8.2845E+06, 8.2963E+06, 8.3081E+06, 8.3200E+06, 8.3319E+06,
     + 8.3437E+06, 8.3556E+06, 8.3675E+06, 8.3794E+06, 8.3914E+06,
     +  8.4033E+06, 8.4152E+06, 8.4272E+06, 8.4392E+06, 8.4512E+06,
     + 8.4632E+06, 8.4752E+06, 8.4872E+06, 8.4993E+06, 8.5113E+06,
     +  8.5234E+06, 8.5355E+06, 8.5476E+06, 8.5597E+06, 8.5718E+06,
     + 8.5840E+06, 8.5961E+06, 8.6083E+06, 8.6205E+06, 8.6327E+06,
     +  8.6449E+06, 8.6571E+06, 8.6693E+06, 8.6816E+06, 8.6938E+06,
     + 8.7061E+06, 8.7184E+06, 8.7307E+06, 8.7430E+06, 8.7553E+06,
     +  8.7676E+06, 8.7800E+06, 8.7923E+06, 8.8047E+06, 8.8171E+06,
     + 8.8295E+06, 8.8419E+06, 8.8544E+06, 8.8668E+06, 8.8793E+06,
     +  8.8917E+06, 8.9042E+06, 8.9167E+06, 8.9292E+06, 8.9417E+06,
     + 8.9543E+06, 8.9668E+06, 8.9794E+06, 8.9920E+06, 9.0046E+06,
     +  9.0172E+06, 9.0298E+06, 9.0424E+06, 9.0551E+06, 9.0677E+06,
     + 9.0804E+06, 9.0931E+06, 9.1058E+06, 9.1185E+06, 9.1312E+06,
     +  9.1440E+06, 9.1567E+06, 9.1695E+06, 9.1823E+06, 9.1951E+06,
     + 9.2079E+06, 9.2207E+06, 9.2336E+06, 9.2464E+06, 9.2593E+06,
     +  9.2721E+06, 9.2850E+06, 9.2979E+06, 9.3109E+06, 9.3238E+06,
     + 9.3367E+06, 9.3497E+06, 9.3627E+06, 9.3757E+06, 9.3887E+06,
     +  9.4017E+06, 9.4147E+06, 9.4278E+06, 9.4408E+06, 9.4539E+06,
     + 9.4670E+06, 9.4801E+06, 9.4932E+06, 9.5063E+06, 9.5194E+06,
     +  9.5326E+06, 9.5458E+06, 9.5589E+06, 9.5721E+06, 9.5853E+06,
     + 9.5986E+06, 9.6118E+06, 9.6251E+06, 9.6383E+06, 9.6516E+06,
     +  9.6649E+06, 9.6782E+06, 9.6915E+06, 9.7049E+06, 9.7182E+06,
     + 9.7316E+06, 9.7449E+06, 9.7583E+06, 9.7717E+06, 9.7852E+06,
     +  9.7986E+06, 9.8120E+06, 9.8255E+06, 9.8390E+06, 9.8525E+06,
     + 9.8660E+06, 9.8795E+06, 9.8930E+06, 9.9066E+06, 9.9201E+06,
     +  9.9337E+06, 9.9473E+06, 9.9609E+06, 9.9745E+06, 9.9881E+06,
     + 1.0002E+07, 1.0015E+07, 1.0029E+07, 1.0043E+07, 1.0057E+07,
     +  1.0070E+07, 1.0084E+07, 1.0098E+07, 1.0111E+07, 1.0125E+07,
     + 1.0139E+07, 1.0153E+07, 1.0167E+07, 1.0180E+07, 1.0194E+07,
     +  1.0208E+07, 1.0222E+07, 1.0236E+07, 1.0250E+07, 1.0264E+07,
     + 1.0278E+07, 1.0292E+07, 1.0306E+07, 1.0320E+07, 1.0334E+07,
     +  1.0348E+07, 1.0362E+07, 1.0376E+07, 1.0390E+07, 1.0404E+07,
     + 1.0418E+07, 1.0432E+07, 1.0446E+07, 1.0460E+07, 1.0474E+07,
     +  1.0488E+07, 1.0503E+07, 1.0517E+07, 1.0531E+07, 1.0545E+07,
     + 1.0559E+07, 1.0574E+07, 1.0588E+07, 1.0602E+07, 1.0616E+07,
     +  1.0631E+07, 1.0645E+07, 1.0659E+07, 1.0674E+07, 1.0688E+07,
     + 1.0702E+07, 1.0717E+07, 1.0731E+07, 1.0746E+07, 1.0760E+07,
     +  1.0774E+07, 1.0789E+07, 1.0803E+07, 1.0818E+07, 1.0832E+07,
     + 1.0847E+07, 1.0861E+07, 1.0876E+07, 1.0890E+07, 1.0905E+07,
     +  1.0920E+07, 1.0934E+07, 1.0949E+07, 1.0963E+07, 1.0978E+07,
     + 1.0993E+07, 1.1007E+07, 1.1022E+07, 1.1037E+07, 1.1052E+07,
     +  1.1066E+07, 1.1081E+07, 1.1096E+07, 1.1111E+07, 1.1125E+07,
     + 1.1140E+07, 1.1155E+07, 1.1170E+07, 1.1185E+07, 1.1200E+07,
     +  1.1214E+07, 1.1229E+07, 1.1244E+07, 1.1259E+07, 1.1274E+07,
     + 1.1289E+07, 1.1304E+07, 1.1319E+07, 1.1334E+07, 1.1349E+07,
     +  1.1364E+07, 1.1379E+07, 1.1394E+07, 1.1409E+07, 1.1424E+07,
     + 1.1439E+07, 1.1455E+07, 1.1470E+07, 1.1485E+07, 1.1500E+07,
     +  1.1515E+07, 1.1530E+07, 1.1546E+07, 1.1561E+07, 1.1576E+07,
     + 1.1591E+07, 1.1607E+07, 1.1622E+07, 1.1637E+07, 1.1653E+07,
     +  1.1668E+07, 1.1683E+07, 1.1699E+07, 1.1714E+07, 1.1729E+07,
     + 1.1745E+07, 1.1760E+07, 1.1776E+07, 1.1791E+07, 1.1807E+07,
     +  1.1822E+07, 1.1838E+07, 1.1853E+07, 1.1869E+07, 1.1884E+07,
     + 1.1900E+07, 1.1915E+07, 1.1931E+07, 1.1946E+07, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_CO2   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj(13),QofT(13,NT),Q(NT)
      data (Tmx(I),I=1,13)/5000.,5000.,5000.,5000.,5000.,5000.,5000.,
     + 5000.,5000.,5000.,5000.,5000.,5000./
      data xgj/ 1.,2.,1.,6.,2.,12.,1.,6.,1.,2.,12.,2.,1./
c...      CO2
c...        --       626
      data (QofT( 1,J),J=1,2408)/ 1.1723E+00, 1.9608E+00, 2.8449E+00,
     +  3.7338E+00, 4.6234E+00, 5.5133E+00, 6.4034E+00, 7.2936E+00,
     + 8.1840E+00, 9.0743E+00, 9.9647E+00, 1.0855E+01, 1.1746E+01,
     +  1.2636E+01, 1.3527E+01, 1.4417E+01, 1.5308E+01, 1.6198E+01,
     + 1.7089E+01, 1.7979E+01, 1.9760E+01, 2.1542E+01, 2.3323E+01,
     +  2.5104E+01, 2.6885E+01, 2.8666E+01, 3.0448E+01, 3.2229E+01,
     + 3.4010E+01, 3.5792E+01, 3.7573E+01, 3.9354E+01, 4.1135E+01,
     +  4.2917E+01, 4.4698E+01, 4.6479E+01, 4.8261E+01, 5.0042E+01,
     + 5.1823E+01, 5.3605E+01, 5.5386E+01, 5.7168E+01, 5.8949E+01,
     +  6.0731E+01, 6.2512E+01, 6.4294E+01, 6.6075E+01, 6.7857E+01,
     + 6.9638E+01, 7.1420E+01, 7.3202E+01, 7.4984E+01, 7.6766E+01,
     +  7.8548E+01, 8.0330E+01, 8.2113E+01, 8.3896E+01, 8.5679E+01,
     + 8.7462E+01, 8.9246E+01, 9.1031E+01, 9.2815E+01, 9.4601E+01,
     +  9.6387E+01, 9.8174E+01, 9.9961E+01, 1.0175E+02, 1.0354E+02,
     + 1.0533E+02, 1.0712E+02, 1.0891E+02, 1.1071E+02, 1.1250E+02,
     +  1.1430E+02, 1.1610E+02, 1.1790E+02, 1.1971E+02, 1.2151E+02,
     + 1.2332E+02, 1.2513E+02, 1.2694E+02, 1.2876E+02, 1.3057E+02,
     +  1.3239E+02, 1.3422E+02, 1.3605E+02, 1.3788E+02, 1.3971E+02,
     + 1.4155E+02, 1.4339E+02, 1.4524E+02, 1.4709E+02, 1.4894E+02,
     +  1.5080E+02, 1.5266E+02, 1.5453E+02, 1.5641E+02, 1.5829E+02,
     + 1.6017E+02, 1.6206E+02, 1.6395E+02, 1.6586E+02, 1.6776E+02,
     +  1.6968E+02, 1.7159E+02, 1.7352E+02, 1.7545E+02, 1.7739E+02,
     + 1.7934E+02, 1.8129E+02, 1.8325E+02, 1.8522E+02, 1.8719E+02,
     +  1.8918E+02, 1.9117E+02, 1.9317E+02, 1.9517E+02, 1.9719E+02,
     + 1.9921E+02, 2.0124E+02, 2.0328E+02, 2.0533E+02, 2.0739E+02,
     +  2.0946E+02, 2.1153E+02, 2.1362E+02, 2.1572E+02, 2.1782E+02,
     + 2.1994E+02, 2.2206E+02, 2.2419E+02, 2.2634E+02, 2.2850E+02,
     +  2.3066E+02, 2.3284E+02, 2.3502E+02, 2.3722E+02, 2.3943E+02,
     + 2.4165E+02, 2.4388E+02, 2.4612E+02, 2.4838E+02, 2.5064E+02,
     +  2.5292E+02, 2.5521E+02, 2.5751E+02, 2.5982E+02, 2.6215E+02,
     + 2.6449E+02, 2.6684E+02, 2.6920E+02, 2.7157E+02, 2.7396E+02,
     +  2.7636E+02, 2.7877E+02, 2.8120E+02, 2.8364E+02, 2.8609E+02,
     + 2.8856E+02, 2.9104E+02, 2.9353E+02, 2.9604E+02, 2.9856E+02,
     +  3.0110E+02, 3.0365E+02, 3.0621E+02, 3.0879E+02, 3.1138E+02,
     + 3.1399E+02, 3.1661E+02, 3.1925E+02, 3.2190E+02, 3.2456E+02,
     +  3.2724E+02, 3.2994E+02, 3.3265E+02, 3.3538E+02, 3.3812E+02,
     + 3.4088E+02, 3.4365E+02, 3.4644E+02, 3.4925E+02, 3.5207E+02,
     +  3.5491E+02, 3.5776E+02, 3.6063E+02, 3.6352E+02, 3.6642E+02,
     + 3.6934E+02, 3.7228E+02, 3.7523E+02, 3.7820E+02, 3.8119E+02,
     +  3.8419E+02, 3.8721E+02, 3.9025E+02, 3.9331E+02, 3.9638E+02,
     + 3.9948E+02, 4.0259E+02, 4.0571E+02, 4.0886E+02, 4.1202E+02,
     +  4.1520E+02, 4.1840E+02, 4.2162E+02, 4.2486E+02, 4.2811E+02,
     + 4.3139E+02, 4.3468E+02, 4.3799E+02, 4.4132E+02, 4.4467E+02,
     +  4.4804E+02, 4.5143E+02, 4.5484E+02, 4.5827E+02, 4.6171E+02,
     + 4.6518E+02, 4.6867E+02, 4.7218E+02, 4.7570E+02, 4.7925E+02,
     +  4.8282E+02, 4.8640E+02, 4.9001E+02, 4.9364E+02, 4.9729E+02,
     + 5.0096E+02, 5.0465E+02, 5.0836E+02, 5.1210E+02, 5.1585E+02,
     +  5.1963E+02, 5.2343E+02, 5.2725E+02, 5.3109E+02, 5.3495E+02,
     + 5.3883E+02, 5.4274E+02, 5.4667E+02, 5.5062E+02, 5.5459E+02,
     +  5.5859E+02, 5.6261E+02, 5.6665E+02, 5.7071E+02, 5.7480E+02,
     + 5.7891E+02, 5.8304E+02, 5.8719E+02, 5.9137E+02, 5.9557E+02,
     +  5.9980E+02, 6.0405E+02, 6.0832E+02, 6.1262E+02, 6.1694E+02,
     + 6.2129E+02, 6.2566E+02, 6.3005E+02, 6.3447E+02, 6.3891E+02,
     +  6.4338E+02, 6.4787E+02, 6.5239E+02, 6.5693E+02, 6.6150E+02,
     + 6.6609E+02, 6.7071E+02, 6.7536E+02, 6.8003E+02, 6.8472E+02,
     +  6.8944E+02, 6.9419E+02, 6.9896E+02, 7.0376E+02, 7.0859E+02,
     + 7.1344E+02, 7.1832E+02, 7.2322E+02, 7.2815E+02, 7.3311E+02,
     +  7.3810E+02, 7.4311E+02, 7.4815E+02, 7.5322E+02, 7.5832E+02,
     + 7.6344E+02, 7.6859E+02, 7.7377E+02, 7.7898E+02, 7.8421E+02,
     +  7.8948E+02, 7.9477E+02, 8.0009E+02, 8.0544E+02, 8.1082E+02,
     + 8.1622E+02, 8.2166E+02, 8.2712E+02, 8.3262E+02, 8.3814E+02,
     +  8.4370E+02, 8.4928E+02, 8.5489E+02, 8.6054E+02, 8.6621E+02,
     + 8.7191E+02, 8.7765E+02, 8.8341E+02, 8.8920E+02, 8.9503E+02,
     +  9.0088E+02, 9.0677E+02, 9.1269E+02, 9.1864E+02, 9.2462E+02,
     + 9.3063E+02, 9.3668E+02, 9.4275E+02, 9.4886E+02, 9.5500E+02,
     +  9.6117E+02, 9.6738E+02, 9.7361E+02, 9.7988E+02, 9.8619E+02,
     + 9.9252E+02, 9.9889E+02, 1.0053E+03, 1.0117E+03, 1.0182E+03,
     +  1.0247E+03, 1.0312E+03, 1.0378E+03, 1.0444E+03, 1.0510E+03,
     + 1.0577E+03, 1.0644E+03, 1.0712E+03, 1.0779E+03, 1.0847E+03,
     +  1.0916E+03, 1.0985E+03, 1.1054E+03, 1.1123E+03, 1.1193E+03,
     + 1.1263E+03, 1.1334E+03, 1.1405E+03, 1.1476E+03, 1.1548E+03,
     +  1.1620E+03, 1.1692E+03, 1.1765E+03, 1.1838E+03, 1.1912E+03,
     + 1.1986E+03, 1.2060E+03, 1.2135E+03, 1.2210E+03, 1.2285E+03,
     +  1.2361E+03, 1.2437E+03, 1.2513E+03, 1.2590E+03, 1.2668E+03,
     + 1.2745E+03, 1.2823E+03, 1.2902E+03, 1.2981E+03, 1.3060E+03,
     +  1.3140E+03, 1.3220E+03, 1.3300E+03, 1.3381E+03, 1.3462E+03,
     + 1.3544E+03, 1.3626E+03, 1.3708E+03, 1.3791E+03, 1.3874E+03,
     +  1.3958E+03, 1.4042E+03, 1.4127E+03, 1.4212E+03, 1.4297E+03,
     + 1.4383E+03, 1.4469E+03, 1.4555E+03, 1.4642E+03, 1.4730E+03,
     +  1.4818E+03, 1.4906E+03, 1.4995E+03, 1.5084E+03, 1.5174E+03,
     + 1.5264E+03, 1.5354E+03, 1.5445E+03, 1.5536E+03, 1.5628E+03,
     +  1.5720E+03, 1.5813E+03, 1.5906E+03, 1.6000E+03, 1.6094E+03,
     + 1.6188E+03, 1.6283E+03, 1.6379E+03, 1.6474E+03, 1.6571E+03,
     +  1.6668E+03, 1.6765E+03, 1.6862E+03, 1.6961E+03, 1.7059E+03,
     + 1.7158E+03, 1.7258E+03, 1.7358E+03, 1.7458E+03, 1.7559E+03,
     +  1.7661E+03, 1.7763E+03, 1.7865E+03, 1.7968E+03, 1.8072E+03,
     + 1.8176E+03, 1.8280E+03, 1.8385E+03, 1.8490E+03, 1.8596E+03,
     +  1.8702E+03, 1.8809E+03, 1.8917E+03, 1.9024E+03, 1.9133E+03,
     + 1.9242E+03, 1.9351E+03, 1.9461E+03, 1.9571E+03, 1.9682E+03,
     +  1.9794E+03, 1.9906E+03, 2.0018E+03, 2.0131E+03, 2.0245E+03,
     + 2.0359E+03, 2.0473E+03, 2.0588E+03, 2.0704E+03, 2.0820E+03,
     +  2.0937E+03, 2.1054E+03, 2.1172E+03, 2.1290E+03, 2.1409E+03,
     + 2.1528E+03, 2.1648E+03, 2.1769E+03, 2.1890E+03, 2.2011E+03,
     +  2.2133E+03, 2.2256E+03, 2.2379E+03, 2.2503E+03, 2.2627E+03,
     + 2.2752E+03, 2.2878E+03, 2.3004E+03, 2.3131E+03, 2.3258E+03,
     +  2.3386E+03, 2.3514E+03, 2.3643E+03, 2.3772E+03, 2.3902E+03,
     + 2.4033E+03, 2.4164E+03, 2.4296E+03, 2.4429E+03, 2.4562E+03,
     +  2.4695E+03, 2.4829E+03, 2.4964E+03, 2.5100E+03, 2.5236E+03,
     + 2.5372E+03, 2.5510E+03, 2.5647E+03, 2.5786E+03, 2.5925E+03,
     +  2.6065E+03, 2.6205E+03, 2.6346E+03, 2.6487E+03, 2.6629E+03,
     + 2.6772E+03, 2.6916E+03, 2.7060E+03, 2.7204E+03, 2.7350E+03,
     +  2.7496E+03, 2.7642E+03, 2.7789E+03, 2.7937E+03, 2.8086E+03,
     + 2.8235E+03, 2.8385E+03, 2.8535E+03, 2.8686E+03, 2.8838E+03,
     +  2.8990E+03, 2.9144E+03, 2.9297E+03, 2.9452E+03, 2.9607E+03,
     + 2.9763E+03, 2.9919E+03, 3.0076E+03, 3.0234E+03, 3.0392E+03,
     +  3.0551E+03, 3.0711E+03, 3.0872E+03, 3.1033E+03, 3.1195E+03,
     + 3.1357E+03, 3.1521E+03, 3.1685E+03, 3.1849E+03, 3.2015E+03,
     +  3.2181E+03, 3.2347E+03, 3.2515E+03, 3.2683E+03, 3.2852E+03,
     + 3.3022E+03, 3.3192E+03, 3.3363E+03, 3.3535E+03, 3.3707E+03,
     +  3.3881E+03, 3.4055E+03, 3.4229E+03, 3.4405E+03, 3.4581E+03,
     + 3.4758E+03, 3.4935E+03, 3.5114E+03, 3.5293E+03, 3.5473E+03,
     +  3.5653E+03, 3.5835E+03, 3.6017E+03, 3.6200E+03, 3.6384E+03,
     + 3.6568E+03, 3.6753E+03, 3.6939E+03, 3.7126E+03, 3.7313E+03,
     +  3.7502E+03, 3.7691E+03, 3.7881E+03, 3.8071E+03, 3.8263E+03,
     + 3.8455E+03, 3.8648E+03, 3.8842E+03, 3.9036E+03, 3.9232E+03,
     +  3.9428E+03, 3.9625E+03, 3.9823E+03, 4.0021E+03, 4.0221E+03,
     + 4.0421E+03, 4.0622E+03, 4.0824E+03, 4.1026E+03, 4.1230E+03,
     +  4.1434E+03, 4.1639E+03, 4.1845E+03, 4.2052E+03, 4.2260E+03,
     + 4.2468E+03, 4.2678E+03, 4.2888E+03, 4.3099E+03, 4.3311E+03,
     +  4.3524E+03, 4.3737E+03, 4.3952E+03, 4.4167E+03, 4.4383E+03,
     + 4.4600E+03, 4.4818E+03, 4.5037E+03, 4.5257E+03, 4.5477E+03,
     +  4.5699E+03, 4.5921E+03, 4.6144E+03, 4.6368E+03, 4.6593E+03,
     + 4.6819E+03, 4.7046E+03, 4.7273E+03, 4.7502E+03, 4.7732E+03,
     +  4.7962E+03, 4.8193E+03, 4.8425E+03, 4.8658E+03, 4.8893E+03,
     + 4.9127E+03, 4.9363E+03, 4.9600E+03, 4.9838E+03, 5.0077E+03,
     +  5.0316E+03, 5.0557E+03, 5.0798E+03, 5.1041E+03, 5.1284E+03,
     + 5.1528E+03, 5.1774E+03, 5.2020E+03, 5.2267E+03, 5.2515E+03,
     +  5.2764E+03, 5.3014E+03, 5.3265E+03, 5.3517E+03, 5.3770E+03,
     + 5.4024E+03, 5.4279E+03, 5.4535E+03, 5.4792E+03, 5.5050E+03,
     +  5.5309E+03, 5.5569E+03, 5.5830E+03, 5.6092E+03, 5.6355E+03,
     + 5.6619E+03, 5.6884E+03, 5.7149E+03, 5.7416E+03, 5.7684E+03,
     +  5.7953E+03, 5.8223E+03, 5.8495E+03, 5.8767E+03, 5.9040E+03,
     + 5.9314E+03, 5.9589E+03, 5.9865E+03, 6.0143E+03, 6.0421E+03,
     +  6.0700E+03, 6.0981E+03, 6.1262E+03, 6.1545E+03, 6.1829E+03,
     + 6.2113E+03, 6.2399E+03, 6.2686E+03, 6.2974E+03, 6.3263E+03,
     +  6.3553E+03, 6.3844E+03, 6.4136E+03, 6.4430E+03, 6.4724E+03,
     + 6.5020E+03, 6.5317E+03, 6.5614E+03, 6.5913E+03, 6.6213E+03,
     +  6.6514E+03, 6.6817E+03, 6.7120E+03, 6.7424E+03, 6.7730E+03,
     + 6.8037E+03, 6.8345E+03, 6.8654E+03, 6.8964E+03, 6.9275E+03,
     +  6.9588E+03, 6.9901E+03, 7.0216E+03, 7.0532E+03, 7.0849E+03,
     + 7.1167E+03, 7.1486E+03, 7.1807E+03, 7.2129E+03, 7.2452E+03,
     +  7.2776E+03, 7.3101E+03, 7.3428E+03, 7.3755E+03, 7.4084E+03,
     + 7.4414E+03, 7.4745E+03, 7.5078E+03, 7.5412E+03, 7.5746E+03,
     +  7.6083E+03, 7.6420E+03, 7.6758E+03, 7.7098E+03, 7.7439E+03,
     + 7.7781E+03, 7.8125E+03, 7.8469E+03, 7.8815E+03, 7.9163E+03,
     +  7.9511E+03, 7.9861E+03, 8.0212E+03, 8.0564E+03, 8.0917E+03,
     + 8.1272E+03, 8.1628E+03, 8.1985E+03, 8.2344E+03, 8.2703E+03,
     +  8.3065E+03, 8.3427E+03, 8.3791E+03, 8.4156E+03, 8.4522E+03,
     + 8.4889E+03, 8.5258E+03, 8.5628E+03, 8.6000E+03, 8.6373E+03,
     +  8.6747E+03, 8.7122E+03, 8.7499E+03, 8.7877E+03, 8.8256E+03,
     + 8.8637E+03, 8.9019E+03, 8.9403E+03, 8.9788E+03, 9.0174E+03,
     +  9.0561E+03, 9.0950E+03, 9.1340E+03, 9.1732E+03, 9.2125E+03,
     + 9.2519E+03, 9.2915E+03, 9.3312E+03, 9.3710E+03, 9.4110E+03,
     +  9.4511E+03, 9.4914E+03, 9.5318E+03, 9.5723E+03, 9.6130E+03,
     + 9.6538E+03, 9.6948E+03, 9.7359E+03, 9.7772E+03, 9.8185E+03,
     +  9.8601E+03, 9.9018E+03, 9.9436E+03, 9.9855E+03, 1.0028E+04,
     + 1.0070E+04, 1.0112E+04, 1.0155E+04, 1.0198E+04, 1.0240E+04,
     +  1.0283E+04, 1.0326E+04, 1.0370E+04, 1.0413E+04, 1.0457E+04,
     + 1.0500E+04, 1.0544E+04, 1.0588E+04, 1.0633E+04, 1.0677E+04,
     +  1.0721E+04, 1.0766E+04, 1.0811E+04, 1.0856E+04, 1.0901E+04,
     + 1.0946E+04, 1.0991E+04, 1.1037E+04, 1.1083E+04, 1.1128E+04,
     +  1.1174E+04, 1.1221E+04, 1.1267E+04, 1.1313E+04, 1.1360E+04,
     + 1.1407E+04, 1.1454E+04, 1.1501E+04, 1.1548E+04, 1.1595E+04,
     +  1.1643E+04, 1.1691E+04, 1.1739E+04, 1.1787E+04, 1.1835E+04,
     + 1.1883E+04, 1.1932E+04, 1.1980E+04, 1.2029E+04, 1.2078E+04,
     +  1.2127E+04, 1.2177E+04, 1.2226E+04, 1.2276E+04, 1.2326E+04,
     + 1.2376E+04, 1.2426E+04, 1.2476E+04, 1.2527E+04, 1.2577E+04,
     +  1.2628E+04, 1.2679E+04, 1.2730E+04, 1.2781E+04, 1.2833E+04,
     + 1.2885E+04, 1.2936E+04, 1.2988E+04, 1.3041E+04, 1.3093E+04,
     +  1.3145E+04, 1.3198E+04, 1.3251E+04, 1.3304E+04, 1.3357E+04,
     + 1.3410E+04, 1.3464E+04, 1.3518E+04, 1.3572E+04, 1.3626E+04,
     +  1.3680E+04, 1.3734E+04, 1.3789E+04, 1.3844E+04, 1.3899E+04,
     + 1.3954E+04, 1.4009E+04, 1.4064E+04, 1.4120E+04, 1.4176E+04,
     +  1.4232E+04, 1.4288E+04, 1.4344E+04, 1.4401E+04, 1.4458E+04,
     + 1.4515E+04, 1.4572E+04, 1.4629E+04, 1.4687E+04, 1.4744E+04,
     +  1.4802E+04, 1.4860E+04, 1.4918E+04, 1.4977E+04, 1.5035E+04,
     + 1.5094E+04, 1.5153E+04, 1.5212E+04, 1.5271E+04, 1.5331E+04,
     +  1.5390E+04, 1.5450E+04, 1.5510E+04, 1.5571E+04, 1.5631E+04,
     + 1.5692E+04, 1.5753E+04, 1.5814E+04, 1.5875E+04, 1.5936E+04,
     +  1.5998E+04, 1.6060E+04, 1.6121E+04, 1.6184E+04, 1.6246E+04,
     + 1.6309E+04, 1.6371E+04, 1.6434E+04, 1.6497E+04, 1.6561E+04,
     +  1.6624E+04, 1.6688E+04, 1.6752E+04, 1.6816E+04, 1.6881E+04,
     + 1.6945E+04, 1.7010E+04, 1.7075E+04, 1.7140E+04, 1.7205E+04,
     +  1.7271E+04, 1.7337E+04, 1.7402E+04, 1.7469E+04, 1.7535E+04,
     + 1.7602E+04, 1.7668E+04, 1.7735E+04, 1.7802E+04, 1.7870E+04,
     +  1.7937E+04, 1.8005E+04, 1.8073E+04, 1.8141E+04, 1.8210E+04,
     + 1.8279E+04, 1.8347E+04, 1.8416E+04, 1.8486E+04, 1.8555E+04,
     +  1.8625E+04, 1.8695E+04, 1.8765E+04, 1.8835E+04, 1.8906E+04,
     + 1.8977E+04, 1.9047E+04, 1.9119E+04, 1.9190E+04, 1.9262E+04,
     +  1.9334E+04, 1.9406E+04, 1.9478E+04, 1.9550E+04, 1.9623E+04,
     + 1.9696E+04, 1.9769E+04, 1.9842E+04, 1.9916E+04, 1.9990E+04,
     +  2.0064E+04, 2.0138E+04, 2.0213E+04, 2.0287E+04, 2.0362E+04,
     + 2.0437E+04, 2.0513E+04, 2.0588E+04, 2.0664E+04, 2.0740E+04,
     +  2.0816E+04, 2.0893E+04, 2.0970E+04, 2.1047E+04, 2.1124E+04,
     + 2.1201E+04, 2.1279E+04, 2.1357E+04, 2.1435E+04, 2.1513E+04,
     +  2.1592E+04, 2.1670E+04, 2.1749E+04, 2.1829E+04, 2.1908E+04,
     + 2.1988E+04, 2.2068E+04, 2.2148E+04, 2.2229E+04, 2.2309E+04,
     +  2.2390E+04, 2.2471E+04, 2.2553E+04, 2.2634E+04, 2.2716E+04,
     + 2.2798E+04, 2.2881E+04, 2.2963E+04, 2.3046E+04, 2.3129E+04,
     +  2.3212E+04, 2.3296E+04, 2.3380E+04, 2.3464E+04, 2.3548E+04,
     + 2.3633E+04, 2.3717E+04, 2.3802E+04, 2.3888E+04, 2.3973E+04,
     +  2.4059E+04, 2.4145E+04, 2.4231E+04, 2.4318E+04, 2.4404E+04,
     + 2.4491E+04, 2.4579E+04, 2.4666E+04, 2.4754E+04, 2.4842E+04,
     +  2.4930E+04, 2.5019E+04, 2.5107E+04, 2.5196E+04, 2.5286E+04,
     + 2.5375E+04, 2.5465E+04, 2.5555E+04, 2.5645E+04, 2.5736E+04,
     +  2.5827E+04, 2.5918E+04, 2.6009E+04, 2.6101E+04, 2.6193E+04,
     + 2.6285E+04, 2.6377E+04, 2.6470E+04, 2.6563E+04, 2.6656E+04,
     +  2.6749E+04, 2.6843E+04, 2.6937E+04, 2.7031E+04, 2.7126E+04,
     + 2.7221E+04, 2.7316E+04, 2.7411E+04, 2.7506E+04, 2.7602E+04,
     +  2.7698E+04, 2.7795E+04, 2.7891E+04, 2.7988E+04, 2.8086E+04,
     + 2.8183E+04, 2.8281E+04, 2.8379E+04, 2.8477E+04, 2.8576E+04,
     +  2.8674E+04, 2.8774E+04, 2.8873E+04, 2.8973E+04, 2.9073E+04,
     + 2.9173E+04, 2.9273E+04, 2.9374E+04, 2.9475E+04, 2.9576E+04,
     +  2.9678E+04, 2.9780E+04, 2.9882E+04, 2.9985E+04, 3.0087E+04,
     + 3.0190E+04, 3.0294E+04, 3.0397E+04, 3.0501E+04, 3.0605E+04,
     +  3.0710E+04, 3.0815E+04, 3.0920E+04, 3.1025E+04, 3.1131E+04,
     + 3.1237E+04, 3.1343E+04, 3.1449E+04, 3.1556E+04, 3.1663E+04,
     +  3.1771E+04, 3.1878E+04, 3.1986E+04, 3.2094E+04, 3.2203E+04,
     + 3.2312E+04, 3.2421E+04, 3.2530E+04, 3.2640E+04, 3.2750E+04,
     +  3.2860E+04, 3.2971E+04, 3.3082E+04, 3.3193E+04, 3.3305E+04,
     + 3.3417E+04, 3.3529E+04, 3.3641E+04, 3.3754E+04, 3.3867E+04,
     +  3.3980E+04, 3.4094E+04, 3.4208E+04, 3.4322E+04, 3.4437E+04,
     + 3.4552E+04, 3.4667E+04, 3.4782E+04, 3.4898E+04, 3.5014E+04,
     +  3.5131E+04, 3.5248E+04, 3.5365E+04, 3.5482E+04, 3.5600E+04,
     + 3.5718E+04, 3.5836E+04, 3.5955E+04, 3.6074E+04, 3.6193E+04,
     +  3.6313E+04, 3.6433E+04, 3.6553E+04, 3.6673E+04, 3.6794E+04,
     + 3.6915E+04, 3.7037E+04, 3.7159E+04, 3.7281E+04, 3.7403E+04,
     +  3.7526E+04, 3.7649E+04, 3.7773E+04, 3.7896E+04, 3.8021E+04,
     + 3.8145E+04, 3.8270E+04, 3.8395E+04, 3.8520E+04, 3.8646E+04,
     +  3.8772E+04, 3.8899E+04, 3.9025E+04, 3.9152E+04, 3.9280E+04,
     + 3.9407E+04, 3.9536E+04, 3.9664E+04, 3.9793E+04, 3.9922E+04,
     +  4.0051E+04, 4.0181E+04, 4.0311E+04, 4.0442E+04, 4.0572E+04,
     + 4.0703E+04, 4.0835E+04, 4.0967E+04, 4.1099E+04, 4.1231E+04,
     +  4.1364E+04, 4.1497E+04, 4.1631E+04, 4.1765E+04, 4.1899E+04,
     + 4.2033E+04, 4.2168E+04, 4.2303E+04, 4.2439E+04, 4.2575E+04,
     +  4.2711E+04, 4.2848E+04, 4.2985E+04, 4.3122E+04, 4.3260E+04,
     + 4.3398E+04, 4.3536E+04, 4.3675E+04, 4.3814E+04, 4.3954E+04,
     +  4.4094E+04, 4.4234E+04, 4.4374E+04, 4.4515E+04, 4.4657E+04,
     + 4.4798E+04, 4.4940E+04, 4.5083E+04, 4.5225E+04, 4.5368E+04,
     +  4.5512E+04, 4.5656E+04, 4.5800E+04, 4.5944E+04, 4.6089E+04,
     + 4.6234E+04, 4.6380E+04, 4.6526E+04, 4.6672E+04, 4.6819E+04,
     +  4.6966E+04, 4.7114E+04, 4.7262E+04, 4.7410E+04, 4.7558E+04,
     + 4.7707E+04, 4.7857E+04, 4.8006E+04, 4.8157E+04, 4.8307E+04,
     +  4.8458E+04, 4.8609E+04, 4.8761E+04, 4.8913E+04, 4.9065E+04,
     + 4.9218E+04, 4.9371E+04, 4.9524E+04, 4.9678E+04, 4.9833E+04,
     +  4.9987E+04, 5.0142E+04, 5.0298E+04, 5.0454E+04, 5.0610E+04,
     + 5.0766E+04, 5.0923E+04, 5.1081E+04, 5.1239E+04, 5.1397E+04,
     +  5.1555E+04, 5.1714E+04, 5.1874E+04, 5.2033E+04, 5.2193E+04,
     + 5.2354E+04, 5.2515E+04, 5.2676E+04, 5.2838E+04, 5.3000E+04,
     +  5.3163E+04, 5.3325E+04, 5.3489E+04, 5.3652E+04, 5.3817E+04,
     + 5.3981E+04, 5.4146E+04, 5.4311E+04, 5.4477E+04, 5.4643E+04,
     +  5.4810E+04, 5.4977E+04, 5.5144E+04, 5.5312E+04, 5.5480E+04,
     + 5.5648E+04, 5.5817E+04, 5.5987E+04, 5.6157E+04, 5.6327E+04,
     +  5.6498E+04, 5.6669E+04, 5.6840E+04, 5.7012E+04, 5.7184E+04,
     + 5.7357E+04, 5.7530E+04, 5.7704E+04, 5.7878E+04, 5.8052E+04,
     +  5.8227E+04, 5.8402E+04, 5.8578E+04, 5.8754E+04, 5.8930E+04,
     + 5.9107E+04, 5.9285E+04, 5.9462E+04, 5.9641E+04, 5.9819E+04,
     +  5.9998E+04, 6.0178E+04, 6.0358E+04, 6.0538E+04, 6.0719E+04,
     + 6.0900E+04, 6.1082E+04, 6.1264E+04, 6.1446E+04, 6.1629E+04,
     +  6.1813E+04, 6.1996E+04, 6.2181E+04, 6.2365E+04, 6.2551E+04,
     + 6.2736E+04, 6.2922E+04, 6.3109E+04, 6.3295E+04, 6.3483E+04,
     +  6.3671E+04, 6.3859E+04, 6.4048E+04, 6.4237E+04, 6.4426E+04,
     + 6.4616E+04, 6.4807E+04, 6.4998E+04, 6.5189E+04, 6.5381E+04,
     +  6.5573E+04, 6.5766E+04, 6.5959E+04, 6.6153E+04, 6.6347E+04,
     + 6.6541E+04, 6.6736E+04, 6.6932E+04, 6.7128E+04, 6.7324E+04,
     +  6.7521E+04, 6.7718E+04, 6.7916E+04, 6.8114E+04, 6.8313E+04,
     + 6.8512E+04, 6.8711E+04, 6.8911E+04, 6.9112E+04, 6.9313E+04,
     +  6.9514E+04, 6.9716E+04, 6.9919E+04, 7.0122E+04, 7.0325E+04,
     + 7.0529E+04, 7.0733E+04, 7.0938E+04, 7.1143E+04, 7.1349E+04,
     +  7.1555E+04, 7.1762E+04, 7.1969E+04, 7.2176E+04, 7.2384E+04,
     + 7.2593E+04, 7.2802E+04, 7.3011E+04, 7.3221E+04, 7.3432E+04,
     +  7.3643E+04, 7.3854E+04, 7.4066E+04, 7.4279E+04, 7.4492E+04,
     + 7.4705E+04, 7.4919E+04, 7.5133E+04, 7.5348E+04, 7.5564E+04,
     +  7.5779E+04, 7.5996E+04, 7.6213E+04, 7.6430E+04, 7.6648E+04,
     + 7.6866E+04, 7.7085E+04, 7.7304E+04, 7.7524E+04, 7.7744E+04,
     +  7.7965E+04, 7.8186E+04, 7.8408E+04, 7.8630E+04, 7.8853E+04,
     + 7.9077E+04, 7.9300E+04, 7.9525E+04, 7.9749E+04, 7.9975E+04,
     +  8.0201E+04, 8.0427E+04, 8.0654E+04, 8.0881E+04, 8.1109E+04,
     + 8.1337E+04, 8.1566E+04, 8.1796E+04, 8.2026E+04, 8.2256E+04,
     +  8.2487E+04, 8.2719E+04, 8.2951E+04, 8.3183E+04, 8.3416E+04,
     + 8.3650E+04, 8.3884E+04, 8.4118E+04, 8.4354E+04, 8.4589E+04,
     +  8.4825E+04, 8.5062E+04, 8.5299E+04, 8.5537E+04, 8.5775E+04,
     + 8.6014E+04, 8.6253E+04, 8.6493E+04, 8.6734E+04, 8.6975E+04,
     +  8.7216E+04, 8.7458E+04, 8.7701E+04, 8.7944E+04, 8.8187E+04,
     + 8.8431E+04, 8.8676E+04, 8.8921E+04, 8.9167E+04, 8.9413E+04,
     +  8.9660E+04, 8.9907E+04, 9.0155E+04, 9.0404E+04, 9.0653E+04,
     + 9.0902E+04, 9.1153E+04, 9.1403E+04, 9.1654E+04, 9.1906E+04,
     +  9.2158E+04, 9.2411E+04, 9.2665E+04, 9.2919E+04, 9.3173E+04,
     + 9.3428E+04, 9.3684E+04, 9.3940E+04, 9.4197E+04, 9.4454E+04,
     +  9.4712E+04, 9.4970E+04, 9.5229E+04, 9.5489E+04, 9.5749E+04,
     + 9.6009E+04, 9.6271E+04, 9.6532E+04, 9.6795E+04, 9.7058E+04,
     +  9.7321E+04, 9.7585E+04, 9.7850E+04, 9.8115E+04, 9.8381E+04,
     + 9.8647E+04, 9.8914E+04, 9.9181E+04, 9.9450E+04, 9.9718E+04,
     +  9.9987E+04, 1.0026E+05, 1.0053E+05, 1.0080E+05, 1.0107E+05,
     + 1.0134E+05, 1.0161E+05, 1.0189E+05, 1.0216E+05, 1.0244E+05,
     +  1.0271E+05, 1.0299E+05, 1.0326E+05, 1.0354E+05, 1.0382E+05,
     + 1.0410E+05, 1.0437E+05, 1.0465E+05, 1.0493E+05, 1.0521E+05,
     +  1.0549E+05, 1.0578E+05, 1.0606E+05, 1.0634E+05, 1.0662E+05,
     + 1.0691E+05, 1.0719E+05, 1.0748E+05, 1.0776E+05, 1.0805E+05,
     +  1.0834E+05, 1.0862E+05, 1.0891E+05, 1.0920E+05, 1.0949E+05,
     + 1.0978E+05, 1.1007E+05, 1.1036E+05, 1.1065E+05, 1.1095E+05,
     +  1.1124E+05, 1.1153E+05, 1.1183E+05, 1.1212E+05, 1.1242E+05,
     + 1.1271E+05, 1.1301E+05, 1.1331E+05, 1.1360E+05, 1.1390E+05,
     +  1.1420E+05, 1.1450E+05, 1.1480E+05, 1.1510E+05, 1.1540E+05,
     + 1.1571E+05, 1.1601E+05, 1.1631E+05, 1.1662E+05, 1.1692E+05,
     +  1.1723E+05, 1.1753E+05, 1.1784E+05, 1.1815E+05, 1.1846E+05,
     + 1.1876E+05, 1.1907E+05, 1.1938E+05, 1.1969E+05, 1.2001E+05,
     +  1.2032E+05, 1.2063E+05, 1.2094E+05, 1.2126E+05, 1.2157E+05,
     + 1.2189E+05, 1.2220E+05, 1.2252E+05, 1.2283E+05, 1.2315E+05,
     +  1.2347E+05, 1.2379E+05, 1.2411E+05, 1.2443E+05, 1.2475E+05,
     + 1.2507E+05, 1.2539E+05, 1.2572E+05, 1.2604E+05, 1.2636E+05,
     +  1.2669E+05, 1.2701E+05, 1.2734E+05, 1.2767E+05, 1.2799E+05,
     + 1.2832E+05, 1.2865E+05, 1.2898E+05, 1.2931E+05, 1.2964E+05,
     +  1.2997E+05, 1.3031E+05, 1.3064E+05, 1.3097E+05, 1.3131E+05,
     + 1.3164E+05, 1.3198E+05, 1.3231E+05, 1.3265E+05, 1.3299E+05,
     +  1.3332E+05, 1.3366E+05, 1.3400E+05, 1.3434E+05, 1.3468E+05,
     + 1.3503E+05, 1.3537E+05, 1.3571E+05, 1.3605E+05, 1.3640E+05,
     +  1.3674E+05, 1.3709E+05, 1.3744E+05, 1.3778E+05, 1.3813E+05,
     + 1.3848E+05, 1.3883E+05, 1.3918E+05, 1.3953E+05, 1.3988E+05,
     +  1.4023E+05, 1.4058E+05, 1.4094E+05, 1.4129E+05, 1.4165E+05,
     + 1.4200E+05, 1.4236E+05, 1.4271E+05, 1.4307E+05, 1.4343E+05,
     +  1.4379E+05, 1.4415E+05, 1.4451E+05, 1.4487E+05, 1.4523E+05,
     + 1.4559E+05, 1.4596E+05, 1.4632E+05, 1.4669E+05, 1.4705E+05,
     +  1.4742E+05, 1.4778E+05, 1.4815E+05, 1.4852E+05, 1.4889E+05,
     + 1.4926E+05, 1.4963E+05, 1.5000E+05, 1.5037E+05, 1.5074E+05,
     +  1.5112E+05, 1.5149E+05, 1.5187E+05, 1.5224E+05, 1.5262E+05,
     + 1.5299E+05, 1.5337E+05, 1.5375E+05, 1.5413E+05, 1.5451E+05,
     +  1.5489E+05, 1.5527E+05, 1.5565E+05, 1.5604E+05, 1.5642E+05,
     + 1.5680E+05, 1.5719E+05, 1.5758E+05, 1.5796E+05, 1.5835E+05,
     +  1.5874E+05, 1.5913E+05, 1.5952E+05, 1.5991E+05, 1.6030E+05,
     + 1.6069E+05, 1.6108E+05, 1.6147E+05, 1.6187E+05, 1.6226E+05,
     +  1.6266E+05, 1.6305E+05, 1.6345E+05, 1.6385E+05, 1.6425E+05,
     + 1.6465E+05, 1.6505E+05, 1.6545E+05, 1.6585E+05, 1.6625E+05,
     +  1.6666E+05, 1.6706E+05, 1.6746E+05, 1.6787E+05, 1.6828E+05,
     + 1.6868E+05, 1.6909E+05, 1.6950E+05, 1.6991E+05, 1.7032E+05,
     +  1.7073E+05, 1.7114E+05, 1.7155E+05, 1.7197E+05, 1.7238E+05,
     + 1.7280E+05, 1.7321E+05, 1.7363E+05, 1.7404E+05, 1.7446E+05,
     +  1.7488E+05, 1.7530E+05, 1.7572E+05, 1.7614E+05, 1.7656E+05,
     + 1.7699E+05, 1.7741E+05, 1.7783E+05, 1.7826E+05, 1.7868E+05,
     +  1.7911E+05, 1.7954E+05, 1.7997E+05, 1.8040E+05, 1.8083E+05,
     + 1.8126E+05, 1.8169E+05, 1.8212E+05, 1.8255E+05, 1.8299E+05,
     +  1.8342E+05, 1.8386E+05, 1.8429E+05, 1.8473E+05, 1.8517E+05,
     + 1.8561E+05, 1.8605E+05, 1.8649E+05, 1.8693E+05, 1.8737E+05,
     +  1.8781E+05, 1.8826E+05, 1.8870E+05, 1.8914E+05, 1.8959E+05,
     + 1.9004E+05, 1.9049E+05, 1.9093E+05, 1.9138E+05, 1.9183E+05,
     +  1.9228E+05, 1.9274E+05, 1.9319E+05, 1.9364E+05, 1.9410E+05,
     + 1.9455E+05, 1.9501E+05, 1.9546E+05, 1.9592E+05, 1.9638E+05,
     +  1.9684E+05, 1.9730E+05, 1.9776E+05, 1.9822E+05, 1.9868E+05,
     + 1.9915E+05, 1.9961E+05, 2.0008E+05, 2.0054E+05, 2.0101E+05,
     +  2.0148E+05, 2.0195E+05, 2.0242E+05, 2.0289E+05, 2.0336E+05,
     + 2.0383E+05, 2.0430E+05, 2.0478E+05, 2.0525E+05, 2.0573E+05,
     +  2.0620E+05, 2.0668E+05, 2.0716E+05, 2.0764E+05, 2.0812E+05,
     + 2.0860E+05, 2.0908E+05, 2.0956E+05, 2.1004E+05, 2.1053E+05,
     +  2.1101E+05, 2.1150E+05, 2.1198E+05, 2.1247E+05, 2.1296E+05,
     + 2.1345E+05, 2.1394E+05, 2.1443E+05, 2.1492E+05, 2.1541E+05,
     +  2.1591E+05, 2.1640E+05, 2.1690E+05, 2.1739E+05, 2.1789E+05,
     + 2.1839E+05, 2.1889E+05, 2.1939E+05, 2.1989E+05, 2.2039E+05,
     +  2.2089E+05, 2.2140E+05, 2.2190E+05, 2.2241E+05, 2.2291E+05,
     + 2.2342E+05, 2.2393E+05, 2.2444E+05, 2.2494E+05, 2.2546E+05,
     +  2.2597E+05, 2.2648E+05, 2.2699E+05, 2.2751E+05, 2.2802E+05,
     + 2.2854E+05, 2.2905E+05, 2.2957E+05, 2.3009E+05, 2.3061E+05,
     +  2.3113E+05, 2.3165E+05, 2.3217E+05, 2.3270E+05, 2.3322E+05,
     + 2.3375E+05, 2.3427E+05, 2.3480E+05, 2.3533E+05, 2.3586E+05,
     +  2.3639E+05, 2.3692E+05, 2.3745E+05, 2.3798E+05, 2.3851E+05,
     + 2.3905E+05, 2.3958E+05, 2.4012E+05, 2.4066E+05, 2.4119E+05,
     +  2.4173E+05, 2.4227E+05, 2.4281E+05, 2.4335E+05, 2.4390E+05,
     + 2.4444E+05, 2.4499E+05, 2.4553E+05, 2.4608E+05, 2.4662E+05,
     +  2.4717E+05, 2.4772E+05, 2.4827E+05, 2.4882E+05, 2.4937E+05,
     + 2.4993E+05, 2.5048E+05, 2.5104E+05, 2.5159E+05, 2.5215E+05,
     +  2.5271E+05, 2.5327E+05, 2.5383E+05, 2.5439E+05, 2.5495E+05,
     + 2.5551E+05, 2.5607E+05, 2.5664E+05, 2.5720E+05, 2.5777E+05,
     +  2.5834E+05, 2.5890E+05, 2.5947E+05, 2.6004E+05, 2.6062E+05,
     + 2.6119E+05, 2.6176E+05, 2.6233E+05, 2.6291E+05, 2.6349E+05,
     +  2.6406E+05, 2.6464E+05, 2.6522E+05, 2.6580E+05, 2.6638E+05,
     + 2.6696E+05, 2.6755E+05, 2.6813E+05, 2.6871E+05, 2.6930E+05,
     +  2.6989E+05, 2.7047E+05, 2.7106E+05, 2.7165E+05, 2.7224E+05,
     + 2.7284E+05, 2.7343E+05, 2.7402E+05, 2.7462E+05, 2.7521E+05,
     +  2.7581E+05, 2.7641E+05, 2.7700E+05, 2.7760E+05, 2.7821E+05,
     + 2.7881E+05, 2.7941E+05, 2.8001E+05, 2.8062E+05, 2.8122E+05,
     +  2.8183E+05, 2.8244E+05, 2.8305E+05, 2.8366E+05, 2.8427E+05,
     + 2.8488E+05, 2.8549E+05, 2.8611E+05, 2.8672E+05, 2.8734E+05,
     +  2.8795E+05, 2.8857E+05, 2.8919E+05, 2.8981E+05, 2.9043E+05,
     + 2.9105E+05, 2.9167E+05, 2.9230E+05, 2.9292E+05, 2.9355E+05,
     +  2.9418E+05, 2.9480E+05, 2.9543E+05, 2.9606E+05, 2.9670E+05,
     + 2.9733E+05, 2.9796E+05, 2.9859E+05, 2.9923E+05, 2.9987E+05,
     +  3.0050E+05, 3.0114E+05, 3.0178E+05, 3.0242E+05, 3.0306E+05,
     + 3.0371E+05, 3.0435E+05, 3.0499E+05, 3.0564E+05, 3.0629E+05,
     +  3.0693E+05, 3.0758E+05, 3.0823E+05, 3.0888E+05, 3.0954E+05,
     + 3.1019E+05, 3.1084E+05, 3.1150E+05, 3.1215E+05, 3.1281E+05,
     +  3.1347E+05, 3.1413E+05, 3.1479E+05, 3.1545E+05, 3.1611E+05,
     + 3.1678E+05, 3.1744E+05, 3.1811E+05, 3.1877E+05, 3.1944E+05,
     +  3.2011E+05, 3.2078E+05, 3.2145E+05, 3.2212E+05, 3.2280E+05,
     + 3.2347E+05, 3.2415E+05, 3.2482E+05, 3.2550E+05, 3.2618E+05,
     +  3.2686E+05, 3.2754E+05, 3.2822E+05, 3.2891E+05, 3.2959E+05,
     + 3.3028E+05, 3.3096E+05, 3.3165E+05, 3.3234E+05, 3.3303E+05,
     +  3.3372E+05, 3.3441E+05, 3.3510E+05, 3.3580E+05, 3.3649E+05,
     + 3.3719E+05, 3.3788E+05, 3.3858E+05, 3.3928E+05, 3.3998E+05,
     +  3.4068E+05, 3.4139E+05, 3.4209E+05, 3.4279E+05, 3.4350E+05,
     + 3.4421E+05, 3.4492E+05, 3.4562E+05, 3.4633E+05, 3.4705E+05,
     +  3.4776E+05, 3.4847E+05, 3.4919E+05, 3.4990E+05, 3.5062E+05,
     + 3.5134E+05, 3.5206E+05, 3.5278E+05, 3.5350E+05, 3.5422E+05,
     +  3.5495E+05, 3.5567E+05, 3.5640E+05, 3.5712E+05, 3.5785E+05,
     + 3.5858E+05, 3.5931E+05, 3.6004E+05, 3.6078E+05, 3.6151E+05,
     +  3.6225E+05, 3.6298E+05, 3.6372E+05, 3.6446E+05, 3.6520E+05,
     + 3.6594E+05, 3.6668E+05, 3.6742E+05, 3.6817E+05, 3.6891E+05,
     +  3.6966E+05, 3.7041E+05, 3.7116E+05, 3.7191E+05, 3.7266E+05,
     + 3.7341E+05, 3.7416E+05, 3.7492E+05, 3.7567E+05, 3.7643E+05,
     +  3.7719E+05, 3.7795E+05, 3.7871E+05, 3.7947E+05, 3.8023E+05,
     + 3.8099E+05, 3.8176E+05, 3.8253E+05, 3.8329E+05, 3.8406E+05,
     +  3.8483E+05, 3.8560E+05, 3.8637E+05, 3.8715E+05, 3.8792E+05,
     + 3.8870E+05, 3.8947E+05, 3.9025E+05, 3.9103E+05, 3.9181E+05,
     +  3.9259E+05, 3.9337E+05, 3.9416E+05, 3.9494E+05, 3.9573E+05,
     + 3.9651E+05, 3.9730E+05, 3.9809E+05, 3.9888E+05, 3.9968E+05,
     +  4.0047E+05, 4.0126E+05, 4.0206E+05, 4.0285E+05, 4.0365E+05,
     + 4.0445E+05, 4.0525E+05, 4.0605E+05, 4.0686E+05, 4.0766E+05,
     +  4.0846E+05, 4.0927E+05, 4.1008E+05, 4.1089E+05, 4.1170E+05,
     + 4.1251E+05, 4.1332E+05, 4.1413E+05, 4.1495E+05, 4.1576E+05,
     +  4.1658E+05, 4.1740E+05, 4.1822E+05, 4.1904E+05, 4.1986E+05,
     + 4.2068E+05, 4.2151E+05, 4.2233E+05, 4.2316E+05, 4.2399E+05,
     +  4.2482E+05, 4.2565E+05, 4.2648E+05, 4.2731E+05, 4.2815E+05,
     + 4.2898E+05, 4.2982E+05, 4.3066E+05, 4.3149E+05, 4.3233E+05,
     +  4.3318E+05, 4.3402E+05, 4.3486E+05, 4.3571E+05, 4.3655E+05,
     + 4.3740E+05, 4.3825E+05, 4.3910E+05, 4.3995E+05, 4.4080E+05,
     +  4.4166E+05, 4.4251E+05, 4.4337E+05, 4.4423E+05, 4.4508E+05,
     + 4.4594E+05, 4.4681E+05, 4.4767E+05, 4.4853E+05, 4.4940E+05,
     +  4.5026E+05, 4.5113E+05, 4.5200E+05, 4.5287E+05, 4.5374E+05,
     + 4.5461E+05, 4.5549E+05, 4.5636E+05, 4.5724E+05, 4.5812E+05,
     +  4.5899E+05, 4.5987E+05, 4.6075E+05, 4.6164E+05, 4.6252E+05,
     + 4.6341E+05, 4.6429E+05, 4.6518E+05, 4.6607E+05, 4.6696E+05,
     +  4.6785E+05, 4.6874E+05, 4.6964E+05, 4.7053E+05, 4.7143E+05,
     + 4.7233E+05, 4.7323E+05, 4.7413E+05, 4.7503E+05, 4.7593E+05,
     +  4.7683E+05, 4.7774E+05, 4.7865E+05, 4.7955E+05, 4.8046E+05,
     + 4.8137E+05, 4.8229E+05, 4.8320E+05, 4.8411E+05, 4.8503E+05,
     +  4.8595E+05, 4.8686E+05, 4.8778E+05, 4.8870E+05, 4.8963E+05,
     + 4.9055E+05, 4.9147E+05, 4.9240E+05, 4.9333E+05, 4.9426E+05,
     +  4.9519E+05, 4.9612E+05, 4.9705E+05, 4.9798E+05, 4.9892E+05,
     + 4.9986E+05, 5.0079E+05, 5.0173E+05, 5.0267E+05, 5.0361E+05,
     +  5.0456E+05, 5.0550E+05, 5.0645E+05, 5.0739E+05, 5.0834E+05,
     + 5.0929E+05, 5.1024E+05, 5.1120E+05, 5.1215E+05, 5.1310E+05,
     +  5.1406E+05, 5.1502E+05, 5.1598E+05, 5.1694E+05, 5.1790E+05,
     + 5.1886E+05, 5.1982E+05, 5.2079E+05, 5.2176E+05, 5.2272E+05,
     +  5.2369E+05, 5.2466E+05, 5.2564E+05, 5.2661E+05, 5.2758E+05,
     + 5.2856E+05, 5.2954E+05, 5.3052E+05, 5.3150E+05, 5.3248E+05,
     +  5.3346E+05, 5.3445E+05, 5.3543E+05, 5.3642E+05, 5.3741E+05,
     + 5.3840E+05, 5.3939E+05, 5.4038E+05, 5.4137E+05, 5.4237E+05,
     +  5.4336E+05, 5.4436E+05, 5.4536E+05, 5.4636E+05, 5.4736E+05,
     + 5.4836E+05, 5.4937E+05, 5.5037E+05, 5.5138E+05, 5.5239E+05,
     +  5.5340E+05, 5.5441E+05, 5.5542E+05, 5.5644E+05, 5.5745E+05,
     + 5.5847E+05, 5.5949E+05, 5.6051E+05, 5.6153E+05, 5.6255E+05,
     +  5.6357E+05, 5.6460E+05, 5.6562E+05, 5.6665E+05, 5.6768E+05,
     + 5.6871E+05, 5.6974E+05, 5.7078E+05, 5.7181E+05, 5.7285E+05,
     +  5.7388E+05, 5.7492E+05, 5.7596E+05, 5.7700E+05, 5.7805E+05,
     + 5.7909E+05, 5.8013E+05, 5.8118E+05, 5.8223E+05, 5.8328E+05,
     +  5.8433E+05, 5.8538E+05, 5.8644E+05, 5.8749E+05, 5.8855E+05,
     + 5.8961E+05, 5.9067E+05, 5.9173E+05, 5.9279E+05, 5.9385E+05,
     +  5.9492E+05, 5.9598E+05, 5.9705E+05, 5.9812E+05, 5.9919E+05,
     + 6.0026E+05, 6.0134E+05, 6.0241E+05, 6.0349E+05, 6.0457E+05,
     +  6.0565E+05, 6.0673E+05, 6.0781E+05, 6.0889E+05, 6.0998E+05,
     + 6.1106E+05, 6.1215E+05, 6.1324E+05, 6.1433E+05, 6.1542E+05,
     +  6.1651E+05, 6.1761E+05, 6.1870E+05, 6.1980E+05, 6.2090E+05,
     + 6.2200E+05, 6.2310E+05, 6.2421E+05, 6.2531E+05, 6.2642E+05,
     +  6.2752E+05, 6.2863E+05, 6.2974E+05, 6.3086E+05, 6.3197E+05,
     + 6.3308E+05, 6.3420E+05, 6.3532E+05, 6.3644E+05, 6.3756E+05,
     +  6.3868E+05, 6.3980E+05, 6.4093E+05, 6.4205E+05, 6.4318E+05,
     + 6.4431E+05, 6.4544E+05, 6.4657E+05, 6.4770E+05, 6.4884E+05,
     +  6.4998E+05, 6.5111E+05, 6.5225E+05, 6.5339E+05, 6.5453E+05,
     + 6.5568E+05, 6.5682E+05, 6.5797E+05, 6.5912E+05, 6.6027E+05,
     +  6.6142E+05, 6.6257E+05, 6.6372E+05, 6.6488E+05, 6.6604E+05,
     + 6.6719E+05, 6.6835E+05, 6.6951E+05, 6.7068E+05, 6.7184E+05,
     +  6.7301E+05, 6.7417E+05, 6.7534E+05, 6.7651E+05, 6.7768E+05,
     + 6.7885E+05, 6.8003E+05, 6.8120E+05, 6.8238E+05, 6.8356E+05,
     +  6.8474E+05, 6.8592E+05, 6.8711E+05, 6.8829E+05, 6.8948E+05,
     + 6.9066E+05, 6.9185E+05, 6.9304E+05, 6.9423E+05, 6.9543E+05,
     +  6.9662E+05, 6.9782E+05, 6.9902E+05, 7.0022E+05, 7.0142E+05,
     + 7.0262E+05, 7.0382E+05, 7.0503E+05, 7.0624E+05, 7.0744E+05,
     +  7.0865E+05, 7.0987E+05, 7.1108E+05, 7.1229E+05, 7.1351E+05,
     + 7.1473E+05, 7.1594E+05, 7.1716E+05, 7.1839E+05, 7.1961E+05,
     +  7.2083E+05, 7.2206E+05, 7.2329E+05, 7.2452E+05, 7.2575E+05,
     + 7.2698E+05, 7.2822E+05, 7.2945E+05, 7.3069E+05, 7.3193E+05,
     +  7.3317E+05, 7.3441E+05, 7.3565E+05, 7.3689E+05, 7.3814E+05,
     + 7.3939E+05, 7.4064E+05, 7.4189E+05, 7.4314E+05, 7.4439E+05,
     +  7.4565E+05, 7.4690E+05, 7.4816E+05, 7.4942E+05, 7.5068E+05/
      data (QofT( 1,J),J=2409,4510)/ 7.5195E+05, 7.5321E+05,
     +  7.5448E+05, 7.5574E+05, 7.5701E+05, 7.5828E+05, 7.5955E+05,
     +  7.6083E+05,
     + 7.6210E+05, 7.6338E+05, 7.6466E+05, 7.6594E+05, 7.6722E+05,
     +  7.6850E+05, 7.6978E+05, 7.7107E+05, 7.7236E+05, 7.7365E+05,
     + 7.7494E+05, 7.7623E+05, 7.7752E+05, 7.7882E+05, 7.8011E+05,
     +  7.8141E+05, 7.8271E+05, 7.8401E+05, 7.8531E+05, 7.8662E+05,
     + 7.8792E+05, 7.8923E+05, 7.9054E+05, 7.9185E+05, 7.9316E+05,
     +  7.9448E+05, 7.9579E+05, 7.9711E+05, 7.9843E+05, 7.9975E+05,
     + 8.0107E+05, 8.0239E+05, 8.0371E+05, 8.0504E+05, 8.0637E+05,
     +  8.0770E+05, 8.0903E+05, 8.1036E+05, 8.1169E+05, 8.1303E+05,
     + 8.1437E+05, 8.1570E+05, 8.1704E+05, 8.1839E+05, 8.1973E+05,
     +  8.2107E+05, 8.2242E+05, 8.2377E+05, 8.2512E+05, 8.2647E+05,
     + 8.2782E+05, 8.2918E+05, 8.3053E+05, 8.3189E+05, 8.3325E+05,
     +  8.3461E+05, 8.3597E+05, 8.3734E+05, 8.3870E+05, 8.4007E+05,
     + 8.4144E+05, 8.4281E+05, 8.4418E+05, 8.4555E+05, 8.4693E+05,
     +  8.4830E+05, 8.4968E+05, 8.5106E+05, 8.5244E+05, 8.5382E+05,
     + 8.5521E+05, 8.5659E+05, 8.5798E+05, 8.5937E+05, 8.6076E+05,
     +  8.6215E+05, 8.6355E+05, 8.6494E+05, 8.6634E+05, 8.6774E+05,
     + 8.6914E+05, 8.7054E+05, 8.7194E+05, 8.7335E+05, 8.7476E+05,
     +  8.7616E+05, 8.7757E+05, 8.7899E+05, 8.8040E+05, 8.8181E+05,
     + 8.8323E+05, 8.8465E+05, 8.8607E+05, 8.8749E+05, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       636
      data (QofT( 2,J),J=1,2408)/ 2.3445E+00, 3.9214E+00, 5.6896E+00,
     +  7.4672E+00, 9.2463E+00, 1.1026E+01, 1.2806E+01, 1.4587E+01,
     + 1.6367E+01, 1.8148E+01, 1.9929E+01, 2.1709E+01, 2.3490E+01,
     +  2.5271E+01, 2.7052E+01, 2.8833E+01, 3.0614E+01, 3.2395E+01,
     + 3.4176E+01, 3.5957E+01, 3.9519E+01, 4.3081E+01, 4.6644E+01,
     +  5.0206E+01, 5.3768E+01, 5.7330E+01, 6.0893E+01, 6.4455E+01,
     + 6.8017E+01, 7.1580E+01, 7.5142E+01, 7.8705E+01, 8.2267E+01,
     +  8.5830E+01, 8.9392E+01, 9.2955E+01, 9.6517E+01, 1.0008E+02,
     + 1.0364E+02, 1.0720E+02, 1.1077E+02, 1.1433E+02, 1.1789E+02,
     +  1.2146E+02, 1.2502E+02, 1.2858E+02, 1.3214E+02, 1.3571E+02,
     + 1.3927E+02, 1.4283E+02, 1.4640E+02, 1.4996E+02, 1.5353E+02,
     +  1.5709E+02, 1.6066E+02, 1.6422E+02, 1.6779E+02, 1.7135E+02,
     + 1.7492E+02, 1.7849E+02, 1.8206E+02, 1.8563E+02, 1.8921E+02,
     +  1.9278E+02, 1.9636E+02, 1.9993E+02, 2.0351E+02, 2.0710E+02,
     + 2.1068E+02, 2.1427E+02, 2.1786E+02, 2.2145E+02, 2.2505E+02,
     +  2.2865E+02, 2.3226E+02, 2.3587E+02, 2.3948E+02, 2.4310E+02,
     + 2.4673E+02, 2.5036E+02, 2.5399E+02, 2.5764E+02, 2.6128E+02,
     +  2.6494E+02, 2.6860E+02, 2.7227E+02, 2.7595E+02, 2.7964E+02,
     + 2.8333E+02, 2.8703E+02, 2.9074E+02, 2.9446E+02, 2.9820E+02,
     +  3.0194E+02, 3.0569E+02, 3.0945E+02, 3.1322E+02, 3.1701E+02,
     + 3.2081E+02, 3.2461E+02, 3.2844E+02, 3.3227E+02, 3.3612E+02,
     +  3.3998E+02, 3.4385E+02, 3.4774E+02, 3.5164E+02, 3.5556E+02,
     + 3.5949E+02, 3.6344E+02, 3.6740E+02, 3.7138E+02, 3.7538E+02,
     +  3.7939E+02, 3.8342E+02, 3.8746E+02, 3.9153E+02, 3.9561E+02,
     + 3.9971E+02, 4.0383E+02, 4.0797E+02, 4.1212E+02, 4.1630E+02,
     +  4.2049E+02, 4.2471E+02, 4.2894E+02, 4.3320E+02, 4.3747E+02,
     + 4.4177E+02, 4.4609E+02, 4.5042E+02, 4.5479E+02, 4.5917E+02,
     +  4.6357E+02, 4.6800E+02, 4.7245E+02, 4.7693E+02, 4.8142E+02,
     + 4.8594E+02, 4.9049E+02, 4.9506E+02, 4.9965E+02, 5.0427E+02,
     +  5.0891E+02, 5.1358E+02, 5.1827E+02, 5.2299E+02, 5.2773E+02,
     + 5.3250E+02, 5.3730E+02, 5.4212E+02, 5.4697E+02, 5.5185E+02,
     +  5.5675E+02, 5.6168E+02, 5.6664E+02, 5.7163E+02, 5.7664E+02,
     + 5.8169E+02, 5.8676E+02, 5.9186E+02, 5.9699E+02, 6.0215E+02,
     +  6.0734E+02, 6.1255E+02, 6.1780E+02, 6.2308E+02, 6.2839E+02,
     + 6.3373E+02, 6.3910E+02, 6.4450E+02, 6.4993E+02, 6.5540E+02,
     +  6.6089E+02, 6.6642E+02, 6.7198E+02, 6.7757E+02, 6.8319E+02,
     + 6.8885E+02, 6.9454E+02, 7.0026E+02, 7.0602E+02, 7.1181E+02,
     +  7.1764E+02, 7.2349E+02, 7.2939E+02, 7.3531E+02, 7.4127E+02,
     + 7.4727E+02, 7.5330E+02, 7.5937E+02, 7.6547E+02, 7.7161E+02,
     +  7.7779E+02, 7.8400E+02, 7.9024E+02, 7.9653E+02, 8.0285E+02,
     + 8.0921E+02, 8.1560E+02, 8.2203E+02, 8.2850E+02, 8.3501E+02,
     +  8.4156E+02, 8.4814E+02, 8.5477E+02, 8.6143E+02, 8.6813E+02,
     + 8.7487E+02, 8.8165E+02, 8.8847E+02, 8.9533E+02, 9.0223E+02,
     +  9.0917E+02, 9.1615E+02, 9.2317E+02, 9.3023E+02, 9.3733E+02,
     + 9.4448E+02, 9.5166E+02, 9.5889E+02, 9.6616E+02, 9.7348E+02,
     +  9.8083E+02, 9.8823E+02, 9.9567E+02, 1.0032E+03, 1.0107E+03,
     + 1.0183E+03, 1.0259E+03, 1.0335E+03, 1.0412E+03, 1.0490E+03,
     +  1.0568E+03, 1.0646E+03, 1.0725E+03, 1.0804E+03, 1.0884E+03,
     + 1.0964E+03, 1.1045E+03, 1.1126E+03, 1.1207E+03, 1.1290E+03,
     +  1.1372E+03, 1.1455E+03, 1.1538E+03, 1.1622E+03, 1.1707E+03,
     + 1.1792E+03, 1.1877E+03, 1.1963E+03, 1.2049E+03, 1.2136E+03,
     +  1.2224E+03, 1.2312E+03, 1.2400E+03, 1.2489E+03, 1.2578E+03,
     + 1.2668E+03, 1.2758E+03, 1.2849E+03, 1.2941E+03, 1.3033E+03,
     +  1.3125E+03, 1.3218E+03, 1.3311E+03, 1.3405E+03, 1.3500E+03,
     + 1.3595E+03, 1.3691E+03, 1.3787E+03, 1.3883E+03, 1.3981E+03,
     +  1.4078E+03, 1.4177E+03, 1.4275E+03, 1.4375E+03, 1.4475E+03,
     + 1.4575E+03, 1.4676E+03, 1.4778E+03, 1.4880E+03, 1.4983E+03,
     +  1.5086E+03, 1.5190E+03, 1.5294E+03, 1.5399E+03, 1.5505E+03,
     + 1.5611E+03, 1.5718E+03, 1.5825E+03, 1.5933E+03, 1.6042E+03,
     +  1.6151E+03, 1.6261E+03, 1.6371E+03, 1.6482E+03, 1.6594E+03,
     + 1.6706E+03, 1.6818E+03, 1.6932E+03, 1.7046E+03, 1.7160E+03,
     +  1.7276E+03, 1.7392E+03, 1.7508E+03, 1.7625E+03, 1.7743E+03,
     + 1.7861E+03, 1.7980E+03, 1.8100E+03, 1.8220E+03, 1.8341E+03,
     +  1.8463E+03, 1.8585E+03, 1.8708E+03, 1.8831E+03, 1.8956E+03,
     + 1.9081E+03, 1.9206E+03, 1.9332E+03, 1.9459E+03, 1.9587E+03,
     +  1.9715E+03, 1.9844E+03, 1.9973E+03, 2.0104E+03, 2.0235E+03,
     + 2.0366E+03, 2.0499E+03, 2.0632E+03, 2.0766E+03, 2.0900E+03,
     +  2.1035E+03, 2.1171E+03, 2.1308E+03, 2.1445E+03, 2.1583E+03,
     + 2.1722E+03, 2.1861E+03, 2.2001E+03, 2.2142E+03, 2.2284E+03,
     +  2.2427E+03, 2.2570E+03, 2.2714E+03, 2.2858E+03, 2.3004E+03,
     + 2.3150E+03, 2.3297E+03, 2.3444E+03, 2.3593E+03, 2.3742E+03,
     +  2.3892E+03, 2.4043E+03, 2.4194E+03, 2.4346E+03, 2.4499E+03,
     + 2.4653E+03, 2.4808E+03, 2.4963E+03, 2.5119E+03, 2.5276E+03,
     +  2.5434E+03, 2.5593E+03, 2.5752E+03, 2.5912E+03, 2.6074E+03,
     + 2.6235E+03, 2.6398E+03, 2.6562E+03, 2.6726E+03, 2.6891E+03,
     +  2.7057E+03, 2.7224E+03, 2.7392E+03, 2.7560E+03, 2.7729E+03,
     + 2.7900E+03, 2.8071E+03, 2.8243E+03, 2.8415E+03, 2.8589E+03,
     +  2.8763E+03, 2.8939E+03, 2.9115E+03, 2.9292E+03, 2.9470E+03,
     + 2.9649E+03, 2.9829E+03, 3.0010E+03, 3.0191E+03, 3.0374E+03,
     +  3.0557E+03, 3.0741E+03, 3.0926E+03, 3.1112E+03, 3.1299E+03,
     + 3.1487E+03, 3.1676E+03, 3.1866E+03, 3.2057E+03, 3.2248E+03,
     +  3.2441E+03, 3.2634E+03, 3.2829E+03, 3.3024E+03, 3.3221E+03,
     + 3.3418E+03, 3.3616E+03, 3.3815E+03, 3.4016E+03, 3.4217E+03,
     +  3.4419E+03, 3.4622E+03, 3.4826E+03, 3.5031E+03, 3.5237E+03,
     + 3.5444E+03, 3.5652E+03, 3.5861E+03, 3.6072E+03, 3.6283E+03,
     +  3.6495E+03, 3.6708E+03, 3.6922E+03, 3.7137E+03, 3.7353E+03,
     + 3.7570E+03, 3.7789E+03, 3.8008E+03, 3.8228E+03, 3.8450E+03,
     +  3.8672E+03, 3.8895E+03, 3.9120E+03, 3.9346E+03, 3.9572E+03,
     + 3.9800E+03, 4.0029E+03, 4.0259E+03, 4.0490E+03, 4.0722E+03,
     +  4.0955E+03, 4.1189E+03, 4.1424E+03, 4.1661E+03, 4.1898E+03,
     + 4.2137E+03, 4.2377E+03, 4.2617E+03, 4.2859E+03, 4.3103E+03,
     +  4.3347E+03, 4.3592E+03, 4.3839E+03, 4.4086E+03, 4.4335E+03,
     + 4.4585E+03, 4.4836E+03, 4.5088E+03, 4.5342E+03, 4.5596E+03,
     +  4.5852E+03, 4.6109E+03, 4.6367E+03, 4.6627E+03, 4.6887E+03,
     + 4.7149E+03, 4.7412E+03, 4.7676E+03, 4.7941E+03, 4.8208E+03,
     +  4.8475E+03, 4.8744E+03, 4.9014E+03, 4.9286E+03, 4.9559E+03,
     + 4.9832E+03, 5.0107E+03, 5.0384E+03, 5.0661E+03, 5.0940E+03,
     +  5.1220E+03, 5.1502E+03, 5.1784E+03, 5.2068E+03, 5.2353E+03,
     + 5.2640E+03, 5.2928E+03, 5.3217E+03, 5.3507E+03, 5.3799E+03,
     +  5.4092E+03, 5.4386E+03, 5.4681E+03, 5.4978E+03, 5.5277E+03,
     + 5.5576E+03, 5.5877E+03, 5.6179E+03, 5.6483E+03, 5.6787E+03,
     +  5.7094E+03, 5.7401E+03, 5.7710E+03, 5.8020E+03, 5.8332E+03,
     + 5.8645E+03, 5.8960E+03, 5.9275E+03, 5.9593E+03, 5.9911E+03,
     +  6.0231E+03, 6.0552E+03, 6.0875E+03, 6.1199E+03, 6.1525E+03,
     + 6.1852E+03, 6.2180E+03, 6.2510E+03, 6.2842E+03, 6.3174E+03,
     +  6.3509E+03, 6.3844E+03, 6.4181E+03, 6.4520E+03, 6.4860E+03,
     + 6.5201E+03, 6.5544E+03, 6.5889E+03, 6.6235E+03, 6.6582E+03,
     +  6.6931E+03, 6.7281E+03, 6.7633E+03, 6.7987E+03, 6.8342E+03,
     + 6.8698E+03, 6.9056E+03, 6.9415E+03, 6.9777E+03, 7.0139E+03,
     +  7.0503E+03, 7.0869E+03, 7.1236E+03, 7.1605E+03, 7.1975E+03,
     + 7.2347E+03, 7.2721E+03, 7.3096E+03, 7.3472E+03, 7.3850E+03,
     +  7.4230E+03, 7.4612E+03, 7.4995E+03, 7.5379E+03, 7.5766E+03,
     + 7.6154E+03, 7.6543E+03, 7.6934E+03, 7.7327E+03, 7.7721E+03,
     +  7.8118E+03, 7.8515E+03, 7.8915E+03, 7.9316E+03, 7.9718E+03,
     + 8.0123E+03, 8.0529E+03, 8.0937E+03, 8.1346E+03, 8.1757E+03,
     +  8.2170E+03, 8.2585E+03, 8.3001E+03, 8.3419E+03, 8.3839E+03,
     + 8.4260E+03, 8.4683E+03, 8.5108E+03, 8.5535E+03, 8.5963E+03,
     +  8.6393E+03, 8.6825E+03, 8.7259E+03, 8.7694E+03, 8.8132E+03,
     + 8.8571E+03, 8.9012E+03, 8.9454E+03, 8.9899E+03, 9.0345E+03,
     +  9.0793E+03, 9.1243E+03, 9.1694E+03, 9.2148E+03, 9.2603E+03,
     + 9.3061E+03, 9.3520E+03, 9.3981E+03, 9.4443E+03, 9.4908E+03,
     +  9.5374E+03, 9.5843E+03, 9.6313E+03, 9.6785E+03, 9.7259E+03,
     + 9.7735E+03, 9.8213E+03, 9.8693E+03, 9.9174E+03, 9.9658E+03,
     +  1.0014E+04, 1.0063E+04, 1.0112E+04, 1.0161E+04, 1.0211E+04,
     + 1.0260E+04, 1.0310E+04, 1.0360E+04, 1.0410E+04, 1.0460E+04,
     +  1.0511E+04, 1.0561E+04, 1.0612E+04, 1.0663E+04, 1.0715E+04,
     + 1.0766E+04, 1.0818E+04, 1.0870E+04, 1.0922E+04, 1.0974E+04,
     +  1.1027E+04, 1.1080E+04, 1.1133E+04, 1.1186E+04, 1.1239E+04,
     + 1.1293E+04, 1.1346E+04, 1.1400E+04, 1.1455E+04, 1.1509E+04,
     +  1.1564E+04, 1.1619E+04, 1.1674E+04, 1.1729E+04, 1.1784E+04,
     + 1.1840E+04, 1.1896E+04, 1.1952E+04, 1.2008E+04, 1.2065E+04,
     +  1.2122E+04, 1.2179E+04, 1.2236E+04, 1.2293E+04, 1.2351E+04,
     + 1.2409E+04, 1.2467E+04, 1.2525E+04, 1.2584E+04, 1.2643E+04,
     +  1.2702E+04, 1.2761E+04, 1.2820E+04, 1.2880E+04, 1.2940E+04,
     + 1.3000E+04, 1.3060E+04, 1.3121E+04, 1.3182E+04, 1.3243E+04,
     +  1.3304E+04, 1.3365E+04, 1.3427E+04, 1.3489E+04, 1.3551E+04,
     + 1.3614E+04, 1.3676E+04, 1.3739E+04, 1.3802E+04, 1.3866E+04,
     +  1.3929E+04, 1.3993E+04, 1.4057E+04, 1.4122E+04, 1.4186E+04,
     + 1.4251E+04, 1.4316E+04, 1.4381E+04, 1.4447E+04, 1.4513E+04,
     +  1.4579E+04, 1.4645E+04, 1.4712E+04, 1.4778E+04, 1.4845E+04,
     + 1.4913E+04, 1.4980E+04, 1.5048E+04, 1.5116E+04, 1.5184E+04,
     +  1.5253E+04, 1.5321E+04, 1.5390E+04, 1.5460E+04, 1.5529E+04,
     + 1.5599E+04, 1.5669E+04, 1.5739E+04, 1.5810E+04, 1.5881E+04,
     +  1.5952E+04, 1.6023E+04, 1.6095E+04, 1.6166E+04, 1.6239E+04,
     + 1.6311E+04, 1.6384E+04, 1.6456E+04, 1.6530E+04, 1.6603E+04,
     +  1.6677E+04, 1.6751E+04, 1.6825E+04, 1.6899E+04, 1.6974E+04,
     + 1.7049E+04, 1.7125E+04, 1.7200E+04, 1.7276E+04, 1.7352E+04,
     +  1.7428E+04, 1.7505E+04, 1.7582E+04, 1.7659E+04, 1.7737E+04,
     + 1.7815E+04, 1.7893E+04, 1.7971E+04, 1.8050E+04, 1.8128E+04,
     +  1.8208E+04, 1.8287E+04, 1.8367E+04, 1.8447E+04, 1.8527E+04,
     + 1.8608E+04, 1.8689E+04, 1.8770E+04, 1.8851E+04, 1.8933E+04,
     +  1.9015E+04, 1.9097E+04, 1.9180E+04, 1.9263E+04, 1.9346E+04,
     + 1.9430E+04, 1.9513E+04, 1.9597E+04, 1.9682E+04, 1.9766E+04,
     +  1.9851E+04, 1.9937E+04, 2.0022E+04, 2.0108E+04, 2.0194E+04,
     + 2.0281E+04, 2.0367E+04, 2.0454E+04, 2.0542E+04, 2.0629E+04,
     +  2.0717E+04, 2.0806E+04, 2.0894E+04, 2.0983E+04, 2.1072E+04,
     + 2.1162E+04, 2.1252E+04, 2.1342E+04, 2.1432E+04, 2.1523E+04,
     +  2.1614E+04, 2.1705E+04, 2.1797E+04, 2.1889E+04, 2.1982E+04,
     + 2.2074E+04, 2.2167E+04, 2.2260E+04, 2.2354E+04, 2.2448E+04,
     +  2.2542E+04, 2.2637E+04, 2.2732E+04, 2.2827E+04, 2.2922E+04,
     + 2.3018E+04, 2.3114E+04, 2.3211E+04, 2.3308E+04, 2.3405E+04,
     +  2.3502E+04, 2.3600E+04, 2.3698E+04, 2.3797E+04, 2.3896E+04,
     + 2.3995E+04, 2.4094E+04, 2.4194E+04, 2.4294E+04, 2.4395E+04,
     +  2.4496E+04, 2.4597E+04, 2.4698E+04, 2.4800E+04, 2.4902E+04,
     + 2.5005E+04, 2.5108E+04, 2.5211E+04, 2.5315E+04, 2.5419E+04,
     +  2.5523E+04, 2.5627E+04, 2.5732E+04, 2.5838E+04, 2.5943E+04,
     + 2.6049E+04, 2.6156E+04, 2.6263E+04, 2.6370E+04, 2.6477E+04,
     +  2.6585E+04, 2.6693E+04, 2.6802E+04, 2.6910E+04, 2.7020E+04,
     + 2.7129E+04, 2.7239E+04, 2.7349E+04, 2.7460E+04, 2.7571E+04,
     +  2.7683E+04, 2.7794E+04, 2.7906E+04, 2.8019E+04, 2.8132E+04,
     + 2.8245E+04, 2.8359E+04, 2.8473E+04, 2.8587E+04, 2.8702E+04,
     +  2.8817E+04, 2.8932E+04, 2.9048E+04, 2.9164E+04, 2.9281E+04,
     + 2.9398E+04, 2.9515E+04, 2.9633E+04, 2.9751E+04, 2.9870E+04,
     +  2.9989E+04, 3.0108E+04, 3.0228E+04, 3.0348E+04, 3.0468E+04,
     + 3.0589E+04, 3.0710E+04, 3.0832E+04, 3.0954E+04, 3.1076E+04,
     +  3.1199E+04, 3.1322E+04, 3.1446E+04, 3.1569E+04, 3.1694E+04,
     + 3.1819E+04, 3.1944E+04, 3.2069E+04, 3.2195E+04, 3.2322E+04,
     +  3.2448E+04, 3.2576E+04, 3.2703E+04, 3.2831E+04, 3.2959E+04,
     + 3.3088E+04, 3.3217E+04, 3.3347E+04, 3.3477E+04, 3.3607E+04,
     +  3.3738E+04, 3.3870E+04, 3.4001E+04, 3.4133E+04, 3.4266E+04,
     + 3.4399E+04, 3.4532E+04, 3.4666E+04, 3.4800E+04, 3.4935E+04,
     +  3.5070E+04, 3.5205E+04, 3.5341E+04, 3.5477E+04, 3.5614E+04,
     + 3.5751E+04, 3.5889E+04, 3.6027E+04, 3.6165E+04, 3.6304E+04,
     +  3.6443E+04, 3.6583E+04, 3.6723E+04, 3.6864E+04, 3.7005E+04,
     + 3.7146E+04, 3.7288E+04, 3.7430E+04, 3.7573E+04, 3.7717E+04,
     +  3.7860E+04, 3.8004E+04, 3.8149E+04, 3.8294E+04, 3.8439E+04,
     + 3.8585E+04, 3.8732E+04, 3.8878E+04, 3.9026E+04, 3.9173E+04,
     +  3.9322E+04, 3.9470E+04, 3.9619E+04, 3.9769E+04, 3.9919E+04,
     + 4.0069E+04, 4.0220E+04, 4.0371E+04, 4.0523E+04, 4.0676E+04,
     +  4.0828E+04, 4.0981E+04, 4.1135E+04, 4.1289E+04, 4.1444E+04,
     + 4.1599E+04, 4.1755E+04, 4.1911E+04, 4.2067E+04, 4.2224E+04,
     +  4.2381E+04, 4.2539E+04, 4.2698E+04, 4.2857E+04, 4.3016E+04,
     + 4.3176E+04, 4.3336E+04, 4.3497E+04, 4.3658E+04, 4.3820E+04,
     +  4.3982E+04, 4.4145E+04, 4.4308E+04, 4.4472E+04, 4.4636E+04,
     + 4.4801E+04, 4.4966E+04, 4.5132E+04, 4.5298E+04, 4.5464E+04,
     +  4.5632E+04, 4.5799E+04, 4.5967E+04, 4.6136E+04, 4.6305E+04,
     + 4.6475E+04, 4.6645E+04, 4.6816E+04, 4.6987E+04, 4.7159E+04,
     +  4.7331E+04, 4.7503E+04, 4.7677E+04, 4.7850E+04, 4.8025E+04,
     + 4.8199E+04, 4.8375E+04, 4.8550E+04, 4.8727E+04, 4.8903E+04,
     +  4.9081E+04, 4.9259E+04, 4.9437E+04, 4.9616E+04, 4.9795E+04,
     + 4.9975E+04, 5.0156E+04, 5.0337E+04, 5.0518E+04, 5.0700E+04,
     +  5.0883E+04, 5.1066E+04, 5.1249E+04, 5.1434E+04, 5.1618E+04,
     + 5.1804E+04, 5.1989E+04, 5.2176E+04, 5.2363E+04, 5.2550E+04,
     +  5.2738E+04, 5.2926E+04, 5.3115E+04, 5.3305E+04, 5.3495E+04,
     + 5.3686E+04, 5.3877E+04, 5.4069E+04, 5.4261E+04, 5.4454E+04,
     +  5.4647E+04, 5.4841E+04, 5.5036E+04, 5.5231E+04, 5.5427E+04,
     + 5.5623E+04, 5.5820E+04, 5.6017E+04, 5.6215E+04, 5.6413E+04,
     +  5.6612E+04, 5.6812E+04, 5.7012E+04, 5.7213E+04, 5.7414E+04,
     + 5.7616E+04, 5.7818E+04, 5.8022E+04, 5.8225E+04, 5.8429E+04,
     +  5.8634E+04, 5.8839E+04, 5.9045E+04, 5.9252E+04, 5.9459E+04,
     + 5.9667E+04, 5.9875E+04, 6.0084E+04, 6.0293E+04, 6.0503E+04,
     +  6.0714E+04, 6.0925E+04, 6.1137E+04, 6.1349E+04, 6.1562E+04,
     + 6.1776E+04, 6.1990E+04, 6.2205E+04, 6.2420E+04, 6.2636E+04,
     +  6.2853E+04, 6.3070E+04, 6.3287E+04, 6.3506E+04, 6.3725E+04,
     + 6.3945E+04, 6.4165E+04, 6.4386E+04, 6.4607E+04, 6.4829E+04,
     +  6.5052E+04, 6.5275E+04, 6.5499E+04, 6.5724E+04, 6.5949E+04,
     + 6.6174E+04, 6.6401E+04, 6.6628E+04, 6.6856E+04, 6.7084E+04,
     +  6.7313E+04, 6.7542E+04, 6.7772E+04, 6.8003E+04, 6.8235E+04,
     + 6.8467E+04, 6.8699E+04, 6.8933E+04, 6.9167E+04, 6.9401E+04,
     +  6.9637E+04, 6.9873E+04, 7.0109E+04, 7.0346E+04, 7.0584E+04,
     + 7.0823E+04, 7.1062E+04, 7.1302E+04, 7.1542E+04, 7.1783E+04,
     +  7.2025E+04, 7.2267E+04, 7.2510E+04, 7.2754E+04, 7.2998E+04,
     + 7.3243E+04, 7.3489E+04, 7.3735E+04, 7.3983E+04, 7.4230E+04,
     +  7.4479E+04, 7.4728E+04, 7.4977E+04, 7.5228E+04, 7.5479E+04,
     + 7.5731E+04, 7.5983E+04, 7.6236E+04, 7.6490E+04, 7.6744E+04,
     +  7.6999E+04, 7.7255E+04, 7.7512E+04, 7.7769E+04, 7.8027E+04,
     + 7.8285E+04, 7.8544E+04, 7.8804E+04, 7.9065E+04, 7.9326E+04,
     +  7.9588E+04, 7.9851E+04, 8.0114E+04, 8.0379E+04, 8.0643E+04,
     + 8.0909E+04, 8.1175E+04, 8.1442E+04, 8.1710E+04, 8.1978E+04,
     +  8.2247E+04, 8.2517E+04, 8.2787E+04, 8.3058E+04, 8.3330E+04,
     + 8.3603E+04, 8.3876E+04, 8.4150E+04, 8.4425E+04, 8.4701E+04,
     +  8.4977E+04, 8.5254E+04, 8.5531E+04, 8.5810E+04, 8.6089E+04,
     + 8.6369E+04, 8.6649E+04, 8.6931E+04, 8.7213E+04, 8.7495E+04,
     +  8.7779E+04, 8.8063E+04, 8.8348E+04, 8.8634E+04, 8.8920E+04,
     + 8.9208E+04, 8.9496E+04, 8.9784E+04, 9.0074E+04, 9.0364E+04,
     +  9.0655E+04, 9.0947E+04, 9.1239E+04, 9.1532E+04, 9.1826E+04,
     + 9.2121E+04, 9.2417E+04, 9.2713E+04, 9.3010E+04, 9.3308E+04,
     +  9.3606E+04, 9.3906E+04, 9.4206E+04, 9.4507E+04, 9.4808E+04,
     + 9.5111E+04, 9.5414E+04, 9.5718E+04, 9.6023E+04, 9.6328E+04,
     +  9.6634E+04, 9.6942E+04, 9.7249E+04, 9.7558E+04, 9.7868E+04,
     + 9.8178E+04, 9.8489E+04, 9.8801E+04, 9.9113E+04, 9.9427E+04,
     +  9.9741E+04, 1.0006E+05, 1.0037E+05, 1.0069E+05, 1.0101E+05,
     + 1.0132E+05, 1.0164E+05, 1.0196E+05, 1.0228E+05, 1.0260E+05,
     +  1.0293E+05, 1.0325E+05, 1.0357E+05, 1.0390E+05, 1.0422E+05,
     + 1.0455E+05, 1.0488E+05, 1.0521E+05, 1.0553E+05, 1.0586E+05,
     +  1.0619E+05, 1.0653E+05, 1.0686E+05, 1.0719E+05, 1.0752E+05,
     + 1.0786E+05, 1.0819E+05, 1.0853E+05, 1.0887E+05, 1.0921E+05,
     +  1.0954E+05, 1.0988E+05, 1.1022E+05, 1.1057E+05, 1.1091E+05,
     + 1.1125E+05, 1.1159E+05, 1.1194E+05, 1.1228E+05, 1.1263E+05,
     +  1.1298E+05, 1.1333E+05, 1.1368E+05, 1.1403E+05, 1.1438E+05,
     + 1.1473E+05, 1.1508E+05, 1.1543E+05, 1.1579E+05, 1.1614E+05,
     +  1.1650E+05, 1.1686E+05, 1.1721E+05, 1.1757E+05, 1.1793E+05,
     + 1.1829E+05, 1.1865E+05, 1.1901E+05, 1.1938E+05, 1.1974E+05,
     +  1.2011E+05, 1.2047E+05, 1.2084E+05, 1.2121E+05, 1.2157E+05,
     + 1.2194E+05, 1.2231E+05, 1.2268E+05, 1.2306E+05, 1.2343E+05,
     +  1.2380E+05, 1.2418E+05, 1.2455E+05, 1.2493E+05, 1.2531E+05,
     + 1.2569E+05, 1.2606E+05, 1.2644E+05, 1.2683E+05, 1.2721E+05,
     +  1.2759E+05, 1.2797E+05, 1.2836E+05, 1.2874E+05, 1.2913E+05,
     + 1.2952E+05, 1.2991E+05, 1.3030E+05, 1.3069E+05, 1.3108E+05,
     +  1.3147E+05, 1.3186E+05, 1.3226E+05, 1.3265E+05, 1.3305E+05,
     + 1.3344E+05, 1.3384E+05, 1.3424E+05, 1.3464E+05, 1.3504E+05,
     +  1.3544E+05, 1.3585E+05, 1.3625E+05, 1.3665E+05, 1.3706E+05,
     + 1.3747E+05, 1.3787E+05, 1.3828E+05, 1.3869E+05, 1.3910E+05,
     +  1.3951E+05, 1.3992E+05, 1.4034E+05, 1.4075E+05, 1.4117E+05,
     + 1.4158E+05, 1.4200E+05, 1.4242E+05, 1.4284E+05, 1.4326E+05,
     +  1.4368E+05, 1.4410E+05, 1.4452E+05, 1.4495E+05, 1.4537E+05,
     + 1.4580E+05, 1.4622E+05, 1.4665E+05, 1.4708E+05, 1.4751E+05,
     +  1.4794E+05, 1.4837E+05, 1.4881E+05, 1.4924E+05, 1.4968E+05,
     + 1.5011E+05, 1.5055E+05, 1.5099E+05, 1.5143E+05, 1.5187E+05,
     +  1.5231E+05, 1.5275E+05, 1.5319E+05, 1.5364E+05, 1.5408E+05,
     + 1.5453E+05, 1.5498E+05, 1.5543E+05, 1.5587E+05, 1.5632E+05,
     +  1.5678E+05, 1.5723E+05, 1.5768E+05, 1.5814E+05, 1.5859E+05,
     + 1.5905E+05, 1.5951E+05, 1.5997E+05, 1.6043E+05, 1.6089E+05,
     +  1.6135E+05, 1.6181E+05, 1.6228E+05, 1.6274E+05, 1.6321E+05,
     + 1.6367E+05, 1.6414E+05, 1.6461E+05, 1.6508E+05, 1.6555E+05,
     +  1.6603E+05, 1.6650E+05, 1.6697E+05, 1.6745E+05, 1.6793E+05,
     + 1.6841E+05, 1.6888E+05, 1.6936E+05, 1.6985E+05, 1.7033E+05,
     +  1.7081E+05, 1.7130E+05, 1.7178E+05, 1.7227E+05, 1.7276E+05,
     + 1.7325E+05, 1.7374E+05, 1.7423E+05, 1.7472E+05, 1.7521E+05,
     +  1.7571E+05, 1.7620E+05, 1.7670E+05, 1.7720E+05, 1.7770E+05,
     + 1.7820E+05, 1.7870E+05, 1.7920E+05, 1.7970E+05, 1.8021E+05,
     +  1.8071E+05, 1.8122E+05, 1.8173E+05, 1.8224E+05, 1.8275E+05,
     + 1.8326E+05, 1.8377E+05, 1.8428E+05, 1.8480E+05, 1.8531E+05,
     +  1.8583E+05, 1.8635E+05, 1.8687E+05, 1.8739E+05, 1.8791E+05,
     + 1.8843E+05, 1.8896E+05, 1.8948E+05, 1.9001E+05, 1.9054E+05,
     +  1.9106E+05, 1.9159E+05, 1.9213E+05, 1.9266E+05, 1.9319E+05,
     + 1.9373E+05, 1.9426E+05, 1.9480E+05, 1.9534E+05, 1.9587E+05,
     +  1.9642E+05, 1.9696E+05, 1.9750E+05, 1.9804E+05, 1.9859E+05,
     + 1.9913E+05, 1.9968E+05, 2.0023E+05, 2.0078E+05, 2.0133E+05,
     +  2.0188E+05, 2.0244E+05, 2.0299E+05, 2.0355E+05, 2.0411E+05,
     + 2.0466E+05, 2.0522E+05, 2.0578E+05, 2.0635E+05, 2.0691E+05,
     +  2.0747E+05, 2.0804E+05, 2.0861E+05, 2.0917E+05, 2.0974E+05,
     + 2.1031E+05, 2.1089E+05, 2.1146E+05, 2.1203E+05, 2.1261E+05,
     +  2.1319E+05, 2.1376E+05, 2.1434E+05, 2.1492E+05, 2.1550E+05,
     + 2.1609E+05, 2.1667E+05, 2.1726E+05, 2.1784E+05, 2.1843E+05,
     +  2.1902E+05, 2.1961E+05, 2.2020E+05, 2.2080E+05, 2.2139E+05,
     + 2.2199E+05, 2.2258E+05, 2.2318E+05, 2.2378E+05, 2.2438E+05,
     +  2.2498E+05, 2.2559E+05, 2.2619E+05, 2.2680E+05, 2.2740E+05,
     + 2.2801E+05, 2.2862E+05, 2.2923E+05, 2.2985E+05, 2.3046E+05,
     +  2.3107E+05, 2.3169E+05, 2.3231E+05, 2.3293E+05, 2.3355E+05,
     + 2.3417E+05, 2.3479E+05, 2.3541E+05, 2.3604E+05, 2.3667E+05,
     +  2.3729E+05, 2.3792E+05, 2.3855E+05, 2.3919E+05, 2.3982E+05,
     + 2.4045E+05, 2.4109E+05, 2.4173E+05, 2.4237E+05, 2.4301E+05,
     +  2.4365E+05, 2.4429E+05, 2.4493E+05, 2.4558E+05, 2.4622E+05,
     + 2.4687E+05, 2.4752E+05, 2.4817E+05, 2.4882E+05, 2.4948E+05,
     +  2.5013E+05, 2.5079E+05, 2.5145E+05, 2.5211E+05, 2.5277E+05,
     + 2.5343E+05, 2.5409E+05, 2.5475E+05, 2.5542E+05, 2.5609E+05,
     +  2.5676E+05, 2.5743E+05, 2.5810E+05, 2.5877E+05, 2.5944E+05,
     + 2.6012E+05, 2.6080E+05, 2.6147E+05, 2.6215E+05, 2.6283E+05,
     +  2.6352E+05, 2.6420E+05, 2.6488E+05, 2.6557E+05, 2.6626E+05,
     + 2.6695E+05, 2.6764E+05, 2.6833E+05, 2.6903E+05, 2.6972E+05,
     +  2.7042E+05, 2.7111E+05, 2.7181E+05, 2.7251E+05, 2.7322E+05,
     + 2.7392E+05, 2.7462E+05, 2.7533E+05, 2.7604E+05, 2.7675E+05,
     +  2.7746E+05, 2.7817E+05, 2.7888E+05, 2.7960E+05, 2.8032E+05,
     + 2.8103E+05, 2.8175E+05, 2.8247E+05, 2.8320E+05, 2.8392E+05,
     +  2.8465E+05, 2.8537E+05, 2.8610E+05, 2.8683E+05, 2.8756E+05,
     + 2.8829E+05, 2.8903E+05, 2.8976E+05, 2.9050E+05, 2.9124E+05,
     +  2.9198E+05, 2.9272E+05, 2.9346E+05, 2.9421E+05, 2.9495E+05,
     + 2.9570E+05, 2.9645E+05, 2.9720E+05, 2.9795E+05, 2.9871E+05,
     +  2.9946E+05, 3.0022E+05, 3.0097E+05, 3.0173E+05, 3.0249E+05,
     + 3.0326E+05, 3.0402E+05, 3.0479E+05, 3.0555E+05, 3.0632E+05,
     +  3.0709E+05, 3.0786E+05, 3.0864E+05, 3.0941E+05, 3.1019E+05,
     + 3.1096E+05, 3.1174E+05, 3.1252E+05, 3.1331E+05, 3.1409E+05,
     +  3.1488E+05, 3.1566E+05, 3.1645E+05, 3.1724E+05, 3.1803E+05,
     + 3.1883E+05, 3.1962E+05, 3.2042E+05, 3.2121E+05, 3.2201E+05,
     +  3.2281E+05, 3.2362E+05, 3.2442E+05, 3.2523E+05, 3.2603E+05,
     + 3.2684E+05, 3.2765E+05, 3.2846E+05, 3.2928E+05, 3.3009E+05,
     +  3.3091E+05, 3.3173E+05, 3.3255E+05, 3.3337E+05, 3.3419E+05,
     + 3.3502E+05, 3.3584E+05, 3.3667E+05, 3.3750E+05, 3.3833E+05,
     +  3.3916E+05, 3.4000E+05, 3.4083E+05, 3.4167E+05, 3.4251E+05,
     + 3.4335E+05, 3.4419E+05, 3.4504E+05, 3.4588E+05, 3.4673E+05,
     +  3.4758E+05, 3.4843E+05, 3.4928E+05, 3.5014E+05, 3.5099E+05,
     + 3.5185E+05, 3.5271E+05, 3.5357E+05, 3.5443E+05, 3.5529E+05,
     +  3.5616E+05, 3.5702E+05, 3.5789E+05, 3.5876E+05, 3.5963E+05,
     + 3.6051E+05, 3.6138E+05, 3.6226E+05, 3.6314E+05, 3.6402E+05,
     +  3.6490E+05, 3.6578E+05, 3.6667E+05, 3.6756E+05, 3.6844E+05,
     + 3.6934E+05, 3.7023E+05, 3.7112E+05, 3.7202E+05, 3.7291E+05,
     +  3.7381E+05, 3.7471E+05, 3.7561E+05, 3.7652E+05, 3.7742E+05,
     + 3.7833E+05, 3.7924E+05, 3.8015E+05, 3.8106E+05, 3.8198E+05,
     +  3.8289E+05, 3.8381E+05, 3.8473E+05, 3.8565E+05, 3.8657E+05,
     + 3.8750E+05, 3.8842E+05, 3.8935E+05, 3.9028E+05, 3.9121E+05,
     +  3.9214E+05, 3.9308E+05, 3.9402E+05, 3.9495E+05, 3.9589E+05,
     + 3.9684E+05, 3.9778E+05, 3.9873E+05, 3.9967E+05, 4.0062E+05,
     +  4.0157E+05, 4.0252E+05, 4.0348E+05, 4.0443E+05, 4.0539E+05,
     + 4.0635E+05, 4.0731E+05, 4.0827E+05, 4.0924E+05, 4.1021E+05,
     +  4.1117E+05, 4.1214E+05, 4.1312E+05, 4.1409E+05, 4.1507E+05,
     + 4.1604E+05, 4.1702E+05, 4.1800E+05, 4.1899E+05, 4.1997E+05,
     +  4.2096E+05, 4.2194E+05, 4.2293E+05, 4.2393E+05, 4.2492E+05,
     + 4.2591E+05, 4.2691E+05, 4.2791E+05, 4.2891E+05, 4.2991E+05,
     +  4.3092E+05, 4.3192E+05, 4.3293E+05, 4.3394E+05, 4.3495E+05,
     + 4.3597E+05, 4.3698E+05, 4.3800E+05, 4.3902E+05, 4.4004E+05,
     +  4.4106E+05, 4.4209E+05, 4.4311E+05, 4.4414E+05, 4.4517E+05,
     + 4.4620E+05, 4.4724E+05, 4.4827E+05, 4.4931E+05, 4.5035E+05,
     +  4.5139E+05, 4.5244E+05, 4.5348E+05, 4.5453E+05, 4.5558E+05,
     + 4.5663E+05, 4.5768E+05, 4.5874E+05, 4.5979E+05, 4.6085E+05,
     +  4.6191E+05, 4.6297E+05, 4.6404E+05, 4.6510E+05, 4.6617E+05,
     + 4.6724E+05, 4.6831E+05, 4.6939E+05, 4.7046E+05, 4.7154E+05,
     +  4.7262E+05, 4.7370E+05, 4.7478E+05, 4.7587E+05, 4.7695E+05,
     + 4.7804E+05, 4.7913E+05, 4.8023E+05, 4.8132E+05, 4.8242E+05,
     +  4.8352E+05, 4.8462E+05, 4.8572E+05, 4.8683E+05, 4.8793E+05,
     + 4.8904E+05, 4.9015E+05, 4.9126E+05, 4.9238E+05, 4.9349E+05,
     +  4.9461E+05, 4.9573E+05, 4.9685E+05, 4.9798E+05, 4.9910E+05,
     + 5.0023E+05, 5.0136E+05, 5.0249E+05, 5.0363E+05, 5.0476E+05,
     +  5.0590E+05, 5.0704E+05, 5.0818E+05, 5.0933E+05, 5.1047E+05,
     + 5.1162E+05, 5.1277E+05, 5.1392E+05, 5.1508E+05, 5.1623E+05,
     +  5.1739E+05, 5.1855E+05, 5.1971E+05, 5.2088E+05, 5.2204E+05,
     + 5.2321E+05, 5.2438E+05, 5.2555E+05, 5.2673E+05, 5.2790E+05,
     +  5.2908E+05, 5.3026E+05, 5.3144E+05, 5.3263E+05, 5.3381E+05,
     + 5.3500E+05, 5.3619E+05, 5.3738E+05, 5.3858E+05, 5.3977E+05,
     +  5.4097E+05, 5.4217E+05, 5.4338E+05, 5.4458E+05, 5.4579E+05,
     + 5.4700E+05, 5.4821E+05, 5.4942E+05, 5.5064E+05, 5.5185E+05,
     +  5.5307E+05, 5.5429E+05, 5.5552E+05, 5.5674E+05, 5.5797E+05,
     + 5.5920E+05, 5.6043E+05, 5.6167E+05, 5.6290E+05, 5.6414E+05,
     +  5.6538E+05, 5.6662E+05, 5.6787E+05, 5.6911E+05, 5.7036E+05,
     + 5.7161E+05, 5.7286E+05, 5.7412E+05, 5.7538E+05, 5.7664E+05,
     +  5.7790E+05, 5.7916E+05, 5.8043E+05, 5.8169E+05, 5.8296E+05,
     + 5.8423E+05, 5.8551E+05, 5.8678E+05, 5.8806E+05, 5.8934E+05,
     +  5.9063E+05, 5.9191E+05, 5.9320E+05, 5.9449E+05, 5.9578E+05,
     + 5.9707E+05, 5.9837E+05, 5.9966E+05, 6.0096E+05, 6.0227E+05,
     +  6.0357E+05, 6.0488E+05, 6.0619E+05, 6.0750E+05, 6.0881E+05,
     + 6.1012E+05, 6.1144E+05, 6.1276E+05, 6.1408E+05, 6.1541E+05,
     +  6.1673E+05, 6.1806E+05, 6.1939E+05, 6.2072E+05, 6.2206E+05,
     + 6.2340E+05, 6.2473E+05, 6.2608E+05, 6.2742E+05, 6.2877E+05,
     +  6.3011E+05, 6.3146E+05, 6.3282E+05, 6.3417E+05, 6.3553E+05,
     + 6.3689E+05, 6.3825E+05, 6.3961E+05, 6.4098E+05, 6.4235E+05,
     +  6.4372E+05, 6.4509E+05, 6.4646E+05, 6.4784E+05, 6.4922E+05,
     + 6.5060E+05, 6.5199E+05, 6.5337E+05, 6.5476E+05, 6.5615E+05,
     +  6.5754E+05, 6.5894E+05, 6.6034E+05, 6.6174E+05, 6.6314E+05,
     + 6.6454E+05, 6.6595E+05, 6.6736E+05, 6.6877E+05, 6.7018E+05,
     +  6.7160E+05, 6.7301E+05, 6.7443E+05, 6.7586E+05, 6.7728E+05,
     + 6.7871E+05, 6.8014E+05, 6.8157E+05, 6.8300E+05, 6.8444E+05,
     +  6.8588E+05, 6.8732E+05, 6.8876E+05, 6.9021E+05, 6.9166E+05,
     + 6.9311E+05, 6.9456E+05, 6.9601E+05, 6.9747E+05, 6.9893E+05,
     +  7.0039E+05, 7.0186E+05, 7.0332E+05, 7.0479E+05, 7.0626E+05,
     + 7.0774E+05, 7.0921E+05, 7.1069E+05, 7.1217E+05, 7.1365E+05,
     +  7.1514E+05, 7.1663E+05, 7.1812E+05, 7.1961E+05, 7.2110E+05,
     + 7.2260E+05, 7.2410E+05, 7.2560E+05, 7.2711E+05, 7.2861E+05,
     +  7.3012E+05, 7.3163E+05, 7.3315E+05, 7.3466E+05, 7.3618E+05,
     + 7.3770E+05, 7.3922E+05, 7.4075E+05, 7.4228E+05, 7.4381E+05,
     +  7.4534E+05, 7.4688E+05, 7.4841E+05, 7.4995E+05, 7.5150E+05,
     + 7.5304E+05, 7.5459E+05, 7.5614E+05, 7.5769E+05, 7.5924E+05,
     +  7.6080E+05, 7.6236E+05, 7.6392E+05, 7.6549E+05, 7.6705E+05,
     + 7.6862E+05, 7.7019E+05, 7.7177E+05, 7.7334E+05, 7.7492E+05,
     +  7.7650E+05, 7.7809E+05, 7.7967E+05, 7.8126E+05, 7.8285E+05,
     + 7.8445E+05, 7.8604E+05, 7.8764E+05, 7.8924E+05, 7.9085E+05,
     +  7.9245E+05, 7.9406E+05, 7.9567E+05, 7.9728E+05, 7.9890E+05,
     + 8.0052E+05, 8.0214E+05, 8.0376E+05, 8.0539E+05, 8.0702E+05,
     +  8.0865E+05, 8.1028E+05, 8.1192E+05, 8.1355E+05, 8.1520E+05,
     + 8.1684E+05, 8.1848E+05, 8.2013E+05, 8.2178E+05, 8.2344E+05,
     +  8.2509E+05, 8.2675E+05, 8.2841E+05, 8.3007E+05, 8.3174E+05,
     + 8.3341E+05, 8.3508E+05, 8.3675E+05, 8.3843E+05, 8.4011E+05,
     +  8.4179E+05, 8.4347E+05, 8.4516E+05, 8.4685E+05, 8.4854E+05,
     + 8.5023E+05, 8.5193E+05, 8.5363E+05, 8.5533E+05, 8.5703E+05,
     +  8.5874E+05, 8.6045E+05, 8.6216E+05, 8.6387E+05, 8.6559E+05,
     + 8.6731E+05, 8.6903E+05, 8.7076E+05, 8.7248E+05, 8.7421E+05,
     +  8.7595E+05, 8.7768E+05, 8.7942E+05, 8.8116E+05, 8.8290E+05,
     + 8.8465E+05, 8.8639E+05, 8.8814E+05, 8.8990E+05, 8.9165E+05,
     +  8.9341E+05, 8.9517E+05, 8.9693E+05, 8.9870E+05, 9.0047E+05,
     + 9.0224E+05, 9.0401E+05, 9.0579E+05, 9.0757E+05, 9.0935E+05,
     +  9.1114E+05, 9.1292E+05, 9.1471E+05, 9.1651E+05, 9.1830E+05,
     + 9.2010E+05, 9.2190E+05, 9.2370E+05, 9.2551E+05, 9.2732E+05,
     +  9.2913E+05, 9.3094E+05, 9.3276E+05, 9.3457E+05, 9.3640E+05,
     + 9.3822E+05, 9.4005E+05, 9.4188E+05, 9.4371E+05, 9.4554E+05,
     +  9.4738E+05, 9.4922E+05, 9.5106E+05, 9.5291E+05, 9.5476E+05,
     + 9.5661E+05, 9.5846E+05, 9.6032E+05, 9.6218E+05, 9.6404E+05,
     +  9.6590E+05, 9.6777E+05, 9.6964E+05, 9.7151E+05, 9.7339E+05,
     + 9.7527E+05, 9.7715E+05, 9.7903E+05, 9.8092E+05, 9.8280E+05,
     +  9.8470E+05, 9.8659E+05, 9.8849E+05, 9.9039E+05, 9.9229E+05,
     + 9.9419E+05, 9.9610E+05, 9.9801E+05, 9.9993E+05, 1.0018E+06,
     +  1.0038E+06, 1.0057E+06, 1.0076E+06, 1.0095E+06, 1.0115E+06,
     + 1.0134E+06, 1.0153E+06, 1.0173E+06, 1.0192E+06, 1.0212E+06,
     +  1.0231E+06, 1.0250E+06, 1.0270E+06, 1.0290E+06, 1.0309E+06,
     + 1.0329E+06, 1.0348E+06, 1.0368E+06, 1.0388E+06, 1.0407E+06,
     +  1.0427E+06, 1.0447E+06, 1.0467E+06, 1.0487E+06, 1.0506E+06,
     + 1.0526E+06, 1.0546E+06, 1.0566E+06, 1.0586E+06, 1.0606E+06,
     +  1.0626E+06, 1.0646E+06, 1.0666E+06, 1.0686E+06, 1.0706E+06,
     + 1.0727E+06, 1.0747E+06, 1.0767E+06, 1.0787E+06, 1.0808E+06,
     +  1.0828E+06, 1.0848E+06, 1.0869E+06, 1.0889E+06, 1.0909E+06,
     + 1.0930E+06, 1.0950E+06, 1.0971E+06, 1.0991E+06, 1.1012E+06,
     +  1.1032E+06, 1.1053E+06, 1.1074E+06, 1.1094E+06, 1.1115E+06,
     + 1.1136E+06, 1.1157E+06, 1.1177E+06, 1.1198E+06, 1.1219E+06,
     +  1.1240E+06, 1.1261E+06, 1.1282E+06, 1.1303E+06, 1.1324E+06,
     + 1.1345E+06, 1.1366E+06, 1.1387E+06, 1.1408E+06, 1.1429E+06,
     +  1.1450E+06, 1.1471E+06, 1.1493E+06, 1.1514E+06, 1.1535E+06,
     + 1.1556E+06, 1.1578E+06, 1.1599E+06, 1.1621E+06, 1.1642E+06,
     +  1.1663E+06, 1.1685E+06, 1.1706E+06, 1.1728E+06, 1.1750E+06,
     + 1.1771E+06, 1.1793E+06, 1.1814E+06, 1.1836E+06, 1.1858E+06,
     +  1.1880E+06, 1.1901E+06, 1.1923E+06, 1.1945E+06, 1.1967E+06,
     + 1.1989E+06, 1.2011E+06, 1.2033E+06, 1.2055E+06, 1.2077E+06,
     +  1.2099E+06, 1.2121E+06, 1.2143E+06, 1.2165E+06, 1.2187E+06,
     + 1.2209E+06, 1.2232E+06, 1.2254E+06, 1.2276E+06, 1.2298E+06,
     +  1.2321E+06, 1.2343E+06, 1.2365E+06, 1.2388E+06, 1.2410E+06,
     + 1.2433E+06, 1.2455E+06, 1.2478E+06, 1.2500E+06, 1.2523E+06,
     +  1.2546E+06, 1.2568E+06, 1.2591E+06, 1.2614E+06, 1.2637E+06,
     + 1.2659E+06, 1.2682E+06, 1.2705E+06, 1.2728E+06, 1.2751E+06,
     +  1.2774E+06, 1.2797E+06, 1.2820E+06, 1.2843E+06, 1.2866E+06,
     + 1.2889E+06, 1.2912E+06, 1.2935E+06, 1.2958E+06, 1.2982E+06,
     +  1.3005E+06, 1.3028E+06, 1.3051E+06, 1.3075E+06, 1.3098E+06,
     + 1.3122E+06, 1.3145E+06, 1.3168E+06, 1.3192E+06, 1.3215E+06,
     +  1.3239E+06, 1.3263E+06, 1.3286E+06, 1.3310E+06, 1.3333E+06,
     + 1.3357E+06, 1.3381E+06, 1.3405E+06, 1.3428E+06, 1.3452E+06,
     +  1.3476E+06, 1.3500E+06, 1.3524E+06, 1.3548E+06, 1.3572E+06,
     + 1.3596E+06, 1.3620E+06, 1.3644E+06, 1.3668E+06, 1.3692E+06,
     +  1.3716E+06, 1.3741E+06, 1.3765E+06, 1.3789E+06, 1.3813E+06,
     + 1.3838E+06, 1.3862E+06, 1.3886E+06, 1.3911E+06, 1.3935E+06,
     +  1.3960E+06, 1.3984E+06, 1.4009E+06, 1.4033E+06, 1.4058E+06,
     + 1.4083E+06, 1.4107E+06, 1.4132E+06, 1.4157E+06, 1.4182E+06,
     +  1.4206E+06, 1.4231E+06, 1.4256E+06, 1.4281E+06, 1.4306E+06,
     + 1.4331E+06, 1.4356E+06, 1.4381E+06, 1.4406E+06, 1.4431E+06,
     +  1.4456E+06, 1.4481E+06, 1.4506E+06, 1.4531E+06, 1.4557E+06,
     + 1.4582E+06, 1.4607E+06, 1.4633E+06, 1.4658E+06, 1.4683E+06,
     +  1.4709E+06, 1.4734E+06, 1.4760E+06, 1.4785E+06, 1.4811E+06,
     + 1.4836E+06, 1.4862E+06, 1.4888E+06, 1.4913E+06, 1.4939E+06,
     +  1.4965E+06, 1.4991E+06, 1.5016E+06, 1.5042E+06, 1.5068E+06,
     + 1.5094E+06, 1.5120E+06, 1.5146E+06, 1.5172E+06, 1.5198E+06,
     +  1.5224E+06, 1.5250E+06, 1.5276E+06, 1.5302E+06, 1.5329E+06,
     + 1.5355E+06, 1.5381E+06, 1.5407E+06, 1.5434E+06, 1.5460E+06,
     +  1.5487E+06, 1.5513E+06, 1.5539E+06, 1.5566E+06, 1.5592E+06,
     + 1.5619E+06, 1.5646E+06, 1.5672E+06, 1.5699E+06, 1.5726E+06,
     +  1.5752E+06, 1.5779E+06, 1.5806E+06, 1.5833E+06, 1.5859E+06,
     + 1.5886E+06, 1.5913E+06, 1.5940E+06, 1.5967E+06, 1.5994E+06,
     +  1.6021E+06, 1.6048E+06, 1.6075E+06, 1.6103E+06, 1.6130E+06/
      data (QofT( 2,J),J=2409,4510)/ 1.6157E+06, 1.6184E+06,
     +  1.6212E+06, 1.6239E+06, 1.6266E+06, 1.6294E+06, 1.6321E+06,
     +  1.6348E+06,
     + 1.6376E+06, 1.6403E+06, 1.6431E+06, 1.6459E+06, 1.6486E+06,
     +  1.6514E+06, 1.6541E+06, 1.6569E+06, 1.6597E+06, 1.6625E+06,
     + 1.6652E+06, 1.6680E+06, 1.6708E+06, 1.6736E+06, 1.6764E+06,
     +  1.6792E+06, 1.6820E+06, 1.6848E+06, 1.6876E+06, 1.6904E+06,
     + 1.6932E+06, 1.6961E+06, 1.6989E+06, 1.7017E+06, 1.7045E+06,
     +  1.7074E+06, 1.7102E+06, 1.7130E+06, 1.7159E+06, 1.7187E+06,
     + 1.7216E+06, 1.7244E+06, 1.7273E+06, 1.7301E+06, 1.7330E+06,
     +  1.7359E+06, 1.7387E+06, 1.7416E+06, 1.7445E+06, 1.7474E+06,
     + 1.7502E+06, 1.7531E+06, 1.7560E+06, 1.7589E+06, 1.7618E+06,
     +  1.7647E+06, 1.7676E+06, 1.7705E+06, 1.7734E+06, 1.7763E+06,
     + 1.7792E+06, 1.7822E+06, 1.7851E+06, 1.7880E+06, 1.7909E+06,
     +  1.7939E+06, 1.7968E+06, 1.7997E+06, 1.8027E+06, 1.8056E+06,
     + 1.8086E+06, 1.8115E+06, 1.8145E+06, 1.8175E+06, 1.8204E+06,
     +  1.8234E+06, 1.8264E+06, 1.8293E+06, 1.8323E+06, 1.8353E+06,
     + 1.8383E+06, 1.8413E+06, 1.8443E+06, 1.8472E+06, 1.8502E+06,
     +  1.8532E+06, 1.8563E+06, 1.8593E+06, 1.8623E+06, 1.8653E+06,
     + 1.8683E+06, 1.8713E+06, 1.8744E+06, 1.8774E+06, 1.8804E+06,
     +  1.8835E+06, 1.8865E+06, 1.8895E+06, 1.8926E+06, 1.8956E+06,
     + 1.8987E+06, 1.9017E+06, 1.9048E+06, 1.9079E+06, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       628
      data (QofT( 3,J),J=1,2408)/ 2.2606E+00, 4.1274E+00, 6.0087E+00,
     +  7.8934E+00, 9.7793E+00, 1.1666E+01, 1.3553E+01, 1.5440E+01,
     + 1.7327E+01, 1.9214E+01, 2.1102E+01, 2.2989E+01, 2.4877E+01,
     +  2.6765E+01, 2.8652E+01, 3.0540E+01, 3.2428E+01, 3.4315E+01,
     + 3.6203E+01, 3.8091E+01, 4.1866E+01, 4.5642E+01, 4.9417E+01,
     +  5.3193E+01, 5.6969E+01, 6.0744E+01, 6.4520E+01, 6.8296E+01,
     + 7.2071E+01, 7.5847E+01, 7.9623E+01, 8.3399E+01, 8.7175E+01,
     +  9.0951E+01, 9.4727E+01, 9.8502E+01, 1.0228E+02, 1.0605E+02,
     + 1.0983E+02, 1.1361E+02, 1.1738E+02, 1.2116E+02, 1.2493E+02,
     +  1.2871E+02, 1.3249E+02, 1.3626E+02, 1.4004E+02, 1.4382E+02,
     + 1.4759E+02, 1.5137E+02, 1.5515E+02, 1.5892E+02, 1.6270E+02,
     +  1.6648E+02, 1.7026E+02, 1.7404E+02, 1.7781E+02, 1.8159E+02,
     + 1.8538E+02, 1.8916E+02, 1.9294E+02, 1.9672E+02, 2.0051E+02,
     +  2.0430E+02, 2.0808E+02, 2.1187E+02, 2.1567E+02, 2.1946E+02,
     + 2.2326E+02, 2.2706E+02, 2.3086E+02, 2.3466E+02, 2.3847E+02,
     +  2.4228E+02, 2.4610E+02, 2.4992E+02, 2.5374E+02, 2.5757E+02,
     + 2.6141E+02, 2.6524E+02, 2.6909E+02, 2.7294E+02, 2.7680E+02,
     +  2.8066E+02, 2.8453E+02, 2.8841E+02, 2.9229E+02, 2.9619E+02,
     + 3.0009E+02, 3.0400E+02, 3.0791E+02, 3.1184E+02, 3.1578E+02,
     +  3.1973E+02, 3.2368E+02, 3.2765E+02, 3.3163E+02, 3.3562E+02,
     + 3.3962E+02, 3.4363E+02, 3.4766E+02, 3.5170E+02, 3.5575E+02,
     +  3.5981E+02, 3.6389E+02, 3.6799E+02, 3.7209E+02, 3.7621E+02,
     + 3.8035E+02, 3.8450E+02, 3.8867E+02, 3.9285E+02, 3.9705E+02,
     +  4.0127E+02, 4.0550E+02, 4.0975E+02, 4.1402E+02, 4.1831E+02,
     + 4.2261E+02, 4.2693E+02, 4.3128E+02, 4.3564E+02, 4.4002E+02,
     +  4.4442E+02, 4.4884E+02, 4.5328E+02, 4.5774E+02, 4.6222E+02,
     + 4.6673E+02, 4.7125E+02, 4.7580E+02, 4.8037E+02, 4.8496E+02,
     +  4.8957E+02, 4.9421E+02, 4.9887E+02, 5.0355E+02, 5.0826E+02,
     + 5.1299E+02, 5.1775E+02, 5.2253E+02, 5.2733E+02, 5.3216E+02,
     +  5.3702E+02, 5.4190E+02, 5.4681E+02, 5.5174E+02, 5.5670E+02,
     + 5.6169E+02, 5.6670E+02, 5.7174E+02, 5.7681E+02, 5.8191E+02,
     +  5.8703E+02, 5.9218E+02, 5.9736E+02, 6.0257E+02, 6.0781E+02,
     + 6.1307E+02, 6.1837E+02, 6.2370E+02, 6.2905E+02, 6.3444E+02,
     +  6.3986E+02, 6.4530E+02, 6.5078E+02, 6.5629E+02, 6.6183E+02,
     + 6.6740E+02, 6.7300E+02, 6.7864E+02, 6.8431E+02, 6.9001E+02,
     +  6.9574E+02, 7.0151E+02, 7.0731E+02, 7.1314E+02, 7.1900E+02,
     + 7.2490E+02, 7.3084E+02, 7.3681E+02, 7.4281E+02, 7.4885E+02,
     +  7.5492E+02, 7.6103E+02, 7.6717E+02, 7.7335E+02, 7.7957E+02,
     + 7.8582E+02, 7.9211E+02, 7.9843E+02, 8.0479E+02, 8.1119E+02,
     +  8.1762E+02, 8.2410E+02, 8.3061E+02, 8.3716E+02, 8.4374E+02,
     + 8.5037E+02, 8.5703E+02, 8.6374E+02, 8.7048E+02, 8.7726E+02,
     +  8.8408E+02, 8.9094E+02, 8.9784E+02, 9.0478E+02, 9.1176E+02,
     + 9.1878E+02, 9.2585E+02, 9.3295E+02, 9.4010E+02, 9.4728E+02,
     +  9.5451E+02, 9.6178E+02, 9.6910E+02, 9.7645E+02, 9.8385E+02,
     + 9.9129E+02, 9.9878E+02, 1.0063E+03, 1.0139E+03, 1.0215E+03,
     +  1.0292E+03, 1.0369E+03, 1.0446E+03, 1.0524E+03, 1.0602E+03,
     + 1.0681E+03, 1.0761E+03, 1.0840E+03, 1.0920E+03, 1.1001E+03,
     +  1.1082E+03, 1.1164E+03, 1.1246E+03, 1.1329E+03, 1.1412E+03,
     + 1.1495E+03, 1.1579E+03, 1.1663E+03, 1.1748E+03, 1.1834E+03,
     +  1.1920E+03, 1.2006E+03, 1.2093E+03, 1.2180E+03, 1.2268E+03,
     + 1.2357E+03, 1.2446E+03, 1.2535E+03, 1.2625E+03, 1.2715E+03,
     +  1.2806E+03, 1.2898E+03, 1.2990E+03, 1.3082E+03, 1.3175E+03,
     + 1.3269E+03, 1.3363E+03, 1.3457E+03, 1.3553E+03, 1.3648E+03,
     +  1.3745E+03, 1.3841E+03, 1.3939E+03, 1.4036E+03, 1.4135E+03,
     + 1.4234E+03, 1.4333E+03, 1.4433E+03, 1.4534E+03, 1.4635E+03,
     +  1.4737E+03, 1.4839E+03, 1.4942E+03, 1.5045E+03, 1.5149E+03,
     + 1.5254E+03, 1.5359E+03, 1.5465E+03, 1.5571E+03, 1.5678E+03,
     +  1.5786E+03, 1.5894E+03, 1.6002E+03, 1.6112E+03, 1.6222E+03,
     + 1.6332E+03, 1.6443E+03, 1.6555E+03, 1.6667E+03, 1.6780E+03,
     +  1.6894E+03, 1.7008E+03, 1.7123E+03, 1.7238E+03, 1.7354E+03,
     + 1.7471E+03, 1.7588E+03, 1.7706E+03, 1.7825E+03, 1.7944E+03,
     +  1.8064E+03, 1.8184E+03, 1.8306E+03, 1.8427E+03, 1.8550E+03,
     + 1.8673E+03, 1.8797E+03, 1.8921E+03, 1.9046E+03, 1.9172E+03,
     +  1.9299E+03, 1.9426E+03, 1.9554E+03, 1.9682E+03, 1.9811E+03,
     + 1.9941E+03, 2.0072E+03, 2.0203E+03, 2.0335E+03, 2.0468E+03,
     +  2.0601E+03, 2.0735E+03, 2.0870E+03, 2.1006E+03, 2.1142E+03,
     + 2.1279E+03, 2.1417E+03, 2.1555E+03, 2.1694E+03, 2.1834E+03,
     +  2.1974E+03, 2.2116E+03, 2.2258E+03, 2.2401E+03, 2.2544E+03,
     + 2.2689E+03, 2.2834E+03, 2.2979E+03, 2.3126E+03, 2.3273E+03,
     +  2.3421E+03, 2.3570E+03, 2.3720E+03, 2.3870E+03, 2.4021E+03,
     + 2.4173E+03, 2.4326E+03, 2.4480E+03, 2.4634E+03, 2.4789E+03,
     +  2.4945E+03, 2.5102E+03, 2.5259E+03, 2.5418E+03, 2.5577E+03,
     + 2.5737E+03, 2.5897E+03, 2.6059E+03, 2.6221E+03, 2.6385E+03,
     +  2.6549E+03, 2.6714E+03, 2.6879E+03, 2.7046E+03, 2.7213E+03,
     + 2.7381E+03, 2.7551E+03, 2.7721E+03, 2.7891E+03, 2.8063E+03,
     +  2.8236E+03, 2.8409E+03, 2.8583E+03, 2.8758E+03, 2.8934E+03,
     + 2.9111E+03, 2.9289E+03, 2.9468E+03, 2.9647E+03, 2.9828E+03,
     +  3.0009E+03, 3.0192E+03, 3.0375E+03, 3.0559E+03, 3.0744E+03,
     + 3.0930E+03, 3.1117E+03, 3.1304E+03, 3.1493E+03, 3.1683E+03,
     +  3.1873E+03, 3.2065E+03, 3.2257E+03, 3.2450E+03, 3.2645E+03,
     + 3.2840E+03, 3.3036E+03, 3.3233E+03, 3.3432E+03, 3.3631E+03,
     +  3.3831E+03, 3.4032E+03, 3.4234E+03, 3.4437E+03, 3.4641E+03,
     + 3.4846E+03, 3.5052E+03, 3.5259E+03, 3.5467E+03, 3.5676E+03,
     +  3.5886E+03, 3.6097E+03, 3.6309E+03, 3.6522E+03, 3.6736E+03,
     + 3.6951E+03, 3.7167E+03, 3.7385E+03, 3.7603E+03, 3.7822E+03,
     +  3.8042E+03, 3.8264E+03, 3.8486E+03, 3.8710E+03, 3.8934E+03,
     + 3.9160E+03, 3.9387E+03, 3.9614E+03, 3.9843E+03, 4.0073E+03,
     +  4.0304E+03, 4.0536E+03, 4.0769E+03, 4.1004E+03, 4.1239E+03,
     + 4.1475E+03, 4.1713E+03, 4.1952E+03, 4.2192E+03, 4.2433E+03,
     +  4.2675E+03, 4.2918E+03, 4.3162E+03, 4.3408E+03, 4.3655E+03,
     + 4.3902E+03, 4.4151E+03, 4.4401E+03, 4.4653E+03, 4.4905E+03,
     +  4.5159E+03, 4.5414E+03, 4.5669E+03, 4.5927E+03, 4.6185E+03,
     + 4.6444E+03, 4.6705E+03, 4.6967E+03, 4.7230E+03, 4.7495E+03,
     +  4.7760E+03, 4.8027E+03, 4.8295E+03, 4.8564E+03, 4.8835E+03,
     + 4.9106E+03, 4.9379E+03, 4.9653E+03, 4.9929E+03, 5.0205E+03,
     +  5.0483E+03, 5.0762E+03, 5.1043E+03, 5.1325E+03, 5.1608E+03,
     + 5.1892E+03, 5.2177E+03, 5.2464E+03, 5.2752E+03, 5.3042E+03,
     +  5.3333E+03, 5.3625E+03, 5.3918E+03, 5.4213E+03, 5.4509E+03,
     + 5.4806E+03, 5.5105E+03, 5.5405E+03, 5.5706E+03, 5.6009E+03,
     +  5.6313E+03, 5.6618E+03, 5.6925E+03, 5.7233E+03, 5.7542E+03,
     + 5.7853E+03, 5.8165E+03, 5.8479E+03, 5.8794E+03, 5.9110E+03,
     +  5.9428E+03, 5.9747E+03, 6.0067E+03, 6.0389E+03, 6.0713E+03,
     + 6.1038E+03, 6.1364E+03, 6.1691E+03, 6.2021E+03, 6.2351E+03,
     +  6.2683E+03, 6.3016E+03, 6.3351E+03, 6.3688E+03, 6.4025E+03,
     + 6.4365E+03, 6.4705E+03, 6.5048E+03, 6.5391E+03, 6.5737E+03,
     +  6.6083E+03, 6.6431E+03, 6.6781E+03, 6.7132E+03, 6.7485E+03,
     + 6.7839E+03, 6.8195E+03, 6.8552E+03, 6.8911E+03, 6.9271E+03,
     +  6.9633E+03, 6.9997E+03, 7.0362E+03, 7.0728E+03, 7.1096E+03,
     + 7.1466E+03, 7.1837E+03, 7.2210E+03, 7.2585E+03, 7.2961E+03,
     +  7.3338E+03, 7.3718E+03, 7.4098E+03, 7.4481E+03, 7.4865E+03,
     + 7.5251E+03, 7.5638E+03, 7.6027E+03, 7.6417E+03, 7.6810E+03,
     +  7.7203E+03, 7.7599E+03, 7.7996E+03, 7.8395E+03, 7.8796E+03,
     + 7.9198E+03, 7.9602E+03, 8.0007E+03, 8.0414E+03, 8.0823E+03,
     +  8.1234E+03, 8.1646E+03, 8.2061E+03, 8.2476E+03, 8.2894E+03,
     + 8.3313E+03, 8.3734E+03, 8.4157E+03, 8.4581E+03, 8.5008E+03,
     +  8.5436E+03, 8.5865E+03, 8.6297E+03, 8.6730E+03, 8.7165E+03,
     + 8.7602E+03, 8.8041E+03, 8.8481E+03, 8.8924E+03, 8.9368E+03,
     +  8.9814E+03, 9.0261E+03, 9.0711E+03, 9.1162E+03, 9.1615E+03,
     + 9.2070E+03, 9.2527E+03, 9.2986E+03, 9.3447E+03, 9.3909E+03,
     +  9.4374E+03, 9.4840E+03, 9.5308E+03, 9.5778E+03, 9.6250E+03,
     + 9.6724E+03, 9.7200E+03, 9.7677E+03, 9.8157E+03, 9.8638E+03,
     +  9.9122E+03, 9.9607E+03, 1.0009E+04, 1.0058E+04, 1.0107E+04,
     + 1.0157E+04, 1.0206E+04, 1.0256E+04, 1.0306E+04, 1.0356E+04,
     +  1.0406E+04, 1.0457E+04, 1.0508E+04, 1.0559E+04, 1.0610E+04,
     + 1.0661E+04, 1.0712E+04, 1.0764E+04, 1.0816E+04, 1.0868E+04,
     +  1.0921E+04, 1.0973E+04, 1.1026E+04, 1.1079E+04, 1.1132E+04,
     + 1.1185E+04, 1.1239E+04, 1.1293E+04, 1.1347E+04, 1.1401E+04,
     +  1.1455E+04, 1.1510E+04, 1.1565E+04, 1.1620E+04, 1.1675E+04,
     + 1.1731E+04, 1.1786E+04, 1.1842E+04, 1.1898E+04, 1.1955E+04,
     +  1.2011E+04, 1.2068E+04, 1.2125E+04, 1.2182E+04, 1.2240E+04,
     + 1.2298E+04, 1.2356E+04, 1.2414E+04, 1.2472E+04, 1.2531E+04,
     +  1.2589E+04, 1.2648E+04, 1.2708E+04, 1.2767E+04, 1.2827E+04,
     + 1.2887E+04, 1.2947E+04, 1.3007E+04, 1.3068E+04, 1.3129E+04,
     +  1.3190E+04, 1.3251E+04, 1.3313E+04, 1.3374E+04, 1.3436E+04,
     + 1.3499E+04, 1.3561E+04, 1.3624E+04, 1.3687E+04, 1.3750E+04,
     +  1.3814E+04, 1.3877E+04, 1.3941E+04, 1.4005E+04, 1.4070E+04,
     + 1.4134E+04, 1.4199E+04, 1.4264E+04, 1.4330E+04, 1.4395E+04,
     +  1.4461E+04, 1.4527E+04, 1.4594E+04, 1.4660E+04, 1.4727E+04,
     + 1.4794E+04, 1.4861E+04, 1.4929E+04, 1.4997E+04, 1.5065E+04,
     +  1.5133E+04, 1.5202E+04, 1.5271E+04, 1.5340E+04, 1.5409E+04,
     + 1.5479E+04, 1.5549E+04, 1.5619E+04, 1.5689E+04, 1.5760E+04,
     +  1.5831E+04, 1.5902E+04, 1.5973E+04, 1.6045E+04, 1.6117E+04,
     + 1.6189E+04, 1.6262E+04, 1.6334E+04, 1.6407E+04, 1.6481E+04,
     +  1.6554E+04, 1.6628E+04, 1.6702E+04, 1.6777E+04, 1.6851E+04,
     + 1.6926E+04, 1.7001E+04, 1.7077E+04, 1.7152E+04, 1.7228E+04,
     +  1.7305E+04, 1.7381E+04, 1.7458E+04, 1.7535E+04, 1.7612E+04,
     + 1.7690E+04, 1.7768E+04, 1.7846E+04, 1.7925E+04, 1.8003E+04,
     +  1.8082E+04, 1.8162E+04, 1.8241E+04, 1.8321E+04, 1.8401E+04,
     + 1.8482E+04, 1.8563E+04, 1.8644E+04, 1.8725E+04, 1.8807E+04,
     +  1.8888E+04, 1.8971E+04, 1.9053E+04, 1.9136E+04, 1.9219E+04,
     + 1.9302E+04, 1.9386E+04, 1.9470E+04, 1.9554E+04, 1.9639E+04,
     +  1.9724E+04, 1.9809E+04, 1.9894E+04, 1.9980E+04, 2.0066E+04,
     + 2.0152E+04, 2.0239E+04, 2.0326E+04, 2.0413E+04, 2.0501E+04,
     +  2.0589E+04, 2.0677E+04, 2.0765E+04, 2.0854E+04, 2.0943E+04,
     + 2.1032E+04, 2.1122E+04, 2.1212E+04, 2.1303E+04, 2.1393E+04,
     +  2.1484E+04, 2.1575E+04, 2.1667E+04, 2.1759E+04, 2.1851E+04,
     + 2.1944E+04, 2.2037E+04, 2.2130E+04, 2.2223E+04, 2.2317E+04,
     +  2.2411E+04, 2.2506E+04, 2.2601E+04, 2.2696E+04, 2.2791E+04,
     + 2.2887E+04, 2.2983E+04, 2.3079E+04, 2.3176E+04, 2.3273E+04,
     +  2.3371E+04, 2.3468E+04, 2.3567E+04, 2.3665E+04, 2.3764E+04,
     + 2.3863E+04, 2.3962E+04, 2.4062E+04, 2.4162E+04, 2.4262E+04,
     +  2.4363E+04, 2.4464E+04, 2.4566E+04, 2.4668E+04, 2.4770E+04,
     + 2.4872E+04, 2.4975E+04, 2.5078E+04, 2.5182E+04, 2.5286E+04,
     +  2.5390E+04, 2.5494E+04, 2.5599E+04, 2.5705E+04, 2.5810E+04,
     + 2.5916E+04, 2.6022E+04, 2.6129E+04, 2.6236E+04, 2.6344E+04,
     +  2.6451E+04, 2.6559E+04, 2.6668E+04, 2.6777E+04, 2.6886E+04,
     + 2.6995E+04, 2.7105E+04, 2.7216E+04, 2.7326E+04, 2.7437E+04,
     +  2.7548E+04, 2.7660E+04, 2.7772E+04, 2.7885E+04, 2.7998E+04,
     + 2.8111E+04, 2.8224E+04, 2.8338E+04, 2.8453E+04, 2.8567E+04,
     +  2.8682E+04, 2.8798E+04, 2.8914E+04, 2.9030E+04, 2.9146E+04,
     + 2.9263E+04, 2.9381E+04, 2.9498E+04, 2.9617E+04, 2.9735E+04,
     +  2.9854E+04, 2.9973E+04, 3.0093E+04, 3.0213E+04, 3.0333E+04,
     + 3.0454E+04, 3.0575E+04, 3.0697E+04, 3.0819E+04, 3.0941E+04,
     +  3.1064E+04, 3.1187E+04, 3.1311E+04, 3.1435E+04, 3.1559E+04,
     + 3.1684E+04, 3.1809E+04, 3.1934E+04, 3.2060E+04, 3.2187E+04,
     +  3.2313E+04, 3.2441E+04, 3.2568E+04, 3.2696E+04, 3.2824E+04,
     + 3.2953E+04, 3.3082E+04, 3.3212E+04, 3.3342E+04, 3.3472E+04,
     +  3.3603E+04, 3.3735E+04, 3.3866E+04, 3.3998E+04, 3.4131E+04,
     + 3.4264E+04, 3.4397E+04, 3.4531E+04, 3.4665E+04, 3.4800E+04,
     +  3.4935E+04, 3.5070E+04, 3.5206E+04, 3.5342E+04, 3.5479E+04,
     + 3.5616E+04, 3.5754E+04, 3.5892E+04, 3.6030E+04, 3.6169E+04,
     +  3.6308E+04, 3.6448E+04, 3.6588E+04, 3.6729E+04, 3.6870E+04,
     + 3.7011E+04, 3.7153E+04, 3.7296E+04, 3.7438E+04, 3.7582E+04,
     +  3.7725E+04, 3.7869E+04, 3.8014E+04, 3.8159E+04, 3.8304E+04,
     + 3.8450E+04, 3.8597E+04, 3.8744E+04, 3.8891E+04, 3.9038E+04,
     +  3.9187E+04, 3.9335E+04, 3.9484E+04, 3.9634E+04, 3.9784E+04,
     + 3.9934E+04, 4.0085E+04, 4.0237E+04, 4.0388E+04, 4.0541E+04,
     +  4.0693E+04, 4.0847E+04, 4.1000E+04, 4.1154E+04, 4.1309E+04,
     + 4.1464E+04, 4.1620E+04, 4.1776E+04, 4.1932E+04, 4.2089E+04,
     +  4.2247E+04, 4.2405E+04, 4.2563E+04, 4.2722E+04, 4.2881E+04,
     + 4.3041E+04, 4.3201E+04, 4.3362E+04, 4.3523E+04, 4.3685E+04,
     +  4.3847E+04, 4.4010E+04, 4.4173E+04, 4.4337E+04, 4.4501E+04,
     + 4.4666E+04, 4.4831E+04, 4.4997E+04, 4.5163E+04, 4.5329E+04,
     +  4.5497E+04, 4.5664E+04, 4.5832E+04, 4.6001E+04, 4.6170E+04,
     + 4.6340E+04, 4.6510E+04, 4.6681E+04, 4.6852E+04, 4.7023E+04,
     +  4.7195E+04, 4.7368E+04, 4.7541E+04, 4.7715E+04, 4.7889E+04,
     + 4.8064E+04, 4.8239E+04, 4.8415E+04, 4.8591E+04, 4.8768E+04,
     +  4.8945E+04, 4.9123E+04, 4.9301E+04, 4.9480E+04, 4.9659E+04,
     + 4.9839E+04, 5.0019E+04, 5.0200E+04, 5.0382E+04, 5.0564E+04,
     +  5.0746E+04, 5.0929E+04, 5.1113E+04, 5.1297E+04, 5.1482E+04,
     + 5.1667E+04, 5.1852E+04, 5.2039E+04, 5.2225E+04, 5.2413E+04,
     +  5.2600E+04, 5.2789E+04, 5.2978E+04, 5.3167E+04, 5.3357E+04,
     + 5.3548E+04, 5.3739E+04, 5.3931E+04, 5.4123E+04, 5.4315E+04,
     +  5.4509E+04, 5.4703E+04, 5.4897E+04, 5.5092E+04, 5.5287E+04,
     + 5.5483E+04, 5.5680E+04, 5.5877E+04, 5.6075E+04, 5.6273E+04,
     +  5.6472E+04, 5.6671E+04, 5.6871E+04, 5.7072E+04, 5.7273E+04,
     + 5.7475E+04, 5.7677E+04, 5.7880E+04, 5.8083E+04, 5.8287E+04,
     +  5.8492E+04, 5.8697E+04, 5.8902E+04, 5.9109E+04, 5.9315E+04,
     + 5.9523E+04, 5.9731E+04, 5.9939E+04, 6.0149E+04, 6.0358E+04,
     +  6.0569E+04, 6.0780E+04, 6.0991E+04, 6.1203E+04, 6.1416E+04,
     + 6.1629E+04, 6.1843E+04, 6.2057E+04, 6.2272E+04, 6.2488E+04,
     +  6.2704E+04, 6.2921E+04, 6.3138E+04, 6.3356E+04, 6.3575E+04,
     + 6.3794E+04, 6.4014E+04, 6.4235E+04, 6.4456E+04, 6.4677E+04,
     +  6.4900E+04, 6.5122E+04, 6.5346E+04, 6.5570E+04, 6.5795E+04,
     + 6.6020E+04, 6.6246E+04, 6.6472E+04, 6.6700E+04, 6.6927E+04,
     +  6.7156E+04, 6.7385E+04, 6.7614E+04, 6.7845E+04, 6.8076E+04,
     + 6.8307E+04, 6.8539E+04, 6.8772E+04, 6.9005E+04, 6.9239E+04,
     +  6.9474E+04, 6.9709E+04, 6.9945E+04, 7.0182E+04, 7.0419E+04,
     + 7.0657E+04, 7.0895E+04, 7.1134E+04, 7.1374E+04, 7.1615E+04,
     +  7.1856E+04, 7.2097E+04, 7.2340E+04, 7.2583E+04, 7.2826E+04,
     + 7.3071E+04, 7.3315E+04, 7.3561E+04, 7.3807E+04, 7.4054E+04,
     +  7.4302E+04, 7.4550E+04, 7.4799E+04, 7.5048E+04, 7.5298E+04,
     + 7.5549E+04, 7.5801E+04, 7.6053E+04, 7.6306E+04, 7.6559E+04,
     +  7.6813E+04, 7.7068E+04, 7.7324E+04, 7.7580E+04, 7.7837E+04,
     + 7.8094E+04, 7.8353E+04, 7.8611E+04, 7.8871E+04, 7.9131E+04,
     +  7.9392E+04, 7.9654E+04, 7.9916E+04, 8.0179E+04, 8.0443E+04,
     + 8.0707E+04, 8.0972E+04, 8.1238E+04, 8.1504E+04, 8.1771E+04,
     +  8.2039E+04, 8.2307E+04, 8.2576E+04, 8.2846E+04, 8.3117E+04,
     + 8.3388E+04, 8.3660E+04, 8.3933E+04, 8.4206E+04, 8.4480E+04,
     +  8.4755E+04, 8.5030E+04, 8.5307E+04, 8.5583E+04, 8.5861E+04,
     + 8.6139E+04, 8.6418E+04, 8.6698E+04, 8.6979E+04, 8.7260E+04,
     +  8.7542E+04, 8.7824E+04, 8.8107E+04, 8.8391E+04, 8.8676E+04,
     + 8.8962E+04, 8.9248E+04, 8.9535E+04, 8.9822E+04, 9.0111E+04,
     +  9.0400E+04, 9.0690E+04, 9.0980E+04, 9.1272E+04, 9.1564E+04,
     + 9.1857E+04, 9.2150E+04, 9.2444E+04, 9.2739E+04, 9.3035E+04,
     +  9.3331E+04, 9.3629E+04, 9.3927E+04, 9.4225E+04, 9.4525E+04,
     + 9.4825E+04, 9.5126E+04, 9.5428E+04, 9.5730E+04, 9.6033E+04,
     +  9.6337E+04, 9.6642E+04, 9.6947E+04, 9.7254E+04, 9.7561E+04,
     + 9.7868E+04, 9.8177E+04, 9.8486E+04, 9.8796E+04, 9.9107E+04,
     +  9.9419E+04, 9.9731E+04, 1.0004E+05, 1.0036E+05, 1.0067E+05,
     + 1.0099E+05, 1.0130E+05, 1.0162E+05, 1.0194E+05, 1.0226E+05,
     +  1.0258E+05, 1.0290E+05, 1.0322E+05, 1.0354E+05, 1.0386E+05,
     + 1.0418E+05, 1.0451E+05, 1.0483E+05, 1.0516E+05, 1.0549E+05,
     +  1.0581E+05, 1.0614E+05, 1.0647E+05, 1.0680E+05, 1.0713E+05,
     + 1.0746E+05, 1.0779E+05, 1.0813E+05, 1.0846E+05, 1.0879E+05,
     +  1.0913E+05, 1.0946E+05, 1.0980E+05, 1.1014E+05, 1.1048E+05,
     + 1.1082E+05, 1.1116E+05, 1.1150E+05, 1.1184E+05, 1.1218E+05,
     +  1.1252E+05, 1.1287E+05, 1.1321E+05, 1.1356E+05, 1.1391E+05,
     + 1.1425E+05, 1.1460E+05, 1.1495E+05, 1.1530E+05, 1.1565E+05,
     +  1.1600E+05, 1.1635E+05, 1.1671E+05, 1.1706E+05, 1.1742E+05,
     + 1.1777E+05, 1.1813E+05, 1.1849E+05, 1.1884E+05, 1.1920E+05,
     +  1.1956E+05, 1.1992E+05, 1.2028E+05, 1.2065E+05, 1.2101E+05,
     + 1.2137E+05, 1.2174E+05, 1.2211E+05, 1.2247E+05, 1.2284E+05,
     +  1.2321E+05, 1.2358E+05, 1.2395E+05, 1.2432E+05, 1.2469E+05,
     + 1.2506E+05, 1.2544E+05, 1.2581E+05, 1.2619E+05, 1.2656E+05,
     +  1.2694E+05, 1.2732E+05, 1.2769E+05, 1.2807E+05, 1.2845E+05,
     + 1.2884E+05, 1.2922E+05, 1.2960E+05, 1.2999E+05, 1.3037E+05,
     +  1.3076E+05, 1.3114E+05, 1.3153E+05, 1.3192E+05, 1.3231E+05,
     + 1.3270E+05, 1.3309E+05, 1.3348E+05, 1.3387E+05, 1.3427E+05,
     +  1.3466E+05, 1.3506E+05, 1.3545E+05, 1.3585E+05, 1.3625E+05,
     + 1.3665E+05, 1.3705E+05, 1.3745E+05, 1.3785E+05, 1.3825E+05,
     +  1.3866E+05, 1.3906E+05, 1.3947E+05, 1.3987E+05, 1.4028E+05,
     + 1.4069E+05, 1.4110E+05, 1.4151E+05, 1.4192E+05, 1.4233E+05,
     +  1.4274E+05, 1.4316E+05, 1.4357E+05, 1.4398E+05, 1.4440E+05,
     + 1.4482E+05, 1.4524E+05, 1.4566E+05, 1.4608E+05, 1.4650E+05,
     +  1.4692E+05, 1.4734E+05, 1.4777E+05, 1.4819E+05, 1.4862E+05,
     + 1.4904E+05, 1.4947E+05, 1.4990E+05, 1.5033E+05, 1.5076E+05,
     +  1.5119E+05, 1.5162E+05, 1.5205E+05, 1.5249E+05, 1.5292E+05,
     + 1.5336E+05, 1.5380E+05, 1.5423E+05, 1.5467E+05, 1.5511E+05,
     +  1.5555E+05, 1.5600E+05, 1.5644E+05, 1.5688E+05, 1.5733E+05,
     + 1.5777E+05, 1.5822E+05, 1.5867E+05, 1.5911E+05, 1.5956E+05,
     +  1.6001E+05, 1.6047E+05, 1.6092E+05, 1.6137E+05, 1.6182E+05,
     + 1.6228E+05, 1.6274E+05, 1.6319E+05, 1.6365E+05, 1.6411E+05,
     +  1.6457E+05, 1.6503E+05, 1.6549E+05, 1.6596E+05, 1.6642E+05,
     + 1.6689E+05, 1.6735E+05, 1.6782E+05, 1.6829E+05, 1.6876E+05,
     +  1.6923E+05, 1.6970E+05, 1.7017E+05, 1.7064E+05, 1.7112E+05,
     + 1.7159E+05, 1.7207E+05, 1.7254E+05, 1.7302E+05, 1.7350E+05,
     +  1.7398E+05, 1.7446E+05, 1.7494E+05, 1.7543E+05, 1.7591E+05,
     + 1.7639E+05, 1.7688E+05, 1.7737E+05, 1.7786E+05, 1.7834E+05,
     +  1.7883E+05, 1.7933E+05, 1.7982E+05, 1.8031E+05, 1.8080E+05,
     + 1.8130E+05, 1.8180E+05, 1.8229E+05, 1.8279E+05, 1.8329E+05,
     +  1.8379E+05, 1.8429E+05, 1.8479E+05, 1.8530E+05, 1.8580E+05,
     + 1.8631E+05, 1.8681E+05, 1.8732E+05, 1.8783E+05, 1.8834E+05,
     +  1.8885E+05, 1.8936E+05, 1.8987E+05, 1.9039E+05, 1.9090E+05,
     + 1.9142E+05, 1.9193E+05, 1.9245E+05, 1.9297E+05, 1.9349E+05,
     +  1.9401E+05, 1.9453E+05, 1.9506E+05, 1.9558E+05, 1.9611E+05,
     + 1.9663E+05, 1.9716E+05, 1.9769E+05, 1.9822E+05, 1.9875E+05,
     +  1.9928E+05, 1.9981E+05, 2.0035E+05, 2.0088E+05, 2.0142E+05,
     + 2.0195E+05, 2.0249E+05, 2.0303E+05, 2.0357E+05, 2.0411E+05,
     +  2.0466E+05, 2.0520E+05, 2.0574E+05, 2.0629E+05, 2.0684E+05,
     + 2.0738E+05, 2.0793E+05, 2.0848E+05, 2.0903E+05, 2.0958E+05,
     +  2.1014E+05, 2.1069E+05, 2.1125E+05, 2.1180E+05, 2.1236E+05,
     + 2.1292E+05, 2.1348E+05, 2.1404E+05, 2.1460E+05, 2.1516E+05,
     +  2.1573E+05, 2.1629E+05, 2.1686E+05, 2.1743E+05, 2.1800E+05,
     + 2.1857E+05, 2.1914E+05, 2.1971E+05, 2.2028E+05, 2.2085E+05,
     +  2.2143E+05, 2.2201E+05, 2.2258E+05, 2.2316E+05, 2.2374E+05,
     + 2.2432E+05, 2.2490E+05, 2.2549E+05, 2.2607E+05, 2.2665E+05,
     +  2.2724E+05, 2.2783E+05, 2.2842E+05, 2.2901E+05, 2.2960E+05,
     + 2.3019E+05, 2.3078E+05, 2.3138E+05, 2.3197E+05, 2.3257E+05,
     +  2.3316E+05, 2.3376E+05, 2.3436E+05, 2.3496E+05, 2.3557E+05,
     + 2.3617E+05, 2.3677E+05, 2.3738E+05, 2.3798E+05, 2.3859E+05,
     +  2.3920E+05, 2.3981E+05, 2.4042E+05, 2.4103E+05, 2.4165E+05,
     + 2.4226E+05, 2.4288E+05, 2.4350E+05, 2.4411E+05, 2.4473E+05,
     +  2.4535E+05, 2.4597E+05, 2.4660E+05, 2.4722E+05, 2.4785E+05,
     + 2.4847E+05, 2.4910E+05, 2.4973E+05, 2.5036E+05, 2.5099E+05,
     +  2.5162E+05, 2.5225E+05, 2.5289E+05, 2.5352E+05, 2.5416E+05,
     + 2.5480E+05, 2.5544E+05, 2.5608E+05, 2.5672E+05, 2.5736E+05,
     +  2.5800E+05, 2.5865E+05, 2.5929E+05, 2.5994E+05, 2.6059E+05,
     + 2.6124E+05, 2.6189E+05, 2.6254E+05, 2.6319E+05, 2.6385E+05,
     +  2.6450E+05, 2.6516E+05, 2.6582E+05, 2.6648E+05, 2.6714E+05,
     + 2.6780E+05, 2.6846E+05, 2.6912E+05, 2.6979E+05, 2.7046E+05,
     +  2.7112E+05, 2.7179E+05, 2.7246E+05, 2.7313E+05, 2.7380E+05,
     + 2.7448E+05, 2.7515E+05, 2.7583E+05, 2.7650E+05, 2.7718E+05,
     +  2.7786E+05, 2.7854E+05, 2.7922E+05, 2.7991E+05, 2.8059E+05,
     + 2.8128E+05, 2.8196E+05, 2.8265E+05, 2.8334E+05, 2.8403E+05,
     +  2.8472E+05, 2.8541E+05, 2.8611E+05, 2.8680E+05, 2.8750E+05,
     + 2.8820E+05, 2.8889E+05, 2.8959E+05, 2.9030E+05, 2.9100E+05,
     +  2.9170E+05, 2.9241E+05, 2.9311E+05, 2.9382E+05, 2.9453E+05,
     + 2.9524E+05, 2.9595E+05, 2.9666E+05, 2.9738E+05, 2.9809E+05,
     +  2.9881E+05, 2.9952E+05, 3.0024E+05, 3.0096E+05, 3.0168E+05,
     + 3.0240E+05, 3.0313E+05, 3.0385E+05, 3.0458E+05, 3.0531E+05,
     +  3.0603E+05, 3.0676E+05, 3.0749E+05, 3.0823E+05, 3.0896E+05,
     + 3.0969E+05, 3.1043E+05, 3.1117E+05, 3.1191E+05, 3.1264E+05,
     +  3.1339E+05, 3.1413E+05, 3.1487E+05, 3.1562E+05, 3.1636E+05,
     + 3.1711E+05, 3.1786E+05, 3.1861E+05, 3.1936E+05, 3.2011E+05,
     +  3.2086E+05, 3.2162E+05, 3.2237E+05, 3.2313E+05, 3.2389E+05,
     + 3.2465E+05, 3.2541E+05, 3.2617E+05, 3.2694E+05, 3.2770E+05,
     +  3.2847E+05, 3.2924E+05, 3.3000E+05, 3.3077E+05, 3.3155E+05,
     + 3.3232E+05, 3.3309E+05, 3.3387E+05, 3.3464E+05, 3.3542E+05,
     +  3.3620E+05, 3.3698E+05, 3.3776E+05, 3.3854E+05, 3.3933E+05,
     + 3.4011E+05, 3.4090E+05, 3.4169E+05, 3.4248E+05, 3.4327E+05,
     +  3.4406E+05, 3.4485E+05, 3.4565E+05, 3.4644E+05, 3.4724E+05,
     + 3.4804E+05, 3.4884E+05, 3.4964E+05, 3.5044E+05, 3.5124E+05,
     +  3.5205E+05, 3.5286E+05, 3.5366E+05, 3.5447E+05, 3.5528E+05,
     + 3.5609E+05, 3.5691E+05, 3.5772E+05, 3.5853E+05, 3.5935E+05,
     +  3.6017E+05, 3.6099E+05, 3.6181E+05, 3.6263E+05, 3.6345E+05,
     + 3.6428E+05, 3.6510E+05, 3.6593E+05, 3.6676E+05, 3.6759E+05,
     +  3.6842E+05, 3.6925E+05, 3.7009E+05, 3.7092E+05, 3.7176E+05,
     + 3.7259E+05, 3.7343E+05, 3.7427E+05, 3.7512E+05, 3.7596E+05,
     +  3.7680E+05, 3.7765E+05, 3.7849E+05, 3.7934E+05, 3.8019E+05,
     + 3.8104E+05, 3.8190E+05, 3.8275E+05, 3.8360E+05, 3.8446E+05,
     +  3.8532E+05, 3.8618E+05, 3.8704E+05, 3.8790E+05, 3.8876E+05,
     + 3.8962E+05, 3.9049E+05, 3.9136E+05, 3.9223E+05, 3.9310E+05,
     +  3.9397E+05, 3.9484E+05, 3.9571E+05, 3.9659E+05, 3.9746E+05,
     + 3.9834E+05, 3.9922E+05, 4.0010E+05, 4.0098E+05, 4.0186E+05,
     +  4.0275E+05, 4.0363E+05, 4.0452E+05, 4.0541E+05, 4.0630E+05,
     + 4.0719E+05, 4.0808E+05, 4.0898E+05, 4.0987E+05, 4.1077E+05,
     +  4.1167E+05, 4.1257E+05, 4.1347E+05, 4.1437E+05, 4.1527E+05,
     + 4.1618E+05, 4.1708E+05, 4.1799E+05, 4.1890E+05, 4.1981E+05,
     +  4.2072E+05, 4.2163E+05, 4.2255E+05, 4.2346E+05, 4.2438E+05,
     + 4.2530E+05, 4.2622E+05, 4.2714E+05, 4.2806E+05, 4.2899E+05,
     +  4.2991E+05, 4.3084E+05, 4.3177E+05, 4.3270E+05, 4.3363E+05,
     + 4.3456E+05, 4.3549E+05, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 3,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       627
      data (QofT( 4,J),J=1,2408)/ 1.3259E+01, 2.4144E+01, 3.5119E+01,
     +  4.6114E+01, 5.7117E+01, 6.8124E+01, 7.9134E+01, 9.0144E+01,
     + 1.0116E+02, 1.1217E+02, 1.2318E+02, 1.3419E+02, 1.4521E+02,
     +  1.5622E+02, 1.6723E+02, 1.7825E+02, 1.8926E+02, 2.0028E+02,
     + 2.1129E+02, 2.2231E+02, 2.4433E+02, 2.6636E+02, 2.8839E+02,
     +  3.1042E+02, 3.3245E+02, 3.5448E+02, 3.7651E+02, 3.9854E+02,
     + 4.2057E+02, 4.4261E+02, 4.6464E+02, 4.8667E+02, 5.0870E+02,
     +  5.3073E+02, 5.5276E+02, 5.7479E+02, 5.9682E+02, 6.1886E+02,
     + 6.4089E+02, 6.6292E+02, 6.8495E+02, 7.0698E+02, 7.2902E+02,
     +  7.5105E+02, 7.7308E+02, 7.9512E+02, 8.1715E+02, 8.3919E+02,
     + 8.6122E+02, 8.8326E+02, 9.0529E+02, 9.2733E+02, 9.4937E+02,
     +  9.7142E+02, 9.9346E+02, 1.0155E+03, 1.0376E+03, 1.0596E+03,
     + 1.0817E+03, 1.1037E+03, 1.1258E+03, 1.1479E+03, 1.1700E+03,
     +  1.1921E+03, 1.2142E+03, 1.2363E+03, 1.2584E+03, 1.2805E+03,
     + 1.3027E+03, 1.3248E+03, 1.3470E+03, 1.3692E+03, 1.3914E+03,
     +  1.4137E+03, 1.4359E+03, 1.4582E+03, 1.4805E+03, 1.5029E+03,
     + 1.5252E+03, 1.5476E+03, 1.5700E+03, 1.5925E+03, 1.6150E+03,
     +  1.6375E+03, 1.6601E+03, 1.6827E+03, 1.7054E+03, 1.7281E+03,
     + 1.7508E+03, 1.7736E+03, 1.7965E+03, 1.8194E+03, 1.8423E+03,
     +  1.8653E+03, 1.8884E+03, 1.9115E+03, 1.9347E+03, 1.9580E+03,
     + 1.9813E+03, 2.0047E+03, 2.0282E+03, 2.0517E+03, 2.0753E+03,
     +  2.0990E+03, 2.1228E+03, 2.1466E+03, 2.1706E+03, 2.1946E+03,
     + 2.2187E+03, 2.2429E+03, 2.2672E+03, 2.2915E+03, 2.3160E+03,
     +  2.3406E+03, 2.3652E+03, 2.3900E+03, 2.4148E+03, 2.4398E+03,
     + 2.4649E+03, 2.4901E+03, 2.5154E+03, 2.5408E+03, 2.5663E+03,
     +  2.5919E+03, 2.6176E+03, 2.6435E+03, 2.6695E+03, 2.6956E+03,
     + 2.7218E+03, 2.7481E+03, 2.7746E+03, 2.8012E+03, 2.8279E+03,
     +  2.8548E+03, 2.8818E+03, 2.9089E+03, 2.9361E+03, 2.9635E+03,
     + 2.9911E+03, 3.0188E+03, 3.0466E+03, 3.0745E+03, 3.1026E+03,
     +  3.1309E+03, 3.1593E+03, 3.1878E+03, 3.2165E+03, 3.2454E+03,
     + 3.2744E+03, 3.3036E+03, 3.3329E+03, 3.3623E+03, 3.3920E+03,
     +  3.4218E+03, 3.4517E+03, 3.4819E+03, 3.5122E+03, 3.5426E+03,
     + 3.5732E+03, 3.6040E+03, 3.6350E+03, 3.6661E+03, 3.6974E+03,
     +  3.7289E+03, 3.7606E+03, 3.7924E+03, 3.8245E+03, 3.8567E+03,
     + 3.8890E+03, 3.9216E+03, 3.9543E+03, 3.9873E+03, 4.0204E+03,
     +  4.0537E+03, 4.0872E+03, 4.1209E+03, 4.1548E+03, 4.1889E+03,
     + 4.2232E+03, 4.2576E+03, 4.2923E+03, 4.3272E+03, 4.3622E+03,
     +  4.3975E+03, 4.4330E+03, 4.4687E+03, 4.5046E+03, 4.5406E+03,
     + 4.5769E+03, 4.6135E+03, 4.6502E+03, 4.6871E+03, 4.7243E+03,
     +  4.7616E+03, 4.7992E+03, 4.8370E+03, 4.8750E+03, 4.9132E+03,
     + 4.9517E+03, 4.9904E+03, 5.0293E+03, 5.0684E+03, 5.1078E+03,
     +  5.1474E+03, 5.1872E+03, 5.2272E+03, 5.2675E+03, 5.3080E+03,
     + 5.3487E+03, 5.3897E+03, 5.4309E+03, 5.4724E+03, 5.5141E+03,
     +  5.5560E+03, 5.5982E+03, 5.6406E+03, 5.6833E+03, 5.7262E+03,
     + 5.7694E+03, 5.8128E+03, 5.8564E+03, 5.9003E+03, 5.9445E+03,
     +  5.9889E+03, 6.0336E+03, 6.0785E+03, 6.1237E+03, 6.1692E+03,
     + 6.2149E+03, 6.2608E+03, 6.3071E+03, 6.3536E+03, 6.4004E+03,
     +  6.4474E+03, 6.4947E+03, 6.5423E+03, 6.5901E+03, 6.6382E+03,
     + 6.6866E+03, 6.7353E+03, 6.7842E+03, 6.8335E+03, 6.8830E+03,
     +  6.9328E+03, 6.9828E+03, 7.0332E+03, 7.0838E+03, 7.1348E+03,
     + 7.1860E+03, 7.2375E+03, 7.2893E+03, 7.3414E+03, 7.3938E+03,
     +  7.4464E+03, 7.4994E+03, 7.5527E+03, 7.6063E+03, 7.6601E+03,
     + 7.7143E+03, 7.7688E+03, 7.8236E+03, 7.8787E+03, 7.9341E+03,
     +  7.9898E+03, 8.0458E+03, 8.1022E+03, 8.1588E+03, 8.2158E+03,
     + 8.2731E+03, 8.3307E+03, 8.3886E+03, 8.4468E+03, 8.5054E+03,
     +  8.5643E+03, 8.6235E+03, 8.6830E+03, 8.7429E+03, 8.8031E+03,
     + 8.8637E+03, 8.9245E+03, 8.9857E+03, 9.0473E+03, 9.1092E+03,
     +  9.1714E+03, 9.2339E+03, 9.2968E+03, 9.3601E+03, 9.4237E+03,
     + 9.4876E+03, 9.5519E+03, 9.6166E+03, 9.6816E+03, 9.7469E+03,
     +  9.8126E+03, 9.8787E+03, 9.9451E+03, 1.0012E+04, 1.0079E+04,
     + 1.0146E+04, 1.0214E+04, 1.0283E+04, 1.0351E+04, 1.0420E+04,
     +  1.0489E+04, 1.0559E+04, 1.0629E+04, 1.0700E+04, 1.0771E+04,
     + 1.0842E+04, 1.0913E+04, 1.0985E+04, 1.1058E+04, 1.1131E+04,
     +  1.1204E+04, 1.1277E+04, 1.1351E+04, 1.1425E+04, 1.1500E+04,
     + 1.1575E+04, 1.1651E+04, 1.1727E+04, 1.1803E+04, 1.1880E+04,
     +  1.1957E+04, 1.2034E+04, 1.2112E+04, 1.2191E+04, 1.2269E+04,
     + 1.2348E+04, 1.2428E+04, 1.2508E+04, 1.2588E+04, 1.2669E+04,
     +  1.2751E+04, 1.2832E+04, 1.2914E+04, 1.2997E+04, 1.3080E+04,
     + 1.3163E+04, 1.3247E+04, 1.3331E+04, 1.3416E+04, 1.3501E+04,
     +  1.3587E+04, 1.3673E+04, 1.3759E+04, 1.3846E+04, 1.3933E+04,
     + 1.4021E+04, 1.4109E+04, 1.4198E+04, 1.4287E+04, 1.4377E+04,
     +  1.4467E+04, 1.4557E+04, 1.4648E+04, 1.4740E+04, 1.4832E+04,
     + 1.4924E+04, 1.5017E+04, 1.5110E+04, 1.5204E+04, 1.5298E+04,
     +  1.5393E+04, 1.5488E+04, 1.5584E+04, 1.5680E+04, 1.5777E+04,
     + 1.5874E+04, 1.5972E+04, 1.6070E+04, 1.6169E+04, 1.6268E+04,
     +  1.6367E+04, 1.6467E+04, 1.6568E+04, 1.6669E+04, 1.6771E+04,
     + 1.6873E+04, 1.6976E+04, 1.7079E+04, 1.7183E+04, 1.7287E+04,
     +  1.7391E+04, 1.7497E+04, 1.7602E+04, 1.7709E+04, 1.7815E+04,
     + 1.7923E+04, 1.8031E+04, 1.8139E+04, 1.8248E+04, 1.8357E+04,
     +  1.8467E+04, 1.8578E+04, 1.8689E+04, 1.8800E+04, 1.8913E+04,
     + 1.9025E+04, 1.9139E+04, 1.9252E+04, 1.9367E+04, 1.9482E+04,
     +  1.9597E+04, 1.9713E+04, 1.9830E+04, 1.9947E+04, 2.0065E+04,
     + 2.0183E+04, 2.0302E+04, 2.0421E+04, 2.0541E+04, 2.0662E+04,
     +  2.0783E+04, 2.0905E+04, 2.1027E+04, 2.1150E+04, 2.1273E+04,
     + 2.1397E+04, 2.1522E+04, 2.1647E+04, 2.1773E+04, 2.1900E+04,
     +  2.2027E+04, 2.2155E+04, 2.2283E+04, 2.2412E+04, 2.2541E+04,
     + 2.2671E+04, 2.2802E+04, 2.2934E+04, 2.3065E+04, 2.3198E+04,
     +  2.3331E+04, 2.3465E+04, 2.3600E+04, 2.3735E+04, 2.3871E+04,
     + 2.4007E+04, 2.4144E+04, 2.4282E+04, 2.4420E+04, 2.4559E+04,
     +  2.4698E+04, 2.4839E+04, 2.4980E+04, 2.5121E+04, 2.5263E+04,
     + 2.5406E+04, 2.5550E+04, 2.5694E+04, 2.5839E+04, 2.5984E+04,
     +  2.6131E+04, 2.6278E+04, 2.6425E+04, 2.6573E+04, 2.6722E+04,
     + 2.6872E+04, 2.7022E+04, 2.7173E+04, 2.7325E+04, 2.7477E+04,
     +  2.7630E+04, 2.7784E+04, 2.7938E+04, 2.8094E+04, 2.8250E+04,
     + 2.8406E+04, 2.8563E+04, 2.8721E+04, 2.8880E+04, 2.9040E+04,
     +  2.9200E+04, 2.9361E+04, 2.9522E+04, 2.9685E+04, 2.9848E+04,
     + 3.0012E+04, 3.0176E+04, 3.0341E+04, 3.0507E+04, 3.0674E+04,
     +  3.0842E+04, 3.1010E+04, 3.1179E+04, 3.1349E+04, 3.1519E+04,
     + 3.1691E+04, 3.1863E+04, 3.2036E+04, 3.2209E+04, 3.2384E+04,
     +  3.2559E+04, 3.2735E+04, 3.2911E+04, 3.3089E+04, 3.3267E+04,
     + 3.3446E+04, 3.3626E+04, 3.3807E+04, 3.3988E+04, 3.4170E+04,
     +  3.4353E+04, 3.4537E+04, 3.4722E+04, 3.4907E+04, 3.5094E+04,
     + 3.5281E+04, 3.5469E+04, 3.5657E+04, 3.5847E+04, 3.6037E+04,
     +  3.6229E+04, 3.6421E+04, 3.6613E+04, 3.6807E+04, 3.7002E+04,
     + 3.7197E+04, 3.7393E+04, 3.7590E+04, 3.7788E+04, 3.7987E+04,
     +  3.8187E+04, 3.8387E+04, 3.8589E+04, 3.8791E+04, 3.8994E+04,
     + 3.9198E+04, 3.9403E+04, 3.9609E+04, 3.9815E+04, 4.0023E+04,
     +  4.0231E+04, 4.0440E+04, 4.0650E+04, 4.0862E+04, 4.1073E+04,
     + 4.1286E+04, 4.1500E+04, 4.1715E+04, 4.1930E+04, 4.2147E+04,
     +  4.2364E+04, 4.2583E+04, 4.2802E+04, 4.3022E+04, 4.3243E+04,
     + 4.3465E+04, 4.3688E+04, 4.3912E+04, 4.4137E+04, 4.4363E+04,
     +  4.4589E+04, 4.4817E+04, 4.5046E+04, 4.5275E+04, 4.5506E+04,
     + 4.5737E+04, 4.5970E+04, 4.6203E+04, 4.6438E+04, 4.6673E+04,
     +  4.6909E+04, 4.7147E+04, 4.7385E+04, 4.7624E+04, 4.7865E+04,
     + 4.8106E+04, 4.8348E+04, 4.8592E+04, 4.8836E+04, 4.9081E+04,
     +  4.9328E+04, 4.9575E+04, 4.9823E+04, 5.0073E+04, 5.0323E+04,
     + 5.0574E+04, 5.0827E+04, 5.1080E+04, 5.1335E+04, 5.1590E+04,
     +  5.1847E+04, 5.2105E+04, 5.2363E+04, 5.2623E+04, 5.2884E+04,
     + 5.3146E+04, 5.3408E+04, 5.3672E+04, 5.3937E+04, 5.4204E+04,
     +  5.4471E+04, 5.4739E+04, 5.5008E+04, 5.5279E+04, 5.5550E+04,
     + 5.5823E+04, 5.6097E+04, 5.6371E+04, 5.6647E+04, 5.6924E+04,
     +  5.7202E+04, 5.7482E+04, 5.7762E+04, 5.8044E+04, 5.8326E+04,
     + 5.8610E+04, 5.8895E+04, 5.9181E+04, 5.9468E+04, 5.9756E+04,
     +  6.0045E+04, 6.0336E+04, 6.0628E+04, 6.0920E+04, 6.1215E+04,
     + 6.1510E+04, 6.1806E+04, 6.2104E+04, 6.2402E+04, 6.2702E+04,
     +  6.3003E+04, 6.3305E+04, 6.3609E+04, 6.3913E+04, 6.4219E+04,
     + 6.4526E+04, 6.4834E+04, 6.5144E+04, 6.5454E+04, 6.5766E+04,
     +  6.6079E+04, 6.6393E+04, 6.6709E+04, 6.7025E+04, 6.7343E+04,
     + 6.7662E+04, 6.7983E+04, 6.8304E+04, 6.8627E+04, 6.8951E+04,
     +  6.9277E+04, 6.9603E+04, 6.9931E+04, 7.0260E+04, 7.0591E+04,
     + 7.0922E+04, 7.1255E+04, 7.1589E+04, 7.1925E+04, 7.2262E+04,
     +  7.2600E+04, 7.2939E+04, 7.3280E+04, 7.3622E+04, 7.3965E+04,
     + 7.4310E+04, 7.4656E+04, 7.5003E+04, 7.5351E+04, 7.5701E+04,
     +  7.6052E+04, 7.6405E+04, 7.6759E+04, 7.7114E+04, 7.7470E+04,
     + 7.7828E+04, 7.8187E+04, 7.8548E+04, 7.8910E+04, 7.9273E+04,
     +  7.9638E+04, 8.0004E+04, 8.0371E+04, 8.0740E+04, 8.1110E+04,
     + 8.1482E+04, 8.1855E+04, 8.2229E+04, 8.2605E+04, 8.2982E+04,
     +  8.3361E+04, 8.3741E+04, 8.4122E+04, 8.4505E+04, 8.4889E+04,
     + 8.5275E+04, 8.5662E+04, 8.6050E+04, 8.6440E+04, 8.6832E+04,
     +  8.7224E+04, 8.7619E+04, 8.8014E+04, 8.8412E+04, 8.8810E+04,
     + 8.9210E+04, 8.9612E+04, 9.0015E+04, 9.0420E+04, 9.0826E+04,
     +  9.1233E+04, 9.1642E+04, 9.2053E+04, 9.2465E+04, 9.2878E+04,
     + 9.3293E+04, 9.3710E+04, 9.4128E+04, 9.4547E+04, 9.4969E+04,
     +  9.5391E+04, 9.5815E+04, 9.6241E+04, 9.6668E+04, 9.7097E+04,
     + 9.7528E+04, 9.7959E+04, 9.8393E+04, 9.8828E+04, 9.9265E+04,
     +  9.9703E+04, 1.0014E+05, 1.0058E+05, 1.0103E+05, 1.0147E+05,
     + 1.0192E+05, 1.0237E+05, 1.0281E+05, 1.0327E+05, 1.0372E+05,
     +  1.0417E+05, 1.0463E+05, 1.0509E+05, 1.0554E+05, 1.0600E+05,
     + 1.0647E+05, 1.0693E+05, 1.0740E+05, 1.0786E+05, 1.0833E+05,
     +  1.0880E+05, 1.0928E+05, 1.0975E+05, 1.1023E+05, 1.1070E+05,
     + 1.1118E+05, 1.1166E+05, 1.1214E+05, 1.1263E+05, 1.1311E+05,
     +  1.1360E+05, 1.1409E+05, 1.1458E+05, 1.1507E+05, 1.1557E+05,
     + 1.1606E+05, 1.1656E+05, 1.1706E+05, 1.1756E+05, 1.1807E+05,
     +  1.1857E+05, 1.1908E+05, 1.1959E+05, 1.2010E+05, 1.2061E+05,
     + 1.2112E+05, 1.2164E+05, 1.2215E+05, 1.2267E+05, 1.2319E+05,
     +  1.2372E+05, 1.2424E+05, 1.2477E+05, 1.2529E+05, 1.2582E+05,
     + 1.2636E+05, 1.2689E+05, 1.2742E+05, 1.2796E+05, 1.2850E+05,
     +  1.2904E+05, 1.2958E+05, 1.3013E+05, 1.3067E+05, 1.3122E+05,
     + 1.3177E+05, 1.3232E+05, 1.3288E+05, 1.3343E+05, 1.3399E+05,
     +  1.3455E+05, 1.3511E+05, 1.3568E+05, 1.3624E+05, 1.3681E+05,
     + 1.3738E+05, 1.3795E+05, 1.3852E+05, 1.3910E+05, 1.3967E+05,
     +  1.4025E+05, 1.4083E+05, 1.4142E+05, 1.4200E+05, 1.4259E+05,
     + 1.4317E+05, 1.4377E+05, 1.4436E+05, 1.4495E+05, 1.4555E+05,
     +  1.4615E+05, 1.4675E+05, 1.4735E+05, 1.4795E+05, 1.4856E+05,
     + 1.4917E+05, 1.4978E+05, 1.5039E+05, 1.5101E+05, 1.5162E+05,
     +  1.5224E+05, 1.5286E+05, 1.5349E+05, 1.5411E+05, 1.5474E+05,
     + 1.5537E+05, 1.5600E+05, 1.5663E+05, 1.5727E+05, 1.5790E+05,
     +  1.5854E+05, 1.5918E+05, 1.5983E+05, 1.6047E+05, 1.6112E+05,
     + 1.6177E+05, 1.6242E+05, 1.6308E+05, 1.6374E+05, 1.6439E+05,
     +  1.6505E+05, 1.6572E+05, 1.6638E+05, 1.6705E+05, 1.6772E+05,
     + 1.6839E+05, 1.6906E+05, 1.6974E+05, 1.7042E+05, 1.7110E+05,
     +  1.7178E+05, 1.7247E+05, 1.7315E+05, 1.7384E+05, 1.7453E+05,
     + 1.7523E+05, 1.7592E+05, 1.7662E+05, 1.7732E+05, 1.7802E+05,
     +  1.7873E+05, 1.7944E+05, 1.8014E+05, 1.8086E+05, 1.8157E+05,
     + 1.8229E+05, 1.8300E+05, 1.8373E+05, 1.8445E+05, 1.8517E+05,
     +  1.8590E+05, 1.8663E+05, 1.8736E+05, 1.8810E+05, 1.8884E+05,
     + 1.8957E+05, 1.9032E+05, 1.9106E+05, 1.9181E+05, 1.9256E+05,
     +  1.9331E+05, 1.9406E+05, 1.9482E+05, 1.9557E+05, 1.9633E+05,
     + 1.9710E+05, 1.9786E+05, 1.9863E+05, 1.9940E+05, 2.0017E+05,
     +  2.0095E+05, 2.0173E+05, 2.0251E+05, 2.0329E+05, 2.0407E+05,
     + 2.0486E+05, 2.0565E+05, 2.0644E+05, 2.0724E+05, 2.0803E+05,
     +  2.0883E+05, 2.0964E+05, 2.1044E+05, 2.1125E+05, 2.1206E+05,
     + 2.1287E+05, 2.1368E+05, 2.1450E+05, 2.1532E+05, 2.1614E+05,
     +  2.1697E+05, 2.1780E+05, 2.1863E+05, 2.1946E+05, 2.2029E+05,
     + 2.2113E+05, 2.2197E+05, 2.2281E+05, 2.2366E+05, 2.2451E+05,
     +  2.2536E+05, 2.2621E+05, 2.2707E+05, 2.2792E+05, 2.2879E+05,
     + 2.2965E+05, 2.3051E+05, 2.3138E+05, 2.3225E+05, 2.3313E+05,
     +  2.3401E+05, 2.3489E+05, 2.3577E+05, 2.3665E+05, 2.3754E+05,
     + 2.3843E+05, 2.3932E+05, 2.4022E+05, 2.4112E+05, 2.4202E+05,
     +  2.4292E+05, 2.4383E+05, 2.4474E+05, 2.4565E+05, 2.4656E+05,
     + 2.4748E+05, 2.4840E+05, 2.4932E+05, 2.5025E+05, 2.5118E+05,
     +  2.5211E+05, 2.5304E+05, 2.5398E+05, 2.5492E+05, 2.5586E+05,
     + 2.5681E+05, 2.5775E+05, 2.5870E+05, 2.5966E+05, 2.6061E+05,
     +  2.6157E+05, 2.6254E+05, 2.6350E+05, 2.6447E+05, 2.6544E+05,
     + 2.6641E+05, 2.6739E+05, 2.6837E+05, 2.6935E+05, 2.7034E+05,
     +  2.7132E+05, 2.7231E+05, 2.7331E+05, 2.7430E+05, 2.7530E+05,
     + 2.7631E+05, 2.7731E+05, 2.7832E+05, 2.7933E+05, 2.8035E+05,
     +  2.8136E+05, 2.8238E+05, 2.8341E+05, 2.8443E+05, 2.8546E+05,
     + 2.8650E+05, 2.8753E+05, 2.8857E+05, 2.8961E+05, 2.9065E+05,
     +  2.9170E+05, 2.9275E+05, 2.9381E+05, 2.9486E+05, 2.9592E+05,
     + 2.9698E+05, 2.9805E+05, 2.9912E+05, 3.0019E+05, 3.0126E+05,
     +  3.0234E+05, 3.0342E+05, 3.0451E+05, 3.0559E+05, 3.0668E+05,
     + 3.0778E+05, 3.0887E+05, 3.0997E+05, 3.1108E+05, 3.1218E+05,
     +  3.1329E+05, 3.1440E+05, 3.1552E+05, 3.1664E+05, 3.1776E+05,
     + 3.1889E+05, 3.2001E+05, 3.2115E+05, 3.2228E+05, 3.2342E+05,
     +  3.2456E+05, 3.2570E+05, 3.2685E+05, 3.2800E+05, 3.2916E+05,
     + 3.3031E+05, 3.3147E+05, 3.3264E+05, 3.3380E+05, 3.3497E+05,
     +  3.3615E+05, 3.3733E+05, 3.3851E+05, 3.3969E+05, 3.4088E+05,
     + 3.4207E+05, 3.4326E+05, 3.4446E+05, 3.4566E+05, 3.4686E+05,
     +  3.4807E+05, 3.4928E+05, 3.5049E+05, 3.5171E+05, 3.5293E+05,
     + 3.5415E+05, 3.5538E+05, 3.5661E+05, 3.5784E+05, 3.5908E+05,
     +  3.6032E+05, 3.6157E+05, 3.6281E+05, 3.6406E+05, 3.6532E+05,
     + 3.6658E+05, 3.6784E+05, 3.6910E+05, 3.7037E+05, 3.7164E+05,
     +  3.7292E+05, 3.7420E+05, 3.7548E+05, 3.7677E+05, 3.7805E+05,
     + 3.7935E+05, 3.8064E+05, 3.8194E+05, 3.8325E+05, 3.8455E+05,
     +  3.8586E+05, 3.8718E+05, 3.8850E+05, 3.8982E+05, 3.9114E+05,
     + 3.9247E+05, 3.9380E+05, 3.9514E+05, 3.9648E+05, 3.9782E+05,
     +  3.9917E+05, 4.0052E+05, 4.0187E+05, 4.0323E+05, 4.0459E+05,
     + 4.0595E+05, 4.0732E+05, 4.0869E+05, 4.1007E+05, 4.1145E+05,
     +  4.1283E+05, 4.1422E+05, 4.1561E+05, 4.1700E+05, 4.1840E+05,
     + 4.1980E+05, 4.2121E+05, 4.2262E+05, 4.2403E+05, 4.2545E+05,
     +  4.2687E+05, 4.2829E+05, 4.2972E+05, 4.3115E+05, 4.3259E+05,
     + 4.3403E+05, 4.3547E+05, 4.3692E+05, 4.3837E+05, 4.3982E+05,
     +  4.4128E+05, 4.4274E+05, 4.4421E+05, 4.4568E+05, 4.4715E+05,
     + 4.4863E+05, 4.5011E+05, 4.5159E+05, 4.5308E+05, 4.5458E+05,
     +  4.5607E+05, 4.5758E+05, 4.5908E+05, 4.6059E+05, 4.6210E+05,
     + 4.6362E+05, 4.6514E+05, 4.6666E+05, 4.6819E+05, 4.6972E+05,
     +  4.7126E+05, 4.7280E+05, 4.7435E+05, 4.7589E+05, 4.7745E+05,
     + 4.7900E+05, 4.8056E+05, 4.8213E+05, 4.8370E+05, 4.8527E+05,
     +  4.8684E+05, 4.8842E+05, 4.9001E+05, 4.9160E+05, 4.9319E+05,
     + 4.9479E+05, 4.9639E+05, 4.9799E+05, 4.9960E+05, 5.0122E+05,
     +  5.0283E+05, 5.0445E+05, 5.0608E+05, 5.0771E+05, 5.0934E+05,
     + 5.1098E+05, 5.1262E+05, 5.1427E+05, 5.1592E+05, 5.1757E+05,
     +  5.1923E+05, 5.2090E+05, 5.2256E+05, 5.2424E+05, 5.2591E+05,
     + 5.2759E+05, 5.2927E+05, 5.3096E+05, 5.3266E+05, 5.3435E+05,
     +  5.3605E+05, 5.3776E+05, 5.3947E+05, 5.4118E+05, 5.4290E+05,
     + 5.4462E+05, 5.4635E+05, 5.4808E+05, 5.4982E+05, 5.5156E+05,
     +  5.5330E+05, 5.5505E+05, 5.5680E+05, 5.5856E+05, 5.6032E+05,
     + 5.6208E+05, 5.6385E+05, 5.6563E+05, 5.6741E+05, 5.6919E+05,
     +  5.7098E+05, 5.7277E+05, 5.7457E+05, 5.7637E+05, 5.7817E+05,
     + 5.7998E+05, 5.8180E+05, 5.8362E+05, 5.8544E+05, 5.8727E+05,
     +  5.8910E+05, 5.9093E+05, 5.9278E+05, 5.9462E+05, 5.9647E+05,
     + 5.9832E+05, 6.0018E+05, 6.0205E+05, 6.0392E+05, 6.0579E+05,
     +  6.0766E+05, 6.0955E+05, 6.1143E+05, 6.1332E+05, 6.1522E+05,
     + 6.1712E+05, 6.1902E+05, 6.2093E+05, 6.2285E+05, 6.2476E+05,
     +  6.2669E+05, 6.2861E+05, 6.3055E+05, 6.3248E+05, 6.3442E+05,
     + 6.3637E+05, 6.3832E+05, 6.4028E+05, 6.4224E+05, 6.4420E+05,
     +  6.4617E+05, 6.4815E+05, 6.5012E+05, 6.5211E+05, 6.5410E+05,
     + 6.5609E+05, 6.5809E+05, 6.6009E+05, 6.6210E+05, 6.6411E+05,
     +  6.6612E+05, 6.6815E+05, 6.7017E+05, 6.7220E+05, 6.7424E+05,
     + 6.7628E+05, 6.7832E+05, 6.8038E+05, 6.8243E+05, 6.8449E+05,
     +  6.8655E+05, 6.8862E+05, 6.9070E+05, 6.9278E+05, 6.9486E+05,
     + 6.9695E+05, 6.9904E+05, 7.0114E+05, 7.0325E+05, 7.0535E+05,
     +  7.0747E+05, 7.0959E+05, 7.1171E+05, 7.1384E+05, 7.1597E+05,
     + 7.1811E+05, 7.2025E+05, 7.2240E+05, 7.2455E+05, 7.2671E+05,
     +  7.2887E+05, 7.3104E+05, 7.3321E+05, 7.3539E+05, 7.3757E+05,
     + 7.3976E+05, 7.4195E+05, 7.4415E+05, 7.4636E+05, 7.4856E+05,
     +  7.5078E+05, 7.5299E+05, 7.5522E+05, 7.5745E+05, 7.5968E+05,
     + 7.6192E+05, 7.6416E+05, 7.6641E+05, 7.6866E+05, 7.7092E+05,
     +  7.7318E+05, 7.7545E+05, 7.7773E+05, 7.8001E+05, 7.8229E+05,
     + 7.8458E+05, 7.8687E+05, 7.8917E+05, 7.9148E+05, 7.9379E+05,
     +  7.9611E+05, 7.9843E+05, 8.0075E+05, 8.0308E+05, 8.0542E+05,
     + 8.0776E+05, 8.1011E+05, 8.1246E+05, 8.1482E+05, 8.1718E+05,
     +  8.1955E+05, 8.2192E+05, 8.2430E+05, 8.2668E+05, 8.2907E+05,
     + 8.3146E+05, 8.3386E+05, 8.3627E+05, 8.3868E+05, 8.4109E+05,
     +  8.4351E+05, 8.4594E+05, 8.4837E+05, 8.5081E+05, 8.5325E+05,
     + 8.5570E+05, 8.5815E+05, 8.6061E+05, 8.6307E+05, 8.6554E+05,
     +  8.6802E+05, 8.7050E+05, 8.7298E+05, 8.7547E+05, 8.7797E+05,
     + 8.8047E+05, 8.8298E+05, 8.8549E+05, 8.8801E+05, 8.9053E+05,
     +  8.9306E+05, 8.9559E+05, 8.9813E+05, 9.0068E+05, 9.0323E+05,
     + 9.0579E+05, 9.0835E+05, 9.1091E+05, 9.1349E+05, 9.1607E+05,
     +  9.1865E+05, 9.2124E+05, 9.2383E+05, 9.2644E+05, 9.2904E+05,
     + 9.3165E+05, 9.3427E+05, 9.3689E+05, 9.3952E+05, 9.4216E+05,
     +  9.4480E+05, 9.4744E+05, 9.5009E+05, 9.5275E+05, 9.5541E+05,
     + 9.5808E+05, 9.6075E+05, 9.6343E+05, 9.6612E+05, 9.6881E+05,
     +  9.7151E+05, 9.7421E+05, 9.7692E+05, 9.7963E+05, 9.8235E+05,
     + 9.8507E+05, 9.8780E+05, 9.9054E+05, 9.9328E+05, 9.9603E+05,
     +  9.9878E+05, 1.0015E+06, 1.0043E+06, 1.0071E+06, 1.0099E+06,
     + 1.0126E+06, 1.0154E+06, 1.0182E+06, 1.0210E+06, 1.0238E+06,
     +  1.0266E+06, 1.0295E+06, 1.0323E+06, 1.0351E+06, 1.0379E+06,
     + 1.0408E+06, 1.0436E+06, 1.0465E+06, 1.0493E+06, 1.0522E+06,
     +  1.0551E+06, 1.0580E+06, 1.0608E+06, 1.0637E+06, 1.0666E+06,
     + 1.0695E+06, 1.0724E+06, 1.0753E+06, 1.0783E+06, 1.0812E+06,
     +  1.0841E+06, 1.0870E+06, 1.0900E+06, 1.0929E+06, 1.0959E+06,
     + 1.0988E+06, 1.1018E+06, 1.1048E+06, 1.1078E+06, 1.1107E+06,
     +  1.1137E+06, 1.1167E+06, 1.1197E+06, 1.1227E+06, 1.1257E+06,
     + 1.1288E+06, 1.1318E+06, 1.1348E+06, 1.1379E+06, 1.1409E+06,
     +  1.1440E+06, 1.1470E+06, 1.1501E+06, 1.1531E+06, 1.1562E+06,
     + 1.1593E+06, 1.1624E+06, 1.1655E+06, 1.1686E+06, 1.1717E+06,
     +  1.1748E+06, 1.1779E+06, 1.1810E+06, 1.1842E+06, 1.1873E+06,
     + 1.1905E+06, 1.1936E+06, 1.1968E+06, 1.1999E+06, 1.2031E+06,
     +  1.2063E+06, 1.2094E+06, 1.2126E+06, 1.2158E+06, 1.2190E+06,
     + 1.2222E+06, 1.2254E+06, 1.2286E+06, 1.2319E+06, 1.2351E+06,
     +  1.2383E+06, 1.2416E+06, 1.2448E+06, 1.2481E+06, 1.2513E+06,
     + 1.2546E+06, 1.2579E+06, 1.2612E+06, 1.2645E+06, 1.2677E+06,
     +  1.2710E+06, 1.2744E+06, 1.2777E+06, 1.2810E+06, 1.2843E+06,
     + 1.2876E+06, 1.2910E+06, 1.2943E+06, 1.2977E+06, 1.3010E+06,
     +  1.3044E+06, 1.3078E+06, 1.3111E+06, 1.3145E+06, 1.3179E+06,
     + 1.3213E+06, 1.3247E+06, 1.3281E+06, 1.3315E+06, 1.3350E+06,
     +  1.3384E+06, 1.3418E+06, 1.3453E+06, 1.3487E+06, 1.3522E+06,
     + 1.3556E+06, 1.3591E+06, 1.3626E+06, 1.3661E+06, 1.3695E+06,
     +  1.3730E+06, 1.3765E+06, 1.3800E+06, 1.3836E+06, 1.3871E+06,
     + 1.3906E+06, 1.3941E+06, 1.3977E+06, 1.4012E+06, 1.4048E+06,
     +  1.4083E+06, 1.4119E+06, 1.4155E+06, 1.4190E+06, 1.4226E+06,
     + 1.4262E+06, 1.4298E+06, 1.4334E+06, 1.4370E+06, 1.4407E+06,
     +  1.4443E+06, 1.4479E+06, 1.4516E+06, 1.4552E+06, 1.4589E+06,
     + 1.4625E+06, 1.4662E+06, 1.4699E+06, 1.4735E+06, 1.4772E+06,
     +  1.4809E+06, 1.4846E+06, 1.4883E+06, 1.4920E+06, 1.4958E+06,
     + 1.4995E+06, 1.5032E+06, 1.5070E+06, 1.5107E+06, 1.5145E+06,
     +  1.5182E+06, 1.5220E+06, 1.5258E+06, 1.5295E+06, 1.5333E+06,
     + 1.5371E+06, 1.5409E+06, 1.5447E+06, 1.5486E+06, 1.5524E+06,
     +  1.5562E+06, 1.5600E+06, 1.5639E+06, 1.5677E+06, 1.5716E+06,
     + 1.5755E+06, 1.5793E+06, 1.5832E+06, 1.5871E+06, 1.5910E+06,
     +  1.5949E+06, 1.5988E+06, 1.6027E+06, 1.6066E+06, 1.6105E+06,
     + 1.6145E+06, 1.6184E+06, 1.6224E+06, 1.6263E+06, 1.6303E+06,
     +  1.6342E+06, 1.6382E+06, 1.6422E+06, 1.6462E+06, 1.6502E+06,
     + 1.6542E+06, 1.6582E+06, 1.6622E+06, 1.6662E+06, 1.6703E+06,
     +  1.6743E+06, 1.6784E+06, 1.6824E+06, 1.6865E+06, 1.6905E+06,
     + 1.6946E+06, 1.6987E+06, 1.7028E+06, 1.7069E+06, 1.7110E+06,
     +  1.7151E+06, 1.7192E+06, 1.7233E+06, 1.7274E+06, 1.7316E+06,
     + 1.7357E+06, 1.7399E+06, 1.7440E+06, 1.7482E+06, 1.7524E+06,
     +  1.7566E+06, 1.7607E+06, 1.7649E+06, 1.7691E+06, 1.7733E+06,
     + 1.7776E+06, 1.7818E+06, 1.7860E+06, 1.7902E+06, 1.7945E+06,
     +  1.7987E+06, 1.8030E+06, 1.8073E+06, 1.8115E+06, 1.8158E+06,
     + 1.8201E+06, 1.8244E+06, 1.8287E+06, 1.8330E+06, 1.8373E+06,
     +  1.8417E+06, 1.8460E+06, 1.8503E+06, 1.8547E+06, 1.8590E+06,
     + 1.8634E+06, 1.8678E+06, 1.8721E+06, 1.8765E+06, 1.8809E+06,
     +  1.8853E+06, 1.8897E+06, 1.8941E+06, 1.8985E+06, 1.9030E+06,
     + 1.9074E+06, 1.9118E+06, 1.9163E+06, 1.9207E+06, 1.9252E+06,
     +  1.9297E+06, 1.9342E+06, 1.9386E+06, 1.9431E+06, 1.9476E+06,
     + 1.9521E+06, 1.9566E+06, 1.9612E+06, 1.9657E+06, 1.9702E+06,
     +  1.9748E+06, 1.9793E+06, 1.9839E+06, 1.9885E+06, 1.9930E+06,
     + 1.9976E+06, 2.0022E+06, 2.0068E+06, 2.0114E+06, 2.0160E+06,
     +  2.0206E+06, 2.0253E+06, 2.0299E+06, 2.0345E+06, 2.0392E+06,
     + 2.0438E+06, 2.0485E+06, 2.0532E+06, 2.0579E+06, 2.0625E+06,
     +  2.0672E+06, 2.0719E+06, 2.0766E+06, 2.0814E+06, 2.0861E+06,
     + 2.0908E+06, 2.0956E+06, 2.1003E+06, 2.1051E+06, 2.1098E+06,
     +  2.1146E+06, 2.1194E+06, 2.1241E+06, 2.1289E+06, 2.1337E+06,
     + 2.1385E+06, 2.1434E+06, 2.1482E+06, 2.1530E+06, 2.1578E+06,
     +  2.1627E+06, 2.1675E+06, 2.1724E+06, 2.1773E+06, 2.1821E+06,
     + 2.1870E+06, 2.1919E+06, 2.1968E+06, 2.2017E+06, 2.2066E+06,
     +  2.2116E+06, 2.2165E+06, 2.2214E+06, 2.2264E+06, 2.2313E+06,
     + 2.2363E+06, 2.2413E+06, 2.2462E+06, 2.2512E+06, 2.2562E+06,
     +  2.2612E+06, 2.2662E+06, 2.2712E+06, 2.2762E+06, 2.2813E+06,
     + 2.2863E+06, 2.2914E+06, 2.2964E+06, 2.3015E+06, 2.3065E+06,
     +  2.3116E+06, 2.3167E+06, 2.3218E+06, 2.3269E+06, 2.3320E+06,
     + 2.3371E+06, 2.3422E+06, 2.3474E+06, 2.3525E+06, 2.3576E+06,
     +  2.3628E+06, 2.3680E+06, 2.3731E+06, 2.3783E+06, 2.3835E+06,
     + 2.3887E+06, 2.3939E+06, 2.3991E+06, 2.4043E+06, 2.4095E+06,
     +  2.4148E+06, 2.4200E+06, 2.4252E+06, 2.4305E+06, 2.4358E+06,
     + 2.4410E+06, 2.4463E+06, 2.4516E+06, 2.4569E+06, 2.4622E+06,
     +  2.4675E+06, 2.4728E+06, 2.4782E+06, 2.4835E+06, 2.4888E+06,
     + 2.4942E+06, 2.4995E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 4,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       638
      data (QofT( 5,J),J=1,2408)/ 4.5211E+00, 8.2548E+00, 1.2017E+01,
     +  1.5787E+01, 1.9558E+01, 2.3331E+01, 2.7105E+01, 3.0879E+01,
     + 3.4654E+01, 3.8429E+01, 4.2204E+01, 4.5979E+01, 4.9754E+01,
     +  5.3529E+01, 5.7304E+01, 6.1079E+01, 6.4855E+01, 6.8630E+01,
     + 7.2405E+01, 7.6181E+01, 8.3732E+01, 9.1283E+01, 9.8834E+01,
     +  1.0639E+02, 1.1394E+02, 1.2149E+02, 1.2904E+02, 1.3659E+02,
     + 1.4414E+02, 1.5169E+02, 1.5925E+02, 1.6680E+02, 1.7435E+02,
     +  1.8190E+02, 1.8945E+02, 1.9700E+02, 2.0456E+02, 2.1211E+02,
     + 2.1966E+02, 2.2721E+02, 2.3476E+02, 2.4232E+02, 2.4987E+02,
     +  2.5742E+02, 2.6497E+02, 2.7252E+02, 2.8008E+02, 2.8763E+02,
     + 2.9518E+02, 3.0274E+02, 3.1029E+02, 3.1785E+02, 3.2540E+02,
     +  3.3296E+02, 3.4052E+02, 3.4808E+02, 3.5564E+02, 3.6320E+02,
     + 3.7076E+02, 3.7833E+02, 3.8590E+02, 3.9347E+02, 4.0104E+02,
     +  4.0862E+02, 4.1620E+02, 4.2379E+02, 4.3138E+02, 4.3898E+02,
     + 4.4658E+02, 4.5419E+02, 4.6181E+02, 4.6943E+02, 4.7706E+02,
     +  4.8470E+02, 4.9234E+02, 5.0000E+02, 5.0767E+02, 5.1535E+02,
     + 5.2304E+02, 5.3074E+02, 5.3845E+02, 5.4618E+02, 5.5392E+02,
     +  5.6168E+02, 5.6945E+02, 5.7724E+02, 5.8505E+02, 5.9287E+02,
     + 6.0071E+02, 6.0858E+02, 6.1646E+02, 6.2436E+02, 6.3228E+02,
     +  6.4023E+02, 6.4819E+02, 6.5619E+02, 6.6420E+02, 6.7224E+02,
     + 6.8031E+02, 6.8840E+02, 6.9652E+02, 7.0467E+02, 7.1285E+02,
     +  7.2105E+02, 7.2929E+02, 7.3755E+02, 7.4585E+02, 7.5418E+02,
     + 7.6254E+02, 7.7094E+02, 7.7937E+02, 7.8783E+02, 7.9633E+02,
     +  8.0487E+02, 8.1344E+02, 8.2205E+02, 8.3070E+02, 8.3939E+02,
     + 8.4811E+02, 8.5688E+02, 8.6569E+02, 8.7453E+02, 8.8342E+02,
     +  8.9236E+02, 9.0133E+02, 9.1035E+02, 9.1942E+02, 9.2853E+02,
     + 9.3768E+02, 9.4688E+02, 9.5613E+02, 9.6542E+02, 9.7476E+02,
     +  9.8415E+02, 9.9359E+02, 1.0031E+03, 1.0126E+03, 1.0222E+03,
     + 1.0319E+03, 1.0415E+03, 1.0513E+03, 1.0611E+03, 1.0709E+03,
     +  1.0808E+03, 1.0908E+03, 1.1008E+03, 1.1109E+03, 1.1210E+03,
     + 1.1312E+03, 1.1414E+03, 1.1517E+03, 1.1621E+03, 1.1725E+03,
     +  1.1830E+03, 1.1935E+03, 1.2041E+03, 1.2147E+03, 1.2255E+03,
     + 1.2362E+03, 1.2471E+03, 1.2580E+03, 1.2689E+03, 1.2800E+03,
     +  1.2911E+03, 1.3022E+03, 1.3134E+03, 1.3247E+03, 1.3361E+03,
     + 1.3475E+03, 1.3590E+03, 1.3705E+03, 1.3821E+03, 1.3938E+03,
     +  1.4056E+03, 1.4174E+03, 1.4293E+03, 1.4413E+03, 1.4533E+03,
     + 1.4654E+03, 1.4776E+03, 1.4899E+03, 1.5022E+03, 1.5146E+03,
     +  1.5271E+03, 1.5396E+03, 1.5522E+03, 1.5649E+03, 1.5777E+03,
     + 1.5905E+03, 1.6035E+03, 1.6165E+03, 1.6295E+03, 1.6427E+03,
     +  1.6559E+03, 1.6692E+03, 1.6826E+03, 1.6961E+03, 1.7097E+03,
     + 1.7233E+03, 1.7370E+03, 1.7508E+03, 1.7647E+03, 1.7786E+03,
     +  1.7927E+03, 1.8068E+03, 1.8210E+03, 1.8353E+03, 1.8497E+03,
     + 1.8642E+03, 1.8787E+03, 1.8933E+03, 1.9081E+03, 1.9229E+03,
     +  1.9378E+03, 1.9528E+03, 1.9678E+03, 1.9830E+03, 1.9983E+03,
     + 2.0136E+03, 2.0290E+03, 2.0446E+03, 2.0602E+03, 2.0759E+03,
     +  2.0917E+03, 2.1076E+03, 2.1236E+03, 2.1397E+03, 2.1558E+03,
     + 2.1721E+03, 2.1885E+03, 2.2050E+03, 2.2215E+03, 2.2382E+03,
     +  2.2549E+03, 2.2718E+03, 2.2887E+03, 2.3058E+03, 2.3229E+03,
     + 2.3402E+03, 2.3575E+03, 2.3750E+03, 2.3925E+03, 2.4102E+03,
     +  2.4280E+03, 2.4458E+03, 2.4638E+03, 2.4818E+03, 2.5000E+03,
     + 2.5183E+03, 2.5367E+03, 2.5552E+03, 2.5738E+03, 2.5925E+03,
     +  2.6113E+03, 2.6302E+03, 2.6492E+03, 2.6684E+03, 2.6876E+03,
     + 2.7070E+03, 2.7264E+03, 2.7460E+03, 2.7657E+03, 2.7855E+03,
     +  2.8054E+03, 2.8255E+03, 2.8456E+03, 2.8659E+03, 2.8862E+03,
     + 2.9067E+03, 2.9273E+03, 2.9481E+03, 2.9689E+03, 2.9899E+03,
     +  3.0109E+03, 3.0321E+03, 3.0535E+03, 3.0749E+03, 3.0964E+03,
     + 3.1181E+03, 3.1399E+03, 3.1618E+03, 3.1839E+03, 3.2060E+03,
     +  3.2283E+03, 3.2508E+03, 3.2733E+03, 3.2960E+03, 3.3188E+03,
     + 3.3417E+03, 3.3647E+03, 3.3879E+03, 3.4112E+03, 3.4346E+03,
     +  3.4582E+03, 3.4819E+03, 3.5057E+03, 3.5297E+03, 3.5538E+03,
     + 3.5780E+03, 3.6023E+03, 3.6268E+03, 3.6514E+03, 3.6762E+03,
     +  3.7011E+03, 3.7261E+03, 3.7513E+03, 3.7766E+03, 3.8020E+03,
     + 3.8276E+03, 3.8533E+03, 3.8792E+03, 3.9052E+03, 3.9313E+03,
     +  3.9576E+03, 3.9840E+03, 4.0106E+03, 4.0373E+03, 4.0642E+03,
     + 4.0912E+03, 4.1183E+03, 4.1456E+03, 4.1731E+03, 4.2006E+03,
     +  4.2284E+03, 4.2563E+03, 4.2843E+03, 4.3125E+03, 4.3408E+03,
     + 4.3693E+03, 4.3979E+03, 4.4267E+03, 4.4557E+03, 4.4848E+03,
     +  4.5140E+03, 4.5434E+03, 4.5730E+03, 4.6027E+03, 4.6326E+03,
     + 4.6626E+03, 4.6928E+03, 4.7231E+03, 4.7537E+03, 4.7843E+03,
     +  4.8152E+03, 4.8462E+03, 4.8773E+03, 4.9086E+03, 4.9401E+03,
     + 4.9718E+03, 5.0036E+03, 5.0356E+03, 5.0677E+03, 5.1000E+03,
     +  5.1325E+03, 5.1652E+03, 5.1980E+03, 5.2310E+03, 5.2641E+03,
     + 5.2975E+03, 5.3310E+03, 5.3646E+03, 5.3985E+03, 5.4325E+03,
     +  5.4667E+03, 5.5011E+03, 5.5356E+03, 5.5704E+03, 5.6053E+03,
     + 5.6404E+03, 5.6756E+03, 5.7111E+03, 5.7467E+03, 5.7825E+03,
     +  5.8185E+03, 5.8547E+03, 5.8910E+03, 5.9276E+03, 5.9643E+03,
     + 6.0012E+03, 6.0383E+03, 6.0756E+03, 6.1131E+03, 6.1508E+03,
     +  6.1886E+03, 6.2267E+03, 6.2649E+03, 6.3033E+03, 6.3419E+03,
     + 6.3808E+03, 6.4198E+03, 6.4590E+03, 6.4984E+03, 6.5380E+03,
     +  6.5778E+03, 6.6177E+03, 6.6579E+03, 6.6983E+03, 6.7389E+03,
     + 6.7797E+03, 6.8207E+03, 6.8619E+03, 6.9033E+03, 6.9449E+03,
     +  6.9867E+03, 7.0287E+03, 7.0710E+03, 7.1134E+03, 7.1560E+03,
     + 7.1989E+03, 7.2419E+03, 7.2852E+03, 7.3287E+03, 7.3724E+03,
     +  7.4163E+03, 7.4604E+03, 7.5048E+03, 7.5493E+03, 7.5941E+03,
     + 7.6391E+03, 7.6843E+03, 7.7297E+03, 7.7753E+03, 7.8212E+03,
     +  7.8673E+03, 7.9136E+03, 7.9601E+03, 8.0069E+03, 8.0539E+03,
     + 8.1011E+03, 8.1485E+03, 8.1962E+03, 8.2441E+03, 8.2922E+03,
     +  8.3405E+03, 8.3891E+03, 8.4379E+03, 8.4870E+03, 8.5362E+03,
     + 8.5858E+03, 8.6355E+03, 8.6855E+03, 8.7357E+03, 8.7862E+03,
     +  8.8369E+03, 8.8878E+03, 8.9390E+03, 8.9904E+03, 9.0421E+03,
     + 9.0940E+03, 9.1461E+03, 9.1985E+03, 9.2512E+03, 9.3041E+03,
     +  9.3572E+03, 9.4106E+03, 9.4642E+03, 9.5181E+03, 9.5722E+03,
     + 9.6266E+03, 9.6813E+03, 9.7362E+03, 9.7913E+03, 9.8467E+03,
     +  9.9024E+03, 9.9583E+03, 1.0014E+04, 1.0071E+04, 1.0128E+04,
     + 1.0185E+04, 1.0242E+04, 1.0299E+04, 1.0357E+04, 1.0415E+04,
     +  1.0473E+04, 1.0532E+04, 1.0591E+04, 1.0650E+04, 1.0709E+04,
     + 1.0769E+04, 1.0829E+04, 1.0889E+04, 1.0949E+04, 1.1010E+04,
     +  1.1071E+04, 1.1132E+04, 1.1194E+04, 1.1256E+04, 1.1318E+04,
     + 1.1380E+04, 1.1443E+04, 1.1506E+04, 1.1569E+04, 1.1633E+04,
     +  1.1697E+04, 1.1761E+04, 1.1825E+04, 1.1890E+04, 1.1955E+04,
     + 1.2020E+04, 1.2086E+04, 1.2151E+04, 1.2218E+04, 1.2284E+04,
     +  1.2351E+04, 1.2418E+04, 1.2485E+04, 1.2553E+04, 1.2621E+04,
     + 1.2689E+04, 1.2757E+04, 1.2826E+04, 1.2895E+04, 1.2965E+04,
     +  1.3035E+04, 1.3105E+04, 1.3175E+04, 1.3246E+04, 1.3317E+04,
     + 1.3388E+04, 1.3460E+04, 1.3532E+04, 1.3604E+04, 1.3676E+04,
     +  1.3749E+04, 1.3822E+04, 1.3896E+04, 1.3970E+04, 1.4044E+04,
     + 1.4118E+04, 1.4193E+04, 1.4268E+04, 1.4344E+04, 1.4420E+04,
     +  1.4496E+04, 1.4572E+04, 1.4649E+04, 1.4726E+04, 1.4803E+04,
     + 1.4881E+04, 1.4959E+04, 1.5038E+04, 1.5116E+04, 1.5196E+04,
     +  1.5275E+04, 1.5355E+04, 1.5435E+04, 1.5515E+04, 1.5596E+04,
     + 1.5677E+04, 1.5759E+04, 1.5841E+04, 1.5923E+04, 1.6005E+04,
     +  1.6088E+04, 1.6172E+04, 1.6255E+04, 1.6339E+04, 1.6424E+04,
     + 1.6508E+04, 1.6593E+04, 1.6679E+04, 1.6764E+04, 1.6850E+04,
     +  1.6937E+04, 1.7024E+04, 1.7111E+04, 1.7199E+04, 1.7286E+04,
     + 1.7375E+04, 1.7463E+04, 1.7553E+04, 1.7642E+04, 1.7732E+04,
     +  1.7822E+04, 1.7912E+04, 1.8003E+04, 1.8095E+04, 1.8186E+04,
     + 1.8278E+04, 1.8371E+04, 1.8464E+04, 1.8557E+04, 1.8650E+04,
     +  1.8744E+04, 1.8839E+04, 1.8933E+04, 1.9029E+04, 1.9124E+04,
     + 1.9220E+04, 1.9316E+04, 1.9413E+04, 1.9510E+04, 1.9608E+04,
     +  1.9705E+04, 1.9804E+04, 1.9902E+04, 2.0002E+04, 2.0101E+04,
     + 2.0201E+04, 2.0301E+04, 2.0402E+04, 2.0503E+04, 2.0605E+04,
     +  2.0707E+04, 2.0809E+04, 2.0912E+04, 2.1015E+04, 2.1119E+04,
     + 2.1223E+04, 2.1327E+04, 2.1432E+04, 2.1537E+04, 2.1643E+04,
     +  2.1749E+04, 2.1856E+04, 2.1963E+04, 2.2070E+04, 2.2178E+04,
     + 2.2286E+04, 2.2395E+04, 2.2504E+04, 2.2614E+04, 2.2724E+04,
     +  2.2834E+04, 2.2945E+04, 2.3056E+04, 2.3168E+04, 2.3280E+04,
     + 2.3393E+04, 2.3506E+04, 2.3620E+04, 2.3734E+04, 2.3848E+04,
     +  2.3963E+04, 2.4078E+04, 2.4194E+04, 2.4311E+04, 2.4427E+04,
     + 2.4544E+04, 2.4662E+04, 2.4780E+04, 2.4899E+04, 2.5018E+04,
     +  2.5137E+04, 2.5257E+04, 2.5378E+04, 2.5499E+04, 2.5620E+04,
     + 2.5742E+04, 2.5864E+04, 2.5987E+04, 2.6110E+04, 2.6234E+04,
     +  2.6358E+04, 2.6483E+04, 2.6608E+04, 2.6734E+04, 2.6860E+04,
     + 2.6986E+04, 2.7114E+04, 2.7241E+04, 2.7369E+04, 2.7498E+04,
     +  2.7627E+04, 2.7757E+04, 2.7887E+04, 2.8017E+04, 2.8148E+04,
     + 2.8280E+04, 2.8412E+04, 2.8545E+04, 2.8678E+04, 2.8811E+04,
     +  2.8945E+04, 2.9080E+04, 2.9215E+04, 2.9351E+04, 2.9487E+04,
     + 2.9624E+04, 2.9761E+04, 2.9898E+04, 3.0037E+04, 3.0175E+04,
     +  3.0315E+04, 3.0455E+04, 3.0595E+04, 3.0736E+04, 3.0877E+04,
     + 3.1019E+04, 3.1161E+04, 3.1304E+04, 3.1448E+04, 3.1592E+04,
     +  3.1737E+04, 3.1882E+04, 3.2027E+04, 3.2174E+04, 3.2320E+04,
     + 3.2468E+04, 3.2615E+04, 3.2764E+04, 3.2913E+04, 3.3062E+04,
     +  3.3212E+04, 3.3363E+04, 3.3514E+04, 3.3666E+04, 3.3818E+04,
     + 3.3971E+04, 3.4124E+04, 3.4278E+04, 3.4433E+04, 3.4588E+04,
     +  3.4744E+04, 3.4900E+04, 3.5057E+04, 3.5214E+04, 3.5372E+04,
     + 3.5531E+04, 3.5690E+04, 3.5849E+04, 3.6010E+04, 3.6171E+04,
     +  3.6332E+04, 3.6494E+04, 3.6657E+04, 3.6820E+04, 3.6984E+04,
     + 3.7148E+04, 3.7313E+04, 3.7479E+04, 3.7645E+04, 3.7812E+04,
     +  3.7979E+04, 3.8147E+04, 3.8316E+04, 3.8485E+04, 3.8655E+04,
     + 3.8825E+04, 3.8996E+04, 3.9168E+04, 3.9340E+04, 3.9513E+04,
     +  3.9687E+04, 3.9861E+04, 4.0036E+04, 4.0211E+04, 4.0387E+04,
     + 4.0564E+04, 4.0741E+04, 4.0919E+04, 4.1098E+04, 4.1277E+04,
     +  4.1457E+04, 4.1637E+04, 4.1818E+04, 4.2000E+04, 4.2182E+04,
     + 4.2365E+04, 4.2549E+04, 4.2733E+04, 4.2918E+04, 4.3104E+04,
     +  4.3290E+04, 4.3477E+04, 4.3664E+04, 4.3853E+04, 4.4042E+04,
     + 4.4231E+04, 4.4421E+04, 4.4612E+04, 4.4804E+04, 4.4996E+04,
     +  4.5189E+04, 4.5382E+04, 4.5577E+04, 4.5772E+04, 4.5967E+04,
     + 4.6163E+04, 4.6360E+04, 4.6558E+04, 4.6756E+04, 4.6955E+04,
     +  4.7155E+04, 4.7355E+04, 4.7556E+04, 4.7758E+04, 4.7961E+04,
     + 4.8164E+04, 4.8368E+04, 4.8572E+04, 4.8778E+04, 4.8983E+04,
     +  4.9190E+04, 4.9398E+04, 4.9606E+04, 4.9814E+04, 5.0024E+04,
     + 5.0234E+04, 5.0445E+04, 5.0657E+04, 5.0869E+04, 5.1082E+04,
     +  5.1296E+04, 5.1511E+04, 5.1726E+04, 5.1942E+04, 5.2159E+04,
     + 5.2376E+04, 5.2595E+04, 5.2814E+04, 5.3033E+04, 5.3254E+04,
     +  5.3475E+04, 5.3697E+04, 5.3920E+04, 5.4143E+04, 5.4367E+04,
     + 5.4592E+04, 5.4818E+04, 5.5045E+04, 5.5272E+04, 5.5500E+04,
     +  5.5729E+04, 5.5958E+04, 5.6188E+04, 5.6419E+04, 5.6651E+04,
     + 5.6884E+04, 5.7117E+04, 5.7351E+04, 5.7586E+04, 5.7822E+04,
     +  5.8058E+04, 5.8296E+04, 5.8534E+04, 5.8773E+04, 5.9012E+04,
     + 5.9253E+04, 5.9494E+04, 5.9736E+04, 5.9979E+04, 6.0222E+04,
     +  6.0467E+04, 6.0712E+04, 6.0958E+04, 6.1205E+04, 6.1453E+04,
     + 6.1701E+04, 6.1950E+04, 6.2201E+04, 6.2452E+04, 6.2703E+04,
     +  6.2956E+04, 6.3209E+04, 6.3463E+04, 6.3718E+04, 6.3974E+04,
     + 6.4231E+04, 6.4489E+04, 6.4747E+04, 6.5006E+04, 6.5266E+04,
     +  6.5527E+04, 6.5789E+04, 6.6051E+04, 6.6315E+04, 6.6579E+04,
     + 6.6844E+04, 6.7110E+04, 6.7377E+04, 6.7645E+04, 6.7914E+04,
     +  6.8183E+04, 6.8453E+04, 6.8724E+04, 6.8996E+04, 6.9269E+04,
     + 6.9543E+04, 6.9818E+04, 7.0093E+04, 7.0370E+04, 7.0647E+04,
     +  7.0925E+04, 7.1204E+04, 7.1484E+04, 7.1765E+04, 7.2047E+04,
     + 7.2330E+04, 7.2613E+04, 7.2898E+04, 7.3183E+04, 7.3469E+04,
     +  7.3756E+04, 7.4044E+04, 7.4333E+04, 7.4623E+04, 7.4914E+04,
     + 7.5206E+04, 7.5498E+04, 7.5792E+04, 7.6086E+04, 7.6382E+04,
     +  7.6678E+04, 7.6975E+04, 7.7273E+04, 7.7573E+04, 7.7873E+04,
     + 7.8174E+04, 7.8476E+04, 7.8778E+04, 7.9082E+04, 7.9387E+04,
     +  7.9693E+04, 7.9999E+04, 8.0307E+04, 8.0616E+04, 8.0925E+04,
     + 8.1236E+04, 8.1547E+04, 8.1859E+04, 8.2173E+04, 8.2487E+04,
     +  8.2802E+04, 8.3119E+04, 8.3436E+04, 8.3754E+04, 8.4073E+04,
     + 8.4394E+04, 8.4715E+04, 8.5037E+04, 8.5360E+04, 8.5684E+04,
     +  8.6009E+04, 8.6336E+04, 8.6663E+04, 8.6991E+04, 8.7320E+04,
     + 8.7650E+04, 8.7981E+04, 8.8313E+04, 8.8647E+04, 8.8981E+04,
     +  8.9316E+04, 8.9652E+04, 8.9989E+04, 9.0328E+04, 9.0667E+04,
     + 9.1007E+04, 9.1349E+04, 9.1691E+04, 9.2034E+04, 9.2379E+04,
     +  9.2724E+04, 9.3071E+04, 9.3418E+04, 9.3767E+04, 9.4117E+04,
     + 9.4467E+04, 9.4819E+04, 9.5172E+04, 9.5526E+04, 9.5881E+04,
     +  9.6237E+04, 9.6594E+04, 9.6952E+04, 9.7311E+04, 9.7671E+04,
     + 9.8033E+04, 9.8395E+04, 9.8758E+04, 9.9123E+04, 9.9489E+04,
     +  9.9855E+04, 1.0022E+05, 1.0059E+05, 1.0096E+05, 1.0133E+05,
     + 1.0171E+05, 1.0208E+05, 1.0245E+05, 1.0283E+05, 1.0321E+05,
     +  1.0358E+05, 1.0396E+05, 1.0434E+05, 1.0472E+05, 1.0510E+05,
     + 1.0549E+05, 1.0587E+05, 1.0626E+05, 1.0664E+05, 1.0703E+05,
     +  1.0742E+05, 1.0781E+05, 1.0820E+05, 1.0859E+05, 1.0899E+05,
     + 1.0938E+05, 1.0978E+05, 1.1018E+05, 1.1057E+05, 1.1097E+05,
     +  1.1137E+05, 1.1177E+05, 1.1218E+05, 1.1258E+05, 1.1299E+05,
     + 1.1339E+05, 1.1380E+05, 1.1421E+05, 1.1462E+05, 1.1503E+05,
     +  1.1544E+05, 1.1585E+05, 1.1627E+05, 1.1668E+05, 1.1710E+05,
     + 1.1752E+05, 1.1794E+05, 1.1836E+05, 1.1878E+05, 1.1920E+05,
     +  1.1963E+05, 1.2005E+05, 1.2048E+05, 1.2091E+05, 1.2134E+05,
     + 1.2177E+05, 1.2220E+05, 1.2263E+05, 1.2306E+05, 1.2350E+05,
     +  1.2393E+05, 1.2437E+05, 1.2481E+05, 1.2525E+05, 1.2569E+05,
     + 1.2613E+05, 1.2658E+05, 1.2702E+05, 1.2747E+05, 1.2792E+05,
     +  1.2836E+05, 1.2881E+05, 1.2927E+05, 1.2972E+05, 1.3017E+05,
     + 1.3063E+05, 1.3108E+05, 1.3154E+05, 1.3200E+05, 1.3246E+05,
     +  1.3292E+05, 1.3338E+05, 1.3385E+05, 1.3431E+05, 1.3478E+05,
     + 1.3525E+05, 1.3572E+05, 1.3619E+05, 1.3666E+05, 1.3713E+05,
     +  1.3760E+05, 1.3808E+05, 1.3856E+05, 1.3904E+05, 1.3951E+05,
     + 1.4000E+05, 1.4048E+05, 1.4096E+05, 1.4145E+05, 1.4193E+05,
     +  1.4242E+05, 1.4291E+05, 1.4340E+05, 1.4389E+05, 1.4438E+05,
     + 1.4488E+05, 1.4537E+05, 1.4587E+05, 1.4637E+05, 1.4687E+05,
     +  1.4737E+05, 1.4787E+05, 1.4837E+05, 1.4888E+05, 1.4938E+05,
     + 1.4989E+05, 1.5040E+05, 1.5091E+05, 1.5142E+05, 1.5194E+05,
     +  1.5245E+05, 1.5297E+05, 1.5349E+05, 1.5400E+05, 1.5452E+05,
     + 1.5505E+05, 1.5557E+05, 1.5609E+05, 1.5662E+05, 1.5715E+05,
     +  1.5767E+05, 1.5820E+05, 1.5874E+05, 1.5927E+05, 1.5980E+05,
     + 1.6034E+05, 1.6088E+05, 1.6141E+05, 1.6195E+05, 1.6250E+05,
     +  1.6304E+05, 1.6358E+05, 1.6413E+05, 1.6468E+05, 1.6522E+05,
     + 1.6577E+05, 1.6633E+05, 1.6688E+05, 1.6743E+05, 1.6799E+05,
     +  1.6855E+05, 1.6910E+05, 1.6967E+05, 1.7023E+05, 1.7079E+05,
     + 1.7135E+05, 1.7192E+05, 1.7249E+05, 1.7306E+05, 1.7363E+05,
     +  1.7420E+05, 1.7477E+05, 1.7535E+05, 1.7593E+05, 1.7650E+05,
     + 1.7708E+05, 1.7766E+05, 1.7825E+05, 1.7883E+05, 1.7942E+05,
     +  1.8000E+05, 1.8059E+05, 1.8118E+05, 1.8177E+05, 1.8237E+05,
     + 1.8296E+05, 1.8356E+05, 1.8416E+05, 1.8475E+05, 1.8536E+05,
     +  1.8596E+05, 1.8656E+05, 1.8717E+05, 1.8777E+05, 1.8838E+05,
     + 1.8899E+05, 1.8960E+05, 1.9022E+05, 1.9083E+05, 1.9145E+05,
     +  1.9207E+05, 1.9269E+05, 1.9331E+05, 1.9393E+05, 1.9455E+05,
     + 1.9518E+05, 1.9581E+05, 1.9644E+05, 1.9707E+05, 1.9770E+05,
     +  1.9833E+05, 1.9897E+05, 1.9961E+05, 2.0024E+05, 2.0088E+05,
     + 2.0153E+05, 2.0217E+05, 2.0281E+05, 2.0346E+05, 2.0411E+05,
     +  2.0476E+05, 2.0541E+05, 2.0606E+05, 2.0672E+05, 2.0738E+05,
     + 2.0803E+05, 2.0869E+05, 2.0935E+05, 2.1002E+05, 2.1068E+05,
     +  2.1135E+05, 2.1202E+05, 2.1269E+05, 2.1336E+05, 2.1403E+05,
     + 2.1470E+05, 2.1538E+05, 2.1606E+05, 2.1674E+05, 2.1742E+05,
     +  2.1810E+05, 2.1879E+05, 2.1947E+05, 2.2016E+05, 2.2085E+05,
     + 2.2154E+05, 2.2223E+05, 2.2293E+05, 2.2362E+05, 2.2432E+05,
     +  2.2502E+05, 2.2572E+05, 2.2643E+05, 2.2713E+05, 2.2784E+05,
     + 2.2855E+05, 2.2926E+05, 2.2997E+05, 2.3068E+05, 2.3140E+05,
     +  2.3211E+05, 2.3283E+05, 2.3355E+05, 2.3428E+05, 2.3500E+05,
     + 2.3573E+05, 2.3645E+05, 2.3718E+05, 2.3791E+05, 2.3865E+05,
     +  2.3938E+05, 2.4012E+05, 2.4085E+05, 2.4159E+05, 2.4234E+05,
     + 2.4308E+05, 2.4382E+05, 2.4457E+05, 2.4532E+05, 2.4607E+05,
     +  2.4682E+05, 2.4758E+05, 2.4833E+05, 2.4909E+05, 2.4985E+05,
     + 2.5061E+05, 2.5137E+05, 2.5214E+05, 2.5290E+05, 2.5367E+05,
     +  2.5444E+05, 2.5521E+05, 2.5599E+05, 2.5676E+05, 2.5754E+05,
     + 2.5832E+05, 2.5910E+05, 2.5988E+05, 2.6067E+05, 2.6146E+05,
     +  2.6224E+05, 2.6303E+05, 2.6383E+05, 2.6462E+05, 2.6542E+05,
     + 2.6621E+05, 2.6701E+05, 2.6781E+05, 2.6862E+05, 2.6942E+05,
     +  2.7023E+05, 2.7104E+05, 2.7185E+05, 2.7266E+05, 2.7348E+05,
     + 2.7429E+05, 2.7511E+05, 2.7593E+05, 2.7675E+05, 2.7758E+05,
     +  2.7840E+05, 2.7923E+05, 2.8006E+05, 2.8089E+05, 2.8172E+05,
     + 2.8256E+05, 2.8340E+05, 2.8424E+05, 2.8508E+05, 2.8592E+05,
     +  2.8677E+05, 2.8761E+05, 2.8846E+05, 2.8931E+05, 2.9016E+05,
     + 2.9102E+05, 2.9187E+05, 2.9273E+05, 2.9359E+05, 2.9446E+05,
     +  2.9532E+05, 2.9619E+05, 2.9705E+05, 2.9792E+05, 2.9880E+05,
     + 2.9967E+05, 3.0055E+05, 3.0142E+05, 3.0230E+05, 3.0319E+05,
     +  3.0407E+05, 3.0496E+05, 3.0584E+05, 3.0673E+05, 3.0763E+05,
     + 3.0852E+05, 3.0942E+05, 3.1031E+05, 3.1121E+05, 3.1211E+05,
     +  3.1302E+05, 3.1392E+05, 3.1483E+05, 3.1574E+05, 3.1665E+05,
     + 3.1757E+05, 3.1848E+05, 3.1940E+05, 3.2032E+05, 3.2124E+05,
     +  3.2217E+05, 3.2309E+05, 3.2402E+05, 3.2495E+05, 3.2588E+05,
     + 3.2682E+05, 3.2775E+05, 3.2869E+05, 3.2963E+05, 3.3057E+05,
     +  3.3152E+05, 3.3246E+05, 3.3341E+05, 3.3436E+05, 3.3532E+05,
     + 3.3627E+05, 3.3723E+05, 3.3819E+05, 3.3915E+05, 3.4011E+05,
     +  3.4108E+05, 3.4204E+05, 3.4301E+05, 3.4398E+05, 3.4496E+05,
     + 3.4593E+05, 3.4691E+05, 3.4789E+05, 3.4887E+05, 3.4986E+05,
     +  3.5084E+05, 3.5183E+05, 3.5282E+05, 3.5381E+05, 3.5481E+05,
     + 3.5580E+05, 3.5680E+05, 3.5780E+05, 3.5881E+05, 3.5981E+05,
     +  3.6082E+05, 3.6183E+05, 3.6284E+05, 3.6385E+05, 3.6487E+05,
     + 3.6589E+05, 3.6691E+05, 3.6793E+05, 3.6896E+05, 3.6998E+05,
     +  3.7101E+05, 3.7204E+05, 3.7308E+05, 3.7411E+05, 3.7515E+05,
     + 3.7619E+05, 3.7723E+05, 3.7827E+05, 3.7932E+05, 3.8037E+05,
     +  3.8142E+05, 3.8247E+05, 3.8353E+05, 3.8458E+05, 3.8564E+05,
     + 3.8670E+05, 3.8777E+05, 3.8883E+05, 3.8990E+05, 3.9097E+05,
     +  3.9205E+05, 3.9312E+05, 3.9420E+05, 3.9528E+05, 3.9636E+05,
     + 3.9744E+05, 3.9853E+05, 3.9962E+05, 4.0071E+05, 4.0180E+05,
     +  4.0289E+05, 4.0399E+05, 4.0509E+05, 4.0619E+05, 4.0730E+05,
     + 4.0840E+05, 4.0951E+05, 4.1062E+05, 4.1173E+05, 4.1285E+05,
     +  4.1397E+05, 4.1509E+05, 4.1621E+05, 4.1733E+05, 4.1846E+05,
     + 4.1959E+05, 4.2072E+05, 4.2185E+05, 4.2299E+05, 4.2413E+05,
     +  4.2527E+05, 4.2641E+05, 4.2756E+05, 4.2870E+05, 4.2985E+05,
     + 4.3100E+05, 4.3216E+05, 4.3331E+05, 4.3447E+05, 4.3563E+05,
     +  4.3680E+05, 4.3796E+05, 4.3913E+05, 4.4030E+05, 4.4147E+05,
     + 4.4265E+05, 4.4383E+05, 4.4501E+05, 4.4619E+05, 4.4737E+05,
     +  4.4856E+05, 4.4975E+05, 4.5094E+05, 4.5213E+05, 4.5333E+05,
     + 4.5453E+05, 4.5573E+05, 4.5693E+05, 4.5814E+05, 4.5934E+05,
     +  4.6055E+05, 4.6177E+05, 4.6298E+05, 4.6420E+05, 4.6542E+05,
     + 4.6664E+05, 4.6787E+05, 4.6909E+05, 4.7032E+05, 4.7155E+05,
     +  4.7279E+05, 4.7402E+05, 4.7526E+05, 4.7650E+05, 4.7775E+05,
     + 4.7899E+05, 4.8024E+05, 4.8149E+05, 4.8275E+05, 4.8400E+05,
     +  4.8526E+05, 4.8652E+05, 4.8778E+05, 4.8905E+05, 4.9032E+05,
     + 4.9159E+05, 4.9286E+05, 4.9413E+05, 4.9541E+05, 4.9669E+05,
     +  4.9797E+05, 4.9926E+05, 5.0055E+05, 5.0184E+05, 5.0313E+05,
     + 5.0442E+05, 5.0572E+05, 5.0702E+05, 5.0832E+05, 5.0963E+05,
     +  5.1093E+05, 5.1224E+05, 5.1356E+05, 5.1487E+05, 5.1619E+05,
     + 5.1751E+05, 5.1883E+05, 5.2015E+05, 5.2148E+05, 5.2281E+05,
     +  5.2414E+05, 5.2547E+05, 5.2681E+05, 5.2815E+05, 5.2949E+05,
     + 5.3084E+05, 5.3218E+05, 5.3353E+05, 5.3488E+05, 5.3624E+05,
     +  5.3759E+05, 5.3895E+05, 5.4032E+05, 5.4168E+05, 5.4305E+05,
     + 5.4442E+05, 5.4579E+05, 5.4716E+05, 5.4854E+05, 5.4992E+05,
     +  5.5130E+05, 5.5269E+05, 5.5407E+05, 5.5546E+05, 5.5685E+05,
     + 5.5825E+05, 5.5965E+05, 5.6105E+05, 5.6245E+05, 5.6385E+05,
     +  5.6526E+05, 5.6667E+05, 5.6808E+05, 5.6950E+05, 5.7092E+05,
     + 5.7234E+05, 5.7376E+05, 5.7519E+05, 5.7661E+05, 5.7804E+05,
     +  5.7948E+05, 5.8091E+05, 5.8235E+05, 5.8379E+05, 5.8524E+05,
     + 5.8668E+05, 5.8813E+05, 5.8958E+05, 5.9104E+05, 5.9249E+05,
     +  5.9395E+05, 5.9541E+05, 5.9688E+05, 5.9835E+05, 5.9982E+05,
     + 6.0129E+05, 6.0276E+05, 6.0424E+05, 6.0572E+05, 6.0720E+05,
     +  6.0869E+05, 6.1018E+05, 6.1167E+05, 6.1316E+05, 6.1466E+05,
     + 6.1615E+05, 6.1766E+05, 6.1916E+05, 6.2067E+05, 6.2217E+05,
     +  6.2369E+05, 6.2520E+05, 6.2672E+05, 6.2824E+05, 6.2976E+05,
     + 6.3128E+05, 6.3281E+05, 6.3434E+05, 6.3588E+05, 6.3741E+05,
     +  6.3895E+05, 6.4049E+05, 6.4203E+05, 6.4358E+05, 6.4513E+05,
     + 6.4668E+05, 6.4824E+05, 6.4979E+05, 6.5135E+05, 6.5292E+05,
     +  6.5448E+05, 6.5605E+05, 6.5762E+05, 6.5919E+05, 6.6077E+05,
     + 6.6235E+05, 6.6393E+05, 6.6551E+05, 6.6710E+05, 6.6869E+05,
     +  6.7028E+05, 6.7188E+05, 6.7347E+05, 6.7507E+05, 6.7668E+05,
     + 6.7828E+05, 6.7989E+05, 6.8150E+05, 6.8312E+05, 6.8473E+05,
     +  6.8635E+05, 6.8797E+05, 6.8960E+05, 6.9123E+05, 6.9286E+05,
     + 6.9449E+05, 6.9613E+05, 6.9777E+05, 6.9941E+05, 7.0105E+05,
     +  7.0270E+05, 7.0435E+05, 7.0600E+05, 7.0765E+05, 7.0931E+05,
     + 7.1097E+05, 7.1264E+05, 7.1430E+05, 7.1597E+05, 7.1764E+05,
     +  7.1932E+05, 7.2100E+05, 7.2268E+05, 7.2436E+05, 7.2604E+05,
     + 7.2773E+05, 7.2942E+05, 7.3112E+05, 7.3281E+05, 7.3451E+05,
     +  7.3622E+05, 7.3792E+05, 7.3963E+05, 7.4134E+05, 7.4305E+05,
     + 7.4477E+05, 7.4649E+05, 7.4821E+05, 7.4994E+05, 7.5166E+05,
     +  7.5339E+05, 7.5513E+05, 7.5686E+05, 7.5860E+05, 7.6034E+05,
     + 7.6209E+05, 7.6384E+05, 7.6559E+05, 7.6734E+05, 7.6910E+05,
     +  7.7085E+05, 7.7262E+05, 7.7438E+05, 7.7615E+05, 7.7792E+05,
     + 7.7969E+05, 7.8147E+05, 7.8324E+05, 7.8502E+05, 7.8681E+05,
     +  7.8860E+05, 7.9039E+05, 7.9218E+05, 7.9397E+05, 7.9577E+05,
     + 7.9757E+05, 7.9938E+05, 8.0119E+05, 8.0300E+05, 8.0481E+05,
     +  8.0662E+05, 8.0844E+05, 8.1026E+05, 8.1209E+05, 8.1392E+05,
     + 8.1575E+05, 8.1758E+05, 8.1941E+05, 8.2125E+05, 8.2309E+05,
     +  8.2494E+05, 8.2679E+05, 8.2864E+05, 8.3049E+05, 8.3235E+05,
     + 8.3420E+05, 8.3607E+05, 8.3793E+05, 8.3980E+05, 8.4167E+05,
     +  8.4354E+05, 8.4542E+05, 8.4730E+05, 8.4918E+05, 8.5107E+05,
     + 8.5295E+05, 8.5484E+05, 8.5674E+05, 8.5863E+05, 8.6053E+05,
     +  8.6244E+05, 8.6434E+05, 8.6625E+05, 8.6816E+05, 8.7008E+05,
     + 8.7199E+05, 8.7391E+05, 8.7584E+05, 8.7776E+05, 8.7969E+05,
     +  8.8162E+05, 8.8356E+05, 8.8550E+05, 8.8744E+05, 8.8938E+05,
     + 8.9133E+05, 8.9328E+05, 8.9523E+05, 8.9718E+05, 8.9914E+05,
     +  9.0110E+05, 9.0307E+05, 9.0504E+05, 9.0701E+05, 9.0898E+05,
     + 9.1096E+05, 9.1293E+05, 9.1492E+05, 9.1690E+05, 9.1889E+05,
     +  9.2088E+05, 9.2287E+05, 9.2487E+05, 9.2687E+05, 9.2887E+05,
     + 9.3088E+05, 9.3289E+05, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 5,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       637
      data (QofT( 6,J),J=1,2408)/ 2.6516E+01, 4.8285E+01, 7.0234E+01,
     +  9.2224E+01, 1.1423E+02, 1.3624E+02, 1.5826E+02, 1.8028E+02,
     + 2.0230E+02, 2.2433E+02, 2.4635E+02, 2.6838E+02, 2.9040E+02,
     +  3.1243E+02, 3.3446E+02, 3.5648E+02, 3.7851E+02, 4.0054E+02,
     + 4.2256E+02, 4.4459E+02, 4.8865E+02, 5.3271E+02, 5.7676E+02,
     +  6.2082E+02, 6.6488E+02, 7.0894E+02, 7.5300E+02, 7.9705E+02,
     + 8.4111E+02, 8.8517E+02, 9.2923E+02, 9.7329E+02, 1.0174E+03,
     +  1.0614E+03, 1.1055E+03, 1.1495E+03, 1.1936E+03, 1.2377E+03,
     + 1.2817E+03, 1.3258E+03, 1.3698E+03, 1.4139E+03, 1.4580E+03,
     +  1.5020E+03, 1.5461E+03, 1.5902E+03, 1.6342E+03, 1.6783E+03,
     + 1.7224E+03, 1.7664E+03, 1.8105E+03, 1.8546E+03, 1.8987E+03,
     +  1.9428E+03, 1.9869E+03, 2.0310E+03, 2.0751E+03, 2.1192E+03,
     + 2.1633E+03, 2.2075E+03, 2.2516E+03, 2.2958E+03, 2.3400E+03,
     +  2.3842E+03, 2.4284E+03, 2.4727E+03, 2.5170E+03, 2.5613E+03,
     + 2.6057E+03, 2.6500E+03, 2.6945E+03, 2.7389E+03, 2.7834E+03,
     +  2.8280E+03, 2.8726E+03, 2.9173E+03, 2.9620E+03, 3.0068E+03,
     + 3.0516E+03, 3.0965E+03, 3.1415E+03, 3.1866E+03, 3.2317E+03,
     +  3.2770E+03, 3.3223E+03, 3.3677E+03, 3.4132E+03, 3.4588E+03,
     + 3.5046E+03, 3.5504E+03, 3.5963E+03, 3.6424E+03, 3.6886E+03,
     +  3.7349E+03, 3.7814E+03, 3.8279E+03, 3.8747E+03, 3.9215E+03,
     + 3.9685E+03, 4.0157E+03, 4.0630E+03, 4.1105E+03, 4.1581E+03,
     +  4.2060E+03, 4.2539E+03, 4.3021E+03, 4.3504E+03, 4.3990E+03,
     + 4.4477E+03, 4.4966E+03, 4.5457E+03, 4.5950E+03, 4.6445E+03,
     +  4.6942E+03, 4.7442E+03, 4.7943E+03, 4.8447E+03, 4.8953E+03,
     + 4.9461E+03, 4.9971E+03, 5.0484E+03, 5.0999E+03, 5.1517E+03,
     +  5.2037E+03, 5.2559E+03, 5.3084E+03, 5.3612E+03, 5.4142E+03,
     + 5.4675E+03, 5.5210E+03, 5.5749E+03, 5.6290E+03, 5.6833E+03,
     +  5.7380E+03, 5.7929E+03, 5.8481E+03, 5.9036E+03, 5.9594E+03,
     + 6.0155E+03, 6.0719E+03, 6.1285E+03, 6.1855E+03, 6.2428E+03,
     +  6.3005E+03, 6.3584E+03, 6.4166E+03, 6.4752E+03, 6.5341E+03,
     + 6.5933E+03, 6.6528E+03, 6.7127E+03, 6.7729E+03, 6.8334E+03,
     +  6.8943E+03, 6.9555E+03, 7.0171E+03, 7.0790E+03, 7.1413E+03,
     + 7.2040E+03, 7.2669E+03, 7.3303E+03, 7.3940E+03, 7.4581E+03,
     +  7.5225E+03, 7.5874E+03, 7.6526E+03, 7.7181E+03, 7.7841E+03,
     + 7.8504E+03, 7.9171E+03, 7.9843E+03, 8.0518E+03, 8.1197E+03,
     +  8.1880E+03, 8.2566E+03, 8.3257E+03, 8.3952E+03, 8.4652E+03,
     + 8.5355E+03, 8.6062E+03, 8.6774E+03, 8.7489E+03, 8.8209E+03,
     +  8.8933E+03, 8.9662E+03, 9.0395E+03, 9.1132E+03, 9.1873E+03,
     + 9.2619E+03, 9.3369E+03, 9.4124E+03, 9.4883E+03, 9.5646E+03,
     +  9.6414E+03, 9.7187E+03, 9.7964E+03, 9.8746E+03, 9.9532E+03,
     + 1.0032E+04, 1.0112E+04, 1.0192E+04, 1.0272E+04, 1.0353E+04,
     +  1.0435E+04, 1.0517E+04, 1.0599E+04, 1.0682E+04, 1.0766E+04,
     + 1.0849E+04, 1.0934E+04, 1.1019E+04, 1.1104E+04, 1.1190E+04,
     +  1.1276E+04, 1.1363E+04, 1.1451E+04, 1.1539E+04, 1.1627E+04,
     + 1.1716E+04, 1.1806E+04, 1.1896E+04, 1.1986E+04, 1.2077E+04,
     +  1.2169E+04, 1.2261E+04, 1.2354E+04, 1.2447E+04, 1.2541E+04,
     + 1.2635E+04, 1.2730E+04, 1.2825E+04, 1.2921E+04, 1.3018E+04,
     +  1.3115E+04, 1.3212E+04, 1.3311E+04, 1.3410E+04, 1.3509E+04,
     + 1.3609E+04, 1.3709E+04, 1.3810E+04, 1.3912E+04, 1.4014E+04,
     +  1.4117E+04, 1.4221E+04, 1.4325E+04, 1.4429E+04, 1.4535E+04,
     + 1.4641E+04, 1.4747E+04, 1.4854E+04, 1.4962E+04, 1.5070E+04,
     +  1.5179E+04, 1.5289E+04, 1.5399E+04, 1.5510E+04, 1.5621E+04,
     + 1.5733E+04, 1.5846E+04, 1.5959E+04, 1.6073E+04, 1.6188E+04,
     +  1.6303E+04, 1.6419E+04, 1.6536E+04, 1.6653E+04, 1.6771E+04,
     + 1.6889E+04, 1.7009E+04, 1.7129E+04, 1.7249E+04, 1.7370E+04,
     +  1.7492E+04, 1.7615E+04, 1.7738E+04, 1.7862E+04, 1.7987E+04,
     + 1.8113E+04, 1.8239E+04, 1.8366E+04, 1.8493E+04, 1.8621E+04,
     +  1.8750E+04, 1.8880E+04, 1.9010E+04, 1.9141E+04, 1.9273E+04,
     + 1.9406E+04, 1.9539E+04, 1.9673E+04, 1.9808E+04, 1.9944E+04,
     +  2.0080E+04, 2.0217E+04, 2.0355E+04, 2.0493E+04, 2.0632E+04,
     + 2.0772E+04, 2.0913E+04, 2.1055E+04, 2.1197E+04, 2.1340E+04,
     +  2.1484E+04, 2.1629E+04, 2.1775E+04, 2.1921E+04, 2.2068E+04,
     + 2.2216E+04, 2.2364E+04, 2.2514E+04, 2.2664E+04, 2.2815E+04,
     +  2.2967E+04, 2.3120E+04, 2.3274E+04, 2.3428E+04, 2.3583E+04,
     + 2.3739E+04, 2.3896E+04, 2.4054E+04, 2.4212E+04, 2.4372E+04,
     +  2.4532E+04, 2.4693E+04, 2.4855E+04, 2.5018E+04, 2.5182E+04,
     + 2.5346E+04, 2.5512E+04, 2.5678E+04, 2.5845E+04, 2.6013E+04,
     +  2.6182E+04, 2.6352E+04, 2.6523E+04, 2.6695E+04, 2.6867E+04,
     + 2.7041E+04, 2.7215E+04, 2.7390E+04, 2.7567E+04, 2.7744E+04,
     +  2.7922E+04, 2.8101E+04, 2.8281E+04, 2.8462E+04, 2.8644E+04,
     + 2.8826E+04, 2.9010E+04, 2.9195E+04, 2.9380E+04, 2.9567E+04,
     +  2.9754E+04, 2.9943E+04, 3.0132E+04, 3.0323E+04, 3.0514E+04,
     + 3.0707E+04, 3.0900E+04, 3.1095E+04, 3.1290E+04, 3.1487E+04,
     +  3.1684E+04, 3.1882E+04, 3.2082E+04, 3.2282E+04, 3.2484E+04,
     + 3.2686E+04, 3.2890E+04, 3.3095E+04, 3.3300E+04, 3.3507E+04,
     +  3.3715E+04, 3.3923E+04, 3.4133E+04, 3.4344E+04, 3.4556E+04,
     + 3.4769E+04, 3.4983E+04, 3.5198E+04, 3.5415E+04, 3.5632E+04,
     +  3.5850E+04, 3.6070E+04, 3.6291E+04, 3.6512E+04, 3.6735E+04,
     + 3.6959E+04, 3.7184E+04, 3.7410E+04, 3.7638E+04, 3.7866E+04,
     +  3.8096E+04, 3.8326E+04, 3.8558E+04, 3.8791E+04, 3.9025E+04,
     + 3.9261E+04, 3.9497E+04, 3.9735E+04, 3.9974E+04, 4.0214E+04,
     +  4.0455E+04, 4.0697E+04, 4.0941E+04, 4.1185E+04, 4.1431E+04,
     + 4.1678E+04, 4.1927E+04, 4.2176E+04, 4.2427E+04, 4.2679E+04,
     +  4.2932E+04, 4.3187E+04, 4.3442E+04, 4.3699E+04, 4.3957E+04,
     + 4.4217E+04, 4.4477E+04, 4.4739E+04, 4.5002E+04, 4.5267E+04,
     +  4.5533E+04, 4.5800E+04, 4.6068E+04, 4.6337E+04, 4.6608E+04,
     + 4.6880E+04, 4.7154E+04, 4.7429E+04, 4.7705E+04, 4.7982E+04,
     +  4.8261E+04, 4.8541E+04, 4.8822E+04, 4.9105E+04, 4.9389E+04,
     + 4.9674E+04, 4.9961E+04, 5.0249E+04, 5.0539E+04, 5.0829E+04,
     +  5.1122E+04, 5.1415E+04, 5.1710E+04, 5.2007E+04, 5.2304E+04,
     + 5.2603E+04, 5.2904E+04, 5.3206E+04, 5.3509E+04, 5.3814E+04,
     +  5.4120E+04, 5.4428E+04, 5.4737E+04, 5.5047E+04, 5.5359E+04,
     + 5.5673E+04, 5.5987E+04, 5.6304E+04, 5.6622E+04, 5.6941E+04,
     +  5.7261E+04, 5.7584E+04, 5.7907E+04, 5.8233E+04, 5.8559E+04,
     + 5.8887E+04, 5.9217E+04, 5.9548E+04, 5.9881E+04, 6.0215E+04,
     +  6.0551E+04, 6.0888E+04, 6.1227E+04, 6.1567E+04, 6.1909E+04,
     + 6.2253E+04, 6.2598E+04, 6.2944E+04, 6.3293E+04, 6.3642E+04,
     +  6.3994E+04, 6.4347E+04, 6.4701E+04, 6.5057E+04, 6.5415E+04,
     + 6.5774E+04, 6.6135E+04, 6.6498E+04, 6.6862E+04, 6.7228E+04,
     +  6.7596E+04, 6.7965E+04, 6.8336E+04, 6.8708E+04, 6.9082E+04,
     + 6.9458E+04, 6.9835E+04, 7.0215E+04, 7.0595E+04, 7.0978E+04,
     +  7.1362E+04, 7.1748E+04, 7.2136E+04, 7.2525E+04, 7.2916E+04,
     + 7.3309E+04, 7.3704E+04, 7.4100E+04, 7.4498E+04, 7.4898E+04,
     +  7.5299E+04, 7.5703E+04, 7.6108E+04, 7.6515E+04, 7.6923E+04,
     + 7.7334E+04, 7.7746E+04, 7.8160E+04, 7.8576E+04, 7.8993E+04,
     +  7.9413E+04, 7.9834E+04, 8.0257E+04, 8.0682E+04, 8.1109E+04,
     + 8.1538E+04, 8.1968E+04, 8.2401E+04, 8.2835E+04, 8.3271E+04,
     +  8.3709E+04, 8.4149E+04, 8.4591E+04, 8.5034E+04, 8.5480E+04,
     + 8.5928E+04, 8.6377E+04, 8.6828E+04, 8.7282E+04, 8.7737E+04,
     +  8.8194E+04, 8.8653E+04, 8.9114E+04, 8.9577E+04, 9.0042E+04,
     + 9.0509E+04, 9.0978E+04, 9.1449E+04, 9.1922E+04, 9.2397E+04,
     +  9.2874E+04, 9.3353E+04, 9.3834E+04, 9.4317E+04, 9.4802E+04,
     + 9.5289E+04, 9.5779E+04, 9.6270E+04, 9.6763E+04, 9.7259E+04,
     +  9.7756E+04, 9.8256E+04, 9.8757E+04, 9.9261E+04, 9.9767E+04,
     + 1.0027E+05, 1.0078E+05, 1.0130E+05, 1.0181E+05, 1.0233E+05,
     +  1.0285E+05, 1.0337E+05, 1.0389E+05, 1.0442E+05, 1.0494E+05,
     + 1.0547E+05, 1.0600E+05, 1.0654E+05, 1.0707E+05, 1.0761E+05,
     +  1.0815E+05, 1.0869E+05, 1.0924E+05, 1.0979E+05, 1.1034E+05,
     + 1.1089E+05, 1.1144E+05, 1.1200E+05, 1.1256E+05, 1.1312E+05,
     +  1.1368E+05, 1.1424E+05, 1.1481E+05, 1.1538E+05, 1.1595E+05,
     + 1.1653E+05, 1.1711E+05, 1.1768E+05, 1.1827E+05, 1.1885E+05,
     +  1.1944E+05, 1.2003E+05, 1.2062E+05, 1.2121E+05, 1.2181E+05,
     + 1.2240E+05, 1.2300E+05, 1.2361E+05, 1.2421E+05, 1.2482E+05,
     +  1.2543E+05, 1.2604E+05, 1.2666E+05, 1.2728E+05, 1.2790E+05,
     + 1.2852E+05, 1.2914E+05, 1.2977E+05, 1.3040E+05, 1.3103E+05,
     +  1.3167E+05, 1.3231E+05, 1.3295E+05, 1.3359E+05, 1.3423E+05,
     + 1.3488E+05, 1.3553E+05, 1.3618E+05, 1.3684E+05, 1.3750E+05,
     +  1.3816E+05, 1.3882E+05, 1.3949E+05, 1.4015E+05, 1.4083E+05,
     + 1.4150E+05, 1.4217E+05, 1.4285E+05, 1.4354E+05, 1.4422E+05,
     +  1.4491E+05, 1.4560E+05, 1.4629E+05, 1.4698E+05, 1.4768E+05,
     + 1.4838E+05, 1.4908E+05, 1.4979E+05, 1.5050E+05, 1.5121E+05,
     +  1.5192E+05, 1.5264E+05, 1.5336E+05, 1.5408E+05, 1.5481E+05,
     + 1.5553E+05, 1.5626E+05, 1.5700E+05, 1.5773E+05, 1.5847E+05,
     +  1.5921E+05, 1.5996E+05, 1.6071E+05, 1.6146E+05, 1.6221E+05,
     + 1.6297E+05, 1.6372E+05, 1.6449E+05, 1.6525E+05, 1.6602E+05,
     +  1.6679E+05, 1.6756E+05, 1.6834E+05, 1.6912E+05, 1.6990E+05,
     + 1.7069E+05, 1.7147E+05, 1.7226E+05, 1.7306E+05, 1.7386E+05,
     +  1.7466E+05, 1.7546E+05, 1.7627E+05, 1.7707E+05, 1.7789E+05,
     + 1.7870E+05, 1.7952E+05, 1.8034E+05, 1.8117E+05, 1.8199E+05,
     +  1.8282E+05, 1.8366E+05, 1.8449E+05, 1.8533E+05, 1.8618E+05,
     + 1.8702E+05, 1.8787E+05, 1.8873E+05, 1.8958E+05, 1.9044E+05,
     +  1.9130E+05, 1.9217E+05, 1.9303E+05, 1.9391E+05, 1.9478E+05,
     + 1.9566E+05, 1.9654E+05, 1.9742E+05, 1.9831E+05, 1.9920E+05,
     +  2.0010E+05, 2.0099E+05, 2.0190E+05, 2.0280E+05, 2.0371E+05,
     + 2.0462E+05, 2.0553E+05, 2.0645E+05, 2.0737E+05, 2.0829E+05,
     +  2.0922E+05, 2.1015E+05, 2.1108E+05, 2.1202E+05, 2.1296E+05,
     + 2.1391E+05, 2.1485E+05, 2.1581E+05, 2.1676E+05, 2.1772E+05,
     +  2.1868E+05, 2.1964E+05, 2.2061E+05, 2.2158E+05, 2.2256E+05,
     + 2.2354E+05, 2.2452E+05, 2.2551E+05, 2.2650E+05, 2.2749E+05,
     +  2.2849E+05, 2.2949E+05, 2.3049E+05, 2.3150E+05, 2.3251E+05,
     + 2.3352E+05, 2.3454E+05, 2.3556E+05, 2.3659E+05, 2.3761E+05,
     +  2.3865E+05, 2.3968E+05, 2.4072E+05, 2.4177E+05, 2.4281E+05,
     + 2.4386E+05, 2.4492E+05, 2.4598E+05, 2.4704E+05, 2.4810E+05,
     +  2.4917E+05, 2.5025E+05, 2.5132E+05, 2.5240E+05, 2.5349E+05,
     + 2.5458E+05, 2.5567E+05, 2.5676E+05, 2.5786E+05, 2.5897E+05,
     +  2.6007E+05, 2.6119E+05, 2.6230E+05, 2.6342E+05, 2.6454E+05,
     + 2.6567E+05, 2.6680E+05, 2.6793E+05, 2.6907E+05, 2.7022E+05,
     +  2.7136E+05, 2.7251E+05, 2.7367E+05, 2.7482E+05, 2.7599E+05,
     + 2.7715E+05, 2.7832E+05, 2.7950E+05, 2.8068E+05, 2.8186E+05,
     +  2.8304E+05, 2.8424E+05, 2.8543E+05, 2.8663E+05, 2.8783E+05,
     + 2.8904E+05, 2.9025E+05, 2.9146E+05, 2.9268E+05, 2.9391E+05,
     +  2.9513E+05, 2.9637E+05, 2.9760E+05, 2.9884E+05, 3.0009E+05,
     + 3.0133E+05, 3.0259E+05, 3.0384E+05, 3.0510E+05, 3.0637E+05,
     +  3.0764E+05, 3.0891E+05, 3.1019E+05, 3.1147E+05, 3.1276E+05,
     + 3.1405E+05, 3.1535E+05, 3.1665E+05, 3.1795E+05, 3.1926E+05,
     +  3.2057E+05, 3.2189E+05, 3.2321E+05, 3.2454E+05, 3.2587E+05,
     + 3.2720E+05, 3.2854E+05, 3.2989E+05, 3.3123E+05, 3.3259E+05,
     +  3.3394E+05, 3.3531E+05, 3.3667E+05, 3.3804E+05, 3.3942E+05,
     + 3.4080E+05, 3.4218E+05, 3.4357E+05, 3.4496E+05, 3.4636E+05,
     +  3.4777E+05, 3.4917E+05, 3.5058E+05, 3.5200E+05, 3.5342E+05,
     + 3.5485E+05, 3.5628E+05, 3.5771E+05, 3.5915E+05, 3.6060E+05,
     +  3.6205E+05, 3.6350E+05, 3.6496E+05, 3.6642E+05, 3.6789E+05,
     + 3.6936E+05, 3.7084E+05, 3.7233E+05, 3.7381E+05, 3.7530E+05,
     +  3.7680E+05, 3.7830E+05, 3.7981E+05, 3.8132E+05, 3.8284E+05,
     + 3.8436E+05, 3.8589E+05, 3.8742E+05, 3.8895E+05, 3.9049E+05,
     +  3.9204E+05, 3.9359E+05, 3.9515E+05, 3.9671E+05, 3.9827E+05,
     + 3.9984E+05, 4.0142E+05, 4.0300E+05, 4.0459E+05, 4.0618E+05,
     +  4.0777E+05, 4.0938E+05, 4.1098E+05, 4.1259E+05, 4.1421E+05,
     + 4.1583E+05, 4.1746E+05, 4.1909E+05, 4.2073E+05, 4.2237E+05,
     +  4.2402E+05, 4.2567E+05, 4.2733E+05, 4.2899E+05, 4.3066E+05,
     + 4.3233E+05, 4.3401E+05, 4.3570E+05, 4.3738E+05, 4.3908E+05,
     +  4.4078E+05, 4.4248E+05, 4.4420E+05, 4.4591E+05, 4.4763E+05,
     + 4.4936E+05, 4.5109E+05, 4.5283E+05, 4.5457E+05, 4.5632E+05,
     +  4.5807E+05, 4.5983E+05, 4.6160E+05, 4.6337E+05, 4.6514E+05,
     + 4.6692E+05, 4.6871E+05, 4.7050E+05, 4.7230E+05, 4.7410E+05,
     +  4.7591E+05, 4.7773E+05, 4.7955E+05, 4.8137E+05, 4.8320E+05,
     + 4.8504E+05, 4.8688E+05, 4.8873E+05, 4.9059E+05, 4.9245E+05,
     +  4.9431E+05, 4.9618E+05, 4.9806E+05, 4.9994E+05, 5.0183E+05,
     + 5.0372E+05, 5.0562E+05, 5.0753E+05, 5.0944E+05, 5.1135E+05,
     +  5.1328E+05, 5.1521E+05, 5.1714E+05, 5.1908E+05, 5.2103E+05,
     + 5.2298E+05, 5.2494E+05, 5.2690E+05, 5.2887E+05, 5.3085E+05,
     +  5.3283E+05, 5.3482E+05, 5.3681E+05, 5.3881E+05, 5.4081E+05,
     + 5.4283E+05, 5.4484E+05, 5.4687E+05, 5.4890E+05, 5.5093E+05,
     +  5.5297E+05, 5.5502E+05, 5.5708E+05, 5.5914E+05, 5.6120E+05,
     + 5.6327E+05, 5.6535E+05, 5.6744E+05, 5.6953E+05, 5.7163E+05,
     +  5.7373E+05, 5.7584E+05, 5.7796E+05, 5.8008E+05, 5.8221E+05,
     + 5.8434E+05, 5.8648E+05, 5.8863E+05, 5.9078E+05, 5.9294E+05,
     +  5.9511E+05, 5.9728E+05, 5.9946E+05, 6.0165E+05, 6.0384E+05,
     + 6.0604E+05, 6.0824E+05, 6.1045E+05, 6.1267E+05, 6.1489E+05,
     +  6.1712E+05, 6.1936E+05, 6.2161E+05, 6.2386E+05, 6.2611E+05,
     + 6.2838E+05, 6.3064E+05, 6.3292E+05, 6.3520E+05, 6.3749E+05,
     +  6.3979E+05, 6.4209E+05, 6.4440E+05, 6.4672E+05, 6.4904E+05,
     + 6.5137E+05, 6.5371E+05, 6.5605E+05, 6.5840E+05, 6.6076E+05,
     +  6.6312E+05, 6.6549E+05, 6.6787E+05, 6.7025E+05, 6.7264E+05,
     + 6.7504E+05, 6.7744E+05, 6.7985E+05, 6.8227E+05, 6.8470E+05,
     +  6.8713E+05, 6.8957E+05, 6.9201E+05, 6.9446E+05, 6.9692E+05,
     + 6.9939E+05, 7.0186E+05, 7.0434E+05, 7.0683E+05, 7.0933E+05,
     +  7.1183E+05, 7.1434E+05, 7.1685E+05, 7.1937E+05, 7.2190E+05,
     + 7.2444E+05, 7.2699E+05, 7.2954E+05, 7.3209E+05, 7.3466E+05,
     +  7.3723E+05, 7.3981E+05, 7.4240E+05, 7.4499E+05, 7.4760E+05,
     + 7.5021E+05, 7.5282E+05, 7.5545E+05, 7.5808E+05, 7.6071E+05,
     +  7.6336E+05, 7.6601E+05, 7.6867E+05, 7.7134E+05, 7.7402E+05,
     + 7.7670E+05, 7.7939E+05, 7.8208E+05, 7.8479E+05, 7.8750E+05,
     +  7.9022E+05, 7.9295E+05, 7.9568E+05, 7.9843E+05, 8.0118E+05,
     + 8.0393E+05, 8.0670E+05, 8.0947E+05, 8.1225E+05, 8.1504E+05,
     +  8.1783E+05, 8.2064E+05, 8.2345E+05, 8.2626E+05, 8.2909E+05,
     + 8.3192E+05, 8.3477E+05, 8.3761E+05, 8.4047E+05, 8.4334E+05,
     +  8.4621E+05, 8.4909E+05, 8.5198E+05, 8.5487E+05, 8.5778E+05,
     + 8.6069E+05, 8.6361E+05, 8.6653E+05, 8.6947E+05, 8.7241E+05,
     +  8.7536E+05, 8.7832E+05, 8.8129E+05, 8.8426E+05, 8.8725E+05,
     + 8.9024E+05, 8.9324E+05, 8.9624E+05, 8.9926E+05, 9.0228E+05,
     +  9.0531E+05, 9.0835E+05, 9.1140E+05, 9.1445E+05, 9.1752E+05,
     + 9.2059E+05, 9.2367E+05, 9.2676E+05, 9.2985E+05, 9.3296E+05,
     +  9.3607E+05, 9.3919E+05, 9.4232E+05, 9.4546E+05, 9.4860E+05,
     + 9.5176E+05, 9.5492E+05, 9.5809E+05, 9.6127E+05, 9.6446E+05,
     +  9.6765E+05, 9.7086E+05, 9.7407E+05, 9.7729E+05, 9.8052E+05,
     + 9.8376E+05, 9.8700E+05, 9.9026E+05, 9.9352E+05, 9.9679E+05,
     +  1.0001E+06, 1.0034E+06, 1.0067E+06, 1.0100E+06, 1.0133E+06,
     + 1.0166E+06, 1.0199E+06, 1.0233E+06, 1.0266E+06, 1.0300E+06,
     +  1.0333E+06, 1.0367E+06, 1.0401E+06, 1.0435E+06, 1.0469E+06,
     + 1.0503E+06, 1.0537E+06, 1.0572E+06, 1.0606E+06, 1.0640E+06,
     +  1.0675E+06, 1.0710E+06, 1.0744E+06, 1.0779E+06, 1.0814E+06,
     + 1.0849E+06, 1.0884E+06, 1.0919E+06, 1.0954E+06, 1.0990E+06,
     +  1.1025E+06, 1.1061E+06, 1.1096E+06, 1.1132E+06, 1.1168E+06,
     + 1.1204E+06, 1.1240E+06, 1.1276E+06, 1.1312E+06, 1.1348E+06,
     +  1.1385E+06, 1.1421E+06, 1.1458E+06, 1.1494E+06, 1.1531E+06,
     + 1.1568E+06, 1.1605E+06, 1.1642E+06, 1.1679E+06, 1.1716E+06,
     +  1.1753E+06, 1.1790E+06, 1.1828E+06, 1.1865E+06, 1.1903E+06,
     + 1.1941E+06, 1.1979E+06, 1.2016E+06, 1.2054E+06, 1.2093E+06,
     +  1.2131E+06, 1.2169E+06, 1.2207E+06, 1.2246E+06, 1.2285E+06,
     + 1.2323E+06, 1.2362E+06, 1.2401E+06, 1.2440E+06, 1.2479E+06,
     +  1.2518E+06, 1.2557E+06, 1.2597E+06, 1.2636E+06, 1.2676E+06,
     + 1.2715E+06, 1.2755E+06, 1.2795E+06, 1.2835E+06, 1.2875E+06,
     +  1.2915E+06, 1.2955E+06, 1.2995E+06, 1.3036E+06, 1.3076E+06,
     + 1.3117E+06, 1.3158E+06, 1.3199E+06, 1.3239E+06, 1.3280E+06,
     +  1.3322E+06, 1.3363E+06, 1.3404E+06, 1.3446E+06, 1.3487E+06,
     + 1.3529E+06, 1.3570E+06, 1.3612E+06, 1.3654E+06, 1.3696E+06,
     +  1.3738E+06, 1.3780E+06, 1.3823E+06, 1.3865E+06, 1.3908E+06,
     + 1.3950E+06, 1.3993E+06, 1.4036E+06, 1.4079E+06, 1.4122E+06,
     +  1.4165E+06, 1.4208E+06, 1.4251E+06, 1.4295E+06, 1.4338E+06,
     + 1.4382E+06, 1.4426E+06, 1.4470E+06, 1.4514E+06, 1.4558E+06,
     +  1.4602E+06, 1.4646E+06, 1.4690E+06, 1.4735E+06, 1.4779E+06,
     + 1.4824E+06, 1.4869E+06, 1.4914E+06, 1.4959E+06, 1.5004E+06,
     +  1.5049E+06, 1.5094E+06, 1.5140E+06, 1.5185E+06, 1.5231E+06,
     + 1.5277E+06, 1.5323E+06, 1.5369E+06, 1.5415E+06, 1.5461E+06,
     +  1.5507E+06, 1.5553E+06, 1.5600E+06, 1.5647E+06, 1.5693E+06,
     + 1.5740E+06, 1.5787E+06, 1.5834E+06, 1.5881E+06, 1.5928E+06,
     +  1.5976E+06, 1.6023E+06, 1.6071E+06, 1.6118E+06, 1.6166E+06,
     + 1.6214E+06, 1.6262E+06, 1.6310E+06, 1.6358E+06, 1.6407E+06,
     +  1.6455E+06, 1.6504E+06, 1.6552E+06, 1.6601E+06, 1.6650E+06,
     + 1.6699E+06, 1.6748E+06, 1.6797E+06, 1.6847E+06, 1.6896E+06,
     +  1.6946E+06, 1.6995E+06, 1.7045E+06, 1.7095E+06, 1.7145E+06,
     + 1.7195E+06, 1.7245E+06, 1.7296E+06, 1.7346E+06, 1.7397E+06,
     +  1.7448E+06, 1.7498E+06, 1.7549E+06, 1.7600E+06, 1.7651E+06,
     + 1.7703E+06, 1.7754E+06, 1.7805E+06, 1.7857E+06, 1.7909E+06,
     +  1.7961E+06, 1.8013E+06, 1.8065E+06, 1.8117E+06, 1.8169E+06,
     + 1.8221E+06, 1.8274E+06, 1.8327E+06, 1.8379E+06, 1.8432E+06,
     +  1.8485E+06, 1.8538E+06, 1.8592E+06, 1.8645E+06, 1.8698E+06,
     + 1.8752E+06, 1.8806E+06, 1.8859E+06, 1.8913E+06, 1.8967E+06,
     +  1.9021E+06, 1.9076E+06, 1.9130E+06, 1.9185E+06, 1.9239E+06,
     + 1.9294E+06, 1.9349E+06, 1.9404E+06, 1.9459E+06, 1.9514E+06,
     +  1.9569E+06, 1.9625E+06, 1.9681E+06, 1.9736E+06, 1.9792E+06,
     + 1.9848E+06, 1.9904E+06, 1.9960E+06, 2.0017E+06, 2.0073E+06,
     +  2.0130E+06, 2.0186E+06, 2.0243E+06, 2.0300E+06, 2.0357E+06,
     + 2.0414E+06, 2.0471E+06, 2.0529E+06, 2.0586E+06, 2.0644E+06,
     +  2.0702E+06, 2.0760E+06, 2.0818E+06, 2.0876E+06, 2.0934E+06,
     + 2.0992E+06, 2.1051E+06, 2.1109E+06, 2.1168E+06, 2.1227E+06,
     +  2.1286E+06, 2.1345E+06, 2.1404E+06, 2.1464E+06, 2.1523E+06,
     + 2.1583E+06, 2.1643E+06, 2.1702E+06, 2.1762E+06, 2.1823E+06,
     +  2.1883E+06, 2.1943E+06, 2.2004E+06, 2.2064E+06, 2.2125E+06,
     + 2.2186E+06, 2.2247E+06, 2.2308E+06, 2.2369E+06, 2.2431E+06,
     +  2.2492E+06, 2.2554E+06, 2.2616E+06, 2.2678E+06, 2.2740E+06,
     + 2.2802E+06, 2.2864E+06, 2.2926E+06, 2.2989E+06, 2.3052E+06,
     +  2.3114E+06, 2.3177E+06, 2.3240E+06, 2.3303E+06, 2.3367E+06,
     + 2.3430E+06, 2.3494E+06, 2.3557E+06, 2.3621E+06, 2.3685E+06,
     +  2.3749E+06, 2.3814E+06, 2.3878E+06, 2.3942E+06, 2.4007E+06,
     + 2.4072E+06, 2.4137E+06, 2.4202E+06, 2.4267E+06, 2.4332E+06,
     +  2.4397E+06, 2.4463E+06, 2.4529E+06, 2.4594E+06, 2.4660E+06,
     + 2.4726E+06, 2.4792E+06, 2.4859E+06, 2.4925E+06, 2.4992E+06,
     +  2.5058E+06, 2.5125E+06, 2.5192E+06, 2.5259E+06, 2.5327E+06,
     + 2.5394E+06, 2.5462E+06, 2.5529E+06, 2.5597E+06, 2.5665E+06,
     +  2.5733E+06, 2.5801E+06, 2.5869E+06, 2.5938E+06, 2.6007E+06,
     + 2.6075E+06, 2.6144E+06, 2.6213E+06, 2.6282E+06, 2.6351E+06,
     +  2.6421E+06, 2.6490E+06, 2.6560E+06, 2.6630E+06, 2.6700E+06,
     + 2.6770E+06, 2.6840E+06, 2.6911E+06, 2.6981E+06, 2.7052E+06,
     +  2.7122E+06, 2.7193E+06, 2.7264E+06, 2.7336E+06, 2.7407E+06,
     + 2.7478E+06, 2.7550E+06, 2.7622E+06, 2.7694E+06, 2.7766E+06,
     +  2.7838E+06, 2.7910E+06, 2.7982E+06, 2.8055E+06, 2.8128E+06,
     + 2.8201E+06, 2.8273E+06, 2.8347E+06, 2.8420E+06, 2.8493E+06,
     +  2.8567E+06, 2.8641E+06, 2.8714E+06, 2.8788E+06, 2.8862E+06,
     + 2.8937E+06, 2.9011E+06, 2.9086E+06, 2.9160E+06, 2.9235E+06,
     +  2.9310E+06, 2.9385E+06, 2.9460E+06, 2.9536E+06, 2.9611E+06,
     + 2.9687E+06, 2.9763E+06, 2.9839E+06, 2.9915E+06, 2.9991E+06,
     +  3.0067E+06, 3.0144E+06, 3.0220E+06, 3.0297E+06, 3.0374E+06,
     + 3.0451E+06, 3.0529E+06, 3.0606E+06, 3.0683E+06, 3.0761E+06,
     +  3.0839E+06, 3.0917E+06, 3.0995E+06, 3.1073E+06, 3.1152E+06,
     + 3.1230E+06, 3.1309E+06, 3.1388E+06, 3.1467E+06, 3.1546E+06,
     +  3.1625E+06, 3.1704E+06, 3.1784E+06, 3.1864E+06, 3.1943E+06,
     + 3.2023E+06, 3.2104E+06, 3.2184E+06, 3.2264E+06, 3.2345E+06,
     +  3.2426E+06, 3.2506E+06, 3.2587E+06, 3.2669E+06, 3.2750E+06,
     + 3.2831E+06, 3.2913E+06, 3.2995E+06, 3.3077E+06, 3.3159E+06,
     +  3.3241E+06, 3.3323E+06, 3.3406E+06, 3.3488E+06, 3.3571E+06,
     + 3.3654E+06, 3.3737E+06, 3.3820E+06, 3.3904E+06, 3.3987E+06,
     +  3.4071E+06, 3.4155E+06, 3.4239E+06, 3.4323E+06, 3.4407E+06,
     + 3.4492E+06, 3.4576E+06, 3.4661E+06, 3.4746E+06, 3.4831E+06,
     +  3.4916E+06, 3.5002E+06, 3.5087E+06, 3.5173E+06, 3.5258E+06,
     + 3.5344E+06, 3.5430E+06, 3.5517E+06, 3.5603E+06, 3.5690E+06,
     +  3.5776E+06, 3.5863E+06, 3.5950E+06, 3.6037E+06, 3.6125E+06,
     + 3.6212E+06, 3.6300E+06, 3.6388E+06, 3.6475E+06, 3.6564E+06,
     +  3.6652E+06, 3.6740E+06, 3.6829E+06, 3.6917E+06, 3.7006E+06,
     + 3.7095E+06, 3.7184E+06, 3.7274E+06, 3.7363E+06, 3.7453E+06,
     +  3.7542E+06, 3.7632E+06, 3.7722E+06, 3.7813E+06, 3.7903E+06,
     + 3.7994E+06, 3.8084E+06, 3.8175E+06, 3.8266E+06, 3.8357E+06,
     +  3.8449E+06, 3.8540E+06, 3.8632E+06, 3.8724E+06, 3.8815E+06,
     + 3.8908E+06, 3.9000E+06, 3.9092E+06, 3.9185E+06, 3.9278E+06,
     +  3.9370E+06, 3.9463E+06, 3.9557E+06, 3.9650E+06, 3.9744E+06,
     + 3.9837E+06, 3.9931E+06, 4.0025E+06, 4.0119E+06, 4.0213E+06,
     +  4.0308E+06, 4.0403E+06, 4.0497E+06, 4.0592E+06, 4.0687E+06,
     + 4.0783E+06, 4.0878E+06, 4.0974E+06, 4.1069E+06, 4.1165E+06,
     +  4.1261E+06, 4.1357E+06, 4.1454E+06, 4.1550E+06, 4.1647E+06,
     + 4.1744E+06, 4.1841E+06, 4.1938E+06, 4.2035E+06, 4.2133E+06,
     +  4.2230E+06, 4.2328E+06, 4.2426E+06, 4.2524E+06, 4.2623E+06,
     + 4.2721E+06, 4.2820E+06, 4.2918E+06, 4.3017E+06, 4.3116E+06,
     +  4.3216E+06, 4.3315E+06, 4.3415E+06, 4.3514E+06, 4.3614E+06,
     + 4.3714E+06, 4.3814E+06, 4.3915E+06, 4.4015E+06, 4.4116E+06,
     +  4.4217E+06, 4.4318E+06, 4.4419E+06, 4.4521E+06, 4.4622E+06,
     + 4.4724E+06, 4.4826E+06, 4.4928E+06, 4.5030E+06, 4.5132E+06,
     +  4.5235E+06, 4.5337E+06, 4.5440E+06, 4.5543E+06, 4.5646E+06,
     + 4.5750E+06, 4.5853E+06, 4.5957E+06, 4.6061E+06, 4.6164E+06,
     +  4.6269E+06, 4.6373E+06, 4.6477E+06, 4.6582E+06, 4.6687E+06,
     + 4.6792E+06, 4.6897E+06, 4.7002E+06, 4.7108E+06, 4.7213E+06,
     +  4.7319E+06, 4.7425E+06, 4.7531E+06, 4.7638E+06, 4.7744E+06,
     + 4.7851E+06, 4.7957E+06, 4.8064E+06, 4.8171E+06, 4.8279E+06,
     +  4.8386E+06, 4.8494E+06, 4.8602E+06, 4.8710E+06, 4.8818E+06,
     + 4.8926E+06, 4.9034E+06, 4.9143E+06, 4.9252E+06, 4.9361E+06,
     +  4.9470E+06, 4.9579E+06, 4.9689E+06, 4.9798E+06, 4.9908E+06,
     + 5.0018E+06, 5.0128E+06, 5.0239E+06, 5.0349E+06, 5.0460E+06,
     +  5.0570E+06, 5.0681E+06, 5.0793E+06, 5.0904E+06, 5.1015E+06,
     + 5.1127E+06, 5.1239E+06, 5.1351E+06, 5.1463E+06, 5.1575E+06,
     +  5.1688E+06, 5.1800E+06, 5.1913E+06, 5.2026E+06, 5.2140E+06,
     + 5.2253E+06, 5.2366E+06, 5.2480E+06, 5.2594E+06, 5.2708E+06,
     +  5.2822E+06, 5.2937E+06, 5.3051E+06, 5.3166E+06, 5.3281E+06,
     + 5.3396E+06, 5.3511E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 6,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       828
      data (QofT( 7,J),J=1,2408)/ 1.2509E+00, 2.1807E+00, 3.1786E+00,
     +  4.1792E+00, 5.1803E+00, 6.1818E+00, 7.1835E+00, 8.1853E+00,
     + 9.1871E+00, 1.0189E+01, 1.1191E+01, 1.2193E+01, 1.3195E+01,
     +  1.4197E+01, 1.5199E+01, 1.6201E+01, 1.7203E+01, 1.8205E+01,
     + 1.9207E+01, 2.0209E+01, 2.2213E+01, 2.4217E+01, 2.6221E+01,
     +  2.8226E+01, 3.0230E+01, 3.2234E+01, 3.4238E+01, 3.6243E+01,
     + 3.8247E+01, 4.0251E+01, 4.2255E+01, 4.4260E+01, 4.6264E+01,
     +  4.8268E+01, 5.0273E+01, 5.2277E+01, 5.4281E+01, 5.6286E+01,
     + 5.8290E+01, 6.0294E+01, 6.2299E+01, 6.4303E+01, 6.6308E+01,
     +  6.8312E+01, 7.0317E+01, 7.2321E+01, 7.4326E+01, 7.6330E+01,
     + 7.8335E+01, 8.0340E+01, 8.2345E+01, 8.4350E+01, 8.6355E+01,
     +  8.8360E+01, 9.0366E+01, 9.2372E+01, 9.4378E+01, 9.6385E+01,
     + 9.8392E+01, 1.0040E+02, 1.0241E+02, 1.0442E+02, 1.0643E+02,
     +  1.0844E+02, 1.1045E+02, 1.1246E+02, 1.1447E+02, 1.1649E+02,
     + 1.1850E+02, 1.2052E+02, 1.2254E+02, 1.2456E+02, 1.2658E+02,
     +  1.2861E+02, 1.3063E+02, 1.3266E+02, 1.3469E+02, 1.3673E+02,
     + 1.3876E+02, 1.4080E+02, 1.4285E+02, 1.4489E+02, 1.4694E+02,
     +  1.4899E+02, 1.5105E+02, 1.5311E+02, 1.5518E+02, 1.5725E+02,
     + 1.5932E+02, 1.6140E+02, 1.6348E+02, 1.6557E+02, 1.6766E+02,
     +  1.6976E+02, 1.7186E+02, 1.7397E+02, 1.7609E+02, 1.7821E+02,
     + 1.8034E+02, 1.8248E+02, 1.8462E+02, 1.8677E+02, 1.8892E+02,
     +  1.9109E+02, 1.9326E+02, 1.9544E+02, 1.9762E+02, 1.9982E+02,
     + 2.0202E+02, 2.0423E+02, 2.0645E+02, 2.0868E+02, 2.1091E+02,
     +  2.1316E+02, 2.1541E+02, 2.1768E+02, 2.1995E+02, 2.2224E+02,
     + 2.2453E+02, 2.2683E+02, 2.2915E+02, 2.3147E+02, 2.3381E+02,
     +  2.3615E+02, 2.3851E+02, 2.4088E+02, 2.4326E+02, 2.4565E+02,
     + 2.4805E+02, 2.5047E+02, 2.5289E+02, 2.5533E+02, 2.5778E+02,
     +  2.6024E+02, 2.6272E+02, 2.6520E+02, 2.6771E+02, 2.7022E+02,
     + 2.7275E+02, 2.7528E+02, 2.7784E+02, 2.8040E+02, 2.8298E+02,
     +  2.8558E+02, 2.8819E+02, 2.9081E+02, 2.9345E+02, 2.9610E+02,
     + 2.9876E+02, 3.0144E+02, 3.0414E+02, 3.0685E+02, 3.0957E+02,
     +  3.1231E+02, 3.1506E+02, 3.1784E+02, 3.2062E+02, 3.2342E+02,
     + 3.2624E+02, 3.2908E+02, 3.3193E+02, 3.3479E+02, 3.3767E+02,
     +  3.4057E+02, 3.4349E+02, 3.4642E+02, 3.4937E+02, 3.5234E+02,
     + 3.5532E+02, 3.5832E+02, 3.6134E+02, 3.6438E+02, 3.6743E+02,
     +  3.7050E+02, 3.7359E+02, 3.7670E+02, 3.7983E+02, 3.8297E+02,
     + 3.8613E+02, 3.8931E+02, 3.9251E+02, 3.9573E+02, 3.9897E+02,
     +  4.0223E+02, 4.0550E+02, 4.0880E+02, 4.1211E+02, 4.1545E+02,
     + 4.1880E+02, 4.2217E+02, 4.2557E+02, 4.2898E+02, 4.3242E+02,
     +  4.3587E+02, 4.3934E+02, 4.4284E+02, 4.4636E+02, 4.4989E+02,
     + 4.5345E+02, 4.5703E+02, 4.6063E+02, 4.6425E+02, 4.6789E+02,
     +  4.7156E+02, 4.7524E+02, 4.7895E+02, 4.8268E+02, 4.8643E+02,
     + 4.9021E+02, 4.9400E+02, 4.9782E+02, 5.0166E+02, 5.0553E+02,
     +  5.0941E+02, 5.1332E+02, 5.1725E+02, 5.2121E+02, 5.2519E+02,
     + 5.2919E+02, 5.3322E+02, 5.3727E+02, 5.4134E+02, 5.4544E+02,
     +  5.4956E+02, 5.5371E+02, 5.5788E+02, 5.6207E+02, 5.6629E+02,
     + 5.7054E+02, 5.7480E+02, 5.7910E+02, 5.8342E+02, 5.8776E+02,
     +  5.9213E+02, 5.9653E+02, 6.0095E+02, 6.0539E+02, 6.0986E+02,
     + 6.1436E+02, 6.1889E+02, 6.2344E+02, 6.2802E+02, 6.3262E+02,
     +  6.3725E+02, 6.4191E+02, 6.4659E+02, 6.5130E+02, 6.5604E+02,
     + 6.6080E+02, 6.6560E+02, 6.7042E+02, 6.7527E+02, 6.8014E+02,
     +  6.8505E+02, 6.8998E+02, 6.9494E+02, 6.9993E+02, 7.0495E+02,
     + 7.0999E+02, 7.1507E+02, 7.2017E+02, 7.2531E+02, 7.3047E+02,
     +  7.3566E+02, 7.4088E+02, 7.4613E+02, 7.5142E+02, 7.5673E+02,
     + 7.6207E+02, 7.6744E+02, 7.7284E+02, 7.7827E+02, 7.8373E+02,
     +  7.8923E+02, 7.9475E+02, 8.0031E+02, 8.0590E+02, 8.1151E+02,
     + 8.1716E+02, 8.2284E+02, 8.2856E+02, 8.3430E+02, 8.4008E+02,
     +  8.4589E+02, 8.5173E+02, 8.5761E+02, 8.6351E+02, 8.6945E+02,
     + 8.7543E+02, 8.8143E+02, 8.8747E+02, 8.9354E+02, 8.9965E+02,
     +  9.0579E+02, 9.1196E+02, 9.1817E+02, 9.2442E+02, 9.3069E+02,
     + 9.3700E+02, 9.4335E+02, 9.4973E+02, 9.5615E+02, 9.6260E+02,
     +  9.6908E+02, 9.7560E+02, 9.8216E+02, 9.8875E+02, 9.9538E+02,
     + 1.0020E+03, 1.0087E+03, 1.0155E+03, 1.0223E+03, 1.0291E+03,
     +  1.0359E+03, 1.0428E+03, 1.0497E+03, 1.0567E+03, 1.0637E+03,
     + 1.0707E+03, 1.0778E+03, 1.0849E+03, 1.0920E+03, 1.0992E+03,
     +  1.1065E+03, 1.1137E+03, 1.1210E+03, 1.1284E+03, 1.1357E+03,
     + 1.1432E+03, 1.1506E+03, 1.1581E+03, 1.1657E+03, 1.1732E+03,
     +  1.1809E+03, 1.1885E+03, 1.1962E+03, 1.2040E+03, 1.2117E+03,
     + 1.2196E+03, 1.2274E+03, 1.2353E+03, 1.2433E+03, 1.2513E+03,
     +  1.2593E+03, 1.2674E+03, 1.2755E+03, 1.2836E+03, 1.2918E+03,
     + 1.3001E+03, 1.3084E+03, 1.3167E+03, 1.3250E+03, 1.3335E+03,
     +  1.3419E+03, 1.3504E+03, 1.3590E+03, 1.3676E+03, 1.3762E+03,
     + 1.3849E+03, 1.3936E+03, 1.4024E+03, 1.4112E+03, 1.4200E+03,
     +  1.4289E+03, 1.4379E+03, 1.4469E+03, 1.4559E+03, 1.4650E+03,
     + 1.4742E+03, 1.4833E+03, 1.4926E+03, 1.5018E+03, 1.5112E+03,
     +  1.5205E+03, 1.5299E+03, 1.5394E+03, 1.5489E+03, 1.5585E+03,
     + 1.5681E+03, 1.5777E+03, 1.5875E+03, 1.5972E+03, 1.6070E+03,
     +  1.6169E+03, 1.6268E+03, 1.6367E+03, 1.6467E+03, 1.6568E+03,
     + 1.6669E+03, 1.6770E+03, 1.6872E+03, 1.6975E+03, 1.7078E+03,
     +  1.7181E+03, 1.7285E+03, 1.7390E+03, 1.7495E+03, 1.7601E+03,
     + 1.7707E+03, 1.7814E+03, 1.7921E+03, 1.8029E+03, 1.8137E+03,
     +  1.8246E+03, 1.8355E+03, 1.8465E+03, 1.8575E+03, 1.8686E+03,
     + 1.8798E+03, 1.8910E+03, 1.9022E+03, 1.9135E+03, 1.9249E+03,
     +  1.9363E+03, 1.9478E+03, 1.9593E+03, 1.9709E+03, 1.9826E+03,
     + 1.9943E+03, 2.0061E+03, 2.0179E+03, 2.0297E+03, 2.0417E+03,
     +  2.0537E+03, 2.0657E+03, 2.0778E+03, 2.0900E+03, 2.1022E+03,
     + 2.1145E+03, 2.1268E+03, 2.1392E+03, 2.1517E+03, 2.1642E+03,
     +  2.1768E+03, 2.1894E+03, 2.2021E+03, 2.2148E+03, 2.2277E+03,
     + 2.2405E+03, 2.2535E+03, 2.2665E+03, 2.2795E+03, 2.2927E+03,
     +  2.3058E+03, 2.3191E+03, 2.3324E+03, 2.3458E+03, 2.3592E+03,
     + 2.3727E+03, 2.3863E+03, 2.3999E+03, 2.4136E+03, 2.4273E+03,
     +  2.4411E+03, 2.4550E+03, 2.4690E+03, 2.4830E+03, 2.4971E+03,
     + 2.5112E+03, 2.5254E+03, 2.5397E+03, 2.5540E+03, 2.5684E+03,
     +  2.5829E+03, 2.5974E+03, 2.6120E+03, 2.6267E+03, 2.6415E+03,
     + 2.6563E+03, 2.6711E+03, 2.6861E+03, 2.7011E+03, 2.7162E+03,
     +  2.7313E+03, 2.7466E+03, 2.7618E+03, 2.7772E+03, 2.7926E+03,
     + 2.8081E+03, 2.8237E+03, 2.8394E+03, 2.8551E+03, 2.8708E+03,
     +  2.8867E+03, 2.9026E+03, 2.9186E+03, 2.9347E+03, 2.9508E+03,
     + 2.9671E+03, 2.9834E+03, 2.9997E+03, 3.0162E+03, 3.0327E+03,
     +  3.0493E+03, 3.0659E+03, 3.0826E+03, 3.0995E+03, 3.1163E+03,
     + 3.1333E+03, 3.1503E+03, 3.1674E+03, 3.1846E+03, 3.2019E+03,
     +  3.2192E+03, 3.2367E+03, 3.2541E+03, 3.2717E+03, 3.2894E+03,
     + 3.3071E+03, 3.3249E+03, 3.3428E+03, 3.3608E+03, 3.3788E+03,
     +  3.3969E+03, 3.4151E+03, 3.4334E+03, 3.4518E+03, 3.4702E+03,
     + 3.4887E+03, 3.5073E+03, 3.5260E+03, 3.5448E+03, 3.5636E+03,
     +  3.5826E+03, 3.6016E+03, 3.6207E+03, 3.6398E+03, 3.6591E+03,
     + 3.6785E+03, 3.6979E+03, 3.7174E+03, 3.7370E+03, 3.7567E+03,
     +  3.7764E+03, 3.7963E+03, 3.8162E+03, 3.8363E+03, 3.8564E+03,
     + 3.8766E+03, 3.8969E+03, 3.9172E+03, 3.9377E+03, 3.9582E+03,
     +  3.9789E+03, 3.9996E+03, 4.0204E+03, 4.0413E+03, 4.0623E+03,
     + 4.0833E+03, 4.1045E+03, 4.1258E+03, 4.1471E+03, 4.1685E+03,
     +  4.1901E+03, 4.2117E+03, 4.2334E+03, 4.2552E+03, 4.2771E+03,
     + 4.2991E+03, 4.3212E+03, 4.3433E+03, 4.3656E+03, 4.3879E+03,
     +  4.4104E+03, 4.4329E+03, 4.4556E+03, 4.4783E+03, 4.5011E+03,
     + 4.5241E+03, 4.5471E+03, 4.5702E+03, 4.5934E+03, 4.6167E+03,
     +  4.6401E+03, 4.6636E+03, 4.6872E+03, 4.7109E+03, 4.7347E+03,
     + 4.7586E+03, 4.7826E+03, 4.8067E+03, 4.8309E+03, 4.8552E+03,
     +  4.8795E+03, 4.9040E+03, 4.9286E+03, 4.9533E+03, 4.9781E+03,
     + 5.0030E+03, 5.0280E+03, 5.0531E+03, 5.0783E+03, 5.1036E+03,
     +  5.1290E+03, 5.1545E+03, 5.1801E+03, 5.2058E+03, 5.2317E+03,
     + 5.2576E+03, 5.2836E+03, 5.3098E+03, 5.3360E+03, 5.3624E+03,
     +  5.3888E+03, 5.4154E+03, 5.4420E+03, 5.4688E+03, 5.4957E+03,
     + 5.5227E+03, 5.5498E+03, 5.5770E+03, 5.6043E+03, 5.6318E+03,
     +  5.6593E+03, 5.6869E+03, 5.7147E+03, 5.7426E+03, 5.7705E+03,
     + 5.7986E+03, 5.8268E+03, 5.8552E+03, 5.8836E+03, 5.9121E+03,
     +  5.9408E+03, 5.9695E+03, 5.9984E+03, 6.0274E+03, 6.0565E+03,
     + 6.0858E+03, 6.1151E+03, 6.1446E+03, 6.1741E+03, 6.2038E+03,
     +  6.2336E+03, 6.2635E+03, 6.2936E+03, 6.3237E+03, 6.3540E+03,
     + 6.3844E+03, 6.4149E+03, 6.4455E+03, 6.4763E+03, 6.5072E+03,
     +  6.5382E+03, 6.5693E+03, 6.6005E+03, 6.6318E+03, 6.6633E+03,
     + 6.6949E+03, 6.7266E+03, 6.7585E+03, 6.7904E+03, 6.8225E+03,
     +  6.8547E+03, 6.8871E+03, 6.9195E+03, 6.9521E+03, 6.9848E+03,
     + 7.0177E+03, 7.0506E+03, 7.0837E+03, 7.1169E+03, 7.1503E+03,
     +  7.1838E+03, 7.2174E+03, 7.2511E+03, 7.2849E+03, 7.3189E+03,
     + 7.3530E+03, 7.3873E+03, 7.4216E+03, 7.4561E+03, 7.4908E+03,
     +  7.5255E+03, 7.5604E+03, 7.5955E+03, 7.6306E+03, 7.6659E+03,
     + 7.7014E+03, 7.7369E+03, 7.7726E+03, 7.8084E+03, 7.8444E+03,
     +  7.8805E+03, 7.9167E+03, 7.9531E+03, 7.9896E+03, 8.0263E+03,
     + 8.0630E+03, 8.1000E+03, 8.1370E+03, 8.1742E+03, 8.2115E+03,
     +  8.2490E+03, 8.2866E+03, 8.3244E+03, 8.3623E+03, 8.4003E+03,
     + 8.4385E+03, 8.4768E+03, 8.5152E+03, 8.5538E+03, 8.5926E+03,
     +  8.6315E+03, 8.6705E+03, 8.7097E+03, 8.7490E+03, 8.7884E+03,
     + 8.8280E+03, 8.8678E+03, 8.9077E+03, 8.9477E+03, 8.9879E+03,
     +  9.0283E+03, 9.0687E+03, 9.1094E+03, 9.1502E+03, 9.1911E+03,
     + 9.2322E+03, 9.2734E+03, 9.3148E+03, 9.3563E+03, 9.3980E+03,
     +  9.4398E+03, 9.4818E+03, 9.5239E+03, 9.5662E+03, 9.6086E+03,
     + 9.6512E+03, 9.6940E+03, 9.7369E+03, 9.7799E+03, 9.8231E+03,
     +  9.8665E+03, 9.9100E+03, 9.9537E+03, 9.9975E+03, 1.0042E+04,
     + 1.0086E+04, 1.0130E+04, 1.0174E+04, 1.0219E+04, 1.0264E+04,
     +  1.0309E+04, 1.0354E+04, 1.0399E+04, 1.0445E+04, 1.0490E+04,
     + 1.0536E+04, 1.0582E+04, 1.0628E+04, 1.0674E+04, 1.0721E+04,
     +  1.0767E+04, 1.0814E+04, 1.0861E+04, 1.0908E+04, 1.0955E+04,
     + 1.1002E+04, 1.1050E+04, 1.1098E+04, 1.1146E+04, 1.1194E+04,
     +  1.1242E+04, 1.1290E+04, 1.1339E+04, 1.1388E+04, 1.1436E+04,
     + 1.1486E+04, 1.1535E+04, 1.1584E+04, 1.1634E+04, 1.1684E+04,
     +  1.1734E+04, 1.1784E+04, 1.1834E+04, 1.1884E+04, 1.1935E+04,
     + 1.1986E+04, 1.2037E+04, 1.2088E+04, 1.2139E+04, 1.2191E+04,
     +  1.2243E+04, 1.2294E+04, 1.2346E+04, 1.2399E+04, 1.2451E+04,
     + 1.2504E+04, 1.2556E+04, 1.2609E+04, 1.2663E+04, 1.2716E+04,
     +  1.2769E+04, 1.2823E+04, 1.2877E+04, 1.2931E+04, 1.2985E+04,
     + 1.3040E+04, 1.3094E+04, 1.3149E+04, 1.3204E+04, 1.3259E+04,
     +  1.3314E+04, 1.3370E+04, 1.3426E+04, 1.3482E+04, 1.3538E+04,
     + 1.3594E+04, 1.3650E+04, 1.3707E+04, 1.3764E+04, 1.3821E+04,
     +  1.3878E+04, 1.3936E+04, 1.3993E+04, 1.4051E+04, 1.4109E+04,
     + 1.4167E+04, 1.4226E+04, 1.4284E+04, 1.4343E+04, 1.4402E+04,
     +  1.4461E+04, 1.4521E+04, 1.4580E+04, 1.4640E+04, 1.4700E+04,
     + 1.4760E+04, 1.4821E+04, 1.4881E+04, 1.4942E+04, 1.5003E+04,
     +  1.5064E+04, 1.5126E+04, 1.5187E+04, 1.5249E+04, 1.5311E+04,
     + 1.5373E+04, 1.5436E+04, 1.5498E+04, 1.5561E+04, 1.5624E+04,
     +  1.5687E+04, 1.5751E+04, 1.5814E+04, 1.5878E+04, 1.5942E+04,
     + 1.6007E+04, 1.6071E+04, 1.6136E+04, 1.6201E+04, 1.6266E+04,
     +  1.6331E+04, 1.6397E+04, 1.6462E+04, 1.6528E+04, 1.6595E+04,
     + 1.6661E+04, 1.6728E+04, 1.6794E+04, 1.6861E+04, 1.6929E+04,
     +  1.6996E+04, 1.7064E+04, 1.7132E+04, 1.7200E+04, 1.7268E+04,
     + 1.7337E+04, 1.7406E+04, 1.7475E+04, 1.7544E+04, 1.7613E+04,
     +  1.7683E+04, 1.7753E+04, 1.7823E+04, 1.7893E+04, 1.7964E+04,
     + 1.8035E+04, 1.8106E+04, 1.8177E+04, 1.8249E+04, 1.8320E+04,
     +  1.8392E+04, 1.8464E+04, 1.8537E+04, 1.8609E+04, 1.8682E+04,
     + 1.8755E+04, 1.8829E+04, 1.8902E+04, 1.8976E+04, 1.9050E+04,
     +  1.9124E+04, 1.9199E+04, 1.9274E+04, 1.9348E+04, 1.9424E+04,
     + 1.9499E+04, 1.9575E+04, 1.9651E+04, 1.9727E+04, 1.9803E+04,
     +  1.9880E+04, 1.9957E+04, 2.0034E+04, 2.0111E+04, 2.0189E+04,
     + 2.0266E+04, 2.0345E+04, 2.0423E+04, 2.0501E+04, 2.0580E+04,
     +  2.0659E+04, 2.0738E+04, 2.0818E+04, 2.0898E+04, 2.0978E+04,
     + 2.1058E+04, 2.1139E+04, 2.1219E+04, 2.1300E+04, 2.1382E+04,
     +  2.1463E+04, 2.1545E+04, 2.1627E+04, 2.1709E+04, 2.1792E+04,
     + 2.1875E+04, 2.1958E+04, 2.2041E+04, 2.2124E+04, 2.2208E+04,
     +  2.2292E+04, 2.2377E+04, 2.2461E+04, 2.2546E+04, 2.2631E+04,
     + 2.2716E+04, 2.2802E+04, 2.2888E+04, 2.2974E+04, 2.3060E+04,
     +  2.3147E+04, 2.3234E+04, 2.3321E+04, 2.3409E+04, 2.3496E+04,
     + 2.3584E+04, 2.3672E+04, 2.3761E+04, 2.3850E+04, 2.3939E+04,
     +  2.4028E+04, 2.4118E+04, 2.4207E+04, 2.4298E+04, 2.4388E+04,
     + 2.4479E+04, 2.4570E+04, 2.4661E+04, 2.4752E+04, 2.4844E+04,
     +  2.4936E+04, 2.5028E+04, 2.5121E+04, 2.5214E+04, 2.5307E+04,
     + 2.5400E+04, 2.5494E+04, 2.5588E+04, 2.5682E+04, 2.5777E+04,
     +  2.5871E+04, 2.5966E+04, 2.6062E+04, 2.6157E+04, 2.6253E+04,
     + 2.6350E+04, 2.6446E+04, 2.6543E+04, 2.6640E+04, 2.6737E+04,
     +  2.6835E+04, 2.6933E+04, 2.7031E+04, 2.7130E+04, 2.7228E+04,
     + 2.7327E+04, 2.7427E+04, 2.7526E+04, 2.7626E+04, 2.7727E+04,
     +  2.7827E+04, 2.7928E+04, 2.8029E+04, 2.8131E+04, 2.8232E+04,
     + 2.8334E+04, 2.8437E+04, 2.8539E+04, 2.8642E+04, 2.8745E+04,
     +  2.8849E+04, 2.8953E+04, 2.9057E+04, 2.9161E+04, 2.9266E+04,
     + 2.9371E+04, 2.9476E+04, 2.9582E+04, 2.9688E+04, 2.9794E+04,
     +  2.9901E+04, 3.0007E+04, 3.0115E+04, 3.0222E+04, 3.0330E+04,
     + 3.0438E+04, 3.0546E+04, 3.0655E+04, 3.0764E+04, 3.0873E+04,
     +  3.0983E+04, 3.1093E+04, 3.1203E+04, 3.1314E+04, 3.1425E+04,
     + 3.1536E+04, 3.1647E+04, 3.1759E+04, 3.1871E+04, 3.1984E+04,
     +  3.2097E+04, 3.2210E+04, 3.2323E+04, 3.2437E+04, 3.2551E+04,
     + 3.2665E+04, 3.2780E+04, 3.2895E+04, 3.3011E+04, 3.3126E+04,
     +  3.3242E+04, 3.3359E+04, 3.3475E+04, 3.3592E+04, 3.3710E+04,
     + 3.3827E+04, 3.3945E+04, 3.4064E+04, 3.4182E+04, 3.4301E+04,
     +  3.4421E+04, 3.4540E+04, 3.4660E+04, 3.4781E+04, 3.4901E+04,
     + 3.5022E+04, 3.5144E+04, 3.5265E+04, 3.5387E+04, 3.5510E+04,
     +  3.5632E+04, 3.5755E+04, 3.5879E+04, 3.6002E+04, 3.6126E+04,
     + 3.6251E+04, 3.6376E+04, 3.6501E+04, 3.6626E+04, 3.6752E+04,
     +  3.6878E+04, 3.7005E+04, 3.7131E+04, 3.7259E+04, 3.7386E+04,
     + 3.7514E+04, 3.7642E+04, 3.7771E+04, 3.7900E+04, 3.8029E+04,
     +  3.8159E+04, 3.8289E+04, 3.8419E+04, 3.8550E+04, 3.8681E+04,
     + 3.8812E+04, 3.8944E+04, 3.9076E+04, 3.9209E+04, 3.9341E+04,
     +  3.9475E+04, 3.9608E+04, 3.9742E+04, 3.9876E+04, 4.0011E+04,
     + 4.0146E+04, 4.0282E+04, 4.0417E+04, 4.0553E+04, 4.0690E+04,
     +  4.0827E+04, 4.0964E+04, 4.1102E+04, 4.1240E+04, 4.1378E+04,
     + 4.1517E+04, 4.1656E+04, 4.1795E+04, 4.1935E+04, 4.2075E+04,
     +  4.2216E+04, 4.2357E+04, 4.2498E+04, 4.2640E+04, 4.2782E+04,
     + 4.2925E+04, 4.3068E+04, 4.3211E+04, 4.3354E+04, 4.3498E+04,
     +  4.3643E+04, 4.3788E+04, 4.3933E+04, 4.4078E+04, 4.4224E+04,
     + 4.4371E+04, 4.4517E+04, 4.4664E+04, 4.4812E+04, 4.4960E+04,
     +  4.5108E+04, 4.5257E+04, 4.5406E+04, 4.5555E+04, 4.5705E+04,
     + 4.5855E+04, 4.6006E+04, 4.6157E+04, 4.6308E+04, 4.6460E+04,
     +  4.6612E+04, 4.6765E+04, 4.6918E+04, 4.7072E+04, 4.7225E+04,
     + 4.7380E+04, 4.7534E+04, 4.7689E+04, 4.7845E+04, 4.8001E+04,
     +  4.8157E+04, 4.8313E+04, 4.8471E+04, 4.8628E+04, 4.8786E+04,
     + 4.8944E+04, 4.9103E+04, 4.9262E+04, 4.9422E+04, 4.9582E+04,
     +  4.9742E+04, 4.9903E+04, 5.0064E+04, 5.0225E+04, 5.0388E+04,
     + 5.0550E+04, 5.0713E+04, 5.0876E+04, 5.1040E+04, 5.1204E+04,
     +  5.1369E+04, 5.1533E+04, 5.1699E+04, 5.1865E+04, 5.2031E+04,
     + 5.2198E+04, 5.2365E+04, 5.2532E+04, 5.2700E+04, 5.2869E+04,
     +  5.3037E+04, 5.3207E+04, 5.3376E+04, 5.3546E+04, 5.3717E+04,
     + 5.3888E+04, 5.4059E+04, 5.4231E+04, 5.4404E+04, 5.4576E+04,
     +  5.4749E+04, 5.4923E+04, 5.5097E+04, 5.5272E+04, 5.5446E+04,
     + 5.5622E+04, 5.5798E+04, 5.5974E+04, 5.6151E+04, 5.6328E+04,
     +  5.6505E+04, 5.6683E+04, 5.6862E+04, 5.7041E+04, 5.7220E+04,
     + 5.7400E+04, 5.7580E+04, 5.7761E+04, 5.7942E+04, 5.8124E+04,
     +  5.8306E+04, 5.8488E+04, 5.8672E+04, 5.8855E+04, 5.9039E+04,
     + 5.9223E+04, 5.9408E+04, 5.9593E+04, 5.9779E+04, 5.9965E+04,
     +  6.0152E+04, 6.0339E+04, 6.0527E+04, 6.0715E+04, 6.0904E+04,
     + 6.1093E+04, 6.1282E+04, 6.1472E+04, 6.1662E+04, 6.1853E+04,
     +  6.2045E+04, 6.2237E+04, 6.2429E+04, 6.2622E+04, 6.2815E+04,
     + 6.3009E+04, 6.3203E+04, 6.3398E+04, 6.3593E+04, 6.3788E+04,
     +  6.3985E+04, 6.4181E+04, 6.4378E+04, 6.4576E+04, 6.4774E+04,
     + 6.4973E+04, 6.5172E+04, 6.5371E+04, 6.5571E+04, 6.5772E+04,
     +  6.5973E+04, 6.6174E+04, 6.6376E+04, 6.6578E+04, 6.6781E+04,
     + 6.6985E+04, 6.7189E+04, 6.7393E+04, 6.7598E+04, 6.7804E+04,
     +  6.8010E+04, 6.8216E+04, 6.8423E+04, 6.8630E+04, 6.8838E+04,
     + 6.9047E+04, 6.9256E+04, 6.9465E+04, 6.9675E+04, 6.9885E+04,
     +  7.0096E+04, 7.0308E+04, 7.0520E+04, 7.0732E+04, 7.0945E+04,
     + 7.1159E+04, 7.1373E+04, 7.1587E+04, 7.1802E+04, 7.2018E+04,
     +  7.2234E+04, 7.2450E+04, 7.2668E+04, 7.2885E+04, 7.3103E+04,
     + 7.3322E+04, 7.3541E+04, 7.3761E+04, 7.3981E+04, 7.4202E+04,
     +  7.4423E+04, 7.4645E+04, 7.4867E+04, 7.5090E+04, 7.5313E+04,
     + 7.5537E+04, 7.5762E+04, 7.5987E+04, 7.6212E+04, 7.6438E+04,
     +  7.6665E+04, 7.6892E+04, 7.7119E+04, 7.7348E+04, 7.7576E+04,
     + 7.7805E+04, 7.8035E+04, 7.8266E+04, 7.8496E+04, 7.8728E+04,
     +  7.8960E+04, 7.9192E+04, 7.9425E+04, 7.9659E+04, 7.9893E+04,
     + 8.0128E+04, 8.0363E+04, 8.0599E+04, 8.0835E+04, 8.1072E+04,
     +  8.1309E+04, 8.1547E+04, 8.1786E+04, 8.2025E+04, 8.2264E+04,
     + 8.2505E+04, 8.2745E+04, 8.2987E+04, 8.3228E+04, 8.3471E+04,
     +  8.3714E+04, 8.3957E+04, 8.4202E+04, 8.4446E+04, 8.4691E+04,
     + 8.4937E+04, 8.5184E+04, 8.5431E+04, 8.5678E+04, 8.5926E+04,
     +  8.6175E+04, 8.6424E+04, 8.6674E+04, 8.6924E+04, 8.7175E+04,
     + 8.7427E+04, 8.7679E+04, 8.7931E+04, 8.8185E+04, 8.8438E+04,
     +  8.8693E+04, 8.8948E+04, 8.9203E+04, 8.9460E+04, 8.9716E+04,
     + 8.9974E+04, 9.0232E+04, 9.0490E+04, 9.0749E+04, 9.1009E+04,
     +  9.1269E+04, 9.1530E+04, 9.1791E+04, 9.2053E+04, 9.2316E+04,
     + 9.2579E+04, 9.2843E+04, 9.3107E+04, 9.3372E+04, 9.3638E+04,
     +  9.3904E+04, 9.4171E+04, 9.4439E+04, 9.4707E+04, 9.4975E+04,
     + 9.5244E+04, 9.5514E+04, 9.5785E+04, 9.6056E+04, 9.6327E+04,
     +  9.6600E+04, 9.6872E+04, 9.7146E+04, 9.7420E+04, 9.7695E+04,
     + 9.7970E+04, 9.8246E+04, 9.8523E+04, 9.8800E+04, 9.9077E+04,
     +  9.9356E+04, 9.9635E+04, 9.9915E+04, 1.0019E+05, 1.0048E+05,
     + 1.0076E+05, 1.0104E+05, 1.0132E+05, 1.0161E+05, 1.0189E+05,
     +  1.0217E+05, 1.0246E+05, 1.0275E+05, 1.0303E+05, 1.0332E+05,
     + 1.0361E+05, 1.0390E+05, 1.0418E+05, 1.0447E+05, 1.0476E+05,
     +  1.0506E+05, 1.0535E+05, 1.0564E+05, 1.0593E+05, 1.0623E+05,
     + 1.0652E+05, 1.0681E+05, 1.0711E+05, 1.0741E+05, 1.0770E+05,
     +  1.0800E+05, 1.0830E+05, 1.0860E+05, 1.0890E+05, 1.0920E+05,
     + 1.0950E+05, 1.0980E+05, 1.1010E+05, 1.1040E+05, 1.1071E+05,
     +  1.1101E+05, 1.1132E+05, 1.1162E+05, 1.1193E+05, 1.1223E+05,
     + 1.1254E+05, 1.1285E+05, 1.1316E+05, 1.1347E+05, 1.1378E+05,
     +  1.1409E+05, 1.1440E+05, 1.1471E+05, 1.1502E+05, 1.1534E+05,
     + 1.1565E+05, 1.1597E+05, 1.1628E+05, 1.1660E+05, 1.1691E+05,
     +  1.1723E+05, 1.1755E+05, 1.1787E+05, 1.1819E+05, 1.1851E+05,
     + 1.1883E+05, 1.1915E+05, 1.1947E+05, 1.1979E+05, 1.2012E+05,
     +  1.2044E+05, 1.2077E+05, 1.2109E+05, 1.2142E+05, 1.2175E+05,
     + 1.2207E+05, 1.2240E+05, 1.2273E+05, 1.2306E+05, 1.2339E+05,
     +  1.2372E+05, 1.2405E+05, 1.2439E+05, 1.2472E+05, 1.2505E+05,
     + 1.2539E+05, 1.2572E+05, 1.2606E+05, 1.2640E+05, 1.2674E+05,
     +  1.2707E+05, 1.2741E+05, 1.2775E+05, 1.2809E+05, 1.2843E+05,
     + 1.2878E+05, 1.2912E+05, 1.2946E+05, 1.2981E+05, 1.3015E+05,
     +  1.3050E+05, 1.3084E+05, 1.3119E+05, 1.3154E+05, 1.3188E+05,
     + 1.3223E+05, 1.3258E+05, 1.3293E+05, 1.3329E+05, 1.3364E+05,
     +  1.3399E+05, 1.3434E+05, 1.3470E+05, 1.3505E+05, 1.3541E+05,
     + 1.3576E+05, 1.3612E+05, 1.3648E+05, 1.3684E+05, 1.3720E+05,
     +  1.3756E+05, 1.3792E+05, 1.3828E+05, 1.3864E+05, 1.3901E+05,
     + 1.3937E+05, 1.3973E+05, 1.4010E+05, 1.4047E+05, 1.4083E+05,
     +  1.4120E+05, 1.4157E+05, 1.4194E+05, 1.4231E+05, 1.4268E+05,
     + 1.4305E+05, 1.4342E+05, 1.4379E+05, 1.4417E+05, 1.4454E+05,
     +  1.4492E+05, 1.4529E+05, 1.4567E+05, 1.4605E+05, 1.4643E+05,
     + 1.4681E+05, 1.4718E+05, 1.4757E+05, 1.4795E+05, 1.4833E+05,
     +  1.4871E+05, 1.4910E+05, 1.4948E+05, 1.4987E+05, 1.5025E+05,
     + 1.5064E+05, 1.5103E+05, 1.5141E+05, 1.5180E+05, 1.5219E+05,
     +  1.5258E+05, 1.5298E+05, 1.5337E+05, 1.5376E+05, 1.5415E+05,
     + 1.5455E+05, 1.5494E+05, 1.5534E+05, 1.5574E+05, 1.5614E+05,
     +  1.5653E+05, 1.5693E+05, 1.5733E+05, 1.5774E+05, 1.5814E+05,
     + 1.5854E+05, 1.5894E+05, 1.5935E+05, 1.5975E+05, 1.6016E+05,
     +  1.6057E+05, 1.6097E+05, 1.6138E+05, 1.6179E+05, 1.6220E+05,
     + 1.6261E+05, 1.6302E+05, 1.6344E+05, 1.6385E+05, 1.6426E+05,
     +  1.6468E+05, 1.6509E+05, 1.6551E+05, 1.6593E+05, 1.6635E+05,
     + 1.6676E+05, 1.6718E+05, 1.6760E+05, 1.6803E+05, 1.6845E+05,
     +  1.6887E+05, 1.6930E+05, 1.6972E+05, 1.7015E+05, 1.7057E+05,
     + 1.7100E+05, 1.7143E+05, 1.7186E+05, 1.7229E+05, 1.7272E+05,
     +  1.7315E+05, 1.7358E+05, 1.7401E+05, 1.7445E+05, 1.7488E+05,
     + 1.7532E+05, 1.7576E+05, 1.7619E+05, 1.7663E+05, 1.7707E+05,
     +  1.7751E+05, 1.7795E+05, 1.7839E+05, 1.7884E+05, 1.7928E+05,
     + 1.7972E+05, 1.8017E+05, 1.8062E+05, 1.8106E+05, 1.8151E+05,
     +  1.8196E+05, 1.8241E+05, 1.8286E+05, 1.8331E+05, 1.8376E+05,
     + 1.8422E+05, 1.8467E+05, 1.8512E+05, 1.8558E+05, 1.8604E+05,
     +  1.8649E+05, 1.8695E+05, 1.8741E+05, 1.8787E+05, 1.8833E+05,
     + 1.8879E+05, 1.8926E+05, 1.8972E+05, 1.9018E+05, 1.9065E+05,
     +  1.9112E+05, 1.9158E+05, 1.9205E+05, 1.9252E+05, 1.9299E+05,
     + 1.9346E+05, 1.9393E+05, 1.9440E+05, 1.9488E+05, 1.9535E+05,
     +  1.9583E+05, 1.9630E+05, 1.9678E+05, 1.9726E+05, 1.9774E+05,
     + 1.9822E+05, 1.9870E+05, 1.9918E+05, 1.9966E+05, 2.0014E+05,
     +  2.0063E+05, 2.0111E+05, 2.0160E+05, 2.0209E+05, 2.0257E+05,
     + 2.0306E+05, 2.0355E+05, 2.0404E+05, 2.0454E+05, 2.0503E+05,
     +  2.0552E+05, 2.0602E+05, 2.0651E+05, 2.0701E+05, 2.0750E+05,
     + 2.0800E+05, 2.0850E+05, 2.0900E+05, 2.0950E+05, 2.1000E+05,
     +  2.1051E+05, 2.1101E+05, 2.1152E+05, 2.1202E+05, 2.1253E+05,
     + 2.1304E+05, 2.1354E+05, 2.1405E+05, 2.1456E+05, 2.1507E+05,
     +  2.1559E+05, 2.1610E+05, 2.1661E+05, 2.1713E+05, 2.1764E+05,
     + 2.1816E+05, 2.1868E+05, 2.1920E+05, 2.1972E+05, 2.2024E+05,
     +  2.2076E+05, 2.2128E+05, 2.2181E+05, 2.2233E+05, 2.2286E+05,
     + 2.2338E+05, 2.2391E+05, 2.2444E+05, 2.2497E+05, 2.2550E+05,
     +  2.2603E+05, 2.2656E+05, 2.2710E+05, 2.2763E+05, 2.2817E+05,
     + 2.2870E+05, 2.2924E+05, 2.2978E+05, 2.3032E+05, 2.3086E+05,
     +  2.3140E+05, 2.3194E+05, 2.3249E+05, 2.3303E+05, 2.3357E+05,
     + 2.3412E+05, 2.3467E+05, 2.3522E+05, 2.3577E+05, 2.3632E+05,
     +  2.3687E+05, 2.3742E+05, 2.3797E+05, 2.3853E+05, 2.3908E+05,
     + 2.3964E+05, 2.4019E+05, 2.4075E+05, 2.4131E+05, 2.4187E+05,
     +  2.4243E+05, 2.4299E+05, 2.4356E+05, 2.4412E+05, 2.4469E+05,
     + 2.4525E+05, 2.4582E+05, 2.4639E+05, 2.4696E+05, 2.4753E+05,
     +  2.4810E+05, 2.4867E+05, 2.4925E+05, 2.4982E+05, 2.5040E+05,
     + 2.5097E+05, 2.5155E+05, 2.5213E+05, 2.5271E+05, 2.5329E+05,
     +  2.5387E+05, 2.5445E+05, 2.5504E+05, 2.5562E+05, 2.5621E+05,
     + 2.5679E+05, 2.5738E+05, 2.5797E+05, 2.5856E+05, 2.5915E+05,
     +  2.5974E+05, 2.6034E+05, 2.6093E+05, 2.6152E+05, 2.6212E+05,
     + 2.6272E+05, 2.6332E+05, 2.6392E+05, 2.6452E+05, 2.6512E+05,
     +  2.6572E+05, 2.6632E+05, 2.6693E+05, 2.6753E+05, 2.6814E+05,
     + 2.6875E+05, 2.6936E+05, 2.6997E+05, 2.7058E+05, 2.7119E+05,
     +  2.7180E+05, 2.7242E+05, 2.7303E+05, 2.7365E+05, 2.7427E+05,
     + 2.7489E+05, 2.7550E+05, 2.7613E+05, 2.7675E+05, 2.7737E+05,
     +  2.7799E+05, 2.7862E+05, 2.7924E+05, 2.7987E+05, 2.8050E+05,
     + 2.8113E+05, 2.8176E+05, 2.8239E+05, 2.8302E+05, 2.8366E+05,
     +  2.8429E+05, 2.8493E+05, 2.8557E+05, 2.8620E+05, 2.8684E+05,
     + 2.8748E+05, 2.8812E+05, 2.8877E+05, 2.8941E+05, 2.9006E+05,
     +  2.9070E+05, 2.9135E+05, 2.9200E+05, 2.9264E+05, 2.9330E+05,
     + 2.9395E+05, 2.9460E+05, 2.9525E+05, 2.9591E+05, 2.9656E+05,
     +  2.9722E+05, 2.9788E+05, 2.9854E+05, 2.9920E+05, 2.9986E+05,
     + 3.0052E+05, 3.0118E+05, 3.0185E+05, 3.0252E+05, 3.0318E+05,
     +  3.0385E+05, 3.0452E+05, 3.0519E+05, 3.0586E+05, 3.0653E+05,
     + 3.0721E+05, 3.0788E+05, 3.0856E+05, 3.0924E+05, 3.0992E+05,
     +  3.1060E+05, 3.1128E+05, 3.1196E+05, 3.1264E+05, 3.1332E+05,
     + 3.1401E+05, 3.1470E+05, 3.1538E+05, 3.1607E+05, 3.1676E+05,
     +  3.1745E+05, 3.1815E+05, 3.1884E+05, 3.1953E+05, 3.2023E+05,
     + 3.2093E+05, 3.2162E+05, 3.2232E+05, 3.2302E+05, 3.2373E+05,
     +  3.2443E+05, 3.2513E+05, 3.2584E+05, 3.2654E+05, 3.2725E+05,
     + 3.2796E+05, 3.2867E+05, 3.2938E+05, 3.3009E+05, 3.3081E+05,
     +  3.3152E+05, 3.3224E+05, 3.3295E+05, 3.3367E+05, 3.3439E+05,
     + 3.3511E+05, 3.3583E+05, 3.3655E+05, 3.3728E+05, 3.3800E+05,
     +  3.3873E+05, 3.3946E+05, 3.4019E+05, 3.4092E+05, 3.4165E+05,
     + 3.4238E+05, 3.4311E+05, 3.4385E+05, 3.4458E+05, 3.4532E+05,
     +  3.4606E+05, 3.4680E+05, 3.4754E+05, 3.4828E+05, 3.4902E+05,
     + 3.4977E+05, 3.5051E+05, 3.5126E+05, 3.5201E+05, 3.5276E+05,
     +  3.5351E+05, 3.5426E+05, 3.5501E+05, 3.5577E+05, 3.5652E+05,
     + 3.5728E+05, 3.5804E+05, 3.5880E+05, 3.5956E+05, 3.6032E+05,
     +  3.6108E+05, 3.6184E+05, 3.6261E+05, 3.6337E+05, 3.6414E+05,
     + 3.6491E+05, 3.6568E+05, 3.6645E+05, 3.6723E+05, 3.6800E+05,
     +  3.6877E+05, 3.6955E+05, 3.7033E+05, 3.7111E+05, 3.7189E+05,
     + 3.7267E+05, 3.7345E+05, 3.7423E+05, 3.7502E+05, 3.7581E+05,
     +  3.7659E+05, 3.7738E+05, 3.7817E+05, 3.7896E+05, 3.7976E+05,
     + 3.8055E+05, 3.8135E+05, 3.8214E+05, 3.8294E+05, 3.8374E+05,
     +  3.8454E+05, 3.8534E+05, 3.8614E+05, 3.8695E+05, 3.8775E+05,
     + 3.8856E+05, 3.8937E+05, 3.9017E+05, 3.9099E+05, 3.9180E+05,
     +  3.9261E+05, 3.9342E+05, 3.9424E+05, 3.9506E+05, 3.9587E+05,
     + 3.9669E+05, 3.9751E+05, 3.9834E+05, 3.9916E+05, 3.9998E+05,
     +  4.0081E+05, 4.0164E+05, 4.0247E+05, 4.0329E+05, 4.0413E+05,
     + 4.0496E+05, 4.0579E+05, 4.0663E+05, 4.0746E+05, 4.0830E+05,
     +  4.0914E+05, 4.0998E+05, 4.1082E+05, 4.1166E+05, 4.1251E+05,
     + 4.1335E+05, 4.1420E+05, 4.1505E+05, 4.1590E+05, 4.1675E+05,
     +  4.1760E+05, 4.1845E+05, 4.1931E+05, 4.2016E+05, 4.2102E+05,
     + 4.2188E+05, 4.2274E+05, 4.2360E+05, 4.2446E+05, 4.2533E+05,
     +  4.2619E+05, 4.2706E+05, 4.2792E+05, 4.2879E+05, 4.2966E+05,
     + 4.3054E+05, 4.3141E+05, 4.3228E+05, 4.3316E+05, 4.3404E+05,
     +  4.3491E+05, 4.3579E+05, 4.3668E+05, 4.3756E+05, 4.3844E+05,
     + 4.3933E+05, 4.4021E+05, 4.4110E+05, 4.4199E+05, 4.4288E+05,
     +  4.4377E+05, 4.4467E+05, 4.4556E+05, 4.4646E+05, 4.4736E+05,
     + 4.4825E+05, 4.4915E+05, 4.5006E+05, 4.5096E+05, 4.5186E+05,
     +  4.5277E+05, 4.5368E+05, 4.5458E+05, 4.5549E+05, 4.5640E+05,
     + 4.5732E+05, 4.5823E+05, 4.5915E+05, 4.6006E+05, 4.6098E+05,
     +  4.6190E+05, 4.6282E+05, 4.6374E+05, 4.6467E+05, 4.6559E+05,
     + 4.6652E+05, 4.6744E+05, 4.6837E+05, 4.6930E+05, 4.7024E+05,
     +  4.7117E+05, 4.7210E+05, 4.7304E+05, 4.7398E+05, 4.7492E+05,
     + 4.7586E+05, 4.7680E+05, 4.7774E+05, 4.7869E+05, 4.7963E+05,
     +  4.8058E+05, 4.8153E+05, 4.8248E+05, 4.8343E+05, 4.8438E+05,
     + 4.8534E+05, 4.8629E+05, 4.8725E+05, 4.8821E+05, 4.8917E+05,
     +  4.9013E+05, 4.9109E+05, 4.9205E+05, 4.9302E+05, 4.9399E+05,
     + 4.9495E+05, 4.9592E+05, 4.9690E+05, 4.9787E+05, 4.9884E+05,
     +  4.9982E+05, 5.0079E+05, 5.0177E+05, 5.0275E+05, 5.0373E+05,
     + 5.0472E+05, 5.0570E+05, 5.0669E+05, 5.0767E+05, 5.0866E+05,
     +  5.0965E+05, 5.1064E+05, 5.1164E+05, 5.1263E+05, 5.1363E+05,
     + 5.1462E+05, 5.1562E+05, 5.1662E+05, 5.1762E+05, 5.1863E+05,
     +  5.1963E+05, 5.2064E+05, 5.2164E+05, 5.2265E+05, 5.2366E+05,
     + 5.2467E+05, 5.2569E+05, 5.2670E+05, 5.2772E+05, 5.2873E+05,
     +  5.2975E+05, 5.3077E+05, 5.3180E+05, 5.3282E+05, 5.3384E+05,
     + 5.3487E+05, 5.3590E+05, 5.3693E+05, 5.3796E+05, 5.3899E+05,
     +  5.4002E+05, 5.4106E+05, 5.4209E+05, 5.4313E+05, 5.4417E+05,
     + 5.4521E+05, 5.4626E+05, 5.4730E+05, 5.4834E+05, 5.4939E+05,
     +  5.5044E+05, 5.5149E+05, 5.5254E+05, 5.5359E+05, 5.5465E+05,
     + 5.5570E+05, 5.5676E+05, 5.5782E+05, 5.5888E+05, 5.5994E+05,
     +  5.6101E+05, 5.6207E+05, 5.6314E+05, 5.6421E+05, 5.6528E+05,
     + 5.6635E+05, 5.6742E+05, 5.6849E+05, 5.6957E+05, 5.7065E+05,
     +  5.7172E+05, 5.7280E+05, 5.7389E+05, 5.7497E+05, 5.7605E+05,
     + 5.7714E+05, 5.7823E+05, 5.7932E+05, 5.8041E+05, 5.8150E+05,
     +  5.8259E+05, 5.8369E+05, 5.8478E+05, 5.8588E+05, 5.8698E+05,
     + 5.8808E+05, 5.8919E+05, 5.9029E+05, 5.9140E+05, 5.9250E+05,
     +  5.9361E+05, 5.9472E+05, 5.9583E+05, 5.9695E+05, 5.9806E+05,
     + 5.9918E+05, 6.0030E+05, 6.0142E+05, 6.0254E+05, 6.0366E+05,
     +  6.0479E+05, 6.0591E+05, 6.0704E+05, 6.0817E+05, 6.0930E+05,
     + 6.1043E+05, 6.1157E+05, 6.1270E+05, 6.1384E+05, 6.1498E+05,
     +  6.1612E+05, 6.1726E+05, 6.1840E+05, 6.1955E+05, 6.2069E+05,
     + 6.2184E+05, 6.2299E+05, 6.2414E+05, 6.2529E+05, 6.2645E+05,
     +  6.2760E+05, 6.2876E+05, 6.2992E+05, 6.3108E+05, 6.3224E+05,
     + 6.3340E+05, 6.3457E+05, 6.3573E+05, 6.3690E+05, 6.3807E+05,
     +  6.3924E+05, 6.4042E+05, 6.4159E+05, 6.4277E+05, 6.4395E+05,
     + 6.4512E+05, 6.4631E+05, 6.4749E+05, 6.4867E+05, 6.4986E+05,
     +  6.5104E+05, 6.5223E+05, 6.5342E+05, 6.5462E+05, 6.5581E+05,
     + 6.5701E+05, 6.5820E+05, 6.5940E+05, 6.6060E+05, 6.6180E+05,
     +  6.6301E+05, 6.6421E+05, 6.6542E+05, 6.6663E+05, 6.6783E+05,
     + 6.6905E+05, 6.7026E+05, 6.7147E+05, 6.7269E+05, 6.7391E+05,
     +  6.7513E+05, 6.7635E+05, 6.7757E+05, 6.7879E+05, 6.8002E+05,
     + 6.8125E+05, 6.8248E+05, 6.8371E+05, 6.8494E+05, 6.8617E+05,
     +  6.8741E+05, 6.8865E+05, 6.8989E+05, 6.9113E+05, 6.9237E+05,
     + 6.9361E+05, 6.9486E+05, 6.9611E+05, 6.9735E+05, 6.9860E+05,
     +  6.9986E+05, 7.0111E+05, 7.0237E+05, 7.0362E+05, 7.0488E+05,
     + 7.0614E+05, 7.0740E+05, 7.0867E+05, 7.0993E+05, 7.1120E+05,
     +  7.1247E+05, 7.1374E+05, 7.1501E+05, 7.1628E+05, 7.1756E+05,
     + 7.1883E+05, 7.2011E+05, 7.2139E+05, 7.2267E+05, 7.2396E+05,
     +  7.2524E+05, 7.2653E+05, 7.2782E+05, 7.2911E+05, 7.3040E+05,
     + 7.3169E+05, 7.3299E+05, 7.3428E+05, 7.3558E+05, 7.3688E+05,
     +  7.3818E+05, 7.3949E+05, 7.4079E+05, 7.4210E+05, 7.4341E+05,
     + 7.4472E+05, 7.4603E+05, 7.4734E+05, 7.4866E+05, 7.4998E+05,
     +  7.5129E+05, 7.5261E+05, 7.5394E+05, 7.5526E+05, 7.5659E+05,
     + 7.5791E+05, 7.5924E+05, 7.6057E+05, 7.6190E+05, 7.6324E+05,
     +  7.6457E+05, 7.6591E+05, 7.6725E+05, 7.6859E+05, 7.6993E+05,
     + 7.7128E+05, 7.7262E+05, 7.7397E+05, 7.7532E+05, 7.7667E+05,
     +  7.7802E+05, 7.7938E+05, 7.8073E+05, 7.8209E+05, 7.8345E+05,
     + 7.8481E+05, 7.8617E+05, 7.8754E+05, 7.8890E+05, 7.9027E+05,
     +  7.9164E+05, 7.9301E+05, 7.9438E+05, 7.9576E+05, 7.9714E+05,
     + 7.9851E+05, 7.9989E+05, 8.0128E+05, 8.0266E+05, 8.0404E+05,
     +  8.0543E+05, 8.0682E+05, 8.0821E+05, 8.0960E+05, 8.1100E+05,
     + 8.1239E+05, 8.1379E+05, 8.1519E+05, 8.1659E+05, 8.1799E+05,
     +  8.1939E+05, 8.2080E+05, 8.2221E+05, 8.2362E+05, 8.2503E+05,
     + 8.2644E+05, 8.2786E+05, 8.2927E+05, 8.3069E+05, 8.3211E+05,
     +  8.3353E+05, 8.3496E+05, 8.3638E+05, 8.3781E+05, 8.3924E+05,
     + 8.4067E+05, 8.4210E+05, 8.4353E+05, 8.4497E+05, 8.4641E+05,
     +  8.4785E+05, 8.4929E+05, 8.5073E+05, 8.5217E+05, 8.5362E+05,
     + 8.5507E+05, 8.5652E+05, 8.5797E+05, 8.5942E+05, 8.6088E+05,
     +  8.6233E+05, 8.6379E+05, 8.6525E+05, 8.6672E+05, 8.6818E+05,
     + 8.6965E+05, 8.7111E+05, 8.7258E+05, 8.7405E+05, 8.7553E+05,
     +  8.7700E+05, 8.7848E+05, 8.7996E+05, 8.8144E+05, 8.8292E+05,
     + 8.8440E+05, 8.8589E+05, 8.8737E+05, 8.8886E+05, 8.9035E+05,
     +  8.9185E+05, 8.9334E+05, 8.9484E+05, 8.9633E+05, 8.9783E+05/
      data (QofT( 7,J),J=2409,4510)/ 8.9934E+05, 9.0084E+05,
     +  9.0234E+05, 9.0385E+05, 9.0536E+05, 9.0687E+05, 9.0838E+05,
     +  9.0990E+05,
     + 9.1141E+05, 9.1293E+05, 9.1445E+05, 9.1597E+05, 9.1749E+05,
     +  9.1902E+05, 9.2055E+05, 9.2207E+05, 9.2361E+05, 9.2514E+05,
     + 9.2667E+05, 9.2821E+05, 9.2975E+05, 9.3128E+05, 9.3283E+05,
     +  9.3437E+05, 9.3591E+05, 9.3746E+05, 9.3901E+05, 9.4056E+05,
     + 9.4211E+05, 9.4367E+05, 9.4522E+05, 9.4678E+05, 9.4834E+05,
     +  9.4990E+05, 9.5146E+05, 9.5303E+05, 9.5460E+05, 9.5616E+05,
     + 9.5774E+05, 9.5931E+05, 9.6088E+05, 9.6246E+05, 9.6404E+05,
     +  9.6562E+05, 9.6720E+05, 9.6878E+05, 9.7037E+05, 9.7195E+05,
     + 9.7354E+05, 9.7513E+05, 9.7673E+05, 9.7832E+05, 9.7992E+05,
     +  9.8151E+05, 9.8311E+05, 9.8472E+05, 9.8632E+05, 9.8793E+05,
     + 9.8953E+05, 9.9114E+05, 9.9275E+05, 9.9437E+05, 9.9598E+05,
     +  9.9760E+05, 9.9922E+05, 1.0008E+06, 1.0025E+06, 1.0041E+06,
     + 1.0057E+06, 1.0073E+06, 1.0090E+06, 1.0106E+06, 1.0122E+06,
     +  1.0139E+06, 1.0155E+06, 1.0171E+06, 1.0188E+06, 1.0204E+06,
     + 1.0221E+06, 1.0237E+06, 1.0254E+06, 1.0270E+06, 1.0287E+06,
     +  1.0303E+06, 1.0320E+06, 1.0336E+06, 1.0353E+06, 1.0370E+06,
     + 1.0386E+06, 1.0403E+06, 1.0420E+06, 1.0436E+06, 1.0453E+06,
     +  1.0470E+06, 1.0486E+06, 1.0503E+06, 1.0520E+06, 1.0537E+06,
     + 1.0554E+06, 1.0571E+06, 1.0587E+06, 1.0604E+06, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       728
      data (QofT( 8,J),J=1,2408)/ 1.3912E+01, 2.5475E+01, 3.7121E+01,
     +  4.8786E+01, 6.0459E+01, 7.2136E+01, 8.3814E+01, 9.5494E+01,
     + 1.0718E+02, 1.1886E+02, 1.3054E+02, 1.4222E+02, 1.5390E+02,
     +  1.6559E+02, 1.7727E+02, 1.8895E+02, 2.0063E+02, 2.1232E+02,
     + 2.2400E+02, 2.3569E+02, 2.5905E+02, 2.8242E+02, 3.0579E+02,
     +  3.2916E+02, 3.5252E+02, 3.7589E+02, 3.9926E+02, 4.2263E+02,
     + 4.4600E+02, 4.6937E+02, 4.9274E+02, 5.1610E+02, 5.3947E+02,
     +  5.6284E+02, 5.8621E+02, 6.0958E+02, 6.3295E+02, 6.5632E+02,
     + 6.7969E+02, 7.0306E+02, 7.2643E+02, 7.4980E+02, 7.7317E+02,
     +  7.9655E+02, 8.1992E+02, 8.4329E+02, 8.6666E+02, 8.9004E+02,
     + 9.1341E+02, 9.3678E+02, 9.6016E+02, 9.8354E+02, 1.0069E+03,
     +  1.0303E+03, 1.0537E+03, 1.0771E+03, 1.1005E+03, 1.1239E+03,
     + 1.1473E+03, 1.1707E+03, 1.1941E+03, 1.2175E+03, 1.2409E+03,
     +  1.2644E+03, 1.2878E+03, 1.3113E+03, 1.3347E+03, 1.3582E+03,
     + 1.3817E+03, 1.4052E+03, 1.4288E+03, 1.4523E+03, 1.4759E+03,
     +  1.4995E+03, 1.5231E+03, 1.5468E+03, 1.5705E+03, 1.5942E+03,
     + 1.6179E+03, 1.6417E+03, 1.6655E+03, 1.6893E+03, 1.7132E+03,
     +  1.7371E+03, 1.7611E+03, 1.7851E+03, 1.8092E+03, 1.8333E+03,
     + 1.8574E+03, 1.8817E+03, 1.9059E+03, 1.9303E+03, 1.9546E+03,
     +  1.9791E+03, 2.0036E+03, 2.0282E+03, 2.0528E+03, 2.0776E+03,
     + 2.1023E+03, 2.1272E+03, 2.1522E+03, 2.1772E+03, 2.2023E+03,
     +  2.2275E+03, 2.2528E+03, 2.2781E+03, 2.3036E+03, 2.3291E+03,
     + 2.3548E+03, 2.3805E+03, 2.4063E+03, 2.4323E+03, 2.4583E+03,
     +  2.4845E+03, 2.5107E+03, 2.5371E+03, 2.5635E+03, 2.5901E+03,
     + 2.6168E+03, 2.6436E+03, 2.6706E+03, 2.6976E+03, 2.7248E+03,
     +  2.7521E+03, 2.7795E+03, 2.8070E+03, 2.8347E+03, 2.8625E+03,
     + 2.8905E+03, 2.9186E+03, 2.9468E+03, 2.9751E+03, 3.0036E+03,
     +  3.0323E+03, 3.0610E+03, 3.0900E+03, 3.1190E+03, 3.1483E+03,
     + 3.1776E+03, 3.2072E+03, 3.2368E+03, 3.2667E+03, 3.2967E+03,
     +  3.3268E+03, 3.3571E+03, 3.3876E+03, 3.4183E+03, 3.4491E+03,
     + 3.4800E+03, 3.5112E+03, 3.5425E+03, 3.5740E+03, 3.6056E+03,
     +  3.6375E+03, 3.6695E+03, 3.7017E+03, 3.7340E+03, 3.7666E+03,
     + 3.7993E+03, 3.8322E+03, 3.8653E+03, 3.8986E+03, 3.9321E+03,
     +  3.9658E+03, 3.9996E+03, 4.0337E+03, 4.0679E+03, 4.1024E+03,
     + 4.1370E+03, 4.1719E+03, 4.2069E+03, 4.2421E+03, 4.2776E+03,
     +  4.3132E+03, 4.3491E+03, 4.3852E+03, 4.4215E+03, 4.4579E+03,
     + 4.4946E+03, 4.5316E+03, 4.5687E+03, 4.6060E+03, 4.6436E+03,
     +  4.6814E+03, 4.7194E+03, 4.7576E+03, 4.7961E+03, 4.8348E+03,
     + 4.8737E+03, 4.9128E+03, 4.9522E+03, 4.9917E+03, 5.0316E+03,
     +  5.0716E+03, 5.1119E+03, 5.1525E+03, 5.1932E+03, 5.2342E+03,
     + 5.2755E+03, 5.3170E+03, 5.3587E+03, 5.4007E+03, 5.4429E+03,
     +  5.4854E+03, 5.5281E+03, 5.5711E+03, 5.6144E+03, 5.6578E+03,
     + 5.7016E+03, 5.7456E+03, 5.7898E+03, 5.8344E+03, 5.8791E+03,
     +  5.9242E+03, 5.9695E+03, 6.0150E+03, 6.0609E+03, 6.1070E+03,
     + 6.1533E+03, 6.2000E+03, 6.2469E+03, 6.2941E+03, 6.3416E+03,
     +  6.3893E+03, 6.4373E+03, 6.4856E+03, 6.5342E+03, 6.5831E+03,
     + 6.6322E+03, 6.6817E+03, 6.7314E+03, 6.7814E+03, 6.8317E+03,
     +  6.8823E+03, 6.9332E+03, 6.9844E+03, 7.0359E+03, 7.0876E+03,
     + 7.1397E+03, 7.1921E+03, 7.2448E+03, 7.2978E+03, 7.3510E+03,
     +  7.4046E+03, 7.4585E+03, 7.5128E+03, 7.5673E+03, 7.6221E+03,
     + 7.6773E+03, 7.7327E+03, 7.7885E+03, 7.8446E+03, 7.9011E+03,
     +  7.9578E+03, 8.0149E+03, 8.0723E+03, 8.1300E+03, 8.1881E+03,
     + 8.2464E+03, 8.3052E+03, 8.3642E+03, 8.4236E+03, 8.4833E+03,
     +  8.5434E+03, 8.6038E+03, 8.6645E+03, 8.7256E+03, 8.7870E+03,
     + 8.8488E+03, 8.9109E+03, 8.9734E+03, 9.0362E+03, 9.0994E+03,
     +  9.1629E+03, 9.2268E+03, 9.2910E+03, 9.3556E+03, 9.4206E+03,
     + 9.4859E+03, 9.5516E+03, 9.6176E+03, 9.6840E+03, 9.7508E+03,
     +  9.8180E+03, 9.8855E+03, 9.9534E+03, 1.0022E+04, 1.0090E+04,
     + 1.0159E+04, 1.0229E+04, 1.0299E+04, 1.0369E+04, 1.0439E+04,
     +  1.0510E+04, 1.0582E+04, 1.0653E+04, 1.0726E+04, 1.0798E+04,
     + 1.0871E+04, 1.0944E+04, 1.1018E+04, 1.1092E+04, 1.1167E+04,
     +  1.1242E+04, 1.1317E+04, 1.1393E+04, 1.1469E+04, 1.1545E+04,
     + 1.1622E+04, 1.1700E+04, 1.1778E+04, 1.1856E+04, 1.1935E+04,
     +  1.2014E+04, 1.2093E+04, 1.2173E+04, 1.2254E+04, 1.2334E+04,
     + 1.2416E+04, 1.2497E+04, 1.2579E+04, 1.2662E+04, 1.2745E+04,
     +  1.2828E+04, 1.2912E+04, 1.2996E+04, 1.3081E+04, 1.3166E+04,
     + 1.3252E+04, 1.3338E+04, 1.3425E+04, 1.3512E+04, 1.3599E+04,
     +  1.3687E+04, 1.3776E+04, 1.3865E+04, 1.3954E+04, 1.4044E+04,
     + 1.4134E+04, 1.4225E+04, 1.4316E+04, 1.4408E+04, 1.4500E+04,
     +  1.4593E+04, 1.4686E+04, 1.4779E+04, 1.4874E+04, 1.4968E+04,
     + 1.5063E+04, 1.5159E+04, 1.5255E+04, 1.5352E+04, 1.5449E+04,
     +  1.5546E+04, 1.5644E+04, 1.5743E+04, 1.5842E+04, 1.5942E+04,
     + 1.6042E+04, 1.6142E+04, 1.6244E+04, 1.6345E+04, 1.6447E+04,
     +  1.6550E+04, 1.6653E+04, 1.6757E+04, 1.6862E+04, 1.6966E+04,
     + 1.7072E+04, 1.7178E+04, 1.7284E+04, 1.7391E+04, 1.7499E+04,
     +  1.7607E+04, 1.7715E+04, 1.7824E+04, 1.7934E+04, 1.8044E+04,
     + 1.8155E+04, 1.8266E+04, 1.8378E+04, 1.8491E+04, 1.8604E+04,
     +  1.8718E+04, 1.8832E+04, 1.8946E+04, 1.9062E+04, 1.9178E+04,
     + 1.9294E+04, 1.9411E+04, 1.9529E+04, 1.9647E+04, 1.9766E+04,
     +  1.9885E+04, 2.0005E+04, 2.0126E+04, 2.0247E+04, 2.0369E+04,
     + 2.0491E+04, 2.0614E+04, 2.0738E+04, 2.0862E+04, 2.0987E+04,
     +  2.1112E+04, 2.1238E+04, 2.1365E+04, 2.1492E+04, 2.1620E+04,
     + 2.1748E+04, 2.1877E+04, 2.2007E+04, 2.2138E+04, 2.2269E+04,
     +  2.2400E+04, 2.2532E+04, 2.2665E+04, 2.2799E+04, 2.2933E+04,
     + 2.3068E+04, 2.3204E+04, 2.3340E+04, 2.3476E+04, 2.3614E+04,
     +  2.3752E+04, 2.3891E+04, 2.4030E+04, 2.4170E+04, 2.4311E+04,
     + 2.4453E+04, 2.4595E+04, 2.4738E+04, 2.4881E+04, 2.5025E+04,
     +  2.5170E+04, 2.5316E+04, 2.5462E+04, 2.5609E+04, 2.5757E+04,
     + 2.5905E+04, 2.6054E+04, 2.6204E+04, 2.6354E+04, 2.6505E+04,
     +  2.6657E+04, 2.6810E+04, 2.6963E+04, 2.7117E+04, 2.7272E+04,
     + 2.7427E+04, 2.7583E+04, 2.7740E+04, 2.7898E+04, 2.8056E+04,
     +  2.8215E+04, 2.8375E+04, 2.8536E+04, 2.8697E+04, 2.8859E+04,
     + 2.9022E+04, 2.9186E+04, 2.9350E+04, 2.9515E+04, 2.9681E+04,
     +  2.9847E+04, 3.0015E+04, 3.0183E+04, 3.0352E+04, 3.0522E+04,
     + 3.0692E+04, 3.0863E+04, 3.1035E+04, 3.1208E+04, 3.1382E+04,
     +  3.1556E+04, 3.1732E+04, 3.1908E+04, 3.2084E+04, 3.2262E+04,
     + 3.2440E+04, 3.2620E+04, 3.2800E+04, 3.2980E+04, 3.3162E+04,
     +  3.3345E+04, 3.3528E+04, 3.3712E+04, 3.3897E+04, 3.4083E+04,
     + 3.4270E+04, 3.4457E+04, 3.4645E+04, 3.4834E+04, 3.5025E+04,
     +  3.5215E+04, 3.5407E+04, 3.5600E+04, 3.5793E+04, 3.5987E+04,
     + 3.6183E+04, 3.6379E+04, 3.6575E+04, 3.6773E+04, 3.6972E+04,
     +  3.7171E+04, 3.7372E+04, 3.7573E+04, 3.7775E+04, 3.7978E+04,
     + 3.8182E+04, 3.8387E+04, 3.8593E+04, 3.8800E+04, 3.9007E+04,
     +  3.9216E+04, 3.9425E+04, 3.9635E+04, 3.9847E+04, 4.0059E+04,
     + 4.0272E+04, 4.0486E+04, 4.0701E+04, 4.0917E+04, 4.1133E+04,
     +  4.1351E+04, 4.1570E+04, 4.1789E+04, 4.2010E+04, 4.2232E+04,
     + 4.2454E+04, 4.2678E+04, 4.2902E+04, 4.3127E+04, 4.3354E+04,
     +  4.3581E+04, 4.3810E+04, 4.4039E+04, 4.4269E+04, 4.4500E+04,
     + 4.4733E+04, 4.4966E+04, 4.5200E+04, 4.5435E+04, 4.5672E+04,
     +  4.5909E+04, 4.6147E+04, 4.6386E+04, 4.6627E+04, 4.6868E+04,
     + 4.7110E+04, 4.7354E+04, 4.7598E+04, 4.7844E+04, 4.8090E+04,
     +  4.8338E+04, 4.8586E+04, 4.8836E+04, 4.9086E+04, 4.9338E+04,
     + 4.9591E+04, 4.9845E+04, 5.0100E+04, 5.0356E+04, 5.0613E+04,
     +  5.0871E+04, 5.1130E+04, 5.1390E+04, 5.1651E+04, 5.1914E+04,
     + 5.2177E+04, 5.2442E+04, 5.2708E+04, 5.2975E+04, 5.3242E+04,
     +  5.3512E+04, 5.3782E+04, 5.4053E+04, 5.4325E+04, 5.4599E+04,
     + 5.4873E+04, 5.5149E+04, 5.5426E+04, 5.5704E+04, 5.5983E+04,
     +  5.6264E+04, 5.6545E+04, 5.6828E+04, 5.7112E+04, 5.7396E+04,
     + 5.7683E+04, 5.7970E+04, 5.8258E+04, 5.8548E+04, 5.8839E+04,
     +  5.9131E+04, 5.9424E+04, 5.9718E+04, 6.0014E+04, 6.0310E+04,
     + 6.0608E+04, 6.0908E+04, 6.1208E+04, 6.1509E+04, 6.1812E+04,
     +  6.2116E+04, 6.2421E+04, 6.2728E+04, 6.3036E+04, 6.3345E+04,
     + 6.3655E+04, 6.3966E+04, 6.4279E+04, 6.4593E+04, 6.4908E+04,
     +  6.5224E+04, 6.5542E+04, 6.5861E+04, 6.6181E+04, 6.6502E+04,
     + 6.6825E+04, 6.7149E+04, 6.7474E+04, 6.7801E+04, 6.8129E+04,
     +  6.8458E+04, 6.8789E+04, 6.9120E+04, 6.9454E+04, 6.9788E+04,
     + 7.0124E+04, 7.0461E+04, 7.0799E+04, 7.1139E+04, 7.1480E+04,
     +  7.1822E+04, 7.2166E+04, 7.2511E+04, 7.2857E+04, 7.3205E+04,
     + 7.3554E+04, 7.3905E+04, 7.4257E+04, 7.4610E+04, 7.4964E+04,
     +  7.5320E+04, 7.5678E+04, 7.6036E+04, 7.6396E+04, 7.6758E+04,
     + 7.7121E+04, 7.7485E+04, 7.7851E+04, 7.8218E+04, 7.8587E+04,
     +  7.8957E+04, 7.9328E+04, 7.9701E+04, 8.0075E+04, 8.0451E+04,
     + 8.0828E+04, 8.1206E+04, 8.1586E+04, 8.1968E+04, 8.2351E+04,
     +  8.2735E+04, 8.3121E+04, 8.3508E+04, 8.3897E+04, 8.4287E+04,
     + 8.4679E+04, 8.5072E+04, 8.5467E+04, 8.5863E+04, 8.6261E+04,
     +  8.6660E+04, 8.7061E+04, 8.7463E+04, 8.7867E+04, 8.8272E+04,
     + 8.8679E+04, 8.9087E+04, 8.9497E+04, 8.9908E+04, 9.0321E+04,
     +  9.0736E+04, 9.1152E+04, 9.1569E+04, 9.1988E+04, 9.2409E+04,
     + 9.2831E+04, 9.3255E+04, 9.3681E+04, 9.4108E+04, 9.4536E+04,
     +  9.4967E+04, 9.5398E+04, 9.5832E+04, 9.6267E+04, 9.6703E+04,
     + 9.7142E+04, 9.7581E+04, 9.8023E+04, 9.8466E+04, 9.8911E+04,
     +  9.9357E+04, 9.9805E+04, 1.0025E+05, 1.0071E+05, 1.0116E+05,
     + 1.0161E+05, 1.0207E+05, 1.0253E+05, 1.0299E+05, 1.0345E+05,
     +  1.0391E+05, 1.0438E+05, 1.0484E+05, 1.0531E+05, 1.0578E+05,
     + 1.0625E+05, 1.0673E+05, 1.0720E+05, 1.0768E+05, 1.0816E+05,
     +  1.0864E+05, 1.0912E+05, 1.0960E+05, 1.1009E+05, 1.1057E+05,
     + 1.1106E+05, 1.1155E+05, 1.1205E+05, 1.1254E+05, 1.1304E+05,
     +  1.1353E+05, 1.1403E+05, 1.1453E+05, 1.1504E+05, 1.1554E+05,
     + 1.1605E+05, 1.1656E+05, 1.1707E+05, 1.1758E+05, 1.1809E+05,
     +  1.1861E+05, 1.1913E+05, 1.1965E+05, 1.2017E+05, 1.2069E+05,
     + 1.2122E+05, 1.2174E+05, 1.2227E+05, 1.2280E+05, 1.2333E+05,
     +  1.2387E+05, 1.2440E+05, 1.2494E+05, 1.2548E+05, 1.2602E+05,
     + 1.2657E+05, 1.2711E+05, 1.2766E+05, 1.2821E+05, 1.2876E+05,
     +  1.2932E+05, 1.2987E+05, 1.3043E+05, 1.3099E+05, 1.3155E+05,
     + 1.3211E+05, 1.3268E+05, 1.3324E+05, 1.3381E+05, 1.3438E+05,
     +  1.3496E+05, 1.3553E+05, 1.3611E+05, 1.3669E+05, 1.3727E+05,
     + 1.3785E+05, 1.3844E+05, 1.3902E+05, 1.3961E+05, 1.4020E+05,
     +  1.4080E+05, 1.4139E+05, 1.4199E+05, 1.4259E+05, 1.4319E+05,
     + 1.4379E+05, 1.4440E+05, 1.4501E+05, 1.4562E+05, 1.4623E+05,
     +  1.4684E+05, 1.4746E+05, 1.4807E+05, 1.4869E+05, 1.4932E+05,
     + 1.4994E+05, 1.5057E+05, 1.5120E+05, 1.5183E+05, 1.5246E+05,
     +  1.5309E+05, 1.5373E+05, 1.5437E+05, 1.5501E+05, 1.5565E+05,
     + 1.5630E+05, 1.5695E+05, 1.5760E+05, 1.5825E+05, 1.5890E+05,
     +  1.5956E+05, 1.6022E+05, 1.6088E+05, 1.6154E+05, 1.6221E+05,
     + 1.6288E+05, 1.6355E+05, 1.6422E+05, 1.6489E+05, 1.6557E+05,
     +  1.6625E+05, 1.6693E+05, 1.6761E+05, 1.6830E+05, 1.6899E+05,
     + 1.6968E+05, 1.7037E+05, 1.7106E+05, 1.7176E+05, 1.7246E+05,
     +  1.7316E+05, 1.7387E+05, 1.7457E+05, 1.7528E+05, 1.7599E+05,
     + 1.7671E+05, 1.7742E+05, 1.7814E+05, 1.7886E+05, 1.7958E+05,
     +  1.8031E+05, 1.8104E+05, 1.8177E+05, 1.8250E+05, 1.8323E+05,
     + 1.8397E+05, 1.8471E+05, 1.8545E+05, 1.8619E+05, 1.8694E+05,
     +  1.8769E+05, 1.8844E+05, 1.8920E+05, 1.8995E+05, 1.9071E+05,
     + 1.9147E+05, 1.9224E+05, 1.9300E+05, 1.9377E+05, 1.9454E+05,
     +  1.9532E+05, 1.9609E+05, 1.9687E+05, 1.9765E+05, 1.9844E+05,
     + 1.9922E+05, 2.0001E+05, 2.0080E+05, 2.0160E+05, 2.0239E+05,
     +  2.0319E+05, 2.0399E+05, 2.0480E+05, 2.0560E+05, 2.0641E+05,
     + 2.0722E+05, 2.0804E+05, 2.0885E+05, 2.0967E+05, 2.1050E+05,
     +  2.1132E+05, 2.1215E+05, 2.1298E+05, 2.1381E+05, 2.1464E+05,
     + 2.1548E+05, 2.1632E+05, 2.1717E+05, 2.1801E+05, 2.1886E+05,
     +  2.1971E+05, 2.2056E+05, 2.2142E+05, 2.2228E+05, 2.2314E+05,
     + 2.2401E+05, 2.2487E+05, 2.2574E+05, 2.2662E+05, 2.2749E+05,
     +  2.2837E+05, 2.2925E+05, 2.3013E+05, 2.3102E+05, 2.3191E+05,
     + 2.3280E+05, 2.3370E+05, 2.3459E+05, 2.3549E+05, 2.3640E+05,
     +  2.3730E+05, 2.3821E+05, 2.3912E+05, 2.4004E+05, 2.4095E+05,
     + 2.4187E+05, 2.4279E+05, 2.4372E+05, 2.4465E+05, 2.4558E+05,
     +  2.4651E+05, 2.4745E+05, 2.4839E+05, 2.4933E+05, 2.5028E+05,
     + 2.5123E+05, 2.5218E+05, 2.5313E+05, 2.5409E+05, 2.5505E+05,
     +  2.5601E+05, 2.5698E+05, 2.5795E+05, 2.5892E+05, 2.5989E+05,
     + 2.6087E+05, 2.6185E+05, 2.6283E+05, 2.6382E+05, 2.6481E+05,
     +  2.6580E+05, 2.6680E+05, 2.6780E+05, 2.6880E+05, 2.6980E+05,
     + 2.7081E+05, 2.7182E+05, 2.7284E+05, 2.7385E+05, 2.7487E+05,
     +  2.7589E+05, 2.7692E+05, 2.7795E+05, 2.7898E+05, 2.8002E+05,
     + 2.8105E+05, 2.8210E+05, 2.8314E+05, 2.8419E+05, 2.8524E+05,
     +  2.8629E+05, 2.8735E+05, 2.8841E+05, 2.8947E+05, 2.9054E+05,
     + 2.9161E+05, 2.9268E+05, 2.9376E+05, 2.9484E+05, 2.9592E+05,
     +  2.9700E+05, 2.9809E+05, 2.9918E+05, 3.0028E+05, 3.0138E+05,
     + 3.0248E+05, 3.0358E+05, 3.0469E+05, 3.0580E+05, 3.0692E+05,
     +  3.0803E+05, 3.0915E+05, 3.1028E+05, 3.1141E+05, 3.1254E+05,
     + 3.1367E+05, 3.1481E+05, 3.1595E+05, 3.1709E+05, 3.1824E+05,
     +  3.1939E+05, 3.2054E+05, 3.2170E+05, 3.2286E+05, 3.2403E+05,
     + 3.2519E+05, 3.2636E+05, 3.2754E+05, 3.2872E+05, 3.2990E+05,
     +  3.3108E+05, 3.3227E+05, 3.3346E+05, 3.3465E+05, 3.3585E+05,
     + 3.3705E+05, 3.3826E+05, 3.3947E+05, 3.4068E+05, 3.4189E+05,
     +  3.4311E+05, 3.4433E+05, 3.4556E+05, 3.4679E+05, 3.4802E+05,
     + 3.4926E+05, 3.5050E+05, 3.5174E+05, 3.5299E+05, 3.5424E+05,
     +  3.5549E+05, 3.5675E+05, 3.5801E+05, 3.5927E+05, 3.6054E+05,
     + 3.6181E+05, 3.6309E+05, 3.6437E+05, 3.6565E+05, 3.6694E+05,
     +  3.6823E+05, 3.6952E+05, 3.7082E+05, 3.7212E+05, 3.7342E+05,
     + 3.7473E+05, 3.7604E+05, 3.7735E+05, 3.7867E+05, 3.8000E+05,
     +  3.8132E+05, 3.8265E+05, 3.8399E+05, 3.8532E+05, 3.8666E+05,
     + 3.8801E+05, 3.8936E+05, 3.9071E+05, 3.9207E+05, 3.9342E+05,
     +  3.9479E+05, 3.9616E+05, 3.9753E+05, 3.9890E+05, 4.0028E+05,
     + 4.0166E+05, 4.0305E+05, 4.0444E+05, 4.0583E+05, 4.0723E+05,
     +  4.0863E+05, 4.1004E+05, 4.1145E+05, 4.1286E+05, 4.1428E+05,
     + 4.1570E+05, 4.1712E+05, 4.1855E+05, 4.1998E+05, 4.2142E+05,
     +  4.2286E+05, 4.2430E+05, 4.2575E+05, 4.2720E+05, 4.2866E+05,
     + 4.3012E+05, 4.3158E+05, 4.3305E+05, 4.3452E+05, 4.3600E+05,
     +  4.3748E+05, 4.3896E+05, 4.4045E+05, 4.4194E+05, 4.4344E+05,
     + 4.4494E+05, 4.4644E+05, 4.4795E+05, 4.4946E+05, 4.5097E+05,
     +  4.5249E+05, 4.5402E+05, 4.5555E+05, 4.5708E+05, 4.5862E+05,
     + 4.6016E+05, 4.6170E+05, 4.6325E+05, 4.6480E+05, 4.6636E+05,
     +  4.6792E+05, 4.6948E+05, 4.7105E+05, 4.7263E+05, 4.7420E+05,
     + 4.7579E+05, 4.7737E+05, 4.7896E+05, 4.8056E+05, 4.8216E+05,
     +  4.8376E+05, 4.8537E+05, 4.8698E+05, 4.8859E+05, 4.9021E+05,
     + 4.9184E+05, 4.9346E+05, 4.9510E+05, 4.9673E+05, 4.9837E+05,
     +  5.0002E+05, 5.0167E+05, 5.0332E+05, 5.0498E+05, 5.0664E+05,
     + 5.0831E+05, 5.0998E+05, 5.1166E+05, 5.1334E+05, 5.1502E+05,
     +  5.1671E+05, 5.1840E+05, 5.2010E+05, 5.2180E+05, 5.2351E+05,
     + 5.2522E+05, 5.2693E+05, 5.2865E+05, 5.3038E+05, 5.3210E+05,
     +  5.3384E+05, 5.3557E+05, 5.3732E+05, 5.3906E+05, 5.4081E+05,
     + 5.4257E+05, 5.4433E+05, 5.4609E+05, 5.4786E+05, 5.4963E+05,
     +  5.5141E+05, 5.5319E+05, 5.5498E+05, 5.5677E+05, 5.5856E+05,
     + 5.6037E+05, 5.6217E+05, 5.6398E+05, 5.6579E+05, 5.6761E+05,
     +  5.6944E+05, 5.7126E+05, 5.7310E+05, 5.7493E+05, 5.7677E+05,
     + 5.7862E+05, 5.8047E+05, 5.8233E+05, 5.8419E+05, 5.8605E+05,
     +  5.8792E+05, 5.8980E+05, 5.9167E+05, 5.9356E+05, 5.9545E+05,
     + 5.9734E+05, 5.9924E+05, 6.0114E+05, 6.0305E+05, 6.0496E+05,
     +  6.0688E+05, 6.0880E+05, 6.1072E+05, 6.1265E+05, 6.1459E+05,
     + 6.1653E+05, 6.1848E+05, 6.2043E+05, 6.2238E+05, 6.2434E+05,
     +  6.2631E+05, 6.2828E+05, 6.3025E+05, 6.3223E+05, 6.3421E+05,
     + 6.3620E+05, 6.3820E+05, 6.4020E+05, 6.4220E+05, 6.4421E+05,
     +  6.4622E+05, 6.4824E+05, 6.5026E+05, 6.5229E+05, 6.5432E+05,
     + 6.5636E+05, 6.5841E+05, 6.6045E+05, 6.6251E+05, 6.6457E+05,
     +  6.6663E+05, 6.6870E+05, 6.7077E+05, 6.7285E+05, 6.7493E+05,
     + 6.7702E+05, 6.7911E+05, 6.8121E+05, 6.8331E+05, 6.8542E+05,
     +  6.8754E+05, 6.8965E+05, 6.9178E+05, 6.9391E+05, 6.9604E+05,
     + 6.9818E+05, 7.0032E+05, 7.0247E+05, 7.0463E+05, 7.0679E+05,
     +  7.0895E+05, 7.1112E+05, 7.1329E+05, 7.1547E+05, 7.1766E+05,
     + 7.1985E+05, 7.2205E+05, 7.2425E+05, 7.2645E+05, 7.2866E+05,
     +  7.3088E+05, 7.3310E+05, 7.3533E+05, 7.3756E+05, 7.3980E+05,
     + 7.4204E+05, 7.4429E+05, 7.4654E+05, 7.4880E+05, 7.5106E+05,
     +  7.5333E+05, 7.5561E+05, 7.5789E+05, 7.6017E+05, 7.6246E+05,
     + 7.6476E+05, 7.6706E+05, 7.6937E+05, 7.7168E+05, 7.7400E+05,
     +  7.7632E+05, 7.7865E+05, 7.8098E+05, 7.8332E+05, 7.8566E+05,
     + 7.8801E+05, 7.9037E+05, 7.9273E+05, 7.9510E+05, 7.9747E+05,
     +  7.9984E+05, 8.0223E+05, 8.0461E+05, 8.0701E+05, 8.0941E+05,
     + 8.1181E+05, 8.1422E+05, 8.1664E+05, 8.1906E+05, 8.2148E+05,
     +  8.2392E+05, 8.2635E+05, 8.2880E+05, 8.3125E+05, 8.3370E+05,
     + 8.3616E+05, 8.3863E+05, 8.4110E+05, 8.4357E+05, 8.4606E+05,
     +  8.4854E+05, 8.5104E+05, 8.5354E+05, 8.5604E+05, 8.5855E+05,
     + 8.6107E+05, 8.6359E+05, 8.6612E+05, 8.6865E+05, 8.7119E+05,
     +  8.7373E+05, 8.7628E+05, 8.7884E+05, 8.8140E+05, 8.8397E+05,
     + 8.8654E+05, 8.8912E+05, 8.9171E+05, 8.9430E+05, 8.9689E+05,
     +  8.9950E+05, 9.0210E+05, 9.0472E+05, 9.0734E+05, 9.0996E+05,
     + 9.1259E+05, 9.1523E+05, 9.1787E+05, 9.2052E+05, 9.2318E+05,
     +  9.2584E+05, 9.2850E+05, 9.3117E+05, 9.3385E+05, 9.3654E+05,
     + 9.3923E+05, 9.4192E+05, 9.4462E+05, 9.4733E+05, 9.5004E+05,
     +  9.5276E+05, 9.5549E+05, 9.5822E+05, 9.6096E+05, 9.6370E+05,
     + 9.6645E+05, 9.6920E+05, 9.7197E+05, 9.7473E+05, 9.7751E+05,
     +  9.8028E+05, 9.8307E+05, 9.8586E+05, 9.8866E+05, 9.9146E+05,
     + 9.9427E+05, 9.9709E+05, 9.9991E+05, 1.0027E+06, 1.0056E+06,
     +  1.0084E+06, 1.0113E+06, 1.0141E+06, 1.0170E+06, 1.0198E+06,
     + 1.0227E+06, 1.0256E+06, 1.0285E+06, 1.0313E+06, 1.0342E+06,
     +  1.0371E+06, 1.0400E+06, 1.0430E+06, 1.0459E+06, 1.0488E+06,
     + 1.0517E+06, 1.0547E+06, 1.0576E+06, 1.0606E+06, 1.0635E+06,
     +  1.0665E+06, 1.0695E+06, 1.0724E+06, 1.0754E+06, 1.0784E+06,
     + 1.0814E+06, 1.0844E+06, 1.0874E+06, 1.0904E+06, 1.0934E+06,
     +  1.0965E+06, 1.0995E+06, 1.1025E+06, 1.1056E+06, 1.1086E+06,
     + 1.1117E+06, 1.1148E+06, 1.1178E+06, 1.1209E+06, 1.1240E+06,
     +  1.1271E+06, 1.1302E+06, 1.1333E+06, 1.1364E+06, 1.1395E+06,
     + 1.1426E+06, 1.1458E+06, 1.1489E+06, 1.1520E+06, 1.1552E+06,
     +  1.1583E+06, 1.1615E+06, 1.1647E+06, 1.1678E+06, 1.1710E+06,
     + 1.1742E+06, 1.1774E+06, 1.1806E+06, 1.1838E+06, 1.1870E+06,
     +  1.1902E+06, 1.1935E+06, 1.1967E+06, 1.1999E+06, 1.2032E+06,
     + 1.2064E+06, 1.2097E+06, 1.2129E+06, 1.2162E+06, 1.2195E+06,
     +  1.2228E+06, 1.2261E+06, 1.2294E+06, 1.2327E+06, 1.2360E+06,
     + 1.2393E+06, 1.2426E+06, 1.2460E+06, 1.2493E+06, 1.2526E+06,
     +  1.2560E+06, 1.2594E+06, 1.2627E+06, 1.2661E+06, 1.2695E+06,
     + 1.2729E+06, 1.2762E+06, 1.2796E+06, 1.2830E+06, 1.2865E+06,
     +  1.2899E+06, 1.2933E+06, 1.2967E+06, 1.3002E+06, 1.3036E+06,
     + 1.3071E+06, 1.3105E+06, 1.3140E+06, 1.3175E+06, 1.3210E+06,
     +  1.3244E+06, 1.3279E+06, 1.3314E+06, 1.3349E+06, 1.3385E+06,
     + 1.3420E+06, 1.3455E+06, 1.3490E+06, 1.3526E+06, 1.3561E+06,
     +  1.3597E+06, 1.3633E+06, 1.3668E+06, 1.3704E+06, 1.3740E+06,
     + 1.3776E+06, 1.3812E+06, 1.3848E+06, 1.3884E+06, 1.3920E+06,
     +  1.3956E+06, 1.3993E+06, 1.4029E+06, 1.4066E+06, 1.4102E+06,
     + 1.4139E+06, 1.4175E+06, 1.4212E+06, 1.4249E+06, 1.4286E+06,
     +  1.4323E+06, 1.4360E+06, 1.4397E+06, 1.4434E+06, 1.4471E+06,
     + 1.4509E+06, 1.4546E+06, 1.4583E+06, 1.4621E+06, 1.4659E+06,
     +  1.4696E+06, 1.4734E+06, 1.4772E+06, 1.4810E+06, 1.4848E+06,
     + 1.4886E+06, 1.4924E+06, 1.4962E+06, 1.5000E+06, 1.5038E+06,
     +  1.5077E+06, 1.5115E+06, 1.5154E+06, 1.5192E+06, 1.5231E+06,
     + 1.5270E+06, 1.5309E+06, 1.5348E+06, 1.5386E+06, 1.5426E+06,
     +  1.5465E+06, 1.5504E+06, 1.5543E+06, 1.5582E+06, 1.5622E+06,
     + 1.5661E+06, 1.5701E+06, 1.5740E+06, 1.5780E+06, 1.5820E+06,
     +  1.5860E+06, 1.5900E+06, 1.5940E+06, 1.5980E+06, 1.6020E+06,
     + 1.6060E+06, 1.6100E+06, 1.6141E+06, 1.6181E+06, 1.6222E+06,
     +  1.6262E+06, 1.6303E+06, 1.6343E+06, 1.6384E+06, 1.6425E+06,
     + 1.6466E+06, 1.6507E+06, 1.6548E+06, 1.6589E+06, 1.6631E+06,
     +  1.6672E+06, 1.6713E+06, 1.6755E+06, 1.6796E+06, 1.6838E+06,
     + 1.6880E+06, 1.6921E+06, 1.6963E+06, 1.7005E+06, 1.7047E+06,
     +  1.7089E+06, 1.7131E+06, 1.7174E+06, 1.7216E+06, 1.7258E+06,
     + 1.7301E+06, 1.7343E+06, 1.7386E+06, 1.7428E+06, 1.7471E+06,
     +  1.7514E+06, 1.7557E+06, 1.7600E+06, 1.7643E+06, 1.7686E+06,
     + 1.7729E+06, 1.7773E+06, 1.7816E+06, 1.7859E+06, 1.7903E+06,
     +  1.7946E+06, 1.7990E+06, 1.8034E+06, 1.8078E+06, 1.8122E+06,
     + 1.8166E+06, 1.8210E+06, 1.8254E+06, 1.8298E+06, 1.8342E+06,
     +  1.8387E+06, 1.8431E+06, 1.8476E+06, 1.8520E+06, 1.8565E+06,
     + 1.8610E+06, 1.8654E+06, 1.8699E+06, 1.8744E+06, 1.8789E+06,
     +  1.8834E+06, 1.8880E+06, 1.8925E+06, 1.8970E+06, 1.9016E+06,
     + 1.9061E+06, 1.9107E+06, 1.9152E+06, 1.9198E+06, 1.9244E+06,
     +  1.9290E+06, 1.9336E+06, 1.9382E+06, 1.9428E+06, 1.9474E+06,
     + 1.9521E+06, 1.9567E+06, 1.9614E+06, 1.9660E+06, 1.9707E+06,
     +  1.9753E+06, 1.9800E+06, 1.9847E+06, 1.9894E+06, 1.9941E+06,
     + 1.9988E+06, 2.0035E+06, 2.0083E+06, 2.0130E+06, 2.0177E+06,
     +  2.0225E+06, 2.0272E+06, 2.0320E+06, 2.0368E+06, 2.0416E+06,
     + 2.0464E+06, 2.0512E+06, 2.0560E+06, 2.0608E+06, 2.0656E+06,
     +  2.0704E+06, 2.0753E+06, 2.0801E+06, 2.0850E+06, 2.0898E+06,
     + 2.0947E+06, 2.0996E+06, 2.1045E+06, 2.1093E+06, 2.1142E+06,
     +  2.1192E+06, 2.1241E+06, 2.1290E+06, 2.1339E+06, 2.1389E+06,
     + 2.1438E+06, 2.1488E+06, 2.1538E+06, 2.1587E+06, 2.1637E+06,
     +  2.1687E+06, 2.1737E+06, 2.1787E+06, 2.1837E+06, 2.1888E+06,
     + 2.1938E+06, 2.1988E+06, 2.2039E+06, 2.2089E+06, 2.2140E+06,
     +  2.2191E+06, 2.2241E+06, 2.2292E+06, 2.2343E+06, 2.2394E+06,
     + 2.2446E+06, 2.2497E+06, 2.2548E+06, 2.2599E+06, 2.2651E+06,
     +  2.2702E+06, 2.2754E+06, 2.2806E+06, 2.2858E+06, 2.2910E+06,
     + 2.2961E+06, 2.3014E+06, 2.3066E+06, 2.3118E+06, 2.3170E+06,
     +  2.3223E+06, 2.3275E+06, 2.3328E+06, 2.3380E+06, 2.3433E+06,
     + 2.3486E+06, 2.3539E+06, 2.3592E+06, 2.3645E+06, 2.3698E+06,
     +  2.3751E+06, 2.3804E+06, 2.3858E+06, 2.3911E+06, 2.3965E+06,
     + 2.4018E+06, 2.4072E+06, 2.4126E+06, 2.4180E+06, 2.4234E+06,
     +  2.4288E+06, 2.4342E+06, 2.4396E+06, 2.4450E+06, 2.4505E+06,
     + 2.4559E+06, 2.4614E+06, 2.4668E+06, 2.4723E+06, 2.4778E+06,
     +  2.4833E+06, 2.4888E+06, 2.4943E+06, 2.4998E+06, 2.5053E+06,
     + 2.5109E+06, 2.5164E+06, 2.5219E+06, 2.5275E+06, 2.5331E+06,
     +  2.5386E+06, 2.5442E+06, 2.5498E+06, 2.5554E+06, 2.5610E+06,
     + 2.5666E+06, 2.5723E+06, 2.5779E+06, 2.5835E+06, 2.5892E+06,
     +  2.5948E+06, 2.6005E+06, 2.6062E+06, 2.6119E+06, 2.6176E+06,
     + 2.6233E+06, 2.6290E+06, 2.6347E+06, 2.6404E+06, 2.6462E+06,
     +  2.6519E+06, 2.6577E+06, 2.6634E+06, 2.6692E+06, 2.6750E+06,
     + 2.6808E+06, 2.6866E+06, 2.6924E+06, 2.6982E+06, 2.7040E+06,
     +  2.7098E+06, 2.7157E+06, 2.7215E+06, 2.7274E+06, 2.7333E+06,
     + 2.7391E+06, 2.7450E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 8,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       727
      data (QofT( 9,J),J=1,2408)/ 4.1244E+01, 7.4487E+01, 1.0843E+02,
     +  1.4245E+02, 1.7648E+02, 2.1053E+02, 2.4459E+02, 2.7865E+02,
     + 3.1271E+02, 3.4678E+02, 3.8084E+02, 4.1491E+02, 4.4898E+02,
     +  4.8305E+02, 5.1711E+02, 5.5118E+02, 5.8525E+02, 6.1932E+02,
     + 6.5339E+02, 6.8747E+02, 7.5561E+02, 8.2375E+02, 8.9189E+02,
     +  9.6004E+02, 1.0282E+03, 1.0963E+03, 1.1645E+03, 1.2326E+03,
     + 1.3008E+03, 1.3689E+03, 1.4371E+03, 1.5052E+03, 1.5734E+03,
     +  1.6415E+03, 1.7097E+03, 1.7778E+03, 1.8460E+03, 1.9141E+03,
     + 1.9823E+03, 2.0504E+03, 2.1186E+03, 2.1867E+03, 2.2549E+03,
     +  2.3230E+03, 2.3912E+03, 2.4593E+03, 2.5275E+03, 2.5956E+03,
     + 2.6638E+03, 2.7320E+03, 2.8001E+03, 2.8683E+03, 2.9365E+03,
     +  3.0047E+03, 3.0729E+03, 3.1411E+03, 3.2093E+03, 3.2775E+03,
     + 3.3457E+03, 3.4140E+03, 3.4823E+03, 3.5506E+03, 3.6189E+03,
     +  3.6872E+03, 3.7556E+03, 3.8240E+03, 3.8924E+03, 3.9609E+03,
     + 4.0294E+03, 4.0980E+03, 4.1666E+03, 4.2353E+03, 4.3040E+03,
     +  4.3728E+03, 4.4417E+03, 4.5107E+03, 4.5797E+03, 4.6488E+03,
     + 4.7180E+03, 4.7873E+03, 4.8567E+03, 4.9262E+03, 4.9958E+03,
     +  5.0655E+03, 5.1354E+03, 5.2054E+03, 5.2755E+03, 5.3458E+03,
     + 5.4162E+03, 5.4867E+03, 5.5575E+03, 5.6284E+03, 5.6994E+03,
     +  5.7707E+03, 5.8421E+03, 5.9137E+03, 5.9855E+03, 6.0576E+03,
     + 6.1298E+03, 6.2022E+03, 6.2749E+03, 6.3478E+03, 6.4209E+03,
     +  6.4943E+03, 6.5679E+03, 6.6418E+03, 6.7159E+03, 6.7903E+03,
     + 6.8650E+03, 6.9399E+03, 7.0152E+03, 7.0907E+03, 7.1665E+03,
     +  7.2426E+03, 7.3190E+03, 7.3958E+03, 7.4728E+03, 7.5502E+03,
     + 7.6279E+03, 7.7060E+03, 7.7843E+03, 7.8631E+03, 7.9421E+03,
     +  8.0216E+03, 8.1014E+03, 8.1816E+03, 8.2621E+03, 8.3430E+03,
     + 8.4243E+03, 8.5060E+03, 8.5881E+03, 8.6706E+03, 8.7535E+03,
     +  8.8368E+03, 8.9205E+03, 9.0046E+03, 9.0892E+03, 9.1742E+03,
     + 9.2596E+03, 9.3455E+03, 9.4318E+03, 9.5185E+03, 9.6058E+03,
     +  9.6934E+03, 9.7816E+03, 9.8702E+03, 9.9592E+03, 1.0049E+04,
     + 1.0139E+04, 1.0229E+04, 1.0320E+04, 1.0412E+04, 1.0504E+04,
     +  1.0596E+04, 1.0689E+04, 1.0783E+04, 1.0877E+04, 1.0972E+04,
     + 1.1067E+04, 1.1162E+04, 1.1258E+04, 1.1355E+04, 1.1452E+04,
     +  1.1550E+04, 1.1649E+04, 1.1748E+04, 1.1847E+04, 1.1947E+04,
     + 1.2048E+04, 1.2149E+04, 1.2251E+04, 1.2353E+04, 1.2456E+04,
     +  1.2559E+04, 1.2664E+04, 1.2768E+04, 1.2874E+04, 1.2979E+04,
     + 1.3086E+04, 1.3193E+04, 1.3301E+04, 1.3409E+04, 1.3518E+04,
     +  1.3628E+04, 1.3738E+04, 1.3849E+04, 1.3961E+04, 1.4073E+04,
     + 1.4186E+04, 1.4300E+04, 1.4414E+04, 1.4529E+04, 1.4644E+04,
     +  1.4760E+04, 1.4877E+04, 1.4995E+04, 1.5113E+04, 1.5232E+04,
     + 1.5352E+04, 1.5472E+04, 1.5593E+04, 1.5715E+04, 1.5837E+04,
     +  1.5961E+04, 1.6085E+04, 1.6209E+04, 1.6335E+04, 1.6461E+04,
     + 1.6587E+04, 1.6715E+04, 1.6843E+04, 1.6972E+04, 1.7102E+04,
     +  1.7233E+04, 1.7364E+04, 1.7496E+04, 1.7629E+04, 1.7763E+04,
     + 1.7897E+04, 1.8032E+04, 1.8168E+04, 1.8305E+04, 1.8442E+04,
     +  1.8581E+04, 1.8720E+04, 1.8860E+04, 1.9001E+04, 1.9142E+04,
     + 1.9285E+04, 1.9428E+04, 1.9572E+04, 1.9717E+04, 1.9863E+04,
     +  2.0009E+04, 2.0156E+04, 2.0305E+04, 2.0454E+04, 2.0604E+04,
     + 2.0755E+04, 2.0906E+04, 2.1059E+04, 2.1212E+04, 2.1367E+04,
     +  2.1522E+04, 2.1678E+04, 2.1835E+04, 2.1993E+04, 2.2151E+04,
     + 2.2311E+04, 2.2472E+04, 2.2633E+04, 2.2796E+04, 2.2959E+04,
     +  2.3123E+04, 2.3288E+04, 2.3455E+04, 2.3622E+04, 2.3790E+04,
     + 2.3959E+04, 2.4129E+04, 2.4299E+04, 2.4471E+04, 2.4644E+04,
     +  2.4818E+04, 2.4993E+04, 2.5168E+04, 2.5345E+04, 2.5523E+04,
     + 2.5702E+04, 2.5881E+04, 2.6062E+04, 2.6244E+04, 2.6427E+04,
     +  2.6610E+04, 2.6795E+04, 2.6981E+04, 2.7168E+04, 2.7356E+04,
     + 2.7545E+04, 2.7735E+04, 2.7926E+04, 2.8118E+04, 2.8311E+04,
     +  2.8505E+04, 2.8700E+04, 2.8897E+04, 2.9094E+04, 2.9293E+04,
     + 2.9492E+04, 2.9693E+04, 2.9895E+04, 3.0098E+04, 3.0302E+04,
     +  3.0507E+04, 3.0713E+04, 3.0921E+04, 3.1129E+04, 3.1339E+04,
     + 3.1550E+04, 3.1762E+04, 3.1975E+04, 3.2189E+04, 3.2404E+04,
     +  3.2621E+04, 3.2839E+04, 3.3058E+04, 3.3278E+04, 3.3499E+04,
     + 3.3722E+04, 3.3945E+04, 3.4170E+04, 3.4396E+04, 3.4624E+04,
     +  3.4852E+04, 3.5082E+04, 3.5313E+04, 3.5545E+04, 3.5778E+04,
     + 3.6013E+04, 3.6249E+04, 3.6486E+04, 3.6725E+04, 3.6964E+04,
     +  3.7205E+04, 3.7448E+04, 3.7691E+04, 3.7936E+04, 3.8182E+04,
     + 3.8430E+04, 3.8678E+04, 3.8928E+04, 3.9180E+04, 3.9432E+04,
     +  3.9686E+04, 3.9942E+04, 4.0198E+04, 4.0457E+04, 4.0716E+04,
     + 4.0977E+04, 4.1239E+04, 4.1502E+04, 4.1767E+04, 4.2033E+04,
     +  4.2301E+04, 4.2570E+04, 4.2840E+04, 4.3112E+04, 4.3385E+04,
     + 4.3659E+04, 4.3935E+04, 4.4213E+04, 4.4492E+04, 4.4772E+04,
     +  4.5054E+04, 4.5337E+04, 4.5621E+04, 4.5908E+04, 4.6195E+04,
     + 4.6484E+04, 4.6775E+04, 4.7067E+04, 4.7360E+04, 4.7655E+04,
     +  4.7951E+04, 4.8249E+04, 4.8549E+04, 4.8850E+04, 4.9152E+04,
     + 4.9456E+04, 4.9762E+04, 5.0069E+04, 5.0378E+04, 5.0688E+04,
     +  5.1000E+04, 5.1313E+04, 5.1628E+04, 5.1944E+04, 5.2262E+04,
     + 5.2582E+04, 5.2903E+04, 5.3226E+04, 5.3551E+04, 5.3877E+04,
     +  5.4205E+04, 5.4534E+04, 5.4865E+04, 5.5198E+04, 5.5532E+04,
     + 5.5868E+04, 5.6205E+04, 5.6545E+04, 5.6886E+04, 5.7228E+04,
     +  5.7573E+04, 5.7919E+04, 5.8266E+04, 5.8616E+04, 5.8967E+04,
     + 5.9320E+04, 5.9675E+04, 6.0031E+04, 6.0389E+04, 6.0749E+04,
     +  6.1110E+04, 6.1474E+04, 6.1839E+04, 6.2206E+04, 6.2574E+04,
     + 6.2945E+04, 6.3317E+04, 6.3691E+04, 6.4067E+04, 6.4445E+04,
     +  6.4824E+04, 6.5206E+04, 6.5589E+04, 6.5974E+04, 6.6361E+04,
     + 6.6749E+04, 6.7140E+04, 6.7533E+04, 6.7927E+04, 6.8323E+04,
     +  6.8721E+04, 6.9121E+04, 6.9523E+04, 6.9927E+04, 7.0333E+04,
     + 7.0741E+04, 7.1150E+04, 7.1562E+04, 7.1975E+04, 7.2391E+04,
     +  7.2808E+04, 7.3228E+04, 7.3649E+04, 7.4073E+04, 7.4498E+04,
     + 7.4925E+04, 7.5355E+04, 7.5786E+04, 7.6220E+04, 7.6655E+04,
     +  7.7093E+04, 7.7532E+04, 7.7974E+04, 7.8418E+04, 7.8863E+04,
     + 7.9311E+04, 7.9761E+04, 8.0213E+04, 8.0667E+04, 8.1124E+04,
     +  8.1582E+04, 8.2042E+04, 8.2505E+04, 8.2970E+04, 8.3437E+04,
     + 8.3906E+04, 8.4377E+04, 8.4850E+04, 8.5326E+04, 8.5803E+04,
     +  8.6283E+04, 8.6766E+04, 8.7250E+04, 8.7736E+04, 8.8225E+04,
     + 8.8716E+04, 8.9209E+04, 8.9705E+04, 9.0203E+04, 9.0703E+04,
     +  9.1205E+04, 9.1710E+04, 9.2216E+04, 9.2726E+04, 9.3237E+04,
     + 9.3751E+04, 9.4267E+04, 9.4785E+04, 9.5306E+04, 9.5829E+04,
     +  9.6355E+04, 9.6882E+04, 9.7413E+04, 9.7945E+04, 9.8480E+04,
     + 9.9018E+04, 9.9557E+04, 1.0010E+05, 1.0064E+05, 1.0119E+05,
     +  1.0174E+05, 1.0229E+05, 1.0285E+05, 1.0340E+05, 1.0396E+05,
     + 1.0452E+05, 1.0509E+05, 1.0566E+05, 1.0622E+05, 1.0680E+05,
     +  1.0737E+05, 1.0795E+05, 1.0853E+05, 1.0911E+05, 1.0969E+05,
     + 1.1028E+05, 1.1087E+05, 1.1146E+05, 1.1206E+05, 1.1265E+05,
     +  1.1325E+05, 1.1386E+05, 1.1446E+05, 1.1507E+05, 1.1568E+05,
     + 1.1629E+05, 1.1691E+05, 1.1753E+05, 1.1815E+05, 1.1877E+05,
     +  1.1940E+05, 1.2003E+05, 1.2066E+05, 1.2130E+05, 1.2193E+05,
     + 1.2257E+05, 1.2322E+05, 1.2386E+05, 1.2451E+05, 1.2516E+05,
     +  1.2582E+05, 1.2647E+05, 1.2713E+05, 1.2780E+05, 1.2846E+05,
     + 1.2913E+05, 1.2980E+05, 1.3047E+05, 1.3115E+05, 1.3183E+05,
     +  1.3251E+05, 1.3320E+05, 1.3389E+05, 1.3458E+05, 1.3527E+05,
     + 1.3597E+05, 1.3667E+05, 1.3737E+05, 1.3808E+05, 1.3879E+05,
     +  1.3950E+05, 1.4022E+05, 1.4093E+05, 1.4165E+05, 1.4238E+05,
     + 1.4311E+05, 1.4384E+05, 1.4457E+05, 1.4530E+05, 1.4604E+05,
     +  1.4679E+05, 1.4753E+05, 1.4828E+05, 1.4903E+05, 1.4979E+05,
     + 1.5054E+05, 1.5131E+05, 1.5207E+05, 1.5284E+05, 1.5361E+05,
     +  1.5438E+05, 1.5516E+05, 1.5594E+05, 1.5672E+05, 1.5751E+05,
     + 1.5830E+05, 1.5909E+05, 1.5989E+05, 1.6069E+05, 1.6149E+05,
     +  1.6229E+05, 1.6310E+05, 1.6392E+05, 1.6473E+05, 1.6555E+05,
     + 1.6637E+05, 1.6720E+05, 1.6803E+05, 1.6886E+05, 1.6970E+05,
     +  1.7054E+05, 1.7138E+05, 1.7223E+05, 1.7308E+05, 1.7393E+05,
     + 1.7479E+05, 1.7565E+05, 1.7651E+05, 1.7738E+05, 1.7825E+05,
     +  1.7912E+05, 1.8000E+05, 1.8088E+05, 1.8176E+05, 1.8265E+05,
     + 1.8354E+05, 1.8444E+05, 1.8534E+05, 1.8624E+05, 1.8714E+05,
     +  1.8805E+05, 1.8897E+05, 1.8988E+05, 1.9080E+05, 1.9173E+05,
     + 1.9266E+05, 1.9359E+05, 1.9452E+05, 1.9546E+05, 1.9640E+05,
     +  1.9735E+05, 1.9830E+05, 1.9925E+05, 2.0021E+05, 2.0117E+05,
     + 2.0214E+05, 2.0310E+05, 2.0408E+05, 2.0505E+05, 2.0603E+05,
     +  2.0702E+05, 2.0800E+05, 2.0900E+05, 2.0999E+05, 2.1099E+05,
     + 2.1199E+05, 2.1300E+05, 2.1401E+05, 2.1503E+05, 2.1605E+05,
     +  2.1707E+05, 2.1810E+05, 2.1913E+05, 2.2016E+05, 2.2120E+05,
     + 2.2224E+05, 2.2329E+05, 2.2434E+05, 2.2540E+05, 2.2645E+05,
     +  2.2752E+05, 2.2858E+05, 2.2965E+05, 2.3073E+05, 2.3181E+05,
     + 2.3289E+05, 2.3398E+05, 2.3507E+05, 2.3617E+05, 2.3727E+05,
     +  2.3837E+05, 2.3948E+05, 2.4059E+05, 2.4171E+05, 2.4283E+05,
     + 2.4396E+05, 2.4509E+05, 2.4622E+05, 2.4736E+05, 2.4850E+05,
     +  2.4965E+05, 2.5080E+05, 2.5195E+05, 2.5311E+05, 2.5428E+05,
     + 2.5545E+05, 2.5662E+05, 2.5780E+05, 2.5898E+05, 2.6016E+05,
     +  2.6136E+05, 2.6255E+05, 2.6375E+05, 2.6495E+05, 2.6616E+05,
     + 2.6737E+05, 2.6859E+05, 2.6981E+05, 2.7104E+05, 2.7227E+05,
     +  2.7351E+05, 2.7475E+05, 2.7599E+05, 2.7724E+05, 2.7850E+05,
     + 2.7975E+05, 2.8102E+05, 2.8229E+05, 2.8356E+05, 2.8484E+05,
     +  2.8612E+05, 2.8740E+05, 2.8870E+05, 2.8999E+05, 2.9129E+05,
     + 2.9260E+05, 2.9391E+05, 2.9522E+05, 2.9654E+05, 2.9787E+05,
     +  2.9920E+05, 3.0053E+05, 3.0187E+05, 3.0322E+05, 3.0456E+05,
     + 3.0592E+05, 3.0728E+05, 3.0864E+05, 3.1001E+05, 3.1138E+05,
     +  3.1276E+05, 3.1415E+05, 3.1553E+05, 3.1693E+05, 3.1833E+05,
     + 3.1973E+05, 3.2114E+05, 3.2255E+05, 3.2397E+05, 3.2540E+05,
     +  3.2682E+05, 3.2826E+05, 3.2970E+05, 3.3114E+05, 3.3259E+05,
     + 3.3405E+05, 3.3551E+05, 3.3697E+05, 3.3844E+05, 3.3992E+05,
     +  3.4140E+05, 3.4288E+05, 3.4438E+05, 3.4587E+05, 3.4737E+05,
     + 3.4888E+05, 3.5039E+05, 3.5191E+05, 3.5344E+05, 3.5496E+05,
     +  3.5650E+05, 3.5804E+05, 3.5958E+05, 3.6113E+05, 3.6269E+05,
     + 3.6425E+05, 3.6582E+05, 3.6739E+05, 3.6897E+05, 3.7055E+05,
     +  3.7214E+05, 3.7373E+05, 3.7533E+05, 3.7694E+05, 3.7855E+05,
     + 3.8016E+05, 3.8179E+05, 3.8341E+05, 3.8505E+05, 3.8669E+05,
     +  3.8833E+05, 3.8998E+05, 3.9164E+05, 3.9330E+05, 3.9497E+05,
     + 3.9664E+05, 3.9832E+05, 4.0001E+05, 4.0170E+05, 4.0339E+05,
     +  4.0509E+05, 4.0680E+05, 4.0852E+05, 4.1024E+05, 4.1196E+05,
     + 4.1369E+05, 4.1543E+05, 4.1718E+05, 4.1893E+05, 4.2068E+05,
     +  4.2244E+05, 4.2421E+05, 4.2598E+05, 4.2776E+05, 4.2955E+05,
     + 4.3134E+05, 4.3314E+05, 4.3494E+05, 4.3675E+05, 4.3857E+05,
     +  4.4039E+05, 4.4222E+05, 4.4405E+05, 4.4589E+05, 4.4774E+05,
     + 4.4959E+05, 4.5145E+05, 4.5332E+05, 4.5519E+05, 4.5707E+05,
     +  4.5895E+05, 4.6084E+05, 4.6274E+05, 4.6465E+05, 4.6656E+05,
     + 4.6847E+05, 4.7039E+05, 4.7232E+05, 4.7426E+05, 4.7620E+05,
     +  4.7815E+05, 4.8010E+05, 4.8207E+05, 4.8403E+05, 4.8601E+05,
     + 4.8799E+05, 4.8998E+05, 4.9197E+05, 4.9397E+05, 4.9598E+05,
     +  4.9799E+05, 5.0001E+05, 5.0204E+05, 5.0407E+05, 5.0611E+05,
     + 5.0816E+05, 5.1022E+05, 5.1228E+05, 5.1434E+05, 5.1642E+05,
     +  5.1850E+05, 5.2059E+05, 5.2268E+05, 5.2478E+05, 5.2689E+05,
     + 5.2901E+05, 5.3113E+05, 5.3326E+05, 5.3539E+05, 5.3754E+05,
     +  5.3969E+05, 5.4184E+05, 5.4401E+05, 5.4618E+05, 5.4836E+05,
     + 5.5054E+05, 5.5273E+05, 5.5493E+05, 5.5714E+05, 5.5935E+05,
     +  5.6157E+05, 5.6380E+05, 5.6603E+05, 5.6828E+05, 5.7053E+05,
     + 5.7278E+05, 5.7505E+05, 5.7732E+05, 5.7959E+05, 5.8188E+05,
     +  5.8417E+05, 5.8647E+05, 5.8878E+05, 5.9109E+05, 5.9342E+05,
     + 5.9575E+05, 5.9808E+05, 6.0043E+05, 6.0278E+05, 6.0514E+05,
     +  6.0751E+05, 6.0988E+05, 6.1226E+05, 6.1465E+05, 6.1705E+05,
     + 6.1945E+05, 6.2186E+05, 6.2428E+05, 6.2671E+05, 6.2915E+05,
     +  6.3159E+05, 6.3404E+05, 6.3650E+05, 6.3896E+05, 6.4144E+05,
     + 6.4392E+05, 6.4641E+05, 6.4890E+05, 6.5141E+05, 6.5392E+05,
     +  6.5644E+05, 6.5897E+05, 6.6151E+05, 6.6405E+05, 6.6660E+05,
     + 6.6916E+05, 6.7173E+05, 6.7431E+05, 6.7689E+05, 6.7948E+05,
     +  6.8208E+05, 6.8469E+05, 6.8730E+05, 6.8993E+05, 6.9256E+05,
     + 6.9520E+05, 6.9785E+05, 7.0051E+05, 7.0317E+05, 7.0584E+05,
     +  7.0853E+05, 7.1122E+05, 7.1391E+05, 7.1662E+05, 7.1933E+05,
     + 7.2206E+05, 7.2479E+05, 7.2753E+05, 7.3027E+05, 7.3303E+05,
     +  7.3580E+05, 7.3857E+05, 7.4135E+05, 7.4414E+05, 7.4694E+05,
     + 7.4975E+05, 7.5256E+05, 7.5538E+05, 7.5822E+05, 7.6106E+05,
     +  7.6391E+05, 7.6677E+05, 7.6963E+05, 7.7251E+05, 7.7539E+05,
     + 7.7829E+05, 7.8119E+05, 7.8410E+05, 7.8702E+05, 7.8995E+05,
     +  7.9288E+05, 7.9583E+05, 7.9879E+05, 8.0175E+05, 8.0472E+05,
     + 8.0770E+05, 8.1069E+05, 8.1369E+05, 8.1670E+05, 8.1972E+05,
     +  8.2274E+05, 8.2578E+05, 8.2882E+05, 8.3188E+05, 8.3494E+05,
     + 8.3801E+05, 8.4109E+05, 8.4418E+05, 8.4728E+05, 8.5039E+05,
     +  8.5351E+05, 8.5663E+05, 8.5977E+05, 8.6291E+05, 8.6607E+05,
     + 8.6923E+05, 8.7241E+05, 8.7559E+05, 8.7878E+05, 8.8198E+05,
     +  8.8519E+05, 8.8841E+05, 8.9164E+05, 8.9488E+05, 8.9813E+05,
     + 9.0139E+05, 9.0466E+05, 9.0793E+05, 9.1122E+05, 9.1452E+05,
     +  9.1782E+05, 9.2114E+05, 9.2446E+05, 9.2780E+05, 9.3114E+05,
     + 9.3450E+05, 9.3786E+05, 9.4124E+05, 9.4462E+05, 9.4801E+05,
     +  9.5142E+05, 9.5483E+05, 9.5825E+05, 9.6169E+05, 9.6513E+05,
     + 9.6858E+05, 9.7205E+05, 9.7552E+05, 9.7900E+05, 9.8249E+05,
     +  9.8600E+05, 9.8951E+05, 9.9303E+05, 9.9657E+05, 1.0001E+06,
     + 1.0037E+06, 1.0072E+06, 1.0108E+06, 1.0144E+06, 1.0180E+06,
     +  1.0216E+06, 1.0252E+06, 1.0288E+06, 1.0325E+06, 1.0361E+06,
     + 1.0398E+06, 1.0434E+06, 1.0471E+06, 1.0508E+06, 1.0545E+06,
     +  1.0582E+06, 1.0619E+06, 1.0657E+06, 1.0694E+06, 1.0731E+06,
     + 1.0769E+06, 1.0807E+06, 1.0845E+06, 1.0883E+06, 1.0921E+06,
     +  1.0959E+06, 1.0997E+06, 1.1035E+06, 1.1074E+06, 1.1112E+06,
     + 1.1151E+06, 1.1190E+06, 1.1229E+06, 1.1268E+06, 1.1307E+06,
     +  1.1346E+06, 1.1386E+06, 1.1425E+06, 1.1465E+06, 1.1504E+06,
     + 1.1544E+06, 1.1584E+06, 1.1624E+06, 1.1664E+06, 1.1704E+06,
     +  1.1745E+06, 1.1785E+06, 1.1826E+06, 1.1866E+06, 1.1907E+06,
     + 1.1948E+06, 1.1989E+06, 1.2030E+06, 1.2071E+06, 1.2113E+06,
     +  1.2154E+06, 1.2196E+06, 1.2237E+06, 1.2279E+06, 1.2321E+06,
     + 1.2363E+06, 1.2405E+06, 1.2448E+06, 1.2490E+06, 1.2533E+06,
     +  1.2575E+06, 1.2618E+06, 1.2661E+06, 1.2704E+06, 1.2747E+06,
     + 1.2790E+06, 1.2833E+06, 1.2877E+06, 1.2920E+06, 1.2964E+06,
     +  1.3008E+06, 1.3052E+06, 1.3096E+06, 1.3140E+06, 1.3184E+06,
     + 1.3229E+06, 1.3273E+06, 1.3318E+06, 1.3363E+06, 1.3408E+06,
     +  1.3453E+06, 1.3498E+06, 1.3543E+06, 1.3588E+06, 1.3634E+06,
     + 1.3679E+06, 1.3725E+06, 1.3771E+06, 1.3817E+06, 1.3863E+06,
     +  1.3909E+06, 1.3956E+06, 1.4002E+06, 1.4049E+06, 1.4095E+06,
     + 1.4142E+06, 1.4189E+06, 1.4236E+06, 1.4284E+06, 1.4331E+06,
     +  1.4378E+06, 1.4426E+06, 1.4474E+06, 1.4522E+06, 1.4570E+06,
     + 1.4618E+06, 1.4666E+06, 1.4714E+06, 1.4763E+06, 1.4811E+06,
     +  1.4860E+06, 1.4909E+06, 1.4958E+06, 1.5007E+06, 1.5057E+06,
     + 1.5106E+06, 1.5155E+06, 1.5205E+06, 1.5255E+06, 1.5305E+06,
     +  1.5355E+06, 1.5405E+06, 1.5455E+06, 1.5506E+06, 1.5556E+06,
     + 1.5607E+06, 1.5658E+06, 1.5709E+06, 1.5760E+06, 1.5811E+06,
     +  1.5863E+06, 1.5914E+06, 1.5966E+06, 1.6018E+06, 1.6070E+06,
     + 1.6122E+06, 1.6174E+06, 1.6226E+06, 1.6279E+06, 1.6331E+06,
     +  1.6384E+06, 1.6437E+06, 1.6490E+06, 1.6543E+06, 1.6596E+06,
     + 1.6650E+06, 1.6703E+06, 1.6757E+06, 1.6811E+06, 1.6865E+06,
     +  1.6919E+06, 1.6973E+06, 1.7027E+06, 1.7082E+06, 1.7137E+06,
     + 1.7191E+06, 1.7246E+06, 1.7301E+06, 1.7357E+06, 1.7412E+06,
     +  1.7468E+06, 1.7523E+06, 1.7579E+06, 1.7635E+06, 1.7691E+06,
     + 1.7747E+06, 1.7804E+06, 1.7860E+06, 1.7917E+06, 1.7974E+06,
     +  1.8031E+06, 1.8088E+06, 1.8145E+06, 1.8202E+06, 1.8260E+06,
     + 1.8317E+06, 1.8375E+06, 1.8433E+06, 1.8491E+06, 1.8550E+06,
     +  1.8608E+06, 1.8667E+06, 1.8725E+06, 1.8784E+06, 1.8843E+06,
     + 1.8902E+06, 1.8962E+06, 1.9021E+06, 1.9081E+06, 1.9140E+06,
     +  1.9200E+06, 1.9260E+06, 1.9320E+06, 1.9381E+06, 1.9441E+06,
     + 1.9502E+06, 1.9563E+06, 1.9624E+06, 1.9685E+06, 1.9746E+06,
     +  1.9807E+06, 1.9869E+06, 1.9931E+06, 1.9992E+06, 2.0054E+06,
     + 2.0117E+06, 2.0179E+06, 2.0241E+06, 2.0304E+06, 2.0367E+06,
     +  2.0430E+06, 2.0493E+06, 2.0556E+06, 2.0620E+06, 2.0683E+06,
     + 2.0747E+06, 2.0811E+06, 2.0875E+06, 2.0939E+06, 2.1003E+06,
     +  2.1068E+06, 2.1132E+06, 2.1197E+06, 2.1262E+06, 2.1327E+06,
     + 2.1393E+06, 2.1458E+06, 2.1524E+06, 2.1590E+06, 2.1656E+06,
     +  2.1722E+06, 2.1788E+06, 2.1854E+06, 2.1921E+06, 2.1988E+06,
     + 2.2055E+06, 2.2122E+06, 2.2189E+06, 2.2256E+06, 2.2324E+06,
     +  2.2392E+06, 2.2460E+06, 2.2528E+06, 2.2596E+06, 2.2664E+06,
     + 2.2733E+06, 2.2802E+06, 2.2870E+06, 2.2940E+06, 2.3009E+06,
     +  2.3078E+06, 2.3148E+06, 2.3217E+06, 2.3287E+06, 2.3357E+06,
     + 2.3428E+06, 2.3498E+06, 2.3569E+06, 2.3639E+06, 2.3710E+06,
     +  2.3781E+06, 2.3853E+06, 2.3924E+06, 2.3996E+06, 2.4067E+06,
     + 2.4139E+06, 2.4211E+06, 2.4284E+06, 2.4356E+06, 2.4429E+06,
     +  2.4501E+06, 2.4574E+06, 2.4648E+06, 2.4721E+06, 2.4794E+06,
     + 2.4868E+06, 2.4942E+06, 2.5016E+06, 2.5090E+06, 2.5164E+06,
     +  2.5239E+06, 2.5314E+06, 2.5388E+06, 2.5464E+06, 2.5539E+06,
     + 2.5614E+06, 2.5690E+06, 2.5766E+06, 2.5842E+06, 2.5918E+06,
     +  2.5994E+06, 2.6070E+06, 2.6147E+06, 2.6224E+06, 2.6301E+06,
     + 2.6378E+06, 2.6456E+06, 2.6533E+06, 2.6611E+06, 2.6689E+06,
     +  2.6767E+06, 2.6845E+06, 2.6924E+06, 2.7002E+06, 2.7081E+06,
     + 2.7160E+06, 2.7240E+06, 2.7319E+06, 2.7399E+06, 2.7478E+06,
     +  2.7558E+06, 2.7638E+06, 2.7719E+06, 2.7799E+06, 2.7880E+06,
     + 2.7961E+06, 2.8042E+06, 2.8123E+06, 2.8205E+06, 2.8286E+06,
     +  2.8368E+06, 2.8450E+06, 2.8532E+06, 2.8615E+06, 2.8697E+06,
     + 2.8780E+06, 2.8863E+06, 2.8946E+06, 2.9029E+06, 2.9113E+06,
     +  2.9197E+06, 2.9281E+06, 2.9365E+06, 2.9449E+06, 2.9533E+06,
     + 2.9618E+06, 2.9703E+06, 2.9788E+06, 2.9873E+06, 2.9959E+06,
     +  3.0044E+06, 3.0130E+06, 3.0216E+06, 3.0303E+06, 3.0389E+06,
     + 3.0476E+06, 3.0562E+06, 3.0649E+06, 3.0737E+06, 3.0824E+06,
     +  3.0912E+06, 3.1000E+06, 3.1088E+06, 3.1176E+06, 3.1264E+06,
     + 3.1353E+06, 3.1442E+06, 3.1531E+06, 3.1620E+06, 3.1709E+06,
     +  3.1799E+06, 3.1889E+06, 3.1979E+06, 3.2069E+06, 3.2159E+06,
     + 3.2250E+06, 3.2341E+06, 3.2432E+06, 3.2523E+06, 3.2614E+06,
     +  3.2706E+06, 3.2798E+06, 3.2890E+06, 3.2982E+06, 3.3075E+06,
     + 3.3167E+06, 3.3260E+06, 3.3353E+06, 3.3447E+06, 3.3540E+06,
     +  3.3634E+06, 3.3728E+06, 3.3822E+06, 3.3916E+06, 3.4011E+06,
     + 3.4105E+06, 3.4200E+06, 3.4295E+06, 3.4391E+06, 3.4486E+06,
     +  3.4582E+06, 3.4678E+06, 3.4774E+06, 3.4871E+06, 3.4967E+06,
     + 3.5064E+06, 3.5161E+06, 3.5259E+06, 3.5356E+06, 3.5454E+06,
     +  3.5552E+06, 3.5650E+06, 3.5748E+06, 3.5847E+06, 3.5946E+06,
     + 3.6045E+06, 3.6144E+06, 3.6243E+06, 3.6343E+06, 3.6443E+06,
     +  3.6543E+06, 3.6643E+06, 3.6744E+06, 3.6844E+06, 3.6945E+06,
     + 3.7047E+06, 3.7148E+06, 3.7250E+06, 3.7351E+06, 3.7453E+06,
     +  3.7556E+06, 3.7658E+06, 3.7761E+06, 3.7864E+06, 3.7967E+06,
     + 3.8070E+06, 3.8174E+06, 3.8278E+06, 3.8382E+06, 3.8486E+06,
     +  3.8591E+06, 3.8696E+06, 3.8800E+06, 3.8906E+06, 3.9011E+06,
     + 3.9117E+06, 3.9223E+06, 3.9329E+06, 3.9435E+06, 3.9542E+06,
     +  3.9648E+06, 3.9755E+06, 3.9863E+06, 3.9970E+06, 4.0078E+06,
     + 4.0186E+06, 4.0294E+06, 4.0402E+06, 4.0511E+06, 4.0620E+06,
     +  4.0729E+06, 4.0838E+06, 4.0947E+06, 4.1057E+06, 4.1167E+06,
     + 4.1277E+06, 4.1388E+06, 4.1499E+06, 4.1610E+06, 4.1721E+06,
     +  4.1832E+06, 4.1944E+06, 4.2056E+06, 4.2168E+06, 4.2280E+06,
     + 4.2393E+06, 4.2506E+06, 4.2619E+06, 4.2732E+06, 4.2845E+06,
     +  4.2959E+06, 4.3073E+06, 4.3188E+06, 4.3302E+06, 4.3417E+06,
     + 4.3532E+06, 4.3647E+06, 4.3762E+06, 4.3878E+06, 4.3994E+06,
     +  4.4110E+06, 4.4227E+06, 4.4343E+06, 4.4460E+06, 4.4577E+06,
     + 4.4695E+06, 4.4812E+06, 4.4930E+06, 4.5048E+06, 4.5167E+06,
     +  4.5285E+06, 4.5404E+06, 4.5523E+06, 4.5643E+06, 4.5762E+06,
     + 4.5882E+06, 4.6002E+06, 4.6123E+06, 4.6243E+06, 4.6364E+06,
     +  4.6485E+06, 4.6607E+06, 4.6728E+06, 4.6850E+06, 4.6972E+06,
     + 4.7094E+06, 4.7217E+06, 4.7340E+06, 4.7463E+06, 4.7586E+06,
     +  4.7710E+06, 4.7834E+06, 4.7958E+06, 4.8082E+06, 4.8207E+06,
     + 4.8332E+06, 4.8457E+06, 4.8582E+06, 4.8708E+06, 4.8834E+06,
     +  4.8960E+06, 4.9087E+06, 4.9213E+06, 4.9340E+06, 4.9467E+06,
     + 4.9595E+06, 4.9723E+06, 4.9851E+06, 4.9979E+06, 5.0107E+06,
     +  5.0236E+06, 5.0365E+06, 5.0494E+06, 5.0624E+06, 5.0754E+06,
     + 5.0884E+06, 5.1014E+06, 5.1145E+06, 5.1275E+06, 5.1407E+06,
     +  5.1538E+06, 5.1670E+06, 5.1801E+06, 5.1934E+06, 5.2066E+06,
     + 5.2199E+06, 5.2332E+06, 5.2465E+06, 5.2598E+06, 5.2732E+06,
     +  5.2866E+06, 5.3001E+06, 5.3135E+06, 5.3270E+06, 5.3405E+06,
     + 5.3541E+06, 5.3676E+06, 5.3812E+06, 5.3948E+06, 5.4085E+06,
     +  5.4221E+06, 5.4358E+06, 5.4496E+06, 5.4633E+06, 5.4771E+06,
     + 5.4909E+06, 5.5048E+06, 5.5186E+06, 5.5325E+06, 5.5464E+06,
     +  5.5604E+06, 5.5744E+06, 5.5884E+06, 5.6024E+06, 5.6164E+06,
     + 5.6305E+06, 5.6446E+06, 5.6588E+06, 5.6730E+06, 5.6872E+06,
     +  5.7014E+06, 5.7156E+06, 5.7299E+06, 5.7442E+06, 5.7586E+06,
     + 5.7729E+06, 5.7873E+06, 5.8017E+06, 5.8162E+06, 5.8307E+06,
     +  5.8452E+06, 5.8597E+06, 5.8743E+06, 5.8889E+06, 5.9035E+06,
     + 5.9181E+06, 5.9328E+06, 5.9475E+06, 5.9622E+06, 5.9770E+06,
     +  5.9918E+06, 6.0066E+06, 6.0215E+06, 6.0363E+06, 6.0512E+06,
     + 6.0662E+06, 6.0811E+06, 6.0961E+06, 6.1112E+06, 6.1262E+06,
     +  6.1413E+06, 6.1564E+06, 6.1715E+06, 6.1867E+06, 6.2019E+06,
     + 6.2171E+06, 6.2324E+06, 6.2477E+06, 6.2630E+06, 6.2783E+06,
     +  6.2937E+06, 6.3091E+06, 6.3245E+06, 6.3400E+06, 6.3555E+06,
     + 6.3710E+06, 6.3866E+06, 6.4021E+06, 6.4177E+06, 6.4334E+06,
     +  6.4490E+06, 6.4647E+06, 6.4805E+06, 6.4962E+06, 6.5120E+06,
     + 6.5278E+06, 6.5437E+06, 6.5596E+06, 6.5755E+06, 6.5914E+06,
     +  6.6074E+06, 6.6234E+06, 6.6394E+06, 6.6555E+06, 6.6716E+06,
     + 6.6877E+06, 6.7038E+06, 6.7200E+06, 6.7362E+06, 6.7525E+06,
     +  6.7688E+06, 6.7851E+06, 6.8014E+06, 6.8178E+06, 6.8342E+06,
     + 6.8506E+06, 6.8671E+06, 6.8835E+06, 6.9001E+06, 6.9166E+06,
     +  6.9332E+06, 6.9498E+06, 6.9665E+06, 6.9831E+06, 6.9998E+06,
     + 7.0166E+06, 7.0334E+06, 7.0502E+06, 7.0670E+06, 7.0839E+06,
     +  7.1007E+06, 7.1177E+06, 7.1346E+06, 7.1516E+06, 7.1686E+06,
     + 7.1857E+06, 7.2028E+06, 7.2199E+06, 7.2370E+06, 7.2542E+06,
     +  7.2714E+06, 7.2887E+06, 7.3060E+06, 7.3233E+06, 7.3406E+06,
     + 7.3580E+06, 7.3754E+06, 7.3928E+06, 7.4103E+06, 7.4278E+06,
     +  7.4453E+06, 7.4629E+06, 7.4805E+06, 7.4981E+06, 7.5158E+06,
     + 7.5335E+06, 7.5512E+06, 7.5689E+06, 7.5867E+06, 7.6046E+06,
     +  7.6224E+06, 7.6403E+06, 7.6582E+06, 7.6762E+06, 7.6942E+06,
     + 7.7122E+06, 7.7302E+06, 7.7483E+06, 7.7664E+06, 7.7846E+06,
     +  7.8028E+06, 7.8210E+06, 7.8393E+06, 7.8575E+06, 7.8759E+06,
     + 7.8942E+06, 7.9126E+06, 7.9310E+06, 7.9495E+06, 7.9679E+06,
     +  7.9865E+06, 8.0050E+06, 8.0236E+06, 8.0422E+06, 8.0609E+06,
     + 8.0795E+06, 8.0983E+06, 8.1170E+06, 8.1358E+06, 8.1546E+06,
     +  8.1735E+06, 8.1924E+06, 8.2113E+06, 8.2302E+06, 8.2492E+06,
     + 8.2683E+06, 8.2873E+06, 8.3064E+06, 8.3255E+06, 8.3447E+06,
     +  8.3639E+06, 8.3831E+06, 8.4024E+06, 8.4217E+06, 8.4410E+06,
     + 8.4604E+06, 8.4798E+06, 8.4992E+06, 8.5187E+06, 8.5382E+06,
     +  8.5577E+06, 8.5773E+06, 8.5969E+06, 8.6166E+06, 8.6362E+06,
     + 8.6559E+06, 8.6757E+06, 8.6955E+06, 8.7153E+06, 8.7351E+06,
     +  8.7550E+06, 8.7750E+06, 8.7949E+06, 8.8149E+06, 8.8349E+06,
     + 8.8550E+06, 8.8751E+06, 8.8952E+06, 8.9154E+06, 8.9356E+06,
     +  8.9558E+06, 8.9761E+06, 8.9964E+06, 9.0168E+06, 9.0372E+06,
     + 9.0576E+06, 9.0780E+06, 9.0985E+06, 9.1190E+06, 9.1396E+06,
     +  9.1602E+06, 9.1808E+06, 9.2015E+06, 9.2222E+06, 9.2430E+06,
     + 9.2637E+06, 9.2845E+06, 9.3054E+06, 9.3263E+06, 9.3472E+06,
     +  9.3682E+06, 9.3892E+06, 9.4102E+06, 9.4313E+06, 9.4524E+06,
     + 9.4735E+06, 9.4947E+06, 9.5159E+06, 9.5371E+06, 9.5584E+06,
     +  9.5798E+06, 9.6011E+06, 9.6225E+06, 9.6439E+06, 9.6654E+06,
     + 9.6869E+06, 9.7085E+06, 9.7300E+06, 9.7517E+06, 9.7733E+06,
     +  9.7950E+06, 9.8167E+06, 9.8385E+06, 9.8603E+06, 9.8822E+06,
     + 9.9040E+06, 9.9260E+06, 9.9479E+06, 9.9699E+06, 9.9919E+06,
     +  1.0014E+07, 1.0036E+07, 1.0058E+07, 1.0080E+07, 1.0103E+07,
     + 1.0125E+07, 1.0147E+07, 1.0170E+07, 1.0192E+07, 1.0214E+07,
     +  1.0237E+07, 1.0259E+07, 1.0282E+07, 1.0304E+07, 1.0327E+07,
     + 1.0350E+07, 1.0372E+07, 1.0395E+07, 1.0418E+07, 1.0440E+07,
     +  1.0463E+07, 1.0486E+07, 1.0509E+07, 1.0532E+07, 1.0555E+07,
     + 1.0578E+07, 1.0601E+07, 1.0624E+07, 1.0647E+07, 1.0670E+07,
     +  1.0694E+07, 1.0717E+07, 1.0740E+07, 1.0763E+07, 1.0787E+07,
     + 1.0810E+07, 1.0834E+07, 1.0857E+07, 1.0881E+07, 1.0904E+07,
     +  1.0928E+07, 1.0952E+07, 1.0975E+07, 1.0999E+07, 1.1023E+07,
     + 1.1046E+07, 1.1070E+07, 1.1094E+07, 1.1118E+07, 1.1142E+07,
     +  1.1166E+07, 1.1190E+07, 1.1214E+07, 1.1238E+07, 1.1262E+07,
     + 1.1287E+07, 1.1311E+07, 1.1335E+07, 1.1359E+07, 1.1384E+07,
     +  1.1408E+07, 1.1433E+07, 1.1457E+07, 1.1482E+07, 1.1506E+07,
     + 1.1531E+07, 1.1555E+07, 1.1580E+07, 1.1605E+07, 1.1630E+07,
     +  1.1654E+07, 1.1679E+07, 1.1704E+07, 1.1729E+07, 1.1754E+07,
     + 1.1779E+07, 1.1804E+07, 1.1829E+07, 1.1854E+07, 1.1879E+07,
     +  1.1905E+07, 1.1930E+07, 1.1955E+07, 1.1980E+07, 1.2006E+07,
     + 1.2031E+07, 1.2057E+07, 1.2082E+07, 1.2108E+07, 1.2133E+07,
     +  1.2159E+07, 1.2184E+07, 1.2210E+07, 1.2236E+07, 1.2262E+07,
     + 1.2287E+07, 1.2313E+07, 1.2339E+07, 1.2365E+07, 1.2391E+07,
     +  1.2417E+07, 1.2443E+07, 1.2469E+07, 1.2496E+07, 1.2522E+07,
     + 1.2548E+07, 1.2574E+07, 1.2601E+07, 1.2627E+07, 1.2653E+07,
     +  1.2680E+07, 1.2706E+07, 1.2733E+07, 1.2759E+07, 1.2786E+07,
     + 1.2813E+07, 1.2839E+07, 1.2866E+07, 1.2893E+07, 1.2920E+07,
     +  1.2947E+07, 1.2973E+07, 1.3000E+07, 1.3027E+07, 1.3054E+07,
     + 1.3082E+07, 1.3109E+07, 1.3136E+07, 1.3163E+07, 1.3190E+07,
     +  1.3218E+07, 1.3245E+07, 1.3272E+07, 1.3300E+07, 1.3327E+07,
     + 1.3355E+07, 1.3382E+07, 1.3410E+07, 1.3438E+07, 1.3465E+07,
     +  1.3493E+07, 1.3521E+07, 1.3549E+07, 1.3576E+07, 1.3604E+07,
     + 1.3632E+07, 1.3660E+07, 1.3688E+07, 1.3716E+07, 1.3745E+07,
     +  1.3773E+07, 1.3801E+07, 1.3829E+07, 1.3857E+07, 1.3886E+07,
     + 1.3914E+07, 1.3943E+07, 1.3971E+07, 1.4000E+07, 1.4028E+07,
     +  1.4057E+07, 1.4085E+07, 1.4114E+07, 1.4143E+07, 1.4172E+07,
     + 1.4200E+07, 1.4229E+07, 1.4258E+07, 1.4287E+07, 1.4316E+07,
     +  1.4345E+07, 1.4374E+07, 1.4403E+07, 1.4433E+07, 1.4462E+07,
     + 1.4491E+07, 1.4520E+07, 1.4550E+07, 1.4579E+07, 1.4609E+07,
     +  1.4638E+07, 1.4668E+07, 1.4697E+07, 1.4727E+07, 1.4757E+07,
     + 1.4786E+07, 1.4816E+07, 1.4846E+07, 1.4876E+07, 1.4906E+07,
     +  1.4936E+07, 1.4966E+07, 1.4996E+07, 1.5026E+07, 1.5056E+07,
     + 1.5086E+07, 1.5116E+07, 1.5147E+07, 1.5177E+07, 1.5207E+07,
     +  1.5238E+07, 1.5268E+07, 1.5299E+07, 1.5329E+07, 1.5360E+07,
     + 1.5390E+07, 1.5421E+07, 1.5452E+07, 1.5483E+07, 1.5513E+07,
     +  1.5544E+07, 1.5575E+07, 1.5606E+07, 1.5637E+07, 1.5668E+07,
     + 1.5699E+07, 1.5730E+07, 1.5762E+07, 1.5793E+07, 1.5824E+07,
     +  1.5855E+07, 1.5887E+07, 1.5918E+07, 1.5950E+07, 1.5981E+07,
     + 1.6013E+07, 1.6044E+07, 1.6076E+07, 1.6108E+07, 1.6140E+07,
     +  1.6171E+07, 1.6203E+07, 1.6235E+07, 1.6267E+07, 1.6299E+07,
     + 1.6331E+07, 1.6363E+07, 1.6395E+07, 1.6427E+07, 1.6460E+07,
     +  1.6492E+07, 1.6524E+07, 1.6557E+07, 1.6589E+07, 1.6621E+07,
     + 1.6654E+07, 1.6687E+07, 1.6719E+07, 1.6752E+07, 1.6785E+07,
     +  1.6817E+07, 1.6850E+07, 1.6883E+07, 1.6916E+07, 1.6949E+07,
     + 1.6982E+07, 1.7015E+07, 1.7048E+07, 1.7081E+07, 1.7114E+07,
     +  1.7147E+07, 1.7181E+07, 1.7214E+07, 1.7247E+07, 1.7281E+07,
     + 1.7314E+07, 1.7348E+07, 1.7381E+07, 1.7415E+07, 1.7449E+07,
     +  1.7482E+07, 1.7516E+07, 1.7550E+07, 1.7584E+07, 1.7618E+07,
     + 1.7652E+07, 1.7686E+07, 1.7720E+07, 1.7754E+07, 1.7788E+07,
     +  1.7822E+07, 1.7856E+07, 1.7891E+07, 1.7925E+07, 1.7960E+07,
     + 1.7994E+07, 1.8028E+07, 1.8063E+07, 1.8098E+07, 1.8132E+07,
     +  1.8167E+07, 1.8202E+07, 1.8237E+07, 1.8271E+07, 1.8306E+07,
     + 1.8341E+07, 1.8376E+07, 1.8411E+07, 1.8446E+07, 1.8481E+07,
     +  1.8517E+07, 1.8552E+07, 1.8587E+07, 1.8623E+07, 1.8658E+07,
     + 1.8693E+07, 1.8729E+07, 1.8764E+07, 1.8800E+07, 1.8836E+07,
     +  1.8871E+07, 1.8907E+07, 1.8943E+07, 1.8979E+07, 1.9015E+07,
     + 1.9051E+07, 1.9087E+07, 1.9123E+07, 1.9159E+07, 1.9195E+07,
     +  1.9231E+07, 1.9267E+07, 1.9304E+07, 1.9340E+07, 1.9376E+07,
     + 1.9413E+07, 1.9449E+07, 1.9486E+07, 1.9523E+07, 1.9559E+07,
     +  1.9596E+07, 1.9633E+07, 1.9670E+07, 1.9706E+07, 1.9743E+07,
     + 1.9780E+07, 1.9817E+07, 1.9854E+07, 1.9891E+07, 1.9929E+07,
     +  1.9966E+07, 2.0003E+07, 2.0040E+07, 2.0078E+07, 2.0115E+07,
     + 2.0153E+07, 2.0190E+07, 2.0228E+07, 2.0266E+07, 2.0303E+07,
     +  2.0341E+07, 2.0379E+07, 2.0417E+07, 2.0454E+07, 2.0492E+07,
     + 2.0530E+07, 2.0568E+07, 2.0607E+07, 2.0645E+07, 2.0683E+07,
     +  2.0721E+07, 2.0760E+07, 2.0798E+07, 2.0836E+07, 2.0875E+07,
     + 2.0913E+07, 2.0952E+07, 2.0991E+07, 2.1029E+07, 2.1068E+07,
     +  2.1107E+07, 2.1146E+07, 2.1184E+07, 2.1223E+07, 2.1262E+07,
     + 2.1301E+07, 2.1341E+07, 2.1380E+07, 2.1419E+07, 2.1458E+07,
     +  2.1498E+07, 2.1537E+07, 2.1576E+07, 2.1616E+07, 2.1655E+07,
     + 2.1695E+07, 2.1735E+07, 2.1774E+07, 2.1814E+07, 2.1854E+07,
     +  2.1894E+07, 2.1934E+07, 2.1973E+07, 2.2013E+07, 2.2054E+07,
     + 2.2094E+07, 2.2134E+07, 2.2174E+07, 2.2214E+07, 2.2255E+07,
     +  2.2295E+07, 2.2335E+07, 2.2376E+07, 2.2417E+07, 2.2457E+07,
     + 2.2498E+07, 2.2539E+07, 2.2579E+07, 2.2620E+07, 2.2661E+07,
     +  2.2702E+07, 2.2743E+07, 2.2784E+07, 2.2825E+07, 2.2866E+07,
     + 2.2907E+07, 2.2949E+07, 2.2990E+07, 2.3031E+07, 2.3073E+07,
     +  2.3114E+07, 2.3156E+07, 2.3197E+07, 2.3239E+07, 2.3281E+07,
     + 2.3322E+07, 2.3364E+07, 2.3406E+07, 2.3448E+07, 2.3490E+07,
     +  2.3532E+07, 2.3574E+07, 2.3616E+07, 2.3658E+07, 2.3700E+07,
     + 2.3743E+07, 2.3785E+07, 2.3828E+07, 2.3870E+07, 2.3913E+07,
     +  2.3955E+07, 2.3998E+07, 2.4040E+07, 2.4083E+07, 2.4126E+07,
     + 2.4169E+07, 2.4212E+07, 2.4255E+07, 2.4298E+07, 2.4341E+07,
     +  2.4384E+07, 2.4427E+07, 2.4470E+07, 2.4514E+07, 2.4557E+07,
     + 2.4600E+07, 2.4644E+07, 2.4687E+07, 2.4731E+07, 2.4774E+07,
     +  2.4818E+07, 2.4862E+07, 2.4906E+07, 2.4950E+07, 2.4993E+07,
     + 2.5037E+07, 2.5081E+07, 2.5126E+07, 2.5170E+07, 2.5214E+07,
     +  2.5258E+07, 2.5302E+07, 2.5347E+07, 2.5391E+07, 2.5436E+07,
     + 2.5480E+07, 2.5525E+07, 2.5569E+07, 2.5614E+07, 2.5659E+07,
     +  2.5704E+07, 2.5749E+07, 2.5794E+07, 2.5839E+07, 2.5884E+07,
     + 2.5929E+07, 2.5974E+07, 2.6019E+07, 2.6064E+07, 2.6110E+07,
     +  2.6155E+07, 2.6200E+07, 2.6246E+07, 2.6292E+07, 2.6337E+07,
     + 2.6383E+07, 2.6429E+07, 2.6474E+07, 2.6520E+07, 2.6566E+07,
     +  2.6612E+07, 2.6658E+07, 2.6704E+07, 2.6750E+07, 2.6796E+07,
     + 2.6843E+07, 2.6889E+07, 2.6935E+07, 2.6982E+07, 2.7028E+07,
     +  2.7075E+07, 2.7121E+07, 2.7168E+07, 2.7215E+07, 2.7262E+07,
     + 2.7308E+07, 2.7355E+07, 2.7402E+07, 2.7449E+07, 2.7496E+07,
     +  2.7543E+07, 2.7591E+07, 2.7638E+07, 2.7685E+07, 2.7732E+07,
     + 2.7780E+07, 2.7827E+07, 2.7875E+07, 2.7922E+07, 2.7970E+07,
     +  2.8018E+07, 2.8066E+07, 2.8113E+07, 2.8161E+07, 2.8209E+07,
     + 2.8257E+07, 2.8305E+07, 2.8353E+07, 2.8402E+07, 2.8450E+07,
     +  2.8498E+07, 2.8546E+07, 2.8595E+07, 2.8643E+07, 2.8692E+07,
     + 2.8740E+07, 2.8789E+07, 2.8838E+07, 2.8887E+07, 2.8935E+07,
     +  2.8984E+07, 2.9033E+07, 2.9082E+07, 2.9131E+07, 2.9180E+07,
     + 2.9230E+07, 2.9279E+07, 2.9328E+07, 2.9377E+07, 2.9427E+07,
     +  2.9476E+07, 2.9526E+07, 2.9576E+07, 2.9625E+07, 2.9675E+07/
      data (QofT( 9,J),J=2409,4510)/ 2.9725E+07, 2.9775E+07,
     +  2.9824E+07, 2.9874E+07, 2.9924E+07, 2.9974E+07, 3.0025E+07,
     +  3.0075E+07,
     + 3.0125E+07, 3.0175E+07, 3.0226E+07, 3.0276E+07, 3.0327E+07,
     +  3.0377E+07, 3.0428E+07, 3.0479E+07, 3.0529E+07, 3.0580E+07,
     + 3.0631E+07, 3.0682E+07, 3.0733E+07, 3.0784E+07, 3.0835E+07,
     +  3.0886E+07, 3.0937E+07, 3.0989E+07, 3.1040E+07, 3.1092E+07,
     + 3.1143E+07, 3.1194E+07, 3.1246E+07, 3.1298E+07, 3.1349E+07,
     +  3.1401E+07, 3.1453E+07, 3.1505E+07, 3.1557E+07, 3.1609E+07,
     + 3.1661E+07, 3.1713E+07, 3.1765E+07, 3.1818E+07, 3.1870E+07,
     +  3.1922E+07, 3.1975E+07, 3.2027E+07, 3.2080E+07, 3.2133E+07,
     + 3.2185E+07, 3.2238E+07, 3.2291E+07, 3.2344E+07, 3.2397E+07,
     +  3.2450E+07, 3.2503E+07, 3.2556E+07, 3.2609E+07, 3.2662E+07,
     + 3.2716E+07, 3.2769E+07, 3.2822E+07, 3.2876E+07, 3.2929E+07,
     +  3.2983E+07, 3.3037E+07, 3.3090E+07, 3.3144E+07, 3.3198E+07,
     + 3.3252E+07, 3.3306E+07, 3.3360E+07, 3.3414E+07, 3.3468E+07,
     +  3.3523E+07, 3.3577E+07, 3.3631E+07, 3.3686E+07, 3.3740E+07,
     + 3.3795E+07, 3.3849E+07, 3.3904E+07, 3.3959E+07, 3.4014E+07,
     +  3.4068E+07, 3.4123E+07, 3.4178E+07, 3.4233E+07, 3.4289E+07,
     + 3.4344E+07, 3.4399E+07, 3.4454E+07, 3.4510E+07, 3.4565E+07,
     +  3.4621E+07, 3.4676E+07, 3.4732E+07, 3.4787E+07, 3.4843E+07,
     + 3.4899E+07, 3.4955E+07, 3.5011E+07, 3.5067E+07, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       838
      data (QofT(10,J),J=1,2408)/ 2.5017E+00, 4.3612E+00, 6.3569E+00,
     +  8.3579E+00, 1.0360E+01, 1.2363E+01, 1.4366E+01, 1.6370E+01,
     + 1.8373E+01, 2.0377E+01, 2.2381E+01, 2.4385E+01, 2.6388E+01,
     +  2.8392E+01, 3.0396E+01, 3.2400E+01, 3.4404E+01, 3.6408E+01,
     + 3.8412E+01, 4.0416E+01, 4.4424E+01, 4.8432E+01, 5.2440E+01,
     +  5.6449E+01, 6.0457E+01, 6.4465E+01, 6.8473E+01, 7.2482E+01,
     + 7.6490E+01, 8.0498E+01, 8.4507E+01, 8.8515E+01, 9.2523E+01,
     +  9.6532E+01, 1.0054E+02, 1.0455E+02, 1.0856E+02, 1.1257E+02,
     + 1.1657E+02, 1.2058E+02, 1.2459E+02, 1.2860E+02, 1.3261E+02,
     +  1.3662E+02, 1.4063E+02, 1.4464E+02, 1.4864E+02, 1.5265E+02,
     + 1.5666E+02, 1.6067E+02, 1.6468E+02, 1.6869E+02, 1.7270E+02,
     +  1.7672E+02, 1.8073E+02, 1.8474E+02, 1.8875E+02, 1.9277E+02,
     + 1.9678E+02, 2.0080E+02, 2.0482E+02, 2.0884E+02, 2.1286E+02,
     +  2.1688E+02, 2.2091E+02, 2.2494E+02, 2.2897E+02, 2.3300E+02,
     + 2.3704E+02, 2.4108E+02, 2.4512E+02, 2.4917E+02, 2.5322E+02,
     +  2.5728E+02, 2.6134E+02, 2.6541E+02, 2.6948E+02, 2.7356E+02,
     + 2.7765E+02, 2.8174E+02, 2.8584E+02, 2.8994E+02, 2.9406E+02,
     +  2.9818E+02, 3.0231E+02, 3.0645E+02, 3.1060E+02, 3.1476E+02,
     + 3.1893E+02, 3.2311E+02, 3.2730E+02, 3.3150E+02, 3.3571E+02,
     +  3.3994E+02, 3.4418E+02, 3.4843E+02, 3.5269E+02, 3.5697E+02,
     + 3.6126E+02, 3.6557E+02, 3.6989E+02, 3.7423E+02, 3.7858E+02,
     +  3.8295E+02, 3.8733E+02, 3.9173E+02, 3.9615E+02, 4.0059E+02,
     + 4.0504E+02, 4.0951E+02, 4.1400E+02, 4.1851E+02, 4.2304E+02,
     +  4.2759E+02, 4.3216E+02, 4.3675E+02, 4.4136E+02, 4.4599E+02,
     + 4.5064E+02, 4.5532E+02, 4.6001E+02, 4.6473E+02, 4.6947E+02,
     +  4.7424E+02, 4.7903E+02, 4.8384E+02, 4.8867E+02, 4.9354E+02,
     + 4.9842E+02, 5.0333E+02, 5.0827E+02, 5.1323E+02, 5.1822E+02,
     +  5.2323E+02, 5.2827E+02, 5.3334E+02, 5.3843E+02, 5.4356E+02,
     + 5.4871E+02, 5.5388E+02, 5.5909E+02, 5.6433E+02, 5.6959E+02,
     +  5.7489E+02, 5.8021E+02, 5.8556E+02, 5.9095E+02, 5.9636E+02,
     + 6.0181E+02, 6.0728E+02, 6.1279E+02, 6.1833E+02, 6.2390E+02,
     +  6.2950E+02, 6.3514E+02, 6.4081E+02, 6.4651E+02, 6.5224E+02,
     + 6.5801E+02, 6.6381E+02, 6.6965E+02, 6.7552E+02, 6.8142E+02,
     +  6.8736E+02, 6.9334E+02, 6.9935E+02, 7.0539E+02, 7.1147E+02,
     + 7.1759E+02, 7.2375E+02, 7.2994E+02, 7.3617E+02, 7.4243E+02,
     +  7.4873E+02, 7.5507E+02, 7.6145E+02, 7.6787E+02, 7.7432E+02,
     + 7.8082E+02, 7.8735E+02, 7.9392E+02, 8.0054E+02, 8.0719E+02,
     +  8.1388E+02, 8.2061E+02, 8.2739E+02, 8.3420E+02, 8.4106E+02,
     + 8.4795E+02, 8.5489E+02, 8.6187E+02, 8.6889E+02, 8.7596E+02,
     +  8.8306E+02, 8.9021E+02, 8.9741E+02, 9.0464E+02, 9.1192E+02,
     + 9.1925E+02, 9.2662E+02, 9.3403E+02, 9.4149E+02, 9.4899E+02,
     +  9.5654E+02, 9.6413E+02, 9.7177E+02, 9.7946E+02, 9.8719E+02,
     + 9.9497E+02, 1.0028E+03, 1.0107E+03, 1.0186E+03, 1.0266E+03,
     +  1.0346E+03, 1.0426E+03, 1.0507E+03, 1.0589E+03, 1.0671E+03,
     + 1.0754E+03, 1.0837E+03, 1.0920E+03, 1.1004E+03, 1.1089E+03,
     +  1.1174E+03, 1.1260E+03, 1.1346E+03, 1.1432E+03, 1.1520E+03,
     + 1.1607E+03, 1.1695E+03, 1.1784E+03, 1.1873E+03, 1.1963E+03,
     +  1.2053E+03, 1.2144E+03, 1.2235E+03, 1.2327E+03, 1.2420E+03,
     + 1.2513E+03, 1.2606E+03, 1.2700E+03, 1.2795E+03, 1.2890E+03,
     +  1.2986E+03, 1.3082E+03, 1.3179E+03, 1.3277E+03, 1.3375E+03,
     + 1.3473E+03, 1.3572E+03, 1.3672E+03, 1.3773E+03, 1.3873E+03,
     +  1.3975E+03, 1.4077E+03, 1.4180E+03, 1.4283E+03, 1.4387E+03,
     + 1.4491E+03, 1.4597E+03, 1.4702E+03, 1.4809E+03, 1.4915E+03,
     +  1.5023E+03, 1.5131E+03, 1.5240E+03, 1.5349E+03, 1.5459E+03,
     + 1.5570E+03, 1.5681E+03, 1.5793E+03, 1.5906E+03, 1.6019E+03,
     +  1.6133E+03, 1.6248E+03, 1.6363E+03, 1.6479E+03, 1.6595E+03,
     + 1.6712E+03, 1.6830E+03, 1.6949E+03, 1.7068E+03, 1.7188E+03,
     +  1.7308E+03, 1.7430E+03, 1.7551E+03, 1.7674E+03, 1.7797E+03,
     + 1.7921E+03, 1.8046E+03, 1.8171E+03, 1.8297E+03, 1.8424E+03,
     +  1.8552E+03, 1.8680E+03, 1.8809E+03, 1.8938E+03, 1.9069E+03,
     + 1.9200E+03, 1.9332E+03, 1.9464E+03, 1.9598E+03, 1.9732E+03,
     +  1.9866E+03, 2.0002E+03, 2.0138E+03, 2.0275E+03, 2.0413E+03,
     + 2.0551E+03, 2.0691E+03, 2.0831E+03, 2.0972E+03, 2.1113E+03,
     +  2.1256E+03, 2.1399E+03, 2.1543E+03, 2.1687E+03, 2.1833E+03,
     + 2.1979E+03, 2.2126E+03, 2.2274E+03, 2.2423E+03, 2.2572E+03,
     +  2.2723E+03, 2.2874E+03, 2.3026E+03, 2.3179E+03, 2.3332E+03,
     + 2.3487E+03, 2.3642E+03, 2.3798E+03, 2.3955E+03, 2.4113E+03,
     +  2.4271E+03, 2.4431E+03, 2.4591E+03, 2.4752E+03, 2.4914E+03,
     + 2.5077E+03, 2.5241E+03, 2.5406E+03, 2.5571E+03, 2.5738E+03,
     +  2.5905E+03, 2.6073E+03, 2.6242E+03, 2.6412E+03, 2.6583E+03,
     + 2.6755E+03, 2.6928E+03, 2.7101E+03, 2.7276E+03, 2.7451E+03,
     +  2.7627E+03, 2.7805E+03, 2.7983E+03, 2.8162E+03, 2.8342E+03,
     + 2.8523E+03, 2.8705E+03, 2.8888E+03, 2.9072E+03, 2.9257E+03,
     +  2.9442E+03, 2.9629E+03, 2.9817E+03, 3.0005E+03, 3.0195E+03,
     + 3.0386E+03, 3.0577E+03, 3.0770E+03, 3.0963E+03, 3.1158E+03,
     +  3.1354E+03, 3.1550E+03, 3.1748E+03, 3.1946E+03, 3.2146E+03,
     + 3.2347E+03, 3.2548E+03, 3.2751E+03, 3.2955E+03, 3.3159E+03,
     +  3.3365E+03, 3.3572E+03, 3.3780E+03, 3.3989E+03, 3.4199E+03,
     + 3.4410E+03, 3.4622E+03, 3.4835E+03, 3.5050E+03, 3.5265E+03,
     +  3.5481E+03, 3.5699E+03, 3.5918E+03, 3.6137E+03, 3.6358E+03,
     + 3.6580E+03, 3.6803E+03, 3.7027E+03, 3.7252E+03, 3.7479E+03,
     +  3.7706E+03, 3.7935E+03, 3.8165E+03, 3.8396E+03, 3.8628E+03,
     + 3.8861E+03, 3.9095E+03, 3.9331E+03, 3.9568E+03, 3.9805E+03,
     +  4.0044E+03, 4.0285E+03, 4.0526E+03, 4.0769E+03, 4.1012E+03,
     + 4.1257E+03, 4.1503E+03, 4.1751E+03, 4.1999E+03, 4.2249E+03,
     +  4.2500E+03, 4.2752E+03, 4.3006E+03, 4.3260E+03, 4.3516E+03,
     + 4.3773E+03, 4.4032E+03, 4.4291E+03, 4.4552E+03, 4.4814E+03,
     +  4.5078E+03, 4.5343E+03, 4.5609E+03, 4.5876E+03, 4.6144E+03,
     + 4.6414E+03, 4.6685E+03, 4.6958E+03, 4.7231E+03, 4.7506E+03,
     +  4.7783E+03, 4.8060E+03, 4.8339E+03, 4.8620E+03, 4.8901E+03,
     + 4.9184E+03, 4.9468E+03, 4.9754E+03, 5.0041E+03, 5.0329E+03,
     +  5.0619E+03, 5.0910E+03, 5.1203E+03, 5.1497E+03, 5.1792E+03,
     + 5.2088E+03, 5.2386E+03, 5.2686E+03, 5.2987E+03, 5.3289E+03,
     +  5.3592E+03, 5.3897E+03, 5.4204E+03, 5.4512E+03, 5.4821E+03,
     + 5.5132E+03, 5.5444E+03, 5.5758E+03, 5.6073E+03, 5.6389E+03,
     +  5.6707E+03, 5.7027E+03, 5.7348E+03, 5.7670E+03, 5.7994E+03,
     + 5.8320E+03, 5.8647E+03, 5.8975E+03, 5.9305E+03, 5.9637E+03,
     +  5.9970E+03, 6.0304E+03, 6.0640E+03, 6.0978E+03, 6.1317E+03,
     + 6.1657E+03, 6.2000E+03, 6.2343E+03, 6.2689E+03, 6.3036E+03,
     +  6.3384E+03, 6.3734E+03, 6.4086E+03, 6.4439E+03, 6.4794E+03,
     + 6.5150E+03, 6.5508E+03, 6.5868E+03, 6.6229E+03, 6.6592E+03,
     +  6.6956E+03, 6.7323E+03, 6.7690E+03, 6.8060E+03, 6.8431E+03,
     + 6.8804E+03, 6.9178E+03, 6.9554E+03, 6.9932E+03, 7.0311E+03,
     +  7.0692E+03, 7.1075E+03, 7.1460E+03, 7.1846E+03, 7.2234E+03,
     + 7.2623E+03, 7.3015E+03, 7.3408E+03, 7.3803E+03, 7.4199E+03,
     +  7.4598E+03, 7.4998E+03, 7.5399E+03, 7.5803E+03, 7.6208E+03,
     + 7.6615E+03, 7.7024E+03, 7.7435E+03, 7.7848E+03, 7.8262E+03,
     +  7.8678E+03, 7.9096E+03, 7.9515E+03, 7.9937E+03, 8.0360E+03,
     + 8.0786E+03, 8.1213E+03, 8.1642E+03, 8.2072E+03, 8.2505E+03,
     +  8.2939E+03, 8.3376E+03, 8.3814E+03, 8.4254E+03, 8.4696E+03,
     + 8.5140E+03, 8.5586E+03, 8.6034E+03, 8.6483E+03, 8.6935E+03,
     +  8.7388E+03, 8.7844E+03, 8.8301E+03, 8.8760E+03, 8.9222E+03,
     + 8.9685E+03, 9.0150E+03, 9.0617E+03, 9.1087E+03, 9.1558E+03,
     +  9.2031E+03, 9.2506E+03, 9.2983E+03, 9.3462E+03, 9.3944E+03,
     + 9.4427E+03, 9.4912E+03, 9.5399E+03, 9.5889E+03, 9.6380E+03,
     +  9.6874E+03, 9.7369E+03, 9.7867E+03, 9.8366E+03, 9.8868E+03,
     + 9.9372E+03, 9.9878E+03, 1.0039E+04, 1.0090E+04, 1.0141E+04,
     +  1.0192E+04, 1.0244E+04, 1.0296E+04, 1.0348E+04, 1.0400E+04,
     + 1.0453E+04, 1.0505E+04, 1.0558E+04, 1.0612E+04, 1.0665E+04,
     +  1.0719E+04, 1.0772E+04, 1.0826E+04, 1.0881E+04, 1.0935E+04,
     + 1.0990E+04, 1.1045E+04, 1.1100E+04, 1.1155E+04, 1.1211E+04,
     +  1.1267E+04, 1.1323E+04, 1.1379E+04, 1.1436E+04, 1.1493E+04,
     + 1.1550E+04, 1.1607E+04, 1.1664E+04, 1.1722E+04, 1.1780E+04,
     +  1.1838E+04, 1.1896E+04, 1.1955E+04, 1.2014E+04, 1.2073E+04,
     + 1.2132E+04, 1.2192E+04, 1.2252E+04, 1.2312E+04, 1.2372E+04,
     +  1.2433E+04, 1.2493E+04, 1.2554E+04, 1.2616E+04, 1.2677E+04,
     + 1.2739E+04, 1.2801E+04, 1.2863E+04, 1.2926E+04, 1.2988E+04,
     +  1.3051E+04, 1.3115E+04, 1.3178E+04, 1.3242E+04, 1.3306E+04,
     + 1.3370E+04, 1.3434E+04, 1.3499E+04, 1.3564E+04, 1.3629E+04,
     +  1.3695E+04, 1.3761E+04, 1.3827E+04, 1.3893E+04, 1.3960E+04,
     + 1.4026E+04, 1.4094E+04, 1.4161E+04, 1.4228E+04, 1.4296E+04,
     +  1.4364E+04, 1.4433E+04, 1.4501E+04, 1.4570E+04, 1.4640E+04,
     + 1.4709E+04, 1.4779E+04, 1.4849E+04, 1.4919E+04, 1.4990E+04,
     +  1.5060E+04, 1.5131E+04, 1.5203E+04, 1.5274E+04, 1.5346E+04,
     + 1.5418E+04, 1.5491E+04, 1.5564E+04, 1.5637E+04, 1.5710E+04,
     +  1.5784E+04, 1.5857E+04, 1.5932E+04, 1.6006E+04, 1.6081E+04,
     + 1.6156E+04, 1.6231E+04, 1.6306E+04, 1.6382E+04, 1.6458E+04,
     +  1.6535E+04, 1.6612E+04, 1.6689E+04, 1.6766E+04, 1.6843E+04,
     + 1.6921E+04, 1.6999E+04, 1.7078E+04, 1.7157E+04, 1.7236E+04,
     +  1.7315E+04, 1.7395E+04, 1.7475E+04, 1.7555E+04, 1.7635E+04,
     + 1.7716E+04, 1.7797E+04, 1.7879E+04, 1.7961E+04, 1.8043E+04,
     +  1.8125E+04, 1.8208E+04, 1.8291E+04, 1.8374E+04, 1.8458E+04,
     + 1.8541E+04, 1.8626E+04, 1.8710E+04, 1.8795E+04, 1.8880E+04,
     +  1.8966E+04, 1.9051E+04, 1.9138E+04, 1.9224E+04, 1.9311E+04,
     + 1.9398E+04, 1.9485E+04, 1.9573E+04, 1.9661E+04, 1.9749E+04,
     +  1.9838E+04, 1.9927E+04, 2.0016E+04, 2.0106E+04, 2.0196E+04,
     + 2.0286E+04, 2.0377E+04, 2.0468E+04, 2.0559E+04, 2.0650E+04,
     +  2.0742E+04, 2.0835E+04, 2.0927E+04, 2.1020E+04, 2.1114E+04,
     + 2.1207E+04, 2.1301E+04, 2.1395E+04, 2.1490E+04, 2.1585E+04,
     +  2.1680E+04, 2.1776E+04, 2.1872E+04, 2.1968E+04, 2.2065E+04,
     + 2.2162E+04, 2.2259E+04, 2.2357E+04, 2.2455E+04, 2.2554E+04,
     +  2.2652E+04, 2.2752E+04, 2.2851E+04, 2.2951E+04, 2.3051E+04,
     + 2.3152E+04, 2.3253E+04, 2.3354E+04, 2.3455E+04, 2.3557E+04,
     +  2.3660E+04, 2.3762E+04, 2.3866E+04, 2.3969E+04, 2.4073E+04,
     + 2.4177E+04, 2.4281E+04, 2.4386E+04, 2.4492E+04, 2.4597E+04,
     +  2.4703E+04, 2.4810E+04, 2.4916E+04, 2.5023E+04, 2.5131E+04,
     + 2.5239E+04, 2.5347E+04, 2.5456E+04, 2.5565E+04, 2.5674E+04,
     +  2.5784E+04, 2.5894E+04, 2.6004E+04, 2.6115E+04, 2.6227E+04,
     + 2.6338E+04, 2.6450E+04, 2.6563E+04, 2.6676E+04, 2.6789E+04,
     +  2.6902E+04, 2.7017E+04, 2.7131E+04, 2.7246E+04, 2.7361E+04,
     + 2.7476E+04, 2.7592E+04, 2.7709E+04, 2.7826E+04, 2.7943E+04,
     +  2.8060E+04, 2.8178E+04, 2.8297E+04, 2.8415E+04, 2.8535E+04,
     + 2.8654E+04, 2.8774E+04, 2.8895E+04, 2.9015E+04, 2.9137E+04,
     +  2.9258E+04, 2.9380E+04, 2.9503E+04, 2.9626E+04, 2.9749E+04,
     + 2.9873E+04, 2.9997E+04, 3.0121E+04, 3.0246E+04, 3.0372E+04,
     +  3.0498E+04, 3.0624E+04, 3.0750E+04, 3.0877E+04, 3.1005E+04,
     + 3.1133E+04, 3.1261E+04, 3.1390E+04, 3.1519E+04, 3.1649E+04,
     +  3.1779E+04, 3.1909E+04, 3.2040E+04, 3.2172E+04, 3.2304E+04,
     + 3.2436E+04, 3.2569E+04, 3.2702E+04, 3.2835E+04, 3.2969E+04,
     +  3.3104E+04, 3.3239E+04, 3.3374E+04, 3.3510E+04, 3.3646E+04,
     + 3.3783E+04, 3.3920E+04, 3.4057E+04, 3.4196E+04, 3.4334E+04,
     +  3.4473E+04, 3.4612E+04, 3.4752E+04, 3.4893E+04, 3.5033E+04,
     + 3.5175E+04, 3.5316E+04, 3.5458E+04, 3.5601E+04, 3.5744E+04,
     +  3.5888E+04, 3.6032E+04, 3.6176E+04, 3.6321E+04, 3.6467E+04,
     + 3.6613E+04, 3.6759E+04, 3.6906E+04, 3.7053E+04, 3.7201E+04,
     +  3.7349E+04, 3.7498E+04, 3.7647E+04, 3.7797E+04, 3.7947E+04,
     + 3.8098E+04, 3.8249E+04, 3.8401E+04, 3.8553E+04, 3.8705E+04,
     +  3.8859E+04, 3.9012E+04, 3.9166E+04, 3.9321E+04, 3.9476E+04,
     + 3.9632E+04, 3.9788E+04, 3.9944E+04, 4.0101E+04, 4.0259E+04,
     +  4.0417E+04, 4.0576E+04, 4.0735E+04, 4.0894E+04, 4.1054E+04,
     + 4.1215E+04, 4.1376E+04, 4.1538E+04, 4.1700E+04, 4.1862E+04,
     +  4.2025E+04, 4.2189E+04, 4.2353E+04, 4.2518E+04, 4.2683E+04,
     + 4.2849E+04, 4.3015E+04, 4.3182E+04, 4.3349E+04, 4.3517E+04,
     +  4.3685E+04, 4.3854E+04, 4.4024E+04, 4.4194E+04, 4.4364E+04,
     + 4.4535E+04, 4.4707E+04, 4.4879E+04, 4.5051E+04, 4.5224E+04,
     +  4.5398E+04, 4.5572E+04, 4.5747E+04, 4.5922E+04, 4.6098E+04,
     + 4.6274E+04, 4.6451E+04, 4.6629E+04, 4.6807E+04, 4.6985E+04,
     +  4.7164E+04, 4.7344E+04, 4.7524E+04, 4.7705E+04, 4.7886E+04,
     + 4.8068E+04, 4.8251E+04, 4.8434E+04, 4.8617E+04, 4.8801E+04,
     +  4.8986E+04, 4.9171E+04, 4.9357E+04, 4.9543E+04, 4.9730E+04,
     + 4.9918E+04, 5.0106E+04, 5.0295E+04, 5.0484E+04, 5.0674E+04,
     +  5.0864E+04, 5.1055E+04, 5.1246E+04, 5.1439E+04, 5.1631E+04,
     + 5.1825E+04, 5.2018E+04, 5.2213E+04, 5.2408E+04, 5.2603E+04,
     +  5.2800E+04, 5.2996E+04, 5.3194E+04, 5.3392E+04, 5.3590E+04,
     + 5.3790E+04, 5.3989E+04, 5.4190E+04, 5.4391E+04, 5.4592E+04,
     +  5.4794E+04, 5.4997E+04, 5.5201E+04, 5.5405E+04, 5.5609E+04,
     + 5.5814E+04, 5.6020E+04, 5.6227E+04, 5.6434E+04, 5.6641E+04,
     +  5.6849E+04, 5.7058E+04, 5.7268E+04, 5.7478E+04, 5.7689E+04,
     + 5.7900E+04, 5.8112E+04, 5.8325E+04, 5.8538E+04, 5.8752E+04,
     +  5.8966E+04, 5.9181E+04, 5.9397E+04, 5.9614E+04, 5.9831E+04,
     + 6.0048E+04, 6.0266E+04, 6.0485E+04, 6.0705E+04, 6.0925E+04,
     +  6.1146E+04, 6.1368E+04, 6.1590E+04, 6.1813E+04, 6.2036E+04,
     + 6.2260E+04, 6.2485E+04, 6.2710E+04, 6.2936E+04, 6.3163E+04,
     +  6.3390E+04, 6.3618E+04, 6.3847E+04, 6.4077E+04, 6.4307E+04,
     + 6.4537E+04, 6.4769E+04, 6.5001E+04, 6.5233E+04, 6.5467E+04,
     +  6.5701E+04, 6.5935E+04, 6.6171E+04, 6.6407E+04, 6.6643E+04,
     + 6.6881E+04, 6.7119E+04, 6.7358E+04, 6.7597E+04, 6.7837E+04,
     +  6.8078E+04, 6.8320E+04, 6.8562E+04, 6.8805E+04, 6.9048E+04,
     + 6.9293E+04, 6.9538E+04, 6.9783E+04, 7.0030E+04, 7.0277E+04,
     +  7.0524E+04, 7.0773E+04, 7.1022E+04, 7.1272E+04, 7.1523E+04,
     + 7.1774E+04, 7.2026E+04, 7.2279E+04, 7.2532E+04, 7.2786E+04,
     +  7.3041E+04, 7.3296E+04, 7.3553E+04, 7.3810E+04, 7.4068E+04,
     + 7.4326E+04, 7.4585E+04, 7.4845E+04, 7.5106E+04, 7.5367E+04,
     +  7.5629E+04, 7.5892E+04, 7.6156E+04, 7.6420E+04, 7.6685E+04,
     + 7.6951E+04, 7.7217E+04, 7.7484E+04, 7.7752E+04, 7.8021E+04,
     +  7.8291E+04, 7.8561E+04, 7.8832E+04, 7.9103E+04, 7.9376E+04,
     + 7.9649E+04, 7.9923E+04, 8.0198E+04, 8.0473E+04, 8.0750E+04,
     +  8.1027E+04, 8.1305E+04, 8.1583E+04, 8.1862E+04, 8.2142E+04,
     + 8.2423E+04, 8.2705E+04, 8.2987E+04, 8.3270E+04, 8.3554E+04,
     +  8.3839E+04, 8.4125E+04, 8.4411E+04, 8.4698E+04, 8.4986E+04,
     + 8.5274E+04, 8.5564E+04, 8.5854E+04, 8.6145E+04, 8.6437E+04,
     +  8.6729E+04, 8.7023E+04, 8.7317E+04, 8.7612E+04, 8.7908E+04,
     + 8.8204E+04, 8.8502E+04, 8.8800E+04, 8.9099E+04, 8.9399E+04,
     +  8.9699E+04, 9.0001E+04, 9.0303E+04, 9.0606E+04, 9.0910E+04,
     + 9.1214E+04, 9.1520E+04, 9.1826E+04, 9.2133E+04, 9.2441E+04,
     +  9.2750E+04, 9.3060E+04, 9.3370E+04, 9.3681E+04, 9.3993E+04,
     + 9.4306E+04, 9.4620E+04, 9.4935E+04, 9.5250E+04, 9.5566E+04,
     +  9.5883E+04, 9.6201E+04, 9.6520E+04, 9.6840E+04, 9.7160E+04,
     + 9.7482E+04, 9.7804E+04, 9.8127E+04, 9.8451E+04, 9.8775E+04,
     +  9.9101E+04, 9.9427E+04, 9.9755E+04, 1.0008E+05, 1.0041E+05,
     + 1.0074E+05, 1.0107E+05, 1.0140E+05, 1.0174E+05, 1.0207E+05,
     +  1.0240E+05, 1.0274E+05, 1.0308E+05, 1.0341E+05, 1.0375E+05,
     + 1.0409E+05, 1.0443E+05, 1.0477E+05, 1.0511E+05, 1.0545E+05,
     +  1.0580E+05, 1.0614E+05, 1.0648E+05, 1.0683E+05, 1.0718E+05,
     + 1.0752E+05, 1.0787E+05, 1.0822E+05, 1.0857E+05, 1.0892E+05,
     +  1.0928E+05, 1.0963E+05, 1.0998E+05, 1.1034E+05, 1.1069E+05,
     + 1.1105E+05, 1.1141E+05, 1.1177E+05, 1.1213E+05, 1.1249E+05,
     +  1.1285E+05, 1.1321E+05, 1.1357E+05, 1.1394E+05, 1.1430E+05,
     + 1.1467E+05, 1.1504E+05, 1.1540E+05, 1.1577E+05, 1.1614E+05,
     +  1.1651E+05, 1.1688E+05, 1.1726E+05, 1.1763E+05, 1.1800E+05,
     + 1.1838E+05, 1.1876E+05, 1.1913E+05, 1.1951E+05, 1.1989E+05,
     +  1.2027E+05, 1.2065E+05, 1.2103E+05, 1.2142E+05, 1.2180E+05,
     + 1.2219E+05, 1.2257E+05, 1.2296E+05, 1.2335E+05, 1.2374E+05,
     +  1.2413E+05, 1.2452E+05, 1.2491E+05, 1.2530E+05, 1.2570E+05,
     + 1.2609E+05, 1.2649E+05, 1.2688E+05, 1.2728E+05, 1.2768E+05,
     +  1.2808E+05, 1.2848E+05, 1.2888E+05, 1.2929E+05, 1.2969E+05,
     + 1.3009E+05, 1.3050E+05, 1.3091E+05, 1.3131E+05, 1.3172E+05,
     +  1.3213E+05, 1.3254E+05, 1.3296E+05, 1.3337E+05, 1.3378E+05,
     + 1.3420E+05, 1.3461E+05, 1.3503E+05, 1.3545E+05, 1.3587E+05,
     +  1.3629E+05, 1.3671E+05, 1.3713E+05, 1.3756E+05, 1.3798E+05,
     + 1.3840E+05, 1.3883E+05, 1.3926E+05, 1.3969E+05, 1.4012E+05,
     +  1.4055E+05, 1.4098E+05, 1.4141E+05, 1.4185E+05, 1.4228E+05,
     + 1.4272E+05, 1.4315E+05, 1.4359E+05, 1.4403E+05, 1.4447E+05,
     +  1.4491E+05, 1.4535E+05, 1.4580E+05, 1.4624E+05, 1.4669E+05,
     + 1.4713E+05, 1.4758E+05, 1.4803E+05, 1.4848E+05, 1.4893E+05,
     +  1.4938E+05, 1.4984E+05, 1.5029E+05, 1.5075E+05, 1.5120E+05,
     + 1.5166E+05, 1.5212E+05, 1.5258E+05, 1.5304E+05, 1.5350E+05,
     +  1.5396E+05, 1.5443E+05, 1.5489E+05, 1.5536E+05, 1.5583E+05,
     + 1.5630E+05, 1.5677E+05, 1.5724E+05, 1.5771E+05, 1.5818E+05,
     +  1.5866E+05, 1.5913E+05, 1.5961E+05, 1.6009E+05, 1.6056E+05,
     + 1.6104E+05, 1.6153E+05, 1.6201E+05, 1.6249E+05, 1.6298E+05,
     +  1.6346E+05, 1.6395E+05, 1.6444E+05, 1.6493E+05, 1.6542E+05,
     + 1.6591E+05, 1.6640E+05, 1.6689E+05, 1.6739E+05, 1.6788E+05,
     +  1.6838E+05, 1.6888E+05, 1.6938E+05, 1.6988E+05, 1.7038E+05,
     + 1.7089E+05, 1.7139E+05, 1.7190E+05, 1.7240E+05, 1.7291E+05,
     +  1.7342E+05, 1.7393E+05, 1.7444E+05, 1.7495E+05, 1.7547E+05,
     + 1.7598E+05, 1.7650E+05, 1.7702E+05, 1.7754E+05, 1.7806E+05,
     +  1.7858E+05, 1.7910E+05, 1.7962E+05, 1.8015E+05, 1.8067E+05,
     + 1.8120E+05, 1.8173E+05, 1.8226E+05, 1.8279E+05, 1.8332E+05,
     +  1.8385E+05, 1.8439E+05, 1.8492E+05, 1.8546E+05, 1.8600E+05,
     + 1.8654E+05, 1.8708E+05, 1.8762E+05, 1.8817E+05, 1.8871E+05,
     +  1.8926E+05, 1.8980E+05, 1.9035E+05, 1.9090E+05, 1.9145E+05,
     + 1.9200E+05, 1.9256E+05, 1.9311E+05, 1.9367E+05, 1.9422E+05,
     +  1.9478E+05, 1.9534E+05, 1.9590E+05, 1.9646E+05, 1.9703E+05,
     + 1.9759E+05, 1.9816E+05, 1.9873E+05, 1.9929E+05, 1.9986E+05,
     +  2.0044E+05, 2.0101E+05, 2.0158E+05, 2.0216E+05, 2.0273E+05,
     + 2.0331E+05, 2.0389E+05, 2.0447E+05, 2.0505E+05, 2.0563E+05,
     +  2.0622E+05, 2.0680E+05, 2.0739E+05, 2.0798E+05, 2.0857E+05,
     + 2.0916E+05, 2.0975E+05, 2.1035E+05, 2.1094E+05, 2.1154E+05,
     +  2.1213E+05, 2.1273E+05, 2.1333E+05, 2.1393E+05, 2.1454E+05,
     + 2.1514E+05, 2.1575E+05, 2.1635E+05, 2.1696E+05, 2.1757E+05,
     +  2.1818E+05, 2.1879E+05, 2.1941E+05, 2.2002E+05, 2.2064E+05,
     + 2.2126E+05, 2.2188E+05, 2.2250E+05, 2.2312E+05, 2.2374E+05,
     +  2.2437E+05, 2.2499E+05, 2.2562E+05, 2.2625E+05, 2.2688E+05,
     + 2.2751E+05, 2.2814E+05, 2.2878E+05, 2.2941E+05, 2.3005E+05,
     +  2.3069E+05, 2.3133E+05, 2.3197E+05, 2.3261E+05, 2.3326E+05,
     + 2.3390E+05, 2.3455E+05, 2.3520E+05, 2.3585E+05, 2.3650E+05,
     +  2.3715E+05, 2.3781E+05, 2.3846E+05, 2.3912E+05, 2.3978E+05,
     + 2.4044E+05, 2.4110E+05, 2.4176E+05, 2.4242E+05, 2.4309E+05,
     +  2.4376E+05, 2.4443E+05, 2.4510E+05, 2.4577E+05, 2.4644E+05,
     + 2.4711E+05, 2.4779E+05, 2.4847E+05, 2.4915E+05, 2.4983E+05,
     +  2.5051E+05, 2.5119E+05, 2.5187E+05, 2.5256E+05, 2.5325E+05,
     + 2.5394E+05, 2.5463E+05, 2.5532E+05, 2.5601E+05, 2.5671E+05,
     +  2.5740E+05, 2.5810E+05, 2.5880E+05, 2.5950E+05, 2.6020E+05,
     + 2.6091E+05, 2.6161E+05, 2.6232E+05, 2.6303E+05, 2.6374E+05,
     +  2.6445E+05, 2.6516E+05, 2.6588E+05, 2.6659E+05, 2.6731E+05,
     + 2.6803E+05, 2.6875E+05, 2.6947E+05, 2.7020E+05, 2.7092E+05,
     +  2.7165E+05, 2.7238E+05, 2.7310E+05, 2.7384E+05, 2.7457E+05,
     + 2.7530E+05, 2.7604E+05, 2.7678E+05, 2.7752E+05, 2.7826E+05,
     +  2.7900E+05, 2.7974E+05, 2.8049E+05, 2.8123E+05, 2.8198E+05,
     + 2.8273E+05, 2.8348E+05, 2.8424E+05, 2.8499E+05, 2.8575E+05,
     +  2.8651E+05, 2.8726E+05, 2.8803E+05, 2.8879E+05, 2.8955E+05,
     + 2.9032E+05, 2.9109E+05, 2.9185E+05, 2.9263E+05, 2.9340E+05,
     +  2.9417E+05, 2.9495E+05, 2.9572E+05, 2.9650E+05, 2.9728E+05,
     + 2.9806E+05, 2.9885E+05, 2.9963E+05, 3.0042E+05, 3.0121E+05,
     +  3.0200E+05, 3.0279E+05, 3.0358E+05, 3.0438E+05, 3.0517E+05,
     + 3.0597E+05, 3.0677E+05, 3.0757E+05, 3.0838E+05, 3.0918E+05,
     +  3.0999E+05, 3.1079E+05, 3.1160E+05, 3.1242E+05, 3.1323E+05,
     + 3.1404E+05, 3.1486E+05, 3.1568E+05, 3.1650E+05, 3.1732E+05,
     +  3.1814E+05, 3.1897E+05, 3.1979E+05, 3.2062E+05, 3.2145E+05,
     + 3.2228E+05, 3.2312E+05, 3.2395E+05, 3.2479E+05, 3.2562E+05,
     +  3.2646E+05, 3.2731E+05, 3.2815E+05, 3.2899E+05, 3.2984E+05,
     + 3.3069E+05, 3.3154E+05, 3.3239E+05, 3.3324E+05, 3.3410E+05,
     +  3.3496E+05, 3.3582E+05, 3.3668E+05, 3.3754E+05, 3.3840E+05,
     + 3.3927E+05, 3.4014E+05, 3.4100E+05, 3.4188E+05, 3.4275E+05,
     +  3.4362E+05, 3.4450E+05, 3.4538E+05, 3.4626E+05, 3.4714E+05,
     + 3.4802E+05, 3.4891E+05, 3.4979E+05, 3.5068E+05, 3.5157E+05,
     +  3.5246E+05, 3.5336E+05, 3.5425E+05, 3.5515E+05, 3.5605E+05,
     + 3.5695E+05, 3.5785E+05, 3.5876E+05, 3.5966E+05, 3.6057E+05,
     +  3.6148E+05, 3.6239E+05, 3.6331E+05, 3.6422E+05, 3.6514E+05,
     + 3.6606E+05, 3.6698E+05, 3.6790E+05, 3.6883E+05, 3.6975E+05,
     +  3.7068E+05, 3.7161E+05, 3.7254E+05, 3.7347E+05, 3.7441E+05,
     + 3.7535E+05, 3.7629E+05, 3.7723E+05, 3.7817E+05, 3.7911E+05,
     +  3.8006E+05, 3.8101E+05, 3.8196E+05, 3.8291E+05, 3.8386E+05,
     + 3.8482E+05, 3.8578E+05, 3.8674E+05, 3.8770E+05, 3.8866E+05,
     +  3.8963E+05, 3.9059E+05, 3.9156E+05, 3.9253E+05, 3.9350E+05,
     + 3.9448E+05, 3.9546E+05, 3.9643E+05, 3.9741E+05, 3.9840E+05,
     +  3.9938E+05, 4.0037E+05, 4.0135E+05, 4.0234E+05, 4.0333E+05,
     + 4.0433E+05, 4.0532E+05, 4.0632E+05, 4.0732E+05, 4.0832E+05,
     +  4.0932E+05, 4.1033E+05, 4.1133E+05, 4.1234E+05, 4.1335E+05,
     + 4.1437E+05, 4.1538E+05, 4.1640E+05, 4.1742E+05, 4.1844E+05,
     +  4.1946E+05, 4.2048E+05, 4.2151E+05, 4.2254E+05, 4.2357E+05,
     + 4.2460E+05, 4.2563E+05, 4.2667E+05, 4.2771E+05, 4.2875E+05,
     +  4.2979E+05, 4.3083E+05, 4.3188E+05, 4.3293E+05, 4.3398E+05,
     + 4.3503E+05, 4.3608E+05, 4.3714E+05, 4.3819E+05, 4.3925E+05,
     +  4.4032E+05, 4.4138E+05, 4.4245E+05, 4.4351E+05, 4.4458E+05,
     + 4.4566E+05, 4.4673E+05, 4.4780E+05, 4.4888E+05, 4.4996E+05,
     +  4.5104E+05, 4.5213E+05, 4.5321E+05, 4.5430E+05, 4.5539E+05,
     + 4.5648E+05, 4.5758E+05, 4.5867E+05, 4.5977E+05, 4.6087E+05,
     +  4.6198E+05, 4.6308E+05, 4.6419E+05, 4.6529E+05, 4.6641E+05,
     + 4.6752E+05, 4.6863E+05, 4.6975E+05, 4.7087E+05, 4.7199E+05,
     +  4.7311E+05, 4.7424E+05, 4.7536E+05, 4.7649E+05, 4.7762E+05,
     + 4.7876E+05, 4.7989E+05, 4.8103E+05, 4.8217E+05, 4.8331E+05,
     +  4.8446E+05, 4.8560E+05, 4.8675E+05, 4.8790E+05, 4.8905E+05,
     + 4.9021E+05, 4.9136E+05, 4.9252E+05, 4.9368E+05, 4.9484E+05,
     +  4.9601E+05, 4.9718E+05, 4.9835E+05, 4.9952E+05, 5.0069E+05,
     + 5.0187E+05, 5.0304E+05, 5.0422E+05, 5.0541E+05, 5.0659E+05,
     +  5.0778E+05, 5.0897E+05, 5.1016E+05, 5.1135E+05, 5.1254E+05,
     + 5.1374E+05, 5.1494E+05, 5.1614E+05, 5.1735E+05, 5.1855E+05,
     +  5.1976E+05, 5.2097E+05, 5.2218E+05, 5.2340E+05, 5.2461E+05,
     + 5.2583E+05, 5.2705E+05, 5.2828E+05, 5.2950E+05, 5.3073E+05,
     +  5.3196E+05, 5.3319E+05, 5.3443E+05, 5.3566E+05, 5.3690E+05,
     + 5.3814E+05, 5.3939E+05, 5.4063E+05, 5.4188E+05, 5.4313E+05,
     +  5.4438E+05, 5.4564E+05, 5.4689E+05, 5.4815E+05, 5.4942E+05,
     + 5.5068E+05, 5.5194E+05, 5.5321E+05, 5.5448E+05, 5.5575E+05,
     +  5.5703E+05, 5.5831E+05, 5.5959E+05, 5.6087E+05, 5.6215E+05,
     + 5.6344E+05, 5.6473E+05, 5.6602E+05, 5.6731E+05, 5.6860E+05,
     +  5.6990E+05, 5.7120E+05, 5.7250E+05, 5.7381E+05, 5.7511E+05,
     + 5.7642E+05, 5.7773E+05, 5.7905E+05, 5.8036E+05, 5.8168E+05,
     +  5.8300E+05, 5.8432E+05, 5.8565E+05, 5.8698E+05, 5.8831E+05,
     + 5.8964E+05, 5.9097E+05, 5.9231E+05, 5.9365E+05, 5.9499E+05,
     +  5.9633E+05, 5.9768E+05, 5.9903E+05, 6.0038E+05, 6.0173E+05,
     + 6.0309E+05, 6.0445E+05, 6.0581E+05, 6.0717E+05, 6.0853E+05,
     +  6.0990E+05, 6.1127E+05, 6.1264E+05, 6.1402E+05, 6.1539E+05,
     + 6.1677E+05, 6.1815E+05, 6.1954E+05, 6.2092E+05, 6.2231E+05,
     +  6.2370E+05, 6.2510E+05, 6.2649E+05, 6.2789E+05, 6.2929E+05,
     + 6.3070E+05, 6.3210E+05, 6.3351E+05, 6.3492E+05, 6.3633E+05,
     +  6.3775E+05, 6.3916E+05, 6.4058E+05, 6.4201E+05, 6.4343E+05,
     + 6.4486E+05, 6.4629E+05, 6.4772E+05, 6.4916E+05, 6.5059E+05,
     +  6.5203E+05, 6.5347E+05, 6.5492E+05, 6.5636E+05, 6.5781E+05,
     + 6.5927E+05, 6.6072E+05, 6.6218E+05, 6.6364E+05, 6.6510E+05,
     +  6.6656E+05, 6.6803E+05, 6.6950E+05, 6.7097E+05, 6.7244E+05,
     + 6.7392E+05, 6.7540E+05, 6.7688E+05, 6.7836E+05, 6.7985E+05,
     +  6.8134E+05, 6.8283E+05, 6.8432E+05, 6.8582E+05, 6.8732E+05,
     + 6.8882E+05, 6.9033E+05, 6.9183E+05, 6.9334E+05, 6.9485E+05,
     +  6.9637E+05, 6.9788E+05, 6.9940E+05, 7.0092E+05, 7.0245E+05,
     + 7.0397E+05, 7.0550E+05, 7.0704E+05, 7.0857E+05, 7.1011E+05,
     +  7.1165E+05, 7.1319E+05, 7.1473E+05, 7.1628E+05, 7.1783E+05,
     + 7.1938E+05, 7.2094E+05, 7.2250E+05, 7.2406E+05, 7.2562E+05,
     +  7.2718E+05, 7.2875E+05, 7.3032E+05, 7.3189E+05, 7.3347E+05,
     + 7.3505E+05, 7.3663E+05, 7.3821E+05, 7.3980E+05, 7.4139E+05,
     +  7.4298E+05, 7.4457E+05, 7.4617E+05, 7.4777E+05, 7.4937E+05,
     + 7.5097E+05, 7.5258E+05, 7.5419E+05, 7.5580E+05, 7.5742E+05,
     +  7.5903E+05, 7.6065E+05, 7.6228E+05, 7.6390E+05, 7.6553E+05,
     + 7.6716E+05, 7.6879E+05, 7.7043E+05, 7.7207E+05, 7.7371E+05,
     +  7.7535E+05, 7.7700E+05, 7.7865E+05, 7.8030E+05, 7.8195E+05,
     + 7.8361E+05, 7.8527E+05, 7.8693E+05, 7.8860E+05, 7.9027E+05,
     +  7.9194E+05, 7.9361E+05, 7.9529E+05, 7.9697E+05, 7.9865E+05,
     + 8.0033E+05, 8.0202E+05, 8.0371E+05, 8.0540E+05, 8.0709E+05,
     +  8.0879E+05, 8.1049E+05, 8.1220E+05, 8.1390E+05, 8.1561E+05,
     + 8.1732E+05, 8.1904E+05, 8.2075E+05, 8.2247E+05, 8.2419E+05,
     +  8.2592E+05, 8.2765E+05, 8.2938E+05, 8.3111E+05, 8.3285E+05,
     + 8.3458E+05, 8.3633E+05, 8.3807E+05, 8.3982E+05, 8.4157E+05,
     +  8.4332E+05, 8.4508E+05, 8.4683E+05, 8.4859E+05, 8.5036E+05,
     + 8.5212E+05, 8.5389E+05, 8.5567E+05, 8.5744E+05, 8.5922E+05,
     +  8.6100E+05, 8.6278E+05, 8.6457E+05, 8.6636E+05, 8.6815E+05,
     + 8.6994E+05, 8.7174E+05, 8.7354E+05, 8.7534E+05, 8.7715E+05,
     +  8.7896E+05, 8.8077E+05, 8.8258E+05, 8.8440E+05, 8.8622E+05,
     + 8.8804E+05, 8.8987E+05, 8.9169E+05, 8.9353E+05, 8.9536E+05,
     +  8.9720E+05, 8.9904E+05, 9.0088E+05, 9.0272E+05, 9.0457E+05,
     + 9.0642E+05, 9.0828E+05, 9.1014E+05, 9.1199E+05, 9.1386E+05,
     +  9.1572E+05, 9.1759E+05, 9.1946E+05, 9.2134E+05, 9.2321E+05,
     + 9.2509E+05, 9.2698E+05, 9.2886E+05, 9.3075E+05, 9.3264E+05,
     +  9.3454E+05, 9.3643E+05, 9.3833E+05, 9.4024E+05, 9.4214E+05,
     + 9.4405E+05, 9.4596E+05, 9.4788E+05, 9.4980E+05, 9.5172E+05,
     +  9.5364E+05, 9.5557E+05, 9.5750E+05, 9.5943E+05, 9.6137E+05,
     + 9.6330E+05, 9.6525E+05, 9.6719E+05, 9.6914E+05, 9.7109E+05,
     +  9.7304E+05, 9.7500E+05, 9.7695E+05, 9.7892E+05, 9.8088E+05,
     + 9.8285E+05, 9.8482E+05, 9.8679E+05, 9.8877E+05, 9.9075E+05,
     +  9.9273E+05, 9.9472E+05, 9.9671E+05, 9.9870E+05, 1.0007E+06,
     + 1.0027E+06, 1.0047E+06, 1.0067E+06, 1.0087E+06, 1.0107E+06,
     +  1.0127E+06, 1.0147E+06, 1.0168E+06, 1.0188E+06, 1.0208E+06,
     + 1.0228E+06, 1.0249E+06, 1.0269E+06, 1.0289E+06, 1.0310E+06,
     +  1.0330E+06, 1.0351E+06, 1.0371E+06, 1.0392E+06, 1.0412E+06,
     + 1.0433E+06, 1.0453E+06, 1.0474E+06, 1.0495E+06, 1.0515E+06,
     +  1.0536E+06, 1.0557E+06, 1.0578E+06, 1.0599E+06, 1.0619E+06,
     + 1.0640E+06, 1.0661E+06, 1.0682E+06, 1.0703E+06, 1.0724E+06,
     +  1.0745E+06, 1.0766E+06, 1.0787E+06, 1.0809E+06, 1.0830E+06,
     + 1.0851E+06, 1.0872E+06, 1.0893E+06, 1.0915E+06, 1.0936E+06,
     +  1.0957E+06, 1.0979E+06, 1.1000E+06, 1.1022E+06, 1.1043E+06,
     + 1.1065E+06, 1.1086E+06, 1.1108E+06, 1.1129E+06, 1.1151E+06,
     +  1.1173E+06, 1.1194E+06, 1.1216E+06, 1.1238E+06, 1.1260E+06,
     + 1.1281E+06, 1.1303E+06, 1.1325E+06, 1.1347E+06, 1.1369E+06,
     +  1.1391E+06, 1.1413E+06, 1.1435E+06, 1.1457E+06, 1.1479E+06,
     + 1.1501E+06, 1.1524E+06, 1.1546E+06, 1.1568E+06, 1.1590E+06,
     +  1.1613E+06, 1.1635E+06, 1.1657E+06, 1.1680E+06, 1.1702E+06,
     + 1.1725E+06, 1.1747E+06, 1.1770E+06, 1.1792E+06, 1.1815E+06,
     +  1.1837E+06, 1.1860E+06, 1.1883E+06, 1.1905E+06, 1.1928E+06,
     + 1.1951E+06, 1.1974E+06, 1.1997E+06, 1.2020E+06, 1.2042E+06,
     +  1.2065E+06, 1.2088E+06, 1.2111E+06, 1.2134E+06, 1.2158E+06,
     + 1.2181E+06, 1.2204E+06, 1.2227E+06, 1.2250E+06, 1.2273E+06,
     +  1.2297E+06, 1.2320E+06, 1.2343E+06, 1.2367E+06, 1.2390E+06,
     + 1.2414E+06, 1.2437E+06, 1.2461E+06, 1.2484E+06, 1.2508E+06,
     +  1.2531E+06, 1.2555E+06, 1.2579E+06, 1.2602E+06, 1.2626E+06,
     + 1.2650E+06, 1.2674E+06, 1.2697E+06, 1.2721E+06, 1.2745E+06,
     +  1.2769E+06, 1.2793E+06, 1.2817E+06, 1.2841E+06, 1.2865E+06,
     + 1.2889E+06, 1.2913E+06, 1.2938E+06, 1.2962E+06, 1.2986E+06,
     +  1.3010E+06, 1.3035E+06, 1.3059E+06, 1.3083E+06, 1.3108E+06,
     + 1.3132E+06, 1.3157E+06, 1.3181E+06, 1.3206E+06, 1.3230E+06,
     +  1.3255E+06, 1.3279E+06, 1.3304E+06, 1.3329E+06, 1.3353E+06,
     + 1.3378E+06, 1.3403E+06, 1.3428E+06, 1.3453E+06, 1.3478E+06,
     +  1.3503E+06, 1.3528E+06, 1.3553E+06, 1.3578E+06, 1.3603E+06,
     + 1.3628E+06, 1.3653E+06, 1.3678E+06, 1.3703E+06, 1.3729E+06,
     +  1.3754E+06, 1.3779E+06, 1.3805E+06, 1.3830E+06, 1.3855E+06,
     + 1.3881E+06, 1.3906E+06, 1.3932E+06, 1.3957E+06, 1.3983E+06,
     +  1.4009E+06, 1.4034E+06, 1.4060E+06, 1.4086E+06, 1.4112E+06,
     + 1.4137E+06, 1.4163E+06, 1.4189E+06, 1.4215E+06, 1.4241E+06,
     +  1.4267E+06, 1.4293E+06, 1.4319E+06, 1.4345E+06, 1.4371E+06,
     + 1.4397E+06, 1.4423E+06, 1.4450E+06, 1.4476E+06, 1.4502E+06,
     +  1.4529E+06, 1.4555E+06, 1.4581E+06, 1.4608E+06, 1.4634E+06,
     + 1.4661E+06, 1.4687E+06, 1.4714E+06, 1.4740E+06, 1.4767E+06,
     +  1.4794E+06, 1.4820E+06, 1.4847E+06, 1.4874E+06, 1.4901E+06,
     + 1.4928E+06, 1.4955E+06, 1.4981E+06, 1.5008E+06, 1.5035E+06,
     +  1.5062E+06, 1.5089E+06, 1.5117E+06, 1.5144E+06, 1.5171E+06,
     + 1.5198E+06, 1.5225E+06, 1.5253E+06, 1.5280E+06, 1.5307E+06,
     +  1.5335E+06, 1.5362E+06, 1.5390E+06, 1.5417E+06, 1.5445E+06,
     + 1.5472E+06, 1.5500E+06, 1.5527E+06, 1.5555E+06, 1.5583E+06,
     +  1.5611E+06, 1.5638E+06, 1.5666E+06, 1.5694E+06, 1.5722E+06,
     + 1.5750E+06, 1.5778E+06, 1.5806E+06, 1.5834E+06, 1.5862E+06,
     +  1.5890E+06, 1.5918E+06, 1.5946E+06, 1.5975E+06, 1.6003E+06,
     + 1.6031E+06, 1.6059E+06, 1.6088E+06, 1.6116E+06, 1.6145E+06,
     +  1.6173E+06, 1.6202E+06, 1.6230E+06, 1.6259E+06, 1.6287E+06,
     + 1.6316E+06, 1.6345E+06, 1.6373E+06, 1.6402E+06, 1.6431E+06,
     +  1.6460E+06, 1.6489E+06, 1.6518E+06, 1.6547E+06, 1.6576E+06,
     + 1.6605E+06, 1.6634E+06, 1.6663E+06, 1.6692E+06, 1.6721E+06,
     +  1.6750E+06, 1.6779E+06, 1.6809E+06, 1.6838E+06, 1.6867E+06,
     + 1.6897E+06, 1.6926E+06, 1.6956E+06, 1.6985E+06, 1.7015E+06,
     +  1.7044E+06, 1.7074E+06, 1.7104E+06, 1.7133E+06, 1.7163E+06,
     + 1.7193E+06, 1.7223E+06, 1.7252E+06, 1.7282E+06, 1.7312E+06,
     +  1.7342E+06, 1.7372E+06, 1.7402E+06, 1.7432E+06, 1.7462E+06,
     + 1.7492E+06, 1.7523E+06, 1.7553E+06, 1.7583E+06, 1.7613E+06,
     +  1.7644E+06, 1.7674E+06, 1.7704E+06, 1.7735E+06, 1.7765E+06,
     + 1.7796E+06, 1.7826E+06, 1.7857E+06, 1.7888E+06, 1.7918E+06,
     +  1.7949E+06, 1.7980E+06, 1.8011E+06, 1.8041E+06, 1.8072E+06,
     + 1.8103E+06, 1.8134E+06, 1.8165E+06, 1.8196E+06, 1.8227E+06,
     +  1.8258E+06, 1.8289E+06, 1.8320E+06, 1.8352E+06, 1.8383E+06,
     + 1.8414E+06, 1.8446E+06, 1.8477E+06, 1.8508E+06, 1.8540E+06,
     +  1.8571E+06, 1.8603E+06, 1.8634E+06, 1.8666E+06, 1.8697E+06,
     + 1.8729E+06, 1.8761E+06, 1.8793E+06, 1.8824E+06, 1.8856E+06,
     +  1.8888E+06, 1.8920E+06, 1.8952E+06, 1.8984E+06, 1.9016E+06,
     + 1.9048E+06, 1.9080E+06, 1.9112E+06, 1.9144E+06, 1.9176E+06,
     +  1.9209E+06, 1.9241E+06, 1.9273E+06, 1.9306E+06, 1.9338E+06/
      data (QofT(10,J),J=2409,4510)/ 1.9370E+06, 1.9403E+06,
     +  1.9435E+06, 1.9468E+06, 1.9501E+06, 1.9533E+06, 1.9566E+06,
     +  1.9599E+06,
     + 1.9631E+06, 1.9664E+06, 1.9697E+06, 1.9730E+06, 1.9763E+06,
     +  1.9796E+06, 1.9829E+06, 1.9862E+06, 1.9895E+06, 1.9928E+06,
     + 1.9961E+06, 1.9994E+06, 2.0027E+06, 2.0061E+06, 2.0094E+06,
     +  2.0127E+06, 2.0161E+06, 2.0194E+06, 2.0227E+06, 2.0261E+06,
     + 2.0295E+06, 2.0328E+06, 2.0362E+06, 2.0395E+06, 2.0429E+06,
     +  2.0463E+06, 2.0497E+06, 2.0530E+06, 2.0564E+06, 2.0598E+06,
     + 2.0632E+06, 2.0666E+06, 2.0700E+06, 2.0734E+06, 2.0768E+06,
     +  2.0802E+06, 2.0836E+06, 2.0871E+06, 2.0905E+06, 2.0939E+06,
     + 2.0974E+06, 2.1008E+06, 2.1042E+06, 2.1077E+06, 2.1111E+06,
     +  2.1146E+06, 2.1180E+06, 2.1215E+06, 2.1250E+06, 2.1284E+06,
     + 2.1319E+06, 2.1354E+06, 2.1389E+06, 2.1424E+06, 2.1458E+06,
     +  2.1493E+06, 2.1528E+06, 2.1563E+06, 2.1598E+06, 2.1633E+06,
     + 2.1669E+06, 2.1704E+06, 2.1739E+06, 2.1774E+06, 2.1810E+06,
     +  2.1845E+06, 2.1880E+06, 2.1916E+06, 2.1951E+06, 2.1987E+06,
     + 2.2022E+06, 2.2058E+06, 2.2093E+06, 2.2129E+06, 2.2165E+06,
     +  2.2200E+06, 2.2236E+06, 2.2272E+06, 2.2308E+06, 2.2344E+06,
     + 2.2380E+06, 2.2416E+06, 2.2452E+06, 2.2488E+06, 2.2524E+06,
     +  2.2560E+06, 2.2596E+06, 2.2633E+06, 2.2669E+06, 2.2705E+06,
     + 2.2742E+06, 2.2778E+06, 2.2814E+06, 2.2851E+06, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       837
      data (QofT(11,J),J=1,2408)/ 2.7823E+01, 5.0948E+01, 7.4240E+01,
     +  9.7570E+01, 1.2092E+02, 1.4427E+02, 1.6762E+02, 1.9098E+02,
     + 2.1434E+02, 2.3771E+02, 2.6107E+02, 2.8443E+02, 3.0780E+02,
     +  3.3116E+02, 3.5453E+02, 3.7789E+02, 4.0126E+02, 4.2462E+02,
     + 4.4799E+02, 4.7136E+02, 5.1809E+02, 5.6482E+02, 6.1156E+02,
     +  6.5829E+02, 7.0502E+02, 7.5176E+02, 7.9850E+02, 8.4523E+02,
     + 8.9197E+02, 9.3870E+02, 9.8544E+02, 1.0322E+03, 1.0789E+03,
     +  1.1257E+03, 1.1724E+03, 1.2191E+03, 1.2659E+03, 1.3126E+03,
     + 1.3593E+03, 1.4061E+03, 1.4528E+03, 1.4996E+03, 1.5463E+03,
     +  1.5930E+03, 1.6398E+03, 1.6865E+03, 1.7333E+03, 1.7800E+03,
     + 1.8268E+03, 1.8735E+03, 1.9203E+03, 1.9670E+03, 2.0138E+03,
     +  2.0606E+03, 2.1073E+03, 2.1541E+03, 2.2009E+03, 2.2477E+03,
     + 2.2945E+03, 2.3414E+03, 2.3882E+03, 2.4351E+03, 2.4820E+03,
     +  2.5289E+03, 2.5758E+03, 2.6228E+03, 2.6698E+03, 2.7168E+03,
     + 2.7638E+03, 2.8109E+03, 2.8581E+03, 2.9053E+03, 2.9525E+03,
     +  2.9998E+03, 3.0471E+03, 3.0945E+03, 3.1420E+03, 3.1896E+03,
     + 3.2372E+03, 3.2849E+03, 3.3326E+03, 3.3805E+03, 3.4284E+03,
     +  3.4765E+03, 3.5246E+03, 3.5728E+03, 3.6212E+03, 3.6697E+03,
     + 3.7182E+03, 3.7669E+03, 3.8157E+03, 3.8647E+03, 3.9138E+03,
     +  3.9630E+03, 4.0124E+03, 4.0619E+03, 4.1116E+03, 4.1614E+03,
     + 4.2114E+03, 4.2615E+03, 4.3118E+03, 4.3623E+03, 4.4130E+03,
     +  4.4639E+03, 4.5149E+03, 4.5662E+03, 4.6176E+03, 4.6692E+03,
     + 4.7211E+03, 4.7731E+03, 4.8254E+03, 4.8779E+03, 4.9306E+03,
     +  4.9835E+03, 5.0367E+03, 5.0901E+03, 5.1437E+03, 5.1976E+03,
     + 5.2518E+03, 5.3061E+03, 5.3608E+03, 5.4157E+03, 5.4708E+03,
     +  5.5263E+03, 5.5820E+03, 5.6379E+03, 5.6942E+03, 5.7507E+03,
     + 5.8075E+03, 5.8646E+03, 5.9220E+03, 5.9797E+03, 6.0377E+03,
     +  6.0960E+03, 6.1546E+03, 6.2135E+03, 6.2727E+03, 6.3323E+03,
     + 6.3922E+03, 6.4523E+03, 6.5129E+03, 6.5737E+03, 6.6349E+03,
     +  6.6964E+03, 6.7583E+03, 6.8205E+03, 6.8830E+03, 6.9459E+03,
     + 7.0092E+03, 7.0728E+03, 7.1368E+03, 7.2012E+03, 7.2659E+03,
     +  7.3309E+03, 7.3964E+03, 7.4622E+03, 7.5284E+03, 7.5950E+03,
     + 7.6620E+03, 7.7294E+03, 7.7971E+03, 7.8653E+03, 7.9339E+03,
     +  8.0028E+03, 8.0722E+03, 8.1420E+03, 8.2121E+03, 8.2827E+03,
     + 8.3537E+03, 8.4252E+03, 8.4970E+03, 8.5693E+03, 8.6420E+03,
     +  8.7152E+03, 8.7887E+03, 8.8628E+03, 8.9372E+03, 9.0121E+03,
     + 9.0875E+03, 9.1632E+03, 9.2395E+03, 9.3162E+03, 9.3934E+03,
     +  9.4710E+03, 9.5491E+03, 9.6276E+03, 9.7067E+03, 9.7862E+03,
     + 9.8661E+03, 9.9466E+03, 1.0028E+04, 1.0109E+04, 1.0191E+04,
     +  1.0273E+04, 1.0356E+04, 1.0440E+04, 1.0523E+04, 1.0608E+04,
     + 1.0693E+04, 1.0778E+04, 1.0864E+04, 1.0951E+04, 1.1037E+04,
     +  1.1125E+04, 1.1213E+04, 1.1302E+04, 1.1391E+04, 1.1480E+04,
     + 1.1570E+04, 1.1661E+04, 1.1752E+04, 1.1844E+04, 1.1936E+04,
     +  1.2029E+04, 1.2122E+04, 1.2216E+04, 1.2311E+04, 1.2406E+04,
     + 1.2502E+04, 1.2598E+04, 1.2695E+04, 1.2792E+04, 1.2890E+04,
     +  1.2989E+04, 1.3088E+04, 1.3187E+04, 1.3288E+04, 1.3389E+04,
     + 1.3490E+04, 1.3592E+04, 1.3695E+04, 1.3798E+04, 1.3902E+04,
     +  1.4006E+04, 1.4112E+04, 1.4217E+04, 1.4324E+04, 1.4431E+04,
     + 1.4538E+04, 1.4647E+04, 1.4755E+04, 1.4865E+04, 1.4975E+04,
     +  1.5086E+04, 1.5197E+04, 1.5309E+04, 1.5422E+04, 1.5536E+04,
     + 1.5650E+04, 1.5764E+04, 1.5880E+04, 1.5996E+04, 1.6113E+04,
     +  1.6230E+04, 1.6348E+04, 1.6467E+04, 1.6586E+04, 1.6707E+04,
     + 1.6828E+04, 1.6949E+04, 1.7071E+04, 1.7194E+04, 1.7318E+04,
     +  1.7442E+04, 1.7567E+04, 1.7693E+04, 1.7820E+04, 1.7947E+04,
     + 1.8075E+04, 1.8204E+04, 1.8333E+04, 1.8463E+04, 1.8594E+04,
     +  1.8726E+04, 1.8858E+04, 1.8991E+04, 1.9125E+04, 1.9260E+04,
     + 1.9395E+04, 1.9532E+04, 1.9669E+04, 1.9806E+04, 1.9945E+04,
     +  2.0084E+04, 2.0224E+04, 2.0365E+04, 2.0507E+04, 2.0649E+04,
     + 2.0792E+04, 2.0936E+04, 2.1081E+04, 2.1227E+04, 2.1373E+04,
     +  2.1521E+04, 2.1669E+04, 2.1818E+04, 2.1967E+04, 2.2118E+04,
     + 2.2269E+04, 2.2422E+04, 2.2575E+04, 2.2729E+04, 2.2884E+04,
     +  2.3039E+04, 2.3196E+04, 2.3353E+04, 2.3511E+04, 2.3670E+04,
     + 2.3830E+04, 2.3991E+04, 2.4153E+04, 2.4316E+04, 2.4479E+04,
     +  2.4644E+04, 2.4809E+04, 2.4975E+04, 2.5142E+04, 2.5310E+04,
     + 2.5479E+04, 2.5649E+04, 2.5820E+04, 2.5991E+04, 2.6164E+04,
     +  2.6337E+04, 2.6512E+04, 2.6687E+04, 2.6864E+04, 2.7041E+04,
     + 2.7219E+04, 2.7398E+04, 2.7578E+04, 2.7760E+04, 2.7942E+04,
     +  2.8125E+04, 2.8309E+04, 2.8494E+04, 2.8680E+04, 2.8867E+04,
     + 2.9055E+04, 2.9244E+04, 2.9434E+04, 2.9625E+04, 2.9817E+04,
     +  3.0010E+04, 3.0204E+04, 3.0399E+04, 3.0595E+04, 3.0792E+04,
     + 3.0990E+04, 3.1189E+04, 3.1390E+04, 3.1591E+04, 3.1793E+04,
     +  3.1997E+04, 3.2201E+04, 3.2407E+04, 3.2613E+04, 3.2821E+04,
     + 3.3030E+04, 3.3239E+04, 3.3450E+04, 3.3662E+04, 3.3876E+04,
     +  3.4090E+04, 3.4305E+04, 3.4522E+04, 3.4739E+04, 3.4958E+04,
     + 3.5178E+04, 3.5398E+04, 3.5620E+04, 3.5844E+04, 3.6068E+04,
     +  3.6294E+04, 3.6520E+04, 3.6748E+04, 3.6977E+04, 3.7207E+04,
     + 3.7438E+04, 3.7671E+04, 3.7904E+04, 3.8139E+04, 3.8375E+04,
     +  3.8612E+04, 3.8851E+04, 3.9091E+04, 3.9331E+04, 3.9573E+04,
     + 3.9817E+04, 4.0061E+04, 4.0307E+04, 4.0554E+04, 4.0802E+04,
     +  4.1051E+04, 4.1302E+04, 4.1554E+04, 4.1807E+04, 4.2062E+04,
     + 4.2317E+04, 4.2574E+04, 4.2833E+04, 4.3092E+04, 4.3353E+04,
     +  4.3615E+04, 4.3879E+04, 4.4143E+04, 4.4409E+04, 4.4677E+04,
     + 4.4945E+04, 4.5215E+04, 4.5487E+04, 4.5759E+04, 4.6033E+04,
     +  4.6309E+04, 4.6585E+04, 4.6863E+04, 4.7143E+04, 4.7424E+04,
     + 4.7706E+04, 4.7989E+04, 4.8274E+04, 4.8560E+04, 4.8848E+04,
     +  4.9137E+04, 4.9428E+04, 4.9720E+04, 5.0013E+04, 5.0308E+04,
     + 5.0604E+04, 5.0901E+04, 5.1200E+04, 5.1501E+04, 5.1803E+04,
     +  5.2106E+04, 5.2411E+04, 5.2717E+04, 5.3025E+04, 5.3334E+04,
     + 5.3644E+04, 5.3957E+04, 5.4270E+04, 5.4585E+04, 5.4902E+04,
     +  5.5220E+04, 5.5540E+04, 5.5861E+04, 5.6184E+04, 5.6508E+04,
     + 5.6833E+04, 5.7161E+04, 5.7490E+04, 5.7820E+04, 5.8152E+04,
     +  5.8485E+04, 5.8820E+04, 5.9157E+04, 5.9495E+04, 5.9835E+04,
     + 6.0176E+04, 6.0519E+04, 6.0864E+04, 6.1210E+04, 6.1558E+04,
     +  6.1907E+04, 6.2258E+04, 6.2611E+04, 6.2966E+04, 6.3321E+04,
     + 6.3679E+04, 6.4038E+04, 6.4399E+04, 6.4762E+04, 6.5126E+04,
     +  6.5492E+04, 6.5860E+04, 6.6229E+04, 6.6600E+04, 6.6973E+04,
     + 6.7347E+04, 6.7723E+04, 6.8101E+04, 6.8481E+04, 6.8862E+04,
     +  6.9245E+04, 6.9630E+04, 7.0017E+04, 7.0405E+04, 7.0795E+04,
     + 7.1187E+04, 7.1581E+04, 7.1976E+04, 7.2374E+04, 7.2773E+04,
     +  7.3173E+04, 7.3576E+04, 7.3980E+04, 7.4387E+04, 7.4795E+04,
     + 7.5205E+04, 7.5617E+04, 7.6030E+04, 7.6446E+04, 7.6863E+04,
     +  7.7282E+04, 7.7703E+04, 7.8126E+04, 7.8551E+04, 7.8978E+04,
     + 7.9407E+04, 7.9837E+04, 8.0270E+04, 8.0704E+04, 8.1140E+04,
     +  8.1579E+04, 8.2019E+04, 8.2461E+04, 8.2905E+04, 8.3351E+04,
     + 8.3799E+04, 8.4249E+04, 8.4701E+04, 8.5155E+04, 8.5611E+04,
     +  8.6069E+04, 8.6529E+04, 8.6991E+04, 8.7455E+04, 8.7921E+04,
     + 8.8389E+04, 8.8859E+04, 8.9331E+04, 8.9806E+04, 9.0282E+04,
     +  9.0760E+04, 9.1241E+04, 9.1723E+04, 9.2208E+04, 9.2694E+04,
     + 9.3183E+04, 9.3674E+04, 9.4167E+04, 9.4662E+04, 9.5159E+04,
     +  9.5659E+04, 9.6160E+04, 9.6664E+04, 9.7170E+04, 9.7678E+04,
     + 9.8188E+04, 9.8701E+04, 9.9215E+04, 9.9732E+04, 1.0025E+05,
     +  1.0077E+05, 1.0130E+05, 1.0182E+05, 1.0235E+05, 1.0288E+05,
     + 1.0341E+05, 1.0395E+05, 1.0448E+05, 1.0502E+05, 1.0556E+05,
     +  1.0611E+05, 1.0665E+05, 1.0720E+05, 1.0775E+05, 1.0831E+05,
     + 1.0886E+05, 1.0942E+05, 1.0998E+05, 1.1054E+05, 1.1111E+05,
     +  1.1167E+05, 1.1224E+05, 1.1281E+05, 1.1339E+05, 1.1396E+05,
     + 1.1454E+05, 1.1512E+05, 1.1571E+05, 1.1629E+05, 1.1688E+05,
     +  1.1747E+05, 1.1807E+05, 1.1866E+05, 1.1926E+05, 1.1986E+05,
     + 1.2047E+05, 1.2107E+05, 1.2168E+05, 1.2229E+05, 1.2290E+05,
     +  1.2352E+05, 1.2414E+05, 1.2476E+05, 1.2538E+05, 1.2601E+05,
     + 1.2664E+05, 1.2727E+05, 1.2790E+05, 1.2854E+05, 1.2918E+05,
     +  1.2982E+05, 1.3046E+05, 1.3111E+05, 1.3176E+05, 1.3241E+05,
     + 1.3306E+05, 1.3372E+05, 1.3438E+05, 1.3504E+05, 1.3571E+05,
     +  1.3638E+05, 1.3705E+05, 1.3772E+05, 1.3840E+05, 1.3908E+05,
     + 1.3976E+05, 1.4044E+05, 1.4113E+05, 1.4182E+05, 1.4251E+05,
     +  1.4320E+05, 1.4390E+05, 1.4460E+05, 1.4531E+05, 1.4601E+05,
     + 1.4672E+05, 1.4743E+05, 1.4815E+05, 1.4887E+05, 1.4959E+05,
     +  1.5031E+05, 1.5104E+05, 1.5176E+05, 1.5250E+05, 1.5323E+05,
     + 1.5397E+05, 1.5471E+05, 1.5545E+05, 1.5620E+05, 1.5695E+05,
     +  1.5770E+05, 1.5845E+05, 1.5921E+05, 1.5997E+05, 1.6074E+05,
     + 1.6150E+05, 1.6227E+05, 1.6305E+05, 1.6382E+05, 1.6460E+05,
     +  1.6538E+05, 1.6617E+05, 1.6696E+05, 1.6775E+05, 1.6854E+05,
     + 1.6934E+05, 1.7014E+05, 1.7094E+05, 1.7175E+05, 1.7256E+05,
     +  1.7337E+05, 1.7419E+05, 1.7501E+05, 1.7583E+05, 1.7665E+05,
     + 1.7748E+05, 1.7831E+05, 1.7915E+05, 1.7999E+05, 1.8083E+05,
     +  1.8167E+05, 1.8252E+05, 1.8337E+05, 1.8423E+05, 1.8508E+05,
     + 1.8594E+05, 1.8681E+05, 1.8767E+05, 1.8854E+05, 1.8942E+05,
     +  1.9030E+05, 1.9118E+05, 1.9206E+05, 1.9295E+05, 1.9384E+05,
     + 1.9473E+05, 1.9563E+05, 1.9653E+05, 1.9743E+05, 1.9834E+05,
     +  1.9925E+05, 2.0016E+05, 2.0108E+05, 2.0200E+05, 2.0293E+05,
     + 2.0385E+05, 2.0478E+05, 2.0572E+05, 2.0666E+05, 2.0760E+05,
     +  2.0854E+05, 2.0949E+05, 2.1044E+05, 2.1140E+05, 2.1236E+05,
     + 2.1332E+05, 2.1429E+05, 2.1526E+05, 2.1623E+05, 2.1721E+05,
     +  2.1819E+05, 2.1917E+05, 2.2016E+05, 2.2115E+05, 2.2215E+05,
     + 2.2315E+05, 2.2415E+05, 2.2515E+05, 2.2616E+05, 2.2718E+05,
     +  2.2819E+05, 2.2922E+05, 2.3024E+05, 2.3127E+05, 2.3230E+05,
     + 2.3334E+05, 2.3438E+05, 2.3542E+05, 2.3647E+05, 2.3752E+05,
     +  2.3857E+05, 2.3963E+05, 2.4069E+05, 2.4176E+05, 2.4283E+05,
     + 2.4390E+05, 2.4498E+05, 2.4606E+05, 2.4715E+05, 2.4824E+05,
     +  2.4933E+05, 2.5043E+05, 2.5153E+05, 2.5263E+05, 2.5374E+05,
     + 2.5486E+05, 2.5597E+05, 2.5709E+05, 2.5822E+05, 2.5935E+05,
     +  2.6048E+05, 2.6162E+05, 2.6276E+05, 2.6390E+05, 2.6505E+05,
     + 2.6621E+05, 2.6736E+05, 2.6852E+05, 2.6969E+05, 2.7086E+05,
     +  2.7203E+05, 2.7321E+05, 2.7439E+05, 2.7558E+05, 2.7677E+05,
     + 2.7796E+05, 2.7916E+05, 2.8037E+05, 2.8157E+05, 2.8278E+05,
     +  2.8400E+05, 2.8522E+05, 2.8644E+05, 2.8767E+05, 2.8890E+05,
     + 2.9014E+05, 2.9138E+05, 2.9263E+05, 2.9388E+05, 2.9513E+05,
     +  2.9639E+05, 2.9765E+05, 2.9892E+05, 3.0019E+05, 3.0147E+05,
     + 3.0275E+05, 3.0403E+05, 3.0532E+05, 3.0662E+05, 3.0792E+05,
     +  3.0922E+05, 3.1053E+05, 3.1184E+05, 3.1315E+05, 3.1447E+05,
     + 3.1580E+05, 3.1713E+05, 3.1846E+05, 3.1980E+05, 3.2115E+05,
     +  3.2249E+05, 3.2385E+05, 3.2520E+05, 3.2656E+05, 3.2793E+05,
     + 3.2930E+05, 3.3068E+05, 3.3206E+05, 3.3344E+05, 3.3483E+05,
     +  3.3623E+05, 3.3763E+05, 3.3903E+05, 3.4044E+05, 3.4185E+05,
     + 3.4327E+05, 3.4469E+05, 3.4612E+05, 3.4755E+05, 3.4899E+05,
     +  3.5043E+05, 3.5188E+05, 3.5333E+05, 3.5479E+05, 3.5625E+05,
     + 3.5772E+05, 3.5919E+05, 3.6066E+05, 3.6214E+05, 3.6363E+05,
     +  3.6512E+05, 3.6662E+05, 3.6812E+05, 3.6962E+05, 3.7114E+05,
     + 3.7265E+05, 3.7417E+05, 3.7570E+05, 3.7723E+05, 3.7877E+05,
     +  3.8031E+05, 3.8185E+05, 3.8340E+05, 3.8496E+05, 3.8652E+05,
     + 3.8809E+05, 3.8966E+05, 3.9124E+05, 3.9282E+05, 3.9441E+05,
     +  3.9600E+05, 3.9760E+05, 3.9920E+05, 4.0081E+05, 4.0242E+05,
     + 4.0404E+05, 4.0566E+05, 4.0729E+05, 4.0893E+05, 4.1057E+05,
     +  4.1221E+05, 4.1386E+05, 4.1552E+05, 4.1718E+05, 4.1885E+05,
     + 4.2052E+05, 4.2220E+05, 4.2388E+05, 4.2557E+05, 4.2726E+05,
     +  4.2896E+05, 4.3066E+05, 4.3237E+05, 4.3409E+05, 4.3581E+05,
     + 4.3754E+05, 4.3927E+05, 4.4101E+05, 4.4275E+05, 4.4450E+05,
     +  4.4625E+05, 4.4801E+05, 4.4978E+05, 4.5155E+05, 4.5333E+05,
     + 4.5511E+05, 4.5690E+05, 4.5869E+05, 4.6049E+05, 4.6229E+05,
     +  4.6410E+05, 4.6592E+05, 4.6774E+05, 4.6957E+05, 4.7141E+05,
     + 4.7325E+05, 4.7509E+05, 4.7694E+05, 4.7880E+05, 4.8066E+05,
     +  4.8253E+05, 4.8441E+05, 4.8629E+05, 4.8817E+05, 4.9006E+05,
     + 4.9196E+05, 4.9387E+05, 4.9578E+05, 4.9769E+05, 4.9962E+05,
     +  5.0154E+05, 5.0348E+05, 5.0542E+05, 5.0736E+05, 5.0932E+05,
     + 5.1127E+05, 5.1324E+05, 5.1521E+05, 5.1719E+05, 5.1917E+05,
     +  5.2116E+05, 5.2315E+05, 5.2515E+05, 5.2716E+05, 5.2917E+05,
     + 5.3119E+05, 5.3322E+05, 5.3525E+05, 5.3729E+05, 5.3933E+05,
     +  5.4138E+05, 5.4344E+05, 5.4550E+05, 5.4757E+05, 5.4965E+05,
     + 5.5173E+05, 5.5382E+05, 5.5592E+05, 5.5802E+05, 5.6013E+05,
     +  5.6224E+05, 5.6436E+05, 5.6649E+05, 5.6862E+05, 5.7076E+05,
     + 5.7291E+05, 5.7506E+05, 5.7722E+05, 5.7939E+05, 5.8156E+05,
     +  5.8374E+05, 5.8592E+05, 5.8812E+05, 5.9032E+05, 5.9252E+05,
     + 5.9473E+05, 5.9695E+05, 5.9918E+05, 6.0141E+05, 6.0365E+05,
     +  6.0589E+05, 6.0815E+05, 6.1041E+05, 6.1267E+05, 6.1494E+05,
     + 6.1722E+05, 6.1951E+05, 6.2180E+05, 6.2410E+05, 6.2641E+05,
     +  6.2872E+05, 6.3104E+05, 6.3337E+05, 6.3570E+05, 6.3804E+05,
     + 6.4039E+05, 6.4275E+05, 6.4511E+05, 6.4748E+05, 6.4985E+05,
     +  6.5223E+05, 6.5462E+05, 6.5702E+05, 6.5943E+05, 6.6184E+05,
     + 6.6425E+05, 6.6668E+05, 6.6911E+05, 6.7155E+05, 6.7400E+05,
     +  6.7645E+05, 6.7891E+05, 6.8138E+05, 6.8385E+05, 6.8634E+05,
     + 6.8882E+05, 6.9132E+05, 6.9383E+05, 6.9634E+05, 6.9886E+05,
     +  7.0138E+05, 7.0391E+05, 7.0645E+05, 7.0900E+05, 7.1156E+05,
     + 7.1412E+05, 7.1669E+05, 7.1927E+05, 7.2185E+05, 7.2444E+05,
     +  7.2704E+05, 7.2965E+05, 7.3227E+05, 7.3489E+05, 7.3752E+05,
     + 7.4015E+05, 7.4280E+05, 7.4545E+05, 7.4811E+05, 7.5078E+05,
     +  7.5345E+05, 7.5614E+05, 7.5883E+05, 7.6152E+05, 7.6423E+05,
     + 7.6694E+05, 7.6966E+05, 7.7239E+05, 7.7513E+05, 7.7787E+05,
     +  7.8063E+05, 7.8338E+05, 7.8615E+05, 7.8893E+05, 7.9171E+05,
     + 7.9450E+05, 7.9730E+05, 8.0011E+05, 8.0292E+05, 8.0574E+05,
     +  8.0858E+05, 8.1141E+05, 8.1426E+05, 8.1711E+05, 8.1998E+05,
     + 8.2285E+05, 8.2572E+05, 8.2861E+05, 8.3151E+05, 8.3441E+05,
     +  8.3732E+05, 8.4024E+05, 8.4316E+05, 8.4610E+05, 8.4904E+05,
     + 8.5199E+05, 8.5495E+05, 8.5792E+05, 8.6090E+05, 8.6388E+05,
     +  8.6687E+05, 8.6987E+05, 8.7288E+05, 8.7590E+05, 8.7892E+05,
     + 8.8196E+05, 8.8500E+05, 8.8805E+05, 8.9111E+05, 8.9418E+05,
     +  8.9725E+05, 9.0033E+05, 9.0343E+05, 9.0653E+05, 9.0964E+05,
     + 9.1275E+05, 9.1588E+05, 9.1902E+05, 9.2216E+05, 9.2531E+05,
     +  9.2847E+05, 9.3164E+05, 9.3482E+05, 9.3800E+05, 9.4120E+05,
     + 9.4440E+05, 9.4761E+05, 9.5084E+05, 9.5406E+05, 9.5730E+05,
     +  9.6055E+05, 9.6381E+05, 9.6707E+05, 9.7034E+05, 9.7363E+05,
     + 9.7692E+05, 9.8022E+05, 9.8352E+05, 9.8684E+05, 9.9017E+05,
     +  9.9350E+05, 9.9685E+05, 1.0002E+06, 1.0036E+06, 1.0069E+06,
     + 1.0103E+06, 1.0137E+06, 1.0171E+06, 1.0205E+06, 1.0239E+06,
     +  1.0273E+06, 1.0308E+06, 1.0342E+06, 1.0377E+06, 1.0411E+06,
     + 1.0446E+06, 1.0481E+06, 1.0516E+06, 1.0551E+06, 1.0586E+06,
     +  1.0621E+06, 1.0656E+06, 1.0692E+06, 1.0727E+06, 1.0763E+06,
     + 1.0798E+06, 1.0834E+06, 1.0870E+06, 1.0906E+06, 1.0942E+06,
     +  1.0978E+06, 1.1014E+06, 1.1050E+06, 1.1087E+06, 1.1123E+06,
     + 1.1160E+06, 1.1196E+06, 1.1233E+06, 1.1270E+06, 1.1307E+06,
     +  1.1344E+06, 1.1381E+06, 1.1418E+06, 1.1456E+06, 1.1493E+06,
     + 1.1531E+06, 1.1568E+06, 1.1606E+06, 1.1644E+06, 1.1682E+06,
     +  1.1720E+06, 1.1758E+06, 1.1796E+06, 1.1834E+06, 1.1873E+06,
     + 1.1911E+06, 1.1950E+06, 1.1988E+06, 1.2027E+06, 1.2066E+06,
     +  1.2105E+06, 1.2144E+06, 1.2183E+06, 1.2223E+06, 1.2262E+06,
     + 1.2302E+06, 1.2341E+06, 1.2381E+06, 1.2421E+06, 1.2460E+06,
     +  1.2500E+06, 1.2541E+06, 1.2581E+06, 1.2621E+06, 1.2661E+06,
     + 1.2702E+06, 1.2743E+06, 1.2783E+06, 1.2824E+06, 1.2865E+06,
     +  1.2906E+06, 1.2947E+06, 1.2988E+06, 1.3030E+06, 1.3071E+06,
     + 1.3112E+06, 1.3154E+06, 1.3196E+06, 1.3238E+06, 1.3280E+06,
     +  1.3322E+06, 1.3364E+06, 1.3406E+06, 1.3448E+06, 1.3491E+06,
     + 1.3533E+06, 1.3576E+06, 1.3619E+06, 1.3662E+06, 1.3705E+06,
     +  1.3748E+06, 1.3791E+06, 1.3834E+06, 1.3878E+06, 1.3921E+06,
     + 1.3965E+06, 1.4009E+06, 1.4052E+06, 1.4096E+06, 1.4140E+06,
     +  1.4184E+06, 1.4229E+06, 1.4273E+06, 1.4318E+06, 1.4362E+06,
     + 1.4407E+06, 1.4452E+06, 1.4497E+06, 1.4542E+06, 1.4587E+06,
     +  1.4632E+06, 1.4677E+06, 1.4723E+06, 1.4768E+06, 1.4814E+06,
     + 1.4860E+06, 1.4906E+06, 1.4952E+06, 1.4998E+06, 1.5044E+06,
     +  1.5090E+06, 1.5137E+06, 1.5183E+06, 1.5230E+06, 1.5277E+06,
     + 1.5324E+06, 1.5371E+06, 1.5418E+06, 1.5465E+06, 1.5512E+06,
     +  1.5560E+06, 1.5607E+06, 1.5655E+06, 1.5703E+06, 1.5751E+06,
     + 1.5799E+06, 1.5847E+06, 1.5895E+06, 1.5944E+06, 1.5992E+06,
     +  1.6041E+06, 1.6089E+06, 1.6138E+06, 1.6187E+06, 1.6236E+06,
     + 1.6285E+06, 1.6335E+06, 1.6384E+06, 1.6434E+06, 1.6483E+06,
     +  1.6533E+06, 1.6583E+06, 1.6633E+06, 1.6683E+06, 1.6733E+06,
     + 1.6783E+06, 1.6834E+06, 1.6884E+06, 1.6935E+06, 1.6986E+06,
     +  1.7037E+06, 1.7088E+06, 1.7139E+06, 1.7190E+06, 1.7242E+06,
     + 1.7293E+06, 1.7345E+06, 1.7397E+06, 1.7448E+06, 1.7500E+06,
     +  1.7553E+06, 1.7605E+06, 1.7657E+06, 1.7710E+06, 1.7762E+06,
     + 1.7815E+06, 1.7868E+06, 1.7921E+06, 1.7974E+06, 1.8027E+06,
     +  1.8080E+06, 1.8134E+06, 1.8187E+06, 1.8241E+06, 1.8295E+06,
     + 1.8349E+06, 1.8403E+06, 1.8457E+06, 1.8511E+06, 1.8565E+06,
     +  1.8620E+06, 1.8675E+06, 1.8729E+06, 1.8784E+06, 1.8839E+06,
     + 1.8894E+06, 1.8950E+06, 1.9005E+06, 1.9061E+06, 1.9116E+06,
     +  1.9172E+06, 1.9228E+06, 1.9284E+06, 1.9340E+06, 1.9396E+06,
     + 1.9453E+06, 1.9509E+06, 1.9566E+06, 1.9623E+06, 1.9680E+06,
     +  1.9737E+06, 1.9794E+06, 1.9851E+06, 1.9908E+06, 1.9966E+06,
     + 2.0024E+06, 2.0081E+06, 2.0139E+06, 2.0197E+06, 2.0256E+06,
     +  2.0314E+06, 2.0372E+06, 2.0431E+06, 2.0489E+06, 2.0548E+06,
     + 2.0607E+06, 2.0666E+06, 2.0725E+06, 2.0785E+06, 2.0844E+06,
     +  2.0904E+06, 2.0964E+06, 2.1023E+06, 2.1083E+06, 2.1143E+06,
     + 2.1204E+06, 2.1264E+06, 2.1325E+06, 2.1385E+06, 2.1446E+06,
     +  2.1507E+06, 2.1568E+06, 2.1629E+06, 2.1690E+06, 2.1752E+06,
     + 2.1813E+06, 2.1875E+06, 2.1937E+06, 2.1999E+06, 2.2061E+06,
     +  2.2123E+06, 2.2185E+06, 2.2248E+06, 2.2310E+06, 2.2373E+06,
     + 2.2436E+06, 2.2499E+06, 2.2562E+06, 2.2626E+06, 2.2689E+06,
     +  2.2753E+06, 2.2816E+06, 2.2880E+06, 2.2944E+06, 2.3008E+06,
     + 2.3072E+06, 2.3137E+06, 2.3201E+06, 2.3266E+06, 2.3331E+06,
     +  2.3396E+06, 2.3461E+06, 2.3526E+06, 2.3591E+06, 2.3657E+06,
     + 2.3722E+06, 2.3788E+06, 2.3854E+06, 2.3920E+06, 2.3986E+06,
     +  2.4052E+06, 2.4119E+06, 2.4185E+06, 2.4252E+06, 2.4319E+06,
     + 2.4386E+06, 2.4453E+06, 2.4520E+06, 2.4587E+06, 2.4655E+06,
     +  2.4723E+06, 2.4790E+06, 2.4858E+06, 2.4926E+06, 2.4995E+06,
     + 2.5063E+06, 2.5132E+06, 2.5200E+06, 2.5269E+06, 2.5338E+06,
     +  2.5407E+06, 2.5476E+06, 2.5546E+06, 2.5615E+06, 2.5685E+06,
     + 2.5755E+06, 2.5824E+06, 2.5895E+06, 2.5965E+06, 2.6035E+06,
     +  2.6106E+06, 2.6176E+06, 2.6247E+06, 2.6318E+06, 2.6389E+06,
     + 2.6460E+06, 2.6532E+06, 2.6603E+06, 2.6675E+06, 2.6747E+06,
     +  2.6819E+06, 2.6891E+06, 2.6963E+06, 2.7035E+06, 2.7108E+06,
     + 2.7180E+06, 2.7253E+06, 2.7326E+06, 2.7399E+06, 2.7473E+06,
     +  2.7546E+06, 2.7619E+06, 2.7693E+06, 2.7767E+06, 2.7841E+06,
     + 2.7915E+06, 2.7989E+06, 2.8064E+06, 2.8138E+06, 2.8213E+06,
     +  2.8288E+06, 2.8363E+06, 2.8438E+06, 2.8513E+06, 2.8589E+06,
     + 2.8664E+06, 2.8740E+06, 2.8816E+06, 2.8892E+06, 2.8968E+06,
     +  2.9045E+06, 2.9121E+06, 2.9198E+06, 2.9275E+06, 2.9352E+06,
     + 2.9429E+06, 2.9506E+06, 2.9583E+06, 2.9661E+06, 2.9739E+06,
     +  2.9816E+06, 2.9894E+06, 2.9973E+06, 3.0051E+06, 3.0129E+06,
     + 3.0208E+06, 3.0287E+06, 3.0366E+06, 3.0445E+06, 3.0524E+06,
     +  3.0603E+06, 3.0683E+06, 3.0762E+06, 3.0842E+06, 3.0922E+06,
     + 3.1002E+06, 3.1083E+06, 3.1163E+06, 3.1244E+06, 3.1324E+06,
     +  3.1405E+06, 3.1486E+06, 3.1568E+06, 3.1649E+06, 3.1731E+06,
     + 3.1812E+06, 3.1894E+06, 3.1976E+06, 3.2058E+06, 3.2140E+06,
     +  3.2223E+06, 3.2305E+06, 3.2388E+06, 3.2471E+06, 3.2554E+06,
     + 3.2637E+06, 3.2721E+06, 3.2804E+06, 3.2888E+06, 3.2972E+06,
     +  3.3056E+06, 3.3140E+06, 3.3224E+06, 3.3309E+06, 3.3394E+06,
     + 3.3478E+06, 3.3563E+06, 3.3648E+06, 3.3734E+06, 3.3819E+06,
     +  3.3905E+06, 3.3990E+06, 3.4076E+06, 3.4162E+06, 3.4249E+06,
     + 3.4335E+06, 3.4422E+06, 3.4508E+06, 3.4595E+06, 3.4682E+06,
     +  3.4769E+06, 3.4857E+06, 3.4944E+06, 3.5032E+06, 3.5120E+06,
     + 3.5208E+06, 3.5296E+06, 3.5384E+06, 3.5473E+06, 3.5561E+06,
     +  3.5650E+06, 3.5739E+06, 3.5828E+06, 3.5917E+06, 3.6007E+06,
     + 3.6096E+06, 3.6186E+06, 3.6276E+06, 3.6366E+06, 3.6456E+06,
     +  3.6547E+06, 3.6637E+06, 3.6728E+06, 3.6819E+06, 3.6910E+06,
     + 3.7001E+06, 3.7093E+06, 3.7184E+06, 3.7276E+06, 3.7368E+06,
     +  3.7460E+06, 3.7552E+06, 3.7644E+06, 3.7737E+06, 3.7830E+06,
     + 3.7923E+06, 3.8016E+06, 3.8109E+06, 3.8202E+06, 3.8296E+06,
     +  3.8389E+06, 3.8483E+06, 3.8577E+06, 3.8671E+06, 3.8766E+06,
     + 3.8860E+06, 3.8955E+06, 3.9050E+06, 3.9145E+06, 3.9240E+06,
     +  3.9335E+06, 3.9431E+06, 3.9527E+06, 3.9623E+06, 3.9719E+06,
     + 3.9815E+06, 3.9911E+06, 4.0008E+06, 4.0104E+06, 4.0201E+06,
     +  4.0298E+06, 4.0395E+06, 4.0493E+06, 4.0590E+06, 4.0688E+06,
     + 4.0786E+06, 4.0884E+06, 4.0982E+06, 4.1081E+06, 4.1179E+06,
     +  4.1278E+06, 4.1377E+06, 4.1476E+06, 4.1575E+06, 4.1674E+06,
     + 4.1774E+06, 4.1874E+06, 4.1974E+06, 4.2074E+06, 4.2174E+06,
     +  4.2275E+06, 4.2375E+06, 4.2476E+06, 4.2577E+06, 4.2678E+06,
     + 4.2779E+06, 4.2881E+06, 4.2982E+06, 4.3084E+06, 4.3186E+06,
     +  4.3288E+06, 4.3391E+06, 4.3493E+06, 4.3596E+06, 4.3699E+06,
     + 4.3802E+06, 4.3905E+06, 4.4008E+06, 4.4112E+06, 4.4215E+06,
     +  4.4319E+06, 4.4423E+06, 4.4528E+06, 4.4632E+06, 4.4736E+06,
     + 4.4841E+06, 4.4946E+06, 4.5051E+06, 4.5157E+06, 4.5262E+06,
     +  4.5368E+06, 4.5473E+06, 4.5579E+06, 4.5686E+06, 4.5792E+06,
     + 4.5898E+06, 4.6005E+06, 4.6112E+06, 4.6219E+06, 4.6326E+06,
     +  4.6433E+06, 4.6541E+06, 4.6649E+06, 4.6757E+06, 4.6865E+06,
     + 4.6973E+06, 4.7081E+06, 4.7190E+06, 4.7299E+06, 4.7408E+06,
     +  4.7517E+06, 4.7626E+06, 4.7736E+06, 4.7845E+06, 4.7955E+06,
     + 4.8065E+06, 4.8176E+06, 4.8286E+06, 4.8396E+06, 4.8507E+06,
     +  4.8618E+06, 4.8729E+06, 4.8841E+06, 4.8952E+06, 4.9064E+06,
     + 4.9175E+06, 4.9287E+06, 4.9400E+06, 4.9512E+06, 4.9625E+06,
     +  4.9737E+06, 4.9850E+06, 4.9963E+06, 5.0076E+06, 5.0190E+06,
     + 5.0304E+06, 5.0417E+06, 5.0531E+06, 5.0645E+06, 5.0760E+06,
     +  5.0874E+06, 5.0989E+06, 5.1104E+06, 5.1219E+06, 5.1334E+06,
     + 5.1450E+06, 5.1565E+06, 5.1681E+06, 5.1797E+06, 5.1913E+06,
     +  5.2029E+06, 5.2146E+06, 5.2263E+06, 5.2380E+06, 5.2497E+06,
     + 5.2614E+06, 5.2731E+06, 5.2849E+06, 5.2967E+06, 5.3085E+06,
     +  5.3203E+06, 5.3321E+06, 5.3440E+06, 5.3558E+06, 5.3677E+06,
     + 5.3796E+06, 5.3916E+06, 5.4035E+06, 5.4155E+06, 5.4274E+06,
     +  5.4394E+06, 5.4515E+06, 5.4635E+06, 5.4756E+06, 5.4876E+06,
     + 5.4997E+06, 5.5118E+06, 5.5240E+06, 5.5361E+06, 5.5483E+06,
     +  5.5605E+06, 5.5727E+06, 5.5849E+06, 5.5971E+06, 5.6094E+06,
     + 5.6217E+06, 5.6340E+06, 5.6463E+06, 5.6586E+06, 5.6710E+06,
     +  5.6833E+06, 5.6957E+06, 5.7081E+06, 5.7205E+06, 5.7330E+06,
     + 5.7455E+06, 5.7579E+06, 5.7704E+06, 5.7830E+06, 5.7955E+06,
     +  5.8080E+06, 5.8206E+06, 5.8332E+06, 5.8458E+06, 5.8585E+06,
     + 5.8711E+06, 5.8838E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT(11,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       737
      data (QofT(12,J),J=1,2408)/ 8.2487E+01, 1.4897E+02, 2.1686E+02,
     +  2.8488E+02, 3.5296E+02, 4.2105E+02, 4.8916E+02, 5.5728E+02,
     + 6.2540E+02, 6.9353E+02, 7.6166E+02, 8.2979E+02, 8.9792E+02,
     +  9.6606E+02, 1.0342E+03, 1.1023E+03, 1.1705E+03, 1.2386E+03,
     + 1.3067E+03, 1.3749E+03, 1.5112E+03, 1.6474E+03, 1.7837E+03,
     +  1.9200E+03, 2.0563E+03, 2.1926E+03, 2.3289E+03, 2.4652E+03,
     + 2.6015E+03, 2.7377E+03, 2.8740E+03, 3.0103E+03, 3.1466E+03,
     +  3.2829E+03, 3.4192E+03, 3.5555E+03, 3.6918E+03, 3.8281E+03,
     + 3.9644E+03, 4.1007E+03, 4.2370E+03, 4.3733E+03, 4.5096E+03,
     +  4.6459E+03, 4.7822E+03, 4.9185E+03, 5.0548E+03, 5.1911E+03,
     + 5.3275E+03, 5.4638E+03, 5.6001E+03, 5.7365E+03, 5.8729E+03,
     +  6.0092E+03, 6.1457E+03, 6.2821E+03, 6.4185E+03, 6.5550E+03,
     + 6.6915E+03, 6.8281E+03, 6.9647E+03, 7.1013E+03, 7.2380E+03,
     +  7.3748E+03, 7.5117E+03, 7.6486E+03, 7.7856E+03, 7.9227E+03,
     + 8.0599E+03, 8.1972E+03, 8.3347E+03, 8.4722E+03, 8.6100E+03,
     +  8.7478E+03, 8.8859E+03, 9.0241E+03, 9.1625E+03, 9.3010E+03,
     + 9.4398E+03, 9.5788E+03, 9.7181E+03, 9.8576E+03, 9.9973E+03,
     +  1.0137E+04, 1.0278E+04, 1.0418E+04, 1.0559E+04, 1.0700E+04,
     + 1.0842E+04, 1.0984E+04, 1.1126E+04, 1.1269E+04, 1.1412E+04,
     +  1.1555E+04, 1.1699E+04, 1.1843E+04, 1.1988E+04, 1.2133E+04,
     + 1.2279E+04, 1.2425E+04, 1.2571E+04, 1.2718E+04, 1.2866E+04,
     +  1.3014E+04, 1.3163E+04, 1.3312E+04, 1.3462E+04, 1.3612E+04,
     + 1.3763E+04, 1.3914E+04, 1.4067E+04, 1.4219E+04, 1.4373E+04,
     +  1.4527E+04, 1.4682E+04, 1.4837E+04, 1.4993E+04, 1.5150E+04,
     + 1.5308E+04, 1.5466E+04, 1.5625E+04, 1.5785E+04, 1.5945E+04,
     +  1.6106E+04, 1.6268E+04, 1.6431E+04, 1.6595E+04, 1.6759E+04,
     + 1.6925E+04, 1.7091E+04, 1.7258E+04, 1.7425E+04, 1.7594E+04,
     +  1.7764E+04, 1.7934E+04, 1.8105E+04, 1.8278E+04, 1.8451E+04,
     + 1.8625E+04, 1.8800E+04, 1.8976E+04, 1.9153E+04, 1.9330E+04,
     +  1.9509E+04, 1.9689E+04, 1.9870E+04, 2.0052E+04, 2.0235E+04,
     + 2.0418E+04, 2.0603E+04, 2.0789E+04, 2.0976E+04, 2.1164E+04,
     +  2.1353E+04, 2.1543E+04, 2.1735E+04, 2.1927E+04, 2.2120E+04,
     + 2.2315E+04, 2.2511E+04, 2.2708E+04, 2.2905E+04, 2.3105E+04,
     +  2.3305E+04, 2.3506E+04, 2.3709E+04, 2.3913E+04, 2.4118E+04,
     + 2.4324E+04, 2.4531E+04, 2.4740E+04, 2.4950E+04, 2.5161E+04,
     +  2.5373E+04, 2.5587E+04, 2.5801E+04, 2.6018E+04, 2.6235E+04,
     + 2.6454E+04, 2.6674E+04, 2.6895E+04, 2.7117E+04, 2.7341E+04,
     +  2.7567E+04, 2.7793E+04, 2.8021E+04, 2.8250E+04, 2.8481E+04,
     + 2.8713E+04, 2.8946E+04, 2.9181E+04, 2.9417E+04, 2.9655E+04,
     +  2.9894E+04, 3.0134E+04, 3.0376E+04, 3.0619E+04, 3.0864E+04,
     + 3.1110E+04, 3.1358E+04, 3.1607E+04, 3.1858E+04, 3.2110E+04,
     +  3.2363E+04, 3.2619E+04, 3.2875E+04, 3.3133E+04, 3.3393E+04,
     + 3.3654E+04, 3.3917E+04, 3.4181E+04, 3.4447E+04, 3.4715E+04,
     +  3.4984E+04, 3.5255E+04, 3.5527E+04, 3.5801E+04, 3.6076E+04,
     + 3.6353E+04, 3.6632E+04, 3.6913E+04, 3.7195E+04, 3.7478E+04,
     +  3.7764E+04, 3.8051E+04, 3.8340E+04, 3.8630E+04, 3.8922E+04,
     + 3.9216E+04, 3.9512E+04, 3.9809E+04, 4.0109E+04, 4.0409E+04,
     +  4.0712E+04, 4.1016E+04, 4.1323E+04, 4.1631E+04, 4.1940E+04,
     + 4.2252E+04, 4.2565E+04, 4.2881E+04, 4.3198E+04, 4.3517E+04,
     +  4.3837E+04, 4.4160E+04, 4.4484E+04, 4.4811E+04, 4.5139E+04,
     + 4.5469E+04, 4.5801E+04, 4.6135E+04, 4.6471E+04, 4.6809E+04,
     +  4.7149E+04, 4.7491E+04, 4.7834E+04, 4.8180E+04, 4.8528E+04,
     + 4.8877E+04, 4.9229E+04, 4.9583E+04, 4.9938E+04, 5.0296E+04,
     +  5.0656E+04, 5.1018E+04, 5.1382E+04, 5.1748E+04, 5.2116E+04,
     + 5.2486E+04, 5.2858E+04, 5.3232E+04, 5.3609E+04, 5.3987E+04,
     +  5.4368E+04, 5.4751E+04, 5.5136E+04, 5.5523E+04, 5.5913E+04,
     + 5.6304E+04, 5.6698E+04, 5.7094E+04, 5.7492E+04, 5.7893E+04,
     +  5.8295E+04, 5.8700E+04, 5.9108E+04, 5.9517E+04, 5.9929E+04,
     + 6.0343E+04, 6.0759E+04, 6.1178E+04, 6.1599E+04, 6.2022E+04,
     +  6.2448E+04, 6.2876E+04, 6.3306E+04, 6.3739E+04, 6.4174E+04,
     + 6.4612E+04, 6.5052E+04, 6.5494E+04, 6.5939E+04, 6.6386E+04,
     +  6.6836E+04, 6.7288E+04, 6.7743E+04, 6.8200E+04, 6.8660E+04,
     + 6.9122E+04, 6.9586E+04, 7.0054E+04, 7.0523E+04, 7.0996E+04,
     +  7.1470E+04, 7.1948E+04, 7.2428E+04, 7.2910E+04, 7.3396E+04,
     + 7.3883E+04, 7.4374E+04, 7.4867E+04, 7.5363E+04, 7.5861E+04,
     +  7.6362E+04, 7.6866E+04, 7.7373E+04, 7.7882E+04, 7.8394E+04,
     + 7.8908E+04, 7.9426E+04, 7.9946E+04, 8.0469E+04, 8.0994E+04,
     +  8.1523E+04, 8.2054E+04, 8.2588E+04, 8.3125E+04, 8.3665E+04,
     + 8.4208E+04, 8.4753E+04, 8.5302E+04, 8.5853E+04, 8.6407E+04,
     +  8.6964E+04, 8.7524E+04, 8.8087E+04, 8.8653E+04, 8.9222E+04,
     + 8.9794E+04, 9.0368E+04, 9.0946E+04, 9.1527E+04, 9.2111E+04,
     +  9.2698E+04, 9.3288E+04, 9.3881E+04, 9.4477E+04, 9.5076E+04,
     + 9.5678E+04, 9.6284E+04, 9.6892E+04, 9.7504E+04, 9.8119E+04,
     +  9.8737E+04, 9.9358E+04, 9.9982E+04, 1.0061E+05, 1.0124E+05,
     + 1.0187E+05, 1.0251E+05, 1.0315E+05, 1.0380E+05, 1.0444E+05,
     +  1.0509E+05, 1.0575E+05, 1.0640E+05, 1.0706E+05, 1.0773E+05,
     + 1.0839E+05, 1.0907E+05, 1.0974E+05, 1.1042E+05, 1.1110E+05,
     +  1.1178E+05, 1.1247E+05, 1.1316E+05, 1.1385E+05, 1.1455E+05,
     + 1.1525E+05, 1.1596E+05, 1.1667E+05, 1.1738E+05, 1.1809E+05,
     +  1.1881E+05, 1.1954E+05, 1.2026E+05, 1.2099E+05, 1.2172E+05,
     + 1.2246E+05, 1.2320E+05, 1.2395E+05, 1.2470E+05, 1.2545E+05,
     +  1.2620E+05, 1.2696E+05, 1.2773E+05, 1.2849E+05, 1.2926E+05,
     + 1.3004E+05, 1.3081E+05, 1.3160E+05, 1.3238E+05, 1.3317E+05,
     +  1.3397E+05, 1.3476E+05, 1.3556E+05, 1.3637E+05, 1.3718E+05,
     + 1.3799E+05, 1.3881E+05, 1.3963E+05, 1.4045E+05, 1.4128E+05,
     +  1.4212E+05, 1.4295E+05, 1.4379E+05, 1.4464E+05, 1.4549E+05,
     + 1.4634E+05, 1.4720E+05, 1.4806E+05, 1.4892E+05, 1.4979E+05,
     +  1.5067E+05, 1.5155E+05, 1.5243E+05, 1.5331E+05, 1.5420E+05,
     + 1.5510E+05, 1.5600E+05, 1.5690E+05, 1.5781E+05, 1.5872E+05,
     +  1.5964E+05, 1.6056E+05, 1.6148E+05, 1.6241E+05, 1.6335E+05,
     + 1.6428E+05, 1.6523E+05, 1.6617E+05, 1.6712E+05, 1.6808E+05,
     +  1.6904E+05, 1.7001E+05, 1.7097E+05, 1.7195E+05, 1.7293E+05,
     + 1.7391E+05, 1.7490E+05, 1.7589E+05, 1.7689E+05, 1.7789E+05,
     +  1.7889E+05, 1.7990E+05, 1.8092E+05, 1.8194E+05, 1.8296E+05,
     + 1.8399E+05, 1.8503E+05, 1.8607E+05, 1.8711E+05, 1.8816E+05,
     +  1.8921E+05, 1.9027E+05, 1.9133E+05, 1.9240E+05, 1.9348E+05,
     + 1.9455E+05, 1.9564E+05, 1.9672E+05, 1.9782E+05, 1.9891E+05,
     +  2.0002E+05, 2.0112E+05, 2.0224E+05, 2.0335E+05, 2.0448E+05,
     + 2.0560E+05, 2.0674E+05, 2.0788E+05, 2.0902E+05, 2.1017E+05,
     +  2.1132E+05, 2.1248E+05, 2.1364E+05, 2.1481E+05, 2.1599E+05,
     + 2.1717E+05, 2.1835E+05, 2.1954E+05, 2.2074E+05, 2.2194E+05,
     +  2.2314E+05, 2.2435E+05, 2.2557E+05, 2.2679E+05, 2.2802E+05,
     + 2.2925E+05, 2.3049E+05, 2.3174E+05, 2.3299E+05, 2.3424E+05,
     +  2.3550E+05, 2.3677E+05, 2.3804E+05, 2.3932E+05, 2.4060E+05,
     + 2.4189E+05, 2.4319E+05, 2.4449E+05, 2.4579E+05, 2.4710E+05,
     +  2.4842E+05, 2.4974E+05, 2.5107E+05, 2.5241E+05, 2.5375E+05,
     + 2.5509E+05, 2.5644E+05, 2.5780E+05, 2.5917E+05, 2.6054E+05,
     +  2.6191E+05, 2.6329E+05, 2.6468E+05, 2.6607E+05, 2.6747E+05,
     + 2.6888E+05, 2.7029E+05, 2.7171E+05, 2.7313E+05, 2.7456E+05,
     +  2.7600E+05, 2.7744E+05, 2.7889E+05, 2.8034E+05, 2.8180E+05,
     + 2.8327E+05, 2.8474E+05, 2.8622E+05, 2.8771E+05, 2.8920E+05,
     +  2.9070E+05, 2.9221E+05, 2.9372E+05, 2.9523E+05, 2.9676E+05,
     + 2.9829E+05, 2.9983E+05, 3.0137E+05, 3.0292E+05, 3.0448E+05,
     +  3.0604E+05, 3.0761E+05, 3.0918E+05, 3.1077E+05, 3.1236E+05,
     + 3.1395E+05, 3.1556E+05, 3.1717E+05, 3.1878E+05, 3.2040E+05,
     +  3.2203E+05, 3.2367E+05, 3.2531E+05, 3.2696E+05, 3.2862E+05,
     + 3.3028E+05, 3.3196E+05, 3.3363E+05, 3.3532E+05, 3.3701E+05,
     +  3.3871E+05, 3.4041E+05, 3.4213E+05, 3.4385E+05, 3.4557E+05,
     + 3.4731E+05, 3.4905E+05, 3.5080E+05, 3.5255E+05, 3.5431E+05,
     +  3.5608E+05, 3.5786E+05, 3.5964E+05, 3.6143E+05, 3.6323E+05,
     + 3.6504E+05, 3.6685E+05, 3.6867E+05, 3.7050E+05, 3.7234E+05,
     +  3.7418E+05, 3.7603E+05, 3.7789E+05, 3.7975E+05, 3.8163E+05,
     + 3.8351E+05, 3.8540E+05, 3.8729E+05, 3.8920E+05, 3.9111E+05,
     +  3.9302E+05, 3.9495E+05, 3.9689E+05, 3.9883E+05, 4.0078E+05,
     + 4.0273E+05, 4.0470E+05, 4.0667E+05, 4.0865E+05, 4.1064E+05,
     +  4.1264E+05, 4.1464E+05, 4.1666E+05, 4.1868E+05, 4.2071E+05,
     + 4.2274E+05, 4.2479E+05, 4.2684E+05, 4.2890E+05, 4.3097E+05,
     +  4.3305E+05, 4.3513E+05, 4.3723E+05, 4.3933E+05, 4.4144E+05,
     + 4.4356E+05, 4.4568E+05, 4.4782E+05, 4.4996E+05, 4.5211E+05,
     +  4.5427E+05, 4.5644E+05, 4.5862E+05, 4.6080E+05, 4.6300E+05,
     + 4.6520E+05, 4.6741E+05, 4.6963E+05, 4.7186E+05, 4.7410E+05,
     +  4.7634E+05, 4.7860E+05, 4.8086E+05, 4.8313E+05, 4.8541E+05,
     + 4.8770E+05, 4.9000E+05, 4.9231E+05, 4.9462E+05, 4.9695E+05,
     +  4.9928E+05, 5.0162E+05, 5.0398E+05, 5.0634E+05, 5.0871E+05,
     + 5.1108E+05, 5.1347E+05, 5.1587E+05, 5.1828E+05, 5.2069E+05,
     +  5.2312E+05, 5.2555E+05, 5.2799E+05, 5.3044E+05, 5.3291E+05,
     + 5.3538E+05, 5.3786E+05, 5.4035E+05, 5.4284E+05, 5.4535E+05,
     +  5.4787E+05, 5.5040E+05, 5.5293E+05, 5.5548E+05, 5.5804E+05,
     + 5.6060E+05, 5.6318E+05, 5.6576E+05, 5.6836E+05, 5.7096E+05,
     +  5.7357E+05, 5.7620E+05, 5.7883E+05, 5.8147E+05, 5.8413E+05,
     + 5.8679E+05, 5.8946E+05, 5.9215E+05, 5.9484E+05, 5.9754E+05,
     +  6.0025E+05, 6.0298E+05, 6.0571E+05, 6.0845E+05, 6.1121E+05,
     + 6.1397E+05, 6.1674E+05, 6.1953E+05, 6.2232E+05, 6.2513E+05,
     +  6.2794E+05, 6.3076E+05, 6.3360E+05, 6.3645E+05, 6.3930E+05,
     + 6.4217E+05, 6.4504E+05, 6.4793E+05, 6.5083E+05, 6.5374E+05,
     +  6.5666E+05, 6.5959E+05, 6.6253E+05, 6.6548E+05, 6.6844E+05,
     + 6.7141E+05, 6.7440E+05, 6.7739E+05, 6.8040E+05, 6.8341E+05,
     +  6.8644E+05, 6.8948E+05, 6.9252E+05, 6.9558E+05, 6.9865E+05,
     + 7.0174E+05, 7.0483E+05, 7.0793E+05, 7.1105E+05, 7.1417E+05,
     +  7.1731E+05, 7.2046E+05, 7.2362E+05, 7.2679E+05, 7.2997E+05,
     + 7.3317E+05, 7.3637E+05, 7.3959E+05, 7.4282E+05, 7.4606E+05,
     +  7.4931E+05, 7.5257E+05, 7.5584E+05, 7.5913E+05, 7.6243E+05,
     + 7.6573E+05, 7.6906E+05, 7.7239E+05, 7.7573E+05, 7.7909E+05,
     +  7.8246E+05, 7.8583E+05, 7.8923E+05, 7.9263E+05, 7.9604E+05,
     + 7.9947E+05, 8.0291E+05, 8.0636E+05, 8.0983E+05, 8.1330E+05,
     +  8.1679E+05, 8.2029E+05, 8.2380E+05, 8.2733E+05, 8.3086E+05,
     + 8.3441E+05, 8.3797E+05, 8.4155E+05, 8.4513E+05, 8.4873E+05,
     +  8.5234E+05, 8.5596E+05, 8.5960E+05, 8.6325E+05, 8.6691E+05,
     + 8.7058E+05, 8.7427E+05, 8.7797E+05, 8.8168E+05, 8.8540E+05,
     +  8.8914E+05, 8.9289E+05, 8.9665E+05, 9.0043E+05, 9.0422E+05,
     + 9.0802E+05, 9.1183E+05, 9.1566E+05, 9.1950E+05, 9.2336E+05,
     +  9.2722E+05, 9.3110E+05, 9.3500E+05, 9.3890E+05, 9.4282E+05,
     + 9.4676E+05, 9.5070E+05, 9.5466E+05, 9.5864E+05, 9.6262E+05,
     +  9.6662E+05, 9.7064E+05, 9.7466E+05, 9.7870E+05, 9.8276E+05,
     + 9.8683E+05, 9.9091E+05, 9.9500E+05, 9.9911E+05, 1.0032E+06,
     +  1.0074E+06, 1.0115E+06, 1.0157E+06, 1.0199E+06, 1.0241E+06,
     + 1.0283E+06, 1.0325E+06, 1.0367E+06, 1.0410E+06, 1.0452E+06,
     +  1.0495E+06, 1.0538E+06, 1.0581E+06, 1.0624E+06, 1.0668E+06,
     + 1.0711E+06, 1.0755E+06, 1.0798E+06, 1.0842E+06, 1.0886E+06,
     +  1.0931E+06, 1.0975E+06, 1.1019E+06, 1.1064E+06, 1.1109E+06,
     + 1.1154E+06, 1.1199E+06, 1.1244E+06, 1.1290E+06, 1.1335E+06,
     +  1.1381E+06, 1.1427E+06, 1.1473E+06, 1.1519E+06, 1.1565E+06,
     + 1.1611E+06, 1.1658E+06, 1.1705E+06, 1.1752E+06, 1.1799E+06,
     +  1.1846E+06, 1.1893E+06, 1.1941E+06, 1.1988E+06, 1.2036E+06,
     + 1.2084E+06, 1.2132E+06, 1.2180E+06, 1.2229E+06, 1.2277E+06,
     +  1.2326E+06, 1.2375E+06, 1.2424E+06, 1.2473E+06, 1.2523E+06,
     + 1.2572E+06, 1.2622E+06, 1.2672E+06, 1.2722E+06, 1.2772E+06,
     +  1.2822E+06, 1.2873E+06, 1.2923E+06, 1.2974E+06, 1.3025E+06,
     + 1.3076E+06, 1.3128E+06, 1.3179E+06, 1.3231E+06, 1.3282E+06,
     +  1.3334E+06, 1.3387E+06, 1.3439E+06, 1.3491E+06, 1.3544E+06,
     + 1.3597E+06, 1.3650E+06, 1.3703E+06, 1.3756E+06, 1.3809E+06,
     +  1.3863E+06, 1.3917E+06, 1.3971E+06, 1.4025E+06, 1.4079E+06,
     + 1.4134E+06, 1.4188E+06, 1.4243E+06, 1.4298E+06, 1.4353E+06,
     +  1.4408E+06, 1.4464E+06, 1.4520E+06, 1.4575E+06, 1.4631E+06,
     + 1.4688E+06, 1.4744E+06, 1.4801E+06, 1.4857E+06, 1.4914E+06,
     +  1.4971E+06, 1.5028E+06, 1.5086E+06, 1.5143E+06, 1.5201E+06,
     + 1.5259E+06, 1.5317E+06, 1.5376E+06, 1.5434E+06, 1.5493E+06,
     +  1.5552E+06, 1.5611E+06, 1.5670E+06, 1.5729E+06, 1.5789E+06,
     + 1.5849E+06, 1.5908E+06, 1.5969E+06, 1.6029E+06, 1.6089E+06,
     +  1.6150E+06, 1.6211E+06, 1.6272E+06, 1.6333E+06, 1.6395E+06,
     + 1.6456E+06, 1.6518E+06, 1.6580E+06, 1.6642E+06, 1.6704E+06,
     +  1.6767E+06, 1.6830E+06, 1.6893E+06, 1.6956E+06, 1.7019E+06,
     + 1.7083E+06, 1.7146E+06, 1.7210E+06, 1.7274E+06, 1.7338E+06,
     +  1.7403E+06, 1.7468E+06, 1.7532E+06, 1.7597E+06, 1.7663E+06,
     + 1.7728E+06, 1.7794E+06, 1.7860E+06, 1.7926E+06, 1.7992E+06,
     +  1.8058E+06, 1.8125E+06, 1.8192E+06, 1.8259E+06, 1.8326E+06,
     + 1.8393E+06, 1.8461E+06, 1.8529E+06, 1.8597E+06, 1.8665E+06,
     +  1.8733E+06, 1.8802E+06, 1.8871E+06, 1.8940E+06, 1.9009E+06,
     + 1.9078E+06, 1.9148E+06, 1.9218E+06, 1.9288E+06, 1.9358E+06,
     +  1.9429E+06, 1.9499E+06, 1.9570E+06, 1.9641E+06, 1.9712E+06,
     + 1.9784E+06, 1.9856E+06, 1.9928E+06, 2.0000E+06, 2.0072E+06,
     +  2.0145E+06, 2.0217E+06, 2.0290E+06, 2.0363E+06, 2.0437E+06,
     + 2.0510E+06, 2.0584E+06, 2.0658E+06, 2.0733E+06, 2.0807E+06,
     +  2.0882E+06, 2.0957E+06, 2.1032E+06, 2.1107E+06, 2.1183E+06,
     + 2.1258E+06, 2.1334E+06, 2.1411E+06, 2.1487E+06, 2.1564E+06,
     +  2.1641E+06, 2.1718E+06, 2.1795E+06, 2.1872E+06, 2.1950E+06,
     + 2.2028E+06, 2.2106E+06, 2.2185E+06, 2.2263E+06, 2.2342E+06,
     +  2.2421E+06, 2.2501E+06, 2.2580E+06, 2.2660E+06, 2.2740E+06,
     + 2.2820E+06, 2.2901E+06, 2.2981E+06, 2.3062E+06, 2.3143E+06,
     +  2.3225E+06, 2.3306E+06, 2.3388E+06, 2.3470E+06, 2.3553E+06,
     + 2.3635E+06, 2.3718E+06, 2.3801E+06, 2.3884E+06, 2.3968E+06,
     +  2.4051E+06, 2.4135E+06, 2.4219E+06, 2.4304E+06, 2.4389E+06,
     + 2.4473E+06, 2.4559E+06, 2.4644E+06, 2.4729E+06, 2.4815E+06,
     +  2.4901E+06, 2.4988E+06, 2.5074E+06, 2.5161E+06, 2.5248E+06,
     + 2.5335E+06, 2.5423E+06, 2.5511E+06, 2.5599E+06, 2.5687E+06,
     +  2.5775E+06, 2.5864E+06, 2.5953E+06, 2.6042E+06, 2.6132E+06,
     + 2.6222E+06, 2.6312E+06, 2.6402E+06, 2.6492E+06, 2.6583E+06,
     +  2.6674E+06, 2.6765E+06, 2.6857E+06, 2.6949E+06, 2.7041E+06,
     + 2.7133E+06, 2.7225E+06, 2.7318E+06, 2.7411E+06, 2.7504E+06,
     +  2.7598E+06, 2.7692E+06, 2.7786E+06, 2.7880E+06, 2.7974E+06,
     + 2.8069E+06, 2.8164E+06, 2.8260E+06, 2.8355E+06, 2.8451E+06,
     +  2.8547E+06, 2.8643E+06, 2.8740E+06, 2.8837E+06, 2.8934E+06,
     + 2.9031E+06, 2.9129E+06, 2.9227E+06, 2.9325E+06, 2.9424E+06,
     +  2.9522E+06, 2.9621E+06, 2.9721E+06, 2.9820E+06, 2.9920E+06,
     + 3.0020E+06, 3.0120E+06, 3.0221E+06, 3.0322E+06, 3.0423E+06,
     +  3.0524E+06, 3.0626E+06, 3.0728E+06, 3.0830E+06, 3.0933E+06,
     + 3.1036E+06, 3.1139E+06, 3.1242E+06, 3.1345E+06, 3.1449E+06,
     +  3.1553E+06, 3.1658E+06, 3.1763E+06, 3.1868E+06, 3.1973E+06,
     + 3.2078E+06, 3.2184E+06, 3.2290E+06, 3.2397E+06, 3.2503E+06,
     +  3.2610E+06, 3.2718E+06, 3.2825E+06, 3.2933E+06, 3.3041E+06,
     + 3.3149E+06, 3.3258E+06, 3.3367E+06, 3.3476E+06, 3.3586E+06,
     +  3.3695E+06, 3.3805E+06, 3.3916E+06, 3.4026E+06, 3.4137E+06,
     + 3.4249E+06, 3.4360E+06, 3.4472E+06, 3.4584E+06, 3.4697E+06,
     +  3.4809E+06, 3.4922E+06, 3.5036E+06, 3.5149E+06, 3.5263E+06,
     + 3.5377E+06, 3.5492E+06, 3.5606E+06, 3.5721E+06, 3.5837E+06,
     +  3.5952E+06, 3.6068E+06, 3.6185E+06, 3.6301E+06, 3.6418E+06,
     + 3.6535E+06, 3.6653E+06, 3.6770E+06, 3.6889E+06, 3.7007E+06,
     +  3.7126E+06, 3.7245E+06, 3.7364E+06, 3.7483E+06, 3.7603E+06,
     + 3.7724E+06, 3.7844E+06, 3.7965E+06, 3.8086E+06, 3.8208E+06,
     +  3.8329E+06, 3.8451E+06, 3.8574E+06, 3.8696E+06, 3.8819E+06,
     + 3.8943E+06, 3.9066E+06, 3.9190E+06, 3.9315E+06, 3.9439E+06,
     +  3.9564E+06, 3.9689E+06, 3.9815E+06, 3.9941E+06, 4.0067E+06,
     + 4.0193E+06, 4.0320E+06, 4.0447E+06, 4.0575E+06, 4.0702E+06,
     +  4.0830E+06, 4.0959E+06, 4.1088E+06, 4.1217E+06, 4.1346E+06,
     + 4.1476E+06, 4.1606E+06, 4.1736E+06, 4.1867E+06, 4.1998E+06,
     +  4.2129E+06, 4.2261E+06, 4.2393E+06, 4.2525E+06, 4.2658E+06,
     + 4.2791E+06, 4.2924E+06, 4.3058E+06, 4.3192E+06, 4.3326E+06,
     +  4.3461E+06, 4.3596E+06, 4.3731E+06, 4.3867E+06, 4.4003E+06,
     + 4.4139E+06, 4.4276E+06, 4.4413E+06, 4.4550E+06, 4.4688E+06,
     +  4.4826E+06, 4.4964E+06, 4.5103E+06, 4.5242E+06, 4.5381E+06,
     + 4.5521E+06, 4.5661E+06, 4.5802E+06, 4.5942E+06, 4.6083E+06,
     +  4.6225E+06, 4.6367E+06, 4.6509E+06, 4.6651E+06, 4.6794E+06,
     + 4.6937E+06, 4.7081E+06, 4.7225E+06, 4.7369E+06, 4.7514E+06,
     +  4.7659E+06, 4.7804E+06, 4.7950E+06, 4.8096E+06, 4.8242E+06,
     + 4.8389E+06, 4.8536E+06, 4.8683E+06, 4.8831E+06, 4.8979E+06,
     +  4.9128E+06, 4.9277E+06, 4.9426E+06, 4.9576E+06, 4.9726E+06,
     + 4.9876E+06, 5.0027E+06, 5.0178E+06, 5.0329E+06, 5.0481E+06,
     +  5.0633E+06, 5.0786E+06, 5.0939E+06, 5.1092E+06, 5.1245E+06,
     + 5.1399E+06, 5.1554E+06, 5.1709E+06, 5.1864E+06, 5.2019E+06,
     +  5.2175E+06, 5.2331E+06, 5.2488E+06, 5.2645E+06, 5.2802E+06,
     + 5.2960E+06, 5.3118E+06, 5.3276E+06, 5.3435E+06, 5.3594E+06,
     +  5.3754E+06, 5.3914E+06, 5.4074E+06, 5.4235E+06, 5.4396E+06,
     + 5.4558E+06, 5.4720E+06, 5.4882E+06, 5.5045E+06, 5.5208E+06,
     +  5.5371E+06, 5.5535E+06, 5.5699E+06, 5.5864E+06, 5.6029E+06,
     + 5.6194E+06, 5.6360E+06, 5.6526E+06, 5.6692E+06, 5.6859E+06,
     +  5.7027E+06, 5.7194E+06, 5.7362E+06, 5.7531E+06, 5.7700E+06,
     + 5.7869E+06, 5.8039E+06, 5.8209E+06, 5.8379E+06, 5.8550E+06,
     +  5.8721E+06, 5.8893E+06, 5.9065E+06, 5.9238E+06, 5.9410E+06,
     + 5.9584E+06, 5.9757E+06, 5.9931E+06, 6.0106E+06, 6.0281E+06,
     +  6.0456E+06, 6.0632E+06, 6.0808E+06, 6.0984E+06, 6.1161E+06,
     + 6.1339E+06, 6.1516E+06, 6.1694E+06, 6.1873E+06, 6.2052E+06,
     +  6.2231E+06, 6.2411E+06, 6.2591E+06, 6.2772E+06, 6.2953E+06,
     + 6.3134E+06, 6.3316E+06, 6.3499E+06, 6.3681E+06, 6.3864E+06,
     +  6.4048E+06, 6.4232E+06, 6.4416E+06, 6.4601E+06, 6.4786E+06,
     + 6.4972E+06, 6.5158E+06, 6.5344E+06, 6.5531E+06, 6.5719E+06,
     +  6.5906E+06, 6.6095E+06, 6.6283E+06, 6.6472E+06, 6.6662E+06,
     + 6.6852E+06, 6.7042E+06, 6.7233E+06, 6.7424E+06, 6.7616E+06,
     +  6.7808E+06, 6.8000E+06, 6.8193E+06, 6.8386E+06, 6.8580E+06,
     + 6.8774E+06, 6.8969E+06, 6.9164E+06, 6.9360E+06, 6.9556E+06,
     +  6.9752E+06, 6.9949E+06, 7.0146E+06, 7.0344E+06, 7.0542E+06,
     + 7.0741E+06, 7.0940E+06, 7.1140E+06, 7.1340E+06, 7.1540E+06,
     +  7.1741E+06, 7.1942E+06, 7.2144E+06, 7.2346E+06, 7.2549E+06,
     + 7.2752E+06, 7.2956E+06, 7.3160E+06, 7.3364E+06, 7.3569E+06,
     +  7.3774E+06, 7.3980E+06, 7.4186E+06, 7.4393E+06, 7.4600E+06,
     + 7.4808E+06, 7.5016E+06, 7.5225E+06, 7.5434E+06, 7.5643E+06,
     +  7.5853E+06, 7.6064E+06, 7.6275E+06, 7.6486E+06, 7.6698E+06,
     + 7.6910E+06, 7.7123E+06, 7.7336E+06, 7.7550E+06, 7.7764E+06,
     +  7.7979E+06, 7.8194E+06, 7.8409E+06, 7.8625E+06, 7.8842E+06,
     + 7.9059E+06, 7.9276E+06, 7.9494E+06, 7.9713E+06, 7.9931E+06,
     +  8.0151E+06, 8.0371E+06, 8.0591E+06, 8.0812E+06, 8.1033E+06,
     + 8.1255E+06, 8.1477E+06, 8.1700E+06, 8.1923E+06, 8.2147E+06,
     +  8.2371E+06, 8.2595E+06, 8.2820E+06, 8.3046E+06, 8.3272E+06,
     + 8.3499E+06, 8.3726E+06, 8.3953E+06, 8.4182E+06, 8.4410E+06,
     +  8.4639E+06, 8.4869E+06, 8.5099E+06, 8.5329E+06, 8.5560E+06,
     + 8.5792E+06, 8.6024E+06, 8.6256E+06, 8.6489E+06, 8.6723E+06,
     +  8.6957E+06, 8.7191E+06, 8.7426E+06, 8.7662E+06, 8.7898E+06,
     + 8.8134E+06, 8.8371E+06, 8.8609E+06, 8.8847E+06, 8.9085E+06,
     +  8.9324E+06, 8.9564E+06, 8.9804E+06, 9.0044E+06, 9.0286E+06,
     + 9.0527E+06, 9.0769E+06, 9.1012E+06, 9.1255E+06, 9.1499E+06,
     +  9.1743E+06, 9.1987E+06, 9.2232E+06, 9.2478E+06, 9.2724E+06,
     + 9.2971E+06, 9.3218E+06, 9.3466E+06, 9.3714E+06, 9.3963E+06,
     +  9.4212E+06, 9.4462E+06, 9.4713E+06, 9.4964E+06, 9.5215E+06,
     + 9.5467E+06, 9.5719E+06, 9.5972E+06, 9.6226E+06, 9.6480E+06,
     +  9.6735E+06, 9.6990E+06, 9.7245E+06, 9.7502E+06, 9.7758E+06,
     + 9.8016E+06, 9.8273E+06, 9.8532E+06, 9.8790E+06, 9.9050E+06,
     +  9.9310E+06, 9.9570E+06, 9.9831E+06, 1.0009E+07, 1.0035E+07,
     + 1.0062E+07, 1.0088E+07, 1.0114E+07, 1.0141E+07, 1.0167E+07,
     +  1.0194E+07, 1.0220E+07, 1.0247E+07, 1.0274E+07, 1.0301E+07,
     + 1.0327E+07, 1.0354E+07, 1.0381E+07, 1.0408E+07, 1.0435E+07,
     +  1.0462E+07, 1.0489E+07, 1.0517E+07, 1.0544E+07, 1.0571E+07,
     + 1.0599E+07, 1.0626E+07, 1.0653E+07, 1.0681E+07, 1.0709E+07,
     +  1.0736E+07, 1.0764E+07, 1.0792E+07, 1.0819E+07, 1.0847E+07,
     + 1.0875E+07, 1.0903E+07, 1.0931E+07, 1.0959E+07, 1.0987E+07,
     +  1.1016E+07, 1.1044E+07, 1.1072E+07, 1.1101E+07, 1.1129E+07,
     + 1.1158E+07, 1.1186E+07, 1.1215E+07, 1.1243E+07, 1.1272E+07,
     +  1.1301E+07, 1.1330E+07, 1.1359E+07, 1.1388E+07, 1.1417E+07,
     + 1.1446E+07, 1.1475E+07, 1.1504E+07, 1.1533E+07, 1.1563E+07,
     +  1.1592E+07, 1.1621E+07, 1.1651E+07, 1.1680E+07, 1.1710E+07,
     + 1.1740E+07, 1.1769E+07, 1.1799E+07, 1.1829E+07, 1.1859E+07,
     +  1.1889E+07, 1.1919E+07, 1.1949E+07, 1.1979E+07, 1.2009E+07,
     + 1.2040E+07, 1.2070E+07, 1.2100E+07, 1.2131E+07, 1.2161E+07,
     +  1.2192E+07, 1.2222E+07, 1.2253E+07, 1.2284E+07, 1.2314E+07,
     + 1.2345E+07, 1.2376E+07, 1.2407E+07, 1.2438E+07, 1.2469E+07,
     +  1.2501E+07, 1.2532E+07, 1.2563E+07, 1.2594E+07, 1.2626E+07,
     + 1.2657E+07, 1.2689E+07, 1.2720E+07, 1.2752E+07, 1.2784E+07,
     +  1.2816E+07, 1.2847E+07, 1.2879E+07, 1.2911E+07, 1.2943E+07,
     + 1.2975E+07, 1.3007E+07, 1.3040E+07, 1.3072E+07, 1.3104E+07,
     +  1.3137E+07, 1.3169E+07, 1.3202E+07, 1.3234E+07, 1.3267E+07,
     + 1.3300E+07, 1.3332E+07, 1.3365E+07, 1.3398E+07, 1.3431E+07,
     +  1.3464E+07, 1.3497E+07, 1.3530E+07, 1.3564E+07, 1.3597E+07,
     + 1.3630E+07, 1.3664E+07, 1.3697E+07, 1.3731E+07, 1.3764E+07,
     +  1.3798E+07, 1.3832E+07, 1.3866E+07, 1.3899E+07, 1.3933E+07,
     + 1.3967E+07, 1.4001E+07, 1.4035E+07, 1.4070E+07, 1.4104E+07,
     +  1.4138E+07, 1.4173E+07, 1.4207E+07, 1.4242E+07, 1.4276E+07,
     + 1.4311E+07, 1.4346E+07, 1.4380E+07, 1.4415E+07, 1.4450E+07,
     +  1.4485E+07, 1.4520E+07, 1.4555E+07, 1.4590E+07, 1.4626E+07,
     + 1.4661E+07, 1.4696E+07, 1.4732E+07, 1.4767E+07, 1.4803E+07,
     +  1.4838E+07, 1.4874E+07, 1.4910E+07, 1.4946E+07, 1.4982E+07,
     + 1.5018E+07, 1.5054E+07, 1.5090E+07, 1.5126E+07, 1.5162E+07,
     +  1.5199E+07, 1.5235E+07, 1.5271E+07, 1.5308E+07, 1.5345E+07,
     + 1.5381E+07, 1.5418E+07, 1.5455E+07, 1.5492E+07, 1.5528E+07,
     +  1.5565E+07, 1.5603E+07, 1.5640E+07, 1.5677E+07, 1.5714E+07,
     + 1.5752E+07, 1.5789E+07, 1.5826E+07, 1.5864E+07, 1.5902E+07,
     +  1.5939E+07, 1.5977E+07, 1.6015E+07, 1.6053E+07, 1.6091E+07,
     + 1.6129E+07, 1.6167E+07, 1.6205E+07, 1.6243E+07, 1.6282E+07,
     +  1.6320E+07, 1.6359E+07, 1.6397E+07, 1.6436E+07, 1.6474E+07,
     + 1.6513E+07, 1.6552E+07, 1.6591E+07, 1.6630E+07, 1.6669E+07,
     +  1.6708E+07, 1.6747E+07, 1.6786E+07, 1.6826E+07, 1.6865E+07,
     + 1.6904E+07, 1.6944E+07, 1.6984E+07, 1.7023E+07, 1.7063E+07,
     +  1.7103E+07, 1.7143E+07, 1.7183E+07, 1.7223E+07, 1.7263E+07,
     + 1.7303E+07, 1.7343E+07, 1.7384E+07, 1.7424E+07, 1.7464E+07,
     +  1.7505E+07, 1.7546E+07, 1.7586E+07, 1.7627E+07, 1.7668E+07,
     + 1.7709E+07, 1.7750E+07, 1.7791E+07, 1.7832E+07, 1.7873E+07,
     +  1.7914E+07, 1.7956E+07, 1.7997E+07, 1.8039E+07, 1.8080E+07,
     + 1.8122E+07, 1.8164E+07, 1.8205E+07, 1.8247E+07, 1.8289E+07,
     +  1.8331E+07, 1.8373E+07, 1.8416E+07, 1.8458E+07, 1.8500E+07,
     + 1.8543E+07, 1.8585E+07, 1.8628E+07, 1.8670E+07, 1.8713E+07,
     +  1.8756E+07, 1.8799E+07, 1.8841E+07, 1.8884E+07, 1.8928E+07,
     + 1.8971E+07, 1.9014E+07, 1.9057E+07, 1.9101E+07, 1.9144E+07,
     +  1.9188E+07, 1.9231E+07, 1.9275E+07, 1.9319E+07, 1.9363E+07,
     + 1.9406E+07, 1.9450E+07, 1.9495E+07, 1.9539E+07, 1.9583E+07,
     +  1.9627E+07, 1.9672E+07, 1.9716E+07, 1.9761E+07, 1.9805E+07,
     + 1.9850E+07, 1.9895E+07, 1.9940E+07, 1.9985E+07, 2.0030E+07,
     +  2.0075E+07, 2.0120E+07, 2.0165E+07, 2.0210E+07, 2.0256E+07,
     + 2.0301E+07, 2.0347E+07, 2.0392E+07, 2.0438E+07, 2.0484E+07,
     +  2.0530E+07, 2.0576E+07, 2.0622E+07, 2.0668E+07, 2.0714E+07,
     + 2.0760E+07, 2.0807E+07, 2.0853E+07, 2.0900E+07, 2.0946E+07,
     +  2.0993E+07, 2.1040E+07, 2.1087E+07, 2.1134E+07, 2.1181E+07,
     + 2.1228E+07, 2.1275E+07, 2.1322E+07, 2.1369E+07, 2.1417E+07,
     +  2.1464E+07, 2.1512E+07, 2.1559E+07, 2.1607E+07, 2.1655E+07,
     + 2.1703E+07, 2.1751E+07, 2.1799E+07, 2.1847E+07, 2.1895E+07,
     +  2.1943E+07, 2.1992E+07, 2.2040E+07, 2.2089E+07, 2.2137E+07,
     + 2.2186E+07, 2.2235E+07, 2.2284E+07, 2.2333E+07, 2.2382E+07,
     +  2.2431E+07, 2.2480E+07, 2.2529E+07, 2.2579E+07, 2.2628E+07,
     + 2.2678E+07, 2.2727E+07, 2.2777E+07, 2.2827E+07, 2.2877E+07,
     +  2.2927E+07, 2.2977E+07, 2.3027E+07, 2.3077E+07, 2.3127E+07,
     + 2.3178E+07, 2.3228E+07, 2.3279E+07, 2.3329E+07, 2.3380E+07,
     +  2.3431E+07, 2.3482E+07, 2.3533E+07, 2.3584E+07, 2.3635E+07,
     + 2.3686E+07, 2.3738E+07, 2.3789E+07, 2.3840E+07, 2.3892E+07,
     +  2.3944E+07, 2.3995E+07, 2.4047E+07, 2.4099E+07, 2.4151E+07,
     + 2.4203E+07, 2.4255E+07, 2.4308E+07, 2.4360E+07, 2.4412E+07,
     +  2.4465E+07, 2.4517E+07, 2.4570E+07, 2.4623E+07, 2.4676E+07,
     + 2.4729E+07, 2.4782E+07, 2.4835E+07, 2.4888E+07, 2.4941E+07,
     +  2.4995E+07, 2.5048E+07, 2.5102E+07, 2.5155E+07, 2.5209E+07,
     + 2.5263E+07, 2.5317E+07, 2.5371E+07, 2.5425E+07, 2.5479E+07,
     +  2.5533E+07, 2.5588E+07, 2.5642E+07, 2.5697E+07, 2.5751E+07,
     + 2.5806E+07, 2.5861E+07, 2.5916E+07, 2.5971E+07, 2.6026E+07,
     +  2.6081E+07, 2.6136E+07, 2.6191E+07, 2.6247E+07, 2.6302E+07,
     + 2.6358E+07, 2.6413E+07, 2.6469E+07, 2.6525E+07, 2.6581E+07,
     +  2.6637E+07, 2.6693E+07, 2.6749E+07, 2.6806E+07, 2.6862E+07,
     + 2.6919E+07, 2.6975E+07, 2.7032E+07, 2.7089E+07, 2.7146E+07,
     +  2.7202E+07, 2.7259E+07, 2.7317E+07, 2.7374E+07, 2.7431E+07,
     + 2.7489E+07, 2.7546E+07, 2.7604E+07, 2.7661E+07, 2.7719E+07,
     +  2.7777E+07, 2.7835E+07, 2.7893E+07, 2.7951E+07, 2.8009E+07,
     + 2.8068E+07, 2.8126E+07, 2.8185E+07, 2.8243E+07, 2.8302E+07,
     +  2.8361E+07, 2.8420E+07, 2.8478E+07, 2.8538E+07, 2.8597E+07,
     + 2.8656E+07, 2.8715E+07, 2.8775E+07, 2.8834E+07, 2.8894E+07,
     +  2.8954E+07, 2.9013E+07, 2.9073E+07, 2.9133E+07, 2.9193E+07,
     + 2.9254E+07, 2.9314E+07, 2.9374E+07, 2.9435E+07, 2.9495E+07,
     +  2.9556E+07, 2.9617E+07, 2.9678E+07, 2.9738E+07, 2.9800E+07,
     + 2.9861E+07, 2.9922E+07, 2.9983E+07, 3.0045E+07, 3.0106E+07,
     +  3.0168E+07, 3.0229E+07, 3.0291E+07, 3.0353E+07, 3.0415E+07,
     + 3.0477E+07, 3.0539E+07, 3.0602E+07, 3.0664E+07, 3.0727E+07,
     +  3.0789E+07, 3.0852E+07, 3.0915E+07, 3.0977E+07, 3.1040E+07,
     + 3.1103E+07, 3.1167E+07, 3.1230E+07, 3.1293E+07, 3.1357E+07,
     +  3.1420E+07, 3.1484E+07, 3.1548E+07, 3.1611E+07, 3.1675E+07,
     + 3.1739E+07, 3.1803E+07, 3.1868E+07, 3.1932E+07, 3.1996E+07,
     +  3.2061E+07, 3.2126E+07, 3.2190E+07, 3.2255E+07, 3.2320E+07,
     + 3.2385E+07, 3.2450E+07, 3.2515E+07, 3.2581E+07, 3.2646E+07,
     +  3.2711E+07, 3.2777E+07, 3.2843E+07, 3.2909E+07, 3.2974E+07,
     + 3.3040E+07, 3.3107E+07, 3.3173E+07, 3.3239E+07, 3.3305E+07,
     +  3.3372E+07, 3.3439E+07, 3.3505E+07, 3.3572E+07, 3.3639E+07,
     + 3.3706E+07, 3.3773E+07, 3.3840E+07, 3.3908E+07, 3.3975E+07,
     +  3.4042E+07, 3.4110E+07, 3.4178E+07, 3.4246E+07, 3.4314E+07,
     + 3.4382E+07, 3.4450E+07, 3.4518E+07, 3.4586E+07, 3.4655E+07,
     +  3.4723E+07, 3.4792E+07, 3.4860E+07, 3.4929E+07, 3.4998E+07,
     + 3.5067E+07, 3.5136E+07, 3.5206E+07, 3.5275E+07, 3.5344E+07,
     +  3.5414E+07, 3.5484E+07, 3.5553E+07, 3.5623E+07, 3.5693E+07,
     + 3.5763E+07, 3.5833E+07, 3.5904E+07, 3.5974E+07, 3.6044E+07,
     +  3.6115E+07, 3.6186E+07, 3.6256E+07, 3.6327E+07, 3.6398E+07,
     + 3.6469E+07, 3.6541E+07, 3.6612E+07, 3.6683E+07, 3.6755E+07,
     +  3.6826E+07, 3.6898E+07, 3.6970E+07, 3.7042E+07, 3.7114E+07,
     + 3.7186E+07, 3.7258E+07, 3.7331E+07, 3.7403E+07, 3.7476E+07,
     +  3.7548E+07, 3.7621E+07, 3.7694E+07, 3.7767E+07, 3.7840E+07,
     + 3.7913E+07, 3.7986E+07, 3.8060E+07, 3.8133E+07, 3.8207E+07,
     +  3.8281E+07, 3.8354E+07, 3.8428E+07, 3.8502E+07, 3.8577E+07,
     + 3.8651E+07, 3.8725E+07, 3.8800E+07, 3.8874E+07, 3.8949E+07,
     +  3.9024E+07, 3.9099E+07, 3.9174E+07, 3.9249E+07, 3.9324E+07,
     + 3.9399E+07, 3.9475E+07, 3.9550E+07, 3.9626E+07, 3.9702E+07,
     +  3.9777E+07, 3.9853E+07, 3.9929E+07, 4.0006E+07, 4.0082E+07,
     + 4.0158E+07, 4.0235E+07, 4.0311E+07, 4.0388E+07, 4.0465E+07,
     +  4.0542E+07, 4.0619E+07, 4.0696E+07, 4.0773E+07, 4.0851E+07,
     + 4.0928E+07, 4.1006E+07, 4.1084E+07, 4.1161E+07, 4.1239E+07,
     +  4.1317E+07, 4.1396E+07, 4.1474E+07, 4.1552E+07, 4.1631E+07,
     + 4.1709E+07, 4.1788E+07, 4.1867E+07, 4.1946E+07, 4.2025E+07,
     +  4.2104E+07, 4.2183E+07, 4.2262E+07, 4.2342E+07, 4.2421E+07,
     + 4.2501E+07, 4.2581E+07, 4.2661E+07, 4.2741E+07, 4.2821E+07,
     +  4.2901E+07, 4.2982E+07, 4.3062E+07, 4.3143E+07, 4.3223E+07,
     + 4.3304E+07, 4.3385E+07, 4.3466E+07, 4.3547E+07, 4.3628E+07,
     +  4.3710E+07, 4.3791E+07, 4.3873E+07, 4.3955E+07, 4.4036E+07,
     + 4.4118E+07, 4.4200E+07, 4.4283E+07, 4.4365E+07, 4.4447E+07,
     +  4.4530E+07, 4.4612E+07, 4.4695E+07, 4.4778E+07, 4.4861E+07,
     + 4.4944E+07, 4.5027E+07, 4.5110E+07, 4.5194E+07, 4.5277E+07,
     +  4.5361E+07, 4.5445E+07, 4.5528E+07, 4.5612E+07, 4.5696E+07,
     + 4.5781E+07, 4.5865E+07, 4.5949E+07, 4.6034E+07, 4.6119E+07,
     +  4.6203E+07, 4.6288E+07, 4.6373E+07, 4.6458E+07, 4.6544E+07,
     + 4.6629E+07, 4.6714E+07, 4.6800E+07, 4.6886E+07, 4.6972E+07,
     +  4.7057E+07, 4.7144E+07, 4.7230E+07, 4.7316E+07, 4.7402E+07,
     + 4.7489E+07, 4.7575E+07, 4.7662E+07, 4.7749E+07, 4.7836E+07,
     +  4.7923E+07, 4.8010E+07, 4.8098E+07, 4.8185E+07, 4.8273E+07,
     + 4.8360E+07, 4.8448E+07, 4.8536E+07, 4.8624E+07, 4.8712E+07,
     +  4.8800E+07, 4.8889E+07, 4.8977E+07, 4.9066E+07, 4.9155E+07,
     + 4.9243E+07, 4.9332E+07, 4.9421E+07, 4.9511E+07, 4.9600E+07,
     +  4.9689E+07, 4.9779E+07, 4.9869E+07, 4.9958E+07, 5.0048E+07,
     + 5.0138E+07, 5.0228E+07, 5.0319E+07, 5.0409E+07, 5.0500E+07,
     +  5.0590E+07, 5.0681E+07, 5.0772E+07, 5.0863E+07, 5.0954E+07,
     + 5.1045E+07, 5.1136E+07, 5.1228E+07, 5.1319E+07, 5.1411E+07,
     +  5.1503E+07, 5.1595E+07, 5.1687E+07, 5.1779E+07, 5.1871E+07,
     + 5.1964E+07, 5.2056E+07, 5.2149E+07, 5.2242E+07, 5.2335E+07,
     +  5.2428E+07, 5.2521E+07, 5.2614E+07, 5.2707E+07, 5.2801E+07,
     + 5.2895E+07, 5.2988E+07, 5.3082E+07, 5.3176E+07, 5.3270E+07,
     +  5.3364E+07, 5.3459E+07, 5.3553E+07, 5.3648E+07, 5.3743E+07,
     + 5.3837E+07, 5.3932E+07, 5.4027E+07, 5.4123E+07, 5.4218E+07,
     +  5.4313E+07, 5.4409E+07, 5.4505E+07, 5.4600E+07, 5.4696E+07,
     + 5.4792E+07, 5.4889E+07, 5.4985E+07, 5.5081E+07, 5.5178E+07,
     +  5.5275E+07, 5.5371E+07, 5.5468E+07, 5.5565E+07, 5.5663E+07,
     + 5.5760E+07, 5.5857E+07, 5.5955E+07, 5.6052E+07, 5.6150E+07,
     +  5.6248E+07, 5.6346E+07, 5.6444E+07, 5.6543E+07, 5.6641E+07,
     + 5.6739E+07, 5.6838E+07, 5.6937E+07, 5.7036E+07, 5.7135E+07,
     +  5.7234E+07, 5.7333E+07, 5.7433E+07, 5.7532E+07, 5.7632E+07,
     + 5.7732E+07, 5.7831E+07, 5.7931E+07, 5.8032E+07, 5.8132E+07,
     +  5.8232E+07, 5.8333E+07, 5.8433E+07, 5.8534E+07, 5.8635E+07,
     + 5.8736E+07, 5.8837E+07, 5.8939E+07, 5.9040E+07, 5.9142E+07,
     +  5.9243E+07, 5.9345E+07, 5.9447E+07, 5.9549E+07, 5.9651E+07,
     + 5.9753E+07, 5.9856E+07, 5.9958E+07, 6.0061E+07, 6.0164E+07,
     +  6.0267E+07, 6.0370E+07, 6.0473E+07, 6.0576E+07, 6.0680E+07,
     + 6.0783E+07, 6.0887E+07, 6.0991E+07, 6.1095E+07, 6.1199E+07,
     +  6.1303E+07, 6.1407E+07, 6.1512E+07, 6.1616E+07, 6.1721E+07,
     + 6.1826E+07, 6.1931E+07, 6.2036E+07, 6.2141E+07, 6.2246E+07,
     +  6.2352E+07, 6.2458E+07, 6.2563E+07, 6.2669E+07, 6.2775E+07,
     + 6.2881E+07, 6.2987E+07, 6.3094E+07, 6.3200E+07, 6.3307E+07,
     +  6.3414E+07, 6.3521E+07, 6.3628E+07, 6.3735E+07, 6.3842E+07/
      data (QofT(12,J),J=2409,4510)/ 6.3950E+07, 6.4057E+07,
     +  6.4165E+07, 6.4273E+07, 6.4380E+07, 6.4489E+07, 6.4597E+07,
     +  6.4705E+07,
     + 6.4813E+07, 6.4922E+07, 6.5031E+07, 6.5140E+07, 6.5249E+07,
     +  6.5358E+07, 6.5467E+07, 6.5576E+07, 6.5686E+07, 6.5795E+07,
     + 6.5905E+07, 6.6015E+07, 6.6125E+07, 6.6235E+07, 6.6346E+07,
     +  6.6456E+07, 6.6566E+07, 6.6677E+07, 6.6788E+07, 6.6899E+07,
     + 6.7010E+07, 6.7121E+07, 6.7232E+07, 6.7344E+07, 6.7456E+07,
     +  6.7567E+07, 6.7679E+07, 6.7791E+07, 6.7903E+07, 6.8016E+07,
     + 6.8128E+07, 6.8240E+07, 6.8353E+07, 6.8466E+07, 6.8579E+07,
     +  6.8692E+07, 6.8805E+07, 6.8918E+07, 6.9032E+07, 6.9145E+07,
     + 6.9259E+07, 6.9373E+07, 6.9487E+07, 6.9601E+07, 6.9715E+07,
     +  6.9830E+07, 6.9944E+07, 7.0059E+07, 7.0174E+07, 7.0289E+07,
     + 7.0404E+07, 7.0519E+07, 7.0634E+07, 7.0750E+07, 7.0865E+07,
     +  7.0981E+07, 7.1097E+07, 7.1213E+07, 7.1329E+07, 7.1445E+07,
     + 7.1562E+07, 7.1678E+07, 7.1795E+07, 7.1912E+07, 7.2029E+07,
     +  7.2146E+07, 7.2263E+07, 7.2380E+07, 7.2498E+07, 7.2615E+07,
     + 7.2733E+07, 7.2851E+07, 7.2969E+07, 7.3087E+07, 7.3205E+07,
     +  7.3324E+07, 7.3442E+07, 7.3561E+07, 7.3680E+07, 7.3799E+07,
     + 7.3918E+07, 7.4037E+07, 7.4157E+07, 7.4276E+07, 7.4396E+07,
     +  7.4516E+07, 7.4635E+07, 7.4755E+07, 7.4876E+07, 7.4996E+07,
     + 7.5116E+07, 7.5237E+07, 7.5358E+07, 7.5479E+07, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       646
      data (QofT(13,J),J=1,2408)/ 1.1723E+00, 1.9606E+00, 2.8447E+00,
     +  3.7334E+00, 4.6230E+00, 5.5128E+00, 6.4029E+00, 7.2930E+00,
     + 8.1833E+00, 9.0736E+00, 9.9639E+00, 1.0854E+01, 1.1745E+01,
     +  1.2635E+01, 1.3526E+01, 1.4416E+01, 1.5306E+01, 1.6197E+01,
     + 1.7087E+01, 1.7978E+01, 1.9759E+01, 2.1540E+01, 2.3321E+01,
     +  2.5102E+01, 2.6883E+01, 2.8664E+01, 3.0445E+01, 3.2226E+01,
     + 3.4007E+01, 3.5788E+01, 3.7570E+01, 3.9351E+01, 4.1132E+01,
     +  4.2913E+01, 4.4694E+01, 4.6475E+01, 4.8257E+01, 5.0038E+01,
     + 5.1819E+01, 5.3600E+01, 5.5382E+01, 5.7163E+01, 5.8944E+01,
     +  6.0725E+01, 6.2507E+01, 6.4288E+01, 6.6070E+01, 6.7851E+01,
     + 6.9633E+01, 7.1415E+01, 7.3197E+01, 7.4979E+01, 7.6761E+01,
     +  7.8543E+01, 8.0326E+01, 8.2109E+01, 8.3893E+01, 8.5677E+01,
     + 8.7461E+01, 8.9247E+01, 9.1032E+01, 9.2819E+01, 9.4606E+01,
     +  9.6395E+01, 9.8184E+01, 9.9974E+01, 1.0177E+02, 1.0356E+02,
     + 1.0535E+02, 1.0715E+02, 1.0895E+02, 1.1075E+02, 1.1255E+02,
     +  1.1435E+02, 1.1616E+02, 1.1797E+02, 1.1978E+02, 1.2160E+02,
     + 1.2341E+02, 1.2523E+02, 1.2706E+02, 1.2888E+02, 1.3071E+02,
     +  1.3255E+02, 1.3439E+02, 1.3623E+02, 1.3808E+02, 1.3993E+02,
     + 1.4178E+02, 1.4365E+02, 1.4551E+02, 1.4738E+02, 1.4926E+02,
     +  1.5114E+02, 1.5303E+02, 1.5492E+02, 1.5682E+02, 1.5873E+02,
     + 1.6064E+02, 1.6256E+02, 1.6449E+02, 1.6642E+02, 1.6836E+02,
     +  1.7030E+02, 1.7226E+02, 1.7422E+02, 1.7619E+02, 1.7817E+02,
     + 1.8015E+02, 1.8215E+02, 1.8415E+02, 1.8616E+02, 1.8818E+02,
     +  1.9021E+02, 1.9225E+02, 1.9430E+02, 1.9635E+02, 1.9842E+02,
     + 2.0050E+02, 2.0258E+02, 2.0468E+02, 2.0678E+02, 2.0890E+02,
     +  2.1102E+02, 2.1316E+02, 2.1531E+02, 2.1747E+02, 2.1964E+02,
     + 2.2182E+02, 2.2401E+02, 2.2621E+02, 2.2843E+02, 2.3065E+02,
     +  2.3289E+02, 2.3514E+02, 2.3740E+02, 2.3968E+02, 2.4197E+02,
     + 2.4426E+02, 2.4658E+02, 2.4890E+02, 2.5124E+02, 2.5359E+02,
     +  2.5595E+02, 2.5833E+02, 2.6072E+02, 2.6312E+02, 2.6554E+02,
     + 2.6797E+02, 2.7041E+02, 2.7287E+02, 2.7534E+02, 2.7783E+02,
     +  2.8033E+02, 2.8285E+02, 2.8538E+02, 2.8792E+02, 2.9048E+02,
     + 2.9305E+02, 2.9564E+02, 2.9825E+02, 3.0087E+02, 3.0350E+02,
     +  3.0615E+02, 3.0882E+02, 3.1150E+02, 3.1420E+02, 3.1691E+02,
     + 3.1964E+02, 3.2238E+02, 3.2514E+02, 3.2792E+02, 3.3072E+02,
     +  3.3353E+02, 3.3635E+02, 3.3920E+02, 3.4206E+02, 3.4494E+02,
     + 3.4783E+02, 3.5075E+02, 3.5368E+02, 3.5662E+02, 3.5959E+02,
     +  3.6257E+02, 3.6557E+02, 3.6859E+02, 3.7163E+02, 3.7468E+02,
     + 3.7776E+02, 3.8085E+02, 3.8396E+02, 3.8709E+02, 3.9023E+02,
     +  3.9340E+02, 3.9658E+02, 3.9979E+02, 4.0301E+02, 4.0625E+02,
     + 4.0951E+02, 4.1279E+02, 4.1609E+02, 4.1941E+02, 4.2275E+02,
     +  4.2611E+02, 4.2949E+02, 4.3289E+02, 4.3631E+02, 4.3975E+02,
     + 4.4322E+02, 4.4670E+02, 4.5020E+02, 4.5372E+02, 4.5727E+02,
     +  4.6083E+02, 4.6442E+02, 4.6803E+02, 4.7165E+02, 4.7531E+02,
     + 4.7898E+02, 4.8267E+02, 4.8639E+02, 4.9012E+02, 4.9388E+02,
     +  4.9767E+02, 5.0147E+02, 5.0530E+02, 5.0915E+02, 5.1302E+02,
     + 5.1691E+02, 5.2083E+02, 5.2477E+02, 5.2873E+02, 5.3272E+02,
     +  5.3673E+02, 5.4076E+02, 5.4482E+02, 5.4890E+02, 5.5301E+02,
     + 5.5713E+02, 5.6129E+02, 5.6546E+02, 5.6966E+02, 5.7389E+02,
     +  5.7814E+02, 5.8242E+02, 5.8671E+02, 5.9104E+02, 5.9539E+02,
     + 5.9976E+02, 6.0416E+02, 6.0859E+02, 6.1304E+02, 6.1752E+02,
     +  6.2202E+02, 6.2655E+02, 6.3110E+02, 6.3568E+02, 6.4029E+02,
     + 6.4492E+02, 6.4958E+02, 6.5427E+02, 6.5898E+02, 6.6372E+02,
     +  6.6849E+02, 6.7328E+02, 6.7811E+02, 6.8295E+02, 6.8783E+02,
     + 6.9274E+02, 6.9767E+02, 7.0263E+02, 7.0762E+02, 7.1263E+02,
     +  7.1768E+02, 7.2275E+02, 7.2785E+02, 7.3298E+02, 7.3814E+02,
     + 7.4333E+02, 7.4855E+02, 7.5380E+02, 7.5907E+02, 7.6438E+02,
     +  7.6972E+02, 7.7508E+02, 7.8048E+02, 7.8590E+02, 7.9136E+02,
     + 7.9685E+02, 8.0236E+02, 8.0791E+02, 8.1349E+02, 8.1910E+02,
     +  8.2474E+02, 8.3041E+02, 8.3611E+02, 8.4185E+02, 8.4761E+02,
     + 8.5341E+02, 8.5924E+02, 8.6510E+02, 8.7100E+02, 8.7692E+02,
     +  8.8288E+02, 8.8887E+02, 8.9490E+02, 9.0095E+02, 9.0705E+02,
     + 9.1317E+02, 9.1933E+02, 9.2552E+02, 9.3174E+02, 9.3800E+02,
     +  9.4429E+02, 9.5062E+02, 9.5698E+02, 9.6337E+02, 9.6980E+02,
     + 9.7627E+02, 9.8276E+02, 9.8930E+02, 9.9587E+02, 1.0025E+03,
     +  1.0091E+03, 1.0158E+03, 1.0225E+03, 1.0292E+03, 1.0360E+03,
     + 1.0428E+03, 1.0497E+03, 1.0566E+03, 1.0635E+03, 1.0705E+03,
     +  1.0775E+03, 1.0845E+03, 1.0916E+03, 1.0987E+03, 1.1059E+03,
     + 1.1131E+03, 1.1203E+03, 1.1276E+03, 1.1349E+03, 1.1422E+03,
     +  1.1496E+03, 1.1570E+03, 1.1645E+03, 1.1720E+03, 1.1795E+03,
     + 1.1871E+03, 1.1947E+03, 1.2024E+03, 1.2101E+03, 1.2178E+03,
     +  1.2256E+03, 1.2334E+03, 1.2413E+03, 1.2492E+03, 1.2571E+03,
     + 1.2651E+03, 1.2731E+03, 1.2812E+03, 1.2893E+03, 1.2974E+03,
     +  1.3056E+03, 1.3138E+03, 1.3221E+03, 1.3304E+03, 1.3388E+03,
     + 1.3472E+03, 1.3556E+03, 1.3641E+03, 1.3727E+03, 1.3812E+03,
     +  1.3899E+03, 1.3985E+03, 1.4072E+03, 1.4160E+03, 1.4248E+03,
     + 1.4336E+03, 1.4425E+03, 1.4514E+03, 1.4604E+03, 1.4694E+03,
     +  1.4785E+03, 1.4876E+03, 1.4968E+03, 1.5060E+03, 1.5152E+03,
     + 1.5245E+03, 1.5338E+03, 1.5432E+03, 1.5527E+03, 1.5622E+03,
     +  1.5717E+03, 1.5813E+03, 1.5909E+03, 1.6006E+03, 1.6103E+03,
     + 1.6201E+03, 1.6299E+03, 1.6397E+03, 1.6497E+03, 1.6596E+03,
     +  1.6696E+03, 1.6797E+03, 1.6898E+03, 1.7000E+03, 1.7102E+03,
     + 1.7205E+03, 1.7308E+03, 1.7411E+03, 1.7516E+03, 1.7620E+03,
     +  1.7725E+03, 1.7831E+03, 1.7937E+03, 1.8044E+03, 1.8151E+03,
     + 1.8259E+03, 1.8367E+03, 1.8476E+03, 1.8585E+03, 1.8695E+03,
     +  1.8806E+03, 1.8917E+03, 1.9028E+03, 1.9140E+03, 1.9253E+03,
     + 1.9366E+03, 1.9479E+03, 1.9594E+03, 1.9708E+03, 1.9824E+03,
     +  1.9939E+03, 2.0056E+03, 2.0173E+03, 2.0290E+03, 2.0408E+03,
     + 2.0527E+03, 2.0646E+03, 2.0766E+03, 2.0886E+03, 2.1007E+03,
     +  2.1129E+03, 2.1251E+03, 2.1373E+03, 2.1497E+03, 2.1620E+03,
     + 2.1745E+03, 2.1870E+03, 2.1995E+03, 2.2121E+03, 2.2248E+03,
     +  2.2375E+03, 2.2503E+03, 2.2632E+03, 2.2761E+03, 2.2891E+03,
     + 2.3021E+03, 2.3152E+03, 2.3283E+03, 2.3416E+03, 2.3548E+03,
     +  2.3682E+03, 2.3816E+03, 2.3950E+03, 2.4086E+03, 2.4222E+03,
     + 2.4358E+03, 2.4495E+03, 2.4633E+03, 2.4771E+03, 2.4910E+03,
     +  2.5050E+03, 2.5190E+03, 2.5331E+03, 2.5473E+03, 2.5615E+03,
     + 2.5758E+03, 2.5902E+03, 2.6046E+03, 2.6191E+03, 2.6336E+03,
     +  2.6483E+03, 2.6630E+03, 2.6777E+03, 2.6925E+03, 2.7074E+03,
     + 2.7224E+03, 2.7374E+03, 2.7525E+03, 2.7677E+03, 2.7829E+03,
     +  2.7982E+03, 2.8135E+03, 2.8290E+03, 2.8445E+03, 2.8601E+03,
     + 2.8757E+03, 2.8914E+03, 2.9072E+03, 2.9231E+03, 2.9390E+03,
     +  2.9550E+03, 2.9710E+03, 2.9872E+03, 3.0034E+03, 3.0197E+03,
     + 3.0360E+03, 3.0525E+03, 3.0690E+03, 3.0855E+03, 3.1022E+03,
     +  3.1189E+03, 3.1357E+03, 3.1526E+03, 3.1695E+03, 3.1865E+03,
     + 3.2036E+03, 3.2208E+03, 3.2381E+03, 3.2554E+03, 3.2728E+03,
     +  3.2902E+03, 3.3078E+03, 3.3254E+03, 3.3431E+03, 3.3609E+03,
     + 3.3788E+03, 3.3967E+03, 3.4147E+03, 3.4328E+03, 3.4510E+03,
     +  3.4692E+03, 3.4875E+03, 3.5059E+03, 3.5244E+03, 3.5430E+03,
     + 3.5616E+03, 3.5804E+03, 3.5992E+03, 3.6181E+03, 3.6370E+03,
     +  3.6561E+03, 3.6752E+03, 3.6944E+03, 3.7137E+03, 3.7331E+03,
     + 3.7526E+03, 3.7721E+03, 3.7917E+03, 3.8115E+03, 3.8313E+03,
     +  3.8511E+03, 3.8711E+03, 3.8912E+03, 3.9113E+03, 3.9315E+03,
     + 3.9518E+03, 3.9722E+03, 3.9927E+03, 4.0132E+03, 4.0339E+03,
     +  4.0546E+03, 4.0755E+03, 4.0964E+03, 4.1174E+03, 4.1385E+03,
     + 4.1596E+03, 4.1809E+03, 4.2023E+03, 4.2237E+03, 4.2452E+03,
     +  4.2669E+03, 4.2886E+03, 4.3104E+03, 4.3323E+03, 4.3543E+03,
     + 4.3764E+03, 4.3985E+03, 4.4208E+03, 4.4431E+03, 4.4656E+03,
     +  4.4881E+03, 4.5108E+03, 4.5335E+03, 4.5563E+03, 4.5792E+03,
     + 4.6022E+03, 4.6253E+03, 4.6485E+03, 4.6718E+03, 4.6952E+03,
     +  4.7187E+03, 4.7423E+03, 4.7660E+03, 4.7897E+03, 4.8136E+03,
     + 4.8376E+03, 4.8616E+03, 4.8858E+03, 4.9101E+03, 4.9344E+03,
     +  4.9589E+03, 4.9834E+03, 5.0081E+03, 5.0329E+03, 5.0577E+03,
     + 5.0827E+03, 5.1077E+03, 5.1329E+03, 5.1582E+03, 5.1835E+03,
     +  5.2090E+03, 5.2345E+03, 5.2602E+03, 5.2860E+03, 5.3119E+03,
     + 5.3379E+03, 5.3639E+03, 5.3901E+03, 5.4164E+03, 5.4428E+03,
     +  5.4693E+03, 5.4959E+03, 5.5226E+03, 5.5495E+03, 5.5764E+03,
     + 5.6034E+03, 5.6306E+03, 5.6578E+03, 5.6852E+03, 5.7126E+03,
     +  5.7402E+03, 5.7679E+03, 5.7957E+03, 5.8236E+03, 5.8516E+03,
     + 5.8797E+03, 5.9080E+03, 5.9363E+03, 5.9648E+03, 5.9933E+03,
     +  6.0220E+03, 6.0508E+03, 6.0797E+03, 6.1087E+03, 6.1378E+03,
     + 6.1671E+03, 6.1964E+03, 6.2259E+03, 6.2555E+03, 6.2852E+03,
     +  6.3150E+03, 6.3449E+03, 6.3750E+03, 6.4051E+03, 6.4354E+03,
     + 6.4658E+03, 6.4963E+03, 6.5269E+03, 6.5577E+03, 6.5886E+03,
     +  6.6195E+03, 6.6506E+03, 6.6819E+03, 6.7132E+03, 6.7447E+03,
     + 6.7763E+03, 6.8080E+03, 6.8398E+03, 6.8717E+03, 6.9038E+03,
     +  6.9360E+03, 6.9683E+03, 7.0008E+03, 7.0333E+03, 7.0660E+03,
     + 7.0988E+03, 7.1317E+03, 7.1648E+03, 7.1980E+03, 7.2313E+03,
     +  7.2647E+03, 7.2983E+03, 7.3320E+03, 7.3658E+03, 7.3997E+03,
     + 7.4338E+03, 7.4680E+03, 7.5023E+03, 7.5368E+03, 7.5714E+03,
     +  7.6061E+03, 7.6409E+03, 7.6759E+03, 7.7110E+03, 7.7462E+03,
     + 7.7816E+03, 7.8171E+03, 7.8527E+03, 7.8885E+03, 7.9244E+03,
     +  7.9604E+03, 7.9966E+03, 8.0329E+03, 8.0693E+03, 8.1059E+03,
     + 8.1426E+03, 8.1794E+03, 8.2164E+03, 8.2535E+03, 8.2907E+03,
     +  8.3281E+03, 8.3656E+03, 8.4033E+03, 8.4411E+03, 8.4790E+03,
     + 8.5171E+03, 8.5553E+03, 8.5937E+03, 8.6322E+03, 8.6708E+03,
     +  8.7096E+03, 8.7485E+03, 8.7875E+03, 8.8267E+03, 8.8661E+03,
     + 8.9056E+03, 8.9452E+03, 8.9850E+03, 9.0249E+03, 9.0650E+03,
     +  9.1052E+03, 9.1455E+03, 9.1860E+03, 9.2267E+03, 9.2675E+03,
     + 9.3084E+03, 9.3495E+03, 9.3907E+03, 9.4321E+03, 9.4736E+03,
     +  9.5153E+03, 9.5571E+03, 9.5991E+03, 9.6413E+03, 9.6835E+03,
     + 9.7260E+03, 9.7686E+03, 9.8113E+03, 9.8542E+03, 9.8972E+03,
     +  9.9404E+03, 9.9838E+03, 1.0027E+04, 1.0071E+04, 1.0115E+04,
     + 1.0159E+04, 1.0203E+04, 1.0247E+04, 1.0292E+04, 1.0336E+04,
     +  1.0381E+04, 1.0426E+04, 1.0471E+04, 1.0516E+04, 1.0561E+04,
     + 1.0607E+04, 1.0653E+04, 1.0699E+04, 1.0745E+04, 1.0791E+04,
     +  1.0837E+04, 1.0884E+04, 1.0930E+04, 1.0977E+04, 1.1024E+04,
     + 1.1071E+04, 1.1119E+04, 1.1166E+04, 1.1214E+04, 1.1262E+04,
     +  1.1310E+04, 1.1358E+04, 1.1406E+04, 1.1455E+04, 1.1503E+04,
     + 1.1552E+04, 1.1601E+04, 1.1650E+04, 1.1699E+04, 1.1749E+04,
     +  1.1799E+04, 1.1848E+04, 1.1898E+04, 1.1949E+04, 1.1999E+04,
     + 1.2050E+04, 1.2100E+04, 1.2151E+04, 1.2202E+04, 1.2253E+04,
     +  1.2305E+04, 1.2356E+04, 1.2408E+04, 1.2460E+04, 1.2512E+04,
     + 1.2564E+04, 1.2617E+04, 1.2670E+04, 1.2722E+04, 1.2775E+04,
     +  1.2829E+04, 1.2882E+04, 1.2935E+04, 1.2989E+04, 1.3043E+04,
     + 1.3097E+04, 1.3151E+04, 1.3206E+04, 1.3260E+04, 1.3315E+04,
     +  1.3370E+04, 1.3425E+04, 1.3481E+04, 1.3536E+04, 1.3592E+04,
     + 1.3648E+04, 1.3704E+04, 1.3760E+04, 1.3817E+04, 1.3874E+04,
     +  1.3930E+04, 1.3988E+04, 1.4045E+04, 1.4102E+04, 1.4160E+04,
     + 1.4218E+04, 1.4276E+04, 1.4334E+04, 1.4392E+04, 1.4451E+04,
     +  1.4510E+04, 1.4569E+04, 1.4628E+04, 1.4687E+04, 1.4747E+04,
     + 1.4806E+04, 1.4866E+04, 1.4927E+04, 1.4987E+04, 1.5047E+04,
     +  1.5108E+04, 1.5169E+04, 1.5230E+04, 1.5292E+04, 1.5353E+04,
     + 1.5415E+04, 1.5477E+04, 1.5539E+04, 1.5601E+04, 1.5664E+04,
     +  1.5727E+04, 1.5790E+04, 1.5853E+04, 1.5916E+04, 1.5980E+04,
     + 1.6044E+04, 1.6108E+04, 1.6172E+04, 1.6236E+04, 1.6301E+04,
     +  1.6366E+04, 1.6431E+04, 1.6496E+04, 1.6561E+04, 1.6627E+04,
     + 1.6693E+04, 1.6759E+04, 1.6825E+04, 1.6892E+04, 1.6959E+04,
     +  1.7025E+04, 1.7093E+04, 1.7160E+04, 1.7228E+04, 1.7295E+04,
     + 1.7363E+04, 1.7432E+04, 1.7500E+04, 1.7569E+04, 1.7638E+04,
     +  1.7707E+04, 1.7776E+04, 1.7846E+04, 1.7915E+04, 1.7985E+04,
     + 1.8056E+04, 1.8126E+04, 1.8197E+04, 1.8268E+04, 1.8339E+04,
     +  1.8410E+04, 1.8482E+04, 1.8553E+04, 1.8625E+04, 1.8698E+04,
     + 1.8770E+04, 1.8843E+04, 1.8916E+04, 1.8989E+04, 1.9062E+04,
     +  1.9136E+04, 1.9210E+04, 1.9284E+04, 1.9358E+04, 1.9432E+04,
     + 1.9507E+04, 1.9582E+04, 1.9657E+04, 1.9733E+04, 1.9809E+04,
     +  1.9885E+04, 1.9961E+04, 2.0037E+04, 2.0114E+04, 2.0191E+04,
     + 2.0268E+04, 2.0345E+04, 2.0423E+04, 2.0501E+04, 2.0579E+04,
     +  2.0657E+04, 2.0736E+04, 2.0814E+04, 2.0893E+04, 2.0973E+04,
     + 2.1052E+04, 2.1132E+04, 2.1212E+04, 2.1292E+04, 2.1373E+04,
     +  2.1453E+04, 2.1534E+04, 2.1616E+04, 2.1697E+04, 2.1779E+04,
     + 2.1861E+04, 2.1943E+04, 2.2026E+04, 2.2108E+04, 2.2191E+04,
     +  2.2275E+04, 2.2358E+04, 2.2442E+04, 2.2526E+04, 2.2610E+04,
     + 2.2695E+04, 2.2780E+04, 2.2865E+04, 2.2950E+04, 2.3035E+04,
     +  2.3121E+04, 2.3207E+04, 2.3294E+04, 2.3380E+04, 2.3467E+04,
     + 2.3554E+04, 2.3642E+04, 2.3729E+04, 2.3817E+04, 2.3905E+04,
     +  2.3994E+04, 2.4082E+04, 2.4171E+04, 2.4261E+04, 2.4350E+04,
     + 2.4440E+04, 2.4530E+04, 2.4620E+04, 2.4711E+04, 2.4801E+04,
     +  2.4893E+04, 2.4984E+04, 2.5076E+04, 2.5167E+04, 2.5260E+04,
     + 2.5352E+04, 2.5445E+04, 2.5538E+04, 2.5631E+04, 2.5725E+04,
     +  2.5819E+04, 2.5913E+04, 2.6007E+04, 2.6102E+04, 2.6197E+04,
     + 2.6292E+04, 2.6387E+04, 2.6483E+04, 2.6579E+04, 2.6676E+04,
     +  2.6772E+04, 2.6869E+04, 2.6966E+04, 2.7064E+04, 2.7161E+04,
     + 2.7260E+04, 2.7358E+04, 2.7456E+04, 2.7555E+04, 2.7655E+04,
     +  2.7754E+04, 2.7854E+04, 2.7954E+04, 2.8054E+04, 2.8155E+04,
     + 2.8256E+04, 2.8357E+04, 2.8459E+04, 2.8560E+04, 2.8662E+04,
     +  2.8765E+04, 2.8868E+04, 2.8971E+04, 2.9074E+04, 2.9177E+04,
     + 2.9281E+04, 2.9386E+04, 2.9490E+04, 2.9595E+04, 2.9700E+04,
     +  2.9805E+04, 2.9911E+04, 3.0017E+04, 3.0123E+04, 3.0230E+04,
     + 3.0337E+04, 3.0444E+04, 3.0551E+04, 3.0659E+04, 3.0767E+04,
     +  3.0876E+04, 3.0985E+04, 3.1094E+04, 3.1203E+04, 3.1313E+04,
     + 3.1423E+04, 3.1533E+04, 3.1644E+04, 3.1754E+04, 3.1866E+04,
     +  3.1977E+04, 3.2089E+04, 3.2201E+04, 3.2314E+04, 3.2427E+04,
     + 3.2540E+04, 3.2653E+04, 3.2767E+04, 3.2881E+04, 3.2995E+04,
     +  3.3110E+04, 3.3225E+04, 3.3341E+04, 3.3456E+04, 3.3572E+04,
     + 3.3689E+04, 3.3805E+04, 3.3922E+04, 3.4040E+04, 3.4157E+04,
     +  3.4275E+04, 3.4394E+04, 3.4512E+04, 3.4631E+04, 3.4750E+04,
     + 3.4870E+04, 3.4990E+04, 3.5110E+04, 3.5231E+04, 3.5352E+04,
     +  3.5473E+04, 3.5595E+04, 3.5717E+04, 3.5839E+04, 3.5962E+04,
     + 3.6085E+04, 3.6208E+04, 3.6332E+04, 3.6456E+04, 3.6580E+04,
     +  3.6705E+04, 3.6830E+04, 3.6955E+04, 3.7081E+04, 3.7207E+04,
     + 3.7333E+04, 3.7460E+04, 3.7587E+04, 3.7715E+04, 3.7842E+04,
     +  3.7971E+04, 3.8099E+04, 3.8228E+04, 3.8357E+04, 3.8487E+04,
     + 3.8617E+04, 3.8747E+04, 3.8877E+04, 3.9008E+04, 3.9140E+04,
     +  3.9271E+04, 3.9403E+04, 3.9536E+04, 3.9668E+04, 3.9802E+04,
     + 3.9935E+04, 4.0069E+04, 4.0203E+04, 4.0338E+04, 4.0472E+04,
     +  4.0608E+04, 4.0743E+04, 4.0879E+04, 4.1016E+04, 4.1152E+04,
     + 4.1290E+04, 4.1427E+04, 4.1565E+04, 4.1703E+04, 4.1842E+04,
     +  4.1981E+04, 4.2120E+04, 4.2259E+04, 4.2400E+04, 4.2540E+04,
     + 4.2681E+04, 4.2822E+04, 4.2964E+04, 4.3105E+04, 4.3248E+04,
     +  4.3390E+04, 4.3533E+04, 4.3677E+04, 4.3821E+04, 4.3965E+04,
     + 4.4110E+04, 4.4255E+04, 4.4400E+04, 4.4546E+04, 4.4692E+04,
     +  4.4838E+04, 4.4985E+04, 4.5132E+04, 4.5280E+04, 4.5428E+04,
     + 4.5577E+04, 4.5725E+04, 4.5875E+04, 4.6024E+04, 4.6174E+04,
     +  4.6325E+04, 4.6475E+04, 4.6627E+04, 4.6778E+04, 4.6930E+04,
     + 4.7083E+04, 4.7235E+04, 4.7389E+04, 4.7542E+04, 4.7696E+04,
     +  4.7850E+04, 4.8005E+04, 4.8160E+04, 4.8316E+04, 4.8472E+04,
     + 4.8628E+04, 4.8785E+04, 4.8942E+04, 4.9100E+04, 4.9258E+04,
     +  4.9416E+04, 4.9575E+04, 4.9734E+04, 4.9894E+04, 5.0054E+04,
     + 5.0215E+04, 5.0376E+04, 5.0537E+04, 5.0699E+04, 5.0861E+04,
     +  5.1023E+04, 5.1186E+04, 5.1350E+04, 5.1513E+04, 5.1678E+04,
     + 5.1842E+04, 5.2007E+04, 5.2173E+04, 5.2339E+04, 5.2505E+04,
     +  5.2672E+04, 5.2839E+04, 5.3007E+04, 5.3175E+04, 5.3343E+04,
     + 5.3512E+04, 5.3681E+04, 5.3851E+04, 5.4021E+04, 5.4192E+04,
     +  5.4363E+04, 5.4535E+04, 5.4706E+04, 5.4879E+04, 5.5052E+04,
     + 5.5225E+04, 5.5398E+04, 5.5573E+04, 5.5747E+04, 5.5922E+04,
     +  5.6097E+04, 5.6273E+04, 5.6450E+04, 5.6626E+04, 5.6804E+04,
     + 5.6981E+04, 5.7159E+04, 5.7338E+04, 5.7517E+04, 5.7696E+04,
     +  5.7876E+04, 5.8056E+04, 5.8237E+04, 5.8418E+04, 5.8600E+04,
     + 5.8782E+04, 5.8965E+04, 5.9148E+04, 5.9331E+04, 5.9515E+04,
     +  5.9700E+04, 5.9885E+04, 6.0070E+04, 6.0256E+04, 6.0442E+04,
     + 6.0629E+04, 6.0816E+04, 6.1003E+04, 6.1191E+04, 6.1380E+04,
     +  6.1569E+04, 6.1758E+04, 6.1948E+04, 6.2139E+04, 6.2330E+04,
     + 6.2521E+04, 6.2713E+04, 6.2905E+04, 6.3098E+04, 6.3291E+04,
     +  6.3485E+04, 6.3679E+04, 6.3874E+04, 6.4069E+04, 6.4265E+04,
     + 6.4461E+04, 6.4657E+04, 6.4854E+04, 6.5052E+04, 6.5250E+04,
     +  6.5449E+04, 6.5648E+04, 6.5847E+04, 6.6047E+04, 6.6247E+04,
     + 6.6448E+04, 6.6650E+04, 6.6852E+04, 6.7054E+04, 6.7257E+04,
     +  6.7460E+04, 6.7664E+04, 6.7869E+04, 6.8073E+04, 6.8279E+04,
     + 6.8485E+04, 6.8691E+04, 6.8898E+04, 6.9105E+04, 6.9313E+04,
     +  6.9521E+04, 6.9730E+04, 6.9940E+04, 7.0150E+04, 7.0360E+04,
     + 7.0571E+04, 7.0782E+04, 7.0994E+04, 7.1206E+04, 7.1419E+04,
     +  7.1633E+04, 7.1847E+04, 7.2061E+04, 7.2276E+04, 7.2491E+04,
     + 7.2707E+04, 7.2924E+04, 7.3141E+04, 7.3358E+04, 7.3576E+04,
     +  7.3795E+04, 7.4014E+04, 7.4233E+04, 7.4454E+04, 7.4674E+04,
     + 7.4895E+04, 7.5117E+04, 7.5339E+04, 7.5562E+04, 7.5785E+04,
     +  7.6009E+04, 7.6233E+04, 7.6458E+04, 7.6683E+04, 7.6909E+04,
     + 7.7136E+04, 7.7363E+04, 7.7590E+04, 7.7818E+04, 7.8047E+04,
     +  7.8276E+04, 7.8505E+04, 7.8736E+04, 7.8966E+04, 7.9197E+04,
     + 7.9429E+04, 7.9662E+04, 7.9894E+04, 8.0128E+04, 8.0362E+04,
     +  8.0596E+04, 8.0831E+04, 8.1067E+04, 8.1303E+04, 8.1540E+04,
     + 8.1777E+04, 8.2015E+04, 8.2253E+04, 8.2492E+04, 8.2732E+04,
     +  8.2972E+04, 8.3212E+04, 8.3453E+04, 8.3695E+04, 8.3937E+04,
     + 8.4180E+04, 8.4423E+04, 8.4667E+04, 8.4912E+04, 8.5157E+04,
     +  8.5402E+04, 8.5649E+04, 8.5895E+04, 8.6143E+04, 8.6390E+04,
     + 8.6639E+04, 8.6888E+04, 8.7137E+04, 8.7388E+04, 8.7638E+04,
     +  8.7890E+04, 8.8142E+04, 8.8394E+04, 8.8647E+04, 8.8901E+04,
     + 8.9155E+04, 8.9410E+04, 8.9665E+04, 8.9921E+04, 9.0177E+04,
     +  9.0434E+04, 9.0692E+04, 9.0950E+04, 9.1209E+04, 9.1469E+04,
     + 9.1729E+04, 9.1989E+04, 9.2250E+04, 9.2512E+04, 9.2775E+04,
     +  9.3038E+04, 9.3301E+04, 9.3565E+04, 9.3830E+04, 9.4095E+04,
     + 9.4361E+04, 9.4628E+04, 9.4895E+04, 9.5163E+04, 9.5431E+04,
     +  9.5700E+04, 9.5970E+04, 9.6240E+04, 9.6511E+04, 9.6782E+04,
     + 9.7054E+04, 9.7327E+04, 9.7600E+04, 9.7874E+04, 9.8148E+04,
     +  9.8423E+04, 9.8699E+04, 9.8975E+04, 9.9252E+04, 9.9529E+04,
     + 9.9807E+04, 1.0009E+05, 1.0037E+05, 1.0065E+05, 1.0093E+05,
     +  1.0121E+05, 1.0149E+05, 1.0177E+05, 1.0205E+05, 1.0234E+05,
     + 1.0262E+05, 1.0291E+05, 1.0319E+05, 1.0348E+05, 1.0377E+05,
     +  1.0405E+05, 1.0434E+05, 1.0463E+05, 1.0492E+05, 1.0521E+05,
     + 1.0550E+05, 1.0579E+05, 1.0608E+05, 1.0638E+05, 1.0667E+05,
     +  1.0696E+05, 1.0726E+05, 1.0755E+05, 1.0785E+05, 1.0815E+05,
     + 1.0844E+05, 1.0874E+05, 1.0904E+05, 1.0934E+05, 1.0964E+05,
     +  1.0994E+05, 1.1024E+05, 1.1054E+05, 1.1084E+05, 1.1115E+05,
     + 1.1145E+05, 1.1175E+05, 1.1206E+05, 1.1236E+05, 1.1267E+05,
     +  1.1298E+05, 1.1329E+05, 1.1359E+05, 1.1390E+05, 1.1421E+05,
     + 1.1452E+05, 1.1483E+05, 1.1515E+05, 1.1546E+05, 1.1577E+05,
     +  1.1608E+05, 1.1640E+05, 1.1671E+05, 1.1703E+05, 1.1735E+05,
     + 1.1766E+05, 1.1798E+05, 1.1830E+05, 1.1862E+05, 1.1894E+05,
     +  1.1926E+05, 1.1958E+05, 1.1990E+05, 1.2022E+05, 1.2055E+05,
     + 1.2087E+05, 1.2119E+05, 1.2152E+05, 1.2185E+05, 1.2217E+05,
     +  1.2250E+05, 1.2283E+05, 1.2316E+05, 1.2349E+05, 1.2382E+05,
     + 1.2415E+05, 1.2448E+05, 1.2481E+05, 1.2514E+05, 1.2548E+05,
     +  1.2581E+05, 1.2615E+05, 1.2648E+05, 1.2682E+05, 1.2716E+05,
     + 1.2749E+05, 1.2783E+05, 1.2817E+05, 1.2851E+05, 1.2885E+05,
     +  1.2919E+05, 1.2954E+05, 1.2988E+05, 1.3022E+05, 1.3057E+05,
     + 1.3091E+05, 1.3126E+05, 1.3160E+05, 1.3195E+05, 1.3230E+05,
     +  1.3265E+05, 1.3300E+05, 1.3335E+05, 1.3370E+05, 1.3405E+05,
     + 1.3440E+05, 1.3475E+05, 1.3511E+05, 1.3546E+05, 1.3582E+05,
     +  1.3617E+05, 1.3653E+05, 1.3689E+05, 1.3725E+05, 1.3760E+05,
     + 1.3796E+05, 1.3832E+05, 1.3869E+05, 1.3905E+05, 1.3941E+05,
     +  1.3977E+05, 1.4014E+05, 1.4050E+05, 1.4087E+05, 1.4123E+05,
     + 1.4160E+05, 1.4197E+05, 1.4234E+05, 1.4271E+05, 1.4308E+05,
     +  1.4345E+05, 1.4382E+05, 1.4419E+05, 1.4457E+05, 1.4494E+05,
     + 1.4531E+05, 1.4569E+05, 1.4607E+05, 1.4644E+05, 1.4682E+05,
     +  1.4720E+05, 1.4758E+05, 1.4796E+05, 1.4834E+05, 1.4872E+05,
     + 1.4910E+05, 1.4949E+05, 1.4987E+05, 1.5026E+05, 1.5064E+05,
     +  1.5103E+05, 1.5141E+05, 1.5180E+05, 1.5219E+05, 1.5258E+05,
     + 1.5297E+05, 1.5336E+05, 1.5375E+05, 1.5415E+05, 1.5454E+05,
     +  1.5493E+05, 1.5533E+05, 1.5572E+05, 1.5612E+05, 1.5652E+05,
     + 1.5692E+05, 1.5732E+05, 1.5772E+05, 1.5812E+05, 1.5852E+05,
     +  1.5892E+05, 1.5932E+05, 1.5973E+05, 1.6013E+05, 1.6054E+05,
     + 1.6094E+05, 1.6135E+05, 1.6176E+05, 1.6217E+05, 1.6258E+05,
     +  1.6299E+05, 1.6340E+05, 1.6381E+05, 1.6422E+05, 1.6463E+05,
     + 1.6505E+05, 1.6546E+05, 1.6588E+05, 1.6630E+05, 1.6671E+05,
     +  1.6713E+05, 1.6755E+05, 1.6797E+05, 1.6839E+05, 1.6882E+05,
     + 1.6924E+05, 1.6966E+05, 1.7009E+05, 1.7051E+05, 1.7094E+05,
     +  1.7136E+05, 1.7179E+05, 1.7222E+05, 1.7265E+05, 1.7308E+05,
     + 1.7351E+05, 1.7394E+05, 1.7437E+05, 1.7481E+05, 1.7524E+05,
     +  1.7568E+05, 1.7611E+05, 1.7655E+05, 1.7699E+05, 1.7743E+05,
     + 1.7787E+05, 1.7831E+05, 1.7875E+05, 1.7919E+05, 1.7963E+05,
     +  1.8008E+05, 1.8052E+05, 1.8097E+05, 1.8141E+05, 1.8186E+05,
     + 1.8231E+05, 1.8276E+05, 1.8321E+05, 1.8366E+05, 1.8411E+05,
     +  1.8456E+05, 1.8501E+05, 1.8547E+05, 1.8592E+05, 1.8638E+05,
     + 1.8684E+05, 1.8729E+05, 1.8775E+05, 1.8821E+05, 1.8867E+05,
     +  1.8913E+05, 1.8960E+05, 1.9006E+05, 1.9052E+05, 1.9099E+05,
     + 1.9145E+05, 1.9192E+05, 1.9239E+05, 1.9286E+05, 1.9332E+05,
     +  1.9379E+05, 1.9427E+05, 1.9474E+05, 1.9521E+05, 1.9568E+05,
     + 1.9616E+05, 1.9663E+05, 1.9711E+05, 1.9759E+05, 1.9807E+05,
     +  1.9855E+05, 1.9903E+05, 1.9951E+05, 1.9999E+05, 2.0047E+05,
     + 2.0095E+05, 2.0144E+05, 2.0192E+05, 2.0241E+05, 2.0290E+05,
     +  2.0339E+05, 2.0388E+05, 2.0437E+05, 2.0486E+05, 2.0535E+05,
     + 2.0584E+05, 2.0633E+05, 2.0683E+05, 2.0733E+05, 2.0782E+05,
     +  2.0832E+05, 2.0882E+05, 2.0932E+05, 2.0982E+05, 2.1032E+05,
     + 2.1082E+05, 2.1132E+05, 2.1183E+05, 2.1233E+05, 2.1284E+05,
     +  2.1335E+05, 2.1385E+05, 2.1436E+05, 2.1487E+05, 2.1538E+05,
     + 2.1589E+05, 2.1641E+05, 2.1692E+05, 2.1743E+05, 2.1795E+05,
     +  2.1847E+05, 2.1898E+05, 2.1950E+05, 2.2002E+05, 2.2054E+05,
     + 2.2106E+05, 2.2158E+05, 2.2211E+05, 2.2263E+05, 2.2316E+05,
     +  2.2368E+05, 2.2421E+05, 2.2474E+05, 2.2527E+05, 2.2580E+05,
     + 2.2633E+05, 2.2686E+05, 2.2739E+05, 2.2792E+05, 2.2846E+05,
     +  2.2900E+05, 2.2953E+05, 2.3007E+05, 2.3061E+05, 2.3115E+05,
     + 2.3169E+05, 2.3223E+05, 2.3277E+05, 2.3332E+05, 2.3386E+05,
     +  2.3441E+05, 2.3495E+05, 2.3550E+05, 2.3605E+05, 2.3660E+05,
     + 2.3715E+05, 2.3770E+05, 2.3825E+05, 2.3881E+05, 2.3936E+05,
     +  2.3992E+05, 2.4047E+05, 2.4103E+05, 2.4159E+05, 2.4215E+05,
     + 2.4271E+05, 2.4327E+05, 2.4383E+05, 2.4440E+05, 2.4496E+05,
     +  2.4553E+05, 2.4609E+05, 2.4666E+05, 2.4723E+05, 2.4780E+05,
     + 2.4837E+05, 2.4894E+05, 2.4951E+05, 2.5009E+05, 2.5066E+05,
     +  2.5124E+05, 2.5181E+05, 2.5239E+05, 2.5297E+05, 2.5355E+05,
     + 2.5413E+05, 2.5471E+05, 2.5530E+05, 2.5588E+05, 2.5647E+05,
     +  2.5705E+05, 2.5764E+05, 2.5823E+05, 2.5882E+05, 2.5941E+05,
     + 2.6000E+05, 2.6059E+05, 2.6118E+05, 2.6178E+05, 2.6237E+05,
     +  2.6297E+05, 2.6357E+05, 2.6417E+05, 2.6477E+05, 2.6537E+05,
     + 2.6597E+05, 2.6657E+05, 2.6718E+05, 2.6778E+05, 2.6839E+05,
     +  2.6899E+05, 2.6960E+05, 2.7021E+05, 2.7082E+05, 2.7143E+05,
     + 2.7205E+05, 2.7266E+05, 2.7327E+05, 2.7389E+05, 2.7451E+05,
     +  2.7512E+05, 2.7574E+05, 2.7636E+05, 2.7698E+05, 2.7761E+05,
     + 2.7823E+05, 2.7885E+05, 2.7948E+05, 2.8011E+05, 2.8073E+05,
     +  2.8136E+05, 2.8199E+05, 2.8262E+05, 2.8325E+05, 2.8389E+05,
     + 2.8452E+05, 2.8516E+05, 2.8579E+05, 2.8643E+05, 2.8707E+05,
     +  2.8771E+05, 2.8835E+05, 2.8899E+05, 2.8963E+05, 2.9028E+05,
     + 2.9092E+05, 2.9157E+05, 2.9222E+05, 2.9286E+05, 2.9351E+05,
     +  2.9417E+05, 2.9482E+05, 2.9547E+05, 2.9612E+05, 2.9678E+05,
     + 2.9744E+05, 2.9809E+05, 2.9875E+05, 2.9941E+05, 3.0007E+05,
     +  3.0073E+05, 3.0140E+05, 3.0206E+05, 3.0273E+05, 3.0339E+05,
     + 3.0406E+05, 3.0473E+05, 3.0540E+05, 3.0607E+05, 3.0674E+05,
     +  3.0741E+05, 3.0809E+05, 3.0876E+05, 3.0944E+05, 3.1012E+05,
     + 3.1080E+05, 3.1148E+05, 3.1216E+05, 3.1284E+05, 3.1352E+05,
     +  3.1421E+05, 3.1489E+05, 3.1558E+05, 3.1627E+05, 3.1696E+05,
     + 3.1765E+05, 3.1834E+05, 3.1903E+05, 3.1973E+05, 3.2042E+05,
     +  3.2112E+05, 3.2181E+05, 3.2251E+05, 3.2321E+05, 3.2391E+05,
     + 3.2462E+05, 3.2532E+05, 3.2602E+05, 3.2673E+05, 3.2744E+05,
     +  3.2814E+05, 3.2885E+05, 3.2956E+05, 3.3027E+05, 3.3099E+05,
     + 3.3170E+05, 3.3242E+05, 3.3313E+05, 3.3385E+05, 3.3457E+05,
     +  3.3529E+05, 3.3601E+05, 3.3673E+05, 3.3745E+05, 3.3818E+05,
     + 3.3890E+05, 3.3963E+05, 3.4036E+05, 3.4109E+05, 3.4182E+05,
     +  3.4255E+05, 3.4328E+05, 3.4402E+05, 3.4475E+05, 3.4549E+05,
     + 3.4622E+05, 3.4696E+05, 3.4770E+05, 3.4844E+05, 3.4919E+05,
     +  3.4993E+05, 3.5068E+05, 3.5142E+05, 3.5217E+05, 3.5292E+05,
     + 3.5367E+05, 3.5442E+05, 3.5517E+05, 3.5592E+05, 3.5668E+05,
     +  3.5743E+05, 3.5819E+05, 3.5895E+05, 3.5971E+05, 3.6047E+05,
     + 3.6123E+05, 3.6199E+05, 3.6276E+05, 3.6352E+05, 3.6429E+05,
     +  3.6506E+05, 3.6583E+05, 3.6660E+05, 3.6737E+05, 3.6814E+05,
     + 3.6892E+05, 3.6969E+05, 3.7047E+05, 3.7125E+05, 3.7203E+05,
     +  3.7281E+05, 3.7359E+05, 3.7437E+05, 3.7516E+05, 3.7594E+05,
     + 3.7673E+05, 3.7752E+05, 3.7831E+05, 3.7910E+05, 3.7989E+05,
     +  3.8068E+05, 3.8148E+05, 3.8227E+05, 3.8307E+05, 3.8387E+05,
     + 3.8467E+05, 3.8547E+05, 3.8627E+05, 3.8707E+05, 3.8788E+05,
     +  3.8868E+05, 3.8949E+05, 3.9030E+05, 3.9111E+05, 3.9192E+05,
     + 3.9273E+05, 3.9355E+05, 3.9436E+05, 3.9518E+05, 3.9599E+05,
     +  3.9681E+05, 3.9763E+05, 3.9845E+05, 3.9928E+05, 4.0010E+05,
     + 4.0093E+05, 4.0175E+05, 4.0258E+05, 4.0341E+05, 4.0424E+05,
     +  4.0507E+05, 4.0590E+05, 4.0674E+05, 4.0757E+05, 4.0841E+05,
     + 4.0925E+05, 4.1009E+05, 4.1093E+05, 4.1177E+05, 4.1261E+05,
     +  4.1346E+05, 4.1430E+05, 4.1515E+05, 4.1600E+05, 4.1685E+05,
     + 4.1770E+05, 4.1855E+05, 4.1941E+05, 4.2026E+05, 4.2112E+05,
     +  4.2198E+05, 4.2283E+05, 4.2369E+05, 4.2456E+05, 4.2542E+05,
     + 4.2628E+05, 4.2715E+05, 4.2802E+05, 4.2888E+05, 4.2975E+05,
     +  4.3063E+05, 4.3150E+05, 4.3237E+05, 4.3325E+05, 4.3412E+05,
     + 4.3500E+05, 4.3588E+05, 4.3676E+05, 4.3764E+05, 4.3853E+05,
     +  4.3941E+05, 4.4030E+05, 4.4118E+05, 4.4207E+05, 4.4296E+05,
     + 4.4385E+05, 4.4474E+05, 4.4564E+05, 4.4653E+05, 4.4743E+05,
     +  4.4833E+05, 4.4923E+05, 4.5013E+05, 4.5103E+05, 4.5193E+05,
     + 4.5284E+05, 4.5375E+05, 4.5465E+05, 4.5556E+05, 4.5647E+05,
     +  4.5738E+05, 4.5830E+05, 4.5921E+05, 4.6013E+05, 4.6104E+05,
     + 4.6196E+05, 4.6288E+05, 4.6380E+05, 4.6473E+05, 4.6565E+05,
     +  4.6658E+05, 4.6750E+05, 4.6843E+05, 4.6936E+05, 4.7029E+05,
     + 4.7122E+05, 4.7216E+05, 4.7309E+05, 4.7403E+05, 4.7497E+05,
     +  4.7591E+05, 4.7685E+05, 4.7779E+05, 4.7873E+05, 4.7968E+05,
     + 4.8063E+05, 4.8157E+05, 4.8252E+05, 4.8347E+05, 4.8443E+05,
     +  4.8538E+05, 4.8633E+05, 4.8729E+05, 4.8825E+05, 4.8921E+05,
     + 4.9017E+05, 4.9113E+05, 4.9209E+05, 4.9306E+05, 4.9402E+05,
     +  4.9499E+05, 4.9596E+05, 4.9693E+05, 4.9790E+05, 4.9888E+05,
     + 4.9985E+05, 5.0083E+05, 5.0180E+05, 5.0278E+05, 5.0376E+05,
     +  5.0475E+05, 5.0573E+05, 5.0671E+05, 5.0770E+05, 5.0869E+05,
     + 5.0968E+05, 5.1067E+05, 5.1166E+05, 5.1265E+05, 5.1365E+05,
     +  5.1464E+05, 5.1564E+05, 5.1664E+05, 5.1764E+05, 5.1864E+05,
     + 5.1965E+05, 5.2065E+05, 5.2166E+05, 5.2267E+05, 5.2367E+05,
     +  5.2469E+05, 5.2570E+05, 5.2671E+05, 5.2773E+05, 5.2874E+05,
     + 5.2976E+05, 5.3078E+05, 5.3180E+05, 5.3282E+05, 5.3385E+05,
     +  5.3487E+05, 5.3590E+05, 5.3693E+05, 5.3796E+05, 5.3899E+05,
     + 5.4002E+05, 5.4106E+05, 5.4209E+05, 5.4313E+05, 5.4417E+05,
     +  5.4521E+05, 5.4625E+05, 5.4729E+05, 5.4834E+05, 5.4938E+05,
     + 5.5043E+05, 5.5148E+05, 5.5253E+05, 5.5358E+05, 5.5464E+05,
     +  5.5569E+05, 5.5675E+05, 5.5781E+05, 5.5887E+05, 5.5993E+05,
     + 5.6099E+05, 5.6205E+05, 5.6312E+05, 5.6419E+05, 5.6526E+05,
     +  5.6633E+05, 5.6740E+05, 5.6847E+05, 5.6954E+05, 5.7062E+05,
     + 5.7170E+05, 5.7278E+05, 5.7386E+05, 5.7494E+05, 5.7602E+05,
     +  5.7711E+05, 5.7820E+05, 5.7928E+05, 5.8037E+05, 5.8147E+05,
     + 5.8256E+05, 5.8365E+05, 5.8475E+05, 5.8585E+05, 5.8694E+05,
     +  5.8804E+05, 5.8915E+05, 5.9025E+05, 5.9136E+05, 5.9246E+05,
     + 5.9357E+05, 5.9468E+05, 5.9579E+05, 5.9690E+05, 5.9802E+05,
     +  5.9913E+05, 6.0025E+05, 6.0137E+05, 6.0249E+05, 6.0361E+05,
     + 6.0474E+05, 6.0586E+05, 6.0699E+05, 6.0812E+05, 6.0925E+05,
     +  6.1038E+05, 6.1151E+05, 6.1264E+05, 6.1378E+05, 6.1492E+05,
     + 6.1606E+05, 6.1720E+05, 6.1834E+05, 6.1948E+05, 6.2063E+05,
     +  6.2177E+05, 6.2292E+05, 6.2407E+05, 6.2522E+05, 6.2638E+05,
     + 6.2753E+05, 6.2869E+05, 6.2985E+05, 6.3101E+05, 6.3217E+05,
     +  6.3333E+05, 6.3449E+05, 6.3566E+05, 6.3683E+05, 6.3800E+05,
     + 6.3917E+05, 6.4034E+05, 6.4151E+05, 6.4269E+05, 6.4386E+05,
     +  6.4504E+05, 6.4622E+05, 6.4740E+05, 6.4859E+05, 6.4977E+05,
     + 6.5096E+05, 6.5215E+05, 6.5334E+05, 6.5453E+05, 6.5572E+05,
     +  6.5691E+05, 6.5811E+05, 6.5931E+05, 6.6051E+05, 6.6171E+05,
     + 6.6291E+05, 6.6411E+05, 6.6532E+05, 6.6653E+05, 6.6773E+05,
     +  6.6895E+05, 6.7016E+05, 6.7137E+05, 6.7259E+05, 6.7380E+05,
     + 6.7502E+05, 6.7624E+05, 6.7746E+05, 6.7869E+05, 6.7991E+05,
     +  6.8114E+05, 6.8237E+05, 6.8360E+05, 6.8483E+05, 6.8606E+05,
     + 6.8729E+05, 6.8853E+05, 6.8977E+05, 6.9101E+05, 6.9225E+05,
     +  6.9349E+05, 6.9474E+05, 6.9598E+05, 6.9723E+05, 6.9848E+05,
     + 6.9973E+05, 7.0098E+05, 7.0224E+05, 7.0349E+05, 7.0475E+05,
     +  7.0601E+05, 7.0727E+05, 7.0853E+05, 7.0980E+05, 7.1106E+05,
     + 7.1233E+05, 7.1360E+05, 7.1487E+05, 7.1614E+05, 7.1742E+05,
     +  7.1869E+05, 7.1997E+05, 7.2125E+05, 7.2253E+05, 7.2381E+05,
     + 7.2510E+05, 7.2638E+05, 7.2767E+05, 7.2896E+05, 7.3025E+05,
     +  7.3154E+05, 7.3284E+05, 7.3413E+05, 7.3543E+05, 7.3673E+05,
     + 7.3803E+05, 7.3933E+05, 7.4064E+05, 7.4194E+05, 7.4325E+05,
     +  7.4456E+05, 7.4587E+05, 7.4718E+05, 7.4850E+05, 7.4981E+05,
     + 7.5113E+05, 7.5245E+05, 7.5377E+05, 7.5509E+05, 7.5642E+05,
     +  7.5774E+05, 7.5907E+05, 7.6040E+05, 7.6173E+05, 7.6306E+05,
     + 7.6440E+05, 7.6573E+05, 7.6707E+05, 7.6841E+05, 7.6975E+05,
     +  7.7109E+05, 7.7244E+05, 7.7378E+05, 7.7513E+05, 7.7648E+05,
     + 7.7783E+05, 7.7919E+05, 7.8054E+05, 7.8190E+05, 7.8326E+05,
     +  7.8462E+05, 7.8598E+05, 7.8734E+05, 7.8871E+05, 7.9007E+05,
     + 7.9144E+05, 7.9281E+05, 7.9418E+05, 7.9556E+05, 7.9693E+05,
     +  7.9831E+05, 7.9969E+05, 8.0107E+05, 8.0245E+05, 8.0384E+05,
     + 8.0522E+05, 8.0661E+05, 8.0800E+05, 8.0939E+05, 8.1078E+05,
     +  8.1218E+05, 8.1357E+05, 8.1497E+05, 8.1637E+05, 8.1777E+05,
     + 8.1917E+05, 8.2058E+05, 8.2199E+05, 8.2339E+05, 8.2480E+05,
     +  8.2622E+05, 8.2763E+05, 8.2904E+05, 8.3046E+05, 8.3188E+05,
     + 8.3330E+05, 8.3472E+05, 8.3615E+05, 8.3757E+05, 8.3900E+05,
     +  8.4043E+05, 8.4186E+05, 8.4329E+05, 8.4473E+05, 8.4616E+05,
     + 8.4760E+05, 8.4904E+05, 8.5048E+05, 8.5193E+05, 8.5337E+05,
     +  8.5482E+05, 8.5627E+05, 8.5772E+05, 8.5917E+05, 8.6062E+05/
      data (QofT(13,J),J=2409,4510)/ 8.6208E+05, 8.6354E+05,
     +  8.6500E+05, 8.6646E+05, 8.6792E+05, 8.6938E+05, 8.7085E+05,
     +  8.7232E+05,
     + 8.7379E+05, 8.7526E+05, 8.7673E+05, 8.7821E+05, 8.7968E+05,
     +  8.8116E+05, 8.8264E+05, 8.8413E+05, 8.8561E+05, 8.8710E+05,
     + 8.8858E+05, 8.9007E+05, 8.9157E+05, 8.9306E+05, 8.9455E+05,
     +  8.9605E+05, 8.9755E+05, 8.9905E+05, 9.0055E+05, 9.0205E+05,
     + 9.0356E+05, 9.0507E+05, 9.0658E+05, 9.0809E+05, 9.0960E+05,
     +  9.1111E+05, 9.1263E+05, 9.1415E+05, 9.1567E+05, 9.1719E+05,
     + 9.1872E+05, 9.2024E+05, 9.2177E+05, 9.2330E+05, 9.2483E+05,
     +  9.2636E+05, 9.2790E+05, 9.2943E+05, 9.3097E+05, 9.3251E+05,
     + 9.3405E+05, 9.3559E+05, 9.3714E+05, 9.3869E+05, 9.4024E+05,
     +  9.4179E+05, 9.4334E+05, 9.4489E+05, 9.4645E+05, 9.4801E+05,
     + 9.4957E+05, 9.5113E+05, 9.5269E+05, 9.5426E+05, 9.5583E+05,
     +  9.5740E+05, 9.5897E+05, 9.6054E+05, 9.6212E+05, 9.6369E+05,
     + 9.6527E+05, 9.6685E+05, 9.6843E+05, 9.7002E+05, 9.7160E+05,
     +  9.7319E+05, 9.7478E+05, 9.7637E+05, 9.7796E+05, 9.7956E+05,
     + 9.8116E+05, 9.8275E+05, 9.8435E+05, 9.8596E+05, 9.8756E+05,
     +  9.8917E+05, 9.9077E+05, 9.9238E+05, 9.9400E+05, 9.9561E+05,
     + 9.9722E+05, 9.9884E+05, 1.0005E+06, 1.0021E+06, 1.0037E+06,
     +  1.0053E+06, 1.0070E+06, 1.0086E+06, 1.0102E+06, 1.0118E+06,
     + 1.0135E+06, 1.0151E+06, 1.0168E+06, 1.0184E+06, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_O3    (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj(18),QofT(18,NT),Q(NT)
      data (Tmx(I),I=1,18)/3500.,3500.,3500.,3500.,3500.,3500.,3500.,
     + 3500.,3500.,
     +                     3500.,3500.,3500.,3500.,3500.,3500.,3500.,
     + 3500.,3500./
      data xgj/ 1.,1.,1.,6.,6.,1.,1.,6.,6.,6.,36.,
     + 1.,1.,6.,6.,36.,1.,6./
c...       O3
c...        --       666
      data (QofT( 1,J),J=1,2408)/ 7.8474E-01, 2.0241E+00, 3.6006E+00,
     +  5.4547E+00, 7.5499E+00, 9.8611E+00, 1.2370E+01, 1.5061E+01,
     + 1.7924E+01, 2.0949E+01, 2.4127E+01, 2.7452E+01, 3.0916E+01,
     +  3.4516E+01, 3.8245E+01, 4.2100E+01, 4.6076E+01, 5.0171E+01,
     + 5.4380E+01, 5.8701E+01, 6.7667E+01, 7.7048E+01, 8.6827E+01,
     +  9.6989E+01, 1.0752E+02, 1.1841E+02, 1.2964E+02, 1.4121E+02,
     + 1.5310E+02, 1.6531E+02, 1.7783E+02, 1.9065E+02, 2.0376E+02,
     +  2.1717E+02, 2.3085E+02, 2.4482E+02, 2.5905E+02, 2.7355E+02,
     + 2.8831E+02, 3.0333E+02, 3.1861E+02, 3.3413E+02, 3.4990E+02,
     +  3.6590E+02, 3.8215E+02, 3.9863E+02, 4.1534E+02, 4.3228E+02,
     + 4.4945E+02, 4.6683E+02, 4.8444E+02, 5.0227E+02, 5.2031E+02,
     +  5.3856E+02, 5.5702E+02, 5.7569E+02, 5.9456E+02, 6.1364E+02,
     + 6.3292E+02, 6.5240E+02, 6.7208E+02, 6.9195E+02, 7.1202E+02,
     +  7.3229E+02, 7.5274E+02, 7.7339E+02, 7.9422E+02, 8.1524E+02,
     + 8.3645E+02, 8.5784E+02, 8.7941E+02, 9.0118E+02, 9.2312E+02,
     +  9.4525E+02, 9.6755E+02, 9.9005E+02, 1.0127E+03, 1.0356E+03,
     + 1.0586E+03, 1.0818E+03, 1.1052E+03, 1.1287E+03, 1.1525E+03,
     +  1.1764E+03, 1.2005E+03, 1.2247E+03, 1.2492E+03, 1.2738E+03,
     + 1.2986E+03, 1.3236E+03, 1.3487E+03, 1.3741E+03, 1.3996E+03,
     +  1.4252E+03, 1.4511E+03, 1.4772E+03, 1.5034E+03, 1.5298E+03,
     + 1.5564E+03, 1.5831E+03, 1.6101E+03, 1.6372E+03, 1.6645E+03,
     +  1.6920E+03, 1.7197E+03, 1.7475E+03, 1.7756E+03, 1.8038E+03,
     + 1.8323E+03, 1.8609E+03, 1.8897E+03, 1.9187E+03, 1.9479E+03,
     +  1.9773E+03, 2.0068E+03, 2.0366E+03, 2.0666E+03, 2.0968E+03,
     + 2.1271E+03, 2.1577E+03, 2.1885E+03, 2.2194E+03, 2.2506E+03,
     +  2.2820E+03, 2.3136E+03, 2.3454E+03, 2.3774E+03, 2.4097E+03,
     + 2.4421E+03, 2.4748E+03, 2.5077E+03, 2.5408E+03, 2.5741E+03,
     +  2.6076E+03, 2.6414E+03, 2.6754E+03, 2.7096E+03, 2.7440E+03,
     + 2.7787E+03, 2.8136E+03, 2.8487E+03, 2.8841E+03, 2.9197E+03,
     +  2.9556E+03, 2.9917E+03, 3.0280E+03, 3.0646E+03, 3.1014E+03,
     + 3.1385E+03, 3.1758E+03, 3.2134E+03, 3.2513E+03, 3.2894E+03,
     +  3.3277E+03, 3.3663E+03, 3.4052E+03, 3.4443E+03, 3.4837E+03,
     + 3.5234E+03, 3.5634E+03, 3.6035E+03, 3.6440E+03, 3.6848E+03,
     +  3.7258E+03, 3.7672E+03, 3.8088E+03, 3.8506E+03, 3.8928E+03,
     + 3.9353E+03, 3.9780E+03, 4.0211E+03, 4.0645E+03, 4.1081E+03,
     +  4.1520E+03, 4.1963E+03, 4.2408E+03, 4.2857E+03, 4.3308E+03,
     + 4.3763E+03, 4.4220E+03, 4.4682E+03, 4.5146E+03, 4.5613E+03,
     +  4.6084E+03, 4.6557E+03, 4.7034E+03, 4.7514E+03, 4.7998E+03,
     + 4.8485E+03, 4.8975E+03, 4.9469E+03, 4.9966E+03, 5.0466E+03,
     +  5.0970E+03, 5.1477E+03, 5.1988E+03, 5.2503E+03, 5.3020E+03,
     + 5.3542E+03, 5.4067E+03, 5.4595E+03, 5.5127E+03, 5.5663E+03,
     +  5.6203E+03, 5.6746E+03, 5.7293E+03, 5.7843E+03, 5.8398E+03,
     + 5.8956E+03, 5.9518E+03, 6.0083E+03, 6.0653E+03, 6.1227E+03,
     +  6.1805E+03, 6.2386E+03, 6.2971E+03, 6.3561E+03, 6.4154E+03,
     + 6.4751E+03, 6.5353E+03, 6.5958E+03, 6.6568E+03, 6.7182E+03,
     +  6.7800E+03, 6.8422E+03, 6.9048E+03, 6.9678E+03, 7.0313E+03,
     + 7.0952E+03, 7.1596E+03, 7.2243E+03, 7.2895E+03, 7.3552E+03,
     +  7.4213E+03, 7.4878E+03, 7.5548E+03, 7.6222E+03, 7.6900E+03,
     + 7.7584E+03, 7.8272E+03, 7.8964E+03, 7.9661E+03, 8.0363E+03,
     +  8.1069E+03, 8.1780E+03, 8.2496E+03, 8.3216E+03, 8.3941E+03,
     + 8.4671E+03, 8.5406E+03, 8.6146E+03, 8.6891E+03, 8.7640E+03,
     +  8.8394E+03, 8.9153E+03, 8.9918E+03, 9.0687E+03, 9.1462E+03,
     + 9.2241E+03, 9.3026E+03, 9.3815E+03, 9.4611E+03, 9.5410E+03,
     +  9.6216E+03, 9.7026E+03, 9.7841E+03, 9.8662E+03, 9.9489E+03,
     + 1.0032E+04, 1.0116E+04, 1.0200E+04, 1.0285E+04, 1.0370E+04,
     +  1.0456E+04, 1.0542E+04, 1.0629E+04, 1.0717E+04, 1.0805E+04,
     + 1.0894E+04, 1.0983E+04, 1.1073E+04, 1.1163E+04, 1.1254E+04,
     +  1.1345E+04, 1.1438E+04, 1.1530E+04, 1.1624E+04, 1.1717E+04,
     + 1.1812E+04, 1.1907E+04, 1.2002E+04, 1.2099E+04, 1.2196E+04,
     +  1.2293E+04, 1.2391E+04, 1.2490E+04, 1.2589E+04, 1.2689E+04,
     + 1.2789E+04, 1.2891E+04, 1.2992E+04, 1.3095E+04, 1.3198E+04,
     +  1.3301E+04, 1.3406E+04, 1.3511E+04, 1.3616E+04, 1.3723E+04,
     + 1.3829E+04, 1.3937E+04, 1.4045E+04, 1.4154E+04, 1.4264E+04,
     +  1.4374E+04, 1.4485E+04, 1.4596E+04, 1.4709E+04, 1.4822E+04,
     + 1.4935E+04, 1.5049E+04, 1.5164E+04, 1.5280E+04, 1.5396E+04,
     +  1.5513E+04, 1.5631E+04, 1.5750E+04, 1.5869E+04, 1.5989E+04,
     + 1.6109E+04, 1.6231E+04, 1.6353E+04, 1.6476E+04, 1.6599E+04,
     +  1.6723E+04, 1.6848E+04, 1.6974E+04, 1.7101E+04, 1.7228E+04,
     + 1.7356E+04, 1.7484E+04, 1.7614E+04, 1.7744E+04, 1.7875E+04,
     +  1.8007E+04, 1.8139E+04, 1.8273E+04, 1.8407E+04, 1.8542E+04,
     + 1.8677E+04, 1.8814E+04, 1.8951E+04, 1.9089E+04, 1.9228E+04,
     +  1.9367E+04, 1.9508E+04, 1.9649E+04, 1.9791E+04, 1.9934E+04,
     + 2.0077E+04, 2.0222E+04, 2.0367E+04, 2.0513E+04, 2.0660E+04,
     +  2.0808E+04, 2.0957E+04, 2.1106E+04, 2.1257E+04, 2.1408E+04,
     + 2.1560E+04, 2.1713E+04, 2.1866E+04, 2.2021E+04, 2.2177E+04,
     +  2.2333E+04, 2.2490E+04, 2.2648E+04, 2.2807E+04, 2.2967E+04,
     + 2.3128E+04, 2.3290E+04, 2.3452E+04, 2.3616E+04, 2.3780E+04,
     +  2.3945E+04, 2.4112E+04, 2.4279E+04, 2.4447E+04, 2.4616E+04,
     + 2.4786E+04, 2.4956E+04, 2.5128E+04, 2.5301E+04, 2.5474E+04,
     +  2.5649E+04, 2.5825E+04, 2.6001E+04, 2.6179E+04, 2.6357E+04,
     + 2.6536E+04, 2.6717E+04, 2.6898E+04, 2.7080E+04, 2.7264E+04,
     +  2.7448E+04, 2.7633E+04, 2.7820E+04, 2.8007E+04, 2.8195E+04,
     + 2.8384E+04, 2.8575E+04, 2.8766E+04, 2.8958E+04, 2.9152E+04,
     +  2.9346E+04, 2.9541E+04, 2.9738E+04, 2.9935E+04, 3.0134E+04,
     + 3.0333E+04, 3.0534E+04, 3.0736E+04, 3.0938E+04, 3.1142E+04,
     +  3.1347E+04, 3.1553E+04, 3.1760E+04, 3.1968E+04, 3.2177E+04,
     + 3.2387E+04, 3.2599E+04, 3.2811E+04, 3.3025E+04, 3.3239E+04,
     +  3.3455E+04, 3.3672E+04, 3.3890E+04, 3.4109E+04, 3.4330E+04,
     + 3.4551E+04, 3.4773E+04, 3.4997E+04, 3.5222E+04, 3.5448E+04,
     +  3.5675E+04, 3.5903E+04, 3.6132E+04, 3.6363E+04, 3.6595E+04,
     + 3.6828E+04, 3.7062E+04, 3.7297E+04, 3.7534E+04, 3.7771E+04,
     +  3.8010E+04, 3.8250E+04, 3.8491E+04, 3.8734E+04, 3.8978E+04,
     + 3.9222E+04, 3.9469E+04, 3.9716E+04, 3.9965E+04, 4.0214E+04,
     +  4.0465E+04, 4.0718E+04, 4.0971E+04, 4.1226E+04, 4.1482E+04,
     + 4.1740E+04, 4.1998E+04, 4.2258E+04, 4.2519E+04, 4.2782E+04,
     +  4.3046E+04, 4.3311E+04, 4.3577E+04, 4.3845E+04, 4.4113E+04,
     + 4.4384E+04, 4.4655E+04, 4.4928E+04, 4.5202E+04, 4.5478E+04,
     +  4.5755E+04, 4.6033E+04, 4.6312E+04, 4.6593E+04, 4.6876E+04,
     + 4.7159E+04, 4.7444E+04, 4.7731E+04, 4.8018E+04, 4.8307E+04,
     +  4.8598E+04, 4.8890E+04, 4.9183E+04, 4.9478E+04, 4.9774E+04,
     + 5.0071E+04, 5.0370E+04, 5.0670E+04, 5.0972E+04, 5.1275E+04,
     +  5.1580E+04, 5.1886E+04, 5.2193E+04, 5.2502E+04, 5.2812E+04,
     + 5.3124E+04, 5.3438E+04, 5.3752E+04, 5.4068E+04, 5.4386E+04,
     +  5.4705E+04, 5.5026E+04, 5.5348E+04, 5.5672E+04, 5.5997E+04,
     + 5.6323E+04, 5.6652E+04, 5.6981E+04, 5.7312E+04, 5.7645E+04,
     +  5.7979E+04, 5.8315E+04, 5.8653E+04, 5.8991E+04, 5.9332E+04,
     + 5.9674E+04, 6.0017E+04, 6.0363E+04, 6.0709E+04, 6.1058E+04,
     +  6.1407E+04, 6.1759E+04, 6.2112E+04, 6.2467E+04, 6.2823E+04,
     + 6.3181E+04, 6.3540E+04, 6.3901E+04, 6.4264E+04, 6.4629E+04,
     +  6.4994E+04, 6.5362E+04, 6.5731E+04, 6.6102E+04, 6.6475E+04,
     + 6.6849E+04, 6.7225E+04, 6.7603E+04, 6.7982E+04, 6.8363E+04,
     +  6.8746E+04, 6.9130E+04, 6.9516E+04, 6.9904E+04, 7.0294E+04,
     + 7.0685E+04, 7.1078E+04, 7.1473E+04, 7.1869E+04, 7.2267E+04,
     +  7.2667E+04, 7.3069E+04, 7.3472E+04, 7.3878E+04, 7.4284E+04,
     + 7.4693E+04, 7.5104E+04, 7.5516E+04, 7.5930E+04, 7.6346E+04,
     +  7.6764E+04, 7.7183E+04, 7.7605E+04, 7.8028E+04, 7.8453E+04,
     + 7.8880E+04, 7.9308E+04, 7.9739E+04, 8.0171E+04, 8.0606E+04,
     +  8.1042E+04, 8.1480E+04, 8.1919E+04, 8.2361E+04, 8.2805E+04,
     + 8.3250E+04, 8.3698E+04, 8.4147E+04, 8.4598E+04, 8.5051E+04,
     +  8.5506E+04, 8.5963E+04, 8.6422E+04, 8.6883E+04, 8.7346E+04,
     + 8.7811E+04, 8.8277E+04, 8.8746E+04, 8.9217E+04, 8.9689E+04,
     +  9.0164E+04, 9.0640E+04, 9.1119E+04, 9.1599E+04, 9.2082E+04,
     + 9.2567E+04, 9.3054E+04, 9.3542E+04, 9.4033E+04, 9.4525E+04,
     +  9.5020E+04, 9.5517E+04, 9.6016E+04, 9.6517E+04, 9.7020E+04,
     + 9.7525E+04, 9.8032E+04, 9.8541E+04, 9.9052E+04, 9.9566E+04,
     +  1.0008E+05, 1.0060E+05, 1.0112E+05, 1.0164E+05, 1.0216E+05,
     + 1.0269E+05, 1.0322E+05, 1.0375E+05, 1.0428E+05, 1.0482E+05,
     +  1.0535E+05, 1.0589E+05, 1.0643E+05, 1.0698E+05, 1.0752E+05,
     + 1.0807E+05, 1.0862E+05, 1.0917E+05, 1.0973E+05, 1.1028E+05,
     +  1.1084E+05, 1.1140E+05, 1.1197E+05, 1.1253E+05, 1.1310E+05,
     + 1.1367E+05, 1.1424E+05, 1.1482E+05, 1.1540E+05, 1.1598E+05,
     +  1.1656E+05, 1.1714E+05, 1.1773E+05, 1.1832E+05, 1.1891E+05,
     + 1.1950E+05, 1.2010E+05, 1.2069E+05, 1.2129E+05, 1.2190E+05,
     +  1.2250E+05, 1.2311E+05, 1.2372E+05, 1.2433E+05, 1.2495E+05,
     + 1.2556E+05, 1.2618E+05, 1.2680E+05, 1.2743E+05, 1.2806E+05,
     +  1.2868E+05, 1.2932E+05, 1.2995E+05, 1.3059E+05, 1.3123E+05,
     + 1.3187E+05, 1.3251E+05, 1.3316E+05, 1.3381E+05, 1.3446E+05,
     +  1.3511E+05, 1.3577E+05, 1.3643E+05, 1.3709E+05, 1.3775E+05,
     + 1.3842E+05, 1.3909E+05, 1.3976E+05, 1.4043E+05, 1.4111E+05,
     +  1.4179E+05, 1.4247E+05, 1.4315E+05, 1.4384E+05, 1.4453E+05,
     + 1.4522E+05, 1.4592E+05, 1.4662E+05, 1.4732E+05, 1.4802E+05,
     +  1.4872E+05, 1.4943E+05, 1.5014E+05, 1.5086E+05, 1.5157E+05,
     + 1.5229E+05, 1.5301E+05, 1.5374E+05, 1.5446E+05, 1.5519E+05,
     +  1.5593E+05, 1.5666E+05, 1.5740E+05, 1.5814E+05, 1.5888E+05,
     + 1.5963E+05, 1.6038E+05, 1.6113E+05, 1.6188E+05, 1.6264E+05,
     +  1.6340E+05, 1.6416E+05, 1.6493E+05, 1.6570E+05, 1.6647E+05,
     + 1.6724E+05, 1.6802E+05, 1.6880E+05, 1.6958E+05, 1.7037E+05,
     +  1.7115E+05, 1.7194E+05, 1.7274E+05, 1.7354E+05, 1.7434E+05,
     + 1.7514E+05, 1.7594E+05, 1.7675E+05, 1.7756E+05, 1.7838E+05,
     +  1.7920E+05, 1.8002E+05, 1.8084E+05, 1.8167E+05, 1.8250E+05,
     + 1.8333E+05, 1.8416E+05, 1.8500E+05, 1.8584E+05, 1.8669E+05,
     +  1.8753E+05, 1.8838E+05, 1.8924E+05, 1.9009E+05, 1.9095E+05,
     + 1.9182E+05, 1.9268E+05, 1.9355E+05, 1.9442E+05, 1.9530E+05,
     +  1.9618E+05, 1.9706E+05, 1.9794E+05, 1.9883E+05, 1.9972E+05,
     + 2.0061E+05, 2.0151E+05, 2.0241E+05, 2.0331E+05, 2.0422E+05,
     +  2.0513E+05, 2.0604E+05, 2.0696E+05, 2.0788E+05, 2.0880E+05,
     + 2.0972E+05, 2.1065E+05, 2.1158E+05, 2.1252E+05, 2.1346E+05,
     +  2.1440E+05, 2.1534E+05, 2.1629E+05, 2.1724E+05, 2.1820E+05,
     + 2.1916E+05, 2.2012E+05, 2.2108E+05, 2.2205E+05, 2.2302E+05,
     +  2.2400E+05, 2.2498E+05, 2.2596E+05, 2.2694E+05, 2.2793E+05,
     + 2.2892E+05, 2.2992E+05, 2.3092E+05, 2.3192E+05, 2.3293E+05,
     +  2.3394E+05, 2.3495E+05, 2.3596E+05, 2.3698E+05, 2.3801E+05,
     + 2.3903E+05, 2.4006E+05, 2.4109E+05, 2.4213E+05, 2.4317E+05,
     +  2.4421E+05, 2.4526E+05, 2.4631E+05, 2.4737E+05, 2.4843E+05,
     + 2.4949E+05, 2.5055E+05, 2.5162E+05, 2.5269E+05, 2.5377E+05,
     +  2.5485E+05, 2.5593E+05, 2.5702E+05, 2.5811E+05, 2.5920E+05,
     + 2.6030E+05, 2.6140E+05, 2.6251E+05, 2.6362E+05, 2.6473E+05,
     +  2.6584E+05, 2.6696E+05, 2.6809E+05, 2.6921E+05, 2.7035E+05,
     + 2.7148E+05, 2.7262E+05, 2.7376E+05, 2.7491E+05, 2.7606E+05,
     +  2.7721E+05, 2.7837E+05, 2.7953E+05, 2.8069E+05, 2.8186E+05,
     + 2.8303E+05, 2.8421E+05, 2.8539E+05, 2.8658E+05, 2.8776E+05,
     +  2.8896E+05, 2.9015E+05, 2.9135E+05, 2.9255E+05, 2.9376E+05,
     + 2.9497E+05, 2.9619E+05, 2.9741E+05, 2.9863E+05, 2.9986E+05,
     +  3.0109E+05, 3.0233E+05, 3.0356E+05, 3.0481E+05, 3.0606E+05,
     + 3.0731E+05, 3.0856E+05, 3.0982E+05, 3.1108E+05, 3.1235E+05,
     +  3.1362E+05, 3.1490E+05, 3.1618E+05, 3.1746E+05, 3.1875E+05,
     + 3.2004E+05, 3.2134E+05, 3.2264E+05, 3.2394E+05, 3.2525E+05,
     +  3.2656E+05, 3.2788E+05, 3.2920E+05, 3.3053E+05, 3.3186E+05,
     + 3.3319E+05, 3.3453E+05, 3.3587E+05, 3.3722E+05, 3.3857E+05,
     +  3.3992E+05, 3.4128E+05, 3.4264E+05, 3.4401E+05, 3.4538E+05,
     + 3.4676E+05, 3.4814E+05, 3.4953E+05, 3.5091E+05, 3.5231E+05,
     +  3.5371E+05, 3.5511E+05, 3.5652E+05, 3.5793E+05, 3.5934E+05,
     + 3.6076E+05, 3.6219E+05, 3.6361E+05, 3.6505E+05, 3.6649E+05,
     +  3.6793E+05, 3.6937E+05, 3.7082E+05, 3.7228E+05, 3.7374E+05,
     + 3.7521E+05, 3.7667E+05, 3.7815E+05, 3.7963E+05, 3.8111E+05,
     +  3.8260E+05, 3.8409E+05, 3.8558E+05, 3.8708E+05, 3.8859E+05,
     + 3.9010E+05, 3.9162E+05, 3.9313E+05, 3.9466E+05, 3.9619E+05,
     +  3.9772E+05, 3.9926E+05, 4.0080E+05, 4.0235E+05, 4.0390E+05,
     + 4.0546E+05, 4.0702E+05, 4.0859E+05, 4.1016E+05, 4.1173E+05,
     +  4.1331E+05, 4.1490E+05, 4.1649E+05, 4.1808E+05, 4.1968E+05,
     + 4.2129E+05, 4.2290E+05, 4.2451E+05, 4.2613E+05, 4.2775E+05,
     +  4.2938E+05, 4.3102E+05, 4.3266E+05, 4.3430E+05, 4.3595E+05,
     + 4.3760E+05, 4.3926E+05, 4.4092E+05, 4.4259E+05, 4.4426E+05,
     +  4.4594E+05, 4.4762E+05, 4.4931E+05, 4.5100E+05, 4.5270E+05,
     + 4.5441E+05, 4.5611E+05, 4.5783E+05, 4.5955E+05, 4.6127E+05,
     +  4.6300E+05, 4.6473E+05, 4.6647E+05, 4.6821E+05, 4.6996E+05,
     + 4.7172E+05, 4.7347E+05, 4.7524E+05, 4.7701E+05, 4.7878E+05,
     +  4.8056E+05, 4.8235E+05, 4.8414E+05, 4.8593E+05, 4.8774E+05,
     + 4.8954E+05, 4.9135E+05, 4.9317E+05, 4.9499E+05, 4.9682E+05,
     +  4.9865E+05, 5.0049E+05, 5.0233E+05, 5.0418E+05, 5.0603E+05,
     + 5.0789E+05, 5.0976E+05, 5.1163E+05, 5.1350E+05, 5.1538E+05,
     +  5.1727E+05, 5.1916E+05, 5.2106E+05, 5.2296E+05, 5.2487E+05,
     + 5.2678E+05, 5.2870E+05, 5.3062E+05, 5.3255E+05, 5.3449E+05,
     +  5.3643E+05, 5.3838E+05, 5.4033E+05, 5.4229E+05, 5.4425E+05,
     + 5.4622E+05, 5.4819E+05, 5.5017E+05, 5.5216E+05, 5.5415E+05,
     +  5.5615E+05, 5.5815E+05, 5.6016E+05, 5.6217E+05, 5.6419E+05,
     + 5.6622E+05, 5.6825E+05, 5.7028E+05, 5.7233E+05, 5.7437E+05,
     +  5.7643E+05, 5.7849E+05, 5.8055E+05, 5.8263E+05, 5.8470E+05,
     + 5.8679E+05, 5.8887E+05, 5.9097E+05, 5.9307E+05, 5.9518E+05,
     +  5.9729E+05, 5.9941E+05, 6.0153E+05, 6.0366E+05, 6.0580E+05,
     + 6.0794E+05, 6.1009E+05, 6.1224E+05, 6.1440E+05, 6.1657E+05,
     +  6.1874E+05, 6.2091E+05, 6.2310E+05, 6.2529E+05, 6.2748E+05,
     + 6.2969E+05, 6.3189E+05, 6.3411E+05, 6.3633E+05, 6.3856E+05,
     +  6.4079E+05, 6.4303E+05, 6.4527E+05, 6.4752E+05, 6.4978E+05,
     + 6.5204E+05, 6.5431E+05, 6.5659E+05, 6.5887E+05, 6.6116E+05,
     +  6.6345E+05, 6.6575E+05, 6.6806E+05, 6.7037E+05, 6.7269E+05,
     + 6.7502E+05, 6.7735E+05, 6.7969E+05, 6.8203E+05, 6.8439E+05,
     +  6.8674E+05, 6.8911E+05, 6.9148E+05, 6.9386E+05, 6.9624E+05,
     + 6.9863E+05, 7.0102E+05, 7.0343E+05, 7.0584E+05, 7.0825E+05,
     +  7.1067E+05, 7.1310E+05, 7.1554E+05, 7.1798E+05, 7.2043E+05,
     + 7.2288E+05, 7.2534E+05, 7.2781E+05, 7.3029E+05, 7.3277E+05,
     +  7.3525E+05, 7.3775E+05, 7.4025E+05, 7.4276E+05, 7.4527E+05,
     + 7.4779E+05, 7.5032E+05, 7.5286E+05, 7.5540E+05, 7.5794E+05,
     +  7.6050E+05, 7.6306E+05, 7.6563E+05, 7.6820E+05, 7.7079E+05,
     + 7.7337E+05, 7.7597E+05, 7.7857E+05, 7.8118E+05, 7.8380E+05,
     +  7.8642E+05, 7.8905E+05, 7.9169E+05, 7.9433E+05, 7.9698E+05,
     + 7.9964E+05, 8.0231E+05, 8.0498E+05, 8.0766E+05, 8.1034E+05,
     +  8.1303E+05, 8.1573E+05, 8.1844E+05, 8.2115E+05, 8.2388E+05,
     + 8.2660E+05, 8.2934E+05, 8.3208E+05, 8.3483E+05, 8.3759E+05,
     +  8.4035E+05, 8.4312E+05, 8.4590E+05, 8.4868E+05, 8.5148E+05,
     + 8.5428E+05, 8.5708E+05, 8.5990E+05, 8.6272E+05, 8.6555E+05,
     +  8.6838E+05, 8.7123E+05, 8.7408E+05, 8.7693E+05, 8.7980E+05,
     + 8.8267E+05, 8.8555E+05, 8.8844E+05, 8.9133E+05, 8.9424E+05,
     +  8.9715E+05, 9.0006E+05, 9.0299E+05, 9.0592E+05, 9.0886E+05,
     + 9.1180E+05, 9.1476E+05, 9.1772E+05, 9.2069E+05, 9.2367E+05,
     +  9.2665E+05, 9.2964E+05, 9.3264E+05, 9.3565E+05, 9.3867E+05,
     + 9.4169E+05, 9.4472E+05, 9.4776E+05, 9.5080E+05, 9.5386E+05,
     +  9.5692E+05, 9.5999E+05, 9.6306E+05, 9.6615E+05, 9.6924E+05,
     + 9.7234E+05, 9.7545E+05, 9.7856E+05, 9.8168E+05, 9.8482E+05,
     +  9.8796E+05, 9.9110E+05, 9.9426E+05, 9.9742E+05, 1.0006E+06,
     + 1.0038E+06, 1.0070E+06, 1.0101E+06, 1.0134E+06, 1.0166E+06,
     +  1.0198E+06, 1.0230E+06, 1.0262E+06, 1.0295E+06, 1.0327E+06,
     + 1.0360E+06, 1.0393E+06, 1.0425E+06, 1.0458E+06, 1.0491E+06,
     +  1.0524E+06, 1.0557E+06, 1.0590E+06, 1.0623E+06, 1.0657E+06,
     + 1.0690E+06, 1.0724E+06, 1.0757E+06, 1.0791E+06, 1.0825E+06,
     +  1.0858E+06, 1.0892E+06, 1.0926E+06, 1.0960E+06, 1.0995E+06,
     + 1.1029E+06, 1.1063E+06, 1.1097E+06, 1.1132E+06, 1.1166E+06,
     +  1.1201E+06, 1.1236E+06, 1.1271E+06, 1.1306E+06, 1.1341E+06,
     + 1.1376E+06, 1.1411E+06, 1.1446E+06, 1.1481E+06, 1.1517E+06,
     +  1.1552E+06, 1.1588E+06, 1.1624E+06, 1.1659E+06, 1.1695E+06,
     + 1.1731E+06, 1.1767E+06, 1.1803E+06, 1.1839E+06, 1.1876E+06,
     +  1.1912E+06, 1.1948E+06, 1.1985E+06, 1.2022E+06, 1.2058E+06,
     + 1.2095E+06, 1.2132E+06, 1.2169E+06, 1.2206E+06, 1.2243E+06,
     +  1.2281E+06, 1.2318E+06, 1.2355E+06, 1.2393E+06, 1.2430E+06,
     + 1.2468E+06, 1.2506E+06, 1.2544E+06, 1.2582E+06, 1.2620E+06,
     +  1.2658E+06, 1.2696E+06, 1.2734E+06, 1.2773E+06, 1.2811E+06,
     + 1.2850E+06, 1.2889E+06, 1.2928E+06, 1.2966E+06, 1.3005E+06,
     +  1.3044E+06, 1.3084E+06, 1.3123E+06, 1.3162E+06, 1.3202E+06,
     + 1.3241E+06, 1.3281E+06, 1.3320E+06, 1.3360E+06, 1.3400E+06,
     +  1.3440E+06, 1.3480E+06, 1.3520E+06, 1.3561E+06, 1.3601E+06,
     + 1.3641E+06, 1.3682E+06, 1.3723E+06, 1.3763E+06, 1.3804E+06,
     +  1.3845E+06, 1.3886E+06, 1.3927E+06, 1.3969E+06, 1.4010E+06,
     + 1.4051E+06, 1.4093E+06, 1.4134E+06, 1.4176E+06, 1.4218E+06,
     +  1.4260E+06, 1.4302E+06, 1.4344E+06, 1.4386E+06, 1.4428E+06,
     + 1.4471E+06, 1.4513E+06, 1.4556E+06, 1.4598E+06, 1.4641E+06,
     +  1.4684E+06, 1.4727E+06, 1.4770E+06, 1.4813E+06, 1.4857E+06,
     + 1.4900E+06, 1.4943E+06, 1.4987E+06, 1.5031E+06, 1.5074E+06,
     +  1.5118E+06, 1.5162E+06, 1.5206E+06, 1.5250E+06, 1.5295E+06,
     + 1.5339E+06, 1.5384E+06, 1.5428E+06, 1.5473E+06, 1.5518E+06,
     +  1.5562E+06, 1.5607E+06, 1.5653E+06, 1.5698E+06, 1.5743E+06,
     + 1.5788E+06, 1.5834E+06, 1.5879E+06, 1.5925E+06, 1.5971E+06,
     +  1.6017E+06, 1.6063E+06, 1.6109E+06, 1.6155E+06, 1.6201E+06,
     + 1.6248E+06, 1.6294E+06, 1.6341E+06, 1.6388E+06, 1.6435E+06,
     +  1.6482E+06, 1.6529E+06, 1.6576E+06, 1.6623E+06, 1.6670E+06,
     + 1.6718E+06, 1.6766E+06, 1.6813E+06, 1.6861E+06, 1.6909E+06,
     +  1.6957E+06, 1.7005E+06, 1.7053E+06, 1.7102E+06, 1.7150E+06,
     + 1.7199E+06, 1.7247E+06, 1.7296E+06, 1.7345E+06, 1.7394E+06,
     +  1.7443E+06, 1.7492E+06, 1.7541E+06, 1.7591E+06, 1.7640E+06,
     + 1.7690E+06, 1.7740E+06, 1.7790E+06, 1.7840E+06, 1.7890E+06,
     +  1.7940E+06, 1.7990E+06, 1.8041E+06, 1.8091E+06, 1.8142E+06,
     + 1.8192E+06, 1.8243E+06, 1.8294E+06, 1.8345E+06, 1.8397E+06,
     +  1.8448E+06, 1.8499E+06, 1.8551E+06, 1.8602E+06, 1.8654E+06,
     + 1.8706E+06, 1.8758E+06, 1.8810E+06, 1.8862E+06, 1.8915E+06,
     +  1.8967E+06, 1.9020E+06, 1.9072E+06, 1.9125E+06, 1.9178E+06,
     + 1.9231E+06, 1.9284E+06, 1.9337E+06, 1.9391E+06, 1.9444E+06,
     +  1.9498E+06, 1.9551E+06, 1.9605E+06, 1.9659E+06, 1.9713E+06,
     + 1.9767E+06, 1.9822E+06, 1.9876E+06, 1.9931E+06, 1.9985E+06,
     +  2.0040E+06, 2.0095E+06, 2.0150E+06, 2.0205E+06, 2.0260E+06,
     + 2.0315E+06, 2.0371E+06, 2.0426E+06, 2.0482E+06, 2.0538E+06,
     +  2.0594E+06, 2.0650E+06, 2.0706E+06, 2.0762E+06, 2.0819E+06,
     + 2.0875E+06, 2.0932E+06, 2.0989E+06, 2.1046E+06, 2.1103E+06,
     +  2.1160E+06, 2.1217E+06, 2.1275E+06, 2.1332E+06, 2.1390E+06,
     + 2.1447E+06, 2.1505E+06, 2.1563E+06, 2.1621E+06, 2.1680E+06,
     +  2.1738E+06, 2.1797E+06, 2.1855E+06, 2.1914E+06, 2.1973E+06,
     + 2.2032E+06, 2.2091E+06, 2.2150E+06, 2.2209E+06, 2.2269E+06,
     +  2.2329E+06, 2.2388E+06, 2.2448E+06, 2.2508E+06, 2.2568E+06,
     + 2.2629E+06, 2.2689E+06, 2.2749E+06, 2.2810E+06, 2.2871E+06,
     +  2.2932E+06, 2.2993E+06, 2.3054E+06, 2.3115E+06, 2.3176E+06,
     + 2.3238E+06, 2.3300E+06, 2.3361E+06, 2.3423E+06, 2.3485E+06,
     +  2.3547E+06, 2.3610E+06, 2.3672E+06, 2.3735E+06, 2.3797E+06,
     + 2.3860E+06, 2.3923E+06, 2.3986E+06, 2.4049E+06, 2.4113E+06,
     +  2.4176E+06, 2.4240E+06, 2.4303E+06, 2.4367E+06, 2.4431E+06,
     + 2.4495E+06, 2.4560E+06, 2.4624E+06, 2.4688E+06, 2.4753E+06,
     +  2.4818E+06, 2.4883E+06, 2.4948E+06, 2.5013E+06, 2.5078E+06,
     + 2.5144E+06, 2.5209E+06, 2.5275E+06, 2.5341E+06, 2.5407E+06,
     +  2.5473E+06, 2.5539E+06, 2.5606E+06, 2.5672E+06, 2.5739E+06,
     + 2.5806E+06, 2.5873E+06, 2.5940E+06, 2.6007E+06, 2.6074E+06,
     +  2.6142E+06, 2.6209E+06, 2.6277E+06, 2.6345E+06, 2.6413E+06,
     + 2.6481E+06, 2.6549E+06, 2.6618E+06, 2.6687E+06, 2.6755E+06,
     +  2.6824E+06, 2.6893E+06, 2.6962E+06, 2.7032E+06, 2.7101E+06,
     + 2.7171E+06, 2.7240E+06, 2.7310E+06, 2.7380E+06, 2.7450E+06,
     +  2.7520E+06, 2.7591E+06, 2.7661E+06, 2.7732E+06, 2.7803E+06,
     + 2.7874E+06, 2.7945E+06, 2.8016E+06, 2.8088E+06, 2.8159E+06,
     +  2.8231E+06, 2.8303E+06, 2.8375E+06, 2.8447E+06, 2.8519E+06,
     + 2.8591E+06, 2.8664E+06, 2.8737E+06, 2.8810E+06, 2.8883E+06,
     +  2.8956E+06, 2.9029E+06, 2.9102E+06, 2.9176E+06, 2.9250E+06,
     + 2.9324E+06, 2.9398E+06, 2.9472E+06, 2.9546E+06, 2.9620E+06,
     +  2.9695E+06, 2.9770E+06, 2.9845E+06, 2.9920E+06, 2.9995E+06,
     + 3.0070E+06, 3.0146E+06, 3.0221E+06, 3.0297E+06, 3.0373E+06,
     +  3.0449E+06, 3.0525E+06, 3.0602E+06, 3.0678E+06, 3.0755E+06,
     + 3.0832E+06, 3.0909E+06, 3.0986E+06, 3.1063E+06, 3.1141E+06,
     +  3.1218E+06, 3.1296E+06, 3.1374E+06, 3.1452E+06, 3.1530E+06,
     + 3.1608E+06, 3.1687E+06, 3.1766E+06, 3.1844E+06, 3.1923E+06,
     +  3.2002E+06, 3.2082E+06, 3.2161E+06, 3.2241E+06, 3.2321E+06,
     + 3.2400E+06, 3.2480E+06, 3.2561E+06, 3.2641E+06, 3.2722E+06,
     +  3.2802E+06, 3.2883E+06, 3.2964E+06, 3.3045E+06, 3.3126E+06,
     + 3.3208E+06, 3.3289E+06, 3.3371E+06, 3.3453E+06, 3.3535E+06,
     +  3.3618E+06, 3.3700E+06, 3.3782E+06, 3.3865E+06, 3.3948E+06,
     + 3.4031E+06, 3.4114E+06, 3.4198E+06, 3.4281E+06, 3.4365E+06,
     +  3.4449E+06, 3.4533E+06, 3.4617E+06, 3.4701E+06, 3.4786E+06,
     + 3.4870E+06, 3.4955E+06, 3.5040E+06, 3.5125E+06, 3.5210E+06,
     +  3.5296E+06, 3.5382E+06, 3.5467E+06, 3.5553E+06, 3.5639E+06,
     + 3.5726E+06, 3.5812E+06, 3.5899E+06, 3.5985E+06, 3.6072E+06,
     +  3.6159E+06, 3.6247E+06, 3.6334E+06, 3.6422E+06, 3.6509E+06,
     + 3.6597E+06, 3.6685E+06, 3.6774E+06, 3.6862E+06, 3.6951E+06,
     +  3.7040E+06, 3.7128E+06, 3.7218E+06, 3.7307E+06, 3.7396E+06,
     + 3.7486E+06, 3.7576E+06, 3.7666E+06, 3.7756E+06, 3.7846E+06,
     +  3.7936E+06, 3.8027E+06, 3.8118E+06, 3.8209E+06, 3.8300E+06,
     + 3.8391E+06, 3.8483E+06, 3.8574E+06, 3.8666E+06, 3.8758E+06,
     +  3.8850E+06, 3.8942E+06, 3.9035E+06, 3.9128E+06, 3.9220E+06,
     + 3.9314E+06, 3.9407E+06, 3.9500E+06, 3.9594E+06, 3.9687E+06,
     +  3.9781E+06, 3.9875E+06, 3.9970E+06, 4.0064E+06, 4.0158E+06,
     + 4.0253E+06, 4.0348E+06, 4.0443E+06, 4.0539E+06, 4.0634E+06,
     +  4.0730E+06, 4.0826E+06, 4.0922E+06, 4.1018E+06, 4.1114E+06,
     + 4.1211E+06, 4.1307E+06, 4.1404E+06, 4.1501E+06, 4.1598E+06,
     +  4.1696E+06, 4.1793E+06, 4.1891E+06, 4.1989E+06, 4.2087E+06,
     + 4.2186E+06, 4.2284E+06, 4.2383E+06, 4.2482E+06, 4.2581E+06,
     +  4.2680E+06, 4.2779E+06, 4.2879E+06, 4.2979E+06, 4.3079E+06,
     + 4.3179E+06, 4.3279E+06, 4.3380E+06, 4.3480E+06, 4.3581E+06,
     +  4.3682E+06, 4.3784E+06, 4.3885E+06, 4.3987E+06, 4.4088E+06,
     + 4.4190E+06, 4.4292E+06, 4.4395E+06, 4.4497E+06, 4.4600E+06,
     +  4.4703E+06, 4.4806E+06, 4.4909E+06, 4.5013E+06, 4.5116E+06,
     + 4.5220E+06, 4.5324E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       668
      data (QofT( 2,J),J=1,2408)/ 1.6596E+00, 4.3002E+00, 7.6625E+00,
     +  1.1618E+01, 1.6089E+01, 2.1022E+01, 2.6376E+01, 3.2122E+01,
     + 3.8233E+01, 4.4689E+01, 5.1474E+01, 5.8571E+01, 6.5967E+01,
     +  7.3652E+01, 8.1613E+01, 8.9843E+01, 9.8333E+01, 1.0707E+02,
     + 1.1606E+02, 1.2529E+02, 1.4443E+02, 1.6446E+02, 1.8534E+02,
     +  2.0704E+02, 2.2952E+02, 2.5277E+02, 2.7675E+02, 3.0145E+02,
     + 3.2684E+02, 3.5291E+02, 3.7964E+02, 4.0702E+02, 4.3502E+02,
     +  4.6364E+02, 4.9287E+02, 5.2268E+02, 5.5308E+02, 5.8404E+02,
     + 6.1556E+02, 6.4763E+02, 6.8024E+02, 7.1339E+02, 7.4705E+02,
     +  7.8124E+02, 8.1593E+02, 8.5112E+02, 8.8680E+02, 9.2297E+02,
     + 9.5962E+02, 9.9675E+02, 1.0344E+03, 1.0724E+03, 1.1109E+03,
     +  1.1499E+03, 1.1893E+03, 1.2292E+03, 1.2695E+03, 1.3102E+03,
     + 1.3514E+03, 1.3930E+03, 1.4350E+03, 1.4775E+03, 1.5203E+03,
     +  1.5636E+03, 1.6073E+03, 1.6514E+03, 1.6959E+03, 1.7408E+03,
     + 1.7861E+03, 1.8318E+03, 1.8779E+03, 1.9244E+03, 1.9712E+03,
     +  2.0185E+03, 2.0661E+03, 2.1142E+03, 2.1626E+03, 2.2114E+03,
     + 2.2606E+03, 2.3102E+03, 2.3602E+03, 2.4105E+03, 2.4613E+03,
     +  2.5124E+03, 2.5639E+03, 2.6157E+03, 2.6680E+03, 2.7206E+03,
     + 2.7736E+03, 2.8271E+03, 2.8808E+03, 2.9350E+03, 2.9896E+03,
     +  3.0445E+03, 3.0998E+03, 3.1556E+03, 3.2116E+03, 3.2681E+03,
     + 3.3250E+03, 3.3823E+03, 3.4400E+03, 3.4981E+03, 3.5565E+03,
     +  3.6154E+03, 3.6747E+03, 3.7344E+03, 3.7944E+03, 3.8549E+03,
     + 3.9158E+03, 3.9771E+03, 4.0389E+03, 4.1010E+03, 4.1636E+03,
     +  4.2266E+03, 4.2900E+03, 4.3538E+03, 4.4181E+03, 4.4828E+03,
     + 4.5479E+03, 4.6135E+03, 4.6795E+03, 4.7460E+03, 4.8129E+03,
     +  4.8803E+03, 4.9481E+03, 5.0164E+03, 5.0852E+03, 5.1544E+03,
     + 5.2240E+03, 5.2942E+03, 5.3648E+03, 5.4359E+03, 5.5075E+03,
     +  5.5796E+03, 5.6522E+03, 5.7252E+03, 5.7988E+03, 5.8728E+03,
     + 5.9474E+03, 6.0225E+03, 6.0981E+03, 6.1742E+03, 6.2509E+03,
     +  6.3280E+03, 6.4057E+03, 6.4839E+03, 6.5627E+03, 6.6419E+03,
     + 6.7218E+03, 6.8022E+03, 6.8831E+03, 6.9647E+03, 7.0467E+03,
     +  7.1294E+03, 7.2126E+03, 7.2963E+03, 7.3807E+03, 7.4657E+03,
     + 7.5512E+03, 7.6374E+03, 7.7241E+03, 7.8114E+03, 7.8994E+03,
     +  7.9879E+03, 8.0771E+03, 8.1668E+03, 8.2572E+03, 8.3483E+03,
     + 8.4400E+03, 8.5323E+03, 8.6252E+03, 8.7189E+03, 8.8132E+03,
     +  8.9081E+03, 9.0036E+03, 9.0998E+03, 9.1968E+03, 9.2944E+03,
     + 9.3926E+03, 9.4916E+03, 9.5913E+03, 9.6916E+03, 9.7927E+03,
     +  9.8945E+03, 9.9969E+03, 1.0100E+04, 1.0204E+04, 1.0309E+04,
     + 1.0414E+04, 1.0520E+04, 1.0627E+04, 1.0735E+04, 1.0843E+04,
     +  1.0952E+04, 1.1062E+04, 1.1172E+04, 1.1284E+04, 1.1396E+04,
     + 1.1509E+04, 1.1623E+04, 1.1737E+04, 1.1853E+04, 1.1969E+04,
     +  1.2086E+04, 1.2203E+04, 1.2322E+04, 1.2441E+04, 1.2562E+04,
     + 1.2682E+04, 1.2804E+04, 1.2927E+04, 1.3051E+04, 1.3175E+04,
     +  1.3300E+04, 1.3427E+04, 1.3554E+04, 1.3681E+04, 1.3810E+04,
     + 1.3940E+04, 1.4071E+04, 1.4202E+04, 1.4334E+04, 1.4468E+04,
     +  1.4602E+04, 1.4737E+04, 1.4873E+04, 1.5010E+04, 1.5148E+04,
     + 1.5287E+04, 1.5426E+04, 1.5567E+04, 1.5709E+04, 1.5851E+04,
     +  1.5995E+04, 1.6140E+04, 1.6285E+04, 1.6432E+04, 1.6579E+04,
     + 1.6728E+04, 1.6878E+04, 1.7028E+04, 1.7180E+04, 1.7332E+04,
     +  1.7486E+04, 1.7641E+04, 1.7796E+04, 1.7953E+04, 1.8111E+04,
     + 1.8270E+04, 1.8430E+04, 1.8591E+04, 1.8753E+04, 1.8916E+04,
     +  1.9080E+04, 1.9245E+04, 1.9412E+04, 1.9579E+04, 1.9748E+04,
     + 1.9918E+04, 2.0089E+04, 2.0260E+04, 2.0434E+04, 2.0608E+04,
     +  2.0783E+04, 2.0960E+04, 2.1138E+04, 2.1316E+04, 2.1497E+04,
     + 2.1678E+04, 2.1860E+04, 2.2044E+04, 2.2229E+04, 2.2415E+04,
     +  2.2602E+04, 2.2790E+04, 2.2980E+04, 2.3171E+04, 2.3363E+04,
     + 2.3556E+04, 2.3751E+04, 2.3947E+04, 2.4144E+04, 2.4342E+04,
     +  2.4542E+04, 2.4743E+04, 2.4945E+04, 2.5148E+04, 2.5353E+04,
     + 2.5559E+04, 2.5767E+04, 2.5975E+04, 2.6185E+04, 2.6397E+04,
     +  2.6609E+04, 2.6823E+04, 2.7039E+04, 2.7256E+04, 2.7474E+04,
     + 2.7693E+04, 2.7914E+04, 2.8136E+04, 2.8360E+04, 2.8585E+04,
     +  2.8811E+04, 2.9039E+04, 2.9268E+04, 2.9499E+04, 2.9731E+04,
     + 2.9965E+04, 3.0200E+04, 3.0436E+04, 3.0674E+04, 3.0914E+04,
     +  3.1154E+04, 3.1397E+04, 3.1641E+04, 3.1886E+04, 3.2133E+04,
     + 3.2381E+04, 3.2631E+04, 3.2882E+04, 3.3135E+04, 3.3389E+04,
     +  3.3645E+04, 3.3903E+04, 3.4162E+04, 3.4423E+04, 3.4685E+04,
     + 3.4949E+04, 3.5214E+04, 3.5481E+04, 3.5750E+04, 3.6020E+04,
     +  3.6291E+04, 3.6565E+04, 3.6840E+04, 3.7117E+04, 3.7395E+04,
     + 3.7675E+04, 3.7957E+04, 3.8240E+04, 3.8525E+04, 3.8812E+04,
     +  3.9100E+04, 3.9390E+04, 3.9682E+04, 3.9975E+04, 4.0270E+04,
     + 4.0567E+04, 4.0866E+04, 4.1166E+04, 4.1469E+04, 4.1772E+04,
     +  4.2078E+04, 4.2386E+04, 4.2695E+04, 4.3006E+04, 4.3319E+04,
     + 4.3633E+04, 4.3950E+04, 4.4268E+04, 4.4588E+04, 4.4910E+04,
     +  4.5234E+04, 4.5559E+04, 4.5887E+04, 4.6216E+04, 4.6548E+04,
     + 4.6881E+04, 4.7216E+04, 4.7553E+04, 4.7892E+04, 4.8232E+04,
     +  4.8575E+04, 4.8920E+04, 4.9266E+04, 4.9615E+04, 4.9965E+04,
     + 5.0318E+04, 5.0672E+04, 5.1028E+04, 5.1387E+04, 5.1747E+04,
     +  5.2110E+04, 5.2474E+04, 5.2841E+04, 5.3209E+04, 5.3579E+04,
     + 5.3952E+04, 5.4327E+04, 5.4703E+04, 5.5082E+04, 5.5463E+04,
     +  5.5846E+04, 5.6231E+04, 5.6618E+04, 5.7007E+04, 5.7399E+04,
     + 5.7792E+04, 5.8188E+04, 5.8586E+04, 5.8986E+04, 5.9388E+04,
     +  5.9792E+04, 6.0199E+04, 6.0608E+04, 6.1019E+04, 6.1432E+04,
     + 6.1847E+04, 6.2265E+04, 6.2685E+04, 6.3107E+04, 6.3531E+04,
     +  6.3958E+04, 6.4387E+04, 6.4818E+04, 6.5251E+04, 6.5687E+04,
     + 6.6126E+04, 6.6566E+04, 6.7009E+04, 6.7454E+04, 6.7901E+04,
     +  6.8351E+04, 6.8803E+04, 6.9258E+04, 6.9715E+04, 7.0174E+04,
     + 7.0636E+04, 7.1100E+04, 7.1567E+04, 7.2036E+04, 7.2507E+04,
     +  7.2981E+04, 7.3458E+04, 7.3937E+04, 7.4418E+04, 7.4902E+04,
     + 7.5388E+04, 7.5877E+04, 7.6368E+04, 7.6862E+04, 7.7359E+04,
     +  7.7858E+04, 7.8359E+04, 7.8863E+04, 7.9370E+04, 7.9879E+04,
     + 8.0391E+04, 8.0906E+04, 8.1423E+04, 8.1943E+04, 8.2465E+04,
     +  8.2990E+04, 8.3518E+04, 8.4048E+04, 8.4581E+04, 8.5117E+04,
     + 8.5655E+04, 8.6196E+04, 8.6740E+04, 8.7287E+04, 8.7836E+04,
     +  8.8388E+04, 8.8943E+04, 8.9500E+04, 9.0061E+04, 9.0624E+04,
     + 9.1190E+04, 9.1759E+04, 9.2330E+04, 9.2905E+04, 9.3482E+04,
     +  9.4062E+04, 9.4645E+04, 9.5231E+04, 9.5819E+04, 9.6411E+04,
     + 9.7005E+04, 9.7603E+04, 9.8203E+04, 9.8806E+04, 9.9413E+04,
     +  1.0002E+05, 1.0063E+05, 1.0125E+05, 1.0187E+05, 1.0249E+05,
     + 1.0311E+05, 1.0374E+05, 1.0437E+05, 1.0500E+05, 1.0564E+05,
     +  1.0628E+05, 1.0692E+05, 1.0757E+05, 1.0821E+05, 1.0887E+05,
     + 1.0952E+05, 1.1018E+05, 1.1084E+05, 1.1150E+05, 1.1217E+05,
     +  1.1284E+05, 1.1351E+05, 1.1419E+05, 1.1487E+05, 1.1555E+05,
     + 1.1624E+05, 1.1693E+05, 1.1762E+05, 1.1832E+05, 1.1902E+05,
     +  1.1972E+05, 1.2043E+05, 1.2114E+05, 1.2185E+05, 1.2257E+05,
     + 1.2329E+05, 1.2401E+05, 1.2473E+05, 1.2546E+05, 1.2620E+05,
     +  1.2693E+05, 1.2767E+05, 1.2841E+05, 1.2916E+05, 1.2991E+05,
     + 1.3066E+05, 1.3142E+05, 1.3218E+05, 1.3295E+05, 1.3371E+05,
     +  1.3448E+05, 1.3526E+05, 1.3604E+05, 1.3682E+05, 1.3760E+05,
     + 1.3839E+05, 1.3918E+05, 1.3998E+05, 1.4078E+05, 1.4158E+05,
     +  1.4239E+05, 1.4320E+05, 1.4401E+05, 1.4483E+05, 1.4565E+05,
     + 1.4647E+05, 1.4730E+05, 1.4813E+05, 1.4897E+05, 1.4981E+05,
     +  1.5065E+05, 1.5150E+05, 1.5235E+05, 1.5321E+05, 1.5407E+05,
     + 1.5493E+05, 1.5579E+05, 1.5666E+05, 1.5754E+05, 1.5842E+05,
     +  1.5930E+05, 1.6018E+05, 1.6107E+05, 1.6197E+05, 1.6286E+05,
     + 1.6377E+05, 1.6467E+05, 1.6558E+05, 1.6649E+05, 1.6741E+05,
     +  1.6833E+05, 1.6926E+05, 1.7019E+05, 1.7112E+05, 1.7206E+05,
     + 1.7300E+05, 1.7394E+05, 1.7489E+05, 1.7585E+05, 1.7680E+05,
     +  1.7777E+05, 1.7873E+05, 1.7970E+05, 1.8068E+05, 1.8165E+05,
     + 1.8264E+05, 1.8362E+05, 1.8461E+05, 1.8561E+05, 1.8661E+05,
     +  1.8761E+05, 1.8862E+05, 1.8963E+05, 1.9065E+05, 1.9167E+05,
     + 1.9270E+05, 1.9373E+05, 1.9476E+05, 1.9580E+05, 1.9684E+05,
     +  1.9789E+05, 1.9894E+05, 2.0000E+05, 2.0106E+05, 2.0212E+05,
     + 2.0319E+05, 2.0427E+05, 2.0534E+05, 2.0643E+05, 2.0751E+05,
     +  2.0861E+05, 2.0970E+05, 2.1080E+05, 2.1191E+05, 2.1302E+05,
     + 2.1413E+05, 2.1525E+05, 2.1638E+05, 2.1751E+05, 2.1864E+05,
     +  2.1978E+05, 2.2092E+05, 2.2207E+05, 2.2322E+05, 2.2438E+05,
     + 2.2554E+05, 2.2670E+05, 2.2788E+05, 2.2905E+05, 2.3023E+05,
     +  2.3142E+05, 2.3261E+05, 2.3380E+05, 2.3500E+05, 2.3621E+05,
     + 2.3742E+05, 2.3863E+05, 2.3985E+05, 2.4108E+05, 2.4231E+05,
     +  2.4354E+05, 2.4478E+05, 2.4602E+05, 2.4727E+05, 2.4853E+05,
     + 2.4979E+05, 2.5105E+05, 2.5232E+05, 2.5359E+05, 2.5487E+05,
     +  2.5616E+05, 2.5745E+05, 2.5874E+05, 2.6004E+05, 2.6135E+05,
     + 2.6266E+05, 2.6398E+05, 2.6530E+05, 2.6662E+05, 2.6795E+05,
     +  2.6929E+05, 2.7063E+05, 2.7198E+05, 2.7333E+05, 2.7469E+05,
     + 2.7605E+05, 2.7742E+05, 2.7880E+05, 2.8017E+05, 2.8156E+05,
     +  2.8295E+05, 2.8434E+05, 2.8574E+05, 2.8715E+05, 2.8856E+05,
     + 2.8998E+05, 2.9140E+05, 2.9283E+05, 2.9426E+05, 2.9570E+05,
     +  2.9715E+05, 2.9860E+05, 3.0006E+05, 3.0152E+05, 3.0298E+05,
     + 3.0446E+05, 3.0594E+05, 3.0742E+05, 3.0891E+05, 3.1041E+05,
     +  3.1191E+05, 3.1341E+05, 3.1493E+05, 3.1645E+05, 3.1797E+05,
     + 3.1950E+05, 3.2104E+05, 3.2258E+05, 3.2412E+05, 3.2568E+05,
     +  3.2724E+05, 3.2880E+05, 3.3037E+05, 3.3195E+05, 3.3353E+05,
     + 3.3512E+05, 3.3672E+05, 3.3832E+05, 3.3993E+05, 3.4154E+05,
     +  3.4316E+05, 3.4478E+05, 3.4641E+05, 3.4805E+05, 3.4970E+05,
     + 3.5134E+05, 3.5300E+05, 3.5466E+05, 3.5633E+05, 3.5800E+05,
     +  3.5968E+05, 3.6137E+05, 3.6306E+05, 3.6476E+05, 3.6647E+05,
     + 3.6818E+05, 3.6990E+05, 3.7162E+05, 3.7335E+05, 3.7509E+05,
     +  3.7683E+05, 3.7858E+05, 3.8034E+05, 3.8210E+05, 3.8387E+05,
     + 3.8565E+05, 3.8743E+05, 3.8922E+05, 3.9101E+05, 3.9281E+05,
     +  3.9462E+05, 3.9644E+05, 3.9826E+05, 4.0008E+05, 4.0192E+05,
     + 4.0376E+05, 4.0561E+05, 4.0746E+05, 4.0932E+05, 4.1119E+05,
     +  4.1307E+05, 4.1495E+05, 4.1684E+05, 4.1873E+05, 4.2063E+05,
     + 4.2254E+05, 4.2445E+05, 4.2638E+05, 4.2830E+05, 4.3024E+05,
     +  4.3218E+05, 4.3413E+05, 4.3609E+05, 4.3805E+05, 4.4002E+05,
     + 4.4200E+05, 4.4399E+05, 4.4598E+05, 4.4798E+05, 4.4998E+05,
     +  4.5199E+05, 4.5401E+05, 4.5604E+05, 4.5807E+05, 4.6012E+05,
     + 4.6216E+05, 4.6422E+05, 4.6628E+05, 4.6835E+05, 4.7043E+05,
     +  4.7251E+05, 4.7461E+05, 4.7670E+05, 4.7881E+05, 4.8092E+05,
     + 4.8305E+05, 4.8517E+05, 4.8731E+05, 4.8945E+05, 4.9160E+05,
     +  4.9376E+05, 4.9593E+05, 4.9810E+05, 5.0028E+05, 5.0247E+05,
     + 5.0466E+05, 5.0687E+05, 5.0908E+05, 5.1130E+05, 5.1352E+05,
     +  5.1576E+05, 5.1800E+05, 5.2025E+05, 5.2250E+05, 5.2477E+05,
     + 5.2704E+05, 5.2932E+05, 5.3161E+05, 5.3390E+05, 5.3621E+05,
     +  5.3852E+05, 5.4084E+05, 5.4316E+05, 5.4550E+05, 5.4784E+05,
     + 5.5019E+05, 5.5255E+05, 5.5492E+05, 5.5729E+05, 5.5967E+05,
     +  5.6206E+05, 5.6446E+05, 5.6687E+05, 5.6928E+05, 5.7170E+05,
     + 5.7414E+05, 5.7657E+05, 5.7902E+05, 5.8148E+05, 5.8394E+05,
     +  5.8641E+05, 5.8889E+05, 5.9138E+05, 5.9388E+05, 5.9638E+05,
     + 5.9889E+05, 6.0141E+05, 6.0395E+05, 6.0648E+05, 6.0903E+05,
     +  6.1158E+05, 6.1415E+05, 6.1672E+05, 6.1930E+05, 6.2189E+05,
     + 6.2449E+05, 6.2709E+05, 6.2971E+05, 6.3233E+05, 6.3496E+05,
     +  6.3760E+05, 6.4025E+05, 6.4291E+05, 6.4558E+05, 6.4825E+05,
     + 6.5094E+05, 6.5363E+05, 6.5633E+05, 6.5904E+05, 6.6176E+05,
     +  6.6449E+05, 6.6722E+05, 6.6997E+05, 6.7272E+05, 6.7549E+05,
     + 6.7826E+05, 6.8104E+05, 6.8383E+05, 6.8663E+05, 6.8944E+05,
     +  6.9225E+05, 6.9508E+05, 6.9792E+05, 7.0076E+05, 7.0361E+05,
     + 7.0648E+05, 7.0935E+05, 7.1223E+05, 7.1512E+05, 7.1802E+05,
     +  7.2093E+05, 7.2385E+05, 7.2677E+05, 7.2971E+05, 7.3266E+05,
     + 7.3561E+05, 7.3858E+05, 7.4155E+05, 7.4454E+05, 7.4753E+05,
     +  7.5053E+05, 7.5354E+05, 7.5656E+05, 7.5960E+05, 7.6264E+05,
     + 7.6569E+05, 7.6875E+05, 7.7182E+05, 7.7490E+05, 7.7798E+05,
     +  7.8108E+05, 7.8419E+05, 7.8731E+05, 7.9044E+05, 7.9357E+05,
     + 7.9672E+05, 7.9988E+05, 8.0305E+05, 8.0622E+05, 8.0941E+05,
     +  8.1261E+05, 8.1581E+05, 8.1903E+05, 8.2226E+05, 8.2549E+05,
     + 8.2874E+05, 8.3200E+05, 8.3526E+05, 8.3854E+05, 8.4183E+05,
     +  8.4512E+05, 8.4843E+05, 8.5175E+05, 8.5508E+05, 8.5841E+05,
     + 8.6176E+05, 8.6512E+05, 8.6849E+05, 8.7187E+05, 8.7526E+05,
     +  8.7866E+05, 8.8207E+05, 8.8549E+05, 8.8892E+05, 8.9236E+05,
     + 8.9581E+05, 8.9927E+05, 9.0275E+05, 9.0623E+05, 9.0972E+05,
     +  9.1323E+05, 9.1674E+05, 9.2027E+05, 9.2381E+05, 9.2735E+05,
     + 9.3091E+05, 9.3448E+05, 9.3806E+05, 9.4165E+05, 9.4525E+05,
     +  9.4886E+05, 9.5249E+05, 9.5612E+05, 9.5977E+05, 9.6342E+05,
     + 9.6709E+05, 9.7076E+05, 9.7445E+05, 9.7815E+05, 9.8186E+05,
     +  9.8558E+05, 9.8932E+05, 9.9306E+05, 9.9682E+05, 1.0006E+06,
     + 1.0044E+06, 1.0081E+06, 1.0119E+06, 1.0158E+06, 1.0196E+06,
     +  1.0234E+06, 1.0273E+06, 1.0311E+06, 1.0350E+06, 1.0389E+06,
     + 1.0428E+06, 1.0467E+06, 1.0506E+06, 1.0545E+06, 1.0584E+06,
     +  1.0624E+06, 1.0663E+06, 1.0703E+06, 1.0743E+06, 1.0783E+06,
     + 1.0823E+06, 1.0863E+06, 1.0903E+06, 1.0944E+06, 1.0984E+06,
     +  1.1025E+06, 1.1066E+06, 1.1107E+06, 1.1148E+06, 1.1189E+06,
     + 1.1230E+06, 1.1271E+06, 1.1313E+06, 1.1355E+06, 1.1396E+06,
     +  1.1438E+06, 1.1480E+06, 1.1522E+06, 1.1564E+06, 1.1607E+06,
     + 1.1649E+06, 1.1692E+06, 1.1734E+06, 1.1777E+06, 1.1820E+06,
     +  1.1863E+06, 1.1906E+06, 1.1950E+06, 1.1993E+06, 1.2037E+06,
     + 1.2080E+06, 1.2124E+06, 1.2168E+06, 1.2212E+06, 1.2256E+06,
     +  1.2301E+06, 1.2345E+06, 1.2390E+06, 1.2434E+06, 1.2479E+06,
     + 1.2524E+06, 1.2569E+06, 1.2614E+06, 1.2660E+06, 1.2705E+06,
     +  1.2751E+06, 1.2797E+06, 1.2842E+06, 1.2888E+06, 1.2934E+06,
     + 1.2981E+06, 1.3027E+06, 1.3074E+06, 1.3120E+06, 1.3167E+06,
     +  1.3214E+06, 1.3261E+06, 1.3308E+06, 1.3355E+06, 1.3403E+06,
     + 1.3450E+06, 1.3498E+06, 1.3546E+06, 1.3594E+06, 1.3642E+06,
     +  1.3690E+06, 1.3738E+06, 1.3787E+06, 1.3835E+06, 1.3884E+06,
     + 1.3933E+06, 1.3982E+06, 1.4031E+06, 1.4080E+06, 1.4130E+06,
     +  1.4179E+06, 1.4229E+06, 1.4279E+06, 1.4329E+06, 1.4379E+06,
     + 1.4429E+06, 1.4480E+06, 1.4530E+06, 1.4581E+06, 1.4632E+06,
     +  1.4682E+06, 1.4734E+06, 1.4785E+06, 1.4836E+06, 1.4888E+06,
     + 1.4939E+06, 1.4991E+06, 1.5043E+06, 1.5095E+06, 1.5147E+06,
     +  1.5200E+06, 1.5252E+06, 1.5305E+06, 1.5357E+06, 1.5410E+06,
     + 1.5463E+06, 1.5517E+06, 1.5570E+06, 1.5623E+06, 1.5677E+06,
     +  1.5731E+06, 1.5785E+06, 1.5839E+06, 1.5893E+06, 1.5947E+06,
     + 1.6002E+06, 1.6057E+06, 1.6111E+06, 1.6166E+06, 1.6221E+06,
     +  1.6277E+06, 1.6332E+06, 1.6388E+06, 1.6443E+06, 1.6499E+06,
     + 1.6555E+06, 1.6611E+06, 1.6668E+06, 1.6724E+06, 1.6781E+06,
     +  1.6837E+06, 1.6894E+06, 1.6951E+06, 1.7008E+06, 1.7066E+06,
     + 1.7123E+06, 1.7181E+06, 1.7239E+06, 1.7297E+06, 1.7355E+06,
     +  1.7413E+06, 1.7471E+06, 1.7530E+06, 1.7589E+06, 1.7647E+06,
     + 1.7707E+06, 1.7766E+06, 1.7825E+06, 1.7885E+06, 1.7944E+06,
     +  1.8004E+06, 1.8064E+06, 1.8124E+06, 1.8184E+06, 1.8245E+06,
     + 1.8305E+06, 1.8366E+06, 1.8427E+06, 1.8488E+06, 1.8549E+06,
     +  1.8611E+06, 1.8672E+06, 1.8734E+06, 1.8796E+06, 1.8858E+06,
     + 1.8920E+06, 1.8982E+06, 1.9045E+06, 1.9108E+06, 1.9170E+06,
     +  1.9233E+06, 1.9297E+06, 1.9360E+06, 1.9423E+06, 1.9487E+06,
     + 1.9551E+06, 1.9615E+06, 1.9679E+06, 1.9743E+06, 1.9808E+06,
     +  1.9872E+06, 1.9937E+06, 2.0002E+06, 2.0067E+06, 2.0133E+06,
     + 2.0198E+06, 2.0264E+06, 2.0329E+06, 2.0395E+06, 2.0462E+06,
     +  2.0528E+06, 2.0594E+06, 2.0661E+06, 2.0728E+06, 2.0795E+06,
     + 2.0862E+06, 2.0929E+06, 2.0997E+06, 2.1064E+06, 2.1132E+06,
     +  2.1200E+06, 2.1268E+06, 2.1337E+06, 2.1405E+06, 2.1474E+06,
     + 2.1543E+06, 2.1612E+06, 2.1681E+06, 2.1751E+06, 2.1820E+06,
     +  2.1890E+06, 2.1960E+06, 2.2030E+06, 2.2100E+06, 2.2171E+06,
     + 2.2241E+06, 2.2312E+06, 2.2383E+06, 2.2454E+06, 2.2525E+06,
     +  2.2597E+06, 2.2669E+06, 2.2741E+06, 2.2813E+06, 2.2885E+06,
     + 2.2957E+06, 2.3030E+06, 2.3103E+06, 2.3176E+06, 2.3249E+06,
     +  2.3322E+06, 2.3395E+06, 2.3469E+06, 2.3543E+06, 2.3617E+06,
     + 2.3691E+06, 2.3766E+06, 2.3840E+06, 2.3915E+06, 2.3990E+06,
     +  2.4065E+06, 2.4140E+06, 2.4216E+06, 2.4292E+06, 2.4368E+06,
     + 2.4444E+06, 2.4520E+06, 2.4596E+06, 2.4673E+06, 2.4750E+06,
     +  2.4827E+06, 2.4904E+06, 2.4981E+06, 2.5059E+06, 2.5137E+06,
     + 2.5215E+06, 2.5293E+06, 2.5371E+06, 2.5450E+06, 2.5528E+06,
     +  2.5607E+06, 2.5686E+06, 2.5766E+06, 2.5845E+06, 2.5925E+06,
     + 2.6005E+06, 2.6085E+06, 2.6165E+06, 2.6245E+06, 2.6326E+06,
     +  2.6407E+06, 2.6488E+06, 2.6569E+06, 2.6651E+06, 2.6732E+06,
     + 2.6814E+06, 2.6896E+06, 2.6978E+06, 2.7061E+06, 2.7143E+06,
     +  2.7226E+06, 2.7309E+06, 2.7392E+06, 2.7476E+06, 2.7559E+06,
     + 2.7643E+06, 2.7727E+06, 2.7811E+06, 2.7896E+06, 2.7980E+06,
     +  2.8065E+06, 2.8150E+06, 2.8235E+06, 2.8321E+06, 2.8406E+06,
     + 2.8492E+06, 2.8578E+06, 2.8664E+06, 2.8751E+06, 2.8837E+06,
     +  2.8924E+06, 2.9011E+06, 2.9098E+06, 2.9186E+06, 2.9274E+06,
     + 2.9361E+06, 2.9449E+06, 2.9538E+06, 2.9626E+06, 2.9715E+06,
     +  2.9804E+06, 2.9893E+06, 2.9982E+06, 3.0072E+06, 3.0162E+06,
     + 3.0252E+06, 3.0342E+06, 3.0432E+06, 3.0523E+06, 3.0613E+06,
     +  3.0704E+06, 3.0796E+06, 3.0887E+06, 3.0979E+06, 3.1071E+06,
     + 3.1163E+06, 3.1255E+06, 3.1348E+06, 3.1440E+06, 3.1533E+06,
     +  3.1626E+06, 3.1720E+06, 3.1813E+06, 3.1907E+06, 3.2001E+06,
     + 3.2095E+06, 3.2190E+06, 3.2285E+06, 3.2379E+06, 3.2475E+06,
     +  3.2570E+06, 3.2665E+06, 3.2761E+06, 3.2857E+06, 3.2953E+06,
     + 3.3050E+06, 3.3147E+06, 3.3243E+06, 3.3341E+06, 3.3438E+06,
     +  3.3535E+06, 3.3633E+06, 3.3731E+06, 3.3829E+06, 3.3928E+06,
     + 3.4027E+06, 3.4125E+06, 3.4225E+06, 3.4324E+06, 3.4423E+06,
     +  3.4523E+06, 3.4623E+06, 3.4724E+06, 3.4824E+06, 3.4925E+06,
     + 3.5026E+06, 3.5127E+06, 3.5228E+06, 3.5330E+06, 3.5432E+06,
     +  3.5534E+06, 3.5636E+06, 3.5739E+06, 3.5842E+06, 3.5945E+06,
     + 3.6048E+06, 3.6151E+06, 3.6255E+06, 3.6359E+06, 3.6463E+06,
     +  3.6568E+06, 3.6672E+06, 3.6777E+06, 3.6882E+06, 3.6988E+06,
     + 3.7093E+06, 3.7199E+06, 3.7305E+06, 3.7411E+06, 3.7518E+06,
     +  3.7625E+06, 3.7732E+06, 3.7839E+06, 3.7947E+06, 3.8054E+06,
     + 3.8162E+06, 3.8271E+06, 3.8379E+06, 3.8488E+06, 3.8597E+06,
     +  3.8706E+06, 3.8815E+06, 3.8925E+06, 3.9035E+06, 3.9145E+06,
     + 3.9256E+06, 3.9366E+06, 3.9477E+06, 3.9588E+06, 3.9700E+06,
     +  3.9811E+06, 3.9923E+06, 4.0035E+06, 4.0148E+06, 4.0260E+06,
     + 4.0373E+06, 4.0486E+06, 4.0600E+06, 4.0713E+06, 4.0827E+06,
     +  4.0941E+06, 4.1056E+06, 4.1170E+06, 4.1285E+06, 4.1400E+06,
     + 4.1516E+06, 4.1631E+06, 4.1747E+06, 4.1863E+06, 4.1980E+06,
     +  4.2096E+06, 4.2213E+06, 4.2330E+06, 4.2448E+06, 4.2566E+06,
     + 4.2683E+06, 4.2802E+06, 4.2920E+06, 4.3039E+06, 4.3158E+06,
     +  4.3277E+06, 4.3396E+06, 4.3516E+06, 4.3636E+06, 4.3756E+06,
     + 4.3877E+06, 4.3998E+06, 4.4119E+06, 4.4240E+06, 4.4362E+06,
     +  4.4483E+06, 4.4605E+06, 4.4728E+06, 4.4850E+06, 4.4973E+06,
     + 4.5096E+06, 4.5220E+06, 4.5344E+06, 4.5467E+06, 4.5592E+06,
     +  4.5716E+06, 4.5841E+06, 4.5966E+06, 4.6091E+06, 4.6217E+06,
     + 4.6342E+06, 4.6469E+06, 4.6595E+06, 4.6722E+06, 4.6848E+06,
     +  4.6976E+06, 4.7103E+06, 4.7231E+06, 4.7359E+06, 4.7487E+06,
     + 4.7615E+06, 4.7744E+06, 4.7873E+06, 4.8003E+06, 4.8132E+06,
     +  4.8262E+06, 4.8392E+06, 4.8523E+06, 4.8653E+06, 4.8784E+06,
     + 4.8916E+06, 4.9047E+06, 4.9179E+06, 4.9311E+06, 4.9444E+06,
     +  4.9576E+06, 4.9709E+06, 4.9842E+06, 4.9976E+06, 5.0110E+06,
     + 5.0244E+06, 5.0378E+06, 5.0513E+06, 5.0648E+06, 5.0783E+06,
     +  5.0918E+06, 5.1054E+06, 5.1190E+06, 5.1326E+06, 5.1463E+06,
     + 5.1600E+06, 5.1737E+06, 5.1875E+06, 5.2012E+06, 5.2150E+06,
     +  5.2289E+06, 5.2427E+06, 5.2566E+06, 5.2705E+06, 5.2845E+06,
     + 5.2985E+06, 5.3125E+06, 5.3265E+06, 5.3406E+06, 5.3547E+06,
     +  5.3688E+06, 5.3829E+06, 5.3971E+06, 5.4113E+06, 5.4256E+06,
     + 5.4399E+06, 5.4542E+06, 5.4685E+06, 5.4828E+06, 5.4972E+06,
     +  5.5116E+06, 5.5261E+06, 5.5406E+06, 5.5551E+06, 5.5696E+06,
     + 5.5842E+06, 5.5988E+06, 5.6134E+06, 5.6281E+06, 5.6428E+06,
     +  5.6575E+06, 5.6722E+06, 5.6870E+06, 5.7018E+06, 5.7166E+06,
     + 5.7315E+06, 5.7464E+06, 5.7613E+06, 5.7763E+06, 5.7913E+06,
     +  5.8063E+06, 5.8214E+06, 5.8365E+06, 5.8516E+06, 5.8667E+06,
     + 5.8819E+06, 5.8971E+06, 5.9123E+06, 5.9276E+06, 5.9429E+06,
     +  5.9582E+06, 5.9736E+06, 5.9890E+06, 6.0044E+06, 6.0198E+06,
     + 6.0353E+06, 6.0509E+06, 6.0664E+06, 6.0820E+06, 6.0976E+06,
     +  6.1132E+06, 6.1289E+06, 6.1446E+06, 6.1603E+06, 6.1761E+06,
     + 6.1919E+06, 6.2077E+06, 6.2236E+06, 6.2395E+06, 6.2554E+06,
     +  6.2714E+06, 6.2874E+06, 6.3034E+06, 6.3195E+06, 6.3356E+06,
     + 6.3517E+06, 6.3678E+06, 6.3840E+06, 6.4002E+06, 6.4165E+06,
     +  6.4328E+06, 6.4491E+06, 6.4654E+06, 6.4818E+06, 6.4982E+06,
     + 6.5147E+06, 6.5311E+06, 6.5476E+06, 6.5642E+06, 6.5808E+06,
     +  6.5974E+06, 6.6140E+06, 6.6307E+06, 6.6474E+06, 6.6641E+06,
     + 6.6809E+06, 6.6977E+06, 6.7146E+06, 6.7314E+06, 6.7483E+06,
     +  6.7653E+06, 6.7822E+06, 6.7993E+06, 6.8163E+06, 6.8334E+06,
     + 6.8505E+06, 6.8676E+06, 6.8848E+06, 6.9020E+06, 6.9192E+06,
     +  6.9365E+06, 6.9538E+06, 6.9712E+06, 6.9885E+06, 7.0060E+06,
     + 7.0234E+06, 7.0409E+06, 7.0584E+06, 7.0760E+06, 7.0935E+06,
     +  7.1112E+06, 7.1288E+06, 7.1465E+06, 7.1642E+06, 7.1820E+06,
     + 7.1998E+06, 7.2176E+06, 7.2354E+06, 7.2533E+06, 7.2713E+06,
     +  7.2892E+06, 7.3072E+06, 7.3253E+06, 7.3433E+06, 7.3614E+06,
     + 7.3796E+06, 7.3977E+06, 7.4159E+06, 7.4342E+06, 7.4525E+06,
     +  7.4708E+06, 7.4891E+06, 7.5075E+06, 7.5259E+06, 7.5444E+06,
     + 7.5629E+06, 7.5814E+06, 7.6000E+06, 7.6186E+06, 7.6372E+06,
     +  7.6559E+06, 7.6746E+06, 7.6933E+06, 7.7121E+06, 7.7309E+06,
     + 7.7498E+06, 7.7687E+06, 7.7876E+06, 7.8065E+06, 7.8255E+06,
     +  7.8446E+06, 7.8636E+06, 7.8827E+06, 7.9019E+06, 7.9211E+06,
     + 7.9403E+06, 7.9595E+06, 7.9788E+06, 7.9981E+06, 8.0175E+06,
     +  8.0369E+06, 8.0563E+06, 8.0758E+06, 8.0953E+06, 8.1148E+06,
     + 8.1344E+06, 8.1540E+06, 8.1737E+06, 8.1934E+06, 8.2131E+06,
     +  8.2329E+06, 8.2527E+06, 8.2725E+06, 8.2924E+06, 8.3123E+06,
     + 8.3323E+06, 8.3523E+06, 8.3723E+06, 8.3924E+06, 8.4125E+06,
     +  8.4326E+06, 8.4528E+06, 8.4730E+06, 8.4933E+06, 8.5136E+06,
     + 8.5339E+06, 8.5543E+06, 8.5747E+06, 8.5952E+06, 8.6157E+06,
     +  8.6362E+06, 8.6568E+06, 8.6774E+06, 8.6980E+06, 8.7187E+06,
     + 8.7394E+06, 8.7601E+06, 8.7809E+06, 8.8018E+06, 8.8227E+06,
     +  8.8436E+06, 8.8645E+06, 8.8855E+06, 8.9065E+06, 8.9276E+06,
     + 8.9487E+06, 8.9699E+06, 8.9910E+06, 9.0123E+06, 9.0335E+06,
     +  9.0548E+06, 9.0762E+06, 9.0976E+06, 9.1190E+06, 9.1405E+06,
     + 9.1620E+06, 9.1835E+06, 9.2051E+06, 9.2267E+06, 9.2484E+06,
     +  9.2701E+06, 9.2918E+06, 9.3136E+06, 9.3354E+06, 9.3573E+06,
     + 9.3792E+06, 9.4011E+06, 9.4231E+06, 9.4451E+06, 9.4672E+06,
     +  9.4893E+06, 9.5114E+06, 9.5336E+06, 9.5559E+06, 9.5781E+06,
     + 9.6004E+06, 9.6228E+06, 9.6452E+06, 9.6676E+06, 9.6901E+06,
     +  9.7126E+06, 9.7351E+06, 9.7577E+06, 9.7804E+06, 9.8031E+06,
     + 9.8258E+06, 9.8485E+06, 9.8713E+06, 9.8942E+06, 9.9171E+06,
     +  9.9400E+06, 9.9630E+06, 9.9860E+06, 1.0009E+07, 1.0032E+07,
     + 1.0055E+07, 1.0078E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       686
      data (QofT( 3,J),J=1,2408)/ 8.1859E-01, 2.1109E+00, 3.7557E+00,
     +  5.6904E+00, 7.8767E+00, 1.0288E+01, 1.2906E+01, 1.5715E+01,
     + 1.8703E+01, 2.1859E+01, 2.5176E+01, 2.8645E+01, 3.2261E+01,
     +  3.6017E+01, 3.9909E+01, 4.3932E+01, 4.8081E+01, 5.2354E+01,
     + 5.6747E+01, 6.1256E+01, 7.0613E+01, 8.0403E+01, 9.0609E+01,
     +  1.0121E+02, 1.1220E+02, 1.2357E+02, 1.3529E+02, 1.4736E+02,
     + 1.5977E+02, 1.7251E+02, 1.8558E+02, 1.9896E+02, 2.1264E+02,
     +  2.2663E+02, 2.4092E+02, 2.5549E+02, 2.7034E+02, 2.8548E+02,
     + 3.0088E+02, 3.1656E+02, 3.3250E+02, 3.4869E+02, 3.6515E+02,
     +  3.8185E+02, 3.9881E+02, 4.1601E+02, 4.3345E+02, 4.5113E+02,
     + 4.6904E+02, 4.8719E+02, 5.0557E+02, 5.2417E+02, 5.4299E+02,
     +  5.6204E+02, 5.8131E+02, 6.0079E+02, 6.2048E+02, 6.4040E+02,
     + 6.6052E+02, 6.8085E+02, 7.0139E+02, 7.2213E+02, 7.4307E+02,
     +  7.6422E+02, 7.8557E+02, 8.0712E+02, 8.2886E+02, 8.5080E+02,
     + 8.7294E+02, 8.9526E+02, 9.1779E+02, 9.4050E+02, 9.6341E+02,
     +  9.8650E+02, 1.0098E+03, 1.0333E+03, 1.0569E+03, 1.0808E+03,
     + 1.1048E+03, 1.1290E+03, 1.1535E+03, 1.1781E+03, 1.2028E+03,
     +  1.2278E+03, 1.2530E+03, 1.2783E+03, 1.3038E+03, 1.3295E+03,
     + 1.3554E+03, 1.3815E+03, 1.4078E+03, 1.4343E+03, 1.4609E+03,
     +  1.4878E+03, 1.5148E+03, 1.5420E+03, 1.5694E+03, 1.5970E+03,
     + 1.6248E+03, 1.6527E+03, 1.6809E+03, 1.7093E+03, 1.7378E+03,
     +  1.7666E+03, 1.7955E+03, 1.8247E+03, 1.8540E+03, 1.8835E+03,
     + 1.9133E+03, 1.9432E+03, 1.9733E+03, 2.0037E+03, 2.0342E+03,
     +  2.0650E+03, 2.0960E+03, 2.1271E+03, 2.1585E+03, 2.1901E+03,
     + 2.2219E+03, 2.2539E+03, 2.2862E+03, 2.3186E+03, 2.3513E+03,
     +  2.3842E+03, 2.4173E+03, 2.4507E+03, 2.4843E+03, 2.5180E+03,
     + 2.5521E+03, 2.5863E+03, 2.6208E+03, 2.6555E+03, 2.6905E+03,
     +  2.7257E+03, 2.7611E+03, 2.7968E+03, 2.8328E+03, 2.8689E+03,
     + 2.9054E+03, 2.9420E+03, 2.9789E+03, 3.0161E+03, 3.0536E+03,
     +  3.0912E+03, 3.1292E+03, 3.1674E+03, 3.2059E+03, 3.2446E+03,
     + 3.2836E+03, 3.3229E+03, 3.3624E+03, 3.4023E+03, 3.4423E+03,
     +  3.4827E+03, 3.5234E+03, 3.5643E+03, 3.6056E+03, 3.6471E+03,
     + 3.6889E+03, 3.7310E+03, 3.7733E+03, 3.8161E+03, 3.8590E+03,
     +  3.9023E+03, 3.9459E+03, 3.9898E+03, 4.0340E+03, 4.0785E+03,
     + 4.1233E+03, 4.1684E+03, 4.2139E+03, 4.2596E+03, 4.3057E+03,
     +  4.3522E+03, 4.3989E+03, 4.4460E+03, 4.4934E+03, 4.5411E+03,
     + 4.5892E+03, 4.6376E+03, 4.6863E+03, 4.7354E+03, 4.7849E+03,
     +  4.8346E+03, 4.8848E+03, 4.9352E+03, 4.9861E+03, 5.0373E+03,
     + 5.0888E+03, 5.1408E+03, 5.1930E+03, 5.2457E+03, 5.2987E+03,
     +  5.3521E+03, 5.4059E+03, 5.4600E+03, 5.5145E+03, 5.5695E+03,
     + 5.6247E+03, 5.6804E+03, 5.7365E+03, 5.7929E+03, 5.8498E+03,
     +  5.9071E+03, 5.9647E+03, 6.0228E+03, 6.0813E+03, 6.1401E+03,
     + 6.1994E+03, 6.2591E+03, 6.3192E+03, 6.3798E+03, 6.4407E+03,
     +  6.5021E+03, 6.5639E+03, 6.6261E+03, 6.6887E+03, 6.7518E+03,
     + 6.8154E+03, 6.8794E+03, 6.9438E+03, 7.0086E+03, 7.0740E+03,
     +  7.1397E+03, 7.2059E+03, 7.2726E+03, 7.3397E+03, 7.4073E+03,
     + 7.4754E+03, 7.5439E+03, 7.6129E+03, 7.6823E+03, 7.7523E+03,
     +  7.8227E+03, 7.8936E+03, 7.9650E+03, 8.0368E+03, 8.1092E+03,
     + 8.1821E+03, 8.2555E+03, 8.3293E+03, 8.4036E+03, 8.4785E+03,
     +  8.5538E+03, 8.6297E+03, 8.7061E+03, 8.7830E+03, 8.8604E+03,
     + 8.9383E+03, 9.0168E+03, 9.0958E+03, 9.1753E+03, 9.2553E+03,
     +  9.3359E+03, 9.4171E+03, 9.4987E+03, 9.5809E+03, 9.6636E+03,
     + 9.7470E+03, 9.8309E+03, 9.9153E+03, 1.0000E+04, 1.0086E+04,
     +  1.0172E+04, 1.0259E+04, 1.0346E+04, 1.0434E+04, 1.0522E+04,
     + 1.0611E+04, 1.0701E+04, 1.0791E+04, 1.0881E+04, 1.0973E+04,
     +  1.1065E+04, 1.1157E+04, 1.1250E+04, 1.1344E+04, 1.1439E+04,
     + 1.1533E+04, 1.1629E+04, 1.1725E+04, 1.1822E+04, 1.1919E+04,
     +  1.2017E+04, 1.2116E+04, 1.2216E+04, 1.2315E+04, 1.2416E+04,
     + 1.2517E+04, 1.2619E+04, 1.2722E+04, 1.2825E+04, 1.2929E+04,
     +  1.3033E+04, 1.3139E+04, 1.3244E+04, 1.3351E+04, 1.3458E+04,
     + 1.3566E+04, 1.3675E+04, 1.3784E+04, 1.3894E+04, 1.4004E+04,
     +  1.4116E+04, 1.4228E+04, 1.4340E+04, 1.4454E+04, 1.4568E+04,
     + 1.4683E+04, 1.4798E+04, 1.4915E+04, 1.5032E+04, 1.5149E+04,
     +  1.5268E+04, 1.5387E+04, 1.5507E+04, 1.5627E+04, 1.5749E+04,
     + 1.5871E+04, 1.5994E+04, 1.6117E+04, 1.6242E+04, 1.6367E+04,
     +  1.6493E+04, 1.6619E+04, 1.6747E+04, 1.6875E+04, 1.7004E+04,
     + 1.7134E+04, 1.7264E+04, 1.7396E+04, 1.7528E+04, 1.7661E+04,
     +  1.7795E+04, 1.7929E+04, 1.8064E+04, 1.8201E+04, 1.8337E+04,
     + 1.8475E+04, 1.8614E+04, 1.8753E+04, 1.8894E+04, 1.9035E+04,
     +  1.9177E+04, 1.9319E+04, 1.9463E+04, 1.9607E+04, 1.9753E+04,
     + 1.9899E+04, 2.0046E+04, 2.0194E+04, 2.0342E+04, 2.0492E+04,
     +  2.0643E+04, 2.0794E+04, 2.0946E+04, 2.1099E+04, 2.1253E+04,
     + 2.1408E+04, 2.1564E+04, 2.1721E+04, 2.1878E+04, 2.2037E+04,
     +  2.2196E+04, 2.2357E+04, 2.2518E+04, 2.2680E+04, 2.2843E+04,
     + 2.3007E+04, 2.3172E+04, 2.3338E+04, 2.3505E+04, 2.3673E+04,
     +  2.3842E+04, 2.4011E+04, 2.4182E+04, 2.4354E+04, 2.4526E+04,
     + 2.4700E+04, 2.4875E+04, 2.5050E+04, 2.5227E+04, 2.5404E+04,
     +  2.5583E+04, 2.5762E+04, 2.5943E+04, 2.6124E+04, 2.6307E+04,
     + 2.6490E+04, 2.6675E+04, 2.6861E+04, 2.7047E+04, 2.7235E+04,
     +  2.7424E+04, 2.7613E+04, 2.7804E+04, 2.7996E+04, 2.8189E+04,
     + 2.8383E+04, 2.8578E+04, 2.8774E+04, 2.8971E+04, 2.9169E+04,
     +  2.9369E+04, 2.9569E+04, 2.9770E+04, 2.9973E+04, 3.0177E+04,
     + 3.0381E+04, 3.0587E+04, 3.0794E+04, 3.1002E+04, 3.1211E+04,
     +  3.1422E+04, 3.1633E+04, 3.1846E+04, 3.2060E+04, 3.2274E+04,
     + 3.2490E+04, 3.2708E+04, 3.2926E+04, 3.3145E+04, 3.3366E+04,
     +  3.3588E+04, 3.3811E+04, 3.4035E+04, 3.4260E+04, 3.4487E+04,
     + 3.4715E+04, 3.4944E+04, 3.5174E+04, 3.5405E+04, 3.5637E+04,
     +  3.5871E+04, 3.6106E+04, 3.6342E+04, 3.6580E+04, 3.6818E+04,
     + 3.7058E+04, 3.7299E+04, 3.7542E+04, 3.7785E+04, 3.8030E+04,
     +  3.8276E+04, 3.8524E+04, 3.8772E+04, 3.9023E+04, 3.9274E+04,
     + 3.9526E+04, 3.9780E+04, 4.0035E+04, 4.0292E+04, 4.0549E+04,
     +  4.0808E+04, 4.1069E+04, 4.1330E+04, 4.1593E+04, 4.1858E+04,
     + 4.2123E+04, 4.2390E+04, 4.2659E+04, 4.2928E+04, 4.3200E+04,
     +  4.3472E+04, 4.3746E+04, 4.4021E+04, 4.4297E+04, 4.4575E+04,
     + 4.4854E+04, 4.5135E+04, 4.5417E+04, 4.5701E+04, 4.5986E+04,
     +  4.6272E+04, 4.6560E+04, 4.6849E+04, 4.7139E+04, 4.7431E+04,
     + 4.7725E+04, 4.8019E+04, 4.8316E+04, 4.8614E+04, 4.8913E+04,
     +  4.9213E+04, 4.9516E+04, 4.9819E+04, 5.0124E+04, 5.0431E+04,
     + 5.0739E+04, 5.1049E+04, 5.1360E+04, 5.1672E+04, 5.1986E+04,
     +  5.2302E+04, 5.2619E+04, 5.2938E+04, 5.3258E+04, 5.3580E+04,
     + 5.3903E+04, 5.4228E+04, 5.4554E+04, 5.4882E+04, 5.5211E+04,
     +  5.5543E+04, 5.5875E+04, 5.6209E+04, 5.6545E+04, 5.6883E+04,
     + 5.7222E+04, 5.7562E+04, 5.7904E+04, 5.8248E+04, 5.8594E+04,
     +  5.8941E+04, 5.9289E+04, 5.9640E+04, 5.9992E+04, 6.0345E+04,
     + 6.0701E+04, 6.1058E+04, 6.1416E+04, 6.1777E+04, 6.2138E+04,
     +  6.2502E+04, 6.2867E+04, 6.3234E+04, 6.3603E+04, 6.3974E+04,
     + 6.4346E+04, 6.4720E+04, 6.5095E+04, 6.5472E+04, 6.5851E+04,
     +  6.6232E+04, 6.6615E+04, 6.6999E+04, 6.7385E+04, 6.7773E+04,
     + 6.8162E+04, 6.8554E+04, 6.8947E+04, 6.9342E+04, 6.9738E+04,
     +  7.0137E+04, 7.0537E+04, 7.0939E+04, 7.1343E+04, 7.1749E+04,
     + 7.2157E+04, 7.2566E+04, 7.2977E+04, 7.3390E+04, 7.3805E+04,
     +  7.4222E+04, 7.4641E+04, 7.5061E+04, 7.5484E+04, 7.5908E+04,
     + 7.6334E+04, 7.6763E+04, 7.7192E+04, 7.7624E+04, 7.8058E+04,
     +  7.8494E+04, 7.8932E+04, 7.9371E+04, 7.9813E+04, 8.0256E+04,
     + 8.0702E+04, 8.1149E+04, 8.1599E+04, 8.2050E+04, 8.2503E+04,
     +  8.2959E+04, 8.3416E+04, 8.3875E+04, 8.4336E+04, 8.4800E+04,
     + 8.5265E+04, 8.5732E+04, 8.6202E+04, 8.6673E+04, 8.7147E+04,
     +  8.7622E+04, 8.8100E+04, 8.8579E+04, 8.9061E+04, 8.9545E+04,
     + 9.0031E+04, 9.0519E+04, 9.1008E+04, 9.1501E+04, 9.1995E+04,
     +  9.2491E+04, 9.2990E+04, 9.3490E+04, 9.3993E+04, 9.4498E+04,
     + 9.5005E+04, 9.5514E+04, 9.6025E+04, 9.6539E+04, 9.7054E+04,
     +  9.7572E+04, 9.8092E+04, 9.8614E+04, 9.9139E+04, 9.9665E+04,
     + 1.0019E+05, 1.0073E+05, 1.0126E+05, 1.0179E+05, 1.0233E+05,
     +  1.0287E+05, 1.0341E+05, 1.0396E+05, 1.0451E+05, 1.0505E+05,
     + 1.0561E+05, 1.0616E+05, 1.0672E+05, 1.0727E+05, 1.0783E+05,
     +  1.0840E+05, 1.0896E+05, 1.0953E+05, 1.1010E+05, 1.1067E+05,
     + 1.1125E+05, 1.1182E+05, 1.1240E+05, 1.1298E+05, 1.1357E+05,
     +  1.1415E+05, 1.1474E+05, 1.1533E+05, 1.1593E+05, 1.1652E+05,
     + 1.1712E+05, 1.1772E+05, 1.1833E+05, 1.1893E+05, 1.1954E+05,
     +  1.2015E+05, 1.2076E+05, 1.2138E+05, 1.2200E+05, 1.2262E+05,
     + 1.2324E+05, 1.2387E+05, 1.2450E+05, 1.2513E+05, 1.2576E+05,
     +  1.2639E+05, 1.2703E+05, 1.2767E+05, 1.2832E+05, 1.2896E+05,
     + 1.2961E+05, 1.3026E+05, 1.3092E+05, 1.3157E+05, 1.3223E+05,
     +  1.3289E+05, 1.3356E+05, 1.3422E+05, 1.3489E+05, 1.3556E+05,
     + 1.3624E+05, 1.3692E+05, 1.3759E+05, 1.3828E+05, 1.3896E+05,
     +  1.3965E+05, 1.4034E+05, 1.4103E+05, 1.4173E+05, 1.4243E+05,
     + 1.4313E+05, 1.4383E+05, 1.4454E+05, 1.4525E+05, 1.4596E+05,
     +  1.4668E+05, 1.4740E+05, 1.4812E+05, 1.4884E+05, 1.4957E+05,
     + 1.5030E+05, 1.5103E+05, 1.5176E+05, 1.5250E+05, 1.5324E+05,
     +  1.5398E+05, 1.5473E+05, 1.5548E+05, 1.5623E+05, 1.5698E+05,
     + 1.5774E+05, 1.5850E+05, 1.5926E+05, 1.6003E+05, 1.6080E+05,
     +  1.6157E+05, 1.6235E+05, 1.6312E+05, 1.6390E+05, 1.6469E+05,
     + 1.6548E+05, 1.6626E+05, 1.6706E+05, 1.6785E+05, 1.6865E+05,
     +  1.6945E+05, 1.7026E+05, 1.7106E+05, 1.7188E+05, 1.7269E+05,
     + 1.7351E+05, 1.7433E+05, 1.7515E+05, 1.7597E+05, 1.7680E+05,
     +  1.7763E+05, 1.7847E+05, 1.7931E+05, 1.8015E+05, 1.8099E+05,
     + 1.8184E+05, 1.8269E+05, 1.8355E+05, 1.8440E+05, 1.8526E+05,
     +  1.8613E+05, 1.8699E+05, 1.8786E+05, 1.8873E+05, 1.8961E+05,
     + 1.9049E+05, 1.9137E+05, 1.9226E+05, 1.9315E+05, 1.9404E+05,
     +  1.9493E+05, 1.9583E+05, 1.9673E+05, 1.9764E+05, 1.9855E+05,
     + 1.9946E+05, 2.0037E+05, 2.0129E+05, 2.0221E+05, 2.0314E+05,
     +  2.0407E+05, 2.0500E+05, 2.0593E+05, 2.0687E+05, 2.0781E+05,
     + 2.0876E+05, 2.0971E+05, 2.1066E+05, 2.1162E+05, 2.1257E+05,
     +  2.1354E+05, 2.1450E+05, 2.1547E+05, 2.1644E+05, 2.1742E+05,
     + 2.1840E+05, 2.1938E+05, 2.2037E+05, 2.2136E+05, 2.2235E+05,
     +  2.2335E+05, 2.2435E+05, 2.2535E+05, 2.2636E+05, 2.2737E+05,
     + 2.2839E+05, 2.2940E+05, 2.3043E+05, 2.3145E+05, 2.3248E+05,
     +  2.3351E+05, 2.3455E+05, 2.3559E+05, 2.3663E+05, 2.3768E+05,
     + 2.3873E+05, 2.3979E+05, 2.4084E+05, 2.4191E+05, 2.4297E+05,
     +  2.4404E+05, 2.4511E+05, 2.4619E+05, 2.4727E+05, 2.4835E+05,
     + 2.4944E+05, 2.5053E+05, 2.5163E+05, 2.5273E+05, 2.5383E+05,
     +  2.5494E+05, 2.5605E+05, 2.5716E+05, 2.5828E+05, 2.5940E+05,
     + 2.6053E+05, 2.6166E+05, 2.6279E+05, 2.6393E+05, 2.6507E+05,
     +  2.6621E+05, 2.6736E+05, 2.6852E+05, 2.6967E+05, 2.7083E+05,
     + 2.7200E+05, 2.7317E+05, 2.7434E+05, 2.7552E+05, 2.7670E+05,
     +  2.7788E+05, 2.7907E+05, 2.8026E+05, 2.8146E+05, 2.8266E+05,
     + 2.8386E+05, 2.8507E+05, 2.8628E+05, 2.8750E+05, 2.8872E+05,
     +  2.8995E+05, 2.9118E+05, 2.9241E+05, 2.9365E+05, 2.9489E+05,
     + 2.9613E+05, 2.9738E+05, 2.9864E+05, 2.9989E+05, 3.0116E+05,
     +  3.0242E+05, 3.0369E+05, 3.0497E+05, 3.0625E+05, 3.0753E+05,
     + 3.0882E+05, 3.1011E+05, 3.1140E+05, 3.1270E+05, 3.1401E+05,
     +  3.1532E+05, 3.1663E+05, 3.1795E+05, 3.1927E+05, 3.2060E+05,
     + 3.2193E+05, 3.2326E+05, 3.2460E+05, 3.2594E+05, 3.2729E+05,
     +  3.2864E+05, 3.3000E+05, 3.3136E+05, 3.3272E+05, 3.3409E+05,
     + 3.3547E+05, 3.3685E+05, 3.3823E+05, 3.3962E+05, 3.4101E+05,
     +  3.4241E+05, 3.4381E+05, 3.4521E+05, 3.4662E+05, 3.4804E+05,
     + 3.4946E+05, 3.5088E+05, 3.5231E+05, 3.5374E+05, 3.5518E+05,
     +  3.5662E+05, 3.5807E+05, 3.5952E+05, 3.6097E+05, 3.6244E+05,
     + 3.6390E+05, 3.6537E+05, 3.6684E+05, 3.6832E+05, 3.6981E+05,
     +  3.7130E+05, 3.7279E+05, 3.7429E+05, 3.7579E+05, 3.7730E+05,
     + 3.7881E+05, 3.8033E+05, 3.8185E+05, 3.8338E+05, 3.8491E+05,
     +  3.8644E+05, 3.8799E+05, 3.8953E+05, 3.9108E+05, 3.9264E+05,
     + 3.9420E+05, 3.9576E+05, 3.9733E+05, 3.9891E+05, 4.0049E+05,
     +  4.0207E+05, 4.0366E+05, 4.0526E+05, 4.0686E+05, 4.0846E+05,
     + 4.1007E+05, 4.1169E+05, 4.1331E+05, 4.1493E+05, 4.1656E+05,
     +  4.1820E+05, 4.1984E+05, 4.2148E+05, 4.2313E+05, 4.2479E+05,
     + 4.2645E+05, 4.2811E+05, 4.2978E+05, 4.3146E+05, 4.3314E+05,
     +  4.3483E+05, 4.3652E+05, 4.3821E+05, 4.3992E+05, 4.4162E+05,
     + 4.4333E+05, 4.4505E+05, 4.4677E+05, 4.4850E+05, 4.5023E+05,
     +  4.5197E+05, 4.5371E+05, 4.5546E+05, 4.5722E+05, 4.5898E+05,
     + 4.6074E+05, 4.6251E+05, 4.6429E+05, 4.6607E+05, 4.6785E+05,
     +  4.6964E+05, 4.7144E+05, 4.7324E+05, 4.7505E+05, 4.7686E+05,
     + 4.7868E+05, 4.8051E+05, 4.8234E+05, 4.8417E+05, 4.8601E+05,
     +  4.8786E+05, 4.8971E+05, 4.9156E+05, 4.9343E+05, 4.9530E+05,
     + 4.9717E+05, 4.9905E+05, 5.0093E+05, 5.0282E+05, 5.0472E+05,
     +  5.0662E+05, 5.0853E+05, 5.1044E+05, 5.1236E+05, 5.1428E+05,
     + 5.1621E+05, 5.1815E+05, 5.2009E+05, 5.2204E+05, 5.2399E+05,
     +  5.2595E+05, 5.2791E+05, 5.2988E+05, 5.3186E+05, 5.3384E+05,
     + 5.3583E+05, 5.3782E+05, 5.3982E+05, 5.4182E+05, 5.4384E+05,
     +  5.4585E+05, 5.4787E+05, 5.4990E+05, 5.5194E+05, 5.5398E+05,
     + 5.5602E+05, 5.5808E+05, 5.6013E+05, 5.6220E+05, 5.6427E+05,
     +  5.6634E+05, 5.6843E+05, 5.7051E+05, 5.7261E+05, 5.7471E+05,
     + 5.7681E+05, 5.7893E+05, 5.8104E+05, 5.8317E+05, 5.8530E+05,
     +  5.8744E+05, 5.8958E+05, 5.9173E+05, 5.9388E+05, 5.9604E+05,
     + 5.9821E+05, 6.0038E+05, 6.0256E+05, 6.0475E+05, 6.0694E+05,
     +  6.0914E+05, 6.1135E+05, 6.1356E+05, 6.1577E+05, 6.1800E+05,
     + 6.2023E+05, 6.2246E+05, 6.2471E+05, 6.2696E+05, 6.2921E+05,
     +  6.3147E+05, 6.3374E+05, 6.3602E+05, 6.3830E+05, 6.4058E+05,
     + 6.4288E+05, 6.4518E+05, 6.4749E+05, 6.4980E+05, 6.5212E+05,
     +  6.5444E+05, 6.5678E+05, 6.5912E+05, 6.6146E+05, 6.6382E+05,
     + 6.6617E+05, 6.6854E+05, 6.7091E+05, 6.7329E+05, 6.7568E+05,
     +  6.7807E+05, 6.8047E+05, 6.8287E+05, 6.8528E+05, 6.8770E+05,
     + 6.9013E+05, 6.9256E+05, 6.9500E+05, 6.9745E+05, 6.9990E+05,
     +  7.0236E+05, 7.0482E+05, 7.0730E+05, 7.0978E+05, 7.1226E+05,
     + 7.1476E+05, 7.1726E+05, 7.1976E+05, 7.2228E+05, 7.2480E+05,
     +  7.2733E+05, 7.2986E+05, 7.3240E+05, 7.3495E+05, 7.3751E+05,
     + 7.4007E+05, 7.4264E+05, 7.4522E+05, 7.4780E+05, 7.5039E+05,
     +  7.5299E+05, 7.5559E+05, 7.5821E+05, 7.6082E+05, 7.6345E+05,
     + 7.6608E+05, 7.6872E+05, 7.7137E+05, 7.7403E+05, 7.7669E+05,
     +  7.7936E+05, 7.8203E+05, 7.8472E+05, 7.8741E+05, 7.9011E+05,
     + 7.9281E+05, 7.9553E+05, 7.9824E+05, 8.0097E+05, 8.0371E+05,
     +  8.0645E+05, 8.0920E+05, 8.1195E+05, 8.1472E+05, 8.1749E+05,
     + 8.2027E+05, 8.2305E+05, 8.2585E+05, 8.2865E+05, 8.3146E+05,
     +  8.3427E+05, 8.3710E+05, 8.3993E+05, 8.4277E+05, 8.4561E+05,
     + 8.4847E+05, 8.5133E+05, 8.5420E+05, 8.5707E+05, 8.5996E+05,
     +  8.6285E+05, 8.6575E+05, 8.6866E+05, 8.7157E+05, 8.7449E+05,
     + 8.7742E+05, 8.8036E+05, 8.8331E+05, 8.8626E+05, 8.8922E+05,
     +  8.9219E+05, 8.9517E+05, 8.9815E+05, 9.0114E+05, 9.0414E+05,
     + 9.0715E+05, 9.1017E+05, 9.1319E+05, 9.1622E+05, 9.1926E+05,
     +  9.2231E+05, 9.2536E+05, 9.2843E+05, 9.3150E+05, 9.3458E+05,
     + 9.3766E+05, 9.4076E+05, 9.4386E+05, 9.4697E+05, 9.5009E+05,
     +  9.5322E+05, 9.5636E+05, 9.5950E+05, 9.6265E+05, 9.6581E+05,
     + 9.6898E+05, 9.7215E+05, 9.7534E+05, 9.7853E+05, 9.8173E+05,
     +  9.8494E+05, 9.8816E+05, 9.9139E+05, 9.9462E+05, 9.9786E+05,
     + 1.0011E+06, 1.0044E+06, 1.0076E+06, 1.0109E+06, 1.0142E+06,
     +  1.0175E+06, 1.0208E+06, 1.0241E+06, 1.0274E+06, 1.0307E+06,
     + 1.0341E+06, 1.0374E+06, 1.0408E+06, 1.0441E+06, 1.0475E+06,
     +  1.0509E+06, 1.0543E+06, 1.0577E+06, 1.0611E+06, 1.0645E+06,
     + 1.0679E+06, 1.0713E+06, 1.0748E+06, 1.0782E+06, 1.0817E+06,
     +  1.0851E+06, 1.0886E+06, 1.0921E+06, 1.0956E+06, 1.0991E+06,
     + 1.1026E+06, 1.1061E+06, 1.1096E+06, 1.1131E+06, 1.1167E+06,
     +  1.1202E+06, 1.1238E+06, 1.1274E+06, 1.1309E+06, 1.1345E+06,
     + 1.1381E+06, 1.1417E+06, 1.1453E+06, 1.1490E+06, 1.1526E+06,
     +  1.1562E+06, 1.1599E+06, 1.1635E+06, 1.1672E+06, 1.1709E+06,
     + 1.1746E+06, 1.1783E+06, 1.1820E+06, 1.1857E+06, 1.1894E+06,
     +  1.1931E+06, 1.1969E+06, 1.2006E+06, 1.2044E+06, 1.2082E+06,
     + 1.2119E+06, 1.2157E+06, 1.2195E+06, 1.2233E+06, 1.2271E+06,
     +  1.2310E+06, 1.2348E+06, 1.2386E+06, 1.2425E+06, 1.2463E+06,
     + 1.2502E+06, 1.2541E+06, 1.2580E+06, 1.2619E+06, 1.2658E+06,
     +  1.2697E+06, 1.2736E+06, 1.2776E+06, 1.2815E+06, 1.2855E+06,
     + 1.2895E+06, 1.2934E+06, 1.2974E+06, 1.3014E+06, 1.3054E+06,
     +  1.3094E+06, 1.3135E+06, 1.3175E+06, 1.3215E+06, 1.3256E+06,
     + 1.3296E+06, 1.3337E+06, 1.3378E+06, 1.3419E+06, 1.3460E+06,
     +  1.3501E+06, 1.3542E+06, 1.3584E+06, 1.3625E+06, 1.3667E+06,
     + 1.3708E+06, 1.3750E+06, 1.3792E+06, 1.3834E+06, 1.3876E+06,
     +  1.3918E+06, 1.3960E+06, 1.4002E+06, 1.4045E+06, 1.4087E+06,
     + 1.4130E+06, 1.4173E+06, 1.4215E+06, 1.4258E+06, 1.4301E+06,
     +  1.4345E+06, 1.4388E+06, 1.4431E+06, 1.4475E+06, 1.4518E+06,
     + 1.4562E+06, 1.4606E+06, 1.4649E+06, 1.4693E+06, 1.4737E+06,
     +  1.4782E+06, 1.4826E+06, 1.4870E+06, 1.4915E+06, 1.4959E+06,
     + 1.5004E+06, 1.5049E+06, 1.5094E+06, 1.5139E+06, 1.5184E+06,
     +  1.5229E+06, 1.5274E+06, 1.5320E+06, 1.5365E+06, 1.5411E+06,
     + 1.5457E+06, 1.5502E+06, 1.5548E+06, 1.5594E+06, 1.5641E+06,
     +  1.5687E+06, 1.5733E+06, 1.5780E+06, 1.5826E+06, 1.5873E+06,
     + 1.5920E+06, 1.5967E+06, 1.6014E+06, 1.6061E+06, 1.6108E+06,
     +  1.6156E+06, 1.6203E+06, 1.6251E+06, 1.6298E+06, 1.6346E+06,
     + 1.6394E+06, 1.6442E+06, 1.6490E+06, 1.6538E+06, 1.6587E+06,
     +  1.6635E+06, 1.6684E+06, 1.6733E+06, 1.6781E+06, 1.6830E+06,
     + 1.6879E+06, 1.6928E+06, 1.6978E+06, 1.7027E+06, 1.7076E+06,
     +  1.7126E+06, 1.7176E+06, 1.7226E+06, 1.7275E+06, 1.7325E+06,
     + 1.7376E+06, 1.7426E+06, 1.7476E+06, 1.7527E+06, 1.7577E+06,
     +  1.7628E+06, 1.7679E+06, 1.7730E+06, 1.7781E+06, 1.7832E+06,
     + 1.7883E+06, 1.7935E+06, 1.7986E+06, 1.8038E+06, 1.8090E+06,
     +  1.8142E+06, 1.8194E+06, 1.8246E+06, 1.8298E+06, 1.8350E+06,
     + 1.8403E+06, 1.8455E+06, 1.8508E+06, 1.8561E+06, 1.8614E+06,
     +  1.8667E+06, 1.8720E+06, 1.8773E+06, 1.8827E+06, 1.8880E+06,
     + 1.8934E+06, 1.8988E+06, 1.9042E+06, 1.9096E+06, 1.9150E+06,
     +  1.9204E+06, 1.9259E+06, 1.9313E+06, 1.9368E+06, 1.9422E+06,
     + 1.9477E+06, 1.9532E+06, 1.9587E+06, 1.9643E+06, 1.9698E+06,
     +  1.9753E+06, 1.9809E+06, 1.9865E+06, 1.9921E+06, 1.9977E+06,
     + 2.0033E+06, 2.0089E+06, 2.0145E+06, 2.0202E+06, 2.0258E+06,
     +  2.0315E+06, 2.0372E+06, 2.0429E+06, 2.0486E+06, 2.0543E+06,
     + 2.0600E+06, 2.0658E+06, 2.0715E+06, 2.0773E+06, 2.0831E+06,
     +  2.0889E+06, 2.0947E+06, 2.1005E+06, 2.1064E+06, 2.1122E+06,
     + 2.1181E+06, 2.1239E+06, 2.1298E+06, 2.1357E+06, 2.1416E+06,
     +  2.1476E+06, 2.1535E+06, 2.1594E+06, 2.1654E+06, 2.1714E+06,
     + 2.1774E+06, 2.1834E+06, 2.1894E+06, 2.1954E+06, 2.2015E+06,
     +  2.2075E+06, 2.2136E+06, 2.2197E+06, 2.2257E+06, 2.2319E+06,
     + 2.2380E+06, 2.2441E+06, 2.2503E+06, 2.2564E+06, 2.2626E+06,
     +  2.2688E+06, 2.2750E+06, 2.2812E+06, 2.2874E+06, 2.2936E+06,
     + 2.2999E+06, 2.3062E+06, 2.3124E+06, 2.3187E+06, 2.3250E+06,
     +  2.3314E+06, 2.3377E+06, 2.3440E+06, 2.3504E+06, 2.3568E+06,
     + 2.3632E+06, 2.3696E+06, 2.3760E+06, 2.3824E+06, 2.3888E+06,
     +  2.3953E+06, 2.4018E+06, 2.4083E+06, 2.4147E+06, 2.4213E+06,
     + 2.4278E+06, 2.4343E+06, 2.4409E+06, 2.4474E+06, 2.4540E+06,
     +  2.4606E+06, 2.4672E+06, 2.4738E+06, 2.4805E+06, 2.4871E+06,
     + 2.4938E+06, 2.5005E+06, 2.5072E+06, 2.5139E+06, 2.5206E+06,
     +  2.5273E+06, 2.5341E+06, 2.5408E+06, 2.5476E+06, 2.5544E+06,
     + 2.5612E+06, 2.5680E+06, 2.5748E+06, 2.5817E+06, 2.5886E+06,
     +  2.5954E+06, 2.6023E+06, 2.6092E+06, 2.6161E+06, 2.6231E+06,
     + 2.6300E+06, 2.6370E+06, 2.6440E+06, 2.6510E+06, 2.6580E+06,
     +  2.6650E+06, 2.6720E+06, 2.6791E+06, 2.6861E+06, 2.6932E+06,
     + 2.7003E+06, 2.7074E+06, 2.7145E+06, 2.7217E+06, 2.7288E+06,
     +  2.7360E+06, 2.7432E+06, 2.7504E+06, 2.7576E+06, 2.7648E+06,
     + 2.7720E+06, 2.7793E+06, 2.7866E+06, 2.7939E+06, 2.8012E+06,
     +  2.8085E+06, 2.8158E+06, 2.8231E+06, 2.8305E+06, 2.8379E+06,
     + 2.8453E+06, 2.8527E+06, 2.8601E+06, 2.8675E+06, 2.8750E+06,
     +  2.8825E+06, 2.8899E+06, 2.8974E+06, 2.9049E+06, 2.9125E+06,
     + 2.9200E+06, 2.9276E+06, 2.9352E+06, 2.9427E+06, 2.9503E+06,
     +  2.9580E+06, 2.9656E+06, 2.9733E+06, 2.9809E+06, 2.9886E+06,
     + 2.9963E+06, 3.0040E+06, 3.0117E+06, 3.0195E+06, 3.0272E+06,
     +  3.0350E+06, 3.0428E+06, 3.0506E+06, 3.0584E+06, 3.0663E+06,
     + 3.0741E+06, 3.0820E+06, 3.0899E+06, 3.0978E+06, 3.1057E+06,
     +  3.1137E+06, 3.1216E+06, 3.1296E+06, 3.1375E+06, 3.1455E+06,
     + 3.1536E+06, 3.1616E+06, 3.1696E+06, 3.1777E+06, 3.1858E+06,
     +  3.1939E+06, 3.2020E+06, 3.2101E+06, 3.2182E+06, 3.2264E+06,
     + 3.2346E+06, 3.2428E+06, 3.2510E+06, 3.2592E+06, 3.2674E+06,
     +  3.2757E+06, 3.2840E+06, 3.2923E+06, 3.3006E+06, 3.3089E+06,
     + 3.3172E+06, 3.3256E+06, 3.3340E+06, 3.3424E+06, 3.3508E+06,
     +  3.3592E+06, 3.3676E+06, 3.3761E+06, 3.3845E+06, 3.3930E+06,
     + 3.4015E+06, 3.4101E+06, 3.4186E+06, 3.4272E+06, 3.4357E+06,
     +  3.4443E+06, 3.4529E+06, 3.4615E+06, 3.4702E+06, 3.4788E+06,
     + 3.4875E+06, 3.4962E+06, 3.5049E+06, 3.5136E+06, 3.5224E+06,
     +  3.5311E+06, 3.5399E+06, 3.5487E+06, 3.5575E+06, 3.5664E+06,
     + 3.5752E+06, 3.5841E+06, 3.5929E+06, 3.6018E+06, 3.6107E+06,
     +  3.6197E+06, 3.6286E+06, 3.6376E+06, 3.6466E+06, 3.6556E+06,
     + 3.6646E+06, 3.6736E+06, 3.6827E+06, 3.6918E+06, 3.7008E+06,
     +  3.7100E+06, 3.7191E+06, 3.7282E+06, 3.7374E+06, 3.7466E+06,
     + 3.7557E+06, 3.7650E+06, 3.7742E+06, 3.7834E+06, 3.7927E+06,
     +  3.8020E+06, 3.8113E+06, 3.8206E+06, 3.8299E+06, 3.8393E+06,
     + 3.8487E+06, 3.8581E+06, 3.8675E+06, 3.8769E+06, 3.8863E+06,
     +  3.8958E+06, 3.9053E+06, 3.9148E+06, 3.9243E+06, 3.9338E+06,
     + 3.9434E+06, 3.9530E+06, 3.9626E+06, 3.9722E+06, 3.9818E+06,
     +  3.9914E+06, 4.0011E+06, 4.0108E+06, 4.0205E+06, 4.0302E+06,
     + 4.0399E+06, 4.0497E+06, 4.0595E+06, 4.0693E+06, 4.0791E+06,
     +  4.0889E+06, 4.0988E+06, 4.1086E+06, 4.1185E+06, 4.1284E+06,
     + 4.1383E+06, 4.1483E+06, 4.1582E+06, 4.1682E+06, 4.1782E+06,
     +  4.1882E+06, 4.1983E+06, 4.2083E+06, 4.2184E+06, 4.2285E+06,
     + 4.2386E+06, 4.2487E+06, 4.2589E+06, 4.2690E+06, 4.2792E+06,
     +  4.2894E+06, 4.2997E+06, 4.3099E+06, 4.3202E+06, 4.3305E+06,
     + 4.3408E+06, 4.3511E+06, 4.3614E+06, 4.3718E+06, 4.3822E+06,
     +  4.3926E+06, 4.4030E+06, 4.4134E+06, 4.4239E+06, 4.4344E+06,
     + 4.4449E+06, 4.4554E+06, 4.4659E+06, 4.4765E+06, 4.4870E+06,
     +  4.4976E+06, 4.5083E+06, 4.5189E+06, 4.5295E+06, 4.5402E+06,
     + 4.5509E+06, 4.5616E+06, 4.5724E+06, 4.5831E+06, 4.5939E+06,
     +  4.6047E+06, 4.6155E+06, 4.6263E+06, 4.6372E+06, 4.6480E+06,
     + 4.6589E+06, 4.6698E+06, 4.6808E+06, 4.6917E+06, 4.7027E+06,
     +  4.7137E+06, 4.7247E+06, 4.7357E+06, 4.7468E+06, 4.7579E+06,
     + 4.7690E+06, 4.7801E+06, 4.7912E+06, 4.8024E+06, 4.8135E+06,
     +  4.8247E+06, 4.8360E+06, 4.8472E+06, 4.8585E+06, 4.8697E+06,
     + 4.8810E+06, 4.8924E+06, 4.9037E+06, 4.9151E+06, 4.9264E+06,
     +  4.9378E+06, 4.9493E+06, 4.9607E+06, 4.9722E+06, 4.9837E+06,
     + 4.9952E+06, 5.0067E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 3,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       667
      data (QofT( 4,J),J=1,2408)/ 9.6939E+00, 2.5063E+01, 4.4624E+01,
     +  6.7633E+01, 9.3636E+01, 1.2232E+02, 1.5346E+02, 1.8687E+02,
     + 2.2241E+02, 2.5996E+02, 2.9941E+02, 3.4068E+02, 3.8369E+02,
     +  4.2837E+02, 4.7466E+02, 5.2252E+02, 5.7188E+02, 6.2272E+02,
     + 6.7497E+02, 7.2861E+02, 8.3992E+02, 9.5638E+02, 1.0778E+03,
     +  1.2040E+03, 1.3347E+03, 1.4699E+03, 1.6093E+03, 1.7529E+03,
     + 1.9006E+03, 2.0521E+03, 2.2076E+03, 2.3667E+03, 2.5296E+03,
     +  2.6960E+03, 2.8659E+03, 3.0392E+03, 3.2160E+03, 3.3960E+03,
     + 3.5793E+03, 3.7657E+03, 3.9554E+03, 4.1481E+03, 4.3438E+03,
     +  4.5426E+03, 4.7443E+03, 4.9489E+03, 5.1563E+03, 5.3666E+03,
     + 5.5798E+03, 5.7956E+03, 6.0143E+03, 6.2356E+03, 6.4595E+03,
     +  6.6861E+03, 6.9153E+03, 7.1471E+03, 7.3815E+03, 7.6183E+03,
     + 7.8577E+03, 8.0996E+03, 8.3439E+03, 8.5906E+03, 8.8398E+03,
     +  9.0914E+03, 9.3454E+03, 9.6017E+03, 9.8603E+03, 1.0121E+04,
     + 1.0385E+04, 1.0650E+04, 1.0918E+04, 1.1188E+04, 1.1461E+04,
     +  1.1736E+04, 1.2013E+04, 1.2292E+04, 1.2574E+04, 1.2857E+04,
     + 1.3143E+04, 1.3431E+04, 1.3722E+04, 1.4014E+04, 1.4309E+04,
     +  1.4606E+04, 1.4905E+04, 1.5207E+04, 1.5511E+04, 1.5817E+04,
     + 1.6125E+04, 1.6435E+04, 1.6747E+04, 1.7062E+04, 1.7379E+04,
     +  1.7698E+04, 1.8020E+04, 1.8343E+04, 1.8669E+04, 1.8997E+04,
     + 1.9328E+04, 1.9660E+04, 1.9995E+04, 2.0332E+04, 2.0672E+04,
     +  2.1014E+04, 2.1358E+04, 2.1704E+04, 2.2053E+04, 2.2404E+04,
     + 2.2758E+04, 2.3114E+04, 2.3472E+04, 2.3833E+04, 2.4196E+04,
     +  2.4561E+04, 2.4929E+04, 2.5300E+04, 2.5673E+04, 2.6048E+04,
     + 2.6426E+04, 2.6807E+04, 2.7190E+04, 2.7575E+04, 2.7963E+04,
     +  2.8354E+04, 2.8747E+04, 2.9143E+04, 2.9542E+04, 2.9943E+04,
     + 3.0347E+04, 3.0754E+04, 3.1163E+04, 3.1575E+04, 3.1991E+04,
     +  3.2408E+04, 3.2829E+04, 3.3252E+04, 3.3679E+04, 3.4108E+04,
     + 3.4540E+04, 3.4975E+04, 3.5413E+04, 3.5854E+04, 3.6297E+04,
     +  3.6744E+04, 3.7195E+04, 3.7647E+04, 3.8103E+04, 3.8563E+04,
     + 3.9025E+04, 3.9490E+04, 3.9959E+04, 4.0431E+04, 4.0906E+04,
     +  4.1384E+04, 4.1866E+04, 4.2351E+04, 4.2839E+04, 4.3331E+04,
     + 4.3826E+04, 4.4324E+04, 4.4826E+04, 4.5331E+04, 4.5840E+04,
     +  4.6352E+04, 4.6868E+04, 4.7387E+04, 4.7911E+04, 4.8437E+04,
     + 4.8967E+04, 4.9501E+04, 5.0039E+04, 5.0580E+04, 5.1125E+04,
     +  5.1674E+04, 5.2226E+04, 5.2783E+04, 5.3343E+04, 5.3907E+04,
     + 5.4475E+04, 5.5047E+04, 5.5623E+04, 5.6203E+04, 5.6787E+04,
     +  5.7375E+04, 5.7967E+04, 5.8563E+04, 5.9164E+04, 5.9768E+04,
     + 6.0377E+04, 6.0990E+04, 6.1607E+04, 6.2228E+04, 6.2854E+04,
     +  6.3484E+04, 6.4118E+04, 6.4757E+04, 6.5400E+04, 6.6048E+04,
     + 6.6700E+04, 6.7357E+04, 6.8018E+04, 6.8683E+04, 6.9354E+04,
     +  7.0029E+04, 7.0708E+04, 7.1393E+04, 7.2082E+04, 7.2776E+04,
     + 7.3474E+04, 7.4177E+04, 7.4886E+04, 7.5599E+04, 7.6316E+04,
     +  7.7040E+04, 7.7768E+04, 7.8500E+04, 7.9238E+04, 7.9981E+04,
     + 8.0729E+04, 8.1482E+04, 8.2240E+04, 8.3003E+04, 8.3772E+04,
     +  8.4546E+04, 8.5326E+04, 8.6110E+04, 8.6899E+04, 8.7695E+04,
     + 8.8495E+04, 8.9301E+04, 9.0112E+04, 9.0929E+04, 9.1752E+04,
     +  9.2579E+04, 9.3414E+04, 9.4253E+04, 9.5097E+04, 9.5948E+04,
     + 9.6804E+04, 9.7666E+04, 9.8534E+04, 9.9408E+04, 1.0029E+05,
     +  1.0117E+05, 1.0206E+05, 1.0296E+05, 1.0386E+05, 1.0477E+05,
     + 1.0569E+05, 1.0661E+05, 1.0754E+05, 1.0847E+05, 1.0941E+05,
     +  1.1036E+05, 1.1131E+05, 1.1227E+05, 1.1323E+05, 1.1420E+05,
     + 1.1518E+05, 1.1617E+05, 1.1716E+05, 1.1815E+05, 1.1916E+05,
     +  1.2017E+05, 1.2118E+05, 1.2221E+05, 1.2324E+05, 1.2427E+05,
     + 1.2532E+05, 1.2637E+05, 1.2742E+05, 1.2849E+05, 1.2956E+05,
     +  1.3064E+05, 1.3172E+05, 1.3281E+05, 1.3391E+05, 1.3502E+05,
     + 1.3613E+05, 1.3725E+05, 1.3838E+05, 1.3951E+05, 1.4065E+05,
     +  1.4180E+05, 1.4296E+05, 1.4412E+05, 1.4529E+05, 1.4647E+05,
     + 1.4765E+05, 1.4885E+05, 1.5005E+05, 1.5126E+05, 1.5247E+05,
     +  1.5370E+05, 1.5493E+05, 1.5617E+05, 1.5741E+05, 1.5867E+05,
     + 1.5993E+05, 1.6120E+05, 1.6248E+05, 1.6377E+05, 1.6506E+05,
     +  1.6636E+05, 1.6767E+05, 1.6899E+05, 1.7032E+05, 1.7165E+05,
     + 1.7300E+05, 1.7435E+05, 1.7571E+05, 1.7707E+05, 1.7845E+05,
     +  1.7984E+05, 1.8123E+05, 1.8263E+05, 1.8404E+05, 1.8546E+05,
     + 1.8689E+05, 1.8832E+05, 1.8977E+05, 1.9122E+05, 1.9269E+05,
     +  1.9416E+05, 1.9564E+05, 1.9713E+05, 1.9862E+05, 2.0013E+05,
     + 2.0165E+05, 2.0317E+05, 2.0471E+05, 2.0625E+05, 2.0780E+05,
     +  2.0937E+05, 2.1094E+05, 2.1252E+05, 2.1411E+05, 2.1571E+05,
     + 2.1732E+05, 2.1894E+05, 2.2056E+05, 2.2220E+05, 2.2385E+05,
     +  2.2550E+05, 2.2717E+05, 2.2885E+05, 2.3053E+05, 2.3223E+05,
     + 2.3393E+05, 2.3565E+05, 2.3738E+05, 2.3911E+05, 2.4086E+05,
     +  2.4261E+05, 2.4438E+05, 2.4616E+05, 2.4794E+05, 2.4974E+05,
     + 2.5155E+05, 2.5336E+05, 2.5519E+05, 2.5703E+05, 2.5888E+05,
     +  2.6074E+05, 2.6261E+05, 2.6449E+05, 2.6638E+05, 2.6828E+05,
     + 2.7020E+05, 2.7212E+05, 2.7406E+05, 2.7600E+05, 2.7796E+05,
     +  2.7993E+05, 2.8190E+05, 2.8389E+05, 2.8590E+05, 2.8791E+05,
     + 2.8993E+05, 2.9197E+05, 2.9401E+05, 2.9607E+05, 2.9814E+05,
     +  3.0022E+05, 3.0231E+05, 3.0441E+05, 3.0653E+05, 3.0866E+05,
     + 3.1079E+05, 3.1295E+05, 3.1511E+05, 3.1728E+05, 3.1947E+05,
     +  3.2166E+05, 3.2387E+05, 3.2610E+05, 3.2833E+05, 3.3058E+05,
     + 3.3284E+05, 3.3511E+05, 3.3739E+05, 3.3969E+05, 3.4199E+05,
     +  3.4431E+05, 3.4665E+05, 3.4899E+05, 3.5135E+05, 3.5372E+05,
     + 3.5611E+05, 3.5850E+05, 3.6091E+05, 3.6333E+05, 3.6577E+05,
     +  3.6822E+05, 3.7068E+05, 3.7315E+05, 3.7564E+05, 3.7814E+05,
     + 3.8065E+05, 3.8318E+05, 3.8572E+05, 3.8827E+05, 3.9084E+05,
     +  3.9342E+05, 3.9602E+05, 3.9862E+05, 4.0124E+05, 4.0388E+05,
     + 4.0653E+05, 4.0919E+05, 4.1187E+05, 4.1456E+05, 4.1726E+05,
     +  4.1998E+05, 4.2271E+05, 4.2546E+05, 4.2822E+05, 4.3100E+05,
     + 4.3378E+05, 4.3659E+05, 4.3941E+05, 4.4224E+05, 4.4509E+05,
     +  4.4795E+05, 4.5082E+05, 4.5371E+05, 4.5662E+05, 4.5954E+05,
     + 4.6248E+05, 4.6543E+05, 4.6839E+05, 4.7137E+05, 4.7437E+05,
     +  4.7738E+05, 4.8040E+05, 4.8344E+05, 4.8650E+05, 4.8957E+05,
     + 4.9266E+05, 4.9576E+05, 4.9888E+05, 5.0201E+05, 5.0516E+05,
     +  5.0833E+05, 5.1151E+05, 5.1470E+05, 5.1792E+05, 5.2115E+05,
     + 5.2439E+05, 5.2765E+05, 5.3093E+05, 5.3422E+05, 5.3753E+05,
     +  5.4085E+05, 5.4419E+05, 5.4755E+05, 5.5092E+05, 5.5432E+05,
     + 5.5772E+05, 5.6115E+05, 5.6459E+05, 5.6805E+05, 5.7152E+05,
     +  5.7501E+05, 5.7852E+05, 5.8204E+05, 5.8559E+05, 5.8915E+05,
     + 5.9272E+05, 5.9632E+05, 5.9993E+05, 6.0355E+05, 6.0720E+05,
     +  6.1086E+05, 6.1454E+05, 6.1824E+05, 6.2196E+05, 6.2569E+05,
     + 6.2945E+05, 6.3321E+05, 6.3700E+05, 6.4081E+05, 6.4463E+05,
     +  6.4847E+05, 6.5233E+05, 6.5621E+05, 6.6011E+05, 6.6402E+05,
     + 6.6795E+05, 6.7191E+05, 6.7588E+05, 6.7987E+05, 6.8387E+05,
     +  6.8790E+05, 6.9194E+05, 6.9601E+05, 7.0009E+05, 7.0419E+05,
     + 7.0831E+05, 7.1245E+05, 7.1661E+05, 7.2079E+05, 7.2499E+05,
     +  7.2921E+05, 7.3344E+05, 7.3770E+05, 7.4197E+05, 7.4627E+05,
     + 7.5058E+05, 7.5492E+05, 7.5927E+05, 7.6365E+05, 7.6804E+05,
     +  7.7246E+05, 7.7689E+05, 7.8135E+05, 7.8582E+05, 7.9032E+05,
     + 7.9483E+05, 7.9937E+05, 8.0393E+05, 8.0851E+05, 8.1311E+05,
     +  8.1772E+05, 8.2236E+05, 8.2702E+05, 8.3170E+05, 8.3641E+05,
     + 8.4113E+05, 8.4588E+05, 8.5064E+05, 8.5543E+05, 8.6024E+05,
     +  8.6507E+05, 8.6992E+05, 8.7479E+05, 8.7969E+05, 8.8460E+05,
     + 8.8954E+05, 8.9450E+05, 8.9949E+05, 9.0449E+05, 9.0952E+05,
     +  9.1456E+05, 9.1963E+05, 9.2473E+05, 9.2984E+05, 9.3498E+05,
     + 9.4014E+05, 9.4532E+05, 9.5053E+05, 9.5576E+05, 9.6101E+05,
     +  9.6628E+05, 9.7158E+05, 9.7690E+05, 9.8224E+05, 9.8761E+05,
     + 9.9300E+05, 9.9841E+05, 1.0038E+06, 1.0093E+06, 1.0148E+06,
     +  1.0203E+06, 1.0258E+06, 1.0314E+06, 1.0369E+06, 1.0426E+06,
     + 1.0482E+06, 1.0538E+06, 1.0595E+06, 1.0652E+06, 1.0709E+06,
     +  1.0767E+06, 1.0824E+06, 1.0882E+06, 1.0941E+06, 1.0999E+06,
     + 1.1058E+06, 1.1117E+06, 1.1176E+06, 1.1235E+06, 1.1295E+06,
     +  1.1355E+06, 1.1415E+06, 1.1475E+06, 1.1536E+06, 1.1597E+06,
     + 1.1658E+06, 1.1720E+06, 1.1781E+06, 1.1843E+06, 1.1906E+06,
     +  1.1968E+06, 1.2031E+06, 1.2094E+06, 1.2157E+06, 1.2221E+06,
     + 1.2284E+06, 1.2349E+06, 1.2413E+06, 1.2477E+06, 1.2542E+06,
     +  1.2607E+06, 1.2673E+06, 1.2738E+06, 1.2804E+06, 1.2871E+06,
     + 1.2937E+06, 1.3004E+06, 1.3071E+06, 1.3138E+06, 1.3206E+06,
     +  1.3273E+06, 1.3342E+06, 1.3410E+06, 1.3479E+06, 1.3548E+06,
     + 1.3617E+06, 1.3686E+06, 1.3756E+06, 1.3826E+06, 1.3896E+06,
     +  1.3967E+06, 1.4038E+06, 1.4109E+06, 1.4181E+06, 1.4252E+06,
     + 1.4324E+06, 1.4397E+06, 1.4469E+06, 1.4542E+06, 1.4616E+06,
     +  1.4689E+06, 1.4763E+06, 1.4837E+06, 1.4911E+06, 1.4986E+06,
     + 1.5061E+06, 1.5136E+06, 1.5212E+06, 1.5288E+06, 1.5364E+06,
     +  1.5440E+06, 1.5517E+06, 1.5594E+06, 1.5671E+06, 1.5749E+06,
     + 1.5827E+06, 1.5905E+06, 1.5984E+06, 1.6063E+06, 1.6142E+06,
     +  1.6221E+06, 1.6301E+06, 1.6381E+06, 1.6462E+06, 1.6543E+06,
     + 1.6624E+06, 1.6705E+06, 1.6787E+06, 1.6869E+06, 1.6951E+06,
     +  1.7034E+06, 1.7117E+06, 1.7200E+06, 1.7283E+06, 1.7367E+06,
     + 1.7452E+06, 1.7536E+06, 1.7621E+06, 1.7706E+06, 1.7792E+06,
     +  1.7878E+06, 1.7964E+06, 1.8050E+06, 1.8137E+06, 1.8224E+06,
     + 1.8312E+06, 1.8400E+06, 1.8488E+06, 1.8576E+06, 1.8665E+06,
     +  1.8754E+06, 1.8844E+06, 1.8934E+06, 1.9024E+06, 1.9114E+06,
     + 1.9205E+06, 1.9296E+06, 1.9388E+06, 1.9480E+06, 1.9572E+06,
     +  1.9665E+06, 1.9758E+06, 1.9851E+06, 1.9944E+06, 2.0038E+06,
     + 2.0133E+06, 2.0227E+06, 2.0322E+06, 2.0418E+06, 2.0513E+06,
     +  2.0610E+06, 2.0706E+06, 2.0803E+06, 2.0900E+06, 2.0997E+06,
     + 2.1095E+06, 2.1194E+06, 2.1292E+06, 2.1391E+06, 2.1490E+06,
     +  2.1590E+06, 2.1690E+06, 2.1790E+06, 2.1891E+06, 2.1992E+06,
     + 2.2094E+06, 2.2196E+06, 2.2298E+06, 2.2401E+06, 2.2504E+06,
     +  2.2607E+06, 2.2711E+06, 2.2815E+06, 2.2919E+06, 2.3024E+06,
     + 2.3129E+06, 2.3235E+06, 2.3341E+06, 2.3447E+06, 2.3554E+06,
     +  2.3661E+06, 2.3769E+06, 2.3877E+06, 2.3985E+06, 2.4094E+06,
     + 2.4203E+06, 2.4312E+06, 2.4422E+06, 2.4532E+06, 2.4643E+06,
     +  2.4754E+06, 2.4866E+06, 2.4977E+06, 2.5090E+06, 2.5202E+06,
     + 2.5315E+06, 2.5429E+06, 2.5543E+06, 2.5657E+06, 2.5771E+06,
     +  2.5886E+06, 2.6002E+06, 2.6118E+06, 2.6234E+06, 2.6351E+06,
     + 2.6468E+06, 2.6585E+06, 2.6703E+06, 2.6821E+06, 2.6940E+06,
     +  2.7059E+06, 2.7179E+06, 2.7299E+06, 2.7419E+06, 2.7540E+06,
     + 2.7661E+06, 2.7783E+06, 2.7905E+06, 2.8027E+06, 2.8150E+06,
     +  2.8273E+06, 2.8397E+06, 2.8521E+06, 2.8646E+06, 2.8771E+06,
     + 2.8896E+06, 2.9022E+06, 2.9149E+06, 2.9276E+06, 2.9403E+06,
     +  2.9530E+06, 2.9658E+06, 2.9787E+06, 2.9916E+06, 3.0045E+06,
     + 3.0175E+06, 3.0305E+06, 3.0436E+06, 3.0567E+06, 3.0699E+06,
     +  3.0831E+06, 3.0963E+06, 3.1096E+06, 3.1230E+06, 3.1364E+06,
     + 3.1498E+06, 3.1633E+06, 3.1768E+06, 3.1904E+06, 3.2040E+06,
     +  3.2176E+06, 3.2313E+06, 3.2451E+06, 3.2589E+06, 3.2727E+06,
     + 3.2866E+06, 3.3005E+06, 3.3145E+06, 3.3286E+06, 3.3426E+06,
     +  3.3567E+06, 3.3709E+06, 3.3851E+06, 3.3994E+06, 3.4137E+06,
     + 3.4281E+06, 3.4425E+06, 3.4569E+06, 3.4714E+06, 3.4860E+06,
     +  3.5006E+06, 3.5152E+06, 3.5299E+06, 3.5446E+06, 3.5594E+06,
     + 3.5743E+06, 3.5892E+06, 3.6041E+06, 3.6191E+06, 3.6341E+06,
     +  3.6492E+06, 3.6643E+06, 3.6795E+06, 3.6948E+06, 3.7100E+06,
     + 3.7254E+06, 3.7407E+06, 3.7562E+06, 3.7717E+06, 3.7872E+06,
     +  3.8028E+06, 3.8184E+06, 3.8341E+06, 3.8498E+06, 3.8656E+06,
     + 3.8814E+06, 3.8973E+06, 3.9133E+06, 3.9293E+06, 3.9453E+06,
     +  3.9614E+06, 3.9775E+06, 3.9937E+06, 4.0100E+06, 4.0263E+06,
     + 4.0426E+06, 4.0590E+06, 4.0755E+06, 4.0920E+06, 4.1086E+06,
     +  4.1252E+06, 4.1418E+06, 4.1586E+06, 4.1753E+06, 4.1922E+06,
     + 4.2090E+06, 4.2260E+06, 4.2430E+06, 4.2600E+06, 4.2771E+06,
     +  4.2943E+06, 4.3115E+06, 4.3287E+06, 4.3460E+06, 4.3634E+06,
     + 4.3808E+06, 4.3983E+06, 4.4158E+06, 4.4334E+06, 4.4510E+06,
     +  4.4687E+06, 4.4865E+06, 4.5043E+06, 4.5222E+06, 4.5401E+06,
     + 4.5581E+06, 4.5761E+06, 4.5942E+06, 4.6123E+06, 4.6305E+06,
     +  4.6488E+06, 4.6671E+06, 4.6855E+06, 4.7039E+06, 4.7224E+06,
     + 4.7409E+06, 4.7595E+06, 4.7782E+06, 4.7969E+06, 4.8156E+06,
     +  4.8345E+06, 4.8534E+06, 4.8723E+06, 4.8913E+06, 4.9104E+06,
     + 4.9295E+06, 4.9487E+06, 4.9679E+06, 4.9872E+06, 5.0066E+06,
     +  5.0260E+06, 5.0454E+06, 5.0650E+06, 5.0846E+06, 5.1042E+06,
     + 5.1239E+06, 5.1437E+06, 5.1635E+06, 5.1834E+06, 5.2034E+06,
     +  5.2234E+06, 5.2435E+06, 5.2636E+06, 5.2838E+06, 5.3040E+06,
     + 5.3243E+06, 5.3447E+06, 5.3652E+06, 5.3857E+06, 5.4062E+06,
     +  5.4268E+06, 5.4475E+06, 5.4683E+06, 5.4891E+06, 5.5100E+06,
     + 5.5309E+06, 5.5519E+06, 5.5729E+06, 5.5941E+06, 5.6153E+06,
     +  5.6365E+06, 5.6578E+06, 5.6792E+06, 5.7006E+06, 5.7221E+06,
     + 5.7437E+06, 5.7653E+06, 5.7870E+06, 5.8088E+06, 5.8306E+06,
     +  5.8525E+06, 5.8744E+06, 5.8965E+06, 5.9186E+06, 5.9407E+06,
     + 5.9629E+06, 5.9852E+06, 6.0075E+06, 6.0299E+06, 6.0524E+06,
     +  6.0750E+06, 6.0976E+06, 6.1202E+06, 6.1430E+06, 6.1658E+06,
     + 6.1886E+06, 6.2116E+06, 6.2346E+06, 6.2577E+06, 6.2808E+06,
     +  6.3040E+06, 6.3273E+06, 6.3506E+06, 6.3740E+06, 6.3975E+06,
     + 6.4211E+06, 6.4447E+06, 6.4684E+06, 6.4921E+06, 6.5159E+06,
     +  6.5398E+06, 6.5638E+06, 6.5878E+06, 6.6119E+06, 6.6361E+06,
     + 6.6603E+06, 6.6846E+06, 6.7090E+06, 6.7334E+06, 6.7579E+06,
     +  6.7825E+06, 6.8071E+06, 6.8319E+06, 6.8567E+06, 6.8815E+06,
     + 6.9065E+06, 6.9315E+06, 6.9566E+06, 6.9817E+06, 7.0069E+06,
     +  7.0322E+06, 7.0576E+06, 7.0830E+06, 7.1085E+06, 7.1341E+06,
     + 7.1598E+06, 7.1855E+06, 7.2113E+06, 7.2372E+06, 7.2631E+06,
     +  7.2891E+06, 7.3152E+06, 7.3414E+06, 7.3676E+06, 7.3939E+06,
     + 7.4203E+06, 7.4468E+06, 7.4733E+06, 7.4999E+06, 7.5266E+06,
     +  7.5534E+06, 7.5802E+06, 7.6071E+06, 7.6341E+06, 7.6611E+06,
     + 7.6883E+06, 7.7155E+06, 7.7428E+06, 7.7701E+06, 7.7976E+06,
     +  7.8251E+06, 7.8527E+06, 7.8803E+06, 7.9081E+06, 7.9359E+06,
     + 7.9638E+06, 7.9918E+06, 8.0198E+06, 8.0479E+06, 8.0761E+06,
     +  8.1044E+06, 8.1328E+06, 8.1612E+06, 8.1898E+06, 8.2183E+06,
     + 8.2470E+06, 8.2758E+06, 8.3046E+06, 8.3335E+06, 8.3625E+06,
     +  8.3916E+06, 8.4207E+06, 8.4499E+06, 8.4793E+06, 8.5086E+06,
     + 8.5381E+06, 8.5677E+06, 8.5973E+06, 8.6270E+06, 8.6568E+06,
     +  8.6867E+06, 8.7166E+06, 8.7466E+06, 8.7768E+06, 8.8070E+06,
     + 8.8372E+06, 8.8676E+06, 8.8980E+06, 8.9286E+06, 8.9592E+06,
     +  8.9898E+06, 9.0206E+06, 9.0515E+06, 9.0824E+06, 9.1134E+06,
     + 9.1445E+06, 9.1757E+06, 9.2070E+06, 9.2383E+06, 9.2698E+06,
     +  9.3013E+06, 9.3329E+06, 9.3646E+06, 9.3964E+06, 9.4282E+06,
     + 9.4602E+06, 9.4922E+06, 9.5243E+06, 9.5565E+06, 9.5888E+06,
     +  9.6212E+06, 9.6536E+06, 9.6862E+06, 9.7188E+06, 9.7515E+06,
     + 9.7843E+06, 9.8172E+06, 9.8502E+06, 9.8833E+06, 9.9164E+06,
     +  9.9497E+06, 9.9830E+06, 1.0016E+07, 1.0050E+07, 1.0083E+07,
     + 1.0117E+07, 1.0151E+07, 1.0185E+07, 1.0219E+07, 1.0253E+07,
     +  1.0287E+07, 1.0321E+07, 1.0355E+07, 1.0390E+07, 1.0424E+07,
     + 1.0459E+07, 1.0493E+07, 1.0528E+07, 1.0563E+07, 1.0598E+07,
     +  1.0633E+07, 1.0668E+07, 1.0703E+07, 1.0739E+07, 1.0774E+07,
     + 1.0810E+07, 1.0845E+07, 1.0881E+07, 1.0917E+07, 1.0952E+07,
     +  1.0988E+07, 1.1024E+07, 1.1060E+07, 1.1097E+07, 1.1133E+07,
     + 1.1169E+07, 1.1206E+07, 1.1242E+07, 1.1279E+07, 1.1316E+07,
     +  1.1353E+07, 1.1390E+07, 1.1427E+07, 1.1464E+07, 1.1501E+07,
     + 1.1539E+07, 1.1576E+07, 1.1614E+07, 1.1651E+07, 1.1689E+07,
     +  1.1727E+07, 1.1765E+07, 1.1803E+07, 1.1841E+07, 1.1879E+07,
     + 1.1917E+07, 1.1956E+07, 1.1994E+07, 1.2033E+07, 1.2072E+07,
     +  1.2110E+07, 1.2149E+07, 1.2188E+07, 1.2227E+07, 1.2267E+07,
     + 1.2306E+07, 1.2345E+07, 1.2385E+07, 1.2424E+07, 1.2464E+07,
     +  1.2504E+07, 1.2544E+07, 1.2584E+07, 1.2624E+07, 1.2664E+07,
     + 1.2704E+07, 1.2745E+07, 1.2785E+07, 1.2826E+07, 1.2866E+07,
     +  1.2907E+07, 1.2948E+07, 1.2989E+07, 1.3030E+07, 1.3071E+07,
     + 1.3113E+07, 1.3154E+07, 1.3196E+07, 1.3237E+07, 1.3279E+07,
     +  1.3321E+07, 1.3363E+07, 1.3405E+07, 1.3447E+07, 1.3489E+07,
     + 1.3531E+07, 1.3574E+07, 1.3616E+07, 1.3659E+07, 1.3702E+07,
     +  1.3745E+07, 1.3788E+07, 1.3831E+07, 1.3874E+07, 1.3917E+07,
     + 1.3960E+07, 1.4004E+07, 1.4048E+07, 1.4091E+07, 1.4135E+07,
     +  1.4179E+07, 1.4223E+07, 1.4267E+07, 1.4311E+07, 1.4356E+07,
     + 1.4400E+07, 1.4445E+07, 1.4489E+07, 1.4534E+07, 1.4579E+07,
     +  1.4624E+07, 1.4669E+07, 1.4714E+07, 1.4760E+07, 1.4805E+07,
     + 1.4851E+07, 1.4896E+07, 1.4942E+07, 1.4988E+07, 1.5034E+07,
     +  1.5080E+07, 1.5126E+07, 1.5173E+07, 1.5219E+07, 1.5266E+07,
     + 1.5312E+07, 1.5359E+07, 1.5406E+07, 1.5453E+07, 1.5500E+07,
     +  1.5547E+07, 1.5595E+07, 1.5642E+07, 1.5690E+07, 1.5737E+07,
     + 1.5785E+07, 1.5833E+07, 1.5881E+07, 1.5929E+07, 1.5978E+07,
     +  1.6026E+07, 1.6074E+07, 1.6123E+07, 1.6172E+07, 1.6220E+07,
     + 1.6269E+07, 1.6318E+07, 1.6368E+07, 1.6417E+07, 1.6466E+07,
     +  1.6516E+07, 1.6566E+07, 1.6615E+07, 1.6665E+07, 1.6715E+07,
     + 1.6765E+07, 1.6815E+07, 1.6866E+07, 1.6916E+07, 1.6967E+07,
     +  1.7018E+07, 1.7068E+07, 1.7119E+07, 1.7170E+07, 1.7222E+07,
     + 1.7273E+07, 1.7324E+07, 1.7376E+07, 1.7427E+07, 1.7479E+07,
     +  1.7531E+07, 1.7583E+07, 1.7635E+07, 1.7688E+07, 1.7740E+07,
     + 1.7793E+07, 1.7845E+07, 1.7898E+07, 1.7951E+07, 1.8004E+07,
     +  1.8057E+07, 1.8110E+07, 1.8164E+07, 1.8217E+07, 1.8271E+07,
     + 1.8324E+07, 1.8378E+07, 1.8432E+07, 1.8486E+07, 1.8541E+07,
     +  1.8595E+07, 1.8649E+07, 1.8704E+07, 1.8759E+07, 1.8814E+07,
     + 1.8869E+07, 1.8924E+07, 1.8979E+07, 1.9034E+07, 1.9090E+07,
     +  1.9145E+07, 1.9201E+07, 1.9257E+07, 1.9313E+07, 1.9369E+07,
     + 1.9426E+07, 1.9482E+07, 1.9538E+07, 1.9595E+07, 1.9652E+07,
     +  1.9709E+07, 1.9766E+07, 1.9823E+07, 1.9880E+07, 1.9938E+07,
     + 1.9995E+07, 2.0053E+07, 2.0111E+07, 2.0169E+07, 2.0227E+07,
     +  2.0285E+07, 2.0343E+07, 2.0402E+07, 2.0460E+07, 2.0519E+07,
     + 2.0578E+07, 2.0637E+07, 2.0696E+07, 2.0755E+07, 2.0815E+07,
     +  2.0874E+07, 2.0934E+07, 2.0994E+07, 2.1054E+07, 2.1114E+07,
     + 2.1174E+07, 2.1234E+07, 2.1295E+07, 2.1356E+07, 2.1416E+07,
     +  2.1477E+07, 2.1538E+07, 2.1599E+07, 2.1661E+07, 2.1722E+07,
     + 2.1784E+07, 2.1845E+07, 2.1907E+07, 2.1969E+07, 2.2031E+07,
     +  2.2094E+07, 2.2156E+07, 2.2218E+07, 2.2281E+07, 2.2344E+07,
     + 2.2407E+07, 2.2470E+07, 2.2533E+07, 2.2597E+07, 2.2660E+07,
     +  2.2724E+07, 2.2788E+07, 2.2852E+07, 2.2916E+07, 2.2980E+07,
     + 2.3044E+07, 2.3109E+07, 2.3173E+07, 2.3238E+07, 2.3303E+07,
     +  2.3368E+07, 2.3433E+07, 2.3499E+07, 2.3564E+07, 2.3630E+07,
     + 2.3695E+07, 2.3761E+07, 2.3827E+07, 2.3894E+07, 2.3960E+07,
     +  2.4027E+07, 2.4093E+07, 2.4160E+07, 2.4227E+07, 2.4294E+07,
     + 2.4361E+07, 2.4429E+07, 2.4496E+07, 2.4564E+07, 2.4632E+07,
     +  2.4700E+07, 2.4768E+07, 2.4836E+07, 2.4904E+07, 2.4973E+07,
     + 2.5042E+07, 2.5110E+07, 2.5179E+07, 2.5249E+07, 2.5318E+07,
     +  2.5387E+07, 2.5457E+07, 2.5527E+07, 2.5597E+07, 2.5667E+07,
     + 2.5737E+07, 2.5807E+07, 2.5878E+07, 2.5948E+07, 2.6019E+07,
     +  2.6090E+07, 2.6161E+07, 2.6232E+07, 2.6304E+07, 2.6375E+07,
     + 2.6447E+07, 2.6519E+07, 2.6591E+07, 2.6663E+07, 2.6736E+07,
     +  2.6808E+07, 2.6881E+07, 2.6953E+07, 2.7026E+07, 2.7100E+07,
     + 2.7173E+07, 2.7246E+07, 2.7320E+07, 2.7394E+07, 2.7467E+07,
     +  2.7541E+07, 2.7616E+07, 2.7690E+07, 2.7765E+07, 2.7839E+07,
     + 2.7914E+07, 2.7989E+07, 2.8064E+07, 2.8139E+07, 2.8215E+07,
     +  2.8291E+07, 2.8366E+07, 2.8442E+07, 2.8518E+07, 2.8595E+07,
     + 2.8671E+07, 2.8748E+07, 2.8824E+07, 2.8901E+07, 2.8978E+07,
     +  2.9055E+07, 2.9133E+07, 2.9210E+07, 2.9288E+07, 2.9366E+07,
     + 2.9444E+07, 2.9522E+07, 2.9601E+07, 2.9679E+07, 2.9758E+07,
     +  2.9837E+07, 2.9916E+07, 2.9995E+07, 3.0074E+07, 3.0154E+07,
     + 3.0233E+07, 3.0313E+07, 3.0393E+07, 3.0473E+07, 3.0554E+07,
     +  3.0634E+07, 3.0715E+07, 3.0796E+07, 3.0877E+07, 3.0958E+07,
     + 3.1039E+07, 3.1121E+07, 3.1202E+07, 3.1284E+07, 3.1366E+07,
     +  3.1448E+07, 3.1531E+07, 3.1613E+07, 3.1696E+07, 3.1779E+07,
     + 3.1862E+07, 3.1945E+07, 3.2029E+07, 3.2112E+07, 3.2196E+07,
     +  3.2280E+07, 3.2364E+07, 3.2448E+07, 3.2532E+07, 3.2617E+07,
     + 3.2702E+07, 3.2787E+07, 3.2872E+07, 3.2957E+07, 3.3042E+07,
     +  3.3128E+07, 3.3214E+07, 3.3300E+07, 3.3386E+07, 3.3472E+07,
     + 3.3559E+07, 3.3645E+07, 3.3732E+07, 3.3819E+07, 3.3906E+07,
     +  3.3994E+07, 3.4081E+07, 3.4169E+07, 3.4257E+07, 3.4345E+07,
     + 3.4433E+07, 3.4522E+07, 3.4610E+07, 3.4699E+07, 3.4788E+07,
     +  3.4877E+07, 3.4967E+07, 3.5056E+07, 3.5146E+07, 3.5236E+07,
     + 3.5326E+07, 3.5416E+07, 3.5506E+07, 3.5597E+07, 3.5688E+07,
     +  3.5779E+07, 3.5870E+07, 3.5961E+07, 3.6053E+07, 3.6144E+07,
     + 3.6236E+07, 3.6328E+07, 3.6420E+07, 3.6513E+07, 3.6605E+07,
     +  3.6698E+07, 3.6791E+07, 3.6884E+07, 3.6978E+07, 3.7071E+07,
     + 3.7165E+07, 3.7259E+07, 3.7353E+07, 3.7447E+07, 3.7542E+07,
     +  3.7636E+07, 3.7731E+07, 3.7826E+07, 3.7921E+07, 3.8017E+07,
     + 3.8112E+07, 3.8208E+07, 3.8304E+07, 3.8400E+07, 3.8497E+07,
     +  3.8593E+07, 3.8690E+07, 3.8787E+07, 3.8884E+07, 3.8981E+07,
     + 3.9079E+07, 3.9176E+07, 3.9274E+07, 3.9372E+07, 3.9471E+07,
     +  3.9569E+07, 3.9668E+07, 3.9767E+07, 3.9866E+07, 3.9965E+07,
     + 4.0064E+07, 4.0164E+07, 4.0264E+07, 4.0364E+07, 4.0464E+07,
     +  4.0564E+07, 4.0665E+07, 4.0766E+07, 4.0867E+07, 4.0968E+07,
     + 4.1069E+07, 4.1171E+07, 4.1273E+07, 4.1375E+07, 4.1477E+07,
     +  4.1579E+07, 4.1682E+07, 4.1784E+07, 4.1887E+07, 4.1991E+07,
     + 4.2094E+07, 4.2197E+07, 4.2301E+07, 4.2405E+07, 4.2509E+07,
     +  4.2614E+07, 4.2718E+07, 4.2823E+07, 4.2928E+07, 4.3033E+07,
     + 4.3139E+07, 4.3244E+07, 4.3350E+07, 4.3456E+07, 4.3562E+07,
     +  4.3668E+07, 4.3775E+07, 4.3882E+07, 4.3989E+07, 4.4096E+07,
     + 4.4203E+07, 4.4311E+07, 4.4419E+07, 4.4527E+07, 4.4635E+07,
     +  4.4744E+07, 4.4852E+07, 4.4961E+07, 4.5070E+07, 4.5179E+07,
     + 4.5289E+07, 4.5399E+07, 4.5509E+07, 4.5619E+07, 4.5729E+07,
     +  4.5840E+07, 4.5950E+07, 4.6061E+07, 4.6172E+07, 4.6284E+07,
     + 4.6395E+07, 4.6507E+07, 4.6619E+07, 4.6731E+07, 4.6844E+07,
     +  4.6956E+07, 4.7069E+07, 4.7182E+07, 4.7296E+07, 4.7409E+07,
     + 4.7523E+07, 4.7637E+07, 4.7751E+07, 4.7865E+07, 4.7980E+07,
     +  4.8095E+07, 4.8210E+07, 4.8325E+07, 4.8440E+07, 4.8556E+07,
     + 4.8672E+07, 4.8788E+07, 4.8904E+07, 4.9021E+07, 4.9137E+07,
     +  4.9254E+07, 4.9371E+07, 4.9489E+07, 4.9606E+07, 4.9724E+07,
     + 4.9842E+07, 4.9961E+07, 5.0079E+07, 5.0198E+07, 5.0317E+07,
     +  5.0436E+07, 5.0555E+07, 5.0675E+07, 5.0795E+07, 5.0915E+07,
     + 5.1035E+07, 5.1156E+07, 5.1276E+07, 5.1397E+07, 5.1518E+07,
     +  5.1640E+07, 5.1761E+07, 5.1883E+07, 5.2005E+07, 5.2128E+07,
     + 5.2250E+07, 5.2373E+07, 5.2496E+07, 5.2619E+07, 5.2742E+07,
     +  5.2866E+07, 5.2990E+07, 5.3114E+07, 5.3238E+07, 5.3363E+07,
     + 5.3488E+07, 5.3613E+07, 5.3738E+07, 5.3863E+07, 5.3989E+07,
     +  5.4115E+07, 5.4241E+07, 5.4368E+07, 5.4494E+07, 5.4621E+07,
     + 5.4748E+07, 5.4876E+07, 5.5003E+07, 5.5131E+07, 5.5259E+07,
     +  5.5387E+07, 5.5516E+07, 5.5644E+07, 5.5773E+07, 5.5903E+07,
     + 5.6032E+07, 5.6162E+07, 5.6292E+07, 5.6422E+07, 5.6552E+07,
     +  5.6683E+07, 5.6814E+07, 5.6945E+07, 5.7076E+07, 5.7208E+07,
     + 5.7339E+07, 5.7471E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 4,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       676
      data (QofT( 5,J),J=1,2408)/ 4.8130E+00, 1.2413E+01, 2.2082E+01,
     +  3.3456E+01, 4.6308E+01, 6.0485E+01, 7.5874E+01, 9.2386E+01,
     + 1.0995E+02, 1.2850E+02, 1.4800E+02, 1.6839E+02, 1.8965E+02,
     +  2.1173E+02, 2.3460E+02, 2.5825E+02, 2.8264E+02, 3.0776E+02,
     + 3.3358E+02, 3.6009E+02, 4.1509E+02, 4.7264E+02, 5.3263E+02,
     +  5.9497E+02, 6.5957E+02, 7.2636E+02, 7.9526E+02, 8.6622E+02,
     + 9.3918E+02, 1.0141E+03, 1.0909E+03, 1.1695E+03, 1.2500E+03,
     +  1.3322E+03, 1.4162E+03, 1.5018E+03, 1.5891E+03, 1.6781E+03,
     + 1.7687E+03, 1.8608E+03, 1.9545E+03, 2.0497E+03, 2.1464E+03,
     +  2.2446E+03, 2.3443E+03, 2.4454E+03, 2.5479E+03, 2.6518E+03,
     + 2.7571E+03, 2.8638E+03, 2.9718E+03, 3.0812E+03, 3.1918E+03,
     +  3.3038E+03, 3.4170E+03, 3.5316E+03, 3.6474E+03, 3.7644E+03,
     + 3.8827E+03, 4.0022E+03, 4.1229E+03, 4.2448E+03, 4.3679E+03,
     +  4.4922E+03, 4.6177E+03, 4.7444E+03, 4.8722E+03, 5.0011E+03,
     + 5.1313E+03, 5.2625E+03, 5.3949E+03, 5.5284E+03, 5.6630E+03,
     +  5.7988E+03, 5.9356E+03, 6.0736E+03, 6.2127E+03, 6.3529E+03,
     + 6.4941E+03, 6.6365E+03, 6.7800E+03, 6.9245E+03, 7.0702E+03,
     +  7.2170E+03, 7.3647E+03, 7.5137E+03, 7.6637E+03, 7.8148E+03,
     + 7.9670E+03, 8.1202E+03, 8.2746E+03, 8.4300E+03, 8.5866E+03,
     +  8.7443E+03, 8.9030E+03, 9.0629E+03, 9.2239E+03, 9.3860E+03,
     + 9.5491E+03, 9.7134E+03, 9.8789E+03, 1.0045E+04, 1.0213E+04,
     +  1.0382E+04, 1.0552E+04, 1.0723E+04, 1.0895E+04, 1.1069E+04,
     + 1.1243E+04, 1.1419E+04, 1.1596E+04, 1.1774E+04, 1.1954E+04,
     +  1.2134E+04, 1.2316E+04, 1.2499E+04, 1.2683E+04, 1.2868E+04,
     + 1.3055E+04, 1.3243E+04, 1.3432E+04, 1.3622E+04, 1.3814E+04,
     +  1.4007E+04, 1.4201E+04, 1.4397E+04, 1.4594E+04, 1.4792E+04,
     + 1.4991E+04, 1.5192E+04, 1.5395E+04, 1.5598E+04, 1.5803E+04,
     +  1.6009E+04, 1.6217E+04, 1.6426E+04, 1.6637E+04, 1.6849E+04,
     + 1.7062E+04, 1.7277E+04, 1.7493E+04, 1.7711E+04, 1.7930E+04,
     +  1.8151E+04, 1.8373E+04, 1.8597E+04, 1.8822E+04, 1.9049E+04,
     + 1.9278E+04, 1.9508E+04, 1.9739E+04, 1.9972E+04, 2.0207E+04,
     +  2.0443E+04, 2.0681E+04, 2.0921E+04, 2.1162E+04, 2.1405E+04,
     + 2.1649E+04, 2.1896E+04, 2.2144E+04, 2.2393E+04, 2.2645E+04,
     +  2.2898E+04, 2.3153E+04, 2.3409E+04, 2.3668E+04, 2.3928E+04,
     + 2.4190E+04, 2.4454E+04, 2.4719E+04, 2.4987E+04, 2.5256E+04,
     +  2.5527E+04, 2.5801E+04, 2.6076E+04, 2.6353E+04, 2.6631E+04,
     + 2.6912E+04, 2.7195E+04, 2.7480E+04, 2.7766E+04, 2.8055E+04,
     +  2.8346E+04, 2.8638E+04, 2.8933E+04, 2.9230E+04, 2.9529E+04,
     + 2.9830E+04, 3.0133E+04, 3.0438E+04, 3.0745E+04, 3.1054E+04,
     +  3.1366E+04, 3.1679E+04, 3.1995E+04, 3.2314E+04, 3.2634E+04,
     + 3.2956E+04, 3.3281E+04, 3.3608E+04, 3.3937E+04, 3.4269E+04,
     +  3.4602E+04, 3.4939E+04, 3.5277E+04, 3.5618E+04, 3.5961E+04,
     + 3.6306E+04, 3.6654E+04, 3.7005E+04, 3.7357E+04, 3.7713E+04,
     +  3.8070E+04, 3.8430E+04, 3.8793E+04, 3.9158E+04, 3.9525E+04,
     + 3.9895E+04, 4.0268E+04, 4.0643E+04, 4.1021E+04, 4.1401E+04,
     +  4.1784E+04, 4.2169E+04, 4.2557E+04, 4.2948E+04, 4.3342E+04,
     + 4.3738E+04, 4.4137E+04, 4.4538E+04, 4.4943E+04, 4.5349E+04,
     +  4.5759E+04, 4.6172E+04, 4.6587E+04, 4.7005E+04, 4.7426E+04,
     + 4.7850E+04, 4.8277E+04, 4.8706E+04, 4.9139E+04, 4.9574E+04,
     +  5.0012E+04, 5.0453E+04, 5.0898E+04, 5.1345E+04, 5.1795E+04,
     + 5.2248E+04, 5.2704E+04, 5.3163E+04, 5.3625E+04, 5.4091E+04,
     +  5.4559E+04, 5.5031E+04, 5.5505E+04, 5.5983E+04, 5.6464E+04,
     + 5.6948E+04, 5.7435E+04, 5.7925E+04, 5.8419E+04, 5.8916E+04,
     +  5.9416E+04, 5.9920E+04, 6.0427E+04, 6.0937E+04, 6.1450E+04,
     + 6.1967E+04, 6.2487E+04, 6.3010E+04, 6.3537E+04, 6.4068E+04,
     +  6.4602E+04, 6.5139E+04, 6.5680E+04, 6.6224E+04, 6.6772E+04,
     + 6.7323E+04, 6.7877E+04, 6.8436E+04, 6.8998E+04, 6.9563E+04,
     +  7.0132E+04, 7.0705E+04, 7.1281E+04, 7.1862E+04, 7.2445E+04,
     + 7.3033E+04, 7.3624E+04, 7.4219E+04, 7.4817E+04, 7.5420E+04,
     +  7.6026E+04, 7.6637E+04, 7.7251E+04, 7.7868E+04, 7.8490E+04,
     + 7.9116E+04, 7.9745E+04, 8.0379E+04, 8.1016E+04, 8.1658E+04,
     +  8.2303E+04, 8.2953E+04, 8.3606E+04, 8.4263E+04, 8.4925E+04,
     + 8.5590E+04, 8.6260E+04, 8.6934E+04, 8.7612E+04, 8.8294E+04,
     +  8.8981E+04, 8.9671E+04, 9.0366E+04, 9.1066E+04, 9.1769E+04,
     + 9.2476E+04, 9.3189E+04, 9.3905E+04, 9.4625E+04, 9.5350E+04,
     +  9.6080E+04, 9.6814E+04, 9.7552E+04, 9.8294E+04, 9.9041E+04,
     + 9.9793E+04, 1.0055E+05, 1.0131E+05, 1.0208E+05, 1.0285E+05,
     +  1.0362E+05, 1.0440E+05, 1.0518E+05, 1.0597E+05, 1.0676E+05,
     + 1.0756E+05, 1.0836E+05, 1.0917E+05, 1.0998E+05, 1.1080E+05,
     +  1.1162E+05, 1.1245E+05, 1.1328E+05, 1.1412E+05, 1.1496E+05,
     + 1.1580E+05, 1.1665E+05, 1.1751E+05, 1.1837E+05, 1.1924E+05,
     +  1.2011E+05, 1.2098E+05, 1.2186E+05, 1.2275E+05, 1.2364E+05,
     + 1.2454E+05, 1.2544E+05, 1.2635E+05, 1.2726E+05, 1.2817E+05,
     +  1.2910E+05, 1.3002E+05, 1.3096E+05, 1.3190E+05, 1.3284E+05,
     + 1.3379E+05, 1.3474E+05, 1.3570E+05, 1.3667E+05, 1.3764E+05,
     +  1.3861E+05, 1.3960E+05, 1.4058E+05, 1.4158E+05, 1.4257E+05,
     + 1.4358E+05, 1.4459E+05, 1.4560E+05, 1.4662E+05, 1.4765E+05,
     +  1.4868E+05, 1.4972E+05, 1.5076E+05, 1.5181E+05, 1.5287E+05,
     + 1.5393E+05, 1.5500E+05, 1.5607E+05, 1.5715E+05, 1.5823E+05,
     +  1.5932E+05, 1.6042E+05, 1.6152E+05, 1.6263E+05, 1.6375E+05,
     + 1.6487E+05, 1.6599E+05, 1.6713E+05, 1.6827E+05, 1.6941E+05,
     +  1.7056E+05, 1.7172E+05, 1.7288E+05, 1.7405E+05, 1.7523E+05,
     + 1.7641E+05, 1.7760E+05, 1.7880E+05, 1.8000E+05, 1.8121E+05,
     +  1.8242E+05, 1.8365E+05, 1.8487E+05, 1.8611E+05, 1.8735E+05,
     + 1.8860E+05, 1.8985E+05, 1.9111E+05, 1.9238E+05, 1.9365E+05,
     +  1.9494E+05, 1.9622E+05, 1.9752E+05, 1.9882E+05, 2.0013E+05,
     + 2.0144E+05, 2.0276E+05, 2.0409E+05, 2.0543E+05, 2.0677E+05,
     +  2.0812E+05, 2.0948E+05, 2.1084E+05, 2.1221E+05, 2.1359E+05,
     + 2.1497E+05, 2.1636E+05, 2.1776E+05, 2.1917E+05, 2.2058E+05,
     +  2.2200E+05, 2.2343E+05, 2.2487E+05, 2.2631E+05, 2.2776E+05,
     + 2.2922E+05, 2.3068E+05, 2.3215E+05, 2.3363E+05, 2.3512E+05,
     +  2.3662E+05, 2.3812E+05, 2.3963E+05, 2.4115E+05, 2.4267E+05,
     + 2.4420E+05, 2.4574E+05, 2.4729E+05, 2.4885E+05, 2.5041E+05,
     +  2.5198E+05, 2.5356E+05, 2.5515E+05, 2.5675E+05, 2.5835E+05,
     + 2.5996E+05, 2.6158E+05, 2.6321E+05, 2.6484E+05, 2.6648E+05,
     +  2.6814E+05, 2.6979E+05, 2.7146E+05, 2.7314E+05, 2.7482E+05,
     + 2.7651E+05, 2.7821E+05, 2.7992E+05, 2.8164E+05, 2.8337E+05,
     +  2.8510E+05, 2.8684E+05, 2.8859E+05, 2.9035E+05, 2.9212E+05,
     + 2.9390E+05, 2.9568E+05, 2.9748E+05, 2.9928E+05, 3.0109E+05,
     +  3.0291E+05, 3.0474E+05, 3.0657E+05, 3.0842E+05, 3.1027E+05,
     + 3.1214E+05, 3.1401E+05, 3.1589E+05, 3.1778E+05, 3.1968E+05,
     +  3.2159E+05, 3.2351E+05, 3.2543E+05, 3.2737E+05, 3.2931E+05,
     + 3.3127E+05, 3.3323E+05, 3.3520E+05, 3.3718E+05, 3.3917E+05,
     +  3.4117E+05, 3.4318E+05, 3.4520E+05, 3.4723E+05, 3.4927E+05,
     + 3.5132E+05, 3.5337E+05, 3.5544E+05, 3.5751E+05, 3.5960E+05,
     +  3.6169E+05, 3.6380E+05, 3.6591E+05, 3.6804E+05, 3.7017E+05,
     + 3.7232E+05, 3.7447E+05, 3.7663E+05, 3.7881E+05, 3.8099E+05,
     +  3.8319E+05, 3.8539E+05, 3.8760E+05, 3.8983E+05, 3.9206E+05,
     + 3.9430E+05, 3.9656E+05, 3.9882E+05, 4.0110E+05, 4.0338E+05,
     +  4.0568E+05, 4.0798E+05, 4.1030E+05, 4.1263E+05, 4.1496E+05,
     + 4.1731E+05, 4.1967E+05, 4.2204E+05, 4.2441E+05, 4.2680E+05,
     +  4.2921E+05, 4.3162E+05, 4.3404E+05, 4.3647E+05, 4.3891E+05,
     + 4.4137E+05, 4.4383E+05, 4.4631E+05, 4.4880E+05, 4.5129E+05,
     +  4.5380E+05, 4.5632E+05, 4.5885E+05, 4.6139E+05, 4.6395E+05,
     + 4.6651E+05, 4.6909E+05, 4.7168E+05, 4.7427E+05, 4.7688E+05,
     +  4.7950E+05, 4.8214E+05, 4.8478E+05, 4.8744E+05, 4.9010E+05,
     + 4.9278E+05, 4.9547E+05, 4.9817E+05, 5.0089E+05, 5.0361E+05,
     +  5.0635E+05, 5.0910E+05, 5.1186E+05, 5.1463E+05, 5.1741E+05,
     + 5.2021E+05, 5.2302E+05, 5.2584E+05, 5.2867E+05, 5.3151E+05,
     +  5.3437E+05, 5.3724E+05, 5.4012E+05, 5.4301E+05, 5.4592E+05,
     + 5.4883E+05, 5.5176E+05, 5.5471E+05, 5.5766E+05, 5.6063E+05,
     +  5.6361E+05, 5.6660E+05, 5.6960E+05, 5.7262E+05, 5.7565E+05,
     + 5.7869E+05, 5.8175E+05, 5.8482E+05, 5.8790E+05, 5.9099E+05,
     +  5.9410E+05, 5.9722E+05, 6.0035E+05, 6.0350E+05, 6.0665E+05,
     + 6.0982E+05, 6.1301E+05, 6.1621E+05, 6.1942E+05, 6.2264E+05,
     +  6.2588E+05, 6.2913E+05, 6.3240E+05, 6.3567E+05, 6.3896E+05,
     + 6.4227E+05, 6.4559E+05, 6.4892E+05, 6.5226E+05, 6.5562E+05,
     +  6.5899E+05, 6.6238E+05, 6.6578E+05, 6.6919E+05, 6.7262E+05,
     + 6.7606E+05, 6.7952E+05, 6.8299E+05, 6.8647E+05, 6.8997E+05,
     +  6.9348E+05, 6.9701E+05, 7.0055E+05, 7.0410E+05, 7.0767E+05,
     + 7.1125E+05, 7.1485E+05, 7.1846E+05, 7.2208E+05, 7.2572E+05,
     +  7.2938E+05, 7.3305E+05, 7.3673E+05, 7.4043E+05, 7.4414E+05,
     + 7.4787E+05, 7.5161E+05, 7.5537E+05, 7.5914E+05, 7.6293E+05,
     +  7.6673E+05, 7.7055E+05, 7.7438E+05, 7.7823E+05, 7.8209E+05,
     + 7.8597E+05, 7.8986E+05, 7.9377E+05, 7.9769E+05, 8.0163E+05,
     +  8.0559E+05, 8.0955E+05, 8.1354E+05, 8.1754E+05, 8.2155E+05,
     + 8.2558E+05, 8.2963E+05, 8.3369E+05, 8.3777E+05, 8.4187E+05,
     +  8.4598E+05, 8.5010E+05, 8.5424E+05, 8.5840E+05, 8.6257E+05,
     + 8.6676E+05, 8.7097E+05, 8.7519E+05, 8.7943E+05, 8.8368E+05,
     +  8.8795E+05, 8.9224E+05, 8.9654E+05, 9.0086E+05, 9.0520E+05,
     + 9.0955E+05, 9.1392E+05, 9.1830E+05, 9.2271E+05, 9.2712E+05,
     +  9.3156E+05, 9.3601E+05, 9.4048E+05, 9.4496E+05, 9.4947E+05,
     + 9.5399E+05, 9.5852E+05, 9.6308E+05, 9.6765E+05, 9.7224E+05,
     +  9.7684E+05, 9.8146E+05, 9.8610E+05, 9.9076E+05, 9.9543E+05,
     + 1.0001E+06, 1.0048E+06, 1.0096E+06, 1.0143E+06, 1.0191E+06,
     +  1.0238E+06, 1.0286E+06, 1.0335E+06, 1.0383E+06, 1.0431E+06,
     + 1.0480E+06, 1.0529E+06, 1.0578E+06, 1.0627E+06, 1.0677E+06,
     +  1.0726E+06, 1.0776E+06, 1.0826E+06, 1.0876E+06, 1.0926E+06,
     + 1.0977E+06, 1.1027E+06, 1.1078E+06, 1.1129E+06, 1.1181E+06,
     +  1.1232E+06, 1.1284E+06, 1.1335E+06, 1.1387E+06, 1.1440E+06,
     + 1.1492E+06, 1.1545E+06, 1.1597E+06, 1.1650E+06, 1.1703E+06,
     +  1.1757E+06, 1.1810E+06, 1.1864E+06, 1.1918E+06, 1.1972E+06,
     + 1.2026E+06, 1.2080E+06, 1.2135E+06, 1.2190E+06, 1.2245E+06,
     +  1.2300E+06, 1.2356E+06, 1.2411E+06, 1.2467E+06, 1.2523E+06,
     + 1.2580E+06, 1.2636E+06, 1.2693E+06, 1.2749E+06, 1.2806E+06,
     +  1.2864E+06, 1.2921E+06, 1.2979E+06, 1.3037E+06, 1.3095E+06,
     + 1.3153E+06, 1.3211E+06, 1.3270E+06, 1.3329E+06, 1.3388E+06,
     +  1.3447E+06, 1.3507E+06, 1.3566E+06, 1.3626E+06, 1.3686E+06,
     + 1.3747E+06, 1.3807E+06, 1.3868E+06, 1.3929E+06, 1.3990E+06,
     +  1.4051E+06, 1.4113E+06, 1.4175E+06, 1.4237E+06, 1.4299E+06,
     + 1.4361E+06, 1.4424E+06, 1.4487E+06, 1.4550E+06, 1.4613E+06,
     +  1.4677E+06, 1.4741E+06, 1.4805E+06, 1.4869E+06, 1.4933E+06,
     + 1.4998E+06, 1.5063E+06, 1.5128E+06, 1.5193E+06, 1.5258E+06,
     +  1.5324E+06, 1.5390E+06, 1.5456E+06, 1.5523E+06, 1.5589E+06,
     + 1.5656E+06, 1.5723E+06, 1.5790E+06, 1.5858E+06, 1.5926E+06,
     +  1.5994E+06, 1.6062E+06, 1.6130E+06, 1.6199E+06, 1.6268E+06,
     + 1.6337E+06, 1.6406E+06, 1.6476E+06, 1.6546E+06, 1.6616E+06,
     +  1.6686E+06, 1.6756E+06, 1.6827E+06, 1.6898E+06, 1.6969E+06,
     + 1.7041E+06, 1.7113E+06, 1.7184E+06, 1.7257E+06, 1.7329E+06,
     +  1.7402E+06, 1.7475E+06, 1.7548E+06, 1.7621E+06, 1.7695E+06,
     + 1.7769E+06, 1.7843E+06, 1.7917E+06, 1.7992E+06, 1.8066E+06,
     +  1.8141E+06, 1.8217E+06, 1.8292E+06, 1.8368E+06, 1.8444E+06,
     + 1.8521E+06, 1.8597E+06, 1.8674E+06, 1.8751E+06, 1.8828E+06,
     +  1.8906E+06, 1.8984E+06, 1.9062E+06, 1.9140E+06, 1.9219E+06,
     + 1.9297E+06, 1.9376E+06, 1.9456E+06, 1.9535E+06, 1.9615E+06,
     +  1.9695E+06, 1.9776E+06, 1.9856E+06, 1.9937E+06, 2.0018E+06,
     + 2.0100E+06, 2.0181E+06, 2.0263E+06, 2.0345E+06, 2.0428E+06,
     +  2.0511E+06, 2.0594E+06, 2.0677E+06, 2.0760E+06, 2.0844E+06,
     + 2.0928E+06, 2.1012E+06, 2.1097E+06, 2.1182E+06, 2.1267E+06,
     +  2.1352E+06, 2.1438E+06, 2.1524E+06, 2.1610E+06, 2.1696E+06,
     + 2.1783E+06, 2.1870E+06, 2.1957E+06, 2.2045E+06, 2.2133E+06,
     +  2.2221E+06, 2.2309E+06, 2.2398E+06, 2.2487E+06, 2.2576E+06,
     + 2.2665E+06, 2.2755E+06, 2.2845E+06, 2.2936E+06, 2.3026E+06,
     +  2.3117E+06, 2.3208E+06, 2.3300E+06, 2.3391E+06, 2.3483E+06,
     + 2.3576E+06, 2.3668E+06, 2.3761E+06, 2.3854E+06, 2.3948E+06,
     +  2.4042E+06, 2.4136E+06, 2.4230E+06, 2.4324E+06, 2.4419E+06,
     + 2.4515E+06, 2.4610E+06, 2.4706E+06, 2.4802E+06, 2.4898E+06,
     +  2.4995E+06, 2.5092E+06, 2.5189E+06, 2.5287E+06, 2.5384E+06,
     + 2.5483E+06, 2.5581E+06, 2.5680E+06, 2.5779E+06, 2.5878E+06,
     +  2.5978E+06, 2.6078E+06, 2.6178E+06, 2.6278E+06, 2.6379E+06,
     + 2.6480E+06, 2.6582E+06, 2.6684E+06, 2.6786E+06, 2.6888E+06,
     +  2.6991E+06, 2.7094E+06, 2.7197E+06, 2.7301E+06, 2.7405E+06,
     + 2.7509E+06, 2.7613E+06, 2.7718E+06, 2.7823E+06, 2.7929E+06,
     +  2.8035E+06, 2.8141E+06, 2.8247E+06, 2.8354E+06, 2.8461E+06,
     + 2.8568E+06, 2.8676E+06, 2.8784E+06, 2.8892E+06, 2.9001E+06,
     +  2.9110E+06, 2.9219E+06, 2.9329E+06, 2.9439E+06, 2.9549E+06,
     + 2.9660E+06, 2.9771E+06, 2.9882E+06, 2.9994E+06, 3.0105E+06,
     +  3.0218E+06, 3.0330E+06, 3.0443E+06, 3.0556E+06, 3.0670E+06,
     + 3.0784E+06, 3.0898E+06, 3.1013E+06, 3.1127E+06, 3.1243E+06,
     +  3.1358E+06, 3.1474E+06, 3.1590E+06, 3.1707E+06, 3.1824E+06,
     + 3.1941E+06, 3.2059E+06, 3.2177E+06, 3.2295E+06, 3.2413E+06,
     +  3.2532E+06, 3.2652E+06, 3.2771E+06, 3.2891E+06, 3.3012E+06,
     + 3.3132E+06, 3.3253E+06, 3.3375E+06, 3.3496E+06, 3.3619E+06,
     +  3.3741E+06, 3.3864E+06, 3.3987E+06, 3.4110E+06, 3.4234E+06,
     + 3.4358E+06, 3.4483E+06, 3.4608E+06, 3.4733E+06, 3.4859E+06,
     +  3.4985E+06, 3.5111E+06, 3.5238E+06, 3.5365E+06, 3.5492E+06,
     + 3.5620E+06, 3.5748E+06, 3.5876E+06, 3.6005E+06, 3.6134E+06,
     +  3.6264E+06, 3.6394E+06, 3.6524E+06, 3.6655E+06, 3.6786E+06,
     + 3.6917E+06, 3.7049E+06, 3.7181E+06, 3.7314E+06, 3.7447E+06,
     +  3.7580E+06, 3.7713E+06, 3.7847E+06, 3.7982E+06, 3.8117E+06,
     + 3.8252E+06, 3.8387E+06, 3.8523E+06, 3.8659E+06, 3.8796E+06,
     +  3.8933E+06, 3.9070E+06, 3.9208E+06, 3.9346E+06, 3.9485E+06,
     + 3.9624E+06, 3.9763E+06, 3.9903E+06, 4.0043E+06, 4.0183E+06,
     +  4.0324E+06, 4.0466E+06, 4.0607E+06, 4.0749E+06, 4.0892E+06,
     + 4.1034E+06, 4.1178E+06, 4.1321E+06, 4.1465E+06, 4.1610E+06,
     +  4.1754E+06, 4.1900E+06, 4.2045E+06, 4.2191E+06, 4.2337E+06,
     + 4.2484E+06, 4.2631E+06, 4.2779E+06, 4.2927E+06, 4.3075E+06,
     +  4.3224E+06, 4.3373E+06, 4.3523E+06, 4.3673E+06, 4.3823E+06,
     + 4.3974E+06, 4.4125E+06, 4.4277E+06, 4.4429E+06, 4.4581E+06,
     +  4.4734E+06, 4.4887E+06, 4.5041E+06, 4.5195E+06, 4.5350E+06,
     + 4.5505E+06, 4.5660E+06, 4.5816E+06, 4.5972E+06, 4.6128E+06,
     +  4.6286E+06, 4.6443E+06, 4.6601E+06, 4.6759E+06, 4.6918E+06,
     + 4.7077E+06, 4.7236E+06, 4.7396E+06, 4.7557E+06, 4.7718E+06,
     +  4.7879E+06, 4.8040E+06, 4.8203E+06, 4.8365E+06, 4.8528E+06,
     + 4.8691E+06, 4.8855E+06, 4.9020E+06, 4.9184E+06, 4.9349E+06,
     +  4.9515E+06, 4.9681E+06, 4.9847E+06, 5.0014E+06, 5.0182E+06,
     + 5.0349E+06, 5.0518E+06, 5.0686E+06, 5.0855E+06, 5.1025E+06,
     +  5.1195E+06, 5.1365E+06, 5.1536E+06, 5.1707E+06, 5.1879E+06,
     + 5.2051E+06, 5.2224E+06, 5.2397E+06, 5.2571E+06, 5.2745E+06,
     +  5.2919E+06, 5.3094E+06, 5.3269E+06, 5.3445E+06, 5.3621E+06,
     + 5.3798E+06, 5.3975E+06, 5.4153E+06, 5.4331E+06, 5.4510E+06,
     +  5.4689E+06, 5.4868E+06, 5.5048E+06, 5.5229E+06, 5.5410E+06,
     + 5.5591E+06, 5.5773E+06, 5.5955E+06, 5.6138E+06, 5.6321E+06,
     +  5.6505E+06, 5.6689E+06, 5.6873E+06, 5.7059E+06, 5.7244E+06,
     + 5.7430E+06, 5.7617E+06, 5.7804E+06, 5.7991E+06, 5.8179E+06,
     +  5.8368E+06, 5.8557E+06, 5.8746E+06, 5.8936E+06, 5.9126E+06,
     + 5.9317E+06, 5.9508E+06, 5.9700E+06, 5.9893E+06, 6.0085E+06,
     +  6.0279E+06, 6.0472E+06, 6.0667E+06, 6.0861E+06, 6.1057E+06,
     + 6.1252E+06, 6.1449E+06, 6.1645E+06, 6.1843E+06, 6.2040E+06,
     +  6.2238E+06, 6.2437E+06, 6.2636E+06, 6.2836E+06, 6.3036E+06,
     + 6.3237E+06, 6.3438E+06, 6.3640E+06, 6.3842E+06, 6.4045E+06,
     +  6.4248E+06, 6.4452E+06, 6.4656E+06, 6.4861E+06, 6.5066E+06,
     + 6.5272E+06, 6.5478E+06, 6.5685E+06, 6.5892E+06, 6.6100E+06,
     +  6.6308E+06, 6.6517E+06, 6.6727E+06, 6.6936E+06, 6.7147E+06,
     + 6.7358E+06, 6.7569E+06, 6.7781E+06, 6.7994E+06, 6.8207E+06,
     +  6.8420E+06, 6.8634E+06, 6.8849E+06, 6.9064E+06, 6.9280E+06,
     + 6.9496E+06, 6.9713E+06, 6.9930E+06, 7.0148E+06, 7.0366E+06,
     +  7.0585E+06, 7.0804E+06, 7.1024E+06, 7.1244E+06, 7.1465E+06,
     + 7.1687E+06, 7.1909E+06, 7.2132E+06, 7.2355E+06, 7.2578E+06,
     +  7.2803E+06, 7.3027E+06, 7.3253E+06, 7.3479E+06, 7.3705E+06,
     + 7.3932E+06, 7.4159E+06, 7.4387E+06, 7.4616E+06, 7.4845E+06,
     +  7.5075E+06, 7.5305E+06, 7.5536E+06, 7.5767E+06, 7.5999E+06,
     + 7.6232E+06, 7.6465E+06, 7.6699E+06, 7.6933E+06, 7.7167E+06,
     +  7.7403E+06, 7.7639E+06, 7.7875E+06, 7.8112E+06, 7.8350E+06,
     + 7.8588E+06, 7.8826E+06, 7.9066E+06, 7.9305E+06, 7.9546E+06,
     +  7.9787E+06, 8.0028E+06, 8.0271E+06, 8.0513E+06, 8.0757E+06,
     + 8.1000E+06, 8.1245E+06, 8.1490E+06, 8.1735E+06, 8.1982E+06,
     +  8.2228E+06, 8.2476E+06, 8.2724E+06, 8.2972E+06, 8.3221E+06,
     + 8.3471E+06, 8.3721E+06, 8.3972E+06, 8.4224E+06, 8.4476E+06,
     +  8.4728E+06, 8.4981E+06, 8.5235E+06, 8.5490E+06, 8.5745E+06,
     + 8.6000E+06, 8.6257E+06, 8.6513E+06, 8.6771E+06, 8.7029E+06,
     +  8.7288E+06, 8.7547E+06, 8.7807E+06, 8.8067E+06, 8.8328E+06,
     + 8.8590E+06, 8.8852E+06, 8.9115E+06, 8.9378E+06, 8.9642E+06,
     +  8.9907E+06, 9.0172E+06, 9.0438E+06, 9.0705E+06, 9.0972E+06,
     + 9.1240E+06, 9.1508E+06, 9.1777E+06, 9.2047E+06, 9.2317E+06,
     +  9.2588E+06, 9.2860E+06, 9.3132E+06, 9.3405E+06, 9.3678E+06,
     + 9.3952E+06, 9.4227E+06, 9.4502E+06, 9.4778E+06, 9.5054E+06,
     +  9.5332E+06, 9.5610E+06, 9.5888E+06, 9.6167E+06, 9.6447E+06,
     + 9.6727E+06, 9.7008E+06, 9.7290E+06, 9.7572E+06, 9.7855E+06,
     +  9.8139E+06, 9.8423E+06, 9.8708E+06, 9.8993E+06, 9.9280E+06,
     + 9.9566E+06, 9.9854E+06, 1.0014E+07, 1.0043E+07, 1.0072E+07,
     +  1.0101E+07, 1.0130E+07, 1.0159E+07, 1.0188E+07, 1.0218E+07,
     + 1.0247E+07, 1.0276E+07, 1.0306E+07, 1.0335E+07, 1.0365E+07,
     +  1.0395E+07, 1.0424E+07, 1.0454E+07, 1.0484E+07, 1.0514E+07,
     + 1.0544E+07, 1.0574E+07, 1.0604E+07, 1.0635E+07, 1.0665E+07,
     +  1.0695E+07, 1.0726E+07, 1.0756E+07, 1.0787E+07, 1.0817E+07,
     + 1.0848E+07, 1.0879E+07, 1.0909E+07, 1.0940E+07, 1.0971E+07,
     +  1.1002E+07, 1.1033E+07, 1.1065E+07, 1.1096E+07, 1.1127E+07,
     + 1.1159E+07, 1.1190E+07, 1.1221E+07, 1.1253E+07, 1.1285E+07,
     +  1.1316E+07, 1.1348E+07, 1.1380E+07, 1.1412E+07, 1.1444E+07,
     + 1.1476E+07, 1.1508E+07, 1.1540E+07, 1.1573E+07, 1.1605E+07,
     +  1.1638E+07, 1.1670E+07, 1.1703E+07, 1.1735E+07, 1.1768E+07,
     + 1.1801E+07, 1.1834E+07, 1.1867E+07, 1.1900E+07, 1.1933E+07,
     +  1.1966E+07, 1.1999E+07, 1.2032E+07, 1.2066E+07, 1.2099E+07,
     + 1.2133E+07, 1.2166E+07, 1.2200E+07, 1.2233E+07, 1.2267E+07,
     +  1.2301E+07, 1.2335E+07, 1.2369E+07, 1.2403E+07, 1.2437E+07,
     + 1.2472E+07, 1.2506E+07, 1.2540E+07, 1.2575E+07, 1.2609E+07,
     +  1.2644E+07, 1.2679E+07, 1.2713E+07, 1.2748E+07, 1.2783E+07,
     + 1.2818E+07, 1.2853E+07, 1.2888E+07, 1.2924E+07, 1.2959E+07,
     +  1.2994E+07, 1.3030E+07, 1.3065E+07, 1.3101E+07, 1.3136E+07,
     + 1.3172E+07, 1.3208E+07, 1.3244E+07, 1.3280E+07, 1.3316E+07,
     +  1.3352E+07, 1.3388E+07, 1.3425E+07, 1.3461E+07, 1.3497E+07,
     + 1.3534E+07, 1.3570E+07, 1.3607E+07, 1.3644E+07, 1.3681E+07,
     +  1.3718E+07, 1.3755E+07, 1.3792E+07, 1.3829E+07, 1.3866E+07,
     + 1.3903E+07, 1.3941E+07, 1.3978E+07, 1.4016E+07, 1.4053E+07,
     +  1.4091E+07, 1.4129E+07, 1.4167E+07, 1.4204E+07, 1.4242E+07,
     + 1.4281E+07, 1.4319E+07, 1.4357E+07, 1.4395E+07, 1.4434E+07,
     +  1.4472E+07, 1.4511E+07, 1.4549E+07, 1.4588E+07, 1.4627E+07,
     + 1.4666E+07, 1.4705E+07, 1.4744E+07, 1.4783E+07, 1.4822E+07,
     +  1.4862E+07, 1.4901E+07, 1.4940E+07, 1.4980E+07, 1.5020E+07,
     + 1.5059E+07, 1.5099E+07, 1.5139E+07, 1.5179E+07, 1.5219E+07,
     +  1.5259E+07, 1.5299E+07, 1.5340E+07, 1.5380E+07, 1.5421E+07,
     + 1.5461E+07, 1.5502E+07, 1.5542E+07, 1.5583E+07, 1.5624E+07,
     +  1.5665E+07, 1.5706E+07, 1.5747E+07, 1.5788E+07, 1.5830E+07,
     + 1.5871E+07, 1.5913E+07, 1.5954E+07, 1.5996E+07, 1.6038E+07,
     +  1.6079E+07, 1.6121E+07, 1.6163E+07, 1.6205E+07, 1.6247E+07,
     + 1.6290E+07, 1.6332E+07, 1.6374E+07, 1.6417E+07, 1.6460E+07,
     +  1.6502E+07, 1.6545E+07, 1.6588E+07, 1.6631E+07, 1.6674E+07,
     + 1.6717E+07, 1.6760E+07, 1.6803E+07, 1.6847E+07, 1.6890E+07,
     +  1.6934E+07, 1.6977E+07, 1.7021E+07, 1.7065E+07, 1.7109E+07,
     + 1.7153E+07, 1.7197E+07, 1.7241E+07, 1.7285E+07, 1.7330E+07,
     +  1.7374E+07, 1.7419E+07, 1.7463E+07, 1.7508E+07, 1.7553E+07,
     + 1.7598E+07, 1.7643E+07, 1.7688E+07, 1.7733E+07, 1.7778E+07,
     +  1.7824E+07, 1.7869E+07, 1.7915E+07, 1.7960E+07, 1.8006E+07,
     + 1.8052E+07, 1.8098E+07, 1.8144E+07, 1.8190E+07, 1.8236E+07,
     +  1.8282E+07, 1.8328E+07, 1.8375E+07, 1.8421E+07, 1.8468E+07,
     + 1.8515E+07, 1.8561E+07, 1.8608E+07, 1.8655E+07, 1.8702E+07,
     +  1.8750E+07, 1.8797E+07, 1.8844E+07, 1.8892E+07, 1.8939E+07,
     + 1.8987E+07, 1.9035E+07, 1.9083E+07, 1.9131E+07, 1.9179E+07,
     +  1.9227E+07, 1.9275E+07, 1.9323E+07, 1.9372E+07, 1.9420E+07,
     + 1.9469E+07, 1.9517E+07, 1.9566E+07, 1.9615E+07, 1.9664E+07,
     +  1.9713E+07, 1.9762E+07, 1.9812E+07, 1.9861E+07, 1.9910E+07,
     + 1.9960E+07, 2.0010E+07, 2.0059E+07, 2.0109E+07, 2.0159E+07,
     +  2.0209E+07, 2.0259E+07, 2.0310E+07, 2.0360E+07, 2.0410E+07,
     + 2.0461E+07, 2.0512E+07, 2.0562E+07, 2.0613E+07, 2.0664E+07,
     +  2.0715E+07, 2.0766E+07, 2.0818E+07, 2.0869E+07, 2.0920E+07,
     + 2.0972E+07, 2.1023E+07, 2.1075E+07, 2.1127E+07, 2.1179E+07,
     +  2.1231E+07, 2.1283E+07, 2.1335E+07, 2.1388E+07, 2.1440E+07,
     + 2.1493E+07, 2.1545E+07, 2.1598E+07, 2.1651E+07, 2.1704E+07,
     +  2.1757E+07, 2.1810E+07, 2.1863E+07, 2.1917E+07, 2.1970E+07,
     + 2.2024E+07, 2.2077E+07, 2.2131E+07, 2.2185E+07, 2.2239E+07,
     +  2.2293E+07, 2.2347E+07, 2.2401E+07, 2.2456E+07, 2.2510E+07,
     + 2.2565E+07, 2.2619E+07, 2.2674E+07, 2.2729E+07, 2.2784E+07,
     +  2.2839E+07, 2.2894E+07, 2.2950E+07, 2.3005E+07, 2.3061E+07,
     + 2.3116E+07, 2.3172E+07, 2.3228E+07, 2.3284E+07, 2.3340E+07,
     +  2.3396E+07, 2.3452E+07, 2.3509E+07, 2.3565E+07, 2.3622E+07,
     + 2.3678E+07, 2.3735E+07, 2.3792E+07, 2.3849E+07, 2.3906E+07,
     +  2.3964E+07, 2.4021E+07, 2.4078E+07, 2.4136E+07, 2.4194E+07,
     + 2.4251E+07, 2.4309E+07, 2.4367E+07, 2.4425E+07, 2.4483E+07,
     +  2.4542E+07, 2.4600E+07, 2.4659E+07, 2.4717E+07, 2.4776E+07,
     + 2.4835E+07, 2.4894E+07, 2.4953E+07, 2.5012E+07, 2.5071E+07,
     +  2.5131E+07, 2.5190E+07, 2.5250E+07, 2.5310E+07, 2.5370E+07,
     + 2.5430E+07, 2.5490E+07, 2.5550E+07, 2.5610E+07, 2.5671E+07,
     +  2.5731E+07, 2.5792E+07, 2.5852E+07, 2.5913E+07, 2.5974E+07,
     + 2.6035E+07, 2.6097E+07, 2.6158E+07, 2.6219E+07, 2.6281E+07,
     +  2.6343E+07, 2.6404E+07, 2.6466E+07, 2.6528E+07, 2.6590E+07,
     + 2.6652E+07, 2.6715E+07, 2.6777E+07, 2.6840E+07, 2.6902E+07,
     +  2.6965E+07, 2.7028E+07, 2.7091E+07, 2.7154E+07, 2.7218E+07,
     + 2.7281E+07, 2.7344E+07, 2.7408E+07, 2.7472E+07, 2.7536E+07,
     +  2.7600E+07, 2.7664E+07, 2.7728E+07, 2.7792E+07, 2.7857E+07,
     + 2.7921E+07, 2.7986E+07, 2.8051E+07, 2.8115E+07, 2.8180E+07,
     +  2.8246E+07, 2.8311E+07, 2.8376E+07, 2.8442E+07, 2.8507E+07,
     + 2.8573E+07, 2.8639E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 5,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       886
      data (QofT( 6,J),J=1,2408)/ 1.7326E+00, 4.4883E+00, 7.9991E+00,
     +  1.2130E+01, 1.6799E+01, 2.1950E+01, 2.7542E+01, 3.3542E+01,
     + 3.9925E+01, 4.6668E+01, 5.3754E+01, 6.1166E+01, 6.8890E+01,
     +  7.6916E+01, 8.5231E+01, 9.3827E+01, 1.0269E+02, 1.1182E+02,
     + 1.2121E+02, 1.3084E+02, 1.5084E+02, 1.7176E+02, 1.9356E+02,
     +  2.1623E+02, 2.3971E+02, 2.6399E+02, 2.8904E+02, 3.1483E+02,
     + 3.4135E+02, 3.6858E+02, 3.9650E+02, 4.2509E+02, 4.5434E+02,
     +  4.8423E+02, 5.1475E+02, 5.4589E+02, 5.7764E+02, 6.0998E+02,
     + 6.4290E+02, 6.7639E+02, 7.1046E+02, 7.4507E+02, 7.8023E+02,
     +  8.1593E+02, 8.5216E+02, 8.8892E+02, 9.2619E+02, 9.6397E+02,
     + 1.0022E+03, 1.0410E+03, 1.0803E+03, 1.1201E+03, 1.1603E+03,
     +  1.2010E+03, 1.2422E+03, 1.2838E+03, 1.3259E+03, 1.3685E+03,
     + 1.4115E+03, 1.4549E+03, 1.4988E+03, 1.5431E+03, 1.5879E+03,
     +  1.6331E+03, 1.6787E+03, 1.7248E+03, 1.7712E+03, 1.8181E+03,
     + 1.8655E+03, 1.9132E+03, 1.9613E+03, 2.0099E+03, 2.0589E+03,
     +  2.1082E+03, 2.1580E+03, 2.2082E+03, 2.2588E+03, 2.3098E+03,
     + 2.3612E+03, 2.4130E+03, 2.4652E+03, 2.5179E+03, 2.5709E+03,
     +  2.6243E+03, 2.6781E+03, 2.7323E+03, 2.7869E+03, 2.8420E+03,
     + 2.8974E+03, 2.9532E+03, 3.0094E+03, 3.0661E+03, 3.1231E+03,
     +  3.1806E+03, 3.2384E+03, 3.2967E+03, 3.3554E+03, 3.4145E+03,
     + 3.4740E+03, 3.5339E+03, 3.5943E+03, 3.6550E+03, 3.7162E+03,
     +  3.7778E+03, 3.8399E+03, 3.9023E+03, 3.9652E+03, 4.0285E+03,
     + 4.0923E+03, 4.1565E+03, 4.2211E+03, 4.2862E+03, 4.3518E+03,
     +  4.4178E+03, 4.4842E+03, 4.5511E+03, 4.6185E+03, 4.6863E+03,
     + 4.7546E+03, 4.8234E+03, 4.8926E+03, 4.9623E+03, 5.0325E+03,
     +  5.1032E+03, 5.1744E+03, 5.2460E+03, 5.3182E+03, 5.3908E+03,
     + 5.4640E+03, 5.5377E+03, 5.6118E+03, 5.6865E+03, 5.7617E+03,
     +  5.8375E+03, 5.9137E+03, 5.9905E+03, 6.0678E+03, 6.1457E+03,
     + 6.2241E+03, 6.3030E+03, 6.3826E+03, 6.4626E+03, 6.5433E+03,
     +  6.6245E+03, 6.7063E+03, 6.7886E+03, 6.8715E+03, 6.9550E+03,
     + 7.0391E+03, 7.1238E+03, 7.2090E+03, 7.2949E+03, 7.3815E+03,
     +  7.4686E+03, 7.5563E+03, 7.6446E+03, 7.7337E+03, 7.8232E+03,
     + 7.9135E+03, 8.0044E+03, 8.0959E+03, 8.1881E+03, 8.2809E+03,
     +  8.3744E+03, 8.4686E+03, 8.5634E+03, 8.6589E+03, 8.7552E+03,
     + 8.8521E+03, 8.9496E+03, 9.0478E+03, 9.1468E+03, 9.2465E+03,
     +  9.3469E+03, 9.4480E+03, 9.5499E+03, 9.6525E+03, 9.7558E+03,
     + 9.8598E+03, 9.9645E+03, 1.0070E+04, 1.0176E+04, 1.0283E+04,
     +  1.0391E+04, 1.0500E+04, 1.0609E+04, 1.0719E+04, 1.0830E+04,
     + 1.0942E+04, 1.1054E+04, 1.1168E+04, 1.1282E+04, 1.1397E+04,
     +  1.1513E+04, 1.1629E+04, 1.1747E+04, 1.1865E+04, 1.1984E+04,
     + 1.2104E+04, 1.2225E+04, 1.2346E+04, 1.2469E+04, 1.2592E+04,
     +  1.2716E+04, 1.2842E+04, 1.2968E+04, 1.3095E+04, 1.3223E+04,
     + 1.3351E+04, 1.3481E+04, 1.3611E+04, 1.3743E+04, 1.3875E+04,
     +  1.4009E+04, 1.4143E+04, 1.4278E+04, 1.4414E+04, 1.4552E+04,
     + 1.4690E+04, 1.4829E+04, 1.4969E+04, 1.5110E+04, 1.5252E+04,
     +  1.5395E+04, 1.5539E+04, 1.5684E+04, 1.5830E+04, 1.5977E+04,
     + 1.6125E+04, 1.6274E+04, 1.6424E+04, 1.6575E+04, 1.6728E+04,
     +  1.6881E+04, 1.7035E+04, 1.7191E+04, 1.7347E+04, 1.7505E+04,
     + 1.7663E+04, 1.7823E+04, 1.7984E+04, 1.8146E+04, 1.8309E+04,
     +  1.8473E+04, 1.8638E+04, 1.8805E+04, 1.8972E+04, 1.9141E+04,
     + 1.9311E+04, 1.9482E+04, 1.9654E+04, 1.9827E+04, 2.0002E+04,
     +  2.0178E+04, 2.0355E+04, 2.0532E+04, 2.0712E+04, 2.0892E+04,
     + 2.1074E+04, 2.1257E+04, 2.1441E+04, 2.1626E+04, 2.1813E+04,
     +  2.2001E+04, 2.2190E+04, 2.2380E+04, 2.2572E+04, 2.2765E+04,
     + 2.2959E+04, 2.3155E+04, 2.3351E+04, 2.3550E+04, 2.3749E+04,
     +  2.3950E+04, 2.4152E+04, 2.4355E+04, 2.4560E+04, 2.4766E+04,
     + 2.4973E+04, 2.5182E+04, 2.5392E+04, 2.5604E+04, 2.5817E+04,
     +  2.6031E+04, 2.6247E+04, 2.6464E+04, 2.6682E+04, 2.6902E+04,
     + 2.7123E+04, 2.7346E+04, 2.7570E+04, 2.7796E+04, 2.8023E+04,
     +  2.8251E+04, 2.8481E+04, 2.8713E+04, 2.8946E+04, 2.9180E+04,
     + 2.9416E+04, 2.9653E+04, 2.9892E+04, 3.0133E+04, 3.0375E+04,
     +  3.0618E+04, 3.0863E+04, 3.1110E+04, 3.1358E+04, 3.1608E+04,
     + 3.1859E+04, 3.2112E+04, 3.2366E+04, 3.2622E+04, 3.2880E+04,
     +  3.3139E+04, 3.3400E+04, 3.3662E+04, 3.3927E+04, 3.4192E+04,
     + 3.4459E+04, 3.4729E+04, 3.4999E+04, 3.5272E+04, 3.5546E+04,
     +  3.5821E+04, 3.6099E+04, 3.6378E+04, 3.6659E+04, 3.6941E+04,
     + 3.7226E+04, 3.7512E+04, 3.7799E+04, 3.8089E+04, 3.8380E+04,
     +  3.8673E+04, 3.8968E+04, 3.9265E+04, 3.9563E+04, 3.9863E+04,
     + 4.0165E+04, 4.0469E+04, 4.0774E+04, 4.1082E+04, 4.1391E+04,
     +  4.1702E+04, 4.2015E+04, 4.2330E+04, 4.2647E+04, 4.2965E+04,
     + 4.3286E+04, 4.3608E+04, 4.3933E+04, 4.4259E+04, 4.4587E+04,
     +  4.4917E+04, 4.5249E+04, 4.5583E+04, 4.5919E+04, 4.6257E+04,
     + 4.6596E+04, 4.6938E+04, 4.7282E+04, 4.7628E+04, 4.7976E+04,
     +  4.8326E+04, 4.8678E+04, 4.9032E+04, 4.9388E+04, 4.9746E+04,
     + 5.0106E+04, 5.0468E+04, 5.0832E+04, 5.1198E+04, 5.1567E+04,
     +  5.1937E+04, 5.2310E+04, 5.2685E+04, 5.3062E+04, 5.3441E+04,
     + 5.3822E+04, 5.4206E+04, 5.4591E+04, 5.4979E+04, 5.5369E+04,
     +  5.5761E+04, 5.6155E+04, 5.6551E+04, 5.6950E+04, 5.7351E+04,
     + 5.7755E+04, 5.8160E+04, 5.8568E+04, 5.8978E+04, 5.9390E+04,
     +  5.9805E+04, 6.0222E+04, 6.0641E+04, 6.1063E+04, 6.1486E+04,
     + 6.1913E+04, 6.2341E+04, 6.2772E+04, 6.3206E+04, 6.3641E+04,
     +  6.4079E+04, 6.4520E+04, 6.4963E+04, 6.5408E+04, 6.5856E+04,
     + 6.6306E+04, 6.6758E+04, 6.7213E+04, 6.7671E+04, 6.8131E+04,
     +  6.8593E+04, 6.9058E+04, 6.9526E+04, 6.9996E+04, 7.0469E+04,
     + 7.0944E+04, 7.1421E+04, 7.1901E+04, 7.2384E+04, 7.2870E+04,
     +  7.3357E+04, 7.3848E+04, 7.4341E+04, 7.4837E+04, 7.5335E+04,
     + 7.5836E+04, 7.6340E+04, 7.6846E+04, 7.7355E+04, 7.7867E+04,
     +  7.8381E+04, 7.8898E+04, 7.9418E+04, 7.9940E+04, 8.0466E+04,
     + 8.0994E+04, 8.1524E+04, 8.2058E+04, 8.2594E+04, 8.3133E+04,
     +  8.3675E+04, 8.4219E+04, 8.4767E+04, 8.5317E+04, 8.5871E+04,
     + 8.6427E+04, 8.6985E+04, 8.7547E+04, 8.8112E+04, 8.8679E+04,
     +  8.9249E+04, 8.9823E+04, 9.0399E+04, 9.0978E+04, 9.1561E+04,
     + 9.2145E+04, 9.2733E+04, 9.3324E+04, 9.3919E+04, 9.4516E+04,
     +  9.5116E+04, 9.5719E+04, 9.6325E+04, 9.6934E+04, 9.7546E+04,
     + 9.8161E+04, 9.8780E+04, 9.9401E+04, 1.0003E+05, 1.0065E+05,
     +  1.0128E+05, 1.0192E+05, 1.0256E+05, 1.0320E+05, 1.0384E+05,
     + 1.0449E+05, 1.0514E+05, 1.0579E+05, 1.0644E+05, 1.0710E+05,
     +  1.0777E+05, 1.0843E+05, 1.0910E+05, 1.0978E+05, 1.1045E+05,
     + 1.1113E+05, 1.1181E+05, 1.1250E+05, 1.1319E+05, 1.1388E+05,
     +  1.1458E+05, 1.1527E+05, 1.1598E+05, 1.1668E+05, 1.1739E+05,
     + 1.1811E+05, 1.1882E+05, 1.1954E+05, 1.2026E+05, 1.2099E+05,
     +  1.2172E+05, 1.2245E+05, 1.2319E+05, 1.2393E+05, 1.2468E+05,
     + 1.2542E+05, 1.2618E+05, 1.2693E+05, 1.2769E+05, 1.2845E+05,
     +  1.2922E+05, 1.2999E+05, 1.3076E+05, 1.3154E+05, 1.3232E+05,
     + 1.3310E+05, 1.3389E+05, 1.3468E+05, 1.3547E+05, 1.3627E+05,
     +  1.3707E+05, 1.3788E+05, 1.3869E+05, 1.3950E+05, 1.4032E+05,
     + 1.4114E+05, 1.4197E+05, 1.4280E+05, 1.4363E+05, 1.4447E+05,
     +  1.4531E+05, 1.4615E+05, 1.4700E+05, 1.4785E+05, 1.4871E+05,
     + 1.4957E+05, 1.5043E+05, 1.5130E+05, 1.5217E+05, 1.5305E+05,
     +  1.5393E+05, 1.5481E+05, 1.5570E+05, 1.5659E+05, 1.5748E+05,
     + 1.5838E+05, 1.5929E+05, 1.6020E+05, 1.6111E+05, 1.6203E+05,
     +  1.6295E+05, 1.6387E+05, 1.6480E+05, 1.6573E+05, 1.6667E+05,
     + 1.6761E+05, 1.6856E+05, 1.6951E+05, 1.7046E+05, 1.7142E+05,
     +  1.7238E+05, 1.7335E+05, 1.7432E+05, 1.7529E+05, 1.7627E+05,
     + 1.7726E+05, 1.7825E+05, 1.7924E+05, 1.8024E+05, 1.8124E+05,
     +  1.8224E+05, 1.8325E+05, 1.8427E+05, 1.8529E+05, 1.8631E+05,
     + 1.8734E+05, 1.8837E+05, 1.8941E+05, 1.9045E+05, 1.9150E+05,
     +  1.9255E+05, 1.9360E+05, 1.9466E+05, 1.9573E+05, 1.9680E+05,
     + 1.9787E+05, 1.9895E+05, 2.0003E+05, 2.0112E+05, 2.0221E+05,
     +  2.0331E+05, 2.0441E+05, 2.0552E+05, 2.0663E+05, 2.0774E+05,
     + 2.0886E+05, 2.0999E+05, 2.1112E+05, 2.1226E+05, 2.1340E+05,
     +  2.1454E+05, 2.1569E+05, 2.1684E+05, 2.1800E+05, 2.1917E+05,
     + 2.2034E+05, 2.2151E+05, 2.2269E+05, 2.2388E+05, 2.2506E+05,
     +  2.2626E+05, 2.2746E+05, 2.2866E+05, 2.2987E+05, 2.3108E+05,
     + 2.3230E+05, 2.3353E+05, 2.3476E+05, 2.3599E+05, 2.3723E+05,
     +  2.3848E+05, 2.3973E+05, 2.4098E+05, 2.4224E+05, 2.4351E+05,
     + 2.4478E+05, 2.4606E+05, 2.4734E+05, 2.4862E+05, 2.4992E+05,
     +  2.5121E+05, 2.5251E+05, 2.5382E+05, 2.5514E+05, 2.5645E+05,
     + 2.5778E+05, 2.5911E+05, 2.6044E+05, 2.6178E+05, 2.6313E+05,
     +  2.6448E+05, 2.6584E+05, 2.6720E+05, 2.6857E+05, 2.6994E+05,
     + 2.7132E+05, 2.7270E+05, 2.7409E+05, 2.7549E+05, 2.7689E+05,
     +  2.7830E+05, 2.7971E+05, 2.8113E+05, 2.8255E+05, 2.8398E+05,
     + 2.8541E+05, 2.8685E+05, 2.8830E+05, 2.8975E+05, 2.9121E+05,
     +  2.9267E+05, 2.9415E+05, 2.9562E+05, 2.9710E+05, 2.9859E+05,
     + 3.0008E+05, 3.0158E+05, 3.0309E+05, 3.0460E+05, 3.0611E+05,
     +  3.0764E+05, 3.0917E+05, 3.1070E+05, 3.1224E+05, 3.1379E+05,
     + 3.1534E+05, 3.1690E+05, 3.1846E+05, 3.2004E+05, 3.2161E+05,
     +  3.2320E+05, 3.2479E+05, 3.2638E+05, 3.2798E+05, 3.2959E+05,
     + 3.3121E+05, 3.3283E+05, 3.3445E+05, 3.3609E+05, 3.3772E+05,
     +  3.3937E+05, 3.4102E+05, 3.4268E+05, 3.4435E+05, 3.4602E+05,
     + 3.4769E+05, 3.4938E+05, 3.5107E+05, 3.5277E+05, 3.5447E+05,
     +  3.5618E+05, 3.5789E+05, 3.5962E+05, 3.6135E+05, 3.6308E+05,
     + 3.6483E+05, 3.6657E+05, 3.6833E+05, 3.7009E+05, 3.7186E+05,
     +  3.7364E+05, 3.7542E+05, 3.7721E+05, 3.7900E+05, 3.8081E+05,
     + 3.8262E+05, 3.8443E+05, 3.8625E+05, 3.8808E+05, 3.8992E+05,
     +  3.9176E+05, 3.9361E+05, 3.9547E+05, 3.9734E+05, 3.9921E+05,
     + 4.0108E+05, 4.0297E+05, 4.0486E+05, 4.0676E+05, 4.0867E+05,
     +  4.1058E+05, 4.1250E+05, 4.1443E+05, 4.1636E+05, 4.1830E+05,
     + 4.2025E+05, 4.2221E+05, 4.2417E+05, 4.2614E+05, 4.2812E+05,
     +  4.3010E+05, 4.3209E+05, 4.3409E+05, 4.3610E+05, 4.3811E+05,
     + 4.4013E+05, 4.4216E+05, 4.4420E+05, 4.4624E+05, 4.4829E+05,
     +  4.5035E+05, 4.5241E+05, 4.5449E+05, 4.5657E+05, 4.5865E+05,
     + 4.6075E+05, 4.6285E+05, 4.6496E+05, 4.6708E+05, 4.6921E+05,
     +  4.7134E+05, 4.7348E+05, 4.7563E+05, 4.7778E+05, 4.7995E+05,
     + 4.8212E+05, 4.8430E+05, 4.8649E+05, 4.8868E+05, 4.9088E+05,
     +  4.9309E+05, 4.9531E+05, 4.9754E+05, 4.9977E+05, 5.0202E+05,
     + 5.0427E+05, 5.0652E+05, 5.0879E+05, 5.1106E+05, 5.1335E+05,
     +  5.1564E+05, 5.1793E+05, 5.2024E+05, 5.2255E+05, 5.2488E+05,
     + 5.2721E+05, 5.2954E+05, 5.3189E+05, 5.3425E+05, 5.3661E+05,
     +  5.3898E+05, 5.4136E+05, 5.4375E+05, 5.4614E+05, 5.4855E+05,
     + 5.5096E+05, 5.5338E+05, 5.5581E+05, 5.5825E+05, 5.6070E+05,
     +  5.6315E+05, 5.6562E+05, 5.6809E+05, 5.7057E+05, 5.7306E+05,
     + 5.7555E+05, 5.7806E+05, 5.8057E+05, 5.8310E+05, 5.8563E+05,
     +  5.8817E+05, 5.9072E+05, 5.9328E+05, 5.9584E+05, 5.9842E+05,
     + 6.0100E+05, 6.0360E+05, 6.0620E+05, 6.0881E+05, 6.1143E+05,
     +  6.1406E+05, 6.1669E+05, 6.1934E+05, 6.2199E+05, 6.2466E+05,
     + 6.2733E+05, 6.3001E+05, 6.3270E+05, 6.3540E+05, 6.3811E+05,
     +  6.4083E+05, 6.4356E+05, 6.4629E+05, 6.4904E+05, 6.5180E+05,
     + 6.5456E+05, 6.5733E+05, 6.6012E+05, 6.6291E+05, 6.6571E+05,
     +  6.6852E+05, 6.7134E+05, 6.7417E+05, 6.7701E+05, 6.7985E+05,
     + 6.8271E+05, 6.8558E+05, 6.8845E+05, 6.9134E+05, 6.9424E+05,
     +  6.9714E+05, 7.0005E+05, 7.0298E+05, 7.0591E+05, 7.0886E+05,
     + 7.1181E+05, 7.1477E+05, 7.1774E+05, 7.2073E+05, 7.2372E+05,
     +  7.2672E+05, 7.2973E+05, 7.3275E+05, 7.3578E+05, 7.3882E+05,
     + 7.4187E+05, 7.4494E+05, 7.4801E+05, 7.5109E+05, 7.5418E+05,
     +  7.5728E+05, 7.6039E+05, 7.6351E+05, 7.6664E+05, 7.6978E+05,
     + 7.7293E+05, 7.7609E+05, 7.7926E+05, 7.8245E+05, 7.8564E+05,
     +  7.8884E+05, 7.9205E+05, 7.9527E+05, 7.9851E+05, 8.0175E+05,
     + 8.0501E+05, 8.0827E+05, 8.1154E+05, 8.1483E+05, 8.1812E+05,
     +  8.2143E+05, 8.2475E+05, 8.2807E+05, 8.3141E+05, 8.3476E+05,
     + 8.3812E+05, 8.4149E+05, 8.4487E+05, 8.4826E+05, 8.5166E+05,
     +  8.5507E+05, 8.5849E+05, 8.6193E+05, 8.6537E+05, 8.6883E+05,
     + 8.7229E+05, 8.7577E+05, 8.7926E+05, 8.8276E+05, 8.8627E+05,
     +  8.8979E+05, 8.9332E+05, 8.9686E+05, 9.0041E+05, 9.0398E+05,
     + 9.0756E+05, 9.1114E+05, 9.1474E+05, 9.1835E+05, 9.2197E+05,
     +  9.2560E+05, 9.2925E+05, 9.3290E+05, 9.3657E+05, 9.4024E+05,
     + 9.4393E+05, 9.4763E+05, 9.5134E+05, 9.5507E+05, 9.5880E+05,
     +  9.6255E+05, 9.6630E+05, 9.7007E+05, 9.7385E+05, 9.7764E+05,
     + 9.8145E+05, 9.8526E+05, 9.8909E+05, 9.9293E+05, 9.9678E+05,
     +  1.0006E+06, 1.0045E+06, 1.0084E+06, 1.0123E+06, 1.0162E+06,
     + 1.0201E+06, 1.0241E+06, 1.0280E+06, 1.0320E+06, 1.0359E+06,
     +  1.0399E+06, 1.0439E+06, 1.0479E+06, 1.0519E+06, 1.0560E+06,
     + 1.0600E+06, 1.0640E+06, 1.0681E+06, 1.0722E+06, 1.0763E+06,
     +  1.0804E+06, 1.0845E+06, 1.0886E+06, 1.0928E+06, 1.0969E+06,
     + 1.1011E+06, 1.1053E+06, 1.1094E+06, 1.1136E+06, 1.1179E+06,
     +  1.1221E+06, 1.1263E+06, 1.1306E+06, 1.1348E+06, 1.1391E+06,
     + 1.1434E+06, 1.1477E+06, 1.1520E+06, 1.1564E+06, 1.1607E+06,
     +  1.1650E+06, 1.1694E+06, 1.1738E+06, 1.1782E+06, 1.1826E+06,
     + 1.1870E+06, 1.1914E+06, 1.1959E+06, 1.2003E+06, 1.2048E+06,
     +  1.2093E+06, 1.2138E+06, 1.2183E+06, 1.2228E+06, 1.2274E+06,
     + 1.2319E+06, 1.2365E+06, 1.2410E+06, 1.2456E+06, 1.2502E+06,
     +  1.2548E+06, 1.2595E+06, 1.2641E+06, 1.2688E+06, 1.2734E+06,
     + 1.2781E+06, 1.2828E+06, 1.2875E+06, 1.2923E+06, 1.2970E+06,
     +  1.3017E+06, 1.3065E+06, 1.3113E+06, 1.3161E+06, 1.3209E+06,
     + 1.3257E+06, 1.3305E+06, 1.3354E+06, 1.3402E+06, 1.3451E+06,
     +  1.3500E+06, 1.3549E+06, 1.3598E+06, 1.3648E+06, 1.3697E+06,
     + 1.3747E+06, 1.3796E+06, 1.3846E+06, 1.3896E+06, 1.3946E+06,
     +  1.3997E+06, 1.4047E+06, 1.4098E+06, 1.4148E+06, 1.4199E+06,
     + 1.4250E+06, 1.4302E+06, 1.4353E+06, 1.4404E+06, 1.4456E+06,
     +  1.4508E+06, 1.4559E+06, 1.4612E+06, 1.4664E+06, 1.4716E+06,
     + 1.4768E+06, 1.4821E+06, 1.4874E+06, 1.4927E+06, 1.4980E+06,
     +  1.5033E+06, 1.5086E+06, 1.5140E+06, 1.5194E+06, 1.5247E+06,
     + 1.5301E+06, 1.5355E+06, 1.5410E+06, 1.5464E+06, 1.5519E+06,
     +  1.5573E+06, 1.5628E+06, 1.5683E+06, 1.5738E+06, 1.5794E+06,
     + 1.5849E+06, 1.5905E+06, 1.5961E+06, 1.6016E+06, 1.6073E+06,
     +  1.6129E+06, 1.6185E+06, 1.6242E+06, 1.6298E+06, 1.6355E+06,
     + 1.6412E+06, 1.6469E+06, 1.6527E+06, 1.6584E+06, 1.6642E+06,
     +  1.6700E+06, 1.6758E+06, 1.6816E+06, 1.6874E+06, 1.6932E+06,
     + 1.6991E+06, 1.7050E+06, 1.7109E+06, 1.7168E+06, 1.7227E+06,
     +  1.7286E+06, 1.7346E+06, 1.7406E+06, 1.7466E+06, 1.7526E+06,
     + 1.7586E+06, 1.7646E+06, 1.7707E+06, 1.7767E+06, 1.7828E+06,
     +  1.7889E+06, 1.7950E+06, 1.8012E+06, 1.8073E+06, 1.8135E+06,
     + 1.8197E+06, 1.8259E+06, 1.8321E+06, 1.8383E+06, 1.8446E+06,
     +  1.8508E+06, 1.8571E+06, 1.8634E+06, 1.8697E+06, 1.8761E+06,
     + 1.8824E+06, 1.8888E+06, 1.8952E+06, 1.9016E+06, 1.9080E+06,
     +  1.9144E+06, 1.9209E+06, 1.9273E+06, 1.9338E+06, 1.9403E+06,
     + 1.9469E+06, 1.9534E+06, 1.9600E+06, 1.9665E+06, 1.9731E+06,
     +  1.9797E+06, 1.9863E+06, 1.9930E+06, 1.9997E+06, 2.0063E+06,
     + 2.0130E+06, 2.0197E+06, 2.0265E+06, 2.0332E+06, 2.0400E+06,
     +  2.0468E+06, 2.0536E+06, 2.0604E+06, 2.0672E+06, 2.0741E+06,
     + 2.0809E+06, 2.0878E+06, 2.0947E+06, 2.1017E+06, 2.1086E+06,
     +  2.1156E+06, 2.1226E+06, 2.1295E+06, 2.1366E+06, 2.1436E+06,
     + 2.1507E+06, 2.1577E+06, 2.1648E+06, 2.1719E+06, 2.1790E+06,
     +  2.1862E+06, 2.1933E+06, 2.2005E+06, 2.2077E+06, 2.2149E+06,
     + 2.2222E+06, 2.2294E+06, 2.2367E+06, 2.2440E+06, 2.2513E+06,
     +  2.2586E+06, 2.2660E+06, 2.2734E+06, 2.2807E+06, 2.2881E+06,
     + 2.2956E+06, 2.3030E+06, 2.3105E+06, 2.3180E+06, 2.3255E+06,
     +  2.3330E+06, 2.3405E+06, 2.3481E+06, 2.3556E+06, 2.3632E+06,
     + 2.3709E+06, 2.3785E+06, 2.3861E+06, 2.3938E+06, 2.4015E+06,
     +  2.4092E+06, 2.4169E+06, 2.4247E+06, 2.4325E+06, 2.4403E+06,
     + 2.4481E+06, 2.4559E+06, 2.4637E+06, 2.4716E+06, 2.4795E+06,
     +  2.4874E+06, 2.4953E+06, 2.5033E+06, 2.5112E+06, 2.5192E+06,
     + 2.5272E+06, 2.5353E+06, 2.5433E+06, 2.5514E+06, 2.5595E+06,
     +  2.5676E+06, 2.5757E+06, 2.5838E+06, 2.5920E+06, 2.6002E+06,
     + 2.6084E+06, 2.6166E+06, 2.6249E+06, 2.6332E+06, 2.6414E+06,
     +  2.6498E+06, 2.6581E+06, 2.6664E+06, 2.6748E+06, 2.6832E+06,
     + 2.6916E+06, 2.7001E+06, 2.7085E+06, 2.7170E+06, 2.7255E+06,
     +  2.7340E+06, 2.7425E+06, 2.7511E+06, 2.7597E+06, 2.7683E+06,
     + 2.7769E+06, 2.7855E+06, 2.7942E+06, 2.8029E+06, 2.8116E+06,
     +  2.8203E+06, 2.8291E+06, 2.8378E+06, 2.8466E+06, 2.8555E+06,
     + 2.8643E+06, 2.8731E+06, 2.8820E+06, 2.8909E+06, 2.8998E+06,
     +  2.9088E+06, 2.9178E+06, 2.9267E+06, 2.9357E+06, 2.9448E+06,
     + 2.9538E+06, 2.9629E+06, 2.9720E+06, 2.9811E+06, 2.9903E+06,
     +  2.9994E+06, 3.0086E+06, 3.0178E+06, 3.0270E+06, 3.0363E+06,
     + 3.0455E+06, 3.0548E+06, 3.0642E+06, 3.0735E+06, 3.0829E+06,
     +  3.0922E+06, 3.1016E+06, 3.1111E+06, 3.1205E+06, 3.1300E+06,
     + 3.1395E+06, 3.1490E+06, 3.1586E+06, 3.1681E+06, 3.1777E+06,
     +  3.1873E+06, 3.1970E+06, 3.2066E+06, 3.2163E+06, 3.2260E+06,
     + 3.2357E+06, 3.2455E+06, 3.2552E+06, 3.2650E+06, 3.2748E+06,
     +  3.2847E+06, 3.2945E+06, 3.3044E+06, 3.3143E+06, 3.3243E+06,
     + 3.3342E+06, 3.3442E+06, 3.3542E+06, 3.3642E+06, 3.3743E+06,
     +  3.3844E+06, 3.3945E+06, 3.4046E+06, 3.4147E+06, 3.4249E+06,
     + 3.4351E+06, 3.4453E+06, 3.4555E+06, 3.4658E+06, 3.4761E+06,
     +  3.4864E+06, 3.4967E+06, 3.5071E+06, 3.5175E+06, 3.5279E+06,
     + 3.5383E+06, 3.5488E+06, 3.5593E+06, 3.5698E+06, 3.5803E+06,
     +  3.5909E+06, 3.6014E+06, 3.6120E+06, 3.6227E+06, 3.6333E+06,
     + 3.6440E+06, 3.6547E+06, 3.6654E+06, 3.6762E+06, 3.6869E+06,
     +  3.6977E+06, 3.7086E+06, 3.7194E+06, 3.7303E+06, 3.7412E+06,
     + 3.7521E+06, 3.7631E+06, 3.7740E+06, 3.7850E+06, 3.7961E+06,
     +  3.8071E+06, 3.8182E+06, 3.8293E+06, 3.8404E+06, 3.8516E+06,
     + 3.8628E+06, 3.8740E+06, 3.8852E+06, 3.8964E+06, 3.9077E+06,
     +  3.9190E+06, 3.9304E+06, 3.9417E+06, 3.9531E+06, 3.9645E+06,
     + 3.9759E+06, 3.9874E+06, 3.9989E+06, 4.0104E+06, 4.0219E+06,
     +  4.0335E+06, 4.0451E+06, 4.0567E+06, 4.0683E+06, 4.0800E+06,
     + 4.0917E+06, 4.1034E+06, 4.1152E+06, 4.1270E+06, 4.1388E+06,
     +  4.1506E+06, 4.1624E+06, 4.1743E+06, 4.1862E+06, 4.1982E+06,
     + 4.2101E+06, 4.2221E+06, 4.2341E+06, 4.2462E+06, 4.2582E+06,
     +  4.2703E+06, 4.2824E+06, 4.2946E+06, 4.3068E+06, 4.3190E+06,
     + 4.3312E+06, 4.3435E+06, 4.3557E+06, 4.3681E+06, 4.3804E+06,
     +  4.3928E+06, 4.4052E+06, 4.4176E+06, 4.4300E+06, 4.4425E+06,
     + 4.4550E+06, 4.4675E+06, 4.4801E+06, 4.4927E+06, 4.5053E+06,
     +  4.5179E+06, 4.5306E+06, 4.5433E+06, 4.5560E+06, 4.5688E+06,
     + 4.5816E+06, 4.5944E+06, 4.6072E+06, 4.6201E+06, 4.6330E+06,
     +  4.6459E+06, 4.6589E+06, 4.6718E+06, 4.6849E+06, 4.6979E+06,
     + 4.7110E+06, 4.7240E+06, 4.7372E+06, 4.7503E+06, 4.7635E+06,
     +  4.7767E+06, 4.7900E+06, 4.8032E+06, 4.8165E+06, 4.8298E+06,
     + 4.8432E+06, 4.8566E+06, 4.8700E+06, 4.8834E+06, 4.8969E+06,
     +  4.9104E+06, 4.9239E+06, 4.9375E+06, 4.9511E+06, 4.9647E+06,
     + 4.9783E+06, 4.9920E+06, 5.0057E+06, 5.0194E+06, 5.0332E+06,
     +  5.0470E+06, 5.0608E+06, 5.0747E+06, 5.0886E+06, 5.1025E+06,
     + 5.1164E+06, 5.1304E+06, 5.1444E+06, 5.1584E+06, 5.1725E+06,
     +  5.1866E+06, 5.2007E+06, 5.2148E+06, 5.2290E+06, 5.2432E+06,
     + 5.2575E+06, 5.2718E+06, 5.2861E+06, 5.3004E+06, 5.3148E+06,
     +  5.3292E+06, 5.3436E+06, 5.3580E+06, 5.3725E+06, 5.3870E+06,
     + 5.4016E+06, 5.4162E+06, 5.4308E+06, 5.4454E+06, 5.4601E+06,
     +  5.4748E+06, 5.4895E+06, 5.5043E+06, 5.5191E+06, 5.5339E+06,
     + 5.5488E+06, 5.5637E+06, 5.5786E+06, 5.5936E+06, 5.6085E+06,
     +  5.6236E+06, 5.6386E+06, 5.6537E+06, 5.6688E+06, 5.6839E+06,
     + 5.6991E+06, 5.7143E+06, 5.7296E+06, 5.7448E+06, 5.7601E+06,
     +  5.7755E+06, 5.7909E+06, 5.8062E+06, 5.8217E+06, 5.8371E+06,
     + 5.8526E+06, 5.8682E+06, 5.8837E+06, 5.8993E+06, 5.9150E+06,
     +  5.9306E+06, 5.9463E+06, 5.9620E+06, 5.9778E+06, 5.9936E+06,
     + 6.0094E+06, 6.0252E+06, 6.0411E+06, 6.0571E+06, 6.0730E+06,
     +  6.0890E+06, 6.1050E+06, 6.1211E+06, 6.1372E+06, 6.1533E+06,
     + 6.1694E+06, 6.1856E+06, 6.2018E+06, 6.2181E+06, 6.2344E+06,
     +  6.2507E+06, 6.2670E+06, 6.2834E+06, 6.2998E+06, 6.3163E+06,
     + 6.3328E+06, 6.3493E+06, 6.3659E+06, 6.3825E+06, 6.3991E+06,
     +  6.4157E+06, 6.4324E+06, 6.4491E+06, 6.4659E+06, 6.4827E+06,
     + 6.4995E+06, 6.5164E+06, 6.5333E+06, 6.5502E+06, 6.5672E+06,
     +  6.5842E+06, 6.6012E+06, 6.6183E+06, 6.6354E+06, 6.6525E+06,
     + 6.6697E+06, 6.6869E+06, 6.7042E+06, 6.7214E+06, 6.7387E+06,
     +  6.7561E+06, 6.7735E+06, 6.7909E+06, 6.8084E+06, 6.8258E+06,
     + 6.8434E+06, 6.8609E+06, 6.8785E+06, 6.8961E+06, 6.9138E+06,
     +  6.9315E+06, 6.9493E+06, 6.9670E+06, 6.9848E+06, 7.0027E+06,
     + 7.0206E+06, 7.0385E+06, 7.0564E+06, 7.0744E+06, 7.0924E+06,
     +  7.1105E+06, 7.1286E+06, 7.1467E+06, 7.1649E+06, 7.1831E+06,
     + 7.2013E+06, 7.2196E+06, 7.2379E+06, 7.2563E+06, 7.2747E+06,
     +  7.2931E+06, 7.3115E+06, 7.3300E+06, 7.3486E+06, 7.3671E+06,
     + 7.3858E+06, 7.4044E+06, 7.4231E+06, 7.4418E+06, 7.4606E+06,
     +  7.4793E+06, 7.4982E+06, 7.5170E+06, 7.5359E+06, 7.5549E+06,
     + 7.5739E+06, 7.5929E+06, 7.6119E+06, 7.6310E+06, 7.6502E+06,
     +  7.6693E+06, 7.6885E+06, 7.7078E+06, 7.7270E+06, 7.7464E+06,
     + 7.7657E+06, 7.7851E+06, 7.8045E+06, 7.8240E+06, 7.8435E+06,
     +  7.8631E+06, 7.8827E+06, 7.9023E+06, 7.9219E+06, 7.9416E+06,
     + 7.9614E+06, 7.9812E+06, 8.0010E+06, 8.0208E+06, 8.0407E+06,
     +  8.0607E+06, 8.0806E+06, 8.1006E+06, 8.1207E+06, 8.1408E+06,
     + 8.1609E+06, 8.1811E+06, 8.2013E+06, 8.2215E+06, 8.2418E+06,
     +  8.2621E+06, 8.2825E+06, 8.3029E+06, 8.3233E+06, 8.3438E+06,
     + 8.3643E+06, 8.3849E+06, 8.4055E+06, 8.4261E+06, 8.4468E+06,
     +  8.4675E+06, 8.4883E+06, 8.5091E+06, 8.5299E+06, 8.5508E+06,
     + 8.5717E+06, 8.5927E+06, 8.6137E+06, 8.6347E+06, 8.6558E+06,
     +  8.6769E+06, 8.6981E+06, 8.7193E+06, 8.7405E+06, 8.7618E+06,
     + 8.7831E+06, 8.8045E+06, 8.8259E+06, 8.8473E+06, 8.8688E+06,
     +  8.8903E+06, 8.9119E+06, 8.9335E+06, 8.9552E+06, 8.9769E+06,
     + 8.9986E+06, 9.0204E+06, 9.0422E+06, 9.0640E+06, 9.0859E+06,
     +  9.1079E+06, 9.1299E+06, 9.1519E+06, 9.1740E+06, 9.1961E+06,
     + 9.2182E+06, 9.2404E+06, 9.2626E+06, 9.2849E+06, 9.3072E+06,
     +  9.3296E+06, 9.3520E+06, 9.3744E+06, 9.3969E+06, 9.4195E+06,
     + 9.4420E+06, 9.4646E+06, 9.4873E+06, 9.5100E+06, 9.5327E+06,
     +  9.5555E+06, 9.5784E+06, 9.6012E+06, 9.6241E+06, 9.6471E+06,
     + 9.6701E+06, 9.6931E+06, 9.7162E+06, 9.7393E+06, 9.7625E+06,
     +  9.7857E+06, 9.8090E+06, 9.8323E+06, 9.8556E+06, 9.8790E+06,
     + 9.9025E+06, 9.9259E+06, 9.9495E+06, 9.9730E+06, 9.9966E+06,
     +  1.0020E+07, 1.0044E+07, 1.0068E+07, 1.0091E+07, 1.0115E+07,
     + 1.0139E+07, 1.0163E+07, 1.0187E+07, 1.0211E+07, 1.0235E+07,
     +  1.0259E+07, 1.0283E+07, 1.0308E+07, 1.0332E+07, 1.0356E+07,
     + 1.0380E+07, 1.0405E+07, 1.0429E+07, 1.0454E+07, 1.0478E+07,
     +  1.0503E+07, 1.0527E+07, 1.0552E+07, 1.0577E+07, 1.0601E+07,
     + 1.0626E+07, 1.0651E+07, 1.0676E+07, 1.0701E+07, 1.0726E+07,
     +  1.0751E+07, 1.0776E+07, 1.0801E+07, 1.0826E+07, 1.0851E+07,
     + 1.0876E+07, 1.0901E+07, 1.0927E+07, 1.0952E+07, 1.0978E+07,
     +  1.1003E+07, 1.1029E+07, 1.1054E+07, 1.1080E+07, 1.1105E+07,
     + 1.1131E+07, 1.1157E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 6,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       868
      data (QofT( 7,J),J=1,2408)/ 8.7893E-01, 2.2874E+00, 4.0827E+00,
     +  6.1955E+00, 8.5839E+00, 1.1219E+01, 1.4080E+01, 1.7150E+01,
     + 2.0416E+01, 2.3866E+01, 2.7492E+01, 3.1285E+01, 3.5238E+01,
     +  3.9344E+01, 4.3600E+01, 4.7998E+01, 5.2536E+01, 5.7208E+01,
     + 6.2011E+01, 6.6942E+01, 7.7173E+01, 8.7879E+01, 9.9040E+01,
     +  1.1064E+02, 1.2266E+02, 1.3508E+02, 1.4790E+02, 1.6110E+02,
     + 1.7467E+02, 1.8861E+02, 2.0290E+02, 2.1753E+02, 2.3250E+02,
     +  2.4780E+02, 2.6342E+02, 2.7936E+02, 2.9560E+02, 3.1215E+02,
     + 3.2900E+02, 3.4615E+02, 3.6358E+02, 3.8130E+02, 3.9929E+02,
     +  4.1756E+02, 4.3611E+02, 4.5492E+02, 4.7399E+02, 4.9333E+02,
     + 5.1292E+02, 5.3277E+02, 5.5287E+02, 5.7321E+02, 5.9380E+02,
     +  6.1464E+02, 6.3571E+02, 6.5702E+02, 6.7857E+02, 7.0035E+02,
     + 7.2236E+02, 7.4460E+02, 7.6706E+02, 7.8976E+02, 8.1267E+02,
     +  8.3580E+02, 8.5915E+02, 8.8273E+02, 9.0652E+02, 9.3053E+02,
     + 9.5475E+02, 9.7918E+02, 1.0038E+03, 1.0287E+03, 1.0538E+03,
     +  1.0790E+03, 1.1045E+03, 1.1302E+03, 1.1561E+03, 1.1822E+03,
     + 1.2086E+03, 1.2351E+03, 1.2618E+03, 1.2888E+03, 1.3159E+03,
     +  1.3433E+03, 1.3708E+03, 1.3986E+03, 1.4265E+03, 1.4547E+03,
     + 1.4831E+03, 1.5117E+03, 1.5405E+03, 1.5695E+03, 1.5987E+03,
     +  1.6281E+03, 1.6578E+03, 1.6876E+03, 1.7177E+03, 1.7480E+03,
     + 1.7784E+03, 1.8091E+03, 1.8400E+03, 1.8712E+03, 1.9025E+03,
     +  1.9341E+03, 1.9659E+03, 1.9979E+03, 2.0301E+03, 2.0625E+03,
     + 2.0952E+03, 2.1281E+03, 2.1612E+03, 2.1946E+03, 2.2281E+03,
     +  2.2620E+03, 2.2960E+03, 2.3303E+03, 2.3648E+03, 2.3996E+03,
     + 2.4346E+03, 2.4698E+03, 2.5053E+03, 2.5410E+03, 2.5769E+03,
     +  2.6131E+03, 2.6496E+03, 2.6863E+03, 2.7233E+03, 2.7605E+03,
     + 2.7980E+03, 2.8357E+03, 2.8737E+03, 2.9120E+03, 2.9505E+03,
     +  2.9893E+03, 3.0284E+03, 3.0677E+03, 3.1073E+03, 3.1472E+03,
     + 3.1874E+03, 3.2278E+03, 3.2685E+03, 3.3095E+03, 3.3508E+03,
     +  3.3924E+03, 3.4343E+03, 3.4765E+03, 3.5189E+03, 3.5617E+03,
     + 3.6048E+03, 3.6481E+03, 3.6918E+03, 3.7358E+03, 3.7801E+03,
     +  3.8247E+03, 3.8696E+03, 3.9148E+03, 3.9604E+03, 4.0062E+03,
     + 4.0524E+03, 4.0990E+03, 4.1458E+03, 4.1930E+03, 4.2405E+03,
     +  4.2884E+03, 4.3366E+03, 4.3851E+03, 4.4340E+03, 4.4832E+03,
     + 4.5328E+03, 4.5827E+03, 4.6330E+03, 4.6836E+03, 4.7346E+03,
     +  4.7860E+03, 4.8377E+03, 4.8898E+03, 4.9422E+03, 4.9951E+03,
     + 5.0483E+03, 5.1019E+03, 5.1558E+03, 5.2102E+03, 5.2649E+03,
     +  5.3200E+03, 5.3756E+03, 5.4315E+03, 5.4878E+03, 5.5445E+03,
     + 5.6016E+03, 5.6591E+03, 5.7171E+03, 5.7754E+03, 5.8341E+03,
     +  5.8933E+03, 5.9529E+03, 6.0129E+03, 6.0733E+03, 6.1342E+03,
     + 6.1955E+03, 6.2572E+03, 6.3193E+03, 6.3819E+03, 6.4450E+03,
     +  6.5085E+03, 6.5724E+03, 6.6367E+03, 6.7016E+03, 6.7669E+03,
     + 6.8326E+03, 6.8988E+03, 6.9655E+03, 7.0326E+03, 7.1002E+03,
     +  7.1683E+03, 7.2369E+03, 7.3060E+03, 7.3755E+03, 7.4455E+03,
     + 7.5160E+03, 7.5870E+03, 7.6585E+03, 7.7304E+03, 7.8029E+03,
     +  7.8759E+03, 7.9494E+03, 8.0235E+03, 8.0980E+03, 8.1730E+03,
     + 8.2485E+03, 8.3246E+03, 8.4012E+03, 8.4783E+03, 8.5560E+03,
     +  8.6342E+03, 8.7129E+03, 8.7922E+03, 8.8720E+03, 8.9523E+03,
     + 9.0333E+03, 9.1148E+03, 9.1968E+03, 9.2793E+03, 9.3625E+03,
     +  9.4462E+03, 9.5305E+03, 9.6154E+03, 9.7008E+03, 9.7868E+03,
     + 9.8734E+03, 9.9605E+03, 1.0048E+04, 1.0137E+04, 1.0226E+04,
     +  1.0315E+04, 1.0405E+04, 1.0496E+04, 1.0587E+04, 1.0679E+04,
     + 1.0772E+04, 1.0865E+04, 1.0959E+04, 1.1054E+04, 1.1149E+04,
     +  1.1244E+04, 1.1341E+04, 1.1438E+04, 1.1535E+04, 1.1634E+04,
     + 1.1733E+04, 1.1832E+04, 1.1932E+04, 1.2033E+04, 1.2135E+04,
     +  1.2237E+04, 1.2340E+04, 1.2443E+04, 1.2548E+04, 1.2653E+04,
     + 1.2758E+04, 1.2865E+04, 1.2972E+04, 1.3079E+04, 1.3188E+04,
     +  1.3297E+04, 1.3407E+04, 1.3517E+04, 1.3628E+04, 1.3740E+04,
     + 1.3853E+04, 1.3966E+04, 1.4080E+04, 1.4195E+04, 1.4311E+04,
     +  1.4427E+04, 1.4544E+04, 1.4662E+04, 1.4780E+04, 1.4900E+04,
     + 1.5020E+04, 1.5140E+04, 1.5262E+04, 1.5384E+04, 1.5507E+04,
     +  1.5631E+04, 1.5756E+04, 1.5881E+04, 1.6008E+04, 1.6135E+04,
     + 1.6263E+04, 1.6391E+04, 1.6521E+04, 1.6651E+04, 1.6782E+04,
     +  1.6914E+04, 1.7046E+04, 1.7180E+04, 1.7314E+04, 1.7449E+04,
     + 1.7585E+04, 1.7722E+04, 1.7860E+04, 1.7998E+04, 1.8137E+04,
     +  1.8278E+04, 1.8419E+04, 1.8561E+04, 1.8703E+04, 1.8847E+04,
     + 1.8992E+04, 1.9137E+04, 1.9283E+04, 1.9430E+04, 1.9578E+04,
     +  1.9727E+04, 1.9877E+04, 2.0028E+04, 2.0180E+04, 2.0332E+04,
     + 2.0486E+04, 2.0640E+04, 2.0795E+04, 2.0952E+04, 2.1109E+04,
     +  2.1267E+04, 2.1426E+04, 2.1586E+04, 2.1747E+04, 2.1909E+04,
     + 2.2071E+04, 2.2235E+04, 2.2400E+04, 2.2566E+04, 2.2732E+04,
     +  2.2900E+04, 2.3069E+04, 2.3238E+04, 2.3409E+04, 2.3581E+04,
     + 2.3753E+04, 2.3927E+04, 2.4101E+04, 2.4277E+04, 2.4454E+04,
     +  2.4631E+04, 2.4810E+04, 2.4990E+04, 2.5171E+04, 2.5352E+04,
     + 2.5535E+04, 2.5719E+04, 2.5904E+04, 2.6090E+04, 2.6277E+04,
     +  2.6465E+04, 2.6655E+04, 2.6845E+04, 2.7036E+04, 2.7229E+04,
     + 2.7422E+04, 2.7617E+04, 2.7813E+04, 2.8009E+04, 2.8207E+04,
     +  2.8406E+04, 2.8607E+04, 2.8808E+04, 2.9010E+04, 2.9214E+04,
     + 2.9418E+04, 2.9624E+04, 2.9831E+04, 3.0039E+04, 3.0249E+04,
     +  3.0459E+04, 3.0671E+04, 3.0883E+04, 3.1097E+04, 3.1312E+04,
     + 3.1529E+04, 3.1746E+04, 3.1965E+04, 3.2185E+04, 3.2406E+04,
     +  3.2628E+04, 3.2851E+04, 3.3076E+04, 3.3302E+04, 3.3529E+04,
     + 3.3758E+04, 3.3987E+04, 3.4218E+04, 3.4450E+04, 3.4683E+04,
     +  3.4918E+04, 3.5154E+04, 3.5391E+04, 3.5630E+04, 3.5869E+04,
     + 3.6110E+04, 3.6352E+04, 3.6596E+04, 3.6841E+04, 3.7087E+04,
     +  3.7334E+04, 3.7583E+04, 3.7833E+04, 3.8084E+04, 3.8337E+04,
     + 3.8591E+04, 3.8847E+04, 3.9103E+04, 3.9362E+04, 3.9621E+04,
     +  3.9882E+04, 4.0144E+04, 4.0407E+04, 4.0672E+04, 4.0939E+04,
     + 4.1206E+04, 4.1475E+04, 4.1746E+04, 4.2018E+04, 4.2291E+04,
     +  4.2566E+04, 4.2842E+04, 4.3119E+04, 4.3398E+04, 4.3679E+04,
     + 4.3960E+04, 4.4244E+04, 4.4529E+04, 4.4815E+04, 4.5102E+04,
     +  4.5391E+04, 4.5682E+04, 4.5974E+04, 4.6268E+04, 4.6563E+04,
     + 4.6859E+04, 4.7157E+04, 4.7457E+04, 4.7758E+04, 4.8060E+04,
     +  4.8364E+04, 4.8670E+04, 4.8977E+04, 4.9286E+04, 4.9596E+04,
     + 4.9908E+04, 5.0221E+04, 5.0536E+04, 5.0852E+04, 5.1170E+04,
     +  5.1490E+04, 5.1811E+04, 5.2134E+04, 5.2458E+04, 5.2784E+04,
     + 5.3112E+04, 5.3441E+04, 5.3772E+04, 5.4104E+04, 5.4439E+04,
     +  5.4774E+04, 5.5112E+04, 5.5451E+04, 5.5791E+04, 5.6134E+04,
     + 5.6478E+04, 5.6823E+04, 5.7171E+04, 5.7520E+04, 5.7870E+04,
     +  5.8223E+04, 5.8577E+04, 5.8933E+04, 5.9290E+04, 5.9650E+04,
     + 6.0011E+04, 6.0373E+04, 6.0738E+04, 6.1104E+04, 6.1472E+04,
     +  6.1842E+04, 6.2213E+04, 6.2587E+04, 6.2962E+04, 6.3338E+04,
     + 6.3717E+04, 6.4098E+04, 6.4480E+04, 6.4864E+04, 6.5250E+04,
     +  6.5637E+04, 6.6026E+04, 6.6418E+04, 6.6811E+04, 6.7206E+04,
     + 6.7603E+04, 6.8002E+04, 6.8402E+04, 6.8805E+04, 6.9209E+04,
     +  6.9615E+04, 7.0023E+04, 7.0433E+04, 7.0845E+04, 7.1259E+04,
     + 7.1675E+04, 7.2092E+04, 7.2512E+04, 7.2933E+04, 7.3357E+04,
     +  7.3782E+04, 7.4209E+04, 7.4639E+04, 7.5070E+04, 7.5503E+04,
     + 7.5939E+04, 7.6376E+04, 7.6815E+04, 7.7256E+04, 7.7700E+04,
     +  7.8145E+04, 7.8592E+04, 7.9041E+04, 7.9493E+04, 7.9946E+04,
     + 8.0401E+04, 8.0859E+04, 8.1318E+04, 8.1780E+04, 8.2244E+04,
     +  8.2709E+04, 8.3177E+04, 8.3647E+04, 8.4119E+04, 8.4593E+04,
     + 8.5070E+04, 8.5548E+04, 8.6029E+04, 8.6511E+04, 8.6996E+04,
     +  8.7483E+04, 8.7972E+04, 8.8464E+04, 8.8957E+04, 8.9453E+04,
     + 8.9950E+04, 9.0450E+04, 9.0953E+04, 9.1457E+04, 9.1964E+04,
     +  9.2473E+04, 9.2984E+04, 9.3497E+04, 9.4013E+04, 9.4531E+04,
     + 9.5051E+04, 9.5573E+04, 9.6097E+04, 9.6624E+04, 9.7153E+04,
     +  9.7685E+04, 9.8219E+04, 9.8755E+04, 9.9293E+04, 9.9834E+04,
     + 1.0038E+05, 1.0092E+05, 1.0147E+05, 1.0202E+05, 1.0257E+05,
     +  1.0313E+05, 1.0368E+05, 1.0424E+05, 1.0481E+05, 1.0537E+05,
     + 1.0594E+05, 1.0651E+05, 1.0708E+05, 1.0765E+05, 1.0823E+05,
     +  1.0881E+05, 1.0939E+05, 1.0997E+05, 1.1056E+05, 1.1115E+05,
     + 1.1174E+05, 1.1233E+05, 1.1293E+05, 1.1353E+05, 1.1413E+05,
     +  1.1473E+05, 1.1534E+05, 1.1595E+05, 1.1656E+05, 1.1717E+05,
     + 1.1779E+05, 1.1841E+05, 1.1903E+05, 1.1965E+05, 1.2028E+05,
     +  1.2091E+05, 1.2154E+05, 1.2218E+05, 1.2281E+05, 1.2345E+05,
     + 1.2410E+05, 1.2474E+05, 1.2539E+05, 1.2604E+05, 1.2669E+05,
     +  1.2735E+05, 1.2801E+05, 1.2867E+05, 1.2933E+05, 1.3000E+05,
     + 1.3067E+05, 1.3134E+05, 1.3201E+05, 1.3269E+05, 1.3337E+05,
     +  1.3405E+05, 1.3474E+05, 1.3543E+05, 1.3612E+05, 1.3681E+05,
     + 1.3751E+05, 1.3821E+05, 1.3891E+05, 1.3962E+05, 1.4033E+05,
     +  1.4104E+05, 1.4175E+05, 1.4247E+05, 1.4319E+05, 1.4391E+05,
     + 1.4463E+05, 1.4536E+05, 1.4609E+05, 1.4683E+05, 1.4756E+05,
     +  1.4830E+05, 1.4905E+05, 1.4979E+05, 1.5054E+05, 1.5129E+05,
     + 1.5205E+05, 1.5280E+05, 1.5356E+05, 1.5433E+05, 1.5509E+05,
     +  1.5586E+05, 1.5664E+05, 1.5741E+05, 1.5819E+05, 1.5897E+05,
     + 1.5975E+05, 1.6054E+05, 1.6133E+05, 1.6213E+05, 1.6292E+05,
     +  1.6372E+05, 1.6453E+05, 1.6533E+05, 1.6614E+05, 1.6695E+05,
     + 1.6777E+05, 1.6859E+05, 1.6941E+05, 1.7024E+05, 1.7106E+05,
     +  1.7190E+05, 1.7273E+05, 1.7357E+05, 1.7441E+05, 1.7525E+05,
     + 1.7610E+05, 1.7695E+05, 1.7780E+05, 1.7866E+05, 1.7952E+05,
     +  1.8039E+05, 1.8125E+05, 1.8212E+05, 1.8300E+05, 1.8387E+05,
     + 1.8475E+05, 1.8564E+05, 1.8652E+05, 1.8741E+05, 1.8831E+05,
     +  1.8920E+05, 1.9010E+05, 1.9101E+05, 1.9191E+05, 1.9282E+05,
     + 1.9374E+05, 1.9466E+05, 1.9558E+05, 1.9650E+05, 1.9743E+05,
     +  1.9836E+05, 1.9929E+05, 2.0023E+05, 2.0117E+05, 2.0212E+05,
     + 2.0306E+05, 2.0402E+05, 2.0497E+05, 2.0593E+05, 2.0689E+05,
     +  2.0786E+05, 2.0883E+05, 2.0980E+05, 2.1078E+05, 2.1176E+05,
     + 2.1274E+05, 2.1373E+05, 2.1472E+05, 2.1572E+05, 2.1671E+05,
     +  2.1772E+05, 2.1872E+05, 2.1973E+05, 2.2074E+05, 2.2176E+05,
     + 2.2278E+05, 2.2381E+05, 2.2483E+05, 2.2586E+05, 2.2690E+05,
     +  2.2794E+05, 2.2898E+05, 2.3003E+05, 2.3108E+05, 2.3213E+05,
     + 2.3319E+05, 2.3425E+05, 2.3532E+05, 2.3639E+05, 2.3746E+05,
     +  2.3854E+05, 2.3962E+05, 2.4070E+05, 2.4179E+05, 2.4288E+05,
     + 2.4398E+05, 2.4508E+05, 2.4618E+05, 2.4729E+05, 2.4840E+05,
     +  2.4952E+05, 2.5064E+05, 2.5176E+05, 2.5289E+05, 2.5402E+05,
     + 2.5516E+05, 2.5630E+05, 2.5744E+05, 2.5859E+05, 2.5974E+05,
     +  2.6090E+05, 2.6206E+05, 2.6322E+05, 2.6439E+05, 2.6556E+05,
     + 2.6674E+05, 2.6792E+05, 2.6910E+05, 2.7029E+05, 2.7148E+05,
     +  2.7268E+05, 2.7388E+05, 2.7508E+05, 2.7629E+05, 2.7751E+05,
     + 2.7873E+05, 2.7995E+05, 2.8117E+05, 2.8240E+05, 2.8364E+05,
     +  2.8488E+05, 2.8612E+05, 2.8737E+05, 2.8862E+05, 2.8988E+05,
     + 2.9114E+05, 2.9240E+05, 2.9367E+05, 2.9494E+05, 2.9622E+05,
     +  2.9750E+05, 2.9879E+05, 3.0008E+05, 3.0137E+05, 3.0267E+05,
     + 3.0398E+05, 3.0529E+05, 3.0660E+05, 3.0792E+05, 3.0924E+05,
     +  3.1056E+05, 3.1189E+05, 3.1323E+05, 3.1457E+05, 3.1591E+05,
     + 3.1726E+05, 3.1862E+05, 3.1997E+05, 3.2134E+05, 3.2270E+05,
     +  3.2407E+05, 3.2545E+05, 3.2683E+05, 3.2822E+05, 3.2961E+05,
     + 3.3100E+05, 3.3240E+05, 3.3380E+05, 3.3521E+05, 3.3662E+05,
     +  3.3804E+05, 3.3947E+05, 3.4089E+05, 3.4232E+05, 3.4376E+05,
     + 3.4520E+05, 3.4665E+05, 3.4810E+05, 3.4956E+05, 3.5102E+05,
     +  3.5248E+05, 3.5395E+05, 3.5543E+05, 3.5691E+05, 3.5839E+05,
     + 3.5988E+05, 3.6138E+05, 3.6287E+05, 3.6438E+05, 3.6589E+05,
     +  3.6740E+05, 3.6892E+05, 3.7045E+05, 3.7197E+05, 3.7351E+05,
     + 3.7505E+05, 3.7659E+05, 3.7814E+05, 3.7969E+05, 3.8125E+05,
     +  3.8282E+05, 3.8439E+05, 3.8596E+05, 3.8754E+05, 3.8912E+05,
     + 3.9071E+05, 3.9231E+05, 3.9391E+05, 3.9551E+05, 3.9712E+05,
     +  3.9874E+05, 4.0036E+05, 4.0198E+05, 4.0361E+05, 4.0525E+05,
     + 4.0689E+05, 4.0853E+05, 4.1019E+05, 4.1184E+05, 4.1350E+05,
     +  4.1517E+05, 4.1684E+05, 4.1852E+05, 4.2021E+05, 4.2189E+05,
     + 4.2359E+05, 4.2529E+05, 4.2699E+05, 4.2870E+05, 4.3042E+05,
     +  4.3214E+05, 4.3386E+05, 4.3559E+05, 4.3733E+05, 4.3907E+05,
     + 4.4082E+05, 4.4257E+05, 4.4433E+05, 4.4610E+05, 4.4787E+05,
     +  4.4964E+05, 4.5142E+05, 4.5321E+05, 4.5500E+05, 4.5680E+05,
     + 4.5860E+05, 4.6041E+05, 4.6223E+05, 4.6405E+05, 4.6587E+05,
     +  4.6770E+05, 4.6954E+05, 4.7138E+05, 4.7323E+05, 4.7509E+05,
     + 4.7695E+05, 4.7881E+05, 4.8068E+05, 4.8256E+05, 4.8444E+05,
     +  4.8633E+05, 4.8822E+05, 4.9012E+05, 4.9203E+05, 4.9394E+05,
     + 4.9586E+05, 4.9778E+05, 4.9971E+05, 5.0165E+05, 5.0359E+05,
     +  5.0554E+05, 5.0749E+05, 5.0945E+05, 5.1141E+05, 5.1338E+05,
     + 5.1536E+05, 5.1734E+05, 5.1933E+05, 5.2133E+05, 5.2333E+05,
     +  5.2533E+05, 5.2735E+05, 5.2937E+05, 5.3139E+05, 5.3342E+05,
     + 5.3546E+05, 5.3750E+05, 5.3955E+05, 5.4161E+05, 5.4367E+05,
     +  5.4574E+05, 5.4781E+05, 5.4989E+05, 5.5198E+05, 5.5407E+05,
     + 5.5617E+05, 5.5828E+05, 5.6039E+05, 5.6251E+05, 5.6463E+05,
     +  5.6676E+05, 5.6890E+05, 5.7104E+05, 5.7319E+05, 5.7534E+05,
     + 5.7751E+05, 5.7968E+05, 5.8185E+05, 5.8403E+05, 5.8622E+05,
     +  5.8841E+05, 5.9061E+05, 5.9282E+05, 5.9504E+05, 5.9726E+05,
     + 5.9948E+05, 6.0172E+05, 6.0396E+05, 6.0620E+05, 6.0845E+05,
     +  6.1071E+05, 6.1298E+05, 6.1525E+05, 6.1753E+05, 6.1982E+05,
     + 6.2211E+05, 6.2441E+05, 6.2672E+05, 6.2903E+05, 6.3135E+05,
     +  6.3367E+05, 6.3601E+05, 6.3835E+05, 6.4069E+05, 6.4305E+05,
     + 6.4541E+05, 6.4777E+05, 6.5015E+05, 6.5253E+05, 6.5491E+05,
     +  6.5731E+05, 6.5971E+05, 6.6212E+05, 6.6453E+05, 6.6695E+05,
     + 6.6938E+05, 6.7182E+05, 6.7426E+05, 6.7671E+05, 6.7917E+05,
     +  6.8163E+05, 6.8410E+05, 6.8658E+05, 6.8906E+05, 6.9156E+05,
     + 6.9405E+05, 6.9656E+05, 6.9907E+05, 7.0159E+05, 7.0412E+05,
     +  7.0666E+05, 7.0920E+05, 7.1175E+05, 7.1430E+05, 7.1687E+05,
     + 7.1944E+05, 7.2201E+05, 7.2460E+05, 7.2719E+05, 7.2979E+05,
     +  7.3240E+05, 7.3501E+05, 7.3763E+05, 7.4026E+05, 7.4290E+05,
     + 7.4554E+05, 7.4819E+05, 7.5085E+05, 7.5352E+05, 7.5619E+05,
     +  7.5887E+05, 7.6156E+05, 7.6426E+05, 7.6696E+05, 7.6967E+05,
     + 7.7239E+05, 7.7511E+05, 7.7785E+05, 7.8059E+05, 7.8334E+05,
     +  7.8609E+05, 7.8886E+05, 7.9163E+05, 7.9441E+05, 7.9719E+05,
     + 7.9999E+05, 8.0279E+05, 8.0560E+05, 8.0842E+05, 8.1124E+05,
     +  8.1408E+05, 8.1692E+05, 8.1977E+05, 8.2262E+05, 8.2549E+05,
     + 8.2836E+05, 8.3124E+05, 8.3413E+05, 8.3702E+05, 8.3993E+05,
     +  8.4284E+05, 8.4576E+05, 8.4869E+05, 8.5162E+05, 8.5456E+05,
     + 8.5752E+05, 8.6048E+05, 8.6344E+05, 8.6642E+05, 8.6940E+05,
     +  8.7239E+05, 8.7539E+05, 8.7840E+05, 8.8142E+05, 8.8444E+05,
     + 8.8747E+05, 8.9051E+05, 8.9356E+05, 8.9662E+05, 8.9968E+05,
     +  9.0276E+05, 9.0584E+05, 9.0893E+05, 9.1203E+05, 9.1513E+05,
     + 9.1825E+05, 9.2137E+05, 9.2450E+05, 9.2764E+05, 9.3079E+05,
     +  9.3395E+05, 9.3711E+05, 9.4029E+05, 9.4347E+05, 9.4666E+05,
     + 9.4986E+05, 9.5306E+05, 9.5628E+05, 9.5950E+05, 9.6274E+05,
     +  9.6598E+05, 9.6923E+05, 9.7249E+05, 9.7575E+05, 9.7903E+05,
     + 9.8231E+05, 9.8561E+05, 9.8891E+05, 9.9222E+05, 9.9554E+05,
     +  9.9887E+05, 1.0022E+06, 1.0055E+06, 1.0089E+06, 1.0123E+06,
     + 1.0156E+06, 1.0190E+06, 1.0224E+06, 1.0258E+06, 1.0292E+06,
     +  1.0326E+06, 1.0361E+06, 1.0395E+06, 1.0429E+06, 1.0464E+06,
     + 1.0498E+06, 1.0533E+06, 1.0568E+06, 1.0603E+06, 1.0638E+06,
     +  1.0673E+06, 1.0708E+06, 1.0743E+06, 1.0779E+06, 1.0814E+06,
     + 1.0849E+06, 1.0885E+06, 1.0921E+06, 1.0957E+06, 1.0992E+06,
     +  1.1028E+06, 1.1065E+06, 1.1101E+06, 1.1137E+06, 1.1173E+06,
     + 1.1210E+06, 1.1246E+06, 1.1283E+06, 1.1320E+06, 1.1356E+06,
     +  1.1393E+06, 1.1430E+06, 1.1467E+06, 1.1505E+06, 1.1542E+06,
     + 1.1579E+06, 1.1617E+06, 1.1654E+06, 1.1692E+06, 1.1730E+06,
     +  1.1768E+06, 1.1806E+06, 1.1844E+06, 1.1882E+06, 1.1920E+06,
     + 1.1958E+06, 1.1997E+06, 1.2035E+06, 1.2074E+06, 1.2113E+06,
     +  1.2152E+06, 1.2190E+06, 1.2229E+06, 1.2269E+06, 1.2308E+06,
     + 1.2347E+06, 1.2387E+06, 1.2426E+06, 1.2466E+06, 1.2505E+06,
     +  1.2545E+06, 1.2585E+06, 1.2625E+06, 1.2665E+06, 1.2705E+06,
     + 1.2746E+06, 1.2786E+06, 1.2827E+06, 1.2867E+06, 1.2908E+06,
     +  1.2949E+06, 1.2990E+06, 1.3031E+06, 1.3072E+06, 1.3113E+06,
     + 1.3154E+06, 1.3196E+06, 1.3237E+06, 1.3279E+06, 1.3321E+06,
     +  1.3363E+06, 1.3405E+06, 1.3447E+06, 1.3489E+06, 1.3531E+06,
     + 1.3573E+06, 1.3616E+06, 1.3658E+06, 1.3701E+06, 1.3744E+06,
     +  1.3787E+06, 1.3830E+06, 1.3873E+06, 1.3916E+06, 1.3959E+06,
     + 1.4003E+06, 1.4046E+06, 1.4090E+06, 1.4134E+06, 1.4177E+06,
     +  1.4221E+06, 1.4265E+06, 1.4310E+06, 1.4354E+06, 1.4398E+06,
     + 1.4443E+06, 1.4487E+06, 1.4532E+06, 1.4577E+06, 1.4622E+06,
     +  1.4667E+06, 1.4712E+06, 1.4757E+06, 1.4802E+06, 1.4848E+06,
     + 1.4893E+06, 1.4939E+06, 1.4985E+06, 1.5031E+06, 1.5077E+06,
     +  1.5123E+06, 1.5169E+06, 1.5215E+06, 1.5262E+06, 1.5308E+06,
     + 1.5355E+06, 1.5402E+06, 1.5449E+06, 1.5496E+06, 1.5543E+06,
     +  1.5590E+06, 1.5637E+06, 1.5685E+06, 1.5732E+06, 1.5780E+06,
     + 1.5828E+06, 1.5876E+06, 1.5924E+06, 1.5972E+06, 1.6020E+06,
     +  1.6069E+06, 1.6117E+06, 1.6166E+06, 1.6214E+06, 1.6263E+06,
     + 1.6312E+06, 1.6361E+06, 1.6410E+06, 1.6460E+06, 1.6509E+06,
     +  1.6559E+06, 1.6608E+06, 1.6658E+06, 1.6708E+06, 1.6758E+06,
     + 1.6808E+06, 1.6858E+06, 1.6909E+06, 1.6959E+06, 1.7010E+06,
     +  1.7060E+06, 1.7111E+06, 1.7162E+06, 1.7213E+06, 1.7264E+06,
     + 1.7316E+06, 1.7367E+06, 1.7419E+06, 1.7470E+06, 1.7522E+06,
     +  1.7574E+06, 1.7626E+06, 1.7678E+06, 1.7730E+06, 1.7783E+06,
     + 1.7835E+06, 1.7888E+06, 1.7941E+06, 1.7994E+06, 1.8047E+06,
     +  1.8100E+06, 1.8153E+06, 1.8206E+06, 1.8260E+06, 1.8313E+06,
     + 1.8367E+06, 1.8421E+06, 1.8475E+06, 1.8529E+06, 1.8583E+06,
     +  1.8638E+06, 1.8692E+06, 1.8747E+06, 1.8801E+06, 1.8856E+06,
     + 1.8911E+06, 1.8966E+06, 1.9022E+06, 1.9077E+06, 1.9132E+06,
     +  1.9188E+06, 1.9244E+06, 1.9300E+06, 1.9356E+06, 1.9412E+06,
     + 1.9468E+06, 1.9524E+06, 1.9581E+06, 1.9638E+06, 1.9694E+06,
     +  1.9751E+06, 1.9808E+06, 1.9865E+06, 1.9923E+06, 1.9980E+06,
     + 2.0038E+06, 2.0095E+06, 2.0153E+06, 2.0211E+06, 2.0269E+06,
     +  2.0327E+06, 2.0386E+06, 2.0444E+06, 2.0503E+06, 2.0561E+06,
     + 2.0620E+06, 2.0679E+06, 2.0738E+06, 2.0798E+06, 2.0857E+06,
     +  2.0916E+06, 2.0976E+06, 2.1036E+06, 2.1096E+06, 2.1156E+06,
     + 2.1216E+06, 2.1276E+06, 2.1337E+06, 2.1397E+06, 2.1458E+06,
     +  2.1519E+06, 2.1580E+06, 2.1641E+06, 2.1702E+06, 2.1764E+06,
     + 2.1825E+06, 2.1887E+06, 2.1949E+06, 2.2011E+06, 2.2073E+06,
     +  2.2135E+06, 2.2197E+06, 2.2260E+06, 2.2322E+06, 2.2385E+06,
     + 2.2448E+06, 2.2511E+06, 2.2574E+06, 2.2638E+06, 2.2701E+06,
     +  2.2765E+06, 2.2829E+06, 2.2892E+06, 2.2956E+06, 2.3021E+06,
     + 2.3085E+06, 2.3149E+06, 2.3214E+06, 2.3279E+06, 2.3344E+06,
     +  2.3409E+06, 2.3474E+06, 2.3539E+06, 2.3605E+06, 2.3670E+06,
     + 2.3736E+06, 2.3802E+06, 2.3868E+06, 2.3934E+06, 2.4000E+06,
     +  2.4067E+06, 2.4133E+06, 2.4200E+06, 2.4267E+06, 2.4334E+06,
     + 2.4401E+06, 2.4469E+06, 2.4536E+06, 2.4604E+06, 2.4671E+06,
     +  2.4739E+06, 2.4807E+06, 2.4876E+06, 2.4944E+06, 2.5012E+06,
     + 2.5081E+06, 2.5150E+06, 2.5219E+06, 2.5288E+06, 2.5357E+06,
     +  2.5427E+06, 2.5496E+06, 2.5566E+06, 2.5636E+06, 2.5706E+06,
     + 2.5776E+06, 2.5846E+06, 2.5916E+06, 2.5987E+06, 2.6058E+06,
     +  2.6129E+06, 2.6200E+06, 2.6271E+06, 2.6342E+06, 2.6414E+06,
     + 2.6485E+06, 2.6557E+06, 2.6629E+06, 2.6701E+06, 2.6774E+06,
     +  2.6846E+06, 2.6919E+06, 2.6991E+06, 2.7064E+06, 2.7137E+06,
     + 2.7210E+06, 2.7284E+06, 2.7357E+06, 2.7431E+06, 2.7505E+06,
     +  2.7579E+06, 2.7653E+06, 2.7727E+06, 2.7801E+06, 2.7876E+06,
     + 2.7951E+06, 2.8026E+06, 2.8101E+06, 2.8176E+06, 2.8251E+06,
     +  2.8327E+06, 2.8403E+06, 2.8478E+06, 2.8555E+06, 2.8631E+06,
     + 2.8707E+06, 2.8784E+06, 2.8860E+06, 2.8937E+06, 2.9014E+06,
     +  2.9091E+06, 2.9168E+06, 2.9246E+06, 2.9323E+06, 2.9401E+06,
     + 2.9479E+06, 2.9557E+06, 2.9636E+06, 2.9714E+06, 2.9793E+06,
     +  2.9871E+06, 2.9950E+06, 3.0029E+06, 3.0109E+06, 3.0188E+06,
     + 3.0268E+06, 3.0347E+06, 3.0427E+06, 3.0507E+06, 3.0588E+06,
     +  3.0668E+06, 3.0749E+06, 3.0829E+06, 3.0910E+06, 3.0991E+06,
     + 3.1073E+06, 3.1154E+06, 3.1236E+06, 3.1317E+06, 3.1399E+06,
     +  3.1481E+06, 3.1564E+06, 3.1646E+06, 3.1729E+06, 3.1811E+06,
     + 3.1894E+06, 3.1977E+06, 3.2061E+06, 3.2144E+06, 3.2228E+06,
     +  3.2311E+06, 3.2395E+06, 3.2479E+06, 3.2564E+06, 3.2648E+06,
     + 3.2733E+06, 3.2818E+06, 3.2903E+06, 3.2988E+06, 3.3073E+06,
     +  3.3159E+06, 3.3244E+06, 3.3330E+06, 3.3416E+06, 3.3502E+06,
     + 3.3589E+06, 3.3675E+06, 3.3762E+06, 3.3849E+06, 3.3936E+06,
     +  3.4023E+06, 3.4111E+06, 3.4198E+06, 3.4286E+06, 3.4374E+06,
     + 3.4462E+06, 3.4550E+06, 3.4639E+06, 3.4728E+06, 3.4816E+06,
     +  3.4905E+06, 3.4995E+06, 3.5084E+06, 3.5174E+06, 3.5263E+06,
     + 3.5353E+06, 3.5443E+06, 3.5534E+06, 3.5624E+06, 3.5715E+06,
     +  3.5806E+06, 3.5897E+06, 3.5988E+06, 3.6079E+06, 3.6171E+06,
     + 3.6262E+06, 3.6354E+06, 3.6446E+06, 3.6539E+06, 3.6631E+06,
     +  3.6724E+06, 3.6817E+06, 3.6910E+06, 3.7003E+06, 3.7096E+06,
     + 3.7190E+06, 3.7284E+06, 3.7377E+06, 3.7472E+06, 3.7566E+06,
     +  3.7660E+06, 3.7755E+06, 3.7850E+06, 3.7945E+06, 3.8040E+06,
     + 3.8136E+06, 3.8231E+06, 3.8327E+06, 3.8423E+06, 3.8519E+06,
     +  3.8616E+06, 3.8712E+06, 3.8809E+06, 3.8906E+06, 3.9003E+06,
     + 3.9101E+06, 3.9198E+06, 3.9296E+06, 3.9394E+06, 3.9492E+06,
     +  3.9590E+06, 3.9689E+06, 3.9787E+06, 3.9886E+06, 3.9985E+06,
     + 4.0084E+06, 4.0184E+06, 4.0283E+06, 4.0383E+06, 4.0483E+06,
     +  4.0584E+06, 4.0684E+06, 4.0785E+06, 4.0885E+06, 4.0986E+06,
     + 4.1088E+06, 4.1189E+06, 4.1291E+06, 4.1392E+06, 4.1494E+06,
     +  4.1597E+06, 4.1699E+06, 4.1802E+06, 4.1904E+06, 4.2007E+06,
     + 4.2110E+06, 4.2214E+06, 4.2317E+06, 4.2421E+06, 4.2525E+06,
     +  4.2629E+06, 4.2734E+06, 4.2838E+06, 4.2943E+06, 4.3048E+06,
     + 4.3153E+06, 4.3259E+06, 4.3364E+06, 4.3470E+06, 4.3576E+06,
     +  4.3682E+06, 4.3789E+06, 4.3895E+06, 4.4002E+06, 4.4109E+06,
     + 4.4216E+06, 4.4324E+06, 4.4431E+06, 4.4539E+06, 4.4647E+06,
     +  4.4755E+06, 4.4864E+06, 4.4972E+06, 4.5081E+06, 4.5190E+06,
     + 4.5300E+06, 4.5409E+06, 4.5519E+06, 4.5629E+06, 4.5739E+06,
     +  4.5849E+06, 4.5960E+06, 4.6070E+06, 4.6181E+06, 4.6292E+06,
     + 4.6404E+06, 4.6515E+06, 4.6627E+06, 4.6739E+06, 4.6851E+06,
     +  4.6964E+06, 4.7076E+06, 4.7189E+06, 4.7302E+06, 4.7416E+06,
     + 4.7529E+06, 4.7643E+06, 4.7757E+06, 4.7871E+06, 4.7985E+06,
     +  4.8100E+06, 4.8214E+06, 4.8329E+06, 4.8445E+06, 4.8560E+06,
     + 4.8676E+06, 4.8792E+06, 4.8908E+06, 4.9024E+06, 4.9140E+06,
     +  4.9257E+06, 4.9374E+06, 4.9491E+06, 4.9608E+06, 4.9726E+06,
     + 4.9844E+06, 4.9962E+06, 5.0080E+06, 5.0199E+06, 5.0317E+06,
     +  5.0436E+06, 5.0555E+06, 5.0675E+06, 5.0794E+06, 5.0914E+06,
     + 5.1034E+06, 5.1154E+06, 5.1275E+06, 5.1395E+06, 5.1516E+06,
     +  5.1637E+06, 5.1759E+06, 5.1880E+06, 5.2002E+06, 5.2124E+06,
     + 5.2246E+06, 5.2369E+06, 5.2492E+06, 5.2615E+06, 5.2738E+06,
     +  5.2861E+06, 5.2985E+06, 5.3109E+06, 5.3233E+06, 5.3357E+06,
     + 5.3481E+06, 5.3606E+06, 5.3731E+06, 5.3856E+06, 5.3982E+06,
     +  5.4107E+06, 5.4233E+06, 5.4359E+06, 5.4486E+06, 5.4612E+06,
     + 5.4739E+06, 5.4866E+06, 5.4993E+06, 5.5121E+06, 5.5249E+06,
     +  5.5377E+06, 5.5505E+06, 5.5633E+06, 5.5762E+06, 5.5891E+06,
     + 5.6020E+06, 5.6149E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 7,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       678
      data (QofT( 8,J),J=1,2408)/ 1.0183E+01, 2.6382E+01, 4.7014E+01,
     +  7.1289E+01, 9.8726E+01, 1.2900E+02, 1.6186E+02, 1.9711E+02,
     + 2.3462E+02, 2.7424E+02, 3.1588E+02, 3.5943E+02, 4.0482E+02,
     +  4.5198E+02, 5.0084E+02, 5.5135E+02, 6.0345E+02, 6.5710E+02,
     + 7.1225E+02, 7.6887E+02, 8.8635E+02, 1.0093E+03, 1.1374E+03,
     +  1.2706E+03, 1.4086E+03, 1.5512E+03, 1.6984E+03, 1.8500E+03,
     + 2.0058E+03, 2.1658E+03, 2.3299E+03, 2.4979E+03, 2.6697E+03,
     +  2.8454E+03, 3.0247E+03, 3.2077E+03, 3.3942E+03, 3.5843E+03,
     + 3.7777E+03, 3.9745E+03, 4.1747E+03, 4.3781E+03, 4.5847E+03,
     +  4.7945E+03, 5.0074E+03, 5.2233E+03, 5.4423E+03, 5.6643E+03,
     + 5.8893E+03, 6.1171E+03, 6.3479E+03, 6.5815E+03, 6.8179E+03,
     +  7.0570E+03, 7.2990E+03, 7.5436E+03, 7.7910E+03, 8.0411E+03,
     + 8.2938E+03, 8.5491E+03, 8.8070E+03, 9.0674E+03, 9.3305E+03,
     +  9.5961E+03, 9.8642E+03, 1.0135E+04, 1.0408E+04, 1.0683E+04,
     + 1.0961E+04, 1.1242E+04, 1.1525E+04, 1.1810E+04, 1.2098E+04,
     +  1.2388E+04, 1.2680E+04, 1.2975E+04, 1.3273E+04, 1.3572E+04,
     + 1.3874E+04, 1.4178E+04, 1.4485E+04, 1.4794E+04, 1.5106E+04,
     +  1.5419E+04, 1.5736E+04, 1.6054E+04, 1.6375E+04, 1.6698E+04,
     + 1.7024E+04, 1.7352E+04, 1.7682E+04, 1.8014E+04, 1.8350E+04,
     +  1.8687E+04, 1.9027E+04, 1.9369E+04, 1.9713E+04, 2.0060E+04,
     + 2.0410E+04, 2.0761E+04, 2.1116E+04, 2.1472E+04, 2.1832E+04,
     +  2.2193E+04, 2.2558E+04, 2.2924E+04, 2.3293E+04, 2.3665E+04,
     + 2.4039E+04, 2.4416E+04, 2.4795E+04, 2.5177E+04, 2.5562E+04,
     +  2.5949E+04, 2.6339E+04, 2.6731E+04, 2.7126E+04, 2.7524E+04,
     + 2.7925E+04, 2.8328E+04, 2.8734E+04, 2.9143E+04, 2.9555E+04,
     +  2.9969E+04, 3.0386E+04, 3.0806E+04, 3.1229E+04, 3.1655E+04,
     + 3.2084E+04, 3.2516E+04, 3.2950E+04, 3.3388E+04, 3.3829E+04,
     +  3.4273E+04, 3.4719E+04, 3.5169E+04, 3.5622E+04, 3.6078E+04,
     + 3.6537E+04, 3.7000E+04, 3.7466E+04, 3.7934E+04, 3.8406E+04,
     +  3.8882E+04, 3.9361E+04, 3.9842E+04, 4.0328E+04, 4.0817E+04,
     + 4.1309E+04, 4.1804E+04, 4.2303E+04, 4.2806E+04, 4.3312E+04,
     +  4.3821E+04, 4.4335E+04, 4.4852E+04, 4.5372E+04, 4.5896E+04,
     + 4.6423E+04, 4.6955E+04, 4.7490E+04, 4.8029E+04, 4.8572E+04,
     +  4.9118E+04, 4.9669E+04, 5.0223E+04, 5.0781E+04, 5.1343E+04,
     + 5.1909E+04, 5.2479E+04, 5.3053E+04, 5.3632E+04, 5.4214E+04,
     +  5.4800E+04, 5.5391E+04, 5.5985E+04, 5.6585E+04, 5.7188E+04,
     + 5.7795E+04, 5.8407E+04, 5.9023E+04, 5.9643E+04, 6.0268E+04,
     +  6.0897E+04, 6.1530E+04, 6.2169E+04, 6.2811E+04, 6.3458E+04,
     + 6.4110E+04, 6.4766E+04, 6.5427E+04, 6.6093E+04, 6.6763E+04,
     +  6.7438E+04, 6.8119E+04, 6.8803E+04, 6.9492E+04, 7.0187E+04,
     + 7.0886E+04, 7.1590E+04, 7.2299E+04, 7.3013E+04, 7.3732E+04,
     +  7.4456E+04, 7.5186E+04, 7.5920E+04, 7.6660E+04, 7.7405E+04,
     + 7.8154E+04, 7.8909E+04, 7.9670E+04, 8.0436E+04, 8.1207E+04,
     +  8.1983E+04, 8.2765E+04, 8.3553E+04, 8.4345E+04, 8.5144E+04,
     + 8.5948E+04, 8.6758E+04, 8.7573E+04, 8.8394E+04, 8.9220E+04,
     +  9.0052E+04, 9.0891E+04, 9.1734E+04, 9.2584E+04, 9.3439E+04,
     + 9.4301E+04, 9.5169E+04, 9.6041E+04, 9.6921E+04, 9.7807E+04,
     +  9.8698E+04, 9.9596E+04, 1.0050E+05, 1.0141E+05, 1.0233E+05,
     + 1.0325E+05, 1.0418E+05, 1.0511E+05, 1.0605E+05, 1.0700E+05,
     +  1.0795E+05, 1.0892E+05, 1.0988E+05, 1.1086E+05, 1.1184E+05,
     + 1.1282E+05, 1.1382E+05, 1.1482E+05, 1.1583E+05, 1.1684E+05,
     +  1.1786E+05, 1.1889E+05, 1.1992E+05, 1.2096E+05, 1.2201E+05,
     + 1.2307E+05, 1.2413E+05, 1.2520E+05, 1.2627E+05, 1.2736E+05,
     +  1.2845E+05, 1.2955E+05, 1.3065E+05, 1.3176E+05, 1.3288E+05,
     + 1.3401E+05, 1.3515E+05, 1.3629E+05, 1.3744E+05, 1.3860E+05,
     +  1.3976E+05, 1.4093E+05, 1.4211E+05, 1.4330E+05, 1.4450E+05,
     + 1.4570E+05, 1.4691E+05, 1.4813E+05, 1.4936E+05, 1.5059E+05,
     +  1.5183E+05, 1.5308E+05, 1.5434E+05, 1.5561E+05, 1.5689E+05,
     + 1.5817E+05, 1.5946E+05, 1.6076E+05, 1.6207E+05, 1.6339E+05,
     +  1.6471E+05, 1.6604E+05, 1.6739E+05, 1.6874E+05, 1.7009E+05,
     + 1.7146E+05, 1.7284E+05, 1.7422E+05, 1.7562E+05, 1.7702E+05,
     +  1.7843E+05, 1.7985E+05, 1.8128E+05, 1.8272E+05, 1.8416E+05,
     + 1.8562E+05, 1.8708E+05, 1.8856E+05, 1.9004E+05, 1.9153E+05,
     +  1.9304E+05, 1.9455E+05, 1.9607E+05, 1.9760E+05, 1.9914E+05,
     + 2.0068E+05, 2.0224E+05, 2.0381E+05, 2.0539E+05, 2.0697E+05,
     +  2.0857E+05, 2.1018E+05, 2.1179E+05, 2.1342E+05, 2.1505E+05,
     + 2.1670E+05, 2.1836E+05, 2.2002E+05, 2.2170E+05, 2.2338E+05,
     +  2.2508E+05, 2.2679E+05, 2.2850E+05, 2.3023E+05, 2.3197E+05,
     + 2.3371E+05, 2.3547E+05, 2.3724E+05, 2.3902E+05, 2.4081E+05,
     +  2.4261E+05, 2.4442E+05, 2.4624E+05, 2.4808E+05, 2.4992E+05,
     + 2.5177E+05, 2.5364E+05, 2.5551E+05, 2.5740E+05, 2.5930E+05,
     +  2.6121E+05, 2.6313E+05, 2.6506E+05, 2.6700E+05, 2.6896E+05,
     + 2.7092E+05, 2.7290E+05, 2.7489E+05, 2.7689E+05, 2.7890E+05,
     +  2.8092E+05, 2.8296E+05, 2.8500E+05, 2.8706E+05, 2.8913E+05,
     + 2.9121E+05, 2.9331E+05, 2.9541E+05, 2.9753E+05, 2.9966E+05,
     +  3.0180E+05, 3.0395E+05, 3.0612E+05, 3.0830E+05, 3.1049E+05,
     + 3.1269E+05, 3.1491E+05, 3.1714E+05, 3.1938E+05, 3.2163E+05,
     +  3.2390E+05, 3.2617E+05, 3.2847E+05, 3.3077E+05, 3.3309E+05,
     + 3.3542E+05, 3.3776E+05, 3.4011E+05, 3.4248E+05, 3.4487E+05,
     +  3.4726E+05, 3.4967E+05, 3.5209E+05, 3.5453E+05, 3.5697E+05,
     + 3.5944E+05, 3.6191E+05, 3.6440E+05, 3.6690E+05, 3.6942E+05,
     +  3.7195E+05, 3.7449E+05, 3.7705E+05, 3.7962E+05, 3.8221E+05,
     + 3.8481E+05, 3.8742E+05, 3.9005E+05, 3.9269E+05, 3.9535E+05,
     +  3.9801E+05, 4.0070E+05, 4.0340E+05, 4.0611E+05, 4.0884E+05,
     + 4.1158E+05, 4.1434E+05, 4.1711E+05, 4.1990E+05, 4.2270E+05,
     +  4.2551E+05, 4.2835E+05, 4.3119E+05, 4.3405E+05, 4.3693E+05,
     + 4.3982E+05, 4.4273E+05, 4.4565E+05, 4.4859E+05, 4.5154E+05,
     +  4.5451E+05, 4.5749E+05, 4.6049E+05, 4.6350E+05, 4.6654E+05,
     + 4.6958E+05, 4.7264E+05, 4.7572E+05, 4.7882E+05, 4.8192E+05,
     +  4.8505E+05, 4.8819E+05, 4.9135E+05, 4.9452E+05, 4.9771E+05,
     + 5.0092E+05, 5.0415E+05, 5.0739E+05, 5.1064E+05, 5.1392E+05,
     +  5.1721E+05, 5.2051E+05, 5.2384E+05, 5.2718E+05, 5.3053E+05,
     + 5.3391E+05, 5.3730E+05, 5.4071E+05, 5.4413E+05, 5.4757E+05,
     +  5.5103E+05, 5.5451E+05, 5.5801E+05, 5.6152E+05, 5.6505E+05,
     + 5.6860E+05, 5.7216E+05, 5.7574E+05, 5.7934E+05, 5.8296E+05,
     +  5.8660E+05, 5.9025E+05, 5.9393E+05, 5.9762E+05, 6.0133E+05,
     + 6.0505E+05, 6.0880E+05, 6.1256E+05, 6.1635E+05, 6.2015E+05,
     +  6.2397E+05, 6.2780E+05, 6.3166E+05, 6.3554E+05, 6.3943E+05,
     + 6.4335E+05, 6.4728E+05, 6.5123E+05, 6.5520E+05, 6.5919E+05,
     +  6.6320E+05, 6.6723E+05, 6.7128E+05, 6.7535E+05, 6.7943E+05,
     + 6.8354E+05, 6.8767E+05, 6.9182E+05, 6.9598E+05, 7.0017E+05,
     +  7.0437E+05, 7.0860E+05, 7.1285E+05, 7.1711E+05, 7.2140E+05,
     + 7.2571E+05, 7.3004E+05, 7.3438E+05, 7.3875E+05, 7.4314E+05,
     +  7.4755E+05, 7.5198E+05, 7.5643E+05, 7.6091E+05, 7.6540E+05,
     + 7.6991E+05, 7.7445E+05, 7.7901E+05, 7.8358E+05, 7.8818E+05,
     +  7.9280E+05, 7.9745E+05, 8.0211E+05, 8.0680E+05, 8.1150E+05,
     + 8.1623E+05, 8.2098E+05, 8.2575E+05, 8.3055E+05, 8.3537E+05,
     +  8.4021E+05, 8.4507E+05, 8.4995E+05, 8.5486E+05, 8.5979E+05,
     + 8.6474E+05, 8.6971E+05, 8.7471E+05, 8.7972E+05, 8.8477E+05,
     +  8.8983E+05, 8.9492E+05, 9.0003E+05, 9.0516E+05, 9.1032E+05,
     + 9.1550E+05, 9.2070E+05, 9.2593E+05, 9.3118E+05, 9.3645E+05,
     +  9.4175E+05, 9.4707E+05, 9.5242E+05, 9.5779E+05, 9.6318E+05,
     + 9.6860E+05, 9.7404E+05, 9.7951E+05, 9.8499E+05, 9.9051E+05,
     +  9.9605E+05, 1.0016E+06, 1.0072E+06, 1.0128E+06, 1.0184E+06,
     + 1.0241E+06, 1.0298E+06, 1.0355E+06, 1.0412E+06, 1.0470E+06,
     +  1.0528E+06, 1.0586E+06, 1.0644E+06, 1.0703E+06, 1.0762E+06,
     + 1.0821E+06, 1.0881E+06, 1.0940E+06, 1.1000E+06, 1.1060E+06,
     +  1.1121E+06, 1.1182E+06, 1.1242E+06, 1.1304E+06, 1.1365E+06,
     + 1.1427E+06, 1.1489E+06, 1.1551E+06, 1.1614E+06, 1.1677E+06,
     +  1.1740E+06, 1.1803E+06, 1.1867E+06, 1.1931E+06, 1.1995E+06,
     + 1.2059E+06, 1.2124E+06, 1.2189E+06, 1.2254E+06, 1.2320E+06,
     +  1.2386E+06, 1.2452E+06, 1.2518E+06, 1.2585E+06, 1.2652E+06,
     + 1.2719E+06, 1.2787E+06, 1.2854E+06, 1.2923E+06, 1.2991E+06,
     +  1.3060E+06, 1.3128E+06, 1.3198E+06, 1.3267E+06, 1.3337E+06,
     + 1.3407E+06, 1.3478E+06, 1.3548E+06, 1.3619E+06, 1.3690E+06,
     +  1.3762E+06, 1.3834E+06, 1.3906E+06, 1.3979E+06, 1.4051E+06,
     + 1.4124E+06, 1.4198E+06, 1.4271E+06, 1.4345E+06, 1.4420E+06,
     +  1.4494E+06, 1.4569E+06, 1.4644E+06, 1.4720E+06, 1.4795E+06,
     + 1.4872E+06, 1.4948E+06, 1.5025E+06, 1.5102E+06, 1.5179E+06,
     +  1.5257E+06, 1.5335E+06, 1.5413E+06, 1.5492E+06, 1.5570E+06,
     + 1.5650E+06, 1.5729E+06, 1.5809E+06, 1.5889E+06, 1.5970E+06,
     +  1.6051E+06, 1.6132E+06, 1.6213E+06, 1.6295E+06, 1.6377E+06,
     + 1.6460E+06, 1.6542E+06, 1.6626E+06, 1.6709E+06, 1.6793E+06,
     +  1.6877E+06, 1.6961E+06, 1.7046E+06, 1.7131E+06, 1.7217E+06,
     + 1.7302E+06, 1.7389E+06, 1.7475E+06, 1.7562E+06, 1.7649E+06,
     +  1.7736E+06, 1.7824E+06, 1.7912E+06, 1.8001E+06, 1.8090E+06,
     + 1.8179E+06, 1.8269E+06, 1.8358E+06, 1.8449E+06, 1.8539E+06,
     +  1.8630E+06, 1.8722E+06, 1.8813E+06, 1.8905E+06, 1.8998E+06,
     + 1.9090E+06, 1.9183E+06, 1.9277E+06, 1.9371E+06, 1.9465E+06,
     +  1.9559E+06, 1.9654E+06, 1.9749E+06, 1.9845E+06, 1.9941E+06,
     + 2.0037E+06, 2.0134E+06, 2.0231E+06, 2.0329E+06, 2.0426E+06,
     +  2.0525E+06, 2.0623E+06, 2.0722E+06, 2.0821E+06, 2.0921E+06,
     + 2.1021E+06, 2.1122E+06, 2.1222E+06, 2.1324E+06, 2.1425E+06,
     +  2.1527E+06, 2.1629E+06, 2.1732E+06, 2.1835E+06, 2.1939E+06,
     + 2.2043E+06, 2.2147E+06, 2.2252E+06, 2.2357E+06, 2.2462E+06,
     +  2.2568E+06, 2.2674E+06, 2.2781E+06, 2.2888E+06, 2.2995E+06,
     + 2.3103E+06, 2.3211E+06, 2.3320E+06, 2.3429E+06, 2.3538E+06,
     +  2.3648E+06, 2.3758E+06, 2.3869E+06, 2.3980E+06, 2.4091E+06,
     + 2.4203E+06, 2.4316E+06, 2.4428E+06, 2.4541E+06, 2.4655E+06,
     +  2.4769E+06, 2.4883E+06, 2.4998E+06, 2.5113E+06, 2.5228E+06,
     + 2.5345E+06, 2.5461E+06, 2.5578E+06, 2.5695E+06, 2.5813E+06,
     +  2.5931E+06, 2.6049E+06, 2.6168E+06, 2.6288E+06, 2.6407E+06,
     + 2.6528E+06, 2.6648E+06, 2.6770E+06, 2.6891E+06, 2.7013E+06,
     +  2.7135E+06, 2.7258E+06, 2.7382E+06, 2.7505E+06, 2.7630E+06,
     + 2.7754E+06, 2.7879E+06, 2.8005E+06, 2.8131E+06, 2.8257E+06,
     +  2.8384E+06, 2.8511E+06, 2.8639E+06, 2.8767E+06, 2.8896E+06,
     + 2.9025E+06, 2.9154E+06, 2.9284E+06, 2.9415E+06, 2.9546E+06,
     +  2.9677E+06, 2.9809E+06, 2.9941E+06, 3.0074E+06, 3.0207E+06,
     + 3.0341E+06, 3.0475E+06, 3.0610E+06, 3.0745E+06, 3.0881E+06,
     +  3.1017E+06, 3.1153E+06, 3.1290E+06, 3.1428E+06, 3.1566E+06,
     + 3.1704E+06, 3.1843E+06, 3.1982E+06, 3.2122E+06, 3.2262E+06,
     +  3.2403E+06, 3.2545E+06, 3.2686E+06, 3.2829E+06, 3.2971E+06,
     + 3.3115E+06, 3.3258E+06, 3.3403E+06, 3.3547E+06, 3.3693E+06,
     +  3.3838E+06, 3.3984E+06, 3.4131E+06, 3.4278E+06, 3.4426E+06,
     + 3.4574E+06, 3.4723E+06, 3.4872E+06, 3.5022E+06, 3.5172E+06,
     +  3.5323E+06, 3.5474E+06, 3.5626E+06, 3.5778E+06, 3.5931E+06,
     + 3.6084E+06, 3.6238E+06, 3.6392E+06, 3.6547E+06, 3.6703E+06,
     +  3.6859E+06, 3.7015E+06, 3.7172E+06, 3.7329E+06, 3.7487E+06,
     + 3.7646E+06, 3.7805E+06, 3.7964E+06, 3.8124E+06, 3.8285E+06,
     +  3.8446E+06, 3.8608E+06, 3.8770E+06, 3.8933E+06, 3.9096E+06,
     + 3.9260E+06, 3.9424E+06, 3.9589E+06, 3.9755E+06, 3.9921E+06,
     +  4.0087E+06, 4.0254E+06, 4.0422E+06, 4.0590E+06, 4.0759E+06,
     + 4.0928E+06, 4.1098E+06, 4.1269E+06, 4.1440E+06, 4.1611E+06,
     +  4.1783E+06, 4.1956E+06, 4.2129E+06, 4.2303E+06, 4.2477E+06,
     + 4.2652E+06, 4.2828E+06, 4.3004E+06, 4.3180E+06, 4.3357E+06,
     +  4.3535E+06, 4.3713E+06, 4.3892E+06, 4.4072E+06, 4.4252E+06,
     + 4.4432E+06, 4.4614E+06, 4.4796E+06, 4.4978E+06, 4.5161E+06,
     +  4.5344E+06, 4.5529E+06, 4.5713E+06, 4.5899E+06, 4.6084E+06,
     + 4.6271E+06, 4.6458E+06, 4.6646E+06, 4.6834E+06, 4.7023E+06,
     +  4.7212E+06, 4.7402E+06, 4.7593E+06, 4.7784E+06, 4.7976E+06,
     + 4.8169E+06, 4.8362E+06, 4.8556E+06, 4.8750E+06, 4.8945E+06,
     +  4.9140E+06, 4.9337E+06, 4.9533E+06, 4.9731E+06, 4.9929E+06,
     + 5.0127E+06, 5.0327E+06, 5.0527E+06, 5.0727E+06, 5.0928E+06,
     +  5.1130E+06, 5.1332E+06, 5.1535E+06, 5.1739E+06, 5.1943E+06,
     + 5.2148E+06, 5.2354E+06, 5.2560E+06, 5.2767E+06, 5.2974E+06,
     +  5.3182E+06, 5.3391E+06, 5.3601E+06, 5.3811E+06, 5.4021E+06,
     + 5.4233E+06, 5.4445E+06, 5.4657E+06, 5.4871E+06, 5.5085E+06,
     +  5.5299E+06, 5.5515E+06, 5.5730E+06, 5.5947E+06, 5.6164E+06,
     + 5.6382E+06, 5.6601E+06, 5.6820E+06, 5.7040E+06, 5.7261E+06,
     +  5.7482E+06, 5.7704E+06, 5.7926E+06, 5.8150E+06, 5.8374E+06,
     + 5.8598E+06, 5.8824E+06, 5.9050E+06, 5.9276E+06, 5.9504E+06,
     +  5.9732E+06, 5.9961E+06, 6.0190E+06, 6.0420E+06, 6.0651E+06,
     + 6.0882E+06, 6.1115E+06, 6.1347E+06, 6.1581E+06, 6.1815E+06,
     +  6.2050E+06, 6.2286E+06, 6.2522E+06, 6.2760E+06, 6.2997E+06,
     + 6.3236E+06, 6.3475E+06, 6.3715E+06, 6.3956E+06, 6.4197E+06,
     +  6.4439E+06, 6.4682E+06, 6.4926E+06, 6.5170E+06, 6.5415E+06,
     + 6.5661E+06, 6.5907E+06, 6.6154E+06, 6.6402E+06, 6.6651E+06,
     +  6.6900E+06, 6.7150E+06, 6.7401E+06, 6.7652E+06, 6.7905E+06,
     + 6.8158E+06, 6.8412E+06, 6.8666E+06, 6.8921E+06, 6.9177E+06,
     +  6.9434E+06, 6.9692E+06, 6.9950E+06, 7.0209E+06, 7.0469E+06,
     + 7.0729E+06, 7.0990E+06, 7.1252E+06, 7.1515E+06, 7.1779E+06,
     +  7.2043E+06, 7.2308E+06, 7.2574E+06, 7.2841E+06, 7.3108E+06,
     + 7.3376E+06, 7.3645E+06, 7.3915E+06, 7.4185E+06, 7.4457E+06,
     +  7.4729E+06, 7.5002E+06, 7.5275E+06, 7.5550E+06, 7.5825E+06,
     + 7.6101E+06, 7.6377E+06, 7.6655E+06, 7.6933E+06, 7.7213E+06,
     +  7.7492E+06, 7.7773E+06, 7.8055E+06, 7.8337E+06, 7.8620E+06,
     + 7.8904E+06, 7.9189E+06, 7.9474E+06, 7.9761E+06, 8.0048E+06,
     +  8.0336E+06, 8.0625E+06, 8.0914E+06, 8.1205E+06, 8.1496E+06,
     + 8.1788E+06, 8.2081E+06, 8.2375E+06, 8.2669E+06, 8.2965E+06,
     +  8.3261E+06, 8.3558E+06, 8.3856E+06, 8.4155E+06, 8.4454E+06,
     + 8.4755E+06, 8.5056E+06, 8.5358E+06, 8.5661E+06, 8.5965E+06,
     +  8.6269E+06, 8.6575E+06, 8.6881E+06, 8.7188E+06, 8.7496E+06,
     + 8.7805E+06, 8.8115E+06, 8.8425E+06, 8.8737E+06, 8.9049E+06,
     +  8.9362E+06, 8.9676E+06, 8.9991E+06, 9.0307E+06, 9.0624E+06,
     + 9.0941E+06, 9.1260E+06, 9.1579E+06, 9.1899E+06, 9.2220E+06,
     +  9.2542E+06, 9.2865E+06, 9.3188E+06, 9.3513E+06, 9.3838E+06,
     + 9.4165E+06, 9.4492E+06, 9.4820E+06, 9.5149E+06, 9.5479E+06,
     +  9.5810E+06, 9.6142E+06, 9.6474E+06, 9.6808E+06, 9.7142E+06,
     + 9.7477E+06, 9.7814E+06, 9.8151E+06, 9.8489E+06, 9.8828E+06,
     +  9.9168E+06, 9.9509E+06, 9.9850E+06, 1.0019E+07, 1.0054E+07,
     + 1.0088E+07, 1.0123E+07, 1.0157E+07, 1.0192E+07, 1.0227E+07,
     +  1.0262E+07, 1.0297E+07, 1.0332E+07, 1.0367E+07, 1.0402E+07,
     + 1.0438E+07, 1.0473E+07, 1.0509E+07, 1.0545E+07, 1.0580E+07,
     +  1.0616E+07, 1.0652E+07, 1.0688E+07, 1.0724E+07, 1.0761E+07,
     + 1.0797E+07, 1.0833E+07, 1.0870E+07, 1.0907E+07, 1.0943E+07,
     +  1.0980E+07, 1.1017E+07, 1.1054E+07, 1.1091E+07, 1.1128E+07,
     + 1.1166E+07, 1.1203E+07, 1.1241E+07, 1.1278E+07, 1.1316E+07,
     +  1.1354E+07, 1.1392E+07, 1.1430E+07, 1.1468E+07, 1.1506E+07,
     + 1.1544E+07, 1.1583E+07, 1.1621E+07, 1.1660E+07, 1.1699E+07,
     +  1.1737E+07, 1.1776E+07, 1.1815E+07, 1.1854E+07, 1.1894E+07,
     + 1.1933E+07, 1.1972E+07, 1.2012E+07, 1.2051E+07, 1.2091E+07,
     +  1.2131E+07, 1.2171E+07, 1.2211E+07, 1.2251E+07, 1.2291E+07,
     + 1.2332E+07, 1.2372E+07, 1.2413E+07, 1.2453E+07, 1.2494E+07,
     +  1.2535E+07, 1.2576E+07, 1.2617E+07, 1.2658E+07, 1.2700E+07,
     + 1.2741E+07, 1.2783E+07, 1.2824E+07, 1.2866E+07, 1.2908E+07,
     +  1.2950E+07, 1.2992E+07, 1.3034E+07, 1.3076E+07, 1.3118E+07,
     + 1.3161E+07, 1.3203E+07, 1.3246E+07, 1.3289E+07, 1.3332E+07,
     +  1.3375E+07, 1.3418E+07, 1.3461E+07, 1.3505E+07, 1.3548E+07,
     + 1.3592E+07, 1.3635E+07, 1.3679E+07, 1.3723E+07, 1.3767E+07,
     +  1.3811E+07, 1.3855E+07, 1.3900E+07, 1.3944E+07, 1.3989E+07,
     + 1.4033E+07, 1.4078E+07, 1.4123E+07, 1.4168E+07, 1.4213E+07,
     +  1.4258E+07, 1.4304E+07, 1.4349E+07, 1.4395E+07, 1.4440E+07,
     + 1.4486E+07, 1.4532E+07, 1.4578E+07, 1.4624E+07, 1.4670E+07,
     +  1.4717E+07, 1.4763E+07, 1.4810E+07, 1.4857E+07, 1.4903E+07,
     + 1.4950E+07, 1.4997E+07, 1.5045E+07, 1.5092E+07, 1.5139E+07,
     +  1.5187E+07, 1.5235E+07, 1.5282E+07, 1.5330E+07, 1.5378E+07,
     + 1.5426E+07, 1.5475E+07, 1.5523E+07, 1.5571E+07, 1.5620E+07,
     +  1.5669E+07, 1.5717E+07, 1.5766E+07, 1.5816E+07, 1.5865E+07,
     + 1.5914E+07, 1.5963E+07, 1.6013E+07, 1.6063E+07, 1.6112E+07,
     +  1.6162E+07, 1.6212E+07, 1.6263E+07, 1.6313E+07, 1.6363E+07,
     + 1.6414E+07, 1.6464E+07, 1.6515E+07, 1.6566E+07, 1.6617E+07,
     +  1.6668E+07, 1.6720E+07, 1.6771E+07, 1.6822E+07, 1.6874E+07,
     + 1.6926E+07, 1.6978E+07, 1.7030E+07, 1.7082E+07, 1.7134E+07,
     +  1.7187E+07, 1.7239E+07, 1.7292E+07, 1.7344E+07, 1.7397E+07,
     + 1.7450E+07, 1.7504E+07, 1.7557E+07, 1.7610E+07, 1.7664E+07,
     +  1.7717E+07, 1.7771E+07, 1.7825E+07, 1.7879E+07, 1.7933E+07,
     + 1.7988E+07, 1.8042E+07, 1.8097E+07, 1.8151E+07, 1.8206E+07,
     +  1.8261E+07, 1.8316E+07, 1.8371E+07, 1.8427E+07, 1.8482E+07,
     + 1.8538E+07, 1.8593E+07, 1.8649E+07, 1.8705E+07, 1.8761E+07,
     +  1.8818E+07, 1.8874E+07, 1.8931E+07, 1.8987E+07, 1.9044E+07,
     + 1.9101E+07, 1.9158E+07, 1.9215E+07, 1.9273E+07, 1.9330E+07,
     +  1.9388E+07, 1.9445E+07, 1.9503E+07, 1.9561E+07, 1.9619E+07,
     + 1.9678E+07, 1.9736E+07, 1.9795E+07, 1.9853E+07, 1.9912E+07,
     +  1.9971E+07, 2.0030E+07, 2.0089E+07, 2.0149E+07, 2.0208E+07,
     + 2.0268E+07, 2.0328E+07, 2.0388E+07, 2.0448E+07, 2.0508E+07,
     +  2.0568E+07, 2.0629E+07, 2.0689E+07, 2.0750E+07, 2.0811E+07,
     + 2.0872E+07, 2.0933E+07, 2.0994E+07, 2.1056E+07, 2.1118E+07,
     +  2.1179E+07, 2.1241E+07, 2.1303E+07, 2.1365E+07, 2.1428E+07,
     + 2.1490E+07, 2.1553E+07, 2.1615E+07, 2.1678E+07, 2.1741E+07,
     +  2.1805E+07, 2.1868E+07, 2.1931E+07, 2.1995E+07, 2.2059E+07,
     + 2.2123E+07, 2.2187E+07, 2.2251E+07, 2.2315E+07, 2.2380E+07,
     +  2.2444E+07, 2.2509E+07, 2.2574E+07, 2.2639E+07, 2.2704E+07,
     + 2.2770E+07, 2.2835E+07, 2.2901E+07, 2.2967E+07, 2.3032E+07,
     +  2.3099E+07, 2.3165E+07, 2.3231E+07, 2.3298E+07, 2.3364E+07,
     + 2.3431E+07, 2.3498E+07, 2.3565E+07, 2.3633E+07, 2.3700E+07,
     +  2.3768E+07, 2.3836E+07, 2.3903E+07, 2.3972E+07, 2.4040E+07,
     + 2.4108E+07, 2.4177E+07, 2.4245E+07, 2.4314E+07, 2.4383E+07,
     +  2.4452E+07, 2.4521E+07, 2.4591E+07, 2.4661E+07, 2.4730E+07,
     + 2.4800E+07, 2.4870E+07, 2.4940E+07, 2.5011E+07, 2.5081E+07,
     +  2.5152E+07, 2.5223E+07, 2.5294E+07, 2.5365E+07, 2.5436E+07,
     + 2.5508E+07, 2.5579E+07, 2.5651E+07, 2.5723E+07, 2.5795E+07,
     +  2.5867E+07, 2.5940E+07, 2.6012E+07, 2.6085E+07, 2.6158E+07,
     + 2.6231E+07, 2.6304E+07, 2.6378E+07, 2.6451E+07, 2.6525E+07,
     +  2.6599E+07, 2.6673E+07, 2.6747E+07, 2.6821E+07, 2.6896E+07,
     + 2.6971E+07, 2.7045E+07, 2.7120E+07, 2.7195E+07, 2.7271E+07,
     +  2.7346E+07, 2.7422E+07, 2.7498E+07, 2.7574E+07, 2.7650E+07,
     + 2.7726E+07, 2.7803E+07, 2.7879E+07, 2.7956E+07, 2.8033E+07,
     +  2.8110E+07, 2.8188E+07, 2.8265E+07, 2.8343E+07, 2.8420E+07,
     + 2.8498E+07, 2.8577E+07, 2.8655E+07, 2.8733E+07, 2.8812E+07,
     +  2.8891E+07, 2.8970E+07, 2.9049E+07, 2.9128E+07, 2.9208E+07,
     + 2.9287E+07, 2.9367E+07, 2.9447E+07, 2.9527E+07, 2.9608E+07,
     +  2.9688E+07, 2.9769E+07, 2.9850E+07, 2.9931E+07, 3.0012E+07,
     + 3.0094E+07, 3.0175E+07, 3.0257E+07, 3.0339E+07, 3.0421E+07,
     +  3.0503E+07, 3.0585E+07, 3.0668E+07, 3.0751E+07, 3.0834E+07,
     + 3.0917E+07, 3.1000E+07, 3.1084E+07, 3.1167E+07, 3.1251E+07,
     +  3.1335E+07, 3.1419E+07, 3.1504E+07, 3.1588E+07, 3.1673E+07,
     + 3.1758E+07, 3.1843E+07, 3.1928E+07, 3.2014E+07, 3.2099E+07,
     +  3.2185E+07, 3.2271E+07, 3.2357E+07, 3.2444E+07, 3.2530E+07,
     + 3.2617E+07, 3.2704E+07, 3.2791E+07, 3.2878E+07, 3.2966E+07,
     +  3.3053E+07, 3.3141E+07, 3.3229E+07, 3.3317E+07, 3.3406E+07,
     + 3.3494E+07, 3.3583E+07, 3.3672E+07, 3.3761E+07, 3.3850E+07,
     +  3.3939E+07, 3.4029E+07, 3.4119E+07, 3.4209E+07, 3.4299E+07,
     + 3.4390E+07, 3.4480E+07, 3.4571E+07, 3.4662E+07, 3.4753E+07,
     +  3.4844E+07, 3.4936E+07, 3.5028E+07, 3.5119E+07, 3.5212E+07,
     + 3.5304E+07, 3.5396E+07, 3.5489E+07, 3.5582E+07, 3.5675E+07,
     +  3.5768E+07, 3.5861E+07, 3.5955E+07, 3.6049E+07, 3.6143E+07,
     + 3.6237E+07, 3.6331E+07, 3.6426E+07, 3.6521E+07, 3.6616E+07,
     +  3.6711E+07, 3.6806E+07, 3.6902E+07, 3.6997E+07, 3.7093E+07,
     + 3.7190E+07, 3.7286E+07, 3.7382E+07, 3.7479E+07, 3.7576E+07,
     +  3.7673E+07, 3.7770E+07, 3.7868E+07, 3.7966E+07, 3.8064E+07,
     + 3.8162E+07, 3.8260E+07, 3.8358E+07, 3.8457E+07, 3.8556E+07,
     +  3.8655E+07, 3.8754E+07, 3.8854E+07, 3.8954E+07, 3.9054E+07,
     + 3.9154E+07, 3.9254E+07, 3.9354E+07, 3.9455E+07, 3.9556E+07,
     +  3.9657E+07, 3.9758E+07, 3.9860E+07, 3.9962E+07, 4.0064E+07,
     + 4.0166E+07, 4.0268E+07, 4.0371E+07, 4.0473E+07, 4.0576E+07,
     +  4.0679E+07, 4.0783E+07, 4.0886E+07, 4.0990E+07, 4.1094E+07,
     + 4.1198E+07, 4.1303E+07, 4.1407E+07, 4.1512E+07, 4.1617E+07,
     +  4.1722E+07, 4.1828E+07, 4.1933E+07, 4.2039E+07, 4.2145E+07,
     + 4.2252E+07, 4.2358E+07, 4.2465E+07, 4.2572E+07, 4.2679E+07,
     +  4.2786E+07, 4.2894E+07, 4.3001E+07, 4.3109E+07, 4.3218E+07,
     + 4.3326E+07, 4.3435E+07, 4.3543E+07, 4.3652E+07, 4.3762E+07,
     +  4.3871E+07, 4.3981E+07, 4.4091E+07, 4.4201E+07, 4.4311E+07,
     + 4.4422E+07, 4.4532E+07, 4.4643E+07, 4.4755E+07, 4.4866E+07,
     +  4.4978E+07, 4.5090E+07, 4.5202E+07, 4.5314E+07, 4.5426E+07,
     + 4.5539E+07, 4.5652E+07, 4.5765E+07, 4.5879E+07, 4.5992E+07,
     +  4.6106E+07, 4.6220E+07, 4.6334E+07, 4.6449E+07, 4.6564E+07,
     + 4.6679E+07, 4.6794E+07, 4.6909E+07, 4.7025E+07, 4.7141E+07,
     +  4.7257E+07, 4.7373E+07, 4.7489E+07, 4.7606E+07, 4.7723E+07,
     + 4.7840E+07, 4.7958E+07, 4.8075E+07, 4.8193E+07, 4.8311E+07,
     +  4.8430E+07, 4.8548E+07, 4.8667E+07, 4.8786E+07, 4.8905E+07,
     + 4.9025E+07, 4.9144E+07, 4.9264E+07, 4.9384E+07, 4.9505E+07,
     +  4.9625E+07, 4.9746E+07, 4.9867E+07, 4.9988E+07, 5.0110E+07,
     + 5.0232E+07, 5.0354E+07, 5.0476E+07, 5.0598E+07, 5.0721E+07,
     +  5.0844E+07, 5.0967E+07, 5.1091E+07, 5.1214E+07, 5.1338E+07,
     + 5.1462E+07, 5.1587E+07, 5.1711E+07, 5.1836E+07, 5.1961E+07,
     +  5.2086E+07, 5.2212E+07, 5.2338E+07, 5.2464E+07, 5.2590E+07,
     + 5.2716E+07, 5.2843E+07, 5.2970E+07, 5.3097E+07, 5.3225E+07,
     +  5.3352E+07, 5.3480E+07, 5.3608E+07, 5.3737E+07, 5.3865E+07,
     + 5.3994E+07, 5.4123E+07, 5.4253E+07, 5.4382E+07, 5.4512E+07,
     +  5.4642E+07, 5.4773E+07, 5.4903E+07, 5.5034E+07, 5.5165E+07,
     + 5.5296E+07, 5.5428E+07, 5.5560E+07, 5.5692E+07, 5.5824E+07,
     +  5.5957E+07, 5.6089E+07, 5.6223E+07, 5.6356E+07, 5.6489E+07,
     + 5.6623E+07, 5.6757E+07, 5.6891E+07, 5.7026E+07, 5.7161E+07,
     +  5.7296E+07, 5.7431E+07, 5.7567E+07, 5.7702E+07, 5.7838E+07,
     + 5.7975E+07, 5.8111E+07, 5.8248E+07, 5.8385E+07, 5.8522E+07,
     +  5.8660E+07, 5.8798E+07, 5.8936E+07, 5.9074E+07, 5.9213E+07,
     + 5.9352E+07, 5.9491E+07, 5.9630E+07, 5.9770E+07, 5.9910E+07,
     +  6.0050E+07, 6.0190E+07, 6.0331E+07, 6.0472E+07, 6.0613E+07,
     + 6.0754E+07, 6.0896E+07, 6.1038E+07, 6.1180E+07, 6.1322E+07,
     +  6.1465E+07, 6.1608E+07, 6.1751E+07, 6.1895E+07, 6.2039E+07,
     + 6.2183E+07, 6.2327E+07, 6.2471E+07, 6.2616E+07, 6.2761E+07,
     +  6.2907E+07, 6.3052E+07, 6.3198E+07, 6.3344E+07, 6.3491E+07,
     + 6.3637E+07, 6.3784E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 8,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       768
      data (QofT( 9,J),J=1,2408)/ 1.0259E+01, 2.6644E+01, 4.7517E+01,
     +  7.2080E+01, 9.9844E+01, 1.3048E+02, 1.6373E+02, 1.9941E+02,
     + 2.3737E+02, 2.7747E+02, 3.1961E+02, 3.6369E+02, 4.0963E+02,
     +  4.5736E+02, 5.0681E+02, 5.5793E+02, 6.1066E+02, 6.6496E+02,
     + 7.2078E+02, 7.7809E+02, 8.9699E+02, 1.0214E+03, 1.1511E+03,
     +  1.2859E+03, 1.4256E+03, 1.5699E+03, 1.7189E+03, 1.8723E+03,
     + 2.0301E+03, 2.1920E+03, 2.3581E+03, 2.5281E+03, 2.7021E+03,
     +  2.8799E+03, 3.0614E+03, 3.2466E+03, 3.4354E+03, 3.6277E+03,
     + 3.8235E+03, 4.0228E+03, 4.2253E+03, 4.4312E+03, 4.6404E+03,
     +  4.8527E+03, 5.0682E+03, 5.2868E+03, 5.5084E+03, 5.7331E+03,
     + 5.9608E+03, 6.1915E+03, 6.4251E+03, 6.6615E+03, 6.9007E+03,
     +  7.1429E+03, 7.3877E+03, 7.6354E+03, 7.8858E+03, 8.1389E+03,
     + 8.3946E+03, 8.6530E+03, 8.9141E+03, 9.1778E+03, 9.4441E+03,
     +  9.7128E+03, 9.9842E+03, 1.0258E+04, 1.0535E+04, 1.0813E+04,
     + 1.1095E+04, 1.1379E+04, 1.1665E+04, 1.1954E+04, 1.2245E+04,
     +  1.2539E+04, 1.2835E+04, 1.3134E+04, 1.3435E+04, 1.3738E+04,
     + 1.4044E+04, 1.4352E+04, 1.4662E+04, 1.4975E+04, 1.5290E+04,
     +  1.5608E+04, 1.5928E+04, 1.6251E+04, 1.6575E+04, 1.6903E+04,
     + 1.7232E+04, 1.7564E+04, 1.7899E+04, 1.8235E+04, 1.8575E+04,
     +  1.8916E+04, 1.9260E+04, 1.9607E+04, 1.9956E+04, 2.0307E+04,
     + 2.0661E+04, 2.1017E+04, 2.1376E+04, 2.1737E+04, 2.2101E+04,
     +  2.2467E+04, 2.2836E+04, 2.3207E+04, 2.3581E+04, 2.3957E+04,
     + 2.4336E+04, 2.4718E+04, 2.5102E+04, 2.5489E+04, 2.5879E+04,
     +  2.6271E+04, 2.6665E+04, 2.7063E+04, 2.7463E+04, 2.7866E+04,
     + 2.8272E+04, 2.8680E+04, 2.9091E+04, 2.9505E+04, 2.9922E+04,
     +  3.0342E+04, 3.0764E+04, 3.1190E+04, 3.1618E+04, 3.2049E+04,
     + 3.2484E+04, 3.2921E+04, 3.3361E+04, 3.3804E+04, 3.4251E+04,
     +  3.4700E+04, 3.5152E+04, 3.5608E+04, 3.6067E+04, 3.6528E+04,
     + 3.6993E+04, 3.7461E+04, 3.7933E+04, 3.8408E+04, 3.8885E+04,
     +  3.9367E+04, 3.9851E+04, 4.0340E+04, 4.0831E+04, 4.1326E+04,
     + 4.1824E+04, 4.2326E+04, 4.2831E+04, 4.3340E+04, 4.3852E+04,
     +  4.4368E+04, 4.4887E+04, 4.5410E+04, 4.5937E+04, 4.6468E+04,
     + 4.7002E+04, 4.7540E+04, 4.8081E+04, 4.8627E+04, 4.9176E+04,
     +  4.9729E+04, 5.0286E+04, 5.0847E+04, 5.1412E+04, 5.1981E+04,
     + 5.2554E+04, 5.3131E+04, 5.3712E+04, 5.4297E+04, 5.4886E+04,
     +  5.5480E+04, 5.6077E+04, 5.6679E+04, 5.7284E+04, 5.7895E+04,
     + 5.8509E+04, 5.9128E+04, 5.9751E+04, 6.0379E+04, 6.1011E+04,
     +  6.1648E+04, 6.2289E+04, 6.2934E+04, 6.3584E+04, 6.4239E+04,
     + 6.4898E+04, 6.5562E+04, 6.6230E+04, 6.6903E+04, 6.7581E+04,
     +  6.8264E+04, 6.8952E+04, 6.9644E+04, 7.0341E+04, 7.1043E+04,
     + 7.1750E+04, 7.2462E+04, 7.3179E+04, 7.3901E+04, 7.4629E+04,
     +  7.5361E+04, 7.6098E+04, 7.6841E+04, 7.7588E+04, 7.8341E+04,
     + 7.9100E+04, 7.9863E+04, 8.0632E+04, 8.1405E+04, 8.2185E+04,
     +  8.2970E+04, 8.3761E+04, 8.4556E+04, 8.5358E+04, 8.6165E+04,
     + 8.6977E+04, 8.7796E+04, 8.8619E+04, 8.9449E+04, 9.0285E+04,
     +  9.1126E+04, 9.1973E+04, 9.2825E+04, 9.3684E+04, 9.4548E+04,
     + 9.5419E+04, 9.6295E+04, 9.7177E+04, 9.8066E+04, 9.8961E+04,
     +  9.9861E+04, 1.0077E+05, 1.0168E+05, 1.0260E+05, 1.0353E+05,
     + 1.0446E+05, 1.0540E+05, 1.0634E+05, 1.0729E+05, 1.0825E+05,
     +  1.0921E+05, 1.1018E+05, 1.1116E+05, 1.1214E+05, 1.1313E+05,
     + 1.1413E+05, 1.1513E+05, 1.1614E+05, 1.1716E+05, 1.1819E+05,
     +  1.1922E+05, 1.2025E+05, 1.2130E+05, 1.2235E+05, 1.2341E+05,
     + 1.2447E+05, 1.2555E+05, 1.2663E+05, 1.2771E+05, 1.2881E+05,
     +  1.2991E+05, 1.3102E+05, 1.3213E+05, 1.3326E+05, 1.3439E+05,
     + 1.3552E+05, 1.3667E+05, 1.3782E+05, 1.3898E+05, 1.4015E+05,
     +  1.4133E+05, 1.4251E+05, 1.4370E+05, 1.4490E+05, 1.4611E+05,
     + 1.4732E+05, 1.4854E+05, 1.4978E+05, 1.5101E+05, 1.5226E+05,
     +  1.5351E+05, 1.5478E+05, 1.5605E+05, 1.5733E+05, 1.5861E+05,
     + 1.5991E+05, 1.6121E+05, 1.6252E+05, 1.6384E+05, 1.6517E+05,
     +  1.6651E+05, 1.6785E+05, 1.6921E+05, 1.7057E+05, 1.7194E+05,
     + 1.7332E+05, 1.7471E+05, 1.7611E+05, 1.7751E+05, 1.7893E+05,
     +  1.8035E+05, 1.8178E+05, 1.8322E+05, 1.8467E+05, 1.8613E+05,
     + 1.8760E+05, 1.8908E+05, 1.9057E+05, 1.9206E+05, 1.9357E+05,
     +  1.9508E+05, 1.9661E+05, 1.9814E+05, 1.9968E+05, 2.0124E+05,
     + 2.0280E+05, 2.0437E+05, 2.0595E+05, 2.0754E+05, 2.0914E+05,
     +  2.1075E+05, 2.1237E+05, 2.1400E+05, 2.1564E+05, 2.1729E+05,
     + 2.1895E+05, 2.2062E+05, 2.2230E+05, 2.2399E+05, 2.2569E+05,
     +  2.2740E+05, 2.2912E+05, 2.3085E+05, 2.3259E+05, 2.3434E+05,
     + 2.3611E+05, 2.3788E+05, 2.3966E+05, 2.4146E+05, 2.4326E+05,
     +  2.4507E+05, 2.4690E+05, 2.4874E+05, 2.5058E+05, 2.5244E+05,
     + 2.5431E+05, 2.5619E+05, 2.5808E+05, 2.5998E+05, 2.6190E+05,
     +  2.6382E+05, 2.6576E+05, 2.6771E+05, 2.6966E+05, 2.7163E+05,
     + 2.7361E+05, 2.7561E+05, 2.7761E+05, 2.7963E+05, 2.8165E+05,
     +  2.8369E+05, 2.8574E+05, 2.8781E+05, 2.8988E+05, 2.9197E+05,
     + 2.9406E+05, 2.9617E+05, 2.9830E+05, 3.0043E+05, 3.0258E+05,
     +  3.0473E+05, 3.0691E+05, 3.0909E+05, 3.1128E+05, 3.1349E+05,
     + 3.1571E+05, 3.1794E+05, 3.2019E+05, 3.2245E+05, 3.2472E+05,
     +  3.2700E+05, 3.2930E+05, 3.3161E+05, 3.3393E+05, 3.3626E+05,
     + 3.3861E+05, 3.4097E+05, 3.4334E+05, 3.4573E+05, 3.4813E+05,
     +  3.5054E+05, 3.5297E+05, 3.5541E+05, 3.5786E+05, 3.6033E+05,
     + 3.6281E+05, 3.6530E+05, 3.6781E+05, 3.7033E+05, 3.7286E+05,
     +  3.7541E+05, 3.7797E+05, 3.8055E+05, 3.8314E+05, 3.8575E+05,
     + 3.8836E+05, 3.9100E+05, 3.9364E+05, 3.9631E+05, 3.9898E+05,
     +  4.0167E+05, 4.0437E+05, 4.0709E+05, 4.0982E+05, 4.1257E+05,
     + 4.1533E+05, 4.1811E+05, 4.2090E+05, 4.2371E+05, 4.2653E+05,
     +  4.2937E+05, 4.3222E+05, 4.3508E+05, 4.3797E+05, 4.4086E+05,
     + 4.4377E+05, 4.4670E+05, 4.4964E+05, 4.5260E+05, 4.5557E+05,
     +  4.5856E+05, 4.6157E+05, 4.6459E+05, 4.6762E+05, 4.7067E+05,
     + 4.7374E+05, 4.7683E+05, 4.7992E+05, 4.8304E+05, 4.8617E+05,
     +  4.8932E+05, 4.9248E+05, 4.9566E+05, 4.9886E+05, 5.0207E+05,
     + 5.0530E+05, 5.0855E+05, 5.1181E+05, 5.1509E+05, 5.1838E+05,
     +  5.2169E+05, 5.2502E+05, 5.2837E+05, 5.3173E+05, 5.3511E+05,
     + 5.3851E+05, 5.4192E+05, 5.4535E+05, 5.4880E+05, 5.5227E+05,
     +  5.5575E+05, 5.5925E+05, 5.6277E+05, 5.6630E+05, 5.6986E+05,
     + 5.7343E+05, 5.7702E+05, 5.8062E+05, 5.8425E+05, 5.8789E+05,
     +  5.9155E+05, 5.9523E+05, 5.9893E+05, 6.0264E+05, 6.0638E+05,
     + 6.1013E+05, 6.1390E+05, 6.1768E+05, 6.2149E+05, 6.2532E+05,
     +  6.2916E+05, 6.3302E+05, 6.3691E+05, 6.4081E+05, 6.4473E+05,
     + 6.4867E+05, 6.5262E+05, 6.5660E+05, 6.6060E+05, 6.6462E+05,
     +  6.6865E+05, 6.7270E+05, 6.7678E+05, 6.8087E+05, 6.8499E+05,
     + 6.8912E+05, 6.9327E+05, 6.9744E+05, 7.0164E+05, 7.0585E+05,
     +  7.1008E+05, 7.1434E+05, 7.1861E+05, 7.2290E+05, 7.2722E+05,
     + 7.3155E+05, 7.3591E+05, 7.4028E+05, 7.4468E+05, 7.4909E+05,
     +  7.5353E+05, 7.5799E+05, 7.6247E+05, 7.6697E+05, 7.7149E+05,
     + 7.7603E+05, 7.8059E+05, 7.8518E+05, 7.8978E+05, 7.9441E+05,
     +  7.9906E+05, 8.0373E+05, 8.0843E+05, 8.1314E+05, 8.1788E+05,
     + 8.2263E+05, 8.2741E+05, 8.3221E+05, 8.3704E+05, 8.4188E+05,
     +  8.4675E+05, 8.5164E+05, 8.5655E+05, 8.6149E+05, 8.6645E+05,
     + 8.7143E+05, 8.7643E+05, 8.8146E+05, 8.8651E+05, 8.9158E+05,
     +  8.9667E+05, 9.0179E+05, 9.0693E+05, 9.1210E+05, 9.1728E+05,
     + 9.2249E+05, 9.2773E+05, 9.3299E+05, 9.3827E+05, 9.4357E+05,
     +  9.4890E+05, 9.5426E+05, 9.5963E+05, 9.6503E+05, 9.7046E+05,
     + 9.7590E+05, 9.8138E+05, 9.8688E+05, 9.9240E+05, 9.9795E+05,
     +  1.0035E+06, 1.0091E+06, 1.0147E+06, 1.0204E+06, 1.0260E+06,
     + 1.0317E+06, 1.0375E+06, 1.0432E+06, 1.0490E+06, 1.0548E+06,
     +  1.0606E+06, 1.0664E+06, 1.0723E+06, 1.0782E+06, 1.0841E+06,
     + 1.0901E+06, 1.0961E+06, 1.1021E+06, 1.1081E+06, 1.1141E+06,
     +  1.1202E+06, 1.1263E+06, 1.1324E+06, 1.1386E+06, 1.1448E+06,
     + 1.1510E+06, 1.1572E+06, 1.1635E+06, 1.1698E+06, 1.1761E+06,
     +  1.1825E+06, 1.1888E+06, 1.1952E+06, 1.2017E+06, 1.2081E+06,
     + 1.2146E+06, 1.2211E+06, 1.2276E+06, 1.2342E+06, 1.2408E+06,
     +  1.2474E+06, 1.2541E+06, 1.2607E+06, 1.2674E+06, 1.2742E+06,
     + 1.2809E+06, 1.2877E+06, 1.2945E+06, 1.3014E+06, 1.3083E+06,
     +  1.3152E+06, 1.3221E+06, 1.3291E+06, 1.3360E+06, 1.3431E+06,
     + 1.3501E+06, 1.3572E+06, 1.3643E+06, 1.3714E+06, 1.3786E+06,
     +  1.3858E+06, 1.3930E+06, 1.4003E+06, 1.4076E+06, 1.4149E+06,
     + 1.4222E+06, 1.4296E+06, 1.4370E+06, 1.4444E+06, 1.4519E+06,
     +  1.4594E+06, 1.4669E+06, 1.4745E+06, 1.4821E+06, 1.4897E+06,
     + 1.4973E+06, 1.5050E+06, 1.5127E+06, 1.5205E+06, 1.5282E+06,
     +  1.5360E+06, 1.5439E+06, 1.5518E+06, 1.5597E+06, 1.5676E+06,
     + 1.5756E+06, 1.5836E+06, 1.5916E+06, 1.5996E+06, 1.6077E+06,
     +  1.6159E+06, 1.6240E+06, 1.6322E+06, 1.6404E+06, 1.6487E+06,
     + 1.6570E+06, 1.6653E+06, 1.6737E+06, 1.6820E+06, 1.6905E+06,
     +  1.6989E+06, 1.7074E+06, 1.7159E+06, 1.7245E+06, 1.7331E+06,
     + 1.7417E+06, 1.7503E+06, 1.7590E+06, 1.7678E+06, 1.7765E+06,
     +  1.7853E+06, 1.7941E+06, 1.8030E+06, 1.8119E+06, 1.8208E+06,
     + 1.8298E+06, 1.8388E+06, 1.8478E+06, 1.8569E+06, 1.8660E+06,
     +  1.8751E+06, 1.8843E+06, 1.8935E+06, 1.9028E+06, 1.9121E+06,
     + 1.9214E+06, 1.9307E+06, 1.9401E+06, 1.9495E+06, 1.9590E+06,
     +  1.9685E+06, 1.9780E+06, 1.9876E+06, 1.9972E+06, 2.0069E+06,
     + 2.0166E+06, 2.0263E+06, 2.0360E+06, 2.0458E+06, 2.0556E+06,
     +  2.0655E+06, 2.0754E+06, 2.0854E+06, 2.0953E+06, 2.1054E+06,
     + 2.1154E+06, 2.1255E+06, 2.1356E+06, 2.1458E+06, 2.1560E+06,
     +  2.1663E+06, 2.1765E+06, 2.1869E+06, 2.1972E+06, 2.2076E+06,
     + 2.2181E+06, 2.2285E+06, 2.2391E+06, 2.2496E+06, 2.2602E+06,
     +  2.2708E+06, 2.2815E+06, 2.2922E+06, 2.3030E+06, 2.3138E+06,
     + 2.3246E+06, 2.3355E+06, 2.3464E+06, 2.3573E+06, 2.3683E+06,
     +  2.3794E+06, 2.3904E+06, 2.4016E+06, 2.4127E+06, 2.4239E+06,
     + 2.4352E+06, 2.4464E+06, 2.4578E+06, 2.4691E+06, 2.4805E+06,
     +  2.4920E+06, 2.5034E+06, 2.5150E+06, 2.5265E+06, 2.5382E+06,
     + 2.5498E+06, 2.5615E+06, 2.5732E+06, 2.5850E+06, 2.5968E+06,
     +  2.6087E+06, 2.6206E+06, 2.6326E+06, 2.6446E+06, 2.6566E+06,
     + 2.6687E+06, 2.6808E+06, 2.6930E+06, 2.7052E+06, 2.7174E+06,
     +  2.7297E+06, 2.7421E+06, 2.7545E+06, 2.7669E+06, 2.7794E+06,
     + 2.7919E+06, 2.8045E+06, 2.8171E+06, 2.8297E+06, 2.8424E+06,
     +  2.8552E+06, 2.8679E+06, 2.8808E+06, 2.8936E+06, 2.9066E+06,
     + 2.9195E+06, 2.9326E+06, 2.9456E+06, 2.9587E+06, 2.9719E+06,
     +  2.9851E+06, 2.9983E+06, 3.0116E+06, 3.0249E+06, 3.0383E+06,
     + 3.0518E+06, 3.0652E+06, 3.0788E+06, 3.0923E+06, 3.1059E+06,
     +  3.1196E+06, 3.1333E+06, 3.1471E+06, 3.1609E+06, 3.1747E+06,
     + 3.1886E+06, 3.2026E+06, 3.2166E+06, 3.2306E+06, 3.2447E+06,
     +  3.2589E+06, 3.2731E+06, 3.2873E+06, 3.3016E+06, 3.3159E+06,
     + 3.3303E+06, 3.3448E+06, 3.3593E+06, 3.3738E+06, 3.3884E+06,
     +  3.4030E+06, 3.4177E+06, 3.4324E+06, 3.4472E+06, 3.4621E+06,
     + 3.4769E+06, 3.4919E+06, 3.5069E+06, 3.5219E+06, 3.5370E+06,
     +  3.5521E+06, 3.5673E+06, 3.5826E+06, 3.5979E+06, 3.6132E+06,
     + 3.6286E+06, 3.6440E+06, 3.6595E+06, 3.6751E+06, 3.6907E+06,
     +  3.7063E+06, 3.7221E+06, 3.7378E+06, 3.7536E+06, 3.7695E+06,
     + 3.7854E+06, 3.8014E+06, 3.8174E+06, 3.8335E+06, 3.8496E+06,
     +  3.8658E+06, 3.8820E+06, 3.8983E+06, 3.9147E+06, 3.9311E+06,
     + 3.9475E+06, 3.9640E+06, 3.9806E+06, 3.9972E+06, 4.0139E+06,
     +  4.0306E+06, 4.0474E+06, 4.0642E+06, 4.0811E+06, 4.0981E+06,
     + 4.1151E+06, 4.1321E+06, 4.1492E+06, 4.1664E+06, 4.1836E+06,
     +  4.2009E+06, 4.2182E+06, 4.2356E+06, 4.2531E+06, 4.2706E+06,
     + 4.2882E+06, 4.3058E+06, 4.3234E+06, 4.3412E+06, 4.3590E+06,
     +  4.3768E+06, 4.3947E+06, 4.4127E+06, 4.4307E+06, 4.4488E+06,
     + 4.4669E+06, 4.4851E+06, 4.5034E+06, 4.5217E+06, 4.5401E+06,
     +  4.5585E+06, 4.5770E+06, 4.5955E+06, 4.6141E+06, 4.6328E+06,
     + 4.6515E+06, 4.6703E+06, 4.6892E+06, 4.7081E+06, 4.7270E+06,
     +  4.7460E+06, 4.7651E+06, 4.7843E+06, 4.8035E+06, 4.8228E+06,
     + 4.8421E+06, 4.8615E+06, 4.8809E+06, 4.9004E+06, 4.9200E+06,
     +  4.9396E+06, 4.9593E+06, 4.9791E+06, 4.9989E+06, 5.0188E+06,
     + 5.0387E+06, 5.0587E+06, 5.0788E+06, 5.0989E+06, 5.1191E+06,
     +  5.1394E+06, 5.1597E+06, 5.1801E+06, 5.2005E+06, 5.2210E+06,
     + 5.2416E+06, 5.2623E+06, 5.2830E+06, 5.3037E+06, 5.3246E+06,
     +  5.3454E+06, 5.3664E+06, 5.3874E+06, 5.4085E+06, 5.4297E+06,
     + 5.4509E+06, 5.4722E+06, 5.4935E+06, 5.5149E+06, 5.5364E+06,
     +  5.5580E+06, 5.5796E+06, 5.6012E+06, 5.6230E+06, 5.6448E+06,
     + 5.6667E+06, 5.6886E+06, 5.7106E+06, 5.7327E+06, 5.7549E+06,
     +  5.7771E+06, 5.7993E+06, 5.8217E+06, 5.8441E+06, 5.8666E+06,
     + 5.8891E+06, 5.9118E+06, 5.9344E+06, 5.9572E+06, 5.9800E+06,
     +  6.0029E+06, 6.0259E+06, 6.0489E+06, 6.0720E+06, 6.0952E+06,
     + 6.1184E+06, 6.1417E+06, 6.1651E+06, 6.1886E+06, 6.2121E+06,
     +  6.2357E+06, 6.2593E+06, 6.2831E+06, 6.3069E+06, 6.3307E+06,
     + 6.3547E+06, 6.3787E+06, 6.4028E+06, 6.4269E+06, 6.4512E+06,
     +  6.4755E+06, 6.4998E+06, 6.5243E+06, 6.5488E+06, 6.5734E+06,
     + 6.5981E+06, 6.6228E+06, 6.6476E+06, 6.6725E+06, 6.6974E+06,
     +  6.7225E+06, 6.7476E+06, 6.7727E+06, 6.7980E+06, 6.8233E+06,
     + 6.8487E+06, 6.8742E+06, 6.8997E+06, 6.9254E+06, 6.9511E+06,
     +  6.9768E+06, 7.0027E+06, 7.0286E+06, 7.0546E+06, 7.0807E+06,
     + 7.1068E+06, 7.1330E+06, 7.1593E+06, 7.1857E+06, 7.2122E+06,
     +  7.2387E+06, 7.2653E+06, 7.2920E+06, 7.3187E+06, 7.3456E+06,
     + 7.3725E+06, 7.3995E+06, 7.4266E+06, 7.4537E+06, 7.4809E+06,
     +  7.5082E+06, 7.5356E+06, 7.5631E+06, 7.5906E+06, 7.6182E+06,
     + 7.6459E+06, 7.6737E+06, 7.7016E+06, 7.7295E+06, 7.7575E+06,
     +  7.7856E+06, 7.8138E+06, 7.8420E+06, 7.8704E+06, 7.8988E+06,
     + 7.9273E+06, 7.9559E+06, 7.9845E+06, 8.0133E+06, 8.0421E+06,
     +  8.0710E+06, 8.1000E+06, 8.1291E+06, 8.1582E+06, 8.1874E+06,
     + 8.2167E+06, 8.2461E+06, 8.2756E+06, 8.3052E+06, 8.3348E+06,
     +  8.3645E+06, 8.3944E+06, 8.4243E+06, 8.4542E+06, 8.4843E+06,
     + 8.5144E+06, 8.5447E+06, 8.5750E+06, 8.6054E+06, 8.6359E+06,
     +  8.6664E+06, 8.6971E+06, 8.7278E+06, 8.7587E+06, 8.7896E+06,
     + 8.8206E+06, 8.8516E+06, 8.8828E+06, 8.9141E+06, 8.9454E+06,
     +  8.9768E+06, 9.0083E+06, 9.0399E+06, 9.0716E+06, 9.1034E+06,
     + 9.1353E+06, 9.1672E+06, 9.1993E+06, 9.2314E+06, 9.2636E+06,
     +  9.2959E+06, 9.3283E+06, 9.3608E+06, 9.3933E+06, 9.4260E+06,
     + 9.4587E+06, 9.4916E+06, 9.5245E+06, 9.5575E+06, 9.5906E+06,
     +  9.6238E+06, 9.6571E+06, 9.6905E+06, 9.7240E+06, 9.7575E+06,
     + 9.7912E+06, 9.8249E+06, 9.8588E+06, 9.8927E+06, 9.9267E+06,
     +  9.9608E+06, 9.9950E+06, 1.0029E+07, 1.0064E+07, 1.0098E+07,
     + 1.0133E+07, 1.0167E+07, 1.0202E+07, 1.0237E+07, 1.0272E+07,
     +  1.0307E+07, 1.0342E+07, 1.0377E+07, 1.0413E+07, 1.0448E+07,
     + 1.0484E+07, 1.0519E+07, 1.0555E+07, 1.0591E+07, 1.0627E+07,
     +  1.0663E+07, 1.0699E+07, 1.0735E+07, 1.0771E+07, 1.0807E+07,
     + 1.0844E+07, 1.0880E+07, 1.0917E+07, 1.0954E+07, 1.0991E+07,
     +  1.1028E+07, 1.1065E+07, 1.1102E+07, 1.1139E+07, 1.1176E+07,
     + 1.1214E+07, 1.1251E+07, 1.1289E+07, 1.1327E+07, 1.1365E+07,
     +  1.1403E+07, 1.1441E+07, 1.1479E+07, 1.1517E+07, 1.1555E+07,
     + 1.1594E+07, 1.1632E+07, 1.1671E+07, 1.1710E+07, 1.1749E+07,
     +  1.1787E+07, 1.1827E+07, 1.1866E+07, 1.1905E+07, 1.1944E+07,
     + 1.1984E+07, 1.2023E+07, 1.2063E+07, 1.2103E+07, 1.2143E+07,
     +  1.2182E+07, 1.2223E+07, 1.2263E+07, 1.2303E+07, 1.2343E+07,
     + 1.2384E+07, 1.2424E+07, 1.2465E+07, 1.2506E+07, 1.2547E+07,
     +  1.2588E+07, 1.2629E+07, 1.2670E+07, 1.2711E+07, 1.2753E+07,
     + 1.2794E+07, 1.2836E+07, 1.2878E+07, 1.2920E+07, 1.2962E+07,
     +  1.3004E+07, 1.3046E+07, 1.3088E+07, 1.3131E+07, 1.3173E+07,
     + 1.3216E+07, 1.3258E+07, 1.3301E+07, 1.3344E+07, 1.3387E+07,
     +  1.3430E+07, 1.3474E+07, 1.3517E+07, 1.3560E+07, 1.3604E+07,
     + 1.3648E+07, 1.3691E+07, 1.3735E+07, 1.3779E+07, 1.3824E+07,
     +  1.3868E+07, 1.3912E+07, 1.3957E+07, 1.4001E+07, 1.4046E+07,
     + 1.4091E+07, 1.4136E+07, 1.4181E+07, 1.4226E+07, 1.4271E+07,
     +  1.4316E+07, 1.4362E+07, 1.4408E+07, 1.4453E+07, 1.4499E+07,
     + 1.4545E+07, 1.4591E+07, 1.4637E+07, 1.4684E+07, 1.4730E+07,
     +  1.4777E+07, 1.4823E+07, 1.4870E+07, 1.4917E+07, 1.4964E+07,
     + 1.5011E+07, 1.5058E+07, 1.5105E+07, 1.5153E+07, 1.5200E+07,
     +  1.5248E+07, 1.5296E+07, 1.5344E+07, 1.5392E+07, 1.5440E+07,
     + 1.5488E+07, 1.5537E+07, 1.5585E+07, 1.5634E+07, 1.5682E+07,
     +  1.5731E+07, 1.5780E+07, 1.5829E+07, 1.5879E+07, 1.5928E+07,
     + 1.5977E+07, 1.6027E+07, 1.6077E+07, 1.6127E+07, 1.6177E+07,
     +  1.6227E+07, 1.6277E+07, 1.6327E+07, 1.6378E+07, 1.6428E+07,
     + 1.6479E+07, 1.6530E+07, 1.6581E+07, 1.6632E+07, 1.6683E+07,
     +  1.6734E+07, 1.6785E+07, 1.6837E+07, 1.6889E+07, 1.6940E+07,
     + 1.6992E+07, 1.7044E+07, 1.7097E+07, 1.7149E+07, 1.7201E+07,
     +  1.7254E+07, 1.7307E+07, 1.7359E+07, 1.7412E+07, 1.7465E+07,
     + 1.7519E+07, 1.7572E+07, 1.7625E+07, 1.7679E+07, 1.7733E+07,
     +  1.7786E+07, 1.7840E+07, 1.7894E+07, 1.7949E+07, 1.8003E+07,
     + 1.8057E+07, 1.8112E+07, 1.8167E+07, 1.8222E+07, 1.8277E+07,
     +  1.8332E+07, 1.8387E+07, 1.8442E+07, 1.8498E+07, 1.8553E+07,
     + 1.8609E+07, 1.8665E+07, 1.8721E+07, 1.8777E+07, 1.8834E+07,
     +  1.8890E+07, 1.8947E+07, 1.9003E+07, 1.9060E+07, 1.9117E+07,
     + 1.9174E+07, 1.9231E+07, 1.9289E+07, 1.9346E+07, 1.9404E+07,
     +  1.9462E+07, 1.9519E+07, 1.9578E+07, 1.9636E+07, 1.9694E+07,
     + 1.9752E+07, 1.9811E+07, 1.9870E+07, 1.9929E+07, 1.9988E+07,
     +  2.0047E+07, 2.0106E+07, 2.0165E+07, 2.0225E+07, 2.0285E+07,
     + 2.0344E+07, 2.0404E+07, 2.0464E+07, 2.0525E+07, 2.0585E+07,
     +  2.0646E+07, 2.0706E+07, 2.0767E+07, 2.0828E+07, 2.0889E+07,
     + 2.0950E+07, 2.1012E+07, 2.1073E+07, 2.1135E+07, 2.1197E+07,
     +  2.1258E+07, 2.1321E+07, 2.1383E+07, 2.1445E+07, 2.1508E+07,
     + 2.1570E+07, 2.1633E+07, 2.1696E+07, 2.1759E+07, 2.1822E+07,
     +  2.1886E+07, 2.1949E+07, 2.2013E+07, 2.2076E+07, 2.2140E+07,
     + 2.2204E+07, 2.2269E+07, 2.2333E+07, 2.2398E+07, 2.2462E+07,
     +  2.2527E+07, 2.2592E+07, 2.2657E+07, 2.2722E+07, 2.2788E+07,
     + 2.2853E+07, 2.2919E+07, 2.2985E+07, 2.3051E+07, 2.3117E+07,
     +  2.3183E+07, 2.3250E+07, 2.3316E+07, 2.3383E+07, 2.3450E+07,
     + 2.3517E+07, 2.3584E+07, 2.3651E+07, 2.3719E+07, 2.3787E+07,
     +  2.3854E+07, 2.3922E+07, 2.3990E+07, 2.4059E+07, 2.4127E+07,
     + 2.4196E+07, 2.4264E+07, 2.4333E+07, 2.4402E+07, 2.4471E+07,
     +  2.4541E+07, 2.4610E+07, 2.4680E+07, 2.4750E+07, 2.4820E+07,
     + 2.4890E+07, 2.4960E+07, 2.5030E+07, 2.5101E+07, 2.5172E+07,
     +  2.5242E+07, 2.5314E+07, 2.5385E+07, 2.5456E+07, 2.5528E+07,
     + 2.5599E+07, 2.5671E+07, 2.5743E+07, 2.5815E+07, 2.5887E+07,
     +  2.5960E+07, 2.6033E+07, 2.6105E+07, 2.6178E+07, 2.6251E+07,
     + 2.6325E+07, 2.6398E+07, 2.6472E+07, 2.6545E+07, 2.6619E+07,
     +  2.6693E+07, 2.6767E+07, 2.6842E+07, 2.6916E+07, 2.6991E+07,
     + 2.7066E+07, 2.7141E+07, 2.7216E+07, 2.7292E+07, 2.7367E+07,
     +  2.7443E+07, 2.7519E+07, 2.7595E+07, 2.7671E+07, 2.7747E+07,
     + 2.7824E+07, 2.7900E+07, 2.7977E+07, 2.8054E+07, 2.8131E+07,
     +  2.8209E+07, 2.8286E+07, 2.8364E+07, 2.8442E+07, 2.8520E+07,
     + 2.8598E+07, 2.8676E+07, 2.8755E+07, 2.8834E+07, 2.8913E+07,
     +  2.8992E+07, 2.9071E+07, 2.9150E+07, 2.9230E+07, 2.9309E+07,
     + 2.9389E+07, 2.9469E+07, 2.9550E+07, 2.9630E+07, 2.9710E+07,
     +  2.9791E+07, 2.9872E+07, 2.9953E+07, 3.0034E+07, 3.0116E+07,
     + 3.0198E+07, 3.0279E+07, 3.0361E+07, 3.0443E+07, 3.0526E+07,
     +  3.0608E+07, 3.0691E+07, 3.0774E+07, 3.0857E+07, 3.0940E+07,
     + 3.1023E+07, 3.1107E+07, 3.1190E+07, 3.1274E+07, 3.1358E+07,
     +  3.1443E+07, 3.1527E+07, 3.1612E+07, 3.1696E+07, 3.1781E+07,
     + 3.1866E+07, 3.1952E+07, 3.2037E+07, 3.2123E+07, 3.2209E+07,
     +  3.2295E+07, 3.2381E+07, 3.2467E+07, 3.2554E+07, 3.2641E+07,
     + 3.2728E+07, 3.2815E+07, 3.2902E+07, 3.2990E+07, 3.3077E+07,
     +  3.3165E+07, 3.3253E+07, 3.3341E+07, 3.3430E+07, 3.3518E+07,
     + 3.3607E+07, 3.3696E+07, 3.3785E+07, 3.3874E+07, 3.3964E+07,
     +  3.4054E+07, 3.4144E+07, 3.4234E+07, 3.4324E+07, 3.4414E+07,
     + 3.4505E+07, 3.4596E+07, 3.4687E+07, 3.4778E+07, 3.4869E+07,
     +  3.4961E+07, 3.5053E+07, 3.5145E+07, 3.5237E+07, 3.5329E+07,
     + 3.5422E+07, 3.5514E+07, 3.5607E+07, 3.5700E+07, 3.5793E+07,
     +  3.5887E+07, 3.5981E+07, 3.6074E+07, 3.6168E+07, 3.6263E+07,
     + 3.6357E+07, 3.6452E+07, 3.6547E+07, 3.6642E+07, 3.6737E+07,
     +  3.6832E+07, 3.6928E+07, 3.7024E+07, 3.7120E+07, 3.7216E+07,
     + 3.7312E+07, 3.7409E+07, 3.7505E+07, 3.7602E+07, 3.7700E+07,
     +  3.7797E+07, 3.7895E+07, 3.7992E+07, 3.8090E+07, 3.8188E+07,
     + 3.8287E+07, 3.8385E+07, 3.8484E+07, 3.8583E+07, 3.8682E+07,
     +  3.8781E+07, 3.8881E+07, 3.8981E+07, 3.9081E+07, 3.9181E+07,
     + 3.9281E+07, 3.9382E+07, 3.9482E+07, 3.9583E+07, 3.9685E+07,
     +  3.9786E+07, 3.9888E+07, 3.9989E+07, 4.0091E+07, 4.0193E+07,
     + 4.0296E+07, 4.0398E+07, 4.0501E+07, 4.0604E+07, 4.0707E+07,
     +  4.0811E+07, 4.0914E+07, 4.1018E+07, 4.1122E+07, 4.1227E+07,
     + 4.1331E+07, 4.1436E+07, 4.1541E+07, 4.1646E+07, 4.1751E+07,
     +  4.1856E+07, 4.1962E+07, 4.2068E+07, 4.2174E+07, 4.2280E+07,
     + 4.2387E+07, 4.2494E+07, 4.2601E+07, 4.2708E+07, 4.2815E+07,
     +  4.2923E+07, 4.3031E+07, 4.3139E+07, 4.3247E+07, 4.3355E+07,
     + 4.3464E+07, 4.3573E+07, 4.3682E+07, 4.3791E+07, 4.3901E+07,
     +  4.4011E+07, 4.4121E+07, 4.4231E+07, 4.4341E+07, 4.4452E+07,
     + 4.4562E+07, 4.4674E+07, 4.4785E+07, 4.4896E+07, 4.5008E+07,
     +  4.5120E+07, 4.5232E+07, 4.5344E+07, 4.5457E+07, 4.5570E+07,
     + 4.5683E+07, 4.5796E+07, 4.5909E+07, 4.6023E+07, 4.6137E+07,
     +  4.6251E+07, 4.6365E+07, 4.6480E+07, 4.6595E+07, 4.6710E+07,
     + 4.6825E+07, 4.6940E+07, 4.7056E+07, 4.7172E+07, 4.7288E+07,
     +  4.7404E+07, 4.7521E+07, 4.7638E+07, 4.7755E+07, 4.7872E+07,
     + 4.7989E+07, 4.8107E+07, 4.8225E+07, 4.8343E+07, 4.8461E+07,
     +  4.8580E+07, 4.8699E+07, 4.8818E+07, 4.8937E+07, 4.9057E+07,
     + 4.9176E+07, 4.9296E+07, 4.9417E+07, 4.9537E+07, 4.9658E+07,
     +  4.9779E+07, 4.9900E+07, 5.0021E+07, 5.0143E+07, 5.0265E+07,
     + 5.0387E+07, 5.0509E+07, 5.0631E+07, 5.0754E+07, 5.0877E+07,
     +  5.1000E+07, 5.1124E+07, 5.1248E+07, 5.1371E+07, 5.1496E+07,
     + 5.1620E+07, 5.1745E+07, 5.1870E+07, 5.1995E+07, 5.2120E+07,
     +  5.2246E+07, 5.2371E+07, 5.2497E+07, 5.2624E+07, 5.2750E+07,
     + 5.2877E+07, 5.3004E+07, 5.3131E+07, 5.3259E+07, 5.3387E+07,
     +  5.3515E+07, 5.3643E+07, 5.3771E+07, 5.3900E+07, 5.4029E+07,
     + 5.4158E+07, 5.4287E+07, 5.4417E+07, 5.4547E+07, 5.4677E+07,
     +  5.4808E+07, 5.4938E+07, 5.5069E+07, 5.5200E+07, 5.5332E+07,
     + 5.5463E+07, 5.5595E+07, 5.5727E+07, 5.5860E+07, 5.5992E+07,
     +  5.6125E+07, 5.6258E+07, 5.6392E+07, 5.6525E+07, 5.6659E+07,
     + 5.6793E+07, 5.6927E+07, 5.7062E+07, 5.7197E+07, 5.7332E+07,
     +  5.7467E+07, 5.7603E+07, 5.7739E+07, 5.7875E+07, 5.8011E+07,
     + 5.8148E+07, 5.8285E+07, 5.8422E+07, 5.8559E+07, 5.8697E+07,
     +  5.8835E+07, 5.8973E+07, 5.9111E+07, 5.9250E+07, 5.9389E+07,
     + 5.9528E+07, 5.9667E+07, 5.9807E+07, 5.9947E+07, 6.0087E+07,
     +  6.0228E+07, 6.0368E+07, 6.0509E+07, 6.0651E+07, 6.0792E+07,
     + 6.0934E+07, 6.1076E+07, 6.1218E+07, 6.1360E+07, 6.1503E+07,
     +  6.1646E+07, 6.1790E+07, 6.1933E+07, 6.2077E+07, 6.2221E+07,
     + 6.2365E+07, 6.2510E+07, 6.2655E+07, 6.2800E+07, 6.2945E+07,
     +  6.3091E+07, 6.3237E+07, 6.3383E+07, 6.3530E+07, 6.3676E+07,
     + 6.3823E+07, 6.3971E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 9,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       786
      data (QofT(10,J),J=1,2408)/ 1.0116E+01, 2.6149E+01, 4.6565E+01,
     +  7.0584E+01, 9.7730E+01, 1.2768E+02, 1.6018E+02, 1.9506E+02,
     + 2.3217E+02, 2.7136E+02, 3.1255E+02, 3.5563E+02, 4.0054E+02,
     +  4.4719E+02, 4.9552E+02, 5.4548E+02, 5.9702E+02, 6.5009E+02,
     + 7.0464E+02, 7.6064E+02, 8.7685E+02, 9.9844E+02, 1.1252E+03,
     +  1.2569E+03, 1.3934E+03, 1.5345E+03, 1.6801E+03, 1.8300E+03,
     + 1.9842E+03, 2.1424E+03, 2.3047E+03, 2.4709E+03, 2.6409E+03,
     +  2.8146E+03, 2.9920E+03, 3.1730E+03, 3.3575E+03, 3.5455E+03,
     + 3.7368E+03, 3.9315E+03, 4.1295E+03, 4.3307E+03, 4.5350E+03,
     +  4.7425E+03, 4.9531E+03, 5.1667E+03, 5.3833E+03, 5.6029E+03,
     + 5.8254E+03, 6.0508E+03, 6.2791E+03, 6.5101E+03, 6.7439E+03,
     +  6.9805E+03, 7.2198E+03, 7.4618E+03, 7.7065E+03, 7.9538E+03,
     + 8.2037E+03, 8.4562E+03, 8.7113E+03, 8.9690E+03, 9.2291E+03,
     +  9.4918E+03, 9.7569E+03, 1.0025E+04, 1.0295E+04, 1.0567E+04,
     + 1.0842E+04, 1.1120E+04, 1.1399E+04, 1.1682E+04, 1.1966E+04,
     +  1.2253E+04, 1.2542E+04, 1.2834E+04, 1.3128E+04, 1.3424E+04,
     + 1.3723E+04, 1.4024E+04, 1.4327E+04, 1.4633E+04, 1.4941E+04,
     +  1.5251E+04, 1.5564E+04, 1.5879E+04, 1.6196E+04, 1.6516E+04,
     + 1.6838E+04, 1.7162E+04, 1.7488E+04, 1.7817E+04, 1.8149E+04,
     +  1.8482E+04, 1.8818E+04, 1.9157E+04, 1.9497E+04, 1.9840E+04,
     + 2.0186E+04, 2.0534E+04, 2.0884E+04, 2.1237E+04, 2.1592E+04,
     +  2.1949E+04, 2.2309E+04, 2.2672E+04, 2.3037E+04, 2.3404E+04,
     + 2.3774E+04, 2.4147E+04, 2.4522E+04, 2.4900E+04, 2.5280E+04,
     +  2.5663E+04, 2.6048E+04, 2.6436E+04, 2.6827E+04, 2.7220E+04,
     + 2.7616E+04, 2.8015E+04, 2.8416E+04, 2.8821E+04, 2.9228E+04,
     +  2.9637E+04, 3.0050E+04, 3.0465E+04, 3.0884E+04, 3.1305E+04,
     + 3.1729E+04, 3.2156E+04, 3.2585E+04, 3.3018E+04, 3.3454E+04,
     +  3.3892E+04, 3.4334E+04, 3.4779E+04, 3.5227E+04, 3.5678E+04,
     + 3.6132E+04, 3.6590E+04, 3.7050E+04, 3.7514E+04, 3.7981E+04,
     +  3.8450E+04, 3.8924E+04, 3.9401E+04, 3.9881E+04, 4.0364E+04,
     + 4.0851E+04, 4.1341E+04, 4.1835E+04, 4.2331E+04, 4.2832E+04,
     +  4.3336E+04, 4.3844E+04, 4.4355E+04, 4.4870E+04, 4.5388E+04,
     + 4.5910E+04, 4.6435E+04, 4.6965E+04, 4.7498E+04, 4.8035E+04,
     +  4.8576E+04, 4.9120E+04, 4.9668E+04, 5.0221E+04, 5.0777E+04,
     + 5.1337E+04, 5.1901E+04, 5.2469E+04, 5.3041E+04, 5.3617E+04,
     +  5.4197E+04, 5.4782E+04, 5.5370E+04, 5.5963E+04, 5.6560E+04,
     + 5.7161E+04, 5.7766E+04, 5.8375E+04, 5.8989E+04, 5.9608E+04,
     +  6.0231E+04, 6.0857E+04, 6.1489E+04, 6.2125E+04, 6.2766E+04,
     + 6.3411E+04, 6.4061E+04, 6.4715E+04, 6.5374E+04, 6.6037E+04,
     +  6.6706E+04, 6.7378E+04, 6.8057E+04, 6.8739E+04, 6.9426E+04,
     + 7.0119E+04, 7.0816E+04, 7.1518E+04, 7.2225E+04, 7.2937E+04,
     +  7.3654E+04, 7.4376E+04, 7.5103E+04, 7.5836E+04, 7.6573E+04,
     + 7.7316E+04, 7.8063E+04, 7.8817E+04, 7.9575E+04, 8.0339E+04,
     +  8.1108E+04, 8.1882E+04, 8.2662E+04, 8.3447E+04, 8.4238E+04,
     + 8.5034E+04, 8.5837E+04, 8.6644E+04, 8.7457E+04, 8.8276E+04,
     +  8.9101E+04, 8.9930E+04, 9.0766E+04, 9.1608E+04, 9.2456E+04,
     + 9.3310E+04, 9.4169E+04, 9.5034E+04, 9.5905E+04, 9.6782E+04,
     +  9.7666E+04, 9.8555E+04, 9.9451E+04, 1.0035E+05, 1.0126E+05,
     + 1.0217E+05, 1.0309E+05, 1.0402E+05, 1.0495E+05, 1.0589E+05,
     +  1.0684E+05, 1.0779E+05, 1.0875E+05, 1.0971E+05, 1.1069E+05,
     + 1.1166E+05, 1.1265E+05, 1.1364E+05, 1.1464E+05, 1.1564E+05,
     +  1.1666E+05, 1.1767E+05, 1.1870E+05, 1.1973E+05, 1.2077E+05,
     + 1.2182E+05, 1.2287E+05, 1.2393E+05, 1.2500E+05, 1.2607E+05,
     +  1.2715E+05, 1.2824E+05, 1.2934E+05, 1.3044E+05, 1.3155E+05,
     + 1.3267E+05, 1.3379E+05, 1.3493E+05, 1.3607E+05, 1.3721E+05,
     +  1.3837E+05, 1.3953E+05, 1.4070E+05, 1.4188E+05, 1.4307E+05,
     + 1.4426E+05, 1.4546E+05, 1.4667E+05, 1.4789E+05, 1.4911E+05,
     +  1.5034E+05, 1.5158E+05, 1.5283E+05, 1.5409E+05, 1.5535E+05,
     + 1.5663E+05, 1.5791E+05, 1.5920E+05, 1.6049E+05, 1.6180E+05,
     +  1.6311E+05, 1.6444E+05, 1.6577E+05, 1.6711E+05, 1.6846E+05,
     + 1.6981E+05, 1.7118E+05, 1.7255E+05, 1.7393E+05, 1.7532E+05,
     +  1.7672E+05, 1.7813E+05, 1.7955E+05, 1.8098E+05, 1.8241E+05,
     + 1.8386E+05, 1.8531E+05, 1.8677E+05, 1.8825E+05, 1.8973E+05,
     +  1.9122E+05, 1.9272E+05, 1.9422E+05, 1.9574E+05, 1.9727E+05,
     + 1.9881E+05, 2.0035E+05, 2.0191E+05, 2.0347E+05, 2.0505E+05,
     +  2.0663E+05, 2.0823E+05, 2.0983E+05, 2.1144E+05, 2.1307E+05,
     + 2.1470E+05, 2.1635E+05, 2.1800E+05, 2.1966E+05, 2.2134E+05,
     +  2.2302E+05, 2.2471E+05, 2.2642E+05, 2.2813E+05, 2.2985E+05,
     + 2.3159E+05, 2.3333E+05, 2.3509E+05, 2.3686E+05, 2.3863E+05,
     +  2.4042E+05, 2.4222E+05, 2.4403E+05, 2.4584E+05, 2.4767E+05,
     + 2.4951E+05, 2.5137E+05, 2.5323E+05, 2.5510E+05, 2.5699E+05,
     +  2.5888E+05, 2.6079E+05, 2.6271E+05, 2.6464E+05, 2.6657E+05,
     + 2.6853E+05, 2.7049E+05, 2.7246E+05, 2.7445E+05, 2.7645E+05,
     +  2.7846E+05, 2.8047E+05, 2.8251E+05, 2.8455E+05, 2.8661E+05,
     + 2.8867E+05, 2.9075E+05, 2.9284E+05, 2.9495E+05, 2.9706E+05,
     +  2.9919E+05, 3.0133E+05, 3.0348E+05, 3.0564E+05, 3.0782E+05,
     + 3.1001E+05, 3.1221E+05, 3.1442E+05, 3.1665E+05, 3.1888E+05,
     +  3.2113E+05, 3.2340E+05, 3.2567E+05, 3.2796E+05, 3.3026E+05,
     + 3.3258E+05, 3.3490E+05, 3.3724E+05, 3.3960E+05, 3.4196E+05,
     +  3.4434E+05, 3.4673E+05, 3.4914E+05, 3.5156E+05, 3.5399E+05,
     + 3.5644E+05, 3.5890E+05, 3.6137E+05, 3.6385E+05, 3.6635E+05,
     +  3.6887E+05, 3.7139E+05, 3.7393E+05, 3.7649E+05, 3.7906E+05,
     + 3.8164E+05, 3.8423E+05, 3.8684E+05, 3.8947E+05, 3.9211E+05,
     +  3.9476E+05, 3.9743E+05, 4.0011E+05, 4.0281E+05, 4.0552E+05,
     + 4.0824E+05, 4.1098E+05, 4.1373E+05, 4.1650E+05, 4.1929E+05,
     +  4.2209E+05, 4.2490E+05, 4.2773E+05, 4.3057E+05, 4.3343E+05,
     + 4.3630E+05, 4.3919E+05, 4.4209E+05, 4.4501E+05, 4.4794E+05,
     +  4.5090E+05, 4.5386E+05, 4.5684E+05, 4.5984E+05, 4.6285E+05,
     + 4.6588E+05, 4.6892E+05, 4.7198E+05, 4.7505E+05, 4.7814E+05,
     +  4.8125E+05, 4.8437E+05, 4.8751E+05, 4.9066E+05, 4.9384E+05,
     + 4.9702E+05, 5.0023E+05, 5.0345E+05, 5.0668E+05, 5.0994E+05,
     +  5.1320E+05, 5.1649E+05, 5.1979E+05, 5.2311E+05, 5.2645E+05,
     + 5.2981E+05, 5.3318E+05, 5.3657E+05, 5.3997E+05, 5.4339E+05,
     +  5.4683E+05, 5.5029E+05, 5.5376E+05, 5.5725E+05, 5.6076E+05,
     + 5.6429E+05, 5.6783E+05, 5.7139E+05, 5.7497E+05, 5.7857E+05,
     +  5.8219E+05, 5.8582E+05, 5.8947E+05, 5.9314E+05, 5.9683E+05,
     + 6.0053E+05, 6.0425E+05, 6.0800E+05, 6.1176E+05, 6.1554E+05,
     +  6.1933E+05, 6.2315E+05, 6.2698E+05, 6.3084E+05, 6.3471E+05,
     + 6.3860E+05, 6.4251E+05, 6.4644E+05, 6.5039E+05, 6.5435E+05,
     +  6.5834E+05, 6.6235E+05, 6.6637E+05, 6.7042E+05, 6.7448E+05,
     + 6.7857E+05, 6.8267E+05, 6.8679E+05, 6.9094E+05, 6.9510E+05,
     +  6.9928E+05, 7.0348E+05, 7.0770E+05, 7.1195E+05, 7.1621E+05,
     + 7.2049E+05, 7.2480E+05, 7.2912E+05, 7.3346E+05, 7.3783E+05,
     +  7.4222E+05, 7.4662E+05, 7.5105E+05, 7.5549E+05, 7.5996E+05,
     + 7.6445E+05, 7.6896E+05, 7.7350E+05, 7.7805E+05, 7.8262E+05,
     +  7.8722E+05, 7.9183E+05, 7.9647E+05, 8.0113E+05, 8.0581E+05,
     + 8.1052E+05, 8.1524E+05, 8.1999E+05, 8.2476E+05, 8.2955E+05,
     +  8.3436E+05, 8.3920E+05, 8.4405E+05, 8.4893E+05, 8.5383E+05,
     + 8.5876E+05, 8.6370E+05, 8.6867E+05, 8.7366E+05, 8.7868E+05,
     +  8.8372E+05, 8.8878E+05, 8.9386E+05, 8.9897E+05, 9.0409E+05,
     + 9.0925E+05, 9.1442E+05, 9.1962E+05, 9.2484E+05, 9.3009E+05,
     +  9.3536E+05, 9.4066E+05, 9.4597E+05, 9.5131E+05, 9.5668E+05,
     + 9.6207E+05, 9.6748E+05, 9.7292E+05, 9.7838E+05, 9.8386E+05,
     +  9.8937E+05, 9.9491E+05, 1.0005E+06, 1.0060E+06, 1.0117E+06,
     + 1.0173E+06, 1.0229E+06, 1.0286E+06, 1.0343E+06, 1.0401E+06,
     +  1.0458E+06, 1.0516E+06, 1.0574E+06, 1.0633E+06, 1.0691E+06,
     + 1.0750E+06, 1.0809E+06, 1.0868E+06, 1.0928E+06, 1.0988E+06,
     +  1.1048E+06, 1.1108E+06, 1.1169E+06, 1.1230E+06, 1.1291E+06,
     + 1.1353E+06, 1.1414E+06, 1.1476E+06, 1.1539E+06, 1.1601E+06,
     +  1.1664E+06, 1.1727E+06, 1.1790E+06, 1.1854E+06, 1.1918E+06,
     + 1.1982E+06, 1.2046E+06, 1.2111E+06, 1.2176E+06, 1.2241E+06,
     +  1.2307E+06, 1.2372E+06, 1.2438E+06, 1.2505E+06, 1.2571E+06,
     + 1.2638E+06, 1.2706E+06, 1.2773E+06, 1.2841E+06, 1.2909E+06,
     +  1.2977E+06, 1.3046E+06, 1.3115E+06, 1.3184E+06, 1.3253E+06,
     + 1.3323E+06, 1.3393E+06, 1.3463E+06, 1.3534E+06, 1.3605E+06,
     +  1.3676E+06, 1.3748E+06, 1.3820E+06, 1.3892E+06, 1.3964E+06,
     + 1.4037E+06, 1.4110E+06, 1.4183E+06, 1.4257E+06, 1.4331E+06,
     +  1.4405E+06, 1.4479E+06, 1.4554E+06, 1.4629E+06, 1.4705E+06,
     + 1.4780E+06, 1.4856E+06, 1.4933E+06, 1.5009E+06, 1.5086E+06,
     +  1.5164E+06, 1.5241E+06, 1.5319E+06, 1.5397E+06, 1.5476E+06,
     + 1.5555E+06, 1.5634E+06, 1.5713E+06, 1.5793E+06, 1.5873E+06,
     +  1.5954E+06, 1.6035E+06, 1.6116E+06, 1.6197E+06, 1.6279E+06,
     + 1.6361E+06, 1.6443E+06, 1.6526E+06, 1.6609E+06, 1.6693E+06,
     +  1.6776E+06, 1.6860E+06, 1.6945E+06, 1.7029E+06, 1.7115E+06,
     + 1.7200E+06, 1.7286E+06, 1.7372E+06, 1.7458E+06, 1.7545E+06,
     +  1.7632E+06, 1.7719E+06, 1.7807E+06, 1.7895E+06, 1.7984E+06,
     + 1.8073E+06, 1.8162E+06, 1.8251E+06, 1.8341E+06, 1.8431E+06,
     +  1.8522E+06, 1.8613E+06, 1.8704E+06, 1.8795E+06, 1.8887E+06,
     + 1.8980E+06, 1.9072E+06, 1.9165E+06, 1.9259E+06, 1.9352E+06,
     +  1.9447E+06, 1.9541E+06, 1.9636E+06, 1.9731E+06, 1.9827E+06,
     + 1.9922E+06, 2.0019E+06, 2.0115E+06, 2.0212E+06, 2.0310E+06,
     +  2.0408E+06, 2.0506E+06, 2.0604E+06, 2.0703E+06, 2.0802E+06,
     + 2.0902E+06, 2.1002E+06, 2.1102E+06, 2.1203E+06, 2.1304E+06,
     +  2.1406E+06, 2.1508E+06, 2.1610E+06, 2.1712E+06, 2.1816E+06,
     + 2.1919E+06, 2.2023E+06, 2.2127E+06, 2.2232E+06, 2.2337E+06,
     +  2.2442E+06, 2.2548E+06, 2.2654E+06, 2.2760E+06, 2.2867E+06,
     + 2.2975E+06, 2.3082E+06, 2.3191E+06, 2.3299E+06, 2.3408E+06,
     +  2.3517E+06, 2.3627E+06, 2.3737E+06, 2.3848E+06, 2.3959E+06,
     + 2.4070E+06, 2.4182E+06, 2.4294E+06, 2.4407E+06, 2.4520E+06,
     +  2.4633E+06, 2.4747E+06, 2.4861E+06, 2.4976E+06, 2.5091E+06,
     + 2.5207E+06, 2.5323E+06, 2.5439E+06, 2.5556E+06, 2.5673E+06,
     +  2.5790E+06, 2.5908E+06, 2.6027E+06, 2.6146E+06, 2.6265E+06,
     + 2.6385E+06, 2.6505E+06, 2.6626E+06, 2.6747E+06, 2.6868E+06,
     +  2.6990E+06, 2.7112E+06, 2.7235E+06, 2.7358E+06, 2.7482E+06,
     + 2.7606E+06, 2.7731E+06, 2.7856E+06, 2.7981E+06, 2.8107E+06,
     +  2.8233E+06, 2.8360E+06, 2.8487E+06, 2.8615E+06, 2.8743E+06,
     + 2.8872E+06, 2.9001E+06, 2.9130E+06, 2.9260E+06, 2.9391E+06,
     +  2.9521E+06, 2.9653E+06, 2.9784E+06, 2.9917E+06, 3.0049E+06,
     + 3.0183E+06, 3.0316E+06, 3.0450E+06, 3.0585E+06, 3.0720E+06,
     +  3.0855E+06, 3.0991E+06, 3.1128E+06, 3.1265E+06, 3.1402E+06,
     + 3.1540E+06, 3.1678E+06, 3.1817E+06, 3.1956E+06, 3.2096E+06,
     +  3.2236E+06, 3.2377E+06, 3.2518E+06, 3.2660E+06, 3.2802E+06,
     + 3.2945E+06, 3.3088E+06, 3.3232E+06, 3.3376E+06, 3.3521E+06,
     +  3.3666E+06, 3.3811E+06, 3.3957E+06, 3.4104E+06, 3.4251E+06,
     + 3.4399E+06, 3.4547E+06, 3.4696E+06, 3.4845E+06, 3.4994E+06,
     +  3.5144E+06, 3.5295E+06, 3.5446E+06, 3.5598E+06, 3.5750E+06,
     + 3.5903E+06, 3.6056E+06, 3.6210E+06, 3.6364E+06, 3.6519E+06,
     +  3.6674E+06, 3.6830E+06, 3.6986E+06, 3.7143E+06, 3.7300E+06,
     + 3.7458E+06, 3.7617E+06, 3.7776E+06, 3.7935E+06, 3.8095E+06,
     +  3.8256E+06, 3.8417E+06, 3.8578E+06, 3.8740E+06, 3.8903E+06,
     + 3.9066E+06, 3.9230E+06, 3.9394E+06, 3.9559E+06, 3.9725E+06,
     +  3.9891E+06, 4.0057E+06, 4.0224E+06, 4.0392E+06, 4.0560E+06,
     + 4.0728E+06, 4.0898E+06, 4.1067E+06, 4.1238E+06, 4.1409E+06,
     +  4.1580E+06, 4.1752E+06, 4.1924E+06, 4.2098E+06, 4.2271E+06,
     + 4.2446E+06, 4.2620E+06, 4.2796E+06, 4.2972E+06, 4.3148E+06,
     +  4.3325E+06, 4.3503E+06, 4.3681E+06, 4.3860E+06, 4.4039E+06,
     + 4.4219E+06, 4.4400E+06, 4.4581E+06, 4.4763E+06, 4.4945E+06,
     +  4.5128E+06, 4.5311E+06, 4.5495E+06, 4.5680E+06, 4.5865E+06,
     + 4.6051E+06, 4.6237E+06, 4.6424E+06, 4.6612E+06, 4.6800E+06,
     +  4.6989E+06, 4.7178E+06, 4.7368E+06, 4.7559E+06, 4.7750E+06,
     + 4.7942E+06, 4.8134E+06, 4.8327E+06, 4.8521E+06, 4.8715E+06,
     +  4.8910E+06, 4.9105E+06, 4.9301E+06, 4.9498E+06, 4.9695E+06,
     + 4.9893E+06, 5.0092E+06, 5.0291E+06, 5.0491E+06, 5.0691E+06,
     +  5.0892E+06, 5.1094E+06, 5.1296E+06, 5.1499E+06, 5.1703E+06,
     + 5.1907E+06, 5.2112E+06, 5.2317E+06, 5.2523E+06, 5.2730E+06,
     +  5.2937E+06, 5.3145E+06, 5.3354E+06, 5.3563E+06, 5.3773E+06,
     + 5.3984E+06, 5.4195E+06, 5.4407E+06, 5.4619E+06, 5.4833E+06,
     +  5.5047E+06, 5.5261E+06, 5.5476E+06, 5.5692E+06, 5.5909E+06,
     + 5.6126E+06, 5.6343E+06, 5.6562E+06, 5.6781E+06, 5.7001E+06,
     +  5.7221E+06, 5.7443E+06, 5.7664E+06, 5.7887E+06, 5.8110E+06,
     + 5.8334E+06, 5.8558E+06, 5.8784E+06, 5.9010E+06, 5.9236E+06,
     +  5.9463E+06, 5.9691E+06, 5.9920E+06, 6.0149E+06, 6.0379E+06,
     + 6.0610E+06, 6.0841E+06, 6.1073E+06, 6.1306E+06, 6.1540E+06,
     +  6.1774E+06, 6.2009E+06, 6.2244E+06, 6.2481E+06, 6.2718E+06,
     + 6.2955E+06, 6.3194E+06, 6.3433E+06, 6.3673E+06, 6.3913E+06,
     +  6.4155E+06, 6.4397E+06, 6.4639E+06, 6.4883E+06, 6.5127E+06,
     + 6.5372E+06, 6.5617E+06, 6.5864E+06, 6.6111E+06, 6.6358E+06,
     +  6.6607E+06, 6.6856E+06, 6.7106E+06, 6.7357E+06, 6.7608E+06,
     + 6.7860E+06, 6.8113E+06, 6.8367E+06, 6.8621E+06, 6.8876E+06,
     +  6.9132E+06, 6.9389E+06, 6.9646E+06, 6.9905E+06, 7.0163E+06,
     + 7.0423E+06, 7.0683E+06, 7.0945E+06, 7.1207E+06, 7.1469E+06,
     +  7.1733E+06, 7.1997E+06, 7.2262E+06, 7.2528E+06, 7.2794E+06,
     + 7.3061E+06, 7.3329E+06, 7.3598E+06, 7.3868E+06, 7.4138E+06,
     +  7.4409E+06, 7.4681E+06, 7.4954E+06, 7.5227E+06, 7.5502E+06,
     + 7.5777E+06, 7.6053E+06, 7.6329E+06, 7.6607E+06, 7.6885E+06,
     +  7.7164E+06, 7.7444E+06, 7.7724E+06, 7.8006E+06, 7.8288E+06,
     + 7.8571E+06, 7.8855E+06, 7.9139E+06, 7.9425E+06, 7.9711E+06,
     +  7.9998E+06, 8.0286E+06, 8.0575E+06, 8.0864E+06, 8.1154E+06,
     + 8.1445E+06, 8.1737E+06, 8.2030E+06, 8.2324E+06, 8.2618E+06,
     +  8.2913E+06, 8.3210E+06, 8.3507E+06, 8.3804E+06, 8.4103E+06,
     + 8.4402E+06, 8.4702E+06, 8.5004E+06, 8.5306E+06, 8.5608E+06,
     +  8.5912E+06, 8.6216E+06, 8.6522E+06, 8.6828E+06, 8.7135E+06,
     + 8.7443E+06, 8.7752E+06, 8.8061E+06, 8.8372E+06, 8.8683E+06,
     +  8.8995E+06, 8.9308E+06, 8.9622E+06, 8.9937E+06, 9.0252E+06,
     + 9.0569E+06, 9.0886E+06, 9.1205E+06, 9.1524E+06, 9.1844E+06,
     +  9.2164E+06, 9.2486E+06, 9.2809E+06, 9.3132E+06, 9.3457E+06,
     + 9.3782E+06, 9.4108E+06, 9.4435E+06, 9.4763E+06, 9.5092E+06,
     +  9.5422E+06, 9.5753E+06, 9.6084E+06, 9.6417E+06, 9.6750E+06,
     + 9.7085E+06, 9.7420E+06, 9.7756E+06, 9.8093E+06, 9.8431E+06,
     +  9.8770E+06, 9.9109E+06, 9.9450E+06, 9.9792E+06, 1.0013E+07,
     + 1.0048E+07, 1.0082E+07, 1.0117E+07, 1.0151E+07, 1.0186E+07,
     +  1.0221E+07, 1.0256E+07, 1.0291E+07, 1.0326E+07, 1.0361E+07,
     + 1.0396E+07, 1.0432E+07, 1.0467E+07, 1.0503E+07, 1.0538E+07,
     +  1.0574E+07, 1.0610E+07, 1.0646E+07, 1.0682E+07, 1.0718E+07,
     + 1.0754E+07, 1.0791E+07, 1.0827E+07, 1.0864E+07, 1.0900E+07,
     +  1.0937E+07, 1.0974E+07, 1.1011E+07, 1.1048E+07, 1.1085E+07,
     + 1.1122E+07, 1.1159E+07, 1.1197E+07, 1.1234E+07, 1.1272E+07,
     +  1.1309E+07, 1.1347E+07, 1.1385E+07, 1.1423E+07, 1.1461E+07,
     + 1.1499E+07, 1.1538E+07, 1.1576E+07, 1.1615E+07, 1.1653E+07,
     +  1.1692E+07, 1.1731E+07, 1.1770E+07, 1.1809E+07, 1.1848E+07,
     + 1.1887E+07, 1.1926E+07, 1.1966E+07, 1.2005E+07, 1.2045E+07,
     +  1.2084E+07, 1.2124E+07, 1.2164E+07, 1.2204E+07, 1.2244E+07,
     + 1.2285E+07, 1.2325E+07, 1.2365E+07, 1.2406E+07, 1.2447E+07,
     +  1.2487E+07, 1.2528E+07, 1.2569E+07, 1.2610E+07, 1.2651E+07,
     + 1.2693E+07, 1.2734E+07, 1.2775E+07, 1.2817E+07, 1.2859E+07,
     +  1.2901E+07, 1.2942E+07, 1.2985E+07, 1.3027E+07, 1.3069E+07,
     + 1.3111E+07, 1.3154E+07, 1.3196E+07, 1.3239E+07, 1.3282E+07,
     +  1.3325E+07, 1.3368E+07, 1.3411E+07, 1.3454E+07, 1.3497E+07,
     + 1.3541E+07, 1.3584E+07, 1.3628E+07, 1.3672E+07, 1.3715E+07,
     +  1.3759E+07, 1.3804E+07, 1.3848E+07, 1.3892E+07, 1.3936E+07,
     + 1.3981E+07, 1.4026E+07, 1.4070E+07, 1.4115E+07, 1.4160E+07,
     +  1.4205E+07, 1.4251E+07, 1.4296E+07, 1.4341E+07, 1.4387E+07,
     + 1.4433E+07, 1.4478E+07, 1.4524E+07, 1.4570E+07, 1.4616E+07,
     +  1.4663E+07, 1.4709E+07, 1.4755E+07, 1.4802E+07, 1.4849E+07,
     + 1.4896E+07, 1.4942E+07, 1.4989E+07, 1.5037E+07, 1.5084E+07,
     +  1.5131E+07, 1.5179E+07, 1.5226E+07, 1.5274E+07, 1.5322E+07,
     + 1.5370E+07, 1.5418E+07, 1.5466E+07, 1.5515E+07, 1.5563E+07,
     +  1.5612E+07, 1.5660E+07, 1.5709E+07, 1.5758E+07, 1.5807E+07,
     + 1.5856E+07, 1.5906E+07, 1.5955E+07, 1.6005E+07, 1.6054E+07,
     +  1.6104E+07, 1.6154E+07, 1.6204E+07, 1.6254E+07, 1.6304E+07,
     + 1.6355E+07, 1.6405E+07, 1.6456E+07, 1.6507E+07, 1.6558E+07,
     +  1.6609E+07, 1.6660E+07, 1.6711E+07, 1.6762E+07, 1.6814E+07,
     + 1.6865E+07, 1.6917E+07, 1.6969E+07, 1.7021E+07, 1.7073E+07,
     +  1.7125E+07, 1.7178E+07, 1.7230E+07, 1.7283E+07, 1.7336E+07,
     + 1.7388E+07, 1.7441E+07, 1.7495E+07, 1.7548E+07, 1.7601E+07,
     +  1.7655E+07, 1.7708E+07, 1.7762E+07, 1.7816E+07, 1.7870E+07,
     + 1.7924E+07, 1.7978E+07, 1.8033E+07, 1.8087E+07, 1.8142E+07,
     +  1.8197E+07, 1.8252E+07, 1.8307E+07, 1.8362E+07, 1.8417E+07,
     + 1.8473E+07, 1.8528E+07, 1.8584E+07, 1.8640E+07, 1.8696E+07,
     +  1.8752E+07, 1.8808E+07, 1.8865E+07, 1.8921E+07, 1.8978E+07,
     + 1.9035E+07, 1.9091E+07, 1.9149E+07, 1.9206E+07, 1.9263E+07,
     +  1.9320E+07, 1.9378E+07, 1.9436E+07, 1.9494E+07, 1.9552E+07,
     + 1.9610E+07, 1.9668E+07, 1.9726E+07, 1.9785E+07, 1.9843E+07,
     +  1.9902E+07, 1.9961E+07, 2.0020E+07, 2.0079E+07, 2.0139E+07,
     + 2.0198E+07, 2.0258E+07, 2.0318E+07, 2.0378E+07, 2.0438E+07,
     +  2.0498E+07, 2.0558E+07, 2.0619E+07, 2.0679E+07, 2.0740E+07,
     + 2.0801E+07, 2.0862E+07, 2.0923E+07, 2.0984E+07, 2.1046E+07,
     +  2.1107E+07, 2.1169E+07, 2.1231E+07, 2.1293E+07, 2.1355E+07,
     + 2.1417E+07, 2.1480E+07, 2.1542E+07, 2.1605E+07, 2.1668E+07,
     +  2.1731E+07, 2.1794E+07, 2.1857E+07, 2.1921E+07, 2.1984E+07,
     + 2.2048E+07, 2.2112E+07, 2.2176E+07, 2.2240E+07, 2.2304E+07,
     +  2.2369E+07, 2.2433E+07, 2.2498E+07, 2.2563E+07, 2.2628E+07,
     + 2.2693E+07, 2.2759E+07, 2.2824E+07, 2.2890E+07, 2.2956E+07,
     +  2.3021E+07, 2.3088E+07, 2.3154E+07, 2.3220E+07, 2.3287E+07,
     + 2.3353E+07, 2.3420E+07, 2.3487E+07, 2.3554E+07, 2.3622E+07,
     +  2.3689E+07, 2.3757E+07, 2.3824E+07, 2.3892E+07, 2.3960E+07,
     + 2.4028E+07, 2.4097E+07, 2.4165E+07, 2.4234E+07, 2.4303E+07,
     +  2.4372E+07, 2.4441E+07, 2.4510E+07, 2.4579E+07, 2.4649E+07,
     + 2.4719E+07, 2.4789E+07, 2.4859E+07, 2.4929E+07, 2.4999E+07,
     +  2.5070E+07, 2.5140E+07, 2.5211E+07, 2.5282E+07, 2.5353E+07,
     + 2.5424E+07, 2.5496E+07, 2.5567E+07, 2.5639E+07, 2.5711E+07,
     +  2.5783E+07, 2.5855E+07, 2.5928E+07, 2.6000E+07, 2.6073E+07,
     + 2.6146E+07, 2.6219E+07, 2.6292E+07, 2.6366E+07, 2.6439E+07,
     +  2.6513E+07, 2.6587E+07, 2.6661E+07, 2.6735E+07, 2.6809E+07,
     + 2.6883E+07, 2.6958E+07, 2.7033E+07, 2.7108E+07, 2.7183E+07,
     +  2.7258E+07, 2.7334E+07, 2.7409E+07, 2.7485E+07, 2.7561E+07,
     + 2.7637E+07, 2.7714E+07, 2.7790E+07, 2.7867E+07, 2.7943E+07,
     +  2.8020E+07, 2.8097E+07, 2.8175E+07, 2.8252E+07, 2.8330E+07,
     + 2.8408E+07, 2.8485E+07, 2.8564E+07, 2.8642E+07, 2.8720E+07,
     +  2.8799E+07, 2.8878E+07, 2.8957E+07, 2.9036E+07, 2.9115E+07,
     + 2.9195E+07, 2.9274E+07, 2.9354E+07, 2.9434E+07, 2.9514E+07,
     +  2.9595E+07, 2.9675E+07, 2.9756E+07, 2.9837E+07, 2.9918E+07,
     + 2.9999E+07, 3.0080E+07, 3.0162E+07, 3.0243E+07, 3.0325E+07,
     +  3.0407E+07, 3.0489E+07, 3.0572E+07, 3.0654E+07, 3.0737E+07,
     + 3.0820E+07, 3.0903E+07, 3.0986E+07, 3.1070E+07, 3.1154E+07,
     +  3.1237E+07, 3.1321E+07, 3.1406E+07, 3.1490E+07, 3.1574E+07,
     + 3.1659E+07, 3.1744E+07, 3.1829E+07, 3.1914E+07, 3.2000E+07,
     +  3.2085E+07, 3.2171E+07, 3.2257E+07, 3.2343E+07, 3.2429E+07,
     + 3.2516E+07, 3.2603E+07, 3.2689E+07, 3.2777E+07, 3.2864E+07,
     +  3.2951E+07, 3.3039E+07, 3.3127E+07, 3.3214E+07, 3.3303E+07,
     + 3.3391E+07, 3.3479E+07, 3.3568E+07, 3.3657E+07, 3.3746E+07,
     +  3.3835E+07, 3.3925E+07, 3.4014E+07, 3.4104E+07, 3.4194E+07,
     + 3.4284E+07, 3.4375E+07, 3.4465E+07, 3.4556E+07, 3.4647E+07,
     +  3.4738E+07, 3.4829E+07, 3.4921E+07, 3.5012E+07, 3.5104E+07,
     + 3.5196E+07, 3.5289E+07, 3.5381E+07, 3.5474E+07, 3.5566E+07,
     +  3.5659E+07, 3.5753E+07, 3.5846E+07, 3.5940E+07, 3.6033E+07,
     + 3.6127E+07, 3.6221E+07, 3.6316E+07, 3.6410E+07, 3.6505E+07,
     +  3.6600E+07, 3.6695E+07, 3.6790E+07, 3.6886E+07, 3.6982E+07,
     + 3.7078E+07, 3.7174E+07, 3.7270E+07, 3.7366E+07, 3.7463E+07,
     +  3.7560E+07, 3.7657E+07, 3.7754E+07, 3.7852E+07, 3.7950E+07,
     + 3.8047E+07, 3.8145E+07, 3.8244E+07, 3.8342E+07, 3.8441E+07,
     +  3.8540E+07, 3.8639E+07, 3.8738E+07, 3.8838E+07, 3.8937E+07,
     + 3.9037E+07, 3.9137E+07, 3.9237E+07, 3.9338E+07, 3.9439E+07,
     +  3.9539E+07, 3.9640E+07, 3.9742E+07, 3.9843E+07, 3.9945E+07,
     + 4.0047E+07, 4.0149E+07, 4.0251E+07, 4.0354E+07, 4.0456E+07,
     +  4.0559E+07, 4.0662E+07, 4.0766E+07, 4.0869E+07, 4.0973E+07,
     + 4.1077E+07, 4.1181E+07, 4.1285E+07, 4.1390E+07, 4.1495E+07,
     +  4.1600E+07, 4.1705E+07, 4.1810E+07, 4.1916E+07, 4.2022E+07,
     + 4.2128E+07, 4.2234E+07, 4.2341E+07, 4.2447E+07, 4.2554E+07,
     +  4.2661E+07, 4.2768E+07, 4.2876E+07, 4.2984E+07, 4.3092E+07,
     + 4.3200E+07, 4.3308E+07, 4.3417E+07, 4.3526E+07, 4.3635E+07,
     +  4.3744E+07, 4.3853E+07, 4.3963E+07, 4.4073E+07, 4.4183E+07,
     + 4.4293E+07, 4.4404E+07, 4.4514E+07, 4.4625E+07, 4.4736E+07,
     +  4.4848E+07, 4.4959E+07, 4.5071E+07, 4.5183E+07, 4.5295E+07,
     + 4.5408E+07, 4.5521E+07, 4.5633E+07, 4.5747E+07, 4.5860E+07,
     +  4.5974E+07, 4.6087E+07, 4.6201E+07, 4.6316E+07, 4.6430E+07,
     + 4.6545E+07, 4.6660E+07, 4.6775E+07, 4.6890E+07, 4.7006E+07,
     +  4.7121E+07, 4.7237E+07, 4.7354E+07, 4.7470E+07, 4.7587E+07,
     + 4.7704E+07, 4.7821E+07, 4.7938E+07, 4.8056E+07, 4.8174E+07,
     +  4.8292E+07, 4.8410E+07, 4.8529E+07, 4.8647E+07, 4.8766E+07,
     + 4.8885E+07, 4.9005E+07, 4.9124E+07, 4.9244E+07, 4.9364E+07,
     +  4.9485E+07, 4.9605E+07, 4.9726E+07, 4.9847E+07, 4.9968E+07,
     + 5.0090E+07, 5.0212E+07, 5.0334E+07, 5.0456E+07, 5.0578E+07,
     +  5.0701E+07, 5.0824E+07, 5.0947E+07, 5.1070E+07, 5.1194E+07,
     + 5.1318E+07, 5.1442E+07, 5.1566E+07, 5.1691E+07, 5.1815E+07,
     +  5.1940E+07, 5.2066E+07, 5.2191E+07, 5.2317E+07, 5.2443E+07,
     + 5.2569E+07, 5.2695E+07, 5.2822E+07, 5.2949E+07, 5.3076E+07,
     +  5.3204E+07, 5.3331E+07, 5.3459E+07, 5.3587E+07, 5.3716E+07,
     + 5.3844E+07, 5.3973E+07, 5.4102E+07, 5.4231E+07, 5.4361E+07,
     +  5.4491E+07, 5.4621E+07, 5.4751E+07, 5.4882E+07, 5.5012E+07,
     + 5.5143E+07, 5.5275E+07, 5.5406E+07, 5.5538E+07, 5.5670E+07,
     +  5.5802E+07, 5.5935E+07, 5.6068E+07, 5.6201E+07, 5.6334E+07,
     + 5.6467E+07, 5.6601E+07, 5.6735E+07, 5.6869E+07, 5.7004E+07,
     +  5.7139E+07, 5.7274E+07, 5.7409E+07, 5.7544E+07, 5.7680E+07,
     + 5.7816E+07, 5.7952E+07, 5.8089E+07, 5.8226E+07, 5.8363E+07,
     +  5.8500E+07, 5.8637E+07, 5.8775E+07, 5.8913E+07, 5.9051E+07,
     + 5.9190E+07, 5.9329E+07, 5.9468E+07, 5.9607E+07, 5.9747E+07,
     +  5.9887E+07, 6.0027E+07, 6.0167E+07, 6.0308E+07, 6.0448E+07,
     + 6.0589E+07, 6.0731E+07, 6.0872E+07, 6.1014E+07, 6.1157E+07,
     +  6.1299E+07, 6.1442E+07, 6.1585E+07, 6.1728E+07, 6.1871E+07,
     + 6.2015E+07, 6.2159E+07, 6.2303E+07, 6.2448E+07, 6.2592E+07,
     +  6.2737E+07, 6.2883E+07, 6.3028E+07, 6.3174E+07, 6.3320E+07,
     + 6.3466E+07, 6.3613E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT(10,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       776
      data (QofT(11,J),J=1,2408)/ 5.9468E+01, 1.5373E+02, 2.7373E+02,
     +  4.1491E+02, 5.7445E+02, 7.5046E+02, 9.4151E+02, 1.1465E+03,
     + 1.3646E+03, 1.5949E+03, 1.8370E+03, 2.0902E+03, 2.3541E+03,
     +  2.6283E+03, 2.9123E+03, 3.2060E+03, 3.5089E+03, 3.8207E+03,
     + 4.1414E+03, 4.4705E+03, 5.1535E+03, 5.8681E+03, 6.6130E+03,
     +  7.3871E+03, 8.1893E+03, 9.0186E+03, 9.8743E+03, 1.0755E+04,
     + 1.1661E+04, 1.2591E+04, 1.3545E+04, 1.4522E+04, 1.5521E+04,
     +  1.6542E+04, 1.7584E+04, 1.8648E+04, 1.9733E+04, 2.0837E+04,
     + 2.1962E+04, 2.3106E+04, 2.4269E+04, 2.5452E+04, 2.6653E+04,
     +  2.7872E+04, 2.9110E+04, 3.0365E+04, 3.1638E+04, 3.2929E+04,
     + 3.4236E+04, 3.5561E+04, 3.6903E+04, 3.8260E+04, 3.9635E+04,
     +  4.1025E+04, 4.2432E+04, 4.3854E+04, 4.5292E+04, 4.6745E+04,
     + 4.8214E+04, 4.9698E+04, 5.1197E+04, 5.2711E+04, 5.4240E+04,
     +  5.5784E+04, 5.7342E+04, 5.8915E+04, 6.0502E+04, 6.2104E+04,
     + 6.3720E+04, 6.5350E+04, 6.6994E+04, 6.8652E+04, 7.0325E+04,
     +  7.2011E+04, 7.3711E+04, 7.5424E+04, 7.7152E+04, 7.8893E+04,
     + 8.0648E+04, 8.2417E+04, 8.4199E+04, 8.5995E+04, 8.7804E+04,
     +  8.9628E+04, 9.1464E+04, 9.3314E+04, 9.5178E+04, 9.7056E+04,
     + 9.8947E+04, 1.0085E+05, 1.0277E+05, 1.0470E+05, 1.0665E+05,
     +  1.0861E+05, 1.1058E+05, 1.1257E+05, 1.1457E+05, 1.1658E+05,
     + 1.1861E+05, 1.2066E+05, 1.2271E+05, 1.2478E+05, 1.2687E+05,
     +  1.2897E+05, 1.3108E+05, 1.3321E+05, 1.3535E+05, 1.3751E+05,
     + 1.3968E+05, 1.4187E+05, 1.4407E+05, 1.4628E+05, 1.4852E+05,
     +  1.5076E+05, 1.5302E+05, 1.5530E+05, 1.5759E+05, 1.5990E+05,
     + 1.6222E+05, 1.6456E+05, 1.6692E+05, 1.6929E+05, 1.7167E+05,
     +  1.7408E+05, 1.7650E+05, 1.7893E+05, 1.8138E+05, 1.8385E+05,
     + 1.8634E+05, 1.8884E+05, 1.9136E+05, 1.9389E+05, 1.9645E+05,
     +  1.9902E+05, 2.0161E+05, 2.0421E+05, 2.0684E+05, 2.0948E+05,
     + 2.1214E+05, 2.1482E+05, 2.1751E+05, 2.2023E+05, 2.2296E+05,
     +  2.2571E+05, 2.2848E+05, 2.3128E+05, 2.3409E+05, 2.3691E+05,
     + 2.3976E+05, 2.4263E+05, 2.4552E+05, 2.4843E+05, 2.5136E+05,
     +  2.5430E+05, 2.5727E+05, 2.6026E+05, 2.6327E+05, 2.6630E+05,
     + 2.6936E+05, 2.7243E+05, 2.7552E+05, 2.7864E+05, 2.8178E+05,
     +  2.8494E+05, 2.8812E+05, 2.9133E+05, 2.9455E+05, 2.9780E+05,
     + 3.0108E+05, 3.0437E+05, 3.0769E+05, 3.1103E+05, 3.1440E+05,
     +  3.1778E+05, 3.2120E+05, 3.2463E+05, 3.2809E+05, 3.3158E+05,
     + 3.3509E+05, 3.3862E+05, 3.4218E+05, 3.4576E+05, 3.4937E+05,
     +  3.5300E+05, 3.5666E+05, 3.6035E+05, 3.6406E+05, 3.6779E+05,
     + 3.7156E+05, 3.7535E+05, 3.7916E+05, 3.8301E+05, 3.8688E+05,
     +  3.9077E+05, 3.9470E+05, 3.9865E+05, 4.0262E+05, 4.0663E+05,
     + 4.1067E+05, 4.1473E+05, 4.1882E+05, 4.2294E+05, 4.2709E+05,
     +  4.3127E+05, 4.3548E+05, 4.3971E+05, 4.4398E+05, 4.4828E+05,
     + 4.5260E+05, 4.5696E+05, 4.6134E+05, 4.6576E+05, 4.7021E+05,
     +  4.7468E+05, 4.7919E+05, 4.8373E+05, 4.8830E+05, 4.9290E+05,
     + 4.9754E+05, 5.0221E+05, 5.0691E+05, 5.1164E+05, 5.1640E+05,
     +  5.2120E+05, 5.2603E+05, 5.3089E+05, 5.3579E+05, 5.4072E+05,
     + 5.4568E+05, 5.5068E+05, 5.5571E+05, 5.6078E+05, 5.6588E+05,
     +  5.7101E+05, 5.7619E+05, 5.8139E+05, 5.8663E+05, 5.9191E+05,
     + 5.9723E+05, 6.0258E+05, 6.0796E+05, 6.1338E+05, 6.1884E+05,
     +  6.2433E+05, 6.2987E+05, 6.3543E+05, 6.4104E+05, 6.4669E+05,
     + 6.5237E+05, 6.5809E+05, 6.6385E+05, 6.6965E+05, 6.7548E+05,
     +  6.8136E+05, 6.8727E+05, 6.9323E+05, 6.9922E+05, 7.0526E+05,
     + 7.1133E+05, 7.1744E+05, 7.2360E+05, 7.2979E+05, 7.3603E+05,
     +  7.4231E+05, 7.4862E+05, 7.5499E+05, 7.6138E+05, 7.6783E+05,
     + 7.7431E+05, 7.8084E+05, 7.8741E+05, 7.9403E+05, 8.0069E+05,
     +  8.0739E+05, 8.1413E+05, 8.2092E+05, 8.2775E+05, 8.3463E+05,
     + 8.4155E+05, 8.4852E+05, 8.5553E+05, 8.6258E+05, 8.6968E+05,
     +  8.7683E+05, 8.8402E+05, 8.9127E+05, 8.9855E+05, 9.0588E+05,
     + 9.1326E+05, 9.2069E+05, 9.2816E+05, 9.3568E+05, 9.4325E+05,
     +  9.5086E+05, 9.5853E+05, 9.6624E+05, 9.7401E+05, 9.8182E+05,
     + 9.8968E+05, 9.9759E+05, 1.0055E+06, 1.0136E+06, 1.0216E+06,
     +  1.0297E+06, 1.0379E+06, 1.0461E+06, 1.0544E+06, 1.0627E+06,
     + 1.0710E+06, 1.0795E+06, 1.0879E+06, 1.0964E+06, 1.1050E+06,
     +  1.1136E+06, 1.1223E+06, 1.1311E+06, 1.1399E+06, 1.1487E+06,
     + 1.1576E+06, 1.1665E+06, 1.1755E+06, 1.1846E+06, 1.1937E+06,
     +  1.2029E+06, 1.2121E+06, 1.2214E+06, 1.2307E+06, 1.2401E+06,
     + 1.2496E+06, 1.2591E+06, 1.2687E+06, 1.2783E+06, 1.2880E+06,
     +  1.2977E+06, 1.3075E+06, 1.3174E+06, 1.3273E+06, 1.3373E+06,
     + 1.3473E+06, 1.3574E+06, 1.3675E+06, 1.3778E+06, 1.3880E+06,
     +  1.3984E+06, 1.4088E+06, 1.4192E+06, 1.4297E+06, 1.4403E+06,
     + 1.4510E+06, 1.4617E+06, 1.4724E+06, 1.4833E+06, 1.4942E+06,
     +  1.5051E+06, 1.5161E+06, 1.5272E+06, 1.5384E+06, 1.5496E+06,
     + 1.5609E+06, 1.5722E+06, 1.5836E+06, 1.5951E+06, 1.6067E+06,
     +  1.6183E+06, 1.6299E+06, 1.6417E+06, 1.6535E+06, 1.6654E+06,
     + 1.6773E+06, 1.6893E+06, 1.7014E+06, 1.7136E+06, 1.7258E+06,
     +  1.7381E+06, 1.7504E+06, 1.7629E+06, 1.7753E+06, 1.7879E+06,
     + 1.8006E+06, 1.8133E+06, 1.8260E+06, 1.8389E+06, 1.8518E+06,
     +  1.8648E+06, 1.8779E+06, 1.8910E+06, 1.9042E+06, 1.9175E+06,
     + 1.9309E+06, 1.9443E+06, 1.9579E+06, 1.9714E+06, 1.9851E+06,
     +  1.9988E+06, 2.0126E+06, 2.0265E+06, 2.0405E+06, 2.0545E+06,
     + 2.0687E+06, 2.0828E+06, 2.0971E+06, 2.1115E+06, 2.1259E+06,
     +  2.1404E+06, 2.1550E+06, 2.1697E+06, 2.1844E+06, 2.1992E+06,
     + 2.2141E+06, 2.2291E+06, 2.2442E+06, 2.2593E+06, 2.2745E+06,
     +  2.2898E+06, 2.3052E+06, 2.3207E+06, 2.3363E+06, 2.3519E+06,
     + 2.3676E+06, 2.3834E+06, 2.3993E+06, 2.4153E+06, 2.4313E+06,
     +  2.4475E+06, 2.4637E+06, 2.4800E+06, 2.4964E+06, 2.5129E+06,
     + 2.5295E+06, 2.5461E+06, 2.5629E+06, 2.5797E+06, 2.5966E+06,
     +  2.6137E+06, 2.6307E+06, 2.6479E+06, 2.6652E+06, 2.6826E+06,
     + 2.7000E+06, 2.7176E+06, 2.7352E+06, 2.7529E+06, 2.7708E+06,
     +  2.7887E+06, 2.8067E+06, 2.8248E+06, 2.8430E+06, 2.8612E+06,
     + 2.8796E+06, 2.8981E+06, 2.9166E+06, 2.9353E+06, 2.9541E+06,
     +  2.9729E+06, 2.9918E+06, 3.0109E+06, 3.0300E+06, 3.0492E+06,
     + 3.0686E+06, 3.0880E+06, 3.1075E+06, 3.1271E+06, 3.1469E+06,
     +  3.1667E+06, 3.1866E+06, 3.2066E+06, 3.2267E+06, 3.2469E+06,
     + 3.2673E+06, 3.2877E+06, 3.3082E+06, 3.3288E+06, 3.3496E+06,
     +  3.3704E+06, 3.3913E+06, 3.4124E+06, 3.4335E+06, 3.4547E+06,
     + 3.4761E+06, 3.4975E+06, 3.5191E+06, 3.5407E+06, 3.5625E+06,
     +  3.5844E+06, 3.6064E+06, 3.6284E+06, 3.6506E+06, 3.6729E+06,
     + 3.6953E+06, 3.7179E+06, 3.7405E+06, 3.7632E+06, 3.7861E+06,
     +  3.8090E+06, 3.8321E+06, 3.8553E+06, 3.8786E+06, 3.9020E+06,
     + 3.9255E+06, 3.9491E+06, 3.9729E+06, 3.9967E+06, 4.0207E+06,
     +  4.0447E+06, 4.0689E+06, 4.0932E+06, 4.1177E+06, 4.1422E+06,
     + 4.1669E+06, 4.1916E+06, 4.2165E+06, 4.2415E+06, 4.2667E+06,
     +  4.2919E+06, 4.3173E+06, 4.3427E+06, 4.3683E+06, 4.3941E+06,
     + 4.4199E+06, 4.4459E+06, 4.4719E+06, 4.4981E+06, 4.5245E+06,
     +  4.5509E+06, 4.5775E+06, 4.6042E+06, 4.6310E+06, 4.6579E+06,
     + 4.6850E+06, 4.7122E+06, 4.7395E+06, 4.7669E+06, 4.7945E+06,
     +  4.8222E+06, 4.8500E+06, 4.8779E+06, 4.9060E+06, 4.9342E+06,
     + 4.9625E+06, 4.9910E+06, 5.0196E+06, 5.0483E+06, 5.0771E+06,
     +  5.1061E+06, 5.1352E+06, 5.1645E+06, 5.1938E+06, 5.2234E+06,
     + 5.2530E+06, 5.2828E+06, 5.3127E+06, 5.3427E+06, 5.3729E+06,
     +  5.4032E+06, 5.4336E+06, 5.4642E+06, 5.4949E+06, 5.5258E+06,
     + 5.5568E+06, 5.5879E+06, 5.6192E+06, 5.6506E+06, 5.6821E+06,
     +  5.7138E+06, 5.7456E+06, 5.7776E+06, 5.8097E+06, 5.8419E+06,
     + 5.8743E+06, 5.9069E+06, 5.9395E+06, 5.9724E+06, 6.0053E+06,
     +  6.0384E+06, 6.0717E+06, 6.1051E+06, 6.1386E+06, 6.1723E+06,
     + 6.2061E+06, 6.2401E+06, 6.2742E+06, 6.3085E+06, 6.3429E+06,
     +  6.3775E+06, 6.4122E+06, 6.4471E+06, 6.4821E+06, 6.5173E+06,
     + 6.5526E+06, 6.5880E+06, 6.6237E+06, 6.6594E+06, 6.6954E+06,
     +  6.7315E+06, 6.7677E+06, 6.8041E+06, 6.8406E+06, 6.8773E+06,
     + 6.9142E+06, 6.9512E+06, 6.9884E+06, 7.0257E+06, 7.0632E+06,
     +  7.1008E+06, 7.1386E+06, 7.1766E+06, 7.2147E+06, 7.2530E+06,
     + 7.2914E+06, 7.3300E+06, 7.3688E+06, 7.4078E+06, 7.4468E+06,
     +  7.4861E+06, 7.5255E+06, 7.5651E+06, 7.6048E+06, 7.6447E+06,
     + 7.6848E+06, 7.7251E+06, 7.7655E+06, 7.8061E+06, 7.8468E+06,
     +  7.8877E+06, 7.9288E+06, 7.9700E+06, 8.0115E+06, 8.0531E+06,
     + 8.0948E+06, 8.1367E+06, 8.1789E+06, 8.2211E+06, 8.2636E+06,
     +  8.3062E+06, 8.3490E+06, 8.3920E+06, 8.4351E+06, 8.4784E+06,
     + 8.5219E+06, 8.5656E+06, 8.6095E+06, 8.6535E+06, 8.6977E+06,
     +  8.7421E+06, 8.7866E+06, 8.8314E+06, 8.8763E+06, 8.9214E+06,
     + 8.9667E+06, 9.0122E+06, 9.0578E+06, 9.1036E+06, 9.1496E+06,
     +  9.1958E+06, 9.2422E+06, 9.2888E+06, 9.3355E+06, 9.3825E+06,
     + 9.4296E+06, 9.4769E+06, 9.5244E+06, 9.5721E+06, 9.6200E+06,
     +  9.6681E+06, 9.7163E+06, 9.7648E+06, 9.8134E+06, 9.8622E+06,
     + 9.9112E+06, 9.9605E+06, 1.0010E+07, 1.0059E+07, 1.0109E+07,
     +  1.0159E+07, 1.0209E+07, 1.0260E+07, 1.0310E+07, 1.0361E+07,
     + 1.0412E+07, 1.0463E+07, 1.0515E+07, 1.0566E+07, 1.0618E+07,
     +  1.0670E+07, 1.0722E+07, 1.0774E+07, 1.0827E+07, 1.0880E+07,
     + 1.0933E+07, 1.0986E+07, 1.1039E+07, 1.1093E+07, 1.1147E+07,
     +  1.1201E+07, 1.1255E+07, 1.1309E+07, 1.1364E+07, 1.1419E+07,
     + 1.1474E+07, 1.1529E+07, 1.1585E+07, 1.1640E+07, 1.1696E+07,
     +  1.1752E+07, 1.1809E+07, 1.1865E+07, 1.1922E+07, 1.1979E+07,
     + 1.2036E+07, 1.2093E+07, 1.2151E+07, 1.2209E+07, 1.2267E+07,
     +  1.2325E+07, 1.2383E+07, 1.2442E+07, 1.2501E+07, 1.2560E+07,
     + 1.2619E+07, 1.2679E+07, 1.2739E+07, 1.2799E+07, 1.2859E+07,
     +  1.2919E+07, 1.2980E+07, 1.3041E+07, 1.3102E+07, 1.3163E+07,
     + 1.3225E+07, 1.3287E+07, 1.3349E+07, 1.3411E+07, 1.3474E+07,
     +  1.3536E+07, 1.3599E+07, 1.3663E+07, 1.3726E+07, 1.3790E+07,
     + 1.3853E+07, 1.3918E+07, 1.3982E+07, 1.4046E+07, 1.4111E+07,
     +  1.4176E+07, 1.4242E+07, 1.4307E+07, 1.4373E+07, 1.4439E+07,
     + 1.4505E+07, 1.4572E+07, 1.4638E+07, 1.4705E+07, 1.4773E+07,
     +  1.4840E+07, 1.4908E+07, 1.4976E+07, 1.5044E+07, 1.5112E+07,
     + 1.5181E+07, 1.5250E+07, 1.5319E+07, 1.5388E+07, 1.5458E+07,
     +  1.5528E+07, 1.5598E+07, 1.5668E+07, 1.5739E+07, 1.5810E+07,
     + 1.5881E+07, 1.5953E+07, 1.6024E+07, 1.6096E+07, 1.6168E+07,
     +  1.6241E+07, 1.6313E+07, 1.6386E+07, 1.6460E+07, 1.6533E+07,
     + 1.6607E+07, 1.6681E+07, 1.6755E+07, 1.6829E+07, 1.6904E+07,
     +  1.6979E+07, 1.7054E+07, 1.7130E+07, 1.7206E+07, 1.7282E+07,
     + 1.7358E+07, 1.7435E+07, 1.7512E+07, 1.7589E+07, 1.7666E+07,
     +  1.7744E+07, 1.7822E+07, 1.7900E+07, 1.7978E+07, 1.8057E+07,
     + 1.8136E+07, 1.8215E+07, 1.8295E+07, 1.8375E+07, 1.8455E+07,
     +  1.8535E+07, 1.8616E+07, 1.8697E+07, 1.8778E+07, 1.8860E+07,
     + 1.8941E+07, 1.9024E+07, 1.9106E+07, 1.9188E+07, 1.9271E+07,
     +  1.9355E+07, 1.9438E+07, 1.9522E+07, 1.9606E+07, 1.9690E+07,
     + 1.9775E+07, 1.9860E+07, 1.9945E+07, 2.0030E+07, 2.0116E+07,
     +  2.0202E+07, 2.0288E+07, 2.0375E+07, 2.0462E+07, 2.0549E+07,
     + 2.0637E+07, 2.0724E+07, 2.0812E+07, 2.0901E+07, 2.0989E+07,
     +  2.1078E+07, 2.1168E+07, 2.1257E+07, 2.1347E+07, 2.1437E+07,
     + 2.1528E+07, 2.1619E+07, 2.1710E+07, 2.1801E+07, 2.1893E+07,
     +  2.1985E+07, 2.2077E+07, 2.2169E+07, 2.2262E+07, 2.2356E+07,
     + 2.2449E+07, 2.2543E+07, 2.2637E+07, 2.2731E+07, 2.2826E+07,
     +  2.2921E+07, 2.3017E+07, 2.3112E+07, 2.3208E+07, 2.3305E+07,
     + 2.3401E+07, 2.3498E+07, 2.3595E+07, 2.3693E+07, 2.3791E+07,
     +  2.3889E+07, 2.3987E+07, 2.4086E+07, 2.4185E+07, 2.4285E+07,
     + 2.4385E+07, 2.4485E+07, 2.4585E+07, 2.4686E+07, 2.4787E+07,
     +  2.4889E+07, 2.4990E+07, 2.5092E+07, 2.5195E+07, 2.5298E+07,
     + 2.5401E+07, 2.5504E+07, 2.5608E+07, 2.5712E+07, 2.5816E+07,
     +  2.5921E+07, 2.6026E+07, 2.6131E+07, 2.6237E+07, 2.6343E+07,
     + 2.6450E+07, 2.6556E+07, 2.6663E+07, 2.6771E+07, 2.6879E+07,
     +  2.6987E+07, 2.7095E+07, 2.7204E+07, 2.7313E+07, 2.7423E+07,
     + 2.7532E+07, 2.7643E+07, 2.7753E+07, 2.7864E+07, 2.7975E+07,
     +  2.8087E+07, 2.8199E+07, 2.8311E+07, 2.8424E+07, 2.8537E+07,
     + 2.8650E+07, 2.8764E+07, 2.8878E+07, 2.8992E+07, 2.9107E+07,
     +  2.9222E+07, 2.9337E+07, 2.9453E+07, 2.9569E+07, 2.9686E+07,
     + 2.9803E+07, 2.9920E+07, 3.0038E+07, 3.0156E+07, 3.0274E+07,
     +  3.0393E+07, 3.0512E+07, 3.0631E+07, 3.0751E+07, 3.0871E+07,
     + 3.0992E+07, 3.1113E+07, 3.1234E+07, 3.1356E+07, 3.1478E+07,
     +  3.1600E+07, 3.1723E+07, 3.1846E+07, 3.1970E+07, 3.2094E+07,
     + 3.2218E+07, 3.2343E+07, 3.2468E+07, 3.2593E+07, 3.2719E+07,
     +  3.2845E+07, 3.2972E+07, 3.3099E+07, 3.3226E+07, 3.3354E+07,
     + 3.3482E+07, 3.3611E+07, 3.3739E+07, 3.3869E+07, 3.3998E+07,
     +  3.4128E+07, 3.4259E+07, 3.4390E+07, 3.4521E+07, 3.4653E+07,
     + 3.4785E+07, 3.4917E+07, 3.5050E+07, 3.5183E+07, 3.5317E+07,
     +  3.5451E+07, 3.5585E+07, 3.5720E+07, 3.5855E+07, 3.5991E+07,
     + 3.6127E+07, 3.6264E+07, 3.6400E+07, 3.6538E+07, 3.6675E+07,
     +  3.6813E+07, 3.6952E+07, 3.7091E+07, 3.7230E+07, 3.7370E+07,
     + 3.7510E+07, 3.7650E+07, 3.7791E+07, 3.7933E+07, 3.8075E+07,
     +  3.8217E+07, 3.8359E+07, 3.8502E+07, 3.8646E+07, 3.8790E+07,
     + 3.8934E+07, 3.9079E+07, 3.9224E+07, 3.9369E+07, 3.9515E+07,
     +  3.9662E+07, 3.9809E+07, 3.9956E+07, 4.0104E+07, 4.0252E+07,
     + 4.0400E+07, 4.0549E+07, 4.0699E+07, 4.0849E+07, 4.0999E+07,
     +  4.1150E+07, 4.1301E+07, 4.1452E+07, 4.1604E+07, 4.1757E+07,
     + 4.1910E+07, 4.2063E+07, 4.2217E+07, 4.2371E+07, 4.2526E+07,
     +  4.2681E+07, 4.2837E+07, 4.2993E+07, 4.3149E+07, 4.3306E+07,
     + 4.3463E+07, 4.3621E+07, 4.3779E+07, 4.3938E+07, 4.4097E+07,
     +  4.4257E+07, 4.4417E+07, 4.4578E+07, 4.4739E+07, 4.4900E+07,
     + 4.5062E+07, 4.5224E+07, 4.5387E+07, 4.5550E+07, 4.5714E+07,
     +  4.5878E+07, 4.6043E+07, 4.6208E+07, 4.6374E+07, 4.6540E+07,
     + 4.6706E+07, 4.6873E+07, 4.7041E+07, 4.7209E+07, 4.7377E+07,
     +  4.7546E+07, 4.7715E+07, 4.7885E+07, 4.8055E+07, 4.8226E+07,
     + 4.8397E+07, 4.8569E+07, 4.8741E+07, 4.8914E+07, 4.9087E+07,
     +  4.9261E+07, 4.9435E+07, 4.9610E+07, 4.9785E+07, 4.9960E+07,
     + 5.0137E+07, 5.0313E+07, 5.0490E+07, 5.0668E+07, 5.0846E+07,
     +  5.1024E+07, 5.1203E+07, 5.1383E+07, 5.1563E+07, 5.1743E+07,
     + 5.1924E+07, 5.2106E+07, 5.2288E+07, 5.2470E+07, 5.2653E+07,
     +  5.2837E+07, 5.3021E+07, 5.3205E+07, 5.3390E+07, 5.3576E+07,
     + 5.3762E+07, 5.3949E+07, 5.4136E+07, 5.4323E+07, 5.4511E+07,
     +  5.4700E+07, 5.4889E+07, 5.5079E+07, 5.5269E+07, 5.5459E+07,
     + 5.5650E+07, 5.5842E+07, 5.6034E+07, 5.6227E+07, 5.6420E+07,
     +  5.6614E+07, 5.6808E+07, 5.7003E+07, 5.7198E+07, 5.7394E+07,
     + 5.7591E+07, 5.7788E+07, 5.7985E+07, 5.8183E+07, 5.8382E+07,
     +  5.8581E+07, 5.8780E+07, 5.8980E+07, 5.9181E+07, 5.9382E+07,
     + 5.9584E+07, 5.9786E+07, 5.9989E+07, 6.0192E+07, 6.0396E+07,
     +  6.0600E+07, 6.0805E+07, 6.1011E+07, 6.1217E+07, 6.1424E+07,
     + 6.1631E+07, 6.1838E+07, 6.2047E+07, 6.2255E+07, 6.2465E+07,
     +  6.2675E+07, 6.2885E+07, 6.3096E+07, 6.3308E+07, 6.3520E+07,
     + 6.3732E+07, 6.3946E+07, 6.4160E+07, 6.4374E+07, 6.4589E+07,
     +  6.4804E+07, 6.5020E+07, 6.5237E+07, 6.5454E+07, 6.5672E+07,
     + 6.5890E+07, 6.6109E+07, 6.6329E+07, 6.6549E+07, 6.6769E+07,
     +  6.6991E+07, 6.7212E+07, 6.7435E+07, 6.7658E+07, 6.7881E+07,
     + 6.8105E+07, 6.8330E+07, 6.8555E+07, 6.8781E+07, 6.9007E+07,
     +  6.9234E+07, 6.9462E+07, 6.9690E+07, 6.9919E+07, 7.0148E+07,
     + 7.0378E+07, 7.0609E+07, 7.0840E+07, 7.1072E+07, 7.1304E+07,
     +  7.1537E+07, 7.1770E+07, 7.2004E+07, 7.2239E+07, 7.2474E+07,
     + 7.2710E+07, 7.2947E+07, 7.3184E+07, 7.3422E+07, 7.3660E+07,
     +  7.3899E+07, 7.4139E+07, 7.4379E+07, 7.4620E+07, 7.4861E+07,
     + 7.5103E+07, 7.5346E+07, 7.5589E+07, 7.5833E+07, 7.6077E+07,
     +  7.6322E+07, 7.6568E+07, 7.6814E+07, 7.7061E+07, 7.7309E+07,
     + 7.7557E+07, 7.7806E+07, 7.8055E+07, 7.8305E+07, 7.8556E+07,
     +  7.8808E+07, 7.9060E+07, 7.9312E+07, 7.9565E+07, 7.9819E+07,
     + 8.0074E+07, 8.0329E+07, 8.0585E+07, 8.0841E+07, 8.1098E+07,
     +  8.1356E+07, 8.1614E+07, 8.1873E+07, 8.2133E+07, 8.2393E+07,
     + 8.2654E+07, 8.2916E+07, 8.3178E+07, 8.3441E+07, 8.3705E+07,
     +  8.3969E+07, 8.4234E+07, 8.4499E+07, 8.4765E+07, 8.5032E+07,
     + 8.5300E+07, 8.5568E+07, 8.5837E+07, 8.6106E+07, 8.6376E+07,
     +  8.6647E+07, 8.6919E+07, 8.7191E+07, 8.7464E+07, 8.7737E+07,
     + 8.8011E+07, 8.8286E+07, 8.8562E+07, 8.8838E+07, 8.9115E+07,
     +  8.9392E+07, 8.9670E+07, 8.9949E+07, 9.0229E+07, 9.0509E+07,
     + 9.0790E+07, 9.1072E+07, 9.1354E+07, 9.1637E+07, 9.1921E+07,
     +  9.2205E+07, 9.2490E+07, 9.2776E+07, 9.3063E+07, 9.3350E+07,
     + 9.3638E+07, 9.3926E+07, 9.4215E+07, 9.4505E+07, 9.4796E+07,
     +  9.5087E+07, 9.5379E+07, 9.5672E+07, 9.5966E+07, 9.6260E+07,
     + 9.6555E+07, 9.6850E+07, 9.7147E+07, 9.7444E+07, 9.7741E+07,
     +  9.8040E+07, 9.8339E+07, 9.8639E+07, 9.8940E+07, 9.9241E+07,
     + 9.9543E+07, 9.9846E+07, 1.0015E+08, 1.0045E+08, 1.0076E+08,
     +  1.0106E+08, 1.0137E+08, 1.0168E+08, 1.0199E+08, 1.0229E+08,
     + 1.0260E+08, 1.0291E+08, 1.0322E+08, 1.0354E+08, 1.0385E+08,
     +  1.0416E+08, 1.0447E+08, 1.0479E+08, 1.0510E+08, 1.0542E+08,
     + 1.0574E+08, 1.0605E+08, 1.0637E+08, 1.0669E+08, 1.0701E+08,
     +  1.0733E+08, 1.0765E+08, 1.0798E+08, 1.0830E+08, 1.0862E+08,
     + 1.0895E+08, 1.0927E+08, 1.0960E+08, 1.0992E+08, 1.1025E+08,
     +  1.1058E+08, 1.1091E+08, 1.1124E+08, 1.1157E+08, 1.1190E+08,
     + 1.1223E+08, 1.1256E+08, 1.1290E+08, 1.1323E+08, 1.1357E+08,
     +  1.1390E+08, 1.1424E+08, 1.1458E+08, 1.1491E+08, 1.1525E+08,
     + 1.1559E+08, 1.1593E+08, 1.1627E+08, 1.1662E+08, 1.1696E+08,
     +  1.1730E+08, 1.1765E+08, 1.1799E+08, 1.1834E+08, 1.1869E+08,
     + 1.1903E+08, 1.1938E+08, 1.1973E+08, 1.2008E+08, 1.2043E+08,
     +  1.2078E+08, 1.2114E+08, 1.2149E+08, 1.2184E+08, 1.2220E+08,
     + 1.2255E+08, 1.2291E+08, 1.2327E+08, 1.2363E+08, 1.2399E+08,
     +  1.2435E+08, 1.2471E+08, 1.2507E+08, 1.2543E+08, 1.2579E+08,
     + 1.2616E+08, 1.2652E+08, 1.2689E+08, 1.2725E+08, 1.2762E+08,
     +  1.2799E+08, 1.2836E+08, 1.2873E+08, 1.2910E+08, 1.2947E+08,
     + 1.2984E+08, 1.3021E+08, 1.3059E+08, 1.3096E+08, 1.3134E+08,
     +  1.3172E+08, 1.3209E+08, 1.3247E+08, 1.3285E+08, 1.3323E+08,
     + 1.3361E+08, 1.3399E+08, 1.3437E+08, 1.3476E+08, 1.3514E+08,
     +  1.3553E+08, 1.3591E+08, 1.3630E+08, 1.3669E+08, 1.3708E+08,
     + 1.3746E+08, 1.3785E+08, 1.3825E+08, 1.3864E+08, 1.3903E+08,
     +  1.3942E+08, 1.3982E+08, 1.4021E+08, 1.4061E+08, 1.4101E+08,
     + 1.4141E+08, 1.4180E+08, 1.4220E+08, 1.4261E+08, 1.4301E+08,
     +  1.4341E+08, 1.4381E+08, 1.4422E+08, 1.4462E+08, 1.4503E+08,
     + 1.4544E+08, 1.4584E+08, 1.4625E+08, 1.4666E+08, 1.4707E+08,
     +  1.4748E+08, 1.4790E+08, 1.4831E+08, 1.4872E+08, 1.4914E+08,
     + 1.4955E+08, 1.4997E+08, 1.5039E+08, 1.5081E+08, 1.5123E+08,
     +  1.5165E+08, 1.5207E+08, 1.5249E+08, 1.5292E+08, 1.5334E+08,
     + 1.5377E+08, 1.5419E+08, 1.5462E+08, 1.5505E+08, 1.5548E+08,
     +  1.5591E+08, 1.5634E+08, 1.5677E+08, 1.5720E+08, 1.5763E+08,
     + 1.5807E+08, 1.5850E+08, 1.5894E+08, 1.5938E+08, 1.5982E+08,
     +  1.6026E+08, 1.6070E+08, 1.6114E+08, 1.6158E+08, 1.6202E+08,
     + 1.6247E+08, 1.6291E+08, 1.6336E+08, 1.6380E+08, 1.6425E+08,
     +  1.6470E+08, 1.6515E+08, 1.6560E+08, 1.6605E+08, 1.6651E+08,
     + 1.6696E+08, 1.6741E+08, 1.6787E+08, 1.6833E+08, 1.6878E+08,
     +  1.6924E+08, 1.6970E+08, 1.7016E+08, 1.7062E+08, 1.7109E+08,
     + 1.7155E+08, 1.7201E+08, 1.7248E+08, 1.7294E+08, 1.7341E+08,
     +  1.7388E+08, 1.7435E+08, 1.7482E+08, 1.7529E+08, 1.7576E+08,
     + 1.7624E+08, 1.7671E+08, 1.7719E+08, 1.7766E+08, 1.7814E+08,
     +  1.7862E+08, 1.7910E+08, 1.7958E+08, 1.8006E+08, 1.8054E+08,
     + 1.8103E+08, 1.8151E+08, 1.8200E+08, 1.8248E+08, 1.8297E+08,
     +  1.8346E+08, 1.8395E+08, 1.8444E+08, 1.8493E+08, 1.8542E+08,
     + 1.8592E+08, 1.8641E+08, 1.8691E+08, 1.8740E+08, 1.8790E+08,
     +  1.8840E+08, 1.8890E+08, 1.8940E+08, 1.8990E+08, 1.9041E+08,
     + 1.9091E+08, 1.9142E+08, 1.9192E+08, 1.9243E+08, 1.9294E+08,
     +  1.9345E+08, 1.9396E+08, 1.9447E+08, 1.9498E+08, 1.9549E+08,
     + 1.9601E+08, 1.9652E+08, 1.9704E+08, 1.9756E+08, 1.9808E+08,
     +  1.9860E+08, 1.9912E+08, 1.9964E+08, 2.0016E+08, 2.0069E+08,
     + 2.0121E+08, 2.0174E+08, 2.0227E+08, 2.0280E+08, 2.0333E+08,
     +  2.0386E+08, 2.0439E+08, 2.0492E+08, 2.0546E+08, 2.0599E+08,
     + 2.0653E+08, 2.0706E+08, 2.0760E+08, 2.0814E+08, 2.0868E+08,
     +  2.0922E+08, 2.0977E+08, 2.1031E+08, 2.1086E+08, 2.1140E+08,
     + 2.1195E+08, 2.1250E+08, 2.1305E+08, 2.1360E+08, 2.1415E+08,
     +  2.1470E+08, 2.1526E+08, 2.1581E+08, 2.1637E+08, 2.1693E+08,
     + 2.1748E+08, 2.1804E+08, 2.1860E+08, 2.1917E+08, 2.1973E+08,
     +  2.2029E+08, 2.2086E+08, 2.2142E+08, 2.2199E+08, 2.2256E+08,
     + 2.2313E+08, 2.2370E+08, 2.2427E+08, 2.2485E+08, 2.2542E+08,
     +  2.2600E+08, 2.2657E+08, 2.2715E+08, 2.2773E+08, 2.2831E+08,
     + 2.2889E+08, 2.2947E+08, 2.3006E+08, 2.3064E+08, 2.3123E+08,
     +  2.3182E+08, 2.3241E+08, 2.3299E+08, 2.3359E+08, 2.3418E+08,
     + 2.3477E+08, 2.3536E+08, 2.3596E+08, 2.3656E+08, 2.3715E+08,
     +  2.3775E+08, 2.3835E+08, 2.3896E+08, 2.3956E+08, 2.4016E+08,
     + 2.4077E+08, 2.4137E+08, 2.4198E+08, 2.4259E+08, 2.4320E+08,
     +  2.4381E+08, 2.4442E+08, 2.4503E+08, 2.4565E+08, 2.4627E+08,
     + 2.4688E+08, 2.4750E+08, 2.4812E+08, 2.4874E+08, 2.4936E+08,
     +  2.4999E+08, 2.5061E+08, 2.5124E+08, 2.5186E+08, 2.5249E+08,
     + 2.5312E+08, 2.5375E+08, 2.5438E+08, 2.5501E+08, 2.5565E+08,
     +  2.5628E+08, 2.5692E+08, 2.5756E+08, 2.5820E+08, 2.5884E+08,
     + 2.5948E+08, 2.6012E+08, 2.6077E+08, 2.6141E+08, 2.6206E+08,
     +  2.6271E+08, 2.6336E+08, 2.6401E+08, 2.6466E+08, 2.6531E+08,
     + 2.6597E+08, 2.6662E+08, 2.6728E+08, 2.6794E+08, 2.6859E+08,
     +  2.6926E+08, 2.6992E+08, 2.7058E+08, 2.7124E+08, 2.7191E+08,
     + 2.7258E+08, 2.7325E+08, 2.7392E+08, 2.7459E+08, 2.7526E+08,
     +  2.7593E+08, 2.7661E+08, 2.7728E+08, 2.7796E+08, 2.7864E+08,
     + 2.7932E+08, 2.8000E+08, 2.8068E+08, 2.8137E+08, 2.8205E+08,
     +  2.8274E+08, 2.8342E+08, 2.8411E+08, 2.8480E+08, 2.8550E+08,
     + 2.8619E+08, 2.8688E+08, 2.8758E+08, 2.8828E+08, 2.8897E+08,
     +  2.8967E+08, 2.9038E+08, 2.9108E+08, 2.9178E+08, 2.9249E+08,
     + 2.9319E+08, 2.9390E+08, 2.9461E+08, 2.9532E+08, 2.9603E+08,
     +  2.9674E+08, 2.9746E+08, 2.9817E+08, 2.9889E+08, 2.9961E+08,
     + 3.0033E+08, 3.0105E+08, 3.0177E+08, 3.0250E+08, 3.0322E+08,
     +  3.0395E+08, 3.0468E+08, 3.0541E+08, 3.0614E+08, 3.0687E+08,
     + 3.0760E+08, 3.0834E+08, 3.0907E+08, 3.0981E+08, 3.1055E+08,
     +  3.1129E+08, 3.1203E+08, 3.1278E+08, 3.1352E+08, 3.1427E+08,
     + 3.1501E+08, 3.1576E+08, 3.1651E+08, 3.1726E+08, 3.1802E+08,
     +  3.1877E+08, 3.1953E+08, 3.2029E+08, 3.2104E+08, 3.2180E+08,
     + 3.2256E+08, 3.2333E+08, 3.2409E+08, 3.2486E+08, 3.2562E+08,
     +  3.2639E+08, 3.2716E+08, 3.2793E+08, 3.2871E+08, 3.2948E+08,
     + 3.3026E+08, 3.3103E+08, 3.3181E+08, 3.3259E+08, 3.3337E+08,
     +  3.3416E+08, 3.3494E+08, 3.3573E+08, 3.3651E+08, 3.3730E+08,
     + 3.3809E+08, 3.3888E+08, 3.3968E+08, 3.4047E+08, 3.4127E+08,
     +  3.4207E+08, 3.4286E+08, 3.4366E+08, 3.4447E+08, 3.4527E+08,
     + 3.4607E+08, 3.4688E+08, 3.4769E+08, 3.4850E+08, 3.4931E+08,
     +  3.5012E+08, 3.5093E+08, 3.5175E+08, 3.5257E+08, 3.5338E+08,
     + 3.5420E+08, 3.5502E+08, 3.5585E+08, 3.5667E+08, 3.5750E+08,
     +  3.5832E+08, 3.5915E+08, 3.5998E+08, 3.6081E+08, 3.6165E+08,
     + 3.6248E+08, 3.6332E+08, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT(11,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       767
      data (QofT(12,J),J=1,2408)/ 2.9951E+01, 7.7617E+01, 1.3832E+02,
     +  2.0973E+02, 2.9044E+02, 3.7949E+02, 4.7615E+02, 5.7987E+02,
     + 6.9020E+02, 8.0676E+02, 9.2924E+02, 1.0574E+03, 1.1909E+03,
     +  1.3296E+03, 1.4734E+03, 1.6219E+03, 1.7752E+03, 1.9330E+03,
     + 2.0953E+03, 2.2618E+03, 2.6074E+03, 2.9690E+03, 3.3459E+03,
     +  3.7377E+03, 4.1436E+03, 4.5632E+03, 4.9962E+03, 5.4421E+03,
     + 5.9005E+03, 6.3712E+03, 6.8538E+03, 7.3480E+03, 7.8536E+03,
     +  8.3702E+03, 8.8978E+03, 9.4361E+03, 9.9848E+03, 1.0544E+04,
     + 1.1113E+04, 1.1692E+04, 1.2281E+04, 1.2879E+04, 1.3487E+04,
     +  1.4104E+04, 1.4730E+04, 1.5365E+04, 1.6010E+04, 1.6663E+04,
     + 1.7324E+04, 1.7995E+04, 1.8674E+04, 1.9361E+04, 2.0056E+04,
     +  2.0760E+04, 2.1471E+04, 2.2191E+04, 2.2919E+04, 2.3654E+04,
     + 2.4397E+04, 2.5148E+04, 2.5907E+04, 2.6673E+04, 2.7447E+04,
     +  2.8228E+04, 2.9017E+04, 2.9813E+04, 3.0616E+04, 3.1427E+04,
     + 3.2244E+04, 3.3069E+04, 3.3902E+04, 3.4741E+04, 3.5587E+04,
     +  3.6440E+04, 3.7301E+04, 3.8168E+04, 3.9042E+04, 3.9924E+04,
     + 4.0812E+04, 4.1707E+04, 4.2609E+04, 4.3518E+04, 4.4434E+04,
     +  4.5357E+04, 4.6286E+04, 4.7223E+04, 4.8167E+04, 4.9117E+04,
     + 5.0074E+04, 5.1038E+04, 5.2009E+04, 5.2987E+04, 5.3972E+04,
     +  5.4964E+04, 5.5963E+04, 5.6969E+04, 5.7982E+04, 5.9002E+04,
     + 6.0029E+04, 6.1063E+04, 6.2105E+04, 6.3153E+04, 6.4209E+04,
     +  6.5272E+04, 6.6342E+04, 6.7419E+04, 6.8504E+04, 6.9596E+04,
     + 7.0696E+04, 7.1802E+04, 7.2917E+04, 7.4039E+04, 7.5168E+04,
     +  7.6306E+04, 7.7450E+04, 7.8603E+04, 7.9764E+04, 8.0932E+04,
     + 8.2109E+04, 8.3292E+04, 8.4484E+04, 8.5685E+04, 8.6893E+04,
     +  8.8109E+04, 8.9334E+04, 9.0567E+04, 9.1809E+04, 9.3058E+04,
     + 9.4316E+04, 9.5583E+04, 9.6858E+04, 9.8143E+04, 9.9435E+04,
     +  1.0074E+05, 1.0205E+05, 1.0337E+05, 1.0469E+05, 1.0603E+05,
     + 1.0738E+05, 1.0873E+05, 1.1010E+05, 1.1147E+05, 1.1286E+05,
     +  1.1425E+05, 1.1565E+05, 1.1707E+05, 1.1849E+05, 1.1992E+05,
     + 1.2136E+05, 1.2281E+05, 1.2428E+05, 1.2575E+05, 1.2723E+05,
     +  1.2872E+05, 1.3022E+05, 1.3174E+05, 1.3326E+05, 1.3479E+05,
     + 1.3634E+05, 1.3789E+05, 1.3946E+05, 1.4104E+05, 1.4263E+05,
     +  1.4423E+05, 1.4584E+05, 1.4746E+05, 1.4909E+05, 1.5073E+05,
     + 1.5239E+05, 1.5406E+05, 1.5574E+05, 1.5743E+05, 1.5913E+05,
     +  1.6084E+05, 1.6257E+05, 1.6431E+05, 1.6606E+05, 1.6782E+05,
     + 1.6959E+05, 1.7138E+05, 1.7318E+05, 1.7499E+05, 1.7682E+05,
     +  1.7866E+05, 1.8051E+05, 1.8237E+05, 1.8425E+05, 1.8614E+05,
     + 1.8804E+05, 1.8996E+05, 1.9189E+05, 1.9383E+05, 1.9579E+05,
     +  1.9776E+05, 1.9974E+05, 2.0174E+05, 2.0375E+05, 2.0578E+05,
     + 2.0782E+05, 2.0987E+05, 2.1194E+05, 2.1402E+05, 2.1612E+05,
     +  2.1823E+05, 2.2036E+05, 2.2250E+05, 2.2466E+05, 2.2683E+05,
     + 2.2901E+05, 2.3121E+05, 2.3343E+05, 2.3566E+05, 2.3791E+05,
     +  2.4017E+05, 2.4245E+05, 2.4475E+05, 2.4706E+05, 2.4938E+05,
     + 2.5172E+05, 2.5408E+05, 2.5646E+05, 2.5885E+05, 2.6125E+05,
     +  2.6368E+05, 2.6612E+05, 2.6858E+05, 2.7105E+05, 2.7354E+05,
     + 2.7605E+05, 2.7857E+05, 2.8111E+05, 2.8367E+05, 2.8625E+05,
     +  2.8885E+05, 2.9146E+05, 2.9409E+05, 2.9673E+05, 2.9940E+05,
     + 3.0208E+05, 3.0478E+05, 3.0750E+05, 3.1024E+05, 3.1300E+05,
     +  3.1578E+05, 3.1857E+05, 3.2138E+05, 3.2421E+05, 3.2706E+05,
     + 3.2993E+05, 3.3282E+05, 3.3573E+05, 3.3866E+05, 3.4160E+05,
     +  3.4457E+05, 3.4756E+05, 3.5056E+05, 3.5359E+05, 3.5663E+05,
     + 3.5970E+05, 3.6279E+05, 3.6589E+05, 3.6902E+05, 3.7217E+05,
     +  3.7534E+05, 3.7853E+05, 3.8174E+05, 3.8497E+05, 3.8822E+05,
     + 3.9150E+05, 3.9479E+05, 3.9811E+05, 4.0145E+05, 4.0481E+05,
     +  4.0819E+05, 4.1159E+05, 4.1502E+05, 4.1847E+05, 4.2194E+05,
     + 4.2543E+05, 4.2894E+05, 4.3248E+05, 4.3604E+05, 4.3963E+05,
     +  4.4323E+05, 4.4686E+05, 4.5052E+05, 4.5419E+05, 4.5789E+05,
     + 4.6161E+05, 4.6536E+05, 4.6913E+05, 4.7293E+05, 4.7674E+05,
     +  4.8059E+05, 4.8445E+05, 4.8835E+05, 4.9226E+05, 4.9620E+05,
     + 5.0017E+05, 5.0416E+05, 5.0817E+05, 5.1221E+05, 5.1628E+05,
     +  5.2037E+05, 5.2448E+05, 5.2863E+05, 5.3279E+05, 5.3699E+05,
     + 5.4121E+05, 5.4545E+05, 5.4972E+05, 5.5402E+05, 5.5835E+05,
     +  5.6270E+05, 5.6708E+05, 5.7148E+05, 5.7591E+05, 5.8037E+05,
     + 5.8486E+05, 5.8937E+05, 5.9391E+05, 5.9848E+05, 6.0308E+05,
     +  6.0770E+05, 6.1236E+05, 6.1704E+05, 6.2175E+05, 6.2648E+05,
     + 6.3125E+05, 6.3604E+05, 6.4086E+05, 6.4572E+05, 6.5060E+05,
     +  6.5551E+05, 6.6045E+05, 6.6542E+05, 6.7041E+05, 6.7544E+05,
     + 6.8050E+05, 6.8559E+05, 6.9071E+05, 6.9586E+05, 7.0104E+05,
     +  7.0625E+05, 7.1149E+05, 7.1676E+05, 7.2206E+05, 7.2740E+05,
     + 7.3276E+05, 7.3816E+05, 7.4359E+05, 7.4904E+05, 7.5454E+05,
     +  7.6006E+05, 7.6561E+05, 7.7120E+05, 7.7682E+05, 7.8247E+05,
     + 7.8816E+05, 7.9387E+05, 7.9963E+05, 8.0541E+05, 8.1123E+05,
     +  8.1708E+05, 8.2296E+05, 8.2888E+05, 8.3483E+05, 8.4081E+05,
     + 8.4683E+05, 8.5289E+05, 8.5897E+05, 8.6509E+05, 8.7125E+05,
     +  8.7745E+05, 8.8367E+05, 8.8993E+05, 8.9623E+05, 9.0256E+05,
     + 9.0893E+05, 9.1533E+05, 9.2177E+05, 9.2825E+05, 9.3476E+05,
     +  9.4131E+05, 9.4789E+05, 9.5452E+05, 9.6117E+05, 9.6786E+05,
     + 9.7460E+05, 9.8137E+05, 9.8817E+05, 9.9502E+05, 1.0019E+06,
     +  1.0088E+06, 1.0158E+06, 1.0228E+06, 1.0298E+06, 1.0369E+06,
     + 1.0440E+06, 1.0511E+06, 1.0583E+06, 1.0656E+06, 1.0728E+06,
     +  1.0801E+06, 1.0875E+06, 1.0949E+06, 1.1023E+06, 1.1098E+06,
     + 1.1173E+06, 1.1248E+06, 1.1324E+06, 1.1400E+06, 1.1477E+06,
     +  1.1554E+06, 1.1631E+06, 1.1709E+06, 1.1788E+06, 1.1866E+06,
     + 1.1946E+06, 1.2025E+06, 1.2105E+06, 1.2186E+06, 1.2266E+06,
     +  1.2348E+06, 1.2429E+06, 1.2512E+06, 1.2594E+06, 1.2677E+06,
     + 1.2761E+06, 1.2845E+06, 1.2929E+06, 1.3014E+06, 1.3099E+06,
     +  1.3184E+06, 1.3271E+06, 1.3357E+06, 1.3444E+06, 1.3531E+06,
     + 1.3619E+06, 1.3708E+06, 1.3796E+06, 1.3886E+06, 1.3975E+06,
     +  1.4066E+06, 1.4156E+06, 1.4247E+06, 1.4339E+06, 1.4431E+06,
     + 1.4523E+06, 1.4616E+06, 1.4710E+06, 1.4804E+06, 1.4898E+06,
     +  1.4993E+06, 1.5088E+06, 1.5184E+06, 1.5281E+06, 1.5377E+06,
     + 1.5475E+06, 1.5572E+06, 1.5671E+06, 1.5769E+06, 1.5869E+06,
     +  1.5968E+06, 1.6069E+06, 1.6169E+06, 1.6271E+06, 1.6372E+06,
     + 1.6475E+06, 1.6578E+06, 1.6681E+06, 1.6785E+06, 1.6889E+06,
     +  1.6994E+06, 1.7099E+06, 1.7205E+06, 1.7311E+06, 1.7418E+06,
     + 1.7526E+06, 1.7634E+06, 1.7742E+06, 1.7851E+06, 1.7961E+06,
     +  1.8071E+06, 1.8181E+06, 1.8292E+06, 1.8404E+06, 1.8516E+06,
     + 1.8629E+06, 1.8742E+06, 1.8856E+06, 1.8971E+06, 1.9086E+06,
     +  1.9201E+06, 1.9317E+06, 1.9434E+06, 1.9551E+06, 1.9669E+06,
     + 1.9787E+06, 1.9906E+06, 2.0025E+06, 2.0145E+06, 2.0266E+06,
     +  2.0387E+06, 2.0509E+06, 2.0631E+06, 2.0754E+06, 2.0878E+06,
     + 2.1002E+06, 2.1126E+06, 2.1252E+06, 2.1377E+06, 2.1504E+06,
     +  2.1631E+06, 2.1758E+06, 2.1887E+06, 2.2015E+06, 2.2145E+06,
     + 2.2275E+06, 2.2405E+06, 2.2537E+06, 2.2668E+06, 2.2801E+06,
     +  2.2934E+06, 2.3067E+06, 2.3202E+06, 2.3337E+06, 2.3472E+06,
     + 2.3608E+06, 2.3745E+06, 2.3882E+06, 2.4020E+06, 2.4159E+06,
     +  2.4298E+06, 2.4438E+06, 2.4579E+06, 2.4720E+06, 2.4862E+06,
     + 2.5004E+06, 2.5148E+06, 2.5291E+06, 2.5436E+06, 2.5581E+06,
     +  2.5727E+06, 2.5873E+06, 2.6020E+06, 2.6168E+06, 2.6316E+06,
     + 2.6465E+06, 2.6615E+06, 2.6765E+06, 2.6917E+06, 2.7068E+06,
     +  2.7221E+06, 2.7374E+06, 2.7528E+06, 2.7682E+06, 2.7837E+06,
     + 2.7993E+06, 2.8150E+06, 2.8307E+06, 2.8465E+06, 2.8624E+06,
     +  2.8783E+06, 2.8943E+06, 2.9104E+06, 2.9265E+06, 2.9427E+06,
     + 2.9590E+06, 2.9754E+06, 2.9918E+06, 3.0083E+06, 3.0249E+06,
     +  3.0415E+06, 3.0583E+06, 3.0750E+06, 3.0919E+06, 3.1089E+06,
     + 3.1259E+06, 3.1429E+06, 3.1601E+06, 3.1773E+06, 3.1946E+06,
     +  3.2120E+06, 3.2295E+06, 3.2470E+06, 3.2646E+06, 3.2823E+06,
     + 3.3001E+06, 3.3179E+06, 3.3358E+06, 3.3538E+06, 3.3719E+06,
     +  3.3900E+06, 3.4082E+06, 3.4265E+06, 3.4449E+06, 3.4634E+06,
     + 3.4819E+06, 3.5005E+06, 3.5192E+06, 3.5380E+06, 3.5568E+06,
     +  3.5757E+06, 3.5947E+06, 3.6138E+06, 3.6330E+06, 3.6522E+06,
     + 3.6716E+06, 3.6910E+06, 3.7105E+06, 3.7300E+06, 3.7497E+06,
     +  3.7694E+06, 3.7892E+06, 3.8091E+06, 3.8291E+06, 3.8492E+06,
     + 3.8693E+06, 3.8895E+06, 3.9099E+06, 3.9303E+06, 3.9507E+06,
     +  3.9713E+06, 3.9920E+06, 4.0127E+06, 4.0335E+06, 4.0544E+06,
     + 4.0754E+06, 4.0965E+06, 4.1177E+06, 4.1389E+06, 4.1603E+06,
     +  4.1817E+06, 4.2032E+06, 4.2248E+06, 4.2465E+06, 4.2683E+06,
     + 4.2901E+06, 4.3121E+06, 4.3341E+06, 4.3562E+06, 4.3785E+06,
     +  4.4008E+06, 4.4232E+06, 4.4457E+06, 4.4682E+06, 4.4909E+06,
     + 4.5137E+06, 4.5365E+06, 4.5595E+06, 4.5825E+06, 4.6056E+06,
     +  4.6288E+06, 4.6522E+06, 4.6756E+06, 4.6991E+06, 4.7226E+06,
     + 4.7463E+06, 4.7701E+06, 4.7940E+06, 4.8180E+06, 4.8420E+06,
     +  4.8662E+06, 4.8904E+06, 4.9148E+06, 4.9392E+06, 4.9638E+06,
     + 4.9884E+06, 5.0131E+06, 5.0380E+06, 5.0629E+06, 5.0879E+06,
     +  5.1130E+06, 5.1382E+06, 5.1636E+06, 5.1890E+06, 5.2145E+06,
     + 5.2401E+06, 5.2658E+06, 5.2916E+06, 5.3175E+06, 5.3436E+06,
     +  5.3697E+06, 5.3959E+06, 5.4222E+06, 5.4486E+06, 5.4751E+06,
     + 5.5018E+06, 5.5285E+06, 5.5553E+06, 5.5822E+06, 5.6093E+06,
     +  5.6364E+06, 5.6636E+06, 5.6910E+06, 5.7184E+06, 5.7460E+06,
     + 5.7736E+06, 5.8014E+06, 5.8293E+06, 5.8572E+06, 5.8853E+06,
     +  5.9135E+06, 5.9418E+06, 5.9702E+06, 5.9987E+06, 6.0273E+06,
     + 6.0560E+06, 6.0848E+06, 6.1138E+06, 6.1428E+06, 6.1720E+06,
     +  6.2012E+06, 6.2306E+06, 6.2601E+06, 6.2897E+06, 6.3194E+06,
     + 6.3492E+06, 6.3791E+06, 6.4092E+06, 6.4393E+06, 6.4696E+06,
     +  6.4999E+06, 6.5304E+06, 6.5610E+06, 6.5917E+06, 6.6226E+06,
     + 6.6535E+06, 6.6845E+06, 6.7157E+06, 6.7470E+06, 6.7784E+06,
     +  6.8099E+06, 6.8415E+06, 6.8733E+06, 6.9051E+06, 6.9371E+06,
     + 6.9692E+06, 7.0014E+06, 7.0337E+06, 7.0662E+06, 7.0987E+06,
     +  7.1314E+06, 7.1642E+06, 7.1971E+06, 7.2301E+06, 7.2633E+06,
     + 7.2966E+06, 7.3300E+06, 7.3635E+06, 7.3971E+06, 7.4309E+06,
     +  7.4648E+06, 7.4988E+06, 7.5329E+06, 7.5671E+06, 7.6015E+06,
     + 7.6360E+06, 7.6706E+06, 7.7053E+06, 7.7402E+06, 7.7752E+06,
     +  7.8103E+06, 7.8456E+06, 7.8809E+06, 7.9164E+06, 7.9520E+06,
     + 7.9878E+06, 8.0236E+06, 8.0596E+06, 8.0957E+06, 8.1320E+06,
     +  8.1684E+06, 8.2049E+06, 8.2415E+06, 8.2783E+06, 8.3152E+06,
     + 8.3522E+06, 8.3894E+06, 8.4266E+06, 8.4641E+06, 8.5016E+06,
     +  8.5393E+06, 8.5771E+06, 8.6150E+06, 8.6531E+06, 8.6913E+06,
     + 8.7296E+06, 8.7681E+06, 8.8067E+06, 8.8454E+06, 8.8843E+06,
     +  8.9233E+06, 8.9625E+06, 9.0017E+06, 9.0412E+06, 9.0807E+06,
     + 9.1204E+06, 9.1602E+06, 9.2002E+06, 9.2403E+06, 9.2805E+06,
     +  9.3209E+06, 9.3614E+06, 9.4020E+06, 9.4428E+06, 9.4837E+06,
     + 9.5248E+06, 9.5660E+06, 9.6074E+06, 9.6489E+06, 9.6905E+06,
     +  9.7323E+06, 9.7742E+06, 9.8162E+06, 9.8584E+06, 9.9008E+06,
     + 9.9433E+06, 9.9859E+06, 1.0029E+07, 1.0072E+07, 1.0115E+07,
     +  1.0158E+07, 1.0201E+07, 1.0245E+07, 1.0288E+07, 1.0332E+07,
     + 1.0376E+07, 1.0420E+07, 1.0464E+07, 1.0509E+07, 1.0553E+07,
     +  1.0598E+07, 1.0643E+07, 1.0688E+07, 1.0733E+07, 1.0778E+07,
     + 1.0824E+07, 1.0869E+07, 1.0915E+07, 1.0961E+07, 1.1007E+07,
     +  1.1053E+07, 1.1099E+07, 1.1146E+07, 1.1192E+07, 1.1239E+07,
     + 1.1286E+07, 1.1333E+07, 1.1381E+07, 1.1428E+07, 1.1476E+07,
     +  1.1523E+07, 1.1571E+07, 1.1619E+07, 1.1667E+07, 1.1716E+07,
     + 1.1764E+07, 1.1813E+07, 1.1862E+07, 1.1911E+07, 1.1960E+07,
     +  1.2009E+07, 1.2059E+07, 1.2108E+07, 1.2158E+07, 1.2208E+07,
     + 1.2258E+07, 1.2308E+07, 1.2359E+07, 1.2409E+07, 1.2460E+07,
     +  1.2511E+07, 1.2562E+07, 1.2613E+07, 1.2665E+07, 1.2716E+07,
     + 1.2768E+07, 1.2820E+07, 1.2872E+07, 1.2924E+07, 1.2977E+07,
     +  1.3029E+07, 1.3082E+07, 1.3135E+07, 1.3188E+07, 1.3241E+07,
     + 1.3295E+07, 1.3348E+07, 1.3402E+07, 1.3456E+07, 1.3510E+07,
     +  1.3565E+07, 1.3619E+07, 1.3674E+07, 1.3728E+07, 1.3783E+07,
     + 1.3838E+07, 1.3894E+07, 1.3949E+07, 1.4005E+07, 1.4061E+07,
     +  1.4117E+07, 1.4173E+07, 1.4229E+07, 1.4286E+07, 1.4343E+07,
     + 1.4399E+07, 1.4457E+07, 1.4514E+07, 1.4571E+07, 1.4629E+07,
     +  1.4687E+07, 1.4745E+07, 1.4803E+07, 1.4861E+07, 1.4919E+07,
     + 1.4978E+07, 1.5037E+07, 1.5096E+07, 1.5155E+07, 1.5215E+07,
     +  1.5274E+07, 1.5334E+07, 1.5394E+07, 1.5454E+07, 1.5515E+07,
     + 1.5575E+07, 1.5636E+07, 1.5697E+07, 1.5758E+07, 1.5819E+07,
     +  1.5880E+07, 1.5942E+07, 1.6004E+07, 1.6066E+07, 1.6128E+07,
     + 1.6191E+07, 1.6253E+07, 1.6316E+07, 1.6379E+07, 1.6442E+07,
     +  1.6505E+07, 1.6569E+07, 1.6633E+07, 1.6697E+07, 1.6761E+07,
     + 1.6825E+07, 1.6890E+07, 1.6954E+07, 1.7019E+07, 1.7084E+07,
     +  1.7150E+07, 1.7215E+07, 1.7281E+07, 1.7347E+07, 1.7413E+07,
     + 1.7479E+07, 1.7545E+07, 1.7612E+07, 1.7679E+07, 1.7746E+07,
     +  1.7813E+07, 1.7881E+07, 1.7948E+07, 1.8016E+07, 1.8084E+07,
     + 1.8153E+07, 1.8221E+07, 1.8290E+07, 1.8359E+07, 1.8428E+07,
     +  1.8497E+07, 1.8567E+07, 1.8636E+07, 1.8706E+07, 1.8776E+07,
     + 1.8847E+07, 1.8917E+07, 1.8988E+07, 1.9059E+07, 1.9130E+07,
     +  1.9201E+07, 1.9273E+07, 1.9345E+07, 1.9417E+07, 1.9489E+07,
     + 1.9561E+07, 1.9634E+07, 1.9707E+07, 1.9780E+07, 1.9853E+07,
     +  1.9927E+07, 2.0000E+07, 2.0074E+07, 2.0149E+07, 2.0223E+07,
     + 2.0297E+07, 2.0372E+07, 2.0447E+07, 2.0522E+07, 2.0598E+07,
     +  2.0673E+07, 2.0749E+07, 2.0825E+07, 2.0902E+07, 2.0978E+07,
     + 2.1055E+07, 2.1132E+07, 2.1209E+07, 2.1286E+07, 2.1364E+07,
     +  2.1442E+07, 2.1520E+07, 2.1598E+07, 2.1677E+07, 2.1756E+07,
     + 2.1835E+07, 2.1914E+07, 2.1993E+07, 2.2073E+07, 2.2153E+07,
     +  2.2233E+07, 2.2313E+07, 2.2394E+07, 2.2474E+07, 2.2555E+07,
     + 2.2637E+07, 2.2718E+07, 2.2800E+07, 2.2882E+07, 2.2964E+07,
     +  2.3046E+07, 2.3129E+07, 2.3212E+07, 2.3295E+07, 2.3378E+07,
     + 2.3462E+07, 2.3546E+07, 2.3630E+07, 2.3714E+07, 2.3798E+07,
     +  2.3883E+07, 2.3968E+07, 2.4053E+07, 2.4139E+07, 2.4225E+07,
     + 2.4311E+07, 2.4397E+07, 2.4483E+07, 2.4570E+07, 2.4657E+07,
     +  2.4744E+07, 2.4831E+07, 2.4919E+07, 2.5007E+07, 2.5095E+07,
     + 2.5183E+07, 2.5272E+07, 2.5361E+07, 2.5450E+07, 2.5539E+07,
     +  2.5629E+07, 2.5718E+07, 2.5809E+07, 2.5899E+07, 2.5989E+07,
     + 2.6080E+07, 2.6171E+07, 2.6263E+07, 2.6354E+07, 2.6446E+07,
     +  2.6538E+07, 2.6631E+07, 2.6723E+07, 2.6816E+07, 2.6909E+07,
     + 2.7002E+07, 2.7096E+07, 2.7190E+07, 2.7284E+07, 2.7378E+07,
     +  2.7473E+07, 2.7568E+07, 2.7663E+07, 2.7758E+07, 2.7854E+07,
     + 2.7950E+07, 2.8046E+07, 2.8142E+07, 2.8239E+07, 2.8336E+07,
     +  2.8433E+07, 2.8531E+07, 2.8629E+07, 2.8727E+07, 2.8825E+07,
     + 2.8923E+07, 2.9022E+07, 2.9121E+07, 2.9220E+07, 2.9320E+07,
     +  2.9420E+07, 2.9520E+07, 2.9620E+07, 2.9721E+07, 2.9822E+07,
     + 2.9923E+07, 3.0025E+07, 3.0126E+07, 3.0228E+07, 3.0331E+07,
     +  3.0433E+07, 3.0536E+07, 3.0639E+07, 3.0743E+07, 3.0846E+07,
     + 3.0950E+07, 3.1054E+07, 3.1159E+07, 3.1264E+07, 3.1369E+07,
     +  3.1474E+07, 3.1580E+07, 3.1685E+07, 3.1792E+07, 3.1898E+07,
     + 3.2005E+07, 3.2112E+07, 3.2219E+07, 3.2326E+07, 3.2434E+07,
     +  3.2542E+07, 3.2651E+07, 3.2759E+07, 3.2868E+07, 3.2978E+07,
     + 3.3087E+07, 3.3197E+07, 3.3307E+07, 3.3418E+07, 3.3528E+07,
     +  3.3639E+07, 3.3750E+07, 3.3862E+07, 3.3974E+07, 3.4086E+07,
     + 3.4198E+07, 3.4311E+07, 3.4424E+07, 3.4537E+07, 3.4651E+07,
     +  3.4765E+07, 3.4879E+07, 3.4993E+07, 3.5108E+07, 3.5223E+07,
     + 3.5339E+07, 3.5454E+07, 3.5570E+07, 3.5686E+07, 3.5803E+07,
     +  3.5920E+07, 3.6037E+07, 3.6154E+07, 3.6272E+07, 3.6390E+07,
     + 3.6509E+07, 3.6627E+07, 3.6746E+07, 3.6865E+07, 3.6985E+07,
     +  3.7105E+07, 3.7225E+07, 3.7346E+07, 3.7466E+07, 3.7587E+07,
     + 3.7709E+07, 3.7830E+07, 3.7952E+07, 3.8075E+07, 3.8197E+07,
     +  3.8320E+07, 3.8444E+07, 3.8567E+07, 3.8691E+07, 3.8815E+07,
     + 3.8940E+07, 3.9065E+07, 3.9190E+07, 3.9315E+07, 3.9441E+07,
     +  3.9567E+07, 3.9693E+07, 3.9820E+07, 3.9947E+07, 4.0075E+07,
     + 4.0202E+07, 4.0330E+07, 4.0458E+07, 4.0587E+07, 4.0716E+07,
     +  4.0845E+07, 4.0975E+07, 4.1105E+07, 4.1235E+07, 4.1366E+07,
     + 4.1497E+07, 4.1628E+07, 4.1759E+07, 4.1891E+07, 4.2023E+07,
     +  4.2156E+07, 4.2289E+07, 4.2422E+07, 4.2555E+07, 4.2689E+07,
     + 4.2823E+07, 4.2958E+07, 4.3093E+07, 4.3228E+07, 4.3363E+07,
     +  4.3499E+07, 4.3635E+07, 4.3772E+07, 4.3909E+07, 4.4046E+07,
     + 4.4184E+07, 4.4321E+07, 4.4460E+07, 4.4598E+07, 4.4737E+07,
     +  4.4876E+07, 4.5016E+07, 4.5156E+07, 4.5296E+07, 4.5436E+07,
     + 4.5577E+07, 4.5719E+07, 4.5860E+07, 4.6002E+07, 4.6144E+07,
     +  4.6287E+07, 4.6430E+07, 4.6573E+07, 4.6717E+07, 4.6861E+07,
     + 4.7005E+07, 4.7150E+07, 4.7295E+07, 4.7441E+07, 4.7586E+07,
     +  4.7733E+07, 4.7879E+07, 4.8026E+07, 4.8173E+07, 4.8321E+07,
     + 4.8469E+07, 4.8617E+07, 4.8765E+07, 4.8914E+07, 4.9064E+07,
     +  4.9213E+07, 4.9363E+07, 4.9514E+07, 4.9665E+07, 4.9816E+07,
     + 4.9967E+07, 5.0119E+07, 5.0271E+07, 5.0424E+07, 5.0577E+07,
     +  5.0730E+07, 5.0884E+07, 5.1038E+07, 5.1192E+07, 5.1347E+07,
     + 5.1502E+07, 5.1658E+07, 5.1813E+07, 5.1970E+07, 5.2126E+07,
     +  5.2283E+07, 5.2441E+07, 5.2598E+07, 5.2756E+07, 5.2915E+07,
     + 5.3074E+07, 5.3233E+07, 5.3393E+07, 5.3553E+07, 5.3713E+07,
     +  5.3874E+07, 5.4035E+07, 5.4196E+07, 5.4358E+07, 5.4520E+07,
     + 5.4683E+07, 5.4846E+07, 5.5009E+07, 5.5173E+07, 5.5337E+07,
     +  5.5502E+07, 5.5667E+07, 5.5832E+07, 5.5998E+07, 5.6164E+07,
     + 5.6330E+07, 5.6497E+07, 5.6664E+07, 5.6832E+07, 5.7000E+07,
     +  5.7168E+07, 5.7337E+07, 5.7506E+07, 5.7676E+07, 5.7846E+07,
     + 5.8016E+07, 5.8187E+07, 5.8358E+07, 5.8530E+07, 5.8702E+07,
     +  5.8874E+07, 5.9047E+07, 5.9220E+07, 5.9393E+07, 5.9567E+07,
     + 5.9742E+07, 5.9916E+07, 6.0092E+07, 6.0267E+07, 6.0443E+07,
     +  6.0619E+07, 6.0796E+07, 6.0973E+07, 6.1151E+07, 6.1329E+07,
     + 6.1507E+07, 6.1686E+07, 6.1865E+07, 6.2045E+07, 6.2225E+07,
     +  6.2405E+07, 6.2586E+07, 6.2767E+07, 6.2949E+07, 6.3131E+07,
     + 6.3314E+07, 6.3497E+07, 6.3680E+07, 6.3864E+07, 6.4048E+07,
     +  6.4232E+07, 6.4417E+07, 6.4603E+07, 6.4788E+07, 6.4975E+07,
     + 6.5161E+07, 6.5348E+07, 6.5536E+07, 6.5724E+07, 6.5912E+07,
     +  6.6101E+07, 6.6290E+07, 6.6480E+07, 6.6670E+07, 6.6860E+07,
     + 6.7051E+07, 6.7243E+07, 6.7434E+07, 6.7627E+07, 6.7819E+07,
     +  6.8012E+07, 6.8206E+07, 6.8400E+07, 6.8594E+07, 6.8789E+07,
     + 6.8984E+07, 6.9180E+07, 6.9376E+07, 6.9572E+07, 6.9769E+07,
     +  6.9967E+07, 7.0165E+07, 7.0363E+07, 7.0562E+07, 7.0761E+07,
     + 7.0960E+07, 7.1160E+07, 7.1361E+07, 7.1562E+07, 7.1763E+07,
     +  7.1965E+07, 7.2167E+07, 7.2370E+07, 7.2573E+07, 7.2777E+07,
     + 7.2981E+07, 7.3185E+07, 7.3390E+07, 7.3596E+07, 7.3802E+07,
     +  7.4008E+07, 7.4215E+07, 7.4422E+07, 7.4630E+07, 7.4838E+07,
     + 7.5046E+07, 7.5255E+07, 7.5465E+07, 7.5675E+07, 7.5885E+07,
     +  7.6096E+07, 7.6307E+07, 7.6519E+07, 7.6731E+07, 7.6944E+07,
     + 7.7157E+07, 7.7371E+07, 7.7585E+07, 7.7800E+07, 7.8015E+07,
     +  7.8230E+07, 7.8446E+07, 7.8663E+07, 7.8880E+07, 7.9097E+07,
     + 7.9315E+07, 7.9533E+07, 7.9752E+07, 7.9971E+07, 8.0191E+07,
     +  8.0411E+07, 8.0632E+07, 8.0853E+07, 8.1075E+07, 8.1297E+07,
     + 8.1520E+07, 8.1743E+07, 8.1966E+07, 8.2190E+07, 8.2415E+07,
     +  8.2640E+07, 8.2865E+07, 8.3091E+07, 8.3318E+07, 8.3545E+07,
     + 8.3772E+07, 8.4000E+07, 8.4229E+07, 8.4458E+07, 8.4687E+07,
     +  8.4917E+07, 8.5147E+07, 8.5378E+07, 8.5609E+07, 8.5841E+07,
     + 8.6074E+07, 8.6306E+07, 8.6540E+07, 8.6773E+07, 8.7008E+07,
     +  8.7243E+07, 8.7478E+07, 8.7714E+07, 8.7950E+07, 8.8187E+07,
     + 8.8424E+07, 8.8662E+07, 8.8900E+07, 8.9139E+07, 8.9378E+07,
     +  8.9618E+07, 8.9859E+07, 9.0099E+07, 9.0341E+07, 9.0583E+07,
     + 9.0825E+07, 9.1068E+07, 9.1311E+07, 9.1555E+07, 9.1800E+07,
     +  9.2044E+07, 9.2290E+07, 9.2536E+07, 9.2782E+07, 9.3029E+07,
     + 9.3277E+07, 9.3525E+07, 9.3773E+07, 9.4022E+07, 9.4272E+07,
     +  9.4522E+07, 9.4773E+07, 9.5024E+07, 9.5275E+07, 9.5528E+07,
     + 9.5780E+07, 9.6034E+07, 9.6287E+07, 9.6542E+07, 9.6796E+07,
     +  9.7052E+07, 9.7308E+07, 9.7564E+07, 9.7821E+07, 9.8078E+07,
     + 9.8336E+07, 9.8595E+07, 9.8854E+07, 9.9114E+07, 9.9374E+07,
     +  9.9634E+07, 9.9896E+07, 1.0016E+08, 1.0042E+08, 1.0068E+08,
     + 1.0095E+08, 1.0121E+08, 1.0147E+08, 1.0174E+08, 1.0200E+08,
     +  1.0227E+08, 1.0254E+08, 1.0280E+08, 1.0307E+08, 1.0334E+08,
     + 1.0361E+08, 1.0388E+08, 1.0415E+08, 1.0442E+08, 1.0469E+08,
     +  1.0496E+08, 1.0523E+08, 1.0551E+08, 1.0578E+08, 1.0605E+08,
     + 1.0633E+08, 1.0660E+08, 1.0688E+08, 1.0715E+08, 1.0743E+08,
     +  1.0771E+08, 1.0799E+08, 1.0826E+08, 1.0854E+08, 1.0882E+08,
     + 1.0910E+08, 1.0938E+08, 1.0966E+08, 1.0995E+08, 1.1023E+08,
     +  1.1051E+08, 1.1079E+08, 1.1108E+08, 1.1136E+08, 1.1165E+08,
     + 1.1193E+08, 1.1222E+08, 1.1251E+08, 1.1279E+08, 1.1308E+08,
     +  1.1337E+08, 1.1366E+08, 1.1395E+08, 1.1424E+08, 1.1453E+08,
     + 1.1482E+08, 1.1511E+08, 1.1541E+08, 1.1570E+08, 1.1599E+08,
     +  1.1629E+08, 1.1658E+08, 1.1688E+08, 1.1717E+08, 1.1747E+08,
     + 1.1777E+08, 1.1807E+08, 1.1836E+08, 1.1866E+08, 1.1896E+08,
     +  1.1926E+08, 1.1956E+08, 1.1987E+08, 1.2017E+08, 1.2047E+08,
     + 1.2077E+08, 1.2108E+08, 1.2138E+08, 1.2169E+08, 1.2199E+08,
     +  1.2230E+08, 1.2261E+08, 1.2291E+08, 1.2322E+08, 1.2353E+08,
     + 1.2384E+08, 1.2415E+08, 1.2446E+08, 1.2477E+08, 1.2508E+08,
     +  1.2539E+08, 1.2571E+08, 1.2602E+08, 1.2633E+08, 1.2665E+08,
     + 1.2696E+08, 1.2728E+08, 1.2760E+08, 1.2791E+08, 1.2823E+08,
     +  1.2855E+08, 1.2887E+08, 1.2919E+08, 1.2951E+08, 1.2983E+08,
     + 1.3015E+08, 1.3048E+08, 1.3080E+08, 1.3112E+08, 1.3145E+08,
     +  1.3177E+08, 1.3210E+08, 1.3242E+08, 1.3275E+08, 1.3308E+08,
     + 1.3340E+08, 1.3373E+08, 1.3406E+08, 1.3439E+08, 1.3472E+08,
     +  1.3505E+08, 1.3538E+08, 1.3572E+08, 1.3605E+08, 1.3638E+08,
     + 1.3672E+08, 1.3705E+08, 1.3739E+08, 1.3772E+08, 1.3806E+08,
     +  1.3840E+08, 1.3874E+08, 1.3908E+08, 1.3942E+08, 1.3976E+08,
     + 1.4010E+08, 1.4044E+08, 1.4078E+08, 1.4112E+08, 1.4147E+08,
     +  1.4181E+08, 1.4216E+08, 1.4250E+08, 1.4285E+08, 1.4319E+08,
     + 1.4354E+08, 1.4389E+08, 1.4424E+08, 1.4459E+08, 1.4494E+08,
     +  1.4529E+08, 1.4564E+08, 1.4599E+08, 1.4634E+08, 1.4670E+08,
     + 1.4705E+08, 1.4741E+08, 1.4776E+08, 1.4812E+08, 1.4847E+08,
     +  1.4883E+08, 1.4919E+08, 1.4955E+08, 1.4991E+08, 1.5027E+08,
     + 1.5063E+08, 1.5099E+08, 1.5135E+08, 1.5171E+08, 1.5208E+08,
     +  1.5244E+08, 1.5281E+08, 1.5317E+08, 1.5354E+08, 1.5391E+08,
     + 1.5427E+08, 1.5464E+08, 1.5501E+08, 1.5538E+08, 1.5575E+08,
     +  1.5612E+08, 1.5649E+08, 1.5687E+08, 1.5724E+08, 1.5761E+08,
     + 1.5799E+08, 1.5836E+08, 1.5874E+08, 1.5912E+08, 1.5949E+08,
     +  1.5987E+08, 1.6025E+08, 1.6063E+08, 1.6101E+08, 1.6139E+08,
     + 1.6177E+08, 1.6215E+08, 1.6254E+08, 1.6292E+08, 1.6331E+08,
     +  1.6369E+08, 1.6408E+08, 1.6446E+08, 1.6485E+08, 1.6524E+08,
     + 1.6563E+08, 1.6602E+08, 1.6641E+08, 1.6680E+08, 1.6719E+08,
     +  1.6758E+08, 1.6798E+08, 1.6837E+08, 1.6876E+08, 1.6916E+08,
     + 1.6955E+08, 1.6995E+08, 1.7035E+08, 1.7075E+08, 1.7115E+08,
     +  1.7155E+08, 1.7195E+08, 1.7235E+08, 1.7275E+08, 1.7315E+08,
     + 1.7355E+08, 1.7396E+08, 1.7436E+08, 1.7477E+08, 1.7518E+08,
     +  1.7558E+08, 1.7599E+08, 1.7640E+08, 1.7681E+08, 1.7722E+08,
     + 1.7763E+08, 1.7804E+08, 1.7845E+08, 1.7887E+08, 1.7928E+08,
     +  1.7969E+08, 1.8011E+08, 1.8052E+08, 1.8094E+08, 1.8136E+08,
     + 1.8178E+08, 1.8220E+08, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT(12,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       888
      data (QofT(13,J),J=1,2408)/ 9.1813E-01, 2.3891E+00, 4.2648E+00,
     +  6.4726E+00, 8.9684E+00, 1.1722E+01, 1.4712E+01, 1.7920E+01,
     + 2.1333E+01, 2.4939E+01, 2.8728E+01, 3.2691E+01, 3.6822E+01,
     +  4.1114E+01, 4.5561E+01, 5.0158E+01, 5.4900E+01, 5.9782E+01,
     + 6.4802E+01, 6.9955E+01, 8.0648E+01, 9.1836E+01, 1.0350E+02,
     +  1.1562E+02, 1.2818E+02, 1.4116E+02, 1.5456E+02, 1.6836E+02,
     + 1.8254E+02, 1.9711E+02, 2.1204E+02, 2.2733E+02, 2.4298E+02,
     +  2.5896E+02, 2.7529E+02, 2.9194E+02, 3.0892E+02, 3.2622E+02,
     + 3.4383E+02, 3.6175E+02, 3.7996E+02, 3.9848E+02, 4.1729E+02,
     +  4.3638E+02, 4.5576E+02, 4.7542E+02, 4.9535E+02, 5.1556E+02,
     + 5.3604E+02, 5.5678E+02, 5.7779E+02, 5.9905E+02, 6.2057E+02,
     +  6.4234E+02, 6.6437E+02, 6.8664E+02, 7.0916E+02, 7.3192E+02,
     + 7.5492E+02, 7.7816E+02, 8.0164E+02, 8.2536E+02, 8.4931E+02,
     +  8.7349E+02, 8.9790E+02, 9.2254E+02, 9.4740E+02, 9.7249E+02,
     + 9.9781E+02, 1.0234E+03, 1.0491E+03, 1.0751E+03, 1.1013E+03,
     +  1.1277E+03, 1.1544E+03, 1.1812E+03, 1.2083E+03, 1.2356E+03,
     + 1.2631E+03, 1.2909E+03, 1.3188E+03, 1.3470E+03, 1.3754E+03,
     +  1.4040E+03, 1.4328E+03, 1.4619E+03, 1.4911E+03, 1.5206E+03,
     + 1.5503E+03, 1.5802E+03, 1.6103E+03, 1.6407E+03, 1.6712E+03,
     +  1.7020E+03, 1.7330E+03, 1.7643E+03, 1.7957E+03, 1.8274E+03,
     + 1.8593E+03, 1.8915E+03, 1.9238E+03, 1.9564E+03, 1.9892E+03,
     +  2.0223E+03, 2.0556E+03, 2.0891E+03, 2.1229E+03, 2.1569E+03,
     + 2.1911E+03, 2.2256E+03, 2.2603E+03, 2.2953E+03, 2.3305E+03,
     +  2.3659E+03, 2.4016E+03, 2.4376E+03, 2.4738E+03, 2.5103E+03,
     + 2.5470E+03, 2.5839E+03, 2.6212E+03, 2.6587E+03, 2.6964E+03,
     +  2.7344E+03, 2.7727E+03, 2.8113E+03, 2.8501E+03, 2.8892E+03,
     + 2.9286E+03, 2.9683E+03, 3.0082E+03, 3.0484E+03, 3.0890E+03,
     +  3.1298E+03, 3.1708E+03, 3.2122E+03, 3.2539E+03, 3.2959E+03,
     + 3.3382E+03, 3.3807E+03, 3.4236E+03, 3.4668E+03, 3.5103E+03,
     +  3.5541E+03, 3.5982E+03, 3.6426E+03, 3.6874E+03, 3.7325E+03,
     + 3.7779E+03, 3.8236E+03, 3.8697E+03, 3.9161E+03, 3.9628E+03,
     +  4.0098E+03, 4.0573E+03, 4.1050E+03, 4.1531E+03, 4.2015E+03,
     + 4.2503E+03, 4.2995E+03, 4.3489E+03, 4.3988E+03, 4.4490E+03,
     +  4.4996E+03, 4.5506E+03, 4.6019E+03, 4.6536E+03, 4.7057E+03,
     + 4.7581E+03, 4.8109E+03, 4.8642E+03, 4.9178E+03, 4.9718E+03,
     +  5.0261E+03, 5.0809E+03, 5.1361E+03, 5.1917E+03, 5.2477E+03,
     + 5.3041E+03, 5.3609E+03, 5.4181E+03, 5.4757E+03, 5.5338E+03,
     +  5.5922E+03, 5.6511E+03, 5.7104E+03, 5.7702E+03, 5.8304E+03,
     + 5.8910E+03, 5.9521E+03, 6.0136E+03, 6.0756E+03, 6.1380E+03,
     +  6.2008E+03, 6.2641E+03, 6.3279E+03, 6.3921E+03, 6.4568E+03,
     + 6.5220E+03, 6.5877E+03, 6.6538E+03, 6.7203E+03, 6.7874E+03,
     +  6.8550E+03, 6.9230E+03, 6.9915E+03, 7.0605E+03, 7.1301E+03,
     + 7.2001E+03, 7.2706E+03, 7.3416E+03, 7.4132E+03, 7.4852E+03,
     +  7.5578E+03, 7.6308E+03, 7.7044E+03, 7.7785E+03, 7.8533E+03,
     + 7.9284E+03, 8.0042E+03, 8.0804E+03, 8.1572E+03, 8.2345E+03,
     +  8.3124E+03, 8.3909E+03, 8.4699E+03, 8.5494E+03, 8.6296E+03,
     + 8.7102E+03, 8.7915E+03, 8.8733E+03, 8.9558E+03, 9.0387E+03,
     +  9.1223E+03, 9.2064E+03, 9.2912E+03, 9.3765E+03, 9.4624E+03,
     + 9.5490E+03, 9.6361E+03, 9.7238E+03, 9.8121E+03, 9.9011E+03,
     +  9.9907E+03, 1.0081E+04, 1.0172E+04, 1.0263E+04, 1.0355E+04,
     + 1.0448E+04, 1.0541E+04, 1.0635E+04, 1.0730E+04, 1.0825E+04,
     +  1.0921E+04, 1.1018E+04, 1.1115E+04, 1.1213E+04, 1.1311E+04,
     + 1.1411E+04, 1.1511E+04, 1.1611E+04, 1.1713E+04, 1.1815E+04,
     +  1.1917E+04, 1.2021E+04, 1.2125E+04, 1.2229E+04, 1.2335E+04,
     + 1.2441E+04, 1.2548E+04, 1.2656E+04, 1.2764E+04, 1.2873E+04,
     +  1.2983E+04, 1.3093E+04, 1.3204E+04, 1.3316E+04, 1.3429E+04,
     + 1.3543E+04, 1.3657E+04, 1.3772E+04, 1.3887E+04, 1.4004E+04,
     +  1.4121E+04, 1.4239E+04, 1.4358E+04, 1.4478E+04, 1.4598E+04,
     + 1.4719E+04, 1.4841E+04, 1.4964E+04, 1.5087E+04, 1.5212E+04,
     +  1.5337E+04, 1.5463E+04, 1.5589E+04, 1.5717E+04, 1.5845E+04,
     + 1.5975E+04, 1.6105E+04, 1.6236E+04, 1.6367E+04, 1.6500E+04,
     +  1.6633E+04, 1.6768E+04, 1.6903E+04, 1.7039E+04, 1.7176E+04,
     + 1.7313E+04, 1.7452E+04, 1.7591E+04, 1.7732E+04, 1.7873E+04,
     +  1.8015E+04, 1.8158E+04, 1.8302E+04, 1.8447E+04, 1.8592E+04,
     + 1.8739E+04, 1.8887E+04, 1.9035E+04, 1.9185E+04, 1.9335E+04,
     +  1.9486E+04, 1.9638E+04, 1.9791E+04, 1.9945E+04, 2.0101E+04,
     + 2.0256E+04, 2.0414E+04, 2.0571E+04, 2.0730E+04, 2.0890E+04,
     +  2.1051E+04, 2.1213E+04, 2.1375E+04, 2.1539E+04, 2.1704E+04,
     + 2.1870E+04, 2.2037E+04, 2.2205E+04, 2.2373E+04, 2.2543E+04,
     +  2.2714E+04, 2.2886E+04, 2.3059E+04, 2.3233E+04, 2.3408E+04,
     + 2.3584E+04, 2.3761E+04, 2.3939E+04, 2.4118E+04, 2.4298E+04,
     +  2.4480E+04, 2.4662E+04, 2.4846E+04, 2.5030E+04, 2.5216E+04,
     + 2.5403E+04, 2.5591E+04, 2.5780E+04, 2.5970E+04, 2.6161E+04,
     +  2.6353E+04, 2.6547E+04, 2.6741E+04, 2.6937E+04, 2.7134E+04,
     + 2.7332E+04, 2.7531E+04, 2.7731E+04, 2.7932E+04, 2.8135E+04,
     +  2.8339E+04, 2.8544E+04, 2.8750E+04, 2.8957E+04, 2.9166E+04,
     + 2.9375E+04, 2.9586E+04, 2.9798E+04, 3.0011E+04, 3.0226E+04,
     +  3.0442E+04, 3.0659E+04, 3.0877E+04, 3.1096E+04, 3.1317E+04,
     + 3.1539E+04, 3.1762E+04, 3.1986E+04, 3.2212E+04, 3.2439E+04,
     +  3.2667E+04, 3.2897E+04, 3.3127E+04, 3.3359E+04, 3.3593E+04,
     + 3.3827E+04, 3.4063E+04, 3.4301E+04, 3.4539E+04, 3.4779E+04,
     +  3.5020E+04, 3.5263E+04, 3.5507E+04, 3.5752E+04, 3.5998E+04,
     + 3.6246E+04, 3.6496E+04, 3.6746E+04, 3.6998E+04, 3.7251E+04,
     +  3.7506E+04, 3.7762E+04, 3.8020E+04, 3.8279E+04, 3.8539E+04,
     + 3.8801E+04, 3.9064E+04, 3.9329E+04, 3.9595E+04, 3.9862E+04,
     +  4.0131E+04, 4.0401E+04, 4.0673E+04, 4.0946E+04, 4.1221E+04,
     + 4.1497E+04, 4.1775E+04, 4.2054E+04, 4.2334E+04, 4.2616E+04,
     +  4.2900E+04, 4.3185E+04, 4.3471E+04, 4.3759E+04, 4.4049E+04,
     + 4.4340E+04, 4.4632E+04, 4.4927E+04, 4.5222E+04, 4.5519E+04,
     +  4.5818E+04, 4.6119E+04, 4.6421E+04, 4.6724E+04, 4.7029E+04,
     + 4.7336E+04, 4.7644E+04, 4.7954E+04, 4.8265E+04, 4.8578E+04,
     +  4.8893E+04, 4.9209E+04, 4.9527E+04, 4.9847E+04, 5.0168E+04,
     + 5.0490E+04, 5.0815E+04, 5.1141E+04, 5.1469E+04, 5.1798E+04,
     +  5.2129E+04, 5.2462E+04, 5.2796E+04, 5.3133E+04, 5.3470E+04,
     + 5.3810E+04, 5.4151E+04, 5.4494E+04, 5.4839E+04, 5.5185E+04,
     +  5.5534E+04, 5.5883E+04, 5.6235E+04, 5.6589E+04, 5.6944E+04,
     + 5.7301E+04, 5.7660E+04, 5.8020E+04, 5.8382E+04, 5.8747E+04,
     +  5.9112E+04, 5.9480E+04, 5.9850E+04, 6.0221E+04, 6.0594E+04,
     + 6.0969E+04, 6.1346E+04, 6.1725E+04, 6.2105E+04, 6.2488E+04,
     +  6.2872E+04, 6.3258E+04, 6.3646E+04, 6.4036E+04, 6.4428E+04,
     + 6.4822E+04, 6.5218E+04, 6.5615E+04, 6.6015E+04, 6.6416E+04,
     +  6.6819E+04, 6.7225E+04, 6.7632E+04, 6.8041E+04, 6.8452E+04,
     + 6.8865E+04, 6.9281E+04, 6.9698E+04, 7.0117E+04, 7.0538E+04,
     +  7.0961E+04, 7.1386E+04, 7.1813E+04, 7.2242E+04, 7.2674E+04,
     + 7.3107E+04, 7.3542E+04, 7.3979E+04, 7.4419E+04, 7.4860E+04,
     +  7.5304E+04, 7.5749E+04, 7.6197E+04, 7.6647E+04, 7.7099E+04,
     + 7.7553E+04, 7.8009E+04, 7.8467E+04, 7.8927E+04, 7.9390E+04,
     +  7.9855E+04, 8.0322E+04, 8.0790E+04, 8.1262E+04, 8.1735E+04,
     + 8.2211E+04, 8.2688E+04, 8.3168E+04, 8.3650E+04, 8.4135E+04,
     +  8.4621E+04, 8.5110E+04, 8.5601E+04, 8.6094E+04, 8.6590E+04,
     + 8.7088E+04, 8.7588E+04, 8.8090E+04, 8.8595E+04, 8.9101E+04,
     +  8.9611E+04, 9.0122E+04, 9.0636E+04, 9.1152E+04, 9.1671E+04,
     + 9.2191E+04, 9.2714E+04, 9.3240E+04, 9.3768E+04, 9.4298E+04,
     +  9.4830E+04, 9.5365E+04, 9.5903E+04, 9.6442E+04, 9.6985E+04,
     + 9.7529E+04, 9.8076E+04, 9.8626E+04, 9.9177E+04, 9.9732E+04,
     +  1.0029E+05, 1.0085E+05, 1.0141E+05, 1.0197E+05, 1.0254E+05,
     + 1.0311E+05, 1.0368E+05, 1.0425E+05, 1.0483E+05, 1.0541E+05,
     +  1.0599E+05, 1.0658E+05, 1.0716E+05, 1.0775E+05, 1.0834E+05,
     + 1.0894E+05, 1.0954E+05, 1.1014E+05, 1.1074E+05, 1.1134E+05,
     +  1.1195E+05, 1.1256E+05, 1.1317E+05, 1.1379E+05, 1.1441E+05,
     + 1.1503E+05, 1.1565E+05, 1.1628E+05, 1.1690E+05, 1.1754E+05,
     +  1.1817E+05, 1.1881E+05, 1.1945E+05, 1.2009E+05, 1.2073E+05,
     + 1.2138E+05, 1.2203E+05, 1.2268E+05, 1.2334E+05, 1.2400E+05,
     +  1.2466E+05, 1.2532E+05, 1.2599E+05, 1.2666E+05, 1.2733E+05,
     + 1.2801E+05, 1.2869E+05, 1.2937E+05, 1.3005E+05, 1.3074E+05,
     +  1.3143E+05, 1.3212E+05, 1.3282E+05, 1.3352E+05, 1.3422E+05,
     + 1.3492E+05, 1.3563E+05, 1.3634E+05, 1.3705E+05, 1.3777E+05,
     +  1.3849E+05, 1.3921E+05, 1.3993E+05, 1.4066E+05, 1.4139E+05,
     + 1.4213E+05, 1.4286E+05, 1.4360E+05, 1.4435E+05, 1.4509E+05,
     +  1.4584E+05, 1.4659E+05, 1.4735E+05, 1.4810E+05, 1.4887E+05,
     + 1.4963E+05, 1.5040E+05, 1.5117E+05, 1.5194E+05, 1.5272E+05,
     +  1.5350E+05, 1.5428E+05, 1.5507E+05, 1.5586E+05, 1.5665E+05,
     + 1.5745E+05, 1.5824E+05, 1.5905E+05, 1.5985E+05, 1.6066E+05,
     +  1.6147E+05, 1.6229E+05, 1.6310E+05, 1.6393E+05, 1.6475E+05,
     + 1.6558E+05, 1.6641E+05, 1.6724E+05, 1.6808E+05, 1.6892E+05,
     +  1.6977E+05, 1.7062E+05, 1.7147E+05, 1.7232E+05, 1.7318E+05,
     + 1.7404E+05, 1.7491E+05, 1.7577E+05, 1.7665E+05, 1.7752E+05,
     +  1.7840E+05, 1.7928E+05, 1.8017E+05, 1.8105E+05, 1.8195E+05,
     + 1.8284E+05, 1.8374E+05, 1.8464E+05, 1.8555E+05, 1.8646E+05,
     +  1.8737E+05, 1.8829E+05, 1.8921E+05, 1.9013E+05, 1.9106E+05,
     + 1.9199E+05, 1.9292E+05, 1.9386E+05, 1.9480E+05, 1.9575E+05,
     +  1.9670E+05, 1.9765E+05, 1.9861E+05, 1.9957E+05, 2.0053E+05,
     + 2.0150E+05, 2.0247E+05, 2.0344E+05, 2.0442E+05, 2.0540E+05,
     +  2.0639E+05, 2.0737E+05, 2.0837E+05, 2.0936E+05, 2.1036E+05,
     + 2.1137E+05, 2.1238E+05, 2.1339E+05, 2.1440E+05, 2.1542E+05,
     +  2.1645E+05, 2.1747E+05, 2.1850E+05, 2.1954E+05, 2.2058E+05,
     + 2.2162E+05, 2.2267E+05, 2.2372E+05, 2.2477E+05, 2.2583E+05,
     +  2.2689E+05, 2.2796E+05, 2.2903E+05, 2.3010E+05, 2.3118E+05,
     + 2.3226E+05, 2.3335E+05, 2.3444E+05, 2.3553E+05, 2.3663E+05,
     +  2.3773E+05, 2.3884E+05, 2.3995E+05, 2.4106E+05, 2.4218E+05,
     + 2.4330E+05, 2.4443E+05, 2.4556E+05, 2.4669E+05, 2.4783E+05,
     +  2.4897E+05, 2.5012E+05, 2.5127E+05, 2.5243E+05, 2.5359E+05,
     + 2.5475E+05, 2.5592E+05, 2.5709E+05, 2.5827E+05, 2.5945E+05,
     +  2.6063E+05, 2.6182E+05, 2.6301E+05, 2.6421E+05, 2.6541E+05,
     + 2.6662E+05, 2.6783E+05, 2.6905E+05, 2.7026E+05, 2.7149E+05,
     +  2.7272E+05, 2.7395E+05, 2.7519E+05, 2.7643E+05, 2.7767E+05,
     + 2.7892E+05, 2.8018E+05, 2.8143E+05, 2.8270E+05, 2.8397E+05,
     +  2.8524E+05, 2.8651E+05, 2.8780E+05, 2.8908E+05, 2.9037E+05,
     + 2.9167E+05, 2.9297E+05, 2.9427E+05, 2.9558E+05, 2.9689E+05,
     +  2.9821E+05, 2.9953E+05, 3.0086E+05, 3.0219E+05, 3.0353E+05,
     + 3.0487E+05, 3.0621E+05, 3.0756E+05, 3.0892E+05, 3.1028E+05,
     +  3.1164E+05, 3.1301E+05, 3.1438E+05, 3.1576E+05, 3.1715E+05,
     + 3.1853E+05, 3.1993E+05, 3.2132E+05, 3.2273E+05, 3.2413E+05,
     +  3.2555E+05, 3.2696E+05, 3.2838E+05, 3.2981E+05, 3.3124E+05,
     + 3.3268E+05, 3.3412E+05, 3.3557E+05, 3.3702E+05, 3.3847E+05,
     +  3.3993E+05, 3.4140E+05, 3.4287E+05, 3.4435E+05, 3.4583E+05,
     + 3.4731E+05, 3.4881E+05, 3.5030E+05, 3.5180E+05, 3.5331E+05,
     +  3.5482E+05, 3.5634E+05, 3.5786E+05, 3.5938E+05, 3.6092E+05,
     + 3.6245E+05, 3.6399E+05, 3.6554E+05, 3.6709E+05, 3.6865E+05,
     +  3.7021E+05, 3.7178E+05, 3.7335E+05, 3.7493E+05, 3.7652E+05,
     + 3.7811E+05, 3.7970E+05, 3.8130E+05, 3.8290E+05, 3.8451E+05,
     +  3.8613E+05, 3.8775E+05, 3.8938E+05, 3.9101E+05, 3.9264E+05,
     + 3.9429E+05, 3.9593E+05, 3.9759E+05, 3.9925E+05, 4.0091E+05,
     +  4.0258E+05, 4.0425E+05, 4.0593E+05, 4.0762E+05, 4.0931E+05,
     + 4.1101E+05, 4.1271E+05, 4.1442E+05, 4.1613E+05, 4.1785E+05,
     +  4.1958E+05, 4.2131E+05, 4.2304E+05, 4.2478E+05, 4.2653E+05,
     + 4.2828E+05, 4.3004E+05, 4.3181E+05, 4.3358E+05, 4.3535E+05,
     +  4.3713E+05, 4.3892E+05, 4.4071E+05, 4.4251E+05, 4.4432E+05,
     + 4.4613E+05, 4.4794E+05, 4.4976E+05, 4.5159E+05, 4.5342E+05,
     +  4.5526E+05, 4.5711E+05, 4.5896E+05, 4.6082E+05, 4.6268E+05,
     + 4.6455E+05, 4.6642E+05, 4.6830E+05, 4.7019E+05, 4.7208E+05,
     +  4.7398E+05, 4.7589E+05, 4.7780E+05, 4.7971E+05, 4.8164E+05,
     + 4.8356E+05, 4.8550E+05, 4.8744E+05, 4.8939E+05, 4.9134E+05,
     +  4.9330E+05, 4.9527E+05, 4.9724E+05, 4.9921E+05, 5.0120E+05,
     + 5.0319E+05, 5.0518E+05, 5.0719E+05, 5.0920E+05, 5.1121E+05,
     +  5.1323E+05, 5.1526E+05, 5.1729E+05, 5.1933E+05, 5.2138E+05,
     + 5.2343E+05, 5.2549E+05, 5.2756E+05, 5.2963E+05, 5.3171E+05,
     +  5.3379E+05, 5.3589E+05, 5.3798E+05, 5.4009E+05, 5.4220E+05,
     + 5.4432E+05, 5.4644E+05, 5.4857E+05, 5.5071E+05, 5.5285E+05,
     +  5.5500E+05, 5.5716E+05, 5.5932E+05, 5.6149E+05, 5.6366E+05,
     + 5.6585E+05, 5.6804E+05, 5.7023E+05, 5.7243E+05, 5.7464E+05,
     +  5.7686E+05, 5.7908E+05, 5.8131E+05, 5.8355E+05, 5.8579E+05,
     + 5.8804E+05, 5.9030E+05, 5.9256E+05, 5.9483E+05, 5.9711E+05,
     +  5.9939E+05, 6.0169E+05, 6.0398E+05, 6.0629E+05, 6.0860E+05,
     + 6.1092E+05, 6.1324E+05, 6.1558E+05, 6.1792E+05, 6.2026E+05,
     +  6.2262E+05, 6.2498E+05, 6.2734E+05, 6.2972E+05, 6.3210E+05,
     + 6.3449E+05, 6.3688E+05, 6.3929E+05, 6.4170E+05, 6.4411E+05,
     +  6.4654E+05, 6.4897E+05, 6.5141E+05, 6.5386E+05, 6.5631E+05,
     + 6.5877E+05, 6.6124E+05, 6.6371E+05, 6.6619E+05, 6.6868E+05,
     +  6.7118E+05, 6.7368E+05, 6.7620E+05, 6.7871E+05, 6.8124E+05,
     + 6.8377E+05, 6.8631E+05, 6.8886E+05, 6.9142E+05, 6.9398E+05,
     +  6.9655E+05, 6.9913E+05, 7.0172E+05, 7.0431E+05, 7.0691E+05,
     + 7.0952E+05, 7.1214E+05, 7.1476E+05, 7.1739E+05, 7.2003E+05,
     +  7.2268E+05, 7.2533E+05, 7.2799E+05, 7.3066E+05, 7.3334E+05,
     + 7.3602E+05, 7.3872E+05, 7.4142E+05, 7.4412E+05, 7.4684E+05,
     +  7.4956E+05, 7.5229E+05, 7.5503E+05, 7.5778E+05, 7.6054E+05,
     + 7.6330E+05, 7.6607E+05, 7.6885E+05, 7.7163E+05, 7.7443E+05,
     +  7.7723E+05, 7.8004E+05, 7.8286E+05, 7.8569E+05, 7.8852E+05,
     + 7.9136E+05, 7.9421E+05, 7.9707E+05, 7.9994E+05, 8.0281E+05,
     +  8.0570E+05, 8.0859E+05, 8.1148E+05, 8.1439E+05, 8.1731E+05,
     + 8.2023E+05, 8.2316E+05, 8.2610E+05, 8.2905E+05, 8.3201E+05,
     +  8.3497E+05, 8.3795E+05, 8.4093E+05, 8.4392E+05, 8.4692E+05,
     + 8.4992E+05, 8.5294E+05, 8.5596E+05, 8.5899E+05, 8.6203E+05,
     +  8.6508E+05, 8.6814E+05, 8.7121E+05, 8.7428E+05, 8.7736E+05,
     + 8.8045E+05, 8.8355E+05, 8.8666E+05, 8.8978E+05, 8.9290E+05,
     +  8.9604E+05, 8.9918E+05, 9.0233E+05, 9.0549E+05, 9.0866E+05,
     + 9.1184E+05, 9.1503E+05, 9.1822E+05, 9.2143E+05, 9.2464E+05,
     +  9.2786E+05, 9.3109E+05, 9.3433E+05, 9.3758E+05, 9.4083E+05,
     + 9.4410E+05, 9.4737E+05, 9.5066E+05, 9.5395E+05, 9.5725E+05,
     +  9.6056E+05, 9.6388E+05, 9.6721E+05, 9.7055E+05, 9.7390E+05,
     + 9.7725E+05, 9.8062E+05, 9.8399E+05, 9.8737E+05, 9.9076E+05,
     +  9.9416E+05, 9.9758E+05, 1.0010E+06, 1.0044E+06, 1.0079E+06,
     + 1.0113E+06, 1.0148E+06, 1.0182E+06, 1.0217E+06, 1.0252E+06,
     +  1.0287E+06, 1.0322E+06, 1.0357E+06, 1.0392E+06, 1.0428E+06,
     + 1.0463E+06, 1.0498E+06, 1.0534E+06, 1.0570E+06, 1.0606E+06,
     +  1.0641E+06, 1.0677E+06, 1.0713E+06, 1.0750E+06, 1.0786E+06,
     + 1.0822E+06, 1.0859E+06, 1.0895E+06, 1.0932E+06, 1.0969E+06,
     +  1.1006E+06, 1.1042E+06, 1.1079E+06, 1.1117E+06, 1.1154E+06,
     + 1.1191E+06, 1.1229E+06, 1.1266E+06, 1.1304E+06, 1.1342E+06,
     +  1.1379E+06, 1.1417E+06, 1.1455E+06, 1.1493E+06, 1.1532E+06,
     + 1.1570E+06, 1.1608E+06, 1.1647E+06, 1.1686E+06, 1.1724E+06,
     +  1.1763E+06, 1.1802E+06, 1.1841E+06, 1.1880E+06, 1.1919E+06,
     + 1.1959E+06, 1.1998E+06, 1.2038E+06, 1.2077E+06, 1.2117E+06,
     +  1.2157E+06, 1.2197E+06, 1.2237E+06, 1.2277E+06, 1.2317E+06,
     + 1.2358E+06, 1.2398E+06, 1.2439E+06, 1.2479E+06, 1.2520E+06,
     +  1.2561E+06, 1.2602E+06, 1.2643E+06, 1.2684E+06, 1.2726E+06,
     + 1.2767E+06, 1.2809E+06, 1.2850E+06, 1.2892E+06, 1.2934E+06,
     +  1.2976E+06, 1.3018E+06, 1.3060E+06, 1.3102E+06, 1.3144E+06,
     + 1.3187E+06, 1.3230E+06, 1.3272E+06, 1.3315E+06, 1.3358E+06,
     +  1.3401E+06, 1.3444E+06, 1.3487E+06, 1.3531E+06, 1.3574E+06,
     + 1.3618E+06, 1.3661E+06, 1.3705E+06, 1.3749E+06, 1.3793E+06,
     +  1.3837E+06, 1.3881E+06, 1.3926E+06, 1.3970E+06, 1.4015E+06,
     + 1.4059E+06, 1.4104E+06, 1.4149E+06, 1.4194E+06, 1.4239E+06,
     +  1.4284E+06, 1.4330E+06, 1.4375E+06, 1.4421E+06, 1.4466E+06,
     + 1.4512E+06, 1.4558E+06, 1.4604E+06, 1.4650E+06, 1.4697E+06,
     +  1.4743E+06, 1.4789E+06, 1.4836E+06, 1.4883E+06, 1.4930E+06,
     + 1.4977E+06, 1.5024E+06, 1.5071E+06, 1.5118E+06, 1.5165E+06,
     +  1.5213E+06, 1.5261E+06, 1.5308E+06, 1.5356E+06, 1.5404E+06,
     + 1.5452E+06, 1.5501E+06, 1.5549E+06, 1.5597E+06, 1.5646E+06,
     +  1.5695E+06, 1.5744E+06, 1.5792E+06, 1.5842E+06, 1.5891E+06,
     + 1.5940E+06, 1.5989E+06, 1.6039E+06, 1.6089E+06, 1.6138E+06,
     +  1.6188E+06, 1.6238E+06, 1.6289E+06, 1.6339E+06, 1.6389E+06,
     + 1.6440E+06, 1.6490E+06, 1.6541E+06, 1.6592E+06, 1.6643E+06,
     +  1.6694E+06, 1.6745E+06, 1.6797E+06, 1.6848E+06, 1.6900E+06,
     + 1.6952E+06, 1.7004E+06, 1.7056E+06, 1.7108E+06, 1.7160E+06,
     +  1.7212E+06, 1.7265E+06, 1.7317E+06, 1.7370E+06, 1.7423E+06,
     + 1.7476E+06, 1.7529E+06, 1.7582E+06, 1.7636E+06, 1.7689E+06,
     +  1.7743E+06, 1.7797E+06, 1.7851E+06, 1.7905E+06, 1.7959E+06,
     + 1.8013E+06, 1.8067E+06, 1.8122E+06, 1.8177E+06, 1.8231E+06,
     +  1.8286E+06, 1.8341E+06, 1.8397E+06, 1.8452E+06, 1.8507E+06,
     + 1.8563E+06, 1.8619E+06, 1.8675E+06, 1.8731E+06, 1.8787E+06,
     +  1.8843E+06, 1.8899E+06, 1.8956E+06, 1.9012E+06, 1.9069E+06,
     + 1.9126E+06, 1.9183E+06, 1.9240E+06, 1.9298E+06, 1.9355E+06,
     +  1.9413E+06, 1.9470E+06, 1.9528E+06, 1.9586E+06, 1.9644E+06,
     + 1.9702E+06, 1.9761E+06, 1.9819E+06, 1.9878E+06, 1.9937E+06,
     +  1.9996E+06, 2.0055E+06, 2.0114E+06, 2.0173E+06, 2.0233E+06,
     + 2.0292E+06, 2.0352E+06, 2.0412E+06, 2.0472E+06, 2.0532E+06,
     +  2.0592E+06, 2.0653E+06, 2.0713E+06, 2.0774E+06, 2.0835E+06,
     + 2.0896E+06, 2.0957E+06, 2.1018E+06, 2.1080E+06, 2.1141E+06,
     +  2.1203E+06, 2.1265E+06, 2.1327E+06, 2.1389E+06, 2.1451E+06,
     + 2.1514E+06, 2.1576E+06, 2.1639E+06, 2.1702E+06, 2.1765E+06,
     +  2.1828E+06, 2.1891E+06, 2.1955E+06, 2.2018E+06, 2.2082E+06,
     + 2.2146E+06, 2.2210E+06, 2.2274E+06, 2.2338E+06, 2.2403E+06,
     +  2.2467E+06, 2.2532E+06, 2.2597E+06, 2.2662E+06, 2.2727E+06,
     + 2.2792E+06, 2.2858E+06, 2.2923E+06, 2.2989E+06, 2.3055E+06,
     +  2.3121E+06, 2.3187E+06, 2.3254E+06, 2.3320E+06, 2.3387E+06,
     + 2.3454E+06, 2.3521E+06, 2.3588E+06, 2.3655E+06, 2.3722E+06,
     +  2.3790E+06, 2.3858E+06, 2.3925E+06, 2.3993E+06, 2.4062E+06,
     + 2.4130E+06, 2.4198E+06, 2.4267E+06, 2.4336E+06, 2.4405E+06,
     +  2.4474E+06, 2.4543E+06, 2.4612E+06, 2.4682E+06, 2.4752E+06,
     + 2.4821E+06, 2.4891E+06, 2.4962E+06, 2.5032E+06, 2.5102E+06,
     +  2.5173E+06, 2.5244E+06, 2.5315E+06, 2.5386E+06, 2.5457E+06,
     + 2.5528E+06, 2.5600E+06, 2.5672E+06, 2.5743E+06, 2.5815E+06,
     +  2.5888E+06, 2.5960E+06, 2.6033E+06, 2.6105E+06, 2.6178E+06,
     + 2.6251E+06, 2.6324E+06, 2.6397E+06, 2.6471E+06, 2.6545E+06,
     +  2.6618E+06, 2.6692E+06, 2.6766E+06, 2.6841E+06, 2.6915E+06,
     + 2.6990E+06, 2.7064E+06, 2.7139E+06, 2.7214E+06, 2.7290E+06,
     +  2.7365E+06, 2.7441E+06, 2.7516E+06, 2.7592E+06, 2.7668E+06,
     + 2.7745E+06, 2.7821E+06, 2.7898E+06, 2.7974E+06, 2.8051E+06,
     +  2.8128E+06, 2.8205E+06, 2.8283E+06, 2.8360E+06, 2.8438E+06,
     + 2.8516E+06, 2.8594E+06, 2.8672E+06, 2.8751E+06, 2.8829E+06,
     +  2.8908E+06, 2.8987E+06, 2.9066E+06, 2.9145E+06, 2.9225E+06,
     + 2.9304E+06, 2.9384E+06, 2.9464E+06, 2.9544E+06, 2.9624E+06,
     +  2.9705E+06, 2.9785E+06, 2.9866E+06, 2.9947E+06, 3.0028E+06,
     + 3.0109E+06, 3.0191E+06, 3.0272E+06, 3.0354E+06, 3.0436E+06,
     +  3.0518E+06, 3.0601E+06, 3.0683E+06, 3.0766E+06, 3.0849E+06,
     + 3.0932E+06, 3.1015E+06, 3.1098E+06, 3.1182E+06, 3.1266E+06,
     +  3.1350E+06, 3.1434E+06, 3.1518E+06, 3.1602E+06, 3.1687E+06,
     + 3.1772E+06, 3.1857E+06, 3.1942E+06, 3.2027E+06, 3.2113E+06,
     +  3.2199E+06, 3.2284E+06, 3.2371E+06, 3.2457E+06, 3.2543E+06,
     + 3.2630E+06, 3.2717E+06, 3.2804E+06, 3.2891E+06, 3.2978E+06,
     +  3.3065E+06, 3.3153E+06, 3.3241E+06, 3.3329E+06, 3.3417E+06,
     + 3.3506E+06, 3.3594E+06, 3.3683E+06, 3.3772E+06, 3.3861E+06,
     +  3.3951E+06, 3.4040E+06, 3.4130E+06, 3.4220E+06, 3.4310E+06,
     + 3.4400E+06, 3.4491E+06, 3.4581E+06, 3.4672E+06, 3.4763E+06,
     +  3.4854E+06, 3.4946E+06, 3.5037E+06, 3.5129E+06, 3.5221E+06,
     + 3.5313E+06, 3.5405E+06, 3.5498E+06, 3.5591E+06, 3.5684E+06,
     +  3.5777E+06, 3.5870E+06, 3.5963E+06, 3.6057E+06, 3.6151E+06,
     + 3.6245E+06, 3.6339E+06, 3.6434E+06, 3.6528E+06, 3.6623E+06,
     +  3.6718E+06, 3.6813E+06, 3.6909E+06, 3.7004E+06, 3.7100E+06,
     + 3.7196E+06, 3.7292E+06, 3.7389E+06, 3.7485E+06, 3.7582E+06,
     +  3.7679E+06, 3.7776E+06, 3.7874E+06, 3.7971E+06, 3.8069E+06,
     + 3.8167E+06, 3.8265E+06, 3.8363E+06, 3.8462E+06, 3.8561E+06,
     +  3.8660E+06, 3.8759E+06, 3.8858E+06, 3.8958E+06, 3.9057E+06,
     + 3.9157E+06, 3.9257E+06, 3.9358E+06, 3.9458E+06, 3.9559E+06,
     +  3.9660E+06, 3.9761E+06, 3.9863E+06, 3.9964E+06, 4.0066E+06,
     + 4.0168E+06, 4.0270E+06, 4.0372E+06, 4.0475E+06, 4.0578E+06,
     +  4.0681E+06, 4.0784E+06, 4.0887E+06, 4.0991E+06, 4.1095E+06,
     + 4.1199E+06, 4.1303E+06, 4.1407E+06, 4.1512E+06, 4.1617E+06,
     +  4.1722E+06, 4.1827E+06, 4.1933E+06, 4.2038E+06, 4.2144E+06,
     + 4.2250E+06, 4.2357E+06, 4.2463E+06, 4.2570E+06, 4.2677E+06,
     +  4.2784E+06, 4.2891E+06, 4.2999E+06, 4.3107E+06, 4.3215E+06,
     + 4.3323E+06, 4.3431E+06, 4.3540E+06, 4.3649E+06, 4.3758E+06,
     +  4.3867E+06, 4.3977E+06, 4.4087E+06, 4.4196E+06, 4.4307E+06,
     + 4.4417E+06, 4.4528E+06, 4.4638E+06, 4.4749E+06, 4.4861E+06,
     +  4.4972E+06, 4.5084E+06, 4.5196E+06, 4.5308E+06, 4.5420E+06,
     + 4.5532E+06, 4.5645E+06, 4.5758E+06, 4.5871E+06, 4.5985E+06,
     +  4.6098E+06, 4.6212E+06, 4.6326E+06, 4.6441E+06, 4.6555E+06,
     + 4.6670E+06, 4.6785E+06, 4.6900E+06, 4.7015E+06, 4.7131E+06,
     +  4.7247E+06, 4.7363E+06, 4.7479E+06, 4.7596E+06, 4.7713E+06,
     + 4.7829E+06, 4.7947E+06, 4.8064E+06, 4.8182E+06, 4.8300E+06,
     +  4.8418E+06, 4.8536E+06, 4.8655E+06, 4.8773E+06, 4.8892E+06,
     + 4.9012E+06, 4.9131E+06, 4.9251E+06, 4.9371E+06, 4.9491E+06,
     +  4.9611E+06, 4.9732E+06, 4.9853E+06, 4.9974E+06, 5.0095E+06,
     + 5.0217E+06, 5.0338E+06, 5.0460E+06, 5.0583E+06, 5.0705E+06,
     +  5.0828E+06, 5.0951E+06, 5.1074E+06, 5.1197E+06, 5.1321E+06,
     + 5.1445E+06, 5.1569E+06, 5.1693E+06, 5.1818E+06, 5.1943E+06,
     +  5.2068E+06, 5.2193E+06, 5.2318E+06, 5.2444E+06, 5.2570E+06,
     + 5.2696E+06, 5.2823E+06, 5.2950E+06, 5.3077E+06, 5.3204E+06,
     +  5.3331E+06, 5.3459E+06, 5.3587E+06, 5.3715E+06, 5.3843E+06,
     + 5.3972E+06, 5.4101E+06, 5.4230E+06, 5.4359E+06, 5.4489E+06,
     +  5.4619E+06, 5.4749E+06, 5.4879E+06, 5.5010E+06, 5.5140E+06,
     + 5.5272E+06, 5.5403E+06, 5.5534E+06, 5.5666E+06, 5.5798E+06,
     +  5.5931E+06, 5.6063E+06, 5.6196E+06, 5.6329E+06, 5.6462E+06,
     + 5.6596E+06, 5.6729E+06, 5.6863E+06, 5.6998E+06, 5.7132E+06,
     +  5.7267E+06, 5.7402E+06, 5.7537E+06, 5.7673E+06, 5.7809E+06,
     + 5.7945E+06, 5.8081E+06, 5.8217E+06, 5.8354E+06, 5.8491E+06,
     +  5.8628E+06, 5.8766E+06, 5.8904E+06, 5.9042E+06, 5.9180E+06,
     + 5.9319E+06, 5.9457E+06, 5.9596E+06, 5.9736E+06, 5.9875E+06,
     +  6.0015E+06, 6.0155E+06, 6.0296E+06, 6.0436E+06, 6.0577E+06,
     + 6.0718E+06, 6.0859E+06, 6.1001E+06, 6.1143E+06, 6.1285E+06,
     +  6.1428E+06, 6.1570E+06, 6.1713E+06, 6.1856E+06, 6.2000E+06,
     + 6.2143E+06, 6.2287E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT(13,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       887
      data (QofT(14,J),J=1,2408)/ 1.0710E+01, 2.7803E+01, 4.9592E+01,
     +  7.5234E+01, 1.0422E+02, 1.3620E+02, 1.7092E+02, 2.0817E+02,
     + 2.4780E+02, 2.8967E+02, 3.3366E+02, 3.7968E+02, 4.2765E+02,
     +  4.7748E+02, 5.2911E+02, 5.8248E+02, 6.3754E+02, 6.9423E+02,
     + 7.5251E+02, 8.1234E+02, 9.3648E+02, 1.0664E+03, 1.2018E+03,
     +  1.3425E+03, 1.4883E+03, 1.6391E+03, 1.7946E+03, 1.9548E+03,
     + 2.1195E+03, 2.2886E+03, 2.4620E+03, 2.6395E+03, 2.8211E+03,
     +  3.0067E+03, 3.1963E+03, 3.3896E+03, 3.5868E+03, 3.7876E+03,
     + 3.9920E+03, 4.2000E+03, 4.4115E+03, 4.6265E+03, 4.8448E+03,
     +  5.0665E+03, 5.2915E+03, 5.5198E+03, 5.7512E+03, 5.9858E+03,
     + 6.2235E+03, 6.4644E+03, 6.7082E+03, 6.9550E+03, 7.2049E+03,
     +  7.4577E+03, 7.7133E+03, 7.9719E+03, 8.2334E+03, 8.4976E+03,
     + 8.7647E+03, 9.0345E+03, 9.3070E+03, 9.5824E+03, 9.8604E+03,
     +  1.0141E+04, 1.0424E+04, 1.0710E+04, 1.0999E+04, 1.1290E+04,
     + 1.1584E+04, 1.1881E+04, 1.2180E+04, 1.2481E+04, 1.2786E+04,
     +  1.3092E+04, 1.3402E+04, 1.3713E+04, 1.4028E+04, 1.4344E+04,
     + 1.4664E+04, 1.4986E+04, 1.5310E+04, 1.5637E+04, 1.5966E+04,
     +  1.6298E+04, 1.6633E+04, 1.6970E+04, 1.7309E+04, 1.7651E+04,
     + 1.7995E+04, 1.8342E+04, 1.8692E+04, 1.9044E+04, 1.9398E+04,
     +  1.9756E+04, 2.0115E+04, 2.0478E+04, 2.0842E+04, 2.1210E+04,
     + 2.1580E+04, 2.1952E+04, 2.2328E+04, 2.2706E+04, 2.3086E+04,
     +  2.3469E+04, 2.3855E+04, 2.4244E+04, 2.4635E+04, 2.5029E+04,
     + 2.5426E+04, 2.5826E+04, 2.6228E+04, 2.6633E+04, 2.7041E+04,
     +  2.7452E+04, 2.7866E+04, 2.8282E+04, 2.8701E+04, 2.9124E+04,
     + 2.9549E+04, 2.9977E+04, 3.0408E+04, 3.0842E+04, 3.1280E+04,
     +  3.1720E+04, 3.2163E+04, 3.2610E+04, 3.3059E+04, 3.3512E+04,
     + 3.3968E+04, 3.4427E+04, 3.4889E+04, 3.5355E+04, 3.5824E+04,
     +  3.6296E+04, 3.6771E+04, 3.7250E+04, 3.7732E+04, 3.8218E+04,
     + 3.8707E+04, 3.9199E+04, 3.9695E+04, 4.0195E+04, 4.0698E+04,
     +  4.1204E+04, 4.1715E+04, 4.2228E+04, 4.2746E+04, 4.3267E+04,
     + 4.3792E+04, 4.4320E+04, 4.4853E+04, 4.5389E+04, 4.5929E+04,
     +  4.6473E+04, 4.7021E+04, 4.7573E+04, 4.8128E+04, 4.8688E+04,
     + 4.9252E+04, 4.9819E+04, 5.0392E+04, 5.0967E+04, 5.1547E+04,
     +  5.2132E+04, 5.2720E+04, 5.3313E+04, 5.3910E+04, 5.4511E+04,
     + 5.5117E+04, 5.5727E+04, 5.6341E+04, 5.6960E+04, 5.7583E+04,
     +  5.8211E+04, 5.8843E+04, 5.9481E+04, 6.0122E+04, 6.0768E+04,
     + 6.1419E+04, 6.2074E+04, 6.2735E+04, 6.3400E+04, 6.4070E+04,
     +  6.4744E+04, 6.5423E+04, 6.6108E+04, 6.6798E+04, 6.7492E+04,
     + 6.8191E+04, 6.8895E+04, 6.9605E+04, 7.0319E+04, 7.1039E+04,
     +  7.1764E+04, 7.2494E+04, 7.3229E+04, 7.3970E+04, 7.4716E+04,
     + 7.5467E+04, 7.6224E+04, 7.6986E+04, 7.7754E+04, 7.8527E+04,
     +  7.9305E+04, 8.0090E+04, 8.0880E+04, 8.1675E+04, 8.2476E+04,
     + 8.3283E+04, 8.4095E+04, 8.4914E+04, 8.5739E+04, 8.6569E+04,
     +  8.7405E+04, 8.8246E+04, 8.9094E+04, 8.9948E+04, 9.0808E+04,
     + 9.1675E+04, 9.2546E+04, 9.3425E+04, 9.4309E+04, 9.5200E+04,
     +  9.6097E+04, 9.7000E+04, 9.7911E+04, 9.8827E+04, 9.9749E+04,
     + 1.0068E+05, 1.0161E+05, 1.0256E+05, 1.0350E+05, 1.0446E+05,
     +  1.0542E+05, 1.0639E+05, 1.0737E+05, 1.0835E+05, 1.0934E+05,
     + 1.1033E+05, 1.1134E+05, 1.1235E+05, 1.1336E+05, 1.1439E+05,
     +  1.1542E+05, 1.1646E+05, 1.1750E+05, 1.1855E+05, 1.1961E+05,
     + 1.2068E+05, 1.2175E+05, 1.2283E+05, 1.2392E+05, 1.2502E+05,
     +  1.2612E+05, 1.2724E+05, 1.2835E+05, 1.2948E+05, 1.3061E+05,
     + 1.3176E+05, 1.3290E+05, 1.3406E+05, 1.3523E+05, 1.3640E+05,
     +  1.3758E+05, 1.3877E+05, 1.3997E+05, 1.4117E+05, 1.4238E+05,
     + 1.4360E+05, 1.4483E+05, 1.4607E+05, 1.4732E+05, 1.4857E+05,
     +  1.4983E+05, 1.5110E+05, 1.5238E+05, 1.5367E+05, 1.5496E+05,
     + 1.5627E+05, 1.5758E+05, 1.5890E+05, 1.6023E+05, 1.6157E+05,
     +  1.6292E+05, 1.6428E+05, 1.6564E+05, 1.6702E+05, 1.6840E+05,
     + 1.6979E+05, 1.7119E+05, 1.7260E+05, 1.7402E+05, 1.7545E+05,
     +  1.7689E+05, 1.7834E+05, 1.7979E+05, 1.8126E+05, 1.8273E+05,
     + 1.8422E+05, 1.8571E+05, 1.8722E+05, 1.8873E+05, 1.9025E+05,
     +  1.9179E+05, 1.9333E+05, 1.9488E+05, 1.9644E+05, 1.9801E+05,
     + 1.9959E+05, 2.0119E+05, 2.0279E+05, 2.0440E+05, 2.0602E+05,
     +  2.0766E+05, 2.0930E+05, 2.1095E+05, 2.1261E+05, 2.1429E+05,
     + 2.1597E+05, 2.1766E+05, 2.1937E+05, 2.2108E+05, 2.2281E+05,
     +  2.2455E+05, 2.2630E+05, 2.2805E+05, 2.2982E+05, 2.3160E+05,
     + 2.3339E+05, 2.3520E+05, 2.3701E+05, 2.3883E+05, 2.4067E+05,
     +  2.4251E+05, 2.4437E+05, 2.4624E+05, 2.4812E+05, 2.5001E+05,
     + 2.5191E+05, 2.5383E+05, 2.5575E+05, 2.5769E+05, 2.5964E+05,
     +  2.6160E+05, 2.6357E+05, 2.6556E+05, 2.6756E+05, 2.6956E+05,
     + 2.7158E+05, 2.7361E+05, 2.7566E+05, 2.7772E+05, 2.7978E+05,
     +  2.8187E+05, 2.8396E+05, 2.8606E+05, 2.8818E+05, 2.9031E+05,
     + 2.9245E+05, 2.9461E+05, 2.9678E+05, 2.9896E+05, 3.0115E+05,
     +  3.0336E+05, 3.0558E+05, 3.0781E+05, 3.1005E+05, 3.1231E+05,
     + 3.1458E+05, 3.1687E+05, 3.1916E+05, 3.2147E+05, 3.2380E+05,
     +  3.2614E+05, 3.2849E+05, 3.3085E+05, 3.3323E+05, 3.3562E+05,
     + 3.3802E+05, 3.4044E+05, 3.4288E+05, 3.4532E+05, 3.4778E+05,
     +  3.5025E+05, 3.5274E+05, 3.5525E+05, 3.5776E+05, 3.6029E+05,
     + 3.6284E+05, 3.6539E+05, 3.6797E+05, 3.7055E+05, 3.7316E+05,
     +  3.7577E+05, 3.7840E+05, 3.8105E+05, 3.8371E+05, 3.8639E+05,
     + 3.8908E+05, 3.9178E+05, 3.9450E+05, 3.9724E+05, 3.9998E+05,
     +  4.0275E+05, 4.0553E+05, 4.0833E+05, 4.1114E+05, 4.1396E+05,
     + 4.1681E+05, 4.1966E+05, 4.2253E+05, 4.2542E+05, 4.2833E+05,
     +  4.3125E+05, 4.3418E+05, 4.3713E+05, 4.4010E+05, 4.4309E+05,
     + 4.4609E+05, 4.4910E+05, 4.5214E+05, 4.5518E+05, 4.5825E+05,
     +  4.6133E+05, 4.6443E+05, 4.6754E+05, 4.7067E+05, 4.7382E+05,
     + 4.7698E+05, 4.8016E+05, 4.8336E+05, 4.8658E+05, 4.8981E+05,
     +  4.9306E+05, 4.9632E+05, 4.9960E+05, 5.0291E+05, 5.0622E+05,
     + 5.0956E+05, 5.1291E+05, 5.1628E+05, 5.1967E+05, 5.2307E+05,
     +  5.2650E+05, 5.2994E+05, 5.3340E+05, 5.3687E+05, 5.4037E+05,
     + 5.4388E+05, 5.4741E+05, 5.5096E+05, 5.5453E+05, 5.5811E+05,
     +  5.6172E+05, 5.6534E+05, 5.6898E+05, 5.7264E+05, 5.7632E+05,
     + 5.8002E+05, 5.8373E+05, 5.8747E+05, 5.9122E+05, 5.9500E+05,
     +  5.9879E+05, 6.0260E+05, 6.0643E+05, 6.1028E+05, 6.1415E+05,
     + 6.1804E+05, 6.2195E+05, 6.2587E+05, 6.2982E+05, 6.3379E+05,
     +  6.3778E+05, 6.4179E+05, 6.4581E+05, 6.4986E+05, 6.5393E+05,
     + 6.5802E+05, 6.6212E+05, 6.6625E+05, 6.7040E+05, 6.7457E+05,
     +  6.7876E+05, 6.8297E+05, 6.8720E+05, 6.9145E+05, 6.9573E+05,
     + 7.0002E+05, 7.0434E+05, 7.0867E+05, 7.1303E+05, 7.1741E+05,
     +  7.2181E+05, 7.2623E+05, 7.3067E+05, 7.3514E+05, 7.3962E+05,
     + 7.4413E+05, 7.4866E+05, 7.5321E+05, 7.5778E+05, 7.6238E+05,
     +  7.6700E+05, 7.7164E+05, 7.7630E+05, 7.8098E+05, 7.8569E+05,
     + 7.9042E+05, 7.9517E+05, 7.9994E+05, 8.0474E+05, 8.0956E+05,
     +  8.1440E+05, 8.1927E+05, 8.2416E+05, 8.2907E+05, 8.3401E+05,
     + 8.3896E+05, 8.4395E+05, 8.4895E+05, 8.5398E+05, 8.5903E+05,
     +  8.6411E+05, 8.6921E+05, 8.7433E+05, 8.7948E+05, 8.8465E+05,
     + 8.8985E+05, 8.9507E+05, 9.0031E+05, 9.0558E+05, 9.1088E+05,
     +  9.1619E+05, 9.2153E+05, 9.2690E+05, 9.3229E+05, 9.3771E+05,
     + 9.4315E+05, 9.4862E+05, 9.5411E+05, 9.5963E+05, 9.6517E+05,
     +  9.7074E+05, 9.7633E+05, 9.8195E+05, 9.8759E+05, 9.9326E+05,
     + 9.9896E+05, 1.0047E+06, 1.0104E+06, 1.0162E+06, 1.0220E+06,
     +  1.0278E+06, 1.0337E+06, 1.0396E+06, 1.0455E+06, 1.0514E+06,
     + 1.0573E+06, 1.0633E+06, 1.0693E+06, 1.0754E+06, 1.0814E+06,
     +  1.0875E+06, 1.0937E+06, 1.0998E+06, 1.1060E+06, 1.1122E+06,
     + 1.1184E+06, 1.1247E+06, 1.1310E+06, 1.1373E+06, 1.1436E+06,
     +  1.1500E+06, 1.1564E+06, 1.1628E+06, 1.1693E+06, 1.1757E+06,
     + 1.1822E+06, 1.1888E+06, 1.1953E+06, 1.2019E+06, 1.2086E+06,
     +  1.2152E+06, 1.2219E+06, 1.2286E+06, 1.2354E+06, 1.2421E+06,
     + 1.2489E+06, 1.2558E+06, 1.2626E+06, 1.2695E+06, 1.2764E+06,
     +  1.2834E+06, 1.2903E+06, 1.2973E+06, 1.3044E+06, 1.3115E+06,
     + 1.3185E+06, 1.3257E+06, 1.3328E+06, 1.3400E+06, 1.3472E+06,
     +  1.3545E+06, 1.3618E+06, 1.3691E+06, 1.3764E+06, 1.3838E+06,
     + 1.3912E+06, 1.3986E+06, 1.4061E+06, 1.4136E+06, 1.4211E+06,
     +  1.4287E+06, 1.4363E+06, 1.4439E+06, 1.4516E+06, 1.4593E+06,
     + 1.4670E+06, 1.4748E+06, 1.4825E+06, 1.4904E+06, 1.4982E+06,
     +  1.5061E+06, 1.5140E+06, 1.5220E+06, 1.5300E+06, 1.5380E+06,
     + 1.5460E+06, 1.5541E+06, 1.5622E+06, 1.5704E+06, 1.5786E+06,
     +  1.5868E+06, 1.5950E+06, 1.6033E+06, 1.6117E+06, 1.6200E+06,
     + 1.6284E+06, 1.6368E+06, 1.6453E+06, 1.6538E+06, 1.6623E+06,
     +  1.6709E+06, 1.6794E+06, 1.6881E+06, 1.6967E+06, 1.7054E+06,
     + 1.7142E+06, 1.7230E+06, 1.7318E+06, 1.7406E+06, 1.7495E+06,
     +  1.7584E+06, 1.7673E+06, 1.7763E+06, 1.7854E+06, 1.7944E+06,
     + 1.8035E+06, 1.8126E+06, 1.8218E+06, 1.8310E+06, 1.8403E+06,
     +  1.8495E+06, 1.8588E+06, 1.8682E+06, 1.8776E+06, 1.8870E+06,
     + 1.8965E+06, 1.9060E+06, 1.9155E+06, 1.9251E+06, 1.9347E+06,
     +  1.9444E+06, 1.9541E+06, 1.9638E+06, 1.9735E+06, 1.9834E+06,
     + 1.9932E+06, 2.0031E+06, 2.0130E+06, 2.0230E+06, 2.0330E+06,
     +  2.0430E+06, 2.0531E+06, 2.0632E+06, 2.0733E+06, 2.0835E+06,
     + 2.0938E+06, 2.1040E+06, 2.1143E+06, 2.1247E+06, 2.1351E+06,
     +  2.1455E+06, 2.1560E+06, 2.1665E+06, 2.1771E+06, 2.1877E+06,
     + 2.1983E+06, 2.2090E+06, 2.2197E+06, 2.2305E+06, 2.2412E+06,
     +  2.2521E+06, 2.2630E+06, 2.2739E+06, 2.2849E+06, 2.2959E+06,
     + 2.3069E+06, 2.3180E+06, 2.3291E+06, 2.3403E+06, 2.3515E+06,
     +  2.3628E+06, 2.3741E+06, 2.3854E+06, 2.3968E+06, 2.4083E+06,
     + 2.4197E+06, 2.4312E+06, 2.4428E+06, 2.4544E+06, 2.4661E+06,
     +  2.4777E+06, 2.4895E+06, 2.5013E+06, 2.5131E+06, 2.5249E+06,
     + 2.5369E+06, 2.5488E+06, 2.5608E+06, 2.5728E+06, 2.5849E+06,
     +  2.5971E+06, 2.6092E+06, 2.6215E+06, 2.6337E+06, 2.6460E+06,
     + 2.6584E+06, 2.6708E+06, 2.6833E+06, 2.6958E+06, 2.7083E+06,
     +  2.7209E+06, 2.7335E+06, 2.7462E+06, 2.7589E+06, 2.7717E+06,
     + 2.7845E+06, 2.7974E+06, 2.8103E+06, 2.8232E+06, 2.8363E+06,
     +  2.8493E+06, 2.8624E+06, 2.8756E+06, 2.8888E+06, 2.9020E+06,
     + 2.9153E+06, 2.9286E+06, 2.9420E+06, 2.9554E+06, 2.9689E+06,
     +  2.9825E+06, 2.9961E+06, 3.0097E+06, 3.0234E+06, 3.0371E+06,
     + 3.0509E+06, 3.0647E+06, 3.0786E+06, 3.0925E+06, 3.1065E+06,
     +  3.1205E+06, 3.1346E+06, 3.1487E+06, 3.1629E+06, 3.1771E+06,
     + 3.1914E+06, 3.2057E+06, 3.2201E+06, 3.2345E+06, 3.2490E+06,
     +  3.2635E+06, 3.2781E+06, 3.2927E+06, 3.3074E+06, 3.3221E+06,
     + 3.3369E+06, 3.3518E+06, 3.3666E+06, 3.3816E+06, 3.3966E+06,
     +  3.4116E+06, 3.4267E+06, 3.4419E+06, 3.4571E+06, 3.4723E+06,
     + 3.4877E+06, 3.5030E+06, 3.5184E+06, 3.5339E+06, 3.5494E+06,
     +  3.5650E+06, 3.5806E+06, 3.5963E+06, 3.6121E+06, 3.6279E+06,
     + 3.6437E+06, 3.6596E+06, 3.6756E+06, 3.6916E+06, 3.7077E+06,
     +  3.7238E+06, 3.7400E+06, 3.7562E+06, 3.7725E+06, 3.7888E+06,
     + 3.8052E+06, 3.8217E+06, 3.8382E+06, 3.8548E+06, 3.8714E+06,
     +  3.8881E+06, 3.9048E+06, 3.9216E+06, 3.9385E+06, 3.9554E+06,
     + 3.9723E+06, 3.9894E+06, 4.0064E+06, 4.0236E+06, 4.0408E+06,
     +  4.0580E+06, 4.0753E+06, 4.0927E+06, 4.1101E+06, 4.1276E+06,
     + 4.1452E+06, 4.1628E+06, 4.1804E+06, 4.1982E+06, 4.2159E+06,
     +  4.2338E+06, 4.2517E+06, 4.2696E+06, 4.2877E+06, 4.3057E+06,
     + 4.3239E+06, 4.3421E+06, 4.3603E+06, 4.3787E+06, 4.3970E+06,
     +  4.4155E+06, 4.4340E+06, 4.4525E+06, 4.4712E+06, 4.4899E+06,
     + 4.5086E+06, 4.5274E+06, 4.5463E+06, 4.5652E+06, 4.5842E+06,
     +  4.6033E+06, 4.6224E+06, 4.6416E+06, 4.6608E+06, 4.6801E+06,
     + 4.6995E+06, 4.7189E+06, 4.7384E+06, 4.7580E+06, 4.7776E+06,
     +  4.7973E+06, 4.8170E+06, 4.8368E+06, 4.8567E+06, 4.8767E+06,
     + 4.8967E+06, 4.9167E+06, 4.9369E+06, 4.9571E+06, 4.9773E+06,
     +  4.9977E+06, 5.0181E+06, 5.0385E+06, 5.0591E+06, 5.0797E+06,
     + 5.1003E+06, 5.1210E+06, 5.1418E+06, 5.1627E+06, 5.1836E+06,
     +  5.2046E+06, 5.2257E+06, 5.2468E+06, 5.2680E+06, 5.2893E+06,
     + 5.3106E+06, 5.3320E+06, 5.3534E+06, 5.3750E+06, 5.3966E+06,
     +  5.4182E+06, 5.4400E+06, 5.4618E+06, 5.4837E+06, 5.5056E+06,
     + 5.5276E+06, 5.5497E+06, 5.5719E+06, 5.5941E+06, 5.6164E+06,
     +  5.6387E+06, 5.6611E+06, 5.6837E+06, 5.7062E+06, 5.7289E+06,
     + 5.7516E+06, 5.7744E+06, 5.7972E+06, 5.8201E+06, 5.8431E+06,
     +  5.8662E+06, 5.8893E+06, 5.9125E+06, 5.9358E+06, 5.9592E+06,
     + 5.9826E+06, 6.0061E+06, 6.0297E+06, 6.0533E+06, 6.0770E+06,
     +  6.1008E+06, 6.1247E+06, 6.1486E+06, 6.1727E+06, 6.1967E+06,
     + 6.2209E+06, 6.2451E+06, 6.2694E+06, 6.2938E+06, 6.3183E+06,
     +  6.3428E+06, 6.3674E+06, 6.3921E+06, 6.4169E+06, 6.4417E+06,
     + 6.4666E+06, 6.4916E+06, 6.5166E+06, 6.5418E+06, 6.5670E+06,
     +  6.5923E+06, 6.6176E+06, 6.6431E+06, 6.6686E+06, 6.6942E+06,
     + 6.7199E+06, 6.7456E+06, 6.7714E+06, 6.7973E+06, 6.8233E+06,
     +  6.8494E+06, 6.8755E+06, 6.9017E+06, 6.9280E+06, 6.9544E+06,
     + 6.9809E+06, 7.0074E+06, 7.0340E+06, 7.0607E+06, 7.0875E+06,
     +  7.1143E+06, 7.1413E+06, 7.1683E+06, 7.1954E+06, 7.2225E+06,
     + 7.2498E+06, 7.2771E+06, 7.3045E+06, 7.3320E+06, 7.3596E+06,
     +  7.3873E+06, 7.4150E+06, 7.4428E+06, 7.4707E+06, 7.4987E+06,
     + 7.5268E+06, 7.5550E+06, 7.5832E+06, 7.6115E+06, 7.6399E+06,
     +  7.6684E+06, 7.6970E+06, 7.7256E+06, 7.7543E+06, 7.7832E+06,
     + 7.8121E+06, 7.8411E+06, 7.8701E+06, 7.8993E+06, 7.9285E+06,
     +  7.9579E+06, 7.9873E+06, 8.0168E+06, 8.0464E+06, 8.0760E+06,
     + 8.1058E+06, 8.1356E+06, 8.1656E+06, 8.1956E+06, 8.2257E+06,
     +  8.2559E+06, 8.2861E+06, 8.3165E+06, 8.3470E+06, 8.3775E+06,
     + 8.4081E+06, 8.4388E+06, 8.4696E+06, 8.5005E+06, 8.5315E+06,
     +  8.5626E+06, 8.5937E+06, 8.6250E+06, 8.6563E+06, 8.6877E+06,
     + 8.7193E+06, 8.7509E+06, 8.7826E+06, 8.8143E+06, 8.8462E+06,
     +  8.8782E+06, 8.9102E+06, 8.9424E+06, 8.9746E+06, 9.0070E+06,
     + 9.0394E+06, 9.0719E+06, 9.1045E+06, 9.1372E+06, 9.1700E+06,
     +  9.2029E+06, 9.2358E+06, 9.2689E+06, 9.3021E+06, 9.3353E+06,
     + 9.3687E+06, 9.4021E+06, 9.4356E+06, 9.4693E+06, 9.5030E+06,
     +  9.5368E+06, 9.5707E+06, 9.6047E+06, 9.6388E+06, 9.6730E+06,
     + 9.7073E+06, 9.7417E+06, 9.7762E+06, 9.8108E+06, 9.8455E+06,
     +  9.8802E+06, 9.9151E+06, 9.9501E+06, 9.9851E+06, 1.0020E+07,
     + 1.0056E+07, 1.0091E+07, 1.0126E+07, 1.0162E+07, 1.0198E+07,
     +  1.0233E+07, 1.0269E+07, 1.0305E+07, 1.0341E+07, 1.0377E+07,
     + 1.0414E+07, 1.0450E+07, 1.0486E+07, 1.0523E+07, 1.0559E+07,
     +  1.0596E+07, 1.0633E+07, 1.0670E+07, 1.0707E+07, 1.0744E+07,
     + 1.0781E+07, 1.0819E+07, 1.0856E+07, 1.0894E+07, 1.0931E+07,
     +  1.0969E+07, 1.1007E+07, 1.1045E+07, 1.1083E+07, 1.1121E+07,
     + 1.1159E+07, 1.1198E+07, 1.1236E+07, 1.1275E+07, 1.1314E+07,
     +  1.1352E+07, 1.1391E+07, 1.1430E+07, 1.1469E+07, 1.1509E+07,
     + 1.1548E+07, 1.1587E+07, 1.1627E+07, 1.1666E+07, 1.1706E+07,
     +  1.1746E+07, 1.1786E+07, 1.1826E+07, 1.1866E+07, 1.1906E+07,
     + 1.1947E+07, 1.1987E+07, 1.2028E+07, 1.2069E+07, 1.2109E+07,
     +  1.2150E+07, 1.2191E+07, 1.2233E+07, 1.2274E+07, 1.2315E+07,
     + 1.2357E+07, 1.2398E+07, 1.2440E+07, 1.2482E+07, 1.2524E+07,
     +  1.2566E+07, 1.2608E+07, 1.2650E+07, 1.2692E+07, 1.2735E+07,
     + 1.2777E+07, 1.2820E+07, 1.2863E+07, 1.2906E+07, 1.2949E+07,
     +  1.2992E+07, 1.3035E+07, 1.3078E+07, 1.3122E+07, 1.3165E+07,
     + 1.3209E+07, 1.3253E+07, 1.3297E+07, 1.3341E+07, 1.3385E+07,
     +  1.3429E+07, 1.3474E+07, 1.3518E+07, 1.3563E+07, 1.3608E+07,
     + 1.3652E+07, 1.3697E+07, 1.3742E+07, 1.3788E+07, 1.3833E+07,
     +  1.3878E+07, 1.3924E+07, 1.3970E+07, 1.4015E+07, 1.4061E+07,
     + 1.4107E+07, 1.4153E+07, 1.4200E+07, 1.4246E+07, 1.4293E+07,
     +  1.4339E+07, 1.4386E+07, 1.4433E+07, 1.4480E+07, 1.4527E+07,
     + 1.4574E+07, 1.4621E+07, 1.4669E+07, 1.4716E+07, 1.4764E+07,
     +  1.4812E+07, 1.4860E+07, 1.4908E+07, 1.4956E+07, 1.5004E+07,
     + 1.5053E+07, 1.5101E+07, 1.5150E+07, 1.5199E+07, 1.5248E+07,
     +  1.5297E+07, 1.5346E+07, 1.5395E+07, 1.5445E+07, 1.5494E+07,
     + 1.5544E+07, 1.5594E+07, 1.5643E+07, 1.5694E+07, 1.5744E+07,
     +  1.5794E+07, 1.5844E+07, 1.5895E+07, 1.5946E+07, 1.5996E+07,
     + 1.6047E+07, 1.6098E+07, 1.6150E+07, 1.6201E+07, 1.6252E+07,
     +  1.6304E+07, 1.6356E+07, 1.6407E+07, 1.6459E+07, 1.6511E+07,
     + 1.6564E+07, 1.6616E+07, 1.6668E+07, 1.6721E+07, 1.6774E+07,
     +  1.6827E+07, 1.6880E+07, 1.6933E+07, 1.6986E+07, 1.7039E+07,
     + 1.7093E+07, 1.7147E+07, 1.7200E+07, 1.7254E+07, 1.7308E+07,
     +  1.7362E+07, 1.7417E+07, 1.7471E+07, 1.7526E+07, 1.7580E+07,
     + 1.7635E+07, 1.7690E+07, 1.7745E+07, 1.7801E+07, 1.7856E+07,
     +  1.7912E+07, 1.7967E+07, 1.8023E+07, 1.8079E+07, 1.8135E+07,
     + 1.8191E+07, 1.8248E+07, 1.8304E+07, 1.8361E+07, 1.8417E+07,
     +  1.8474E+07, 1.8531E+07, 1.8588E+07, 1.8646E+07, 1.8703E+07,
     + 1.8761E+07, 1.8819E+07, 1.8876E+07, 1.8934E+07, 1.8993E+07,
     +  1.9051E+07, 1.9109E+07, 1.9168E+07, 1.9227E+07, 1.9285E+07,
     + 1.9344E+07, 1.9403E+07, 1.9463E+07, 1.9522E+07, 1.9582E+07,
     +  1.9641E+07, 1.9701E+07, 1.9761E+07, 1.9821E+07, 1.9882E+07,
     + 1.9942E+07, 2.0003E+07, 2.0063E+07, 2.0124E+07, 2.0185E+07,
     +  2.0246E+07, 2.0308E+07, 2.0369E+07, 2.0431E+07, 2.0492E+07,
     + 2.0554E+07, 2.0616E+07, 2.0678E+07, 2.0741E+07, 2.0803E+07,
     +  2.0866E+07, 2.0928E+07, 2.0991E+07, 2.1054E+07, 2.1118E+07,
     + 2.1181E+07, 2.1244E+07, 2.1308E+07, 2.1372E+07, 2.1436E+07,
     +  2.1500E+07, 2.1564E+07, 2.1629E+07, 2.1693E+07, 2.1758E+07,
     + 2.1823E+07, 2.1888E+07, 2.1953E+07, 2.2018E+07, 2.2084E+07,
     +  2.2149E+07, 2.2215E+07, 2.2281E+07, 2.2347E+07, 2.2413E+07,
     + 2.2479E+07, 2.2546E+07, 2.2613E+07, 2.2680E+07, 2.2747E+07,
     +  2.2814E+07, 2.2881E+07, 2.2948E+07, 2.3016E+07, 2.3084E+07,
     + 2.3152E+07, 2.3220E+07, 2.3288E+07, 2.3357E+07, 2.3425E+07,
     +  2.3494E+07, 2.3563E+07, 2.3632E+07, 2.3701E+07, 2.3770E+07,
     + 2.3840E+07, 2.3910E+07, 2.3979E+07, 2.4049E+07, 2.4120E+07,
     +  2.4190E+07, 2.4260E+07, 2.4331E+07, 2.4402E+07, 2.4473E+07,
     + 2.4544E+07, 2.4615E+07, 2.4687E+07, 2.4758E+07, 2.4830E+07,
     +  2.4902E+07, 2.4974E+07, 2.5046E+07, 2.5119E+07, 2.5191E+07,
     + 2.5264E+07, 2.5337E+07, 2.5410E+07, 2.5483E+07, 2.5557E+07,
     +  2.5630E+07, 2.5704E+07, 2.5778E+07, 2.5852E+07, 2.5926E+07,
     + 2.6001E+07, 2.6075E+07, 2.6150E+07, 2.6225E+07, 2.6300E+07,
     +  2.6376E+07, 2.6451E+07, 2.6527E+07, 2.6602E+07, 2.6678E+07,
     + 2.6754E+07, 2.6831E+07, 2.6907E+07, 2.6984E+07, 2.7061E+07,
     +  2.7138E+07, 2.7215E+07, 2.7292E+07, 2.7370E+07, 2.7447E+07,
     + 2.7525E+07, 2.7603E+07, 2.7681E+07, 2.7760E+07, 2.7838E+07,
     +  2.7917E+07, 2.7996E+07, 2.8075E+07, 2.8154E+07, 2.8233E+07,
     + 2.8313E+07, 2.8393E+07, 2.8473E+07, 2.8553E+07, 2.8633E+07,
     +  2.8714E+07, 2.8794E+07, 2.8875E+07, 2.8956E+07, 2.9037E+07,
     + 2.9119E+07, 2.9200E+07, 2.9282E+07, 2.9364E+07, 2.9446E+07,
     +  2.9528E+07, 2.9610E+07, 2.9693E+07, 2.9776E+07, 2.9859E+07,
     + 2.9942E+07, 3.0025E+07, 3.0109E+07, 3.0193E+07, 3.0276E+07,
     +  3.0361E+07, 3.0445E+07, 3.0529E+07, 3.0614E+07, 3.0699E+07,
     + 3.0784E+07, 3.0869E+07, 3.0954E+07, 3.1040E+07, 3.1125E+07,
     +  3.1211E+07, 3.1297E+07, 3.1384E+07, 3.1470E+07, 3.1557E+07,
     + 3.1644E+07, 3.1731E+07, 3.1818E+07, 3.1905E+07, 3.1993E+07,
     +  3.2081E+07, 3.2169E+07, 3.2257E+07, 3.2345E+07, 3.2434E+07,
     + 3.2523E+07, 3.2612E+07, 3.2701E+07, 3.2790E+07, 3.2880E+07,
     +  3.2969E+07, 3.3059E+07, 3.3149E+07, 3.3240E+07, 3.3330E+07,
     + 3.3421E+07, 3.3512E+07, 3.3603E+07, 3.3694E+07, 3.3785E+07,
     +  3.3877E+07, 3.3969E+07, 3.4061E+07, 3.4153E+07, 3.4245E+07,
     + 3.4338E+07, 3.4431E+07, 3.4524E+07, 3.4617E+07, 3.4710E+07,
     +  3.4804E+07, 3.4898E+07, 3.4992E+07, 3.5086E+07, 3.5180E+07,
     + 3.5275E+07, 3.5370E+07, 3.5465E+07, 3.5560E+07, 3.5656E+07,
     +  3.5751E+07, 3.5847E+07, 3.5943E+07, 3.6039E+07, 3.6136E+07,
     + 3.6232E+07, 3.6329E+07, 3.6426E+07, 3.6523E+07, 3.6621E+07,
     +  3.6718E+07, 3.6816E+07, 3.6914E+07, 3.7012E+07, 3.7111E+07,
     + 3.7209E+07, 3.7308E+07, 3.7407E+07, 3.7507E+07, 3.7606E+07,
     +  3.7706E+07, 3.7806E+07, 3.7906E+07, 3.8006E+07, 3.8107E+07,
     + 3.8207E+07, 3.8308E+07, 3.8409E+07, 3.8511E+07, 3.8612E+07,
     +  3.8714E+07, 3.8816E+07, 3.8918E+07, 3.9021E+07, 3.9123E+07,
     + 3.9226E+07, 3.9329E+07, 3.9433E+07, 3.9536E+07, 3.9640E+07,
     +  3.9744E+07, 3.9848E+07, 3.9952E+07, 4.0057E+07, 4.0161E+07,
     + 4.0266E+07, 4.0371E+07, 4.0477E+07, 4.0582E+07, 4.0688E+07,
     +  4.0794E+07, 4.0901E+07, 4.1007E+07, 4.1114E+07, 4.1221E+07,
     + 4.1328E+07, 4.1435E+07, 4.1543E+07, 4.1651E+07, 4.1759E+07,
     +  4.1867E+07, 4.1975E+07, 4.2084E+07, 4.2193E+07, 4.2302E+07,
     + 4.2411E+07, 4.2521E+07, 4.2631E+07, 4.2741E+07, 4.2851E+07,
     +  4.2961E+07, 4.3072E+07, 4.3183E+07, 4.3294E+07, 4.3406E+07,
     + 4.3517E+07, 4.3629E+07, 4.3741E+07, 4.3853E+07, 4.3966E+07,
     +  4.4078E+07, 4.4191E+07, 4.4304E+07, 4.4418E+07, 4.4532E+07,
     + 4.4645E+07, 4.4759E+07, 4.4874E+07, 4.4988E+07, 4.5103E+07,
     +  4.5218E+07, 4.5333E+07, 4.5449E+07, 4.5564E+07, 4.5680E+07,
     + 4.5796E+07, 4.5913E+07, 4.6029E+07, 4.6146E+07, 4.6263E+07,
     +  4.6381E+07, 4.6498E+07, 4.6616E+07, 4.6734E+07, 4.6852E+07,
     + 4.6971E+07, 4.7089E+07, 4.7208E+07, 4.7328E+07, 4.7447E+07,
     +  4.7567E+07, 4.7687E+07, 4.7807E+07, 4.7927E+07, 4.8048E+07,
     + 4.8169E+07, 4.8290E+07, 4.8411E+07, 4.8533E+07, 4.8654E+07,
     +  4.8776E+07, 4.8899E+07, 4.9021E+07, 4.9144E+07, 4.9267E+07,
     + 4.9390E+07, 4.9514E+07, 4.9638E+07, 4.9762E+07, 4.9886E+07,
     +  5.0010E+07, 5.0135E+07, 5.0260E+07, 5.0385E+07, 5.0511E+07,
     + 5.0636E+07, 5.0762E+07, 5.0888E+07, 5.1015E+07, 5.1142E+07,
     +  5.1268E+07, 5.1396E+07, 5.1523E+07, 5.1651E+07, 5.1779E+07,
     + 5.1907E+07, 5.2035E+07, 5.2164E+07, 5.2293E+07, 5.2422E+07,
     +  5.2551E+07, 5.2681E+07, 5.2811E+07, 5.2941E+07, 5.3072E+07,
     + 5.3202E+07, 5.3333E+07, 5.3464E+07, 5.3596E+07, 5.3728E+07,
     +  5.3859E+07, 5.3992E+07, 5.4124E+07, 5.4257E+07, 5.4390E+07,
     + 5.4523E+07, 5.4657E+07, 5.4790E+07, 5.4924E+07, 5.5059E+07,
     +  5.5193E+07, 5.5328E+07, 5.5463E+07, 5.5598E+07, 5.5734E+07,
     + 5.5869E+07, 5.6006E+07, 5.6142E+07, 5.6278E+07, 5.6415E+07,
     +  5.6552E+07, 5.6690E+07, 5.6827E+07, 5.6965E+07, 5.7103E+07,
     + 5.7242E+07, 5.7380E+07, 5.7519E+07, 5.7659E+07, 5.7798E+07,
     +  5.7938E+07, 5.8078E+07, 5.8218E+07, 5.8359E+07, 5.8499E+07,
     + 5.8641E+07, 5.8782E+07, 5.8924E+07, 5.9065E+07, 5.9208E+07,
     +  5.9350E+07, 5.9493E+07, 5.9636E+07, 5.9779E+07, 5.9922E+07,
     + 6.0066E+07, 6.0210E+07, 6.0354E+07, 6.0499E+07, 6.0644E+07,
     +  6.0789E+07, 6.0934E+07, 6.1080E+07, 6.1226E+07, 6.1372E+07,
     + 6.1519E+07, 6.1665E+07, 6.1812E+07, 6.1960E+07, 6.2107E+07,
     +  6.2255E+07, 6.2403E+07, 6.2552E+07, 6.2700E+07, 6.2849E+07,
     + 6.2999E+07, 6.3148E+07, 6.3298E+07, 6.3448E+07, 6.3598E+07,
     +  6.3749E+07, 6.3900E+07, 6.4051E+07, 6.4203E+07, 6.4354E+07,
     + 6.4507E+07, 6.4659E+07, 6.4812E+07, 6.4964E+07, 6.5118E+07,
     +  6.5271E+07, 6.5425E+07, 6.5579E+07, 6.5733E+07, 6.5888E+07,
     + 6.6043E+07, 6.6198E+07, 6.6353E+07, 6.6509E+07, 6.6665E+07,
     +  6.6821E+07, 6.6978E+07, 6.7135E+07, 6.7292E+07, 6.7450E+07,
     + 6.7607E+07, 6.7765E+07, 6.7924E+07, 6.8082E+07, 6.8241E+07,
     +  6.8401E+07, 6.8560E+07, 6.8720E+07, 6.8880E+07, 6.9040E+07,
     + 6.9201E+07, 6.9362E+07, 6.9523E+07, 6.9685E+07, 6.9847E+07,
     +  7.0009E+07, 7.0171E+07, 7.0334E+07, 7.0497E+07, 7.0661E+07,
     + 7.0824E+07, 7.0988E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT(14,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       878
      data (QofT(15,J),J=1,2408)/ 5.3946E+00, 1.4038E+01, 2.5058E+01,
     +  3.8028E+01, 5.2690E+01, 6.8867E+01, 8.6430E+01, 1.0528E+02,
     + 1.2532E+02, 1.4651E+02, 1.6876E+02, 1.9205E+02, 2.1631E+02,
     +  2.4153E+02, 2.6765E+02, 2.9465E+02, 3.2251E+02, 3.5119E+02,
     + 3.8067E+02, 4.1094E+02, 4.7375E+02, 5.3948E+02, 6.0799E+02,
     +  6.7919E+02, 7.5297E+02, 8.2925E+02, 9.0795E+02, 9.8899E+02,
     + 1.0723E+03, 1.1579E+03, 1.2456E+03, 1.3354E+03, 1.4273E+03,
     +  1.5212E+03, 1.6171E+03, 1.7150E+03, 1.8147E+03, 1.9163E+03,
     + 2.0198E+03, 2.1250E+03, 2.2320E+03, 2.3408E+03, 2.4513E+03,
     +  2.5634E+03, 2.6773E+03, 2.7927E+03, 2.9098E+03, 3.0285E+03,
     + 3.1488E+03, 3.2707E+03, 3.3941E+03, 3.5190E+03, 3.6454E+03,
     +  3.7733E+03, 3.9026E+03, 4.0335E+03, 4.1658E+03, 4.2995E+03,
     + 4.4346E+03, 4.5711E+03, 4.7091E+03, 4.8483E+03, 4.9890E+03,
     +  5.1311E+03, 5.2744E+03, 5.4192E+03, 5.5652E+03, 5.7126E+03,
     + 5.8613E+03, 6.0113E+03, 6.1627E+03, 6.3153E+03, 6.4692E+03,
     +  6.6244E+03, 6.7809E+03, 6.9387E+03, 7.0977E+03, 7.2581E+03,
     + 7.4197E+03, 7.5826E+03, 7.7468E+03, 7.9121E+03, 8.0789E+03,
     +  8.2468E+03, 8.4161E+03, 8.5866E+03, 8.7584E+03, 8.9315E+03,
     + 9.1058E+03, 9.2814E+03, 9.4583E+03, 9.6365E+03, 9.8160E+03,
     +  9.9967E+03, 1.0179E+04, 1.0362E+04, 1.0547E+04, 1.0733E+04,
     + 1.0920E+04, 1.1109E+04, 1.1299E+04, 1.1490E+04, 1.1683E+04,
     +  1.1877E+04, 1.2072E+04, 1.2269E+04, 1.2467E+04, 1.2666E+04,
     + 1.2867E+04, 1.3069E+04, 1.3273E+04, 1.3478E+04, 1.3685E+04,
     +  1.3892E+04, 1.4102E+04, 1.4313E+04, 1.4525E+04, 1.4739E+04,
     + 1.4954E+04, 1.5171E+04, 1.5389E+04, 1.5609E+04, 1.5830E+04,
     +  1.6053E+04, 1.6278E+04, 1.6504E+04, 1.6731E+04, 1.6960E+04,
     + 1.7191E+04, 1.7423E+04, 1.7657E+04, 1.7893E+04, 1.8130E+04,
     +  1.8369E+04, 1.8610E+04, 1.8852E+04, 1.9096E+04, 1.9342E+04,
     + 1.9590E+04, 1.9839E+04, 2.0090E+04, 2.0343E+04, 2.0597E+04,
     +  2.0854E+04, 2.1112E+04, 2.1372E+04, 2.1634E+04, 2.1898E+04,
     + 2.2163E+04, 2.2431E+04, 2.2700E+04, 2.2971E+04, 2.3245E+04,
     +  2.3520E+04, 2.3797E+04, 2.4076E+04, 2.4358E+04, 2.4641E+04,
     + 2.4926E+04, 2.5213E+04, 2.5502E+04, 2.5794E+04, 2.6087E+04,
     +  2.6383E+04, 2.6680E+04, 2.6980E+04, 2.7282E+04, 2.7586E+04,
     + 2.7893E+04, 2.8201E+04, 2.8512E+04, 2.8825E+04, 2.9140E+04,
     +  2.9458E+04, 2.9778E+04, 3.0100E+04, 3.0424E+04, 3.0751E+04,
     + 3.1080E+04, 3.1412E+04, 3.1746E+04, 3.2082E+04, 3.2421E+04,
     +  3.2762E+04, 3.3106E+04, 3.3452E+04, 3.3800E+04, 3.4151E+04,
     + 3.4505E+04, 3.4861E+04, 3.5220E+04, 3.5581E+04, 3.5945E+04,
     +  3.6311E+04, 3.6680E+04, 3.7052E+04, 3.7426E+04, 3.7804E+04,
     + 3.8183E+04, 3.8566E+04, 3.8951E+04, 3.9339E+04, 3.9730E+04,
     +  4.0123E+04, 4.0520E+04, 4.0919E+04, 4.1321E+04, 4.1726E+04,
     + 4.2133E+04, 4.2544E+04, 4.2958E+04, 4.3374E+04, 4.3793E+04,
     +  4.4216E+04, 4.4642E+04, 4.5070E+04, 4.5502E+04, 4.5936E+04,
     + 4.6373E+04, 4.6814E+04, 4.7258E+04, 4.7705E+04, 4.8155E+04,
     +  4.8608E+04, 4.9064E+04, 4.9524E+04, 4.9987E+04, 5.0453E+04,
     + 5.0922E+04, 5.1395E+04, 5.1871E+04, 5.2350E+04, 5.2833E+04,
     +  5.3318E+04, 5.3807E+04, 5.4300E+04, 5.4796E+04, 5.5296E+04,
     + 5.5799E+04, 5.6305E+04, 5.6815E+04, 5.7328E+04, 5.7845E+04,
     +  5.8366E+04, 5.8890E+04, 5.9418E+04, 5.9949E+04, 6.0484E+04,
     + 6.1023E+04, 6.1565E+04, 6.2111E+04, 6.2661E+04, 6.3214E+04,
     +  6.3772E+04, 6.4332E+04, 6.4897E+04, 6.5466E+04, 6.6038E+04,
     + 6.6615E+04, 6.7195E+04, 6.7779E+04, 6.8367E+04, 6.8959E+04,
     +  6.9555E+04, 7.0155E+04, 7.0759E+04, 7.1367E+04, 7.1979E+04,
     + 7.2595E+04, 7.3216E+04, 7.3840E+04, 7.4469E+04, 7.5102E+04,
     +  7.5738E+04, 7.6379E+04, 7.7024E+04, 7.7674E+04, 7.8328E+04,
     + 7.8986E+04, 7.9649E+04, 8.0315E+04, 8.0987E+04, 8.1662E+04,
     +  8.2342E+04, 8.3027E+04, 8.3715E+04, 8.4409E+04, 8.5107E+04,
     + 8.5809E+04, 8.6516E+04, 8.7228E+04, 8.7944E+04, 8.8664E+04,
     +  8.9390E+04, 9.0120E+04, 9.0855E+04, 9.1594E+04, 9.2339E+04,
     + 9.3087E+04, 9.3841E+04, 9.4600E+04, 9.5363E+04, 9.6131E+04,
     +  9.6904E+04, 9.7682E+04, 9.8465E+04, 9.9253E+04, 1.0005E+05,
     + 1.0084E+05, 1.0165E+05, 1.0245E+05, 1.0327E+05, 1.0408E+05,
     +  1.0491E+05, 1.0574E+05, 1.0657E+05, 1.0741E+05, 1.0825E+05,
     + 1.0910E+05, 1.0996E+05, 1.1082E+05, 1.1168E+05, 1.1255E+05,
     +  1.1343E+05, 1.1431E+05, 1.1519E+05, 1.1609E+05, 1.1698E+05,
     + 1.1789E+05, 1.1879E+05, 1.1971E+05, 1.2063E+05, 1.2155E+05,
     +  1.2248E+05, 1.2342E+05, 1.2436E+05, 1.2531E+05, 1.2626E+05,
     + 1.2722E+05, 1.2819E+05, 1.2916E+05, 1.3014E+05, 1.3112E+05,
     +  1.3211E+05, 1.3310E+05, 1.3410E+05, 1.3511E+05, 1.3612E+05,
     + 1.3714E+05, 1.3816E+05, 1.3919E+05, 1.4023E+05, 1.4127E+05,
     +  1.4232E+05, 1.4338E+05, 1.4444E+05, 1.4550E+05, 1.4658E+05,
     + 1.4766E+05, 1.4874E+05, 1.4984E+05, 1.5094E+05, 1.5204E+05,
     +  1.5315E+05, 1.5427E+05, 1.5540E+05, 1.5653E+05, 1.5767E+05,
     + 1.5881E+05, 1.5996E+05, 1.6112E+05, 1.6228E+05, 1.6345E+05,
     +  1.6463E+05, 1.6582E+05, 1.6701E+05, 1.6820E+05, 1.6941E+05,
     + 1.7062E+05, 1.7184E+05, 1.7307E+05, 1.7430E+05, 1.7554E+05,
     +  1.7678E+05, 1.7804E+05, 1.7930E+05, 1.8057E+05, 1.8184E+05,
     + 1.8312E+05, 1.8441E+05, 1.8571E+05, 1.8701E+05, 1.8832E+05,
     +  1.8964E+05, 1.9097E+05, 1.9230E+05, 1.9364E+05, 1.9499E+05,
     + 1.9634E+05, 1.9770E+05, 1.9907E+05, 2.0045E+05, 2.0184E+05,
     +  2.0323E+05, 2.0463E+05, 2.0604E+05, 2.0745E+05, 2.0888E+05,
     + 2.1031E+05, 2.1175E+05, 2.1319E+05, 2.1465E+05, 2.1611E+05,
     +  2.1758E+05, 2.1906E+05, 2.2055E+05, 2.2204E+05, 2.2355E+05,
     + 2.2506E+05, 2.2658E+05, 2.2810E+05, 2.2964E+05, 2.3118E+05,
     +  2.3273E+05, 2.3429E+05, 2.3586E+05, 2.3744E+05, 2.3902E+05,
     + 2.4062E+05, 2.4222E+05, 2.4383E+05, 2.4545E+05, 2.4707E+05,
     +  2.4871E+05, 2.5035E+05, 2.5201E+05, 2.5367E+05, 2.5534E+05,
     + 2.5702E+05, 2.5871E+05, 2.6040E+05, 2.6211E+05, 2.6382E+05,
     +  2.6555E+05, 2.6728E+05, 2.6902E+05, 2.7077E+05, 2.7253E+05,
     + 2.7430E+05, 2.7608E+05, 2.7786E+05, 2.7966E+05, 2.8147E+05,
     +  2.8328E+05, 2.8510E+05, 2.8694E+05, 2.8878E+05, 2.9063E+05,
     + 2.9249E+05, 2.9436E+05, 2.9625E+05, 2.9813E+05, 3.0003E+05,
     +  3.0194E+05, 3.0386E+05, 3.0579E+05, 3.0773E+05, 3.0968E+05,
     + 3.1163E+05, 3.1360E+05, 3.1558E+05, 3.1757E+05, 3.1956E+05,
     +  3.2157E+05, 3.2359E+05, 3.2561E+05, 3.2765E+05, 3.2970E+05,
     + 3.3176E+05, 3.3382E+05, 3.3590E+05, 3.3799E+05, 3.4009E+05,
     +  3.4220E+05, 3.4432E+05, 3.4645E+05, 3.4859E+05, 3.5074E+05,
     + 3.5290E+05, 3.5507E+05, 3.5725E+05, 3.5945E+05, 3.6165E+05,
     +  3.6386E+05, 3.6609E+05, 3.6832E+05, 3.7057E+05, 3.7283E+05,
     + 3.7510E+05, 3.7738E+05, 3.7967E+05, 3.8197E+05, 3.8428E+05,
     +  3.8660E+05, 3.8894E+05, 3.9129E+05, 3.9364E+05, 3.9601E+05,
     + 3.9839E+05, 4.0078E+05, 4.0318E+05, 4.0560E+05, 4.0802E+05,
     +  4.1046E+05, 4.1291E+05, 4.1537E+05, 4.1784E+05, 4.2032E+05,
     + 4.2282E+05, 4.2533E+05, 4.2784E+05, 4.3037E+05, 4.3292E+05,
     +  4.3547E+05, 4.3804E+05, 4.4061E+05, 4.4320E+05, 4.4581E+05,
     + 4.4842E+05, 4.5105E+05, 4.5369E+05, 4.5634E+05, 4.5900E+05,
     +  4.6167E+05, 4.6436E+05, 4.6706E+05, 4.6978E+05, 4.7250E+05,
     + 4.7524E+05, 4.7799E+05, 4.8075E+05, 4.8353E+05, 4.8631E+05,
     +  4.8912E+05, 4.9193E+05, 4.9476E+05, 4.9760E+05, 5.0045E+05,
     + 5.0331E+05, 5.0619E+05, 5.0908E+05, 5.1199E+05, 5.1490E+05,
     +  5.1783E+05, 5.2078E+05, 5.2373E+05, 5.2671E+05, 5.2969E+05,
     + 5.3269E+05, 5.3570E+05, 5.3872E+05, 5.4176E+05, 5.4481E+05,
     +  5.4787E+05, 5.5095E+05, 5.5404E+05, 5.5715E+05, 5.6027E+05,
     + 5.6340E+05, 5.6655E+05, 5.6971E+05, 5.7288E+05, 5.7607E+05,
     +  5.7928E+05, 5.8250E+05, 5.8573E+05, 5.8897E+05, 5.9223E+05,
     + 5.9550E+05, 5.9879E+05, 6.0210E+05, 6.0541E+05, 6.0874E+05,
     +  6.1209E+05, 6.1545E+05, 6.1883E+05, 6.2222E+05, 6.2562E+05,
     + 6.2904E+05, 6.3248E+05, 6.3592E+05, 6.3939E+05, 6.4287E+05,
     +  6.4636E+05, 6.4987E+05, 6.5339E+05, 6.5693E+05, 6.6049E+05,
     + 6.6406E+05, 6.6764E+05, 6.7124E+05, 6.7486E+05, 6.7849E+05,
     +  6.8213E+05, 6.8579E+05, 6.8947E+05, 6.9316E+05, 6.9687E+05,
     + 7.0060E+05, 7.0434E+05, 7.0809E+05, 7.1186E+05, 7.1565E+05,
     +  7.1945E+05, 7.2327E+05, 7.2711E+05, 7.3096E+05, 7.3483E+05,
     + 7.3871E+05, 7.4261E+05, 7.4653E+05, 7.5046E+05, 7.5441E+05,
     +  7.5837E+05, 7.6235E+05, 7.6635E+05, 7.7037E+05, 7.7440E+05,
     + 7.7845E+05, 7.8251E+05, 7.8659E+05, 7.9069E+05, 7.9481E+05,
     +  7.9894E+05, 8.0309E+05, 8.0725E+05, 8.1144E+05, 8.1564E+05,
     + 8.1986E+05, 8.2409E+05, 8.2834E+05, 8.3261E+05, 8.3690E+05,
     +  8.4120E+05, 8.4552E+05, 8.4986E+05, 8.5422E+05, 8.5860E+05,
     + 8.6299E+05, 8.6740E+05, 8.7182E+05, 8.7627E+05, 8.8073E+05,
     +  8.8522E+05, 8.8971E+05, 8.9423E+05, 8.9877E+05, 9.0332E+05,
     + 9.0789E+05, 9.1248E+05, 9.1709E+05, 9.2172E+05, 9.2636E+05,
     +  9.3103E+05, 9.3571E+05, 9.4041E+05, 9.4513E+05, 9.4987E+05,
     + 9.5463E+05, 9.5940E+05, 9.6420E+05, 9.6901E+05, 9.7384E+05,
     +  9.7870E+05, 9.8357E+05, 9.8846E+05, 9.9337E+05, 9.9829E+05,
     + 1.0032E+06, 1.0082E+06, 1.0132E+06, 1.0182E+06, 1.0232E+06,
     +  1.0283E+06, 1.0333E+06, 1.0384E+06, 1.0435E+06, 1.0487E+06,
     + 1.0538E+06, 1.0590E+06, 1.0641E+06, 1.0693E+06, 1.0746E+06,
     +  1.0798E+06, 1.0851E+06, 1.0904E+06, 1.0957E+06, 1.1010E+06,
     + 1.1063E+06, 1.1117E+06, 1.1171E+06, 1.1225E+06, 1.1279E+06,
     +  1.1334E+06, 1.1388E+06, 1.1443E+06, 1.1498E+06, 1.1554E+06,
     + 1.1609E+06, 1.1665E+06, 1.1721E+06, 1.1777E+06, 1.1833E+06,
     +  1.1890E+06, 1.1947E+06, 1.2004E+06, 1.2061E+06, 1.2119E+06,
     + 1.2176E+06, 1.2234E+06, 1.2292E+06, 1.2351E+06, 1.2409E+06,
     +  1.2468E+06, 1.2527E+06, 1.2586E+06, 1.2645E+06, 1.2705E+06,
     + 1.2765E+06, 1.2825E+06, 1.2885E+06, 1.2946E+06, 1.3006E+06,
     +  1.3067E+06, 1.3129E+06, 1.3190E+06, 1.3252E+06, 1.3314E+06,
     + 1.3376E+06, 1.3438E+06, 1.3501E+06, 1.3563E+06, 1.3626E+06,
     +  1.3690E+06, 1.3753E+06, 1.3817E+06, 1.3881E+06, 1.3945E+06,
     + 1.4009E+06, 1.4074E+06, 1.4139E+06, 1.4204E+06, 1.4269E+06,
     +  1.4335E+06, 1.4401E+06, 1.4467E+06, 1.4533E+06, 1.4600E+06,
     + 1.4666E+06, 1.4734E+06, 1.4801E+06, 1.4868E+06, 1.4936E+06,
     +  1.5004E+06, 1.5072E+06, 1.5141E+06, 1.5209E+06, 1.5278E+06,
     + 1.5348E+06, 1.5417E+06, 1.5487E+06, 1.5557E+06, 1.5627E+06,
     +  1.5698E+06, 1.5768E+06, 1.5839E+06, 1.5910E+06, 1.5982E+06,
     + 1.6054E+06, 1.6126E+06, 1.6198E+06, 1.6270E+06, 1.6343E+06,
     +  1.6416E+06, 1.6489E+06, 1.6563E+06, 1.6637E+06, 1.6711E+06,
     + 1.6785E+06, 1.6859E+06, 1.6934E+06, 1.7009E+06, 1.7085E+06,
     +  1.7160E+06, 1.7236E+06, 1.7312E+06, 1.7389E+06, 1.7465E+06,
     + 1.7542E+06, 1.7619E+06, 1.7697E+06, 1.7775E+06, 1.7853E+06,
     +  1.7931E+06, 1.8009E+06, 1.8088E+06, 1.8167E+06, 1.8247E+06,
     + 1.8326E+06, 1.8406E+06, 1.8486E+06, 1.8567E+06, 1.8647E+06,
     +  1.8728E+06, 1.8810E+06, 1.8891E+06, 1.8973E+06, 1.9055E+06,
     + 1.9138E+06, 1.9220E+06, 1.9303E+06, 1.9386E+06, 1.9470E+06,
     +  1.9554E+06, 1.9638E+06, 1.9722E+06, 1.9807E+06, 1.9892E+06,
     + 1.9977E+06, 2.0063E+06, 2.0148E+06, 2.0234E+06, 2.0321E+06,
     +  2.0407E+06, 2.0494E+06, 2.0582E+06, 2.0669E+06, 2.0757E+06,
     + 2.0845E+06, 2.0934E+06, 2.1022E+06, 2.1111E+06, 2.1201E+06,
     +  2.1290E+06, 2.1380E+06, 2.1470E+06, 2.1561E+06, 2.1652E+06,
     + 2.1743E+06, 2.1834E+06, 2.1926E+06, 2.2018E+06, 2.2110E+06,
     +  2.2203E+06, 2.2296E+06, 2.2389E+06, 2.2483E+06, 2.2576E+06,
     + 2.2671E+06, 2.2765E+06, 2.2860E+06, 2.2955E+06, 2.3050E+06,
     +  2.3146E+06, 2.3242E+06, 2.3338E+06, 2.3435E+06, 2.3532E+06,
     + 2.3629E+06, 2.3727E+06, 2.3825E+06, 2.3923E+06, 2.4022E+06,
     +  2.4121E+06, 2.4220E+06, 2.4319E+06, 2.4419E+06, 2.4519E+06,
     + 2.4620E+06, 2.4721E+06, 2.4822E+06, 2.4923E+06, 2.5025E+06,
     +  2.5127E+06, 2.5229E+06, 2.5332E+06, 2.5435E+06, 2.5539E+06,
     + 2.5643E+06, 2.5747E+06, 2.5851E+06, 2.5956E+06, 2.6061E+06,
     +  2.6166E+06, 2.6272E+06, 2.6378E+06, 2.6485E+06, 2.6591E+06,
     + 2.6698E+06, 2.6806E+06, 2.6914E+06, 2.7022E+06, 2.7130E+06,
     +  2.7239E+06, 2.7348E+06, 2.7458E+06, 2.7568E+06, 2.7678E+06,
     + 2.7788E+06, 2.7899E+06, 2.8011E+06, 2.8122E+06, 2.8234E+06,
     +  2.8346E+06, 2.8459E+06, 2.8572E+06, 2.8685E+06, 2.8799E+06,
     + 2.8913E+06, 2.9027E+06, 2.9142E+06, 2.9257E+06, 2.9373E+06,
     +  2.9489E+06, 2.9605E+06, 2.9721E+06, 2.9838E+06, 2.9956E+06,
     + 3.0073E+06, 3.0191E+06, 3.0310E+06, 3.0428E+06, 3.0547E+06,
     +  3.0667E+06, 3.0787E+06, 3.0907E+06, 3.1028E+06, 3.1148E+06,
     + 3.1270E+06, 3.1391E+06, 3.1514E+06, 3.1636E+06, 3.1759E+06,
     +  3.1882E+06, 3.2006E+06, 3.2129E+06, 3.2254E+06, 3.2378E+06,
     + 3.2503E+06, 3.2629E+06, 3.2755E+06, 3.2881E+06, 3.3008E+06,
     +  3.3135E+06, 3.3262E+06, 3.3390E+06, 3.3518E+06, 3.3646E+06,
     + 3.3775E+06, 3.3904E+06, 3.4034E+06, 3.4164E+06, 3.4295E+06,
     +  3.4426E+06, 3.4557E+06, 3.4688E+06, 3.4820E+06, 3.4953E+06,
     + 3.5086E+06, 3.5219E+06, 3.5353E+06, 3.5487E+06, 3.5621E+06,
     +  3.5756E+06, 3.5891E+06, 3.6027E+06, 3.6163E+06, 3.6299E+06,
     + 3.6436E+06, 3.6573E+06, 3.6711E+06, 3.6849E+06, 3.6987E+06,
     +  3.7126E+06, 3.7266E+06, 3.7405E+06, 3.7545E+06, 3.7686E+06,
     + 3.7827E+06, 3.7968E+06, 3.8110E+06, 3.8252E+06, 3.8395E+06,
     +  3.8538E+06, 3.8681E+06, 3.8825E+06, 3.8969E+06, 3.9114E+06,
     + 3.9259E+06, 3.9404E+06, 3.9550E+06, 3.9697E+06, 3.9844E+06,
     +  3.9991E+06, 4.0139E+06, 4.0287E+06, 4.0435E+06, 4.0584E+06,
     + 4.0734E+06, 4.0883E+06, 4.1034E+06, 4.1184E+06, 4.1335E+06,
     +  4.1487E+06, 4.1639E+06, 4.1791E+06, 4.1944E+06, 4.2098E+06,
     + 4.2251E+06, 4.2406E+06, 4.2560E+06, 4.2715E+06, 4.2871E+06,
     +  4.3027E+06, 4.3183E+06, 4.3340E+06, 4.3497E+06, 4.3655E+06,
     + 4.3813E+06, 4.3972E+06, 4.4131E+06, 4.4291E+06, 4.4451E+06,
     +  4.4611E+06, 4.4772E+06, 4.4933E+06, 4.5095E+06, 4.5258E+06,
     + 4.5420E+06, 4.5583E+06, 4.5747E+06, 4.5911E+06, 4.6076E+06,
     +  4.6241E+06, 4.6406E+06, 4.6572E+06, 4.6739E+06, 4.6906E+06,
     + 4.7073E+06, 4.7241E+06, 4.7410E+06, 4.7578E+06, 4.7748E+06,
     +  4.7917E+06, 4.8088E+06, 4.8258E+06, 4.8429E+06, 4.8601E+06,
     + 4.8773E+06, 4.8946E+06, 4.9119E+06, 4.9293E+06, 4.9467E+06,
     +  4.9641E+06, 4.9816E+06, 4.9992E+06, 5.0168E+06, 5.0344E+06,
     + 5.0521E+06, 5.0699E+06, 5.0877E+06, 5.1055E+06, 5.1234E+06,
     +  5.1413E+06, 5.1593E+06, 5.1774E+06, 5.1955E+06, 5.2136E+06,
     + 5.2318E+06, 5.2500E+06, 5.2683E+06, 5.2867E+06, 5.3051E+06,
     +  5.3235E+06, 5.3420E+06, 5.3605E+06, 5.3791E+06, 5.3978E+06,
     + 5.4165E+06, 5.4352E+06, 5.4540E+06, 5.4729E+06, 5.4918E+06,
     +  5.5107E+06, 5.5297E+06, 5.5488E+06, 5.5679E+06, 5.5870E+06,
     + 5.6062E+06, 5.6255E+06, 5.6448E+06, 5.6642E+06, 5.6836E+06,
     +  5.7030E+06, 5.7226E+06, 5.7421E+06, 5.7618E+06, 5.7814E+06,
     + 5.8012E+06, 5.8210E+06, 5.8408E+06, 5.8607E+06, 5.8806E+06,
     +  5.9006E+06, 5.9207E+06, 5.9408E+06, 5.9609E+06, 5.9811E+06,
     + 6.0014E+06, 6.0217E+06, 6.0421E+06, 6.0625E+06, 6.0830E+06,
     +  6.1035E+06, 6.1241E+06, 6.1448E+06, 6.1655E+06, 6.1862E+06,
     + 6.2070E+06, 6.2279E+06, 6.2488E+06, 6.2698E+06, 6.2908E+06,
     +  6.3119E+06, 6.3331E+06, 6.3543E+06, 6.3755E+06, 6.3968E+06,
     + 6.4182E+06, 6.4396E+06, 6.4611E+06, 6.4826E+06, 6.5042E+06,
     +  6.5258E+06, 6.5476E+06, 6.5693E+06, 6.5911E+06, 6.6130E+06,
     + 6.6349E+06, 6.6569E+06, 6.6790E+06, 6.7011E+06, 6.7232E+06,
     +  6.7454E+06, 6.7677E+06, 6.7900E+06, 6.8124E+06, 6.8349E+06,
     + 6.8574E+06, 6.8799E+06, 6.9026E+06, 6.9252E+06, 6.9480E+06,
     +  6.9708E+06, 6.9936E+06, 7.0166E+06, 7.0395E+06, 7.0626E+06,
     + 7.0857E+06, 7.1088E+06, 7.1320E+06, 7.1553E+06, 7.1786E+06,
     +  7.2020E+06, 7.2255E+06, 7.2490E+06, 7.2726E+06, 7.2962E+06,
     + 7.3199E+06, 7.3436E+06, 7.3674E+06, 7.3913E+06, 7.4152E+06,
     +  7.4392E+06, 7.4633E+06, 7.4874E+06, 7.5116E+06, 7.5358E+06,
     + 7.5601E+06, 7.5845E+06, 7.6089E+06, 7.6334E+06, 7.6579E+06,
     +  7.6826E+06, 7.7072E+06, 7.7320E+06, 7.7568E+06, 7.7816E+06,
     + 7.8065E+06, 7.8315E+06, 7.8566E+06, 7.8817E+06, 7.9069E+06,
     +  7.9321E+06, 7.9574E+06, 7.9828E+06, 8.0082E+06, 8.0337E+06,
     + 8.0592E+06, 8.0848E+06, 8.1105E+06, 8.1363E+06, 8.1621E+06,
     +  8.1880E+06, 8.2139E+06, 8.2399E+06, 8.2660E+06, 8.2921E+06,
     + 8.3183E+06, 8.3446E+06, 8.3709E+06, 8.3973E+06, 8.4237E+06,
     +  8.4503E+06, 8.4769E+06, 8.5035E+06, 8.5302E+06, 8.5570E+06,
     + 8.5839E+06, 8.6108E+06, 8.6378E+06, 8.6648E+06, 8.6919E+06,
     +  8.7191E+06, 8.7464E+06, 8.7737E+06, 8.8011E+06, 8.8285E+06,
     + 8.8561E+06, 8.8837E+06, 8.9113E+06, 8.9390E+06, 8.9668E+06,
     +  8.9947E+06, 9.0226E+06, 9.0506E+06, 9.0787E+06, 9.1068E+06,
     + 9.1350E+06, 9.1633E+06, 9.1916E+06, 9.2200E+06, 9.2485E+06,
     +  9.2770E+06, 9.3056E+06, 9.3343E+06, 9.3631E+06, 9.3919E+06,
     + 9.4208E+06, 9.4498E+06, 9.4788E+06, 9.5079E+06, 9.5371E+06,
     +  9.5663E+06, 9.5956E+06, 9.6250E+06, 9.6545E+06, 9.6840E+06,
     + 9.7136E+06, 9.7432E+06, 9.7730E+06, 9.8028E+06, 9.8327E+06,
     +  9.8626E+06, 9.8926E+06, 9.9227E+06, 9.9529E+06, 9.9831E+06,
     + 1.0013E+07, 1.0044E+07, 1.0074E+07, 1.0105E+07, 1.0135E+07,
     +  1.0166E+07, 1.0197E+07, 1.0228E+07, 1.0259E+07, 1.0290E+07,
     + 1.0321E+07, 1.0352E+07, 1.0383E+07, 1.0414E+07, 1.0445E+07,
     +  1.0477E+07, 1.0508E+07, 1.0540E+07, 1.0572E+07, 1.0603E+07,
     + 1.0635E+07, 1.0667E+07, 1.0699E+07, 1.0731E+07, 1.0763E+07,
     +  1.0795E+07, 1.0827E+07, 1.0860E+07, 1.0892E+07, 1.0924E+07,
     + 1.0957E+07, 1.0990E+07, 1.1022E+07, 1.1055E+07, 1.1088E+07,
     +  1.1121E+07, 1.1154E+07, 1.1187E+07, 1.1220E+07, 1.1253E+07,
     + 1.1287E+07, 1.1320E+07, 1.1353E+07, 1.1387E+07, 1.1421E+07,
     +  1.1454E+07, 1.1488E+07, 1.1522E+07, 1.1556E+07, 1.1590E+07,
     + 1.1624E+07, 1.1658E+07, 1.1692E+07, 1.1727E+07, 1.1761E+07,
     +  1.1795E+07, 1.1830E+07, 1.1865E+07, 1.1899E+07, 1.1934E+07,
     + 1.1969E+07, 1.2004E+07, 1.2039E+07, 1.2074E+07, 1.2109E+07,
     +  1.2145E+07, 1.2180E+07, 1.2215E+07, 1.2251E+07, 1.2287E+07,
     + 1.2322E+07, 1.2358E+07, 1.2394E+07, 1.2430E+07, 1.2466E+07,
     +  1.2502E+07, 1.2538E+07, 1.2574E+07, 1.2611E+07, 1.2647E+07,
     + 1.2684E+07, 1.2720E+07, 1.2757E+07, 1.2794E+07, 1.2830E+07,
     +  1.2867E+07, 1.2904E+07, 1.2941E+07, 1.2979E+07, 1.3016E+07,
     + 1.3053E+07, 1.3091E+07, 1.3128E+07, 1.3166E+07, 1.3203E+07,
     +  1.3241E+07, 1.3279E+07, 1.3317E+07, 1.3355E+07, 1.3393E+07,
     + 1.3431E+07, 1.3469E+07, 1.3508E+07, 1.3546E+07, 1.3585E+07,
     +  1.3623E+07, 1.3662E+07, 1.3701E+07, 1.3740E+07, 1.3779E+07,
     + 1.3818E+07, 1.3857E+07, 1.3896E+07, 1.3935E+07, 1.3975E+07,
     +  1.4014E+07, 1.4054E+07, 1.4093E+07, 1.4133E+07, 1.4173E+07,
     + 1.4213E+07, 1.4253E+07, 1.4293E+07, 1.4333E+07, 1.4374E+07,
     +  1.4414E+07, 1.4454E+07, 1.4495E+07, 1.4536E+07, 1.4576E+07,
     + 1.4617E+07, 1.4658E+07, 1.4699E+07, 1.4740E+07, 1.4781E+07,
     +  1.4822E+07, 1.4864E+07, 1.4905E+07, 1.4947E+07, 1.4988E+07,
     + 1.5030E+07, 1.5072E+07, 1.5114E+07, 1.5156E+07, 1.5198E+07,
     +  1.5240E+07, 1.5282E+07, 1.5325E+07, 1.5367E+07, 1.5410E+07,
     + 1.5452E+07, 1.5495E+07, 1.5538E+07, 1.5581E+07, 1.5624E+07,
     +  1.5667E+07, 1.5710E+07, 1.5753E+07, 1.5797E+07, 1.5840E+07,
     + 1.5884E+07, 1.5927E+07, 1.5971E+07, 1.6015E+07, 1.6059E+07,
     +  1.6103E+07, 1.6147E+07, 1.6191E+07, 1.6236E+07, 1.6280E+07,
     + 1.6325E+07, 1.6369E+07, 1.6414E+07, 1.6459E+07, 1.6504E+07,
     +  1.6549E+07, 1.6594E+07, 1.6639E+07, 1.6684E+07, 1.6730E+07,
     + 1.6775E+07, 1.6821E+07, 1.6866E+07, 1.6912E+07, 1.6958E+07,
     +  1.7004E+07, 1.7050E+07, 1.7096E+07, 1.7142E+07, 1.7189E+07,
     + 1.7235E+07, 1.7282E+07, 1.7328E+07, 1.7375E+07, 1.7422E+07,
     +  1.7469E+07, 1.7516E+07, 1.7563E+07, 1.7610E+07, 1.7658E+07,
     + 1.7705E+07, 1.7753E+07, 1.7800E+07, 1.7848E+07, 1.7896E+07,
     +  1.7944E+07, 1.7992E+07, 1.8040E+07, 1.8088E+07, 1.8137E+07,
     + 1.8185E+07, 1.8234E+07, 1.8282E+07, 1.8331E+07, 1.8380E+07,
     +  1.8429E+07, 1.8478E+07, 1.8527E+07, 1.8577E+07, 1.8626E+07,
     + 1.8675E+07, 1.8725E+07, 1.8775E+07, 1.8824E+07, 1.8874E+07,
     +  1.8924E+07, 1.8974E+07, 1.9025E+07, 1.9075E+07, 1.9125E+07,
     + 1.9176E+07, 1.9227E+07, 1.9277E+07, 1.9328E+07, 1.9379E+07,
     +  1.9430E+07, 1.9481E+07, 1.9533E+07, 1.9584E+07, 1.9635E+07,
     + 1.9687E+07, 1.9739E+07, 1.9790E+07, 1.9842E+07, 1.9894E+07,
     +  1.9946E+07, 1.9999E+07, 2.0051E+07, 2.0103E+07, 2.0156E+07,
     + 2.0209E+07, 2.0261E+07, 2.0314E+07, 2.0367E+07, 2.0420E+07,
     +  2.0473E+07, 2.0527E+07, 2.0580E+07, 2.0634E+07, 2.0687E+07,
     + 2.0741E+07, 2.0795E+07, 2.0849E+07, 2.0903E+07, 2.0957E+07,
     +  2.1011E+07, 2.1066E+07, 2.1120E+07, 2.1175E+07, 2.1230E+07,
     + 2.1284E+07, 2.1339E+07, 2.1394E+07, 2.1450E+07, 2.1505E+07,
     +  2.1560E+07, 2.1616E+07, 2.1671E+07, 2.1727E+07, 2.1783E+07,
     + 2.1839E+07, 2.1895E+07, 2.1951E+07, 2.2008E+07, 2.2064E+07,
     +  2.2120E+07, 2.2177E+07, 2.2234E+07, 2.2291E+07, 2.2348E+07,
     + 2.2405E+07, 2.2462E+07, 2.2519E+07, 2.2577E+07, 2.2634E+07,
     +  2.2692E+07, 2.2750E+07, 2.2808E+07, 2.2866E+07, 2.2924E+07,
     + 2.2982E+07, 2.3040E+07, 2.3099E+07, 2.3158E+07, 2.3216E+07,
     +  2.3275E+07, 2.3334E+07, 2.3393E+07, 2.3452E+07, 2.3512E+07,
     + 2.3571E+07, 2.3631E+07, 2.3690E+07, 2.3750E+07, 2.3810E+07,
     +  2.3870E+07, 2.3930E+07, 2.3990E+07, 2.4051E+07, 2.4111E+07,
     + 2.4172E+07, 2.4232E+07, 2.4293E+07, 2.4354E+07, 2.4415E+07,
     +  2.4477E+07, 2.4538E+07, 2.4599E+07, 2.4661E+07, 2.4723E+07,
     + 2.4784E+07, 2.4846E+07, 2.4908E+07, 2.4971E+07, 2.5033E+07,
     +  2.5095E+07, 2.5158E+07, 2.5221E+07, 2.5283E+07, 2.5346E+07,
     + 2.5409E+07, 2.5472E+07, 2.5536E+07, 2.5599E+07, 2.5663E+07,
     +  2.5726E+07, 2.5790E+07, 2.5854E+07, 2.5918E+07, 2.5982E+07,
     + 2.6046E+07, 2.6111E+07, 2.6175E+07, 2.6240E+07, 2.6305E+07,
     +  2.6370E+07, 2.6435E+07, 2.6500E+07, 2.6565E+07, 2.6631E+07,
     + 2.6696E+07, 2.6762E+07, 2.6827E+07, 2.6893E+07, 2.6959E+07,
     +  2.7026E+07, 2.7092E+07, 2.7158E+07, 2.7225E+07, 2.7292E+07,
     + 2.7358E+07, 2.7425E+07, 2.7492E+07, 2.7560E+07, 2.7627E+07,
     +  2.7694E+07, 2.7762E+07, 2.7830E+07, 2.7897E+07, 2.7965E+07,
     + 2.8034E+07, 2.8102E+07, 2.8170E+07, 2.8239E+07, 2.8307E+07,
     +  2.8376E+07, 2.8445E+07, 2.8514E+07, 2.8583E+07, 2.8652E+07,
     + 2.8722E+07, 2.8791E+07, 2.8861E+07, 2.8931E+07, 2.9001E+07,
     +  2.9071E+07, 2.9141E+07, 2.9211E+07, 2.9282E+07, 2.9352E+07,
     + 2.9423E+07, 2.9494E+07, 2.9565E+07, 2.9636E+07, 2.9707E+07,
     +  2.9779E+07, 2.9850E+07, 2.9922E+07, 2.9994E+07, 3.0066E+07,
     + 3.0138E+07, 3.0210E+07, 3.0283E+07, 3.0355E+07, 3.0428E+07,
     +  3.0500E+07, 3.0573E+07, 3.0646E+07, 3.0720E+07, 3.0793E+07,
     + 3.0866E+07, 3.0940E+07, 3.1014E+07, 3.1087E+07, 3.1161E+07,
     +  3.1236E+07, 3.1310E+07, 3.1384E+07, 3.1459E+07, 3.1534E+07,
     + 3.1608E+07, 3.1683E+07, 3.1759E+07, 3.1834E+07, 3.1909E+07,
     +  3.1985E+07, 3.2060E+07, 3.2136E+07, 3.2212E+07, 3.2288E+07,
     + 3.2365E+07, 3.2441E+07, 3.2517E+07, 3.2594E+07, 3.2671E+07,
     +  3.2748E+07, 3.2825E+07, 3.2902E+07, 3.2980E+07, 3.3057E+07,
     + 3.3135E+07, 3.3213E+07, 3.3291E+07, 3.3369E+07, 3.3447E+07,
     +  3.3525E+07, 3.3604E+07, 3.3682E+07, 3.3761E+07, 3.3840E+07,
     + 3.3919E+07, 3.3999E+07, 3.4078E+07, 3.4158E+07, 3.4237E+07,
     +  3.4317E+07, 3.4397E+07, 3.4477E+07, 3.4558E+07, 3.4638E+07,
     + 3.4719E+07, 3.4799E+07, 3.4880E+07, 3.4961E+07, 3.5042E+07,
     +  3.5124E+07, 3.5205E+07, 3.5287E+07, 3.5368E+07, 3.5450E+07,
     + 3.5532E+07, 3.5615E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT(15,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       778
      data (QofT(16,J),J=1,2408)/ 6.2959E+01, 1.6349E+02, 2.9160E+02,
     +  4.4235E+02, 6.1276E+02, 8.0077E+02, 1.0049E+03, 1.2239E+03,
     + 1.4569E+03, 1.7030E+03, 1.9616E+03, 2.2322E+03, 2.5142E+03,
     +  2.8072E+03, 3.1107E+03, 3.4245E+03, 3.7481E+03, 4.0814E+03,
     + 4.4241E+03, 4.7758E+03, 5.5056E+03, 6.2693E+03, 7.0654E+03,
     +  7.8927E+03, 8.7500E+03, 9.6363E+03, 1.0551E+04, 1.1492E+04,
     + 1.2461E+04, 1.3455E+04, 1.4474E+04, 1.5518E+04, 1.6585E+04,
     +  1.7677E+04, 1.8791E+04, 1.9928E+04, 2.1086E+04, 2.2267E+04,
     + 2.3469E+04, 2.4692E+04, 2.5935E+04, 2.7199E+04, 2.8483E+04,
     +  2.9786E+04, 3.1109E+04, 3.2450E+04, 3.3811E+04, 3.5190E+04,
     + 3.6588E+04, 3.8004E+04, 3.9437E+04, 4.0888E+04, 4.2357E+04,
     +  4.3843E+04, 4.5346E+04, 4.6867E+04, 4.8403E+04, 4.9957E+04,
     + 5.1527E+04, 5.3113E+04, 5.4716E+04, 5.6334E+04, 5.7968E+04,
     +  5.9618E+04, 6.1284E+04, 6.2965E+04, 6.4662E+04, 6.6374E+04,
     + 6.8102E+04, 6.9845E+04, 7.1603E+04, 7.3376E+04, 7.5164E+04,
     +  7.6967E+04, 7.8785E+04, 8.0617E+04, 8.2465E+04, 8.4327E+04,
     + 8.6204E+04, 8.8096E+04, 9.0002E+04, 9.1924E+04, 9.3860E+04,
     +  9.5810E+04, 9.7775E+04, 9.9755E+04, 1.0175E+05, 1.0376E+05,
     + 1.0578E+05, 1.0782E+05, 1.0988E+05, 1.1194E+05, 1.1403E+05,
     +  1.1613E+05, 1.1824E+05, 1.2037E+05, 1.2251E+05, 1.2467E+05,
     + 1.2684E+05, 1.2903E+05, 1.3123E+05, 1.3345E+05, 1.3569E+05,
     +  1.3794E+05, 1.4021E+05, 1.4249E+05, 1.4479E+05, 1.4710E+05,
     + 1.4943E+05, 1.5177E+05, 1.5414E+05, 1.5651E+05, 1.5891E+05,
     +  1.6132E+05, 1.6375E+05, 1.6619E+05, 1.6865E+05, 1.7113E+05,
     + 1.7362E+05, 1.7614E+05, 1.7867E+05, 1.8121E+05, 1.8378E+05,
     +  1.8636E+05, 1.8896E+05, 1.9158E+05, 1.9421E+05, 1.9687E+05,
     + 1.9954E+05, 2.0223E+05, 2.0494E+05, 2.0767E+05, 2.1042E+05,
     +  2.1319E+05, 2.1597E+05, 2.1878E+05, 2.2160E+05, 2.2445E+05,
     + 2.2731E+05, 2.3020E+05, 2.3310E+05, 2.3603E+05, 2.3897E+05,
     +  2.4194E+05, 2.4493E+05, 2.4793E+05, 2.5096E+05, 2.5401E+05,
     + 2.5709E+05, 2.6018E+05, 2.6330E+05, 2.6643E+05, 2.6959E+05,
     +  2.7278E+05, 2.7598E+05, 2.7921E+05, 2.8246E+05, 2.8573E+05,
     + 2.8903E+05, 2.9235E+05, 2.9569E+05, 2.9906E+05, 3.0245E+05,
     +  3.0587E+05, 3.0930E+05, 3.1277E+05, 3.1626E+05, 3.1977E+05,
     + 3.2331E+05, 3.2688E+05, 3.3046E+05, 3.3408E+05, 3.3772E+05,
     +  3.4139E+05, 3.4508E+05, 3.4880E+05, 3.5255E+05, 3.5632E+05,
     + 3.6012E+05, 3.6394E+05, 3.6780E+05, 3.7168E+05, 3.7559E+05,
     +  3.7953E+05, 3.8349E+05, 3.8748E+05, 3.9150E+05, 3.9555E+05,
     + 3.9964E+05, 4.0375E+05, 4.0788E+05, 4.1205E+05, 4.1625E+05,
     +  4.2047E+05, 4.2473E+05, 4.2902E+05, 4.3333E+05, 4.3768E+05,
     + 4.4206E+05, 4.4647E+05, 4.5091E+05, 4.5539E+05, 4.5989E+05,
     +  4.6443E+05, 4.6900E+05, 4.7360E+05, 4.7823E+05, 4.8290E+05,
     + 4.8760E+05, 4.9233E+05, 4.9710E+05, 5.0190E+05, 5.0673E+05,
     +  5.1160E+05, 5.1650E+05, 5.2144E+05, 5.2641E+05, 5.3141E+05,
     + 5.3645E+05, 5.4153E+05, 5.4664E+05, 5.5179E+05, 5.5697E+05,
     +  5.6219E+05, 5.6744E+05, 5.7274E+05, 5.7807E+05, 5.8343E+05,
     + 5.8884E+05, 5.9428E+05, 5.9976E+05, 6.0527E+05, 6.1083E+05,
     +  6.1643E+05, 6.2206E+05, 6.2772E+05, 6.3344E+05, 6.3919E+05,
     + 6.4497E+05, 6.5080E+05, 6.5667E+05, 6.6258E+05, 6.6853E+05,
     +  6.7452E+05, 6.8055E+05, 6.8662E+05, 6.9273E+05, 6.9889E+05,
     + 7.0509E+05, 7.1133E+05, 7.1761E+05, 7.2393E+05, 7.3030E+05,
     +  7.3671E+05, 7.4316E+05, 7.4966E+05, 7.5620E+05, 7.6278E+05,
     + 7.6941E+05, 7.7608E+05, 7.8280E+05, 7.8956E+05, 7.9637E+05,
     +  8.0322E+05, 8.1012E+05, 8.1706E+05, 8.2406E+05, 8.3109E+05,
     + 8.3818E+05, 8.4531E+05, 8.5249E+05, 8.5971E+05, 8.6698E+05,
     +  8.7430E+05, 8.8167E+05, 8.8909E+05, 8.9656E+05, 9.0407E+05,
     + 9.1163E+05, 9.1925E+05, 9.2691E+05, 9.3462E+05, 9.4239E+05,
     +  9.5020E+05, 9.5807E+05, 9.6598E+05, 9.7394E+05, 9.8196E+05,
     + 9.9003E+05, 9.9815E+05, 1.0063E+06, 1.0146E+06, 1.0228E+06,
     +  1.0312E+06, 1.0396E+06, 1.0480E+06, 1.0565E+06, 1.0650E+06,
     + 1.0736E+06, 1.0823E+06, 1.0910E+06, 1.0998E+06, 1.1086E+06,
     +  1.1175E+06, 1.1264E+06, 1.1354E+06, 1.1444E+06, 1.1535E+06,
     + 1.1627E+06, 1.1719E+06, 1.1812E+06, 1.1905E+06, 1.1999E+06,
     +  1.2094E+06, 1.2189E+06, 1.2285E+06, 1.2381E+06, 1.2478E+06,
     + 1.2575E+06, 1.2673E+06, 1.2772E+06, 1.2871E+06, 1.2971E+06,
     +  1.3072E+06, 1.3173E+06, 1.3274E+06, 1.3377E+06, 1.3480E+06,
     + 1.3583E+06, 1.3688E+06, 1.3793E+06, 1.3898E+06, 1.4004E+06,
     +  1.4111E+06, 1.4218E+06, 1.4327E+06, 1.4435E+06, 1.4545E+06,
     + 1.4655E+06, 1.4765E+06, 1.4877E+06, 1.4989E+06, 1.5102E+06,
     +  1.5215E+06, 1.5329E+06, 1.5444E+06, 1.5559E+06, 1.5675E+06,
     + 1.5792E+06, 1.5910E+06, 1.6028E+06, 1.6147E+06, 1.6266E+06,
     +  1.6386E+06, 1.6507E+06, 1.6629E+06, 1.6752E+06, 1.6875E+06,
     + 1.6999E+06, 1.7123E+06, 1.7248E+06, 1.7374E+06, 1.7501E+06,
     +  1.7629E+06, 1.7757E+06, 1.7886E+06, 1.8015E+06, 1.8146E+06,
     + 1.8277E+06, 1.8409E+06, 1.8542E+06, 1.8675E+06, 1.8810E+06,
     +  1.8945E+06, 1.9080E+06, 1.9217E+06, 1.9354E+06, 1.9492E+06,
     + 1.9631E+06, 1.9771E+06, 1.9911E+06, 2.0053E+06, 2.0195E+06,
     +  2.0338E+06, 2.0481E+06, 2.0626E+06, 2.0771E+06, 2.0917E+06,
     + 2.1064E+06, 2.1212E+06, 2.1360E+06, 2.1510E+06, 2.1660E+06,
     +  2.1811E+06, 2.1963E+06, 2.2115E+06, 2.2269E+06, 2.2423E+06,
     + 2.2579E+06, 2.2735E+06, 2.2892E+06, 2.3050E+06, 2.3208E+06,
     +  2.3368E+06, 2.3528E+06, 2.3690E+06, 2.3852E+06, 2.4015E+06,
     + 2.4179E+06, 2.4344E+06, 2.4509E+06, 2.4676E+06, 2.4844E+06,
     +  2.5012E+06, 2.5181E+06, 2.5352E+06, 2.5523E+06, 2.5695E+06,
     + 2.5868E+06, 2.6042E+06, 2.6217E+06, 2.6393E+06, 2.6569E+06,
     +  2.6747E+06, 2.6926E+06, 2.7105E+06, 2.7286E+06, 2.7467E+06,
     + 2.7650E+06, 2.7833E+06, 2.8018E+06, 2.8203E+06, 2.8389E+06,
     +  2.8577E+06, 2.8765E+06, 2.8954E+06, 2.9145E+06, 2.9336E+06,
     + 2.9528E+06, 2.9721E+06, 2.9916E+06, 3.0111E+06, 3.0307E+06,
     +  3.0505E+06, 3.0703E+06, 3.0902E+06, 3.1103E+06, 3.1304E+06,
     + 3.1507E+06, 3.1710E+06, 3.1915E+06, 3.2120E+06, 3.2327E+06,
     +  3.2535E+06, 3.2743E+06, 3.2953E+06, 3.3164E+06, 3.3376E+06,
     + 3.3589E+06, 3.3803E+06, 3.4018E+06, 3.4235E+06, 3.4452E+06,
     +  3.4671E+06, 3.4890E+06, 3.5111E+06, 3.5333E+06, 3.5556E+06,
     + 3.5780E+06, 3.6005E+06, 3.6231E+06, 3.6458E+06, 3.6687E+06,
     +  3.6917E+06, 3.7147E+06, 3.7379E+06, 3.7612E+06, 3.7847E+06,
     + 3.8082E+06, 3.8319E+06, 3.8557E+06, 3.8795E+06, 3.9036E+06,
     +  3.9277E+06, 3.9519E+06, 3.9763E+06, 4.0008E+06, 4.0254E+06,
     + 4.0501E+06, 4.0750E+06, 4.0999E+06, 4.1250E+06, 4.1502E+06,
     +  4.1756E+06, 4.2010E+06, 4.2266E+06, 4.2523E+06, 4.2781E+06,
     + 4.3040E+06, 4.3301E+06, 4.3563E+06, 4.3827E+06, 4.4091E+06,
     +  4.4357E+06, 4.4624E+06, 4.4892E+06, 4.5162E+06, 4.5433E+06,
     + 4.5705E+06, 4.5978E+06, 4.6253E+06, 4.6529E+06, 4.6806E+06,
     +  4.7085E+06, 4.7365E+06, 4.7647E+06, 4.7929E+06, 4.8213E+06,
     + 4.8498E+06, 4.8785E+06, 4.9073E+06, 4.9362E+06, 4.9653E+06,
     +  4.9945E+06, 5.0239E+06, 5.0533E+06, 5.0830E+06, 5.1127E+06,
     + 5.1426E+06, 5.1726E+06, 5.2028E+06, 5.2331E+06, 5.2636E+06,
     +  5.2942E+06, 5.3249E+06, 5.3558E+06, 5.3868E+06, 5.4179E+06,
     + 5.4492E+06, 5.4807E+06, 5.5123E+06, 5.5440E+06, 5.5758E+06,
     +  5.6079E+06, 5.6400E+06, 5.6723E+06, 5.7048E+06, 5.7374E+06,
     + 5.7702E+06, 5.8031E+06, 5.8361E+06, 5.8693E+06, 5.9027E+06,
     +  5.9361E+06, 5.9698E+06, 6.0036E+06, 6.0376E+06, 6.0717E+06,
     + 6.1059E+06, 6.1403E+06, 6.1749E+06, 6.2096E+06, 6.2445E+06,
     +  6.2795E+06, 6.3147E+06, 6.3500E+06, 6.3855E+06, 6.4212E+06,
     + 6.4570E+06, 6.4930E+06, 6.5291E+06, 6.5654E+06, 6.6018E+06,
     +  6.6384E+06, 6.6752E+06, 6.7121E+06, 6.7492E+06, 6.7864E+06,
     + 6.8238E+06, 6.8614E+06, 6.8992E+06, 6.9371E+06, 6.9752E+06,
     +  7.0134E+06, 7.0518E+06, 7.0903E+06, 7.1291E+06, 7.1680E+06,
     + 7.2071E+06, 7.2463E+06, 7.2857E+06, 7.3253E+06, 7.3650E+06,
     +  7.4049E+06, 7.4450E+06, 7.4853E+06, 7.5257E+06, 7.5663E+06,
     + 7.6071E+06, 7.6481E+06, 7.6892E+06, 7.7305E+06, 7.7720E+06,
     +  7.8136E+06, 7.8555E+06, 7.8975E+06, 7.9396E+06, 7.9820E+06,
     + 8.0246E+06, 8.0673E+06, 8.1102E+06, 8.1533E+06, 8.1965E+06,
     +  8.2400E+06, 8.2836E+06, 8.3274E+06, 8.3714E+06, 8.4156E+06,
     + 8.4599E+06, 8.5045E+06, 8.5492E+06, 8.5941E+06, 8.6392E+06,
     +  8.6845E+06, 8.7300E+06, 8.7757E+06, 8.8215E+06, 8.8676E+06,
     + 8.9138E+06, 8.9602E+06, 9.0068E+06, 9.0536E+06, 9.1006E+06,
     +  9.1478E+06, 9.1952E+06, 9.2428E+06, 9.2906E+06, 9.3385E+06,
     + 9.3867E+06, 9.4351E+06, 9.4836E+06, 9.5324E+06, 9.5813E+06,
     +  9.6305E+06, 9.6798E+06, 9.7294E+06, 9.7791E+06, 9.8291E+06,
     + 9.8793E+06, 9.9296E+06, 9.9802E+06, 1.0031E+07, 1.0082E+07,
     +  1.0133E+07, 1.0184E+07, 1.0236E+07, 1.0288E+07, 1.0340E+07,
     + 1.0392E+07, 1.0444E+07, 1.0497E+07, 1.0550E+07, 1.0603E+07,
     +  1.0656E+07, 1.0710E+07, 1.0763E+07, 1.0817E+07, 1.0871E+07,
     + 1.0926E+07, 1.0980E+07, 1.1035E+07, 1.1090E+07, 1.1145E+07,
     +  1.1200E+07, 1.1256E+07, 1.1312E+07, 1.1368E+07, 1.1424E+07,
     + 1.1481E+07, 1.1537E+07, 1.1594E+07, 1.1651E+07, 1.1709E+07,
     +  1.1766E+07, 1.1824E+07, 1.1882E+07, 1.1941E+07, 1.1999E+07,
     + 1.2058E+07, 1.2117E+07, 1.2176E+07, 1.2235E+07, 1.2295E+07,
     +  1.2355E+07, 1.2415E+07, 1.2475E+07, 1.2536E+07, 1.2596E+07,
     + 1.2657E+07, 1.2719E+07, 1.2780E+07, 1.2842E+07, 1.2904E+07,
     +  1.2966E+07, 1.3028E+07, 1.3091E+07, 1.3154E+07, 1.3217E+07,
     + 1.3281E+07, 1.3344E+07, 1.3408E+07, 1.3472E+07, 1.3536E+07,
     +  1.3601E+07, 1.3666E+07, 1.3731E+07, 1.3796E+07, 1.3862E+07,
     + 1.3928E+07, 1.3994E+07, 1.4060E+07, 1.4126E+07, 1.4193E+07,
     +  1.4260E+07, 1.4328E+07, 1.4395E+07, 1.4463E+07, 1.4531E+07,
     + 1.4599E+07, 1.4668E+07, 1.4737E+07, 1.4806E+07, 1.4875E+07,
     +  1.4945E+07, 1.5014E+07, 1.5084E+07, 1.5155E+07, 1.5225E+07,
     + 1.5296E+07, 1.5367E+07, 1.5439E+07, 1.5510E+07, 1.5582E+07,
     +  1.5655E+07, 1.5727E+07, 1.5800E+07, 1.5873E+07, 1.5946E+07,
     + 1.6019E+07, 1.6093E+07, 1.6167E+07, 1.6241E+07, 1.6316E+07,
     +  1.6391E+07, 1.6466E+07, 1.6541E+07, 1.6617E+07, 1.6693E+07,
     + 1.6769E+07, 1.6846E+07, 1.6922E+07, 1.6999E+07, 1.7077E+07,
     +  1.7154E+07, 1.7232E+07, 1.7310E+07, 1.7389E+07, 1.7467E+07,
     + 1.7546E+07, 1.7625E+07, 1.7705E+07, 1.7785E+07, 1.7865E+07,
     +  1.7945E+07, 1.8026E+07, 1.8107E+07, 1.8188E+07, 1.8270E+07,
     + 1.8351E+07, 1.8434E+07, 1.8516E+07, 1.8599E+07, 1.8682E+07,
     +  1.8765E+07, 1.8848E+07, 1.8932E+07, 1.9016E+07, 1.9101E+07,
     + 1.9186E+07, 1.9271E+07, 1.9356E+07, 1.9442E+07, 1.9528E+07,
     +  1.9614E+07, 1.9700E+07, 1.9787E+07, 1.9874E+07, 1.9962E+07,
     + 2.0049E+07, 2.0137E+07, 2.0226E+07, 2.0314E+07, 2.0403E+07,
     +  2.0493E+07, 2.0582E+07, 2.0672E+07, 2.0762E+07, 2.0853E+07,
     + 2.0944E+07, 2.1035E+07, 2.1126E+07, 2.1218E+07, 2.1310E+07,
     +  2.1402E+07, 2.1495E+07, 2.1588E+07, 2.1681E+07, 2.1775E+07,
     + 2.1869E+07, 2.1963E+07, 2.2058E+07, 2.2153E+07, 2.2248E+07,
     +  2.2343E+07, 2.2439E+07, 2.2536E+07, 2.2632E+07, 2.2729E+07,
     + 2.2826E+07, 2.2924E+07, 2.3022E+07, 2.3120E+07, 2.3218E+07,
     +  2.3317E+07, 2.3416E+07, 2.3516E+07, 2.3615E+07, 2.3716E+07,
     + 2.3816E+07, 2.3917E+07, 2.4018E+07, 2.4120E+07, 2.4221E+07,
     +  2.4324E+07, 2.4426E+07, 2.4529E+07, 2.4632E+07, 2.4736E+07,
     + 2.4840E+07, 2.4944E+07, 2.5048E+07, 2.5153E+07, 2.5259E+07,
     +  2.5364E+07, 2.5470E+07, 2.5576E+07, 2.5683E+07, 2.5790E+07,
     + 2.5897E+07, 2.6005E+07, 2.6113E+07, 2.6222E+07, 2.6330E+07,
     +  2.6440E+07, 2.6549E+07, 2.6659E+07, 2.6769E+07, 2.6880E+07,
     + 2.6991E+07, 2.7102E+07, 2.7213E+07, 2.7325E+07, 2.7438E+07,
     +  2.7551E+07, 2.7664E+07, 2.7777E+07, 2.7891E+07, 2.8005E+07,
     + 2.8120E+07, 2.8235E+07, 2.8350E+07, 2.8465E+07, 2.8582E+07,
     +  2.8698E+07, 2.8815E+07, 2.8932E+07, 2.9049E+07, 2.9167E+07,
     + 2.9286E+07, 2.9404E+07, 2.9523E+07, 2.9643E+07, 2.9762E+07,
     +  2.9883E+07, 3.0003E+07, 3.0124E+07, 3.0245E+07, 3.0367E+07,
     + 3.0489E+07, 3.0612E+07, 3.0735E+07, 3.0858E+07, 3.0981E+07,
     +  3.1106E+07, 3.1230E+07, 3.1355E+07, 3.1480E+07, 3.1606E+07,
     + 3.1732E+07, 3.1858E+07, 3.1985E+07, 3.2112E+07, 3.2240E+07,
     +  3.2368E+07, 3.2496E+07, 3.2625E+07, 3.2754E+07, 3.2883E+07,
     + 3.3013E+07, 3.3144E+07, 3.3275E+07, 3.3406E+07, 3.3537E+07,
     +  3.3669E+07, 3.3802E+07, 3.3935E+07, 3.4068E+07, 3.4202E+07,
     + 3.4336E+07, 3.4470E+07, 3.4605E+07, 3.4740E+07, 3.4876E+07,
     +  3.5012E+07, 3.5149E+07, 3.5286E+07, 3.5423E+07, 3.5561E+07,
     + 3.5699E+07, 3.5838E+07, 3.5977E+07, 3.6117E+07, 3.6257E+07,
     +  3.6397E+07, 3.6538E+07, 3.6679E+07, 3.6821E+07, 3.6963E+07,
     + 3.7105E+07, 3.7248E+07, 3.7392E+07, 3.7535E+07, 3.7680E+07,
     +  3.7824E+07, 3.7969E+07, 3.8115E+07, 3.8261E+07, 3.8408E+07,
     + 3.8554E+07, 3.8702E+07, 3.8849E+07, 3.8998E+07, 3.9146E+07,
     +  3.9295E+07, 3.9445E+07, 3.9595E+07, 3.9745E+07, 3.9896E+07,
     + 4.0048E+07, 4.0200E+07, 4.0352E+07, 4.0504E+07, 4.0658E+07,
     +  4.0811E+07, 4.0965E+07, 4.1120E+07, 4.1275E+07, 4.1430E+07,
     + 4.1586E+07, 4.1743E+07, 4.1899E+07, 4.2057E+07, 4.2214E+07,
     +  4.2373E+07, 4.2531E+07, 4.2691E+07, 4.2850E+07, 4.3010E+07,
     + 4.3171E+07, 4.3332E+07, 4.3493E+07, 4.3655E+07, 4.3818E+07,
     +  4.3981E+07, 4.4144E+07, 4.4308E+07, 4.4472E+07, 4.4637E+07,
     + 4.4803E+07, 4.4968E+07, 4.5135E+07, 4.5301E+07, 4.5469E+07,
     +  4.5636E+07, 4.5805E+07, 4.5973E+07, 4.6143E+07, 4.6312E+07,
     + 4.6483E+07, 4.6653E+07, 4.6824E+07, 4.6996E+07, 4.7168E+07,
     +  4.7341E+07, 4.7514E+07, 4.7688E+07, 4.7862E+07, 4.8036E+07,
     + 4.8212E+07, 4.8387E+07, 4.8563E+07, 4.8740E+07, 4.8917E+07,
     +  4.9095E+07, 4.9273E+07, 4.9452E+07, 4.9631E+07, 4.9811E+07,
     + 4.9991E+07, 5.0172E+07, 5.0353E+07, 5.0535E+07, 5.0717E+07,
     +  5.0900E+07, 5.1083E+07, 5.1267E+07, 5.1451E+07, 5.1636E+07,
     + 5.1822E+07, 5.2007E+07, 5.2194E+07, 5.2381E+07, 5.2568E+07,
     +  5.2756E+07, 5.2945E+07, 5.3134E+07, 5.3324E+07, 5.3514E+07,
     + 5.3704E+07, 5.3896E+07, 5.4087E+07, 5.4280E+07, 5.4472E+07,
     +  5.4666E+07, 5.4860E+07, 5.5054E+07, 5.5249E+07, 5.5445E+07,
     + 5.5641E+07, 5.5837E+07, 5.6035E+07, 5.6232E+07, 5.6430E+07,
     +  5.6629E+07, 5.6829E+07, 5.7028E+07, 5.7229E+07, 5.7430E+07,
     + 5.7632E+07, 5.7834E+07, 5.8036E+07, 5.8240E+07, 5.8443E+07,
     +  5.8648E+07, 5.8853E+07, 5.9058E+07, 5.9264E+07, 5.9471E+07,
     + 5.9678E+07, 5.9886E+07, 6.0094E+07, 6.0303E+07, 6.0512E+07,
     +  6.0722E+07, 6.0933E+07, 6.1144E+07, 6.1356E+07, 6.1568E+07,
     + 6.1781E+07, 6.1995E+07, 6.2209E+07, 6.2423E+07, 6.2639E+07,
     +  6.2854E+07, 6.3071E+07, 6.3288E+07, 6.3505E+07, 6.3723E+07,
     + 6.3942E+07, 6.4161E+07, 6.4381E+07, 6.4602E+07, 6.4823E+07,
     +  6.5045E+07, 6.5267E+07, 6.5490E+07, 6.5713E+07, 6.5937E+07,
     + 6.6162E+07, 6.6387E+07, 6.6613E+07, 6.6840E+07, 6.7067E+07,
     +  6.7295E+07, 6.7523E+07, 6.7752E+07, 6.7982E+07, 6.8212E+07,
     + 6.8442E+07, 6.8674E+07, 6.8906E+07, 6.9138E+07, 6.9372E+07,
     +  6.9605E+07, 6.9840E+07, 7.0075E+07, 7.0311E+07, 7.0547E+07,
     + 7.0784E+07, 7.1022E+07, 7.1260E+07, 7.1499E+07, 7.1738E+07,
     +  7.1978E+07, 7.2219E+07, 7.2460E+07, 7.2702E+07, 7.2945E+07,
     + 7.3188E+07, 7.3432E+07, 7.3677E+07, 7.3922E+07, 7.4168E+07,
     +  7.4414E+07, 7.4661E+07, 7.4909E+07, 7.5157E+07, 7.5406E+07,
     + 7.5656E+07, 7.5906E+07, 7.6157E+07, 7.6409E+07, 7.6661E+07,
     +  7.6914E+07, 7.7168E+07, 7.7422E+07, 7.7677E+07, 7.7933E+07,
     + 7.8189E+07, 7.8446E+07, 7.8703E+07, 7.8962E+07, 7.9220E+07,
     +  7.9480E+07, 7.9740E+07, 8.0001E+07, 8.0263E+07, 8.0525E+07,
     + 8.0788E+07, 8.1051E+07, 8.1316E+07, 8.1581E+07, 8.1846E+07,
     +  8.2113E+07, 8.2380E+07, 8.2647E+07, 8.2916E+07, 8.3185E+07,
     + 8.3454E+07, 8.3725E+07, 8.3996E+07, 8.4268E+07, 8.4540E+07,
     +  8.4813E+07, 8.5087E+07, 8.5362E+07, 8.5637E+07, 8.5913E+07,
     + 8.6190E+07, 8.6467E+07, 8.6745E+07, 8.7024E+07, 8.7303E+07,
     +  8.7583E+07, 8.7864E+07, 8.8146E+07, 8.8428E+07, 8.8711E+07,
     + 8.8995E+07, 8.9279E+07, 8.9564E+07, 8.9850E+07, 9.0137E+07,
     +  9.0424E+07, 9.0712E+07, 9.1001E+07, 9.1290E+07, 9.1580E+07,
     + 9.1871E+07, 9.2163E+07, 9.2455E+07, 9.2748E+07, 9.3042E+07,
     +  9.3337E+07, 9.3632E+07, 9.3928E+07, 9.4225E+07, 9.4522E+07,
     + 9.4820E+07, 9.5119E+07, 9.5419E+07, 9.5720E+07, 9.6021E+07,
     +  9.6323E+07, 9.6625E+07, 9.6929E+07, 9.7233E+07, 9.7538E+07,
     + 9.7844E+07, 9.8150E+07, 9.8457E+07, 9.8765E+07, 9.9074E+07,
     +  9.9383E+07, 9.9693E+07, 1.0000E+08, 1.0032E+08, 1.0063E+08,
     + 1.0094E+08, 1.0126E+08, 1.0157E+08, 1.0189E+08, 1.0220E+08,
     +  1.0252E+08, 1.0284E+08, 1.0316E+08, 1.0348E+08, 1.0380E+08,
     + 1.0412E+08, 1.0444E+08, 1.0476E+08, 1.0508E+08, 1.0541E+08,
     +  1.0573E+08, 1.0606E+08, 1.0639E+08, 1.0671E+08, 1.0704E+08,
     + 1.0737E+08, 1.0770E+08, 1.0803E+08, 1.0836E+08, 1.0869E+08,
     +  1.0903E+08, 1.0936E+08, 1.0969E+08, 1.1003E+08, 1.1037E+08,
     + 1.1070E+08, 1.1104E+08, 1.1138E+08, 1.1172E+08, 1.1206E+08,
     +  1.1240E+08, 1.1274E+08, 1.1308E+08, 1.1343E+08, 1.1377E+08,
     + 1.1412E+08, 1.1446E+08, 1.1481E+08, 1.1515E+08, 1.1550E+08,
     +  1.1585E+08, 1.1620E+08, 1.1655E+08, 1.1690E+08, 1.1726E+08,
     + 1.1761E+08, 1.1796E+08, 1.1832E+08, 1.1868E+08, 1.1903E+08,
     +  1.1939E+08, 1.1975E+08, 1.2011E+08, 1.2047E+08, 1.2083E+08,
     + 1.2119E+08, 1.2155E+08, 1.2192E+08, 1.2228E+08, 1.2264E+08,
     +  1.2301E+08, 1.2338E+08, 1.2374E+08, 1.2411E+08, 1.2448E+08,
     + 1.2485E+08, 1.2522E+08, 1.2560E+08, 1.2597E+08, 1.2634E+08,
     +  1.2672E+08, 1.2709E+08, 1.2747E+08, 1.2785E+08, 1.2822E+08,
     + 1.2860E+08, 1.2898E+08, 1.2936E+08, 1.2975E+08, 1.3013E+08,
     +  1.3051E+08, 1.3090E+08, 1.3128E+08, 1.3167E+08, 1.3205E+08,
     + 1.3244E+08, 1.3283E+08, 1.3322E+08, 1.3361E+08, 1.3400E+08,
     +  1.3439E+08, 1.3479E+08, 1.3518E+08, 1.3558E+08, 1.3597E+08,
     + 1.3637E+08, 1.3677E+08, 1.3717E+08, 1.3757E+08, 1.3797E+08,
     +  1.3837E+08, 1.3877E+08, 1.3917E+08, 1.3958E+08, 1.3998E+08,
     + 1.4039E+08, 1.4080E+08, 1.4120E+08, 1.4161E+08, 1.4202E+08,
     +  1.4243E+08, 1.4284E+08, 1.4326E+08, 1.4367E+08, 1.4408E+08,
     + 1.4450E+08, 1.4492E+08, 1.4533E+08, 1.4575E+08, 1.4617E+08,
     +  1.4659E+08, 1.4701E+08, 1.4743E+08, 1.4786E+08, 1.4828E+08,
     + 1.4871E+08, 1.4913E+08, 1.4956E+08, 1.4999E+08, 1.5041E+08,
     +  1.5084E+08, 1.5127E+08, 1.5171E+08, 1.5214E+08, 1.5257E+08,
     + 1.5301E+08, 1.5344E+08, 1.5388E+08, 1.5432E+08, 1.5475E+08,
     +  1.5519E+08, 1.5563E+08, 1.5608E+08, 1.5652E+08, 1.5696E+08,
     + 1.5741E+08, 1.5785E+08, 1.5830E+08, 1.5874E+08, 1.5919E+08,
     +  1.5964E+08, 1.6009E+08, 1.6054E+08, 1.6099E+08, 1.6145E+08,
     + 1.6190E+08, 1.6236E+08, 1.6281E+08, 1.6327E+08, 1.6373E+08,
     +  1.6419E+08, 1.6465E+08, 1.6511E+08, 1.6557E+08, 1.6604E+08,
     + 1.6650E+08, 1.6696E+08, 1.6743E+08, 1.6790E+08, 1.6837E+08,
     +  1.6884E+08, 1.6931E+08, 1.6978E+08, 1.7025E+08, 1.7072E+08,
     + 1.7120E+08, 1.7167E+08, 1.7215E+08, 1.7263E+08, 1.7311E+08,
     +  1.7359E+08, 1.7407E+08, 1.7455E+08, 1.7503E+08, 1.7552E+08,
     + 1.7600E+08, 1.7649E+08, 1.7698E+08, 1.7746E+08, 1.7795E+08,
     +  1.7844E+08, 1.7893E+08, 1.7943E+08, 1.7992E+08, 1.8041E+08,
     + 1.8091E+08, 1.8141E+08, 1.8191E+08, 1.8240E+08, 1.8290E+08,
     +  1.8340E+08, 1.8391E+08, 1.8441E+08, 1.8491E+08, 1.8542E+08,
     + 1.8593E+08, 1.8643E+08, 1.8694E+08, 1.8745E+08, 1.8796E+08,
     +  1.8847E+08, 1.8899E+08, 1.8950E+08, 1.9002E+08, 1.9053E+08,
     + 1.9105E+08, 1.9157E+08, 1.9209E+08, 1.9261E+08, 1.9313E+08,
     +  1.9365E+08, 1.9418E+08, 1.9470E+08, 1.9523E+08, 1.9576E+08,
     + 1.9628E+08, 1.9681E+08, 1.9734E+08, 1.9788E+08, 1.9841E+08,
     +  1.9894E+08, 1.9948E+08, 2.0001E+08, 2.0055E+08, 2.0109E+08,
     + 2.0163E+08, 2.0217E+08, 2.0271E+08, 2.0326E+08, 2.0380E+08,
     +  2.0435E+08, 2.0489E+08, 2.0544E+08, 2.0599E+08, 2.0654E+08,
     + 2.0709E+08, 2.0764E+08, 2.0820E+08, 2.0875E+08, 2.0931E+08,
     +  2.0986E+08, 2.1042E+08, 2.1098E+08, 2.1154E+08, 2.1210E+08,
     + 2.1267E+08, 2.1323E+08, 2.1380E+08, 2.1436E+08, 2.1493E+08,
     +  2.1550E+08, 2.1607E+08, 2.1664E+08, 2.1721E+08, 2.1779E+08,
     + 2.1836E+08, 2.1894E+08, 2.1951E+08, 2.2009E+08, 2.2067E+08,
     +  2.2125E+08, 2.2183E+08, 2.2242E+08, 2.2300E+08, 2.2359E+08,
     + 2.2417E+08, 2.2476E+08, 2.2535E+08, 2.2594E+08, 2.2653E+08,
     +  2.2712E+08, 2.2772E+08, 2.2831E+08, 2.2891E+08, 2.2951E+08,
     + 2.3011E+08, 2.3071E+08, 2.3131E+08, 2.3191E+08, 2.3251E+08,
     +  2.3312E+08, 2.3372E+08, 2.3433E+08, 2.3494E+08, 2.3555E+08,
     + 2.3616E+08, 2.3677E+08, 2.3739E+08, 2.3800E+08, 2.3862E+08,
     +  2.3924E+08, 2.3985E+08, 2.4047E+08, 2.4110E+08, 2.4172E+08,
     + 2.4234E+08, 2.4297E+08, 2.4359E+08, 2.4422E+08, 2.4485E+08,
     +  2.4548E+08, 2.4611E+08, 2.4674E+08, 2.4738E+08, 2.4801E+08,
     + 2.4865E+08, 2.4929E+08, 2.4993E+08, 2.5057E+08, 2.5121E+08,
     +  2.5185E+08, 2.5249E+08, 2.5314E+08, 2.5379E+08, 2.5443E+08,
     + 2.5508E+08, 2.5573E+08, 2.5639E+08, 2.5704E+08, 2.5769E+08,
     +  2.5835E+08, 2.5901E+08, 2.5967E+08, 2.6033E+08, 2.6099E+08,
     + 2.6165E+08, 2.6231E+08, 2.6298E+08, 2.6364E+08, 2.6431E+08,
     +  2.6498E+08, 2.6565E+08, 2.6632E+08, 2.6700E+08, 2.6767E+08,
     + 2.6835E+08, 2.6902E+08, 2.6970E+08, 2.7038E+08, 2.7106E+08,
     +  2.7175E+08, 2.7243E+08, 2.7312E+08, 2.7380E+08, 2.7449E+08,
     + 2.7518E+08, 2.7587E+08, 2.7656E+08, 2.7725E+08, 2.7795E+08,
     +  2.7865E+08, 2.7934E+08, 2.8004E+08, 2.8074E+08, 2.8144E+08,
     + 2.8215E+08, 2.8285E+08, 2.8356E+08, 2.8426E+08, 2.8497E+08,
     +  2.8568E+08, 2.8639E+08, 2.8711E+08, 2.8782E+08, 2.8854E+08,
     + 2.8925E+08, 2.8997E+08, 2.9069E+08, 2.9141E+08, 2.9213E+08,
     +  2.9286E+08, 2.9358E+08, 2.9431E+08, 2.9504E+08, 2.9577E+08,
     + 2.9650E+08, 2.9723E+08, 2.9796E+08, 2.9870E+08, 2.9944E+08,
     +  3.0017E+08, 3.0091E+08, 3.0165E+08, 3.0240E+08, 3.0314E+08,
     + 3.0388E+08, 3.0463E+08, 3.0538E+08, 3.0613E+08, 3.0688E+08,
     +  3.0763E+08, 3.0839E+08, 3.0914E+08, 3.0990E+08, 3.1066E+08,
     + 3.1142E+08, 3.1218E+08, 3.1294E+08, 3.1370E+08, 3.1447E+08,
     +  3.1524E+08, 3.1601E+08, 3.1677E+08, 3.1755E+08, 3.1832E+08,
     + 3.1909E+08, 3.1987E+08, 3.2065E+08, 3.2143E+08, 3.2221E+08,
     +  3.2299E+08, 3.2377E+08, 3.2456E+08, 3.2534E+08, 3.2613E+08,
     + 3.2692E+08, 3.2771E+08, 3.2850E+08, 3.2930E+08, 3.3009E+08,
     +  3.3089E+08, 3.3169E+08, 3.3249E+08, 3.3329E+08, 3.3409E+08,
     + 3.3489E+08, 3.3570E+08, 3.3651E+08, 3.3732E+08, 3.3813E+08,
     +  3.3894E+08, 3.3975E+08, 3.4057E+08, 3.4138E+08, 3.4220E+08,
     + 3.4302E+08, 3.4384E+08, 3.4467E+08, 3.4549E+08, 3.4632E+08,
     +  3.4714E+08, 3.4797E+08, 3.4880E+08, 3.4963E+08, 3.5047E+08,
     + 3.5130E+08, 3.5214E+08, 3.5298E+08, 3.5382E+08, 3.5466E+08,
     +  3.5550E+08, 3.5635E+08, 3.5719E+08, 3.5804E+08, 3.5889E+08,
     + 3.5974E+08, 3.6059E+08, 3.6145E+08, 3.6230E+08, 3.6316E+08,
     +  3.6402E+08, 3.6488E+08, 3.6574E+08, 3.6661E+08, 3.6747E+08,
     + 3.6834E+08, 3.6921E+08, 3.7008E+08, 3.7095E+08, 3.7182E+08,
     +  3.7270E+08, 3.7357E+08, 3.7445E+08, 3.7533E+08, 3.7621E+08,
     + 3.7710E+08, 3.7798E+08, 3.7887E+08, 3.7976E+08, 3.8064E+08,
     +  3.8154E+08, 3.8243E+08, 3.8332E+08, 3.8422E+08, 3.8512E+08,
     + 3.8602E+08, 3.8692E+08, 3.8782E+08, 3.8873E+08, 3.8963E+08,
     +  3.9054E+08, 3.9145E+08, 3.9236E+08, 3.9327E+08, 3.9419E+08,
     + 3.9510E+08, 3.9602E+08, 3.9694E+08, 3.9786E+08, 3.9878E+08,
     +  3.9971E+08, 4.0063E+08, 4.0156E+08, 4.0249E+08, 4.0342E+08,
     + 4.0436E+08, 4.0529E+08, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT(16,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       787
      data (QofT(17,J),J=1,2408)/ 3.1266E+01, 8.1008E+01, 1.4438E+02,
     +  2.1895E+02, 3.0324E+02, 3.9623E+02, 4.9718E+02, 6.0549E+02,
     + 7.2071E+02, 8.4244E+02, 9.7035E+02, 1.1042E+03, 1.2436E+03,
     +  1.3885E+03, 1.5386E+03, 1.6938E+03, 1.8538E+03, 2.0186E+03,
     + 2.1881E+03, 2.3620E+03, 2.7229E+03, 3.1006E+03, 3.4943E+03,
     +  3.9033E+03, 4.3273E+03, 4.7656E+03, 5.2177E+03, 5.6834E+03,
     + 6.1622E+03, 6.6537E+03, 7.1577E+03, 7.6739E+03, 8.2019E+03,
     +  8.7415E+03, 9.2925E+03, 9.8546E+03, 1.0428E+04, 1.1011E+04,
     + 1.1606E+04, 1.2211E+04, 1.2825E+04, 1.3450E+04, 1.4085E+04,
     +  1.4730E+04, 1.5384E+04, 1.6047E+04, 1.6720E+04, 1.7402E+04,
     + 1.8093E+04, 1.8793E+04, 1.9502E+04, 2.0220E+04, 2.0946E+04,
     +  2.1681E+04, 2.2424E+04, 2.3176E+04, 2.3936E+04, 2.4704E+04,
     + 2.5480E+04, 2.6264E+04, 2.7057E+04, 2.7857E+04, 2.8665E+04,
     +  2.9481E+04, 3.0305E+04, 3.1136E+04, 3.1975E+04, 3.2822E+04,
     + 3.3676E+04, 3.4538E+04, 3.5407E+04, 3.6284E+04, 3.7168E+04,
     +  3.8059E+04, 3.8958E+04, 3.9864E+04, 4.0778E+04, 4.1698E+04,
     + 4.2626E+04, 4.3561E+04, 4.4504E+04, 4.5454E+04, 4.6411E+04,
     +  4.7375E+04, 4.8347E+04, 4.9326E+04, 5.0312E+04, 5.1305E+04,
     + 5.2306E+04, 5.3314E+04, 5.4329E+04, 5.5351E+04, 5.6382E+04,
     +  5.7419E+04, 5.8463E+04, 5.9515E+04, 6.0575E+04, 6.1642E+04,
     + 6.2716E+04, 6.3798E+04, 6.4888E+04, 6.5985E+04, 6.7089E+04,
     +  6.8201E+04, 6.9322E+04, 7.0450E+04, 7.1585E+04, 7.2729E+04,
     + 7.3880E+04, 7.5039E+04, 7.6207E+04, 7.7382E+04, 7.8565E+04,
     +  7.9757E+04, 8.0957E+04, 8.2165E+04, 8.3381E+04, 8.4607E+04,
     + 8.5840E+04, 8.7081E+04, 8.8332E+04, 8.9590E+04, 9.0858E+04,
     +  9.2135E+04, 9.3420E+04, 9.4714E+04, 9.6017E+04, 9.7329E+04,
     + 9.8651E+04, 9.9980E+04, 1.0132E+05, 1.0267E+05, 1.0403E+05,
     +  1.0539E+05, 1.0677E+05, 1.0816E+05, 1.0956E+05, 1.1096E+05,
     + 1.1238E+05, 1.1380E+05, 1.1524E+05, 1.1669E+05, 1.1814E+05,
     +  1.1961E+05, 1.2109E+05, 1.2257E+05, 1.2407E+05, 1.2558E+05,
     + 1.2710E+05, 1.2863E+05, 1.3017E+05, 1.3172E+05, 1.3328E+05,
     +  1.3486E+05, 1.3644E+05, 1.3804E+05, 1.3964E+05, 1.4126E+05,
     + 1.4289E+05, 1.4454E+05, 1.4619E+05, 1.4785E+05, 1.4953E+05,
     +  1.5122E+05, 1.5292E+05, 1.5464E+05, 1.5636E+05, 1.5810E+05,
     + 1.5985E+05, 1.6161E+05, 1.6339E+05, 1.6518E+05, 1.6698E+05,
     +  1.6879E+05, 1.7062E+05, 1.7246E+05, 1.7431E+05, 1.7618E+05,
     + 1.7806E+05, 1.7995E+05, 1.8186E+05, 1.8378E+05, 1.8571E+05,
     +  1.8766E+05, 1.8962E+05, 1.9160E+05, 1.9359E+05, 1.9559E+05,
     + 1.9761E+05, 1.9965E+05, 2.0169E+05, 2.0376E+05, 2.0583E+05,
     +  2.0793E+05, 2.1003E+05, 2.1215E+05, 2.1429E+05, 2.1644E+05,
     + 2.1861E+05, 2.2079E+05, 2.2299E+05, 2.2521E+05, 2.2744E+05,
     +  2.2968E+05, 2.3194E+05, 2.3422E+05, 2.3652E+05, 2.3883E+05,
     + 2.4115E+05, 2.4350E+05, 2.4586E+05, 2.4823E+05, 2.5063E+05,
     +  2.5304E+05, 2.5546E+05, 2.5791E+05, 2.6037E+05, 2.6285E+05,
     + 2.6534E+05, 2.6786E+05, 2.7039E+05, 2.7294E+05, 2.7550E+05,
     +  2.7809E+05, 2.8069E+05, 2.8331E+05, 2.8595E+05, 2.8861E+05,
     + 2.9129E+05, 2.9399E+05, 2.9670E+05, 2.9943E+05, 3.0218E+05,
     +  3.0496E+05, 3.0775E+05, 3.1056E+05, 3.1338E+05, 3.1623E+05,
     + 3.1910E+05, 3.2199E+05, 3.2490E+05, 3.2782E+05, 3.3077E+05,
     +  3.3374E+05, 3.3673E+05, 3.3974E+05, 3.4276E+05, 3.4581E+05,
     + 3.4888E+05, 3.5198E+05, 3.5509E+05, 3.5822E+05, 3.6138E+05,
     +  3.6455E+05, 3.6775E+05, 3.7097E+05, 3.7422E+05, 3.7748E+05,
     + 3.8076E+05, 3.8407E+05, 3.8740E+05, 3.9075E+05, 3.9413E+05,
     +  3.9752E+05, 4.0094E+05, 4.0439E+05, 4.0785E+05, 4.1134E+05,
     + 4.1485E+05, 4.1839E+05, 4.2194E+05, 4.2552E+05, 4.2913E+05,
     +  4.3276E+05, 4.3641E+05, 4.4009E+05, 4.4379E+05, 4.4752E+05,
     + 4.5127E+05, 4.5504E+05, 4.5884E+05, 4.6267E+05, 4.6652E+05,
     +  4.7039E+05, 4.7429E+05, 4.7822E+05, 4.8217E+05, 4.8614E+05,
     + 4.9015E+05, 4.9417E+05, 4.9823E+05, 5.0231E+05, 5.0642E+05,
     +  5.1055E+05, 5.1471E+05, 5.1889E+05, 5.2310E+05, 5.2734E+05,
     + 5.3161E+05, 5.3591E+05, 5.4022E+05, 5.4457E+05, 5.4895E+05,
     +  5.5335E+05, 5.5778E+05, 5.6225E+05, 5.6673E+05, 5.7125E+05,
     + 5.7579E+05, 5.8037E+05, 5.8497E+05, 5.8960E+05, 5.9426E+05,
     +  5.9895E+05, 6.0366E+05, 6.0841E+05, 6.1319E+05, 6.1799E+05,
     + 6.2283E+05, 6.2769E+05, 6.3259E+05, 6.3752E+05, 6.4247E+05,
     +  6.4746E+05, 6.5248E+05, 6.5753E+05, 6.6261E+05, 6.6772E+05,
     + 6.7286E+05, 6.7803E+05, 6.8324E+05, 6.8848E+05, 6.9374E+05,
     +  6.9904E+05, 7.0438E+05, 7.0974E+05, 7.1514E+05, 7.2057E+05,
     + 7.2603E+05, 7.3152E+05, 7.3705E+05, 7.4261E+05, 7.4821E+05,
     +  7.5383E+05, 7.5950E+05, 7.6519E+05, 7.7092E+05, 7.7668E+05,
     + 7.8248E+05, 7.8831E+05, 7.9418E+05, 8.0008E+05, 8.0602E+05,
     +  8.1199E+05, 8.1799E+05, 8.2403E+05, 8.3011E+05, 8.3622E+05,
     + 8.4237E+05, 8.4856E+05, 8.5478E+05, 8.6103E+05, 8.6732E+05,
     +  8.7365E+05, 8.8002E+05, 8.8642E+05, 8.9286E+05, 8.9934E+05,
     + 9.0586E+05, 9.1241E+05, 9.1899E+05, 9.2562E+05, 9.3229E+05,
     +  9.3899E+05, 9.4573E+05, 9.5252E+05, 9.5933E+05, 9.6619E+05,
     + 9.7309E+05, 9.8002E+05, 9.8700E+05, 9.9401E+05, 1.0011E+06,
     +  1.0082E+06, 1.0153E+06, 1.0225E+06, 1.0297E+06, 1.0369E+06,
     + 1.0442E+06, 1.0516E+06, 1.0590E+06, 1.0664E+06, 1.0738E+06,
     +  1.0813E+06, 1.0889E+06, 1.0965E+06, 1.1041E+06, 1.1118E+06,
     + 1.1195E+06, 1.1272E+06, 1.1350E+06, 1.1429E+06, 1.1507E+06,
     +  1.1587E+06, 1.1666E+06, 1.1746E+06, 1.1827E+06, 1.1908E+06,
     + 1.1990E+06, 1.2071E+06, 1.2154E+06, 1.2237E+06, 1.2320E+06,
     +  1.2403E+06, 1.2488E+06, 1.2572E+06, 1.2657E+06, 1.2743E+06,
     + 1.2829E+06, 1.2915E+06, 1.3002E+06, 1.3089E+06, 1.3177E+06,
     +  1.3265E+06, 1.3354E+06, 1.3443E+06, 1.3533E+06, 1.3623E+06,
     + 1.3714E+06, 1.3805E+06, 1.3897E+06, 1.3989E+06, 1.4081E+06,
     +  1.4174E+06, 1.4268E+06, 1.4362E+06, 1.4457E+06, 1.4552E+06,
     + 1.4647E+06, 1.4743E+06, 1.4840E+06, 1.4937E+06, 1.5034E+06,
     +  1.5132E+06, 1.5231E+06, 1.5330E+06, 1.5430E+06, 1.5530E+06,
     + 1.5630E+06, 1.5731E+06, 1.5833E+06, 1.5935E+06, 1.6038E+06,
     +  1.6141E+06, 1.6245E+06, 1.6349E+06, 1.6454E+06, 1.6559E+06,
     + 1.6665E+06, 1.6772E+06, 1.6879E+06, 1.6986E+06, 1.7094E+06,
     +  1.7203E+06, 1.7312E+06, 1.7422E+06, 1.7532E+06, 1.7643E+06,
     + 1.7754E+06, 1.7866E+06, 1.7978E+06, 1.8091E+06, 1.8205E+06,
     +  1.8319E+06, 1.8434E+06, 1.8549E+06, 1.8665E+06, 1.8782E+06,
     + 1.8898E+06, 1.9016E+06, 1.9134E+06, 1.9253E+06, 1.9372E+06,
     +  1.9492E+06, 1.9613E+06, 1.9734E+06, 1.9856E+06, 1.9978E+06,
     + 2.0101E+06, 2.0225E+06, 2.0349E+06, 2.0473E+06, 2.0599E+06,
     +  2.0725E+06, 2.0851E+06, 2.0978E+06, 2.1106E+06, 2.1234E+06,
     + 2.1363E+06, 2.1493E+06, 2.1623E+06, 2.1754E+06, 2.1886E+06,
     +  2.2018E+06, 2.2151E+06, 2.2284E+06, 2.2418E+06, 2.2553E+06,
     + 2.2688E+06, 2.2824E+06, 2.2961E+06, 2.3098E+06, 2.3236E+06,
     +  2.3374E+06, 2.3514E+06, 2.3654E+06, 2.3794E+06, 2.3935E+06,
     + 2.4077E+06, 2.4220E+06, 2.4363E+06, 2.4507E+06, 2.4651E+06,
     +  2.4797E+06, 2.4942E+06, 2.5089E+06, 2.5236E+06, 2.5384E+06,
     + 2.5533E+06, 2.5682E+06, 2.5832E+06, 2.5983E+06, 2.6135E+06,
     +  2.6287E+06, 2.6439E+06, 2.6593E+06, 2.6747E+06, 2.6902E+06,
     + 2.7058E+06, 2.7214E+06, 2.7371E+06, 2.7529E+06, 2.7688E+06,
     +  2.7847E+06, 2.8007E+06, 2.8167E+06, 2.8329E+06, 2.8491E+06,
     + 2.8654E+06, 2.8818E+06, 2.8982E+06, 2.9147E+06, 2.9313E+06,
     +  2.9480E+06, 2.9647E+06, 2.9815E+06, 2.9984E+06, 3.0154E+06,
     + 3.0324E+06, 3.0495E+06, 3.0667E+06, 3.0840E+06, 3.1013E+06,
     +  3.1187E+06, 3.1362E+06, 3.1538E+06, 3.1715E+06, 3.1892E+06,
     + 3.2070E+06, 3.2249E+06, 3.2429E+06, 3.2609E+06, 3.2790E+06,
     +  3.2973E+06, 3.3156E+06, 3.3339E+06, 3.3524E+06, 3.3709E+06,
     + 3.3895E+06, 3.4082E+06, 3.4270E+06, 3.4458E+06, 3.4648E+06,
     +  3.4838E+06, 3.5029E+06, 3.5221E+06, 3.5414E+06, 3.5607E+06,
     + 3.5802E+06, 3.5997E+06, 3.6193E+06, 3.6390E+06, 3.6587E+06,
     +  3.6786E+06, 3.6985E+06, 3.7186E+06, 3.7387E+06, 3.7589E+06,
     + 3.7792E+06, 3.7996E+06, 3.8200E+06, 3.8406E+06, 3.8612E+06,
     +  3.8819E+06, 3.9027E+06, 3.9236E+06, 3.9446E+06, 3.9657E+06,
     + 3.9869E+06, 4.0081E+06, 4.0295E+06, 4.0509E+06, 4.0724E+06,
     +  4.0940E+06, 4.1158E+06, 4.1376E+06, 4.1594E+06, 4.1814E+06,
     + 4.2035E+06, 4.2257E+06, 4.2479E+06, 4.2703E+06, 4.2927E+06,
     +  4.3152E+06, 4.3379E+06, 4.3606E+06, 4.3834E+06, 4.4063E+06,
     + 4.4293E+06, 4.4524E+06, 4.4756E+06, 4.4989E+06, 4.5223E+06,
     +  4.5458E+06, 4.5694E+06, 4.5931E+06, 4.6168E+06, 4.6407E+06,
     + 4.6647E+06, 4.6887E+06, 4.7129E+06, 4.7372E+06, 4.7615E+06,
     +  4.7860E+06, 4.8106E+06, 4.8352E+06, 4.8600E+06, 4.8848E+06,
     + 4.9098E+06, 4.9349E+06, 4.9600E+06, 4.9853E+06, 5.0107E+06,
     +  5.0361E+06, 5.0617E+06, 5.0874E+06, 5.1131E+06, 5.1390E+06,
     + 5.1650E+06, 5.1911E+06, 5.2173E+06, 5.2436E+06, 5.2700E+06,
     +  5.2965E+06, 5.3231E+06, 5.3498E+06, 5.3766E+06, 5.4035E+06,
     + 5.4306E+06, 5.4577E+06, 5.4850E+06, 5.5123E+06, 5.5398E+06,
     +  5.5673E+06, 5.5950E+06, 5.6228E+06, 5.6507E+06, 5.6787E+06,
     + 5.7068E+06, 5.7350E+06, 5.7634E+06, 5.7918E+06, 5.8204E+06,
     +  5.8491E+06, 5.8778E+06, 5.9067E+06, 5.9357E+06, 5.9648E+06,
     + 5.9941E+06, 6.0234E+06, 6.0529E+06, 6.0824E+06, 6.1121E+06,
     +  6.1419E+06, 6.1718E+06, 6.2019E+06, 6.2320E+06, 6.2622E+06,
     + 6.2926E+06, 6.3231E+06, 6.3537E+06, 6.3844E+06, 6.4153E+06,
     +  6.4462E+06, 6.4773E+06, 6.5085E+06, 6.5398E+06, 6.5712E+06,
     + 6.6028E+06, 6.6345E+06, 6.6663E+06, 6.6982E+06, 6.7302E+06,
     +  6.7623E+06, 6.7946E+06, 6.8270E+06, 6.8595E+06, 6.8922E+06,
     + 6.9249E+06, 6.9578E+06, 6.9908E+06, 7.0239E+06, 7.0572E+06,
     +  7.0906E+06, 7.1241E+06, 7.1577E+06, 7.1914E+06, 7.2253E+06,
     + 7.2593E+06, 7.2935E+06, 7.3277E+06, 7.3621E+06, 7.3966E+06,
     +  7.4313E+06, 7.4660E+06, 7.5009E+06, 7.5359E+06, 7.5711E+06,
     + 7.6064E+06, 7.6418E+06, 7.6773E+06, 7.7130E+06, 7.7488E+06,
     +  7.7847E+06, 7.8208E+06, 7.8570E+06, 7.8933E+06, 7.9298E+06,
     + 7.9664E+06, 8.0031E+06, 8.0400E+06, 8.0769E+06, 8.1141E+06,
     +  8.1513E+06, 8.1887E+06, 8.2263E+06, 8.2639E+06, 8.3017E+06,
     + 8.3397E+06, 8.3777E+06, 8.4160E+06, 8.4543E+06, 8.4928E+06,
     +  8.5314E+06, 8.5702E+06, 8.6091E+06, 8.6481E+06, 8.6873E+06,
     + 8.7266E+06, 8.7661E+06, 8.8057E+06, 8.8455E+06, 8.8854E+06,
     +  8.9254E+06, 8.9656E+06, 9.0059E+06, 9.0463E+06, 9.0869E+06,
     + 9.1277E+06, 9.1686E+06, 9.2096E+06, 9.2508E+06, 9.2921E+06,
     +  9.3336E+06, 9.3752E+06, 9.4169E+06, 9.4588E+06, 9.5009E+06,
     + 9.5431E+06, 9.5854E+06, 9.6279E+06, 9.6706E+06, 9.7133E+06,
     +  9.7563E+06, 9.7994E+06, 9.8426E+06, 9.8860E+06, 9.9296E+06,
     + 9.9732E+06, 1.0017E+07, 1.0061E+07, 1.0105E+07, 1.0150E+07,
     +  1.0194E+07, 1.0239E+07, 1.0283E+07, 1.0328E+07, 1.0373E+07,
     + 1.0419E+07, 1.0464E+07, 1.0509E+07, 1.0555E+07, 1.0601E+07,
     +  1.0647E+07, 1.0693E+07, 1.0740E+07, 1.0786E+07, 1.0833E+07,
     + 1.0879E+07, 1.0926E+07, 1.0974E+07, 1.1021E+07, 1.1068E+07,
     +  1.1116E+07, 1.1164E+07, 1.1211E+07, 1.1260E+07, 1.1308E+07,
     + 1.1356E+07, 1.1405E+07, 1.1454E+07, 1.1502E+07, 1.1551E+07,
     +  1.1601E+07, 1.1650E+07, 1.1700E+07, 1.1749E+07, 1.1799E+07,
     + 1.1849E+07, 1.1900E+07, 1.1950E+07, 1.2000E+07, 1.2051E+07,
     +  1.2102E+07, 1.2153E+07, 1.2204E+07, 1.2256E+07, 1.2307E+07,
     + 1.2359E+07, 1.2411E+07, 1.2463E+07, 1.2515E+07, 1.2568E+07,
     +  1.2620E+07, 1.2673E+07, 1.2726E+07, 1.2779E+07, 1.2833E+07,
     + 1.2886E+07, 1.2940E+07, 1.2994E+07, 1.3048E+07, 1.3102E+07,
     +  1.3156E+07, 1.3211E+07, 1.3265E+07, 1.3320E+07, 1.3375E+07,
     + 1.3431E+07, 1.3486E+07, 1.3542E+07, 1.3597E+07, 1.3653E+07,
     +  1.3710E+07, 1.3766E+07, 1.3822E+07, 1.3879E+07, 1.3936E+07,
     + 1.3993E+07, 1.4050E+07, 1.4108E+07, 1.4165E+07, 1.4223E+07,
     +  1.4281E+07, 1.4339E+07, 1.4398E+07, 1.4456E+07, 1.4515E+07,
     + 1.4574E+07, 1.4633E+07, 1.4692E+07, 1.4752E+07, 1.4811E+07,
     +  1.4871E+07, 1.4931E+07, 1.4992E+07, 1.5052E+07, 1.5113E+07,
     + 1.5174E+07, 1.5235E+07, 1.5296E+07, 1.5357E+07, 1.5419E+07,
     +  1.5481E+07, 1.5543E+07, 1.5605E+07, 1.5667E+07, 1.5730E+07,
     + 1.5792E+07, 1.5855E+07, 1.5919E+07, 1.5982E+07, 1.6046E+07,
     +  1.6109E+07, 1.6173E+07, 1.6237E+07, 1.6302E+07, 1.6366E+07,
     + 1.6431E+07, 1.6496E+07, 1.6561E+07, 1.6627E+07, 1.6692E+07,
     +  1.6758E+07, 1.6824E+07, 1.6890E+07, 1.6957E+07, 1.7023E+07,
     + 1.7090E+07, 1.7157E+07, 1.7224E+07, 1.7292E+07, 1.7359E+07,
     +  1.7427E+07, 1.7495E+07, 1.7563E+07, 1.7632E+07, 1.7700E+07,
     + 1.7769E+07, 1.7838E+07, 1.7908E+07, 1.7977E+07, 1.8047E+07,
     +  1.8117E+07, 1.8187E+07, 1.8258E+07, 1.8328E+07, 1.8399E+07,
     + 1.8470E+07, 1.8541E+07, 1.8613E+07, 1.8684E+07, 1.8756E+07,
     +  1.8828E+07, 1.8901E+07, 1.8973E+07, 1.9046E+07, 1.9119E+07,
     + 1.9192E+07, 1.9265E+07, 1.9339E+07, 1.9413E+07, 1.9487E+07,
     +  1.9561E+07, 1.9636E+07, 1.9711E+07, 1.9786E+07, 1.9861E+07,
     + 1.9936E+07, 2.0012E+07, 2.0088E+07, 2.0164E+07, 2.0240E+07,
     +  2.0317E+07, 2.0393E+07, 2.0470E+07, 2.0548E+07, 2.0625E+07,
     + 2.0703E+07, 2.0781E+07, 2.0859E+07, 2.0937E+07, 2.1016E+07,
     +  2.1095E+07, 2.1174E+07, 2.1253E+07, 2.1333E+07, 2.1412E+07,
     + 2.1492E+07, 2.1573E+07, 2.1653E+07, 2.1734E+07, 2.1815E+07,
     +  2.1896E+07, 2.1978E+07, 2.2059E+07, 2.2141E+07, 2.2223E+07,
     + 2.2306E+07, 2.2388E+07, 2.2471E+07, 2.2554E+07, 2.2638E+07,
     +  2.2721E+07, 2.2805E+07, 2.2889E+07, 2.2973E+07, 2.3058E+07,
     + 2.3143E+07, 2.3228E+07, 2.3313E+07, 2.3399E+07, 2.3485E+07,
     +  2.3571E+07, 2.3657E+07, 2.3743E+07, 2.3830E+07, 2.3917E+07,
     + 2.4005E+07, 2.4092E+07, 2.4180E+07, 2.4268E+07, 2.4356E+07,
     +  2.4445E+07, 2.4534E+07, 2.4623E+07, 2.4712E+07, 2.4802E+07,
     + 2.4891E+07, 2.4981E+07, 2.5072E+07, 2.5162E+07, 2.5253E+07,
     +  2.5344E+07, 2.5436E+07, 2.5527E+07, 2.5619E+07, 2.5711E+07,
     + 2.5804E+07, 2.5896E+07, 2.5989E+07, 2.6082E+07, 2.6176E+07,
     +  2.6270E+07, 2.6364E+07, 2.6458E+07, 2.6552E+07, 2.6647E+07,
     + 2.6742E+07, 2.6837E+07, 2.6933E+07, 2.7029E+07, 2.7125E+07,
     +  2.7221E+07, 2.7318E+07, 2.7415E+07, 2.7512E+07, 2.7610E+07,
     + 2.7707E+07, 2.7805E+07, 2.7904E+07, 2.8002E+07, 2.8101E+07,
     +  2.8200E+07, 2.8299E+07, 2.8399E+07, 2.8499E+07, 2.8599E+07,
     + 2.8700E+07, 2.8800E+07, 2.8901E+07, 2.9003E+07, 2.9104E+07,
     +  2.9206E+07, 2.9308E+07, 2.9411E+07, 2.9513E+07, 2.9616E+07,
     + 2.9720E+07, 2.9823E+07, 2.9927E+07, 3.0031E+07, 3.0136E+07,
     +  3.0240E+07, 3.0345E+07, 3.0450E+07, 3.0556E+07, 3.0662E+07,
     + 3.0768E+07, 3.0874E+07, 3.0981E+07, 3.1088E+07, 3.1195E+07,
     +  3.1303E+07, 3.1411E+07, 3.1519E+07, 3.1627E+07, 3.1736E+07,
     + 3.1845E+07, 3.1955E+07, 3.2064E+07, 3.2174E+07, 3.2284E+07,
     +  3.2395E+07, 3.2506E+07, 3.2617E+07, 3.2728E+07, 3.2840E+07,
     + 3.2952E+07, 3.3064E+07, 3.3177E+07, 3.3290E+07, 3.3403E+07,
     +  3.3516E+07, 3.3630E+07, 3.3744E+07, 3.3859E+07, 3.3974E+07,
     + 3.4089E+07, 3.4204E+07, 3.4320E+07, 3.4436E+07, 3.4552E+07,
     +  3.4668E+07, 3.4785E+07, 3.4902E+07, 3.5020E+07, 3.5138E+07,
     + 3.5256E+07, 3.5374E+07, 3.5493E+07, 3.5612E+07, 3.5731E+07,
     +  3.5851E+07, 3.5971E+07, 3.6091E+07, 3.6212E+07, 3.6333E+07,
     + 3.6454E+07, 3.6576E+07, 3.6698E+07, 3.6820E+07, 3.6943E+07,
     +  3.7065E+07, 3.7189E+07, 3.7312E+07, 3.7436E+07, 3.7560E+07,
     + 3.7685E+07, 3.7809E+07, 3.7935E+07, 3.8060E+07, 3.8186E+07,
     +  3.8312E+07, 3.8438E+07, 3.8565E+07, 3.8692E+07, 3.8820E+07,
     + 3.8947E+07, 3.9075E+07, 3.9204E+07, 3.9333E+07, 3.9462E+07,
     +  3.9591E+07, 3.9721E+07, 3.9851E+07, 3.9981E+07, 4.0112E+07,
     + 4.0243E+07, 4.0375E+07, 4.0506E+07, 4.0638E+07, 4.0771E+07,
     +  4.0904E+07, 4.1037E+07, 4.1170E+07, 4.1304E+07, 4.1438E+07,
     + 4.1573E+07, 4.1707E+07, 4.1843E+07, 4.1978E+07, 4.2114E+07,
     +  4.2250E+07, 4.2387E+07, 4.2524E+07, 4.2661E+07, 4.2798E+07,
     + 4.2936E+07, 4.3075E+07, 4.3213E+07, 4.3352E+07, 4.3492E+07,
     +  4.3631E+07, 4.3771E+07, 4.3912E+07, 4.4052E+07, 4.4193E+07,
     + 4.4335E+07, 4.4477E+07, 4.4619E+07, 4.4761E+07, 4.4904E+07,
     +  4.5048E+07, 4.5191E+07, 4.5335E+07, 4.5479E+07, 4.5624E+07,
     + 4.5769E+07, 4.5915E+07, 4.6060E+07, 4.6206E+07, 4.6353E+07,
     +  4.6500E+07, 4.6647E+07, 4.6795E+07, 4.6943E+07, 4.7091E+07,
     + 4.7240E+07, 4.7389E+07, 4.7538E+07, 4.7688E+07, 4.7838E+07,
     +  4.7989E+07, 4.8140E+07, 4.8291E+07, 4.8442E+07, 4.8595E+07,
     + 4.8747E+07, 4.8900E+07, 4.9053E+07, 4.9206E+07, 4.9360E+07,
     +  4.9515E+07, 4.9669E+07, 4.9824E+07, 4.9980E+07, 5.0136E+07,
     + 5.0292E+07, 5.0448E+07, 5.0605E+07, 5.0763E+07, 5.0920E+07,
     +  5.1079E+07, 5.1237E+07, 5.1396E+07, 5.1555E+07, 5.1715E+07,
     + 5.1875E+07, 5.2035E+07, 5.2196E+07, 5.2357E+07, 5.2519E+07,
     +  5.2681E+07, 5.2843E+07, 5.3006E+07, 5.3169E+07, 5.3333E+07,
     + 5.3497E+07, 5.3661E+07, 5.3826E+07, 5.3991E+07, 5.4157E+07,
     +  5.4323E+07, 5.4489E+07, 5.4656E+07, 5.4823E+07, 5.4991E+07,
     + 5.5159E+07, 5.5327E+07, 5.5496E+07, 5.5665E+07, 5.5835E+07,
     +  5.6005E+07, 5.6175E+07, 5.6346E+07, 5.6517E+07, 5.6689E+07,
     + 5.6861E+07, 5.7033E+07, 5.7206E+07, 5.7379E+07, 5.7553E+07,
     +  5.7727E+07, 5.7901E+07, 5.8076E+07, 5.8252E+07, 5.8427E+07,
     + 5.8604E+07, 5.8780E+07, 5.8957E+07, 5.9135E+07, 5.9312E+07,
     +  5.9491E+07, 5.9669E+07, 5.9848E+07, 6.0028E+07, 6.0208E+07,
     + 6.0388E+07, 6.0569E+07, 6.0750E+07, 6.0932E+07, 6.1114E+07,
     +  6.1296E+07, 6.1479E+07, 6.1663E+07, 6.1847E+07, 6.2031E+07,
     + 6.2215E+07, 6.2400E+07, 6.2586E+07, 6.2772E+07, 6.2958E+07,
     +  6.3145E+07, 6.3332E+07, 6.3520E+07, 6.3708E+07, 6.3897E+07,
     + 6.4086E+07, 6.4275E+07, 6.4465E+07, 6.4655E+07, 6.4846E+07,
     +  6.5037E+07, 6.5229E+07, 6.5421E+07, 6.5614E+07, 6.5807E+07,
     + 6.6000E+07, 6.6194E+07, 6.6388E+07, 6.6583E+07, 6.6778E+07,
     +  6.6974E+07, 6.7170E+07, 6.7366E+07, 6.7563E+07, 6.7761E+07,
     + 6.7959E+07, 6.8157E+07, 6.8356E+07, 6.8555E+07, 6.8755E+07,
     +  6.8955E+07, 6.9156E+07, 6.9357E+07, 6.9559E+07, 6.9761E+07,
     + 6.9963E+07, 7.0166E+07, 7.0370E+07, 7.0573E+07, 7.0778E+07,
     +  7.0983E+07, 7.1188E+07, 7.1394E+07, 7.1600E+07, 7.1807E+07,
     + 7.2014E+07, 7.2221E+07, 7.2429E+07, 7.2638E+07, 7.2847E+07,
     +  7.3057E+07, 7.3266E+07, 7.3477E+07, 7.3688E+07, 7.3899E+07,
     + 7.4111E+07, 7.4324E+07, 7.4536E+07, 7.4750E+07, 7.4963E+07,
     +  7.5178E+07, 7.5392E+07, 7.5608E+07, 7.5823E+07, 7.6040E+07,
     + 7.6256E+07, 7.6473E+07, 7.6691E+07, 7.6909E+07, 7.7128E+07,
     +  7.7347E+07, 7.7567E+07, 7.7787E+07, 7.8007E+07, 7.8228E+07,
     + 7.8450E+07, 7.8672E+07, 7.8895E+07, 7.9118E+07, 7.9341E+07,
     +  7.9565E+07, 7.9790E+07, 8.0015E+07, 8.0240E+07, 8.0466E+07,
     + 8.0693E+07, 8.0920E+07, 8.1148E+07, 8.1376E+07, 8.1604E+07,
     +  8.1833E+07, 8.2063E+07, 8.2293E+07, 8.2523E+07, 8.2754E+07,
     + 8.2986E+07, 8.3218E+07, 8.3451E+07, 8.3684E+07, 8.3917E+07,
     +  8.4152E+07, 8.4386E+07, 8.4621E+07, 8.4857E+07, 8.5093E+07,
     + 8.5330E+07, 8.5567E+07, 8.5805E+07, 8.6043E+07, 8.6282E+07,
     +  8.6521E+07, 8.6761E+07, 8.7002E+07, 8.7242E+07, 8.7484E+07,
     + 8.7726E+07, 8.7968E+07, 8.8211E+07, 8.8455E+07, 8.8699E+07,
     +  8.8943E+07, 8.9188E+07, 8.9434E+07, 8.9680E+07, 8.9927E+07,
     + 9.0174E+07, 9.0422E+07, 9.0670E+07, 9.0919E+07, 9.1168E+07,
     +  9.1418E+07, 9.1669E+07, 9.1919E+07, 9.2171E+07, 9.2423E+07,
     + 9.2676E+07, 9.2929E+07, 9.3182E+07, 9.3437E+07, 9.3691E+07,
     +  9.3947E+07, 9.4203E+07, 9.4459E+07, 9.4716E+07, 9.4973E+07,
     + 9.5232E+07, 9.5490E+07, 9.5749E+07, 9.6009E+07, 9.6269E+07,
     +  9.6530E+07, 9.6791E+07, 9.7053E+07, 9.7316E+07, 9.7579E+07,
     + 9.7842E+07, 9.8107E+07, 9.8371E+07, 9.8637E+07, 9.8902E+07,
     +  9.9169E+07, 9.9436E+07, 9.9703E+07, 9.9971E+07, 1.0024E+08,
     + 1.0051E+08, 1.0078E+08, 1.0105E+08, 1.0132E+08, 1.0159E+08,
     +  1.0186E+08, 1.0214E+08, 1.0241E+08, 1.0268E+08, 1.0296E+08,
     + 1.0323E+08, 1.0351E+08, 1.0378E+08, 1.0406E+08, 1.0434E+08,
     +  1.0462E+08, 1.0490E+08, 1.0517E+08, 1.0545E+08, 1.0573E+08,
     + 1.0601E+08, 1.0630E+08, 1.0658E+08, 1.0686E+08, 1.0714E+08,
     +  1.0743E+08, 1.0771E+08, 1.0800E+08, 1.0828E+08, 1.0857E+08,
     + 1.0885E+08, 1.0914E+08, 1.0943E+08, 1.0972E+08, 1.1001E+08,
     +  1.1030E+08, 1.1059E+08, 1.1088E+08, 1.1117E+08, 1.1146E+08,
     + 1.1175E+08, 1.1205E+08, 1.1234E+08, 1.1264E+08, 1.1293E+08,
     +  1.1323E+08, 1.1352E+08, 1.1382E+08, 1.1412E+08, 1.1442E+08,
     + 1.1471E+08, 1.1501E+08, 1.1531E+08, 1.1561E+08, 1.1591E+08,
     +  1.1622E+08, 1.1652E+08, 1.1682E+08, 1.1713E+08, 1.1743E+08,
     + 1.1773E+08, 1.1804E+08, 1.1835E+08, 1.1865E+08, 1.1896E+08,
     +  1.1927E+08, 1.1958E+08, 1.1989E+08, 1.2020E+08, 1.2051E+08,
     + 1.2082E+08, 1.2113E+08, 1.2144E+08, 1.2176E+08, 1.2207E+08,
     +  1.2238E+08, 1.2270E+08, 1.2301E+08, 1.2333E+08, 1.2365E+08,
     + 1.2396E+08, 1.2428E+08, 1.2460E+08, 1.2492E+08, 1.2524E+08,
     +  1.2556E+08, 1.2588E+08, 1.2620E+08, 1.2653E+08, 1.2685E+08,
     + 1.2717E+08, 1.2750E+08, 1.2782E+08, 1.2815E+08, 1.2848E+08,
     +  1.2880E+08, 1.2913E+08, 1.2946E+08, 1.2979E+08, 1.3012E+08,
     + 1.3045E+08, 1.3078E+08, 1.3111E+08, 1.3145E+08, 1.3178E+08,
     +  1.3211E+08, 1.3245E+08, 1.3278E+08, 1.3312E+08, 1.3345E+08,
     + 1.3379E+08, 1.3413E+08, 1.3447E+08, 1.3481E+08, 1.3515E+08,
     +  1.3549E+08, 1.3583E+08, 1.3617E+08, 1.3651E+08, 1.3686E+08,
     + 1.3720E+08, 1.3754E+08, 1.3789E+08, 1.3824E+08, 1.3858E+08,
     +  1.3893E+08, 1.3928E+08, 1.3963E+08, 1.3998E+08, 1.4033E+08,
     + 1.4068E+08, 1.4103E+08, 1.4138E+08, 1.4173E+08, 1.4209E+08,
     +  1.4244E+08, 1.4279E+08, 1.4315E+08, 1.4351E+08, 1.4386E+08,
     + 1.4422E+08, 1.4458E+08, 1.4494E+08, 1.4530E+08, 1.4566E+08,
     +  1.4602E+08, 1.4638E+08, 1.4674E+08, 1.4711E+08, 1.4747E+08,
     + 1.4784E+08, 1.4820E+08, 1.4857E+08, 1.4893E+08, 1.4930E+08,
     +  1.4967E+08, 1.5004E+08, 1.5041E+08, 1.5078E+08, 1.5115E+08,
     + 1.5152E+08, 1.5189E+08, 1.5227E+08, 1.5264E+08, 1.5302E+08,
     +  1.5339E+08, 1.5377E+08, 1.5414E+08, 1.5452E+08, 1.5490E+08,
     + 1.5528E+08, 1.5566E+08, 1.5604E+08, 1.5642E+08, 1.5680E+08,
     +  1.5719E+08, 1.5757E+08, 1.5795E+08, 1.5834E+08, 1.5872E+08,
     + 1.5911E+08, 1.5950E+08, 1.5988E+08, 1.6027E+08, 1.6066E+08,
     +  1.6105E+08, 1.6144E+08, 1.6184E+08, 1.6223E+08, 1.6262E+08,
     + 1.6301E+08, 1.6341E+08, 1.6380E+08, 1.6420E+08, 1.6460E+08,
     +  1.6499E+08, 1.6539E+08, 1.6579E+08, 1.6619E+08, 1.6659E+08,
     + 1.6699E+08, 1.6740E+08, 1.6780E+08, 1.6820E+08, 1.6861E+08,
     +  1.6901E+08, 1.6942E+08, 1.6982E+08, 1.7023E+08, 1.7064E+08,
     + 1.7105E+08, 1.7146E+08, 1.7187E+08, 1.7228E+08, 1.7269E+08,
     +  1.7311E+08, 1.7352E+08, 1.7393E+08, 1.7435E+08, 1.7476E+08,
     + 1.7518E+08, 1.7560E+08, 1.7602E+08, 1.7644E+08, 1.7686E+08,
     +  1.7728E+08, 1.7770E+08, 1.7812E+08, 1.7854E+08, 1.7897E+08,
     + 1.7939E+08, 1.7982E+08, 1.8024E+08, 1.8067E+08, 1.8110E+08,
     +  1.8153E+08, 1.8196E+08, 1.8239E+08, 1.8282E+08, 1.8325E+08,
     + 1.8368E+08, 1.8411E+08, 1.8455E+08, 1.8498E+08, 1.8542E+08,
     +  1.8586E+08, 1.8629E+08, 1.8673E+08, 1.8717E+08, 1.8761E+08,
     + 1.8805E+08, 1.8849E+08, 1.8893E+08, 1.8938E+08, 1.8982E+08,
     +  1.9027E+08, 1.9071E+08, 1.9116E+08, 1.9161E+08, 1.9205E+08,
     + 1.9250E+08, 1.9295E+08, 1.9340E+08, 1.9385E+08, 1.9431E+08,
     +  1.9476E+08, 1.9521E+08, 1.9567E+08, 1.9612E+08, 1.9658E+08,
     + 1.9704E+08, 1.9749E+08, 1.9795E+08, 1.9841E+08, 1.9887E+08,
     +  1.9933E+08, 1.9980E+08, 2.0026E+08, 2.0072E+08, 2.0119E+08,
     + 2.0165E+08, 2.0212E+08, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT(17,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       777
      data (QofT(18,J),J=1,2408)/ 1.8376E+02, 4.7616E+02, 8.4860E+02,
     +  1.2868E+03, 1.7821E+03, 2.3285E+03, 2.9217E+03, 3.5582E+03,
     + 4.2353E+03, 4.9505E+03, 5.7022E+03, 6.4884E+03, 7.3078E+03,
     +  8.1592E+03, 9.0412E+03, 9.9530E+03, 1.0894E+04, 1.1862E+04,
     + 1.2858E+04, 1.3880E+04, 1.6000E+04, 1.8220E+04, 2.0533E+04,
     +  2.2937E+04, 2.5428E+04, 2.8003E+04, 3.0660E+04, 3.3396E+04,
     + 3.6210E+04, 3.9098E+04, 4.2059E+04, 4.5092E+04, 4.8195E+04,
     +  5.1366E+04, 5.4603E+04, 5.7906E+04, 6.1274E+04, 6.4704E+04,
     + 6.8197E+04, 7.1750E+04, 7.5363E+04, 7.9035E+04, 8.2765E+04,
     +  8.6552E+04, 9.0395E+04, 9.4294E+04, 9.8247E+04, 1.0225E+05,
     + 1.0632E+05, 1.1043E+05, 1.1459E+05, 1.1881E+05, 1.2308E+05,
     +  1.2740E+05, 1.3176E+05, 1.3618E+05, 1.4065E+05, 1.4516E+05,
     + 1.4972E+05, 1.5433E+05, 1.5899E+05, 1.6369E+05, 1.6844E+05,
     +  1.7323E+05, 1.7807E+05, 1.8296E+05, 1.8789E+05, 1.9286E+05,
     + 1.9788E+05, 2.0294E+05, 2.0805E+05, 2.1320E+05, 2.1839E+05,
     +  2.2363E+05, 2.2891E+05, 2.3424E+05, 2.3960E+05, 2.4501E+05,
     + 2.5046E+05, 2.5596E+05, 2.6149E+05, 2.6707E+05, 2.7270E+05,
     +  2.7836E+05, 2.8407E+05, 2.8982E+05, 2.9561E+05, 3.0144E+05,
     + 3.0732E+05, 3.1324E+05, 3.1920E+05, 3.2521E+05, 3.3126E+05,
     +  3.3735E+05, 3.4348E+05, 3.4966E+05, 3.5588E+05, 3.6214E+05,
     + 3.6845E+05, 3.7480E+05, 3.8120E+05, 3.8764E+05, 3.9412E+05,
     +  4.0065E+05, 4.0723E+05, 4.1385E+05, 4.2051E+05, 4.2722E+05,
     + 4.3398E+05, 4.4078E+05, 4.4763E+05, 4.5453E+05, 4.6147E+05,
     +  4.6846E+05, 4.7550E+05, 4.8259E+05, 4.8972E+05, 4.9690E+05,
     + 5.0414E+05, 5.1142E+05, 5.1875E+05, 5.2613E+05, 5.3356E+05,
     +  5.4105E+05, 5.4858E+05, 5.5617E+05, 5.6380E+05, 5.7149E+05,
     + 5.7923E+05, 5.8703E+05, 5.9488E+05, 6.0278E+05, 6.1074E+05,
     +  6.1875E+05, 6.2682E+05, 6.3494E+05, 6.4312E+05, 6.5136E+05,
     + 6.5965E+05, 6.6800E+05, 6.7641E+05, 6.8488E+05, 6.9340E+05,
     +  7.0198E+05, 7.1063E+05, 7.1933E+05, 7.2809E+05, 7.3692E+05,
     + 7.4580E+05, 7.5476E+05, 7.6377E+05, 7.7284E+05, 7.8198E+05,
     +  7.9118E+05, 8.0045E+05, 8.0978E+05, 8.1918E+05, 8.2865E+05,
     + 8.3817E+05, 8.4777E+05, 8.5743E+05, 8.6717E+05, 8.7697E+05,
     +  8.8684E+05, 8.9677E+05, 9.0679E+05, 9.1687E+05, 9.2702E+05,
     + 9.3725E+05, 9.4754E+05, 9.5791E+05, 9.6835E+05, 9.7887E+05,
     +  9.8945E+05, 1.0001E+06, 1.0109E+06, 1.0217E+06, 1.0326E+06,
     + 1.0435E+06, 1.0546E+06, 1.0657E+06, 1.0769E+06, 1.0882E+06,
     +  1.0996E+06, 1.1110E+06, 1.1225E+06, 1.1341E+06, 1.1458E+06,
     + 1.1576E+06, 1.1694E+06, 1.1814E+06, 1.1934E+06, 1.2055E+06,
     +  1.2177E+06, 1.2300E+06, 1.2424E+06, 1.2548E+06, 1.2673E+06,
     + 1.2800E+06, 1.2927E+06, 1.3055E+06, 1.3184E+06, 1.3314E+06,
     +  1.3445E+06, 1.3576E+06, 1.3709E+06, 1.3843E+06, 1.3977E+06,
     + 1.4113E+06, 1.4249E+06, 1.4386E+06, 1.4525E+06, 1.4664E+06,
     +  1.4804E+06, 1.4945E+06, 1.5088E+06, 1.5231E+06, 1.5375E+06,
     + 1.5520E+06, 1.5667E+06, 1.5814E+06, 1.5962E+06, 1.6111E+06,
     +  1.6262E+06, 1.6413E+06, 1.6566E+06, 1.6719E+06, 1.6874E+06,
     + 1.7029E+06, 1.7186E+06, 1.7344E+06, 1.7503E+06, 1.7663E+06,
     +  1.7824E+06, 1.7986E+06, 1.8149E+06, 1.8313E+06, 1.8479E+06,
     + 1.8645E+06, 1.8813E+06, 1.8982E+06, 1.9152E+06, 1.9323E+06,
     +  1.9496E+06, 1.9669E+06, 1.9844E+06, 2.0020E+06, 2.0197E+06,
     + 2.0375E+06, 2.0555E+06, 2.0736E+06, 2.0917E+06, 2.1101E+06,
     +  2.1285E+06, 2.1471E+06, 2.1657E+06, 2.1846E+06, 2.2035E+06,
     + 2.2226E+06, 2.2418E+06, 2.2611E+06, 2.2805E+06, 2.3001E+06,
     +  2.3198E+06, 2.3396E+06, 2.3596E+06, 2.3797E+06, 2.3999E+06,
     + 2.4203E+06, 2.4408E+06, 2.4614E+06, 2.4822E+06, 2.5031E+06,
     +  2.5242E+06, 2.5454E+06, 2.5667E+06, 2.5881E+06, 2.6097E+06,
     + 2.6315E+06, 2.6534E+06, 2.6754E+06, 2.6975E+06, 2.7199E+06,
     +  2.7423E+06, 2.7649E+06, 2.7877E+06, 2.8106E+06, 2.8336E+06,
     + 2.8568E+06, 2.8801E+06, 2.9036E+06, 2.9272E+06, 2.9510E+06,
     +  2.9749E+06, 2.9990E+06, 3.0233E+06, 3.0477E+06, 3.0722E+06,
     + 3.0969E+06, 3.1218E+06, 3.1468E+06, 3.1720E+06, 3.1973E+06,
     +  3.2228E+06, 3.2485E+06, 3.2743E+06, 3.3003E+06, 3.3264E+06,
     + 3.3527E+06, 3.3792E+06, 3.4058E+06, 3.4326E+06, 3.4596E+06,
     +  3.4867E+06, 3.5140E+06, 3.5415E+06, 3.5691E+06, 3.5969E+06,
     + 3.6249E+06, 3.6531E+06, 3.6814E+06, 3.7099E+06, 3.7386E+06,
     +  3.7674E+06, 3.7965E+06, 3.8256E+06, 3.8550E+06, 3.8846E+06,
     + 3.9143E+06, 3.9442E+06, 3.9743E+06, 4.0046E+06, 4.0351E+06,
     +  4.0657E+06, 4.0966E+06, 4.1276E+06, 4.1588E+06, 4.1902E+06,
     + 4.2218E+06, 4.2535E+06, 4.2855E+06, 4.3176E+06, 4.3500E+06,
     +  4.3825E+06, 4.4152E+06, 4.4481E+06, 4.4812E+06, 4.5145E+06,
     + 4.5481E+06, 4.5818E+06, 4.6157E+06, 4.6497E+06, 4.6840E+06,
     +  4.7185E+06, 4.7532E+06, 4.7881E+06, 4.8233E+06, 4.8586E+06,
     + 4.8941E+06, 4.9298E+06, 4.9657E+06, 5.0019E+06, 5.0382E+06,
     +  5.0748E+06, 5.1115E+06, 5.1485E+06, 5.1857E+06, 5.2231E+06,
     + 5.2607E+06, 5.2986E+06, 5.3366E+06, 5.3749E+06, 5.4133E+06,
     +  5.4521E+06, 5.4910E+06, 5.5301E+06, 5.5695E+06, 5.6091E+06,
     + 5.6489E+06, 5.6889E+06, 5.7292E+06, 5.7697E+06, 5.8104E+06,
     +  5.8513E+06, 5.8925E+06, 5.9339E+06, 5.9756E+06, 6.0174E+06,
     + 6.0596E+06, 6.1019E+06, 6.1444E+06, 6.1873E+06, 6.2303E+06,
     +  6.2736E+06, 6.3171E+06, 6.3609E+06, 6.4049E+06, 6.4491E+06,
     + 6.4936E+06, 6.5383E+06, 6.5833E+06, 6.6285E+06, 6.6740E+06,
     +  6.7197E+06, 6.7657E+06, 6.8119E+06, 6.8584E+06, 6.9051E+06,
     + 6.9521E+06, 6.9993E+06, 7.0468E+06, 7.0945E+06, 7.1425E+06,
     +  7.1908E+06, 7.2393E+06, 7.2881E+06, 7.3371E+06, 7.3864E+06,
     + 7.4360E+06, 7.4858E+06, 7.5359E+06, 7.5862E+06, 7.6369E+06,
     +  7.6878E+06, 7.7389E+06, 7.7904E+06, 7.8421E+06, 7.8941E+06,
     + 7.9463E+06, 7.9988E+06, 8.0517E+06, 8.1047E+06, 8.1581E+06,
     +  8.2117E+06, 8.2656E+06, 8.3198E+06, 8.3743E+06, 8.4291E+06,
     + 8.4841E+06, 8.5395E+06, 8.5951E+06, 8.6510E+06, 8.7072E+06,
     +  8.7637E+06, 8.8205E+06, 8.8776E+06, 8.9349E+06, 8.9926E+06,
     + 9.0505E+06, 9.1088E+06, 9.1674E+06, 9.2262E+06, 9.2854E+06,
     +  9.3448E+06, 9.4046E+06, 9.4647E+06, 9.5250E+06, 9.5857E+06,
     + 9.6467E+06, 9.7080E+06, 9.7696E+06, 9.8315E+06, 9.8937E+06,
     +  9.9562E+06, 1.0019E+07, 1.0082E+07, 1.0146E+07, 1.0209E+07,
     + 1.0274E+07, 1.0338E+07, 1.0403E+07, 1.0468E+07, 1.0533E+07,
     +  1.0599E+07, 1.0665E+07, 1.0731E+07, 1.0798E+07, 1.0865E+07,
     + 1.0933E+07, 1.1000E+07, 1.1068E+07, 1.1137E+07, 1.1205E+07,
     +  1.1274E+07, 1.1344E+07, 1.1413E+07, 1.1483E+07, 1.1554E+07,
     + 1.1625E+07, 1.1696E+07, 1.1767E+07, 1.1839E+07, 1.1911E+07,
     +  1.1983E+07, 1.2056E+07, 1.2129E+07, 1.2203E+07, 1.2277E+07,
     + 1.2351E+07, 1.2426E+07, 1.2501E+07, 1.2576E+07, 1.2652E+07,
     +  1.2728E+07, 1.2804E+07, 1.2881E+07, 1.2958E+07, 1.3035E+07,
     + 1.3113E+07, 1.3191E+07, 1.3270E+07, 1.3349E+07, 1.3428E+07,
     +  1.3508E+07, 1.3588E+07, 1.3669E+07, 1.3749E+07, 1.3831E+07,
     + 1.3912E+07, 1.3994E+07, 1.4077E+07, 1.4159E+07, 1.4242E+07,
     +  1.4326E+07, 1.4410E+07, 1.4494E+07, 1.4579E+07, 1.4664E+07,
     + 1.4749E+07, 1.4835E+07, 1.4921E+07, 1.5008E+07, 1.5095E+07,
     +  1.5183E+07, 1.5271E+07, 1.5359E+07, 1.5448E+07, 1.5537E+07,
     + 1.5626E+07, 1.5716E+07, 1.5806E+07, 1.5897E+07, 1.5988E+07,
     +  1.6080E+07, 1.6172E+07, 1.6264E+07, 1.6357E+07, 1.6450E+07,
     + 1.6544E+07, 1.6638E+07, 1.6732E+07, 1.6827E+07, 1.6922E+07,
     +  1.7018E+07, 1.7114E+07, 1.7211E+07, 1.7308E+07, 1.7405E+07,
     + 1.7503E+07, 1.7602E+07, 1.7700E+07, 1.7800E+07, 1.7899E+07,
     +  1.7999E+07, 1.8100E+07, 1.8201E+07, 1.8303E+07, 1.8404E+07,
     + 1.8507E+07, 1.8610E+07, 1.8713E+07, 1.8817E+07, 1.8921E+07,
     +  1.9025E+07, 1.9130E+07, 1.9236E+07, 1.9342E+07, 1.9448E+07,
     + 1.9555E+07, 1.9663E+07, 1.9770E+07, 1.9879E+07, 1.9988E+07,
     +  2.0097E+07, 2.0207E+07, 2.0317E+07, 2.0427E+07, 2.0539E+07,
     + 2.0650E+07, 2.0762E+07, 2.0875E+07, 2.0988E+07, 2.1102E+07,
     +  2.1216E+07, 2.1330E+07, 2.1445E+07, 2.1561E+07, 2.1677E+07,
     + 2.1793E+07, 2.1910E+07, 2.2028E+07, 2.2146E+07, 2.2264E+07,
     +  2.2383E+07, 2.2503E+07, 2.2623E+07, 2.2743E+07, 2.2864E+07,
     + 2.2986E+07, 2.3108E+07, 2.3230E+07, 2.3354E+07, 2.3477E+07,
     +  2.3601E+07, 2.3726E+07, 2.3851E+07, 2.3977E+07, 2.4103E+07,
     + 2.4230E+07, 2.4357E+07, 2.4485E+07, 2.4613E+07, 2.4742E+07,
     +  2.4871E+07, 2.5001E+07, 2.5131E+07, 2.5262E+07, 2.5394E+07,
     + 2.5526E+07, 2.5659E+07, 2.5792E+07, 2.5925E+07, 2.6060E+07,
     +  2.6194E+07, 2.6330E+07, 2.6466E+07, 2.6602E+07, 2.6739E+07,
     + 2.6877E+07, 2.7015E+07, 2.7153E+07, 2.7293E+07, 2.7432E+07,
     +  2.7573E+07, 2.7714E+07, 2.7855E+07, 2.7997E+07, 2.8140E+07,
     + 2.8283E+07, 2.8427E+07, 2.8571E+07, 2.8716E+07, 2.8862E+07,
     +  2.9008E+07, 2.9155E+07, 2.9302E+07, 2.9450E+07, 2.9598E+07,
     + 2.9747E+07, 2.9897E+07, 3.0047E+07, 3.0198E+07, 3.0350E+07,
     +  3.0502E+07, 3.0654E+07, 3.0808E+07, 3.0961E+07, 3.1116E+07,
     + 3.1271E+07, 3.1427E+07, 3.1583E+07, 3.1740E+07, 3.1897E+07,
     +  3.2056E+07, 3.2214E+07, 3.2374E+07, 3.2534E+07, 3.2694E+07,
     + 3.2856E+07, 3.3017E+07, 3.3180E+07, 3.3343E+07, 3.3507E+07,
     +  3.3671E+07, 3.3836E+07, 3.4002E+07, 3.4168E+07, 3.4335E+07,
     + 3.4503E+07, 3.4671E+07, 3.4840E+07, 3.5010E+07, 3.5180E+07,
     +  3.5351E+07, 3.5522E+07, 3.5694E+07, 3.5867E+07, 3.6041E+07,
     + 3.6215E+07, 3.6390E+07, 3.6565E+07, 3.6741E+07, 3.6918E+07,
     +  3.7096E+07, 3.7274E+07, 3.7453E+07, 3.7632E+07, 3.7812E+07,
     + 3.7993E+07, 3.8175E+07, 3.8357E+07, 3.8540E+07, 3.8724E+07,
     +  3.8908E+07, 3.9093E+07, 3.9279E+07, 3.9465E+07, 3.9652E+07,
     + 3.9840E+07, 4.0028E+07, 4.0218E+07, 4.0408E+07, 4.0598E+07,
     +  4.0789E+07, 4.0981E+07, 4.1174E+07, 4.1368E+07, 4.1562E+07,
     + 4.1757E+07, 4.1952E+07, 4.2149E+07, 4.2346E+07, 4.2544E+07,
     +  4.2742E+07, 4.2941E+07, 4.3141E+07, 4.3342E+07, 4.3544E+07,
     + 4.3746E+07, 4.3949E+07, 4.4152E+07, 4.4357E+07, 4.4562E+07,
     +  4.4768E+07, 4.4975E+07, 4.5182E+07, 4.5390E+07, 4.5599E+07,
     + 4.5809E+07, 4.6019E+07, 4.6230E+07, 4.6442E+07, 4.6655E+07,
     +  4.6869E+07, 4.7083E+07, 4.7298E+07, 4.7514E+07, 4.7730E+07,
     + 4.7948E+07, 4.8166E+07, 4.8385E+07, 4.8605E+07, 4.8825E+07,
     +  4.9047E+07, 4.9269E+07, 4.9492E+07, 4.9715E+07, 4.9940E+07,
     + 5.0165E+07, 5.0391E+07, 5.0618E+07, 5.0846E+07, 5.1074E+07,
     +  5.1303E+07, 5.1534E+07, 5.1765E+07, 5.1996E+07, 5.2229E+07,
     + 5.2462E+07, 5.2696E+07, 5.2931E+07, 5.3167E+07, 5.3404E+07,
     +  5.3642E+07, 5.3880E+07, 5.4119E+07, 5.4359E+07, 5.4600E+07,
     + 5.4842E+07, 5.5084E+07, 5.5328E+07, 5.5572E+07, 5.5817E+07,
     +  5.6063E+07, 5.6310E+07, 5.6557E+07, 5.6806E+07, 5.7055E+07,
     + 5.7305E+07, 5.7557E+07, 5.7809E+07, 5.8061E+07, 5.8315E+07,
     +  5.8570E+07, 5.8825E+07, 5.9081E+07, 5.9339E+07, 5.9597E+07,
     + 5.9856E+07, 6.0116E+07, 6.0376E+07, 6.0638E+07, 6.0901E+07,
     +  6.1164E+07, 6.1428E+07, 6.1694E+07, 6.1960E+07, 6.2227E+07,
     + 6.2495E+07, 6.2764E+07, 6.3033E+07, 6.3304E+07, 6.3576E+07,
     +  6.3848E+07, 6.4121E+07, 6.4396E+07, 6.4671E+07, 6.4947E+07,
     + 6.5225E+07, 6.5503E+07, 6.5782E+07, 6.6061E+07, 6.6342E+07,
     +  6.6624E+07, 6.6907E+07, 6.7191E+07, 6.7475E+07, 6.7761E+07,
     + 6.8048E+07, 6.8335E+07, 6.8623E+07, 6.8913E+07, 6.9203E+07,
     +  6.9495E+07, 6.9787E+07, 7.0080E+07, 7.0374E+07, 7.0670E+07,
     + 7.0966E+07, 7.1263E+07, 7.1561E+07, 7.1860E+07, 7.2160E+07,
     +  7.2462E+07, 7.2764E+07, 7.3067E+07, 7.3371E+07, 7.3676E+07,
     + 7.3982E+07, 7.4289E+07, 7.4597E+07, 7.4906E+07, 7.5216E+07,
     +  7.5528E+07, 7.5840E+07, 7.6153E+07, 7.6467E+07, 7.6782E+07,
     + 7.7098E+07, 7.7416E+07, 7.7734E+07, 7.8053E+07, 7.8373E+07,
     +  7.8695E+07, 7.9017E+07, 7.9341E+07, 7.9665E+07, 7.9991E+07,
     + 8.0317E+07, 8.0645E+07, 8.0973E+07, 8.1303E+07, 8.1634E+07,
     +  8.1966E+07, 8.2299E+07, 8.2633E+07, 8.2968E+07, 8.3304E+07,
     + 8.3641E+07, 8.3979E+07, 8.4318E+07, 8.4659E+07, 8.5000E+07,
     +  8.5343E+07, 8.5687E+07, 8.6031E+07, 8.6377E+07, 8.6724E+07,
     + 8.7072E+07, 8.7421E+07, 8.7771E+07, 8.8123E+07, 8.8475E+07,
     +  8.8829E+07, 8.9183E+07, 8.9539E+07, 8.9896E+07, 9.0254E+07,
     + 9.0613E+07, 9.0973E+07, 9.1335E+07, 9.1697E+07, 9.2061E+07,
     +  9.2426E+07, 9.2792E+07, 9.3159E+07, 9.3527E+07, 9.3896E+07,
     + 9.4267E+07, 9.4638E+07, 9.5011E+07, 9.5385E+07, 9.5760E+07,
     +  9.6136E+07, 9.6514E+07, 9.6893E+07, 9.7272E+07, 9.7653E+07,
     + 9.8035E+07, 9.8419E+07, 9.8803E+07, 9.9189E+07, 9.9576E+07,
     +  9.9964E+07, 1.0035E+08, 1.0074E+08, 1.0113E+08, 1.0153E+08,
     + 1.0192E+08, 1.0232E+08, 1.0271E+08, 1.0311E+08, 1.0351E+08,
     +  1.0391E+08, 1.0431E+08, 1.0471E+08, 1.0512E+08, 1.0552E+08,
     + 1.0593E+08, 1.0634E+08, 1.0674E+08, 1.0715E+08, 1.0756E+08,
     +  1.0798E+08, 1.0839E+08, 1.0881E+08, 1.0922E+08, 1.0964E+08,
     + 1.1006E+08, 1.1048E+08, 1.1090E+08, 1.1132E+08, 1.1174E+08,
     +  1.1217E+08, 1.1260E+08, 1.1302E+08, 1.1345E+08, 1.1388E+08,
     + 1.1431E+08, 1.1475E+08, 1.1518E+08, 1.1561E+08, 1.1605E+08,
     +  1.1649E+08, 1.1693E+08, 1.1737E+08, 1.1781E+08, 1.1825E+08,
     + 1.1870E+08, 1.1914E+08, 1.1959E+08, 1.2004E+08, 1.2049E+08,
     +  1.2094E+08, 1.2139E+08, 1.2184E+08, 1.2230E+08, 1.2275E+08,
     + 1.2321E+08, 1.2367E+08, 1.2413E+08, 1.2459E+08, 1.2505E+08,
     +  1.2552E+08, 1.2598E+08, 1.2645E+08, 1.2692E+08, 1.2739E+08,
     + 1.2786E+08, 1.2833E+08, 1.2881E+08, 1.2928E+08, 1.2976E+08,
     +  1.3024E+08, 1.3072E+08, 1.3120E+08, 1.3168E+08, 1.3216E+08,
     + 1.3265E+08, 1.3313E+08, 1.3362E+08, 1.3411E+08, 1.3460E+08,
     +  1.3509E+08, 1.3559E+08, 1.3608E+08, 1.3658E+08, 1.3707E+08,
     + 1.3757E+08, 1.3807E+08, 1.3857E+08, 1.3908E+08, 1.3958E+08,
     +  1.4009E+08, 1.4060E+08, 1.4111E+08, 1.4162E+08, 1.4213E+08,
     + 1.4264E+08, 1.4316E+08, 1.4367E+08, 1.4419E+08, 1.4471E+08,
     +  1.4523E+08, 1.4575E+08, 1.4628E+08, 1.4680E+08, 1.4733E+08,
     + 1.4786E+08, 1.4839E+08, 1.4892E+08, 1.4945E+08, 1.4998E+08,
     +  1.5052E+08, 1.5106E+08, 1.5159E+08, 1.5213E+08, 1.5268E+08,
     + 1.5322E+08, 1.5376E+08, 1.5431E+08, 1.5486E+08, 1.5541E+08,
     +  1.5596E+08, 1.5651E+08, 1.5706E+08, 1.5762E+08, 1.5818E+08,
     + 1.5873E+08, 1.5929E+08, 1.5986E+08, 1.6042E+08, 1.6098E+08,
     +  1.6155E+08, 1.6212E+08, 1.6269E+08, 1.6326E+08, 1.6383E+08,
     + 1.6440E+08, 1.6498E+08, 1.6556E+08, 1.6614E+08, 1.6672E+08,
     +  1.6730E+08, 1.6788E+08, 1.6847E+08, 1.6905E+08, 1.6964E+08,
     + 1.7023E+08, 1.7082E+08, 1.7142E+08, 1.7201E+08, 1.7261E+08,
     +  1.7321E+08, 1.7381E+08, 1.7441E+08, 1.7501E+08, 1.7562E+08,
     + 1.7622E+08, 1.7683E+08, 1.7744E+08, 1.7805E+08, 1.7866E+08,
     +  1.7928E+08, 1.7990E+08, 1.8051E+08, 1.8113E+08, 1.8175E+08,
     + 1.8238E+08, 1.8300E+08, 1.8363E+08, 1.8426E+08, 1.8489E+08,
     +  1.8552E+08, 1.8615E+08, 1.8678E+08, 1.8742E+08, 1.8806E+08,
     + 1.8870E+08, 1.8934E+08, 1.8998E+08, 1.9063E+08, 1.9128E+08,
     +  1.9192E+08, 1.9257E+08, 1.9323E+08, 1.9388E+08, 1.9454E+08,
     + 1.9519E+08, 1.9585E+08, 1.9651E+08, 1.9717E+08, 1.9784E+08,
     +  1.9850E+08, 1.9917E+08, 1.9984E+08, 2.0051E+08, 2.0119E+08,
     + 2.0186E+08, 2.0254E+08, 2.0321E+08, 2.0389E+08, 2.0458E+08,
     +  2.0526E+08, 2.0595E+08, 2.0663E+08, 2.0732E+08, 2.0801E+08,
     + 2.0871E+08, 2.0940E+08, 2.1010E+08, 2.1079E+08, 2.1149E+08,
     +  2.1220E+08, 2.1290E+08, 2.1360E+08, 2.1431E+08, 2.1502E+08,
     + 2.1573E+08, 2.1644E+08, 2.1716E+08, 2.1788E+08, 2.1859E+08,
     +  2.1931E+08, 2.2004E+08, 2.2076E+08, 2.2149E+08, 2.2221E+08,
     + 2.2294E+08, 2.2368E+08, 2.2441E+08, 2.2514E+08, 2.2588E+08,
     +  2.2662E+08, 2.2736E+08, 2.2810E+08, 2.2885E+08, 2.2959E+08,
     + 2.3034E+08, 2.3109E+08, 2.3185E+08, 2.3260E+08, 2.3336E+08,
     +  2.3412E+08, 2.3488E+08, 2.3564E+08, 2.3640E+08, 2.3717E+08,
     + 2.3794E+08, 2.3870E+08, 2.3948E+08, 2.4025E+08, 2.4103E+08,
     +  2.4180E+08, 2.4258E+08, 2.4337E+08, 2.4415E+08, 2.4493E+08,
     + 2.4572E+08, 2.4651E+08, 2.4730E+08, 2.4810E+08, 2.4889E+08,
     +  2.4969E+08, 2.5049E+08, 2.5129E+08, 2.5209E+08, 2.5290E+08,
     + 2.5371E+08, 2.5452E+08, 2.5533E+08, 2.5614E+08, 2.5696E+08,
     +  2.5778E+08, 2.5860E+08, 2.5942E+08, 2.6024E+08, 2.6107E+08,
     + 2.6190E+08, 2.6273E+08, 2.6356E+08, 2.6439E+08, 2.6523E+08,
     +  2.6607E+08, 2.6691E+08, 2.6775E+08, 2.6860E+08, 2.6944E+08,
     + 2.7029E+08, 2.7114E+08, 2.7200E+08, 2.7285E+08, 2.7371E+08,
     +  2.7457E+08, 2.7543E+08, 2.7629E+08, 2.7716E+08, 2.7803E+08,
     + 2.7890E+08, 2.7977E+08, 2.8064E+08, 2.8152E+08, 2.8240E+08,
     +  2.8328E+08, 2.8416E+08, 2.8505E+08, 2.8593E+08, 2.8682E+08,
     + 2.8772E+08, 2.8861E+08, 2.8951E+08, 2.9040E+08, 2.9130E+08,
     +  2.9221E+08, 2.9311E+08, 2.9402E+08, 2.9493E+08, 2.9584E+08,
     + 2.9675E+08, 2.9767E+08, 2.9859E+08, 2.9951E+08, 3.0043E+08,
     +  3.0135E+08, 3.0228E+08, 3.0321E+08, 3.0414E+08, 3.0508E+08,
     + 3.0601E+08, 3.0695E+08, 3.0789E+08, 3.0883E+08, 3.0978E+08,
     +  3.1073E+08, 3.1167E+08, 3.1263E+08, 3.1358E+08, 3.1454E+08,
     + 3.1550E+08, 3.1646E+08, 3.1742E+08, 3.1839E+08, 3.1935E+08,
     +  3.2032E+08, 3.2130E+08, 3.2227E+08, 3.2325E+08, 3.2423E+08,
     + 3.2521E+08, 3.2619E+08, 3.2718E+08, 3.2817E+08, 3.2916E+08,
     +  3.3015E+08, 3.3115E+08, 3.3215E+08, 3.3315E+08, 3.3415E+08,
     + 3.3516E+08, 3.3616E+08, 3.3717E+08, 3.3819E+08, 3.3920E+08,
     +  3.4022E+08, 3.4124E+08, 3.4226E+08, 3.4329E+08, 3.4431E+08,
     + 3.4534E+08, 3.4637E+08, 3.4741E+08, 3.4844E+08, 3.4948E+08,
     +  3.5052E+08, 3.5157E+08, 3.5262E+08, 3.5366E+08, 3.5472E+08,
     + 3.5577E+08, 3.5682E+08, 3.5788E+08, 3.5894E+08, 3.6001E+08,
     +  3.6107E+08, 3.6214E+08, 3.6321E+08, 3.6429E+08, 3.6536E+08,
     + 3.6644E+08, 3.6752E+08, 3.6861E+08, 3.6969E+08, 3.7078E+08,
     +  3.7187E+08, 3.7296E+08, 3.7406E+08, 3.7516E+08, 3.7626E+08,
     + 3.7736E+08, 3.7847E+08, 3.7958E+08, 3.8069E+08, 3.8180E+08,
     +  3.8292E+08, 3.8404E+08, 3.8516E+08, 3.8629E+08, 3.8741E+08,
     + 3.8854E+08, 3.8967E+08, 3.9081E+08, 3.9195E+08, 3.9308E+08,
     +  3.9423E+08, 3.9537E+08, 3.9652E+08, 3.9767E+08, 3.9882E+08,
     + 3.9998E+08, 4.0114E+08, 4.0230E+08, 4.0346E+08, 4.0463E+08,
     +  4.0579E+08, 4.0696E+08, 4.0814E+08, 4.0932E+08, 4.1049E+08,
     + 4.1168E+08, 4.1286E+08, 4.1405E+08, 4.1524E+08, 4.1643E+08,
     +  4.1763E+08, 4.1882E+08, 4.2002E+08, 4.2123E+08, 4.2243E+08,
     + 4.2364E+08, 4.2485E+08, 4.2607E+08, 4.2728E+08, 4.2850E+08,
     +  4.2973E+08, 4.3095E+08, 4.3218E+08, 4.3341E+08, 4.3464E+08,
     + 4.3588E+08, 4.3712E+08, 4.3836E+08, 4.3961E+08, 4.4085E+08,
     +  4.4210E+08, 4.4336E+08, 4.4461E+08, 4.4587E+08, 4.4713E+08,
     + 4.4839E+08, 4.4966E+08, 4.5093E+08, 4.5220E+08, 4.5348E+08,
     +  4.5476E+08, 4.5604E+08, 4.5732E+08, 4.5861E+08, 4.5990E+08,
     + 4.6119E+08, 4.6248E+08, 4.6378E+08, 4.6508E+08, 4.6639E+08,
     +  4.6769E+08, 4.6900E+08, 4.7031E+08, 4.7163E+08, 4.7295E+08,
     + 4.7427E+08, 4.7559E+08, 4.7692E+08, 4.7825E+08, 4.7958E+08,
     +  4.8092E+08, 4.8226E+08, 4.8360E+08, 4.8494E+08, 4.8629E+08,
     + 4.8764E+08, 4.8899E+08, 4.9035E+08, 4.9171E+08, 4.9307E+08,
     +  4.9443E+08, 4.9580E+08, 4.9717E+08, 4.9855E+08, 4.9992E+08,
     + 5.0130E+08, 5.0269E+08, 5.0407E+08, 5.0546E+08, 5.0685E+08,
     +  5.0825E+08, 5.0964E+08, 5.1105E+08, 5.1245E+08, 5.1386E+08,
     + 5.1527E+08, 5.1668E+08, 5.1810E+08, 5.1951E+08, 5.2094E+08,
     +  5.2236E+08, 5.2379E+08, 5.2522E+08, 5.2666E+08, 5.2809E+08,
     + 5.2953E+08, 5.3098E+08, 5.3242E+08, 5.3387E+08, 5.3533E+08,
     +  5.3678E+08, 5.3824E+08, 5.3970E+08, 5.4117E+08, 5.4264E+08,
     + 5.4411E+08, 5.4559E+08, 5.4706E+08, 5.4854E+08, 5.5003E+08,
     +  5.5152E+08, 5.5301E+08, 5.5450E+08, 5.5600E+08, 5.5750E+08,
     + 5.5900E+08, 5.6051E+08, 5.6202E+08, 5.6353E+08, 5.6504E+08,
     +  5.6656E+08, 5.6809E+08, 5.6961E+08, 5.7114E+08, 5.7267E+08,
     + 5.7421E+08, 5.7575E+08, 5.7729E+08, 5.7883E+08, 5.8038E+08,
     +  5.8193E+08, 5.8349E+08, 5.8505E+08, 5.8661E+08, 5.8817E+08,
     + 5.8974E+08, 5.9131E+08, 5.9289E+08, 5.9446E+08, 5.9604E+08,
     +  5.9763E+08, 5.9922E+08, 6.0081E+08, 6.0240E+08, 6.0400E+08,
     + 6.0560E+08, 6.0720E+08, 6.0881E+08, 6.1042E+08, 6.1204E+08,
     +  6.1365E+08, 6.1528E+08, 6.1690E+08, 6.1853E+08, 6.2016E+08,
     + 6.2179E+08, 6.2343E+08, 6.2507E+08, 6.2672E+08, 6.2837E+08,
     +  6.3002E+08, 6.3167E+08, 6.3333E+08, 6.3499E+08, 6.3666E+08,
     + 6.3833E+08, 6.4000E+08, 6.4167E+08, 6.4335E+08, 6.4503E+08,
     +  6.4672E+08, 6.4841E+08, 6.5010E+08, 6.5180E+08, 6.5350E+08,
     + 6.5520E+08, 6.5691E+08, 6.5862E+08, 6.6033E+08, 6.6205E+08,
     +  6.6377E+08, 6.6549E+08, 6.6722E+08, 6.6895E+08, 6.7069E+08,
     + 6.7242E+08, 6.7417E+08, 6.7591E+08, 6.7766E+08, 6.7941E+08,
     +  6.8117E+08, 6.8293E+08, 6.8469E+08, 6.8646E+08, 6.8823E+08,
     + 6.9000E+08, 6.9178E+08, 6.9356E+08, 6.9535E+08, 6.9713E+08,
     +  6.9893E+08, 7.0072E+08, 7.0252E+08, 7.0432E+08, 7.0613E+08,
     + 7.0794E+08, 7.0975E+08, 7.1157E+08, 7.1339E+08, 7.1522E+08,
     +  7.1705E+08, 7.1888E+08, 7.2071E+08, 7.2255E+08, 7.2440E+08,
     + 7.2624E+08, 7.2809E+08, 7.2995E+08, 7.3180E+08, 7.3367E+08,
     +  7.3553E+08, 7.3740E+08, 7.3927E+08, 7.4115E+08, 7.4303E+08,
     + 7.4491E+08, 7.4680E+08, 7.4869E+08, 7.5059E+08, 7.5249E+08,
     +  7.5439E+08, 7.5630E+08, 7.5821E+08, 7.6012E+08, 7.6204E+08,
     + 7.6396E+08, 7.6589E+08, 7.6782E+08, 7.6975E+08, 7.7169E+08,
     +  7.7363E+08, 7.7557E+08, 7.7752E+08, 7.7947E+08, 7.8143E+08,
     + 7.8339E+08, 7.8535E+08, 7.8732E+08, 7.8929E+08, 7.9127E+08,
     +  7.9325E+08, 7.9523E+08, 7.9722E+08, 7.9921E+08, 8.0120E+08,
     + 8.0320E+08, 8.0521E+08, 8.0721E+08, 8.0922E+08, 8.1124E+08,
     +  8.1326E+08, 8.1528E+08, 8.1731E+08, 8.1934E+08, 8.2137E+08,
     + 8.2341E+08, 8.2545E+08, 8.2750E+08, 8.2955E+08, 8.3160E+08,
     +  8.3366E+08, 8.3572E+08, 8.3779E+08, 8.3986E+08, 8.4193E+08,
     + 8.4401E+08, 8.4610E+08, 8.4818E+08, 8.5027E+08, 8.5237E+08,
     +  8.5447E+08, 8.5657E+08, 8.5868E+08, 8.6079E+08, 8.6290E+08,
     + 8.6502E+08, 8.6714E+08, 8.6927E+08, 8.7140E+08, 8.7354E+08,
     +  8.7568E+08, 8.7782E+08, 8.7997E+08, 8.8212E+08, 8.8428E+08,
     + 8.8644E+08, 8.8860E+08, 8.9077E+08, 8.9294E+08, 8.9512E+08,
     +  8.9730E+08, 8.9948E+08, 9.0167E+08, 9.0387E+08, 9.0606E+08,
     + 9.0827E+08, 9.1047E+08, 9.1268E+08, 9.1490E+08, 9.1712E+08,
     +  9.1934E+08, 9.2157E+08, 9.2380E+08, 9.2603E+08, 9.2827E+08,
     + 9.3052E+08, 9.3276E+08, 9.3502E+08, 9.3727E+08, 9.3954E+08,
     +  9.4180E+08, 9.4407E+08, 9.4635E+08, 9.4862E+08, 9.5091E+08,
     + 9.5319E+08, 9.5549E+08, 9.5778E+08, 9.6008E+08, 9.6239E+08,
     +  9.6469E+08, 9.6701E+08, 9.6932E+08, 9.7165E+08, 9.7397E+08,
     + 9.7630E+08, 9.7864E+08, 9.8098E+08, 9.8332E+08, 9.8567E+08,
     +  9.8802E+08, 9.9038E+08, 9.9274E+08, 9.9511E+08, 9.9748E+08,
     + 9.9985E+08, 1.0022E+09, 1.0046E+09, 1.0070E+09, 1.0094E+09,
     +  1.0118E+09, 1.0142E+09, 1.0166E+09, 1.0190E+09, 1.0214E+09,
     + 1.0238E+09, 1.0263E+09, 1.0287E+09, 1.0311E+09, 1.0336E+09,
     +  1.0360E+09, 1.0385E+09, 1.0409E+09, 1.0434E+09, 1.0458E+09,
     + 1.0483E+09, 1.0508E+09, 1.0532E+09, 1.0557E+09, 1.0582E+09,
     +  1.0607E+09, 1.0632E+09, 1.0657E+09, 1.0682E+09, 1.0707E+09,
     + 1.0732E+09, 1.0757E+09, 1.0782E+09, 1.0807E+09, 1.0833E+09,
     +  1.0858E+09, 1.0883E+09, 1.0909E+09, 1.0934E+09, 1.0960E+09,
     + 1.0985E+09, 1.1011E+09, 1.1037E+09, 1.1062E+09, 1.1088E+09,
     +  1.1114E+09, 1.1140E+09, 1.1166E+09, 1.1192E+09, 1.1218E+09,
     + 1.1244E+09, 1.1270E+09, 1.1296E+09, 1.1322E+09, 1.1348E+09,
     +  1.1375E+09, 1.1401E+09, 1.1427E+09, 1.1454E+09, 1.1480E+09,
     + 1.1507E+09, 1.1533E+09, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT(18,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_N2O   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 5),QofT( 5,NT),Q(NT)
      data (Tmx(I),I=1,5)/5000.,3500.,3500.,3500.,3500./
      data xgj/ 9.,6.,6.,9.,54. /
c...      N2O
c...        --       446
      data (QofT( 1,J),J=1,2408)/ 1.8341E+01, 3.3050E+01, 4.7912E+01,
     +  6.2808E+01, 7.7718E+01, 9.2635E+01, 1.0755E+02, 1.2248E+02,
     + 1.3740E+02, 1.5233E+02, 1.6725E+02, 1.8218E+02, 1.9711E+02,
     +  2.1203E+02, 2.2696E+02, 2.4189E+02, 2.5682E+02, 2.7174E+02,
     + 2.8667E+02, 3.0160E+02, 3.3146E+02, 3.6132E+02, 3.9118E+02,
     +  4.2103E+02, 4.5089E+02, 4.8075E+02, 5.1061E+02, 5.4047E+02,
     + 5.7033E+02, 6.0019E+02, 6.3005E+02, 6.5991E+02, 6.8977E+02,
     +  7.1964E+02, 7.4950E+02, 7.7936E+02, 8.0922E+02, 8.3908E+02,
     + 8.6894E+02, 8.9881E+02, 9.2867E+02, 9.5853E+02, 9.8840E+02,
     +  1.0183E+03, 1.0481E+03, 1.0780E+03, 1.1079E+03, 1.1377E+03,
     + 1.1676E+03, 1.1975E+03, 1.2274E+03, 1.2573E+03, 1.2872E+03,
     +  1.3171E+03, 1.3470E+03, 1.3769E+03, 1.4068E+03, 1.4368E+03,
     + 1.4667E+03, 1.4967E+03, 1.5267E+03, 1.5567E+03, 1.5868E+03,
     +  1.6168E+03, 1.6469E+03, 1.6770E+03, 1.7072E+03, 1.7373E+03,
     + 1.7676E+03, 1.7978E+03, 1.8281E+03, 1.8585E+03, 1.8889E+03,
     +  1.9193E+03, 1.9498E+03, 1.9804E+03, 2.0110E+03, 2.0417E+03,
     + 2.0725E+03, 2.1033E+03, 2.1342E+03, 2.1652E+03, 2.1962E+03,
     +  2.2274E+03, 2.2586E+03, 2.2899E+03, 2.3214E+03, 2.3529E+03,
     + 2.3845E+03, 2.4162E+03, 2.4481E+03, 2.4800E+03, 2.5121E+03,
     +  2.5443E+03, 2.5766E+03, 2.6090E+03, 2.6416E+03, 2.6743E+03,
     + 2.7071E+03, 2.7400E+03, 2.7732E+03, 2.8064E+03, 2.8398E+03,
     +  2.8733E+03, 2.9071E+03, 2.9409E+03, 2.9749E+03, 3.0091E+03,
     + 3.0435E+03, 3.0780E+03, 3.1127E+03, 3.1475E+03, 3.1826E+03,
     +  3.2178E+03, 3.2532E+03, 3.2888E+03, 3.3246E+03, 3.3605E+03,
     + 3.3967E+03, 3.4331E+03, 3.4696E+03, 3.5064E+03, 3.5434E+03,
     +  3.5805E+03, 3.6179E+03, 3.6555E+03, 3.6933E+03, 3.7314E+03,
     + 3.7696E+03, 3.8081E+03, 3.8468E+03, 3.8857E+03, 3.9248E+03,
     +  3.9642E+03, 4.0038E+03, 4.0437E+03, 4.0838E+03, 4.1241E+03,
     + 4.1647E+03, 4.2055E+03, 4.2466E+03, 4.2879E+03, 4.3295E+03,
     +  4.3713E+03, 4.4134E+03, 4.4557E+03, 4.4983E+03, 4.5412E+03,
     + 4.5844E+03, 4.6278E+03, 4.6714E+03, 4.7154E+03, 4.7596E+03,
     +  4.8041E+03, 4.8489E+03, 4.8939E+03, 4.9393E+03, 4.9849E+03,
     + 5.0308E+03, 5.0770E+03, 5.1235E+03, 5.1703E+03, 5.2174E+03,
     +  5.2648E+03, 5.3124E+03, 5.3604E+03, 5.4087E+03, 5.4573E+03,
     + 5.5062E+03, 5.5554E+03, 5.6049E+03, 5.6548E+03, 5.7049E+03,
     +  5.7554E+03, 5.8062E+03, 5.8573E+03, 5.9088E+03, 5.9605E+03,
     + 6.0126E+03, 6.0650E+03, 6.1178E+03, 6.1709E+03, 6.2243E+03,
     +  6.2781E+03, 6.3322E+03, 6.3867E+03, 6.4415E+03, 6.4966E+03,
     + 6.5521E+03, 6.6080E+03, 6.6642E+03, 6.7208E+03, 6.7777E+03,
     +  6.8350E+03, 6.8926E+03, 6.9506E+03, 7.0090E+03, 7.0677E+03,
     + 7.1268E+03, 7.1863E+03, 7.2462E+03, 7.3064E+03, 7.3670E+03,
     +  7.4280E+03, 7.4894E+03, 7.5512E+03, 7.6133E+03, 7.6758E+03,
     + 7.7388E+03, 7.8021E+03, 7.8658E+03, 7.9299E+03, 7.9945E+03,
     +  8.0594E+03, 8.1247E+03, 8.1905E+03, 8.2566E+03, 8.3232E+03,
     + 8.3901E+03, 8.4575E+03, 8.5253E+03, 8.5935E+03, 8.6622E+03,
     +  8.7313E+03, 8.8007E+03, 8.8707E+03, 8.9410E+03, 9.0118E+03,
     + 9.0830E+03, 9.1547E+03, 9.2268E+03, 9.2994E+03, 9.3723E+03,
     +  9.4458E+03, 9.5197E+03, 9.5940E+03, 9.6688E+03, 9.7440E+03,
     + 9.8198E+03, 9.8959E+03, 9.9726E+03, 1.0050E+04, 1.0127E+04,
     +  1.0205E+04, 1.0284E+04, 1.0363E+04, 1.0442E+04, 1.0522E+04,
     + 1.0603E+04, 1.0683E+04, 1.0765E+04, 1.0847E+04, 1.0929E+04,
     +  1.1012E+04, 1.1095E+04, 1.1179E+04, 1.1264E+04, 1.1348E+04,
     + 1.1434E+04, 1.1520E+04, 1.1606E+04, 1.1693E+04, 1.1780E+04,
     +  1.1868E+04, 1.1957E+04, 1.2046E+04, 1.2135E+04, 1.2225E+04,
     + 1.2316E+04, 1.2407E+04, 1.2499E+04, 1.2591E+04, 1.2684E+04,
     +  1.2777E+04, 1.2871E+04, 1.2965E+04, 1.3060E+04, 1.3155E+04,
     + 1.3251E+04, 1.3348E+04, 1.3445E+04, 1.3543E+04, 1.3641E+04,
     +  1.3740E+04, 1.3839E+04, 1.3939E+04, 1.4040E+04, 1.4141E+04,
     + 1.4243E+04, 1.4345E+04, 1.4448E+04, 1.4552E+04, 1.4656E+04,
     +  1.4760E+04, 1.4866E+04, 1.4972E+04, 1.5078E+04, 1.5185E+04,
     + 1.5293E+04, 1.5401E+04, 1.5510E+04, 1.5620E+04, 1.5730E+04,
     +  1.5841E+04, 1.5952E+04, 1.6064E+04, 1.6177E+04, 1.6290E+04,
     + 1.6404E+04, 1.6519E+04, 1.6634E+04, 1.6750E+04, 1.6867E+04,
     +  1.6984E+04, 1.7102E+04, 1.7220E+04, 1.7340E+04, 1.7459E+04,
     + 1.7580E+04, 1.7701E+04, 1.7823E+04, 1.7946E+04, 1.8069E+04,
     +  1.8193E+04, 1.8317E+04, 1.8443E+04, 1.8569E+04, 1.8695E+04,
     + 1.8823E+04, 1.8951E+04, 1.9080E+04, 1.9209E+04, 1.9339E+04,
     +  1.9470E+04, 1.9602E+04, 1.9734E+04, 1.9867E+04, 2.0001E+04,
     + 2.0136E+04, 2.0271E+04, 2.0407E+04, 2.0544E+04, 2.0681E+04,
     +  2.0820E+04, 2.0959E+04, 2.1098E+04, 2.1239E+04, 2.1380E+04,
     + 2.1522E+04, 2.1665E+04, 2.1808E+04, 2.1953E+04, 2.2098E+04,
     +  2.2244E+04, 2.2390E+04, 2.2538E+04, 2.2686E+04, 2.2835E+04,
     + 2.2985E+04, 2.3135E+04, 2.3287E+04, 2.3439E+04, 2.3592E+04,
     +  2.3746E+04, 2.3900E+04, 2.4056E+04, 2.4212E+04, 2.4369E+04,
     + 2.4527E+04, 2.4686E+04, 2.4846E+04, 2.5006E+04, 2.5167E+04,
     +  2.5329E+04, 2.5492E+04, 2.5656E+04, 2.5821E+04, 2.5987E+04,
     + 2.6153E+04, 2.6320E+04, 2.6488E+04, 2.6657E+04, 2.6827E+04,
     +  2.6998E+04, 2.7170E+04, 2.7342E+04, 2.7516E+04, 2.7690E+04,
     + 2.7865E+04, 2.8041E+04, 2.8219E+04, 2.8397E+04, 2.8575E+04,
     +  2.8755E+04, 2.8936E+04, 2.9118E+04, 2.9300E+04, 2.9484E+04,
     + 2.9668E+04, 2.9853E+04, 3.0040E+04, 3.0227E+04, 3.0415E+04,
     +  3.0604E+04, 3.0795E+04, 3.0986E+04, 3.1178E+04, 3.1371E+04,
     + 3.1565E+04, 3.1760E+04, 3.1956E+04, 3.2153E+04, 3.2351E+04,
     +  3.2550E+04, 3.2750E+04, 3.2950E+04, 3.3152E+04, 3.3355E+04,
     + 3.3559E+04, 3.3764E+04, 3.3970E+04, 3.4177E+04, 3.4385E+04,
     +  3.4595E+04, 3.4805E+04, 3.5016E+04, 3.5228E+04, 3.5441E+04,
     + 3.5656E+04, 3.5871E+04, 3.6087E+04, 3.6305E+04, 3.6524E+04,
     +  3.6743E+04, 3.6964E+04, 3.7186E+04, 3.7409E+04, 3.7633E+04,
     + 3.7858E+04, 3.8084E+04, 3.8311E+04, 3.8540E+04, 3.8769E+04,
     +  3.9000E+04, 3.9231E+04, 3.9464E+04, 3.9698E+04, 3.9933E+04,
     + 4.0170E+04, 4.0407E+04, 4.0646E+04, 4.0885E+04, 4.1126E+04,
     +  4.1368E+04, 4.1611E+04, 4.1856E+04, 4.2101E+04, 4.2348E+04,
     + 4.2596E+04, 4.2845E+04, 4.3095E+04, 4.3347E+04, 4.3599E+04,
     +  4.3853E+04, 4.4108E+04, 4.4365E+04, 4.4622E+04, 4.4881E+04,
     + 4.5141E+04, 4.5402E+04, 4.5664E+04, 4.5928E+04, 4.6193E+04,
     +  4.6459E+04, 4.6726E+04, 4.6995E+04, 4.7265E+04, 4.7536E+04,
     + 4.7808E+04, 4.8082E+04, 4.8357E+04, 4.8633E+04, 4.8911E+04,
     +  4.9190E+04, 4.9470E+04, 4.9752E+04, 5.0034E+04, 5.0318E+04,
     + 5.0604E+04, 5.0891E+04, 5.1179E+04, 5.1468E+04, 5.1759E+04,
     +  5.2051E+04, 5.2344E+04, 5.2639E+04, 5.2935E+04, 5.3233E+04,
     + 5.3531E+04, 5.3832E+04, 5.4133E+04, 5.4436E+04, 5.4741E+04,
     +  5.5046E+04, 5.5354E+04, 5.5662E+04, 5.5972E+04, 5.6283E+04,
     + 5.6596E+04, 5.6910E+04, 5.7226E+04, 5.7543E+04, 5.7862E+04,
     +  5.8182E+04, 5.8503E+04, 5.8826E+04, 5.9150E+04, 5.9476E+04,
     + 5.9803E+04, 6.0132E+04, 6.0462E+04, 6.0794E+04, 6.1127E+04,
     +  6.1461E+04, 6.1797E+04, 6.2135E+04, 6.2474E+04, 6.2815E+04,
     + 6.3157E+04, 6.3501E+04, 6.3846E+04, 6.4193E+04, 6.4541E+04,
     +  6.4891E+04, 6.5242E+04, 6.5595E+04, 6.5950E+04, 6.6306E+04,
     + 6.6663E+04, 6.7022E+04, 6.7383E+04, 6.7746E+04, 6.8109E+04,
     +  6.8475E+04, 6.8842E+04, 6.9211E+04, 6.9581E+04, 6.9953E+04,
     + 7.0327E+04, 7.0702E+04, 7.1079E+04, 7.1458E+04, 7.1838E+04,
     +  7.2219E+04, 7.2603E+04, 7.2988E+04, 7.3375E+04, 7.3763E+04,
     + 7.4153E+04, 7.4545E+04, 7.4939E+04, 7.5334E+04, 7.5731E+04,
     +  7.6130E+04, 7.6530E+04, 7.6932E+04, 7.7336E+04, 7.7741E+04,
     + 7.8148E+04, 7.8557E+04, 7.8968E+04, 7.9381E+04, 7.9795E+04,
     +  8.0211E+04, 8.0629E+04, 8.1048E+04, 8.1470E+04, 8.1893E+04,
     + 8.2318E+04, 8.2744E+04, 8.3173E+04, 8.3603E+04, 8.4035E+04,
     +  8.4469E+04, 8.4905E+04, 8.5342E+04, 8.5782E+04, 8.6223E+04,
     + 8.6666E+04, 8.7111E+04, 8.7558E+04, 8.8007E+04, 8.8457E+04,
     +  8.8910E+04, 8.9364E+04, 8.9820E+04, 9.0279E+04, 9.0739E+04,
     + 9.1201E+04, 9.1664E+04, 9.2130E+04, 9.2598E+04, 9.3067E+04,
     +  9.3539E+04, 9.4013E+04, 9.4488E+04, 9.4965E+04, 9.5445E+04,
     + 9.5926E+04, 9.6410E+04, 9.6895E+04, 9.7382E+04, 9.7871E+04,
     +  9.8363E+04, 9.8856E+04, 9.9351E+04, 9.9849E+04, 1.0035E+05,
     + 1.0085E+05, 1.0135E+05, 1.0186E+05, 1.0237E+05, 1.0288E+05,
     +  1.0339E+05, 1.0390E+05, 1.0442E+05, 1.0493E+05, 1.0545E+05,
     + 1.0598E+05, 1.0650E+05, 1.0703E+05, 1.0755E+05, 1.0809E+05,
     +  1.0862E+05, 1.0915E+05, 1.0969E+05, 1.1023E+05, 1.1077E+05,
     + 1.1131E+05, 1.1186E+05, 1.1241E+05, 1.1296E+05, 1.1351E+05,
     +  1.1406E+05, 1.1462E+05, 1.1518E+05, 1.1574E+05, 1.1630E+05,
     + 1.1687E+05, 1.1743E+05, 1.1800E+05, 1.1857E+05, 1.1915E+05,
     +  1.1973E+05, 1.2030E+05, 1.2088E+05, 1.2147E+05, 1.2205E+05,
     + 1.2264E+05, 1.2323E+05, 1.2382E+05, 1.2442E+05, 1.2501E+05,
     +  1.2561E+05, 1.2622E+05, 1.2682E+05, 1.2743E+05, 1.2803E+05,
     + 1.2865E+05, 1.2926E+05, 1.2987E+05, 1.3049E+05, 1.3111E+05,
     +  1.3174E+05, 1.3236E+05, 1.3299E+05, 1.3362E+05, 1.3425E+05,
     + 1.3488E+05, 1.3552E+05, 1.3616E+05, 1.3680E+05, 1.3745E+05,
     +  1.3810E+05, 1.3874E+05, 1.3940E+05, 1.4005E+05, 1.4071E+05,
     + 1.4137E+05, 1.4203E+05, 1.4269E+05, 1.4336E+05, 1.4403E+05,
     +  1.4470E+05, 1.4538E+05, 1.4605E+05, 1.4673E+05, 1.4741E+05,
     + 1.4810E+05, 1.4878E+05, 1.4947E+05, 1.5017E+05, 1.5086E+05,
     +  1.5156E+05, 1.5226E+05, 1.5296E+05, 1.5367E+05, 1.5437E+05,
     + 1.5508E+05, 1.5580E+05, 1.5651E+05, 1.5723E+05, 1.5795E+05,
     +  1.5868E+05, 1.5940E+05, 1.6013E+05, 1.6086E+05, 1.6160E+05,
     + 1.6233E+05, 1.6307E+05, 1.6381E+05, 1.6456E+05, 1.6531E+05,
     +  1.6606E+05, 1.6681E+05, 1.6757E+05, 1.6833E+05, 1.6909E+05,
     + 1.6985E+05, 1.7062E+05, 1.7139E+05, 1.7216E+05, 1.7294E+05,
     +  1.7371E+05, 1.7449E+05, 1.7528E+05, 1.7606E+05, 1.7685E+05,
     + 1.7765E+05, 1.7844E+05, 1.7924E+05, 1.8004E+05, 1.8084E+05,
     +  1.8165E+05, 1.8246E+05, 1.8327E+05, 1.8408E+05, 1.8490E+05,
     + 1.8572E+05, 1.8655E+05, 1.8737E+05, 1.8820E+05, 1.8903E+05,
     +  1.8987E+05, 1.9071E+05, 1.9155E+05, 1.9239E+05, 1.9324E+05,
     + 1.9409E+05, 1.9494E+05, 1.9580E+05, 1.9666E+05, 1.9752E+05,
     +  1.9838E+05, 1.9925E+05, 2.0012E+05, 2.0100E+05, 2.0187E+05,
     + 2.0275E+05, 2.0364E+05, 2.0452E+05, 2.0541E+05, 2.0631E+05,
     +  2.0720E+05, 2.0810E+05, 2.0900E+05, 2.0991E+05, 2.1081E+05,
     + 2.1172E+05, 2.1264E+05, 2.1355E+05, 2.1447E+05, 2.1540E+05,
     +  2.1632E+05, 2.1725E+05, 2.1819E+05, 2.1912E+05, 2.2006E+05,
     + 2.2100E+05, 2.2195E+05, 2.2290E+05, 2.2385E+05, 2.2480E+05,
     +  2.2576E+05, 2.2672E+05, 2.2769E+05, 2.2866E+05, 2.2963E+05,
     + 2.3060E+05, 2.3158E+05, 2.3256E+05, 2.3354E+05, 2.3453E+05,
     +  2.3552E+05, 2.3652E+05, 2.3751E+05, 2.3851E+05, 2.3952E+05,
     + 2.4052E+05, 2.4154E+05, 2.4255E+05, 2.4357E+05, 2.4459E+05,
     +  2.4561E+05, 2.4664E+05, 2.4767E+05, 2.4870E+05, 2.4974E+05,
     + 2.5078E+05, 2.5182E+05, 2.5287E+05, 2.5392E+05, 2.5498E+05,
     +  2.5603E+05, 2.5710E+05, 2.5816E+05, 2.5923E+05, 2.6030E+05,
     + 2.6137E+05, 2.6245E+05, 2.6354E+05, 2.6462E+05, 2.6571E+05,
     +  2.6680E+05, 2.6790E+05, 2.6900E+05, 2.7010E+05, 2.7121E+05,
     + 2.7232E+05, 2.7343E+05, 2.7455E+05, 2.7567E+05, 2.7679E+05,
     +  2.7792E+05, 2.7905E+05, 2.8018E+05, 2.8132E+05, 2.8246E+05,
     + 2.8361E+05, 2.8476E+05, 2.8591E+05, 2.8707E+05, 2.8823E+05,
     +  2.8939E+05, 2.9056E+05, 2.9173E+05, 2.9290E+05, 2.9408E+05,
     + 2.9526E+05, 2.9645E+05, 2.9764E+05, 2.9883E+05, 3.0003E+05,
     +  3.0123E+05, 3.0243E+05, 3.0364E+05, 3.0485E+05, 3.0607E+05,
     + 3.0729E+05, 3.0851E+05, 3.0974E+05, 3.1097E+05, 3.1220E+05,
     +  3.1344E+05, 3.1468E+05, 3.1592E+05, 3.1717E+05, 3.1843E+05,
     + 3.1968E+05, 3.2094E+05, 3.2221E+05, 3.2347E+05, 3.2475E+05,
     +  3.2602E+05, 3.2730E+05, 3.2859E+05, 3.2987E+05, 3.3116E+05,
     + 3.3246E+05, 3.3376E+05, 3.3506E+05, 3.3637E+05, 3.3768E+05,
     +  3.3899E+05, 3.4031E+05, 3.4163E+05, 3.4296E+05, 3.4429E+05,
     + 3.4562E+05, 3.4696E+05, 3.4830E+05, 3.4965E+05, 3.5100E+05,
     +  3.5235E+05, 3.5371E+05, 3.5507E+05, 3.5644E+05, 3.5781E+05,
     + 3.5918E+05, 3.6056E+05, 3.6194E+05, 3.6333E+05, 3.6472E+05,
     +  3.6611E+05, 3.6751E+05, 3.6891E+05, 3.7031E+05, 3.7172E+05,
     + 3.7314E+05, 3.7456E+05, 3.7598E+05, 3.7741E+05, 3.7884E+05,
     +  3.8027E+05, 3.8171E+05, 3.8315E+05, 3.8460E+05, 3.8605E+05,
     + 3.8750E+05, 3.8896E+05, 3.9043E+05, 3.9190E+05, 3.9337E+05,
     +  3.9484E+05, 3.9632E+05, 3.9781E+05, 3.9930E+05, 4.0079E+05,
     + 4.0228E+05, 4.0379E+05, 4.0529E+05, 4.0680E+05, 4.0831E+05,
     +  4.0983E+05, 4.1135E+05, 4.1288E+05, 4.1441E+05, 4.1595E+05,
     + 4.1748E+05, 4.1903E+05, 4.2058E+05, 4.2213E+05, 4.2368E+05,
     +  4.2524E+05, 4.2681E+05, 4.2838E+05, 4.2995E+05, 4.3153E+05,
     + 4.3311E+05, 4.3470E+05, 4.3629E+05, 4.3788E+05, 4.3948E+05,
     +  4.4109E+05, 4.4269E+05, 4.4431E+05, 4.4592E+05, 4.4754E+05,
     + 4.4917E+05, 4.5080E+05, 4.5243E+05, 4.5407E+05, 4.5572E+05,
     +  4.5736E+05, 4.5902E+05, 4.6067E+05, 4.6233E+05, 4.6400E+05,
     + 4.6567E+05, 4.6734E+05, 4.6902E+05, 4.7070E+05, 4.7239E+05,
     +  4.7408E+05, 4.7578E+05, 4.7748E+05, 4.7919E+05, 4.8090E+05,
     + 4.8261E+05, 4.8433E+05, 4.8605E+05, 4.8778E+05, 4.8951E+05,
     +  4.9125E+05, 4.9299E+05, 4.9474E+05, 4.9649E+05, 4.9824E+05,
     + 5.0000E+05, 5.0177E+05, 5.0354E+05, 5.0531E+05, 5.0709E+05,
     +  5.0887E+05, 5.1066E+05, 5.1245E+05, 5.1424E+05, 5.1605E+05,
     + 5.1785E+05, 5.1966E+05, 5.2148E+05, 5.2330E+05, 5.2512E+05,
     +  5.2695E+05, 5.2878E+05, 5.3062E+05, 5.3246E+05, 5.3431E+05,
     + 5.3616E+05, 5.3802E+05, 5.3988E+05, 5.4175E+05, 5.4362E+05,
     +  5.4549E+05, 5.4737E+05, 5.4926E+05, 5.5115E+05, 5.5304E+05,
     + 5.5494E+05, 5.5684E+05, 5.5875E+05, 5.6067E+05, 5.6258E+05,
     +  5.6451E+05, 5.6643E+05, 5.6837E+05, 5.7030E+05, 5.7224E+05,
     + 5.7419E+05, 5.7614E+05, 5.7810E+05, 5.8006E+05, 5.8202E+05,
     +  5.8400E+05, 5.8597E+05, 5.8795E+05, 5.8994E+05, 5.9193E+05,
     + 5.9392E+05, 5.9592E+05, 5.9792E+05, 5.9993E+05, 6.0195E+05,
     +  6.0397E+05, 6.0599E+05, 6.0802E+05, 6.1005E+05, 6.1209E+05,
     + 6.1413E+05, 6.1618E+05, 6.1823E+05, 6.2029E+05, 6.2235E+05,
     +  6.2442E+05, 6.2649E+05, 6.2857E+05, 6.3065E+05, 6.3274E+05,
     + 6.3483E+05, 6.3693E+05, 6.3903E+05, 6.4114E+05, 6.4325E+05,
     +  6.4537E+05, 6.4749E+05, 6.4962E+05, 6.5175E+05, 6.5388E+05,
     + 6.5603E+05, 6.5817E+05, 6.6032E+05, 6.6248E+05, 6.6464E+05,
     +  6.6681E+05, 6.6898E+05, 6.7116E+05, 6.7334E+05, 6.7552E+05,
     + 6.7772E+05, 6.7991E+05, 6.8211E+05, 6.8432E+05, 6.8653E+05,
     +  6.8875E+05, 6.9097E+05, 6.9320E+05, 6.9543E+05, 6.9767E+05,
     + 6.9991E+05, 7.0216E+05, 7.0441E+05, 7.0667E+05, 7.0893E+05,
     +  7.1120E+05, 7.1347E+05, 7.1575E+05, 7.1803E+05, 7.2032E+05,
     + 7.2261E+05, 7.2491E+05, 7.2721E+05, 7.2952E+05, 7.3183E+05,
     +  7.3415E+05, 7.3647E+05, 7.3880E+05, 7.4114E+05, 7.4348E+05,
     + 7.4582E+05, 7.4817E+05, 7.5052E+05, 7.5288E+05, 7.5525E+05,
     +  7.5762E+05, 7.5999E+05, 7.6237E+05, 7.6476E+05, 7.6715E+05,
     + 7.6955E+05, 7.7195E+05, 7.7435E+05, 7.7677E+05, 7.7918E+05,
     +  7.8160E+05, 7.8403E+05, 7.8646E+05, 7.8890E+05, 7.9134E+05,
     + 7.9379E+05, 7.9625E+05, 7.9870E+05, 8.0117E+05, 8.0364E+05,
     +  8.0611E+05, 8.0859E+05, 8.1107E+05, 8.1356E+05, 8.1606E+05,
     + 8.1856E+05, 8.2107E+05, 8.2358E+05, 8.2609E+05, 8.2862E+05,
     +  8.3114E+05, 8.3367E+05, 8.3621E+05, 8.3875E+05, 8.4130E+05,
     + 8.4386E+05, 8.4641E+05, 8.4898E+05, 8.5155E+05, 8.5412E+05,
     +  8.5670E+05, 8.5929E+05, 8.6188E+05, 8.6447E+05, 8.6707E+05,
     + 8.6968E+05, 8.7229E+05, 8.7491E+05, 8.7753E+05, 8.8016E+05,
     +  8.8279E+05, 8.8543E+05, 8.8808E+05, 8.9073E+05, 8.9338E+05,
     + 8.9604E+05, 8.9871E+05, 9.0138E+05, 9.0405E+05, 9.0673E+05,
     +  9.0942E+05, 9.1211E+05, 9.1481E+05, 9.1751E+05, 9.2022E+05,
     + 9.2294E+05, 9.2566E+05, 9.2838E+05, 9.3111E+05, 9.3385E+05,
     +  9.3659E+05, 9.3933E+05, 9.4209E+05, 9.4484E+05, 9.4761E+05,
     + 9.5037E+05, 9.5315E+05, 9.5593E+05, 9.5871E+05, 9.6150E+05,
     +  9.6429E+05, 9.6710E+05, 9.6990E+05, 9.7271E+05, 9.7553E+05,
     + 9.7835E+05, 9.8118E+05, 9.8401E+05, 9.8685E+05, 9.8970E+05,
     +  9.9255E+05, 9.9540E+05, 9.9826E+05, 1.0011E+06, 1.0040E+06,
     + 1.0069E+06, 1.0098E+06, 1.0126E+06, 1.0155E+06, 1.0184E+06,
     +  1.0213E+06, 1.0243E+06, 1.0272E+06, 1.0301E+06, 1.0330E+06,
     + 1.0360E+06, 1.0389E+06, 1.0418E+06, 1.0448E+06, 1.0477E+06,
     +  1.0507E+06, 1.0537E+06, 1.0566E+06, 1.0596E+06, 1.0626E+06,
     + 1.0656E+06, 1.0686E+06, 1.0716E+06, 1.0746E+06, 1.0776E+06,
     +  1.0806E+06, 1.0836E+06, 1.0866E+06, 1.0897E+06, 1.0927E+06,
     + 1.0958E+06, 1.0988E+06, 1.1019E+06, 1.1049E+06, 1.1080E+06,
     +  1.1111E+06, 1.1141E+06, 1.1172E+06, 1.1203E+06, 1.1234E+06,
     + 1.1265E+06, 1.1296E+06, 1.1327E+06, 1.1358E+06, 1.1389E+06,
     +  1.1421E+06, 1.1452E+06, 1.1483E+06, 1.1515E+06, 1.1546E+06,
     + 1.1578E+06, 1.1609E+06, 1.1641E+06, 1.1673E+06, 1.1705E+06,
     +  1.1736E+06, 1.1768E+06, 1.1800E+06, 1.1832E+06, 1.1864E+06,
     + 1.1896E+06, 1.1928E+06, 1.1961E+06, 1.1993E+06, 1.2025E+06,
     +  1.2058E+06, 1.2090E+06, 1.2123E+06, 1.2155E+06, 1.2188E+06,
     + 1.2220E+06, 1.2253E+06, 1.2286E+06, 1.2319E+06, 1.2352E+06,
     +  1.2385E+06, 1.2418E+06, 1.2451E+06, 1.2484E+06, 1.2517E+06,
     + 1.2550E+06, 1.2583E+06, 1.2617E+06, 1.2650E+06, 1.2684E+06,
     +  1.2717E+06, 1.2751E+06, 1.2784E+06, 1.2818E+06, 1.2852E+06,
     + 1.2886E+06, 1.2919E+06, 1.2953E+06, 1.2987E+06, 1.3021E+06,
     +  1.3055E+06, 1.3090E+06, 1.3124E+06, 1.3158E+06, 1.3192E+06,
     + 1.3227E+06, 1.3261E+06, 1.3296E+06, 1.3330E+06, 1.3365E+06,
     +  1.3399E+06, 1.3434E+06, 1.3469E+06, 1.3504E+06, 1.3538E+06,
     + 1.3573E+06, 1.3608E+06, 1.3643E+06, 1.3678E+06, 1.3714E+06,
     +  1.3749E+06, 1.3784E+06, 1.3819E+06, 1.3855E+06, 1.3890E+06,
     + 1.3926E+06, 1.3961E+06, 1.3997E+06, 1.4033E+06, 1.4068E+06,
     +  1.4104E+06, 1.4140E+06, 1.4176E+06, 1.4212E+06, 1.4248E+06,
     + 1.4284E+06, 1.4320E+06, 1.4356E+06, 1.4392E+06, 1.4429E+06,
     +  1.4465E+06, 1.4501E+06, 1.4538E+06, 1.4574E+06, 1.4611E+06,
     + 1.4648E+06, 1.4684E+06, 1.4721E+06, 1.4758E+06, 1.4795E+06,
     +  1.4832E+06, 1.4869E+06, 1.4906E+06, 1.4943E+06, 1.4980E+06,
     + 1.5017E+06, 1.5054E+06, 1.5092E+06, 1.5129E+06, 1.5167E+06,
     +  1.5204E+06, 1.5242E+06, 1.5279E+06, 1.5317E+06, 1.5355E+06,
     + 1.5392E+06, 1.5430E+06, 1.5468E+06, 1.5506E+06, 1.5544E+06,
     +  1.5582E+06, 1.5620E+06, 1.5658E+06, 1.5697E+06, 1.5735E+06,
     + 1.5773E+06, 1.5812E+06, 1.5850E+06, 1.5889E+06, 1.5927E+06,
     +  1.5966E+06, 1.6005E+06, 1.6043E+06, 1.6082E+06, 1.6121E+06,
     + 1.6160E+06, 1.6199E+06, 1.6238E+06, 1.6277E+06, 1.6316E+06,
     +  1.6355E+06, 1.6395E+06, 1.6434E+06, 1.6473E+06, 1.6513E+06,
     + 1.6552E+06, 1.6592E+06, 1.6631E+06, 1.6671E+06, 1.6711E+06,
     +  1.6751E+06, 1.6790E+06, 1.6830E+06, 1.6870E+06, 1.6910E+06,
     + 1.6950E+06, 1.6990E+06, 1.7031E+06, 1.7071E+06, 1.7111E+06,
     +  1.7152E+06, 1.7192E+06, 1.7232E+06, 1.7273E+06, 1.7313E+06,
     + 1.7354E+06, 1.7395E+06, 1.7436E+06, 1.7476E+06, 1.7517E+06,
     +  1.7558E+06, 1.7599E+06, 1.7640E+06, 1.7681E+06, 1.7722E+06,
     + 1.7764E+06, 1.7805E+06, 1.7846E+06, 1.7888E+06, 1.7929E+06,
     +  1.7971E+06, 1.8012E+06, 1.8054E+06, 1.8095E+06, 1.8137E+06,
     + 1.8179E+06, 1.8221E+06, 1.8263E+06, 1.8305E+06, 1.8347E+06,
     +  1.8389E+06, 1.8431E+06, 1.8473E+06, 1.8515E+06, 1.8557E+06,
     + 1.8600E+06, 1.8642E+06, 1.8685E+06, 1.8727E+06, 1.8770E+06,
     +  1.8812E+06, 1.8855E+06, 1.8898E+06, 1.8941E+06, 1.8983E+06,
     + 1.9026E+06, 1.9069E+06, 1.9112E+06, 1.9155E+06, 1.9199E+06,
     +  1.9242E+06, 1.9285E+06, 1.9328E+06, 1.9372E+06, 1.9415E+06,
     + 1.9459E+06, 1.9502E+06, 1.9546E+06, 1.9589E+06, 1.9633E+06,
     +  1.9677E+06, 1.9721E+06, 1.9765E+06, 1.9809E+06, 1.9853E+06,
     + 1.9897E+06, 1.9941E+06, 1.9985E+06, 2.0029E+06, 2.0073E+06,
     +  2.0118E+06, 2.0162E+06, 2.0207E+06, 2.0251E+06, 2.0296E+06,
     + 2.0340E+06, 2.0385E+06, 2.0430E+06, 2.0475E+06, 2.0519E+06,
     +  2.0564E+06, 2.0609E+06, 2.0654E+06, 2.0699E+06, 2.0744E+06,
     + 2.0790E+06, 2.0835E+06, 2.0880E+06, 2.0926E+06, 2.0971E+06,
     +  2.1016E+06, 2.1062E+06, 2.1108E+06, 2.1153E+06, 2.1199E+06,
     + 2.1245E+06, 2.1290E+06, 2.1336E+06, 2.1382E+06, 2.1428E+06,
     +  2.1474E+06, 2.1520E+06, 2.1566E+06, 2.1613E+06, 2.1659E+06,
     + 2.1705E+06, 2.1752E+06, 2.1798E+06, 2.1845E+06, 2.1891E+06,
     +  2.1938E+06, 2.1984E+06, 2.2031E+06, 2.2078E+06, 2.2125E+06,
     + 2.2171E+06, 2.2218E+06, 2.2265E+06, 2.2312E+06, 2.2360E+06,
     +  2.2407E+06, 2.2454E+06, 2.2501E+06, 2.2549E+06, 2.2596E+06,
     + 2.2643E+06, 2.2691E+06, 2.2738E+06, 2.2786E+06, 2.2834E+06,
     +  2.2881E+06, 2.2929E+06, 2.2977E+06, 2.3025E+06, 2.3073E+06,
     + 2.3121E+06, 2.3169E+06, 2.3217E+06, 2.3265E+06, 2.3313E+06,
     +  2.3362E+06, 2.3410E+06, 2.3458E+06, 2.3507E+06, 2.3555E+06,
     + 2.3604E+06, 2.3652E+06, 2.3701E+06, 2.3750E+06, 2.3799E+06,
     +  2.3847E+06, 2.3896E+06, 2.3945E+06, 2.3994E+06, 2.4043E+06,
     + 2.4092E+06, 2.4142E+06, 2.4191E+06, 2.4240E+06, 2.4289E+06,
     +  2.4339E+06, 2.4388E+06, 2.4438E+06, 2.4487E+06, 2.4537E+06,
     + 2.4587E+06, 2.4636E+06, 2.4686E+06, 2.4736E+06, 2.4786E+06,
     +  2.4836E+06, 2.4886E+06, 2.4936E+06, 2.4986E+06, 2.5036E+06,
     + 2.5086E+06, 2.5137E+06, 2.5187E+06, 2.5237E+06, 2.5288E+06,
     +  2.5338E+06, 2.5389E+06, 2.5439E+06, 2.5490E+06, 2.5541E+06,
     + 2.5592E+06, 2.5642E+06, 2.5693E+06, 2.5744E+06, 2.5795E+06,
     +  2.5846E+06, 2.5897E+06, 2.5948E+06, 2.6000E+06, 2.6051E+06,
     + 2.6102E+06, 2.6154E+06, 2.6205E+06, 2.6256E+06, 2.6308E+06,
     +  2.6360E+06, 2.6411E+06, 2.6463E+06, 2.6515E+06, 2.6566E+06,
     + 2.6618E+06, 2.6670E+06, 2.6722E+06, 2.6774E+06, 2.6826E+06,
     +  2.6878E+06, 2.6931E+06, 2.6983E+06, 2.7035E+06, 2.7088E+06,
     + 2.7140E+06, 2.7192E+06, 2.7245E+06, 2.7297E+06, 2.7350E+06,
     +  2.7403E+06, 2.7455E+06, 2.7508E+06, 2.7561E+06, 2.7614E+06,
     + 2.7667E+06, 2.7720E+06, 2.7773E+06, 2.7826E+06, 2.7879E+06,
     +  2.7932E+06, 2.7986E+06, 2.8039E+06, 2.8092E+06, 2.8146E+06,
     + 2.8199E+06, 2.8253E+06, 2.8306E+06, 2.8360E+06, 2.8414E+06,
     +  2.8468E+06, 2.8521E+06, 2.8575E+06, 2.8629E+06, 2.8683E+06,
     + 2.8737E+06, 2.8791E+06, 2.8845E+06, 2.8900E+06, 2.8954E+06,
     +  2.9008E+06, 2.9062E+06, 2.9117E+06, 2.9171E+06, 2.9226E+06,
     + 2.9280E+06, 2.9335E+06, 2.9390E+06, 2.9444E+06, 2.9499E+06,
     +  2.9554E+06, 2.9609E+06, 2.9664E+06, 2.9719E+06, 2.9774E+06,
     + 2.9829E+06, 2.9884E+06, 2.9939E+06, 2.9994E+06, 3.0050E+06,
     +  3.0105E+06, 3.0160E+06, 3.0216E+06, 3.0271E+06, 3.0327E+06,
     + 3.0382E+06, 3.0438E+06, 3.0494E+06, 3.0550E+06, 3.0605E+06,
     +  3.0661E+06, 3.0717E+06, 3.0773E+06, 3.0829E+06, 3.0885E+06,
     + 3.0941E+06, 3.0998E+06, 3.1054E+06, 3.1110E+06, 3.1167E+06,
     +  3.1223E+06, 3.1279E+06, 3.1336E+06, 3.1392E+06, 3.1449E+06,
     + 3.1506E+06, 3.1562E+06, 3.1619E+06, 3.1676E+06, 3.1733E+06,
     +  3.1790E+06, 3.1847E+06, 3.1904E+06, 3.1961E+06, 3.2018E+06,
     + 3.2075E+06, 3.2132E+06, 3.2190E+06, 3.2247E+06, 3.2305E+06,
     +  3.2362E+06, 3.2419E+06, 3.2477E+06, 3.2535E+06, 3.2592E+06,
     + 3.2650E+06, 3.2708E+06, 3.2765E+06, 3.2823E+06, 3.2881E+06,
     +  3.2939E+06, 3.2997E+06, 3.3055E+06, 3.3113E+06, 3.3172E+06,
     + 3.3230E+06, 3.3288E+06, 3.3346E+06, 3.3405E+06, 3.3463E+06,
     +  3.3522E+06, 3.3580E+06, 3.3639E+06, 3.3697E+06, 3.3756E+06,
     + 3.3815E+06, 3.3873E+06, 3.3932E+06, 3.3991E+06, 3.4050E+06,
     +  3.4109E+06, 3.4168E+06, 3.4227E+06, 3.4286E+06, 3.4346E+06,
     + 3.4405E+06, 3.4464E+06, 3.4523E+06, 3.4583E+06, 3.4642E+06,
     +  3.4702E+06, 3.4761E+06, 3.4821E+06, 3.4880E+06, 3.4940E+06,
     + 3.5000E+06, 3.5060E+06, 3.5120E+06, 3.5179E+06, 3.5239E+06,
     +  3.5299E+06, 3.5359E+06, 3.5419E+06, 3.5480E+06, 3.5540E+06,
     + 3.5600E+06, 3.5660E+06, 3.5721E+06, 3.5781E+06, 3.5842E+06,
     +  3.5902E+06, 3.5963E+06, 3.6023E+06, 3.6084E+06, 3.6144E+06,
     + 3.6205E+06, 3.6266E+06, 3.6327E+06, 3.6388E+06, 3.6449E+06,
     +  3.6510E+06, 3.6571E+06, 3.6632E+06, 3.6693E+06, 3.6754E+06,
     + 3.6815E+06, 3.6877E+06, 3.6938E+06, 3.6999E+06, 3.7061E+06,
     +  3.7122E+06, 3.7184E+06, 3.7245E+06, 3.7307E+06, 3.7369E+06,
     + 3.7430E+06, 3.7492E+06, 3.7554E+06, 3.7616E+06, 3.7678E+06,
     +  3.7740E+06, 3.7802E+06, 3.7864E+06, 3.7926E+06, 3.7988E+06,
     + 3.8050E+06, 3.8113E+06, 3.8175E+06, 3.8237E+06, 3.8300E+06,
     +  3.8362E+06, 3.8425E+06, 3.8487E+06, 3.8550E+06, 3.8613E+06,
     + 3.8675E+06, 3.8738E+06, 3.8801E+06, 3.8864E+06, 3.8927E+06,
     +  3.8990E+06, 3.9053E+06, 3.9116E+06, 3.9179E+06, 3.9242E+06,
     + 3.9305E+06, 3.9368E+06, 3.9432E+06, 3.9495E+06, 3.9558E+06,
     +  3.9622E+06, 3.9685E+06, 3.9749E+06, 3.9812E+06, 3.9876E+06,
     + 3.9940E+06, 4.0003E+06, 4.0067E+06, 4.0131E+06, 4.0195E+06,
     +  4.0259E+06, 4.0323E+06, 4.0387E+06, 4.0451E+06, 4.0515E+06,
     + 4.0579E+06, 4.0643E+06, 4.0707E+06, 4.0772E+06, 4.0836E+06,
     +  4.0900E+06, 4.0965E+06, 4.1029E+06, 4.1094E+06, 4.1158E+06,
     + 4.1223E+06, 4.1288E+06, 4.1352E+06, 4.1417E+06, 4.1482E+06,
     +  4.1547E+06, 4.1612E+06, 4.1676E+06, 4.1741E+06, 4.1807E+06,
     + 4.1872E+06, 4.1937E+06, 4.2002E+06, 4.2067E+06, 4.2132E+06,
     +  4.2198E+06, 4.2263E+06, 4.2328E+06, 4.2394E+06, 4.2459E+06,
     + 4.2525E+06, 4.2591E+06, 4.2656E+06, 4.2722E+06, 4.2788E+06,
     +  4.2853E+06, 4.2919E+06, 4.2985E+06, 4.3051E+06, 4.3117E+06,
     + 4.3183E+06, 4.3249E+06, 4.3315E+06, 4.3381E+06, 4.3448E+06,
     +  4.3514E+06, 4.3580E+06, 4.3646E+06, 4.3713E+06, 4.3779E+06,
     + 4.3846E+06, 4.3912E+06, 4.3979E+06, 4.4045E+06, 4.4112E+06,
     +  4.4179E+06, 4.4245E+06, 4.4312E+06, 4.4379E+06, 4.4446E+06,
     + 4.4513E+06, 4.4580E+06, 4.4647E+06, 4.4714E+06, 4.4781E+06,
     +  4.4848E+06, 4.4915E+06, 4.4983E+06, 4.5050E+06, 4.5117E+06,
     + 4.5185E+06, 4.5252E+06, 4.5319E+06, 4.5387E+06, 4.5455E+06,
     +  4.5522E+06, 4.5590E+06, 4.5657E+06, 4.5725E+06, 4.5793E+06,
     + 4.5861E+06, 4.5929E+06, 4.5997E+06, 4.6065E+06, 4.6133E+06,
     +  4.6201E+06, 4.6269E+06, 4.6337E+06, 4.6405E+06, 4.6473E+06,
     + 4.6541E+06, 4.6610E+06, 4.6678E+06, 4.6747E+06, 4.6815E+06,
     +  4.6883E+06, 4.6952E+06, 4.7021E+06, 4.7089E+06, 4.7158E+06,
     + 4.7227E+06, 4.7295E+06, 4.7364E+06, 4.7433E+06, 4.7502E+06,
     +  4.7571E+06, 4.7640E+06, 4.7709E+06, 4.7778E+06, 4.7847E+06,
     + 4.7916E+06, 4.7985E+06, 4.8054E+06, 4.8124E+06, 4.8193E+06,
     +  4.8262E+06, 4.8332E+06, 4.8401E+06, 4.8471E+06, 4.8540E+06,
     + 4.8610E+06, 4.8679E+06, 4.8749E+06, 4.8819E+06, 4.8889E+06,
     +  4.8958E+06, 4.9028E+06, 4.9098E+06, 4.9168E+06, 4.9238E+06,
     + 4.9308E+06, 4.9378E+06, 4.9448E+06, 4.9518E+06, 4.9588E+06,
     +  4.9659E+06, 4.9729E+06, 4.9799E+06, 4.9870E+06, 4.9940E+06,
     + 5.0010E+06, 5.0081E+06, 5.0151E+06, 5.0222E+06, 5.0292E+06,
     +  5.0363E+06, 5.0434E+06, 5.0505E+06, 5.0575E+06, 5.0646E+06,
     + 5.0717E+06, 5.0788E+06, 5.0859E+06, 5.0930E+06, 5.1001E+06,
     +  5.1072E+06, 5.1143E+06, 5.1214E+06, 5.1285E+06, 5.1356E+06,
     + 5.1428E+06, 5.1499E+06, 5.1570E+06, 5.1642E+06, 5.1713E+06,
     +  5.1785E+06, 5.1856E+06, 5.1928E+06, 5.1999E+06, 5.2071E+06,
     + 5.2143E+06, 5.2214E+06, 5.2286E+06, 5.2358E+06, 5.2430E+06,
     +  5.2502E+06, 5.2574E+06, 5.2646E+06, 5.2718E+06, 5.2790E+06,
     + 5.2862E+06, 5.2934E+06, 5.3006E+06, 5.3078E+06, 5.3151E+06,
     +  5.3223E+06, 5.3295E+06, 5.3368E+06, 5.3440E+06, 5.3512E+06,
     + 5.3585E+06, 5.3658E+06, 5.3730E+06, 5.3803E+06, 5.3875E+06,
     +  5.3948E+06, 5.4021E+06, 5.4094E+06, 5.4166E+06, 5.4239E+06,
     + 5.4312E+06, 5.4385E+06, 5.4458E+06, 5.4531E+06, 5.4604E+06,
     +  5.4677E+06, 5.4750E+06, 5.4824E+06, 5.4897E+06, 5.4970E+06,
     + 5.5043E+06, 5.5117E+06, 5.5190E+06, 5.5264E+06, 5.5337E+06,
     +  5.5411E+06, 5.5484E+06, 5.5558E+06, 5.5631E+06, 5.5705E+06,
     + 5.5779E+06, 5.5852E+06, 5.5926E+06, 5.6000E+06, 5.6074E+06,
     +  5.6148E+06, 5.6222E+06, 5.6296E+06, 5.6370E+06, 5.6444E+06,
     + 5.6518E+06, 5.6592E+06, 5.6666E+06, 5.6740E+06, 5.6814E+06,
     +  5.6889E+06, 5.6963E+06, 5.7037E+06, 5.7112E+06, 5.7186E+06,
     + 5.7261E+06, 5.7335E+06, 5.7410E+06, 5.7484E+06, 5.7559E+06,
     +  5.7634E+06, 5.7708E+06, 5.7783E+06, 5.7858E+06, 5.7933E+06,
     + 5.8008E+06, 5.8083E+06, 5.8157E+06, 5.8232E+06, 5.8307E+06,
     +  5.8382E+06, 5.8458E+06, 5.8533E+06, 5.8608E+06, 5.8683E+06,
     + 5.8758E+06, 5.8834E+06, 5.8909E+06, 5.8984E+06, 5.9060E+06,
     +  5.9135E+06, 5.9211E+06, 5.9286E+06, 5.9362E+06, 5.9437E+06,
     + 5.9513E+06, 5.9588E+06, 5.9664E+06, 5.9740E+06, 5.9816E+06,
     +  5.9891E+06, 5.9967E+06, 6.0043E+06, 6.0119E+06, 6.0195E+06,
     + 6.0271E+06, 6.0347E+06, 6.0423E+06, 6.0499E+06, 6.0575E+06,
     +  6.0651E+06, 6.0728E+06, 6.0804E+06, 6.0880E+06, 6.0957E+06,
     + 6.1033E+06, 6.1109E+06, 6.1186E+06, 6.1262E+06, 6.1339E+06,
     +  6.1415E+06, 6.1492E+06, 6.1568E+06, 6.1645E+06, 6.1722E+06,
     + 6.1798E+06, 6.1875E+06, 6.1952E+06, 6.2029E+06, 6.2106E+06,
     +  6.2183E+06, 6.2260E+06, 6.2336E+06, 6.2414E+06, 6.2491E+06,
     + 6.2568E+06, 6.2645E+06, 6.2722E+06, 6.2799E+06, 6.2876E+06,
     +  6.2954E+06, 6.3031E+06, 6.3108E+06, 6.3186E+06, 6.3263E+06,
     + 6.3340E+06, 6.3418E+06, 6.3495E+06, 6.3573E+06, 6.3651E+06,
     +  6.3728E+06, 6.3806E+06, 6.3883E+06, 6.3961E+06, 6.4039E+06,
     + 6.4117E+06, 6.4195E+06, 6.4272E+06, 6.4350E+06, 6.4428E+06,
     +  6.4506E+06, 6.4584E+06, 6.4662E+06, 6.4740E+06, 6.4818E+06,
     + 6.4897E+06, 6.4975E+06, 6.5053E+06, 6.5131E+06, 6.5210E+06,
     +  6.5288E+06, 6.5366E+06, 6.5445E+06, 6.5523E+06, 6.5601E+06,
     + 6.5680E+06, 6.5758E+06, 6.5837E+06, 6.5916E+06, 6.5994E+06,
     +  6.6073E+06, 6.6152E+06, 6.6230E+06, 6.6309E+06, 6.6388E+06,
     + 6.6467E+06, 6.6546E+06, 6.6624E+06, 6.6703E+06, 6.6782E+06,
     +  6.6861E+06, 6.6940E+06, 6.7019E+06, 6.7099E+06, 6.7178E+06,
     + 6.7257E+06, 6.7336E+06, 6.7415E+06, 6.7495E+06, 6.7574E+06,
     +  6.7653E+06, 6.7733E+06, 6.7812E+06, 6.7891E+06, 6.7971E+06,
     + 6.8050E+06, 6.8130E+06, 6.8209E+06, 6.8289E+06, 6.8369E+06,
     +  6.8448E+06, 6.8528E+06, 6.8608E+06, 6.8688E+06, 6.8767E+06,
     + 6.8847E+06, 6.8927E+06, 6.9007E+06, 6.9087E+06, 6.9167E+06,
     +  6.9247E+06, 6.9327E+06, 6.9407E+06, 6.9487E+06, 6.9567E+06,
     + 6.9647E+06, 6.9728E+06, 6.9808E+06, 6.9888E+06, 6.9968E+06,
     +  7.0049E+06, 7.0129E+06, 7.0209E+06, 7.0290E+06, 7.0370E+06,
     + 7.0451E+06, 7.0531E+06, 7.0612E+06, 7.0692E+06, 7.0773E+06,
     +  7.0854E+06, 7.0934E+06, 7.1015E+06, 7.1096E+06, 7.1177E+06,
     + 7.1257E+06, 7.1338E+06, 7.1419E+06, 7.1500E+06, 7.1581E+06,
     +  7.1662E+06, 7.1743E+06, 7.1824E+06, 7.1905E+06, 7.1986E+06,
     + 7.2067E+06, 7.2148E+06, 7.2229E+06, 7.2311E+06, 7.2392E+06,
     +  7.2473E+06, 7.2554E+06, 7.2636E+06, 7.2717E+06, 7.2799E+06,
     + 7.2880E+06, 7.2961E+06, 7.3043E+06, 7.3124E+06, 7.3206E+06,
     +  7.3288E+06, 7.3369E+06, 7.3451E+06, 7.3533E+06, 7.3614E+06,
     + 7.3696E+06, 7.3778E+06, 7.3860E+06, 7.3941E+06, 7.4023E+06,
     +  7.4105E+06, 7.4187E+06, 7.4269E+06, 7.4351E+06, 7.4433E+06,
     + 7.4515E+06, 7.4597E+06, 7.4679E+06, 7.4761E+06, 7.4844E+06,
     +  7.4926E+06, 7.5008E+06, 7.5090E+06, 7.5173E+06, 7.5255E+06,
     + 7.5337E+06, 7.5420E+06, 7.5502E+06, 7.5584E+06, 7.5667E+06,
     +  7.5749E+06, 7.5832E+06, 7.5914E+06, 7.5997E+06, 7.6080E+06,
     + 7.6162E+06, 7.6245E+06, 7.6328E+06, 7.6410E+06, 7.6493E+06,
     +  7.6576E+06, 7.6659E+06, 7.6742E+06, 7.6825E+06, 7.6907E+06,
     + 7.6990E+06, 7.7073E+06, 7.7156E+06, 7.7239E+06, 7.7322E+06,
     +  7.7406E+06, 7.7489E+06, 7.7572E+06, 7.7655E+06, 7.7738E+06,
     + 7.7821E+06, 7.7905E+06, 7.7988E+06, 7.8071E+06, 7.8155E+06,
     +  7.8238E+06, 7.8321E+06, 7.8405E+06, 7.8488E+06, 7.8572E+06,
     + 7.8655E+06, 7.8739E+06, 7.8822E+06, 7.8906E+06, 7.8990E+06,
     +  7.9073E+06, 7.9157E+06, 7.9241E+06, 7.9324E+06, 7.9408E+06,
     + 7.9492E+06, 7.9576E+06, 7.9660E+06, 7.9744E+06, 7.9827E+06,
     +  7.9911E+06, 7.9995E+06, 8.0079E+06, 8.0163E+06, 8.0247E+06,
     + 8.0331E+06, 8.0416E+06, 8.0500E+06, 8.0584E+06, 8.0668E+06,
     +  8.0752E+06, 8.0837E+06, 8.0921E+06, 8.1005E+06, 8.1089E+06/
      data (QofT( 1,J),J=2409,4510)/ 8.1174E+06, 8.1258E+06,
     +  8.1343E+06, 8.1427E+06, 8.1511E+06, 8.1596E+06, 8.1680E+06,
     +  8.1765E+06,
     + 8.1850E+06, 8.1934E+06, 8.2019E+06, 8.2103E+06, 8.2188E+06,
     +  8.2273E+06, 8.2358E+06, 8.2442E+06, 8.2527E+06, 8.2612E+06,
     + 8.2697E+06, 8.2782E+06, 8.2867E+06, 8.2952E+06, 8.3036E+06,
     +  8.3121E+06, 8.3206E+06, 8.3291E+06, 8.3377E+06, 8.3462E+06,
     + 8.3547E+06, 8.3632E+06, 8.3717E+06, 8.3802E+06, 8.3887E+06,
     +  8.3973E+06, 8.4058E+06, 8.4143E+06, 8.4229E+06, 8.4314E+06,
     + 8.4399E+06, 8.4485E+06, 8.4570E+06, 8.4656E+06, 8.4741E+06,
     +  8.4826E+06, 8.4912E+06, 8.4998E+06, 8.5083E+06, 8.5169E+06,
     + 8.5254E+06, 8.5340E+06, 8.5426E+06, 8.5511E+06, 8.5597E+06,
     +  8.5683E+06, 8.5769E+06, 8.5855E+06, 8.5940E+06, 8.6026E+06,
     + 8.6112E+06, 8.6198E+06, 8.6284E+06, 8.6370E+06, 8.6456E+06,
     +  8.6542E+06, 8.6628E+06, 8.6714E+06, 8.6800E+06, 8.6886E+06,
     + 8.6972E+06, 8.7059E+06, 8.7145E+06, 8.7231E+06, 8.7317E+06,
     +  8.7403E+06, 8.7490E+06, 8.7576E+06, 8.7662E+06, 8.7749E+06,
     + 8.7835E+06, 8.7922E+06, 8.8008E+06, 8.8094E+06, 8.8181E+06,
     +  8.8267E+06, 8.8354E+06, 8.8441E+06, 8.8527E+06, 8.8614E+06,
     + 8.8700E+06, 8.8787E+06, 8.8874E+06, 8.8960E+06, 8.9047E+06,
     +  8.9134E+06, 8.9221E+06, 8.9307E+06, 8.9394E+06, 8.9481E+06,
     + 8.9568E+06, 8.9655E+06, 8.9742E+06, 8.9829E+06, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       456
      data (QofT( 2,J),J=1,2408)/ 1.2229E+01, 2.2037E+01, 3.1947E+01,
     +  4.1881E+01, 5.1823E+01, 6.1770E+01, 7.1720E+01, 8.1671E+01,
     + 9.1624E+01, 1.0158E+02, 1.1153E+02, 1.2149E+02, 1.3144E+02,
     +  1.4140E+02, 1.5135E+02, 1.6131E+02, 1.7127E+02, 1.8122E+02,
     + 1.9118E+02, 2.0114E+02, 2.2106E+02, 2.4098E+02, 2.6089E+02,
     +  2.8081E+02, 3.0073E+02, 3.2066E+02, 3.4058E+02, 3.6050E+02,
     + 3.8043E+02, 4.0035E+02, 4.2028E+02, 4.4021E+02, 4.6014E+02,
     +  4.8007E+02, 5.0000E+02, 5.1993E+02, 5.3986E+02, 5.5979E+02,
     + 5.7973E+02, 5.9966E+02, 6.1960E+02, 6.3953E+02, 6.5948E+02,
     +  6.7942E+02, 6.9936E+02, 7.1930E+02, 7.3925E+02, 7.5920E+02,
     + 7.7916E+02, 7.9911E+02, 8.1907E+02, 8.3904E+02, 8.5901E+02,
     +  8.7899E+02, 8.9898E+02, 9.1897E+02, 9.3898E+02, 9.5899E+02,
     + 9.7902E+02, 9.9906E+02, 1.0191E+03, 1.0392E+03, 1.0593E+03,
     +  1.0794E+03, 1.0995E+03, 1.1197E+03, 1.1399E+03, 1.1601E+03,
     + 1.1803E+03, 1.2006E+03, 1.2209E+03, 1.2412E+03, 1.2616E+03,
     +  1.2820E+03, 1.3024E+03, 1.3229E+03, 1.3435E+03, 1.3641E+03,
     + 1.3847E+03, 1.4054E+03, 1.4261E+03, 1.4470E+03, 1.4678E+03,
     +  1.4888E+03, 1.5098E+03, 1.5308E+03, 1.5520E+03, 1.5732E+03,
     + 1.5945E+03, 1.6158E+03, 1.6373E+03, 1.6588E+03, 1.6804E+03,
     +  1.7021E+03, 1.7239E+03, 1.7458E+03, 1.7677E+03, 1.7898E+03,
     + 1.8120E+03, 1.8342E+03, 1.8566E+03, 1.8790E+03, 1.9016E+03,
     +  1.9243E+03, 1.9471E+03, 1.9700E+03, 1.9930E+03, 2.0162E+03,
     + 2.0394E+03, 2.0628E+03, 2.0863E+03, 2.1099E+03, 2.1337E+03,
     +  2.1575E+03, 2.1815E+03, 2.2057E+03, 2.2300E+03, 2.2544E+03,
     + 2.2789E+03, 2.3036E+03, 2.3285E+03, 2.3534E+03, 2.3785E+03,
     +  2.4038E+03, 2.4292E+03, 2.4548E+03, 2.4805E+03, 2.5064E+03,
     + 2.5324E+03, 2.5586E+03, 2.5850E+03, 2.6115E+03, 2.6381E+03,
     +  2.6650E+03, 2.6919E+03, 2.7191E+03, 2.7465E+03, 2.7740E+03,
     + 2.8016E+03, 2.8295E+03, 2.8575E+03, 2.8857E+03, 2.9140E+03,
     +  2.9426E+03, 2.9713E+03, 3.0003E+03, 3.0294E+03, 3.0586E+03,
     + 3.0881E+03, 3.1178E+03, 3.1476E+03, 3.1777E+03, 3.2079E+03,
     +  3.2383E+03, 3.2689E+03, 3.2998E+03, 3.3308E+03, 3.3620E+03,
     + 3.3934E+03, 3.4251E+03, 3.4569E+03, 3.4889E+03, 3.5211E+03,
     +  3.5536E+03, 3.5862E+03, 3.6191E+03, 3.6522E+03, 3.6855E+03,
     + 3.7190E+03, 3.7528E+03, 3.7867E+03, 3.8209E+03, 3.8553E+03,
     +  3.8899E+03, 3.9247E+03, 3.9598E+03, 3.9951E+03, 4.0306E+03,
     + 4.0663E+03, 4.1024E+03, 4.1386E+03, 4.1750E+03, 4.2117E+03,
     +  4.2486E+03, 4.2858E+03, 4.3232E+03, 4.3608E+03, 4.3987E+03,
     + 4.4368E+03, 4.4752E+03, 4.5138E+03, 4.5527E+03, 4.5919E+03,
     +  4.6312E+03, 4.6708E+03, 4.7107E+03, 4.7509E+03, 4.7913E+03,
     + 4.8319E+03, 4.8728E+03, 4.9140E+03, 4.9554E+03, 4.9972E+03,
     +  5.0391E+03, 5.0814E+03, 5.1239E+03, 5.1667E+03, 5.2097E+03,
     + 5.2531E+03, 5.2966E+03, 5.3405E+03, 5.3847E+03, 5.4292E+03,
     +  5.4738E+03, 5.5189E+03, 5.5641E+03, 5.6097E+03, 5.6556E+03,
     + 5.7017E+03, 5.7482E+03, 5.7949E+03, 5.8419E+03, 5.8893E+03,
     +  5.9369E+03, 5.9848E+03, 6.0330E+03, 6.0815E+03, 6.1304E+03,
     + 6.1795E+03, 6.2289E+03, 6.2786E+03, 6.3287E+03, 6.3791E+03,
     +  6.4297E+03, 6.4807E+03, 6.5320E+03, 6.5836E+03, 6.6356E+03,
     + 6.6879E+03, 6.7404E+03, 6.7933E+03, 6.8466E+03, 6.9001E+03,
     +  6.9540E+03, 7.0082E+03, 7.0628E+03, 7.1177E+03, 7.1729E+03,
     + 7.2284E+03, 7.2843E+03, 7.3406E+03, 7.3972E+03, 7.4541E+03,
     +  7.5113E+03, 7.5690E+03, 7.6269E+03, 7.6852E+03, 7.7439E+03,
     + 7.8029E+03, 7.8623E+03, 7.9220E+03, 7.9821E+03, 8.0426E+03,
     +  8.1034E+03, 8.1646E+03, 8.2262E+03, 8.2881E+03, 8.3503E+03,
     + 8.4130E+03, 8.4761E+03, 8.5395E+03, 8.6032E+03, 8.6674E+03,
     +  8.7320E+03, 8.7969E+03, 8.8622E+03, 8.9279E+03, 8.9940E+03,
     + 9.0605E+03, 9.1274E+03, 9.1946E+03, 9.2622E+03, 9.3303E+03,
     +  9.3988E+03, 9.4676E+03, 9.5369E+03, 9.6065E+03, 9.6766E+03,
     + 9.7471E+03, 9.8180E+03, 9.8893E+03, 9.9610E+03, 1.0033E+04,
     +  1.0106E+04, 1.0179E+04, 1.0252E+04, 1.0326E+04, 1.0400E+04,
     + 1.0475E+04, 1.0550E+04, 1.0625E+04, 1.0701E+04, 1.0778E+04,
     +  1.0855E+04, 1.0932E+04, 1.1010E+04, 1.1088E+04, 1.1166E+04,
     + 1.1245E+04, 1.1325E+04, 1.1405E+04, 1.1485E+04, 1.1566E+04,
     +  1.1648E+04, 1.1729E+04, 1.1812E+04, 1.1894E+04, 1.1978E+04,
     + 1.2061E+04, 1.2145E+04, 1.2230E+04, 1.2315E+04, 1.2401E+04,
     +  1.2487E+04, 1.2573E+04, 1.2660E+04, 1.2747E+04, 1.2835E+04,
     + 1.2924E+04, 1.3013E+04, 1.3102E+04, 1.3192E+04, 1.3283E+04,
     +  1.3374E+04, 1.3465E+04, 1.3557E+04, 1.3649E+04, 1.3742E+04,
     + 1.3836E+04, 1.3930E+04, 1.4024E+04, 1.4119E+04, 1.4215E+04,
     +  1.4311E+04, 1.4407E+04, 1.4505E+04, 1.4602E+04, 1.4700E+04,
     + 1.4799E+04, 1.4898E+04, 1.4998E+04, 1.5098E+04, 1.5199E+04,
     +  1.5301E+04, 1.5403E+04, 1.5505E+04, 1.5608E+04, 1.5712E+04,
     + 1.5816E+04, 1.5921E+04, 1.6026E+04, 1.6132E+04, 1.6238E+04,
     +  1.6345E+04, 1.6453E+04, 1.6561E+04, 1.6670E+04, 1.6779E+04,
     + 1.6889E+04, 1.6999E+04, 1.7110E+04, 1.7222E+04, 1.7334E+04,
     +  1.7447E+04, 1.7560E+04, 1.7674E+04, 1.7789E+04, 1.7904E+04,
     + 1.8020E+04, 1.8136E+04, 1.8253E+04, 1.8371E+04, 1.8489E+04,
     +  1.8608E+04, 1.8728E+04, 1.8848E+04, 1.8969E+04, 1.9090E+04,
     + 1.9212E+04, 1.9335E+04, 1.9458E+04, 1.9582E+04, 1.9706E+04,
     +  1.9832E+04, 1.9957E+04, 2.0084E+04, 2.0211E+04, 2.0339E+04,
     + 2.0467E+04, 2.0596E+04, 2.0726E+04, 2.0857E+04, 2.0988E+04,
     +  2.1120E+04, 2.1252E+04, 2.1385E+04, 2.1519E+04, 2.1653E+04,
     + 2.1789E+04, 2.1924E+04, 2.2061E+04, 2.2198E+04, 2.2336E+04,
     +  2.2475E+04, 2.2614E+04, 2.2754E+04, 2.2895E+04, 2.3036E+04,
     + 2.3178E+04, 2.3321E+04, 2.3465E+04, 2.3609E+04, 2.3754E+04,
     +  2.3900E+04, 2.4046E+04, 2.4194E+04, 2.4342E+04, 2.4490E+04,
     + 2.4640E+04, 2.4790E+04, 2.4941E+04, 2.5092E+04, 2.5245E+04,
     +  2.5398E+04, 2.5552E+04, 2.5706E+04, 2.5862E+04, 2.6018E+04,
     + 2.6175E+04, 2.6333E+04, 2.6491E+04, 2.6651E+04, 2.6811E+04,
     +  2.6971E+04, 2.7133E+04, 2.7296E+04, 2.7459E+04, 2.7623E+04,
     + 2.7788E+04, 2.7953E+04, 2.8120E+04, 2.8287E+04, 2.8455E+04,
     +  2.8624E+04, 2.8793E+04, 2.8964E+04, 2.9135E+04, 2.9307E+04,
     + 2.9480E+04, 2.9654E+04, 2.9828E+04, 3.0004E+04, 3.0180E+04,
     +  3.0357E+04, 3.0535E+04, 3.0714E+04, 3.0894E+04, 3.1074E+04,
     + 3.1256E+04, 3.1438E+04, 3.1621E+04, 3.1805E+04, 3.1990E+04,
     +  3.2176E+04, 3.2362E+04, 3.2550E+04, 3.2738E+04, 3.2927E+04,
     + 3.3118E+04, 3.3309E+04, 3.3501E+04, 3.3693E+04, 3.3887E+04,
     +  3.4082E+04, 3.4278E+04, 3.4474E+04, 3.4671E+04, 3.4870E+04,
     + 3.5069E+04, 3.5269E+04, 3.5470E+04, 3.5672E+04, 3.5875E+04,
     +  3.6079E+04, 3.6284E+04, 3.6490E+04, 3.6697E+04, 3.6904E+04,
     + 3.7113E+04, 3.7323E+04, 3.7533E+04, 3.7745E+04, 3.7957E+04,
     +  3.8171E+04, 3.8385E+04, 3.8601E+04, 3.8817E+04, 3.9035E+04,
     + 3.9253E+04, 3.9473E+04, 3.9693E+04, 3.9915E+04, 4.0137E+04,
     +  4.0361E+04, 4.0585E+04, 4.0811E+04, 4.1037E+04, 4.1265E+04,
     + 4.1493E+04, 4.1723E+04, 4.1953E+04, 4.2185E+04, 4.2418E+04,
     +  4.2651E+04, 4.2886E+04, 4.3122E+04, 4.3359E+04, 4.3597E+04,
     + 4.3836E+04, 4.4076E+04, 4.4317E+04, 4.4560E+04, 4.4803E+04,
     +  4.5047E+04, 4.5293E+04, 4.5539E+04, 4.5787E+04, 4.6036E+04,
     + 4.6286E+04, 4.6537E+04, 4.6789E+04, 4.7042E+04, 4.7296E+04,
     +  4.7552E+04, 4.7808E+04, 4.8066E+04, 4.8325E+04, 4.8585E+04,
     + 4.8846E+04, 4.9108E+04, 4.9372E+04, 4.9636E+04, 4.9902E+04,
     +  5.0169E+04, 5.0437E+04, 5.0706E+04, 5.0976E+04, 5.1248E+04,
     + 5.1520E+04, 5.1794E+04, 5.2069E+04, 5.2346E+04, 5.2623E+04,
     +  5.2902E+04, 5.3182E+04, 5.3463E+04, 5.3745E+04, 5.4028E+04,
     + 5.4313E+04, 5.4599E+04, 5.4886E+04, 5.5175E+04, 5.5464E+04,
     +  5.5755E+04, 5.6047E+04, 5.6340E+04, 5.6635E+04, 5.6931E+04,
     + 5.7228E+04, 5.7526E+04, 5.7826E+04, 5.8127E+04, 5.8429E+04,
     +  5.8732E+04, 5.9037E+04, 5.9343E+04, 5.9650E+04, 5.9959E+04,
     + 6.0269E+04, 6.0580E+04, 6.0893E+04, 6.1206E+04, 6.1522E+04,
     +  6.1838E+04, 6.2156E+04, 6.2475E+04, 6.2795E+04, 6.3117E+04,
     + 6.3440E+04, 6.3765E+04, 6.4091E+04, 6.4418E+04, 6.4746E+04,
     +  6.5076E+04, 6.5407E+04, 6.5740E+04, 6.6074E+04, 6.6410E+04,
     + 6.6746E+04, 6.7084E+04, 6.7424E+04, 6.7765E+04, 6.8107E+04,
     +  6.8451E+04, 6.8796E+04, 6.9143E+04, 6.9491E+04, 6.9840E+04,
     + 7.0191E+04, 7.0544E+04, 7.0897E+04, 7.1253E+04, 7.1609E+04,
     +  7.1967E+04, 7.2327E+04, 7.2688E+04, 7.3050E+04, 7.3414E+04,
     + 7.3780E+04, 7.4147E+04, 7.4515E+04, 7.4885E+04, 7.5256E+04,
     +  7.5629E+04, 7.6004E+04, 7.6379E+04, 7.6757E+04, 7.7136E+04,
     + 7.7516E+04, 7.7898E+04, 7.8282E+04, 7.8667E+04, 7.9053E+04,
     +  7.9441E+04, 7.9831E+04, 8.0222E+04, 8.0615E+04, 8.1009E+04,
     + 8.1405E+04, 8.1803E+04, 8.2202E+04, 8.2602E+04, 8.3005E+04,
     +  8.3409E+04, 8.3814E+04, 8.4221E+04, 8.4630E+04, 8.5040E+04,
     + 8.5452E+04, 8.5865E+04, 8.6280E+04, 8.6697E+04, 8.7115E+04,
     +  8.7535E+04, 8.7957E+04, 8.8380E+04, 8.8805E+04, 8.9232E+04,
     + 8.9660E+04, 9.0090E+04, 9.0522E+04, 9.0955E+04, 9.1390E+04,
     +  9.1827E+04, 9.2265E+04, 9.2706E+04, 9.3147E+04, 9.3591E+04,
     + 9.4036E+04, 9.4483E+04, 9.4932E+04, 9.5382E+04, 9.5834E+04,
     +  9.6288E+04, 9.6744E+04, 9.7201E+04, 9.7660E+04, 9.8121E+04,
     + 9.8584E+04, 9.9048E+04, 9.9515E+04, 9.9983E+04, 1.0045E+05,
     +  1.0092E+05, 1.0140E+05, 1.0187E+05, 1.0235E+05, 1.0283E+05,
     + 1.0331E+05, 1.0379E+05, 1.0428E+05, 1.0476E+05, 1.0525E+05,
     +  1.0574E+05, 1.0623E+05, 1.0673E+05, 1.0722E+05, 1.0772E+05,
     + 1.0822E+05, 1.0872E+05, 1.0922E+05, 1.0973E+05, 1.1023E+05,
     +  1.1074E+05, 1.1125E+05, 1.1176E+05, 1.1228E+05, 1.1279E+05,
     + 1.1331E+05, 1.1383E+05, 1.1435E+05, 1.1488E+05, 1.1540E+05,
     +  1.1593E+05, 1.1646E+05, 1.1699E+05, 1.1753E+05, 1.1806E+05,
     + 1.1860E+05, 1.1914E+05, 1.1968E+05, 1.2023E+05, 1.2077E+05,
     +  1.2132E+05, 1.2187E+05, 1.2242E+05, 1.2298E+05, 1.2353E+05,
     + 1.2409E+05, 1.2465E+05, 1.2521E+05, 1.2578E+05, 1.2634E+05,
     +  1.2691E+05, 1.2748E+05, 1.2805E+05, 1.2863E+05, 1.2920E+05,
     + 1.2978E+05, 1.3036E+05, 1.3095E+05, 1.3153E+05, 1.3212E+05,
     +  1.3271E+05, 1.3330E+05, 1.3389E+05, 1.3449E+05, 1.3509E+05,
     + 1.3569E+05, 1.3629E+05, 1.3689E+05, 1.3750E+05, 1.3811E+05,
     +  1.3872E+05, 1.3933E+05, 1.3995E+05, 1.4056E+05, 1.4118E+05,
     + 1.4180E+05, 1.4243E+05, 1.4305E+05, 1.4368E+05, 1.4431E+05,
     +  1.4495E+05, 1.4558E+05, 1.4622E+05, 1.4686E+05, 1.4750E+05,
     + 1.4815E+05, 1.4879E+05, 1.4944E+05, 1.5009E+05, 1.5075E+05,
     +  1.5140E+05, 1.5206E+05, 1.5272E+05, 1.5338E+05, 1.5405E+05,
     + 1.5472E+05, 1.5539E+05, 1.5606E+05, 1.5673E+05, 1.5741E+05,
     +  1.5809E+05, 1.5877E+05, 1.5945E+05, 1.6014E+05, 1.6083E+05,
     + 1.6152E+05, 1.6221E+05, 1.6291E+05, 1.6361E+05, 1.6431E+05,
     +  1.6501E+05, 1.6572E+05, 1.6643E+05, 1.6714E+05, 1.6785E+05,
     + 1.6857E+05, 1.6928E+05, 1.7000E+05, 1.7073E+05, 1.7145E+05,
     +  1.7218E+05, 1.7291E+05, 1.7364E+05, 1.7438E+05, 1.7512E+05,
     + 1.7586E+05, 1.7660E+05, 1.7735E+05, 1.7810E+05, 1.7885E+05,
     +  1.7960E+05, 1.8036E+05, 1.8111E+05, 1.8187E+05, 1.8264E+05,
     + 1.8340E+05, 1.8417E+05, 1.8495E+05, 1.8572E+05, 1.8650E+05,
     +  1.8728E+05, 1.8806E+05, 1.8884E+05, 1.8963E+05, 1.9042E+05,
     + 1.9121E+05, 1.9201E+05, 1.9281E+05, 1.9361E+05, 1.9441E+05,
     +  1.9522E+05, 1.9602E+05, 1.9684E+05, 1.9765E+05, 1.9847E+05,
     + 1.9929E+05, 2.0011E+05, 2.0093E+05, 2.0176E+05, 2.0259E+05,
     +  2.0343E+05, 2.0426E+05, 2.0510E+05, 2.0594E+05, 2.0679E+05,
     + 2.0764E+05, 2.0849E+05, 2.0934E+05, 2.1020E+05, 2.1105E+05,
     +  2.1192E+05, 2.1278E+05, 2.1365E+05, 2.1452E+05, 2.1539E+05,
     + 2.1627E+05, 2.1715E+05, 2.1803E+05, 2.1891E+05, 2.1980E+05,
     +  2.2069E+05, 2.2158E+05, 2.2248E+05, 2.2338E+05, 2.2428E+05,
     + 2.2519E+05, 2.2609E+05, 2.2701E+05, 2.2792E+05, 2.2884E+05,
     +  2.2976E+05, 2.3068E+05, 2.3161E+05, 2.3253E+05, 2.3347E+05,
     + 2.3440E+05, 2.3534E+05, 2.3628E+05, 2.3723E+05, 2.3817E+05,
     +  2.3912E+05, 2.4008E+05, 2.4103E+05, 2.4199E+05, 2.4295E+05,
     + 2.4392E+05, 2.4489E+05, 2.4586E+05, 2.4684E+05, 2.4781E+05,
     +  2.4880E+05, 2.4978E+05, 2.5077E+05, 2.5176E+05, 2.5275E+05,
     + 2.5375E+05, 2.5475E+05, 2.5575E+05, 2.5676E+05, 2.5777E+05,
     +  2.5878E+05, 2.5980E+05, 2.6082E+05, 2.6184E+05, 2.6287E+05,
     + 2.6390E+05, 2.6493E+05, 2.6597E+05, 2.6701E+05, 2.6805E+05,
     +  2.6909E+05, 2.7014E+05, 2.7120E+05, 2.7225E+05, 2.7331E+05,
     + 2.7437E+05, 2.7544E+05, 2.7651E+05, 2.7758E+05, 2.7866E+05,
     +  2.7974E+05, 2.8082E+05, 2.8190E+05, 2.8299E+05, 2.8409E+05,
     + 2.8518E+05, 2.8628E+05, 2.8739E+05, 2.8849E+05, 2.8960E+05,
     +  2.9071E+05, 2.9183E+05, 2.9295E+05, 2.9408E+05, 2.9520E+05,
     + 2.9633E+05, 2.9747E+05, 2.9861E+05, 2.9975E+05, 3.0089E+05,
     +  3.0204E+05, 3.0319E+05, 3.0435E+05, 3.0551E+05, 3.0667E+05,
     + 3.0784E+05, 3.0901E+05, 3.1018E+05, 3.1136E+05, 3.1254E+05,
     +  3.1372E+05, 3.1491E+05, 3.1610E+05, 3.1730E+05, 3.1849E+05,
     + 3.1970E+05, 3.2090E+05, 3.2211E+05, 3.2333E+05, 3.2454E+05,
     +  3.2576E+05, 3.2699E+05, 3.2822E+05, 3.2945E+05, 3.3069E+05,
     + 3.3193E+05, 3.3317E+05, 3.3442E+05, 3.3567E+05, 3.3692E+05,
     +  3.3818E+05, 3.3944E+05, 3.4071E+05, 3.4198E+05, 3.4325E+05,
     + 3.4453E+05, 3.4581E+05, 3.4710E+05, 3.4839E+05, 3.4968E+05,
     +  3.5098E+05, 3.5228E+05, 3.5358E+05, 3.5489E+05, 3.5620E+05,
     + 3.5752E+05, 3.5884E+05, 3.6016E+05, 3.6149E+05, 3.6282E+05,
     +  3.6416E+05, 3.6550E+05, 3.6684E+05, 3.6819E+05, 3.6955E+05,
     + 3.7090E+05, 3.7226E+05, 3.7363E+05, 3.7499E+05, 3.7637E+05,
     +  3.7774E+05, 3.7912E+05, 3.8051E+05, 3.8190E+05, 3.8329E+05,
     + 3.8469E+05, 3.8609E+05, 3.8749E+05, 3.8890E+05, 3.9032E+05,
     +  3.9173E+05, 3.9315E+05, 3.9458E+05, 3.9601E+05, 3.9745E+05,
     + 3.9888E+05, 4.0033E+05, 4.0177E+05, 4.0322E+05, 4.0468E+05,
     +  4.0614E+05, 4.0760E+05, 4.0907E+05, 4.1054E+05, 4.1202E+05,
     + 4.1350E+05, 4.1499E+05, 4.1648E+05, 4.1797E+05, 4.1947E+05,
     +  4.2097E+05, 4.2248E+05, 4.2399E+05, 4.2551E+05, 4.2703E+05,
     + 4.2855E+05, 4.3008E+05, 4.3161E+05, 4.3315E+05, 4.3469E+05,
     +  4.3624E+05, 4.3779E+05, 4.3935E+05, 4.4091E+05, 4.4247E+05,
     + 4.4404E+05, 4.4561E+05, 4.4719E+05, 4.4877E+05, 4.5036E+05,
     +  4.5195E+05, 4.5355E+05, 4.5515E+05, 4.5675E+05, 4.5836E+05,
     + 4.5998E+05, 4.6160E+05, 4.6322E+05, 4.6485E+05, 4.6648E+05,
     +  4.6812E+05, 4.6976E+05, 4.7141E+05, 4.7306E+05, 4.7472E+05,
     + 4.7638E+05, 4.7804E+05, 4.7971E+05, 4.8139E+05, 4.8307E+05,
     +  4.8475E+05, 4.8644E+05, 4.8813E+05, 4.8983E+05, 4.9154E+05,
     + 4.9324E+05, 4.9496E+05, 4.9667E+05, 4.9840E+05, 5.0012E+05,
     +  5.0186E+05, 5.0359E+05, 5.0534E+05, 5.0708E+05, 5.0883E+05,
     + 5.1059E+05, 5.1235E+05, 5.1412E+05, 5.1589E+05, 5.1767E+05,
     +  5.1945E+05, 5.2123E+05, 5.2303E+05, 5.2482E+05, 5.2662E+05,
     + 5.2843E+05, 5.3024E+05, 5.3206E+05, 5.3388E+05, 5.3570E+05,
     +  5.3753E+05, 5.3937E+05, 5.4121E+05, 5.4306E+05, 5.4491E+05,
     + 5.4677E+05, 5.4863E+05, 5.5050E+05, 5.5237E+05, 5.5425E+05,
     +  5.5613E+05, 5.5801E+05, 5.5991E+05, 5.6181E+05, 5.6371E+05,
     + 5.6562E+05, 5.6753E+05, 5.6945E+05, 5.7137E+05, 5.7330E+05,
     +  5.7524E+05, 5.7718E+05, 5.7912E+05, 5.8107E+05, 5.8303E+05,
     + 5.8499E+05, 5.8695E+05, 5.8893E+05, 5.9090E+05, 5.9289E+05,
     +  5.9487E+05, 5.9687E+05, 5.9886E+05, 6.0087E+05, 6.0288E+05,
     + 6.0489E+05, 6.0691E+05, 6.0894E+05, 6.1097E+05, 6.1301E+05,
     +  6.1505E+05, 6.1710E+05, 6.1915E+05, 6.2121E+05, 6.2327E+05,
     + 6.2534E+05, 6.2742E+05, 6.2950E+05, 6.3158E+05, 6.3367E+05,
     +  6.3577E+05, 6.3788E+05, 6.3998E+05, 6.4210E+05, 6.4422E+05,
     + 6.4634E+05, 6.4848E+05, 6.5061E+05, 6.5275E+05, 6.5490E+05,
     +  6.5706E+05, 6.5922E+05, 6.6138E+05, 6.6355E+05, 6.6573E+05,
     + 6.6791E+05, 6.7010E+05, 6.7230E+05, 6.7450E+05, 6.7670E+05,
     +  6.7891E+05, 6.8113E+05, 6.8336E+05, 6.8559E+05, 6.8782E+05,
     + 6.9006E+05, 6.9231E+05, 6.9456E+05, 6.9682E+05, 6.9909E+05,
     +  7.0136E+05, 7.0363E+05, 7.0592E+05, 7.0820E+05, 7.1050E+05,
     + 7.1280E+05, 7.1511E+05, 7.1742E+05, 7.1974E+05, 7.2206E+05,
     +  7.2439E+05, 7.2673E+05, 7.2907E+05, 7.3142E+05, 7.3378E+05,
     + 7.3614E+05, 7.3851E+05, 7.4088E+05, 7.4326E+05, 7.4565E+05,
     +  7.4804E+05, 7.5044E+05, 7.5284E+05, 7.5525E+05, 7.5767E+05,
     + 7.6009E+05, 7.6252E+05, 7.6496E+05, 7.6740E+05, 7.6985E+05,
     +  7.7230E+05, 7.7476E+05, 7.7723E+05, 7.7970E+05, 7.8218E+05,
     + 7.8467E+05, 7.8716E+05, 7.8966E+05, 7.9217E+05, 7.9468E+05,
     +  7.9720E+05, 7.9972E+05, 8.0225E+05, 8.0479E+05, 8.0734E+05,
     + 8.0989E+05, 8.1244E+05, 8.1501E+05, 8.1758E+05, 8.2016E+05,
     +  8.2274E+05, 8.2533E+05, 8.2793E+05, 8.3053E+05, 8.3314E+05,
     + 8.3575E+05, 8.3838E+05, 8.4101E+05, 8.4364E+05, 8.4629E+05,
     +  8.4894E+05, 8.5159E+05, 8.5426E+05, 8.5693E+05, 8.5960E+05,
     + 8.6229E+05, 8.6498E+05, 8.6767E+05, 8.7038E+05, 8.7309E+05,
     +  8.7580E+05, 8.7853E+05, 8.8126E+05, 8.8400E+05, 8.8674E+05,
     + 8.8949E+05, 8.9225E+05, 8.9502E+05, 8.9779E+05, 9.0057E+05,
     +  9.0335E+05, 9.0615E+05, 9.0895E+05, 9.1175E+05, 9.1457E+05,
     + 9.1739E+05, 9.2022E+05, 9.2305E+05, 9.2589E+05, 9.2874E+05,
     +  9.3160E+05, 9.3446E+05, 9.3733E+05, 9.4021E+05, 9.4310E+05,
     + 9.4599E+05, 9.4889E+05, 9.5179E+05, 9.5471E+05, 9.5763E+05,
     +  9.6055E+05, 9.6349E+05, 9.6643E+05, 9.6938E+05, 9.7234E+05,
     + 9.7530E+05, 9.7827E+05, 9.8125E+05, 9.8424E+05, 9.8723E+05,
     +  9.9023E+05, 9.9324E+05, 9.9625E+05, 9.9928E+05, 1.0023E+06,
     + 1.0053E+06, 1.0084E+06, 1.0114E+06, 1.0145E+06, 1.0176E+06,
     +  1.0206E+06, 1.0237E+06, 1.0268E+06, 1.0299E+06, 1.0330E+06,
     + 1.0361E+06, 1.0393E+06, 1.0424E+06, 1.0455E+06, 1.0487E+06,
     +  1.0518E+06, 1.0550E+06, 1.0581E+06, 1.0613E+06, 1.0645E+06,
     + 1.0677E+06, 1.0709E+06, 1.0741E+06, 1.0773E+06, 1.0805E+06,
     +  1.0837E+06, 1.0870E+06, 1.0902E+06, 1.0935E+06, 1.0967E+06,
     + 1.1000E+06, 1.1033E+06, 1.1065E+06, 1.1098E+06, 1.1131E+06,
     +  1.1164E+06, 1.1198E+06, 1.1231E+06, 1.1264E+06, 1.1297E+06,
     + 1.1331E+06, 1.1364E+06, 1.1398E+06, 1.1432E+06, 1.1466E+06,
     +  1.1499E+06, 1.1533E+06, 1.1567E+06, 1.1601E+06, 1.1636E+06,
     + 1.1670E+06, 1.1704E+06, 1.1739E+06, 1.1773E+06, 1.1808E+06,
     +  1.1842E+06, 1.1877E+06, 1.1912E+06, 1.1947E+06, 1.1982E+06,
     + 1.2017E+06, 1.2052E+06, 1.2087E+06, 1.2123E+06, 1.2158E+06,
     +  1.2194E+06, 1.2229E+06, 1.2265E+06, 1.2301E+06, 1.2336E+06,
     + 1.2372E+06, 1.2408E+06, 1.2445E+06, 1.2481E+06, 1.2517E+06,
     +  1.2553E+06, 1.2590E+06, 1.2626E+06, 1.2663E+06, 1.2699E+06,
     + 1.2736E+06, 1.2773E+06, 1.2810E+06, 1.2847E+06, 1.2884E+06,
     +  1.2921E+06, 1.2959E+06, 1.2996E+06, 1.3034E+06, 1.3071E+06,
     + 1.3109E+06, 1.3146E+06, 1.3184E+06, 1.3222E+06, 1.3260E+06,
     +  1.3298E+06, 1.3336E+06, 1.3375E+06, 1.3413E+06, 1.3451E+06,
     + 1.3490E+06, 1.3528E+06, 1.3567E+06, 1.3606E+06, 1.3645E+06,
     +  1.3684E+06, 1.3723E+06, 1.3762E+06, 1.3801E+06, 1.3841E+06,
     + 1.3880E+06, 1.3919E+06, 1.3959E+06, 1.3999E+06, 1.4039E+06,
     +  1.4078E+06, 1.4118E+06, 1.4158E+06, 1.4199E+06, 1.4239E+06,
     + 1.4279E+06, 1.4320E+06, 1.4360E+06, 1.4401E+06, 1.4441E+06,
     +  1.4482E+06, 1.4523E+06, 1.4564E+06, 1.4605E+06, 1.4646E+06,
     + 1.4687E+06, 1.4729E+06, 1.4770E+06, 1.4812E+06, 1.4853E+06,
     +  1.4895E+06, 1.4937E+06, 1.4979E+06, 1.5021E+06, 1.5063E+06,
     + 1.5105E+06, 1.5147E+06, 1.5190E+06, 1.5232E+06, 1.5275E+06,
     +  1.5318E+06, 1.5360E+06, 1.5403E+06, 1.5446E+06, 1.5489E+06,
     + 1.5532E+06, 1.5576E+06, 1.5619E+06, 1.5662E+06, 1.5706E+06,
     +  1.5750E+06, 1.5793E+06, 1.5837E+06, 1.5881E+06, 1.5925E+06,
     + 1.5969E+06, 1.6014E+06, 1.6058E+06, 1.6102E+06, 1.6147E+06,
     +  1.6192E+06, 1.6236E+06, 1.6281E+06, 1.6326E+06, 1.6371E+06,
     + 1.6416E+06, 1.6461E+06, 1.6507E+06, 1.6552E+06, 1.6598E+06,
     +  1.6643E+06, 1.6689E+06, 1.6735E+06, 1.6781E+06, 1.6827E+06,
     + 1.6873E+06, 1.6919E+06, 1.6966E+06, 1.7012E+06, 1.7059E+06,
     +  1.7105E+06, 1.7152E+06, 1.7199E+06, 1.7246E+06, 1.7293E+06,
     + 1.7340E+06, 1.7387E+06, 1.7435E+06, 1.7482E+06, 1.7530E+06,
     +  1.7577E+06, 1.7625E+06, 1.7673E+06, 1.7721E+06, 1.7769E+06,
     + 1.7818E+06, 1.7866E+06, 1.7914E+06, 1.7963E+06, 1.8011E+06,
     +  1.8060E+06, 1.8109E+06, 1.8158E+06, 1.8207E+06, 1.8256E+06,
     + 1.8305E+06, 1.8355E+06, 1.8404E+06, 1.8454E+06, 1.8504E+06,
     +  1.8553E+06, 1.8603E+06, 1.8653E+06, 1.8704E+06, 1.8754E+06,
     + 1.8804E+06, 1.8855E+06, 1.8905E+06, 1.8956E+06, 1.9007E+06,
     +  1.9058E+06, 1.9109E+06, 1.9160E+06, 1.9211E+06, 1.9262E+06,
     + 1.9314E+06, 1.9365E+06, 1.9417E+06, 1.9469E+06, 1.9521E+06,
     +  1.9573E+06, 1.9625E+06, 1.9677E+06, 1.9729E+06, 1.9782E+06,
     + 1.9834E+06, 1.9887E+06, 1.9940E+06, 1.9993E+06, 2.0046E+06,
     +  2.0099E+06, 2.0152E+06, 2.0206E+06, 2.0259E+06, 2.0313E+06,
     + 2.0366E+06, 2.0420E+06, 2.0474E+06, 2.0528E+06, 2.0582E+06,
     +  2.0637E+06, 2.0691E+06, 2.0746E+06, 2.0800E+06, 2.0855E+06,
     + 2.0910E+06, 2.0965E+06, 2.1020E+06, 2.1075E+06, 2.1130E+06,
     +  2.1186E+06, 2.1241E+06, 2.1297E+06, 2.1353E+06, 2.1409E+06,
     + 2.1465E+06, 2.1521E+06, 2.1577E+06, 2.1634E+06, 2.1690E+06,
     +  2.1747E+06, 2.1804E+06, 2.1860E+06, 2.1917E+06, 2.1975E+06,
     + 2.2032E+06, 2.2089E+06, 2.2147E+06, 2.2204E+06, 2.2262E+06,
     +  2.2320E+06, 2.2378E+06, 2.2436E+06, 2.2494E+06, 2.2552E+06,
     + 2.2611E+06, 2.2669E+06, 2.2728E+06, 2.2787E+06, 2.2846E+06,
     +  2.2905E+06, 2.2964E+06, 2.3023E+06, 2.3083E+06, 2.3142E+06,
     + 2.3202E+06, 2.3262E+06, 2.3321E+06, 2.3381E+06, 2.3442E+06,
     +  2.3502E+06, 2.3562E+06, 2.3623E+06, 2.3684E+06, 2.3744E+06,
     + 2.3805E+06, 2.3866E+06, 2.3927E+06, 2.3989E+06, 2.4050E+06,
     +  2.4112E+06, 2.4173E+06, 2.4235E+06, 2.4297E+06, 2.4359E+06,
     + 2.4421E+06, 2.4484E+06, 2.4546E+06, 2.4609E+06, 2.4671E+06,
     +  2.4734E+06, 2.4797E+06, 2.4860E+06, 2.4923E+06, 2.4987E+06,
     + 2.5050E+06, 2.5114E+06, 2.5178E+06, 2.5241E+06, 2.5305E+06,
     +  2.5369E+06, 2.5434E+06, 2.5498E+06, 2.5563E+06, 2.5627E+06,
     + 2.5692E+06, 2.5757E+06, 2.5822E+06, 2.5887E+06, 2.5952E+06,
     +  2.6018E+06, 2.6084E+06, 2.6149E+06, 2.6215E+06, 2.6281E+06,
     + 2.6347E+06, 2.6413E+06, 2.6480E+06, 2.6546E+06, 2.6613E+06,
     +  2.6680E+06, 2.6747E+06, 2.6814E+06, 2.6881E+06, 2.6948E+06,
     + 2.7016E+06, 2.7083E+06, 2.7151E+06, 2.7219E+06, 2.7287E+06,
     +  2.7355E+06, 2.7423E+06, 2.7492E+06, 2.7560E+06, 2.7629E+06,
     + 2.7698E+06, 2.7767E+06, 2.7836E+06, 2.7905E+06, 2.7974E+06,
     +  2.8044E+06, 2.8114E+06, 2.8183E+06, 2.8253E+06, 2.8323E+06,
     + 2.8394E+06, 2.8464E+06, 2.8534E+06, 2.8605E+06, 2.8676E+06,
     +  2.8747E+06, 2.8818E+06, 2.8889E+06, 2.8960E+06, 2.9032E+06,
     + 2.9104E+06, 2.9175E+06, 2.9247E+06, 2.9319E+06, 2.9392E+06,
     +  2.9464E+06, 2.9536E+06, 2.9609E+06, 2.9682E+06, 2.9755E+06,
     + 2.9828E+06, 2.9901E+06, 2.9974E+06, 3.0048E+06, 3.0121E+06,
     +  3.0195E+06, 3.0269E+06, 3.0343E+06, 3.0418E+06, 3.0492E+06,
     + 3.0566E+06, 3.0641E+06, 3.0716E+06, 3.0791E+06, 3.0866E+06,
     +  3.0941E+06, 3.1017E+06, 3.1092E+06, 3.1168E+06, 3.1244E+06,
     + 3.1320E+06, 3.1396E+06, 3.1472E+06, 3.1549E+06, 3.1625E+06,
     +  3.1702E+06, 3.1779E+06, 3.1856E+06, 3.1933E+06, 3.2011E+06,
     + 3.2088E+06, 3.2166E+06, 3.2243E+06, 3.2321E+06, 3.2400E+06,
     +  3.2478E+06, 3.2556E+06, 3.2635E+06, 3.2713E+06, 3.2792E+06,
     + 3.2871E+06, 3.2951E+06, 3.3030E+06, 3.3109E+06, 3.3189E+06,
     +  3.3269E+06, 3.3349E+06, 3.3429E+06, 3.3509E+06, 3.3590E+06,
     + 3.3670E+06, 3.3751E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       546
      data (QofT( 3,J),J=1,2408)/ 1.2565E+01, 2.2727E+01, 3.2986E+01,
     +  4.3267E+01, 5.3558E+01, 6.3852E+01, 7.4149E+01, 8.4448E+01,
     + 9.4748E+01, 1.0505E+02, 1.1535E+02, 1.2565E+02, 1.3596E+02,
     +  1.4626E+02, 1.5656E+02, 1.6687E+02, 1.7717E+02, 1.8748E+02,
     + 1.9778E+02, 2.0809E+02, 2.2870E+02, 2.4931E+02, 2.6992E+02,
     +  2.9054E+02, 3.1116E+02, 3.3177E+02, 3.5239E+02, 3.7301E+02,
     + 3.9363E+02, 4.1425E+02, 4.3487E+02, 4.5549E+02, 4.7612E+02,
     +  4.9674E+02, 5.1737E+02, 5.3800E+02, 5.5862E+02, 5.7925E+02,
     + 5.9988E+02, 6.2051E+02, 6.4114E+02, 6.6177E+02, 6.8241E+02,
     +  7.0305E+02, 7.2368E+02, 7.4433E+02, 7.6497E+02, 7.8561E+02,
     + 8.0626E+02, 8.2691E+02, 8.4757E+02, 8.6823E+02, 8.8889E+02,
     +  9.0956E+02, 9.3024E+02, 9.5093E+02, 9.7162E+02, 9.9233E+02,
     + 1.0130E+03, 1.0338E+03, 1.0545E+03, 1.0753E+03, 1.0960E+03,
     +  1.1168E+03, 1.1377E+03, 1.1585E+03, 1.1793E+03, 1.2002E+03,
     + 1.2211E+03, 1.2421E+03, 1.2631E+03, 1.2841E+03, 1.3051E+03,
     +  1.3262E+03, 1.3473E+03, 1.3685E+03, 1.3897E+03, 1.4109E+03,
     + 1.4322E+03, 1.4536E+03, 1.4750E+03, 1.4965E+03, 1.5180E+03,
     +  1.5396E+03, 1.5612E+03, 1.5830E+03, 1.6047E+03, 1.6266E+03,
     + 1.6485E+03, 1.6705E+03, 1.6926E+03, 1.7148E+03, 1.7370E+03,
     +  1.7593E+03, 1.7818E+03, 1.8043E+03, 1.8269E+03, 1.8495E+03,
     + 1.8723E+03, 1.8952E+03, 1.9182E+03, 1.9413E+03, 1.9645E+03,
     +  1.9878E+03, 2.0112E+03, 2.0347E+03, 2.0584E+03, 2.0821E+03,
     + 2.1060E+03, 2.1300E+03, 2.1541E+03, 2.1783E+03, 2.2027E+03,
     +  2.2272E+03, 2.2518E+03, 2.2766E+03, 2.3014E+03, 2.3265E+03,
     + 2.3516E+03, 2.3769E+03, 2.4024E+03, 2.4279E+03, 2.4537E+03,
     +  2.4795E+03, 2.5056E+03, 2.5317E+03, 2.5581E+03, 2.5845E+03,
     + 2.6112E+03, 2.6380E+03, 2.6649E+03, 2.6920E+03, 2.7193E+03,
     +  2.7468E+03, 2.7744E+03, 2.8021E+03, 2.8301E+03, 2.8582E+03,
     + 2.8865E+03, 2.9149E+03, 2.9435E+03, 2.9723E+03, 3.0013E+03,
     +  3.0305E+03, 3.0598E+03, 3.0894E+03, 3.1191E+03, 3.1490E+03,
     + 3.1791E+03, 3.2094E+03, 3.2398E+03, 3.2705E+03, 3.3013E+03,
     +  3.3324E+03, 3.3636E+03, 3.3951E+03, 3.4267E+03, 3.4586E+03,
     + 3.4906E+03, 3.5229E+03, 3.5553E+03, 3.5880E+03, 3.6209E+03,
     +  3.6540E+03, 3.6873E+03, 3.7208E+03, 3.7545E+03, 3.7885E+03,
     + 3.8226E+03, 3.8570E+03, 3.8916E+03, 3.9264E+03, 3.9615E+03,
     +  3.9968E+03, 4.0322E+03, 4.0680E+03, 4.1039E+03, 4.1401E+03,
     + 4.1765E+03, 4.2132E+03, 4.2501E+03, 4.2872E+03, 4.3246E+03,
     +  4.3622E+03, 4.4000E+03, 4.4381E+03, 4.4764E+03, 4.5150E+03,
     + 4.5538E+03, 4.5929E+03, 4.6322E+03, 4.6718E+03, 4.7116E+03,
     +  4.7517E+03, 4.7920E+03, 4.8326E+03, 4.8734E+03, 4.9146E+03,
     + 4.9559E+03, 4.9976E+03, 5.0394E+03, 5.0816E+03, 5.1241E+03,
     +  5.1668E+03, 5.2098E+03, 5.2530E+03, 5.2965E+03, 5.3403E+03,
     + 5.3844E+03, 5.4287E+03, 5.4733E+03, 5.5183E+03, 5.5635E+03,
     +  5.6089E+03, 5.6547E+03, 5.7008E+03, 5.7471E+03, 5.7937E+03,
     + 5.8406E+03, 5.8879E+03, 5.9354E+03, 5.9832E+03, 6.0313E+03,
     +  6.0797E+03, 6.1284E+03, 6.1774E+03, 6.2267E+03, 6.2764E+03,
     + 6.3263E+03, 6.3766E+03, 6.4271E+03, 6.4780E+03, 6.5291E+03,
     +  6.5807E+03, 6.6324E+03, 6.6846E+03, 6.7370E+03, 6.7898E+03,
     + 6.8429E+03, 6.8963E+03, 6.9501E+03, 7.0042E+03, 7.0586E+03,
     +  7.1133E+03, 7.1684E+03, 7.2238E+03, 7.2796E+03, 7.3357E+03,
     + 7.3921E+03, 7.4489E+03, 7.5060E+03, 7.5635E+03, 7.6213E+03,
     +  7.6794E+03, 7.7379E+03, 7.7968E+03, 7.8560E+03, 7.9156E+03,
     + 7.9755E+03, 8.0358E+03, 8.0965E+03, 8.1575E+03, 8.2189E+03,
     +  8.2807E+03, 8.3428E+03, 8.4052E+03, 8.4681E+03, 8.5313E+03,
     + 8.5950E+03, 8.6590E+03, 8.7233E+03, 8.7881E+03, 8.8532E+03,
     +  8.9187E+03, 8.9846E+03, 9.0509E+03, 9.1176E+03, 9.1847E+03,
     + 9.2521E+03, 9.3200E+03, 9.3883E+03, 9.4570E+03, 9.5260E+03,
     +  9.5955E+03, 9.6653E+03, 9.7356E+03, 9.8063E+03, 9.8774E+03,
     + 9.9489E+03, 1.0021E+04, 1.0093E+04, 1.0166E+04, 1.0239E+04,
     +  1.0313E+04, 1.0387E+04, 1.0461E+04, 1.0536E+04, 1.0611E+04,
     + 1.0687E+04, 1.0763E+04, 1.0840E+04, 1.0917E+04, 1.0994E+04,
     +  1.1072E+04, 1.1151E+04, 1.1230E+04, 1.1309E+04, 1.1389E+04,
     + 1.1469E+04, 1.1549E+04, 1.1631E+04, 1.1712E+04, 1.1794E+04,
     +  1.1877E+04, 1.1960E+04, 1.2043E+04, 1.2127E+04, 1.2211E+04,
     + 1.2296E+04, 1.2381E+04, 1.2467E+04, 1.2553E+04, 1.2640E+04,
     +  1.2727E+04, 1.2815E+04, 1.2903E+04, 1.2992E+04, 1.3081E+04,
     + 1.3171E+04, 1.3261E+04, 1.3351E+04, 1.3443E+04, 1.3534E+04,
     +  1.3626E+04, 1.3719E+04, 1.3812E+04, 1.3906E+04, 1.4000E+04,
     + 1.4095E+04, 1.4190E+04, 1.4286E+04, 1.4382E+04, 1.4479E+04,
     +  1.4576E+04, 1.4674E+04, 1.4772E+04, 1.4871E+04, 1.4971E+04,
     + 1.5071E+04, 1.5171E+04, 1.5273E+04, 1.5374E+04, 1.5476E+04,
     +  1.5579E+04, 1.5682E+04, 1.5786E+04, 1.5891E+04, 1.5996E+04,
     + 1.6101E+04, 1.6207E+04, 1.6314E+04, 1.6421E+04, 1.6529E+04,
     +  1.6637E+04, 1.6746E+04, 1.6856E+04, 1.6966E+04, 1.7077E+04,
     + 1.7188E+04, 1.7300E+04, 1.7412E+04, 1.7525E+04, 1.7639E+04,
     +  1.7753E+04, 1.7868E+04, 1.7983E+04, 1.8099E+04, 1.8216E+04,
     + 1.8333E+04, 1.8451E+04, 1.8570E+04, 1.8689E+04, 1.8809E+04,
     +  1.8929E+04, 1.9050E+04, 1.9172E+04, 1.9294E+04, 1.9417E+04,
     + 1.9540E+04, 1.9664E+04, 1.9789E+04, 1.9915E+04, 2.0041E+04,
     +  2.0167E+04, 2.0295E+04, 2.0423E+04, 2.0552E+04, 2.0681E+04,
     + 2.0811E+04, 2.0942E+04, 2.1073E+04, 2.1205E+04, 2.1338E+04,
     +  2.1471E+04, 2.1605E+04, 2.1740E+04, 2.1875E+04, 2.2011E+04,
     + 2.2148E+04, 2.2286E+04, 2.2424E+04, 2.2563E+04, 2.2702E+04,
     +  2.2842E+04, 2.2984E+04, 2.3125E+04, 2.3268E+04, 2.3411E+04,
     + 2.3554E+04, 2.3699E+04, 2.3844E+04, 2.3990E+04, 2.4137E+04,
     +  2.4285E+04, 2.4433E+04, 2.4582E+04, 2.4731E+04, 2.4882E+04,
     + 2.5033E+04, 2.5185E+04, 2.5337E+04, 2.5491E+04, 2.5645E+04,
     +  2.5800E+04, 2.5956E+04, 2.6112E+04, 2.6269E+04, 2.6427E+04,
     + 2.6586E+04, 2.6745E+04, 2.6906E+04, 2.7067E+04, 2.7229E+04,
     +  2.7391E+04, 2.7555E+04, 2.7719E+04, 2.7884E+04, 2.8050E+04,
     + 2.8217E+04, 2.8384E+04, 2.8552E+04, 2.8721E+04, 2.8891E+04,
     +  2.9062E+04, 2.9234E+04, 2.9406E+04, 2.9579E+04, 2.9753E+04,
     + 2.9928E+04, 3.0104E+04, 3.0280E+04, 3.0458E+04, 3.0636E+04,
     +  3.0815E+04, 3.0995E+04, 3.1176E+04, 3.1357E+04, 3.1540E+04,
     + 3.1723E+04, 3.1907E+04, 3.2093E+04, 3.2279E+04, 3.2465E+04,
     +  3.2653E+04, 3.2842E+04, 3.3031E+04, 3.3222E+04, 3.3413E+04,
     + 3.3605E+04, 3.3798E+04, 3.3992E+04, 3.4187E+04, 3.4383E+04,
     +  3.4580E+04, 3.4777E+04, 3.4976E+04, 3.5176E+04, 3.5376E+04,
     + 3.5577E+04, 3.5780E+04, 3.5983E+04, 3.6187E+04, 3.6392E+04,
     +  3.6598E+04, 3.6805E+04, 3.7013E+04, 3.7222E+04, 3.7432E+04,
     + 3.7643E+04, 3.7855E+04, 3.8068E+04, 3.8281E+04, 3.8496E+04,
     +  3.8712E+04, 3.8928E+04, 3.9146E+04, 3.9365E+04, 3.9585E+04,
     + 3.9805E+04, 4.0027E+04, 4.0250E+04, 4.0473E+04, 4.0698E+04,
     +  4.0924E+04, 4.1151E+04, 4.1378E+04, 4.1607E+04, 4.1837E+04,
     + 4.2068E+04, 4.2300E+04, 4.2533E+04, 4.2767E+04, 4.3002E+04,
     +  4.3238E+04, 4.3475E+04, 4.3713E+04, 4.3953E+04, 4.4193E+04,
     + 4.4434E+04, 4.4677E+04, 4.4920E+04, 4.5165E+04, 4.5411E+04,
     +  4.5658E+04, 4.5906E+04, 4.6155E+04, 4.6405E+04, 4.6656E+04,
     + 4.6908E+04, 4.7162E+04, 4.7416E+04, 4.7672E+04, 4.7929E+04,
     +  4.8187E+04, 4.8446E+04, 4.8706E+04, 4.8967E+04, 4.9230E+04,
     + 4.9493E+04, 4.9758E+04, 5.0024E+04, 5.0291E+04, 5.0559E+04,
     +  5.0829E+04, 5.1099E+04, 5.1371E+04, 5.1644E+04, 5.1918E+04,
     + 5.2194E+04, 5.2470E+04, 5.2748E+04, 5.3027E+04, 5.3307E+04,
     +  5.3588E+04, 5.3871E+04, 5.4154E+04, 5.4439E+04, 5.4725E+04,
     + 5.5013E+04, 5.5301E+04, 5.5591E+04, 5.5882E+04, 5.6174E+04,
     +  5.6468E+04, 5.6763E+04, 5.7059E+04, 5.7356E+04, 5.7655E+04,
     + 5.7955E+04, 5.8256E+04, 5.8558E+04, 5.8862E+04, 5.9167E+04,
     +  5.9473E+04, 5.9780E+04, 6.0089E+04, 6.0399E+04, 6.0711E+04,
     + 6.1024E+04, 6.1338E+04, 6.1653E+04, 6.1970E+04, 6.2288E+04,
     +  6.2607E+04, 6.2928E+04, 6.3250E+04, 6.3573E+04, 6.3898E+04,
     + 6.4224E+04, 6.4551E+04, 6.4880E+04, 6.5210E+04, 6.5541E+04,
     +  6.5874E+04, 6.6208E+04, 6.6544E+04, 6.6881E+04, 6.7220E+04,
     + 6.7559E+04, 6.7901E+04, 6.8243E+04, 6.8587E+04, 6.8932E+04,
     +  6.9279E+04, 6.9628E+04, 6.9977E+04, 7.0328E+04, 7.0681E+04,
     + 7.1035E+04, 7.1390E+04, 7.1747E+04, 7.2105E+04, 7.2465E+04,
     +  7.2826E+04, 7.3189E+04, 7.3553E+04, 7.3919E+04, 7.4286E+04,
     + 7.4654E+04, 7.5024E+04, 7.5396E+04, 7.5769E+04, 7.6144E+04,
     +  7.6520E+04, 7.6897E+04, 7.7276E+04, 7.7657E+04, 7.8039E+04,
     + 7.8423E+04, 7.8808E+04, 7.9195E+04, 7.9583E+04, 7.9973E+04,
     +  8.0364E+04, 8.0757E+04, 8.1152E+04, 8.1548E+04, 8.1946E+04,
     + 8.2345E+04, 8.2746E+04, 8.3148E+04, 8.3552E+04, 8.3958E+04,
     +  8.4365E+04, 8.4774E+04, 8.5184E+04, 8.5596E+04, 8.6010E+04,
     + 8.6425E+04, 8.6842E+04, 8.7261E+04, 8.7681E+04, 8.8103E+04,
     +  8.8526E+04, 8.8951E+04, 8.9378E+04, 8.9806E+04, 9.0236E+04,
     + 9.0668E+04, 9.1102E+04, 9.1537E+04, 9.1974E+04, 9.2412E+04,
     +  9.2853E+04, 9.3294E+04, 9.3738E+04, 9.4184E+04, 9.4631E+04,
     + 9.5080E+04, 9.5530E+04, 9.5982E+04, 9.6436E+04, 9.6892E+04,
     +  9.7350E+04, 9.7809E+04, 9.8270E+04, 9.8733E+04, 9.9198E+04,
     + 9.9664E+04, 1.0013E+05, 1.0060E+05, 1.0107E+05, 1.0155E+05,
     +  1.0202E+05, 1.0250E+05, 1.0298E+05, 1.0346E+05, 1.0394E+05,
     + 1.0443E+05, 1.0491E+05, 1.0540E+05, 1.0589E+05, 1.0638E+05,
     +  1.0688E+05, 1.0737E+05, 1.0787E+05, 1.0837E+05, 1.0887E+05,
     + 1.0937E+05, 1.0988E+05, 1.1038E+05, 1.1089E+05, 1.1140E+05,
     +  1.1192E+05, 1.1243E+05, 1.1295E+05, 1.1347E+05, 1.1399E+05,
     + 1.1451E+05, 1.1503E+05, 1.1556E+05, 1.1609E+05, 1.1662E+05,
     +  1.1715E+05, 1.1768E+05, 1.1822E+05, 1.1876E+05, 1.1929E+05,
     + 1.1984E+05, 1.2038E+05, 1.2093E+05, 1.2147E+05, 1.2202E+05,
     +  1.2258E+05, 1.2313E+05, 1.2369E+05, 1.2424E+05, 1.2480E+05,
     + 1.2537E+05, 1.2593E+05, 1.2650E+05, 1.2706E+05, 1.2763E+05,
     +  1.2821E+05, 1.2878E+05, 1.2936E+05, 1.2994E+05, 1.3052E+05,
     + 1.3110E+05, 1.3169E+05, 1.3227E+05, 1.3286E+05, 1.3345E+05,
     +  1.3405E+05, 1.3464E+05, 1.3524E+05, 1.3584E+05, 1.3644E+05,
     + 1.3705E+05, 1.3765E+05, 1.3826E+05, 1.3887E+05, 1.3949E+05,
     +  1.4010E+05, 1.4072E+05, 1.4134E+05, 1.4196E+05, 1.4258E+05,
     + 1.4321E+05, 1.4384E+05, 1.4447E+05, 1.4510E+05, 1.4574E+05,
     +  1.4638E+05, 1.4701E+05, 1.4766E+05, 1.4830E+05, 1.4895E+05,
     + 1.4960E+05, 1.5025E+05, 1.5090E+05, 1.5156E+05, 1.5222E+05,
     +  1.5288E+05, 1.5354E+05, 1.5420E+05, 1.5487E+05, 1.5554E+05,
     + 1.5621E+05, 1.5689E+05, 1.5756E+05, 1.5824E+05, 1.5892E+05,
     +  1.5961E+05, 1.6030E+05, 1.6098E+05, 1.6168E+05, 1.6237E+05,
     + 1.6306E+05, 1.6376E+05, 1.6446E+05, 1.6517E+05, 1.6587E+05,
     +  1.6658E+05, 1.6729E+05, 1.6800E+05, 1.6872E+05, 1.6944E+05,
     + 1.7016E+05, 1.7088E+05, 1.7161E+05, 1.7233E+05, 1.7306E+05,
     +  1.7380E+05, 1.7453E+05, 1.7527E+05, 1.7601E+05, 1.7675E+05,
     + 1.7750E+05, 1.7825E+05, 1.7900E+05, 1.7975E+05, 1.8051E+05,
     +  1.8127E+05, 1.8203E+05, 1.8279E+05, 1.8356E+05, 1.8433E+05,
     + 1.8510E+05, 1.8587E+05, 1.8665E+05, 1.8743E+05, 1.8821E+05,
     +  1.8899E+05, 1.8978E+05, 1.9057E+05, 1.9136E+05, 1.9216E+05,
     + 1.9296E+05, 1.9376E+05, 1.9456E+05, 1.9537E+05, 1.9618E+05,
     +  1.9699E+05, 1.9780E+05, 1.9862E+05, 1.9944E+05, 2.0026E+05,
     + 2.0108E+05, 2.0191E+05, 2.0274E+05, 2.0358E+05, 2.0441E+05,
     +  2.0525E+05, 2.0609E+05, 2.0694E+05, 2.0778E+05, 2.0863E+05,
     + 2.0949E+05, 2.1034E+05, 2.1120E+05, 2.1206E+05, 2.1293E+05,
     +  2.1379E+05, 2.1466E+05, 2.1554E+05, 2.1641E+05, 2.1729E+05,
     + 2.1817E+05, 2.1906E+05, 2.1995E+05, 2.2084E+05, 2.2173E+05,
     +  2.2262E+05, 2.2352E+05, 2.2443E+05, 2.2533E+05, 2.2624E+05,
     + 2.2715E+05, 2.2806E+05, 2.2898E+05, 2.2990E+05, 2.3082E+05,
     +  2.3175E+05, 2.3268E+05, 2.3361E+05, 2.3454E+05, 2.3548E+05,
     + 2.3642E+05, 2.3737E+05, 2.3831E+05, 2.3926E+05, 2.4022E+05,
     +  2.4117E+05, 2.4213E+05, 2.4309E+05, 2.4406E+05, 2.4503E+05,
     + 2.4600E+05, 2.4697E+05, 2.4795E+05, 2.4893E+05, 2.4992E+05,
     +  2.5090E+05, 2.5190E+05, 2.5289E+05, 2.5389E+05, 2.5489E+05,
     + 2.5589E+05, 2.5690E+05, 2.5790E+05, 2.5892E+05, 2.5993E+05,
     +  2.6095E+05, 2.6198E+05, 2.6300E+05, 2.6403E+05, 2.6506E+05,
     + 2.6610E+05, 2.6714E+05, 2.6818E+05, 2.6922E+05, 2.7027E+05,
     +  2.7133E+05, 2.7238E+05, 2.7344E+05, 2.7450E+05, 2.7557E+05,
     + 2.7664E+05, 2.7771E+05, 2.7878E+05, 2.7986E+05, 2.8094E+05,
     +  2.8203E+05, 2.8312E+05, 2.8421E+05, 2.8531E+05, 2.8641E+05,
     + 2.8751E+05, 2.8861E+05, 2.8972E+05, 2.9084E+05, 2.9195E+05,
     +  2.9307E+05, 2.9420E+05, 2.9532E+05, 2.9645E+05, 2.9759E+05,
     + 2.9872E+05, 2.9986E+05, 3.0101E+05, 3.0216E+05, 3.0331E+05,
     +  3.0446E+05, 3.0562E+05, 3.0678E+05, 3.0795E+05, 3.0912E+05,
     + 3.1029E+05, 3.1147E+05, 3.1265E+05, 3.1383E+05, 3.1502E+05,
     +  3.1621E+05, 3.1741E+05, 3.1860E+05, 3.1981E+05, 3.2101E+05,
     + 3.2222E+05, 3.2343E+05, 3.2465E+05, 3.2587E+05, 3.2709E+05,
     +  3.2832E+05, 3.2955E+05, 3.3079E+05, 3.3203E+05, 3.3327E+05,
     + 3.3452E+05, 3.3577E+05, 3.3702E+05, 3.3828E+05, 3.3954E+05,
     +  3.4081E+05, 3.4208E+05, 3.4335E+05, 3.4463E+05, 3.4591E+05,
     + 3.4719E+05, 3.4848E+05, 3.4977E+05, 3.5107E+05, 3.5237E+05,
     +  3.5367E+05, 3.5498E+05, 3.5629E+05, 3.5761E+05, 3.5893E+05,
     + 3.6025E+05, 3.6158E+05, 3.6291E+05, 3.6425E+05, 3.6559E+05,
     +  3.6693E+05, 3.6828E+05, 3.6963E+05, 3.7098E+05, 3.7234E+05,
     + 3.7371E+05, 3.7507E+05, 3.7645E+05, 3.7782E+05, 3.7920E+05,
     +  3.8059E+05, 3.8197E+05, 3.8337E+05, 3.8476E+05, 3.8616E+05,
     + 3.8757E+05, 3.8898E+05, 3.9039E+05, 3.9180E+05, 3.9323E+05,
     +  3.9465E+05, 3.9608E+05, 3.9751E+05, 3.9895E+05, 4.0039E+05,
     + 4.0184E+05, 4.0329E+05, 4.0474E+05, 4.0620E+05, 4.0767E+05,
     +  4.0913E+05, 4.1060E+05, 4.1208E+05, 4.1356E+05, 4.1505E+05,
     + 4.1653E+05, 4.1803E+05, 4.1952E+05, 4.2103E+05, 4.2253E+05,
     +  4.2404E+05, 4.2556E+05, 4.2708E+05, 4.2860E+05, 4.3013E+05,
     + 4.3166E+05, 4.3320E+05, 4.3474E+05, 4.3628E+05, 4.3783E+05,
     +  4.3939E+05, 4.4095E+05, 4.4251E+05, 4.4408E+05, 4.4565E+05,
     + 4.4723E+05, 4.4881E+05, 4.5040E+05, 4.5199E+05, 4.5358E+05,
     +  4.5518E+05, 4.5679E+05, 4.5839E+05, 4.6001E+05, 4.6163E+05,
     + 4.6325E+05, 4.6488E+05, 4.6651E+05, 4.6814E+05, 4.6978E+05,
     +  4.7143E+05, 4.7308E+05, 4.7474E+05, 4.7639E+05, 4.7806E+05,
     + 4.7973E+05, 4.8140E+05, 4.8308E+05, 4.8476E+05, 4.8645E+05,
     +  4.8814E+05, 4.8984E+05, 4.9154E+05, 4.9325E+05, 4.9496E+05,
     + 4.9668E+05, 4.9840E+05, 5.0013E+05, 5.0186E+05, 5.0359E+05,
     +  5.0534E+05, 5.0708E+05, 5.0883E+05, 5.1059E+05, 5.1235E+05,
     + 5.1411E+05, 5.1588E+05, 5.1766E+05, 5.1944E+05, 5.2122E+05,
     +  5.2301E+05, 5.2481E+05, 5.2661E+05, 5.2841E+05, 5.3022E+05,
     + 5.3204E+05, 5.3386E+05, 5.3568E+05, 5.3751E+05, 5.3935E+05,
     +  5.4119E+05, 5.4303E+05, 5.4488E+05, 5.4674E+05, 5.4860E+05,
     + 5.5046E+05, 5.5233E+05, 5.5421E+05, 5.5609E+05, 5.5798E+05,
     +  5.5987E+05, 5.6176E+05, 5.6367E+05, 5.6557E+05, 5.6749E+05,
     + 5.6940E+05, 5.7133E+05, 5.7325E+05, 5.7519E+05, 5.7712E+05,
     +  5.7907E+05, 5.8102E+05, 5.8297E+05, 5.8493E+05, 5.8690E+05,
     + 5.8887E+05, 5.9084E+05, 5.9282E+05, 5.9481E+05, 5.9680E+05,
     +  5.9880E+05, 6.0080E+05, 6.0281E+05, 6.0482E+05, 6.0684E+05,
     + 6.0886E+05, 6.1089E+05, 6.1293E+05, 6.1497E+05, 6.1701E+05,
     +  6.1906E+05, 6.2112E+05, 6.2318E+05, 6.2525E+05, 6.2733E+05,
     + 6.2940E+05, 6.3149E+05, 6.3358E+05, 6.3568E+05, 6.3778E+05,
     +  6.3988E+05, 6.4200E+05, 6.4411E+05, 6.4624E+05, 6.4837E+05,
     + 6.5050E+05, 6.5264E+05, 6.5479E+05, 6.5694E+05, 6.5910E+05,
     +  6.6127E+05, 6.6343E+05, 6.6561E+05, 6.6779E+05, 6.6998E+05,
     + 6.7217E+05, 6.7437E+05, 6.7657E+05, 6.7878E+05, 6.8100E+05,
     +  6.8322E+05, 6.8545E+05, 6.8768E+05, 6.8992E+05, 6.9217E+05,
     + 6.9442E+05, 6.9667E+05, 6.9894E+05, 7.0121E+05, 7.0348E+05,
     +  7.0576E+05, 7.0805E+05, 7.1034E+05, 7.1264E+05, 7.1494E+05,
     + 7.1726E+05, 7.1957E+05, 7.2189E+05, 7.2422E+05, 7.2656E+05,
     +  7.2890E+05, 7.3125E+05, 7.3360E+05, 7.3596E+05, 7.3832E+05,
     + 7.4070E+05, 7.4307E+05, 7.4546E+05, 7.4785E+05, 7.5024E+05,
     +  7.5265E+05, 7.5506E+05, 7.5747E+05, 7.5989E+05, 7.6232E+05,
     + 7.6475E+05, 7.6719E+05, 7.6964E+05, 7.7209E+05, 7.7455E+05,
     +  7.7702E+05, 7.7949E+05, 7.8197E+05, 7.8445E+05, 7.8694E+05,
     + 7.8944E+05, 7.9194E+05, 7.9445E+05, 7.9697E+05, 7.9949E+05,
     +  8.0202E+05, 8.0455E+05, 8.0710E+05, 8.0964E+05, 8.1220E+05,
     + 8.1476E+05, 8.1733E+05, 8.1990E+05, 8.2248E+05, 8.2507E+05,
     +  8.2767E+05, 8.3027E+05, 8.3287E+05, 8.3549E+05, 8.3811E+05,
     + 8.4074E+05, 8.4337E+05, 8.4601E+05, 8.4866E+05, 8.5131E+05,
     +  8.5397E+05, 8.5664E+05, 8.5931E+05, 8.6200E+05, 8.6468E+05,
     + 8.6738E+05, 8.7008E+05, 8.7279E+05, 8.7550E+05, 8.7822E+05,
     +  8.8095E+05, 8.8369E+05, 8.8643E+05, 8.8918E+05, 8.9193E+05,
     + 8.9470E+05, 8.9747E+05, 9.0024E+05, 9.0303E+05, 9.0582E+05,
     +  9.0861E+05, 9.1142E+05, 9.1423E+05, 9.1705E+05, 9.1987E+05,
     + 9.2271E+05, 9.2554E+05, 9.2839E+05, 9.3124E+05, 9.3411E+05,
     +  9.3697E+05, 9.3985E+05, 9.4273E+05, 9.4562E+05, 9.4852E+05,
     + 9.5142E+05, 9.5433E+05, 9.5725E+05, 9.6017E+05, 9.6310E+05,
     +  9.6604E+05, 9.6899E+05, 9.7194E+05, 9.7491E+05, 9.7787E+05,
     + 9.8085E+05, 9.8383E+05, 9.8682E+05, 9.8982E+05, 9.9283E+05,
     +  9.9584E+05, 9.9886E+05, 1.0019E+06, 1.0049E+06, 1.0080E+06,
     + 1.0110E+06, 1.0141E+06, 1.0171E+06, 1.0202E+06, 1.0233E+06,
     +  1.0264E+06, 1.0295E+06, 1.0326E+06, 1.0357E+06, 1.0388E+06,
     + 1.0419E+06, 1.0451E+06, 1.0482E+06, 1.0513E+06, 1.0545E+06,
     +  1.0577E+06, 1.0608E+06, 1.0640E+06, 1.0672E+06, 1.0704E+06,
     + 1.0736E+06, 1.0768E+06, 1.0800E+06, 1.0832E+06, 1.0865E+06,
     +  1.0897E+06, 1.0930E+06, 1.0962E+06, 1.0995E+06, 1.1027E+06,
     + 1.1060E+06, 1.1093E+06, 1.1126E+06, 1.1159E+06, 1.1192E+06,
     +  1.1225E+06, 1.1259E+06, 1.1292E+06, 1.1325E+06, 1.1359E+06,
     + 1.1392E+06, 1.1426E+06, 1.1460E+06, 1.1494E+06, 1.1528E+06,
     +  1.1562E+06, 1.1596E+06, 1.1630E+06, 1.1664E+06, 1.1698E+06,
     + 1.1733E+06, 1.1767E+06, 1.1802E+06, 1.1836E+06, 1.1871E+06,
     +  1.1906E+06, 1.1941E+06, 1.1976E+06, 1.2011E+06, 1.2046E+06,
     + 1.2081E+06, 1.2116E+06, 1.2152E+06, 1.2187E+06, 1.2223E+06,
     +  1.2258E+06, 1.2294E+06, 1.2330E+06, 1.2366E+06, 1.2402E+06,
     + 1.2438E+06, 1.2474E+06, 1.2510E+06, 1.2546E+06, 1.2583E+06,
     +  1.2619E+06, 1.2656E+06, 1.2693E+06, 1.2729E+06, 1.2766E+06,
     + 1.2803E+06, 1.2840E+06, 1.2877E+06, 1.2914E+06, 1.2951E+06,
     +  1.2989E+06, 1.3026E+06, 1.3064E+06, 1.3101E+06, 1.3139E+06,
     + 1.3177E+06, 1.3215E+06, 1.3253E+06, 1.3291E+06, 1.3329E+06,
     +  1.3367E+06, 1.3405E+06, 1.3444E+06, 1.3482E+06, 1.3521E+06,
     + 1.3559E+06, 1.3598E+06, 1.3637E+06, 1.3676E+06, 1.3715E+06,
     +  1.3754E+06, 1.3793E+06, 1.3832E+06, 1.3872E+06, 1.3911E+06,
     + 1.3951E+06, 1.3990E+06, 1.4030E+06, 1.4070E+06, 1.4110E+06,
     +  1.4150E+06, 1.4190E+06, 1.4230E+06, 1.4270E+06, 1.4311E+06,
     + 1.4351E+06, 1.4392E+06, 1.4433E+06, 1.4473E+06, 1.4514E+06,
     +  1.4555E+06, 1.4596E+06, 1.4637E+06, 1.4678E+06, 1.4720E+06,
     + 1.4761E+06, 1.4803E+06, 1.4844E+06, 1.4886E+06, 1.4928E+06,
     +  1.4969E+06, 1.5011E+06, 1.5053E+06, 1.5096E+06, 1.5138E+06,
     + 1.5180E+06, 1.5223E+06, 1.5265E+06, 1.5308E+06, 1.5351E+06,
     +  1.5393E+06, 1.5436E+06, 1.5479E+06, 1.5522E+06, 1.5566E+06,
     + 1.5609E+06, 1.5652E+06, 1.5696E+06, 1.5739E+06, 1.5783E+06,
     +  1.5827E+06, 1.5871E+06, 1.5915E+06, 1.5959E+06, 1.6003E+06,
     + 1.6047E+06, 1.6092E+06, 1.6136E+06, 1.6181E+06, 1.6225E+06,
     +  1.6270E+06, 1.6315E+06, 1.6360E+06, 1.6405E+06, 1.6450E+06,
     + 1.6496E+06, 1.6541E+06, 1.6586E+06, 1.6632E+06, 1.6678E+06,
     +  1.6724E+06, 1.6769E+06, 1.6815E+06, 1.6861E+06, 1.6908E+06,
     + 1.6954E+06, 1.7000E+06, 1.7047E+06, 1.7093E+06, 1.7140E+06,
     +  1.7187E+06, 1.7234E+06, 1.7281E+06, 1.7328E+06, 1.7375E+06,
     + 1.7423E+06, 1.7470E+06, 1.7518E+06, 1.7565E+06, 1.7613E+06,
     +  1.7661E+06, 1.7709E+06, 1.7757E+06, 1.7805E+06, 1.7853E+06,
     + 1.7902E+06, 1.7950E+06, 1.7999E+06, 1.8047E+06, 1.8096E+06,
     +  1.8145E+06, 1.8194E+06, 1.8243E+06, 1.8292E+06, 1.8342E+06,
     + 1.8391E+06, 1.8441E+06, 1.8490E+06, 1.8540E+06, 1.8590E+06,
     +  1.8640E+06, 1.8690E+06, 1.8740E+06, 1.8790E+06, 1.8841E+06,
     + 1.8891E+06, 1.8942E+06, 1.8993E+06, 1.9044E+06, 1.9095E+06,
     +  1.9146E+06, 1.9197E+06, 1.9248E+06, 1.9299E+06, 1.9351E+06,
     + 1.9403E+06, 1.9454E+06, 1.9506E+06, 1.9558E+06, 1.9610E+06,
     +  1.9662E+06, 1.9715E+06, 1.9767E+06, 1.9819E+06, 1.9872E+06,
     + 1.9925E+06, 1.9978E+06, 2.0031E+06, 2.0084E+06, 2.0137E+06,
     +  2.0190E+06, 2.0244E+06, 2.0297E+06, 2.0351E+06, 2.0405E+06,
     + 2.0458E+06, 2.0512E+06, 2.0567E+06, 2.0621E+06, 2.0675E+06,
     +  2.0730E+06, 2.0784E+06, 2.0839E+06, 2.0894E+06, 2.0948E+06,
     + 2.1004E+06, 2.1059E+06, 2.1114E+06, 2.1169E+06, 2.1225E+06,
     +  2.1280E+06, 2.1336E+06, 2.1392E+06, 2.1448E+06, 2.1504E+06,
     + 2.1560E+06, 2.1617E+06, 2.1673E+06, 2.1730E+06, 2.1786E+06,
     +  2.1843E+06, 2.1900E+06, 2.1957E+06, 2.2014E+06, 2.2071E+06,
     + 2.2129E+06, 2.2186E+06, 2.2244E+06, 2.2302E+06, 2.2360E+06,
     +  2.2418E+06, 2.2476E+06, 2.2534E+06, 2.2592E+06, 2.2651E+06,
     + 2.2709E+06, 2.2768E+06, 2.2827E+06, 2.2886E+06, 2.2945E+06,
     +  2.3004E+06, 2.3064E+06, 2.3123E+06, 2.3183E+06, 2.3242E+06,
     + 2.3302E+06, 2.3362E+06, 2.3422E+06, 2.3482E+06, 2.3543E+06,
     +  2.3603E+06, 2.3664E+06, 2.3725E+06, 2.3785E+06, 2.3846E+06,
     + 2.3907E+06, 2.3969E+06, 2.4030E+06, 2.4091E+06, 2.4153E+06,
     +  2.4215E+06, 2.4277E+06, 2.4339E+06, 2.4401E+06, 2.4463E+06,
     + 2.4525E+06, 2.4588E+06, 2.4650E+06, 2.4713E+06, 2.4776E+06,
     +  2.4839E+06, 2.4902E+06, 2.4965E+06, 2.5029E+06, 2.5092E+06,
     + 2.5156E+06, 2.5220E+06, 2.5284E+06, 2.5348E+06, 2.5412E+06,
     +  2.5476E+06, 2.5541E+06, 2.5605E+06, 2.5670E+06, 2.5735E+06,
     + 2.5800E+06, 2.5865E+06, 2.5930E+06, 2.5995E+06, 2.6061E+06,
     +  2.6126E+06, 2.6192E+06, 2.6258E+06, 2.6324E+06, 2.6390E+06,
     + 2.6457E+06, 2.6523E+06, 2.6590E+06, 2.6656E+06, 2.6723E+06,
     +  2.6790E+06, 2.6857E+06, 2.6924E+06, 2.6992E+06, 2.7059E+06,
     + 2.7127E+06, 2.7195E+06, 2.7263E+06, 2.7331E+06, 2.7399E+06,
     +  2.7467E+06, 2.7536E+06, 2.7604E+06, 2.7673E+06, 2.7742E+06,
     + 2.7811E+06, 2.7880E+06, 2.7949E+06, 2.8019E+06, 2.8088E+06,
     +  2.8158E+06, 2.8228E+06, 2.8298E+06, 2.8368E+06, 2.8438E+06,
     + 2.8509E+06, 2.8579E+06, 2.8650E+06, 2.8721E+06, 2.8792E+06,
     +  2.8863E+06, 2.8934E+06, 2.9005E+06, 2.9077E+06, 2.9149E+06,
     + 2.9221E+06, 2.9292E+06, 2.9365E+06, 2.9437E+06, 2.9509E+06,
     +  2.9582E+06, 2.9654E+06, 2.9727E+06, 2.9800E+06, 2.9873E+06,
     + 2.9947E+06, 3.0020E+06, 3.0094E+06, 3.0167E+06, 3.0241E+06,
     +  3.0315E+06, 3.0389E+06, 3.0464E+06, 3.0538E+06, 3.0613E+06,
     + 3.0687E+06, 3.0762E+06, 3.0837E+06, 3.0912E+06, 3.0988E+06,
     +  3.1063E+06, 3.1139E+06, 3.1214E+06, 3.1290E+06, 3.1366E+06,
     + 3.1443E+06, 3.1519E+06, 3.1595E+06, 3.1672E+06, 3.1749E+06,
     +  3.1826E+06, 3.1903E+06, 3.1980E+06, 3.2058E+06, 3.2135E+06,
     + 3.2213E+06, 3.2291E+06, 3.2369E+06, 3.2447E+06, 3.2525E+06,
     +  3.2604E+06, 3.2682E+06, 3.2761E+06, 3.2840E+06, 3.2919E+06,
     + 3.2998E+06, 3.3078E+06, 3.3157E+06, 3.3237E+06, 3.3317E+06,
     +  3.3397E+06, 3.3477E+06, 3.3557E+06, 3.3637E+06, 3.3718E+06,
     + 3.3799E+06, 3.3880E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 3,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       448
      data (QofT( 4,J),J=1,2408)/ 1.9200E+01, 3.4811E+01, 5.0564E+01,
     +  6.6349E+01, 8.2147E+01, 9.7952E+01, 1.1376E+02, 1.2957E+02,
     + 1.4538E+02, 1.6120E+02, 1.7701E+02, 1.9283E+02, 2.0865E+02,
     +  2.2446E+02, 2.4028E+02, 2.5610E+02, 2.7192E+02, 2.8774E+02,
     + 3.0356E+02, 3.1938E+02, 3.5102E+02, 3.8267E+02, 4.1432E+02,
     +  4.4596E+02, 4.7761E+02, 5.0926E+02, 5.4092E+02, 5.7257E+02,
     + 6.0422E+02, 6.3588E+02, 6.6754E+02, 6.9920E+02, 7.3086E+02,
     +  7.6252E+02, 7.9419E+02, 8.2585E+02, 8.5752E+02, 8.8919E+02,
     + 9.2086E+02, 9.5253E+02, 9.8421E+02, 1.0159E+03, 1.0476E+03,
     +  1.0792E+03, 1.1109E+03, 1.1426E+03, 1.1743E+03, 1.2060E+03,
     + 1.2377E+03, 1.2694E+03, 1.3011E+03, 1.3328E+03, 1.3645E+03,
     +  1.3963E+03, 1.4280E+03, 1.4598E+03, 1.4916E+03, 1.5234E+03,
     + 1.5552E+03, 1.5870E+03, 1.6188E+03, 1.6507E+03, 1.6826E+03,
     +  1.7145E+03, 1.7465E+03, 1.7785E+03, 1.8105E+03, 1.8426E+03,
     + 1.8747E+03, 1.9068E+03, 1.9390E+03, 1.9713E+03, 2.0036E+03,
     +  2.0360E+03, 2.0684E+03, 2.1009E+03, 2.1335E+03, 2.1661E+03,
     + 2.1988E+03, 2.2316E+03, 2.2645E+03, 2.2975E+03, 2.3305E+03,
     +  2.3637E+03, 2.3969E+03, 2.4303E+03, 2.4637E+03, 2.4973E+03,
     + 2.5310E+03, 2.5648E+03, 2.5987E+03, 2.6328E+03, 2.6669E+03,
     +  2.7012E+03, 2.7357E+03, 2.7703E+03, 2.8050E+03, 2.8398E+03,
     + 2.8749E+03, 2.9100E+03, 2.9453E+03, 2.9808E+03, 3.0164E+03,
     +  3.0523E+03, 3.0882E+03, 3.1244E+03, 3.1607E+03, 3.1972E+03,
     + 3.2339E+03, 3.2708E+03, 3.3079E+03, 3.3451E+03, 3.3826E+03,
     +  3.4202E+03, 3.4580E+03, 3.4961E+03, 3.5344E+03, 3.5728E+03,
     + 3.6115E+03, 3.6504E+03, 3.6895E+03, 3.7289E+03, 3.7684E+03,
     +  3.8082E+03, 3.8482E+03, 3.8885E+03, 3.9290E+03, 3.9697E+03,
     + 4.0106E+03, 4.0519E+03, 4.0933E+03, 4.1350E+03, 4.1770E+03,
     +  4.2192E+03, 4.2616E+03, 4.3044E+03, 4.3473E+03, 4.3906E+03,
     + 4.4341E+03, 4.4779E+03, 4.5219E+03, 4.5662E+03, 4.6109E+03,
     +  4.6558E+03, 4.7009E+03, 4.7464E+03, 4.7921E+03, 4.8381E+03,
     + 4.8844E+03, 4.9310E+03, 4.9779E+03, 5.0251E+03, 5.0726E+03,
     +  5.1204E+03, 5.1685E+03, 5.2169E+03, 5.2657E+03, 5.3147E+03,
     + 5.3641E+03, 5.4138E+03, 5.4637E+03, 5.5141E+03, 5.5647E+03,
     +  5.6156E+03, 5.6669E+03, 5.7185E+03, 5.7705E+03, 5.8228E+03,
     + 5.8754E+03, 5.9284E+03, 5.9817E+03, 6.0353E+03, 6.0893E+03,
     +  6.1436E+03, 6.1984E+03, 6.2534E+03, 6.3088E+03, 6.3646E+03,
     + 6.4207E+03, 6.4772E+03, 6.5340E+03, 6.5912E+03, 6.6489E+03,
     +  6.7068E+03, 6.7651E+03, 6.8238E+03, 6.8829E+03, 6.9424E+03,
     + 7.0022E+03, 7.0625E+03, 7.1231E+03, 7.1841E+03, 7.2455E+03,
     +  7.3073E+03, 7.3695E+03, 7.4321E+03, 7.4951E+03, 7.5585E+03,
     + 7.6223E+03, 7.6865E+03, 7.7511E+03, 7.8162E+03, 7.8816E+03,
     +  7.9475E+03, 8.0138E+03, 8.0805E+03, 8.1476E+03, 8.2152E+03,
     + 8.2832E+03, 8.3516E+03, 8.4205E+03, 8.4898E+03, 8.5595E+03,
     +  8.6297E+03, 8.7003E+03, 8.7714E+03, 8.8429E+03, 8.9149E+03,
     + 8.9873E+03, 9.0602E+03, 9.1335E+03, 9.2073E+03, 9.2816E+03,
     +  9.3563E+03, 9.4315E+03, 9.5072E+03, 9.5833E+03, 9.6599E+03,
     + 9.7370E+03, 9.8146E+03, 9.8926E+03, 9.9711E+03, 1.0050E+04,
     +  1.0130E+04, 1.0210E+04, 1.0290E+04, 1.0371E+04, 1.0453E+04,
     + 1.0535E+04, 1.0617E+04, 1.0700E+04, 1.0784E+04, 1.0868E+04,
     +  1.0952E+04, 1.1037E+04, 1.1123E+04, 1.1209E+04, 1.1296E+04,
     + 1.1383E+04, 1.1471E+04, 1.1559E+04, 1.1648E+04, 1.1737E+04,
     +  1.1827E+04, 1.1917E+04, 1.2008E+04, 1.2100E+04, 1.2192E+04,
     + 1.2285E+04, 1.2378E+04, 1.2471E+04, 1.2566E+04, 1.2661E+04,
     +  1.2756E+04, 1.2852E+04, 1.2949E+04, 1.3046E+04, 1.3144E+04,
     + 1.3242E+04, 1.3341E+04, 1.3440E+04, 1.3540E+04, 1.3641E+04,
     +  1.3742E+04, 1.3844E+04, 1.3947E+04, 1.4050E+04, 1.4154E+04,
     + 1.4258E+04, 1.4363E+04, 1.4468E+04, 1.4575E+04, 1.4681E+04,
     +  1.4789E+04, 1.4897E+04, 1.5005E+04, 1.5115E+04, 1.5225E+04,
     + 1.5335E+04, 1.5447E+04, 1.5558E+04, 1.5671E+04, 1.5784E+04,
     +  1.5898E+04, 1.6013E+04, 1.6128E+04, 1.6244E+04, 1.6360E+04,
     + 1.6477E+04, 1.6595E+04, 1.6714E+04, 1.6833E+04, 1.6953E+04,
     +  1.7073E+04, 1.7194E+04, 1.7317E+04, 1.7439E+04, 1.7563E+04,
     + 1.7687E+04, 1.7811E+04, 1.7937E+04, 1.8063E+04, 1.8190E+04,
     +  1.8317E+04, 1.8446E+04, 1.8575E+04, 1.8705E+04, 1.8835E+04,
     + 1.8966E+04, 1.9098E+04, 1.9231E+04, 1.9365E+04, 1.9499E+04,
     +  1.9634E+04, 1.9769E+04, 1.9906E+04, 2.0043E+04, 2.0181E+04,
     + 2.0320E+04, 2.0460E+04, 2.0600E+04, 2.0741E+04, 2.0883E+04,
     +  2.1026E+04, 2.1169E+04, 2.1313E+04, 2.1458E+04, 2.1604E+04,
     + 2.1751E+04, 2.1898E+04, 2.2046E+04, 2.2196E+04, 2.2345E+04,
     +  2.2496E+04, 2.2648E+04, 2.2800E+04, 2.2953E+04, 2.3107E+04,
     + 2.3262E+04, 2.3418E+04, 2.3574E+04, 2.3732E+04, 2.3890E+04,
     +  2.4049E+04, 2.4209E+04, 2.4370E+04, 2.4531E+04, 2.4694E+04,
     + 2.4857E+04, 2.5022E+04, 2.5187E+04, 2.5353E+04, 2.5520E+04,
     +  2.5688E+04, 2.5856E+04, 2.6026E+04, 2.6196E+04, 2.6368E+04,
     + 2.6540E+04, 2.6714E+04, 2.6888E+04, 2.7063E+04, 2.7239E+04,
     +  2.7416E+04, 2.7594E+04, 2.7773E+04, 2.7952E+04, 2.8133E+04,
     + 2.8315E+04, 2.8497E+04, 2.8681E+04, 2.8865E+04, 2.9051E+04,
     +  2.9237E+04, 2.9425E+04, 2.9613E+04, 2.9802E+04, 2.9993E+04,
     + 3.0184E+04, 3.0377E+04, 3.0570E+04, 3.0764E+04, 3.0960E+04,
     +  3.1156E+04, 3.1353E+04, 3.1552E+04, 3.1751E+04, 3.1952E+04,
     + 3.2153E+04, 3.2356E+04, 3.2559E+04, 3.2764E+04, 3.2969E+04,
     +  3.3176E+04, 3.3384E+04, 3.3592E+04, 3.3802E+04, 3.4013E+04,
     + 3.4225E+04, 3.4438E+04, 3.4652E+04, 3.4867E+04, 3.5084E+04,
     +  3.5301E+04, 3.5520E+04, 3.5739E+04, 3.5960E+04, 3.6182E+04,
     + 3.6405E+04, 3.6629E+04, 3.6854E+04, 3.7080E+04, 3.7307E+04,
     +  3.7536E+04, 3.7766E+04, 3.7996E+04, 3.8228E+04, 3.8462E+04,
     + 3.8696E+04, 3.8931E+04, 3.9168E+04, 3.9406E+04, 3.9645E+04,
     +  3.9885E+04, 4.0126E+04, 4.0369E+04, 4.0612E+04, 4.0857E+04,
     + 4.1103E+04, 4.1351E+04, 4.1599E+04, 4.1849E+04, 4.2100E+04,
     +  4.2352E+04, 4.2605E+04, 4.2860E+04, 4.3116E+04, 4.3373E+04,
     + 4.3631E+04, 4.3891E+04, 4.4152E+04, 4.4414E+04, 4.4677E+04,
     +  4.4942E+04, 4.5208E+04, 4.5475E+04, 4.5744E+04, 4.6014E+04,
     + 4.6285E+04, 4.6557E+04, 4.6831E+04, 4.7106E+04, 4.7382E+04,
     +  4.7660E+04, 4.7939E+04, 4.8219E+04, 4.8501E+04, 4.8784E+04,
     + 4.9068E+04, 4.9354E+04, 4.9641E+04, 4.9929E+04, 5.0219E+04,
     +  5.0510E+04, 5.0803E+04, 5.1097E+04, 5.1392E+04, 5.1689E+04,
     + 5.1987E+04, 5.2286E+04, 5.2587E+04, 5.2889E+04, 5.3193E+04,
     +  5.3498E+04, 5.3805E+04, 5.4112E+04, 5.4422E+04, 5.4733E+04,
     + 5.5045E+04, 5.5359E+04, 5.5674E+04, 5.5991E+04, 5.6309E+04,
     +  5.6628E+04, 5.6949E+04, 5.7272E+04, 5.7596E+04, 5.7921E+04,
     + 5.8249E+04, 5.8577E+04, 5.8907E+04, 5.9239E+04, 5.9572E+04,
     +  5.9906E+04, 6.0242E+04, 6.0580E+04, 6.0919E+04, 6.1260E+04,
     + 6.1602E+04, 6.1946E+04, 6.2292E+04, 6.2639E+04, 6.2987E+04,
     +  6.3337E+04, 6.3689E+04, 6.4043E+04, 6.4398E+04, 6.4754E+04,
     + 6.5112E+04, 6.5472E+04, 6.5833E+04, 6.6196E+04, 6.6561E+04,
     +  6.6927E+04, 6.7295E+04, 6.7665E+04, 6.8036E+04, 6.8409E+04,
     + 6.8783E+04, 6.9159E+04, 6.9537E+04, 6.9917E+04, 7.0298E+04,
     +  7.0681E+04, 7.1066E+04, 7.1452E+04, 7.1840E+04, 7.2230E+04,
     + 7.2621E+04, 7.3015E+04, 7.3410E+04, 7.3806E+04, 7.4205E+04,
     +  7.4605E+04, 7.5007E+04, 7.5410E+04, 7.5816E+04, 7.6223E+04,
     + 7.6632E+04, 7.7043E+04, 7.7456E+04, 7.7870E+04, 7.8286E+04,
     +  7.8704E+04, 7.9124E+04, 7.9546E+04, 7.9969E+04, 8.0395E+04,
     + 8.0822E+04, 8.1251E+04, 8.1682E+04, 8.2115E+04, 8.2549E+04,
     +  8.2986E+04, 8.3424E+04, 8.3864E+04, 8.4306E+04, 8.4750E+04,
     + 8.5196E+04, 8.5644E+04, 8.6094E+04, 8.6545E+04, 8.6999E+04,
     +  8.7455E+04, 8.7912E+04, 8.8372E+04, 8.8833E+04, 8.9296E+04,
     + 8.9762E+04, 9.0229E+04, 9.0698E+04, 9.1169E+04, 9.1643E+04,
     +  9.2118E+04, 9.2595E+04, 9.3074E+04, 9.3556E+04, 9.4039E+04,
     + 9.4524E+04, 9.5012E+04, 9.5501E+04, 9.5993E+04, 9.6486E+04,
     +  9.6982E+04, 9.7479E+04, 9.7979E+04, 9.8481E+04, 9.8985E+04,
     + 9.9491E+04, 9.9999E+04, 1.0051E+05, 1.0102E+05, 1.0154E+05,
     +  1.0205E+05, 1.0257E+05, 1.0309E+05, 1.0362E+05, 1.0414E+05,
     + 1.0467E+05, 1.0520E+05, 1.0573E+05, 1.0626E+05, 1.0680E+05,
     +  1.0734E+05, 1.0788E+05, 1.0842E+05, 1.0897E+05, 1.0951E+05,
     + 1.1006E+05, 1.1061E+05, 1.1117E+05, 1.1172E+05, 1.1228E+05,
     +  1.1284E+05, 1.1341E+05, 1.1397E+05, 1.1454E+05, 1.1511E+05,
     + 1.1568E+05, 1.1626E+05, 1.1683E+05, 1.1741E+05, 1.1799E+05,
     +  1.1858E+05, 1.1916E+05, 1.1975E+05, 1.2034E+05, 1.2094E+05,
     + 1.2153E+05, 1.2213E+05, 1.2273E+05, 1.2333E+05, 1.2394E+05,
     +  1.2455E+05, 1.2516E+05, 1.2577E+05, 1.2638E+05, 1.2700E+05,
     + 1.2762E+05, 1.2824E+05, 1.2887E+05, 1.2949E+05, 1.3012E+05,
     +  1.3076E+05, 1.3139E+05, 1.3203E+05, 1.3267E+05, 1.3331E+05,
     + 1.3395E+05, 1.3460E+05, 1.3525E+05, 1.3590E+05, 1.3656E+05,
     +  1.3722E+05, 1.3788E+05, 1.3854E+05, 1.3920E+05, 1.3987E+05,
     + 1.4054E+05, 1.4122E+05, 1.4189E+05, 1.4257E+05, 1.4325E+05,
     +  1.4393E+05, 1.4462E+05, 1.4531E+05, 1.4600E+05, 1.4669E+05,
     + 1.4739E+05, 1.4809E+05, 1.4879E+05, 1.4950E+05, 1.5021E+05,
     +  1.5092E+05, 1.5163E+05, 1.5235E+05, 1.5306E+05, 1.5379E+05,
     + 1.5451E+05, 1.5524E+05, 1.5597E+05, 1.5670E+05, 1.5743E+05,
     +  1.5817E+05, 1.5891E+05, 1.5966E+05, 1.6040E+05, 1.6115E+05,
     + 1.6190E+05, 1.6266E+05, 1.6342E+05, 1.6418E+05, 1.6494E+05,
     +  1.6571E+05, 1.6648E+05, 1.6725E+05, 1.6803E+05, 1.6880E+05,
     + 1.6959E+05, 1.7037E+05, 1.7116E+05, 1.7195E+05, 1.7274E+05,
     +  1.7354E+05, 1.7433E+05, 1.7514E+05, 1.7594E+05, 1.7675E+05,
     + 1.7756E+05, 1.7837E+05, 1.7919E+05, 1.8001E+05, 1.8083E+05,
     +  1.8166E+05, 1.8249E+05, 1.8332E+05, 1.8416E+05, 1.8499E+05,
     + 1.8584E+05, 1.8668E+05, 1.8753E+05, 1.8838E+05, 1.8923E+05,
     +  1.9009E+05, 1.9095E+05, 1.9181E+05, 1.9268E+05, 1.9355E+05,
     + 1.9442E+05, 1.9530E+05, 1.9618E+05, 1.9706E+05, 1.9795E+05,
     +  1.9884E+05, 1.9973E+05, 2.0062E+05, 2.0152E+05, 2.0242E+05,
     + 2.0333E+05, 2.0424E+05, 2.0515E+05, 2.0606E+05, 2.0698E+05,
     +  2.0790E+05, 2.0883E+05, 2.0976E+05, 2.1069E+05, 2.1163E+05,
     + 2.1256E+05, 2.1351E+05, 2.1445E+05, 2.1540E+05, 2.1635E+05,
     +  2.1731E+05, 2.1827E+05, 2.1923E+05, 2.2020E+05, 2.2116E+05,
     + 2.2214E+05, 2.2311E+05, 2.2409E+05, 2.2508E+05, 2.2606E+05,
     +  2.2705E+05, 2.2805E+05, 2.2904E+05, 2.3005E+05, 2.3105E+05,
     + 2.3206E+05, 2.3307E+05, 2.3408E+05, 2.3510E+05, 2.3612E+05,
     +  2.3715E+05, 2.3818E+05, 2.3921E+05, 2.4025E+05, 2.4129E+05,
     + 2.4233E+05, 2.4338E+05, 2.4443E+05, 2.4549E+05, 2.4655E+05,
     +  2.4761E+05, 2.4868E+05, 2.4975E+05, 2.5082E+05, 2.5190E+05,
     + 2.5298E+05, 2.5406E+05, 2.5515E+05, 2.5624E+05, 2.5734E+05,
     +  2.5844E+05, 2.5954E+05, 2.6065E+05, 2.6176E+05, 2.6288E+05,
     + 2.6400E+05, 2.6512E+05, 2.6625E+05, 2.6738E+05, 2.6851E+05,
     +  2.6965E+05, 2.7079E+05, 2.7194E+05, 2.7309E+05, 2.7424E+05,
     + 2.7540E+05, 2.7657E+05, 2.7773E+05, 2.7890E+05, 2.8008E+05,
     +  2.8125E+05, 2.8244E+05, 2.8362E+05, 2.8481E+05, 2.8601E+05,
     + 2.8721E+05, 2.8841E+05, 2.8961E+05, 2.9083E+05, 2.9204E+05,
     +  2.9326E+05, 2.9448E+05, 2.9571E+05, 2.9694E+05, 2.9818E+05,
     + 2.9942E+05, 3.0066E+05, 3.0191E+05, 3.0316E+05, 3.0442E+05,
     +  3.0568E+05, 3.0694E+05, 3.0821E+05, 3.0948E+05, 3.1076E+05,
     + 3.1204E+05, 3.1333E+05, 3.1462E+05, 3.1591E+05, 3.1721E+05,
     +  3.1852E+05, 3.1982E+05, 3.2114E+05, 3.2245E+05, 3.2377E+05,
     + 3.2510E+05, 3.2643E+05, 3.2776E+05, 3.2910E+05, 3.3044E+05,
     +  3.3179E+05, 3.3314E+05, 3.3450E+05, 3.3586E+05, 3.3722E+05,
     + 3.3859E+05, 3.3997E+05, 3.4135E+05, 3.4273E+05, 3.4412E+05,
     +  3.4551E+05, 3.4691E+05, 3.4831E+05, 3.4971E+05, 3.5112E+05,
     + 3.5254E+05, 3.5396E+05, 3.5538E+05, 3.5681E+05, 3.5825E+05,
     +  3.5969E+05, 3.6113E+05, 3.6258E+05, 3.6403E+05, 3.6549E+05,
     + 3.6695E+05, 3.6841E+05, 3.6989E+05, 3.7136E+05, 3.7284E+05,
     +  3.7433E+05, 3.7582E+05, 3.7731E+05, 3.7882E+05, 3.8032E+05,
     + 3.8183E+05, 3.8334E+05, 3.8486E+05, 3.8639E+05, 3.8792E+05,
     +  3.8945E+05, 3.9099E+05, 3.9254E+05, 3.9408E+05, 3.9564E+05,
     + 3.9720E+05, 3.9876E+05, 4.0033E+05, 4.0190E+05, 4.0348E+05,
     +  4.0507E+05, 4.0666E+05, 4.0825E+05, 4.0985E+05, 4.1145E+05,
     + 4.1306E+05, 4.1468E+05, 4.1630E+05, 4.1792E+05, 4.1955E+05,
     +  4.2118E+05, 4.2282E+05, 4.2447E+05, 4.2612E+05, 4.2777E+05,
     + 4.2944E+05, 4.3110E+05, 4.3277E+05, 4.3445E+05, 4.3613E+05,
     +  4.3782E+05, 4.3951E+05, 4.4121E+05, 4.4291E+05, 4.4462E+05,
     + 4.4633E+05, 4.4805E+05, 4.4977E+05, 4.5150E+05, 4.5324E+05,
     +  4.5498E+05, 4.5673E+05, 4.5848E+05, 4.6023E+05, 4.6199E+05,
     + 4.6376E+05, 4.6554E+05, 4.6731E+05, 4.6910E+05, 4.7089E+05,
     +  4.7268E+05, 4.7448E+05, 4.7629E+05, 4.7810E+05, 4.7992E+05,
     + 4.8174E+05, 4.8357E+05, 4.8540E+05, 4.8724E+05, 4.8909E+05,
     +  4.9094E+05, 4.9279E+05, 4.9466E+05, 4.9652E+05, 4.9840E+05,
     + 5.0028E+05, 5.0216E+05, 5.0405E+05, 5.0595E+05, 5.0785E+05,
     +  5.0976E+05, 5.1168E+05, 5.1360E+05, 5.1552E+05, 5.1745E+05,
     + 5.1939E+05, 5.2133E+05, 5.2328E+05, 5.2524E+05, 5.2720E+05,
     +  5.2916E+05, 5.3114E+05, 5.3312E+05, 5.3510E+05, 5.3709E+05,
     + 5.3909E+05, 5.4109E+05, 5.4310E+05, 5.4511E+05, 5.4713E+05,
     +  5.4916E+05, 5.5119E+05, 5.5323E+05, 5.5528E+05, 5.5733E+05,
     + 5.5939E+05, 5.6145E+05, 5.6352E+05, 5.6560E+05, 5.6768E+05,
     +  5.6977E+05, 5.7186E+05, 5.7396E+05, 5.7607E+05, 5.7818E+05,
     + 5.8030E+05, 5.8243E+05, 5.8456E+05, 5.8670E+05, 5.8884E+05,
     +  5.9099E+05, 5.9315E+05, 5.9531E+05, 5.9748E+05, 5.9966E+05,
     + 6.0184E+05, 6.0403E+05, 6.0623E+05, 6.0843E+05, 6.1064E+05,
     +  6.1285E+05, 6.1507E+05, 6.1730E+05, 6.1954E+05, 6.2178E+05,
     + 6.2403E+05, 6.2628E+05, 6.2854E+05, 6.3081E+05, 6.3308E+05,
     +  6.3536E+05, 6.3765E+05, 6.3994E+05, 6.4224E+05, 6.4455E+05,
     + 6.4687E+05, 6.4919E+05, 6.5152E+05, 6.5385E+05, 6.5619E+05,
     +  6.5854E+05, 6.6089E+05, 6.6325E+05, 6.6562E+05, 6.6800E+05,
     + 6.7038E+05, 6.7277E+05, 6.7516E+05, 6.7757E+05, 6.7998E+05,
     +  6.8239E+05, 6.8482E+05, 6.8725E+05, 6.8968E+05, 6.9213E+05,
     + 6.9458E+05, 6.9704E+05, 6.9950E+05, 7.0198E+05, 7.0445E+05,
     +  7.0694E+05, 7.0943E+05, 7.1194E+05, 7.1444E+05, 7.1696E+05,
     + 7.1948E+05, 7.2201E+05, 7.2455E+05, 7.2709E+05, 7.2964E+05,
     +  7.3220E+05, 7.3476E+05, 7.3734E+05, 7.3992E+05, 7.4250E+05,
     + 7.4510E+05, 7.4770E+05, 7.5031E+05, 7.5293E+05, 7.5555E+05,
     +  7.5818E+05, 7.6082E+05, 7.6347E+05, 7.6612E+05, 7.6878E+05,
     + 7.7145E+05, 7.7413E+05, 7.7681E+05, 7.7950E+05, 7.8220E+05,
     +  7.8491E+05, 7.8762E+05, 7.9034E+05, 7.9307E+05, 7.9581E+05,
     + 7.9855E+05, 8.0130E+05, 8.0406E+05, 8.0683E+05, 8.0960E+05,
     +  8.1239E+05, 8.1518E+05, 8.1798E+05, 8.2078E+05, 8.2359E+05,
     + 8.2642E+05, 8.2924E+05, 8.3208E+05, 8.3493E+05, 8.3778E+05,
     +  8.4064E+05, 8.4351E+05, 8.4638E+05, 8.4927E+05, 8.5216E+05,
     + 8.5506E+05, 8.5797E+05, 8.6089E+05, 8.6381E+05, 8.6674E+05,
     +  8.6968E+05, 8.7263E+05, 8.7559E+05, 8.7855E+05, 8.8152E+05,
     + 8.8450E+05, 8.8749E+05, 8.9049E+05, 8.9350E+05, 8.9651E+05,
     +  8.9953E+05, 9.0256E+05, 9.0560E+05, 9.0865E+05, 9.1170E+05,
     + 9.1476E+05, 9.1783E+05, 9.2091E+05, 9.2400E+05, 9.2710E+05,
     +  9.3020E+05, 9.3332E+05, 9.3644E+05, 9.3957E+05, 9.4270E+05,
     + 9.4585E+05, 9.4901E+05, 9.5217E+05, 9.5534E+05, 9.5852E+05,
     +  9.6171E+05, 9.6491E+05, 9.6812E+05, 9.7133E+05, 9.7456E+05,
     + 9.7779E+05, 9.8103E+05, 9.8428E+05, 9.8754E+05, 9.9081E+05,
     +  9.9408E+05, 9.9737E+05, 1.0007E+06, 1.0040E+06, 1.0073E+06,
     + 1.0106E+06, 1.0139E+06, 1.0173E+06, 1.0206E+06, 1.0240E+06,
     +  1.0273E+06, 1.0307E+06, 1.0341E+06, 1.0375E+06, 1.0409E+06,
     + 1.0443E+06, 1.0477E+06, 1.0511E+06, 1.0546E+06, 1.0580E+06,
     +  1.0615E+06, 1.0649E+06, 1.0684E+06, 1.0719E+06, 1.0754E+06,
     + 1.0789E+06, 1.0824E+06, 1.0859E+06, 1.0894E+06, 1.0930E+06,
     +  1.0965E+06, 1.1001E+06, 1.1036E+06, 1.1072E+06, 1.1108E+06,
     + 1.1144E+06, 1.1180E+06, 1.1216E+06, 1.1252E+06, 1.1288E+06,
     +  1.1325E+06, 1.1361E+06, 1.1398E+06, 1.1435E+06, 1.1471E+06,
     + 1.1508E+06, 1.1545E+06, 1.1582E+06, 1.1619E+06, 1.1657E+06,
     +  1.1694E+06, 1.1732E+06, 1.1769E+06, 1.1807E+06, 1.1844E+06,
     + 1.1882E+06, 1.1920E+06, 1.1958E+06, 1.1996E+06, 1.2035E+06,
     +  1.2073E+06, 1.2111E+06, 1.2150E+06, 1.2189E+06, 1.2227E+06,
     + 1.2266E+06, 1.2305E+06, 1.2344E+06, 1.2383E+06, 1.2422E+06,
     +  1.2462E+06, 1.2501E+06, 1.2541E+06, 1.2580E+06, 1.2620E+06,
     + 1.2660E+06, 1.2700E+06, 1.2740E+06, 1.2780E+06, 1.2820E+06,
     +  1.2861E+06, 1.2901E+06, 1.2942E+06, 1.2982E+06, 1.3023E+06,
     + 1.3064E+06, 1.3105E+06, 1.3146E+06, 1.3187E+06, 1.3228E+06,
     +  1.3270E+06, 1.3311E+06, 1.3353E+06, 1.3394E+06, 1.3436E+06,
     + 1.3478E+06, 1.3520E+06, 1.3562E+06, 1.3604E+06, 1.3647E+06,
     +  1.3689E+06, 1.3732E+06, 1.3774E+06, 1.3817E+06, 1.3860E+06,
     + 1.3903E+06, 1.3946E+06, 1.3989E+06, 1.4032E+06, 1.4076E+06,
     +  1.4119E+06, 1.4163E+06, 1.4207E+06, 1.4250E+06, 1.4294E+06,
     + 1.4338E+06, 1.4383E+06, 1.4427E+06, 1.4471E+06, 1.4516E+06,
     +  1.4560E+06, 1.4605E+06, 1.4650E+06, 1.4695E+06, 1.4740E+06,
     + 1.4785E+06, 1.4830E+06, 1.4876E+06, 1.4921E+06, 1.4967E+06,
     +  1.5012E+06, 1.5058E+06, 1.5104E+06, 1.5150E+06, 1.5196E+06,
     + 1.5243E+06, 1.5289E+06, 1.5335E+06, 1.5382E+06, 1.5429E+06,
     +  1.5476E+06, 1.5523E+06, 1.5570E+06, 1.5617E+06, 1.5664E+06,
     + 1.5712E+06, 1.5759E+06, 1.5807E+06, 1.5855E+06, 1.5902E+06,
     +  1.5950E+06, 1.5999E+06, 1.6047E+06, 1.6095E+06, 1.6144E+06,
     + 1.6192E+06, 1.6241E+06, 1.6290E+06, 1.6339E+06, 1.6388E+06,
     +  1.6437E+06, 1.6486E+06, 1.6536E+06, 1.6585E+06, 1.6635E+06,
     + 1.6685E+06, 1.6735E+06, 1.6785E+06, 1.6835E+06, 1.6885E+06,
     +  1.6935E+06, 1.6986E+06, 1.7036E+06, 1.7087E+06, 1.7138E+06,
     + 1.7189E+06, 1.7240E+06, 1.7291E+06, 1.7343E+06, 1.7394E+06,
     +  1.7446E+06, 1.7498E+06, 1.7549E+06, 1.7601E+06, 1.7654E+06,
     + 1.7706E+06, 1.7758E+06, 1.7811E+06, 1.7863E+06, 1.7916E+06,
     +  1.7969E+06, 1.8022E+06, 1.8075E+06, 1.8128E+06, 1.8181E+06,
     + 1.8235E+06, 1.8289E+06, 1.8342E+06, 1.8396E+06, 1.8450E+06,
     +  1.8504E+06, 1.8559E+06, 1.8613E+06, 1.8667E+06, 1.8722E+06,
     + 1.8777E+06, 1.8832E+06, 1.8887E+06, 1.8942E+06, 1.8997E+06,
     +  1.9053E+06, 1.9108E+06, 1.9164E+06, 1.9220E+06, 1.9275E+06,
     + 1.9332E+06, 1.9388E+06, 1.9444E+06, 1.9501E+06, 1.9557E+06,
     +  1.9614E+06, 1.9671E+06, 1.9728E+06, 1.9785E+06, 1.9842E+06,
     + 1.9899E+06, 1.9957E+06, 2.0015E+06, 2.0072E+06, 2.0130E+06,
     +  2.0188E+06, 2.0247E+06, 2.0305E+06, 2.0363E+06, 2.0422E+06,
     + 2.0481E+06, 2.0540E+06, 2.0599E+06, 2.0658E+06, 2.0717E+06,
     +  2.0777E+06, 2.0836E+06, 2.0896E+06, 2.0956E+06, 2.1016E+06,
     + 2.1076E+06, 2.1136E+06, 2.1196E+06, 2.1257E+06, 2.1318E+06,
     +  2.1378E+06, 2.1439E+06, 2.1500E+06, 2.1562E+06, 2.1623E+06,
     + 2.1685E+06, 2.1746E+06, 2.1808E+06, 2.1870E+06, 2.1932E+06,
     +  2.1994E+06, 2.2057E+06, 2.2119E+06, 2.2182E+06, 2.2245E+06,
     + 2.2308E+06, 2.2371E+06, 2.2434E+06, 2.2497E+06, 2.2561E+06,
     +  2.2624E+06, 2.2688E+06, 2.2752E+06, 2.2816E+06, 2.2880E+06,
     + 2.2945E+06, 2.3009E+06, 2.3074E+06, 2.3139E+06, 2.3204E+06,
     +  2.3269E+06, 2.3334E+06, 2.3400E+06, 2.3465E+06, 2.3531E+06,
     + 2.3597E+06, 2.3663E+06, 2.3729E+06, 2.3795E+06, 2.3862E+06,
     +  2.3928E+06, 2.3995E+06, 2.4062E+06, 2.4129E+06, 2.4196E+06,
     + 2.4264E+06, 2.4331E+06, 2.4399E+06, 2.4466E+06, 2.4534E+06,
     +  2.4603E+06, 2.4671E+06, 2.4739E+06, 2.4808E+06, 2.4877E+06,
     + 2.4945E+06, 2.5014E+06, 2.5084E+06, 2.5153E+06, 2.5222E+06,
     +  2.5292E+06, 2.5362E+06, 2.5432E+06, 2.5502E+06, 2.5572E+06,
     + 2.5643E+06, 2.5713E+06, 2.5784E+06, 2.5855E+06, 2.5926E+06,
     +  2.5997E+06, 2.6069E+06, 2.6140E+06, 2.6212E+06, 2.6284E+06,
     + 2.6356E+06, 2.6428E+06, 2.6500E+06, 2.6573E+06, 2.6645E+06,
     +  2.6718E+06, 2.6791E+06, 2.6864E+06, 2.6937E+06, 2.7011E+06,
     + 2.7084E+06, 2.7158E+06, 2.7232E+06, 2.7306E+06, 2.7381E+06,
     +  2.7455E+06, 2.7530E+06, 2.7604E+06, 2.7679E+06, 2.7754E+06,
     + 2.7829E+06, 2.7905E+06, 2.7980E+06, 2.8056E+06, 2.8132E+06,
     +  2.8208E+06, 2.8284E+06, 2.8361E+06, 2.8437E+06, 2.8514E+06,
     + 2.8591E+06, 2.8668E+06, 2.8745E+06, 2.8823E+06, 2.8900E+06,
     +  2.8978E+06, 2.9056E+06, 2.9134E+06, 2.9212E+06, 2.9291E+06,
     + 2.9369E+06, 2.9448E+06, 2.9527E+06, 2.9606E+06, 2.9685E+06,
     +  2.9765E+06, 2.9844E+06, 2.9924E+06, 3.0004E+06, 3.0084E+06,
     + 3.0164E+06, 3.0245E+06, 3.0325E+06, 3.0406E+06, 3.0487E+06,
     +  3.0568E+06, 3.0650E+06, 3.0731E+06, 3.0813E+06, 3.0895E+06,
     + 3.0977E+06, 3.1059E+06, 3.1141E+06, 3.1224E+06, 3.1307E+06,
     +  3.1390E+06, 3.1473E+06, 3.1556E+06, 3.1639E+06, 3.1723E+06,
     + 3.1807E+06, 3.1891E+06, 3.1975E+06, 3.2059E+06, 3.2144E+06,
     +  3.2228E+06, 3.2313E+06, 3.2398E+06, 3.2484E+06, 3.2569E+06,
     + 3.2655E+06, 3.2740E+06, 3.2826E+06, 3.2912E+06, 3.2999E+06,
     +  3.3085E+06, 3.3172E+06, 3.3259E+06, 3.3346E+06, 3.3433E+06,
     + 3.3521E+06, 3.3608E+06, 3.3696E+06, 3.3784E+06, 3.3872E+06,
     +  3.3960E+06, 3.4049E+06, 3.4138E+06, 3.4227E+06, 3.4316E+06,
     + 3.4405E+06, 3.4494E+06, 3.4584E+06, 3.4674E+06, 3.4764E+06,
     +  3.4854E+06, 3.4945E+06, 3.5035E+06, 3.5126E+06, 3.5217E+06,
     + 3.5308E+06, 3.5400E+06, 3.5491E+06, 3.5583E+06, 3.5675E+06,
     +  3.5767E+06, 3.5859E+06, 3.5952E+06, 3.6044E+06, 3.6137E+06,
     + 3.6230E+06, 3.6324E+06, 3.6417E+06, 3.6511E+06, 3.6605E+06,
     +  3.6699E+06, 3.6793E+06, 3.6887E+06, 3.6982E+06, 3.7077E+06,
     + 3.7172E+06, 3.7267E+06, 3.7363E+06, 3.7458E+06, 3.7554E+06,
     +  3.7650E+06, 3.7746E+06, 3.7843E+06, 3.7939E+06, 3.8036E+06,
     + 3.8133E+06, 3.8230E+06, 3.8328E+06, 3.8425E+06, 3.8523E+06,
     +  3.8621E+06, 3.8719E+06, 3.8818E+06, 3.8916E+06, 3.9015E+06,
     + 3.9114E+06, 3.9213E+06, 3.9313E+06, 3.9412E+06, 3.9512E+06,
     +  3.9612E+06, 3.9712E+06, 3.9813E+06, 3.9914E+06, 4.0014E+06,
     + 4.0115E+06, 4.0217E+06, 4.0318E+06, 4.0420E+06, 4.0522E+06,
     +  4.0624E+06, 4.0726E+06, 4.0829E+06, 4.0931E+06, 4.1034E+06,
     + 4.1137E+06, 4.1241E+06, 4.1344E+06, 4.1448E+06, 4.1552E+06,
     +  4.1656E+06, 4.1761E+06, 4.1865E+06, 4.1970E+06, 4.2075E+06,
     + 4.2180E+06, 4.2286E+06, 4.2391E+06, 4.2497E+06, 4.2603E+06,
     +  4.2710E+06, 4.2816E+06, 4.2923E+06, 4.3030E+06, 4.3137E+06,
     + 4.3244E+06, 4.3352E+06, 4.3460E+06, 4.3568E+06, 4.3676E+06,
     +  4.3784E+06, 4.3893E+06, 4.4002E+06, 4.4111E+06, 4.4220E+06,
     + 4.4330E+06, 4.4440E+06, 4.4550E+06, 4.4660E+06, 4.4770E+06,
     +  4.4881E+06, 4.4992E+06, 4.5103E+06, 4.5214E+06, 4.5326E+06,
     + 4.5437E+06, 4.5549E+06, 4.5661E+06, 4.5774E+06, 4.5886E+06,
     +  4.5999E+06, 4.6112E+06, 4.6226E+06, 4.6339E+06, 4.6453E+06,
     + 4.6567E+06, 4.6681E+06, 4.6796E+06, 4.6910E+06, 4.7025E+06,
     +  4.7140E+06, 4.7255E+06, 4.7371E+06, 4.7487E+06, 4.7603E+06,
     + 4.7719E+06, 4.7836E+06, 4.7952E+06, 4.8069E+06, 4.8186E+06,
     +  4.8304E+06, 4.8421E+06, 4.8539E+06, 4.8657E+06, 4.8776E+06,
     + 4.8894E+06, 4.9013E+06, 4.9132E+06, 4.9251E+06, 4.9371E+06,
     +  4.9490E+06, 4.9610E+06, 4.9731E+06, 4.9851E+06, 4.9972E+06,
     + 5.0092E+06, 5.0214E+06, 5.0335E+06, 5.0457E+06, 5.0578E+06,
     +  5.0700E+06, 5.0823E+06, 5.0945E+06, 5.1068E+06, 5.1191E+06,
     + 5.1314E+06, 5.1438E+06, 5.1562E+06, 5.1686E+06, 5.1810E+06,
     +  5.1934E+06, 5.2059E+06, 5.2184E+06, 5.2309E+06, 5.2434E+06,
     + 5.2560E+06, 5.2686E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 4,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       447
      data (QofT( 5,J),J=1,2408)/ 1.1269E+02, 2.0372E+02, 2.9564E+02,
     +  3.8776E+02, 4.7995E+02, 5.7219E+02, 6.6445E+02, 7.5673E+02,
     + 8.4901E+02, 9.4131E+02, 1.0336E+03, 1.1259E+03, 1.2182E+03,
     +  1.3105E+03, 1.4029E+03, 1.4952E+03, 1.5875E+03, 1.6798E+03,
     + 1.7722E+03, 1.8645E+03, 2.0492E+03, 2.2339E+03, 2.4186E+03,
     +  2.6033E+03, 2.7880E+03, 2.9727E+03, 3.1574E+03, 3.3422E+03,
     + 3.5269E+03, 3.7117E+03, 3.8965E+03, 4.0812E+03, 4.2660E+03,
     +  4.4508E+03, 4.6356E+03, 4.8204E+03, 5.0053E+03, 5.1901E+03,
     + 5.3749E+03, 5.5598E+03, 5.7446E+03, 5.9295E+03, 6.1144E+03,
     +  6.2993E+03, 6.4842E+03, 6.6692E+03, 6.8541E+03, 7.0390E+03,
     + 7.2240E+03, 7.4091E+03, 7.5942E+03, 7.7793E+03, 7.9644E+03,
     +  8.1496E+03, 8.3349E+03, 8.5202E+03, 8.7056E+03, 8.8912E+03,
     + 9.0768E+03, 9.2624E+03, 9.4483E+03, 9.6343E+03, 9.8204E+03,
     +  1.0007E+04, 1.0193E+04, 1.0380E+04, 1.0567E+04, 1.0754E+04,
     + 1.0941E+04, 1.1129E+04, 1.1317E+04, 1.1505E+04, 1.1693E+04,
     +  1.1882E+04, 1.2071E+04, 1.2261E+04, 1.2451E+04, 1.2641E+04,
     + 1.2832E+04, 1.3024E+04, 1.3215E+04, 1.3408E+04, 1.3600E+04,
     +  1.3794E+04, 1.3988E+04, 1.4182E+04, 1.4377E+04, 1.4573E+04,
     + 1.4769E+04, 1.4966E+04, 1.5164E+04, 1.5363E+04, 1.5562E+04,
     +  1.5762E+04, 1.5962E+04, 1.6164E+04, 1.6366E+04, 1.6570E+04,
     + 1.6774E+04, 1.6979E+04, 1.7184E+04, 1.7391E+04, 1.7599E+04,
     +  1.7807E+04, 1.8017E+04, 1.8228E+04, 1.8439E+04, 1.8652E+04,
     + 1.8866E+04, 1.9081E+04, 1.9296E+04, 1.9513E+04, 1.9731E+04,
     +  1.9951E+04, 2.0171E+04, 2.0393E+04, 2.0616E+04, 2.0839E+04,
     + 2.1065E+04, 2.1291E+04, 2.1519E+04, 2.1748E+04, 2.1978E+04,
     +  2.2210E+04, 2.2443E+04, 2.2677E+04, 2.2912E+04, 2.3150E+04,
     + 2.3388E+04, 2.3628E+04, 2.3869E+04, 2.4112E+04, 2.4356E+04,
     +  2.4601E+04, 2.4848E+04, 2.5097E+04, 2.5347E+04, 2.5599E+04,
     + 2.5852E+04, 2.6106E+04, 2.6362E+04, 2.6620E+04, 2.6880E+04,
     +  2.7141E+04, 2.7403E+04, 2.7668E+04, 2.7934E+04, 2.8201E+04,
     + 2.8470E+04, 2.8741E+04, 2.9014E+04, 2.9288E+04, 2.9565E+04,
     +  2.9842E+04, 3.0122E+04, 3.0404E+04, 3.0687E+04, 3.0972E+04,
     + 3.1258E+04, 3.1547E+04, 3.1838E+04, 3.2130E+04, 3.2424E+04,
     +  3.2720E+04, 3.3018E+04, 3.3318E+04, 3.3620E+04, 3.3924E+04,
     + 3.4229E+04, 3.4537E+04, 3.4847E+04, 3.5158E+04, 3.5472E+04,
     +  3.5787E+04, 3.6105E+04, 3.6425E+04, 3.6746E+04, 3.7070E+04,
     + 3.7396E+04, 3.7724E+04, 3.8054E+04, 3.8386E+04, 3.8721E+04,
     +  3.9057E+04, 3.9396E+04, 3.9736E+04, 4.0079E+04, 4.0425E+04,
     + 4.0772E+04, 4.1121E+04, 4.1473E+04, 4.1827E+04, 4.2184E+04,
     +  4.2542E+04, 4.2903E+04, 4.3266E+04, 4.3632E+04, 4.4000E+04,
     + 4.4370E+04, 4.4743E+04, 4.5117E+04, 4.5495E+04, 4.5874E+04,
     +  4.6256E+04, 4.6641E+04, 4.7028E+04, 4.7417E+04, 4.7809E+04,
     + 4.8204E+04, 4.8600E+04, 4.9000E+04, 4.9401E+04, 4.9806E+04,
     +  5.0213E+04, 5.0622E+04, 5.1034E+04, 5.1449E+04, 5.1866E+04,
     + 5.2286E+04, 5.2708E+04, 5.3133E+04, 5.3561E+04, 5.3992E+04,
     +  5.4425E+04, 5.4861E+04, 5.5299E+04, 5.5740E+04, 5.6184E+04,
     + 5.6631E+04, 5.7081E+04, 5.7533E+04, 5.7988E+04, 5.8446E+04,
     +  5.8907E+04, 5.9370E+04, 5.9837E+04, 6.0306E+04, 6.0778E+04,
     + 6.1253E+04, 6.1731E+04, 6.2212E+04, 6.2696E+04, 6.3183E+04,
     +  6.3672E+04, 6.4165E+04, 6.4661E+04, 6.5159E+04, 6.5661E+04,
     + 6.6166E+04, 6.6674E+04, 6.7185E+04, 6.7699E+04, 6.8216E+04,
     +  6.8737E+04, 6.9260E+04, 6.9787E+04, 7.0317E+04, 7.0850E+04,
     + 7.1386E+04, 7.1925E+04, 7.2468E+04, 7.3014E+04, 7.3563E+04,
     +  7.4115E+04, 7.4671E+04, 7.5230E+04, 7.5792E+04, 7.6358E+04,
     + 7.6927E+04, 7.7499E+04, 7.8075E+04, 7.8655E+04, 7.9237E+04,
     +  7.9823E+04, 8.0413E+04, 8.1006E+04, 8.1602E+04, 8.2202E+04,
     + 8.2806E+04, 8.3413E+04, 8.4023E+04, 8.4638E+04, 8.5255E+04,
     +  8.5877E+04, 8.6502E+04, 8.7130E+04, 8.7763E+04, 8.8398E+04,
     + 8.9038E+04, 8.9681E+04, 9.0329E+04, 9.0979E+04, 9.1634E+04,
     +  9.2293E+04, 9.2955E+04, 9.3621E+04, 9.4290E+04, 9.4964E+04,
     + 9.5642E+04, 9.6323E+04, 9.7008E+04, 9.7698E+04, 9.8391E+04,
     +  9.9088E+04, 9.9789E+04, 1.0049E+05, 1.0120E+05, 1.0192E+05,
     + 1.0263E+05, 1.0335E+05, 1.0408E+05, 1.0481E+05, 1.0554E+05,
     +  1.0628E+05, 1.0702E+05, 1.0777E+05, 1.0852E+05, 1.0927E+05,
     + 1.1003E+05, 1.1079E+05, 1.1156E+05, 1.1233E+05, 1.1311E+05,
     +  1.1389E+05, 1.1467E+05, 1.1546E+05, 1.1625E+05, 1.1705E+05,
     + 1.1785E+05, 1.1866E+05, 1.1947E+05, 1.2028E+05, 1.2110E+05,
     +  1.2193E+05, 1.2276E+05, 1.2359E+05, 1.2443E+05, 1.2527E+05,
     + 1.2612E+05, 1.2697E+05, 1.2783E+05, 1.2869E+05, 1.2955E+05,
     +  1.3042E+05, 1.3130E+05, 1.3218E+05, 1.3306E+05, 1.3395E+05,
     + 1.3485E+05, 1.3574E+05, 1.3665E+05, 1.3756E+05, 1.3847E+05,
     +  1.3939E+05, 1.4031E+05, 1.4124E+05, 1.4218E+05, 1.4311E+05,
     + 1.4406E+05, 1.4501E+05, 1.4596E+05, 1.4692E+05, 1.4788E+05,
     +  1.4885E+05, 1.4983E+05, 1.5080E+05, 1.5179E+05, 1.5278E+05,
     + 1.5377E+05, 1.5477E+05, 1.5578E+05, 1.5679E+05, 1.5781E+05,
     +  1.5883E+05, 1.5985E+05, 1.6089E+05, 1.6192E+05, 1.6297E+05,
     + 1.6401E+05, 1.6507E+05, 1.6613E+05, 1.6719E+05, 1.6826E+05,
     +  1.6934E+05, 1.7042E+05, 1.7151E+05, 1.7260E+05, 1.7370E+05,
     + 1.7480E+05, 1.7591E+05, 1.7703E+05, 1.7815E+05, 1.7928E+05,
     +  1.8041E+05, 1.8155E+05, 1.8269E+05, 1.8384E+05, 1.8500E+05,
     + 1.8616E+05, 1.8733E+05, 1.8851E+05, 1.8969E+05, 1.9087E+05,
     +  1.9206E+05, 1.9326E+05, 1.9447E+05, 1.9568E+05, 1.9689E+05,
     + 1.9812E+05, 1.9935E+05, 2.0058E+05, 2.0182E+05, 2.0307E+05,
     +  2.0432E+05, 2.0558E+05, 2.0685E+05, 2.0812E+05, 2.0940E+05,
     + 2.1069E+05, 2.1198E+05, 2.1328E+05, 2.1458E+05, 2.1589E+05,
     +  2.1721E+05, 2.1854E+05, 2.1987E+05, 2.2120E+05, 2.2255E+05,
     + 2.2390E+05, 2.2526E+05, 2.2662E+05, 2.2799E+05, 2.2937E+05,
     +  2.3075E+05, 2.3215E+05, 2.3354E+05, 2.3495E+05, 2.3636E+05,
     + 2.3778E+05, 2.3920E+05, 2.4064E+05, 2.4208E+05, 2.4352E+05,
     +  2.4498E+05, 2.4644E+05, 2.4791E+05, 2.4938E+05, 2.5086E+05,
     + 2.5235E+05, 2.5385E+05, 2.5535E+05, 2.5686E+05, 2.5838E+05,
     +  2.5991E+05, 2.6144E+05, 2.6298E+05, 2.6453E+05, 2.6608E+05,
     + 2.6764E+05, 2.6922E+05, 2.7079E+05, 2.7238E+05, 2.7397E+05,
     +  2.7557E+05, 2.7718E+05, 2.7879E+05, 2.8042E+05, 2.8205E+05,
     + 2.8369E+05, 2.8533E+05, 2.8699E+05, 2.8865E+05, 2.9032E+05,
     +  2.9199E+05, 2.9368E+05, 2.9537E+05, 2.9707E+05, 2.9878E+05,
     + 3.0050E+05, 3.0223E+05, 3.0396E+05, 3.0570E+05, 3.0745E+05,
     +  3.0921E+05, 3.1097E+05, 3.1275E+05, 3.1453E+05, 3.1632E+05,
     + 3.1812E+05, 3.1993E+05, 3.2174E+05, 3.2357E+05, 3.2540E+05,
     +  3.2724E+05, 3.2909E+05, 3.3095E+05, 3.3281E+05, 3.3469E+05,
     + 3.3657E+05, 3.3847E+05, 3.4037E+05, 3.4228E+05, 3.4419E+05,
     +  3.4612E+05, 3.4806E+05, 3.5000E+05, 3.5196E+05, 3.5392E+05,
     + 3.5589E+05, 3.5787E+05, 3.5986E+05, 3.6186E+05, 3.6386E+05,
     +  3.6588E+05, 3.6791E+05, 3.6994E+05, 3.7199E+05, 3.7404E+05,
     + 3.7610E+05, 3.7817E+05, 3.8025E+05, 3.8234E+05, 3.8444E+05,
     +  3.8655E+05, 3.8867E+05, 3.9080E+05, 3.9294E+05, 3.9508E+05,
     + 3.9724E+05, 3.9940E+05, 4.0158E+05, 4.0377E+05, 4.0596E+05,
     +  4.0816E+05, 4.1038E+05, 4.1260E+05, 4.1484E+05, 4.1708E+05,
     + 4.1934E+05, 4.2160E+05, 4.2387E+05, 4.2616E+05, 4.2845E+05,
     +  4.3075E+05, 4.3307E+05, 4.3539E+05, 4.3772E+05, 4.4007E+05,
     + 4.4242E+05, 4.4479E+05, 4.4716E+05, 4.4955E+05, 4.5194E+05,
     +  4.5435E+05, 4.5677E+05, 4.5919E+05, 4.6163E+05, 4.6408E+05,
     + 4.6654E+05, 4.6901E+05, 4.7149E+05, 4.7398E+05, 4.7648E+05,
     +  4.7899E+05, 4.8151E+05, 4.8405E+05, 4.8659E+05, 4.8915E+05,
     + 4.9171E+05, 4.9429E+05, 4.9688E+05, 4.9948E+05, 5.0209E+05,
     +  5.0471E+05, 5.0734E+05, 5.0998E+05, 5.1264E+05, 5.1531E+05,
     + 5.1798E+05, 5.2067E+05, 5.2337E+05, 5.2608E+05, 5.2881E+05,
     +  5.3154E+05, 5.3429E+05, 5.3704E+05, 5.3981E+05, 5.4259E+05,
     + 5.4539E+05, 5.4819E+05, 5.5101E+05, 5.5383E+05, 5.5667E+05,
     +  5.5952E+05, 5.6239E+05, 5.6526E+05, 5.6815E+05, 5.7105E+05,
     + 5.7396E+05, 5.7688E+05, 5.7982E+05, 5.8277E+05, 5.8572E+05,
     +  5.8870E+05, 5.9168E+05, 5.9468E+05, 5.9769E+05, 6.0071E+05,
     + 6.0374E+05, 6.0679E+05, 6.0985E+05, 6.1292E+05, 6.1600E+05,
     +  6.1910E+05, 6.2221E+05, 6.2533E+05, 6.2846E+05, 6.3161E+05,
     + 6.3477E+05, 6.3794E+05, 6.4113E+05, 6.4433E+05, 6.4754E+05,
     +  6.5076E+05, 6.5400E+05, 6.5725E+05, 6.6052E+05, 6.6379E+05,
     + 6.6708E+05, 6.7039E+05, 6.7370E+05, 6.7704E+05, 6.8038E+05,
     +  6.8374E+05, 6.8711E+05, 6.9049E+05, 6.9389E+05, 6.9730E+05,
     + 7.0073E+05, 7.0416E+05, 7.0762E+05, 7.1108E+05, 7.1456E+05,
     +  7.1806E+05, 7.2156E+05, 7.2509E+05, 7.2862E+05, 7.3217E+05,
     + 7.3573E+05, 7.3931E+05, 7.4290E+05, 7.4651E+05, 7.5013E+05,
     +  7.5377E+05, 7.5742E+05, 7.6108E+05, 7.6476E+05, 7.6845E+05,
     + 7.7215E+05, 7.7588E+05, 7.7961E+05, 7.8336E+05, 7.8713E+05,
     +  7.9091E+05, 7.9470E+05, 7.9851E+05, 8.0234E+05, 8.0617E+05,
     + 8.1003E+05, 8.1390E+05, 8.1778E+05, 8.2168E+05, 8.2560E+05,
     +  8.2953E+05, 8.3347E+05, 8.3743E+05, 8.4141E+05, 8.4540E+05,
     + 8.4940E+05, 8.5342E+05, 8.5746E+05, 8.6151E+05, 8.6558E+05,
     +  8.6966E+05, 8.7376E+05, 8.7788E+05, 8.8201E+05, 8.8615E+05,
     + 8.9032E+05, 8.9450E+05, 8.9869E+05, 9.0290E+05, 9.0712E+05,
     +  9.1137E+05, 9.1563E+05, 9.1990E+05, 9.2419E+05, 9.2850E+05,
     + 9.3282E+05, 9.3716E+05, 9.4152E+05, 9.4589E+05, 9.5028E+05,
     +  9.5468E+05, 9.5910E+05, 9.6354E+05, 9.6800E+05, 9.7247E+05,
     + 9.7696E+05, 9.8147E+05, 9.8599E+05, 9.9053E+05, 9.9509E+05,
     +  9.9966E+05, 1.0042E+06, 1.0089E+06, 1.0135E+06, 1.0181E+06,
     + 1.0228E+06, 1.0275E+06, 1.0322E+06, 1.0369E+06, 1.0416E+06,
     +  1.0463E+06, 1.0511E+06, 1.0559E+06, 1.0607E+06, 1.0655E+06,
     + 1.0703E+06, 1.0752E+06, 1.0801E+06, 1.0849E+06, 1.0899E+06,
     +  1.0948E+06, 1.0997E+06, 1.1047E+06, 1.1097E+06, 1.1147E+06,
     + 1.1197E+06, 1.1247E+06, 1.1298E+06, 1.1348E+06, 1.1399E+06,
     +  1.1450E+06, 1.1502E+06, 1.1553E+06, 1.1605E+06, 1.1656E+06,
     + 1.1708E+06, 1.1761E+06, 1.1813E+06, 1.1866E+06, 1.1918E+06,
     +  1.1971E+06, 1.2024E+06, 1.2078E+06, 1.2131E+06, 1.2185E+06,
     + 1.2239E+06, 1.2293E+06, 1.2347E+06, 1.2402E+06, 1.2457E+06,
     +  1.2511E+06, 1.2567E+06, 1.2622E+06, 1.2677E+06, 1.2733E+06,
     + 1.2789E+06, 1.2845E+06, 1.2901E+06, 1.2958E+06, 1.3014E+06,
     +  1.3071E+06, 1.3128E+06, 1.3186E+06, 1.3243E+06, 1.3301E+06,
     + 1.3359E+06, 1.3417E+06, 1.3475E+06, 1.3534E+06, 1.3592E+06,
     +  1.3651E+06, 1.3710E+06, 1.3770E+06, 1.3829E+06, 1.3889E+06,
     + 1.3949E+06, 1.4009E+06, 1.4069E+06, 1.4130E+06, 1.4191E+06,
     +  1.4252E+06, 1.4313E+06, 1.4375E+06, 1.4436E+06, 1.4498E+06,
     + 1.4560E+06, 1.4622E+06, 1.4685E+06, 1.4748E+06, 1.4811E+06,
     +  1.4874E+06, 1.4937E+06, 1.5001E+06, 1.5065E+06, 1.5129E+06,
     + 1.5193E+06, 1.5257E+06, 1.5322E+06, 1.5387E+06, 1.5452E+06,
     +  1.5518E+06, 1.5583E+06, 1.5649E+06, 1.5715E+06, 1.5781E+06,
     + 1.5848E+06, 1.5915E+06, 1.5982E+06, 1.6049E+06, 1.6116E+06,
     +  1.6184E+06, 1.6252E+06, 1.6320E+06, 1.6388E+06, 1.6457E+06,
     + 1.6526E+06, 1.6595E+06, 1.6664E+06, 1.6733E+06, 1.6803E+06,
     +  1.6873E+06, 1.6943E+06, 1.7014E+06, 1.7084E+06, 1.7155E+06,
     + 1.7227E+06, 1.7298E+06, 1.7370E+06, 1.7442E+06, 1.7514E+06,
     +  1.7586E+06, 1.7659E+06, 1.7732E+06, 1.7805E+06, 1.7878E+06,
     + 1.7952E+06, 1.8025E+06, 1.8099E+06, 1.8174E+06, 1.8248E+06,
     +  1.8323E+06, 1.8398E+06, 1.8474E+06, 1.8549E+06, 1.8625E+06,
     + 1.8701E+06, 1.8777E+06, 1.8854E+06, 1.8931E+06, 1.9008E+06,
     +  1.9085E+06, 1.9163E+06, 1.9241E+06, 1.9319E+06, 1.9397E+06,
     + 1.9476E+06, 1.9555E+06, 1.9634E+06, 1.9713E+06, 1.9793E+06,
     +  1.9873E+06, 1.9953E+06, 2.0033E+06, 2.0114E+06, 2.0195E+06,
     + 2.0276E+06, 2.0358E+06, 2.0439E+06, 2.0521E+06, 2.0604E+06,
     +  2.0686E+06, 2.0769E+06, 2.0852E+06, 2.0936E+06, 2.1019E+06,
     + 2.1103E+06, 2.1187E+06, 2.1272E+06, 2.1356E+06, 2.1441E+06,
     +  2.1527E+06, 2.1612E+06, 2.1698E+06, 2.1784E+06, 2.1871E+06,
     + 2.1957E+06, 2.2044E+06, 2.2131E+06, 2.2219E+06, 2.2307E+06,
     +  2.2395E+06, 2.2483E+06, 2.2572E+06, 2.2661E+06, 2.2750E+06,
     + 2.2839E+06, 2.2929E+06, 2.3019E+06, 2.3109E+06, 2.3200E+06,
     +  2.3291E+06, 2.3382E+06, 2.3473E+06, 2.3565E+06, 2.3657E+06,
     + 2.3750E+06, 2.3842E+06, 2.3935E+06, 2.4028E+06, 2.4122E+06,
     +  2.4216E+06, 2.4310E+06, 2.4404E+06, 2.4499E+06, 2.4594E+06,
     + 2.4689E+06, 2.4785E+06, 2.4881E+06, 2.4977E+06, 2.5073E+06,
     +  2.5170E+06, 2.5267E+06, 2.5365E+06, 2.5462E+06, 2.5560E+06,
     + 2.5659E+06, 2.5757E+06, 2.5856E+06, 2.5955E+06, 2.6055E+06,
     +  2.6155E+06, 2.6255E+06, 2.6355E+06, 2.6456E+06, 2.6557E+06,
     + 2.6659E+06, 2.6760E+06, 2.6862E+06, 2.6965E+06, 2.7067E+06,
     +  2.7170E+06, 2.7274E+06, 2.7377E+06, 2.7481E+06, 2.7586E+06,
     + 2.7690E+06, 2.7795E+06, 2.7900E+06, 2.8006E+06, 2.8112E+06,
     +  2.8218E+06, 2.8324E+06, 2.8431E+06, 2.8538E+06, 2.8646E+06,
     + 2.8754E+06, 2.8862E+06, 2.8970E+06, 2.9079E+06, 2.9188E+06,
     +  2.9298E+06, 2.9408E+06, 2.9518E+06, 2.9628E+06, 2.9739E+06,
     + 2.9850E+06, 2.9962E+06, 3.0073E+06, 3.0185E+06, 3.0298E+06,
     +  3.0411E+06, 3.0524E+06, 3.0637E+06, 3.0751E+06, 3.0866E+06,
     + 3.0980E+06, 3.1095E+06, 3.1210E+06, 3.1326E+06, 3.1442E+06,
     +  3.1558E+06, 3.1674E+06, 3.1791E+06, 3.1909E+06, 3.2026E+06,
     + 3.2144E+06, 3.2263E+06, 3.2381E+06, 3.2501E+06, 3.2620E+06,
     +  3.2740E+06, 3.2860E+06, 3.2980E+06, 3.3101E+06, 3.3222E+06,
     + 3.3344E+06, 3.3466E+06, 3.3588E+06, 3.3711E+06, 3.3834E+06,
     +  3.3957E+06, 3.4081E+06, 3.4205E+06, 3.4329E+06, 3.4454E+06,
     + 3.4579E+06, 3.4705E+06, 3.4831E+06, 3.4957E+06, 3.5084E+06,
     +  3.5211E+06, 3.5338E+06, 3.5466E+06, 3.5594E+06, 3.5723E+06,
     + 3.5852E+06, 3.5981E+06, 3.6111E+06, 3.6241E+06, 3.6371E+06,
     +  3.6502E+06, 3.6633E+06, 3.6765E+06, 3.6897E+06, 3.7029E+06,
     + 3.7162E+06, 3.7295E+06, 3.7428E+06, 3.7562E+06, 3.7696E+06,
     +  3.7831E+06, 3.7966E+06, 3.8102E+06, 3.8237E+06, 3.8374E+06,
     + 3.8510E+06, 3.8647E+06, 3.8785E+06, 3.8922E+06, 3.9060E+06,
     +  3.9199E+06, 3.9338E+06, 3.9477E+06, 3.9617E+06, 3.9757E+06,
     + 3.9898E+06, 4.0039E+06, 4.0180E+06, 4.0322E+06, 4.0464E+06,
     +  4.0607E+06, 4.0750E+06, 4.0893E+06, 4.1037E+06, 4.1181E+06,
     + 4.1326E+06, 4.1471E+06, 4.1616E+06, 4.1762E+06, 4.1908E+06,
     +  4.2055E+06, 4.2202E+06, 4.2350E+06, 4.2498E+06, 4.2646E+06,
     + 4.2795E+06, 4.2944E+06, 4.3094E+06, 4.3244E+06, 4.3394E+06,
     +  4.3545E+06, 4.3696E+06, 4.3848E+06, 4.4000E+06, 4.4153E+06,
     + 4.4306E+06, 4.4459E+06, 4.4613E+06, 4.4767E+06, 4.4922E+06,
     +  4.5077E+06, 4.5233E+06, 4.5389E+06, 4.5545E+06, 4.5702E+06,
     + 4.5859E+06, 4.6017E+06, 4.6175E+06, 4.6334E+06, 4.6493E+06,
     +  4.6653E+06, 4.6813E+06, 4.6973E+06, 4.7134E+06, 4.7295E+06,
     + 4.7457E+06, 4.7619E+06, 4.7782E+06, 4.7945E+06, 4.8108E+06,
     +  4.8272E+06, 4.8437E+06, 4.8602E+06, 4.8767E+06, 4.8933E+06,
     + 4.9099E+06, 4.9266E+06, 4.9433E+06, 4.9601E+06, 4.9769E+06,
     +  4.9937E+06, 5.0106E+06, 5.0276E+06, 5.0446E+06, 5.0616E+06,
     + 5.0787E+06, 5.0958E+06, 5.1130E+06, 5.1303E+06, 5.1475E+06,
     +  5.1649E+06, 5.1822E+06, 5.1996E+06, 5.2171E+06, 5.2346E+06,
     + 5.2522E+06, 5.2698E+06, 5.2874E+06, 5.3051E+06, 5.3229E+06,
     +  5.3407E+06, 5.3585E+06, 5.3764E+06, 5.3944E+06, 5.4123E+06,
     + 5.4304E+06, 5.4485E+06, 5.4666E+06, 5.4848E+06, 5.5030E+06,
     +  5.5213E+06, 5.5396E+06, 5.5580E+06, 5.5764E+06, 5.5949E+06,
     + 5.6134E+06, 5.6320E+06, 5.6507E+06, 5.6693E+06, 5.6881E+06,
     +  5.7068E+06, 5.7257E+06, 5.7445E+06, 5.7635E+06, 5.7824E+06,
     + 5.8015E+06, 5.8205E+06, 5.8397E+06, 5.8589E+06, 5.8781E+06,
     +  5.8974E+06, 5.9167E+06, 5.9361E+06, 5.9555E+06, 5.9750E+06,
     + 5.9946E+06, 6.0141E+06, 6.0338E+06, 6.0535E+06, 6.0732E+06,
     +  6.0930E+06, 6.1129E+06, 6.1328E+06, 6.1527E+06, 6.1727E+06,
     + 6.1928E+06, 6.2129E+06, 6.2331E+06, 6.2533E+06, 6.2736E+06,
     +  6.2939E+06, 6.3143E+06, 6.3347E+06, 6.3552E+06, 6.3757E+06,
     + 6.3963E+06, 6.4169E+06, 6.4376E+06, 6.4584E+06, 6.4792E+06,
     +  6.5001E+06, 6.5210E+06, 6.5419E+06, 6.5630E+06, 6.5841E+06,
     + 6.6052E+06, 6.6264E+06, 6.6476E+06, 6.6689E+06, 6.6903E+06,
     +  6.7117E+06, 6.7331E+06, 6.7547E+06, 6.7762E+06, 6.7979E+06,
     + 6.8195E+06, 6.8413E+06, 6.8631E+06, 6.8849E+06, 6.9068E+06,
     +  6.9288E+06, 6.9508E+06, 6.9729E+06, 6.9950E+06, 7.0172E+06,
     + 7.0395E+06, 7.0618E+06, 7.0842E+06, 7.1066E+06, 7.1290E+06,
     +  7.1516E+06, 7.1742E+06, 7.1968E+06, 7.2195E+06, 7.2423E+06,
     + 7.2651E+06, 7.2880E+06, 7.3109E+06, 7.3339E+06, 7.3570E+06,
     +  7.3801E+06, 7.4033E+06, 7.4265E+06, 7.4498E+06, 7.4732E+06,
     + 7.4966E+06, 7.5200E+06, 7.5436E+06, 7.5672E+06, 7.5908E+06,
     +  7.6145E+06, 7.6383E+06, 7.6621E+06, 7.6860E+06, 7.7099E+06,
     + 7.7339E+06, 7.7580E+06, 7.7821E+06, 7.8063E+06, 7.8306E+06,
     +  7.8549E+06, 7.8792E+06, 7.9037E+06, 7.9282E+06, 7.9527E+06,
     + 7.9773E+06, 8.0020E+06, 8.0268E+06, 8.0515E+06, 8.0764E+06,
     +  8.1013E+06, 8.1263E+06, 8.1514E+06, 8.1765E+06, 8.2017E+06,
     + 8.2269E+06, 8.2522E+06, 8.2775E+06, 8.3030E+06, 8.3285E+06,
     +  8.3540E+06, 8.3796E+06, 8.4053E+06, 8.4310E+06, 8.4568E+06,
     + 8.4827E+06, 8.5086E+06, 8.5346E+06, 8.5607E+06, 8.5868E+06,
     +  8.6130E+06, 8.6393E+06, 8.6656E+06, 8.6920E+06, 8.7184E+06,
     + 8.7449E+06, 8.7715E+06, 8.7981E+06, 8.8248E+06, 8.8516E+06,
     +  8.8784E+06, 8.9054E+06, 8.9323E+06, 8.9594E+06, 8.9865E+06,
     + 9.0136E+06, 9.0409E+06, 9.0682E+06, 9.0955E+06, 9.1230E+06,
     +  9.1505E+06, 9.1780E+06, 9.2057E+06, 9.2334E+06, 9.2611E+06,
     + 9.2890E+06, 9.3169E+06, 9.3448E+06, 9.3729E+06, 9.4010E+06,
     +  9.4292E+06, 9.4574E+06, 9.4857E+06, 9.5141E+06, 9.5425E+06,
     + 9.5711E+06, 9.5996E+06, 9.6283E+06, 9.6570E+06, 9.6858E+06,
     +  9.7147E+06, 9.7436E+06, 9.7726E+06, 9.8017E+06, 9.8308E+06,
     + 9.8600E+06, 9.8893E+06, 9.9187E+06, 9.9481E+06, 9.9776E+06,
     +  1.0007E+07, 1.0037E+07, 1.0066E+07, 1.0096E+07, 1.0126E+07,
     + 1.0156E+07, 1.0186E+07, 1.0216E+07, 1.0246E+07, 1.0276E+07,
     +  1.0307E+07, 1.0337E+07, 1.0367E+07, 1.0398E+07, 1.0428E+07,
     + 1.0459E+07, 1.0490E+07, 1.0521E+07, 1.0551E+07, 1.0582E+07,
     +  1.0613E+07, 1.0644E+07, 1.0676E+07, 1.0707E+07, 1.0738E+07,
     + 1.0769E+07, 1.0801E+07, 1.0832E+07, 1.0864E+07, 1.0896E+07,
     +  1.0927E+07, 1.0959E+07, 1.0991E+07, 1.1023E+07, 1.1055E+07,
     + 1.1087E+07, 1.1119E+07, 1.1152E+07, 1.1184E+07, 1.1216E+07,
     +  1.1249E+07, 1.1282E+07, 1.1314E+07, 1.1347E+07, 1.1380E+07,
     + 1.1413E+07, 1.1445E+07, 1.1479E+07, 1.1512E+07, 1.1545E+07,
     +  1.1578E+07, 1.1611E+07, 1.1645E+07, 1.1678E+07, 1.1712E+07,
     + 1.1745E+07, 1.1779E+07, 1.1813E+07, 1.1847E+07, 1.1881E+07,
     +  1.1915E+07, 1.1949E+07, 1.1983E+07, 1.2017E+07, 1.2052E+07,
     + 1.2086E+07, 1.2121E+07, 1.2155E+07, 1.2190E+07, 1.2225E+07,
     +  1.2260E+07, 1.2294E+07, 1.2329E+07, 1.2365E+07, 1.2400E+07,
     + 1.2435E+07, 1.2470E+07, 1.2506E+07, 1.2541E+07, 1.2577E+07,
     +  1.2612E+07, 1.2648E+07, 1.2684E+07, 1.2720E+07, 1.2756E+07,
     + 1.2792E+07, 1.2828E+07, 1.2864E+07, 1.2900E+07, 1.2937E+07,
     +  1.2973E+07, 1.3010E+07, 1.3046E+07, 1.3083E+07, 1.3120E+07,
     + 1.3157E+07, 1.3194E+07, 1.3231E+07, 1.3268E+07, 1.3305E+07,
     +  1.3342E+07, 1.3380E+07, 1.3417E+07, 1.3455E+07, 1.3492E+07,
     + 1.3530E+07, 1.3568E+07, 1.3606E+07, 1.3644E+07, 1.3682E+07,
     +  1.3720E+07, 1.3758E+07, 1.3796E+07, 1.3835E+07, 1.3873E+07,
     + 1.3912E+07, 1.3950E+07, 1.3989E+07, 1.4028E+07, 1.4067E+07,
     +  1.4106E+07, 1.4145E+07, 1.4184E+07, 1.4223E+07, 1.4263E+07,
     + 1.4302E+07, 1.4342E+07, 1.4381E+07, 1.4421E+07, 1.4461E+07,
     +  1.4501E+07, 1.4541E+07, 1.4581E+07, 1.4621E+07, 1.4661E+07,
     + 1.4702E+07, 1.4742E+07, 1.4782E+07, 1.4823E+07, 1.4864E+07,
     +  1.4905E+07, 1.4945E+07, 1.4986E+07, 1.5027E+07, 1.5069E+07,
     + 1.5110E+07, 1.5151E+07, 1.5193E+07, 1.5234E+07, 1.5276E+07,
     +  1.5317E+07, 1.5359E+07, 1.5401E+07, 1.5443E+07, 1.5485E+07,
     + 1.5527E+07, 1.5569E+07, 1.5612E+07, 1.5654E+07, 1.5697E+07,
     +  1.5739E+07, 1.5782E+07, 1.5825E+07, 1.5868E+07, 1.5911E+07,
     + 1.5954E+07, 1.5997E+07, 1.6040E+07, 1.6083E+07, 1.6127E+07,
     +  1.6170E+07, 1.6214E+07, 1.6258E+07, 1.6302E+07, 1.6346E+07,
     + 1.6390E+07, 1.6434E+07, 1.6478E+07, 1.6522E+07, 1.6567E+07,
     +  1.6611E+07, 1.6656E+07, 1.6701E+07, 1.6745E+07, 1.6790E+07,
     + 1.6835E+07, 1.6880E+07, 1.6925E+07, 1.6971E+07, 1.7016E+07,
     +  1.7062E+07, 1.7107E+07, 1.7153E+07, 1.7199E+07, 1.7244E+07,
     + 1.7290E+07, 1.7337E+07, 1.7383E+07, 1.7429E+07, 1.7475E+07,
     +  1.7522E+07, 1.7568E+07, 1.7615E+07, 1.7662E+07, 1.7709E+07,
     + 1.7756E+07, 1.7803E+07, 1.7850E+07, 1.7897E+07, 1.7944E+07,
     +  1.7992E+07, 1.8039E+07, 1.8087E+07, 1.8135E+07, 1.8183E+07,
     + 1.8231E+07, 1.8279E+07, 1.8327E+07, 1.8375E+07, 1.8424E+07,
     +  1.8472E+07, 1.8521E+07, 1.8569E+07, 1.8618E+07, 1.8667E+07,
     + 1.8716E+07, 1.8765E+07, 1.8814E+07, 1.8864E+07, 1.8913E+07,
     +  1.8963E+07, 1.9012E+07, 1.9062E+07, 1.9112E+07, 1.9162E+07,
     + 1.9212E+07, 1.9262E+07, 1.9312E+07, 1.9363E+07, 1.9413E+07,
     +  1.9464E+07, 1.9514E+07, 1.9565E+07, 1.9616E+07, 1.9667E+07,
     + 1.9718E+07, 1.9769E+07, 1.9821E+07, 1.9872E+07, 1.9924E+07,
     +  1.9975E+07, 2.0027E+07, 2.0079E+07, 2.0131E+07, 2.0183E+07,
     + 2.0235E+07, 2.0288E+07, 2.0340E+07, 2.0392E+07, 2.0445E+07,
     +  2.0498E+07, 2.0551E+07, 2.0604E+07, 2.0657E+07, 2.0710E+07,
     + 2.0763E+07, 2.0817E+07, 2.0870E+07, 2.0924E+07, 2.0977E+07,
     +  2.1031E+07, 2.1085E+07, 2.1139E+07, 2.1193E+07, 2.1248E+07,
     + 2.1302E+07, 2.1357E+07, 2.1411E+07, 2.1466E+07, 2.1521E+07,
     +  2.1576E+07, 2.1631E+07, 2.1686E+07, 2.1741E+07, 2.1797E+07,
     + 2.1852E+07, 2.1908E+07, 2.1964E+07, 2.2020E+07, 2.2076E+07,
     +  2.2132E+07, 2.2188E+07, 2.2244E+07, 2.2301E+07, 2.2357E+07,
     + 2.2414E+07, 2.2471E+07, 2.2528E+07, 2.2585E+07, 2.2642E+07,
     +  2.2699E+07, 2.2756E+07, 2.2814E+07, 2.2871E+07, 2.2929E+07,
     + 2.2987E+07, 2.3045E+07, 2.3103E+07, 2.3161E+07, 2.3220E+07,
     +  2.3278E+07, 2.3337E+07, 2.3395E+07, 2.3454E+07, 2.3513E+07,
     + 2.3572E+07, 2.3631E+07, 2.3690E+07, 2.3750E+07, 2.3809E+07,
     +  2.3869E+07, 2.3929E+07, 2.3988E+07, 2.4048E+07, 2.4109E+07,
     + 2.4169E+07, 2.4229E+07, 2.4290E+07, 2.4350E+07, 2.4411E+07,
     +  2.4472E+07, 2.4533E+07, 2.4594E+07, 2.4655E+07, 2.4716E+07,
     + 2.4778E+07, 2.4839E+07, 2.4901E+07, 2.4963E+07, 2.5025E+07,
     +  2.5087E+07, 2.5149E+07, 2.5211E+07, 2.5274E+07, 2.5336E+07,
     + 2.5399E+07, 2.5462E+07, 2.5525E+07, 2.5588E+07, 2.5651E+07,
     +  2.5714E+07, 2.5778E+07, 2.5841E+07, 2.5905E+07, 2.5969E+07,
     + 2.6033E+07, 2.6097E+07, 2.6161E+07, 2.6226E+07, 2.6290E+07,
     +  2.6355E+07, 2.6419E+07, 2.6484E+07, 2.6549E+07, 2.6614E+07,
     + 2.6679E+07, 2.6745E+07, 2.6810E+07, 2.6876E+07, 2.6942E+07,
     +  2.7008E+07, 2.7074E+07, 2.7140E+07, 2.7206E+07, 2.7272E+07,
     + 2.7339E+07, 2.7406E+07, 2.7472E+07, 2.7539E+07, 2.7606E+07,
     +  2.7673E+07, 2.7741E+07, 2.7808E+07, 2.7876E+07, 2.7944E+07,
     + 2.8011E+07, 2.8079E+07, 2.8147E+07, 2.8216E+07, 2.8284E+07,
     +  2.8353E+07, 2.8421E+07, 2.8490E+07, 2.8559E+07, 2.8628E+07,
     + 2.8697E+07, 2.8766E+07, 2.8836E+07, 2.8905E+07, 2.8975E+07,
     +  2.9045E+07, 2.9115E+07, 2.9185E+07, 2.9255E+07, 2.9326E+07,
     + 2.9396E+07, 2.9467E+07, 2.9538E+07, 2.9609E+07, 2.9680E+07,
     +  2.9751E+07, 2.9822E+07, 2.9894E+07, 2.9966E+07, 3.0037E+07,
     + 3.0109E+07, 3.0181E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 5,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_CO    (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 9),QofT( 9,NT),Q(NT)
      data (Tmx(I),I=1,9)/9000.,9000.,9000.,9000.,9000.,9000.,9000.,
     + 9000.,9000./
      data xgj/ 1.,2.,1.,6.,2.,12.,1.,1.,6./
c...       CO
c...        --        26
      data (QofT( 1,J),J=1,2408)/ 1.0119E+00, 1.1900E+00, 1.4944E+00,
     +  1.8331E+00, 2.1824E+00, 2.5363E+00, 2.8926E+00, 3.2503E+00,
     + 3.6089E+00, 3.9681E+00, 4.3278E+00, 4.6878E+00, 5.0480E+00,
     +  5.4085E+00, 5.7691E+00, 6.1298E+00, 6.4906E+00, 6.8515E+00,
     + 7.2125E+00, 7.5735E+00, 8.2957E+00, 9.0181E+00, 9.7406E+00,
     +  1.0463E+01, 1.1186E+01, 1.1909E+01, 1.2631E+01, 1.3354E+01,
     + 1.4077E+01, 1.4800E+01, 1.5523E+01, 1.6246E+01, 1.6969E+01,
     +  1.7692E+01, 1.8415E+01, 1.9138E+01, 1.9861E+01, 2.0584E+01,
     + 2.1307E+01, 2.2031E+01, 2.2754E+01, 2.3477E+01, 2.4200E+01,
     +  2.4923E+01, 2.5646E+01, 2.6370E+01, 2.7093E+01, 2.7816E+01,
     + 2.8539E+01, 2.9262E+01, 2.9986E+01, 3.0709E+01, 3.1432E+01,
     +  3.2156E+01, 3.2879E+01, 3.3602E+01, 3.4325E+01, 3.5049E+01,
     + 3.5772E+01, 3.6495E+01, 3.7219E+01, 3.7942E+01, 3.8665E+01,
     +  3.9389E+01, 4.0112E+01, 4.0836E+01, 4.1559E+01, 4.2282E+01,
     + 4.3006E+01, 4.3729E+01, 4.4453E+01, 4.5176E+01, 4.5899E+01,
     +  4.6623E+01, 4.7346E+01, 4.8070E+01, 4.8793E+01, 4.9517E+01,
     + 5.0240E+01, 5.0964E+01, 5.1687E+01, 5.2411E+01, 5.3134E+01,
     +  5.3858E+01, 5.4581E+01, 5.5305E+01, 5.6028E+01, 5.6752E+01,
     + 5.7475E+01, 5.8199E+01, 5.8922E+01, 5.9646E+01, 6.0370E+01,
     +  6.1093E+01, 6.1817E+01, 6.2540E+01, 6.3264E+01, 6.3988E+01,
     + 6.4711E+01, 6.5435E+01, 6.6158E+01, 6.6882E+01, 6.7606E+01,
     +  6.8329E+01, 6.9053E+01, 6.9777E+01, 7.0500E+01, 7.1224E+01,
     + 7.1948E+01, 7.2671E+01, 7.3395E+01, 7.4119E+01, 7.4842E+01,
     +  7.5566E+01, 7.6290E+01, 7.7014E+01, 7.7737E+01, 7.8461E+01,
     + 7.9185E+01, 7.9909E+01, 8.0632E+01, 8.1356E+01, 8.2080E+01,
     +  8.2804E+01, 8.3528E+01, 8.4252E+01, 8.4975E+01, 8.5699E+01,
     + 8.6423E+01, 8.7147E+01, 8.7871E+01, 8.8595E+01, 8.9318E+01,
     +  9.0042E+01, 9.0766E+01, 9.1490E+01, 9.2214E+01, 9.2938E+01,
     + 9.3662E+01, 9.4386E+01, 9.5110E+01, 9.5834E+01, 9.6558E+01,
     +  9.7282E+01, 9.8006E+01, 9.8730E+01, 9.9454E+01, 1.0018E+02,
     + 1.0090E+02, 1.0163E+02, 1.0235E+02, 1.0307E+02, 1.0380E+02,
     +  1.0452E+02, 1.0525E+02, 1.0597E+02, 1.0670E+02, 1.0742E+02,
     + 1.0814E+02, 1.0887E+02, 1.0959E+02, 1.1032E+02, 1.1104E+02,
     +  1.1177E+02, 1.1249E+02, 1.1321E+02, 1.1394E+02, 1.1466E+02,
     + 1.1539E+02, 1.1611E+02, 1.1684E+02, 1.1756E+02, 1.1829E+02,
     +  1.1901E+02, 1.1974E+02, 1.2046E+02, 1.2119E+02, 1.2191E+02,
     + 1.2264E+02, 1.2336E+02, 1.2409E+02, 1.2481E+02, 1.2554E+02,
     +  1.2626E+02, 1.2699E+02, 1.2771E+02, 1.2844E+02, 1.2916E+02,
     + 1.2989E+02, 1.3061E+02, 1.3134E+02, 1.3207E+02, 1.3279E+02,
     +  1.3352E+02, 1.3424E+02, 1.3497E+02, 1.3569E+02, 1.3642E+02,
     + 1.3715E+02, 1.3787E+02, 1.3860E+02, 1.3933E+02, 1.4005E+02,
     +  1.4078E+02, 1.4151E+02, 1.4223E+02, 1.4296E+02, 1.4369E+02,
     + 1.4441E+02, 1.4514E+02, 1.4587E+02, 1.4660E+02, 1.4732E+02,
     +  1.4805E+02, 1.4878E+02, 1.4951E+02, 1.5023E+02, 1.5096E+02,
     + 1.5169E+02, 1.5242E+02, 1.5315E+02, 1.5388E+02, 1.5460E+02,
     +  1.5533E+02, 1.5606E+02, 1.5679E+02, 1.5752E+02, 1.5825E+02,
     + 1.5898E+02, 1.5971E+02, 1.6044E+02, 1.6117E+02, 1.6190E+02,
     +  1.6263E+02, 1.6336E+02, 1.6409E+02, 1.6482E+02, 1.6555E+02,
     + 1.6628E+02, 1.6702E+02, 1.6775E+02, 1.6848E+02, 1.6921E+02,
     +  1.6994E+02, 1.7068E+02, 1.7141E+02, 1.7214E+02, 1.7287E+02,
     + 1.7361E+02, 1.7434E+02, 1.7507E+02, 1.7581E+02, 1.7654E+02,
     +  1.7728E+02, 1.7801E+02, 1.7875E+02, 1.7948E+02, 1.8022E+02,
     + 1.8095E+02, 1.8169E+02, 1.8242E+02, 1.8316E+02, 1.8389E+02,
     +  1.8463E+02, 1.8537E+02, 1.8611E+02, 1.8684E+02, 1.8758E+02,
     + 1.8832E+02, 1.8906E+02, 1.8979E+02, 1.9053E+02, 1.9127E+02,
     +  1.9201E+02, 1.9275E+02, 1.9349E+02, 1.9423E+02, 1.9497E+02,
     + 1.9571E+02, 1.9645E+02, 1.9720E+02, 1.9794E+02, 1.9868E+02,
     +  1.9942E+02, 2.0016E+02, 2.0091E+02, 2.0165E+02, 2.0239E+02,
     + 2.0314E+02, 2.0388E+02, 2.0463E+02, 2.0537E+02, 2.0612E+02,
     +  2.0686E+02, 2.0761E+02, 2.0835E+02, 2.0910E+02, 2.0985E+02,
     + 2.1060E+02, 2.1134E+02, 2.1209E+02, 2.1284E+02, 2.1359E+02,
     +  2.1434E+02, 2.1509E+02, 2.1584E+02, 2.1659E+02, 2.1734E+02,
     + 2.1809E+02, 2.1884E+02, 2.1959E+02, 2.2034E+02, 2.2110E+02,
     +  2.2185E+02, 2.2260E+02, 2.2336E+02, 2.2411E+02, 2.2487E+02,
     + 2.2562E+02, 2.2638E+02, 2.2713E+02, 2.2789E+02, 2.2865E+02,
     +  2.2940E+02, 2.3016E+02, 2.3092E+02, 2.3168E+02, 2.3244E+02,
     + 2.3320E+02, 2.3396E+02, 2.3472E+02, 2.3548E+02, 2.3624E+02,
     +  2.3700E+02, 2.3776E+02, 2.3853E+02, 2.3929E+02, 2.4005E+02,
     + 2.4082E+02, 2.4158E+02, 2.4235E+02, 2.4311E+02, 2.4388E+02,
     +  2.4464E+02, 2.4541E+02, 2.4618E+02, 2.4695E+02, 2.4771E+02,
     + 2.4848E+02, 2.4925E+02, 2.5002E+02, 2.5079E+02, 2.5156E+02,
     +  2.5234E+02, 2.5311E+02, 2.5388E+02, 2.5465E+02, 2.5543E+02,
     + 2.5620E+02, 2.5698E+02, 2.5775E+02, 2.5853E+02, 2.5930E+02,
     +  2.6008E+02, 2.6086E+02, 2.6163E+02, 2.6241E+02, 2.6319E+02,
     + 2.6397E+02, 2.6475E+02, 2.6553E+02, 2.6631E+02, 2.6709E+02,
     +  2.6788E+02, 2.6866E+02, 2.6944E+02, 2.7023E+02, 2.7101E+02,
     + 2.7180E+02, 2.7258E+02, 2.7337E+02, 2.7415E+02, 2.7494E+02,
     +  2.7573E+02, 2.7652E+02, 2.7731E+02, 2.7810E+02, 2.7889E+02,
     + 2.7968E+02, 2.8047E+02, 2.8126E+02, 2.8205E+02, 2.8285E+02,
     +  2.8364E+02, 2.8444E+02, 2.8523E+02, 2.8603E+02, 2.8682E+02,
     + 2.8762E+02, 2.8842E+02, 2.8922E+02, 2.9001E+02, 2.9081E+02,
     +  2.9161E+02, 2.9241E+02, 2.9322E+02, 2.9402E+02, 2.9482E+02,
     + 2.9562E+02, 2.9643E+02, 2.9723E+02, 2.9804E+02, 2.9884E+02,
     +  2.9965E+02, 3.0046E+02, 3.0126E+02, 3.0207E+02, 3.0288E+02,
     + 3.0369E+02, 3.0450E+02, 3.0531E+02, 3.0612E+02, 3.0693E+02,
     +  3.0775E+02, 3.0856E+02, 3.0938E+02, 3.1019E+02, 3.1101E+02,
     + 3.1182E+02, 3.1264E+02, 3.1346E+02, 3.1427E+02, 3.1509E+02,
     +  3.1591E+02, 3.1673E+02, 3.1755E+02, 3.1838E+02, 3.1920E+02,
     + 3.2002E+02, 3.2085E+02, 3.2167E+02, 3.2249E+02, 3.2332E+02,
     +  3.2415E+02, 3.2497E+02, 3.2580E+02, 3.2663E+02, 3.2746E+02,
     + 3.2829E+02, 3.2912E+02, 3.2995E+02, 3.3078E+02, 3.3162E+02,
     +  3.3245E+02, 3.3328E+02, 3.3412E+02, 3.3495E+02, 3.3579E+02,
     + 3.3663E+02, 3.3747E+02, 3.3830E+02, 3.3914E+02, 3.3998E+02,
     +  3.4082E+02, 3.4166E+02, 3.4251E+02, 3.4335E+02, 3.4419E+02,
     + 3.4504E+02, 3.4588E+02, 3.4673E+02, 3.4757E+02, 3.4842E+02,
     +  3.4927E+02, 3.5012E+02, 3.5097E+02, 3.5182E+02, 3.5267E+02,
     + 3.5352E+02, 3.5437E+02, 3.5523E+02, 3.5608E+02, 3.5693E+02,
     +  3.5779E+02, 3.5865E+02, 3.5950E+02, 3.6036E+02, 3.6122E+02,
     + 3.6208E+02, 3.6294E+02, 3.6380E+02, 3.6466E+02, 3.6552E+02,
     +  3.6639E+02, 3.6725E+02, 3.6811E+02, 3.6898E+02, 3.6984E+02,
     + 3.7071E+02, 3.7158E+02, 3.7245E+02, 3.7332E+02, 3.7419E+02,
     +  3.7506E+02, 3.7593E+02, 3.7680E+02, 3.7767E+02, 3.7855E+02,
     + 3.7942E+02, 3.8030E+02, 3.8117E+02, 3.8205E+02, 3.8293E+02,
     +  3.8381E+02, 3.8469E+02, 3.8557E+02, 3.8645E+02, 3.8733E+02,
     + 3.8821E+02, 3.8909E+02, 3.8998E+02, 3.9086E+02, 3.9175E+02,
     +  3.9263E+02, 3.9352E+02, 3.9441E+02, 3.9530E+02, 3.9619E+02,
     + 3.9708E+02, 3.9797E+02, 3.9886E+02, 3.9975E+02, 4.0065E+02,
     +  4.0154E+02, 4.0244E+02, 4.0333E+02, 4.0423E+02, 4.0513E+02,
     + 4.0602E+02, 4.0692E+02, 4.0782E+02, 4.0872E+02, 4.0963E+02,
     +  4.1053E+02, 4.1143E+02, 4.1234E+02, 4.1324E+02, 4.1415E+02,
     + 4.1505E+02, 4.1596E+02, 4.1687E+02, 4.1778E+02, 4.1869E+02,
     +  4.1960E+02, 4.2051E+02, 4.2142E+02, 4.2233E+02, 4.2325E+02,
     + 4.2416E+02, 4.2508E+02, 4.2599E+02, 4.2691E+02, 4.2783E+02,
     +  4.2875E+02, 4.2967E+02, 4.3059E+02, 4.3151E+02, 4.3243E+02,
     + 4.3335E+02, 4.3428E+02, 4.3520E+02, 4.3613E+02, 4.3706E+02,
     +  4.3798E+02, 4.3891E+02, 4.3984E+02, 4.4077E+02, 4.4170E+02,
     + 4.4263E+02, 4.4356E+02, 4.4450E+02, 4.4543E+02, 4.4636E+02,
     +  4.4730E+02, 4.4824E+02, 4.4917E+02, 4.5011E+02, 4.5105E+02,
     + 4.5199E+02, 4.5293E+02, 4.5387E+02, 4.5482E+02, 4.5576E+02,
     +  4.5670E+02, 4.5765E+02, 4.5859E+02, 4.5954E+02, 4.6049E+02,
     + 4.6144E+02, 4.6239E+02, 4.6334E+02, 4.6429E+02, 4.6524E+02,
     +  4.6619E+02, 4.6714E+02, 4.6810E+02, 4.6905E+02, 4.7001E+02,
     + 4.7097E+02, 4.7192E+02, 4.7288E+02, 4.7384E+02, 4.7480E+02,
     +  4.7576E+02, 4.7673E+02, 4.7769E+02, 4.7865E+02, 4.7962E+02,
     + 4.8058E+02, 4.8155E+02, 4.8252E+02, 4.8348E+02, 4.8445E+02,
     +  4.8542E+02, 4.8639E+02, 4.8737E+02, 4.8834E+02, 4.8931E+02,
     + 4.9029E+02, 4.9126E+02, 4.9224E+02, 4.9321E+02, 4.9419E+02,
     +  4.9517E+02, 4.9615E+02, 4.9713E+02, 4.9811E+02, 4.9909E+02,
     + 5.0008E+02, 5.0106E+02, 5.0205E+02, 5.0303E+02, 5.0402E+02,
     +  5.0500E+02, 5.0599E+02, 5.0698E+02, 5.0797E+02, 5.0896E+02,
     + 5.0995E+02, 5.1095E+02, 5.1194E+02, 5.1294E+02, 5.1393E+02,
     +  5.1493E+02, 5.1592E+02, 5.1692E+02, 5.1792E+02, 5.1892E+02,
     + 5.1992E+02, 5.2092E+02, 5.2193E+02, 5.2293E+02, 5.2393E+02,
     +  5.2494E+02, 5.2594E+02, 5.2695E+02, 5.2796E+02, 5.2897E+02,
     + 5.2998E+02, 5.3099E+02, 5.3200E+02, 5.3301E+02, 5.3402E+02,
     +  5.3504E+02, 5.3605E+02, 5.3707E+02, 5.3808E+02, 5.3910E+02,
     + 5.4012E+02, 5.4114E+02, 5.4216E+02, 5.4318E+02, 5.4420E+02,
     +  5.4523E+02, 5.4625E+02, 5.4728E+02, 5.4830E+02, 5.4933E+02,
     + 5.5036E+02, 5.5138E+02, 5.5241E+02, 5.5344E+02, 5.5447E+02,
     +  5.5551E+02, 5.5654E+02, 5.5757E+02, 5.5861E+02, 5.5964E+02,
     + 5.6068E+02, 5.6172E+02, 5.6276E+02, 5.6379E+02, 5.6483E+02,
     +  5.6588E+02, 5.6692E+02, 5.6796E+02, 5.6900E+02, 5.7005E+02,
     + 5.7109E+02, 5.7214E+02, 5.7319E+02, 5.7424E+02, 5.7529E+02,
     +  5.7634E+02, 5.7739E+02, 5.7844E+02, 5.7949E+02, 5.8054E+02,
     + 5.8160E+02, 5.8266E+02, 5.8371E+02, 5.8477E+02, 5.8583E+02,
     +  5.8689E+02, 5.8795E+02, 5.8901E+02, 5.9007E+02, 5.9113E+02,
     + 5.9220E+02, 5.9326E+02, 5.9433E+02, 5.9539E+02, 5.9646E+02,
     +  5.9753E+02, 5.9860E+02, 5.9967E+02, 6.0074E+02, 6.0181E+02,
     + 6.0288E+02, 6.0396E+02, 6.0503E+02, 6.0611E+02, 6.0719E+02,
     +  6.0826E+02, 6.0934E+02, 6.1042E+02, 6.1150E+02, 6.1258E+02,
     + 6.1366E+02, 6.1475E+02, 6.1583E+02, 6.1692E+02, 6.1800E+02,
     +  6.1909E+02, 6.2018E+02, 6.2127E+02, 6.2235E+02, 6.2345E+02,
     + 6.2454E+02, 6.2563E+02, 6.2672E+02, 6.2782E+02, 6.2891E+02,
     +  6.3001E+02, 6.3110E+02, 6.3220E+02, 6.3330E+02, 6.3440E+02,
     + 6.3550E+02, 6.3660E+02, 6.3771E+02, 6.3881E+02, 6.3991E+02,
     +  6.4102E+02, 6.4212E+02, 6.4323E+02, 6.4434E+02, 6.4545E+02,
     + 6.4656E+02, 6.4767E+02, 6.4878E+02, 6.4989E+02, 6.5101E+02,
     +  6.5212E+02, 6.5324E+02, 6.5435E+02, 6.5547E+02, 6.5659E+02,
     + 6.5771E+02, 6.5883E+02, 6.5995E+02, 6.6107E+02, 6.6220E+02,
     +  6.6332E+02, 6.6444E+02, 6.6557E+02, 6.6670E+02, 6.6782E+02,
     + 6.6895E+02, 6.7008E+02, 6.7121E+02, 6.7234E+02, 6.7348E+02,
     +  6.7461E+02, 6.7574E+02, 6.7688E+02, 6.7801E+02, 6.7915E+02,
     + 6.8029E+02, 6.8143E+02, 6.8257E+02, 6.8371E+02, 6.8485E+02,
     +  6.8599E+02, 6.8714E+02, 6.8828E+02, 6.8943E+02, 6.9057E+02,
     + 6.9172E+02, 6.9287E+02, 6.9402E+02, 6.9517E+02, 6.9632E+02,
     +  6.9747E+02, 6.9862E+02, 6.9978E+02, 7.0093E+02, 7.0209E+02,
     + 7.0325E+02, 7.0440E+02, 7.0556E+02, 7.0672E+02, 7.0788E+02,
     +  7.0904E+02, 7.1021E+02, 7.1137E+02, 7.1253E+02, 7.1370E+02,
     + 7.1486E+02, 7.1603E+02, 7.1720E+02, 7.1837E+02, 7.1954E+02,
     +  7.2071E+02, 7.2188E+02, 7.2305E+02, 7.2423E+02, 7.2540E+02,
     + 7.2658E+02, 7.2775E+02, 7.2893E+02, 7.3011E+02, 7.3129E+02,
     +  7.3247E+02, 7.3365E+02, 7.3483E+02, 7.3602E+02, 7.3720E+02,
     + 7.3839E+02, 7.3957E+02, 7.4076E+02, 7.4195E+02, 7.4314E+02,
     +  7.4433E+02, 7.4552E+02, 7.4671E+02, 7.4790E+02, 7.4909E+02,
     + 7.5029E+02, 7.5148E+02, 7.5268E+02, 7.5388E+02, 7.5508E+02,
     +  7.5628E+02, 7.5748E+02, 7.5868E+02, 7.5988E+02, 7.6108E+02,
     + 7.6229E+02, 7.6349E+02, 7.6470E+02, 7.6590E+02, 7.6711E+02,
     +  7.6832E+02, 7.6953E+02, 7.7074E+02, 7.7195E+02, 7.7317E+02,
     + 7.7438E+02, 7.7559E+02, 7.7681E+02, 7.7803E+02, 7.7924E+02,
     +  7.8046E+02, 7.8168E+02, 7.8290E+02, 7.8412E+02, 7.8535E+02,
     + 7.8657E+02, 7.8779E+02, 7.8902E+02, 7.9024E+02, 7.9147E+02,
     +  7.9270E+02, 7.9393E+02, 7.9516E+02, 7.9639E+02, 7.9762E+02,
     + 7.9885E+02, 8.0009E+02, 8.0132E+02, 8.0256E+02, 8.0379E+02,
     +  8.0503E+02, 8.0627E+02, 8.0751E+02, 8.0875E+02, 8.0999E+02,
     + 8.1123E+02, 8.1248E+02, 8.1372E+02, 8.1496E+02, 8.1621E+02,
     +  8.1746E+02, 8.1871E+02, 8.1996E+02, 8.2120E+02, 8.2246E+02,
     + 8.2371E+02, 8.2496E+02, 8.2621E+02, 8.2747E+02, 8.2872E+02,
     +  8.2998E+02, 8.3124E+02, 8.3250E+02, 8.3376E+02, 8.3502E+02,
     + 8.3628E+02, 8.3754E+02, 8.3881E+02, 8.4007E+02, 8.4134E+02,
     +  8.4260E+02, 8.4387E+02, 8.4514E+02, 8.4641E+02, 8.4768E+02,
     + 8.4895E+02, 8.5022E+02, 8.5149E+02, 8.5277E+02, 8.5404E+02,
     +  8.5532E+02, 8.5659E+02, 8.5787E+02, 8.5915E+02, 8.6043E+02,
     + 8.6171E+02, 8.6299E+02, 8.6428E+02, 8.6556E+02, 8.6684E+02,
     +  8.6813E+02, 8.6942E+02, 8.7070E+02, 8.7199E+02, 8.7328E+02,
     + 8.7457E+02, 8.7586E+02, 8.7716E+02, 8.7845E+02, 8.7974E+02,
     +  8.8104E+02, 8.8234E+02, 8.8363E+02, 8.8493E+02, 8.8623E+02,
     + 8.8753E+02, 8.8883E+02, 8.9013E+02, 8.9144E+02, 8.9274E+02,
     +  8.9404E+02, 8.9535E+02, 8.9666E+02, 8.9797E+02, 8.9927E+02,
     + 9.0058E+02, 9.0189E+02, 9.0321E+02, 9.0452E+02, 9.0583E+02,
     +  9.0715E+02, 9.0846E+02, 9.0978E+02, 9.1110E+02, 9.1242E+02,
     + 9.1373E+02, 9.1506E+02, 9.1638E+02, 9.1770E+02, 9.1902E+02,
     +  9.2035E+02, 9.2167E+02, 9.2300E+02, 9.2433E+02, 9.2565E+02,
     + 9.2698E+02, 9.2831E+02, 9.2964E+02, 9.3098E+02, 9.3231E+02,
     +  9.3364E+02, 9.3498E+02, 9.3631E+02, 9.3765E+02, 9.3899E+02,
     + 9.4033E+02, 9.4167E+02, 9.4301E+02, 9.4435E+02, 9.4569E+02,
     +  9.4704E+02, 9.4838E+02, 9.4973E+02, 9.5107E+02, 9.5242E+02,
     + 9.5377E+02, 9.5512E+02, 9.5647E+02, 9.5782E+02, 9.5917E+02,
     +  9.6053E+02, 9.6188E+02, 9.6324E+02, 9.6459E+02, 9.6595E+02,
     + 9.6731E+02, 9.6867E+02, 9.7003E+02, 9.7139E+02, 9.7275E+02,
     +  9.7412E+02, 9.7548E+02, 9.7685E+02, 9.7821E+02, 9.7958E+02,
     + 9.8095E+02, 9.8232E+02, 9.8369E+02, 9.8506E+02, 9.8643E+02,
     +  9.8780E+02, 9.8918E+02, 9.9055E+02, 9.9193E+02, 9.9331E+02,
     + 9.9468E+02, 9.9606E+02, 9.9744E+02, 9.9882E+02, 1.0002E+03,
     +  1.0016E+03, 1.0030E+03, 1.0044E+03, 1.0057E+03, 1.0071E+03,
     + 1.0085E+03, 1.0099E+03, 1.0113E+03, 1.0127E+03, 1.0141E+03,
     +  1.0155E+03, 1.0169E+03, 1.0183E+03, 1.0197E+03, 1.0210E+03,
     + 1.0224E+03, 1.0238E+03, 1.0252E+03, 1.0266E+03, 1.0280E+03,
     +  1.0294E+03, 1.0309E+03, 1.0323E+03, 1.0337E+03, 1.0351E+03,
     + 1.0365E+03, 1.0379E+03, 1.0393E+03, 1.0407E+03, 1.0421E+03,
     +  1.0435E+03, 1.0449E+03, 1.0464E+03, 1.0478E+03, 1.0492E+03,
     + 1.0506E+03, 1.0520E+03, 1.0534E+03, 1.0549E+03, 1.0563E+03,
     +  1.0577E+03, 1.0591E+03, 1.0605E+03, 1.0620E+03, 1.0634E+03,
     + 1.0648E+03, 1.0663E+03, 1.0677E+03, 1.0691E+03, 1.0705E+03,
     +  1.0720E+03, 1.0734E+03, 1.0748E+03, 1.0763E+03, 1.0777E+03,
     + 1.0792E+03, 1.0806E+03, 1.0820E+03, 1.0835E+03, 1.0849E+03,
     +  1.0863E+03, 1.0878E+03, 1.0892E+03, 1.0907E+03, 1.0921E+03,
     + 1.0936E+03, 1.0950E+03, 1.0965E+03, 1.0979E+03, 1.0994E+03,
     +  1.1008E+03, 1.1023E+03, 1.1037E+03, 1.1052E+03, 1.1066E+03,
     + 1.1081E+03, 1.1096E+03, 1.1110E+03, 1.1125E+03, 1.1139E+03,
     +  1.1154E+03, 1.1169E+03, 1.1183E+03, 1.1198E+03, 1.1212E+03,
     + 1.1227E+03, 1.1242E+03, 1.1256E+03, 1.1271E+03, 1.1286E+03,
     +  1.1301E+03, 1.1315E+03, 1.1330E+03, 1.1345E+03, 1.1360E+03,
     + 1.1374E+03, 1.1389E+03, 1.1404E+03, 1.1419E+03, 1.1433E+03,
     +  1.1448E+03, 1.1463E+03, 1.1478E+03, 1.1493E+03, 1.1508E+03,
     + 1.1522E+03, 1.1537E+03, 1.1552E+03, 1.1567E+03, 1.1582E+03,
     +  1.1597E+03, 1.1612E+03, 1.1627E+03, 1.1642E+03, 1.1657E+03,
     + 1.1672E+03, 1.1687E+03, 1.1702E+03, 1.1717E+03, 1.1732E+03,
     +  1.1747E+03, 1.1762E+03, 1.1777E+03, 1.1792E+03, 1.1807E+03,
     + 1.1822E+03, 1.1837E+03, 1.1852E+03, 1.1867E+03, 1.1882E+03,
     +  1.1897E+03, 1.1912E+03, 1.1927E+03, 1.1943E+03, 1.1958E+03,
     + 1.1973E+03, 1.1988E+03, 1.2003E+03, 1.2018E+03, 1.2034E+03,
     +  1.2049E+03, 1.2064E+03, 1.2079E+03, 1.2095E+03, 1.2110E+03,
     + 1.2125E+03, 1.2140E+03, 1.2156E+03, 1.2171E+03, 1.2186E+03,
     +  1.2201E+03, 1.2217E+03, 1.2232E+03, 1.2247E+03, 1.2263E+03,
     + 1.2278E+03, 1.2294E+03, 1.2309E+03, 1.2324E+03, 1.2340E+03,
     +  1.2355E+03, 1.2370E+03, 1.2386E+03, 1.2401E+03, 1.2417E+03,
     + 1.2432E+03, 1.2448E+03, 1.2463E+03, 1.2479E+03, 1.2494E+03,
     +  1.2510E+03, 1.2525E+03, 1.2541E+03, 1.2556E+03, 1.2572E+03,
     + 1.2587E+03, 1.2603E+03, 1.2618E+03, 1.2634E+03, 1.2650E+03,
     +  1.2665E+03, 1.2681E+03, 1.2697E+03, 1.2712E+03, 1.2728E+03,
     + 1.2743E+03, 1.2759E+03, 1.2775E+03, 1.2790E+03, 1.2806E+03,
     +  1.2822E+03, 1.2838E+03, 1.2853E+03, 1.2869E+03, 1.2885E+03,
     + 1.2901E+03, 1.2916E+03, 1.2932E+03, 1.2948E+03, 1.2964E+03,
     +  1.2979E+03, 1.2995E+03, 1.3011E+03, 1.3027E+03, 1.3043E+03,
     + 1.3059E+03, 1.3075E+03, 1.3090E+03, 1.3106E+03, 1.3122E+03,
     +  1.3138E+03, 1.3154E+03, 1.3170E+03, 1.3186E+03, 1.3202E+03,
     + 1.3218E+03, 1.3234E+03, 1.3250E+03, 1.3266E+03, 1.3282E+03,
     +  1.3298E+03, 1.3314E+03, 1.3330E+03, 1.3346E+03, 1.3362E+03,
     + 1.3378E+03, 1.3394E+03, 1.3410E+03, 1.3426E+03, 1.3442E+03,
     +  1.3458E+03, 1.3474E+03, 1.3490E+03, 1.3507E+03, 1.3523E+03,
     + 1.3539E+03, 1.3555E+03, 1.3571E+03, 1.3587E+03, 1.3604E+03,
     +  1.3620E+03, 1.3636E+03, 1.3652E+03, 1.3669E+03, 1.3685E+03,
     + 1.3701E+03, 1.3717E+03, 1.3734E+03, 1.3750E+03, 1.3766E+03,
     +  1.3782E+03, 1.3799E+03, 1.3815E+03, 1.3831E+03, 1.3848E+03,
     + 1.3864E+03, 1.3880E+03, 1.3897E+03, 1.3913E+03, 1.3930E+03,
     +  1.3946E+03, 1.3962E+03, 1.3979E+03, 1.3995E+03, 1.4012E+03,
     + 1.4028E+03, 1.4045E+03, 1.4061E+03, 1.4078E+03, 1.4094E+03,
     +  1.4111E+03, 1.4127E+03, 1.4144E+03, 1.4160E+03, 1.4177E+03,
     + 1.4193E+03, 1.4210E+03, 1.4226E+03, 1.4243E+03, 1.4260E+03,
     +  1.4276E+03, 1.4293E+03, 1.4309E+03, 1.4326E+03, 1.4343E+03,
     + 1.4359E+03, 1.4376E+03, 1.4393E+03, 1.4409E+03, 1.4426E+03,
     +  1.4443E+03, 1.4460E+03, 1.4476E+03, 1.4493E+03, 1.4510E+03,
     + 1.4527E+03, 1.4543E+03, 1.4560E+03, 1.4577E+03, 1.4594E+03,
     +  1.4610E+03, 1.4627E+03, 1.4644E+03, 1.4661E+03, 1.4678E+03,
     + 1.4695E+03, 1.4711E+03, 1.4728E+03, 1.4745E+03, 1.4762E+03,
     +  1.4779E+03, 1.4796E+03, 1.4813E+03, 1.4830E+03, 1.4847E+03,
     + 1.4864E+03, 1.4881E+03, 1.4898E+03, 1.4915E+03, 1.4932E+03,
     +  1.4949E+03, 1.4966E+03, 1.4983E+03, 1.5000E+03, 1.5017E+03,
     + 1.5034E+03, 1.5051E+03, 1.5068E+03, 1.5085E+03, 1.5102E+03,
     +  1.5119E+03, 1.5136E+03, 1.5154E+03, 1.5171E+03, 1.5188E+03,
     + 1.5205E+03, 1.5222E+03, 1.5239E+03, 1.5257E+03, 1.5274E+03,
     +  1.5291E+03, 1.5308E+03, 1.5325E+03, 1.5343E+03, 1.5360E+03,
     + 1.5377E+03, 1.5394E+03, 1.5412E+03, 1.5429E+03, 1.5446E+03,
     +  1.5464E+03, 1.5481E+03, 1.5498E+03, 1.5516E+03, 1.5533E+03,
     + 1.5550E+03, 1.5568E+03, 1.5585E+03, 1.5602E+03, 1.5620E+03,
     +  1.5637E+03, 1.5655E+03, 1.5672E+03, 1.5690E+03, 1.5707E+03,
     + 1.5724E+03, 1.5742E+03, 1.5759E+03, 1.5777E+03, 1.5794E+03,
     +  1.5812E+03, 1.5829E+03, 1.5847E+03, 1.5864E+03, 1.5882E+03,
     + 1.5900E+03, 1.5917E+03, 1.5935E+03, 1.5952E+03, 1.5970E+03,
     +  1.5988E+03, 1.6005E+03, 1.6023E+03, 1.6040E+03, 1.6058E+03,
     + 1.6076E+03, 1.6093E+03, 1.6111E+03, 1.6129E+03, 1.6147E+03,
     +  1.6164E+03, 1.6182E+03, 1.6200E+03, 1.6217E+03, 1.6235E+03,
     + 1.6253E+03, 1.6271E+03, 1.6289E+03, 1.6306E+03, 1.6324E+03,
     +  1.6342E+03, 1.6360E+03, 1.6378E+03, 1.6395E+03, 1.6413E+03,
     + 1.6431E+03, 1.6449E+03, 1.6467E+03, 1.6485E+03, 1.6503E+03,
     +  1.6521E+03, 1.6539E+03, 1.6557E+03, 1.6574E+03, 1.6592E+03,
     + 1.6610E+03, 1.6628E+03, 1.6646E+03, 1.6664E+03, 1.6682E+03,
     +  1.6700E+03, 1.6718E+03, 1.6736E+03, 1.6755E+03, 1.6773E+03,
     + 1.6791E+03, 1.6809E+03, 1.6827E+03, 1.6845E+03, 1.6863E+03,
     +  1.6881E+03, 1.6899E+03, 1.6917E+03, 1.6936E+03, 1.6954E+03,
     + 1.6972E+03, 1.6990E+03, 1.7008E+03, 1.7026E+03, 1.7045E+03,
     +  1.7063E+03, 1.7081E+03, 1.7099E+03, 1.7118E+03, 1.7136E+03,
     + 1.7154E+03, 1.7172E+03, 1.7191E+03, 1.7209E+03, 1.7227E+03,
     +  1.7246E+03, 1.7264E+03, 1.7282E+03, 1.7301E+03, 1.7319E+03,
     + 1.7337E+03, 1.7356E+03, 1.7374E+03, 1.7393E+03, 1.7411E+03,
     +  1.7429E+03, 1.7448E+03, 1.7466E+03, 1.7485E+03, 1.7503E+03,
     + 1.7522E+03, 1.7540E+03, 1.7559E+03, 1.7577E+03, 1.7596E+03,
     +  1.7614E+03, 1.7633E+03, 1.7651E+03, 1.7670E+03, 1.7688E+03,
     + 1.7707E+03, 1.7726E+03, 1.7744E+03, 1.7763E+03, 1.7781E+03,
     +  1.7800E+03, 1.7819E+03, 1.7837E+03, 1.7856E+03, 1.7875E+03,
     + 1.7893E+03, 1.7912E+03, 1.7931E+03, 1.7949E+03, 1.7968E+03,
     +  1.7987E+03, 1.8006E+03, 1.8024E+03, 1.8043E+03, 1.8062E+03,
     + 1.8081E+03, 1.8100E+03, 1.8118E+03, 1.8137E+03, 1.8156E+03,
     +  1.8175E+03, 1.8194E+03, 1.8212E+03, 1.8231E+03, 1.8250E+03,
     + 1.8269E+03, 1.8288E+03, 1.8307E+03, 1.8326E+03, 1.8345E+03,
     +  1.8364E+03, 1.8383E+03, 1.8402E+03, 1.8421E+03, 1.8440E+03,
     + 1.8459E+03, 1.8478E+03, 1.8497E+03, 1.8516E+03, 1.8535E+03,
     +  1.8554E+03, 1.8573E+03, 1.8592E+03, 1.8611E+03, 1.8630E+03,
     + 1.8649E+03, 1.8668E+03, 1.8687E+03, 1.8706E+03, 1.8725E+03,
     +  1.8745E+03, 1.8764E+03, 1.8783E+03, 1.8802E+03, 1.8821E+03,
     + 1.8840E+03, 1.8860E+03, 1.8879E+03, 1.8898E+03, 1.8917E+03,
     +  1.8936E+03, 1.8956E+03, 1.8975E+03, 1.8994E+03, 1.9014E+03,
     + 1.9033E+03, 1.9052E+03, 1.9071E+03, 1.9091E+03, 1.9110E+03,
     +  1.9129E+03, 1.9149E+03, 1.9168E+03, 1.9188E+03, 1.9207E+03,
     + 1.9226E+03, 1.9246E+03, 1.9265E+03, 1.9285E+03, 1.9304E+03,
     +  1.9323E+03, 1.9343E+03, 1.9362E+03, 1.9382E+03, 1.9401E+03,
     + 1.9421E+03, 1.9440E+03, 1.9460E+03, 1.9479E+03, 1.9499E+03,
     +  1.9518E+03, 1.9538E+03, 1.9558E+03, 1.9577E+03, 1.9597E+03,
     + 1.9616E+03, 1.9636E+03, 1.9656E+03, 1.9675E+03, 1.9695E+03,
     +  1.9715E+03, 1.9734E+03, 1.9754E+03, 1.9774E+03, 1.9793E+03,
     + 1.9813E+03, 1.9833E+03, 1.9852E+03, 1.9872E+03, 1.9892E+03,
     +  1.9912E+03, 1.9931E+03, 1.9951E+03, 1.9971E+03, 1.9991E+03,
     + 2.0011E+03, 2.0030E+03, 2.0050E+03, 2.0070E+03, 2.0090E+03,
     +  2.0110E+03, 2.0130E+03, 2.0149E+03, 2.0169E+03, 2.0189E+03,
     + 2.0209E+03, 2.0229E+03, 2.0249E+03, 2.0269E+03, 2.0289E+03,
     +  2.0309E+03, 2.0329E+03, 2.0349E+03, 2.0369E+03, 2.0389E+03,
     + 2.0409E+03, 2.0429E+03, 2.0449E+03, 2.0469E+03, 2.0489E+03,
     +  2.0509E+03, 2.0529E+03, 2.0549E+03, 2.0569E+03, 2.0589E+03,
     + 2.0609E+03, 2.0630E+03, 2.0650E+03, 2.0670E+03, 2.0690E+03,
     +  2.0710E+03, 2.0730E+03, 2.0751E+03, 2.0771E+03, 2.0791E+03,
     + 2.0811E+03, 2.0831E+03, 2.0852E+03, 2.0872E+03, 2.0892E+03,
     +  2.0912E+03, 2.0933E+03, 2.0953E+03, 2.0973E+03, 2.0994E+03,
     + 2.1014E+03, 2.1034E+03, 2.1055E+03, 2.1075E+03, 2.1095E+03,
     +  2.1116E+03, 2.1136E+03, 2.1156E+03, 2.1177E+03, 2.1197E+03,
     + 2.1218E+03, 2.1238E+03, 2.1259E+03, 2.1279E+03, 2.1299E+03,
     +  2.1320E+03, 2.1340E+03, 2.1361E+03, 2.1381E+03, 2.1402E+03,
     + 2.1422E+03, 2.1443E+03, 2.1464E+03, 2.1484E+03, 2.1505E+03,
     +  2.1525E+03, 2.1546E+03, 2.1566E+03, 2.1587E+03, 2.1608E+03,
     + 2.1628E+03, 2.1649E+03, 2.1670E+03, 2.1690E+03, 2.1711E+03,
     +  2.1732E+03, 2.1752E+03, 2.1773E+03, 2.1794E+03, 2.1814E+03,
     + 2.1835E+03, 2.1856E+03, 2.1877E+03, 2.1897E+03, 2.1918E+03,
     +  2.1939E+03, 2.1960E+03, 2.1981E+03, 2.2001E+03, 2.2022E+03,
     + 2.2043E+03, 2.2064E+03, 2.2085E+03, 2.2106E+03, 2.2126E+03,
     +  2.2147E+03, 2.2168E+03, 2.2189E+03, 2.2210E+03, 2.2231E+03,
     + 2.2252E+03, 2.2273E+03, 2.2294E+03, 2.2315E+03, 2.2336E+03,
     +  2.2357E+03, 2.2378E+03, 2.2399E+03, 2.2420E+03, 2.2441E+03,
     + 2.2462E+03, 2.2483E+03, 2.2504E+03, 2.2525E+03, 2.2546E+03,
     +  2.2567E+03, 2.2588E+03, 2.2610E+03, 2.2631E+03, 2.2652E+03,
     + 2.2673E+03, 2.2694E+03, 2.2715E+03, 2.2736E+03, 2.2758E+03,
     +  2.2779E+03, 2.2800E+03, 2.2821E+03, 2.2842E+03, 2.2864E+03,
     + 2.2885E+03, 2.2906E+03, 2.2927E+03, 2.2949E+03, 2.2970E+03,
     +  2.2991E+03, 2.3013E+03, 2.3034E+03, 2.3055E+03, 2.3077E+03,
     + 2.3098E+03, 2.3119E+03, 2.3141E+03, 2.3162E+03, 2.3184E+03,
     +  2.3205E+03, 2.3226E+03, 2.3248E+03, 2.3269E+03, 2.3291E+03,
     + 2.3312E+03, 2.3334E+03, 2.3355E+03, 2.3377E+03, 2.3398E+03,
     +  2.3420E+03, 2.3441E+03, 2.3463E+03, 2.3484E+03, 2.3506E+03,
     + 2.3527E+03, 2.3549E+03, 2.3570E+03, 2.3592E+03, 2.3614E+03,
     +  2.3635E+03, 2.3657E+03, 2.3679E+03, 2.3700E+03, 2.3722E+03,
     + 2.3743E+03, 2.3765E+03, 2.3787E+03, 2.3809E+03, 2.3830E+03,
     +  2.3852E+03, 2.3874E+03, 2.3895E+03, 2.3917E+03, 2.3939E+03,
     + 2.3961E+03, 2.3982E+03, 2.4004E+03, 2.4026E+03, 2.4048E+03,
     +  2.4070E+03, 2.4092E+03, 2.4113E+03, 2.4135E+03, 2.4157E+03,
     + 2.4179E+03, 2.4201E+03, 2.4223E+03, 2.4245E+03, 2.4267E+03,
     +  2.4289E+03, 2.4310E+03, 2.4332E+03, 2.4354E+03, 2.4376E+03,
     + 2.4398E+03, 2.4420E+03, 2.4442E+03, 2.4464E+03, 2.4486E+03,
     +  2.4508E+03, 2.4530E+03, 2.4552E+03, 2.4574E+03, 2.4597E+03,
     + 2.4619E+03, 2.4641E+03, 2.4663E+03, 2.4685E+03, 2.4707E+03,
     +  2.4729E+03, 2.4751E+03, 2.4774E+03, 2.4796E+03, 2.4818E+03,
     + 2.4840E+03, 2.4862E+03, 2.4884E+03, 2.4907E+03, 2.4929E+03,
     +  2.4951E+03, 2.4973E+03, 2.4996E+03, 2.5018E+03, 2.5040E+03,
     + 2.5062E+03, 2.5085E+03, 2.5107E+03, 2.5129E+03, 2.5152E+03,
     +  2.5174E+03, 2.5196E+03, 2.5219E+03, 2.5241E+03, 2.5264E+03,
     + 2.5286E+03, 2.5308E+03, 2.5331E+03, 2.5353E+03, 2.5376E+03,
     +  2.5398E+03, 2.5421E+03, 2.5443E+03, 2.5465E+03, 2.5488E+03,
     + 2.5510E+03, 2.5533E+03, 2.5556E+03, 2.5578E+03, 2.5601E+03,
     +  2.5623E+03, 2.5646E+03, 2.5668E+03, 2.5691E+03, 2.5713E+03,
     + 2.5736E+03, 2.5759E+03, 2.5781E+03, 2.5804E+03, 2.5827E+03,
     +  2.5849E+03, 2.5872E+03, 2.5895E+03, 2.5917E+03, 2.5940E+03,
     + 2.5963E+03, 2.5985E+03, 2.6008E+03, 2.6031E+03, 2.6054E+03,
     +  2.6076E+03, 2.6099E+03, 2.6122E+03, 2.6145E+03, 2.6168E+03,
     + 2.6190E+03, 2.6213E+03, 2.6236E+03, 2.6259E+03, 2.6282E+03,
     +  2.6305E+03, 2.6327E+03, 2.6350E+03, 2.6373E+03, 2.6396E+03,
     + 2.6419E+03, 2.6442E+03, 2.6465E+03, 2.6488E+03, 2.6511E+03,
     +  2.6534E+03, 2.6557E+03, 2.6580E+03, 2.6603E+03, 2.6626E+03,
     + 2.6649E+03, 2.6672E+03, 2.6695E+03, 2.6718E+03, 2.6741E+03,
     +  2.6764E+03, 2.6787E+03, 2.6810E+03, 2.6833E+03, 2.6856E+03,
     + 2.6880E+03, 2.6903E+03, 2.6926E+03, 2.6949E+03, 2.6972E+03,
     +  2.6995E+03, 2.7019E+03, 2.7042E+03, 2.7065E+03, 2.7088E+03,
     + 2.7111E+03, 2.7135E+03, 2.7158E+03, 2.7181E+03, 2.7205E+03,
     +  2.7228E+03, 2.7251E+03, 2.7274E+03, 2.7298E+03, 2.7321E+03,
     + 2.7344E+03, 2.7368E+03, 2.7391E+03, 2.7414E+03, 2.7438E+03,
     +  2.7461E+03, 2.7485E+03, 2.7508E+03, 2.7531E+03, 2.7555E+03,
     + 2.7578E+03, 2.7602E+03, 2.7625E+03, 2.7649E+03, 2.7672E+03,
     +  2.7696E+03, 2.7719E+03, 2.7743E+03, 2.7766E+03, 2.7790E+03,
     + 2.7813E+03, 2.7837E+03, 2.7861E+03, 2.7884E+03, 2.7908E+03,
     +  2.7931E+03, 2.7955E+03, 2.7979E+03, 2.8002E+03, 2.8026E+03,
     + 2.8049E+03, 2.8073E+03, 2.8097E+03, 2.8120E+03, 2.8144E+03,
     +  2.8168E+03, 2.8192E+03, 2.8215E+03, 2.8239E+03, 2.8263E+03,
     + 2.8287E+03, 2.8310E+03, 2.8334E+03, 2.8358E+03, 2.8382E+03,
     +  2.8406E+03, 2.8429E+03, 2.8453E+03, 2.8477E+03, 2.8501E+03,
     + 2.8525E+03, 2.8549E+03, 2.8572E+03, 2.8596E+03, 2.8620E+03,
     +  2.8644E+03, 2.8668E+03, 2.8692E+03, 2.8716E+03, 2.8740E+03,
     + 2.8764E+03, 2.8788E+03, 2.8812E+03, 2.8836E+03, 2.8860E+03,
     +  2.8884E+03, 2.8908E+03, 2.8932E+03, 2.8956E+03, 2.8980E+03,
     + 2.9004E+03, 2.9028E+03, 2.9052E+03, 2.9077E+03, 2.9101E+03,
     +  2.9125E+03, 2.9149E+03, 2.9173E+03, 2.9197E+03, 2.9221E+03,
     + 2.9246E+03, 2.9270E+03, 2.9294E+03, 2.9318E+03, 2.9342E+03,
     +  2.9367E+03, 2.9391E+03, 2.9415E+03, 2.9439E+03, 2.9464E+03,
     + 2.9488E+03, 2.9512E+03, 2.9537E+03, 2.9561E+03, 2.9585E+03,
     +  2.9610E+03, 2.9634E+03, 2.9658E+03, 2.9683E+03, 2.9707E+03,
     + 2.9731E+03, 2.9756E+03, 2.9780E+03, 2.9805E+03, 2.9829E+03,
     +  2.9854E+03, 2.9878E+03, 2.9902E+03, 2.9927E+03, 2.9951E+03,
     + 2.9976E+03, 3.0000E+03, 3.0025E+03, 3.0049E+03, 3.0074E+03,
     +  3.0099E+03, 3.0123E+03, 3.0148E+03, 3.0172E+03, 3.0197E+03,
     + 3.0222E+03, 3.0246E+03, 3.0271E+03, 3.0295E+03, 3.0320E+03,
     +  3.0345E+03, 3.0369E+03, 3.0394E+03, 3.0419E+03, 3.0443E+03,
     + 3.0468E+03, 3.0493E+03, 3.0518E+03, 3.0542E+03, 3.0567E+03,
     +  3.0592E+03, 3.0617E+03, 3.0641E+03, 3.0666E+03, 3.0691E+03,
     + 3.0716E+03, 3.0741E+03, 3.0765E+03, 3.0790E+03, 3.0815E+03,
     +  3.0840E+03, 3.0865E+03, 3.0890E+03, 3.0915E+03, 3.0940E+03,
     + 3.0965E+03, 3.0990E+03, 3.1014E+03, 3.1039E+03, 3.1064E+03,
     +  3.1089E+03, 3.1114E+03, 3.1139E+03, 3.1164E+03, 3.1189E+03,
     + 3.1214E+03, 3.1239E+03, 3.1264E+03, 3.1290E+03, 3.1315E+03,
     +  3.1340E+03, 3.1365E+03, 3.1390E+03, 3.1415E+03, 3.1440E+03,
     + 3.1465E+03, 3.1490E+03, 3.1516E+03, 3.1541E+03, 3.1566E+03,
     +  3.1591E+03, 3.1616E+03, 3.1642E+03, 3.1667E+03, 3.1692E+03,
     + 3.1717E+03, 3.1742E+03, 3.1768E+03, 3.1793E+03, 3.1818E+03,
     +  3.1844E+03, 3.1869E+03, 3.1894E+03, 3.1920E+03, 3.1945E+03,
     + 3.1970E+03, 3.1996E+03, 3.2021E+03, 3.2046E+03, 3.2072E+03,
     +  3.2097E+03, 3.2123E+03, 3.2148E+03, 3.2173E+03, 3.2199E+03,
     + 3.2224E+03, 3.2250E+03, 3.2275E+03, 3.2301E+03, 3.2326E+03,
     +  3.2352E+03, 3.2377E+03, 3.2403E+03, 3.2428E+03, 3.2454E+03,
     + 3.2479E+03, 3.2505E+03, 3.2531E+03, 3.2556E+03, 3.2582E+03,
     +  3.2607E+03, 3.2633E+03, 3.2659E+03, 3.2684E+03, 3.2710E+03,
     + 3.2736E+03, 3.2761E+03, 3.2787E+03, 3.2813E+03, 3.2838E+03,
     +  3.2864E+03, 3.2890E+03, 3.2916E+03, 3.2941E+03, 3.2967E+03,
     + 3.2993E+03, 3.3019E+03, 3.3044E+03, 3.3070E+03, 3.3096E+03,
     +  3.3122E+03, 3.3148E+03, 3.3174E+03, 3.3199E+03, 3.3225E+03,
     + 3.3251E+03, 3.3277E+03, 3.3303E+03, 3.3329E+03, 3.3355E+03,
     +  3.3381E+03, 3.3407E+03, 3.3433E+03, 3.3459E+03, 3.3485E+03,
     + 3.3511E+03, 3.3537E+03, 3.3563E+03, 3.3589E+03, 3.3615E+03,
     +  3.3641E+03, 3.3667E+03, 3.3693E+03, 3.3719E+03, 3.3745E+03,
     + 3.3771E+03, 3.3797E+03, 3.3823E+03, 3.3849E+03, 3.3876E+03,
     +  3.3902E+03, 3.3928E+03, 3.3954E+03, 3.3980E+03, 3.4006E+03,
     + 3.4033E+03, 3.4059E+03, 3.4085E+03, 3.4111E+03, 3.4137E+03,
     +  3.4164E+03, 3.4190E+03, 3.4216E+03, 3.4243E+03, 3.4269E+03,
     + 3.4295E+03, 3.4321E+03, 3.4348E+03, 3.4374E+03, 3.4401E+03,
     +  3.4427E+03, 3.4453E+03, 3.4480E+03, 3.4506E+03, 3.4532E+03,
     + 3.4559E+03, 3.4585E+03, 3.4612E+03, 3.4638E+03, 3.4665E+03,
     +  3.4691E+03, 3.4718E+03, 3.4744E+03, 3.4771E+03, 3.4797E+03,
     + 3.4824E+03, 3.4850E+03, 3.4877E+03, 3.4903E+03, 3.4930E+03,
     +  3.4956E+03, 3.4983E+03, 3.5010E+03, 3.5036E+03, 3.5063E+03,
     + 3.5089E+03, 3.5116E+03, 3.5143E+03, 3.5169E+03, 3.5196E+03,
     +  3.5223E+03, 3.5249E+03, 3.5276E+03, 3.5303E+03, 3.5330E+03,
     + 3.5356E+03, 3.5383E+03, 3.5410E+03, 3.5437E+03, 3.5463E+03,
     +  3.5490E+03, 3.5517E+03, 3.5544E+03, 3.5571E+03, 3.5597E+03,
     + 3.5624E+03, 3.5651E+03, 3.5678E+03, 3.5705E+03, 3.5732E+03,
     +  3.5759E+03, 3.5786E+03, 3.5812E+03, 3.5839E+03, 3.5866E+03,
     + 3.5893E+03, 3.5920E+03, 3.5947E+03, 3.5974E+03, 3.6001E+03,
     +  3.6028E+03, 3.6055E+03, 3.6082E+03, 3.6109E+03, 3.6136E+03,
     + 3.6163E+03, 3.6190E+03, 3.6218E+03, 3.6245E+03, 3.6272E+03,
     +  3.6299E+03, 3.6326E+03, 3.6353E+03, 3.6380E+03, 3.6407E+03,
     + 3.6435E+03, 3.6462E+03, 3.6489E+03, 3.6516E+03, 3.6543E+03,
     +  3.6571E+03, 3.6598E+03, 3.6625E+03, 3.6652E+03, 3.6680E+03,
     + 3.6707E+03, 3.6734E+03, 3.6761E+03, 3.6789E+03, 3.6816E+03,
     +  3.6843E+03, 3.6871E+03, 3.6898E+03, 3.6925E+03, 3.6953E+03,
     + 3.6980E+03, 3.7008E+03, 3.7035E+03, 3.7062E+03, 3.7090E+03,
     +  3.7117E+03, 3.7145E+03, 3.7172E+03, 3.7200E+03, 3.7227E+03,
     + 3.7255E+03, 3.7282E+03, 3.7310E+03, 3.7337E+03, 3.7365E+03,
     +  3.7392E+03, 3.7420E+03, 3.7447E+03, 3.7475E+03, 3.7503E+03,
     + 3.7530E+03, 3.7558E+03, 3.7585E+03, 3.7613E+03, 3.7641E+03,
     +  3.7668E+03, 3.7696E+03, 3.7724E+03, 3.7751E+03, 3.7779E+03/
      data (QofT( 1,J),J=2409,4510)/ 3.7807E+03, 3.7834E+03,
     +  3.7862E+03, 3.7890E+03, 3.7918E+03, 3.7945E+03, 3.7973E+03,
     +  3.8001E+03,
     + 3.8029E+03, 3.8057E+03, 3.8084E+03, 3.8112E+03, 3.8140E+03,
     +  3.8168E+03, 3.8196E+03, 3.8224E+03, 3.8251E+03, 3.8279E+03,
     + 3.8307E+03, 3.8335E+03, 3.8363E+03, 3.8391E+03, 3.8419E+03,
     +  3.8447E+03, 3.8475E+03, 3.8503E+03, 3.8531E+03, 3.8559E+03,
     + 3.8587E+03, 3.8615E+03, 3.8643E+03, 3.8671E+03, 3.8699E+03,
     +  3.8727E+03, 3.8755E+03, 3.8783E+03, 3.8811E+03, 3.8839E+03,
     + 3.8868E+03, 3.8896E+03, 3.8924E+03, 3.8952E+03, 3.8980E+03,
     +  3.9008E+03, 3.9036E+03, 3.9065E+03, 3.9093E+03, 3.9121E+03,
     + 3.9149E+03, 3.9178E+03, 3.9206E+03, 3.9234E+03, 3.9262E+03,
     +  3.9291E+03, 3.9319E+03, 3.9347E+03, 3.9375E+03, 3.9404E+03,
     + 3.9432E+03, 3.9460E+03, 3.9489E+03, 3.9517E+03, 3.9546E+03,
     +  3.9574E+03, 3.9602E+03, 3.9631E+03, 3.9659E+03, 3.9688E+03,
     + 3.9716E+03, 3.9745E+03, 3.9773E+03, 3.9801E+03, 3.9830E+03,
     +  3.9858E+03, 3.9887E+03, 3.9915E+03, 3.9944E+03, 3.9973E+03,
     + 4.0001E+03, 4.0030E+03, 4.0058E+03, 4.0087E+03, 4.0115E+03,
     +  4.0144E+03, 4.0173E+03, 4.0201E+03, 4.0230E+03, 4.0259E+03,
     + 4.0287E+03, 4.0316E+03, 4.0345E+03, 4.0373E+03, 4.0402E+03,
     +  4.0431E+03, 4.0459E+03, 4.0488E+03, 4.0517E+03, 4.0546E+03,
     + 4.0574E+03, 4.0603E+03, 4.0632E+03, 4.0661E+03, 4.0690E+03,
     +  4.0718E+03, 4.0747E+03, 4.0776E+03, 4.0805E+03, 4.0834E+03,
     + 4.0863E+03, 4.0892E+03, 4.0920E+03, 4.0949E+03, 4.0978E+03,
     +  4.1007E+03, 4.1036E+03, 4.1065E+03, 4.1094E+03, 4.1123E+03,
     + 4.1152E+03, 4.1181E+03, 4.1210E+03, 4.1239E+03, 4.1268E+03,
     +  4.1297E+03, 4.1326E+03, 4.1355E+03, 4.1384E+03, 4.1413E+03,
     + 4.1442E+03, 4.1472E+03, 4.1501E+03, 4.1530E+03, 4.1559E+03,
     +  4.1588E+03, 4.1617E+03, 4.1646E+03, 4.1676E+03, 4.1705E+03,
     + 4.1734E+03, 4.1763E+03, 4.1792E+03, 4.1822E+03, 4.1851E+03,
     +  4.1880E+03, 4.1909E+03, 4.1939E+03, 4.1968E+03, 4.1997E+03,
     + 4.2027E+03, 4.2056E+03, 4.2085E+03, 4.2115E+03, 4.2144E+03,
     +  4.2173E+03, 4.2203E+03, 4.2232E+03, 4.2261E+03, 4.2291E+03,
     + 4.2320E+03, 4.2350E+03, 4.2379E+03, 4.2409E+03, 4.2438E+03,
     +  4.2468E+03, 4.2497E+03, 4.2527E+03, 4.2556E+03, 4.2586E+03,
     + 4.2615E+03, 4.2645E+03, 4.2674E+03, 4.2704E+03, 4.2733E+03,
     +  4.2763E+03, 4.2793E+03, 4.2822E+03, 4.2852E+03, 4.2881E+03,
     + 4.2911E+03, 4.2941E+03, 4.2970E+03, 4.3000E+03, 4.3030E+03,
     +  4.3059E+03, 4.3089E+03, 4.3119E+03, 4.3148E+03, 4.3178E+03,
     + 4.3208E+03, 4.3238E+03, 4.3268E+03, 4.3297E+03, 4.3327E+03,
     +  4.3357E+03, 4.3387E+03, 4.3417E+03, 4.3446E+03, 4.3476E+03,
     + 4.3506E+03, 4.3536E+03, 4.3566E+03, 4.3596E+03, 4.3626E+03,
     +  4.3656E+03, 4.3685E+03, 4.3715E+03, 4.3745E+03, 4.3775E+03,
     + 4.3805E+03, 4.3835E+03, 4.3865E+03, 4.3895E+03, 4.3925E+03,
     +  4.3955E+03, 4.3985E+03, 4.4015E+03, 4.4045E+03, 4.4075E+03,
     + 4.4105E+03, 4.4136E+03, 4.4166E+03, 4.4196E+03, 4.4226E+03,
     +  4.4256E+03, 4.4286E+03, 4.4316E+03, 4.4346E+03, 4.4377E+03,
     + 4.4407E+03, 4.4437E+03, 4.4467E+03, 4.4497E+03, 4.4528E+03,
     +  4.4558E+03, 4.4588E+03, 4.4618E+03, 4.4649E+03, 4.4679E+03,
     + 4.4709E+03, 4.4740E+03, 4.4770E+03, 4.4800E+03, 4.4831E+03,
     +  4.4861E+03, 4.4891E+03, 4.4922E+03, 4.4952E+03, 4.4982E+03,
     + 4.5013E+03, 4.5043E+03, 4.5074E+03, 4.5104E+03, 4.5135E+03,
     +  4.5165E+03, 4.5196E+03, 4.5226E+03, 4.5256E+03, 4.5287E+03,
     + 4.5318E+03, 4.5348E+03, 4.5379E+03, 4.5409E+03, 4.5440E+03,
     +  4.5470E+03, 4.5501E+03, 4.5531E+03, 4.5562E+03, 4.5593E+03,
     + 4.5623E+03, 4.5654E+03, 4.5685E+03, 4.5715E+03, 4.5746E+03,
     +  4.5777E+03, 4.5807E+03, 4.5838E+03, 4.5869E+03, 4.5899E+03,
     + 4.5930E+03, 4.5961E+03, 4.5992E+03, 4.6022E+03, 4.6053E+03,
     +  4.6084E+03, 4.6115E+03, 4.6146E+03, 4.6176E+03, 4.6207E+03,
     + 4.6238E+03, 4.6269E+03, 4.6300E+03, 4.6331E+03, 4.6362E+03,
     +  4.6392E+03, 4.6423E+03, 4.6454E+03, 4.6485E+03, 4.6516E+03,
     + 4.6547E+03, 4.6578E+03, 4.6609E+03, 4.6640E+03, 4.6671E+03,
     +  4.6702E+03, 4.6733E+03, 4.6764E+03, 4.6795E+03, 4.6826E+03,
     + 4.6857E+03, 4.6888E+03, 4.6919E+03, 4.6950E+03, 4.6982E+03,
     +  4.7013E+03, 4.7044E+03, 4.7075E+03, 4.7106E+03, 4.7137E+03,
     + 4.7168E+03, 4.7200E+03, 4.7231E+03, 4.7262E+03, 4.7293E+03,
     +  4.7324E+03, 4.7356E+03, 4.7387E+03, 4.7418E+03, 4.7450E+03,
     + 4.7481E+03, 4.7512E+03, 4.7543E+03, 4.7575E+03, 4.7606E+03,
     +  4.7637E+03, 4.7669E+03, 4.7700E+03, 4.7731E+03, 4.7763E+03,
     + 4.7794E+03, 4.7826E+03, 4.7857E+03, 4.7889E+03, 4.7920E+03,
     +  4.7951E+03, 4.7983E+03, 4.8014E+03, 4.8046E+03, 4.8077E+03,
     + 4.8109E+03, 4.8140E+03, 4.8172E+03, 4.8203E+03, 4.8235E+03,
     +  4.8266E+03, 4.8298E+03, 4.8330E+03, 4.8361E+03, 4.8393E+03,
     + 4.8424E+03, 4.8456E+03, 4.8488E+03, 4.8519E+03, 4.8551E+03,
     +  4.8583E+03, 4.8614E+03, 4.8646E+03, 4.8678E+03, 4.8709E+03,
     + 4.8741E+03, 4.8773E+03, 4.8805E+03, 4.8836E+03, 4.8868E+03,
     +  4.8900E+03, 4.8932E+03, 4.8964E+03, 4.8995E+03, 4.9027E+03,
     + 4.9059E+03, 4.9091E+03, 4.9123E+03, 4.9155E+03, 4.9187E+03,
     +  4.9218E+03, 4.9250E+03, 4.9282E+03, 4.9314E+03, 4.9346E+03,
     + 4.9378E+03, 4.9410E+03, 4.9442E+03, 4.9474E+03, 4.9506E+03,
     +  4.9538E+03, 4.9570E+03, 4.9602E+03, 4.9634E+03, 4.9666E+03,
     + 4.9698E+03, 4.9730E+03, 4.9762E+03, 4.9794E+03, 4.9826E+03,
     +  4.9859E+03, 4.9891E+03, 4.9923E+03, 4.9955E+03, 4.9987E+03,
     + 5.0019E+03, 5.0051E+03, 5.0084E+03, 5.0116E+03, 5.0148E+03,
     +  5.0180E+03, 5.0213E+03, 5.0245E+03, 5.0277E+03, 5.0309E+03,
     + 5.0342E+03, 5.0374E+03, 5.0406E+03, 5.0438E+03, 5.0471E+03,
     +  5.0503E+03, 5.0535E+03, 5.0568E+03, 5.0600E+03, 5.0633E+03,
     + 5.0665E+03, 5.0697E+03, 5.0730E+03, 5.0762E+03, 5.0795E+03,
     +  5.0827E+03, 5.0860E+03, 5.0892E+03, 5.0924E+03, 5.0957E+03,
     + 5.0989E+03, 5.1022E+03, 5.1055E+03, 5.1087E+03, 5.1120E+03,
     +  5.1152E+03, 5.1185E+03, 5.1217E+03, 5.1250E+03, 5.1282E+03,
     + 5.1315E+03, 5.1348E+03, 5.1380E+03, 5.1413E+03, 5.1446E+03,
     +  5.1478E+03, 5.1511E+03, 5.1544E+03, 5.1576E+03, 5.1609E+03,
     + 5.1642E+03, 5.1675E+03, 5.1707E+03, 5.1740E+03, 5.1773E+03,
     +  5.1806E+03, 5.1838E+03, 5.1871E+03, 5.1904E+03, 5.1937E+03,
     + 5.1970E+03, 5.2002E+03, 5.2035E+03, 5.2068E+03, 5.2101E+03,
     +  5.2134E+03, 5.2167E+03, 5.2200E+03, 5.2233E+03, 5.2266E+03,
     + 5.2299E+03, 5.2332E+03, 5.2364E+03, 5.2397E+03, 5.2430E+03,
     +  5.2463E+03, 5.2496E+03, 5.2529E+03, 5.2563E+03, 5.2596E+03,
     + 5.2629E+03, 5.2662E+03, 5.2695E+03, 5.2728E+03, 5.2761E+03,
     +  5.2794E+03, 5.2827E+03, 5.2860E+03, 5.2893E+03, 5.2927E+03,
     + 5.2960E+03, 5.2993E+03, 5.3026E+03, 5.3059E+03, 5.3093E+03,
     +  5.3126E+03, 5.3159E+03, 5.3192E+03, 5.3226E+03, 5.3259E+03,
     + 5.3292E+03, 5.3325E+03, 5.3359E+03, 5.3392E+03, 5.3425E+03,
     +  5.3459E+03, 5.3492E+03, 5.3525E+03, 5.3559E+03, 5.3592E+03,
     + 5.3625E+03, 5.3659E+03, 5.3692E+03, 5.3726E+03, 5.3759E+03,
     +  5.3793E+03, 5.3826E+03, 5.3860E+03, 5.3893E+03, 5.3926E+03,
     + 5.3960E+03, 5.3994E+03, 5.4027E+03, 5.4061E+03, 5.4094E+03,
     +  5.4128E+03, 5.4161E+03, 5.4195E+03, 5.4228E+03, 5.4262E+03,
     + 5.4296E+03, 5.4329E+03, 5.4363E+03, 5.4397E+03, 5.4430E+03,
     +  5.4464E+03, 5.4498E+03, 5.4531E+03, 5.4565E+03, 5.4599E+03,
     + 5.4632E+03, 5.4666E+03, 5.4700E+03, 5.4734E+03, 5.4767E+03,
     +  5.4801E+03, 5.4835E+03, 5.4869E+03, 5.4903E+03, 5.4936E+03,
     + 5.4970E+03, 5.5004E+03, 5.5038E+03, 5.5072E+03, 5.5106E+03,
     +  5.5140E+03, 5.5173E+03, 5.5207E+03, 5.5241E+03, 5.5275E+03,
     + 5.5309E+03, 5.5343E+03, 5.5377E+03, 5.5411E+03, 5.5445E+03,
     +  5.5479E+03, 5.5513E+03, 5.5547E+03, 5.5581E+03, 5.5615E+03,
     + 5.5649E+03, 5.5683E+03, 5.5717E+03, 5.5752E+03, 5.5786E+03,
     +  5.5820E+03, 5.5854E+03, 5.5888E+03, 5.5922E+03, 5.5956E+03,
     + 5.5990E+03, 5.6025E+03, 5.6059E+03, 5.6093E+03, 5.6127E+03,
     +  5.6162E+03, 5.6196E+03, 5.6230E+03, 5.6264E+03, 5.6299E+03,
     + 5.6333E+03, 5.6367E+03, 5.6401E+03, 5.6436E+03, 5.6470E+03,
     +  5.6504E+03, 5.6539E+03, 5.6573E+03, 5.6607E+03, 5.6642E+03,
     + 5.6676E+03, 5.6711E+03, 5.6745E+03, 5.6780E+03, 5.6814E+03,
     +  5.6848E+03, 5.6883E+03, 5.6917E+03, 5.6952E+03, 5.6986E+03,
     + 5.7021E+03, 5.7055E+03, 5.7090E+03, 5.7124E+03, 5.7159E+03,
     +  5.7194E+03, 5.7228E+03, 5.7263E+03, 5.7297E+03, 5.7332E+03,
     + 5.7367E+03, 5.7401E+03, 5.7436E+03, 5.7470E+03, 5.7505E+03,
     +  5.7540E+03, 5.7574E+03, 5.7609E+03, 5.7644E+03, 5.7679E+03,
     + 5.7713E+03, 5.7748E+03, 5.7783E+03, 5.7818E+03, 5.7852E+03,
     +  5.7887E+03, 5.7922E+03, 5.7957E+03, 5.7992E+03, 5.8026E+03,
     + 5.8061E+03, 5.8096E+03, 5.8131E+03, 5.8166E+03, 5.8201E+03,
     +  5.8236E+03, 5.8271E+03, 5.8306E+03, 5.8340E+03, 5.8375E+03,
     + 5.8410E+03, 5.8445E+03, 5.8480E+03, 5.8515E+03, 5.8550E+03,
     +  5.8585E+03, 5.8620E+03, 5.8655E+03, 5.8690E+03, 5.8725E+03,
     + 5.8761E+03, 5.8796E+03, 5.8831E+03, 5.8866E+03, 5.8901E+03,
     +  5.8936E+03, 5.8971E+03, 5.9006E+03, 5.9041E+03, 5.9077E+03,
     + 5.9112E+03, 5.9147E+03, 5.9182E+03, 5.9217E+03, 5.9253E+03,
     +  5.9288E+03, 5.9323E+03, 5.9358E+03, 5.9394E+03, 5.9429E+03,
     + 5.9464E+03, 5.9500E+03, 5.9535E+03, 5.9570E+03, 5.9606E+03,
     +  5.9641E+03, 5.9676E+03, 5.9712E+03, 5.9747E+03, 5.9782E+03,
     + 5.9818E+03, 5.9853E+03, 5.9889E+03, 5.9924E+03, 5.9960E+03,
     +  5.9995E+03, 6.0030E+03, 6.0066E+03, 6.0101E+03, 6.0137E+03,
     + 6.0173E+03, 6.0208E+03, 6.0244E+03, 6.0279E+03, 6.0315E+03,
     +  6.0350E+03, 6.0386E+03, 6.0421E+03, 6.0457E+03, 6.0493E+03,
     + 6.0528E+03, 6.0564E+03, 6.0600E+03, 6.0635E+03, 6.0671E+03,
     +  6.0707E+03, 6.0742E+03, 6.0778E+03, 6.0814E+03, 6.0850E+03,
     + 6.0885E+03, 6.0921E+03, 6.0957E+03, 6.0993E+03, 6.1028E+03,
     +  6.1064E+03, 6.1100E+03, 6.1136E+03, 6.1172E+03, 6.1207E+03,
     + 6.1243E+03, 6.1279E+03, 6.1315E+03, 6.1351E+03, 6.1387E+03,
     +  6.1423E+03, 6.1459E+03, 6.1495E+03, 6.1531E+03, 6.1567E+03,
     + 6.1603E+03, 6.1638E+03, 6.1674E+03, 6.1710E+03, 6.1746E+03,
     +  6.1783E+03, 6.1819E+03, 6.1855E+03, 6.1891E+03, 6.1927E+03,
     + 6.1963E+03, 6.1999E+03, 6.2035E+03, 6.2071E+03, 6.2107E+03,
     +  6.2143E+03, 6.2180E+03, 6.2216E+03, 6.2252E+03, 6.2288E+03,
     + 6.2324E+03, 6.2360E+03, 6.2397E+03, 6.2433E+03, 6.2469E+03,
     +  6.2505E+03, 6.2542E+03, 6.2578E+03, 6.2614E+03, 6.2651E+03,
     + 6.2687E+03, 6.2723E+03, 6.2760E+03, 6.2796E+03, 6.2832E+03,
     +  6.2869E+03, 6.2905E+03, 6.2941E+03, 6.2978E+03, 6.3014E+03,
     + 6.3051E+03, 6.3087E+03, 6.3123E+03, 6.3160E+03, 6.3196E+03,
     +  6.3233E+03, 6.3269E+03, 6.3306E+03, 6.3342E+03, 6.3379E+03,
     + 6.3415E+03, 6.3452E+03, 6.3489E+03, 6.3525E+03, 6.3562E+03,
     +  6.3598E+03, 6.3635E+03, 6.3671E+03, 6.3708E+03, 6.3745E+03,
     + 6.3781E+03, 6.3818E+03, 6.3855E+03, 6.3891E+03, 6.3928E+03,
     +  6.3965E+03, 6.4002E+03, 6.4038E+03, 6.4075E+03, 6.4112E+03,
     + 6.4148E+03, 6.4185E+03, 6.4222E+03, 6.4259E+03, 6.4296E+03,
     +  6.4332E+03, 6.4369E+03, 6.4406E+03, 6.4443E+03, 6.4480E+03,
     + 6.4517E+03, 6.4554E+03, 6.4591E+03, 6.4627E+03, 6.4664E+03,
     +  6.4701E+03, 6.4738E+03, 6.4775E+03, 6.4812E+03, 6.4849E+03,
     + 6.4886E+03, 6.4923E+03, 6.4960E+03, 6.4997E+03, 6.5034E+03,
     +  6.5071E+03, 6.5108E+03, 6.5145E+03, 6.5182E+03, 6.5220E+03,
     + 6.5257E+03, 6.5294E+03, 6.5331E+03, 6.5368E+03, 6.5405E+03,
     +  6.5442E+03, 6.5479E+03, 6.5517E+03, 6.5554E+03, 6.5591E+03,
     + 6.5628E+03, 6.5665E+03, 6.5703E+03, 6.5740E+03, 6.5777E+03,
     +  6.5815E+03, 6.5852E+03, 6.5889E+03, 6.5926E+03, 6.5964E+03,
     + 6.6001E+03, 6.6038E+03, 6.6076E+03, 6.6113E+03, 6.6150E+03,
     +  6.6188E+03, 6.6225E+03, 6.6263E+03, 6.6300E+03, 6.6338E+03,
     + 6.6375E+03, 6.6412E+03, 6.6450E+03, 6.6487E+03, 6.6525E+03,
     +  6.6562E+03, 6.6600E+03, 6.6637E+03, 6.6675E+03, 6.6712E+03,
     + 6.6750E+03, 6.6788E+03, 6.6825E+03, 6.6863E+03, 6.6900E+03,
     +  6.6938E+03, 6.6976E+03, 6.7013E+03, 6.7051E+03, 6.7089E+03,
     + 6.7126E+03, 6.7164E+03, 6.7202E+03, 6.7239E+03, 6.7277E+03,
     +  6.7315E+03, 6.7352E+03, 6.7390E+03, 6.7428E+03, 6.7466E+03,
     + 6.7504E+03, 6.7541E+03, 6.7579E+03, 6.7617E+03, 6.7655E+03,
     +  6.7693E+03, 6.7730E+03, 6.7768E+03, 6.7806E+03, 6.7844E+03,
     + 6.7882E+03, 6.7920E+03, 6.7958E+03, 6.7996E+03, 6.8034E+03,
     +  6.8072E+03, 6.8110E+03, 6.8148E+03, 6.8186E+03, 6.8224E+03,
     + 6.8262E+03, 6.8300E+03, 6.8338E+03, 6.8376E+03, 6.8414E+03,
     +  6.8452E+03, 6.8490E+03, 6.8528E+03, 6.8566E+03, 6.8604E+03,
     + 6.8642E+03, 6.8680E+03, 6.8719E+03, 6.8757E+03, 6.8795E+03,
     +  6.8833E+03, 6.8871E+03, 6.8910E+03, 6.8948E+03, 6.8986E+03,
     + 6.9024E+03, 6.9062E+03, 6.9101E+03, 6.9139E+03, 6.9177E+03,
     +  6.9216E+03, 6.9254E+03, 6.9292E+03, 6.9331E+03, 6.9369E+03,
     + 6.9407E+03, 6.9446E+03, 6.9484E+03, 6.9522E+03, 6.9561E+03,
     +  6.9599E+03, 6.9638E+03, 6.9676E+03, 6.9714E+03, 6.9753E+03,
     + 6.9791E+03, 6.9830E+03, 6.9868E+03, 6.9907E+03, 6.9945E+03,
     +  6.9984E+03, 7.0022E+03, 7.0061E+03, 7.0100E+03, 7.0138E+03,
     + 7.0177E+03, 7.0215E+03, 7.0254E+03, 7.0293E+03, 7.0331E+03,
     +  7.0370E+03, 7.0408E+03, 7.0447E+03, 7.0486E+03, 7.0524E+03,
     + 7.0563E+03, 7.0602E+03, 7.0641E+03, 7.0679E+03, 7.0718E+03,
     +  7.0757E+03, 7.0796E+03, 7.0834E+03, 7.0873E+03, 7.0912E+03,
     + 7.0951E+03, 7.0990E+03, 7.1028E+03, 7.1067E+03, 7.1106E+03,
     +  7.1145E+03, 7.1184E+03, 7.1223E+03, 7.1262E+03, 7.1301E+03,
     + 7.1339E+03, 7.1378E+03, 7.1417E+03, 7.1456E+03, 7.1495E+03,
     +  7.1534E+03, 7.1573E+03, 7.1612E+03, 7.1651E+03, 7.1690E+03,
     + 7.1729E+03, 7.1768E+03, 7.1807E+03, 7.1847E+03, 7.1886E+03,
     +  7.1925E+03, 7.1964E+03, 7.2003E+03, 7.2042E+03, 7.2081E+03,
     + 7.2120E+03, 7.2160E+03, 7.2199E+03, 7.2238E+03, 7.2277E+03,
     +  7.2316E+03, 7.2356E+03, 7.2395E+03, 7.2434E+03, 7.2473E+03,
     + 7.2513E+03, 7.2552E+03, 7.2591E+03, 7.2630E+03, 7.2670E+03,
     +  7.2709E+03, 7.2748E+03, 7.2788E+03, 7.2827E+03, 7.2866E+03,
     + 7.2906E+03, 7.2945E+03, 7.2985E+03, 7.3024E+03, 7.3064E+03,
     +  7.3103E+03, 7.3142E+03, 7.3182E+03, 7.3221E+03, 7.3261E+03,
     + 7.3300E+03, 7.3340E+03, 7.3379E+03, 7.3419E+03, 7.3458E+03,
     +  7.3498E+03, 7.3538E+03, 7.3577E+03, 7.3617E+03, 7.3656E+03,
     + 7.3696E+03, 7.3736E+03, 7.3775E+03, 7.3815E+03, 7.3855E+03,
     +  7.3894E+03, 7.3934E+03, 7.3974E+03, 7.4013E+03, 7.4053E+03,
     + 7.4093E+03, 7.4132E+03, 7.4172E+03, 7.4212E+03, 7.4252E+03,
     +  7.4292E+03, 7.4331E+03, 7.4371E+03, 7.4411E+03, 7.4451E+03,
     + 7.4491E+03, 7.4530E+03, 7.4570E+03, 7.4610E+03, 7.4650E+03,
     +  7.4690E+03, 7.4730E+03, 7.4770E+03, 7.4810E+03, 7.4850E+03,
     + 7.4890E+03, 7.4930E+03, 7.4970E+03, 7.5010E+03, 7.5050E+03,
     +  7.5090E+03, 7.5130E+03, 7.5170E+03, 7.5210E+03, 7.5250E+03,
     + 7.5290E+03, 7.5330E+03, 7.5370E+03, 7.5410E+03, 7.5450E+03,
     +  7.5490E+03, 7.5531E+03, 7.5571E+03, 7.5611E+03, 7.5651E+03,
     + 7.5691E+03, 7.5731E+03, 7.5772E+03, 7.5812E+03, 7.5852E+03,
     +  7.5892E+03, 7.5933E+03, 7.5973E+03, 7.6013E+03, 7.6053E+03,
     + 7.6094E+03, 7.6134E+03, 7.6174E+03, 7.6215E+03, 7.6255E+03,
     +  7.6295E+03, 7.6336E+03, 7.6376E+03, 7.6417E+03, 7.6457E+03,
     + 7.6497E+03, 7.6538E+03, 7.6578E+03, 7.6619E+03, 7.6659E+03,
     +  7.6700E+03, 7.6740E+03, 7.6781E+03, 7.6821E+03, 7.6862E+03,
     + 7.6902E+03, 7.6943E+03, 7.6983E+03, 7.7024E+03, 7.7064E+03,
     +  7.7105E+03, 7.7146E+03, 7.7186E+03, 7.7227E+03, 7.7267E+03,
     + 7.7308E+03, 7.7349E+03, 7.7389E+03, 7.7430E+03, 7.7471E+03,
     +  7.7511E+03, 7.7552E+03, 7.7593E+03, 7.7634E+03, 7.7674E+03,
     + 7.7715E+03, 7.7756E+03, 7.7797E+03, 7.7838E+03, 7.7878E+03,
     +  7.7919E+03, 7.7960E+03, 7.8001E+03, 7.8042E+03, 7.8083E+03,
     + 7.8123E+03, 7.8164E+03, 7.8205E+03, 7.8246E+03, 7.8287E+03,
     +  7.8328E+03, 7.8369E+03, 7.8410E+03, 7.8451E+03, 7.8492E+03,
     + 7.8533E+03, 7.8574E+03, 7.8615E+03, 7.8656E+03, 7.8697E+03,
     +  7.8738E+03, 7.8779E+03, 7.8820E+03, 7.8861E+03, 7.8902E+03,
     + 7.8943E+03, 7.8985E+03, 7.9026E+03, 7.9067E+03, 7.9108E+03,
     +  7.9149E+03, 7.9190E+03, 7.9231E+03, 7.9273E+03, 7.9314E+03,
     + 7.9355E+03, 7.9396E+03, 7.9438E+03, 7.9479E+03, 7.9520E+03,
     +  7.9561E+03, 7.9603E+03, 7.9644E+03, 7.9685E+03, 7.9727E+03,
     + 7.9768E+03, 7.9809E+03, 7.9851E+03, 7.9892E+03, 7.9933E+03,
     +  7.9975E+03, 8.0016E+03, 8.0058E+03, 8.0099E+03, 8.0141E+03,
     + 8.0182E+03, 8.0223E+03, 8.0265E+03, 8.0306E+03, 8.0348E+03,
     +  8.0389E+03, 8.0431E+03, 8.0472E+03, 8.0514E+03, 8.0556E+03,
     + 8.0597E+03, 8.0639E+03, 8.0680E+03, 8.0722E+03, 8.0764E+03,
     +  8.0805E+03, 8.0847E+03, 8.0888E+03, 8.0930E+03, 8.0972E+03,
     + 8.1014E+03, 8.1055E+03, 8.1097E+03, 8.1139E+03, 8.1180E+03,
     +  8.1222E+03, 8.1264E+03, 8.1306E+03, 8.1347E+03, 8.1389E+03,
     + 8.1431E+03, 8.1473E+03, 8.1515E+03, 8.1556E+03, 8.1598E+03,
     +  8.1640E+03, 8.1682E+03, 8.1724E+03, 8.1766E+03, 8.1808E+03,
     + 8.1850E+03, 8.1892E+03, 8.1933E+03, 8.1975E+03, 8.2017E+03,
     +  8.2059E+03, 8.2101E+03, 8.2143E+03, 8.2185E+03, 8.2227E+03,
     + 8.2269E+03, 8.2311E+03, 8.2354E+03, 8.2396E+03, 8.2438E+03,
     +  8.2480E+03, 8.2522E+03, 8.2564E+03, 8.2606E+03, 8.2648E+03,
     + 8.2690E+03, 8.2733E+03, 8.2775E+03, 8.2817E+03, 8.2859E+03,
     +  8.2901E+03, 8.2944E+03, 8.2986E+03, 8.3028E+03, 8.3070E+03,
     + 8.3112E+03, 8.3155E+03, 8.3197E+03, 8.3239E+03, 8.3282E+03,
     +  8.3324E+03, 8.3366E+03, 8.3409E+03, 8.3451E+03, 8.3493E+03,
     + 8.3536E+03, 8.3578E+03, 8.3621E+03, 8.3663E+03, 8.3705E+03,
     +  8.3748E+03, 8.3790E+03, 8.3833E+03, 8.3875E+03, 8.3918E+03,
     + 8.3960E+03, 8.4003E+03, 8.4045E+03, 8.4088E+03, 8.4130E+03,
     +  8.4173E+03, 8.4215E+03, 8.4258E+03, 8.4301E+03, 8.4343E+03,
     + 8.4386E+03, 8.4428E+03, 8.4471E+03, 8.4514E+03, 8.4556E+03,
     +  8.4599E+03, 8.4642E+03, 8.4684E+03, 8.4727E+03, 8.4770E+03,
     + 8.4813E+03, 8.4855E+03, 8.4898E+03, 8.4941E+03, 8.4984E+03,
     +  8.5026E+03, 8.5069E+03, 8.5112E+03, 8.5155E+03, 8.5198E+03,
     + 8.5240E+03, 8.5283E+03, 8.5326E+03, 8.5369E+03, 8.5412E+03,
     +  8.5455E+03, 8.5498E+03, 8.5541E+03, 8.5584E+03, 8.5627E+03,
     + 8.5670E+03, 8.5712E+03, 8.5755E+03, 8.5798E+03, 8.5841E+03,
     +  8.5884E+03, 8.5928E+03, 8.5971E+03, 8.6014E+03, 8.6057E+03,
     + 8.6100E+03, 8.6143E+03, 8.6186E+03, 8.6229E+03, 8.6272E+03,
     +  8.6315E+03, 8.6358E+03, 8.6402E+03, 8.6445E+03, 8.6488E+03,
     + 8.6531E+03, 8.6574E+03, 8.6618E+03, 8.6661E+03, 8.6704E+03,
     +  8.6747E+03, 8.6791E+03, 8.6834E+03, 8.6877E+03, 8.6920E+03,
     + 8.6964E+03, 8.7007E+03, 8.7050E+03, 8.7094E+03, 8.7137E+03,
     +  8.7180E+03, 8.7224E+03, 8.7267E+03, 8.7311E+03, 8.7354E+03,
     + 8.7397E+03, 8.7441E+03, 8.7484E+03, 8.7528E+03, 8.7571E+03,
     +  8.7615E+03, 8.7658E+03, 8.7702E+03, 8.7745E+03, 8.7789E+03,
     + 8.7832E+03, 8.7876E+03, 8.7919E+03, 8.7963E+03, 8.8007E+03,
     +  8.8050E+03, 8.8094E+03, 8.8137E+03, 8.8181E+03, 8.8225E+03,
     + 8.8268E+03, 8.8312E+03, 8.8356E+03, 8.8399E+03, 8.8443E+03,
     +  8.8487E+03, 8.8531E+03, 8.8574E+03, 8.8618E+03, 8.8662E+03,
     + 8.8706E+03, 8.8749E+03, 8.8793E+03, 8.8837E+03, 8.8881E+03,
     +  8.8925E+03, 8.8968E+03, 8.9012E+03, 8.9056E+03, 8.9100E+03,
     + 8.9144E+03, 8.9188E+03, 8.9232E+03, 8.9276E+03, 8.9320E+03,
     +  8.9364E+03, 8.9407E+03, 8.9451E+03, 8.9495E+03, 8.9539E+03,
     + 8.9583E+03, 8.9627E+03, 8.9671E+03, 8.9716E+03, 8.9760E+03,
     +  8.9804E+03, 8.9848E+03, 8.9892E+03, 8.9936E+03, 8.9980E+03,
     + 9.0024E+03, 9.0068E+03, 9.0112E+03, 9.0157E+03, 9.0201E+03,
     +  9.0245E+03, 9.0289E+03, 9.0333E+03, 9.0378E+03, 9.0422E+03,
     + 9.0466E+03, 9.0510E+03, 9.0554E+03, 9.0599E+03, 9.0643E+03,
     +  9.0687E+03, 9.0732E+03, 9.0776E+03, 9.0820E+03, 9.0865E+03,
     + 9.0909E+03, 9.0953E+03, 9.0998E+03, 9.1042E+03, 9.1087E+03,
     +  9.1131E+03, 9.1175E+03, 9.1220E+03, 9.1264E+03, 9.1309E+03,
     + 9.1353E+03, 9.1398E+03, 9.1442E+03, 9.1487E+03, 9.1531E+03,
     +  9.1576E+03, 9.1620E+03, 9.1665E+03, 9.1709E+03, 9.1754E+03,
     + 9.1799E+03, 9.1843E+03, 9.1888E+03, 9.1932E+03, 9.1977E+03,
     +  9.2022E+03, 9.2066E+03, 9.2111E+03, 9.2156E+03, 9.2200E+03,
     + 9.2245E+03, 9.2290E+03, 9.2335E+03, 9.2379E+03, 9.2424E+03,
     +  9.2469E+03, 9.2514E+03, 9.2558E+03, 9.2603E+03, 9.2648E+03,
     + 9.2693E+03, 9.2738E+03, 9.2782E+03, 9.2827E+03, 9.2872E+03,
     +  9.2917E+03, 9.2962E+03, 9.3007E+03, 9.3052E+03, 9.3097E+03,
     + 9.3142E+03, 9.3187E+03, 9.3232E+03, 9.3276E+03, 9.3321E+03,
     +  9.3366E+03, 9.3411E+03, 9.3457E+03, 9.3502E+03, 9.3547E+03,
     + 9.3592E+03, 9.3637E+03, 9.3682E+03, 9.3727E+03, 9.3772E+03,
     +  9.3817E+03, 9.3862E+03, 9.3907E+03, 9.3952E+03, 9.3998E+03,
     + 9.4043E+03, 9.4088E+03, 9.4133E+03, 9.4178E+03, 9.4224E+03,
     +  9.4269E+03, 9.4314E+03, 9.4359E+03, 9.4405E+03, 9.4450E+03,
     + 9.4495E+03, 9.4540E+03, 9.4586E+03, 9.4631E+03, 9.4676E+03,
     +  9.4722E+03, 9.4767E+03, 9.4813E+03, 9.4858E+03, 9.4903E+03,
     + 9.4949E+03, 9.4994E+03, 9.5040E+03, 9.5085E+03, 9.5130E+03,
     +  9.5176E+03, 9.5221E+03, 9.5267E+03, 9.5312E+03, 9.5358E+03,
     + 9.5403E+03, 9.5449E+03, 9.5494E+03, 9.5540E+03, 9.5586E+03,
     +  9.5631E+03, 9.5677E+03, 9.5722E+03, 9.5768E+03, 9.5814E+03,
     + 9.5859E+03, 9.5905E+03, 9.5951E+03, 9.5996E+03, 9.6042E+03,
     +  9.6088E+03, 9.6133E+03, 9.6179E+03, 9.6225E+03, 9.6271E+03,
     + 9.6316E+03, 9.6362E+03, 9.6408E+03, 9.6454E+03, 9.6499E+03,
     +  9.6545E+03, 9.6591E+03, 9.6637E+03, 9.6683E+03, 9.6729E+03,
     + 9.6775E+03, 9.6820E+03, 9.6866E+03, 9.6912E+03, 9.6958E+03,
     +  9.7004E+03, 9.7050E+03, 9.7096E+03, 9.7142E+03, 9.7188E+03,
     + 9.7234E+03, 9.7280E+03, 9.7326E+03, 9.7372E+03, 9.7418E+03,
     +  9.7464E+03, 9.7510E+03, 9.7556E+03, 9.7602E+03, 9.7648E+03,
     + 9.7694E+03, 9.7741E+03, 9.7787E+03, 9.7833E+03, 9.7879E+03,
     +  9.7925E+03, 9.7971E+03, 9.8018E+03, 9.8064E+03, 9.8110E+03,
     + 9.8156E+03, 9.8202E+03, 9.8249E+03, 9.8295E+03, 9.8341E+03,
     +  9.8387E+03, 9.8434E+03, 9.8480E+03, 9.8526E+03, 9.8573E+03,
     + 9.8619E+03, 9.8665E+03, 9.8712E+03, 9.8758E+03, 9.8805E+03,
     +  9.8851E+03, 9.8897E+03, 9.8944E+03, 9.8990E+03, 9.9037E+03,
     + 9.9083E+03, 9.9130E+03, 9.9176E+03, 9.9223E+03, 9.9269E+03,
     +  9.9316E+03, 9.9362E+03, 9.9409E+03, 9.9455E+03, 9.9502E+03,
     + 9.9548E+03, 9.9595E+03, 9.9642E+03, 9.9688E+03, 9.9735E+03,
     +  9.9781E+03, 9.9828E+03, 9.9875E+03, 9.9921E+03, 9.9968E+03,
     + 1.0001E+04, 1.0006E+04, 1.0011E+04, 1.0015E+04, 1.0020E+04,
     +  1.0025E+04, 1.0030E+04, 1.0034E+04, 1.0039E+04, 1.0044E+04,
     + 1.0048E+04, 1.0053E+04, 1.0058E+04, 1.0062E+04, 1.0067E+04,
     +  1.0072E+04, 1.0076E+04, 1.0081E+04, 1.0086E+04, 1.0090E+04,
     + 1.0095E+04, 1.0100E+04, 1.0104E+04, 1.0109E+04, 1.0114E+04,
     +  1.0119E+04, 1.0123E+04, 1.0128E+04, 1.0133E+04, 1.0137E+04,
     + 1.0142E+04, 1.0147E+04, 1.0152E+04, 1.0156E+04, 1.0161E+04,
     +  1.0166E+04, 1.0170E+04, 1.0175E+04, 1.0180E+04, 1.0184E+04,
     + 1.0189E+04, 1.0194E+04, 1.0199E+04, 1.0203E+04, 1.0208E+04,
     +  1.0213E+04, 1.0218E+04, 1.0222E+04, 1.0227E+04, 1.0232E+04,
     + 1.0236E+04, 1.0241E+04, 1.0246E+04, 1.0251E+04, 1.0255E+04,
     +  1.0260E+04, 1.0265E+04, 1.0270E+04, 1.0274E+04, 1.0279E+04,
     + 1.0284E+04, 1.0289E+04, 1.0293E+04, 1.0298E+04, 1.0303E+04,
     +  1.0307E+04, 1.0312E+04, 1.0317E+04, 1.0322E+04, 1.0326E+04,
     + 1.0331E+04, 1.0336E+04, 1.0341E+04, 1.0346E+04, 1.0350E+04,
     +  1.0355E+04, 1.0360E+04, 1.0365E+04, 1.0369E+04, 1.0374E+04,
     + 1.0379E+04, 1.0384E+04, 1.0388E+04, 1.0393E+04, 1.0398E+04,
     +  1.0403E+04, 1.0407E+04, 1.0412E+04, 1.0417E+04, 1.0422E+04,
     + 1.0427E+04, 1.0431E+04, 1.0436E+04, 1.0441E+04, 1.0446E+04,
     +  1.0450E+04, 1.0455E+04, 1.0460E+04, 1.0465E+04, 1.0470E+04,
     + 1.0474E+04, 1.0479E+04, 1.0484E+04, 1.0489E+04, 1.0493E+04,
     +  1.0498E+04, 1.0503E+04, 1.0508E+04, 1.0513E+04, 1.0517E+04,
     + 1.0522E+04, 1.0527E+04, 1.0532E+04, 1.0537E+04, 1.0541E+04,
     +  1.0546E+04, 1.0551E+04, 1.0556E+04, 1.0561E+04, 1.0565E+04,
     + 1.0570E+04, 1.0575E+04, 1.0580E+04, 1.0585E+04, 1.0590E+04,
     +  1.0594E+04, 1.0599E+04, 1.0604E+04, 1.0609E+04, 1.0614E+04,
     + 1.0618E+04, 1.0623E+04, 1.0628E+04, 1.0633E+04, 1.0638E+04,
     +  1.0643E+04, 1.0647E+04, 1.0652E+04, 1.0657E+04, 1.0662E+04,
     + 1.0667E+04, 1.0672E+04, 1.0676E+04, 1.0681E+04, 1.0686E+04,
     +  1.0691E+04, 1.0696E+04, 1.0701E+04, 1.0705E+04, 1.0710E+04,
     + 1.0715E+04, 1.0720E+04, 1.0725E+04, 1.0730E+04, 1.0735E+04,
     +  1.0739E+04, 1.0744E+04, 1.0749E+04, 1.0754E+04, 1.0759E+04,
     + 1.0764E+04, 1.0769E+04, 1.0773E+04, 1.0778E+04, 1.0783E+04,
     +  1.0788E+04, 1.0793E+04, 1.0798E+04, 1.0803E+04, 1.0807E+04,
     + 1.0812E+04, 1.0817E+04, 1.0822E+04, 1.0827E+04, 1.0832E+04,
     +  1.0837E+04, 1.0842E+04, 1.0846E+04, 1.0851E+04, 1.0856E+04,
     + 1.0861E+04, 1.0866E+04, 1.0871E+04, 1.0876E+04, 1.0881E+04,
     +  1.0885E+04, 1.0890E+04, 1.0895E+04, 1.0900E+04, 1.0905E+04,
     + 1.0910E+04, 1.0915E+04, 1.0920E+04, 1.0925E+04, 1.0930E+04,
     +  1.0934E+04, 1.0939E+04, 1.0944E+04, 1.0949E+04, 1.0954E+04,
     + 1.0959E+04, 1.0964E+04, 1.0969E+04, 1.0974E+04, 1.0979E+04,
     +  1.0983E+04, 1.0988E+04, 1.0993E+04, 1.0998E+04, 1.1003E+04,
     + 1.1008E+04, 1.1013E+04, 1.1018E+04, 1.1023E+04, 1.1028E+04,
     +  1.1033E+04, 1.1038E+04, 1.1042E+04, 1.1047E+04, 1.1052E+04,
     + 1.1057E+04, 1.1062E+04, 1.1067E+04, 1.1072E+04, 1.1077E+04,
     +  1.1082E+04, 1.1087E+04, 1.1092E+04, 1.1097E+04, 1.1102E+04,
     + 1.1107E+04, 1.1112E+04, 1.1116E+04, 1.1121E+04, 1.1126E+04,
     +  1.1131E+04, 1.1136E+04, 1.1141E+04, 1.1146E+04, 1.1151E+04,
     + 1.1156E+04, 1.1161E+04, 1.1166E+04, 1.1171E+04, 1.1176E+04,
     +  1.1181E+04, 1.1186E+04, 1.1191E+04, 1.1196E+04, 1.1201E+04,
     + 1.1206E+04, 1.1211E+04, 1.1216E+04, 1.1221E+04, 1.1226E+04,
     +  1.1230E+04, 1.1235E+04, 1.1240E+04, 1.1245E+04, 1.1250E+04,
     + 1.1255E+04, 1.1260E+04, 1.1265E+04, 1.1270E+04, 1.1275E+04,
     +  1.1280E+04, 1.1285E+04, 1.1290E+04, 1.1295E+04, 1.1300E+04,
     + 1.1305E+04, 1.1310E+04, 1.1315E+04, 1.1320E+04, 1.1325E+04,
     +  1.1330E+04, 1.1335E+04, 1.1340E+04, 1.1345E+04, 1.1350E+04,
     + 1.1355E+04, 1.1360E+04, 1.1365E+04, 1.1370E+04, 1.1375E+04,
     +  1.1380E+04, 1.1385E+04, 1.1390E+04, 1.1395E+04, 1.1400E+04,
     + 1.1405E+04, 1.1410E+04, 1.1415E+04, 1.1420E+04, 1.1425E+04,
     +  1.1430E+04, 1.1435E+04, 1.1440E+04, 1.1445E+04, 1.1450E+04,
     + 1.1455E+04, 1.1460E+04, 1.1465E+04, 1.1470E+04, 1.1475E+04,
     +  1.1480E+04, 1.1485E+04, 1.1491E+04, 1.1496E+04, 1.1501E+04,
     + 1.1506E+04, 1.1511E+04, 1.1516E+04, 1.1521E+04, 1.1526E+04,
     +  1.1531E+04, 1.1536E+04, 1.1541E+04, 1.1546E+04, 1.1551E+04,
     + 1.1556E+04, 1.1561E+04, 1.1566E+04, 1.1571E+04, 1.1576E+04,
     +  1.1581E+04, 1.1586E+04, 1.1591E+04, 1.1596E+04, 1.1601E+04,
     + 1.1607E+04, 1.1612E+04, 1.1617E+04, 1.1622E+04, 1.1627E+04,
     +  1.1632E+04, 1.1637E+04, 1.1642E+04, 1.1647E+04, 1.1652E+04,
     + 1.1657E+04, 1.1662E+04, 1.1667E+04, 1.1672E+04, 1.1677E+04,
     +  1.1683E+04, 1.1688E+04, 1.1693E+04, 1.1698E+04, 1.1703E+04,
     + 1.1708E+04, 1.1713E+04, 1.1718E+04, 1.1723E+04, 1.1728E+04,
     +  1.1733E+04, 1.1738E+04, 1.1744E+04, 1.1749E+04, 1.1754E+04,
     + 1.1759E+04, 1.1764E+04, 1.1769E+04, 1.1774E+04, 1.1779E+04,
     +  1.1784E+04, 1.1789E+04, 1.1794E+04, 1.1800E+04, 1.1805E+04,
     + 1.1810E+04, 1.1815E+04, 1.1820E+04, 1.1825E+04, 1.1830E+04,
     +  1.1835E+04, 1.1840E+04, 1.1846E+04, 1.1851E+04, 1.1856E+04,
     + 1.1861E+04, 1.1866E+04, 1.1871E+04, 1.1876E+04, 1.1881E+04,
     +  1.1887E+04, 1.1892E+04, 1.1897E+04, 1.1902E+04, 1.1907E+04,
     + 1.1912E+04, 1.1917E+04, 1.1922E+04, 1.1928E+04, 1.1933E+04,
     +  1.1938E+04, 1.1943E+04, 1.1948E+04, 1.1953E+04, 1.1958E+04,
     + 1.1963E+04, 1.1969E+04, 1.1974E+04, 1.1979E+04, 1.1984E+04,
     +  1.1989E+04, 1.1994E+04, 1.2000E+04, 1.2005E+04, 1.2010E+04,
     + 1.2015E+04, 1.2020E+04, 1.2025E+04, 1.2030E+04, 1.2036E+04,
     +  1.2041E+04, 1.2046E+04, 1.2051E+04, 1.2056E+04, 1.2061E+04,
     + 1.2067E+04, 1.2072E+04, 1.2077E+04, 1.2082E+04/
c...        --        36
      data (QofT( 2,J),J=1,2408)/ 2.0303E+00, 2.4299E+00, 3.0801E+00,
     +  3.7937E+00, 4.5270E+00, 5.2689E+00, 6.0153E+00, 6.7643E+00,
     + 7.5151E+00, 8.2670E+00, 9.0198E+00, 9.7732E+00, 1.0527E+01,
     +  1.1281E+01, 1.2036E+01, 1.2791E+01, 1.3546E+01, 1.4301E+01,
     + 1.5056E+01, 1.5811E+01, 1.7323E+01, 1.8834E+01, 2.0346E+01,
     +  2.1857E+01, 2.3369E+01, 2.4881E+01, 2.6393E+01, 2.7906E+01,
     + 2.9418E+01, 3.0930E+01, 3.2443E+01, 3.3955E+01, 3.5468E+01,
     +  3.6981E+01, 3.8493E+01, 4.0006E+01, 4.1519E+01, 4.3031E+01,
     + 4.4544E+01, 4.6057E+01, 4.7570E+01, 4.9083E+01, 5.0596E+01,
     +  5.2108E+01, 5.3621E+01, 5.5134E+01, 5.6647E+01, 5.8160E+01,
     + 5.9673E+01, 6.1186E+01, 6.2699E+01, 6.4213E+01, 6.5726E+01,
     +  6.7239E+01, 6.8752E+01, 7.0265E+01, 7.1778E+01, 7.3291E+01,
     + 7.4805E+01, 7.6318E+01, 7.7831E+01, 7.9344E+01, 8.0858E+01,
     +  8.2371E+01, 8.3884E+01, 8.5397E+01, 8.6911E+01, 8.8424E+01,
     + 8.9938E+01, 9.1451E+01, 9.2964E+01, 9.4478E+01, 9.5991E+01,
     +  9.7505E+01, 9.9018E+01, 1.0053E+02, 1.0204E+02, 1.0356E+02,
     + 1.0507E+02, 1.0659E+02, 1.0810E+02, 1.0961E+02, 1.1113E+02,
     +  1.1264E+02, 1.1415E+02, 1.1567E+02, 1.1718E+02, 1.1869E+02,
     + 1.2021E+02, 1.2172E+02, 1.2324E+02, 1.2475E+02, 1.2626E+02,
     +  1.2778E+02, 1.2929E+02, 1.3080E+02, 1.3232E+02, 1.3383E+02,
     + 1.3535E+02, 1.3686E+02, 1.3837E+02, 1.3989E+02, 1.4140E+02,
     +  1.4291E+02, 1.4443E+02, 1.4594E+02, 1.4746E+02, 1.4897E+02,
     + 1.5048E+02, 1.5200E+02, 1.5351E+02, 1.5503E+02, 1.5654E+02,
     +  1.5805E+02, 1.5957E+02, 1.6108E+02, 1.6260E+02, 1.6411E+02,
     + 1.6563E+02, 1.6714E+02, 1.6865E+02, 1.7017E+02, 1.7168E+02,
     +  1.7320E+02, 1.7471E+02, 1.7622E+02, 1.7774E+02, 1.7925E+02,
     + 1.8077E+02, 1.8228E+02, 1.8380E+02, 1.8531E+02, 1.8682E+02,
     +  1.8834E+02, 1.8985E+02, 1.9137E+02, 1.9288E+02, 1.9440E+02,
     + 1.9591E+02, 1.9743E+02, 1.9894E+02, 2.0046E+02, 2.0197E+02,
     +  2.0348E+02, 2.0500E+02, 2.0651E+02, 2.0803E+02, 2.0954E+02,
     + 2.1106E+02, 2.1257E+02, 2.1409E+02, 2.1560E+02, 2.1712E+02,
     +  2.1863E+02, 2.2015E+02, 2.2166E+02, 2.2318E+02, 2.2469E+02,
     + 2.2621E+02, 2.2773E+02, 2.2924E+02, 2.3076E+02, 2.3227E+02,
     +  2.3379E+02, 2.3530E+02, 2.3682E+02, 2.3833E+02, 2.3985E+02,
     + 2.4137E+02, 2.4288E+02, 2.4440E+02, 2.4592E+02, 2.4743E+02,
     +  2.4895E+02, 2.5046E+02, 2.5198E+02, 2.5350E+02, 2.5501E+02,
     + 2.5653E+02, 2.5805E+02, 2.5957E+02, 2.6108E+02, 2.6260E+02,
     +  2.6412E+02, 2.6564E+02, 2.6715E+02, 2.6867E+02, 2.7019E+02,
     + 2.7171E+02, 2.7323E+02, 2.7474E+02, 2.7626E+02, 2.7778E+02,
     +  2.7930E+02, 2.8082E+02, 2.8234E+02, 2.8386E+02, 2.8538E+02,
     + 2.8690E+02, 2.8842E+02, 2.8994E+02, 2.9146E+02, 2.9298E+02,
     +  2.9450E+02, 2.9602E+02, 2.9754E+02, 2.9906E+02, 3.0059E+02,
     + 3.0211E+02, 3.0363E+02, 3.0515E+02, 3.0668E+02, 3.0820E+02,
     +  3.0972E+02, 3.1125E+02, 3.1277E+02, 3.1429E+02, 3.1582E+02,
     + 3.1734E+02, 3.1887E+02, 3.2039E+02, 3.2192E+02, 3.2344E+02,
     +  3.2497E+02, 3.2650E+02, 3.2802E+02, 3.2955E+02, 3.3108E+02,
     + 3.3261E+02, 3.3413E+02, 3.3566E+02, 3.3719E+02, 3.3872E+02,
     +  3.4025E+02, 3.4178E+02, 3.4331E+02, 3.4484E+02, 3.4637E+02,
     + 3.4790E+02, 3.4944E+02, 3.5097E+02, 3.5250E+02, 3.5404E+02,
     +  3.5557E+02, 3.5710E+02, 3.5864E+02, 3.6017E+02, 3.6171E+02,
     + 3.6325E+02, 3.6478E+02, 3.6632E+02, 3.6786E+02, 3.6939E+02,
     +  3.7093E+02, 3.7247E+02, 3.7401E+02, 3.7555E+02, 3.7709E+02,
     + 3.7863E+02, 3.8018E+02, 3.8172E+02, 3.8326E+02, 3.8480E+02,
     +  3.8635E+02, 3.8789E+02, 3.8944E+02, 3.9098E+02, 3.9253E+02,
     + 3.9408E+02, 3.9563E+02, 3.9717E+02, 3.9872E+02, 4.0027E+02,
     +  4.0182E+02, 4.0337E+02, 4.0492E+02, 4.0648E+02, 4.0803E+02,
     + 4.0958E+02, 4.1114E+02, 4.1269E+02, 4.1425E+02, 4.1580E+02,
     +  4.1736E+02, 4.1892E+02, 4.2048E+02, 4.2204E+02, 4.2360E+02,
     + 4.2516E+02, 4.2672E+02, 4.2828E+02, 4.2984E+02, 4.3141E+02,
     +  4.3297E+02, 4.3454E+02, 4.3610E+02, 4.3767E+02, 4.3924E+02,
     + 4.4080E+02, 4.4237E+02, 4.4394E+02, 4.4551E+02, 4.4709E+02,
     +  4.4866E+02, 4.5023E+02, 4.5181E+02, 4.5338E+02, 4.5496E+02,
     + 4.5653E+02, 4.5811E+02, 4.5969E+02, 4.6127E+02, 4.6285E+02,
     +  4.6443E+02, 4.6601E+02, 4.6760E+02, 4.6918E+02, 4.7077E+02,
     + 4.7235E+02, 4.7394E+02, 4.7553E+02, 4.7711E+02, 4.7870E+02,
     +  4.8030E+02, 4.8189E+02, 4.8348E+02, 4.8507E+02, 4.8667E+02,
     + 4.8826E+02, 4.8986E+02, 4.9146E+02, 4.9306E+02, 4.9465E+02,
     +  4.9626E+02, 4.9786E+02, 4.9946E+02, 5.0106E+02, 5.0267E+02,
     + 5.0427E+02, 5.0588E+02, 5.0749E+02, 5.0910E+02, 5.1071E+02,
     +  5.1232E+02, 5.1393E+02, 5.1554E+02, 5.1716E+02, 5.1877E+02,
     + 5.2039E+02, 5.2201E+02, 5.2363E+02, 5.2525E+02, 5.2687E+02,
     +  5.2849E+02, 5.3011E+02, 5.3174E+02, 5.3336E+02, 5.3499E+02,
     + 5.3662E+02, 5.3825E+02, 5.3988E+02, 5.4151E+02, 5.4314E+02,
     +  5.4478E+02, 5.4641E+02, 5.4805E+02, 5.4969E+02, 5.5132E+02,
     + 5.5296E+02, 5.5461E+02, 5.5625E+02, 5.5789E+02, 5.5954E+02,
     +  5.6118E+02, 5.6283E+02, 5.6448E+02, 5.6613E+02, 5.6778E+02,
     + 5.6943E+02, 5.7108E+02, 5.7274E+02, 5.7439E+02, 5.7605E+02,
     +  5.7771E+02, 5.7937E+02, 5.8103E+02, 5.8269E+02, 5.8436E+02,
     + 5.8602E+02, 5.8769E+02, 5.8936E+02, 5.9103E+02, 5.9270E+02,
     +  5.9437E+02, 5.9604E+02, 5.9772E+02, 5.9939E+02, 6.0107E+02,
     + 6.0275E+02, 6.0443E+02, 6.0611E+02, 6.0779E+02, 6.0947E+02,
     +  6.1116E+02, 6.1285E+02, 6.1453E+02, 6.1622E+02, 6.1791E+02,
     + 6.1961E+02, 6.2130E+02, 6.2299E+02, 6.2469E+02, 6.2639E+02,
     +  6.2809E+02, 6.2979E+02, 6.3149E+02, 6.3319E+02, 6.3490E+02,
     + 6.3660E+02, 6.3831E+02, 6.4002E+02, 6.4173E+02, 6.4344E+02,
     +  6.4516E+02, 6.4687E+02, 6.4859E+02, 6.5030E+02, 6.5202E+02,
     + 6.5374E+02, 6.5547E+02, 6.5719E+02, 6.5892E+02, 6.6064E+02,
     +  6.6237E+02, 6.6410E+02, 6.6583E+02, 6.6756E+02, 6.6930E+02,
     + 6.7103E+02, 6.7277E+02, 6.7451E+02, 6.7625E+02, 6.7799E+02,
     +  6.7973E+02, 6.8148E+02, 6.8322E+02, 6.8497E+02, 6.8672E+02,
     + 6.8847E+02, 6.9022E+02, 6.9198E+02, 6.9373E+02, 6.9549E+02,
     +  6.9725E+02, 6.9901E+02, 7.0077E+02, 7.0253E+02, 7.0430E+02,
     + 7.0606E+02, 7.0783E+02, 7.0960E+02, 7.1137E+02, 7.1314E+02,
     +  7.1492E+02, 7.1669E+02, 7.1847E+02, 7.2025E+02, 7.2203E+02,
     + 7.2381E+02, 7.2559E+02, 7.2738E+02, 7.2917E+02, 7.3095E+02,
     +  7.3274E+02, 7.3453E+02, 7.3633E+02, 7.3812E+02, 7.3992E+02,
     + 7.4172E+02, 7.4352E+02, 7.4532E+02, 7.4712E+02, 7.4893E+02,
     +  7.5073E+02, 7.5254E+02, 7.5435E+02, 7.5616E+02, 7.5797E+02,
     + 7.5979E+02, 7.6160E+02, 7.6342E+02, 7.6524E+02, 7.6706E+02,
     +  7.6888E+02, 7.7071E+02, 7.7253E+02, 7.7436E+02, 7.7619E+02,
     + 7.7802E+02, 7.7985E+02, 7.8169E+02, 7.8352E+02, 7.8536E+02,
     +  7.8720E+02, 7.8904E+02, 7.9088E+02, 7.9273E+02, 7.9457E+02,
     + 7.9642E+02, 7.9827E+02, 8.0012E+02, 8.0197E+02, 8.0383E+02,
     +  8.0568E+02, 8.0754E+02, 8.0940E+02, 8.1126E+02, 8.1313E+02,
     + 8.1499E+02, 8.1686E+02, 8.1872E+02, 8.2059E+02, 8.2247E+02,
     +  8.2434E+02, 8.2621E+02, 8.2809E+02, 8.2997E+02, 8.3185E+02,
     + 8.3373E+02, 8.3562E+02, 8.3750E+02, 8.3939E+02, 8.4128E+02,
     +  8.4317E+02, 8.4506E+02, 8.4695E+02, 8.4885E+02, 8.5075E+02,
     + 8.5265E+02, 8.5455E+02, 8.5645E+02, 8.5835E+02, 8.6026E+02,
     +  8.6217E+02, 8.6408E+02, 8.6599E+02, 8.6790E+02, 8.6982E+02,
     + 8.7174E+02, 8.7365E+02, 8.7557E+02, 8.7750E+02, 8.7942E+02,
     +  8.8135E+02, 8.8327E+02, 8.8520E+02, 8.8713E+02, 8.8907E+02,
     + 8.9100E+02, 8.9294E+02, 8.9488E+02, 8.9682E+02, 8.9876E+02,
     +  9.0070E+02, 9.0265E+02, 9.0460E+02, 9.0654E+02, 9.0849E+02,
     + 9.1045E+02, 9.1240E+02, 9.1436E+02, 9.1632E+02, 9.1828E+02,
     +  9.2024E+02, 9.2220E+02, 9.2417E+02, 9.2613E+02, 9.2810E+02,
     + 9.3007E+02, 9.3205E+02, 9.3402E+02, 9.3600E+02, 9.3797E+02,
     +  9.3995E+02, 9.4194E+02, 9.4392E+02, 9.4590E+02, 9.4789E+02,
     + 9.4988E+02, 9.5187E+02, 9.5386E+02, 9.5586E+02, 9.5785E+02,
     +  9.5985E+02, 9.6185E+02, 9.6385E+02, 9.6586E+02, 9.6786E+02,
     + 9.6987E+02, 9.7188E+02, 9.7389E+02, 9.7590E+02, 9.7792E+02,
     +  9.7994E+02, 9.8195E+02, 9.8397E+02, 9.8600E+02, 9.8802E+02,
     + 9.9005E+02, 9.9207E+02, 9.9410E+02, 9.9613E+02, 9.9817E+02,
     +  1.0002E+03, 1.0022E+03, 1.0043E+03, 1.0063E+03, 1.0084E+03,
     + 1.0104E+03, 1.0125E+03, 1.0145E+03, 1.0166E+03, 1.0186E+03,
     +  1.0207E+03, 1.0227E+03, 1.0248E+03, 1.0268E+03, 1.0289E+03,
     + 1.0310E+03, 1.0330E+03, 1.0351E+03, 1.0372E+03, 1.0392E+03,
     +  1.0413E+03, 1.0434E+03, 1.0454E+03, 1.0475E+03, 1.0496E+03,
     + 1.0517E+03, 1.0538E+03, 1.0559E+03, 1.0579E+03, 1.0600E+03,
     +  1.0621E+03, 1.0642E+03, 1.0663E+03, 1.0684E+03, 1.0705E+03,
     + 1.0726E+03, 1.0747E+03, 1.0768E+03, 1.0789E+03, 1.0810E+03,
     +  1.0831E+03, 1.0853E+03, 1.0874E+03, 1.0895E+03, 1.0916E+03,
     + 1.0937E+03, 1.0959E+03, 1.0980E+03, 1.1001E+03, 1.1022E+03,
     +  1.1044E+03, 1.1065E+03, 1.1086E+03, 1.1108E+03, 1.1129E+03,
     + 1.1150E+03, 1.1172E+03, 1.1193E+03, 1.1215E+03, 1.1236E+03,
     +  1.1258E+03, 1.1279E+03, 1.1301E+03, 1.1322E+03, 1.1344E+03,
     + 1.1365E+03, 1.1387E+03, 1.1409E+03, 1.1430E+03, 1.1452E+03,
     +  1.1474E+03, 1.1495E+03, 1.1517E+03, 1.1539E+03, 1.1560E+03,
     + 1.1582E+03, 1.1604E+03, 1.1626E+03, 1.1648E+03, 1.1670E+03,
     +  1.1691E+03, 1.1713E+03, 1.1735E+03, 1.1757E+03, 1.1779E+03,
     + 1.1801E+03, 1.1823E+03, 1.1845E+03, 1.1867E+03, 1.1889E+03,
     +  1.1911E+03, 1.1933E+03, 1.1955E+03, 1.1978E+03, 1.2000E+03,
     + 1.2022E+03, 1.2044E+03, 1.2066E+03, 1.2088E+03, 1.2111E+03,
     +  1.2133E+03, 1.2155E+03, 1.2178E+03, 1.2200E+03, 1.2222E+03,
     + 1.2245E+03, 1.2267E+03, 1.2289E+03, 1.2312E+03, 1.2334E+03,
     +  1.2357E+03, 1.2379E+03, 1.2402E+03, 1.2424E+03, 1.2447E+03,
     + 1.2469E+03, 1.2492E+03, 1.2515E+03, 1.2537E+03, 1.2560E+03,
     +  1.2582E+03, 1.2605E+03, 1.2628E+03, 1.2651E+03, 1.2673E+03,
     + 1.2696E+03, 1.2719E+03, 1.2742E+03, 1.2764E+03, 1.2787E+03,
     +  1.2810E+03, 1.2833E+03, 1.2856E+03, 1.2879E+03, 1.2902E+03,
     + 1.2925E+03, 1.2948E+03, 1.2971E+03, 1.2994E+03, 1.3017E+03,
     +  1.3040E+03, 1.3063E+03, 1.3086E+03, 1.3109E+03, 1.3132E+03,
     + 1.3155E+03, 1.3178E+03, 1.3202E+03, 1.3225E+03, 1.3248E+03,
     +  1.3271E+03, 1.3295E+03, 1.3318E+03, 1.3341E+03, 1.3365E+03,
     + 1.3388E+03, 1.3411E+03, 1.3435E+03, 1.3458E+03, 1.3481E+03,
     +  1.3505E+03, 1.3528E+03, 1.3552E+03, 1.3575E+03, 1.3599E+03,
     + 1.3622E+03, 1.3646E+03, 1.3670E+03, 1.3693E+03, 1.3717E+03,
     +  1.3741E+03, 1.3764E+03, 1.3788E+03, 1.3812E+03, 1.3835E+03,
     + 1.3859E+03, 1.3883E+03, 1.3907E+03, 1.3930E+03, 1.3954E+03,
     +  1.3978E+03, 1.4002E+03, 1.4026E+03, 1.4050E+03, 1.4074E+03,
     + 1.4098E+03, 1.4122E+03, 1.4146E+03, 1.4170E+03, 1.4194E+03,
     +  1.4218E+03, 1.4242E+03, 1.4266E+03, 1.4290E+03, 1.4314E+03,
     + 1.4338E+03, 1.4362E+03, 1.4387E+03, 1.4411E+03, 1.4435E+03,
     +  1.4459E+03, 1.4484E+03, 1.4508E+03, 1.4532E+03, 1.4557E+03,
     + 1.4581E+03, 1.4605E+03, 1.4630E+03, 1.4654E+03, 1.4679E+03,
     +  1.4703E+03, 1.4727E+03, 1.4752E+03, 1.4776E+03, 1.4801E+03,
     + 1.4826E+03, 1.4850E+03, 1.4875E+03, 1.4899E+03, 1.4924E+03,
     +  1.4949E+03, 1.4973E+03, 1.4998E+03, 1.5023E+03, 1.5047E+03,
     + 1.5072E+03, 1.5097E+03, 1.5122E+03, 1.5147E+03, 1.5171E+03,
     +  1.5196E+03, 1.5221E+03, 1.5246E+03, 1.5271E+03, 1.5296E+03,
     + 1.5321E+03, 1.5346E+03, 1.5371E+03, 1.5396E+03, 1.5421E+03,
     +  1.5446E+03, 1.5471E+03, 1.5496E+03, 1.5521E+03, 1.5546E+03,
     + 1.5572E+03, 1.5597E+03, 1.5622E+03, 1.5647E+03, 1.5672E+03,
     +  1.5698E+03, 1.5723E+03, 1.5748E+03, 1.5774E+03, 1.5799E+03,
     + 1.5824E+03, 1.5850E+03, 1.5875E+03, 1.5901E+03, 1.5926E+03,
     +  1.5951E+03, 1.5977E+03, 1.6002E+03, 1.6028E+03, 1.6054E+03,
     + 1.6079E+03, 1.6105E+03, 1.6130E+03, 1.6156E+03, 1.6182E+03,
     +  1.6207E+03, 1.6233E+03, 1.6259E+03, 1.6284E+03, 1.6310E+03,
     + 1.6336E+03, 1.6362E+03, 1.6388E+03, 1.6413E+03, 1.6439E+03,
     +  1.6465E+03, 1.6491E+03, 1.6517E+03, 1.6543E+03, 1.6569E+03,
     + 1.6595E+03, 1.6621E+03, 1.6647E+03, 1.6673E+03, 1.6699E+03,
     +  1.6725E+03, 1.6751E+03, 1.6777E+03, 1.6803E+03, 1.6830E+03,
     + 1.6856E+03, 1.6882E+03, 1.6908E+03, 1.6934E+03, 1.6961E+03,
     +  1.6987E+03, 1.7013E+03, 1.7040E+03, 1.7066E+03, 1.7092E+03,
     + 1.7119E+03, 1.7145E+03, 1.7172E+03, 1.7198E+03, 1.7224E+03,
     +  1.7251E+03, 1.7277E+03, 1.7304E+03, 1.7331E+03, 1.7357E+03,
     + 1.7384E+03, 1.7410E+03, 1.7437E+03, 1.7464E+03, 1.7490E+03,
     +  1.7517E+03, 1.7544E+03, 1.7570E+03, 1.7597E+03, 1.7624E+03,
     + 1.7651E+03, 1.7678E+03, 1.7705E+03, 1.7731E+03, 1.7758E+03,
     +  1.7785E+03, 1.7812E+03, 1.7839E+03, 1.7866E+03, 1.7893E+03,
     + 1.7920E+03, 1.7947E+03, 1.7974E+03, 1.8001E+03, 1.8028E+03,
     +  1.8055E+03, 1.8082E+03, 1.8110E+03, 1.8137E+03, 1.8164E+03,
     + 1.8191E+03, 1.8218E+03, 1.8246E+03, 1.8273E+03, 1.8300E+03,
     +  1.8328E+03, 1.8355E+03, 1.8382E+03, 1.8410E+03, 1.8437E+03,
     + 1.8464E+03, 1.8492E+03, 1.8519E+03, 1.8547E+03, 1.8574E+03,
     +  1.8602E+03, 1.8629E+03, 1.8657E+03, 1.8685E+03, 1.8712E+03,
     + 1.8740E+03, 1.8767E+03, 1.8795E+03, 1.8823E+03, 1.8851E+03,
     +  1.8878E+03, 1.8906E+03, 1.8934E+03, 1.8962E+03, 1.8989E+03,
     + 1.9017E+03, 1.9045E+03, 1.9073E+03, 1.9101E+03, 1.9129E+03,
     +  1.9157E+03, 1.9185E+03, 1.9213E+03, 1.9241E+03, 1.9269E+03,
     + 1.9297E+03, 1.9325E+03, 1.9353E+03, 1.9381E+03, 1.9409E+03,
     +  1.9437E+03, 1.9465E+03, 1.9494E+03, 1.9522E+03, 1.9550E+03,
     + 1.9578E+03, 1.9607E+03, 1.9635E+03, 1.9663E+03, 1.9692E+03,
     +  1.9720E+03, 1.9748E+03, 1.9777E+03, 1.9805E+03, 1.9834E+03,
     + 1.9862E+03, 1.9890E+03, 1.9919E+03, 1.9948E+03, 1.9976E+03,
     +  2.0005E+03, 2.0033E+03, 2.0062E+03, 2.0090E+03, 2.0119E+03,
     + 2.0148E+03, 2.0176E+03, 2.0205E+03, 2.0234E+03, 2.0263E+03,
     +  2.0291E+03, 2.0320E+03, 2.0349E+03, 2.0378E+03, 2.0407E+03,
     + 2.0436E+03, 2.0464E+03, 2.0493E+03, 2.0522E+03, 2.0551E+03,
     +  2.0580E+03, 2.0609E+03, 2.0638E+03, 2.0667E+03, 2.0696E+03,
     + 2.0726E+03, 2.0755E+03, 2.0784E+03, 2.0813E+03, 2.0842E+03,
     +  2.0871E+03, 2.0900E+03, 2.0930E+03, 2.0959E+03, 2.0988E+03,
     + 2.1018E+03, 2.1047E+03, 2.1076E+03, 2.1106E+03, 2.1135E+03,
     +  2.1164E+03, 2.1194E+03, 2.1223E+03, 2.1253E+03, 2.1282E+03,
     + 2.1312E+03, 2.1341E+03, 2.1371E+03, 2.1400E+03, 2.1430E+03,
     +  2.1459E+03, 2.1489E+03, 2.1519E+03, 2.1548E+03, 2.1578E+03,
     + 2.1608E+03, 2.1638E+03, 2.1667E+03, 2.1697E+03, 2.1727E+03,
     +  2.1757E+03, 2.1787E+03, 2.1816E+03, 2.1846E+03, 2.1876E+03,
     + 2.1906E+03, 2.1936E+03, 2.1966E+03, 2.1996E+03, 2.2026E+03,
     +  2.2056E+03, 2.2086E+03, 2.2116E+03, 2.2146E+03, 2.2176E+03,
     + 2.2207E+03, 2.2237E+03, 2.2267E+03, 2.2297E+03, 2.2327E+03,
     +  2.2358E+03, 2.2388E+03, 2.2418E+03, 2.2448E+03, 2.2479E+03,
     + 2.2509E+03, 2.2539E+03, 2.2570E+03, 2.2600E+03, 2.2631E+03,
     +  2.2661E+03, 2.2692E+03, 2.2722E+03, 2.2753E+03, 2.2783E+03,
     + 2.2814E+03, 2.2844E+03, 2.2875E+03, 2.2905E+03, 2.2936E+03,
     +  2.2967E+03, 2.2997E+03, 2.3028E+03, 2.3059E+03, 2.3090E+03,
     + 2.3120E+03, 2.3151E+03, 2.3182E+03, 2.3213E+03, 2.3244E+03,
     +  2.3275E+03, 2.3305E+03, 2.3336E+03, 2.3367E+03, 2.3398E+03,
     + 2.3429E+03, 2.3460E+03, 2.3491E+03, 2.3522E+03, 2.3553E+03,
     +  2.3584E+03, 2.3616E+03, 2.3647E+03, 2.3678E+03, 2.3709E+03,
     + 2.3740E+03, 2.3771E+03, 2.3803E+03, 2.3834E+03, 2.3865E+03,
     +  2.3896E+03, 2.3928E+03, 2.3959E+03, 2.3990E+03, 2.4022E+03,
     + 2.4053E+03, 2.4085E+03, 2.4116E+03, 2.4148E+03, 2.4179E+03,
     +  2.4211E+03, 2.4242E+03, 2.4274E+03, 2.4305E+03, 2.4337E+03,
     + 2.4368E+03, 2.4400E+03, 2.4432E+03, 2.4463E+03, 2.4495E+03,
     +  2.4527E+03, 2.4559E+03, 2.4590E+03, 2.4622E+03, 2.4654E+03,
     + 2.4686E+03, 2.4718E+03, 2.4749E+03, 2.4781E+03, 2.4813E+03,
     +  2.4845E+03, 2.4877E+03, 2.4909E+03, 2.4941E+03, 2.4973E+03,
     + 2.5005E+03, 2.5037E+03, 2.5069E+03, 2.5101E+03, 2.5133E+03,
     +  2.5166E+03, 2.5198E+03, 2.5230E+03, 2.5262E+03, 2.5294E+03,
     + 2.5327E+03, 2.5359E+03, 2.5391E+03, 2.5423E+03, 2.5456E+03,
     +  2.5488E+03, 2.5521E+03, 2.5553E+03, 2.5585E+03, 2.5618E+03,
     + 2.5650E+03, 2.5683E+03, 2.5715E+03, 2.5748E+03, 2.5780E+03,
     +  2.5813E+03, 2.5845E+03, 2.5878E+03, 2.5911E+03, 2.5943E+03,
     + 2.5976E+03, 2.6009E+03, 2.6041E+03, 2.6074E+03, 2.6107E+03,
     +  2.6140E+03, 2.6173E+03, 2.6205E+03, 2.6238E+03, 2.6271E+03,
     + 2.6304E+03, 2.6337E+03, 2.6370E+03, 2.6403E+03, 2.6436E+03,
     +  2.6469E+03, 2.6502E+03, 2.6535E+03, 2.6568E+03, 2.6601E+03,
     + 2.6634E+03, 2.6667E+03, 2.6700E+03, 2.6733E+03, 2.6767E+03,
     +  2.6800E+03, 2.6833E+03, 2.6866E+03, 2.6900E+03, 2.6933E+03,
     + 2.6966E+03, 2.6999E+03, 2.7033E+03, 2.7066E+03, 2.7100E+03,
     +  2.7133E+03, 2.7166E+03, 2.7200E+03, 2.7233E+03, 2.7267E+03,
     + 2.7300E+03, 2.7334E+03, 2.7368E+03, 2.7401E+03, 2.7435E+03,
     +  2.7468E+03, 2.7502E+03, 2.7536E+03, 2.7569E+03, 2.7603E+03,
     + 2.7637E+03, 2.7671E+03, 2.7704E+03, 2.7738E+03, 2.7772E+03,
     +  2.7806E+03, 2.7840E+03, 2.7874E+03, 2.7908E+03, 2.7942E+03,
     + 2.7975E+03, 2.8009E+03, 2.8043E+03, 2.8077E+03, 2.8111E+03,
     +  2.8146E+03, 2.8180E+03, 2.8214E+03, 2.8248E+03, 2.8282E+03,
     + 2.8316E+03, 2.8350E+03, 2.8385E+03, 2.8419E+03, 2.8453E+03,
     +  2.8487E+03, 2.8522E+03, 2.8556E+03, 2.8590E+03, 2.8625E+03,
     + 2.8659E+03, 2.8693E+03, 2.8728E+03, 2.8762E+03, 2.8797E+03,
     +  2.8831E+03, 2.8866E+03, 2.8900E+03, 2.8935E+03, 2.8969E+03,
     + 2.9004E+03, 2.9039E+03, 2.9073E+03, 2.9108E+03, 2.9143E+03,
     +  2.9177E+03, 2.9212E+03, 2.9247E+03, 2.9282E+03, 2.9316E+03,
     + 2.9351E+03, 2.9386E+03, 2.9421E+03, 2.9456E+03, 2.9491E+03,
     +  2.9525E+03, 2.9560E+03, 2.9595E+03, 2.9630E+03, 2.9665E+03,
     + 2.9700E+03, 2.9735E+03, 2.9771E+03, 2.9806E+03, 2.9841E+03,
     +  2.9876E+03, 2.9911E+03, 2.9946E+03, 2.9981E+03, 3.0017E+03,
     + 3.0052E+03, 3.0087E+03, 3.0122E+03, 3.0158E+03, 3.0193E+03,
     +  3.0228E+03, 3.0264E+03, 3.0299E+03, 3.0335E+03, 3.0370E+03,
     + 3.0405E+03, 3.0441E+03, 3.0476E+03, 3.0512E+03, 3.0547E+03,
     +  3.0583E+03, 3.0619E+03, 3.0654E+03, 3.0690E+03, 3.0725E+03,
     + 3.0761E+03, 3.0797E+03, 3.0833E+03, 3.0868E+03, 3.0904E+03,
     +  3.0940E+03, 3.0976E+03, 3.1011E+03, 3.1047E+03, 3.1083E+03,
     + 3.1119E+03, 3.1155E+03, 3.1191E+03, 3.1227E+03, 3.1263E+03,
     +  3.1299E+03, 3.1335E+03, 3.1371E+03, 3.1407E+03, 3.1443E+03,
     + 3.1479E+03, 3.1515E+03, 3.1551E+03, 3.1587E+03, 3.1624E+03,
     +  3.1660E+03, 3.1696E+03, 3.1732E+03, 3.1769E+03, 3.1805E+03,
     + 3.1841E+03, 3.1878E+03, 3.1914E+03, 3.1950E+03, 3.1987E+03,
     +  3.2023E+03, 3.2060E+03, 3.2096E+03, 3.2132E+03, 3.2169E+03,
     + 3.2206E+03, 3.2242E+03, 3.2279E+03, 3.2315E+03, 3.2352E+03,
     +  3.2388E+03, 3.2425E+03, 3.2462E+03, 3.2499E+03, 3.2535E+03,
     + 3.2572E+03, 3.2609E+03, 3.2646E+03, 3.2682E+03, 3.2719E+03,
     +  3.2756E+03, 3.2793E+03, 3.2830E+03, 3.2867E+03, 3.2904E+03,
     + 3.2941E+03, 3.2978E+03, 3.3015E+03, 3.3052E+03, 3.3089E+03,
     +  3.3126E+03, 3.3163E+03, 3.3200E+03, 3.3237E+03, 3.3274E+03,
     + 3.3311E+03, 3.3349E+03, 3.3386E+03, 3.3423E+03, 3.3460E+03,
     +  3.3498E+03, 3.3535E+03, 3.3572E+03, 3.3610E+03, 3.3647E+03,
     + 3.3684E+03, 3.3722E+03, 3.3759E+03, 3.3797E+03, 3.3834E+03,
     +  3.3872E+03, 3.3909E+03, 3.3947E+03, 3.3984E+03, 3.4022E+03,
     + 3.4060E+03, 3.4097E+03, 3.4135E+03, 3.4173E+03, 3.4210E+03,
     +  3.4248E+03, 3.4286E+03, 3.4323E+03, 3.4361E+03, 3.4399E+03,
     + 3.4437E+03, 3.4475E+03, 3.4513E+03, 3.4551E+03, 3.4588E+03,
     +  3.4626E+03, 3.4664E+03, 3.4702E+03, 3.4740E+03, 3.4778E+03,
     + 3.4816E+03, 3.4854E+03, 3.4893E+03, 3.4931E+03, 3.4969E+03,
     +  3.5007E+03, 3.5045E+03, 3.5083E+03, 3.5122E+03, 3.5160E+03,
     + 3.5198E+03, 3.5236E+03, 3.5275E+03, 3.5313E+03, 3.5351E+03,
     +  3.5390E+03, 3.5428E+03, 3.5466E+03, 3.5505E+03, 3.5543E+03,
     + 3.5582E+03, 3.5620E+03, 3.5659E+03, 3.5697E+03, 3.5736E+03,
     +  3.5775E+03, 3.5813E+03, 3.5852E+03, 3.5890E+03, 3.5929E+03,
     + 3.5968E+03, 3.6007E+03, 3.6045E+03, 3.6084E+03, 3.6123E+03,
     +  3.6162E+03, 3.6200E+03, 3.6239E+03, 3.6278E+03, 3.6317E+03,
     + 3.6356E+03, 3.6395E+03, 3.6434E+03, 3.6473E+03, 3.6512E+03,
     +  3.6551E+03, 3.6590E+03, 3.6629E+03, 3.6668E+03, 3.6707E+03,
     + 3.6746E+03, 3.6785E+03, 3.6825E+03, 3.6864E+03, 3.6903E+03,
     +  3.6942E+03, 3.6982E+03, 3.7021E+03, 3.7060E+03, 3.7099E+03,
     + 3.7139E+03, 3.7178E+03, 3.7218E+03, 3.7257E+03, 3.7296E+03,
     +  3.7336E+03, 3.7375E+03, 3.7415E+03, 3.7454E+03, 3.7494E+03,
     + 3.7534E+03, 3.7573E+03, 3.7613E+03, 3.7652E+03, 3.7692E+03,
     +  3.7732E+03, 3.7771E+03, 3.7811E+03, 3.7851E+03, 3.7891E+03,
     + 3.7930E+03, 3.7970E+03, 3.8010E+03, 3.8050E+03, 3.8090E+03,
     +  3.8130E+03, 3.8170E+03, 3.8209E+03, 3.8249E+03, 3.8289E+03,
     + 3.8329E+03, 3.8369E+03, 3.8410E+03, 3.8450E+03, 3.8490E+03,
     +  3.8530E+03, 3.8570E+03, 3.8610E+03, 3.8650E+03, 3.8690E+03,
     + 3.8731E+03, 3.8771E+03, 3.8811E+03, 3.8851E+03, 3.8892E+03,
     +  3.8932E+03, 3.8972E+03, 3.9013E+03, 3.9053E+03, 3.9094E+03,
     + 3.9134E+03, 3.9175E+03, 3.9215E+03, 3.9256E+03, 3.9296E+03,
     +  3.9337E+03, 3.9377E+03, 3.9418E+03, 3.9458E+03, 3.9499E+03,
     + 3.9540E+03, 3.9580E+03, 3.9621E+03, 3.9662E+03, 3.9703E+03,
     +  3.9743E+03, 3.9784E+03, 3.9825E+03, 3.9866E+03, 3.9907E+03,
     + 3.9947E+03, 3.9988E+03, 4.0029E+03, 4.0070E+03, 4.0111E+03,
     +  4.0152E+03, 4.0193E+03, 4.0234E+03, 4.0275E+03, 4.0316E+03,
     + 4.0357E+03, 4.0399E+03, 4.0440E+03, 4.0481E+03, 4.0522E+03,
     +  4.0563E+03, 4.0605E+03, 4.0646E+03, 4.0687E+03, 4.0728E+03,
     + 4.0770E+03, 4.0811E+03, 4.0852E+03, 4.0894E+03, 4.0935E+03,
     +  4.0977E+03, 4.1018E+03, 4.1059E+03, 4.1101E+03, 4.1142E+03,
     + 4.1184E+03, 4.1226E+03, 4.1267E+03, 4.1309E+03, 4.1350E+03,
     +  4.1392E+03, 4.1434E+03, 4.1475E+03, 4.1517E+03, 4.1559E+03,
     + 4.1601E+03, 4.1642E+03, 4.1684E+03, 4.1726E+03, 4.1768E+03,
     +  4.1810E+03, 4.1852E+03, 4.1893E+03, 4.1935E+03, 4.1977E+03,
     + 4.2019E+03, 4.2061E+03, 4.2103E+03, 4.2145E+03, 4.2187E+03,
     +  4.2230E+03, 4.2272E+03, 4.2314E+03, 4.2356E+03, 4.2398E+03,
     + 4.2440E+03, 4.2482E+03, 4.2525E+03, 4.2567E+03, 4.2609E+03,
     +  4.2652E+03, 4.2694E+03, 4.2736E+03, 4.2779E+03, 4.2821E+03,
     + 4.2863E+03, 4.2906E+03, 4.2948E+03, 4.2991E+03, 4.3033E+03,
     +  4.3076E+03, 4.3118E+03, 4.3161E+03, 4.3204E+03, 4.3246E+03,
     + 4.3289E+03, 4.3331E+03, 4.3374E+03, 4.3417E+03, 4.3460E+03,
     +  4.3502E+03, 4.3545E+03, 4.3588E+03, 4.3631E+03, 4.3673E+03,
     + 4.3716E+03, 4.3759E+03, 4.3802E+03, 4.3845E+03, 4.3888E+03,
     +  4.3931E+03, 4.3974E+03, 4.4017E+03, 4.4060E+03, 4.4103E+03,
     + 4.4146E+03, 4.4189E+03, 4.4232E+03, 4.4275E+03, 4.4319E+03,
     +  4.4362E+03, 4.4405E+03, 4.4448E+03, 4.4491E+03, 4.4535E+03,
     + 4.4578E+03, 4.4621E+03, 4.4665E+03, 4.4708E+03, 4.4751E+03,
     +  4.4795E+03, 4.4838E+03, 4.4882E+03, 4.4925E+03, 4.4969E+03,
     + 4.5012E+03, 4.5056E+03, 4.5099E+03, 4.5143E+03, 4.5186E+03,
     +  4.5230E+03, 4.5274E+03, 4.5317E+03, 4.5361E+03, 4.5405E+03,
     + 4.5449E+03, 4.5492E+03, 4.5536E+03, 4.5580E+03, 4.5624E+03,
     +  4.5668E+03, 4.5711E+03, 4.5755E+03, 4.5799E+03, 4.5843E+03,
     + 4.5887E+03, 4.5931E+03, 4.5975E+03, 4.6019E+03, 4.6063E+03,
     +  4.6107E+03, 4.6151E+03, 4.6195E+03, 4.6240E+03, 4.6284E+03,
     + 4.6328E+03, 4.6372E+03, 4.6416E+03, 4.6461E+03, 4.6505E+03,
     +  4.6549E+03, 4.6593E+03, 4.6638E+03, 4.6682E+03, 4.6727E+03,
     + 4.6771E+03, 4.6815E+03, 4.6860E+03, 4.6904E+03, 4.6949E+03,
     +  4.6993E+03, 4.7038E+03, 4.7082E+03, 4.7127E+03, 4.7172E+03,
     + 4.7216E+03, 4.7261E+03, 4.7305E+03, 4.7350E+03, 4.7395E+03,
     +  4.7440E+03, 4.7484E+03, 4.7529E+03, 4.7574E+03, 4.7619E+03,
     + 4.7664E+03, 4.7708E+03, 4.7753E+03, 4.7798E+03, 4.7843E+03,
     +  4.7888E+03, 4.7933E+03, 4.7978E+03, 4.8023E+03, 4.8068E+03,
     + 4.8113E+03, 4.8158E+03, 4.8203E+03, 4.8249E+03, 4.8294E+03,
     +  4.8339E+03, 4.8384E+03, 4.8429E+03, 4.8475E+03, 4.8520E+03,
     + 4.8565E+03, 4.8610E+03, 4.8656E+03, 4.8701E+03, 4.8746E+03,
     +  4.8792E+03, 4.8837E+03, 4.8883E+03, 4.8928E+03, 4.8974E+03,
     + 4.9019E+03, 4.9065E+03, 4.9110E+03, 4.9156E+03, 4.9201E+03,
     +  4.9247E+03, 4.9293E+03, 4.9338E+03, 4.9384E+03, 4.9430E+03,
     + 4.9475E+03, 4.9521E+03, 4.9567E+03, 4.9613E+03, 4.9659E+03,
     +  4.9704E+03, 4.9750E+03, 4.9796E+03, 4.9842E+03, 4.9888E+03,
     + 4.9934E+03, 4.9980E+03, 5.0026E+03, 5.0072E+03, 5.0118E+03,
     +  5.0164E+03, 5.0210E+03, 5.0256E+03, 5.0302E+03, 5.0349E+03,
     + 5.0395E+03, 5.0441E+03, 5.0487E+03, 5.0533E+03, 5.0580E+03,
     +  5.0626E+03, 5.0672E+03, 5.0719E+03, 5.0765E+03, 5.0811E+03,
     + 5.0858E+03, 5.0904E+03, 5.0951E+03, 5.0997E+03, 5.1044E+03,
     +  5.1090E+03, 5.1137E+03, 5.1183E+03, 5.1230E+03, 5.1276E+03,
     + 5.1323E+03, 5.1370E+03, 5.1416E+03, 5.1463E+03, 5.1510E+03,
     +  5.1556E+03, 5.1603E+03, 5.1650E+03, 5.1697E+03, 5.1743E+03,
     + 5.1790E+03, 5.1837E+03, 5.1884E+03, 5.1931E+03, 5.1978E+03,
     +  5.2025E+03, 5.2072E+03, 5.2119E+03, 5.2166E+03, 5.2213E+03,
     + 5.2260E+03, 5.2307E+03, 5.2354E+03, 5.2401E+03, 5.2448E+03,
     +  5.2496E+03, 5.2543E+03, 5.2590E+03, 5.2637E+03, 5.2685E+03,
     + 5.2732E+03, 5.2779E+03, 5.2827E+03, 5.2874E+03, 5.2921E+03,
     +  5.2969E+03, 5.3016E+03, 5.3064E+03, 5.3111E+03, 5.3158E+03,
     + 5.3206E+03, 5.3254E+03, 5.3301E+03, 5.3349E+03, 5.3396E+03,
     +  5.3444E+03, 5.3492E+03, 5.3539E+03, 5.3587E+03, 5.3635E+03,
     + 5.3682E+03, 5.3730E+03, 5.3778E+03, 5.3826E+03, 5.3873E+03,
     +  5.3921E+03, 5.3969E+03, 5.4017E+03, 5.4065E+03, 5.4113E+03,
     + 5.4161E+03, 5.4209E+03, 5.4257E+03, 5.4305E+03, 5.4353E+03,
     +  5.4401E+03, 5.4449E+03, 5.4497E+03, 5.4545E+03, 5.4594E+03,
     + 5.4642E+03, 5.4690E+03, 5.4738E+03, 5.4786E+03, 5.4835E+03,
     +  5.4883E+03, 5.4931E+03, 5.4980E+03, 5.5028E+03, 5.5076E+03,
     + 5.5125E+03, 5.5173E+03, 5.5222E+03, 5.5270E+03, 5.5319E+03,
     +  5.5367E+03, 5.5416E+03, 5.5464E+03, 5.5513E+03, 5.5561E+03,
     + 5.5610E+03, 5.5659E+03, 5.5707E+03, 5.5756E+03, 5.5805E+03,
     +  5.5854E+03, 5.5902E+03, 5.5951E+03, 5.6000E+03, 5.6049E+03,
     + 5.6098E+03, 5.6147E+03, 5.6195E+03, 5.6244E+03, 5.6293E+03,
     +  5.6342E+03, 5.6391E+03, 5.6440E+03, 5.6489E+03, 5.6538E+03,
     + 5.6587E+03, 5.6637E+03, 5.6686E+03, 5.6735E+03, 5.6784E+03,
     +  5.6833E+03, 5.6882E+03, 5.6932E+03, 5.6981E+03, 5.7030E+03,
     + 5.7079E+03, 5.7129E+03, 5.7178E+03, 5.7228E+03, 5.7277E+03,
     +  5.7326E+03, 5.7376E+03, 5.7425E+03, 5.7475E+03, 5.7524E+03,
     + 5.7574E+03, 5.7623E+03, 5.7673E+03, 5.7722E+03, 5.7772E+03,
     +  5.7822E+03, 5.7871E+03, 5.7921E+03, 5.7971E+03, 5.8021E+03,
     + 5.8070E+03, 5.8120E+03, 5.8170E+03, 5.8220E+03, 5.8269E+03,
     +  5.8319E+03, 5.8369E+03, 5.8419E+03, 5.8469E+03, 5.8519E+03,
     + 5.8569E+03, 5.8619E+03, 5.8669E+03, 5.8719E+03, 5.8769E+03,
     +  5.8819E+03, 5.8869E+03, 5.8920E+03, 5.8970E+03, 5.9020E+03,
     + 5.9070E+03, 5.9120E+03, 5.9171E+03, 5.9221E+03, 5.9271E+03,
     +  5.9321E+03, 5.9372E+03, 5.9422E+03, 5.9472E+03, 5.9523E+03,
     + 5.9573E+03, 5.9624E+03, 5.9674E+03, 5.9725E+03, 5.9775E+03,
     +  5.9826E+03, 5.9876E+03, 5.9927E+03, 5.9978E+03, 6.0028E+03,
     + 6.0079E+03, 6.0130E+03, 6.0180E+03, 6.0231E+03, 6.0282E+03,
     +  6.0332E+03, 6.0383E+03, 6.0434E+03, 6.0485E+03, 6.0536E+03,
     + 6.0587E+03, 6.0638E+03, 6.0688E+03, 6.0739E+03, 6.0790E+03,
     +  6.0841E+03, 6.0892E+03, 6.0943E+03, 6.0994E+03, 6.1046E+03,
     + 6.1097E+03, 6.1148E+03, 6.1199E+03, 6.1250E+03, 6.1301E+03,
     +  6.1353E+03, 6.1404E+03, 6.1455E+03, 6.1506E+03, 6.1558E+03,
     + 6.1609E+03, 6.1660E+03, 6.1712E+03, 6.1763E+03, 6.1814E+03,
     +  6.1866E+03, 6.1917E+03, 6.1969E+03, 6.2020E+03, 6.2072E+03,
     + 6.2124E+03, 6.2175E+03, 6.2227E+03, 6.2278E+03, 6.2330E+03,
     +  6.2382E+03, 6.2433E+03, 6.2485E+03, 6.2537E+03, 6.2589E+03,
     + 6.2640E+03, 6.2692E+03, 6.2744E+03, 6.2796E+03, 6.2848E+03,
     +  6.2900E+03, 6.2951E+03, 6.3003E+03, 6.3055E+03, 6.3107E+03,
     + 6.3159E+03, 6.3211E+03, 6.3263E+03, 6.3316E+03, 6.3368E+03,
     +  6.3420E+03, 6.3472E+03, 6.3524E+03, 6.3576E+03, 6.3628E+03,
     + 6.3681E+03, 6.3733E+03, 6.3785E+03, 6.3838E+03, 6.3890E+03,
     +  6.3942E+03, 6.3995E+03, 6.4047E+03, 6.4099E+03, 6.4152E+03,
     + 6.4204E+03, 6.4257E+03, 6.4309E+03, 6.4362E+03, 6.4414E+03,
     +  6.4467E+03, 6.4520E+03, 6.4572E+03, 6.4625E+03, 6.4678E+03,
     + 6.4730E+03, 6.4783E+03, 6.4836E+03, 6.4888E+03, 6.4941E+03,
     +  6.4994E+03, 6.5047E+03, 6.5100E+03, 6.5153E+03, 6.5205E+03,
     + 6.5258E+03, 6.5311E+03, 6.5364E+03, 6.5417E+03, 6.5470E+03,
     +  6.5523E+03, 6.5576E+03, 6.5629E+03, 6.5682E+03, 6.5736E+03,
     + 6.5789E+03, 6.5842E+03, 6.5895E+03, 6.5948E+03, 6.6002E+03,
     +  6.6055E+03, 6.6108E+03, 6.6161E+03, 6.6215E+03, 6.6268E+03,
     + 6.6321E+03, 6.6375E+03, 6.6428E+03, 6.6482E+03, 6.6535E+03,
     +  6.6589E+03, 6.6642E+03, 6.6696E+03, 6.6749E+03, 6.6803E+03,
     + 6.6856E+03, 6.6910E+03, 6.6964E+03, 6.7017E+03, 6.7071E+03,
     +  6.7125E+03, 6.7178E+03, 6.7232E+03, 6.7286E+03, 6.7340E+03,
     + 6.7394E+03, 6.7447E+03, 6.7501E+03, 6.7555E+03, 6.7609E+03,
     +  6.7663E+03, 6.7717E+03, 6.7771E+03, 6.7825E+03, 6.7879E+03,
     + 6.7933E+03, 6.7987E+03, 6.8041E+03, 6.8095E+03, 6.8149E+03,
     +  6.8204E+03, 6.8258E+03, 6.8312E+03, 6.8366E+03, 6.8421E+03,
     + 6.8475E+03, 6.8529E+03, 6.8583E+03, 6.8638E+03, 6.8692E+03,
     +  6.8747E+03, 6.8801E+03, 6.8855E+03, 6.8910E+03, 6.8964E+03,
     + 6.9019E+03, 6.9073E+03, 6.9128E+03, 6.9183E+03, 6.9237E+03,
     +  6.9292E+03, 6.9346E+03, 6.9401E+03, 6.9456E+03, 6.9510E+03,
     + 6.9565E+03, 6.9620E+03, 6.9675E+03, 6.9730E+03, 6.9784E+03,
     +  6.9839E+03, 6.9894E+03, 6.9949E+03, 7.0004E+03, 7.0059E+03,
     + 7.0114E+03, 7.0169E+03, 7.0224E+03, 7.0279E+03, 7.0334E+03,
     +  7.0389E+03, 7.0444E+03, 7.0499E+03, 7.0554E+03, 7.0610E+03,
     + 7.0665E+03, 7.0720E+03, 7.0775E+03, 7.0830E+03, 7.0886E+03,
     +  7.0941E+03, 7.0996E+03, 7.1052E+03, 7.1107E+03, 7.1162E+03,
     + 7.1218E+03, 7.1273E+03, 7.1329E+03, 7.1384E+03, 7.1440E+03,
     +  7.1495E+03, 7.1551E+03, 7.1606E+03, 7.1662E+03, 7.1718E+03,
     + 7.1773E+03, 7.1829E+03, 7.1885E+03, 7.1940E+03, 7.1996E+03,
     +  7.2052E+03, 7.2108E+03, 7.2164E+03, 7.2219E+03, 7.2275E+03,
     + 7.2331E+03, 7.2387E+03, 7.2443E+03, 7.2499E+03, 7.2555E+03,
     +  7.2611E+03, 7.2667E+03, 7.2723E+03, 7.2779E+03, 7.2835E+03,
     + 7.2891E+03, 7.2947E+03, 7.3003E+03, 7.3060E+03, 7.3116E+03,
     +  7.3172E+03, 7.3228E+03, 7.3285E+03, 7.3341E+03, 7.3397E+03,
     + 7.3453E+03, 7.3510E+03, 7.3566E+03, 7.3623E+03, 7.3679E+03,
     +  7.3735E+03, 7.3792E+03, 7.3848E+03, 7.3905E+03, 7.3962E+03,
     + 7.4018E+03, 7.4075E+03, 7.4131E+03, 7.4188E+03, 7.4245E+03,
     +  7.4301E+03, 7.4358E+03, 7.4415E+03, 7.4471E+03, 7.4528E+03,
     + 7.4585E+03, 7.4642E+03, 7.4699E+03, 7.4756E+03, 7.4812E+03,
     +  7.4869E+03, 7.4926E+03, 7.4983E+03, 7.5040E+03, 7.5097E+03,
     + 7.5154E+03, 7.5211E+03, 7.5268E+03, 7.5325E+03, 7.5383E+03,
     +  7.5440E+03, 7.5497E+03, 7.5554E+03, 7.5611E+03, 7.5668E+03,
     + 7.5726E+03, 7.5783E+03, 7.5840E+03, 7.5898E+03, 7.5955E+03,
     +  7.6012E+03, 7.6070E+03, 7.6127E+03, 7.6185E+03, 7.6242E+03,
     + 7.6299E+03, 7.6357E+03, 7.6415E+03, 7.6472E+03, 7.6530E+03,
     +  7.6587E+03, 7.6645E+03, 7.6703E+03, 7.6760E+03, 7.6818E+03,
     + 7.6876E+03, 7.6933E+03, 7.6991E+03, 7.7049E+03, 7.7107E+03,
     +  7.7164E+03, 7.7222E+03, 7.7280E+03, 7.7338E+03, 7.7396E+03,
     + 7.7454E+03, 7.7512E+03, 7.7570E+03, 7.7628E+03, 7.7686E+03,
     +  7.7744E+03, 7.7802E+03, 7.7860E+03, 7.7918E+03, 7.7976E+03,
     + 7.8035E+03, 7.8093E+03, 7.8151E+03, 7.8209E+03, 7.8268E+03,
     +  7.8326E+03, 7.8384E+03, 7.8443E+03, 7.8501E+03, 7.8559E+03,
     + 7.8618E+03, 7.8676E+03, 7.8735E+03, 7.8793E+03, 7.8852E+03,
     +  7.8910E+03, 7.8969E+03, 7.9027E+03, 7.9086E+03, 7.9144E+03,
     + 7.9203E+03, 7.9262E+03, 7.9320E+03, 7.9379E+03, 7.9438E+03,
     +  7.9496E+03, 7.9555E+03, 7.9614E+03, 7.9673E+03, 7.9732E+03,
     + 7.9791E+03, 7.9849E+03, 7.9908E+03, 7.9967E+03, 8.0026E+03,
     +  8.0085E+03, 8.0144E+03, 8.0203E+03, 8.0262E+03, 8.0321E+03/
      data (QofT( 2,J),J=2409,4510)/ 8.0380E+03, 8.0440E+03,
     +  8.0499E+03, 8.0558E+03, 8.0617E+03, 8.0676E+03, 8.0736E+03,
     +  8.0795E+03,
     + 8.0854E+03, 8.0913E+03, 8.0973E+03, 8.1032E+03, 8.1091E+03,
     +  8.1151E+03, 8.1210E+03, 8.1270E+03, 8.1329E+03, 8.1389E+03,
     + 8.1448E+03, 8.1508E+03, 8.1567E+03, 8.1627E+03, 8.1686E+03,
     +  8.1746E+03, 8.1806E+03, 8.1865E+03, 8.1925E+03, 8.1985E+03,
     + 8.2045E+03, 8.2104E+03, 8.2164E+03, 8.2224E+03, 8.2284E+03,
     +  8.2344E+03, 8.2403E+03, 8.2463E+03, 8.2523E+03, 8.2583E+03,
     + 8.2643E+03, 8.2703E+03, 8.2763E+03, 8.2823E+03, 8.2883E+03,
     +  8.2943E+03, 8.3004E+03, 8.3064E+03, 8.3124E+03, 8.3184E+03,
     + 8.3244E+03, 8.3304E+03, 8.3365E+03, 8.3425E+03, 8.3485E+03,
     +  8.3546E+03, 8.3606E+03, 8.3666E+03, 8.3727E+03, 8.3787E+03,
     + 8.3848E+03, 8.3908E+03, 8.3969E+03, 8.4029E+03, 8.4090E+03,
     +  8.4150E+03, 8.4211E+03, 8.4271E+03, 8.4332E+03, 8.4393E+03,
     + 8.4453E+03, 8.4514E+03, 8.4575E+03, 8.4635E+03, 8.4696E+03,
     +  8.4757E+03, 8.4818E+03, 8.4879E+03, 8.4939E+03, 8.5000E+03,
     + 8.5061E+03, 8.5122E+03, 8.5183E+03, 8.5244E+03, 8.5305E+03,
     +  8.5366E+03, 8.5427E+03, 8.5488E+03, 8.5549E+03, 8.5610E+03,
     + 8.5671E+03, 8.5733E+03, 8.5794E+03, 8.5855E+03, 8.5916E+03,
     +  8.5977E+03, 8.6039E+03, 8.6100E+03, 8.6161E+03, 8.6223E+03,
     + 8.6284E+03, 8.6345E+03, 8.6407E+03, 8.6468E+03, 8.6530E+03,
     +  8.6591E+03, 8.6653E+03, 8.6714E+03, 8.6776E+03, 8.6837E+03,
     + 8.6899E+03, 8.6961E+03, 8.7022E+03, 8.7084E+03, 8.7146E+03,
     +  8.7207E+03, 8.7269E+03, 8.7331E+03, 8.7393E+03, 8.7454E+03,
     + 8.7516E+03, 8.7578E+03, 8.7640E+03, 8.7702E+03, 8.7764E+03,
     +  8.7826E+03, 8.7888E+03, 8.7950E+03, 8.8012E+03, 8.8074E+03,
     + 8.8136E+03, 8.8198E+03, 8.8260E+03, 8.8322E+03, 8.8384E+03,
     +  8.8446E+03, 8.8509E+03, 8.8571E+03, 8.8633E+03, 8.8695E+03,
     + 8.8758E+03, 8.8820E+03, 8.8882E+03, 8.8945E+03, 8.9007E+03,
     +  8.9070E+03, 8.9132E+03, 8.9194E+03, 8.9257E+03, 8.9319E+03,
     + 8.9382E+03, 8.9445E+03, 8.9507E+03, 8.9570E+03, 8.9632E+03,
     +  8.9695E+03, 8.9758E+03, 8.9820E+03, 8.9883E+03, 8.9946E+03,
     + 9.0009E+03, 9.0071E+03, 9.0134E+03, 9.0197E+03, 9.0260E+03,
     +  9.0323E+03, 9.0386E+03, 9.0448E+03, 9.0511E+03, 9.0574E+03,
     + 9.0637E+03, 9.0700E+03, 9.0763E+03, 9.0827E+03, 9.0890E+03,
     +  9.0953E+03, 9.1016E+03, 9.1079E+03, 9.1142E+03, 9.1205E+03,
     + 9.1269E+03, 9.1332E+03, 9.1395E+03, 9.1458E+03, 9.1522E+03,
     +  9.1585E+03, 9.1648E+03, 9.1712E+03, 9.1775E+03, 9.1839E+03,
     + 9.1902E+03, 9.1966E+03, 9.2029E+03, 9.2093E+03, 9.2156E+03,
     +  9.2220E+03, 9.2283E+03, 9.2347E+03, 9.2411E+03, 9.2474E+03,
     + 9.2538E+03, 9.2602E+03, 9.2666E+03, 9.2729E+03, 9.2793E+03,
     +  9.2857E+03, 9.2921E+03, 9.2985E+03, 9.3048E+03, 9.3112E+03,
     + 9.3176E+03, 9.3240E+03, 9.3304E+03, 9.3368E+03, 9.3432E+03,
     +  9.3496E+03, 9.3560E+03, 9.3624E+03, 9.3689E+03, 9.3753E+03,
     + 9.3817E+03, 9.3881E+03, 9.3945E+03, 9.4010E+03, 9.4074E+03,
     +  9.4138E+03, 9.4202E+03, 9.4267E+03, 9.4331E+03, 9.4395E+03,
     + 9.4460E+03, 9.4524E+03, 9.4589E+03, 9.4653E+03, 9.4718E+03,
     +  9.4782E+03, 9.4847E+03, 9.4911E+03, 9.4976E+03, 9.5040E+03,
     + 9.5105E+03, 9.5170E+03, 9.5234E+03, 9.5299E+03, 9.5364E+03,
     +  9.5429E+03, 9.5493E+03, 9.5558E+03, 9.5623E+03, 9.5688E+03,
     + 9.5753E+03, 9.5818E+03, 9.5882E+03, 9.5947E+03, 9.6012E+03,
     +  9.6077E+03, 9.6142E+03, 9.6207E+03, 9.6272E+03, 9.6337E+03,
     + 9.6403E+03, 9.6468E+03, 9.6533E+03, 9.6598E+03, 9.6663E+03,
     +  9.6728E+03, 9.6794E+03, 9.6859E+03, 9.6924E+03, 9.6990E+03,
     + 9.7055E+03, 9.7120E+03, 9.7186E+03, 9.7251E+03, 9.7316E+03,
     +  9.7382E+03, 9.7447E+03, 9.7513E+03, 9.7578E+03, 9.7644E+03,
     + 9.7709E+03, 9.7775E+03, 9.7841E+03, 9.7906E+03, 9.7972E+03,
     +  9.8038E+03, 9.8103E+03, 9.8169E+03, 9.8235E+03, 9.8301E+03,
     + 9.8366E+03, 9.8432E+03, 9.8498E+03, 9.8564E+03, 9.8630E+03,
     +  9.8696E+03, 9.8762E+03, 9.8828E+03, 9.8894E+03, 9.8960E+03,
     + 9.9026E+03, 9.9092E+03, 9.9158E+03, 9.9224E+03, 9.9290E+03,
     +  9.9356E+03, 9.9422E+03, 9.9489E+03, 9.9555E+03, 9.9621E+03,
     + 9.9687E+03, 9.9754E+03, 9.9820E+03, 9.9886E+03, 9.9953E+03,
     +  1.0002E+04, 1.0009E+04, 1.0015E+04, 1.0022E+04, 1.0028E+04,
     + 1.0035E+04, 1.0042E+04, 1.0048E+04, 1.0055E+04, 1.0062E+04,
     +  1.0068E+04, 1.0075E+04, 1.0082E+04, 1.0088E+04, 1.0095E+04,
     + 1.0102E+04, 1.0108E+04, 1.0115E+04, 1.0122E+04, 1.0129E+04,
     +  1.0135E+04, 1.0142E+04, 1.0149E+04, 1.0155E+04, 1.0162E+04,
     + 1.0169E+04, 1.0175E+04, 1.0182E+04, 1.0189E+04, 1.0195E+04,
     +  1.0202E+04, 1.0209E+04, 1.0216E+04, 1.0222E+04, 1.0229E+04,
     + 1.0236E+04, 1.0242E+04, 1.0249E+04, 1.0256E+04, 1.0263E+04,
     +  1.0269E+04, 1.0276E+04, 1.0283E+04, 1.0290E+04, 1.0296E+04,
     + 1.0303E+04, 1.0310E+04, 1.0317E+04, 1.0323E+04, 1.0330E+04,
     +  1.0337E+04, 1.0344E+04, 1.0350E+04, 1.0357E+04, 1.0364E+04,
     + 1.0371E+04, 1.0377E+04, 1.0384E+04, 1.0391E+04, 1.0398E+04,
     +  1.0405E+04, 1.0411E+04, 1.0418E+04, 1.0425E+04, 1.0432E+04,
     + 1.0439E+04, 1.0445E+04, 1.0452E+04, 1.0459E+04, 1.0466E+04,
     +  1.0473E+04, 1.0479E+04, 1.0486E+04, 1.0493E+04, 1.0500E+04,
     + 1.0507E+04, 1.0513E+04, 1.0520E+04, 1.0527E+04, 1.0534E+04,
     +  1.0541E+04, 1.0548E+04, 1.0554E+04, 1.0561E+04, 1.0568E+04,
     + 1.0575E+04, 1.0582E+04, 1.0589E+04, 1.0595E+04, 1.0602E+04,
     +  1.0609E+04, 1.0616E+04, 1.0623E+04, 1.0630E+04, 1.0637E+04,
     + 1.0643E+04, 1.0650E+04, 1.0657E+04, 1.0664E+04, 1.0671E+04,
     +  1.0678E+04, 1.0685E+04, 1.0691E+04, 1.0698E+04, 1.0705E+04,
     + 1.0712E+04, 1.0719E+04, 1.0726E+04, 1.0733E+04, 1.0740E+04,
     +  1.0747E+04, 1.0754E+04, 1.0760E+04, 1.0767E+04, 1.0774E+04,
     + 1.0781E+04, 1.0788E+04, 1.0795E+04, 1.0802E+04, 1.0809E+04,
     +  1.0816E+04, 1.0823E+04, 1.0830E+04, 1.0837E+04, 1.0843E+04,
     + 1.0850E+04, 1.0857E+04, 1.0864E+04, 1.0871E+04, 1.0878E+04,
     +  1.0885E+04, 1.0892E+04, 1.0899E+04, 1.0906E+04, 1.0913E+04,
     + 1.0920E+04, 1.0927E+04, 1.0934E+04, 1.0941E+04, 1.0948E+04,
     +  1.0955E+04, 1.0962E+04, 1.0969E+04, 1.0976E+04, 1.0983E+04,
     + 1.0990E+04, 1.0997E+04, 1.1004E+04, 1.1011E+04, 1.1018E+04,
     +  1.1025E+04, 1.1032E+04, 1.1039E+04, 1.1046E+04, 1.1053E+04,
     + 1.1060E+04, 1.1067E+04, 1.1074E+04, 1.1081E+04, 1.1088E+04,
     +  1.1095E+04, 1.1102E+04, 1.1109E+04, 1.1116E+04, 1.1123E+04,
     + 1.1130E+04, 1.1137E+04, 1.1144E+04, 1.1151E+04, 1.1158E+04,
     +  1.1165E+04, 1.1172E+04, 1.1179E+04, 1.1186E+04, 1.1193E+04,
     + 1.1200E+04, 1.1207E+04, 1.1214E+04, 1.1221E+04, 1.1228E+04,
     +  1.1235E+04, 1.1242E+04, 1.1250E+04, 1.1257E+04, 1.1264E+04,
     + 1.1271E+04, 1.1278E+04, 1.1285E+04, 1.1292E+04, 1.1299E+04,
     +  1.1306E+04, 1.1313E+04, 1.1320E+04, 1.1327E+04, 1.1335E+04,
     + 1.1342E+04, 1.1349E+04, 1.1356E+04, 1.1363E+04, 1.1370E+04,
     +  1.1377E+04, 1.1384E+04, 1.1391E+04, 1.1399E+04, 1.1406E+04,
     + 1.1413E+04, 1.1420E+04, 1.1427E+04, 1.1434E+04, 1.1441E+04,
     +  1.1448E+04, 1.1456E+04, 1.1463E+04, 1.1470E+04, 1.1477E+04,
     + 1.1484E+04, 1.1491E+04, 1.1498E+04, 1.1506E+04, 1.1513E+04,
     +  1.1520E+04, 1.1527E+04, 1.1534E+04, 1.1541E+04, 1.1549E+04,
     + 1.1556E+04, 1.1563E+04, 1.1570E+04, 1.1577E+04, 1.1585E+04,
     +  1.1592E+04, 1.1599E+04, 1.1606E+04, 1.1613E+04, 1.1620E+04,
     + 1.1628E+04, 1.1635E+04, 1.1642E+04, 1.1649E+04, 1.1656E+04,
     +  1.1664E+04, 1.1671E+04, 1.1678E+04, 1.1685E+04, 1.1693E+04,
     + 1.1700E+04, 1.1707E+04, 1.1714E+04, 1.1721E+04, 1.1729E+04,
     +  1.1736E+04, 1.1743E+04, 1.1750E+04, 1.1758E+04, 1.1765E+04,
     + 1.1772E+04, 1.1779E+04, 1.1787E+04, 1.1794E+04, 1.1801E+04,
     +  1.1808E+04, 1.1816E+04, 1.1823E+04, 1.1830E+04, 1.1837E+04,
     + 1.1845E+04, 1.1852E+04, 1.1859E+04, 1.1866E+04, 1.1874E+04,
     +  1.1881E+04, 1.1888E+04, 1.1896E+04, 1.1903E+04, 1.1910E+04,
     + 1.1917E+04, 1.1925E+04, 1.1932E+04, 1.1939E+04, 1.1947E+04,
     +  1.1954E+04, 1.1961E+04, 1.1969E+04, 1.1976E+04, 1.1983E+04,
     + 1.1990E+04, 1.1998E+04, 1.2005E+04, 1.2012E+04, 1.2020E+04,
     +  1.2027E+04, 1.2034E+04, 1.2042E+04, 1.2049E+04, 1.2056E+04,
     + 1.2064E+04, 1.2071E+04, 1.2078E+04, 1.2086E+04, 1.2093E+04,
     +  1.2100E+04, 1.2108E+04, 1.2115E+04, 1.2123E+04, 1.2130E+04,
     + 1.2137E+04, 1.2145E+04, 1.2152E+04, 1.2159E+04, 1.2167E+04,
     +  1.2174E+04, 1.2182E+04, 1.2189E+04, 1.2196E+04, 1.2204E+04,
     + 1.2211E+04, 1.2218E+04, 1.2226E+04, 1.2233E+04, 1.2241E+04,
     +  1.2248E+04, 1.2255E+04, 1.2263E+04, 1.2270E+04, 1.2278E+04,
     + 1.2285E+04, 1.2292E+04, 1.2300E+04, 1.2307E+04, 1.2315E+04,
     +  1.2322E+04, 1.2330E+04, 1.2337E+04, 1.2344E+04, 1.2352E+04,
     + 1.2359E+04, 1.2367E+04, 1.2374E+04, 1.2382E+04, 1.2389E+04,
     +  1.2396E+04, 1.2404E+04, 1.2411E+04, 1.2419E+04, 1.2426E+04,
     + 1.2434E+04, 1.2441E+04, 1.2449E+04, 1.2456E+04, 1.2464E+04,
     +  1.2471E+04, 1.2479E+04, 1.2486E+04, 1.2494E+04, 1.2501E+04,
     + 1.2508E+04, 1.2516E+04, 1.2523E+04, 1.2531E+04, 1.2538E+04,
     +  1.2546E+04, 1.2553E+04, 1.2561E+04, 1.2568E+04, 1.2576E+04,
     + 1.2583E+04, 1.2591E+04, 1.2598E+04, 1.2606E+04, 1.2613E+04,
     +  1.2621E+04, 1.2629E+04, 1.2636E+04, 1.2644E+04, 1.2651E+04,
     + 1.2659E+04, 1.2666E+04, 1.2674E+04, 1.2681E+04, 1.2689E+04,
     +  1.2696E+04, 1.2704E+04, 1.2711E+04, 1.2719E+04, 1.2727E+04,
     + 1.2734E+04, 1.2742E+04, 1.2749E+04, 1.2757E+04, 1.2764E+04,
     +  1.2772E+04, 1.2779E+04, 1.2787E+04, 1.2795E+04, 1.2802E+04,
     + 1.2810E+04, 1.2817E+04, 1.2825E+04, 1.2833E+04, 1.2840E+04,
     +  1.2848E+04, 1.2855E+04, 1.2863E+04, 1.2870E+04, 1.2878E+04,
     + 1.2886E+04, 1.2893E+04, 1.2901E+04, 1.2909E+04, 1.2916E+04,
     +  1.2924E+04, 1.2931E+04, 1.2939E+04, 1.2947E+04, 1.2954E+04,
     + 1.2962E+04, 1.2969E+04, 1.2977E+04, 1.2985E+04, 1.2992E+04,
     +  1.3000E+04, 1.3008E+04, 1.3015E+04, 1.3023E+04, 1.3031E+04,
     + 1.3038E+04, 1.3046E+04, 1.3054E+04, 1.3061E+04, 1.3069E+04,
     +  1.3077E+04, 1.3084E+04, 1.3092E+04, 1.3100E+04, 1.3107E+04,
     + 1.3115E+04, 1.3123E+04, 1.3130E+04, 1.3138E+04, 1.3146E+04,
     +  1.3153E+04, 1.3161E+04, 1.3169E+04, 1.3176E+04, 1.3184E+04,
     + 1.3192E+04, 1.3199E+04, 1.3207E+04, 1.3215E+04, 1.3223E+04,
     +  1.3230E+04, 1.3238E+04, 1.3246E+04, 1.3253E+04, 1.3261E+04,
     + 1.3269E+04, 1.3277E+04, 1.3284E+04, 1.3292E+04, 1.3300E+04,
     +  1.3308E+04, 1.3315E+04, 1.3323E+04, 1.3331E+04, 1.3339E+04,
     + 1.3346E+04, 1.3354E+04, 1.3362E+04, 1.3370E+04, 1.3377E+04,
     +  1.3385E+04, 1.3393E+04, 1.3401E+04, 1.3408E+04, 1.3416E+04,
     + 1.3424E+04, 1.3432E+04, 1.3439E+04, 1.3447E+04, 1.3455E+04,
     +  1.3463E+04, 1.3471E+04, 1.3478E+04, 1.3486E+04, 1.3494E+04,
     + 1.3502E+04, 1.3510E+04, 1.3517E+04, 1.3525E+04, 1.3533E+04,
     +  1.3541E+04, 1.3549E+04, 1.3556E+04, 1.3564E+04, 1.3572E+04,
     + 1.3580E+04, 1.3588E+04, 1.3595E+04, 1.3603E+04, 1.3611E+04,
     +  1.3619E+04, 1.3627E+04, 1.3635E+04, 1.3642E+04, 1.3650E+04,
     + 1.3658E+04, 1.3666E+04, 1.3674E+04, 1.3682E+04, 1.3690E+04,
     +  1.3697E+04, 1.3705E+04, 1.3713E+04, 1.3721E+04, 1.3729E+04,
     + 1.3737E+04, 1.3745E+04, 1.3752E+04, 1.3760E+04, 1.3768E+04,
     +  1.3776E+04, 1.3784E+04, 1.3792E+04, 1.3800E+04, 1.3808E+04,
     + 1.3816E+04, 1.3823E+04, 1.3831E+04, 1.3839E+04, 1.3847E+04,
     +  1.3855E+04, 1.3863E+04, 1.3871E+04, 1.3879E+04, 1.3887E+04,
     + 1.3895E+04, 1.3903E+04, 1.3910E+04, 1.3918E+04, 1.3926E+04,
     +  1.3934E+04, 1.3942E+04, 1.3950E+04, 1.3958E+04, 1.3966E+04,
     + 1.3974E+04, 1.3982E+04, 1.3990E+04, 1.3998E+04, 1.4006E+04,
     +  1.4014E+04, 1.4022E+04, 1.4030E+04, 1.4037E+04, 1.4045E+04,
     + 1.4053E+04, 1.4061E+04, 1.4069E+04, 1.4077E+04, 1.4085E+04,
     +  1.4093E+04, 1.4101E+04, 1.4109E+04, 1.4117E+04, 1.4125E+04,
     + 1.4133E+04, 1.4141E+04, 1.4149E+04, 1.4157E+04, 1.4165E+04,
     +  1.4173E+04, 1.4181E+04, 1.4189E+04, 1.4197E+04, 1.4205E+04,
     + 1.4213E+04, 1.4221E+04, 1.4229E+04, 1.4237E+04, 1.4245E+04,
     +  1.4253E+04, 1.4261E+04, 1.4269E+04, 1.4277E+04, 1.4285E+04,
     + 1.4293E+04, 1.4302E+04, 1.4310E+04, 1.4318E+04, 1.4326E+04,
     +  1.4334E+04, 1.4342E+04, 1.4350E+04, 1.4358E+04, 1.4366E+04,
     + 1.4374E+04, 1.4382E+04, 1.4390E+04, 1.4398E+04, 1.4406E+04,
     +  1.4414E+04, 1.4422E+04, 1.4431E+04, 1.4439E+04, 1.4447E+04,
     + 1.4455E+04, 1.4463E+04, 1.4471E+04, 1.4479E+04, 1.4487E+04,
     +  1.4495E+04, 1.4503E+04, 1.4511E+04, 1.4520E+04, 1.4528E+04,
     + 1.4536E+04, 1.4544E+04, 1.4552E+04, 1.4560E+04, 1.4568E+04,
     +  1.4576E+04, 1.4584E+04, 1.4593E+04, 1.4601E+04, 1.4609E+04,
     + 1.4617E+04, 1.4625E+04, 1.4633E+04, 1.4641E+04, 1.4650E+04,
     +  1.4658E+04, 1.4666E+04, 1.4674E+04, 1.4682E+04, 1.4690E+04,
     + 1.4698E+04, 1.4707E+04, 1.4715E+04, 1.4723E+04, 1.4731E+04,
     +  1.4739E+04, 1.4747E+04, 1.4756E+04, 1.4764E+04, 1.4772E+04,
     + 1.4780E+04, 1.4788E+04, 1.4797E+04, 1.4805E+04, 1.4813E+04,
     +  1.4821E+04, 1.4829E+04, 1.4838E+04, 1.4846E+04, 1.4854E+04,
     + 1.4862E+04, 1.4870E+04, 1.4879E+04, 1.4887E+04, 1.4895E+04,
     +  1.4903E+04, 1.4911E+04, 1.4920E+04, 1.4928E+04, 1.4936E+04,
     + 1.4944E+04, 1.4953E+04, 1.4961E+04, 1.4969E+04, 1.4977E+04,
     +  1.4986E+04, 1.4994E+04, 1.5002E+04, 1.5010E+04, 1.5019E+04,
     + 1.5027E+04, 1.5035E+04, 1.5043E+04, 1.5052E+04, 1.5060E+04,
     +  1.5068E+04, 1.5076E+04, 1.5085E+04, 1.5093E+04, 1.5101E+04,
     + 1.5110E+04, 1.5118E+04, 1.5126E+04, 1.5134E+04, 1.5143E+04,
     +  1.5151E+04, 1.5159E+04, 1.5168E+04, 1.5176E+04, 1.5184E+04,
     + 1.5192E+04, 1.5201E+04, 1.5209E+04, 1.5217E+04, 1.5226E+04,
     +  1.5234E+04, 1.5242E+04, 1.5251E+04, 1.5259E+04, 1.5267E+04,
     + 1.5276E+04, 1.5284E+04, 1.5292E+04, 1.5301E+04, 1.5309E+04,
     +  1.5317E+04, 1.5326E+04, 1.5334E+04, 1.5342E+04, 1.5351E+04,
     + 1.5359E+04, 1.5367E+04, 1.5376E+04, 1.5384E+04, 1.5393E+04,
     +  1.5401E+04, 1.5409E+04, 1.5418E+04, 1.5426E+04, 1.5434E+04,
     + 1.5443E+04, 1.5451E+04, 1.5460E+04, 1.5468E+04, 1.5476E+04,
     +  1.5485E+04, 1.5493E+04, 1.5501E+04, 1.5510E+04, 1.5518E+04,
     + 1.5527E+04, 1.5535E+04, 1.5543E+04, 1.5552E+04, 1.5560E+04,
     +  1.5569E+04, 1.5577E+04, 1.5586E+04, 1.5594E+04, 1.5602E+04,
     + 1.5611E+04, 1.5619E+04, 1.5628E+04, 1.5636E+04, 1.5645E+04,
     +  1.5653E+04, 1.5661E+04, 1.5670E+04, 1.5678E+04, 1.5687E+04,
     + 1.5695E+04, 1.5704E+04, 1.5712E+04, 1.5721E+04, 1.5729E+04,
     +  1.5738E+04, 1.5746E+04, 1.5754E+04, 1.5763E+04, 1.5771E+04,
     + 1.5780E+04, 1.5788E+04, 1.5797E+04, 1.5805E+04, 1.5814E+04,
     +  1.5822E+04, 1.5831E+04, 1.5839E+04, 1.5848E+04, 1.5856E+04,
     + 1.5865E+04, 1.5873E+04, 1.5882E+04, 1.5890E+04, 1.5899E+04,
     +  1.5907E+04, 1.5916E+04, 1.5924E+04, 1.5933E+04, 1.5941E+04,
     + 1.5950E+04, 1.5958E+04, 1.5967E+04, 1.5976E+04, 1.5984E+04,
     +  1.5993E+04, 1.6001E+04, 1.6010E+04, 1.6018E+04, 1.6027E+04,
     + 1.6035E+04, 1.6044E+04, 1.6052E+04, 1.6061E+04, 1.6070E+04,
     +  1.6078E+04, 1.6087E+04, 1.6095E+04, 1.6104E+04, 1.6112E+04,
     + 1.6121E+04, 1.6130E+04, 1.6138E+04, 1.6147E+04, 1.6155E+04,
     +  1.6164E+04, 1.6172E+04, 1.6181E+04, 1.6190E+04, 1.6198E+04,
     + 1.6207E+04, 1.6215E+04, 1.6224E+04, 1.6233E+04, 1.6241E+04,
     +  1.6250E+04, 1.6258E+04, 1.6267E+04, 1.6276E+04, 1.6284E+04,
     + 1.6293E+04, 1.6302E+04, 1.6310E+04, 1.6319E+04, 1.6327E+04,
     +  1.6336E+04, 1.6345E+04, 1.6353E+04, 1.6362E+04, 1.6371E+04,
     + 1.6379E+04, 1.6388E+04, 1.6397E+04, 1.6405E+04, 1.6414E+04,
     +  1.6423E+04, 1.6431E+04, 1.6440E+04, 1.6449E+04, 1.6457E+04,
     + 1.6466E+04, 1.6475E+04, 1.6483E+04, 1.6492E+04, 1.6501E+04,
     +  1.6509E+04, 1.6518E+04, 1.6527E+04, 1.6535E+04, 1.6544E+04,
     + 1.6553E+04, 1.6561E+04, 1.6570E+04, 1.6579E+04, 1.6588E+04,
     +  1.6596E+04, 1.6605E+04, 1.6614E+04, 1.6622E+04, 1.6631E+04,
     + 1.6640E+04, 1.6649E+04, 1.6657E+04, 1.6666E+04, 1.6675E+04,
     +  1.6683E+04, 1.6692E+04, 1.6701E+04, 1.6710E+04, 1.6718E+04,
     + 1.6727E+04, 1.6736E+04, 1.6745E+04, 1.6753E+04, 1.6762E+04,
     +  1.6771E+04, 1.6780E+04, 1.6788E+04, 1.6797E+04, 1.6806E+04,
     + 1.6815E+04, 1.6824E+04, 1.6832E+04, 1.6841E+04, 1.6850E+04,
     +  1.6859E+04, 1.6867E+04, 1.6876E+04, 1.6885E+04, 1.6894E+04,
     + 1.6903E+04, 1.6911E+04, 1.6920E+04, 1.6929E+04, 1.6938E+04,
     +  1.6947E+04, 1.6955E+04, 1.6964E+04, 1.6973E+04, 1.6982E+04,
     + 1.6991E+04, 1.7000E+04, 1.7008E+04, 1.7017E+04, 1.7026E+04,
     +  1.7035E+04, 1.7044E+04, 1.7052E+04, 1.7061E+04, 1.7070E+04,
     + 1.7079E+04, 1.7088E+04, 1.7097E+04, 1.7106E+04, 1.7114E+04,
     +  1.7123E+04, 1.7132E+04, 1.7141E+04, 1.7150E+04, 1.7159E+04,
     + 1.7168E+04, 1.7176E+04, 1.7185E+04, 1.7194E+04, 1.7203E+04,
     +  1.7212E+04, 1.7221E+04, 1.7230E+04, 1.7239E+04, 1.7248E+04,
     + 1.7256E+04, 1.7265E+04, 1.7274E+04, 1.7283E+04, 1.7292E+04,
     +  1.7301E+04, 1.7310E+04, 1.7319E+04, 1.7328E+04, 1.7337E+04,
     + 1.7345E+04, 1.7354E+04, 1.7363E+04, 1.7372E+04, 1.7381E+04,
     +  1.7390E+04, 1.7399E+04, 1.7408E+04, 1.7417E+04, 1.7426E+04,
     + 1.7435E+04, 1.7444E+04, 1.7453E+04, 1.7462E+04, 1.7471E+04,
     +  1.7480E+04, 1.7488E+04, 1.7497E+04, 1.7506E+04, 1.7515E+04,
     + 1.7524E+04, 1.7533E+04, 1.7542E+04, 1.7551E+04, 1.7560E+04,
     +  1.7569E+04, 1.7578E+04, 1.7587E+04, 1.7596E+04, 1.7605E+04,
     + 1.7614E+04, 1.7623E+04, 1.7632E+04, 1.7641E+04, 1.7650E+04,
     +  1.7659E+04, 1.7668E+04, 1.7677E+04, 1.7686E+04, 1.7695E+04,
     + 1.7704E+04, 1.7713E+04, 1.7722E+04, 1.7731E+04, 1.7740E+04,
     +  1.7749E+04, 1.7758E+04, 1.7767E+04, 1.7776E+04, 1.7785E+04,
     + 1.7794E+04, 1.7804E+04, 1.7813E+04, 1.7822E+04, 1.7831E+04,
     +  1.7840E+04, 1.7849E+04, 1.7858E+04, 1.7867E+04, 1.7876E+04,
     + 1.7885E+04, 1.7894E+04, 1.7903E+04, 1.7912E+04, 1.7921E+04,
     +  1.7930E+04, 1.7939E+04, 1.7949E+04, 1.7958E+04, 1.7967E+04,
     + 1.7976E+04, 1.7985E+04, 1.7994E+04, 1.8003E+04, 1.8012E+04,
     +  1.8021E+04, 1.8030E+04, 1.8039E+04, 1.8049E+04, 1.8058E+04,
     + 1.8067E+04, 1.8076E+04, 1.8085E+04, 1.8094E+04, 1.8103E+04,
     +  1.8112E+04, 1.8122E+04, 1.8131E+04, 1.8140E+04, 1.8149E+04,
     + 1.8158E+04, 1.8167E+04, 1.8176E+04, 1.8186E+04, 1.8195E+04,
     +  1.8204E+04, 1.8213E+04, 1.8222E+04, 1.8231E+04, 1.8240E+04,
     + 1.8250E+04, 1.8259E+04, 1.8268E+04, 1.8277E+04, 1.8286E+04,
     +  1.8295E+04, 1.8305E+04, 1.8314E+04, 1.8323E+04, 1.8332E+04,
     + 1.8341E+04, 1.8351E+04, 1.8360E+04, 1.8369E+04, 1.8378E+04,
     +  1.8387E+04, 1.8397E+04, 1.8406E+04, 1.8415E+04, 1.8424E+04,
     + 1.8433E+04, 1.8443E+04, 1.8452E+04, 1.8461E+04, 1.8470E+04,
     +  1.8480E+04, 1.8489E+04, 1.8498E+04, 1.8507E+04, 1.8516E+04,
     + 1.8526E+04, 1.8535E+04, 1.8544E+04, 1.8553E+04, 1.8563E+04,
     +  1.8572E+04, 1.8581E+04, 1.8590E+04, 1.8600E+04, 1.8609E+04,
     + 1.8618E+04, 1.8627E+04, 1.8637E+04, 1.8646E+04, 1.8655E+04,
     +  1.8665E+04, 1.8674E+04, 1.8683E+04, 1.8692E+04, 1.8702E+04,
     + 1.8711E+04, 1.8720E+04, 1.8730E+04, 1.8739E+04, 1.8748E+04,
     +  1.8757E+04, 1.8767E+04, 1.8776E+04, 1.8785E+04, 1.8795E+04,
     + 1.8804E+04, 1.8813E+04, 1.8823E+04, 1.8832E+04, 1.8841E+04,
     +  1.8851E+04, 1.8860E+04, 1.8869E+04, 1.8879E+04, 1.8888E+04,
     + 1.8897E+04, 1.8907E+04, 1.8916E+04, 1.8925E+04, 1.8935E+04,
     +  1.8944E+04, 1.8953E+04, 1.8963E+04, 1.8972E+04, 1.8981E+04,
     + 1.8991E+04, 1.9000E+04, 1.9009E+04, 1.9019E+04, 1.9028E+04,
     +  1.9038E+04, 1.9047E+04, 1.9056E+04, 1.9066E+04, 1.9075E+04,
     + 1.9084E+04, 1.9094E+04, 1.9103E+04, 1.9113E+04, 1.9122E+04,
     +  1.9131E+04, 1.9141E+04, 1.9150E+04, 1.9160E+04, 1.9169E+04,
     + 1.9178E+04, 1.9188E+04, 1.9197E+04, 1.9207E+04, 1.9216E+04,
     +  1.9226E+04, 1.9235E+04, 1.9244E+04, 1.9254E+04, 1.9263E+04,
     + 1.9273E+04, 1.9282E+04, 1.9292E+04, 1.9301E+04, 1.9310E+04,
     +  1.9320E+04, 1.9329E+04, 1.9339E+04, 1.9348E+04, 1.9358E+04,
     + 1.9367E+04, 1.9377E+04, 1.9386E+04, 1.9396E+04, 1.9405E+04,
     +  1.9415E+04, 1.9424E+04, 1.9433E+04, 1.9443E+04, 1.9452E+04,
     + 1.9462E+04, 1.9471E+04, 1.9481E+04, 1.9490E+04, 1.9500E+04,
     +  1.9509E+04, 1.9519E+04, 1.9528E+04, 1.9538E+04, 1.9547E+04,
     + 1.9557E+04, 1.9566E+04, 1.9576E+04, 1.9585E+04, 1.9595E+04,
     +  1.9604E+04, 1.9614E+04, 1.9624E+04, 1.9633E+04, 1.9643E+04,
     + 1.9652E+04, 1.9662E+04, 1.9671E+04, 1.9681E+04, 1.9690E+04,
     +  1.9700E+04, 1.9709E+04, 1.9719E+04, 1.9728E+04, 1.9738E+04,
     + 1.9748E+04, 1.9757E+04, 1.9767E+04, 1.9776E+04, 1.9786E+04,
     +  1.9795E+04, 1.9805E+04, 1.9815E+04, 1.9824E+04, 1.9834E+04,
     + 1.9843E+04, 1.9853E+04, 1.9863E+04, 1.9872E+04, 1.9882E+04,
     +  1.9891E+04, 1.9901E+04, 1.9910E+04, 1.9920E+04, 1.9930E+04,
     + 1.9939E+04, 1.9949E+04, 1.9959E+04, 1.9968E+04, 1.9978E+04,
     +  1.9987E+04, 1.9997E+04, 2.0007E+04, 2.0016E+04, 2.0026E+04,
     + 2.0036E+04, 2.0045E+04, 2.0055E+04, 2.0064E+04, 2.0074E+04,
     +  2.0084E+04, 2.0093E+04, 2.0103E+04, 2.0113E+04, 2.0122E+04,
     + 2.0132E+04, 2.0142E+04, 2.0151E+04, 2.0161E+04, 2.0171E+04,
     +  2.0180E+04, 2.0190E+04, 2.0200E+04, 2.0209E+04, 2.0219E+04,
     + 2.0229E+04, 2.0238E+04, 2.0248E+04, 2.0258E+04, 2.0267E+04,
     +  2.0277E+04, 2.0287E+04, 2.0297E+04, 2.0306E+04, 2.0316E+04,
     + 2.0326E+04, 2.0335E+04, 2.0345E+04, 2.0355E+04, 2.0365E+04,
     +  2.0374E+04, 2.0384E+04, 2.0394E+04, 2.0403E+04, 2.0413E+04,
     + 2.0423E+04, 2.0433E+04, 2.0442E+04, 2.0452E+04, 2.0462E+04,
     +  2.0472E+04, 2.0481E+04, 2.0491E+04, 2.0501E+04, 2.0511E+04,
     + 2.0520E+04, 2.0530E+04, 2.0540E+04, 2.0550E+04, 2.0559E+04,
     +  2.0569E+04, 2.0579E+04, 2.0589E+04, 2.0599E+04, 2.0608E+04,
     + 2.0618E+04, 2.0628E+04, 2.0638E+04, 2.0647E+04, 2.0657E+04,
     +  2.0667E+04, 2.0677E+04, 2.0687E+04, 2.0696E+04, 2.0706E+04,
     + 2.0716E+04, 2.0726E+04, 2.0736E+04, 2.0745E+04, 2.0755E+04,
     +  2.0765E+04, 2.0775E+04, 2.0785E+04, 2.0795E+04, 2.0804E+04,
     + 2.0814E+04, 2.0824E+04, 2.0834E+04, 2.0844E+04, 2.0854E+04,
     +  2.0863E+04, 2.0873E+04, 2.0883E+04, 2.0893E+04, 2.0903E+04,
     + 2.0913E+04, 2.0923E+04, 2.0932E+04, 2.0942E+04, 2.0952E+04,
     +  2.0962E+04, 2.0972E+04, 2.0982E+04, 2.0992E+04, 2.1002E+04,
     + 2.1011E+04, 2.1021E+04, 2.1031E+04, 2.1041E+04, 2.1051E+04,
     +  2.1061E+04, 2.1071E+04, 2.1081E+04, 2.1091E+04, 2.1100E+04,
     + 2.1110E+04, 2.1120E+04, 2.1130E+04, 2.1140E+04, 2.1150E+04,
     +  2.1160E+04, 2.1170E+04, 2.1180E+04, 2.1190E+04, 2.1200E+04,
     + 2.1210E+04, 2.1219E+04, 2.1229E+04, 2.1239E+04, 2.1249E+04,
     +  2.1259E+04, 2.1269E+04, 2.1279E+04, 2.1289E+04, 2.1299E+04,
     + 2.1309E+04, 2.1319E+04, 2.1329E+04, 2.1339E+04, 2.1349E+04,
     +  2.1359E+04, 2.1369E+04, 2.1379E+04, 2.1389E+04, 2.1399E+04,
     + 2.1409E+04, 2.1419E+04, 2.1429E+04, 2.1439E+04, 2.1449E+04,
     +  2.1459E+04, 2.1469E+04, 2.1479E+04, 2.1489E+04, 2.1499E+04,
     + 2.1509E+04, 2.1519E+04, 2.1529E+04, 2.1539E+04, 2.1549E+04,
     +  2.1559E+04, 2.1569E+04, 2.1579E+04, 2.1589E+04, 2.1599E+04,
     + 2.1609E+04, 2.1619E+04, 2.1629E+04, 2.1639E+04, 2.1649E+04,
     +  2.1659E+04, 2.1669E+04, 2.1679E+04, 2.1689E+04, 2.1699E+04,
     + 2.1709E+04, 2.1719E+04, 2.1729E+04, 2.1739E+04, 2.1750E+04,
     +  2.1760E+04, 2.1770E+04, 2.1780E+04, 2.1790E+04, 2.1800E+04,
     + 2.1810E+04, 2.1820E+04, 2.1830E+04, 2.1840E+04, 2.1850E+04,
     +  2.1860E+04, 2.1871E+04, 2.1881E+04, 2.1891E+04, 2.1901E+04,
     + 2.1911E+04, 2.1921E+04, 2.1931E+04, 2.1941E+04, 2.1951E+04,
     +  2.1961E+04, 2.1972E+04, 2.1982E+04, 2.1992E+04, 2.2002E+04,
     + 2.2012E+04, 2.2022E+04, 2.2032E+04, 2.2043E+04, 2.2053E+04,
     +  2.2063E+04, 2.2073E+04, 2.2083E+04, 2.2093E+04, 2.2103E+04,
     + 2.2114E+04, 2.2124E+04, 2.2134E+04, 2.2144E+04, 2.2154E+04,
     +  2.2164E+04, 2.2175E+04, 2.2185E+04, 2.2195E+04, 2.2205E+04,
     + 2.2215E+04, 2.2225E+04, 2.2236E+04, 2.2246E+04, 2.2256E+04,
     +  2.2266E+04, 2.2276E+04, 2.2287E+04, 2.2297E+04, 2.2307E+04,
     + 2.2317E+04, 2.2327E+04, 2.2338E+04, 2.2348E+04, 2.2358E+04,
     +  2.2368E+04, 2.2378E+04, 2.2389E+04, 2.2399E+04, 2.2409E+04,
     + 2.2419E+04, 2.2430E+04, 2.2440E+04, 2.2450E+04, 2.2460E+04,
     +  2.2471E+04, 2.2481E+04, 2.2491E+04, 2.2501E+04, 2.2512E+04,
     + 2.2522E+04, 2.2532E+04, 2.2542E+04, 2.2553E+04, 2.2563E+04,
     +  2.2573E+04, 2.2583E+04, 2.2594E+04, 2.2604E+04, 2.2614E+04,
     + 2.2624E+04, 2.2635E+04, 2.2645E+04, 2.2655E+04, 2.2666E+04,
     +  2.2676E+04, 2.2686E+04, 2.2696E+04, 2.2707E+04, 2.2717E+04,
     + 2.2727E+04, 2.2738E+04, 2.2748E+04, 2.2758E+04, 2.2769E+04,
     +  2.2779E+04, 2.2789E+04, 2.2800E+04, 2.2810E+04, 2.2820E+04,
     + 2.2831E+04, 2.2841E+04, 2.2851E+04, 2.2862E+04, 2.2872E+04,
     +  2.2882E+04, 2.2893E+04, 2.2903E+04, 2.2913E+04, 2.2924E+04,
     + 2.2934E+04, 2.2944E+04, 2.2955E+04, 2.2965E+04, 2.2975E+04,
     +  2.2986E+04, 2.2996E+04, 2.3007E+04, 2.3017E+04, 2.3027E+04,
     + 2.3038E+04, 2.3048E+04, 2.3058E+04, 2.3069E+04, 2.3079E+04,
     +  2.3090E+04, 2.3100E+04, 2.3110E+04, 2.3121E+04, 2.3131E+04,
     + 2.3142E+04, 2.3152E+04, 2.3162E+04, 2.3173E+04, 2.3183E+04,
     +  2.3194E+04, 2.3204E+04, 2.3214E+04, 2.3225E+04, 2.3235E+04,
     + 2.3246E+04, 2.3256E+04, 2.3267E+04, 2.3277E+04, 2.3288E+04,
     +  2.3298E+04, 2.3308E+04, 2.3319E+04, 2.3329E+04, 2.3340E+04,
     + 2.3350E+04, 2.3361E+04, 2.3371E+04, 2.3382E+04, 2.3392E+04,
     +  2.3403E+04, 2.3413E+04, 2.3423E+04, 2.3434E+04, 2.3444E+04,
     + 2.3455E+04, 2.3465E+04, 2.3476E+04, 2.3486E+04, 2.3497E+04,
     +  2.3507E+04, 2.3518E+04, 2.3528E+04, 2.3539E+04, 2.3549E+04,
     + 2.3560E+04, 2.3570E+04, 2.3581E+04, 2.3591E+04, 2.3602E+04,
     +  2.3612E+04, 2.3623E+04, 2.3633E+04, 2.3644E+04, 2.3654E+04,
     + 2.3665E+04, 2.3676E+04, 2.3686E+04, 2.3697E+04, 2.3707E+04,
     +  2.3718E+04, 2.3728E+04, 2.3739E+04, 2.3749E+04, 2.3760E+04,
     + 2.3770E+04, 2.3781E+04, 2.3792E+04, 2.3802E+04, 2.3813E+04,
     +  2.3823E+04, 2.3834E+04, 2.3844E+04, 2.3855E+04, 2.3865E+04,
     + 2.3876E+04, 2.3887E+04, 2.3897E+04, 2.3908E+04, 2.3918E+04,
     +  2.3929E+04, 2.3940E+04, 2.3950E+04, 2.3961E+04, 2.3971E+04,
     + 2.3982E+04, 2.3993E+04, 2.4003E+04, 2.4014E+04, 2.4024E+04,
     +  2.4035E+04, 2.4046E+04, 2.4056E+04, 2.4067E+04, 2.4078E+04,
     + 2.4088E+04, 2.4099E+04, 2.4109E+04, 2.4120E+04, 2.4131E+04,
     +  2.4141E+04, 2.4152E+04, 2.4163E+04, 2.4173E+04, 2.4184E+04,
     + 2.4195E+04, 2.4205E+04, 2.4216E+04, 2.4227E+04, 2.4237E+04,
     +  2.4248E+04, 2.4259E+04, 2.4269E+04, 2.4280E+04, 2.4291E+04,
     + 2.4301E+04, 2.4312E+04, 2.4323E+04, 2.4333E+04, 2.4344E+04,
     +  2.4355E+04, 2.4365E+04, 2.4376E+04, 2.4387E+04, 2.4397E+04,
     + 2.4408E+04, 2.4419E+04, 2.4430E+04, 2.4440E+04, 2.4451E+04,
     +  2.4462E+04, 2.4472E+04, 2.4483E+04, 2.4494E+04, 2.4505E+04,
     + 2.4515E+04, 2.4526E+04, 2.4537E+04, 2.4548E+04, 2.4558E+04,
     +  2.4569E+04, 2.4580E+04, 2.4590E+04, 2.4601E+04, 2.4612E+04,
     + 2.4623E+04, 2.4633E+04, 2.4644E+04, 2.4655E+04, 2.4666E+04,
     +  2.4677E+04, 2.4687E+04, 2.4698E+04, 2.4709E+04, 2.4720E+04,
     + 2.4730E+04, 2.4741E+04, 2.4752E+04, 2.4763E+04, 2.4774E+04,
     +  2.4784E+04, 2.4795E+04, 2.4806E+04, 2.4817E+04, 2.4828E+04,
     + 2.4838E+04, 2.4849E+04, 2.4860E+04, 2.4871E+04, 2.4882E+04,
     +  2.4892E+04, 2.4903E+04, 2.4914E+04, 2.4925E+04, 2.4936E+04,
     + 2.4946E+04, 2.4957E+04, 2.4968E+04, 2.4979E+04, 2.4990E+04,
     +  2.5001E+04, 2.5012E+04, 2.5022E+04, 2.5033E+04, 2.5044E+04,
     + 2.5055E+04, 2.5066E+04, 2.5077E+04, 2.5087E+04, 2.5098E+04,
     +  2.5109E+04, 2.5120E+04, 2.5131E+04, 2.5142E+04, 2.5153E+04,
     + 2.5164E+04, 2.5174E+04, 2.5185E+04, 2.5196E+04, 2.5207E+04,
     +  2.5218E+04, 2.5229E+04, 2.5240E+04, 2.5251E+04, 2.5262E+04,
     + 2.5272E+04, 2.5283E+04, 2.5294E+04, 2.5305E+04, 2.5316E+04,
     +  2.5327E+04, 2.5338E+04, 2.5349E+04, 2.5360E+04, 2.5371E+04,
     + 2.5382E+04, 2.5393E+04, 2.5403E+04, 2.5414E+04, 2.5425E+04,
     +  2.5436E+04, 2.5447E+04, 2.5458E+04, 2.5469E+04, 2.5480E+04,
     + 2.5491E+04, 2.5502E+04, 2.5513E+04, 2.5524E+04, 2.5535E+04,
     +  2.5546E+04, 2.5557E+04, 2.5568E+04, 2.5579E+04, 2.5590E+04,
     + 2.5601E+04, 2.5612E+04, 2.5623E+04, 2.5634E+04, 2.5645E+04,
     +  2.5656E+04, 2.5667E+04, 2.5678E+04, 2.5689E+04, 2.5700E+04,
     + 2.5711E+04, 2.5722E+04, 2.5733E+04, 2.5744E+04/
c...        --        28
      data (QofT( 3,J),J=1,2408)/ 1.0155E+00, 1.2171E+00, 1.5440E+00,
     +  1.9024E+00, 2.2706E+00, 2.6430E+00, 3.0176E+00, 3.3936E+00,
     + 3.7705E+00, 4.1479E+00, 4.5257E+00, 4.9039E+00, 5.2823E+00,
     +  5.6609E+00, 6.0396E+00, 6.4184E+00, 6.7973E+00, 7.1764E+00,
     + 7.5554E+00, 7.9346E+00, 8.6930E+00, 9.4516E+00, 1.0210E+01,
     +  1.0969E+01, 1.1728E+01, 1.2487E+01, 1.3246E+01, 1.4005E+01,
     + 1.4764E+01, 1.5523E+01, 1.6282E+01, 1.7041E+01, 1.7800E+01,
     +  1.8560E+01, 1.9319E+01, 2.0078E+01, 2.0837E+01, 2.1597E+01,
     + 2.2356E+01, 2.3115E+01, 2.3874E+01, 2.4634E+01, 2.5393E+01,
     +  2.6152E+01, 2.6912E+01, 2.7671E+01, 2.8431E+01, 2.9190E+01,
     + 2.9949E+01, 3.0709E+01, 3.1468E+01, 3.2228E+01, 3.2987E+01,
     +  3.3746E+01, 3.4506E+01, 3.5265E+01, 3.6025E+01, 3.6784E+01,
     + 3.7544E+01, 3.8303E+01, 3.9063E+01, 3.9822E+01, 4.0582E+01,
     +  4.1341E+01, 4.2101E+01, 4.2860E+01, 4.3620E+01, 4.4380E+01,
     + 4.5139E+01, 4.5899E+01, 4.6658E+01, 4.7418E+01, 4.8178E+01,
     +  4.8937E+01, 4.9697E+01, 5.0456E+01, 5.1216E+01, 5.1976E+01,
     + 5.2735E+01, 5.3495E+01, 5.4255E+01, 5.5014E+01, 5.5774E+01,
     +  5.6534E+01, 5.7293E+01, 5.8053E+01, 5.8813E+01, 5.9573E+01,
     + 6.0332E+01, 6.1092E+01, 6.1852E+01, 6.2611E+01, 6.3371E+01,
     +  6.4131E+01, 6.4891E+01, 6.5651E+01, 6.6410E+01, 6.7170E+01,
     + 6.7930E+01, 6.8690E+01, 6.9450E+01, 7.0209E+01, 7.0969E+01,
     +  7.1729E+01, 7.2489E+01, 7.3249E+01, 7.4009E+01, 7.4768E+01,
     + 7.5528E+01, 7.6288E+01, 7.7048E+01, 7.7808E+01, 7.8568E+01,
     +  7.9328E+01, 8.0088E+01, 8.0848E+01, 8.1608E+01, 8.2367E+01,
     + 8.3127E+01, 8.3887E+01, 8.4647E+01, 8.5407E+01, 8.6167E+01,
     +  8.6927E+01, 8.7687E+01, 8.8447E+01, 8.9207E+01, 8.9967E+01,
     + 9.0727E+01, 9.1488E+01, 9.2248E+01, 9.3008E+01, 9.3768E+01,
     +  9.4528E+01, 9.5288E+01, 9.6048E+01, 9.6808E+01, 9.7568E+01,
     + 9.8329E+01, 9.9089E+01, 9.9849E+01, 1.0061E+02, 1.0137E+02,
     +  1.0213E+02, 1.0289E+02, 1.0365E+02, 1.0441E+02, 1.0517E+02,
     + 1.0593E+02, 1.0669E+02, 1.0745E+02, 1.0821E+02, 1.0897E+02,
     +  1.0973E+02, 1.1049E+02, 1.1125E+02, 1.1201E+02, 1.1277E+02,
     + 1.1354E+02, 1.1430E+02, 1.1506E+02, 1.1582E+02, 1.1658E+02,
     +  1.1734E+02, 1.1810E+02, 1.1886E+02, 1.1962E+02, 1.2038E+02,
     + 1.2114E+02, 1.2190E+02, 1.2266E+02, 1.2343E+02, 1.2419E+02,
     +  1.2495E+02, 1.2571E+02, 1.2647E+02, 1.2723E+02, 1.2799E+02,
     + 1.2875E+02, 1.2952E+02, 1.3028E+02, 1.3104E+02, 1.3180E+02,
     +  1.3256E+02, 1.3332E+02, 1.3409E+02, 1.3485E+02, 1.3561E+02,
     + 1.3637E+02, 1.3713E+02, 1.3790E+02, 1.3866E+02, 1.3942E+02,
     +  1.4018E+02, 1.4095E+02, 1.4171E+02, 1.4247E+02, 1.4323E+02,
     + 1.4400E+02, 1.4476E+02, 1.4552E+02, 1.4629E+02, 1.4705E+02,
     +  1.4781E+02, 1.4858E+02, 1.4934E+02, 1.5010E+02, 1.5087E+02,
     + 1.5163E+02, 1.5239E+02, 1.5316E+02, 1.5392E+02, 1.5469E+02,
     +  1.5545E+02, 1.5622E+02, 1.5698E+02, 1.5775E+02, 1.5851E+02,
     + 1.5928E+02, 1.6004E+02, 1.6081E+02, 1.6157E+02, 1.6234E+02,
     +  1.6311E+02, 1.6387E+02, 1.6464E+02, 1.6541E+02, 1.6617E+02,
     + 1.6694E+02, 1.6771E+02, 1.6847E+02, 1.6924E+02, 1.7001E+02,
     +  1.7078E+02, 1.7154E+02, 1.7231E+02, 1.7308E+02, 1.7385E+02,
     + 1.7462E+02, 1.7539E+02, 1.7616E+02, 1.7693E+02, 1.7770E+02,
     +  1.7847E+02, 1.7924E+02, 1.8001E+02, 1.8078E+02, 1.8155E+02,
     + 1.8232E+02, 1.8309E+02, 1.8386E+02, 1.8463E+02, 1.8541E+02,
     +  1.8618E+02, 1.8695E+02, 1.8772E+02, 1.8850E+02, 1.8927E+02,
     + 1.9004E+02, 1.9082E+02, 1.9159E+02, 1.9237E+02, 1.9314E+02,
     +  1.9392E+02, 1.9469E+02, 1.9547E+02, 1.9624E+02, 1.9702E+02,
     + 1.9780E+02, 1.9857E+02, 1.9935E+02, 2.0013E+02, 2.0091E+02,
     +  2.0168E+02, 2.0246E+02, 2.0324E+02, 2.0402E+02, 2.0480E+02,
     + 2.0558E+02, 2.0636E+02, 2.0714E+02, 2.0792E+02, 2.0870E+02,
     +  2.0948E+02, 2.1027E+02, 2.1105E+02, 2.1183E+02, 2.1262E+02,
     + 2.1340E+02, 2.1418E+02, 2.1497E+02, 2.1575E+02, 2.1654E+02,
     +  2.1732E+02, 2.1811E+02, 2.1889E+02, 2.1968E+02, 2.2047E+02,
     + 2.2125E+02, 2.2204E+02, 2.2283E+02, 2.2362E+02, 2.2441E+02,
     +  2.2520E+02, 2.2599E+02, 2.2678E+02, 2.2757E+02, 2.2836E+02,
     + 2.2915E+02, 2.2994E+02, 2.3074E+02, 2.3153E+02, 2.3232E+02,
     +  2.3312E+02, 2.3391E+02, 2.3471E+02, 2.3550E+02, 2.3630E+02,
     + 2.3709E+02, 2.3789E+02, 2.3869E+02, 2.3948E+02, 2.4028E+02,
     +  2.4108E+02, 2.4188E+02, 2.4268E+02, 2.4348E+02, 2.4428E+02,
     + 2.4508E+02, 2.4588E+02, 2.4669E+02, 2.4749E+02, 2.4829E+02,
     +  2.4910E+02, 2.4990E+02, 2.5070E+02, 2.5151E+02, 2.5232E+02,
     + 2.5312E+02, 2.5393E+02, 2.5474E+02, 2.5554E+02, 2.5635E+02,
     +  2.5716E+02, 2.5797E+02, 2.5878E+02, 2.5959E+02, 2.6040E+02,
     + 2.6121E+02, 2.6203E+02, 2.6284E+02, 2.6365E+02, 2.6447E+02,
     +  2.6528E+02, 2.6610E+02, 2.6691E+02, 2.6773E+02, 2.6855E+02,
     + 2.6936E+02, 2.7018E+02, 2.7100E+02, 2.7182E+02, 2.7264E+02,
     +  2.7346E+02, 2.7428E+02, 2.7510E+02, 2.7592E+02, 2.7675E+02,
     + 2.7757E+02, 2.7839E+02, 2.7922E+02, 2.8004E+02, 2.8087E+02,
     +  2.8170E+02, 2.8252E+02, 2.8335E+02, 2.8418E+02, 2.8501E+02,
     + 2.8584E+02, 2.8667E+02, 2.8750E+02, 2.8833E+02, 2.8916E+02,
     +  2.9000E+02, 2.9083E+02, 2.9166E+02, 2.9250E+02, 2.9334E+02,
     + 2.9417E+02, 2.9501E+02, 2.9585E+02, 2.9668E+02, 2.9752E+02,
     +  2.9836E+02, 2.9920E+02, 3.0004E+02, 3.0088E+02, 3.0173E+02,
     + 3.0257E+02, 3.0341E+02, 3.0426E+02, 3.0510E+02, 3.0595E+02,
     +  3.0679E+02, 3.0764E+02, 3.0849E+02, 3.0934E+02, 3.1019E+02,
     + 3.1104E+02, 3.1189E+02, 3.1274E+02, 3.1359E+02, 3.1444E+02,
     +  3.1530E+02, 3.1615E+02, 3.1700E+02, 3.1786E+02, 3.1872E+02,
     + 3.1957E+02, 3.2043E+02, 3.2129E+02, 3.2215E+02, 3.2301E+02,
     +  3.2387E+02, 3.2473E+02, 3.2559E+02, 3.2645E+02, 3.2732E+02,
     + 3.2818E+02, 3.2905E+02, 3.2991E+02, 3.3078E+02, 3.3165E+02,
     +  3.3251E+02, 3.3338E+02, 3.3425E+02, 3.3512E+02, 3.3599E+02,
     + 3.3687E+02, 3.3774E+02, 3.3861E+02, 3.3948E+02, 3.4036E+02,
     +  3.4123E+02, 3.4211E+02, 3.4299E+02, 3.4387E+02, 3.4474E+02,
     + 3.4562E+02, 3.4650E+02, 3.4738E+02, 3.4827E+02, 3.4915E+02,
     +  3.5003E+02, 3.5092E+02, 3.5180E+02, 3.5269E+02, 3.5357E+02,
     + 3.5446E+02, 3.5535E+02, 3.5624E+02, 3.5713E+02, 3.5802E+02,
     +  3.5891E+02, 3.5980E+02, 3.6069E+02, 3.6158E+02, 3.6248E+02,
     + 3.6337E+02, 3.6427E+02, 3.6517E+02, 3.6606E+02, 3.6696E+02,
     +  3.6786E+02, 3.6876E+02, 3.6966E+02, 3.7056E+02, 3.7147E+02,
     + 3.7237E+02, 3.7327E+02, 3.7418E+02, 3.7508E+02, 3.7599E+02,
     +  3.7690E+02, 3.7780E+02, 3.7871E+02, 3.7962E+02, 3.8053E+02,
     + 3.8144E+02, 3.8236E+02, 3.8327E+02, 3.8418E+02, 3.8510E+02,
     +  3.8601E+02, 3.8693E+02, 3.8785E+02, 3.8877E+02, 3.8968E+02,
     + 3.9060E+02, 3.9152E+02, 3.9245E+02, 3.9337E+02, 3.9429E+02,
     +  3.9522E+02, 3.9614E+02, 3.9707E+02, 3.9799E+02, 3.9892E+02,
     + 3.9985E+02, 4.0078E+02, 4.0171E+02, 4.0264E+02, 4.0357E+02,
     +  4.0450E+02, 4.0543E+02, 4.0637E+02, 4.0730E+02, 4.0824E+02,
     + 4.0918E+02, 4.1011E+02, 4.1105E+02, 4.1199E+02, 4.1293E+02,
     +  4.1387E+02, 4.1481E+02, 4.1576E+02, 4.1670E+02, 4.1764E+02,
     + 4.1859E+02, 4.1954E+02, 4.2048E+02, 4.2143E+02, 4.2238E+02,
     +  4.2333E+02, 4.2428E+02, 4.2523E+02, 4.2618E+02, 4.2714E+02,
     + 4.2809E+02, 4.2905E+02, 4.3000E+02, 4.3096E+02, 4.3192E+02,
     +  4.3288E+02, 4.3384E+02, 4.3480E+02, 4.3576E+02, 4.3672E+02,
     + 4.3768E+02, 4.3865E+02, 4.3961E+02, 4.4058E+02, 4.4154E+02,
     +  4.4251E+02, 4.4348E+02, 4.4445E+02, 4.4542E+02, 4.4639E+02,
     + 4.4736E+02, 4.4833E+02, 4.4931E+02, 4.5028E+02, 4.5126E+02,
     +  4.5223E+02, 4.5321E+02, 4.5419E+02, 4.5517E+02, 4.5615E+02,
     + 4.5713E+02, 4.5811E+02, 4.5910E+02, 4.6008E+02, 4.6106E+02,
     +  4.6205E+02, 4.6304E+02, 4.6402E+02, 4.6501E+02, 4.6600E+02,
     + 4.6699E+02, 4.6798E+02, 4.6897E+02, 4.6997E+02, 4.7096E+02,
     +  4.7195E+02, 4.7295E+02, 4.7395E+02, 4.7494E+02, 4.7594E+02,
     + 4.7694E+02, 4.7794E+02, 4.7894E+02, 4.7994E+02, 4.8095E+02,
     +  4.8195E+02, 4.8296E+02, 4.8396E+02, 4.8497E+02, 4.8598E+02,
     + 4.8698E+02, 4.8799E+02, 4.8900E+02, 4.9002E+02, 4.9103E+02,
     +  4.9204E+02, 4.9306E+02, 4.9407E+02, 4.9509E+02, 4.9610E+02,
     + 4.9712E+02, 4.9814E+02, 4.9916E+02, 5.0018E+02, 5.0120E+02,
     +  5.0222E+02, 5.0325E+02, 5.0427E+02, 5.0530E+02, 5.0632E+02,
     + 5.0735E+02, 5.0838E+02, 5.0941E+02, 5.1044E+02, 5.1147E+02,
     +  5.1250E+02, 5.1353E+02, 5.1457E+02, 5.1560E+02, 5.1664E+02,
     + 5.1767E+02, 5.1871E+02, 5.1975E+02, 5.2079E+02, 5.2183E+02,
     +  5.2287E+02, 5.2391E+02, 5.2496E+02, 5.2600E+02, 5.2705E+02,
     + 5.2809E+02, 5.2914E+02, 5.3019E+02, 5.3124E+02, 5.3229E+02,
     +  5.3334E+02, 5.3439E+02, 5.3544E+02, 5.3650E+02, 5.3755E+02,
     + 5.3861E+02, 5.3966E+02, 5.4072E+02, 5.4178E+02, 5.4284E+02,
     +  5.4390E+02, 5.4496E+02, 5.4602E+02, 5.4709E+02, 5.4815E+02,
     + 5.4922E+02, 5.5028E+02, 5.5135E+02, 5.5242E+02, 5.5349E+02,
     +  5.5456E+02, 5.5563E+02, 5.5670E+02, 5.5777E+02, 5.5885E+02,
     + 5.5992E+02, 5.6100E+02, 5.6207E+02, 5.6315E+02, 5.6423E+02,
     +  5.6531E+02, 5.6639E+02, 5.6747E+02, 5.6855E+02, 5.6964E+02,
     + 5.7072E+02, 5.7181E+02, 5.7289E+02, 5.7398E+02, 5.7507E+02,
     +  5.7616E+02, 5.7725E+02, 5.7834E+02, 5.7943E+02, 5.8053E+02,
     + 5.8162E+02, 5.8272E+02, 5.8381E+02, 5.8491E+02, 5.8601E+02,
     +  5.8711E+02, 5.8821E+02, 5.8931E+02, 5.9041E+02, 5.9151E+02,
     + 5.9262E+02, 5.9372E+02, 5.9483E+02, 5.9594E+02, 5.9704E+02,
     +  5.9815E+02, 5.9926E+02, 6.0037E+02, 6.0149E+02, 6.0260E+02,
     + 6.0371E+02, 6.0483E+02, 6.0594E+02, 6.0706E+02, 6.0818E+02,
     +  6.0930E+02, 6.1042E+02, 6.1154E+02, 6.1266E+02, 6.1378E+02,
     + 6.1490E+02, 6.1603E+02, 6.1716E+02, 6.1828E+02, 6.1941E+02,
     +  6.2054E+02, 6.2167E+02, 6.2280E+02, 6.2393E+02, 6.2506E+02,
     + 6.2620E+02, 6.2733E+02, 6.2847E+02, 6.2960E+02, 6.3074E+02,
     +  6.3188E+02, 6.3302E+02, 6.3416E+02, 6.3530E+02, 6.3644E+02,
     + 6.3759E+02, 6.3873E+02, 6.3988E+02, 6.4102E+02, 6.4217E+02,
     +  6.4332E+02, 6.4447E+02, 6.4562E+02, 6.4677E+02, 6.4792E+02,
     + 6.4908E+02, 6.5023E+02, 6.5139E+02, 6.5254E+02, 6.5370E+02,
     +  6.5486E+02, 6.5602E+02, 6.5718E+02, 6.5834E+02, 6.5950E+02,
     + 6.6066E+02, 6.6183E+02, 6.6299E+02, 6.6416E+02, 6.6533E+02,
     +  6.6650E+02, 6.6767E+02, 6.6884E+02, 6.7001E+02, 6.7118E+02,
     + 6.7235E+02, 6.7353E+02, 6.7470E+02, 6.7588E+02, 6.7706E+02,
     +  6.7824E+02, 6.7941E+02, 6.8059E+02, 6.8178E+02, 6.8296E+02,
     + 6.8414E+02, 6.8533E+02, 6.8651E+02, 6.8770E+02, 6.8889E+02,
     +  6.9007E+02, 6.9126E+02, 6.9245E+02, 6.9365E+02, 6.9484E+02,
     + 6.9603E+02, 6.9723E+02, 6.9842E+02, 6.9962E+02, 7.0082E+02,
     +  7.0201E+02, 7.0321E+02, 7.0441E+02, 7.0562E+02, 7.0682E+02,
     + 7.0802E+02, 7.0923E+02, 7.1043E+02, 7.1164E+02, 7.1285E+02,
     +  7.1405E+02, 7.1526E+02, 7.1647E+02, 7.1769E+02, 7.1890E+02,
     + 7.2011E+02, 7.2133E+02, 7.2254E+02, 7.2376E+02, 7.2498E+02,
     +  7.2620E+02, 7.2742E+02, 7.2864E+02, 7.2986E+02, 7.3108E+02,
     + 7.3231E+02, 7.3353E+02, 7.3476E+02, 7.3598E+02, 7.3721E+02,
     +  7.3844E+02, 7.3967E+02, 7.4090E+02, 7.4213E+02, 7.4336E+02,
     + 7.4460E+02, 7.4583E+02, 7.4707E+02, 7.4831E+02, 7.4954E+02,
     +  7.5078E+02, 7.5202E+02, 7.5326E+02, 7.5451E+02, 7.5575E+02,
     + 7.5699E+02, 7.5824E+02, 7.5949E+02, 7.6073E+02, 7.6198E+02,
     +  7.6323E+02, 7.6448E+02, 7.6573E+02, 7.6698E+02, 7.6824E+02,
     + 7.6949E+02, 7.7075E+02, 7.7200E+02, 7.7326E+02, 7.7452E+02,
     +  7.7578E+02, 7.7704E+02, 7.7830E+02, 7.7956E+02, 7.8083E+02,
     + 7.8209E+02, 7.8336E+02, 7.8462E+02, 7.8589E+02, 7.8716E+02,
     +  7.8843E+02, 7.8970E+02, 7.9097E+02, 7.9224E+02, 7.9352E+02,
     + 7.9479E+02, 7.9607E+02, 7.9735E+02, 7.9862E+02, 7.9990E+02,
     +  8.0118E+02, 8.0246E+02, 8.0374E+02, 8.0503E+02, 8.0631E+02,
     + 8.0760E+02, 8.0888E+02, 8.1017E+02, 8.1146E+02, 8.1275E+02,
     +  8.1404E+02, 8.1533E+02, 8.1662E+02, 8.1791E+02, 8.1921E+02,
     + 8.2050E+02, 8.2180E+02, 8.2310E+02, 8.2440E+02, 8.2570E+02,
     +  8.2700E+02, 8.2830E+02, 8.2960E+02, 8.3090E+02, 8.3221E+02,
     + 8.3351E+02, 8.3482E+02, 8.3613E+02, 8.3744E+02, 8.3875E+02,
     +  8.4006E+02, 8.4137E+02, 8.4268E+02, 8.4400E+02, 8.4531E+02,
     + 8.4663E+02, 8.4794E+02, 8.4926E+02, 8.5058E+02, 8.5190E+02,
     +  8.5322E+02, 8.5454E+02, 8.5587E+02, 8.5719E+02, 8.5852E+02,
     + 8.5984E+02, 8.6117E+02, 8.6250E+02, 8.6383E+02, 8.6516E+02,
     +  8.6649E+02, 8.6782E+02, 8.6915E+02, 8.7049E+02, 8.7183E+02,
     + 8.7316E+02, 8.7450E+02, 8.7584E+02, 8.7718E+02, 8.7852E+02,
     +  8.7986E+02, 8.8120E+02, 8.8255E+02, 8.8389E+02, 8.8524E+02,
     + 8.8658E+02, 8.8793E+02, 8.8928E+02, 8.9063E+02, 8.9198E+02,
     +  8.9333E+02, 8.9469E+02, 8.9604E+02, 8.9740E+02, 8.9875E+02,
     + 9.0011E+02, 9.0147E+02, 9.0283E+02, 9.0419E+02, 9.0555E+02,
     +  9.0691E+02, 9.0828E+02, 9.0964E+02, 9.1101E+02, 9.1237E+02,
     + 9.1374E+02, 9.1511E+02, 9.1648E+02, 9.1785E+02, 9.1922E+02,
     +  9.2060E+02, 9.2197E+02, 9.2334E+02, 9.2472E+02, 9.2610E+02,
     + 9.2748E+02, 9.2886E+02, 9.3024E+02, 9.3162E+02, 9.3300E+02,
     +  9.3438E+02, 9.3577E+02, 9.3715E+02, 9.3854E+02, 9.3993E+02,
     + 9.4131E+02, 9.4270E+02, 9.4409E+02, 9.4549E+02, 9.4688E+02,
     +  9.4827E+02, 9.4967E+02, 9.5106E+02, 9.5246E+02, 9.5386E+02,
     + 9.5526E+02, 9.5666E+02, 9.5806E+02, 9.5946E+02, 9.6086E+02,
     +  9.6227E+02, 9.6367E+02, 9.6508E+02, 9.6649E+02, 9.6790E+02,
     + 9.6930E+02, 9.7072E+02, 9.7213E+02, 9.7354E+02, 9.7495E+02,
     +  9.7637E+02, 9.7778E+02, 9.7920E+02, 9.8062E+02, 9.8204E+02,
     + 9.8346E+02, 9.8488E+02, 9.8630E+02, 9.8772E+02, 9.8915E+02,
     +  9.9057E+02, 9.9200E+02, 9.9343E+02, 9.9485E+02, 9.9628E+02,
     + 9.9771E+02, 9.9914E+02, 1.0006E+03, 1.0020E+03, 1.0034E+03,
     +  1.0049E+03, 1.0063E+03, 1.0078E+03, 1.0092E+03, 1.0106E+03,
     + 1.0121E+03, 1.0135E+03, 1.0150E+03, 1.0164E+03, 1.0178E+03,
     +  1.0193E+03, 1.0207E+03, 1.0222E+03, 1.0236E+03, 1.0251E+03,
     + 1.0265E+03, 1.0280E+03, 1.0294E+03, 1.0309E+03, 1.0324E+03,
     +  1.0338E+03, 1.0353E+03, 1.0367E+03, 1.0382E+03, 1.0397E+03,
     + 1.0411E+03, 1.0426E+03, 1.0440E+03, 1.0455E+03, 1.0470E+03,
     +  1.0484E+03, 1.0499E+03, 1.0514E+03, 1.0528E+03, 1.0543E+03,
     + 1.0558E+03, 1.0573E+03, 1.0587E+03, 1.0602E+03, 1.0617E+03,
     +  1.0632E+03, 1.0646E+03, 1.0661E+03, 1.0676E+03, 1.0691E+03,
     + 1.0706E+03, 1.0721E+03, 1.0735E+03, 1.0750E+03, 1.0765E+03,
     +  1.0780E+03, 1.0795E+03, 1.0810E+03, 1.0825E+03, 1.0840E+03,
     + 1.0855E+03, 1.0870E+03, 1.0884E+03, 1.0899E+03, 1.0914E+03,
     +  1.0929E+03, 1.0944E+03, 1.0959E+03, 1.0974E+03, 1.0989E+03,
     + 1.1005E+03, 1.1020E+03, 1.1035E+03, 1.1050E+03, 1.1065E+03,
     +  1.1080E+03, 1.1095E+03, 1.1110E+03, 1.1125E+03, 1.1140E+03,
     + 1.1156E+03, 1.1171E+03, 1.1186E+03, 1.1201E+03, 1.1216E+03,
     +  1.1231E+03, 1.1247E+03, 1.1262E+03, 1.1277E+03, 1.1292E+03,
     + 1.1308E+03, 1.1323E+03, 1.1338E+03, 1.1353E+03, 1.1369E+03,
     +  1.1384E+03, 1.1399E+03, 1.1415E+03, 1.1430E+03, 1.1445E+03,
     + 1.1461E+03, 1.1476E+03, 1.1491E+03, 1.1507E+03, 1.1522E+03,
     +  1.1538E+03, 1.1553E+03, 1.1568E+03, 1.1584E+03, 1.1599E+03,
     + 1.1615E+03, 1.1630E+03, 1.1646E+03, 1.1661E+03, 1.1677E+03,
     +  1.1692E+03, 1.1708E+03, 1.1723E+03, 1.1739E+03, 1.1754E+03,
     + 1.1770E+03, 1.1786E+03, 1.1801E+03, 1.1817E+03, 1.1832E+03,
     +  1.1848E+03, 1.1864E+03, 1.1879E+03, 1.1895E+03, 1.1911E+03,
     + 1.1926E+03, 1.1942E+03, 1.1958E+03, 1.1973E+03, 1.1989E+03,
     +  1.2005E+03, 1.2021E+03, 1.2036E+03, 1.2052E+03, 1.2068E+03,
     + 1.2084E+03, 1.2099E+03, 1.2115E+03, 1.2131E+03, 1.2147E+03,
     +  1.2163E+03, 1.2179E+03, 1.2194E+03, 1.2210E+03, 1.2226E+03,
     + 1.2242E+03, 1.2258E+03, 1.2274E+03, 1.2290E+03, 1.2306E+03,
     +  1.2322E+03, 1.2338E+03, 1.2354E+03, 1.2370E+03, 1.2386E+03,
     + 1.2402E+03, 1.2418E+03, 1.2434E+03, 1.2450E+03, 1.2466E+03,
     +  1.2482E+03, 1.2498E+03, 1.2514E+03, 1.2530E+03, 1.2546E+03,
     + 1.2562E+03, 1.2578E+03, 1.2594E+03, 1.2610E+03, 1.2627E+03,
     +  1.2643E+03, 1.2659E+03, 1.2675E+03, 1.2691E+03, 1.2707E+03,
     + 1.2724E+03, 1.2740E+03, 1.2756E+03, 1.2772E+03, 1.2789E+03,
     +  1.2805E+03, 1.2821E+03, 1.2837E+03, 1.2854E+03, 1.2870E+03,
     + 1.2886E+03, 1.2903E+03, 1.2919E+03, 1.2935E+03, 1.2952E+03,
     +  1.2968E+03, 1.2984E+03, 1.3001E+03, 1.3017E+03, 1.3034E+03,
     + 1.3050E+03, 1.3067E+03, 1.3083E+03, 1.3099E+03, 1.3116E+03,
     +  1.3132E+03, 1.3149E+03, 1.3165E+03, 1.3182E+03, 1.3198E+03,
     + 1.3215E+03, 1.3231E+03, 1.3248E+03, 1.3265E+03, 1.3281E+03,
     +  1.3298E+03, 1.3314E+03, 1.3331E+03, 1.3348E+03, 1.3364E+03,
     + 1.3381E+03, 1.3397E+03, 1.3414E+03, 1.3431E+03, 1.3447E+03,
     +  1.3464E+03, 1.3481E+03, 1.3498E+03, 1.3514E+03, 1.3531E+03,
     + 1.3548E+03, 1.3565E+03, 1.3581E+03, 1.3598E+03, 1.3615E+03,
     +  1.3632E+03, 1.3648E+03, 1.3665E+03, 1.3682E+03, 1.3699E+03,
     + 1.3716E+03, 1.3733E+03, 1.3750E+03, 1.3766E+03, 1.3783E+03,
     +  1.3800E+03, 1.3817E+03, 1.3834E+03, 1.3851E+03, 1.3868E+03,
     + 1.3885E+03, 1.3902E+03, 1.3919E+03, 1.3936E+03, 1.3953E+03,
     +  1.3970E+03, 1.3987E+03, 1.4004E+03, 1.4021E+03, 1.4038E+03,
     + 1.4055E+03, 1.4072E+03, 1.4089E+03, 1.4106E+03, 1.4123E+03,
     +  1.4141E+03, 1.4158E+03, 1.4175E+03, 1.4192E+03, 1.4209E+03,
     + 1.4226E+03, 1.4244E+03, 1.4261E+03, 1.4278E+03, 1.4295E+03,
     +  1.4312E+03, 1.4330E+03, 1.4347E+03, 1.4364E+03, 1.4381E+03,
     + 1.4399E+03, 1.4416E+03, 1.4433E+03, 1.4451E+03, 1.4468E+03,
     +  1.4485E+03, 1.4503E+03, 1.4520E+03, 1.4537E+03, 1.4555E+03,
     + 1.4572E+03, 1.4589E+03, 1.4607E+03, 1.4624E+03, 1.4642E+03,
     +  1.4659E+03, 1.4677E+03, 1.4694E+03, 1.4712E+03, 1.4729E+03,
     + 1.4747E+03, 1.4764E+03, 1.4782E+03, 1.4799E+03, 1.4817E+03,
     +  1.4834E+03, 1.4852E+03, 1.4869E+03, 1.4887E+03, 1.4905E+03,
     + 1.4922E+03, 1.4940E+03, 1.4957E+03, 1.4975E+03, 1.4993E+03,
     +  1.5010E+03, 1.5028E+03, 1.5046E+03, 1.5063E+03, 1.5081E+03,
     + 1.5099E+03, 1.5116E+03, 1.5134E+03, 1.5152E+03, 1.5170E+03,
     +  1.5187E+03, 1.5205E+03, 1.5223E+03, 1.5241E+03, 1.5259E+03,
     + 1.5276E+03, 1.5294E+03, 1.5312E+03, 1.5330E+03, 1.5348E+03,
     +  1.5366E+03, 1.5384E+03, 1.5402E+03, 1.5419E+03, 1.5437E+03,
     + 1.5455E+03, 1.5473E+03, 1.5491E+03, 1.5509E+03, 1.5527E+03,
     +  1.5545E+03, 1.5563E+03, 1.5581E+03, 1.5599E+03, 1.5617E+03,
     + 1.5635E+03, 1.5653E+03, 1.5671E+03, 1.5689E+03, 1.5707E+03,
     +  1.5726E+03, 1.5744E+03, 1.5762E+03, 1.5780E+03, 1.5798E+03,
     + 1.5816E+03, 1.5834E+03, 1.5852E+03, 1.5871E+03, 1.5889E+03,
     +  1.5907E+03, 1.5925E+03, 1.5943E+03, 1.5962E+03, 1.5980E+03,
     + 1.5998E+03, 1.6016E+03, 1.6035E+03, 1.6053E+03, 1.6071E+03,
     +  1.6090E+03, 1.6108E+03, 1.6126E+03, 1.6145E+03, 1.6163E+03,
     + 1.6181E+03, 1.6200E+03, 1.6218E+03, 1.6236E+03, 1.6255E+03,
     +  1.6273E+03, 1.6292E+03, 1.6310E+03, 1.6329E+03, 1.6347E+03,
     + 1.6366E+03, 1.6384E+03, 1.6403E+03, 1.6421E+03, 1.6440E+03,
     +  1.6458E+03, 1.6477E+03, 1.6495E+03, 1.6514E+03, 1.6532E+03,
     + 1.6551E+03, 1.6569E+03, 1.6588E+03, 1.6607E+03, 1.6625E+03,
     +  1.6644E+03, 1.6663E+03, 1.6681E+03, 1.6700E+03, 1.6719E+03,
     + 1.6737E+03, 1.6756E+03, 1.6775E+03, 1.6793E+03, 1.6812E+03,
     +  1.6831E+03, 1.6850E+03, 1.6868E+03, 1.6887E+03, 1.6906E+03,
     + 1.6925E+03, 1.6944E+03, 1.6962E+03, 1.6981E+03, 1.7000E+03,
     +  1.7019E+03, 1.7038E+03, 1.7057E+03, 1.7075E+03, 1.7094E+03,
     + 1.7113E+03, 1.7132E+03, 1.7151E+03, 1.7170E+03, 1.7189E+03,
     +  1.7208E+03, 1.7227E+03, 1.7246E+03, 1.7265E+03, 1.7284E+03,
     + 1.7303E+03, 1.7322E+03, 1.7341E+03, 1.7360E+03, 1.7379E+03,
     +  1.7398E+03, 1.7417E+03, 1.7436E+03, 1.7455E+03, 1.7475E+03,
     + 1.7494E+03, 1.7513E+03, 1.7532E+03, 1.7551E+03, 1.7570E+03,
     +  1.7589E+03, 1.7609E+03, 1.7628E+03, 1.7647E+03, 1.7666E+03,
     + 1.7686E+03, 1.7705E+03, 1.7724E+03, 1.7743E+03, 1.7763E+03,
     +  1.7782E+03, 1.7801E+03, 1.7820E+03, 1.7840E+03, 1.7859E+03,
     + 1.7878E+03, 1.7898E+03, 1.7917E+03, 1.7937E+03, 1.7956E+03,
     +  1.7975E+03, 1.7995E+03, 1.8014E+03, 1.8034E+03, 1.8053E+03,
     + 1.8072E+03, 1.8092E+03, 1.8111E+03, 1.8131E+03, 1.8150E+03,
     +  1.8170E+03, 1.8189E+03, 1.8209E+03, 1.8228E+03, 1.8248E+03,
     + 1.8268E+03, 1.8287E+03, 1.8307E+03, 1.8326E+03, 1.8346E+03,
     +  1.8366E+03, 1.8385E+03, 1.8405E+03, 1.8424E+03, 1.8444E+03,
     + 1.8464E+03, 1.8483E+03, 1.8503E+03, 1.8523E+03, 1.8543E+03,
     +  1.8562E+03, 1.8582E+03, 1.8602E+03, 1.8622E+03, 1.8641E+03,
     + 1.8661E+03, 1.8681E+03, 1.8701E+03, 1.8721E+03, 1.8740E+03,
     +  1.8760E+03, 1.8780E+03, 1.8800E+03, 1.8820E+03, 1.8840E+03,
     + 1.8860E+03, 1.8879E+03, 1.8899E+03, 1.8919E+03, 1.8939E+03,
     +  1.8959E+03, 1.8979E+03, 1.8999E+03, 1.9019E+03, 1.9039E+03,
     + 1.9059E+03, 1.9079E+03, 1.9099E+03, 1.9119E+03, 1.9139E+03,
     +  1.9159E+03, 1.9179E+03, 1.9199E+03, 1.9219E+03, 1.9239E+03,
     + 1.9260E+03, 1.9280E+03, 1.9300E+03, 1.9320E+03, 1.9340E+03,
     +  1.9360E+03, 1.9380E+03, 1.9401E+03, 1.9421E+03, 1.9441E+03,
     + 1.9461E+03, 1.9482E+03, 1.9502E+03, 1.9522E+03, 1.9542E+03,
     +  1.9563E+03, 1.9583E+03, 1.9603E+03, 1.9623E+03, 1.9644E+03,
     + 1.9664E+03, 1.9684E+03, 1.9705E+03, 1.9725E+03, 1.9746E+03,
     +  1.9766E+03, 1.9786E+03, 1.9807E+03, 1.9827E+03, 1.9848E+03,
     + 1.9868E+03, 1.9888E+03, 1.9909E+03, 1.9929E+03, 1.9950E+03,
     +  1.9970E+03, 1.9991E+03, 2.0011E+03, 2.0032E+03, 2.0052E+03,
     + 2.0073E+03, 2.0094E+03, 2.0114E+03, 2.0135E+03, 2.0155E+03,
     +  2.0176E+03, 2.0196E+03, 2.0217E+03, 2.0238E+03, 2.0258E+03,
     + 2.0279E+03, 2.0300E+03, 2.0320E+03, 2.0341E+03, 2.0362E+03,
     +  2.0383E+03, 2.0403E+03, 2.0424E+03, 2.0445E+03, 2.0465E+03,
     + 2.0486E+03, 2.0507E+03, 2.0528E+03, 2.0549E+03, 2.0569E+03,
     +  2.0590E+03, 2.0611E+03, 2.0632E+03, 2.0653E+03, 2.0674E+03,
     + 2.0695E+03, 2.0715E+03, 2.0736E+03, 2.0757E+03, 2.0778E+03,
     +  2.0799E+03, 2.0820E+03, 2.0841E+03, 2.0862E+03, 2.0883E+03,
     + 2.0904E+03, 2.0925E+03, 2.0946E+03, 2.0967E+03, 2.0988E+03,
     +  2.1009E+03, 2.1030E+03, 2.1051E+03, 2.1072E+03, 2.1093E+03,
     + 2.1114E+03, 2.1136E+03, 2.1157E+03, 2.1178E+03, 2.1199E+03,
     +  2.1220E+03, 2.1241E+03, 2.1262E+03, 2.1284E+03, 2.1305E+03,
     + 2.1326E+03, 2.1347E+03, 2.1369E+03, 2.1390E+03, 2.1411E+03,
     +  2.1432E+03, 2.1454E+03, 2.1475E+03, 2.1496E+03, 2.1517E+03,
     + 2.1539E+03, 2.1560E+03, 2.1581E+03, 2.1603E+03, 2.1624E+03,
     +  2.1646E+03, 2.1667E+03, 2.1688E+03, 2.1710E+03, 2.1731E+03,
     + 2.1753E+03, 2.1774E+03, 2.1795E+03, 2.1817E+03, 2.1838E+03,
     +  2.1860E+03, 2.1881E+03, 2.1903E+03, 2.1924E+03, 2.1946E+03,
     + 2.1967E+03, 2.1989E+03, 2.2011E+03, 2.2032E+03, 2.2054E+03,
     +  2.2075E+03, 2.2097E+03, 2.2119E+03, 2.2140E+03, 2.2162E+03,
     + 2.2184E+03, 2.2205E+03, 2.2227E+03, 2.2249E+03, 2.2270E+03,
     +  2.2292E+03, 2.2314E+03, 2.2335E+03, 2.2357E+03, 2.2379E+03,
     + 2.2401E+03, 2.2422E+03, 2.2444E+03, 2.2466E+03, 2.2488E+03,
     +  2.2510E+03, 2.2531E+03, 2.2553E+03, 2.2575E+03, 2.2597E+03,
     + 2.2619E+03, 2.2641E+03, 2.2663E+03, 2.2685E+03, 2.2707E+03,
     +  2.2728E+03, 2.2750E+03, 2.2772E+03, 2.2794E+03, 2.2816E+03,
     + 2.2838E+03, 2.2860E+03, 2.2882E+03, 2.2904E+03, 2.2926E+03,
     +  2.2948E+03, 2.2970E+03, 2.2992E+03, 2.3015E+03, 2.3037E+03,
     + 2.3059E+03, 2.3081E+03, 2.3103E+03, 2.3125E+03, 2.3147E+03,
     +  2.3169E+03, 2.3192E+03, 2.3214E+03, 2.3236E+03, 2.3258E+03,
     + 2.3280E+03, 2.3303E+03, 2.3325E+03, 2.3347E+03, 2.3369E+03,
     +  2.3392E+03, 2.3414E+03, 2.3436E+03, 2.3458E+03, 2.3481E+03,
     + 2.3503E+03, 2.3525E+03, 2.3548E+03, 2.3570E+03, 2.3592E+03,
     +  2.3615E+03, 2.3637E+03, 2.3660E+03, 2.3682E+03, 2.3704E+03,
     + 2.3727E+03, 2.3749E+03, 2.3772E+03, 2.3794E+03, 2.3817E+03,
     +  2.3839E+03, 2.3862E+03, 2.3884E+03, 2.3907E+03, 2.3929E+03,
     + 2.3952E+03, 2.3974E+03, 2.3997E+03, 2.4019E+03, 2.4042E+03,
     +  2.4065E+03, 2.4087E+03, 2.4110E+03, 2.4132E+03, 2.4155E+03,
     + 2.4178E+03, 2.4200E+03, 2.4223E+03, 2.4246E+03, 2.4269E+03,
     +  2.4291E+03, 2.4314E+03, 2.4337E+03, 2.4359E+03, 2.4382E+03,
     + 2.4405E+03, 2.4428E+03, 2.4450E+03, 2.4473E+03, 2.4496E+03,
     +  2.4519E+03, 2.4542E+03, 2.4565E+03, 2.4587E+03, 2.4610E+03,
     + 2.4633E+03, 2.4656E+03, 2.4679E+03, 2.4702E+03, 2.4725E+03,
     +  2.4748E+03, 2.4771E+03, 2.4794E+03, 2.4817E+03, 2.4840E+03,
     + 2.4863E+03, 2.4886E+03, 2.4909E+03, 2.4932E+03, 2.4955E+03,
     +  2.4978E+03, 2.5001E+03, 2.5024E+03, 2.5047E+03, 2.5070E+03,
     + 2.5093E+03, 2.5116E+03, 2.5139E+03, 2.5162E+03, 2.5186E+03,
     +  2.5209E+03, 2.5232E+03, 2.5255E+03, 2.5278E+03, 2.5301E+03,
     + 2.5325E+03, 2.5348E+03, 2.5371E+03, 2.5394E+03, 2.5418E+03,
     +  2.5441E+03, 2.5464E+03, 2.5487E+03, 2.5511E+03, 2.5534E+03,
     + 2.5557E+03, 2.5581E+03, 2.5604E+03, 2.5627E+03, 2.5651E+03,
     +  2.5674E+03, 2.5698E+03, 2.5721E+03, 2.5744E+03, 2.5768E+03,
     + 2.5791E+03, 2.5815E+03, 2.5838E+03, 2.5862E+03, 2.5885E+03,
     +  2.5909E+03, 2.5932E+03, 2.5956E+03, 2.5979E+03, 2.6003E+03,
     + 2.6026E+03, 2.6050E+03, 2.6073E+03, 2.6097E+03, 2.6121E+03,
     +  2.6144E+03, 2.6168E+03, 2.6191E+03, 2.6215E+03, 2.6239E+03,
     + 2.6262E+03, 2.6286E+03, 2.6310E+03, 2.6333E+03, 2.6357E+03,
     +  2.6381E+03, 2.6405E+03, 2.6428E+03, 2.6452E+03, 2.6476E+03,
     + 2.6500E+03, 2.6523E+03, 2.6547E+03, 2.6571E+03, 2.6595E+03,
     +  2.6619E+03, 2.6642E+03, 2.6666E+03, 2.6690E+03, 2.6714E+03,
     + 2.6738E+03, 2.6762E+03, 2.6786E+03, 2.6810E+03, 2.6834E+03,
     +  2.6858E+03, 2.6881E+03, 2.6905E+03, 2.6929E+03, 2.6953E+03,
     + 2.6977E+03, 2.7001E+03, 2.7025E+03, 2.7049E+03, 2.7073E+03,
     +  2.7098E+03, 2.7122E+03, 2.7146E+03, 2.7170E+03, 2.7194E+03,
     + 2.7218E+03, 2.7242E+03, 2.7266E+03, 2.7290E+03, 2.7315E+03,
     +  2.7339E+03, 2.7363E+03, 2.7387E+03, 2.7411E+03, 2.7435E+03,
     + 2.7460E+03, 2.7484E+03, 2.7508E+03, 2.7532E+03, 2.7557E+03,
     +  2.7581E+03, 2.7605E+03, 2.7630E+03, 2.7654E+03, 2.7678E+03,
     + 2.7703E+03, 2.7727E+03, 2.7751E+03, 2.7776E+03, 2.7800E+03,
     +  2.7824E+03, 2.7849E+03, 2.7873E+03, 2.7898E+03, 2.7922E+03,
     + 2.7946E+03, 2.7971E+03, 2.7995E+03, 2.8020E+03, 2.8044E+03,
     +  2.8069E+03, 2.8093E+03, 2.8118E+03, 2.8142E+03, 2.8167E+03,
     + 2.8192E+03, 2.8216E+03, 2.8241E+03, 2.8265E+03, 2.8290E+03,
     +  2.8315E+03, 2.8339E+03, 2.8364E+03, 2.8388E+03, 2.8413E+03,
     + 2.8438E+03, 2.8462E+03, 2.8487E+03, 2.8512E+03, 2.8537E+03,
     +  2.8561E+03, 2.8586E+03, 2.8611E+03, 2.8636E+03, 2.8660E+03,
     + 2.8685E+03, 2.8710E+03, 2.8735E+03, 2.8760E+03, 2.8784E+03,
     +  2.8809E+03, 2.8834E+03, 2.8859E+03, 2.8884E+03, 2.8909E+03,
     + 2.8934E+03, 2.8959E+03, 2.8983E+03, 2.9008E+03, 2.9033E+03,
     +  2.9058E+03, 2.9083E+03, 2.9108E+03, 2.9133E+03, 2.9158E+03,
     + 2.9183E+03, 2.9208E+03, 2.9233E+03, 2.9258E+03, 2.9283E+03,
     +  2.9308E+03, 2.9334E+03, 2.9359E+03, 2.9384E+03, 2.9409E+03,
     + 2.9434E+03, 2.9459E+03, 2.9484E+03, 2.9509E+03, 2.9535E+03,
     +  2.9560E+03, 2.9585E+03, 2.9610E+03, 2.9635E+03, 2.9661E+03,
     + 2.9686E+03, 2.9711E+03, 2.9736E+03, 2.9762E+03, 2.9787E+03,
     +  2.9812E+03, 2.9837E+03, 2.9863E+03, 2.9888E+03, 2.9913E+03,
     + 2.9939E+03, 2.9964E+03, 2.9990E+03, 3.0015E+03, 3.0040E+03,
     +  3.0066E+03, 3.0091E+03, 3.0117E+03, 3.0142E+03, 3.0168E+03,
     + 3.0193E+03, 3.0218E+03, 3.0244E+03, 3.0269E+03, 3.0295E+03,
     +  3.0320E+03, 3.0346E+03, 3.0372E+03, 3.0397E+03, 3.0423E+03,
     + 3.0448E+03, 3.0474E+03, 3.0499E+03, 3.0525E+03, 3.0551E+03,
     +  3.0576E+03, 3.0602E+03, 3.0628E+03, 3.0653E+03, 3.0679E+03,
     + 3.0705E+03, 3.0730E+03, 3.0756E+03, 3.0782E+03, 3.0808E+03,
     +  3.0833E+03, 3.0859E+03, 3.0885E+03, 3.0911E+03, 3.0936E+03,
     + 3.0962E+03, 3.0988E+03, 3.1014E+03, 3.1040E+03, 3.1066E+03,
     +  3.1091E+03, 3.1117E+03, 3.1143E+03, 3.1169E+03, 3.1195E+03,
     + 3.1221E+03, 3.1247E+03, 3.1273E+03, 3.1299E+03, 3.1325E+03,
     +  3.1351E+03, 3.1377E+03, 3.1403E+03, 3.1429E+03, 3.1455E+03,
     + 3.1481E+03, 3.1507E+03, 3.1533E+03, 3.1559E+03, 3.1585E+03,
     +  3.1611E+03, 3.1637E+03, 3.1663E+03, 3.1689E+03, 3.1715E+03,
     + 3.1742E+03, 3.1768E+03, 3.1794E+03, 3.1820E+03, 3.1846E+03,
     +  3.1873E+03, 3.1899E+03, 3.1925E+03, 3.1951E+03, 3.1977E+03,
     + 3.2004E+03, 3.2030E+03, 3.2056E+03, 3.2083E+03, 3.2109E+03,
     +  3.2135E+03, 3.2161E+03, 3.2188E+03, 3.2214E+03, 3.2241E+03,
     + 3.2267E+03, 3.2293E+03, 3.2320E+03, 3.2346E+03, 3.2373E+03,
     +  3.2399E+03, 3.2425E+03, 3.2452E+03, 3.2478E+03, 3.2505E+03,
     + 3.2531E+03, 3.2558E+03, 3.2584E+03, 3.2611E+03, 3.2637E+03,
     +  3.2664E+03, 3.2690E+03, 3.2717E+03, 3.2744E+03, 3.2770E+03,
     + 3.2797E+03, 3.2823E+03, 3.2850E+03, 3.2877E+03, 3.2903E+03,
     +  3.2930E+03, 3.2957E+03, 3.2983E+03, 3.3010E+03, 3.3037E+03,
     + 3.3063E+03, 3.3090E+03, 3.3117E+03, 3.3144E+03, 3.3170E+03,
     +  3.3197E+03, 3.3224E+03, 3.3251E+03, 3.3278E+03, 3.3304E+03,
     + 3.3331E+03, 3.3358E+03, 3.3385E+03, 3.3412E+03, 3.3439E+03,
     +  3.3466E+03, 3.3492E+03, 3.3519E+03, 3.3546E+03, 3.3573E+03,
     + 3.3600E+03, 3.3627E+03, 3.3654E+03, 3.3681E+03, 3.3708E+03,
     +  3.3735E+03, 3.3762E+03, 3.3789E+03, 3.3816E+03, 3.3843E+03,
     + 3.3870E+03, 3.3897E+03, 3.3924E+03, 3.3951E+03, 3.3979E+03,
     +  3.4006E+03, 3.4033E+03, 3.4060E+03, 3.4087E+03, 3.4114E+03,
     + 3.4141E+03, 3.4169E+03, 3.4196E+03, 3.4223E+03, 3.4250E+03,
     +  3.4277E+03, 3.4305E+03, 3.4332E+03, 3.4359E+03, 3.4387E+03,
     + 3.4414E+03, 3.4441E+03, 3.4468E+03, 3.4496E+03, 3.4523E+03,
     +  3.4550E+03, 3.4578E+03, 3.4605E+03, 3.4633E+03, 3.4660E+03,
     + 3.4687E+03, 3.4715E+03, 3.4742E+03, 3.4770E+03, 3.4797E+03,
     +  3.4824E+03, 3.4852E+03, 3.4879E+03, 3.4907E+03, 3.4934E+03,
     + 3.4962E+03, 3.4989E+03, 3.5017E+03, 3.5045E+03, 3.5072E+03,
     +  3.5100E+03, 3.5127E+03, 3.5155E+03, 3.5183E+03, 3.5210E+03,
     + 3.5238E+03, 3.5265E+03, 3.5293E+03, 3.5321E+03, 3.5348E+03,
     +  3.5376E+03, 3.5404E+03, 3.5431E+03, 3.5459E+03, 3.5487E+03,
     + 3.5515E+03, 3.5542E+03, 3.5570E+03, 3.5598E+03, 3.5626E+03,
     +  3.5654E+03, 3.5681E+03, 3.5709E+03, 3.5737E+03, 3.5765E+03,
     + 3.5793E+03, 3.5821E+03, 3.5849E+03, 3.5876E+03, 3.5904E+03,
     +  3.5932E+03, 3.5960E+03, 3.5988E+03, 3.6016E+03, 3.6044E+03,
     + 3.6072E+03, 3.6100E+03, 3.6128E+03, 3.6156E+03, 3.6184E+03,
     +  3.6212E+03, 3.6240E+03, 3.6268E+03, 3.6296E+03, 3.6324E+03,
     + 3.6352E+03, 3.6381E+03, 3.6409E+03, 3.6437E+03, 3.6465E+03,
     +  3.6493E+03, 3.6521E+03, 3.6549E+03, 3.6578E+03, 3.6606E+03,
     + 3.6634E+03, 3.6662E+03, 3.6690E+03, 3.6719E+03, 3.6747E+03,
     +  3.6775E+03, 3.6803E+03, 3.6832E+03, 3.6860E+03, 3.6888E+03,
     + 3.6917E+03, 3.6945E+03, 3.6973E+03, 3.7002E+03, 3.7030E+03,
     +  3.7058E+03, 3.7087E+03, 3.7115E+03, 3.7144E+03, 3.7172E+03,
     + 3.7201E+03, 3.7229E+03, 3.7257E+03, 3.7286E+03, 3.7314E+03,
     +  3.7343E+03, 3.7371E+03, 3.7400E+03, 3.7428E+03, 3.7457E+03,
     + 3.7486E+03, 3.7514E+03, 3.7543E+03, 3.7571E+03, 3.7600E+03,
     +  3.7628E+03, 3.7657E+03, 3.7686E+03, 3.7714E+03, 3.7743E+03,
     + 3.7772E+03, 3.7800E+03, 3.7829E+03, 3.7858E+03, 3.7886E+03,
     +  3.7915E+03, 3.7944E+03, 3.7973E+03, 3.8001E+03, 3.8030E+03,
     + 3.8059E+03, 3.8088E+03, 3.8117E+03, 3.8145E+03, 3.8174E+03,
     +  3.8203E+03, 3.8232E+03, 3.8261E+03, 3.8290E+03, 3.8319E+03,
     + 3.8347E+03, 3.8376E+03, 3.8405E+03, 3.8434E+03, 3.8463E+03,
     +  3.8492E+03, 3.8521E+03, 3.8550E+03, 3.8579E+03, 3.8608E+03,
     + 3.8637E+03, 3.8666E+03, 3.8695E+03, 3.8724E+03, 3.8753E+03,
     +  3.8782E+03, 3.8811E+03, 3.8841E+03, 3.8870E+03, 3.8899E+03,
     + 3.8928E+03, 3.8957E+03, 3.8986E+03, 3.9015E+03, 3.9044E+03,
     +  3.9074E+03, 3.9103E+03, 3.9132E+03, 3.9161E+03, 3.9191E+03,
     + 3.9220E+03, 3.9249E+03, 3.9278E+03, 3.9308E+03, 3.9337E+03,
     +  3.9366E+03, 3.9396E+03, 3.9425E+03, 3.9454E+03, 3.9484E+03,
     + 3.9513E+03, 3.9542E+03, 3.9572E+03, 3.9601E+03, 3.9630E+03,
     +  3.9660E+03, 3.9689E+03, 3.9719E+03, 3.9748E+03, 3.9778E+03,
     + 3.9807E+03, 3.9837E+03, 3.9866E+03, 3.9896E+03, 3.9925E+03,
     +  3.9955E+03, 3.9984E+03, 4.0014E+03, 4.0043E+03, 4.0073E+03,
     + 4.0103E+03, 4.0132E+03, 4.0162E+03, 4.0191E+03, 4.0221E+03,
     +  4.0251E+03, 4.0280E+03, 4.0310E+03, 4.0340E+03, 4.0369E+03/
      data (QofT( 3,J),J=2409,4510)/ 4.0399E+03, 4.0429E+03,
     +  4.0459E+03, 4.0488E+03, 4.0518E+03, 4.0548E+03, 4.0578E+03,
     +  4.0607E+03,
     + 4.0637E+03, 4.0667E+03, 4.0697E+03, 4.0727E+03, 4.0757E+03,
     +  4.0786E+03, 4.0816E+03, 4.0846E+03, 4.0876E+03, 4.0906E+03,
     + 4.0936E+03, 4.0966E+03, 4.0996E+03, 4.1026E+03, 4.1056E+03,
     +  4.1086E+03, 4.1116E+03, 4.1146E+03, 4.1176E+03, 4.1206E+03,
     + 4.1236E+03, 4.1266E+03, 4.1296E+03, 4.1326E+03, 4.1356E+03,
     +  4.1386E+03, 4.1416E+03, 4.1446E+03, 4.1476E+03, 4.1507E+03,
     + 4.1537E+03, 4.1567E+03, 4.1597E+03, 4.1627E+03, 4.1657E+03,
     +  4.1688E+03, 4.1718E+03, 4.1748E+03, 4.1778E+03, 4.1809E+03,
     + 4.1839E+03, 4.1869E+03, 4.1899E+03, 4.1930E+03, 4.1960E+03,
     +  4.1990E+03, 4.2021E+03, 4.2051E+03, 4.2081E+03, 4.2112E+03,
     + 4.2142E+03, 4.2173E+03, 4.2203E+03, 4.2233E+03, 4.2264E+03,
     +  4.2294E+03, 4.2325E+03, 4.2355E+03, 4.2386E+03, 4.2416E+03,
     + 4.2447E+03, 4.2477E+03, 4.2508E+03, 4.2538E+03, 4.2569E+03,
     +  4.2599E+03, 4.2630E+03, 4.2660E+03, 4.2691E+03, 4.2722E+03,
     + 4.2752E+03, 4.2783E+03, 4.2814E+03, 4.2844E+03, 4.2875E+03,
     +  4.2906E+03, 4.2936E+03, 4.2967E+03, 4.2998E+03, 4.3028E+03,
     + 4.3059E+03, 4.3090E+03, 4.3121E+03, 4.3151E+03, 4.3182E+03,
     +  4.3213E+03, 4.3244E+03, 4.3275E+03, 4.3305E+03, 4.3336E+03,
     + 4.3367E+03, 4.3398E+03, 4.3429E+03, 4.3460E+03, 4.3491E+03,
     +  4.3521E+03, 4.3552E+03, 4.3583E+03, 4.3614E+03, 4.3645E+03,
     + 4.3676E+03, 4.3707E+03, 4.3738E+03, 4.3769E+03, 4.3800E+03,
     +  4.3831E+03, 4.3862E+03, 4.3893E+03, 4.3924E+03, 4.3955E+03,
     + 4.3987E+03, 4.4018E+03, 4.4049E+03, 4.4080E+03, 4.4111E+03,
     +  4.4142E+03, 4.4173E+03, 4.4204E+03, 4.4236E+03, 4.4267E+03,
     + 4.4298E+03, 4.4329E+03, 4.4360E+03, 4.4392E+03, 4.4423E+03,
     +  4.4454E+03, 4.4485E+03, 4.4517E+03, 4.4548E+03, 4.4579E+03,
     + 4.4611E+03, 4.4642E+03, 4.4673E+03, 4.4705E+03, 4.4736E+03,
     +  4.4767E+03, 4.4799E+03, 4.4830E+03, 4.4862E+03, 4.4893E+03,
     + 4.4924E+03, 4.4956E+03, 4.4987E+03, 4.5019E+03, 4.5050E+03,
     +  4.5082E+03, 4.5113E+03, 4.5145E+03, 4.5176E+03, 4.5208E+03,
     + 4.5239E+03, 4.5271E+03, 4.5303E+03, 4.5334E+03, 4.5366E+03,
     +  4.5397E+03, 4.5429E+03, 4.5461E+03, 4.5492E+03, 4.5524E+03,
     + 4.5556E+03, 4.5587E+03, 4.5619E+03, 4.5651E+03, 4.5682E+03,
     +  4.5714E+03, 4.5746E+03, 4.5778E+03, 4.5809E+03, 4.5841E+03,
     + 4.5873E+03, 4.5905E+03, 4.5937E+03, 4.5968E+03, 4.6000E+03,
     +  4.6032E+03, 4.6064E+03, 4.6096E+03, 4.6128E+03, 4.6160E+03,
     + 4.6191E+03, 4.6223E+03, 4.6255E+03, 4.6287E+03, 4.6319E+03,
     +  4.6351E+03, 4.6383E+03, 4.6415E+03, 4.6447E+03, 4.6479E+03,
     + 4.6511E+03, 4.6543E+03, 4.6575E+03, 4.6607E+03, 4.6639E+03,
     +  4.6671E+03, 4.6704E+03, 4.6736E+03, 4.6768E+03, 4.6800E+03,
     + 4.6832E+03, 4.6864E+03, 4.6896E+03, 4.6929E+03, 4.6961E+03,
     +  4.6993E+03, 4.7025E+03, 4.7057E+03, 4.7090E+03, 4.7122E+03,
     + 4.7154E+03, 4.7186E+03, 4.7219E+03, 4.7251E+03, 4.7283E+03,
     +  4.7316E+03, 4.7348E+03, 4.7380E+03, 4.7413E+03, 4.7445E+03,
     + 4.7477E+03, 4.7510E+03, 4.7542E+03, 4.7574E+03, 4.7607E+03,
     +  4.7639E+03, 4.7672E+03, 4.7704E+03, 4.7737E+03, 4.7769E+03,
     + 4.7802E+03, 4.7834E+03, 4.7867E+03, 4.7899E+03, 4.7932E+03,
     +  4.7964E+03, 4.7997E+03, 4.8029E+03, 4.8062E+03, 4.8095E+03,
     + 4.8127E+03, 4.8160E+03, 4.8193E+03, 4.8225E+03, 4.8258E+03,
     +  4.8290E+03, 4.8323E+03, 4.8356E+03, 4.8389E+03, 4.8421E+03,
     + 4.8454E+03, 4.8487E+03, 4.8520E+03, 4.8552E+03, 4.8585E+03,
     +  4.8618E+03, 4.8651E+03, 4.8683E+03, 4.8716E+03, 4.8749E+03,
     + 4.8782E+03, 4.8815E+03, 4.8848E+03, 4.8881E+03, 4.8913E+03,
     +  4.8946E+03, 4.8979E+03, 4.9012E+03, 4.9045E+03, 4.9078E+03,
     + 4.9111E+03, 4.9144E+03, 4.9177E+03, 4.9210E+03, 4.9243E+03,
     +  4.9276E+03, 4.9309E+03, 4.9342E+03, 4.9375E+03, 4.9408E+03,
     + 4.9441E+03, 4.9474E+03, 4.9508E+03, 4.9541E+03, 4.9574E+03,
     +  4.9607E+03, 4.9640E+03, 4.9673E+03, 4.9706E+03, 4.9740E+03,
     + 4.9773E+03, 4.9806E+03, 4.9839E+03, 4.9872E+03, 4.9906E+03,
     +  4.9939E+03, 4.9972E+03, 5.0006E+03, 5.0039E+03, 5.0072E+03,
     + 5.0105E+03, 5.0139E+03, 5.0172E+03, 5.0205E+03, 5.0239E+03,
     +  5.0272E+03, 5.0306E+03, 5.0339E+03, 5.0372E+03, 5.0406E+03,
     + 5.0439E+03, 5.0473E+03, 5.0506E+03, 5.0540E+03, 5.0573E+03,
     +  5.0607E+03, 5.0640E+03, 5.0674E+03, 5.0707E+03, 5.0741E+03,
     + 5.0774E+03, 5.0808E+03, 5.0841E+03, 5.0875E+03, 5.0909E+03,
     +  5.0942E+03, 5.0976E+03, 5.1010E+03, 5.1043E+03, 5.1077E+03,
     + 5.1111E+03, 5.1144E+03, 5.1178E+03, 5.1212E+03, 5.1245E+03,
     +  5.1279E+03, 5.1313E+03, 5.1347E+03, 5.1380E+03, 5.1414E+03,
     + 5.1448E+03, 5.1482E+03, 5.1516E+03, 5.1549E+03, 5.1583E+03,
     +  5.1617E+03, 5.1651E+03, 5.1685E+03, 5.1719E+03, 5.1753E+03,
     + 5.1786E+03, 5.1820E+03, 5.1854E+03, 5.1888E+03, 5.1922E+03,
     +  5.1956E+03, 5.1990E+03, 5.2024E+03, 5.2058E+03, 5.2092E+03,
     + 5.2126E+03, 5.2160E+03, 5.2194E+03, 5.2228E+03, 5.2262E+03,
     +  5.2297E+03, 5.2331E+03, 5.2365E+03, 5.2399E+03, 5.2433E+03,
     + 5.2467E+03, 5.2501E+03, 5.2536E+03, 5.2570E+03, 5.2604E+03,
     +  5.2638E+03, 5.2672E+03, 5.2707E+03, 5.2741E+03, 5.2775E+03,
     + 5.2809E+03, 5.2844E+03, 5.2878E+03, 5.2912E+03, 5.2947E+03,
     +  5.2981E+03, 5.3015E+03, 5.3050E+03, 5.3084E+03, 5.3118E+03,
     + 5.3153E+03, 5.3187E+03, 5.3221E+03, 5.3256E+03, 5.3290E+03,
     +  5.3325E+03, 5.3359E+03, 5.3394E+03, 5.3428E+03, 5.3463E+03,
     + 5.3497E+03, 5.3532E+03, 5.3566E+03, 5.3601E+03, 5.3635E+03,
     +  5.3670E+03, 5.3704E+03, 5.3739E+03, 5.3774E+03, 5.3808E+03,
     + 5.3843E+03, 5.3878E+03, 5.3912E+03, 5.3947E+03, 5.3981E+03,
     +  5.4016E+03, 5.4051E+03, 5.4086E+03, 5.4120E+03, 5.4155E+03,
     + 5.4190E+03, 5.4225E+03, 5.4259E+03, 5.4294E+03, 5.4329E+03,
     +  5.4364E+03, 5.4398E+03, 5.4433E+03, 5.4468E+03, 5.4503E+03,
     + 5.4538E+03, 5.4573E+03, 5.4608E+03, 5.4642E+03, 5.4677E+03,
     +  5.4712E+03, 5.4747E+03, 5.4782E+03, 5.4817E+03, 5.4852E+03,
     + 5.4887E+03, 5.4922E+03, 5.4957E+03, 5.4992E+03, 5.5027E+03,
     +  5.5062E+03, 5.5097E+03, 5.5132E+03, 5.5167E+03, 5.5202E+03,
     + 5.5238E+03, 5.5273E+03, 5.5308E+03, 5.5343E+03, 5.5378E+03,
     +  5.5413E+03, 5.5448E+03, 5.5484E+03, 5.5519E+03, 5.5554E+03,
     + 5.5589E+03, 5.5624E+03, 5.5660E+03, 5.5695E+03, 5.5730E+03,
     +  5.5766E+03, 5.5801E+03, 5.5836E+03, 5.5871E+03, 5.5907E+03,
     + 5.5942E+03, 5.5977E+03, 5.6013E+03, 5.6048E+03, 5.6084E+03,
     +  5.6119E+03, 5.6154E+03, 5.6190E+03, 5.6225E+03, 5.6261E+03,
     + 5.6296E+03, 5.6332E+03, 5.6367E+03, 5.6403E+03, 5.6438E+03,
     +  5.6474E+03, 5.6509E+03, 5.6545E+03, 5.6580E+03, 5.6616E+03,
     + 5.6651E+03, 5.6687E+03, 5.6723E+03, 5.6758E+03, 5.6794E+03,
     +  5.6829E+03, 5.6865E+03, 5.6901E+03, 5.6936E+03, 5.6972E+03,
     + 5.7008E+03, 5.7044E+03, 5.7079E+03, 5.7115E+03, 5.7151E+03,
     +  5.7186E+03, 5.7222E+03, 5.7258E+03, 5.7294E+03, 5.7330E+03,
     + 5.7365E+03, 5.7401E+03, 5.7437E+03, 5.7473E+03, 5.7509E+03,
     +  5.7545E+03, 5.7581E+03, 5.7617E+03, 5.7652E+03, 5.7688E+03,
     + 5.7724E+03, 5.7760E+03, 5.7796E+03, 5.7832E+03, 5.7868E+03,
     +  5.7904E+03, 5.7940E+03, 5.7976E+03, 5.8012E+03, 5.8048E+03,
     + 5.8084E+03, 5.8120E+03, 5.8156E+03, 5.8193E+03, 5.8229E+03,
     +  5.8265E+03, 5.8301E+03, 5.8337E+03, 5.8373E+03, 5.8409E+03,
     + 5.8446E+03, 5.8482E+03, 5.8518E+03, 5.8554E+03, 5.8590E+03,
     +  5.8627E+03, 5.8663E+03, 5.8699E+03, 5.8735E+03, 5.8772E+03,
     + 5.8808E+03, 5.8844E+03, 5.8881E+03, 5.8917E+03, 5.8953E+03,
     +  5.8990E+03, 5.9026E+03, 5.9062E+03, 5.9099E+03, 5.9135E+03,
     + 5.9172E+03, 5.9208E+03, 5.9244E+03, 5.9281E+03, 5.9317E+03,
     +  5.9354E+03, 5.9390E+03, 5.9427E+03, 5.9463E+03, 5.9500E+03,
     + 5.9536E+03, 5.9573E+03, 5.9610E+03, 5.9646E+03, 5.9683E+03,
     +  5.9719E+03, 5.9756E+03, 5.9792E+03, 5.9829E+03, 5.9866E+03,
     + 5.9902E+03, 5.9939E+03, 5.9976E+03, 6.0012E+03, 6.0049E+03,
     +  6.0086E+03, 6.0123E+03, 6.0159E+03, 6.0196E+03, 6.0233E+03,
     + 6.0270E+03, 6.0306E+03, 6.0343E+03, 6.0380E+03, 6.0417E+03,
     +  6.0454E+03, 6.0491E+03, 6.0527E+03, 6.0564E+03, 6.0601E+03,
     + 6.0638E+03, 6.0675E+03, 6.0712E+03, 6.0749E+03, 6.0786E+03,
     +  6.0823E+03, 6.0860E+03, 6.0897E+03, 6.0934E+03, 6.0971E+03,
     + 6.1008E+03, 6.1045E+03, 6.1082E+03, 6.1119E+03, 6.1156E+03,
     +  6.1193E+03, 6.1230E+03, 6.1267E+03, 6.1304E+03, 6.1341E+03,
     + 6.1379E+03, 6.1416E+03, 6.1453E+03, 6.1490E+03, 6.1527E+03,
     +  6.1564E+03, 6.1602E+03, 6.1639E+03, 6.1676E+03, 6.1713E+03,
     + 6.1751E+03, 6.1788E+03, 6.1825E+03, 6.1862E+03, 6.1900E+03,
     +  6.1937E+03, 6.1974E+03, 6.2012E+03, 6.2049E+03, 6.2086E+03,
     + 6.2124E+03, 6.2161E+03, 6.2199E+03, 6.2236E+03, 6.2273E+03,
     +  6.2311E+03, 6.2348E+03, 6.2386E+03, 6.2423E+03, 6.2461E+03,
     + 6.2498E+03, 6.2536E+03, 6.2573E+03, 6.2611E+03, 6.2648E+03,
     +  6.2686E+03, 6.2724E+03, 6.2761E+03, 6.2799E+03, 6.2836E+03,
     + 6.2874E+03, 6.2912E+03, 6.2949E+03, 6.2987E+03, 6.3025E+03,
     +  6.3062E+03, 6.3100E+03, 6.3138E+03, 6.3175E+03, 6.3213E+03,
     + 6.3251E+03, 6.3289E+03, 6.3326E+03, 6.3364E+03, 6.3402E+03,
     +  6.3440E+03, 6.3477E+03, 6.3515E+03, 6.3553E+03, 6.3591E+03,
     + 6.3629E+03, 6.3667E+03, 6.3705E+03, 6.3742E+03, 6.3780E+03,
     +  6.3818E+03, 6.3856E+03, 6.3894E+03, 6.3932E+03, 6.3970E+03,
     + 6.4008E+03, 6.4046E+03, 6.4084E+03, 6.4122E+03, 6.4160E+03,
     +  6.4198E+03, 6.4236E+03, 6.4274E+03, 6.4312E+03, 6.4350E+03,
     + 6.4389E+03, 6.4427E+03, 6.4465E+03, 6.4503E+03, 6.4541E+03,
     +  6.4579E+03, 6.4617E+03, 6.4656E+03, 6.4694E+03, 6.4732E+03,
     + 6.4770E+03, 6.4809E+03, 6.4847E+03, 6.4885E+03, 6.4923E+03,
     +  6.4962E+03, 6.5000E+03, 6.5038E+03, 6.5077E+03, 6.5115E+03,
     + 6.5153E+03, 6.5192E+03, 6.5230E+03, 6.5268E+03, 6.5307E+03,
     +  6.5345E+03, 6.5384E+03, 6.5422E+03, 6.5460E+03, 6.5499E+03,
     + 6.5537E+03, 6.5576E+03, 6.5614E+03, 6.5653E+03, 6.5691E+03,
     +  6.5730E+03, 6.5768E+03, 6.5807E+03, 6.5845E+03, 6.5884E+03,
     + 6.5923E+03, 6.5961E+03, 6.6000E+03, 6.6038E+03, 6.6077E+03,
     +  6.6116E+03, 6.6154E+03, 6.6193E+03, 6.6232E+03, 6.6270E+03,
     + 6.6309E+03, 6.6348E+03, 6.6387E+03, 6.6425E+03, 6.6464E+03,
     +  6.6503E+03, 6.6542E+03, 6.6580E+03, 6.6619E+03, 6.6658E+03,
     + 6.6697E+03, 6.6736E+03, 6.6775E+03, 6.6813E+03, 6.6852E+03,
     +  6.6891E+03, 6.6930E+03, 6.6969E+03, 6.7008E+03, 6.7047E+03,
     + 6.7086E+03, 6.7125E+03, 6.7164E+03, 6.7203E+03, 6.7242E+03,
     +  6.7281E+03, 6.7320E+03, 6.7359E+03, 6.7398E+03, 6.7437E+03,
     + 6.7476E+03, 6.7515E+03, 6.7554E+03, 6.7593E+03, 6.7632E+03,
     +  6.7672E+03, 6.7711E+03, 6.7750E+03, 6.7789E+03, 6.7828E+03,
     + 6.7867E+03, 6.7907E+03, 6.7946E+03, 6.7985E+03, 6.8024E+03,
     +  6.8064E+03, 6.8103E+03, 6.8142E+03, 6.8181E+03, 6.8221E+03,
     + 6.8260E+03, 6.8299E+03, 6.8339E+03, 6.8378E+03, 6.8417E+03,
     +  6.8457E+03, 6.8496E+03, 6.8536E+03, 6.8575E+03, 6.8614E+03,
     + 6.8654E+03, 6.8693E+03, 6.8733E+03, 6.8772E+03, 6.8812E+03,
     +  6.8851E+03, 6.8891E+03, 6.8930E+03, 6.8970E+03, 6.9009E+03,
     + 6.9049E+03, 6.9088E+03, 6.9128E+03, 6.9168E+03, 6.9207E+03,
     +  6.9247E+03, 6.9286E+03, 6.9326E+03, 6.9366E+03, 6.9405E+03,
     + 6.9445E+03, 6.9485E+03, 6.9524E+03, 6.9564E+03, 6.9604E+03,
     +  6.9644E+03, 6.9683E+03, 6.9723E+03, 6.9763E+03, 6.9803E+03,
     + 6.9842E+03, 6.9882E+03, 6.9922E+03, 6.9962E+03, 7.0002E+03,
     +  7.0042E+03, 7.0082E+03, 7.0121E+03, 7.0161E+03, 7.0201E+03,
     + 7.0241E+03, 7.0281E+03, 7.0321E+03, 7.0361E+03, 7.0401E+03,
     +  7.0441E+03, 7.0481E+03, 7.0521E+03, 7.0561E+03, 7.0601E+03,
     + 7.0641E+03, 7.0681E+03, 7.0721E+03, 7.0761E+03, 7.0801E+03,
     +  7.0841E+03, 7.0882E+03, 7.0922E+03, 7.0962E+03, 7.1002E+03,
     + 7.1042E+03, 7.1082E+03, 7.1123E+03, 7.1163E+03, 7.1203E+03,
     +  7.1243E+03, 7.1283E+03, 7.1324E+03, 7.1364E+03, 7.1404E+03,
     + 7.1444E+03, 7.1485E+03, 7.1525E+03, 7.1565E+03, 7.1606E+03,
     +  7.1646E+03, 7.1686E+03, 7.1727E+03, 7.1767E+03, 7.1808E+03,
     + 7.1848E+03, 7.1888E+03, 7.1929E+03, 7.1969E+03, 7.2010E+03,
     +  7.2050E+03, 7.2091E+03, 7.2131E+03, 7.2172E+03, 7.2212E+03,
     + 7.2253E+03, 7.2293E+03, 7.2334E+03, 7.2374E+03, 7.2415E+03,
     +  7.2456E+03, 7.2496E+03, 7.2537E+03, 7.2577E+03, 7.2618E+03,
     + 7.2659E+03, 7.2699E+03, 7.2740E+03, 7.2781E+03, 7.2821E+03,
     +  7.2862E+03, 7.2903E+03, 7.2944E+03, 7.2984E+03, 7.3025E+03,
     + 7.3066E+03, 7.3107E+03, 7.3148E+03, 7.3188E+03, 7.3229E+03,
     +  7.3270E+03, 7.3311E+03, 7.3352E+03, 7.3393E+03, 7.3433E+03,
     + 7.3474E+03, 7.3515E+03, 7.3556E+03, 7.3597E+03, 7.3638E+03,
     +  7.3679E+03, 7.3720E+03, 7.3761E+03, 7.3802E+03, 7.3843E+03,
     + 7.3884E+03, 7.3925E+03, 7.3966E+03, 7.4007E+03, 7.4048E+03,
     +  7.4089E+03, 7.4130E+03, 7.4171E+03, 7.4213E+03, 7.4254E+03,
     + 7.4295E+03, 7.4336E+03, 7.4377E+03, 7.4418E+03, 7.4460E+03,
     +  7.4501E+03, 7.4542E+03, 7.4583E+03, 7.4624E+03, 7.4666E+03,
     + 7.4707E+03, 7.4748E+03, 7.4790E+03, 7.4831E+03, 7.4872E+03,
     +  7.4913E+03, 7.4955E+03, 7.4996E+03, 7.5038E+03, 7.5079E+03,
     + 7.5120E+03, 7.5162E+03, 7.5203E+03, 7.5245E+03, 7.5286E+03,
     +  7.5327E+03, 7.5369E+03, 7.5410E+03, 7.5452E+03, 7.5493E+03,
     + 7.5535E+03, 7.5576E+03, 7.5618E+03, 7.5659E+03, 7.5701E+03,
     +  7.5743E+03, 7.5784E+03, 7.5826E+03, 7.5867E+03, 7.5909E+03,
     + 7.5951E+03, 7.5992E+03, 7.6034E+03, 7.6076E+03, 7.6117E+03,
     +  7.6159E+03, 7.6201E+03, 7.6242E+03, 7.6284E+03, 7.6326E+03,
     + 7.6368E+03, 7.6409E+03, 7.6451E+03, 7.6493E+03, 7.6535E+03,
     +  7.6577E+03, 7.6618E+03, 7.6660E+03, 7.6702E+03, 7.6744E+03,
     + 7.6786E+03, 7.6828E+03, 7.6870E+03, 7.6911E+03, 7.6953E+03,
     +  7.6995E+03, 7.7037E+03, 7.7079E+03, 7.7121E+03, 7.7163E+03,
     + 7.7205E+03, 7.7247E+03, 7.7289E+03, 7.7331E+03, 7.7373E+03,
     +  7.7415E+03, 7.7457E+03, 7.7500E+03, 7.7542E+03, 7.7584E+03,
     + 7.7626E+03, 7.7668E+03, 7.7710E+03, 7.7752E+03, 7.7794E+03,
     +  7.7837E+03, 7.7879E+03, 7.7921E+03, 7.7963E+03, 7.8006E+03,
     + 7.8048E+03, 7.8090E+03, 7.8132E+03, 7.8175E+03, 7.8217E+03,
     +  7.8259E+03, 7.8301E+03, 7.8344E+03, 7.8386E+03, 7.8428E+03,
     + 7.8471E+03, 7.8513E+03, 7.8556E+03, 7.8598E+03, 7.8640E+03,
     +  7.8683E+03, 7.8725E+03, 7.8768E+03, 7.8810E+03, 7.8853E+03,
     + 7.8895E+03, 7.8938E+03, 7.8980E+03, 7.9023E+03, 7.9065E+03,
     +  7.9108E+03, 7.9150E+03, 7.9193E+03, 7.9236E+03, 7.9278E+03,
     + 7.9321E+03, 7.9363E+03, 7.9406E+03, 7.9449E+03, 7.9491E+03,
     +  7.9534E+03, 7.9577E+03, 7.9619E+03, 7.9662E+03, 7.9705E+03,
     + 7.9748E+03, 7.9790E+03, 7.9833E+03, 7.9876E+03, 7.9919E+03,
     +  7.9961E+03, 8.0004E+03, 8.0047E+03, 8.0090E+03, 8.0133E+03,
     + 8.0176E+03, 8.0218E+03, 8.0261E+03, 8.0304E+03, 8.0347E+03,
     +  8.0390E+03, 8.0433E+03, 8.0476E+03, 8.0519E+03, 8.0562E+03,
     + 8.0605E+03, 8.0648E+03, 8.0691E+03, 8.0734E+03, 8.0777E+03,
     +  8.0820E+03, 8.0863E+03, 8.0906E+03, 8.0949E+03, 8.0992E+03,
     + 8.1035E+03, 8.1078E+03, 8.1122E+03, 8.1165E+03, 8.1208E+03,
     +  8.1251E+03, 8.1294E+03, 8.1337E+03, 8.1381E+03, 8.1424E+03,
     + 8.1467E+03, 8.1510E+03, 8.1554E+03, 8.1597E+03, 8.1640E+03,
     +  8.1683E+03, 8.1727E+03, 8.1770E+03, 8.1813E+03, 8.1857E+03,
     + 8.1900E+03, 8.1943E+03, 8.1987E+03, 8.2030E+03, 8.2074E+03,
     +  8.2117E+03, 8.2160E+03, 8.2204E+03, 8.2247E+03, 8.2291E+03,
     + 8.2334E+03, 8.2378E+03, 8.2421E+03, 8.2465E+03, 8.2508E+03,
     +  8.2552E+03, 8.2595E+03, 8.2639E+03, 8.2682E+03, 8.2726E+03,
     + 8.2770E+03, 8.2813E+03, 8.2857E+03, 8.2900E+03, 8.2944E+03,
     +  8.2988E+03, 8.3031E+03, 8.3075E+03, 8.3119E+03, 8.3162E+03,
     + 8.3206E+03, 8.3250E+03, 8.3294E+03, 8.3337E+03, 8.3381E+03,
     +  8.3425E+03, 8.3469E+03, 8.3513E+03, 8.3556E+03, 8.3600E+03,
     + 8.3644E+03, 8.3688E+03, 8.3732E+03, 8.3776E+03, 8.3820E+03,
     +  8.3863E+03, 8.3907E+03, 8.3951E+03, 8.3995E+03, 8.4039E+03,
     + 8.4083E+03, 8.4127E+03, 8.4171E+03, 8.4215E+03, 8.4259E+03,
     +  8.4303E+03, 8.4347E+03, 8.4391E+03, 8.4435E+03, 8.4479E+03,
     + 8.4524E+03, 8.4568E+03, 8.4612E+03, 8.4656E+03, 8.4700E+03,
     +  8.4744E+03, 8.4788E+03, 8.4832E+03, 8.4877E+03, 8.4921E+03,
     + 8.4965E+03, 8.5009E+03, 8.5054E+03, 8.5098E+03, 8.5142E+03,
     +  8.5186E+03, 8.5231E+03, 8.5275E+03, 8.5319E+03, 8.5364E+03,
     + 8.5408E+03, 8.5452E+03, 8.5497E+03, 8.5541E+03, 8.5585E+03,
     +  8.5630E+03, 8.5674E+03, 8.5719E+03, 8.5763E+03, 8.5807E+03,
     + 8.5852E+03, 8.5896E+03, 8.5941E+03, 8.5985E+03, 8.6030E+03,
     +  8.6074E+03, 8.6119E+03, 8.6163E+03, 8.6208E+03, 8.6253E+03,
     + 8.6297E+03, 8.6342E+03, 8.6386E+03, 8.6431E+03, 8.6476E+03,
     +  8.6520E+03, 8.6565E+03, 8.6610E+03, 8.6654E+03, 8.6699E+03,
     + 8.6744E+03, 8.6788E+03, 8.6833E+03, 8.6878E+03, 8.6923E+03,
     +  8.6967E+03, 8.7012E+03, 8.7057E+03, 8.7102E+03, 8.7147E+03,
     + 8.7191E+03, 8.7236E+03, 8.7281E+03, 8.7326E+03, 8.7371E+03,
     +  8.7416E+03, 8.7461E+03, 8.7506E+03, 8.7551E+03, 8.7595E+03,
     + 8.7640E+03, 8.7685E+03, 8.7730E+03, 8.7775E+03, 8.7820E+03,
     +  8.7865E+03, 8.7910E+03, 8.7955E+03, 8.8001E+03, 8.8046E+03,
     + 8.8091E+03, 8.8136E+03, 8.8181E+03, 8.8226E+03, 8.8271E+03,
     +  8.8316E+03, 8.8361E+03, 8.8407E+03, 8.8452E+03, 8.8497E+03,
     + 8.8542E+03, 8.8587E+03, 8.8633E+03, 8.8678E+03, 8.8723E+03,
     +  8.8768E+03, 8.8814E+03, 8.8859E+03, 8.8904E+03, 8.8949E+03,
     + 8.8995E+03, 8.9040E+03, 8.9085E+03, 8.9131E+03, 8.9176E+03,
     +  8.9222E+03, 8.9267E+03, 8.9312E+03, 8.9358E+03, 8.9403E+03,
     + 8.9449E+03, 8.9494E+03, 8.9540E+03, 8.9585E+03, 8.9631E+03,
     +  8.9676E+03, 8.9722E+03, 8.9767E+03, 8.9813E+03, 8.9858E+03,
     + 8.9904E+03, 8.9949E+03, 8.9995E+03, 9.0041E+03, 9.0086E+03,
     +  9.0132E+03, 9.0178E+03, 9.0223E+03, 9.0269E+03, 9.0315E+03,
     + 9.0360E+03, 9.0406E+03, 9.0452E+03, 9.0497E+03, 9.0543E+03,
     +  9.0589E+03, 9.0635E+03, 9.0681E+03, 9.0726E+03, 9.0772E+03,
     + 9.0818E+03, 9.0864E+03, 9.0910E+03, 9.0955E+03, 9.1001E+03,
     +  9.1047E+03, 9.1093E+03, 9.1139E+03, 9.1185E+03, 9.1231E+03,
     + 9.1277E+03, 9.1323E+03, 9.1369E+03, 9.1415E+03, 9.1461E+03,
     +  9.1507E+03, 9.1553E+03, 9.1599E+03, 9.1645E+03, 9.1691E+03,
     + 9.1737E+03, 9.1783E+03, 9.1829E+03, 9.1875E+03, 9.1921E+03,
     +  9.1967E+03, 9.2014E+03, 9.2060E+03, 9.2106E+03, 9.2152E+03,
     + 9.2198E+03, 9.2245E+03, 9.2291E+03, 9.2337E+03, 9.2383E+03,
     +  9.2429E+03, 9.2476E+03, 9.2522E+03, 9.2568E+03, 9.2615E+03,
     + 9.2661E+03, 9.2707E+03, 9.2754E+03, 9.2800E+03, 9.2846E+03,
     +  9.2893E+03, 9.2939E+03, 9.2985E+03, 9.3032E+03, 9.3078E+03,
     + 9.3125E+03, 9.3171E+03, 9.3218E+03, 9.3264E+03, 9.3311E+03,
     +  9.3357E+03, 9.3404E+03, 9.3450E+03, 9.3497E+03, 9.3543E+03,
     + 9.3590E+03, 9.3636E+03, 9.3683E+03, 9.3730E+03, 9.3776E+03,
     +  9.3823E+03, 9.3869E+03, 9.3916E+03, 9.3963E+03, 9.4009E+03,
     + 9.4056E+03, 9.4103E+03, 9.4150E+03, 9.4196E+03, 9.4243E+03,
     +  9.4290E+03, 9.4337E+03, 9.4383E+03, 9.4430E+03, 9.4477E+03,
     + 9.4524E+03, 9.4571E+03, 9.4617E+03, 9.4664E+03, 9.4711E+03,
     +  9.4758E+03, 9.4805E+03, 9.4852E+03, 9.4899E+03, 9.4946E+03,
     + 9.4993E+03, 9.5040E+03, 9.5086E+03, 9.5133E+03, 9.5180E+03,
     +  9.5227E+03, 9.5274E+03, 9.5321E+03, 9.5368E+03, 9.5416E+03,
     + 9.5463E+03, 9.5510E+03, 9.5557E+03, 9.5604E+03, 9.5651E+03,
     +  9.5698E+03, 9.5745E+03, 9.5792E+03, 9.5840E+03, 9.5887E+03,
     + 9.5934E+03, 9.5981E+03, 9.6028E+03, 9.6076E+03, 9.6123E+03,
     +  9.6170E+03, 9.6217E+03, 9.6265E+03, 9.6312E+03, 9.6359E+03,
     + 9.6406E+03, 9.6454E+03, 9.6501E+03, 9.6548E+03, 9.6596E+03,
     +  9.6643E+03, 9.6691E+03, 9.6738E+03, 9.6785E+03, 9.6833E+03,
     + 9.6880E+03, 9.6928E+03, 9.6975E+03, 9.7023E+03, 9.7070E+03,
     +  9.7118E+03, 9.7165E+03, 9.7213E+03, 9.7260E+03, 9.7308E+03,
     + 9.7355E+03, 9.7403E+03, 9.7450E+03, 9.7498E+03, 9.7546E+03,
     +  9.7593E+03, 9.7641E+03, 9.7688E+03, 9.7736E+03, 9.7784E+03,
     + 9.7831E+03, 9.7879E+03, 9.7927E+03, 9.7975E+03, 9.8022E+03,
     +  9.8070E+03, 9.8118E+03, 9.8166E+03, 9.8213E+03, 9.8261E+03,
     + 9.8309E+03, 9.8357E+03, 9.8405E+03, 9.8452E+03, 9.8500E+03,
     +  9.8548E+03, 9.8596E+03, 9.8644E+03, 9.8692E+03, 9.8740E+03,
     + 9.8788E+03, 9.8836E+03, 9.8884E+03, 9.8932E+03, 9.8980E+03,
     +  9.9028E+03, 9.9076E+03, 9.9124E+03, 9.9172E+03, 9.9220E+03,
     + 9.9268E+03, 9.9316E+03, 9.9364E+03, 9.9412E+03, 9.9460E+03,
     +  9.9508E+03, 9.9556E+03, 9.9604E+03, 9.9653E+03, 9.9701E+03,
     + 9.9749E+03, 9.9797E+03, 9.9845E+03, 9.9894E+03, 9.9942E+03,
     +  9.9990E+03, 1.0004E+04, 1.0009E+04, 1.0013E+04, 1.0018E+04,
     + 1.0023E+04, 1.0028E+04, 1.0033E+04, 1.0038E+04, 1.0042E+04,
     +  1.0047E+04, 1.0052E+04, 1.0057E+04, 1.0062E+04, 1.0067E+04,
     + 1.0072E+04, 1.0076E+04, 1.0081E+04, 1.0086E+04, 1.0091E+04,
     +  1.0096E+04, 1.0101E+04, 1.0105E+04, 1.0110E+04, 1.0115E+04,
     + 1.0120E+04, 1.0125E+04, 1.0130E+04, 1.0135E+04, 1.0139E+04,
     +  1.0144E+04, 1.0149E+04, 1.0154E+04, 1.0159E+04, 1.0164E+04,
     + 1.0169E+04, 1.0174E+04, 1.0178E+04, 1.0183E+04, 1.0188E+04,
     +  1.0193E+04, 1.0198E+04, 1.0203E+04, 1.0208E+04, 1.0213E+04,
     + 1.0217E+04, 1.0222E+04, 1.0227E+04, 1.0232E+04, 1.0237E+04,
     +  1.0242E+04, 1.0247E+04, 1.0252E+04, 1.0256E+04, 1.0261E+04,
     + 1.0266E+04, 1.0271E+04, 1.0276E+04, 1.0281E+04, 1.0286E+04,
     +  1.0291E+04, 1.0296E+04, 1.0301E+04, 1.0305E+04, 1.0310E+04,
     + 1.0315E+04, 1.0320E+04, 1.0325E+04, 1.0330E+04, 1.0335E+04,
     +  1.0340E+04, 1.0345E+04, 1.0350E+04, 1.0355E+04, 1.0359E+04,
     + 1.0364E+04, 1.0369E+04, 1.0374E+04, 1.0379E+04, 1.0384E+04,
     +  1.0389E+04, 1.0394E+04, 1.0399E+04, 1.0404E+04, 1.0409E+04,
     + 1.0414E+04, 1.0419E+04, 1.0423E+04, 1.0428E+04, 1.0433E+04,
     +  1.0438E+04, 1.0443E+04, 1.0448E+04, 1.0453E+04, 1.0458E+04,
     + 1.0463E+04, 1.0468E+04, 1.0473E+04, 1.0478E+04, 1.0483E+04,
     +  1.0488E+04, 1.0493E+04, 1.0498E+04, 1.0503E+04, 1.0508E+04,
     + 1.0512E+04, 1.0517E+04, 1.0522E+04, 1.0527E+04, 1.0532E+04,
     +  1.0537E+04, 1.0542E+04, 1.0547E+04, 1.0552E+04, 1.0557E+04,
     + 1.0562E+04, 1.0567E+04, 1.0572E+04, 1.0577E+04, 1.0582E+04,
     +  1.0587E+04, 1.0592E+04, 1.0597E+04, 1.0602E+04, 1.0607E+04,
     + 1.0612E+04, 1.0617E+04, 1.0622E+04, 1.0627E+04, 1.0632E+04,
     +  1.0637E+04, 1.0642E+04, 1.0647E+04, 1.0652E+04, 1.0657E+04,
     + 1.0662E+04, 1.0667E+04, 1.0672E+04, 1.0677E+04, 1.0682E+04,
     +  1.0687E+04, 1.0692E+04, 1.0697E+04, 1.0702E+04, 1.0707E+04,
     + 1.0712E+04, 1.0717E+04, 1.0722E+04, 1.0727E+04, 1.0732E+04,
     +  1.0737E+04, 1.0742E+04, 1.0747E+04, 1.0752E+04, 1.0757E+04,
     + 1.0762E+04, 1.0767E+04, 1.0772E+04, 1.0777E+04, 1.0782E+04,
     +  1.0787E+04, 1.0792E+04, 1.0797E+04, 1.0802E+04, 1.0807E+04,
     + 1.0812E+04, 1.0817E+04, 1.0822E+04, 1.0827E+04, 1.0832E+04,
     +  1.0837E+04, 1.0842E+04, 1.0847E+04, 1.0852E+04, 1.0857E+04,
     + 1.0862E+04, 1.0867E+04, 1.0873E+04, 1.0878E+04, 1.0883E+04,
     +  1.0888E+04, 1.0893E+04, 1.0898E+04, 1.0903E+04, 1.0908E+04,
     + 1.0913E+04, 1.0918E+04, 1.0923E+04, 1.0928E+04, 1.0933E+04,
     +  1.0938E+04, 1.0943E+04, 1.0948E+04, 1.0953E+04, 1.0958E+04,
     + 1.0964E+04, 1.0969E+04, 1.0974E+04, 1.0979E+04, 1.0984E+04,
     +  1.0989E+04, 1.0994E+04, 1.0999E+04, 1.1004E+04, 1.1009E+04,
     + 1.1014E+04, 1.1019E+04, 1.1024E+04, 1.1030E+04, 1.1035E+04,
     +  1.1040E+04, 1.1045E+04, 1.1050E+04, 1.1055E+04, 1.1060E+04,
     + 1.1065E+04, 1.1070E+04, 1.1075E+04, 1.1080E+04, 1.1086E+04,
     +  1.1091E+04, 1.1096E+04, 1.1101E+04, 1.1106E+04, 1.1111E+04,
     + 1.1116E+04, 1.1121E+04, 1.1126E+04, 1.1131E+04, 1.1137E+04,
     +  1.1142E+04, 1.1147E+04, 1.1152E+04, 1.1157E+04, 1.1162E+04,
     + 1.1167E+04, 1.1172E+04, 1.1178E+04, 1.1183E+04, 1.1188E+04,
     +  1.1193E+04, 1.1198E+04, 1.1203E+04, 1.1208E+04, 1.1213E+04,
     + 1.1219E+04, 1.1224E+04, 1.1229E+04, 1.1234E+04, 1.1239E+04,
     +  1.1244E+04, 1.1249E+04, 1.1254E+04, 1.1260E+04, 1.1265E+04,
     + 1.1270E+04, 1.1275E+04, 1.1280E+04, 1.1285E+04, 1.1290E+04,
     +  1.1296E+04, 1.1301E+04, 1.1306E+04, 1.1311E+04, 1.1316E+04,
     + 1.1321E+04, 1.1327E+04, 1.1332E+04, 1.1337E+04, 1.1342E+04,
     +  1.1347E+04, 1.1352E+04, 1.1358E+04, 1.1363E+04, 1.1368E+04,
     + 1.1373E+04, 1.1378E+04, 1.1383E+04, 1.1389E+04, 1.1394E+04,
     +  1.1399E+04, 1.1404E+04, 1.1409E+04, 1.1414E+04, 1.1420E+04,
     + 1.1425E+04, 1.1430E+04, 1.1435E+04, 1.1440E+04, 1.1445E+04,
     +  1.1451E+04, 1.1456E+04, 1.1461E+04, 1.1466E+04, 1.1471E+04,
     + 1.1477E+04, 1.1482E+04, 1.1487E+04, 1.1492E+04, 1.1497E+04,
     +  1.1503E+04, 1.1508E+04, 1.1513E+04, 1.1518E+04, 1.1523E+04,
     + 1.1529E+04, 1.1534E+04, 1.1539E+04, 1.1544E+04, 1.1549E+04,
     +  1.1555E+04, 1.1560E+04, 1.1565E+04, 1.1570E+04, 1.1576E+04,
     + 1.1581E+04, 1.1586E+04, 1.1591E+04, 1.1596E+04, 1.1602E+04,
     +  1.1607E+04, 1.1612E+04, 1.1617E+04, 1.1623E+04, 1.1628E+04,
     + 1.1633E+04, 1.1638E+04, 1.1643E+04, 1.1649E+04, 1.1654E+04,
     +  1.1659E+04, 1.1664E+04, 1.1670E+04, 1.1675E+04, 1.1680E+04,
     + 1.1685E+04, 1.1691E+04, 1.1696E+04, 1.1701E+04, 1.1706E+04,
     +  1.1712E+04, 1.1717E+04, 1.1722E+04, 1.1727E+04, 1.1733E+04,
     + 1.1738E+04, 1.1743E+04, 1.1748E+04, 1.1754E+04, 1.1759E+04,
     +  1.1764E+04, 1.1769E+04, 1.1775E+04, 1.1780E+04, 1.1785E+04,
     + 1.1790E+04, 1.1796E+04, 1.1801E+04, 1.1806E+04, 1.1812E+04,
     +  1.1817E+04, 1.1822E+04, 1.1827E+04, 1.1833E+04, 1.1838E+04,
     + 1.1843E+04, 1.1848E+04, 1.1854E+04, 1.1859E+04, 1.1864E+04,
     +  1.1870E+04, 1.1875E+04, 1.1880E+04, 1.1885E+04, 1.1891E+04,
     + 1.1896E+04, 1.1901E+04, 1.1907E+04, 1.1912E+04, 1.1917E+04,
     +  1.1923E+04, 1.1928E+04, 1.1933E+04, 1.1938E+04, 1.1944E+04,
     + 1.1949E+04, 1.1954E+04, 1.1960E+04, 1.1965E+04, 1.1970E+04,
     +  1.1976E+04, 1.1981E+04, 1.1986E+04, 1.1992E+04, 1.1997E+04,
     + 1.2002E+04, 1.2008E+04, 1.2013E+04, 1.2018E+04, 1.2023E+04,
     +  1.2029E+04, 1.2034E+04, 1.2039E+04, 1.2045E+04, 1.2050E+04,
     + 1.2055E+04, 1.2061E+04, 1.2066E+04, 1.2071E+04, 1.2077E+04,
     +  1.2082E+04, 1.2087E+04, 1.2093E+04, 1.2098E+04, 1.2103E+04,
     + 1.2109E+04, 1.2114E+04, 1.2119E+04, 1.2125E+04, 1.2130E+04,
     +  1.2136E+04, 1.2141E+04, 1.2146E+04, 1.2152E+04, 1.2157E+04,
     + 1.2162E+04, 1.2168E+04, 1.2173E+04, 1.2178E+04, 1.2184E+04,
     +  1.2189E+04, 1.2194E+04, 1.2200E+04, 1.2205E+04, 1.2211E+04,
     + 1.2216E+04, 1.2221E+04, 1.2227E+04, 1.2232E+04, 1.2237E+04,
     +  1.2243E+04, 1.2248E+04, 1.2254E+04, 1.2259E+04, 1.2264E+04,
     + 1.2270E+04, 1.2275E+04, 1.2280E+04, 1.2286E+04, 1.2291E+04,
     +  1.2297E+04, 1.2302E+04, 1.2307E+04, 1.2313E+04, 1.2318E+04,
     + 1.2324E+04, 1.2329E+04, 1.2334E+04, 1.2340E+04, 1.2345E+04,
     +  1.2351E+04, 1.2356E+04, 1.2361E+04, 1.2367E+04, 1.2372E+04,
     + 1.2378E+04, 1.2383E+04, 1.2388E+04, 1.2394E+04, 1.2399E+04,
     +  1.2405E+04, 1.2410E+04, 1.2415E+04, 1.2421E+04, 1.2426E+04,
     + 1.2432E+04, 1.2437E+04, 1.2442E+04, 1.2448E+04, 1.2453E+04,
     +  1.2459E+04, 1.2464E+04, 1.2470E+04, 1.2475E+04, 1.2480E+04,
     + 1.2486E+04, 1.2491E+04, 1.2497E+04, 1.2502E+04, 1.2508E+04,
     +  1.2513E+04, 1.2519E+04, 1.2524E+04, 1.2529E+04, 1.2535E+04,
     + 1.2540E+04, 1.2546E+04, 1.2551E+04, 1.2557E+04, 1.2562E+04,
     +  1.2568E+04, 1.2573E+04, 1.2578E+04, 1.2584E+04, 1.2589E+04,
     + 1.2595E+04, 1.2600E+04, 1.2606E+04, 1.2611E+04, 1.2617E+04,
     +  1.2622E+04, 1.2628E+04, 1.2633E+04, 1.2638E+04, 1.2644E+04,
     + 1.2649E+04, 1.2655E+04, 1.2660E+04, 1.2666E+04, 1.2671E+04,
     +  1.2677E+04, 1.2682E+04, 1.2688E+04, 1.2693E+04, 1.2699E+04,
     + 1.2704E+04, 1.2710E+04, 1.2715E+04, 1.2721E+04, 1.2726E+04,
     +  1.2732E+04, 1.2737E+04, 1.2743E+04, 1.2748E+04, 1.2754E+04,
     + 1.2759E+04, 1.2765E+04, 1.2770E+04, 1.2775E+04, 1.2781E+04,
     +  1.2786E+04, 1.2792E+04, 1.2797E+04, 1.2803E+04, 1.2809E+04,
     + 1.2814E+04, 1.2820E+04, 1.2825E+04, 1.2831E+04, 1.2836E+04,
     +  1.2842E+04, 1.2847E+04, 1.2853E+04, 1.2858E+04, 1.2864E+04,
     + 1.2869E+04, 1.2875E+04, 1.2880E+04, 1.2886E+04, 1.2891E+04,
     +  1.2897E+04, 1.2902E+04, 1.2908E+04, 1.2913E+04, 1.2919E+04,
     + 1.2924E+04, 1.2930E+04, 1.2935E+04, 1.2941E+04/
c...        --        27
      data (QofT( 4,J),J=1,2408)/ 6.0819E+00, 7.2235E+00, 9.1203E+00,
     +  1.1214E+01, 1.3368E+01, 1.5549E+01, 1.7744E+01, 1.9947E+01,
     + 2.2156E+01, 2.4368E+01, 2.6582E+01, 2.8799E+01, 3.1017E+01,
     +  3.3236E+01, 3.5456E+01, 3.7676E+01, 3.9897E+01, 4.2119E+01,
     + 4.4341E+01, 4.6564E+01, 5.1010E+01, 5.5457E+01, 5.9904E+01,
     +  6.4352E+01, 6.8801E+01, 7.3250E+01, 7.7699E+01, 8.2149E+01,
     + 8.6599E+01, 9.1049E+01, 9.5499E+01, 9.9949E+01, 1.0440E+02,
     +  1.0885E+02, 1.1330E+02, 1.1775E+02, 1.2220E+02, 1.2665E+02,
     + 1.3111E+02, 1.3556E+02, 1.4001E+02, 1.4446E+02, 1.4891E+02,
     +  1.5336E+02, 1.5781E+02, 1.6227E+02, 1.6672E+02, 1.7117E+02,
     + 1.7562E+02, 1.8007E+02, 1.8452E+02, 1.8898E+02, 1.9343E+02,
     +  1.9788E+02, 2.0233E+02, 2.0679E+02, 2.1124E+02, 2.1569E+02,
     + 2.2014E+02, 2.2459E+02, 2.2905E+02, 2.3350E+02, 2.3795E+02,
     +  2.4241E+02, 2.4686E+02, 2.5131E+02, 2.5576E+02, 2.6022E+02,
     + 2.6467E+02, 2.6912E+02, 2.7358E+02, 2.7803E+02, 2.8248E+02,
     +  2.8693E+02, 2.9139E+02, 2.9584E+02, 3.0029E+02, 3.0475E+02,
     + 3.0920E+02, 3.1365E+02, 3.1811E+02, 3.2256E+02, 3.2701E+02,
     +  3.3147E+02, 3.3592E+02, 3.4038E+02, 3.4483E+02, 3.4928E+02,
     + 3.5374E+02, 3.5819E+02, 3.6264E+02, 3.6710E+02, 3.7155E+02,
     +  3.7601E+02, 3.8046E+02, 3.8491E+02, 3.8937E+02, 3.9382E+02,
     + 3.9828E+02, 4.0273E+02, 4.0719E+02, 4.1164E+02, 4.1609E+02,
     +  4.2055E+02, 4.2500E+02, 4.2946E+02, 4.3391E+02, 4.3837E+02,
     + 4.4282E+02, 4.4728E+02, 4.5173E+02, 4.5619E+02, 4.6064E+02,
     +  4.6510E+02, 4.6955E+02, 4.7401E+02, 4.7846E+02, 4.8292E+02,
     + 4.8737E+02, 4.9183E+02, 4.9628E+02, 5.0074E+02, 5.0519E+02,
     +  5.0965E+02, 5.1410E+02, 5.1856E+02, 5.2301E+02, 5.2747E+02,
     + 5.3192E+02, 5.3638E+02, 5.4084E+02, 5.4529E+02, 5.4975E+02,
     +  5.5420E+02, 5.5866E+02, 5.6312E+02, 5.6757E+02, 5.7203E+02,
     + 5.7648E+02, 5.8094E+02, 5.8540E+02, 5.8985E+02, 5.9431E+02,
     +  5.9877E+02, 6.0322E+02, 6.0768E+02, 6.1214E+02, 6.1660E+02,
     + 6.2105E+02, 6.2551E+02, 6.2997E+02, 6.3443E+02, 6.3888E+02,
     +  6.4334E+02, 6.4780E+02, 6.5226E+02, 6.5671E+02, 6.6117E+02,
     + 6.6563E+02, 6.7009E+02, 6.7455E+02, 6.7901E+02, 6.8347E+02,
     +  6.8793E+02, 6.9239E+02, 6.9685E+02, 7.0131E+02, 7.0577E+02,
     + 7.1023E+02, 7.1469E+02, 7.1915E+02, 7.2361E+02, 7.2807E+02,
     +  7.3253E+02, 7.3699E+02, 7.4146E+02, 7.4592E+02, 7.5038E+02,
     + 7.5484E+02, 7.5931E+02, 7.6377E+02, 7.6823E+02, 7.7270E+02,
     +  7.7716E+02, 7.8163E+02, 7.8609E+02, 7.9056E+02, 7.9503E+02,
     + 7.9949E+02, 8.0396E+02, 8.0843E+02, 8.1289E+02, 8.1736E+02,
     +  8.2183E+02, 8.2630E+02, 8.3077E+02, 8.3524E+02, 8.3971E+02,
     + 8.4418E+02, 8.4865E+02, 8.5312E+02, 8.5760E+02, 8.6207E+02,
     +  8.6654E+02, 8.7102E+02, 8.7549E+02, 8.7997E+02, 8.8445E+02,
     + 8.8892E+02, 8.9340E+02, 8.9788E+02, 9.0236E+02, 9.0684E+02,
     +  9.1132E+02, 9.1580E+02, 9.2028E+02, 9.2477E+02, 9.2925E+02,
     + 9.3373E+02, 9.3822E+02, 9.4271E+02, 9.4719E+02, 9.5168E+02,
     +  9.5617E+02, 9.6066E+02, 9.6515E+02, 9.6964E+02, 9.7413E+02,
     + 9.7863E+02, 9.8312E+02, 9.8762E+02, 9.9211E+02, 9.9661E+02,
     +  1.0011E+03, 1.0056E+03, 1.0101E+03, 1.0146E+03, 1.0191E+03,
     + 1.0236E+03, 1.0281E+03, 1.0326E+03, 1.0371E+03, 1.0416E+03,
     +  1.0462E+03, 1.0507E+03, 1.0552E+03, 1.0597E+03, 1.0642E+03,
     + 1.0687E+03, 1.0732E+03, 1.0778E+03, 1.0823E+03, 1.0868E+03,
     +  1.0913E+03, 1.0959E+03, 1.1004E+03, 1.1049E+03, 1.1094E+03,
     + 1.1140E+03, 1.1185E+03, 1.1230E+03, 1.1276E+03, 1.1321E+03,
     +  1.1367E+03, 1.1412E+03, 1.1457E+03, 1.1503E+03, 1.1548E+03,
     + 1.1594E+03, 1.1639E+03, 1.1685E+03, 1.1730E+03, 1.1776E+03,
     +  1.1821E+03, 1.1867E+03, 1.1913E+03, 1.1958E+03, 1.2004E+03,
     + 1.2050E+03, 1.2095E+03, 1.2141E+03, 1.2187E+03, 1.2232E+03,
     +  1.2278E+03, 1.2324E+03, 1.2370E+03, 1.2416E+03, 1.2461E+03,
     + 1.2507E+03, 1.2553E+03, 1.2599E+03, 1.2645E+03, 1.2691E+03,
     +  1.2737E+03, 1.2783E+03, 1.2829E+03, 1.2875E+03, 1.2921E+03,
     + 1.2967E+03, 1.3013E+03, 1.3059E+03, 1.3105E+03, 1.3152E+03,
     +  1.3198E+03, 1.3244E+03, 1.3290E+03, 1.3337E+03, 1.3383E+03,
     + 1.3429E+03, 1.3476E+03, 1.3522E+03, 1.3568E+03, 1.3615E+03,
     +  1.3661E+03, 1.3708E+03, 1.3754E+03, 1.3801E+03, 1.3847E+03,
     + 1.3894E+03, 1.3940E+03, 1.3987E+03, 1.4034E+03, 1.4080E+03,
     +  1.4127E+03, 1.4174E+03, 1.4221E+03, 1.4267E+03, 1.4314E+03,
     + 1.4361E+03, 1.4408E+03, 1.4455E+03, 1.4502E+03, 1.4549E+03,
     +  1.4596E+03, 1.4643E+03, 1.4690E+03, 1.4737E+03, 1.4784E+03,
     + 1.4831E+03, 1.4879E+03, 1.4926E+03, 1.4973E+03, 1.5020E+03,
     +  1.5068E+03, 1.5115E+03, 1.5162E+03, 1.5210E+03, 1.5257E+03,
     + 1.5305E+03, 1.5352E+03, 1.5400E+03, 1.5447E+03, 1.5495E+03,
     +  1.5542E+03, 1.5590E+03, 1.5638E+03, 1.5685E+03, 1.5733E+03,
     + 1.5781E+03, 1.5829E+03, 1.5877E+03, 1.5925E+03, 1.5973E+03,
     +  1.6020E+03, 1.6068E+03, 1.6117E+03, 1.6165E+03, 1.6213E+03,
     + 1.6261E+03, 1.6309E+03, 1.6357E+03, 1.6405E+03, 1.6454E+03,
     +  1.6502E+03, 1.6550E+03, 1.6599E+03, 1.6647E+03, 1.6696E+03,
     + 1.6744E+03, 1.6793E+03, 1.6841E+03, 1.6890E+03, 1.6938E+03,
     +  1.6987E+03, 1.7036E+03, 1.7084E+03, 1.7133E+03, 1.7182E+03,
     + 1.7231E+03, 1.7280E+03, 1.7329E+03, 1.7378E+03, 1.7427E+03,
     +  1.7476E+03, 1.7525E+03, 1.7574E+03, 1.7623E+03, 1.7672E+03,
     + 1.7722E+03, 1.7771E+03, 1.7820E+03, 1.7870E+03, 1.7919E+03,
     +  1.7968E+03, 1.8018E+03, 1.8067E+03, 1.8117E+03, 1.8167E+03,
     + 1.8216E+03, 1.8266E+03, 1.8316E+03, 1.8365E+03, 1.8415E+03,
     +  1.8465E+03, 1.8515E+03, 1.8565E+03, 1.8615E+03, 1.8665E+03,
     + 1.8715E+03, 1.8765E+03, 1.8815E+03, 1.8865E+03, 1.8915E+03,
     +  1.8966E+03, 1.9016E+03, 1.9066E+03, 1.9117E+03, 1.9167E+03,
     + 1.9217E+03, 1.9268E+03, 1.9318E+03, 1.9369E+03, 1.9420E+03,
     +  1.9470E+03, 1.9521E+03, 1.9572E+03, 1.9623E+03, 1.9674E+03,
     + 1.9724E+03, 1.9775E+03, 1.9826E+03, 1.9877E+03, 1.9928E+03,
     +  1.9979E+03, 2.0031E+03, 2.0082E+03, 2.0133E+03, 2.0184E+03,
     + 2.0236E+03, 2.0287E+03, 2.0338E+03, 2.0390E+03, 2.0441E+03,
     +  2.0493E+03, 2.0545E+03, 2.0596E+03, 2.0648E+03, 2.0700E+03,
     + 2.0751E+03, 2.0803E+03, 2.0855E+03, 2.0907E+03, 2.0959E+03,
     +  2.1011E+03, 2.1063E+03, 2.1115E+03, 2.1167E+03, 2.1219E+03,
     + 2.1272E+03, 2.1324E+03, 2.1376E+03, 2.1429E+03, 2.1481E+03,
     +  2.1533E+03, 2.1586E+03, 2.1639E+03, 2.1691E+03, 2.1744E+03,
     + 2.1796E+03, 2.1849E+03, 2.1902E+03, 2.1955E+03, 2.2008E+03,
     +  2.2061E+03, 2.2114E+03, 2.2167E+03, 2.2220E+03, 2.2273E+03,
     + 2.2326E+03, 2.2379E+03, 2.2432E+03, 2.2486E+03, 2.2539E+03,
     +  2.2593E+03, 2.2646E+03, 2.2699E+03, 2.2753E+03, 2.2807E+03,
     + 2.2860E+03, 2.2914E+03, 2.2968E+03, 2.3021E+03, 2.3075E+03,
     +  2.3129E+03, 2.3183E+03, 2.3237E+03, 2.3291E+03, 2.3345E+03,
     + 2.3399E+03, 2.3454E+03, 2.3508E+03, 2.3562E+03, 2.3616E+03,
     +  2.3671E+03, 2.3725E+03, 2.3780E+03, 2.3834E+03, 2.3889E+03,
     + 2.3943E+03, 2.3998E+03, 2.4053E+03, 2.4108E+03, 2.4162E+03,
     +  2.4217E+03, 2.4272E+03, 2.4327E+03, 2.4382E+03, 2.4437E+03,
     + 2.4492E+03, 2.4548E+03, 2.4603E+03, 2.4658E+03, 2.4713E+03,
     +  2.4769E+03, 2.4824E+03, 2.4880E+03, 2.4935E+03, 2.4991E+03,
     + 2.5046E+03, 2.5102E+03, 2.5158E+03, 2.5214E+03, 2.5269E+03,
     +  2.5325E+03, 2.5381E+03, 2.5437E+03, 2.5493E+03, 2.5549E+03,
     + 2.5606E+03, 2.5662E+03, 2.5718E+03, 2.5774E+03, 2.5831E+03,
     +  2.5887E+03, 2.5943E+03, 2.6000E+03, 2.6056E+03, 2.6113E+03,
     + 2.6170E+03, 2.6226E+03, 2.6283E+03, 2.6340E+03, 2.6397E+03,
     +  2.6454E+03, 2.6511E+03, 2.6568E+03, 2.6625E+03, 2.6682E+03,
     + 2.6739E+03, 2.6796E+03, 2.6854E+03, 2.6911E+03, 2.6968E+03,
     +  2.7026E+03, 2.7083E+03, 2.7141E+03, 2.7198E+03, 2.7256E+03,
     + 2.7314E+03, 2.7372E+03, 2.7429E+03, 2.7487E+03, 2.7545E+03,
     +  2.7603E+03, 2.7661E+03, 2.7719E+03, 2.7777E+03, 2.7835E+03,
     + 2.7894E+03, 2.7952E+03, 2.8010E+03, 2.8069E+03, 2.8127E+03,
     +  2.8186E+03, 2.8244E+03, 2.8303E+03, 2.8361E+03, 2.8420E+03,
     + 2.8479E+03, 2.8538E+03, 2.8597E+03, 2.8656E+03, 2.8715E+03,
     +  2.8774E+03, 2.8833E+03, 2.8892E+03, 2.8951E+03, 2.9010E+03,
     + 2.9070E+03, 2.9129E+03, 2.9188E+03, 2.9248E+03, 2.9307E+03,
     +  2.9367E+03, 2.9426E+03, 2.9486E+03, 2.9546E+03, 2.9606E+03,
     + 2.9666E+03, 2.9725E+03, 2.9785E+03, 2.9845E+03, 2.9905E+03,
     +  2.9966E+03, 3.0026E+03, 3.0086E+03, 3.0146E+03, 3.0207E+03,
     + 3.0267E+03, 3.0327E+03, 3.0388E+03, 3.0448E+03, 3.0509E+03,
     +  3.0570E+03, 3.0630E+03, 3.0691E+03, 3.0752E+03, 3.0813E+03,
     + 3.0874E+03, 3.0935E+03, 3.0996E+03, 3.1057E+03, 3.1118E+03,
     +  3.1179E+03, 3.1241E+03, 3.1302E+03, 3.1363E+03, 3.1425E+03,
     + 3.1486E+03, 3.1548E+03, 3.1609E+03, 3.1671E+03, 3.1733E+03,
     +  3.1794E+03, 3.1856E+03, 3.1918E+03, 3.1980E+03, 3.2042E+03,
     + 3.2104E+03, 3.2166E+03, 3.2228E+03, 3.2290E+03, 3.2353E+03,
     +  3.2415E+03, 3.2477E+03, 3.2540E+03, 3.2602E+03, 3.2665E+03,
     + 3.2727E+03, 3.2790E+03, 3.2853E+03, 3.2916E+03, 3.2978E+03,
     +  3.3041E+03, 3.3104E+03, 3.3167E+03, 3.3230E+03, 3.3293E+03,
     + 3.3357E+03, 3.3420E+03, 3.3483E+03, 3.3546E+03, 3.3610E+03,
     +  3.3673E+03, 3.3737E+03, 3.3800E+03, 3.3864E+03, 3.3927E+03,
     + 3.3991E+03, 3.4055E+03, 3.4119E+03, 3.4183E+03, 3.4247E+03,
     +  3.4311E+03, 3.4375E+03, 3.4439E+03, 3.4503E+03, 3.4567E+03,
     + 3.4631E+03, 3.4696E+03, 3.4760E+03, 3.4825E+03, 3.4889E+03,
     +  3.4954E+03, 3.5018E+03, 3.5083E+03, 3.5148E+03, 3.5213E+03,
     + 3.5277E+03, 3.5342E+03, 3.5407E+03, 3.5472E+03, 3.5537E+03,
     +  3.5602E+03, 3.5668E+03, 3.5733E+03, 3.5798E+03, 3.5864E+03,
     + 3.5929E+03, 3.5994E+03, 3.6060E+03, 3.6126E+03, 3.6191E+03,
     +  3.6257E+03, 3.6323E+03, 3.6388E+03, 3.6454E+03, 3.6520E+03,
     + 3.6586E+03, 3.6652E+03, 3.6718E+03, 3.6785E+03, 3.6851E+03,
     +  3.6917E+03, 3.6983E+03, 3.7050E+03, 3.7116E+03, 3.7183E+03,
     + 3.7249E+03, 3.7316E+03, 3.7383E+03, 3.7449E+03, 3.7516E+03,
     +  3.7583E+03, 3.7650E+03, 3.7717E+03, 3.7784E+03, 3.7851E+03,
     + 3.7918E+03, 3.7985E+03, 3.8052E+03, 3.8120E+03, 3.8187E+03,
     +  3.8255E+03, 3.8322E+03, 3.8390E+03, 3.8457E+03, 3.8525E+03,
     + 3.8593E+03, 3.8660E+03, 3.8728E+03, 3.8796E+03, 3.8864E+03,
     +  3.8932E+03, 3.9000E+03, 3.9068E+03, 3.9136E+03, 3.9205E+03,
     + 3.9273E+03, 3.9341E+03, 3.9410E+03, 3.9478E+03, 3.9547E+03,
     +  3.9615E+03, 3.9684E+03, 3.9753E+03, 3.9821E+03, 3.9890E+03,
     + 3.9959E+03, 4.0028E+03, 4.0097E+03, 4.0166E+03, 4.0235E+03,
     +  4.0304E+03, 4.0373E+03, 4.0443E+03, 4.0512E+03, 4.0581E+03,
     + 4.0651E+03, 4.0720E+03, 4.0790E+03, 4.0860E+03, 4.0929E+03,
     +  4.0999E+03, 4.1069E+03, 4.1139E+03, 4.1209E+03, 4.1279E+03,
     + 4.1349E+03, 4.1419E+03, 4.1489E+03, 4.1559E+03, 4.1629E+03,
     +  4.1700E+03, 4.1770E+03, 4.1840E+03, 4.1911E+03, 4.1982E+03,
     + 4.2052E+03, 4.2123E+03, 4.2194E+03, 4.2264E+03, 4.2335E+03,
     +  4.2406E+03, 4.2477E+03, 4.2548E+03, 4.2619E+03, 4.2690E+03,
     + 4.2762E+03, 4.2833E+03, 4.2904E+03, 4.2976E+03, 4.3047E+03,
     +  4.3119E+03, 4.3190E+03, 4.3262E+03, 4.3333E+03, 4.3405E+03,
     + 4.3477E+03, 4.3549E+03, 4.3621E+03, 4.3693E+03, 4.3765E+03,
     +  4.3837E+03, 4.3909E+03, 4.3981E+03, 4.4053E+03, 4.4126E+03,
     + 4.4198E+03, 4.4271E+03, 4.4343E+03, 4.4416E+03, 4.4488E+03,
     +  4.4561E+03, 4.4634E+03, 4.4706E+03, 4.4779E+03, 4.4852E+03,
     + 4.4925E+03, 4.4998E+03, 4.5071E+03, 4.5144E+03, 4.5218E+03,
     +  4.5291E+03, 4.5364E+03, 4.5438E+03, 4.5511E+03, 4.5585E+03,
     + 4.5658E+03, 4.5732E+03, 4.5805E+03, 4.5879E+03, 4.5953E+03,
     +  4.6027E+03, 4.6101E+03, 4.6175E+03, 4.6249E+03, 4.6323E+03,
     + 4.6397E+03, 4.6471E+03, 4.6546E+03, 4.6620E+03, 4.6694E+03,
     +  4.6769E+03, 4.6843E+03, 4.6918E+03, 4.6992E+03, 4.7067E+03,
     + 4.7142E+03, 4.7217E+03, 4.7292E+03, 4.7366E+03, 4.7441E+03,
     +  4.7517E+03, 4.7592E+03, 4.7667E+03, 4.7742E+03, 4.7817E+03,
     + 4.7893E+03, 4.7968E+03, 4.8044E+03, 4.8119E+03, 4.8195E+03,
     +  4.8270E+03, 4.8346E+03, 4.8422E+03, 4.8498E+03, 4.8573E+03,
     + 4.8649E+03, 4.8725E+03, 4.8801E+03, 4.8878E+03, 4.8954E+03,
     +  4.9030E+03, 4.9106E+03, 4.9183E+03, 4.9259E+03, 4.9336E+03,
     + 4.9412E+03, 4.9489E+03, 4.9565E+03, 4.9642E+03, 4.9719E+03,
     +  4.9796E+03, 4.9873E+03, 4.9950E+03, 5.0027E+03, 5.0104E+03,
     + 5.0181E+03, 5.0258E+03, 5.0335E+03, 5.0413E+03, 5.0490E+03,
     +  5.0567E+03, 5.0645E+03, 5.0722E+03, 5.0800E+03, 5.0878E+03,
     + 5.0955E+03, 5.1033E+03, 5.1111E+03, 5.1189E+03, 5.1267E+03,
     +  5.1345E+03, 5.1423E+03, 5.1501E+03, 5.1579E+03, 5.1658E+03,
     + 5.1736E+03, 5.1814E+03, 5.1893E+03, 5.1971E+03, 5.2050E+03,
     +  5.2129E+03, 5.2207E+03, 5.2286E+03, 5.2365E+03, 5.2444E+03,
     + 5.2523E+03, 5.2602E+03, 5.2681E+03, 5.2760E+03, 5.2839E+03,
     +  5.2918E+03, 5.2998E+03, 5.3077E+03, 5.3156E+03, 5.3236E+03,
     + 5.3315E+03, 5.3395E+03, 5.3475E+03, 5.3554E+03, 5.3634E+03,
     +  5.3714E+03, 5.3794E+03, 5.3874E+03, 5.3954E+03, 5.4034E+03,
     + 5.4114E+03, 5.4194E+03, 5.4275E+03, 5.4355E+03, 5.4435E+03,
     +  5.4516E+03, 5.4596E+03, 5.4677E+03, 5.4757E+03, 5.4838E+03,
     + 5.4919E+03, 5.5000E+03, 5.5080E+03, 5.5161E+03, 5.5242E+03,
     +  5.5323E+03, 5.5405E+03, 5.5486E+03, 5.5567E+03, 5.5648E+03,
     + 5.5730E+03, 5.5811E+03, 5.5892E+03, 5.5974E+03, 5.6056E+03,
     +  5.6137E+03, 5.6219E+03, 5.6301E+03, 5.6383E+03, 5.6464E+03,
     + 5.6546E+03, 5.6628E+03, 5.6710E+03, 5.6793E+03, 5.6875E+03,
     +  5.6957E+03, 5.7039E+03, 5.7122E+03, 5.7204E+03, 5.7287E+03,
     + 5.7369E+03, 5.7452E+03, 5.7535E+03, 5.7617E+03, 5.7700E+03,
     +  5.7783E+03, 5.7866E+03, 5.7949E+03, 5.8032E+03, 5.8115E+03,
     + 5.8198E+03, 5.8281E+03, 5.8365E+03, 5.8448E+03, 5.8531E+03,
     +  5.8615E+03, 5.8698E+03, 5.8782E+03, 5.8866E+03, 5.8949E+03,
     + 5.9033E+03, 5.9117E+03, 5.9201E+03, 5.9285E+03, 5.9369E+03,
     +  5.9453E+03, 5.9537E+03, 5.9621E+03, 5.9706E+03, 5.9790E+03,
     + 5.9874E+03, 5.9959E+03, 6.0043E+03, 6.0128E+03, 6.0212E+03,
     +  6.0297E+03, 6.0382E+03, 6.0467E+03, 6.0551E+03, 6.0636E+03,
     + 6.0721E+03, 6.0806E+03, 6.0892E+03, 6.0977E+03, 6.1062E+03,
     +  6.1147E+03, 6.1233E+03, 6.1318E+03, 6.1403E+03, 6.1489E+03,
     + 6.1575E+03, 6.1660E+03, 6.1746E+03, 6.1832E+03, 6.1918E+03,
     +  6.2004E+03, 6.2089E+03, 6.2176E+03, 6.2262E+03, 6.2348E+03,
     + 6.2434E+03, 6.2520E+03, 6.2607E+03, 6.2693E+03, 6.2779E+03,
     +  6.2866E+03, 6.2952E+03, 6.3039E+03, 6.3126E+03, 6.3213E+03,
     + 6.3299E+03, 6.3386E+03, 6.3473E+03, 6.3560E+03, 6.3647E+03,
     +  6.3734E+03, 6.3822E+03, 6.3909E+03, 6.3996E+03, 6.4084E+03,
     + 6.4171E+03, 6.4258E+03, 6.4346E+03, 6.4434E+03, 6.4521E+03,
     +  6.4609E+03, 6.4697E+03, 6.4785E+03, 6.4873E+03, 6.4961E+03,
     + 6.5049E+03, 6.5137E+03, 6.5225E+03, 6.5313E+03, 6.5401E+03,
     +  6.5490E+03, 6.5578E+03, 6.5667E+03, 6.5755E+03, 6.5844E+03,
     + 6.5932E+03, 6.6021E+03, 6.6110E+03, 6.6199E+03, 6.6288E+03,
     +  6.6377E+03, 6.6466E+03, 6.6555E+03, 6.6644E+03, 6.6733E+03,
     + 6.6822E+03, 6.6912E+03, 6.7001E+03, 6.7091E+03, 6.7180E+03,
     +  6.7270E+03, 6.7359E+03, 6.7449E+03, 6.7539E+03, 6.7629E+03,
     + 6.7719E+03, 6.7808E+03, 6.7898E+03, 6.7989E+03, 6.8079E+03,
     +  6.8169E+03, 6.8259E+03, 6.8349E+03, 6.8440E+03, 6.8530E+03,
     + 6.8621E+03, 6.8711E+03, 6.8802E+03, 6.8893E+03, 6.8983E+03,
     +  6.9074E+03, 6.9165E+03, 6.9256E+03, 6.9347E+03, 6.9438E+03,
     + 6.9529E+03, 6.9620E+03, 6.9712E+03, 6.9803E+03, 6.9894E+03,
     +  6.9986E+03, 7.0077E+03, 7.0169E+03, 7.0260E+03, 7.0352E+03,
     + 7.0444E+03, 7.0535E+03, 7.0627E+03, 7.0719E+03, 7.0811E+03,
     +  7.0903E+03, 7.0995E+03, 7.1087E+03, 7.1180E+03, 7.1272E+03,
     + 7.1364E+03, 7.1457E+03, 7.1549E+03, 7.1642E+03, 7.1734E+03,
     +  7.1827E+03, 7.1920E+03, 7.2013E+03, 7.2105E+03, 7.2198E+03,
     + 7.2291E+03, 7.2384E+03, 7.2477E+03, 7.2570E+03, 7.2664E+03,
     +  7.2757E+03, 7.2850E+03, 7.2944E+03, 7.3037E+03, 7.3131E+03,
     + 7.3224E+03, 7.3318E+03, 7.3412E+03, 7.3505E+03, 7.3599E+03,
     +  7.3693E+03, 7.3787E+03, 7.3881E+03, 7.3975E+03, 7.4069E+03,
     + 7.4163E+03, 7.4258E+03, 7.4352E+03, 7.4446E+03, 7.4541E+03,
     +  7.4635E+03, 7.4730E+03, 7.4825E+03, 7.4919E+03, 7.5014E+03,
     + 7.5109E+03, 7.5204E+03, 7.5299E+03, 7.5394E+03, 7.5489E+03,
     +  7.5584E+03, 7.5679E+03, 7.5774E+03, 7.5870E+03, 7.5965E+03,
     + 7.6060E+03, 7.6156E+03, 7.6252E+03, 7.6347E+03, 7.6443E+03,
     +  7.6539E+03, 7.6634E+03, 7.6730E+03, 7.6826E+03, 7.6922E+03,
     + 7.7018E+03, 7.7114E+03, 7.7211E+03, 7.7307E+03, 7.7403E+03,
     +  7.7500E+03, 7.7596E+03, 7.7692E+03, 7.7789E+03, 7.7886E+03,
     + 7.7982E+03, 7.8079E+03, 7.8176E+03, 7.8273E+03, 7.8370E+03,
     +  7.8467E+03, 7.8564E+03, 7.8661E+03, 7.8758E+03, 7.8855E+03,
     + 7.8952E+03, 7.9050E+03, 7.9147E+03, 7.9245E+03, 7.9342E+03,
     +  7.9440E+03, 7.9538E+03, 7.9635E+03, 7.9733E+03, 7.9831E+03,
     + 7.9929E+03, 8.0027E+03, 8.0125E+03, 8.0223E+03, 8.0321E+03,
     +  8.0419E+03, 8.0518E+03, 8.0616E+03, 8.0715E+03, 8.0813E+03,
     + 8.0912E+03, 8.1010E+03, 8.1109E+03, 8.1208E+03, 8.1306E+03,
     +  8.1405E+03, 8.1504E+03, 8.1603E+03, 8.1702E+03, 8.1801E+03,
     + 8.1900E+03, 8.2000E+03, 8.2099E+03, 8.2198E+03, 8.2298E+03,
     +  8.2397E+03, 8.2497E+03, 8.2596E+03, 8.2696E+03, 8.2796E+03,
     + 8.2896E+03, 8.2995E+03, 8.3095E+03, 8.3195E+03, 8.3295E+03,
     +  8.3395E+03, 8.3496E+03, 8.3596E+03, 8.3696E+03, 8.3796E+03,
     + 8.3897E+03, 8.3997E+03, 8.4098E+03, 8.4198E+03, 8.4299E+03,
     +  8.4400E+03, 8.4501E+03, 8.4601E+03, 8.4702E+03, 8.4803E+03,
     + 8.4904E+03, 8.5005E+03, 8.5107E+03, 8.5208E+03, 8.5309E+03,
     +  8.5411E+03, 8.5512E+03, 8.5613E+03, 8.5715E+03, 8.5817E+03,
     + 8.5918E+03, 8.6020E+03, 8.6122E+03, 8.6224E+03, 8.6325E+03,
     +  8.6427E+03, 8.6529E+03, 8.6632E+03, 8.6734E+03, 8.6836E+03,
     + 8.6938E+03, 8.7041E+03, 8.7143E+03, 8.7245E+03, 8.7348E+03,
     +  8.7451E+03, 8.7553E+03, 8.7656E+03, 8.7759E+03, 8.7862E+03,
     + 8.7965E+03, 8.8068E+03, 8.8171E+03, 8.8274E+03, 8.8377E+03,
     +  8.8480E+03, 8.8583E+03, 8.8687E+03, 8.8790E+03, 8.8894E+03,
     + 8.8997E+03, 8.9101E+03, 8.9204E+03, 8.9308E+03, 8.9412E+03,
     +  8.9516E+03, 8.9620E+03, 8.9724E+03, 8.9828E+03, 8.9932E+03,
     + 9.0036E+03, 9.0140E+03, 9.0244E+03, 9.0349E+03, 9.0453E+03,
     +  9.0558E+03, 9.0662E+03, 9.0767E+03, 9.0872E+03, 9.0976E+03,
     + 9.1081E+03, 9.1186E+03, 9.1291E+03, 9.1396E+03, 9.1501E+03,
     +  9.1606E+03, 9.1711E+03, 9.1816E+03, 9.1922E+03, 9.2027E+03,
     + 9.2132E+03, 9.2238E+03, 9.2343E+03, 9.2449E+03, 9.2555E+03,
     +  9.2660E+03, 9.2766E+03, 9.2872E+03, 9.2978E+03, 9.3084E+03,
     + 9.3190E+03, 9.3296E+03, 9.3402E+03, 9.3509E+03, 9.3615E+03,
     +  9.3721E+03, 9.3828E+03, 9.3934E+03, 9.4041E+03, 9.4147E+03,
     + 9.4254E+03, 9.4361E+03, 9.4468E+03, 9.4574E+03, 9.4681E+03,
     +  9.4788E+03, 9.4895E+03, 9.5003E+03, 9.5110E+03, 9.5217E+03,
     + 9.5324E+03, 9.5432E+03, 9.5539E+03, 9.5647E+03, 9.5754E+03,
     +  9.5862E+03, 9.5969E+03, 9.6077E+03, 9.6185E+03, 9.6293E+03,
     + 9.6401E+03, 9.6509E+03, 9.6617E+03, 9.6725E+03, 9.6833E+03,
     +  9.6941E+03, 9.7050E+03, 9.7158E+03, 9.7266E+03, 9.7375E+03,
     + 9.7484E+03, 9.7592E+03, 9.7701E+03, 9.7810E+03, 9.7918E+03,
     +  9.8027E+03, 9.8136E+03, 9.8245E+03, 9.8354E+03, 9.8463E+03,
     + 9.8573E+03, 9.8682E+03, 9.8791E+03, 9.8901E+03, 9.9010E+03,
     +  9.9120E+03, 9.9229E+03, 9.9339E+03, 9.9448E+03, 9.9558E+03,
     + 9.9668E+03, 9.9778E+03, 9.9888E+03, 9.9998E+03, 1.0011E+04,
     +  1.0022E+04, 1.0033E+04, 1.0044E+04, 1.0055E+04, 1.0066E+04,
     + 1.0077E+04, 1.0088E+04, 1.0099E+04, 1.0110E+04, 1.0121E+04,
     +  1.0132E+04, 1.0143E+04, 1.0154E+04, 1.0166E+04, 1.0177E+04,
     + 1.0188E+04, 1.0199E+04, 1.0210E+04, 1.0221E+04, 1.0232E+04,
     +  1.0243E+04, 1.0255E+04, 1.0266E+04, 1.0277E+04, 1.0288E+04,
     + 1.0299E+04, 1.0310E+04, 1.0322E+04, 1.0333E+04, 1.0344E+04,
     +  1.0355E+04, 1.0366E+04, 1.0378E+04, 1.0389E+04, 1.0400E+04,
     + 1.0411E+04, 1.0423E+04, 1.0434E+04, 1.0445E+04, 1.0456E+04,
     +  1.0468E+04, 1.0479E+04, 1.0490E+04, 1.0501E+04, 1.0513E+04,
     + 1.0524E+04, 1.0535E+04, 1.0547E+04, 1.0558E+04, 1.0569E+04,
     +  1.0581E+04, 1.0592E+04, 1.0603E+04, 1.0615E+04, 1.0626E+04,
     + 1.0637E+04, 1.0649E+04, 1.0660E+04, 1.0671E+04, 1.0683E+04,
     +  1.0694E+04, 1.0706E+04, 1.0717E+04, 1.0728E+04, 1.0740E+04,
     + 1.0751E+04, 1.0763E+04, 1.0774E+04, 1.0786E+04, 1.0797E+04,
     +  1.0809E+04, 1.0820E+04, 1.0831E+04, 1.0843E+04, 1.0854E+04,
     + 1.0866E+04, 1.0877E+04, 1.0889E+04, 1.0900E+04, 1.0912E+04,
     +  1.0923E+04, 1.0935E+04, 1.0947E+04, 1.0958E+04, 1.0970E+04,
     + 1.0981E+04, 1.0993E+04, 1.1004E+04, 1.1016E+04, 1.1027E+04,
     +  1.1039E+04, 1.1051E+04, 1.1062E+04, 1.1074E+04, 1.1085E+04,
     + 1.1097E+04, 1.1109E+04, 1.1120E+04, 1.1132E+04, 1.1144E+04,
     +  1.1155E+04, 1.1167E+04, 1.1179E+04, 1.1190E+04, 1.1202E+04,
     + 1.1214E+04, 1.1225E+04, 1.1237E+04, 1.1249E+04, 1.1260E+04,
     +  1.1272E+04, 1.1284E+04, 1.1295E+04, 1.1307E+04, 1.1319E+04,
     + 1.1331E+04, 1.1342E+04, 1.1354E+04, 1.1366E+04, 1.1378E+04,
     +  1.1389E+04, 1.1401E+04, 1.1413E+04, 1.1425E+04, 1.1437E+04,
     + 1.1448E+04, 1.1460E+04, 1.1472E+04, 1.1484E+04, 1.1496E+04,
     +  1.1508E+04, 1.1519E+04, 1.1531E+04, 1.1543E+04, 1.1555E+04,
     + 1.1567E+04, 1.1579E+04, 1.1591E+04, 1.1603E+04, 1.1614E+04,
     +  1.1626E+04, 1.1638E+04, 1.1650E+04, 1.1662E+04, 1.1674E+04,
     + 1.1686E+04, 1.1698E+04, 1.1710E+04, 1.1722E+04, 1.1734E+04,
     +  1.1746E+04, 1.1758E+04, 1.1770E+04, 1.1782E+04, 1.1794E+04,
     + 1.1806E+04, 1.1818E+04, 1.1830E+04, 1.1842E+04, 1.1854E+04,
     +  1.1866E+04, 1.1878E+04, 1.1890E+04, 1.1902E+04, 1.1914E+04,
     + 1.1926E+04, 1.1938E+04, 1.1950E+04, 1.1962E+04, 1.1974E+04,
     +  1.1986E+04, 1.1998E+04, 1.2011E+04, 1.2023E+04, 1.2035E+04,
     + 1.2047E+04, 1.2059E+04, 1.2071E+04, 1.2083E+04, 1.2095E+04,
     +  1.2108E+04, 1.2120E+04, 1.2132E+04, 1.2144E+04, 1.2156E+04,
     + 1.2169E+04, 1.2181E+04, 1.2193E+04, 1.2205E+04, 1.2217E+04,
     +  1.2230E+04, 1.2242E+04, 1.2254E+04, 1.2266E+04, 1.2279E+04,
     + 1.2291E+04, 1.2303E+04, 1.2315E+04, 1.2328E+04, 1.2340E+04,
     +  1.2352E+04, 1.2364E+04, 1.2377E+04, 1.2389E+04, 1.2401E+04,
     + 1.2414E+04, 1.2426E+04, 1.2438E+04, 1.2451E+04, 1.2463E+04,
     +  1.2475E+04, 1.2488E+04, 1.2500E+04, 1.2512E+04, 1.2525E+04,
     + 1.2537E+04, 1.2550E+04, 1.2562E+04, 1.2574E+04, 1.2587E+04,
     +  1.2599E+04, 1.2612E+04, 1.2624E+04, 1.2636E+04, 1.2649E+04,
     + 1.2661E+04, 1.2674E+04, 1.2686E+04, 1.2699E+04, 1.2711E+04,
     +  1.2724E+04, 1.2736E+04, 1.2749E+04, 1.2761E+04, 1.2774E+04,
     + 1.2786E+04, 1.2799E+04, 1.2811E+04, 1.2824E+04, 1.2836E+04,
     +  1.2849E+04, 1.2861E+04, 1.2874E+04, 1.2887E+04, 1.2899E+04,
     + 1.2912E+04, 1.2924E+04, 1.2937E+04, 1.2949E+04, 1.2962E+04,
     +  1.2975E+04, 1.2987E+04, 1.3000E+04, 1.3012E+04, 1.3025E+04,
     + 1.3038E+04, 1.3050E+04, 1.3063E+04, 1.3076E+04, 1.3088E+04,
     +  1.3101E+04, 1.3114E+04, 1.3126E+04, 1.3139E+04, 1.3152E+04,
     + 1.3164E+04, 1.3177E+04, 1.3190E+04, 1.3203E+04, 1.3215E+04,
     +  1.3228E+04, 1.3241E+04, 1.3254E+04, 1.3266E+04, 1.3279E+04,
     + 1.3292E+04, 1.3305E+04, 1.3317E+04, 1.3330E+04, 1.3343E+04,
     +  1.3356E+04, 1.3369E+04, 1.3381E+04, 1.3394E+04, 1.3407E+04,
     + 1.3420E+04, 1.3433E+04, 1.3446E+04, 1.3458E+04, 1.3471E+04,
     +  1.3484E+04, 1.3497E+04, 1.3510E+04, 1.3523E+04, 1.3536E+04,
     + 1.3549E+04, 1.3561E+04, 1.3574E+04, 1.3587E+04, 1.3600E+04,
     +  1.3613E+04, 1.3626E+04, 1.3639E+04, 1.3652E+04, 1.3665E+04,
     + 1.3678E+04, 1.3691E+04, 1.3704E+04, 1.3717E+04, 1.3730E+04,
     +  1.3743E+04, 1.3756E+04, 1.3769E+04, 1.3782E+04, 1.3795E+04,
     + 1.3808E+04, 1.3821E+04, 1.3834E+04, 1.3847E+04, 1.3860E+04,
     +  1.3873E+04, 1.3886E+04, 1.3899E+04, 1.3912E+04, 1.3925E+04,
     + 1.3938E+04, 1.3952E+04, 1.3965E+04, 1.3978E+04, 1.3991E+04,
     +  1.4004E+04, 1.4017E+04, 1.4030E+04, 1.4043E+04, 1.4057E+04,
     + 1.4070E+04, 1.4083E+04, 1.4096E+04, 1.4109E+04, 1.4122E+04,
     +  1.4136E+04, 1.4149E+04, 1.4162E+04, 1.4175E+04, 1.4188E+04,
     + 1.4202E+04, 1.4215E+04, 1.4228E+04, 1.4241E+04, 1.4255E+04,
     +  1.4268E+04, 1.4281E+04, 1.4294E+04, 1.4308E+04, 1.4321E+04,
     + 1.4334E+04, 1.4347E+04, 1.4361E+04, 1.4374E+04, 1.4387E+04,
     +  1.4401E+04, 1.4414E+04, 1.4427E+04, 1.4441E+04, 1.4454E+04,
     + 1.4467E+04, 1.4481E+04, 1.4494E+04, 1.4507E+04, 1.4521E+04,
     +  1.4534E+04, 1.4548E+04, 1.4561E+04, 1.4574E+04, 1.4588E+04,
     + 1.4601E+04, 1.4615E+04, 1.4628E+04, 1.4641E+04, 1.4655E+04,
     +  1.4668E+04, 1.4682E+04, 1.4695E+04, 1.4709E+04, 1.4722E+04,
     + 1.4736E+04, 1.4749E+04, 1.4763E+04, 1.4776E+04, 1.4790E+04,
     +  1.4803E+04, 1.4817E+04, 1.4830E+04, 1.4844E+04, 1.4857E+04,
     + 1.4871E+04, 1.4884E+04, 1.4898E+04, 1.4911E+04, 1.4925E+04,
     +  1.4939E+04, 1.4952E+04, 1.4966E+04, 1.4979E+04, 1.4993E+04,
     + 1.5007E+04, 1.5020E+04, 1.5034E+04, 1.5047E+04, 1.5061E+04,
     +  1.5075E+04, 1.5088E+04, 1.5102E+04, 1.5116E+04, 1.5129E+04,
     + 1.5143E+04, 1.5157E+04, 1.5170E+04, 1.5184E+04, 1.5198E+04,
     +  1.5211E+04, 1.5225E+04, 1.5239E+04, 1.5253E+04, 1.5266E+04,
     + 1.5280E+04, 1.5294E+04, 1.5308E+04, 1.5321E+04, 1.5335E+04,
     +  1.5349E+04, 1.5363E+04, 1.5376E+04, 1.5390E+04, 1.5404E+04,
     + 1.5418E+04, 1.5432E+04, 1.5445E+04, 1.5459E+04, 1.5473E+04,
     +  1.5487E+04, 1.5501E+04, 1.5515E+04, 1.5528E+04, 1.5542E+04,
     + 1.5556E+04, 1.5570E+04, 1.5584E+04, 1.5598E+04, 1.5612E+04,
     +  1.5626E+04, 1.5640E+04, 1.5653E+04, 1.5667E+04, 1.5681E+04,
     + 1.5695E+04, 1.5709E+04, 1.5723E+04, 1.5737E+04, 1.5751E+04,
     +  1.5765E+04, 1.5779E+04, 1.5793E+04, 1.5807E+04, 1.5821E+04,
     + 1.5835E+04, 1.5849E+04, 1.5863E+04, 1.5877E+04, 1.5891E+04,
     +  1.5905E+04, 1.5919E+04, 1.5933E+04, 1.5947E+04, 1.5961E+04,
     + 1.5975E+04, 1.5989E+04, 1.6003E+04, 1.6017E+04, 1.6032E+04,
     +  1.6046E+04, 1.6060E+04, 1.6074E+04, 1.6088E+04, 1.6102E+04,
     + 1.6116E+04, 1.6130E+04, 1.6145E+04, 1.6159E+04, 1.6173E+04,
     +  1.6187E+04, 1.6201E+04, 1.6215E+04, 1.6230E+04, 1.6244E+04,
     + 1.6258E+04, 1.6272E+04, 1.6286E+04, 1.6301E+04, 1.6315E+04,
     +  1.6329E+04, 1.6343E+04, 1.6357E+04, 1.6372E+04, 1.6386E+04,
     + 1.6400E+04, 1.6414E+04, 1.6429E+04, 1.6443E+04, 1.6457E+04,
     +  1.6472E+04, 1.6486E+04, 1.6500E+04, 1.6514E+04, 1.6529E+04,
     + 1.6543E+04, 1.6557E+04, 1.6572E+04, 1.6586E+04, 1.6600E+04,
     +  1.6615E+04, 1.6629E+04, 1.6644E+04, 1.6658E+04, 1.6672E+04,
     + 1.6687E+04, 1.6701E+04, 1.6716E+04, 1.6730E+04, 1.6744E+04,
     +  1.6759E+04, 1.6773E+04, 1.6788E+04, 1.6802E+04, 1.6817E+04,
     + 1.6831E+04, 1.6845E+04, 1.6860E+04, 1.6874E+04, 1.6889E+04,
     +  1.6903E+04, 1.6918E+04, 1.6932E+04, 1.6947E+04, 1.6961E+04,
     + 1.6976E+04, 1.6990E+04, 1.7005E+04, 1.7020E+04, 1.7034E+04,
     +  1.7049E+04, 1.7063E+04, 1.7078E+04, 1.7092E+04, 1.7107E+04,
     + 1.7121E+04, 1.7136E+04, 1.7151E+04, 1.7165E+04, 1.7180E+04,
     +  1.7195E+04, 1.7209E+04, 1.7224E+04, 1.7238E+04, 1.7253E+04,
     + 1.7268E+04, 1.7282E+04, 1.7297E+04, 1.7312E+04, 1.7326E+04,
     +  1.7341E+04, 1.7356E+04, 1.7370E+04, 1.7385E+04, 1.7400E+04,
     + 1.7415E+04, 1.7429E+04, 1.7444E+04, 1.7459E+04, 1.7474E+04,
     +  1.7488E+04, 1.7503E+04, 1.7518E+04, 1.7533E+04, 1.7547E+04,
     + 1.7562E+04, 1.7577E+04, 1.7592E+04, 1.7607E+04, 1.7621E+04,
     +  1.7636E+04, 1.7651E+04, 1.7666E+04, 1.7681E+04, 1.7695E+04,
     + 1.7710E+04, 1.7725E+04, 1.7740E+04, 1.7755E+04, 1.7770E+04,
     +  1.7785E+04, 1.7800E+04, 1.7814E+04, 1.7829E+04, 1.7844E+04,
     + 1.7859E+04, 1.7874E+04, 1.7889E+04, 1.7904E+04, 1.7919E+04,
     +  1.7934E+04, 1.7949E+04, 1.7964E+04, 1.7979E+04, 1.7994E+04,
     + 1.8009E+04, 1.8024E+04, 1.8039E+04, 1.8054E+04, 1.8069E+04,
     +  1.8084E+04, 1.8099E+04, 1.8114E+04, 1.8129E+04, 1.8144E+04,
     + 1.8159E+04, 1.8174E+04, 1.8189E+04, 1.8204E+04, 1.8219E+04,
     +  1.8234E+04, 1.8249E+04, 1.8264E+04, 1.8279E+04, 1.8295E+04,
     + 1.8310E+04, 1.8325E+04, 1.8340E+04, 1.8355E+04, 1.8370E+04,
     +  1.8385E+04, 1.8400E+04, 1.8416E+04, 1.8431E+04, 1.8446E+04,
     + 1.8461E+04, 1.8476E+04, 1.8492E+04, 1.8507E+04, 1.8522E+04,
     +  1.8537E+04, 1.8552E+04, 1.8568E+04, 1.8583E+04, 1.8598E+04,
     + 1.8613E+04, 1.8629E+04, 1.8644E+04, 1.8659E+04, 1.8674E+04,
     +  1.8690E+04, 1.8705E+04, 1.8720E+04, 1.8735E+04, 1.8751E+04,
     + 1.8766E+04, 1.8781E+04, 1.8797E+04, 1.8812E+04, 1.8827E+04,
     +  1.8843E+04, 1.8858E+04, 1.8873E+04, 1.8889E+04, 1.8904E+04,
     + 1.8920E+04, 1.8935E+04, 1.8950E+04, 1.8966E+04, 1.8981E+04,
     +  1.8997E+04, 1.9012E+04, 1.9027E+04, 1.9043E+04, 1.9058E+04,
     + 1.9074E+04, 1.9089E+04, 1.9105E+04, 1.9120E+04, 1.9136E+04,
     +  1.9151E+04, 1.9166E+04, 1.9182E+04, 1.9197E+04, 1.9213E+04,
     + 1.9228E+04, 1.9244E+04, 1.9259E+04, 1.9275E+04, 1.9291E+04,
     +  1.9306E+04, 1.9322E+04, 1.9337E+04, 1.9353E+04, 1.9368E+04,
     + 1.9384E+04, 1.9399E+04, 1.9415E+04, 1.9431E+04, 1.9446E+04,
     +  1.9462E+04, 1.9477E+04, 1.9493E+04, 1.9509E+04, 1.9524E+04,
     + 1.9540E+04, 1.9556E+04, 1.9571E+04, 1.9587E+04, 1.9603E+04,
     +  1.9618E+04, 1.9634E+04, 1.9650E+04, 1.9665E+04, 1.9681E+04,
     + 1.9697E+04, 1.9713E+04, 1.9728E+04, 1.9744E+04, 1.9760E+04,
     +  1.9775E+04, 1.9791E+04, 1.9807E+04, 1.9823E+04, 1.9838E+04,
     + 1.9854E+04, 1.9870E+04, 1.9886E+04, 1.9902E+04, 1.9917E+04,
     +  1.9933E+04, 1.9949E+04, 1.9965E+04, 1.9981E+04, 1.9996E+04,
     + 2.0012E+04, 2.0028E+04, 2.0044E+04, 2.0060E+04, 2.0076E+04,
     +  2.0092E+04, 2.0108E+04, 2.0123E+04, 2.0139E+04, 2.0155E+04,
     + 2.0171E+04, 2.0187E+04, 2.0203E+04, 2.0219E+04, 2.0235E+04,
     +  2.0251E+04, 2.0267E+04, 2.0283E+04, 2.0299E+04, 2.0315E+04,
     + 2.0331E+04, 2.0346E+04, 2.0362E+04, 2.0378E+04, 2.0394E+04,
     +  2.0410E+04, 2.0426E+04, 2.0443E+04, 2.0459E+04, 2.0475E+04,
     + 2.0491E+04, 2.0507E+04, 2.0523E+04, 2.0539E+04, 2.0555E+04,
     +  2.0571E+04, 2.0587E+04, 2.0603E+04, 2.0619E+04, 2.0635E+04,
     + 2.0651E+04, 2.0667E+04, 2.0684E+04, 2.0700E+04, 2.0716E+04,
     +  2.0732E+04, 2.0748E+04, 2.0764E+04, 2.0780E+04, 2.0797E+04,
     + 2.0813E+04, 2.0829E+04, 2.0845E+04, 2.0861E+04, 2.0878E+04,
     +  2.0894E+04, 2.0910E+04, 2.0926E+04, 2.0942E+04, 2.0959E+04,
     + 2.0975E+04, 2.0991E+04, 2.1007E+04, 2.1024E+04, 2.1040E+04,
     +  2.1056E+04, 2.1072E+04, 2.1089E+04, 2.1105E+04, 2.1121E+04,
     + 2.1138E+04, 2.1154E+04, 2.1170E+04, 2.1187E+04, 2.1203E+04,
     +  2.1219E+04, 2.1236E+04, 2.1252E+04, 2.1268E+04, 2.1285E+04,
     + 2.1301E+04, 2.1317E+04, 2.1334E+04, 2.1350E+04, 2.1367E+04,
     +  2.1383E+04, 2.1399E+04, 2.1416E+04, 2.1432E+04, 2.1449E+04,
     + 2.1465E+04, 2.1482E+04, 2.1498E+04, 2.1514E+04, 2.1531E+04,
     +  2.1547E+04, 2.1564E+04, 2.1580E+04, 2.1597E+04, 2.1613E+04,
     + 2.1630E+04, 2.1646E+04, 2.1663E+04, 2.1679E+04, 2.1696E+04,
     +  2.1713E+04, 2.1729E+04, 2.1746E+04, 2.1762E+04, 2.1779E+04,
     + 2.1795E+04, 2.1812E+04, 2.1828E+04, 2.1845E+04, 2.1862E+04,
     +  2.1878E+04, 2.1895E+04, 2.1912E+04, 2.1928E+04, 2.1945E+04,
     + 2.1961E+04, 2.1978E+04, 2.1995E+04, 2.2011E+04, 2.2028E+04,
     +  2.2045E+04, 2.2061E+04, 2.2078E+04, 2.2095E+04, 2.2111E+04,
     + 2.2128E+04, 2.2145E+04, 2.2162E+04, 2.2178E+04, 2.2195E+04,
     +  2.2212E+04, 2.2229E+04, 2.2245E+04, 2.2262E+04, 2.2279E+04,
     + 2.2296E+04, 2.2312E+04, 2.2329E+04, 2.2346E+04, 2.2363E+04,
     +  2.2380E+04, 2.2396E+04, 2.2413E+04, 2.2430E+04, 2.2447E+04,
     + 2.2464E+04, 2.2481E+04, 2.2497E+04, 2.2514E+04, 2.2531E+04,
     +  2.2548E+04, 2.2565E+04, 2.2582E+04, 2.2599E+04, 2.2616E+04,
     + 2.2632E+04, 2.2649E+04, 2.2666E+04, 2.2683E+04, 2.2700E+04,
     +  2.2717E+04, 2.2734E+04, 2.2751E+04, 2.2768E+04, 2.2785E+04,
     + 2.2802E+04, 2.2819E+04, 2.2836E+04, 2.2853E+04, 2.2870E+04,
     +  2.2887E+04, 2.2904E+04, 2.2921E+04, 2.2938E+04, 2.2955E+04,
     + 2.2972E+04, 2.2989E+04, 2.3006E+04, 2.3023E+04, 2.3040E+04,
     +  2.3057E+04, 2.3074E+04, 2.3092E+04, 2.3109E+04, 2.3126E+04,
     + 2.3143E+04, 2.3160E+04, 2.3177E+04, 2.3194E+04, 2.3211E+04,
     +  2.3228E+04, 2.3246E+04, 2.3263E+04, 2.3280E+04, 2.3297E+04,
     + 2.3314E+04, 2.3331E+04, 2.3349E+04, 2.3366E+04, 2.3383E+04,
     +  2.3400E+04, 2.3417E+04, 2.3435E+04, 2.3452E+04, 2.3469E+04/
      data (QofT( 4,J),J=2409,4510)/ 2.3486E+04, 2.3504E+04,
     +  2.3521E+04, 2.3538E+04, 2.3555E+04, 2.3573E+04, 2.3590E+04,
     +  2.3607E+04,
     + 2.3625E+04, 2.3642E+04, 2.3659E+04, 2.3677E+04, 2.3694E+04,
     +  2.3711E+04, 2.3729E+04, 2.3746E+04, 2.3763E+04, 2.3781E+04,
     + 2.3798E+04, 2.3815E+04, 2.3833E+04, 2.3850E+04, 2.3867E+04,
     +  2.3885E+04, 2.3902E+04, 2.3920E+04, 2.3937E+04, 2.3955E+04,
     + 2.3972E+04, 2.3989E+04, 2.4007E+04, 2.4024E+04, 2.4042E+04,
     +  2.4059E+04, 2.4077E+04, 2.4094E+04, 2.4112E+04, 2.4129E+04,
     + 2.4147E+04, 2.4164E+04, 2.4182E+04, 2.4199E+04, 2.4217E+04,
     +  2.4234E+04, 2.4252E+04, 2.4269E+04, 2.4287E+04, 2.4304E+04,
     + 2.4322E+04, 2.4340E+04, 2.4357E+04, 2.4375E+04, 2.4392E+04,
     +  2.4410E+04, 2.4428E+04, 2.4445E+04, 2.4463E+04, 2.4480E+04,
     + 2.4498E+04, 2.4516E+04, 2.4533E+04, 2.4551E+04, 2.4569E+04,
     +  2.4586E+04, 2.4604E+04, 2.4622E+04, 2.4639E+04, 2.4657E+04,
     + 2.4675E+04, 2.4692E+04, 2.4710E+04, 2.4728E+04, 2.4746E+04,
     +  2.4763E+04, 2.4781E+04, 2.4799E+04, 2.4817E+04, 2.4834E+04,
     + 2.4852E+04, 2.4870E+04, 2.4888E+04, 2.4906E+04, 2.4923E+04,
     +  2.4941E+04, 2.4959E+04, 2.4977E+04, 2.4995E+04, 2.5012E+04,
     + 2.5030E+04, 2.5048E+04, 2.5066E+04, 2.5084E+04, 2.5102E+04,
     +  2.5120E+04, 2.5137E+04, 2.5155E+04, 2.5173E+04, 2.5191E+04,
     + 2.5209E+04, 2.5227E+04, 2.5245E+04, 2.5263E+04, 2.5281E+04,
     +  2.5299E+04, 2.5317E+04, 2.5335E+04, 2.5352E+04, 2.5370E+04,
     + 2.5388E+04, 2.5406E+04, 2.5424E+04, 2.5442E+04, 2.5460E+04,
     +  2.5478E+04, 2.5496E+04, 2.5514E+04, 2.5532E+04, 2.5550E+04,
     + 2.5568E+04, 2.5587E+04, 2.5605E+04, 2.5623E+04, 2.5641E+04,
     +  2.5659E+04, 2.5677E+04, 2.5695E+04, 2.5713E+04, 2.5731E+04,
     + 2.5749E+04, 2.5767E+04, 2.5786E+04, 2.5804E+04, 2.5822E+04,
     +  2.5840E+04, 2.5858E+04, 2.5876E+04, 2.5894E+04, 2.5913E+04,
     + 2.5931E+04, 2.5949E+04, 2.5967E+04, 2.5985E+04, 2.6003E+04,
     +  2.6022E+04, 2.6040E+04, 2.6058E+04, 2.6076E+04, 2.6095E+04,
     + 2.6113E+04, 2.6131E+04, 2.6149E+04, 2.6168E+04, 2.6186E+04,
     +  2.6204E+04, 2.6222E+04, 2.6241E+04, 2.6259E+04, 2.6277E+04,
     + 2.6296E+04, 2.6314E+04, 2.6332E+04, 2.6351E+04, 2.6369E+04,
     +  2.6387E+04, 2.6406E+04, 2.6424E+04, 2.6442E+04, 2.6461E+04,
     + 2.6479E+04, 2.6498E+04, 2.6516E+04, 2.6534E+04, 2.6553E+04,
     +  2.6571E+04, 2.6590E+04, 2.6608E+04, 2.6626E+04, 2.6645E+04,
     + 2.6663E+04, 2.6682E+04, 2.6700E+04, 2.6719E+04, 2.6737E+04,
     +  2.6756E+04, 2.6774E+04, 2.6793E+04, 2.6811E+04, 2.6830E+04,
     + 2.6848E+04, 2.6867E+04, 2.6885E+04, 2.6904E+04, 2.6922E+04,
     +  2.6941E+04, 2.6959E+04, 2.6978E+04, 2.6997E+04, 2.7015E+04,
     + 2.7034E+04, 2.7052E+04, 2.7071E+04, 2.7090E+04, 2.7108E+04,
     +  2.7127E+04, 2.7145E+04, 2.7164E+04, 2.7183E+04, 2.7201E+04,
     + 2.7220E+04, 2.7239E+04, 2.7257E+04, 2.7276E+04, 2.7295E+04,
     +  2.7313E+04, 2.7332E+04, 2.7351E+04, 2.7369E+04, 2.7388E+04,
     + 2.7407E+04, 2.7426E+04, 2.7444E+04, 2.7463E+04, 2.7482E+04,
     +  2.7501E+04, 2.7519E+04, 2.7538E+04, 2.7557E+04, 2.7576E+04,
     + 2.7594E+04, 2.7613E+04, 2.7632E+04, 2.7651E+04, 2.7670E+04,
     +  2.7688E+04, 2.7707E+04, 2.7726E+04, 2.7745E+04, 2.7764E+04,
     + 2.7783E+04, 2.7802E+04, 2.7820E+04, 2.7839E+04, 2.7858E+04,
     +  2.7877E+04, 2.7896E+04, 2.7915E+04, 2.7934E+04, 2.7953E+04,
     + 2.7972E+04, 2.7991E+04, 2.8009E+04, 2.8028E+04, 2.8047E+04,
     +  2.8066E+04, 2.8085E+04, 2.8104E+04, 2.8123E+04, 2.8142E+04,
     + 2.8161E+04, 2.8180E+04, 2.8199E+04, 2.8218E+04, 2.8237E+04,
     +  2.8256E+04, 2.8275E+04, 2.8294E+04, 2.8313E+04, 2.8332E+04,
     + 2.8352E+04, 2.8371E+04, 2.8390E+04, 2.8409E+04, 2.8428E+04,
     +  2.8447E+04, 2.8466E+04, 2.8485E+04, 2.8504E+04, 2.8523E+04,
     + 2.8543E+04, 2.8562E+04, 2.8581E+04, 2.8600E+04, 2.8619E+04,
     +  2.8638E+04, 2.8657E+04, 2.8677E+04, 2.8696E+04, 2.8715E+04,
     + 2.8734E+04, 2.8753E+04, 2.8773E+04, 2.8792E+04, 2.8811E+04,
     +  2.8830E+04, 2.8850E+04, 2.8869E+04, 2.8888E+04, 2.8907E+04,
     + 2.8927E+04, 2.8946E+04, 2.8965E+04, 2.8984E+04, 2.9004E+04,
     +  2.9023E+04, 2.9042E+04, 2.9062E+04, 2.9081E+04, 2.9100E+04,
     + 2.9120E+04, 2.9139E+04, 2.9158E+04, 2.9178E+04, 2.9197E+04,
     +  2.9216E+04, 2.9236E+04, 2.9255E+04, 2.9275E+04, 2.9294E+04,
     + 2.9313E+04, 2.9333E+04, 2.9352E+04, 2.9372E+04, 2.9391E+04,
     +  2.9410E+04, 2.9430E+04, 2.9449E+04, 2.9469E+04, 2.9488E+04,
     + 2.9508E+04, 2.9527E+04, 2.9547E+04, 2.9566E+04, 2.9586E+04,
     +  2.9605E+04, 2.9625E+04, 2.9644E+04, 2.9664E+04, 2.9683E+04,
     + 2.9703E+04, 2.9722E+04, 2.9742E+04, 2.9762E+04, 2.9781E+04,
     +  2.9801E+04, 2.9820E+04, 2.9840E+04, 2.9859E+04, 2.9879E+04,
     + 2.9899E+04, 2.9918E+04, 2.9938E+04, 2.9958E+04, 2.9977E+04,
     +  2.9997E+04, 3.0017E+04, 3.0036E+04, 3.0056E+04, 3.0076E+04,
     + 3.0095E+04, 3.0115E+04, 3.0135E+04, 3.0154E+04, 3.0174E+04,
     +  3.0194E+04, 3.0213E+04, 3.0233E+04, 3.0253E+04, 3.0273E+04,
     + 3.0292E+04, 3.0312E+04, 3.0332E+04, 3.0352E+04, 3.0371E+04,
     +  3.0391E+04, 3.0411E+04, 3.0431E+04, 3.0451E+04, 3.0470E+04,
     + 3.0490E+04, 3.0510E+04, 3.0530E+04, 3.0550E+04, 3.0570E+04,
     +  3.0589E+04, 3.0609E+04, 3.0629E+04, 3.0649E+04, 3.0669E+04,
     + 3.0689E+04, 3.0709E+04, 3.0729E+04, 3.0748E+04, 3.0768E+04,
     +  3.0788E+04, 3.0808E+04, 3.0828E+04, 3.0848E+04, 3.0868E+04,
     + 3.0888E+04, 3.0908E+04, 3.0928E+04, 3.0948E+04, 3.0968E+04,
     +  3.0988E+04, 3.1008E+04, 3.1028E+04, 3.1048E+04, 3.1068E+04,
     + 3.1088E+04, 3.1108E+04, 3.1128E+04, 3.1148E+04, 3.1168E+04,
     +  3.1188E+04, 3.1208E+04, 3.1228E+04, 3.1248E+04, 3.1268E+04,
     + 3.1289E+04, 3.1309E+04, 3.1329E+04, 3.1349E+04, 3.1369E+04,
     +  3.1389E+04, 3.1409E+04, 3.1429E+04, 3.1450E+04, 3.1470E+04,
     + 3.1490E+04, 3.1510E+04, 3.1530E+04, 3.1550E+04, 3.1571E+04,
     +  3.1591E+04, 3.1611E+04, 3.1631E+04, 3.1651E+04, 3.1672E+04,
     + 3.1692E+04, 3.1712E+04, 3.1732E+04, 3.1753E+04, 3.1773E+04,
     +  3.1793E+04, 3.1813E+04, 3.1834E+04, 3.1854E+04, 3.1874E+04,
     + 3.1895E+04, 3.1915E+04, 3.1935E+04, 3.1955E+04, 3.1976E+04,
     +  3.1996E+04, 3.2016E+04, 3.2037E+04, 3.2057E+04, 3.2078E+04,
     + 3.2098E+04, 3.2118E+04, 3.2139E+04, 3.2159E+04, 3.2179E+04,
     +  3.2200E+04, 3.2220E+04, 3.2241E+04, 3.2261E+04, 3.2282E+04,
     + 3.2302E+04, 3.2322E+04, 3.2343E+04, 3.2363E+04, 3.2384E+04,
     +  3.2404E+04, 3.2425E+04, 3.2445E+04, 3.2466E+04, 3.2486E+04,
     + 3.2507E+04, 3.2527E+04, 3.2548E+04, 3.2568E+04, 3.2589E+04,
     +  3.2609E+04, 3.2630E+04, 3.2650E+04, 3.2671E+04, 3.2692E+04,
     + 3.2712E+04, 3.2733E+04, 3.2753E+04, 3.2774E+04, 3.2795E+04,
     +  3.2815E+04, 3.2836E+04, 3.2856E+04, 3.2877E+04, 3.2898E+04,
     + 3.2918E+04, 3.2939E+04, 3.2960E+04, 3.2980E+04, 3.3001E+04,
     +  3.3022E+04, 3.3042E+04, 3.3063E+04, 3.3084E+04, 3.3104E+04,
     + 3.3125E+04, 3.3146E+04, 3.3167E+04, 3.3187E+04, 3.3208E+04,
     +  3.3229E+04, 3.3250E+04, 3.3270E+04, 3.3291E+04, 3.3312E+04,
     + 3.3333E+04, 3.3354E+04, 3.3374E+04, 3.3395E+04, 3.3416E+04,
     +  3.3437E+04, 3.3458E+04, 3.3478E+04, 3.3499E+04, 3.3520E+04,
     + 3.3541E+04, 3.3562E+04, 3.3583E+04, 3.3604E+04, 3.3624E+04,
     +  3.3645E+04, 3.3666E+04, 3.3687E+04, 3.3708E+04, 3.3729E+04,
     + 3.3750E+04, 3.3771E+04, 3.3792E+04, 3.3813E+04, 3.3834E+04,
     +  3.3855E+04, 3.3876E+04, 3.3897E+04, 3.3918E+04, 3.3939E+04,
     + 3.3960E+04, 3.3981E+04, 3.4002E+04, 3.4023E+04, 3.4044E+04,
     +  3.4065E+04, 3.4086E+04, 3.4107E+04, 3.4128E+04, 3.4149E+04,
     + 3.4170E+04, 3.4191E+04, 3.4212E+04, 3.4233E+04, 3.4254E+04,
     +  3.4275E+04, 3.4296E+04, 3.4317E+04, 3.4339E+04, 3.4360E+04,
     + 3.4381E+04, 3.4402E+04, 3.4423E+04, 3.4444E+04, 3.4465E+04,
     +  3.4487E+04, 3.4508E+04, 3.4529E+04, 3.4550E+04, 3.4571E+04,
     + 3.4593E+04, 3.4614E+04, 3.4635E+04, 3.4656E+04, 3.4677E+04,
     +  3.4699E+04, 3.4720E+04, 3.4741E+04, 3.4762E+04, 3.4784E+04,
     + 3.4805E+04, 3.4826E+04, 3.4848E+04, 3.4869E+04, 3.4890E+04,
     +  3.4911E+04, 3.4933E+04, 3.4954E+04, 3.4975E+04, 3.4997E+04,
     + 3.5018E+04, 3.5039E+04, 3.5061E+04, 3.5082E+04, 3.5103E+04,
     +  3.5125E+04, 3.5146E+04, 3.5168E+04, 3.5189E+04, 3.5210E+04,
     + 3.5232E+04, 3.5253E+04, 3.5275E+04, 3.5296E+04, 3.5318E+04,
     +  3.5339E+04, 3.5360E+04, 3.5382E+04, 3.5403E+04, 3.5425E+04,
     + 3.5446E+04, 3.5468E+04, 3.5489E+04, 3.5511E+04, 3.5532E+04,
     +  3.5554E+04, 3.5575E+04, 3.5597E+04, 3.5618E+04, 3.5640E+04,
     + 3.5662E+04, 3.5683E+04, 3.5705E+04, 3.5726E+04, 3.5748E+04,
     +  3.5769E+04, 3.5791E+04, 3.5813E+04, 3.5834E+04, 3.5856E+04,
     + 3.5877E+04, 3.5899E+04, 3.5921E+04, 3.5942E+04, 3.5964E+04,
     +  3.5986E+04, 3.6007E+04, 3.6029E+04, 3.6051E+04, 3.6072E+04,
     + 3.6094E+04, 3.6116E+04, 3.6137E+04, 3.6159E+04, 3.6181E+04,
     +  3.6203E+04, 3.6224E+04, 3.6246E+04, 3.6268E+04, 3.6290E+04,
     + 3.6311E+04, 3.6333E+04, 3.6355E+04, 3.6377E+04, 3.6398E+04,
     +  3.6420E+04, 3.6442E+04, 3.6464E+04, 3.6486E+04, 3.6507E+04,
     + 3.6529E+04, 3.6551E+04, 3.6573E+04, 3.6595E+04, 3.6617E+04,
     +  3.6639E+04, 3.6660E+04, 3.6682E+04, 3.6704E+04, 3.6726E+04,
     + 3.6748E+04, 3.6770E+04, 3.6792E+04, 3.6814E+04, 3.6836E+04,
     +  3.6858E+04, 3.6880E+04, 3.6901E+04, 3.6923E+04, 3.6945E+04,
     + 3.6967E+04, 3.6989E+04, 3.7011E+04, 3.7033E+04, 3.7055E+04,
     +  3.7077E+04, 3.7099E+04, 3.7121E+04, 3.7143E+04, 3.7165E+04,
     + 3.7187E+04, 3.7210E+04, 3.7232E+04, 3.7254E+04, 3.7276E+04,
     +  3.7298E+04, 3.7320E+04, 3.7342E+04, 3.7364E+04, 3.7386E+04,
     + 3.7408E+04, 3.7430E+04, 3.7453E+04, 3.7475E+04, 3.7497E+04,
     +  3.7519E+04, 3.7541E+04, 3.7563E+04, 3.7585E+04, 3.7608E+04,
     + 3.7630E+04, 3.7652E+04, 3.7674E+04, 3.7696E+04, 3.7719E+04,
     +  3.7741E+04, 3.7763E+04, 3.7785E+04, 3.7807E+04, 3.7830E+04,
     + 3.7852E+04, 3.7874E+04, 3.7896E+04, 3.7919E+04, 3.7941E+04,
     +  3.7963E+04, 3.7986E+04, 3.8008E+04, 3.8030E+04, 3.8053E+04,
     + 3.8075E+04, 3.8097E+04, 3.8120E+04, 3.8142E+04, 3.8164E+04,
     +  3.8187E+04, 3.8209E+04, 3.8231E+04, 3.8254E+04, 3.8276E+04,
     + 3.8298E+04, 3.8321E+04, 3.8343E+04, 3.8366E+04, 3.8388E+04,
     +  3.8411E+04, 3.8433E+04, 3.8455E+04, 3.8478E+04, 3.8500E+04,
     + 3.8523E+04, 3.8545E+04, 3.8568E+04, 3.8590E+04, 3.8613E+04,
     +  3.8635E+04, 3.8658E+04, 3.8680E+04, 3.8703E+04, 3.8725E+04,
     + 3.8748E+04, 3.8770E+04, 3.8793E+04, 3.8815E+04, 3.8838E+04,
     +  3.8861E+04, 3.8883E+04, 3.8906E+04, 3.8928E+04, 3.8951E+04,
     + 3.8973E+04, 3.8996E+04, 3.9019E+04, 3.9041E+04, 3.9064E+04,
     +  3.9087E+04, 3.9109E+04, 3.9132E+04, 3.9155E+04, 3.9177E+04,
     + 3.9200E+04, 3.9223E+04, 3.9245E+04, 3.9268E+04, 3.9291E+04,
     +  3.9313E+04, 3.9336E+04, 3.9359E+04, 3.9382E+04, 3.9404E+04,
     + 3.9427E+04, 3.9450E+04, 3.9473E+04, 3.9495E+04, 3.9518E+04,
     +  3.9541E+04, 3.9564E+04, 3.9586E+04, 3.9609E+04, 3.9632E+04,
     + 3.9655E+04, 3.9678E+04, 3.9700E+04, 3.9723E+04, 3.9746E+04,
     +  3.9769E+04, 3.9792E+04, 3.9815E+04, 3.9838E+04, 3.9860E+04,
     + 3.9883E+04, 3.9906E+04, 3.9929E+04, 3.9952E+04, 3.9975E+04,
     +  3.9998E+04, 4.0021E+04, 4.0044E+04, 4.0067E+04, 4.0090E+04,
     + 4.0113E+04, 4.0136E+04, 4.0159E+04, 4.0181E+04, 4.0204E+04,
     +  4.0227E+04, 4.0250E+04, 4.0273E+04, 4.0296E+04, 4.0319E+04,
     + 4.0343E+04, 4.0366E+04, 4.0389E+04, 4.0412E+04, 4.0435E+04,
     +  4.0458E+04, 4.0481E+04, 4.0504E+04, 4.0527E+04, 4.0550E+04,
     + 4.0573E+04, 4.0596E+04, 4.0619E+04, 4.0642E+04, 4.0666E+04,
     +  4.0689E+04, 4.0712E+04, 4.0735E+04, 4.0758E+04, 4.0781E+04,
     + 4.0805E+04, 4.0828E+04, 4.0851E+04, 4.0874E+04, 4.0897E+04,
     +  4.0920E+04, 4.0944E+04, 4.0967E+04, 4.0990E+04, 4.1013E+04,
     + 4.1037E+04, 4.1060E+04, 4.1083E+04, 4.1106E+04, 4.1130E+04,
     +  4.1153E+04, 4.1176E+04, 4.1199E+04, 4.1223E+04, 4.1246E+04,
     + 4.1269E+04, 4.1293E+04, 4.1316E+04, 4.1339E+04, 4.1363E+04,
     +  4.1386E+04, 4.1409E+04, 4.1433E+04, 4.1456E+04, 4.1479E+04,
     + 4.1503E+04, 4.1526E+04, 4.1550E+04, 4.1573E+04, 4.1596E+04,
     +  4.1620E+04, 4.1643E+04, 4.1667E+04, 4.1690E+04, 4.1714E+04,
     + 4.1737E+04, 4.1760E+04, 4.1784E+04, 4.1807E+04, 4.1831E+04,
     +  4.1854E+04, 4.1878E+04, 4.1901E+04, 4.1925E+04, 4.1948E+04,
     + 4.1972E+04, 4.1995E+04, 4.2019E+04, 4.2042E+04, 4.2066E+04,
     +  4.2090E+04, 4.2113E+04, 4.2137E+04, 4.2160E+04, 4.2184E+04,
     + 4.2207E+04, 4.2231E+04, 4.2255E+04, 4.2278E+04, 4.2302E+04,
     +  4.2326E+04, 4.2349E+04, 4.2373E+04, 4.2396E+04, 4.2420E+04,
     + 4.2444E+04, 4.2467E+04, 4.2491E+04, 4.2515E+04, 4.2538E+04,
     +  4.2562E+04, 4.2586E+04, 4.2610E+04, 4.2633E+04, 4.2657E+04,
     + 4.2681E+04, 4.2705E+04, 4.2728E+04, 4.2752E+04, 4.2776E+04,
     +  4.2800E+04, 4.2823E+04, 4.2847E+04, 4.2871E+04, 4.2895E+04,
     + 4.2919E+04, 4.2942E+04, 4.2966E+04, 4.2990E+04, 4.3014E+04,
     +  4.3038E+04, 4.3061E+04, 4.3085E+04, 4.3109E+04, 4.3133E+04,
     + 4.3157E+04, 4.3181E+04, 4.3205E+04, 4.3229E+04, 4.3253E+04,
     +  4.3276E+04, 4.3300E+04, 4.3324E+04, 4.3348E+04, 4.3372E+04,
     + 4.3396E+04, 4.3420E+04, 4.3444E+04, 4.3468E+04, 4.3492E+04,
     +  4.3516E+04, 4.3540E+04, 4.3564E+04, 4.3588E+04, 4.3612E+04,
     + 4.3636E+04, 4.3660E+04, 4.3684E+04, 4.3708E+04, 4.3732E+04,
     +  4.3756E+04, 4.3780E+04, 4.3804E+04, 4.3828E+04, 4.3852E+04,
     + 4.3876E+04, 4.3901E+04, 4.3925E+04, 4.3949E+04, 4.3973E+04,
     +  4.3997E+04, 4.4021E+04, 4.4045E+04, 4.4069E+04, 4.4094E+04,
     + 4.4118E+04, 4.4142E+04, 4.4166E+04, 4.4190E+04, 4.4214E+04,
     +  4.4239E+04, 4.4263E+04, 4.4287E+04, 4.4311E+04, 4.4336E+04,
     + 4.4360E+04, 4.4384E+04, 4.4408E+04, 4.4432E+04, 4.4457E+04,
     +  4.4481E+04, 4.4505E+04, 4.4530E+04, 4.4554E+04, 4.4578E+04,
     + 4.4602E+04, 4.4627E+04, 4.4651E+04, 4.4675E+04, 4.4700E+04,
     +  4.4724E+04, 4.4748E+04, 4.4773E+04, 4.4797E+04, 4.4821E+04,
     + 4.4846E+04, 4.4870E+04, 4.4895E+04, 4.4919E+04, 4.4943E+04,
     +  4.4968E+04, 4.4992E+04, 4.5017E+04, 4.5041E+04, 4.5066E+04,
     + 4.5090E+04, 4.5114E+04, 4.5139E+04, 4.5163E+04, 4.5188E+04,
     +  4.5212E+04, 4.5237E+04, 4.5261E+04, 4.5286E+04, 4.5310E+04,
     + 4.5335E+04, 4.5359E+04, 4.5384E+04, 4.5408E+04, 4.5433E+04,
     +  4.5457E+04, 4.5482E+04, 4.5507E+04, 4.5531E+04, 4.5556E+04,
     + 4.5580E+04, 4.5605E+04, 4.5630E+04, 4.5654E+04, 4.5679E+04,
     +  4.5703E+04, 4.5728E+04, 4.5753E+04, 4.5777E+04, 4.5802E+04,
     + 4.5827E+04, 4.5851E+04, 4.5876E+04, 4.5901E+04, 4.5925E+04,
     +  4.5950E+04, 4.5975E+04, 4.5999E+04, 4.6024E+04, 4.6049E+04,
     + 4.6074E+04, 4.6098E+04, 4.6123E+04, 4.6148E+04, 4.6173E+04,
     +  4.6197E+04, 4.6222E+04, 4.6247E+04, 4.6272E+04, 4.6296E+04,
     + 4.6321E+04, 4.6346E+04, 4.6371E+04, 4.6396E+04, 4.6421E+04,
     +  4.6445E+04, 4.6470E+04, 4.6495E+04, 4.6520E+04, 4.6545E+04,
     + 4.6570E+04, 4.6595E+04, 4.6619E+04, 4.6644E+04, 4.6669E+04,
     +  4.6694E+04, 4.6719E+04, 4.6744E+04, 4.6769E+04, 4.6794E+04,
     + 4.6819E+04, 4.6844E+04, 4.6869E+04, 4.6894E+04, 4.6919E+04,
     +  4.6944E+04, 4.6969E+04, 4.6994E+04, 4.7019E+04, 4.7044E+04,
     + 4.7069E+04, 4.7094E+04, 4.7119E+04, 4.7144E+04, 4.7169E+04,
     +  4.7194E+04, 4.7219E+04, 4.7244E+04, 4.7269E+04, 4.7294E+04,
     + 4.7319E+04, 4.7344E+04, 4.7369E+04, 4.7394E+04, 4.7420E+04,
     +  4.7445E+04, 4.7470E+04, 4.7495E+04, 4.7520E+04, 4.7545E+04,
     + 4.7570E+04, 4.7596E+04, 4.7621E+04, 4.7646E+04, 4.7671E+04,
     +  4.7696E+04, 4.7722E+04, 4.7747E+04, 4.7772E+04, 4.7797E+04,
     + 4.7822E+04, 4.7848E+04, 4.7873E+04, 4.7898E+04, 4.7923E+04,
     +  4.7949E+04, 4.7974E+04, 4.7999E+04, 4.8024E+04, 4.8050E+04,
     + 4.8075E+04, 4.8100E+04, 4.8126E+04, 4.8151E+04, 4.8176E+04,
     +  4.8202E+04, 4.8227E+04, 4.8252E+04, 4.8278E+04, 4.8303E+04,
     + 4.8328E+04, 4.8354E+04, 4.8379E+04, 4.8405E+04, 4.8430E+04,
     +  4.8455E+04, 4.8481E+04, 4.8506E+04, 4.8532E+04, 4.8557E+04,
     + 4.8583E+04, 4.8608E+04, 4.8634E+04, 4.8659E+04, 4.8684E+04,
     +  4.8710E+04, 4.8735E+04, 4.8761E+04, 4.8786E+04, 4.8812E+04,
     + 4.8837E+04, 4.8863E+04, 4.8888E+04, 4.8914E+04, 4.8940E+04,
     +  4.8965E+04, 4.8991E+04, 4.9016E+04, 4.9042E+04, 4.9067E+04,
     + 4.9093E+04, 4.9119E+04, 4.9144E+04, 4.9170E+04, 4.9195E+04,
     +  4.9221E+04, 4.9247E+04, 4.9272E+04, 4.9298E+04, 4.9324E+04,
     + 4.9349E+04, 4.9375E+04, 4.9401E+04, 4.9426E+04, 4.9452E+04,
     +  4.9478E+04, 4.9503E+04, 4.9529E+04, 4.9555E+04, 4.9581E+04,
     + 4.9606E+04, 4.9632E+04, 4.9658E+04, 4.9683E+04, 4.9709E+04,
     +  4.9735E+04, 4.9761E+04, 4.9787E+04, 4.9812E+04, 4.9838E+04,
     + 4.9864E+04, 4.9890E+04, 4.9916E+04, 4.9941E+04, 4.9967E+04,
     +  4.9993E+04, 5.0019E+04, 5.0045E+04, 5.0071E+04, 5.0096E+04,
     + 5.0122E+04, 5.0148E+04, 5.0174E+04, 5.0200E+04, 5.0226E+04,
     +  5.0252E+04, 5.0278E+04, 5.0304E+04, 5.0330E+04, 5.0356E+04,
     + 5.0382E+04, 5.0407E+04, 5.0433E+04, 5.0459E+04, 5.0485E+04,
     +  5.0511E+04, 5.0537E+04, 5.0563E+04, 5.0589E+04, 5.0615E+04,
     + 5.0641E+04, 5.0667E+04, 5.0693E+04, 5.0719E+04, 5.0746E+04,
     +  5.0772E+04, 5.0798E+04, 5.0824E+04, 5.0850E+04, 5.0876E+04,
     + 5.0902E+04, 5.0928E+04, 5.0954E+04, 5.0980E+04, 5.1006E+04,
     +  5.1033E+04, 5.1059E+04, 5.1085E+04, 5.1111E+04, 5.1137E+04,
     + 5.1163E+04, 5.1189E+04, 5.1216E+04, 5.1242E+04, 5.1268E+04,
     +  5.1294E+04, 5.1320E+04, 5.1347E+04, 5.1373E+04, 5.1399E+04,
     + 5.1425E+04, 5.1452E+04, 5.1478E+04, 5.1504E+04, 5.1530E+04,
     +  5.1557E+04, 5.1583E+04, 5.1609E+04, 5.1635E+04, 5.1662E+04,
     + 5.1688E+04, 5.1714E+04, 5.1741E+04, 5.1767E+04, 5.1793E+04,
     +  5.1820E+04, 5.1846E+04, 5.1872E+04, 5.1899E+04, 5.1925E+04,
     + 5.1951E+04, 5.1978E+04, 5.2004E+04, 5.2031E+04, 5.2057E+04,
     +  5.2083E+04, 5.2110E+04, 5.2136E+04, 5.2163E+04, 5.2189E+04,
     + 5.2216E+04, 5.2242E+04, 5.2269E+04, 5.2295E+04, 5.2322E+04,
     +  5.2348E+04, 5.2374E+04, 5.2401E+04, 5.2427E+04, 5.2454E+04,
     + 5.2481E+04, 5.2507E+04, 5.2534E+04, 5.2560E+04, 5.2587E+04,
     +  5.2613E+04, 5.2640E+04, 5.2666E+04, 5.2693E+04, 5.2720E+04,
     + 5.2746E+04, 5.2773E+04, 5.2799E+04, 5.2826E+04, 5.2853E+04,
     +  5.2879E+04, 5.2906E+04, 5.2932E+04, 5.2959E+04, 5.2986E+04,
     + 5.3012E+04, 5.3039E+04, 5.3066E+04, 5.3093E+04, 5.3119E+04,
     +  5.3146E+04, 5.3173E+04, 5.3199E+04, 5.3226E+04, 5.3253E+04,
     + 5.3280E+04, 5.3306E+04, 5.3333E+04, 5.3360E+04, 5.3387E+04,
     +  5.3413E+04, 5.3440E+04, 5.3467E+04, 5.3494E+04, 5.3521E+04,
     + 5.3547E+04, 5.3574E+04, 5.3601E+04, 5.3628E+04, 5.3655E+04,
     +  5.3681E+04, 5.3708E+04, 5.3735E+04, 5.3762E+04, 5.3789E+04,
     + 5.3816E+04, 5.3843E+04, 5.3870E+04, 5.3897E+04, 5.3923E+04,
     +  5.3950E+04, 5.3977E+04, 5.4004E+04, 5.4031E+04, 5.4058E+04,
     + 5.4085E+04, 5.4112E+04, 5.4139E+04, 5.4166E+04, 5.4193E+04,
     +  5.4220E+04, 5.4247E+04, 5.4274E+04, 5.4301E+04, 5.4328E+04,
     + 5.4355E+04, 5.4382E+04, 5.4409E+04, 5.4436E+04, 5.4463E+04,
     +  5.4490E+04, 5.4517E+04, 5.4544E+04, 5.4571E+04, 5.4599E+04,
     + 5.4626E+04, 5.4653E+04, 5.4680E+04, 5.4707E+04, 5.4734E+04,
     +  5.4761E+04, 5.4788E+04, 5.4816E+04, 5.4843E+04, 5.4870E+04,
     + 5.4897E+04, 5.4924E+04, 5.4951E+04, 5.4979E+04, 5.5006E+04,
     +  5.5033E+04, 5.5060E+04, 5.5087E+04, 5.5115E+04, 5.5142E+04,
     + 5.5169E+04, 5.5196E+04, 5.5224E+04, 5.5251E+04, 5.5278E+04,
     +  5.5305E+04, 5.5333E+04, 5.5360E+04, 5.5387E+04, 5.5415E+04,
     + 5.5442E+04, 5.5469E+04, 5.5497E+04, 5.5524E+04, 5.5551E+04,
     +  5.5579E+04, 5.5606E+04, 5.5633E+04, 5.5661E+04, 5.5688E+04,
     + 5.5716E+04, 5.5743E+04, 5.5770E+04, 5.5798E+04, 5.5825E+04,
     +  5.5853E+04, 5.5880E+04, 5.5907E+04, 5.5935E+04, 5.5962E+04,
     + 5.5990E+04, 5.6017E+04, 5.6045E+04, 5.6072E+04, 5.6100E+04,
     +  5.6127E+04, 5.6155E+04, 5.6182E+04, 5.6210E+04, 5.6237E+04,
     + 5.6265E+04, 5.6292E+04, 5.6320E+04, 5.6347E+04, 5.6375E+04,
     +  5.6403E+04, 5.6430E+04, 5.6458E+04, 5.6485E+04, 5.6513E+04,
     + 5.6541E+04, 5.6568E+04, 5.6596E+04, 5.6623E+04, 5.6651E+04,
     +  5.6679E+04, 5.6706E+04, 5.6734E+04, 5.6762E+04, 5.6789E+04,
     + 5.6817E+04, 5.6845E+04, 5.6872E+04, 5.6900E+04, 5.6928E+04,
     +  5.6955E+04, 5.6983E+04, 5.7011E+04, 5.7039E+04, 5.7066E+04,
     + 5.7094E+04, 5.7122E+04, 5.7150E+04, 5.7177E+04, 5.7205E+04,
     +  5.7233E+04, 5.7261E+04, 5.7289E+04, 5.7316E+04, 5.7344E+04,
     + 5.7372E+04, 5.7400E+04, 5.7428E+04, 5.7456E+04, 5.7483E+04,
     +  5.7511E+04, 5.7539E+04, 5.7567E+04, 5.7595E+04, 5.7623E+04,
     + 5.7651E+04, 5.7679E+04, 5.7706E+04, 5.7734E+04, 5.7762E+04,
     +  5.7790E+04, 5.7818E+04, 5.7846E+04, 5.7874E+04, 5.7902E+04,
     + 5.7930E+04, 5.7958E+04, 5.7986E+04, 5.8014E+04, 5.8042E+04,
     +  5.8070E+04, 5.8098E+04, 5.8126E+04, 5.8154E+04, 5.8182E+04,
     + 5.8210E+04, 5.8238E+04, 5.8266E+04, 5.8294E+04, 5.8322E+04,
     +  5.8350E+04, 5.8378E+04, 5.8407E+04, 5.8435E+04, 5.8463E+04,
     + 5.8491E+04, 5.8519E+04, 5.8547E+04, 5.8575E+04, 5.8603E+04,
     +  5.8632E+04, 5.8660E+04, 5.8688E+04, 5.8716E+04, 5.8744E+04,
     + 5.8772E+04, 5.8801E+04, 5.8829E+04, 5.8857E+04, 5.8885E+04,
     +  5.8913E+04, 5.8942E+04, 5.8970E+04, 5.8998E+04, 5.9026E+04,
     + 5.9055E+04, 5.9083E+04, 5.9111E+04, 5.9139E+04, 5.9168E+04,
     +  5.9196E+04, 5.9224E+04, 5.9253E+04, 5.9281E+04, 5.9309E+04,
     + 5.9338E+04, 5.9366E+04, 5.9394E+04, 5.9423E+04, 5.9451E+04,
     +  5.9479E+04, 5.9508E+04, 5.9536E+04, 5.9564E+04, 5.9593E+04,
     + 5.9621E+04, 5.9650E+04, 5.9678E+04, 5.9707E+04, 5.9735E+04,
     +  5.9763E+04, 5.9792E+04, 5.9820E+04, 5.9849E+04, 5.9877E+04,
     + 5.9906E+04, 5.9934E+04, 5.9963E+04, 5.9991E+04, 6.0020E+04,
     +  6.0048E+04, 6.0077E+04, 6.0105E+04, 6.0134E+04, 6.0162E+04,
     + 6.0191E+04, 6.0219E+04, 6.0248E+04, 6.0277E+04, 6.0305E+04,
     +  6.0334E+04, 6.0362E+04, 6.0391E+04, 6.0419E+04, 6.0448E+04,
     + 6.0477E+04, 6.0505E+04, 6.0534E+04, 6.0563E+04, 6.0591E+04,
     +  6.0620E+04, 6.0649E+04, 6.0677E+04, 6.0706E+04, 6.0735E+04,
     + 6.0763E+04, 6.0792E+04, 6.0821E+04, 6.0849E+04, 6.0878E+04,
     +  6.0907E+04, 6.0936E+04, 6.0964E+04, 6.0993E+04, 6.1022E+04,
     + 6.1051E+04, 6.1079E+04, 6.1108E+04, 6.1137E+04, 6.1166E+04,
     +  6.1195E+04, 6.1223E+04, 6.1252E+04, 6.1281E+04, 6.1310E+04,
     + 6.1339E+04, 6.1368E+04, 6.1396E+04, 6.1425E+04, 6.1454E+04,
     +  6.1483E+04, 6.1512E+04, 6.1541E+04, 6.1570E+04, 6.1599E+04,
     + 6.1627E+04, 6.1656E+04, 6.1685E+04, 6.1714E+04, 6.1743E+04,
     +  6.1772E+04, 6.1801E+04, 6.1830E+04, 6.1859E+04, 6.1888E+04,
     + 6.1917E+04, 6.1946E+04, 6.1975E+04, 6.2004E+04, 6.2033E+04,
     +  6.2062E+04, 6.2091E+04, 6.2120E+04, 6.2149E+04, 6.2178E+04,
     + 6.2207E+04, 6.2236E+04, 6.2265E+04, 6.2294E+04, 6.2324E+04,
     +  6.2353E+04, 6.2382E+04, 6.2411E+04, 6.2440E+04, 6.2469E+04,
     + 6.2498E+04, 6.2527E+04, 6.2556E+04, 6.2586E+04, 6.2615E+04,
     +  6.2644E+04, 6.2673E+04, 6.2702E+04, 6.2731E+04, 6.2761E+04,
     + 6.2790E+04, 6.2819E+04, 6.2848E+04, 6.2878E+04, 6.2907E+04,
     +  6.2936E+04, 6.2965E+04, 6.2994E+04, 6.3024E+04, 6.3053E+04,
     + 6.3082E+04, 6.3112E+04, 6.3141E+04, 6.3170E+04, 6.3199E+04,
     +  6.3229E+04, 6.3258E+04, 6.3287E+04, 6.3317E+04, 6.3346E+04,
     + 6.3375E+04, 6.3405E+04, 6.3434E+04, 6.3463E+04, 6.3493E+04,
     +  6.3522E+04, 6.3552E+04, 6.3581E+04, 6.3610E+04, 6.3640E+04,
     + 6.3669E+04, 6.3699E+04, 6.3728E+04, 6.3758E+04, 6.3787E+04,
     +  6.3816E+04, 6.3846E+04, 6.3875E+04, 6.3905E+04, 6.3934E+04,
     + 6.3964E+04, 6.3993E+04, 6.4023E+04, 6.4052E+04, 6.4082E+04,
     +  6.4111E+04, 6.4141E+04, 6.4171E+04, 6.4200E+04, 6.4230E+04,
     + 6.4259E+04, 6.4289E+04, 6.4318E+04, 6.4348E+04, 6.4378E+04,
     +  6.4407E+04, 6.4437E+04, 6.4466E+04, 6.4496E+04, 6.4526E+04,
     + 6.4555E+04, 6.4585E+04, 6.4615E+04, 6.4644E+04, 6.4674E+04,
     +  6.4704E+04, 6.4733E+04, 6.4763E+04, 6.4793E+04, 6.4822E+04,
     + 6.4852E+04, 6.4882E+04, 6.4911E+04, 6.4941E+04, 6.4971E+04,
     +  6.5001E+04, 6.5030E+04, 6.5060E+04, 6.5090E+04, 6.5120E+04,
     + 6.5150E+04, 6.5179E+04, 6.5209E+04, 6.5239E+04, 6.5269E+04,
     +  6.5299E+04, 6.5328E+04, 6.5358E+04, 6.5388E+04, 6.5418E+04,
     + 6.5448E+04, 6.5478E+04, 6.5508E+04, 6.5537E+04, 6.5567E+04,
     +  6.5597E+04, 6.5627E+04, 6.5657E+04, 6.5687E+04, 6.5717E+04,
     + 6.5747E+04, 6.5777E+04, 6.5807E+04, 6.5837E+04, 6.5867E+04,
     +  6.5896E+04, 6.5926E+04, 6.5956E+04, 6.5986E+04, 6.6016E+04,
     + 6.6046E+04, 6.6076E+04, 6.6106E+04, 6.6136E+04, 6.6167E+04,
     +  6.6197E+04, 6.6227E+04, 6.6257E+04, 6.6287E+04, 6.6317E+04,
     + 6.6347E+04, 6.6377E+04, 6.6407E+04, 6.6437E+04, 6.6467E+04,
     +  6.6497E+04, 6.6527E+04, 6.6558E+04, 6.6588E+04, 6.6618E+04,
     + 6.6648E+04, 6.6678E+04, 6.6708E+04, 6.6739E+04, 6.6769E+04,
     +  6.6799E+04, 6.6829E+04, 6.6859E+04, 6.6889E+04, 6.6920E+04,
     + 6.6950E+04, 6.6980E+04, 6.7010E+04, 6.7041E+04, 6.7071E+04,
     +  6.7101E+04, 6.7131E+04, 6.7162E+04, 6.7192E+04, 6.7222E+04,
     + 6.7253E+04, 6.7283E+04, 6.7313E+04, 6.7343E+04, 6.7374E+04,
     +  6.7404E+04, 6.7434E+04, 6.7465E+04, 6.7495E+04, 6.7525E+04,
     + 6.7556E+04, 6.7586E+04, 6.7617E+04, 6.7647E+04, 6.7677E+04,
     +  6.7708E+04, 6.7738E+04, 6.7769E+04, 6.7799E+04, 6.7829E+04,
     + 6.7860E+04, 6.7890E+04, 6.7921E+04, 6.7951E+04, 6.7982E+04,
     +  6.8012E+04, 6.8043E+04, 6.8073E+04, 6.8104E+04, 6.8134E+04,
     + 6.8165E+04, 6.8195E+04, 6.8226E+04, 6.8256E+04, 6.8287E+04,
     +  6.8317E+04, 6.8348E+04, 6.8379E+04, 6.8409E+04, 6.8440E+04,
     + 6.8470E+04, 6.8501E+04, 6.8531E+04, 6.8562E+04, 6.8593E+04,
     +  6.8623E+04, 6.8654E+04, 6.8685E+04, 6.8715E+04, 6.8746E+04,
     + 6.8777E+04, 6.8807E+04, 6.8838E+04, 6.8869E+04, 6.8899E+04,
     +  6.8930E+04, 6.8961E+04, 6.8991E+04, 6.9022E+04, 6.9053E+04,
     + 6.9083E+04, 6.9114E+04, 6.9145E+04, 6.9176E+04, 6.9206E+04,
     +  6.9237E+04, 6.9268E+04, 6.9299E+04, 6.9330E+04, 6.9360E+04,
     + 6.9391E+04, 6.9422E+04, 6.9453E+04, 6.9484E+04, 6.9515E+04,
     +  6.9545E+04, 6.9576E+04, 6.9607E+04, 6.9638E+04, 6.9669E+04,
     + 6.9700E+04, 6.9731E+04, 6.9761E+04, 6.9792E+04, 6.9823E+04,
     +  6.9854E+04, 6.9885E+04, 6.9916E+04, 6.9947E+04, 6.9978E+04,
     + 7.0009E+04, 7.0040E+04, 7.0071E+04, 7.0102E+04, 7.0133E+04,
     +  7.0164E+04, 7.0195E+04, 7.0226E+04, 7.0257E+04, 7.0288E+04,
     + 7.0319E+04, 7.0350E+04, 7.0381E+04, 7.0412E+04, 7.0443E+04,
     +  7.0474E+04, 7.0505E+04, 7.0536E+04, 7.0567E+04, 7.0598E+04,
     + 7.0629E+04, 7.0660E+04, 7.0692E+04, 7.0723E+04, 7.0754E+04,
     +  7.0785E+04, 7.0816E+04, 7.0847E+04, 7.0878E+04, 7.0910E+04,
     + 7.0941E+04, 7.0972E+04, 7.1003E+04, 7.1034E+04, 7.1066E+04,
     +  7.1097E+04, 7.1128E+04, 7.1159E+04, 7.1190E+04, 7.1222E+04,
     + 7.1253E+04, 7.1284E+04, 7.1315E+04, 7.1347E+04, 7.1378E+04,
     +  7.1409E+04, 7.1440E+04, 7.1472E+04, 7.1503E+04, 7.1534E+04,
     + 7.1566E+04, 7.1597E+04, 7.1628E+04, 7.1660E+04, 7.1691E+04,
     +  7.1722E+04, 7.1754E+04, 7.1785E+04, 7.1816E+04, 7.1848E+04,
     + 7.1879E+04, 7.1911E+04, 7.1942E+04, 7.1973E+04, 7.2005E+04,
     +  7.2036E+04, 7.2068E+04, 7.2099E+04, 7.2131E+04, 7.2162E+04,
     + 7.2194E+04, 7.2225E+04, 7.2256E+04, 7.2288E+04, 7.2319E+04,
     +  7.2351E+04, 7.2382E+04, 7.2414E+04, 7.2445E+04, 7.2477E+04,
     + 7.2509E+04, 7.2540E+04, 7.2572E+04, 7.2603E+04, 7.2635E+04,
     +  7.2666E+04, 7.2698E+04, 7.2730E+04, 7.2761E+04, 7.2793E+04,
     + 7.2824E+04, 7.2856E+04, 7.2888E+04, 7.2919E+04, 7.2951E+04,
     +  7.2982E+04, 7.3014E+04, 7.3046E+04, 7.3077E+04, 7.3109E+04,
     + 7.3141E+04, 7.3172E+04, 7.3204E+04, 7.3236E+04, 7.3268E+04,
     +  7.3299E+04, 7.3331E+04, 7.3363E+04, 7.3394E+04, 7.3426E+04,
     + 7.3458E+04, 7.3490E+04, 7.3522E+04, 7.3553E+04, 7.3585E+04,
     +  7.3617E+04, 7.3649E+04, 7.3680E+04, 7.3712E+04, 7.3744E+04,
     + 7.3776E+04, 7.3808E+04, 7.3840E+04, 7.3871E+04, 7.3903E+04,
     +  7.3935E+04, 7.3967E+04, 7.3999E+04, 7.4031E+04, 7.4063E+04,
     + 7.4095E+04, 7.4127E+04, 7.4158E+04, 7.4190E+04, 7.4222E+04,
     +  7.4254E+04, 7.4286E+04, 7.4318E+04, 7.4350E+04, 7.4382E+04,
     + 7.4414E+04, 7.4446E+04, 7.4478E+04, 7.4510E+04, 7.4542E+04,
     +  7.4574E+04, 7.4606E+04, 7.4638E+04, 7.4670E+04, 7.4702E+04,
     + 7.4734E+04, 7.4766E+04, 7.4798E+04, 7.4830E+04, 7.4862E+04,
     +  7.4895E+04, 7.4927E+04, 7.4959E+04, 7.4991E+04, 7.5023E+04,
     + 7.5055E+04, 7.5087E+04, 7.5119E+04, 7.5151E+04/
c...        --        38
      data (QofT( 5,J),J=1,2408)/ 2.0394E+00, 2.4916E+00, 3.1900E+00,
     +  3.9464E+00, 4.7209E+00, 5.5033E+00, 6.2899E+00, 7.0790E+00,
     + 7.8698E+00, 8.6616E+00, 9.4543E+00, 1.0247E+01, 1.1041E+01,
     +  1.1835E+01, 1.2629E+01, 1.3424E+01, 1.4219E+01, 1.5014E+01,
     + 1.5809E+01, 1.6604E+01, 1.8194E+01, 1.9785E+01, 2.1376E+01,
     +  2.2967E+01, 2.4558E+01, 2.6150E+01, 2.7741E+01, 2.9333E+01,
     + 3.0924E+01, 3.2516E+01, 3.4108E+01, 3.5700E+01, 3.7292E+01,
     +  3.8884E+01, 4.0476E+01, 4.2068E+01, 4.3660E+01, 4.5252E+01,
     + 4.6844E+01, 4.8436E+01, 5.0028E+01, 5.1621E+01, 5.3213E+01,
     +  5.4805E+01, 5.6397E+01, 5.7990E+01, 5.9582E+01, 6.1174E+01,
     + 6.2767E+01, 6.4359E+01, 6.5951E+01, 6.7544E+01, 6.9136E+01,
     +  7.0729E+01, 7.2321E+01, 7.3914E+01, 7.5506E+01, 7.7099E+01,
     + 7.8691E+01, 8.0284E+01, 8.1876E+01, 8.3469E+01, 8.5062E+01,
     +  8.6654E+01, 8.8247E+01, 8.9840E+01, 9.1432E+01, 9.3025E+01,
     + 9.4618E+01, 9.6210E+01, 9.7803E+01, 9.9396E+01, 1.0099E+02,
     +  1.0258E+02, 1.0417E+02, 1.0577E+02, 1.0736E+02, 1.0895E+02,
     + 1.1055E+02, 1.1214E+02, 1.1373E+02, 1.1532E+02, 1.1692E+02,
     +  1.1851E+02, 1.2010E+02, 1.2170E+02, 1.2329E+02, 1.2488E+02,
     + 1.2648E+02, 1.2807E+02, 1.2966E+02, 1.3125E+02, 1.3285E+02,
     +  1.3444E+02, 1.3603E+02, 1.3763E+02, 1.3922E+02, 1.4081E+02,
     + 1.4241E+02, 1.4400E+02, 1.4559E+02, 1.4719E+02, 1.4878E+02,
     +  1.5037E+02, 1.5197E+02, 1.5356E+02, 1.5515E+02, 1.5675E+02,
     + 1.5834E+02, 1.5993E+02, 1.6153E+02, 1.6312E+02, 1.6471E+02,
     +  1.6631E+02, 1.6790E+02, 1.6949E+02, 1.7109E+02, 1.7268E+02,
     + 1.7427E+02, 1.7587E+02, 1.7746E+02, 1.7905E+02, 1.8065E+02,
     +  1.8224E+02, 1.8384E+02, 1.8543E+02, 1.8702E+02, 1.8862E+02,
     + 1.9021E+02, 1.9180E+02, 1.9340E+02, 1.9499E+02, 1.9659E+02,
     +  1.9818E+02, 1.9977E+02, 2.0137E+02, 2.0296E+02, 2.0455E+02,
     + 2.0615E+02, 2.0774E+02, 2.0934E+02, 2.1093E+02, 2.1252E+02,
     +  2.1412E+02, 2.1571E+02, 2.1731E+02, 2.1890E+02, 2.2050E+02,
     + 2.2209E+02, 2.2369E+02, 2.2528E+02, 2.2687E+02, 2.2847E+02,
     +  2.3006E+02, 2.3166E+02, 2.3325E+02, 2.3485E+02, 2.3644E+02,
     + 2.3804E+02, 2.3963E+02, 2.4123E+02, 2.4282E+02, 2.4442E+02,
     +  2.4601E+02, 2.4761E+02, 2.4920E+02, 2.5080E+02, 2.5240E+02,
     + 2.5399E+02, 2.5559E+02, 2.5718E+02, 2.5878E+02, 2.6038E+02,
     +  2.6197E+02, 2.6357E+02, 2.6516E+02, 2.6676E+02, 2.6836E+02,
     + 2.6996E+02, 2.7155E+02, 2.7315E+02, 2.7475E+02, 2.7634E+02,
     +  2.7794E+02, 2.7954E+02, 2.8114E+02, 2.8274E+02, 2.8433E+02,
     + 2.8593E+02, 2.8753E+02, 2.8913E+02, 2.9073E+02, 2.9233E+02,
     +  2.9393E+02, 2.9553E+02, 2.9713E+02, 2.9873E+02, 3.0033E+02,
     + 3.0193E+02, 3.0353E+02, 3.0513E+02, 3.0673E+02, 3.0833E+02,
     +  3.0993E+02, 3.1154E+02, 3.1314E+02, 3.1474E+02, 3.1634E+02,
     + 3.1795E+02, 3.1955E+02, 3.2115E+02, 3.2276E+02, 3.2436E+02,
     +  3.2597E+02, 3.2757E+02, 3.2918E+02, 3.3078E+02, 3.3239E+02,
     + 3.3399E+02, 3.3560E+02, 3.3721E+02, 3.3881E+02, 3.4042E+02,
     +  3.4203E+02, 3.4364E+02, 3.4525E+02, 3.4686E+02, 3.4847E+02,
     + 3.5008E+02, 3.5169E+02, 3.5330E+02, 3.5491E+02, 3.5652E+02,
     +  3.5813E+02, 3.5975E+02, 3.6136E+02, 3.6297E+02, 3.6459E+02,
     + 3.6620E+02, 3.6782E+02, 3.6943E+02, 3.7105E+02, 3.7266E+02,
     +  3.7428E+02, 3.7590E+02, 3.7752E+02, 3.7914E+02, 3.8076E+02,
     + 3.8237E+02, 3.8400E+02, 3.8562E+02, 3.8724E+02, 3.8886E+02,
     +  3.9048E+02, 3.9211E+02, 3.9373E+02, 3.9535E+02, 3.9698E+02,
     + 3.9860E+02, 4.0023E+02, 4.0186E+02, 4.0349E+02, 4.0511E+02,
     +  4.0674E+02, 4.0837E+02, 4.1000E+02, 4.1163E+02, 4.1327E+02,
     + 4.1490E+02, 4.1653E+02, 4.1817E+02, 4.1980E+02, 4.2144E+02,
     +  4.2307E+02, 4.2471E+02, 4.2635E+02, 4.2798E+02, 4.2962E+02,
     + 4.3126E+02, 4.3290E+02, 4.3455E+02, 4.3619E+02, 4.3783E+02,
     +  4.3947E+02, 4.4112E+02, 4.4276E+02, 4.4441E+02, 4.4606E+02,
     + 4.4771E+02, 4.4936E+02, 4.5101E+02, 4.5266E+02, 4.5431E+02,
     +  4.5596E+02, 4.5761E+02, 4.5927E+02, 4.6092E+02, 4.6258E+02,
     + 4.6424E+02, 4.6589E+02, 4.6755E+02, 4.6921E+02, 4.7087E+02,
     +  4.7253E+02, 4.7420E+02, 4.7586E+02, 4.7753E+02, 4.7919E+02,
     + 4.8086E+02, 4.8253E+02, 4.8419E+02, 4.8586E+02, 4.8753E+02,
     +  4.8921E+02, 4.9088E+02, 4.9255E+02, 4.9423E+02, 4.9590E+02,
     + 4.9758E+02, 4.9926E+02, 5.0094E+02, 5.0262E+02, 5.0430E+02,
     +  5.0598E+02, 5.0766E+02, 5.0935E+02, 5.1103E+02, 5.1272E+02,
     + 5.1441E+02, 5.1610E+02, 5.1779E+02, 5.1948E+02, 5.2117E+02,
     +  5.2286E+02, 5.2456E+02, 5.2625E+02, 5.2795E+02, 5.2965E+02,
     + 5.3135E+02, 5.3305E+02, 5.3475E+02, 5.3645E+02, 5.3816E+02,
     +  5.3986E+02, 5.4157E+02, 5.4328E+02, 5.4499E+02, 5.4670E+02,
     + 5.4841E+02, 5.5012E+02, 5.5184E+02, 5.5355E+02, 5.5527E+02,
     +  5.5699E+02, 5.5870E+02, 5.6042E+02, 5.6215E+02, 5.6387E+02,
     + 5.6559E+02, 5.6732E+02, 5.6905E+02, 5.7077E+02, 5.7250E+02,
     +  5.7424E+02, 5.7597E+02, 5.7770E+02, 5.7944E+02, 5.8117E+02,
     + 5.8291E+02, 5.8465E+02, 5.8639E+02, 5.8813E+02, 5.8987E+02,
     +  5.9162E+02, 5.9336E+02, 5.9511E+02, 5.9686E+02, 5.9861E+02,
     + 6.0036E+02, 6.0211E+02, 6.0387E+02, 6.0562E+02, 6.0738E+02,
     +  6.0914E+02, 6.1090E+02, 6.1266E+02, 6.1442E+02, 6.1619E+02,
     + 6.1795E+02, 6.1972E+02, 6.2149E+02, 6.2326E+02, 6.2503E+02,
     +  6.2680E+02, 6.2858E+02, 6.3036E+02, 6.3213E+02, 6.3391E+02,
     + 6.3569E+02, 6.3747E+02, 6.3926E+02, 6.4104E+02, 6.4283E+02,
     +  6.4462E+02, 6.4641E+02, 6.4820E+02, 6.4999E+02, 6.5178E+02,
     + 6.5358E+02, 6.5538E+02, 6.5718E+02, 6.5898E+02, 6.6078E+02,
     +  6.6258E+02, 6.6439E+02, 6.6619E+02, 6.6800E+02, 6.6981E+02,
     + 6.7162E+02, 6.7344E+02, 6.7525E+02, 6.7707E+02, 6.7888E+02,
     +  6.8070E+02, 6.8253E+02, 6.8435E+02, 6.8617E+02, 6.8800E+02,
     + 6.8982E+02, 6.9165E+02, 6.9348E+02, 6.9532E+02, 6.9715E+02,
     +  6.9899E+02, 7.0082E+02, 7.0266E+02, 7.0450E+02, 7.0634E+02,
     + 7.0819E+02, 7.1003E+02, 7.1188E+02, 7.1373E+02, 7.1558E+02,
     +  7.1743E+02, 7.1929E+02, 7.2114E+02, 7.2300E+02, 7.2486E+02,
     + 7.2672E+02, 7.2858E+02, 7.3044E+02, 7.3231E+02, 7.3418E+02,
     +  7.3605E+02, 7.3792E+02, 7.3979E+02, 7.4166E+02, 7.4354E+02,
     + 7.4542E+02, 7.4730E+02, 7.4918E+02, 7.5106E+02, 7.5294E+02,
     +  7.5483E+02, 7.5672E+02, 7.5861E+02, 7.6050E+02, 7.6239E+02,
     + 7.6429E+02, 7.6618E+02, 7.6808E+02, 7.6998E+02, 7.7188E+02,
     +  7.7379E+02, 7.7569E+02, 7.7760E+02, 7.7951E+02, 7.8142E+02,
     + 7.8333E+02, 7.8525E+02, 7.8716E+02, 7.8908E+02, 7.9100E+02,
     +  7.9292E+02, 7.9485E+02, 7.9677E+02, 7.9870E+02, 8.0063E+02,
     + 8.0256E+02, 8.0449E+02, 8.0642E+02, 8.0836E+02, 8.1030E+02,
     +  8.1224E+02, 8.1418E+02, 8.1612E+02, 8.1807E+02, 8.2001E+02,
     + 8.2196E+02, 8.2391E+02, 8.2587E+02, 8.2782E+02, 8.2978E+02,
     +  8.3173E+02, 8.3369E+02, 8.3565E+02, 8.3762E+02, 8.3958E+02,
     + 8.4155E+02, 8.4352E+02, 8.4549E+02, 8.4746E+02, 8.4944E+02,
     +  8.5141E+02, 8.5339E+02, 8.5537E+02, 8.5735E+02, 8.5934E+02,
     + 8.6132E+02, 8.6331E+02, 8.6530E+02, 8.6729E+02, 8.6928E+02,
     +  8.7128E+02, 8.7328E+02, 8.7528E+02, 8.7728E+02, 8.7928E+02,
     + 8.8128E+02, 8.8329E+02, 8.8530E+02, 8.8731E+02, 8.8932E+02,
     +  8.9134E+02, 8.9335E+02, 8.9537E+02, 8.9739E+02, 8.9941E+02,
     + 9.0143E+02, 9.0346E+02, 9.0549E+02, 9.0752E+02, 9.0955E+02,
     +  9.1158E+02, 9.1362E+02, 9.1565E+02, 9.1769E+02, 9.1973E+02,
     + 9.2178E+02, 9.2382E+02, 9.2587E+02, 9.2792E+02, 9.2997E+02,
     +  9.3202E+02, 9.3407E+02, 9.3613E+02, 9.3819E+02, 9.4025E+02,
     + 9.4231E+02, 9.4438E+02, 9.4644E+02, 9.4851E+02, 9.5058E+02,
     +  9.5265E+02, 9.5473E+02, 9.5680E+02, 9.5888E+02, 9.6096E+02,
     + 9.6304E+02, 9.6513E+02, 9.6721E+02, 9.6930E+02, 9.7139E+02,
     +  9.7348E+02, 9.7557E+02, 9.7767E+02, 9.7977E+02, 9.8187E+02,
     + 9.8397E+02, 9.8607E+02, 9.8818E+02, 9.9028E+02, 9.9239E+02,
     +  9.9451E+02, 9.9662E+02, 9.9873E+02, 1.0009E+03, 1.0030E+03,
     + 1.0051E+03, 1.0072E+03, 1.0093E+03, 1.0115E+03, 1.0136E+03,
     +  1.0157E+03, 1.0179E+03, 1.0200E+03, 1.0221E+03, 1.0243E+03,
     + 1.0264E+03, 1.0286E+03, 1.0307E+03, 1.0329E+03, 1.0350E+03,
     +  1.0372E+03, 1.0393E+03, 1.0415E+03, 1.0436E+03, 1.0458E+03,
     + 1.0479E+03, 1.0501E+03, 1.0523E+03, 1.0544E+03, 1.0566E+03,
     +  1.0588E+03, 1.0610E+03, 1.0631E+03, 1.0653E+03, 1.0675E+03,
     + 1.0697E+03, 1.0719E+03, 1.0740E+03, 1.0762E+03, 1.0784E+03,
     +  1.0806E+03, 1.0828E+03, 1.0850E+03, 1.0872E+03, 1.0894E+03,
     + 1.0916E+03, 1.0938E+03, 1.0960E+03, 1.0982E+03, 1.1004E+03,
     +  1.1026E+03, 1.1049E+03, 1.1071E+03, 1.1093E+03, 1.1115E+03,
     + 1.1137E+03, 1.1160E+03, 1.1182E+03, 1.1204E+03, 1.1227E+03,
     +  1.1249E+03, 1.1271E+03, 1.1294E+03, 1.1316E+03, 1.1338E+03,
     + 1.1361E+03, 1.1383E+03, 1.1406E+03, 1.1428E+03, 1.1451E+03,
     +  1.1473E+03, 1.1496E+03, 1.1518E+03, 1.1541E+03, 1.1564E+03,
     + 1.1586E+03, 1.1609E+03, 1.1632E+03, 1.1654E+03, 1.1677E+03,
     +  1.1700E+03, 1.1723E+03, 1.1745E+03, 1.1768E+03, 1.1791E+03,
     + 1.1814E+03, 1.1837E+03, 1.1860E+03, 1.1883E+03, 1.1905E+03,
     +  1.1928E+03, 1.1951E+03, 1.1974E+03, 1.1997E+03, 1.2020E+03,
     + 1.2043E+03, 1.2067E+03, 1.2090E+03, 1.2113E+03, 1.2136E+03,
     +  1.2159E+03, 1.2182E+03, 1.2205E+03, 1.2229E+03, 1.2252E+03,
     + 1.2275E+03, 1.2299E+03, 1.2322E+03, 1.2345E+03, 1.2369E+03,
     +  1.2392E+03, 1.2415E+03, 1.2439E+03, 1.2462E+03, 1.2486E+03,
     + 1.2509E+03, 1.2533E+03, 1.2556E+03, 1.2580E+03, 1.2603E+03,
     +  1.2627E+03, 1.2650E+03, 1.2674E+03, 1.2698E+03, 1.2721E+03,
     + 1.2745E+03, 1.2769E+03, 1.2792E+03, 1.2816E+03, 1.2840E+03,
     +  1.2864E+03, 1.2888E+03, 1.2911E+03, 1.2935E+03, 1.2959E+03,
     + 1.2983E+03, 1.3007E+03, 1.3031E+03, 1.3055E+03, 1.3079E+03,
     +  1.3103E+03, 1.3127E+03, 1.3151E+03, 1.3175E+03, 1.3199E+03,
     + 1.3223E+03, 1.3247E+03, 1.3272E+03, 1.3296E+03, 1.3320E+03,
     +  1.3344E+03, 1.3368E+03, 1.3393E+03, 1.3417E+03, 1.3441E+03,
     + 1.3466E+03, 1.3490E+03, 1.3514E+03, 1.3539E+03, 1.3563E+03,
     +  1.3588E+03, 1.3612E+03, 1.3637E+03, 1.3661E+03, 1.3686E+03,
     + 1.3710E+03, 1.3735E+03, 1.3759E+03, 1.3784E+03, 1.3809E+03,
     +  1.3833E+03, 1.3858E+03, 1.3883E+03, 1.3907E+03, 1.3932E+03,
     + 1.3957E+03, 1.3982E+03, 1.4006E+03, 1.4031E+03, 1.4056E+03,
     +  1.4081E+03, 1.4106E+03, 1.4131E+03, 1.4156E+03, 1.4181E+03,
     + 1.4206E+03, 1.4231E+03, 1.4256E+03, 1.4281E+03, 1.4306E+03,
     +  1.4331E+03, 1.4356E+03, 1.4381E+03, 1.4406E+03, 1.4431E+03,
     + 1.4456E+03, 1.4482E+03, 1.4507E+03, 1.4532E+03, 1.4557E+03,
     +  1.4583E+03, 1.4608E+03, 1.4633E+03, 1.4659E+03, 1.4684E+03,
     + 1.4710E+03, 1.4735E+03, 1.4760E+03, 1.4786E+03, 1.4811E+03,
     +  1.4837E+03, 1.4862E+03, 1.4888E+03, 1.4914E+03, 1.4939E+03,
     + 1.4965E+03, 1.4990E+03, 1.5016E+03, 1.5042E+03, 1.5068E+03,
     +  1.5093E+03, 1.5119E+03, 1.5145E+03, 1.5171E+03, 1.5196E+03,
     + 1.5222E+03, 1.5248E+03, 1.5274E+03, 1.5300E+03, 1.5326E+03,
     +  1.5352E+03, 1.5378E+03, 1.5404E+03, 1.5430E+03, 1.5456E+03,
     + 1.5482E+03, 1.5508E+03, 1.5534E+03, 1.5560E+03, 1.5586E+03,
     +  1.5613E+03, 1.5639E+03, 1.5665E+03, 1.5691E+03, 1.5717E+03,
     + 1.5744E+03, 1.5770E+03, 1.5796E+03, 1.5823E+03, 1.5849E+03,
     +  1.5875E+03, 1.5902E+03, 1.5928E+03, 1.5955E+03, 1.5981E+03,
     + 1.6008E+03, 1.6034E+03, 1.6061E+03, 1.6087E+03, 1.6114E+03,
     +  1.6141E+03, 1.6167E+03, 1.6194E+03, 1.6220E+03, 1.6247E+03,
     + 1.6274E+03, 1.6301E+03, 1.6327E+03, 1.6354E+03, 1.6381E+03,
     +  1.6408E+03, 1.6435E+03, 1.6462E+03, 1.6488E+03, 1.6515E+03,
     + 1.6542E+03, 1.6569E+03, 1.6596E+03, 1.6623E+03, 1.6650E+03,
     +  1.6677E+03, 1.6704E+03, 1.6731E+03, 1.6759E+03, 1.6786E+03,
     + 1.6813E+03, 1.6840E+03, 1.6867E+03, 1.6894E+03, 1.6922E+03,
     +  1.6949E+03, 1.6976E+03, 1.7004E+03, 1.7031E+03, 1.7058E+03,
     + 1.7086E+03, 1.7113E+03, 1.7141E+03, 1.7168E+03, 1.7195E+03,
     +  1.7223E+03, 1.7250E+03, 1.7278E+03, 1.7306E+03, 1.7333E+03,
     + 1.7361E+03, 1.7388E+03, 1.7416E+03, 1.7444E+03, 1.7471E+03,
     +  1.7499E+03, 1.7527E+03, 1.7554E+03, 1.7582E+03, 1.7610E+03,
     + 1.7638E+03, 1.7666E+03, 1.7694E+03, 1.7721E+03, 1.7749E+03,
     +  1.7777E+03, 1.7805E+03, 1.7833E+03, 1.7861E+03, 1.7889E+03,
     + 1.7917E+03, 1.7945E+03, 1.7973E+03, 1.8002E+03, 1.8030E+03,
     +  1.8058E+03, 1.8086E+03, 1.8114E+03, 1.8142E+03, 1.8171E+03,
     + 1.8199E+03, 1.8227E+03, 1.8256E+03, 1.8284E+03, 1.8312E+03,
     +  1.8341E+03, 1.8369E+03, 1.8397E+03, 1.8426E+03, 1.8454E+03,
     + 1.8483E+03, 1.8511E+03, 1.8540E+03, 1.8568E+03, 1.8597E+03,
     +  1.8626E+03, 1.8654E+03, 1.8683E+03, 1.8712E+03, 1.8740E+03,
     + 1.8769E+03, 1.8798E+03, 1.8826E+03, 1.8855E+03, 1.8884E+03,
     +  1.8913E+03, 1.8942E+03, 1.8971E+03, 1.8999E+03, 1.9028E+03,
     + 1.9057E+03, 1.9086E+03, 1.9115E+03, 1.9144E+03, 1.9173E+03,
     +  1.9202E+03, 1.9231E+03, 1.9260E+03, 1.9290E+03, 1.9319E+03,
     + 1.9348E+03, 1.9377E+03, 1.9406E+03, 1.9435E+03, 1.9465E+03,
     +  1.9494E+03, 1.9523E+03, 1.9553E+03, 1.9582E+03, 1.9611E+03,
     + 1.9641E+03, 1.9670E+03, 1.9699E+03, 1.9729E+03, 1.9758E+03,
     +  1.9788E+03, 1.9817E+03, 1.9847E+03, 1.9877E+03, 1.9906E+03,
     + 1.9936E+03, 1.9965E+03, 1.9995E+03, 2.0025E+03, 2.0054E+03,
     +  2.0084E+03, 2.0114E+03, 2.0144E+03, 2.0173E+03, 2.0203E+03,
     + 2.0233E+03, 2.0263E+03, 2.0293E+03, 2.0323E+03, 2.0353E+03,
     +  2.0382E+03, 2.0412E+03, 2.0442E+03, 2.0472E+03, 2.0502E+03,
     + 2.0533E+03, 2.0563E+03, 2.0593E+03, 2.0623E+03, 2.0653E+03,
     +  2.0683E+03, 2.0713E+03, 2.0744E+03, 2.0774E+03, 2.0804E+03,
     + 2.0834E+03, 2.0865E+03, 2.0895E+03, 2.0925E+03, 2.0956E+03,
     +  2.0986E+03, 2.1016E+03, 2.1047E+03, 2.1077E+03, 2.1108E+03,
     + 2.1138E+03, 2.1169E+03, 2.1199E+03, 2.1230E+03, 2.1261E+03,
     +  2.1291E+03, 2.1322E+03, 2.1352E+03, 2.1383E+03, 2.1414E+03,
     + 2.1445E+03, 2.1475E+03, 2.1506E+03, 2.1537E+03, 2.1568E+03,
     +  2.1599E+03, 2.1629E+03, 2.1660E+03, 2.1691E+03, 2.1722E+03,
     + 2.1753E+03, 2.1784E+03, 2.1815E+03, 2.1846E+03, 2.1877E+03,
     +  2.1908E+03, 2.1939E+03, 2.1970E+03, 2.2002E+03, 2.2033E+03,
     + 2.2064E+03, 2.2095E+03, 2.2126E+03, 2.2158E+03, 2.2189E+03,
     +  2.2220E+03, 2.2252E+03, 2.2283E+03, 2.2314E+03, 2.2346E+03,
     + 2.2377E+03, 2.2408E+03, 2.2440E+03, 2.2471E+03, 2.2503E+03,
     +  2.2534E+03, 2.2566E+03, 2.2598E+03, 2.2629E+03, 2.2661E+03,
     + 2.2692E+03, 2.2724E+03, 2.2756E+03, 2.2787E+03, 2.2819E+03,
     +  2.2851E+03, 2.2883E+03, 2.2914E+03, 2.2946E+03, 2.2978E+03,
     + 2.3010E+03, 2.3042E+03, 2.3074E+03, 2.3106E+03, 2.3138E+03,
     +  2.3170E+03, 2.3202E+03, 2.3234E+03, 2.3266E+03, 2.3298E+03,
     + 2.3330E+03, 2.3362E+03, 2.3394E+03, 2.3426E+03, 2.3458E+03,
     +  2.3491E+03, 2.3523E+03, 2.3555E+03, 2.3587E+03, 2.3620E+03,
     + 2.3652E+03, 2.3684E+03, 2.3717E+03, 2.3749E+03, 2.3781E+03,
     +  2.3814E+03, 2.3846E+03, 2.3879E+03, 2.3911E+03, 2.3944E+03,
     + 2.3976E+03, 2.4009E+03, 2.4041E+03, 2.4074E+03, 2.4107E+03,
     +  2.4139E+03, 2.4172E+03, 2.4205E+03, 2.4237E+03, 2.4270E+03,
     + 2.4303E+03, 2.4336E+03, 2.4369E+03, 2.4401E+03, 2.4434E+03,
     +  2.4467E+03, 2.4500E+03, 2.4533E+03, 2.4566E+03, 2.4599E+03,
     + 2.4632E+03, 2.4665E+03, 2.4698E+03, 2.4731E+03, 2.4764E+03,
     +  2.4797E+03, 2.4830E+03, 2.4864E+03, 2.4897E+03, 2.4930E+03,
     + 2.4963E+03, 2.4996E+03, 2.5030E+03, 2.5063E+03, 2.5096E+03,
     +  2.5130E+03, 2.5163E+03, 2.5196E+03, 2.5230E+03, 2.5263E+03,
     + 2.5297E+03, 2.5330E+03, 2.5364E+03, 2.5397E+03, 2.5431E+03,
     +  2.5464E+03, 2.5498E+03, 2.5531E+03, 2.5565E+03, 2.5599E+03,
     + 2.5632E+03, 2.5666E+03, 2.5700E+03, 2.5734E+03, 2.5767E+03,
     +  2.5801E+03, 2.5835E+03, 2.5869E+03, 2.5903E+03, 2.5937E+03,
     + 2.5971E+03, 2.6004E+03, 2.6038E+03, 2.6072E+03, 2.6106E+03,
     +  2.6140E+03, 2.6175E+03, 2.6209E+03, 2.6243E+03, 2.6277E+03,
     + 2.6311E+03, 2.6345E+03, 2.6379E+03, 2.6413E+03, 2.6448E+03,
     +  2.6482E+03, 2.6516E+03, 2.6551E+03, 2.6585E+03, 2.6619E+03,
     + 2.6654E+03, 2.6688E+03, 2.6722E+03, 2.6757E+03, 2.6791E+03,
     +  2.6826E+03, 2.6860E+03, 2.6895E+03, 2.6929E+03, 2.6964E+03,
     + 2.6999E+03, 2.7033E+03, 2.7068E+03, 2.7102E+03, 2.7137E+03,
     +  2.7172E+03, 2.7207E+03, 2.7241E+03, 2.7276E+03, 2.7311E+03,
     + 2.7346E+03, 2.7381E+03, 2.7416E+03, 2.7450E+03, 2.7485E+03,
     +  2.7520E+03, 2.7555E+03, 2.7590E+03, 2.7625E+03, 2.7660E+03,
     + 2.7695E+03, 2.7730E+03, 2.7766E+03, 2.7801E+03, 2.7836E+03,
     +  2.7871E+03, 2.7906E+03, 2.7941E+03, 2.7977E+03, 2.8012E+03,
     + 2.8047E+03, 2.8082E+03, 2.8118E+03, 2.8153E+03, 2.8189E+03,
     +  2.8224E+03, 2.8259E+03, 2.8295E+03, 2.8330E+03, 2.8366E+03,
     + 2.8401E+03, 2.8437E+03, 2.8472E+03, 2.8508E+03, 2.8544E+03,
     +  2.8579E+03, 2.8615E+03, 2.8651E+03, 2.8686E+03, 2.8722E+03,
     + 2.8758E+03, 2.8794E+03, 2.8829E+03, 2.8865E+03, 2.8901E+03,
     +  2.8937E+03, 2.8973E+03, 2.9009E+03, 2.9045E+03, 2.9081E+03,
     + 2.9116E+03, 2.9152E+03, 2.9189E+03, 2.9225E+03, 2.9261E+03,
     +  2.9297E+03, 2.9333E+03, 2.9369E+03, 2.9405E+03, 2.9441E+03,
     + 2.9478E+03, 2.9514E+03, 2.9550E+03, 2.9586E+03, 2.9623E+03,
     +  2.9659E+03, 2.9695E+03, 2.9732E+03, 2.9768E+03, 2.9804E+03,
     + 2.9841E+03, 2.9877E+03, 2.9914E+03, 2.9950E+03, 2.9987E+03,
     +  3.0023E+03, 3.0060E+03, 3.0097E+03, 3.0133E+03, 3.0170E+03,
     + 3.0207E+03, 3.0243E+03, 3.0280E+03, 3.0317E+03, 3.0353E+03,
     +  3.0390E+03, 3.0427E+03, 3.0464E+03, 3.0501E+03, 3.0538E+03,
     + 3.0575E+03, 3.0611E+03, 3.0648E+03, 3.0685E+03, 3.0722E+03,
     +  3.0759E+03, 3.0796E+03, 3.0833E+03, 3.0871E+03, 3.0908E+03,
     + 3.0945E+03, 3.0982E+03, 3.1019E+03, 3.1056E+03, 3.1094E+03,
     +  3.1131E+03, 3.1168E+03, 3.1205E+03, 3.1243E+03, 3.1280E+03,
     + 3.1317E+03, 3.1355E+03, 3.1392E+03, 3.1430E+03, 3.1467E+03,
     +  3.1505E+03, 3.1542E+03, 3.1580E+03, 3.1617E+03, 3.1655E+03,
     + 3.1692E+03, 3.1730E+03, 3.1768E+03, 3.1805E+03, 3.1843E+03,
     +  3.1881E+03, 3.1918E+03, 3.1956E+03, 3.1994E+03, 3.2032E+03,
     + 3.2070E+03, 3.2107E+03, 3.2145E+03, 3.2183E+03, 3.2221E+03,
     +  3.2259E+03, 3.2297E+03, 3.2335E+03, 3.2373E+03, 3.2411E+03,
     + 3.2449E+03, 3.2487E+03, 3.2525E+03, 3.2563E+03, 3.2602E+03,
     +  3.2640E+03, 3.2678E+03, 3.2716E+03, 3.2754E+03, 3.2793E+03,
     + 3.2831E+03, 3.2869E+03, 3.2908E+03, 3.2946E+03, 3.2984E+03,
     +  3.3023E+03, 3.3061E+03, 3.3100E+03, 3.3138E+03, 3.3177E+03,
     + 3.3215E+03, 3.3254E+03, 3.3292E+03, 3.3331E+03, 3.3370E+03,
     +  3.3408E+03, 3.3447E+03, 3.3486E+03, 3.3524E+03, 3.3563E+03,
     + 3.3602E+03, 3.3641E+03, 3.3679E+03, 3.3718E+03, 3.3757E+03,
     +  3.3796E+03, 3.3835E+03, 3.3874E+03, 3.3913E+03, 3.3952E+03,
     + 3.3991E+03, 3.4030E+03, 3.4069E+03, 3.4108E+03, 3.4147E+03,
     +  3.4186E+03, 3.4225E+03, 3.4264E+03, 3.4303E+03, 3.4343E+03,
     + 3.4382E+03, 3.4421E+03, 3.4460E+03, 3.4500E+03, 3.4539E+03,
     +  3.4578E+03, 3.4618E+03, 3.4657E+03, 3.4696E+03, 3.4736E+03,
     + 3.4775E+03, 3.4815E+03, 3.4854E+03, 3.4894E+03, 3.4933E+03,
     +  3.4973E+03, 3.5013E+03, 3.5052E+03, 3.5092E+03, 3.5131E+03,
     + 3.5171E+03, 3.5211E+03, 3.5251E+03, 3.5290E+03, 3.5330E+03,
     +  3.5370E+03, 3.5410E+03, 3.5450E+03, 3.5490E+03, 3.5529E+03,
     + 3.5569E+03, 3.5609E+03, 3.5649E+03, 3.5689E+03, 3.5729E+03,
     +  3.5769E+03, 3.5809E+03, 3.5849E+03, 3.5890E+03, 3.5930E+03,
     + 3.5970E+03, 3.6010E+03, 3.6050E+03, 3.6091E+03, 3.6131E+03,
     +  3.6171E+03, 3.6211E+03, 3.6252E+03, 3.6292E+03, 3.6332E+03,
     + 3.6373E+03, 3.6413E+03, 3.6454E+03, 3.6494E+03, 3.6535E+03,
     +  3.6575E+03, 3.6616E+03, 3.6656E+03, 3.6697E+03, 3.6737E+03,
     + 3.6778E+03, 3.6819E+03, 3.6859E+03, 3.6900E+03, 3.6941E+03,
     +  3.6981E+03, 3.7022E+03, 3.7063E+03, 3.7104E+03, 3.7145E+03,
     + 3.7185E+03, 3.7226E+03, 3.7267E+03, 3.7308E+03, 3.7349E+03,
     +  3.7390E+03, 3.7431E+03, 3.7472E+03, 3.7513E+03, 3.7554E+03,
     + 3.7595E+03, 3.7636E+03, 3.7678E+03, 3.7719E+03, 3.7760E+03,
     +  3.7801E+03, 3.7842E+03, 3.7884E+03, 3.7925E+03, 3.7966E+03,
     + 3.8008E+03, 3.8049E+03, 3.8090E+03, 3.8132E+03, 3.8173E+03,
     +  3.8215E+03, 3.8256E+03, 3.8298E+03, 3.8339E+03, 3.8381E+03,
     + 3.8422E+03, 3.8464E+03, 3.8505E+03, 3.8547E+03, 3.8589E+03,
     +  3.8630E+03, 3.8672E+03, 3.8714E+03, 3.8755E+03, 3.8797E+03,
     + 3.8839E+03, 3.8881E+03, 3.8923E+03, 3.8965E+03, 3.9006E+03,
     +  3.9048E+03, 3.9090E+03, 3.9132E+03, 3.9174E+03, 3.9216E+03,
     + 3.9258E+03, 3.9300E+03, 3.9342E+03, 3.9385E+03, 3.9427E+03,
     +  3.9469E+03, 3.9511E+03, 3.9553E+03, 3.9595E+03, 3.9638E+03,
     + 3.9680E+03, 3.9722E+03, 3.9765E+03, 3.9807E+03, 3.9849E+03,
     +  3.9892E+03, 3.9934E+03, 3.9976E+03, 4.0019E+03, 4.0061E+03,
     + 4.0104E+03, 4.0146E+03, 4.0189E+03, 4.0232E+03, 4.0274E+03,
     +  4.0317E+03, 4.0359E+03, 4.0402E+03, 4.0445E+03, 4.0487E+03,
     + 4.0530E+03, 4.0573E+03, 4.0616E+03, 4.0659E+03, 4.0701E+03,
     +  4.0744E+03, 4.0787E+03, 4.0830E+03, 4.0873E+03, 4.0916E+03,
     + 4.0959E+03, 4.1002E+03, 4.1045E+03, 4.1088E+03, 4.1131E+03,
     +  4.1174E+03, 4.1217E+03, 4.1260E+03, 4.1304E+03, 4.1347E+03,
     + 4.1390E+03, 4.1433E+03, 4.1476E+03, 4.1520E+03, 4.1563E+03,
     +  4.1606E+03, 4.1650E+03, 4.1693E+03, 4.1736E+03, 4.1780E+03,
     + 4.1823E+03, 4.1867E+03, 4.1910E+03, 4.1954E+03, 4.1997E+03,
     +  4.2041E+03, 4.2085E+03, 4.2128E+03, 4.2172E+03, 4.2215E+03,
     + 4.2259E+03, 4.2303E+03, 4.2347E+03, 4.2390E+03, 4.2434E+03,
     +  4.2478E+03, 4.2522E+03, 4.2566E+03, 4.2609E+03, 4.2653E+03,
     + 4.2697E+03, 4.2741E+03, 4.2785E+03, 4.2829E+03, 4.2873E+03,
     +  4.2917E+03, 4.2961E+03, 4.3005E+03, 4.3049E+03, 4.3094E+03,
     + 4.3138E+03, 4.3182E+03, 4.3226E+03, 4.3270E+03, 4.3315E+03,
     +  4.3359E+03, 4.3403E+03, 4.3447E+03, 4.3492E+03, 4.3536E+03,
     + 4.3581E+03, 4.3625E+03, 4.3669E+03, 4.3714E+03, 4.3758E+03,
     +  4.3803E+03, 4.3847E+03, 4.3892E+03, 4.3937E+03, 4.3981E+03,
     + 4.4026E+03, 4.4070E+03, 4.4115E+03, 4.4160E+03, 4.4204E+03,
     +  4.4249E+03, 4.4294E+03, 4.4339E+03, 4.4384E+03, 4.4428E+03,
     + 4.4473E+03, 4.4518E+03, 4.4563E+03, 4.4608E+03, 4.4653E+03,
     +  4.4698E+03, 4.4743E+03, 4.4788E+03, 4.4833E+03, 4.4878E+03,
     + 4.4923E+03, 4.4968E+03, 4.5014E+03, 4.5059E+03, 4.5104E+03,
     +  4.5149E+03, 4.5194E+03, 4.5240E+03, 4.5285E+03, 4.5330E+03,
     + 4.5376E+03, 4.5421E+03, 4.5466E+03, 4.5512E+03, 4.5557E+03,
     +  4.5603E+03, 4.5648E+03, 4.5694E+03, 4.5739E+03, 4.5785E+03,
     + 4.5830E+03, 4.5876E+03, 4.5922E+03, 4.5967E+03, 4.6013E+03,
     +  4.6059E+03, 4.6104E+03, 4.6150E+03, 4.6196E+03, 4.6242E+03,
     + 4.6287E+03, 4.6333E+03, 4.6379E+03, 4.6425E+03, 4.6471E+03,
     +  4.6517E+03, 4.6563E+03, 4.6609E+03, 4.6655E+03, 4.6701E+03,
     + 4.6747E+03, 4.6793E+03, 4.6839E+03, 4.6885E+03, 4.6931E+03,
     +  4.6977E+03, 4.7024E+03, 4.7070E+03, 4.7116E+03, 4.7162E+03,
     + 4.7209E+03, 4.7255E+03, 4.7301E+03, 4.7348E+03, 4.7394E+03,
     +  4.7440E+03, 4.7487E+03, 4.7533E+03, 4.7580E+03, 4.7626E+03,
     + 4.7673E+03, 4.7719E+03, 4.7766E+03, 4.7813E+03, 4.7859E+03,
     +  4.7906E+03, 4.7952E+03, 4.7999E+03, 4.8046E+03, 4.8093E+03,
     + 4.8139E+03, 4.8186E+03, 4.8233E+03, 4.8280E+03, 4.8327E+03,
     +  4.8374E+03, 4.8421E+03, 4.8467E+03, 4.8514E+03, 4.8561E+03,
     + 4.8608E+03, 4.8655E+03, 4.8702E+03, 4.8750E+03, 4.8797E+03,
     +  4.8844E+03, 4.8891E+03, 4.8938E+03, 4.8985E+03, 4.9032E+03,
     + 4.9080E+03, 4.9127E+03, 4.9174E+03, 4.9222E+03, 4.9269E+03,
     +  4.9316E+03, 4.9364E+03, 4.9411E+03, 4.9459E+03, 4.9506E+03,
     + 4.9553E+03, 4.9601E+03, 4.9649E+03, 4.9696E+03, 4.9744E+03,
     +  4.9791E+03, 4.9839E+03, 4.9887E+03, 4.9934E+03, 4.9982E+03,
     + 5.0030E+03, 5.0077E+03, 5.0125E+03, 5.0173E+03, 5.0221E+03,
     +  5.0269E+03, 5.0316E+03, 5.0364E+03, 5.0412E+03, 5.0460E+03,
     + 5.0508E+03, 5.0556E+03, 5.0604E+03, 5.0652E+03, 5.0700E+03,
     +  5.0748E+03, 5.0796E+03, 5.0844E+03, 5.0893E+03, 5.0941E+03,
     + 5.0989E+03, 5.1037E+03, 5.1085E+03, 5.1134E+03, 5.1182E+03,
     +  5.1230E+03, 5.1279E+03, 5.1327E+03, 5.1375E+03, 5.1424E+03,
     + 5.1472E+03, 5.1521E+03, 5.1569E+03, 5.1618E+03, 5.1666E+03,
     +  5.1715E+03, 5.1763E+03, 5.1812E+03, 5.1861E+03, 5.1909E+03,
     + 5.1958E+03, 5.2007E+03, 5.2055E+03, 5.2104E+03, 5.2153E+03,
     +  5.2202E+03, 5.2251E+03, 5.2299E+03, 5.2348E+03, 5.2397E+03,
     + 5.2446E+03, 5.2495E+03, 5.2544E+03, 5.2593E+03, 5.2642E+03,
     +  5.2691E+03, 5.2740E+03, 5.2789E+03, 5.2838E+03, 5.2887E+03,
     + 5.2937E+03, 5.2986E+03, 5.3035E+03, 5.3084E+03, 5.3133E+03,
     +  5.3183E+03, 5.3232E+03, 5.3281E+03, 5.3331E+03, 5.3380E+03,
     + 5.3429E+03, 5.3479E+03, 5.3528E+03, 5.3578E+03, 5.3627E+03,
     +  5.3677E+03, 5.3726E+03, 5.3776E+03, 5.3826E+03, 5.3875E+03,
     + 5.3925E+03, 5.3974E+03, 5.4024E+03, 5.4074E+03, 5.4124E+03,
     +  5.4173E+03, 5.4223E+03, 5.4273E+03, 5.4323E+03, 5.4373E+03,
     + 5.4422E+03, 5.4472E+03, 5.4522E+03, 5.4572E+03, 5.4622E+03,
     +  5.4672E+03, 5.4722E+03, 5.4772E+03, 5.4822E+03, 5.4872E+03,
     + 5.4923E+03, 5.4973E+03, 5.5023E+03, 5.5073E+03, 5.5123E+03,
     +  5.5173E+03, 5.5224E+03, 5.5274E+03, 5.5324E+03, 5.5375E+03,
     + 5.5425E+03, 5.5475E+03, 5.5526E+03, 5.5576E+03, 5.5627E+03,
     +  5.5677E+03, 5.5728E+03, 5.5778E+03, 5.5829E+03, 5.5879E+03,
     + 5.5930E+03, 5.5981E+03, 5.6031E+03, 5.6082E+03, 5.6133E+03,
     +  5.6183E+03, 5.6234E+03, 5.6285E+03, 5.6336E+03, 5.6386E+03,
     + 5.6437E+03, 5.6488E+03, 5.6539E+03, 5.6590E+03, 5.6641E+03,
     +  5.6692E+03, 5.6743E+03, 5.6794E+03, 5.6845E+03, 5.6896E+03,
     + 5.6947E+03, 5.6998E+03, 5.7049E+03, 5.7100E+03, 5.7152E+03,
     +  5.7203E+03, 5.7254E+03, 5.7305E+03, 5.7356E+03, 5.7408E+03,
     + 5.7459E+03, 5.7510E+03, 5.7562E+03, 5.7613E+03, 5.7665E+03,
     +  5.7716E+03, 5.7768E+03, 5.7819E+03, 5.7871E+03, 5.7922E+03,
     + 5.7974E+03, 5.8025E+03, 5.8077E+03, 5.8128E+03, 5.8180E+03,
     +  5.8232E+03, 5.8284E+03, 5.8335E+03, 5.8387E+03, 5.8439E+03,
     + 5.8491E+03, 5.8542E+03, 5.8594E+03, 5.8646E+03, 5.8698E+03,
     +  5.8750E+03, 5.8802E+03, 5.8854E+03, 5.8906E+03, 5.8958E+03,
     + 5.9010E+03, 5.9062E+03, 5.9114E+03, 5.9166E+03, 5.9218E+03,
     +  5.9271E+03, 5.9323E+03, 5.9375E+03, 5.9427E+03, 5.9479E+03,
     + 5.9532E+03, 5.9584E+03, 5.9636E+03, 5.9689E+03, 5.9741E+03,
     +  5.9794E+03, 5.9846E+03, 5.9898E+03, 5.9951E+03, 6.0003E+03,
     + 6.0056E+03, 6.0109E+03, 6.0161E+03, 6.0214E+03, 6.0266E+03,
     +  6.0319E+03, 6.0372E+03, 6.0424E+03, 6.0477E+03, 6.0530E+03,
     + 6.0583E+03, 6.0635E+03, 6.0688E+03, 6.0741E+03, 6.0794E+03,
     +  6.0847E+03, 6.0900E+03, 6.0953E+03, 6.1006E+03, 6.1059E+03,
     + 6.1112E+03, 6.1165E+03, 6.1218E+03, 6.1271E+03, 6.1324E+03,
     +  6.1377E+03, 6.1430E+03, 6.1483E+03, 6.1537E+03, 6.1590E+03,
     + 6.1643E+03, 6.1696E+03, 6.1750E+03, 6.1803E+03, 6.1856E+03,
     +  6.1910E+03, 6.1963E+03, 6.2017E+03, 6.2070E+03, 6.2124E+03,
     + 6.2177E+03, 6.2231E+03, 6.2284E+03, 6.2338E+03, 6.2391E+03,
     +  6.2445E+03, 6.2499E+03, 6.2552E+03, 6.2606E+03, 6.2660E+03,
     + 6.2713E+03, 6.2767E+03, 6.2821E+03, 6.2875E+03, 6.2929E+03,
     +  6.2982E+03, 6.3036E+03, 6.3090E+03, 6.3144E+03, 6.3198E+03,
     + 6.3252E+03, 6.3306E+03, 6.3360E+03, 6.3414E+03, 6.3468E+03,
     +  6.3522E+03, 6.3577E+03, 6.3631E+03, 6.3685E+03, 6.3739E+03,
     + 6.3793E+03, 6.3848E+03, 6.3902E+03, 6.3956E+03, 6.4010E+03,
     +  6.4065E+03, 6.4119E+03, 6.4174E+03, 6.4228E+03, 6.4282E+03,
     + 6.4337E+03, 6.4391E+03, 6.4446E+03, 6.4501E+03, 6.4555E+03,
     +  6.4610E+03, 6.4664E+03, 6.4719E+03, 6.4774E+03, 6.4828E+03,
     + 6.4883E+03, 6.4938E+03, 6.4993E+03, 6.5047E+03, 6.5102E+03,
     +  6.5157E+03, 6.5212E+03, 6.5267E+03, 6.5322E+03, 6.5377E+03,
     + 6.5432E+03, 6.5486E+03, 6.5541E+03, 6.5597E+03, 6.5652E+03,
     +  6.5707E+03, 6.5762E+03, 6.5817E+03, 6.5872E+03, 6.5927E+03,
     + 6.5982E+03, 6.6038E+03, 6.6093E+03, 6.6148E+03, 6.6203E+03,
     +  6.6259E+03, 6.6314E+03, 6.6370E+03, 6.6425E+03, 6.6480E+03,
     + 6.6536E+03, 6.6591E+03, 6.6647E+03, 6.6702E+03, 6.6758E+03,
     +  6.6813E+03, 6.6869E+03, 6.6925E+03, 6.6980E+03, 6.7036E+03,
     + 6.7092E+03, 6.7147E+03, 6.7203E+03, 6.7259E+03, 6.7315E+03,
     +  6.7370E+03, 6.7426E+03, 6.7482E+03, 6.7538E+03, 6.7594E+03,
     + 6.7650E+03, 6.7706E+03, 6.7762E+03, 6.7818E+03, 6.7874E+03,
     +  6.7930E+03, 6.7986E+03, 6.8042E+03, 6.8098E+03, 6.8154E+03,
     + 6.8211E+03, 6.8267E+03, 6.8323E+03, 6.8379E+03, 6.8436E+03,
     +  6.8492E+03, 6.8548E+03, 6.8604E+03, 6.8661E+03, 6.8717E+03,
     + 6.8774E+03, 6.8830E+03, 6.8887E+03, 6.8943E+03, 6.9000E+03,
     +  6.9056E+03, 6.9113E+03, 6.9169E+03, 6.9226E+03, 6.9283E+03,
     + 6.9339E+03, 6.9396E+03, 6.9453E+03, 6.9509E+03, 6.9566E+03,
     +  6.9623E+03, 6.9680E+03, 6.9737E+03, 6.9793E+03, 6.9850E+03,
     + 6.9907E+03, 6.9964E+03, 7.0021E+03, 7.0078E+03, 7.0135E+03,
     +  7.0192E+03, 7.0249E+03, 7.0306E+03, 7.0363E+03, 7.0421E+03,
     + 7.0478E+03, 7.0535E+03, 7.0592E+03, 7.0649E+03, 7.0707E+03,
     +  7.0764E+03, 7.0821E+03, 7.0879E+03, 7.0936E+03, 7.0993E+03,
     + 7.1051E+03, 7.1108E+03, 7.1166E+03, 7.1223E+03, 7.1281E+03,
     +  7.1338E+03, 7.1396E+03, 7.1453E+03, 7.1511E+03, 7.1568E+03,
     + 7.1626E+03, 7.1684E+03, 7.1741E+03, 7.1799E+03, 7.1857E+03,
     +  7.1915E+03, 7.1972E+03, 7.2030E+03, 7.2088E+03, 7.2146E+03,
     + 7.2204E+03, 7.2262E+03, 7.2320E+03, 7.2378E+03, 7.2436E+03,
     +  7.2494E+03, 7.2552E+03, 7.2610E+03, 7.2668E+03, 7.2726E+03,
     + 7.2784E+03, 7.2842E+03, 7.2901E+03, 7.2959E+03, 7.3017E+03,
     +  7.3075E+03, 7.3134E+03, 7.3192E+03, 7.3250E+03, 7.3309E+03,
     + 7.3367E+03, 7.3425E+03, 7.3484E+03, 7.3542E+03, 7.3601E+03,
     +  7.3659E+03, 7.3718E+03, 7.3776E+03, 7.3835E+03, 7.3894E+03,
     + 7.3952E+03, 7.4011E+03, 7.4069E+03, 7.4128E+03, 7.4187E+03,
     +  7.4246E+03, 7.4304E+03, 7.4363E+03, 7.4422E+03, 7.4481E+03,
     + 7.4540E+03, 7.4599E+03, 7.4658E+03, 7.4717E+03, 7.4776E+03,
     +  7.4835E+03, 7.4894E+03, 7.4953E+03, 7.5012E+03, 7.5071E+03,
     + 7.5130E+03, 7.5189E+03, 7.5248E+03, 7.5307E+03, 7.5367E+03,
     +  7.5426E+03, 7.5485E+03, 7.5544E+03, 7.5604E+03, 7.5663E+03,
     + 7.5723E+03, 7.5782E+03, 7.5841E+03, 7.5901E+03, 7.5960E+03,
     +  7.6020E+03, 7.6079E+03, 7.6139E+03, 7.6198E+03, 7.6258E+03,
     + 7.6318E+03, 7.6377E+03, 7.6437E+03, 7.6497E+03, 7.6556E+03,
     +  7.6616E+03, 7.6676E+03, 7.6736E+03, 7.6795E+03, 7.6855E+03,
     + 7.6915E+03, 7.6975E+03, 7.7035E+03, 7.7095E+03, 7.7155E+03,
     +  7.7215E+03, 7.7275E+03, 7.7335E+03, 7.7395E+03, 7.7455E+03,
     + 7.7515E+03, 7.7575E+03, 7.7635E+03, 7.7696E+03, 7.7756E+03,
     +  7.7816E+03, 7.7876E+03, 7.7937E+03, 7.7997E+03, 7.8057E+03,
     + 7.8118E+03, 7.8178E+03, 7.8238E+03, 7.8299E+03, 7.8359E+03,
     +  7.8420E+03, 7.8480E+03, 7.8541E+03, 7.8601E+03, 7.8662E+03,
     + 7.8722E+03, 7.8783E+03, 7.8844E+03, 7.8904E+03, 7.8965E+03,
     +  7.9026E+03, 7.9087E+03, 7.9147E+03, 7.9208E+03, 7.9269E+03,
     + 7.9330E+03, 7.9391E+03, 7.9452E+03, 7.9513E+03, 7.9574E+03,
     +  7.9634E+03, 7.9695E+03, 7.9756E+03, 7.9818E+03, 7.9879E+03,
     + 7.9940E+03, 8.0001E+03, 8.0062E+03, 8.0123E+03, 8.0184E+03,
     +  8.0246E+03, 8.0307E+03, 8.0368E+03, 8.0429E+03, 8.0491E+03,
     + 8.0552E+03, 8.0613E+03, 8.0675E+03, 8.0736E+03, 8.0798E+03,
     +  8.0859E+03, 8.0921E+03, 8.0982E+03, 8.1044E+03, 8.1105E+03,
     + 8.1167E+03, 8.1229E+03, 8.1290E+03, 8.1352E+03, 8.1414E+03,
     +  8.1475E+03, 8.1537E+03, 8.1599E+03, 8.1661E+03, 8.1722E+03,
     + 8.1784E+03, 8.1846E+03, 8.1908E+03, 8.1970E+03, 8.2032E+03,
     +  8.2094E+03, 8.2156E+03, 8.2218E+03, 8.2280E+03, 8.2342E+03,
     + 8.2404E+03, 8.2466E+03, 8.2528E+03, 8.2590E+03, 8.2653E+03,
     +  8.2715E+03, 8.2777E+03, 8.2839E+03, 8.2902E+03, 8.2964E+03,
     + 8.3026E+03, 8.3089E+03, 8.3151E+03, 8.3213E+03, 8.3276E+03,
     +  8.3338E+03, 8.3401E+03, 8.3463E+03, 8.3526E+03, 8.3588E+03,
     + 8.3651E+03, 8.3714E+03, 8.3776E+03, 8.3839E+03, 8.3902E+03,
     +  8.3964E+03, 8.4027E+03, 8.4090E+03, 8.4153E+03, 8.4215E+03,
     + 8.4278E+03, 8.4341E+03, 8.4404E+03, 8.4467E+03, 8.4530E+03,
     +  8.4593E+03, 8.4656E+03, 8.4719E+03, 8.4782E+03, 8.4845E+03,
     + 8.4908E+03, 8.4971E+03, 8.5034E+03, 8.5097E+03, 8.5161E+03,
     +  8.5224E+03, 8.5287E+03, 8.5350E+03, 8.5413E+03, 8.5477E+03,
     + 8.5540E+03, 8.5603E+03, 8.5667E+03, 8.5730E+03, 8.5794E+03,
     +  8.5857E+03, 8.5921E+03, 8.5984E+03, 8.6048E+03, 8.6111E+03/
      data (QofT( 5,J),J=2409,4510)/ 8.6175E+03, 8.6238E+03,
     +  8.6302E+03, 8.6366E+03, 8.6429E+03, 8.6493E+03, 8.6557E+03,
     +  8.6621E+03,
     + 8.6684E+03, 8.6748E+03, 8.6812E+03, 8.6876E+03, 8.6940E+03,
     +  8.7004E+03, 8.7068E+03, 8.7132E+03, 8.7196E+03, 8.7260E+03,
     + 8.7324E+03, 8.7388E+03, 8.7452E+03, 8.7516E+03, 8.7580E+03,
     +  8.7644E+03, 8.7708E+03, 8.7772E+03, 8.7837E+03, 8.7901E+03,
     + 8.7965E+03, 8.8030E+03, 8.8094E+03, 8.8158E+03, 8.8223E+03,
     +  8.8287E+03, 8.8351E+03, 8.8416E+03, 8.8480E+03, 8.8545E+03,
     + 8.8609E+03, 8.8674E+03, 8.8739E+03, 8.8803E+03, 8.8868E+03,
     +  8.8932E+03, 8.8997E+03, 8.9062E+03, 8.9127E+03, 8.9191E+03,
     + 8.9256E+03, 8.9321E+03, 8.9386E+03, 8.9451E+03, 8.9515E+03,
     +  8.9580E+03, 8.9645E+03, 8.9710E+03, 8.9775E+03, 8.9840E+03,
     + 8.9905E+03, 8.9970E+03, 9.0035E+03, 9.0100E+03, 9.0166E+03,
     +  9.0231E+03, 9.0296E+03, 9.0361E+03, 9.0426E+03, 9.0492E+03,
     + 9.0557E+03, 9.0622E+03, 9.0688E+03, 9.0753E+03, 9.0818E+03,
     +  9.0884E+03, 9.0949E+03, 9.1015E+03, 9.1080E+03, 9.1145E+03,
     + 9.1211E+03, 9.1277E+03, 9.1342E+03, 9.1408E+03, 9.1473E+03,
     +  9.1539E+03, 9.1605E+03, 9.1670E+03, 9.1736E+03, 9.1802E+03,
     + 9.1868E+03, 9.1933E+03, 9.1999E+03, 9.2065E+03, 9.2131E+03,
     +  9.2197E+03, 9.2263E+03, 9.2329E+03, 9.2395E+03, 9.2461E+03,
     + 9.2527E+03, 9.2593E+03, 9.2659E+03, 9.2725E+03, 9.2791E+03,
     +  9.2857E+03, 9.2924E+03, 9.2990E+03, 9.3056E+03, 9.3122E+03,
     + 9.3189E+03, 9.3255E+03, 9.3321E+03, 9.3388E+03, 9.3454E+03,
     +  9.3520E+03, 9.3587E+03, 9.3653E+03, 9.3720E+03, 9.3786E+03,
     + 9.3853E+03, 9.3919E+03, 9.3986E+03, 9.4052E+03, 9.4119E+03,
     +  9.4186E+03, 9.4252E+03, 9.4319E+03, 9.4386E+03, 9.4453E+03,
     + 9.4519E+03, 9.4586E+03, 9.4653E+03, 9.4720E+03, 9.4787E+03,
     +  9.4854E+03, 9.4921E+03, 9.4988E+03, 9.5055E+03, 9.5122E+03,
     + 9.5189E+03, 9.5256E+03, 9.5323E+03, 9.5390E+03, 9.5457E+03,
     +  9.5524E+03, 9.5591E+03, 9.5659E+03, 9.5726E+03, 9.5793E+03,
     + 9.5860E+03, 9.5928E+03, 9.5995E+03, 9.6062E+03, 9.6130E+03,
     +  9.6197E+03, 9.6265E+03, 9.6332E+03, 9.6399E+03, 9.6467E+03,
     + 9.6535E+03, 9.6602E+03, 9.6670E+03, 9.6737E+03, 9.6805E+03,
     +  9.6873E+03, 9.6940E+03, 9.7008E+03, 9.7076E+03, 9.7143E+03,
     + 9.7211E+03, 9.7279E+03, 9.7347E+03, 9.7415E+03, 9.7483E+03,
     +  9.7551E+03, 9.7618E+03, 9.7686E+03, 9.7754E+03, 9.7822E+03,
     + 9.7890E+03, 9.7959E+03, 9.8027E+03, 9.8095E+03, 9.8163E+03,
     +  9.8231E+03, 9.8299E+03, 9.8367E+03, 9.8436E+03, 9.8504E+03,
     + 9.8572E+03, 9.8641E+03, 9.8709E+03, 9.8777E+03, 9.8846E+03,
     +  9.8914E+03, 9.8983E+03, 9.9051E+03, 9.9119E+03, 9.9188E+03,
     + 9.9257E+03, 9.9325E+03, 9.9394E+03, 9.9462E+03, 9.9531E+03,
     +  9.9600E+03, 9.9668E+03, 9.9737E+03, 9.9806E+03, 9.9875E+03,
     + 9.9943E+03, 1.0001E+04, 1.0008E+04, 1.0015E+04, 1.0022E+04,
     +  1.0029E+04, 1.0036E+04, 1.0043E+04, 1.0049E+04, 1.0056E+04,
     + 1.0063E+04, 1.0070E+04, 1.0077E+04, 1.0084E+04, 1.0091E+04,
     +  1.0098E+04, 1.0105E+04, 1.0112E+04, 1.0119E+04, 1.0126E+04,
     + 1.0132E+04, 1.0139E+04, 1.0146E+04, 1.0153E+04, 1.0160E+04,
     +  1.0167E+04, 1.0174E+04, 1.0181E+04, 1.0188E+04, 1.0195E+04,
     + 1.0202E+04, 1.0209E+04, 1.0216E+04, 1.0223E+04, 1.0230E+04,
     +  1.0237E+04, 1.0244E+04, 1.0251E+04, 1.0258E+04, 1.0265E+04,
     + 1.0272E+04, 1.0279E+04, 1.0286E+04, 1.0293E+04, 1.0300E+04,
     +  1.0307E+04, 1.0314E+04, 1.0321E+04, 1.0328E+04, 1.0335E+04,
     + 1.0342E+04, 1.0349E+04, 1.0356E+04, 1.0363E+04, 1.0370E+04,
     +  1.0377E+04, 1.0384E+04, 1.0391E+04, 1.0398E+04, 1.0405E+04,
     + 1.0412E+04, 1.0419E+04, 1.0426E+04, 1.0433E+04, 1.0440E+04,
     +  1.0447E+04, 1.0454E+04, 1.0461E+04, 1.0468E+04, 1.0475E+04,
     + 1.0482E+04, 1.0489E+04, 1.0496E+04, 1.0503E+04, 1.0510E+04,
     +  1.0517E+04, 1.0525E+04, 1.0532E+04, 1.0539E+04, 1.0546E+04,
     + 1.0553E+04, 1.0560E+04, 1.0567E+04, 1.0574E+04, 1.0581E+04,
     +  1.0588E+04, 1.0595E+04, 1.0603E+04, 1.0610E+04, 1.0617E+04,
     + 1.0624E+04, 1.0631E+04, 1.0638E+04, 1.0645E+04, 1.0652E+04,
     +  1.0659E+04, 1.0667E+04, 1.0674E+04, 1.0681E+04, 1.0688E+04,
     + 1.0695E+04, 1.0702E+04, 1.0709E+04, 1.0716E+04, 1.0724E+04,
     +  1.0731E+04, 1.0738E+04, 1.0745E+04, 1.0752E+04, 1.0759E+04,
     + 1.0766E+04, 1.0774E+04, 1.0781E+04, 1.0788E+04, 1.0795E+04,
     +  1.0802E+04, 1.0809E+04, 1.0817E+04, 1.0824E+04, 1.0831E+04,
     + 1.0838E+04, 1.0845E+04, 1.0853E+04, 1.0860E+04, 1.0867E+04,
     +  1.0874E+04, 1.0881E+04, 1.0889E+04, 1.0896E+04, 1.0903E+04,
     + 1.0910E+04, 1.0917E+04, 1.0925E+04, 1.0932E+04, 1.0939E+04,
     +  1.0946E+04, 1.0953E+04, 1.0961E+04, 1.0968E+04, 1.0975E+04,
     + 1.0982E+04, 1.0990E+04, 1.0997E+04, 1.1004E+04, 1.1011E+04,
     +  1.1019E+04, 1.1026E+04, 1.1033E+04, 1.1040E+04, 1.1048E+04,
     + 1.1055E+04, 1.1062E+04, 1.1069E+04, 1.1077E+04, 1.1084E+04,
     +  1.1091E+04, 1.1098E+04, 1.1106E+04, 1.1113E+04, 1.1120E+04,
     + 1.1128E+04, 1.1135E+04, 1.1142E+04, 1.1149E+04, 1.1157E+04,
     +  1.1164E+04, 1.1171E+04, 1.1179E+04, 1.1186E+04, 1.1193E+04,
     + 1.1201E+04, 1.1208E+04, 1.1215E+04, 1.1223E+04, 1.1230E+04,
     +  1.1237E+04, 1.1244E+04, 1.1252E+04, 1.1259E+04, 1.1266E+04,
     + 1.1274E+04, 1.1281E+04, 1.1289E+04, 1.1296E+04, 1.1303E+04,
     +  1.1311E+04, 1.1318E+04, 1.1325E+04, 1.1333E+04, 1.1340E+04,
     + 1.1347E+04, 1.1355E+04, 1.1362E+04, 1.1369E+04, 1.1377E+04,
     +  1.1384E+04, 1.1392E+04, 1.1399E+04, 1.1406E+04, 1.1414E+04,
     + 1.1421E+04, 1.1428E+04, 1.1436E+04, 1.1443E+04, 1.1451E+04,
     +  1.1458E+04, 1.1465E+04, 1.1473E+04, 1.1480E+04, 1.1488E+04,
     + 1.1495E+04, 1.1502E+04, 1.1510E+04, 1.1517E+04, 1.1525E+04,
     +  1.1532E+04, 1.1540E+04, 1.1547E+04, 1.1554E+04, 1.1562E+04,
     + 1.1569E+04, 1.1577E+04, 1.1584E+04, 1.1592E+04, 1.1599E+04,
     +  1.1607E+04, 1.1614E+04, 1.1621E+04, 1.1629E+04, 1.1636E+04,
     + 1.1644E+04, 1.1651E+04, 1.1659E+04, 1.1666E+04, 1.1674E+04,
     +  1.1681E+04, 1.1689E+04, 1.1696E+04, 1.1704E+04, 1.1711E+04,
     + 1.1719E+04, 1.1726E+04, 1.1734E+04, 1.1741E+04, 1.1749E+04,
     +  1.1756E+04, 1.1764E+04, 1.1771E+04, 1.1779E+04, 1.1786E+04,
     + 1.1794E+04, 1.1801E+04, 1.1809E+04, 1.1816E+04, 1.1824E+04,
     +  1.1831E+04, 1.1839E+04, 1.1846E+04, 1.1854E+04, 1.1861E+04,
     + 1.1869E+04, 1.1876E+04, 1.1884E+04, 1.1892E+04, 1.1899E+04,
     +  1.1907E+04, 1.1914E+04, 1.1922E+04, 1.1929E+04, 1.1937E+04,
     + 1.1944E+04, 1.1952E+04, 1.1960E+04, 1.1967E+04, 1.1975E+04,
     +  1.1982E+04, 1.1990E+04, 1.1997E+04, 1.2005E+04, 1.2013E+04,
     + 1.2020E+04, 1.2028E+04, 1.2035E+04, 1.2043E+04, 1.2051E+04,
     +  1.2058E+04, 1.2066E+04, 1.2073E+04, 1.2081E+04, 1.2089E+04,
     + 1.2096E+04, 1.2104E+04, 1.2112E+04, 1.2119E+04, 1.2127E+04,
     +  1.2134E+04, 1.2142E+04, 1.2150E+04, 1.2157E+04, 1.2165E+04,
     + 1.2173E+04, 1.2180E+04, 1.2188E+04, 1.2196E+04, 1.2203E+04,
     +  1.2211E+04, 1.2219E+04, 1.2226E+04, 1.2234E+04, 1.2241E+04,
     + 1.2249E+04, 1.2257E+04, 1.2265E+04, 1.2272E+04, 1.2280E+04,
     +  1.2288E+04, 1.2295E+04, 1.2303E+04, 1.2311E+04, 1.2318E+04,
     + 1.2326E+04, 1.2334E+04, 1.2341E+04, 1.2349E+04, 1.2357E+04,
     +  1.2364E+04, 1.2372E+04, 1.2380E+04, 1.2388E+04, 1.2395E+04,
     + 1.2403E+04, 1.2411E+04, 1.2418E+04, 1.2426E+04, 1.2434E+04,
     +  1.2442E+04, 1.2449E+04, 1.2457E+04, 1.2465E+04, 1.2473E+04,
     + 1.2480E+04, 1.2488E+04, 1.2496E+04, 1.2504E+04, 1.2511E+04,
     +  1.2519E+04, 1.2527E+04, 1.2535E+04, 1.2542E+04, 1.2550E+04,
     + 1.2558E+04, 1.2566E+04, 1.2574E+04, 1.2581E+04, 1.2589E+04,
     +  1.2597E+04, 1.2605E+04, 1.2612E+04, 1.2620E+04, 1.2628E+04,
     + 1.2636E+04, 1.2644E+04, 1.2651E+04, 1.2659E+04, 1.2667E+04,
     +  1.2675E+04, 1.2683E+04, 1.2690E+04, 1.2698E+04, 1.2706E+04,
     + 1.2714E+04, 1.2722E+04, 1.2730E+04, 1.2737E+04, 1.2745E+04,
     +  1.2753E+04, 1.2761E+04, 1.2769E+04, 1.2777E+04, 1.2784E+04,
     + 1.2792E+04, 1.2800E+04, 1.2808E+04, 1.2816E+04, 1.2824E+04,
     +  1.2832E+04, 1.2839E+04, 1.2847E+04, 1.2855E+04, 1.2863E+04,
     + 1.2871E+04, 1.2879E+04, 1.2887E+04, 1.2895E+04, 1.2902E+04,
     +  1.2910E+04, 1.2918E+04, 1.2926E+04, 1.2934E+04, 1.2942E+04,
     + 1.2950E+04, 1.2958E+04, 1.2966E+04, 1.2973E+04, 1.2981E+04,
     +  1.2989E+04, 1.2997E+04, 1.3005E+04, 1.3013E+04, 1.3021E+04,
     + 1.3029E+04, 1.3037E+04, 1.3045E+04, 1.3053E+04, 1.3061E+04,
     +  1.3069E+04, 1.3076E+04, 1.3084E+04, 1.3092E+04, 1.3100E+04,
     + 1.3108E+04, 1.3116E+04, 1.3124E+04, 1.3132E+04, 1.3140E+04,
     +  1.3148E+04, 1.3156E+04, 1.3164E+04, 1.3172E+04, 1.3180E+04,
     + 1.3188E+04, 1.3196E+04, 1.3204E+04, 1.3212E+04, 1.3220E+04,
     +  1.3228E+04, 1.3236E+04, 1.3244E+04, 1.3252E+04, 1.3260E+04,
     + 1.3268E+04, 1.3276E+04, 1.3284E+04, 1.3292E+04, 1.3300E+04,
     +  1.3308E+04, 1.3316E+04, 1.3324E+04, 1.3332E+04, 1.3340E+04,
     + 1.3348E+04, 1.3356E+04, 1.3364E+04, 1.3372E+04, 1.3380E+04,
     +  1.3388E+04, 1.3396E+04, 1.3404E+04, 1.3412E+04, 1.3420E+04,
     + 1.3428E+04, 1.3436E+04, 1.3445E+04, 1.3453E+04, 1.3461E+04,
     +  1.3469E+04, 1.3477E+04, 1.3485E+04, 1.3493E+04, 1.3501E+04,
     + 1.3509E+04, 1.3517E+04, 1.3525E+04, 1.3533E+04, 1.3541E+04,
     +  1.3549E+04, 1.3558E+04, 1.3566E+04, 1.3574E+04, 1.3582E+04,
     + 1.3590E+04, 1.3598E+04, 1.3606E+04, 1.3614E+04, 1.3622E+04,
     +  1.3631E+04, 1.3639E+04, 1.3647E+04, 1.3655E+04, 1.3663E+04,
     + 1.3671E+04, 1.3679E+04, 1.3687E+04, 1.3696E+04, 1.3704E+04,
     +  1.3712E+04, 1.3720E+04, 1.3728E+04, 1.3736E+04, 1.3744E+04,
     + 1.3753E+04, 1.3761E+04, 1.3769E+04, 1.3777E+04, 1.3785E+04,
     +  1.3793E+04, 1.3802E+04, 1.3810E+04, 1.3818E+04, 1.3826E+04,
     + 1.3834E+04, 1.3843E+04, 1.3851E+04, 1.3859E+04, 1.3867E+04,
     +  1.3875E+04, 1.3883E+04, 1.3892E+04, 1.3900E+04, 1.3908E+04,
     + 1.3916E+04, 1.3925E+04, 1.3933E+04, 1.3941E+04, 1.3949E+04,
     +  1.3957E+04, 1.3966E+04, 1.3974E+04, 1.3982E+04, 1.3990E+04,
     + 1.3999E+04, 1.4007E+04, 1.4015E+04, 1.4023E+04, 1.4032E+04,
     +  1.4040E+04, 1.4048E+04, 1.4056E+04, 1.4065E+04, 1.4073E+04,
     + 1.4081E+04, 1.4089E+04, 1.4098E+04, 1.4106E+04, 1.4114E+04,
     +  1.4122E+04, 1.4131E+04, 1.4139E+04, 1.4147E+04, 1.4155E+04,
     + 1.4164E+04, 1.4172E+04, 1.4180E+04, 1.4189E+04, 1.4197E+04,
     +  1.4205E+04, 1.4214E+04, 1.4222E+04, 1.4230E+04, 1.4238E+04,
     + 1.4247E+04, 1.4255E+04, 1.4263E+04, 1.4272E+04, 1.4280E+04,
     +  1.4288E+04, 1.4297E+04, 1.4305E+04, 1.4313E+04, 1.4322E+04,
     + 1.4330E+04, 1.4338E+04, 1.4347E+04, 1.4355E+04, 1.4363E+04,
     +  1.4372E+04, 1.4380E+04, 1.4388E+04, 1.4397E+04, 1.4405E+04,
     + 1.4414E+04, 1.4422E+04, 1.4430E+04, 1.4439E+04, 1.4447E+04,
     +  1.4455E+04, 1.4464E+04, 1.4472E+04, 1.4481E+04, 1.4489E+04,
     + 1.4497E+04, 1.4506E+04, 1.4514E+04, 1.4523E+04, 1.4531E+04,
     +  1.4539E+04, 1.4548E+04, 1.4556E+04, 1.4565E+04, 1.4573E+04,
     + 1.4581E+04, 1.4590E+04, 1.4598E+04, 1.4607E+04, 1.4615E+04,
     +  1.4624E+04, 1.4632E+04, 1.4640E+04, 1.4649E+04, 1.4657E+04,
     + 1.4666E+04, 1.4674E+04, 1.4683E+04, 1.4691E+04, 1.4699E+04,
     +  1.4708E+04, 1.4716E+04, 1.4725E+04, 1.4733E+04, 1.4742E+04,
     + 1.4750E+04, 1.4759E+04, 1.4767E+04, 1.4776E+04, 1.4784E+04,
     +  1.4793E+04, 1.4801E+04, 1.4810E+04, 1.4818E+04, 1.4827E+04,
     + 1.4835E+04, 1.4844E+04, 1.4852E+04, 1.4861E+04, 1.4869E+04,
     +  1.4878E+04, 1.4886E+04, 1.4895E+04, 1.4903E+04, 1.4912E+04,
     + 1.4920E+04, 1.4929E+04, 1.4937E+04, 1.4946E+04, 1.4954E+04,
     +  1.4963E+04, 1.4971E+04, 1.4980E+04, 1.4988E+04, 1.4997E+04,
     + 1.5005E+04, 1.5014E+04, 1.5023E+04, 1.5031E+04, 1.5040E+04,
     +  1.5048E+04, 1.5057E+04, 1.5065E+04, 1.5074E+04, 1.5083E+04,
     + 1.5091E+04, 1.5100E+04, 1.5108E+04, 1.5117E+04, 1.5125E+04,
     +  1.5134E+04, 1.5143E+04, 1.5151E+04, 1.5160E+04, 1.5168E+04,
     + 1.5177E+04, 1.5186E+04, 1.5194E+04, 1.5203E+04, 1.5211E+04,
     +  1.5220E+04, 1.5229E+04, 1.5237E+04, 1.5246E+04, 1.5254E+04,
     + 1.5263E+04, 1.5272E+04, 1.5280E+04, 1.5289E+04, 1.5298E+04,
     +  1.5306E+04, 1.5315E+04, 1.5324E+04, 1.5332E+04, 1.5341E+04,
     + 1.5349E+04, 1.5358E+04, 1.5367E+04, 1.5375E+04, 1.5384E+04,
     +  1.5393E+04, 1.5401E+04, 1.5410E+04, 1.5419E+04, 1.5427E+04,
     + 1.5436E+04, 1.5445E+04, 1.5453E+04, 1.5462E+04, 1.5471E+04,
     +  1.5480E+04, 1.5488E+04, 1.5497E+04, 1.5506E+04, 1.5514E+04,
     + 1.5523E+04, 1.5532E+04, 1.5540E+04, 1.5549E+04, 1.5558E+04,
     +  1.5567E+04, 1.5575E+04, 1.5584E+04, 1.5593E+04, 1.5601E+04,
     + 1.5610E+04, 1.5619E+04, 1.5628E+04, 1.5636E+04, 1.5645E+04,
     +  1.5654E+04, 1.5663E+04, 1.5671E+04, 1.5680E+04, 1.5689E+04,
     + 1.5698E+04, 1.5706E+04, 1.5715E+04, 1.5724E+04, 1.5733E+04,
     +  1.5741E+04, 1.5750E+04, 1.5759E+04, 1.5768E+04, 1.5776E+04,
     + 1.5785E+04, 1.5794E+04, 1.5803E+04, 1.5812E+04, 1.5820E+04,
     +  1.5829E+04, 1.5838E+04, 1.5847E+04, 1.5856E+04, 1.5864E+04,
     + 1.5873E+04, 1.5882E+04, 1.5891E+04, 1.5900E+04, 1.5908E+04,
     +  1.5917E+04, 1.5926E+04, 1.5935E+04, 1.5944E+04, 1.5953E+04,
     + 1.5961E+04, 1.5970E+04, 1.5979E+04, 1.5988E+04, 1.5997E+04,
     +  1.6006E+04, 1.6014E+04, 1.6023E+04, 1.6032E+04, 1.6041E+04,
     + 1.6050E+04, 1.6059E+04, 1.6068E+04, 1.6076E+04, 1.6085E+04,
     +  1.6094E+04, 1.6103E+04, 1.6112E+04, 1.6121E+04, 1.6130E+04,
     + 1.6139E+04, 1.6148E+04, 1.6156E+04, 1.6165E+04, 1.6174E+04,
     +  1.6183E+04, 1.6192E+04, 1.6201E+04, 1.6210E+04, 1.6219E+04,
     + 1.6228E+04, 1.6237E+04, 1.6245E+04, 1.6254E+04, 1.6263E+04,
     +  1.6272E+04, 1.6281E+04, 1.6290E+04, 1.6299E+04, 1.6308E+04,
     + 1.6317E+04, 1.6326E+04, 1.6335E+04, 1.6344E+04, 1.6353E+04,
     +  1.6362E+04, 1.6371E+04, 1.6380E+04, 1.6388E+04, 1.6397E+04,
     + 1.6406E+04, 1.6415E+04, 1.6424E+04, 1.6433E+04, 1.6442E+04,
     +  1.6451E+04, 1.6460E+04, 1.6469E+04, 1.6478E+04, 1.6487E+04,
     + 1.6496E+04, 1.6505E+04, 1.6514E+04, 1.6523E+04, 1.6532E+04,
     +  1.6541E+04, 1.6550E+04, 1.6559E+04, 1.6568E+04, 1.6577E+04,
     + 1.6586E+04, 1.6595E+04, 1.6604E+04, 1.6613E+04, 1.6622E+04,
     +  1.6631E+04, 1.6640E+04, 1.6649E+04, 1.6658E+04, 1.6667E+04,
     + 1.6676E+04, 1.6686E+04, 1.6695E+04, 1.6704E+04, 1.6713E+04,
     +  1.6722E+04, 1.6731E+04, 1.6740E+04, 1.6749E+04, 1.6758E+04,
     + 1.6767E+04, 1.6776E+04, 1.6785E+04, 1.6794E+04, 1.6803E+04,
     +  1.6812E+04, 1.6822E+04, 1.6831E+04, 1.6840E+04, 1.6849E+04,
     + 1.6858E+04, 1.6867E+04, 1.6876E+04, 1.6885E+04, 1.6894E+04,
     +  1.6903E+04, 1.6913E+04, 1.6922E+04, 1.6931E+04, 1.6940E+04,
     + 1.6949E+04, 1.6958E+04, 1.6967E+04, 1.6976E+04, 1.6985E+04,
     +  1.6995E+04, 1.7004E+04, 1.7013E+04, 1.7022E+04, 1.7031E+04,
     + 1.7040E+04, 1.7049E+04, 1.7059E+04, 1.7068E+04, 1.7077E+04,
     +  1.7086E+04, 1.7095E+04, 1.7104E+04, 1.7114E+04, 1.7123E+04,
     + 1.7132E+04, 1.7141E+04, 1.7150E+04, 1.7159E+04, 1.7169E+04,
     +  1.7178E+04, 1.7187E+04, 1.7196E+04, 1.7205E+04, 1.7215E+04,
     + 1.7224E+04, 1.7233E+04, 1.7242E+04, 1.7251E+04, 1.7261E+04,
     +  1.7270E+04, 1.7279E+04, 1.7288E+04, 1.7297E+04, 1.7307E+04,
     + 1.7316E+04, 1.7325E+04, 1.7334E+04, 1.7344E+04, 1.7353E+04,
     +  1.7362E+04, 1.7371E+04, 1.7381E+04, 1.7390E+04, 1.7399E+04,
     + 1.7408E+04, 1.7418E+04, 1.7427E+04, 1.7436E+04, 1.7445E+04,
     +  1.7455E+04, 1.7464E+04, 1.7473E+04, 1.7482E+04, 1.7492E+04,
     + 1.7501E+04, 1.7510E+04, 1.7520E+04, 1.7529E+04, 1.7538E+04,
     +  1.7547E+04, 1.7557E+04, 1.7566E+04, 1.7575E+04, 1.7585E+04,
     + 1.7594E+04, 1.7603E+04, 1.7613E+04, 1.7622E+04, 1.7631E+04,
     +  1.7640E+04, 1.7650E+04, 1.7659E+04, 1.7668E+04, 1.7678E+04,
     + 1.7687E+04, 1.7696E+04, 1.7706E+04, 1.7715E+04, 1.7724E+04,
     +  1.7734E+04, 1.7743E+04, 1.7752E+04, 1.7762E+04, 1.7771E+04,
     + 1.7781E+04, 1.7790E+04, 1.7799E+04, 1.7809E+04, 1.7818E+04,
     +  1.7827E+04, 1.7837E+04, 1.7846E+04, 1.7855E+04, 1.7865E+04,
     + 1.7874E+04, 1.7884E+04, 1.7893E+04, 1.7902E+04, 1.7912E+04,
     +  1.7921E+04, 1.7931E+04, 1.7940E+04, 1.7949E+04, 1.7959E+04,
     + 1.7968E+04, 1.7978E+04, 1.7987E+04, 1.7996E+04, 1.8006E+04,
     +  1.8015E+04, 1.8025E+04, 1.8034E+04, 1.8044E+04, 1.8053E+04,
     + 1.8062E+04, 1.8072E+04, 1.8081E+04, 1.8091E+04, 1.8100E+04,
     +  1.8110E+04, 1.8119E+04, 1.8129E+04, 1.8138E+04, 1.8148E+04,
     + 1.8157E+04, 1.8166E+04, 1.8176E+04, 1.8185E+04, 1.8195E+04,
     +  1.8204E+04, 1.8214E+04, 1.8223E+04, 1.8233E+04, 1.8242E+04,
     + 1.8252E+04, 1.8261E+04, 1.8271E+04, 1.8280E+04, 1.8290E+04,
     +  1.8299E+04, 1.8309E+04, 1.8318E+04, 1.8328E+04, 1.8337E+04,
     + 1.8347E+04, 1.8356E+04, 1.8366E+04, 1.8375E+04, 1.8385E+04,
     +  1.8394E+04, 1.8404E+04, 1.8413E+04, 1.8423E+04, 1.8432E+04,
     + 1.8442E+04, 1.8452E+04, 1.8461E+04, 1.8471E+04, 1.8480E+04,
     +  1.8490E+04, 1.8499E+04, 1.8509E+04, 1.8518E+04, 1.8528E+04,
     + 1.8538E+04, 1.8547E+04, 1.8557E+04, 1.8566E+04, 1.8576E+04,
     +  1.8585E+04, 1.8595E+04, 1.8605E+04, 1.8614E+04, 1.8624E+04,
     + 1.8633E+04, 1.8643E+04, 1.8653E+04, 1.8662E+04, 1.8672E+04,
     +  1.8681E+04, 1.8691E+04, 1.8701E+04, 1.8710E+04, 1.8720E+04,
     + 1.8729E+04, 1.8739E+04, 1.8749E+04, 1.8758E+04, 1.8768E+04,
     +  1.8778E+04, 1.8787E+04, 1.8797E+04, 1.8807E+04, 1.8816E+04,
     + 1.8826E+04, 1.8835E+04, 1.8845E+04, 1.8855E+04, 1.8864E+04,
     +  1.8874E+04, 1.8884E+04, 1.8893E+04, 1.8903E+04, 1.8913E+04,
     + 1.8922E+04, 1.8932E+04, 1.8942E+04, 1.8951E+04, 1.8961E+04,
     +  1.8971E+04, 1.8981E+04, 1.8990E+04, 1.9000E+04, 1.9010E+04,
     + 1.9019E+04, 1.9029E+04, 1.9039E+04, 1.9048E+04, 1.9058E+04,
     +  1.9068E+04, 1.9078E+04, 1.9087E+04, 1.9097E+04, 1.9107E+04,
     + 1.9116E+04, 1.9126E+04, 1.9136E+04, 1.9146E+04, 1.9155E+04,
     +  1.9165E+04, 1.9175E+04, 1.9185E+04, 1.9194E+04, 1.9204E+04,
     + 1.9214E+04, 1.9224E+04, 1.9233E+04, 1.9243E+04, 1.9253E+04,
     +  1.9263E+04, 1.9272E+04, 1.9282E+04, 1.9292E+04, 1.9302E+04,
     + 1.9311E+04, 1.9321E+04, 1.9331E+04, 1.9341E+04, 1.9351E+04,
     +  1.9360E+04, 1.9370E+04, 1.9380E+04, 1.9390E+04, 1.9400E+04,
     + 1.9409E+04, 1.9419E+04, 1.9429E+04, 1.9439E+04, 1.9449E+04,
     +  1.9458E+04, 1.9468E+04, 1.9478E+04, 1.9488E+04, 1.9498E+04,
     + 1.9508E+04, 1.9517E+04, 1.9527E+04, 1.9537E+04, 1.9547E+04,
     +  1.9557E+04, 1.9567E+04, 1.9576E+04, 1.9586E+04, 1.9596E+04,
     + 1.9606E+04, 1.9616E+04, 1.9626E+04, 1.9636E+04, 1.9645E+04,
     +  1.9655E+04, 1.9665E+04, 1.9675E+04, 1.9685E+04, 1.9695E+04,
     + 1.9705E+04, 1.9715E+04, 1.9725E+04, 1.9734E+04, 1.9744E+04,
     +  1.9754E+04, 1.9764E+04, 1.9774E+04, 1.9784E+04, 1.9794E+04,
     + 1.9804E+04, 1.9814E+04, 1.9824E+04, 1.9833E+04, 1.9843E+04,
     +  1.9853E+04, 1.9863E+04, 1.9873E+04, 1.9883E+04, 1.9893E+04,
     + 1.9903E+04, 1.9913E+04, 1.9923E+04, 1.9933E+04, 1.9943E+04,
     +  1.9953E+04, 1.9963E+04, 1.9973E+04, 1.9983E+04, 1.9993E+04,
     + 2.0002E+04, 2.0012E+04, 2.0022E+04, 2.0032E+04, 2.0042E+04,
     +  2.0052E+04, 2.0062E+04, 2.0072E+04, 2.0082E+04, 2.0092E+04,
     + 2.0102E+04, 2.0112E+04, 2.0122E+04, 2.0132E+04, 2.0142E+04,
     +  2.0152E+04, 2.0162E+04, 2.0172E+04, 2.0182E+04, 2.0192E+04,
     + 2.0202E+04, 2.0212E+04, 2.0222E+04, 2.0232E+04, 2.0242E+04,
     +  2.0252E+04, 2.0262E+04, 2.0272E+04, 2.0283E+04, 2.0293E+04,
     + 2.0303E+04, 2.0313E+04, 2.0323E+04, 2.0333E+04, 2.0343E+04,
     +  2.0353E+04, 2.0363E+04, 2.0373E+04, 2.0383E+04, 2.0393E+04,
     + 2.0403E+04, 2.0413E+04, 2.0423E+04, 2.0433E+04, 2.0443E+04,
     +  2.0454E+04, 2.0464E+04, 2.0474E+04, 2.0484E+04, 2.0494E+04,
     + 2.0504E+04, 2.0514E+04, 2.0524E+04, 2.0534E+04, 2.0544E+04,
     +  2.0554E+04, 2.0565E+04, 2.0575E+04, 2.0585E+04, 2.0595E+04,
     + 2.0605E+04, 2.0615E+04, 2.0625E+04, 2.0635E+04, 2.0646E+04,
     +  2.0656E+04, 2.0666E+04, 2.0676E+04, 2.0686E+04, 2.0696E+04,
     + 2.0706E+04, 2.0717E+04, 2.0727E+04, 2.0737E+04, 2.0747E+04,
     +  2.0757E+04, 2.0767E+04, 2.0778E+04, 2.0788E+04, 2.0798E+04,
     + 2.0808E+04, 2.0818E+04, 2.0828E+04, 2.0839E+04, 2.0849E+04,
     +  2.0859E+04, 2.0869E+04, 2.0879E+04, 2.0890E+04, 2.0900E+04,
     + 2.0910E+04, 2.0920E+04, 2.0930E+04, 2.0941E+04, 2.0951E+04,
     +  2.0961E+04, 2.0971E+04, 2.0981E+04, 2.0992E+04, 2.1002E+04,
     + 2.1012E+04, 2.1022E+04, 2.1033E+04, 2.1043E+04, 2.1053E+04,
     +  2.1063E+04, 2.1074E+04, 2.1084E+04, 2.1094E+04, 2.1104E+04,
     + 2.1115E+04, 2.1125E+04, 2.1135E+04, 2.1145E+04, 2.1156E+04,
     +  2.1166E+04, 2.1176E+04, 2.1186E+04, 2.1197E+04, 2.1207E+04,
     + 2.1217E+04, 2.1227E+04, 2.1238E+04, 2.1248E+04, 2.1258E+04,
     +  2.1269E+04, 2.1279E+04, 2.1289E+04, 2.1300E+04, 2.1310E+04,
     + 2.1320E+04, 2.1330E+04, 2.1341E+04, 2.1351E+04, 2.1361E+04,
     +  2.1372E+04, 2.1382E+04, 2.1392E+04, 2.1403E+04, 2.1413E+04,
     + 2.1423E+04, 2.1434E+04, 2.1444E+04, 2.1454E+04, 2.1465E+04,
     +  2.1475E+04, 2.1485E+04, 2.1496E+04, 2.1506E+04, 2.1516E+04,
     + 2.1527E+04, 2.1537E+04, 2.1548E+04, 2.1558E+04, 2.1568E+04,
     +  2.1579E+04, 2.1589E+04, 2.1599E+04, 2.1610E+04, 2.1620E+04,
     + 2.1631E+04, 2.1641E+04, 2.1651E+04, 2.1662E+04, 2.1672E+04,
     +  2.1683E+04, 2.1693E+04, 2.1703E+04, 2.1714E+04, 2.1724E+04,
     + 2.1735E+04, 2.1745E+04, 2.1755E+04, 2.1766E+04, 2.1776E+04,
     +  2.1787E+04, 2.1797E+04, 2.1808E+04, 2.1818E+04, 2.1828E+04,
     + 2.1839E+04, 2.1849E+04, 2.1860E+04, 2.1870E+04, 2.1881E+04,
     +  2.1891E+04, 2.1902E+04, 2.1912E+04, 2.1922E+04, 2.1933E+04,
     + 2.1943E+04, 2.1954E+04, 2.1964E+04, 2.1975E+04, 2.1985E+04,
     +  2.1996E+04, 2.2006E+04, 2.2017E+04, 2.2027E+04, 2.2038E+04,
     + 2.2048E+04, 2.2059E+04, 2.2069E+04, 2.2080E+04, 2.2090E+04,
     +  2.2101E+04, 2.2111E+04, 2.2122E+04, 2.2132E+04, 2.2143E+04,
     + 2.2153E+04, 2.2164E+04, 2.2174E+04, 2.2185E+04, 2.2195E+04,
     +  2.2206E+04, 2.2216E+04, 2.2227E+04, 2.2237E+04, 2.2248E+04,
     + 2.2259E+04, 2.2269E+04, 2.2280E+04, 2.2290E+04, 2.2301E+04,
     +  2.2311E+04, 2.2322E+04, 2.2332E+04, 2.2343E+04, 2.2354E+04,
     + 2.2364E+04, 2.2375E+04, 2.2385E+04, 2.2396E+04, 2.2406E+04,
     +  2.2417E+04, 2.2428E+04, 2.2438E+04, 2.2449E+04, 2.2459E+04,
     + 2.2470E+04, 2.2481E+04, 2.2491E+04, 2.2502E+04, 2.2512E+04,
     +  2.2523E+04, 2.2534E+04, 2.2544E+04, 2.2555E+04, 2.2565E+04,
     + 2.2576E+04, 2.2587E+04, 2.2597E+04, 2.2608E+04, 2.2619E+04,
     +  2.2629E+04, 2.2640E+04, 2.2651E+04, 2.2661E+04, 2.2672E+04,
     + 2.2682E+04, 2.2693E+04, 2.2704E+04, 2.2714E+04, 2.2725E+04,
     +  2.2736E+04, 2.2746E+04, 2.2757E+04, 2.2768E+04, 2.2778E+04,
     + 2.2789E+04, 2.2800E+04, 2.2810E+04, 2.2821E+04, 2.2832E+04,
     +  2.2843E+04, 2.2853E+04, 2.2864E+04, 2.2875E+04, 2.2885E+04,
     + 2.2896E+04, 2.2907E+04, 2.2917E+04, 2.2928E+04, 2.2939E+04,
     +  2.2950E+04, 2.2960E+04, 2.2971E+04, 2.2982E+04, 2.2992E+04,
     + 2.3003E+04, 2.3014E+04, 2.3025E+04, 2.3035E+04, 2.3046E+04,
     +  2.3057E+04, 2.3068E+04, 2.3078E+04, 2.3089E+04, 2.3100E+04,
     + 2.3111E+04, 2.3121E+04, 2.3132E+04, 2.3143E+04, 2.3154E+04,
     +  2.3164E+04, 2.3175E+04, 2.3186E+04, 2.3197E+04, 2.3208E+04,
     + 2.3218E+04, 2.3229E+04, 2.3240E+04, 2.3251E+04, 2.3262E+04,
     +  2.3272E+04, 2.3283E+04, 2.3294E+04, 2.3305E+04, 2.3316E+04,
     + 2.3326E+04, 2.3337E+04, 2.3348E+04, 2.3359E+04, 2.3370E+04,
     +  2.3380E+04, 2.3391E+04, 2.3402E+04, 2.3413E+04, 2.3424E+04,
     + 2.3435E+04, 2.3445E+04, 2.3456E+04, 2.3467E+04, 2.3478E+04,
     +  2.3489E+04, 2.3500E+04, 2.3511E+04, 2.3521E+04, 2.3532E+04,
     + 2.3543E+04, 2.3554E+04, 2.3565E+04, 2.3576E+04, 2.3587E+04,
     +  2.3597E+04, 2.3608E+04, 2.3619E+04, 2.3630E+04, 2.3641E+04,
     + 2.3652E+04, 2.3663E+04, 2.3674E+04, 2.3685E+04, 2.3695E+04,
     +  2.3706E+04, 2.3717E+04, 2.3728E+04, 2.3739E+04, 2.3750E+04,
     + 2.3761E+04, 2.3772E+04, 2.3783E+04, 2.3794E+04, 2.3805E+04,
     +  2.3816E+04, 2.3826E+04, 2.3837E+04, 2.3848E+04, 2.3859E+04,
     + 2.3870E+04, 2.3881E+04, 2.3892E+04, 2.3903E+04, 2.3914E+04,
     +  2.3925E+04, 2.3936E+04, 2.3947E+04, 2.3958E+04, 2.3969E+04,
     + 2.3980E+04, 2.3991E+04, 2.4002E+04, 2.4013E+04, 2.4024E+04,
     +  2.4035E+04, 2.4046E+04, 2.4057E+04, 2.4068E+04, 2.4079E+04,
     + 2.4090E+04, 2.4101E+04, 2.4112E+04, 2.4123E+04, 2.4134E+04,
     +  2.4145E+04, 2.4156E+04, 2.4167E+04, 2.4178E+04, 2.4189E+04,
     + 2.4200E+04, 2.4211E+04, 2.4222E+04, 2.4233E+04, 2.4244E+04,
     +  2.4255E+04, 2.4266E+04, 2.4277E+04, 2.4288E+04, 2.4299E+04,
     + 2.4310E+04, 2.4321E+04, 2.4332E+04, 2.4343E+04, 2.4354E+04,
     +  2.4365E+04, 2.4376E+04, 2.4387E+04, 2.4399E+04, 2.4410E+04,
     + 2.4421E+04, 2.4432E+04, 2.4443E+04, 2.4454E+04, 2.4465E+04,
     +  2.4476E+04, 2.4487E+04, 2.4498E+04, 2.4509E+04, 2.4520E+04,
     + 2.4532E+04, 2.4543E+04, 2.4554E+04, 2.4565E+04, 2.4576E+04,
     +  2.4587E+04, 2.4598E+04, 2.4609E+04, 2.4620E+04, 2.4632E+04,
     + 2.4643E+04, 2.4654E+04, 2.4665E+04, 2.4676E+04, 2.4687E+04,
     +  2.4698E+04, 2.4710E+04, 2.4721E+04, 2.4732E+04, 2.4743E+04,
     + 2.4754E+04, 2.4765E+04, 2.4776E+04, 2.4788E+04, 2.4799E+04,
     +  2.4810E+04, 2.4821E+04, 2.4832E+04, 2.4843E+04, 2.4855E+04,
     + 2.4866E+04, 2.4877E+04, 2.4888E+04, 2.4899E+04, 2.4911E+04,
     +  2.4922E+04, 2.4933E+04, 2.4944E+04, 2.4955E+04, 2.4967E+04,
     + 2.4978E+04, 2.4989E+04, 2.5000E+04, 2.5011E+04, 2.5023E+04,
     +  2.5034E+04, 2.5045E+04, 2.5056E+04, 2.5068E+04, 2.5079E+04,
     + 2.5090E+04, 2.5101E+04, 2.5112E+04, 2.5124E+04, 2.5135E+04,
     +  2.5146E+04, 2.5157E+04, 2.5169E+04, 2.5180E+04, 2.5191E+04,
     + 2.5202E+04, 2.5214E+04, 2.5225E+04, 2.5236E+04, 2.5248E+04,
     +  2.5259E+04, 2.5270E+04, 2.5281E+04, 2.5293E+04, 2.5304E+04,
     + 2.5315E+04, 2.5327E+04, 2.5338E+04, 2.5349E+04, 2.5360E+04,
     +  2.5372E+04, 2.5383E+04, 2.5394E+04, 2.5406E+04, 2.5417E+04,
     + 2.5428E+04, 2.5440E+04, 2.5451E+04, 2.5462E+04, 2.5474E+04,
     +  2.5485E+04, 2.5496E+04, 2.5508E+04, 2.5519E+04, 2.5530E+04,
     + 2.5542E+04, 2.5553E+04, 2.5564E+04, 2.5576E+04, 2.5587E+04,
     +  2.5598E+04, 2.5610E+04, 2.5621E+04, 2.5632E+04, 2.5644E+04,
     + 2.5655E+04, 2.5666E+04, 2.5678E+04, 2.5689E+04, 2.5701E+04,
     +  2.5712E+04, 2.5723E+04, 2.5735E+04, 2.5746E+04, 2.5757E+04,
     + 2.5769E+04, 2.5780E+04, 2.5792E+04, 2.5803E+04, 2.5814E+04,
     +  2.5826E+04, 2.5837E+04, 2.5849E+04, 2.5860E+04, 2.5872E+04,
     + 2.5883E+04, 2.5894E+04, 2.5906E+04, 2.5917E+04, 2.5929E+04,
     +  2.5940E+04, 2.5952E+04, 2.5963E+04, 2.5974E+04, 2.5986E+04,
     + 2.5997E+04, 2.6009E+04, 2.6020E+04, 2.6032E+04, 2.6043E+04,
     +  2.6055E+04, 2.6066E+04, 2.6077E+04, 2.6089E+04, 2.6100E+04,
     + 2.6112E+04, 2.6123E+04, 2.6135E+04, 2.6146E+04, 2.6158E+04,
     +  2.6169E+04, 2.6181E+04, 2.6192E+04, 2.6204E+04, 2.6215E+04,
     + 2.6227E+04, 2.6238E+04, 2.6250E+04, 2.6261E+04, 2.6273E+04,
     +  2.6284E+04, 2.6296E+04, 2.6307E+04, 2.6319E+04, 2.6330E+04,
     + 2.6342E+04, 2.6353E+04, 2.6365E+04, 2.6376E+04, 2.6388E+04,
     +  2.6400E+04, 2.6411E+04, 2.6423E+04, 2.6434E+04, 2.6446E+04,
     + 2.6457E+04, 2.6469E+04, 2.6480E+04, 2.6492E+04, 2.6503E+04,
     +  2.6515E+04, 2.6527E+04, 2.6538E+04, 2.6550E+04, 2.6561E+04,
     + 2.6573E+04, 2.6584E+04, 2.6596E+04, 2.6608E+04, 2.6619E+04,
     +  2.6631E+04, 2.6642E+04, 2.6654E+04, 2.6666E+04, 2.6677E+04,
     + 2.6689E+04, 2.6700E+04, 2.6712E+04, 2.6724E+04, 2.6735E+04,
     +  2.6747E+04, 2.6759E+04, 2.6770E+04, 2.6782E+04, 2.6793E+04,
     + 2.6805E+04, 2.6817E+04, 2.6828E+04, 2.6840E+04, 2.6852E+04,
     +  2.6863E+04, 2.6875E+04, 2.6887E+04, 2.6898E+04, 2.6910E+04,
     + 2.6921E+04, 2.6933E+04, 2.6945E+04, 2.6956E+04, 2.6968E+04,
     +  2.6980E+04, 2.6991E+04, 2.7003E+04, 2.7015E+04, 2.7027E+04,
     + 2.7038E+04, 2.7050E+04, 2.7062E+04, 2.7073E+04, 2.7085E+04,
     +  2.7097E+04, 2.7108E+04, 2.7120E+04, 2.7132E+04, 2.7143E+04,
     + 2.7155E+04, 2.7167E+04, 2.7179E+04, 2.7190E+04, 2.7202E+04,
     +  2.7214E+04, 2.7226E+04, 2.7237E+04, 2.7249E+04, 2.7261E+04,
     + 2.7272E+04, 2.7284E+04, 2.7296E+04, 2.7308E+04, 2.7319E+04,
     +  2.7331E+04, 2.7343E+04, 2.7355E+04, 2.7366E+04, 2.7378E+04,
     + 2.7390E+04, 2.7402E+04, 2.7413E+04, 2.7425E+04, 2.7437E+04,
     +  2.7449E+04, 2.7461E+04, 2.7472E+04, 2.7484E+04, 2.7496E+04,
     + 2.7508E+04, 2.7520E+04, 2.7531E+04, 2.7543E+04, 2.7555E+04,
     +  2.7567E+04, 2.7579E+04, 2.7590E+04, 2.7602E+04, 2.7614E+04,
     + 2.7626E+04, 2.7638E+04, 2.7649E+04, 2.7661E+04/
c...        --        37
      data (QofT( 6,J),J=1,2408)/ 1.2209E+01, 1.4769E+01, 1.8821E+01,
     +  2.3236E+01, 2.7764E+01, 3.2341E+01, 3.6945E+01, 4.1563E+01,
     + 4.6192E+01, 5.0828E+01, 5.5468E+01, 6.0112E+01, 6.4759E+01,
     +  6.9409E+01, 7.4059E+01, 7.8711E+01, 8.3365E+01, 8.8019E+01,
     + 9.2674E+01, 9.7330E+01, 1.0664E+02, 1.1596E+02, 1.2527E+02,
     +  1.3459E+02, 1.4391E+02, 1.5323E+02, 1.6255E+02, 1.7187E+02,
     + 1.8119E+02, 1.9051E+02, 1.9983E+02, 2.0915E+02, 2.1847E+02,
     +  2.2780E+02, 2.3712E+02, 2.4644E+02, 2.5576E+02, 2.6509E+02,
     + 2.7441E+02, 2.8373E+02, 2.9306E+02, 3.0238E+02, 3.1170E+02,
     +  3.2103E+02, 3.3035E+02, 3.3968E+02, 3.4900E+02, 3.5832E+02,
     + 3.6765E+02, 3.7697E+02, 3.8630E+02, 3.9562E+02, 4.0495E+02,
     +  4.1427E+02, 4.2360E+02, 4.3292E+02, 4.4225E+02, 4.5158E+02,
     + 4.6090E+02, 4.7023E+02, 4.7955E+02, 4.8888E+02, 4.9821E+02,
     +  5.0753E+02, 5.1686E+02, 5.2618E+02, 5.3551E+02, 5.4484E+02,
     + 5.5416E+02, 5.6349E+02, 5.7282E+02, 5.8214E+02, 5.9147E+02,
     +  6.0080E+02, 6.1013E+02, 6.1945E+02, 6.2878E+02, 6.3811E+02,
     + 6.4744E+02, 6.5676E+02, 6.6609E+02, 6.7542E+02, 6.8475E+02,
     +  6.9408E+02, 7.0340E+02, 7.1273E+02, 7.2206E+02, 7.3139E+02,
     + 7.4072E+02, 7.5005E+02, 7.5938E+02, 7.6870E+02, 7.7803E+02,
     +  7.8736E+02, 7.9669E+02, 8.0602E+02, 8.1535E+02, 8.2468E+02,
     + 8.3401E+02, 8.4334E+02, 8.5267E+02, 8.6200E+02, 8.7133E+02,
     +  8.8066E+02, 8.8999E+02, 8.9932E+02, 9.0865E+02, 9.1798E+02,
     + 9.2731E+02, 9.3664E+02, 9.4597E+02, 9.5530E+02, 9.6463E+02,
     +  9.7396E+02, 9.8329E+02, 9.9262E+02, 1.0020E+03, 1.0113E+03,
     + 1.0206E+03, 1.0299E+03, 1.0393E+03, 1.0486E+03, 1.0579E+03,
     +  1.0673E+03, 1.0766E+03, 1.0859E+03, 1.0953E+03, 1.1046E+03,
     + 1.1139E+03, 1.1233E+03, 1.1326E+03, 1.1419E+03, 1.1513E+03,
     +  1.1606E+03, 1.1699E+03, 1.1793E+03, 1.1886E+03, 1.1979E+03,
     + 1.2073E+03, 1.2166E+03, 1.2259E+03, 1.2353E+03, 1.2446E+03,
     +  1.2539E+03, 1.2633E+03, 1.2726E+03, 1.2820E+03, 1.2913E+03,
     + 1.3006E+03, 1.3100E+03, 1.3193E+03, 1.3286E+03, 1.3380E+03,
     +  1.3473E+03, 1.3566E+03, 1.3660E+03, 1.3753E+03, 1.3847E+03,
     + 1.3940E+03, 1.4033E+03, 1.4127E+03, 1.4220E+03, 1.4314E+03,
     +  1.4407E+03, 1.4500E+03, 1.4594E+03, 1.4687E+03, 1.4781E+03,
     + 1.4874E+03, 1.4968E+03, 1.5061E+03, 1.5155E+03, 1.5248E+03,
     +  1.5341E+03, 1.5435E+03, 1.5528E+03, 1.5622E+03, 1.5715E+03,
     + 1.5809E+03, 1.5902E+03, 1.5996E+03, 1.6089E+03, 1.6183E+03,
     +  1.6276E+03, 1.6370E+03, 1.6464E+03, 1.6557E+03, 1.6651E+03,
     + 1.6744E+03, 1.6838E+03, 1.6932E+03, 1.7025E+03, 1.7119E+03,
     +  1.7212E+03, 1.7306E+03, 1.7400E+03, 1.7493E+03, 1.7587E+03,
     + 1.7681E+03, 1.7774E+03, 1.7868E+03, 1.7962E+03, 1.8056E+03,
     +  1.8149E+03, 1.8243E+03, 1.8337E+03, 1.8431E+03, 1.8525E+03,
     + 1.8618E+03, 1.8712E+03, 1.8806E+03, 1.8900E+03, 1.8994E+03,
     +  1.9088E+03, 1.9182E+03, 1.9276E+03, 1.9370E+03, 1.9464E+03,
     + 1.9558E+03, 1.9652E+03, 1.9746E+03, 1.9840E+03, 1.9934E+03,
     +  2.0028E+03, 2.0122E+03, 2.0216E+03, 2.0311E+03, 2.0405E+03,
     + 2.0499E+03, 2.0593E+03, 2.0687E+03, 2.0782E+03, 2.0876E+03,
     +  2.0970E+03, 2.1065E+03, 2.1159E+03, 2.1254E+03, 2.1348E+03,
     + 2.1443E+03, 2.1537E+03, 2.1632E+03, 2.1726E+03, 2.1821E+03,
     +  2.1915E+03, 2.2010E+03, 2.2105E+03, 2.2199E+03, 2.2294E+03,
     + 2.2389E+03, 2.2484E+03, 2.2578E+03, 2.2673E+03, 2.2768E+03,
     +  2.2863E+03, 2.2958E+03, 2.3053E+03, 2.3148E+03, 2.3243E+03,
     + 2.3338E+03, 2.3433E+03, 2.3529E+03, 2.3624E+03, 2.3719E+03,
     +  2.3814E+03, 2.3910E+03, 2.4005E+03, 2.4100E+03, 2.4196E+03,
     + 2.4291E+03, 2.4387E+03, 2.4482E+03, 2.4578E+03, 2.4674E+03,
     +  2.4769E+03, 2.4865E+03, 2.4961E+03, 2.5056E+03, 2.5152E+03,
     + 2.5248E+03, 2.5344E+03, 2.5440E+03, 2.5536E+03, 2.5632E+03,
     +  2.5728E+03, 2.5824E+03, 2.5921E+03, 2.6017E+03, 2.6113E+03,
     + 2.6210E+03, 2.6306E+03, 2.6402E+03, 2.6499E+03, 2.6595E+03,
     +  2.6692E+03, 2.6789E+03, 2.6885E+03, 2.6982E+03, 2.7079E+03,
     + 2.7176E+03, 2.7273E+03, 2.7370E+03, 2.7467E+03, 2.7564E+03,
     +  2.7661E+03, 2.7758E+03, 2.7855E+03, 2.7952E+03, 2.8050E+03,
     + 2.8147E+03, 2.8245E+03, 2.8342E+03, 2.8440E+03, 2.8537E+03,
     +  2.8635E+03, 2.8733E+03, 2.8831E+03, 2.8928E+03, 2.9026E+03,
     + 2.9124E+03, 2.9222E+03, 2.9320E+03, 2.9419E+03, 2.9517E+03,
     +  2.9615E+03, 2.9713E+03, 2.9812E+03, 2.9910E+03, 3.0009E+03,
     + 3.0107E+03, 3.0206E+03, 3.0305E+03, 3.0403E+03, 3.0502E+03,
     +  3.0601E+03, 3.0700E+03, 3.0799E+03, 3.0898E+03, 3.0998E+03,
     + 3.1097E+03, 3.1196E+03, 3.1296E+03, 3.1395E+03, 3.1494E+03,
     +  3.1594E+03, 3.1694E+03, 3.1793E+03, 3.1893E+03, 3.1993E+03,
     + 3.2093E+03, 3.2193E+03, 3.2293E+03, 3.2393E+03, 3.2494E+03,
     +  3.2594E+03, 3.2694E+03, 3.2795E+03, 3.2895E+03, 3.2996E+03,
     + 3.3096E+03, 3.3197E+03, 3.3298E+03, 3.3399E+03, 3.3500E+03,
     +  3.3601E+03, 3.3702E+03, 3.3803E+03, 3.3904E+03, 3.4006E+03,
     + 3.4107E+03, 3.4209E+03, 3.4310E+03, 3.4412E+03, 3.4514E+03,
     +  3.4615E+03, 3.4717E+03, 3.4819E+03, 3.4921E+03, 3.5023E+03,
     + 3.5126E+03, 3.5228E+03, 3.5330E+03, 3.5433E+03, 3.5535E+03,
     +  3.5638E+03, 3.5741E+03, 3.5843E+03, 3.5946E+03, 3.6049E+03,
     + 3.6152E+03, 3.6255E+03, 3.6358E+03, 3.6462E+03, 3.6565E+03,
     +  3.6668E+03, 3.6772E+03, 3.6876E+03, 3.6979E+03, 3.7083E+03,
     + 3.7187E+03, 3.7291E+03, 3.7395E+03, 3.7499E+03, 3.7603E+03,
     +  3.7708E+03, 3.7812E+03, 3.7916E+03, 3.8021E+03, 3.8126E+03,
     + 3.8230E+03, 3.8335E+03, 3.8440E+03, 3.8545E+03, 3.8650E+03,
     +  3.8755E+03, 3.8861E+03, 3.8966E+03, 3.9071E+03, 3.9177E+03,
     + 3.9283E+03, 3.9388E+03, 3.9494E+03, 3.9600E+03, 3.9706E+03,
     +  3.9812E+03, 3.9918E+03, 4.0025E+03, 4.0131E+03, 4.0237E+03,
     + 4.0344E+03, 4.0451E+03, 4.0557E+03, 4.0664E+03, 4.0771E+03,
     +  4.0878E+03, 4.0985E+03, 4.1092E+03, 4.1200E+03, 4.1307E+03,
     + 4.1414E+03, 4.1522E+03, 4.1630E+03, 4.1737E+03, 4.1845E+03,
     +  4.1953E+03, 4.2061E+03, 4.2169E+03, 4.2278E+03, 4.2386E+03,
     + 4.2495E+03, 4.2603E+03, 4.2712E+03, 4.2820E+03, 4.2929E+03,
     +  4.3038E+03, 4.3147E+03, 4.3256E+03, 4.3366E+03, 4.3475E+03,
     + 4.3584E+03, 4.3694E+03, 4.3803E+03, 4.3913E+03, 4.4023E+03,
     +  4.4133E+03, 4.4243E+03, 4.4353E+03, 4.4463E+03, 4.4573E+03,
     + 4.4684E+03, 4.4794E+03, 4.4905E+03, 4.5016E+03, 4.5127E+03,
     +  4.5237E+03, 4.5348E+03, 4.5460E+03, 4.5571E+03, 4.5682E+03,
     + 4.5794E+03, 4.5905E+03, 4.6017E+03, 4.6128E+03, 4.6240E+03,
     +  4.6352E+03, 4.6464E+03, 4.6576E+03, 4.6689E+03, 4.6801E+03,
     + 4.6913E+03, 4.7026E+03, 4.7139E+03, 4.7251E+03, 4.7364E+03,
     +  4.7477E+03, 4.7590E+03, 4.7704E+03, 4.7817E+03, 4.7930E+03,
     + 4.8044E+03, 4.8157E+03, 4.8271E+03, 4.8385E+03, 4.8499E+03,
     +  4.8613E+03, 4.8727E+03, 4.8841E+03, 4.8955E+03, 4.9070E+03,
     + 4.9184E+03, 4.9299E+03, 4.9414E+03, 4.9529E+03, 4.9644E+03,
     +  4.9759E+03, 4.9874E+03, 4.9989E+03, 5.0105E+03, 5.0220E+03,
     + 5.0336E+03, 5.0452E+03, 5.0567E+03, 5.0683E+03, 5.0799E+03,
     +  5.0916E+03, 5.1032E+03, 5.1148E+03, 5.1265E+03, 5.1381E+03,
     + 5.1498E+03, 5.1615E+03, 5.1732E+03, 5.1849E+03, 5.1966E+03,
     +  5.2083E+03, 5.2201E+03, 5.2318E+03, 5.2436E+03, 5.2553E+03,
     + 5.2671E+03, 5.2789E+03, 5.2907E+03, 5.3025E+03, 5.3143E+03,
     +  5.3262E+03, 5.3380E+03, 5.3499E+03, 5.3618E+03, 5.3736E+03,
     + 5.3855E+03, 5.3974E+03, 5.4093E+03, 5.4213E+03, 5.4332E+03,
     +  5.4452E+03, 5.4571E+03, 5.4691E+03, 5.4811E+03, 5.4931E+03,
     + 5.5051E+03, 5.5171E+03, 5.5291E+03, 5.5411E+03, 5.5532E+03,
     +  5.5652E+03, 5.5773E+03, 5.5894E+03, 5.6015E+03, 5.6136E+03,
     + 5.6257E+03, 5.6378E+03, 5.6500E+03, 5.6621E+03, 5.6743E+03,
     +  5.6864E+03, 5.6986E+03, 5.7108E+03, 5.7230E+03, 5.7352E+03,
     + 5.7475E+03, 5.7597E+03, 5.7720E+03, 5.7842E+03, 5.7965E+03,
     +  5.8088E+03, 5.8211E+03, 5.8334E+03, 5.8457E+03, 5.8581E+03,
     + 5.8704E+03, 5.8828E+03, 5.8951E+03, 5.9075E+03, 5.9199E+03,
     +  5.9323E+03, 5.9447E+03, 5.9571E+03, 5.9696E+03, 5.9820E+03,
     + 5.9945E+03, 6.0069E+03, 6.0194E+03, 6.0319E+03, 6.0444E+03,
     +  6.0569E+03, 6.0695E+03, 6.0820E+03, 6.0946E+03, 6.1071E+03,
     + 6.1197E+03, 6.1323E+03, 6.1449E+03, 6.1575E+03, 6.1701E+03,
     +  6.1827E+03, 6.1954E+03, 6.2080E+03, 6.2207E+03, 6.2334E+03,
     + 6.2461E+03, 6.2588E+03, 6.2715E+03, 6.2842E+03, 6.2970E+03,
     +  6.3097E+03, 6.3225E+03, 6.3353E+03, 6.3480E+03, 6.3608E+03,
     + 6.3737E+03, 6.3865E+03, 6.3993E+03, 6.4122E+03, 6.4250E+03,
     +  6.4379E+03, 6.4508E+03, 6.4636E+03, 6.4766E+03, 6.4895E+03,
     + 6.5024E+03, 6.5153E+03, 6.5283E+03, 6.5412E+03, 6.5542E+03,
     +  6.5672E+03, 6.5802E+03, 6.5932E+03, 6.6062E+03, 6.6193E+03,
     + 6.6323E+03, 6.6454E+03, 6.6585E+03, 6.6715E+03, 6.6846E+03,
     +  6.6977E+03, 6.7109E+03, 6.7240E+03, 6.7371E+03, 6.7503E+03,
     + 6.7634E+03, 6.7766E+03, 6.7898E+03, 6.8030E+03, 6.8162E+03,
     +  6.8295E+03, 6.8427E+03, 6.8559E+03, 6.8692E+03, 6.8825E+03,
     + 6.8958E+03, 6.9091E+03, 6.9224E+03, 6.9357E+03, 6.9490E+03,
     +  6.9624E+03, 6.9757E+03, 6.9891E+03, 7.0025E+03, 7.0159E+03,
     + 7.0293E+03, 7.0427E+03, 7.0561E+03, 7.0696E+03, 7.0830E+03,
     +  7.0965E+03, 7.1100E+03, 7.1235E+03, 7.1370E+03, 7.1505E+03,
     + 7.1640E+03, 7.1776E+03, 7.1911E+03, 7.2047E+03, 7.2183E+03,
     +  7.2318E+03, 7.2454E+03, 7.2591E+03, 7.2727E+03, 7.2863E+03,
     + 7.3000E+03, 7.3136E+03, 7.3273E+03, 7.3410E+03, 7.3547E+03,
     +  7.3684E+03, 7.3821E+03, 7.3959E+03, 7.4096E+03, 7.4234E+03,
     + 7.4371E+03, 7.4509E+03, 7.4647E+03, 7.4785E+03, 7.4923E+03,
     +  7.5062E+03, 7.5200E+03, 7.5339E+03, 7.5478E+03, 7.5616E+03,
     + 7.5755E+03, 7.5894E+03, 7.6034E+03, 7.6173E+03, 7.6312E+03,
     +  7.6452E+03, 7.6591E+03, 7.6731E+03, 7.6871E+03, 7.7011E+03,
     + 7.7151E+03, 7.7292E+03, 7.7432E+03, 7.7573E+03, 7.7713E+03,
     +  7.7854E+03, 7.7995E+03, 7.8136E+03, 7.8277E+03, 7.8418E+03,
     + 7.8560E+03, 7.8701E+03, 7.8843E+03, 7.8985E+03, 7.9127E+03,
     +  7.9269E+03, 7.9411E+03, 7.9553E+03, 7.9696E+03, 7.9838E+03,
     + 7.9981E+03, 8.0123E+03, 8.0266E+03, 8.0409E+03, 8.0553E+03,
     +  8.0696E+03, 8.0839E+03, 8.0983E+03, 8.1126E+03, 8.1270E+03,
     + 8.1414E+03, 8.1558E+03, 8.1702E+03, 8.1846E+03, 8.1991E+03,
     +  8.2135E+03, 8.2280E+03, 8.2425E+03, 8.2569E+03, 8.2714E+03,
     + 8.2860E+03, 8.3005E+03, 8.3150E+03, 8.3296E+03, 8.3441E+03,
     +  8.3587E+03, 8.3733E+03, 8.3879E+03, 8.4025E+03, 8.4171E+03,
     + 8.4318E+03, 8.4464E+03, 8.4611E+03, 8.4758E+03, 8.4904E+03,
     +  8.5051E+03, 8.5199E+03, 8.5346E+03, 8.5493E+03, 8.5641E+03,
     + 8.5788E+03, 8.5936E+03, 8.6084E+03, 8.6232E+03, 8.6380E+03,
     +  8.6528E+03, 8.6677E+03, 8.6825E+03, 8.6974E+03, 8.7123E+03,
     + 8.7271E+03, 8.7420E+03, 8.7570E+03, 8.7719E+03, 8.7868E+03,
     +  8.8018E+03, 8.8167E+03, 8.8317E+03, 8.8467E+03, 8.8617E+03,
     + 8.8767E+03, 8.8917E+03, 8.9068E+03, 8.9218E+03, 8.9369E+03,
     +  8.9520E+03, 8.9671E+03, 8.9822E+03, 8.9973E+03, 9.0124E+03,
     + 9.0275E+03, 9.0427E+03, 9.0579E+03, 9.0730E+03, 9.0882E+03,
     +  9.1034E+03, 9.1186E+03, 9.1339E+03, 9.1491E+03, 9.1644E+03,
     + 9.1796E+03, 9.1949E+03, 9.2102E+03, 9.2255E+03, 9.2408E+03,
     +  9.2562E+03, 9.2715E+03, 9.2868E+03, 9.3022E+03, 9.3176E+03,
     + 9.3330E+03, 9.3484E+03, 9.3638E+03, 9.3792E+03, 9.3947E+03,
     +  9.4101E+03, 9.4256E+03, 9.4411E+03, 9.4566E+03, 9.4721E+03,
     + 9.4876E+03, 9.5031E+03, 9.5187E+03, 9.5342E+03, 9.5498E+03,
     +  9.5654E+03, 9.5810E+03, 9.5966E+03, 9.6122E+03, 9.6279E+03,
     + 9.6435E+03, 9.6592E+03, 9.6748E+03, 9.6905E+03, 9.7062E+03,
     +  9.7219E+03, 9.7376E+03, 9.7534E+03, 9.7691E+03, 9.7849E+03,
     + 9.8007E+03, 9.8165E+03, 9.8323E+03, 9.8481E+03, 9.8639E+03,
     +  9.8797E+03, 9.8956E+03, 9.9114E+03, 9.9273E+03, 9.9432E+03,
     + 9.9591E+03, 9.9750E+03, 9.9909E+03, 1.0007E+04, 1.0023E+04,
     +  1.0039E+04, 1.0055E+04, 1.0071E+04, 1.0087E+04, 1.0103E+04,
     + 1.0119E+04, 1.0135E+04, 1.0151E+04, 1.0167E+04, 1.0183E+04,
     +  1.0199E+04, 1.0215E+04, 1.0231E+04, 1.0248E+04, 1.0264E+04,
     + 1.0280E+04, 1.0296E+04, 1.0312E+04, 1.0328E+04, 1.0345E+04,
     +  1.0361E+04, 1.0377E+04, 1.0393E+04, 1.0410E+04, 1.0426E+04,
     + 1.0442E+04, 1.0458E+04, 1.0475E+04, 1.0491E+04, 1.0507E+04,
     +  1.0524E+04, 1.0540E+04, 1.0556E+04, 1.0573E+04, 1.0589E+04,
     + 1.0606E+04, 1.0622E+04, 1.0639E+04, 1.0655E+04, 1.0671E+04,
     +  1.0688E+04, 1.0704E+04, 1.0721E+04, 1.0737E+04, 1.0754E+04,
     + 1.0770E+04, 1.0787E+04, 1.0804E+04, 1.0820E+04, 1.0837E+04,
     +  1.0853E+04, 1.0870E+04, 1.0887E+04, 1.0903E+04, 1.0920E+04,
     + 1.0937E+04, 1.0953E+04, 1.0970E+04, 1.0987E+04, 1.1003E+04,
     +  1.1020E+04, 1.1037E+04, 1.1054E+04, 1.1070E+04, 1.1087E+04,
     + 1.1104E+04, 1.1121E+04, 1.1138E+04, 1.1155E+04, 1.1171E+04,
     +  1.1188E+04, 1.1205E+04, 1.1222E+04, 1.1239E+04, 1.1256E+04,
     + 1.1273E+04, 1.1290E+04, 1.1307E+04, 1.1324E+04, 1.1341E+04,
     +  1.1358E+04, 1.1375E+04, 1.1392E+04, 1.1409E+04, 1.1426E+04,
     + 1.1443E+04, 1.1460E+04, 1.1477E+04, 1.1494E+04, 1.1511E+04,
     +  1.1528E+04, 1.1545E+04, 1.1563E+04, 1.1580E+04, 1.1597E+04,
     + 1.1614E+04, 1.1631E+04, 1.1648E+04, 1.1666E+04, 1.1683E+04,
     +  1.1700E+04, 1.1717E+04, 1.1735E+04, 1.1752E+04, 1.1769E+04,
     + 1.1787E+04, 1.1804E+04, 1.1821E+04, 1.1839E+04, 1.1856E+04,
     +  1.1873E+04, 1.1891E+04, 1.1908E+04, 1.1926E+04, 1.1943E+04,
     + 1.1961E+04, 1.1978E+04, 1.1995E+04, 1.2013E+04, 1.2030E+04,
     +  1.2048E+04, 1.2066E+04, 1.2083E+04, 1.2101E+04, 1.2118E+04,
     + 1.2136E+04, 1.2153E+04, 1.2171E+04, 1.2189E+04, 1.2206E+04,
     +  1.2224E+04, 1.2241E+04, 1.2259E+04, 1.2277E+04, 1.2295E+04,
     + 1.2312E+04, 1.2330E+04, 1.2348E+04, 1.2365E+04, 1.2383E+04,
     +  1.2401E+04, 1.2419E+04, 1.2437E+04, 1.2454E+04, 1.2472E+04,
     + 1.2490E+04, 1.2508E+04, 1.2526E+04, 1.2544E+04, 1.2562E+04,
     +  1.2579E+04, 1.2597E+04, 1.2615E+04, 1.2633E+04, 1.2651E+04,
     + 1.2669E+04, 1.2687E+04, 1.2705E+04, 1.2723E+04, 1.2741E+04,
     +  1.2759E+04, 1.2777E+04, 1.2795E+04, 1.2813E+04, 1.2831E+04,
     + 1.2850E+04, 1.2868E+04, 1.2886E+04, 1.2904E+04, 1.2922E+04,
     +  1.2940E+04, 1.2958E+04, 1.2977E+04, 1.2995E+04, 1.3013E+04,
     + 1.3031E+04, 1.3049E+04, 1.3068E+04, 1.3086E+04, 1.3104E+04,
     +  1.3123E+04, 1.3141E+04, 1.3159E+04, 1.3178E+04, 1.3196E+04,
     + 1.3214E+04, 1.3233E+04, 1.3251E+04, 1.3269E+04, 1.3288E+04,
     +  1.3306E+04, 1.3325E+04, 1.3343E+04, 1.3362E+04, 1.3380E+04,
     + 1.3399E+04, 1.3417E+04, 1.3436E+04, 1.3454E+04, 1.3473E+04,
     +  1.3491E+04, 1.3510E+04, 1.3528E+04, 1.3547E+04, 1.3566E+04,
     + 1.3584E+04, 1.3603E+04, 1.3622E+04, 1.3640E+04, 1.3659E+04,
     +  1.3678E+04, 1.3696E+04, 1.3715E+04, 1.3734E+04, 1.3752E+04,
     + 1.3771E+04, 1.3790E+04, 1.3809E+04, 1.3828E+04, 1.3846E+04,
     +  1.3865E+04, 1.3884E+04, 1.3903E+04, 1.3922E+04, 1.3941E+04,
     + 1.3959E+04, 1.3978E+04, 1.3997E+04, 1.4016E+04, 1.4035E+04,
     +  1.4054E+04, 1.4073E+04, 1.4092E+04, 1.4111E+04, 1.4130E+04,
     + 1.4149E+04, 1.4168E+04, 1.4187E+04, 1.4206E+04, 1.4225E+04,
     +  1.4244E+04, 1.4263E+04, 1.4283E+04, 1.4302E+04, 1.4321E+04,
     + 1.4340E+04, 1.4359E+04, 1.4378E+04, 1.4397E+04, 1.4417E+04,
     +  1.4436E+04, 1.4455E+04, 1.4474E+04, 1.4494E+04, 1.4513E+04,
     + 1.4532E+04, 1.4551E+04, 1.4571E+04, 1.4590E+04, 1.4609E+04,
     +  1.4629E+04, 1.4648E+04, 1.4668E+04, 1.4687E+04, 1.4706E+04,
     + 1.4726E+04, 1.4745E+04, 1.4765E+04, 1.4784E+04, 1.4804E+04,
     +  1.4823E+04, 1.4842E+04, 1.4862E+04, 1.4882E+04, 1.4901E+04,
     + 1.4921E+04, 1.4940E+04, 1.4960E+04, 1.4979E+04, 1.4999E+04,
     +  1.5019E+04, 1.5038E+04, 1.5058E+04, 1.5077E+04, 1.5097E+04,
     + 1.5117E+04, 1.5136E+04, 1.5156E+04, 1.5176E+04, 1.5196E+04,
     +  1.5215E+04, 1.5235E+04, 1.5255E+04, 1.5275E+04, 1.5294E+04,
     + 1.5314E+04, 1.5334E+04, 1.5354E+04, 1.5374E+04, 1.5394E+04,
     +  1.5414E+04, 1.5433E+04, 1.5453E+04, 1.5473E+04, 1.5493E+04,
     + 1.5513E+04, 1.5533E+04, 1.5553E+04, 1.5573E+04, 1.5593E+04,
     +  1.5613E+04, 1.5633E+04, 1.5653E+04, 1.5673E+04, 1.5693E+04,
     + 1.5713E+04, 1.5733E+04, 1.5753E+04, 1.5774E+04, 1.5794E+04,
     +  1.5814E+04, 1.5834E+04, 1.5854E+04, 1.5874E+04, 1.5895E+04,
     + 1.5915E+04, 1.5935E+04, 1.5955E+04, 1.5975E+04, 1.5996E+04,
     +  1.6016E+04, 1.6036E+04, 1.6057E+04, 1.6077E+04, 1.6097E+04,
     + 1.6118E+04, 1.6138E+04, 1.6158E+04, 1.6179E+04, 1.6199E+04,
     +  1.6219E+04, 1.6240E+04, 1.6260E+04, 1.6281E+04, 1.6301E+04,
     + 1.6322E+04, 1.6342E+04, 1.6363E+04, 1.6383E+04, 1.6404E+04,
     +  1.6424E+04, 1.6445E+04, 1.6465E+04, 1.6486E+04, 1.6507E+04,
     + 1.6527E+04, 1.6548E+04, 1.6568E+04, 1.6589E+04, 1.6610E+04,
     +  1.6630E+04, 1.6651E+04, 1.6672E+04, 1.6693E+04, 1.6713E+04,
     + 1.6734E+04, 1.6755E+04, 1.6775E+04, 1.6796E+04, 1.6817E+04,
     +  1.6838E+04, 1.6859E+04, 1.6880E+04, 1.6900E+04, 1.6921E+04,
     + 1.6942E+04, 1.6963E+04, 1.6984E+04, 1.7005E+04, 1.7026E+04,
     +  1.7047E+04, 1.7068E+04, 1.7089E+04, 1.7110E+04, 1.7131E+04,
     + 1.7152E+04, 1.7173E+04, 1.7194E+04, 1.7215E+04, 1.7236E+04,
     +  1.7257E+04, 1.7278E+04, 1.7299E+04, 1.7320E+04, 1.7341E+04,
     + 1.7362E+04, 1.7384E+04, 1.7405E+04, 1.7426E+04, 1.7447E+04,
     +  1.7468E+04, 1.7490E+04, 1.7511E+04, 1.7532E+04, 1.7553E+04,
     + 1.7575E+04, 1.7596E+04, 1.7617E+04, 1.7638E+04, 1.7660E+04,
     +  1.7681E+04, 1.7702E+04, 1.7724E+04, 1.7745E+04, 1.7767E+04,
     + 1.7788E+04, 1.7809E+04, 1.7831E+04, 1.7852E+04, 1.7874E+04,
     +  1.7895E+04, 1.7917E+04, 1.7938E+04, 1.7960E+04, 1.7981E+04,
     + 1.8003E+04, 1.8024E+04, 1.8046E+04, 1.8068E+04, 1.8089E+04,
     +  1.8111E+04, 1.8132E+04, 1.8154E+04, 1.8176E+04, 1.8197E+04,
     + 1.8219E+04, 1.8241E+04, 1.8262E+04, 1.8284E+04, 1.8306E+04,
     +  1.8328E+04, 1.8349E+04, 1.8371E+04, 1.8393E+04, 1.8415E+04,
     + 1.8436E+04, 1.8458E+04, 1.8480E+04, 1.8502E+04, 1.8524E+04,
     +  1.8546E+04, 1.8568E+04, 1.8589E+04, 1.8611E+04, 1.8633E+04,
     + 1.8655E+04, 1.8677E+04, 1.8699E+04, 1.8721E+04, 1.8743E+04,
     +  1.8765E+04, 1.8787E+04, 1.8809E+04, 1.8831E+04, 1.8853E+04,
     + 1.8875E+04, 1.8898E+04, 1.8920E+04, 1.8942E+04, 1.8964E+04,
     +  1.8986E+04, 1.9008E+04, 1.9030E+04, 1.9053E+04, 1.9075E+04,
     + 1.9097E+04, 1.9119E+04, 1.9141E+04, 1.9164E+04, 1.9186E+04,
     +  1.9208E+04, 1.9231E+04, 1.9253E+04, 1.9275E+04, 1.9297E+04,
     + 1.9320E+04, 1.9342E+04, 1.9365E+04, 1.9387E+04, 1.9409E+04,
     +  1.9432E+04, 1.9454E+04, 1.9477E+04, 1.9499E+04, 1.9522E+04,
     + 1.9544E+04, 1.9567E+04, 1.9589E+04, 1.9612E+04, 1.9634E+04,
     +  1.9657E+04, 1.9679E+04, 1.9702E+04, 1.9724E+04, 1.9747E+04,
     + 1.9770E+04, 1.9792E+04, 1.9815E+04, 1.9838E+04, 1.9860E+04,
     +  1.9883E+04, 1.9906E+04, 1.9928E+04, 1.9951E+04, 1.9974E+04,
     + 1.9996E+04, 2.0019E+04, 2.0042E+04, 2.0065E+04, 2.0088E+04,
     +  2.0110E+04, 2.0133E+04, 2.0156E+04, 2.0179E+04, 2.0202E+04,
     + 2.0225E+04, 2.0248E+04, 2.0271E+04, 2.0293E+04, 2.0316E+04,
     +  2.0339E+04, 2.0362E+04, 2.0385E+04, 2.0408E+04, 2.0431E+04,
     + 2.0454E+04, 2.0477E+04, 2.0500E+04, 2.0523E+04, 2.0547E+04,
     +  2.0570E+04, 2.0593E+04, 2.0616E+04, 2.0639E+04, 2.0662E+04,
     + 2.0685E+04, 2.0708E+04, 2.0732E+04, 2.0755E+04, 2.0778E+04,
     +  2.0801E+04, 2.0824E+04, 2.0848E+04, 2.0871E+04, 2.0894E+04,
     + 2.0918E+04, 2.0941E+04, 2.0964E+04, 2.0988E+04, 2.1011E+04,
     +  2.1034E+04, 2.1058E+04, 2.1081E+04, 2.1104E+04, 2.1128E+04,
     + 2.1151E+04, 2.1175E+04, 2.1198E+04, 2.1222E+04, 2.1245E+04,
     +  2.1269E+04, 2.1292E+04, 2.1316E+04, 2.1339E+04, 2.1363E+04,
     + 2.1386E+04, 2.1410E+04, 2.1433E+04, 2.1457E+04, 2.1481E+04,
     +  2.1504E+04, 2.1528E+04, 2.1552E+04, 2.1575E+04, 2.1599E+04,
     + 2.1623E+04, 2.1646E+04, 2.1670E+04, 2.1694E+04, 2.1717E+04,
     +  2.1741E+04, 2.1765E+04, 2.1789E+04, 2.1813E+04, 2.1836E+04,
     + 2.1860E+04, 2.1884E+04, 2.1908E+04, 2.1932E+04, 2.1956E+04,
     +  2.1980E+04, 2.2004E+04, 2.2027E+04, 2.2051E+04, 2.2075E+04,
     + 2.2099E+04, 2.2123E+04, 2.2147E+04, 2.2171E+04, 2.2195E+04,
     +  2.2219E+04, 2.2243E+04, 2.2268E+04, 2.2292E+04, 2.2316E+04,
     + 2.2340E+04, 2.2364E+04, 2.2388E+04, 2.2412E+04, 2.2436E+04,
     +  2.2460E+04, 2.2485E+04, 2.2509E+04, 2.2533E+04, 2.2557E+04,
     + 2.2582E+04, 2.2606E+04, 2.2630E+04, 2.2654E+04, 2.2679E+04,
     +  2.2703E+04, 2.2727E+04, 2.2752E+04, 2.2776E+04, 2.2800E+04,
     + 2.2825E+04, 2.2849E+04, 2.2873E+04, 2.2898E+04, 2.2922E+04,
     +  2.2947E+04, 2.2971E+04, 2.2996E+04, 2.3020E+04, 2.3045E+04,
     + 2.3069E+04, 2.3094E+04, 2.3118E+04, 2.3143E+04, 2.3167E+04,
     +  2.3192E+04, 2.3217E+04, 2.3241E+04, 2.3266E+04, 2.3290E+04,
     + 2.3315E+04, 2.3340E+04, 2.3364E+04, 2.3389E+04, 2.3414E+04,
     +  2.3439E+04, 2.3463E+04, 2.3488E+04, 2.3513E+04, 2.3538E+04,
     + 2.3562E+04, 2.3587E+04, 2.3612E+04, 2.3637E+04, 2.3662E+04,
     +  2.3686E+04, 2.3711E+04, 2.3736E+04, 2.3761E+04, 2.3786E+04,
     + 2.3811E+04, 2.3836E+04, 2.3861E+04, 2.3886E+04, 2.3911E+04,
     +  2.3936E+04, 2.3961E+04, 2.3986E+04, 2.4011E+04, 2.4036E+04,
     + 2.4061E+04, 2.4086E+04, 2.4111E+04, 2.4136E+04, 2.4161E+04,
     +  2.4186E+04, 2.4211E+04, 2.4237E+04, 2.4262E+04, 2.4287E+04,
     + 2.4312E+04, 2.4337E+04, 2.4363E+04, 2.4388E+04, 2.4413E+04,
     +  2.4438E+04, 2.4464E+04, 2.4489E+04, 2.4514E+04, 2.4539E+04,
     + 2.4565E+04, 2.4590E+04, 2.4616E+04, 2.4641E+04, 2.4666E+04,
     +  2.4692E+04, 2.4717E+04, 2.4742E+04, 2.4768E+04, 2.4793E+04,
     + 2.4819E+04, 2.4844E+04, 2.4870E+04, 2.4895E+04, 2.4921E+04,
     +  2.4946E+04, 2.4972E+04, 2.4998E+04, 2.5023E+04, 2.5049E+04,
     + 2.5074E+04, 2.5100E+04, 2.5126E+04, 2.5151E+04, 2.5177E+04,
     +  2.5202E+04, 2.5228E+04, 2.5254E+04, 2.5280E+04, 2.5305E+04,
     + 2.5331E+04, 2.5357E+04, 2.5383E+04, 2.5408E+04, 2.5434E+04,
     +  2.5460E+04, 2.5486E+04, 2.5512E+04, 2.5537E+04, 2.5563E+04,
     + 2.5589E+04, 2.5615E+04, 2.5641E+04, 2.5667E+04, 2.5693E+04,
     +  2.5719E+04, 2.5745E+04, 2.5771E+04, 2.5797E+04, 2.5823E+04,
     + 2.5849E+04, 2.5875E+04, 2.5901E+04, 2.5927E+04, 2.5953E+04,
     +  2.5979E+04, 2.6005E+04, 2.6031E+04, 2.6057E+04, 2.6083E+04,
     + 2.6110E+04, 2.6136E+04, 2.6162E+04, 2.6188E+04, 2.6214E+04,
     +  2.6241E+04, 2.6267E+04, 2.6293E+04, 2.6319E+04, 2.6346E+04,
     + 2.6372E+04, 2.6398E+04, 2.6424E+04, 2.6451E+04, 2.6477E+04,
     +  2.6504E+04, 2.6530E+04, 2.6556E+04, 2.6583E+04, 2.6609E+04,
     + 2.6636E+04, 2.6662E+04, 2.6688E+04, 2.6715E+04, 2.6741E+04,
     +  2.6768E+04, 2.6794E+04, 2.6821E+04, 2.6847E+04, 2.6874E+04,
     + 2.6901E+04, 2.6927E+04, 2.6954E+04, 2.6980E+04, 2.7007E+04,
     +  2.7034E+04, 2.7060E+04, 2.7087E+04, 2.7114E+04, 2.7140E+04,
     + 2.7167E+04, 2.7194E+04, 2.7220E+04, 2.7247E+04, 2.7274E+04,
     +  2.7301E+04, 2.7327E+04, 2.7354E+04, 2.7381E+04, 2.7408E+04,
     + 2.7435E+04, 2.7462E+04, 2.7488E+04, 2.7515E+04, 2.7542E+04,
     +  2.7569E+04, 2.7596E+04, 2.7623E+04, 2.7650E+04, 2.7677E+04,
     + 2.7704E+04, 2.7731E+04, 2.7758E+04, 2.7785E+04, 2.7812E+04,
     +  2.7839E+04, 2.7866E+04, 2.7893E+04, 2.7920E+04, 2.7947E+04,
     + 2.7974E+04, 2.8001E+04, 2.8029E+04, 2.8056E+04, 2.8083E+04,
     +  2.8110E+04, 2.8137E+04, 2.8165E+04, 2.8192E+04, 2.8219E+04,
     + 2.8246E+04, 2.8273E+04, 2.8301E+04, 2.8328E+04, 2.8355E+04,
     +  2.8383E+04, 2.8410E+04, 2.8437E+04, 2.8465E+04, 2.8492E+04,
     + 2.8519E+04, 2.8547E+04, 2.8574E+04, 2.8602E+04, 2.8629E+04,
     +  2.8657E+04, 2.8684E+04, 2.8712E+04, 2.8739E+04, 2.8767E+04,
     + 2.8794E+04, 2.8822E+04, 2.8849E+04, 2.8877E+04, 2.8904E+04,
     +  2.8932E+04, 2.8960E+04, 2.8987E+04, 2.9015E+04, 2.9043E+04,
     + 2.9070E+04, 2.9098E+04, 2.9126E+04, 2.9153E+04, 2.9181E+04,
     +  2.9209E+04, 2.9236E+04, 2.9264E+04, 2.9292E+04, 2.9320E+04,
     + 2.9348E+04, 2.9375E+04, 2.9403E+04, 2.9431E+04, 2.9459E+04,
     +  2.9487E+04, 2.9515E+04, 2.9543E+04, 2.9571E+04, 2.9598E+04,
     + 2.9626E+04, 2.9654E+04, 2.9682E+04, 2.9710E+04, 2.9738E+04,
     +  2.9766E+04, 2.9794E+04, 2.9822E+04, 2.9850E+04, 2.9879E+04,
     + 2.9907E+04, 2.9935E+04, 2.9963E+04, 2.9991E+04, 3.0019E+04,
     +  3.0047E+04, 3.0075E+04, 3.0104E+04, 3.0132E+04, 3.0160E+04,
     + 3.0188E+04, 3.0216E+04, 3.0245E+04, 3.0273E+04, 3.0301E+04,
     +  3.0329E+04, 3.0358E+04, 3.0386E+04, 3.0414E+04, 3.0443E+04,
     + 3.0471E+04, 3.0499E+04, 3.0528E+04, 3.0556E+04, 3.0585E+04,
     +  3.0613E+04, 3.0642E+04, 3.0670E+04, 3.0698E+04, 3.0727E+04,
     + 3.0755E+04, 3.0784E+04, 3.0812E+04, 3.0841E+04, 3.0870E+04,
     +  3.0898E+04, 3.0927E+04, 3.0955E+04, 3.0984E+04, 3.1013E+04,
     + 3.1041E+04, 3.1070E+04, 3.1098E+04, 3.1127E+04, 3.1156E+04,
     +  3.1185E+04, 3.1213E+04, 3.1242E+04, 3.1271E+04, 3.1300E+04,
     + 3.1328E+04, 3.1357E+04, 3.1386E+04, 3.1415E+04, 3.1444E+04,
     +  3.1472E+04, 3.1501E+04, 3.1530E+04, 3.1559E+04, 3.1588E+04,
     + 3.1617E+04, 3.1646E+04, 3.1675E+04, 3.1704E+04, 3.1733E+04,
     +  3.1762E+04, 3.1791E+04, 3.1820E+04, 3.1849E+04, 3.1878E+04,
     + 3.1907E+04, 3.1936E+04, 3.1965E+04, 3.1994E+04, 3.2023E+04,
     +  3.2052E+04, 3.2081E+04, 3.2110E+04, 3.2140E+04, 3.2169E+04,
     + 3.2198E+04, 3.2227E+04, 3.2256E+04, 3.2286E+04, 3.2315E+04,
     +  3.2344E+04, 3.2373E+04, 3.2403E+04, 3.2432E+04, 3.2461E+04,
     + 3.2491E+04, 3.2520E+04, 3.2549E+04, 3.2579E+04, 3.2608E+04,
     +  3.2638E+04, 3.2667E+04, 3.2696E+04, 3.2726E+04, 3.2755E+04,
     + 3.2785E+04, 3.2814E+04, 3.2844E+04, 3.2873E+04, 3.2903E+04,
     +  3.2932E+04, 3.2962E+04, 3.2991E+04, 3.3021E+04, 3.3051E+04,
     + 3.3080E+04, 3.3110E+04, 3.3139E+04, 3.3169E+04, 3.3199E+04,
     +  3.3228E+04, 3.3258E+04, 3.3288E+04, 3.3318E+04, 3.3347E+04,
     + 3.3377E+04, 3.3407E+04, 3.3437E+04, 3.3466E+04, 3.3496E+04,
     +  3.3526E+04, 3.3556E+04, 3.3586E+04, 3.3616E+04, 3.3645E+04,
     + 3.3675E+04, 3.3705E+04, 3.3735E+04, 3.3765E+04, 3.3795E+04,
     +  3.3825E+04, 3.3855E+04, 3.3885E+04, 3.3915E+04, 3.3945E+04,
     + 3.3975E+04, 3.4005E+04, 3.4035E+04, 3.4065E+04, 3.4095E+04,
     +  3.4125E+04, 3.4155E+04, 3.4186E+04, 3.4216E+04, 3.4246E+04,
     + 3.4276E+04, 3.4306E+04, 3.4336E+04, 3.4367E+04, 3.4397E+04,
     +  3.4427E+04, 3.4457E+04, 3.4488E+04, 3.4518E+04, 3.4548E+04,
     + 3.4579E+04, 3.4609E+04, 3.4639E+04, 3.4670E+04, 3.4700E+04,
     +  3.4730E+04, 3.4761E+04, 3.4791E+04, 3.4821E+04, 3.4852E+04,
     + 3.4882E+04, 3.4913E+04, 3.4943E+04, 3.4974E+04, 3.5004E+04,
     +  3.5035E+04, 3.5065E+04, 3.5096E+04, 3.5126E+04, 3.5157E+04,
     + 3.5188E+04, 3.5218E+04, 3.5249E+04, 3.5279E+04, 3.5310E+04,
     +  3.5341E+04, 3.5371E+04, 3.5402E+04, 3.5433E+04, 3.5464E+04,
     + 3.5494E+04, 3.5525E+04, 3.5556E+04, 3.5587E+04, 3.5617E+04,
     +  3.5648E+04, 3.5679E+04, 3.5710E+04, 3.5741E+04, 3.5771E+04,
     + 3.5802E+04, 3.5833E+04, 3.5864E+04, 3.5895E+04, 3.5926E+04,
     +  3.5957E+04, 3.5988E+04, 3.6019E+04, 3.6050E+04, 3.6081E+04,
     + 3.6112E+04, 3.6143E+04, 3.6174E+04, 3.6205E+04, 3.6236E+04,
     +  3.6267E+04, 3.6298E+04, 3.6329E+04, 3.6360E+04, 3.6392E+04,
     + 3.6423E+04, 3.6454E+04, 3.6485E+04, 3.6516E+04, 3.6547E+04,
     +  3.6579E+04, 3.6610E+04, 3.6641E+04, 3.6672E+04, 3.6704E+04,
     + 3.6735E+04, 3.6766E+04, 3.6798E+04, 3.6829E+04, 3.6860E+04,
     +  3.6892E+04, 3.6923E+04, 3.6954E+04, 3.6986E+04, 3.7017E+04,
     + 3.7049E+04, 3.7080E+04, 3.7112E+04, 3.7143E+04, 3.7174E+04,
     +  3.7206E+04, 3.7238E+04, 3.7269E+04, 3.7301E+04, 3.7332E+04,
     + 3.7364E+04, 3.7395E+04, 3.7427E+04, 3.7459E+04, 3.7490E+04,
     +  3.7522E+04, 3.7553E+04, 3.7585E+04, 3.7617E+04, 3.7648E+04,
     + 3.7680E+04, 3.7712E+04, 3.7744E+04, 3.7775E+04, 3.7807E+04,
     +  3.7839E+04, 3.7871E+04, 3.7903E+04, 3.7934E+04, 3.7966E+04,
     + 3.7998E+04, 3.8030E+04, 3.8062E+04, 3.8094E+04, 3.8126E+04,
     +  3.8158E+04, 3.8190E+04, 3.8221E+04, 3.8253E+04, 3.8285E+04,
     + 3.8317E+04, 3.8349E+04, 3.8381E+04, 3.8413E+04, 3.8446E+04,
     +  3.8478E+04, 3.8510E+04, 3.8542E+04, 3.8574E+04, 3.8606E+04,
     + 3.8638E+04, 3.8670E+04, 3.8702E+04, 3.8735E+04, 3.8767E+04,
     +  3.8799E+04, 3.8831E+04, 3.8863E+04, 3.8896E+04, 3.8928E+04,
     + 3.8960E+04, 3.8993E+04, 3.9025E+04, 3.9057E+04, 3.9090E+04,
     +  3.9122E+04, 3.9154E+04, 3.9187E+04, 3.9219E+04, 3.9251E+04,
     + 3.9284E+04, 3.9316E+04, 3.9349E+04, 3.9381E+04, 3.9414E+04,
     +  3.9446E+04, 3.9479E+04, 3.9511E+04, 3.9544E+04, 3.9576E+04,
     + 3.9609E+04, 3.9641E+04, 3.9674E+04, 3.9707E+04, 3.9739E+04,
     +  3.9772E+04, 3.9804E+04, 3.9837E+04, 3.9870E+04, 3.9902E+04,
     + 3.9935E+04, 3.9968E+04, 4.0001E+04, 4.0033E+04, 4.0066E+04,
     +  4.0099E+04, 4.0132E+04, 4.0164E+04, 4.0197E+04, 4.0230E+04,
     + 4.0263E+04, 4.0296E+04, 4.0329E+04, 4.0362E+04, 4.0394E+04,
     +  4.0427E+04, 4.0460E+04, 4.0493E+04, 4.0526E+04, 4.0559E+04,
     + 4.0592E+04, 4.0625E+04, 4.0658E+04, 4.0691E+04, 4.0724E+04,
     +  4.0757E+04, 4.0790E+04, 4.0823E+04, 4.0857E+04, 4.0890E+04,
     + 4.0923E+04, 4.0956E+04, 4.0989E+04, 4.1022E+04, 4.1055E+04,
     +  4.1089E+04, 4.1122E+04, 4.1155E+04, 4.1188E+04, 4.1222E+04,
     + 4.1255E+04, 4.1288E+04, 4.1321E+04, 4.1355E+04, 4.1388E+04,
     +  4.1421E+04, 4.1455E+04, 4.1488E+04, 4.1521E+04, 4.1555E+04,
     + 4.1588E+04, 4.1622E+04, 4.1655E+04, 4.1689E+04, 4.1722E+04,
     +  4.1756E+04, 4.1789E+04, 4.1823E+04, 4.1856E+04, 4.1890E+04,
     + 4.1923E+04, 4.1957E+04, 4.1990E+04, 4.2024E+04, 4.2058E+04,
     +  4.2091E+04, 4.2125E+04, 4.2158E+04, 4.2192E+04, 4.2226E+04,
     + 4.2259E+04, 4.2293E+04, 4.2327E+04, 4.2361E+04, 4.2394E+04,
     +  4.2428E+04, 4.2462E+04, 4.2496E+04, 4.2530E+04, 4.2563E+04,
     + 4.2597E+04, 4.2631E+04, 4.2665E+04, 4.2699E+04, 4.2733E+04,
     +  4.2767E+04, 4.2800E+04, 4.2834E+04, 4.2868E+04, 4.2902E+04,
     + 4.2936E+04, 4.2970E+04, 4.3004E+04, 4.3038E+04, 4.3072E+04,
     +  4.3106E+04, 4.3140E+04, 4.3175E+04, 4.3209E+04, 4.3243E+04,
     + 4.3277E+04, 4.3311E+04, 4.3345E+04, 4.3379E+04, 4.3414E+04,
     +  4.3448E+04, 4.3482E+04, 4.3516E+04, 4.3550E+04, 4.3585E+04,
     + 4.3619E+04, 4.3653E+04, 4.3687E+04, 4.3722E+04, 4.3756E+04,
     +  4.3790E+04, 4.3825E+04, 4.3859E+04, 4.3894E+04, 4.3928E+04,
     + 4.3962E+04, 4.3997E+04, 4.4031E+04, 4.4066E+04, 4.4100E+04,
     +  4.4135E+04, 4.4169E+04, 4.4204E+04, 4.4238E+04, 4.4273E+04,
     + 4.4307E+04, 4.4342E+04, 4.4376E+04, 4.4411E+04, 4.4445E+04,
     +  4.4480E+04, 4.4515E+04, 4.4549E+04, 4.4584E+04, 4.4619E+04,
     + 4.4653E+04, 4.4688E+04, 4.4723E+04, 4.4758E+04, 4.4792E+04,
     +  4.4827E+04, 4.4862E+04, 4.4897E+04, 4.4931E+04, 4.4966E+04,
     + 4.5001E+04, 4.5036E+04, 4.5071E+04, 4.5106E+04, 4.5141E+04,
     +  4.5175E+04, 4.5210E+04, 4.5245E+04, 4.5280E+04, 4.5315E+04,
     + 4.5350E+04, 4.5385E+04, 4.5420E+04, 4.5455E+04, 4.5490E+04,
     +  4.5525E+04, 4.5560E+04, 4.5595E+04, 4.5631E+04, 4.5666E+04,
     + 4.5701E+04, 4.5736E+04, 4.5771E+04, 4.5806E+04, 4.5841E+04,
     +  4.5877E+04, 4.5912E+04, 4.5947E+04, 4.5982E+04, 4.6018E+04,
     + 4.6053E+04, 4.6088E+04, 4.6123E+04, 4.6159E+04, 4.6194E+04,
     +  4.6229E+04, 4.6265E+04, 4.6300E+04, 4.6335E+04, 4.6371E+04,
     + 4.6406E+04, 4.6442E+04, 4.6477E+04, 4.6513E+04, 4.6548E+04,
     +  4.6583E+04, 4.6619E+04, 4.6654E+04, 4.6690E+04, 4.6726E+04,
     + 4.6761E+04, 4.6797E+04, 4.6832E+04, 4.6868E+04, 4.6903E+04,
     +  4.6939E+04, 4.6975E+04, 4.7010E+04, 4.7046E+04, 4.7082E+04,
     + 4.7117E+04, 4.7153E+04, 4.7189E+04, 4.7224E+04, 4.7260E+04,
     +  4.7296E+04, 4.7332E+04, 4.7368E+04, 4.7403E+04, 4.7439E+04,
     + 4.7475E+04, 4.7511E+04, 4.7547E+04, 4.7583E+04, 4.7619E+04,
     +  4.7654E+04, 4.7690E+04, 4.7726E+04, 4.7762E+04, 4.7798E+04,
     + 4.7834E+04, 4.7870E+04, 4.7906E+04, 4.7942E+04, 4.7978E+04,
     +  4.8014E+04, 4.8050E+04, 4.8086E+04, 4.8123E+04, 4.8159E+04,
     + 4.8195E+04, 4.8231E+04, 4.8267E+04, 4.8303E+04, 4.8339E+04,
     +  4.8376E+04, 4.8412E+04, 4.8448E+04, 4.8484E+04, 4.8521E+04,
     + 4.8557E+04, 4.8593E+04, 4.8629E+04, 4.8666E+04, 4.8702E+04,
     +  4.8738E+04, 4.8775E+04, 4.8811E+04, 4.8847E+04, 4.8884E+04,
     + 4.8920E+04, 4.8957E+04, 4.8993E+04, 4.9030E+04, 4.9066E+04,
     +  4.9103E+04, 4.9139E+04, 4.9176E+04, 4.9212E+04, 4.9249E+04,
     + 4.9285E+04, 4.9322E+04, 4.9358E+04, 4.9395E+04, 4.9432E+04,
     +  4.9468E+04, 4.9505E+04, 4.9541E+04, 4.9578E+04, 4.9615E+04,
     + 4.9652E+04, 4.9688E+04, 4.9725E+04, 4.9762E+04, 4.9798E+04,
     +  4.9835E+04, 4.9872E+04, 4.9909E+04, 4.9946E+04, 4.9982E+04/
      data (QofT( 6,J),J=2409,4510)/ 5.0019E+04, 5.0056E+04,
     +  5.0093E+04, 5.0130E+04, 5.0167E+04, 5.0204E+04, 5.0241E+04,
     +  5.0278E+04,
     + 5.0315E+04, 5.0352E+04, 5.0389E+04, 5.0426E+04, 5.0463E+04,
     +  5.0500E+04, 5.0537E+04, 5.0574E+04, 5.0611E+04, 5.0648E+04,
     + 5.0685E+04, 5.0722E+04, 5.0759E+04, 5.0796E+04, 5.0834E+04,
     +  5.0871E+04, 5.0908E+04, 5.0945E+04, 5.0982E+04, 5.1020E+04,
     + 5.1057E+04, 5.1094E+04, 5.1131E+04, 5.1169E+04, 5.1206E+04,
     +  5.1243E+04, 5.1281E+04, 5.1318E+04, 5.1355E+04, 5.1393E+04,
     + 5.1430E+04, 5.1467E+04, 5.1505E+04, 5.1542E+04, 5.1580E+04,
     +  5.1617E+04, 5.1655E+04, 5.1692E+04, 5.1730E+04, 5.1767E+04,
     + 5.1805E+04, 5.1842E+04, 5.1880E+04, 5.1917E+04, 5.1955E+04,
     +  5.1993E+04, 5.2030E+04, 5.2068E+04, 5.2106E+04, 5.2143E+04,
     + 5.2181E+04, 5.2219E+04, 5.2256E+04, 5.2294E+04, 5.2332E+04,
     +  5.2370E+04, 5.2407E+04, 5.2445E+04, 5.2483E+04, 5.2521E+04,
     + 5.2559E+04, 5.2596E+04, 5.2634E+04, 5.2672E+04, 5.2710E+04,
     +  5.2748E+04, 5.2786E+04, 5.2824E+04, 5.2862E+04, 5.2900E+04,
     + 5.2938E+04, 5.2976E+04, 5.3014E+04, 5.3052E+04, 5.3090E+04,
     +  5.3128E+04, 5.3166E+04, 5.3204E+04, 5.3242E+04, 5.3280E+04,
     + 5.3318E+04, 5.3356E+04, 5.3394E+04, 5.3432E+04, 5.3471E+04,
     +  5.3509E+04, 5.3547E+04, 5.3585E+04, 5.3624E+04, 5.3662E+04,
     + 5.3700E+04, 5.3738E+04, 5.3777E+04, 5.3815E+04, 5.3853E+04,
     +  5.3892E+04, 5.3930E+04, 5.3968E+04, 5.4007E+04, 5.4045E+04,
     + 5.4083E+04, 5.4122E+04, 5.4160E+04, 5.4199E+04, 5.4237E+04,
     +  5.4276E+04, 5.4314E+04, 5.4353E+04, 5.4391E+04, 5.4430E+04,
     + 5.4468E+04, 5.4507E+04, 5.4545E+04, 5.4584E+04, 5.4623E+04,
     +  5.4661E+04, 5.4700E+04, 5.4739E+04, 5.4777E+04, 5.4816E+04,
     + 5.4855E+04, 5.4893E+04, 5.4932E+04, 5.4971E+04, 5.5009E+04,
     +  5.5048E+04, 5.5087E+04, 5.5126E+04, 5.5165E+04, 5.5203E+04,
     + 5.5242E+04, 5.5281E+04, 5.5320E+04, 5.5359E+04, 5.5398E+04,
     +  5.5437E+04, 5.5476E+04, 5.5515E+04, 5.5554E+04, 5.5593E+04,
     + 5.5632E+04, 5.5671E+04, 5.5710E+04, 5.5749E+04, 5.5788E+04,
     +  5.5827E+04, 5.5866E+04, 5.5905E+04, 5.5944E+04, 5.5983E+04,
     + 5.6022E+04, 5.6061E+04, 5.6100E+04, 5.6140E+04, 5.6179E+04,
     +  5.6218E+04, 5.6257E+04, 5.6296E+04, 5.6336E+04, 5.6375E+04,
     + 5.6414E+04, 5.6454E+04, 5.6493E+04, 5.6532E+04, 5.6572E+04,
     +  5.6611E+04, 5.6650E+04, 5.6690E+04, 5.6729E+04, 5.6768E+04,
     + 5.6808E+04, 5.6847E+04, 5.6887E+04, 5.6926E+04, 5.6966E+04,
     +  5.7005E+04, 5.7045E+04, 5.7084E+04, 5.7124E+04, 5.7163E+04,
     + 5.7203E+04, 5.7242E+04, 5.7282E+04, 5.7322E+04, 5.7361E+04,
     +  5.7401E+04, 5.7441E+04, 5.7480E+04, 5.7520E+04, 5.7560E+04,
     + 5.7599E+04, 5.7639E+04, 5.7679E+04, 5.7719E+04, 5.7758E+04,
     +  5.7798E+04, 5.7838E+04, 5.7878E+04, 5.7918E+04, 5.7957E+04,
     + 5.7997E+04, 5.8037E+04, 5.8077E+04, 5.8117E+04, 5.8157E+04,
     +  5.8197E+04, 5.8237E+04, 5.8277E+04, 5.8317E+04, 5.8357E+04,
     + 5.8397E+04, 5.8437E+04, 5.8477E+04, 5.8517E+04, 5.8557E+04,
     +  5.8597E+04, 5.8637E+04, 5.8677E+04, 5.8717E+04, 5.8757E+04,
     + 5.8798E+04, 5.8838E+04, 5.8878E+04, 5.8918E+04, 5.8958E+04,
     +  5.8999E+04, 5.9039E+04, 5.9079E+04, 5.9119E+04, 5.9160E+04,
     + 5.9200E+04, 5.9240E+04, 5.9281E+04, 5.9321E+04, 5.9361E+04,
     +  5.9402E+04, 5.9442E+04, 5.9482E+04, 5.9523E+04, 5.9563E+04,
     + 5.9604E+04, 5.9644E+04, 5.9685E+04, 5.9725E+04, 5.9766E+04,
     +  5.9806E+04, 5.9847E+04, 5.9887E+04, 5.9928E+04, 5.9968E+04,
     + 6.0009E+04, 6.0050E+04, 6.0090E+04, 6.0131E+04, 6.0171E+04,
     +  6.0212E+04, 6.0253E+04, 6.0294E+04, 6.0334E+04, 6.0375E+04,
     + 6.0416E+04, 6.0456E+04, 6.0497E+04, 6.0538E+04, 6.0579E+04,
     +  6.0620E+04, 6.0660E+04, 6.0701E+04, 6.0742E+04, 6.0783E+04,
     + 6.0824E+04, 6.0865E+04, 6.0906E+04, 6.0947E+04, 6.0988E+04,
     +  6.1028E+04, 6.1069E+04, 6.1110E+04, 6.1151E+04, 6.1192E+04,
     + 6.1233E+04, 6.1275E+04, 6.1316E+04, 6.1357E+04, 6.1398E+04,
     +  6.1439E+04, 6.1480E+04, 6.1521E+04, 6.1562E+04, 6.1603E+04,
     + 6.1645E+04, 6.1686E+04, 6.1727E+04, 6.1768E+04, 6.1809E+04,
     +  6.1851E+04, 6.1892E+04, 6.1933E+04, 6.1975E+04, 6.2016E+04,
     + 6.2057E+04, 6.2099E+04, 6.2140E+04, 6.2181E+04, 6.2223E+04,
     +  6.2264E+04, 6.2305E+04, 6.2347E+04, 6.2388E+04, 6.2430E+04,
     + 6.2471E+04, 6.2513E+04, 6.2554E+04, 6.2596E+04, 6.2637E+04,
     +  6.2679E+04, 6.2720E+04, 6.2762E+04, 6.2804E+04, 6.2845E+04,
     + 6.2887E+04, 6.2928E+04, 6.2970E+04, 6.3012E+04, 6.3053E+04,
     +  6.3095E+04, 6.3137E+04, 6.3178E+04, 6.3220E+04, 6.3262E+04,
     + 6.3304E+04, 6.3346E+04, 6.3387E+04, 6.3429E+04, 6.3471E+04,
     +  6.3513E+04, 6.3555E+04, 6.3596E+04, 6.3638E+04, 6.3680E+04,
     + 6.3722E+04, 6.3764E+04, 6.3806E+04, 6.3848E+04, 6.3890E+04,
     +  6.3932E+04, 6.3974E+04, 6.4016E+04, 6.4058E+04, 6.4100E+04,
     + 6.4142E+04, 6.4184E+04, 6.4226E+04, 6.4268E+04, 6.4310E+04,
     +  6.4353E+04, 6.4395E+04, 6.4437E+04, 6.4479E+04, 6.4521E+04,
     + 6.4564E+04, 6.4606E+04, 6.4648E+04, 6.4690E+04, 6.4732E+04,
     +  6.4775E+04, 6.4817E+04, 6.4859E+04, 6.4902E+04, 6.4944E+04,
     + 6.4986E+04, 6.5029E+04, 6.5071E+04, 6.5114E+04, 6.5156E+04,
     +  6.5198E+04, 6.5241E+04, 6.5283E+04, 6.5326E+04, 6.5368E+04,
     + 6.5411E+04, 6.5453E+04, 6.5496E+04, 6.5538E+04, 6.5581E+04,
     +  6.5623E+04, 6.5666E+04, 6.5709E+04, 6.5751E+04, 6.5794E+04,
     + 6.5837E+04, 6.5879E+04, 6.5922E+04, 6.5965E+04, 6.6007E+04,
     +  6.6050E+04, 6.6093E+04, 6.6136E+04, 6.6178E+04, 6.6221E+04,
     + 6.6264E+04, 6.6307E+04, 6.6349E+04, 6.6392E+04, 6.6435E+04,
     +  6.6478E+04, 6.6521E+04, 6.6564E+04, 6.6607E+04, 6.6650E+04,
     + 6.6693E+04, 6.6736E+04, 6.6779E+04, 6.6822E+04, 6.6865E+04,
     +  6.6908E+04, 6.6951E+04, 6.6994E+04, 6.7037E+04, 6.7080E+04,
     + 6.7123E+04, 6.7166E+04, 6.7209E+04, 6.7252E+04, 6.7295E+04,
     +  6.7339E+04, 6.7382E+04, 6.7425E+04, 6.7468E+04, 6.7511E+04,
     + 6.7555E+04, 6.7598E+04, 6.7641E+04, 6.7684E+04, 6.7728E+04,
     +  6.7771E+04, 6.7814E+04, 6.7858E+04, 6.7901E+04, 6.7944E+04,
     + 6.7988E+04, 6.8031E+04, 6.8075E+04, 6.8118E+04, 6.8161E+04,
     +  6.8205E+04, 6.8248E+04, 6.8292E+04, 6.8335E+04, 6.8379E+04,
     + 6.8422E+04, 6.8466E+04, 6.8510E+04, 6.8553E+04, 6.8597E+04,
     +  6.8640E+04, 6.8684E+04, 6.8728E+04, 6.8771E+04, 6.8815E+04,
     + 6.8859E+04, 6.8902E+04, 6.8946E+04, 6.8990E+04, 6.9034E+04,
     +  6.9077E+04, 6.9121E+04, 6.9165E+04, 6.9209E+04, 6.9252E+04,
     + 6.9296E+04, 6.9340E+04, 6.9384E+04, 6.9428E+04, 6.9472E+04,
     +  6.9516E+04, 6.9560E+04, 6.9604E+04, 6.9647E+04, 6.9691E+04,
     + 6.9735E+04, 6.9779E+04, 6.9823E+04, 6.9867E+04, 6.9911E+04,
     +  6.9956E+04, 7.0000E+04, 7.0044E+04, 7.0088E+04, 7.0132E+04,
     + 7.0176E+04, 7.0220E+04, 7.0264E+04, 7.0308E+04, 7.0353E+04,
     +  7.0397E+04, 7.0441E+04, 7.0485E+04, 7.0530E+04, 7.0574E+04,
     + 7.0618E+04, 7.0662E+04, 7.0707E+04, 7.0751E+04, 7.0795E+04,
     +  7.0840E+04, 7.0884E+04, 7.0928E+04, 7.0973E+04, 7.1017E+04,
     + 7.1062E+04, 7.1106E+04, 7.1151E+04, 7.1195E+04, 7.1240E+04,
     +  7.1284E+04, 7.1329E+04, 7.1373E+04, 7.1418E+04, 7.1462E+04,
     + 7.1507E+04, 7.1551E+04, 7.1596E+04, 7.1641E+04, 7.1685E+04,
     +  7.1730E+04, 7.1774E+04, 7.1819E+04, 7.1864E+04, 7.1909E+04,
     + 7.1953E+04, 7.1998E+04, 7.2043E+04, 7.2088E+04, 7.2132E+04,
     +  7.2177E+04, 7.2222E+04, 7.2267E+04, 7.2312E+04, 7.2356E+04,
     + 7.2401E+04, 7.2446E+04, 7.2491E+04, 7.2536E+04, 7.2581E+04,
     +  7.2626E+04, 7.2671E+04, 7.2716E+04, 7.2761E+04, 7.2806E+04,
     + 7.2851E+04, 7.2896E+04, 7.2941E+04, 7.2986E+04, 7.3031E+04,
     +  7.3076E+04, 7.3121E+04, 7.3166E+04, 7.3212E+04, 7.3257E+04,
     + 7.3302E+04, 7.3347E+04, 7.3392E+04, 7.3437E+04, 7.3483E+04,
     +  7.3528E+04, 7.3573E+04, 7.3618E+04, 7.3664E+04, 7.3709E+04,
     + 7.3754E+04, 7.3800E+04, 7.3845E+04, 7.3890E+04, 7.3936E+04,
     +  7.3981E+04, 7.4027E+04, 7.4072E+04, 7.4117E+04, 7.4163E+04,
     + 7.4208E+04, 7.4254E+04, 7.4299E+04, 7.4345E+04, 7.4390E+04,
     +  7.4436E+04, 7.4481E+04, 7.4527E+04, 7.4573E+04, 7.4618E+04,
     + 7.4664E+04, 7.4709E+04, 7.4755E+04, 7.4801E+04, 7.4846E+04,
     +  7.4892E+04, 7.4938E+04, 7.4984E+04, 7.5029E+04, 7.5075E+04,
     + 7.5121E+04, 7.5167E+04, 7.5212E+04, 7.5258E+04, 7.5304E+04,
     +  7.5350E+04, 7.5396E+04, 7.5442E+04, 7.5487E+04, 7.5533E+04,
     + 7.5579E+04, 7.5625E+04, 7.5671E+04, 7.5717E+04, 7.5763E+04,
     +  7.5809E+04, 7.5855E+04, 7.5901E+04, 7.5947E+04, 7.5993E+04,
     + 7.6039E+04, 7.6085E+04, 7.6131E+04, 7.6177E+04, 7.6224E+04,
     +  7.6270E+04, 7.6316E+04, 7.6362E+04, 7.6408E+04, 7.6454E+04,
     + 7.6501E+04, 7.6547E+04, 7.6593E+04, 7.6639E+04, 7.6686E+04,
     +  7.6732E+04, 7.6778E+04, 7.6825E+04, 7.6871E+04, 7.6917E+04,
     + 7.6964E+04, 7.7010E+04, 7.7056E+04, 7.7103E+04, 7.7149E+04,
     +  7.7196E+04, 7.7242E+04, 7.7289E+04, 7.7335E+04, 7.7381E+04,
     + 7.7428E+04, 7.7475E+04, 7.7521E+04, 7.7568E+04, 7.7614E+04,
     +  7.7661E+04, 7.7707E+04, 7.7754E+04, 7.7801E+04, 7.7847E+04,
     + 7.7894E+04, 7.7941E+04, 7.7987E+04, 7.8034E+04, 7.8081E+04,
     +  7.8127E+04, 7.8174E+04, 7.8221E+04, 7.8268E+04, 7.8315E+04,
     + 7.8361E+04, 7.8408E+04, 7.8455E+04, 7.8502E+04, 7.8549E+04,
     +  7.8596E+04, 7.8643E+04, 7.8689E+04, 7.8736E+04, 7.8783E+04,
     + 7.8830E+04, 7.8877E+04, 7.8924E+04, 7.8971E+04, 7.9018E+04,
     +  7.9065E+04, 7.9112E+04, 7.9159E+04, 7.9206E+04, 7.9254E+04,
     + 7.9301E+04, 7.9348E+04, 7.9395E+04, 7.9442E+04, 7.9489E+04,
     +  7.9536E+04, 7.9584E+04, 7.9631E+04, 7.9678E+04, 7.9725E+04,
     + 7.9773E+04, 7.9820E+04, 7.9867E+04, 7.9914E+04, 7.9962E+04,
     +  8.0009E+04, 8.0056E+04, 8.0104E+04, 8.0151E+04, 8.0199E+04,
     + 8.0246E+04, 8.0293E+04, 8.0341E+04, 8.0388E+04, 8.0436E+04,
     +  8.0483E+04, 8.0531E+04, 8.0578E+04, 8.0626E+04, 8.0673E+04,
     + 8.0721E+04, 8.0769E+04, 8.0816E+04, 8.0864E+04, 8.0911E+04,
     +  8.0959E+04, 8.1007E+04, 8.1054E+04, 8.1102E+04, 8.1150E+04,
     + 8.1197E+04, 8.1245E+04, 8.1293E+04, 8.1341E+04, 8.1388E+04,
     +  8.1436E+04, 8.1484E+04, 8.1532E+04, 8.1580E+04, 8.1627E+04,
     + 8.1675E+04, 8.1723E+04, 8.1771E+04, 8.1819E+04, 8.1867E+04,
     +  8.1915E+04, 8.1963E+04, 8.2011E+04, 8.2059E+04, 8.2107E+04,
     + 8.2155E+04, 8.2203E+04, 8.2251E+04, 8.2299E+04, 8.2347E+04,
     +  8.2395E+04, 8.2443E+04, 8.2491E+04, 8.2539E+04, 8.2587E+04,
     + 8.2636E+04, 8.2684E+04, 8.2732E+04, 8.2780E+04, 8.2828E+04,
     +  8.2877E+04, 8.2925E+04, 8.2973E+04, 8.3021E+04, 8.3070E+04,
     + 8.3118E+04, 8.3166E+04, 8.3215E+04, 8.3263E+04, 8.3311E+04,
     +  8.3360E+04, 8.3408E+04, 8.3457E+04, 8.3505E+04, 8.3553E+04,
     + 8.3602E+04, 8.3650E+04, 8.3699E+04, 8.3747E+04, 8.3796E+04,
     +  8.3844E+04, 8.3893E+04, 8.3942E+04, 8.3990E+04, 8.4039E+04,
     + 8.4087E+04, 8.4136E+04, 8.4185E+04, 8.4233E+04, 8.4282E+04,
     +  8.4331E+04, 8.4379E+04, 8.4428E+04, 8.4477E+04, 8.4526E+04,
     + 8.4574E+04, 8.4623E+04, 8.4672E+04, 8.4721E+04, 8.4770E+04,
     +  8.4818E+04, 8.4867E+04, 8.4916E+04, 8.4965E+04, 8.5014E+04,
     + 8.5063E+04, 8.5112E+04, 8.5161E+04, 8.5210E+04, 8.5259E+04,
     +  8.5308E+04, 8.5357E+04, 8.5406E+04, 8.5455E+04, 8.5504E+04,
     + 8.5553E+04, 8.5602E+04, 8.5651E+04, 8.5700E+04, 8.5749E+04,
     +  8.5798E+04, 8.5847E+04, 8.5897E+04, 8.5946E+04, 8.5995E+04,
     + 8.6044E+04, 8.6093E+04, 8.6143E+04, 8.6192E+04, 8.6241E+04,
     +  8.6290E+04, 8.6340E+04, 8.6389E+04, 8.6438E+04, 8.6488E+04,
     + 8.6537E+04, 8.6586E+04, 8.6636E+04, 8.6685E+04, 8.6735E+04,
     +  8.6784E+04, 8.6834E+04, 8.6883E+04, 8.6933E+04, 8.6982E+04,
     + 8.7032E+04, 8.7081E+04, 8.7131E+04, 8.7180E+04, 8.7230E+04,
     +  8.7279E+04, 8.7329E+04, 8.7379E+04, 8.7428E+04, 8.7478E+04,
     + 8.7528E+04, 8.7577E+04, 8.7627E+04, 8.7677E+04, 8.7726E+04,
     +  8.7776E+04, 8.7826E+04, 8.7876E+04, 8.7925E+04, 8.7975E+04,
     + 8.8025E+04, 8.8075E+04, 8.8125E+04, 8.8175E+04, 8.8224E+04,
     +  8.8274E+04, 8.8324E+04, 8.8374E+04, 8.8424E+04, 8.8474E+04,
     + 8.8524E+04, 8.8574E+04, 8.8624E+04, 8.8674E+04, 8.8724E+04,
     +  8.8774E+04, 8.8824E+04, 8.8874E+04, 8.8924E+04, 8.8974E+04,
     + 8.9025E+04, 8.9075E+04, 8.9125E+04, 8.9175E+04, 8.9225E+04,
     +  8.9275E+04, 8.9326E+04, 8.9376E+04, 8.9426E+04, 8.9476E+04,
     + 8.9527E+04, 8.9577E+04, 8.9627E+04, 8.9677E+04, 8.9728E+04,
     +  8.9778E+04, 8.9828E+04, 8.9879E+04, 8.9929E+04, 8.9980E+04,
     + 9.0030E+04, 9.0081E+04, 9.0131E+04, 9.0181E+04, 9.0232E+04,
     +  9.0282E+04, 9.0333E+04, 9.0383E+04, 9.0434E+04, 9.0485E+04,
     + 9.0535E+04, 9.0586E+04, 9.0636E+04, 9.0687E+04, 9.0738E+04,
     +  9.0788E+04, 9.0839E+04, 9.0890E+04, 9.0940E+04, 9.0991E+04,
     + 9.1042E+04, 9.1092E+04, 9.1143E+04, 9.1194E+04, 9.1245E+04,
     +  9.1296E+04, 9.1346E+04, 9.1397E+04, 9.1448E+04, 9.1499E+04,
     + 9.1550E+04, 9.1601E+04, 9.1652E+04, 9.1702E+04, 9.1753E+04,
     +  9.1804E+04, 9.1855E+04, 9.1906E+04, 9.1957E+04, 9.2008E+04,
     + 9.2059E+04, 9.2110E+04, 9.2161E+04, 9.2212E+04, 9.2264E+04,
     +  9.2315E+04, 9.2366E+04, 9.2417E+04, 9.2468E+04, 9.2519E+04,
     + 9.2570E+04, 9.2622E+04, 9.2673E+04, 9.2724E+04, 9.2775E+04,
     +  9.2827E+04, 9.2878E+04, 9.2929E+04, 9.2980E+04, 9.3032E+04,
     + 9.3083E+04, 9.3134E+04, 9.3186E+04, 9.3237E+04, 9.3288E+04,
     +  9.3340E+04, 9.3391E+04, 9.3443E+04, 9.3494E+04, 9.3546E+04,
     + 9.3597E+04, 9.3649E+04, 9.3700E+04, 9.3752E+04, 9.3803E+04,
     +  9.3855E+04, 9.3906E+04, 9.3958E+04, 9.4010E+04, 9.4061E+04,
     + 9.4113E+04, 9.4164E+04, 9.4216E+04, 9.4268E+04, 9.4319E+04,
     +  9.4371E+04, 9.4423E+04, 9.4475E+04, 9.4526E+04, 9.4578E+04,
     + 9.4630E+04, 9.4682E+04, 9.4734E+04, 9.4785E+04, 9.4837E+04,
     +  9.4889E+04, 9.4941E+04, 9.4993E+04, 9.5045E+04, 9.5097E+04,
     + 9.5149E+04, 9.5201E+04, 9.5253E+04, 9.5305E+04, 9.5357E+04,
     +  9.5409E+04, 9.5461E+04, 9.5513E+04, 9.5565E+04, 9.5617E+04,
     + 9.5669E+04, 9.5721E+04, 9.5773E+04, 9.5825E+04, 9.5877E+04,
     +  9.5930E+04, 9.5982E+04, 9.6034E+04, 9.6086E+04, 9.6138E+04,
     + 9.6191E+04, 9.6243E+04, 9.6295E+04, 9.6347E+04, 9.6400E+04,
     +  9.6452E+04, 9.6504E+04, 9.6557E+04, 9.6609E+04, 9.6661E+04,
     + 9.6714E+04, 9.6766E+04, 9.6819E+04, 9.6871E+04, 9.6924E+04,
     +  9.6976E+04, 9.7028E+04, 9.7081E+04, 9.7133E+04, 9.7186E+04,
     + 9.7239E+04, 9.7291E+04, 9.7344E+04, 9.7396E+04, 9.7449E+04,
     +  9.7501E+04, 9.7554E+04, 9.7607E+04, 9.7659E+04, 9.7712E+04,
     + 9.7765E+04, 9.7818E+04, 9.7870E+04, 9.7923E+04, 9.7976E+04,
     +  9.8029E+04, 9.8081E+04, 9.8134E+04, 9.8187E+04, 9.8240E+04,
     + 9.8293E+04, 9.8345E+04, 9.8398E+04, 9.8451E+04, 9.8504E+04,
     +  9.8557E+04, 9.8610E+04, 9.8663E+04, 9.8716E+04, 9.8769E+04,
     + 9.8822E+04, 9.8875E+04, 9.8928E+04, 9.8981E+04, 9.9034E+04,
     +  9.9087E+04, 9.9140E+04, 9.9193E+04, 9.9246E+04, 9.9300E+04,
     + 9.9353E+04, 9.9406E+04, 9.9459E+04, 9.9512E+04, 9.9566E+04,
     +  9.9619E+04, 9.9672E+04, 9.9725E+04, 9.9779E+04, 9.9832E+04,
     + 9.9885E+04, 9.9938E+04, 9.9992E+04, 1.0005E+05, 1.0010E+05,
     +  1.0015E+05, 1.0021E+05, 1.0026E+05, 1.0031E+05, 1.0037E+05,
     + 1.0042E+05, 1.0047E+05, 1.0053E+05, 1.0058E+05, 1.0063E+05,
     +  1.0069E+05, 1.0074E+05, 1.0079E+05, 1.0085E+05, 1.0090E+05,
     + 1.0095E+05, 1.0101E+05, 1.0106E+05, 1.0112E+05, 1.0117E+05,
     +  1.0122E+05, 1.0128E+05, 1.0133E+05, 1.0138E+05, 1.0144E+05,
     + 1.0149E+05, 1.0155E+05, 1.0160E+05, 1.0165E+05, 1.0171E+05,
     +  1.0176E+05, 1.0181E+05, 1.0187E+05, 1.0192E+05, 1.0198E+05,
     + 1.0203E+05, 1.0208E+05, 1.0214E+05, 1.0219E+05, 1.0225E+05,
     +  1.0230E+05, 1.0235E+05, 1.0241E+05, 1.0246E+05, 1.0252E+05,
     + 1.0257E+05, 1.0262E+05, 1.0268E+05, 1.0273E+05, 1.0279E+05,
     +  1.0284E+05, 1.0289E+05, 1.0295E+05, 1.0300E+05, 1.0306E+05,
     + 1.0311E+05, 1.0317E+05, 1.0322E+05, 1.0327E+05, 1.0333E+05,
     +  1.0338E+05, 1.0344E+05, 1.0349E+05, 1.0355E+05, 1.0360E+05,
     + 1.0365E+05, 1.0371E+05, 1.0376E+05, 1.0382E+05, 1.0387E+05,
     +  1.0393E+05, 1.0398E+05, 1.0404E+05, 1.0409E+05, 1.0414E+05,
     + 1.0420E+05, 1.0425E+05, 1.0431E+05, 1.0436E+05, 1.0442E+05,
     +  1.0447E+05, 1.0453E+05, 1.0458E+05, 1.0464E+05, 1.0469E+05,
     + 1.0474E+05, 1.0480E+05, 1.0485E+05, 1.0491E+05, 1.0496E+05,
     +  1.0502E+05, 1.0507E+05, 1.0513E+05, 1.0518E+05, 1.0524E+05,
     + 1.0529E+05, 1.0535E+05, 1.0540E+05, 1.0546E+05, 1.0551E+05,
     +  1.0557E+05, 1.0562E+05, 1.0568E+05, 1.0573E+05, 1.0579E+05,
     + 1.0584E+05, 1.0590E+05, 1.0595E+05, 1.0601E+05, 1.0606E+05,
     +  1.0612E+05, 1.0617E+05, 1.0623E+05, 1.0628E+05, 1.0634E+05,
     + 1.0639E+05, 1.0645E+05, 1.0650E+05, 1.0656E+05, 1.0661E+05,
     +  1.0667E+05, 1.0672E+05, 1.0678E+05, 1.0683E+05, 1.0689E+05,
     + 1.0694E+05, 1.0700E+05, 1.0706E+05, 1.0711E+05, 1.0717E+05,
     +  1.0722E+05, 1.0728E+05, 1.0733E+05, 1.0739E+05, 1.0744E+05,
     + 1.0750E+05, 1.0755E+05, 1.0761E+05, 1.0766E+05, 1.0772E+05,
     +  1.0778E+05, 1.0783E+05, 1.0789E+05, 1.0794E+05, 1.0800E+05,
     + 1.0805E+05, 1.0811E+05, 1.0816E+05, 1.0822E+05, 1.0828E+05,
     +  1.0833E+05, 1.0839E+05, 1.0844E+05, 1.0850E+05, 1.0855E+05,
     + 1.0861E+05, 1.0867E+05, 1.0872E+05, 1.0878E+05, 1.0883E+05,
     +  1.0889E+05, 1.0895E+05, 1.0900E+05, 1.0906E+05, 1.0911E+05,
     + 1.0917E+05, 1.0922E+05, 1.0928E+05, 1.0934E+05, 1.0939E+05,
     +  1.0945E+05, 1.0950E+05, 1.0956E+05, 1.0962E+05, 1.0967E+05,
     + 1.0973E+05, 1.0978E+05, 1.0984E+05, 1.0990E+05, 1.0995E+05,
     +  1.1001E+05, 1.1007E+05, 1.1012E+05, 1.1018E+05, 1.1023E+05,
     + 1.1029E+05, 1.1035E+05, 1.1040E+05, 1.1046E+05, 1.1051E+05,
     +  1.1057E+05, 1.1063E+05, 1.1068E+05, 1.1074E+05, 1.1080E+05,
     + 1.1085E+05, 1.1091E+05, 1.1097E+05, 1.1102E+05, 1.1108E+05,
     +  1.1113E+05, 1.1119E+05, 1.1125E+05, 1.1130E+05, 1.1136E+05,
     + 1.1142E+05, 1.1147E+05, 1.1153E+05, 1.1159E+05, 1.1164E+05,
     +  1.1170E+05, 1.1176E+05, 1.1181E+05, 1.1187E+05, 1.1193E+05,
     + 1.1198E+05, 1.1204E+05, 1.1210E+05, 1.1215E+05, 1.1221E+05,
     +  1.1227E+05, 1.1232E+05, 1.1238E+05, 1.1244E+05, 1.1249E+05,
     + 1.1255E+05, 1.1261E+05, 1.1266E+05, 1.1272E+05, 1.1278E+05,
     +  1.1284E+05, 1.1289E+05, 1.1295E+05, 1.1301E+05, 1.1306E+05,
     + 1.1312E+05, 1.1318E+05, 1.1323E+05, 1.1329E+05, 1.1335E+05,
     +  1.1340E+05, 1.1346E+05, 1.1352E+05, 1.1358E+05, 1.1363E+05,
     + 1.1369E+05, 1.1375E+05, 1.1380E+05, 1.1386E+05, 1.1392E+05,
     +  1.1398E+05, 1.1403E+05, 1.1409E+05, 1.1415E+05, 1.1421E+05,
     + 1.1426E+05, 1.1432E+05, 1.1438E+05, 1.1443E+05, 1.1449E+05,
     +  1.1455E+05, 1.1461E+05, 1.1466E+05, 1.1472E+05, 1.1478E+05,
     + 1.1484E+05, 1.1489E+05, 1.1495E+05, 1.1501E+05, 1.1507E+05,
     +  1.1512E+05, 1.1518E+05, 1.1524E+05, 1.1530E+05, 1.1535E+05,
     + 1.1541E+05, 1.1547E+05, 1.1553E+05, 1.1558E+05, 1.1564E+05,
     +  1.1570E+05, 1.1576E+05, 1.1581E+05, 1.1587E+05, 1.1593E+05,
     + 1.1599E+05, 1.1605E+05, 1.1610E+05, 1.1616E+05, 1.1622E+05,
     +  1.1628E+05, 1.1633E+05, 1.1639E+05, 1.1645E+05, 1.1651E+05,
     + 1.1657E+05, 1.1662E+05, 1.1668E+05, 1.1674E+05, 1.1680E+05,
     +  1.1686E+05, 1.1691E+05, 1.1697E+05, 1.1703E+05, 1.1709E+05,
     + 1.1715E+05, 1.1720E+05, 1.1726E+05, 1.1732E+05, 1.1738E+05,
     +  1.1744E+05, 1.1749E+05, 1.1755E+05, 1.1761E+05, 1.1767E+05,
     + 1.1773E+05, 1.1779E+05, 1.1784E+05, 1.1790E+05, 1.1796E+05,
     +  1.1802E+05, 1.1808E+05, 1.1814E+05, 1.1819E+05, 1.1825E+05,
     + 1.1831E+05, 1.1837E+05, 1.1843E+05, 1.1849E+05, 1.1854E+05,
     +  1.1860E+05, 1.1866E+05, 1.1872E+05, 1.1878E+05, 1.1884E+05,
     + 1.1889E+05, 1.1895E+05, 1.1901E+05, 1.1907E+05, 1.1913E+05,
     +  1.1919E+05, 1.1925E+05, 1.1930E+05, 1.1936E+05, 1.1942E+05,
     + 1.1948E+05, 1.1954E+05, 1.1960E+05, 1.1966E+05, 1.1972E+05,
     +  1.1977E+05, 1.1983E+05, 1.1989E+05, 1.1995E+05, 1.2001E+05,
     + 1.2007E+05, 1.2013E+05, 1.2019E+05, 1.2024E+05, 1.2030E+05,
     +  1.2036E+05, 1.2042E+05, 1.2048E+05, 1.2054E+05, 1.2060E+05,
     + 1.2066E+05, 1.2072E+05, 1.2077E+05, 1.2083E+05, 1.2089E+05,
     +  1.2095E+05, 1.2101E+05, 1.2107E+05, 1.2113E+05, 1.2119E+05,
     + 1.2125E+05, 1.2131E+05, 1.2137E+05, 1.2142E+05, 1.2148E+05,
     +  1.2154E+05, 1.2160E+05, 1.2166E+05, 1.2172E+05, 1.2178E+05,
     + 1.2184E+05, 1.2190E+05, 1.2196E+05, 1.2202E+05, 1.2208E+05,
     +  1.2214E+05, 1.2220E+05, 1.2225E+05, 1.2231E+05, 1.2237E+05,
     + 1.2243E+05, 1.2249E+05, 1.2255E+05, 1.2261E+05, 1.2267E+05,
     +  1.2273E+05, 1.2279E+05, 1.2285E+05, 1.2291E+05, 1.2297E+05,
     + 1.2303E+05, 1.2309E+05, 1.2315E+05, 1.2321E+05, 1.2327E+05,
     +  1.2333E+05, 1.2339E+05, 1.2345E+05, 1.2351E+05, 1.2357E+05,
     + 1.2362E+05, 1.2368E+05, 1.2374E+05, 1.2380E+05, 1.2386E+05,
     +  1.2392E+05, 1.2398E+05, 1.2404E+05, 1.2410E+05, 1.2416E+05,
     + 1.2422E+05, 1.2428E+05, 1.2434E+05, 1.2440E+05, 1.2446E+05,
     +  1.2452E+05, 1.2458E+05, 1.2464E+05, 1.2470E+05, 1.2476E+05,
     + 1.2482E+05, 1.2488E+05, 1.2494E+05, 1.2500E+05, 1.2506E+05,
     +  1.2512E+05, 1.2518E+05, 1.2524E+05, 1.2530E+05, 1.2536E+05,
     + 1.2542E+05, 1.2548E+05, 1.2554E+05, 1.2560E+05, 1.2567E+05,
     +  1.2573E+05, 1.2579E+05, 1.2585E+05, 1.2591E+05, 1.2597E+05,
     + 1.2603E+05, 1.2609E+05, 1.2615E+05, 1.2621E+05, 1.2627E+05,
     +  1.2633E+05, 1.2639E+05, 1.2645E+05, 1.2651E+05, 1.2657E+05,
     + 1.2663E+05, 1.2669E+05, 1.2675E+05, 1.2681E+05, 1.2687E+05,
     +  1.2693E+05, 1.2699E+05, 1.2706E+05, 1.2712E+05, 1.2718E+05,
     + 1.2724E+05, 1.2730E+05, 1.2736E+05, 1.2742E+05, 1.2748E+05,
     +  1.2754E+05, 1.2760E+05, 1.2766E+05, 1.2772E+05, 1.2778E+05,
     + 1.2784E+05, 1.2791E+05, 1.2797E+05, 1.2803E+05, 1.2809E+05,
     +  1.2815E+05, 1.2821E+05, 1.2827E+05, 1.2833E+05, 1.2839E+05,
     + 1.2845E+05, 1.2851E+05, 1.2858E+05, 1.2864E+05, 1.2870E+05,
     +  1.2876E+05, 1.2882E+05, 1.2888E+05, 1.2894E+05, 1.2900E+05,
     + 1.2906E+05, 1.2913E+05, 1.2919E+05, 1.2925E+05, 1.2931E+05,
     +  1.2937E+05, 1.2943E+05, 1.2949E+05, 1.2955E+05, 1.2962E+05,
     + 1.2968E+05, 1.2974E+05, 1.2980E+05, 1.2986E+05, 1.2992E+05,
     +  1.2998E+05, 1.3004E+05, 1.3011E+05, 1.3017E+05, 1.3023E+05,
     + 1.3029E+05, 1.3035E+05, 1.3041E+05, 1.3047E+05, 1.3054E+05,
     +  1.3060E+05, 1.3066E+05, 1.3072E+05, 1.3078E+05, 1.3084E+05,
     + 1.3091E+05, 1.3097E+05, 1.3103E+05, 1.3109E+05, 1.3115E+05,
     +  1.3121E+05, 1.3128E+05, 1.3134E+05, 1.3140E+05, 1.3146E+05,
     + 1.3152E+05, 1.3158E+05, 1.3165E+05, 1.3171E+05, 1.3177E+05,
     +  1.3183E+05, 1.3189E+05, 1.3195E+05, 1.3202E+05, 1.3208E+05,
     + 1.3214E+05, 1.3220E+05, 1.3226E+05, 1.3233E+05, 1.3239E+05,
     +  1.3245E+05, 1.3251E+05, 1.3257E+05, 1.3264E+05, 1.3270E+05,
     + 1.3276E+05, 1.3282E+05, 1.3288E+05, 1.3295E+05, 1.3301E+05,
     +  1.3307E+05, 1.3313E+05, 1.3319E+05, 1.3326E+05, 1.3332E+05,
     + 1.3338E+05, 1.3344E+05, 1.3351E+05, 1.3357E+05, 1.3363E+05,
     +  1.3369E+05, 1.3376E+05, 1.3382E+05, 1.3388E+05, 1.3394E+05,
     + 1.3400E+05, 1.3407E+05, 1.3413E+05, 1.3419E+05, 1.3425E+05,
     +  1.3432E+05, 1.3438E+05, 1.3444E+05, 1.3450E+05, 1.3457E+05,
     + 1.3463E+05, 1.3469E+05, 1.3475E+05, 1.3482E+05, 1.3488E+05,
     +  1.3494E+05, 1.3500E+05, 1.3507E+05, 1.3513E+05, 1.3519E+05,
     + 1.3525E+05, 1.3532E+05, 1.3538E+05, 1.3544E+05, 1.3551E+05,
     +  1.3557E+05, 1.3563E+05, 1.3569E+05, 1.3576E+05, 1.3582E+05,
     + 1.3588E+05, 1.3595E+05, 1.3601E+05, 1.3607E+05, 1.3613E+05,
     +  1.3620E+05, 1.3626E+05, 1.3632E+05, 1.3639E+05, 1.3645E+05,
     + 1.3651E+05, 1.3657E+05, 1.3664E+05, 1.3670E+05, 1.3676E+05,
     +  1.3683E+05, 1.3689E+05, 1.3695E+05, 1.3702E+05, 1.3708E+05,
     + 1.3714E+05, 1.3721E+05, 1.3727E+05, 1.3733E+05, 1.3739E+05,
     +  1.3746E+05, 1.3752E+05, 1.3758E+05, 1.3765E+05, 1.3771E+05,
     + 1.3777E+05, 1.3784E+05, 1.3790E+05, 1.3796E+05, 1.3803E+05,
     +  1.3809E+05, 1.3815E+05, 1.3822E+05, 1.3828E+05, 1.3834E+05,
     + 1.3841E+05, 1.3847E+05, 1.3853E+05, 1.3860E+05, 1.3866E+05,
     +  1.3873E+05, 1.3879E+05, 1.3885E+05, 1.3892E+05, 1.3898E+05,
     + 1.3904E+05, 1.3911E+05, 1.3917E+05, 1.3923E+05, 1.3930E+05,
     +  1.3936E+05, 1.3942E+05, 1.3949E+05, 1.3955E+05, 1.3962E+05,
     + 1.3968E+05, 1.3974E+05, 1.3981E+05, 1.3987E+05, 1.3993E+05,
     +  1.4000E+05, 1.4006E+05, 1.4013E+05, 1.4019E+05, 1.4025E+05,
     + 1.4032E+05, 1.4038E+05, 1.4045E+05, 1.4051E+05, 1.4057E+05,
     +  1.4064E+05, 1.4070E+05, 1.4077E+05, 1.4083E+05, 1.4089E+05,
     + 1.4096E+05, 1.4102E+05, 1.4109E+05, 1.4115E+05, 1.4121E+05,
     +  1.4128E+05, 1.4134E+05, 1.4141E+05, 1.4147E+05, 1.4154E+05,
     + 1.4160E+05, 1.4166E+05, 1.4173E+05, 1.4179E+05, 1.4186E+05,
     +  1.4192E+05, 1.4198E+05, 1.4205E+05, 1.4211E+05, 1.4218E+05,
     + 1.4224E+05, 1.4231E+05, 1.4237E+05, 1.4244E+05, 1.4250E+05,
     +  1.4256E+05, 1.4263E+05, 1.4269E+05, 1.4276E+05, 1.4282E+05,
     + 1.4289E+05, 1.4295E+05, 1.4302E+05, 1.4308E+05, 1.4315E+05,
     +  1.4321E+05, 1.4327E+05, 1.4334E+05, 1.4340E+05, 1.4347E+05,
     + 1.4353E+05, 1.4360E+05, 1.4366E+05, 1.4373E+05, 1.4379E+05,
     +  1.4386E+05, 1.4392E+05, 1.4399E+05, 1.4405E+05, 1.4412E+05,
     + 1.4418E+05, 1.4425E+05, 1.4431E+05, 1.4437E+05, 1.4444E+05,
     +  1.4450E+05, 1.4457E+05, 1.4463E+05, 1.4470E+05, 1.4476E+05,
     + 1.4483E+05, 1.4489E+05, 1.4496E+05, 1.4502E+05, 1.4509E+05,
     +  1.4515E+05, 1.4522E+05, 1.4528E+05, 1.4535E+05, 1.4542E+05,
     + 1.4548E+05, 1.4555E+05, 1.4561E+05, 1.4568E+05, 1.4574E+05,
     +  1.4581E+05, 1.4587E+05, 1.4594E+05, 1.4600E+05, 1.4607E+05,
     + 1.4613E+05, 1.4620E+05, 1.4626E+05, 1.4633E+05, 1.4639E+05,
     +  1.4646E+05, 1.4652E+05, 1.4659E+05, 1.4666E+05, 1.4672E+05,
     + 1.4679E+05, 1.4685E+05, 1.4692E+05, 1.4698E+05, 1.4705E+05,
     +  1.4711E+05, 1.4718E+05, 1.4724E+05, 1.4731E+05, 1.4738E+05,
     + 1.4744E+05, 1.4751E+05, 1.4757E+05, 1.4764E+05, 1.4770E+05,
     +  1.4777E+05, 1.4784E+05, 1.4790E+05, 1.4797E+05, 1.4803E+05,
     + 1.4810E+05, 1.4816E+05, 1.4823E+05, 1.4830E+05, 1.4836E+05,
     +  1.4843E+05, 1.4849E+05, 1.4856E+05, 1.4862E+05, 1.4869E+05,
     + 1.4876E+05, 1.4882E+05, 1.4889E+05, 1.4895E+05, 1.4902E+05,
     +  1.4909E+05, 1.4915E+05, 1.4922E+05, 1.4928E+05, 1.4935E+05,
     + 1.4942E+05, 1.4948E+05, 1.4955E+05, 1.4961E+05, 1.4968E+05,
     +  1.4975E+05, 1.4981E+05, 1.4988E+05, 1.4995E+05, 1.5001E+05,
     + 1.5008E+05, 1.5014E+05, 1.5021E+05, 1.5028E+05, 1.5034E+05,
     +  1.5041E+05, 1.5048E+05, 1.5054E+05, 1.5061E+05, 1.5067E+05,
     + 1.5074E+05, 1.5081E+05, 1.5087E+05, 1.5094E+05, 1.5101E+05,
     +  1.5107E+05, 1.5114E+05, 1.5121E+05, 1.5127E+05, 1.5134E+05,
     + 1.5141E+05, 1.5147E+05, 1.5154E+05, 1.5161E+05, 1.5167E+05,
     +  1.5174E+05, 1.5181E+05, 1.5187E+05, 1.5194E+05, 1.5201E+05,
     + 1.5207E+05, 1.5214E+05, 1.5221E+05, 1.5227E+05, 1.5234E+05,
     +  1.5241E+05, 1.5247E+05, 1.5254E+05, 1.5261E+05, 1.5267E+05,
     + 1.5274E+05, 1.5281E+05, 1.5287E+05, 1.5294E+05, 1.5301E+05,
     +  1.5307E+05, 1.5314E+05, 1.5321E+05, 1.5327E+05, 1.5334E+05,
     + 1.5341E+05, 1.5348E+05, 1.5354E+05, 1.5361E+05, 1.5368E+05,
     +  1.5374E+05, 1.5381E+05, 1.5388E+05, 1.5395E+05, 1.5401E+05,
     + 1.5408E+05, 1.5415E+05, 1.5421E+05, 1.5428E+05, 1.5435E+05,
     +  1.5442E+05, 1.5448E+05, 1.5455E+05, 1.5462E+05, 1.5468E+05,
     + 1.5475E+05, 1.5482E+05, 1.5489E+05, 1.5495E+05, 1.5502E+05,
     +  1.5509E+05, 1.5516E+05, 1.5522E+05, 1.5529E+05, 1.5536E+05,
     + 1.5543E+05, 1.5549E+05, 1.5556E+05, 1.5563E+05, 1.5570E+05,
     +  1.5576E+05, 1.5583E+05, 1.5590E+05, 1.5597E+05, 1.5603E+05,
     + 1.5610E+05, 1.5617E+05, 1.5624E+05, 1.5630E+05, 1.5637E+05,
     +  1.5644E+05, 1.5651E+05, 1.5657E+05, 1.5664E+05, 1.5671E+05,
     + 1.5678E+05, 1.5685E+05, 1.5691E+05, 1.5698E+05, 1.5705E+05,
     +  1.5712E+05, 1.5718E+05, 1.5725E+05, 1.5732E+05, 1.5739E+05,
     + 1.5746E+05, 1.5752E+05, 1.5759E+05, 1.5766E+05, 1.5773E+05,
     +  1.5780E+05, 1.5786E+05, 1.5793E+05, 1.5800E+05, 1.5807E+05,
     + 1.5814E+05, 1.5820E+05, 1.5827E+05, 1.5834E+05, 1.5841E+05,
     +  1.5848E+05, 1.5855E+05, 1.5861E+05, 1.5868E+05, 1.5875E+05,
     + 1.5882E+05, 1.5889E+05, 1.5895E+05, 1.5902E+05, 1.5909E+05,
     +  1.5916E+05, 1.5923E+05, 1.5930E+05, 1.5936E+05, 1.5943E+05,
     + 1.5950E+05, 1.5957E+05, 1.5964E+05, 1.5971E+05, 1.5977E+05,
     +  1.5984E+05, 1.5991E+05, 1.5998E+05, 1.6005E+05, 1.6012E+05,
     + 1.6019E+05, 1.6025E+05, 1.6032E+05, 1.6039E+05/
c...        --        46
      data (QofT( 7,J),J=1,2408)/ 1.0186E+00, 1.2389E+00, 1.5829E+00,
     +  1.9564E+00, 2.3391E+00, 2.7259E+00, 3.1147E+00, 3.5049E+00,
     + 3.8959E+00, 4.2874E+00, 4.6793E+00, 5.0716E+00, 5.4640E+00,
     +  5.8567E+00, 6.2495E+00, 6.6424E+00, 7.0353E+00, 7.4284E+00,
     + 7.8215E+00, 8.2147E+00, 9.0012E+00, 9.7879E+00, 1.0575E+01,
     +  1.1361E+01, 1.2148E+01, 1.2935E+01, 1.3722E+01, 1.4510E+01,
     + 1.5297E+01, 1.6084E+01, 1.6871E+01, 1.7658E+01, 1.8445E+01,
     +  1.9233E+01, 2.0020E+01, 2.0807E+01, 2.1594E+01, 2.2382E+01,
     + 2.3169E+01, 2.3956E+01, 2.4744E+01, 2.5531E+01, 2.6319E+01,
     +  2.7106E+01, 2.7893E+01, 2.8681E+01, 2.9468E+01, 3.0256E+01,
     + 3.1043E+01, 3.1831E+01, 3.2618E+01, 3.3406E+01, 3.4193E+01,
     +  3.4981E+01, 3.5768E+01, 3.6556E+01, 3.7343E+01, 3.8131E+01,
     + 3.8918E+01, 3.9706E+01, 4.0493E+01, 4.1281E+01, 4.2069E+01,
     +  4.2856E+01, 4.3644E+01, 4.4431E+01, 4.5219E+01, 4.6007E+01,
     + 4.6794E+01, 4.7582E+01, 4.8370E+01, 4.9157E+01, 4.9945E+01,
     +  5.0733E+01, 5.1520E+01, 5.2308E+01, 5.3096E+01, 5.3883E+01,
     + 5.4671E+01, 5.5459E+01, 5.6247E+01, 5.7034E+01, 5.7822E+01,
     +  5.8610E+01, 5.9398E+01, 6.0185E+01, 6.0973E+01, 6.1761E+01,
     + 6.2549E+01, 6.3336E+01, 6.4124E+01, 6.4912E+01, 6.5700E+01,
     +  6.6488E+01, 6.7276E+01, 6.8063E+01, 6.8851E+01, 6.9639E+01,
     + 7.0427E+01, 7.1215E+01, 7.2003E+01, 7.2791E+01, 7.3578E+01,
     +  7.4366E+01, 7.5154E+01, 7.5942E+01, 7.6730E+01, 7.7518E+01,
     + 7.8306E+01, 7.9094E+01, 7.9882E+01, 8.0670E+01, 8.1458E+01,
     +  8.2246E+01, 8.3034E+01, 8.3822E+01, 8.4610E+01, 8.5398E+01,
     + 8.6186E+01, 8.6974E+01, 8.7762E+01, 8.8550E+01, 8.9338E+01,
     +  9.0126E+01, 9.0914E+01, 9.1702E+01, 9.2491E+01, 9.3279E+01,
     + 9.4067E+01, 9.4855E+01, 9.5643E+01, 9.6431E+01, 9.7219E+01,
     +  9.8008E+01, 9.8796E+01, 9.9584E+01, 1.0037E+02, 1.0116E+02,
     + 1.0195E+02, 1.0274E+02, 1.0353E+02, 1.0431E+02, 1.0510E+02,
     +  1.0589E+02, 1.0668E+02, 1.0747E+02, 1.0826E+02, 1.0904E+02,
     + 1.0983E+02, 1.1062E+02, 1.1141E+02, 1.1220E+02, 1.1299E+02,
     +  1.1378E+02, 1.1456E+02, 1.1535E+02, 1.1614E+02, 1.1693E+02,
     + 1.1772E+02, 1.1851E+02, 1.1930E+02, 1.2008E+02, 1.2087E+02,
     +  1.2166E+02, 1.2245E+02, 1.2324E+02, 1.2403E+02, 1.2482E+02,
     + 1.2561E+02, 1.2640E+02, 1.2719E+02, 1.2798E+02, 1.2876E+02,
     +  1.2955E+02, 1.3034E+02, 1.3113E+02, 1.3192E+02, 1.3271E+02,
     + 1.3350E+02, 1.3429E+02, 1.3508E+02, 1.3587E+02, 1.3666E+02,
     +  1.3745E+02, 1.3824E+02, 1.3903E+02, 1.3982E+02, 1.4061E+02,
     + 1.4140E+02, 1.4219E+02, 1.4298E+02, 1.4377E+02, 1.4456E+02,
     +  1.4536E+02, 1.4615E+02, 1.4694E+02, 1.4773E+02, 1.4852E+02,
     + 1.4931E+02, 1.5010E+02, 1.5089E+02, 1.5169E+02, 1.5248E+02,
     +  1.5327E+02, 1.5406E+02, 1.5485E+02, 1.5565E+02, 1.5644E+02,
     + 1.5723E+02, 1.5802E+02, 1.5882E+02, 1.5961E+02, 1.6040E+02,
     +  1.6120E+02, 1.6199E+02, 1.6278E+02, 1.6358E+02, 1.6437E+02,
     + 1.6517E+02, 1.6596E+02, 1.6675E+02, 1.6755E+02, 1.6834E+02,
     +  1.6914E+02, 1.6993E+02, 1.7073E+02, 1.7153E+02, 1.7232E+02,
     + 1.7312E+02, 1.7391E+02, 1.7471E+02, 1.7551E+02, 1.7630E+02,
     +  1.7710E+02, 1.7790E+02, 1.7869E+02, 1.7949E+02, 1.8029E+02,
     + 1.8109E+02, 1.8189E+02, 1.8269E+02, 1.8348E+02, 1.8428E+02,
     +  1.8508E+02, 1.8588E+02, 1.8668E+02, 1.8748E+02, 1.8828E+02,
     + 1.8908E+02, 1.8988E+02, 1.9069E+02, 1.9149E+02, 1.9229E+02,
     +  1.9309E+02, 1.9389E+02, 1.9470E+02, 1.9550E+02, 1.9630E+02,
     + 1.9711E+02, 1.9791E+02, 1.9871E+02, 1.9952E+02, 2.0032E+02,
     +  2.0113E+02, 2.0193E+02, 2.0274E+02, 2.0355E+02, 2.0435E+02,
     + 2.0516E+02, 2.0597E+02, 2.0677E+02, 2.0758E+02, 2.0839E+02,
     +  2.0920E+02, 2.1001E+02, 2.1082E+02, 2.1163E+02, 2.1244E+02,
     + 2.1325E+02, 2.1406E+02, 2.1487E+02, 2.1568E+02, 2.1649E+02,
     +  2.1730E+02, 2.1812E+02, 2.1893E+02, 2.1974E+02, 2.2056E+02,
     + 2.2137E+02, 2.2219E+02, 2.2300E+02, 2.2382E+02, 2.2463E+02,
     +  2.2545E+02, 2.2627E+02, 2.2708E+02, 2.2790E+02, 2.2872E+02,
     + 2.2954E+02, 2.3036E+02, 2.3118E+02, 2.3200E+02, 2.3282E+02,
     +  2.3364E+02, 2.3446E+02, 2.3528E+02, 2.3611E+02, 2.3693E+02,
     + 2.3775E+02, 2.3858E+02, 2.3940E+02, 2.4022E+02, 2.4105E+02,
     +  2.4188E+02, 2.4270E+02, 2.4353E+02, 2.4436E+02, 2.4518E+02,
     + 2.4601E+02, 2.4684E+02, 2.4767E+02, 2.4850E+02, 2.4933E+02,
     +  2.5016E+02, 2.5099E+02, 2.5183E+02, 2.5266E+02, 2.5349E+02,
     + 2.5433E+02, 2.5516E+02, 2.5599E+02, 2.5683E+02, 2.5767E+02,
     +  2.5850E+02, 2.5934E+02, 2.6018E+02, 2.6102E+02, 2.6185E+02,
     + 2.6269E+02, 2.6353E+02, 2.6437E+02, 2.6522E+02, 2.6606E+02,
     +  2.6690E+02, 2.6774E+02, 2.6859E+02, 2.6943E+02, 2.7027E+02,
     + 2.7112E+02, 2.7197E+02, 2.7281E+02, 2.7366E+02, 2.7451E+02,
     +  2.7536E+02, 2.7620E+02, 2.7705E+02, 2.7790E+02, 2.7875E+02,
     + 2.7961E+02, 2.8046E+02, 2.8131E+02, 2.8216E+02, 2.8302E+02,
     +  2.8387E+02, 2.8473E+02, 2.8559E+02, 2.8644E+02, 2.8730E+02,
     + 2.8816E+02, 2.8902E+02, 2.8988E+02, 2.9074E+02, 2.9160E+02,
     +  2.9246E+02, 2.9332E+02, 2.9418E+02, 2.9505E+02, 2.9591E+02,
     + 2.9677E+02, 2.9764E+02, 2.9851E+02, 2.9937E+02, 3.0024E+02,
     +  3.0111E+02, 3.0198E+02, 3.0285E+02, 3.0372E+02, 3.0459E+02,
     + 3.0546E+02, 3.0633E+02, 3.0721E+02, 3.0808E+02, 3.0895E+02,
     +  3.0983E+02, 3.1071E+02, 3.1158E+02, 3.1246E+02, 3.1334E+02,
     + 3.1422E+02, 3.1510E+02, 3.1598E+02, 3.1686E+02, 3.1774E+02,
     +  3.1862E+02, 3.1951E+02, 3.2039E+02, 3.2128E+02, 3.2216E+02,
     + 3.2305E+02, 3.2393E+02, 3.2482E+02, 3.2571E+02, 3.2660E+02,
     +  3.2749E+02, 3.2838E+02, 3.2927E+02, 3.3017E+02, 3.3106E+02,
     + 3.3195E+02, 3.3285E+02, 3.3374E+02, 3.3464E+02, 3.3554E+02,
     +  3.3643E+02, 3.3733E+02, 3.3823E+02, 3.3913E+02, 3.4003E+02,
     + 3.4094E+02, 3.4184E+02, 3.4274E+02, 3.4365E+02, 3.4455E+02,
     +  3.4546E+02, 3.4636E+02, 3.4727E+02, 3.4818E+02, 3.4909E+02,
     + 3.5000E+02, 3.5091E+02, 3.5182E+02, 3.5273E+02, 3.5365E+02,
     +  3.5456E+02, 3.5548E+02, 3.5639E+02, 3.5731E+02, 3.5823E+02,
     + 3.5914E+02, 3.6006E+02, 3.6098E+02, 3.6190E+02, 3.6282E+02,
     +  3.6375E+02, 3.6467E+02, 3.6559E+02, 3.6652E+02, 3.6744E+02,
     + 3.6837E+02, 3.6930E+02, 3.7023E+02, 3.7115E+02, 3.7208E+02,
     +  3.7302E+02, 3.7395E+02, 3.7488E+02, 3.7581E+02, 3.7675E+02,
     + 3.7768E+02, 3.7862E+02, 3.7955E+02, 3.8049E+02, 3.8143E+02,
     +  3.8237E+02, 3.8331E+02, 3.8425E+02, 3.8519E+02, 3.8613E+02,
     + 3.8708E+02, 3.8802E+02, 3.8897E+02, 3.8991E+02, 3.9086E+02,
     +  3.9181E+02, 3.9276E+02, 3.9371E+02, 3.9466E+02, 3.9561E+02,
     + 3.9656E+02, 3.9752E+02, 3.9847E+02, 3.9943E+02, 4.0038E+02,
     +  4.0134E+02, 4.0230E+02, 4.0325E+02, 4.0421E+02, 4.0517E+02,
     + 4.0614E+02, 4.0710E+02, 4.0806E+02, 4.0902E+02, 4.0999E+02,
     +  4.1096E+02, 4.1192E+02, 4.1289E+02, 4.1386E+02, 4.1483E+02,
     + 4.1580E+02, 4.1677E+02, 4.1774E+02, 4.1871E+02, 4.1969E+02,
     +  4.2066E+02, 4.2164E+02, 4.2262E+02, 4.2359E+02, 4.2457E+02,
     + 4.2555E+02, 4.2653E+02, 4.2751E+02, 4.2850E+02, 4.2948E+02,
     +  4.3046E+02, 4.3145E+02, 4.3243E+02, 4.3342E+02, 4.3441E+02,
     + 4.3540E+02, 4.3639E+02, 4.3738E+02, 4.3837E+02, 4.3936E+02,
     +  4.4035E+02, 4.4135E+02, 4.4234E+02, 4.4334E+02, 4.4434E+02,
     + 4.4534E+02, 4.4633E+02, 4.4733E+02, 4.4833E+02, 4.4934E+02,
     +  4.5034E+02, 4.5134E+02, 4.5235E+02, 4.5335E+02, 4.5436E+02,
     + 4.5537E+02, 4.5638E+02, 4.5739E+02, 4.5840E+02, 4.5941E+02,
     +  4.6042E+02, 4.6143E+02, 4.6245E+02, 4.6346E+02, 4.6448E+02,
     + 4.6549E+02, 4.6651E+02, 4.6753E+02, 4.6855E+02, 4.6957E+02,
     +  4.7059E+02, 4.7162E+02, 4.7264E+02, 4.7367E+02, 4.7469E+02,
     + 4.7572E+02, 4.7675E+02, 4.7777E+02, 4.7880E+02, 4.7983E+02,
     +  4.8087E+02, 4.8190E+02, 4.8293E+02, 4.8397E+02, 4.8500E+02,
     + 4.8604E+02, 4.8707E+02, 4.8811E+02, 4.8915E+02, 4.9019E+02,
     +  4.9123E+02, 4.9228E+02, 4.9332E+02, 4.9436E+02, 4.9541E+02,
     + 4.9645E+02, 4.9750E+02, 4.9855E+02, 4.9960E+02, 5.0065E+02,
     +  5.0170E+02, 5.0275E+02, 5.0380E+02, 5.0486E+02, 5.0591E+02,
     + 5.0697E+02, 5.0803E+02, 5.0908E+02, 5.1014E+02, 5.1120E+02,
     +  5.1226E+02, 5.1332E+02, 5.1439E+02, 5.1545E+02, 5.1652E+02,
     + 5.1758E+02, 5.1865E+02, 5.1972E+02, 5.2078E+02, 5.2185E+02,
     +  5.2293E+02, 5.2400E+02, 5.2507E+02, 5.2614E+02, 5.2722E+02,
     + 5.2829E+02, 5.2937E+02, 5.3045E+02, 5.3153E+02, 5.3261E+02,
     +  5.3369E+02, 5.3477E+02, 5.3585E+02, 5.3694E+02, 5.3802E+02,
     + 5.3911E+02, 5.4019E+02, 5.4128E+02, 5.4237E+02, 5.4346E+02,
     +  5.4455E+02, 5.4564E+02, 5.4674E+02, 5.4783E+02, 5.4892E+02,
     + 5.5002E+02, 5.5112E+02, 5.5221E+02, 5.5331E+02, 5.5441E+02,
     +  5.5551E+02, 5.5662E+02, 5.5772E+02, 5.5882E+02, 5.5993E+02,
     + 5.6103E+02, 5.6214E+02, 5.6325E+02, 5.6436E+02, 5.6547E+02,
     +  5.6658E+02, 5.6769E+02, 5.6880E+02, 5.6992E+02, 5.7103E+02,
     + 5.7215E+02, 5.7327E+02, 5.7438E+02, 5.7550E+02, 5.7662E+02,
     +  5.7774E+02, 5.7887E+02, 5.7999E+02, 5.8111E+02, 5.8224E+02,
     + 5.8337E+02, 5.8449E+02, 5.8562E+02, 5.8675E+02, 5.8788E+02,
     +  5.8901E+02, 5.9015E+02, 5.9128E+02, 5.9241E+02, 5.9355E+02,
     + 5.9469E+02, 5.9582E+02, 5.9696E+02, 5.9810E+02, 5.9924E+02,
     +  6.0038E+02, 6.0153E+02, 6.0267E+02, 6.0382E+02, 6.0496E+02,
     + 6.0611E+02, 6.0726E+02, 6.0841E+02, 6.0956E+02, 6.1071E+02,
     +  6.1186E+02, 6.1301E+02, 6.1417E+02, 6.1532E+02, 6.1648E+02,
     + 6.1764E+02, 6.1879E+02, 6.1995E+02, 6.2111E+02, 6.2228E+02,
     +  6.2344E+02, 6.2460E+02, 6.2577E+02, 6.2693E+02, 6.2810E+02,
     + 6.2927E+02, 6.3043E+02, 6.3160E+02, 6.3277E+02, 6.3395E+02,
     +  6.3512E+02, 6.3629E+02, 6.3747E+02, 6.3864E+02, 6.3982E+02,
     + 6.4100E+02, 6.4218E+02, 6.4336E+02, 6.4454E+02, 6.4572E+02,
     +  6.4691E+02, 6.4809E+02, 6.4928E+02, 6.5046E+02, 6.5165E+02,
     + 6.5284E+02, 6.5403E+02, 6.5522E+02, 6.5641E+02, 6.5760E+02,
     +  6.5880E+02, 6.5999E+02, 6.6119E+02, 6.6238E+02, 6.6358E+02,
     + 6.6478E+02, 6.6598E+02, 6.6718E+02, 6.6839E+02, 6.6959E+02,
     +  6.7079E+02, 6.7200E+02, 6.7320E+02, 6.7441E+02, 6.7562E+02,
     + 6.7683E+02, 6.7804E+02, 6.7925E+02, 6.8047E+02, 6.8168E+02,
     +  6.8289E+02, 6.8411E+02, 6.8533E+02, 6.8654E+02, 6.8776E+02,
     + 6.8898E+02, 6.9021E+02, 6.9143E+02, 6.9265E+02, 6.9388E+02,
     +  6.9510E+02, 6.9633E+02, 6.9755E+02, 6.9878E+02, 7.0001E+02,
     + 7.0124E+02, 7.0248E+02, 7.0371E+02, 7.0494E+02, 7.0618E+02,
     +  7.0741E+02, 7.0865E+02, 7.0989E+02, 7.1113E+02, 7.1237E+02,
     + 7.1361E+02, 7.1485E+02, 7.1610E+02, 7.1734E+02, 7.1859E+02,
     +  7.1983E+02, 7.2108E+02, 7.2233E+02, 7.2358E+02, 7.2483E+02,
     + 7.2608E+02, 7.2734E+02, 7.2859E+02, 7.2984E+02, 7.3110E+02,
     +  7.3236E+02, 7.3362E+02, 7.3488E+02, 7.3614E+02, 7.3740E+02,
     + 7.3866E+02, 7.3992E+02, 7.4119E+02, 7.4246E+02, 7.4372E+02,
     +  7.4499E+02, 7.4626E+02, 7.4753E+02, 7.4880E+02, 7.5007E+02,
     + 7.5135E+02, 7.5262E+02, 7.5390E+02, 7.5517E+02, 7.5645E+02,
     +  7.5773E+02, 7.5901E+02, 7.6029E+02, 7.6157E+02, 7.6286E+02,
     + 7.6414E+02, 7.6542E+02, 7.6671E+02, 7.6800E+02, 7.6929E+02,
     +  7.7058E+02, 7.7187E+02, 7.7316E+02, 7.7445E+02, 7.7574E+02,
     + 7.7704E+02, 7.7834E+02, 7.7963E+02, 7.8093E+02, 7.8223E+02,
     +  7.8353E+02, 7.8483E+02, 7.8613E+02, 7.8744E+02, 7.8874E+02,
     + 7.9005E+02, 7.9135E+02, 7.9266E+02, 7.9397E+02, 7.9528E+02,
     +  7.9659E+02, 7.9790E+02, 7.9922E+02, 8.0053E+02, 8.0185E+02,
     + 8.0316E+02, 8.0448E+02, 8.0580E+02, 8.0712E+02, 8.0844E+02,
     +  8.0976E+02, 8.1108E+02, 8.1241E+02, 8.1373E+02, 8.1506E+02,
     + 8.1639E+02, 8.1771E+02, 8.1904E+02, 8.2037E+02, 8.2171E+02,
     +  8.2304E+02, 8.2437E+02, 8.2571E+02, 8.2704E+02, 8.2838E+02,
     + 8.2972E+02, 8.3106E+02, 8.3240E+02, 8.3374E+02, 8.3508E+02,
     +  8.3642E+02, 8.3777E+02, 8.3911E+02, 8.4046E+02, 8.4181E+02,
     + 8.4316E+02, 8.4451E+02, 8.4586E+02, 8.4721E+02, 8.4856E+02,
     +  8.4992E+02, 8.5127E+02, 8.5263E+02, 8.5399E+02, 8.5535E+02,
     + 8.5671E+02, 8.5807E+02, 8.5943E+02, 8.6079E+02, 8.6216E+02,
     +  8.6352E+02, 8.6489E+02, 8.6626E+02, 8.6762E+02, 8.6899E+02,
     + 8.7036E+02, 8.7174E+02, 8.7311E+02, 8.7448E+02, 8.7586E+02,
     +  8.7723E+02, 8.7861E+02, 8.7999E+02, 8.8137E+02, 8.8275E+02,
     + 8.8413E+02, 8.8551E+02, 8.8690E+02, 8.8828E+02, 8.8967E+02,
     +  8.9106E+02, 8.9244E+02, 8.9383E+02, 8.9522E+02, 8.9661E+02,
     + 8.9801E+02, 8.9940E+02, 9.0079E+02, 9.0219E+02, 9.0359E+02,
     +  9.0499E+02, 9.0638E+02, 9.0778E+02, 9.0919E+02, 9.1059E+02,
     + 9.1199E+02, 9.1340E+02, 9.1480E+02, 9.1621E+02, 9.1762E+02,
     +  9.1903E+02, 9.2044E+02, 9.2185E+02, 9.2326E+02, 9.2467E+02,
     + 9.2609E+02, 9.2750E+02, 9.2892E+02, 9.3034E+02, 9.3175E+02,
     +  9.3317E+02, 9.3460E+02, 9.3602E+02, 9.3744E+02, 9.3887E+02,
     + 9.4029E+02, 9.4172E+02, 9.4314E+02, 9.4457E+02, 9.4600E+02,
     +  9.4743E+02, 9.4887E+02, 9.5030E+02, 9.5173E+02, 9.5317E+02,
     + 9.5460E+02, 9.5604E+02, 9.5748E+02, 9.5892E+02, 9.6036E+02,
     +  9.6180E+02, 9.6325E+02, 9.6469E+02, 9.6613E+02, 9.6758E+02,
     + 9.6903E+02, 9.7048E+02, 9.7193E+02, 9.7338E+02, 9.7483E+02,
     +  9.7628E+02, 9.7774E+02, 9.7919E+02, 9.8065E+02, 9.8210E+02,
     + 9.8356E+02, 9.8502E+02, 9.8648E+02, 9.8794E+02, 9.8941E+02,
     +  9.9087E+02, 9.9234E+02, 9.9380E+02, 9.9527E+02, 9.9674E+02,
     + 9.9821E+02, 9.9968E+02, 1.0011E+03, 1.0026E+03, 1.0041E+03,
     +  1.0056E+03, 1.0070E+03, 1.0085E+03, 1.0100E+03, 1.0115E+03,
     + 1.0130E+03, 1.0144E+03, 1.0159E+03, 1.0174E+03, 1.0189E+03,
     +  1.0204E+03, 1.0219E+03, 1.0234E+03, 1.0248E+03, 1.0263E+03,
     + 1.0278E+03, 1.0293E+03, 1.0308E+03, 1.0323E+03, 1.0338E+03,
     +  1.0353E+03, 1.0368E+03, 1.0383E+03, 1.0398E+03, 1.0413E+03,
     + 1.0428E+03, 1.0443E+03, 1.0458E+03, 1.0473E+03, 1.0488E+03,
     +  1.0503E+03, 1.0518E+03, 1.0534E+03, 1.0549E+03, 1.0564E+03,
     + 1.0579E+03, 1.0594E+03, 1.0609E+03, 1.0624E+03, 1.0640E+03,
     +  1.0655E+03, 1.0670E+03, 1.0685E+03, 1.0700E+03, 1.0716E+03,
     + 1.0731E+03, 1.0746E+03, 1.0761E+03, 1.0777E+03, 1.0792E+03,
     +  1.0807E+03, 1.0823E+03, 1.0838E+03, 1.0853E+03, 1.0869E+03,
     + 1.0884E+03, 1.0899E+03, 1.0915E+03, 1.0930E+03, 1.0945E+03,
     +  1.0961E+03, 1.0976E+03, 1.0992E+03, 1.1007E+03, 1.1023E+03,
     + 1.1038E+03, 1.1054E+03, 1.1069E+03, 1.1085E+03, 1.1100E+03,
     +  1.1116E+03, 1.1131E+03, 1.1147E+03, 1.1162E+03, 1.1178E+03,
     + 1.1193E+03, 1.1209E+03, 1.1225E+03, 1.1240E+03, 1.1256E+03,
     +  1.1272E+03, 1.1287E+03, 1.1303E+03, 1.1319E+03, 1.1334E+03,
     + 1.1350E+03, 1.1366E+03, 1.1381E+03, 1.1397E+03, 1.1413E+03,
     +  1.1429E+03, 1.1444E+03, 1.1460E+03, 1.1476E+03, 1.1492E+03,
     + 1.1507E+03, 1.1523E+03, 1.1539E+03, 1.1555E+03, 1.1571E+03,
     +  1.1587E+03, 1.1602E+03, 1.1618E+03, 1.1634E+03, 1.1650E+03,
     + 1.1666E+03, 1.1682E+03, 1.1698E+03, 1.1714E+03, 1.1730E+03,
     +  1.1746E+03, 1.1762E+03, 1.1778E+03, 1.1794E+03, 1.1810E+03,
     + 1.1826E+03, 1.1842E+03, 1.1858E+03, 1.1874E+03, 1.1890E+03,
     +  1.1906E+03, 1.1922E+03, 1.1938E+03, 1.1954E+03, 1.1971E+03,
     + 1.1987E+03, 1.2003E+03, 1.2019E+03, 1.2035E+03, 1.2051E+03,
     +  1.2068E+03, 1.2084E+03, 1.2100E+03, 1.2116E+03, 1.2133E+03,
     + 1.2149E+03, 1.2165E+03, 1.2181E+03, 1.2198E+03, 1.2214E+03,
     +  1.2230E+03, 1.2246E+03, 1.2263E+03, 1.2279E+03, 1.2296E+03,
     + 1.2312E+03, 1.2328E+03, 1.2345E+03, 1.2361E+03, 1.2377E+03,
     +  1.2394E+03, 1.2410E+03, 1.2427E+03, 1.2443E+03, 1.2460E+03,
     + 1.2476E+03, 1.2493E+03, 1.2509E+03, 1.2526E+03, 1.2542E+03,
     +  1.2559E+03, 1.2575E+03, 1.2592E+03, 1.2608E+03, 1.2625E+03,
     + 1.2642E+03, 1.2658E+03, 1.2675E+03, 1.2691E+03, 1.2708E+03,
     +  1.2725E+03, 1.2741E+03, 1.2758E+03, 1.2775E+03, 1.2791E+03,
     + 1.2808E+03, 1.2825E+03, 1.2841E+03, 1.2858E+03, 1.2875E+03,
     +  1.2892E+03, 1.2908E+03, 1.2925E+03, 1.2942E+03, 1.2959E+03,
     + 1.2976E+03, 1.2992E+03, 1.3009E+03, 1.3026E+03, 1.3043E+03,
     +  1.3060E+03, 1.3077E+03, 1.3094E+03, 1.3111E+03, 1.3127E+03,
     + 1.3144E+03, 1.3161E+03, 1.3178E+03, 1.3195E+03, 1.3212E+03,
     +  1.3229E+03, 1.3246E+03, 1.3263E+03, 1.3280E+03, 1.3297E+03,
     + 1.3314E+03, 1.3331E+03, 1.3348E+03, 1.3365E+03, 1.3383E+03,
     +  1.3400E+03, 1.3417E+03, 1.3434E+03, 1.3451E+03, 1.3468E+03,
     + 1.3485E+03, 1.3502E+03, 1.3520E+03, 1.3537E+03, 1.3554E+03,
     +  1.3571E+03, 1.3588E+03, 1.3606E+03, 1.3623E+03, 1.3640E+03,
     + 1.3657E+03, 1.3675E+03, 1.3692E+03, 1.3709E+03, 1.3727E+03,
     +  1.3744E+03, 1.3761E+03, 1.3779E+03, 1.3796E+03, 1.3813E+03,
     + 1.3831E+03, 1.3848E+03, 1.3865E+03, 1.3883E+03, 1.3900E+03,
     +  1.3918E+03, 1.3935E+03, 1.3953E+03, 1.3970E+03, 1.3988E+03,
     + 1.4005E+03, 1.4023E+03, 1.4040E+03, 1.4058E+03, 1.4075E+03,
     +  1.4093E+03, 1.4110E+03, 1.4128E+03, 1.4145E+03, 1.4163E+03,
     + 1.4181E+03, 1.4198E+03, 1.4216E+03, 1.4233E+03, 1.4251E+03,
     +  1.4269E+03, 1.4286E+03, 1.4304E+03, 1.4322E+03, 1.4340E+03,
     + 1.4357E+03, 1.4375E+03, 1.4393E+03, 1.4411E+03, 1.4428E+03,
     +  1.4446E+03, 1.4464E+03, 1.4482E+03, 1.4499E+03, 1.4517E+03,
     + 1.4535E+03, 1.4553E+03, 1.4571E+03, 1.4589E+03, 1.4607E+03,
     +  1.4624E+03, 1.4642E+03, 1.4660E+03, 1.4678E+03, 1.4696E+03,
     + 1.4714E+03, 1.4732E+03, 1.4750E+03, 1.4768E+03, 1.4786E+03,
     +  1.4804E+03, 1.4822E+03, 1.4840E+03, 1.4858E+03, 1.4876E+03,
     + 1.4894E+03, 1.4912E+03, 1.4930E+03, 1.4948E+03, 1.4966E+03,
     +  1.4985E+03, 1.5003E+03, 1.5021E+03, 1.5039E+03, 1.5057E+03,
     + 1.5075E+03, 1.5093E+03, 1.5112E+03, 1.5130E+03, 1.5148E+03,
     +  1.5166E+03, 1.5185E+03, 1.5203E+03, 1.5221E+03, 1.5239E+03,
     + 1.5258E+03, 1.5276E+03, 1.5294E+03, 1.5313E+03, 1.5331E+03,
     +  1.5349E+03, 1.5368E+03, 1.5386E+03, 1.5404E+03, 1.5423E+03,
     + 1.5441E+03, 1.5460E+03, 1.5478E+03, 1.5496E+03, 1.5515E+03,
     +  1.5533E+03, 1.5552E+03, 1.5570E+03, 1.5589E+03, 1.5607E+03,
     + 1.5626E+03, 1.5644E+03, 1.5663E+03, 1.5681E+03, 1.5700E+03,
     +  1.5718E+03, 1.5737E+03, 1.5756E+03, 1.5774E+03, 1.5793E+03,
     + 1.5811E+03, 1.5830E+03, 1.5849E+03, 1.5867E+03, 1.5886E+03,
     +  1.5905E+03, 1.5923E+03, 1.5942E+03, 1.5961E+03, 1.5980E+03,
     + 1.5998E+03, 1.6017E+03, 1.6036E+03, 1.6055E+03, 1.6073E+03,
     +  1.6092E+03, 1.6111E+03, 1.6130E+03, 1.6149E+03, 1.6168E+03,
     + 1.6186E+03, 1.6205E+03, 1.6224E+03, 1.6243E+03, 1.6262E+03,
     +  1.6281E+03, 1.6300E+03, 1.6319E+03, 1.6338E+03, 1.6357E+03,
     + 1.6376E+03, 1.6395E+03, 1.6414E+03, 1.6433E+03, 1.6452E+03,
     +  1.6471E+03, 1.6490E+03, 1.6509E+03, 1.6528E+03, 1.6547E+03,
     + 1.6566E+03, 1.6585E+03, 1.6604E+03, 1.6623E+03, 1.6642E+03,
     +  1.6662E+03, 1.6681E+03, 1.6700E+03, 1.6719E+03, 1.6738E+03,
     + 1.6757E+03, 1.6777E+03, 1.6796E+03, 1.6815E+03, 1.6834E+03,
     +  1.6854E+03, 1.6873E+03, 1.6892E+03, 1.6911E+03, 1.6931E+03,
     + 1.6950E+03, 1.6969E+03, 1.6989E+03, 1.7008E+03, 1.7027E+03,
     +  1.7047E+03, 1.7066E+03, 1.7086E+03, 1.7105E+03, 1.7124E+03,
     + 1.7144E+03, 1.7163E+03, 1.7183E+03, 1.7202E+03, 1.7222E+03,
     +  1.7241E+03, 1.7261E+03, 1.7280E+03, 1.7300E+03, 1.7319E+03,
     + 1.7339E+03, 1.7358E+03, 1.7378E+03, 1.7397E+03, 1.7417E+03,
     +  1.7437E+03, 1.7456E+03, 1.7476E+03, 1.7495E+03, 1.7515E+03,
     + 1.7535E+03, 1.7554E+03, 1.7574E+03, 1.7594E+03, 1.7613E+03,
     +  1.7633E+03, 1.7653E+03, 1.7673E+03, 1.7692E+03, 1.7712E+03,
     + 1.7732E+03, 1.7752E+03, 1.7772E+03, 1.7791E+03, 1.7811E+03,
     +  1.7831E+03, 1.7851E+03, 1.7871E+03, 1.7891E+03, 1.7910E+03,
     + 1.7930E+03, 1.7950E+03, 1.7970E+03, 1.7990E+03, 1.8010E+03,
     +  1.8030E+03, 1.8050E+03, 1.8070E+03, 1.8090E+03, 1.8110E+03,
     + 1.8130E+03, 1.8150E+03, 1.8170E+03, 1.8190E+03, 1.8210E+03,
     +  1.8230E+03, 1.8250E+03, 1.8270E+03, 1.8290E+03, 1.8310E+03,
     + 1.8330E+03, 1.8351E+03, 1.8371E+03, 1.8391E+03, 1.8411E+03,
     +  1.8431E+03, 1.8451E+03, 1.8472E+03, 1.8492E+03, 1.8512E+03,
     + 1.8532E+03, 1.8553E+03, 1.8573E+03, 1.8593E+03, 1.8613E+03,
     +  1.8634E+03, 1.8654E+03, 1.8674E+03, 1.8695E+03, 1.8715E+03,
     + 1.8735E+03, 1.8756E+03, 1.8776E+03, 1.8796E+03, 1.8817E+03,
     +  1.8837E+03, 1.8858E+03, 1.8878E+03, 1.8898E+03, 1.8919E+03,
     + 1.8939E+03, 1.8960E+03, 1.8980E+03, 1.9001E+03, 1.9021E+03,
     +  1.9042E+03, 1.9062E+03, 1.9083E+03, 1.9103E+03, 1.9124E+03,
     + 1.9145E+03, 1.9165E+03, 1.9186E+03, 1.9206E+03, 1.9227E+03,
     +  1.9248E+03, 1.9268E+03, 1.9289E+03, 1.9310E+03, 1.9330E+03,
     + 1.9351E+03, 1.9372E+03, 1.9392E+03, 1.9413E+03, 1.9434E+03,
     +  1.9455E+03, 1.9475E+03, 1.9496E+03, 1.9517E+03, 1.9538E+03,
     + 1.9559E+03, 1.9579E+03, 1.9600E+03, 1.9621E+03, 1.9642E+03,
     +  1.9663E+03, 1.9684E+03, 1.9705E+03, 1.9726E+03, 1.9746E+03,
     + 1.9767E+03, 1.9788E+03, 1.9809E+03, 1.9830E+03, 1.9851E+03,
     +  1.9872E+03, 1.9893E+03, 1.9914E+03, 1.9935E+03, 1.9956E+03,
     + 1.9977E+03, 1.9998E+03, 2.0019E+03, 2.0040E+03, 2.0062E+03,
     +  2.0083E+03, 2.0104E+03, 2.0125E+03, 2.0146E+03, 2.0167E+03,
     + 2.0188E+03, 2.0209E+03, 2.0231E+03, 2.0252E+03, 2.0273E+03,
     +  2.0294E+03, 2.0316E+03, 2.0337E+03, 2.0358E+03, 2.0379E+03,
     + 2.0401E+03, 2.0422E+03, 2.0443E+03, 2.0464E+03, 2.0486E+03,
     +  2.0507E+03, 2.0528E+03, 2.0550E+03, 2.0571E+03, 2.0593E+03,
     + 2.0614E+03, 2.0635E+03, 2.0657E+03, 2.0678E+03, 2.0700E+03,
     +  2.0721E+03, 2.0742E+03, 2.0764E+03, 2.0785E+03, 2.0807E+03,
     + 2.0828E+03, 2.0850E+03, 2.0871E+03, 2.0893E+03, 2.0915E+03,
     +  2.0936E+03, 2.0958E+03, 2.0979E+03, 2.1001E+03, 2.1023E+03,
     + 2.1044E+03, 2.1066E+03, 2.1087E+03, 2.1109E+03, 2.1131E+03,
     +  2.1152E+03, 2.1174E+03, 2.1196E+03, 2.1218E+03, 2.1239E+03,
     + 2.1261E+03, 2.1283E+03, 2.1305E+03, 2.1326E+03, 2.1348E+03,
     +  2.1370E+03, 2.1392E+03, 2.1413E+03, 2.1435E+03, 2.1457E+03,
     + 2.1479E+03, 2.1501E+03, 2.1523E+03, 2.1545E+03, 2.1567E+03,
     +  2.1588E+03, 2.1610E+03, 2.1632E+03, 2.1654E+03, 2.1676E+03,
     + 2.1698E+03, 2.1720E+03, 2.1742E+03, 2.1764E+03, 2.1786E+03,
     +  2.1808E+03, 2.1830E+03, 2.1852E+03, 2.1874E+03, 2.1896E+03,
     + 2.1919E+03, 2.1941E+03, 2.1963E+03, 2.1985E+03, 2.2007E+03,
     +  2.2029E+03, 2.2051E+03, 2.2073E+03, 2.2096E+03, 2.2118E+03,
     + 2.2140E+03, 2.2162E+03, 2.2185E+03, 2.2207E+03, 2.2229E+03,
     +  2.2251E+03, 2.2274E+03, 2.2296E+03, 2.2318E+03, 2.2340E+03,
     + 2.2363E+03, 2.2385E+03, 2.2407E+03, 2.2430E+03, 2.2452E+03,
     +  2.2475E+03, 2.2497E+03, 2.2519E+03, 2.2542E+03, 2.2564E+03,
     + 2.2587E+03, 2.2609E+03, 2.2632E+03, 2.2654E+03, 2.2676E+03,
     +  2.2699E+03, 2.2721E+03, 2.2744E+03, 2.2767E+03, 2.2789E+03,
     + 2.2812E+03, 2.2834E+03, 2.2857E+03, 2.2879E+03, 2.2902E+03,
     +  2.2925E+03, 2.2947E+03, 2.2970E+03, 2.2992E+03, 2.3015E+03,
     + 2.3038E+03, 2.3060E+03, 2.3083E+03, 2.3106E+03, 2.3129E+03,
     +  2.3151E+03, 2.3174E+03, 2.3197E+03, 2.3220E+03, 2.3242E+03,
     + 2.3265E+03, 2.3288E+03, 2.3311E+03, 2.3334E+03, 2.3356E+03,
     +  2.3379E+03, 2.3402E+03, 2.3425E+03, 2.3448E+03, 2.3471E+03,
     + 2.3494E+03, 2.3517E+03, 2.3540E+03, 2.3562E+03, 2.3585E+03,
     +  2.3608E+03, 2.3631E+03, 2.3654E+03, 2.3677E+03, 2.3700E+03,
     + 2.3723E+03, 2.3746E+03, 2.3769E+03, 2.3793E+03, 2.3816E+03,
     +  2.3839E+03, 2.3862E+03, 2.3885E+03, 2.3908E+03, 2.3931E+03,
     + 2.3954E+03, 2.3977E+03, 2.4001E+03, 2.4024E+03, 2.4047E+03,
     +  2.4070E+03, 2.4093E+03, 2.4117E+03, 2.4140E+03, 2.4163E+03,
     + 2.4186E+03, 2.4210E+03, 2.4233E+03, 2.4256E+03, 2.4279E+03,
     +  2.4303E+03, 2.4326E+03, 2.4349E+03, 2.4373E+03, 2.4396E+03,
     + 2.4420E+03, 2.4443E+03, 2.4466E+03, 2.4490E+03, 2.4513E+03,
     +  2.4537E+03, 2.4560E+03, 2.4583E+03, 2.4607E+03, 2.4630E+03,
     + 2.4654E+03, 2.4677E+03, 2.4701E+03, 2.4724E+03, 2.4748E+03,
     +  2.4772E+03, 2.4795E+03, 2.4819E+03, 2.4842E+03, 2.4866E+03,
     + 2.4889E+03, 2.4913E+03, 2.4937E+03, 2.4960E+03, 2.4984E+03,
     +  2.5008E+03, 2.5031E+03, 2.5055E+03, 2.5079E+03, 2.5102E+03,
     + 2.5126E+03, 2.5150E+03, 2.5174E+03, 2.5197E+03, 2.5221E+03,
     +  2.5245E+03, 2.5269E+03, 2.5293E+03, 2.5316E+03, 2.5340E+03,
     + 2.5364E+03, 2.5388E+03, 2.5412E+03, 2.5436E+03, 2.5460E+03,
     +  2.5484E+03, 2.5507E+03, 2.5531E+03, 2.5555E+03, 2.5579E+03,
     + 2.5603E+03, 2.5627E+03, 2.5651E+03, 2.5675E+03, 2.5699E+03,
     +  2.5723E+03, 2.5747E+03, 2.5771E+03, 2.5795E+03, 2.5819E+03,
     + 2.5843E+03, 2.5868E+03, 2.5892E+03, 2.5916E+03, 2.5940E+03,
     +  2.5964E+03, 2.5988E+03, 2.6012E+03, 2.6037E+03, 2.6061E+03,
     + 2.6085E+03, 2.6109E+03, 2.6133E+03, 2.6158E+03, 2.6182E+03,
     +  2.6206E+03, 2.6230E+03, 2.6255E+03, 2.6279E+03, 2.6303E+03,
     + 2.6328E+03, 2.6352E+03, 2.6376E+03, 2.6401E+03, 2.6425E+03,
     +  2.6449E+03, 2.6474E+03, 2.6498E+03, 2.6523E+03, 2.6547E+03,
     + 2.6571E+03, 2.6596E+03, 2.6620E+03, 2.6645E+03, 2.6669E+03,
     +  2.6694E+03, 2.6718E+03, 2.6743E+03, 2.6767E+03, 2.6792E+03,
     + 2.6816E+03, 2.6841E+03, 2.6866E+03, 2.6890E+03, 2.6915E+03,
     +  2.6939E+03, 2.6964E+03, 2.6989E+03, 2.7013E+03, 2.7038E+03,
     + 2.7063E+03, 2.7087E+03, 2.7112E+03, 2.7137E+03, 2.7161E+03,
     +  2.7186E+03, 2.7211E+03, 2.7236E+03, 2.7260E+03, 2.7285E+03,
     + 2.7310E+03, 2.7335E+03, 2.7360E+03, 2.7384E+03, 2.7409E+03,
     +  2.7434E+03, 2.7459E+03, 2.7484E+03, 2.7509E+03, 2.7534E+03,
     + 2.7558E+03, 2.7583E+03, 2.7608E+03, 2.7633E+03, 2.7658E+03,
     +  2.7683E+03, 2.7708E+03, 2.7733E+03, 2.7758E+03, 2.7783E+03,
     + 2.7808E+03, 2.7833E+03, 2.7858E+03, 2.7883E+03, 2.7908E+03,
     +  2.7933E+03, 2.7959E+03, 2.7984E+03, 2.8009E+03, 2.8034E+03,
     + 2.8059E+03, 2.8084E+03, 2.8109E+03, 2.8135E+03, 2.8160E+03,
     +  2.8185E+03, 2.8210E+03, 2.8235E+03, 2.8261E+03, 2.8286E+03,
     + 2.8311E+03, 2.8336E+03, 2.8362E+03, 2.8387E+03, 2.8412E+03,
     +  2.8438E+03, 2.8463E+03, 2.8488E+03, 2.8514E+03, 2.8539E+03,
     + 2.8564E+03, 2.8590E+03, 2.8615E+03, 2.8641E+03, 2.8666E+03,
     +  2.8692E+03, 2.8717E+03, 2.8742E+03, 2.8768E+03, 2.8793E+03,
     + 2.8819E+03, 2.8844E+03, 2.8870E+03, 2.8896E+03, 2.8921E+03,
     +  2.8947E+03, 2.8972E+03, 2.8998E+03, 2.9023E+03, 2.9049E+03,
     + 2.9075E+03, 2.9100E+03, 2.9126E+03, 2.9152E+03, 2.9177E+03,
     +  2.9203E+03, 2.9229E+03, 2.9254E+03, 2.9280E+03, 2.9306E+03,
     + 2.9331E+03, 2.9357E+03, 2.9383E+03, 2.9409E+03, 2.9435E+03,
     +  2.9460E+03, 2.9486E+03, 2.9512E+03, 2.9538E+03, 2.9564E+03,
     + 2.9590E+03, 2.9615E+03, 2.9641E+03, 2.9667E+03, 2.9693E+03,
     +  2.9719E+03, 2.9745E+03, 2.9771E+03, 2.9797E+03, 2.9823E+03,
     + 2.9849E+03, 2.9875E+03, 2.9901E+03, 2.9927E+03, 2.9953E+03,
     +  2.9979E+03, 3.0005E+03, 3.0031E+03, 3.0057E+03, 3.0083E+03,
     + 3.0109E+03, 3.0135E+03, 3.0161E+03, 3.0188E+03, 3.0214E+03,
     +  3.0240E+03, 3.0266E+03, 3.0292E+03, 3.0318E+03, 3.0345E+03,
     + 3.0371E+03, 3.0397E+03, 3.0423E+03, 3.0449E+03, 3.0476E+03,
     +  3.0502E+03, 3.0528E+03, 3.0555E+03, 3.0581E+03, 3.0607E+03,
     + 3.0634E+03, 3.0660E+03, 3.0686E+03, 3.0713E+03, 3.0739E+03,
     +  3.0765E+03, 3.0792E+03, 3.0818E+03, 3.0845E+03, 3.0871E+03,
     + 3.0898E+03, 3.0924E+03, 3.0951E+03, 3.0977E+03, 3.1004E+03,
     +  3.1030E+03, 3.1057E+03, 3.1083E+03, 3.1110E+03, 3.1136E+03,
     + 3.1163E+03, 3.1189E+03, 3.1216E+03, 3.1243E+03, 3.1269E+03,
     +  3.1296E+03, 3.1322E+03, 3.1349E+03, 3.1376E+03, 3.1402E+03,
     + 3.1429E+03, 3.1456E+03, 3.1483E+03, 3.1509E+03, 3.1536E+03,
     +  3.1563E+03, 3.1590E+03, 3.1616E+03, 3.1643E+03, 3.1670E+03,
     + 3.1697E+03, 3.1724E+03, 3.1750E+03, 3.1777E+03, 3.1804E+03,
     +  3.1831E+03, 3.1858E+03, 3.1885E+03, 3.1912E+03, 3.1939E+03,
     + 3.1966E+03, 3.1992E+03, 3.2019E+03, 3.2046E+03, 3.2073E+03,
     +  3.2100E+03, 3.2127E+03, 3.2154E+03, 3.2181E+03, 3.2208E+03,
     + 3.2236E+03, 3.2263E+03, 3.2290E+03, 3.2317E+03, 3.2344E+03,
     +  3.2371E+03, 3.2398E+03, 3.2425E+03, 3.2452E+03, 3.2480E+03,
     + 3.2507E+03, 3.2534E+03, 3.2561E+03, 3.2588E+03, 3.2616E+03,
     +  3.2643E+03, 3.2670E+03, 3.2697E+03, 3.2725E+03, 3.2752E+03,
     + 3.2779E+03, 3.2806E+03, 3.2834E+03, 3.2861E+03, 3.2888E+03,
     +  3.2916E+03, 3.2943E+03, 3.2970E+03, 3.2998E+03, 3.3025E+03,
     + 3.3053E+03, 3.3080E+03, 3.3108E+03, 3.3135E+03, 3.3162E+03,
     +  3.3190E+03, 3.3217E+03, 3.3245E+03, 3.3272E+03, 3.3300E+03,
     + 3.3327E+03, 3.3355E+03, 3.3383E+03, 3.3410E+03, 3.3438E+03,
     +  3.3465E+03, 3.3493E+03, 3.3520E+03, 3.3548E+03, 3.3576E+03,
     + 3.3603E+03, 3.3631E+03, 3.3659E+03, 3.3686E+03, 3.3714E+03,
     +  3.3742E+03, 3.3770E+03, 3.3797E+03, 3.3825E+03, 3.3853E+03,
     + 3.3881E+03, 3.3908E+03, 3.3936E+03, 3.3964E+03, 3.3992E+03,
     +  3.4020E+03, 3.4047E+03, 3.4075E+03, 3.4103E+03, 3.4131E+03,
     + 3.4159E+03, 3.4187E+03, 3.4215E+03, 3.4243E+03, 3.4271E+03,
     +  3.4299E+03, 3.4327E+03, 3.4355E+03, 3.4383E+03, 3.4411E+03,
     + 3.4439E+03, 3.4467E+03, 3.4495E+03, 3.4523E+03, 3.4551E+03,
     +  3.4579E+03, 3.4607E+03, 3.4635E+03, 3.4663E+03, 3.4691E+03,
     + 3.4719E+03, 3.4748E+03, 3.4776E+03, 3.4804E+03, 3.4832E+03,
     +  3.4860E+03, 3.4888E+03, 3.4917E+03, 3.4945E+03, 3.4973E+03,
     + 3.5001E+03, 3.5030E+03, 3.5058E+03, 3.5086E+03, 3.5114E+03,
     +  3.5143E+03, 3.5171E+03, 3.5199E+03, 3.5228E+03, 3.5256E+03,
     + 3.5285E+03, 3.5313E+03, 3.5341E+03, 3.5370E+03, 3.5398E+03,
     +  3.5427E+03, 3.5455E+03, 3.5484E+03, 3.5512E+03, 3.5540E+03,
     + 3.5569E+03, 3.5597E+03, 3.5626E+03, 3.5655E+03, 3.5683E+03,
     +  3.5712E+03, 3.5740E+03, 3.5769E+03, 3.5797E+03, 3.5826E+03,
     + 3.5855E+03, 3.5883E+03, 3.5912E+03, 3.5941E+03, 3.5969E+03,
     +  3.5998E+03, 3.6027E+03, 3.6055E+03, 3.6084E+03, 3.6113E+03,
     + 3.6141E+03, 3.6170E+03, 3.6199E+03, 3.6228E+03, 3.6256E+03,
     +  3.6285E+03, 3.6314E+03, 3.6343E+03, 3.6372E+03, 3.6401E+03,
     + 3.6429E+03, 3.6458E+03, 3.6487E+03, 3.6516E+03, 3.6545E+03,
     +  3.6574E+03, 3.6603E+03, 3.6632E+03, 3.6661E+03, 3.6690E+03,
     + 3.6719E+03, 3.6748E+03, 3.6777E+03, 3.6806E+03, 3.6835E+03,
     +  3.6864E+03, 3.6893E+03, 3.6922E+03, 3.6951E+03, 3.6980E+03,
     + 3.7009E+03, 3.7038E+03, 3.7067E+03, 3.7097E+03, 3.7126E+03,
     +  3.7155E+03, 3.7184E+03, 3.7213E+03, 3.7242E+03, 3.7272E+03,
     + 3.7301E+03, 3.7330E+03, 3.7359E+03, 3.7389E+03, 3.7418E+03,
     +  3.7447E+03, 3.7476E+03, 3.7506E+03, 3.7535E+03, 3.7564E+03,
     + 3.7594E+03, 3.7623E+03, 3.7652E+03, 3.7682E+03, 3.7711E+03,
     +  3.7741E+03, 3.7770E+03, 3.7799E+03, 3.7829E+03, 3.7858E+03,
     + 3.7888E+03, 3.7917E+03, 3.7947E+03, 3.7976E+03, 3.8006E+03,
     +  3.8035E+03, 3.8065E+03, 3.8094E+03, 3.8124E+03, 3.8154E+03,
     + 3.8183E+03, 3.8213E+03, 3.8242E+03, 3.8272E+03, 3.8302E+03,
     +  3.8331E+03, 3.8361E+03, 3.8391E+03, 3.8420E+03, 3.8450E+03,
     + 3.8480E+03, 3.8509E+03, 3.8539E+03, 3.8569E+03, 3.8599E+03,
     +  3.8628E+03, 3.8658E+03, 3.8688E+03, 3.8718E+03, 3.8748E+03,
     + 3.8777E+03, 3.8807E+03, 3.8837E+03, 3.8867E+03, 3.8897E+03,
     +  3.8927E+03, 3.8957E+03, 3.8987E+03, 3.9016E+03, 3.9046E+03,
     + 3.9076E+03, 3.9106E+03, 3.9136E+03, 3.9166E+03, 3.9196E+03,
     +  3.9226E+03, 3.9256E+03, 3.9286E+03, 3.9316E+03, 3.9346E+03,
     + 3.9377E+03, 3.9407E+03, 3.9437E+03, 3.9467E+03, 3.9497E+03,
     +  3.9527E+03, 3.9557E+03, 3.9587E+03, 3.9618E+03, 3.9648E+03,
     + 3.9678E+03, 3.9708E+03, 3.9738E+03, 3.9769E+03, 3.9799E+03,
     +  3.9829E+03, 3.9859E+03, 3.9890E+03, 3.9920E+03, 3.9950E+03,
     + 3.9981E+03, 4.0011E+03, 4.0041E+03, 4.0072E+03, 4.0102E+03,
     +  4.0132E+03, 4.0163E+03, 4.0193E+03, 4.0223E+03, 4.0254E+03,
     + 4.0284E+03, 4.0315E+03, 4.0345E+03, 4.0376E+03, 4.0406E+03,
     +  4.0437E+03, 4.0467E+03, 4.0498E+03, 4.0528E+03, 4.0559E+03,
     + 4.0589E+03, 4.0620E+03, 4.0651E+03, 4.0681E+03, 4.0712E+03,
     +  4.0742E+03, 4.0773E+03, 4.0804E+03, 4.0834E+03, 4.0865E+03,
     + 4.0896E+03, 4.0926E+03, 4.0957E+03, 4.0988E+03, 4.1019E+03,
     +  4.1049E+03, 4.1080E+03, 4.1111E+03, 4.1142E+03, 4.1172E+03,
     + 4.1203E+03, 4.1234E+03, 4.1265E+03, 4.1296E+03, 4.1326E+03,
     +  4.1357E+03, 4.1388E+03, 4.1419E+03, 4.1450E+03, 4.1481E+03,
     + 4.1512E+03, 4.1543E+03, 4.1574E+03, 4.1605E+03, 4.1636E+03,
     +  4.1667E+03, 4.1698E+03, 4.1729E+03, 4.1760E+03, 4.1791E+03,
     + 4.1822E+03, 4.1853E+03, 4.1884E+03, 4.1915E+03, 4.1946E+03,
     +  4.1977E+03, 4.2008E+03, 4.2039E+03, 4.2071E+03, 4.2102E+03,
     + 4.2133E+03, 4.2164E+03, 4.2195E+03, 4.2227E+03, 4.2258E+03,
     +  4.2289E+03, 4.2320E+03, 4.2351E+03, 4.2383E+03, 4.2414E+03/
      data (QofT( 7,J),J=2409,4510)/ 4.2445E+03, 4.2477E+03,
     +  4.2508E+03, 4.2539E+03, 4.2571E+03, 4.2602E+03, 4.2633E+03,
     +  4.2665E+03,
     + 4.2696E+03, 4.2728E+03, 4.2759E+03, 4.2790E+03, 4.2822E+03,
     +  4.2853E+03, 4.2885E+03, 4.2916E+03, 4.2948E+03, 4.2979E+03,
     + 4.3011E+03, 4.3042E+03, 4.3074E+03, 4.3105E+03, 4.3137E+03,
     +  4.3168E+03, 4.3200E+03, 4.3232E+03, 4.3263E+03, 4.3295E+03,
     + 4.3327E+03, 4.3358E+03, 4.3390E+03, 4.3421E+03, 4.3453E+03,
     +  4.3485E+03, 4.3517E+03, 4.3548E+03, 4.3580E+03, 4.3612E+03,
     + 4.3644E+03, 4.3675E+03, 4.3707E+03, 4.3739E+03, 4.3771E+03,
     +  4.3803E+03, 4.3834E+03, 4.3866E+03, 4.3898E+03, 4.3930E+03,
     + 4.3962E+03, 4.3994E+03, 4.4026E+03, 4.4058E+03, 4.4089E+03,
     +  4.4121E+03, 4.4153E+03, 4.4185E+03, 4.4217E+03, 4.4249E+03,
     + 4.4281E+03, 4.4313E+03, 4.4345E+03, 4.4377E+03, 4.4409E+03,
     +  4.4442E+03, 4.4474E+03, 4.4506E+03, 4.4538E+03, 4.4570E+03,
     + 4.4602E+03, 4.4634E+03, 4.4666E+03, 4.4698E+03, 4.4731E+03,
     +  4.4763E+03, 4.4795E+03, 4.4827E+03, 4.4860E+03, 4.4892E+03,
     + 4.4924E+03, 4.4956E+03, 4.4989E+03, 4.5021E+03, 4.5053E+03,
     +  4.5085E+03, 4.5118E+03, 4.5150E+03, 4.5182E+03, 4.5215E+03,
     + 4.5247E+03, 4.5280E+03, 4.5312E+03, 4.5344E+03, 4.5377E+03,
     +  4.5409E+03, 4.5442E+03, 4.5474E+03, 4.5507E+03, 4.5539E+03,
     + 4.5572E+03, 4.5604E+03, 4.5637E+03, 4.5669E+03, 4.5702E+03,
     +  4.5734E+03, 4.5767E+03, 4.5799E+03, 4.5832E+03, 4.5865E+03,
     + 4.5897E+03, 4.5930E+03, 4.5963E+03, 4.5995E+03, 4.6028E+03,
     +  4.6061E+03, 4.6093E+03, 4.6126E+03, 4.6159E+03, 4.6191E+03,
     + 4.6224E+03, 4.6257E+03, 4.6290E+03, 4.6322E+03, 4.6355E+03,
     +  4.6388E+03, 4.6421E+03, 4.6454E+03, 4.6487E+03, 4.6519E+03,
     + 4.6552E+03, 4.6585E+03, 4.6618E+03, 4.6651E+03, 4.6684E+03,
     +  4.6717E+03, 4.6750E+03, 4.6783E+03, 4.6816E+03, 4.6849E+03,
     + 4.6882E+03, 4.6915E+03, 4.6948E+03, 4.6981E+03, 4.7014E+03,
     +  4.7047E+03, 4.7080E+03, 4.7113E+03, 4.7146E+03, 4.7179E+03,
     + 4.7212E+03, 4.7245E+03, 4.7278E+03, 4.7312E+03, 4.7345E+03,
     +  4.7378E+03, 4.7411E+03, 4.7444E+03, 4.7478E+03, 4.7511E+03,
     + 4.7544E+03, 4.7577E+03, 4.7611E+03, 4.7644E+03, 4.7677E+03,
     +  4.7710E+03, 4.7744E+03, 4.7777E+03, 4.7810E+03, 4.7844E+03,
     + 4.7877E+03, 4.7910E+03, 4.7944E+03, 4.7977E+03, 4.8011E+03,
     +  4.8044E+03, 4.8078E+03, 4.8111E+03, 4.8144E+03, 4.8178E+03,
     + 4.8211E+03, 4.8245E+03, 4.8278E+03, 4.8312E+03, 4.8345E+03,
     +  4.8379E+03, 4.8413E+03, 4.8446E+03, 4.8480E+03, 4.8513E+03,
     + 4.8547E+03, 4.8581E+03, 4.8614E+03, 4.8648E+03, 4.8681E+03,
     +  4.8715E+03, 4.8749E+03, 4.8783E+03, 4.8816E+03, 4.8850E+03,
     + 4.8884E+03, 4.8917E+03, 4.8951E+03, 4.8985E+03, 4.9019E+03,
     +  4.9053E+03, 4.9086E+03, 4.9120E+03, 4.9154E+03, 4.9188E+03,
     + 4.9222E+03, 4.9256E+03, 4.9289E+03, 4.9323E+03, 4.9357E+03,
     +  4.9391E+03, 4.9425E+03, 4.9459E+03, 4.9493E+03, 4.9527E+03,
     + 4.9561E+03, 4.9595E+03, 4.9629E+03, 4.9663E+03, 4.9697E+03,
     +  4.9731E+03, 4.9765E+03, 4.9799E+03, 4.9833E+03, 4.9867E+03,
     + 4.9901E+03, 4.9936E+03, 4.9970E+03, 5.0004E+03, 5.0038E+03,
     +  5.0072E+03, 5.0106E+03, 5.0141E+03, 5.0175E+03, 5.0209E+03,
     + 5.0243E+03, 5.0277E+03, 5.0312E+03, 5.0346E+03, 5.0380E+03,
     +  5.0415E+03, 5.0449E+03, 5.0483E+03, 5.0517E+03, 5.0552E+03,
     + 5.0586E+03, 5.0621E+03, 5.0655E+03, 5.0689E+03, 5.0724E+03,
     +  5.0758E+03, 5.0793E+03, 5.0827E+03, 5.0861E+03, 5.0896E+03,
     + 5.0930E+03, 5.0965E+03, 5.0999E+03, 5.1034E+03, 5.1068E+03,
     +  5.1103E+03, 5.1138E+03, 5.1172E+03, 5.1207E+03, 5.1241E+03,
     + 5.1276E+03, 5.1310E+03, 5.1345E+03, 5.1380E+03, 5.1414E+03,
     +  5.1449E+03, 5.1484E+03, 5.1518E+03, 5.1553E+03, 5.1588E+03,
     + 5.1623E+03, 5.1657E+03, 5.1692E+03, 5.1727E+03, 5.1762E+03,
     +  5.1796E+03, 5.1831E+03, 5.1866E+03, 5.1901E+03, 5.1936E+03,
     + 5.1971E+03, 5.2005E+03, 5.2040E+03, 5.2075E+03, 5.2110E+03,
     +  5.2145E+03, 5.2180E+03, 5.2215E+03, 5.2250E+03, 5.2285E+03,
     + 5.2320E+03, 5.2355E+03, 5.2390E+03, 5.2425E+03, 5.2460E+03,
     +  5.2495E+03, 5.2530E+03, 5.2565E+03, 5.2600E+03, 5.2635E+03,
     + 5.2670E+03, 5.2705E+03, 5.2740E+03, 5.2776E+03, 5.2811E+03,
     +  5.2846E+03, 5.2881E+03, 5.2916E+03, 5.2951E+03, 5.2987E+03,
     + 5.3022E+03, 5.3057E+03, 5.3092E+03, 5.3128E+03, 5.3163E+03,
     +  5.3198E+03, 5.3234E+03, 5.3269E+03, 5.3304E+03, 5.3339E+03,
     + 5.3375E+03, 5.3410E+03, 5.3446E+03, 5.3481E+03, 5.3516E+03,
     +  5.3552E+03, 5.3587E+03, 5.3623E+03, 5.3658E+03, 5.3694E+03,
     + 5.3729E+03, 5.3765E+03, 5.3800E+03, 5.3836E+03, 5.3871E+03,
     +  5.3907E+03, 5.3942E+03, 5.3978E+03, 5.4013E+03, 5.4049E+03,
     + 5.4084E+03, 5.4120E+03, 5.4156E+03, 5.4191E+03, 5.4227E+03,
     +  5.4263E+03, 5.4298E+03, 5.4334E+03, 5.4370E+03, 5.4405E+03,
     + 5.4441E+03, 5.4477E+03, 5.4513E+03, 5.4548E+03, 5.4584E+03,
     +  5.4620E+03, 5.4656E+03, 5.4692E+03, 5.4727E+03, 5.4763E+03,
     + 5.4799E+03, 5.4835E+03, 5.4871E+03, 5.4907E+03, 5.4943E+03,
     +  5.4979E+03, 5.5014E+03, 5.5050E+03, 5.5086E+03, 5.5122E+03,
     + 5.5158E+03, 5.5194E+03, 5.5230E+03, 5.5266E+03, 5.5302E+03,
     +  5.5338E+03, 5.5374E+03, 5.5410E+03, 5.5447E+03, 5.5483E+03,
     + 5.5519E+03, 5.5555E+03, 5.5591E+03, 5.5627E+03, 5.5663E+03,
     +  5.5699E+03, 5.5736E+03, 5.5772E+03, 5.5808E+03, 5.5844E+03,
     + 5.5880E+03, 5.5917E+03, 5.5953E+03, 5.5989E+03, 5.6025E+03,
     +  5.6062E+03, 5.6098E+03, 5.6134E+03, 5.6171E+03, 5.6207E+03,
     + 5.6243E+03, 5.6280E+03, 5.6316E+03, 5.6353E+03, 5.6389E+03,
     +  5.6425E+03, 5.6462E+03, 5.6498E+03, 5.6535E+03, 5.6571E+03,
     + 5.6608E+03, 5.6644E+03, 5.6681E+03, 5.6717E+03, 5.6754E+03,
     +  5.6790E+03, 5.6827E+03, 5.6863E+03, 5.6900E+03, 5.6936E+03,
     + 5.6973E+03, 5.7010E+03, 5.7046E+03, 5.7083E+03, 5.7120E+03,
     +  5.7156E+03, 5.7193E+03, 5.7230E+03, 5.7266E+03, 5.7303E+03,
     + 5.7340E+03, 5.7377E+03, 5.7413E+03, 5.7450E+03, 5.7487E+03,
     +  5.7524E+03, 5.7560E+03, 5.7597E+03, 5.7634E+03, 5.7671E+03,
     + 5.7708E+03, 5.7745E+03, 5.7782E+03, 5.7818E+03, 5.7855E+03,
     +  5.7892E+03, 5.7929E+03, 5.7966E+03, 5.8003E+03, 5.8040E+03,
     + 5.8077E+03, 5.8114E+03, 5.8151E+03, 5.8188E+03, 5.8225E+03,
     +  5.8262E+03, 5.8299E+03, 5.8336E+03, 5.8373E+03, 5.8410E+03,
     + 5.8448E+03, 5.8485E+03, 5.8522E+03, 5.8559E+03, 5.8596E+03,
     +  5.8633E+03, 5.8670E+03, 5.8708E+03, 5.8745E+03, 5.8782E+03,
     + 5.8819E+03, 5.8856E+03, 5.8894E+03, 5.8931E+03, 5.8968E+03,
     +  5.9006E+03, 5.9043E+03, 5.9080E+03, 5.9118E+03, 5.9155E+03,
     + 5.9192E+03, 5.9230E+03, 5.9267E+03, 5.9304E+03, 5.9342E+03,
     +  5.9379E+03, 5.9417E+03, 5.9454E+03, 5.9492E+03, 5.9529E+03,
     + 5.9567E+03, 5.9604E+03, 5.9642E+03, 5.9679E+03, 5.9717E+03,
     +  5.9754E+03, 5.9792E+03, 5.9829E+03, 5.9867E+03, 5.9904E+03,
     + 5.9942E+03, 5.9980E+03, 6.0017E+03, 6.0055E+03, 6.0093E+03,
     +  6.0130E+03, 6.0168E+03, 6.0206E+03, 6.0243E+03, 6.0281E+03,
     + 6.0319E+03, 6.0357E+03, 6.0394E+03, 6.0432E+03, 6.0470E+03,
     +  6.0508E+03, 6.0546E+03, 6.0583E+03, 6.0621E+03, 6.0659E+03,
     + 6.0697E+03, 6.0735E+03, 6.0773E+03, 6.0811E+03, 6.0848E+03,
     +  6.0886E+03, 6.0924E+03, 6.0962E+03, 6.1000E+03, 6.1038E+03,
     + 6.1076E+03, 6.1114E+03, 6.1152E+03, 6.1190E+03, 6.1228E+03,
     +  6.1266E+03, 6.1304E+03, 6.1342E+03, 6.1381E+03, 6.1419E+03,
     + 6.1457E+03, 6.1495E+03, 6.1533E+03, 6.1571E+03, 6.1609E+03,
     +  6.1648E+03, 6.1686E+03, 6.1724E+03, 6.1762E+03, 6.1800E+03,
     + 6.1839E+03, 6.1877E+03, 6.1915E+03, 6.1953E+03, 6.1992E+03,
     +  6.2030E+03, 6.2068E+03, 6.2107E+03, 6.2145E+03, 6.2183E+03,
     + 6.2222E+03, 6.2260E+03, 6.2298E+03, 6.2337E+03, 6.2375E+03,
     +  6.2414E+03, 6.2452E+03, 6.2491E+03, 6.2529E+03, 6.2568E+03,
     + 6.2606E+03, 6.2645E+03, 6.2683E+03, 6.2722E+03, 6.2760E+03,
     +  6.2799E+03, 6.2837E+03, 6.2876E+03, 6.2914E+03, 6.2953E+03,
     + 6.2992E+03, 6.3030E+03, 6.3069E+03, 6.3108E+03, 6.3146E+03,
     +  6.3185E+03, 6.3224E+03, 6.3262E+03, 6.3301E+03, 6.3340E+03,
     + 6.3379E+03, 6.3417E+03, 6.3456E+03, 6.3495E+03, 6.3534E+03,
     +  6.3573E+03, 6.3611E+03, 6.3650E+03, 6.3689E+03, 6.3728E+03,
     + 6.3767E+03, 6.3806E+03, 6.3845E+03, 6.3883E+03, 6.3922E+03,
     +  6.3961E+03, 6.4000E+03, 6.4039E+03, 6.4078E+03, 6.4117E+03,
     + 6.4156E+03, 6.4195E+03, 6.4234E+03, 6.4273E+03, 6.4312E+03,
     +  6.4351E+03, 6.4390E+03, 6.4430E+03, 6.4469E+03, 6.4508E+03,
     + 6.4547E+03, 6.4586E+03, 6.4625E+03, 6.4664E+03, 6.4704E+03,
     +  6.4743E+03, 6.4782E+03, 6.4821E+03, 6.4860E+03, 6.4900E+03,
     + 6.4939E+03, 6.4978E+03, 6.5017E+03, 6.5057E+03, 6.5096E+03,
     +  6.5135E+03, 6.5175E+03, 6.5214E+03, 6.5253E+03, 6.5293E+03,
     + 6.5332E+03, 6.5371E+03, 6.5411E+03, 6.5450E+03, 6.5490E+03,
     +  6.5529E+03, 6.5569E+03, 6.5608E+03, 6.5648E+03, 6.5687E+03,
     + 6.5727E+03, 6.5766E+03, 6.5806E+03, 6.5845E+03, 6.5885E+03,
     +  6.5924E+03, 6.5964E+03, 6.6003E+03, 6.6043E+03, 6.6083E+03,
     + 6.6122E+03, 6.6162E+03, 6.6202E+03, 6.6241E+03, 6.6281E+03,
     +  6.6321E+03, 6.6360E+03, 6.6400E+03, 6.6440E+03, 6.6480E+03,
     + 6.6519E+03, 6.6559E+03, 6.6599E+03, 6.6639E+03, 6.6679E+03,
     +  6.6718E+03, 6.6758E+03, 6.6798E+03, 6.6838E+03, 6.6878E+03,
     + 6.6918E+03, 6.6958E+03, 6.6998E+03, 6.7037E+03, 6.7077E+03,
     +  6.7117E+03, 6.7157E+03, 6.7197E+03, 6.7237E+03, 6.7277E+03,
     + 6.7317E+03, 6.7357E+03, 6.7397E+03, 6.7437E+03, 6.7477E+03,
     +  6.7518E+03, 6.7558E+03, 6.7598E+03, 6.7638E+03, 6.7678E+03,
     + 6.7718E+03, 6.7758E+03, 6.7798E+03, 6.7839E+03, 6.7879E+03,
     +  6.7919E+03, 6.7959E+03, 6.8000E+03, 6.8040E+03, 6.8080E+03,
     + 6.8120E+03, 6.8161E+03, 6.8201E+03, 6.8241E+03, 6.8281E+03,
     +  6.8322E+03, 6.8362E+03, 6.8403E+03, 6.8443E+03, 6.8483E+03,
     + 6.8524E+03, 6.8564E+03, 6.8605E+03, 6.8645E+03, 6.8685E+03,
     +  6.8726E+03, 6.8766E+03, 6.8807E+03, 6.8847E+03, 6.8888E+03,
     + 6.8928E+03, 6.8969E+03, 6.9009E+03, 6.9050E+03, 6.9091E+03,
     +  6.9131E+03, 6.9172E+03, 6.9212E+03, 6.9253E+03, 6.9294E+03,
     + 6.9334E+03, 6.9375E+03, 6.9416E+03, 6.9456E+03, 6.9497E+03,
     +  6.9538E+03, 6.9578E+03, 6.9619E+03, 6.9660E+03, 6.9701E+03,
     + 6.9742E+03, 6.9782E+03, 6.9823E+03, 6.9864E+03, 6.9905E+03,
     +  6.9946E+03, 6.9986E+03, 7.0027E+03, 7.0068E+03, 7.0109E+03,
     + 7.0150E+03, 7.0191E+03, 7.0232E+03, 7.0273E+03, 7.0314E+03,
     +  7.0355E+03, 7.0396E+03, 7.0437E+03, 7.0478E+03, 7.0519E+03,
     + 7.0560E+03, 7.0601E+03, 7.0642E+03, 7.0683E+03, 7.0724E+03,
     +  7.0765E+03, 7.0806E+03, 7.0847E+03, 7.0889E+03, 7.0930E+03,
     + 7.0971E+03, 7.1012E+03, 7.1053E+03, 7.1094E+03, 7.1136E+03,
     +  7.1177E+03, 7.1218E+03, 7.1259E+03, 7.1301E+03, 7.1342E+03,
     + 7.1383E+03, 7.1425E+03, 7.1466E+03, 7.1507E+03, 7.1549E+03,
     +  7.1590E+03, 7.1631E+03, 7.1673E+03, 7.1714E+03, 7.1755E+03,
     + 7.1797E+03, 7.1838E+03, 7.1880E+03, 7.1921E+03, 7.1963E+03,
     +  7.2004E+03, 7.2046E+03, 7.2087E+03, 7.2129E+03, 7.2170E+03,
     + 7.2212E+03, 7.2253E+03, 7.2295E+03, 7.2336E+03, 7.2378E+03,
     +  7.2420E+03, 7.2461E+03, 7.2503E+03, 7.2545E+03, 7.2586E+03,
     + 7.2628E+03, 7.2670E+03, 7.2711E+03, 7.2753E+03, 7.2795E+03,
     +  7.2836E+03, 7.2878E+03, 7.2920E+03, 7.2962E+03, 7.3004E+03,
     + 7.3045E+03, 7.3087E+03, 7.3129E+03, 7.3171E+03, 7.3213E+03,
     +  7.3255E+03, 7.3296E+03, 7.3338E+03, 7.3380E+03, 7.3422E+03,
     + 7.3464E+03, 7.3506E+03, 7.3548E+03, 7.3590E+03, 7.3632E+03,
     +  7.3674E+03, 7.3716E+03, 7.3758E+03, 7.3800E+03, 7.3842E+03,
     + 7.3884E+03, 7.3926E+03, 7.3968E+03, 7.4010E+03, 7.4052E+03,
     +  7.4095E+03, 7.4137E+03, 7.4179E+03, 7.4221E+03, 7.4263E+03,
     + 7.4305E+03, 7.4348E+03, 7.4390E+03, 7.4432E+03, 7.4474E+03,
     +  7.4516E+03, 7.4559E+03, 7.4601E+03, 7.4643E+03, 7.4686E+03,
     + 7.4728E+03, 7.4770E+03, 7.4813E+03, 7.4855E+03, 7.4897E+03,
     +  7.4940E+03, 7.4982E+03, 7.5024E+03, 7.5067E+03, 7.5109E+03,
     + 7.5152E+03, 7.5194E+03, 7.5237E+03, 7.5279E+03, 7.5322E+03,
     +  7.5364E+03, 7.5407E+03, 7.5449E+03, 7.5492E+03, 7.5534E+03,
     + 7.5577E+03, 7.5619E+03, 7.5662E+03, 7.5705E+03, 7.5747E+03,
     +  7.5790E+03, 7.5833E+03, 7.5875E+03, 7.5918E+03, 7.5961E+03,
     + 7.6003E+03, 7.6046E+03, 7.6089E+03, 7.6132E+03, 7.6174E+03,
     +  7.6217E+03, 7.6260E+03, 7.6303E+03, 7.6345E+03, 7.6388E+03,
     + 7.6431E+03, 7.6474E+03, 7.6517E+03, 7.6560E+03, 7.6602E+03,
     +  7.6645E+03, 7.6688E+03, 7.6731E+03, 7.6774E+03, 7.6817E+03,
     + 7.6860E+03, 7.6903E+03, 7.6946E+03, 7.6989E+03, 7.7032E+03,
     +  7.7075E+03, 7.7118E+03, 7.7161E+03, 7.7204E+03, 7.7247E+03,
     + 7.7290E+03, 7.7333E+03, 7.7376E+03, 7.7420E+03, 7.7463E+03,
     +  7.7506E+03, 7.7549E+03, 7.7592E+03, 7.7635E+03, 7.7679E+03,
     + 7.7722E+03, 7.7765E+03, 7.7808E+03, 7.7852E+03, 7.7895E+03,
     +  7.7938E+03, 7.7981E+03, 7.8025E+03, 7.8068E+03, 7.8111E+03,
     + 7.8155E+03, 7.8198E+03, 7.8241E+03, 7.8285E+03, 7.8328E+03,
     +  7.8372E+03, 7.8415E+03, 7.8458E+03, 7.8502E+03, 7.8545E+03,
     + 7.8589E+03, 7.8632E+03, 7.8676E+03, 7.8719E+03, 7.8763E+03,
     +  7.8806E+03, 7.8850E+03, 7.8893E+03, 7.8937E+03, 7.8981E+03,
     + 7.9024E+03, 7.9068E+03, 7.9111E+03, 7.9155E+03, 7.9199E+03,
     +  7.9242E+03, 7.9286E+03, 7.9330E+03, 7.9373E+03, 7.9417E+03,
     + 7.9461E+03, 7.9505E+03, 7.9548E+03, 7.9592E+03, 7.9636E+03,
     +  7.9680E+03, 7.9724E+03, 7.9767E+03, 7.9811E+03, 7.9855E+03,
     + 7.9899E+03, 7.9943E+03, 7.9987E+03, 8.0031E+03, 8.0074E+03,
     +  8.0118E+03, 8.0162E+03, 8.0206E+03, 8.0250E+03, 8.0294E+03,
     + 8.0338E+03, 8.0382E+03, 8.0426E+03, 8.0470E+03, 8.0514E+03,
     +  8.0558E+03, 8.0602E+03, 8.0646E+03, 8.0691E+03, 8.0735E+03,
     + 8.0779E+03, 8.0823E+03, 8.0867E+03, 8.0911E+03, 8.0955E+03,
     +  8.0999E+03, 8.1044E+03, 8.1088E+03, 8.1132E+03, 8.1176E+03,
     + 8.1221E+03, 8.1265E+03, 8.1309E+03, 8.1353E+03, 8.1398E+03,
     +  8.1442E+03, 8.1486E+03, 8.1531E+03, 8.1575E+03, 8.1619E+03,
     + 8.1664E+03, 8.1708E+03, 8.1753E+03, 8.1797E+03, 8.1841E+03,
     +  8.1886E+03, 8.1930E+03, 8.1975E+03, 8.2019E+03, 8.2064E+03,
     + 8.2108E+03, 8.2153E+03, 8.2197E+03, 8.2242E+03, 8.2286E+03,
     +  8.2331E+03, 8.2375E+03, 8.2420E+03, 8.2465E+03, 8.2509E+03,
     + 8.2554E+03, 8.2599E+03, 8.2643E+03, 8.2688E+03, 8.2733E+03,
     +  8.2777E+03, 8.2822E+03, 8.2867E+03, 8.2911E+03, 8.2956E+03,
     + 8.3001E+03, 8.3046E+03, 8.3090E+03, 8.3135E+03, 8.3180E+03,
     +  8.3225E+03, 8.3270E+03, 8.3315E+03, 8.3359E+03, 8.3404E+03,
     + 8.3449E+03, 8.3494E+03, 8.3539E+03, 8.3584E+03, 8.3629E+03,
     +  8.3674E+03, 8.3719E+03, 8.3764E+03, 8.3809E+03, 8.3854E+03,
     + 8.3899E+03, 8.3944E+03, 8.3989E+03, 8.4034E+03, 8.4079E+03,
     +  8.4124E+03, 8.4169E+03, 8.4214E+03, 8.4259E+03, 8.4305E+03,
     + 8.4350E+03, 8.4395E+03, 8.4440E+03, 8.4485E+03, 8.4530E+03,
     +  8.4576E+03, 8.4621E+03, 8.4666E+03, 8.4711E+03, 8.4757E+03,
     + 8.4802E+03, 8.4847E+03, 8.4892E+03, 8.4938E+03, 8.4983E+03,
     +  8.5028E+03, 8.5074E+03, 8.5119E+03, 8.5164E+03, 8.5210E+03,
     + 8.5255E+03, 8.5301E+03, 8.5346E+03, 8.5392E+03, 8.5437E+03,
     +  8.5483E+03, 8.5528E+03, 8.5573E+03, 8.5619E+03, 8.5665E+03,
     + 8.5710E+03, 8.5756E+03, 8.5801E+03, 8.5847E+03, 8.5892E+03,
     +  8.5938E+03, 8.5984E+03, 8.6029E+03, 8.6075E+03, 8.6120E+03,
     + 8.6166E+03, 8.6212E+03, 8.6257E+03, 8.6303E+03, 8.6349E+03,
     +  8.6395E+03, 8.6440E+03, 8.6486E+03, 8.6532E+03, 8.6578E+03,
     + 8.6623E+03, 8.6669E+03, 8.6715E+03, 8.6761E+03, 8.6807E+03,
     +  8.6853E+03, 8.6898E+03, 8.6944E+03, 8.6990E+03, 8.7036E+03,
     + 8.7082E+03, 8.7128E+03, 8.7174E+03, 8.7220E+03, 8.7266E+03,
     +  8.7312E+03, 8.7358E+03, 8.7404E+03, 8.7450E+03, 8.7496E+03,
     + 8.7542E+03, 8.7588E+03, 8.7634E+03, 8.7680E+03, 8.7726E+03,
     +  8.7772E+03, 8.7819E+03, 8.7865E+03, 8.7911E+03, 8.7957E+03,
     + 8.8003E+03, 8.8049E+03, 8.8096E+03, 8.8142E+03, 8.8188E+03,
     +  8.8234E+03, 8.8281E+03, 8.8327E+03, 8.8373E+03, 8.8419E+03,
     + 8.8466E+03, 8.8512E+03, 8.8558E+03, 8.8605E+03, 8.8651E+03,
     +  8.8698E+03, 8.8744E+03, 8.8790E+03, 8.8837E+03, 8.8883E+03,
     + 8.8930E+03, 8.8976E+03, 8.9022E+03, 8.9069E+03, 8.9115E+03,
     +  8.9162E+03, 8.9208E+03, 8.9255E+03, 8.9302E+03, 8.9348E+03,
     + 8.9395E+03, 8.9441E+03, 8.9488E+03, 8.9534E+03, 8.9581E+03,
     +  8.9628E+03, 8.9674E+03, 8.9721E+03, 8.9768E+03, 8.9814E+03,
     + 8.9861E+03, 8.9908E+03, 8.9955E+03, 9.0001E+03, 9.0048E+03,
     +  9.0095E+03, 9.0142E+03, 9.0188E+03, 9.0235E+03, 9.0282E+03,
     + 9.0329E+03, 9.0376E+03, 9.0423E+03, 9.0469E+03, 9.0516E+03,
     +  9.0563E+03, 9.0610E+03, 9.0657E+03, 9.0704E+03, 9.0751E+03,
     + 9.0798E+03, 9.0845E+03, 9.0892E+03, 9.0939E+03, 9.0986E+03,
     +  9.1033E+03, 9.1080E+03, 9.1127E+03, 9.1174E+03, 9.1221E+03,
     + 9.1268E+03, 9.1315E+03, 9.1362E+03, 9.1409E+03, 9.1457E+03,
     +  9.1504E+03, 9.1551E+03, 9.1598E+03, 9.1645E+03, 9.1692E+03,
     + 9.1740E+03, 9.1787E+03, 9.1834E+03, 9.1881E+03, 9.1929E+03,
     +  9.1976E+03, 9.2023E+03, 9.2071E+03, 9.2118E+03, 9.2165E+03,
     + 9.2213E+03, 9.2260E+03, 9.2307E+03, 9.2355E+03, 9.2402E+03,
     +  9.2450E+03, 9.2497E+03, 9.2544E+03, 9.2592E+03, 9.2639E+03,
     + 9.2687E+03, 9.2734E+03, 9.2782E+03, 9.2829E+03, 9.2877E+03,
     +  9.2924E+03, 9.2972E+03, 9.3020E+03, 9.3067E+03, 9.3115E+03,
     + 9.3162E+03, 9.3210E+03, 9.3258E+03, 9.3305E+03, 9.3353E+03,
     +  9.3401E+03, 9.3448E+03, 9.3496E+03, 9.3544E+03, 9.3591E+03,
     + 9.3639E+03, 9.3687E+03, 9.3735E+03, 9.3782E+03, 9.3830E+03,
     +  9.3878E+03, 9.3926E+03, 9.3974E+03, 9.4021E+03, 9.4069E+03,
     + 9.4117E+03, 9.4165E+03, 9.4213E+03, 9.4261E+03, 9.4309E+03,
     +  9.4357E+03, 9.4405E+03, 9.4453E+03, 9.4501E+03, 9.4549E+03,
     + 9.4597E+03, 9.4645E+03, 9.4693E+03, 9.4741E+03, 9.4789E+03,
     +  9.4837E+03, 9.4885E+03, 9.4933E+03, 9.4981E+03, 9.5029E+03,
     + 9.5077E+03, 9.5125E+03, 9.5174E+03, 9.5222E+03, 9.5270E+03,
     +  9.5318E+03, 9.5366E+03, 9.5415E+03, 9.5463E+03, 9.5511E+03,
     + 9.5559E+03, 9.5608E+03, 9.5656E+03, 9.5704E+03, 9.5752E+03,
     +  9.5801E+03, 9.5849E+03, 9.5897E+03, 9.5946E+03, 9.5994E+03,
     + 9.6043E+03, 9.6091E+03, 9.6139E+03, 9.6188E+03, 9.6236E+03,
     +  9.6285E+03, 9.6333E+03, 9.6382E+03, 9.6430E+03, 9.6479E+03,
     + 9.6527E+03, 9.6576E+03, 9.6624E+03, 9.6673E+03, 9.6721E+03,
     +  9.6770E+03, 9.6818E+03, 9.6867E+03, 9.6916E+03, 9.6964E+03,
     + 9.7013E+03, 9.7062E+03, 9.7110E+03, 9.7159E+03, 9.7208E+03,
     +  9.7256E+03, 9.7305E+03, 9.7354E+03, 9.7403E+03, 9.7451E+03,
     + 9.7500E+03, 9.7549E+03, 9.7598E+03, 9.7647E+03, 9.7695E+03,
     +  9.7744E+03, 9.7793E+03, 9.7842E+03, 9.7891E+03, 9.7940E+03,
     + 9.7989E+03, 9.8038E+03, 9.8087E+03, 9.8135E+03, 9.8184E+03,
     +  9.8233E+03, 9.8282E+03, 9.8331E+03, 9.8380E+03, 9.8429E+03,
     + 9.8478E+03, 9.8528E+03, 9.8577E+03, 9.8626E+03, 9.8675E+03,
     +  9.8724E+03, 9.8773E+03, 9.8822E+03, 9.8871E+03, 9.8920E+03,
     + 9.8970E+03, 9.9019E+03, 9.9068E+03, 9.9117E+03, 9.9166E+03,
     +  9.9216E+03, 9.9265E+03, 9.9314E+03, 9.9363E+03, 9.9413E+03,
     + 9.9462E+03, 9.9511E+03, 9.9561E+03, 9.9610E+03, 9.9659E+03,
     +  9.9709E+03, 9.9758E+03, 9.9807E+03, 9.9857E+03, 9.9906E+03,
     + 9.9956E+03, 1.0001E+04, 1.0005E+04, 1.0010E+04, 1.0015E+04,
     +  1.0020E+04, 1.0025E+04, 1.0030E+04, 1.0035E+04, 1.0040E+04,
     + 1.0045E+04, 1.0050E+04, 1.0055E+04, 1.0060E+04, 1.0065E+04,
     +  1.0070E+04, 1.0075E+04, 1.0080E+04, 1.0085E+04, 1.0090E+04,
     + 1.0095E+04, 1.0100E+04, 1.0105E+04, 1.0110E+04, 1.0115E+04,
     +  1.0120E+04, 1.0125E+04, 1.0130E+04, 1.0134E+04, 1.0139E+04,
     + 1.0144E+04, 1.0149E+04, 1.0154E+04, 1.0159E+04, 1.0164E+04,
     +  1.0169E+04, 1.0174E+04, 1.0179E+04, 1.0184E+04, 1.0189E+04,
     + 1.0194E+04, 1.0199E+04, 1.0204E+04, 1.0209E+04, 1.0214E+04,
     +  1.0219E+04, 1.0224E+04, 1.0229E+04, 1.0234E+04, 1.0239E+04,
     + 1.0244E+04, 1.0249E+04, 1.0254E+04, 1.0259E+04, 1.0264E+04,
     +  1.0269E+04, 1.0274E+04, 1.0279E+04, 1.0284E+04, 1.0290E+04,
     + 1.0295E+04, 1.0300E+04, 1.0305E+04, 1.0310E+04, 1.0315E+04,
     +  1.0320E+04, 1.0325E+04, 1.0330E+04, 1.0335E+04, 1.0340E+04,
     + 1.0345E+04, 1.0350E+04, 1.0355E+04, 1.0360E+04, 1.0365E+04,
     +  1.0370E+04, 1.0375E+04, 1.0380E+04, 1.0385E+04, 1.0390E+04,
     + 1.0395E+04, 1.0400E+04, 1.0405E+04, 1.0410E+04, 1.0415E+04,
     +  1.0420E+04, 1.0426E+04, 1.0431E+04, 1.0436E+04, 1.0441E+04,
     + 1.0446E+04, 1.0451E+04, 1.0456E+04, 1.0461E+04, 1.0466E+04,
     +  1.0471E+04, 1.0476E+04, 1.0481E+04, 1.0486E+04, 1.0491E+04,
     + 1.0496E+04, 1.0502E+04, 1.0507E+04, 1.0512E+04, 1.0517E+04,
     +  1.0522E+04, 1.0527E+04, 1.0532E+04, 1.0537E+04, 1.0542E+04,
     + 1.0547E+04, 1.0552E+04, 1.0557E+04, 1.0563E+04, 1.0568E+04,
     +  1.0573E+04, 1.0578E+04, 1.0583E+04, 1.0588E+04, 1.0593E+04,
     + 1.0598E+04, 1.0603E+04, 1.0608E+04, 1.0613E+04, 1.0619E+04,
     +  1.0624E+04, 1.0629E+04, 1.0634E+04, 1.0639E+04, 1.0644E+04,
     + 1.0649E+04, 1.0654E+04, 1.0659E+04, 1.0665E+04, 1.0670E+04,
     +  1.0675E+04, 1.0680E+04, 1.0685E+04, 1.0690E+04, 1.0695E+04,
     + 1.0700E+04, 1.0706E+04, 1.0711E+04, 1.0716E+04, 1.0721E+04,
     +  1.0726E+04, 1.0731E+04, 1.0736E+04, 1.0741E+04, 1.0747E+04,
     + 1.0752E+04, 1.0757E+04, 1.0762E+04, 1.0767E+04, 1.0772E+04,
     +  1.0777E+04, 1.0783E+04, 1.0788E+04, 1.0793E+04, 1.0798E+04,
     + 1.0803E+04, 1.0808E+04, 1.0814E+04, 1.0819E+04, 1.0824E+04,
     +  1.0829E+04, 1.0834E+04, 1.0839E+04, 1.0844E+04, 1.0850E+04,
     + 1.0855E+04, 1.0860E+04, 1.0865E+04, 1.0870E+04, 1.0875E+04,
     +  1.0881E+04, 1.0886E+04, 1.0891E+04, 1.0896E+04, 1.0901E+04,
     + 1.0907E+04, 1.0912E+04, 1.0917E+04, 1.0922E+04, 1.0927E+04,
     +  1.0932E+04, 1.0938E+04, 1.0943E+04, 1.0948E+04, 1.0953E+04,
     + 1.0958E+04, 1.0964E+04, 1.0969E+04, 1.0974E+04, 1.0979E+04,
     +  1.0984E+04, 1.0990E+04, 1.0995E+04, 1.1000E+04, 1.1005E+04,
     + 1.1010E+04, 1.1016E+04, 1.1021E+04, 1.1026E+04, 1.1031E+04,
     +  1.1036E+04, 1.1042E+04, 1.1047E+04, 1.1052E+04, 1.1057E+04,
     + 1.1062E+04, 1.1068E+04, 1.1073E+04, 1.1078E+04, 1.1083E+04,
     +  1.1089E+04, 1.1094E+04, 1.1099E+04, 1.1104E+04, 1.1109E+04,
     + 1.1115E+04, 1.1120E+04, 1.1125E+04, 1.1130E+04, 1.1136E+04,
     +  1.1141E+04, 1.1146E+04, 1.1151E+04, 1.1157E+04, 1.1162E+04,
     + 1.1167E+04, 1.1172E+04, 1.1178E+04, 1.1183E+04, 1.1188E+04,
     +  1.1193E+04, 1.1199E+04, 1.1204E+04, 1.1209E+04, 1.1214E+04,
     + 1.1220E+04, 1.1225E+04, 1.1230E+04, 1.1235E+04, 1.1241E+04,
     +  1.1246E+04, 1.1251E+04, 1.1256E+04, 1.1262E+04, 1.1267E+04,
     + 1.1272E+04, 1.1277E+04, 1.1283E+04, 1.1288E+04, 1.1293E+04,
     +  1.1299E+04, 1.1304E+04, 1.1309E+04, 1.1314E+04, 1.1320E+04,
     + 1.1325E+04, 1.1330E+04, 1.1336E+04, 1.1341E+04, 1.1346E+04,
     +  1.1351E+04, 1.1357E+04, 1.1362E+04, 1.1367E+04, 1.1373E+04,
     + 1.1378E+04, 1.1383E+04, 1.1388E+04, 1.1394E+04, 1.1399E+04,
     +  1.1404E+04, 1.1410E+04, 1.1415E+04, 1.1420E+04, 1.1426E+04,
     + 1.1431E+04, 1.1436E+04, 1.1441E+04, 1.1447E+04, 1.1452E+04,
     +  1.1457E+04, 1.1463E+04, 1.1468E+04, 1.1473E+04, 1.1479E+04,
     + 1.1484E+04, 1.1489E+04, 1.1495E+04, 1.1500E+04, 1.1505E+04,
     +  1.1511E+04, 1.1516E+04, 1.1521E+04, 1.1527E+04, 1.1532E+04,
     + 1.1537E+04, 1.1543E+04, 1.1548E+04, 1.1553E+04, 1.1559E+04,
     +  1.1564E+04, 1.1569E+04, 1.1575E+04, 1.1580E+04, 1.1585E+04,
     + 1.1591E+04, 1.1596E+04, 1.1601E+04, 1.1607E+04, 1.1612E+04,
     +  1.1617E+04, 1.1623E+04, 1.1628E+04, 1.1634E+04, 1.1639E+04,
     + 1.1644E+04, 1.1650E+04, 1.1655E+04, 1.1660E+04, 1.1666E+04,
     +  1.1671E+04, 1.1676E+04, 1.1682E+04, 1.1687E+04, 1.1693E+04,
     + 1.1698E+04, 1.1703E+04, 1.1709E+04, 1.1714E+04, 1.1719E+04,
     +  1.1725E+04, 1.1730E+04, 1.1736E+04, 1.1741E+04, 1.1746E+04,
     + 1.1752E+04, 1.1757E+04, 1.1763E+04, 1.1768E+04, 1.1773E+04,
     +  1.1779E+04, 1.1784E+04, 1.1790E+04, 1.1795E+04, 1.1800E+04,
     + 1.1806E+04, 1.1811E+04, 1.1817E+04, 1.1822E+04, 1.1827E+04,
     +  1.1833E+04, 1.1838E+04, 1.1844E+04, 1.1849E+04, 1.1854E+04,
     + 1.1860E+04, 1.1865E+04, 1.1871E+04, 1.1876E+04, 1.1881E+04,
     +  1.1887E+04, 1.1892E+04, 1.1898E+04, 1.1903E+04, 1.1909E+04,
     + 1.1914E+04, 1.1919E+04, 1.1925E+04, 1.1930E+04, 1.1936E+04,
     +  1.1941E+04, 1.1947E+04, 1.1952E+04, 1.1957E+04, 1.1963E+04,
     + 1.1968E+04, 1.1974E+04, 1.1979E+04, 1.1985E+04, 1.1990E+04,
     +  1.1996E+04, 1.2001E+04, 1.2006E+04, 1.2012E+04, 1.2017E+04,
     + 1.2023E+04, 1.2028E+04, 1.2034E+04, 1.2039E+04, 1.2045E+04,
     +  1.2050E+04, 1.2056E+04, 1.2061E+04, 1.2066E+04, 1.2072E+04,
     + 1.2077E+04, 1.2083E+04, 1.2088E+04, 1.2094E+04, 1.2099E+04,
     +  1.2105E+04, 1.2110E+04, 1.2116E+04, 1.2121E+04, 1.2127E+04,
     + 1.2132E+04, 1.2138E+04, 1.2143E+04, 1.2149E+04, 1.2154E+04,
     +  1.2159E+04, 1.2165E+04, 1.2170E+04, 1.2176E+04, 1.2181E+04,
     + 1.2187E+04, 1.2192E+04, 1.2198E+04, 1.2203E+04, 1.2209E+04,
     +  1.2214E+04, 1.2220E+04, 1.2225E+04, 1.2231E+04, 1.2236E+04,
     + 1.2242E+04, 1.2247E+04, 1.2253E+04, 1.2258E+04, 1.2264E+04,
     +  1.2269E+04, 1.2275E+04, 1.2280E+04, 1.2286E+04, 1.2292E+04,
     + 1.2297E+04, 1.2303E+04, 1.2308E+04, 1.2314E+04, 1.2319E+04,
     +  1.2325E+04, 1.2330E+04, 1.2336E+04, 1.2341E+04, 1.2347E+04,
     + 1.2352E+04, 1.2358E+04, 1.2363E+04, 1.2369E+04, 1.2374E+04,
     +  1.2380E+04, 1.2386E+04, 1.2391E+04, 1.2397E+04, 1.2402E+04,
     + 1.2408E+04, 1.2413E+04, 1.2419E+04, 1.2424E+04, 1.2430E+04,
     +  1.2435E+04, 1.2441E+04, 1.2447E+04, 1.2452E+04, 1.2458E+04,
     + 1.2463E+04, 1.2469E+04, 1.2474E+04, 1.2480E+04, 1.2485E+04,
     +  1.2491E+04, 1.2497E+04, 1.2502E+04, 1.2508E+04, 1.2513E+04,
     + 1.2519E+04, 1.2524E+04, 1.2530E+04, 1.2536E+04, 1.2541E+04,
     +  1.2547E+04, 1.2552E+04, 1.2558E+04, 1.2563E+04, 1.2569E+04,
     + 1.2575E+04, 1.2580E+04, 1.2586E+04, 1.2591E+04, 1.2597E+04,
     +  1.2603E+04, 1.2608E+04, 1.2614E+04, 1.2619E+04, 1.2625E+04,
     + 1.2630E+04, 1.2636E+04, 1.2642E+04, 1.2647E+04, 1.2653E+04,
     +  1.2658E+04, 1.2664E+04, 1.2670E+04, 1.2675E+04, 1.2681E+04,
     + 1.2687E+04, 1.2692E+04, 1.2698E+04, 1.2703E+04, 1.2709E+04,
     +  1.2715E+04, 1.2720E+04, 1.2726E+04, 1.2731E+04, 1.2737E+04,
     + 1.2743E+04, 1.2748E+04, 1.2754E+04, 1.2760E+04, 1.2765E+04,
     +  1.2771E+04, 1.2776E+04, 1.2782E+04, 1.2788E+04, 1.2793E+04,
     + 1.2799E+04, 1.2805E+04, 1.2810E+04, 1.2816E+04, 1.2822E+04,
     +  1.2827E+04, 1.2833E+04, 1.2838E+04, 1.2844E+04, 1.2850E+04,
     + 1.2855E+04, 1.2861E+04, 1.2867E+04, 1.2872E+04, 1.2878E+04,
     +  1.2884E+04, 1.2889E+04, 1.2895E+04, 1.2901E+04, 1.2906E+04,
     + 1.2912E+04, 1.2918E+04, 1.2923E+04, 1.2929E+04, 1.2935E+04,
     +  1.2940E+04, 1.2946E+04, 1.2952E+04, 1.2957E+04, 1.2963E+04,
     + 1.2969E+04, 1.2974E+04, 1.2980E+04, 1.2986E+04, 1.2991E+04,
     +  1.2997E+04, 1.3003E+04, 1.3008E+04, 1.3014E+04, 1.3020E+04,
     + 1.3025E+04, 1.3031E+04, 1.3037E+04, 1.3043E+04, 1.3048E+04,
     +  1.3054E+04, 1.3060E+04, 1.3065E+04, 1.3071E+04, 1.3077E+04,
     + 1.3082E+04, 1.3088E+04, 1.3094E+04, 1.3100E+04, 1.3105E+04,
     +  1.3111E+04, 1.3117E+04, 1.3122E+04, 1.3128E+04, 1.3134E+04,
     + 1.3139E+04, 1.3145E+04, 1.3151E+04, 1.3157E+04, 1.3162E+04,
     +  1.3168E+04, 1.3174E+04, 1.3180E+04, 1.3185E+04, 1.3191E+04,
     + 1.3197E+04, 1.3202E+04, 1.3208E+04, 1.3214E+04, 1.3220E+04,
     +  1.3225E+04, 1.3231E+04, 1.3237E+04, 1.3243E+04, 1.3248E+04,
     + 1.3254E+04, 1.3260E+04, 1.3266E+04, 1.3271E+04, 1.3277E+04,
     +  1.3283E+04, 1.3288E+04, 1.3294E+04, 1.3300E+04, 1.3306E+04,
     + 1.3311E+04, 1.3317E+04, 1.3323E+04, 1.3329E+04, 1.3335E+04,
     +  1.3340E+04, 1.3346E+04, 1.3352E+04, 1.3358E+04, 1.3363E+04,
     + 1.3369E+04, 1.3375E+04, 1.3381E+04, 1.3386E+04, 1.3392E+04,
     +  1.3398E+04, 1.3404E+04, 1.3409E+04, 1.3415E+04, 1.3421E+04,
     + 1.3427E+04, 1.3433E+04, 1.3438E+04, 1.3444E+04, 1.3450E+04,
     +  1.3456E+04, 1.3462E+04, 1.3467E+04, 1.3473E+04, 1.3479E+04,
     + 1.3485E+04, 1.3490E+04, 1.3496E+04, 1.3502E+04, 1.3508E+04,
     +  1.3514E+04, 1.3519E+04, 1.3525E+04, 1.3531E+04, 1.3537E+04,
     + 1.3543E+04, 1.3548E+04, 1.3554E+04, 1.3560E+04, 1.3566E+04,
     +  1.3572E+04, 1.3578E+04, 1.3583E+04, 1.3589E+04, 1.3595E+04,
     + 1.3601E+04, 1.3607E+04, 1.3612E+04, 1.3618E+04/
c...        --        48
      data (QofT( 8,J),J=1,2408)/ 1.0243E+00, 1.2734E+00, 1.6431E+00,
     +  2.0396E+00, 2.4446E+00, 2.8533E+00, 3.2640E+00, 3.6759E+00,
     + 4.0886E+00, 4.5018E+00, 4.9153E+00, 5.3292E+00, 5.7432E+00,
     +  6.1575E+00, 6.5718E+00, 6.9863E+00, 7.4008E+00, 7.8155E+00,
     + 8.2302E+00, 8.6449E+00, 9.4745E+00, 1.0304E+01, 1.1134E+01,
     +  1.1964E+01, 1.2794E+01, 1.3624E+01, 1.4454E+01, 1.5284E+01,
     + 1.6115E+01, 1.6945E+01, 1.7775E+01, 1.8605E+01, 1.9436E+01,
     +  2.0266E+01, 2.1096E+01, 2.1927E+01, 2.2757E+01, 2.3588E+01,
     + 2.4418E+01, 2.5248E+01, 2.6079E+01, 2.6909E+01, 2.7740E+01,
     +  2.8570E+01, 2.9401E+01, 3.0231E+01, 3.1062E+01, 3.1892E+01,
     + 3.2723E+01, 3.3553E+01, 3.4384E+01, 3.5214E+01, 3.6045E+01,
     +  3.6876E+01, 3.7706E+01, 3.8537E+01, 3.9367E+01, 4.0198E+01,
     + 4.1029E+01, 4.1859E+01, 4.2690E+01, 4.3521E+01, 4.4351E+01,
     +  4.5182E+01, 4.6013E+01, 4.6843E+01, 4.7674E+01, 4.8505E+01,
     + 4.9336E+01, 5.0166E+01, 5.0997E+01, 5.1828E+01, 5.2659E+01,
     +  5.3489E+01, 5.4320E+01, 5.5151E+01, 5.5982E+01, 5.6812E+01,
     + 5.7643E+01, 5.8474E+01, 5.9305E+01, 6.0136E+01, 6.0966E+01,
     +  6.1797E+01, 6.2628E+01, 6.3459E+01, 6.4290E+01, 6.5121E+01,
     + 6.5952E+01, 6.6783E+01, 6.7613E+01, 6.8444E+01, 6.9275E+01,
     +  7.0106E+01, 7.0937E+01, 7.1768E+01, 7.2599E+01, 7.3430E+01,
     + 7.4261E+01, 7.5092E+01, 7.5923E+01, 7.6754E+01, 7.7585E+01,
     +  7.8416E+01, 7.9247E+01, 8.0078E+01, 8.0909E+01, 8.1740E+01,
     + 8.2571E+01, 8.3402E+01, 8.4233E+01, 8.5064E+01, 8.5895E+01,
     +  8.6726E+01, 8.7558E+01, 8.8389E+01, 8.9220E+01, 9.0051E+01,
     + 9.0882E+01, 9.1713E+01, 9.2544E+01, 9.3376E+01, 9.4207E+01,
     +  9.5038E+01, 9.5869E+01, 9.6700E+01, 9.7532E+01, 9.8363E+01,
     + 9.9194E+01, 1.0003E+02, 1.0086E+02, 1.0169E+02, 1.0252E+02,
     +  1.0335E+02, 1.0418E+02, 1.0501E+02, 1.0584E+02, 1.0668E+02,
     + 1.0751E+02, 1.0834E+02, 1.0917E+02, 1.1000E+02, 1.1083E+02,
     +  1.1166E+02, 1.1250E+02, 1.1333E+02, 1.1416E+02, 1.1499E+02,
     + 1.1582E+02, 1.1665E+02, 1.1749E+02, 1.1832E+02, 1.1915E+02,
     +  1.1998E+02, 1.2081E+02, 1.2165E+02, 1.2248E+02, 1.2331E+02,
     + 1.2414E+02, 1.2497E+02, 1.2581E+02, 1.2664E+02, 1.2747E+02,
     +  1.2830E+02, 1.2913E+02, 1.2997E+02, 1.3080E+02, 1.3163E+02,
     + 1.3246E+02, 1.3330E+02, 1.3413E+02, 1.3496E+02, 1.3579E+02,
     +  1.3663E+02, 1.3746E+02, 1.3829E+02, 1.3913E+02, 1.3996E+02,
     + 1.4079E+02, 1.4162E+02, 1.4246E+02, 1.4329E+02, 1.4413E+02,
     +  1.4496E+02, 1.4579E+02, 1.4663E+02, 1.4746E+02, 1.4829E+02,
     + 1.4913E+02, 1.4996E+02, 1.5080E+02, 1.5163E+02, 1.5247E+02,
     +  1.5330E+02, 1.5413E+02, 1.5497E+02, 1.5580E+02, 1.5664E+02,
     + 1.5747E+02, 1.5831E+02, 1.5915E+02, 1.5998E+02, 1.6082E+02,
     +  1.6165E+02, 1.6249E+02, 1.6333E+02, 1.6416E+02, 1.6500E+02,
     + 1.6584E+02, 1.6667E+02, 1.6751E+02, 1.6835E+02, 1.6918E+02,
     +  1.7002E+02, 1.7086E+02, 1.7170E+02, 1.7254E+02, 1.7337E+02,
     + 1.7421E+02, 1.7505E+02, 1.7589E+02, 1.7673E+02, 1.7757E+02,
     +  1.7841E+02, 1.7925E+02, 1.8009E+02, 1.8093E+02, 1.8177E+02,
     + 1.8261E+02, 1.8345E+02, 1.8429E+02, 1.8513E+02, 1.8598E+02,
     +  1.8682E+02, 1.8766E+02, 1.8850E+02, 1.8934E+02, 1.9019E+02,
     + 1.9103E+02, 1.9187E+02, 1.9272E+02, 1.9356E+02, 1.9441E+02,
     +  1.9525E+02, 1.9610E+02, 1.9694E+02, 1.9779E+02, 1.9863E+02,
     + 1.9948E+02, 2.0033E+02, 2.0117E+02, 2.0202E+02, 2.0287E+02,
     +  2.0372E+02, 2.0457E+02, 2.0541E+02, 2.0626E+02, 2.0711E+02,
     + 2.0796E+02, 2.0881E+02, 2.0966E+02, 2.1051E+02, 2.1136E+02,
     +  2.1222E+02, 2.1307E+02, 2.1392E+02, 2.1477E+02, 2.1563E+02,
     + 2.1648E+02, 2.1733E+02, 2.1819E+02, 2.1904E+02, 2.1990E+02,
     +  2.2075E+02, 2.2161E+02, 2.2246E+02, 2.2332E+02, 2.2418E+02,
     + 2.2504E+02, 2.2589E+02, 2.2675E+02, 2.2761E+02, 2.2847E+02,
     +  2.2933E+02, 2.3019E+02, 2.3105E+02, 2.3191E+02, 2.3278E+02,
     + 2.3364E+02, 2.3450E+02, 2.3536E+02, 2.3623E+02, 2.3709E+02,
     +  2.3796E+02, 2.3882E+02, 2.3969E+02, 2.4055E+02, 2.4142E+02,
     + 2.4229E+02, 2.4315E+02, 2.4402E+02, 2.4489E+02, 2.4576E+02,
     +  2.4663E+02, 2.4750E+02, 2.4837E+02, 2.4924E+02, 2.5012E+02,
     + 2.5099E+02, 2.5186E+02, 2.5273E+02, 2.5361E+02, 2.5448E+02,
     +  2.5536E+02, 2.5623E+02, 2.5711E+02, 2.5799E+02, 2.5887E+02,
     + 2.5974E+02, 2.6062E+02, 2.6150E+02, 2.6238E+02, 2.6326E+02,
     +  2.6414E+02, 2.6503E+02, 2.6591E+02, 2.6679E+02, 2.6768E+02,
     + 2.6856E+02, 2.6944E+02, 2.7033E+02, 2.7122E+02, 2.7210E+02,
     +  2.7299E+02, 2.7388E+02, 2.7477E+02, 2.7566E+02, 2.7655E+02,
     + 2.7744E+02, 2.7833E+02, 2.7922E+02, 2.8011E+02, 2.8101E+02,
     +  2.8190E+02, 2.8279E+02, 2.8369E+02, 2.8459E+02, 2.8548E+02,
     + 2.8638E+02, 2.8728E+02, 2.8818E+02, 2.8908E+02, 2.8998E+02,
     +  2.9088E+02, 2.9178E+02, 2.9268E+02, 2.9358E+02, 2.9449E+02,
     + 2.9539E+02, 2.9630E+02, 2.9720E+02, 2.9811E+02, 2.9902E+02,
     +  2.9992E+02, 3.0083E+02, 3.0174E+02, 3.0265E+02, 3.0356E+02,
     + 3.0448E+02, 3.0539E+02, 3.0630E+02, 3.0721E+02, 3.0813E+02,
     +  3.0904E+02, 3.0996E+02, 3.1088E+02, 3.1180E+02, 3.1271E+02,
     + 3.1363E+02, 3.1455E+02, 3.1547E+02, 3.1640E+02, 3.1732E+02,
     +  3.1824E+02, 3.1916E+02, 3.2009E+02, 3.2101E+02, 3.2194E+02,
     + 3.2287E+02, 3.2380E+02, 3.2472E+02, 3.2565E+02, 3.2658E+02,
     +  3.2752E+02, 3.2845E+02, 3.2938E+02, 3.3031E+02, 3.3125E+02,
     + 3.3218E+02, 3.3312E+02, 3.3405E+02, 3.3499E+02, 3.3593E+02,
     +  3.3687E+02, 3.3781E+02, 3.3875E+02, 3.3969E+02, 3.4064E+02,
     + 3.4158E+02, 3.4252E+02, 3.4347E+02, 3.4441E+02, 3.4536E+02,
     +  3.4631E+02, 3.4726E+02, 3.4821E+02, 3.4916E+02, 3.5011E+02,
     + 3.5106E+02, 3.5201E+02, 3.5296E+02, 3.5392E+02, 3.5487E+02,
     +  3.5583E+02, 3.5679E+02, 3.5775E+02, 3.5870E+02, 3.5966E+02,
     + 3.6062E+02, 3.6159E+02, 3.6255E+02, 3.6351E+02, 3.6448E+02,
     +  3.6544E+02, 3.6641E+02, 3.6737E+02, 3.6834E+02, 3.6931E+02,
     + 3.7028E+02, 3.7125E+02, 3.7222E+02, 3.7319E+02, 3.7417E+02,
     +  3.7514E+02, 3.7611E+02, 3.7709E+02, 3.7807E+02, 3.7904E+02,
     + 3.8002E+02, 3.8100E+02, 3.8198E+02, 3.8296E+02, 3.8395E+02,
     +  3.8493E+02, 3.8591E+02, 3.8690E+02, 3.8788E+02, 3.8887E+02,
     + 3.8986E+02, 3.9085E+02, 3.9184E+02, 3.9283E+02, 3.9382E+02,
     +  3.9481E+02, 3.9581E+02, 3.9680E+02, 3.9779E+02, 3.9879E+02,
     + 3.9979E+02, 4.0079E+02, 4.0179E+02, 4.0279E+02, 4.0379E+02,
     +  4.0479E+02, 4.0579E+02, 4.0679E+02, 4.0780E+02, 4.0881E+02,
     + 4.0981E+02, 4.1082E+02, 4.1183E+02, 4.1284E+02, 4.1385E+02,
     +  4.1486E+02, 4.1587E+02, 4.1689E+02, 4.1790E+02, 4.1892E+02,
     + 4.1993E+02, 4.2095E+02, 4.2197E+02, 4.2299E+02, 4.2401E+02,
     +  4.2503E+02, 4.2605E+02, 4.2707E+02, 4.2810E+02, 4.2912E+02,
     + 4.3015E+02, 4.3118E+02, 4.3220E+02, 4.3323E+02, 4.3426E+02,
     +  4.3529E+02, 4.3633E+02, 4.3736E+02, 4.3839E+02, 4.3943E+02,
     + 4.4046E+02, 4.4150E+02, 4.4254E+02, 4.4358E+02, 4.4462E+02,
     +  4.4566E+02, 4.4670E+02, 4.4775E+02, 4.4879E+02, 4.4983E+02,
     + 4.5088E+02, 4.5193E+02, 4.5298E+02, 4.5403E+02, 4.5508E+02,
     +  4.5613E+02, 4.5718E+02, 4.5823E+02, 4.5929E+02, 4.6034E+02,
     + 4.6140E+02, 4.6245E+02, 4.6351E+02, 4.6457E+02, 4.6563E+02,
     +  4.6669E+02, 4.6776E+02, 4.6882E+02, 4.6988E+02, 4.7095E+02,
     + 4.7202E+02, 4.7308E+02, 4.7415E+02, 4.7522E+02, 4.7629E+02,
     +  4.7736E+02, 4.7844E+02, 4.7951E+02, 4.8059E+02, 4.8166E+02,
     + 4.8274E+02, 4.8382E+02, 4.8490E+02, 4.8598E+02, 4.8706E+02,
     +  4.8814E+02, 4.8922E+02, 4.9031E+02, 4.9139E+02, 4.9248E+02,
     + 4.9356E+02, 4.9465E+02, 4.9574E+02, 4.9683E+02, 4.9792E+02,
     +  4.9902E+02, 5.0011E+02, 5.0120E+02, 5.0230E+02, 5.0340E+02,
     + 5.0449E+02, 5.0559E+02, 5.0669E+02, 5.0779E+02, 5.0890E+02,
     +  5.1000E+02, 5.1110E+02, 5.1221E+02, 5.1332E+02, 5.1442E+02,
     + 5.1553E+02, 5.1664E+02, 5.1775E+02, 5.1886E+02, 5.1998E+02,
     +  5.2109E+02, 5.2220E+02, 5.2332E+02, 5.2444E+02, 5.2555E+02,
     + 5.2667E+02, 5.2779E+02, 5.2892E+02, 5.3004E+02, 5.3116E+02,
     +  5.3229E+02, 5.3341E+02, 5.3454E+02, 5.3567E+02, 5.3679E+02,
     + 5.3792E+02, 5.3905E+02, 5.4019E+02, 5.4132E+02, 5.4245E+02,
     +  5.4359E+02, 5.4473E+02, 5.4586E+02, 5.4700E+02, 5.4814E+02,
     + 5.4928E+02, 5.5042E+02, 5.5157E+02, 5.5271E+02, 5.5385E+02,
     +  5.5500E+02, 5.5615E+02, 5.5730E+02, 5.5845E+02, 5.5960E+02,
     + 5.6075E+02, 5.6190E+02, 5.6305E+02, 5.6421E+02, 5.6536E+02,
     +  5.6652E+02, 5.6768E+02, 5.6884E+02, 5.7000E+02, 5.7116E+02,
     + 5.7232E+02, 5.7349E+02, 5.7465E+02, 5.7582E+02, 5.7698E+02,
     +  5.7815E+02, 5.7932E+02, 5.8049E+02, 5.8166E+02, 5.8283E+02,
     + 5.8401E+02, 5.8518E+02, 5.8636E+02, 5.8753E+02, 5.8871E+02,
     +  5.8989E+02, 5.9107E+02, 5.9225E+02, 5.9343E+02, 5.9462E+02,
     + 5.9580E+02, 5.9699E+02, 5.9817E+02, 5.9936E+02, 6.0055E+02,
     +  6.0174E+02, 6.0293E+02, 6.0412E+02, 6.0532E+02, 6.0651E+02,
     + 6.0771E+02, 6.0890E+02, 6.1010E+02, 6.1130E+02, 6.1250E+02,
     +  6.1370E+02, 6.1490E+02, 6.1611E+02, 6.1731E+02, 6.1852E+02,
     + 6.1972E+02, 6.2093E+02, 6.2214E+02, 6.2335E+02, 6.2456E+02,
     +  6.2577E+02, 6.2699E+02, 6.2820E+02, 6.2942E+02, 6.3064E+02,
     + 6.3185E+02, 6.3307E+02, 6.3429E+02, 6.3551E+02, 6.3674E+02,
     +  6.3796E+02, 6.3918E+02, 6.4041E+02, 6.4164E+02, 6.4286E+02,
     + 6.4409E+02, 6.4532E+02, 6.4656E+02, 6.4779E+02, 6.4902E+02,
     +  6.5026E+02, 6.5149E+02, 6.5273E+02, 6.5397E+02, 6.5521E+02,
     + 6.5645E+02, 6.5769E+02, 6.5893E+02, 6.6018E+02, 6.6142E+02,
     +  6.6267E+02, 6.6391E+02, 6.6516E+02, 6.6641E+02, 6.6766E+02,
     + 6.6891E+02, 6.7017E+02, 6.7142E+02, 6.7268E+02, 6.7393E+02,
     +  6.7519E+02, 6.7645E+02, 6.7771E+02, 6.7897E+02, 6.8023E+02,
     + 6.8149E+02, 6.8276E+02, 6.8402E+02, 6.8529E+02, 6.8656E+02,
     +  6.8783E+02, 6.8910E+02, 6.9037E+02, 6.9164E+02, 6.9291E+02,
     + 6.9419E+02, 6.9546E+02, 6.9674E+02, 6.9802E+02, 6.9930E+02,
     +  7.0058E+02, 7.0186E+02, 7.0314E+02, 7.0442E+02, 7.0571E+02,
     + 7.0699E+02, 7.0828E+02, 7.0957E+02, 7.1086E+02, 7.1215E+02,
     +  7.1344E+02, 7.1473E+02, 7.1603E+02, 7.1732E+02, 7.1862E+02,
     + 7.1992E+02, 7.2121E+02, 7.2251E+02, 7.2381E+02, 7.2512E+02,
     +  7.2642E+02, 7.2772E+02, 7.2903E+02, 7.3034E+02, 7.3164E+02,
     + 7.3295E+02, 7.3426E+02, 7.3557E+02, 7.3689E+02, 7.3820E+02,
     +  7.3951E+02, 7.4083E+02, 7.4215E+02, 7.4346E+02, 7.4478E+02,
     + 7.4610E+02, 7.4742E+02, 7.4875E+02, 7.5007E+02, 7.5140E+02,
     +  7.5272E+02, 7.5405E+02, 7.5538E+02, 7.5671E+02, 7.5804E+02,
     + 7.5937E+02, 7.6070E+02, 7.6204E+02, 7.6337E+02, 7.6471E+02,
     +  7.6605E+02, 7.6739E+02, 7.6873E+02, 7.7007E+02, 7.7141E+02,
     + 7.7275E+02, 7.7410E+02, 7.7544E+02, 7.7679E+02, 7.7814E+02,
     +  7.7949E+02, 7.8084E+02, 7.8219E+02, 7.8354E+02, 7.8490E+02,
     + 7.8625E+02, 7.8761E+02, 7.8896E+02, 7.9032E+02, 7.9168E+02,
     +  7.9304E+02, 7.9441E+02, 7.9577E+02, 7.9713E+02, 7.9850E+02,
     + 7.9987E+02, 8.0123E+02, 8.0260E+02, 8.0397E+02, 8.0534E+02,
     +  8.0672E+02, 8.0809E+02, 8.0946E+02, 8.1084E+02, 8.1222E+02,
     + 8.1360E+02, 8.1498E+02, 8.1636E+02, 8.1774E+02, 8.1912E+02,
     +  8.2051E+02, 8.2189E+02, 8.2328E+02, 8.2466E+02, 8.2605E+02,
     + 8.2744E+02, 8.2883E+02, 8.3023E+02, 8.3162E+02, 8.3302E+02,
     +  8.3441E+02, 8.3581E+02, 8.3721E+02, 8.3861E+02, 8.4001E+02,
     + 8.4141E+02, 8.4281E+02, 8.4421E+02, 8.4562E+02, 8.4703E+02,
     +  8.4843E+02, 8.4984E+02, 8.5125E+02, 8.5266E+02, 8.5408E+02,
     + 8.5549E+02, 8.5690E+02, 8.5832E+02, 8.5974E+02, 8.6115E+02,
     +  8.6257E+02, 8.6399E+02, 8.6542E+02, 8.6684E+02, 8.6826E+02,
     + 8.6969E+02, 8.7111E+02, 8.7254E+02, 8.7397E+02, 8.7540E+02,
     +  8.7683E+02, 8.7826E+02, 8.7970E+02, 8.8113E+02, 8.8257E+02,
     + 8.8400E+02, 8.8544E+02, 8.8688E+02, 8.8832E+02, 8.8976E+02,
     +  8.9121E+02, 8.9265E+02, 8.9410E+02, 8.9554E+02, 8.9699E+02,
     + 8.9844E+02, 8.9989E+02, 9.0134E+02, 9.0279E+02, 9.0424E+02,
     +  9.0570E+02, 9.0715E+02, 9.0861E+02, 9.1007E+02, 9.1153E+02,
     + 9.1299E+02, 9.1445E+02, 9.1591E+02, 9.1738E+02, 9.1884E+02,
     +  9.2031E+02, 9.2178E+02, 9.2325E+02, 9.2472E+02, 9.2619E+02,
     + 9.2766E+02, 9.2913E+02, 9.3061E+02, 9.3208E+02, 9.3356E+02,
     +  9.3504E+02, 9.3652E+02, 9.3800E+02, 9.3948E+02, 9.4096E+02,
     + 9.4245E+02, 9.4393E+02, 9.4542E+02, 9.4691E+02, 9.4840E+02,
     +  9.4989E+02, 9.5138E+02, 9.5287E+02, 9.5436E+02, 9.5586E+02,
     + 9.5735E+02, 9.5885E+02, 9.6035E+02, 9.6185E+02, 9.6335E+02,
     +  9.6485E+02, 9.6636E+02, 9.6786E+02, 9.6937E+02, 9.7087E+02,
     + 9.7238E+02, 9.7389E+02, 9.7540E+02, 9.7691E+02, 9.7842E+02,
     +  9.7994E+02, 9.8145E+02, 9.8297E+02, 9.8449E+02, 9.8600E+02,
     + 9.8752E+02, 9.8905E+02, 9.9057E+02, 9.9209E+02, 9.9362E+02,
     +  9.9514E+02, 9.9667E+02, 9.9820E+02, 9.9973E+02, 1.0013E+03,
     + 1.0028E+03, 1.0043E+03, 1.0059E+03, 1.0074E+03, 1.0089E+03,
     +  1.0105E+03, 1.0120E+03, 1.0135E+03, 1.0151E+03, 1.0166E+03,
     + 1.0182E+03, 1.0197E+03, 1.0213E+03, 1.0228E+03, 1.0244E+03,
     +  1.0259E+03, 1.0275E+03, 1.0290E+03, 1.0306E+03, 1.0321E+03,
     + 1.0337E+03, 1.0352E+03, 1.0368E+03, 1.0383E+03, 1.0399E+03,
     +  1.0415E+03, 1.0430E+03, 1.0446E+03, 1.0462E+03, 1.0477E+03,
     + 1.0493E+03, 1.0509E+03, 1.0524E+03, 1.0540E+03, 1.0556E+03,
     +  1.0571E+03, 1.0587E+03, 1.0603E+03, 1.0619E+03, 1.0635E+03,
     + 1.0650E+03, 1.0666E+03, 1.0682E+03, 1.0698E+03, 1.0714E+03,
     +  1.0729E+03, 1.0745E+03, 1.0761E+03, 1.0777E+03, 1.0793E+03,
     + 1.0809E+03, 1.0825E+03, 1.0841E+03, 1.0857E+03, 1.0873E+03,
     +  1.0889E+03, 1.0905E+03, 1.0921E+03, 1.0937E+03, 1.0953E+03,
     + 1.0969E+03, 1.0985E+03, 1.1001E+03, 1.1017E+03, 1.1033E+03,
     +  1.1049E+03, 1.1065E+03, 1.1081E+03, 1.1097E+03, 1.1114E+03,
     + 1.1130E+03, 1.1146E+03, 1.1162E+03, 1.1178E+03, 1.1194E+03,
     +  1.1211E+03, 1.1227E+03, 1.1243E+03, 1.1259E+03, 1.1276E+03,
     + 1.1292E+03, 1.1308E+03, 1.1325E+03, 1.1341E+03, 1.1357E+03,
     +  1.1373E+03, 1.1390E+03, 1.1406E+03, 1.1423E+03, 1.1439E+03,
     + 1.1455E+03, 1.1472E+03, 1.1488E+03, 1.1505E+03, 1.1521E+03,
     +  1.1537E+03, 1.1554E+03, 1.1570E+03, 1.1587E+03, 1.1603E+03,
     + 1.1620E+03, 1.1636E+03, 1.1653E+03, 1.1670E+03, 1.1686E+03,
     +  1.1703E+03, 1.1719E+03, 1.1736E+03, 1.1753E+03, 1.1769E+03,
     + 1.1786E+03, 1.1802E+03, 1.1819E+03, 1.1836E+03, 1.1852E+03,
     +  1.1869E+03, 1.1886E+03, 1.1903E+03, 1.1919E+03, 1.1936E+03,
     + 1.1953E+03, 1.1970E+03, 1.1986E+03, 1.2003E+03, 1.2020E+03,
     +  1.2037E+03, 1.2054E+03, 1.2070E+03, 1.2087E+03, 1.2104E+03,
     + 1.2121E+03, 1.2138E+03, 1.2155E+03, 1.2172E+03, 1.2189E+03,
     +  1.2206E+03, 1.2223E+03, 1.2240E+03, 1.2256E+03, 1.2273E+03,
     + 1.2290E+03, 1.2307E+03, 1.2324E+03, 1.2342E+03, 1.2359E+03,
     +  1.2376E+03, 1.2393E+03, 1.2410E+03, 1.2427E+03, 1.2444E+03,
     + 1.2461E+03, 1.2478E+03, 1.2495E+03, 1.2513E+03, 1.2530E+03,
     +  1.2547E+03, 1.2564E+03, 1.2581E+03, 1.2599E+03, 1.2616E+03,
     + 1.2633E+03, 1.2650E+03, 1.2667E+03, 1.2685E+03, 1.2702E+03,
     +  1.2719E+03, 1.2737E+03, 1.2754E+03, 1.2771E+03, 1.2789E+03,
     + 1.2806E+03, 1.2823E+03, 1.2841E+03, 1.2858E+03, 1.2876E+03,
     +  1.2893E+03, 1.2910E+03, 1.2928E+03, 1.2945E+03, 1.2963E+03,
     + 1.2980E+03, 1.2998E+03, 1.3015E+03, 1.3033E+03, 1.3050E+03,
     +  1.3068E+03, 1.3085E+03, 1.3103E+03, 1.3121E+03, 1.3138E+03,
     + 1.3156E+03, 1.3173E+03, 1.3191E+03, 1.3209E+03, 1.3226E+03,
     +  1.3244E+03, 1.3262E+03, 1.3279E+03, 1.3297E+03, 1.3315E+03,
     + 1.3333E+03, 1.3350E+03, 1.3368E+03, 1.3386E+03, 1.3404E+03,
     +  1.3421E+03, 1.3439E+03, 1.3457E+03, 1.3475E+03, 1.3493E+03,
     + 1.3511E+03, 1.3528E+03, 1.3546E+03, 1.3564E+03, 1.3582E+03,
     +  1.3600E+03, 1.3618E+03, 1.3636E+03, 1.3654E+03, 1.3672E+03,
     + 1.3690E+03, 1.3708E+03, 1.3726E+03, 1.3744E+03, 1.3762E+03,
     +  1.3780E+03, 1.3798E+03, 1.3816E+03, 1.3834E+03, 1.3852E+03,
     + 1.3870E+03, 1.3888E+03, 1.3906E+03, 1.3924E+03, 1.3943E+03,
     +  1.3961E+03, 1.3979E+03, 1.3997E+03, 1.4015E+03, 1.4033E+03,
     + 1.4052E+03, 1.4070E+03, 1.4088E+03, 1.4106E+03, 1.4125E+03,
     +  1.4143E+03, 1.4161E+03, 1.4179E+03, 1.4198E+03, 1.4216E+03,
     + 1.4234E+03, 1.4253E+03, 1.4271E+03, 1.4290E+03, 1.4308E+03,
     +  1.4326E+03, 1.4345E+03, 1.4363E+03, 1.4382E+03, 1.4400E+03,
     + 1.4418E+03, 1.4437E+03, 1.4455E+03, 1.4474E+03, 1.4492E+03,
     +  1.4511E+03, 1.4529E+03, 1.4548E+03, 1.4567E+03, 1.4585E+03,
     + 1.4604E+03, 1.4622E+03, 1.4641E+03, 1.4660E+03, 1.4678E+03,
     +  1.4697E+03, 1.4715E+03, 1.4734E+03, 1.4753E+03, 1.4772E+03,
     + 1.4790E+03, 1.4809E+03, 1.4828E+03, 1.4846E+03, 1.4865E+03,
     +  1.4884E+03, 1.4903E+03, 1.4922E+03, 1.4940E+03, 1.4959E+03,
     + 1.4978E+03, 1.4997E+03, 1.5016E+03, 1.5035E+03, 1.5053E+03,
     +  1.5072E+03, 1.5091E+03, 1.5110E+03, 1.5129E+03, 1.5148E+03,
     + 1.5167E+03, 1.5186E+03, 1.5205E+03, 1.5224E+03, 1.5243E+03,
     +  1.5262E+03, 1.5281E+03, 1.5300E+03, 1.5319E+03, 1.5338E+03,
     + 1.5357E+03, 1.5376E+03, 1.5395E+03, 1.5414E+03, 1.5433E+03,
     +  1.5453E+03, 1.5472E+03, 1.5491E+03, 1.5510E+03, 1.5529E+03,
     + 1.5548E+03, 1.5568E+03, 1.5587E+03, 1.5606E+03, 1.5625E+03,
     +  1.5645E+03, 1.5664E+03, 1.5683E+03, 1.5702E+03, 1.5722E+03,
     + 1.5741E+03, 1.5760E+03, 1.5780E+03, 1.5799E+03, 1.5818E+03,
     +  1.5838E+03, 1.5857E+03, 1.5877E+03, 1.5896E+03, 1.5915E+03,
     + 1.5935E+03, 1.5954E+03, 1.5974E+03, 1.5993E+03, 1.6013E+03,
     +  1.6032E+03, 1.6052E+03, 1.6071E+03, 1.6091E+03, 1.6110E+03,
     + 1.6130E+03, 1.6150E+03, 1.6169E+03, 1.6189E+03, 1.6208E+03,
     +  1.6228E+03, 1.6248E+03, 1.6267E+03, 1.6287E+03, 1.6307E+03,
     + 1.6326E+03, 1.6346E+03, 1.6366E+03, 1.6385E+03, 1.6405E+03,
     +  1.6425E+03, 1.6445E+03, 1.6465E+03, 1.6484E+03, 1.6504E+03,
     + 1.6524E+03, 1.6544E+03, 1.6564E+03, 1.6583E+03, 1.6603E+03,
     +  1.6623E+03, 1.6643E+03, 1.6663E+03, 1.6683E+03, 1.6703E+03,
     + 1.6723E+03, 1.6743E+03, 1.6763E+03, 1.6783E+03, 1.6803E+03,
     +  1.6823E+03, 1.6843E+03, 1.6863E+03, 1.6883E+03, 1.6903E+03,
     + 1.6923E+03, 1.6943E+03, 1.6963E+03, 1.6983E+03, 1.7003E+03,
     +  1.7023E+03, 1.7043E+03, 1.7063E+03, 1.7084E+03, 1.7104E+03,
     + 1.7124E+03, 1.7144E+03, 1.7164E+03, 1.7185E+03, 1.7205E+03,
     +  1.7225E+03, 1.7245E+03, 1.7266E+03, 1.7286E+03, 1.7306E+03,
     + 1.7326E+03, 1.7347E+03, 1.7367E+03, 1.7387E+03, 1.7408E+03,
     +  1.7428E+03, 1.7449E+03, 1.7469E+03, 1.7489E+03, 1.7510E+03,
     + 1.7530E+03, 1.7551E+03, 1.7571E+03, 1.7592E+03, 1.7612E+03,
     +  1.7633E+03, 1.7653E+03, 1.7674E+03, 1.7694E+03, 1.7715E+03,
     + 1.7735E+03, 1.7756E+03, 1.7776E+03, 1.7797E+03, 1.7818E+03,
     +  1.7838E+03, 1.7859E+03, 1.7879E+03, 1.7900E+03, 1.7921E+03,
     + 1.7941E+03, 1.7962E+03, 1.7983E+03, 1.8003E+03, 1.8024E+03,
     +  1.8045E+03, 1.8066E+03, 1.8086E+03, 1.8107E+03, 1.8128E+03,
     + 1.8149E+03, 1.8170E+03, 1.8190E+03, 1.8211E+03, 1.8232E+03,
     +  1.8253E+03, 1.8274E+03, 1.8295E+03, 1.8316E+03, 1.8337E+03,
     + 1.8358E+03, 1.8378E+03, 1.8399E+03, 1.8420E+03, 1.8441E+03,
     +  1.8462E+03, 1.8483E+03, 1.8504E+03, 1.8525E+03, 1.8546E+03,
     + 1.8567E+03, 1.8589E+03, 1.8610E+03, 1.8631E+03, 1.8652E+03,
     +  1.8673E+03, 1.8694E+03, 1.8715E+03, 1.8736E+03, 1.8757E+03,
     + 1.8779E+03, 1.8800E+03, 1.8821E+03, 1.8842E+03, 1.8863E+03,
     +  1.8885E+03, 1.8906E+03, 1.8927E+03, 1.8949E+03, 1.8970E+03,
     + 1.8991E+03, 1.9012E+03, 1.9034E+03, 1.9055E+03, 1.9076E+03,
     +  1.9098E+03, 1.9119E+03, 1.9141E+03, 1.9162E+03, 1.9183E+03,
     + 1.9205E+03, 1.9226E+03, 1.9248E+03, 1.9269E+03, 1.9291E+03,
     +  1.9312E+03, 1.9334E+03, 1.9355E+03, 1.9377E+03, 1.9398E+03,
     + 1.9420E+03, 1.9441E+03, 1.9463E+03, 1.9484E+03, 1.9506E+03,
     +  1.9528E+03, 1.9549E+03, 1.9571E+03, 1.9593E+03, 1.9614E+03,
     + 1.9636E+03, 1.9658E+03, 1.9679E+03, 1.9701E+03, 1.9723E+03,
     +  1.9744E+03, 1.9766E+03, 1.9788E+03, 1.9810E+03, 1.9831E+03,
     + 1.9853E+03, 1.9875E+03, 1.9897E+03, 1.9919E+03, 1.9941E+03,
     +  1.9962E+03, 1.9984E+03, 2.0006E+03, 2.0028E+03, 2.0050E+03,
     + 2.0072E+03, 2.0094E+03, 2.0116E+03, 2.0138E+03, 2.0160E+03,
     +  2.0182E+03, 2.0204E+03, 2.0226E+03, 2.0248E+03, 2.0270E+03,
     + 2.0292E+03, 2.0314E+03, 2.0336E+03, 2.0358E+03, 2.0380E+03,
     +  2.0402E+03, 2.0424E+03, 2.0447E+03, 2.0469E+03, 2.0491E+03,
     + 2.0513E+03, 2.0535E+03, 2.0557E+03, 2.0580E+03, 2.0602E+03,
     +  2.0624E+03, 2.0646E+03, 2.0669E+03, 2.0691E+03, 2.0713E+03,
     + 2.0735E+03, 2.0758E+03, 2.0780E+03, 2.0802E+03, 2.0825E+03,
     +  2.0847E+03, 2.0869E+03, 2.0892E+03, 2.0914E+03, 2.0937E+03,
     + 2.0959E+03, 2.0982E+03, 2.1004E+03, 2.1026E+03, 2.1049E+03,
     +  2.1071E+03, 2.1094E+03, 2.1116E+03, 2.1139E+03, 2.1161E+03,
     + 2.1184E+03, 2.1207E+03, 2.1229E+03, 2.1252E+03, 2.1274E+03,
     +  2.1297E+03, 2.1320E+03, 2.1342E+03, 2.1365E+03, 2.1388E+03,
     + 2.1410E+03, 2.1433E+03, 2.1456E+03, 2.1478E+03, 2.1501E+03,
     +  2.1524E+03, 2.1546E+03, 2.1569E+03, 2.1592E+03, 2.1615E+03,
     + 2.1638E+03, 2.1660E+03, 2.1683E+03, 2.1706E+03, 2.1729E+03,
     +  2.1752E+03, 2.1775E+03, 2.1798E+03, 2.1820E+03, 2.1843E+03,
     + 2.1866E+03, 2.1889E+03, 2.1912E+03, 2.1935E+03, 2.1958E+03,
     +  2.1981E+03, 2.2004E+03, 2.2027E+03, 2.2050E+03, 2.2073E+03,
     + 2.2096E+03, 2.2119E+03, 2.2142E+03, 2.2165E+03, 2.2189E+03,
     +  2.2212E+03, 2.2235E+03, 2.2258E+03, 2.2281E+03, 2.2304E+03,
     + 2.2327E+03, 2.2351E+03, 2.2374E+03, 2.2397E+03, 2.2420E+03,
     +  2.2443E+03, 2.2467E+03, 2.2490E+03, 2.2513E+03, 2.2537E+03,
     + 2.2560E+03, 2.2583E+03, 2.2607E+03, 2.2630E+03, 2.2653E+03,
     +  2.2677E+03, 2.2700E+03, 2.2723E+03, 2.2747E+03, 2.2770E+03,
     + 2.2794E+03, 2.2817E+03, 2.2840E+03, 2.2864E+03, 2.2887E+03,
     +  2.2911E+03, 2.2934E+03, 2.2958E+03, 2.2981E+03, 2.3005E+03,
     + 2.3029E+03, 2.3052E+03, 2.3076E+03, 2.3099E+03, 2.3123E+03,
     +  2.3147E+03, 2.3170E+03, 2.3194E+03, 2.3217E+03, 2.3241E+03,
     + 2.3265E+03, 2.3289E+03, 2.3312E+03, 2.3336E+03, 2.3360E+03,
     +  2.3383E+03, 2.3407E+03, 2.3431E+03, 2.3455E+03, 2.3479E+03,
     + 2.3502E+03, 2.3526E+03, 2.3550E+03, 2.3574E+03, 2.3598E+03,
     +  2.3622E+03, 2.3645E+03, 2.3669E+03, 2.3693E+03, 2.3717E+03,
     + 2.3741E+03, 2.3765E+03, 2.3789E+03, 2.3813E+03, 2.3837E+03,
     +  2.3861E+03, 2.3885E+03, 2.3909E+03, 2.3933E+03, 2.3957E+03,
     + 2.3981E+03, 2.4005E+03, 2.4029E+03, 2.4053E+03, 2.4077E+03,
     +  2.4102E+03, 2.4126E+03, 2.4150E+03, 2.4174E+03, 2.4198E+03,
     + 2.4222E+03, 2.4247E+03, 2.4271E+03, 2.4295E+03, 2.4319E+03,
     +  2.4344E+03, 2.4368E+03, 2.4392E+03, 2.4416E+03, 2.4441E+03,
     + 2.4465E+03, 2.4489E+03, 2.4514E+03, 2.4538E+03, 2.4562E+03,
     +  2.4587E+03, 2.4611E+03, 2.4635E+03, 2.4660E+03, 2.4684E+03,
     + 2.4709E+03, 2.4733E+03, 2.4758E+03, 2.4782E+03, 2.4807E+03,
     +  2.4831E+03, 2.4856E+03, 2.4880E+03, 2.4905E+03, 2.4929E+03,
     + 2.4954E+03, 2.4978E+03, 2.5003E+03, 2.5028E+03, 2.5052E+03,
     +  2.5077E+03, 2.5102E+03, 2.5126E+03, 2.5151E+03, 2.5176E+03,
     + 2.5200E+03, 2.5225E+03, 2.5250E+03, 2.5274E+03, 2.5299E+03,
     +  2.5324E+03, 2.5349E+03, 2.5373E+03, 2.5398E+03, 2.5423E+03,
     + 2.5448E+03, 2.5473E+03, 2.5497E+03, 2.5522E+03, 2.5547E+03,
     +  2.5572E+03, 2.5597E+03, 2.5622E+03, 2.5647E+03, 2.5672E+03,
     + 2.5697E+03, 2.5722E+03, 2.5747E+03, 2.5772E+03, 2.5797E+03,
     +  2.5822E+03, 2.5847E+03, 2.5872E+03, 2.5897E+03, 2.5922E+03,
     + 2.5947E+03, 2.5972E+03, 2.5997E+03, 2.6022E+03, 2.6047E+03,
     +  2.6072E+03, 2.6098E+03, 2.6123E+03, 2.6148E+03, 2.6173E+03,
     + 2.6198E+03, 2.6224E+03, 2.6249E+03, 2.6274E+03, 2.6299E+03,
     +  2.6325E+03, 2.6350E+03, 2.6375E+03, 2.6400E+03, 2.6426E+03,
     + 2.6451E+03, 2.6476E+03, 2.6502E+03, 2.6527E+03, 2.6552E+03,
     +  2.6578E+03, 2.6603E+03, 2.6629E+03, 2.6654E+03, 2.6680E+03,
     + 2.6705E+03, 2.6730E+03, 2.6756E+03, 2.6781E+03, 2.6807E+03,
     +  2.6832E+03, 2.6858E+03, 2.6884E+03, 2.6909E+03, 2.6935E+03,
     + 2.6960E+03, 2.6986E+03, 2.7011E+03, 2.7037E+03, 2.7063E+03,
     +  2.7088E+03, 2.7114E+03, 2.7140E+03, 2.7165E+03, 2.7191E+03,
     + 2.7217E+03, 2.7242E+03, 2.7268E+03, 2.7294E+03, 2.7320E+03,
     +  2.7346E+03, 2.7371E+03, 2.7397E+03, 2.7423E+03, 2.7449E+03,
     + 2.7475E+03, 2.7500E+03, 2.7526E+03, 2.7552E+03, 2.7578E+03,
     +  2.7604E+03, 2.7630E+03, 2.7656E+03, 2.7682E+03, 2.7708E+03,
     + 2.7734E+03, 2.7760E+03, 2.7786E+03, 2.7812E+03, 2.7838E+03,
     +  2.7864E+03, 2.7890E+03, 2.7916E+03, 2.7942E+03, 2.7968E+03,
     + 2.7994E+03, 2.8020E+03, 2.8046E+03, 2.8072E+03, 2.8099E+03,
     +  2.8125E+03, 2.8151E+03, 2.8177E+03, 2.8203E+03, 2.8229E+03,
     + 2.8256E+03, 2.8282E+03, 2.8308E+03, 2.8334E+03, 2.8361E+03,
     +  2.8387E+03, 2.8413E+03, 2.8440E+03, 2.8466E+03, 2.8492E+03,
     + 2.8519E+03, 2.8545E+03, 2.8571E+03, 2.8598E+03, 2.8624E+03,
     +  2.8650E+03, 2.8677E+03, 2.8703E+03, 2.8730E+03, 2.8756E+03,
     + 2.8783E+03, 2.8809E+03, 2.8836E+03, 2.8862E+03, 2.8889E+03,
     +  2.8915E+03, 2.8942E+03, 2.8968E+03, 2.8995E+03, 2.9022E+03,
     + 2.9048E+03, 2.9075E+03, 2.9101E+03, 2.9128E+03, 2.9155E+03,
     +  2.9181E+03, 2.9208E+03, 2.9235E+03, 2.9262E+03, 2.9288E+03,
     + 2.9315E+03, 2.9342E+03, 2.9368E+03, 2.9395E+03, 2.9422E+03,
     +  2.9449E+03, 2.9476E+03, 2.9502E+03, 2.9529E+03, 2.9556E+03,
     + 2.9583E+03, 2.9610E+03, 2.9637E+03, 2.9664E+03, 2.9691E+03,
     +  2.9717E+03, 2.9744E+03, 2.9771E+03, 2.9798E+03, 2.9825E+03,
     + 2.9852E+03, 2.9879E+03, 2.9906E+03, 2.9933E+03, 2.9960E+03,
     +  2.9987E+03, 3.0015E+03, 3.0042E+03, 3.0069E+03, 3.0096E+03,
     + 3.0123E+03, 3.0150E+03, 3.0177E+03, 3.0204E+03, 3.0231E+03,
     +  3.0259E+03, 3.0286E+03, 3.0313E+03, 3.0340E+03, 3.0368E+03,
     + 3.0395E+03, 3.0422E+03, 3.0449E+03, 3.0477E+03, 3.0504E+03,
     +  3.0531E+03, 3.0559E+03, 3.0586E+03, 3.0613E+03, 3.0641E+03,
     + 3.0668E+03, 3.0695E+03, 3.0723E+03, 3.0750E+03, 3.0778E+03,
     +  3.0805E+03, 3.0832E+03, 3.0860E+03, 3.0887E+03, 3.0915E+03,
     + 3.0942E+03, 3.0970E+03, 3.0997E+03, 3.1025E+03, 3.1053E+03,
     +  3.1080E+03, 3.1108E+03, 3.1135E+03, 3.1163E+03, 3.1191E+03,
     + 3.1218E+03, 3.1246E+03, 3.1273E+03, 3.1301E+03, 3.1329E+03,
     +  3.1357E+03, 3.1384E+03, 3.1412E+03, 3.1440E+03, 3.1467E+03,
     + 3.1495E+03, 3.1523E+03, 3.1551E+03, 3.1579E+03, 3.1606E+03,
     +  3.1634E+03, 3.1662E+03, 3.1690E+03, 3.1718E+03, 3.1746E+03,
     + 3.1773E+03, 3.1801E+03, 3.1829E+03, 3.1857E+03, 3.1885E+03,
     +  3.1913E+03, 3.1941E+03, 3.1969E+03, 3.1997E+03, 3.2025E+03,
     + 3.2053E+03, 3.2081E+03, 3.2109E+03, 3.2137E+03, 3.2165E+03,
     +  3.2193E+03, 3.2221E+03, 3.2250E+03, 3.2278E+03, 3.2306E+03,
     + 3.2334E+03, 3.2362E+03, 3.2390E+03, 3.2419E+03, 3.2447E+03,
     +  3.2475E+03, 3.2503E+03, 3.2531E+03, 3.2560E+03, 3.2588E+03,
     + 3.2616E+03, 3.2644E+03, 3.2673E+03, 3.2701E+03, 3.2729E+03,
     +  3.2758E+03, 3.2786E+03, 3.2814E+03, 3.2843E+03, 3.2871E+03,
     + 3.2900E+03, 3.2928E+03, 3.2957E+03, 3.2985E+03, 3.3013E+03,
     +  3.3042E+03, 3.3070E+03, 3.3099E+03, 3.3127E+03, 3.3156E+03,
     + 3.3184E+03, 3.3213E+03, 3.3242E+03, 3.3270E+03, 3.3299E+03,
     +  3.3327E+03, 3.3356E+03, 3.3385E+03, 3.3413E+03, 3.3442E+03,
     + 3.3471E+03, 3.3499E+03, 3.3528E+03, 3.3557E+03, 3.3585E+03,
     +  3.3614E+03, 3.3643E+03, 3.3672E+03, 3.3700E+03, 3.3729E+03,
     + 3.3758E+03, 3.3787E+03, 3.3815E+03, 3.3844E+03, 3.3873E+03,
     +  3.3902E+03, 3.3931E+03, 3.3960E+03, 3.3989E+03, 3.4018E+03,
     + 3.4047E+03, 3.4075E+03, 3.4104E+03, 3.4133E+03, 3.4162E+03,
     +  3.4191E+03, 3.4220E+03, 3.4249E+03, 3.4278E+03, 3.4307E+03,
     + 3.4336E+03, 3.4366E+03, 3.4395E+03, 3.4424E+03, 3.4453E+03,
     +  3.4482E+03, 3.4511E+03, 3.4540E+03, 3.4569E+03, 3.4599E+03,
     + 3.4628E+03, 3.4657E+03, 3.4686E+03, 3.4715E+03, 3.4745E+03,
     +  3.4774E+03, 3.4803E+03, 3.4832E+03, 3.4862E+03, 3.4891E+03,
     + 3.4920E+03, 3.4950E+03, 3.4979E+03, 3.5008E+03, 3.5038E+03,
     +  3.5067E+03, 3.5096E+03, 3.5126E+03, 3.5155E+03, 3.5185E+03,
     + 3.5214E+03, 3.5244E+03, 3.5273E+03, 3.5303E+03, 3.5332E+03,
     +  3.5362E+03, 3.5391E+03, 3.5421E+03, 3.5450E+03, 3.5480E+03,
     + 3.5509E+03, 3.5539E+03, 3.5568E+03, 3.5598E+03, 3.5628E+03,
     +  3.5657E+03, 3.5687E+03, 3.5717E+03, 3.5746E+03, 3.5776E+03,
     + 3.5806E+03, 3.5835E+03, 3.5865E+03, 3.5895E+03, 3.5925E+03,
     +  3.5954E+03, 3.5984E+03, 3.6014E+03, 3.6044E+03, 3.6074E+03,
     + 3.6103E+03, 3.6133E+03, 3.6163E+03, 3.6193E+03, 3.6223E+03,
     +  3.6253E+03, 3.6283E+03, 3.6313E+03, 3.6343E+03, 3.6373E+03,
     + 3.6402E+03, 3.6432E+03, 3.6462E+03, 3.6492E+03, 3.6522E+03,
     +  3.6553E+03, 3.6583E+03, 3.6613E+03, 3.6643E+03, 3.6673E+03,
     + 3.6703E+03, 3.6733E+03, 3.6763E+03, 3.6793E+03, 3.6823E+03,
     +  3.6854E+03, 3.6884E+03, 3.6914E+03, 3.6944E+03, 3.6974E+03,
     + 3.7004E+03, 3.7035E+03, 3.7065E+03, 3.7095E+03, 3.7126E+03,
     +  3.7156E+03, 3.7186E+03, 3.7216E+03, 3.7247E+03, 3.7277E+03,
     + 3.7307E+03, 3.7338E+03, 3.7368E+03, 3.7399E+03, 3.7429E+03,
     +  3.7459E+03, 3.7490E+03, 3.7520E+03, 3.7551E+03, 3.7581E+03,
     + 3.7612E+03, 3.7642E+03, 3.7673E+03, 3.7703E+03, 3.7734E+03,
     +  3.7764E+03, 3.7795E+03, 3.7825E+03, 3.7856E+03, 3.7887E+03,
     + 3.7917E+03, 3.7948E+03, 3.7979E+03, 3.8009E+03, 3.8040E+03,
     +  3.8071E+03, 3.8101E+03, 3.8132E+03, 3.8163E+03, 3.8193E+03,
     + 3.8224E+03, 3.8255E+03, 3.8286E+03, 3.8316E+03, 3.8347E+03,
     +  3.8378E+03, 3.8409E+03, 3.8440E+03, 3.8471E+03, 3.8501E+03,
     + 3.8532E+03, 3.8563E+03, 3.8594E+03, 3.8625E+03, 3.8656E+03,
     +  3.8687E+03, 3.8718E+03, 3.8749E+03, 3.8780E+03, 3.8811E+03,
     + 3.8842E+03, 3.8873E+03, 3.8904E+03, 3.8935E+03, 3.8966E+03,
     +  3.8997E+03, 3.9028E+03, 3.9059E+03, 3.9090E+03, 3.9121E+03,
     + 3.9153E+03, 3.9184E+03, 3.9215E+03, 3.9246E+03, 3.9277E+03,
     +  3.9309E+03, 3.9340E+03, 3.9371E+03, 3.9402E+03, 3.9433E+03,
     + 3.9465E+03, 3.9496E+03, 3.9527E+03, 3.9559E+03, 3.9590E+03,
     +  3.9621E+03, 3.9653E+03, 3.9684E+03, 3.9715E+03, 3.9747E+03,
     + 3.9778E+03, 3.9810E+03, 3.9841E+03, 3.9872E+03, 3.9904E+03,
     +  3.9935E+03, 3.9967E+03, 3.9998E+03, 4.0030E+03, 4.0061E+03,
     + 4.0093E+03, 4.0124E+03, 4.0156E+03, 4.0188E+03, 4.0219E+03,
     +  4.0251E+03, 4.0282E+03, 4.0314E+03, 4.0346E+03, 4.0377E+03,
     + 4.0409E+03, 4.0441E+03, 4.0472E+03, 4.0504E+03, 4.0536E+03,
     +  4.0567E+03, 4.0599E+03, 4.0631E+03, 4.0663E+03, 4.0694E+03,
     + 4.0726E+03, 4.0758E+03, 4.0790E+03, 4.0822E+03, 4.0854E+03,
     +  4.0885E+03, 4.0917E+03, 4.0949E+03, 4.0981E+03, 4.1013E+03,
     + 4.1045E+03, 4.1077E+03, 4.1109E+03, 4.1141E+03, 4.1173E+03,
     +  4.1205E+03, 4.1237E+03, 4.1269E+03, 4.1301E+03, 4.1333E+03,
     + 4.1365E+03, 4.1397E+03, 4.1429E+03, 4.1461E+03, 4.1493E+03,
     +  4.1525E+03, 4.1558E+03, 4.1590E+03, 4.1622E+03, 4.1654E+03,
     + 4.1686E+03, 4.1718E+03, 4.1751E+03, 4.1783E+03, 4.1815E+03,
     +  4.1847E+03, 4.1880E+03, 4.1912E+03, 4.1944E+03, 4.1977E+03,
     + 4.2009E+03, 4.2041E+03, 4.2074E+03, 4.2106E+03, 4.2138E+03,
     +  4.2171E+03, 4.2203E+03, 4.2235E+03, 4.2268E+03, 4.2300E+03,
     + 4.2333E+03, 4.2365E+03, 4.2398E+03, 4.2430E+03, 4.2463E+03,
     +  4.2495E+03, 4.2528E+03, 4.2560E+03, 4.2593E+03, 4.2625E+03,
     + 4.2658E+03, 4.2691E+03, 4.2723E+03, 4.2756E+03, 4.2788E+03,
     +  4.2821E+03, 4.2854E+03, 4.2886E+03, 4.2919E+03, 4.2952E+03,
     + 4.2985E+03, 4.3017E+03, 4.3050E+03, 4.3083E+03, 4.3116E+03,
     +  4.3148E+03, 4.3181E+03, 4.3214E+03, 4.3247E+03, 4.3280E+03,
     + 4.3312E+03, 4.3345E+03, 4.3378E+03, 4.3411E+03, 4.3444E+03,
     +  4.3477E+03, 4.3510E+03, 4.3543E+03, 4.3576E+03, 4.3609E+03,
     + 4.3642E+03, 4.3675E+03, 4.3708E+03, 4.3741E+03, 4.3774E+03,
     +  4.3807E+03, 4.3840E+03, 4.3873E+03, 4.3906E+03, 4.3939E+03,
     + 4.3972E+03, 4.4005E+03, 4.4038E+03, 4.4072E+03, 4.4105E+03,
     +  4.4138E+03, 4.4171E+03, 4.4204E+03, 4.4238E+03, 4.4271E+03,
     + 4.4304E+03, 4.4337E+03, 4.4371E+03, 4.4404E+03, 4.4437E+03,
     +  4.4470E+03, 4.4504E+03, 4.4537E+03, 4.4570E+03, 4.4604E+03,
     + 4.4637E+03, 4.4671E+03, 4.4704E+03, 4.4737E+03, 4.4771E+03,
     +  4.4804E+03, 4.4838E+03, 4.4871E+03, 4.4905E+03, 4.4938E+03,
     + 4.4972E+03, 4.5005E+03, 4.5039E+03, 4.5072E+03, 4.5106E+03,
     +  4.5139E+03, 4.5173E+03, 4.5207E+03, 4.5240E+03, 4.5274E+03,
     + 4.5308E+03, 4.5341E+03, 4.5375E+03, 4.5409E+03, 4.5442E+03,
     +  4.5476E+03, 4.5510E+03, 4.5543E+03, 4.5577E+03, 4.5611E+03/
      data (QofT( 8,J),J=2409,4510)/ 4.5645E+03, 4.5678E+03,
     +  4.5712E+03, 4.5746E+03, 4.5780E+03, 4.5814E+03, 4.5848E+03,
     +  4.5881E+03,
     + 4.5915E+03, 4.5949E+03, 4.5983E+03, 4.6017E+03, 4.6051E+03,
     +  4.6085E+03, 4.6119E+03, 4.6153E+03, 4.6187E+03, 4.6221E+03,
     + 4.6255E+03, 4.6289E+03, 4.6323E+03, 4.6357E+03, 4.6391E+03,
     +  4.6425E+03, 4.6459E+03, 4.6493E+03, 4.6527E+03, 4.6562E+03,
     + 4.6596E+03, 4.6630E+03, 4.6664E+03, 4.6698E+03, 4.6732E+03,
     +  4.6767E+03, 4.6801E+03, 4.6835E+03, 4.6869E+03, 4.6904E+03,
     + 4.6938E+03, 4.6972E+03, 4.7007E+03, 4.7041E+03, 4.7075E+03,
     +  4.7110E+03, 4.7144E+03, 4.7178E+03, 4.7213E+03, 4.7247E+03,
     + 4.7281E+03, 4.7316E+03, 4.7350E+03, 4.7385E+03, 4.7419E+03,
     +  4.7454E+03, 4.7488E+03, 4.7523E+03, 4.7557E+03, 4.7592E+03,
     + 4.7626E+03, 4.7661E+03, 4.7695E+03, 4.7730E+03, 4.7765E+03,
     +  4.7799E+03, 4.7834E+03, 4.7868E+03, 4.7903E+03, 4.7938E+03,
     + 4.7972E+03, 4.8007E+03, 4.8042E+03, 4.8077E+03, 4.8111E+03,
     +  4.8146E+03, 4.8181E+03, 4.8216E+03, 4.8250E+03, 4.8285E+03,
     + 4.8320E+03, 4.8355E+03, 4.8390E+03, 4.8424E+03, 4.8459E+03,
     +  4.8494E+03, 4.8529E+03, 4.8564E+03, 4.8599E+03, 4.8634E+03,
     + 4.8669E+03, 4.8704E+03, 4.8739E+03, 4.8774E+03, 4.8809E+03,
     +  4.8844E+03, 4.8879E+03, 4.8914E+03, 4.8949E+03, 4.8984E+03,
     + 4.9019E+03, 4.9054E+03, 4.9089E+03, 4.9124E+03, 4.9159E+03,
     +  4.9195E+03, 4.9230E+03, 4.9265E+03, 4.9300E+03, 4.9335E+03,
     + 4.9370E+03, 4.9406E+03, 4.9441E+03, 4.9476E+03, 4.9511E+03,
     +  4.9547E+03, 4.9582E+03, 4.9617E+03, 4.9653E+03, 4.9688E+03,
     + 4.9723E+03, 4.9759E+03, 4.9794E+03, 4.9829E+03, 4.9865E+03,
     +  4.9900E+03, 4.9936E+03, 4.9971E+03, 5.0007E+03, 5.0042E+03,
     + 5.0077E+03, 5.0113E+03, 5.0148E+03, 5.0184E+03, 5.0220E+03,
     +  5.0255E+03, 5.0291E+03, 5.0326E+03, 5.0362E+03, 5.0397E+03,
     + 5.0433E+03, 5.0469E+03, 5.0504E+03, 5.0540E+03, 5.0576E+03,
     +  5.0611E+03, 5.0647E+03, 5.0683E+03, 5.0718E+03, 5.0754E+03,
     + 5.0790E+03, 5.0826E+03, 5.0861E+03, 5.0897E+03, 5.0933E+03,
     +  5.0969E+03, 5.1005E+03, 5.1040E+03, 5.1076E+03, 5.1112E+03,
     + 5.1148E+03, 5.1184E+03, 5.1220E+03, 5.1256E+03, 5.1292E+03,
     +  5.1328E+03, 5.1364E+03, 5.1400E+03, 5.1435E+03, 5.1471E+03,
     + 5.1507E+03, 5.1544E+03, 5.1580E+03, 5.1616E+03, 5.1652E+03,
     +  5.1688E+03, 5.1724E+03, 5.1760E+03, 5.1796E+03, 5.1832E+03,
     + 5.1868E+03, 5.1905E+03, 5.1941E+03, 5.1977E+03, 5.2013E+03,
     +  5.2049E+03, 5.2086E+03, 5.2122E+03, 5.2158E+03, 5.2194E+03,
     + 5.2231E+03, 5.2267E+03, 5.2303E+03, 5.2339E+03, 5.2376E+03,
     +  5.2412E+03, 5.2448E+03, 5.2485E+03, 5.2521E+03, 5.2558E+03,
     + 5.2594E+03, 5.2630E+03, 5.2667E+03, 5.2703E+03, 5.2740E+03,
     +  5.2776E+03, 5.2813E+03, 5.2849E+03, 5.2886E+03, 5.2922E+03,
     + 5.2959E+03, 5.2996E+03, 5.3032E+03, 5.3069E+03, 5.3105E+03,
     +  5.3142E+03, 5.3179E+03, 5.3215E+03, 5.3252E+03, 5.3288E+03,
     + 5.3325E+03, 5.3362E+03, 5.3399E+03, 5.3435E+03, 5.3472E+03,
     +  5.3509E+03, 5.3546E+03, 5.3582E+03, 5.3619E+03, 5.3656E+03,
     + 5.3693E+03, 5.3730E+03, 5.3766E+03, 5.3803E+03, 5.3840E+03,
     +  5.3877E+03, 5.3914E+03, 5.3951E+03, 5.3988E+03, 5.4025E+03,
     + 5.4062E+03, 5.4099E+03, 5.4136E+03, 5.4173E+03, 5.4210E+03,
     +  5.4247E+03, 5.4284E+03, 5.4321E+03, 5.4358E+03, 5.4395E+03,
     + 5.4432E+03, 5.4469E+03, 5.4506E+03, 5.4543E+03, 5.4580E+03,
     +  5.4617E+03, 5.4655E+03, 5.4692E+03, 5.4729E+03, 5.4766E+03,
     + 5.4803E+03, 5.4841E+03, 5.4878E+03, 5.4915E+03, 5.4952E+03,
     +  5.4990E+03, 5.5027E+03, 5.5064E+03, 5.5102E+03, 5.5139E+03,
     + 5.5176E+03, 5.5214E+03, 5.5251E+03, 5.5288E+03, 5.5326E+03,
     +  5.5363E+03, 5.5401E+03, 5.5438E+03, 5.5476E+03, 5.5513E+03,
     + 5.5551E+03, 5.5588E+03, 5.5626E+03, 5.5663E+03, 5.5701E+03,
     +  5.5738E+03, 5.5776E+03, 5.5813E+03, 5.5851E+03, 5.5889E+03,
     + 5.5926E+03, 5.5964E+03, 5.6002E+03, 5.6039E+03, 5.6077E+03,
     +  5.6115E+03, 5.6152E+03, 5.6190E+03, 5.6228E+03, 5.6265E+03,
     + 5.6303E+03, 5.6341E+03, 5.6379E+03, 5.6417E+03, 5.6454E+03,
     +  5.6492E+03, 5.6530E+03, 5.6568E+03, 5.6606E+03, 5.6644E+03,
     + 5.6682E+03, 5.6719E+03, 5.6757E+03, 5.6795E+03, 5.6833E+03,
     +  5.6871E+03, 5.6909E+03, 5.6947E+03, 5.6985E+03, 5.7023E+03,
     + 5.7061E+03, 5.7099E+03, 5.7137E+03, 5.7175E+03, 5.7213E+03,
     +  5.7252E+03, 5.7290E+03, 5.7328E+03, 5.7366E+03, 5.7404E+03,
     + 5.7442E+03, 5.7480E+03, 5.7519E+03, 5.7557E+03, 5.7595E+03,
     +  5.7633E+03, 5.7671E+03, 5.7710E+03, 5.7748E+03, 5.7786E+03,
     + 5.7825E+03, 5.7863E+03, 5.7901E+03, 5.7940E+03, 5.7978E+03,
     +  5.8016E+03, 5.8055E+03, 5.8093E+03, 5.8131E+03, 5.8170E+03,
     + 5.8208E+03, 5.8247E+03, 5.8285E+03, 5.8324E+03, 5.8362E+03,
     +  5.8401E+03, 5.8439E+03, 5.8478E+03, 5.8516E+03, 5.8555E+03,
     + 5.8593E+03, 5.8632E+03, 5.8670E+03, 5.8709E+03, 5.8748E+03,
     +  5.8786E+03, 5.8825E+03, 5.8864E+03, 5.8902E+03, 5.8941E+03,
     + 5.8980E+03, 5.9018E+03, 5.9057E+03, 5.9096E+03, 5.9135E+03,
     +  5.9173E+03, 5.9212E+03, 5.9251E+03, 5.9290E+03, 5.9329E+03,
     + 5.9368E+03, 5.9406E+03, 5.9445E+03, 5.9484E+03, 5.9523E+03,
     +  5.9562E+03, 5.9601E+03, 5.9640E+03, 5.9679E+03, 5.9718E+03,
     + 5.9757E+03, 5.9796E+03, 5.9835E+03, 5.9874E+03, 5.9913E+03,
     +  5.9952E+03, 5.9991E+03, 6.0030E+03, 6.0069E+03, 6.0108E+03,
     + 6.0147E+03, 6.0186E+03, 6.0225E+03, 6.0264E+03, 6.0304E+03,
     +  6.0343E+03, 6.0382E+03, 6.0421E+03, 6.0460E+03, 6.0500E+03,
     + 6.0539E+03, 6.0578E+03, 6.0617E+03, 6.0657E+03, 6.0696E+03,
     +  6.0735E+03, 6.0775E+03, 6.0814E+03, 6.0853E+03, 6.0893E+03,
     + 6.0932E+03, 6.0971E+03, 6.1011E+03, 6.1050E+03, 6.1090E+03,
     +  6.1129E+03, 6.1169E+03, 6.1208E+03, 6.1248E+03, 6.1287E+03,
     + 6.1327E+03, 6.1366E+03, 6.1406E+03, 6.1445E+03, 6.1485E+03,
     +  6.1524E+03, 6.1564E+03, 6.1604E+03, 6.1643E+03, 6.1683E+03,
     + 6.1723E+03, 6.1762E+03, 6.1802E+03, 6.1842E+03, 6.1881E+03,
     +  6.1921E+03, 6.1961E+03, 6.2000E+03, 6.2040E+03, 6.2080E+03,
     + 6.2120E+03, 6.2160E+03, 6.2199E+03, 6.2239E+03, 6.2279E+03,
     +  6.2319E+03, 6.2359E+03, 6.2399E+03, 6.2439E+03, 6.2478E+03,
     + 6.2518E+03, 6.2558E+03, 6.2598E+03, 6.2638E+03, 6.2678E+03,
     +  6.2718E+03, 6.2758E+03, 6.2798E+03, 6.2838E+03, 6.2878E+03,
     + 6.2918E+03, 6.2958E+03, 6.2998E+03, 6.3039E+03, 6.3079E+03,
     +  6.3119E+03, 6.3159E+03, 6.3199E+03, 6.3239E+03, 6.3279E+03,
     + 6.3320E+03, 6.3360E+03, 6.3400E+03, 6.3440E+03, 6.3481E+03,
     +  6.3521E+03, 6.3561E+03, 6.3601E+03, 6.3642E+03, 6.3682E+03,
     + 6.3722E+03, 6.3763E+03, 6.3803E+03, 6.3843E+03, 6.3884E+03,
     +  6.3924E+03, 6.3965E+03, 6.4005E+03, 6.4045E+03, 6.4086E+03,
     + 6.4126E+03, 6.4167E+03, 6.4207E+03, 6.4248E+03, 6.4288E+03,
     +  6.4329E+03, 6.4369E+03, 6.4410E+03, 6.4451E+03, 6.4491E+03,
     + 6.4532E+03, 6.4572E+03, 6.4613E+03, 6.4654E+03, 6.4694E+03,
     +  6.4735E+03, 6.4776E+03, 6.4816E+03, 6.4857E+03, 6.4898E+03,
     + 6.4939E+03, 6.4979E+03, 6.5020E+03, 6.5061E+03, 6.5102E+03,
     +  6.5142E+03, 6.5183E+03, 6.5224E+03, 6.5265E+03, 6.5306E+03,
     + 6.5347E+03, 6.5388E+03, 6.5428E+03, 6.5469E+03, 6.5510E+03,
     +  6.5551E+03, 6.5592E+03, 6.5633E+03, 6.5674E+03, 6.5715E+03,
     + 6.5756E+03, 6.5797E+03, 6.5838E+03, 6.5879E+03, 6.5920E+03,
     +  6.5961E+03, 6.6002E+03, 6.6044E+03, 6.6085E+03, 6.6126E+03,
     + 6.6167E+03, 6.6208E+03, 6.6249E+03, 6.6290E+03, 6.6332E+03,
     +  6.6373E+03, 6.6414E+03, 6.6455E+03, 6.6497E+03, 6.6538E+03,
     + 6.6579E+03, 6.6620E+03, 6.6662E+03, 6.6703E+03, 6.6744E+03,
     +  6.6786E+03, 6.6827E+03, 6.6869E+03, 6.6910E+03, 6.6951E+03,
     + 6.6993E+03, 6.7034E+03, 6.7076E+03, 6.7117E+03, 6.7159E+03,
     +  6.7200E+03, 6.7242E+03, 6.7283E+03, 6.7325E+03, 6.7366E+03,
     + 6.7408E+03, 6.7449E+03, 6.7491E+03, 6.7532E+03, 6.7574E+03,
     +  6.7616E+03, 6.7657E+03, 6.7699E+03, 6.7741E+03, 6.7782E+03,
     + 6.7824E+03, 6.7866E+03, 6.7907E+03, 6.7949E+03, 6.7991E+03,
     +  6.8033E+03, 6.8074E+03, 6.8116E+03, 6.8158E+03, 6.8200E+03,
     + 6.8242E+03, 6.8284E+03, 6.8325E+03, 6.8367E+03, 6.8409E+03,
     +  6.8451E+03, 6.8493E+03, 6.8535E+03, 6.8577E+03, 6.8619E+03,
     + 6.8661E+03, 6.8703E+03, 6.8745E+03, 6.8787E+03, 6.8829E+03,
     +  6.8871E+03, 6.8913E+03, 6.8955E+03, 6.8997E+03, 6.9039E+03,
     + 6.9081E+03, 6.9123E+03, 6.9165E+03, 6.9208E+03, 6.9250E+03,
     +  6.9292E+03, 6.9334E+03, 6.9376E+03, 6.9418E+03, 6.9461E+03,
     + 6.9503E+03, 6.9545E+03, 6.9587E+03, 6.9630E+03, 6.9672E+03,
     +  6.9714E+03, 6.9757E+03, 6.9799E+03, 6.9841E+03, 6.9884E+03,
     + 6.9926E+03, 6.9968E+03, 7.0011E+03, 7.0053E+03, 7.0096E+03,
     +  7.0138E+03, 7.0181E+03, 7.0223E+03, 7.0266E+03, 7.0308E+03,
     + 7.0351E+03, 7.0393E+03, 7.0436E+03, 7.0478E+03, 7.0521E+03,
     +  7.0563E+03, 7.0606E+03, 7.0648E+03, 7.0691E+03, 7.0734E+03,
     + 7.0776E+03, 7.0819E+03, 7.0862E+03, 7.0904E+03, 7.0947E+03,
     +  7.0990E+03, 7.1033E+03, 7.1075E+03, 7.1118E+03, 7.1161E+03,
     + 7.1204E+03, 7.1246E+03, 7.1289E+03, 7.1332E+03, 7.1375E+03,
     +  7.1418E+03, 7.1461E+03, 7.1504E+03, 7.1546E+03, 7.1589E+03,
     + 7.1632E+03, 7.1675E+03, 7.1718E+03, 7.1761E+03, 7.1804E+03,
     +  7.1847E+03, 7.1890E+03, 7.1933E+03, 7.1976E+03, 7.2019E+03,
     + 7.2062E+03, 7.2105E+03, 7.2148E+03, 7.2192E+03, 7.2235E+03,
     +  7.2278E+03, 7.2321E+03, 7.2364E+03, 7.2407E+03, 7.2450E+03,
     + 7.2494E+03, 7.2537E+03, 7.2580E+03, 7.2623E+03, 7.2667E+03,
     +  7.2710E+03, 7.2753E+03, 7.2796E+03, 7.2840E+03, 7.2883E+03,
     + 7.2926E+03, 7.2970E+03, 7.3013E+03, 7.3056E+03, 7.3100E+03,
     +  7.3143E+03, 7.3187E+03, 7.3230E+03, 7.3273E+03, 7.3317E+03,
     + 7.3360E+03, 7.3404E+03, 7.3447E+03, 7.3491E+03, 7.3534E+03,
     +  7.3578E+03, 7.3622E+03, 7.3665E+03, 7.3709E+03, 7.3752E+03,
     + 7.3796E+03, 7.3840E+03, 7.3883E+03, 7.3927E+03, 7.3970E+03,
     +  7.4014E+03, 7.4058E+03, 7.4102E+03, 7.4145E+03, 7.4189E+03,
     + 7.4233E+03, 7.4277E+03, 7.4320E+03, 7.4364E+03, 7.4408E+03,
     +  7.4452E+03, 7.4496E+03, 7.4539E+03, 7.4583E+03, 7.4627E+03,
     + 7.4671E+03, 7.4715E+03, 7.4759E+03, 7.4803E+03, 7.4847E+03,
     +  7.4891E+03, 7.4935E+03, 7.4979E+03, 7.5023E+03, 7.5067E+03,
     + 7.5111E+03, 7.5155E+03, 7.5199E+03, 7.5243E+03, 7.5287E+03,
     +  7.5331E+03, 7.5375E+03, 7.5419E+03, 7.5463E+03, 7.5507E+03,
     + 7.5552E+03, 7.5596E+03, 7.5640E+03, 7.5684E+03, 7.5728E+03,
     +  7.5773E+03, 7.5817E+03, 7.5861E+03, 7.5905E+03, 7.5950E+03,
     + 7.5994E+03, 7.6038E+03, 7.6082E+03, 7.6127E+03, 7.6171E+03,
     +  7.6216E+03, 7.6260E+03, 7.6304E+03, 7.6349E+03, 7.6393E+03,
     + 7.6438E+03, 7.6482E+03, 7.6526E+03, 7.6571E+03, 7.6615E+03,
     +  7.6660E+03, 7.6704E+03, 7.6749E+03, 7.6794E+03, 7.6838E+03,
     + 7.6883E+03, 7.6927E+03, 7.6972E+03, 7.7016E+03, 7.7061E+03,
     +  7.7106E+03, 7.7150E+03, 7.7195E+03, 7.7240E+03, 7.7284E+03,
     + 7.7329E+03, 7.7374E+03, 7.7419E+03, 7.7463E+03, 7.7508E+03,
     +  7.7553E+03, 7.7598E+03, 7.7642E+03, 7.7687E+03, 7.7732E+03,
     + 7.7777E+03, 7.7822E+03, 7.7867E+03, 7.7912E+03, 7.7957E+03,
     +  7.8001E+03, 7.8046E+03, 7.8091E+03, 7.8136E+03, 7.8181E+03,
     + 7.8226E+03, 7.8271E+03, 7.8316E+03, 7.8361E+03, 7.8406E+03,
     +  7.8451E+03, 7.8496E+03, 7.8541E+03, 7.8587E+03, 7.8632E+03,
     + 7.8677E+03, 7.8722E+03, 7.8767E+03, 7.8812E+03, 7.8857E+03,
     +  7.8903E+03, 7.8948E+03, 7.8993E+03, 7.9038E+03, 7.9084E+03,
     + 7.9129E+03, 7.9174E+03, 7.9219E+03, 7.9265E+03, 7.9310E+03,
     +  7.9355E+03, 7.9401E+03, 7.9446E+03, 7.9491E+03, 7.9537E+03,
     + 7.9582E+03, 7.9628E+03, 7.9673E+03, 7.9718E+03, 7.9764E+03,
     +  7.9809E+03, 7.9855E+03, 7.9900E+03, 7.9946E+03, 7.9991E+03,
     + 8.0037E+03, 8.0082E+03, 8.0128E+03, 8.0174E+03, 8.0219E+03,
     +  8.0265E+03, 8.0310E+03, 8.0356E+03, 8.0402E+03, 8.0447E+03,
     + 8.0493E+03, 8.0539E+03, 8.0584E+03, 8.0630E+03, 8.0676E+03,
     +  8.0722E+03, 8.0767E+03, 8.0813E+03, 8.0859E+03, 8.0905E+03,
     + 8.0951E+03, 8.0996E+03, 8.1042E+03, 8.1088E+03, 8.1134E+03,
     +  8.1180E+03, 8.1226E+03, 8.1272E+03, 8.1318E+03, 8.1363E+03,
     + 8.1409E+03, 8.1455E+03, 8.1501E+03, 8.1547E+03, 8.1593E+03,
     +  8.1639E+03, 8.1685E+03, 8.1731E+03, 8.1777E+03, 8.1824E+03,
     + 8.1870E+03, 8.1916E+03, 8.1962E+03, 8.2008E+03, 8.2054E+03,
     +  8.2100E+03, 8.2146E+03, 8.2193E+03, 8.2239E+03, 8.2285E+03,
     + 8.2331E+03, 8.2378E+03, 8.2424E+03, 8.2470E+03, 8.2516E+03,
     +  8.2563E+03, 8.2609E+03, 8.2655E+03, 8.2702E+03, 8.2748E+03,
     + 8.2794E+03, 8.2841E+03, 8.2887E+03, 8.2933E+03, 8.2980E+03,
     +  8.3026E+03, 8.3073E+03, 8.3119E+03, 8.3166E+03, 8.3212E+03,
     + 8.3259E+03, 8.3305E+03, 8.3352E+03, 8.3398E+03, 8.3445E+03,
     +  8.3491E+03, 8.3538E+03, 8.3585E+03, 8.3631E+03, 8.3678E+03,
     + 8.3725E+03, 8.3771E+03, 8.3818E+03, 8.3865E+03, 8.3911E+03,
     +  8.3958E+03, 8.4005E+03, 8.4051E+03, 8.4098E+03, 8.4145E+03,
     + 8.4192E+03, 8.4238E+03, 8.4285E+03, 8.4332E+03, 8.4379E+03,
     +  8.4426E+03, 8.4473E+03, 8.4520E+03, 8.4566E+03, 8.4613E+03,
     + 8.4660E+03, 8.4707E+03, 8.4754E+03, 8.4801E+03, 8.4848E+03,
     +  8.4895E+03, 8.4942E+03, 8.4989E+03, 8.5036E+03, 8.5083E+03,
     + 8.5130E+03, 8.5177E+03, 8.5224E+03, 8.5271E+03, 8.5319E+03,
     +  8.5366E+03, 8.5413E+03, 8.5460E+03, 8.5507E+03, 8.5554E+03,
     + 8.5601E+03, 8.5649E+03, 8.5696E+03, 8.5743E+03, 8.5790E+03,
     +  8.5838E+03, 8.5885E+03, 8.5932E+03, 8.5980E+03, 8.6027E+03,
     + 8.6074E+03, 8.6122E+03, 8.6169E+03, 8.6216E+03, 8.6264E+03,
     +  8.6311E+03, 8.6358E+03, 8.6406E+03, 8.6453E+03, 8.6501E+03,
     + 8.6548E+03, 8.6596E+03, 8.6643E+03, 8.6691E+03, 8.6738E+03,
     +  8.6786E+03, 8.6833E+03, 8.6881E+03, 8.6929E+03, 8.6976E+03,
     + 8.7024E+03, 8.7071E+03, 8.7119E+03, 8.7167E+03, 8.7214E+03,
     +  8.7262E+03, 8.7310E+03, 8.7357E+03, 8.7405E+03, 8.7453E+03,
     + 8.7501E+03, 8.7548E+03, 8.7596E+03, 8.7644E+03, 8.7692E+03,
     +  8.7740E+03, 8.7787E+03, 8.7835E+03, 8.7883E+03, 8.7931E+03,
     + 8.7979E+03, 8.8027E+03, 8.8075E+03, 8.8123E+03, 8.8171E+03,
     +  8.8219E+03, 8.8266E+03, 8.8314E+03, 8.8362E+03, 8.8410E+03,
     + 8.8458E+03, 8.8507E+03, 8.8555E+03, 8.8603E+03, 8.8651E+03,
     +  8.8699E+03, 8.8747E+03, 8.8795E+03, 8.8843E+03, 8.8891E+03,
     + 8.8940E+03, 8.8988E+03, 8.9036E+03, 8.9084E+03, 8.9132E+03,
     +  8.9181E+03, 8.9229E+03, 8.9277E+03, 8.9325E+03, 8.9374E+03,
     + 8.9422E+03, 8.9470E+03, 8.9519E+03, 8.9567E+03, 8.9615E+03,
     +  8.9664E+03, 8.9712E+03, 8.9760E+03, 8.9809E+03, 8.9857E+03,
     + 8.9906E+03, 8.9954E+03, 9.0003E+03, 9.0051E+03, 9.0100E+03,
     +  9.0148E+03, 9.0197E+03, 9.0245E+03, 9.0294E+03, 9.0342E+03,
     + 9.0391E+03, 9.0439E+03, 9.0488E+03, 9.0537E+03, 9.0585E+03,
     +  9.0634E+03, 9.0683E+03, 9.0731E+03, 9.0780E+03, 9.0829E+03,
     + 9.0877E+03, 9.0926E+03, 9.0975E+03, 9.1024E+03, 9.1072E+03,
     +  9.1121E+03, 9.1170E+03, 9.1219E+03, 9.1268E+03, 9.1317E+03,
     + 9.1365E+03, 9.1414E+03, 9.1463E+03, 9.1512E+03, 9.1561E+03,
     +  9.1610E+03, 9.1659E+03, 9.1708E+03, 9.1757E+03, 9.1806E+03,
     + 9.1855E+03, 9.1904E+03, 9.1953E+03, 9.2002E+03, 9.2051E+03,
     +  9.2100E+03, 9.2149E+03, 9.2198E+03, 9.2247E+03, 9.2296E+03,
     + 9.2345E+03, 9.2395E+03, 9.2444E+03, 9.2493E+03, 9.2542E+03,
     +  9.2591E+03, 9.2641E+03, 9.2690E+03, 9.2739E+03, 9.2788E+03,
     + 9.2838E+03, 9.2887E+03, 9.2936E+03, 9.2986E+03, 9.3035E+03,
     +  9.3084E+03, 9.3134E+03, 9.3183E+03, 9.3232E+03, 9.3282E+03,
     + 9.3331E+03, 9.3381E+03, 9.3430E+03, 9.3479E+03, 9.3529E+03,
     +  9.3578E+03, 9.3628E+03, 9.3677E+03, 9.3727E+03, 9.3776E+03,
     + 9.3826E+03, 9.3876E+03, 9.3925E+03, 9.3975E+03, 9.4024E+03,
     +  9.4074E+03, 9.4124E+03, 9.4173E+03, 9.4223E+03, 9.4273E+03,
     + 9.4322E+03, 9.4372E+03, 9.4422E+03, 9.4471E+03, 9.4521E+03,
     +  9.4571E+03, 9.4621E+03, 9.4671E+03, 9.4720E+03, 9.4770E+03,
     + 9.4820E+03, 9.4870E+03, 9.4920E+03, 9.4970E+03, 9.5019E+03,
     +  9.5069E+03, 9.5119E+03, 9.5169E+03, 9.5219E+03, 9.5269E+03,
     + 9.5319E+03, 9.5369E+03, 9.5419E+03, 9.5469E+03, 9.5519E+03,
     +  9.5569E+03, 9.5619E+03, 9.5669E+03, 9.5719E+03, 9.5769E+03,
     + 9.5819E+03, 9.5870E+03, 9.5920E+03, 9.5970E+03, 9.6020E+03,
     +  9.6070E+03, 9.6120E+03, 9.6171E+03, 9.6221E+03, 9.6271E+03,
     + 9.6321E+03, 9.6372E+03, 9.6422E+03, 9.6472E+03, 9.6522E+03,
     +  9.6573E+03, 9.6623E+03, 9.6673E+03, 9.6724E+03, 9.6774E+03,
     + 9.6825E+03, 9.6875E+03, 9.6925E+03, 9.6976E+03, 9.7026E+03,
     +  9.7077E+03, 9.7127E+03, 9.7178E+03, 9.7228E+03, 9.7279E+03,
     + 9.7329E+03, 9.7380E+03, 9.7430E+03, 9.7481E+03, 9.7531E+03,
     +  9.7582E+03, 9.7633E+03, 9.7683E+03, 9.7734E+03, 9.7785E+03,
     + 9.7835E+03, 9.7886E+03, 9.7937E+03, 9.7987E+03, 9.8038E+03,
     +  9.8089E+03, 9.8140E+03, 9.8190E+03, 9.8241E+03, 9.8292E+03,
     + 9.8343E+03, 9.8393E+03, 9.8444E+03, 9.8495E+03, 9.8546E+03,
     +  9.8597E+03, 9.8648E+03, 9.8699E+03, 9.8750E+03, 9.8800E+03,
     + 9.8851E+03, 9.8902E+03, 9.8953E+03, 9.9004E+03, 9.9055E+03,
     +  9.9106E+03, 9.9157E+03, 9.9208E+03, 9.9259E+03, 9.9311E+03,
     + 9.9362E+03, 9.9413E+03, 9.9464E+03, 9.9515E+03, 9.9566E+03,
     +  9.9617E+03, 9.9668E+03, 9.9720E+03, 9.9771E+03, 9.9822E+03,
     + 9.9873E+03, 9.9924E+03, 9.9976E+03, 1.0003E+04, 1.0008E+04,
     +  1.0013E+04, 1.0018E+04, 1.0023E+04, 1.0028E+04, 1.0033E+04,
     + 1.0039E+04, 1.0044E+04, 1.0049E+04, 1.0054E+04, 1.0059E+04,
     +  1.0064E+04, 1.0069E+04, 1.0075E+04, 1.0080E+04, 1.0085E+04,
     + 1.0090E+04, 1.0095E+04, 1.0100E+04, 1.0106E+04, 1.0111E+04,
     +  1.0116E+04, 1.0121E+04, 1.0126E+04, 1.0131E+04, 1.0136E+04,
     + 1.0142E+04, 1.0147E+04, 1.0152E+04, 1.0157E+04, 1.0162E+04,
     +  1.0167E+04, 1.0173E+04, 1.0178E+04, 1.0183E+04, 1.0188E+04,
     + 1.0193E+04, 1.0199E+04, 1.0204E+04, 1.0209E+04, 1.0214E+04,
     +  1.0219E+04, 1.0224E+04, 1.0230E+04, 1.0235E+04, 1.0240E+04,
     + 1.0245E+04, 1.0250E+04, 1.0256E+04, 1.0261E+04, 1.0266E+04,
     +  1.0271E+04, 1.0276E+04, 1.0282E+04, 1.0287E+04, 1.0292E+04,
     + 1.0297E+04, 1.0302E+04, 1.0308E+04, 1.0313E+04, 1.0318E+04,
     +  1.0323E+04, 1.0328E+04, 1.0334E+04, 1.0339E+04, 1.0344E+04,
     + 1.0349E+04, 1.0355E+04, 1.0360E+04, 1.0365E+04, 1.0370E+04,
     +  1.0375E+04, 1.0381E+04, 1.0386E+04, 1.0391E+04, 1.0396E+04,
     + 1.0402E+04, 1.0407E+04, 1.0412E+04, 1.0417E+04, 1.0423E+04,
     +  1.0428E+04, 1.0433E+04, 1.0438E+04, 1.0444E+04, 1.0449E+04,
     + 1.0454E+04, 1.0459E+04, 1.0465E+04, 1.0470E+04, 1.0475E+04,
     +  1.0480E+04, 1.0486E+04, 1.0491E+04, 1.0496E+04, 1.0501E+04,
     + 1.0507E+04, 1.0512E+04, 1.0517E+04, 1.0522E+04, 1.0528E+04,
     +  1.0533E+04, 1.0538E+04, 1.0543E+04, 1.0549E+04, 1.0554E+04,
     + 1.0559E+04, 1.0565E+04, 1.0570E+04, 1.0575E+04, 1.0580E+04,
     +  1.0586E+04, 1.0591E+04, 1.0596E+04, 1.0602E+04, 1.0607E+04,
     + 1.0612E+04, 1.0617E+04, 1.0623E+04, 1.0628E+04, 1.0633E+04,
     +  1.0639E+04, 1.0644E+04, 1.0649E+04, 1.0654E+04, 1.0660E+04,
     + 1.0665E+04, 1.0670E+04, 1.0676E+04, 1.0681E+04, 1.0686E+04,
     +  1.0692E+04, 1.0697E+04, 1.0702E+04, 1.0708E+04, 1.0713E+04,
     + 1.0718E+04, 1.0723E+04, 1.0729E+04, 1.0734E+04, 1.0739E+04,
     +  1.0745E+04, 1.0750E+04, 1.0755E+04, 1.0761E+04, 1.0766E+04,
     + 1.0771E+04, 1.0777E+04, 1.0782E+04, 1.0787E+04, 1.0793E+04,
     +  1.0798E+04, 1.0803E+04, 1.0809E+04, 1.0814E+04, 1.0819E+04,
     + 1.0825E+04, 1.0830E+04, 1.0836E+04, 1.0841E+04, 1.0846E+04,
     +  1.0852E+04, 1.0857E+04, 1.0862E+04, 1.0868E+04, 1.0873E+04,
     + 1.0878E+04, 1.0884E+04, 1.0889E+04, 1.0894E+04, 1.0900E+04,
     +  1.0905E+04, 1.0911E+04, 1.0916E+04, 1.0921E+04, 1.0927E+04,
     + 1.0932E+04, 1.0937E+04, 1.0943E+04, 1.0948E+04, 1.0954E+04,
     +  1.0959E+04, 1.0964E+04, 1.0970E+04, 1.0975E+04, 1.0980E+04,
     + 1.0986E+04, 1.0991E+04, 1.0997E+04, 1.1002E+04, 1.1007E+04,
     +  1.1013E+04, 1.1018E+04, 1.1024E+04, 1.1029E+04, 1.1034E+04,
     + 1.1040E+04, 1.1045E+04, 1.1051E+04, 1.1056E+04, 1.1061E+04,
     +  1.1067E+04, 1.1072E+04, 1.1078E+04, 1.1083E+04, 1.1088E+04,
     + 1.1094E+04, 1.1099E+04, 1.1105E+04, 1.1110E+04, 1.1116E+04,
     +  1.1121E+04, 1.1126E+04, 1.1132E+04, 1.1137E+04, 1.1143E+04,
     + 1.1148E+04, 1.1153E+04, 1.1159E+04, 1.1164E+04, 1.1170E+04,
     +  1.1175E+04, 1.1181E+04, 1.1186E+04, 1.1192E+04, 1.1197E+04,
     + 1.1202E+04, 1.1208E+04, 1.1213E+04, 1.1219E+04, 1.1224E+04,
     +  1.1230E+04, 1.1235E+04, 1.1241E+04, 1.1246E+04, 1.1251E+04,
     + 1.1257E+04, 1.1262E+04, 1.1268E+04, 1.1273E+04, 1.1279E+04,
     +  1.1284E+04, 1.1290E+04, 1.1295E+04, 1.1301E+04, 1.1306E+04,
     + 1.1312E+04, 1.1317E+04, 1.1323E+04, 1.1328E+04, 1.1333E+04,
     +  1.1339E+04, 1.1344E+04, 1.1350E+04, 1.1355E+04, 1.1361E+04,
     + 1.1366E+04, 1.1372E+04, 1.1377E+04, 1.1383E+04, 1.1388E+04,
     +  1.1394E+04, 1.1399E+04, 1.1405E+04, 1.1410E+04, 1.1416E+04,
     + 1.1421E+04, 1.1427E+04, 1.1432E+04, 1.1438E+04, 1.1443E+04,
     +  1.1449E+04, 1.1454E+04, 1.1460E+04, 1.1465E+04, 1.1471E+04,
     + 1.1476E+04, 1.1482E+04, 1.1487E+04, 1.1493E+04, 1.1498E+04,
     +  1.1504E+04, 1.1509E+04, 1.1515E+04, 1.1520E+04, 1.1526E+04,
     + 1.1532E+04, 1.1537E+04, 1.1543E+04, 1.1548E+04, 1.1554E+04,
     +  1.1559E+04, 1.1565E+04, 1.1570E+04, 1.1576E+04, 1.1581E+04,
     + 1.1587E+04, 1.1592E+04, 1.1598E+04, 1.1604E+04, 1.1609E+04,
     +  1.1615E+04, 1.1620E+04, 1.1626E+04, 1.1631E+04, 1.1637E+04,
     + 1.1642E+04, 1.1648E+04, 1.1653E+04, 1.1659E+04, 1.1665E+04,
     +  1.1670E+04, 1.1676E+04, 1.1681E+04, 1.1687E+04, 1.1692E+04,
     + 1.1698E+04, 1.1704E+04, 1.1709E+04, 1.1715E+04, 1.1720E+04,
     +  1.1726E+04, 1.1731E+04, 1.1737E+04, 1.1743E+04, 1.1748E+04,
     + 1.1754E+04, 1.1759E+04, 1.1765E+04, 1.1770E+04, 1.1776E+04,
     +  1.1782E+04, 1.1787E+04, 1.1793E+04, 1.1798E+04, 1.1804E+04,
     + 1.1810E+04, 1.1815E+04, 1.1821E+04, 1.1826E+04, 1.1832E+04,
     +  1.1838E+04, 1.1843E+04, 1.1849E+04, 1.1854E+04, 1.1860E+04,
     + 1.1866E+04, 1.1871E+04, 1.1877E+04, 1.1883E+04, 1.1888E+04,
     +  1.1894E+04, 1.1899E+04, 1.1905E+04, 1.1911E+04, 1.1916E+04,
     + 1.1922E+04, 1.1927E+04, 1.1933E+04, 1.1939E+04, 1.1944E+04,
     +  1.1950E+04, 1.1956E+04, 1.1961E+04, 1.1967E+04, 1.1973E+04,
     + 1.1978E+04, 1.1984E+04, 1.1989E+04, 1.1995E+04, 1.2001E+04,
     +  1.2006E+04, 1.2012E+04, 1.2018E+04, 1.2023E+04, 1.2029E+04,
     + 1.2035E+04, 1.2040E+04, 1.2046E+04, 1.2052E+04, 1.2057E+04,
     +  1.2063E+04, 1.2069E+04, 1.2074E+04, 1.2080E+04, 1.2086E+04,
     + 1.2091E+04, 1.2097E+04, 1.2103E+04, 1.2108E+04, 1.2114E+04,
     +  1.2120E+04, 1.2125E+04, 1.2131E+04, 1.2137E+04, 1.2142E+04,
     + 1.2148E+04, 1.2154E+04, 1.2159E+04, 1.2165E+04, 1.2171E+04,
     +  1.2176E+04, 1.2182E+04, 1.2188E+04, 1.2193E+04, 1.2199E+04,
     + 1.2205E+04, 1.2211E+04, 1.2216E+04, 1.2222E+04, 1.2228E+04,
     +  1.2233E+04, 1.2239E+04, 1.2245E+04, 1.2250E+04, 1.2256E+04,
     + 1.2262E+04, 1.2268E+04, 1.2273E+04, 1.2279E+04, 1.2285E+04,
     +  1.2290E+04, 1.2296E+04, 1.2302E+04, 1.2308E+04, 1.2313E+04,
     + 1.2319E+04, 1.2325E+04, 1.2330E+04, 1.2336E+04, 1.2342E+04,
     +  1.2348E+04, 1.2353E+04, 1.2359E+04, 1.2365E+04, 1.2371E+04,
     + 1.2376E+04, 1.2382E+04, 1.2388E+04, 1.2394E+04, 1.2399E+04,
     +  1.2405E+04, 1.2411E+04, 1.2416E+04, 1.2422E+04, 1.2428E+04,
     + 1.2434E+04, 1.2439E+04, 1.2445E+04, 1.2451E+04, 1.2457E+04,
     +  1.2463E+04, 1.2468E+04, 1.2474E+04, 1.2480E+04, 1.2486E+04,
     + 1.2491E+04, 1.2497E+04, 1.2503E+04, 1.2509E+04, 1.2514E+04,
     +  1.2520E+04, 1.2526E+04, 1.2532E+04, 1.2537E+04, 1.2543E+04,
     + 1.2549E+04, 1.2555E+04, 1.2561E+04, 1.2566E+04, 1.2572E+04,
     +  1.2578E+04, 1.2584E+04, 1.2589E+04, 1.2595E+04, 1.2601E+04,
     + 1.2607E+04, 1.2613E+04, 1.2618E+04, 1.2624E+04, 1.2630E+04,
     +  1.2636E+04, 1.2642E+04, 1.2647E+04, 1.2653E+04, 1.2659E+04,
     + 1.2665E+04, 1.2671E+04, 1.2676E+04, 1.2682E+04, 1.2688E+04,
     +  1.2694E+04, 1.2700E+04, 1.2706E+04, 1.2711E+04, 1.2717E+04,
     + 1.2723E+04, 1.2729E+04, 1.2735E+04, 1.2740E+04, 1.2746E+04,
     +  1.2752E+04, 1.2758E+04, 1.2764E+04, 1.2770E+04, 1.2775E+04,
     + 1.2781E+04, 1.2787E+04, 1.2793E+04, 1.2799E+04, 1.2805E+04,
     +  1.2810E+04, 1.2816E+04, 1.2822E+04, 1.2828E+04, 1.2834E+04,
     + 1.2840E+04, 1.2846E+04, 1.2851E+04, 1.2857E+04, 1.2863E+04,
     +  1.2869E+04, 1.2875E+04, 1.2881E+04, 1.2886E+04, 1.2892E+04,
     + 1.2898E+04, 1.2904E+04, 1.2910E+04, 1.2916E+04, 1.2922E+04,
     +  1.2928E+04, 1.2933E+04, 1.2939E+04, 1.2945E+04, 1.2951E+04,
     + 1.2957E+04, 1.2963E+04, 1.2969E+04, 1.2975E+04, 1.2980E+04,
     +  1.2986E+04, 1.2992E+04, 1.2998E+04, 1.3004E+04, 1.3010E+04,
     + 1.3016E+04, 1.3022E+04, 1.3028E+04, 1.3033E+04, 1.3039E+04,
     +  1.3045E+04, 1.3051E+04, 1.3057E+04, 1.3063E+04, 1.3069E+04,
     + 1.3075E+04, 1.3081E+04, 1.3087E+04, 1.3092E+04, 1.3098E+04,
     +  1.3104E+04, 1.3110E+04, 1.3116E+04, 1.3122E+04, 1.3128E+04,
     + 1.3134E+04, 1.3140E+04, 1.3146E+04, 1.3152E+04, 1.3157E+04,
     +  1.3163E+04, 1.3169E+04, 1.3175E+04, 1.3181E+04, 1.3187E+04,
     + 1.3193E+04, 1.3199E+04, 1.3205E+04, 1.3211E+04, 1.3217E+04,
     +  1.3223E+04, 1.3229E+04, 1.3235E+04, 1.3241E+04, 1.3246E+04,
     + 1.3252E+04, 1.3258E+04, 1.3264E+04, 1.3270E+04, 1.3276E+04,
     +  1.3282E+04, 1.3288E+04, 1.3294E+04, 1.3300E+04, 1.3306E+04,
     + 1.3312E+04, 1.3318E+04, 1.3324E+04, 1.3330E+04, 1.3336E+04,
     +  1.3342E+04, 1.3348E+04, 1.3354E+04, 1.3360E+04, 1.3366E+04,
     + 1.3372E+04, 1.3378E+04, 1.3384E+04, 1.3390E+04, 1.3396E+04,
     +  1.3401E+04, 1.3407E+04, 1.3413E+04, 1.3419E+04, 1.3425E+04,
     + 1.3431E+04, 1.3437E+04, 1.3443E+04, 1.3449E+04, 1.3455E+04,
     +  1.3461E+04, 1.3467E+04, 1.3473E+04, 1.3479E+04, 1.3485E+04,
     + 1.3491E+04, 1.3497E+04, 1.3503E+04, 1.3509E+04, 1.3515E+04,
     +  1.3521E+04, 1.3527E+04, 1.3533E+04, 1.3539E+04, 1.3545E+04,
     + 1.3551E+04, 1.3557E+04, 1.3563E+04, 1.3569E+04, 1.3576E+04,
     +  1.3582E+04, 1.3588E+04, 1.3594E+04, 1.3600E+04, 1.3606E+04,
     + 1.3612E+04, 1.3618E+04, 1.3624E+04, 1.3630E+04, 1.3636E+04,
     +  1.3642E+04, 1.3648E+04, 1.3654E+04, 1.3660E+04, 1.3666E+04,
     + 1.3672E+04, 1.3678E+04, 1.3684E+04, 1.3690E+04, 1.3696E+04,
     +  1.3702E+04, 1.3708E+04, 1.3714E+04, 1.3720E+04, 1.3726E+04,
     + 1.3733E+04, 1.3739E+04, 1.3745E+04, 1.3751E+04, 1.3757E+04,
     +  1.3763E+04, 1.3769E+04, 1.3775E+04, 1.3781E+04, 1.3787E+04,
     + 1.3793E+04, 1.3799E+04, 1.3805E+04, 1.3811E+04, 1.3817E+04,
     +  1.3824E+04, 1.3830E+04, 1.3836E+04, 1.3842E+04, 1.3848E+04,
     + 1.3854E+04, 1.3860E+04, 1.3866E+04, 1.3872E+04, 1.3878E+04,
     +  1.3884E+04, 1.3890E+04, 1.3897E+04, 1.3903E+04, 1.3909E+04,
     + 1.3915E+04, 1.3921E+04, 1.3927E+04, 1.3933E+04, 1.3939E+04,
     +  1.3945E+04, 1.3951E+04, 1.3958E+04, 1.3964E+04, 1.3970E+04,
     + 1.3976E+04, 1.3982E+04, 1.3988E+04, 1.3994E+04, 1.4000E+04,
     +  1.4006E+04, 1.4013E+04, 1.4019E+04, 1.4025E+04, 1.4031E+04,
     + 1.4037E+04, 1.4043E+04, 1.4049E+04, 1.4055E+04, 1.4062E+04,
     +  1.4068E+04, 1.4074E+04, 1.4080E+04, 1.4086E+04, 1.4092E+04,
     + 1.4098E+04, 1.4105E+04, 1.4111E+04, 1.4117E+04, 1.4123E+04,
     +  1.4129E+04, 1.4135E+04, 1.4141E+04, 1.4148E+04, 1.4154E+04,
     + 1.4160E+04, 1.4166E+04, 1.4172E+04, 1.4178E+04, 1.4185E+04,
     +  1.4191E+04, 1.4197E+04, 1.4203E+04, 1.4209E+04, 1.4215E+04,
     + 1.4221E+04, 1.4228E+04, 1.4234E+04, 1.4240E+04, 1.4246E+04,
     +  1.4252E+04, 1.4259E+04, 1.4265E+04, 1.4271E+04, 1.4277E+04,
     + 1.4283E+04, 1.4289E+04, 1.4296E+04, 1.4302E+04, 1.4308E+04,
     +  1.4314E+04, 1.4320E+04, 1.4327E+04, 1.4333E+04, 1.4339E+04,
     + 1.4345E+04, 1.4351E+04, 1.4358E+04, 1.4364E+04, 1.4370E+04,
     +  1.4376E+04, 1.4382E+04, 1.4389E+04, 1.4395E+04, 1.4401E+04,
     + 1.4407E+04, 1.4413E+04, 1.4420E+04, 1.4426E+04, 1.4432E+04,
     +  1.4438E+04, 1.4444E+04, 1.4451E+04, 1.4457E+04, 1.4463E+04,
     + 1.4469E+04, 1.4476E+04, 1.4482E+04, 1.4488E+04, 1.4494E+04,
     +  1.4500E+04, 1.4507E+04, 1.4513E+04, 1.4519E+04, 1.4525E+04,
     + 1.4532E+04, 1.4538E+04, 1.4544E+04, 1.4550E+04, 1.4557E+04,
     +  1.4563E+04, 1.4569E+04, 1.4575E+04, 1.4582E+04, 1.4588E+04,
     + 1.4594E+04, 1.4600E+04, 1.4607E+04, 1.4613E+04, 1.4619E+04,
     +  1.4625E+04, 1.4632E+04, 1.4638E+04, 1.4644E+04, 1.4650E+04,
     + 1.4657E+04, 1.4663E+04, 1.4669E+04, 1.4675E+04/
c...        --        47
      data (QofT( 9,J),J=1,2408)/ 6.1286E+00, 7.5397E+00, 9.6835E+00,
     +  1.1996E+01, 1.4362E+01, 1.6751E+01, 1.9152E+01, 2.1560E+01,
     + 2.3973E+01, 2.6390E+01, 2.8808E+01, 3.1229E+01, 3.3651E+01,
     +  3.6073E+01, 3.8497E+01, 4.0921E+01, 4.3346E+01, 4.5772E+01,
     + 4.8197E+01, 5.0623E+01, 5.5476E+01, 6.0330E+01, 6.5184E+01,
     +  7.0039E+01, 7.4894E+01, 7.9750E+01, 8.4606E+01, 8.9462E+01,
     + 9.4318E+01, 9.9175E+01, 1.0403E+02, 1.0889E+02, 1.1375E+02,
     +  1.1860E+02, 1.2346E+02, 1.2832E+02, 1.3317E+02, 1.3803E+02,
     + 1.4289E+02, 1.4775E+02, 1.5261E+02, 1.5746E+02, 1.6232E+02,
     +  1.6718E+02, 1.7204E+02, 1.7690E+02, 1.8175E+02, 1.8661E+02,
     + 1.9147E+02, 1.9633E+02, 2.0119E+02, 2.0605E+02, 2.1091E+02,
     +  2.1576E+02, 2.2062E+02, 2.2548E+02, 2.3034E+02, 2.3520E+02,
     + 2.4006E+02, 2.4492E+02, 2.4978E+02, 2.5464E+02, 2.5950E+02,
     +  2.6435E+02, 2.6921E+02, 2.7407E+02, 2.7893E+02, 2.8379E+02,
     + 2.8865E+02, 2.9351E+02, 2.9837E+02, 3.0323E+02, 3.0809E+02,
     +  3.1295E+02, 3.1781E+02, 3.2267E+02, 3.2753E+02, 3.3239E+02,
     + 3.3725E+02, 3.4211E+02, 3.4697E+02, 3.5183E+02, 3.5669E+02,
     +  3.6155E+02, 3.6641E+02, 3.7127E+02, 3.7613E+02, 3.8099E+02,
     + 3.8585E+02, 3.9071E+02, 3.9557E+02, 4.0043E+02, 4.0529E+02,
     +  4.1016E+02, 4.1502E+02, 4.1988E+02, 4.2474E+02, 4.2960E+02,
     + 4.3446E+02, 4.3932E+02, 4.4418E+02, 4.4904E+02, 4.5390E+02,
     +  4.5877E+02, 4.6363E+02, 4.6849E+02, 4.7335E+02, 4.7821E+02,
     + 4.8307E+02, 4.8793E+02, 4.9279E+02, 4.9766E+02, 5.0252E+02,
     +  5.0738E+02, 5.1224E+02, 5.1710E+02, 5.2197E+02, 5.2683E+02,
     + 5.3169E+02, 5.3655E+02, 5.4141E+02, 5.4628E+02, 5.5114E+02,
     +  5.5600E+02, 5.6086E+02, 5.6572E+02, 5.7059E+02, 5.7545E+02,
     + 5.8031E+02, 5.8517E+02, 5.9004E+02, 5.9490E+02, 5.9976E+02,
     +  6.0463E+02, 6.0949E+02, 6.1435E+02, 6.1922E+02, 6.2408E+02,
     + 6.2894E+02, 6.3381E+02, 6.3867E+02, 6.4353E+02, 6.4840E+02,
     +  6.5326E+02, 6.5813E+02, 6.6299E+02, 6.6785E+02, 6.7272E+02,
     + 6.7758E+02, 6.8245E+02, 6.8731E+02, 6.9218E+02, 6.9704E+02,
     +  7.0191E+02, 7.0678E+02, 7.1164E+02, 7.1651E+02, 7.2137E+02,
     + 7.2624E+02, 7.3111E+02, 7.3598E+02, 7.4084E+02, 7.4571E+02,
     +  7.5058E+02, 7.5545E+02, 7.6031E+02, 7.6518E+02, 7.7005E+02,
     + 7.7492E+02, 7.7979E+02, 7.8466E+02, 7.8953E+02, 7.9440E+02,
     +  7.9927E+02, 8.0414E+02, 8.0901E+02, 8.1389E+02, 8.1876E+02,
     + 8.2363E+02, 8.2851E+02, 8.3338E+02, 8.3825E+02, 8.4313E+02,
     +  8.4800E+02, 8.5288E+02, 8.5775E+02, 8.6263E+02, 8.6751E+02,
     + 8.7239E+02, 8.7727E+02, 8.8214E+02, 8.8702E+02, 8.9190E+02,
     +  8.9678E+02, 9.0167E+02, 9.0655E+02, 9.1143E+02, 9.1631E+02,
     + 9.2120E+02, 9.2608E+02, 9.3097E+02, 9.3586E+02, 9.4074E+02,
     +  9.4563E+02, 9.5052E+02, 9.5541E+02, 9.6030E+02, 9.6519E+02,
     + 9.7009E+02, 9.7498E+02, 9.7987E+02, 9.8477E+02, 9.8966E+02,
     +  9.9456E+02, 9.9946E+02, 1.0044E+03, 1.0093E+03, 1.0142E+03,
     + 1.0191E+03, 1.0240E+03, 1.0289E+03, 1.0338E+03, 1.0387E+03,
     +  1.0436E+03, 1.0485E+03, 1.0534E+03, 1.0583E+03, 1.0632E+03,
     + 1.0682E+03, 1.0731E+03, 1.0780E+03, 1.0829E+03, 1.0878E+03,
     +  1.0927E+03, 1.0977E+03, 1.1026E+03, 1.1075E+03, 1.1124E+03,
     + 1.1174E+03, 1.1223E+03, 1.1272E+03, 1.1322E+03, 1.1371E+03,
     +  1.1420E+03, 1.1470E+03, 1.1519E+03, 1.1569E+03, 1.1618E+03,
     + 1.1668E+03, 1.1717E+03, 1.1767E+03, 1.1816E+03, 1.1866E+03,
     +  1.1915E+03, 1.1965E+03, 1.2014E+03, 1.2064E+03, 1.2113E+03,
     + 1.2163E+03, 1.2213E+03, 1.2262E+03, 1.2312E+03, 1.2362E+03,
     +  1.2412E+03, 1.2461E+03, 1.2511E+03, 1.2561E+03, 1.2611E+03,
     + 1.2661E+03, 1.2711E+03, 1.2760E+03, 1.2810E+03, 1.2860E+03,
     +  1.2910E+03, 1.2960E+03, 1.3010E+03, 1.3060E+03, 1.3110E+03,
     + 1.3160E+03, 1.3211E+03, 1.3261E+03, 1.3311E+03, 1.3361E+03,
     +  1.3411E+03, 1.3462E+03, 1.3512E+03, 1.3562E+03, 1.3612E+03,
     + 1.3663E+03, 1.3713E+03, 1.3764E+03, 1.3814E+03, 1.3864E+03,
     +  1.3915E+03, 1.3965E+03, 1.4016E+03, 1.4066E+03, 1.4117E+03,
     + 1.4168E+03, 1.4218E+03, 1.4269E+03, 1.4320E+03, 1.4370E+03,
     +  1.4421E+03, 1.4472E+03, 1.4523E+03, 1.4574E+03, 1.4625E+03,
     + 1.4676E+03, 1.4727E+03, 1.4778E+03, 1.4829E+03, 1.4880E+03,
     +  1.4931E+03, 1.4982E+03, 1.5033E+03, 1.5084E+03, 1.5135E+03,
     + 1.5187E+03, 1.5238E+03, 1.5289E+03, 1.5341E+03, 1.5392E+03,
     +  1.5443E+03, 1.5495E+03, 1.5546E+03, 1.5598E+03, 1.5649E+03,
     + 1.5701E+03, 1.5753E+03, 1.5804E+03, 1.5856E+03, 1.5908E+03,
     +  1.5959E+03, 1.6011E+03, 1.6063E+03, 1.6115E+03, 1.6167E+03,
     + 1.6219E+03, 1.6271E+03, 1.6323E+03, 1.6375E+03, 1.6427E+03,
     +  1.6479E+03, 1.6531E+03, 1.6583E+03, 1.6636E+03, 1.6688E+03,
     + 1.6740E+03, 1.6793E+03, 1.6845E+03, 1.6897E+03, 1.6950E+03,
     +  1.7003E+03, 1.7055E+03, 1.7108E+03, 1.7160E+03, 1.7213E+03,
     + 1.7266E+03, 1.7318E+03, 1.7371E+03, 1.7424E+03, 1.7477E+03,
     +  1.7530E+03, 1.7583E+03, 1.7636E+03, 1.7689E+03, 1.7742E+03,
     + 1.7795E+03, 1.7848E+03, 1.7902E+03, 1.7955E+03, 1.8008E+03,
     +  1.8062E+03, 1.8115E+03, 1.8168E+03, 1.8222E+03, 1.8275E+03,
     + 1.8329E+03, 1.8383E+03, 1.8436E+03, 1.8490E+03, 1.8544E+03,
     +  1.8597E+03, 1.8651E+03, 1.8705E+03, 1.8759E+03, 1.8813E+03,
     + 1.8867E+03, 1.8921E+03, 1.8975E+03, 1.9029E+03, 1.9084E+03,
     +  1.9138E+03, 1.9192E+03, 1.9246E+03, 1.9301E+03, 1.9355E+03,
     + 1.9410E+03, 1.9464E+03, 1.9519E+03, 1.9573E+03, 1.9628E+03,
     +  1.9683E+03, 1.9738E+03, 1.9792E+03, 1.9847E+03, 1.9902E+03,
     + 1.9957E+03, 2.0012E+03, 2.0067E+03, 2.0122E+03, 2.0177E+03,
     +  2.0233E+03, 2.0288E+03, 2.0343E+03, 2.0398E+03, 2.0454E+03,
     + 2.0509E+03, 2.0565E+03, 2.0620E+03, 2.0676E+03, 2.0731E+03,
     +  2.0787E+03, 2.0843E+03, 2.0899E+03, 2.0954E+03, 2.1010E+03,
     + 2.1066E+03, 2.1122E+03, 2.1178E+03, 2.1234E+03, 2.1290E+03,
     +  2.1347E+03, 2.1403E+03, 2.1459E+03, 2.1515E+03, 2.1572E+03,
     + 2.1628E+03, 2.1685E+03, 2.1741E+03, 2.1798E+03, 2.1855E+03,
     +  2.1911E+03, 2.1968E+03, 2.2025E+03, 2.2082E+03, 2.2139E+03,
     + 2.2195E+03, 2.2252E+03, 2.2310E+03, 2.2367E+03, 2.2424E+03,
     +  2.2481E+03, 2.2538E+03, 2.2596E+03, 2.2653E+03, 2.2710E+03,
     + 2.2768E+03, 2.2825E+03, 2.2883E+03, 2.2941E+03, 2.2998E+03,
     +  2.3056E+03, 2.3114E+03, 2.3172E+03, 2.3230E+03, 2.3288E+03,
     + 2.3346E+03, 2.3404E+03, 2.3462E+03, 2.3520E+03, 2.3578E+03,
     +  2.3637E+03, 2.3695E+03, 2.3753E+03, 2.3812E+03, 2.3870E+03,
     + 2.3929E+03, 2.3987E+03, 2.4046E+03, 2.4105E+03, 2.4164E+03,
     +  2.4222E+03, 2.4281E+03, 2.4340E+03, 2.4399E+03, 2.4458E+03,
     + 2.4518E+03, 2.4577E+03, 2.4636E+03, 2.4695E+03, 2.4755E+03,
     +  2.4814E+03, 2.4873E+03, 2.4933E+03, 2.4993E+03, 2.5052E+03,
     + 2.5112E+03, 2.5172E+03, 2.5231E+03, 2.5291E+03, 2.5351E+03,
     +  2.5411E+03, 2.5471E+03, 2.5531E+03, 2.5591E+03, 2.5652E+03,
     + 2.5712E+03, 2.5772E+03, 2.5832E+03, 2.5893E+03, 2.5953E+03,
     +  2.6014E+03, 2.6074E+03, 2.6135E+03, 2.6196E+03, 2.6257E+03,
     + 2.6317E+03, 2.6378E+03, 2.6439E+03, 2.6500E+03, 2.6561E+03,
     +  2.6622E+03, 2.6684E+03, 2.6745E+03, 2.6806E+03, 2.6867E+03,
     + 2.6929E+03, 2.6990E+03, 2.7052E+03, 2.7113E+03, 2.7175E+03,
     +  2.7237E+03, 2.7299E+03, 2.7360E+03, 2.7422E+03, 2.7484E+03,
     + 2.7546E+03, 2.7608E+03, 2.7670E+03, 2.7733E+03, 2.7795E+03,
     +  2.7857E+03, 2.7919E+03, 2.7982E+03, 2.8044E+03, 2.8107E+03,
     + 2.8169E+03, 2.8232E+03, 2.8295E+03, 2.8358E+03, 2.8420E+03,
     +  2.8483E+03, 2.8546E+03, 2.8609E+03, 2.8672E+03, 2.8736E+03,
     + 2.8799E+03, 2.8862E+03, 2.8925E+03, 2.8989E+03, 2.9052E+03,
     +  2.9116E+03, 2.9179E+03, 2.9243E+03, 2.9306E+03, 2.9370E+03,
     + 2.9434E+03, 2.9498E+03, 2.9562E+03, 2.9626E+03, 2.9690E+03,
     +  2.9754E+03, 2.9818E+03, 2.9882E+03, 2.9946E+03, 3.0011E+03,
     + 3.0075E+03, 3.0140E+03, 3.0204E+03, 3.0269E+03, 3.0333E+03,
     +  3.0398E+03, 3.0463E+03, 3.0528E+03, 3.0593E+03, 3.0658E+03,
     + 3.0723E+03, 3.0788E+03, 3.0853E+03, 3.0918E+03, 3.0983E+03,
     +  3.1049E+03, 3.1114E+03, 3.1180E+03, 3.1245E+03, 3.1311E+03,
     + 3.1376E+03, 3.1442E+03, 3.1508E+03, 3.1574E+03, 3.1639E+03,
     +  3.1705E+03, 3.1771E+03, 3.1837E+03, 3.1904E+03, 3.1970E+03,
     + 3.2036E+03, 3.2102E+03, 3.2169E+03, 3.2235E+03, 3.2302E+03,
     +  3.2368E+03, 3.2435E+03, 3.2502E+03, 3.2568E+03, 3.2635E+03,
     + 3.2702E+03, 3.2769E+03, 3.2836E+03, 3.2903E+03, 3.2970E+03,
     +  3.3037E+03, 3.3105E+03, 3.3172E+03, 3.3239E+03, 3.3307E+03,
     + 3.3374E+03, 3.3442E+03, 3.3510E+03, 3.3577E+03, 3.3645E+03,
     +  3.3713E+03, 3.3781E+03, 3.3849E+03, 3.3917E+03, 3.3985E+03,
     + 3.4053E+03, 3.4121E+03, 3.4189E+03, 3.4258E+03, 3.4326E+03,
     +  3.4395E+03, 3.4463E+03, 3.4532E+03, 3.4600E+03, 3.4669E+03,
     + 3.4738E+03, 3.4807E+03, 3.4876E+03, 3.4945E+03, 3.5014E+03,
     +  3.5083E+03, 3.5152E+03, 3.5221E+03, 3.5291E+03, 3.5360E+03,
     + 3.5429E+03, 3.5499E+03, 3.5568E+03, 3.5638E+03, 3.5708E+03,
     +  3.5777E+03, 3.5847E+03, 3.5917E+03, 3.5987E+03, 3.6057E+03,
     + 3.6127E+03, 3.6197E+03, 3.6267E+03, 3.6338E+03, 3.6408E+03,
     +  3.6478E+03, 3.6549E+03, 3.6619E+03, 3.6690E+03, 3.6760E+03,
     + 3.6831E+03, 3.6902E+03, 3.6973E+03, 3.7044E+03, 3.7115E+03,
     +  3.7186E+03, 3.7257E+03, 3.7328E+03, 3.7399E+03, 3.7470E+03,
     + 3.7542E+03, 3.7613E+03, 3.7685E+03, 3.7756E+03, 3.7828E+03,
     +  3.7900E+03, 3.7971E+03, 3.8043E+03, 3.8115E+03, 3.8187E+03,
     + 3.8259E+03, 3.8331E+03, 3.8403E+03, 3.8475E+03, 3.8548E+03,
     +  3.8620E+03, 3.8692E+03, 3.8765E+03, 3.8837E+03, 3.8910E+03,
     + 3.8983E+03, 3.9055E+03, 3.9128E+03, 3.9201E+03, 3.9274E+03,
     +  3.9347E+03, 3.9420E+03, 3.9493E+03, 3.9566E+03, 3.9639E+03,
     + 3.9713E+03, 3.9786E+03, 3.9860E+03, 3.9933E+03, 4.0007E+03,
     +  4.0080E+03, 4.0154E+03, 4.0228E+03, 4.0302E+03, 4.0375E+03,
     + 4.0449E+03, 4.0523E+03, 4.0598E+03, 4.0672E+03, 4.0746E+03,
     +  4.0820E+03, 4.0895E+03, 4.0969E+03, 4.1044E+03, 4.1118E+03,
     + 4.1193E+03, 4.1267E+03, 4.1342E+03, 4.1417E+03, 4.1492E+03,
     +  4.1567E+03, 4.1642E+03, 4.1717E+03, 4.1792E+03, 4.1867E+03,
     + 4.1943E+03, 4.2018E+03, 4.2093E+03, 4.2169E+03, 4.2244E+03,
     +  4.2320E+03, 4.2396E+03, 4.2471E+03, 4.2547E+03, 4.2623E+03,
     + 4.2699E+03, 4.2775E+03, 4.2851E+03, 4.2927E+03, 4.3004E+03,
     +  4.3080E+03, 4.3156E+03, 4.3233E+03, 4.3309E+03, 4.3386E+03,
     + 4.3462E+03, 4.3539E+03, 4.3616E+03, 4.3692E+03, 4.3769E+03,
     +  4.3846E+03, 4.3923E+03, 4.4000E+03, 4.4078E+03, 4.4155E+03,
     + 4.4232E+03, 4.4309E+03, 4.4387E+03, 4.4464E+03, 4.4542E+03,
     +  4.4619E+03, 4.4697E+03, 4.4775E+03, 4.4853E+03, 4.4930E+03,
     + 4.5008E+03, 4.5086E+03, 4.5165E+03, 4.5243E+03, 4.5321E+03,
     +  4.5399E+03, 4.5478E+03, 4.5556E+03, 4.5634E+03, 4.5713E+03,
     + 4.5792E+03, 4.5870E+03, 4.5949E+03, 4.6028E+03, 4.6107E+03,
     +  4.6186E+03, 4.6265E+03, 4.6344E+03, 4.6423E+03, 4.6502E+03,
     + 4.6581E+03, 4.6661E+03, 4.6740E+03, 4.6820E+03, 4.6899E+03,
     +  4.6979E+03, 4.7058E+03, 4.7138E+03, 4.7218E+03, 4.7298E+03,
     + 4.7378E+03, 4.7458E+03, 4.7538E+03, 4.7618E+03, 4.7698E+03,
     +  4.7779E+03, 4.7859E+03, 4.7939E+03, 4.8020E+03, 4.8101E+03,
     + 4.8181E+03, 4.8262E+03, 4.8343E+03, 4.8423E+03, 4.8504E+03,
     +  4.8585E+03, 4.8666E+03, 4.8747E+03, 4.8829E+03, 4.8910E+03,
     + 4.8991E+03, 4.9073E+03, 4.9154E+03, 4.9235E+03, 4.9317E+03,
     +  4.9399E+03, 4.9480E+03, 4.9562E+03, 4.9644E+03, 4.9726E+03,
     + 4.9808E+03, 4.9890E+03, 4.9972E+03, 5.0054E+03, 5.0137E+03,
     +  5.0219E+03, 5.0301E+03, 5.0384E+03, 5.0466E+03, 5.0549E+03,
     + 5.0631E+03, 5.0714E+03, 5.0797E+03, 5.0880E+03, 5.0963E+03,
     +  5.1046E+03, 5.1129E+03, 5.1212E+03, 5.1295E+03, 5.1378E+03,
     + 5.1462E+03, 5.1545E+03, 5.1629E+03, 5.1712E+03, 5.1796E+03,
     +  5.1879E+03, 5.1963E+03, 5.2047E+03, 5.2131E+03, 5.2215E+03,
     + 5.2299E+03, 5.2383E+03, 5.2467E+03, 5.2551E+03, 5.2636E+03,
     +  5.2720E+03, 5.2804E+03, 5.2889E+03, 5.2973E+03, 5.3058E+03,
     + 5.3143E+03, 5.3228E+03, 5.3312E+03, 5.3397E+03, 5.3482E+03,
     +  5.3567E+03, 5.3652E+03, 5.3738E+03, 5.3823E+03, 5.3908E+03,
     + 5.3993E+03, 5.4079E+03, 5.4164E+03, 5.4250E+03, 5.4336E+03,
     +  5.4421E+03, 5.4507E+03, 5.4593E+03, 5.4679E+03, 5.4765E+03,
     + 5.4851E+03, 5.4937E+03, 5.5023E+03, 5.5110E+03, 5.5196E+03,
     +  5.5282E+03, 5.5369E+03, 5.5455E+03, 5.5542E+03, 5.5629E+03,
     + 5.5715E+03, 5.5802E+03, 5.5889E+03, 5.5976E+03, 5.6063E+03,
     +  5.6150E+03, 5.6237E+03, 5.6325E+03, 5.6412E+03, 5.6499E+03,
     + 5.6587E+03, 5.6674E+03, 5.6762E+03, 5.6849E+03, 5.6937E+03,
     +  5.7025E+03, 5.7113E+03, 5.7201E+03, 5.7288E+03, 5.7377E+03,
     + 5.7465E+03, 5.7553E+03, 5.7641E+03, 5.7729E+03, 5.7818E+03,
     +  5.7906E+03, 5.7995E+03, 5.8083E+03, 5.8172E+03, 5.8261E+03,
     + 5.8350E+03, 5.8438E+03, 5.8527E+03, 5.8616E+03, 5.8706E+03,
     +  5.8795E+03, 5.8884E+03, 5.8973E+03, 5.9063E+03, 5.9152E+03,
     + 5.9241E+03, 5.9331E+03, 5.9421E+03, 5.9510E+03, 5.9600E+03,
     +  5.9690E+03, 5.9780E+03, 5.9870E+03, 5.9960E+03, 6.0050E+03,
     + 6.0140E+03, 6.0230E+03, 6.0321E+03, 6.0411E+03, 6.0502E+03,
     +  6.0592E+03, 6.0683E+03, 6.0773E+03, 6.0864E+03, 6.0955E+03,
     + 6.1046E+03, 6.1137E+03, 6.1228E+03, 6.1319E+03, 6.1410E+03,
     +  6.1501E+03, 6.1592E+03, 6.1684E+03, 6.1775E+03, 6.1867E+03,
     + 6.1958E+03, 6.2050E+03, 6.2142E+03, 6.2233E+03, 6.2325E+03,
     +  6.2417E+03, 6.2509E+03, 6.2601E+03, 6.2693E+03, 6.2785E+03,
     + 6.2878E+03, 6.2970E+03, 6.3062E+03, 6.3155E+03, 6.3247E+03,
     +  6.3340E+03, 6.3433E+03, 6.3525E+03, 6.3618E+03, 6.3711E+03,
     + 6.3804E+03, 6.3897E+03, 6.3990E+03, 6.4083E+03, 6.4176E+03,
     +  6.4270E+03, 6.4363E+03, 6.4457E+03, 6.4550E+03, 6.4644E+03,
     + 6.4737E+03, 6.4831E+03, 6.4925E+03, 6.5019E+03, 6.5112E+03,
     +  6.5206E+03, 6.5301E+03, 6.5395E+03, 6.5489E+03, 6.5583E+03,
     + 6.5677E+03, 6.5772E+03, 6.5866E+03, 6.5961E+03, 6.6055E+03,
     +  6.6150E+03, 6.6245E+03, 6.6340E+03, 6.6435E+03, 6.6530E+03,
     + 6.6625E+03, 6.6720E+03, 6.6815E+03, 6.6910E+03, 6.7005E+03,
     +  6.7101E+03, 6.7196E+03, 6.7292E+03, 6.7387E+03, 6.7483E+03,
     + 6.7579E+03, 6.7674E+03, 6.7770E+03, 6.7866E+03, 6.7962E+03,
     +  6.8058E+03, 6.8154E+03, 6.8251E+03, 6.8347E+03, 6.8443E+03,
     + 6.8540E+03, 6.8636E+03, 6.8733E+03, 6.8829E+03, 6.8926E+03,
     +  6.9023E+03, 6.9120E+03, 6.9217E+03, 6.9314E+03, 6.9411E+03,
     + 6.9508E+03, 6.9605E+03, 6.9702E+03, 6.9800E+03, 6.9897E+03,
     +  6.9994E+03, 7.0092E+03, 7.0189E+03, 7.0287E+03, 7.0385E+03,
     + 7.0483E+03, 7.0581E+03, 7.0679E+03, 7.0777E+03, 7.0875E+03,
     +  7.0973E+03, 7.1071E+03, 7.1169E+03, 7.1268E+03, 7.1366E+03,
     + 7.1465E+03, 7.1563E+03, 7.1662E+03, 7.1761E+03, 7.1859E+03,
     +  7.1958E+03, 7.2057E+03, 7.2156E+03, 7.2255E+03, 7.2354E+03,
     + 7.2454E+03, 7.2553E+03, 7.2652E+03, 7.2752E+03, 7.2851E+03,
     +  7.2951E+03, 7.3050E+03, 7.3150E+03, 7.3250E+03, 7.3350E+03,
     + 7.3450E+03, 7.3550E+03, 7.3650E+03, 7.3750E+03, 7.3850E+03,
     +  7.3950E+03, 7.4051E+03, 7.4151E+03, 7.4251E+03, 7.4352E+03,
     + 7.4453E+03, 7.4553E+03, 7.4654E+03, 7.4755E+03, 7.4856E+03,
     +  7.4957E+03, 7.5058E+03, 7.5159E+03, 7.5260E+03, 7.5361E+03,
     + 7.5463E+03, 7.5564E+03, 7.5665E+03, 7.5767E+03, 7.5869E+03,
     +  7.5970E+03, 7.6072E+03, 7.6174E+03, 7.6276E+03, 7.6378E+03,
     + 7.6480E+03, 7.6582E+03, 7.6684E+03, 7.6786E+03, 7.6888E+03,
     +  7.6991E+03, 7.7093E+03, 7.7196E+03, 7.7298E+03, 7.7401E+03,
     + 7.7504E+03, 7.7606E+03, 7.7709E+03, 7.7812E+03, 7.7915E+03,
     +  7.8018E+03, 7.8121E+03, 7.8225E+03, 7.8328E+03, 7.8431E+03,
     + 7.8535E+03, 7.8638E+03, 7.8742E+03, 7.8845E+03, 7.8949E+03,
     +  7.9053E+03, 7.9157E+03, 7.9260E+03, 7.9364E+03, 7.9469E+03,
     + 7.9573E+03, 7.9677E+03, 7.9781E+03, 7.9885E+03, 7.9990E+03,
     +  8.0094E+03, 8.0199E+03, 8.0303E+03, 8.0408E+03, 8.0513E+03,
     + 8.0618E+03, 8.0723E+03, 8.0828E+03, 8.0933E+03, 8.1038E+03,
     +  8.1143E+03, 8.1248E+03, 8.1354E+03, 8.1459E+03, 8.1564E+03,
     + 8.1670E+03, 8.1776E+03, 8.1881E+03, 8.1987E+03, 8.2093E+03,
     +  8.2199E+03, 8.2305E+03, 8.2411E+03, 8.2517E+03, 8.2623E+03,
     + 8.2729E+03, 8.2835E+03, 8.2942E+03, 8.3048E+03, 8.3155E+03,
     +  8.3261E+03, 8.3368E+03, 8.3475E+03, 8.3582E+03, 8.3688E+03,
     + 8.3795E+03, 8.3902E+03, 8.4010E+03, 8.4117E+03, 8.4224E+03,
     +  8.4331E+03, 8.4439E+03, 8.4546E+03, 8.4654E+03, 8.4761E+03,
     + 8.4869E+03, 8.4976E+03, 8.5084E+03, 8.5192E+03, 8.5300E+03,
     +  8.5408E+03, 8.5516E+03, 8.5624E+03, 8.5733E+03, 8.5841E+03,
     + 8.5949E+03, 8.6058E+03, 8.6166E+03, 8.6275E+03, 8.6383E+03,
     +  8.6492E+03, 8.6601E+03, 8.6710E+03, 8.6819E+03, 8.6928E+03,
     + 8.7037E+03, 8.7146E+03, 8.7255E+03, 8.7364E+03, 8.7474E+03,
     +  8.7583E+03, 8.7693E+03, 8.7802E+03, 8.7912E+03, 8.8021E+03,
     + 8.8131E+03, 8.8241E+03, 8.8351E+03, 8.8461E+03, 8.8571E+03,
     +  8.8681E+03, 8.8791E+03, 8.8902E+03, 8.9012E+03, 8.9122E+03,
     + 8.9233E+03, 8.9343E+03, 8.9454E+03, 8.9565E+03, 8.9676E+03,
     +  8.9786E+03, 8.9897E+03, 9.0008E+03, 9.0119E+03, 9.0230E+03,
     + 9.0342E+03, 9.0453E+03, 9.0564E+03, 9.0676E+03, 9.0787E+03,
     +  9.0899E+03, 9.1010E+03, 9.1122E+03, 9.1234E+03, 9.1346E+03,
     + 9.1457E+03, 9.1569E+03, 9.1682E+03, 9.1794E+03, 9.1906E+03,
     +  9.2018E+03, 9.2130E+03, 9.2243E+03, 9.2355E+03, 9.2468E+03,
     + 9.2580E+03, 9.2693E+03, 9.2806E+03, 9.2919E+03, 9.3032E+03,
     +  9.3145E+03, 9.3258E+03, 9.3371E+03, 9.3484E+03, 9.3597E+03,
     + 9.3711E+03, 9.3824E+03, 9.3937E+03, 9.4051E+03, 9.4165E+03,
     +  9.4278E+03, 9.4392E+03, 9.4506E+03, 9.4620E+03, 9.4734E+03,
     + 9.4848E+03, 9.4962E+03, 9.5076E+03, 9.5190E+03, 9.5305E+03,
     +  9.5419E+03, 9.5533E+03, 9.5648E+03, 9.5763E+03, 9.5877E+03,
     + 9.5992E+03, 9.6107E+03, 9.6222E+03, 9.6337E+03, 9.6452E+03,
     +  9.6567E+03, 9.6682E+03, 9.6797E+03, 9.6913E+03, 9.7028E+03,
     + 9.7143E+03, 9.7259E+03, 9.7375E+03, 9.7490E+03, 9.7606E+03,
     +  9.7722E+03, 9.7838E+03, 9.7954E+03, 9.8070E+03, 9.8186E+03,
     + 9.8302E+03, 9.8418E+03, 9.8535E+03, 9.8651E+03, 9.8767E+03,
     +  9.8884E+03, 9.9001E+03, 9.9117E+03, 9.9234E+03, 9.9351E+03,
     + 9.9468E+03, 9.9585E+03, 9.9702E+03, 9.9819E+03, 9.9936E+03,
     +  1.0005E+04, 1.0017E+04, 1.0029E+04, 1.0041E+04, 1.0052E+04,
     + 1.0064E+04, 1.0076E+04, 1.0088E+04, 1.0099E+04, 1.0111E+04,
     +  1.0123E+04, 1.0135E+04, 1.0147E+04, 1.0158E+04, 1.0170E+04,
     + 1.0182E+04, 1.0194E+04, 1.0206E+04, 1.0218E+04, 1.0229E+04,
     +  1.0241E+04, 1.0253E+04, 1.0265E+04, 1.0277E+04, 1.0289E+04,
     + 1.0301E+04, 1.0313E+04, 1.0325E+04, 1.0337E+04, 1.0348E+04,
     +  1.0360E+04, 1.0372E+04, 1.0384E+04, 1.0396E+04, 1.0408E+04,
     + 1.0420E+04, 1.0432E+04, 1.0444E+04, 1.0456E+04, 1.0468E+04,
     +  1.0480E+04, 1.0492E+04, 1.0504E+04, 1.0516E+04, 1.0528E+04,
     + 1.0540E+04, 1.0552E+04, 1.0564E+04, 1.0577E+04, 1.0589E+04,
     +  1.0601E+04, 1.0613E+04, 1.0625E+04, 1.0637E+04, 1.0649E+04,
     + 1.0661E+04, 1.0673E+04, 1.0685E+04, 1.0698E+04, 1.0710E+04,
     +  1.0722E+04, 1.0734E+04, 1.0746E+04, 1.0758E+04, 1.0771E+04,
     + 1.0783E+04, 1.0795E+04, 1.0807E+04, 1.0819E+04, 1.0832E+04,
     +  1.0844E+04, 1.0856E+04, 1.0868E+04, 1.0881E+04, 1.0893E+04,
     + 1.0905E+04, 1.0917E+04, 1.0930E+04, 1.0942E+04, 1.0954E+04,
     +  1.0967E+04, 1.0979E+04, 1.0991E+04, 1.1003E+04, 1.1016E+04,
     + 1.1028E+04, 1.1040E+04, 1.1053E+04, 1.1065E+04, 1.1078E+04,
     +  1.1090E+04, 1.1102E+04, 1.1115E+04, 1.1127E+04, 1.1139E+04,
     + 1.1152E+04, 1.1164E+04, 1.1177E+04, 1.1189E+04, 1.1202E+04,
     +  1.1214E+04, 1.1226E+04, 1.1239E+04, 1.1251E+04, 1.1264E+04,
     + 1.1276E+04, 1.1289E+04, 1.1301E+04, 1.1314E+04, 1.1326E+04,
     +  1.1339E+04, 1.1351E+04, 1.1364E+04, 1.1376E+04, 1.1389E+04,
     + 1.1402E+04, 1.1414E+04, 1.1427E+04, 1.1439E+04, 1.1452E+04,
     +  1.1464E+04, 1.1477E+04, 1.1490E+04, 1.1502E+04, 1.1515E+04,
     + 1.1527E+04, 1.1540E+04, 1.1553E+04, 1.1565E+04, 1.1578E+04,
     +  1.1591E+04, 1.1603E+04, 1.1616E+04, 1.1629E+04, 1.1641E+04,
     + 1.1654E+04, 1.1667E+04, 1.1679E+04, 1.1692E+04, 1.1705E+04,
     +  1.1718E+04, 1.1730E+04, 1.1743E+04, 1.1756E+04, 1.1769E+04,
     + 1.1781E+04, 1.1794E+04, 1.1807E+04, 1.1820E+04, 1.1833E+04,
     +  1.1845E+04, 1.1858E+04, 1.1871E+04, 1.1884E+04, 1.1897E+04,
     + 1.1909E+04, 1.1922E+04, 1.1935E+04, 1.1948E+04, 1.1961E+04,
     +  1.1974E+04, 1.1987E+04, 1.2000E+04, 1.2012E+04, 1.2025E+04,
     + 1.2038E+04, 1.2051E+04, 1.2064E+04, 1.2077E+04, 1.2090E+04,
     +  1.2103E+04, 1.2116E+04, 1.2129E+04, 1.2142E+04, 1.2155E+04,
     + 1.2168E+04, 1.2181E+04, 1.2194E+04, 1.2207E+04, 1.2220E+04,
     +  1.2233E+04, 1.2246E+04, 1.2259E+04, 1.2272E+04, 1.2285E+04,
     + 1.2298E+04, 1.2311E+04, 1.2324E+04, 1.2337E+04, 1.2350E+04,
     +  1.2363E+04, 1.2376E+04, 1.2390E+04, 1.2403E+04, 1.2416E+04,
     + 1.2429E+04, 1.2442E+04, 1.2455E+04, 1.2468E+04, 1.2482E+04,
     +  1.2495E+04, 1.2508E+04, 1.2521E+04, 1.2534E+04, 1.2547E+04,
     + 1.2561E+04, 1.2574E+04, 1.2587E+04, 1.2600E+04, 1.2614E+04,
     +  1.2627E+04, 1.2640E+04, 1.2653E+04, 1.2667E+04, 1.2680E+04,
     + 1.2693E+04, 1.2706E+04, 1.2720E+04, 1.2733E+04, 1.2746E+04,
     +  1.2760E+04, 1.2773E+04, 1.2786E+04, 1.2800E+04, 1.2813E+04,
     + 1.2826E+04, 1.2840E+04, 1.2853E+04, 1.2866E+04, 1.2880E+04,
     +  1.2893E+04, 1.2906E+04, 1.2920E+04, 1.2933E+04, 1.2947E+04,
     + 1.2960E+04, 1.2973E+04, 1.2987E+04, 1.3000E+04, 1.3014E+04,
     +  1.3027E+04, 1.3041E+04, 1.3054E+04, 1.3068E+04, 1.3081E+04,
     + 1.3095E+04, 1.3108E+04, 1.3122E+04, 1.3135E+04, 1.3149E+04,
     +  1.3162E+04, 1.3176E+04, 1.3189E+04, 1.3203E+04, 1.3216E+04,
     + 1.3230E+04, 1.3244E+04, 1.3257E+04, 1.3271E+04, 1.3284E+04,
     +  1.3298E+04, 1.3311E+04, 1.3325E+04, 1.3339E+04, 1.3352E+04,
     + 1.3366E+04, 1.3380E+04, 1.3393E+04, 1.3407E+04, 1.3421E+04,
     +  1.3434E+04, 1.3448E+04, 1.3462E+04, 1.3475E+04, 1.3489E+04,
     + 1.3503E+04, 1.3516E+04, 1.3530E+04, 1.3544E+04, 1.3558E+04,
     +  1.3571E+04, 1.3585E+04, 1.3599E+04, 1.3613E+04, 1.3626E+04,
     + 1.3640E+04, 1.3654E+04, 1.3668E+04, 1.3682E+04, 1.3695E+04,
     +  1.3709E+04, 1.3723E+04, 1.3737E+04, 1.3751E+04, 1.3765E+04,
     + 1.3779E+04, 1.3792E+04, 1.3806E+04, 1.3820E+04, 1.3834E+04,
     +  1.3848E+04, 1.3862E+04, 1.3876E+04, 1.3890E+04, 1.3904E+04,
     + 1.3917E+04, 1.3931E+04, 1.3945E+04, 1.3959E+04, 1.3973E+04,
     +  1.3987E+04, 1.4001E+04, 1.4015E+04, 1.4029E+04, 1.4043E+04,
     + 1.4057E+04, 1.4071E+04, 1.4085E+04, 1.4099E+04, 1.4113E+04,
     +  1.4127E+04, 1.4141E+04, 1.4155E+04, 1.4169E+04, 1.4184E+04,
     + 1.4198E+04, 1.4212E+04, 1.4226E+04, 1.4240E+04, 1.4254E+04,
     +  1.4268E+04, 1.4282E+04, 1.4296E+04, 1.4310E+04, 1.4325E+04,
     + 1.4339E+04, 1.4353E+04, 1.4367E+04, 1.4381E+04, 1.4395E+04,
     +  1.4410E+04, 1.4424E+04, 1.4438E+04, 1.4452E+04, 1.4466E+04,
     + 1.4481E+04, 1.4495E+04, 1.4509E+04, 1.4523E+04, 1.4538E+04,
     +  1.4552E+04, 1.4566E+04, 1.4580E+04, 1.4595E+04, 1.4609E+04,
     + 1.4623E+04, 1.4638E+04, 1.4652E+04, 1.4666E+04, 1.4680E+04,
     +  1.4695E+04, 1.4709E+04, 1.4724E+04, 1.4738E+04, 1.4752E+04,
     + 1.4767E+04, 1.4781E+04, 1.4795E+04, 1.4810E+04, 1.4824E+04,
     +  1.4839E+04, 1.4853E+04, 1.4867E+04, 1.4882E+04, 1.4896E+04,
     + 1.4911E+04, 1.4925E+04, 1.4940E+04, 1.4954E+04, 1.4969E+04,
     +  1.4983E+04, 1.4997E+04, 1.5012E+04, 1.5026E+04, 1.5041E+04,
     + 1.5055E+04, 1.5070E+04, 1.5085E+04, 1.5099E+04, 1.5114E+04,
     +  1.5128E+04, 1.5143E+04, 1.5157E+04, 1.5172E+04, 1.5186E+04,
     + 1.5201E+04, 1.5216E+04, 1.5230E+04, 1.5245E+04, 1.5259E+04,
     +  1.5274E+04, 1.5289E+04, 1.5303E+04, 1.5318E+04, 1.5333E+04,
     + 1.5347E+04, 1.5362E+04, 1.5377E+04, 1.5391E+04, 1.5406E+04,
     +  1.5421E+04, 1.5435E+04, 1.5450E+04, 1.5465E+04, 1.5480E+04,
     + 1.5494E+04, 1.5509E+04, 1.5524E+04, 1.5539E+04, 1.5553E+04,
     +  1.5568E+04, 1.5583E+04, 1.5598E+04, 1.5612E+04, 1.5627E+04,
     + 1.5642E+04, 1.5657E+04, 1.5672E+04, 1.5687E+04, 1.5701E+04,
     +  1.5716E+04, 1.5731E+04, 1.5746E+04, 1.5761E+04, 1.5776E+04,
     + 1.5791E+04, 1.5805E+04, 1.5820E+04, 1.5835E+04, 1.5850E+04,
     +  1.5865E+04, 1.5880E+04, 1.5895E+04, 1.5910E+04, 1.5925E+04,
     + 1.5940E+04, 1.5955E+04, 1.5970E+04, 1.5985E+04, 1.6000E+04,
     +  1.6015E+04, 1.6030E+04, 1.6045E+04, 1.6060E+04, 1.6075E+04,
     + 1.6090E+04, 1.6105E+04, 1.6120E+04, 1.6135E+04, 1.6150E+04,
     +  1.6165E+04, 1.6180E+04, 1.6195E+04, 1.6210E+04, 1.6225E+04,
     + 1.6240E+04, 1.6256E+04, 1.6271E+04, 1.6286E+04, 1.6301E+04,
     +  1.6316E+04, 1.6331E+04, 1.6346E+04, 1.6362E+04, 1.6377E+04,
     + 1.6392E+04, 1.6407E+04, 1.6422E+04, 1.6438E+04, 1.6453E+04,
     +  1.6468E+04, 1.6483E+04, 1.6498E+04, 1.6514E+04, 1.6529E+04,
     + 1.6544E+04, 1.6559E+04, 1.6575E+04, 1.6590E+04, 1.6605E+04,
     +  1.6620E+04, 1.6636E+04, 1.6651E+04, 1.6666E+04, 1.6682E+04,
     + 1.6697E+04, 1.6712E+04, 1.6728E+04, 1.6743E+04, 1.6758E+04,
     +  1.6774E+04, 1.6789E+04, 1.6805E+04, 1.6820E+04, 1.6835E+04,
     + 1.6851E+04, 1.6866E+04, 1.6882E+04, 1.6897E+04, 1.6912E+04,
     +  1.6928E+04, 1.6943E+04, 1.6959E+04, 1.6974E+04, 1.6990E+04,
     + 1.7005E+04, 1.7021E+04, 1.7036E+04, 1.7052E+04, 1.7067E+04,
     +  1.7083E+04, 1.7098E+04, 1.7114E+04, 1.7129E+04, 1.7145E+04,
     + 1.7160E+04, 1.7176E+04, 1.7191E+04, 1.7207E+04, 1.7222E+04,
     +  1.7238E+04, 1.7254E+04, 1.7269E+04, 1.7285E+04, 1.7300E+04,
     + 1.7316E+04, 1.7332E+04, 1.7347E+04, 1.7363E+04, 1.7379E+04,
     +  1.7394E+04, 1.7410E+04, 1.7426E+04, 1.7441E+04, 1.7457E+04,
     + 1.7473E+04, 1.7488E+04, 1.7504E+04, 1.7520E+04, 1.7536E+04,
     +  1.7551E+04, 1.7567E+04, 1.7583E+04, 1.7599E+04, 1.7614E+04,
     + 1.7630E+04, 1.7646E+04, 1.7662E+04, 1.7677E+04, 1.7693E+04,
     +  1.7709E+04, 1.7725E+04, 1.7741E+04, 1.7757E+04, 1.7772E+04,
     + 1.7788E+04, 1.7804E+04, 1.7820E+04, 1.7836E+04, 1.7852E+04,
     +  1.7868E+04, 1.7883E+04, 1.7899E+04, 1.7915E+04, 1.7931E+04,
     + 1.7947E+04, 1.7963E+04, 1.7979E+04, 1.7995E+04, 1.8011E+04,
     +  1.8027E+04, 1.8043E+04, 1.8059E+04, 1.8075E+04, 1.8091E+04,
     + 1.8107E+04, 1.8123E+04, 1.8139E+04, 1.8155E+04, 1.8171E+04,
     +  1.8187E+04, 1.8203E+04, 1.8219E+04, 1.8235E+04, 1.8251E+04,
     + 1.8267E+04, 1.8283E+04, 1.8299E+04, 1.8315E+04, 1.8331E+04,
     +  1.8347E+04, 1.8364E+04, 1.8380E+04, 1.8396E+04, 1.8412E+04,
     + 1.8428E+04, 1.8444E+04, 1.8460E+04, 1.8477E+04, 1.8493E+04,
     +  1.8509E+04, 1.8525E+04, 1.8541E+04, 1.8557E+04, 1.8574E+04,
     + 1.8590E+04, 1.8606E+04, 1.8622E+04, 1.8639E+04, 1.8655E+04,
     +  1.8671E+04, 1.8687E+04, 1.8704E+04, 1.8720E+04, 1.8736E+04,
     + 1.8752E+04, 1.8769E+04, 1.8785E+04, 1.8801E+04, 1.8818E+04,
     +  1.8834E+04, 1.8850E+04, 1.8867E+04, 1.8883E+04, 1.8899E+04,
     + 1.8916E+04, 1.8932E+04, 1.8949E+04, 1.8965E+04, 1.8981E+04,
     +  1.8998E+04, 1.9014E+04, 1.9031E+04, 1.9047E+04, 1.9063E+04,
     + 1.9080E+04, 1.9096E+04, 1.9113E+04, 1.9129E+04, 1.9146E+04,
     +  1.9162E+04, 1.9179E+04, 1.9195E+04, 1.9212E+04, 1.9228E+04,
     + 1.9245E+04, 1.9261E+04, 1.9278E+04, 1.9294E+04, 1.9311E+04,
     +  1.9327E+04, 1.9344E+04, 1.9360E+04, 1.9377E+04, 1.9394E+04,
     + 1.9410E+04, 1.9427E+04, 1.9443E+04, 1.9460E+04, 1.9477E+04,
     +  1.9493E+04, 1.9510E+04, 1.9527E+04, 1.9543E+04, 1.9560E+04,
     + 1.9577E+04, 1.9593E+04, 1.9610E+04, 1.9627E+04, 1.9643E+04,
     +  1.9660E+04, 1.9677E+04, 1.9693E+04, 1.9710E+04, 1.9727E+04,
     + 1.9744E+04, 1.9760E+04, 1.9777E+04, 1.9794E+04, 1.9811E+04,
     +  1.9827E+04, 1.9844E+04, 1.9861E+04, 1.9878E+04, 1.9895E+04,
     + 1.9911E+04, 1.9928E+04, 1.9945E+04, 1.9962E+04, 1.9979E+04,
     +  1.9996E+04, 2.0012E+04, 2.0029E+04, 2.0046E+04, 2.0063E+04,
     + 2.0080E+04, 2.0097E+04, 2.0114E+04, 2.0131E+04, 2.0148E+04,
     +  2.0164E+04, 2.0181E+04, 2.0198E+04, 2.0215E+04, 2.0232E+04,
     + 2.0249E+04, 2.0266E+04, 2.0283E+04, 2.0300E+04, 2.0317E+04,
     +  2.0334E+04, 2.0351E+04, 2.0368E+04, 2.0385E+04, 2.0402E+04,
     + 2.0419E+04, 2.0436E+04, 2.0453E+04, 2.0470E+04, 2.0487E+04,
     +  2.0505E+04, 2.0522E+04, 2.0539E+04, 2.0556E+04, 2.0573E+04,
     + 2.0590E+04, 2.0607E+04, 2.0624E+04, 2.0641E+04, 2.0659E+04,
     +  2.0676E+04, 2.0693E+04, 2.0710E+04, 2.0727E+04, 2.0744E+04,
     + 2.0762E+04, 2.0779E+04, 2.0796E+04, 2.0813E+04, 2.0830E+04,
     +  2.0848E+04, 2.0865E+04, 2.0882E+04, 2.0899E+04, 2.0917E+04,
     + 2.0934E+04, 2.0951E+04, 2.0968E+04, 2.0986E+04, 2.1003E+04,
     +  2.1020E+04, 2.1038E+04, 2.1055E+04, 2.1072E+04, 2.1090E+04,
     + 2.1107E+04, 2.1124E+04, 2.1142E+04, 2.1159E+04, 2.1176E+04,
     +  2.1194E+04, 2.1211E+04, 2.1229E+04, 2.1246E+04, 2.1263E+04,
     + 2.1281E+04, 2.1298E+04, 2.1316E+04, 2.1333E+04, 2.1351E+04,
     +  2.1368E+04, 2.1385E+04, 2.1403E+04, 2.1420E+04, 2.1438E+04,
     + 2.1455E+04, 2.1473E+04, 2.1490E+04, 2.1508E+04, 2.1525E+04,
     +  2.1543E+04, 2.1560E+04, 2.1578E+04, 2.1596E+04, 2.1613E+04,
     + 2.1631E+04, 2.1648E+04, 2.1666E+04, 2.1683E+04, 2.1701E+04,
     +  2.1719E+04, 2.1736E+04, 2.1754E+04, 2.1771E+04, 2.1789E+04,
     + 2.1807E+04, 2.1824E+04, 2.1842E+04, 2.1860E+04, 2.1877E+04,
     +  2.1895E+04, 2.1913E+04, 2.1930E+04, 2.1948E+04, 2.1966E+04,
     + 2.1984E+04, 2.2001E+04, 2.2019E+04, 2.2037E+04, 2.2054E+04,
     +  2.2072E+04, 2.2090E+04, 2.2108E+04, 2.2126E+04, 2.2143E+04,
     + 2.2161E+04, 2.2179E+04, 2.2197E+04, 2.2215E+04, 2.2232E+04,
     +  2.2250E+04, 2.2268E+04, 2.2286E+04, 2.2304E+04, 2.2322E+04,
     + 2.2339E+04, 2.2357E+04, 2.2375E+04, 2.2393E+04, 2.2411E+04,
     +  2.2429E+04, 2.2447E+04, 2.2465E+04, 2.2483E+04, 2.2501E+04,
     + 2.2519E+04, 2.2537E+04, 2.2554E+04, 2.2572E+04, 2.2590E+04,
     +  2.2608E+04, 2.2626E+04, 2.2644E+04, 2.2662E+04, 2.2680E+04,
     + 2.2698E+04, 2.2716E+04, 2.2734E+04, 2.2753E+04, 2.2771E+04,
     +  2.2789E+04, 2.2807E+04, 2.2825E+04, 2.2843E+04, 2.2861E+04,
     + 2.2879E+04, 2.2897E+04, 2.2915E+04, 2.2933E+04, 2.2951E+04,
     +  2.2970E+04, 2.2988E+04, 2.3006E+04, 2.3024E+04, 2.3042E+04,
     + 2.3060E+04, 2.3079E+04, 2.3097E+04, 2.3115E+04, 2.3133E+04,
     +  2.3151E+04, 2.3170E+04, 2.3188E+04, 2.3206E+04, 2.3224E+04,
     + 2.3242E+04, 2.3261E+04, 2.3279E+04, 2.3297E+04, 2.3316E+04,
     +  2.3334E+04, 2.3352E+04, 2.3370E+04, 2.3389E+04, 2.3407E+04,
     + 2.3425E+04, 2.3444E+04, 2.3462E+04, 2.3480E+04, 2.3499E+04,
     +  2.3517E+04, 2.3535E+04, 2.3554E+04, 2.3572E+04, 2.3591E+04,
     + 2.3609E+04, 2.3627E+04, 2.3646E+04, 2.3664E+04, 2.3683E+04,
     +  2.3701E+04, 2.3720E+04, 2.3738E+04, 2.3756E+04, 2.3775E+04,
     + 2.3793E+04, 2.3812E+04, 2.3830E+04, 2.3849E+04, 2.3867E+04,
     +  2.3886E+04, 2.3904E+04, 2.3923E+04, 2.3941E+04, 2.3960E+04,
     + 2.3979E+04, 2.3997E+04, 2.4016E+04, 2.4034E+04, 2.4053E+04,
     +  2.4071E+04, 2.4090E+04, 2.4109E+04, 2.4127E+04, 2.4146E+04,
     + 2.4164E+04, 2.4183E+04, 2.4202E+04, 2.4220E+04, 2.4239E+04,
     +  2.4258E+04, 2.4276E+04, 2.4295E+04, 2.4314E+04, 2.4332E+04,
     + 2.4351E+04, 2.4370E+04, 2.4389E+04, 2.4407E+04, 2.4426E+04,
     +  2.4445E+04, 2.4463E+04, 2.4482E+04, 2.4501E+04, 2.4520E+04,
     + 2.4539E+04, 2.4557E+04, 2.4576E+04, 2.4595E+04, 2.4614E+04,
     +  2.4633E+04, 2.4651E+04, 2.4670E+04, 2.4689E+04, 2.4708E+04,
     + 2.4727E+04, 2.4746E+04, 2.4764E+04, 2.4783E+04, 2.4802E+04,
     +  2.4821E+04, 2.4840E+04, 2.4859E+04, 2.4878E+04, 2.4897E+04,
     + 2.4916E+04, 2.4935E+04, 2.4954E+04, 2.4973E+04, 2.4992E+04,
     +  2.5010E+04, 2.5029E+04, 2.5048E+04, 2.5067E+04, 2.5086E+04,
     + 2.5105E+04, 2.5124E+04, 2.5143E+04, 2.5162E+04, 2.5182E+04,
     +  2.5201E+04, 2.5220E+04, 2.5239E+04, 2.5258E+04, 2.5277E+04,
     + 2.5296E+04, 2.5315E+04, 2.5334E+04, 2.5353E+04, 2.5372E+04,
     +  2.5391E+04, 2.5411E+04, 2.5430E+04, 2.5449E+04, 2.5468E+04,
     + 2.5487E+04, 2.5506E+04, 2.5525E+04, 2.5545E+04, 2.5564E+04,
     +  2.5583E+04, 2.5602E+04, 2.5621E+04, 2.5641E+04, 2.5660E+04,
     + 2.5679E+04, 2.5698E+04, 2.5718E+04, 2.5737E+04, 2.5756E+04,
     +  2.5775E+04, 2.5795E+04, 2.5814E+04, 2.5833E+04, 2.5853E+04,
     + 2.5872E+04, 2.5891E+04, 2.5911E+04, 2.5930E+04, 2.5949E+04,
     +  2.5969E+04, 2.5988E+04, 2.6007E+04, 2.6027E+04, 2.6046E+04,
     + 2.6065E+04, 2.6085E+04, 2.6104E+04, 2.6124E+04, 2.6143E+04,
     +  2.6163E+04, 2.6182E+04, 2.6201E+04, 2.6221E+04, 2.6240E+04,
     + 2.6260E+04, 2.6279E+04, 2.6299E+04, 2.6318E+04, 2.6338E+04,
     +  2.6357E+04, 2.6377E+04, 2.6396E+04, 2.6416E+04, 2.6435E+04/
      data (QofT( 9,J),J=2409,4510)/ 2.6455E+04, 2.6474E+04,
     +  2.6494E+04, 2.6514E+04, 2.6533E+04, 2.6553E+04, 2.6572E+04,
     +  2.6592E+04,
     + 2.6611E+04, 2.6631E+04, 2.6651E+04, 2.6670E+04, 2.6690E+04,
     +  2.6710E+04, 2.6729E+04, 2.6749E+04, 2.6769E+04, 2.6788E+04,
     + 2.6808E+04, 2.6828E+04, 2.6847E+04, 2.6867E+04, 2.6887E+04,
     +  2.6906E+04, 2.6926E+04, 2.6946E+04, 2.6966E+04, 2.6985E+04,
     + 2.7005E+04, 2.7025E+04, 2.7045E+04, 2.7064E+04, 2.7084E+04,
     +  2.7104E+04, 2.7124E+04, 2.7144E+04, 2.7163E+04, 2.7183E+04,
     + 2.7203E+04, 2.7223E+04, 2.7243E+04, 2.7263E+04, 2.7283E+04,
     +  2.7302E+04, 2.7322E+04, 2.7342E+04, 2.7362E+04, 2.7382E+04,
     + 2.7402E+04, 2.7422E+04, 2.7442E+04, 2.7462E+04, 2.7482E+04,
     +  2.7502E+04, 2.7521E+04, 2.7541E+04, 2.7561E+04, 2.7581E+04,
     + 2.7601E+04, 2.7621E+04, 2.7641E+04, 2.7661E+04, 2.7681E+04,
     +  2.7701E+04, 2.7721E+04, 2.7741E+04, 2.7762E+04, 2.7782E+04,
     + 2.7802E+04, 2.7822E+04, 2.7842E+04, 2.7862E+04, 2.7882E+04,
     +  2.7902E+04, 2.7922E+04, 2.7942E+04, 2.7962E+04, 2.7983E+04,
     + 2.8003E+04, 2.8023E+04, 2.8043E+04, 2.8063E+04, 2.8083E+04,
     +  2.8104E+04, 2.8124E+04, 2.8144E+04, 2.8164E+04, 2.8184E+04,
     + 2.8205E+04, 2.8225E+04, 2.8245E+04, 2.8265E+04, 2.8285E+04,
     +  2.8306E+04, 2.8326E+04, 2.8346E+04, 2.8367E+04, 2.8387E+04,
     + 2.8407E+04, 2.8427E+04, 2.8448E+04, 2.8468E+04, 2.8488E+04,
     +  2.8509E+04, 2.8529E+04, 2.8549E+04, 2.8570E+04, 2.8590E+04,
     + 2.8611E+04, 2.8631E+04, 2.8651E+04, 2.8672E+04, 2.8692E+04,
     +  2.8712E+04, 2.8733E+04, 2.8753E+04, 2.8774E+04, 2.8794E+04,
     + 2.8815E+04, 2.8835E+04, 2.8856E+04, 2.8876E+04, 2.8897E+04,
     +  2.8917E+04, 2.8938E+04, 2.8958E+04, 2.8979E+04, 2.8999E+04,
     + 2.9020E+04, 2.9040E+04, 2.9061E+04, 2.9081E+04, 2.9102E+04,
     +  2.9122E+04, 2.9143E+04, 2.9163E+04, 2.9184E+04, 2.9205E+04,
     + 2.9225E+04, 2.9246E+04, 2.9266E+04, 2.9287E+04, 2.9308E+04,
     +  2.9328E+04, 2.9349E+04, 2.9370E+04, 2.9390E+04, 2.9411E+04,
     + 2.9432E+04, 2.9452E+04, 2.9473E+04, 2.9494E+04, 2.9514E+04,
     +  2.9535E+04, 2.9556E+04, 2.9577E+04, 2.9597E+04, 2.9618E+04,
     + 2.9639E+04, 2.9660E+04, 2.9680E+04, 2.9701E+04, 2.9722E+04,
     +  2.9743E+04, 2.9764E+04, 2.9784E+04, 2.9805E+04, 2.9826E+04,
     + 2.9847E+04, 2.9868E+04, 2.9889E+04, 2.9909E+04, 2.9930E+04,
     +  2.9951E+04, 2.9972E+04, 2.9993E+04, 3.0014E+04, 3.0035E+04,
     + 3.0056E+04, 3.0077E+04, 3.0098E+04, 3.0118E+04, 3.0139E+04,
     +  3.0160E+04, 3.0181E+04, 3.0202E+04, 3.0223E+04, 3.0244E+04,
     + 3.0265E+04, 3.0286E+04, 3.0307E+04, 3.0328E+04, 3.0349E+04,
     +  3.0370E+04, 3.0391E+04, 3.0412E+04, 3.0433E+04, 3.0454E+04,
     + 3.0476E+04, 3.0497E+04, 3.0518E+04, 3.0539E+04, 3.0560E+04,
     +  3.0581E+04, 3.0602E+04, 3.0623E+04, 3.0644E+04, 3.0666E+04,
     + 3.0687E+04, 3.0708E+04, 3.0729E+04, 3.0750E+04, 3.0771E+04,
     +  3.0792E+04, 3.0814E+04, 3.0835E+04, 3.0856E+04, 3.0877E+04,
     + 3.0899E+04, 3.0920E+04, 3.0941E+04, 3.0962E+04, 3.0983E+04,
     +  3.1005E+04, 3.1026E+04, 3.1047E+04, 3.1069E+04, 3.1090E+04,
     + 3.1111E+04, 3.1132E+04, 3.1154E+04, 3.1175E+04, 3.1196E+04,
     +  3.1218E+04, 3.1239E+04, 3.1260E+04, 3.1282E+04, 3.1303E+04,
     + 3.1325E+04, 3.1346E+04, 3.1367E+04, 3.1389E+04, 3.1410E+04,
     +  3.1432E+04, 3.1453E+04, 3.1474E+04, 3.1496E+04, 3.1517E+04,
     + 3.1539E+04, 3.1560E+04, 3.1582E+04, 3.1603E+04, 3.1625E+04,
     +  3.1646E+04, 3.1668E+04, 3.1689E+04, 3.1711E+04, 3.1732E+04,
     + 3.1754E+04, 3.1775E+04, 3.1797E+04, 3.1818E+04, 3.1840E+04,
     +  3.1862E+04, 3.1883E+04, 3.1905E+04, 3.1926E+04, 3.1948E+04,
     + 3.1970E+04, 3.1991E+04, 3.2013E+04, 3.2034E+04, 3.2056E+04,
     +  3.2078E+04, 3.2099E+04, 3.2121E+04, 3.2143E+04, 3.2164E+04,
     + 3.2186E+04, 3.2208E+04, 3.2229E+04, 3.2251E+04, 3.2273E+04,
     +  3.2295E+04, 3.2316E+04, 3.2338E+04, 3.2360E+04, 3.2382E+04,
     + 3.2403E+04, 3.2425E+04, 3.2447E+04, 3.2469E+04, 3.2491E+04,
     +  3.2512E+04, 3.2534E+04, 3.2556E+04, 3.2578E+04, 3.2600E+04,
     + 3.2621E+04, 3.2643E+04, 3.2665E+04, 3.2687E+04, 3.2709E+04,
     +  3.2731E+04, 3.2753E+04, 3.2775E+04, 3.2796E+04, 3.2818E+04,
     + 3.2840E+04, 3.2862E+04, 3.2884E+04, 3.2906E+04, 3.2928E+04,
     +  3.2950E+04, 3.2972E+04, 3.2994E+04, 3.3016E+04, 3.3038E+04,
     + 3.3060E+04, 3.3082E+04, 3.3104E+04, 3.3126E+04, 3.3148E+04,
     +  3.3170E+04, 3.3192E+04, 3.3214E+04, 3.3236E+04, 3.3258E+04,
     + 3.3280E+04, 3.3302E+04, 3.3325E+04, 3.3347E+04, 3.3369E+04,
     +  3.3391E+04, 3.3413E+04, 3.3435E+04, 3.3457E+04, 3.3479E+04,
     + 3.3502E+04, 3.3524E+04, 3.3546E+04, 3.3568E+04, 3.3590E+04,
     +  3.3613E+04, 3.3635E+04, 3.3657E+04, 3.3679E+04, 3.3701E+04,
     + 3.3724E+04, 3.3746E+04, 3.3768E+04, 3.3790E+04, 3.3813E+04,
     +  3.3835E+04, 3.3857E+04, 3.3879E+04, 3.3902E+04, 3.3924E+04,
     + 3.3946E+04, 3.3969E+04, 3.3991E+04, 3.4013E+04, 3.4036E+04,
     +  3.4058E+04, 3.4080E+04, 3.4103E+04, 3.4125E+04, 3.4148E+04,
     + 3.4170E+04, 3.4192E+04, 3.4215E+04, 3.4237E+04, 3.4260E+04,
     +  3.4282E+04, 3.4304E+04, 3.4327E+04, 3.4349E+04, 3.4372E+04,
     + 3.4394E+04, 3.4417E+04, 3.4439E+04, 3.4462E+04, 3.4484E+04,
     +  3.4507E+04, 3.4529E+04, 3.4552E+04, 3.4574E+04, 3.4597E+04,
     + 3.4619E+04, 3.4642E+04, 3.4664E+04, 3.4687E+04, 3.4710E+04,
     +  3.4732E+04, 3.4755E+04, 3.4777E+04, 3.4800E+04, 3.4823E+04,
     + 3.4845E+04, 3.4868E+04, 3.4891E+04, 3.4913E+04, 3.4936E+04,
     +  3.4959E+04, 3.4981E+04, 3.5004E+04, 3.5027E+04, 3.5049E+04,
     + 3.5072E+04, 3.5095E+04, 3.5117E+04, 3.5140E+04, 3.5163E+04,
     +  3.5186E+04, 3.5208E+04, 3.5231E+04, 3.5254E+04, 3.5277E+04,
     + 3.5299E+04, 3.5322E+04, 3.5345E+04, 3.5368E+04, 3.5391E+04,
     +  3.5413E+04, 3.5436E+04, 3.5459E+04, 3.5482E+04, 3.5505E+04,
     + 3.5528E+04, 3.5551E+04, 3.5573E+04, 3.5596E+04, 3.5619E+04,
     +  3.5642E+04, 3.5665E+04, 3.5688E+04, 3.5711E+04, 3.5734E+04,
     + 3.5757E+04, 3.5780E+04, 3.5803E+04, 3.5826E+04, 3.5849E+04,
     +  3.5872E+04, 3.5895E+04, 3.5918E+04, 3.5940E+04, 3.5964E+04,
     + 3.5987E+04, 3.6010E+04, 3.6033E+04, 3.6056E+04, 3.6079E+04,
     +  3.6102E+04, 3.6125E+04, 3.6148E+04, 3.6171E+04, 3.6194E+04,
     + 3.6217E+04, 3.6240E+04, 3.6263E+04, 3.6286E+04, 3.6310E+04,
     +  3.6333E+04, 3.6356E+04, 3.6379E+04, 3.6402E+04, 3.6425E+04,
     + 3.6449E+04, 3.6472E+04, 3.6495E+04, 3.6518E+04, 3.6541E+04,
     +  3.6565E+04, 3.6588E+04, 3.6611E+04, 3.6634E+04, 3.6657E+04,
     + 3.6681E+04, 3.6704E+04, 3.6727E+04, 3.6751E+04, 3.6774E+04,
     +  3.6797E+04, 3.6820E+04, 3.6844E+04, 3.6867E+04, 3.6890E+04,
     + 3.6914E+04, 3.6937E+04, 3.6960E+04, 3.6984E+04, 3.7007E+04,
     +  3.7030E+04, 3.7054E+04, 3.7077E+04, 3.7101E+04, 3.7124E+04,
     + 3.7147E+04, 3.7171E+04, 3.7194E+04, 3.7218E+04, 3.7241E+04,
     +  3.7265E+04, 3.7288E+04, 3.7312E+04, 3.7335E+04, 3.7358E+04,
     + 3.7382E+04, 3.7405E+04, 3.7429E+04, 3.7452E+04, 3.7476E+04,
     +  3.7500E+04, 3.7523E+04, 3.7547E+04, 3.7570E+04, 3.7594E+04,
     + 3.7617E+04, 3.7641E+04, 3.7664E+04, 3.7688E+04, 3.7712E+04,
     +  3.7735E+04, 3.7759E+04, 3.7782E+04, 3.7806E+04, 3.7830E+04,
     + 3.7853E+04, 3.7877E+04, 3.7901E+04, 3.7924E+04, 3.7948E+04,
     +  3.7972E+04, 3.7995E+04, 3.8019E+04, 3.8043E+04, 3.8067E+04,
     + 3.8090E+04, 3.8114E+04, 3.8138E+04, 3.8162E+04, 3.8185E+04,
     +  3.8209E+04, 3.8233E+04, 3.8257E+04, 3.8280E+04, 3.8304E+04,
     + 3.8328E+04, 3.8352E+04, 3.8376E+04, 3.8399E+04, 3.8423E+04,
     +  3.8447E+04, 3.8471E+04, 3.8495E+04, 3.8519E+04, 3.8543E+04,
     + 3.8566E+04, 3.8590E+04, 3.8614E+04, 3.8638E+04, 3.8662E+04,
     +  3.8686E+04, 3.8710E+04, 3.8734E+04, 3.8758E+04, 3.8782E+04,
     + 3.8806E+04, 3.8830E+04, 3.8854E+04, 3.8878E+04, 3.8902E+04,
     +  3.8926E+04, 3.8950E+04, 3.8974E+04, 3.8998E+04, 3.9022E+04,
     + 3.9046E+04, 3.9070E+04, 3.9094E+04, 3.9118E+04, 3.9142E+04,
     +  3.9166E+04, 3.9190E+04, 3.9214E+04, 3.9238E+04, 3.9262E+04,
     + 3.9287E+04, 3.9311E+04, 3.9335E+04, 3.9359E+04, 3.9383E+04,
     +  3.9407E+04, 3.9431E+04, 3.9456E+04, 3.9480E+04, 3.9504E+04,
     + 3.9528E+04, 3.9552E+04, 3.9577E+04, 3.9601E+04, 3.9625E+04,
     +  3.9649E+04, 3.9674E+04, 3.9698E+04, 3.9722E+04, 3.9746E+04,
     + 3.9771E+04, 3.9795E+04, 3.9819E+04, 3.9844E+04, 3.9868E+04,
     +  3.9892E+04, 3.9917E+04, 3.9941E+04, 3.9965E+04, 3.9990E+04,
     + 4.0014E+04, 4.0038E+04, 4.0063E+04, 4.0087E+04, 4.0111E+04,
     +  4.0136E+04, 4.0160E+04, 4.0185E+04, 4.0209E+04, 4.0233E+04,
     + 4.0258E+04, 4.0282E+04, 4.0307E+04, 4.0331E+04, 4.0356E+04,
     +  4.0380E+04, 4.0405E+04, 4.0429E+04, 4.0454E+04, 4.0478E+04,
     + 4.0503E+04, 4.0527E+04, 4.0552E+04, 4.0576E+04, 4.0601E+04,
     +  4.0625E+04, 4.0650E+04, 4.0675E+04, 4.0699E+04, 4.0724E+04,
     + 4.0748E+04, 4.0773E+04, 4.0798E+04, 4.0822E+04, 4.0847E+04,
     +  4.0871E+04, 4.0896E+04, 4.0921E+04, 4.0945E+04, 4.0970E+04,
     + 4.0995E+04, 4.1019E+04, 4.1044E+04, 4.1069E+04, 4.1093E+04,
     +  4.1118E+04, 4.1143E+04, 4.1168E+04, 4.1192E+04, 4.1217E+04,
     + 4.1242E+04, 4.1267E+04, 4.1291E+04, 4.1316E+04, 4.1341E+04,
     +  4.1366E+04, 4.1391E+04, 4.1415E+04, 4.1440E+04, 4.1465E+04,
     + 4.1490E+04, 4.1515E+04, 4.1540E+04, 4.1564E+04, 4.1589E+04,
     +  4.1614E+04, 4.1639E+04, 4.1664E+04, 4.1689E+04, 4.1714E+04,
     + 4.1739E+04, 4.1764E+04, 4.1788E+04, 4.1813E+04, 4.1838E+04,
     +  4.1863E+04, 4.1888E+04, 4.1913E+04, 4.1938E+04, 4.1963E+04,
     + 4.1988E+04, 4.2013E+04, 4.2038E+04, 4.2063E+04, 4.2088E+04,
     +  4.2113E+04, 4.2138E+04, 4.2163E+04, 4.2188E+04, 4.2213E+04,
     + 4.2239E+04, 4.2264E+04, 4.2289E+04, 4.2314E+04, 4.2339E+04,
     +  4.2364E+04, 4.2389E+04, 4.2414E+04, 4.2439E+04, 4.2465E+04,
     + 4.2490E+04, 4.2515E+04, 4.2540E+04, 4.2565E+04, 4.2590E+04,
     +  4.2616E+04, 4.2641E+04, 4.2666E+04, 4.2691E+04, 4.2716E+04,
     + 4.2742E+04, 4.2767E+04, 4.2792E+04, 4.2817E+04, 4.2843E+04,
     +  4.2868E+04, 4.2893E+04, 4.2918E+04, 4.2944E+04, 4.2969E+04,
     + 4.2994E+04, 4.3020E+04, 4.3045E+04, 4.3070E+04, 4.3096E+04,
     +  4.3121E+04, 4.3146E+04, 4.3172E+04, 4.3197E+04, 4.3222E+04,
     + 4.3248E+04, 4.3273E+04, 4.3299E+04, 4.3324E+04, 4.3350E+04,
     +  4.3375E+04, 4.3400E+04, 4.3426E+04, 4.3451E+04, 4.3477E+04,
     + 4.3502E+04, 4.3528E+04, 4.3553E+04, 4.3579E+04, 4.3604E+04,
     +  4.3630E+04, 4.3655E+04, 4.3681E+04, 4.3706E+04, 4.3732E+04,
     + 4.3757E+04, 4.3783E+04, 4.3808E+04, 4.3834E+04, 4.3860E+04,
     +  4.3885E+04, 4.3911E+04, 4.3936E+04, 4.3962E+04, 4.3988E+04,
     + 4.4013E+04, 4.4039E+04, 4.4065E+04, 4.4090E+04, 4.4116E+04,
     +  4.4142E+04, 4.4167E+04, 4.4193E+04, 4.4219E+04, 4.4244E+04,
     + 4.4270E+04, 4.4296E+04, 4.4321E+04, 4.4347E+04, 4.4373E+04,
     +  4.4399E+04, 4.4424E+04, 4.4450E+04, 4.4476E+04, 4.4502E+04,
     + 4.4527E+04, 4.4553E+04, 4.4579E+04, 4.4605E+04, 4.4631E+04,
     +  4.4657E+04, 4.4682E+04, 4.4708E+04, 4.4734E+04, 4.4760E+04,
     + 4.4786E+04, 4.4812E+04, 4.4838E+04, 4.4863E+04, 4.4889E+04,
     +  4.4915E+04, 4.4941E+04, 4.4967E+04, 4.4993E+04, 4.5019E+04,
     + 4.5045E+04, 4.5071E+04, 4.5097E+04, 4.5123E+04, 4.5149E+04,
     +  4.5175E+04, 4.5201E+04, 4.5227E+04, 4.5253E+04, 4.5279E+04,
     + 4.5305E+04, 4.5331E+04, 4.5357E+04, 4.5383E+04, 4.5409E+04,
     +  4.5435E+04, 4.5461E+04, 4.5487E+04, 4.5513E+04, 4.5539E+04,
     + 4.5566E+04, 4.5592E+04, 4.5618E+04, 4.5644E+04, 4.5670E+04,
     +  4.5696E+04, 4.5722E+04, 4.5748E+04, 4.5775E+04, 4.5801E+04,
     + 4.5827E+04, 4.5853E+04, 4.5879E+04, 4.5906E+04, 4.5932E+04,
     +  4.5958E+04, 4.5984E+04, 4.6011E+04, 4.6037E+04, 4.6063E+04,
     + 4.6089E+04, 4.6116E+04, 4.6142E+04, 4.6168E+04, 4.6194E+04,
     +  4.6221E+04, 4.6247E+04, 4.6273E+04, 4.6300E+04, 4.6326E+04,
     + 4.6352E+04, 4.6379E+04, 4.6405E+04, 4.6432E+04, 4.6458E+04,
     +  4.6484E+04, 4.6511E+04, 4.6537E+04, 4.6563E+04, 4.6590E+04,
     + 4.6616E+04, 4.6643E+04, 4.6669E+04, 4.6696E+04, 4.6722E+04,
     +  4.6749E+04, 4.6775E+04, 4.6802E+04, 4.6828E+04, 4.6855E+04,
     + 4.6881E+04, 4.6908E+04, 4.6934E+04, 4.6961E+04, 4.6987E+04,
     +  4.7014E+04, 4.7040E+04, 4.7067E+04, 4.7093E+04, 4.7120E+04,
     + 4.7147E+04, 4.7173E+04, 4.7200E+04, 4.7226E+04, 4.7253E+04,
     +  4.7280E+04, 4.7306E+04, 4.7333E+04, 4.7359E+04, 4.7386E+04,
     + 4.7413E+04, 4.7439E+04, 4.7466E+04, 4.7493E+04, 4.7520E+04,
     +  4.7546E+04, 4.7573E+04, 4.7600E+04, 4.7626E+04, 4.7653E+04,
     + 4.7680E+04, 4.7707E+04, 4.7733E+04, 4.7760E+04, 4.7787E+04,
     +  4.7814E+04, 4.7841E+04, 4.7867E+04, 4.7894E+04, 4.7921E+04,
     + 4.7948E+04, 4.7975E+04, 4.8001E+04, 4.8028E+04, 4.8055E+04,
     +  4.8082E+04, 4.8109E+04, 4.8136E+04, 4.8163E+04, 4.8190E+04,
     + 4.8217E+04, 4.8243E+04, 4.8270E+04, 4.8297E+04, 4.8324E+04,
     +  4.8351E+04, 4.8378E+04, 4.8405E+04, 4.8432E+04, 4.8459E+04,
     + 4.8486E+04, 4.8513E+04, 4.8540E+04, 4.8567E+04, 4.8594E+04,
     +  4.8621E+04, 4.8648E+04, 4.8675E+04, 4.8702E+04, 4.8729E+04,
     + 4.8756E+04, 4.8783E+04, 4.8810E+04, 4.8838E+04, 4.8865E+04,
     +  4.8892E+04, 4.8919E+04, 4.8946E+04, 4.8973E+04, 4.9000E+04,
     + 4.9027E+04, 4.9055E+04, 4.9082E+04, 4.9109E+04, 4.9136E+04,
     +  4.9163E+04, 4.9190E+04, 4.9218E+04, 4.9245E+04, 4.9272E+04,
     + 4.9299E+04, 4.9327E+04, 4.9354E+04, 4.9381E+04, 4.9408E+04,
     +  4.9436E+04, 4.9463E+04, 4.9490E+04, 4.9517E+04, 4.9545E+04,
     + 4.9572E+04, 4.9599E+04, 4.9627E+04, 4.9654E+04, 4.9681E+04,
     +  4.9709E+04, 4.9736E+04, 4.9763E+04, 4.9791E+04, 4.9818E+04,
     + 4.9846E+04, 4.9873E+04, 4.9900E+04, 4.9928E+04, 4.9955E+04,
     +  4.9983E+04, 5.0010E+04, 5.0038E+04, 5.0065E+04, 5.0092E+04,
     + 5.0120E+04, 5.0147E+04, 5.0175E+04, 5.0202E+04, 5.0230E+04,
     +  5.0257E+04, 5.0285E+04, 5.0312E+04, 5.0340E+04, 5.0367E+04,
     + 5.0395E+04, 5.0423E+04, 5.0450E+04, 5.0478E+04, 5.0505E+04,
     +  5.0533E+04, 5.0560E+04, 5.0588E+04, 5.0616E+04, 5.0643E+04,
     + 5.0671E+04, 5.0699E+04, 5.0726E+04, 5.0754E+04, 5.0782E+04,
     +  5.0809E+04, 5.0837E+04, 5.0865E+04, 5.0892E+04, 5.0920E+04,
     + 5.0948E+04, 5.0975E+04, 5.1003E+04, 5.1031E+04, 5.1059E+04,
     +  5.1086E+04, 5.1114E+04, 5.1142E+04, 5.1170E+04, 5.1197E+04,
     + 5.1225E+04, 5.1253E+04, 5.1281E+04, 5.1309E+04, 5.1336E+04,
     +  5.1364E+04, 5.1392E+04, 5.1420E+04, 5.1448E+04, 5.1476E+04,
     + 5.1504E+04, 5.1531E+04, 5.1559E+04, 5.1587E+04, 5.1615E+04,
     +  5.1643E+04, 5.1671E+04, 5.1699E+04, 5.1727E+04, 5.1755E+04,
     + 5.1783E+04, 5.1811E+04, 5.1839E+04, 5.1867E+04, 5.1895E+04,
     +  5.1923E+04, 5.1951E+04, 5.1979E+04, 5.2007E+04, 5.2035E+04,
     + 5.2063E+04, 5.2091E+04, 5.2119E+04, 5.2147E+04, 5.2175E+04,
     +  5.2203E+04, 5.2231E+04, 5.2259E+04, 5.2287E+04, 5.2315E+04,
     + 5.2343E+04, 5.2371E+04, 5.2400E+04, 5.2428E+04, 5.2456E+04,
     +  5.2484E+04, 5.2512E+04, 5.2540E+04, 5.2569E+04, 5.2597E+04,
     + 5.2625E+04, 5.2653E+04, 5.2681E+04, 5.2710E+04, 5.2738E+04,
     +  5.2766E+04, 5.2794E+04, 5.2822E+04, 5.2851E+04, 5.2879E+04,
     + 5.2907E+04, 5.2936E+04, 5.2964E+04, 5.2992E+04, 5.3020E+04,
     +  5.3049E+04, 5.3077E+04, 5.3105E+04, 5.3134E+04, 5.3162E+04,
     + 5.3190E+04, 5.3219E+04, 5.3247E+04, 5.3276E+04, 5.3304E+04,
     +  5.3332E+04, 5.3361E+04, 5.3389E+04, 5.3418E+04, 5.3446E+04,
     + 5.3474E+04, 5.3503E+04, 5.3531E+04, 5.3560E+04, 5.3588E+04,
     +  5.3617E+04, 5.3645E+04, 5.3674E+04, 5.3702E+04, 5.3731E+04,
     + 5.3759E+04, 5.3788E+04, 5.3816E+04, 5.3845E+04, 5.3873E+04,
     +  5.3902E+04, 5.3930E+04, 5.3959E+04, 5.3988E+04, 5.4016E+04,
     + 5.4045E+04, 5.4073E+04, 5.4102E+04, 5.4131E+04, 5.4159E+04,
     +  5.4188E+04, 5.4216E+04, 5.4245E+04, 5.4274E+04, 5.4302E+04,
     + 5.4331E+04, 5.4360E+04, 5.4388E+04, 5.4417E+04, 5.4446E+04,
     +  5.4475E+04, 5.4503E+04, 5.4532E+04, 5.4561E+04, 5.4590E+04,
     + 5.4618E+04, 5.4647E+04, 5.4676E+04, 5.4705E+04, 5.4733E+04,
     +  5.4762E+04, 5.4791E+04, 5.4820E+04, 5.4849E+04, 5.4877E+04,
     + 5.4906E+04, 5.4935E+04, 5.4964E+04, 5.4993E+04, 5.5022E+04,
     +  5.5051E+04, 5.5079E+04, 5.5108E+04, 5.5137E+04, 5.5166E+04,
     + 5.5195E+04, 5.5224E+04, 5.5253E+04, 5.5282E+04, 5.5311E+04,
     +  5.5340E+04, 5.5369E+04, 5.5398E+04, 5.5427E+04, 5.5456E+04,
     + 5.5485E+04, 5.5514E+04, 5.5543E+04, 5.5572E+04, 5.5601E+04,
     +  5.5630E+04, 5.5659E+04, 5.5688E+04, 5.5717E+04, 5.5746E+04,
     + 5.5775E+04, 5.5804E+04, 5.5833E+04, 5.5862E+04, 5.5892E+04,
     +  5.5921E+04, 5.5950E+04, 5.5979E+04, 5.6008E+04, 5.6037E+04,
     + 5.6066E+04, 5.6096E+04, 5.6125E+04, 5.6154E+04, 5.6183E+04,
     +  5.6212E+04, 5.6241E+04, 5.6271E+04, 5.6300E+04, 5.6329E+04,
     + 5.6358E+04, 5.6388E+04, 5.6417E+04, 5.6446E+04, 5.6475E+04,
     +  5.6505E+04, 5.6534E+04, 5.6563E+04, 5.6593E+04, 5.6622E+04,
     + 5.6651E+04, 5.6681E+04, 5.6710E+04, 5.6739E+04, 5.6769E+04,
     +  5.6798E+04, 5.6827E+04, 5.6857E+04, 5.6886E+04, 5.6915E+04,
     + 5.6945E+04, 5.6974E+04, 5.7004E+04, 5.7033E+04, 5.7063E+04,
     +  5.7092E+04, 5.7121E+04, 5.7151E+04, 5.7180E+04, 5.7210E+04,
     + 5.7239E+04, 5.7269E+04, 5.7298E+04, 5.7328E+04, 5.7357E+04,
     +  5.7387E+04, 5.7416E+04, 5.7446E+04, 5.7475E+04, 5.7505E+04,
     + 5.7535E+04, 5.7564E+04, 5.7594E+04, 5.7623E+04, 5.7653E+04,
     +  5.7683E+04, 5.7712E+04, 5.7742E+04, 5.7771E+04, 5.7801E+04,
     + 5.7831E+04, 5.7860E+04, 5.7890E+04, 5.7920E+04, 5.7949E+04,
     +  5.7979E+04, 5.8009E+04, 5.8038E+04, 5.8068E+04, 5.8098E+04,
     + 5.8128E+04, 5.8157E+04, 5.8187E+04, 5.8217E+04, 5.8246E+04,
     +  5.8276E+04, 5.8306E+04, 5.8336E+04, 5.8366E+04, 5.8395E+04,
     + 5.8425E+04, 5.8455E+04, 5.8485E+04, 5.8515E+04, 5.8545E+04,
     +  5.8574E+04, 5.8604E+04, 5.8634E+04, 5.8664E+04, 5.8694E+04,
     + 5.8724E+04, 5.8754E+04, 5.8783E+04, 5.8813E+04, 5.8843E+04,
     +  5.8873E+04, 5.8903E+04, 5.8933E+04, 5.8963E+04, 5.8993E+04,
     + 5.9023E+04, 5.9053E+04, 5.9083E+04, 5.9113E+04, 5.9143E+04,
     +  5.9173E+04, 5.9203E+04, 5.9233E+04, 5.9263E+04, 5.9293E+04,
     + 5.9323E+04, 5.9353E+04, 5.9383E+04, 5.9413E+04, 5.9443E+04,
     +  5.9473E+04, 5.9504E+04, 5.9534E+04, 5.9564E+04, 5.9594E+04,
     + 5.9624E+04, 5.9654E+04, 5.9684E+04, 5.9714E+04, 5.9745E+04,
     +  5.9775E+04, 5.9805E+04, 5.9835E+04, 5.9865E+04, 5.9896E+04,
     + 5.9926E+04, 5.9956E+04, 5.9986E+04, 6.0016E+04, 6.0047E+04,
     +  6.0077E+04, 6.0107E+04, 6.0137E+04, 6.0168E+04, 6.0198E+04,
     + 6.0228E+04, 6.0259E+04, 6.0289E+04, 6.0319E+04, 6.0350E+04,
     +  6.0380E+04, 6.0410E+04, 6.0441E+04, 6.0471E+04, 6.0501E+04,
     + 6.0532E+04, 6.0562E+04, 6.0592E+04, 6.0623E+04, 6.0653E+04,
     +  6.0684E+04, 6.0714E+04, 6.0744E+04, 6.0775E+04, 6.0805E+04,
     + 6.0836E+04, 6.0866E+04, 6.0897E+04, 6.0927E+04, 6.0958E+04,
     +  6.0988E+04, 6.1019E+04, 6.1049E+04, 6.1080E+04, 6.1110E+04,
     + 6.1141E+04, 6.1171E+04, 6.1202E+04, 6.1232E+04, 6.1263E+04,
     +  6.1294E+04, 6.1324E+04, 6.1355E+04, 6.1385E+04, 6.1416E+04,
     + 6.1447E+04, 6.1477E+04, 6.1508E+04, 6.1538E+04, 6.1569E+04,
     +  6.1600E+04, 6.1630E+04, 6.1661E+04, 6.1692E+04, 6.1722E+04,
     + 6.1753E+04, 6.1784E+04, 6.1815E+04, 6.1845E+04, 6.1876E+04,
     +  6.1907E+04, 6.1937E+04, 6.1968E+04, 6.1999E+04, 6.2030E+04,
     + 6.2061E+04, 6.2091E+04, 6.2122E+04, 6.2153E+04, 6.2184E+04,
     +  6.2215E+04, 6.2245E+04, 6.2276E+04, 6.2307E+04, 6.2338E+04,
     + 6.2369E+04, 6.2400E+04, 6.2430E+04, 6.2461E+04, 6.2492E+04,
     +  6.2523E+04, 6.2554E+04, 6.2585E+04, 6.2616E+04, 6.2647E+04,
     + 6.2678E+04, 6.2709E+04, 6.2740E+04, 6.2771E+04, 6.2802E+04,
     +  6.2833E+04, 6.2864E+04, 6.2895E+04, 6.2926E+04, 6.2957E+04,
     + 6.2988E+04, 6.3019E+04, 6.3050E+04, 6.3081E+04, 6.3112E+04,
     +  6.3143E+04, 6.3174E+04, 6.3205E+04, 6.3236E+04, 6.3267E+04,
     + 6.3298E+04, 6.3329E+04, 6.3360E+04, 6.3392E+04, 6.3423E+04,
     +  6.3454E+04, 6.3485E+04, 6.3516E+04, 6.3547E+04, 6.3578E+04,
     + 6.3610E+04, 6.3641E+04, 6.3672E+04, 6.3703E+04, 6.3734E+04,
     +  6.3766E+04, 6.3797E+04, 6.3828E+04, 6.3859E+04, 6.3891E+04,
     + 6.3922E+04, 6.3953E+04, 6.3984E+04, 6.4016E+04, 6.4047E+04,
     +  6.4078E+04, 6.4110E+04, 6.4141E+04, 6.4172E+04, 6.4204E+04,
     + 6.4235E+04, 6.4266E+04, 6.4298E+04, 6.4329E+04, 6.4360E+04,
     +  6.4392E+04, 6.4423E+04, 6.4455E+04, 6.4486E+04, 6.4517E+04,
     + 6.4549E+04, 6.4580E+04, 6.4612E+04, 6.4643E+04, 6.4675E+04,
     +  6.4706E+04, 6.4738E+04, 6.4769E+04, 6.4801E+04, 6.4832E+04,
     + 6.4864E+04, 6.4895E+04, 6.4927E+04, 6.4958E+04, 6.4990E+04,
     +  6.5021E+04, 6.5053E+04, 6.5084E+04, 6.5116E+04, 6.5147E+04,
     + 6.5179E+04, 6.5211E+04, 6.5242E+04, 6.5274E+04, 6.5305E+04,
     +  6.5337E+04, 6.5369E+04, 6.5400E+04, 6.5432E+04, 6.5464E+04,
     + 6.5495E+04, 6.5527E+04, 6.5559E+04, 6.5590E+04, 6.5622E+04,
     +  6.5654E+04, 6.5686E+04, 6.5717E+04, 6.5749E+04, 6.5781E+04,
     + 6.5812E+04, 6.5844E+04, 6.5876E+04, 6.5908E+04, 6.5940E+04,
     +  6.5971E+04, 6.6003E+04, 6.6035E+04, 6.6067E+04, 6.6099E+04,
     + 6.6130E+04, 6.6162E+04, 6.6194E+04, 6.6226E+04, 6.6258E+04,
     +  6.6290E+04, 6.6322E+04, 6.6353E+04, 6.6385E+04, 6.6417E+04,
     + 6.6449E+04, 6.6481E+04, 6.6513E+04, 6.6545E+04, 6.6577E+04,
     +  6.6609E+04, 6.6641E+04, 6.6673E+04, 6.6705E+04, 6.6737E+04,
     + 6.6769E+04, 6.6801E+04, 6.6833E+04, 6.6865E+04, 6.6897E+04,
     +  6.6929E+04, 6.6961E+04, 6.6993E+04, 6.7025E+04, 6.7057E+04,
     + 6.7089E+04, 6.7121E+04, 6.7153E+04, 6.7185E+04, 6.7217E+04,
     +  6.7250E+04, 6.7282E+04, 6.7314E+04, 6.7346E+04, 6.7378E+04,
     + 6.7410E+04, 6.7442E+04, 6.7475E+04, 6.7507E+04, 6.7539E+04,
     +  6.7571E+04, 6.7603E+04, 6.7636E+04, 6.7668E+04, 6.7700E+04,
     + 6.7732E+04, 6.7765E+04, 6.7797E+04, 6.7829E+04, 6.7861E+04,
     +  6.7894E+04, 6.7926E+04, 6.7958E+04, 6.7990E+04, 6.8023E+04,
     + 6.8055E+04, 6.8087E+04, 6.8120E+04, 6.8152E+04, 6.8184E+04,
     +  6.8217E+04, 6.8249E+04, 6.8282E+04, 6.8314E+04, 6.8346E+04,
     + 6.8379E+04, 6.8411E+04, 6.8443E+04, 6.8476E+04, 6.8508E+04,
     +  6.8541E+04, 6.8573E+04, 6.8606E+04, 6.8638E+04, 6.8671E+04,
     + 6.8703E+04, 6.8736E+04, 6.8768E+04, 6.8801E+04, 6.8833E+04,
     +  6.8866E+04, 6.8898E+04, 6.8931E+04, 6.8963E+04, 6.8996E+04,
     + 6.9028E+04, 6.9061E+04, 6.9093E+04, 6.9126E+04, 6.9159E+04,
     +  6.9191E+04, 6.9224E+04, 6.9256E+04, 6.9289E+04, 6.9322E+04,
     + 6.9354E+04, 6.9387E+04, 6.9420E+04, 6.9452E+04, 6.9485E+04,
     +  6.9518E+04, 6.9550E+04, 6.9583E+04, 6.9616E+04, 6.9649E+04,
     + 6.9681E+04, 6.9714E+04, 6.9747E+04, 6.9779E+04, 6.9812E+04,
     +  6.9845E+04, 6.9878E+04, 6.9911E+04, 6.9943E+04, 6.9976E+04,
     + 7.0009E+04, 7.0042E+04, 7.0075E+04, 7.0107E+04, 7.0140E+04,
     +  7.0173E+04, 7.0206E+04, 7.0239E+04, 7.0272E+04, 7.0304E+04,
     + 7.0337E+04, 7.0370E+04, 7.0403E+04, 7.0436E+04, 7.0469E+04,
     +  7.0502E+04, 7.0535E+04, 7.0568E+04, 7.0601E+04, 7.0634E+04,
     + 7.0667E+04, 7.0700E+04, 7.0733E+04, 7.0766E+04, 7.0799E+04,
     +  7.0832E+04, 7.0865E+04, 7.0898E+04, 7.0931E+04, 7.0964E+04,
     + 7.0997E+04, 7.1030E+04, 7.1063E+04, 7.1096E+04, 7.1129E+04,
     +  7.1162E+04, 7.1195E+04, 7.1228E+04, 7.1261E+04, 7.1295E+04,
     + 7.1328E+04, 7.1361E+04, 7.1394E+04, 7.1427E+04, 7.1460E+04,
     +  7.1493E+04, 7.1527E+04, 7.1560E+04, 7.1593E+04, 7.1626E+04,
     + 7.1659E+04, 7.1693E+04, 7.1726E+04, 7.1759E+04, 7.1792E+04,
     +  7.1826E+04, 7.1859E+04, 7.1892E+04, 7.1925E+04, 7.1959E+04,
     + 7.1992E+04, 7.2025E+04, 7.2059E+04, 7.2092E+04, 7.2125E+04,
     +  7.2159E+04, 7.2192E+04, 7.2225E+04, 7.2259E+04, 7.2292E+04,
     + 7.2325E+04, 7.2359E+04, 7.2392E+04, 7.2425E+04, 7.2459E+04,
     +  7.2492E+04, 7.2526E+04, 7.2559E+04, 7.2593E+04, 7.2626E+04,
     + 7.2659E+04, 7.2693E+04, 7.2726E+04, 7.2760E+04, 7.2793E+04,
     +  7.2827E+04, 7.2860E+04, 7.2894E+04, 7.2927E+04, 7.2961E+04,
     + 7.2994E+04, 7.3028E+04, 7.3062E+04, 7.3095E+04, 7.3129E+04,
     +  7.3162E+04, 7.3196E+04, 7.3229E+04, 7.3263E+04, 7.3297E+04,
     + 7.3330E+04, 7.3364E+04, 7.3397E+04, 7.3431E+04, 7.3465E+04,
     +  7.3498E+04, 7.3532E+04, 7.3566E+04, 7.3599E+04, 7.3633E+04,
     + 7.3667E+04, 7.3701E+04, 7.3734E+04, 7.3768E+04, 7.3802E+04,
     +  7.3835E+04, 7.3869E+04, 7.3903E+04, 7.3937E+04, 7.3970E+04,
     + 7.4004E+04, 7.4038E+04, 7.4072E+04, 7.4106E+04, 7.4139E+04,
     +  7.4173E+04, 7.4207E+04, 7.4241E+04, 7.4275E+04, 7.4309E+04,
     + 7.4342E+04, 7.4376E+04, 7.4410E+04, 7.4444E+04, 7.4478E+04,
     +  7.4512E+04, 7.4546E+04, 7.4580E+04, 7.4614E+04, 7.4648E+04,
     + 7.4681E+04, 7.4715E+04, 7.4749E+04, 7.4783E+04, 7.4817E+04,
     +  7.4851E+04, 7.4885E+04, 7.4919E+04, 7.4953E+04, 7.4987E+04,
     + 7.5021E+04, 7.5055E+04, 7.5089E+04, 7.5123E+04, 7.5157E+04,
     +  7.5192E+04, 7.5226E+04, 7.5260E+04, 7.5294E+04, 7.5328E+04,
     + 7.5362E+04, 7.5396E+04, 7.5430E+04, 7.5464E+04, 7.5498E+04,
     +  7.5533E+04, 7.5567E+04, 7.5601E+04, 7.5635E+04, 7.5669E+04,
     + 7.5703E+04, 7.5738E+04, 7.5772E+04, 7.5806E+04, 7.5840E+04,
     +  7.5874E+04, 7.5909E+04, 7.5943E+04, 7.5977E+04, 7.6011E+04,
     + 7.6046E+04, 7.6080E+04, 7.6114E+04, 7.6149E+04, 7.6183E+04,
     +  7.6217E+04, 7.6251E+04, 7.6286E+04, 7.6320E+04, 7.6354E+04,
     + 7.6389E+04, 7.6423E+04, 7.6458E+04, 7.6492E+04, 7.6526E+04,
     +  7.6561E+04, 7.6595E+04, 7.6629E+04, 7.6664E+04, 7.6698E+04,
     + 7.6733E+04, 7.6767E+04, 7.6802E+04, 7.6836E+04, 7.6870E+04,
     +  7.6905E+04, 7.6939E+04, 7.6974E+04, 7.7008E+04, 7.7043E+04,
     + 7.7077E+04, 7.7112E+04, 7.7146E+04, 7.7181E+04, 7.7216E+04,
     +  7.7250E+04, 7.7285E+04, 7.7319E+04, 7.7354E+04, 7.7388E+04,
     + 7.7423E+04, 7.7458E+04, 7.7492E+04, 7.7527E+04, 7.7561E+04,
     +  7.7596E+04, 7.7631E+04, 7.7665E+04, 7.7700E+04, 7.7735E+04,
     + 7.7769E+04, 7.7804E+04, 7.7839E+04, 7.7873E+04, 7.7908E+04,
     +  7.7943E+04, 7.7977E+04, 7.8012E+04, 7.8047E+04, 7.8082E+04,
     + 7.8116E+04, 7.8151E+04, 7.8186E+04, 7.8221E+04, 7.8255E+04,
     +  7.8290E+04, 7.8325E+04, 7.8360E+04, 7.8395E+04, 7.8430E+04,
     + 7.8464E+04, 7.8499E+04, 7.8534E+04, 7.8569E+04, 7.8604E+04,
     +  7.8639E+04, 7.8674E+04, 7.8708E+04, 7.8743E+04, 7.8778E+04,
     + 7.8813E+04, 7.8848E+04, 7.8883E+04, 7.8918E+04, 7.8953E+04,
     +  7.8988E+04, 7.9023E+04, 7.9058E+04, 7.9093E+04, 7.9128E+04,
     + 7.9163E+04, 7.9198E+04, 7.9233E+04, 7.9268E+04, 7.9303E+04,
     +  7.9338E+04, 7.9373E+04, 7.9408E+04, 7.9443E+04, 7.9478E+04,
     + 7.9513E+04, 7.9548E+04, 7.9583E+04, 7.9618E+04, 7.9653E+04,
     +  7.9689E+04, 7.9724E+04, 7.9759E+04, 7.9794E+04, 7.9829E+04,
     + 7.9864E+04, 7.9899E+04, 7.9935E+04, 7.9970E+04, 8.0005E+04,
     +  8.0040E+04, 8.0075E+04, 8.0111E+04, 8.0146E+04, 8.0181E+04,
     + 8.0216E+04, 8.0252E+04, 8.0287E+04, 8.0322E+04, 8.0357E+04,
     +  8.0393E+04, 8.0428E+04, 8.0463E+04, 8.0498E+04, 8.0534E+04,
     + 8.0569E+04, 8.0604E+04, 8.0640E+04, 8.0675E+04, 8.0710E+04,
     +  8.0746E+04, 8.0781E+04, 8.0817E+04, 8.0852E+04, 8.0887E+04,
     + 8.0923E+04, 8.0958E+04, 8.0994E+04, 8.1029E+04, 8.1064E+04,
     +  8.1100E+04, 8.1135E+04, 8.1171E+04, 8.1206E+04, 8.1242E+04,
     + 8.1277E+04, 8.1313E+04, 8.1348E+04, 8.1384E+04, 8.1419E+04,
     +  8.1455E+04, 8.1490E+04, 8.1526E+04, 8.1561E+04, 8.1597E+04,
     + 8.1632E+04, 8.1668E+04, 8.1704E+04, 8.1739E+04, 8.1775E+04,
     +  8.1810E+04, 8.1846E+04, 8.1882E+04, 8.1917E+04, 8.1953E+04,
     + 8.1988E+04, 8.2024E+04, 8.2060E+04, 8.2095E+04, 8.2131E+04,
     +  8.2167E+04, 8.2202E+04, 8.2238E+04, 8.2274E+04, 8.2310E+04,
     + 8.2345E+04, 8.2381E+04, 8.2417E+04, 8.2453E+04, 8.2488E+04,
     +  8.2524E+04, 8.2560E+04, 8.2596E+04, 8.2631E+04, 8.2667E+04,
     + 8.2703E+04, 8.2739E+04, 8.2775E+04, 8.2810E+04, 8.2846E+04,
     +  8.2882E+04, 8.2918E+04, 8.2954E+04, 8.2990E+04, 8.3026E+04,
     + 8.3061E+04, 8.3097E+04, 8.3133E+04, 8.3169E+04, 8.3205E+04,
     +  8.3241E+04, 8.3277E+04, 8.3313E+04, 8.3349E+04, 8.3385E+04,
     + 8.3421E+04, 8.3457E+04, 8.3493E+04, 8.3529E+04, 8.3565E+04,
     +  8.3601E+04, 8.3637E+04, 8.3673E+04, 8.3709E+04, 8.3745E+04,
     + 8.3781E+04, 8.3817E+04, 8.3853E+04, 8.3889E+04, 8.3925E+04,
     +  8.3961E+04, 8.3997E+04, 8.4033E+04, 8.4070E+04, 8.4106E+04,
     + 8.4142E+04, 8.4178E+04, 8.4214E+04, 8.4250E+04, 8.4286E+04,
     +  8.4323E+04, 8.4359E+04, 8.4395E+04, 8.4431E+04, 8.4467E+04,
     + 8.4504E+04, 8.4540E+04, 8.4576E+04, 8.4612E+04, 8.4648E+04,
     +  8.4685E+04, 8.4721E+04, 8.4757E+04, 8.4794E+04, 8.4830E+04,
     + 8.4866E+04, 8.4902E+04, 8.4939E+04, 8.4975E+04/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_CH4   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 4),QofT( 4,NT),Q(NT)
      data (Tmx(I),I=1,4)/3500.,3500.,4500.,3500./
      data xgj/ 1.,2.,3.,6./
c...      CH4
c...        --       211
      data (QofT( 1,J),J=1,2408)/ 5.0000E+00, 5.0048E+00, 5.0590E+00,
     +  5.2077E+00, 5.4438E+00, 5.7421E+00, 6.0829E+00, 6.4547E+00,
     + 6.8519E+00, 7.2721E+00, 7.7144E+00, 8.1788E+00, 8.6655E+00,
     +  9.1750E+00, 9.7080E+00, 1.0265E+01, 1.0846E+01, 1.1452E+01,
     + 1.2082E+01, 1.2737E+01, 1.4121E+01, 1.5602E+01, 1.7176E+01,
     +  1.8841E+01, 2.0593E+01, 2.2425E+01, 2.4336E+01, 2.6320E+01,
     + 2.8374E+01, 3.0495E+01, 3.2678E+01, 3.4922E+01, 3.7223E+01,
     +  3.9580E+01, 4.1991E+01, 4.4453E+01, 4.6965E+01, 4.9526E+01,
     + 5.2133E+01, 5.4788E+01, 5.7487E+01, 6.0230E+01, 6.3017E+01,
     +  6.5847E+01, 6.8718E+01, 7.1630E+01, 7.4582E+01, 7.7575E+01,
     + 8.0607E+01, 8.3678E+01, 8.6786E+01, 8.9933E+01, 9.3117E+01,
     +  9.6338E+01, 9.9595E+01, 1.0289E+02, 1.0622E+02, 1.0958E+02,
     + 1.1298E+02, 1.1642E+02, 1.1988E+02, 1.2339E+02, 1.2692E+02,
     +  1.3049E+02, 1.3409E+02, 1.3773E+02, 1.4139E+02, 1.4509E+02,
     + 1.4882E+02, 1.5258E+02, 1.5638E+02, 1.6020E+02, 1.6406E+02,
     +  1.6794E+02, 1.7186E+02, 1.7581E+02, 1.7978E+02, 1.8379E+02,
     + 1.8782E+02, 1.9189E+02, 1.9598E+02, 2.0010E+02, 2.0425E+02,
     +  2.0843E+02, 2.1264E+02, 2.1688E+02, 2.2114E+02, 2.2543E+02,
     + 2.2975E+02, 2.3410E+02, 2.3848E+02, 2.4288E+02, 2.4731E+02,
     +  2.5176E+02, 2.5624E+02, 2.6075E+02, 2.6529E+02, 2.6985E+02,
     + 2.7444E+02, 2.7905E+02, 2.8369E+02, 2.8836E+02, 2.9305E+02,
     +  2.9777E+02, 3.0251E+02, 3.0728E+02, 3.1208E+02, 3.1690E+02,
     + 3.2175E+02, 3.2662E+02, 3.3151E+02, 3.3644E+02, 3.4138E+02,
     +  3.4636E+02, 3.5135E+02, 3.5638E+02, 3.6142E+02, 3.6650E+02,
     + 3.7159E+02, 3.7672E+02, 3.8187E+02, 3.8704E+02, 3.9224E+02,
     +  3.9746E+02, 4.0271E+02, 4.0798E+02, 4.1328E+02, 4.1861E+02,
     + 4.2396E+02, 4.2933E+02, 4.3473E+02, 4.4016E+02, 4.4561E+02,
     +  4.5109E+02, 4.5659E+02, 4.6212E+02, 4.6767E+02, 4.7325E+02,
     + 4.7886E+02, 4.8449E+02, 4.9015E+02, 4.9583E+02, 5.0154E+02,
     +  5.0728E+02, 5.1305E+02, 5.1884E+02, 5.2466E+02, 5.3050E+02,
     + 5.3637E+02, 5.4227E+02, 5.4820E+02, 5.5416E+02, 5.6014E+02,
     +  5.6615E+02, 5.7219E+02, 5.7826E+02, 5.8435E+02, 5.9048E+02,
     + 5.9663E+02, 6.0282E+02, 6.0903E+02, 6.1527E+02, 6.2154E+02,
     +  6.2784E+02, 6.3418E+02, 6.4054E+02, 6.4693E+02, 6.5335E+02,
     + 6.5981E+02, 6.6630E+02, 6.7281E+02, 6.7936E+02, 6.8595E+02,
     +  6.9256E+02, 6.9921E+02, 7.0589E+02, 7.1260E+02, 7.1935E+02,
     + 7.2612E+02, 7.3294E+02, 7.3979E+02, 7.4667E+02, 7.5359E+02,
     +  7.6054E+02, 7.6752E+02, 7.7455E+02, 7.8161E+02, 7.8870E+02,
     + 7.9583E+02, 8.0300E+02, 8.1021E+02, 8.1745E+02, 8.2473E+02,
     +  8.3205E+02, 8.3941E+02, 8.4680E+02, 8.5424E+02, 8.6171E+02,
     + 8.6923E+02, 8.7678E+02, 8.8437E+02, 8.9201E+02, 8.9969E+02,
     +  9.0740E+02, 9.1516E+02, 9.2296E+02, 9.3081E+02, 9.3869E+02,
     + 9.4662E+02, 9.5460E+02, 9.6261E+02, 9.7067E+02, 9.7878E+02,
     +  9.8693E+02, 9.9513E+02, 1.0034E+03, 1.0117E+03, 1.0200E+03,
     + 1.0284E+03, 1.0368E+03, 1.0453E+03, 1.0538E+03, 1.0624E+03,
     +  1.0710E+03, 1.0797E+03, 1.0884E+03, 1.0972E+03, 1.1060E+03,
     + 1.1149E+03, 1.1238E+03, 1.1328E+03, 1.1418E+03, 1.1509E+03,
     +  1.1600E+03, 1.1692E+03, 1.1784E+03, 1.1877E+03, 1.1971E+03,
     + 1.2065E+03, 1.2159E+03, 1.2254E+03, 1.2350E+03, 1.2446E+03,
     +  1.2543E+03, 1.2641E+03, 1.2739E+03, 1.2837E+03, 1.2937E+03,
     + 1.3036E+03, 1.3137E+03, 1.3238E+03, 1.3339E+03, 1.3442E+03,
     +  1.3544E+03, 1.3648E+03, 1.3752E+03, 1.3857E+03, 1.3962E+03,
     + 1.4068E+03, 1.4175E+03, 1.4282E+03, 1.4390E+03, 1.4499E+03,
     +  1.4608E+03, 1.4718E+03, 1.4829E+03, 1.4940E+03, 1.5052E+03,
     + 1.5165E+03, 1.5278E+03, 1.5392E+03, 1.5507E+03, 1.5623E+03,
     +  1.5739E+03, 1.5856E+03, 1.5974E+03, 1.6092E+03, 1.6212E+03,
     + 1.6331E+03, 1.6452E+03, 1.6574E+03, 1.6696E+03, 1.6819E+03,
     +  1.6943E+03, 1.7067E+03, 1.7193E+03, 1.7319E+03, 1.7446E+03,
     + 1.7573E+03, 1.7702E+03, 1.7831E+03, 1.7961E+03, 1.8092E+03,
     +  1.8224E+03, 1.8357E+03, 1.8491E+03, 1.8625E+03, 1.8760E+03,
     + 1.8896E+03, 1.9033E+03, 1.9171E+03, 1.9310E+03, 1.9449E+03,
     +  1.9590E+03, 1.9731E+03, 1.9874E+03, 2.0017E+03, 2.0161E+03,
     + 2.0306E+03, 2.0452E+03, 2.0599E+03, 2.0747E+03, 2.0896E+03,
     +  2.1046E+03, 2.1196E+03, 2.1348E+03, 2.1501E+03, 2.1655E+03,
     + 2.1809E+03, 2.1965E+03, 2.2122E+03, 2.2279E+03, 2.2438E+03,
     +  2.2598E+03, 2.2759E+03, 2.2921E+03, 2.3084E+03, 2.3248E+03,
     + 2.3413E+03, 2.3579E+03, 2.3746E+03, 2.3914E+03, 2.4084E+03,
     +  2.4254E+03, 2.4426E+03, 2.4598E+03, 2.4772E+03, 2.4947E+03,
     + 2.5123E+03, 2.5300E+03, 2.5479E+03, 2.5658E+03, 2.5839E+03,
     +  2.6021E+03, 2.6204E+03, 2.6388E+03, 2.6574E+03, 2.6760E+03,
     + 2.6948E+03, 2.7138E+03, 2.7328E+03, 2.7520E+03, 2.7712E+03,
     +  2.7907E+03, 2.8102E+03, 2.8299E+03, 2.8497E+03, 2.8696E+03,
     + 2.8896E+03, 2.9098E+03, 2.9302E+03, 2.9506E+03, 2.9712E+03,
     +  2.9919E+03, 3.0128E+03, 3.0338E+03, 3.0549E+03, 3.0762E+03,
     + 3.0976E+03, 3.1191E+03, 3.1408E+03, 3.1627E+03, 3.1846E+03,
     +  3.2068E+03, 3.2290E+03, 3.2515E+03, 3.2740E+03, 3.2967E+03,
     + 3.3196E+03, 3.3426E+03, 3.3657E+03, 3.3891E+03, 3.4125E+03,
     +  3.4361E+03, 3.4599E+03, 3.4838E+03, 3.5079E+03, 3.5322E+03,
     + 3.5566E+03, 3.5811E+03, 3.6058E+03, 3.6307E+03, 3.6558E+03,
     +  3.6810E+03, 3.7064E+03, 3.7319E+03, 3.7576E+03, 3.7835E+03,
     + 3.8096E+03, 3.8358E+03, 3.8622E+03, 3.8887E+03, 3.9155E+03,
     +  3.9424E+03, 3.9695E+03, 3.9968E+03, 4.0242E+03, 4.0519E+03,
     + 4.0797E+03, 4.1077E+03, 4.1358E+03, 4.1642E+03, 4.1928E+03,
     +  4.2215E+03, 4.2504E+03, 4.2795E+03, 4.3088E+03, 4.3383E+03,
     + 4.3680E+03, 4.3979E+03, 4.4280E+03, 4.4583E+03, 4.4887E+03,
     +  4.5194E+03, 4.5503E+03, 4.5814E+03, 4.6126E+03, 4.6441E+03,
     + 4.6758E+03, 4.7077E+03, 4.7398E+03, 4.7722E+03, 4.8047E+03,
     +  4.8374E+03, 4.8704E+03, 4.9036E+03, 4.9370E+03, 4.9706E+03,
     + 5.0044E+03, 5.0384E+03, 5.0727E+03, 5.1072E+03, 5.1419E+03,
     +  5.1769E+03, 5.2121E+03, 5.2475E+03, 5.2831E+03, 5.3190E+03,
     + 5.3551E+03, 5.3914E+03, 5.4280E+03, 5.4648E+03, 5.5019E+03,
     +  5.5392E+03, 5.5767E+03, 5.6145E+03, 5.6525E+03, 5.6908E+03,
     + 5.7293E+03, 5.7681E+03, 5.8072E+03, 5.8464E+03, 5.8860E+03,
     +  5.9258E+03, 5.9658E+03, 6.0062E+03, 6.0468E+03, 6.0876E+03,
     + 6.1287E+03, 6.1701E+03, 6.2117E+03, 6.2537E+03, 6.2959E+03,
     +  6.3383E+03, 6.3811E+03, 6.4241E+03, 6.4674E+03, 6.5110E+03,
     + 6.5548E+03, 6.5990E+03, 6.6434E+03, 6.6881E+03, 6.7331E+03,
     +  6.7785E+03, 6.8241E+03, 6.8699E+03, 6.9161E+03, 6.9626E+03,
     + 7.0094E+03, 7.0565E+03, 7.1039E+03, 7.1516E+03, 7.1996E+03,
     +  7.2480E+03, 7.2966E+03, 7.3456E+03, 7.3948E+03, 7.4444E+03,
     + 7.4943E+03, 7.5445E+03, 7.5951E+03, 7.6460E+03, 7.6972E+03,
     +  7.7487E+03, 7.8006E+03, 7.8528E+03, 7.9054E+03, 7.9582E+03,
     + 8.0115E+03, 8.0650E+03, 8.1189E+03, 8.1732E+03, 8.2278E+03,
     +  8.2828E+03, 8.3381E+03, 8.3937E+03, 8.4498E+03, 8.5062E+03,
     + 8.5629E+03, 8.6200E+03, 8.6775E+03, 8.7353E+03, 8.7936E+03,
     +  8.8522E+03, 8.9111E+03, 8.9705E+03, 9.0302E+03, 9.0903E+03,
     + 9.1508E+03, 9.2117E+03, 9.2730E+03, 9.3346E+03, 9.3967E+03,
     +  9.4592E+03, 9.5220E+03, 9.5853E+03, 9.6490E+03, 9.7130E+03,
     + 9.7775E+03, 9.8424E+03, 9.9077E+03, 9.9734E+03, 1.0040E+04,
     +  1.0106E+04, 1.0173E+04, 1.0241E+04, 1.0308E+04, 1.0377E+04,
     + 1.0445E+04, 1.0515E+04, 1.0584E+04, 1.0654E+04, 1.0725E+04,
     +  1.0796E+04, 1.0867E+04, 1.0939E+04, 1.1011E+04, 1.1084E+04,
     + 1.1157E+04, 1.1231E+04, 1.1305E+04, 1.1380E+04, 1.1455E+04,
     +  1.1530E+04, 1.1606E+04, 1.1683E+04, 1.1760E+04, 1.1837E+04,
     + 1.1915E+04, 1.1994E+04, 1.2073E+04, 1.2152E+04, 1.2232E+04,
     +  1.2313E+04, 1.2394E+04, 1.2475E+04, 1.2557E+04, 1.2640E+04,
     + 1.2723E+04, 1.2806E+04, 1.2891E+04, 1.2975E+04, 1.3060E+04,
     +  1.3146E+04, 1.3232E+04, 1.3319E+04, 1.3407E+04, 1.3494E+04,
     + 1.3583E+04, 1.3672E+04, 1.3761E+04, 1.3852E+04, 1.3942E+04,
     +  1.4034E+04, 1.4125E+04, 1.4218E+04, 1.4311E+04, 1.4404E+04,
     + 1.4499E+04, 1.4593E+04, 1.4689E+04, 1.4785E+04, 1.4881E+04,
     +  1.4978E+04, 1.5076E+04, 1.5174E+04, 1.5273E+04, 1.5373E+04,
     + 1.5473E+04, 1.5574E+04, 1.5676E+04, 1.5778E+04, 1.5880E+04,
     +  1.5984E+04, 1.6088E+04, 1.6193E+04, 1.6298E+04, 1.6404E+04,
     + 1.6511E+04, 1.6618E+04, 1.6726E+04, 1.6835E+04, 1.6944E+04,
     +  1.7054E+04, 1.7165E+04, 1.7276E+04, 1.7388E+04, 1.7501E+04,
     + 1.7614E+04, 1.7728E+04, 1.7843E+04, 1.7959E+04, 1.8075E+04,
     +  1.8192E+04, 1.8310E+04, 1.8429E+04, 1.8548E+04, 1.8668E+04,
     + 1.8788E+04, 1.8910E+04, 1.9032E+04, 1.9155E+04, 1.9279E+04,
     +  1.9403E+04, 1.9528E+04, 1.9654E+04, 1.9781E+04, 1.9909E+04,
     + 2.0037E+04, 2.0166E+04, 2.0296E+04, 2.0427E+04, 2.0558E+04,
     +  2.0691E+04, 2.0824E+04, 2.0958E+04, 2.1093E+04, 2.1228E+04,
     + 2.1365E+04, 2.1502E+04, 2.1640E+04, 2.1779E+04, 2.1919E+04,
     +  2.2060E+04, 2.2201E+04, 2.2344E+04, 2.2487E+04, 2.2631E+04,
     + 2.2776E+04, 2.2922E+04, 2.3069E+04, 2.3217E+04, 2.3365E+04,
     +  2.3515E+04, 2.3665E+04, 2.3817E+04, 2.3969E+04, 2.4122E+04,
     + 2.4276E+04, 2.4431E+04, 2.4587E+04, 2.4744E+04, 2.4902E+04,
     +  2.5061E+04, 2.5221E+04, 2.5382E+04, 2.5543E+04, 2.5706E+04,
     + 2.5870E+04, 2.6034E+04, 2.6200E+04, 2.6367E+04, 2.6535E+04,
     +  2.6703E+04, 2.6873E+04, 2.7044E+04, 2.7216E+04, 2.7389E+04,
     + 2.7563E+04, 2.7738E+04, 2.7914E+04, 2.8091E+04, 2.8269E+04,
     +  2.8448E+04, 2.8628E+04, 2.8810E+04, 2.8992E+04, 2.9176E+04,
     + 2.9360E+04, 2.9546E+04, 2.9768E+04, 2.9956E+04, 3.0146E+04,
     +  3.0337E+04, 3.0529E+04, 3.0722E+04, 3.0916E+04, 3.1112E+04,
     + 3.1309E+04, 3.1507E+04, 3.1706E+04, 3.1906E+04, 3.2107E+04,
     +  3.2310E+04, 3.2514E+04, 3.2719E+04, 3.2925E+04, 3.3133E+04,
     + 3.3342E+04, 3.3552E+04, 3.3763E+04, 3.3976E+04, 3.4190E+04,
     +  3.4405E+04, 3.4621E+04, 3.4839E+04, 3.5058E+04, 3.5278E+04,
     + 3.5500E+04, 3.5723E+04, 3.5947E+04, 3.6173E+04, 3.6400E+04,
     +  3.6628E+04, 3.6858E+04, 3.7089E+04, 3.7321E+04, 3.7555E+04,
     + 3.7790E+04, 3.8027E+04, 3.8265E+04, 3.8504E+04, 3.8745E+04,
     +  3.8988E+04, 3.9231E+04, 3.9476E+04, 3.9723E+04, 3.9971E+04,
     + 4.0221E+04, 4.0472E+04, 4.0724E+04, 4.0978E+04, 4.1234E+04,
     +  4.1491E+04, 4.1749E+04, 4.2009E+04, 4.2271E+04, 4.2534E+04,
     + 4.2799E+04, 4.3065E+04, 4.3333E+04, 4.3602E+04, 4.3873E+04,
     +  4.4146E+04, 4.4420E+04, 4.4695E+04, 4.4973E+04, 4.5252E+04,
     + 4.5533E+04, 4.5815E+04, 4.6099E+04, 4.6384E+04, 4.6672E+04,
     +  4.6961E+04, 4.7251E+04, 4.7544E+04, 4.7838E+04, 4.8134E+04,
     + 4.8431E+04, 4.8731E+04, 4.9032E+04, 4.9334E+04, 4.9639E+04,
     +  4.9945E+04, 5.0253E+04, 5.0563E+04, 5.0875E+04, 5.1189E+04,
     + 5.1504E+04, 5.1821E+04, 5.2140E+04, 5.2461E+04, 5.2784E+04,
     +  5.3108E+04, 5.3435E+04, 5.3763E+04, 5.4094E+04, 5.4426E+04,
     + 5.4760E+04, 5.5096E+04, 5.5434E+04, 5.5774E+04, 5.6116E+04,
     +  5.6460E+04, 5.6806E+04, 5.7154E+04, 5.7504E+04, 5.7856E+04,
     + 5.8210E+04, 5.8566E+04, 5.8924E+04, 5.9284E+04, 5.9647E+04,
     +  6.0011E+04, 6.0377E+04, 6.0746E+04, 6.1117E+04, 6.1489E+04,
     + 6.1864E+04, 6.2241E+04, 6.2621E+04, 6.3002E+04, 6.3386E+04,
     +  6.3772E+04, 6.4160E+04, 6.4550E+04, 6.4942E+04, 6.5337E+04,
     + 6.5734E+04, 6.6133E+04, 6.6535E+04, 6.6939E+04, 6.7345E+04,
     +  6.7753E+04, 6.8164E+04, 6.8577E+04, 6.8993E+04, 6.9411E+04,
     + 6.9831E+04, 7.0254E+04, 7.0679E+04, 7.1106E+04, 7.1536E+04,
     +  7.1969E+04, 7.2403E+04, 7.2841E+04, 7.3280E+04, 7.3723E+04,
     + 7.4168E+04, 7.4615E+04, 7.5065E+04, 7.5517E+04, 7.5972E+04,
     +  7.6430E+04, 7.6890E+04, 7.7352E+04, 7.7818E+04, 7.8286E+04,
     + 7.8756E+04, 7.9230E+04, 7.9706E+04, 8.0184E+04, 8.0666E+04,
     +  8.1150E+04, 8.1636E+04, 8.2126E+04, 8.2618E+04, 8.3113E+04,
     + 8.3611E+04, 8.4112E+04, 8.4615E+04, 8.5122E+04, 8.5631E+04,
     +  8.6143E+04, 8.6658E+04, 8.7175E+04, 8.7696E+04, 8.8219E+04,
     + 8.8746E+04, 8.9275E+04, 8.9808E+04, 9.0343E+04, 9.0882E+04,
     +  9.1423E+04, 9.1967E+04, 9.2515E+04, 9.3065E+04, 9.3619E+04,
     + 9.4176E+04, 9.4735E+04, 9.5298E+04, 9.5864E+04, 9.6434E+04,
     +  9.7006E+04, 9.7581E+04, 9.8160E+04, 9.8742E+04, 9.9327E+04,
     + 9.9916E+04, 1.0051E+05, 1.0110E+05, 1.0170E+05, 1.0230E+05,
     +  1.0291E+05, 1.0352E+05, 1.0413E+05, 1.0474E+05, 1.0536E+05,
     + 1.0598E+05, 1.0661E+05, 1.0724E+05, 1.0787E+05, 1.0850E+05,
     +  1.0914E+05, 1.0979E+05, 1.1043E+05, 1.1108E+05, 1.1174E+05,
     + 1.1239E+05, 1.1305E+05, 1.1372E+05, 1.1439E+05, 1.1506E+05,
     +  1.1573E+05, 1.1641E+05, 1.1709E+05, 1.1778E+05, 1.1847E+05,
     + 1.1917E+05, 1.1986E+05, 1.2056E+05, 1.2127E+05, 1.2198E+05,
     +  1.2269E+05, 1.2341E+05, 1.2413E+05, 1.2486E+05, 1.2558E+05,
     + 1.2632E+05, 1.2705E+05, 1.2780E+05, 1.2854E+05, 1.2929E+05,
     +  1.3004E+05, 1.3080E+05, 1.3156E+05, 1.3233E+05, 1.3310E+05,
     + 1.3387E+05, 1.3465E+05, 1.3543E+05, 1.3622E+05, 1.3701E+05,
     +  1.3780E+05, 1.3860E+05, 1.3941E+05, 1.4021E+05, 1.4103E+05,
     + 1.4184E+05, 1.4266E+05, 1.4349E+05, 1.4432E+05, 1.4515E+05,
     +  1.4599E+05, 1.4684E+05, 1.4768E+05, 1.4854E+05, 1.4939E+05,
     + 1.5026E+05, 1.5112E+05, 1.5199E+05, 1.5287E+05, 1.5375E+05,
     +  1.5464E+05, 1.5553E+05, 1.5642E+05, 1.5732E+05, 1.5822E+05,
     + 1.5913E+05, 1.6005E+05, 1.6097E+05, 1.6189E+05, 1.6282E+05,
     +  1.6375E+05, 1.6469E+05, 1.6564E+05, 1.6659E+05, 1.6754E+05,
     + 1.6850E+05, 1.6946E+05, 1.7043E+05, 1.7141E+05, 1.7239E+05,
     +  1.7337E+05, 1.7436E+05, 1.7536E+05, 1.7636E+05, 1.7736E+05,
     + 1.7838E+05, 1.7939E+05, 1.8042E+05, 1.8144E+05, 1.8248E+05,
     +  1.8351E+05, 1.8456E+05, 1.8561E+05, 1.8666E+05, 1.8772E+05,
     + 1.8879E+05, 1.8986E+05, 1.9094E+05, 1.9202E+05, 1.9311E+05,
     +  1.9421E+05, 1.9531E+05, 1.9641E+05, 1.9753E+05, 1.9864E+05,
     + 1.9977E+05, 2.0090E+05, 2.0203E+05, 2.0318E+05, 2.0432E+05,
     +  2.0548E+05, 2.0664E+05, 2.0780E+05, 2.0897E+05, 2.1015E+05,
     + 2.1134E+05, 2.1253E+05, 2.1372E+05, 2.1493E+05, 2.1614E+05,
     +  2.1735E+05, 2.1857E+05, 2.1980E+05, 2.2104E+05, 2.2228E+05,
     + 2.2353E+05, 2.2478E+05, 2.2604E+05, 2.2731E+05, 2.2858E+05,
     +  2.2986E+05, 2.3115E+05, 2.3245E+05, 2.3375E+05, 2.3505E+05,
     + 2.3637E+05, 2.3769E+05, 2.3902E+05, 2.4035E+05, 2.4169E+05,
     +  2.4304E+05, 2.4440E+05, 2.4576E+05, 2.4713E+05, 2.4851E+05,
     + 2.4989E+05, 2.5128E+05, 2.5268E+05, 2.5409E+05, 2.5550E+05,
     +  2.5692E+05, 2.5835E+05, 2.5978E+05, 2.6122E+05, 2.6267E+05,
     + 2.6413E+05, 2.6559E+05, 2.6707E+05, 2.6855E+05, 2.7003E+05,
     +  2.7153E+05, 2.7303E+05, 2.7454E+05, 2.7606E+05, 2.7758E+05,
     + 2.7912E+05, 2.8066E+05, 2.8221E+05, 2.8377E+05, 2.8533E+05,
     +  2.8691E+05, 2.8849E+05, 2.9008E+05, 2.9167E+05, 2.9328E+05,
     + 2.9489E+05, 2.9651E+05, 2.9815E+05, 2.9978E+05, 3.0143E+05,
     +  3.0309E+05, 3.0475E+05, 3.0642E+05, 3.0810E+05, 3.0979E+05,
     + 3.1149E+05, 3.1320E+05, 3.1491E+05, 3.1664E+05, 3.1837E+05,
     +  3.2011E+05, 3.2186E+05, 3.2362E+05, 3.2539E+05, 3.2716E+05,
     + 3.2895E+05, 3.3074E+05, 3.3255E+05, 3.3436E+05, 3.3618E+05,
     +  3.3801E+05, 3.3986E+05, 3.4171E+05, 3.4356E+05, 3.4543E+05,
     + 3.4731E+05, 3.4920E+05, 3.5110E+05, 3.5300E+05, 3.5492E+05,
     +  3.5684E+05, 3.5878E+05, 3.6072E+05, 3.6268E+05, 3.6464E+05,
     + 3.6662E+05, 3.6860E+05, 3.7060E+05, 3.7260E+05, 3.7461E+05,
     +  3.7664E+05, 3.7867E+05, 3.8072E+05, 3.8277E+05, 3.8484E+05,
     + 3.8691E+05, 3.8900E+05, 3.9110E+05, 3.9320E+05, 3.9532E+05,
     +  3.9745E+05, 3.9958E+05, 4.0173E+05, 4.0389E+05, 4.0606E+05,
     + 4.0824E+05, 4.1044E+05, 4.1264E+05, 4.1485E+05, 4.1708E+05,
     +  4.1931E+05, 4.2156E+05, 4.2382E+05, 4.2609E+05, 4.2837E+05,
     + 4.3066E+05, 4.3296E+05, 4.3527E+05, 4.3760E+05, 4.3994E+05,
     +  4.4228E+05, 4.4464E+05, 4.4702E+05, 4.4940E+05, 4.5179E+05,
     + 4.5420E+05, 4.5662E+05, 4.5905E+05, 4.6149E+05, 4.6395E+05,
     +  4.6641E+05, 4.6889E+05, 4.7138E+05, 4.7389E+05, 4.7640E+05,
     + 4.7893E+05, 4.8147E+05, 4.8402E+05, 4.8659E+05, 4.8917E+05,
     +  4.9176E+05, 4.9436E+05, 4.9697E+05, 4.9960E+05, 5.0224E+05,
     + 5.0490E+05, 5.0756E+05, 5.1024E+05, 5.1294E+05, 5.1564E+05,
     +  5.1836E+05, 5.2109E+05, 5.2384E+05, 5.2660E+05, 5.2937E+05,
     + 5.3216E+05, 5.3496E+05, 5.3777E+05, 5.4060E+05, 5.4344E+05,
     +  5.4629E+05, 5.4916E+05, 5.5204E+05, 5.5493E+05, 5.5784E+05,
     + 5.6077E+05, 5.6371E+05, 5.6666E+05, 5.6962E+05, 5.7260E+05,
     +  5.7560E+05, 5.7861E+05, 5.8163E+05, 5.8467E+05, 5.8772E+05,
     + 5.9079E+05, 5.9387E+05, 5.9697E+05, 6.0008E+05, 6.0321E+05,
     +  6.0635E+05, 6.0951E+05, 6.1268E+05, 6.1587E+05, 6.1907E+05,
     + 6.2229E+05, 6.2552E+05, 6.2877E+05, 6.3204E+05, 6.3532E+05,
     +  6.3861E+05, 6.4192E+05, 6.4525E+05, 6.4859E+05, 6.5195E+05,
     + 6.5533E+05, 6.5872E+05, 6.6213E+05, 6.6555E+05, 6.6899E+05,
     +  6.7245E+05, 6.7592E+05, 6.7941E+05, 6.8291E+05, 6.8644E+05,
     + 6.8998E+05, 6.9353E+05, 6.9710E+05, 7.0069E+05, 7.0430E+05,
     +  7.0792E+05, 7.1157E+05, 7.1522E+05, 7.1890E+05, 7.2259E+05,
     + 7.2630E+05, 7.3003E+05, 7.3377E+05, 7.3754E+05, 7.4132E+05,
     +  7.4512E+05, 7.4893E+05, 7.5277E+05, 7.5662E+05, 7.6049E+05,
     + 7.6438E+05, 7.6829E+05, 7.7221E+05, 7.7616E+05, 7.8012E+05,
     +  7.8410E+05, 7.8810E+05, 7.9212E+05, 7.9615E+05, 8.0021E+05,
     + 8.0428E+05, 8.0838E+05, 8.1249E+05, 8.1662E+05, 8.2078E+05,
     +  8.2495E+05, 8.2914E+05, 8.3335E+05, 8.3758E+05, 8.4183E+05,
     + 8.4610E+05, 8.5039E+05, 8.5469E+05, 8.5902E+05, 8.6337E+05,
     +  8.6774E+05, 8.7213E+05, 8.7654E+05, 8.8097E+05, 8.8543E+05,
     + 8.8990E+05, 8.9439E+05, 8.9890E+05, 9.0344E+05, 9.0799E+05,
     +  9.1257E+05, 9.1717E+05, 9.2179E+05, 9.2643E+05, 9.3109E+05,
     + 9.3577E+05, 9.4048E+05, 9.4520E+05, 9.4995E+05, 9.5472E+05,
     +  9.5951E+05, 9.6433E+05, 9.6916E+05, 9.7402E+05, 9.7890E+05,
     + 9.8381E+05, 9.8873E+05, 9.9368E+05, 9.9865E+05, 1.0036E+06,
     +  1.0087E+06, 1.0137E+06, 1.0188E+06, 1.0239E+06, 1.0290E+06,
     + 1.0341E+06, 1.0393E+06, 1.0444E+06, 1.0496E+06, 1.0549E+06,
     +  1.0601E+06, 1.0654E+06, 1.0707E+06, 1.0760E+06, 1.0814E+06,
     + 1.0867E+06, 1.0921E+06, 1.0976E+06, 1.1030E+06, 1.1085E+06,
     +  1.1140E+06, 1.1195E+06, 1.1250E+06, 1.1306E+06, 1.1362E+06,
     + 1.1418E+06, 1.1475E+06, 1.1531E+06, 1.1588E+06, 1.1646E+06,
     +  1.1703E+06, 1.1761E+06, 1.1819E+06, 1.1877E+06, 1.1936E+06,
     + 1.1994E+06, 1.2054E+06, 1.2113E+06, 1.2172E+06, 1.2232E+06,
     +  1.2293E+06, 1.2353E+06, 1.2414E+06, 1.2475E+06, 1.2536E+06,
     + 1.2597E+06, 1.2659E+06, 1.2721E+06, 1.2784E+06, 1.2846E+06,
     +  1.2909E+06, 1.2972E+06, 1.3036E+06, 1.3099E+06, 1.3163E+06,
     + 1.3228E+06, 1.3292E+06, 1.3357E+06, 1.3423E+06, 1.3488E+06,
     +  1.3554E+06, 1.3620E+06, 1.3686E+06, 1.3753E+06, 1.3820E+06,
     + 1.3887E+06, 1.3955E+06, 1.4022E+06, 1.4091E+06, 1.4159E+06,
     +  1.4228E+06, 1.4297E+06, 1.4366E+06, 1.4436E+06, 1.4506E+06,
     + 1.4576E+06, 1.4647E+06, 1.4718E+06, 1.4789E+06, 1.4861E+06,
     +  1.4932E+06, 1.5005E+06, 1.5077E+06, 1.5150E+06, 1.5223E+06,
     + 1.5297E+06, 1.5370E+06, 1.5445E+06, 1.5519E+06, 1.5594E+06,
     +  1.5669E+06, 1.5744E+06, 1.5820E+06, 1.5896E+06, 1.5973E+06,
     + 1.6049E+06, 1.6126E+06, 1.6204E+06, 1.6282E+06, 1.6360E+06,
     +  1.6438E+06, 1.6517E+06, 1.6596E+06, 1.6676E+06, 1.6756E+06,
     + 1.6836E+06, 1.6916E+06, 1.6997E+06, 1.7079E+06, 1.7160E+06,
     +  1.7242E+06, 1.7324E+06, 1.7407E+06, 1.7490E+06, 1.7574E+06,
     + 1.7657E+06, 1.7742E+06, 1.7826E+06, 1.7911E+06, 1.7996E+06,
     +  1.8082E+06, 1.8168E+06, 1.8254E+06, 1.8341E+06, 1.8428E+06,
     + 1.8515E+06, 1.8603E+06, 1.8691E+06, 1.8780E+06, 1.8869E+06,
     +  1.8959E+06, 1.9048E+06, 1.9138E+06, 1.9229E+06, 1.9320E+06,
     + 1.9411E+06, 1.9503E+06, 1.9595E+06, 1.9688E+06, 1.9781E+06,
     +  1.9874E+06, 1.9968E+06, 2.0062E+06, 2.0156E+06, 2.0251E+06,
     + 2.0347E+06, 2.0442E+06, 2.0539E+06, 2.0635E+06, 2.0732E+06,
     +  2.0830E+06, 2.0927E+06, 2.1026E+06, 2.1124E+06, 2.1223E+06,
     + 2.1323E+06, 2.1423E+06, 2.1523E+06, 2.1624E+06, 2.1725E+06,
     +  2.1827E+06, 2.1929E+06, 2.2031E+06, 2.2134E+06, 2.2238E+06,
     + 2.2342E+06, 2.2446E+06, 2.2551E+06, 2.2656E+06, 2.2761E+06,
     +  2.2868E+06, 2.2974E+06, 2.3081E+06, 2.3188E+06, 2.3296E+06,
     + 2.3405E+06, 2.3513E+06, 2.3623E+06, 2.3732E+06, 2.3843E+06,
     +  2.3953E+06, 2.4064E+06, 2.4176E+06, 2.4288E+06, 2.4401E+06,
     + 2.4514E+06, 2.4627E+06, 2.4741E+06, 2.4855E+06, 2.4970E+06,
     +  2.5086E+06, 2.5202E+06, 2.5318E+06, 2.5435E+06, 2.5552E+06,
     + 2.5670E+06, 2.5788E+06, 2.5907E+06, 2.6027E+06, 2.6147E+06,
     +  2.6267E+06, 2.6388E+06, 2.6509E+06, 2.6631E+06, 2.6753E+06,
     + 2.6876E+06, 2.7000E+06, 2.7124E+06, 2.7248E+06, 2.7373E+06,
     +  2.7498E+06, 2.7624E+06, 2.7751E+06, 2.7878E+06, 2.8006E+06,
     + 2.8134E+06, 2.8262E+06, 2.8391E+06, 2.8521E+06, 2.8651E+06,
     +  2.8782E+06, 2.8914E+06, 2.9045E+06, 2.9178E+06, 2.9311E+06,
     + 2.9444E+06, 2.9578E+06, 2.9713E+06, 2.9848E+06, 2.9984E+06,
     +  3.0120E+06, 3.0257E+06, 3.0395E+06, 3.0533E+06, 3.0671E+06,
     + 3.0810E+06, 3.0950E+06, 3.1090E+06, 3.1231E+06, 3.1373E+06,
     +  3.1515E+06, 3.1658E+06, 3.1801E+06, 3.1945E+06, 3.2089E+06,
     + 3.2234E+06, 3.2379E+06, 3.2526E+06, 3.2672E+06, 3.2820E+06,
     +  3.2968E+06, 3.3116E+06, 3.3266E+06, 3.3415E+06, 3.3566E+06,
     + 3.3717E+06, 3.3869E+06, 3.4021E+06, 3.4174E+06, 3.4327E+06,
     +  3.4481E+06, 3.4636E+06, 3.4792E+06, 3.4948E+06, 3.5104E+06,
     + 3.5262E+06, 3.5420E+06, 3.5578E+06, 3.5737E+06, 3.5897E+06,
     +  3.6058E+06, 3.6219E+06, 3.6381E+06, 3.6543E+06, 3.6707E+06,
     + 3.6870E+06, 3.7035E+06, 3.7200E+06, 3.7366E+06, 3.7532E+06,
     +  3.7700E+06, 3.7867E+06, 3.8036E+06, 3.8205E+06, 3.8375E+06,
     + 3.8546E+06, 3.8717E+06, 3.8889E+06, 3.9061E+06, 3.9235E+06,
     +  3.9409E+06, 3.9584E+06, 3.9759E+06, 3.9935E+06, 4.0112E+06,
     + 4.0290E+06, 4.0468E+06, 4.0647E+06, 4.0827E+06, 4.1007E+06,
     +  4.1188E+06, 4.1370E+06, 4.1553E+06, 4.1736E+06, 4.1920E+06,
     + 4.2105E+06, 4.2290E+06, 4.2477E+06, 4.2664E+06, 4.2852E+06,
     +  4.3040E+06, 4.3229E+06, 4.3419E+06, 4.3610E+06, 4.3802E+06,
     + 4.3994E+06, 4.4187E+06, 4.4381E+06, 4.4576E+06, 4.4771E+06,
     +  4.4967E+06, 4.5164E+06, 4.5362E+06, 4.5560E+06, 4.5760E+06,
     + 4.5960E+06, 4.6161E+06, 4.6362E+06, 4.6565E+06, 4.6768E+06,
     +  4.6972E+06, 4.7177E+06, 4.7383E+06, 4.7589E+06, 4.7797E+06,
     + 4.8005E+06, 4.8214E+06, 4.8424E+06, 4.8634E+06, 4.8846E+06,
     +  4.9058E+06, 4.9271E+06, 4.9485E+06, 4.9700E+06, 4.9915E+06,
     + 5.0132E+06, 5.0349E+06, 5.0567E+06, 5.0786E+06, 5.1006E+06,
     +  5.1227E+06, 5.1449E+06, 5.1671E+06, 5.1895E+06, 5.2119E+06,
     + 5.2344E+06, 5.2570E+06, 5.2797E+06, 5.3025E+06, 5.3253E+06,
     +  5.3483E+06, 5.3713E+06, 5.3945E+06, 5.4177E+06, 5.4410E+06,
     + 5.4644E+06, 5.4879E+06, 5.5115E+06, 5.5352E+06, 5.5590E+06,
     +  5.5828E+06, 5.6068E+06, 5.6308E+06, 5.6550E+06, 5.6792E+06,
     + 5.7036E+06, 5.7280E+06, 5.7525E+06, 5.7771E+06, 5.8018E+06,
     +  5.8266E+06, 5.8516E+06, 5.8766E+06, 5.9016E+06, 5.9268E+06,
     + 5.9521E+06, 5.9775E+06, 6.0030E+06, 6.0286E+06, 6.0543E+06,
     +  6.0801E+06, 6.1059E+06, 6.1319E+06, 6.1580E+06, 6.1842E+06,
     + 6.2105E+06, 6.2368E+06, 6.2633E+06, 6.2899E+06, 6.3166E+06,
     +  6.3434E+06, 6.3703E+06, 6.3973E+06, 6.4244E+06, 6.4516E+06,
     + 6.4789E+06, 6.5063E+06, 6.5338E+06, 6.5614E+06, 6.5892E+06,
     +  6.6170E+06, 6.6449E+06, 6.6730E+06, 6.7011E+06, 6.7294E+06,
     + 6.7578E+06, 6.7862E+06, 6.8148E+06, 6.8435E+06, 6.8723E+06,
     +  6.9012E+06, 6.9302E+06, 6.9594E+06, 6.9886E+06, 7.0180E+06,
     + 7.0474E+06, 7.0770E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       311
      data (QofT( 2,J),J=1,2408)/ 1.0000E+01, 1.0010E+01, 1.0118E+01,
     +  1.0415E+01, 1.0888E+01, 1.1484E+01, 1.2166E+01, 1.2909E+01,
     + 1.3704E+01, 1.4544E+01, 1.5429E+01, 1.6358E+01, 1.7331E+01,
     +  1.8350E+01, 1.9416E+01, 2.0530E+01, 2.1692E+01, 2.2903E+01,
     + 2.4164E+01, 2.5474E+01, 2.8242E+01, 3.1204E+01, 3.4354E+01,
     +  3.7684E+01, 4.1186E+01, 4.4851E+01, 4.8673E+01, 5.2641E+01,
     + 5.6749E+01, 6.0990E+01, 6.5357E+01, 6.9844E+01, 7.4447E+01,
     +  7.9161E+01, 8.3982E+01, 8.8906E+01, 9.3930E+01, 9.9051E+01,
     + 1.0427E+02, 1.0958E+02, 1.1497E+02, 1.2046E+02, 1.2603E+02,
     +  1.3169E+02, 1.3744E+02, 1.4326E+02, 1.4917E+02, 1.5515E+02,
     + 1.6121E+02, 1.6736E+02, 1.7357E+02, 1.7987E+02, 1.8623E+02,
     +  1.9268E+02, 1.9919E+02, 2.0578E+02, 2.1244E+02, 2.1916E+02,
     + 2.2596E+02, 2.3283E+02, 2.3977E+02, 2.4677E+02, 2.5384E+02,
     +  2.6098E+02, 2.6818E+02, 2.7545E+02, 2.8278E+02, 2.9018E+02,
     + 2.9764E+02, 3.0517E+02, 3.1275E+02, 3.2040E+02, 3.2811E+02,
     +  3.3589E+02, 3.4372E+02, 3.5161E+02, 3.5956E+02, 3.6757E+02,
     + 3.7564E+02, 3.8377E+02, 3.9196E+02, 4.0021E+02, 4.0851E+02,
     +  4.1687E+02, 4.2528E+02, 4.3376E+02, 4.4229E+02, 4.5087E+02,
     + 4.5951E+02, 4.6820E+02, 4.7695E+02, 4.8575E+02, 4.9461E+02,
     +  5.0352E+02, 5.1249E+02, 5.2151E+02, 5.3058E+02, 5.3970E+02,
     + 5.4887E+02, 5.5810E+02, 5.6739E+02, 5.7672E+02, 5.8610E+02,
     +  5.9554E+02, 6.0502E+02, 6.1456E+02, 6.2415E+02, 6.3379E+02,
     + 6.4349E+02, 6.5323E+02, 6.6302E+02, 6.7287E+02, 6.8276E+02,
     +  6.9271E+02, 7.0270E+02, 7.1274E+02, 7.2284E+02, 7.3299E+02,
     + 7.4318E+02, 7.5343E+02, 7.6372E+02, 7.7407E+02, 7.8446E+02,
     +  7.9491E+02, 8.0540E+02, 8.1595E+02, 8.2655E+02, 8.3719E+02,
     + 8.4790E+02, 8.5864E+02, 8.6944E+02, 8.8029E+02, 8.9119E+02,
     +  9.0214E+02, 9.1315E+02, 9.2420E+02, 9.3530E+02, 9.4646E+02,
     + 9.5767E+02, 9.6893E+02, 9.8024E+02, 9.9161E+02, 1.0030E+03,
     +  1.0145E+03, 1.0260E+03, 1.0376E+03, 1.0492E+03, 1.0609E+03,
     + 1.0727E+03, 1.0845E+03, 1.0963E+03, 1.1082E+03, 1.1202E+03,
     +  1.1322E+03, 1.1443E+03, 1.1564E+03, 1.1686E+03, 1.1808E+03,
     + 1.1931E+03, 1.2055E+03, 1.2179E+03, 1.2304E+03, 1.2429E+03,
     +  1.2555E+03, 1.2682E+03, 1.2809E+03, 1.2937E+03, 1.3065E+03,
     + 1.3194E+03, 1.3324E+03, 1.3454E+03, 1.3585E+03, 1.3716E+03,
     +  1.3848E+03, 1.3981E+03, 1.4115E+03, 1.4249E+03, 1.4384E+03,
     + 1.4519E+03, 1.4655E+03, 1.4792E+03, 1.4930E+03, 1.5068E+03,
     +  1.5207E+03, 1.5346E+03, 1.5487E+03, 1.5628E+03, 1.5769E+03,
     + 1.5912E+03, 1.6055E+03, 1.6199E+03, 1.6344E+03, 1.6489E+03,
     +  1.6635E+03, 1.6782E+03, 1.6930E+03, 1.7078E+03, 1.7228E+03,
     + 1.7378E+03, 1.7529E+03, 1.7680E+03, 1.7833E+03, 1.7986E+03,
     +  1.8140E+03, 1.8295E+03, 1.8451E+03, 1.8607E+03, 1.8765E+03,
     + 1.8923E+03, 1.9082E+03, 1.9242E+03, 1.9403E+03, 1.9565E+03,
     +  1.9728E+03, 1.9891E+03, 2.0056E+03, 2.0221E+03, 2.0388E+03,
     + 2.0555E+03, 2.0723E+03, 2.0893E+03, 2.1063E+03, 2.1234E+03,
     +  2.1406E+03, 2.1579E+03, 2.1753E+03, 2.1928E+03, 2.2104E+03,
     + 2.2281E+03, 2.2459E+03, 2.2638E+03, 2.2818E+03, 2.2999E+03,
     +  2.3181E+03, 2.3365E+03, 2.3549E+03, 2.3734E+03, 2.3921E+03,
     + 2.4109E+03, 2.4297E+03, 2.4487E+03, 2.4678E+03, 2.4870E+03,
     +  2.5063E+03, 2.5257E+03, 2.5453E+03, 2.5650E+03, 2.5848E+03,
     + 2.6046E+03, 2.6247E+03, 2.6448E+03, 2.6651E+03, 2.6854E+03,
     +  2.7059E+03, 2.7266E+03, 2.7473E+03, 2.7682E+03, 2.7892E+03,
     + 2.8103E+03, 2.8316E+03, 2.8530E+03, 2.8745E+03, 2.8962E+03,
     +  2.9180E+03, 2.9399E+03, 2.9619E+03, 2.9841E+03, 3.0064E+03,
     + 3.0289E+03, 3.0515E+03, 3.0742E+03, 3.0971E+03, 3.1201E+03,
     +  3.1433E+03, 3.1666E+03, 3.1901E+03, 3.2136E+03, 3.2374E+03,
     + 3.2613E+03, 3.2853E+03, 3.3095E+03, 3.3339E+03, 3.3583E+03,
     +  3.3830E+03, 3.4078E+03, 3.4328E+03, 3.4579E+03, 3.4831E+03,
     + 3.5086E+03, 3.5342E+03, 3.5599E+03, 3.5858E+03, 3.6119E+03,
     +  3.6382E+03, 3.6646E+03, 3.6911E+03, 3.7179E+03, 3.7448E+03,
     + 3.7719E+03, 3.7991E+03, 3.8266E+03, 3.8541E+03, 3.8819E+03,
     +  3.9099E+03, 3.9380E+03, 3.9663E+03, 3.9948E+03, 4.0235E+03,
     + 4.0524E+03, 4.0814E+03, 4.1106E+03, 4.1400E+03, 4.1696E+03,
     +  4.1994E+03, 4.2294E+03, 4.2596E+03, 4.2900E+03, 4.3205E+03,
     + 4.3513E+03, 4.3822E+03, 4.4134E+03, 4.4447E+03, 4.4763E+03,
     +  4.5080E+03, 4.5400E+03, 4.5722E+03, 4.6046E+03, 4.6371E+03,
     + 4.6700E+03, 4.7030E+03, 4.7362E+03, 4.7696E+03, 4.8032E+03,
     +  4.8371E+03, 4.8712E+03, 4.9055E+03, 4.9400E+03, 4.9748E+03,
     + 5.0098E+03, 5.0449E+03, 5.0804E+03, 5.1160E+03, 5.1519E+03,
     +  5.1880E+03, 5.2244E+03, 5.2610E+03, 5.2978E+03, 5.3349E+03,
     + 5.3721E+03, 5.4097E+03, 5.4475E+03, 5.4855E+03, 5.5238E+03,
     +  5.5623E+03, 5.6011E+03, 5.6401E+03, 5.6794E+03, 5.7189E+03,
     + 5.7587E+03, 5.7988E+03, 5.8391E+03, 5.8797E+03, 5.9205E+03,
     +  5.9616E+03, 6.0030E+03, 6.0446E+03, 6.0866E+03, 6.1287E+03,
     + 6.1712E+03, 6.2139E+03, 6.2569E+03, 6.3002E+03, 6.3438E+03,
     +  6.3876E+03, 6.4318E+03, 6.4762E+03, 6.5209E+03, 6.5659E+03,
     + 6.6112E+03, 6.6568E+03, 6.7027E+03, 6.7489E+03, 6.7954E+03,
     +  6.8421E+03, 6.8892E+03, 6.9366E+03, 6.9843E+03, 7.0323E+03,
     + 7.0807E+03, 7.1293E+03, 7.1783E+03, 7.2275E+03, 7.2771E+03,
     +  7.3270E+03, 7.3773E+03, 7.4279E+03, 7.4788E+03, 7.5300E+03,
     + 7.5815E+03, 7.6334E+03, 7.6857E+03, 7.7382E+03, 7.7911E+03,
     +  7.8444E+03, 7.8980E+03, 7.9520E+03, 8.0062E+03, 8.0609E+03,
     + 8.1159E+03, 8.1713E+03, 8.2270E+03, 8.2831E+03, 8.3396E+03,
     +  8.3963E+03, 8.4535E+03, 8.5111E+03, 8.5690E+03, 8.6273E+03,
     + 8.6860E+03, 8.7451E+03, 8.8045E+03, 8.8644E+03, 8.9246E+03,
     +  8.9852E+03, 9.0462E+03, 9.1076E+03, 9.1694E+03, 9.2316E+03,
     + 9.2941E+03, 9.3572E+03, 9.4206E+03, 9.4844E+03, 9.5486E+03,
     +  9.6133E+03, 9.6783E+03, 9.7438E+03, 9.8098E+03, 9.8761E+03,
     + 9.9428E+03, 1.0010E+04, 1.0078E+04, 1.0146E+04, 1.0214E+04,
     +  1.0283E+04, 1.0353E+04, 1.0422E+04, 1.0493E+04, 1.0564E+04,
     + 1.0635E+04, 1.0706E+04, 1.0778E+04, 1.0851E+04, 1.0924E+04,
     +  1.0998E+04, 1.1072E+04, 1.1146E+04, 1.1221E+04, 1.1297E+04,
     + 1.1373E+04, 1.1449E+04, 1.1526E+04, 1.1603E+04, 1.1681E+04,
     +  1.1760E+04, 1.1838E+04, 1.1918E+04, 1.1998E+04, 1.2078E+04,
     + 1.2159E+04, 1.2241E+04, 1.2323E+04, 1.2405E+04, 1.2488E+04,
     +  1.2572E+04, 1.2656E+04, 1.2741E+04, 1.2826E+04, 1.2912E+04,
     + 1.2998E+04, 1.3085E+04, 1.3172E+04, 1.3260E+04, 1.3349E+04,
     +  1.3438E+04, 1.3527E+04, 1.3618E+04, 1.3708E+04, 1.3800E+04,
     + 1.3892E+04, 1.3984E+04, 1.4078E+04, 1.4171E+04, 1.4266E+04,
     +  1.4361E+04, 1.4456E+04, 1.4552E+04, 1.4649E+04, 1.4747E+04,
     + 1.4845E+04, 1.4943E+04, 1.5043E+04, 1.5142E+04, 1.5243E+04,
     +  1.5344E+04, 1.5446E+04, 1.5549E+04, 1.5652E+04, 1.5755E+04,
     + 1.5860E+04, 1.5965E+04, 1.6071E+04, 1.6177E+04, 1.6284E+04,
     +  1.6392E+04, 1.6501E+04, 1.6610E+04, 1.6720E+04, 1.6830E+04,
     + 1.6942E+04, 1.7054E+04, 1.7166E+04, 1.7280E+04, 1.7394E+04,
     +  1.7508E+04, 1.7624E+04, 1.7740E+04, 1.7857E+04, 1.7975E+04,
     + 1.8094E+04, 1.8213E+04, 1.8333E+04, 1.8454E+04, 1.8575E+04,
     +  1.8697E+04, 1.8821E+04, 1.8944E+04, 1.9069E+04, 1.9194E+04,
     + 1.9321E+04, 1.9448E+04, 1.9575E+04, 1.9704E+04, 1.9833E+04,
     +  1.9964E+04, 2.0095E+04, 2.0226E+04, 2.0359E+04, 2.0493E+04,
     + 2.0627E+04, 2.0762E+04, 2.0898E+04, 2.1035E+04, 2.1173E+04,
     +  2.1311E+04, 2.1451E+04, 2.1591E+04, 2.1733E+04, 2.1875E+04,
     + 2.2018E+04, 2.2162E+04, 2.2306E+04, 2.2452E+04, 2.2599E+04,
     +  2.2746E+04, 2.2895E+04, 2.3044E+04, 2.3194E+04, 2.3346E+04,
     + 2.3498E+04, 2.3651E+04, 2.3805E+04, 2.3960E+04, 2.4116E+04,
     +  2.4273E+04, 2.4431E+04, 2.4590E+04, 2.4750E+04, 2.4911E+04,
     + 2.5073E+04, 2.5235E+04, 2.5399E+04, 2.5564E+04, 2.5730E+04,
     +  2.5897E+04, 2.6065E+04, 2.6234E+04, 2.6405E+04, 2.6576E+04,
     + 2.6748E+04, 2.6921E+04, 2.7096E+04, 2.7271E+04, 2.7448E+04,
     +  2.7625E+04, 2.7804E+04, 2.7984E+04, 2.8165E+04, 2.8347E+04,
     + 2.8530E+04, 2.8714E+04, 2.8900E+04, 2.9086E+04, 2.9274E+04,
     +  2.9463E+04, 2.9653E+04, 2.9844E+04, 3.0036E+04, 3.0230E+04,
     + 3.0425E+04, 3.0621E+04, 3.0818E+04, 3.1016E+04, 3.1216E+04,
     +  3.1416E+04, 3.1618E+04, 3.1822E+04, 3.2026E+04, 3.2232E+04,
     + 3.2439E+04, 3.2647E+04, 3.2857E+04, 3.3068E+04, 3.3280E+04,
     +  3.3493E+04, 3.3708E+04, 3.3924E+04, 3.4141E+04, 3.4360E+04,
     + 3.4580E+04, 3.4801E+04, 3.5024E+04, 3.5248E+04, 3.5473E+04,
     +  3.5700E+04, 3.5928E+04, 3.6158E+04, 3.6389E+04, 3.6621E+04,
     + 3.6855E+04, 3.7090E+04, 3.7327E+04, 3.7565E+04, 3.7804E+04,
     +  3.8045E+04, 3.8287E+04, 3.8531E+04, 3.8776E+04, 3.9023E+04,
     + 3.9272E+04, 3.9521E+04, 3.9773E+04, 4.0025E+04, 4.0280E+04,
     +  4.0536E+04, 4.0793E+04, 4.1052E+04, 4.1313E+04, 4.1575E+04,
     + 4.1838E+04, 4.2104E+04, 4.2370E+04, 4.2639E+04, 4.2909E+04,
     +  4.3181E+04, 4.3454E+04, 4.3729E+04, 4.4006E+04, 4.4284E+04,
     + 4.4564E+04, 4.4845E+04, 4.5129E+04, 4.5414E+04, 4.5700E+04,
     +  4.5989E+04, 4.6279E+04, 4.6571E+04, 4.6865E+04, 4.7160E+04,
     + 4.7457E+04, 4.7756E+04, 4.8057E+04, 4.8359E+04, 4.8664E+04,
     +  4.8970E+04, 4.9278E+04, 4.9588E+04, 4.9899E+04, 5.0213E+04,
     + 5.0528E+04, 5.0845E+04, 5.1164E+04, 5.1485E+04, 5.1808E+04,
     +  5.2133E+04, 5.2460E+04, 5.2788E+04, 5.3119E+04, 5.3452E+04,
     + 5.3786E+04, 5.4123E+04, 5.4461E+04, 5.4802E+04, 5.5144E+04,
     +  5.5489E+04, 5.5836E+04, 5.6184E+04, 5.6535E+04, 5.6888E+04,
     + 5.7243E+04, 5.7600E+04, 5.7959E+04, 5.8320E+04, 5.8683E+04,
     +  5.9049E+04, 5.9416E+04, 5.9786E+04, 6.0158E+04, 6.0532E+04,
     + 6.0908E+04, 6.1287E+04, 6.1668E+04, 6.2050E+04, 6.2436E+04,
     +  6.2823E+04, 6.3213E+04, 6.3605E+04, 6.3999E+04, 6.4396E+04,
     + 6.4795E+04, 6.5196E+04, 6.5600E+04, 6.6006E+04, 6.6414E+04,
     +  6.6825E+04, 6.7238E+04, 6.7654E+04, 6.8071E+04, 6.8492E+04,
     + 6.8915E+04, 6.9340E+04, 6.9768E+04, 7.0198E+04, 7.0631E+04,
     +  7.1066E+04, 7.1504E+04, 7.1944E+04, 7.2387E+04, 7.2833E+04,
     + 7.3281E+04, 7.3731E+04, 7.4185E+04, 7.4640E+04, 7.5099E+04,
     +  7.5560E+04, 7.6024E+04, 7.6491E+04, 7.6960E+04, 7.7432E+04,
     + 7.7906E+04, 7.8384E+04, 7.8864E+04, 7.9347E+04, 7.9833E+04,
     +  8.0321E+04, 8.0813E+04, 8.1307E+04, 8.1804E+04, 8.2304E+04,
     + 8.2806E+04, 8.3312E+04, 8.3821E+04, 8.4332E+04, 8.4846E+04,
     +  8.5364E+04, 8.5884E+04, 8.6407E+04, 8.6933E+04, 8.7463E+04,
     + 8.7995E+04, 8.8530E+04, 8.9069E+04, 8.9610E+04, 9.0155E+04,
     +  9.0703E+04, 9.1254E+04, 9.1808E+04, 9.2365E+04, 9.2925E+04,
     + 9.3489E+04, 9.4055E+04, 9.4625E+04, 9.5198E+04, 9.5775E+04,
     +  9.6355E+04, 9.6938E+04, 9.7524E+04, 9.8114E+04, 9.8707E+04,
     + 9.9303E+04, 9.9903E+04, 1.0051E+05, 1.0111E+05, 1.0172E+05,
     +  1.0234E+05, 1.0295E+05, 1.0357E+05, 1.0420E+05, 1.0482E+05,
     + 1.0546E+05, 1.0609E+05, 1.0673E+05, 1.0737E+05, 1.0802E+05,
     +  1.0866E+05, 1.0932E+05, 1.0997E+05, 1.1063E+05, 1.1130E+05,
     + 1.1197E+05, 1.1264E+05, 1.1331E+05, 1.1399E+05, 1.1467E+05,
     +  1.1536E+05, 1.1605E+05, 1.1674E+05, 1.1744E+05, 1.1814E+05,
     + 1.1885E+05, 1.1956E+05, 1.2027E+05, 1.2099E+05, 1.2171E+05,
     +  1.2244E+05, 1.2317E+05, 1.2390E+05, 1.2464E+05, 1.2538E+05,
     + 1.2613E+05, 1.2688E+05, 1.2764E+05, 1.2839E+05, 1.2916E+05,
     +  1.2992E+05, 1.3070E+05, 1.3147E+05, 1.3225E+05, 1.3304E+05,
     + 1.3383E+05, 1.3462E+05, 1.3542E+05, 1.3622E+05, 1.3703E+05,
     +  1.3784E+05, 1.3865E+05, 1.3947E+05, 1.4030E+05, 1.4113E+05,
     + 1.4196E+05, 1.4280E+05, 1.4364E+05, 1.4449E+05, 1.4534E+05,
     +  1.4620E+05, 1.4706E+05, 1.4793E+05, 1.4880E+05, 1.4967E+05,
     + 1.5055E+05, 1.5144E+05, 1.5233E+05, 1.5323E+05, 1.5413E+05,
     +  1.5503E+05, 1.5594E+05, 1.5686E+05, 1.5778E+05, 1.5870E+05,
     + 1.5963E+05, 1.6057E+05, 1.6151E+05, 1.6246E+05, 1.6341E+05,
     +  1.6436E+05, 1.6532E+05, 1.6629E+05, 1.6726E+05, 1.6824E+05,
     + 1.6922E+05, 1.7021E+05, 1.7120E+05, 1.7220E+05, 1.7321E+05,
     +  1.7422E+05, 1.7523E+05, 1.7625E+05, 1.7728E+05, 1.7831E+05,
     + 1.7935E+05, 1.8039E+05, 1.8144E+05, 1.8249E+05, 1.8355E+05,
     +  1.8462E+05, 1.8569E+05, 1.8677E+05, 1.8785E+05, 1.8894E+05,
     + 1.9003E+05, 1.9114E+05, 1.9224E+05, 1.9336E+05, 1.9447E+05,
     +  1.9560E+05, 1.9673E+05, 1.9787E+05, 1.9901E+05, 2.0016E+05,
     + 2.0131E+05, 2.0248E+05, 2.0364E+05, 2.0482E+05, 2.0600E+05,
     +  2.0719E+05, 2.0838E+05, 2.0958E+05, 2.1078E+05, 2.1200E+05,
     + 2.1322E+05, 2.1444E+05, 2.1567E+05, 2.1691E+05, 2.1816E+05,
     +  2.1941E+05, 2.2067E+05, 2.2193E+05, 2.2321E+05, 2.2448E+05,
     + 2.2577E+05, 2.2706E+05, 2.2836E+05, 2.2967E+05, 2.3098E+05,
     +  2.3230E+05, 2.3363E+05, 2.3496E+05, 2.3631E+05, 2.3765E+05,
     + 2.3901E+05, 2.4037E+05, 2.4174E+05, 2.4312E+05, 2.4451E+05,
     +  2.4590E+05, 2.4730E+05, 2.4870E+05, 2.5012E+05, 2.5154E+05,
     + 2.5297E+05, 2.5441E+05, 2.5585E+05, 2.5730E+05, 2.5876E+05,
     +  2.6023E+05, 2.6171E+05, 2.6319E+05, 2.6468E+05, 2.6618E+05,
     + 2.6769E+05, 2.6920E+05, 2.7072E+05, 2.7225E+05, 2.7379E+05,
     +  2.7534E+05, 2.7689E+05, 2.7846E+05, 2.8003E+05, 2.8161E+05,
     + 2.8319E+05, 2.8479E+05, 2.8639E+05, 2.8801E+05, 2.8963E+05,
     +  2.9126E+05, 2.9290E+05, 2.9454E+05, 2.9620E+05, 2.9786E+05,
     + 2.9953E+05, 3.0122E+05, 3.0291E+05, 3.0460E+05, 3.0631E+05,
     +  3.0803E+05, 3.0975E+05, 3.1149E+05, 3.1323E+05, 3.1499E+05,
     + 3.1675E+05, 3.1852E+05, 3.2030E+05, 3.2209E+05, 3.2389E+05,
     +  3.2569E+05, 3.2751E+05, 3.2934E+05, 3.3117E+05, 3.3302E+05,
     + 3.3487E+05, 3.3674E+05, 3.3861E+05, 3.4050E+05, 3.4239E+05,
     +  3.4430E+05, 3.4621E+05, 3.4813E+05, 3.5007E+05, 3.5201E+05,
     + 3.5396E+05, 3.5592E+05, 3.5790E+05, 3.5988E+05, 3.6188E+05,
     +  3.6388E+05, 3.6589E+05, 3.6792E+05, 3.6995E+05, 3.7200E+05,
     + 3.7405E+05, 3.7612E+05, 3.7820E+05, 3.8028E+05, 3.8238E+05,
     +  3.8449E+05, 3.8661E+05, 3.8874E+05, 3.9088E+05, 3.9304E+05,
     + 3.9520E+05, 3.9737E+05, 3.9956E+05, 4.0176E+05, 4.0396E+05,
     +  4.0618E+05, 4.0841E+05, 4.1066E+05, 4.1291E+05, 4.1517E+05,
     + 4.1745E+05, 4.1974E+05, 4.2204E+05, 4.2435E+05, 4.2667E+05,
     +  4.2900E+05, 4.3135E+05, 4.3371E+05, 4.3608E+05, 4.3846E+05,
     + 4.4085E+05, 4.4326E+05, 4.4568E+05, 4.4811E+05, 4.5055E+05,
     +  4.5301E+05, 4.5548E+05, 4.5796E+05, 4.6045E+05, 4.6295E+05,
     + 4.6547E+05, 4.6800E+05, 4.7054E+05, 4.7310E+05, 4.7567E+05,
     +  4.7825E+05, 4.8085E+05, 4.8345E+05, 4.8607E+05, 4.8871E+05,
     + 4.9136E+05, 4.9402E+05, 4.9669E+05, 4.9938E+05, 5.0208E+05,
     +  5.0479E+05, 5.0752E+05, 5.1026E+05, 5.1302E+05, 5.1578E+05,
     + 5.1857E+05, 5.2136E+05, 5.2417E+05, 5.2700E+05, 5.2984E+05,
     +  5.3269E+05, 5.3556E+05, 5.3844E+05, 5.4133E+05, 5.4424E+05,
     + 5.4717E+05, 5.5011E+05, 5.5306E+05, 5.5603E+05, 5.5901E+05,
     +  5.6201E+05, 5.6502E+05, 5.6805E+05, 5.7109E+05, 5.7415E+05,
     + 5.7722E+05, 5.8031E+05, 5.8341E+05, 5.8653E+05, 5.8966E+05,
     +  5.9281E+05, 5.9598E+05, 5.9916E+05, 6.0235E+05, 6.0557E+05,
     + 6.0879E+05, 6.1204E+05, 6.1530E+05, 6.1857E+05, 6.2186E+05,
     +  6.2517E+05, 6.2850E+05, 6.3184E+05, 6.3519E+05, 6.3857E+05,
     + 6.4196E+05, 6.4536E+05, 6.4879E+05, 6.5223E+05, 6.5568E+05,
     +  6.5916E+05, 6.6265E+05, 6.6616E+05, 6.6968E+05, 6.7322E+05,
     + 6.7678E+05, 6.8036E+05, 6.8395E+05, 6.8757E+05, 6.9120E+05,
     +  6.9484E+05, 6.9851E+05, 7.0219E+05, 7.0589E+05, 7.0961E+05,
     + 7.1335E+05, 7.1710E+05, 7.2088E+05, 7.2467E+05, 7.2848E+05,
     +  7.3231E+05, 7.3615E+05, 7.4002E+05, 7.4390E+05, 7.4781E+05,
     + 7.5173E+05, 7.5567E+05, 7.5963E+05, 7.6361E+05, 7.6761E+05,
     +  7.7163E+05, 7.7566E+05, 7.7972E+05, 7.8380E+05, 7.8789E+05,
     + 7.9201E+05, 7.9614E+05, 8.0030E+05, 8.0448E+05, 8.0867E+05,
     +  8.1289E+05, 8.1712E+05, 8.2138E+05, 8.2566E+05, 8.2995E+05,
     + 8.3427E+05, 8.3861E+05, 8.4297E+05, 8.4735E+05, 8.5175E+05,
     +  8.5618E+05, 8.6062E+05, 8.6509E+05, 8.6957E+05, 8.7408E+05,
     + 8.7861E+05, 8.8316E+05, 8.8773E+05, 8.9233E+05, 8.9694E+05,
     +  9.0158E+05, 9.0624E+05, 9.1093E+05, 9.1563E+05, 9.2036E+05,
     + 9.2511E+05, 9.2988E+05, 9.3468E+05, 9.3950E+05, 9.4434E+05,
     +  9.4920E+05, 9.5409E+05, 9.5900E+05, 9.6393E+05, 9.6889E+05,
     + 9.7387E+05, 9.7888E+05, 9.8390E+05, 9.8895E+05, 9.9403E+05,
     +  9.9913E+05, 1.0043E+06, 1.0094E+06, 1.0146E+06, 1.0198E+06,
     + 1.0250E+06, 1.0302E+06, 1.0355E+06, 1.0408E+06, 1.0461E+06,
     +  1.0515E+06, 1.0568E+06, 1.0622E+06, 1.0677E+06, 1.0731E+06,
     + 1.0786E+06, 1.0841E+06, 1.0896E+06, 1.0951E+06, 1.1007E+06,
     +  1.1063E+06, 1.1120E+06, 1.1176E+06, 1.1233E+06, 1.1290E+06,
     + 1.1347E+06, 1.1405E+06, 1.1463E+06, 1.1521E+06, 1.1579E+06,
     +  1.1638E+06, 1.1697E+06, 1.1756E+06, 1.1816E+06, 1.1876E+06,
     + 1.1936E+06, 1.1996E+06, 1.2057E+06, 1.2118E+06, 1.2179E+06,
     +  1.2240E+06, 1.2302E+06, 1.2364E+06, 1.2426E+06, 1.2489E+06,
     + 1.2552E+06, 1.2615E+06, 1.2679E+06, 1.2743E+06, 1.2807E+06,
     +  1.2871E+06, 1.2936E+06, 1.3001E+06, 1.3066E+06, 1.3132E+06,
     + 1.3198E+06, 1.3264E+06, 1.3330E+06, 1.3397E+06, 1.3464E+06,
     +  1.3532E+06, 1.3599E+06, 1.3667E+06, 1.3736E+06, 1.3804E+06,
     + 1.3873E+06, 1.3943E+06, 1.4012E+06, 1.4082E+06, 1.4153E+06,
     +  1.4223E+06, 1.4294E+06, 1.4365E+06, 1.4437E+06, 1.4509E+06,
     + 1.4581E+06, 1.4654E+06, 1.4727E+06, 1.4800E+06, 1.4873E+06,
     +  1.4947E+06, 1.5022E+06, 1.5096E+06, 1.5171E+06, 1.5246E+06,
     + 1.5322E+06, 1.5398E+06, 1.5474E+06, 1.5551E+06, 1.5628E+06,
     +  1.5705E+06, 1.5783E+06, 1.5861E+06, 1.5939E+06, 1.6018E+06,
     + 1.6097E+06, 1.6177E+06, 1.6257E+06, 1.6337E+06, 1.6417E+06,
     +  1.6498E+06, 1.6579E+06, 1.6661E+06, 1.6743E+06, 1.6826E+06,
     + 1.6908E+06, 1.6992E+06, 1.7075E+06, 1.7159E+06, 1.7243E+06,
     +  1.7328E+06, 1.7413E+06, 1.7498E+06, 1.7584E+06, 1.7670E+06,
     + 1.7757E+06, 1.7844E+06, 1.7931E+06, 1.8019E+06, 1.8107E+06,
     +  1.8196E+06, 1.8285E+06, 1.8374E+06, 1.8464E+06, 1.8554E+06,
     + 1.8645E+06, 1.8736E+06, 1.8827E+06, 1.8919E+06, 1.9011E+06,
     +  1.9103E+06, 1.9196E+06, 1.9290E+06, 1.9384E+06, 1.9478E+06,
     + 1.9573E+06, 1.9668E+06, 1.9763E+06, 1.9859E+06, 1.9956E+06,
     +  2.0053E+06, 2.0150E+06, 2.0248E+06, 2.0346E+06, 2.0444E+06,
     + 2.0543E+06, 2.0643E+06, 2.0743E+06, 2.0843E+06, 2.0944E+06,
     +  2.1045E+06, 2.1147E+06, 2.1249E+06, 2.1351E+06, 2.1454E+06,
     + 2.1558E+06, 2.1662E+06, 2.1766E+06, 2.1871E+06, 2.1976E+06,
     +  2.2082E+06, 2.2188E+06, 2.2295E+06, 2.2402E+06, 2.2510E+06,
     + 2.2618E+06, 2.2727E+06, 2.2836E+06, 2.2945E+06, 2.3055E+06,
     +  2.3166E+06, 2.3277E+06, 2.3389E+06, 2.3501E+06, 2.3613E+06,
     + 2.3726E+06, 2.3840E+06, 2.3954E+06, 2.4068E+06, 2.4183E+06,
     +  2.4299E+06, 2.4415E+06, 2.4531E+06, 2.4648E+06, 2.4766E+06,
     + 2.4884E+06, 2.5002E+06, 2.5122E+06, 2.5241E+06, 2.5361E+06,
     +  2.5482E+06, 2.5603E+06, 2.5725E+06, 2.5847E+06, 2.5970E+06,
     + 2.6093E+06, 2.6217E+06, 2.6341E+06, 2.6466E+06, 2.6592E+06,
     +  2.6718E+06, 2.6844E+06, 2.6972E+06, 2.7099E+06, 2.7227E+06,
     + 2.7356E+06, 2.7486E+06, 2.7615E+06, 2.7746E+06, 2.7877E+06,
     +  2.8009E+06, 2.8141E+06, 2.8273E+06, 2.8407E+06, 2.8541E+06,
     + 2.8675E+06, 2.8810E+06, 2.8946E+06, 2.9082E+06, 2.9219E+06,
     +  2.9356E+06, 2.9494E+06, 2.9633E+06, 2.9772E+06, 2.9912E+06,
     + 3.0052E+06, 3.0193E+06, 3.0335E+06, 3.0477E+06, 3.0620E+06,
     +  3.0763E+06, 3.0907E+06, 3.1052E+06, 3.1197E+06, 3.1343E+06,
     + 3.1490E+06, 3.1637E+06, 3.1785E+06, 3.1933E+06, 3.2082E+06,
     +  3.2232E+06, 3.2382E+06, 3.2533E+06, 3.2685E+06, 3.2837E+06,
     + 3.2990E+06, 3.3143E+06, 3.3298E+06, 3.3453E+06, 3.3608E+06,
     +  3.3764E+06, 3.3921E+06, 3.4079E+06, 3.4237E+06, 3.4396E+06,
     + 3.4555E+06, 3.4716E+06, 3.4877E+06, 3.5038E+06, 3.5201E+06,
     +  3.5364E+06, 3.5527E+06, 3.5692E+06, 3.5857E+06, 3.6022E+06,
     + 3.6189E+06, 3.6356E+06, 3.6524E+06, 3.6692E+06, 3.6862E+06,
     +  3.7032E+06, 3.7203E+06, 3.7374E+06, 3.7546E+06, 3.7719E+06,
     + 3.7893E+06, 3.8067E+06, 3.8242E+06, 3.8418E+06, 3.8595E+06,
     +  3.8772E+06, 3.8950E+06, 3.9129E+06, 3.9308E+06, 3.9489E+06,
     + 3.9670E+06, 3.9852E+06, 4.0034E+06, 4.0218E+06, 4.0402E+06,
     +  4.0587E+06, 4.0772E+06, 4.0959E+06, 4.1146E+06, 4.1334E+06,
     + 4.1523E+06, 4.1713E+06, 4.1903E+06, 4.2094E+06, 4.2286E+06,
     +  4.2479E+06, 4.2673E+06, 4.2867E+06, 4.3062E+06, 4.3258E+06,
     + 4.3455E+06, 4.3653E+06, 4.3852E+06, 4.4051E+06, 4.4251E+06,
     +  4.4452E+06, 4.4654E+06, 4.4856E+06, 4.5060E+06, 4.5264E+06,
     + 4.5470E+06, 4.5676E+06, 4.5883E+06, 4.6090E+06, 4.6299E+06,
     +  4.6508E+06, 4.6719E+06, 4.6930E+06, 4.7142E+06, 4.7355E+06,
     + 4.7569E+06, 4.7784E+06, 4.7999E+06, 4.8216E+06, 4.8433E+06,
     +  4.8652E+06, 4.8871E+06, 4.9091E+06, 4.9312E+06, 4.9534E+06,
     + 4.9757E+06, 4.9981E+06, 5.0205E+06, 5.0431E+06, 5.0658E+06,
     +  5.0885E+06, 5.1114E+06, 5.1343E+06, 5.1573E+06, 5.1804E+06,
     + 5.2037E+06, 5.2270E+06, 5.2504E+06, 5.2739E+06, 5.2975E+06,
     +  5.3212E+06, 5.3450E+06, 5.3689E+06, 5.3929E+06, 5.4170E+06,
     + 5.4412E+06, 5.4654E+06, 5.4898E+06, 5.5143E+06, 5.5389E+06,
     +  5.5636E+06, 5.5884E+06, 5.6133E+06, 5.6383E+06, 5.6633E+06,
     + 5.6885E+06, 5.7138E+06, 5.7392E+06, 5.7647E+06, 5.7903E+06,
     +  5.8161E+06, 5.8419E+06, 5.8678E+06, 5.8938E+06, 5.9199E+06,
     + 5.9462E+06, 5.9725E+06, 5.9990E+06, 6.0255E+06, 6.0522E+06,
     +  6.0790E+06, 6.1058E+06, 6.1328E+06, 6.1599E+06, 6.1871E+06,
     + 6.2145E+06, 6.2419E+06, 6.2694E+06, 6.2971E+06, 6.3248E+06,
     +  6.3527E+06, 6.3807E+06, 6.4088E+06, 6.4370E+06, 6.4653E+06,
     + 6.4938E+06, 6.5223E+06, 6.5510E+06, 6.5798E+06, 6.6087E+06,
     +  6.6377E+06, 6.6668E+06, 6.6961E+06, 6.7254E+06, 6.7549E+06,
     + 6.7845E+06, 6.8142E+06, 6.8441E+06, 6.8740E+06, 6.9041E+06,
     +  6.9343E+06, 6.9646E+06, 6.9951E+06, 7.0256E+06, 7.0563E+06,
     + 7.0871E+06, 7.1181E+06, 7.1491E+06, 7.1803E+06, 7.2116E+06,
     +  7.2430E+06, 7.2746E+06, 7.3063E+06, 7.3381E+06, 7.3700E+06,
     + 7.4020E+06, 7.4342E+06, 7.4665E+06, 7.4990E+06, 7.5316E+06,
     +  7.5643E+06, 7.5971E+06, 7.6300E+06, 7.6631E+06, 7.6964E+06,
     + 7.7297E+06, 7.7632E+06, 7.7968E+06, 7.8306E+06, 7.8645E+06,
     +  7.8985E+06, 7.9326E+06, 7.9669E+06, 8.0013E+06, 8.0359E+06,
     + 8.0706E+06, 8.1054E+06, 8.1404E+06, 8.1755E+06, 8.2108E+06,
     +  8.2461E+06, 8.2817E+06, 8.3173E+06, 8.3531E+06, 8.3891E+06,
     + 8.4252E+06, 8.4614E+06, 8.4978E+06, 8.5343E+06, 8.5710E+06,
     +  8.6078E+06, 8.6447E+06, 8.6818E+06, 8.7190E+06, 8.7564E+06,
     + 8.7939E+06, 8.8316E+06, 8.8694E+06, 8.9074E+06, 8.9455E+06,
     +  8.9838E+06, 9.0222E+06, 9.0608E+06, 9.0995E+06, 9.1384E+06,
     + 9.1774E+06, 9.2166E+06, 9.2559E+06, 9.2954E+06, 9.3350E+06,
     +  9.3748E+06, 9.4148E+06, 9.4549E+06, 9.4951E+06, 9.5356E+06,
     + 9.5761E+06, 9.6168E+06, 9.6577E+06, 9.6988E+06, 9.7400E+06,
     +  9.7814E+06, 9.8229E+06, 9.8646E+06, 9.9064E+06, 9.9484E+06,
     + 9.9906E+06, 1.0033E+07, 1.0075E+07, 1.0118E+07, 1.0161E+07,
     +  1.0204E+07, 1.0247E+07, 1.0290E+07, 1.0334E+07, 1.0378E+07,
     + 1.0421E+07, 1.0465E+07, 1.0510E+07, 1.0554E+07, 1.0598E+07,
     +  1.0643E+07, 1.0688E+07, 1.0733E+07, 1.0778E+07, 1.0824E+07,
     + 1.0869E+07, 1.0915E+07, 1.0961E+07, 1.1007E+07, 1.1053E+07,
     +  1.1100E+07, 1.1146E+07, 1.1193E+07, 1.1240E+07, 1.1287E+07,
     + 1.1334E+07, 1.1382E+07, 1.1430E+07, 1.1477E+07, 1.1525E+07,
     +  1.1574E+07, 1.1622E+07, 1.1671E+07, 1.1720E+07, 1.1769E+07,
     + 1.1818E+07, 1.1867E+07, 1.1917E+07, 1.1966E+07, 1.2016E+07,
     +  1.2066E+07, 1.2117E+07, 1.2167E+07, 1.2218E+07, 1.2269E+07,
     + 1.2320E+07, 1.2371E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       212
      data (QofT( 3,J),J=1,2408)/ 8.1368E+00, 9.2079E+00, 1.1853E+01,
     +  1.4940E+01, 1.8323E+01, 2.1954E+01, 2.5805E+01, 2.9857E+01,
     + 3.4098E+01, 3.8517E+01, 4.3109E+01, 4.7870E+01, 5.2799E+01,
     +  5.7894E+01, 6.3155E+01, 6.8580E+01, 7.4168E+01, 7.9918E+01,
     + 8.5830E+01, 9.1901E+01, 1.0451E+02, 1.1774E+02, 1.3156E+02,
     +  1.4596E+02, 1.6091E+02, 1.7639E+02, 1.9239E+02, 2.0889E+02,
     + 2.2587E+02, 2.4332E+02, 2.6122E+02, 2.7956E+02, 2.9832E+02,
     +  3.1750E+02, 3.3709E+02, 3.5707E+02, 3.7744E+02, 3.9818E+02,
     + 4.1930E+02, 4.4079E+02, 4.6263E+02, 4.8483E+02, 5.0737E+02,
     +  5.3026E+02, 5.5348E+02, 5.7703E+02, 6.0090E+02, 6.2510E+02,
     + 6.4962E+02, 6.7445E+02, 6.9959E+02, 7.2503E+02, 7.5078E+02,
     +  7.7683E+02, 8.0317E+02, 8.2980E+02, 8.5672E+02, 8.8393E+02,
     + 9.1142E+02, 9.3920E+02, 9.6725E+02, 9.9557E+02, 1.0242E+03,
     +  1.0530E+03, 1.0822E+03, 1.1116E+03, 1.1412E+03, 1.1711E+03,
     + 1.2013E+03, 1.2318E+03, 1.2624E+03, 1.2934E+03, 1.3246E+03,
     +  1.3560E+03, 1.3877E+03, 1.4196E+03, 1.4518E+03, 1.4842E+03,
     + 1.5168E+03, 1.5497E+03, 1.5829E+03, 1.6162E+03, 1.6498E+03,
     +  1.6836E+03, 1.7177E+03, 1.7520E+03, 1.7865E+03, 1.8212E+03,
     + 1.8562E+03, 1.8914E+03, 1.9268E+03, 1.9624E+03, 1.9982E+03,
     +  2.0343E+03, 2.0706E+03, 2.1071E+03, 2.1438E+03, 2.1808E+03,
     + 2.2179E+03, 2.2553E+03, 2.2929E+03, 2.3307E+03, 2.3687E+03,
     +  2.4070E+03, 2.4454E+03, 2.4840E+03, 2.5229E+03, 2.5620E+03,
     + 2.6013E+03, 2.6408E+03, 2.6805E+03, 2.7204E+03, 2.7605E+03,
     +  2.8009E+03, 2.8415E+03, 2.8822E+03, 2.9232E+03, 2.9644E+03,
     + 3.0058E+03, 3.0474E+03, 3.0892E+03, 3.1313E+03, 3.1735E+03,
     +  3.2160E+03, 3.2587E+03, 3.3016E+03, 3.3447E+03, 3.3881E+03,
     + 3.4316E+03, 3.4754E+03, 3.5194E+03, 3.5636E+03, 3.6080E+03,
     +  3.6527E+03, 3.6976E+03, 3.7427E+03, 3.7880E+03, 3.8336E+03,
     + 3.8793E+03, 3.9254E+03, 3.9716E+03, 4.0181E+03, 4.0648E+03,
     +  4.1117E+03, 4.1589E+03, 4.2064E+03, 4.2540E+03, 4.3019E+03,
     + 4.3501E+03, 4.3985E+03, 4.4471E+03, 4.4960E+03, 4.5452E+03,
     +  4.5946E+03, 4.6442E+03, 4.6941E+03, 4.7443E+03, 4.7947E+03,
     + 4.8454E+03, 4.8964E+03, 4.9476E+03, 4.9991E+03, 5.0509E+03,
     +  5.1029E+03, 5.1552E+03, 5.2078E+03, 5.2607E+03, 5.3138E+03,
     + 5.3673E+03, 5.4210E+03, 5.4750E+03, 5.5294E+03, 5.5840E+03,
     +  5.6389E+03, 5.6941E+03, 5.7496E+03, 5.8054E+03, 5.8616E+03,
     + 5.9180E+03, 5.9748E+03, 6.0319E+03, 6.0893E+03, 6.1470E+03,
     +  6.2050E+03, 6.2634E+03, 6.3221E+03, 6.3811E+03, 6.4405E+03,
     + 6.5002E+03, 6.5603E+03, 6.6207E+03, 6.6815E+03, 6.7426E+03,
     +  6.8041E+03, 6.8659E+03, 6.9281E+03, 6.9907E+03, 7.0536E+03,
     + 7.1169E+03, 7.1806E+03, 7.2447E+03, 7.3091E+03, 7.3739E+03,
     +  7.4391E+03, 7.5048E+03, 7.5708E+03, 7.6372E+03, 7.7040E+03,
     + 7.7712E+03, 7.8388E+03, 7.9069E+03, 7.9753E+03, 8.0442E+03,
     +  8.1135E+03, 8.1833E+03, 8.2534E+03, 8.3240E+03, 8.3951E+03,
     + 8.4666E+03, 8.5385E+03, 8.6109E+03, 8.6838E+03, 8.7571E+03,
     +  8.8308E+03, 8.9051E+03, 8.9798E+03, 9.0549E+03, 9.1306E+03,
     + 9.2068E+03, 9.2834E+03, 9.3605E+03, 9.4381E+03, 9.5162E+03,
     +  9.5948E+03, 9.6740E+03, 9.7536E+03, 9.8338E+03, 9.9144E+03,
     + 9.9956E+03, 1.0077E+04, 1.0160E+04, 1.0242E+04, 1.0326E+04,
     +  1.0410E+04, 1.0494E+04, 1.0579E+04, 1.0665E+04, 1.0751E+04,
     + 1.0837E+04, 1.0925E+04, 1.1013E+04, 1.1101E+04, 1.1190E+04,
     +  1.1280E+04, 1.1370E+04, 1.1461E+04, 1.1552E+04, 1.1644E+04,
     + 1.1737E+04, 1.1830E+04, 1.1924E+04, 1.2018E+04, 1.2113E+04,
     +  1.2209E+04, 1.2306E+04, 1.2403E+04, 1.2500E+04, 1.2599E+04,
     + 1.2698E+04, 1.2798E+04, 1.2898E+04, 1.2999E+04, 1.3101E+04,
     +  1.3203E+04, 1.3307E+04, 1.3411E+04, 1.3515E+04, 1.3620E+04,
     + 1.3726E+04, 1.3833E+04, 1.3941E+04, 1.4049E+04, 1.4158E+04,
     +  1.4267E+04, 1.4378E+04, 1.4489E+04, 1.4601E+04, 1.4714E+04,
     + 1.4827E+04, 1.4942E+04, 1.5057E+04, 1.5173E+04, 1.5289E+04,
     +  1.5407E+04, 1.5525E+04, 1.5644E+04, 1.5764E+04, 1.5885E+04,
     + 1.6006E+04, 1.6129E+04, 1.6252E+04, 1.6376E+04, 1.6501E+04,
     +  1.6627E+04, 1.6754E+04, 1.6881E+04, 1.7010E+04, 1.7139E+04,
     + 1.7270E+04, 1.7401E+04, 1.7533E+04, 1.7666E+04, 1.7800E+04,
     +  1.7935E+04, 1.8071E+04, 1.8207E+04, 1.8345E+04, 1.8484E+04,
     + 1.8623E+04, 1.8764E+04, 1.8906E+04, 1.9048E+04, 1.9192E+04,
     +  1.9336E+04, 1.9482E+04, 1.9628E+04, 1.9776E+04, 1.9925E+04,
     + 2.0074E+04, 2.0225E+04, 2.0377E+04, 2.0530E+04, 2.0684E+04,
     +  2.0839E+04, 2.0995E+04, 2.1152E+04, 2.1310E+04, 2.1470E+04,
     + 2.1630E+04, 2.1792E+04, 2.1954E+04, 2.2118E+04, 2.2283E+04,
     +  2.2450E+04, 2.2617E+04, 2.2785E+04, 2.2955E+04, 2.3126E+04,
     + 2.3298E+04, 2.3471E+04, 2.3646E+04, 2.3822E+04, 2.3999E+04,
     +  2.4177E+04, 2.4356E+04, 2.4537E+04, 2.4719E+04, 2.4902E+04,
     + 2.5087E+04, 2.5272E+04, 2.5459E+04, 2.5648E+04, 2.5838E+04,
     +  2.6029E+04, 2.6221E+04, 2.6415E+04, 2.6610E+04, 2.6806E+04,
     + 2.7004E+04, 2.7203E+04, 2.7404E+04, 2.7606E+04, 2.7809E+04,
     +  2.8014E+04, 2.8221E+04, 2.8428E+04, 2.8637E+04, 2.8848E+04,
     + 2.9060E+04, 2.9274E+04, 2.9489E+04, 2.9706E+04, 2.9924E+04,
     +  3.0143E+04, 3.0364E+04, 3.0587E+04, 3.0811E+04, 3.1037E+04,
     + 3.1265E+04, 3.1494E+04, 3.1724E+04, 3.1956E+04, 3.2190E+04,
     +  3.2426E+04, 3.2663E+04, 3.2902E+04, 3.3142E+04, 3.3384E+04,
     + 3.3628E+04, 3.3873E+04, 3.4120E+04, 3.4369E+04, 3.4620E+04,
     +  3.4872E+04, 3.5127E+04, 3.5382E+04, 3.5640E+04, 3.5900E+04,
     + 3.6161E+04, 3.6424E+04, 3.6689E+04, 3.6956E+04, 3.7224E+04,
     +  3.7495E+04, 3.7767E+04, 3.8042E+04, 3.8318E+04, 3.8596E+04,
     + 3.8876E+04, 3.9158E+04, 3.9442E+04, 3.9728E+04, 4.0016E+04,
     +  4.0305E+04, 4.0597E+04, 4.0891E+04, 4.1187E+04, 4.1485E+04,
     + 4.1785E+04, 4.2087E+04, 4.2392E+04, 4.2698E+04, 4.3006E+04,
     +  4.3317E+04, 4.3630E+04, 4.3945E+04, 4.4262E+04, 4.4581E+04,
     + 4.4902E+04, 4.5226E+04, 4.5552E+04, 4.5880E+04, 4.6211E+04,
     +  4.6543E+04, 4.6878E+04, 4.7216E+04, 4.7555E+04, 4.7897E+04,
     + 4.8241E+04, 4.8588E+04, 4.8937E+04, 4.9289E+04, 4.9643E+04,
     +  4.9999E+04, 5.0358E+04, 5.0719E+04, 5.1083E+04, 5.1449E+04,
     + 5.1818E+04, 5.2189E+04, 5.2563E+04, 5.2939E+04, 5.3318E+04,
     +  5.3700E+04, 5.4084E+04, 5.4471E+04, 5.4860E+04, 5.5252E+04,
     + 5.5647E+04, 5.6045E+04, 5.6445E+04, 5.6848E+04, 5.7254E+04,
     +  5.7662E+04, 5.8073E+04, 5.8487E+04, 5.8904E+04, 5.9324E+04,
     + 5.9747E+04, 6.0172E+04, 6.0601E+04, 6.1032E+04, 6.1466E+04,
     +  6.1903E+04, 6.2344E+04, 6.2787E+04, 6.3233E+04, 6.3682E+04,
     + 6.4135E+04, 6.4590E+04, 6.5048E+04, 6.5510E+04, 6.5975E+04,
     +  6.6443E+04, 6.6914E+04, 6.7388E+04, 6.7865E+04, 6.8346E+04,
     + 6.8830E+04, 6.9317E+04, 6.9808E+04, 7.0302E+04, 7.0799E+04,
     +  7.1299E+04, 7.1803E+04, 7.2310E+04, 7.2821E+04, 7.3335E+04,
     + 7.3853E+04, 7.4374E+04, 7.4899E+04, 7.5427E+04, 7.5959E+04,
     +  7.6494E+04, 7.7033E+04, 7.7576E+04, 7.8122E+04, 7.8672E+04,
     + 7.9225E+04, 7.9783E+04, 8.0344E+04, 8.0909E+04, 8.1477E+04,
     +  8.2050E+04, 8.2626E+04, 8.3206E+04, 8.3790E+04, 8.4378E+04,
     + 8.4970E+04, 8.5566E+04, 8.6166E+04, 8.6770E+04, 8.7377E+04,
     +  8.7989E+04, 8.8605E+04, 8.9226E+04, 8.9850E+04, 9.0478E+04,
     + 9.1111E+04, 9.1748E+04, 9.2389E+04, 9.3034E+04, 9.3684E+04,
     +  9.4338E+04, 9.4996E+04, 9.5659E+04, 9.6326E+04, 9.6998E+04,
     + 9.7674E+04, 9.8355E+04, 9.9040E+04, 9.9729E+04, 1.0042E+05,
     +  1.0112E+05, 1.0183E+05, 1.0253E+05, 1.0325E+05, 1.0396E+05,
     + 1.0469E+05, 1.0541E+05, 1.0614E+05, 1.0688E+05, 1.0762E+05,
     +  1.0837E+05, 1.0912E+05, 1.0988E+05, 1.1064E+05, 1.1140E+05,
     + 1.1218E+05, 1.1295E+05, 1.1373E+05, 1.1452E+05, 1.1531E+05,
     +  1.1611E+05, 1.1691E+05, 1.1772E+05, 1.1853E+05, 1.1935E+05,
     + 1.2017E+05, 1.2100E+05, 1.2184E+05, 1.2268E+05, 1.2352E+05,
     +  1.2437E+05, 1.2523E+05, 1.2609E+05, 1.2696E+05, 1.2783E+05,
     + 1.2871E+05, 1.2960E+05, 1.3049E+05, 1.3138E+05, 1.3229E+05,
     +  1.3320E+05, 1.3411E+05, 1.3503E+05, 1.3596E+05, 1.3689E+05,
     + 1.3783E+05, 1.3877E+05, 1.3972E+05, 1.4068E+05, 1.4164E+05,
     +  1.4261E+05, 1.4358E+05, 1.4457E+05, 1.4555E+05, 1.4655E+05,
     + 1.4755E+05, 1.4856E+05, 1.4957E+05, 1.5059E+05, 1.5162E+05,
     +  1.5265E+05, 1.5369E+05, 1.5474E+05, 1.5579E+05, 1.5685E+05,
     + 1.5792E+05, 1.5899E+05, 1.6008E+05, 1.6116E+05, 1.6226E+05,
     +  1.6336E+05, 1.6447E+05, 1.6559E+05, 1.6671E+05, 1.6784E+05,
     + 1.6898E+05, 1.7013E+05, 1.7128E+05, 1.7244E+05, 1.7361E+05,
     +  1.7478E+05, 1.7596E+05, 1.7715E+05, 1.7835E+05, 1.7956E+05,
     + 1.8077E+05, 1.8199E+05, 1.8322E+05, 1.8446E+05, 1.8570E+05,
     +  1.8695E+05, 1.8821E+05, 1.8948E+05, 1.9076E+05, 1.9204E+05,
     + 1.9334E+05, 1.9464E+05, 1.9595E+05, 1.9726E+05, 1.9859E+05,
     +  1.9992E+05, 2.0127E+05, 2.0262E+05, 2.0398E+05, 2.0535E+05,
     + 2.0672E+05, 2.0811E+05, 2.0951E+05, 2.1091E+05, 2.1232E+05,
     +  2.1374E+05, 2.1517E+05, 2.1661E+05, 2.1806E+05, 2.1952E+05,
     + 2.2099E+05, 2.2246E+05, 2.2395E+05, 2.2544E+05, 2.2695E+05,
     +  2.2846E+05, 2.2998E+05, 2.3151E+05, 2.3306E+05, 2.3461E+05,
     + 2.3617E+05, 2.3774E+05, 2.3932E+05, 2.4091E+05, 2.4252E+05,
     +  2.4413E+05, 2.4575E+05, 2.4738E+05, 2.4902E+05, 2.5067E+05,
     + 2.5234E+05, 2.5401E+05, 2.5569E+05, 2.5739E+05, 2.5909E+05,
     +  2.6080E+05, 2.6253E+05, 2.6427E+05, 2.6601E+05, 2.6777E+05,
     + 2.6954E+05, 2.7132E+05, 2.7311E+05, 2.7491E+05, 2.7673E+05,
     +  2.7855E+05, 2.8039E+05, 2.8223E+05, 2.8409E+05, 2.8596E+05,
     + 2.8784E+05, 2.8974E+05, 2.9164E+05, 2.9356E+05, 2.9549E+05,
     +  2.9743E+05, 2.9938E+05, 3.0135E+05, 3.0332E+05, 3.0531E+05,
     + 3.0731E+05, 3.0933E+05, 3.1135E+05, 3.1339E+05, 3.1544E+05,
     +  3.1750E+05, 3.1958E+05, 3.2167E+05, 3.2377E+05, 3.2589E+05,
     + 3.2801E+05, 3.3015E+05, 3.3231E+05, 3.3447E+05, 3.3665E+05,
     +  3.3885E+05, 3.4106E+05, 3.4328E+05, 3.4551E+05, 3.4776E+05,
     + 3.5002E+05, 3.5229E+05, 3.5458E+05, 3.5689E+05, 3.5920E+05,
     +  3.6153E+05, 3.6388E+05, 3.6624E+05, 3.6861E+05, 3.7100E+05,
     + 3.7340E+05, 3.7582E+05, 3.7825E+05, 3.8070E+05, 3.8316E+05,
     +  3.8564E+05, 3.8813E+05, 3.9064E+05, 3.9316E+05, 3.9570E+05,
     + 3.9825E+05, 4.0082E+05, 4.0340E+05, 4.0600E+05, 4.0861E+05,
     +  4.1125E+05, 4.1389E+05, 4.1655E+05, 4.1923E+05, 4.2193E+05,
     + 4.2464E+05, 4.2736E+05, 4.3011E+05, 4.3287E+05, 4.3564E+05,
     +  4.3844E+05, 4.4125E+05, 4.4407E+05, 4.4692E+05, 4.4978E+05,
     + 4.5266E+05, 4.5555E+05, 4.5846E+05, 4.6139E+05, 4.6434E+05,
     +  4.6731E+05, 4.7029E+05, 4.7329E+05, 4.7631E+05, 4.7934E+05,
     + 4.8240E+05, 4.8547E+05, 4.8856E+05, 4.9167E+05, 4.9480E+05,
     +  4.9795E+05, 5.0111E+05, 5.0430E+05, 5.0750E+05, 5.1072E+05,
     + 5.1396E+05, 5.1722E+05, 5.2050E+05, 5.2380E+05, 5.2712E+05,
     +  5.3046E+05, 5.3382E+05, 5.3719E+05, 5.4059E+05, 5.4401E+05,
     + 5.4745E+05, 5.5091E+05, 5.5438E+05, 5.5788E+05, 5.6140E+05,
     +  5.6494E+05, 5.6851E+05, 5.7209E+05, 5.7569E+05, 5.7932E+05,
     + 5.8296E+05, 5.8663E+05, 5.9032E+05, 5.9403E+05, 5.9776E+05,
     +  6.0152E+05, 6.0529E+05, 6.0909E+05, 6.1291E+05, 6.1675E+05,
     + 6.2062E+05, 6.2451E+05, 6.2842E+05, 6.3235E+05, 6.3631E+05,
     +  6.4029E+05, 6.4429E+05, 6.4832E+05, 6.5237E+05, 6.5644E+05,
     + 6.6054E+05, 6.6466E+05, 6.6880E+05, 6.7297E+05, 6.7716E+05,
     +  6.8138E+05, 6.8562E+05, 6.8989E+05, 6.9418E+05, 6.9849E+05,
     + 7.0284E+05, 7.0720E+05, 7.1159E+05, 7.1601E+05, 7.2045E+05,
     +  7.2492E+05, 7.2941E+05, 7.3393E+05, 7.3848E+05, 7.4305E+05,
     + 7.4765E+05, 7.5228E+05, 7.5693E+05, 7.6161E+05, 7.6631E+05,
     +  7.7104E+05, 7.7580E+05, 7.8059E+05, 7.8540E+05, 7.9025E+05,
     + 7.9512E+05, 8.0001E+05, 8.0494E+05, 8.0989E+05, 8.1488E+05,
     +  8.1989E+05, 8.2493E+05, 8.2999E+05, 8.3509E+05, 8.4022E+05,
     + 8.4537E+05, 8.5056E+05, 8.5577E+05, 8.6102E+05, 8.6629E+05,
     +  8.7160E+05, 8.7693E+05, 8.8230E+05, 8.8769E+05, 8.9312E+05,
     + 8.9857E+05, 9.0406E+05, 9.0958E+05, 9.1513E+05, 9.2071E+05,
     +  9.2633E+05, 9.3197E+05, 9.3765E+05, 9.4336E+05, 9.4910E+05,
     + 9.5488E+05, 9.6068E+05, 9.6652E+05, 9.7240E+05, 9.7830E+05,
     +  9.8424E+05, 9.9021E+05, 9.9622E+05, 1.0023E+06, 1.0083E+06,
     + 1.0144E+06, 1.0206E+06, 1.0268E+06, 1.0330E+06, 1.0392E+06,
     +  1.0455E+06, 1.0518E+06, 1.0582E+06, 1.0646E+06, 1.0710E+06,
     + 1.0774E+06, 1.0839E+06, 1.0905E+06, 1.0970E+06, 1.1037E+06,
     +  1.1103E+06, 1.1170E+06, 1.1237E+06, 1.1304E+06, 1.1372E+06,
     + 1.1441E+06, 1.1509E+06, 1.1578E+06, 1.1648E+06, 1.1718E+06,
     +  1.1788E+06, 1.1859E+06, 1.1930E+06, 1.2001E+06, 1.2073E+06,
     + 1.2145E+06, 1.2218E+06, 1.2291E+06, 1.2364E+06, 1.2438E+06,
     +  1.2512E+06, 1.2587E+06, 1.2662E+06, 1.2737E+06, 1.2813E+06,
     + 1.2889E+06, 1.2966E+06, 1.3043E+06, 1.3121E+06, 1.3199E+06,
     +  1.3277E+06, 1.3356E+06, 1.3435E+06, 1.3515E+06, 1.3595E+06,
     + 1.3676E+06, 1.3757E+06, 1.3838E+06, 1.3920E+06, 1.4003E+06,
     +  1.4085E+06, 1.4169E+06, 1.4253E+06, 1.4337E+06, 1.4421E+06,
     + 1.4507E+06, 1.4592E+06, 1.4678E+06, 1.4765E+06, 1.4852E+06,
     +  1.4939E+06, 1.5027E+06, 1.5116E+06, 1.5205E+06, 1.5294E+06,
     + 1.5384E+06, 1.5474E+06, 1.5565E+06, 1.5656E+06, 1.5748E+06,
     +  1.5841E+06, 1.5934E+06, 1.6027E+06, 1.6121E+06, 1.6215E+06,
     + 1.6310E+06, 1.6405E+06, 1.6501E+06, 1.6598E+06, 1.6695E+06,
     +  1.6792E+06, 1.6890E+06, 1.6989E+06, 1.7088E+06, 1.7188E+06,
     + 1.7288E+06, 1.7388E+06, 1.7490E+06, 1.7591E+06, 1.7694E+06,
     +  1.7797E+06, 1.7900E+06, 1.8004E+06, 1.8109E+06, 1.8214E+06,
     + 1.8319E+06, 1.8426E+06, 1.8532E+06, 1.8640E+06, 1.8748E+06,
     +  1.8856E+06, 1.8965E+06, 1.9075E+06, 1.9185E+06, 1.9296E+06,
     + 1.9408E+06, 1.9520E+06, 1.9632E+06, 1.9746E+06, 1.9859E+06,
     +  1.9974E+06, 2.0089E+06, 2.0205E+06, 2.0321E+06, 2.0438E+06,
     + 2.0555E+06, 2.0674E+06, 2.0792E+06, 2.0912E+06, 2.1032E+06,
     +  2.1153E+06, 2.1274E+06, 2.1396E+06, 2.1519E+06, 2.1642E+06,
     + 2.1766E+06, 2.1890E+06, 2.2016E+06, 2.2142E+06, 2.2268E+06,
     +  2.2395E+06, 2.2523E+06, 2.2652E+06, 2.2781E+06, 2.2911E+06,
     + 2.3042E+06, 2.3173E+06, 2.3305E+06, 2.3438E+06, 2.3571E+06,
     +  2.3705E+06, 2.3840E+06, 2.3976E+06, 2.4112E+06, 2.4249E+06,
     + 2.4386E+06, 2.4525E+06, 2.4664E+06, 2.4804E+06, 2.4944E+06,
     +  2.5086E+06, 2.5228E+06, 2.5370E+06, 2.5514E+06, 2.5658E+06,
     + 2.5803E+06, 2.5949E+06, 2.6095E+06, 2.6243E+06, 2.6391E+06,
     +  2.6540E+06, 2.6689E+06, 2.6840E+06, 2.6991E+06, 2.7143E+06,
     + 2.7296E+06, 2.7449E+06, 2.7604E+06, 2.7759E+06, 2.7915E+06,
     +  2.8071E+06, 2.8229E+06, 2.8387E+06, 2.8547E+06, 2.8707E+06,
     + 2.8867E+06, 2.9029E+06, 2.9192E+06, 2.9355E+06, 2.9519E+06,
     +  2.9684E+06, 2.9850E+06, 3.0017E+06, 3.0185E+06, 3.0353E+06,
     + 3.0522E+06, 3.0693E+06, 3.0864E+06, 3.1036E+06, 3.1208E+06,
     +  3.1382E+06, 3.1557E+06, 3.1732E+06, 3.1909E+06, 3.2086E+06,
     + 3.2264E+06, 3.2444E+06, 3.2624E+06, 3.2805E+06, 3.2987E+06,
     +  3.3169E+06, 3.3353E+06, 3.3538E+06, 3.3724E+06, 3.3910E+06,
     + 3.4098E+06, 3.4286E+06, 3.4476E+06, 3.4666E+06, 3.4858E+06,
     +  3.5050E+06, 3.5243E+06, 3.5438E+06, 3.5633E+06, 3.5829E+06,
     + 3.6027E+06, 3.6225E+06, 3.6424E+06, 3.6625E+06, 3.6826E+06,
     +  3.7028E+06, 3.7232E+06, 3.7436E+06, 3.7642E+06, 3.7848E+06,
     + 3.8056E+06, 3.8264E+06, 3.8474E+06, 3.8684E+06, 3.8896E+06,
     +  3.9109E+06, 3.9323E+06, 3.9538E+06, 3.9754E+06, 3.9971E+06,
     + 4.0189E+06, 4.0409E+06, 4.0629E+06, 4.0851E+06, 4.1073E+06,
     +  4.1297E+06, 4.1522E+06, 4.1748E+06, 4.1975E+06, 4.2203E+06,
     + 4.2433E+06, 4.2663E+06, 4.2895E+06, 4.3128E+06, 4.3362E+06,
     +  4.3597E+06, 4.3834E+06, 4.4071E+06, 4.4310E+06, 4.4550E+06,
     + 4.4791E+06, 4.5033E+06, 4.5277E+06, 4.5522E+06, 4.5768E+06,
     +  4.6015E+06, 4.6263E+06, 4.6513E+06, 4.6764E+06, 4.7016E+06,
     + 4.7270E+06, 4.7524E+06, 4.7780E+06, 4.8037E+06, 4.8296E+06,
     +  4.8556E+06, 4.8817E+06, 4.9079E+06, 4.9342E+06, 4.9607E+06,
     + 4.9874E+06, 5.0141E+06, 5.0410E+06, 5.0680E+06, 5.0952E+06,
     +  5.1225E+06, 5.1499E+06, 5.1774E+06, 5.2051E+06, 5.2329E+06,
     + 5.2609E+06, 5.2890E+06, 5.3172E+06, 5.3456E+06, 5.3741E+06,
     +  5.4028E+06, 5.4316E+06, 5.4605E+06, 5.4896E+06, 5.5188E+06,
     + 5.5482E+06, 5.5777E+06, 5.6073E+06, 5.6371E+06, 5.6671E+06,
     +  5.6972E+06, 5.7274E+06, 5.7578E+06, 5.7883E+06, 5.8190E+06,
     + 5.8498E+06, 5.8808E+06, 5.9119E+06, 5.9432E+06, 5.9747E+06,
     +  6.0062E+06, 6.0380E+06, 6.0699E+06, 6.1019E+06, 6.1341E+06,
     + 6.1665E+06, 6.1990E+06, 6.2317E+06, 6.2645E+06, 6.2975E+06,
     +  6.3307E+06, 6.3640E+06, 6.3974E+06, 6.4311E+06, 6.4649E+06,
     + 6.4988E+06, 6.5329E+06, 6.5672E+06, 6.6017E+06, 6.6363E+06,
     +  6.6711E+06, 6.7060E+06, 6.7412E+06, 6.7765E+06, 6.8119E+06,
     + 6.8476E+06, 6.8834E+06, 6.9193E+06, 6.9555E+06, 6.9918E+06,
     +  7.0283E+06, 7.0650E+06, 7.1018E+06, 7.1388E+06, 7.1760E+06,
     + 7.2134E+06, 7.2510E+06, 7.2887E+06, 7.3266E+06, 7.3647E+06,
     +  7.4030E+06, 7.4414E+06, 7.4801E+06, 7.5189E+06, 7.5579E+06,
     + 7.5971E+06, 7.6365E+06, 7.6761E+06, 7.7158E+06, 7.7558E+06,
     +  7.7959E+06, 7.8363E+06, 7.8768E+06, 7.9175E+06, 7.9584E+06,
     + 7.9995E+06, 8.0408E+06, 8.0823E+06, 8.1240E+06, 8.1659E+06,
     +  8.2079E+06, 8.2502E+06, 8.2927E+06, 8.3354E+06, 8.3783E+06,
     + 8.4214E+06, 8.4646E+06, 8.5081E+06, 8.5518E+06, 8.5958E+06,
     +  8.6399E+06, 8.6842E+06, 8.7287E+06, 8.7734E+06, 8.8184E+06,
     + 8.8636E+06, 8.9089E+06, 8.9545E+06, 9.0003E+06, 9.0463E+06,
     +  9.0926E+06, 9.1390E+06, 9.1857E+06, 9.2325E+06, 9.2796E+06,
     + 9.3270E+06, 9.3745E+06, 9.4223E+06, 9.4703E+06, 9.5185E+06,
     +  9.5669E+06, 9.6156E+06, 9.6645E+06, 9.7136E+06, 9.7629E+06,
     + 9.8125E+06, 9.8623E+06, 9.9124E+06, 9.9626E+06, 1.0013E+07,
     +  1.0064E+07, 1.0115E+07, 1.0166E+07, 1.0218E+07, 1.0269E+07,
     + 1.0321E+07, 1.0373E+07, 1.0426E+07, 1.0478E+07, 1.0531E+07,
     +  1.0584E+07, 1.0638E+07, 1.0691E+07, 1.0745E+07, 1.0799E+07,
     + 1.0854E+07, 1.0909E+07, 1.0963E+07, 1.1019E+07, 1.1074E+07,
     +  1.1130E+07, 1.1185E+07, 1.1242E+07, 1.1298E+07, 1.1355E+07,
     + 1.1412E+07, 1.1469E+07, 1.1526E+07, 1.1584E+07, 1.1642E+07,
     +  1.1700E+07, 1.1759E+07, 1.1818E+07, 1.1877E+07, 1.1936E+07,
     + 1.1996E+07, 1.2056E+07, 1.2116E+07, 1.2176E+07, 1.2237E+07,
     +  1.2298E+07, 1.2359E+07, 1.2421E+07, 1.2482E+07, 1.2545E+07,
     + 1.2607E+07, 1.2670E+07, 1.2733E+07, 1.2796E+07, 1.2859E+07,
     +  1.2923E+07, 1.2987E+07, 1.3052E+07, 1.3116E+07, 1.3181E+07,
     + 1.3247E+07, 1.3312E+07, 1.3378E+07, 1.3444E+07, 1.3511E+07,
     +  1.3578E+07, 1.3645E+07, 1.3712E+07, 1.3780E+07, 1.3848E+07,
     + 1.3916E+07, 1.3985E+07, 1.4054E+07, 1.4123E+07, 1.4192E+07,
     +  1.4262E+07, 1.4332E+07, 1.4403E+07, 1.4474E+07, 1.4545E+07,
     + 1.4616E+07, 1.4688E+07, 1.4760E+07, 1.4832E+07, 1.4905E+07,
     +  1.4978E+07, 1.5052E+07, 1.5125E+07, 1.5199E+07, 1.5274E+07,
     + 1.5349E+07, 1.5424E+07, 1.5499E+07, 1.5575E+07, 1.5651E+07,
     +  1.5727E+07, 1.5804E+07, 1.5881E+07, 1.5958E+07, 1.6036E+07,
     + 1.6114E+07, 1.6193E+07, 1.6272E+07, 1.6351E+07, 1.6430E+07,
     +  1.6510E+07, 1.6590E+07, 1.6671E+07, 1.6752E+07, 1.6833E+07,
     + 1.6915E+07, 1.6997E+07, 1.7079E+07, 1.7162E+07, 1.7245E+07,
     +  1.7329E+07, 1.7413E+07, 1.7497E+07, 1.7582E+07, 1.7667E+07,
     + 1.7752E+07, 1.7838E+07, 1.7924E+07, 1.8010E+07, 1.8097E+07,
     +  1.8184E+07, 1.8272E+07, 1.8360E+07, 1.8449E+07, 1.8537E+07,
     + 1.8627E+07, 1.8716E+07, 1.8806E+07, 1.8897E+07, 1.8987E+07,
     +  1.9079E+07, 1.9170E+07, 1.9262E+07, 1.9355E+07, 1.9447E+07,
     + 1.9541E+07, 1.9634E+07, 1.9728E+07, 1.9823E+07, 1.9917E+07,
     +  2.0013E+07, 2.0108E+07, 2.0205E+07, 2.0301E+07, 2.0398E+07,
     + 2.0495E+07, 2.0593E+07, 2.0691E+07, 2.0790E+07, 2.0889E+07,
     +  2.0989E+07, 2.1089E+07, 2.1189E+07, 2.1290E+07, 2.1391E+07,
     + 2.1493E+07, 2.1595E+07, 2.1697E+07, 2.1800E+07, 2.1904E+07,
     +  2.2008E+07, 2.2112E+07, 2.2217E+07, 2.2322E+07, 2.2428E+07,
     + 2.2534E+07, 2.2641E+07, 2.2748E+07, 2.2856E+07, 2.2964E+07,
     +  2.3072E+07, 2.3181E+07, 2.3291E+07, 2.3401E+07, 2.3511E+07,
     + 2.3622E+07, 2.3733E+07, 2.3845E+07, 2.3957E+07, 2.4070E+07,
     +  2.4183E+07, 2.4297E+07, 2.4411E+07, 2.4526E+07, 2.4641E+07,
     + 2.4757E+07, 2.4873E+07, 2.4990E+07, 2.5107E+07, 2.5225E+07,
     +  2.5343E+07, 2.5462E+07, 2.5581E+07, 2.5701E+07, 2.5822E+07,
     + 2.5942E+07, 2.6064E+07, 2.6185E+07, 2.6308E+07, 2.6431E+07,
     +  2.6554E+07, 2.6678E+07, 2.6802E+07, 2.6927E+07, 2.7053E+07,
     + 2.7179E+07, 2.7306E+07, 2.7433E+07, 2.7560E+07, 2.7689E+07,
     +  2.7817E+07, 2.7947E+07, 2.8077E+07, 2.8207E+07, 2.8338E+07,
     + 2.8470E+07, 2.8602E+07, 2.8734E+07, 2.8867E+07, 2.9001E+07,
     +  2.9136E+07, 2.9270E+07, 2.9406E+07, 2.9542E+07, 2.9679E+07,
     + 2.9816E+07, 2.9954E+07, 3.0092E+07, 3.0231E+07, 3.0370E+07,
     +  3.0511E+07, 3.0651E+07, 3.0793E+07, 3.0934E+07, 3.1077E+07,
     + 3.1220E+07, 3.1364E+07, 3.1508E+07, 3.1653E+07, 3.1798E+07,
     +  3.1945E+07, 3.2091E+07, 3.2239E+07, 3.2387E+07, 3.2535E+07,
     + 3.2684E+07, 3.2834E+07, 3.2985E+07, 3.3136E+07, 3.3288E+07,
     +  3.3440E+07, 3.3593E+07, 3.3747E+07, 3.3901E+07, 3.4056E+07,
     + 3.4211E+07, 3.4368E+07, 3.4524E+07, 3.4682E+07, 3.4840E+07,
     +  3.4999E+07, 3.5159E+07, 3.5319E+07, 3.5480E+07, 3.5641E+07,
     + 3.5803E+07, 3.5966E+07, 3.6130E+07, 3.6294E+07, 3.6459E+07,
     +  3.6624E+07, 3.6791E+07, 3.6958E+07, 3.7125E+07, 3.7294E+07,
     + 3.7463E+07, 3.7632E+07, 3.7803E+07, 3.7974E+07, 3.8146E+07,
     +  3.8318E+07, 3.8492E+07, 3.8666E+07, 3.8840E+07, 3.9016E+07,
     + 3.9192E+07, 3.9369E+07, 3.9546E+07, 3.9725E+07, 3.9904E+07,
     +  4.0084E+07, 4.0264E+07, 4.0445E+07, 4.0627E+07, 4.0810E+07,
     + 4.0994E+07, 4.1178E+07, 4.1363E+07, 4.1549E+07, 4.1735E+07,
     +  4.1923E+07, 4.2111E+07, 4.2300E+07, 4.2489E+07, 4.2680E+07,
     + 4.2871E+07, 4.3063E+07, 4.3256E+07, 4.3449E+07, 4.3644E+07,
     +  4.3839E+07, 4.4035E+07, 4.4232E+07, 4.4429E+07, 4.4627E+07,
     + 4.4827E+07, 4.5027E+07, 4.5227E+07, 4.5429E+07, 4.5631E+07,
     +  4.5835E+07, 4.6039E+07, 4.6244E+07, 4.6449E+07, 4.6656E+07,
     + 4.6863E+07, 4.7071E+07, 4.7281E+07, 4.7490E+07, 4.7701E+07,
     +  4.7913E+07, 4.8125E+07, 4.8339E+07, 4.8553E+07, 4.8768E+07,
     + 4.8984E+07, 4.9201E+07, 4.9419E+07, 4.9637E+07, 4.9857E+07,
     +  5.0077E+07, 5.0298E+07, 5.0520E+07, 5.0743E+07, 5.0967E+07,
     + 5.1192E+07, 5.1418E+07, 5.1645E+07, 5.1872E+07, 5.2101E+07,
     +  5.2330E+07, 5.2560E+07, 5.2791E+07, 5.3024E+07, 5.3257E+07,
     + 5.3491E+07, 5.3726E+07, 5.3962E+07, 5.4198E+07, 5.4436E+07,
     +  5.4675E+07, 5.4915E+07, 5.5155E+07, 5.5397E+07, 5.5640E+07,
     + 5.5883E+07, 5.6128E+07, 5.6373E+07, 5.6620E+07, 5.6867E+07,
     +  5.7116E+07, 5.7365E+07, 5.7616E+07, 5.7867E+07, 5.8120E+07,
     + 5.8373E+07, 5.8628E+07, 5.8883E+07, 5.9139E+07, 5.9397E+07,
     +  5.9656E+07, 5.9915E+07, 6.0176E+07, 6.0437E+07, 6.0700E+07,
     + 6.0964E+07, 6.1228E+07, 6.1494E+07, 6.1761E+07, 6.2029E+07,
     +  6.2298E+07, 6.2568E+07, 6.2839E+07, 6.3111E+07, 6.3385E+07,
     + 6.3659E+07, 6.3934E+07, 6.4211E+07, 6.4488E+07, 6.4767E+07,
     +  6.5047E+07, 6.5328E+07, 6.5610E+07, 6.5893E+07, 6.6177E+07,
     + 6.6462E+07, 6.6749E+07, 6.7036E+07, 6.7325E+07, 6.7615E+07,
     +  6.7906E+07, 6.8198E+07, 6.8491E+07, 6.8786E+07, 6.9081E+07,
     + 6.9378E+07, 6.9676E+07, 6.9975E+07, 7.0275E+07, 7.0577E+07,
     +  7.0879E+07, 7.1183E+07, 7.1488E+07, 7.1794E+07, 7.2102E+07,
     + 7.2410E+07, 7.2720E+07, 7.3031E+07, 7.3343E+07, 7.3656E+07,
     +  7.3971E+07, 7.4287E+07, 7.4604E+07, 7.4922E+07, 7.5242E+07,
     + 7.5563E+07, 7.5885E+07, 7.6208E+07, 7.6532E+07, 7.6858E+07,
     +  7.7185E+07, 7.7514E+07, 7.7843E+07, 7.8174E+07, 7.8506E+07,
     + 7.8840E+07, 7.9175E+07, 7.9511E+07, 7.9848E+07, 8.0187E+07,
     +  8.0527E+07, 8.0868E+07, 8.1210E+07, 8.1554E+07, 8.1900E+07,
     + 8.2246E+07, 8.2594E+07, 8.2943E+07, 8.3294E+07, 8.3646E+07,
     +  8.3999E+07, 8.4354E+07, 8.4710E+07, 8.5067E+07, 8.5426E+07,
     + 8.5786E+07, 8.6148E+07, 8.6511E+07, 8.6875E+07, 8.7241E+07,
     +  8.7608E+07, 8.7976E+07, 8.8346E+07, 8.8718E+07, 8.9090E+07,
     + 8.9465E+07, 8.9840E+07, 9.0217E+07, 9.0596E+07, 9.0976E+07,
     +  9.1357E+07, 9.1740E+07, 9.2125E+07, 9.2510E+07, 9.2898E+07,
     + 9.3287E+07, 9.3677E+07, 9.4069E+07, 9.4462E+07, 9.4857E+07,
     +  9.5253E+07, 9.5651E+07, 9.6050E+07, 9.6451E+07, 9.6853E+07,
     + 9.7257E+07, 9.7662E+07, 9.8069E+07, 9.8477E+07, 9.8887E+07,
     +  9.9299E+07, 9.9712E+07, 1.0013E+08, 1.0054E+08, 1.0096E+08,
     + 1.0138E+08, 1.0180E+08, 1.0222E+08, 1.0265E+08, 1.0307E+08,
     +  1.0350E+08, 1.0393E+08, 1.0436E+08, 1.0479E+08, 1.0523E+08,
     + 1.0566E+08, 1.0610E+08, 1.0654E+08, 1.0698E+08, 1.0742E+08,
     +  1.0787E+08, 1.0831E+08, 1.0876E+08, 1.0921E+08, 1.0966E+08,
     + 1.1011E+08, 1.1056E+08, 1.1102E+08, 1.1148E+08, 1.1194E+08,
     +  1.1240E+08, 1.1286E+08, 1.1332E+08, 1.1379E+08, 1.1426E+08,
     + 1.1473E+08, 1.1520E+08, 1.1567E+08, 1.1615E+08, 1.1662E+08,
     +  1.1710E+08, 1.1758E+08, 1.1806E+08, 1.1855E+08, 1.1903E+08,
     + 1.1952E+08, 1.2001E+08, 1.2050E+08, 1.2099E+08, 1.2149E+08,
     +  1.2198E+08, 1.2248E+08, 1.2298E+08, 1.2348E+08, 1.2399E+08,
     + 1.2449E+08, 1.2500E+08, 1.2551E+08, 1.2602E+08, 1.2654E+08,
     +  1.2705E+08, 1.2757E+08, 1.2809E+08, 1.2861E+08, 1.2913E+08,
     + 1.2966E+08, 1.3018E+08, 1.3071E+08, 1.3124E+08, 1.3177E+08,
     +  1.3231E+08, 1.3285E+08, 1.3338E+08, 1.3393E+08, 1.3447E+08,
     + 1.3501E+08, 1.3556E+08, 1.3611E+08, 1.3666E+08, 1.3721E+08,
     +  1.3777E+08, 1.3832E+08, 1.3888E+08, 1.3944E+08, 1.4001E+08,
     + 1.4057E+08, 1.4114E+08, 1.4171E+08, 1.4228E+08, 1.4285E+08,
     +  1.4343E+08, 1.4400E+08, 1.4458E+08, 1.4517E+08, 1.4575E+08,
     + 1.4634E+08, 1.4692E+08, 1.4751E+08, 1.4811E+08, 1.4870E+08,
     +  1.4930E+08, 1.4990E+08, 1.5050E+08, 1.5110E+08, 1.5171E+08,
     + 1.5232E+08, 1.5293E+08, 1.5354E+08, 1.5415E+08, 1.5477E+08,
     +  1.5539E+08, 1.5601E+08, 1.5663E+08, 1.5726E+08, 1.5789E+08,
     + 1.5852E+08, 1.5915E+08, 1.5979E+08, 1.6042E+08, 1.6106E+08,
     +  1.6170E+08, 1.6235E+08, 1.6300E+08, 1.6364E+08, 1.6430E+08,
     + 1.6495E+08, 1.6561E+08, 1.6626E+08, 1.6692E+08, 1.6759E+08,
     +  1.6825E+08, 1.6892E+08, 1.6959E+08, 1.7026E+08, 1.7094E+08,
     + 1.7162E+08, 1.7230E+08, 1.7298E+08, 1.7367E+08, 1.7435E+08,
     +  1.7504E+08, 1.7574E+08, 1.7643E+08, 1.7713E+08, 1.7783E+08,
     + 1.7853E+08, 1.7924E+08, 1.7994E+08, 1.8065E+08, 1.8137E+08,
     +  1.8208E+08, 1.8280E+08, 1.8352E+08, 1.8424E+08, 1.8497E+08,
     + 1.8570E+08, 1.8643E+08, 1.8716E+08, 1.8790E+08, 1.8864E+08,
     +  1.8938E+08, 1.9012E+08, 1.9087E+08, 1.9162E+08, 1.9237E+08,
     + 1.9312E+08, 1.9388E+08, 1.9464E+08, 1.9540E+08, 1.9617E+08,
     +  1.9694E+08, 1.9771E+08, 1.9848E+08, 1.9926E+08, 2.0004E+08,
     + 2.0082E+08, 2.0160E+08, 2.0239E+08, 2.0318E+08, 2.0398E+08,
     +  2.0477E+08, 2.0557E+08, 2.0637E+08, 2.0718E+08, 2.0798E+08,
     + 2.0879E+08, 2.0961E+08, 2.1042E+08, 2.1124E+08, 2.1206E+08,
     +  2.1289E+08, 2.1372E+08, 2.1455E+08, 2.1538E+08, 2.1622E+08,
     + 2.1706E+08, 2.1790E+08, 2.1874E+08, 2.1959E+08, 2.2044E+08,
     +  2.2130E+08, 2.2215E+08, 2.2302E+08, 2.2388E+08, 2.2475E+08,
     + 2.2561E+08, 2.2649E+08, 2.2736E+08, 2.2824E+08, 2.2912E+08,
     +  2.3001E+08, 2.3090E+08, 2.3179E+08, 2.3268E+08, 2.3358E+08,
     + 2.3448E+08, 2.3538E+08, 2.3629E+08, 2.3720E+08, 2.3811E+08,
     +  2.3903E+08, 2.3995E+08, 2.4087E+08, 2.4180E+08, 2.4273E+08,
     + 2.4366E+08, 2.4460E+08, 2.4554E+08, 2.4648E+08, 2.4742E+08,
     +  2.4837E+08, 2.4933E+08, 2.5028E+08, 2.5124E+08, 2.5220E+08,
     + 2.5317E+08, 2.5414E+08, 2.5511E+08, 2.5609E+08, 2.5706E+08,
     +  2.5805E+08, 2.5903E+08, 2.6002E+08, 2.6102E+08, 2.6201E+08,
     + 2.6301E+08, 2.6401E+08, 2.6502E+08, 2.6603E+08, 2.6705E+08,
     +  2.6806E+08, 2.6908E+08, 2.7011E+08, 2.7114E+08, 2.7217E+08,
     + 2.7320E+08, 2.7424E+08, 2.7528E+08, 2.7633E+08, 2.7738E+08,
     +  2.7843E+08, 2.7949E+08, 2.8055E+08, 2.8161E+08, 2.8268E+08,
     + 2.8375E+08, 2.8483E+08, 2.8591E+08, 2.8699E+08, 2.8807E+08,
     +  2.8916E+08, 2.9026E+08, 2.9136E+08, 2.9246E+08, 2.9356E+08,
     + 2.9467E+08, 2.9578E+08, 2.9690E+08, 2.9802E+08, 2.9914E+08,
     +  3.0027E+08, 3.0140E+08, 3.0254E+08, 3.0368E+08, 3.0482E+08,
     + 3.0597E+08, 3.0712E+08, 3.0828E+08, 3.0944E+08, 3.1060E+08,
     +  3.1177E+08, 3.1294E+08, 3.1411E+08, 3.1529E+08, 3.1648E+08,
     + 3.1766E+08, 3.1886E+08, 3.2005E+08, 3.2125E+08, 3.2245E+08,
     +  3.2366E+08, 3.2487E+08, 3.2609E+08, 3.2731E+08, 3.2853E+08,
     + 3.2976E+08, 3.3100E+08, 3.3223E+08, 3.3347E+08, 3.3472E+08,
     +  3.3597E+08, 3.3722E+08, 3.3848E+08, 3.3974E+08, 3.4101E+08,
     + 3.4228E+08, 3.4355E+08, 3.4483E+08, 3.4612E+08, 3.4741E+08,
     +  3.4870E+08, 3.4999E+08, 3.5130E+08, 3.5260E+08, 3.5391E+08,
     + 3.5523E+08, 3.5655E+08, 3.5787E+08, 3.5920E+08, 3.6053E+08,
     +  3.6187E+08, 3.6321E+08, 3.6455E+08, 3.6590E+08, 3.6726E+08,
     + 3.6862E+08, 3.6998E+08, 3.7135E+08, 3.7272E+08, 3.7410E+08,
     +  3.7548E+08, 3.7687E+08, 3.7826E+08, 3.7966E+08, 3.8106E+08,
     + 3.8247E+08, 3.8388E+08, 3.8529E+08, 3.8671E+08, 3.8814E+08,
     +  3.8957E+08, 3.9100E+08, 3.9244E+08, 3.9389E+08, 3.9534E+08,
     + 3.9679E+08, 3.9825E+08, 3.9971E+08, 4.0118E+08, 4.0265E+08,
     +  4.0413E+08, 4.0561E+08, 4.0710E+08, 4.0860E+08, 4.1009E+08,
     + 4.1160E+08, 4.1310E+08, 4.1462E+08, 4.1614E+08, 4.1766E+08,
     +  4.1919E+08, 4.2072E+08, 4.2226E+08, 4.2380E+08, 4.2535E+08,
     + 4.2691E+08, 4.2846E+08, 4.3003E+08, 4.3160E+08, 4.3317E+08,
     +  4.3475E+08, 4.3634E+08, 4.3793E+08, 4.3952E+08, 4.4112E+08,
     + 4.4273E+08, 4.4434E+08, 4.4596E+08, 4.4758E+08, 4.4921E+08,
     +  4.5084E+08, 4.5248E+08, 4.5412E+08, 4.5577E+08, 4.5743E+08,
     + 4.5909E+08, 4.6075E+08, 4.6243E+08, 4.6410E+08, 4.6578E+08,
     +  4.6747E+08, 4.6916E+08, 4.7086E+08, 4.7257E+08, 4.7428E+08,
     + 4.7599E+08, 4.7771E+08, 4.7944E+08, 4.8117E+08, 4.8291E+08,
     +  4.8466E+08, 4.8641E+08, 4.8816E+08, 4.8992E+08, 4.9169E+08,
     + 4.9346E+08, 4.9524E+08, 4.9703E+08, 4.9882E+08, 5.0061E+08,
     +  5.0242E+08, 5.0422E+08, 5.0604E+08, 5.0786E+08, 5.0968E+08,
     + 5.1151E+08, 5.1335E+08, 5.1520E+08, 5.1705E+08, 5.1890E+08,
     +  5.2076E+08, 5.2263E+08, 5.2451E+08, 5.2639E+08, 5.2827E+08,
     + 5.3016E+08, 5.3206E+08, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 3,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       312
      data (QofT( 4,J),J=1,2408)/ 1.6270E+01, 1.8421E+01, 2.3716E+01,
     +  2.9896E+01, 3.6668E+01, 4.3936E+01, 5.1644E+01, 5.9756E+01,
     + 6.8244E+01, 7.7089E+01, 8.6281E+01, 9.5812E+01, 1.0568E+02,
     +  1.1588E+02, 1.2641E+02, 1.3727E+02, 1.4845E+02, 1.5996E+02,
     + 1.7180E+02, 1.8395E+02, 2.0920E+02, 2.3567E+02, 2.6334E+02,
     +  2.9215E+02, 3.2208E+02, 3.5307E+02, 3.8510E+02, 4.1813E+02,
     + 4.5212E+02, 4.8704E+02, 5.2287E+02, 5.5957E+02, 5.9713E+02,
     +  6.3553E+02, 6.7473E+02, 7.1473E+02, 7.5550E+02, 7.9703E+02,
     + 8.3930E+02, 8.8231E+02, 9.2604E+02, 9.7047E+02, 1.0156E+03,
     +  1.0614E+03, 1.1079E+03, 1.1550E+03, 1.2028E+03, 1.2512E+03,
     + 1.3003E+03, 1.3500E+03, 1.4003E+03, 1.4513E+03, 1.5028E+03,
     +  1.5550E+03, 1.6077E+03, 1.6610E+03, 1.7149E+03, 1.7693E+03,
     + 1.8244E+03, 1.8800E+03, 1.9361E+03, 1.9928E+03, 2.0500E+03,
     +  2.1078E+03, 2.1661E+03, 2.2250E+03, 2.2844E+03, 2.3442E+03,
     + 2.4047E+03, 2.4656E+03, 2.5270E+03, 2.5889E+03, 2.6514E+03,
     +  2.7143E+03, 2.7777E+03, 2.8416E+03, 2.9060E+03, 2.9709E+03,
     + 3.0362E+03, 3.1021E+03, 3.1684E+03, 3.2352E+03, 3.3024E+03,
     +  3.3701E+03, 3.4383E+03, 3.5069E+03, 3.5760E+03, 3.6455E+03,
     + 3.7155E+03, 3.7859E+03, 3.8568E+03, 3.9281E+03, 3.9999E+03,
     +  4.0721E+03, 4.1447E+03, 4.2178E+03, 4.2913E+03, 4.3653E+03,
     + 4.4396E+03, 4.5145E+03, 4.5897E+03, 4.6654E+03, 4.7415E+03,
     +  4.8180E+03, 4.8950E+03, 4.9723E+03, 5.0502E+03, 5.1284E+03,
     + 5.2070E+03, 5.2861E+03, 5.3656E+03, 5.4456E+03, 5.5259E+03,
     +  5.6067E+03, 5.6879E+03, 5.7695E+03, 5.8515E+03, 5.9340E+03,
     + 6.0169E+03, 6.1002E+03, 6.1839E+03, 6.2681E+03, 6.3527E+03,
     +  6.4378E+03, 6.5232E+03, 6.6091E+03, 6.6954E+03, 6.7822E+03,
     + 6.8694E+03, 6.9570E+03, 7.0451E+03, 7.1337E+03, 7.2227E+03,
     +  7.3120E+03, 7.4019E+03, 7.4922E+03, 7.5830E+03, 7.6742E+03,
     + 7.7659E+03, 7.8581E+03, 7.9507E+03, 8.0437E+03, 8.1373E+03,
     +  8.2313E+03, 8.3258E+03, 8.4208E+03, 8.5163E+03, 8.6122E+03,
     + 8.7087E+03, 8.8055E+03, 8.9030E+03, 9.0009E+03, 9.0993E+03,
     +  9.1982E+03, 9.2977E+03, 9.3977E+03, 9.4981E+03, 9.5992E+03,
     + 9.7007E+03, 9.8028E+03, 9.9053E+03, 1.0009E+04, 1.0112E+04,
     +  1.0216E+04, 1.0321E+04, 1.0427E+04, 1.0533E+04, 1.0639E+04,
     + 1.0746E+04, 1.0854E+04, 1.0962E+04, 1.1071E+04, 1.1180E+04,
     +  1.1290E+04, 1.1401E+04, 1.1512E+04, 1.1624E+04, 1.1736E+04,
     + 1.1849E+04, 1.1963E+04, 1.2077E+04, 1.2192E+04, 1.2308E+04,
     +  1.2424E+04, 1.2541E+04, 1.2659E+04, 1.2777E+04, 1.2896E+04,
     + 1.3016E+04, 1.3136E+04, 1.3257E+04, 1.3379E+04, 1.3502E+04,
     +  1.3625E+04, 1.3749E+04, 1.3873E+04, 1.3999E+04, 1.4125E+04,
     + 1.4252E+04, 1.4379E+04, 1.4508E+04, 1.4637E+04, 1.4767E+04,
     +  1.4898E+04, 1.5029E+04, 1.5161E+04, 1.5294E+04, 1.5428E+04,
     + 1.5563E+04, 1.5699E+04, 1.5835E+04, 1.5972E+04, 1.6110E+04,
     +  1.6249E+04, 1.6389E+04, 1.6530E+04, 1.6671E+04, 1.6814E+04,
     + 1.6957E+04, 1.7101E+04, 1.7246E+04, 1.7392E+04, 1.7539E+04,
     +  1.7687E+04, 1.7836E+04, 1.7986E+04, 1.8136E+04, 1.8288E+04,
     + 1.8441E+04, 1.8594E+04, 1.8749E+04, 1.8905E+04, 1.9061E+04,
     +  1.9219E+04, 1.9377E+04, 1.9537E+04, 1.9698E+04, 1.9859E+04,
     + 2.0022E+04, 2.0186E+04, 2.0351E+04, 2.0517E+04, 2.0684E+04,
     +  2.0852E+04, 2.1022E+04, 2.1192E+04, 2.1364E+04, 2.1536E+04,
     + 2.1710E+04, 2.1885E+04, 2.2061E+04, 2.2238E+04, 2.2417E+04,
     +  2.2597E+04, 2.2777E+04, 2.2960E+04, 2.3143E+04, 2.3327E+04,
     + 2.3513E+04, 2.3700E+04, 2.3888E+04, 2.4078E+04, 2.4269E+04,
     +  2.4461E+04, 2.4654E+04, 2.4849E+04, 2.5045E+04, 2.5242E+04,
     + 2.5441E+04, 2.5641E+04, 2.5842E+04, 2.6045E+04, 2.6249E+04,
     +  2.6455E+04, 2.6661E+04, 2.6870E+04, 2.7080E+04, 2.7291E+04,
     + 2.7503E+04, 2.7717E+04, 2.7933E+04, 2.8150E+04, 2.8368E+04,
     +  2.8588E+04, 2.8810E+04, 2.9033E+04, 2.9258E+04, 2.9484E+04,
     + 2.9711E+04, 2.9941E+04, 3.0171E+04, 3.0404E+04, 3.0638E+04,
     +  3.0873E+04, 3.1111E+04, 3.1350E+04, 3.1590E+04, 3.1832E+04,
     + 3.2076E+04, 3.2322E+04, 3.2569E+04, 3.2818E+04, 3.3069E+04,
     +  3.3321E+04, 3.3575E+04, 3.3831E+04, 3.4089E+04, 3.4349E+04,
     + 3.4610E+04, 3.4873E+04, 3.5138E+04, 3.5405E+04, 3.5674E+04,
     +  3.5944E+04, 3.6216E+04, 3.6491E+04, 3.6767E+04, 3.7045E+04,
     + 3.7325E+04, 3.7607E+04, 3.7891E+04, 3.8177E+04, 3.8465E+04,
     +  3.8755E+04, 3.9048E+04, 3.9342E+04, 3.9638E+04, 3.9936E+04,
     + 4.0236E+04, 4.0539E+04, 4.0843E+04, 4.1149E+04, 4.1458E+04,
     +  4.1769E+04, 4.2083E+04, 4.2398E+04, 4.2715E+04, 4.3035E+04,
     + 4.3357E+04, 4.3681E+04, 4.4007E+04, 4.4336E+04, 4.4667E+04,
     +  4.5001E+04, 4.5336E+04, 4.5674E+04, 4.6015E+04, 4.6358E+04,
     + 4.6703E+04, 4.7050E+04, 4.7401E+04, 4.7753E+04, 4.8108E+04,
     +  4.8465E+04, 4.8825E+04, 4.9188E+04, 4.9553E+04, 4.9921E+04,
     + 5.0291E+04, 5.0663E+04, 5.1039E+04, 5.1417E+04, 5.1797E+04,
     +  5.2181E+04, 5.2567E+04, 5.2955E+04, 5.3347E+04, 5.3741E+04,
     + 5.4137E+04, 5.4537E+04, 5.4940E+04, 5.5345E+04, 5.5753E+04,
     +  5.6164E+04, 5.6578E+04, 5.6995E+04, 5.7414E+04, 5.7837E+04,
     + 5.8262E+04, 5.8691E+04, 5.9123E+04, 5.9557E+04, 5.9995E+04,
     +  6.0435E+04, 6.0879E+04, 6.1326E+04, 6.1776E+04, 6.2228E+04,
     + 6.2685E+04, 6.3144E+04, 6.3607E+04, 6.4073E+04, 6.4542E+04,
     +  6.5014E+04, 6.5490E+04, 6.5968E+04, 6.6451E+04, 6.6937E+04,
     + 6.7426E+04, 6.7918E+04, 6.8414E+04, 6.8914E+04, 6.9416E+04,
     +  6.9923E+04, 7.0432E+04, 7.0946E+04, 7.1463E+04, 7.1984E+04,
     + 7.2508E+04, 7.3035E+04, 7.3567E+04, 7.4102E+04, 7.4641E+04,
     +  7.5184E+04, 7.5730E+04, 7.6280E+04, 7.6835E+04, 7.7392E+04,
     + 7.7954E+04, 7.8520E+04, 7.9089E+04, 7.9663E+04, 8.0240E+04,
     +  8.0822E+04, 8.1408E+04, 8.1997E+04, 8.2591E+04, 8.3189E+04,
     + 8.3791E+04, 8.4397E+04, 8.5007E+04, 8.5622E+04, 8.6241E+04,
     +  8.6864E+04, 8.7491E+04, 8.8123E+04, 8.8759E+04, 8.9399E+04,
     + 9.0044E+04, 9.0693E+04, 9.1347E+04, 9.2006E+04, 9.2669E+04,
     +  9.3336E+04, 9.4008E+04, 9.4685E+04, 9.5366E+04, 9.6052E+04,
     + 9.6743E+04, 9.7438E+04, 9.8139E+04, 9.8844E+04, 9.9554E+04,
     +  1.0027E+05, 1.0099E+05, 1.0171E+05, 1.0244E+05, 1.0318E+05,
     + 1.0392E+05, 1.0466E+05, 1.0541E+05, 1.0617E+05, 1.0693E+05,
     +  1.0769E+05, 1.0846E+05, 1.0924E+05, 1.1002E+05, 1.1081E+05,
     + 1.1160E+05, 1.1240E+05, 1.1320E+05, 1.1401E+05, 1.1482E+05,
     +  1.1564E+05, 1.1647E+05, 1.1730E+05, 1.1813E+05, 1.1898E+05,
     + 1.1982E+05, 1.2068E+05, 1.2154E+05, 1.2240E+05, 1.2327E+05,
     +  1.2415E+05, 1.2503E+05, 1.2592E+05, 1.2682E+05, 1.2772E+05,
     + 1.2863E+05, 1.2954E+05, 1.3046E+05, 1.3139E+05, 1.3232E+05,
     +  1.3326E+05, 1.3420E+05, 1.3515E+05, 1.3611E+05, 1.3707E+05,
     + 1.3804E+05, 1.3902E+05, 1.4001E+05, 1.4100E+05, 1.4199E+05,
     +  1.4300E+05, 1.4401E+05, 1.4503E+05, 1.4605E+05, 1.4708E+05,
     + 1.4812E+05, 1.4917E+05, 1.5022E+05, 1.5128E+05, 1.5235E+05,
     +  1.5342E+05, 1.5450E+05, 1.5559E+05, 1.5668E+05, 1.5779E+05,
     + 1.5890E+05, 1.6002E+05, 1.6114E+05, 1.6227E+05, 1.6341E+05,
     +  1.6456E+05, 1.6572E+05, 1.6688E+05, 1.6805E+05, 1.6923E+05,
     + 1.7042E+05, 1.7162E+05, 1.7282E+05, 1.7403E+05, 1.7525E+05,
     +  1.7648E+05, 1.7771E+05, 1.7896E+05, 1.8021E+05, 1.8147E+05,
     + 1.8274E+05, 1.8402E+05, 1.8530E+05, 1.8660E+05, 1.8790E+05,
     +  1.8921E+05, 1.9053E+05, 1.9186E+05, 1.9320E+05, 1.9455E+05,
     + 1.9590E+05, 1.9727E+05, 1.9864E+05, 2.0003E+05, 2.0142E+05,
     +  2.0282E+05, 2.0423E+05, 2.0565E+05, 2.0708E+05, 2.0852E+05,
     + 2.0997E+05, 2.1143E+05, 2.1289E+05, 2.1437E+05, 2.1586E+05,
     +  2.1736E+05, 2.1886E+05, 2.2038E+05, 2.2191E+05, 2.2344E+05,
     + 2.2499E+05, 2.2655E+05, 2.2812E+05, 2.2969E+05, 2.3128E+05,
     +  2.3288E+05, 2.3449E+05, 2.3611E+05, 2.3774E+05, 2.3938E+05,
     + 2.4103E+05, 2.4270E+05, 2.4437E+05, 2.4605E+05, 2.4775E+05,
     +  2.4946E+05, 2.5117E+05, 2.5290E+05, 2.5464E+05, 2.5639E+05,
     + 2.5816E+05, 2.5993E+05, 2.6172E+05, 2.6352E+05, 2.6533E+05,
     +  2.6715E+05, 2.6898E+05, 2.7083E+05, 2.7268E+05, 2.7455E+05,
     + 2.7643E+05, 2.7833E+05, 2.8023E+05, 2.8215E+05, 2.8408E+05,
     +  2.8603E+05, 2.8798E+05, 2.8995E+05, 2.9193E+05, 2.9393E+05,
     + 2.9594E+05, 2.9796E+05, 2.9999E+05, 3.0204E+05, 3.0409E+05,
     +  3.0617E+05, 3.0825E+05, 3.1035E+05, 3.1247E+05, 3.1459E+05,
     + 3.1674E+05, 3.1889E+05, 3.2106E+05, 3.2324E+05, 3.2544E+05,
     +  3.2765E+05, 3.2987E+05, 3.3211E+05, 3.3436E+05, 3.3663E+05,
     + 3.3891E+05, 3.4121E+05, 3.4352E+05, 3.4585E+05, 3.4819E+05,
     +  3.5055E+05, 3.5292E+05, 3.5531E+05, 3.5771E+05, 3.6013E+05,
     + 3.6256E+05, 3.6501E+05, 3.6747E+05, 3.6995E+05, 3.7245E+05,
     +  3.7496E+05, 3.7749E+05, 3.8003E+05, 3.8259E+05, 3.8517E+05,
     + 3.8776E+05, 3.9037E+05, 3.9299E+05, 3.9564E+05, 3.9829E+05,
     +  4.0097E+05, 4.0366E+05, 4.0637E+05, 4.0910E+05, 4.1184E+05,
     + 4.1461E+05, 4.1739E+05, 4.2018E+05, 4.2300E+05, 4.2583E+05,
     +  4.2868E+05, 4.3155E+05, 4.3443E+05, 4.3734E+05, 4.4026E+05,
     + 4.4320E+05, 4.4616E+05, 4.4914E+05, 4.5214E+05, 4.5515E+05,
     +  4.5819E+05, 4.6124E+05, 4.6432E+05, 4.6741E+05, 4.7052E+05,
     + 4.7365E+05, 4.7680E+05, 4.7997E+05, 4.8316E+05, 4.8637E+05,
     +  4.8961E+05, 4.9286E+05, 4.9613E+05, 4.9942E+05, 5.0273E+05,
     + 5.0607E+05, 5.0942E+05, 5.1279E+05, 5.1619E+05, 5.1961E+05,
     +  5.2305E+05, 5.2651E+05, 5.2999E+05, 5.3349E+05, 5.3701E+05,
     + 5.4056E+05, 5.4413E+05, 5.4772E+05, 5.5133E+05, 5.5497E+05,
     +  5.5863E+05, 5.6231E+05, 5.6601E+05, 5.6974E+05, 5.7349E+05,
     + 5.7726E+05, 5.8106E+05, 5.8488E+05, 5.8872E+05, 5.9259E+05,
     +  5.9648E+05, 6.0039E+05, 6.0433E+05, 6.0829E+05, 6.1228E+05,
     + 6.1629E+05, 6.2033E+05, 6.2439E+05, 6.2848E+05, 6.3259E+05,
     +  6.3673E+05, 6.4089E+05, 6.4508E+05, 6.4929E+05, 6.5353E+05,
     + 6.5780E+05, 6.6209E+05, 6.6641E+05, 6.7075E+05, 6.7512E+05,
     +  6.7952E+05, 6.8395E+05, 6.8840E+05, 6.9288E+05, 6.9738E+05,
     + 7.0192E+05, 7.0648E+05, 7.1107E+05, 7.1568E+05, 7.2033E+05,
     +  7.2500E+05, 7.2971E+05, 7.3444E+05, 7.3920E+05, 7.4398E+05,
     + 7.4880E+05, 7.5365E+05, 7.5852E+05, 7.6343E+05, 7.6836E+05,
     +  7.7333E+05, 7.7832E+05, 7.8335E+05, 7.8841E+05, 7.9349E+05,
     + 7.9861E+05, 8.0376E+05, 8.0894E+05, 8.1415E+05, 8.1939E+05,
     +  8.2466E+05, 8.2997E+05, 8.3531E+05, 8.4067E+05, 8.4608E+05,
     + 8.5151E+05, 8.5698E+05, 8.6248E+05, 8.6801E+05, 8.7357E+05,
     +  8.7917E+05, 8.8481E+05, 8.9047E+05, 8.9617E+05, 9.0191E+05,
     + 9.0768E+05, 9.1348E+05, 9.1932E+05, 9.2519E+05, 9.3110E+05,
     +  9.3705E+05, 9.4302E+05, 9.4904E+05, 9.5509E+05, 9.6118E+05,
     + 9.6730E+05, 9.7346E+05, 9.7966E+05, 9.8589E+05, 9.9216E+05,
     +  9.9847E+05, 1.0048E+06, 1.0112E+06, 1.0176E+06, 1.0241E+06,
     + 1.0306E+06, 1.0371E+06, 1.0437E+06, 1.0503E+06, 1.0569E+06,
     +  1.0636E+06, 1.0704E+06, 1.0771E+06, 1.0839E+06, 1.0908E+06,
     + 1.0977E+06, 1.1046E+06, 1.1116E+06, 1.1186E+06, 1.1257E+06,
     +  1.1328E+06, 1.1399E+06, 1.1471E+06, 1.1543E+06, 1.1616E+06,
     + 1.1689E+06, 1.1762E+06, 1.1836E+06, 1.1911E+06, 1.1985E+06,
     +  1.2061E+06, 1.2136E+06, 1.2212E+06, 1.2289E+06, 1.2366E+06,
     + 1.2444E+06, 1.2521E+06, 1.2600E+06, 1.2679E+06, 1.2758E+06,
     +  1.2838E+06, 1.2918E+06, 1.2999E+06, 1.3080E+06, 1.3161E+06,
     + 1.3244E+06, 1.3326E+06, 1.3409E+06, 1.3493E+06, 1.3577E+06,
     +  1.3661E+06, 1.3746E+06, 1.3832E+06, 1.3918E+06, 1.4004E+06,
     + 1.4091E+06, 1.4179E+06, 1.4267E+06, 1.4355E+06, 1.4444E+06,
     +  1.4534E+06, 1.4624E+06, 1.4715E+06, 1.4806E+06, 1.4897E+06,
     + 1.4990E+06, 1.5082E+06, 1.5176E+06, 1.5269E+06, 1.5364E+06,
     +  1.5458E+06, 1.5554E+06, 1.5650E+06, 1.5746E+06, 1.5843E+06,
     + 1.5941E+06, 1.6039E+06, 1.6138E+06, 1.6237E+06, 1.6337E+06,
     +  1.6437E+06, 1.6538E+06, 1.6640E+06, 1.6742E+06, 1.6845E+06,
     + 1.6948E+06, 1.7052E+06, 1.7157E+06, 1.7262E+06, 1.7367E+06,
     +  1.7474E+06, 1.7581E+06, 1.7688E+06, 1.7796E+06, 1.7905E+06,
     + 1.8014E+06, 1.8124E+06, 1.8235E+06, 1.8346E+06, 1.8458E+06,
     +  1.8570E+06, 1.8684E+06, 1.8797E+06, 1.8912E+06, 1.9027E+06,
     + 1.9143E+06, 1.9259E+06, 1.9376E+06, 1.9494E+06, 1.9612E+06,
     +  1.9731E+06, 1.9851E+06, 1.9971E+06, 2.0092E+06, 2.0214E+06,
     + 2.0336E+06, 2.0459E+06, 2.0583E+06, 2.0708E+06, 2.0833E+06,
     +  2.0959E+06, 2.1085E+06, 2.1213E+06, 2.1341E+06, 2.1469E+06,
     + 2.1599E+06, 2.1729E+06, 2.1860E+06, 2.1991E+06, 2.2124E+06,
     +  2.2257E+06, 2.2391E+06, 2.2525E+06, 2.2661E+06, 2.2797E+06,
     + 2.2934E+06, 2.3071E+06, 2.3210E+06, 2.3349E+06, 2.3489E+06,
     +  2.3630E+06, 2.3771E+06, 2.3913E+06, 2.4056E+06, 2.4200E+06,
     + 2.4345E+06, 2.4490E+06, 2.4637E+06, 2.4784E+06, 2.4932E+06,
     +  2.5080E+06, 2.5230E+06, 2.5380E+06, 2.5532E+06, 2.5684E+06,
     + 2.5836E+06, 2.5990E+06, 2.6145E+06, 2.6300E+06, 2.6456E+06,
     +  2.6614E+06, 2.6772E+06, 2.6930E+06, 2.7090E+06, 2.7251E+06,
     + 2.7412E+06, 2.7575E+06, 2.7738E+06, 2.7902E+06, 2.8067E+06,
     +  2.8233E+06, 2.8400E+06, 2.8568E+06, 2.8737E+06, 2.8906E+06,
     + 2.9077E+06, 2.9248E+06, 2.9421E+06, 2.9594E+06, 2.9768E+06,
     +  2.9944E+06, 3.0120E+06, 3.0297E+06, 3.0475E+06, 3.0654E+06,
     + 3.0835E+06, 3.1016E+06, 3.1198E+06, 3.1381E+06, 3.1565E+06,
     +  3.1750E+06, 3.1936E+06, 3.2123E+06, 3.2311E+06, 3.2500E+06,
     + 3.2690E+06, 3.2881E+06, 3.3074E+06, 3.3267E+06, 3.3461E+06,
     +  3.3656E+06, 3.3853E+06, 3.4050E+06, 3.4249E+06, 3.4448E+06,
     + 3.4649E+06, 3.4851E+06, 3.5053E+06, 3.5257E+06, 3.5462E+06,
     +  3.5668E+06, 3.5876E+06, 3.6084E+06, 3.6293E+06, 3.6504E+06,
     + 3.6716E+06, 3.6928E+06, 3.7142E+06, 3.7358E+06, 3.7574E+06,
     +  3.7791E+06, 3.8010E+06, 3.8230E+06, 3.8451E+06, 3.8673E+06,
     + 3.8896E+06, 3.9120E+06, 3.9346E+06, 3.9573E+06, 3.9801E+06,
     +  4.0030E+06, 4.0261E+06, 4.0493E+06, 4.0726E+06, 4.0960E+06,
     + 4.1195E+06, 4.1432E+06, 4.1670E+06, 4.1909E+06, 4.2150E+06,
     +  4.2392E+06, 4.2635E+06, 4.2879E+06, 4.3125E+06, 4.3372E+06,
     + 4.3620E+06, 4.3870E+06, 4.4121E+06, 4.4373E+06, 4.4626E+06,
     +  4.4881E+06, 4.5137E+06, 4.5395E+06, 4.5654E+06, 4.5914E+06,
     + 4.6176E+06, 4.6439E+06, 4.6703E+06, 4.6969E+06, 4.7236E+06,
     +  4.7505E+06, 4.7775E+06, 4.8047E+06, 4.8319E+06, 4.8594E+06,
     + 4.8870E+06, 4.9147E+06, 4.9425E+06, 4.9705E+06, 4.9987E+06,
     +  5.0270E+06, 5.0555E+06, 5.0841E+06, 5.1128E+06, 5.1417E+06,
     + 5.1708E+06, 5.2000E+06, 5.2293E+06, 5.2588E+06, 5.2885E+06,
     +  5.3183E+06, 5.3483E+06, 5.3784E+06, 5.4087E+06, 5.4391E+06,
     + 5.4697E+06, 5.5005E+06, 5.5314E+06, 5.5625E+06, 5.5937E+06,
     +  5.6251E+06, 5.6567E+06, 5.6884E+06, 5.7203E+06, 5.7523E+06,
     + 5.7845E+06, 5.8169E+06, 5.8495E+06, 5.8822E+06, 5.9151E+06,
     +  5.9482E+06, 5.9814E+06, 6.0148E+06, 6.0484E+06, 6.0821E+06,
     + 6.1160E+06, 6.1501E+06, 6.1844E+06, 6.2188E+06, 6.2534E+06,
     +  6.2882E+06, 6.3232E+06, 6.3584E+06, 6.3937E+06, 6.4292E+06,
     + 6.4649E+06, 6.5008E+06, 6.5369E+06, 6.5731E+06, 6.6096E+06,
     +  6.6462E+06, 6.6830E+06, 6.7200E+06, 6.7572E+06, 6.7946E+06,
     + 6.8321E+06, 6.8699E+06, 6.9078E+06, 6.9460E+06, 6.9843E+06,
     +  7.0228E+06, 7.0616E+06, 7.1005E+06, 7.1396E+06, 7.1789E+06,
     + 7.2184E+06, 7.2582E+06, 7.2981E+06, 7.3382E+06, 7.3785E+06,
     +  7.4191E+06, 7.4598E+06, 7.5007E+06, 7.5419E+06, 7.5832E+06,
     + 7.6248E+06, 7.6666E+06, 7.7086E+06, 7.7508E+06, 7.7932E+06,
     +  7.8358E+06, 7.8786E+06, 7.9217E+06, 7.9650E+06, 8.0084E+06,
     + 8.0521E+06, 8.0961E+06, 8.1402E+06, 8.1846E+06, 8.2292E+06,
     +  8.2740E+06, 8.3190E+06, 8.3643E+06, 8.4098E+06, 8.4555E+06,
     + 8.5014E+06, 8.5476E+06, 8.5940E+06, 8.6407E+06, 8.6875E+06,
     +  8.7346E+06, 8.7820E+06, 8.8296E+06, 8.8774E+06, 8.9254E+06,
     + 8.9737E+06, 9.0223E+06, 9.0710E+06, 9.1200E+06, 9.1693E+06,
     +  9.2188E+06, 9.2686E+06, 9.3186E+06, 9.3688E+06, 9.4193E+06,
     + 9.4700E+06, 9.5210E+06, 9.5723E+06, 9.6238E+06, 9.6756E+06,
     +  9.7276E+06, 9.7798E+06, 9.8324E+06, 9.8852E+06, 9.9382E+06,
     + 9.9915E+06, 1.0045E+07, 1.0099E+07, 1.0153E+07, 1.0207E+07,
     +  1.0262E+07, 1.0317E+07, 1.0372E+07, 1.0428E+07, 1.0483E+07,
     + 1.0539E+07, 1.0596E+07, 1.0652E+07, 1.0709E+07, 1.0766E+07,
     +  1.0823E+07, 1.0881E+07, 1.0939E+07, 1.0997E+07, 1.1056E+07,
     + 1.1115E+07, 1.1174E+07, 1.1233E+07, 1.1293E+07, 1.1353E+07,
     +  1.1413E+07, 1.1473E+07, 1.1534E+07, 1.1595E+07, 1.1657E+07,
     + 1.1719E+07, 1.1781E+07, 1.1843E+07, 1.1906E+07, 1.1969E+07,
     +  1.2032E+07, 1.2095E+07, 1.2159E+07, 1.2223E+07, 1.2288E+07,
     + 1.2353E+07, 1.2418E+07, 1.2483E+07, 1.2549E+07, 1.2615E+07,
     +  1.2681E+07, 1.2748E+07, 1.2815E+07, 1.2882E+07, 1.2950E+07,
     + 1.3018E+07, 1.3086E+07, 1.3155E+07, 1.3224E+07, 1.3293E+07,
     +  1.3363E+07, 1.3433E+07, 1.3503E+07, 1.3574E+07, 1.3645E+07,
     + 1.3716E+07, 1.3788E+07, 1.3860E+07, 1.3932E+07, 1.4005E+07,
     +  1.4078E+07, 1.4152E+07, 1.4225E+07, 1.4300E+07, 1.4374E+07,
     + 1.4449E+07, 1.4524E+07, 1.4600E+07, 1.4675E+07, 1.4752E+07,
     +  1.4828E+07, 1.4905E+07, 1.4983E+07, 1.5061E+07, 1.5139E+07,
     + 1.5217E+07, 1.5296E+07, 1.5375E+07, 1.5455E+07, 1.5535E+07,
     +  1.5615E+07, 1.5696E+07, 1.5777E+07, 1.5859E+07, 1.5940E+07,
     + 1.6023E+07, 1.6105E+07, 1.6188E+07, 1.6272E+07, 1.6356E+07,
     +  1.6440E+07, 1.6525E+07, 1.6610E+07, 1.6695E+07, 1.6781E+07,
     + 1.6867E+07, 1.6954E+07, 1.7041E+07, 1.7129E+07, 1.7216E+07,
     +  1.7305E+07, 1.7394E+07, 1.7483E+07, 1.7572E+07, 1.7662E+07,
     + 1.7753E+07, 1.7843E+07, 1.7935E+07, 1.8026E+07, 1.8119E+07,
     +  1.8211E+07, 1.8304E+07, 1.8398E+07, 1.8491E+07, 1.8586E+07,
     + 1.8680E+07, 1.8776E+07, 1.8871E+07, 1.8967E+07, 1.9064E+07,
     +  1.9161E+07, 1.9258E+07, 1.9356E+07, 1.9454E+07, 1.9553E+07,
     + 1.9652E+07, 1.9752E+07, 1.9852E+07, 1.9953E+07, 2.0054E+07,
     +  2.0156E+07, 2.0258E+07, 2.0360E+07, 2.0463E+07, 2.0567E+07,
     + 2.0671E+07, 2.0775E+07, 2.0880E+07, 2.0986E+07, 2.1091E+07,
     +  2.1198E+07, 2.1305E+07, 2.1412E+07, 2.1520E+07, 2.1628E+07,
     + 2.1737E+07, 2.1847E+07, 2.1956E+07, 2.2067E+07, 2.2178E+07,
     +  2.2289E+07, 2.2401E+07, 2.2513E+07, 2.2626E+07, 2.2740E+07,
     + 2.2854E+07, 2.2968E+07, 2.3083E+07, 2.3199E+07, 2.3315E+07,
     +  2.3432E+07, 2.3549E+07, 2.3667E+07, 2.3785E+07, 2.3904E+07,
     + 2.4023E+07, 2.4143E+07, 2.4263E+07, 2.4384E+07, 2.4506E+07,
     +  2.4628E+07, 2.4750E+07, 2.4874E+07, 2.4997E+07, 2.5122E+07,
     + 2.5247E+07, 2.5372E+07, 2.5498E+07, 2.5625E+07, 2.5752E+07,
     +  2.5880E+07, 2.6008E+07, 2.6137E+07, 2.6266E+07, 2.6396E+07,
     + 2.6527E+07, 2.6658E+07, 2.6790E+07, 2.6923E+07, 2.7056E+07,
     +  2.7189E+07, 2.7324E+07, 2.7459E+07, 2.7594E+07, 2.7730E+07,
     + 2.7867E+07, 2.8004E+07, 2.8142E+07, 2.8281E+07, 2.8420E+07,
     +  2.8560E+07, 2.8700E+07, 2.8841E+07, 2.8983E+07, 2.9125E+07,
     + 2.9268E+07, 2.9412E+07, 2.9556E+07, 2.9701E+07, 2.9847E+07,
     +  2.9993E+07, 3.0140E+07, 3.0288E+07, 3.0436E+07, 3.0585E+07,
     + 3.0734E+07, 3.0884E+07, 3.1035E+07, 3.1187E+07, 3.1339E+07,
     +  3.1492E+07, 3.1646E+07, 3.1800E+07, 3.1955E+07, 3.2111E+07,
     + 3.2267E+07, 3.2424E+07, 3.2582E+07, 3.2740E+07, 3.2899E+07,
     +  3.3059E+07, 3.3220E+07, 3.3381E+07, 3.3543E+07, 3.3706E+07,
     + 3.3870E+07, 3.4034E+07, 3.4199E+07, 3.4364E+07, 3.4531E+07,
     +  3.4698E+07, 3.4866E+07, 3.5034E+07, 3.5204E+07, 3.5374E+07,
     + 3.5545E+07, 3.5716E+07, 3.5889E+07, 3.6062E+07, 3.6236E+07,
     +  3.6410E+07, 3.6586E+07, 3.6762E+07, 3.6939E+07, 3.7117E+07,
     + 3.7295E+07, 3.7474E+07, 3.7655E+07, 3.7836E+07, 3.8017E+07,
     +  3.8200E+07, 3.8383E+07, 3.8567E+07, 3.8752E+07, 3.8938E+07,
     + 3.9124E+07, 3.9312E+07, 3.9500E+07, 3.9689E+07, 3.9879E+07,
     +  4.0069E+07, 4.0261E+07, 4.0453E+07, 4.0646E+07, 4.0840E+07,
     + 4.1035E+07, 4.1231E+07, 4.1428E+07, 4.1625E+07, 4.1823E+07,
     +  4.2022E+07, 4.2223E+07, 4.2423E+07, 4.2625E+07, 4.2828E+07,
     + 4.3031E+07, 4.3236E+07, 4.3441E+07, 4.3647E+07, 4.3854E+07,
     +  4.4062E+07, 4.4271E+07, 4.4481E+07, 4.4692E+07, 4.4903E+07,
     + 4.5116E+07, 4.5329E+07, 4.5544E+07, 4.5759E+07, 4.5975E+07,
     +  4.6192E+07, 4.6410E+07, 4.6630E+07, 4.6849E+07, 4.7070E+07,
     + 4.7292E+07, 4.7515E+07, 4.7739E+07, 4.7964E+07, 4.8189E+07,
     +  4.8416E+07, 4.8644E+07, 4.8873E+07, 4.9102E+07, 4.9333E+07,
     + 4.9564E+07, 4.9797E+07, 5.0031E+07, 5.0265E+07, 5.0501E+07,
     +  5.0738E+07, 5.0975E+07, 5.1214E+07, 5.1454E+07, 5.1694E+07,
     + 5.1936E+07, 5.2179E+07, 5.2423E+07, 5.2668E+07, 5.2913E+07,
     +  5.3160E+07, 5.3408E+07, 5.3657E+07, 5.3908E+07, 5.4159E+07,
     + 5.4411E+07, 5.4664E+07, 5.4919E+07, 5.5174E+07, 5.5431E+07,
     +  5.5688E+07, 5.5947E+07, 5.6207E+07, 5.6468E+07, 5.6730E+07,
     + 5.6993E+07, 5.7258E+07, 5.7523E+07, 5.7790E+07, 5.8057E+07,
     +  5.8326E+07, 5.8596E+07, 5.8867E+07, 5.9139E+07, 5.9413E+07,
     + 5.9687E+07, 5.9963E+07, 6.0240E+07, 6.0518E+07, 6.0797E+07,
     +  6.1077E+07, 6.1359E+07, 6.1642E+07, 6.1926E+07, 6.2211E+07,
     + 6.2497E+07, 6.2785E+07, 6.3073E+07, 6.3363E+07, 6.3655E+07,
     +  6.3947E+07, 6.4241E+07, 6.4535E+07, 6.4832E+07, 6.5129E+07,
     + 6.5427E+07, 6.5727E+07, 6.6028E+07, 6.6331E+07, 6.6634E+07,
     +  6.6939E+07, 6.7245E+07, 6.7553E+07, 6.7862E+07, 6.8172E+07,
     + 6.8483E+07, 6.8795E+07, 6.9109E+07, 6.9425E+07, 6.9741E+07,
     +  7.0059E+07, 7.0378E+07, 7.0699E+07, 7.1020E+07, 7.1344E+07,
     + 7.1668E+07, 7.1994E+07, 7.2321E+07, 7.2650E+07, 7.2980E+07,
     +  7.3311E+07, 7.3644E+07, 7.3978E+07, 7.4313E+07, 7.4650E+07,
     + 7.4988E+07, 7.5328E+07, 7.5669E+07, 7.6011E+07, 7.6355E+07,
     +  7.6700E+07, 7.7047E+07, 7.7395E+07, 7.7745E+07, 7.8096E+07,
     + 7.8448E+07, 7.8802E+07, 7.9157E+07, 7.9514E+07, 7.9873E+07,
     +  8.0232E+07, 8.0594E+07, 8.0956E+07, 8.1321E+07, 8.1686E+07,
     + 8.2053E+07, 8.2422E+07, 8.2792E+07, 8.3164E+07, 8.3537E+07,
     +  8.3912E+07, 8.4289E+07, 8.4667E+07, 8.5046E+07, 8.5427E+07,
     + 8.5810E+07, 8.6194E+07, 8.6579E+07, 8.6967E+07, 8.7355E+07,
     +  8.7746E+07, 8.8138E+07, 8.8531E+07, 8.8927E+07, 8.9324E+07,
     + 8.9722E+07, 9.0122E+07, 9.0524E+07, 9.0927E+07, 9.1332E+07,
     +  9.1739E+07, 9.2147E+07, 9.2557E+07, 9.2969E+07, 9.3382E+07,
     + 9.3797E+07, 9.4213E+07, 9.4632E+07, 9.5052E+07, 9.5473E+07,
     +  9.5897E+07, 9.6322E+07, 9.6749E+07, 9.7177E+07, 9.7608E+07,
     + 9.8040E+07, 9.8474E+07, 9.8909E+07, 9.9346E+07, 9.9786E+07,
     +  1.0023E+08, 1.0067E+08, 1.0111E+08, 1.0156E+08, 1.0201E+08,
     + 1.0246E+08, 1.0291E+08, 1.0336E+08, 1.0382E+08, 1.0427E+08,
     +  1.0473E+08, 1.0519E+08, 1.0566E+08, 1.0612E+08, 1.0659E+08,
     + 1.0706E+08, 1.0753E+08, 1.0800E+08, 1.0847E+08, 1.0895E+08,
     +  1.0943E+08, 1.0991E+08, 1.1039E+08, 1.1087E+08, 1.1136E+08,
     + 1.1184E+08, 1.1233E+08, 1.1282E+08, 1.1332E+08, 1.1381E+08,
     +  1.1431E+08, 1.1481E+08, 1.1531E+08, 1.1581E+08, 1.1632E+08,
     + 1.1682E+08, 1.1733E+08, 1.1784E+08, 1.1836E+08, 1.1887E+08,
     +  1.1939E+08, 1.1991E+08, 1.2043E+08, 1.2095E+08, 1.2148E+08,
     + 1.2201E+08, 1.2254E+08, 1.2307E+08, 1.2360E+08, 1.2414E+08,
     +  1.2468E+08, 1.2522E+08, 1.2576E+08, 1.2630E+08, 1.2685E+08,
     + 1.2740E+08, 1.2795E+08, 1.2850E+08, 1.2906E+08, 1.2961E+08,
     +  1.3017E+08, 1.3074E+08, 1.3130E+08, 1.3187E+08, 1.3244E+08,
     + 1.3301E+08, 1.3358E+08, 1.3415E+08, 1.3473E+08, 1.3531E+08,
     +  1.3589E+08, 1.3648E+08, 1.3706E+08, 1.3765E+08, 1.3825E+08,
     + 1.3884E+08, 1.3943E+08, 1.4003E+08, 1.4063E+08, 1.4124E+08,
     +  1.4184E+08, 1.4245E+08, 1.4306E+08, 1.4367E+08, 1.4429E+08,
     + 1.4490E+08, 1.4552E+08, 1.4615E+08, 1.4677E+08, 1.4740E+08,
     +  1.4803E+08, 1.4866E+08, 1.4929E+08, 1.4993E+08, 1.5057E+08,
     + 1.5121E+08, 1.5185E+08, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 4,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_O2    (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 6),QofT( 6,NT),Q(NT)
      data (Tmx(I),I=1,6)/7500.,7500.,7500.,7500.,7500.,7500./
      data xgj/ 1.,1.,1.,1.,1.,1./
c...       O2
c...        --        66
      data (QofT( 1,J),J=1,2408)/ 1.2593E+00, 2.2900E+00, 3.2955E+00,
     +  4.1330E+00, 4.8674E+00, 5.5593E+00, 6.2404E+00, 6.9233E+00,
     + 7.6116E+00, 8.3058E+00, 9.0049E+00, 9.7081E+00, 1.0415E+01,
     +  1.1124E+01, 1.1835E+01, 1.2548E+01, 1.3262E+01, 1.3978E+01,
     + 1.4694E+01, 1.5412E+01, 1.6849E+01, 1.8288E+01, 1.9729E+01,
     +  2.1171E+01, 2.2615E+01, 2.4059E+01, 2.5504E+01, 2.6950E+01,
     + 2.8396E+01, 2.9843E+01, 3.1290E+01, 3.2737E+01, 3.4185E+01,
     +  3.5633E+01, 3.7082E+01, 3.8530E+01, 3.9979E+01, 4.1428E+01,
     + 4.2877E+01, 4.4326E+01, 4.5775E+01, 4.7224E+01, 4.8674E+01,
     +  5.0123E+01, 5.1573E+01, 5.3023E+01, 5.4473E+01, 5.5923E+01,
     + 5.7373E+01, 5.8823E+01, 6.0273E+01, 6.1723E+01, 6.3174E+01,
     +  6.4624E+01, 6.6074E+01, 6.7525E+01, 6.8975E+01, 7.0426E+01,
     + 7.1877E+01, 7.3327E+01, 7.4778E+01, 7.6229E+01, 7.7679E+01,
     +  7.9130E+01, 8.0581E+01, 8.2032E+01, 8.3483E+01, 8.4934E+01,
     + 8.6385E+01, 8.7836E+01, 8.9287E+01, 9.0738E+01, 9.2189E+01,
     +  9.3640E+01, 9.5091E+01, 9.6543E+01, 9.7994E+01, 9.9445E+01,
     + 1.0090E+02, 1.0235E+02, 1.0380E+02, 1.0525E+02, 1.0670E+02,
     +  1.0815E+02, 1.0960E+02, 1.1106E+02, 1.1251E+02, 1.1396E+02,
     + 1.1541E+02, 1.1686E+02, 1.1831E+02, 1.1977E+02, 1.2122E+02,
     +  1.2267E+02, 1.2412E+02, 1.2557E+02, 1.2702E+02, 1.2848E+02,
     + 1.2993E+02, 1.3138E+02, 1.3283E+02, 1.3428E+02, 1.3574E+02,
     +  1.3719E+02, 1.3864E+02, 1.4009E+02, 1.4154E+02, 1.4300E+02,
     + 1.4445E+02, 1.4590E+02, 1.4735E+02, 1.4881E+02, 1.5026E+02,
     +  1.5171E+02, 1.5316E+02, 1.5462E+02, 1.5607E+02, 1.5752E+02,
     + 1.5897E+02, 1.6043E+02, 1.6188E+02, 1.6333E+02, 1.6479E+02,
     +  1.6624E+02, 1.6769E+02, 1.6915E+02, 1.7060E+02, 1.7205E+02,
     + 1.7351E+02, 1.7496E+02, 1.7641E+02, 1.7787E+02, 1.7932E+02,
     +  1.8078E+02, 1.8223E+02, 1.8369E+02, 1.8514E+02, 1.8660E+02,
     + 1.8805E+02, 1.8951E+02, 1.9096E+02, 1.9242E+02, 1.9387E+02,
     +  1.9533E+02, 1.9678E+02, 1.9824E+02, 1.9970E+02, 2.0115E+02,
     + 2.0261E+02, 2.0407E+02, 2.0553E+02, 2.0698E+02, 2.0844E+02,
     +  2.0990E+02, 2.1136E+02, 2.1282E+02, 2.1428E+02, 2.1573E+02,
     + 2.1719E+02, 2.1865E+02, 2.2011E+02, 2.2157E+02, 2.2304E+02,
     +  2.2450E+02, 2.2596E+02, 2.2742E+02, 2.2888E+02, 2.3035E+02,
     + 2.3181E+02, 2.3327E+02, 2.3474E+02, 2.3620E+02, 2.3767E+02,
     +  2.3913E+02, 2.4060E+02, 2.4206E+02, 2.4353E+02, 2.4500E+02,
     + 2.4646E+02, 2.4793E+02, 2.4940E+02, 2.5087E+02, 2.5234E+02,
     +  2.5381E+02, 2.5528E+02, 2.5675E+02, 2.5822E+02, 2.5970E+02,
     + 2.6117E+02, 2.6264E+02, 2.6412E+02, 2.6559E+02, 2.6707E+02,
     +  2.6854E+02, 2.7002E+02, 2.7150E+02, 2.7298E+02, 2.7445E+02,
     + 2.7593E+02, 2.7741E+02, 2.7889E+02, 2.8038E+02, 2.8186E+02,
     +  2.8334E+02, 2.8482E+02, 2.8631E+02, 2.8779E+02, 2.8928E+02,
     + 2.9077E+02, 2.9225E+02, 2.9374E+02, 2.9523E+02, 2.9672E+02,
     +  2.9821E+02, 2.9970E+02, 3.0120E+02, 3.0269E+02, 3.0418E+02,
     + 3.0568E+02, 3.0717E+02, 3.0867E+02, 3.1017E+02, 3.1167E+02,
     +  3.1317E+02, 3.1467E+02, 3.1617E+02, 3.1767E+02, 3.1917E+02,
     + 3.2068E+02, 3.2218E+02, 3.2369E+02, 3.2519E+02, 3.2670E+02,
     +  3.2821E+02, 3.2972E+02, 3.3123E+02, 3.3274E+02, 3.3426E+02,
     + 3.3577E+02, 3.3729E+02, 3.3880E+02, 3.4032E+02, 3.4184E+02,
     +  3.4335E+02, 3.4487E+02, 3.4640E+02, 3.4792E+02, 3.4944E+02,
     + 3.5097E+02, 3.5249E+02, 3.5402E+02, 3.5555E+02, 3.5707E+02,
     +  3.5860E+02, 3.6013E+02, 3.6167E+02, 3.6320E+02, 3.6473E+02,
     + 3.6627E+02, 3.6781E+02, 3.6934E+02, 3.7088E+02, 3.7242E+02,
     +  3.7396E+02, 3.7551E+02, 3.7705E+02, 3.7859E+02, 3.8014E+02,
     + 3.8169E+02, 3.8324E+02, 3.8479E+02, 3.8634E+02, 3.8789E+02,
     +  3.8944E+02, 3.9099E+02, 3.9255E+02, 3.9411E+02, 3.9566E+02,
     + 3.9722E+02, 3.9878E+02, 4.0035E+02, 4.0191E+02, 4.0347E+02,
     +  4.0504E+02, 4.0660E+02, 4.0817E+02, 4.0974E+02, 4.1131E+02,
     + 4.1288E+02, 4.1446E+02, 4.1603E+02, 4.1760E+02, 4.1918E+02,
     +  4.2076E+02, 4.2234E+02, 4.2392E+02, 4.2550E+02, 4.2708E+02,
     + 4.2867E+02, 4.3025E+02, 4.3184E+02, 4.3343E+02, 4.3502E+02,
     +  4.3661E+02, 4.3820E+02, 4.3979E+02, 4.4138E+02, 4.4298E+02,
     + 4.4458E+02, 4.4618E+02, 4.4778E+02, 4.4938E+02, 4.5098E+02,
     +  4.5258E+02, 4.5419E+02, 4.5579E+02, 4.5740E+02, 4.5901E+02,
     + 4.6062E+02, 4.6223E+02, 4.6384E+02, 4.6546E+02, 4.6707E+02,
     +  4.6869E+02, 4.7030E+02, 4.7192E+02, 4.7354E+02, 4.7517E+02,
     + 4.7679E+02, 4.7841E+02, 4.8004E+02, 4.8167E+02, 4.8329E+02,
     +  4.8492E+02, 4.8655E+02, 4.8819E+02, 4.8982E+02, 4.9145E+02,
     + 4.9309E+02, 4.9473E+02, 4.9637E+02, 4.9801E+02, 4.9965E+02,
     +  5.0129E+02, 5.0293E+02, 5.0458E+02, 5.0623E+02, 5.0787E+02,
     + 5.0952E+02, 5.1117E+02, 5.1282E+02, 5.1448E+02, 5.1613E+02,
     +  5.1779E+02, 5.1944E+02, 5.2110E+02, 5.2276E+02, 5.2442E+02,
     + 5.2609E+02, 5.2775E+02, 5.2941E+02, 5.3108E+02, 5.3275E+02,
     +  5.3442E+02, 5.3609E+02, 5.3776E+02, 5.3943E+02, 5.4110E+02,
     + 5.4278E+02, 5.4445E+02, 5.4613E+02, 5.4781E+02, 5.4949E+02,
     +  5.5117E+02, 5.5285E+02, 5.5454E+02, 5.5622E+02, 5.5791E+02,
     + 5.5960E+02, 5.6129E+02, 5.6298E+02, 5.6467E+02, 5.6636E+02,
     +  5.6806E+02, 5.6975E+02, 5.7145E+02, 5.7315E+02, 5.7485E+02,
     + 5.7655E+02, 5.7825E+02, 5.7995E+02, 5.8166E+02, 5.8336E+02,
     +  5.8507E+02, 5.8678E+02, 5.8849E+02, 5.9020E+02, 5.9191E+02,
     + 5.9362E+02, 5.9533E+02, 5.9705E+02, 5.9877E+02, 6.0049E+02,
     +  6.0220E+02, 6.0392E+02, 6.0565E+02, 6.0737E+02, 6.0909E+02,
     + 6.1082E+02, 6.1255E+02, 6.1427E+02, 6.1600E+02, 6.1773E+02,
     +  6.1946E+02, 6.2120E+02, 6.2293E+02, 6.2467E+02, 6.2640E+02,
     + 6.2814E+02, 6.2988E+02, 6.3162E+02, 6.3336E+02, 6.3510E+02,
     +  6.3684E+02, 6.3859E+02, 6.4033E+02, 6.4208E+02, 6.4383E+02,
     + 6.4558E+02, 6.4733E+02, 6.4908E+02, 6.5083E+02, 6.5259E+02,
     +  6.5434E+02, 6.5610E+02, 6.5786E+02, 6.5962E+02, 6.6138E+02,
     + 6.6314E+02, 6.6490E+02, 6.6666E+02, 6.6843E+02, 6.7019E+02,
     +  6.7196E+02, 6.7373E+02, 6.7550E+02, 6.7727E+02, 6.7904E+02,
     + 6.8081E+02, 6.8259E+02, 6.8436E+02, 6.8614E+02, 6.8791E+02,
     +  6.8969E+02, 6.9147E+02, 6.9325E+02, 6.9503E+02, 6.9682E+02,
     + 6.9860E+02, 7.0039E+02, 7.0217E+02, 7.0396E+02, 7.0575E+02,
     +  7.0754E+02, 7.0933E+02, 7.1112E+02, 7.1291E+02, 7.1470E+02,
     + 7.1650E+02, 7.1830E+02, 7.2009E+02, 7.2189E+02, 7.2369E+02,
     +  7.2549E+02, 7.2729E+02, 7.2909E+02, 7.3090E+02, 7.3270E+02,
     + 7.3451E+02, 7.3631E+02, 7.3812E+02, 7.3993E+02, 7.4174E+02,
     +  7.4355E+02, 7.4536E+02, 7.4718E+02, 7.4899E+02, 7.5081E+02,
     + 7.5262E+02, 7.5444E+02, 7.5626E+02, 7.5808E+02, 7.5990E+02,
     +  7.6172E+02, 7.6354E+02, 7.6537E+02, 7.6719E+02, 7.6902E+02,
     + 7.7084E+02, 7.7267E+02, 7.7450E+02, 7.7633E+02, 7.7816E+02,
     +  7.7999E+02, 7.8182E+02, 7.8366E+02, 7.8549E+02, 7.8733E+02,
     + 7.8917E+02, 7.9100E+02, 7.9284E+02, 7.9468E+02, 7.9652E+02,
     +  7.9836E+02, 8.0021E+02, 8.0205E+02, 8.0390E+02, 8.0574E+02,
     + 8.0759E+02, 8.0944E+02, 8.1128E+02, 8.1313E+02, 8.1498E+02,
     +  8.1684E+02, 8.1869E+02, 8.2054E+02, 8.2240E+02, 8.2425E+02,
     + 8.2611E+02, 8.2796E+02, 8.2982E+02, 8.3168E+02, 8.3354E+02,
     +  8.3540E+02, 8.3727E+02, 8.3913E+02, 8.4099E+02, 8.4286E+02,
     + 8.4472E+02, 8.4659E+02, 8.4846E+02, 8.5033E+02, 8.5220E+02,
     +  8.5407E+02, 8.5594E+02, 8.5781E+02, 8.5968E+02, 8.6156E+02,
     + 8.6343E+02, 8.6531E+02, 8.6718E+02, 8.6906E+02, 8.7094E+02,
     +  8.7282E+02, 8.7470E+02, 8.7658E+02, 8.7846E+02, 8.8035E+02,
     + 8.8223E+02, 8.8412E+02, 8.8600E+02, 8.8789E+02, 8.8978E+02,
     +  8.9167E+02, 8.9356E+02, 8.9545E+02, 8.9734E+02, 8.9923E+02,
     + 9.0112E+02, 9.0302E+02, 9.0491E+02, 9.0681E+02, 9.0870E+02,
     +  9.1060E+02, 9.1250E+02, 9.1440E+02, 9.1630E+02, 9.1820E+02,
     + 9.2010E+02, 9.2200E+02, 9.2390E+02, 9.2581E+02, 9.2771E+02,
     +  9.2962E+02, 9.3153E+02, 9.3343E+02, 9.3534E+02, 9.3725E+02,
     + 9.3916E+02, 9.4107E+02, 9.4298E+02, 9.4490E+02, 9.4681E+02,
     +  9.4872E+02, 9.5064E+02, 9.5255E+02, 9.5447E+02, 9.5639E+02,
     + 9.5831E+02, 9.6023E+02, 9.6215E+02, 9.6407E+02, 9.6599E+02,
     +  9.6791E+02, 9.6983E+02, 9.7176E+02, 9.7368E+02, 9.7561E+02,
     + 9.7753E+02, 9.7946E+02, 9.8139E+02, 9.8332E+02, 9.8525E+02,
     +  9.8718E+02, 9.8911E+02, 9.9104E+02, 9.9297E+02, 9.9491E+02,
     + 9.9684E+02, 9.9878E+02, 1.0007E+03, 1.0026E+03, 1.0046E+03,
     +  1.0065E+03, 1.0085E+03, 1.0104E+03, 1.0123E+03, 1.0143E+03,
     + 1.0162E+03, 1.0182E+03, 1.0201E+03, 1.0221E+03, 1.0240E+03,
     +  1.0259E+03, 1.0279E+03, 1.0298E+03, 1.0318E+03, 1.0337E+03,
     + 1.0357E+03, 1.0376E+03, 1.0396E+03, 1.0415E+03, 1.0435E+03,
     +  1.0454E+03, 1.0474E+03, 1.0494E+03, 1.0513E+03, 1.0533E+03,
     + 1.0552E+03, 1.0572E+03, 1.0591E+03, 1.0611E+03, 1.0631E+03,
     +  1.0650E+03, 1.0670E+03, 1.0689E+03, 1.0709E+03, 1.0729E+03,
     + 1.0748E+03, 1.0768E+03, 1.0788E+03, 1.0807E+03, 1.0827E+03,
     +  1.0847E+03, 1.0866E+03, 1.0886E+03, 1.0906E+03, 1.0925E+03,
     + 1.0945E+03, 1.0965E+03, 1.0985E+03, 1.1004E+03, 1.1024E+03,
     +  1.1044E+03, 1.1064E+03, 1.1083E+03, 1.1103E+03, 1.1123E+03,
     + 1.1143E+03, 1.1162E+03, 1.1182E+03, 1.1202E+03, 1.1222E+03,
     +  1.1242E+03, 1.1262E+03, 1.1281E+03, 1.1301E+03, 1.1321E+03,
     + 1.1341E+03, 1.1361E+03, 1.1381E+03, 1.1400E+03, 1.1420E+03,
     +  1.1440E+03, 1.1460E+03, 1.1480E+03, 1.1500E+03, 1.1520E+03,
     + 1.1540E+03, 1.1560E+03, 1.1580E+03, 1.1600E+03, 1.1620E+03,
     +  1.1639E+03, 1.1659E+03, 1.1679E+03, 1.1699E+03, 1.1719E+03,
     + 1.1739E+03, 1.1759E+03, 1.1779E+03, 1.1799E+03, 1.1819E+03,
     +  1.1839E+03, 1.1859E+03, 1.1879E+03, 1.1899E+03, 1.1920E+03,
     + 1.1940E+03, 1.1960E+03, 1.1980E+03, 1.2000E+03, 1.2020E+03,
     +  1.2040E+03, 1.2060E+03, 1.2080E+03, 1.2100E+03, 1.2120E+03,
     + 1.2140E+03, 1.2161E+03, 1.2181E+03, 1.2201E+03, 1.2221E+03,
     +  1.2241E+03, 1.2261E+03, 1.2281E+03, 1.2302E+03, 1.2322E+03,
     + 1.2342E+03, 1.2362E+03, 1.2382E+03, 1.2402E+03, 1.2423E+03,
     +  1.2443E+03, 1.2463E+03, 1.2483E+03, 1.2504E+03, 1.2524E+03,
     + 1.2544E+03, 1.2564E+03, 1.2584E+03, 1.2605E+03, 1.2625E+03,
     +  1.2645E+03, 1.2666E+03, 1.2686E+03, 1.2706E+03, 1.2726E+03,
     + 1.2747E+03, 1.2767E+03, 1.2787E+03, 1.2808E+03, 1.2828E+03,
     +  1.2848E+03, 1.2869E+03, 1.2889E+03, 1.2909E+03, 1.2930E+03,
     + 1.2950E+03, 1.2970E+03, 1.2991E+03, 1.3011E+03, 1.3032E+03,
     +  1.3052E+03, 1.3072E+03, 1.3093E+03, 1.3113E+03, 1.3134E+03,
     + 1.3154E+03, 1.3174E+03, 1.3195E+03, 1.3215E+03, 1.3236E+03,
     +  1.3256E+03, 1.3277E+03, 1.3297E+03, 1.3317E+03, 1.3338E+03,
     + 1.3358E+03, 1.3379E+03, 1.3399E+03, 1.3420E+03, 1.3440E+03,
     +  1.3461E+03, 1.3481E+03, 1.3502E+03, 1.3522E+03, 1.3543E+03,
     + 1.3563E+03, 1.3584E+03, 1.3605E+03, 1.3625E+03, 1.3646E+03,
     +  1.3666E+03, 1.3687E+03, 1.3707E+03, 1.3728E+03, 1.3749E+03,
     + 1.3769E+03, 1.3790E+03, 1.3810E+03, 1.3831E+03, 1.3852E+03,
     +  1.3872E+03, 1.3893E+03, 1.3913E+03, 1.3934E+03, 1.3955E+03,
     + 1.3975E+03, 1.3996E+03, 1.4017E+03, 1.4037E+03, 1.4058E+03,
     +  1.4079E+03, 1.4099E+03, 1.4120E+03, 1.4141E+03, 1.4161E+03,
     + 1.4182E+03, 1.4203E+03, 1.4223E+03, 1.4244E+03, 1.4265E+03,
     +  1.4286E+03, 1.4306E+03, 1.4327E+03, 1.4348E+03, 1.4369E+03,
     + 1.4389E+03, 1.4410E+03, 1.4431E+03, 1.4452E+03, 1.4472E+03,
     +  1.4493E+03, 1.4514E+03, 1.4535E+03, 1.4556E+03, 1.4576E+03,
     + 1.4597E+03, 1.4618E+03, 1.4639E+03, 1.4660E+03, 1.4681E+03,
     +  1.4701E+03, 1.4722E+03, 1.4743E+03, 1.4764E+03, 1.4785E+03,
     + 1.4806E+03, 1.4826E+03, 1.4847E+03, 1.4868E+03, 1.4889E+03,
     +  1.4910E+03, 1.4931E+03, 1.4952E+03, 1.4973E+03, 1.4994E+03,
     + 1.5015E+03, 1.5035E+03, 1.5056E+03, 1.5077E+03, 1.5098E+03,
     +  1.5119E+03, 1.5140E+03, 1.5161E+03, 1.5182E+03, 1.5203E+03,
     + 1.5224E+03, 1.5245E+03, 1.5266E+03, 1.5287E+03, 1.5308E+03,
     +  1.5329E+03, 1.5350E+03, 1.5371E+03, 1.5392E+03, 1.5413E+03,
     + 1.5434E+03, 1.5455E+03, 1.5476E+03, 1.5497E+03, 1.5518E+03,
     +  1.5539E+03, 1.5560E+03, 1.5581E+03, 1.5602E+03, 1.5623E+03,
     + 1.5644E+03, 1.5665E+03, 1.5687E+03, 1.5708E+03, 1.5729E+03,
     +  1.5750E+03, 1.5771E+03, 1.5792E+03, 1.5813E+03, 1.5834E+03,
     + 1.5855E+03, 1.5877E+03, 1.5898E+03, 1.5919E+03, 1.5940E+03,
     +  1.5961E+03, 1.5982E+03, 1.6003E+03, 1.6025E+03, 1.6046E+03,
     + 1.6067E+03, 1.6088E+03, 1.6109E+03, 1.6130E+03, 1.6152E+03,
     +  1.6173E+03, 1.6194E+03, 1.6215E+03, 1.6236E+03, 1.6258E+03,
     + 1.6279E+03, 1.6300E+03, 1.6321E+03, 1.6343E+03, 1.6364E+03,
     +  1.6385E+03, 1.6406E+03, 1.6428E+03, 1.6449E+03, 1.6470E+03,
     + 1.6491E+03, 1.6513E+03, 1.6534E+03, 1.6555E+03, 1.6577E+03,
     +  1.6598E+03, 1.6619E+03, 1.6640E+03, 1.6662E+03, 1.6683E+03,
     + 1.6704E+03, 1.6726E+03, 1.6747E+03, 1.6768E+03, 1.6790E+03,
     +  1.6811E+03, 1.6832E+03, 1.6854E+03, 1.6875E+03, 1.6897E+03,
     + 1.6918E+03, 1.6939E+03, 1.6961E+03, 1.6982E+03, 1.7003E+03,
     +  1.7025E+03, 1.7046E+03, 1.7068E+03, 1.7089E+03, 1.7111E+03,
     + 1.7132E+03, 1.7153E+03, 1.7175E+03, 1.7196E+03, 1.7218E+03,
     +  1.7239E+03, 1.7261E+03, 1.7282E+03, 1.7303E+03, 1.7325E+03,
     + 1.7346E+03, 1.7368E+03, 1.7389E+03, 1.7411E+03, 1.7432E+03,
     +  1.7454E+03, 1.7475E+03, 1.7497E+03, 1.7518E+03, 1.7540E+03,
     + 1.7561E+03, 1.7583E+03, 1.7604E+03, 1.7626E+03, 1.7648E+03,
     +  1.7669E+03, 1.7691E+03, 1.7712E+03, 1.7734E+03, 1.7755E+03,
     + 1.7777E+03, 1.7798E+03, 1.7820E+03, 1.7842E+03, 1.7863E+03,
     +  1.7885E+03, 1.7906E+03, 1.7928E+03, 1.7950E+03, 1.7971E+03,
     + 1.7993E+03, 1.8014E+03, 1.8036E+03, 1.8058E+03, 1.8079E+03,
     +  1.8101E+03, 1.8123E+03, 1.8144E+03, 1.8166E+03, 1.8188E+03,
     + 1.8209E+03, 1.8231E+03, 1.8253E+03, 1.8274E+03, 1.8296E+03,
     +  1.8318E+03, 1.8339E+03, 1.8361E+03, 1.8383E+03, 1.8405E+03,
     + 1.8426E+03, 1.8448E+03, 1.8470E+03, 1.8491E+03, 1.8513E+03,
     +  1.8535E+03, 1.8557E+03, 1.8578E+03, 1.8600E+03, 1.8622E+03,
     + 1.8644E+03, 1.8665E+03, 1.8687E+03, 1.8709E+03, 1.8731E+03,
     +  1.8753E+03, 1.8774E+03, 1.8796E+03, 1.8818E+03, 1.8840E+03,
     + 1.8862E+03, 1.8883E+03, 1.8905E+03, 1.8927E+03, 1.8949E+03,
     +  1.8971E+03, 1.8993E+03, 1.9014E+03, 1.9036E+03, 1.9058E+03,
     + 1.9080E+03, 1.9102E+03, 1.9124E+03, 1.9146E+03, 1.9168E+03,
     +  1.9189E+03, 1.9211E+03, 1.9233E+03, 1.9255E+03, 1.9277E+03,
     + 1.9299E+03, 1.9321E+03, 1.9343E+03, 1.9365E+03, 1.9387E+03,
     +  1.9409E+03, 1.9430E+03, 1.9452E+03, 1.9474E+03, 1.9496E+03,
     + 1.9518E+03, 1.9540E+03, 1.9562E+03, 1.9584E+03, 1.9606E+03,
     +  1.9628E+03, 1.9650E+03, 1.9672E+03, 1.9694E+03, 1.9716E+03,
     + 1.9738E+03, 1.9760E+03, 1.9782E+03, 1.9804E+03, 1.9826E+03,
     +  1.9848E+03, 1.9870E+03, 1.9892E+03, 1.9914E+03, 1.9936E+03,
     + 1.9958E+03, 1.9980E+03, 2.0003E+03, 2.0025E+03, 2.0047E+03,
     +  2.0069E+03, 2.0091E+03, 2.0113E+03, 2.0135E+03, 2.0157E+03,
     + 2.0179E+03, 2.0201E+03, 2.0223E+03, 2.0246E+03, 2.0268E+03,
     +  2.0290E+03, 2.0312E+03, 2.0334E+03, 2.0356E+03, 2.0378E+03,
     + 2.0401E+03, 2.0423E+03, 2.0445E+03, 2.0467E+03, 2.0489E+03,
     +  2.0511E+03, 2.0534E+03, 2.0556E+03, 2.0578E+03, 2.0600E+03,
     + 2.0622E+03, 2.0644E+03, 2.0667E+03, 2.0689E+03, 2.0711E+03,
     +  2.0733E+03, 2.0756E+03, 2.0778E+03, 2.0800E+03, 2.0822E+03,
     + 2.0845E+03, 2.0867E+03, 2.0889E+03, 2.0911E+03, 2.0934E+03,
     +  2.0956E+03, 2.0978E+03, 2.1000E+03, 2.1023E+03, 2.1045E+03,
     + 2.1067E+03, 2.1090E+03, 2.1112E+03, 2.1134E+03, 2.1157E+03,
     +  2.1179E+03, 2.1201E+03, 2.1223E+03, 2.1246E+03, 2.1268E+03,
     + 2.1291E+03, 2.1313E+03, 2.1335E+03, 2.1358E+03, 2.1380E+03,
     +  2.1402E+03, 2.1425E+03, 2.1447E+03, 2.1469E+03, 2.1492E+03,
     + 2.1514E+03, 2.1537E+03, 2.1559E+03, 2.1581E+03, 2.1604E+03,
     +  2.1626E+03, 2.1649E+03, 2.1671E+03, 2.1694E+03, 2.1716E+03,
     + 2.1738E+03, 2.1761E+03, 2.1783E+03, 2.1806E+03, 2.1828E+03,
     +  2.1851E+03, 2.1873E+03, 2.1896E+03, 2.1918E+03, 2.1941E+03,
     + 2.1963E+03, 2.1986E+03, 2.2008E+03, 2.2031E+03, 2.2053E+03,
     +  2.2076E+03, 2.2098E+03, 2.2121E+03, 2.2143E+03, 2.2166E+03,
     + 2.2188E+03, 2.2211E+03, 2.2233E+03, 2.2256E+03, 2.2278E+03,
     +  2.2301E+03, 2.2324E+03, 2.2346E+03, 2.2369E+03, 2.2391E+03,
     + 2.2414E+03, 2.2437E+03, 2.2459E+03, 2.2482E+03, 2.2504E+03,
     +  2.2527E+03, 2.2550E+03, 2.2572E+03, 2.2595E+03, 2.2617E+03,
     + 2.2640E+03, 2.2663E+03, 2.2685E+03, 2.2708E+03, 2.2731E+03,
     +  2.2753E+03, 2.2776E+03, 2.2799E+03, 2.2821E+03, 2.2844E+03,
     + 2.2867E+03, 2.2889E+03, 2.2912E+03, 2.2935E+03, 2.2957E+03,
     +  2.2980E+03, 2.3003E+03, 2.3026E+03, 2.3048E+03, 2.3071E+03,
     + 2.3094E+03, 2.3117E+03, 2.3139E+03, 2.3162E+03, 2.3185E+03,
     +  2.3208E+03, 2.3230E+03, 2.3253E+03, 2.3276E+03, 2.3299E+03,
     + 2.3321E+03, 2.3344E+03, 2.3367E+03, 2.3390E+03, 2.3413E+03,
     +  2.3435E+03, 2.3458E+03, 2.3481E+03, 2.3504E+03, 2.3527E+03,
     + 2.3550E+03, 2.3572E+03, 2.3595E+03, 2.3618E+03, 2.3641E+03,
     +  2.3664E+03, 2.3687E+03, 2.3710E+03, 2.3732E+03, 2.3755E+03,
     + 2.3778E+03, 2.3801E+03, 2.3824E+03, 2.3847E+03, 2.3870E+03,
     +  2.3893E+03, 2.3916E+03, 2.3938E+03, 2.3961E+03, 2.3984E+03,
     + 2.4007E+03, 2.4030E+03, 2.4053E+03, 2.4076E+03, 2.4099E+03,
     +  2.4122E+03, 2.4145E+03, 2.4168E+03, 2.4191E+03, 2.4214E+03,
     + 2.4237E+03, 2.4260E+03, 2.4283E+03, 2.4306E+03, 2.4329E+03,
     +  2.4352E+03, 2.4375E+03, 2.4398E+03, 2.4421E+03, 2.4444E+03,
     + 2.4467E+03, 2.4490E+03, 2.4513E+03, 2.4536E+03, 2.4559E+03,
     +  2.4582E+03, 2.4605E+03, 2.4628E+03, 2.4651E+03, 2.4674E+03,
     + 2.4698E+03, 2.4721E+03, 2.4744E+03, 2.4767E+03, 2.4790E+03,
     +  2.4813E+03, 2.4836E+03, 2.4859E+03, 2.4882E+03, 2.4906E+03,
     + 2.4929E+03, 2.4952E+03, 2.4975E+03, 2.4998E+03, 2.5021E+03,
     +  2.5044E+03, 2.5068E+03, 2.5091E+03, 2.5114E+03, 2.5137E+03,
     + 2.5160E+03, 2.5183E+03, 2.5207E+03, 2.5230E+03, 2.5253E+03,
     +  2.5276E+03, 2.5299E+03, 2.5323E+03, 2.5346E+03, 2.5369E+03,
     + 2.5392E+03, 2.5416E+03, 2.5439E+03, 2.5462E+03, 2.5485E+03,
     +  2.5509E+03, 2.5532E+03, 2.5555E+03, 2.5578E+03, 2.5602E+03,
     + 2.5625E+03, 2.5648E+03, 2.5672E+03, 2.5695E+03, 2.5718E+03,
     +  2.5741E+03, 2.5765E+03, 2.5788E+03, 2.5811E+03, 2.5835E+03,
     + 2.5858E+03, 2.5881E+03, 2.5905E+03, 2.5928E+03, 2.5952E+03,
     +  2.5975E+03, 2.5998E+03, 2.6022E+03, 2.6045E+03, 2.6068E+03,
     + 2.6092E+03, 2.6115E+03, 2.6139E+03, 2.6162E+03, 2.6185E+03,
     +  2.6209E+03, 2.6232E+03, 2.6256E+03, 2.6279E+03, 2.6302E+03,
     + 2.6326E+03, 2.6349E+03, 2.6373E+03, 2.6396E+03, 2.6420E+03,
     +  2.6443E+03, 2.6467E+03, 2.6490E+03, 2.6514E+03, 2.6537E+03,
     + 2.6561E+03, 2.6584E+03, 2.6608E+03, 2.6631E+03, 2.6655E+03,
     +  2.6678E+03, 2.6702E+03, 2.6725E+03, 2.6749E+03, 2.6772E+03,
     + 2.6796E+03, 2.6819E+03, 2.6843E+03, 2.6866E+03, 2.6890E+03,
     +  2.6913E+03, 2.6937E+03, 2.6961E+03, 2.6984E+03, 2.7008E+03,
     + 2.7031E+03, 2.7055E+03, 2.7079E+03, 2.7102E+03, 2.7126E+03,
     +  2.7149E+03, 2.7173E+03, 2.7197E+03, 2.7220E+03, 2.7244E+03,
     + 2.7268E+03, 2.7291E+03, 2.7315E+03, 2.7339E+03, 2.7362E+03,
     +  2.7386E+03, 2.7410E+03, 2.7433E+03, 2.7457E+03, 2.7481E+03,
     + 2.7504E+03, 2.7528E+03, 2.7552E+03, 2.7575E+03, 2.7599E+03,
     +  2.7623E+03, 2.7647E+03, 2.7670E+03, 2.7694E+03, 2.7718E+03,
     + 2.7742E+03, 2.7765E+03, 2.7789E+03, 2.7813E+03, 2.7837E+03,
     +  2.7860E+03, 2.7884E+03, 2.7908E+03, 2.7932E+03, 2.7955E+03,
     + 2.7979E+03, 2.8003E+03, 2.8027E+03, 2.8051E+03, 2.8075E+03,
     +  2.8098E+03, 2.8122E+03, 2.8146E+03, 2.8170E+03, 2.8194E+03,
     + 2.8218E+03, 2.8241E+03, 2.8265E+03, 2.8289E+03, 2.8313E+03,
     +  2.8337E+03, 2.8361E+03, 2.8385E+03, 2.8409E+03, 2.8432E+03,
     + 2.8456E+03, 2.8480E+03, 2.8504E+03, 2.8528E+03, 2.8552E+03,
     +  2.8576E+03, 2.8600E+03, 2.8624E+03, 2.8648E+03, 2.8672E+03,
     + 2.8696E+03, 2.8720E+03, 2.8744E+03, 2.8768E+03, 2.8792E+03,
     +  2.8816E+03, 2.8840E+03, 2.8864E+03, 2.8888E+03, 2.8912E+03,
     + 2.8936E+03, 2.8960E+03, 2.8984E+03, 2.9008E+03, 2.9032E+03,
     +  2.9056E+03, 2.9080E+03, 2.9104E+03, 2.9128E+03, 2.9152E+03,
     + 2.9176E+03, 2.9200E+03, 2.9224E+03, 2.9248E+03, 2.9272E+03,
     +  2.9296E+03, 2.9320E+03, 2.9344E+03, 2.9369E+03, 2.9393E+03,
     + 2.9417E+03, 2.9441E+03, 2.9465E+03, 2.9489E+03, 2.9513E+03,
     +  2.9537E+03, 2.9562E+03, 2.9586E+03, 2.9610E+03, 2.9634E+03,
     + 2.9658E+03, 2.9682E+03, 2.9707E+03, 2.9731E+03, 2.9755E+03,
     +  2.9779E+03, 2.9803E+03, 2.9828E+03, 2.9852E+03, 2.9876E+03,
     + 2.9900E+03, 2.9924E+03, 2.9949E+03, 2.9973E+03, 2.9997E+03,
     +  3.0021E+03, 3.0046E+03, 3.0070E+03, 3.0094E+03, 3.0118E+03,
     + 3.0143E+03, 3.0167E+03, 3.0191E+03, 3.0216E+03, 3.0240E+03,
     +  3.0264E+03, 3.0288E+03, 3.0313E+03, 3.0337E+03, 3.0361E+03,
     + 3.0386E+03, 3.0410E+03, 3.0434E+03, 3.0459E+03, 3.0483E+03,
     +  3.0507E+03, 3.0532E+03, 3.0556E+03, 3.0581E+03, 3.0605E+03,
     + 3.0629E+03, 3.0654E+03, 3.0678E+03, 3.0703E+03, 3.0727E+03,
     +  3.0751E+03, 3.0776E+03, 3.0800E+03, 3.0825E+03, 3.0849E+03,
     + 3.0873E+03, 3.0898E+03, 3.0922E+03, 3.0947E+03, 3.0971E+03,
     +  3.0996E+03, 3.1020E+03, 3.1045E+03, 3.1069E+03, 3.1094E+03,
     + 3.1118E+03, 3.1143E+03, 3.1167E+03, 3.1192E+03, 3.1216E+03,
     +  3.1241E+03, 3.1265E+03, 3.1290E+03, 3.1314E+03, 3.1339E+03,
     + 3.1363E+03, 3.1388E+03, 3.1412E+03, 3.1437E+03, 3.1462E+03,
     +  3.1486E+03, 3.1511E+03, 3.1535E+03, 3.1560E+03, 3.1584E+03,
     + 3.1609E+03, 3.1634E+03, 3.1658E+03, 3.1683E+03, 3.1708E+03,
     +  3.1732E+03, 3.1757E+03, 3.1781E+03, 3.1806E+03, 3.1831E+03,
     + 3.1855E+03, 3.1880E+03, 3.1905E+03, 3.1929E+03, 3.1954E+03,
     +  3.1979E+03, 3.2003E+03, 3.2028E+03, 3.2053E+03, 3.2077E+03,
     + 3.2102E+03, 3.2127E+03, 3.2152E+03, 3.2176E+03, 3.2201E+03,
     +  3.2226E+03, 3.2251E+03, 3.2275E+03, 3.2300E+03, 3.2325E+03,
     + 3.2350E+03, 3.2374E+03, 3.2399E+03, 3.2424E+03, 3.2449E+03,
     +  3.2473E+03, 3.2498E+03, 3.2523E+03, 3.2548E+03, 3.2573E+03,
     + 3.2597E+03, 3.2622E+03, 3.2647E+03, 3.2672E+03, 3.2697E+03,
     +  3.2722E+03, 3.2747E+03, 3.2771E+03, 3.2796E+03, 3.2821E+03,
     + 3.2846E+03, 3.2871E+03, 3.2896E+03, 3.2921E+03, 3.2946E+03,
     +  3.2970E+03, 3.2995E+03, 3.3020E+03, 3.3045E+03, 3.3070E+03,
     + 3.3095E+03, 3.3120E+03, 3.3145E+03, 3.3170E+03, 3.3195E+03,
     +  3.3220E+03, 3.3245E+03, 3.3270E+03, 3.3295E+03, 3.3320E+03,
     + 3.3345E+03, 3.3370E+03, 3.3395E+03, 3.3420E+03, 3.3445E+03,
     +  3.3470E+03, 3.3495E+03, 3.3520E+03, 3.3545E+03, 3.3570E+03,
     + 3.3595E+03, 3.3620E+03, 3.3645E+03, 3.3670E+03, 3.3695E+03,
     +  3.3720E+03, 3.3745E+03, 3.3770E+03, 3.3795E+03, 3.3820E+03,
     + 3.3845E+03, 3.3870E+03, 3.3896E+03, 3.3921E+03, 3.3946E+03,
     +  3.3971E+03, 3.3996E+03, 3.4021E+03, 3.4046E+03, 3.4071E+03,
     + 3.4097E+03, 3.4122E+03, 3.4147E+03, 3.4172E+03, 3.4197E+03,
     +  3.4222E+03, 3.4248E+03, 3.4273E+03, 3.4298E+03, 3.4323E+03,
     + 3.4348E+03, 3.4374E+03, 3.4399E+03, 3.4424E+03, 3.4449E+03,
     +  3.4475E+03, 3.4500E+03, 3.4525E+03, 3.4550E+03, 3.4576E+03,
     + 3.4601E+03, 3.4626E+03, 3.4651E+03, 3.4677E+03, 3.4702E+03,
     +  3.4727E+03, 3.4752E+03, 3.4778E+03, 3.4803E+03, 3.4828E+03,
     + 3.4854E+03, 3.4879E+03, 3.4904E+03, 3.4930E+03, 3.4955E+03,
     +  3.4980E+03, 3.5006E+03, 3.5031E+03, 3.5056E+03, 3.5082E+03,
     + 3.5107E+03, 3.5133E+03, 3.5158E+03, 3.5183E+03, 3.5209E+03,
     +  3.5234E+03, 3.5260E+03, 3.5285E+03, 3.5310E+03, 3.5336E+03,
     + 3.5361E+03, 3.5387E+03, 3.5412E+03, 3.5438E+03, 3.5463E+03,
     +  3.5488E+03, 3.5514E+03, 3.5539E+03, 3.5565E+03, 3.5590E+03,
     + 3.5616E+03, 3.5641E+03, 3.5667E+03, 3.5692E+03, 3.5718E+03,
     +  3.5743E+03, 3.5769E+03, 3.5794E+03, 3.5820E+03, 3.5845E+03,
     + 3.5871E+03, 3.5897E+03, 3.5922E+03, 3.5948E+03, 3.5973E+03,
     +  3.5999E+03, 3.6024E+03, 3.6050E+03, 3.6076E+03, 3.6101E+03,
     + 3.6127E+03, 3.6152E+03, 3.6178E+03, 3.6204E+03, 3.6229E+03,
     +  3.6255E+03, 3.6280E+03, 3.6306E+03, 3.6332E+03, 3.6357E+03,
     + 3.6383E+03, 3.6409E+03, 3.6434E+03, 3.6460E+03, 3.6486E+03,
     +  3.6511E+03, 3.6537E+03, 3.6563E+03, 3.6588E+03, 3.6614E+03,
     + 3.6640E+03, 3.6666E+03, 3.6691E+03, 3.6717E+03, 3.6743E+03,
     +  3.6768E+03, 3.6794E+03, 3.6820E+03, 3.6846E+03, 3.6871E+03,
     + 3.6897E+03, 3.6923E+03, 3.6949E+03, 3.6975E+03, 3.7000E+03,
     +  3.7026E+03, 3.7052E+03, 3.7078E+03, 3.7104E+03, 3.7129E+03,
     + 3.7155E+03, 3.7181E+03, 3.7207E+03, 3.7233E+03, 3.7259E+03,
     +  3.7284E+03, 3.7310E+03, 3.7336E+03, 3.7362E+03, 3.7388E+03,
     + 3.7414E+03, 3.7440E+03, 3.7466E+03, 3.7491E+03, 3.7517E+03,
     +  3.7543E+03, 3.7569E+03, 3.7595E+03, 3.7621E+03, 3.7647E+03,
     + 3.7673E+03, 3.7699E+03, 3.7725E+03, 3.7751E+03, 3.7777E+03,
     +  3.7803E+03, 3.7829E+03, 3.7855E+03, 3.7881E+03, 3.7907E+03,
     + 3.7933E+03, 3.7959E+03, 3.7985E+03, 3.8011E+03, 3.8037E+03,
     +  3.8063E+03, 3.8089E+03, 3.8115E+03, 3.8141E+03, 3.8167E+03,
     + 3.8193E+03, 3.8219E+03, 3.8245E+03, 3.8271E+03, 3.8297E+03,
     +  3.8323E+03, 3.8349E+03, 3.8375E+03, 3.8402E+03, 3.8428E+03,
     + 3.8454E+03, 3.8480E+03, 3.8506E+03, 3.8532E+03, 3.8558E+03,
     +  3.8584E+03, 3.8611E+03, 3.8637E+03, 3.8663E+03, 3.8689E+03,
     + 3.8715E+03, 3.8741E+03, 3.8768E+03, 3.8794E+03, 3.8820E+03,
     +  3.8846E+03, 3.8872E+03, 3.8898E+03, 3.8925E+03, 3.8951E+03,
     + 3.8977E+03, 3.9003E+03, 3.9030E+03, 3.9056E+03, 3.9082E+03,
     +  3.9108E+03, 3.9135E+03, 3.9161E+03, 3.9187E+03, 3.9213E+03,
     + 3.9240E+03, 3.9266E+03, 3.9292E+03, 3.9319E+03, 3.9345E+03,
     +  3.9371E+03, 3.9398E+03, 3.9424E+03, 3.9450E+03, 3.9477E+03,
     + 3.9503E+03, 3.9529E+03, 3.9556E+03, 3.9582E+03, 3.9608E+03,
     +  3.9635E+03, 3.9661E+03, 3.9687E+03, 3.9714E+03, 3.9740E+03,
     + 3.9767E+03, 3.9793E+03, 3.9819E+03, 3.9846E+03, 3.9872E+03,
     +  3.9899E+03, 3.9925E+03, 3.9952E+03, 3.9978E+03, 4.0004E+03,
     + 4.0031E+03, 4.0057E+03, 4.0084E+03, 4.0110E+03, 4.0137E+03,
     +  4.0163E+03, 4.0190E+03, 4.0216E+03, 4.0243E+03, 4.0269E+03,
     + 4.0296E+03, 4.0322E+03, 4.0349E+03, 4.0375E+03, 4.0402E+03,
     +  4.0428E+03, 4.0455E+03, 4.0481E+03, 4.0508E+03, 4.0535E+03,
     + 4.0561E+03, 4.0588E+03, 4.0614E+03, 4.0641E+03, 4.0668E+03,
     +  4.0694E+03, 4.0721E+03, 4.0747E+03, 4.0774E+03, 4.0801E+03,
     + 4.0827E+03, 4.0854E+03, 4.0880E+03, 4.0907E+03, 4.0934E+03,
     +  4.0960E+03, 4.0987E+03, 4.1014E+03, 4.1040E+03, 4.1067E+03,
     + 4.1094E+03, 4.1120E+03, 4.1147E+03, 4.1174E+03, 4.1201E+03,
     +  4.1227E+03, 4.1254E+03, 4.1281E+03, 4.1307E+03, 4.1334E+03,
     + 4.1361E+03, 4.1388E+03, 4.1414E+03, 4.1441E+03, 4.1468E+03,
     +  4.1495E+03, 4.1522E+03, 4.1548E+03, 4.1575E+03, 4.1602E+03,
     + 4.1629E+03, 4.1656E+03, 4.1682E+03, 4.1709E+03, 4.1736E+03,
     +  4.1763E+03, 4.1790E+03, 4.1816E+03, 4.1843E+03, 4.1870E+03,
     + 4.1897E+03, 4.1924E+03, 4.1951E+03, 4.1978E+03, 4.2005E+03,
     +  4.2031E+03, 4.2058E+03, 4.2085E+03, 4.2112E+03, 4.2139E+03,
     + 4.2166E+03, 4.2193E+03, 4.2220E+03, 4.2247E+03, 4.2274E+03,
     +  4.2301E+03, 4.2328E+03, 4.2355E+03, 4.2382E+03, 4.2409E+03,
     + 4.2435E+03, 4.2462E+03, 4.2489E+03, 4.2516E+03, 4.2543E+03,
     +  4.2570E+03, 4.2597E+03, 4.2624E+03, 4.2651E+03, 4.2679E+03,
     + 4.2706E+03, 4.2733E+03, 4.2760E+03, 4.2787E+03, 4.2814E+03,
     +  4.2841E+03, 4.2868E+03, 4.2895E+03, 4.2922E+03, 4.2949E+03,
     + 4.2976E+03, 4.3003E+03, 4.3030E+03, 4.3058E+03, 4.3085E+03,
     +  4.3112E+03, 4.3139E+03, 4.3166E+03, 4.3193E+03, 4.3220E+03,
     + 4.3247E+03, 4.3275E+03, 4.3302E+03, 4.3329E+03, 4.3356E+03,
     +  4.3383E+03, 4.3410E+03, 4.3438E+03, 4.3465E+03, 4.3492E+03,
     + 4.3519E+03, 4.3547E+03, 4.3574E+03, 4.3601E+03, 4.3628E+03,
     +  4.3655E+03, 4.3683E+03, 4.3710E+03, 4.3737E+03, 4.3764E+03,
     + 4.3792E+03, 4.3819E+03, 4.3846E+03, 4.3874E+03, 4.3901E+03,
     +  4.3928E+03, 4.3955E+03, 4.3983E+03, 4.4010E+03, 4.4037E+03,
     + 4.4065E+03, 4.4092E+03, 4.4119E+03, 4.4147E+03, 4.4174E+03,
     +  4.4201E+03, 4.4229E+03, 4.4256E+03, 4.4284E+03, 4.4311E+03,
     + 4.4338E+03, 4.4366E+03, 4.4393E+03, 4.4420E+03, 4.4448E+03,
     +  4.4475E+03, 4.4503E+03, 4.4530E+03, 4.4558E+03, 4.4585E+03,
     + 4.4612E+03, 4.4640E+03, 4.4667E+03, 4.4695E+03, 4.4722E+03,
     +  4.4750E+03, 4.4777E+03, 4.4805E+03, 4.4832E+03, 4.4860E+03,
     + 4.4887E+03, 4.4915E+03, 4.4942E+03, 4.4970E+03, 4.4997E+03,
     +  4.5025E+03, 4.5052E+03, 4.5080E+03, 4.5107E+03, 4.5135E+03,
     + 4.5163E+03, 4.5190E+03, 4.5218E+03, 4.5245E+03, 4.5273E+03,
     +  4.5300E+03, 4.5328E+03, 4.5356E+03, 4.5383E+03, 4.5411E+03,
     + 4.5438E+03, 4.5466E+03, 4.5494E+03, 4.5521E+03, 4.5549E+03,
     +  4.5577E+03, 4.5604E+03, 4.5632E+03, 4.5660E+03, 4.5687E+03,
     + 4.5715E+03, 4.5743E+03, 4.5770E+03, 4.5798E+03, 4.5826E+03,
     +  4.5853E+03, 4.5881E+03, 4.5909E+03, 4.5937E+03, 4.5964E+03,
     + 4.5992E+03, 4.6020E+03, 4.6048E+03, 4.6075E+03, 4.6103E+03,
     +  4.6131E+03, 4.6159E+03, 4.6186E+03, 4.6214E+03, 4.6242E+03,
     + 4.6270E+03, 4.6297E+03, 4.6325E+03, 4.6353E+03, 4.6381E+03,
     +  4.6409E+03, 4.6437E+03, 4.6464E+03, 4.6492E+03, 4.6520E+03,
     + 4.6548E+03, 4.6576E+03, 4.6604E+03, 4.6632E+03, 4.6659E+03,
     +  4.6687E+03, 4.6715E+03, 4.6743E+03, 4.6771E+03, 4.6799E+03,
     + 4.6827E+03, 4.6855E+03, 4.6883E+03, 4.6911E+03, 4.6938E+03,
     +  4.6966E+03, 4.6994E+03, 4.7022E+03, 4.7050E+03, 4.7078E+03,
     + 4.7106E+03, 4.7134E+03, 4.7162E+03, 4.7190E+03, 4.7218E+03,
     +  4.7246E+03, 4.7274E+03, 4.7302E+03, 4.7330E+03, 4.7358E+03,
     + 4.7386E+03, 4.7414E+03, 4.7442E+03, 4.7470E+03, 4.7498E+03,
     +  4.7526E+03, 4.7555E+03, 4.7583E+03, 4.7611E+03, 4.7639E+03,
     + 4.7667E+03, 4.7695E+03, 4.7723E+03, 4.7751E+03, 4.7779E+03,
     +  4.7807E+03, 4.7835E+03, 4.7864E+03, 4.7892E+03, 4.7920E+03,
     + 4.7948E+03, 4.7976E+03, 4.8004E+03, 4.8032E+03, 4.8061E+03,
     +  4.8089E+03, 4.8117E+03, 4.8145E+03, 4.8173E+03, 4.8202E+03,
     + 4.8230E+03, 4.8258E+03, 4.8286E+03, 4.8314E+03, 4.8343E+03,
     +  4.8371E+03, 4.8399E+03, 4.8427E+03, 4.8456E+03, 4.8484E+03,
     + 4.8512E+03, 4.8540E+03, 4.8569E+03, 4.8597E+03, 4.8625E+03,
     +  4.8654E+03, 4.8682E+03, 4.8710E+03, 4.8738E+03, 4.8767E+03,
     + 4.8795E+03, 4.8823E+03, 4.8852E+03, 4.8880E+03, 4.8908E+03,
     +  4.8937E+03, 4.8965E+03, 4.8993E+03, 4.9022E+03, 4.9050E+03,
     + 4.9079E+03, 4.9107E+03, 4.9135E+03, 4.9164E+03, 4.9192E+03,
     +  4.9221E+03, 4.9249E+03, 4.9277E+03, 4.9306E+03, 4.9334E+03,
     + 4.9363E+03, 4.9391E+03, 4.9420E+03, 4.9448E+03, 4.9477E+03,
     +  4.9505E+03, 4.9533E+03, 4.9562E+03, 4.9590E+03, 4.9619E+03,
     + 4.9647E+03, 4.9676E+03, 4.9704E+03, 4.9733E+03, 4.9761E+03,
     +  4.9790E+03, 4.9819E+03, 4.9847E+03, 4.9876E+03, 4.9904E+03,
     + 4.9933E+03, 4.9961E+03, 4.9990E+03, 5.0018E+03, 5.0047E+03,
     +  5.0076E+03, 5.0104E+03, 5.0133E+03, 5.0161E+03, 5.0190E+03,
     + 5.0219E+03, 5.0247E+03, 5.0276E+03, 5.0304E+03, 5.0333E+03,
     +  5.0362E+03, 5.0390E+03, 5.0419E+03, 5.0448E+03, 5.0476E+03,
     + 5.0505E+03, 5.0534E+03, 5.0562E+03, 5.0591E+03, 5.0620E+03,
     +  5.0648E+03, 5.0677E+03, 5.0706E+03, 5.0735E+03, 5.0763E+03,
     + 5.0792E+03, 5.0821E+03, 5.0850E+03, 5.0878E+03, 5.0907E+03,
     +  5.0936E+03, 5.0965E+03, 5.0993E+03, 5.1022E+03, 5.1051E+03,
     + 5.1080E+03, 5.1108E+03, 5.1137E+03, 5.1166E+03, 5.1195E+03,
     +  5.1224E+03, 5.1253E+03, 5.1281E+03, 5.1310E+03, 5.1339E+03,
     + 5.1368E+03, 5.1397E+03, 5.1426E+03, 5.1454E+03, 5.1483E+03,
     +  5.1512E+03, 5.1541E+03, 5.1570E+03, 5.1599E+03, 5.1628E+03,
     + 5.1657E+03, 5.1686E+03, 5.1714E+03, 5.1743E+03, 5.1772E+03,
     +  5.1801E+03, 5.1830E+03, 5.1859E+03, 5.1888E+03, 5.1917E+03,
     + 5.1946E+03, 5.1975E+03, 5.2004E+03, 5.2033E+03, 5.2062E+03,
     +  5.2091E+03, 5.2120E+03, 5.2149E+03, 5.2178E+03, 5.2207E+03,
     + 5.2236E+03, 5.2265E+03, 5.2294E+03, 5.2323E+03, 5.2352E+03,
     +  5.2381E+03, 5.2410E+03, 5.2439E+03, 5.2468E+03, 5.2497E+03,
     + 5.2527E+03, 5.2556E+03, 5.2585E+03, 5.2614E+03, 5.2643E+03,
     +  5.2672E+03, 5.2701E+03, 5.2730E+03, 5.2759E+03, 5.2789E+03,
     + 5.2818E+03, 5.2847E+03, 5.2876E+03, 5.2905E+03, 5.2934E+03,
     +  5.2963E+03, 5.2993E+03, 5.3022E+03, 5.3051E+03, 5.3080E+03,
     + 5.3109E+03, 5.3139E+03, 5.3168E+03, 5.3197E+03, 5.3226E+03,
     +  5.3255E+03, 5.3285E+03, 5.3314E+03, 5.3343E+03, 5.3372E+03/
      data (QofT( 1,J),J=2409,4510)/ 5.3402E+03, 5.3431E+03,
     +  5.3460E+03, 5.3489E+03, 5.3519E+03, 5.3548E+03, 5.3577E+03,
     +  5.3607E+03,
     + 5.3636E+03, 5.3665E+03, 5.3695E+03, 5.3724E+03, 5.3753E+03,
     +  5.3782E+03, 5.3812E+03, 5.3841E+03, 5.3871E+03, 5.3900E+03,
     + 5.3929E+03, 5.3959E+03, 5.3988E+03, 5.4017E+03, 5.4047E+03,
     +  5.4076E+03, 5.4106E+03, 5.4135E+03, 5.4164E+03, 5.4194E+03,
     + 5.4223E+03, 5.4253E+03, 5.4282E+03, 5.4311E+03, 5.4341E+03,
     +  5.4370E+03, 5.4400E+03, 5.4429E+03, 5.4459E+03, 5.4488E+03,
     + 5.4518E+03, 5.4547E+03, 5.4577E+03, 5.4606E+03, 5.4636E+03,
     +  5.4665E+03, 5.4695E+03, 5.4724E+03, 5.4754E+03, 5.4783E+03,
     + 5.4813E+03, 5.4842E+03, 5.4872E+03, 5.4901E+03, 5.4931E+03,
     +  5.4961E+03, 5.4990E+03, 5.5020E+03, 5.5049E+03, 5.5079E+03,
     + 5.5108E+03, 5.5138E+03, 5.5168E+03, 5.5197E+03, 5.5227E+03,
     +  5.5256E+03, 5.5286E+03, 5.5316E+03, 5.5345E+03, 5.5375E+03,
     + 5.5405E+03, 5.5434E+03, 5.5464E+03, 5.5494E+03, 5.5523E+03,
     +  5.5553E+03, 5.5583E+03, 5.5612E+03, 5.5642E+03, 5.5672E+03,
     + 5.5701E+03, 5.5731E+03, 5.5761E+03, 5.5791E+03, 5.5820E+03,
     +  5.5850E+03, 5.5880E+03, 5.5910E+03, 5.5939E+03, 5.5969E+03,
     + 5.5999E+03, 5.6029E+03, 5.6058E+03, 5.6088E+03, 5.6118E+03,
     +  5.6148E+03, 5.6178E+03, 5.6207E+03, 5.6237E+03, 5.6267E+03,
     + 5.6297E+03, 5.6327E+03, 5.6357E+03, 5.6386E+03, 5.6416E+03,
     +  5.6446E+03, 5.6476E+03, 5.6506E+03, 5.6536E+03, 5.6566E+03,
     + 5.6595E+03, 5.6625E+03, 5.6655E+03, 5.6685E+03, 5.6715E+03,
     +  5.6745E+03, 5.6775E+03, 5.6805E+03, 5.6835E+03, 5.6865E+03,
     + 5.6895E+03, 5.6924E+03, 5.6954E+03, 5.6984E+03, 5.7014E+03,
     +  5.7044E+03, 5.7074E+03, 5.7104E+03, 5.7134E+03, 5.7164E+03,
     + 5.7194E+03, 5.7224E+03, 5.7254E+03, 5.7284E+03, 5.7314E+03,
     +  5.7344E+03, 5.7374E+03, 5.7404E+03, 5.7434E+03, 5.7464E+03,
     + 5.7495E+03, 5.7525E+03, 5.7555E+03, 5.7585E+03, 5.7615E+03,
     +  5.7645E+03, 5.7675E+03, 5.7705E+03, 5.7735E+03, 5.7765E+03,
     + 5.7795E+03, 5.7826E+03, 5.7856E+03, 5.7886E+03, 5.7916E+03,
     +  5.7946E+03, 5.7976E+03, 5.8006E+03, 5.8037E+03, 5.8067E+03,
     + 5.8097E+03, 5.8127E+03, 5.8157E+03, 5.8187E+03, 5.8218E+03,
     +  5.8248E+03, 5.8278E+03, 5.8308E+03, 5.8338E+03, 5.8369E+03,
     + 5.8399E+03, 5.8429E+03, 5.8459E+03, 5.8490E+03, 5.8520E+03,
     +  5.8550E+03, 5.8580E+03, 5.8611E+03, 5.8641E+03, 5.8671E+03,
     + 5.8701E+03, 5.8732E+03, 5.8762E+03, 5.8792E+03, 5.8823E+03,
     +  5.8853E+03, 5.8883E+03, 5.8914E+03, 5.8944E+03, 5.8974E+03,
     + 5.9005E+03, 5.9035E+03, 5.9065E+03, 5.9096E+03, 5.9126E+03,
     +  5.9156E+03, 5.9187E+03, 5.9217E+03, 5.9248E+03, 5.9278E+03,
     + 5.9308E+03, 5.9339E+03, 5.9369E+03, 5.9400E+03, 5.9430E+03,
     +  5.9460E+03, 5.9491E+03, 5.9521E+03, 5.9552E+03, 5.9582E+03,
     + 5.9613E+03, 5.9643E+03, 5.9674E+03, 5.9704E+03, 5.9735E+03,
     +  5.9765E+03, 5.9796E+03, 5.9826E+03, 5.9857E+03, 5.9887E+03,
     + 5.9918E+03, 5.9948E+03, 5.9979E+03, 6.0009E+03, 6.0040E+03,
     +  6.0070E+03, 6.0101E+03, 6.0131E+03, 6.0162E+03, 6.0192E+03,
     + 6.0223E+03, 6.0254E+03, 6.0284E+03, 6.0315E+03, 6.0345E+03,
     +  6.0376E+03, 6.0407E+03, 6.0437E+03, 6.0468E+03, 6.0498E+03,
     + 6.0529E+03, 6.0560E+03, 6.0590E+03, 6.0621E+03, 6.0652E+03,
     +  6.0682E+03, 6.0713E+03, 6.0744E+03, 6.0774E+03, 6.0805E+03,
     + 6.0836E+03, 6.0866E+03, 6.0897E+03, 6.0928E+03, 6.0958E+03,
     +  6.0989E+03, 6.1020E+03, 6.1051E+03, 6.1081E+03, 6.1112E+03,
     + 6.1143E+03, 6.1174E+03, 6.1204E+03, 6.1235E+03, 6.1266E+03,
     +  6.1297E+03, 6.1327E+03, 6.1358E+03, 6.1389E+03, 6.1420E+03,
     + 6.1451E+03, 6.1481E+03, 6.1512E+03, 6.1543E+03, 6.1574E+03,
     +  6.1605E+03, 6.1635E+03, 6.1666E+03, 6.1697E+03, 6.1728E+03,
     + 6.1759E+03, 6.1790E+03, 6.1821E+03, 6.1851E+03, 6.1882E+03,
     +  6.1913E+03, 6.1944E+03, 6.1975E+03, 6.2006E+03, 6.2037E+03,
     + 6.2068E+03, 6.2099E+03, 6.2130E+03, 6.2160E+03, 6.2191E+03,
     +  6.2222E+03, 6.2253E+03, 6.2284E+03, 6.2315E+03, 6.2346E+03,
     + 6.2377E+03, 6.2408E+03, 6.2439E+03, 6.2470E+03, 6.2501E+03,
     +  6.2532E+03, 6.2563E+03, 6.2594E+03, 6.2625E+03, 6.2656E+03,
     + 6.2687E+03, 6.2718E+03, 6.2749E+03, 6.2780E+03, 6.2811E+03,
     +  6.2842E+03, 6.2873E+03, 6.2904E+03, 6.2936E+03, 6.2967E+03,
     + 6.2998E+03, 6.3029E+03, 6.3060E+03, 6.3091E+03, 6.3122E+03,
     +  6.3153E+03, 6.3184E+03, 6.3215E+03, 6.3247E+03, 6.3278E+03,
     + 6.3309E+03, 6.3340E+03, 6.3371E+03, 6.3402E+03, 6.3433E+03,
     +  6.3465E+03, 6.3496E+03, 6.3527E+03, 6.3558E+03, 6.3589E+03,
     + 6.3621E+03, 6.3652E+03, 6.3683E+03, 6.3714E+03, 6.3745E+03,
     +  6.3777E+03, 6.3808E+03, 6.3839E+03, 6.3870E+03, 6.3902E+03,
     + 6.3933E+03, 6.3964E+03, 6.3995E+03, 6.4027E+03, 6.4058E+03,
     +  6.4089E+03, 6.4120E+03, 6.4152E+03, 6.4183E+03, 6.4214E+03,
     + 6.4246E+03, 6.4277E+03, 6.4308E+03, 6.4340E+03, 6.4371E+03,
     +  6.4402E+03, 6.4434E+03, 6.4465E+03, 6.4496E+03, 6.4528E+03,
     + 6.4559E+03, 6.4590E+03, 6.4622E+03, 6.4653E+03, 6.4684E+03,
     +  6.4716E+03, 6.4747E+03, 6.4779E+03, 6.4810E+03, 6.4842E+03,
     + 6.4873E+03, 6.4904E+03, 6.4936E+03, 6.4967E+03, 6.4999E+03,
     +  6.5030E+03, 6.5062E+03, 6.5093E+03, 6.5124E+03, 6.5156E+03,
     + 6.5187E+03, 6.5219E+03, 6.5250E+03, 6.5282E+03, 6.5313E+03,
     +  6.5345E+03, 6.5376E+03, 6.5408E+03, 6.5439E+03, 6.5471E+03,
     + 6.5502E+03, 6.5534E+03, 6.5566E+03, 6.5597E+03, 6.5629E+03,
     +  6.5660E+03, 6.5692E+03, 6.5723E+03, 6.5755E+03, 6.5787E+03,
     + 6.5818E+03, 6.5850E+03, 6.5881E+03, 6.5913E+03, 6.5944E+03,
     +  6.5976E+03, 6.6008E+03, 6.6039E+03, 6.6071E+03, 6.6103E+03,
     + 6.6134E+03, 6.6166E+03, 6.6198E+03, 6.6229E+03, 6.6261E+03,
     +  6.6293E+03, 6.6324E+03, 6.6356E+03, 6.6388E+03, 6.6419E+03,
     + 6.6451E+03, 6.6483E+03, 6.6514E+03, 6.6546E+03, 6.6578E+03,
     +  6.6610E+03, 6.6641E+03, 6.6673E+03, 6.6705E+03, 6.6737E+03,
     + 6.6768E+03, 6.6800E+03, 6.6832E+03, 6.6864E+03, 6.6895E+03,
     +  6.6927E+03, 6.6959E+03, 6.6991E+03, 6.7022E+03, 6.7054E+03,
     + 6.7086E+03, 6.7118E+03, 6.7150E+03, 6.7182E+03, 6.7213E+03,
     +  6.7245E+03, 6.7277E+03, 6.7309E+03, 6.7341E+03, 6.7373E+03,
     + 6.7404E+03, 6.7436E+03, 6.7468E+03, 6.7500E+03, 6.7532E+03,
     +  6.7564E+03, 6.7596E+03, 6.7628E+03, 6.7660E+03, 6.7691E+03,
     + 6.7723E+03, 6.7755E+03, 6.7787E+03, 6.7819E+03, 6.7851E+03,
     +  6.7883E+03, 6.7915E+03, 6.7947E+03, 6.7979E+03, 6.8011E+03,
     + 6.8043E+03, 6.8075E+03, 6.8107E+03, 6.8139E+03, 6.8171E+03,
     +  6.8203E+03, 6.8235E+03, 6.8267E+03, 6.8299E+03, 6.8331E+03,
     + 6.8363E+03, 6.8395E+03, 6.8427E+03, 6.8459E+03, 6.8491E+03,
     +  6.8523E+03, 6.8555E+03, 6.8587E+03, 6.8619E+03, 6.8651E+03,
     + 6.8684E+03, 6.8716E+03, 6.8748E+03, 6.8780E+03, 6.8812E+03,
     +  6.8844E+03, 6.8876E+03, 6.8908E+03, 6.8940E+03, 6.8973E+03,
     + 6.9005E+03, 6.9037E+03, 6.9069E+03, 6.9101E+03, 6.9133E+03,
     +  6.9165E+03, 6.9198E+03, 6.9230E+03, 6.9262E+03, 6.9294E+03,
     + 6.9326E+03, 6.9359E+03, 6.9391E+03, 6.9423E+03, 6.9455E+03,
     +  6.9487E+03, 6.9520E+03, 6.9552E+03, 6.9584E+03, 6.9616E+03,
     + 6.9649E+03, 6.9681E+03, 6.9713E+03, 6.9745E+03, 6.9778E+03,
     +  6.9810E+03, 6.9842E+03, 6.9874E+03, 6.9907E+03, 6.9939E+03,
     + 6.9971E+03, 7.0004E+03, 7.0036E+03, 7.0068E+03, 7.0101E+03,
     +  7.0133E+03, 7.0165E+03, 7.0198E+03, 7.0230E+03, 7.0262E+03,
     + 7.0295E+03, 7.0327E+03, 7.0359E+03, 7.0392E+03, 7.0424E+03,
     +  7.0457E+03, 7.0489E+03, 7.0521E+03, 7.0554E+03, 7.0586E+03,
     + 7.0619E+03, 7.0651E+03, 7.0683E+03, 7.0716E+03, 7.0748E+03,
     +  7.0781E+03, 7.0813E+03, 7.0846E+03, 7.0878E+03, 7.0910E+03,
     + 7.0943E+03, 7.0975E+03, 7.1008E+03, 7.1040E+03, 7.1073E+03,
     +  7.1105E+03, 7.1138E+03, 7.1170E+03, 7.1203E+03, 7.1235E+03,
     + 7.1268E+03, 7.1300E+03, 7.1333E+03, 7.1365E+03, 7.1398E+03,
     +  7.1431E+03, 7.1463E+03, 7.1496E+03, 7.1528E+03, 7.1561E+03,
     + 7.1593E+03, 7.1626E+03, 7.1659E+03, 7.1691E+03, 7.1724E+03,
     +  7.1756E+03, 7.1789E+03, 7.1822E+03, 7.1854E+03, 7.1887E+03,
     + 7.1919E+03, 7.1952E+03, 7.1985E+03, 7.2017E+03, 7.2050E+03,
     +  7.2083E+03, 7.2115E+03, 7.2148E+03, 7.2181E+03, 7.2213E+03,
     + 7.2246E+03, 7.2279E+03, 7.2311E+03, 7.2344E+03, 7.2377E+03,
     +  7.2409E+03, 7.2442E+03, 7.2475E+03, 7.2508E+03, 7.2540E+03,
     + 7.2573E+03, 7.2606E+03, 7.2638E+03, 7.2671E+03, 7.2704E+03,
     +  7.2737E+03, 7.2770E+03, 7.2802E+03, 7.2835E+03, 7.2868E+03,
     + 7.2901E+03, 7.2933E+03, 7.2966E+03, 7.2999E+03, 7.3032E+03,
     +  7.3065E+03, 7.3097E+03, 7.3130E+03, 7.3163E+03, 7.3196E+03,
     + 7.3229E+03, 7.3262E+03, 7.3295E+03, 7.3327E+03, 7.3360E+03,
     +  7.3393E+03, 7.3426E+03, 7.3459E+03, 7.3492E+03, 7.3525E+03,
     + 7.3558E+03, 7.3590E+03, 7.3623E+03, 7.3656E+03, 7.3689E+03,
     +  7.3722E+03, 7.3755E+03, 7.3788E+03, 7.3821E+03, 7.3854E+03,
     + 7.3887E+03, 7.3920E+03, 7.3953E+03, 7.3986E+03, 7.4019E+03,
     +  7.4052E+03, 7.4085E+03, 7.4118E+03, 7.4150E+03, 7.4183E+03,
     + 7.4216E+03, 7.4249E+03, 7.4283E+03, 7.4316E+03, 7.4349E+03,
     +  7.4382E+03, 7.4415E+03, 7.4448E+03, 7.4481E+03, 7.4514E+03,
     + 7.4547E+03, 7.4580E+03, 7.4613E+03, 7.4646E+03, 7.4679E+03,
     +  7.4712E+03, 7.4745E+03, 7.4778E+03, 7.4811E+03, 7.4844E+03,
     + 7.4878E+03, 7.4911E+03, 7.4944E+03, 7.4977E+03, 7.5010E+03,
     +  7.5043E+03, 7.5076E+03, 7.5109E+03, 7.5143E+03, 7.5176E+03,
     + 7.5209E+03, 7.5242E+03, 7.5275E+03, 7.5308E+03, 7.5342E+03,
     +  7.5375E+03, 7.5408E+03, 7.5441E+03, 7.5474E+03, 7.5508E+03,
     + 7.5541E+03, 7.5574E+03, 7.5607E+03, 7.5640E+03, 7.5674E+03,
     +  7.5707E+03, 7.5740E+03, 7.5773E+03, 7.5807E+03, 7.5840E+03,
     + 7.5873E+03, 7.5906E+03, 7.5940E+03, 7.5973E+03, 7.6006E+03,
     +  7.6040E+03, 7.6073E+03, 7.6106E+03, 7.6139E+03, 7.6173E+03,
     + 7.6206E+03, 7.6239E+03, 7.6273E+03, 7.6306E+03, 7.6339E+03,
     +  7.6373E+03, 7.6406E+03, 7.6439E+03, 7.6473E+03, 7.6506E+03,
     + 7.6540E+03, 7.6573E+03, 7.6606E+03, 7.6640E+03, 7.6673E+03,
     +  7.6706E+03, 7.6740E+03, 7.6773E+03, 7.6807E+03, 7.6840E+03,
     + 7.6874E+03, 7.6907E+03, 7.6940E+03, 7.6974E+03, 7.7007E+03,
     +  7.7041E+03, 7.7074E+03, 7.7108E+03, 7.7141E+03, 7.7175E+03,
     + 7.7208E+03, 7.7241E+03, 7.7275E+03, 7.7308E+03, 7.7342E+03,
     +  7.7375E+03, 7.7409E+03, 7.7442E+03, 7.7476E+03, 7.7509E+03,
     + 7.7543E+03, 7.7577E+03, 7.7610E+03, 7.7644E+03, 7.7677E+03,
     +  7.7711E+03, 7.7744E+03, 7.7778E+03, 7.7811E+03, 7.7845E+03,
     + 7.7879E+03, 7.7912E+03, 7.7946E+03, 7.7979E+03, 7.8013E+03,
     +  7.8046E+03, 7.8080E+03, 7.8114E+03, 7.8147E+03, 7.8181E+03,
     + 7.8215E+03, 7.8248E+03, 7.8282E+03, 7.8315E+03, 7.8349E+03,
     +  7.8383E+03, 7.8416E+03, 7.8450E+03, 7.8484E+03, 7.8517E+03,
     + 7.8551E+03, 7.8585E+03, 7.8618E+03, 7.8652E+03, 7.8686E+03,
     +  7.8720E+03, 7.8753E+03, 7.8787E+03, 7.8821E+03, 7.8854E+03,
     + 7.8888E+03, 7.8922E+03, 7.8956E+03, 7.8989E+03, 7.9023E+03,
     +  7.9057E+03, 7.9091E+03, 7.9124E+03, 7.9158E+03, 7.9192E+03,
     + 7.9226E+03, 7.9260E+03, 7.9293E+03, 7.9327E+03, 7.9361E+03,
     +  7.9395E+03, 7.9429E+03, 7.9462E+03, 7.9496E+03, 7.9530E+03,
     + 7.9564E+03, 7.9598E+03, 7.9632E+03, 7.9665E+03, 7.9699E+03,
     +  7.9733E+03, 7.9767E+03, 7.9801E+03, 7.9835E+03, 7.9869E+03,
     + 7.9902E+03, 7.9936E+03, 7.9970E+03, 8.0004E+03, 8.0038E+03,
     +  8.0072E+03, 8.0106E+03, 8.0140E+03, 8.0174E+03, 8.0208E+03,
     + 8.0242E+03, 8.0275E+03, 8.0309E+03, 8.0343E+03, 8.0377E+03,
     +  8.0411E+03, 8.0445E+03, 8.0479E+03, 8.0513E+03, 8.0547E+03,
     + 8.0581E+03, 8.0615E+03, 8.0649E+03, 8.0683E+03, 8.0717E+03,
     +  8.0751E+03, 8.0785E+03, 8.0819E+03, 8.0853E+03, 8.0887E+03,
     + 8.0921E+03, 8.0955E+03, 8.0989E+03, 8.1023E+03, 8.1057E+03,
     +  8.1092E+03, 8.1126E+03, 8.1160E+03, 8.1194E+03, 8.1228E+03,
     + 8.1262E+03, 8.1296E+03, 8.1330E+03, 8.1364E+03, 8.1398E+03,
     +  8.1432E+03, 8.1466E+03, 8.1501E+03, 8.1535E+03, 8.1569E+03,
     + 8.1603E+03, 8.1637E+03, 8.1671E+03, 8.1705E+03, 8.1740E+03,
     +  8.1774E+03, 8.1808E+03, 8.1842E+03, 8.1876E+03, 8.1910E+03,
     + 8.1945E+03, 8.1979E+03, 8.2013E+03, 8.2047E+03, 8.2081E+03,
     +  8.2116E+03, 8.2150E+03, 8.2184E+03, 8.2218E+03, 8.2253E+03,
     + 8.2287E+03, 8.2321E+03, 8.2355E+03, 8.2390E+03, 8.2424E+03,
     +  8.2458E+03, 8.2492E+03, 8.2527E+03, 8.2561E+03, 8.2595E+03,
     + 8.2629E+03, 8.2664E+03, 8.2698E+03, 8.2732E+03, 8.2767E+03,
     +  8.2801E+03, 8.2835E+03, 8.2870E+03, 8.2904E+03, 8.2938E+03,
     + 8.2973E+03, 8.3007E+03, 8.3041E+03, 8.3076E+03, 8.3110E+03,
     +  8.3144E+03, 8.3179E+03, 8.3213E+03, 8.3247E+03, 8.3282E+03,
     + 8.3316E+03, 8.3351E+03, 8.3385E+03, 8.3419E+03, 8.3454E+03,
     +  8.3488E+03, 8.3523E+03, 8.3557E+03, 8.3591E+03, 8.3626E+03,
     + 8.3660E+03, 8.3695E+03, 8.3729E+03, 8.3764E+03, 8.3798E+03,
     +  8.3833E+03, 8.3867E+03, 8.3901E+03, 8.3936E+03, 8.3970E+03,
     + 8.4005E+03, 8.4039E+03, 8.4074E+03, 8.4108E+03, 8.4143E+03,
     +  8.4177E+03, 8.4212E+03, 8.4246E+03, 8.4281E+03, 8.4315E+03,
     + 8.4350E+03, 8.4385E+03, 8.4419E+03, 8.4454E+03, 8.4488E+03,
     +  8.4523E+03, 8.4557E+03, 8.4592E+03, 8.4626E+03, 8.4661E+03,
     + 8.4696E+03, 8.4730E+03, 8.4765E+03, 8.4799E+03, 8.4834E+03,
     +  8.4869E+03, 8.4903E+03, 8.4938E+03, 8.4972E+03, 8.5007E+03,
     + 8.5042E+03, 8.5076E+03, 8.5111E+03, 8.5146E+03, 8.5180E+03,
     +  8.5215E+03, 8.5250E+03, 8.5284E+03, 8.5319E+03, 8.5354E+03,
     + 8.5388E+03, 8.5423E+03, 8.5458E+03, 8.5492E+03, 8.5527E+03,
     +  8.5562E+03, 8.5596E+03, 8.5631E+03, 8.5666E+03, 8.5701E+03,
     + 8.5735E+03, 8.5770E+03, 8.5805E+03, 8.5840E+03, 8.5874E+03,
     +  8.5909E+03, 8.5944E+03, 8.5979E+03, 8.6013E+03, 8.6048E+03,
     + 8.6083E+03, 8.6118E+03, 8.6152E+03, 8.6187E+03, 8.6222E+03,
     +  8.6257E+03, 8.6292E+03, 8.6326E+03, 8.6361E+03, 8.6396E+03,
     + 8.6431E+03, 8.6466E+03, 8.6501E+03, 8.6535E+03, 8.6570E+03,
     +  8.6605E+03, 8.6640E+03, 8.6675E+03, 8.6710E+03, 8.6745E+03,
     + 8.6779E+03, 8.6814E+03, 8.6849E+03, 8.6884E+03, 8.6919E+03,
     +  8.6954E+03, 8.6989E+03, 8.7024E+03, 8.7059E+03, 8.7093E+03,
     + 8.7128E+03, 8.7163E+03, 8.7198E+03, 8.7233E+03, 8.7268E+03,
     +  8.7303E+03, 8.7338E+03, 8.7373E+03, 8.7408E+03, 8.7443E+03,
     + 8.7478E+03, 8.7513E+03, 8.7548E+03, 8.7583E+03, 8.7618E+03,
     +  8.7653E+03, 8.7688E+03, 8.7723E+03, 8.7758E+03, 8.7793E+03,
     + 8.7828E+03, 8.7863E+03, 8.7898E+03, 8.7933E+03, 8.7968E+03,
     +  8.8003E+03, 8.8038E+03, 8.8073E+03, 8.8108E+03, 8.8143E+03,
     + 8.8178E+03, 8.8213E+03, 8.8248E+03, 8.8283E+03, 8.8319E+03,
     +  8.8354E+03, 8.8389E+03, 8.8424E+03, 8.8459E+03, 8.8494E+03,
     + 8.8529E+03, 8.8564E+03, 8.8599E+03, 8.8635E+03, 8.8670E+03,
     +  8.8705E+03, 8.8740E+03, 8.8775E+03, 8.8810E+03, 8.8845E+03,
     + 8.8881E+03, 8.8916E+03, 8.8951E+03, 8.8986E+03, 8.9021E+03,
     +  8.9056E+03, 8.9092E+03, 8.9127E+03, 8.9162E+03, 8.9197E+03,
     + 8.9232E+03, 8.9268E+03, 8.9303E+03, 8.9338E+03, 8.9373E+03,
     +  8.9408E+03, 8.9444E+03, 8.9479E+03, 8.9514E+03, 8.9549E+03,
     + 8.9585E+03, 8.9620E+03, 8.9655E+03, 8.9690E+03, 8.9726E+03,
     +  8.9761E+03, 8.9796E+03, 8.9832E+03, 8.9867E+03, 8.9902E+03,
     + 8.9938E+03, 8.9973E+03, 9.0008E+03, 9.0043E+03, 9.0079E+03,
     +  9.0114E+03, 9.0149E+03, 9.0185E+03, 9.0220E+03, 9.0255E+03,
     + 9.0291E+03, 9.0326E+03, 9.0361E+03, 9.0397E+03, 9.0432E+03,
     +  9.0468E+03, 9.0503E+03, 9.0538E+03, 9.0574E+03, 9.0609E+03,
     + 9.0644E+03, 9.0680E+03, 9.0715E+03, 9.0751E+03, 9.0786E+03,
     +  9.0822E+03, 9.0857E+03, 9.0892E+03, 9.0928E+03, 9.0963E+03,
     + 9.0999E+03, 9.1034E+03, 9.1070E+03, 9.1105E+03, 9.1140E+03,
     +  9.1176E+03, 9.1211E+03, 9.1247E+03, 9.1282E+03, 9.1318E+03,
     + 9.1353E+03, 9.1389E+03, 9.1424E+03, 9.1460E+03, 9.1495E+03,
     +  9.1531E+03, 9.1566E+03, 9.1602E+03, 9.1637E+03, 9.1673E+03,
     + 9.1708E+03, 9.1744E+03, 9.1780E+03, 9.1815E+03, 9.1851E+03,
     +  9.1886E+03, 9.1922E+03, 9.1957E+03, 9.1993E+03, 9.2028E+03,
     + 9.2064E+03, 9.2100E+03, 9.2135E+03, 9.2171E+03, 9.2206E+03,
     +  9.2242E+03, 9.2278E+03, 9.2313E+03, 9.2349E+03, 9.2384E+03,
     + 9.2420E+03, 9.2456E+03, 9.2491E+03, 9.2527E+03, 9.2563E+03,
     +  9.2598E+03, 9.2634E+03, 9.2670E+03, 9.2705E+03, 9.2741E+03,
     + 9.2777E+03, 9.2812E+03, 9.2848E+03, 9.2884E+03, 9.2919E+03,
     +  9.2955E+03, 9.2991E+03, 9.3026E+03, 9.3062E+03, 9.3098E+03,
     + 9.3134E+03, 9.3169E+03, 9.3205E+03, 9.3241E+03, 9.3276E+03,
     +  9.3312E+03, 9.3348E+03, 9.3384E+03, 9.3419E+03, 9.3455E+03,
     + 9.3491E+03, 9.3527E+03, 9.3562E+03, 9.3598E+03, 9.3634E+03,
     +  9.3670E+03, 9.3706E+03, 9.3741E+03, 9.3777E+03, 9.3813E+03,
     + 9.3849E+03, 9.3885E+03, 9.3920E+03, 9.3956E+03, 9.3992E+03,
     +  9.4028E+03, 9.4064E+03, 9.4100E+03, 9.4135E+03, 9.4171E+03,
     + 9.4207E+03, 9.4243E+03, 9.4279E+03, 9.4315E+03, 9.4351E+03,
     +  9.4386E+03, 9.4422E+03, 9.4458E+03, 9.4494E+03, 9.4530E+03,
     + 9.4566E+03, 9.4602E+03, 9.4638E+03, 9.4674E+03, 9.4710E+03,
     +  9.4745E+03, 9.4781E+03, 9.4817E+03, 9.4853E+03, 9.4889E+03,
     + 9.4925E+03, 9.4961E+03, 9.4997E+03, 9.5033E+03, 9.5069E+03,
     +  9.5105E+03, 9.5141E+03, 9.5177E+03, 9.5213E+03, 9.5249E+03,
     + 9.5285E+03, 9.5321E+03, 9.5357E+03, 9.5393E+03, 9.5429E+03,
     +  9.5465E+03, 9.5501E+03, 9.5537E+03, 9.5573E+03, 9.5609E+03,
     + 9.5645E+03, 9.5681E+03, 9.5717E+03, 9.5753E+03, 9.5789E+03,
     +  9.5825E+03, 9.5861E+03, 9.5897E+03, 9.5933E+03, 9.5969E+03,
     + 9.6005E+03, 9.6041E+03, 9.6078E+03, 9.6114E+03, 9.6150E+03,
     +  9.6186E+03, 9.6222E+03, 9.6258E+03, 9.6294E+03, 9.6330E+03,
     + 9.6366E+03, 9.6403E+03, 9.6439E+03, 9.6475E+03, 9.6511E+03,
     +  9.6547E+03, 9.6583E+03, 9.6619E+03, 9.6656E+03, 9.6692E+03,
     + 9.6728E+03, 9.6764E+03, 9.6800E+03, 9.6836E+03, 9.6873E+03,
     +  9.6909E+03, 9.6945E+03, 9.6981E+03, 9.7017E+03, 9.7053E+03,
     + 9.7090E+03, 9.7126E+03, 9.7162E+03, 9.7198E+03, 9.7235E+03,
     +  9.7271E+03, 9.7307E+03, 9.7343E+03, 9.7380E+03, 9.7416E+03,
     + 9.7452E+03, 9.7488E+03, 9.7525E+03, 9.7561E+03, 9.7597E+03,
     +  9.7633E+03, 9.7670E+03, 9.7706E+03, 9.7742E+03, 9.7778E+03,
     + 9.7815E+03, 9.7851E+03, 9.7887E+03, 9.7924E+03, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        68
      data (QofT( 2,J),J=1,2408)/ 3.5624E+00, 4.5717E+00, 5.7902E+00,
     +  7.1107E+00, 8.4926E+00, 9.9138E+00, 1.1361E+01, 1.2827E+01,
     + 1.4305E+01, 1.5794E+01, 1.7290E+01, 1.8792E+01, 2.0299E+01,
     +  2.1809E+01, 2.3323E+01, 2.4839E+01, 2.6357E+01, 2.7877E+01,
     + 2.9398E+01, 3.0921E+01, 3.3970E+01, 3.7023E+01, 4.0078E+01,
     +  4.3136E+01, 4.6196E+01, 4.9257E+01, 5.2319E+01, 5.5382E+01,
     + 5.8447E+01, 6.1512E+01, 6.4577E+01, 6.7644E+01, 7.0711E+01,
     +  7.3778E+01, 7.6846E+01, 7.9914E+01, 8.2982E+01, 8.6051E+01,
     + 8.9120E+01, 9.2189E+01, 9.5259E+01, 9.8328E+01, 1.0140E+02,
     +  1.0447E+02, 1.0754E+02, 1.1061E+02, 1.1368E+02, 1.1675E+02,
     + 1.1982E+02, 1.2289E+02, 1.2596E+02, 1.2903E+02, 1.3211E+02,
     +  1.3518E+02, 1.3825E+02, 1.4132E+02, 1.4439E+02, 1.4746E+02,
     + 1.5054E+02, 1.5361E+02, 1.5668E+02, 1.5975E+02, 1.6282E+02,
     +  1.6590E+02, 1.6897E+02, 1.7204E+02, 1.7511E+02, 1.7819E+02,
     + 1.8126E+02, 1.8433E+02, 1.8741E+02, 1.9048E+02, 1.9355E+02,
     +  1.9662E+02, 1.9970E+02, 2.0277E+02, 2.0584E+02, 2.0892E+02,
     + 2.1199E+02, 2.1506E+02, 2.1814E+02, 2.2121E+02, 2.2428E+02,
     +  2.2736E+02, 2.3043E+02, 2.3351E+02, 2.3658E+02, 2.3965E+02,
     + 2.4273E+02, 2.4580E+02, 2.4888E+02, 2.5195E+02, 2.5502E+02,
     +  2.5810E+02, 2.6117E+02, 2.6425E+02, 2.6732E+02, 2.7040E+02,
     + 2.7347E+02, 2.7654E+02, 2.7962E+02, 2.8269E+02, 2.8577E+02,
     +  2.8884E+02, 2.9192E+02, 2.9499E+02, 2.9807E+02, 3.0114E+02,
     + 3.0422E+02, 3.0730E+02, 3.1037E+02, 3.1345E+02, 3.1652E+02,
     +  3.1960E+02, 3.2267E+02, 3.2575E+02, 3.2883E+02, 3.3190E+02,
     + 3.3498E+02, 3.3806E+02, 3.4114E+02, 3.4421E+02, 3.4729E+02,
     +  3.5037E+02, 3.5345E+02, 3.5652E+02, 3.5960E+02, 3.6268E+02,
     + 3.6576E+02, 3.6884E+02, 3.7192E+02, 3.7500E+02, 3.7808E+02,
     +  3.8116E+02, 3.8424E+02, 3.8732E+02, 3.9040E+02, 3.9348E+02,
     + 3.9657E+02, 3.9965E+02, 4.0273E+02, 4.0582E+02, 4.0890E+02,
     +  4.1199E+02, 4.1507E+02, 4.1816E+02, 4.2124E+02, 4.2433E+02,
     + 4.2742E+02, 4.3050E+02, 4.3359E+02, 4.3668E+02, 4.3977E+02,
     +  4.4286E+02, 4.4595E+02, 4.4904E+02, 4.5214E+02, 4.5523E+02,
     + 4.5832E+02, 4.6142E+02, 4.6451E+02, 4.6761E+02, 4.7071E+02,
     +  4.7381E+02, 4.7691E+02, 4.8001E+02, 4.8311E+02, 4.8621E+02,
     + 4.8931E+02, 4.9242E+02, 4.9552E+02, 4.9863E+02, 5.0173E+02,
     +  5.0484E+02, 5.0795E+02, 5.1106E+02, 5.1417E+02, 5.1729E+02,
     + 5.2040E+02, 5.2352E+02, 5.2663E+02, 5.2975E+02, 5.3287E+02,
     +  5.3599E+02, 5.3911E+02, 5.4224E+02, 5.4536E+02, 5.4849E+02,
     + 5.5162E+02, 5.5475E+02, 5.5788E+02, 5.6101E+02, 5.6414E+02,
     +  5.6728E+02, 5.7042E+02, 5.7356E+02, 5.7670E+02, 5.7984E+02,
     + 5.8299E+02, 5.8613E+02, 5.8928E+02, 5.9243E+02, 5.9558E+02,
     +  5.9874E+02, 6.0189E+02, 6.0505E+02, 6.0821E+02, 6.1137E+02,
     + 6.1453E+02, 6.1770E+02, 6.2087E+02, 6.2404E+02, 6.2721E+02,
     +  6.3038E+02, 6.3356E+02, 6.3674E+02, 6.3992E+02, 6.4310E+02,
     + 6.4629E+02, 6.4947E+02, 6.5266E+02, 6.5585E+02, 6.5905E+02,
     +  6.6225E+02, 6.6545E+02, 6.6865E+02, 6.7185E+02, 6.7506E+02,
     + 6.7827E+02, 6.8148E+02, 6.8469E+02, 6.8791E+02, 6.9113E+02,
     +  6.9435E+02, 6.9758E+02, 7.0081E+02, 7.0404E+02, 7.0727E+02,
     + 7.1051E+02, 7.1374E+02, 7.1699E+02, 7.2023E+02, 7.2348E+02,
     +  7.2673E+02, 7.2998E+02, 7.3324E+02, 7.3649E+02, 7.3976E+02,
     + 7.4302E+02, 7.4629E+02, 7.4956E+02, 7.5283E+02, 7.5611E+02,
     +  7.5939E+02, 7.6267E+02, 7.6596E+02, 7.6925E+02, 7.7254E+02,
     + 7.7584E+02, 7.7914E+02, 7.8244E+02, 7.8575E+02, 7.8905E+02,
     +  7.9237E+02, 7.9568E+02, 7.9900E+02, 8.0232E+02, 8.0565E+02,
     + 8.0898E+02, 8.1231E+02, 8.1565E+02, 8.1899E+02, 8.2233E+02,
     +  8.2567E+02, 8.2902E+02, 8.3238E+02, 8.3573E+02, 8.3910E+02,
     + 8.4246E+02, 8.4583E+02, 8.4920E+02, 8.5257E+02, 8.5595E+02,
     +  8.5933E+02, 8.6272E+02, 8.6611E+02, 8.6950E+02, 8.7290E+02,
     + 8.7630E+02, 8.7971E+02, 8.8312E+02, 8.8653E+02, 8.8994E+02,
     +  8.9336E+02, 8.9679E+02, 9.0022E+02, 9.0365E+02, 9.0708E+02,
     + 9.1052E+02, 9.1397E+02, 9.1741E+02, 9.2087E+02, 9.2432E+02,
     +  9.2778E+02, 9.3124E+02, 9.3471E+02, 9.3818E+02, 9.4166E+02,
     + 9.4514E+02, 9.4862E+02, 9.5211E+02, 9.5560E+02, 9.5910E+02,
     +  9.6260E+02, 9.6610E+02, 9.6961E+02, 9.7312E+02, 9.7664E+02,
     + 9.8016E+02, 9.8369E+02, 9.8722E+02, 9.9075E+02, 9.9429E+02,
     +  9.9783E+02, 1.0014E+03, 1.0049E+03, 1.0085E+03, 1.0120E+03,
     + 1.0156E+03, 1.0192E+03, 1.0228E+03, 1.0263E+03, 1.0299E+03,
     +  1.0335E+03, 1.0371E+03, 1.0407E+03, 1.0443E+03, 1.0479E+03,
     + 1.0515E+03, 1.0551E+03, 1.0587E+03, 1.0623E+03, 1.0660E+03,
     +  1.0696E+03, 1.0732E+03, 1.0769E+03, 1.0805E+03, 1.0842E+03,
     + 1.0878E+03, 1.0915E+03, 1.0951E+03, 1.0988E+03, 1.1025E+03,
     +  1.1062E+03, 1.1098E+03, 1.1135E+03, 1.1172E+03, 1.1209E+03,
     + 1.1246E+03, 1.1283E+03, 1.1320E+03, 1.1357E+03, 1.1395E+03,
     +  1.1432E+03, 1.1469E+03, 1.1506E+03, 1.1544E+03, 1.1581E+03,
     + 1.1619E+03, 1.1656E+03, 1.1694E+03, 1.1731E+03, 1.1769E+03,
     +  1.1807E+03, 1.1844E+03, 1.1882E+03, 1.1920E+03, 1.1958E+03,
     + 1.1996E+03, 1.2034E+03, 1.2072E+03, 1.2110E+03, 1.2148E+03,
     +  1.2186E+03, 1.2225E+03, 1.2263E+03, 1.2301E+03, 1.2340E+03,
     + 1.2378E+03, 1.2417E+03, 1.2455E+03, 1.2494E+03, 1.2532E+03,
     +  1.2571E+03, 1.2610E+03, 1.2648E+03, 1.2687E+03, 1.2726E+03,
     + 1.2765E+03, 1.2804E+03, 1.2843E+03, 1.2882E+03, 1.2921E+03,
     +  1.2961E+03, 1.3000E+03, 1.3039E+03, 1.3078E+03, 1.3118E+03,
     + 1.3157E+03, 1.3197E+03, 1.3236E+03, 1.3276E+03, 1.3315E+03,
     +  1.3355E+03, 1.3395E+03, 1.3435E+03, 1.3475E+03, 1.3514E+03,
     + 1.3554E+03, 1.3594E+03, 1.3634E+03, 1.3674E+03, 1.3715E+03,
     +  1.3755E+03, 1.3795E+03, 1.3835E+03, 1.3876E+03, 1.3916E+03,
     + 1.3957E+03, 1.3997E+03, 1.4038E+03, 1.4078E+03, 1.4119E+03,
     +  1.4160E+03, 1.4200E+03, 1.4241E+03, 1.4282E+03, 1.4323E+03,
     + 1.4364E+03, 1.4405E+03, 1.4446E+03, 1.4487E+03, 1.4528E+03,
     +  1.4570E+03, 1.4611E+03, 1.4652E+03, 1.4694E+03, 1.4735E+03,
     + 1.4777E+03, 1.4818E+03, 1.4860E+03, 1.4902E+03, 1.4943E+03,
     +  1.4985E+03, 1.5027E+03, 1.5069E+03, 1.5111E+03, 1.5153E+03,
     + 1.5195E+03, 1.5237E+03, 1.5279E+03, 1.5321E+03, 1.5363E+03,
     +  1.5406E+03, 1.5448E+03, 1.5490E+03, 1.5533E+03, 1.5575E+03,
     + 1.5618E+03, 1.5661E+03, 1.5703E+03, 1.5746E+03, 1.5789E+03,
     +  1.5832E+03, 1.5874E+03, 1.5917E+03, 1.5960E+03, 1.6003E+03,
     + 1.6047E+03, 1.6090E+03, 1.6133E+03, 1.6176E+03, 1.6220E+03,
     +  1.6263E+03, 1.6306E+03, 1.6350E+03, 1.6393E+03, 1.6437E+03,
     + 1.6481E+03, 1.6524E+03, 1.6568E+03, 1.6612E+03, 1.6656E+03,
     +  1.6700E+03, 1.6744E+03, 1.6788E+03, 1.6832E+03, 1.6876E+03,
     + 1.6920E+03, 1.6964E+03, 1.7009E+03, 1.7053E+03, 1.7097E+03,
     +  1.7142E+03, 1.7186E+03, 1.7231E+03, 1.7276E+03, 1.7320E+03,
     + 1.7365E+03, 1.7410E+03, 1.7455E+03, 1.7500E+03, 1.7545E+03,
     +  1.7590E+03, 1.7635E+03, 1.7680E+03, 1.7725E+03, 1.7770E+03,
     + 1.7816E+03, 1.7861E+03, 1.7906E+03, 1.7952E+03, 1.7997E+03,
     +  1.8043E+03, 1.8089E+03, 1.8134E+03, 1.8180E+03, 1.8226E+03,
     + 1.8272E+03, 1.8318E+03, 1.8363E+03, 1.8410E+03, 1.8456E+03,
     +  1.8502E+03, 1.8548E+03, 1.8594E+03, 1.8640E+03, 1.8687E+03,
     + 1.8733E+03, 1.8780E+03, 1.8826E+03, 1.8873E+03, 1.8919E+03,
     +  1.8966E+03, 1.9013E+03, 1.9060E+03, 1.9107E+03, 1.9153E+03,
     + 1.9200E+03, 1.9247E+03, 1.9295E+03, 1.9342E+03, 1.9389E+03,
     +  1.9436E+03, 1.9483E+03, 1.9531E+03, 1.9578E+03, 1.9626E+03,
     + 1.9673E+03, 1.9721E+03, 1.9769E+03, 1.9816E+03, 1.9864E+03,
     +  1.9912E+03, 1.9960E+03, 2.0008E+03, 2.0056E+03, 2.0104E+03,
     + 2.0152E+03, 2.0200E+03, 2.0248E+03, 2.0296E+03, 2.0345E+03,
     +  2.0393E+03, 2.0442E+03, 2.0490E+03, 2.0539E+03, 2.0587E+03,
     + 2.0636E+03, 2.0685E+03, 2.0733E+03, 2.0782E+03, 2.0831E+03,
     +  2.0880E+03, 2.0929E+03, 2.0978E+03, 2.1027E+03, 2.1077E+03,
     + 2.1126E+03, 2.1175E+03, 2.1224E+03, 2.1274E+03, 2.1323E+03,
     +  2.1373E+03, 2.1422E+03, 2.1472E+03, 2.1522E+03, 2.1572E+03,
     + 2.1621E+03, 2.1671E+03, 2.1721E+03, 2.1771E+03, 2.1821E+03,
     +  2.1871E+03, 2.1921E+03, 2.1972E+03, 2.2022E+03, 2.2072E+03,
     + 2.2123E+03, 2.2173E+03, 2.2224E+03, 2.2274E+03, 2.2325E+03,
     +  2.2376E+03, 2.2426E+03, 2.2477E+03, 2.2528E+03, 2.2579E+03,
     + 2.2630E+03, 2.2681E+03, 2.2732E+03, 2.2783E+03, 2.2834E+03,
     +  2.2886E+03, 2.2937E+03, 2.2988E+03, 2.3040E+03, 2.3091E+03,
     + 2.3143E+03, 2.3194E+03, 2.3246E+03, 2.3298E+03, 2.3350E+03,
     +  2.3401E+03, 2.3453E+03, 2.3505E+03, 2.3557E+03, 2.3609E+03,
     + 2.3661E+03, 2.3714E+03, 2.3766E+03, 2.3818E+03, 2.3871E+03,
     +  2.3923E+03, 2.3975E+03, 2.4028E+03, 2.4081E+03, 2.4133E+03,
     + 2.4186E+03, 2.4239E+03, 2.4292E+03, 2.4344E+03, 2.4397E+03,
     +  2.4450E+03, 2.4503E+03, 2.4557E+03, 2.4610E+03, 2.4663E+03,
     + 2.4716E+03, 2.4770E+03, 2.4823E+03, 2.4877E+03, 2.4930E+03,
     +  2.4984E+03, 2.5037E+03, 2.5091E+03, 2.5145E+03, 2.5199E+03,
     + 2.5253E+03, 2.5307E+03, 2.5361E+03, 2.5415E+03, 2.5469E+03,
     +  2.5523E+03, 2.5577E+03, 2.5631E+03, 2.5686E+03, 2.5740E+03,
     + 2.5795E+03, 2.5849E+03, 2.5904E+03, 2.5959E+03, 2.6013E+03,
     +  2.6068E+03, 2.6123E+03, 2.6178E+03, 2.6233E+03, 2.6288E+03,
     + 2.6343E+03, 2.6398E+03, 2.6453E+03, 2.6508E+03, 2.6564E+03,
     +  2.6619E+03, 2.6674E+03, 2.6730E+03, 2.6786E+03, 2.6841E+03,
     + 2.6897E+03, 2.6952E+03, 2.7008E+03, 2.7064E+03, 2.7120E+03,
     +  2.7176E+03, 2.7232E+03, 2.7288E+03, 2.7344E+03, 2.7400E+03,
     + 2.7457E+03, 2.7513E+03, 2.7569E+03, 2.7626E+03, 2.7682E+03,
     +  2.7739E+03, 2.7796E+03, 2.7852E+03, 2.7909E+03, 2.7966E+03,
     + 2.8023E+03, 2.8079E+03, 2.8136E+03, 2.8194E+03, 2.8251E+03,
     +  2.8308E+03, 2.8365E+03, 2.8422E+03, 2.8480E+03, 2.8537E+03,
     + 2.8594E+03, 2.8652E+03, 2.8710E+03, 2.8767E+03, 2.8825E+03,
     +  2.8883E+03, 2.8940E+03, 2.8998E+03, 2.9056E+03, 2.9114E+03,
     + 2.9172E+03, 2.9230E+03, 2.9289E+03, 2.9347E+03, 2.9405E+03,
     +  2.9464E+03, 2.9522E+03, 2.9580E+03, 2.9639E+03, 2.9698E+03,
     + 2.9756E+03, 2.9815E+03, 2.9874E+03, 2.9933E+03, 2.9991E+03,
     +  3.0050E+03, 3.0109E+03, 3.0169E+03, 3.0228E+03, 3.0287E+03,
     + 3.0346E+03, 3.0405E+03, 3.0465E+03, 3.0524E+03, 3.0584E+03,
     +  3.0643E+03, 3.0703E+03, 3.0763E+03, 3.0822E+03, 3.0882E+03,
     + 3.0942E+03, 3.1002E+03, 3.1062E+03, 3.1122E+03, 3.1182E+03,
     +  3.1242E+03, 3.1303E+03, 3.1363E+03, 3.1423E+03, 3.1484E+03,
     + 3.1544E+03, 3.1605E+03, 3.1665E+03, 3.1726E+03, 3.1787E+03,
     +  3.1848E+03, 3.1908E+03, 3.1969E+03, 3.2030E+03, 3.2091E+03,
     + 3.2152E+03, 3.2213E+03, 3.2275E+03, 3.2336E+03, 3.2397E+03,
     +  3.2459E+03, 3.2520E+03, 3.2582E+03, 3.2643E+03, 3.2705E+03,
     + 3.2767E+03, 3.2828E+03, 3.2890E+03, 3.2952E+03, 3.3014E+03,
     +  3.3076E+03, 3.3138E+03, 3.3200E+03, 3.3262E+03, 3.3325E+03,
     + 3.3387E+03, 3.3449E+03, 3.3512E+03, 3.3574E+03, 3.3637E+03,
     +  3.3700E+03, 3.3762E+03, 3.3825E+03, 3.3888E+03, 3.3951E+03,
     + 3.4014E+03, 3.4077E+03, 3.4140E+03, 3.4203E+03, 3.4266E+03,
     +  3.4329E+03, 3.4392E+03, 3.4456E+03, 3.4519E+03, 3.4583E+03,
     + 3.4646E+03, 3.4710E+03, 3.4774E+03, 3.4837E+03, 3.4901E+03,
     +  3.4965E+03, 3.5029E+03, 3.5093E+03, 3.5157E+03, 3.5221E+03,
     + 3.5285E+03, 3.5349E+03, 3.5414E+03, 3.5478E+03, 3.5543E+03,
     +  3.5607E+03, 3.5672E+03, 3.5736E+03, 3.5801E+03, 3.5866E+03,
     + 3.5930E+03, 3.5995E+03, 3.6060E+03, 3.6125E+03, 3.6190E+03,
     +  3.6255E+03, 3.6320E+03, 3.6386E+03, 3.6451E+03, 3.6516E+03,
     + 3.6582E+03, 3.6647E+03, 3.6713E+03, 3.6778E+03, 3.6844E+03,
     +  3.6910E+03, 3.6976E+03, 3.7041E+03, 3.7107E+03, 3.7173E+03,
     + 3.7239E+03, 3.7305E+03, 3.7372E+03, 3.7438E+03, 3.7504E+03,
     +  3.7571E+03, 3.7637E+03, 3.7703E+03, 3.7770E+03, 3.7837E+03,
     + 3.7903E+03, 3.7970E+03, 3.8037E+03, 3.8104E+03, 3.8171E+03,
     +  3.8238E+03, 3.8305E+03, 3.8372E+03, 3.8439E+03, 3.8506E+03,
     + 3.8573E+03, 3.8641E+03, 3.8708E+03, 3.8776E+03, 3.8843E+03,
     +  3.8911E+03, 3.8979E+03, 3.9046E+03, 3.9114E+03, 3.9182E+03,
     + 3.9250E+03, 3.9318E+03, 3.9386E+03, 3.9454E+03, 3.9522E+03,
     +  3.9591E+03, 3.9659E+03, 3.9727E+03, 3.9796E+03, 3.9864E+03,
     + 3.9933E+03, 4.0001E+03, 4.0070E+03, 4.0139E+03, 4.0208E+03,
     +  4.0277E+03, 4.0346E+03, 4.0415E+03, 4.0484E+03, 4.0553E+03,
     + 4.0622E+03, 4.0691E+03, 4.0761E+03, 4.0830E+03, 4.0899E+03,
     +  4.0969E+03, 4.1039E+03, 4.1108E+03, 4.1178E+03, 4.1248E+03,
     + 4.1318E+03, 4.1387E+03, 4.1457E+03, 4.1527E+03, 4.1598E+03,
     +  4.1668E+03, 4.1738E+03, 4.1808E+03, 4.1879E+03, 4.1949E+03,
     + 4.2019E+03, 4.2090E+03, 4.2161E+03, 4.2231E+03, 4.2302E+03,
     +  4.2373E+03, 4.2444E+03, 4.2515E+03, 4.2586E+03, 4.2657E+03,
     + 4.2728E+03, 4.2799E+03, 4.2870E+03, 4.2942E+03, 4.3013E+03,
     +  4.3084E+03, 4.3156E+03, 4.3228E+03, 4.3299E+03, 4.3371E+03,
     + 4.3443E+03, 4.3515E+03, 4.3586E+03, 4.3658E+03, 4.3730E+03,
     +  4.3802E+03, 4.3875E+03, 4.3947E+03, 4.4019E+03, 4.4092E+03,
     + 4.4164E+03, 4.4236E+03, 4.4309E+03, 4.4382E+03, 4.4454E+03,
     +  4.4527E+03, 4.4600E+03, 4.4673E+03, 4.4746E+03, 4.4819E+03,
     + 4.4892E+03, 4.4965E+03, 4.5038E+03, 4.5111E+03, 4.5185E+03,
     +  4.5258E+03, 4.5331E+03, 4.5405E+03, 4.5479E+03, 4.5552E+03,
     + 4.5626E+03, 4.5700E+03, 4.5774E+03, 4.5847E+03, 4.5921E+03,
     +  4.5996E+03, 4.6070E+03, 4.6144E+03, 4.6218E+03, 4.6292E+03,
     + 4.6367E+03, 4.6441E+03, 4.6516E+03, 4.6590E+03, 4.6665E+03,
     +  4.6740E+03, 4.6814E+03, 4.6889E+03, 4.6964E+03, 4.7039E+03,
     + 4.7114E+03, 4.7189E+03, 4.7264E+03, 4.7340E+03, 4.7415E+03,
     +  4.7490E+03, 4.7566E+03, 4.7641E+03, 4.7717E+03, 4.7792E+03,
     + 4.7868E+03, 4.7944E+03, 4.8019E+03, 4.8095E+03, 4.8171E+03,
     +  4.8247E+03, 4.8323E+03, 4.8400E+03, 4.8476E+03, 4.8552E+03,
     + 4.8628E+03, 4.8705E+03, 4.8781E+03, 4.8858E+03, 4.8934E+03,
     +  4.9011E+03, 4.9088E+03, 4.9165E+03, 4.9242E+03, 4.9318E+03,
     + 4.9395E+03, 4.9473E+03, 4.9550E+03, 4.9627E+03, 4.9704E+03,
     +  4.9781E+03, 4.9859E+03, 4.9936E+03, 5.0014E+03, 5.0092E+03,
     + 5.0169E+03, 5.0247E+03, 5.0325E+03, 5.0403E+03, 5.0481E+03,
     +  5.0559E+03, 5.0637E+03, 5.0715E+03, 5.0793E+03, 5.0871E+03,
     + 5.0950E+03, 5.1028E+03, 5.1106E+03, 5.1185E+03, 5.1264E+03,
     +  5.1342E+03, 5.1421E+03, 5.1500E+03, 5.1579E+03, 5.1658E+03,
     + 5.1737E+03, 5.1816E+03, 5.1895E+03, 5.1974E+03, 5.2053E+03,
     +  5.2133E+03, 5.2212E+03, 5.2292E+03, 5.2371E+03, 5.2451E+03,
     + 5.2530E+03, 5.2610E+03, 5.2690E+03, 5.2770E+03, 5.2850E+03,
     +  5.2930E+03, 5.3010E+03, 5.3090E+03, 5.3170E+03, 5.3250E+03,
     + 5.3331E+03, 5.3411E+03, 5.3492E+03, 5.3572E+03, 5.3653E+03,
     +  5.3734E+03, 5.3814E+03, 5.3895E+03, 5.3976E+03, 5.4057E+03,
     + 5.4138E+03, 5.4219E+03, 5.4300E+03, 5.4381E+03, 5.4463E+03,
     +  5.4544E+03, 5.4626E+03, 5.4707E+03, 5.4789E+03, 5.4870E+03,
     + 5.4952E+03, 5.5034E+03, 5.5116E+03, 5.5197E+03, 5.5279E+03,
     +  5.5361E+03, 5.5444E+03, 5.5526E+03, 5.5608E+03, 5.5690E+03,
     + 5.5773E+03, 5.5855E+03, 5.5938E+03, 5.6020E+03, 5.6103E+03,
     +  5.6186E+03, 5.6268E+03, 5.6351E+03, 5.6434E+03, 5.6517E+03,
     + 5.6600E+03, 5.6683E+03, 5.6767E+03, 5.6850E+03, 5.6933E+03,
     +  5.7017E+03, 5.7100E+03, 5.7184E+03, 5.7267E+03, 5.7351E+03,
     + 5.7435E+03, 5.7518E+03, 5.7602E+03, 5.7686E+03, 5.7770E+03,
     +  5.7854E+03, 5.7939E+03, 5.8023E+03, 5.8107E+03, 5.8191E+03,
     + 5.8276E+03, 5.8360E+03, 5.8445E+03, 5.8530E+03, 5.8614E+03,
     +  5.8699E+03, 5.8784E+03, 5.8869E+03, 5.8954E+03, 5.9039E+03,
     + 5.9124E+03, 5.9209E+03, 5.9294E+03, 5.9380E+03, 5.9465E+03,
     +  5.9551E+03, 5.9636E+03, 5.9722E+03, 5.9807E+03, 5.9893E+03,
     + 5.9979E+03, 6.0065E+03, 6.0151E+03, 6.0237E+03, 6.0323E+03,
     +  6.0409E+03, 6.0495E+03, 6.0582E+03, 6.0668E+03, 6.0754E+03,
     + 6.0841E+03, 6.0928E+03, 6.1014E+03, 6.1101E+03, 6.1188E+03,
     +  6.1275E+03, 6.1361E+03, 6.1448E+03, 6.1535E+03, 6.1623E+03,
     + 6.1710E+03, 6.1797E+03, 6.1884E+03, 6.1972E+03, 6.2059E+03,
     +  6.2147E+03, 6.2234E+03, 6.2322E+03, 6.2410E+03, 6.2498E+03,
     + 6.2586E+03, 6.2674E+03, 6.2762E+03, 6.2850E+03, 6.2938E+03,
     +  6.3026E+03, 6.3114E+03, 6.3203E+03, 6.3291E+03, 6.3380E+03,
     + 6.3468E+03, 6.3557E+03, 6.3646E+03, 6.3735E+03, 6.3824E+03,
     +  6.3912E+03, 6.4001E+03, 6.4091E+03, 6.4180E+03, 6.4269E+03,
     + 6.4358E+03, 6.4448E+03, 6.4537E+03, 6.4627E+03, 6.4716E+03,
     +  6.4806E+03, 6.4895E+03, 6.4985E+03, 6.5075E+03, 6.5165E+03,
     + 6.5255E+03, 6.5345E+03, 6.5435E+03, 6.5525E+03, 6.5616E+03,
     +  6.5706E+03, 6.5797E+03, 6.5887E+03, 6.5978E+03, 6.6068E+03,
     + 6.6159E+03, 6.6250E+03, 6.6341E+03, 6.6431E+03, 6.6522E+03,
     +  6.6614E+03, 6.6705E+03, 6.6796E+03, 6.6887E+03, 6.6978E+03,
     + 6.7070E+03, 6.7161E+03, 6.7253E+03, 6.7345E+03, 6.7436E+03,
     +  6.7528E+03, 6.7620E+03, 6.7712E+03, 6.7804E+03, 6.7896E+03,
     + 6.7988E+03, 6.8080E+03, 6.8172E+03, 6.8265E+03, 6.8357E+03,
     +  6.8450E+03, 6.8542E+03, 6.8635E+03, 6.8727E+03, 6.8820E+03,
     + 6.8913E+03, 6.9006E+03, 6.9099E+03, 6.9192E+03, 6.9285E+03,
     +  6.9378E+03, 6.9472E+03, 6.9565E+03, 6.9658E+03, 6.9752E+03,
     + 6.9845E+03, 6.9939E+03, 7.0033E+03, 7.0126E+03, 7.0220E+03,
     +  7.0314E+03, 7.0408E+03, 7.0502E+03, 7.0596E+03, 7.0690E+03,
     + 7.0785E+03, 7.0879E+03, 7.0973E+03, 7.1068E+03, 7.1163E+03,
     +  7.1257E+03, 7.1352E+03, 7.1447E+03, 7.1541E+03, 7.1636E+03,
     + 7.1731E+03, 7.1826E+03, 7.1922E+03, 7.2017E+03, 7.2112E+03,
     +  7.2207E+03, 7.2303E+03, 7.2398E+03, 7.2494E+03, 7.2589E+03,
     + 7.2685E+03, 7.2781E+03, 7.2877E+03, 7.2973E+03, 7.3069E+03,
     +  7.3165E+03, 7.3261E+03, 7.3357E+03, 7.3453E+03, 7.3550E+03,
     + 7.3646E+03, 7.3743E+03, 7.3839E+03, 7.3936E+03, 7.4033E+03,
     +  7.4129E+03, 7.4226E+03, 7.4323E+03, 7.4420E+03, 7.4517E+03,
     + 7.4614E+03, 7.4712E+03, 7.4809E+03, 7.4906E+03, 7.5004E+03,
     +  7.5101E+03, 7.5199E+03, 7.5296E+03, 7.5394E+03, 7.5492E+03,
     + 7.5590E+03, 7.5688E+03, 7.5786E+03, 7.5884E+03, 7.5982E+03,
     +  7.6080E+03, 7.6179E+03, 7.6277E+03, 7.6375E+03, 7.6474E+03,
     + 7.6573E+03, 7.6671E+03, 7.6770E+03, 7.6869E+03, 7.6968E+03,
     +  7.7067E+03, 7.7166E+03, 7.7265E+03, 7.7364E+03, 7.7463E+03,
     + 7.7563E+03, 7.7662E+03, 7.7762E+03, 7.7861E+03, 7.7961E+03,
     +  7.8060E+03, 7.8160E+03, 7.8260E+03, 7.8360E+03, 7.8460E+03,
     + 7.8560E+03, 7.8660E+03, 7.8760E+03, 7.8861E+03, 7.8961E+03,
     +  7.9061E+03, 7.9162E+03, 7.9263E+03, 7.9363E+03, 7.9464E+03,
     + 7.9565E+03, 7.9666E+03, 7.9767E+03, 7.9868E+03, 7.9969E+03,
     +  8.0070E+03, 8.0171E+03, 8.0272E+03, 8.0374E+03, 8.0475E+03,
     + 8.0577E+03, 8.0678E+03, 8.0780E+03, 8.0882E+03, 8.0984E+03,
     +  8.1086E+03, 8.1188E+03, 8.1290E+03, 8.1392E+03, 8.1494E+03,
     + 8.1596E+03, 8.1699E+03, 8.1801E+03, 8.1904E+03, 8.2006E+03,
     +  8.2109E+03, 8.2211E+03, 8.2314E+03, 8.2417E+03, 8.2520E+03,
     + 8.2623E+03, 8.2726E+03, 8.2829E+03, 8.2933E+03, 8.3036E+03,
     +  8.3139E+03, 8.3243E+03, 8.3346E+03, 8.3450E+03, 8.3554E+03,
     + 8.3657E+03, 8.3761E+03, 8.3865E+03, 8.3969E+03, 8.4073E+03,
     +  8.4177E+03, 8.4281E+03, 8.4386E+03, 8.4490E+03, 8.4594E+03,
     + 8.4699E+03, 8.4804E+03, 8.4908E+03, 8.5013E+03, 8.5118E+03,
     +  8.5223E+03, 8.5328E+03, 8.5433E+03, 8.5538E+03, 8.5643E+03,
     + 8.5748E+03, 8.5853E+03, 8.5959E+03, 8.6064E+03, 8.6170E+03,
     +  8.6275E+03, 8.6381E+03, 8.6487E+03, 8.6593E+03, 8.6699E+03,
     + 8.6805E+03, 8.6911E+03, 8.7017E+03, 8.7123E+03, 8.7229E+03,
     +  8.7336E+03, 8.7442E+03, 8.7549E+03, 8.7655E+03, 8.7762E+03,
     + 8.7869E+03, 8.7976E+03, 8.8082E+03, 8.8189E+03, 8.8296E+03,
     +  8.8404E+03, 8.8511E+03, 8.8618E+03, 8.8725E+03, 8.8833E+03,
     + 8.8940E+03, 8.9048E+03, 8.9155E+03, 8.9263E+03, 8.9371E+03,
     +  8.9479E+03, 8.9587E+03, 8.9695E+03, 8.9803E+03, 8.9911E+03,
     + 9.0019E+03, 9.0128E+03, 9.0236E+03, 9.0345E+03, 9.0453E+03,
     +  9.0562E+03, 9.0670E+03, 9.0779E+03, 9.0888E+03, 9.0997E+03,
     + 9.1106E+03, 9.1215E+03, 9.1324E+03, 9.1433E+03, 9.1543E+03,
     +  9.1652E+03, 9.1762E+03, 9.1871E+03, 9.1981E+03, 9.2090E+03,
     + 9.2200E+03, 9.2310E+03, 9.2420E+03, 9.2530E+03, 9.2640E+03,
     +  9.2750E+03, 9.2860E+03, 9.2971E+03, 9.3081E+03, 9.3191E+03,
     + 9.3302E+03, 9.3413E+03, 9.3523E+03, 9.3634E+03, 9.3745E+03,
     +  9.3856E+03, 9.3967E+03, 9.4078E+03, 9.4189E+03, 9.4300E+03,
     + 9.4411E+03, 9.4523E+03, 9.4634E+03, 9.4746E+03, 9.4857E+03,
     +  9.4969E+03, 9.5081E+03, 9.5192E+03, 9.5304E+03, 9.5416E+03,
     + 9.5528E+03, 9.5640E+03, 9.5753E+03, 9.5865E+03, 9.5977E+03,
     +  9.6090E+03, 9.6202E+03, 9.6315E+03, 9.6427E+03, 9.6540E+03,
     + 9.6653E+03, 9.6766E+03, 9.6879E+03, 9.6992E+03, 9.7105E+03,
     +  9.7218E+03, 9.7331E+03, 9.7445E+03, 9.7558E+03, 9.7671E+03,
     + 9.7785E+03, 9.7899E+03, 9.8012E+03, 9.8126E+03, 9.8240E+03,
     +  9.8354E+03, 9.8468E+03, 9.8582E+03, 9.8696E+03, 9.8811E+03,
     + 9.8925E+03, 9.9039E+03, 9.9154E+03, 9.9268E+03, 9.9383E+03,
     +  9.9498E+03, 9.9613E+03, 9.9727E+03, 9.9842E+03, 9.9957E+03,
     + 1.0007E+04, 1.0019E+04, 1.0030E+04, 1.0042E+04, 1.0053E+04,
     +  1.0065E+04, 1.0076E+04, 1.0088E+04, 1.0100E+04, 1.0111E+04,
     + 1.0123E+04, 1.0134E+04, 1.0146E+04, 1.0158E+04, 1.0169E+04,
     +  1.0181E+04, 1.0192E+04, 1.0204E+04, 1.0216E+04, 1.0227E+04,
     + 1.0239E+04, 1.0251E+04, 1.0262E+04, 1.0274E+04, 1.0286E+04,
     +  1.0298E+04, 1.0309E+04, 1.0321E+04, 1.0333E+04, 1.0344E+04,
     + 1.0356E+04, 1.0368E+04, 1.0380E+04, 1.0391E+04, 1.0403E+04,
     +  1.0415E+04, 1.0427E+04, 1.0439E+04, 1.0450E+04, 1.0462E+04,
     + 1.0474E+04, 1.0486E+04, 1.0498E+04, 1.0509E+04, 1.0521E+04,
     +  1.0533E+04, 1.0545E+04, 1.0557E+04, 1.0569E+04, 1.0581E+04,
     + 1.0593E+04, 1.0605E+04, 1.0616E+04, 1.0628E+04, 1.0640E+04,
     +  1.0652E+04, 1.0664E+04, 1.0676E+04, 1.0688E+04, 1.0700E+04,
     + 1.0712E+04, 1.0724E+04, 1.0736E+04, 1.0748E+04, 1.0760E+04,
     +  1.0772E+04, 1.0784E+04, 1.0796E+04, 1.0808E+04, 1.0820E+04,
     + 1.0832E+04, 1.0844E+04, 1.0856E+04, 1.0868E+04, 1.0880E+04,
     +  1.0893E+04, 1.0905E+04, 1.0917E+04, 1.0929E+04, 1.0941E+04,
     + 1.0953E+04, 1.0965E+04, 1.0977E+04, 1.0990E+04, 1.1002E+04,
     +  1.1014E+04, 1.1026E+04, 1.1038E+04, 1.1050E+04, 1.1063E+04,
     + 1.1075E+04, 1.1087E+04, 1.1099E+04, 1.1112E+04, 1.1124E+04,
     +  1.1136E+04, 1.1148E+04, 1.1161E+04, 1.1173E+04, 1.1185E+04,
     + 1.1197E+04, 1.1210E+04, 1.1222E+04, 1.1234E+04, 1.1247E+04,
     +  1.1259E+04, 1.1271E+04, 1.1284E+04, 1.1296E+04, 1.1308E+04,
     + 1.1321E+04, 1.1333E+04, 1.1345E+04, 1.1358E+04, 1.1370E+04,
     +  1.1383E+04, 1.1395E+04, 1.1407E+04, 1.1420E+04, 1.1432E+04,
     + 1.1445E+04, 1.1457E+04, 1.1470E+04, 1.1482E+04, 1.1495E+04,
     +  1.1507E+04, 1.1520E+04, 1.1532E+04, 1.1545E+04, 1.1557E+04,
     + 1.1570E+04, 1.1582E+04, 1.1595E+04, 1.1607E+04, 1.1620E+04,
     +  1.1632E+04, 1.1645E+04, 1.1658E+04, 1.1670E+04, 1.1683E+04,
     + 1.1695E+04, 1.1708E+04, 1.1721E+04, 1.1733E+04, 1.1746E+04,
     +  1.1758E+04, 1.1771E+04, 1.1784E+04, 1.1796E+04, 1.1809E+04,
     + 1.1822E+04, 1.1834E+04, 1.1847E+04, 1.1860E+04, 1.1873E+04,
     +  1.1885E+04, 1.1898E+04, 1.1911E+04, 1.1923E+04, 1.1936E+04,
     + 1.1949E+04, 1.1962E+04, 1.1975E+04, 1.1987E+04, 1.2000E+04,
     +  1.2013E+04, 1.2026E+04, 1.2039E+04, 1.2051E+04, 1.2064E+04,
     + 1.2077E+04, 1.2090E+04, 1.2103E+04, 1.2116E+04, 1.2128E+04,
     +  1.2141E+04, 1.2154E+04, 1.2167E+04, 1.2180E+04, 1.2193E+04,
     + 1.2206E+04, 1.2219E+04, 1.2232E+04, 1.2245E+04, 1.2258E+04,
     +  1.2271E+04, 1.2284E+04, 1.2296E+04, 1.2309E+04, 1.2322E+04,
     + 1.2335E+04, 1.2348E+04, 1.2361E+04, 1.2375E+04, 1.2388E+04,
     +  1.2401E+04, 1.2414E+04, 1.2427E+04, 1.2440E+04, 1.2453E+04,
     + 1.2466E+04, 1.2479E+04, 1.2492E+04, 1.2505E+04, 1.2518E+04,
     +  1.2531E+04, 1.2545E+04, 1.2558E+04, 1.2571E+04, 1.2584E+04,
     + 1.2597E+04, 1.2610E+04, 1.2623E+04, 1.2637E+04, 1.2650E+04,
     +  1.2663E+04, 1.2676E+04, 1.2689E+04, 1.2703E+04, 1.2716E+04,
     + 1.2729E+04, 1.2742E+04, 1.2756E+04, 1.2769E+04, 1.2782E+04,
     +  1.2795E+04, 1.2809E+04, 1.2822E+04, 1.2835E+04, 1.2849E+04,
     + 1.2862E+04, 1.2875E+04, 1.2889E+04, 1.2902E+04, 1.2915E+04,
     +  1.2929E+04, 1.2942E+04, 1.2955E+04, 1.2969E+04, 1.2982E+04,
     + 1.2996E+04, 1.3009E+04, 1.3022E+04, 1.3036E+04, 1.3049E+04,
     +  1.3063E+04, 1.3076E+04, 1.3090E+04, 1.3103E+04, 1.3116E+04,
     + 1.3130E+04, 1.3143E+04, 1.3157E+04, 1.3170E+04, 1.3184E+04,
     +  1.3197E+04, 1.3211E+04, 1.3225E+04, 1.3238E+04, 1.3252E+04,
     + 1.3265E+04, 1.3279E+04, 1.3292E+04, 1.3306E+04, 1.3319E+04,
     +  1.3333E+04, 1.3347E+04, 1.3360E+04, 1.3374E+04, 1.3388E+04,
     + 1.3401E+04, 1.3415E+04, 1.3428E+04, 1.3442E+04, 1.3456E+04,
     +  1.3470E+04, 1.3483E+04, 1.3497E+04, 1.3511E+04, 1.3524E+04,
     + 1.3538E+04, 1.3552E+04, 1.3565E+04, 1.3579E+04, 1.3593E+04,
     +  1.3607E+04, 1.3621E+04, 1.3634E+04, 1.3648E+04, 1.3662E+04,
     + 1.3676E+04, 1.3689E+04, 1.3703E+04, 1.3717E+04, 1.3731E+04,
     +  1.3745E+04, 1.3759E+04, 1.3772E+04, 1.3786E+04, 1.3800E+04,
     + 1.3814E+04, 1.3828E+04, 1.3842E+04, 1.3856E+04, 1.3870E+04,
     +  1.3884E+04, 1.3898E+04, 1.3911E+04, 1.3925E+04, 1.3939E+04,
     + 1.3953E+04, 1.3967E+04, 1.3981E+04, 1.3995E+04, 1.4009E+04,
     +  1.4023E+04, 1.4037E+04, 1.4051E+04, 1.4065E+04, 1.4079E+04,
     + 1.4093E+04, 1.4107E+04, 1.4122E+04, 1.4136E+04, 1.4150E+04,
     +  1.4164E+04, 1.4178E+04, 1.4192E+04, 1.4206E+04, 1.4220E+04,
     + 1.4234E+04, 1.4248E+04, 1.4263E+04, 1.4277E+04, 1.4291E+04,
     +  1.4305E+04, 1.4319E+04, 1.4333E+04, 1.4348E+04, 1.4362E+04,
     + 1.4376E+04, 1.4390E+04, 1.4404E+04, 1.4419E+04, 1.4433E+04,
     +  1.4447E+04, 1.4461E+04, 1.4476E+04, 1.4490E+04, 1.4504E+04,
     + 1.4518E+04, 1.4533E+04, 1.4547E+04, 1.4561E+04, 1.4576E+04,
     +  1.4590E+04, 1.4604E+04, 1.4619E+04, 1.4633E+04, 1.4647E+04,
     + 1.4662E+04, 1.4676E+04, 1.4690E+04, 1.4705E+04, 1.4719E+04,
     +  1.4734E+04, 1.4748E+04, 1.4763E+04, 1.4777E+04, 1.4791E+04,
     + 1.4806E+04, 1.4820E+04, 1.4835E+04, 1.4849E+04, 1.4864E+04,
     +  1.4878E+04, 1.4893E+04, 1.4907E+04, 1.4922E+04, 1.4936E+04,
     + 1.4951E+04, 1.4965E+04, 1.4980E+04, 1.4994E+04, 1.5009E+04,
     +  1.5024E+04, 1.5038E+04, 1.5053E+04, 1.5067E+04, 1.5082E+04,
     + 1.5097E+04, 1.5111E+04, 1.5126E+04, 1.5140E+04, 1.5155E+04,
     +  1.5170E+04, 1.5184E+04, 1.5199E+04, 1.5214E+04, 1.5228E+04,
     + 1.5243E+04, 1.5258E+04, 1.5273E+04, 1.5287E+04, 1.5302E+04,
     +  1.5317E+04, 1.5331E+04, 1.5346E+04, 1.5361E+04, 1.5376E+04,
     + 1.5390E+04, 1.5405E+04, 1.5420E+04, 1.5435E+04, 1.5450E+04,
     +  1.5464E+04, 1.5479E+04, 1.5494E+04, 1.5509E+04, 1.5524E+04,
     + 1.5539E+04, 1.5554E+04, 1.5568E+04, 1.5583E+04, 1.5598E+04,
     +  1.5613E+04, 1.5628E+04, 1.5643E+04, 1.5658E+04, 1.5673E+04,
     + 1.5688E+04, 1.5703E+04, 1.5718E+04, 1.5733E+04, 1.5748E+04,
     +  1.5763E+04, 1.5777E+04, 1.5792E+04, 1.5807E+04, 1.5823E+04,
     + 1.5838E+04, 1.5853E+04, 1.5868E+04, 1.5883E+04, 1.5898E+04,
     +  1.5913E+04, 1.5928E+04, 1.5943E+04, 1.5958E+04, 1.5973E+04,
     + 1.5988E+04, 1.6003E+04, 1.6018E+04, 1.6034E+04, 1.6049E+04,
     +  1.6064E+04, 1.6079E+04, 1.6094E+04, 1.6109E+04, 1.6124E+04,
     + 1.6140E+04, 1.6155E+04, 1.6170E+04, 1.6185E+04, 1.6200E+04,
     +  1.6216E+04, 1.6231E+04, 1.6246E+04, 1.6261E+04, 1.6277E+04,
     + 1.6292E+04, 1.6307E+04, 1.6323E+04, 1.6338E+04, 1.6353E+04,
     +  1.6368E+04, 1.6384E+04, 1.6399E+04, 1.6414E+04, 1.6430E+04,
     + 1.6445E+04, 1.6460E+04, 1.6476E+04, 1.6491E+04, 1.6507E+04,
     +  1.6522E+04, 1.6537E+04, 1.6553E+04, 1.6568E+04, 1.6584E+04,
     + 1.6599E+04, 1.6614E+04, 1.6630E+04, 1.6645E+04, 1.6661E+04,
     +  1.6676E+04, 1.6692E+04, 1.6707E+04, 1.6723E+04, 1.6738E+04,
     + 1.6754E+04, 1.6769E+04, 1.6785E+04, 1.6800E+04, 1.6816E+04,
     +  1.6831E+04, 1.6847E+04, 1.6863E+04, 1.6878E+04, 1.6894E+04,
     + 1.6909E+04, 1.6925E+04, 1.6941E+04, 1.6956E+04, 1.6972E+04,
     +  1.6987E+04, 1.7003E+04, 1.7019E+04, 1.7034E+04, 1.7050E+04,
     + 1.7066E+04, 1.7081E+04, 1.7097E+04, 1.7113E+04, 1.7129E+04,
     +  1.7144E+04, 1.7160E+04, 1.7176E+04, 1.7192E+04, 1.7207E+04,
     + 1.7223E+04, 1.7239E+04, 1.7255E+04, 1.7270E+04, 1.7286E+04,
     +  1.7302E+04, 1.7318E+04, 1.7334E+04, 1.7349E+04, 1.7365E+04,
     + 1.7381E+04, 1.7397E+04, 1.7413E+04, 1.7429E+04, 1.7445E+04,
     +  1.7460E+04, 1.7476E+04, 1.7492E+04, 1.7508E+04, 1.7524E+04,
     + 1.7540E+04, 1.7556E+04, 1.7572E+04, 1.7588E+04, 1.7604E+04,
     +  1.7620E+04, 1.7636E+04, 1.7652E+04, 1.7668E+04, 1.7684E+04,
     + 1.7700E+04, 1.7716E+04, 1.7732E+04, 1.7748E+04, 1.7764E+04,
     +  1.7780E+04, 1.7796E+04, 1.7812E+04, 1.7828E+04, 1.7844E+04,
     + 1.7860E+04, 1.7876E+04, 1.7892E+04, 1.7909E+04, 1.7925E+04,
     +  1.7941E+04, 1.7957E+04, 1.7973E+04, 1.7989E+04, 1.8005E+04,
     + 1.8022E+04, 1.8038E+04, 1.8054E+04, 1.8070E+04, 1.8086E+04,
     +  1.8103E+04, 1.8119E+04, 1.8135E+04, 1.8151E+04, 1.8168E+04,
     + 1.8184E+04, 1.8200E+04, 1.8216E+04, 1.8233E+04, 1.8249E+04,
     +  1.8265E+04, 1.8282E+04, 1.8298E+04, 1.8314E+04, 1.8331E+04,
     + 1.8347E+04, 1.8363E+04, 1.8380E+04, 1.8396E+04, 1.8412E+04,
     +  1.8429E+04, 1.8445E+04, 1.8461E+04, 1.8478E+04, 1.8494E+04,
     + 1.8511E+04, 1.8527E+04, 1.8544E+04, 1.8560E+04, 1.8577E+04,
     +  1.8593E+04, 1.8609E+04, 1.8626E+04, 1.8642E+04, 1.8659E+04,
     + 1.8675E+04, 1.8692E+04, 1.8708E+04, 1.8725E+04, 1.8742E+04,
     +  1.8758E+04, 1.8775E+04, 1.8791E+04, 1.8808E+04, 1.8824E+04,
     + 1.8841E+04, 1.8858E+04, 1.8874E+04, 1.8891E+04, 1.8907E+04,
     +  1.8924E+04, 1.8941E+04, 1.8957E+04, 1.8974E+04, 1.8991E+04,
     + 1.9007E+04, 1.9024E+04, 1.9041E+04, 1.9057E+04, 1.9074E+04,
     +  1.9091E+04, 1.9108E+04, 1.9124E+04, 1.9141E+04, 1.9158E+04,
     + 1.9174E+04, 1.9191E+04, 1.9208E+04, 1.9225E+04, 1.9242E+04,
     +  1.9258E+04, 1.9275E+04, 1.9292E+04, 1.9309E+04, 1.9326E+04,
     + 1.9342E+04, 1.9359E+04, 1.9376E+04, 1.9393E+04, 1.9410E+04,
     +  1.9427E+04, 1.9444E+04, 1.9461E+04, 1.9477E+04, 1.9494E+04,
     + 1.9511E+04, 1.9528E+04, 1.9545E+04, 1.9562E+04, 1.9579E+04,
     +  1.9596E+04, 1.9613E+04, 1.9630E+04, 1.9647E+04, 1.9664E+04,
     + 1.9681E+04, 1.9698E+04, 1.9715E+04, 1.9732E+04, 1.9749E+04,
     +  1.9766E+04, 1.9783E+04, 1.9800E+04, 1.9817E+04, 1.9834E+04,
     + 1.9852E+04, 1.9869E+04, 1.9886E+04, 1.9903E+04, 1.9920E+04,
     +  1.9937E+04, 1.9954E+04, 1.9971E+04, 1.9989E+04, 2.0006E+04,
     + 2.0023E+04, 2.0040E+04, 2.0057E+04, 2.0074E+04, 2.0092E+04,
     +  2.0109E+04, 2.0126E+04, 2.0143E+04, 2.0161E+04, 2.0178E+04,
     + 2.0195E+04, 2.0212E+04, 2.0230E+04, 2.0247E+04, 2.0264E+04,
     +  2.0281E+04, 2.0299E+04, 2.0316E+04, 2.0333E+04, 2.0351E+04,
     + 2.0368E+04, 2.0385E+04, 2.0403E+04, 2.0420E+04, 2.0437E+04,
     +  2.0455E+04, 2.0472E+04, 2.0490E+04, 2.0507E+04, 2.0524E+04,
     + 2.0542E+04, 2.0559E+04, 2.0577E+04, 2.0594E+04, 2.0612E+04,
     +  2.0629E+04, 2.0647E+04, 2.0664E+04, 2.0682E+04, 2.0699E+04,
     + 2.0717E+04, 2.0734E+04, 2.0752E+04, 2.0769E+04, 2.0787E+04,
     +  2.0804E+04, 2.0822E+04, 2.0839E+04, 2.0857E+04, 2.0874E+04,
     + 2.0892E+04, 2.0910E+04, 2.0927E+04, 2.0945E+04, 2.0962E+04,
     +  2.0980E+04, 2.0998E+04, 2.1015E+04, 2.1033E+04, 2.1051E+04,
     + 2.1068E+04, 2.1086E+04, 2.1104E+04, 2.1121E+04, 2.1139E+04,
     +  2.1157E+04, 2.1175E+04, 2.1192E+04, 2.1210E+04, 2.1228E+04,
     + 2.1246E+04, 2.1263E+04, 2.1281E+04, 2.1299E+04, 2.1317E+04,
     +  2.1334E+04, 2.1352E+04, 2.1370E+04, 2.1388E+04, 2.1406E+04,
     + 2.1424E+04, 2.1441E+04, 2.1459E+04, 2.1477E+04, 2.1495E+04,
     +  2.1513E+04, 2.1531E+04, 2.1549E+04, 2.1567E+04, 2.1584E+04,
     + 2.1602E+04, 2.1620E+04, 2.1638E+04, 2.1656E+04, 2.1674E+04,
     +  2.1692E+04, 2.1710E+04, 2.1728E+04, 2.1746E+04, 2.1764E+04,
     + 2.1782E+04, 2.1800E+04, 2.1818E+04, 2.1836E+04, 2.1854E+04,
     +  2.1872E+04, 2.1890E+04, 2.1908E+04, 2.1926E+04, 2.1945E+04,
     + 2.1963E+04, 2.1981E+04, 2.1999E+04, 2.2017E+04, 2.2035E+04,
     +  2.2053E+04, 2.2071E+04, 2.2089E+04, 2.2108E+04, 2.2126E+04,
     + 2.2144E+04, 2.2162E+04, 2.2180E+04, 2.2199E+04, 2.2217E+04,
     +  2.2235E+04, 2.2253E+04, 2.2271E+04, 2.2290E+04, 2.2308E+04,
     + 2.2326E+04, 2.2344E+04, 2.2363E+04, 2.2381E+04, 2.2399E+04,
     +  2.2418E+04, 2.2436E+04, 2.2454E+04, 2.2472E+04, 2.2491E+04/
      data (QofT( 2,J),J=2409,4510)/ 2.2509E+04, 2.2527E+04,
     +  2.2546E+04, 2.2564E+04, 2.2583E+04, 2.2601E+04, 2.2619E+04,
     +  2.2638E+04,
     + 2.2656E+04, 2.2675E+04, 2.2693E+04, 2.2711E+04, 2.2730E+04,
     +  2.2748E+04, 2.2767E+04, 2.2785E+04, 2.2804E+04, 2.2822E+04,
     + 2.2841E+04, 2.2859E+04, 2.2878E+04, 2.2896E+04, 2.2915E+04,
     +  2.2933E+04, 2.2952E+04, 2.2970E+04, 2.2989E+04, 2.3007E+04,
     + 2.3026E+04, 2.3045E+04, 2.3063E+04, 2.3082E+04, 2.3100E+04,
     +  2.3119E+04, 2.3138E+04, 2.3156E+04, 2.3175E+04, 2.3194E+04,
     + 2.3212E+04, 2.3231E+04, 2.3250E+04, 2.3268E+04, 2.3287E+04,
     +  2.3306E+04, 2.3324E+04, 2.3343E+04, 2.3362E+04, 2.3380E+04,
     + 2.3399E+04, 2.3418E+04, 2.3437E+04, 2.3455E+04, 2.3474E+04,
     +  2.3493E+04, 2.3512E+04, 2.3531E+04, 2.3549E+04, 2.3568E+04,
     + 2.3587E+04, 2.3606E+04, 2.3625E+04, 2.3644E+04, 2.3662E+04,
     +  2.3681E+04, 2.3700E+04, 2.3719E+04, 2.3738E+04, 2.3757E+04,
     + 2.3776E+04, 2.3795E+04, 2.3814E+04, 2.3833E+04, 2.3851E+04,
     +  2.3870E+04, 2.3889E+04, 2.3908E+04, 2.3927E+04, 2.3946E+04,
     + 2.3965E+04, 2.3984E+04, 2.4003E+04, 2.4022E+04, 2.4041E+04,
     +  2.4060E+04, 2.4079E+04, 2.4098E+04, 2.4117E+04, 2.4137E+04,
     + 2.4156E+04, 2.4175E+04, 2.4194E+04, 2.4213E+04, 2.4232E+04,
     +  2.4251E+04, 2.4270E+04, 2.4289E+04, 2.4309E+04, 2.4328E+04,
     + 2.4347E+04, 2.4366E+04, 2.4385E+04, 2.4404E+04, 2.4424E+04,
     +  2.4443E+04, 2.4462E+04, 2.4481E+04, 2.4500E+04, 2.4520E+04,
     + 2.4539E+04, 2.4558E+04, 2.4577E+04, 2.4597E+04, 2.4616E+04,
     +  2.4635E+04, 2.4654E+04, 2.4674E+04, 2.4693E+04, 2.4712E+04,
     + 2.4732E+04, 2.4751E+04, 2.4770E+04, 2.4790E+04, 2.4809E+04,
     +  2.4828E+04, 2.4848E+04, 2.4867E+04, 2.4887E+04, 2.4906E+04,
     + 2.4925E+04, 2.4945E+04, 2.4964E+04, 2.4984E+04, 2.5003E+04,
     +  2.5023E+04, 2.5042E+04, 2.5061E+04, 2.5081E+04, 2.5100E+04,
     + 2.5120E+04, 2.5139E+04, 2.5159E+04, 2.5178E+04, 2.5198E+04,
     +  2.5217E+04, 2.5237E+04, 2.5257E+04, 2.5276E+04, 2.5296E+04,
     + 2.5315E+04, 2.5335E+04, 2.5354E+04, 2.5374E+04, 2.5394E+04,
     +  2.5413E+04, 2.5433E+04, 2.5452E+04, 2.5472E+04, 2.5492E+04,
     + 2.5511E+04, 2.5531E+04, 2.5551E+04, 2.5570E+04, 2.5590E+04,
     +  2.5610E+04, 2.5630E+04, 2.5649E+04, 2.5669E+04, 2.5689E+04,
     + 2.5708E+04, 2.5728E+04, 2.5748E+04, 2.5768E+04, 2.5788E+04,
     +  2.5807E+04, 2.5827E+04, 2.5847E+04, 2.5867E+04, 2.5886E+04,
     + 2.5906E+04, 2.5926E+04, 2.5946E+04, 2.5966E+04, 2.5986E+04,
     +  2.6006E+04, 2.6025E+04, 2.6045E+04, 2.6065E+04, 2.6085E+04,
     + 2.6105E+04, 2.6125E+04, 2.6145E+04, 2.6165E+04, 2.6185E+04,
     +  2.6205E+04, 2.6225E+04, 2.6245E+04, 2.6265E+04, 2.6285E+04,
     + 2.6305E+04, 2.6325E+04, 2.6345E+04, 2.6365E+04, 2.6385E+04,
     +  2.6405E+04, 2.6425E+04, 2.6445E+04, 2.6465E+04, 2.6485E+04,
     + 2.6505E+04, 2.6525E+04, 2.6545E+04, 2.6565E+04, 2.6585E+04,
     +  2.6605E+04, 2.6625E+04, 2.6646E+04, 2.6666E+04, 2.6686E+04,
     + 2.6706E+04, 2.6726E+04, 2.6746E+04, 2.6767E+04, 2.6787E+04,
     +  2.6807E+04, 2.6827E+04, 2.6847E+04, 2.6868E+04, 2.6888E+04,
     + 2.6908E+04, 2.6928E+04, 2.6949E+04, 2.6969E+04, 2.6989E+04,
     +  2.7009E+04, 2.7030E+04, 2.7050E+04, 2.7070E+04, 2.7091E+04,
     + 2.7111E+04, 2.7131E+04, 2.7152E+04, 2.7172E+04, 2.7192E+04,
     +  2.7213E+04, 2.7233E+04, 2.7253E+04, 2.7274E+04, 2.7294E+04,
     + 2.7315E+04, 2.7335E+04, 2.7355E+04, 2.7376E+04, 2.7396E+04,
     +  2.7417E+04, 2.7437E+04, 2.7458E+04, 2.7478E+04, 2.7499E+04,
     + 2.7519E+04, 2.7540E+04, 2.7560E+04, 2.7581E+04, 2.7601E+04,
     +  2.7622E+04, 2.7642E+04, 2.7663E+04, 2.7683E+04, 2.7704E+04,
     + 2.7724E+04, 2.7745E+04, 2.7766E+04, 2.7786E+04, 2.7807E+04,
     +  2.7827E+04, 2.7848E+04, 2.7869E+04, 2.7889E+04, 2.7910E+04,
     + 2.7931E+04, 2.7951E+04, 2.7972E+04, 2.7993E+04, 2.8013E+04,
     +  2.8034E+04, 2.8055E+04, 2.8075E+04, 2.8096E+04, 2.8117E+04,
     + 2.8138E+04, 2.8158E+04, 2.8179E+04, 2.8200E+04, 2.8221E+04,
     +  2.8241E+04, 2.8262E+04, 2.8283E+04, 2.8304E+04, 2.8325E+04,
     + 2.8345E+04, 2.8366E+04, 2.8387E+04, 2.8408E+04, 2.8429E+04,
     +  2.8450E+04, 2.8470E+04, 2.8491E+04, 2.8512E+04, 2.8533E+04,
     + 2.8554E+04, 2.8575E+04, 2.8596E+04, 2.8617E+04, 2.8638E+04,
     +  2.8659E+04, 2.8680E+04, 2.8701E+04, 2.8722E+04, 2.8743E+04,
     + 2.8764E+04, 2.8785E+04, 2.8806E+04, 2.8827E+04, 2.8848E+04,
     +  2.8869E+04, 2.8890E+04, 2.8911E+04, 2.8932E+04, 2.8953E+04,
     + 2.8974E+04, 2.8995E+04, 2.9016E+04, 2.9037E+04, 2.9058E+04,
     +  2.9079E+04, 2.9100E+04, 2.9122E+04, 2.9143E+04, 2.9164E+04,
     + 2.9185E+04, 2.9206E+04, 2.9227E+04, 2.9248E+04, 2.9270E+04,
     +  2.9291E+04, 2.9312E+04, 2.9333E+04, 2.9354E+04, 2.9376E+04,
     + 2.9397E+04, 2.9418E+04, 2.9439E+04, 2.9461E+04, 2.9482E+04,
     +  2.9503E+04, 2.9524E+04, 2.9546E+04, 2.9567E+04, 2.9588E+04,
     + 2.9610E+04, 2.9631E+04, 2.9652E+04, 2.9674E+04, 2.9695E+04,
     +  2.9716E+04, 2.9738E+04, 2.9759E+04, 2.9781E+04, 2.9802E+04,
     + 2.9823E+04, 2.9845E+04, 2.9866E+04, 2.9888E+04, 2.9909E+04,
     +  2.9930E+04, 2.9952E+04, 2.9973E+04, 2.9995E+04, 3.0016E+04,
     + 3.0038E+04, 3.0059E+04, 3.0081E+04, 3.0102E+04, 3.0124E+04,
     +  3.0145E+04, 3.0167E+04, 3.0188E+04, 3.0210E+04, 3.0231E+04,
     + 3.0253E+04, 3.0275E+04, 3.0296E+04, 3.0318E+04, 3.0339E+04,
     +  3.0361E+04, 3.0383E+04, 3.0404E+04, 3.0426E+04, 3.0447E+04,
     + 3.0469E+04, 3.0491E+04, 3.0512E+04, 3.0534E+04, 3.0556E+04,
     +  3.0577E+04, 3.0599E+04, 3.0621E+04, 3.0643E+04, 3.0664E+04,
     + 3.0686E+04, 3.0708E+04, 3.0729E+04, 3.0751E+04, 3.0773E+04,
     +  3.0795E+04, 3.0817E+04, 3.0838E+04, 3.0860E+04, 3.0882E+04,
     + 3.0904E+04, 3.0926E+04, 3.0947E+04, 3.0969E+04, 3.0991E+04,
     +  3.1013E+04, 3.1035E+04, 3.1057E+04, 3.1078E+04, 3.1100E+04,
     + 3.1122E+04, 3.1144E+04, 3.1166E+04, 3.1188E+04, 3.1210E+04,
     +  3.1232E+04, 3.1254E+04, 3.1276E+04, 3.1298E+04, 3.1320E+04,
     + 3.1342E+04, 3.1364E+04, 3.1386E+04, 3.1408E+04, 3.1430E+04,
     +  3.1452E+04, 3.1474E+04, 3.1496E+04, 3.1518E+04, 3.1540E+04,
     + 3.1562E+04, 3.1584E+04, 3.1606E+04, 3.1628E+04, 3.1650E+04,
     +  3.1672E+04, 3.1694E+04, 3.1716E+04, 3.1739E+04, 3.1761E+04,
     + 3.1783E+04, 3.1805E+04, 3.1827E+04, 3.1849E+04, 3.1871E+04,
     +  3.1894E+04, 3.1916E+04, 3.1938E+04, 3.1960E+04, 3.1982E+04,
     + 3.2005E+04, 3.2027E+04, 3.2049E+04, 3.2071E+04, 3.2094E+04,
     +  3.2116E+04, 3.2138E+04, 3.2160E+04, 3.2183E+04, 3.2205E+04,
     + 3.2227E+04, 3.2250E+04, 3.2272E+04, 3.2294E+04, 3.2317E+04,
     +  3.2339E+04, 3.2361E+04, 3.2384E+04, 3.2406E+04, 3.2428E+04,
     + 3.2451E+04, 3.2473E+04, 3.2496E+04, 3.2518E+04, 3.2540E+04,
     +  3.2563E+04, 3.2585E+04, 3.2608E+04, 3.2630E+04, 3.2653E+04,
     + 3.2675E+04, 3.2698E+04, 3.2720E+04, 3.2742E+04, 3.2765E+04,
     +  3.2787E+04, 3.2810E+04, 3.2833E+04, 3.2855E+04, 3.2878E+04,
     + 3.2900E+04, 3.2923E+04, 3.2945E+04, 3.2968E+04, 3.2990E+04,
     +  3.3013E+04, 3.3036E+04, 3.3058E+04, 3.3081E+04, 3.3103E+04,
     + 3.3126E+04, 3.3149E+04, 3.3171E+04, 3.3194E+04, 3.3217E+04,
     +  3.3239E+04, 3.3262E+04, 3.3285E+04, 3.3307E+04, 3.3330E+04,
     + 3.3353E+04, 3.3376E+04, 3.3398E+04, 3.3421E+04, 3.3444E+04,
     +  3.3466E+04, 3.3489E+04, 3.3512E+04, 3.3535E+04, 3.3558E+04,
     + 3.3580E+04, 3.3603E+04, 3.3626E+04, 3.3649E+04, 3.3672E+04,
     +  3.3694E+04, 3.3717E+04, 3.3740E+04, 3.3763E+04, 3.3786E+04,
     + 3.3809E+04, 3.3832E+04, 3.3854E+04, 3.3877E+04, 3.3900E+04,
     +  3.3923E+04, 3.3946E+04, 3.3969E+04, 3.3992E+04, 3.4015E+04,
     + 3.4038E+04, 3.4061E+04, 3.4084E+04, 3.4107E+04, 3.4130E+04,
     +  3.4153E+04, 3.4176E+04, 3.4199E+04, 3.4222E+04, 3.4245E+04,
     + 3.4268E+04, 3.4291E+04, 3.4314E+04, 3.4337E+04, 3.4360E+04,
     +  3.4383E+04, 3.4406E+04, 3.4429E+04, 3.4452E+04, 3.4476E+04,
     + 3.4499E+04, 3.4522E+04, 3.4545E+04, 3.4568E+04, 3.4591E+04,
     +  3.4614E+04, 3.4638E+04, 3.4661E+04, 3.4684E+04, 3.4707E+04,
     + 3.4730E+04, 3.4753E+04, 3.4777E+04, 3.4800E+04, 3.4823E+04,
     +  3.4846E+04, 3.4870E+04, 3.4893E+04, 3.4916E+04, 3.4939E+04,
     + 3.4963E+04, 3.4986E+04, 3.5009E+04, 3.5033E+04, 3.5056E+04,
     +  3.5079E+04, 3.5103E+04, 3.5126E+04, 3.5149E+04, 3.5173E+04,
     + 3.5196E+04, 3.5219E+04, 3.5243E+04, 3.5266E+04, 3.5289E+04,
     +  3.5313E+04, 3.5336E+04, 3.5360E+04, 3.5383E+04, 3.5406E+04,
     + 3.5430E+04, 3.5453E+04, 3.5477E+04, 3.5500E+04, 3.5524E+04,
     +  3.5547E+04, 3.5571E+04, 3.5594E+04, 3.5618E+04, 3.5641E+04,
     + 3.5665E+04, 3.5688E+04, 3.5712E+04, 3.5735E+04, 3.5759E+04,
     +  3.5782E+04, 3.5806E+04, 3.5830E+04, 3.5853E+04, 3.5877E+04,
     + 3.5900E+04, 3.5924E+04, 3.5948E+04, 3.5971E+04, 3.5995E+04,
     +  3.6019E+04, 3.6042E+04, 3.6066E+04, 3.6090E+04, 3.6113E+04,
     + 3.6137E+04, 3.6161E+04, 3.6184E+04, 3.6208E+04, 3.6232E+04,
     +  3.6255E+04, 3.6279E+04, 3.6303E+04, 3.6327E+04, 3.6350E+04,
     + 3.6374E+04, 3.6398E+04, 3.6422E+04, 3.6445E+04, 3.6469E+04,
     +  3.6493E+04, 3.6517E+04, 3.6541E+04, 3.6565E+04, 3.6588E+04,
     + 3.6612E+04, 3.6636E+04, 3.6660E+04, 3.6684E+04, 3.6708E+04,
     +  3.6732E+04, 3.6755E+04, 3.6779E+04, 3.6803E+04, 3.6827E+04,
     + 3.6851E+04, 3.6875E+04, 3.6899E+04, 3.6923E+04, 3.6947E+04,
     +  3.6971E+04, 3.6995E+04, 3.7019E+04, 3.7043E+04, 3.7067E+04,
     + 3.7091E+04, 3.7115E+04, 3.7139E+04, 3.7163E+04, 3.7187E+04,
     +  3.7211E+04, 3.7235E+04, 3.7259E+04, 3.7283E+04, 3.7307E+04,
     + 3.7331E+04, 3.7355E+04, 3.7379E+04, 3.7404E+04, 3.7428E+04,
     +  3.7452E+04, 3.7476E+04, 3.7500E+04, 3.7524E+04, 3.7548E+04,
     + 3.7572E+04, 3.7597E+04, 3.7621E+04, 3.7645E+04, 3.7669E+04,
     +  3.7693E+04, 3.7718E+04, 3.7742E+04, 3.7766E+04, 3.7790E+04,
     + 3.7815E+04, 3.7839E+04, 3.7863E+04, 3.7887E+04, 3.7912E+04,
     +  3.7936E+04, 3.7960E+04, 3.7984E+04, 3.8009E+04, 3.8033E+04,
     + 3.8057E+04, 3.8082E+04, 3.8106E+04, 3.8130E+04, 3.8155E+04,
     +  3.8179E+04, 3.8204E+04, 3.8228E+04, 3.8252E+04, 3.8277E+04,
     + 3.8301E+04, 3.8325E+04, 3.8350E+04, 3.8374E+04, 3.8399E+04,
     +  3.8423E+04, 3.8448E+04, 3.8472E+04, 3.8497E+04, 3.8521E+04,
     + 3.8546E+04, 3.8570E+04, 3.8595E+04, 3.8619E+04, 3.8644E+04,
     +  3.8668E+04, 3.8693E+04, 3.8717E+04, 3.8742E+04, 3.8766E+04,
     + 3.8791E+04, 3.8815E+04, 3.8840E+04, 3.8865E+04, 3.8889E+04,
     +  3.8914E+04, 3.8938E+04, 3.8963E+04, 3.8988E+04, 3.9012E+04,
     + 3.9037E+04, 3.9062E+04, 3.9086E+04, 3.9111E+04, 3.9136E+04,
     +  3.9160E+04, 3.9185E+04, 3.9210E+04, 3.9234E+04, 3.9259E+04,
     + 3.9284E+04, 3.9308E+04, 3.9333E+04, 3.9358E+04, 3.9383E+04,
     +  3.9407E+04, 3.9432E+04, 3.9457E+04, 3.9482E+04, 3.9507E+04,
     + 3.9531E+04, 3.9556E+04, 3.9581E+04, 3.9606E+04, 3.9631E+04,
     +  3.9655E+04, 3.9680E+04, 3.9705E+04, 3.9730E+04, 3.9755E+04,
     + 3.9780E+04, 3.9805E+04, 3.9830E+04, 3.9854E+04, 3.9879E+04,
     +  3.9904E+04, 3.9929E+04, 3.9954E+04, 3.9979E+04, 4.0004E+04,
     + 4.0029E+04, 4.0054E+04, 4.0079E+04, 4.0104E+04, 4.0129E+04,
     +  4.0154E+04, 4.0179E+04, 4.0204E+04, 4.0229E+04, 4.0254E+04,
     + 4.0279E+04, 4.0304E+04, 4.0329E+04, 4.0354E+04, 4.0379E+04,
     +  4.0404E+04, 4.0429E+04, 4.0454E+04, 4.0480E+04, 4.0505E+04,
     + 4.0530E+04, 4.0555E+04, 4.0580E+04, 4.0605E+04, 4.0630E+04,
     +  4.0655E+04, 4.0681E+04, 4.0706E+04, 4.0731E+04, 4.0756E+04,
     + 4.0781E+04, 4.0807E+04, 4.0832E+04, 4.0857E+04, 4.0882E+04,
     +  4.0907E+04, 4.0933E+04, 4.0958E+04, 4.0983E+04, 4.1008E+04,
     + 4.1034E+04, 4.1059E+04, 4.1084E+04, 4.1110E+04, 4.1135E+04,
     +  4.1160E+04, 4.1186E+04, 4.1211E+04, 4.1236E+04, 4.1262E+04,
     + 4.1287E+04, 4.1312E+04, 4.1338E+04, 4.1363E+04, 4.1388E+04,
     +  4.1414E+04, 4.1439E+04, 4.1465E+04, 4.1490E+04, 4.1515E+04,
     + 4.1541E+04, 4.1566E+04, 4.1592E+04, 4.1617E+04, 4.1643E+04,
     +  4.1668E+04, 4.1694E+04, 4.1719E+04, 4.1745E+04, 4.1770E+04,
     + 4.1796E+04, 4.1821E+04, 4.1847E+04, 4.1872E+04, 4.1898E+04,
     +  4.1923E+04, 4.1949E+04, 4.1974E+04, 4.2000E+04, 4.2025E+04,
     + 4.2051E+04, 4.2077E+04, 4.2102E+04, 4.2128E+04, 4.2154E+04,
     +  4.2179E+04, 4.2205E+04, 4.2230E+04, 4.2256E+04, 4.2282E+04,
     + 4.2307E+04, 4.2333E+04, 4.2359E+04, 4.2384E+04, 4.2410E+04,
     +  4.2436E+04, 4.2462E+04, 4.2487E+04, 4.2513E+04, 4.2539E+04,
     + 4.2564E+04, 4.2590E+04, 4.2616E+04, 4.2642E+04, 4.2667E+04,
     +  4.2693E+04, 4.2719E+04, 4.2745E+04, 4.2771E+04, 4.2796E+04,
     + 4.2822E+04, 4.2848E+04, 4.2874E+04, 4.2900E+04, 4.2926E+04,
     +  4.2951E+04, 4.2977E+04, 4.3003E+04, 4.3029E+04, 4.3055E+04,
     + 4.3081E+04, 4.3107E+04, 4.3133E+04, 4.3159E+04, 4.3185E+04,
     +  4.3210E+04, 4.3236E+04, 4.3262E+04, 4.3288E+04, 4.3314E+04,
     + 4.3340E+04, 4.3366E+04, 4.3392E+04, 4.3418E+04, 4.3444E+04,
     +  4.3470E+04, 4.3496E+04, 4.3522E+04, 4.3548E+04, 4.3574E+04,
     + 4.3600E+04, 4.3627E+04, 4.3653E+04, 4.3679E+04, 4.3705E+04,
     +  4.3731E+04, 4.3757E+04, 4.3783E+04, 4.3809E+04, 4.3835E+04,
     + 4.3861E+04, 4.3888E+04, 4.3914E+04, 4.3940E+04, 4.3966E+04,
     +  4.3992E+04, 4.4018E+04, 4.4045E+04, 4.4071E+04, 4.4097E+04,
     + 4.4123E+04, 4.4149E+04, 4.4176E+04, 4.4202E+04, 4.4228E+04,
     +  4.4254E+04, 4.4281E+04, 4.4307E+04, 4.4333E+04, 4.4359E+04,
     + 4.4386E+04, 4.4412E+04, 4.4438E+04, 4.4464E+04, 4.4491E+04,
     +  4.4517E+04, 4.4543E+04, 4.4570E+04, 4.4596E+04, 4.4622E+04,
     + 4.4649E+04, 4.4675E+04, 4.4702E+04, 4.4728E+04, 4.4754E+04,
     +  4.4781E+04, 4.4807E+04, 4.4834E+04, 4.4860E+04, 4.4886E+04,
     + 4.4913E+04, 4.4939E+04, 4.4966E+04, 4.4992E+04, 4.5019E+04,
     +  4.5045E+04, 4.5072E+04, 4.5098E+04, 4.5125E+04, 4.5151E+04,
     + 4.5178E+04, 4.5204E+04, 4.5231E+04, 4.5257E+04, 4.5284E+04,
     +  4.5310E+04, 4.5337E+04, 4.5363E+04, 4.5390E+04, 4.5417E+04,
     + 4.5443E+04, 4.5470E+04, 4.5496E+04, 4.5523E+04, 4.5550E+04,
     +  4.5576E+04, 4.5603E+04, 4.5630E+04, 4.5656E+04, 4.5683E+04,
     + 4.5710E+04, 4.5736E+04, 4.5763E+04, 4.5790E+04, 4.5816E+04,
     +  4.5843E+04, 4.5870E+04, 4.5896E+04, 4.5923E+04, 4.5950E+04,
     + 4.5977E+04, 4.6003E+04, 4.6030E+04, 4.6057E+04, 4.6084E+04,
     +  4.6110E+04, 4.6137E+04, 4.6164E+04, 4.6191E+04, 4.6218E+04,
     + 4.6244E+04, 4.6271E+04, 4.6298E+04, 4.6325E+04, 4.6352E+04,
     +  4.6379E+04, 4.6406E+04, 4.6432E+04, 4.6459E+04, 4.6486E+04,
     + 4.6513E+04, 4.6540E+04, 4.6567E+04, 4.6594E+04, 4.6621E+04,
     +  4.6648E+04, 4.6675E+04, 4.6702E+04, 4.6729E+04, 4.6756E+04,
     + 4.6782E+04, 4.6809E+04, 4.6836E+04, 4.6863E+04, 4.6890E+04,
     +  4.6917E+04, 4.6944E+04, 4.6971E+04, 4.6999E+04, 4.7026E+04,
     + 4.7053E+04, 4.7080E+04, 4.7107E+04, 4.7134E+04, 4.7161E+04,
     +  4.7188E+04, 4.7215E+04, 4.7242E+04, 4.7269E+04, 4.7296E+04,
     + 4.7323E+04, 4.7351E+04, 4.7378E+04, 4.7405E+04, 4.7432E+04,
     +  4.7459E+04, 4.7486E+04, 4.7514E+04, 4.7541E+04, 4.7568E+04,
     + 4.7595E+04, 4.7622E+04, 4.7650E+04, 4.7677E+04, 4.7704E+04,
     +  4.7731E+04, 4.7758E+04, 4.7786E+04, 4.7813E+04, 4.7840E+04,
     + 4.7868E+04, 4.7895E+04, 4.7922E+04, 4.7949E+04, 4.7977E+04,
     +  4.8004E+04, 4.8031E+04, 4.8059E+04, 4.8086E+04, 4.8113E+04,
     + 4.8141E+04, 4.8168E+04, 4.8195E+04, 4.8223E+04, 4.8250E+04,
     +  4.8278E+04, 4.8305E+04, 4.8332E+04, 4.8360E+04, 4.8387E+04,
     + 4.8415E+04, 4.8442E+04, 4.8469E+04, 4.8497E+04, 4.8524E+04,
     +  4.8552E+04, 4.8579E+04, 4.8607E+04, 4.8634E+04, 4.8662E+04,
     + 4.8689E+04, 4.8717E+04, 4.8744E+04, 4.8772E+04, 4.8799E+04,
     +  4.8827E+04, 4.8854E+04, 4.8882E+04, 4.8909E+04, 4.8937E+04,
     + 4.8965E+04, 4.8992E+04, 4.9020E+04, 4.9047E+04, 4.9075E+04,
     +  4.9103E+04, 4.9130E+04, 4.9158E+04, 4.9185E+04, 4.9213E+04,
     + 4.9241E+04, 4.9268E+04, 4.9296E+04, 4.9324E+04, 4.9351E+04,
     +  4.9379E+04, 4.9407E+04, 4.9435E+04, 4.9462E+04, 4.9490E+04,
     + 4.9518E+04, 4.9545E+04, 4.9573E+04, 4.9601E+04, 4.9629E+04,
     +  4.9656E+04, 4.9684E+04, 4.9712E+04, 4.9740E+04, 4.9767E+04,
     + 4.9795E+04, 4.9823E+04, 4.9851E+04, 4.9879E+04, 4.9907E+04,
     +  4.9934E+04, 4.9962E+04, 4.9990E+04, 5.0018E+04, 5.0046E+04,
     + 5.0074E+04, 5.0102E+04, 5.0129E+04, 5.0157E+04, 5.0185E+04,
     +  5.0213E+04, 5.0241E+04, 5.0269E+04, 5.0297E+04, 5.0325E+04,
     + 5.0353E+04, 5.0381E+04, 5.0409E+04, 5.0437E+04, 5.0465E+04,
     +  5.0493E+04, 5.0521E+04, 5.0549E+04, 5.0577E+04, 5.0605E+04,
     + 5.0633E+04, 5.0661E+04, 5.0689E+04, 5.0717E+04, 5.0745E+04,
     +  5.0773E+04, 5.0801E+04, 5.0829E+04, 5.0857E+04, 5.0885E+04,
     + 5.0913E+04, 5.0941E+04, 5.0969E+04, 5.0998E+04, 5.1026E+04,
     +  5.1054E+04, 5.1082E+04, 5.1110E+04, 5.1138E+04, 5.1166E+04,
     + 5.1195E+04, 5.1223E+04, 5.1251E+04, 5.1279E+04, 5.1307E+04,
     +  5.1336E+04, 5.1364E+04, 5.1392E+04, 5.1420E+04, 5.1448E+04,
     + 5.1477E+04, 5.1505E+04, 5.1533E+04, 5.1561E+04, 5.1590E+04,
     +  5.1618E+04, 5.1646E+04, 5.1675E+04, 5.1703E+04, 5.1731E+04,
     + 5.1759E+04, 5.1788E+04, 5.1816E+04, 5.1844E+04, 5.1873E+04,
     +  5.1901E+04, 5.1930E+04, 5.1958E+04, 5.1986E+04, 5.2015E+04,
     + 5.2043E+04, 5.2071E+04, 5.2100E+04, 5.2128E+04, 5.2157E+04,
     +  5.2185E+04, 5.2214E+04, 5.2242E+04, 5.2270E+04, 5.2299E+04,
     + 5.2327E+04, 5.2356E+04, 5.2384E+04, 5.2413E+04, 5.2441E+04,
     +  5.2470E+04, 5.2498E+04, 5.2527E+04, 5.2555E+04, 5.2584E+04,
     + 5.2612E+04, 5.2641E+04, 5.2669E+04, 5.2698E+04, 5.2727E+04,
     +  5.2755E+04, 5.2784E+04, 5.2812E+04, 5.2841E+04, 5.2869E+04,
     + 5.2898E+04, 5.2927E+04, 5.2955E+04, 5.2984E+04, 5.3013E+04,
     +  5.3041E+04, 5.3070E+04, 5.3098E+04, 5.3127E+04, 5.3156E+04,
     + 5.3185E+04, 5.3213E+04, 5.3242E+04, 5.3271E+04, 5.3299E+04,
     +  5.3328E+04, 5.3357E+04, 5.3385E+04, 5.3414E+04, 5.3443E+04,
     + 5.3472E+04, 5.3500E+04, 5.3529E+04, 5.3558E+04, 5.3587E+04,
     +  5.3616E+04, 5.3644E+04, 5.3673E+04, 5.3702E+04, 5.3731E+04,
     + 5.3760E+04, 5.3788E+04, 5.3817E+04, 5.3846E+04, 5.3875E+04,
     +  5.3904E+04, 5.3933E+04, 5.3962E+04, 5.3990E+04, 5.4019E+04,
     + 5.4048E+04, 5.4077E+04, 5.4106E+04, 5.4135E+04, 5.4164E+04,
     +  5.4193E+04, 5.4222E+04, 5.4251E+04, 5.4280E+04, 5.4309E+04,
     + 5.4338E+04, 5.4367E+04, 5.4396E+04, 5.4425E+04, 5.4454E+04,
     +  5.4483E+04, 5.4512E+04, 5.4541E+04, 5.4570E+04, 5.4599E+04,
     + 5.4628E+04, 5.4657E+04, 5.4686E+04, 5.4715E+04, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        67
      data (QofT( 3,J),J=1,2408)/ 2.0923E+01, 2.6875E+01, 3.4025E+01,
     +  4.1754E+01, 4.9834E+01, 5.8137E+01, 6.6592E+01, 7.5152E+01,
     + 8.3788E+01, 9.2481E+01, 1.0122E+02, 1.0999E+02, 1.1879E+02,
     +  1.2760E+02, 1.3644E+02, 1.4529E+02, 1.5416E+02, 1.6303E+02,
     + 1.7192E+02, 1.8081E+02, 1.9861E+02, 2.1644E+02, 2.3428E+02,
     +  2.5213E+02, 2.7000E+02, 2.8787E+02, 3.0575E+02, 3.2364E+02,
     + 3.4153E+02, 3.5942E+02, 3.7732E+02, 3.9523E+02, 4.1313E+02,
     +  4.3104E+02, 4.4896E+02, 4.6687E+02, 4.8479E+02, 5.0270E+02,
     + 5.2062E+02, 5.3854E+02, 5.5647E+02, 5.7439E+02, 5.9231E+02,
     +  6.1024E+02, 6.2817E+02, 6.4610E+02, 6.6402E+02, 6.8195E+02,
     + 6.9988E+02, 7.1782E+02, 7.3575E+02, 7.5368E+02, 7.7161E+02,
     +  7.8955E+02, 8.0748E+02, 8.2542E+02, 8.4335E+02, 8.6129E+02,
     + 8.7922E+02, 8.9716E+02, 9.1510E+02, 9.3304E+02, 9.5098E+02,
     +  9.6891E+02, 9.8685E+02, 1.0048E+03, 1.0227E+03, 1.0407E+03,
     + 1.0586E+03, 1.0766E+03, 1.0945E+03, 1.1124E+03, 1.1304E+03,
     +  1.1483E+03, 1.1663E+03, 1.1842E+03, 1.2022E+03, 1.2201E+03,
     + 1.2380E+03, 1.2560E+03, 1.2739E+03, 1.2919E+03, 1.3098E+03,
     +  1.3278E+03, 1.3457E+03, 1.3637E+03, 1.3816E+03, 1.3996E+03,
     + 1.4175E+03, 1.4355E+03, 1.4534E+03, 1.4714E+03, 1.4893E+03,
     +  1.5073E+03, 1.5252E+03, 1.5432E+03, 1.5611E+03, 1.5791E+03,
     + 1.5970E+03, 1.6150E+03, 1.6329E+03, 1.6509E+03, 1.6688E+03,
     +  1.6868E+03, 1.7047E+03, 1.7227E+03, 1.7406E+03, 1.7586E+03,
     + 1.7766E+03, 1.7945E+03, 1.8125E+03, 1.8304E+03, 1.8484E+03,
     +  1.8663E+03, 1.8843E+03, 1.9023E+03, 1.9202E+03, 1.9382E+03,
     + 1.9562E+03, 1.9741E+03, 1.9921E+03, 2.0101E+03, 2.0280E+03,
     +  2.0460E+03, 2.0640E+03, 2.0819E+03, 2.0999E+03, 2.1179E+03,
     + 2.1359E+03, 2.1538E+03, 2.1718E+03, 2.1898E+03, 2.2078E+03,
     +  2.2258E+03, 2.2437E+03, 2.2617E+03, 2.2797E+03, 2.2977E+03,
     + 2.3157E+03, 2.3337E+03, 2.3517E+03, 2.3697E+03, 2.3877E+03,
     +  2.4057E+03, 2.4237E+03, 2.4417E+03, 2.4597E+03, 2.4778E+03,
     + 2.4958E+03, 2.5138E+03, 2.5318E+03, 2.5499E+03, 2.5679E+03,
     +  2.5859E+03, 2.6040E+03, 2.6220E+03, 2.6401E+03, 2.6581E+03,
     + 2.6762E+03, 2.6942E+03, 2.7123E+03, 2.7304E+03, 2.7485E+03,
     +  2.7665E+03, 2.7846E+03, 2.8027E+03, 2.8208E+03, 2.8389E+03,
     + 2.8570E+03, 2.8751E+03, 2.8932E+03, 2.9114E+03, 2.9295E+03,
     +  2.9476E+03, 2.9658E+03, 2.9839E+03, 3.0021E+03, 3.0202E+03,
     + 3.0384E+03, 3.0566E+03, 3.0748E+03, 3.0929E+03, 3.1111E+03,
     +  3.1293E+03, 3.1476E+03, 3.1658E+03, 3.1840E+03, 3.2022E+03,
     + 3.2205E+03, 3.2387E+03, 3.2570E+03, 3.2753E+03, 3.2935E+03,
     +  3.3118E+03, 3.3301E+03, 3.3484E+03, 3.3668E+03, 3.3851E+03,
     + 3.4034E+03, 3.4218E+03, 3.4401E+03, 3.4585E+03, 3.4769E+03,
     +  3.4952E+03, 3.5136E+03, 3.5320E+03, 3.5505E+03, 3.5689E+03,
     + 3.5873E+03, 3.6058E+03, 3.6243E+03, 3.6427E+03, 3.6612E+03,
     +  3.6797E+03, 3.6982E+03, 3.7168E+03, 3.7353E+03, 3.7538E+03,
     + 3.7724E+03, 3.7910E+03, 3.8096E+03, 3.8282E+03, 3.8468E+03,
     +  3.8654E+03, 3.8840E+03, 3.9027E+03, 3.9214E+03, 3.9401E+03,
     + 3.9588E+03, 3.9775E+03, 3.9962E+03, 4.0149E+03, 4.0337E+03,
     +  4.0525E+03, 4.0712E+03, 4.0900E+03, 4.1089E+03, 4.1277E+03,
     + 4.1465E+03, 4.1654E+03, 4.1843E+03, 4.2032E+03, 4.2221E+03,
     +  4.2410E+03, 4.2599E+03, 4.2789E+03, 4.2979E+03, 4.3169E+03,
     + 4.3359E+03, 4.3549E+03, 4.3739E+03, 4.3930E+03, 4.4121E+03,
     +  4.4312E+03, 4.4503E+03, 4.4694E+03, 4.4886E+03, 4.5077E+03,
     + 4.5269E+03, 4.5461E+03, 4.5653E+03, 4.5846E+03, 4.6038E+03,
     +  4.6231E+03, 4.6424E+03, 4.6617E+03, 4.6810E+03, 4.7004E+03,
     + 4.7198E+03, 4.7392E+03, 4.7586E+03, 4.7780E+03, 4.7974E+03,
     +  4.8169E+03, 4.8364E+03, 4.8559E+03, 4.8754E+03, 4.8950E+03,
     + 4.9146E+03, 4.9341E+03, 4.9538E+03, 4.9734E+03, 4.9930E+03,
     +  5.0127E+03, 5.0324E+03, 5.0521E+03, 5.0719E+03, 5.0916E+03,
     + 5.1114E+03, 5.1312E+03, 5.1510E+03, 5.1708E+03, 5.1907E+03,
     +  5.2106E+03, 5.2305E+03, 5.2504E+03, 5.2704E+03, 5.2904E+03,
     + 5.3104E+03, 5.3304E+03, 5.3504E+03, 5.3705E+03, 5.3906E+03,
     +  5.4107E+03, 5.4308E+03, 5.4510E+03, 5.4711E+03, 5.4913E+03,
     + 5.5116E+03, 5.5318E+03, 5.5521E+03, 5.5724E+03, 5.5927E+03,
     +  5.6130E+03, 5.6334E+03, 5.6538E+03, 5.6742E+03, 5.6946E+03,
     + 5.7151E+03, 5.7356E+03, 5.7561E+03, 5.7766E+03, 5.7972E+03,
     +  5.8178E+03, 5.8384E+03, 5.8590E+03, 5.8797E+03, 5.9003E+03,
     + 5.9211E+03, 5.9418E+03, 5.9625E+03, 5.9833E+03, 6.0041E+03,
     +  6.0249E+03, 6.0458E+03, 6.0667E+03, 6.0876E+03, 6.1085E+03,
     + 6.1295E+03, 6.1505E+03, 6.1715E+03, 6.1925E+03, 6.2135E+03,
     +  6.2346E+03, 6.2557E+03, 6.2769E+03, 6.2980E+03, 6.3192E+03,
     + 6.3404E+03, 6.3617E+03, 6.3829E+03, 6.4042E+03, 6.4255E+03,
     +  6.4469E+03, 6.4683E+03, 6.4897E+03, 6.5111E+03, 6.5325E+03,
     + 6.5540E+03, 6.5755E+03, 6.5970E+03, 6.6186E+03, 6.6402E+03,
     +  6.6618E+03, 6.6834E+03, 6.7051E+03, 6.7268E+03, 6.7485E+03,
     + 6.7703E+03, 6.7920E+03, 6.8138E+03, 6.8357E+03, 6.8575E+03,
     +  6.8794E+03, 6.9013E+03, 6.9232E+03, 6.9452E+03, 6.9672E+03,
     + 6.9892E+03, 7.0113E+03, 7.0333E+03, 7.0554E+03, 7.0776E+03,
     +  7.0997E+03, 7.1219E+03, 7.1441E+03, 7.1664E+03, 7.1887E+03,
     + 7.2110E+03, 7.2333E+03, 7.2556E+03, 7.2780E+03, 7.3004E+03,
     +  7.3229E+03, 7.3453E+03, 7.3678E+03, 7.3904E+03, 7.4129E+03,
     + 7.4355E+03, 7.4581E+03, 7.4808E+03, 7.5034E+03, 7.5261E+03,
     +  7.5489E+03, 7.5716E+03, 7.5944E+03, 7.6172E+03, 7.6400E+03,
     + 7.6629E+03, 7.6858E+03, 7.7087E+03, 7.7317E+03, 7.7547E+03,
     +  7.7777E+03, 7.8008E+03, 7.8238E+03, 7.8469E+03, 7.8701E+03,
     + 7.8932E+03, 7.9164E+03, 7.9396E+03, 7.9629E+03, 7.9862E+03,
     +  8.0095E+03, 8.0328E+03, 8.0562E+03, 8.0796E+03, 8.1030E+03,
     + 8.1265E+03, 8.1500E+03, 8.1735E+03, 8.1970E+03, 8.2206E+03,
     +  8.2442E+03, 8.2678E+03, 8.2915E+03, 8.3152E+03, 8.3389E+03,
     + 8.3627E+03, 8.3865E+03, 8.4103E+03, 8.4341E+03, 8.4580E+03,
     +  8.4819E+03, 8.5058E+03, 8.5298E+03, 8.5538E+03, 8.5778E+03,
     + 8.6019E+03, 8.6260E+03, 8.6501E+03, 8.6742E+03, 8.6984E+03,
     +  8.7226E+03, 8.7469E+03, 8.7711E+03, 8.7954E+03, 8.8198E+03,
     + 8.8441E+03, 8.8685E+03, 8.8929E+03, 8.9174E+03, 8.9419E+03,
     +  8.9664E+03, 8.9909E+03, 9.0155E+03, 9.0401E+03, 9.0648E+03,
     + 9.0894E+03, 9.1141E+03, 9.1388E+03, 9.1636E+03, 9.1884E+03,
     +  9.2132E+03, 9.2381E+03, 9.2630E+03, 9.2879E+03, 9.3128E+03,
     + 9.3378E+03, 9.3628E+03, 9.3878E+03, 9.4129E+03, 9.4380E+03,
     +  9.4631E+03, 9.4883E+03, 9.5135E+03, 9.5387E+03, 9.5640E+03,
     + 9.5893E+03, 9.6146E+03, 9.6399E+03, 9.6653E+03, 9.6907E+03,
     +  9.7162E+03, 9.7416E+03, 9.7672E+03, 9.7927E+03, 9.8183E+03,
     + 9.8439E+03, 9.8695E+03, 9.8952E+03, 9.9209E+03, 9.9466E+03,
     +  9.9723E+03, 9.9981E+03, 1.0024E+04, 1.0050E+04, 1.0076E+04,
     + 1.0102E+04, 1.0128E+04, 1.0154E+04, 1.0180E+04, 1.0206E+04,
     +  1.0232E+04, 1.0258E+04, 1.0284E+04, 1.0310E+04, 1.0336E+04,
     + 1.0363E+04, 1.0389E+04, 1.0415E+04, 1.0441E+04, 1.0468E+04,
     +  1.0494E+04, 1.0521E+04, 1.0547E+04, 1.0574E+04, 1.0600E+04,
     + 1.0627E+04, 1.0653E+04, 1.0680E+04, 1.0706E+04, 1.0733E+04,
     +  1.0760E+04, 1.0787E+04, 1.0813E+04, 1.0840E+04, 1.0867E+04,
     + 1.0894E+04, 1.0921E+04, 1.0948E+04, 1.0975E+04, 1.1002E+04,
     +  1.1029E+04, 1.1056E+04, 1.1083E+04, 1.1110E+04, 1.1137E+04,
     + 1.1164E+04, 1.1192E+04, 1.1219E+04, 1.1246E+04, 1.1274E+04,
     +  1.1301E+04, 1.1328E+04, 1.1356E+04, 1.1383E+04, 1.1411E+04,
     + 1.1438E+04, 1.1466E+04, 1.1493E+04, 1.1521E+04, 1.1549E+04,
     +  1.1576E+04, 1.1604E+04, 1.1632E+04, 1.1660E+04, 1.1687E+04,
     + 1.1715E+04, 1.1743E+04, 1.1771E+04, 1.1799E+04, 1.1827E+04,
     +  1.1855E+04, 1.1883E+04, 1.1911E+04, 1.1939E+04, 1.1967E+04,
     + 1.1995E+04, 1.2024E+04, 1.2052E+04, 1.2080E+04, 1.2108E+04,
     +  1.2137E+04, 1.2165E+04, 1.2194E+04, 1.2222E+04, 1.2250E+04,
     + 1.2279E+04, 1.2308E+04, 1.2336E+04, 1.2365E+04, 1.2393E+04,
     +  1.2422E+04, 1.2451E+04, 1.2479E+04, 1.2508E+04, 1.2537E+04,
     + 1.2566E+04, 1.2595E+04, 1.2624E+04, 1.2653E+04, 1.2681E+04,
     +  1.2710E+04, 1.2740E+04, 1.2769E+04, 1.2798E+04, 1.2827E+04,
     + 1.2856E+04, 1.2885E+04, 1.2914E+04, 1.2944E+04, 1.2973E+04,
     +  1.3002E+04, 1.3032E+04, 1.3061E+04, 1.3090E+04, 1.3120E+04,
     + 1.3149E+04, 1.3179E+04, 1.3209E+04, 1.3238E+04, 1.3268E+04,
     +  1.3297E+04, 1.3327E+04, 1.3357E+04, 1.3387E+04, 1.3416E+04,
     + 1.3446E+04, 1.3476E+04, 1.3506E+04, 1.3536E+04, 1.3566E+04,
     +  1.3596E+04, 1.3626E+04, 1.3656E+04, 1.3686E+04, 1.3716E+04,
     + 1.3746E+04, 1.3777E+04, 1.3807E+04, 1.3837E+04, 1.3867E+04,
     +  1.3898E+04, 1.3928E+04, 1.3958E+04, 1.3989E+04, 1.4019E+04,
     + 1.4050E+04, 1.4080E+04, 1.4111E+04, 1.4142E+04, 1.4172E+04,
     +  1.4203E+04, 1.4234E+04, 1.4264E+04, 1.4295E+04, 1.4326E+04,
     + 1.4357E+04, 1.4388E+04, 1.4418E+04, 1.4449E+04, 1.4480E+04,
     +  1.4511E+04, 1.4542E+04, 1.4573E+04, 1.4605E+04, 1.4636E+04,
     + 1.4667E+04, 1.4698E+04, 1.4729E+04, 1.4761E+04, 1.4792E+04,
     +  1.4823E+04, 1.4855E+04, 1.4886E+04, 1.4918E+04, 1.4949E+04,
     + 1.4981E+04, 1.5012E+04, 1.5044E+04, 1.5075E+04, 1.5107E+04,
     +  1.5139E+04, 1.5170E+04, 1.5202E+04, 1.5234E+04, 1.5266E+04,
     + 1.5298E+04, 1.5329E+04, 1.5361E+04, 1.5393E+04, 1.5425E+04,
     +  1.5457E+04, 1.5489E+04, 1.5521E+04, 1.5554E+04, 1.5586E+04,
     + 1.5618E+04, 1.5650E+04, 1.5682E+04, 1.5715E+04, 1.5747E+04,
     +  1.5779E+04, 1.5812E+04, 1.5844E+04, 1.5877E+04, 1.5909E+04,
     + 1.5942E+04, 1.5974E+04, 1.6007E+04, 1.6040E+04, 1.6072E+04,
     +  1.6105E+04, 1.6138E+04, 1.6171E+04, 1.6203E+04, 1.6236E+04,
     + 1.6269E+04, 1.6302E+04, 1.6335E+04, 1.6368E+04, 1.6401E+04,
     +  1.6434E+04, 1.6467E+04, 1.6500E+04, 1.6533E+04, 1.6567E+04,
     + 1.6600E+04, 1.6633E+04, 1.6666E+04, 1.6700E+04, 1.6733E+04,
     +  1.6766E+04, 1.6800E+04, 1.6833E+04, 1.6867E+04, 1.6900E+04,
     + 1.6934E+04, 1.6968E+04, 1.7001E+04, 1.7035E+04, 1.7069E+04,
     +  1.7102E+04, 1.7136E+04, 1.7170E+04, 1.7204E+04, 1.7238E+04,
     + 1.7272E+04, 1.7306E+04, 1.7340E+04, 1.7374E+04, 1.7408E+04,
     +  1.7442E+04, 1.7476E+04, 1.7510E+04, 1.7544E+04, 1.7578E+04,
     + 1.7613E+04, 1.7647E+04, 1.7681E+04, 1.7716E+04, 1.7750E+04,
     +  1.7785E+04, 1.7819E+04, 1.7854E+04, 1.7888E+04, 1.7923E+04,
     + 1.7957E+04, 1.7992E+04, 1.8027E+04, 1.8061E+04, 1.8096E+04,
     +  1.8131E+04, 1.8166E+04, 1.8201E+04, 1.8236E+04, 1.8270E+04,
     + 1.8305E+04, 1.8340E+04, 1.8375E+04, 1.8411E+04, 1.8446E+04,
     +  1.8481E+04, 1.8516E+04, 1.8551E+04, 1.8586E+04, 1.8622E+04,
     + 1.8657E+04, 1.8692E+04, 1.8728E+04, 1.8763E+04, 1.8799E+04,
     +  1.8834E+04, 1.8870E+04, 1.8905E+04, 1.8941E+04, 1.8976E+04,
     + 1.9012E+04, 1.9048E+04, 1.9084E+04, 1.9119E+04, 1.9155E+04,
     +  1.9191E+04, 1.9227E+04, 1.9263E+04, 1.9299E+04, 1.9335E+04,
     + 1.9371E+04, 1.9407E+04, 1.9443E+04, 1.9479E+04, 1.9515E+04,
     +  1.9551E+04, 1.9588E+04, 1.9624E+04, 1.9660E+04, 1.9697E+04,
     + 1.9733E+04, 1.9769E+04, 1.9806E+04, 1.9842E+04, 1.9879E+04,
     +  1.9915E+04, 1.9952E+04, 1.9989E+04, 2.0025E+04, 2.0062E+04,
     + 2.0099E+04, 2.0136E+04, 2.0172E+04, 2.0209E+04, 2.0246E+04,
     +  2.0283E+04, 2.0320E+04, 2.0357E+04, 2.0394E+04, 2.0431E+04,
     + 2.0468E+04, 2.0505E+04, 2.0542E+04, 2.0580E+04, 2.0617E+04,
     +  2.0654E+04, 2.0691E+04, 2.0729E+04, 2.0766E+04, 2.0804E+04,
     + 2.0841E+04, 2.0879E+04, 2.0916E+04, 2.0954E+04, 2.0991E+04,
     +  2.1029E+04, 2.1066E+04, 2.1104E+04, 2.1142E+04, 2.1180E+04,
     + 2.1218E+04, 2.1255E+04, 2.1293E+04, 2.1331E+04, 2.1369E+04,
     +  2.1407E+04, 2.1445E+04, 2.1483E+04, 2.1521E+04, 2.1559E+04,
     + 2.1598E+04, 2.1636E+04, 2.1674E+04, 2.1712E+04, 2.1751E+04,
     +  2.1789E+04, 2.1827E+04, 2.1866E+04, 2.1904E+04, 2.1943E+04,
     + 2.1981E+04, 2.2020E+04, 2.2059E+04, 2.2097E+04, 2.2136E+04,
     +  2.2175E+04, 2.2213E+04, 2.2252E+04, 2.2291E+04, 2.2330E+04,
     + 2.2369E+04, 2.2408E+04, 2.2447E+04, 2.2486E+04, 2.2525E+04,
     +  2.2564E+04, 2.2603E+04, 2.2642E+04, 2.2681E+04, 2.2720E+04,
     + 2.2760E+04, 2.2799E+04, 2.2838E+04, 2.2878E+04, 2.2917E+04,
     +  2.2956E+04, 2.2996E+04, 2.3035E+04, 2.3075E+04, 2.3115E+04,
     + 2.3154E+04, 2.3194E+04, 2.3234E+04, 2.3273E+04, 2.3313E+04,
     +  2.3353E+04, 2.3393E+04, 2.3433E+04, 2.3473E+04, 2.3513E+04,
     + 2.3553E+04, 2.3593E+04, 2.3633E+04, 2.3673E+04, 2.3713E+04,
     +  2.3753E+04, 2.3793E+04, 2.3834E+04, 2.3874E+04, 2.3914E+04,
     + 2.3954E+04, 2.3995E+04, 2.4035E+04, 2.4076E+04, 2.4116E+04,
     +  2.4157E+04, 2.4197E+04, 2.4238E+04, 2.4279E+04, 2.4319E+04,
     + 2.4360E+04, 2.4401E+04, 2.4442E+04, 2.4483E+04, 2.4523E+04,
     +  2.4564E+04, 2.4605E+04, 2.4646E+04, 2.4687E+04, 2.4728E+04,
     + 2.4769E+04, 2.4811E+04, 2.4852E+04, 2.4893E+04, 2.4934E+04,
     +  2.4976E+04, 2.5017E+04, 2.5058E+04, 2.5100E+04, 2.5141E+04,
     + 2.5183E+04, 2.5224E+04, 2.5266E+04, 2.5307E+04, 2.5349E+04,
     +  2.5390E+04, 2.5432E+04, 2.5474E+04, 2.5516E+04, 2.5557E+04,
     + 2.5599E+04, 2.5641E+04, 2.5683E+04, 2.5725E+04, 2.5767E+04,
     +  2.5809E+04, 2.5851E+04, 2.5893E+04, 2.5935E+04, 2.5978E+04,
     + 2.6020E+04, 2.6062E+04, 2.6104E+04, 2.6147E+04, 2.6189E+04,
     +  2.6231E+04, 2.6274E+04, 2.6316E+04, 2.6359E+04, 2.6401E+04,
     + 2.6444E+04, 2.6487E+04, 2.6529E+04, 2.6572E+04, 2.6615E+04,
     +  2.6658E+04, 2.6700E+04, 2.6743E+04, 2.6786E+04, 2.6829E+04,
     + 2.6872E+04, 2.6915E+04, 2.6958E+04, 2.7001E+04, 2.7044E+04,
     +  2.7087E+04, 2.7131E+04, 2.7174E+04, 2.7217E+04, 2.7260E+04,
     + 2.7304E+04, 2.7347E+04, 2.7391E+04, 2.7434E+04, 2.7478E+04,
     +  2.7521E+04, 2.7565E+04, 2.7608E+04, 2.7652E+04, 2.7696E+04,
     + 2.7739E+04, 2.7783E+04, 2.7827E+04, 2.7871E+04, 2.7915E+04,
     +  2.7959E+04, 2.8003E+04, 2.8047E+04, 2.8091E+04, 2.8135E+04,
     + 2.8179E+04, 2.8223E+04, 2.8267E+04, 2.8311E+04, 2.8356E+04,
     +  2.8400E+04, 2.8444E+04, 2.8489E+04, 2.8533E+04, 2.8577E+04,
     + 2.8622E+04, 2.8666E+04, 2.8711E+04, 2.8756E+04, 2.8800E+04,
     +  2.8845E+04, 2.8890E+04, 2.8934E+04, 2.8979E+04, 2.9024E+04,
     + 2.9069E+04, 2.9114E+04, 2.9159E+04, 2.9204E+04, 2.9249E+04,
     +  2.9294E+04, 2.9339E+04, 2.9384E+04, 2.9429E+04, 2.9475E+04,
     + 2.9520E+04, 2.9565E+04, 2.9610E+04, 2.9656E+04, 2.9701E+04,
     +  2.9747E+04, 2.9792E+04, 2.9838E+04, 2.9883E+04, 2.9929E+04,
     + 2.9974E+04, 3.0020E+04, 3.0066E+04, 3.0112E+04, 3.0157E+04,
     +  3.0203E+04, 3.0249E+04, 3.0295E+04, 3.0341E+04, 3.0387E+04,
     + 3.0433E+04, 3.0479E+04, 3.0525E+04, 3.0571E+04, 3.0618E+04,
     +  3.0664E+04, 3.0710E+04, 3.0756E+04, 3.0803E+04, 3.0849E+04,
     + 3.0895E+04, 3.0942E+04, 3.0988E+04, 3.1035E+04, 3.1081E+04,
     +  3.1128E+04, 3.1175E+04, 3.1221E+04, 3.1268E+04, 3.1315E+04,
     + 3.1362E+04, 3.1409E+04, 3.1455E+04, 3.1502E+04, 3.1549E+04,
     +  3.1596E+04, 3.1643E+04, 3.1690E+04, 3.1738E+04, 3.1785E+04,
     + 3.1832E+04, 3.1879E+04, 3.1926E+04, 3.1974E+04, 3.2021E+04,
     +  3.2069E+04, 3.2116E+04, 3.2163E+04, 3.2211E+04, 3.2258E+04,
     + 3.2306E+04, 3.2354E+04, 3.2401E+04, 3.2449E+04, 3.2497E+04,
     +  3.2545E+04, 3.2592E+04, 3.2640E+04, 3.2688E+04, 3.2736E+04,
     + 3.2784E+04, 3.2832E+04, 3.2880E+04, 3.2928E+04, 3.2976E+04,
     +  3.3025E+04, 3.3073E+04, 3.3121E+04, 3.3169E+04, 3.3218E+04,
     + 3.3266E+04, 3.3315E+04, 3.3363E+04, 3.3411E+04, 3.3460E+04,
     +  3.3509E+04, 3.3557E+04, 3.3606E+04, 3.3655E+04, 3.3703E+04,
     + 3.3752E+04, 3.3801E+04, 3.3850E+04, 3.3899E+04, 3.3948E+04,
     +  3.3996E+04, 3.4046E+04, 3.4095E+04, 3.4144E+04, 3.4193E+04,
     + 3.4242E+04, 3.4291E+04, 3.4340E+04, 3.4390E+04, 3.4439E+04,
     +  3.4488E+04, 3.4538E+04, 3.4587E+04, 3.4637E+04, 3.4686E+04,
     + 3.4736E+04, 3.4785E+04, 3.4835E+04, 3.4885E+04, 3.4934E+04,
     +  3.4984E+04, 3.5034E+04, 3.5084E+04, 3.5134E+04, 3.5184E+04,
     + 3.5234E+04, 3.5284E+04, 3.5334E+04, 3.5384E+04, 3.5434E+04,
     +  3.5484E+04, 3.5534E+04, 3.5585E+04, 3.5635E+04, 3.5685E+04,
     + 3.5736E+04, 3.5786E+04, 3.5836E+04, 3.5887E+04, 3.5937E+04,
     +  3.5988E+04, 3.6039E+04, 3.6089E+04, 3.6140E+04, 3.6191E+04,
     + 3.6241E+04, 3.6292E+04, 3.6343E+04, 3.6394E+04, 3.6445E+04,
     +  3.6496E+04, 3.6547E+04, 3.6598E+04, 3.6649E+04, 3.6700E+04,
     + 3.6751E+04, 3.6803E+04, 3.6854E+04, 3.6905E+04, 3.6956E+04,
     +  3.7008E+04, 3.7059E+04, 3.7111E+04, 3.7162E+04, 3.7214E+04,
     + 3.7265E+04, 3.7317E+04, 3.7369E+04, 3.7420E+04, 3.7472E+04,
     +  3.7524E+04, 3.7576E+04, 3.7628E+04, 3.7679E+04, 3.7731E+04,
     + 3.7783E+04, 3.7835E+04, 3.7887E+04, 3.7940E+04, 3.7992E+04,
     +  3.8044E+04, 3.8096E+04, 3.8148E+04, 3.8201E+04, 3.8253E+04,
     + 3.8305E+04, 3.8358E+04, 3.8410E+04, 3.8463E+04, 3.8515E+04,
     +  3.8568E+04, 3.8621E+04, 3.8673E+04, 3.8726E+04, 3.8779E+04,
     + 3.8832E+04, 3.8884E+04, 3.8937E+04, 3.8990E+04, 3.9043E+04,
     +  3.9096E+04, 3.9149E+04, 3.9202E+04, 3.9255E+04, 3.9309E+04,
     + 3.9362E+04, 3.9415E+04, 3.9468E+04, 3.9522E+04, 3.9575E+04,
     +  3.9629E+04, 3.9682E+04, 3.9735E+04, 3.9789E+04, 3.9843E+04,
     + 3.9896E+04, 3.9950E+04, 4.0004E+04, 4.0057E+04, 4.0111E+04,
     +  4.0165E+04, 4.0219E+04, 4.0273E+04, 4.0327E+04, 4.0381E+04,
     + 4.0435E+04, 4.0489E+04, 4.0543E+04, 4.0597E+04, 4.0651E+04,
     +  4.0705E+04, 4.0760E+04, 4.0814E+04, 4.0868E+04, 4.0923E+04,
     + 4.0977E+04, 4.1032E+04, 4.1086E+04, 4.1141E+04, 4.1195E+04,
     +  4.1250E+04, 4.1305E+04, 4.1360E+04, 4.1414E+04, 4.1469E+04,
     + 4.1524E+04, 4.1579E+04, 4.1634E+04, 4.1689E+04, 4.1744E+04,
     +  4.1799E+04, 4.1854E+04, 4.1909E+04, 4.1964E+04, 4.2020E+04,
     + 4.2075E+04, 4.2130E+04, 4.2186E+04, 4.2241E+04, 4.2296E+04,
     +  4.2352E+04, 4.2407E+04, 4.2463E+04, 4.2519E+04, 4.2574E+04,
     + 4.2630E+04, 4.2686E+04, 4.2742E+04, 4.2797E+04, 4.2853E+04,
     +  4.2909E+04, 4.2965E+04, 4.3021E+04, 4.3077E+04, 4.3133E+04,
     + 4.3189E+04, 4.3245E+04, 4.3302E+04, 4.3358E+04, 4.3414E+04,
     +  4.3470E+04, 4.3527E+04, 4.3583E+04, 4.3640E+04, 4.3696E+04,
     + 4.3753E+04, 4.3809E+04, 4.3866E+04, 4.3923E+04, 4.3979E+04,
     +  4.4036E+04, 4.4093E+04, 4.4150E+04, 4.4206E+04, 4.4263E+04,
     + 4.4320E+04, 4.4377E+04, 4.4434E+04, 4.4491E+04, 4.4549E+04,
     +  4.4606E+04, 4.4663E+04, 4.4720E+04, 4.4777E+04, 4.4835E+04,
     + 4.4892E+04, 4.4950E+04, 4.5007E+04, 4.5065E+04, 4.5122E+04,
     +  4.5180E+04, 4.5237E+04, 4.5295E+04, 4.5353E+04, 4.5410E+04,
     + 4.5468E+04, 4.5526E+04, 4.5584E+04, 4.5642E+04, 4.5700E+04,
     +  4.5758E+04, 4.5816E+04, 4.5874E+04, 4.5932E+04, 4.5990E+04,
     + 4.6049E+04, 4.6107E+04, 4.6165E+04, 4.6223E+04, 4.6282E+04,
     +  4.6340E+04, 4.6399E+04, 4.6457E+04, 4.6516E+04, 4.6574E+04,
     + 4.6633E+04, 4.6692E+04, 4.6751E+04, 4.6809E+04, 4.6868E+04,
     +  4.6927E+04, 4.6986E+04, 4.7045E+04, 4.7104E+04, 4.7163E+04,
     + 4.7222E+04, 4.7281E+04, 4.7340E+04, 4.7399E+04, 4.7459E+04,
     +  4.7518E+04, 4.7577E+04, 4.7637E+04, 4.7696E+04, 4.7756E+04,
     + 4.7815E+04, 4.7875E+04, 4.7934E+04, 4.7994E+04, 4.8053E+04,
     +  4.8113E+04, 4.8173E+04, 4.8233E+04, 4.8293E+04, 4.8352E+04,
     + 4.8412E+04, 4.8472E+04, 4.8532E+04, 4.8592E+04, 4.8653E+04,
     +  4.8713E+04, 4.8773E+04, 4.8833E+04, 4.8893E+04, 4.8954E+04,
     + 4.9014E+04, 4.9074E+04, 4.9135E+04, 4.9195E+04, 4.9256E+04,
     +  4.9316E+04, 4.9377E+04, 4.9438E+04, 4.9498E+04, 4.9559E+04,
     + 4.9620E+04, 4.9681E+04, 4.9742E+04, 4.9803E+04, 4.9864E+04,
     +  4.9925E+04, 4.9986E+04, 5.0047E+04, 5.0108E+04, 5.0169E+04,
     + 5.0230E+04, 5.0292E+04, 5.0353E+04, 5.0414E+04, 5.0476E+04,
     +  5.0537E+04, 5.0599E+04, 5.0660E+04, 5.0722E+04, 5.0783E+04,
     + 5.0845E+04, 5.0907E+04, 5.0968E+04, 5.1030E+04, 5.1092E+04,
     +  5.1154E+04, 5.1216E+04, 5.1278E+04, 5.1340E+04, 5.1402E+04,
     + 5.1464E+04, 5.1526E+04, 5.1588E+04, 5.1650E+04, 5.1713E+04,
     +  5.1775E+04, 5.1837E+04, 5.1900E+04, 5.1962E+04, 5.2025E+04,
     + 5.2087E+04, 5.2150E+04, 5.2212E+04, 5.2275E+04, 5.2338E+04,
     +  5.2400E+04, 5.2463E+04, 5.2526E+04, 5.2589E+04, 5.2652E+04,
     + 5.2715E+04, 5.2778E+04, 5.2841E+04, 5.2904E+04, 5.2967E+04,
     +  5.3030E+04, 5.3094E+04, 5.3157E+04, 5.3220E+04, 5.3283E+04,
     + 5.3347E+04, 5.3410E+04, 5.3474E+04, 5.3537E+04, 5.3601E+04,
     +  5.3665E+04, 5.3728E+04, 5.3792E+04, 5.3856E+04, 5.3919E+04,
     + 5.3983E+04, 5.4047E+04, 5.4111E+04, 5.4175E+04, 5.4239E+04,
     +  5.4303E+04, 5.4367E+04, 5.4431E+04, 5.4496E+04, 5.4560E+04,
     + 5.4624E+04, 5.4688E+04, 5.4753E+04, 5.4817E+04, 5.4882E+04,
     +  5.4946E+04, 5.5011E+04, 5.5075E+04, 5.5140E+04, 5.5205E+04,
     + 5.5269E+04, 5.5334E+04, 5.5399E+04, 5.5464E+04, 5.5529E+04,
     +  5.5594E+04, 5.5659E+04, 5.5724E+04, 5.5789E+04, 5.5854E+04,
     + 5.5919E+04, 5.5984E+04, 5.6049E+04, 5.6115E+04, 5.6180E+04,
     +  5.6245E+04, 5.6311E+04, 5.6376E+04, 5.6442E+04, 5.6507E+04,
     + 5.6573E+04, 5.6639E+04, 5.6704E+04, 5.6770E+04, 5.6836E+04,
     +  5.6902E+04, 5.6967E+04, 5.7033E+04, 5.7099E+04, 5.7165E+04,
     + 5.7231E+04, 5.7297E+04, 5.7364E+04, 5.7430E+04, 5.7496E+04,
     +  5.7562E+04, 5.7629E+04, 5.7695E+04, 5.7761E+04, 5.7828E+04,
     + 5.7894E+04, 5.7961E+04, 5.8027E+04, 5.8094E+04, 5.8161E+04,
     +  5.8227E+04, 5.8294E+04, 5.8361E+04, 5.8428E+04, 5.8495E+04,
     + 5.8562E+04, 5.8629E+04, 5.8696E+04, 5.8763E+04, 5.8830E+04,
     +  5.8897E+04, 5.8964E+04, 5.9031E+04, 5.9099E+04, 5.9166E+04,
     + 5.9233E+04, 5.9301E+04, 5.9368E+04, 5.9436E+04, 5.9503E+04,
     +  5.9571E+04, 5.9639E+04, 5.9706E+04, 5.9774E+04, 5.9842E+04,
     + 5.9910E+04, 5.9978E+04, 6.0046E+04, 6.0114E+04, 6.0182E+04,
     +  6.0250E+04, 6.0318E+04, 6.0386E+04, 6.0454E+04, 6.0522E+04,
     + 6.0591E+04, 6.0659E+04, 6.0727E+04, 6.0796E+04, 6.0864E+04,
     +  6.0933E+04, 6.1001E+04, 6.1070E+04, 6.1138E+04, 6.1207E+04,
     + 6.1276E+04, 6.1345E+04, 6.1413E+04, 6.1482E+04, 6.1551E+04,
     +  6.1620E+04, 6.1689E+04, 6.1758E+04, 6.1827E+04, 6.1896E+04,
     + 6.1966E+04, 6.2035E+04, 6.2104E+04, 6.2173E+04, 6.2243E+04,
     +  6.2312E+04, 6.2382E+04, 6.2451E+04, 6.2521E+04, 6.2590E+04,
     + 6.2660E+04, 6.2729E+04, 6.2799E+04, 6.2869E+04, 6.2939E+04,
     +  6.3009E+04, 6.3078E+04, 6.3148E+04, 6.3218E+04, 6.3288E+04,
     + 6.3359E+04, 6.3429E+04, 6.3499E+04, 6.3569E+04, 6.3639E+04,
     +  6.3710E+04, 6.3780E+04, 6.3850E+04, 6.3921E+04, 6.3991E+04,
     + 6.4062E+04, 6.4132E+04, 6.4203E+04, 6.4274E+04, 6.4344E+04,
     +  6.4415E+04, 6.4486E+04, 6.4557E+04, 6.4628E+04, 6.4699E+04,
     + 6.4770E+04, 6.4841E+04, 6.4912E+04, 6.4983E+04, 6.5054E+04,
     +  6.5125E+04, 6.5197E+04, 6.5268E+04, 6.5339E+04, 6.5411E+04,
     + 6.5482E+04, 6.5554E+04, 6.5625E+04, 6.5697E+04, 6.5768E+04,
     +  6.5840E+04, 6.5912E+04, 6.5983E+04, 6.6055E+04, 6.6127E+04,
     + 6.6199E+04, 6.6271E+04, 6.6343E+04, 6.6415E+04, 6.6487E+04,
     +  6.6559E+04, 6.6631E+04, 6.6704E+04, 6.6776E+04, 6.6848E+04,
     + 6.6920E+04, 6.6993E+04, 6.7065E+04, 6.7138E+04, 6.7210E+04,
     +  6.7283E+04, 6.7356E+04, 6.7428E+04, 6.7501E+04, 6.7574E+04,
     + 6.7646E+04, 6.7719E+04, 6.7792E+04, 6.7865E+04, 6.7938E+04,
     +  6.8011E+04, 6.8084E+04, 6.8157E+04, 6.8231E+04, 6.8304E+04,
     + 6.8377E+04, 6.8450E+04, 6.8524E+04, 6.8597E+04, 6.8671E+04,
     +  6.8744E+04, 6.8818E+04, 6.8891E+04, 6.8965E+04, 6.9039E+04,
     + 6.9112E+04, 6.9186E+04, 6.9260E+04, 6.9334E+04, 6.9408E+04,
     +  6.9482E+04, 6.9556E+04, 6.9630E+04, 6.9704E+04, 6.9778E+04,
     + 6.9852E+04, 6.9926E+04, 7.0001E+04, 7.0075E+04, 7.0149E+04,
     +  7.0224E+04, 7.0298E+04, 7.0373E+04, 7.0447E+04, 7.0522E+04,
     + 7.0596E+04, 7.0671E+04, 7.0746E+04, 7.0821E+04, 7.0895E+04,
     +  7.0970E+04, 7.1045E+04, 7.1120E+04, 7.1195E+04, 7.1270E+04,
     + 7.1345E+04, 7.1421E+04, 7.1496E+04, 7.1571E+04, 7.1646E+04,
     +  7.1722E+04, 7.1797E+04, 7.1872E+04, 7.1948E+04, 7.2023E+04,
     + 7.2099E+04, 7.2175E+04, 7.2250E+04, 7.2326E+04, 7.2402E+04,
     +  7.2478E+04, 7.2553E+04, 7.2629E+04, 7.2705E+04, 7.2781E+04,
     + 7.2857E+04, 7.2933E+04, 7.3009E+04, 7.3086E+04, 7.3162E+04,
     +  7.3238E+04, 7.3314E+04, 7.3391E+04, 7.3467E+04, 7.3544E+04,
     + 7.3620E+04, 7.3697E+04, 7.3773E+04, 7.3850E+04, 7.3927E+04,
     +  7.4003E+04, 7.4080E+04, 7.4157E+04, 7.4234E+04, 7.4311E+04,
     + 7.4388E+04, 7.4465E+04, 7.4542E+04, 7.4619E+04, 7.4696E+04,
     +  7.4773E+04, 7.4850E+04, 7.4928E+04, 7.5005E+04, 7.5082E+04,
     + 7.5160E+04, 7.5237E+04, 7.5315E+04, 7.5392E+04, 7.5470E+04,
     +  7.5547E+04, 7.5625E+04, 7.5703E+04, 7.5781E+04, 7.5859E+04,
     + 7.5936E+04, 7.6014E+04, 7.6092E+04, 7.6170E+04, 7.6248E+04,
     +  7.6327E+04, 7.6405E+04, 7.6483E+04, 7.6561E+04, 7.6640E+04,
     + 7.6718E+04, 7.6796E+04, 7.6875E+04, 7.6953E+04, 7.7032E+04,
     +  7.7110E+04, 7.7189E+04, 7.7268E+04, 7.7346E+04, 7.7425E+04,
     + 7.7504E+04, 7.7583E+04, 7.7662E+04, 7.7741E+04, 7.7820E+04,
     +  7.7899E+04, 7.7978E+04, 7.8057E+04, 7.8136E+04, 7.8215E+04,
     + 7.8295E+04, 7.8374E+04, 7.8453E+04, 7.8533E+04, 7.8612E+04,
     +  7.8692E+04, 7.8771E+04, 7.8851E+04, 7.8931E+04, 7.9010E+04,
     + 7.9090E+04, 7.9170E+04, 7.9250E+04, 7.9330E+04, 7.9410E+04,
     +  7.9490E+04, 7.9570E+04, 7.9650E+04, 7.9730E+04, 7.9810E+04,
     + 7.9890E+04, 7.9971E+04, 8.0051E+04, 8.0131E+04, 8.0212E+04,
     +  8.0292E+04, 8.0373E+04, 8.0453E+04, 8.0534E+04, 8.0614E+04,
     + 8.0695E+04, 8.0776E+04, 8.0857E+04, 8.0937E+04, 8.1018E+04,
     +  8.1099E+04, 8.1180E+04, 8.1261E+04, 8.1342E+04, 8.1423E+04,
     + 8.1505E+04, 8.1586E+04, 8.1667E+04, 8.1748E+04, 8.1830E+04,
     +  8.1911E+04, 8.1992E+04, 8.2074E+04, 8.2156E+04, 8.2237E+04,
     + 8.2319E+04, 8.2400E+04, 8.2482E+04, 8.2564E+04, 8.2646E+04,
     +  8.2728E+04, 8.2809E+04, 8.2891E+04, 8.2973E+04, 8.3056E+04,
     + 8.3138E+04, 8.3220E+04, 8.3302E+04, 8.3384E+04, 8.3466E+04,
     +  8.3549E+04, 8.3631E+04, 8.3714E+04, 8.3796E+04, 8.3879E+04,
     + 8.3961E+04, 8.4044E+04, 8.4126E+04, 8.4209E+04, 8.4292E+04,
     +  8.4375E+04, 8.4458E+04, 8.4541E+04, 8.4623E+04, 8.4706E+04,
     + 8.4790E+04, 8.4873E+04, 8.4956E+04, 8.5039E+04, 8.5122E+04,
     +  8.5205E+04, 8.5289E+04, 8.5372E+04, 8.5456E+04, 8.5539E+04,
     + 8.5623E+04, 8.5706E+04, 8.5790E+04, 8.5873E+04, 8.5957E+04,
     +  8.6041E+04, 8.6125E+04, 8.6209E+04, 8.6292E+04, 8.6376E+04,
     + 8.6460E+04, 8.6544E+04, 8.6628E+04, 8.6713E+04, 8.6797E+04,
     +  8.6881E+04, 8.6965E+04, 8.7050E+04, 8.7134E+04, 8.7218E+04,
     + 8.7303E+04, 8.7387E+04, 8.7472E+04, 8.7557E+04, 8.7641E+04,
     +  8.7726E+04, 8.7811E+04, 8.7895E+04, 8.7980E+04, 8.8065E+04,
     + 8.8150E+04, 8.8235E+04, 8.8320E+04, 8.8405E+04, 8.8490E+04,
     +  8.8575E+04, 8.8661E+04, 8.8746E+04, 8.8831E+04, 8.8917E+04,
     + 8.9002E+04, 8.9088E+04, 8.9173E+04, 8.9259E+04, 8.9344E+04,
     +  8.9430E+04, 8.9516E+04, 8.9601E+04, 8.9687E+04, 8.9773E+04,
     + 8.9859E+04, 8.9945E+04, 9.0031E+04, 9.0117E+04, 9.0203E+04,
     +  9.0289E+04, 9.0375E+04, 9.0461E+04, 9.0548E+04, 9.0634E+04,
     + 9.0720E+04, 9.0807E+04, 9.0893E+04, 9.0980E+04, 9.1066E+04,
     +  9.1153E+04, 9.1239E+04, 9.1326E+04, 9.1413E+04, 9.1500E+04,
     + 9.1586E+04, 9.1673E+04, 9.1760E+04, 9.1847E+04, 9.1934E+04,
     +  9.2021E+04, 9.2109E+04, 9.2196E+04, 9.2283E+04, 9.2370E+04,
     + 9.2457E+04, 9.2545E+04, 9.2632E+04, 9.2720E+04, 9.2807E+04,
     +  9.2895E+04, 9.2982E+04, 9.3070E+04, 9.3158E+04, 9.3245E+04,
     + 9.3333E+04, 9.3421E+04, 9.3509E+04, 9.3597E+04, 9.3685E+04,
     +  9.3773E+04, 9.3861E+04, 9.3949E+04, 9.4037E+04, 9.4126E+04,
     + 9.4214E+04, 9.4302E+04, 9.4390E+04, 9.4479E+04, 9.4567E+04,
     +  9.4656E+04, 9.4744E+04, 9.4833E+04, 9.4922E+04, 9.5010E+04,
     + 9.5099E+04, 9.5188E+04, 9.5277E+04, 9.5366E+04, 9.5455E+04,
     +  9.5544E+04, 9.5633E+04, 9.5722E+04, 9.5811E+04, 9.5900E+04,
     + 9.5989E+04, 9.6078E+04, 9.6168E+04, 9.6257E+04, 9.6347E+04,
     +  9.6436E+04, 9.6526E+04, 9.6615E+04, 9.6705E+04, 9.6794E+04,
     + 9.6884E+04, 9.6974E+04, 9.7064E+04, 9.7153E+04, 9.7243E+04,
     +  9.7333E+04, 9.7423E+04, 9.7513E+04, 9.7603E+04, 9.7694E+04,
     + 9.7784E+04, 9.7874E+04, 9.7964E+04, 9.8055E+04, 9.8145E+04,
     +  9.8235E+04, 9.8326E+04, 9.8416E+04, 9.8507E+04, 9.8597E+04,
     + 9.8688E+04, 9.8779E+04, 9.8870E+04, 9.8960E+04, 9.9051E+04,
     +  9.9142E+04, 9.9233E+04, 9.9324E+04, 9.9415E+04, 9.9506E+04,
     + 9.9597E+04, 9.9689E+04, 9.9780E+04, 9.9871E+04, 9.9962E+04,
     +  1.0005E+05, 1.0015E+05, 1.0024E+05, 1.0033E+05, 1.0042E+05,
     + 1.0051E+05, 1.0060E+05, 1.0069E+05, 1.0079E+05, 1.0088E+05,
     +  1.0097E+05, 1.0106E+05, 1.0115E+05, 1.0125E+05, 1.0134E+05,
     + 1.0143E+05, 1.0152E+05, 1.0161E+05, 1.0171E+05, 1.0180E+05,
     +  1.0189E+05, 1.0198E+05, 1.0208E+05, 1.0217E+05, 1.0226E+05,
     + 1.0235E+05, 1.0245E+05, 1.0254E+05, 1.0263E+05, 1.0272E+05,
     +  1.0282E+05, 1.0291E+05, 1.0300E+05, 1.0310E+05, 1.0319E+05,
     + 1.0328E+05, 1.0338E+05, 1.0347E+05, 1.0356E+05, 1.0366E+05,
     +  1.0375E+05, 1.0384E+05, 1.0394E+05, 1.0403E+05, 1.0412E+05,
     + 1.0422E+05, 1.0431E+05, 1.0440E+05, 1.0450E+05, 1.0459E+05,
     +  1.0468E+05, 1.0478E+05, 1.0487E+05, 1.0497E+05, 1.0506E+05,
     + 1.0515E+05, 1.0525E+05, 1.0534E+05, 1.0544E+05, 1.0553E+05,
     +  1.0562E+05, 1.0572E+05, 1.0581E+05, 1.0591E+05, 1.0600E+05,
     + 1.0610E+05, 1.0619E+05, 1.0629E+05, 1.0638E+05, 1.0648E+05,
     +  1.0657E+05, 1.0666E+05, 1.0676E+05, 1.0685E+05, 1.0695E+05,
     + 1.0704E+05, 1.0714E+05, 1.0723E+05, 1.0733E+05, 1.0742E+05,
     +  1.0752E+05, 1.0762E+05, 1.0771E+05, 1.0781E+05, 1.0790E+05,
     + 1.0800E+05, 1.0809E+05, 1.0819E+05, 1.0828E+05, 1.0838E+05,
     +  1.0847E+05, 1.0857E+05, 1.0867E+05, 1.0876E+05, 1.0886E+05,
     + 1.0895E+05, 1.0905E+05, 1.0915E+05, 1.0924E+05, 1.0934E+05,
     +  1.0943E+05, 1.0953E+05, 1.0963E+05, 1.0972E+05, 1.0982E+05,
     + 1.0992E+05, 1.1001E+05, 1.1011E+05, 1.1021E+05, 1.1030E+05,
     +  1.1040E+05, 1.1050E+05, 1.1059E+05, 1.1069E+05, 1.1079E+05,
     + 1.1088E+05, 1.1098E+05, 1.1108E+05, 1.1117E+05, 1.1127E+05,
     +  1.1137E+05, 1.1147E+05, 1.1156E+05, 1.1166E+05, 1.1176E+05,
     + 1.1186E+05, 1.1195E+05, 1.1205E+05, 1.1215E+05, 1.1225E+05,
     +  1.1234E+05, 1.1244E+05, 1.1254E+05, 1.1264E+05, 1.1273E+05,
     + 1.1283E+05, 1.1293E+05, 1.1303E+05, 1.1313E+05, 1.1322E+05,
     +  1.1332E+05, 1.1342E+05, 1.1352E+05, 1.1362E+05, 1.1372E+05,
     + 1.1381E+05, 1.1391E+05, 1.1401E+05, 1.1411E+05, 1.1421E+05,
     +  1.1431E+05, 1.1440E+05, 1.1450E+05, 1.1460E+05, 1.1470E+05,
     + 1.1480E+05, 1.1490E+05, 1.1500E+05, 1.1510E+05, 1.1520E+05,
     +  1.1529E+05, 1.1539E+05, 1.1549E+05, 1.1559E+05, 1.1569E+05,
     + 1.1579E+05, 1.1589E+05, 1.1599E+05, 1.1609E+05, 1.1619E+05,
     +  1.1629E+05, 1.1639E+05, 1.1649E+05, 1.1659E+05, 1.1669E+05,
     + 1.1679E+05, 1.1689E+05, 1.1699E+05, 1.1709E+05, 1.1719E+05,
     +  1.1729E+05, 1.1739E+05, 1.1749E+05, 1.1759E+05, 1.1769E+05,
     + 1.1779E+05, 1.1789E+05, 1.1799E+05, 1.1809E+05, 1.1819E+05,
     +  1.1829E+05, 1.1839E+05, 1.1849E+05, 1.1859E+05, 1.1869E+05,
     + 1.1879E+05, 1.1889E+05, 1.1900E+05, 1.1910E+05, 1.1920E+05,
     +  1.1930E+05, 1.1940E+05, 1.1950E+05, 1.1960E+05, 1.1970E+05,
     + 1.1980E+05, 1.1991E+05, 1.2001E+05, 1.2011E+05, 1.2021E+05,
     +  1.2031E+05, 1.2041E+05, 1.2051E+05, 1.2062E+05, 1.2072E+05,
     + 1.2082E+05, 1.2092E+05, 1.2102E+05, 1.2113E+05, 1.2123E+05,
     +  1.2133E+05, 1.2143E+05, 1.2153E+05, 1.2164E+05, 1.2174E+05,
     + 1.2184E+05, 1.2194E+05, 1.2204E+05, 1.2215E+05, 1.2225E+05,
     +  1.2235E+05, 1.2245E+05, 1.2256E+05, 1.2266E+05, 1.2276E+05,
     + 1.2287E+05, 1.2297E+05, 1.2307E+05, 1.2317E+05, 1.2328E+05,
     +  1.2338E+05, 1.2348E+05, 1.2359E+05, 1.2369E+05, 1.2379E+05,
     + 1.2390E+05, 1.2400E+05, 1.2410E+05, 1.2421E+05, 1.2431E+05,
     +  1.2441E+05, 1.2452E+05, 1.2462E+05, 1.2472E+05, 1.2483E+05,
     + 1.2493E+05, 1.2503E+05, 1.2514E+05, 1.2524E+05, 1.2535E+05,
     +  1.2545E+05, 1.2555E+05, 1.2566E+05, 1.2576E+05, 1.2587E+05,
     + 1.2597E+05, 1.2607E+05, 1.2618E+05, 1.2628E+05, 1.2639E+05,
     +  1.2649E+05, 1.2660E+05, 1.2670E+05, 1.2680E+05, 1.2691E+05,
     + 1.2701E+05, 1.2712E+05, 1.2722E+05, 1.2733E+05, 1.2743E+05,
     +  1.2754E+05, 1.2764E+05, 1.2775E+05, 1.2785E+05, 1.2796E+05,
     + 1.2806E+05, 1.2817E+05, 1.2827E+05, 1.2838E+05, 1.2848E+05,
     +  1.2859E+05, 1.2870E+05, 1.2880E+05, 1.2891E+05, 1.2901E+05,
     + 1.2912E+05, 1.2922E+05, 1.2933E+05, 1.2944E+05, 1.2954E+05,
     +  1.2965E+05, 1.2975E+05, 1.2986E+05, 1.2996E+05, 1.3007E+05/
      data (QofT( 3,J),J=2409,4510)/ 1.3018E+05, 1.3028E+05,
     +  1.3039E+05, 1.3050E+05, 1.3060E+05, 1.3071E+05, 1.3081E+05,
     +  1.3092E+05,
     + 1.3103E+05, 1.3113E+05, 1.3124E+05, 1.3135E+05, 1.3145E+05,
     +  1.3156E+05, 1.3167E+05, 1.3177E+05, 1.3188E+05, 1.3199E+05,
     + 1.3210E+05, 1.3220E+05, 1.3231E+05, 1.3242E+05, 1.3252E+05,
     +  1.3263E+05, 1.3274E+05, 1.3285E+05, 1.3295E+05, 1.3306E+05,
     + 1.3317E+05, 1.3328E+05, 1.3338E+05, 1.3349E+05, 1.3360E+05,
     +  1.3371E+05, 1.3381E+05, 1.3392E+05, 1.3403E+05, 1.3414E+05,
     + 1.3425E+05, 1.3435E+05, 1.3446E+05, 1.3457E+05, 1.3468E+05,
     +  1.3479E+05, 1.3490E+05, 1.3500E+05, 1.3511E+05, 1.3522E+05,
     + 1.3533E+05, 1.3544E+05, 1.3555E+05, 1.3565E+05, 1.3576E+05,
     +  1.3587E+05, 1.3598E+05, 1.3609E+05, 1.3620E+05, 1.3631E+05,
     + 1.3642E+05, 1.3652E+05, 1.3663E+05, 1.3674E+05, 1.3685E+05,
     +  1.3696E+05, 1.3707E+05, 1.3718E+05, 1.3729E+05, 1.3740E+05,
     + 1.3751E+05, 1.3762E+05, 1.3773E+05, 1.3784E+05, 1.3795E+05,
     +  1.3806E+05, 1.3817E+05, 1.3828E+05, 1.3839E+05, 1.3850E+05,
     + 1.3861E+05, 1.3872E+05, 1.3883E+05, 1.3894E+05, 1.3905E+05,
     +  1.3916E+05, 1.3927E+05, 1.3938E+05, 1.3949E+05, 1.3960E+05,
     + 1.3971E+05, 1.3982E+05, 1.3993E+05, 1.4004E+05, 1.4015E+05,
     +  1.4026E+05, 1.4037E+05, 1.4048E+05, 1.4059E+05, 1.4070E+05,
     + 1.4081E+05, 1.4092E+05, 1.4104E+05, 1.4115E+05, 1.4126E+05,
     +  1.4137E+05, 1.4148E+05, 1.4159E+05, 1.4170E+05, 1.4181E+05,
     + 1.4193E+05, 1.4204E+05, 1.4215E+05, 1.4226E+05, 1.4237E+05,
     +  1.4248E+05, 1.4259E+05, 1.4271E+05, 1.4282E+05, 1.4293E+05,
     + 1.4304E+05, 1.4315E+05, 1.4327E+05, 1.4338E+05, 1.4349E+05,
     +  1.4360E+05, 1.4371E+05, 1.4383E+05, 1.4394E+05, 1.4405E+05,
     + 1.4416E+05, 1.4428E+05, 1.4439E+05, 1.4450E+05, 1.4461E+05,
     +  1.4473E+05, 1.4484E+05, 1.4495E+05, 1.4506E+05, 1.4518E+05,
     + 1.4529E+05, 1.4540E+05, 1.4552E+05, 1.4563E+05, 1.4574E+05,
     +  1.4585E+05, 1.4597E+05, 1.4608E+05, 1.4619E+05, 1.4631E+05,
     + 1.4642E+05, 1.4653E+05, 1.4665E+05, 1.4676E+05, 1.4687E+05,
     +  1.4699E+05, 1.4710E+05, 1.4722E+05, 1.4733E+05, 1.4744E+05,
     + 1.4756E+05, 1.4767E+05, 1.4778E+05, 1.4790E+05, 1.4801E+05,
     +  1.4813E+05, 1.4824E+05, 1.4835E+05, 1.4847E+05, 1.4858E+05,
     + 1.4870E+05, 1.4881E+05, 1.4893E+05, 1.4904E+05, 1.4915E+05,
     +  1.4927E+05, 1.4938E+05, 1.4950E+05, 1.4961E+05, 1.4973E+05,
     + 1.4984E+05, 1.4996E+05, 1.5007E+05, 1.5019E+05, 1.5030E+05,
     +  1.5042E+05, 1.5053E+05, 1.5065E+05, 1.5076E+05, 1.5088E+05,
     + 1.5099E+05, 1.5111E+05, 1.5122E+05, 1.5134E+05, 1.5145E+05,
     +  1.5157E+05, 1.5169E+05, 1.5180E+05, 1.5192E+05, 1.5203E+05,
     + 1.5215E+05, 1.5226E+05, 1.5238E+05, 1.5250E+05, 1.5261E+05,
     +  1.5273E+05, 1.5284E+05, 1.5296E+05, 1.5308E+05, 1.5319E+05,
     + 1.5331E+05, 1.5342E+05, 1.5354E+05, 1.5366E+05, 1.5377E+05,
     +  1.5389E+05, 1.5401E+05, 1.5412E+05, 1.5424E+05, 1.5436E+05,
     + 1.5447E+05, 1.5459E+05, 1.5471E+05, 1.5482E+05, 1.5494E+05,
     +  1.5506E+05, 1.5517E+05, 1.5529E+05, 1.5541E+05, 1.5553E+05,
     + 1.5564E+05, 1.5576E+05, 1.5588E+05, 1.5600E+05, 1.5611E+05,
     +  1.5623E+05, 1.5635E+05, 1.5647E+05, 1.5658E+05, 1.5670E+05,
     + 1.5682E+05, 1.5694E+05, 1.5705E+05, 1.5717E+05, 1.5729E+05,
     +  1.5741E+05, 1.5753E+05, 1.5764E+05, 1.5776E+05, 1.5788E+05,
     + 1.5800E+05, 1.5812E+05, 1.5823E+05, 1.5835E+05, 1.5847E+05,
     +  1.5859E+05, 1.5871E+05, 1.5883E+05, 1.5894E+05, 1.5906E+05,
     + 1.5918E+05, 1.5930E+05, 1.5942E+05, 1.5954E+05, 1.5966E+05,
     +  1.5978E+05, 1.5990E+05, 1.6001E+05, 1.6013E+05, 1.6025E+05,
     + 1.6037E+05, 1.6049E+05, 1.6061E+05, 1.6073E+05, 1.6085E+05,
     +  1.6097E+05, 1.6109E+05, 1.6121E+05, 1.6133E+05, 1.6145E+05,
     + 1.6157E+05, 1.6169E+05, 1.6180E+05, 1.6192E+05, 1.6204E+05,
     +  1.6216E+05, 1.6228E+05, 1.6240E+05, 1.6252E+05, 1.6264E+05,
     + 1.6276E+05, 1.6288E+05, 1.6300E+05, 1.6312E+05, 1.6325E+05,
     +  1.6337E+05, 1.6349E+05, 1.6361E+05, 1.6373E+05, 1.6385E+05,
     + 1.6397E+05, 1.6409E+05, 1.6421E+05, 1.6433E+05, 1.6445E+05,
     +  1.6457E+05, 1.6469E+05, 1.6481E+05, 1.6493E+05, 1.6506E+05,
     + 1.6518E+05, 1.6530E+05, 1.6542E+05, 1.6554E+05, 1.6566E+05,
     +  1.6578E+05, 1.6590E+05, 1.6603E+05, 1.6615E+05, 1.6627E+05,
     + 1.6639E+05, 1.6651E+05, 1.6663E+05, 1.6675E+05, 1.6688E+05,
     +  1.6700E+05, 1.6712E+05, 1.6724E+05, 1.6736E+05, 1.6749E+05,
     + 1.6761E+05, 1.6773E+05, 1.6785E+05, 1.6797E+05, 1.6810E+05,
     +  1.6822E+05, 1.6834E+05, 1.6846E+05, 1.6859E+05, 1.6871E+05,
     + 1.6883E+05, 1.6895E+05, 1.6908E+05, 1.6920E+05, 1.6932E+05,
     +  1.6944E+05, 1.6957E+05, 1.6969E+05, 1.6981E+05, 1.6994E+05,
     + 1.7006E+05, 1.7018E+05, 1.7031E+05, 1.7043E+05, 1.7055E+05,
     +  1.7067E+05, 1.7080E+05, 1.7092E+05, 1.7104E+05, 1.7117E+05,
     + 1.7129E+05, 1.7142E+05, 1.7154E+05, 1.7166E+05, 1.7179E+05,
     +  1.7191E+05, 1.7203E+05, 1.7216E+05, 1.7228E+05, 1.7241E+05,
     + 1.7253E+05, 1.7265E+05, 1.7278E+05, 1.7290E+05, 1.7303E+05,
     +  1.7315E+05, 1.7327E+05, 1.7340E+05, 1.7352E+05, 1.7365E+05,
     + 1.7377E+05, 1.7390E+05, 1.7402E+05, 1.7415E+05, 1.7427E+05,
     +  1.7439E+05, 1.7452E+05, 1.7464E+05, 1.7477E+05, 1.7489E+05,
     + 1.7502E+05, 1.7514E+05, 1.7527E+05, 1.7539E+05, 1.7552E+05,
     +  1.7564E+05, 1.7577E+05, 1.7589E+05, 1.7602E+05, 1.7615E+05,
     + 1.7627E+05, 1.7640E+05, 1.7652E+05, 1.7665E+05, 1.7677E+05,
     +  1.7690E+05, 1.7702E+05, 1.7715E+05, 1.7728E+05, 1.7740E+05,
     + 1.7753E+05, 1.7765E+05, 1.7778E+05, 1.7791E+05, 1.7803E+05,
     +  1.7816E+05, 1.7828E+05, 1.7841E+05, 1.7854E+05, 1.7866E+05,
     + 1.7879E+05, 1.7892E+05, 1.7904E+05, 1.7917E+05, 1.7930E+05,
     +  1.7942E+05, 1.7955E+05, 1.7968E+05, 1.7980E+05, 1.7993E+05,
     + 1.8006E+05, 1.8018E+05, 1.8031E+05, 1.8044E+05, 1.8056E+05,
     +  1.8069E+05, 1.8082E+05, 1.8094E+05, 1.8107E+05, 1.8120E+05,
     + 1.8133E+05, 1.8145E+05, 1.8158E+05, 1.8171E+05, 1.8184E+05,
     +  1.8196E+05, 1.8209E+05, 1.8222E+05, 1.8235E+05, 1.8247E+05,
     + 1.8260E+05, 1.8273E+05, 1.8286E+05, 1.8299E+05, 1.8311E+05,
     +  1.8324E+05, 1.8337E+05, 1.8350E+05, 1.8363E+05, 1.8375E+05,
     + 1.8388E+05, 1.8401E+05, 1.8414E+05, 1.8427E+05, 1.8440E+05,
     +  1.8453E+05, 1.8465E+05, 1.8478E+05, 1.8491E+05, 1.8504E+05,
     + 1.8517E+05, 1.8530E+05, 1.8543E+05, 1.8556E+05, 1.8568E+05,
     +  1.8581E+05, 1.8594E+05, 1.8607E+05, 1.8620E+05, 1.8633E+05,
     + 1.8646E+05, 1.8659E+05, 1.8672E+05, 1.8685E+05, 1.8698E+05,
     +  1.8711E+05, 1.8724E+05, 1.8737E+05, 1.8749E+05, 1.8762E+05,
     + 1.8775E+05, 1.8788E+05, 1.8801E+05, 1.8814E+05, 1.8827E+05,
     +  1.8840E+05, 1.8853E+05, 1.8866E+05, 1.8879E+05, 1.8892E+05,
     + 1.8905E+05, 1.8918E+05, 1.8931E+05, 1.8944E+05, 1.8958E+05,
     +  1.8971E+05, 1.8984E+05, 1.8997E+05, 1.9010E+05, 1.9023E+05,
     + 1.9036E+05, 1.9049E+05, 1.9062E+05, 1.9075E+05, 1.9088E+05,
     +  1.9101E+05, 1.9114E+05, 1.9127E+05, 1.9141E+05, 1.9154E+05,
     + 1.9167E+05, 1.9180E+05, 1.9193E+05, 1.9206E+05, 1.9219E+05,
     +  1.9232E+05, 1.9246E+05, 1.9259E+05, 1.9272E+05, 1.9285E+05,
     + 1.9298E+05, 1.9311E+05, 1.9325E+05, 1.9338E+05, 1.9351E+05,
     +  1.9364E+05, 1.9377E+05, 1.9391E+05, 1.9404E+05, 1.9417E+05,
     + 1.9430E+05, 1.9443E+05, 1.9457E+05, 1.9470E+05, 1.9483E+05,
     +  1.9496E+05, 1.9509E+05, 1.9523E+05, 1.9536E+05, 1.9549E+05,
     + 1.9562E+05, 1.9576E+05, 1.9589E+05, 1.9602E+05, 1.9616E+05,
     +  1.9629E+05, 1.9642E+05, 1.9655E+05, 1.9669E+05, 1.9682E+05,
     + 1.9695E+05, 1.9709E+05, 1.9722E+05, 1.9735E+05, 1.9749E+05,
     +  1.9762E+05, 1.9775E+05, 1.9789E+05, 1.9802E+05, 1.9815E+05,
     + 1.9829E+05, 1.9842E+05, 1.9855E+05, 1.9869E+05, 1.9882E+05,
     +  1.9895E+05, 1.9909E+05, 1.9922E+05, 1.9936E+05, 1.9949E+05,
     + 1.9962E+05, 1.9976E+05, 1.9989E+05, 2.0003E+05, 2.0016E+05,
     +  2.0030E+05, 2.0043E+05, 2.0056E+05, 2.0070E+05, 2.0083E+05,
     + 2.0097E+05, 2.0110E+05, 2.0124E+05, 2.0137E+05, 2.0151E+05,
     +  2.0164E+05, 2.0178E+05, 2.0191E+05, 2.0204E+05, 2.0218E+05,
     + 2.0231E+05, 2.0245E+05, 2.0258E+05, 2.0272E+05, 2.0285E+05,
     +  2.0299E+05, 2.0313E+05, 2.0326E+05, 2.0340E+05, 2.0353E+05,
     + 2.0367E+05, 2.0380E+05, 2.0394E+05, 2.0407E+05, 2.0421E+05,
     +  2.0434E+05, 2.0448E+05, 2.0462E+05, 2.0475E+05, 2.0489E+05,
     + 2.0502E+05, 2.0516E+05, 2.0530E+05, 2.0543E+05, 2.0557E+05,
     +  2.0570E+05, 2.0584E+05, 2.0598E+05, 2.0611E+05, 2.0625E+05,
     + 2.0639E+05, 2.0652E+05, 2.0666E+05, 2.0679E+05, 2.0693E+05,
     +  2.0707E+05, 2.0720E+05, 2.0734E+05, 2.0748E+05, 2.0761E+05,
     + 2.0775E+05, 2.0789E+05, 2.0803E+05, 2.0816E+05, 2.0830E+05,
     +  2.0844E+05, 2.0857E+05, 2.0871E+05, 2.0885E+05, 2.0899E+05,
     + 2.0912E+05, 2.0926E+05, 2.0940E+05, 2.0953E+05, 2.0967E+05,
     +  2.0981E+05, 2.0995E+05, 2.1009E+05, 2.1022E+05, 2.1036E+05,
     + 2.1050E+05, 2.1064E+05, 2.1077E+05, 2.1091E+05, 2.1105E+05,
     +  2.1119E+05, 2.1133E+05, 2.1146E+05, 2.1160E+05, 2.1174E+05,
     + 2.1188E+05, 2.1202E+05, 2.1216E+05, 2.1229E+05, 2.1243E+05,
     +  2.1257E+05, 2.1271E+05, 2.1285E+05, 2.1299E+05, 2.1313E+05,
     + 2.1326E+05, 2.1340E+05, 2.1354E+05, 2.1368E+05, 2.1382E+05,
     +  2.1396E+05, 2.1410E+05, 2.1424E+05, 2.1438E+05, 2.1451E+05,
     + 2.1465E+05, 2.1479E+05, 2.1493E+05, 2.1507E+05, 2.1521E+05,
     +  2.1535E+05, 2.1549E+05, 2.1563E+05, 2.1577E+05, 2.1591E+05,
     + 2.1605E+05, 2.1619E+05, 2.1633E+05, 2.1647E+05, 2.1661E+05,
     +  2.1675E+05, 2.1689E+05, 2.1703E+05, 2.1717E+05, 2.1731E+05,
     + 2.1745E+05, 2.1759E+05, 2.1773E+05, 2.1787E+05, 2.1801E+05,
     +  2.1815E+05, 2.1829E+05, 2.1843E+05, 2.1857E+05, 2.1871E+05,
     + 2.1885E+05, 2.1899E+05, 2.1913E+05, 2.1927E+05, 2.1941E+05,
     +  2.1956E+05, 2.1970E+05, 2.1984E+05, 2.1998E+05, 2.2012E+05,
     + 2.2026E+05, 2.2040E+05, 2.2054E+05, 2.2068E+05, 2.2082E+05,
     +  2.2097E+05, 2.2111E+05, 2.2125E+05, 2.2139E+05, 2.2153E+05,
     + 2.2167E+05, 2.2181E+05, 2.2196E+05, 2.2210E+05, 2.2224E+05,
     +  2.2238E+05, 2.2252E+05, 2.2267E+05, 2.2281E+05, 2.2295E+05,
     + 2.2309E+05, 2.2323E+05, 2.2338E+05, 2.2352E+05, 2.2366E+05,
     +  2.2380E+05, 2.2394E+05, 2.2409E+05, 2.2423E+05, 2.2437E+05,
     + 2.2451E+05, 2.2466E+05, 2.2480E+05, 2.2494E+05, 2.2508E+05,
     +  2.2523E+05, 2.2537E+05, 2.2551E+05, 2.2566E+05, 2.2580E+05,
     + 2.2594E+05, 2.2608E+05, 2.2623E+05, 2.2637E+05, 2.2651E+05,
     +  2.2666E+05, 2.2680E+05, 2.2694E+05, 2.2709E+05, 2.2723E+05,
     + 2.2737E+05, 2.2752E+05, 2.2766E+05, 2.2780E+05, 2.2795E+05,
     +  2.2809E+05, 2.2823E+05, 2.2838E+05, 2.2852E+05, 2.2867E+05,
     + 2.2881E+05, 2.2895E+05, 2.2910E+05, 2.2924E+05, 2.2939E+05,
     +  2.2953E+05, 2.2967E+05, 2.2982E+05, 2.2996E+05, 2.3011E+05,
     + 2.3025E+05, 2.3039E+05, 2.3054E+05, 2.3068E+05, 2.3083E+05,
     +  2.3097E+05, 2.3112E+05, 2.3126E+05, 2.3141E+05, 2.3155E+05,
     + 2.3170E+05, 2.3184E+05, 2.3199E+05, 2.3213E+05, 2.3228E+05,
     +  2.3242E+05, 2.3257E+05, 2.3271E+05, 2.3286E+05, 2.3300E+05,
     + 2.3315E+05, 2.3329E+05, 2.3344E+05, 2.3358E+05, 2.3373E+05,
     +  2.3387E+05, 2.3402E+05, 2.3416E+05, 2.3431E+05, 2.3446E+05,
     + 2.3460E+05, 2.3475E+05, 2.3489E+05, 2.3504E+05, 2.3518E+05,
     +  2.3533E+05, 2.3548E+05, 2.3562E+05, 2.3577E+05, 2.3591E+05,
     + 2.3606E+05, 2.3621E+05, 2.3635E+05, 2.3650E+05, 2.3665E+05,
     +  2.3679E+05, 2.3694E+05, 2.3709E+05, 2.3723E+05, 2.3738E+05,
     + 2.3753E+05, 2.3767E+05, 2.3782E+05, 2.3797E+05, 2.3811E+05,
     +  2.3826E+05, 2.3841E+05, 2.3855E+05, 2.3870E+05, 2.3885E+05,
     + 2.3899E+05, 2.3914E+05, 2.3929E+05, 2.3944E+05, 2.3958E+05,
     +  2.3973E+05, 2.3988E+05, 2.4002E+05, 2.4017E+05, 2.4032E+05,
     + 2.4047E+05, 2.4062E+05, 2.4076E+05, 2.4091E+05, 2.4106E+05,
     +  2.4121E+05, 2.4135E+05, 2.4150E+05, 2.4165E+05, 2.4180E+05,
     + 2.4195E+05, 2.4209E+05, 2.4224E+05, 2.4239E+05, 2.4254E+05,
     +  2.4269E+05, 2.4283E+05, 2.4298E+05, 2.4313E+05, 2.4328E+05,
     + 2.4343E+05, 2.4358E+05, 2.4373E+05, 2.4387E+05, 2.4402E+05,
     +  2.4417E+05, 2.4432E+05, 2.4447E+05, 2.4462E+05, 2.4477E+05,
     + 2.4492E+05, 2.4506E+05, 2.4521E+05, 2.4536E+05, 2.4551E+05,
     +  2.4566E+05, 2.4581E+05, 2.4596E+05, 2.4611E+05, 2.4626E+05,
     + 2.4641E+05, 2.4656E+05, 2.4671E+05, 2.4686E+05, 2.4700E+05,
     +  2.4715E+05, 2.4730E+05, 2.4745E+05, 2.4760E+05, 2.4775E+05,
     + 2.4790E+05, 2.4805E+05, 2.4820E+05, 2.4835E+05, 2.4850E+05,
     +  2.4865E+05, 2.4880E+05, 2.4895E+05, 2.4910E+05, 2.4925E+05,
     + 2.4940E+05, 2.4955E+05, 2.4970E+05, 2.4986E+05, 2.5001E+05,
     +  2.5016E+05, 2.5031E+05, 2.5046E+05, 2.5061E+05, 2.5076E+05,
     + 2.5091E+05, 2.5106E+05, 2.5121E+05, 2.5136E+05, 2.5151E+05,
     +  2.5166E+05, 2.5181E+05, 2.5197E+05, 2.5212E+05, 2.5227E+05,
     + 2.5242E+05, 2.5257E+05, 2.5272E+05, 2.5287E+05, 2.5302E+05,
     +  2.5318E+05, 2.5333E+05, 2.5348E+05, 2.5363E+05, 2.5378E+05,
     + 2.5393E+05, 2.5409E+05, 2.5424E+05, 2.5439E+05, 2.5454E+05,
     +  2.5469E+05, 2.5484E+05, 2.5500E+05, 2.5515E+05, 2.5530E+05,
     + 2.5545E+05, 2.5560E+05, 2.5576E+05, 2.5591E+05, 2.5606E+05,
     +  2.5621E+05, 2.5637E+05, 2.5652E+05, 2.5667E+05, 2.5682E+05,
     + 2.5698E+05, 2.5713E+05, 2.5728E+05, 2.5743E+05, 2.5759E+05,
     +  2.5774E+05, 2.5789E+05, 2.5805E+05, 2.5820E+05, 2.5835E+05,
     + 2.5850E+05, 2.5866E+05, 2.5881E+05, 2.5896E+05, 2.5912E+05,
     +  2.5927E+05, 2.5942E+05, 2.5958E+05, 2.5973E+05, 2.5988E+05,
     + 2.6004E+05, 2.6019E+05, 2.6034E+05, 2.6050E+05, 2.6065E+05,
     +  2.6080E+05, 2.6096E+05, 2.6111E+05, 2.6127E+05, 2.6142E+05,
     + 2.6157E+05, 2.6173E+05, 2.6188E+05, 2.6204E+05, 2.6219E+05,
     +  2.6234E+05, 2.6250E+05, 2.6265E+05, 2.6281E+05, 2.6296E+05,
     + 2.6312E+05, 2.6327E+05, 2.6342E+05, 2.6358E+05, 2.6373E+05,
     +  2.6389E+05, 2.6404E+05, 2.6420E+05, 2.6435E+05, 2.6451E+05,
     + 2.6466E+05, 2.6482E+05, 2.6497E+05, 2.6513E+05, 2.6528E+05,
     +  2.6544E+05, 2.6559E+05, 2.6575E+05, 2.6590E+05, 2.6606E+05,
     + 2.6621E+05, 2.6637E+05, 2.6652E+05, 2.6668E+05, 2.6683E+05,
     +  2.6699E+05, 2.6714E+05, 2.6730E+05, 2.6746E+05, 2.6761E+05,
     + 2.6777E+05, 2.6792E+05, 2.6808E+05, 2.6823E+05, 2.6839E+05,
     +  2.6855E+05, 2.6870E+05, 2.6886E+05, 2.6901E+05, 2.6917E+05,
     + 2.6933E+05, 2.6948E+05, 2.6964E+05, 2.6980E+05, 2.6995E+05,
     +  2.7011E+05, 2.7026E+05, 2.7042E+05, 2.7058E+05, 2.7073E+05,
     + 2.7089E+05, 2.7105E+05, 2.7120E+05, 2.7136E+05, 2.7152E+05,
     +  2.7167E+05, 2.7183E+05, 2.7199E+05, 2.7214E+05, 2.7230E+05,
     + 2.7246E+05, 2.7262E+05, 2.7277E+05, 2.7293E+05, 2.7309E+05,
     +  2.7324E+05, 2.7340E+05, 2.7356E+05, 2.7372E+05, 2.7387E+05,
     + 2.7403E+05, 2.7419E+05, 2.7435E+05, 2.7450E+05, 2.7466E+05,
     +  2.7482E+05, 2.7498E+05, 2.7513E+05, 2.7529E+05, 2.7545E+05,
     + 2.7561E+05, 2.7577E+05, 2.7592E+05, 2.7608E+05, 2.7624E+05,
     +  2.7640E+05, 2.7656E+05, 2.7672E+05, 2.7687E+05, 2.7703E+05,
     + 2.7719E+05, 2.7735E+05, 2.7751E+05, 2.7767E+05, 2.7782E+05,
     +  2.7798E+05, 2.7814E+05, 2.7830E+05, 2.7846E+05, 2.7862E+05,
     + 2.7878E+05, 2.7894E+05, 2.7909E+05, 2.7925E+05, 2.7941E+05,
     +  2.7957E+05, 2.7973E+05, 2.7989E+05, 2.8005E+05, 2.8021E+05,
     + 2.8037E+05, 2.8053E+05, 2.8069E+05, 2.8084E+05, 2.8100E+05,
     +  2.8116E+05, 2.8132E+05, 2.8148E+05, 2.8164E+05, 2.8180E+05,
     + 2.8196E+05, 2.8212E+05, 2.8228E+05, 2.8244E+05, 2.8260E+05,
     +  2.8276E+05, 2.8292E+05, 2.8308E+05, 2.8324E+05, 2.8340E+05,
     + 2.8356E+05, 2.8372E+05, 2.8388E+05, 2.8404E+05, 2.8420E+05,
     +  2.8436E+05, 2.8452E+05, 2.8468E+05, 2.8484E+05, 2.8500E+05,
     + 2.8516E+05, 2.8533E+05, 2.8549E+05, 2.8565E+05, 2.8581E+05,
     +  2.8597E+05, 2.8613E+05, 2.8629E+05, 2.8645E+05, 2.8661E+05,
     + 2.8677E+05, 2.8693E+05, 2.8709E+05, 2.8726E+05, 2.8742E+05,
     +  2.8758E+05, 2.8774E+05, 2.8790E+05, 2.8806E+05, 2.8822E+05,
     + 2.8839E+05, 2.8855E+05, 2.8871E+05, 2.8887E+05, 2.8903E+05,
     +  2.8919E+05, 2.8935E+05, 2.8952E+05, 2.8968E+05, 2.8984E+05,
     + 2.9000E+05, 2.9016E+05, 2.9033E+05, 2.9049E+05, 2.9065E+05,
     +  2.9081E+05, 2.9097E+05, 2.9114E+05, 2.9130E+05, 2.9146E+05,
     + 2.9162E+05, 2.9179E+05, 2.9195E+05, 2.9211E+05, 2.9227E+05,
     +  2.9244E+05, 2.9260E+05, 2.9276E+05, 2.9292E+05, 2.9309E+05,
     + 2.9325E+05, 2.9341E+05, 2.9357E+05, 2.9374E+05, 2.9390E+05,
     +  2.9406E+05, 2.9423E+05, 2.9439E+05, 2.9455E+05, 2.9472E+05,
     + 2.9488E+05, 2.9504E+05, 2.9521E+05, 2.9537E+05, 2.9553E+05,
     +  2.9570E+05, 2.9586E+05, 2.9602E+05, 2.9619E+05, 2.9635E+05,
     + 2.9651E+05, 2.9668E+05, 2.9684E+05, 2.9700E+05, 2.9717E+05,
     +  2.9733E+05, 2.9750E+05, 2.9766E+05, 2.9782E+05, 2.9799E+05,
     + 2.9815E+05, 2.9832E+05, 2.9848E+05, 2.9864E+05, 2.9881E+05,
     +  2.9897E+05, 2.9914E+05, 2.9930E+05, 2.9947E+05, 2.9963E+05,
     + 2.9979E+05, 2.9996E+05, 3.0012E+05, 3.0029E+05, 3.0045E+05,
     +  3.0062E+05, 3.0078E+05, 3.0095E+05, 3.0111E+05, 3.0128E+05,
     + 3.0144E+05, 3.0161E+05, 3.0177E+05, 3.0194E+05, 3.0210E+05,
     +  3.0227E+05, 3.0243E+05, 3.0260E+05, 3.0276E+05, 3.0293E+05,
     + 3.0309E+05, 3.0326E+05, 3.0342E+05, 3.0359E+05, 3.0375E+05,
     +  3.0392E+05, 3.0409E+05, 3.0425E+05, 3.0442E+05, 3.0458E+05,
     + 3.0475E+05, 3.0491E+05, 3.0508E+05, 3.0525E+05, 3.0541E+05,
     +  3.0558E+05, 3.0574E+05, 3.0591E+05, 3.0608E+05, 3.0624E+05,
     + 3.0641E+05, 3.0657E+05, 3.0674E+05, 3.0691E+05, 3.0707E+05,
     +  3.0724E+05, 3.0741E+05, 3.0757E+05, 3.0774E+05, 3.0791E+05,
     + 3.0807E+05, 3.0824E+05, 3.0841E+05, 3.0857E+05, 3.0874E+05,
     +  3.0891E+05, 3.0907E+05, 3.0924E+05, 3.0941E+05, 3.0957E+05,
     + 3.0974E+05, 3.0991E+05, 3.1008E+05, 3.1024E+05, 3.1041E+05,
     +  3.1058E+05, 3.1074E+05, 3.1091E+05, 3.1108E+05, 3.1125E+05,
     + 3.1141E+05, 3.1158E+05, 3.1175E+05, 3.1192E+05, 3.1208E+05,
     +  3.1225E+05, 3.1242E+05, 3.1259E+05, 3.1276E+05, 3.1292E+05,
     + 3.1309E+05, 3.1326E+05, 3.1343E+05, 3.1360E+05, 3.1376E+05,
     +  3.1393E+05, 3.1410E+05, 3.1427E+05, 3.1444E+05, 3.1460E+05,
     + 3.1477E+05, 3.1494E+05, 3.1511E+05, 3.1528E+05, 3.1545E+05,
     +  3.1561E+05, 3.1578E+05, 3.1595E+05, 3.1612E+05, 3.1629E+05,
     + 3.1646E+05, 3.1663E+05, 3.1680E+05, 3.1696E+05, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        88
      data (QofT( 4,J),J=1,2408)/ 1.2710E+00, 2.3161E+00, 3.3314E+00,
     +  4.1979E+00, 4.9877E+00, 5.7544E+00, 6.5207E+00, 7.2935E+00,
     + 8.0735E+00, 8.8596E+00, 9.6506E+00, 1.0446E+01, 1.1243E+01,
     +  1.2044E+01, 1.2846E+01, 1.3650E+01, 1.4456E+01, 1.5262E+01,
     + 1.6069E+01, 1.6878E+01, 1.8496E+01, 2.0117E+01, 2.1739E+01,
     +  2.3363E+01, 2.4988E+01, 2.6614E+01, 2.8240E+01, 2.9867E+01,
     + 3.1495E+01, 3.3123E+01, 3.4751E+01, 3.6380E+01, 3.8009E+01,
     +  3.9638E+01, 4.1268E+01, 4.2897E+01, 4.4527E+01, 4.6157E+01,
     + 4.7788E+01, 4.9418E+01, 5.1049E+01, 5.2679E+01, 5.4310E+01,
     +  5.5941E+01, 5.7572E+01, 5.9203E+01, 6.0834E+01, 6.2465E+01,
     + 6.4097E+01, 6.5728E+01, 6.7360E+01, 6.8991E+01, 7.0623E+01,
     +  7.2254E+01, 7.3886E+01, 7.5518E+01, 7.7150E+01, 7.8781E+01,
     + 8.0413E+01, 8.2045E+01, 8.3677E+01, 8.5309E+01, 8.6941E+01,
     +  8.8574E+01, 9.0206E+01, 9.1838E+01, 9.3470E+01, 9.5103E+01,
     + 9.6735E+01, 9.8367E+01, 1.0000E+02, 1.0163E+02, 1.0326E+02,
     +  1.0490E+02, 1.0653E+02, 1.0816E+02, 1.0979E+02, 1.1143E+02,
     + 1.1306E+02, 1.1469E+02, 1.1633E+02, 1.1796E+02, 1.1959E+02,
     +  1.2122E+02, 1.2286E+02, 1.2449E+02, 1.2612E+02, 1.2776E+02,
     + 1.2939E+02, 1.3102E+02, 1.3265E+02, 1.3429E+02, 1.3592E+02,
     +  1.3755E+02, 1.3919E+02, 1.4082E+02, 1.4245E+02, 1.4409E+02,
     + 1.4572E+02, 1.4735E+02, 1.4899E+02, 1.5062E+02, 1.5225E+02,
     +  1.5389E+02, 1.5552E+02, 1.5715E+02, 1.5879E+02, 1.6042E+02,
     + 1.6206E+02, 1.6369E+02, 1.6532E+02, 1.6696E+02, 1.6859E+02,
     +  1.7023E+02, 1.7186E+02, 1.7350E+02, 1.7513E+02, 1.7676E+02,
     + 1.7840E+02, 1.8003E+02, 1.8167E+02, 1.8330E+02, 1.8494E+02,
     +  1.8658E+02, 1.8821E+02, 1.8985E+02, 1.9148E+02, 1.9312E+02,
     + 1.9475E+02, 1.9639E+02, 1.9803E+02, 1.9966E+02, 2.0130E+02,
     +  2.0294E+02, 2.0458E+02, 2.0621E+02, 2.0785E+02, 2.0949E+02,
     + 2.1113E+02, 2.1277E+02, 2.1440E+02, 2.1604E+02, 2.1768E+02,
     +  2.1932E+02, 2.2096E+02, 2.2260E+02, 2.2424E+02, 2.2588E+02,
     + 2.2753E+02, 2.2917E+02, 2.3081E+02, 2.3245E+02, 2.3409E+02,
     +  2.3574E+02, 2.3738E+02, 2.3903E+02, 2.4067E+02, 2.4232E+02,
     + 2.4396E+02, 2.4561E+02, 2.4725E+02, 2.4890E+02, 2.5055E+02,
     +  2.5220E+02, 2.5385E+02, 2.5549E+02, 2.5714E+02, 2.5879E+02,
     + 2.6045E+02, 2.6210E+02, 2.6375E+02, 2.6540E+02, 2.6706E+02,
     +  2.6871E+02, 2.7037E+02, 2.7202E+02, 2.7368E+02, 2.7533E+02,
     + 2.7699E+02, 2.7865E+02, 2.8031E+02, 2.8197E+02, 2.8363E+02,
     +  2.8529E+02, 2.8696E+02, 2.8862E+02, 2.9028E+02, 2.9195E+02,
     + 2.9362E+02, 2.9528E+02, 2.9695E+02, 2.9862E+02, 3.0029E+02,
     +  3.0196E+02, 3.0363E+02, 3.0531E+02, 3.0698E+02, 3.0866E+02,
     + 3.1033E+02, 3.1201E+02, 3.1369E+02, 3.1537E+02, 3.1705E+02,
     +  3.1873E+02, 3.2041E+02, 3.2209E+02, 3.2378E+02, 3.2546E+02,
     + 3.2715E+02, 3.2884E+02, 3.3053E+02, 3.3222E+02, 3.3391E+02,
     +  3.3561E+02, 3.3730E+02, 3.3900E+02, 3.4069E+02, 3.4239E+02,
     + 3.4409E+02, 3.4579E+02, 3.4750E+02, 3.4920E+02, 3.5091E+02,
     +  3.5261E+02, 3.5432E+02, 3.5603E+02, 3.5774E+02, 3.5946E+02,
     + 3.6117E+02, 3.6288E+02, 3.6460E+02, 3.6632E+02, 3.6804E+02,
     +  3.6976E+02, 3.7149E+02, 3.7321E+02, 3.7494E+02, 3.7666E+02,
     + 3.7839E+02, 3.8012E+02, 3.8186E+02, 3.8359E+02, 3.8533E+02,
     +  3.8707E+02, 3.8880E+02, 3.9055E+02, 3.9229E+02, 3.9403E+02,
     + 3.9578E+02, 3.9753E+02, 3.9928E+02, 4.0103E+02, 4.0278E+02,
     +  4.0454E+02, 4.0629E+02, 4.0805E+02, 4.0981E+02, 4.1157E+02,
     + 4.1334E+02, 4.1510E+02, 4.1687E+02, 4.1864E+02, 4.2041E+02,
     +  4.2219E+02, 4.2396E+02, 4.2574E+02, 4.2752E+02, 4.2930E+02,
     + 4.3108E+02, 4.3287E+02, 4.3465E+02, 4.3644E+02, 4.3823E+02,
     +  4.4003E+02, 4.4182E+02, 4.4362E+02, 4.4542E+02, 4.4722E+02,
     + 4.4902E+02, 4.5082E+02, 4.5263E+02, 4.5444E+02, 4.5625E+02,
     +  4.5807E+02, 4.5988E+02, 4.6170E+02, 4.6352E+02, 4.6534E+02,
     + 4.6716E+02, 4.6899E+02, 4.7082E+02, 4.7265E+02, 4.7448E+02,
     +  4.7632E+02, 4.7815E+02, 4.7999E+02, 4.8183E+02, 4.8368E+02,
     + 4.8552E+02, 4.8737E+02, 4.8922E+02, 4.9107E+02, 4.9293E+02,
     +  4.9478E+02, 4.9664E+02, 4.9851E+02, 5.0037E+02, 5.0224E+02,
     + 5.0410E+02, 5.0597E+02, 5.0785E+02, 5.0972E+02, 5.1160E+02,
     +  5.1348E+02, 5.1536E+02, 5.1725E+02, 5.1913E+02, 5.2102E+02,
     + 5.2292E+02, 5.2481E+02, 5.2671E+02, 5.2860E+02, 5.3051E+02,
     +  5.3241E+02, 5.3432E+02, 5.3622E+02, 5.3814E+02, 5.4005E+02,
     + 5.4196E+02, 5.4388E+02, 5.4580E+02, 5.4773E+02, 5.4965E+02,
     +  5.5158E+02, 5.5351E+02, 5.5544E+02, 5.5738E+02, 5.5932E+02,
     + 5.6126E+02, 5.6320E+02, 5.6515E+02, 5.6710E+02, 5.6905E+02,
     +  5.7100E+02, 5.7295E+02, 5.7491E+02, 5.7687E+02, 5.7884E+02,
     + 5.8080E+02, 5.8277E+02, 5.8474E+02, 5.8672E+02, 5.8869E+02,
     +  5.9067E+02, 5.9265E+02, 5.9464E+02, 5.9662E+02, 5.9861E+02,
     + 6.0060E+02, 6.0260E+02, 6.0459E+02, 6.0659E+02, 6.0860E+02,
     +  6.1060E+02, 6.1261E+02, 6.1462E+02, 6.1663E+02, 6.1865E+02,
     + 6.2066E+02, 6.2268E+02, 6.2471E+02, 6.2673E+02, 6.2876E+02,
     +  6.3079E+02, 6.3283E+02, 6.3486E+02, 6.3690E+02, 6.3895E+02,
     + 6.4099E+02, 6.4304E+02, 6.4509E+02, 6.4714E+02, 6.4920E+02,
     +  6.5125E+02, 6.5332E+02, 6.5538E+02, 6.5745E+02, 6.5951E+02,
     + 6.6159E+02, 6.6366E+02, 6.6574E+02, 6.6782E+02, 6.6990E+02,
     +  6.7199E+02, 6.7408E+02, 6.7617E+02, 6.7826E+02, 6.8036E+02,
     + 6.8246E+02, 6.8456E+02, 6.8666E+02, 6.8877E+02, 6.9088E+02,
     +  6.9300E+02, 6.9511E+02, 6.9723E+02, 6.9935E+02, 7.0148E+02,
     + 7.0360E+02, 7.0573E+02, 7.0787E+02, 7.1000E+02, 7.1214E+02,
     +  7.1428E+02, 7.1643E+02, 7.1857E+02, 7.2072E+02, 7.2288E+02,
     + 7.2503E+02, 7.2719E+02, 7.2935E+02, 7.3152E+02, 7.3368E+02,
     +  7.3585E+02, 7.3803E+02, 7.4020E+02, 7.4238E+02, 7.4456E+02,
     + 7.4674E+02, 7.4893E+02, 7.5112E+02, 7.5331E+02, 7.5551E+02,
     +  7.5771E+02, 7.5991E+02, 7.6211E+02, 7.6432E+02, 7.6653E+02,
     + 7.6874E+02, 7.7096E+02, 7.7318E+02, 7.7540E+02, 7.7762E+02,
     +  7.7985E+02, 7.8208E+02, 7.8432E+02, 7.8655E+02, 7.8879E+02,
     + 7.9103E+02, 7.9328E+02, 7.9553E+02, 7.9778E+02, 8.0003E+02,
     +  8.0229E+02, 8.0455E+02, 8.0681E+02, 8.0907E+02, 8.1134E+02,
     + 8.1361E+02, 8.1589E+02, 8.1816E+02, 8.2045E+02, 8.2273E+02,
     +  8.2501E+02, 8.2730E+02, 8.2959E+02, 8.3189E+02, 8.3419E+02,
     + 8.3649E+02, 8.3879E+02, 8.4110E+02, 8.4341E+02, 8.4572E+02,
     +  8.4804E+02, 8.5036E+02, 8.5268E+02, 8.5500E+02, 8.5733E+02,
     + 8.5966E+02, 8.6199E+02, 8.6433E+02, 8.6667E+02, 8.6901E+02,
     +  8.7136E+02, 8.7371E+02, 8.7606E+02, 8.7841E+02, 8.8077E+02,
     + 8.8313E+02, 8.8549E+02, 8.8786E+02, 8.9023E+02, 8.9260E+02,
     +  8.9498E+02, 8.9736E+02, 8.9974E+02, 9.0212E+02, 9.0451E+02,
     + 9.0690E+02, 9.0929E+02, 9.1169E+02, 9.1409E+02, 9.1649E+02,
     +  9.1890E+02, 9.2131E+02, 9.2372E+02, 9.2613E+02, 9.2855E+02,
     + 9.3097E+02, 9.3340E+02, 9.3582E+02, 9.3825E+02, 9.4069E+02,
     +  9.4312E+02, 9.4556E+02, 9.4800E+02, 9.5045E+02, 9.5290E+02,
     + 9.5535E+02, 9.5780E+02, 9.6026E+02, 9.6272E+02, 9.6518E+02,
     +  9.6765E+02, 9.7012E+02, 9.7259E+02, 9.7507E+02, 9.7755E+02,
     + 9.8003E+02, 9.8252E+02, 9.8500E+02, 9.8749E+02, 9.8999E+02,
     +  9.9249E+02, 9.9499E+02, 9.9749E+02, 1.0000E+03, 1.0025E+03,
     + 1.0050E+03, 1.0075E+03, 1.0101E+03, 1.0126E+03, 1.0151E+03,
     +  1.0176E+03, 1.0202E+03, 1.0227E+03, 1.0252E+03, 1.0278E+03,
     + 1.0303E+03, 1.0329E+03, 1.0354E+03, 1.0380E+03, 1.0405E+03,
     +  1.0431E+03, 1.0456E+03, 1.0482E+03, 1.0508E+03, 1.0533E+03,
     + 1.0559E+03, 1.0585E+03, 1.0611E+03, 1.0637E+03, 1.0663E+03,
     +  1.0688E+03, 1.0714E+03, 1.0740E+03, 1.0766E+03, 1.0792E+03,
     + 1.0818E+03, 1.0844E+03, 1.0871E+03, 1.0897E+03, 1.0923E+03,
     +  1.0949E+03, 1.0975E+03, 1.1002E+03, 1.1028E+03, 1.1054E+03,
     + 1.1081E+03, 1.1107E+03, 1.1134E+03, 1.1160E+03, 1.1186E+03,
     +  1.1213E+03, 1.1240E+03, 1.1266E+03, 1.1293E+03, 1.1319E+03,
     + 1.1346E+03, 1.1373E+03, 1.1400E+03, 1.1426E+03, 1.1453E+03,
     +  1.1480E+03, 1.1507E+03, 1.1534E+03, 1.1561E+03, 1.1588E+03,
     + 1.1615E+03, 1.1642E+03, 1.1669E+03, 1.1696E+03, 1.1723E+03,
     +  1.1750E+03, 1.1777E+03, 1.1805E+03, 1.1832E+03, 1.1859E+03,
     + 1.1886E+03, 1.1914E+03, 1.1941E+03, 1.1969E+03, 1.1996E+03,
     +  1.2023E+03, 1.2051E+03, 1.2079E+03, 1.2106E+03, 1.2134E+03,
     + 1.2161E+03, 1.2189E+03, 1.2217E+03, 1.2244E+03, 1.2272E+03,
     +  1.2300E+03, 1.2328E+03, 1.2356E+03, 1.2384E+03, 1.2411E+03,
     + 1.2439E+03, 1.2467E+03, 1.2495E+03, 1.2523E+03, 1.2552E+03,
     +  1.2580E+03, 1.2608E+03, 1.2636E+03, 1.2664E+03, 1.2692E+03,
     + 1.2721E+03, 1.2749E+03, 1.2777E+03, 1.2806E+03, 1.2834E+03,
     +  1.2863E+03, 1.2891E+03, 1.2919E+03, 1.2948E+03, 1.2977E+03,
     + 1.3005E+03, 1.3034E+03, 1.3062E+03, 1.3091E+03, 1.3120E+03,
     +  1.3149E+03, 1.3177E+03, 1.3206E+03, 1.3235E+03, 1.3264E+03,
     + 1.3293E+03, 1.3322E+03, 1.3351E+03, 1.3380E+03, 1.3409E+03,
     +  1.3438E+03, 1.3467E+03, 1.3496E+03, 1.3525E+03, 1.3555E+03,
     + 1.3584E+03, 1.3613E+03, 1.3642E+03, 1.3672E+03, 1.3701E+03,
     +  1.3730E+03, 1.3760E+03, 1.3789E+03, 1.3819E+03, 1.3848E+03,
     + 1.3878E+03, 1.3907E+03, 1.3937E+03, 1.3967E+03, 1.3996E+03,
     +  1.4026E+03, 1.4056E+03, 1.4086E+03, 1.4116E+03, 1.4145E+03,
     + 1.4175E+03, 1.4205E+03, 1.4235E+03, 1.4265E+03, 1.4295E+03,
     +  1.4325E+03, 1.4355E+03, 1.4385E+03, 1.4416E+03, 1.4446E+03,
     + 1.4476E+03, 1.4506E+03, 1.4536E+03, 1.4567E+03, 1.4597E+03,
     +  1.4627E+03, 1.4658E+03, 1.4688E+03, 1.4719E+03, 1.4749E+03,
     + 1.4780E+03, 1.4810E+03, 1.4841E+03, 1.4872E+03, 1.4902E+03,
     +  1.4933E+03, 1.4964E+03, 1.4994E+03, 1.5025E+03, 1.5056E+03,
     + 1.5087E+03, 1.5118E+03, 1.5149E+03, 1.5180E+03, 1.5211E+03,
     +  1.5242E+03, 1.5273E+03, 1.5304E+03, 1.5335E+03, 1.5366E+03,
     + 1.5397E+03, 1.5429E+03, 1.5460E+03, 1.5491E+03, 1.5522E+03,
     +  1.5554E+03, 1.5585E+03, 1.5617E+03, 1.5648E+03, 1.5680E+03,
     + 1.5711E+03, 1.5743E+03, 1.5774E+03, 1.5806E+03, 1.5838E+03,
     +  1.5869E+03, 1.5901E+03, 1.5933E+03, 1.5964E+03, 1.5996E+03,
     + 1.6028E+03, 1.6060E+03, 1.6092E+03, 1.6124E+03, 1.6156E+03,
     +  1.6188E+03, 1.6220E+03, 1.6252E+03, 1.6284E+03, 1.6316E+03,
     + 1.6348E+03, 1.6381E+03, 1.6413E+03, 1.6445E+03, 1.6478E+03,
     +  1.6510E+03, 1.6542E+03, 1.6575E+03, 1.6607E+03, 1.6640E+03,
     + 1.6672E+03, 1.6705E+03, 1.6737E+03, 1.6770E+03, 1.6803E+03,
     +  1.6835E+03, 1.6868E+03, 1.6901E+03, 1.6933E+03, 1.6966E+03,
     + 1.6999E+03, 1.7032E+03, 1.7065E+03, 1.7098E+03, 1.7131E+03,
     +  1.7164E+03, 1.7197E+03, 1.7230E+03, 1.7263E+03, 1.7296E+03,
     + 1.7329E+03, 1.7363E+03, 1.7396E+03, 1.7429E+03, 1.7463E+03,
     +  1.7496E+03, 1.7529E+03, 1.7563E+03, 1.7596E+03, 1.7630E+03,
     + 1.7663E+03, 1.7697E+03, 1.7730E+03, 1.7764E+03, 1.7798E+03,
     +  1.7831E+03, 1.7865E+03, 1.7899E+03, 1.7932E+03, 1.7966E+03,
     + 1.8000E+03, 1.8034E+03, 1.8068E+03, 1.8102E+03, 1.8136E+03,
     +  1.8170E+03, 1.8204E+03, 1.8238E+03, 1.8272E+03, 1.8306E+03,
     + 1.8341E+03, 1.8375E+03, 1.8409E+03, 1.8443E+03, 1.8478E+03,
     +  1.8512E+03, 1.8546E+03, 1.8581E+03, 1.8615E+03, 1.8650E+03,
     + 1.8684E+03, 1.8719E+03, 1.8753E+03, 1.8788E+03, 1.8823E+03,
     +  1.8857E+03, 1.8892E+03, 1.8927E+03, 1.8962E+03, 1.8997E+03,
     + 1.9031E+03, 1.9066E+03, 1.9101E+03, 1.9136E+03, 1.9171E+03,
     +  1.9206E+03, 1.9241E+03, 1.9277E+03, 1.9312E+03, 1.9347E+03,
     + 1.9382E+03, 1.9417E+03, 1.9453E+03, 1.9488E+03, 1.9523E+03,
     +  1.9559E+03, 1.9594E+03, 1.9629E+03, 1.9665E+03, 1.9700E+03,
     + 1.9736E+03, 1.9772E+03, 1.9807E+03, 1.9843E+03, 1.9879E+03,
     +  1.9914E+03, 1.9950E+03, 1.9986E+03, 2.0022E+03, 2.0058E+03,
     + 2.0093E+03, 2.0129E+03, 2.0165E+03, 2.0201E+03, 2.0237E+03,
     +  2.0273E+03, 2.0310E+03, 2.0346E+03, 2.0382E+03, 2.0418E+03,
     + 2.0454E+03, 2.0491E+03, 2.0527E+03, 2.0563E+03, 2.0600E+03,
     +  2.0636E+03, 2.0672E+03, 2.0709E+03, 2.0745E+03, 2.0782E+03,
     + 2.0819E+03, 2.0855E+03, 2.0892E+03, 2.0928E+03, 2.0965E+03,
     +  2.1002E+03, 2.1039E+03, 2.1076E+03, 2.1112E+03, 2.1149E+03,
     + 2.1186E+03, 2.1223E+03, 2.1260E+03, 2.1297E+03, 2.1334E+03,
     +  2.1371E+03, 2.1409E+03, 2.1446E+03, 2.1483E+03, 2.1520E+03,
     + 2.1557E+03, 2.1595E+03, 2.1632E+03, 2.1669E+03, 2.1707E+03,
     +  2.1744E+03, 2.1782E+03, 2.1819E+03, 2.1857E+03, 2.1894E+03,
     + 2.1932E+03, 2.1970E+03, 2.2007E+03, 2.2045E+03, 2.2083E+03,
     +  2.2121E+03, 2.2159E+03, 2.2196E+03, 2.2234E+03, 2.2272E+03,
     + 2.2310E+03, 2.2348E+03, 2.2386E+03, 2.2424E+03, 2.2462E+03,
     +  2.2501E+03, 2.2539E+03, 2.2577E+03, 2.2615E+03, 2.2654E+03,
     + 2.2692E+03, 2.2730E+03, 2.2769E+03, 2.2807E+03, 2.2845E+03,
     +  2.2884E+03, 2.2923E+03, 2.2961E+03, 2.3000E+03, 2.3038E+03,
     + 2.3077E+03, 2.3116E+03, 2.3154E+03, 2.3193E+03, 2.3232E+03,
     +  2.3271E+03, 2.3310E+03, 2.3349E+03, 2.3388E+03, 2.3427E+03,
     + 2.3466E+03, 2.3505E+03, 2.3544E+03, 2.3583E+03, 2.3622E+03,
     +  2.3661E+03, 2.3701E+03, 2.3740E+03, 2.3779E+03, 2.3818E+03,
     + 2.3858E+03, 2.3897E+03, 2.3937E+03, 2.3976E+03, 2.4016E+03,
     +  2.4055E+03, 2.4095E+03, 2.4134E+03, 2.4174E+03, 2.4214E+03,
     + 2.4254E+03, 2.4293E+03, 2.4333E+03, 2.4373E+03, 2.4413E+03,
     +  2.4453E+03, 2.4493E+03, 2.4533E+03, 2.4573E+03, 2.4613E+03,
     + 2.4653E+03, 2.4693E+03, 2.4733E+03, 2.4773E+03, 2.4814E+03,
     +  2.4854E+03, 2.4894E+03, 2.4934E+03, 2.4975E+03, 2.5015E+03,
     + 2.5056E+03, 2.5096E+03, 2.5137E+03, 2.5177E+03, 2.5218E+03,
     +  2.5258E+03, 2.5299E+03, 2.5340E+03, 2.5380E+03, 2.5421E+03,
     + 2.5462E+03, 2.5503E+03, 2.5544E+03, 2.5585E+03, 2.5626E+03,
     +  2.5667E+03, 2.5708E+03, 2.5749E+03, 2.5790E+03, 2.5831E+03,
     + 2.5872E+03, 2.5913E+03, 2.5954E+03, 2.5996E+03, 2.6037E+03,
     +  2.6078E+03, 2.6120E+03, 2.6161E+03, 2.6203E+03, 2.6244E+03,
     + 2.6286E+03, 2.6327E+03, 2.6369E+03, 2.6410E+03, 2.6452E+03,
     +  2.6494E+03, 2.6535E+03, 2.6577E+03, 2.6619E+03, 2.6661E+03,
     + 2.6703E+03, 2.6745E+03, 2.6787E+03, 2.6829E+03, 2.6871E+03,
     +  2.6913E+03, 2.6955E+03, 2.6997E+03, 2.7039E+03, 2.7081E+03,
     + 2.7124E+03, 2.7166E+03, 2.7208E+03, 2.7250E+03, 2.7293E+03,
     +  2.7335E+03, 2.7378E+03, 2.7420E+03, 2.7463E+03, 2.7505E+03,
     + 2.7548E+03, 2.7591E+03, 2.7633E+03, 2.7676E+03, 2.7719E+03,
     +  2.7762E+03, 2.7804E+03, 2.7847E+03, 2.7890E+03, 2.7933E+03,
     + 2.7976E+03, 2.8019E+03, 2.8062E+03, 2.8105E+03, 2.8148E+03,
     +  2.8191E+03, 2.8235E+03, 2.8278E+03, 2.8321E+03, 2.8364E+03,
     + 2.8408E+03, 2.8451E+03, 2.8494E+03, 2.8538E+03, 2.8581E+03,
     +  2.8625E+03, 2.8668E+03, 2.8712E+03, 2.8756E+03, 2.8799E+03,
     + 2.8843E+03, 2.8887E+03, 2.8931E+03, 2.8974E+03, 2.9018E+03,
     +  2.9062E+03, 2.9106E+03, 2.9150E+03, 2.9194E+03, 2.9238E+03,
     + 2.9282E+03, 2.9326E+03, 2.9370E+03, 2.9414E+03, 2.9459E+03,
     +  2.9503E+03, 2.9547E+03, 2.9592E+03, 2.9636E+03, 2.9680E+03,
     + 2.9725E+03, 2.9769E+03, 2.9814E+03, 2.9858E+03, 2.9903E+03,
     +  2.9948E+03, 2.9992E+03, 3.0037E+03, 3.0082E+03, 3.0126E+03,
     + 3.0171E+03, 3.0216E+03, 3.0261E+03, 3.0306E+03, 3.0351E+03,
     +  3.0396E+03, 3.0441E+03, 3.0486E+03, 3.0531E+03, 3.0576E+03,
     + 3.0621E+03, 3.0666E+03, 3.0712E+03, 3.0757E+03, 3.0802E+03,
     +  3.0848E+03, 3.0893E+03, 3.0939E+03, 3.0984E+03, 3.1030E+03,
     + 3.1075E+03, 3.1121E+03, 3.1166E+03, 3.1212E+03, 3.1258E+03,
     +  3.1303E+03, 3.1349E+03, 3.1395E+03, 3.1441E+03, 3.1487E+03,
     + 3.1533E+03, 3.1579E+03, 3.1625E+03, 3.1671E+03, 3.1717E+03,
     +  3.1763E+03, 3.1809E+03, 3.1855E+03, 3.1901E+03, 3.1948E+03,
     + 3.1994E+03, 3.2040E+03, 3.2087E+03, 3.2133E+03, 3.2179E+03,
     +  3.2226E+03, 3.2272E+03, 3.2319E+03, 3.2366E+03, 3.2412E+03,
     + 3.2459E+03, 3.2506E+03, 3.2552E+03, 3.2599E+03, 3.2646E+03,
     +  3.2693E+03, 3.2740E+03, 3.2787E+03, 3.2834E+03, 3.2881E+03,
     + 3.2928E+03, 3.2975E+03, 3.3022E+03, 3.3069E+03, 3.3116E+03,
     +  3.3163E+03, 3.3211E+03, 3.3258E+03, 3.3305E+03, 3.3353E+03,
     + 3.3400E+03, 3.3448E+03, 3.3495E+03, 3.3543E+03, 3.3590E+03,
     +  3.3638E+03, 3.3686E+03, 3.3733E+03, 3.3781E+03, 3.3829E+03,
     + 3.3877E+03, 3.3924E+03, 3.3972E+03, 3.4020E+03, 3.4068E+03,
     +  3.4116E+03, 3.4164E+03, 3.4212E+03, 3.4260E+03, 3.4308E+03,
     + 3.4357E+03, 3.4405E+03, 3.4453E+03, 3.4501E+03, 3.4550E+03,
     +  3.4598E+03, 3.4647E+03, 3.4695E+03, 3.4743E+03, 3.4792E+03,
     + 3.4841E+03, 3.4889E+03, 3.4938E+03, 3.4986E+03, 3.5035E+03,
     +  3.5084E+03, 3.5133E+03, 3.5182E+03, 3.5230E+03, 3.5279E+03,
     + 3.5328E+03, 3.5377E+03, 3.5426E+03, 3.5475E+03, 3.5524E+03,
     +  3.5574E+03, 3.5623E+03, 3.5672E+03, 3.5721E+03, 3.5770E+03,
     + 3.5820E+03, 3.5869E+03, 3.5919E+03, 3.5968E+03, 3.6017E+03,
     +  3.6067E+03, 3.6117E+03, 3.6166E+03, 3.6216E+03, 3.6265E+03,
     + 3.6315E+03, 3.6365E+03, 3.6415E+03, 3.6465E+03, 3.6514E+03,
     +  3.6564E+03, 3.6614E+03, 3.6664E+03, 3.6714E+03, 3.6764E+03,
     + 3.6814E+03, 3.6864E+03, 3.6915E+03, 3.6965E+03, 3.7015E+03,
     +  3.7065E+03, 3.7116E+03, 3.7166E+03, 3.7216E+03, 3.7267E+03,
     + 3.7317E+03, 3.7368E+03, 3.7418E+03, 3.7469E+03, 3.7520E+03,
     +  3.7570E+03, 3.7621E+03, 3.7672E+03, 3.7723E+03, 3.7773E+03,
     + 3.7824E+03, 3.7875E+03, 3.7926E+03, 3.7977E+03, 3.8028E+03,
     +  3.8079E+03, 3.8130E+03, 3.8181E+03, 3.8233E+03, 3.8284E+03,
     + 3.8335E+03, 3.8386E+03, 3.8438E+03, 3.8489E+03, 3.8541E+03,
     +  3.8592E+03, 3.8643E+03, 3.8695E+03, 3.8747E+03, 3.8798E+03,
     + 3.8850E+03, 3.8901E+03, 3.8953E+03, 3.9005E+03, 3.9057E+03,
     +  3.9109E+03, 3.9160E+03, 3.9212E+03, 3.9264E+03, 3.9316E+03,
     + 3.9368E+03, 3.9420E+03, 3.9473E+03, 3.9525E+03, 3.9577E+03,
     +  3.9629E+03, 3.9681E+03, 3.9734E+03, 3.9786E+03, 3.9838E+03,
     + 3.9891E+03, 3.9943E+03, 3.9996E+03, 4.0048E+03, 4.0101E+03,
     +  4.0154E+03, 4.0206E+03, 4.0259E+03, 4.0312E+03, 4.0364E+03,
     + 4.0417E+03, 4.0470E+03, 4.0523E+03, 4.0576E+03, 4.0629E+03,
     +  4.0682E+03, 4.0735E+03, 4.0788E+03, 4.0841E+03, 4.0894E+03,
     + 4.0947E+03, 4.1001E+03, 4.1054E+03, 4.1107E+03, 4.1161E+03,
     +  4.1214E+03, 4.1268E+03, 4.1321E+03, 4.1375E+03, 4.1428E+03,
     + 4.1482E+03, 4.1535E+03, 4.1589E+03, 4.1643E+03, 4.1696E+03,
     +  4.1750E+03, 4.1804E+03, 4.1858E+03, 4.1912E+03, 4.1966E+03,
     + 4.2020E+03, 4.2074E+03, 4.2128E+03, 4.2182E+03, 4.2236E+03,
     +  4.2290E+03, 4.2345E+03, 4.2399E+03, 4.2453E+03, 4.2508E+03,
     + 4.2562E+03, 4.2616E+03, 4.2671E+03, 4.2725E+03, 4.2780E+03,
     +  4.2834E+03, 4.2889E+03, 4.2944E+03, 4.2998E+03, 4.3053E+03,
     + 4.3108E+03, 4.3163E+03, 4.3218E+03, 4.3273E+03, 4.3328E+03,
     +  4.3383E+03, 4.3438E+03, 4.3493E+03, 4.3548E+03, 4.3603E+03,
     + 4.3658E+03, 4.3713E+03, 4.3769E+03, 4.3824E+03, 4.3879E+03,
     +  4.3935E+03, 4.3990E+03, 4.4045E+03, 4.4101E+03, 4.4156E+03,
     + 4.4212E+03, 4.4268E+03, 4.4323E+03, 4.4379E+03, 4.4435E+03,
     +  4.4491E+03, 4.4546E+03, 4.4602E+03, 4.4658E+03, 4.4714E+03,
     + 4.4770E+03, 4.4826E+03, 4.4882E+03, 4.4938E+03, 4.4994E+03,
     +  4.5051E+03, 4.5107E+03, 4.5163E+03, 4.5219E+03, 4.5276E+03,
     + 4.5332E+03, 4.5388E+03, 4.5445E+03, 4.5501E+03, 4.5558E+03,
     +  4.5615E+03, 4.5671E+03, 4.5728E+03, 4.5785E+03, 4.5841E+03,
     + 4.5898E+03, 4.5955E+03, 4.6012E+03, 4.6069E+03, 4.6126E+03,
     +  4.6183E+03, 4.6240E+03, 4.6297E+03, 4.6354E+03, 4.6411E+03,
     + 4.6468E+03, 4.6525E+03, 4.6583E+03, 4.6640E+03, 4.6697E+03,
     +  4.6755E+03, 4.6812E+03, 4.6869E+03, 4.6927E+03, 4.6984E+03,
     + 4.7042E+03, 4.7100E+03, 4.7157E+03, 4.7215E+03, 4.7273E+03,
     +  4.7331E+03, 4.7388E+03, 4.7446E+03, 4.7504E+03, 4.7562E+03,
     + 4.7620E+03, 4.7678E+03, 4.7736E+03, 4.7794E+03, 4.7852E+03,
     +  4.7911E+03, 4.7969E+03, 4.8027E+03, 4.8085E+03, 4.8144E+03,
     + 4.8202E+03, 4.8261E+03, 4.8319E+03, 4.8378E+03, 4.8436E+03,
     +  4.8495E+03, 4.8553E+03, 4.8612E+03, 4.8671E+03, 4.8730E+03,
     + 4.8788E+03, 4.8847E+03, 4.8906E+03, 4.8965E+03, 4.9024E+03,
     +  4.9083E+03, 4.9142E+03, 4.9201E+03, 4.9260E+03, 4.9319E+03,
     + 4.9379E+03, 4.9438E+03, 4.9497E+03, 4.9556E+03, 4.9616E+03,
     +  4.9675E+03, 4.9735E+03, 4.9794E+03, 4.9854E+03, 4.9913E+03,
     + 4.9973E+03, 5.0032E+03, 5.0092E+03, 5.0152E+03, 5.0212E+03,
     +  5.0271E+03, 5.0331E+03, 5.0391E+03, 5.0451E+03, 5.0511E+03,
     + 5.0571E+03, 5.0631E+03, 5.0691E+03, 5.0751E+03, 5.0812E+03,
     +  5.0872E+03, 5.0932E+03, 5.0992E+03, 5.1053E+03, 5.1113E+03,
     + 5.1174E+03, 5.1234E+03, 5.1294E+03, 5.1355E+03, 5.1416E+03,
     +  5.1476E+03, 5.1537E+03, 5.1598E+03, 5.1658E+03, 5.1719E+03,
     + 5.1780E+03, 5.1841E+03, 5.1902E+03, 5.1963E+03, 5.2024E+03,
     +  5.2085E+03, 5.2146E+03, 5.2207E+03, 5.2268E+03, 5.2329E+03,
     + 5.2391E+03, 5.2452E+03, 5.2513E+03, 5.2575E+03, 5.2636E+03,
     +  5.2697E+03, 5.2759E+03, 5.2820E+03, 5.2882E+03, 5.2944E+03,
     + 5.3005E+03, 5.3067E+03, 5.3129E+03, 5.3190E+03, 5.3252E+03,
     +  5.3314E+03, 5.3376E+03, 5.3438E+03, 5.3500E+03, 5.3562E+03,
     + 5.3624E+03, 5.3686E+03, 5.3748E+03, 5.3811E+03, 5.3873E+03,
     +  5.3935E+03, 5.3997E+03, 5.4060E+03, 5.4122E+03, 5.4184E+03,
     + 5.4247E+03, 5.4309E+03, 5.4372E+03, 5.4435E+03, 5.4497E+03,
     +  5.4560E+03, 5.4623E+03, 5.4685E+03, 5.4748E+03, 5.4811E+03,
     + 5.4874E+03, 5.4937E+03, 5.5000E+03, 5.5063E+03, 5.5126E+03,
     +  5.5189E+03, 5.5252E+03, 5.5315E+03, 5.5379E+03, 5.5442E+03,
     + 5.5505E+03, 5.5569E+03, 5.5632E+03, 5.5695E+03, 5.5759E+03,
     +  5.5822E+03, 5.5886E+03, 5.5949E+03, 5.6013E+03, 5.6077E+03,
     + 5.6140E+03, 5.6204E+03, 5.6268E+03, 5.6332E+03, 5.6396E+03,
     +  5.6460E+03, 5.6524E+03, 5.6588E+03, 5.6652E+03, 5.6716E+03,
     + 5.6780E+03, 5.6844E+03, 5.6908E+03, 5.6973E+03, 5.7037E+03,
     +  5.7101E+03, 5.7166E+03, 5.7230E+03, 5.7294E+03, 5.7359E+03,
     + 5.7424E+03, 5.7488E+03, 5.7553E+03, 5.7617E+03, 5.7682E+03,
     +  5.7747E+03, 5.7812E+03, 5.7877E+03, 5.7941E+03, 5.8006E+03,
     + 5.8071E+03, 5.8136E+03, 5.8201E+03, 5.8266E+03, 5.8332E+03,
     +  5.8397E+03, 5.8462E+03, 5.8527E+03, 5.8592E+03, 5.8658E+03,
     + 5.8723E+03, 5.8789E+03, 5.8854E+03, 5.8920E+03, 5.8985E+03,
     +  5.9051E+03, 5.9116E+03, 5.9182E+03, 5.9248E+03, 5.9313E+03,
     + 5.9379E+03, 5.9445E+03, 5.9511E+03, 5.9577E+03, 5.9643E+03,
     +  5.9709E+03, 5.9775E+03, 5.9841E+03, 5.9907E+03, 5.9973E+03,
     + 6.0040E+03, 6.0106E+03, 6.0172E+03, 6.0238E+03, 6.0305E+03,
     +  6.0371E+03, 6.0438E+03, 6.0504E+03, 6.0571E+03, 6.0637E+03,
     + 6.0704E+03, 6.0771E+03, 6.0837E+03, 6.0904E+03, 6.0971E+03,
     +  6.1038E+03, 6.1105E+03, 6.1172E+03, 6.1239E+03, 6.1306E+03,
     + 6.1373E+03, 6.1440E+03, 6.1507E+03, 6.1574E+03, 6.1641E+03,
     +  6.1709E+03, 6.1776E+03, 6.1843E+03, 6.1911E+03, 6.1978E+03,
     + 6.2045E+03, 6.2113E+03, 6.2181E+03, 6.2248E+03, 6.2316E+03,
     +  6.2383E+03, 6.2451E+03, 6.2519E+03, 6.2587E+03, 6.2655E+03,
     + 6.2723E+03, 6.2790E+03, 6.2858E+03, 6.2926E+03, 6.2995E+03,
     +  6.3063E+03, 6.3131E+03, 6.3199E+03, 6.3267E+03, 6.3335E+03,
     + 6.3404E+03, 6.3472E+03, 6.3541E+03, 6.3609E+03, 6.3677E+03,
     +  6.3746E+03, 6.3815E+03, 6.3883E+03, 6.3952E+03, 6.4021E+03,
     + 6.4089E+03, 6.4158E+03, 6.4227E+03, 6.4296E+03, 6.4365E+03,
     +  6.4434E+03, 6.4503E+03, 6.4572E+03, 6.4641E+03, 6.4710E+03,
     + 6.4779E+03, 6.4848E+03, 6.4917E+03, 6.4987E+03, 6.5056E+03,
     +  6.5125E+03, 6.5195E+03, 6.5264E+03, 6.5334E+03, 6.5403E+03,
     + 6.5473E+03, 6.5542E+03, 6.5612E+03, 6.5682E+03, 6.5752E+03,
     +  6.5821E+03, 6.5891E+03, 6.5961E+03, 6.6031E+03, 6.6101E+03,
     + 6.6171E+03, 6.6241E+03, 6.6311E+03, 6.6381E+03, 6.6451E+03,
     +  6.6522E+03, 6.6592E+03, 6.6662E+03, 6.6733E+03, 6.6803E+03,
     + 6.6873E+03, 6.6944E+03, 6.7014E+03, 6.7085E+03, 6.7156E+03,
     +  6.7226E+03, 6.7297E+03, 6.7368E+03, 6.7438E+03, 6.7509E+03,
     + 6.7580E+03, 6.7651E+03, 6.7722E+03, 6.7793E+03, 6.7864E+03,
     +  6.7935E+03, 6.8006E+03, 6.8077E+03, 6.8148E+03, 6.8220E+03,
     + 6.8291E+03, 6.8362E+03, 6.8434E+03, 6.8505E+03, 6.8576E+03,
     +  6.8648E+03, 6.8719E+03, 6.8791E+03, 6.8863E+03, 6.8934E+03,
     + 6.9006E+03, 6.9078E+03, 6.9150E+03, 6.9221E+03, 6.9293E+03,
     +  6.9365E+03, 6.9437E+03, 6.9509E+03, 6.9581E+03, 6.9653E+03,
     + 6.9725E+03, 6.9798E+03, 6.9870E+03, 6.9942E+03, 7.0014E+03,
     +  7.0087E+03, 7.0159E+03, 7.0231E+03, 7.0304E+03, 7.0376E+03,
     + 7.0449E+03, 7.0522E+03, 7.0594E+03, 7.0667E+03, 7.0740E+03,
     +  7.0812E+03, 7.0885E+03, 7.0958E+03, 7.1031E+03, 7.1104E+03,
     + 7.1177E+03, 7.1250E+03, 7.1323E+03, 7.1396E+03, 7.1469E+03,
     +  7.1542E+03, 7.1616E+03, 7.1689E+03, 7.1762E+03, 7.1836E+03,
     + 7.1909E+03, 7.1983E+03, 7.2056E+03, 7.2130E+03, 7.2203E+03,
     +  7.2277E+03, 7.2350E+03, 7.2424E+03, 7.2498E+03, 7.2572E+03,
     + 7.2646E+03, 7.2719E+03, 7.2793E+03, 7.2867E+03, 7.2941E+03,
     +  7.3015E+03, 7.3089E+03, 7.3164E+03, 7.3238E+03, 7.3312E+03,
     + 7.3386E+03, 7.3461E+03, 7.3535E+03, 7.3609E+03, 7.3684E+03,
     +  7.3758E+03, 7.3833E+03, 7.3907E+03, 7.3982E+03, 7.4057E+03,
     + 7.4131E+03, 7.4206E+03, 7.4281E+03, 7.4356E+03, 7.4430E+03,
     +  7.4505E+03, 7.4580E+03, 7.4655E+03, 7.4730E+03, 7.4805E+03,
     + 7.4881E+03, 7.4956E+03, 7.5031E+03, 7.5106E+03, 7.5181E+03,
     +  7.5257E+03, 7.5332E+03, 7.5408E+03, 7.5483E+03, 7.5559E+03,
     + 7.5634E+03, 7.5710E+03, 7.5785E+03, 7.5861E+03, 7.5937E+03,
     +  7.6013E+03, 7.6088E+03, 7.6164E+03, 7.6240E+03, 7.6316E+03,
     + 7.6392E+03, 7.6468E+03, 7.6544E+03, 7.6620E+03, 7.6696E+03,
     +  7.6773E+03, 7.6849E+03, 7.6925E+03, 7.7001E+03, 7.7078E+03,
     + 7.7154E+03, 7.7231E+03, 7.7307E+03, 7.7384E+03, 7.7460E+03,
     +  7.7537E+03, 7.7614E+03, 7.7690E+03, 7.7767E+03, 7.7844E+03,
     + 7.7921E+03, 7.7998E+03, 7.8075E+03, 7.8151E+03, 7.8229E+03,
     +  7.8306E+03, 7.8383E+03, 7.8460E+03, 7.8537E+03, 7.8614E+03,
     + 7.8691E+03, 7.8769E+03, 7.8846E+03, 7.8924E+03, 7.9001E+03,
     +  7.9079E+03, 7.9156E+03, 7.9234E+03, 7.9311E+03, 7.9389E+03,
     + 7.9467E+03, 7.9544E+03, 7.9622E+03, 7.9700E+03, 7.9778E+03,
     +  7.9856E+03, 7.9934E+03, 8.0012E+03, 8.0090E+03, 8.0168E+03,
     + 8.0246E+03, 8.0324E+03, 8.0402E+03, 8.0481E+03, 8.0559E+03,
     +  8.0637E+03, 8.0716E+03, 8.0794E+03, 8.0873E+03, 8.0951E+03,
     + 8.1030E+03, 8.1108E+03, 8.1187E+03, 8.1266E+03, 8.1344E+03,
     +  8.1423E+03, 8.1502E+03, 8.1581E+03, 8.1660E+03, 8.1739E+03,
     + 8.1818E+03, 8.1897E+03, 8.1976E+03, 8.2055E+03, 8.2134E+03,
     +  8.2213E+03, 8.2293E+03, 8.2372E+03, 8.2451E+03, 8.2531E+03,
     + 8.2610E+03, 8.2690E+03, 8.2769E+03, 8.2849E+03, 8.2928E+03,
     +  8.3008E+03, 8.3088E+03, 8.3167E+03, 8.3247E+03, 8.3327E+03,
     + 8.3407E+03, 8.3487E+03, 8.3567E+03, 8.3647E+03, 8.3727E+03,
     +  8.3807E+03, 8.3887E+03, 8.3967E+03, 8.4047E+03, 8.4128E+03,
     + 8.4208E+03, 8.4288E+03, 8.4369E+03, 8.4449E+03, 8.4530E+03,
     +  8.4610E+03, 8.4691E+03, 8.4771E+03, 8.4852E+03, 8.4933E+03,
     + 8.5013E+03, 8.5094E+03, 8.5175E+03, 8.5256E+03, 8.5337E+03,
     +  8.5418E+03, 8.5499E+03, 8.5580E+03, 8.5661E+03, 8.5742E+03,
     + 8.5823E+03, 8.5904E+03, 8.5985E+03, 8.6067E+03, 8.6148E+03,
     +  8.6230E+03, 8.6311E+03, 8.6392E+03, 8.6474E+03, 8.6555E+03,
     + 8.6637E+03, 8.6719E+03, 8.6800E+03, 8.6882E+03, 8.6964E+03,
     +  8.7046E+03, 8.7128E+03, 8.7209E+03, 8.7291E+03, 8.7373E+03,
     + 8.7455E+03, 8.7538E+03, 8.7620E+03, 8.7702E+03, 8.7784E+03,
     +  8.7866E+03, 8.7949E+03, 8.8031E+03, 8.8113E+03, 8.8196E+03,
     + 8.8278E+03, 8.8361E+03, 8.8443E+03, 8.8526E+03, 8.8609E+03,
     +  8.8691E+03, 8.8774E+03, 8.8857E+03, 8.8940E+03, 8.9022E+03,
     + 8.9105E+03, 8.9188E+03, 8.9271E+03, 8.9354E+03, 8.9437E+03,
     +  8.9520E+03, 8.9604E+03, 8.9687E+03, 8.9770E+03, 8.9853E+03,
     + 8.9937E+03, 9.0020E+03, 9.0103E+03, 9.0187E+03, 9.0270E+03,
     +  9.0354E+03, 9.0438E+03, 9.0521E+03, 9.0605E+03, 9.0689E+03,
     + 9.0772E+03, 9.0856E+03, 9.0940E+03, 9.1024E+03, 9.1108E+03,
     +  9.1192E+03, 9.1276E+03, 9.1360E+03, 9.1444E+03, 9.1528E+03,
     + 9.1613E+03, 9.1697E+03, 9.1781E+03, 9.1865E+03, 9.1950E+03,
     +  9.2034E+03, 9.2119E+03, 9.2203E+03, 9.2288E+03, 9.2372E+03,
     + 9.2457E+03, 9.2542E+03, 9.2626E+03, 9.2711E+03, 9.2796E+03,
     +  9.2881E+03, 9.2966E+03, 9.3051E+03, 9.3136E+03, 9.3221E+03,
     + 9.3306E+03, 9.3391E+03, 9.3476E+03, 9.3561E+03, 9.3647E+03,
     +  9.3732E+03, 9.3817E+03, 9.3903E+03, 9.3988E+03, 9.4074E+03,
     + 9.4159E+03, 9.4245E+03, 9.4330E+03, 9.4416E+03, 9.4502E+03,
     +  9.4587E+03, 9.4673E+03, 9.4759E+03, 9.4845E+03, 9.4931E+03,
     + 9.5017E+03, 9.5103E+03, 9.5189E+03, 9.5275E+03, 9.5361E+03,
     +  9.5447E+03, 9.5533E+03, 9.5620E+03, 9.5706E+03, 9.5792E+03,
     + 9.5879E+03, 9.5965E+03, 9.6051E+03, 9.6138E+03, 9.6225E+03,
     +  9.6311E+03, 9.6398E+03, 9.6484E+03, 9.6571E+03, 9.6658E+03,
     + 9.6745E+03, 9.6832E+03, 9.6919E+03, 9.7006E+03, 9.7093E+03,
     +  9.7180E+03, 9.7267E+03, 9.7354E+03, 9.7441E+03, 9.7528E+03,
     + 9.7615E+03, 9.7703E+03, 9.7790E+03, 9.7878E+03, 9.7965E+03,
     +  9.8052E+03, 9.8140E+03, 9.8228E+03, 9.8315E+03, 9.8403E+03,
     + 9.8490E+03, 9.8578E+03, 9.8666E+03, 9.8754E+03, 9.8842E+03,
     +  9.8930E+03, 9.9018E+03, 9.9106E+03, 9.9194E+03, 9.9282E+03,
     + 9.9370E+03, 9.9458E+03, 9.9546E+03, 9.9634E+03, 9.9723E+03,
     +  9.9811E+03, 9.9900E+03, 9.9988E+03, 1.0008E+04, 1.0016E+04,
     + 1.0025E+04, 1.0034E+04, 1.0043E+04, 1.0052E+04, 1.0061E+04,
     +  1.0070E+04, 1.0079E+04, 1.0087E+04, 1.0096E+04, 1.0105E+04,
     + 1.0114E+04, 1.0123E+04, 1.0132E+04, 1.0141E+04, 1.0150E+04,
     +  1.0159E+04, 1.0168E+04, 1.0177E+04, 1.0186E+04, 1.0194E+04,
     + 1.0203E+04, 1.0212E+04, 1.0221E+04, 1.0230E+04, 1.0239E+04,
     +  1.0248E+04, 1.0257E+04, 1.0266E+04, 1.0275E+04, 1.0284E+04,
     + 1.0293E+04, 1.0302E+04, 1.0311E+04, 1.0320E+04, 1.0329E+04,
     +  1.0338E+04, 1.0347E+04, 1.0356E+04, 1.0365E+04, 1.0374E+04,
     + 1.0383E+04, 1.0392E+04, 1.0401E+04, 1.0410E+04, 1.0419E+04,
     +  1.0428E+04, 1.0438E+04, 1.0447E+04, 1.0456E+04, 1.0465E+04,
     + 1.0474E+04, 1.0483E+04, 1.0492E+04, 1.0501E+04, 1.0510E+04,
     +  1.0519E+04, 1.0528E+04, 1.0537E+04, 1.0547E+04, 1.0556E+04,
     + 1.0565E+04, 1.0574E+04, 1.0583E+04, 1.0592E+04, 1.0601E+04,
     +  1.0610E+04, 1.0620E+04, 1.0629E+04, 1.0638E+04, 1.0647E+04,
     + 1.0656E+04, 1.0665E+04, 1.0675E+04, 1.0684E+04, 1.0693E+04,
     +  1.0702E+04, 1.0711E+04, 1.0720E+04, 1.0730E+04, 1.0739E+04,
     + 1.0748E+04, 1.0757E+04, 1.0767E+04, 1.0776E+04, 1.0785E+04,
     +  1.0794E+04, 1.0803E+04, 1.0813E+04, 1.0822E+04, 1.0831E+04,
     + 1.0840E+04, 1.0850E+04, 1.0859E+04, 1.0868E+04, 1.0877E+04,
     +  1.0887E+04, 1.0896E+04, 1.0905E+04, 1.0915E+04, 1.0924E+04,
     + 1.0933E+04, 1.0942E+04, 1.0952E+04, 1.0961E+04, 1.0970E+04,
     +  1.0980E+04, 1.0989E+04, 1.0998E+04, 1.1008E+04, 1.1017E+04,
     + 1.1026E+04, 1.1036E+04, 1.1045E+04, 1.1054E+04, 1.1064E+04,
     +  1.1073E+04, 1.1082E+04, 1.1092E+04, 1.1101E+04, 1.1111E+04,
     + 1.1120E+04, 1.1129E+04, 1.1139E+04, 1.1148E+04, 1.1158E+04,
     +  1.1167E+04, 1.1176E+04, 1.1186E+04, 1.1195E+04, 1.1205E+04,
     + 1.1214E+04, 1.1223E+04, 1.1233E+04, 1.1242E+04, 1.1252E+04,
     +  1.1261E+04, 1.1271E+04, 1.1280E+04, 1.1290E+04, 1.1299E+04,
     + 1.1308E+04, 1.1318E+04, 1.1327E+04, 1.1337E+04, 1.1346E+04,
     +  1.1356E+04, 1.1365E+04, 1.1375E+04, 1.1384E+04, 1.1394E+04,
     + 1.1403E+04, 1.1413E+04, 1.1422E+04, 1.1432E+04, 1.1442E+04,
     +  1.1451E+04, 1.1461E+04, 1.1470E+04, 1.1480E+04, 1.1489E+04,
     + 1.1499E+04, 1.1508E+04, 1.1518E+04, 1.1527E+04, 1.1537E+04,
     +  1.1547E+04, 1.1556E+04, 1.1566E+04, 1.1575E+04, 1.1585E+04,
     + 1.1595E+04, 1.1604E+04, 1.1614E+04, 1.1623E+04, 1.1633E+04,
     +  1.1643E+04, 1.1652E+04, 1.1662E+04, 1.1672E+04, 1.1681E+04,
     + 1.1691E+04, 1.1700E+04, 1.1710E+04, 1.1720E+04, 1.1729E+04,
     +  1.1739E+04, 1.1749E+04, 1.1758E+04, 1.1768E+04, 1.1778E+04,
     + 1.1788E+04, 1.1797E+04, 1.1807E+04, 1.1817E+04, 1.1826E+04,
     +  1.1836E+04, 1.1846E+04, 1.1855E+04, 1.1865E+04, 1.1875E+04,
     + 1.1885E+04, 1.1894E+04, 1.1904E+04, 1.1914E+04, 1.1924E+04,
     +  1.1933E+04, 1.1943E+04, 1.1953E+04, 1.1963E+04, 1.1972E+04,
     + 1.1982E+04, 1.1992E+04, 1.2002E+04, 1.2012E+04, 1.2021E+04,
     +  1.2031E+04, 1.2041E+04, 1.2051E+04, 1.2061E+04, 1.2070E+04,
     + 1.2080E+04, 1.2090E+04, 1.2100E+04, 1.2110E+04, 1.2120E+04,
     +  1.2129E+04, 1.2139E+04, 1.2149E+04, 1.2159E+04, 1.2169E+04/
      data (QofT( 4,J),J=2409,4510)/ 1.2179E+04, 1.2188E+04,
     +  1.2198E+04, 1.2208E+04, 1.2218E+04, 1.2228E+04, 1.2238E+04,
     +  1.2248E+04,
     + 1.2258E+04, 1.2268E+04, 1.2277E+04, 1.2287E+04, 1.2297E+04,
     +  1.2307E+04, 1.2317E+04, 1.2327E+04, 1.2337E+04, 1.2347E+04,
     + 1.2357E+04, 1.2367E+04, 1.2377E+04, 1.2387E+04, 1.2397E+04,
     +  1.2407E+04, 1.2417E+04, 1.2427E+04, 1.2437E+04, 1.2447E+04,
     + 1.2457E+04, 1.2467E+04, 1.2477E+04, 1.2487E+04, 1.2497E+04,
     +  1.2507E+04, 1.2517E+04, 1.2527E+04, 1.2537E+04, 1.2547E+04,
     + 1.2557E+04, 1.2567E+04, 1.2577E+04, 1.2587E+04, 1.2597E+04,
     +  1.2607E+04, 1.2617E+04, 1.2627E+04, 1.2637E+04, 1.2647E+04,
     + 1.2657E+04, 1.2667E+04, 1.2677E+04, 1.2687E+04, 1.2697E+04,
     +  1.2708E+04, 1.2718E+04, 1.2728E+04, 1.2738E+04, 1.2748E+04,
     + 1.2758E+04, 1.2768E+04, 1.2778E+04, 1.2788E+04, 1.2799E+04,
     +  1.2809E+04, 1.2819E+04, 1.2829E+04, 1.2839E+04, 1.2849E+04,
     + 1.2860E+04, 1.2870E+04, 1.2880E+04, 1.2890E+04, 1.2900E+04,
     +  1.2910E+04, 1.2921E+04, 1.2931E+04, 1.2941E+04, 1.2951E+04,
     + 1.2961E+04, 1.2972E+04, 1.2982E+04, 1.2992E+04, 1.3002E+04,
     +  1.3012E+04, 1.3023E+04, 1.3033E+04, 1.3043E+04, 1.3053E+04,
     + 1.3064E+04, 1.3074E+04, 1.3084E+04, 1.3094E+04, 1.3105E+04,
     +  1.3115E+04, 1.3125E+04, 1.3135E+04, 1.3146E+04, 1.3156E+04,
     + 1.3166E+04, 1.3177E+04, 1.3187E+04, 1.3197E+04, 1.3208E+04,
     +  1.3218E+04, 1.3228E+04, 1.3238E+04, 1.3249E+04, 1.3259E+04,
     + 1.3269E+04, 1.3280E+04, 1.3290E+04, 1.3300E+04, 1.3311E+04,
     +  1.3321E+04, 1.3332E+04, 1.3342E+04, 1.3352E+04, 1.3363E+04,
     + 1.3373E+04, 1.3383E+04, 1.3394E+04, 1.3404E+04, 1.3415E+04,
     +  1.3425E+04, 1.3435E+04, 1.3446E+04, 1.3456E+04, 1.3467E+04,
     + 1.3477E+04, 1.3487E+04, 1.3498E+04, 1.3508E+04, 1.3519E+04,
     +  1.3529E+04, 1.3540E+04, 1.3550E+04, 1.3561E+04, 1.3571E+04,
     + 1.3581E+04, 1.3592E+04, 1.3602E+04, 1.3613E+04, 1.3623E+04,
     +  1.3634E+04, 1.3644E+04, 1.3655E+04, 1.3665E+04, 1.3676E+04,
     + 1.3686E+04, 1.3697E+04, 1.3707E+04, 1.3718E+04, 1.3728E+04,
     +  1.3739E+04, 1.3750E+04, 1.3760E+04, 1.3771E+04, 1.3781E+04,
     + 1.3792E+04, 1.3802E+04, 1.3813E+04, 1.3823E+04, 1.3834E+04,
     +  1.3844E+04, 1.3855E+04, 1.3866E+04, 1.3876E+04, 1.3887E+04,
     + 1.3897E+04, 1.3908E+04, 1.3919E+04, 1.3929E+04, 1.3940E+04,
     +  1.3950E+04, 1.3961E+04, 1.3972E+04, 1.3982E+04, 1.3993E+04,
     + 1.4004E+04, 1.4014E+04, 1.4025E+04, 1.4036E+04, 1.4046E+04,
     +  1.4057E+04, 1.4068E+04, 1.4078E+04, 1.4089E+04, 1.4100E+04,
     + 1.4110E+04, 1.4121E+04, 1.4132E+04, 1.4142E+04, 1.4153E+04,
     +  1.4164E+04, 1.4174E+04, 1.4185E+04, 1.4196E+04, 1.4207E+04,
     + 1.4217E+04, 1.4228E+04, 1.4239E+04, 1.4249E+04, 1.4260E+04,
     +  1.4271E+04, 1.4282E+04, 1.4292E+04, 1.4303E+04, 1.4314E+04,
     + 1.4325E+04, 1.4336E+04, 1.4346E+04, 1.4357E+04, 1.4368E+04,
     +  1.4379E+04, 1.4389E+04, 1.4400E+04, 1.4411E+04, 1.4422E+04,
     + 1.4433E+04, 1.4443E+04, 1.4454E+04, 1.4465E+04, 1.4476E+04,
     +  1.4487E+04, 1.4498E+04, 1.4508E+04, 1.4519E+04, 1.4530E+04,
     + 1.4541E+04, 1.4552E+04, 1.4563E+04, 1.4574E+04, 1.4584E+04,
     +  1.4595E+04, 1.4606E+04, 1.4617E+04, 1.4628E+04, 1.4639E+04,
     + 1.4650E+04, 1.4661E+04, 1.4672E+04, 1.4683E+04, 1.4693E+04,
     +  1.4704E+04, 1.4715E+04, 1.4726E+04, 1.4737E+04, 1.4748E+04,
     + 1.4759E+04, 1.4770E+04, 1.4781E+04, 1.4792E+04, 1.4803E+04,
     +  1.4814E+04, 1.4825E+04, 1.4836E+04, 1.4847E+04, 1.4858E+04,
     + 1.4869E+04, 1.4880E+04, 1.4891E+04, 1.4902E+04, 1.4913E+04,
     +  1.4924E+04, 1.4935E+04, 1.4946E+04, 1.4957E+04, 1.4968E+04,
     + 1.4979E+04, 1.4990E+04, 1.5001E+04, 1.5012E+04, 1.5023E+04,
     +  1.5034E+04, 1.5045E+04, 1.5056E+04, 1.5067E+04, 1.5078E+04,
     + 1.5089E+04, 1.5100E+04, 1.5111E+04, 1.5122E+04, 1.5134E+04,
     +  1.5145E+04, 1.5156E+04, 1.5167E+04, 1.5178E+04, 1.5189E+04,
     + 1.5200E+04, 1.5211E+04, 1.5222E+04, 1.5234E+04, 1.5245E+04,
     +  1.5256E+04, 1.5267E+04, 1.5278E+04, 1.5289E+04, 1.5300E+04,
     + 1.5312E+04, 1.5323E+04, 1.5334E+04, 1.5345E+04, 1.5356E+04,
     +  1.5367E+04, 1.5379E+04, 1.5390E+04, 1.5401E+04, 1.5412E+04,
     + 1.5423E+04, 1.5435E+04, 1.5446E+04, 1.5457E+04, 1.5468E+04,
     +  1.5479E+04, 1.5491E+04, 1.5502E+04, 1.5513E+04, 1.5524E+04,
     + 1.5536E+04, 1.5547E+04, 1.5558E+04, 1.5569E+04, 1.5581E+04,
     +  1.5592E+04, 1.5603E+04, 1.5614E+04, 1.5626E+04, 1.5637E+04,
     + 1.5648E+04, 1.5660E+04, 1.5671E+04, 1.5682E+04, 1.5693E+04,
     +  1.5705E+04, 1.5716E+04, 1.5727E+04, 1.5739E+04, 1.5750E+04,
     + 1.5761E+04, 1.5773E+04, 1.5784E+04, 1.5795E+04, 1.5807E+04,
     +  1.5818E+04, 1.5829E+04, 1.5841E+04, 1.5852E+04, 1.5863E+04,
     + 1.5875E+04, 1.5886E+04, 1.5898E+04, 1.5909E+04, 1.5920E+04,
     +  1.5932E+04, 1.5943E+04, 1.5955E+04, 1.5966E+04, 1.5977E+04,
     + 1.5989E+04, 1.6000E+04, 1.6012E+04, 1.6023E+04, 1.6035E+04,
     +  1.6046E+04, 1.6057E+04, 1.6069E+04, 1.6080E+04, 1.6092E+04,
     + 1.6103E+04, 1.6115E+04, 1.6126E+04, 1.6138E+04, 1.6149E+04,
     +  1.6161E+04, 1.6172E+04, 1.6184E+04, 1.6195E+04, 1.6207E+04,
     + 1.6218E+04, 1.6230E+04, 1.6241E+04, 1.6253E+04, 1.6264E+04,
     +  1.6276E+04, 1.6287E+04, 1.6299E+04, 1.6310E+04, 1.6322E+04,
     + 1.6333E+04, 1.6345E+04, 1.6356E+04, 1.6368E+04, 1.6379E+04,
     +  1.6391E+04, 1.6403E+04, 1.6414E+04, 1.6426E+04, 1.6437E+04,
     + 1.6449E+04, 1.6460E+04, 1.6472E+04, 1.6484E+04, 1.6495E+04,
     +  1.6507E+04, 1.6518E+04, 1.6530E+04, 1.6542E+04, 1.6553E+04,
     + 1.6565E+04, 1.6577E+04, 1.6588E+04, 1.6600E+04, 1.6612E+04,
     +  1.6623E+04, 1.6635E+04, 1.6646E+04, 1.6658E+04, 1.6670E+04,
     + 1.6681E+04, 1.6693E+04, 1.6705E+04, 1.6717E+04, 1.6728E+04,
     +  1.6740E+04, 1.6752E+04, 1.6763E+04, 1.6775E+04, 1.6787E+04,
     + 1.6798E+04, 1.6810E+04, 1.6822E+04, 1.6834E+04, 1.6845E+04,
     +  1.6857E+04, 1.6869E+04, 1.6881E+04, 1.6892E+04, 1.6904E+04,
     + 1.6916E+04, 1.6928E+04, 1.6939E+04, 1.6951E+04, 1.6963E+04,
     +  1.6975E+04, 1.6986E+04, 1.6998E+04, 1.7010E+04, 1.7022E+04,
     + 1.7034E+04, 1.7045E+04, 1.7057E+04, 1.7069E+04, 1.7081E+04,
     +  1.7093E+04, 1.7104E+04, 1.7116E+04, 1.7128E+04, 1.7140E+04,
     + 1.7152E+04, 1.7164E+04, 1.7175E+04, 1.7187E+04, 1.7199E+04,
     +  1.7211E+04, 1.7223E+04, 1.7235E+04, 1.7247E+04, 1.7259E+04,
     + 1.7270E+04, 1.7282E+04, 1.7294E+04, 1.7306E+04, 1.7318E+04,
     +  1.7330E+04, 1.7342E+04, 1.7354E+04, 1.7366E+04, 1.7378E+04,
     + 1.7390E+04, 1.7401E+04, 1.7413E+04, 1.7425E+04, 1.7437E+04,
     +  1.7449E+04, 1.7461E+04, 1.7473E+04, 1.7485E+04, 1.7497E+04,
     + 1.7509E+04, 1.7521E+04, 1.7533E+04, 1.7545E+04, 1.7557E+04,
     +  1.7569E+04, 1.7581E+04, 1.7593E+04, 1.7605E+04, 1.7617E+04,
     + 1.7629E+04, 1.7641E+04, 1.7653E+04, 1.7665E+04, 1.7677E+04,
     +  1.7689E+04, 1.7701E+04, 1.7713E+04, 1.7725E+04, 1.7737E+04,
     + 1.7749E+04, 1.7761E+04, 1.7773E+04, 1.7785E+04, 1.7797E+04,
     +  1.7810E+04, 1.7822E+04, 1.7834E+04, 1.7846E+04, 1.7858E+04,
     + 1.7870E+04, 1.7882E+04, 1.7894E+04, 1.7906E+04, 1.7918E+04,
     +  1.7931E+04, 1.7943E+04, 1.7955E+04, 1.7967E+04, 1.7979E+04,
     + 1.7991E+04, 1.8003E+04, 1.8015E+04, 1.8028E+04, 1.8040E+04,
     +  1.8052E+04, 1.8064E+04, 1.8076E+04, 1.8088E+04, 1.8101E+04,
     + 1.8113E+04, 1.8125E+04, 1.8137E+04, 1.8149E+04, 1.8162E+04,
     +  1.8174E+04, 1.8186E+04, 1.8198E+04, 1.8210E+04, 1.8223E+04,
     + 1.8235E+04, 1.8247E+04, 1.8259E+04, 1.8271E+04, 1.8284E+04,
     +  1.8296E+04, 1.8308E+04, 1.8320E+04, 1.8333E+04, 1.8345E+04,
     + 1.8357E+04, 1.8369E+04, 1.8382E+04, 1.8394E+04, 1.8406E+04,
     +  1.8419E+04, 1.8431E+04, 1.8443E+04, 1.8455E+04, 1.8468E+04,
     + 1.8480E+04, 1.8492E+04, 1.8505E+04, 1.8517E+04, 1.8529E+04,
     +  1.8542E+04, 1.8554E+04, 1.8566E+04, 1.8579E+04, 1.8591E+04,
     + 1.8603E+04, 1.8616E+04, 1.8628E+04, 1.8640E+04, 1.8653E+04,
     +  1.8665E+04, 1.8678E+04, 1.8690E+04, 1.8702E+04, 1.8715E+04,
     + 1.8727E+04, 1.8739E+04, 1.8752E+04, 1.8764E+04, 1.8777E+04,
     +  1.8789E+04, 1.8801E+04, 1.8814E+04, 1.8826E+04, 1.8839E+04,
     + 1.8851E+04, 1.8864E+04, 1.8876E+04, 1.8888E+04, 1.8901E+04,
     +  1.8913E+04, 1.8926E+04, 1.8938E+04, 1.8951E+04, 1.8963E+04,
     + 1.8976E+04, 1.8988E+04, 1.9001E+04, 1.9013E+04, 1.9026E+04,
     +  1.9038E+04, 1.9051E+04, 1.9063E+04, 1.9076E+04, 1.9088E+04,
     + 1.9101E+04, 1.9113E+04, 1.9126E+04, 1.9138E+04, 1.9151E+04,
     +  1.9163E+04, 1.9176E+04, 1.9188E+04, 1.9201E+04, 1.9213E+04,
     + 1.9226E+04, 1.9239E+04, 1.9251E+04, 1.9264E+04, 1.9276E+04,
     +  1.9289E+04, 1.9301E+04, 1.9314E+04, 1.9327E+04, 1.9339E+04,
     + 1.9352E+04, 1.9364E+04, 1.9377E+04, 1.9390E+04, 1.9402E+04,
     +  1.9415E+04, 1.9427E+04, 1.9440E+04, 1.9453E+04, 1.9465E+04,
     + 1.9478E+04, 1.9491E+04, 1.9503E+04, 1.9516E+04, 1.9529E+04,
     +  1.9541E+04, 1.9554E+04, 1.9567E+04, 1.9579E+04, 1.9592E+04,
     + 1.9605E+04, 1.9617E+04, 1.9630E+04, 1.9643E+04, 1.9655E+04,
     +  1.9668E+04, 1.9681E+04, 1.9693E+04, 1.9706E+04, 1.9719E+04,
     + 1.9732E+04, 1.9744E+04, 1.9757E+04, 1.9770E+04, 1.9783E+04,
     +  1.9795E+04, 1.9808E+04, 1.9821E+04, 1.9834E+04, 1.9846E+04,
     + 1.9859E+04, 1.9872E+04, 1.9885E+04, 1.9897E+04, 1.9910E+04,
     +  1.9923E+04, 1.9936E+04, 1.9949E+04, 1.9961E+04, 1.9974E+04,
     + 1.9987E+04, 2.0000E+04, 2.0013E+04, 2.0025E+04, 2.0038E+04,
     +  2.0051E+04, 2.0064E+04, 2.0077E+04, 2.0089E+04, 2.0102E+04,
     + 2.0115E+04, 2.0128E+04, 2.0141E+04, 2.0154E+04, 2.0167E+04,
     +  2.0179E+04, 2.0192E+04, 2.0205E+04, 2.0218E+04, 2.0231E+04,
     + 2.0244E+04, 2.0257E+04, 2.0270E+04, 2.0283E+04, 2.0295E+04,
     +  2.0308E+04, 2.0321E+04, 2.0334E+04, 2.0347E+04, 2.0360E+04,
     + 2.0373E+04, 2.0386E+04, 2.0399E+04, 2.0412E+04, 2.0425E+04,
     +  2.0438E+04, 2.0451E+04, 2.0464E+04, 2.0477E+04, 2.0489E+04,
     + 2.0502E+04, 2.0515E+04, 2.0528E+04, 2.0541E+04, 2.0554E+04,
     +  2.0567E+04, 2.0580E+04, 2.0593E+04, 2.0606E+04, 2.0619E+04,
     + 2.0632E+04, 2.0645E+04, 2.0658E+04, 2.0671E+04, 2.0684E+04,
     +  2.0697E+04, 2.0710E+04, 2.0724E+04, 2.0737E+04, 2.0750E+04,
     + 2.0763E+04, 2.0776E+04, 2.0789E+04, 2.0802E+04, 2.0815E+04,
     +  2.0828E+04, 2.0841E+04, 2.0854E+04, 2.0867E+04, 2.0880E+04,
     + 2.0893E+04, 2.0906E+04, 2.0920E+04, 2.0933E+04, 2.0946E+04,
     +  2.0959E+04, 2.0972E+04, 2.0985E+04, 2.0998E+04, 2.1011E+04,
     + 2.1025E+04, 2.1038E+04, 2.1051E+04, 2.1064E+04, 2.1077E+04,
     +  2.1090E+04, 2.1103E+04, 2.1117E+04, 2.1130E+04, 2.1143E+04,
     + 2.1156E+04, 2.1169E+04, 2.1182E+04, 2.1196E+04, 2.1209E+04,
     +  2.1222E+04, 2.1235E+04, 2.1248E+04, 2.1262E+04, 2.1275E+04,
     + 2.1288E+04, 2.1301E+04, 2.1314E+04, 2.1328E+04, 2.1341E+04,
     +  2.1354E+04, 2.1367E+04, 2.1381E+04, 2.1394E+04, 2.1407E+04,
     + 2.1420E+04, 2.1434E+04, 2.1447E+04, 2.1460E+04, 2.1473E+04,
     +  2.1487E+04, 2.1500E+04, 2.1513E+04, 2.1527E+04, 2.1540E+04,
     + 2.1553E+04, 2.1566E+04, 2.1580E+04, 2.1593E+04, 2.1606E+04,
     +  2.1620E+04, 2.1633E+04, 2.1646E+04, 2.1660E+04, 2.1673E+04,
     + 2.1686E+04, 2.1700E+04, 2.1713E+04, 2.1726E+04, 2.1740E+04,
     +  2.1753E+04, 2.1766E+04, 2.1780E+04, 2.1793E+04, 2.1807E+04,
     + 2.1820E+04, 2.1833E+04, 2.1847E+04, 2.1860E+04, 2.1873E+04,
     +  2.1887E+04, 2.1900E+04, 2.1914E+04, 2.1927E+04, 2.1940E+04,
     + 2.1954E+04, 2.1967E+04, 2.1981E+04, 2.1994E+04, 2.2008E+04,
     +  2.2021E+04, 2.2034E+04, 2.2048E+04, 2.2061E+04, 2.2075E+04,
     + 2.2088E+04, 2.2102E+04, 2.2115E+04, 2.2129E+04, 2.2142E+04,
     +  2.2156E+04, 2.2169E+04, 2.2183E+04, 2.2196E+04, 2.2210E+04,
     + 2.2223E+04, 2.2237E+04, 2.2250E+04, 2.2264E+04, 2.2277E+04,
     +  2.2291E+04, 2.2304E+04, 2.2318E+04, 2.2331E+04, 2.2345E+04,
     + 2.2358E+04, 2.2372E+04, 2.2385E+04, 2.2399E+04, 2.2412E+04,
     +  2.2426E+04, 2.2440E+04, 2.2453E+04, 2.2467E+04, 2.2480E+04,
     + 2.2494E+04, 2.2507E+04, 2.2521E+04, 2.2535E+04, 2.2548E+04,
     +  2.2562E+04, 2.2575E+04, 2.2589E+04, 2.2603E+04, 2.2616E+04,
     + 2.2630E+04, 2.2643E+04, 2.2657E+04, 2.2671E+04, 2.2684E+04,
     +  2.2698E+04, 2.2712E+04, 2.2725E+04, 2.2739E+04, 2.2753E+04,
     + 2.2766E+04, 2.2780E+04, 2.2794E+04, 2.2807E+04, 2.2821E+04,
     +  2.2835E+04, 2.2848E+04, 2.2862E+04, 2.2876E+04, 2.2889E+04,
     + 2.2903E+04, 2.2917E+04, 2.2930E+04, 2.2944E+04, 2.2958E+04,
     +  2.2972E+04, 2.2985E+04, 2.2999E+04, 2.3013E+04, 2.3026E+04,
     + 2.3040E+04, 2.3054E+04, 2.3068E+04, 2.3081E+04, 2.3095E+04,
     +  2.3109E+04, 2.3123E+04, 2.3136E+04, 2.3150E+04, 2.3164E+04,
     + 2.3178E+04, 2.3192E+04, 2.3205E+04, 2.3219E+04, 2.3233E+04,
     +  2.3247E+04, 2.3261E+04, 2.3274E+04, 2.3288E+04, 2.3302E+04,
     + 2.3316E+04, 2.3330E+04, 2.3343E+04, 2.3357E+04, 2.3371E+04,
     +  2.3385E+04, 2.3399E+04, 2.3413E+04, 2.3426E+04, 2.3440E+04,
     + 2.3454E+04, 2.3468E+04, 2.3482E+04, 2.3496E+04, 2.3510E+04,
     +  2.3524E+04, 2.3537E+04, 2.3551E+04, 2.3565E+04, 2.3579E+04,
     + 2.3593E+04, 2.3607E+04, 2.3621E+04, 2.3635E+04, 2.3649E+04,
     +  2.3663E+04, 2.3676E+04, 2.3690E+04, 2.3704E+04, 2.3718E+04,
     + 2.3732E+04, 2.3746E+04, 2.3760E+04, 2.3774E+04, 2.3788E+04,
     +  2.3802E+04, 2.3816E+04, 2.3830E+04, 2.3844E+04, 2.3858E+04,
     + 2.3872E+04, 2.3886E+04, 2.3900E+04, 2.3914E+04, 2.3928E+04,
     +  2.3942E+04, 2.3956E+04, 2.3970E+04, 2.3984E+04, 2.3998E+04,
     + 2.4012E+04, 2.4026E+04, 2.4040E+04, 2.4054E+04, 2.4068E+04,
     +  2.4082E+04, 2.4096E+04, 2.4110E+04, 2.4124E+04, 2.4138E+04,
     + 2.4152E+04, 2.4166E+04, 2.4180E+04, 2.4194E+04, 2.4208E+04,
     +  2.4222E+04, 2.4236E+04, 2.4251E+04, 2.4265E+04, 2.4279E+04,
     + 2.4293E+04, 2.4307E+04, 2.4321E+04, 2.4335E+04, 2.4349E+04,
     +  2.4363E+04, 2.4377E+04, 2.4392E+04, 2.4406E+04, 2.4420E+04,
     + 2.4434E+04, 2.4448E+04, 2.4462E+04, 2.4476E+04, 2.4490E+04,
     +  2.4505E+04, 2.4519E+04, 2.4533E+04, 2.4547E+04, 2.4561E+04,
     + 2.4575E+04, 2.4590E+04, 2.4604E+04, 2.4618E+04, 2.4632E+04,
     +  2.4646E+04, 2.4661E+04, 2.4675E+04, 2.4689E+04, 2.4703E+04,
     + 2.4717E+04, 2.4732E+04, 2.4746E+04, 2.4760E+04, 2.4774E+04,
     +  2.4788E+04, 2.4803E+04, 2.4817E+04, 2.4831E+04, 2.4845E+04,
     + 2.4860E+04, 2.4874E+04, 2.4888E+04, 2.4902E+04, 2.4917E+04,
     +  2.4931E+04, 2.4945E+04, 2.4960E+04, 2.4974E+04, 2.4988E+04,
     + 2.5002E+04, 2.5017E+04, 2.5031E+04, 2.5045E+04, 2.5060E+04,
     +  2.5074E+04, 2.5088E+04, 2.5103E+04, 2.5117E+04, 2.5131E+04,
     + 2.5145E+04, 2.5160E+04, 2.5174E+04, 2.5188E+04, 2.5203E+04,
     +  2.5217E+04, 2.5232E+04, 2.5246E+04, 2.5260E+04, 2.5275E+04,
     + 2.5289E+04, 2.5303E+04, 2.5318E+04, 2.5332E+04, 2.5346E+04,
     +  2.5361E+04, 2.5375E+04, 2.5390E+04, 2.5404E+04, 2.5418E+04,
     + 2.5433E+04, 2.5447E+04, 2.5462E+04, 2.5476E+04, 2.5490E+04,
     +  2.5505E+04, 2.5519E+04, 2.5534E+04, 2.5548E+04, 2.5563E+04,
     + 2.5577E+04, 2.5592E+04, 2.5606E+04, 2.5620E+04, 2.5635E+04,
     +  2.5649E+04, 2.5664E+04, 2.5678E+04, 2.5693E+04, 2.5707E+04,
     + 2.5722E+04, 2.5736E+04, 2.5751E+04, 2.5765E+04, 2.5780E+04,
     +  2.5794E+04, 2.5809E+04, 2.5823E+04, 2.5838E+04, 2.5852E+04,
     + 2.5867E+04, 2.5881E+04, 2.5896E+04, 2.5910E+04, 2.5925E+04,
     +  2.5939E+04, 2.5954E+04, 2.5968E+04, 2.5983E+04, 2.5998E+04,
     + 2.6012E+04, 2.6027E+04, 2.6041E+04, 2.6056E+04, 2.6070E+04,
     +  2.6085E+04, 2.6100E+04, 2.6114E+04, 2.6129E+04, 2.6143E+04,
     + 2.6158E+04, 2.6172E+04, 2.6187E+04, 2.6202E+04, 2.6216E+04,
     +  2.6231E+04, 2.6246E+04, 2.6260E+04, 2.6275E+04, 2.6289E+04,
     + 2.6304E+04, 2.6319E+04, 2.6333E+04, 2.6348E+04, 2.6363E+04,
     +  2.6377E+04, 2.6392E+04, 2.6407E+04, 2.6421E+04, 2.6436E+04,
     + 2.6451E+04, 2.6465E+04, 2.6480E+04, 2.6495E+04, 2.6509E+04,
     +  2.6524E+04, 2.6539E+04, 2.6553E+04, 2.6568E+04, 2.6583E+04,
     + 2.6598E+04, 2.6612E+04, 2.6627E+04, 2.6642E+04, 2.6656E+04,
     +  2.6671E+04, 2.6686E+04, 2.6701E+04, 2.6715E+04, 2.6730E+04,
     + 2.6745E+04, 2.6760E+04, 2.6774E+04, 2.6789E+04, 2.6804E+04,
     +  2.6819E+04, 2.6833E+04, 2.6848E+04, 2.6863E+04, 2.6878E+04,
     + 2.6893E+04, 2.6907E+04, 2.6922E+04, 2.6937E+04, 2.6952E+04,
     +  2.6966E+04, 2.6981E+04, 2.6996E+04, 2.7011E+04, 2.7026E+04,
     + 2.7041E+04, 2.7055E+04, 2.7070E+04, 2.7085E+04, 2.7100E+04,
     +  2.7115E+04, 2.7130E+04, 2.7144E+04, 2.7159E+04, 2.7174E+04,
     + 2.7189E+04, 2.7204E+04, 2.7219E+04, 2.7234E+04, 2.7248E+04,
     +  2.7263E+04, 2.7278E+04, 2.7293E+04, 2.7308E+04, 2.7323E+04,
     + 2.7338E+04, 2.7353E+04, 2.7368E+04, 2.7383E+04, 2.7397E+04,
     +  2.7412E+04, 2.7427E+04, 2.7442E+04, 2.7457E+04, 2.7472E+04,
     + 2.7487E+04, 2.7502E+04, 2.7517E+04, 2.7532E+04, 2.7547E+04,
     +  2.7562E+04, 2.7577E+04, 2.7592E+04, 2.7607E+04, 2.7622E+04,
     + 2.7637E+04, 2.7652E+04, 2.7667E+04, 2.7682E+04, 2.7696E+04,
     +  2.7711E+04, 2.7726E+04, 2.7741E+04, 2.7756E+04, 2.7771E+04,
     + 2.7786E+04, 2.7802E+04, 2.7817E+04, 2.7832E+04, 2.7847E+04,
     +  2.7862E+04, 2.7877E+04, 2.7892E+04, 2.7907E+04, 2.7922E+04,
     + 2.7937E+04, 2.7952E+04, 2.7967E+04, 2.7982E+04, 2.7997E+04,
     +  2.8012E+04, 2.8027E+04, 2.8042E+04, 2.8057E+04, 2.8072E+04,
     + 2.8087E+04, 2.8103E+04, 2.8118E+04, 2.8133E+04, 2.8148E+04,
     +  2.8163E+04, 2.8178E+04, 2.8193E+04, 2.8208E+04, 2.8223E+04,
     + 2.8239E+04, 2.8254E+04, 2.8269E+04, 2.8284E+04, 2.8299E+04,
     +  2.8314E+04, 2.8329E+04, 2.8344E+04, 2.8360E+04, 2.8375E+04,
     + 2.8390E+04, 2.8405E+04, 2.8420E+04, 2.8435E+04, 2.8451E+04,
     +  2.8466E+04, 2.8481E+04, 2.8496E+04, 2.8511E+04, 2.8527E+04,
     + 2.8542E+04, 2.8557E+04, 2.8572E+04, 2.8587E+04, 2.8603E+04,
     +  2.8618E+04, 2.8633E+04, 2.8648E+04, 2.8663E+04, 2.8679E+04,
     + 2.8694E+04, 2.8709E+04, 2.8724E+04, 2.8740E+04, 2.8755E+04,
     +  2.8770E+04, 2.8785E+04, 2.8801E+04, 2.8816E+04, 2.8831E+04,
     + 2.8846E+04, 2.8862E+04, 2.8877E+04, 2.8892E+04, 2.8908E+04,
     +  2.8923E+04, 2.8938E+04, 2.8953E+04, 2.8969E+04, 2.8984E+04,
     + 2.8999E+04, 2.9015E+04, 2.9030E+04, 2.9045E+04, 2.9061E+04,
     +  2.9076E+04, 2.9091E+04, 2.9107E+04, 2.9122E+04, 2.9137E+04,
     + 2.9153E+04, 2.9168E+04, 2.9183E+04, 2.9199E+04, 2.9214E+04,
     +  2.9229E+04, 2.9245E+04, 2.9260E+04, 2.9275E+04, 2.9291E+04,
     + 2.9306E+04, 2.9322E+04, 2.9337E+04, 2.9352E+04, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        87
      data (QofT( 5,J),J=1,2408)/ 2.1575E+01, 2.7866E+01, 3.5401E+01,
     +  4.3570E+01, 5.2123E+01, 6.0921E+01, 6.9883E+01, 7.8958E+01,
     + 8.8116E+01, 9.7334E+01, 1.0660E+02, 1.1590E+02, 1.2523E+02,
     +  1.3459E+02, 1.4396E+02, 1.5335E+02, 1.6275E+02, 1.7216E+02,
     + 1.8159E+02, 1.9102E+02, 2.0990E+02, 2.2881E+02, 2.4773E+02,
     +  2.6667E+02, 2.8561E+02, 3.0457E+02, 3.2354E+02, 3.4251E+02,
     + 3.6149E+02, 3.8047E+02, 3.9946E+02, 4.1845E+02, 4.3744E+02,
     +  4.5644E+02, 4.7543E+02, 4.9444E+02, 5.1344E+02, 5.3244E+02,
     + 5.5145E+02, 5.7046E+02, 5.8947E+02, 6.0848E+02, 6.2749E+02,
     +  6.4651E+02, 6.6552E+02, 6.8454E+02, 7.0355E+02, 7.2257E+02,
     + 7.4159E+02, 7.6061E+02, 7.7963E+02, 7.9865E+02, 8.1767E+02,
     +  8.3669E+02, 8.5572E+02, 8.7474E+02, 8.9376E+02, 9.1279E+02,
     + 9.3181E+02, 9.5084E+02, 9.6987E+02, 9.8889E+02, 1.0079E+03,
     +  1.0269E+03, 1.0460E+03, 1.0650E+03, 1.0840E+03, 1.1031E+03,
     + 1.1221E+03, 1.1411E+03, 1.1602E+03, 1.1792E+03, 1.1982E+03,
     +  1.2172E+03, 1.2363E+03, 1.2553E+03, 1.2743E+03, 1.2934E+03,
     + 1.3124E+03, 1.3314E+03, 1.3505E+03, 1.3695E+03, 1.3886E+03,
     +  1.4076E+03, 1.4266E+03, 1.4457E+03, 1.4647E+03, 1.4837E+03,
     + 1.5028E+03, 1.5218E+03, 1.5408E+03, 1.5599E+03, 1.5789E+03,
     +  1.5980E+03, 1.6170E+03, 1.6360E+03, 1.6551E+03, 1.6741E+03,
     + 1.6932E+03, 1.7122E+03, 1.7313E+03, 1.7503E+03, 1.7693E+03,
     +  1.7884E+03, 1.8074E+03, 1.8265E+03, 1.8455E+03, 1.8646E+03,
     + 1.8836E+03, 1.9027E+03, 1.9217E+03, 1.9408E+03, 1.9598E+03,
     +  1.9789E+03, 1.9979E+03, 2.0170E+03, 2.0360E+03, 2.0551E+03,
     + 2.0741E+03, 2.0932E+03, 2.1122E+03, 2.1313E+03, 2.1504E+03,
     +  2.1694E+03, 2.1885E+03, 2.2076E+03, 2.2266E+03, 2.2457E+03,
     + 2.2648E+03, 2.2838E+03, 2.3029E+03, 2.3220E+03, 2.3411E+03,
     +  2.3602E+03, 2.3792E+03, 2.3983E+03, 2.4174E+03, 2.4365E+03,
     + 2.4556E+03, 2.4747E+03, 2.4938E+03, 2.5129E+03, 2.5320E+03,
     +  2.5511E+03, 2.5702E+03, 2.5893E+03, 2.6085E+03, 2.6276E+03,
     + 2.6467E+03, 2.6659E+03, 2.6850E+03, 2.7041E+03, 2.7233E+03,
     +  2.7424E+03, 2.7616E+03, 2.7807E+03, 2.7999E+03, 2.8191E+03,
     + 2.8382E+03, 2.8574E+03, 2.8766E+03, 2.8958E+03, 2.9150E+03,
     +  2.9342E+03, 2.9534E+03, 2.9726E+03, 2.9918E+03, 3.0110E+03,
     + 3.0303E+03, 3.0495E+03, 3.0688E+03, 3.0880E+03, 3.1073E+03,
     +  3.1265E+03, 3.1458E+03, 3.1651E+03, 3.1844E+03, 3.2037E+03,
     + 3.2230E+03, 3.2423E+03, 3.2616E+03, 3.2810E+03, 3.3003E+03,
     +  3.3197E+03, 3.3390E+03, 3.3584E+03, 3.3778E+03, 3.3972E+03,
     + 3.4166E+03, 3.4360E+03, 3.4554E+03, 3.4748E+03, 3.4943E+03,
     +  3.5137E+03, 3.5332E+03, 3.5526E+03, 3.5721E+03, 3.5916E+03,
     + 3.6111E+03, 3.6307E+03, 3.6502E+03, 3.6697E+03, 3.6893E+03,
     +  3.7088E+03, 3.7284E+03, 3.7480E+03, 3.7676E+03, 3.7872E+03,
     + 3.8069E+03, 3.8265E+03, 3.8462E+03, 3.8659E+03, 3.8855E+03,
     +  3.9052E+03, 3.9250E+03, 3.9447E+03, 3.9644E+03, 3.9842E+03,
     + 4.0040E+03, 4.0237E+03, 4.0435E+03, 4.0634E+03, 4.0832E+03,
     +  4.1030E+03, 4.1229E+03, 4.1428E+03, 4.1627E+03, 4.1826E+03,
     + 4.2025E+03, 4.2225E+03, 4.2424E+03, 4.2624E+03, 4.2824E+03,
     +  4.3024E+03, 4.3225E+03, 4.3425E+03, 4.3626E+03, 4.3827E+03,
     + 4.4028E+03, 4.4229E+03, 4.4430E+03, 4.4632E+03, 4.4834E+03,
     +  4.5035E+03, 4.5238E+03, 4.5440E+03, 4.5642E+03, 4.5845E+03,
     + 4.6048E+03, 4.6251E+03, 4.6454E+03, 4.6658E+03, 4.6862E+03,
     +  4.7065E+03, 4.7269E+03, 4.7474E+03, 4.7678E+03, 4.7883E+03,
     + 4.8088E+03, 4.8293E+03, 4.8498E+03, 4.8704E+03, 4.8910E+03,
     +  4.9115E+03, 4.9322E+03, 4.9528E+03, 4.9735E+03, 4.9941E+03,
     + 5.0148E+03, 5.0356E+03, 5.0563E+03, 5.0771E+03, 5.0979E+03,
     +  5.1187E+03, 5.1395E+03, 5.1604E+03, 5.1813E+03, 5.2022E+03,
     + 5.2231E+03, 5.2441E+03, 5.2650E+03, 5.2860E+03, 5.3071E+03,
     +  5.3281E+03, 5.3492E+03, 5.3703E+03, 5.3914E+03, 5.4125E+03,
     + 5.4337E+03, 5.4549E+03, 5.4761E+03, 5.4973E+03, 5.5186E+03,
     +  5.5399E+03, 5.5612E+03, 5.5825E+03, 5.6039E+03, 5.6253E+03,
     + 5.6467E+03, 5.6681E+03, 5.6896E+03, 5.7111E+03, 5.7326E+03,
     +  5.7541E+03, 5.7757E+03, 5.7973E+03, 5.8189E+03, 5.8405E+03,
     + 5.8622E+03, 5.8839E+03, 5.9056E+03, 5.9274E+03, 5.9491E+03,
     +  5.9709E+03, 5.9928E+03, 6.0146E+03, 6.0365E+03, 6.0584E+03,
     + 6.0803E+03, 6.1023E+03, 6.1243E+03, 6.1463E+03, 6.1683E+03,
     +  6.1904E+03, 6.2125E+03, 6.2346E+03, 6.2568E+03, 6.2790E+03,
     + 6.3012E+03, 6.3234E+03, 6.3457E+03, 6.3680E+03, 6.3903E+03,
     +  6.4126E+03, 6.4350E+03, 6.4574E+03, 6.4798E+03, 6.5023E+03,
     + 6.5248E+03, 6.5473E+03, 6.5698E+03, 6.5924E+03, 6.6150E+03,
     +  6.6376E+03, 6.6603E+03, 6.6830E+03, 6.7057E+03, 6.7285E+03,
     + 6.7512E+03, 6.7740E+03, 6.7969E+03, 6.8197E+03, 6.8426E+03,
     +  6.8655E+03, 6.8885E+03, 6.9115E+03, 6.9345E+03, 6.9575E+03,
     + 6.9806E+03, 7.0037E+03, 7.0268E+03, 7.0500E+03, 7.0732E+03,
     +  7.0964E+03, 7.1196E+03, 7.1429E+03, 7.1662E+03, 7.1896E+03,
     + 7.2129E+03, 7.2363E+03, 7.2598E+03, 7.2832E+03, 7.3067E+03,
     +  7.3302E+03, 7.3538E+03, 7.3774E+03, 7.4010E+03, 7.4246E+03,
     + 7.4483E+03, 7.4720E+03, 7.4957E+03, 7.5195E+03, 7.5433E+03,
     +  7.5671E+03, 7.5910E+03, 7.6149E+03, 7.6388E+03, 7.6627E+03,
     + 7.6867E+03, 7.7107E+03, 7.7348E+03, 7.7589E+03, 7.7830E+03,
     +  7.8071E+03, 7.8313E+03, 7.8555E+03, 7.8797E+03, 7.9040E+03,
     + 7.9283E+03, 7.9526E+03, 7.9770E+03, 8.0014E+03, 8.0258E+03,
     +  8.0502E+03, 8.0747E+03, 8.0993E+03, 8.1238E+03, 8.1484E+03,
     + 8.1730E+03, 8.1977E+03, 8.2223E+03, 8.2470E+03, 8.2718E+03,
     +  8.2966E+03, 8.3214E+03, 8.3462E+03, 8.3711E+03, 8.3960E+03,
     + 8.4209E+03, 8.4459E+03, 8.4709E+03, 8.4959E+03, 8.5210E+03,
     +  8.5461E+03, 8.5713E+03, 8.5964E+03, 8.6216E+03, 8.6469E+03,
     + 8.6721E+03, 8.6974E+03, 8.7227E+03, 8.7481E+03, 8.7735E+03,
     +  8.7989E+03, 8.8244E+03, 8.8499E+03, 8.8754E+03, 8.9010E+03,
     + 8.9266E+03, 8.9522E+03, 8.9779E+03, 9.0036E+03, 9.0293E+03,
     +  9.0550E+03, 9.0808E+03, 9.1067E+03, 9.1325E+03, 9.1584E+03,
     + 9.1843E+03, 9.2103E+03, 9.2363E+03, 9.2623E+03, 9.2884E+03,
     +  9.3145E+03, 9.3406E+03, 9.3668E+03, 9.3930E+03, 9.4192E+03,
     + 9.4454E+03, 9.4717E+03, 9.4981E+03, 9.5244E+03, 9.5508E+03,
     +  9.5773E+03, 9.6037E+03, 9.6302E+03, 9.6567E+03, 9.6833E+03,
     + 9.7099E+03, 9.7365E+03, 9.7632E+03, 9.7899E+03, 9.8166E+03,
     +  9.8434E+03, 9.8702E+03, 9.8971E+03, 9.9239E+03, 9.9508E+03,
     + 9.9778E+03, 1.0005E+04, 1.0032E+04, 1.0059E+04, 1.0086E+04,
     +  1.0113E+04, 1.0140E+04, 1.0167E+04, 1.0195E+04, 1.0222E+04,
     + 1.0249E+04, 1.0276E+04, 1.0304E+04, 1.0331E+04, 1.0359E+04,
     +  1.0386E+04, 1.0413E+04, 1.0441E+04, 1.0469E+04, 1.0496E+04,
     + 1.0524E+04, 1.0551E+04, 1.0579E+04, 1.0607E+04, 1.0635E+04,
     +  1.0662E+04, 1.0690E+04, 1.0718E+04, 1.0746E+04, 1.0774E+04,
     + 1.0802E+04, 1.0830E+04, 1.0858E+04, 1.0886E+04, 1.0914E+04,
     +  1.0942E+04, 1.0971E+04, 1.0999E+04, 1.1027E+04, 1.1055E+04,
     + 1.1084E+04, 1.1112E+04, 1.1140E+04, 1.1169E+04, 1.1197E+04,
     +  1.1226E+04, 1.1254E+04, 1.1283E+04, 1.1311E+04, 1.1340E+04,
     + 1.1369E+04, 1.1397E+04, 1.1426E+04, 1.1455E+04, 1.1484E+04,
     +  1.1513E+04, 1.1542E+04, 1.1570E+04, 1.1599E+04, 1.1628E+04,
     + 1.1657E+04, 1.1687E+04, 1.1716E+04, 1.1745E+04, 1.1774E+04,
     +  1.1803E+04, 1.1832E+04, 1.1862E+04, 1.1891E+04, 1.1920E+04,
     + 1.1950E+04, 1.1979E+04, 1.2009E+04, 1.2038E+04, 1.2068E+04,
     +  1.2097E+04, 1.2127E+04, 1.2156E+04, 1.2186E+04, 1.2216E+04,
     + 1.2245E+04, 1.2275E+04, 1.2305E+04, 1.2335E+04, 1.2365E+04,
     +  1.2395E+04, 1.2425E+04, 1.2455E+04, 1.2485E+04, 1.2515E+04,
     + 1.2545E+04, 1.2575E+04, 1.2605E+04, 1.2635E+04, 1.2665E+04,
     +  1.2696E+04, 1.2726E+04, 1.2756E+04, 1.2787E+04, 1.2817E+04,
     + 1.2848E+04, 1.2878E+04, 1.2909E+04, 1.2939E+04, 1.2970E+04,
     +  1.3000E+04, 1.3031E+04, 1.3062E+04, 1.3093E+04, 1.3123E+04,
     + 1.3154E+04, 1.3185E+04, 1.3216E+04, 1.3247E+04, 1.3278E+04,
     +  1.3309E+04, 1.3340E+04, 1.3371E+04, 1.3402E+04, 1.3433E+04,
     + 1.3464E+04, 1.3495E+04, 1.3527E+04, 1.3558E+04, 1.3589E+04,
     +  1.3621E+04, 1.3652E+04, 1.3684E+04, 1.3715E+04, 1.3746E+04,
     + 1.3778E+04, 1.3810E+04, 1.3841E+04, 1.3873E+04, 1.3905E+04,
     +  1.3936E+04, 1.3968E+04, 1.4000E+04, 1.4032E+04, 1.4063E+04,
     + 1.4095E+04, 1.4127E+04, 1.4159E+04, 1.4191E+04, 1.4223E+04,
     +  1.4255E+04, 1.4288E+04, 1.4320E+04, 1.4352E+04, 1.4384E+04,
     + 1.4416E+04, 1.4449E+04, 1.4481E+04, 1.4513E+04, 1.4546E+04,
     +  1.4578E+04, 1.4611E+04, 1.4643E+04, 1.4676E+04, 1.4708E+04,
     + 1.4741E+04, 1.4774E+04, 1.4806E+04, 1.4839E+04, 1.4872E+04,
     +  1.4905E+04, 1.4938E+04, 1.4971E+04, 1.5003E+04, 1.5036E+04,
     + 1.5069E+04, 1.5102E+04, 1.5136E+04, 1.5169E+04, 1.5202E+04,
     +  1.5235E+04, 1.5268E+04, 1.5301E+04, 1.5335E+04, 1.5368E+04,
     + 1.5401E+04, 1.5435E+04, 1.5468E+04, 1.5502E+04, 1.5535E+04,
     +  1.5569E+04, 1.5602E+04, 1.5636E+04, 1.5670E+04, 1.5703E+04,
     + 1.5737E+04, 1.5771E+04, 1.5805E+04, 1.5839E+04, 1.5873E+04,
     +  1.5906E+04, 1.5940E+04, 1.5974E+04, 1.6008E+04, 1.6043E+04,
     + 1.6077E+04, 1.6111E+04, 1.6145E+04, 1.6179E+04, 1.6213E+04,
     +  1.6248E+04, 1.6282E+04, 1.6316E+04, 1.6351E+04, 1.6385E+04,
     + 1.6420E+04, 1.6454E+04, 1.6489E+04, 1.6523E+04, 1.6558E+04,
     +  1.6593E+04, 1.6627E+04, 1.6662E+04, 1.6697E+04, 1.6732E+04,
     + 1.6767E+04, 1.6802E+04, 1.6837E+04, 1.6871E+04, 1.6906E+04,
     +  1.6942E+04, 1.6977E+04, 1.7012E+04, 1.7047E+04, 1.7082E+04,
     + 1.7117E+04, 1.7153E+04, 1.7188E+04, 1.7223E+04, 1.7259E+04,
     +  1.7294E+04, 1.7330E+04, 1.7365E+04, 1.7401E+04, 1.7436E+04,
     + 1.7472E+04, 1.7507E+04, 1.7543E+04, 1.7579E+04, 1.7615E+04,
     +  1.7650E+04, 1.7686E+04, 1.7722E+04, 1.7758E+04, 1.7794E+04,
     + 1.7830E+04, 1.7866E+04, 1.7902E+04, 1.7938E+04, 1.7974E+04,
     +  1.8010E+04, 1.8047E+04, 1.8083E+04, 1.8119E+04, 1.8155E+04,
     + 1.8192E+04, 1.8228E+04, 1.8265E+04, 1.8301E+04, 1.8338E+04,
     +  1.8374E+04, 1.8411E+04, 1.8447E+04, 1.8484E+04, 1.8521E+04,
     + 1.8558E+04, 1.8594E+04, 1.8631E+04, 1.8668E+04, 1.8705E+04,
     +  1.8742E+04, 1.8779E+04, 1.8816E+04, 1.8853E+04, 1.8890E+04,
     + 1.8927E+04, 1.8964E+04, 1.9001E+04, 1.9039E+04, 1.9076E+04,
     +  1.9113E+04, 1.9151E+04, 1.9188E+04, 1.9225E+04, 1.9263E+04,
     + 1.9300E+04, 1.9338E+04, 1.9376E+04, 1.9413E+04, 1.9451E+04,
     +  1.9488E+04, 1.9526E+04, 1.9564E+04, 1.9602E+04, 1.9640E+04,
     + 1.9678E+04, 1.9715E+04, 1.9753E+04, 1.9791E+04, 1.9829E+04,
     +  1.9868E+04, 1.9906E+04, 1.9944E+04, 1.9982E+04, 2.0020E+04,
     + 2.0059E+04, 2.0097E+04, 2.0135E+04, 2.0174E+04, 2.0212E+04,
     +  2.0250E+04, 2.0289E+04, 2.0328E+04, 2.0366E+04, 2.0405E+04,
     + 2.0443E+04, 2.0482E+04, 2.0521E+04, 2.0560E+04, 2.0598E+04,
     +  2.0637E+04, 2.0676E+04, 2.0715E+04, 2.0754E+04, 2.0793E+04,
     + 2.0832E+04, 2.0871E+04, 2.0910E+04, 2.0949E+04, 2.0989E+04,
     +  2.1028E+04, 2.1067E+04, 2.1106E+04, 2.1146E+04, 2.1185E+04,
     + 2.1225E+04, 2.1264E+04, 2.1304E+04, 2.1343E+04, 2.1383E+04,
     +  2.1422E+04, 2.1462E+04, 2.1502E+04, 2.1542E+04, 2.1581E+04,
     + 2.1621E+04, 2.1661E+04, 2.1701E+04, 2.1741E+04, 2.1781E+04,
     +  2.1821E+04, 2.1861E+04, 2.1901E+04, 2.1941E+04, 2.1981E+04,
     + 2.2021E+04, 2.2062E+04, 2.2102E+04, 2.2142E+04, 2.2183E+04,
     +  2.2223E+04, 2.2264E+04, 2.2304E+04, 2.2345E+04, 2.2385E+04,
     + 2.2426E+04, 2.2466E+04, 2.2507E+04, 2.2548E+04, 2.2589E+04,
     +  2.2629E+04, 2.2670E+04, 2.2711E+04, 2.2752E+04, 2.2793E+04,
     + 2.2834E+04, 2.2875E+04, 2.2916E+04, 2.2957E+04, 2.2998E+04,
     +  2.3039E+04, 2.3081E+04, 2.3122E+04, 2.3163E+04, 2.3205E+04,
     + 2.3246E+04, 2.3287E+04, 2.3329E+04, 2.3370E+04, 2.3412E+04,
     +  2.3454E+04, 2.3495E+04, 2.3537E+04, 2.3579E+04, 2.3620E+04,
     + 2.3662E+04, 2.3704E+04, 2.3746E+04, 2.3788E+04, 2.3830E+04,
     +  2.3872E+04, 2.3914E+04, 2.3956E+04, 2.3998E+04, 2.4040E+04,
     + 2.4082E+04, 2.4124E+04, 2.4167E+04, 2.4209E+04, 2.4251E+04,
     +  2.4294E+04, 2.4336E+04, 2.4378E+04, 2.4421E+04, 2.4463E+04,
     + 2.4506E+04, 2.4549E+04, 2.4591E+04, 2.4634E+04, 2.4677E+04,
     +  2.4720E+04, 2.4762E+04, 2.4805E+04, 2.4848E+04, 2.4891E+04,
     + 2.4934E+04, 2.4977E+04, 2.5020E+04, 2.5063E+04, 2.5106E+04,
     +  2.5149E+04, 2.5193E+04, 2.5236E+04, 2.5279E+04, 2.5323E+04,
     + 2.5366E+04, 2.5409E+04, 2.5453E+04, 2.5496E+04, 2.5540E+04,
     +  2.5583E+04, 2.5627E+04, 2.5671E+04, 2.5714E+04, 2.5758E+04,
     + 2.5802E+04, 2.5846E+04, 2.5890E+04, 2.5933E+04, 2.5977E+04,
     +  2.6021E+04, 2.6065E+04, 2.6109E+04, 2.6154E+04, 2.6198E+04,
     + 2.6242E+04, 2.6286E+04, 2.6330E+04, 2.6375E+04, 2.6419E+04,
     +  2.6463E+04, 2.6508E+04, 2.6552E+04, 2.6597E+04, 2.6641E+04,
     + 2.6686E+04, 2.6731E+04, 2.6775E+04, 2.6820E+04, 2.6865E+04,
     +  2.6909E+04, 2.6954E+04, 2.6999E+04, 2.7044E+04, 2.7089E+04,
     + 2.7134E+04, 2.7179E+04, 2.7224E+04, 2.7269E+04, 2.7314E+04,
     +  2.7359E+04, 2.7405E+04, 2.7450E+04, 2.7495E+04, 2.7541E+04,
     + 2.7586E+04, 2.7631E+04, 2.7677E+04, 2.7722E+04, 2.7768E+04,
     +  2.7814E+04, 2.7859E+04, 2.7905E+04, 2.7951E+04, 2.7996E+04,
     + 2.8042E+04, 2.8088E+04, 2.8134E+04, 2.8180E+04, 2.8226E+04,
     +  2.8272E+04, 2.8318E+04, 2.8364E+04, 2.8410E+04, 2.8456E+04,
     + 2.8502E+04, 2.8549E+04, 2.8595E+04, 2.8641E+04, 2.8688E+04,
     +  2.8734E+04, 2.8781E+04, 2.8827E+04, 2.8874E+04, 2.8920E+04,
     + 2.8967E+04, 2.9013E+04, 2.9060E+04, 2.9107E+04, 2.9154E+04,
     +  2.9201E+04, 2.9247E+04, 2.9294E+04, 2.9341E+04, 2.9388E+04,
     + 2.9435E+04, 2.9482E+04, 2.9529E+04, 2.9577E+04, 2.9624E+04,
     +  2.9671E+04, 2.9718E+04, 2.9766E+04, 2.9813E+04, 2.9860E+04,
     + 2.9908E+04, 2.9955E+04, 3.0003E+04, 3.0050E+04, 3.0098E+04,
     +  3.0146E+04, 3.0193E+04, 3.0241E+04, 3.0289E+04, 3.0337E+04,
     + 3.0385E+04, 3.0432E+04, 3.0480E+04, 3.0528E+04, 3.0576E+04,
     +  3.0624E+04, 3.0673E+04, 3.0721E+04, 3.0769E+04, 3.0817E+04,
     + 3.0865E+04, 3.0914E+04, 3.0962E+04, 3.1010E+04, 3.1059E+04,
     +  3.1107E+04, 3.1156E+04, 3.1204E+04, 3.1253E+04, 3.1302E+04,
     + 3.1350E+04, 3.1399E+04, 3.1448E+04, 3.1497E+04, 3.1546E+04,
     +  3.1594E+04, 3.1643E+04, 3.1692E+04, 3.1741E+04, 3.1790E+04,
     + 3.1840E+04, 3.1889E+04, 3.1938E+04, 3.1987E+04, 3.2036E+04,
     +  3.2086E+04, 3.2135E+04, 3.2185E+04, 3.2234E+04, 3.2283E+04,
     + 3.2333E+04, 3.2383E+04, 3.2432E+04, 3.2482E+04, 3.2531E+04,
     +  3.2581E+04, 3.2631E+04, 3.2681E+04, 3.2731E+04, 3.2781E+04,
     + 3.2831E+04, 3.2881E+04, 3.2931E+04, 3.2981E+04, 3.3031E+04,
     +  3.3081E+04, 3.3131E+04, 3.3181E+04, 3.3232E+04, 3.3282E+04,
     + 3.3332E+04, 3.3383E+04, 3.3433E+04, 3.3484E+04, 3.3534E+04,
     +  3.3585E+04, 3.3635E+04, 3.3686E+04, 3.3737E+04, 3.3788E+04,
     + 3.3838E+04, 3.3889E+04, 3.3940E+04, 3.3991E+04, 3.4042E+04,
     +  3.4093E+04, 3.4144E+04, 3.4195E+04, 3.4246E+04, 3.4297E+04,
     + 3.4349E+04, 3.4400E+04, 3.4451E+04, 3.4502E+04, 3.4554E+04,
     +  3.4605E+04, 3.4657E+04, 3.4708E+04, 3.4760E+04, 3.4811E+04,
     + 3.4863E+04, 3.4915E+04, 3.4966E+04, 3.5018E+04, 3.5070E+04,
     +  3.5122E+04, 3.5174E+04, 3.5226E+04, 3.5278E+04, 3.5330E+04,
     + 3.5382E+04, 3.5434E+04, 3.5486E+04, 3.5538E+04, 3.5591E+04,
     +  3.5643E+04, 3.5695E+04, 3.5747E+04, 3.5800E+04, 3.5852E+04,
     + 3.5905E+04, 3.5957E+04, 3.6010E+04, 3.6063E+04, 3.6115E+04,
     +  3.6168E+04, 3.6221E+04, 3.6274E+04, 3.6326E+04, 3.6379E+04,
     + 3.6432E+04, 3.6485E+04, 3.6538E+04, 3.6591E+04, 3.6644E+04,
     +  3.6697E+04, 3.6751E+04, 3.6804E+04, 3.6857E+04, 3.6910E+04,
     + 3.6964E+04, 3.7017E+04, 3.7071E+04, 3.7124E+04, 3.7178E+04,
     +  3.7231E+04, 3.7285E+04, 3.7339E+04, 3.7392E+04, 3.7446E+04,
     + 3.7500E+04, 3.7554E+04, 3.7607E+04, 3.7661E+04, 3.7715E+04,
     +  3.7769E+04, 3.7823E+04, 3.7877E+04, 3.7932E+04, 3.7986E+04,
     + 3.8040E+04, 3.8094E+04, 3.8149E+04, 3.8203E+04, 3.8257E+04,
     +  3.8312E+04, 3.8366E+04, 3.8421E+04, 3.8475E+04, 3.8530E+04,
     + 3.8585E+04, 3.8639E+04, 3.8694E+04, 3.8749E+04, 3.8804E+04,
     +  3.8859E+04, 3.8913E+04, 3.8968E+04, 3.9023E+04, 3.9079E+04,
     + 3.9134E+04, 3.9189E+04, 3.9244E+04, 3.9299E+04, 3.9354E+04,
     +  3.9410E+04, 3.9465E+04, 3.9520E+04, 3.9576E+04, 3.9631E+04,
     + 3.9687E+04, 3.9742E+04, 3.9798E+04, 3.9854E+04, 3.9909E+04,
     +  3.9965E+04, 4.0021E+04, 4.0077E+04, 4.0133E+04, 4.0189E+04,
     + 4.0245E+04, 4.0301E+04, 4.0357E+04, 4.0413E+04, 4.0469E+04,
     +  4.0525E+04, 4.0581E+04, 4.0638E+04, 4.0694E+04, 4.0750E+04,
     + 4.0807E+04, 4.0863E+04, 4.0920E+04, 4.0976E+04, 4.1033E+04,
     +  4.1089E+04, 4.1146E+04, 4.1203E+04, 4.1260E+04, 4.1316E+04,
     + 4.1373E+04, 4.1430E+04, 4.1487E+04, 4.1544E+04, 4.1601E+04,
     +  4.1658E+04, 4.1715E+04, 4.1772E+04, 4.1830E+04, 4.1887E+04,
     + 4.1944E+04, 4.2002E+04, 4.2059E+04, 4.2116E+04, 4.2174E+04,
     +  4.2231E+04, 4.2289E+04, 4.2346E+04, 4.2404E+04, 4.2462E+04,
     + 4.2520E+04, 4.2577E+04, 4.2635E+04, 4.2693E+04, 4.2751E+04,
     +  4.2809E+04, 4.2867E+04, 4.2925E+04, 4.2983E+04, 4.3041E+04,
     + 4.3099E+04, 4.3158E+04, 4.3216E+04, 4.3274E+04, 4.3332E+04,
     +  4.3391E+04, 4.3449E+04, 4.3508E+04, 4.3566E+04, 4.3625E+04,
     + 4.3684E+04, 4.3742E+04, 4.3801E+04, 4.3860E+04, 4.3919E+04,
     +  4.3977E+04, 4.4036E+04, 4.4095E+04, 4.4154E+04, 4.4213E+04,
     + 4.4272E+04, 4.4331E+04, 4.4391E+04, 4.4450E+04, 4.4509E+04,
     +  4.4568E+04, 4.4628E+04, 4.4687E+04, 4.4746E+04, 4.4806E+04,
     + 4.4865E+04, 4.4925E+04, 4.4985E+04, 4.5044E+04, 4.5104E+04,
     +  4.5164E+04, 4.5224E+04, 4.5283E+04, 4.5343E+04, 4.5403E+04,
     + 4.5463E+04, 4.5523E+04, 4.5583E+04, 4.5643E+04, 4.5703E+04,
     +  4.5764E+04, 4.5824E+04, 4.5884E+04, 4.5945E+04, 4.6005E+04,
     + 4.6065E+04, 4.6126E+04, 4.6186E+04, 4.6247E+04, 4.6307E+04,
     +  4.6368E+04, 4.6429E+04, 4.6490E+04, 4.6550E+04, 4.6611E+04,
     + 4.6672E+04, 4.6733E+04, 4.6794E+04, 4.6855E+04, 4.6916E+04,
     +  4.6977E+04, 4.7038E+04, 4.7099E+04, 4.7161E+04, 4.7222E+04,
     + 4.7283E+04, 4.7345E+04, 4.7406E+04, 4.7467E+04, 4.7529E+04,
     +  4.7591E+04, 4.7652E+04, 4.7714E+04, 4.7775E+04, 4.7837E+04,
     + 4.7899E+04, 4.7961E+04, 4.8023E+04, 4.8085E+04, 4.8147E+04,
     +  4.8209E+04, 4.8271E+04, 4.8333E+04, 4.8395E+04, 4.8457E+04,
     + 4.8519E+04, 4.8582E+04, 4.8644E+04, 4.8706E+04, 4.8769E+04,
     +  4.8831E+04, 4.8894E+04, 4.8956E+04, 4.9019E+04, 4.9081E+04,
     + 4.9144E+04, 4.9207E+04, 4.9270E+04, 4.9332E+04, 4.9395E+04,
     +  4.9458E+04, 4.9521E+04, 4.9584E+04, 4.9647E+04, 4.9710E+04,
     + 4.9774E+04, 4.9837E+04, 4.9900E+04, 4.9963E+04, 5.0027E+04,
     +  5.0090E+04, 5.0153E+04, 5.0217E+04, 5.0280E+04, 5.0344E+04,
     + 5.0408E+04, 5.0471E+04, 5.0535E+04, 5.0599E+04, 5.0662E+04,
     +  5.0726E+04, 5.0790E+04, 5.0854E+04, 5.0918E+04, 5.0982E+04,
     + 5.1046E+04, 5.1110E+04, 5.1174E+04, 5.1239E+04, 5.1303E+04,
     +  5.1367E+04, 5.1432E+04, 5.1496E+04, 5.1560E+04, 5.1625E+04,
     + 5.1689E+04, 5.1754E+04, 5.1819E+04, 5.1883E+04, 5.1948E+04,
     +  5.2013E+04, 5.2077E+04, 5.2142E+04, 5.2207E+04, 5.2272E+04,
     + 5.2337E+04, 5.2402E+04, 5.2467E+04, 5.2532E+04, 5.2598E+04,
     +  5.2663E+04, 5.2728E+04, 5.2793E+04, 5.2859E+04, 5.2924E+04,
     + 5.2990E+04, 5.3055E+04, 5.3121E+04, 5.3186E+04, 5.3252E+04,
     +  5.3318E+04, 5.3383E+04, 5.3449E+04, 5.3515E+04, 5.3581E+04,
     + 5.3647E+04, 5.3713E+04, 5.3779E+04, 5.3845E+04, 5.3911E+04,
     +  5.3977E+04, 5.4043E+04, 5.4109E+04, 5.4176E+04, 5.4242E+04,
     + 5.4308E+04, 5.4375E+04, 5.4441E+04, 5.4508E+04, 5.4574E+04,
     +  5.4641E+04, 5.4708E+04, 5.4774E+04, 5.4841E+04, 5.4908E+04,
     + 5.4975E+04, 5.5042E+04, 5.5109E+04, 5.5176E+04, 5.5243E+04,
     +  5.5310E+04, 5.5377E+04, 5.5444E+04, 5.5511E+04, 5.5579E+04,
     + 5.5646E+04, 5.5713E+04, 5.5781E+04, 5.5848E+04, 5.5916E+04,
     +  5.5983E+04, 5.6051E+04, 5.6118E+04, 5.6186E+04, 5.6254E+04,
     + 5.6322E+04, 5.6389E+04, 5.6457E+04, 5.6525E+04, 5.6593E+04,
     +  5.6661E+04, 5.6729E+04, 5.6797E+04, 5.6866E+04, 5.6934E+04,
     + 5.7002E+04, 5.7070E+04, 5.7139E+04, 5.7207E+04, 5.7275E+04,
     +  5.7344E+04, 5.7412E+04, 5.7481E+04, 5.7550E+04, 5.7618E+04,
     + 5.7687E+04, 5.7756E+04, 5.7825E+04, 5.7894E+04, 5.7962E+04,
     +  5.8031E+04, 5.8100E+04, 5.8169E+04, 5.8239E+04, 5.8308E+04,
     + 5.8377E+04, 5.8446E+04, 5.8515E+04, 5.8585E+04, 5.8654E+04,
     +  5.8724E+04, 5.8793E+04, 5.8863E+04, 5.8932E+04, 5.9002E+04,
     + 5.9071E+04, 5.9141E+04, 5.9211E+04, 5.9281E+04, 5.9351E+04,
     +  5.9420E+04, 5.9490E+04, 5.9560E+04, 5.9630E+04, 5.9701E+04,
     + 5.9771E+04, 5.9841E+04, 5.9911E+04, 5.9981E+04, 6.0052E+04,
     +  6.0122E+04, 6.0192E+04, 6.0263E+04, 6.0333E+04, 6.0404E+04,
     + 6.0475E+04, 6.0545E+04, 6.0616E+04, 6.0687E+04, 6.0758E+04,
     +  6.0828E+04, 6.0899E+04, 6.0970E+04, 6.1041E+04, 6.1112E+04,
     + 6.1183E+04, 6.1255E+04, 6.1326E+04, 6.1397E+04, 6.1468E+04,
     +  6.1540E+04, 6.1611E+04, 6.1682E+04, 6.1754E+04, 6.1825E+04,
     + 6.1897E+04, 6.1968E+04, 6.2040E+04, 6.2112E+04, 6.2184E+04,
     +  6.2255E+04, 6.2327E+04, 6.2399E+04, 6.2471E+04, 6.2543E+04,
     + 6.2615E+04, 6.2687E+04, 6.2759E+04, 6.2832E+04, 6.2904E+04,
     +  6.2976E+04, 6.3048E+04, 6.3121E+04, 6.3193E+04, 6.3266E+04,
     + 6.3338E+04, 6.3411E+04, 6.3483E+04, 6.3556E+04, 6.3629E+04,
     +  6.3701E+04, 6.3774E+04, 6.3847E+04, 6.3920E+04, 6.3993E+04,
     + 6.4066E+04, 6.4139E+04, 6.4212E+04, 6.4285E+04, 6.4358E+04,
     +  6.4432E+04, 6.4505E+04, 6.4578E+04, 6.4652E+04, 6.4725E+04,
     + 6.4799E+04, 6.4872E+04, 6.4946E+04, 6.5019E+04, 6.5093E+04,
     +  6.5167E+04, 6.5240E+04, 6.5314E+04, 6.5388E+04, 6.5462E+04,
     + 6.5536E+04, 6.5610E+04, 6.5684E+04, 6.5758E+04, 6.5832E+04,
     +  6.5906E+04, 6.5981E+04, 6.6055E+04, 6.6129E+04, 6.6204E+04,
     + 6.6278E+04, 6.6353E+04, 6.6427E+04, 6.6502E+04, 6.6576E+04,
     +  6.6651E+04, 6.6726E+04, 6.6801E+04, 6.6875E+04, 6.6950E+04,
     + 6.7025E+04, 6.7100E+04, 6.7175E+04, 6.7250E+04, 6.7325E+04,
     +  6.7400E+04, 6.7476E+04, 6.7551E+04, 6.7626E+04, 6.7702E+04,
     + 6.7777E+04, 6.7852E+04, 6.7928E+04, 6.8003E+04, 6.8079E+04,
     +  6.8155E+04, 6.8230E+04, 6.8306E+04, 6.8382E+04, 6.8458E+04,
     + 6.8534E+04, 6.8610E+04, 6.8686E+04, 6.8762E+04, 6.8838E+04,
     +  6.8914E+04, 6.8990E+04, 6.9066E+04, 6.9142E+04, 6.9219E+04,
     + 6.9295E+04, 6.9372E+04, 6.9448E+04, 6.9525E+04, 6.9601E+04,
     +  6.9678E+04, 6.9754E+04, 6.9831E+04, 6.9908E+04, 6.9985E+04,
     + 7.0061E+04, 7.0138E+04, 7.0215E+04, 7.0292E+04, 7.0369E+04,
     +  7.0446E+04, 7.0524E+04, 7.0601E+04, 7.0678E+04, 7.0755E+04,
     + 7.0833E+04, 7.0910E+04, 7.0988E+04, 7.1065E+04, 7.1143E+04,
     +  7.1220E+04, 7.1298E+04, 7.1375E+04, 7.1453E+04, 7.1531E+04,
     + 7.1609E+04, 7.1687E+04, 7.1765E+04, 7.1843E+04, 7.1921E+04,
     +  7.1999E+04, 7.2077E+04, 7.2155E+04, 7.2233E+04, 7.2311E+04,
     + 7.2390E+04, 7.2468E+04, 7.2546E+04, 7.2625E+04, 7.2703E+04,
     +  7.2782E+04, 7.2861E+04, 7.2939E+04, 7.3018E+04, 7.3097E+04,
     + 7.3176E+04, 7.3254E+04, 7.3333E+04, 7.3412E+04, 7.3491E+04,
     +  7.3570E+04, 7.3649E+04, 7.3728E+04, 7.3808E+04, 7.3887E+04,
     + 7.3966E+04, 7.4046E+04, 7.4125E+04, 7.4204E+04, 7.4284E+04,
     +  7.4363E+04, 7.4443E+04, 7.4523E+04, 7.4602E+04, 7.4682E+04,
     + 7.4762E+04, 7.4842E+04, 7.4921E+04, 7.5001E+04, 7.5081E+04,
     +  7.5161E+04, 7.5241E+04, 7.5322E+04, 7.5402E+04, 7.5482E+04,
     + 7.5562E+04, 7.5643E+04, 7.5723E+04, 7.5803E+04, 7.5884E+04,
     +  7.5964E+04, 7.6045E+04, 7.6126E+04, 7.6206E+04, 7.6287E+04,
     + 7.6368E+04, 7.6448E+04, 7.6529E+04, 7.6610E+04, 7.6691E+04,
     +  7.6772E+04, 7.6853E+04, 7.6934E+04, 7.7016E+04, 7.7097E+04,
     + 7.7178E+04, 7.7259E+04, 7.7341E+04, 7.7422E+04, 7.7503E+04,
     +  7.7585E+04, 7.7666E+04, 7.7748E+04, 7.7830E+04, 7.7911E+04,
     + 7.7993E+04, 7.8075E+04, 7.8157E+04, 7.8239E+04, 7.8321E+04,
     +  7.8403E+04, 7.8485E+04, 7.8567E+04, 7.8649E+04, 7.8731E+04,
     + 7.8813E+04, 7.8896E+04, 7.8978E+04, 7.9060E+04, 7.9143E+04,
     +  7.9225E+04, 7.9308E+04, 7.9390E+04, 7.9473E+04, 7.9556E+04,
     + 7.9638E+04, 7.9721E+04, 7.9804E+04, 7.9887E+04, 7.9970E+04,
     +  8.0053E+04, 8.0136E+04, 8.0219E+04, 8.0302E+04, 8.0385E+04,
     + 8.0468E+04, 8.0552E+04, 8.0635E+04, 8.0718E+04, 8.0802E+04,
     +  8.0885E+04, 8.0969E+04, 8.1052E+04, 8.1136E+04, 8.1220E+04,
     + 8.1303E+04, 8.1387E+04, 8.1471E+04, 8.1555E+04, 8.1639E+04,
     +  8.1723E+04, 8.1807E+04, 8.1891E+04, 8.1975E+04, 8.2059E+04,
     + 8.2143E+04, 8.2228E+04, 8.2312E+04, 8.2396E+04, 8.2481E+04,
     +  8.2565E+04, 8.2650E+04, 8.2734E+04, 8.2819E+04, 8.2903E+04,
     + 8.2988E+04, 8.3073E+04, 8.3158E+04, 8.3243E+04, 8.3327E+04,
     +  8.3412E+04, 8.3497E+04, 8.3582E+04, 8.3668E+04, 8.3753E+04,
     + 8.3838E+04, 8.3923E+04, 8.4009E+04, 8.4094E+04, 8.4179E+04,
     +  8.4265E+04, 8.4350E+04, 8.4436E+04, 8.4521E+04, 8.4607E+04,
     + 8.4693E+04, 8.4778E+04, 8.4864E+04, 8.4950E+04, 8.5036E+04,
     +  8.5122E+04, 8.5208E+04, 8.5294E+04, 8.5380E+04, 8.5466E+04,
     + 8.5552E+04, 8.5639E+04, 8.5725E+04, 8.5811E+04, 8.5898E+04,
     +  8.5984E+04, 8.6071E+04, 8.6157E+04, 8.6244E+04, 8.6331E+04,
     + 8.6417E+04, 8.6504E+04, 8.6591E+04, 8.6678E+04, 8.6765E+04,
     +  8.6851E+04, 8.6938E+04, 8.7026E+04, 8.7113E+04, 8.7200E+04,
     + 8.7287E+04, 8.7374E+04, 8.7461E+04, 8.7549E+04, 8.7636E+04,
     +  8.7724E+04, 8.7811E+04, 8.7899E+04, 8.7986E+04, 8.8074E+04,
     + 8.8162E+04, 8.8249E+04, 8.8337E+04, 8.8425E+04, 8.8513E+04,
     +  8.8601E+04, 8.8689E+04, 8.8777E+04, 8.8865E+04, 8.8953E+04,
     + 8.9041E+04, 8.9130E+04, 8.9218E+04, 8.9306E+04, 8.9395E+04,
     +  8.9483E+04, 8.9572E+04, 8.9660E+04, 8.9749E+04, 8.9837E+04,
     + 8.9926E+04, 9.0015E+04, 9.0104E+04, 9.0192E+04, 9.0281E+04,
     +  9.0370E+04, 9.0459E+04, 9.0548E+04, 9.0637E+04, 9.0727E+04,
     + 9.0816E+04, 9.0905E+04, 9.0994E+04, 9.1084E+04, 9.1173E+04,
     +  9.1262E+04, 9.1352E+04, 9.1442E+04, 9.1531E+04, 9.1621E+04,
     + 9.1710E+04, 9.1800E+04, 9.1890E+04, 9.1980E+04, 9.2070E+04,
     +  9.2160E+04, 9.2250E+04, 9.2340E+04, 9.2430E+04, 9.2520E+04,
     + 9.2610E+04, 9.2700E+04, 9.2791E+04, 9.2881E+04, 9.2971E+04,
     +  9.3062E+04, 9.3152E+04, 9.3243E+04, 9.3334E+04, 9.3424E+04,
     + 9.3515E+04, 9.3606E+04, 9.3696E+04, 9.3787E+04, 9.3878E+04,
     +  9.3969E+04, 9.4060E+04, 9.4151E+04, 9.4242E+04, 9.4333E+04,
     + 9.4425E+04, 9.4516E+04, 9.4607E+04, 9.4699E+04, 9.4790E+04,
     +  9.4881E+04, 9.4973E+04, 9.5064E+04, 9.5156E+04, 9.5248E+04,
     + 9.5339E+04, 9.5431E+04, 9.5523E+04, 9.5615E+04, 9.5707E+04,
     +  9.5799E+04, 9.5891E+04, 9.5983E+04, 9.6075E+04, 9.6167E+04,
     + 9.6259E+04, 9.6351E+04, 9.6444E+04, 9.6536E+04, 9.6629E+04,
     +  9.6721E+04, 9.6813E+04, 9.6906E+04, 9.6999E+04, 9.7091E+04,
     + 9.7184E+04, 9.7277E+04, 9.7370E+04, 9.7462E+04, 9.7555E+04,
     +  9.7648E+04, 9.7741E+04, 9.7834E+04, 9.7928E+04, 9.8021E+04,
     + 9.8114E+04, 9.8207E+04, 9.8300E+04, 9.8394E+04, 9.8487E+04,
     +  9.8581E+04, 9.8674E+04, 9.8768E+04, 9.8861E+04, 9.8955E+04,
     + 9.9049E+04, 9.9143E+04, 9.9236E+04, 9.9330E+04, 9.9424E+04,
     +  9.9518E+04, 9.9612E+04, 9.9706E+04, 9.9800E+04, 9.9895E+04,
     + 9.9989E+04, 1.0008E+05, 1.0018E+05, 1.0027E+05, 1.0037E+05,
     +  1.0046E+05, 1.0056E+05, 1.0065E+05, 1.0074E+05, 1.0084E+05,
     + 1.0093E+05, 1.0103E+05, 1.0112E+05, 1.0122E+05, 1.0131E+05,
     +  1.0141E+05, 1.0150E+05, 1.0160E+05, 1.0169E+05, 1.0179E+05,
     + 1.0188E+05, 1.0198E+05, 1.0207E+05, 1.0217E+05, 1.0227E+05,
     +  1.0236E+05, 1.0246E+05, 1.0255E+05, 1.0265E+05, 1.0274E+05,
     + 1.0284E+05, 1.0293E+05, 1.0303E+05, 1.0313E+05, 1.0322E+05,
     +  1.0332E+05, 1.0341E+05, 1.0351E+05, 1.0361E+05, 1.0370E+05,
     + 1.0380E+05, 1.0390E+05, 1.0399E+05, 1.0409E+05, 1.0418E+05,
     +  1.0428E+05, 1.0438E+05, 1.0447E+05, 1.0457E+05, 1.0467E+05,
     + 1.0476E+05, 1.0486E+05, 1.0496E+05, 1.0505E+05, 1.0515E+05,
     +  1.0525E+05, 1.0535E+05, 1.0544E+05, 1.0554E+05, 1.0564E+05,
     + 1.0573E+05, 1.0583E+05, 1.0593E+05, 1.0603E+05, 1.0612E+05,
     +  1.0622E+05, 1.0632E+05, 1.0642E+05, 1.0651E+05, 1.0661E+05,
     + 1.0671E+05, 1.0681E+05, 1.0691E+05, 1.0700E+05, 1.0710E+05,
     +  1.0720E+05, 1.0730E+05, 1.0740E+05, 1.0749E+05, 1.0759E+05,
     + 1.0769E+05, 1.0779E+05, 1.0789E+05, 1.0799E+05, 1.0808E+05,
     +  1.0818E+05, 1.0828E+05, 1.0838E+05, 1.0848E+05, 1.0858E+05,
     + 1.0868E+05, 1.0877E+05, 1.0887E+05, 1.0897E+05, 1.0907E+05,
     +  1.0917E+05, 1.0927E+05, 1.0937E+05, 1.0947E+05, 1.0957E+05,
     + 1.0967E+05, 1.0977E+05, 1.0987E+05, 1.0996E+05, 1.1006E+05,
     +  1.1016E+05, 1.1026E+05, 1.1036E+05, 1.1046E+05, 1.1056E+05,
     + 1.1066E+05, 1.1076E+05, 1.1086E+05, 1.1096E+05, 1.1106E+05,
     +  1.1116E+05, 1.1126E+05, 1.1136E+05, 1.1146E+05, 1.1156E+05,
     + 1.1166E+05, 1.1176E+05, 1.1186E+05, 1.1196E+05, 1.1206E+05,
     +  1.1216E+05, 1.1227E+05, 1.1237E+05, 1.1247E+05, 1.1257E+05,
     + 1.1267E+05, 1.1277E+05, 1.1287E+05, 1.1297E+05, 1.1307E+05,
     +  1.1317E+05, 1.1327E+05, 1.1338E+05, 1.1348E+05, 1.1358E+05,
     + 1.1368E+05, 1.1378E+05, 1.1388E+05, 1.1398E+05, 1.1409E+05,
     +  1.1419E+05, 1.1429E+05, 1.1439E+05, 1.1449E+05, 1.1459E+05,
     + 1.1470E+05, 1.1480E+05, 1.1490E+05, 1.1500E+05, 1.1510E+05,
     +  1.1521E+05, 1.1531E+05, 1.1541E+05, 1.1551E+05, 1.1561E+05,
     + 1.1572E+05, 1.1582E+05, 1.1592E+05, 1.1602E+05, 1.1613E+05,
     +  1.1623E+05, 1.1633E+05, 1.1643E+05, 1.1654E+05, 1.1664E+05,
     + 1.1674E+05, 1.1685E+05, 1.1695E+05, 1.1705E+05, 1.1715E+05,
     +  1.1726E+05, 1.1736E+05, 1.1746E+05, 1.1757E+05, 1.1767E+05,
     + 1.1777E+05, 1.1788E+05, 1.1798E+05, 1.1808E+05, 1.1819E+05,
     +  1.1829E+05, 1.1839E+05, 1.1850E+05, 1.1860E+05, 1.1871E+05,
     + 1.1881E+05, 1.1891E+05, 1.1902E+05, 1.1912E+05, 1.1923E+05,
     +  1.1933E+05, 1.1943E+05, 1.1954E+05, 1.1964E+05, 1.1975E+05,
     + 1.1985E+05, 1.1996E+05, 1.2006E+05, 1.2016E+05, 1.2027E+05,
     +  1.2037E+05, 1.2048E+05, 1.2058E+05, 1.2069E+05, 1.2079E+05,
     + 1.2090E+05, 1.2100E+05, 1.2111E+05, 1.2121E+05, 1.2132E+05,
     +  1.2142E+05, 1.2153E+05, 1.2163E+05, 1.2174E+05, 1.2184E+05,
     + 1.2195E+05, 1.2205E+05, 1.2216E+05, 1.2227E+05, 1.2237E+05,
     +  1.2248E+05, 1.2258E+05, 1.2269E+05, 1.2279E+05, 1.2290E+05,
     + 1.2301E+05, 1.2311E+05, 1.2322E+05, 1.2332E+05, 1.2343E+05,
     +  1.2354E+05, 1.2364E+05, 1.2375E+05, 1.2385E+05, 1.2396E+05,
     + 1.2407E+05, 1.2417E+05, 1.2428E+05, 1.2439E+05, 1.2449E+05,
     +  1.2460E+05, 1.2471E+05, 1.2481E+05, 1.2492E+05, 1.2503E+05,
     + 1.2513E+05, 1.2524E+05, 1.2535E+05, 1.2546E+05, 1.2556E+05,
     +  1.2567E+05, 1.2578E+05, 1.2588E+05, 1.2599E+05, 1.2610E+05,
     + 1.2621E+05, 1.2631E+05, 1.2642E+05, 1.2653E+05, 1.2664E+05,
     +  1.2674E+05, 1.2685E+05, 1.2696E+05, 1.2707E+05, 1.2718E+05,
     + 1.2728E+05, 1.2739E+05, 1.2750E+05, 1.2761E+05, 1.2772E+05,
     +  1.2782E+05, 1.2793E+05, 1.2804E+05, 1.2815E+05, 1.2826E+05,
     + 1.2837E+05, 1.2847E+05, 1.2858E+05, 1.2869E+05, 1.2880E+05,
     +  1.2891E+05, 1.2902E+05, 1.2913E+05, 1.2923E+05, 1.2934E+05,
     + 1.2945E+05, 1.2956E+05, 1.2967E+05, 1.2978E+05, 1.2989E+05,
     +  1.3000E+05, 1.3011E+05, 1.3022E+05, 1.3033E+05, 1.3044E+05,
     + 1.3055E+05, 1.3065E+05, 1.3076E+05, 1.3087E+05, 1.3098E+05,
     +  1.3109E+05, 1.3120E+05, 1.3131E+05, 1.3142E+05, 1.3153E+05,
     + 1.3164E+05, 1.3175E+05, 1.3186E+05, 1.3197E+05, 1.3208E+05,
     +  1.3219E+05, 1.3230E+05, 1.3241E+05, 1.3252E+05, 1.3263E+05,
     + 1.3275E+05, 1.3286E+05, 1.3297E+05, 1.3308E+05, 1.3319E+05,
     +  1.3330E+05, 1.3341E+05, 1.3352E+05, 1.3363E+05, 1.3374E+05,
     + 1.3385E+05, 1.3396E+05, 1.3407E+05, 1.3419E+05, 1.3430E+05,
     +  1.3441E+05, 1.3452E+05, 1.3463E+05, 1.3474E+05, 1.3485E+05,
     + 1.3497E+05, 1.3508E+05, 1.3519E+05, 1.3530E+05, 1.3541E+05,
     +  1.3552E+05, 1.3564E+05, 1.3575E+05, 1.3586E+05, 1.3597E+05,
     + 1.3608E+05, 1.3620E+05, 1.3631E+05, 1.3642E+05, 1.3653E+05,
     +  1.3664E+05, 1.3676E+05, 1.3687E+05, 1.3698E+05, 1.3709E+05,
     + 1.3721E+05, 1.3732E+05, 1.3743E+05, 1.3754E+05, 1.3766E+05,
     +  1.3777E+05, 1.3788E+05, 1.3800E+05, 1.3811E+05, 1.3822E+05,
     + 1.3833E+05, 1.3845E+05, 1.3856E+05, 1.3867E+05, 1.3879E+05,
     +  1.3890E+05, 1.3901E+05, 1.3913E+05, 1.3924E+05, 1.3935E+05,
     + 1.3947E+05, 1.3958E+05, 1.3969E+05, 1.3981E+05, 1.3992E+05,
     +  1.4004E+05, 1.4015E+05, 1.4026E+05, 1.4038E+05, 1.4049E+05/
      data (QofT( 5,J),J=2409,4510)/ 1.4061E+05, 1.4072E+05,
     +  1.4083E+05, 1.4095E+05, 1.4106E+05, 1.4118E+05, 1.4129E+05,
     +  1.4141E+05,
     + 1.4152E+05, 1.4163E+05, 1.4175E+05, 1.4186E+05, 1.4198E+05,
     +  1.4209E+05, 1.4221E+05, 1.4232E+05, 1.4244E+05, 1.4255E+05,
     + 1.4267E+05, 1.4278E+05, 1.4290E+05, 1.4301E+05, 1.4313E+05,
     +  1.4324E+05, 1.4336E+05, 1.4347E+05, 1.4359E+05, 1.4370E+05,
     + 1.4382E+05, 1.4394E+05, 1.4405E+05, 1.4417E+05, 1.4428E+05,
     +  1.4440E+05, 1.4451E+05, 1.4463E+05, 1.4475E+05, 1.4486E+05,
     + 1.4498E+05, 1.4509E+05, 1.4521E+05, 1.4533E+05, 1.4544E+05,
     +  1.4556E+05, 1.4567E+05, 1.4579E+05, 1.4591E+05, 1.4602E+05,
     + 1.4614E+05, 1.4626E+05, 1.4637E+05, 1.4649E+05, 1.4661E+05,
     +  1.4672E+05, 1.4684E+05, 1.4696E+05, 1.4707E+05, 1.4719E+05,
     + 1.4731E+05, 1.4742E+05, 1.4754E+05, 1.4766E+05, 1.4778E+05,
     +  1.4789E+05, 1.4801E+05, 1.4813E+05, 1.4825E+05, 1.4836E+05,
     + 1.4848E+05, 1.4860E+05, 1.4872E+05, 1.4883E+05, 1.4895E+05,
     +  1.4907E+05, 1.4919E+05, 1.4930E+05, 1.4942E+05, 1.4954E+05,
     + 1.4966E+05, 1.4978E+05, 1.4989E+05, 1.5001E+05, 1.5013E+05,
     +  1.5025E+05, 1.5037E+05, 1.5049E+05, 1.5060E+05, 1.5072E+05,
     + 1.5084E+05, 1.5096E+05, 1.5108E+05, 1.5120E+05, 1.5132E+05,
     +  1.5143E+05, 1.5155E+05, 1.5167E+05, 1.5179E+05, 1.5191E+05,
     + 1.5203E+05, 1.5215E+05, 1.5227E+05, 1.5239E+05, 1.5251E+05,
     +  1.5263E+05, 1.5274E+05, 1.5286E+05, 1.5298E+05, 1.5310E+05,
     + 1.5322E+05, 1.5334E+05, 1.5346E+05, 1.5358E+05, 1.5370E+05,
     +  1.5382E+05, 1.5394E+05, 1.5406E+05, 1.5418E+05, 1.5430E+05,
     + 1.5442E+05, 1.5454E+05, 1.5466E+05, 1.5478E+05, 1.5490E+05,
     +  1.5502E+05, 1.5514E+05, 1.5526E+05, 1.5538E+05, 1.5550E+05,
     + 1.5562E+05, 1.5575E+05, 1.5587E+05, 1.5599E+05, 1.5611E+05,
     +  1.5623E+05, 1.5635E+05, 1.5647E+05, 1.5659E+05, 1.5671E+05,
     + 1.5683E+05, 1.5695E+05, 1.5708E+05, 1.5720E+05, 1.5732E+05,
     +  1.5744E+05, 1.5756E+05, 1.5768E+05, 1.5780E+05, 1.5792E+05,
     + 1.5805E+05, 1.5817E+05, 1.5829E+05, 1.5841E+05, 1.5853E+05,
     +  1.5866E+05, 1.5878E+05, 1.5890E+05, 1.5902E+05, 1.5914E+05,
     + 1.5927E+05, 1.5939E+05, 1.5951E+05, 1.5963E+05, 1.5975E+05,
     +  1.5988E+05, 1.6000E+05, 1.6012E+05, 1.6024E+05, 1.6037E+05,
     + 1.6049E+05, 1.6061E+05, 1.6073E+05, 1.6086E+05, 1.6098E+05,
     +  1.6110E+05, 1.6123E+05, 1.6135E+05, 1.6147E+05, 1.6159E+05,
     + 1.6172E+05, 1.6184E+05, 1.6196E+05, 1.6209E+05, 1.6221E+05,
     +  1.6233E+05, 1.6246E+05, 1.6258E+05, 1.6270E+05, 1.6283E+05,
     + 1.6295E+05, 1.6307E+05, 1.6320E+05, 1.6332E+05, 1.6345E+05,
     +  1.6357E+05, 1.6369E+05, 1.6382E+05, 1.6394E+05, 1.6407E+05,
     + 1.6419E+05, 1.6431E+05, 1.6444E+05, 1.6456E+05, 1.6469E+05,
     +  1.6481E+05, 1.6494E+05, 1.6506E+05, 1.6518E+05, 1.6531E+05,
     + 1.6543E+05, 1.6556E+05, 1.6568E+05, 1.6581E+05, 1.6593E+05,
     +  1.6606E+05, 1.6618E+05, 1.6631E+05, 1.6643E+05, 1.6656E+05,
     + 1.6668E+05, 1.6681E+05, 1.6693E+05, 1.6706E+05, 1.6718E+05,
     +  1.6731E+05, 1.6743E+05, 1.6756E+05, 1.6769E+05, 1.6781E+05,
     + 1.6794E+05, 1.6806E+05, 1.6819E+05, 1.6831E+05, 1.6844E+05,
     +  1.6857E+05, 1.6869E+05, 1.6882E+05, 1.6894E+05, 1.6907E+05,
     + 1.6920E+05, 1.6932E+05, 1.6945E+05, 1.6958E+05, 1.6970E+05,
     +  1.6983E+05, 1.6995E+05, 1.7008E+05, 1.7021E+05, 1.7033E+05,
     + 1.7046E+05, 1.7059E+05, 1.7071E+05, 1.7084E+05, 1.7097E+05,
     +  1.7109E+05, 1.7122E+05, 1.7135E+05, 1.7148E+05, 1.7160E+05,
     + 1.7173E+05, 1.7186E+05, 1.7198E+05, 1.7211E+05, 1.7224E+05,
     +  1.7237E+05, 1.7249E+05, 1.7262E+05, 1.7275E+05, 1.7288E+05,
     + 1.7300E+05, 1.7313E+05, 1.7326E+05, 1.7339E+05, 1.7352E+05,
     +  1.7364E+05, 1.7377E+05, 1.7390E+05, 1.7403E+05, 1.7416E+05,
     + 1.7428E+05, 1.7441E+05, 1.7454E+05, 1.7467E+05, 1.7480E+05,
     +  1.7492E+05, 1.7505E+05, 1.7518E+05, 1.7531E+05, 1.7544E+05,
     + 1.7557E+05, 1.7570E+05, 1.7583E+05, 1.7595E+05, 1.7608E+05,
     +  1.7621E+05, 1.7634E+05, 1.7647E+05, 1.7660E+05, 1.7673E+05,
     + 1.7686E+05, 1.7699E+05, 1.7712E+05, 1.7724E+05, 1.7737E+05,
     +  1.7750E+05, 1.7763E+05, 1.7776E+05, 1.7789E+05, 1.7802E+05,
     + 1.7815E+05, 1.7828E+05, 1.7841E+05, 1.7854E+05, 1.7867E+05,
     +  1.7880E+05, 1.7893E+05, 1.7906E+05, 1.7919E+05, 1.7932E+05,
     + 1.7945E+05, 1.7958E+05, 1.7971E+05, 1.7984E+05, 1.7997E+05,
     +  1.8010E+05, 1.8023E+05, 1.8036E+05, 1.8049E+05, 1.8062E+05,
     + 1.8076E+05, 1.8089E+05, 1.8102E+05, 1.8115E+05, 1.8128E+05,
     +  1.8141E+05, 1.8154E+05, 1.8167E+05, 1.8180E+05, 1.8193E+05,
     + 1.8206E+05, 1.8220E+05, 1.8233E+05, 1.8246E+05, 1.8259E+05,
     +  1.8272E+05, 1.8285E+05, 1.8298E+05, 1.8312E+05, 1.8325E+05,
     + 1.8338E+05, 1.8351E+05, 1.8364E+05, 1.8377E+05, 1.8391E+05,
     +  1.8404E+05, 1.8417E+05, 1.8430E+05, 1.8443E+05, 1.8457E+05,
     + 1.8470E+05, 1.8483E+05, 1.8496E+05, 1.8510E+05, 1.8523E+05,
     +  1.8536E+05, 1.8549E+05, 1.8563E+05, 1.8576E+05, 1.8589E+05,
     + 1.8602E+05, 1.8616E+05, 1.8629E+05, 1.8642E+05, 1.8655E+05,
     +  1.8669E+05, 1.8682E+05, 1.8695E+05, 1.8709E+05, 1.8722E+05,
     + 1.8735E+05, 1.8749E+05, 1.8762E+05, 1.8775E+05, 1.8789E+05,
     +  1.8802E+05, 1.8815E+05, 1.8829E+05, 1.8842E+05, 1.8855E+05,
     + 1.8869E+05, 1.8882E+05, 1.8896E+05, 1.8909E+05, 1.8922E+05,
     +  1.8936E+05, 1.8949E+05, 1.8962E+05, 1.8976E+05, 1.8989E+05,
     + 1.9003E+05, 1.9016E+05, 1.9030E+05, 1.9043E+05, 1.9056E+05,
     +  1.9070E+05, 1.9083E+05, 1.9097E+05, 1.9110E+05, 1.9124E+05,
     + 1.9137E+05, 1.9151E+05, 1.9164E+05, 1.9178E+05, 1.9191E+05,
     +  1.9205E+05, 1.9218E+05, 1.9232E+05, 1.9245E+05, 1.9259E+05,
     + 1.9272E+05, 1.9286E+05, 1.9299E+05, 1.9313E+05, 1.9326E+05,
     +  1.9340E+05, 1.9353E+05, 1.9367E+05, 1.9380E+05, 1.9394E+05,
     + 1.9408E+05, 1.9421E+05, 1.9435E+05, 1.9448E+05, 1.9462E+05,
     +  1.9476E+05, 1.9489E+05, 1.9503E+05, 1.9516E+05, 1.9530E+05,
     + 1.9544E+05, 1.9557E+05, 1.9571E+05, 1.9584E+05, 1.9598E+05,
     +  1.9612E+05, 1.9625E+05, 1.9639E+05, 1.9653E+05, 1.9666E+05,
     + 1.9680E+05, 1.9694E+05, 1.9707E+05, 1.9721E+05, 1.9735E+05,
     +  1.9748E+05, 1.9762E+05, 1.9776E+05, 1.9790E+05, 1.9803E+05,
     + 1.9817E+05, 1.9831E+05, 1.9844E+05, 1.9858E+05, 1.9872E+05,
     +  1.9886E+05, 1.9899E+05, 1.9913E+05, 1.9927E+05, 1.9941E+05,
     + 1.9954E+05, 1.9968E+05, 1.9982E+05, 1.9996E+05, 2.0010E+05,
     +  2.0023E+05, 2.0037E+05, 2.0051E+05, 2.0065E+05, 2.0079E+05,
     + 2.0092E+05, 2.0106E+05, 2.0120E+05, 2.0134E+05, 2.0148E+05,
     +  2.0162E+05, 2.0175E+05, 2.0189E+05, 2.0203E+05, 2.0217E+05,
     + 2.0231E+05, 2.0245E+05, 2.0259E+05, 2.0272E+05, 2.0286E+05,
     +  2.0300E+05, 2.0314E+05, 2.0328E+05, 2.0342E+05, 2.0356E+05,
     + 2.0370E+05, 2.0384E+05, 2.0398E+05, 2.0411E+05, 2.0425E+05,
     +  2.0439E+05, 2.0453E+05, 2.0467E+05, 2.0481E+05, 2.0495E+05,
     + 2.0509E+05, 2.0523E+05, 2.0537E+05, 2.0551E+05, 2.0565E+05,
     +  2.0579E+05, 2.0593E+05, 2.0607E+05, 2.0621E+05, 2.0635E+05,
     + 2.0649E+05, 2.0663E+05, 2.0677E+05, 2.0691E+05, 2.0705E+05,
     +  2.0719E+05, 2.0733E+05, 2.0747E+05, 2.0761E+05, 2.0775E+05,
     + 2.0789E+05, 2.0804E+05, 2.0818E+05, 2.0832E+05, 2.0846E+05,
     +  2.0860E+05, 2.0874E+05, 2.0888E+05, 2.0902E+05, 2.0916E+05,
     + 2.0930E+05, 2.0945E+05, 2.0959E+05, 2.0973E+05, 2.0987E+05,
     +  2.1001E+05, 2.1015E+05, 2.1029E+05, 2.1043E+05, 2.1058E+05,
     + 2.1072E+05, 2.1086E+05, 2.1100E+05, 2.1114E+05, 2.1129E+05,
     +  2.1143E+05, 2.1157E+05, 2.1171E+05, 2.1185E+05, 2.1199E+05,
     + 2.1214E+05, 2.1228E+05, 2.1242E+05, 2.1256E+05, 2.1271E+05,
     +  2.1285E+05, 2.1299E+05, 2.1313E+05, 2.1328E+05, 2.1342E+05,
     + 2.1356E+05, 2.1370E+05, 2.1385E+05, 2.1399E+05, 2.1413E+05,
     +  2.1427E+05, 2.1442E+05, 2.1456E+05, 2.1470E+05, 2.1485E+05,
     + 2.1499E+05, 2.1513E+05, 2.1528E+05, 2.1542E+05, 2.1556E+05,
     +  2.1571E+05, 2.1585E+05, 2.1599E+05, 2.1614E+05, 2.1628E+05,
     + 2.1642E+05, 2.1657E+05, 2.1671E+05, 2.1685E+05, 2.1700E+05,
     +  2.1714E+05, 2.1729E+05, 2.1743E+05, 2.1757E+05, 2.1772E+05,
     + 2.1786E+05, 2.1801E+05, 2.1815E+05, 2.1829E+05, 2.1844E+05,
     +  2.1858E+05, 2.1873E+05, 2.1887E+05, 2.1902E+05, 2.1916E+05,
     + 2.1931E+05, 2.1945E+05, 2.1960E+05, 2.1974E+05, 2.1988E+05,
     +  2.2003E+05, 2.2017E+05, 2.2032E+05, 2.2046E+05, 2.2061E+05,
     + 2.2075E+05, 2.2090E+05, 2.2104E+05, 2.2119E+05, 2.2133E+05,
     +  2.2148E+05, 2.2163E+05, 2.2177E+05, 2.2192E+05, 2.2206E+05,
     + 2.2221E+05, 2.2235E+05, 2.2250E+05, 2.2264E+05, 2.2279E+05,
     +  2.2294E+05, 2.2308E+05, 2.2323E+05, 2.2337E+05, 2.2352E+05,
     + 2.2367E+05, 2.2381E+05, 2.2396E+05, 2.2410E+05, 2.2425E+05,
     +  2.2440E+05, 2.2454E+05, 2.2469E+05, 2.2484E+05, 2.2498E+05,
     + 2.2513E+05, 2.2528E+05, 2.2542E+05, 2.2557E+05, 2.2572E+05,
     +  2.2586E+05, 2.2601E+05, 2.2616E+05, 2.2630E+05, 2.2645E+05,
     + 2.2660E+05, 2.2674E+05, 2.2689E+05, 2.2704E+05, 2.2719E+05,
     +  2.2733E+05, 2.2748E+05, 2.2763E+05, 2.2777E+05, 2.2792E+05,
     + 2.2807E+05, 2.2822E+05, 2.2836E+05, 2.2851E+05, 2.2866E+05,
     +  2.2881E+05, 2.2896E+05, 2.2910E+05, 2.2925E+05, 2.2940E+05,
     + 2.2955E+05, 2.2970E+05, 2.2984E+05, 2.2999E+05, 2.3014E+05,
     +  2.3029E+05, 2.3044E+05, 2.3058E+05, 2.3073E+05, 2.3088E+05,
     + 2.3103E+05, 2.3118E+05, 2.3133E+05, 2.3148E+05, 2.3162E+05,
     +  2.3177E+05, 2.3192E+05, 2.3207E+05, 2.3222E+05, 2.3237E+05,
     + 2.3252E+05, 2.3267E+05, 2.3282E+05, 2.3296E+05, 2.3311E+05,
     +  2.3326E+05, 2.3341E+05, 2.3356E+05, 2.3371E+05, 2.3386E+05,
     + 2.3401E+05, 2.3416E+05, 2.3431E+05, 2.3446E+05, 2.3461E+05,
     +  2.3476E+05, 2.3491E+05, 2.3506E+05, 2.3521E+05, 2.3536E+05,
     + 2.3551E+05, 2.3566E+05, 2.3581E+05, 2.3596E+05, 2.3611E+05,
     +  2.3626E+05, 2.3641E+05, 2.3656E+05, 2.3671E+05, 2.3686E+05,
     + 2.3701E+05, 2.3716E+05, 2.3731E+05, 2.3746E+05, 2.3761E+05,
     +  2.3776E+05, 2.3791E+05, 2.3806E+05, 2.3821E+05, 2.3836E+05,
     + 2.3851E+05, 2.3867E+05, 2.3882E+05, 2.3897E+05, 2.3912E+05,
     +  2.3927E+05, 2.3942E+05, 2.3957E+05, 2.3972E+05, 2.3987E+05,
     + 2.4003E+05, 2.4018E+05, 2.4033E+05, 2.4048E+05, 2.4063E+05,
     +  2.4078E+05, 2.4094E+05, 2.4109E+05, 2.4124E+05, 2.4139E+05,
     + 2.4154E+05, 2.4169E+05, 2.4185E+05, 2.4200E+05, 2.4215E+05,
     +  2.4230E+05, 2.4245E+05, 2.4261E+05, 2.4276E+05, 2.4291E+05,
     + 2.4306E+05, 2.4322E+05, 2.4337E+05, 2.4352E+05, 2.4367E+05,
     +  2.4383E+05, 2.4398E+05, 2.4413E+05, 2.4428E+05, 2.4444E+05,
     + 2.4459E+05, 2.4474E+05, 2.4490E+05, 2.4505E+05, 2.4520E+05,
     +  2.4535E+05, 2.4551E+05, 2.4566E+05, 2.4581E+05, 2.4597E+05,
     + 2.4612E+05, 2.4627E+05, 2.4643E+05, 2.4658E+05, 2.4673E+05,
     +  2.4689E+05, 2.4704E+05, 2.4719E+05, 2.4735E+05, 2.4750E+05,
     + 2.4766E+05, 2.4781E+05, 2.4796E+05, 2.4812E+05, 2.4827E+05,
     +  2.4843E+05, 2.4858E+05, 2.4873E+05, 2.4889E+05, 2.4904E+05,
     + 2.4920E+05, 2.4935E+05, 2.4950E+05, 2.4966E+05, 2.4981E+05,
     +  2.4997E+05, 2.5012E+05, 2.5028E+05, 2.5043E+05, 2.5059E+05,
     + 2.5074E+05, 2.5090E+05, 2.5105E+05, 2.5121E+05, 2.5136E+05,
     +  2.5152E+05, 2.5167E+05, 2.5183E+05, 2.5198E+05, 2.5214E+05,
     + 2.5229E+05, 2.5245E+05, 2.5260E+05, 2.5276E+05, 2.5291E+05,
     +  2.5307E+05, 2.5322E+05, 2.5338E+05, 2.5353E+05, 2.5369E+05,
     + 2.5384E+05, 2.5400E+05, 2.5416E+05, 2.5431E+05, 2.5447E+05,
     +  2.5462E+05, 2.5478E+05, 2.5494E+05, 2.5509E+05, 2.5525E+05,
     + 2.5540E+05, 2.5556E+05, 2.5572E+05, 2.5587E+05, 2.5603E+05,
     +  2.5619E+05, 2.5634E+05, 2.5650E+05, 2.5665E+05, 2.5681E+05,
     + 2.5697E+05, 2.5712E+05, 2.5728E+05, 2.5744E+05, 2.5760E+05,
     +  2.5775E+05, 2.5791E+05, 2.5807E+05, 2.5822E+05, 2.5838E+05,
     + 2.5854E+05, 2.5869E+05, 2.5885E+05, 2.5901E+05, 2.5917E+05,
     +  2.5932E+05, 2.5948E+05, 2.5964E+05, 2.5980E+05, 2.5995E+05,
     + 2.6011E+05, 2.6027E+05, 2.6043E+05, 2.6058E+05, 2.6074E+05,
     +  2.6090E+05, 2.6106E+05, 2.6121E+05, 2.6137E+05, 2.6153E+05,
     + 2.6169E+05, 2.6185E+05, 2.6200E+05, 2.6216E+05, 2.6232E+05,
     +  2.6248E+05, 2.6264E+05, 2.6280E+05, 2.6295E+05, 2.6311E+05,
     + 2.6327E+05, 2.6343E+05, 2.6359E+05, 2.6375E+05, 2.6391E+05,
     +  2.6406E+05, 2.6422E+05, 2.6438E+05, 2.6454E+05, 2.6470E+05,
     + 2.6486E+05, 2.6502E+05, 2.6518E+05, 2.6534E+05, 2.6550E+05,
     +  2.6565E+05, 2.6581E+05, 2.6597E+05, 2.6613E+05, 2.6629E+05,
     + 2.6645E+05, 2.6661E+05, 2.6677E+05, 2.6693E+05, 2.6709E+05,
     +  2.6725E+05, 2.6741E+05, 2.6757E+05, 2.6773E+05, 2.6789E+05,
     + 2.6805E+05, 2.6821E+05, 2.6837E+05, 2.6853E+05, 2.6869E+05,
     +  2.6885E+05, 2.6901E+05, 2.6917E+05, 2.6933E+05, 2.6949E+05,
     + 2.6965E+05, 2.6981E+05, 2.6997E+05, 2.7013E+05, 2.7029E+05,
     +  2.7045E+05, 2.7061E+05, 2.7077E+05, 2.7093E+05, 2.7109E+05,
     + 2.7126E+05, 2.7142E+05, 2.7158E+05, 2.7174E+05, 2.7190E+05,
     +  2.7206E+05, 2.7222E+05, 2.7238E+05, 2.7254E+05, 2.7271E+05,
     + 2.7287E+05, 2.7303E+05, 2.7319E+05, 2.7335E+05, 2.7351E+05,
     +  2.7367E+05, 2.7384E+05, 2.7400E+05, 2.7416E+05, 2.7432E+05,
     + 2.7448E+05, 2.7464E+05, 2.7481E+05, 2.7497E+05, 2.7513E+05,
     +  2.7529E+05, 2.7545E+05, 2.7562E+05, 2.7578E+05, 2.7594E+05,
     + 2.7610E+05, 2.7626E+05, 2.7643E+05, 2.7659E+05, 2.7675E+05,
     +  2.7691E+05, 2.7708E+05, 2.7724E+05, 2.7740E+05, 2.7756E+05,
     + 2.7773E+05, 2.7789E+05, 2.7805E+05, 2.7822E+05, 2.7838E+05,
     +  2.7854E+05, 2.7870E+05, 2.7887E+05, 2.7903E+05, 2.7919E+05,
     + 2.7936E+05, 2.7952E+05, 2.7968E+05, 2.7985E+05, 2.8001E+05,
     +  2.8017E+05, 2.8034E+05, 2.8050E+05, 2.8066E+05, 2.8083E+05,
     + 2.8099E+05, 2.8115E+05, 2.8132E+05, 2.8148E+05, 2.8165E+05,
     +  2.8181E+05, 2.8197E+05, 2.8214E+05, 2.8230E+05, 2.8247E+05,
     + 2.8263E+05, 2.8279E+05, 2.8296E+05, 2.8312E+05, 2.8329E+05,
     +  2.8345E+05, 2.8361E+05, 2.8378E+05, 2.8394E+05, 2.8411E+05,
     + 2.8427E+05, 2.8444E+05, 2.8460E+05, 2.8477E+05, 2.8493E+05,
     +  2.8510E+05, 2.8526E+05, 2.8543E+05, 2.8559E+05, 2.8576E+05,
     + 2.8592E+05, 2.8609E+05, 2.8625E+05, 2.8642E+05, 2.8658E+05,
     +  2.8675E+05, 2.8691E+05, 2.8708E+05, 2.8724E+05, 2.8741E+05,
     + 2.8757E+05, 2.8774E+05, 2.8790E+05, 2.8807E+05, 2.8824E+05,
     +  2.8840E+05, 2.8857E+05, 2.8873E+05, 2.8890E+05, 2.8906E+05,
     + 2.8923E+05, 2.8940E+05, 2.8956E+05, 2.8973E+05, 2.8989E+05,
     +  2.9006E+05, 2.9023E+05, 2.9039E+05, 2.9056E+05, 2.9073E+05,
     + 2.9089E+05, 2.9106E+05, 2.9122E+05, 2.9139E+05, 2.9156E+05,
     +  2.9172E+05, 2.9189E+05, 2.9206E+05, 2.9222E+05, 2.9239E+05,
     + 2.9256E+05, 2.9272E+05, 2.9289E+05, 2.9306E+05, 2.9323E+05,
     +  2.9339E+05, 2.9356E+05, 2.9373E+05, 2.9389E+05, 2.9406E+05,
     + 2.9423E+05, 2.9440E+05, 2.9456E+05, 2.9473E+05, 2.9490E+05,
     +  2.9507E+05, 2.9523E+05, 2.9540E+05, 2.9557E+05, 2.9574E+05,
     + 2.9590E+05, 2.9607E+05, 2.9624E+05, 2.9641E+05, 2.9658E+05,
     +  2.9674E+05, 2.9691E+05, 2.9708E+05, 2.9725E+05, 2.9742E+05,
     + 2.9758E+05, 2.9775E+05, 2.9792E+05, 2.9809E+05, 2.9826E+05,
     +  2.9843E+05, 2.9859E+05, 2.9876E+05, 2.9893E+05, 2.9910E+05,
     + 2.9927E+05, 2.9944E+05, 2.9961E+05, 2.9977E+05, 2.9994E+05,
     +  3.0011E+05, 3.0028E+05, 3.0045E+05, 3.0062E+05, 3.0079E+05,
     + 3.0096E+05, 3.0113E+05, 3.0130E+05, 3.0146E+05, 3.0163E+05,
     +  3.0180E+05, 3.0197E+05, 3.0214E+05, 3.0231E+05, 3.0248E+05,
     + 3.0265E+05, 3.0282E+05, 3.0299E+05, 3.0316E+05, 3.0333E+05,
     +  3.0350E+05, 3.0367E+05, 3.0384E+05, 3.0401E+05, 3.0418E+05,
     + 3.0435E+05, 3.0452E+05, 3.0469E+05, 3.0486E+05, 3.0503E+05,
     +  3.0520E+05, 3.0537E+05, 3.0554E+05, 3.0571E+05, 3.0588E+05,
     + 3.0605E+05, 3.0622E+05, 3.0639E+05, 3.0656E+05, 3.0673E+05,
     +  3.0690E+05, 3.0707E+05, 3.0724E+05, 3.0742E+05, 3.0759E+05,
     + 3.0776E+05, 3.0793E+05, 3.0810E+05, 3.0827E+05, 3.0844E+05,
     +  3.0861E+05, 3.0878E+05, 3.0895E+05, 3.0913E+05, 3.0930E+05,
     + 3.0947E+05, 3.0964E+05, 3.0981E+05, 3.0998E+05, 3.1015E+05,
     +  3.1033E+05, 3.1050E+05, 3.1067E+05, 3.1084E+05, 3.1101E+05,
     + 3.1118E+05, 3.1136E+05, 3.1153E+05, 3.1170E+05, 3.1187E+05,
     +  3.1204E+05, 3.1222E+05, 3.1239E+05, 3.1256E+05, 3.1273E+05,
     + 3.1290E+05, 3.1308E+05, 3.1325E+05, 3.1342E+05, 3.1359E+05,
     +  3.1377E+05, 3.1394E+05, 3.1411E+05, 3.1428E+05, 3.1446E+05,
     + 3.1463E+05, 3.1480E+05, 3.1497E+05, 3.1515E+05, 3.1532E+05,
     +  3.1549E+05, 3.1567E+05, 3.1584E+05, 3.1601E+05, 3.1619E+05,
     + 3.1636E+05, 3.1653E+05, 3.1670E+05, 3.1688E+05, 3.1705E+05,
     +  3.1722E+05, 3.1740E+05, 3.1757E+05, 3.1774E+05, 3.1792E+05,
     + 3.1809E+05, 3.1827E+05, 3.1844E+05, 3.1861E+05, 3.1879E+05,
     +  3.1896E+05, 3.1913E+05, 3.1931E+05, 3.1948E+05, 3.1966E+05,
     + 3.1983E+05, 3.2000E+05, 3.2018E+05, 3.2035E+05, 3.2053E+05,
     +  3.2070E+05, 3.2087E+05, 3.2105E+05, 3.2122E+05, 3.2140E+05,
     + 3.2157E+05, 3.2175E+05, 3.2192E+05, 3.2210E+05, 3.2227E+05,
     +  3.2245E+05, 3.2262E+05, 3.2279E+05, 3.2297E+05, 3.2314E+05,
     + 3.2332E+05, 3.2349E+05, 3.2367E+05, 3.2384E+05, 3.2402E+05,
     +  3.2419E+05, 3.2437E+05, 3.2454E+05, 3.2472E+05, 3.2489E+05,
     + 3.2507E+05, 3.2525E+05, 3.2542E+05, 3.2560E+05, 3.2577E+05,
     +  3.2595E+05, 3.2612E+05, 3.2630E+05, 3.2647E+05, 3.2665E+05,
     + 3.2683E+05, 3.2700E+05, 3.2718E+05, 3.2735E+05, 3.2753E+05,
     +  3.2771E+05, 3.2788E+05, 3.2806E+05, 3.2823E+05, 3.2841E+05,
     + 3.2859E+05, 3.2876E+05, 3.2894E+05, 3.2911E+05, 3.2929E+05,
     +  3.2947E+05, 3.2964E+05, 3.2982E+05, 3.3000E+05, 3.3017E+05,
     + 3.3035E+05, 3.3053E+05, 3.3070E+05, 3.3088E+05, 3.3106E+05,
     +  3.3123E+05, 3.3141E+05, 3.3159E+05, 3.3176E+05, 3.3194E+05,
     + 3.3212E+05, 3.3230E+05, 3.3247E+05, 3.3265E+05, 3.3283E+05,
     +  3.3300E+05, 3.3318E+05, 3.3336E+05, 3.3354E+05, 3.3371E+05,
     + 3.3389E+05, 3.3407E+05, 3.3425E+05, 3.3442E+05, 3.3460E+05,
     +  3.3478E+05, 3.3496E+05, 3.3514E+05, 3.3531E+05, 3.3549E+05,
     + 3.3567E+05, 3.3585E+05, 3.3602E+05, 3.3620E+05, 3.3638E+05,
     +  3.3656E+05, 3.3674E+05, 3.3692E+05, 3.3709E+05, 3.3727E+05,
     + 3.3745E+05, 3.3763E+05, 3.3781E+05, 3.3799E+05, 3.3816E+05,
     +  3.3834E+05, 3.3852E+05, 3.3870E+05, 3.3888E+05, 3.3906E+05,
     + 3.3924E+05, 3.3942E+05, 3.3959E+05, 3.3977E+05, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        77
      data (QofT( 6,J),J=1,2408)/ 7.1974E+01, 8.7215E+01, 1.0687E+02,
     +  1.2945E+02, 1.5384E+02, 1.7927E+02, 2.0533E+02, 2.3177E+02,
     + 2.5847E+02, 2.8536E+02, 3.1238E+02, 3.3951E+02, 3.6673E+02,
     +  3.9401E+02, 4.2135E+02, 4.4874E+02, 4.7616E+02, 5.0362E+02,
     + 5.3111E+02, 5.5862E+02, 6.1370E+02, 6.6885E+02, 7.2404E+02,
     +  7.7928E+02, 8.3455E+02, 8.8985E+02, 9.4517E+02, 1.0005E+03,
     + 1.0559E+03, 1.1112E+03, 1.1666E+03, 1.2220E+03, 1.2774E+03,
     +  1.3328E+03, 1.3882E+03, 1.4437E+03, 1.4991E+03, 1.5545E+03,
     + 1.6100E+03, 1.6654E+03, 1.7209E+03, 1.7763E+03, 1.8318E+03,
     +  1.8872E+03, 1.9427E+03, 1.9982E+03, 2.0536E+03, 2.1091E+03,
     + 2.1646E+03, 2.2201E+03, 2.2755E+03, 2.3310E+03, 2.3865E+03,
     +  2.4420E+03, 2.4975E+03, 2.5530E+03, 2.6085E+03, 2.6639E+03,
     + 2.7194E+03, 2.7749E+03, 2.8304E+03, 2.8859E+03, 2.9414E+03,
     +  2.9969E+03, 3.0524E+03, 3.1079E+03, 3.1634E+03, 3.2190E+03,
     + 3.2745E+03, 3.3300E+03, 3.3855E+03, 3.4410E+03, 3.4965E+03,
     +  3.5520E+03, 3.6075E+03, 3.6631E+03, 3.7186E+03, 3.7741E+03,
     + 3.8296E+03, 3.8851E+03, 3.9406E+03, 3.9962E+03, 4.0517E+03,
     +  4.1072E+03, 4.1627E+03, 4.2183E+03, 4.2738E+03, 4.3293E+03,
     + 4.3849E+03, 4.4404E+03, 4.4959E+03, 4.5515E+03, 4.6070E+03,
     +  4.6625E+03, 4.7181E+03, 4.7736E+03, 4.8291E+03, 4.8847E+03,
     + 4.9402E+03, 4.9958E+03, 5.0513E+03, 5.1068E+03, 5.1624E+03,
     +  5.2179E+03, 5.2735E+03, 5.3290E+03, 5.3846E+03, 5.4402E+03,
     + 5.4957E+03, 5.5513E+03, 5.6068E+03, 5.6624E+03, 5.7180E+03,
     +  5.7735E+03, 5.8291E+03, 5.8847E+03, 5.9403E+03, 5.9958E+03,
     + 6.0514E+03, 6.1070E+03, 6.1626E+03, 6.2182E+03, 6.2738E+03,
     +  6.3294E+03, 6.3850E+03, 6.4406E+03, 6.4962E+03, 6.5518E+03,
     + 6.6074E+03, 6.6631E+03, 6.7187E+03, 6.7743E+03, 6.8300E+03,
     +  6.8856E+03, 6.9413E+03, 6.9969E+03, 7.0526E+03, 7.1083E+03,
     + 7.1640E+03, 7.2197E+03, 7.2754E+03, 7.3311E+03, 7.3868E+03,
     +  7.4425E+03, 7.4982E+03, 7.5540E+03, 7.6097E+03, 7.6655E+03,
     + 7.7213E+03, 7.7771E+03, 7.8329E+03, 7.8887E+03, 7.9445E+03,
     +  8.0003E+03, 8.0561E+03, 8.1120E+03, 8.1679E+03, 8.2238E+03,
     + 8.2796E+03, 8.3356E+03, 8.3915E+03, 8.4474E+03, 8.5034E+03,
     +  8.5594E+03, 8.6153E+03, 8.6713E+03, 8.7274E+03, 8.7834E+03,
     + 8.8395E+03, 8.8955E+03, 8.9516E+03, 9.0078E+03, 9.0639E+03,
     +  9.1201E+03, 9.1762E+03, 9.2324E+03, 9.2887E+03, 9.3449E+03,
     + 9.4012E+03, 9.4574E+03, 9.5138E+03, 9.5701E+03, 9.6265E+03,
     +  9.6828E+03, 9.7392E+03, 9.7957E+03, 9.8521E+03, 9.9086E+03,
     + 9.9652E+03, 1.0022E+04, 1.0078E+04, 1.0135E+04, 1.0192E+04,
     +  1.0248E+04, 1.0305E+04, 1.0362E+04, 1.0418E+04, 1.0475E+04,
     + 1.0532E+04, 1.0589E+04, 1.0646E+04, 1.0703E+04, 1.0760E+04,
     +  1.0816E+04, 1.0874E+04, 1.0931E+04, 1.0988E+04, 1.1045E+04,
     + 1.1102E+04, 1.1159E+04, 1.1216E+04, 1.1274E+04, 1.1331E+04,
     +  1.1388E+04, 1.1446E+04, 1.1503E+04, 1.1561E+04, 1.1618E+04,
     + 1.1676E+04, 1.1733E+04, 1.1791E+04, 1.1849E+04, 1.1906E+04,
     +  1.1964E+04, 1.2022E+04, 1.2080E+04, 1.2138E+04, 1.2196E+04,
     + 1.2254E+04, 1.2312E+04, 1.2370E+04, 1.2428E+04, 1.2486E+04,
     +  1.2544E+04, 1.2603E+04, 1.2661E+04, 1.2719E+04, 1.2778E+04,
     + 1.2836E+04, 1.2895E+04, 1.2953E+04, 1.3012E+04, 1.3071E+04,
     +  1.3129E+04, 1.3188E+04, 1.3247E+04, 1.3306E+04, 1.3365E+04,
     + 1.3424E+04, 1.3483E+04, 1.3542E+04, 1.3601E+04, 1.3660E+04,
     +  1.3720E+04, 1.3779E+04, 1.3838E+04, 1.3898E+04, 1.3957E+04,
     + 1.4017E+04, 1.4076E+04, 1.4136E+04, 1.4196E+04, 1.4256E+04,
     +  1.4316E+04, 1.4375E+04, 1.4435E+04, 1.4495E+04, 1.4556E+04,
     + 1.4616E+04, 1.4676E+04, 1.4736E+04, 1.4797E+04, 1.4857E+04,
     +  1.4918E+04, 1.4978E+04, 1.5039E+04, 1.5099E+04, 1.5160E+04,
     + 1.5221E+04, 1.5282E+04, 1.5343E+04, 1.5404E+04, 1.5465E+04,
     +  1.5526E+04, 1.5587E+04, 1.5648E+04, 1.5710E+04, 1.5771E+04,
     + 1.5833E+04, 1.5894E+04, 1.5956E+04, 1.6017E+04, 1.6079E+04,
     +  1.6141E+04, 1.6203E+04, 1.6265E+04, 1.6327E+04, 1.6389E+04,
     + 1.6451E+04, 1.6513E+04, 1.6576E+04, 1.6638E+04, 1.6701E+04,
     +  1.6763E+04, 1.6826E+04, 1.6888E+04, 1.6951E+04, 1.7014E+04,
     + 1.7077E+04, 1.7140E+04, 1.7203E+04, 1.7266E+04, 1.7329E+04,
     +  1.7392E+04, 1.7456E+04, 1.7519E+04, 1.7583E+04, 1.7646E+04,
     + 1.7710E+04, 1.7774E+04, 1.7837E+04, 1.7901E+04, 1.7965E+04,
     +  1.8029E+04, 1.8093E+04, 1.8158E+04, 1.8222E+04, 1.8286E+04,
     + 1.8351E+04, 1.8415E+04, 1.8480E+04, 1.8544E+04, 1.8609E+04,
     +  1.8674E+04, 1.8739E+04, 1.8804E+04, 1.8869E+04, 1.8934E+04,
     + 1.8999E+04, 1.9065E+04, 1.9130E+04, 1.9196E+04, 1.9261E+04,
     +  1.9327E+04, 1.9392E+04, 1.9458E+04, 1.9524E+04, 1.9590E+04,
     + 1.9656E+04, 1.9722E+04, 1.9788E+04, 1.9855E+04, 1.9921E+04,
     +  1.9988E+04, 2.0054E+04, 2.0121E+04, 2.0187E+04, 2.0254E+04,
     + 2.0321E+04, 2.0388E+04, 2.0455E+04, 2.0522E+04, 2.0590E+04,
     +  2.0657E+04, 2.0724E+04, 2.0792E+04, 2.0859E+04, 2.0927E+04,
     + 2.0995E+04, 2.1062E+04, 2.1130E+04, 2.1198E+04, 2.1266E+04,
     +  2.1335E+04, 2.1403E+04, 2.1471E+04, 2.1540E+04, 2.1608E+04,
     + 2.1677E+04, 2.1745E+04, 2.1814E+04, 2.1883E+04, 2.1952E+04,
     +  2.2021E+04, 2.2090E+04, 2.2160E+04, 2.2229E+04, 2.2298E+04,
     + 2.2368E+04, 2.2437E+04, 2.2507E+04, 2.2577E+04, 2.2647E+04,
     +  2.2717E+04, 2.2787E+04, 2.2857E+04, 2.2927E+04, 2.2997E+04,
     + 2.3068E+04, 2.3138E+04, 2.3209E+04, 2.3279E+04, 2.3350E+04,
     +  2.3421E+04, 2.3492E+04, 2.3563E+04, 2.3634E+04, 2.3705E+04,
     + 2.3776E+04, 2.3848E+04, 2.3919E+04, 2.3991E+04, 2.4063E+04,
     +  2.4134E+04, 2.4206E+04, 2.4278E+04, 2.4350E+04, 2.4422E+04,
     + 2.4495E+04, 2.4567E+04, 2.4639E+04, 2.4712E+04, 2.4784E+04,
     +  2.4857E+04, 2.4930E+04, 2.5003E+04, 2.5076E+04, 2.5149E+04,
     + 2.5222E+04, 2.5295E+04, 2.5368E+04, 2.5442E+04, 2.5515E+04,
     +  2.5589E+04, 2.5663E+04, 2.5737E+04, 2.5810E+04, 2.5884E+04,
     + 2.5959E+04, 2.6033E+04, 2.6107E+04, 2.6181E+04, 2.6256E+04,
     +  2.6331E+04, 2.6405E+04, 2.6480E+04, 2.6555E+04, 2.6630E+04,
     + 2.6705E+04, 2.6780E+04, 2.6855E+04, 2.6931E+04, 2.7006E+04,
     +  2.7081E+04, 2.7157E+04, 2.7233E+04, 2.7309E+04, 2.7385E+04,
     + 2.7461E+04, 2.7537E+04, 2.7613E+04, 2.7689E+04, 2.7766E+04,
     +  2.7842E+04, 2.7919E+04, 2.7995E+04, 2.8072E+04, 2.8149E+04,
     + 2.8226E+04, 2.8303E+04, 2.8380E+04, 2.8457E+04, 2.8535E+04,
     +  2.8612E+04, 2.8690E+04, 2.8768E+04, 2.8845E+04, 2.8923E+04,
     + 2.9001E+04, 2.9079E+04, 2.9157E+04, 2.9236E+04, 2.9314E+04,
     +  2.9392E+04, 2.9471E+04, 2.9549E+04, 2.9628E+04, 2.9707E+04,
     + 2.9786E+04, 2.9865E+04, 2.9944E+04, 3.0023E+04, 3.0103E+04,
     +  3.0182E+04, 3.0262E+04, 3.0341E+04, 3.0421E+04, 3.0501E+04,
     + 3.0581E+04, 3.0661E+04, 3.0741E+04, 3.0821E+04, 3.0901E+04,
     +  3.0982E+04, 3.1062E+04, 3.1143E+04, 3.1224E+04, 3.1305E+04,
     + 3.1385E+04, 3.1466E+04, 3.1548E+04, 3.1629E+04, 3.1710E+04,
     +  3.1792E+04, 3.1873E+04, 3.1955E+04, 3.2036E+04, 3.2118E+04,
     + 3.2200E+04, 3.2282E+04, 3.2364E+04, 3.2447E+04, 3.2529E+04,
     +  3.2611E+04, 3.2694E+04, 3.2776E+04, 3.2859E+04, 3.2942E+04,
     + 3.3025E+04, 3.3108E+04, 3.3191E+04, 3.3274E+04, 3.3358E+04,
     +  3.3441E+04, 3.3525E+04, 3.3608E+04, 3.3692E+04, 3.3776E+04,
     + 3.3860E+04, 3.3944E+04, 3.4028E+04, 3.4112E+04, 3.4196E+04,
     +  3.4281E+04, 3.4365E+04, 3.4450E+04, 3.4535E+04, 3.4620E+04,
     + 3.4705E+04, 3.4790E+04, 3.4875E+04, 3.4960E+04, 3.5045E+04,
     +  3.5131E+04, 3.5216E+04, 3.5302E+04, 3.5388E+04, 3.5474E+04,
     + 3.5560E+04, 3.5646E+04, 3.5732E+04, 3.5818E+04, 3.5905E+04,
     +  3.5991E+04, 3.6078E+04, 3.6164E+04, 3.6251E+04, 3.6338E+04,
     + 3.6425E+04, 3.6512E+04, 3.6599E+04, 3.6687E+04, 3.6774E+04,
     +  3.6862E+04, 3.6949E+04, 3.7037E+04, 3.7125E+04, 3.7213E+04,
     + 3.7301E+04, 3.7389E+04, 3.7477E+04, 3.7565E+04, 3.7654E+04,
     +  3.7742E+04, 3.7831E+04, 3.7920E+04, 3.8009E+04, 3.8098E+04,
     + 3.8187E+04, 3.8276E+04, 3.8365E+04, 3.8455E+04, 3.8544E+04,
     +  3.8634E+04, 3.8723E+04, 3.8813E+04, 3.8903E+04, 3.8993E+04,
     + 3.9083E+04, 3.9173E+04, 3.9264E+04, 3.9354E+04, 3.9444E+04,
     +  3.9535E+04, 3.9626E+04, 3.9717E+04, 3.9808E+04, 3.9899E+04,
     + 3.9990E+04, 4.0081E+04, 4.0172E+04, 4.0264E+04, 4.0355E+04,
     +  4.0447E+04, 4.0539E+04, 4.0631E+04, 4.0723E+04, 4.0815E+04,
     + 4.0907E+04, 4.0999E+04, 4.1092E+04, 4.1184E+04, 4.1277E+04,
     +  4.1369E+04, 4.1462E+04, 4.1555E+04, 4.1648E+04, 4.1741E+04,
     + 4.1835E+04, 4.1928E+04, 4.2021E+04, 4.2115E+04, 4.2209E+04,
     +  4.2302E+04, 4.2396E+04, 4.2490E+04, 4.2584E+04, 4.2678E+04,
     + 4.2773E+04, 4.2867E+04, 4.2962E+04, 4.3056E+04, 4.3151E+04,
     +  4.3246E+04, 4.3341E+04, 4.3436E+04, 4.3531E+04, 4.3626E+04,
     + 4.3721E+04, 4.3817E+04, 4.3912E+04, 4.4008E+04, 4.4104E+04,
     +  4.4200E+04, 4.4296E+04, 4.4392E+04, 4.4488E+04, 4.4584E+04,
     + 4.4681E+04, 4.4777E+04, 4.4874E+04, 4.4971E+04, 4.5067E+04,
     +  4.5164E+04, 4.5261E+04, 4.5359E+04, 4.5456E+04, 4.5553E+04,
     + 4.5651E+04, 4.5748E+04, 4.5846E+04, 4.5944E+04, 4.6042E+04,
     +  4.6140E+04, 4.6238E+04, 4.6336E+04, 4.6434E+04, 4.6533E+04,
     + 4.6631E+04, 4.6730E+04, 4.6829E+04, 4.6928E+04, 4.7026E+04,
     +  4.7126E+04, 4.7225E+04, 4.7324E+04, 4.7423E+04, 4.7523E+04,
     + 4.7623E+04, 4.7722E+04, 4.7822E+04, 4.7922E+04, 4.8022E+04,
     +  4.8122E+04, 4.8222E+04, 4.8323E+04, 4.8423E+04, 4.8524E+04,
     + 4.8625E+04, 4.8725E+04, 4.8826E+04, 4.8927E+04, 4.9028E+04,
     +  4.9130E+04, 4.9231E+04, 4.9332E+04, 4.9434E+04, 4.9536E+04,
     + 4.9637E+04, 4.9739E+04, 4.9841E+04, 4.9943E+04, 5.0045E+04,
     +  5.0148E+04, 5.0250E+04, 5.0353E+04, 5.0455E+04, 5.0558E+04,
     + 5.0661E+04, 5.0764E+04, 5.0867E+04, 5.0970E+04, 5.1073E+04,
     +  5.1177E+04, 5.1280E+04, 5.1384E+04, 5.1487E+04, 5.1591E+04,
     + 5.1695E+04, 5.1799E+04, 5.1903E+04, 5.2008E+04, 5.2112E+04,
     +  5.2216E+04, 5.2321E+04, 5.2426E+04, 5.2531E+04, 5.2635E+04,
     + 5.2740E+04, 5.2846E+04, 5.2951E+04, 5.3056E+04, 5.3162E+04,
     +  5.3267E+04, 5.3373E+04, 5.3479E+04, 5.3584E+04, 5.3690E+04,
     + 5.3797E+04, 5.3903E+04, 5.4009E+04, 5.4116E+04, 5.4222E+04,
     +  5.4329E+04, 5.4436E+04, 5.4542E+04, 5.4649E+04, 5.4756E+04,
     + 5.4864E+04, 5.4971E+04, 5.5078E+04, 5.5186E+04, 5.5294E+04,
     +  5.5401E+04, 5.5509E+04, 5.5617E+04, 5.5725E+04, 5.5833E+04,
     + 5.5942E+04, 5.6050E+04, 5.6159E+04, 5.6267E+04, 5.6376E+04,
     +  5.6485E+04, 5.6594E+04, 5.6703E+04, 5.6812E+04, 5.6921E+04,
     + 5.7031E+04, 5.7140E+04, 5.7250E+04, 5.7360E+04, 5.7469E+04,
     +  5.7579E+04, 5.7689E+04, 5.7800E+04, 5.7910E+04, 5.8020E+04,
     + 5.8131E+04, 5.8241E+04, 5.8352E+04, 5.8463E+04, 5.8574E+04,
     +  5.8685E+04, 5.8796E+04, 5.8907E+04, 5.9019E+04, 5.9130E+04,
     + 5.9242E+04, 5.9354E+04, 5.9465E+04, 5.9577E+04, 5.9689E+04,
     +  5.9802E+04, 5.9914E+04, 6.0026E+04, 6.0139E+04, 6.0251E+04,
     + 6.0364E+04, 6.0477E+04, 6.0590E+04, 6.0703E+04, 6.0816E+04,
     +  6.0929E+04, 6.1043E+04, 6.1156E+04, 6.1270E+04, 6.1384E+04,
     + 6.1497E+04, 6.1611E+04, 6.1725E+04, 6.1840E+04, 6.1954E+04,
     +  6.2068E+04, 6.2183E+04, 6.2297E+04, 6.2412E+04, 6.2527E+04,
     + 6.2642E+04, 6.2757E+04, 6.2872E+04, 6.2988E+04, 6.3103E+04,
     +  6.3218E+04, 6.3334E+04, 6.3450E+04, 6.3566E+04, 6.3682E+04,
     + 6.3798E+04, 6.3914E+04, 6.4030E+04, 6.4147E+04, 6.4263E+04,
     +  6.4380E+04, 6.4497E+04, 6.4613E+04, 6.4730E+04, 6.4847E+04,
     + 6.4965E+04, 6.5082E+04, 6.5199E+04, 6.5317E+04, 6.5435E+04,
     +  6.5552E+04, 6.5670E+04, 6.5788E+04, 6.5906E+04, 6.6025E+04,
     + 6.6143E+04, 6.6261E+04, 6.6380E+04, 6.6499E+04, 6.6617E+04,
     +  6.6736E+04, 6.6855E+04, 6.6974E+04, 6.7094E+04, 6.7213E+04,
     + 6.7332E+04, 6.7452E+04, 6.7572E+04, 6.7692E+04, 6.7811E+04,
     +  6.7931E+04, 6.8052E+04, 6.8172E+04, 6.8292E+04, 6.8413E+04,
     + 6.8533E+04, 6.8654E+04, 6.8775E+04, 6.8896E+04, 6.9017E+04,
     +  6.9138E+04, 6.9259E+04, 6.9381E+04, 6.9502E+04, 6.9624E+04,
     + 6.9746E+04, 6.9867E+04, 6.9989E+04, 7.0112E+04, 7.0234E+04,
     +  7.0356E+04, 7.0478E+04, 7.0601E+04, 7.0724E+04, 7.0846E+04,
     + 7.0969E+04, 7.1092E+04, 7.1215E+04, 7.1339E+04, 7.1462E+04,
     +  7.1585E+04, 7.1709E+04, 7.1833E+04, 7.1957E+04, 7.2080E+04,
     + 7.2204E+04, 7.2329E+04, 7.2453E+04, 7.2577E+04, 7.2702E+04,
     +  7.2826E+04, 7.2951E+04, 7.3076E+04, 7.3201E+04, 7.3326E+04,
     + 7.3451E+04, 7.3576E+04, 7.3702E+04, 7.3827E+04, 7.3953E+04,
     +  7.4079E+04, 7.4205E+04, 7.4331E+04, 7.4457E+04, 7.4583E+04,
     + 7.4709E+04, 7.4836E+04, 7.4962E+04, 7.5089E+04, 7.5216E+04,
     +  7.5343E+04, 7.5470E+04, 7.5597E+04, 7.5724E+04, 7.5852E+04,
     + 7.5979E+04, 7.6107E+04, 7.6234E+04, 7.6362E+04, 7.6490E+04,
     +  7.6618E+04, 7.6747E+04, 7.6875E+04, 7.7003E+04, 7.7132E+04,
     + 7.7260E+04, 7.7389E+04, 7.7518E+04, 7.7647E+04, 7.7776E+04,
     +  7.7906E+04, 7.8035E+04, 7.8164E+04, 7.8294E+04, 7.8424E+04,
     + 7.8553E+04, 7.8683E+04, 7.8813E+04, 7.8944E+04, 7.9074E+04,
     +  7.9204E+04, 7.9335E+04, 7.9466E+04, 7.9596E+04, 7.9727E+04,
     + 7.9858E+04, 7.9989E+04, 8.0121E+04, 8.0252E+04, 8.0383E+04,
     +  8.0515E+04, 8.0647E+04, 8.0778E+04, 8.0910E+04, 8.1042E+04,
     + 8.1175E+04, 8.1307E+04, 8.1439E+04, 8.1572E+04, 8.1704E+04,
     +  8.1837E+04, 8.1970E+04, 8.2103E+04, 8.2236E+04, 8.2369E+04,
     + 8.2503E+04, 8.2636E+04, 8.2770E+04, 8.2904E+04, 8.3037E+04,
     +  8.3171E+04, 8.3305E+04, 8.3440E+04, 8.3574E+04, 8.3708E+04,
     + 8.3843E+04, 8.3977E+04, 8.4112E+04, 8.4247E+04, 8.4382E+04,
     +  8.4517E+04, 8.4652E+04, 8.4788E+04, 8.4923E+04, 8.5059E+04,
     + 8.5195E+04, 8.5330E+04, 8.5466E+04, 8.5602E+04, 8.5739E+04,
     +  8.5875E+04, 8.6011E+04, 8.6148E+04, 8.6285E+04, 8.6421E+04,
     + 8.6558E+04, 8.6695E+04, 8.6832E+04, 8.6970E+04, 8.7107E+04,
     +  8.7245E+04, 8.7382E+04, 8.7520E+04, 8.7658E+04, 8.7796E+04,
     + 8.7934E+04, 8.8072E+04, 8.8210E+04, 8.8349E+04, 8.8487E+04,
     +  8.8626E+04, 8.8765E+04, 8.8904E+04, 8.9043E+04, 8.9182E+04,
     + 8.9321E+04, 8.9461E+04, 8.9600E+04, 8.9740E+04, 8.9880E+04,
     +  9.0020E+04, 9.0160E+04, 9.0300E+04, 9.0440E+04, 9.0580E+04,
     + 9.0721E+04, 9.0862E+04, 9.1002E+04, 9.1143E+04, 9.1284E+04,
     +  9.1425E+04, 9.1566E+04, 9.1708E+04, 9.1849E+04, 9.1991E+04,
     + 9.2133E+04, 9.2274E+04, 9.2416E+04, 9.2558E+04, 9.2701E+04,
     +  9.2843E+04, 9.2985E+04, 9.3128E+04, 9.3270E+04, 9.3413E+04,
     + 9.3556E+04, 9.3699E+04, 9.3842E+04, 9.3986E+04, 9.4129E+04,
     +  9.4273E+04, 9.4416E+04, 9.4560E+04, 9.4704E+04, 9.4848E+04,
     + 9.4992E+04, 9.5136E+04, 9.5281E+04, 9.5425E+04, 9.5570E+04,
     +  9.5714E+04, 9.5859E+04, 9.6004E+04, 9.6149E+04, 9.6295E+04,
     + 9.6440E+04, 9.6585E+04, 9.6731E+04, 9.6877E+04, 9.7022E+04,
     +  9.7168E+04, 9.7314E+04, 9.7461E+04, 9.7607E+04, 9.7753E+04,
     + 9.7900E+04, 9.8047E+04, 9.8193E+04, 9.8340E+04, 9.8487E+04,
     +  9.8635E+04, 9.8782E+04, 9.8929E+04, 9.9077E+04, 9.9224E+04,
     + 9.9372E+04, 9.9520E+04, 9.9668E+04, 9.9816E+04, 9.9965E+04,
     +  1.0011E+05, 1.0026E+05, 1.0041E+05, 1.0056E+05, 1.0071E+05,
     + 1.0086E+05, 1.0101E+05, 1.0116E+05, 1.0130E+05, 1.0145E+05,
     +  1.0160E+05, 1.0175E+05, 1.0190E+05, 1.0205E+05, 1.0220E+05,
     + 1.0235E+05, 1.0250E+05, 1.0265E+05, 1.0281E+05, 1.0296E+05,
     +  1.0311E+05, 1.0326E+05, 1.0341E+05, 1.0356E+05, 1.0371E+05,
     + 1.0386E+05, 1.0401E+05, 1.0417E+05, 1.0432E+05, 1.0447E+05,
     +  1.0462E+05, 1.0477E+05, 1.0493E+05, 1.0508E+05, 1.0523E+05,
     + 1.0538E+05, 1.0554E+05, 1.0569E+05, 1.0584E+05, 1.0600E+05,
     +  1.0615E+05, 1.0630E+05, 1.0646E+05, 1.0661E+05, 1.0676E+05,
     + 1.0692E+05, 1.0707E+05, 1.0723E+05, 1.0738E+05, 1.0754E+05,
     +  1.0769E+05, 1.0785E+05, 1.0800E+05, 1.0815E+05, 1.0831E+05,
     + 1.0847E+05, 1.0862E+05, 1.0878E+05, 1.0893E+05, 1.0909E+05,
     +  1.0924E+05, 1.0940E+05, 1.0956E+05, 1.0971E+05, 1.0987E+05,
     + 1.1002E+05, 1.1018E+05, 1.1034E+05, 1.1049E+05, 1.1065E+05,
     +  1.1081E+05, 1.1097E+05, 1.1112E+05, 1.1128E+05, 1.1144E+05,
     + 1.1160E+05, 1.1175E+05, 1.1191E+05, 1.1207E+05, 1.1223E+05,
     +  1.1239E+05, 1.1254E+05, 1.1270E+05, 1.1286E+05, 1.1302E+05,
     + 1.1318E+05, 1.1334E+05, 1.1350E+05, 1.1366E+05, 1.1382E+05,
     +  1.1398E+05, 1.1414E+05, 1.1430E+05, 1.1446E+05, 1.1462E+05,
     + 1.1478E+05, 1.1494E+05, 1.1510E+05, 1.1526E+05, 1.1542E+05,
     +  1.1558E+05, 1.1574E+05, 1.1590E+05, 1.1606E+05, 1.1622E+05,
     + 1.1639E+05, 1.1655E+05, 1.1671E+05, 1.1687E+05, 1.1703E+05,
     +  1.1720E+05, 1.1736E+05, 1.1752E+05, 1.1768E+05, 1.1785E+05,
     + 1.1801E+05, 1.1817E+05, 1.1833E+05, 1.1850E+05, 1.1866E+05,
     +  1.1882E+05, 1.1899E+05, 1.1915E+05, 1.1932E+05, 1.1948E+05,
     + 1.1964E+05, 1.1981E+05, 1.1997E+05, 1.2014E+05, 1.2030E+05,
     +  1.2047E+05, 1.2063E+05, 1.2080E+05, 1.2096E+05, 1.2113E+05,
     + 1.2129E+05, 1.2146E+05, 1.2162E+05, 1.2179E+05, 1.2195E+05,
     +  1.2212E+05, 1.2229E+05, 1.2245E+05, 1.2262E+05, 1.2278E+05,
     + 1.2295E+05, 1.2312E+05, 1.2328E+05, 1.2345E+05, 1.2362E+05,
     +  1.2379E+05, 1.2395E+05, 1.2412E+05, 1.2429E+05, 1.2446E+05,
     + 1.2462E+05, 1.2479E+05, 1.2496E+05, 1.2513E+05, 1.2530E+05,
     +  1.2547E+05, 1.2563E+05, 1.2580E+05, 1.2597E+05, 1.2614E+05,
     + 1.2631E+05, 1.2648E+05, 1.2665E+05, 1.2682E+05, 1.2699E+05,
     +  1.2716E+05, 1.2733E+05, 1.2750E+05, 1.2767E+05, 1.2784E+05,
     + 1.2801E+05, 1.2818E+05, 1.2835E+05, 1.2852E+05, 1.2869E+05,
     +  1.2886E+05, 1.2904E+05, 1.2921E+05, 1.2938E+05, 1.2955E+05,
     + 1.2972E+05, 1.2989E+05, 1.3007E+05, 1.3024E+05, 1.3041E+05,
     +  1.3058E+05, 1.3076E+05, 1.3093E+05, 1.3110E+05, 1.3127E+05,
     + 1.3145E+05, 1.3162E+05, 1.3179E+05, 1.3197E+05, 1.3214E+05,
     +  1.3231E+05, 1.3249E+05, 1.3266E+05, 1.3284E+05, 1.3301E+05,
     + 1.3318E+05, 1.3336E+05, 1.3353E+05, 1.3371E+05, 1.3388E+05,
     +  1.3406E+05, 1.3423E+05, 1.3441E+05, 1.3458E+05, 1.3476E+05,
     + 1.3494E+05, 1.3511E+05, 1.3529E+05, 1.3546E+05, 1.3564E+05,
     +  1.3582E+05, 1.3599E+05, 1.3617E+05, 1.3635E+05, 1.3652E+05,
     + 1.3670E+05, 1.3688E+05, 1.3705E+05, 1.3723E+05, 1.3741E+05,
     +  1.3759E+05, 1.3776E+05, 1.3794E+05, 1.3812E+05, 1.3830E+05,
     + 1.3848E+05, 1.3866E+05, 1.3883E+05, 1.3901E+05, 1.3919E+05,
     +  1.3937E+05, 1.3955E+05, 1.3973E+05, 1.3991E+05, 1.4009E+05,
     + 1.4027E+05, 1.4045E+05, 1.4063E+05, 1.4081E+05, 1.4099E+05,
     +  1.4117E+05, 1.4135E+05, 1.4153E+05, 1.4171E+05, 1.4189E+05,
     + 1.4207E+05, 1.4225E+05, 1.4243E+05, 1.4262E+05, 1.4280E+05,
     +  1.4298E+05, 1.4316E+05, 1.4334E+05, 1.4352E+05, 1.4371E+05,
     + 1.4389E+05, 1.4407E+05, 1.4425E+05, 1.4444E+05, 1.4462E+05,
     +  1.4480E+05, 1.4499E+05, 1.4517E+05, 1.4535E+05, 1.4554E+05,
     + 1.4572E+05, 1.4590E+05, 1.4609E+05, 1.4627E+05, 1.4645E+05,
     +  1.4664E+05, 1.4682E+05, 1.4701E+05, 1.4719E+05, 1.4738E+05,
     + 1.4756E+05, 1.4775E+05, 1.4793E+05, 1.4812E+05, 1.4830E+05,
     +  1.4849E+05, 1.4868E+05, 1.4886E+05, 1.4905E+05, 1.4923E+05,
     + 1.4942E+05, 1.4961E+05, 1.4979E+05, 1.4998E+05, 1.5017E+05,
     +  1.5035E+05, 1.5054E+05, 1.5073E+05, 1.5091E+05, 1.5110E+05,
     + 1.5129E+05, 1.5148E+05, 1.5167E+05, 1.5185E+05, 1.5204E+05,
     +  1.5223E+05, 1.5242E+05, 1.5261E+05, 1.5280E+05, 1.5298E+05,
     + 1.5317E+05, 1.5336E+05, 1.5355E+05, 1.5374E+05, 1.5393E+05,
     +  1.5412E+05, 1.5431E+05, 1.5450E+05, 1.5469E+05, 1.5488E+05,
     + 1.5507E+05, 1.5526E+05, 1.5545E+05, 1.5564E+05, 1.5583E+05,
     +  1.5602E+05, 1.5622E+05, 1.5641E+05, 1.5660E+05, 1.5679E+05,
     + 1.5698E+05, 1.5717E+05, 1.5737E+05, 1.5756E+05, 1.5775E+05,
     +  1.5794E+05, 1.5813E+05, 1.5833E+05, 1.5852E+05, 1.5871E+05,
     + 1.5891E+05, 1.5910E+05, 1.5929E+05, 1.5949E+05, 1.5968E+05,
     +  1.5987E+05, 1.6007E+05, 1.6026E+05, 1.6045E+05, 1.6065E+05,
     + 1.6084E+05, 1.6104E+05, 1.6123E+05, 1.6143E+05, 1.6162E+05,
     +  1.6182E+05, 1.6201E+05, 1.6221E+05, 1.6240E+05, 1.6260E+05,
     + 1.6279E+05, 1.6299E+05, 1.6319E+05, 1.6338E+05, 1.6358E+05,
     +  1.6378E+05, 1.6397E+05, 1.6417E+05, 1.6437E+05, 1.6456E+05,
     + 1.6476E+05, 1.6496E+05, 1.6515E+05, 1.6535E+05, 1.6555E+05,
     +  1.6575E+05, 1.6594E+05, 1.6614E+05, 1.6634E+05, 1.6654E+05,
     + 1.6674E+05, 1.6694E+05, 1.6714E+05, 1.6733E+05, 1.6753E+05,
     +  1.6773E+05, 1.6793E+05, 1.6813E+05, 1.6833E+05, 1.6853E+05,
     + 1.6873E+05, 1.6893E+05, 1.6913E+05, 1.6933E+05, 1.6953E+05,
     +  1.6973E+05, 1.6993E+05, 1.7013E+05, 1.7033E+05, 1.7054E+05,
     + 1.7074E+05, 1.7094E+05, 1.7114E+05, 1.7134E+05, 1.7154E+05,
     +  1.7174E+05, 1.7195E+05, 1.7215E+05, 1.7235E+05, 1.7255E+05,
     + 1.7276E+05, 1.7296E+05, 1.7316E+05, 1.7336E+05, 1.7357E+05,
     +  1.7377E+05, 1.7397E+05, 1.7418E+05, 1.7438E+05, 1.7459E+05,
     + 1.7479E+05, 1.7499E+05, 1.7520E+05, 1.7540E+05, 1.7561E+05,
     +  1.7581E+05, 1.7602E+05, 1.7622E+05, 1.7643E+05, 1.7663E+05,
     + 1.7684E+05, 1.7704E+05, 1.7725E+05, 1.7745E+05, 1.7766E+05,
     +  1.7787E+05, 1.7807E+05, 1.7828E+05, 1.7848E+05, 1.7869E+05,
     + 1.7890E+05, 1.7910E+05, 1.7931E+05, 1.7952E+05, 1.7973E+05,
     +  1.7993E+05, 1.8014E+05, 1.8035E+05, 1.8056E+05, 1.8076E+05,
     + 1.8097E+05, 1.8118E+05, 1.8139E+05, 1.8160E+05, 1.8181E+05,
     +  1.8202E+05, 1.8222E+05, 1.8243E+05, 1.8264E+05, 1.8285E+05,
     + 1.8306E+05, 1.8327E+05, 1.8348E+05, 1.8369E+05, 1.8390E+05,
     +  1.8411E+05, 1.8432E+05, 1.8453E+05, 1.8474E+05, 1.8495E+05,
     + 1.8516E+05, 1.8538E+05, 1.8559E+05, 1.8580E+05, 1.8601E+05,
     +  1.8622E+05, 1.8643E+05, 1.8664E+05, 1.8686E+05, 1.8707E+05,
     + 1.8728E+05, 1.8749E+05, 1.8771E+05, 1.8792E+05, 1.8813E+05,
     +  1.8834E+05, 1.8856E+05, 1.8877E+05, 1.8898E+05, 1.8920E+05,
     + 1.8941E+05, 1.8963E+05, 1.8984E+05, 1.9005E+05, 1.9027E+05,
     +  1.9048E+05, 1.9070E+05, 1.9091E+05, 1.9113E+05, 1.9134E+05,
     + 1.9156E+05, 1.9177E+05, 1.9199E+05, 1.9220E+05, 1.9242E+05,
     +  1.9263E+05, 1.9285E+05, 1.9307E+05, 1.9328E+05, 1.9350E+05,
     + 1.9371E+05, 1.9393E+05, 1.9415E+05, 1.9436E+05, 1.9458E+05,
     +  1.9480E+05, 1.9502E+05, 1.9523E+05, 1.9545E+05, 1.9567E+05,
     + 1.9589E+05, 1.9611E+05, 1.9632E+05, 1.9654E+05, 1.9676E+05,
     +  1.9698E+05, 1.9720E+05, 1.9742E+05, 1.9764E+05, 1.9785E+05,
     + 1.9807E+05, 1.9829E+05, 1.9851E+05, 1.9873E+05, 1.9895E+05,
     +  1.9917E+05, 1.9939E+05, 1.9961E+05, 1.9983E+05, 2.0005E+05,
     + 2.0027E+05, 2.0050E+05, 2.0072E+05, 2.0094E+05, 2.0116E+05,
     +  2.0138E+05, 2.0160E+05, 2.0182E+05, 2.0205E+05, 2.0227E+05,
     + 2.0249E+05, 2.0271E+05, 2.0293E+05, 2.0316E+05, 2.0338E+05,
     +  2.0360E+05, 2.0383E+05, 2.0405E+05, 2.0427E+05, 2.0450E+05,
     + 2.0472E+05, 2.0494E+05, 2.0517E+05, 2.0539E+05, 2.0561E+05,
     +  2.0584E+05, 2.0606E+05, 2.0629E+05, 2.0651E+05, 2.0674E+05,
     + 2.0696E+05, 2.0719E+05, 2.0741E+05, 2.0764E+05, 2.0786E+05,
     +  2.0809E+05, 2.0831E+05, 2.0854E+05, 2.0877E+05, 2.0899E+05,
     + 2.0922E+05, 2.0945E+05, 2.0967E+05, 2.0990E+05, 2.1013E+05,
     +  2.1035E+05, 2.1058E+05, 2.1081E+05, 2.1104E+05, 2.1126E+05,
     + 2.1149E+05, 2.1172E+05, 2.1195E+05, 2.1218E+05, 2.1240E+05,
     +  2.1263E+05, 2.1286E+05, 2.1309E+05, 2.1332E+05, 2.1355E+05,
     + 2.1378E+05, 2.1401E+05, 2.1424E+05, 2.1447E+05, 2.1470E+05,
     +  2.1493E+05, 2.1516E+05, 2.1539E+05, 2.1562E+05, 2.1585E+05,
     + 2.1608E+05, 2.1631E+05, 2.1654E+05, 2.1677E+05, 2.1700E+05,
     +  2.1723E+05, 2.1746E+05, 2.1770E+05, 2.1793E+05, 2.1816E+05,
     + 2.1839E+05, 2.1862E+05, 2.1886E+05, 2.1909E+05, 2.1932E+05,
     +  2.1955E+05, 2.1979E+05, 2.2002E+05, 2.2025E+05, 2.2049E+05,
     + 2.2072E+05, 2.2095E+05, 2.2119E+05, 2.2142E+05, 2.2166E+05,
     +  2.2189E+05, 2.2212E+05, 2.2236E+05, 2.2259E+05, 2.2283E+05,
     + 2.2306E+05, 2.2330E+05, 2.2353E+05, 2.2377E+05, 2.2401E+05,
     +  2.2424E+05, 2.2448E+05, 2.2471E+05, 2.2495E+05, 2.2518E+05,
     + 2.2542E+05, 2.2566E+05, 2.2589E+05, 2.2613E+05, 2.2637E+05,
     +  2.2661E+05, 2.2684E+05, 2.2708E+05, 2.2732E+05, 2.2756E+05,
     + 2.2779E+05, 2.2803E+05, 2.2827E+05, 2.2851E+05, 2.2875E+05,
     +  2.2898E+05, 2.2922E+05, 2.2946E+05, 2.2970E+05, 2.2994E+05,
     + 2.3018E+05, 2.3042E+05, 2.3066E+05, 2.3090E+05, 2.3114E+05,
     +  2.3138E+05, 2.3162E+05, 2.3186E+05, 2.3210E+05, 2.3234E+05,
     + 2.3258E+05, 2.3282E+05, 2.3306E+05, 2.3330E+05, 2.3354E+05,
     +  2.3379E+05, 2.3403E+05, 2.3427E+05, 2.3451E+05, 2.3475E+05,
     + 2.3500E+05, 2.3524E+05, 2.3548E+05, 2.3572E+05, 2.3597E+05,
     +  2.3621E+05, 2.3645E+05, 2.3669E+05, 2.3694E+05, 2.3718E+05,
     + 2.3742E+05, 2.3767E+05, 2.3791E+05, 2.3816E+05, 2.3840E+05,
     +  2.3865E+05, 2.3889E+05, 2.3913E+05, 2.3938E+05, 2.3962E+05,
     + 2.3987E+05, 2.4011E+05, 2.4036E+05, 2.4061E+05, 2.4085E+05,
     +  2.4110E+05, 2.4134E+05, 2.4159E+05, 2.4183E+05, 2.4208E+05,
     + 2.4233E+05, 2.4257E+05, 2.4282E+05, 2.4307E+05, 2.4332E+05,
     +  2.4356E+05, 2.4381E+05, 2.4406E+05, 2.4430E+05, 2.4455E+05,
     + 2.4480E+05, 2.4505E+05, 2.4530E+05, 2.4555E+05, 2.4579E+05,
     +  2.4604E+05, 2.4629E+05, 2.4654E+05, 2.4679E+05, 2.4704E+05,
     + 2.4729E+05, 2.4754E+05, 2.4779E+05, 2.4804E+05, 2.4829E+05,
     +  2.4854E+05, 2.4879E+05, 2.4904E+05, 2.4929E+05, 2.4954E+05,
     + 2.4979E+05, 2.5004E+05, 2.5029E+05, 2.5054E+05, 2.5080E+05,
     +  2.5105E+05, 2.5130E+05, 2.5155E+05, 2.5180E+05, 2.5205E+05,
     + 2.5231E+05, 2.5256E+05, 2.5281E+05, 2.5307E+05, 2.5332E+05,
     +  2.5357E+05, 2.5382E+05, 2.5408E+05, 2.5433E+05, 2.5458E+05,
     + 2.5484E+05, 2.5509E+05, 2.5535E+05, 2.5560E+05, 2.5586E+05,
     +  2.5611E+05, 2.5636E+05, 2.5662E+05, 2.5687E+05, 2.5713E+05,
     + 2.5738E+05, 2.5764E+05, 2.5790E+05, 2.5815E+05, 2.5841E+05,
     +  2.5866E+05, 2.5892E+05, 2.5918E+05, 2.5943E+05, 2.5969E+05,
     + 2.5995E+05, 2.6020E+05, 2.6046E+05, 2.6072E+05, 2.6097E+05,
     +  2.6123E+05, 2.6149E+05, 2.6175E+05, 2.6200E+05, 2.6226E+05,
     + 2.6252E+05, 2.6278E+05, 2.6304E+05, 2.6330E+05, 2.6355E+05,
     +  2.6381E+05, 2.6407E+05, 2.6433E+05, 2.6459E+05, 2.6485E+05,
     + 2.6511E+05, 2.6537E+05, 2.6563E+05, 2.6589E+05, 2.6615E+05,
     +  2.6641E+05, 2.6667E+05, 2.6693E+05, 2.6719E+05, 2.6745E+05,
     + 2.6771E+05, 2.6798E+05, 2.6824E+05, 2.6850E+05, 2.6876E+05,
     +  2.6902E+05, 2.6928E+05, 2.6955E+05, 2.6981E+05, 2.7007E+05,
     + 2.7033E+05, 2.7060E+05, 2.7086E+05, 2.7112E+05, 2.7139E+05,
     +  2.7165E+05, 2.7191E+05, 2.7218E+05, 2.7244E+05, 2.7270E+05,
     + 2.7297E+05, 2.7323E+05, 2.7350E+05, 2.7376E+05, 2.7403E+05,
     +  2.7429E+05, 2.7456E+05, 2.7482E+05, 2.7509E+05, 2.7535E+05,
     + 2.7562E+05, 2.7588E+05, 2.7615E+05, 2.7641E+05, 2.7668E+05,
     +  2.7695E+05, 2.7721E+05, 2.7748E+05, 2.7775E+05, 2.7801E+05,
     + 2.7828E+05, 2.7855E+05, 2.7881E+05, 2.7908E+05, 2.7935E+05,
     +  2.7962E+05, 2.7989E+05, 2.8015E+05, 2.8042E+05, 2.8069E+05,
     + 2.8096E+05, 2.8123E+05, 2.8150E+05, 2.8176E+05, 2.8203E+05,
     +  2.8230E+05, 2.8257E+05, 2.8284E+05, 2.8311E+05, 2.8338E+05,
     + 2.8365E+05, 2.8392E+05, 2.8419E+05, 2.8446E+05, 2.8473E+05,
     +  2.8500E+05, 2.8527E+05, 2.8554E+05, 2.8582E+05, 2.8609E+05,
     + 2.8636E+05, 2.8663E+05, 2.8690E+05, 2.8717E+05, 2.8745E+05,
     +  2.8772E+05, 2.8799E+05, 2.8826E+05, 2.8854E+05, 2.8881E+05,
     + 2.8908E+05, 2.8935E+05, 2.8963E+05, 2.8990E+05, 2.9017E+05,
     +  2.9045E+05, 2.9072E+05, 2.9100E+05, 2.9127E+05, 2.9154E+05,
     + 2.9182E+05, 2.9209E+05, 2.9237E+05, 2.9264E+05, 2.9292E+05,
     +  2.9319E+05, 2.9347E+05, 2.9374E+05, 2.9402E+05, 2.9429E+05,
     + 2.9457E+05, 2.9485E+05, 2.9512E+05, 2.9540E+05, 2.9568E+05,
     +  2.9595E+05, 2.9623E+05, 2.9651E+05, 2.9678E+05, 2.9706E+05,
     + 2.9734E+05, 2.9761E+05, 2.9789E+05, 2.9817E+05, 2.9845E+05,
     +  2.9873E+05, 2.9900E+05, 2.9928E+05, 2.9956E+05, 2.9984E+05,
     + 3.0012E+05, 3.0040E+05, 3.0068E+05, 3.0096E+05, 3.0124E+05,
     +  3.0152E+05, 3.0180E+05, 3.0208E+05, 3.0236E+05, 3.0264E+05,
     + 3.0292E+05, 3.0320E+05, 3.0348E+05, 3.0376E+05, 3.0404E+05,
     +  3.0432E+05, 3.0460E+05, 3.0488E+05, 3.0516E+05, 3.0545E+05,
     + 3.0573E+05, 3.0601E+05, 3.0629E+05, 3.0657E+05, 3.0686E+05,
     +  3.0714E+05, 3.0742E+05, 3.0770E+05, 3.0799E+05, 3.0827E+05,
     + 3.0855E+05, 3.0884E+05, 3.0912E+05, 3.0940E+05, 3.0969E+05,
     +  3.0997E+05, 3.1026E+05, 3.1054E+05, 3.1082E+05, 3.1111E+05,
     + 3.1139E+05, 3.1168E+05, 3.1196E+05, 3.1225E+05, 3.1253E+05,
     +  3.1282E+05, 3.1311E+05, 3.1339E+05, 3.1368E+05, 3.1396E+05,
     + 3.1425E+05, 3.1454E+05, 3.1482E+05, 3.1511E+05, 3.1540E+05,
     +  3.1568E+05, 3.1597E+05, 3.1626E+05, 3.1655E+05, 3.1683E+05,
     + 3.1712E+05, 3.1741E+05, 3.1770E+05, 3.1799E+05, 3.1827E+05,
     +  3.1856E+05, 3.1885E+05, 3.1914E+05, 3.1943E+05, 3.1972E+05,
     + 3.2001E+05, 3.2030E+05, 3.2059E+05, 3.2088E+05, 3.2117E+05,
     +  3.2146E+05, 3.2175E+05, 3.2204E+05, 3.2233E+05, 3.2262E+05,
     + 3.2291E+05, 3.2320E+05, 3.2349E+05, 3.2378E+05, 3.2407E+05,
     +  3.2436E+05, 3.2466E+05, 3.2495E+05, 3.2524E+05, 3.2553E+05,
     + 3.2582E+05, 3.2612E+05, 3.2641E+05, 3.2670E+05, 3.2699E+05,
     +  3.2729E+05, 3.2758E+05, 3.2787E+05, 3.2817E+05, 3.2846E+05,
     + 3.2875E+05, 3.2905E+05, 3.2934E+05, 3.2964E+05, 3.2993E+05,
     +  3.3022E+05, 3.3052E+05, 3.3081E+05, 3.3111E+05, 3.3140E+05,
     + 3.3170E+05, 3.3199E+05, 3.3229E+05, 3.3259E+05, 3.3288E+05,
     +  3.3318E+05, 3.3347E+05, 3.3377E+05, 3.3407E+05, 3.3436E+05,
     + 3.3466E+05, 3.3496E+05, 3.3525E+05, 3.3555E+05, 3.3585E+05,
     +  3.3615E+05, 3.3644E+05, 3.3674E+05, 3.3704E+05, 3.3734E+05,
     + 3.3764E+05, 3.3793E+05, 3.3823E+05, 3.3853E+05, 3.3883E+05,
     +  3.3913E+05, 3.3943E+05, 3.3973E+05, 3.4003E+05, 3.4033E+05,
     + 3.4063E+05, 3.4093E+05, 3.4123E+05, 3.4153E+05, 3.4183E+05,
     +  3.4213E+05, 3.4243E+05, 3.4273E+05, 3.4303E+05, 3.4333E+05,
     + 3.4363E+05, 3.4393E+05, 3.4423E+05, 3.4454E+05, 3.4484E+05,
     +  3.4514E+05, 3.4544E+05, 3.4574E+05, 3.4605E+05, 3.4635E+05,
     + 3.4665E+05, 3.4695E+05, 3.4726E+05, 3.4756E+05, 3.4786E+05,
     +  3.4817E+05, 3.4847E+05, 3.4877E+05, 3.4908E+05, 3.4938E+05,
     + 3.4969E+05, 3.4999E+05, 3.5030E+05, 3.5060E+05, 3.5091E+05,
     +  3.5121E+05, 3.5152E+05, 3.5182E+05, 3.5213E+05, 3.5243E+05,
     + 3.5274E+05, 3.5304E+05, 3.5335E+05, 3.5366E+05, 3.5396E+05,
     +  3.5427E+05, 3.5457E+05, 3.5488E+05, 3.5519E+05, 3.5550E+05,
     + 3.5580E+05, 3.5611E+05, 3.5642E+05, 3.5672E+05, 3.5703E+05,
     +  3.5734E+05, 3.5765E+05, 3.5796E+05, 3.5827E+05, 3.5857E+05,
     + 3.5888E+05, 3.5919E+05, 3.5950E+05, 3.5981E+05, 3.6012E+05,
     +  3.6043E+05, 3.6074E+05, 3.6105E+05, 3.6136E+05, 3.6167E+05,
     + 3.6198E+05, 3.6229E+05, 3.6260E+05, 3.6291E+05, 3.6322E+05,
     +  3.6353E+05, 3.6384E+05, 3.6415E+05, 3.6446E+05, 3.6478E+05,
     + 3.6509E+05, 3.6540E+05, 3.6571E+05, 3.6602E+05, 3.6634E+05,
     +  3.6665E+05, 3.6696E+05, 3.6727E+05, 3.6759E+05, 3.6790E+05,
     + 3.6821E+05, 3.6853E+05, 3.6884E+05, 3.6915E+05, 3.6947E+05,
     +  3.6978E+05, 3.7010E+05, 3.7041E+05, 3.7072E+05, 3.7104E+05,
     + 3.7135E+05, 3.7167E+05, 3.7198E+05, 3.7230E+05, 3.7261E+05,
     +  3.7293E+05, 3.7324E+05, 3.7356E+05, 3.7388E+05, 3.7419E+05,
     + 3.7451E+05, 3.7483E+05, 3.7514E+05, 3.7546E+05, 3.7578E+05,
     +  3.7609E+05, 3.7641E+05, 3.7673E+05, 3.7704E+05, 3.7736E+05,
     + 3.7768E+05, 3.7800E+05, 3.7832E+05, 3.7863E+05, 3.7895E+05,
     +  3.7927E+05, 3.7959E+05, 3.7991E+05, 3.8023E+05, 3.8055E+05,
     + 3.8086E+05, 3.8118E+05, 3.8150E+05, 3.8182E+05, 3.8214E+05,
     +  3.8246E+05, 3.8278E+05, 3.8310E+05, 3.8342E+05, 3.8374E+05,
     + 3.8406E+05, 3.8439E+05, 3.8471E+05, 3.8503E+05, 3.8535E+05,
     +  3.8567E+05, 3.8599E+05, 3.8631E+05, 3.8664E+05, 3.8696E+05,
     + 3.8728E+05, 3.8760E+05, 3.8793E+05, 3.8825E+05, 3.8857E+05,
     +  3.8889E+05, 3.8922E+05, 3.8954E+05, 3.8986E+05, 3.9019E+05,
     + 3.9051E+05, 3.9083E+05, 3.9116E+05, 3.9148E+05, 3.9181E+05,
     +  3.9213E+05, 3.9246E+05, 3.9278E+05, 3.9311E+05, 3.9343E+05,
     + 3.9376E+05, 3.9408E+05, 3.9441E+05, 3.9473E+05, 3.9506E+05,
     +  3.9538E+05, 3.9571E+05, 3.9604E+05, 3.9636E+05, 3.9669E+05,
     + 3.9702E+05, 3.9734E+05, 3.9767E+05, 3.9800E+05, 3.9832E+05,
     +  3.9865E+05, 3.9898E+05, 3.9931E+05, 3.9964E+05, 3.9996E+05,
     + 4.0029E+05, 4.0062E+05, 4.0095E+05, 4.0128E+05, 4.0161E+05,
     +  4.0194E+05, 4.0226E+05, 4.0259E+05, 4.0292E+05, 4.0325E+05,
     + 4.0358E+05, 4.0391E+05, 4.0424E+05, 4.0457E+05, 4.0490E+05,
     +  4.0523E+05, 4.0556E+05, 4.0589E+05, 4.0623E+05, 4.0656E+05/
      data (QofT( 6,J),J=2409,4510)/ 4.0689E+05, 4.0722E+05,
     +  4.0755E+05, 4.0788E+05, 4.0821E+05, 4.0855E+05, 4.0888E+05,
     +  4.0921E+05,
     + 4.0954E+05, 4.0988E+05, 4.1021E+05, 4.1054E+05, 4.1087E+05,
     +  4.1121E+05, 4.1154E+05, 4.1187E+05, 4.1221E+05, 4.1254E+05,
     + 4.1288E+05, 4.1321E+05, 4.1354E+05, 4.1388E+05, 4.1421E+05,
     +  4.1455E+05, 4.1488E+05, 4.1522E+05, 4.1555E+05, 4.1589E+05,
     + 4.1622E+05, 4.1656E+05, 4.1689E+05, 4.1723E+05, 4.1757E+05,
     +  4.1790E+05, 4.1824E+05, 4.1858E+05, 4.1891E+05, 4.1925E+05,
     + 4.1959E+05, 4.1992E+05, 4.2026E+05, 4.2060E+05, 4.2094E+05,
     +  4.2127E+05, 4.2161E+05, 4.2195E+05, 4.2229E+05, 4.2263E+05,
     + 4.2296E+05, 4.2330E+05, 4.2364E+05, 4.2398E+05, 4.2432E+05,
     +  4.2466E+05, 4.2500E+05, 4.2534E+05, 4.2568E+05, 4.2602E+05,
     + 4.2636E+05, 4.2670E+05, 4.2704E+05, 4.2738E+05, 4.2772E+05,
     +  4.2806E+05, 4.2840E+05, 4.2874E+05, 4.2908E+05, 4.2942E+05,
     + 4.2976E+05, 4.3011E+05, 4.3045E+05, 4.3079E+05, 4.3113E+05,
     +  4.3147E+05, 4.3182E+05, 4.3216E+05, 4.3250E+05, 4.3284E+05,
     + 4.3319E+05, 4.3353E+05, 4.3387E+05, 4.3422E+05, 4.3456E+05,
     +  4.3490E+05, 4.3525E+05, 4.3559E+05, 4.3594E+05, 4.3628E+05,
     + 4.3663E+05, 4.3697E+05, 4.3731E+05, 4.3766E+05, 4.3800E+05,
     +  4.3835E+05, 4.3870E+05, 4.3904E+05, 4.3939E+05, 4.3973E+05,
     + 4.4008E+05, 4.4042E+05, 4.4077E+05, 4.4112E+05, 4.4146E+05,
     +  4.4181E+05, 4.4216E+05, 4.4250E+05, 4.4285E+05, 4.4320E+05,
     + 4.4355E+05, 4.4389E+05, 4.4424E+05, 4.4459E+05, 4.4494E+05,
     +  4.4529E+05, 4.4563E+05, 4.4598E+05, 4.4633E+05, 4.4668E+05,
     + 4.4703E+05, 4.4738E+05, 4.4773E+05, 4.4808E+05, 4.4843E+05,
     +  4.4878E+05, 4.4913E+05, 4.4948E+05, 4.4983E+05, 4.5018E+05,
     + 4.5053E+05, 4.5088E+05, 4.5123E+05, 4.5158E+05, 4.5193E+05,
     +  4.5228E+05, 4.5263E+05, 4.5298E+05, 4.5334E+05, 4.5369E+05,
     + 4.5404E+05, 4.5439E+05, 4.5474E+05, 4.5510E+05, 4.5545E+05,
     +  4.5580E+05, 4.5615E+05, 4.5651E+05, 4.5686E+05, 4.5721E+05,
     + 4.5757E+05, 4.5792E+05, 4.5827E+05, 4.5863E+05, 4.5898E+05,
     +  4.5934E+05, 4.5969E+05, 4.6005E+05, 4.6040E+05, 4.6076E+05,
     + 4.6111E+05, 4.6147E+05, 4.6182E+05, 4.6218E+05, 4.6253E+05,
     +  4.6289E+05, 4.6324E+05, 4.6360E+05, 4.6396E+05, 4.6431E+05,
     + 4.6467E+05, 4.6502E+05, 4.6538E+05, 4.6574E+05, 4.6610E+05,
     +  4.6645E+05, 4.6681E+05, 4.6717E+05, 4.6753E+05, 4.6788E+05,
     + 4.6824E+05, 4.6860E+05, 4.6896E+05, 4.6932E+05, 4.6967E+05,
     +  4.7003E+05, 4.7039E+05, 4.7075E+05, 4.7111E+05, 4.7147E+05,
     + 4.7183E+05, 4.7219E+05, 4.7255E+05, 4.7291E+05, 4.7327E+05,
     +  4.7363E+05, 4.7399E+05, 4.7435E+05, 4.7471E+05, 4.7507E+05,
     + 4.7543E+05, 4.7579E+05, 4.7615E+05, 4.7652E+05, 4.7688E+05,
     +  4.7724E+05, 4.7760E+05, 4.7796E+05, 4.7833E+05, 4.7869E+05,
     + 4.7905E+05, 4.7941E+05, 4.7978E+05, 4.8014E+05, 4.8050E+05,
     +  4.8086E+05, 4.8123E+05, 4.8159E+05, 4.8196E+05, 4.8232E+05,
     + 4.8268E+05, 4.8305E+05, 4.8341E+05, 4.8378E+05, 4.8414E+05,
     +  4.8450E+05, 4.8487E+05, 4.8523E+05, 4.8560E+05, 4.8597E+05,
     + 4.8633E+05, 4.8670E+05, 4.8706E+05, 4.8743E+05, 4.8779E+05,
     +  4.8816E+05, 4.8853E+05, 4.8889E+05, 4.8926E+05, 4.8963E+05,
     + 4.8999E+05, 4.9036E+05, 4.9073E+05, 4.9110E+05, 4.9146E+05,
     +  4.9183E+05, 4.9220E+05, 4.9257E+05, 4.9293E+05, 4.9330E+05,
     + 4.9367E+05, 4.9404E+05, 4.9441E+05, 4.9478E+05, 4.9515E+05,
     +  4.9552E+05, 4.9588E+05, 4.9625E+05, 4.9662E+05, 4.9699E+05,
     + 4.9736E+05, 4.9773E+05, 4.9810E+05, 4.9847E+05, 4.9885E+05,
     +  4.9922E+05, 4.9959E+05, 4.9996E+05, 5.0033E+05, 5.0070E+05,
     + 5.0107E+05, 5.0144E+05, 5.0181E+05, 5.0219E+05, 5.0256E+05,
     +  5.0293E+05, 5.0330E+05, 5.0368E+05, 5.0405E+05, 5.0442E+05,
     + 5.0479E+05, 5.0517E+05, 5.0554E+05, 5.0591E+05, 5.0629E+05,
     +  5.0666E+05, 5.0703E+05, 5.0741E+05, 5.0778E+05, 5.0816E+05,
     + 5.0853E+05, 5.0891E+05, 5.0928E+05, 5.0966E+05, 5.1003E+05,
     +  5.1041E+05, 5.1078E+05, 5.1116E+05, 5.1153E+05, 5.1191E+05,
     + 5.1228E+05, 5.1266E+05, 5.1304E+05, 5.1341E+05, 5.1379E+05,
     +  5.1417E+05, 5.1454E+05, 5.1492E+05, 5.1530E+05, 5.1567E+05,
     + 5.1605E+05, 5.1643E+05, 5.1681E+05, 5.1718E+05, 5.1756E+05,
     +  5.1794E+05, 5.1832E+05, 5.1870E+05, 5.1908E+05, 5.1945E+05,
     + 5.1983E+05, 5.2021E+05, 5.2059E+05, 5.2097E+05, 5.2135E+05,
     +  5.2173E+05, 5.2211E+05, 5.2249E+05, 5.2287E+05, 5.2325E+05,
     + 5.2363E+05, 5.2401E+05, 5.2439E+05, 5.2477E+05, 5.2515E+05,
     +  5.2553E+05, 5.2592E+05, 5.2630E+05, 5.2668E+05, 5.2706E+05,
     + 5.2744E+05, 5.2782E+05, 5.2821E+05, 5.2859E+05, 5.2897E+05,
     +  5.2935E+05, 5.2974E+05, 5.3012E+05, 5.3050E+05, 5.3089E+05,
     + 5.3127E+05, 5.3165E+05, 5.3204E+05, 5.3242E+05, 5.3280E+05,
     +  5.3319E+05, 5.3357E+05, 5.3396E+05, 5.3434E+05, 5.3473E+05,
     + 5.3511E+05, 5.3550E+05, 5.3588E+05, 5.3627E+05, 5.3665E+05,
     +  5.3704E+05, 5.3742E+05, 5.3781E+05, 5.3820E+05, 5.3858E+05,
     + 5.3897E+05, 5.3935E+05, 5.3974E+05, 5.4013E+05, 5.4051E+05,
     +  5.4090E+05, 5.4129E+05, 5.4168E+05, 5.4206E+05, 5.4245E+05,
     + 5.4284E+05, 5.4323E+05, 5.4362E+05, 5.4400E+05, 5.4439E+05,
     +  5.4478E+05, 5.4517E+05, 5.4556E+05, 5.4595E+05, 5.4634E+05,
     + 5.4673E+05, 5.4712E+05, 5.4750E+05, 5.4789E+05, 5.4828E+05,
     +  5.4867E+05, 5.4906E+05, 5.4946E+05, 5.4985E+05, 5.5024E+05,
     + 5.5063E+05, 5.5102E+05, 5.5141E+05, 5.5180E+05, 5.5219E+05,
     +  5.5258E+05, 5.5297E+05, 5.5337E+05, 5.5376E+05, 5.5415E+05,
     + 5.5454E+05, 5.5494E+05, 5.5533E+05, 5.5572E+05, 5.5611E+05,
     +  5.5651E+05, 5.5690E+05, 5.5729E+05, 5.5769E+05, 5.5808E+05,
     + 5.5847E+05, 5.5887E+05, 5.5926E+05, 5.5966E+05, 5.6005E+05,
     +  5.6044E+05, 5.6084E+05, 5.6123E+05, 5.6163E+05, 5.6202E+05,
     + 5.6242E+05, 5.6281E+05, 5.6321E+05, 5.6361E+05, 5.6400E+05,
     +  5.6440E+05, 5.6479E+05, 5.6519E+05, 5.6559E+05, 5.6598E+05,
     + 5.6638E+05, 5.6678E+05, 5.6717E+05, 5.6757E+05, 5.6797E+05,
     +  5.6837E+05, 5.6876E+05, 5.6916E+05, 5.6956E+05, 5.6996E+05,
     + 5.7035E+05, 5.7075E+05, 5.7115E+05, 5.7155E+05, 5.7195E+05,
     +  5.7235E+05, 5.7275E+05, 5.7315E+05, 5.7355E+05, 5.7395E+05,
     + 5.7434E+05, 5.7474E+05, 5.7514E+05, 5.7554E+05, 5.7594E+05,
     +  5.7635E+05, 5.7675E+05, 5.7715E+05, 5.7755E+05, 5.7795E+05,
     + 5.7835E+05, 5.7875E+05, 5.7915E+05, 5.7955E+05, 5.7996E+05,
     +  5.8036E+05, 5.8076E+05, 5.8116E+05, 5.8156E+05, 5.8197E+05,
     + 5.8237E+05, 5.8277E+05, 5.8317E+05, 5.8358E+05, 5.8398E+05,
     +  5.8438E+05, 5.8479E+05, 5.8519E+05, 5.8560E+05, 5.8600E+05,
     + 5.8640E+05, 5.8681E+05, 5.8721E+05, 5.8762E+05, 5.8802E+05,
     +  5.8843E+05, 5.8883E+05, 5.8924E+05, 5.8964E+05, 5.9005E+05,
     + 5.9045E+05, 5.9086E+05, 5.9126E+05, 5.9167E+05, 5.9208E+05,
     +  5.9248E+05, 5.9289E+05, 5.9330E+05, 5.9370E+05, 5.9411E+05,
     + 5.9452E+05, 5.9492E+05, 5.9533E+05, 5.9574E+05, 5.9615E+05,
     +  5.9655E+05, 5.9696E+05, 5.9737E+05, 5.9778E+05, 5.9819E+05,
     + 5.9859E+05, 5.9900E+05, 5.9941E+05, 5.9982E+05, 6.0023E+05,
     +  6.0064E+05, 6.0105E+05, 6.0146E+05, 6.0187E+05, 6.0228E+05,
     + 6.0269E+05, 6.0310E+05, 6.0351E+05, 6.0392E+05, 6.0433E+05,
     +  6.0474E+05, 6.0515E+05, 6.0556E+05, 6.0597E+05, 6.0638E+05,
     + 6.0679E+05, 6.0721E+05, 6.0762E+05, 6.0803E+05, 6.0844E+05,
     +  6.0885E+05, 6.0927E+05, 6.0968E+05, 6.1009E+05, 6.1050E+05,
     + 6.1092E+05, 6.1133E+05, 6.1174E+05, 6.1216E+05, 6.1257E+05,
     +  6.1298E+05, 6.1340E+05, 6.1381E+05, 6.1423E+05, 6.1464E+05,
     + 6.1505E+05, 6.1547E+05, 6.1588E+05, 6.1630E+05, 6.1671E+05,
     +  6.1713E+05, 6.1754E+05, 6.1796E+05, 6.1837E+05, 6.1879E+05,
     + 6.1921E+05, 6.1962E+05, 6.2004E+05, 6.2045E+05, 6.2087E+05,
     +  6.2129E+05, 6.2170E+05, 6.2212E+05, 6.2254E+05, 6.2295E+05,
     + 6.2337E+05, 6.2379E+05, 6.2421E+05, 6.2462E+05, 6.2504E+05,
     +  6.2546E+05, 6.2588E+05, 6.2630E+05, 6.2672E+05, 6.2713E+05,
     + 6.2755E+05, 6.2797E+05, 6.2839E+05, 6.2881E+05, 6.2923E+05,
     +  6.2965E+05, 6.3007E+05, 6.3049E+05, 6.3091E+05, 6.3133E+05,
     + 6.3175E+05, 6.3217E+05, 6.3259E+05, 6.3301E+05, 6.3343E+05,
     +  6.3385E+05, 6.3427E+05, 6.3469E+05, 6.3511E+05, 6.3554E+05,
     + 6.3596E+05, 6.3638E+05, 6.3680E+05, 6.3722E+05, 6.3765E+05,
     +  6.3807E+05, 6.3849E+05, 6.3891E+05, 6.3934E+05, 6.3976E+05,
     + 6.4018E+05, 6.4061E+05, 6.4103E+05, 6.4145E+05, 6.4188E+05,
     +  6.4230E+05, 6.4272E+05, 6.4315E+05, 6.4357E+05, 6.4400E+05,
     + 6.4442E+05, 6.4485E+05, 6.4527E+05, 6.4570E+05, 6.4612E+05,
     +  6.4655E+05, 6.4697E+05, 6.4740E+05, 6.4782E+05, 6.4825E+05,
     + 6.4867E+05, 6.4910E+05, 6.4953E+05, 6.4995E+05, 6.5038E+05,
     +  6.5081E+05, 6.5123E+05, 6.5166E+05, 6.5209E+05, 6.5252E+05,
     + 6.5294E+05, 6.5337E+05, 6.5380E+05, 6.5423E+05, 6.5465E+05,
     +  6.5508E+05, 6.5551E+05, 6.5594E+05, 6.5637E+05, 6.5680E+05,
     + 6.5723E+05, 6.5765E+05, 6.5808E+05, 6.5851E+05, 6.5894E+05,
     +  6.5937E+05, 6.5980E+05, 6.6023E+05, 6.6066E+05, 6.6109E+05,
     + 6.6152E+05, 6.6195E+05, 6.6238E+05, 6.6281E+05, 6.6325E+05,
     +  6.6368E+05, 6.6411E+05, 6.6454E+05, 6.6497E+05, 6.6540E+05,
     + 6.6583E+05, 6.6627E+05, 6.6670E+05, 6.6713E+05, 6.6756E+05,
     +  6.6799E+05, 6.6843E+05, 6.6886E+05, 6.6929E+05, 6.6973E+05,
     + 6.7016E+05, 6.7059E+05, 6.7103E+05, 6.7146E+05, 6.7189E+05,
     +  6.7233E+05, 6.7276E+05, 6.7320E+05, 6.7363E+05, 6.7407E+05,
     + 6.7450E+05, 6.7493E+05, 6.7537E+05, 6.7580E+05, 6.7624E+05,
     +  6.7668E+05, 6.7711E+05, 6.7755E+05, 6.7798E+05, 6.7842E+05,
     + 6.7885E+05, 6.7929E+05, 6.7973E+05, 6.8016E+05, 6.8060E+05,
     +  6.8104E+05, 6.8147E+05, 6.8191E+05, 6.8235E+05, 6.8279E+05,
     + 6.8322E+05, 6.8366E+05, 6.8410E+05, 6.8454E+05, 6.8498E+05,
     +  6.8541E+05, 6.8585E+05, 6.8629E+05, 6.8673E+05, 6.8717E+05,
     + 6.8761E+05, 6.8805E+05, 6.8849E+05, 6.8893E+05, 6.8936E+05,
     +  6.8980E+05, 6.9024E+05, 6.9068E+05, 6.9112E+05, 6.9156E+05,
     + 6.9201E+05, 6.9245E+05, 6.9289E+05, 6.9333E+05, 6.9377E+05,
     +  6.9421E+05, 6.9465E+05, 6.9509E+05, 6.9553E+05, 6.9598E+05,
     + 6.9642E+05, 6.9686E+05, 6.9730E+05, 6.9774E+05, 6.9819E+05,
     +  6.9863E+05, 6.9907E+05, 6.9951E+05, 6.9996E+05, 7.0040E+05,
     + 7.0084E+05, 7.0129E+05, 7.0173E+05, 7.0217E+05, 7.0262E+05,
     +  7.0306E+05, 7.0351E+05, 7.0395E+05, 7.0440E+05, 7.0484E+05,
     + 7.0528E+05, 7.0573E+05, 7.0617E+05, 7.0662E+05, 7.0707E+05,
     +  7.0751E+05, 7.0796E+05, 7.0840E+05, 7.0885E+05, 7.0929E+05,
     + 7.0974E+05, 7.1019E+05, 7.1063E+05, 7.1108E+05, 7.1153E+05,
     +  7.1197E+05, 7.1242E+05, 7.1287E+05, 7.1331E+05, 7.1376E+05,
     + 7.1421E+05, 7.1466E+05, 7.1510E+05, 7.1555E+05, 7.1600E+05,
     +  7.1645E+05, 7.1690E+05, 7.1735E+05, 7.1780E+05, 7.1824E+05,
     + 7.1869E+05, 7.1914E+05, 7.1959E+05, 7.2004E+05, 7.2049E+05,
     +  7.2094E+05, 7.2139E+05, 7.2184E+05, 7.2229E+05, 7.2274E+05,
     + 7.2319E+05, 7.2364E+05, 7.2409E+05, 7.2454E+05, 7.2499E+05,
     +  7.2544E+05, 7.2590E+05, 7.2635E+05, 7.2680E+05, 7.2725E+05,
     + 7.2770E+05, 7.2815E+05, 7.2861E+05, 7.2906E+05, 7.2951E+05,
     +  7.2996E+05, 7.3042E+05, 7.3087E+05, 7.3132E+05, 7.3178E+05,
     + 7.3223E+05, 7.3268E+05, 7.3314E+05, 7.3359E+05, 7.3404E+05,
     +  7.3450E+05, 7.3495E+05, 7.3541E+05, 7.3586E+05, 7.3631E+05,
     + 7.3677E+05, 7.3722E+05, 7.3768E+05, 7.3813E+05, 7.3859E+05,
     +  7.3904E+05, 7.3950E+05, 7.3996E+05, 7.4041E+05, 7.4087E+05,
     + 7.4132E+05, 7.4178E+05, 7.4224E+05, 7.4269E+05, 7.4315E+05,
     +  7.4361E+05, 7.4406E+05, 7.4452E+05, 7.4498E+05, 7.4544E+05,
     + 7.4589E+05, 7.4635E+05, 7.4681E+05, 7.4727E+05, 7.4772E+05,
     +  7.4818E+05, 7.4864E+05, 7.4910E+05, 7.4956E+05, 7.5002E+05,
     + 7.5048E+05, 7.5093E+05, 7.5139E+05, 7.5185E+05, 7.5231E+05,
     +  7.5277E+05, 7.5323E+05, 7.5369E+05, 7.5415E+05, 7.5461E+05,
     + 7.5507E+05, 7.5553E+05, 7.5599E+05, 7.5645E+05, 7.5692E+05,
     +  7.5738E+05, 7.5784E+05, 7.5830E+05, 7.5876E+05, 7.5922E+05,
     + 7.5968E+05, 7.6015E+05, 7.6061E+05, 7.6107E+05, 7.6153E+05,
     +  7.6199E+05, 7.6246E+05, 7.6292E+05, 7.6338E+05, 7.6385E+05,
     + 7.6431E+05, 7.6477E+05, 7.6524E+05, 7.6570E+05, 7.6616E+05,
     +  7.6663E+05, 7.6709E+05, 7.6755E+05, 7.6802E+05, 7.6848E+05,
     + 7.6895E+05, 7.6941E+05, 7.6988E+05, 7.7034E+05, 7.7081E+05,
     +  7.7127E+05, 7.7174E+05, 7.7220E+05, 7.7267E+05, 7.7313E+05,
     + 7.7360E+05, 7.7407E+05, 7.7453E+05, 7.7500E+05, 7.7547E+05,
     +  7.7593E+05, 7.7640E+05, 7.7687E+05, 7.7733E+05, 7.7780E+05,
     + 7.7827E+05, 7.7874E+05, 7.7920E+05, 7.7967E+05, 7.8014E+05,
     +  7.8061E+05, 7.8107E+05, 7.8154E+05, 7.8201E+05, 7.8248E+05,
     + 7.8295E+05, 7.8342E+05, 7.8389E+05, 7.8436E+05, 7.8482E+05,
     +  7.8529E+05, 7.8576E+05, 7.8623E+05, 7.8670E+05, 7.8717E+05,
     + 7.8764E+05, 7.8811E+05, 7.8858E+05, 7.8905E+05, 7.8953E+05,
     +  7.9000E+05, 7.9047E+05, 7.9094E+05, 7.9141E+05, 7.9188E+05,
     + 7.9235E+05, 7.9282E+05, 7.9330E+05, 7.9377E+05, 7.9424E+05,
     +  7.9471E+05, 7.9518E+05, 7.9566E+05, 7.9613E+05, 7.9660E+05,
     + 7.9708E+05, 7.9755E+05, 7.9802E+05, 7.9849E+05, 7.9897E+05,
     +  7.9944E+05, 7.9992E+05, 8.0039E+05, 8.0086E+05, 8.0134E+05,
     + 8.0181E+05, 8.0229E+05, 8.0276E+05, 8.0324E+05, 8.0371E+05,
     +  8.0419E+05, 8.0466E+05, 8.0514E+05, 8.0561E+05, 8.0609E+05,
     + 8.0656E+05, 8.0704E+05, 8.0751E+05, 8.0799E+05, 8.0847E+05,
     +  8.0894E+05, 8.0942E+05, 8.0990E+05, 8.1037E+05, 8.1085E+05,
     + 8.1133E+05, 8.1180E+05, 8.1228E+05, 8.1276E+05, 8.1324E+05,
     +  8.1371E+05, 8.1419E+05, 8.1467E+05, 8.1515E+05, 8.1563E+05,
     + 8.1610E+05, 8.1658E+05, 8.1706E+05, 8.1754E+05, 8.1802E+05,
     +  8.1850E+05, 8.1898E+05, 8.1946E+05, 8.1994E+05, 8.2042E+05,
     + 8.2090E+05, 8.2138E+05, 8.2186E+05, 8.2234E+05, 8.2282E+05,
     +  8.2330E+05, 8.2378E+05, 8.2426E+05, 8.2474E+05, 8.2522E+05,
     + 8.2570E+05, 8.2618E+05, 8.2667E+05, 8.2715E+05, 8.2763E+05,
     +  8.2811E+05, 8.2859E+05, 8.2907E+05, 8.2956E+05, 8.3004E+05,
     + 8.3052E+05, 8.3100E+05, 8.3149E+05, 8.3197E+05, 8.3245E+05,
     +  8.3294E+05, 8.3342E+05, 8.3390E+05, 8.3439E+05, 8.3487E+05,
     + 8.3536E+05, 8.3584E+05, 8.3632E+05, 8.3681E+05, 8.3729E+05,
     +  8.3778E+05, 8.3826E+05, 8.3875E+05, 8.3923E+05, 8.3972E+05,
     + 8.4020E+05, 8.4069E+05, 8.4117E+05, 8.4166E+05, 8.4214E+05,
     +  8.4263E+05, 8.4312E+05, 8.4360E+05, 8.4409E+05, 8.4458E+05,
     + 8.4506E+05, 8.4555E+05, 8.4604E+05, 8.4652E+05, 8.4701E+05,
     +  8.4750E+05, 8.4799E+05, 8.4847E+05, 8.4896E+05, 8.4945E+05,
     + 8.4994E+05, 8.5042E+05, 8.5091E+05, 8.5140E+05, 8.5189E+05,
     +  8.5238E+05, 8.5287E+05, 8.5336E+05, 8.5385E+05, 8.5433E+05,
     + 8.5482E+05, 8.5531E+05, 8.5580E+05, 8.5629E+05, 8.5678E+05,
     +  8.5727E+05, 8.5776E+05, 8.5825E+05, 8.5874E+05, 8.5924E+05,
     + 8.5973E+05, 8.6022E+05, 8.6071E+05, 8.6120E+05, 8.6169E+05,
     +  8.6218E+05, 8.6267E+05, 8.6316E+05, 8.6366E+05, 8.6415E+05,
     + 8.6464E+05, 8.6513E+05, 8.6563E+05, 8.6612E+05, 8.6661E+05,
     +  8.6710E+05, 8.6760E+05, 8.6809E+05, 8.6858E+05, 8.6908E+05,
     + 8.6957E+05, 8.7006E+05, 8.7056E+05, 8.7105E+05, 8.7154E+05,
     +  8.7204E+05, 8.7253E+05, 8.7303E+05, 8.7352E+05, 8.7402E+05,
     + 8.7451E+05, 8.7501E+05, 8.7550E+05, 8.7600E+05, 8.7649E+05,
     +  8.7699E+05, 8.7748E+05, 8.7798E+05, 8.7847E+05, 8.7897E+05,
     + 8.7947E+05, 8.7996E+05, 8.8046E+05, 8.8096E+05, 8.8145E+05,
     +  8.8195E+05, 8.8245E+05, 8.8294E+05, 8.8344E+05, 8.8394E+05,
     + 8.8444E+05, 8.8493E+05, 8.8543E+05, 8.8593E+05, 8.8643E+05,
     +  8.8693E+05, 8.8742E+05, 8.8792E+05, 8.8842E+05, 8.8892E+05,
     + 8.8942E+05, 8.8992E+05, 8.9042E+05, 8.9092E+05, 8.9142E+05,
     +  8.9191E+05, 8.9241E+05, 8.9291E+05, 8.9341E+05, 8.9391E+05,
     + 8.9441E+05, 8.9491E+05, 8.9542E+05, 8.9592E+05, 8.9642E+05,
     +  8.9692E+05, 8.9742E+05, 8.9792E+05, 8.9842E+05, 8.9892E+05,
     + 8.9942E+05, 8.9993E+05, 9.0043E+05, 9.0093E+05, 9.0143E+05,
     +  9.0193E+05, 9.0244E+05, 9.0294E+05, 9.0344E+05, 9.0394E+05,
     + 9.0445E+05, 9.0495E+05, 9.0545E+05, 9.0596E+05, 9.0646E+05,
     +  9.0696E+05, 9.0747E+05, 9.0797E+05, 9.0847E+05, 9.0898E+05,
     + 9.0948E+05, 9.0999E+05, 9.1049E+05, 9.1100E+05, 9.1150E+05,
     +  9.1201E+05, 9.1251E+05, 9.1302E+05, 9.1352E+05, 9.1403E+05,
     + 9.1453E+05, 9.1504E+05, 9.1554E+05, 9.1605E+05, 9.1656E+05,
     +  9.1706E+05, 9.1757E+05, 9.1807E+05, 9.1858E+05, 9.1909E+05,
     + 9.1959E+05, 9.2010E+05, 9.2061E+05, 9.2112E+05, 9.2162E+05,
     +  9.2213E+05, 9.2264E+05, 9.2315E+05, 9.2365E+05, 9.2416E+05,
     + 9.2467E+05, 9.2518E+05, 9.2569E+05, 9.2620E+05, 9.2670E+05,
     +  9.2721E+05, 9.2772E+05, 9.2823E+05, 9.2874E+05, 9.2925E+05,
     + 9.2976E+05, 9.3027E+05, 9.3078E+05, 9.3129E+05, 9.3180E+05,
     +  9.3231E+05, 9.3282E+05, 9.3333E+05, 9.3384E+05, 9.3435E+05,
     + 9.3486E+05, 9.3537E+05, 9.3588E+05, 9.3640E+05, 9.3691E+05,
     +  9.3742E+05, 9.3793E+05, 9.3844E+05, 9.3895E+05, 9.3947E+05,
     + 9.3998E+05, 9.4049E+05, 9.4100E+05, 9.4151E+05, 9.4203E+05,
     +  9.4254E+05, 9.4305E+05, 9.4357E+05, 9.4408E+05, 9.4459E+05,
     + 9.4511E+05, 9.4562E+05, 9.4613E+05, 9.4665E+05, 9.4716E+05,
     +  9.4768E+05, 9.4819E+05, 9.4870E+05, 9.4922E+05, 9.4973E+05,
     + 9.5025E+05, 9.5076E+05, 9.5128E+05, 9.5179E+05, 9.5231E+05,
     +  9.5282E+05, 9.5334E+05, 9.5386E+05, 9.5437E+05, 9.5489E+05,
     + 9.5540E+05, 9.5592E+05, 9.5644E+05, 9.5695E+05, 9.5747E+05,
     +  9.5799E+05, 9.5850E+05, 9.5902E+05, 9.5954E+05, 9.6005E+05,
     + 9.6057E+05, 9.6109E+05, 9.6161E+05, 9.6212E+05, 9.6264E+05,
     +  9.6316E+05, 9.6368E+05, 9.6420E+05, 9.6472E+05, 9.6523E+05,
     + 9.6575E+05, 9.6627E+05, 9.6679E+05, 9.6731E+05, 9.6783E+05,
     +  9.6835E+05, 9.6887E+05, 9.6939E+05, 9.6991E+05, 9.7043E+05,
     + 9.7095E+05, 9.7147E+05, 9.7199E+05, 9.7251E+05, 9.7303E+05,
     +  9.7355E+05, 9.7407E+05, 9.7459E+05, 9.7511E+05, 9.7563E+05,
     + 9.7615E+05, 9.7668E+05, 9.7720E+05, 9.7772E+05, 9.7824E+05,
     +  9.7876E+05, 9.7928E+05, 9.7981E+05, 9.8033E+05, 9.8085E+05,
     + 9.8137E+05, 9.8190E+05, 9.8242E+05, 9.8294E+05, 9.8347E+05,
     +  9.8399E+05, 9.8451E+05, 9.8504E+05, 9.8556E+05, 9.8608E+05,
     + 9.8661E+05, 9.8713E+05, 9.8766E+05, 9.8818E+05, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_NO    (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 3),QofT( 3,NT),Q(NT)
      data (Tmx(I),I=1,3)/3500.,3500.,3500./
      data xgj/ 3.,2.,3./
c...       NO
c...        --        46
      data (QofT( 1,J),J=1,2408)/ 1.1879E+01, 1.2547E+01, 1.4091E+01,
     +  1.6081E+01, 1.8266E+01, 2.0547E+01, 2.2881E+01, 2.5248E+01,
     + 2.7636E+01, 3.0039E+01, 3.2452E+01, 3.4873E+01, 3.7300E+01,
     +  3.9732E+01, 4.2168E+01, 4.4607E+01, 4.7049E+01, 4.9493E+01,
     + 5.1940E+01, 5.4389E+01, 5.9297E+01, 6.4217E+01, 6.9156E+01,
     +  7.4117E+01, 7.9106E+01, 8.4130E+01, 8.9193E+01, 9.4302E+01,
     + 9.9461E+01, 1.0468E+02, 1.0995E+02, 1.1529E+02, 1.2069E+02,
     +  1.2617E+02, 1.3171E+02, 1.3733E+02, 1.4302E+02, 1.4879E+02,
     + 1.5463E+02, 1.6055E+02, 1.6654E+02, 1.7261E+02, 1.7876E+02,
     +  1.8498E+02, 1.9127E+02, 1.9763E+02, 2.0407E+02, 2.1057E+02,
     + 2.1714E+02, 2.2378E+02, 2.3049E+02, 2.3726E+02, 2.4410E+02,
     +  2.5100E+02, 2.5796E+02, 2.6497E+02, 2.7205E+02, 2.7918E+02,
     + 2.8637E+02, 2.9361E+02, 3.0090E+02, 3.0825E+02, 3.1564E+02,
     +  3.2308E+02, 3.3080E+02, 3.3843E+02, 3.4611E+02, 3.5382E+02,
     + 3.6158E+02, 3.6938E+02, 3.7722E+02, 3.8510E+02, 3.9301E+02,
     +  4.0096E+02, 4.0895E+02, 4.1697E+02, 4.2503E+02, 4.3312E+02,
     + 4.4124E+02, 4.4940E+02, 4.5758E+02, 4.6580E+02, 4.7404E+02,
     +  4.8231E+02, 4.9061E+02, 4.9894E+02, 5.0730E+02, 5.1568E+02,
     + 5.2408E+02, 5.3251E+02, 5.4096E+02, 5.4944E+02, 5.5794E+02,
     +  5.6646E+02, 5.7500E+02, 5.8357E+02, 5.9215E+02, 6.0076E+02,
     + 6.0938E+02, 6.1803E+02, 6.2669E+02, 6.3537E+02, 6.4407E+02,
     +  6.5279E+02, 6.6153E+02, 6.7028E+02, 6.7905E+02, 6.8783E+02,
     + 6.9663E+02, 7.0545E+02, 7.1428E+02, 7.2313E+02, 7.3199E+02,
     +  7.4086E+02, 7.4975E+02, 7.5866E+02, 7.6757E+02, 7.7650E+02,
     + 7.8544E+02, 7.9440E+02, 8.0337E+02, 8.1235E+02, 8.2134E+02,
     +  8.3035E+02, 8.3936E+02, 8.4839E+02, 8.5742E+02, 8.6647E+02,
     + 8.7553E+02, 8.8460E+02, 8.9367E+02, 9.0277E+02, 9.1187E+02,
     +  9.2098E+02, 9.3009E+02, 9.3922E+02, 9.4835E+02, 9.5751E+02,
     + 9.6666E+02, 9.7582E+02, 9.8499E+02, 9.9418E+02, 1.0034E+03,
     +  1.0126E+03, 1.0218E+03, 1.0310E+03, 1.0402E+03, 1.0494E+03,
     + 1.0587E+03, 1.0679E+03, 1.0772E+03, 1.0864E+03, 1.0957E+03,
     +  1.1050E+03, 1.1142E+03, 1.1235E+03, 1.1328E+03, 1.1421E+03,
     + 1.1514E+03, 1.1608E+03, 1.1701E+03, 1.1794E+03, 1.1887E+03,
     +  1.1981E+03, 1.2074E+03, 1.2168E+03, 1.2261E+03, 1.2355E+03,
     + 1.2449E+03, 1.2543E+03, 1.2636E+03, 1.2730E+03, 1.2824E+03,
     +  1.2918E+03, 1.3012E+03, 1.3106E+03, 1.3201E+03, 1.3295E+03,
     + 1.3389E+03, 1.3484E+03, 1.3578E+03, 1.3672E+03, 1.3767E+03,
     +  1.3861E+03, 1.3956E+03, 1.4051E+03, 1.4145E+03, 1.4240E+03,
     + 1.4335E+03, 1.4430E+03, 1.4525E+03, 1.4620E+03, 1.4715E+03,
     +  1.4810E+03, 1.4905E+03, 1.5001E+03, 1.5096E+03, 1.5191E+03,
     + 1.5286E+03, 1.5382E+03, 1.5477E+03, 1.5573E+03, 1.5668E+03,
     +  1.5764E+03, 1.5860E+03, 1.5955E+03, 1.6051E+03, 1.6147E+03,
     + 1.6243E+03, 1.6339E+03, 1.6435E+03, 1.6531E+03, 1.6627E+03,
     +  1.6723E+03, 1.6820E+03, 1.6916E+03, 1.7012E+03, 1.7108E+03,
     + 1.7205E+03, 1.7301E+03, 1.7398E+03, 1.7494E+03, 1.7591E+03,
     +  1.7688E+03, 1.7785E+03, 1.7881E+03, 1.7978E+03, 1.8075E+03,
     + 1.8172E+03, 1.8269E+03, 1.8366E+03, 1.8463E+03, 1.8561E+03,
     +  1.8658E+03, 1.8755E+03, 1.8853E+03, 1.8950E+03, 1.9048E+03,
     + 1.9145E+03, 1.9243E+03, 1.9340E+03, 1.9438E+03, 1.9536E+03,
     +  1.9634E+03, 1.9732E+03, 1.9830E+03, 1.9928E+03, 2.0026E+03,
     + 2.0124E+03, 2.0222E+03, 2.0320E+03, 2.0418E+03, 2.0517E+03,
     +  2.0615E+03, 2.0714E+03, 2.0812E+03, 2.0911E+03, 2.1010E+03,
     + 2.1109E+03, 2.1207E+03, 2.1306E+03, 2.1405E+03, 2.1504E+03,
     +  2.1603E+03, 2.1703E+03, 2.1802E+03, 2.1901E+03, 2.2000E+03,
     + 2.2100E+03, 2.2199E+03, 2.2299E+03, 2.2399E+03, 2.2498E+03,
     +  2.2598E+03, 2.2698E+03, 2.2798E+03, 2.2898E+03, 2.2998E+03,
     + 2.3098E+03, 2.3198E+03, 2.3298E+03, 2.3399E+03, 2.3499E+03,
     +  2.3600E+03, 2.3700E+03, 2.3801E+03, 2.3901E+03, 2.4002E+03,
     + 2.4103E+03, 2.4204E+03, 2.4305E+03, 2.4406E+03, 2.4507E+03,
     +  2.4609E+03, 2.4710E+03, 2.4811E+03, 2.4913E+03, 2.5014E+03,
     + 2.5116E+03, 2.5217E+03, 2.5319E+03, 2.5421E+03, 2.5523E+03,
     +  2.5625E+03, 2.5727E+03, 2.5830E+03, 2.5932E+03, 2.6034E+03,
     + 2.6137E+03, 2.6239E+03, 2.6342E+03, 2.6444E+03, 2.6547E+03,
     +  2.6650E+03, 2.6753E+03, 2.6856E+03, 2.6959E+03, 2.7062E+03,
     + 2.7165E+03, 2.7269E+03, 2.7372E+03, 2.7476E+03, 2.7579E+03,
     +  2.7683E+03, 2.7787E+03, 2.7891E+03, 2.7994E+03, 2.8099E+03,
     + 2.8203E+03, 2.8307E+03, 2.8411E+03, 2.8516E+03, 2.8620E+03,
     +  2.8725E+03, 2.8829E+03, 2.8934E+03, 2.9039E+03, 2.9144E+03,
     + 2.9249E+03, 2.9354E+03, 2.9459E+03, 2.9564E+03, 2.9670E+03,
     +  2.9775E+03, 2.9881E+03, 2.9987E+03, 3.0092E+03, 3.0199E+03,
     + 3.0304E+03, 3.0410E+03, 3.0517E+03, 3.0623E+03, 3.0729E+03,
     +  3.0836E+03, 3.0942E+03, 3.1049E+03, 3.1156E+03, 3.1262E+03,
     + 3.1369E+03, 3.1476E+03, 3.1584E+03, 3.1691E+03, 3.1798E+03,
     +  3.1906E+03, 3.2013E+03, 3.2121E+03, 3.2228E+03, 3.2336E+03,
     + 3.2444E+03, 3.2552E+03, 3.2660E+03, 3.2768E+03, 3.2877E+03,
     +  3.2985E+03, 3.3094E+03, 3.3202E+03, 3.3311E+03, 3.3420E+03,
     + 3.3529E+03, 3.3638E+03, 3.3747E+03, 3.3856E+03, 3.3966E+03,
     +  3.4075E+03, 3.4185E+03, 3.4295E+03, 3.4404E+03, 3.4514E+03,
     + 3.4624E+03, 3.4734E+03, 3.4845E+03, 3.4955E+03, 3.5065E+03,
     +  3.5176E+03, 3.5286E+03, 3.5397E+03, 3.5508E+03, 3.5619E+03,
     + 3.5730E+03, 3.5841E+03, 3.5952E+03, 3.6064E+03, 3.6175E+03,
     +  3.6287E+03, 3.6399E+03, 3.6511E+03, 3.6622E+03, 3.6734E+03,
     + 3.6847E+03, 3.6959E+03, 3.7071E+03, 3.7184E+03, 3.7296E+03,
     +  3.7409E+03, 3.7522E+03, 3.7635E+03, 3.7748E+03, 3.7861E+03,
     + 3.7974E+03, 3.8088E+03, 3.8201E+03, 3.8315E+03, 3.8429E+03,
     +  3.8543E+03, 3.8656E+03, 3.8770E+03, 3.8885E+03, 3.8999E+03,
     + 3.9113E+03, 3.9228E+03, 3.9343E+03, 3.9457E+03, 3.9572E+03,
     +  3.9687E+03, 3.9802E+03, 3.9918E+03, 4.0033E+03, 4.0148E+03,
     + 4.0264E+03, 4.0379E+03, 4.0495E+03, 4.0611E+03, 4.0727E+03,
     +  4.0843E+03, 4.0960E+03, 4.1076E+03, 4.1193E+03, 4.1309E+03,
     + 4.1426E+03, 4.1543E+03, 4.1660E+03, 4.1777E+03, 4.1894E+03,
     +  4.2011E+03, 4.2129E+03, 4.2247E+03, 4.2364E+03, 4.2482E+03,
     + 4.2600E+03, 4.2718E+03, 4.2836E+03, 4.2955E+03, 4.3073E+03,
     +  4.3191E+03, 4.3310E+03, 4.3429E+03, 4.3548E+03, 4.3667E+03,
     + 4.3786E+03, 4.3905E+03, 4.4024E+03, 4.4144E+03, 4.4264E+03,
     +  4.4384E+03, 4.4503E+03, 4.4623E+03, 4.4743E+03, 4.4864E+03,
     + 4.4984E+03, 4.5105E+03, 4.5225E+03, 4.5346E+03, 4.5467E+03,
     +  4.5588E+03, 4.5709E+03, 4.5830E+03, 4.5952E+03, 4.6073E+03,
     + 4.6194E+03, 4.6316E+03, 4.6438E+03, 4.6560E+03, 4.6682E+03,
     +  4.6804E+03, 4.6927E+03, 4.7049E+03, 4.7172E+03, 4.7295E+03,
     + 4.7418E+03, 4.7540E+03, 4.7663E+03, 4.7787E+03, 4.7910E+03,
     +  4.8033E+03, 4.8157E+03, 4.8281E+03, 4.8405E+03, 4.8529E+03,
     + 4.8653E+03, 4.8777E+03, 4.8901E+03, 4.9026E+03, 4.9151E+03,
     +  4.9275E+03, 4.9400E+03, 4.9525E+03, 4.9651E+03, 4.9776E+03,
     + 4.9901E+03, 5.0026E+03, 5.0152E+03, 5.0278E+03, 5.0404E+03,
     +  5.0530E+03, 5.0656E+03, 5.0782E+03, 5.0909E+03, 5.1035E+03,
     + 5.1162E+03, 5.1289E+03, 5.1416E+03, 5.1543E+03, 5.1670E+03,
     +  5.1797E+03, 5.1925E+03, 5.2053E+03, 5.2180E+03, 5.2308E+03,
     + 5.2436E+03, 5.2564E+03, 5.2692E+03, 5.2820E+03, 5.2949E+03,
     +  5.3077E+03, 5.3206E+03, 5.3335E+03, 5.3464E+03, 5.3594E+03,
     + 5.3723E+03, 5.3852E+03, 5.3982E+03, 5.4111E+03, 5.4241E+03,
     +  5.4371E+03, 5.4501E+03, 5.4631E+03, 5.4762E+03, 5.4892E+03,
     + 5.5023E+03, 5.5153E+03, 5.5284E+03, 5.5415E+03, 5.5546E+03,
     +  5.5677E+03, 5.5809E+03, 5.5940E+03, 5.6072E+03, 5.6204E+03,
     + 5.6336E+03, 5.6468E+03, 5.6600E+03, 5.6733E+03, 5.6865E+03,
     +  5.6997E+03, 5.7130E+03, 5.7263E+03, 5.7396E+03, 5.7529E+03,
     + 5.7662E+03, 5.7795E+03, 5.7929E+03, 5.8063E+03, 5.8197E+03,
     +  5.8331E+03, 5.8464E+03, 5.8599E+03, 5.8733E+03, 5.8868E+03,
     + 5.9002E+03, 5.9137E+03, 5.9272E+03, 5.9406E+03, 5.9542E+03,
     +  5.9677E+03, 5.9812E+03, 5.9948E+03, 6.0083E+03, 6.0219E+03,
     + 6.0355E+03, 6.0491E+03, 6.0627E+03, 6.0764E+03, 6.0900E+03,
     +  6.1037E+03, 6.1174E+03, 6.1310E+03, 6.1447E+03, 6.1585E+03,
     + 6.1722E+03, 6.1859E+03, 6.1997E+03, 6.2134E+03, 6.2272E+03,
     +  6.2411E+03, 6.2548E+03, 6.2687E+03, 6.2825E+03, 6.2964E+03,
     + 6.3102E+03, 6.3241E+03, 6.3380E+03, 6.3519E+03, 6.3658E+03,
     +  6.3797E+03, 6.3937E+03, 6.4076E+03, 6.4216E+03, 6.4356E+03,
     + 6.4496E+03, 6.4636E+03, 6.4776E+03, 6.4917E+03, 6.5058E+03,
     +  6.5198E+03, 6.5339E+03, 6.5480E+03, 6.5621E+03, 6.5762E+03,
     + 6.5904E+03, 6.6045E+03, 6.6187E+03, 6.6329E+03, 6.6470E+03,
     +  6.6613E+03, 6.6755E+03, 6.6897E+03, 6.7039E+03, 6.7182E+03,
     + 6.7325E+03, 6.7467E+03, 6.7611E+03, 6.7754E+03, 6.7897E+03,
     +  6.8041E+03, 6.8184E+03, 6.8328E+03, 6.8472E+03, 6.8616E+03,
     + 6.8760E+03, 6.8905E+03, 6.9049E+03, 6.9193E+03, 6.9338E+03,
     +  6.9482E+03, 6.9628E+03, 6.9773E+03, 6.9918E+03, 7.0063E+03,
     + 7.0209E+03, 7.0355E+03, 7.0500E+03, 7.0646E+03, 7.0792E+03,
     +  7.0939E+03, 7.1085E+03, 7.1231E+03, 7.1378E+03, 7.1525E+03,
     + 7.1672E+03, 7.1819E+03, 7.1966E+03, 7.2113E+03, 7.2261E+03,
     +  7.2408E+03, 7.2556E+03, 7.2704E+03, 7.2852E+03, 7.3000E+03,
     + 7.3148E+03, 7.3297E+03, 7.3445E+03, 7.3594E+03, 7.3743E+03,
     +  7.3892E+03, 7.4041E+03, 7.4191E+03, 7.4340E+03, 7.4489E+03,
     + 7.4639E+03, 7.4789E+03, 7.4939E+03, 7.5089E+03, 7.5239E+03,
     +  7.5390E+03, 7.5540E+03, 7.5691E+03, 7.5842E+03, 7.5993E+03,
     + 7.6144E+03, 7.6295E+03, 7.6446E+03, 7.6598E+03, 7.6749E+03,
     +  7.6901E+03, 7.7053E+03, 7.7205E+03, 7.7358E+03, 7.7510E+03,
     + 7.7662E+03, 7.7815E+03, 7.7968E+03, 7.8121E+03, 7.8274E+03,
     +  7.8427E+03, 7.8580E+03, 7.8734E+03, 7.8888E+03, 7.9041E+03,
     + 7.9195E+03, 7.9350E+03, 7.9503E+03, 7.9658E+03, 7.9812E+03,
     +  7.9967E+03, 8.0122E+03, 8.0277E+03, 8.0432E+03, 8.0587E+03,
     + 8.0742E+03, 8.0897E+03, 8.1054E+03, 8.1209E+03, 8.1365E+03,
     +  8.1521E+03, 8.1677E+03, 8.1833E+03, 8.1990E+03, 8.2146E+03,
     + 8.2304E+03, 8.2460E+03, 8.2617E+03, 8.2774E+03, 8.2932E+03,
     +  8.3089E+03, 8.3247E+03, 8.3404E+03, 8.3562E+03, 8.3720E+03,
     + 8.3878E+03, 8.4037E+03, 8.4195E+03, 8.4354E+03, 8.4512E+03,
     +  8.4671E+03, 8.4830E+03, 8.4990E+03, 8.5149E+03, 8.5308E+03,
     + 8.5468E+03, 8.5628E+03, 8.5787E+03, 8.5948E+03, 8.6108E+03,
     +  8.6268E+03, 8.6428E+03, 8.6589E+03, 8.6750E+03, 8.6911E+03,
     + 8.7071E+03, 8.7233E+03, 8.7394E+03, 8.7556E+03, 8.7717E+03,
     +  8.7878E+03, 8.8040E+03, 8.8202E+03, 8.8365E+03, 8.8527E+03,
     + 8.8689E+03, 8.8852E+03, 8.9014E+03, 8.9178E+03, 8.9341E+03,
     +  8.9504E+03, 8.9667E+03, 8.9831E+03, 8.9994E+03, 9.0158E+03,
     + 9.0322E+03, 9.0486E+03, 9.0650E+03, 9.0814E+03, 9.0979E+03,
     +  9.1143E+03, 9.1308E+03, 9.1473E+03, 9.1638E+03, 9.1803E+03,
     + 9.1968E+03, 9.2134E+03, 9.2299E+03, 9.2465E+03, 9.2631E+03,
     +  9.2797E+03, 9.2963E+03, 9.3129E+03, 9.3295E+03, 9.3462E+03,
     + 9.3629E+03, 9.3796E+03, 9.3963E+03, 9.4130E+03, 9.4297E+03,
     +  9.4465E+03, 9.4633E+03, 9.4800E+03, 9.4968E+03, 9.5136E+03,
     + 9.5304E+03, 9.5472E+03, 9.5641E+03, 9.5809E+03, 9.5979E+03,
     +  9.6147E+03, 9.6316E+03, 9.6485E+03, 9.6655E+03, 9.6824E+03,
     + 9.6994E+03, 9.7164E+03, 9.7334E+03, 9.7504E+03, 9.7673E+03,
     +  9.7844E+03, 9.8014E+03, 9.8185E+03, 9.8356E+03, 9.8527E+03,
     + 9.8698E+03, 9.8869E+03, 9.9040E+03, 9.9212E+03, 9.9383E+03,
     +  9.9555E+03, 9.9727E+03, 9.9899E+03, 1.0007E+04, 1.0024E+04,
     + 1.0042E+04, 1.0059E+04, 1.0076E+04, 1.0093E+04, 1.0111E+04,
     +  1.0128E+04, 1.0145E+04, 1.0163E+04, 1.0180E+04, 1.0197E+04,
     + 1.0215E+04, 1.0232E+04, 1.0250E+04, 1.0267E+04, 1.0285E+04,
     +  1.0302E+04, 1.0320E+04, 1.0337E+04, 1.0355E+04, 1.0372E+04,
     + 1.0390E+04, 1.0407E+04, 1.0425E+04, 1.0442E+04, 1.0460E+04,
     +  1.0478E+04, 1.0495E+04, 1.0513E+04, 1.0530E+04, 1.0548E+04,
     + 1.0566E+04, 1.0583E+04, 1.0601E+04, 1.0619E+04, 1.0637E+04,
     +  1.0654E+04, 1.0672E+04, 1.0690E+04, 1.0708E+04, 1.0726E+04,
     + 1.0743E+04, 1.0761E+04, 1.0779E+04, 1.0797E+04, 1.0815E+04,
     +  1.0833E+04, 1.0851E+04, 1.0869E+04, 1.0887E+04, 1.0905E+04,
     + 1.0923E+04, 1.0940E+04, 1.0958E+04, 1.0977E+04, 1.0994E+04,
     +  1.1013E+04, 1.1031E+04, 1.1049E+04, 1.1067E+04, 1.1085E+04,
     + 1.1103E+04, 1.1121E+04, 1.1139E+04, 1.1157E+04, 1.1176E+04,
     +  1.1194E+04, 1.1212E+04, 1.1230E+04, 1.1248E+04, 1.1267E+04,
     + 1.1285E+04, 1.1303E+04, 1.1322E+04, 1.1340E+04, 1.1358E+04,
     +  1.1377E+04, 1.1395E+04, 1.1413E+04, 1.1432E+04, 1.1450E+04,
     + 1.1468E+04, 1.1487E+04, 1.1505E+04, 1.1524E+04, 1.1542E+04,
     +  1.1561E+04, 1.1579E+04, 1.1598E+04, 1.1616E+04, 1.1635E+04,
     + 1.1653E+04, 1.1672E+04, 1.1690E+04, 1.1709E+04, 1.1728E+04,
     +  1.1746E+04, 1.1765E+04, 1.1784E+04, 1.1802E+04, 1.1821E+04,
     + 1.1840E+04, 1.1858E+04, 1.1877E+04, 1.1896E+04, 1.1915E+04,
     +  1.1933E+04, 1.1952E+04, 1.1971E+04, 1.1990E+04, 1.2009E+04,
     + 1.2027E+04, 1.2046E+04, 1.2065E+04, 1.2084E+04, 1.2103E+04,
     +  1.2122E+04, 1.2141E+04, 1.2160E+04, 1.2179E+04, 1.2198E+04,
     + 1.2217E+04, 1.2236E+04, 1.2255E+04, 1.2274E+04, 1.2293E+04,
     +  1.2312E+04, 1.2331E+04, 1.2350E+04, 1.2369E+04, 1.2388E+04,
     + 1.2407E+04, 1.2426E+04, 1.2446E+04, 1.2465E+04, 1.2484E+04,
     +  1.2503E+04, 1.2523E+04, 1.2542E+04, 1.2561E+04, 1.2580E+04,
     + 1.2600E+04, 1.2619E+04, 1.2638E+04, 1.2657E+04, 1.2677E+04,
     +  1.2696E+04, 1.2716E+04, 1.2735E+04, 1.2754E+04, 1.2774E+04,
     + 1.2793E+04, 1.2813E+04, 1.2832E+04, 1.2852E+04, 1.2871E+04,
     +  1.2891E+04, 1.2910E+04, 1.2930E+04, 1.2949E+04, 1.2969E+04,
     + 1.2988E+04, 1.3008E+04, 1.3027E+04, 1.3047E+04, 1.3067E+04,
     +  1.3086E+04, 1.3106E+04, 1.3126E+04, 1.3145E+04, 1.3165E+04,
     + 1.3185E+04, 1.3205E+04, 1.3224E+04, 1.3244E+04, 1.3264E+04,
     +  1.3284E+04, 1.3303E+04, 1.3323E+04, 1.3343E+04, 1.3363E+04,
     + 1.3383E+04, 1.3403E+04, 1.3423E+04, 1.3443E+04, 1.3463E+04,
     +  1.3482E+04, 1.3502E+04, 1.3522E+04, 1.3542E+04, 1.3562E+04,
     + 1.3582E+04, 1.3602E+04, 1.3622E+04, 1.3642E+04, 1.3663E+04,
     +  1.3683E+04, 1.3703E+04, 1.3723E+04, 1.3743E+04, 1.3763E+04,
     + 1.3783E+04, 1.3803E+04, 1.3824E+04, 1.3844E+04, 1.3864E+04,
     +  1.3884E+04, 1.3905E+04, 1.3925E+04, 1.3945E+04, 1.3965E+04,
     + 1.3986E+04, 1.4006E+04, 1.4026E+04, 1.4047E+04, 1.4067E+04,
     +  1.4088E+04, 1.4108E+04, 1.4128E+04, 1.4149E+04, 1.4169E+04,
     + 1.4190E+04, 1.4210E+04, 1.4231E+04, 1.4251E+04, 1.4272E+04,
     +  1.4292E+04, 1.4313E+04, 1.4333E+04, 1.4354E+04, 1.4374E+04,
     + 1.4395E+04, 1.4416E+04, 1.4436E+04, 1.4457E+04, 1.4478E+04,
     +  1.4498E+04, 1.4519E+04, 1.4540E+04, 1.4560E+04, 1.4581E+04,
     + 1.4602E+04, 1.4623E+04, 1.4643E+04, 1.4664E+04, 1.4685E+04,
     +  1.4706E+04, 1.4727E+04, 1.4747E+04, 1.4768E+04, 1.4789E+04,
     + 1.4810E+04, 1.4831E+04, 1.4852E+04, 1.4873E+04, 1.4894E+04,
     +  1.4915E+04, 1.4936E+04, 1.4957E+04, 1.4978E+04, 1.4999E+04,
     + 1.5020E+04, 1.5041E+04, 1.5062E+04, 1.5083E+04, 1.5104E+04,
     +  1.5125E+04, 1.5147E+04, 1.5168E+04, 1.5189E+04, 1.5210E+04,
     + 1.5231E+04, 1.5252E+04, 1.5274E+04, 1.5295E+04, 1.5316E+04,
     +  1.5337E+04, 1.5359E+04, 1.5380E+04, 1.5401E+04, 1.5423E+04,
     + 1.5444E+04, 1.5465E+04, 1.5487E+04, 1.5508E+04, 1.5529E+04,
     +  1.5551E+04, 1.5572E+04, 1.5594E+04, 1.5615E+04, 1.5637E+04,
     + 1.5658E+04, 1.5680E+04, 1.5701E+04, 1.5723E+04, 1.5744E+04,
     +  1.5766E+04, 1.5787E+04, 1.5809E+04, 1.5831E+04, 1.5852E+04,
     + 1.5874E+04, 1.5895E+04, 1.5917E+04, 1.5939E+04, 1.5960E+04,
     +  1.5982E+04, 1.6004E+04, 1.6026E+04, 1.6047E+04, 1.6069E+04,
     + 1.6091E+04, 1.6113E+04, 1.6135E+04, 1.6156E+04, 1.6178E+04,
     +  1.6200E+04, 1.6222E+04, 1.6244E+04, 1.6266E+04, 1.6288E+04,
     + 1.6310E+04, 1.6331E+04, 1.6353E+04, 1.6375E+04, 1.6397E+04,
     +  1.6419E+04, 1.6441E+04, 1.6463E+04, 1.6485E+04, 1.6508E+04,
     + 1.6530E+04, 1.6552E+04, 1.6574E+04, 1.6596E+04, 1.6618E+04,
     +  1.6640E+04, 1.6662E+04, 1.6685E+04, 1.6707E+04, 1.6729E+04,
     + 1.6751E+04, 1.6773E+04, 1.6796E+04, 1.6818E+04, 1.6840E+04,
     +  1.6863E+04, 1.6885E+04, 1.6907E+04, 1.6929E+04, 1.6952E+04,
     + 1.6974E+04, 1.6997E+04, 1.7019E+04, 1.7041E+04, 1.7064E+04,
     +  1.7086E+04, 1.7109E+04, 1.7131E+04, 1.7154E+04, 1.7176E+04,
     + 1.7199E+04, 1.7221E+04, 1.7244E+04, 1.7266E+04, 1.7289E+04,
     +  1.7312E+04, 1.7334E+04, 1.7357E+04, 1.7379E+04, 1.7402E+04,
     + 1.7425E+04, 1.7447E+04, 1.7470E+04, 1.7493E+04, 1.7516E+04,
     +  1.7538E+04, 1.7561E+04, 1.7584E+04, 1.7607E+04, 1.7630E+04,
     + 1.7652E+04, 1.7675E+04, 1.7698E+04, 1.7721E+04, 1.7744E+04,
     +  1.7767E+04, 1.7789E+04, 1.7812E+04, 1.7835E+04, 1.7858E+04,
     + 1.7881E+04, 1.7904E+04, 1.7927E+04, 1.7950E+04, 1.7973E+04,
     +  1.7996E+04, 1.8019E+04, 1.8042E+04, 1.8065E+04, 1.8089E+04,
     + 1.8112E+04, 1.8135E+04, 1.8158E+04, 1.8181E+04, 1.8204E+04,
     +  1.8227E+04, 1.8251E+04, 1.8274E+04, 1.8297E+04, 1.8320E+04,
     + 1.8344E+04, 1.8367E+04, 1.8390E+04, 1.8413E+04, 1.8437E+04,
     +  1.8460E+04, 1.8484E+04, 1.8507E+04, 1.8530E+04, 1.8554E+04,
     + 1.8577E+04, 1.8600E+04, 1.8624E+04, 1.8647E+04, 1.8671E+04,
     +  1.8694E+04, 1.8718E+04, 1.8741E+04, 1.8765E+04, 1.8788E+04,
     + 1.8812E+04, 1.8836E+04, 1.8859E+04, 1.8883E+04, 1.8906E+04,
     +  1.8930E+04, 1.8954E+04, 1.8977E+04, 1.9001E+04, 1.9025E+04,
     + 1.9048E+04, 1.9072E+04, 1.9096E+04, 1.9120E+04, 1.9143E+04,
     +  1.9167E+04, 1.9191E+04, 1.9215E+04, 1.9239E+04, 1.9262E+04,
     + 1.9286E+04, 1.9310E+04, 1.9334E+04, 1.9358E+04, 1.9382E+04,
     +  1.9406E+04, 1.9430E+04, 1.9454E+04, 1.9478E+04, 1.9502E+04,
     + 1.9526E+04, 1.9549E+04, 1.9574E+04, 1.9598E+04, 1.9622E+04,
     +  1.9646E+04, 1.9670E+04, 1.9694E+04, 1.9718E+04, 1.9742E+04,
     + 1.9766E+04, 1.9791E+04, 1.9815E+04, 1.9839E+04, 1.9863E+04,
     +  1.9887E+04, 1.9912E+04, 1.9936E+04, 1.9960E+04, 1.9984E+04,
     + 2.0009E+04, 2.0033E+04, 2.0057E+04, 2.0082E+04, 2.0106E+04,
     +  2.0130E+04, 2.0155E+04, 2.0179E+04, 2.0204E+04, 2.0228E+04,
     + 2.0252E+04, 2.0277E+04, 2.0301E+04, 2.0326E+04, 2.0350E+04,
     +  2.0375E+04, 2.0399E+04, 2.0424E+04, 2.0449E+04, 2.0473E+04,
     + 2.0498E+04, 2.0522E+04, 2.0547E+04, 2.0572E+04, 2.0596E+04,
     +  2.0621E+04, 2.0646E+04, 2.0670E+04, 2.0695E+04, 2.0720E+04,
     + 2.0744E+04, 2.0769E+04, 2.0794E+04, 2.0819E+04, 2.0844E+04,
     +  2.0868E+04, 2.0893E+04, 2.0918E+04, 2.0943E+04, 2.0968E+04,
     + 2.0993E+04, 2.1018E+04, 2.1043E+04, 2.1068E+04, 2.1093E+04,
     +  2.1117E+04, 2.1142E+04, 2.1167E+04, 2.1192E+04, 2.1218E+04,
     + 2.1242E+04, 2.1268E+04, 2.1293E+04, 2.1318E+04, 2.1343E+04,
     +  2.1368E+04, 2.1393E+04, 2.1418E+04, 2.1443E+04, 2.1469E+04,
     + 2.1494E+04, 2.1519E+04, 2.1544E+04, 2.1569E+04, 2.1595E+04,
     +  2.1620E+04, 2.1645E+04, 2.1670E+04, 2.1696E+04, 2.1721E+04,
     + 2.1746E+04, 2.1772E+04, 2.1797E+04, 2.1823E+04, 2.1848E+04,
     +  2.1873E+04, 2.1899E+04, 2.1924E+04, 2.1950E+04, 2.1975E+04,
     + 2.2001E+04, 2.2026E+04, 2.2052E+04, 2.2077E+04, 2.2103E+04,
     +  2.2128E+04, 2.2154E+04, 2.2180E+04, 2.2205E+04, 2.2231E+04,
     + 2.2256E+04, 2.2282E+04, 2.2308E+04, 2.2333E+04, 2.2359E+04,
     +  2.2385E+04, 2.2411E+04, 2.2436E+04, 2.2462E+04, 2.2488E+04,
     + 2.2514E+04, 2.2539E+04, 2.2565E+04, 2.2591E+04, 2.2617E+04,
     +  2.2643E+04, 2.2669E+04, 2.2694E+04, 2.2720E+04, 2.2746E+04,
     + 2.2772E+04, 2.2798E+04, 2.2824E+04, 2.2850E+04, 2.2876E+04,
     +  2.2902E+04, 2.2928E+04, 2.2954E+04, 2.2980E+04, 2.3006E+04,
     + 2.3032E+04, 2.3058E+04, 2.3085E+04, 2.3111E+04, 2.3137E+04,
     +  2.3163E+04, 2.3189E+04, 2.3215E+04, 2.3242E+04, 2.3268E+04,
     + 2.3294E+04, 2.3320E+04, 2.3346E+04, 2.3373E+04, 2.3399E+04,
     +  2.3425E+04, 2.3452E+04, 2.3478E+04, 2.3504E+04, 2.3531E+04,
     + 2.3557E+04, 2.3584E+04, 2.3610E+04, 2.3636E+04, 2.3663E+04,
     +  2.3689E+04, 2.3716E+04, 2.3742E+04, 2.3769E+04, 2.3795E+04,
     + 2.3822E+04, 2.3848E+04, 2.3875E+04, 2.3901E+04, 2.3928E+04,
     +  2.3955E+04, 2.3981E+04, 2.4008E+04, 2.4035E+04, 2.4061E+04,
     + 2.4088E+04, 2.4115E+04, 2.4141E+04, 2.4168E+04, 2.4195E+04,
     +  2.4222E+04, 2.4248E+04, 2.4275E+04, 2.4302E+04, 2.4329E+04,
     + 2.4356E+04, 2.4382E+04, 2.4409E+04, 2.4436E+04, 2.4463E+04,
     +  2.4490E+04, 2.4517E+04, 2.4544E+04, 2.4571E+04, 2.4598E+04,
     + 2.4625E+04, 2.4652E+04, 2.4679E+04, 2.4706E+04, 2.4733E+04,
     +  2.4760E+04, 2.4787E+04, 2.4814E+04, 2.4841E+04, 2.4868E+04,
     + 2.4895E+04, 2.4922E+04, 2.4949E+04, 2.4977E+04, 2.5004E+04,
     +  2.5031E+04, 2.5058E+04, 2.5086E+04, 2.5113E+04, 2.5140E+04,
     + 2.5167E+04, 2.5195E+04, 2.5222E+04, 2.5249E+04, 2.5277E+04,
     +  2.5304E+04, 2.5331E+04, 2.5359E+04, 2.5386E+04, 2.5413E+04,
     + 2.5441E+04, 2.5468E+04, 2.5496E+04, 2.5523E+04, 2.5551E+04,
     +  2.5578E+04, 2.5606E+04, 2.5633E+04, 2.5661E+04, 2.5688E+04,
     + 2.5716E+04, 2.5744E+04, 2.5771E+04, 2.5799E+04, 2.5826E+04,
     +  2.5854E+04, 2.5882E+04, 2.5909E+04, 2.5937E+04, 2.5965E+04,
     + 2.5993E+04, 2.6020E+04, 2.6048E+04, 2.6076E+04, 2.6104E+04,
     +  2.6131E+04, 2.6159E+04, 2.6187E+04, 2.6215E+04, 2.6243E+04,
     + 2.6271E+04, 2.6298E+04, 2.6326E+04, 2.6354E+04, 2.6382E+04,
     +  2.6410E+04, 2.6438E+04, 2.6466E+04, 2.6494E+04, 2.6522E+04,
     + 2.6550E+04, 2.6578E+04, 2.6606E+04, 2.6634E+04, 2.6662E+04,
     +  2.6691E+04, 2.6719E+04, 2.6747E+04, 2.6775E+04, 2.6803E+04,
     + 2.6831E+04, 2.6859E+04, 2.6888E+04, 2.6916E+04, 2.6944E+04,
     +  2.6972E+04, 2.7001E+04, 2.7029E+04, 2.7057E+04, 2.7085E+04,
     + 2.7114E+04, 2.7142E+04, 2.7170E+04, 2.7199E+04, 2.7227E+04,
     +  2.7256E+04, 2.7284E+04, 2.7312E+04, 2.7341E+04, 2.7369E+04,
     + 2.7398E+04, 2.7426E+04, 2.7455E+04, 2.7483E+04, 2.7512E+04,
     +  2.7540E+04, 2.7569E+04, 2.7598E+04, 2.7626E+04, 2.7655E+04,
     + 2.7683E+04, 2.7712E+04, 2.7741E+04, 2.7769E+04, 2.7798E+04,
     +  2.7827E+04, 2.7855E+04, 2.7884E+04, 2.7913E+04, 2.7942E+04,
     + 2.7970E+04, 2.7999E+04, 2.8028E+04, 2.8057E+04, 2.8086E+04,
     +  2.8114E+04, 2.8143E+04, 2.8172E+04, 2.8201E+04, 2.8230E+04,
     + 2.8259E+04, 2.8288E+04, 2.8317E+04, 2.8346E+04, 2.8375E+04,
     +  2.8404E+04, 2.8433E+04, 2.8462E+04, 2.8491E+04, 2.8520E+04,
     + 2.8549E+04, 2.8578E+04, 2.8607E+04, 2.8636E+04, 2.8665E+04,
     +  2.8695E+04, 2.8724E+04, 2.8753E+04, 2.8782E+04, 2.8811E+04,
     + 2.8841E+04, 2.8870E+04, 2.8899E+04, 2.8928E+04, 2.8958E+04,
     +  2.8987E+04, 2.9016E+04, 2.9045E+04, 2.9075E+04, 2.9104E+04,
     + 2.9134E+04, 2.9163E+04, 2.9192E+04, 2.9222E+04, 2.9251E+04,
     +  2.9281E+04, 2.9310E+04, 2.9339E+04, 2.9369E+04, 2.9399E+04,
     + 2.9428E+04, 2.9458E+04, 2.9487E+04, 2.9517E+04, 2.9546E+04,
     +  2.9576E+04, 2.9605E+04, 2.9635E+04, 2.9665E+04, 2.9694E+04,
     + 2.9724E+04, 2.9754E+04, 2.9783E+04, 2.9813E+04, 2.9843E+04,
     +  2.9873E+04, 2.9902E+04, 2.9932E+04, 2.9962E+04, 2.9992E+04,
     + 3.0022E+04, 3.0051E+04, 3.0081E+04, 3.0111E+04, 3.0141E+04,
     +  3.0171E+04, 3.0201E+04, 3.0231E+04, 3.0261E+04, 3.0291E+04,
     + 3.0321E+04, 3.0351E+04, 3.0381E+04, 3.0411E+04, 3.0441E+04,
     +  3.0471E+04, 3.0501E+04, 3.0531E+04, 3.0561E+04, 3.0591E+04,
     + 3.0621E+04, 3.0651E+04, 3.0681E+04, 3.0711E+04, 3.0742E+04,
     +  3.0772E+04, 3.0802E+04, 3.0832E+04, 3.0863E+04, 3.0893E+04,
     + 3.0923E+04, 3.0953E+04, 3.0984E+04, 3.1014E+04, 3.1044E+04,
     +  3.1075E+04, 3.1105E+04, 3.1135E+04, 3.1166E+04, 3.1196E+04,
     + 3.1227E+04, 3.1257E+04, 3.1287E+04, 3.1318E+04, 3.1348E+04,
     +  3.1379E+04, 3.1409E+04, 3.1440E+04, 3.1470E+04, 3.1501E+04,
     + 3.1532E+04, 3.1562E+04, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        56
      data (QofT( 2,J),J=1,2408)/ 7.9413E+00, 8.4319E+00, 9.5361E+00,
     +  1.0934E+01, 1.2457E+01, 1.4042E+01, 1.5661E+01, 1.7302E+01,
     + 1.8955E+01, 2.0619E+01, 2.2289E+01, 2.3964E+01, 2.5643E+01,
     +  2.7325E+01, 2.9009E+01, 3.0696E+01, 3.2384E+01, 3.4074E+01,
     + 3.5766E+01, 3.7459E+01, 4.0852E+01, 4.4253E+01, 4.7667E+01,
     +  5.1096E+01, 5.4544E+01, 5.8016E+01, 6.1516E+01, 6.5047E+01,
     + 6.8612E+01, 7.2216E+01, 7.5861E+01, 7.9550E+01, 8.3285E+01,
     +  8.7067E+01, 9.0899E+01, 9.4782E+01, 9.8715E+01, 1.0270E+02,
     + 1.0674E+02, 1.1083E+02, 1.1497E+02, 1.1916E+02, 1.2341E+02,
     +  1.2771E+02, 1.3205E+02, 1.3645E+02, 1.4090E+02, 1.4539E+02,
     + 1.4994E+02, 1.5453E+02, 1.5916E+02, 1.6384E+02, 1.6856E+02,
     +  1.7333E+02, 1.7814E+02, 1.8299E+02, 1.8788E+02, 1.9280E+02,
     + 1.9777E+02, 2.0277E+02, 2.0781E+02, 2.1289E+02, 2.1811E+02,
     +  2.2332E+02, 2.2856E+02, 2.3383E+02, 2.3913E+02, 2.4446E+02,
     + 2.4983E+02, 2.5521E+02, 2.6063E+02, 2.6608E+02, 2.7155E+02,
     +  2.7704E+02, 2.8256E+02, 2.8811E+02, 2.9367E+02, 2.9927E+02,
     + 3.0488E+02, 3.1051E+02, 3.1617E+02, 3.2185E+02, 3.2754E+02,
     +  3.3326E+02, 3.3900E+02, 3.4475E+02, 3.5052E+02, 3.5631E+02,
     + 3.6212E+02, 3.6795E+02, 3.7379E+02, 3.7965E+02, 3.8552E+02,
     +  3.9141E+02, 3.9731E+02, 4.0323E+02, 4.0917E+02, 4.1511E+02,
     + 4.2107E+02, 4.2705E+02, 4.3304E+02, 4.3904E+02, 4.4505E+02,
     +  4.5107E+02, 4.5711E+02, 4.6316E+02, 4.6922E+02, 4.7529E+02,
     + 4.8137E+02, 4.8746E+02, 4.9357E+02, 4.9968E+02, 5.0580E+02,
     +  5.1194E+02, 5.1808E+02, 5.2423E+02, 5.3039E+02, 5.3657E+02,
     + 5.4275E+02, 5.4894E+02, 5.5514E+02, 5.6134E+02, 5.6756E+02,
     +  5.7378E+02, 5.8001E+02, 5.8624E+02, 5.9249E+02, 5.9874E+02,
     + 6.0500E+02, 6.1127E+02, 6.1755E+02, 6.2383E+02, 6.3011E+02,
     +  6.3641E+02, 6.4271E+02, 6.4902E+02, 6.5534E+02, 6.6165E+02,
     + 6.6798E+02, 6.7432E+02, 6.8066E+02, 6.8700E+02, 6.9335E+02,
     +  6.9971E+02, 7.0607E+02, 7.1244E+02, 7.1881E+02, 7.2519E+02,
     + 7.3157E+02, 7.3797E+02, 7.4436E+02, 7.5076E+02, 7.5716E+02,
     +  7.6357E+02, 7.6999E+02, 7.7640E+02, 7.8283E+02, 7.8926E+02,
     + 7.9570E+02, 8.0214E+02, 8.0857E+02, 8.1502E+02, 8.2147E+02,
     +  8.2793E+02, 8.3439E+02, 8.4085E+02, 8.4733E+02, 8.5380E+02,
     + 8.6028E+02, 8.6676E+02, 8.7325E+02, 8.7974E+02, 8.8623E+02,
     +  8.9273E+02, 8.9923E+02, 9.0574E+02, 9.1225E+02, 9.1876E+02,
     + 9.2528E+02, 9.3181E+02, 9.3833E+02, 9.4487E+02, 9.5140E+02,
     +  9.5794E+02, 9.6448E+02, 9.7103E+02, 9.7758E+02, 9.8413E+02,
     + 9.9069E+02, 9.9725E+02, 1.0038E+03, 1.0104E+03, 1.0170E+03,
     +  1.0235E+03, 1.0301E+03, 1.0367E+03, 1.0433E+03, 1.0499E+03,
     + 1.0565E+03, 1.0631E+03, 1.0697E+03, 1.0763E+03, 1.0829E+03,
     +  1.0895E+03, 1.0961E+03, 1.1027E+03, 1.1094E+03, 1.1160E+03,
     + 1.1226E+03, 1.1293E+03, 1.1359E+03, 1.1425E+03, 1.1492E+03,
     +  1.1559E+03, 1.1625E+03, 1.1692E+03, 1.1758E+03, 1.1825E+03,
     + 1.1892E+03, 1.1958E+03, 1.2025E+03, 1.2092E+03, 1.2159E+03,
     +  1.2226E+03, 1.2293E+03, 1.2360E+03, 1.2427E+03, 1.2494E+03,
     + 1.2561E+03, 1.2628E+03, 1.2695E+03, 1.2763E+03, 1.2830E+03,
     +  1.2897E+03, 1.2964E+03, 1.3032E+03, 1.3099E+03, 1.3167E+03,
     + 1.3234E+03, 1.3302E+03, 1.3369E+03, 1.3437E+03, 1.3505E+03,
     +  1.3572E+03, 1.3640E+03, 1.3708E+03, 1.3776E+03, 1.3844E+03,
     + 1.3912E+03, 1.3980E+03, 1.4048E+03, 1.4116E+03, 1.4184E+03,
     +  1.4252E+03, 1.4320E+03, 1.4389E+03, 1.4457E+03, 1.4525E+03,
     + 1.4594E+03, 1.4662E+03, 1.4731E+03, 1.4799E+03, 1.4868E+03,
     +  1.4936E+03, 1.5005E+03, 1.5074E+03, 1.5143E+03, 1.5211E+03,
     + 1.5280E+03, 1.5349E+03, 1.5418E+03, 1.5487E+03, 1.5556E+03,
     +  1.5625E+03, 1.5695E+03, 1.5764E+03, 1.5833E+03, 1.5902E+03,
     + 1.5972E+03, 1.6041E+03, 1.6111E+03, 1.6180E+03, 1.6250E+03,
     +  1.6320E+03, 1.6389E+03, 1.6459E+03, 1.6529E+03, 1.6599E+03,
     + 1.6669E+03, 1.6739E+03, 1.6809E+03, 1.6879E+03, 1.6949E+03,
     +  1.7019E+03, 1.7089E+03, 1.7159E+03, 1.7230E+03, 1.7300E+03,
     + 1.7371E+03, 1.7441E+03, 1.7512E+03, 1.7582E+03, 1.7653E+03,
     +  1.7724E+03, 1.7795E+03, 1.7866E+03, 1.7937E+03, 1.8008E+03,
     + 1.8079E+03, 1.8150E+03, 1.8221E+03, 1.8292E+03, 1.8363E+03,
     +  1.8435E+03, 1.8506E+03, 1.8578E+03, 1.8649E+03, 1.8721E+03,
     + 1.8792E+03, 1.8864E+03, 1.8936E+03, 1.9008E+03, 1.9079E+03,
     +  1.9151E+03, 1.9223E+03, 1.9295E+03, 1.9368E+03, 1.9440E+03,
     + 1.9512E+03, 1.9584E+03, 1.9657E+03, 1.9729E+03, 1.9802E+03,
     +  1.9874E+03, 1.9947E+03, 2.0020E+03, 2.0093E+03, 2.0166E+03,
     + 2.0238E+03, 2.0311E+03, 2.0385E+03, 2.0458E+03, 2.0531E+03,
     +  2.0604E+03, 2.0677E+03, 2.0751E+03, 2.0824E+03, 2.0898E+03,
     + 2.0971E+03, 2.1045E+03, 2.1119E+03, 2.1192E+03, 2.1266E+03,
     +  2.1340E+03, 2.1414E+03, 2.1488E+03, 2.1563E+03, 2.1637E+03,
     + 2.1711E+03, 2.1785E+03, 2.1860E+03, 2.1934E+03, 2.2009E+03,
     +  2.2084E+03, 2.2158E+03, 2.2233E+03, 2.2308E+03, 2.2383E+03,
     + 2.2458E+03, 2.2533E+03, 2.2608E+03, 2.2683E+03, 2.2759E+03,
     +  2.2834E+03, 2.2909E+03, 2.2985E+03, 2.3060E+03, 2.3136E+03,
     + 2.3212E+03, 2.3288E+03, 2.3363E+03, 2.3439E+03, 2.3515E+03,
     +  2.3592E+03, 2.3668E+03, 2.3744E+03, 2.3820E+03, 2.3897E+03,
     + 2.3973E+03, 2.4050E+03, 2.4126E+03, 2.4203E+03, 2.4280E+03,
     +  2.4357E+03, 2.4434E+03, 2.4511E+03, 2.4588E+03, 2.4665E+03,
     + 2.4742E+03, 2.4819E+03, 2.4897E+03, 2.4974E+03, 2.5052E+03,
     +  2.5129E+03, 2.5207E+03, 2.5285E+03, 2.5363E+03, 2.5441E+03,
     + 2.5519E+03, 2.5597E+03, 2.5675E+03, 2.5753E+03, 2.5832E+03,
     +  2.5910E+03, 2.5989E+03, 2.6067E+03, 2.6146E+03, 2.6225E+03,
     + 2.6303E+03, 2.6382E+03, 2.6461E+03, 2.6540E+03, 2.6619E+03,
     +  2.6699E+03, 2.6778E+03, 2.6857E+03, 2.6937E+03, 2.7016E+03,
     + 2.7096E+03, 2.7176E+03, 2.7255E+03, 2.7335E+03, 2.7415E+03,
     +  2.7495E+03, 2.7575E+03, 2.7656E+03, 2.7736E+03, 2.7816E+03,
     + 2.7897E+03, 2.7977E+03, 2.8058E+03, 2.8139E+03, 2.8219E+03,
     +  2.8300E+03, 2.8381E+03, 2.8462E+03, 2.8543E+03, 2.8625E+03,
     + 2.8706E+03, 2.8787E+03, 2.8869E+03, 2.8950E+03, 2.9032E+03,
     +  2.9114E+03, 2.9195E+03, 2.9277E+03, 2.9359E+03, 2.9441E+03,
     + 2.9523E+03, 2.9606E+03, 2.9688E+03, 2.9770E+03, 2.9853E+03,
     +  2.9935E+03, 3.0018E+03, 3.0101E+03, 3.0184E+03, 3.0267E+03,
     + 3.0349E+03, 3.0432E+03, 3.0516E+03, 3.0599E+03, 3.0682E+03,
     +  3.0766E+03, 3.0849E+03, 3.0933E+03, 3.1017E+03, 3.1100E+03,
     + 3.1184E+03, 3.1268E+03, 3.1352E+03, 3.1436E+03, 3.1520E+03,
     +  3.1605E+03, 3.1689E+03, 3.1774E+03, 3.1858E+03, 3.1943E+03,
     + 3.2027E+03, 3.2112E+03, 3.2197E+03, 3.2282E+03, 3.2367E+03,
     +  3.2452E+03, 3.2538E+03, 3.2623E+03, 3.2709E+03, 3.2794E+03,
     + 3.2880E+03, 3.2966E+03, 3.3051E+03, 3.3137E+03, 3.3223E+03,
     +  3.3309E+03, 3.3395E+03, 3.3482E+03, 3.3568E+03, 3.3654E+03,
     + 3.3741E+03, 3.3827E+03, 3.3914E+03, 3.4001E+03, 3.4088E+03,
     +  3.4175E+03, 3.4262E+03, 3.4349E+03, 3.4436E+03, 3.4524E+03,
     + 3.4611E+03, 3.4699E+03, 3.4786E+03, 3.4874E+03, 3.4962E+03,
     +  3.5050E+03, 3.5138E+03, 3.5226E+03, 3.5314E+03, 3.5402E+03,
     + 3.5490E+03, 3.5579E+03, 3.5667E+03, 3.5756E+03, 3.5845E+03,
     +  3.5933E+03, 3.6022E+03, 3.6111E+03, 3.6200E+03, 3.6290E+03,
     + 3.6379E+03, 3.6468E+03, 3.6557E+03, 3.6647E+03, 3.6737E+03,
     +  3.6826E+03, 3.6916E+03, 3.7006E+03, 3.7096E+03, 3.7186E+03,
     + 3.7276E+03, 3.7367E+03, 3.7457E+03, 3.7547E+03, 3.7638E+03,
     +  3.7729E+03, 3.7819E+03, 3.7910E+03, 3.8001E+03, 3.8092E+03,
     + 3.8183E+03, 3.8274E+03, 3.8366E+03, 3.8457E+03, 3.8549E+03,
     +  3.8640E+03, 3.8732E+03, 3.8823E+03, 3.8915E+03, 3.9007E+03,
     + 3.9099E+03, 3.9192E+03, 3.9284E+03, 3.9376E+03, 3.9469E+03,
     +  3.9561E+03, 3.9654E+03, 3.9746E+03, 3.9839E+03, 3.9932E+03,
     + 4.0025E+03, 4.0118E+03, 4.0211E+03, 4.0305E+03, 4.0398E+03,
     +  4.0491E+03, 4.0585E+03, 4.0679E+03, 4.0772E+03, 4.0866E+03,
     + 4.0960E+03, 4.1054E+03, 4.1148E+03, 4.1242E+03, 4.1337E+03,
     +  4.1431E+03, 4.1526E+03, 4.1620E+03, 4.1715E+03, 4.1810E+03,
     + 4.1905E+03, 4.1999E+03, 4.2094E+03, 4.2190E+03, 4.2285E+03,
     +  4.2380E+03, 4.2476E+03, 4.2571E+03, 4.2667E+03, 4.2763E+03,
     + 4.2859E+03, 4.2955E+03, 4.3051E+03, 4.3147E+03, 4.3243E+03,
     +  4.3339E+03, 4.3436E+03, 4.3532E+03, 4.3629E+03, 4.3725E+03,
     + 4.3822E+03, 4.3919E+03, 4.4016E+03, 4.4113E+03, 4.4210E+03,
     +  4.4308E+03, 4.4405E+03, 4.4502E+03, 4.4600E+03, 4.4698E+03,
     + 4.4795E+03, 4.4894E+03, 4.4991E+03, 4.5089E+03, 4.5188E+03,
     +  4.5286E+03, 4.5384E+03, 4.5483E+03, 4.5581E+03, 4.5680E+03,
     + 4.5778E+03, 4.5877E+03, 4.5976E+03, 4.6075E+03, 4.6174E+03,
     +  4.6274E+03, 4.6373E+03, 4.6472E+03, 4.6572E+03, 4.6671E+03,
     + 4.6771E+03, 4.6871E+03, 4.6971E+03, 4.7071E+03, 4.7171E+03,
     +  4.7271E+03, 4.7371E+03, 4.7472E+03, 4.7572E+03, 4.7673E+03,
     + 4.7773E+03, 4.7874E+03, 4.7975E+03, 4.8076E+03, 4.8177E+03,
     +  4.8278E+03, 4.8379E+03, 4.8481E+03, 4.8582E+03, 4.8684E+03,
     + 4.8785E+03, 4.8887E+03, 4.8989E+03, 4.9091E+03, 4.9193E+03,
     +  4.9295E+03, 4.9398E+03, 4.9500E+03, 4.9603E+03, 4.9705E+03,
     + 4.9807E+03, 4.9910E+03, 5.0013E+03, 5.0116E+03, 5.0219E+03,
     +  5.0322E+03, 5.0425E+03, 5.0529E+03, 5.0632E+03, 5.0736E+03,
     + 5.0839E+03, 5.0943E+03, 5.1047E+03, 5.1151E+03, 5.1255E+03,
     +  5.1359E+03, 5.1463E+03, 5.1567E+03, 5.1672E+03, 5.1776E+03,
     + 5.1881E+03, 5.1986E+03, 5.2090E+03, 5.2195E+03, 5.2300E+03,
     +  5.2405E+03, 5.2510E+03, 5.2616E+03, 5.2721E+03, 5.2827E+03,
     + 5.2932E+03, 5.3038E+03, 5.3144E+03, 5.3250E+03, 5.3356E+03,
     +  5.3462E+03, 5.3568E+03, 5.3674E+03, 5.3781E+03, 5.3887E+03,
     + 5.3994E+03, 5.4100E+03, 5.4207E+03, 5.4314E+03, 5.4421E+03,
     +  5.4528E+03, 5.4635E+03, 5.4743E+03, 5.4850E+03, 5.4958E+03,
     + 5.5065E+03, 5.5173E+03, 5.5281E+03, 5.5388E+03, 5.5496E+03,
     +  5.5605E+03, 5.5713E+03, 5.5821E+03, 5.5929E+03, 5.6038E+03,
     + 5.6147E+03, 5.6255E+03, 5.6364E+03, 5.6473E+03, 5.6582E+03,
     +  5.6691E+03, 5.6800E+03, 5.6909E+03, 5.7018E+03, 5.7128E+03,
     + 5.7238E+03, 5.7348E+03, 5.7457E+03, 5.7567E+03, 5.7677E+03,
     +  5.7787E+03, 5.7897E+03, 5.8007E+03, 5.8118E+03, 5.8228E+03,
     + 5.8339E+03, 5.8450E+03, 5.8560E+03, 5.8671E+03, 5.8782E+03,
     +  5.8893E+03, 5.9004E+03, 5.9116E+03, 5.9227E+03, 5.9338E+03,
     + 5.9450E+03, 5.9562E+03, 5.9674E+03, 5.9786E+03, 5.9898E+03,
     +  6.0010E+03, 6.0122E+03, 6.0234E+03, 6.0346E+03, 6.0459E+03,
     + 6.0572E+03, 6.0684E+03, 6.0797E+03, 6.0910E+03, 6.1023E+03,
     +  6.1136E+03, 6.1249E+03, 6.1362E+03, 6.1476E+03, 6.1589E+03,
     + 6.1703E+03, 6.1816E+03, 6.1930E+03, 6.2044E+03, 6.2158E+03,
     +  6.2272E+03, 6.2386E+03, 6.2501E+03, 6.2615E+03, 6.2730E+03,
     + 6.2844E+03, 6.2959E+03, 6.3074E+03, 6.3189E+03, 6.3304E+03,
     +  6.3419E+03, 6.3534E+03, 6.3649E+03, 6.3765E+03, 6.3880E+03,
     + 6.3996E+03, 6.4112E+03, 6.4227E+03, 6.4343E+03, 6.4459E+03,
     +  6.4575E+03, 6.4692E+03, 6.4808E+03, 6.4924E+03, 6.5041E+03,
     + 6.5157E+03, 6.5274E+03, 6.5391E+03, 6.5508E+03, 6.5625E+03,
     +  6.5742E+03, 6.5859E+03, 6.5976E+03, 6.6094E+03, 6.6212E+03,
     + 6.6329E+03, 6.6447E+03, 6.6565E+03, 6.6683E+03, 6.6801E+03,
     +  6.6919E+03, 6.7037E+03, 6.7155E+03, 6.7274E+03, 6.7392E+03,
     + 6.7511E+03, 6.7630E+03, 6.7749E+03, 6.7867E+03, 6.7987E+03,
     +  6.8106E+03, 6.8225E+03, 6.8344E+03, 6.8464E+03, 6.8583E+03,
     + 6.8703E+03, 6.8823E+03, 6.8942E+03, 6.9063E+03, 6.9183E+03,
     +  6.9303E+03, 6.9423E+03, 6.9544E+03, 6.9664E+03, 6.9784E+03,
     + 6.9905E+03, 7.0026E+03, 7.0147E+03, 7.0268E+03, 7.0389E+03,
     +  7.0510E+03, 7.0631E+03, 7.0752E+03, 7.0874E+03, 7.0995E+03,
     + 7.1117E+03, 7.1239E+03, 7.1361E+03, 7.1483E+03, 7.1605E+03,
     +  7.1727E+03, 7.1850E+03, 7.1972E+03, 7.2094E+03, 7.2217E+03,
     + 7.2340E+03, 7.2463E+03, 7.2586E+03, 7.2709E+03, 7.2831E+03,
     +  7.2955E+03, 7.3078E+03, 7.3202E+03, 7.3325E+03, 7.3449E+03,
     + 7.3573E+03, 7.3696E+03, 7.3820E+03, 7.3944E+03, 7.4068E+03,
     +  7.4193E+03, 7.4317E+03, 7.4442E+03, 7.4566E+03, 7.4690E+03,
     + 7.4815E+03, 7.4940E+03, 7.5065E+03, 7.5190E+03, 7.5315E+03,
     +  7.5441E+03, 7.5566E+03, 7.5691E+03, 7.5817E+03, 7.5942E+03,
     + 7.6068E+03, 7.6195E+03, 7.6320E+03, 7.6446E+03, 7.6572E+03,
     +  7.6699E+03, 7.6825E+03, 7.6952E+03, 7.7078E+03, 7.7205E+03,
     + 7.7332E+03, 7.7458E+03, 7.7586E+03, 7.7713E+03, 7.7840E+03,
     +  7.7967E+03, 7.8095E+03, 7.8222E+03, 7.8350E+03, 7.8477E+03,
     + 7.8605E+03, 7.8733E+03, 7.8861E+03, 7.8989E+03, 7.9118E+03,
     +  7.9246E+03, 7.9374E+03, 7.9503E+03, 7.9632E+03, 7.9760E+03,
     + 7.9889E+03, 8.0018E+03, 8.0147E+03, 8.0276E+03, 8.0406E+03,
     +  8.0535E+03, 8.0664E+03, 8.0794E+03, 8.0924E+03, 8.1053E+03,
     + 8.1183E+03, 8.1313E+03, 8.1443E+03, 8.1573E+03, 8.1703E+03,
     +  8.1834E+03, 8.1965E+03, 8.2095E+03, 8.2226E+03, 8.2356E+03,
     + 8.2488E+03, 8.2618E+03, 8.2749E+03, 8.2881E+03, 8.3012E+03,
     +  8.3143E+03, 8.3275E+03, 8.3406E+03, 8.3538E+03, 8.3670E+03,
     + 8.3802E+03, 8.3934E+03, 8.4066E+03, 8.4198E+03, 8.4331E+03,
     +  8.4463E+03, 8.4596E+03, 8.4728E+03, 8.4861E+03, 8.4994E+03,
     + 8.5126E+03, 8.5260E+03, 8.5393E+03, 8.5526E+03, 8.5659E+03,
     +  8.5793E+03, 8.5927E+03, 8.6060E+03, 8.6194E+03, 8.6327E+03,
     + 8.6461E+03, 8.6596E+03, 8.6730E+03, 8.6864E+03, 8.6999E+03,
     +  8.7133E+03, 8.7268E+03, 8.7402E+03, 8.7537E+03, 8.7672E+03,
     + 8.7807E+03, 8.7942E+03, 8.8078E+03, 8.8213E+03, 8.8348E+03,
     +  8.8484E+03, 8.8619E+03, 8.8755E+03, 8.8891E+03, 8.9027E+03,
     + 8.9163E+03, 8.9299E+03, 8.9435E+03, 8.9571E+03, 8.9708E+03,
     +  8.9845E+03, 8.9981E+03, 9.0118E+03, 9.0255E+03, 9.0391E+03,
     + 9.0529E+03, 9.0666E+03, 9.0803E+03, 9.0940E+03, 9.1078E+03,
     +  9.1215E+03, 9.1353E+03, 9.1491E+03, 9.1629E+03, 9.1767E+03,
     + 9.1905E+03, 9.2043E+03, 9.2182E+03, 9.2320E+03, 9.2458E+03,
     +  9.2597E+03, 9.2736E+03, 9.2874E+03, 9.3013E+03, 9.3153E+03,
     + 9.3292E+03, 9.3430E+03, 9.3570E+03, 9.3709E+03, 9.3849E+03,
     +  9.3989E+03, 9.4129E+03, 9.4268E+03, 9.4408E+03, 9.4548E+03,
     + 9.4689E+03, 9.4829E+03, 9.4969E+03, 9.5110E+03, 9.5250E+03,
     +  9.5391E+03, 9.5532E+03, 9.5672E+03, 9.5813E+03, 9.5954E+03,
     + 9.6095E+03, 9.6237E+03, 9.6379E+03, 9.6520E+03, 9.6661E+03,
     +  9.6803E+03, 9.6945E+03, 9.7087E+03, 9.7229E+03, 9.7371E+03,
     + 9.7513E+03, 9.7656E+03, 9.7798E+03, 9.7941E+03, 9.8083E+03,
     +  9.8226E+03, 9.8369E+03, 9.8511E+03, 9.8655E+03, 9.8798E+03,
     + 9.8941E+03, 9.9084E+03, 9.9228E+03, 9.9371E+03, 9.9515E+03,
     +  9.9659E+03, 9.9803E+03, 9.9947E+03, 1.0009E+04, 1.0023E+04,
     + 1.0038E+04, 1.0052E+04, 1.0067E+04, 1.0081E+04, 1.0096E+04,
     +  1.0110E+04, 1.0125E+04, 1.0139E+04, 1.0154E+04, 1.0168E+04,
     + 1.0183E+04, 1.0197E+04, 1.0212E+04, 1.0226E+04, 1.0241E+04,
     +  1.0256E+04, 1.0270E+04, 1.0285E+04, 1.0299E+04, 1.0314E+04,
     + 1.0329E+04, 1.0343E+04, 1.0358E+04, 1.0373E+04, 1.0387E+04,
     +  1.0402E+04, 1.0417E+04, 1.0431E+04, 1.0446E+04, 1.0461E+04,
     + 1.0476E+04, 1.0490E+04, 1.0505E+04, 1.0520E+04, 1.0535E+04,
     +  1.0549E+04, 1.0564E+04, 1.0579E+04, 1.0594E+04, 1.0609E+04,
     + 1.0624E+04, 1.0638E+04, 1.0653E+04, 1.0668E+04, 1.0683E+04,
     +  1.0698E+04, 1.0713E+04, 1.0728E+04, 1.0743E+04, 1.0758E+04,
     + 1.0773E+04, 1.0788E+04, 1.0802E+04, 1.0818E+04, 1.0832E+04,
     +  1.0847E+04, 1.0863E+04, 1.0877E+04, 1.0893E+04, 1.0908E+04,
     + 1.0923E+04, 1.0938E+04, 1.0953E+04, 1.0968E+04, 1.0983E+04,
     +  1.0998E+04, 1.1013E+04, 1.1028E+04, 1.1043E+04, 1.1059E+04,
     + 1.1074E+04, 1.1089E+04, 1.1104E+04, 1.1119E+04, 1.1134E+04,
     +  1.1150E+04, 1.1165E+04, 1.1180E+04, 1.1195E+04, 1.1211E+04,
     + 1.1226E+04, 1.1241E+04, 1.1256E+04, 1.1272E+04, 1.1287E+04,
     +  1.1302E+04, 1.1318E+04, 1.1333E+04, 1.1348E+04, 1.1364E+04,
     + 1.1379E+04, 1.1394E+04, 1.1410E+04, 1.1425E+04, 1.1441E+04,
     +  1.1456E+04, 1.1471E+04, 1.1487E+04, 1.1502E+04, 1.1518E+04,
     + 1.1533E+04, 1.1549E+04, 1.1564E+04, 1.1580E+04, 1.1595E+04,
     +  1.1611E+04, 1.1626E+04, 1.1642E+04, 1.1657E+04, 1.1673E+04,
     + 1.1688E+04, 1.1704E+04, 1.1720E+04, 1.1735E+04, 1.1751E+04,
     +  1.1766E+04, 1.1782E+04, 1.1798E+04, 1.1813E+04, 1.1829E+04,
     + 1.1845E+04, 1.1860E+04, 1.1876E+04, 1.1892E+04, 1.1907E+04,
     +  1.1923E+04, 1.1939E+04, 1.1955E+04, 1.1970E+04, 1.1986E+04,
     + 1.2002E+04, 1.2018E+04, 1.2034E+04, 1.2049E+04, 1.2065E+04,
     +  1.2081E+04, 1.2097E+04, 1.2113E+04, 1.2129E+04, 1.2144E+04,
     + 1.2160E+04, 1.2176E+04, 1.2192E+04, 1.2208E+04, 1.2224E+04,
     +  1.2240E+04, 1.2256E+04, 1.2272E+04, 1.2288E+04, 1.2304E+04,
     + 1.2320E+04, 1.2336E+04, 1.2352E+04, 1.2368E+04, 1.2384E+04,
     +  1.2400E+04, 1.2416E+04, 1.2432E+04, 1.2448E+04, 1.2464E+04,
     + 1.2480E+04, 1.2496E+04, 1.2512E+04, 1.2529E+04, 1.2545E+04,
     +  1.2561E+04, 1.2577E+04, 1.2593E+04, 1.2609E+04, 1.2625E+04,
     + 1.2642E+04, 1.2658E+04, 1.2674E+04, 1.2690E+04, 1.2707E+04,
     +  1.2723E+04, 1.2739E+04, 1.2755E+04, 1.2772E+04, 1.2788E+04,
     + 1.2804E+04, 1.2821E+04, 1.2837E+04, 1.2853E+04, 1.2869E+04,
     +  1.2886E+04, 1.2902E+04, 1.2919E+04, 1.2935E+04, 1.2951E+04,
     + 1.2968E+04, 1.2984E+04, 1.3001E+04, 1.3017E+04, 1.3033E+04,
     +  1.3050E+04, 1.3066E+04, 1.3083E+04, 1.3099E+04, 1.3116E+04,
     + 1.3132E+04, 1.3149E+04, 1.3165E+04, 1.3182E+04, 1.3199E+04,
     +  1.3215E+04, 1.3232E+04, 1.3248E+04, 1.3265E+04, 1.3281E+04,
     + 1.3298E+04, 1.3315E+04, 1.3331E+04, 1.3348E+04, 1.3365E+04,
     +  1.3381E+04, 1.3398E+04, 1.3415E+04, 1.3431E+04, 1.3448E+04,
     + 1.3465E+04, 1.3481E+04, 1.3498E+04, 1.3515E+04, 1.3532E+04,
     +  1.3548E+04, 1.3565E+04, 1.3582E+04, 1.3599E+04, 1.3616E+04,
     + 1.3632E+04, 1.3649E+04, 1.3666E+04, 1.3683E+04, 1.3700E+04,
     +  1.3717E+04, 1.3734E+04, 1.3750E+04, 1.3767E+04, 1.3784E+04,
     + 1.3801E+04, 1.3818E+04, 1.3835E+04, 1.3852E+04, 1.3869E+04,
     +  1.3886E+04, 1.3903E+04, 1.3920E+04, 1.3937E+04, 1.3954E+04,
     + 1.3971E+04, 1.3988E+04, 1.4005E+04, 1.4022E+04, 1.4039E+04,
     +  1.4056E+04, 1.4073E+04, 1.4091E+04, 1.4108E+04, 1.4125E+04,
     + 1.4142E+04, 1.4159E+04, 1.4176E+04, 1.4193E+04, 1.4211E+04,
     +  1.4228E+04, 1.4245E+04, 1.4262E+04, 1.4279E+04, 1.4297E+04,
     + 1.4314E+04, 1.4331E+04, 1.4348E+04, 1.4366E+04, 1.4383E+04,
     +  1.4400E+04, 1.4418E+04, 1.4435E+04, 1.4452E+04, 1.4469E+04,
     + 1.4487E+04, 1.4504E+04, 1.4522E+04, 1.4539E+04, 1.4556E+04,
     +  1.4574E+04, 1.4591E+04, 1.4608E+04, 1.4626E+04, 1.4643E+04,
     + 1.4661E+04, 1.4678E+04, 1.4696E+04, 1.4713E+04, 1.4731E+04,
     +  1.4748E+04, 1.4766E+04, 1.4783E+04, 1.4801E+04, 1.4818E+04,
     + 1.4836E+04, 1.4853E+04, 1.4871E+04, 1.4889E+04, 1.4906E+04,
     +  1.4924E+04, 1.4941E+04, 1.4959E+04, 1.4977E+04, 1.4994E+04,
     + 1.5012E+04, 1.5030E+04, 1.5047E+04, 1.5065E+04, 1.5083E+04,
     +  1.5100E+04, 1.5118E+04, 1.5136E+04, 1.5154E+04, 1.5171E+04,
     + 1.5189E+04, 1.5207E+04, 1.5225E+04, 1.5242E+04, 1.5260E+04,
     +  1.5278E+04, 1.5296E+04, 1.5314E+04, 1.5332E+04, 1.5349E+04,
     + 1.5367E+04, 1.5385E+04, 1.5403E+04, 1.5421E+04, 1.5439E+04,
     +  1.5457E+04, 1.5475E+04, 1.5493E+04, 1.5511E+04, 1.5529E+04,
     + 1.5547E+04, 1.5565E+04, 1.5583E+04, 1.5601E+04, 1.5619E+04,
     +  1.5637E+04, 1.5655E+04, 1.5673E+04, 1.5691E+04, 1.5709E+04,
     + 1.5727E+04, 1.5745E+04, 1.5763E+04, 1.5781E+04, 1.5799E+04,
     +  1.5817E+04, 1.5836E+04, 1.5854E+04, 1.5872E+04, 1.5890E+04,
     + 1.5908E+04, 1.5926E+04, 1.5945E+04, 1.5963E+04, 1.5981E+04,
     +  1.5999E+04, 1.6018E+04, 1.6036E+04, 1.6054E+04, 1.6072E+04,
     + 1.6091E+04, 1.6109E+04, 1.6127E+04, 1.6145E+04, 1.6164E+04,
     +  1.6182E+04, 1.6200E+04, 1.6219E+04, 1.6237E+04, 1.6256E+04,
     + 1.6274E+04, 1.6292E+04, 1.6311E+04, 1.6329E+04, 1.6348E+04,
     +  1.6366E+04, 1.6385E+04, 1.6403E+04, 1.6422E+04, 1.6440E+04,
     + 1.6458E+04, 1.6477E+04, 1.6495E+04, 1.6514E+04, 1.6533E+04,
     +  1.6551E+04, 1.6570E+04, 1.6588E+04, 1.6607E+04, 1.6625E+04,
     + 1.6644E+04, 1.6663E+04, 1.6681E+04, 1.6700E+04, 1.6719E+04,
     +  1.6737E+04, 1.6756E+04, 1.6775E+04, 1.6793E+04, 1.6812E+04,
     + 1.6831E+04, 1.6849E+04, 1.6868E+04, 1.6887E+04, 1.6905E+04,
     +  1.6924E+04, 1.6943E+04, 1.6962E+04, 1.6981E+04, 1.6999E+04,
     + 1.7018E+04, 1.7037E+04, 1.7056E+04, 1.7075E+04, 1.7094E+04,
     +  1.7112E+04, 1.7131E+04, 1.7150E+04, 1.7169E+04, 1.7188E+04,
     + 1.7207E+04, 1.7226E+04, 1.7245E+04, 1.7264E+04, 1.7283E+04,
     +  1.7302E+04, 1.7321E+04, 1.7340E+04, 1.7359E+04, 1.7378E+04,
     + 1.7397E+04, 1.7416E+04, 1.7435E+04, 1.7454E+04, 1.7473E+04,
     +  1.7492E+04, 1.7511E+04, 1.7530E+04, 1.7549E+04, 1.7568E+04,
     + 1.7587E+04, 1.7607E+04, 1.7626E+04, 1.7645E+04, 1.7664E+04,
     +  1.7683E+04, 1.7702E+04, 1.7722E+04, 1.7741E+04, 1.7760E+04,
     + 1.7779E+04, 1.7799E+04, 1.7818E+04, 1.7837E+04, 1.7856E+04,
     +  1.7876E+04, 1.7895E+04, 1.7914E+04, 1.7933E+04, 1.7953E+04,
     + 1.7972E+04, 1.7991E+04, 1.8011E+04, 1.8030E+04, 1.8050E+04,
     +  1.8069E+04, 1.8088E+04, 1.8108E+04, 1.8127E+04, 1.8147E+04,
     + 1.8166E+04, 1.8186E+04, 1.8205E+04, 1.8224E+04, 1.8244E+04,
     +  1.8263E+04, 1.8283E+04, 1.8302E+04, 1.8322E+04, 1.8341E+04,
     + 1.8361E+04, 1.8381E+04, 1.8400E+04, 1.8420E+04, 1.8439E+04,
     +  1.8459E+04, 1.8479E+04, 1.8498E+04, 1.8518E+04, 1.8537E+04,
     + 1.8557E+04, 1.8577E+04, 1.8596E+04, 1.8616E+04, 1.8636E+04,
     +  1.8656E+04, 1.8675E+04, 1.8695E+04, 1.8715E+04, 1.8734E+04,
     + 1.8754E+04, 1.8774E+04, 1.8794E+04, 1.8813E+04, 1.8833E+04,
     +  1.8853E+04, 1.8873E+04, 1.8893E+04, 1.8913E+04, 1.8932E+04,
     + 1.8952E+04, 1.8972E+04, 1.8992E+04, 1.9012E+04, 1.9032E+04,
     +  1.9052E+04, 1.9072E+04, 1.9092E+04, 1.9112E+04, 1.9131E+04,
     + 1.9152E+04, 1.9171E+04, 1.9191E+04, 1.9212E+04, 1.9231E+04,
     +  1.9251E+04, 1.9272E+04, 1.9292E+04, 1.9312E+04, 1.9332E+04,
     + 1.9352E+04, 1.9372E+04, 1.9392E+04, 1.9412E+04, 1.9432E+04,
     +  1.9452E+04, 1.9472E+04, 1.9493E+04, 1.9513E+04, 1.9533E+04,
     + 1.9553E+04, 1.9573E+04, 1.9593E+04, 1.9614E+04, 1.9634E+04,
     +  1.9654E+04, 1.9674E+04, 1.9695E+04, 1.9715E+04, 1.9735E+04,
     + 1.9755E+04, 1.9776E+04, 1.9796E+04, 1.9816E+04, 1.9837E+04,
     +  1.9857E+04, 1.9877E+04, 1.9898E+04, 1.9918E+04, 1.9938E+04,
     + 1.9959E+04, 1.9979E+04, 2.0000E+04, 2.0020E+04, 2.0040E+04,
     +  2.0061E+04, 2.0081E+04, 2.0102E+04, 2.0122E+04, 2.0143E+04,
     + 2.0163E+04, 2.0184E+04, 2.0204E+04, 2.0225E+04, 2.0245E+04,
     +  2.0266E+04, 2.0286E+04, 2.0307E+04, 2.0328E+04, 2.0348E+04,
     + 2.0369E+04, 2.0389E+04, 2.0410E+04, 2.0431E+04, 2.0451E+04,
     +  2.0472E+04, 2.0493E+04, 2.0513E+04, 2.0534E+04, 2.0555E+04,
     + 2.0575E+04, 2.0596E+04, 2.0617E+04, 2.0637E+04, 2.0658E+04,
     +  2.0679E+04, 2.0700E+04, 2.0721E+04, 2.0741E+04, 2.0762E+04,
     + 2.0783E+04, 2.0804E+04, 2.0825E+04, 2.0845E+04, 2.0866E+04,
     +  2.0887E+04, 2.0908E+04, 2.0929E+04, 2.0950E+04, 2.0971E+04,
     + 2.0992E+04, 2.1013E+04, 2.1033E+04, 2.1054E+04, 2.1075E+04,
     +  2.1096E+04, 2.1117E+04, 2.1138E+04, 2.1159E+04, 2.1180E+04,
     + 2.1201E+04, 2.1222E+04, 2.1243E+04, 2.1264E+04, 2.1285E+04,
     +  2.1307E+04, 2.1328E+04, 2.1349E+04, 2.1370E+04, 2.1391E+04,
     + 2.1412E+04, 2.1433E+04, 2.1454E+04, 2.1476E+04, 2.1497E+04,
     +  2.1518E+04, 2.1539E+04, 2.1560E+04, 2.1582E+04, 2.1603E+04,
     + 2.1624E+04, 2.1645E+04, 2.1666E+04, 2.1688E+04, 2.1709E+04,
     +  2.1730E+04, 2.1752E+04, 2.1773E+04, 2.1794E+04, 2.1815E+04,
     + 2.1837E+04, 2.1858E+04, 2.1880E+04, 2.1901E+04, 2.1922E+04,
     +  2.1944E+04, 2.1965E+04, 2.1986E+04, 2.2008E+04, 2.2029E+04,
     + 2.2051E+04, 2.2072E+04, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        48
      data (QofT( 3,J),J=1,2408)/ 1.1892E+01, 1.2680E+01, 1.4396E+01,
     +  1.6547E+01, 1.8882E+01, 2.1307E+01, 2.3783E+01, 2.6290E+01,
     + 2.8816E+01, 3.1357E+01, 3.3907E+01, 3.6465E+01, 3.9028E+01,
     +  4.1596E+01, 4.4168E+01, 4.6742E+01, 4.9319E+01, 5.1899E+01,
     + 5.4481E+01, 5.7066E+01, 6.2244E+01, 6.7435E+01, 7.2645E+01,
     +  7.7879E+01, 8.3142E+01, 8.8441E+01, 9.3782E+01, 9.9170E+01,
     + 1.0461E+02, 1.1011E+02, 1.1567E+02, 1.2130E+02, 1.2700E+02,
     +  1.3278E+02, 1.3862E+02, 1.4455E+02, 1.5055E+02, 1.5663E+02,
     + 1.6279E+02, 1.6903E+02, 1.7535E+02, 1.8175E+02, 1.8823E+02,
     +  1.9479E+02, 2.0142E+02, 2.0813E+02, 2.1492E+02, 2.2178E+02,
     + 2.2871E+02, 2.3571E+02, 2.4279E+02, 2.4993E+02, 2.5713E+02,
     +  2.6441E+02, 2.7175E+02, 2.7914E+02, 2.8661E+02, 2.9413E+02,
     + 3.0170E+02, 3.0934E+02, 3.1703E+02, 3.2477E+02, 3.3257E+02,
     +  3.4076E+02, 3.4876E+02, 3.5680E+02, 3.6489E+02, 3.7303E+02,
     + 3.8121E+02, 3.8944E+02, 3.9771E+02, 4.0601E+02, 4.1436E+02,
     +  4.2275E+02, 4.3117E+02, 4.3963E+02, 4.4813E+02, 4.5666E+02,
     + 4.6523E+02, 4.7383E+02, 4.8246E+02, 4.9112E+02, 4.9982E+02,
     +  5.0854E+02, 5.1729E+02, 5.2608E+02, 5.3489E+02, 5.4372E+02,
     + 5.5259E+02, 5.6148E+02, 5.7039E+02, 5.7933E+02, 5.8829E+02,
     +  5.9728E+02, 6.0629E+02, 6.1532E+02, 6.2438E+02, 6.3345E+02,
     + 6.4255E+02, 6.5167E+02, 6.6080E+02, 6.6996E+02, 6.7914E+02,
     +  6.8833E+02, 6.9754E+02, 7.0677E+02, 7.1602E+02, 7.2529E+02,
     + 7.3457E+02, 7.4387E+02, 7.5318E+02, 7.6251E+02, 7.7185E+02,
     +  7.8121E+02, 7.9059E+02, 7.9998E+02, 8.0938E+02, 8.1881E+02,
     + 8.2824E+02, 8.3768E+02, 8.4714E+02, 8.5661E+02, 8.6609E+02,
     +  8.7559E+02, 8.8509E+02, 8.9461E+02, 9.0414E+02, 9.1368E+02,
     + 9.2324E+02, 9.3281E+02, 9.4238E+02, 9.5197E+02, 9.6156E+02,
     +  9.7117E+02, 9.8079E+02, 9.9042E+02, 1.0000E+03, 1.0097E+03,
     + 1.0194E+03, 1.0290E+03, 1.0387E+03, 1.0484E+03, 1.0581E+03,
     +  1.0678E+03, 1.0775E+03, 1.0872E+03, 1.0969E+03, 1.1067E+03,
     + 1.1164E+03, 1.1262E+03, 1.1359E+03, 1.1457E+03, 1.1555E+03,
     +  1.1652E+03, 1.1750E+03, 1.1848E+03, 1.1946E+03, 1.2044E+03,
     + 1.2143E+03, 1.2241E+03, 1.2339E+03, 1.2438E+03, 1.2536E+03,
     +  1.2635E+03, 1.2733E+03, 1.2832E+03, 1.2931E+03, 1.3029E+03,
     + 1.3128E+03, 1.3227E+03, 1.3326E+03, 1.3425E+03, 1.3524E+03,
     +  1.3624E+03, 1.3723E+03, 1.3822E+03, 1.3922E+03, 1.4021E+03,
     + 1.4121E+03, 1.4220E+03, 1.4320E+03, 1.4419E+03, 1.4519E+03,
     +  1.4619E+03, 1.4719E+03, 1.4819E+03, 1.4919E+03, 1.5019E+03,
     + 1.5119E+03, 1.5219E+03, 1.5319E+03, 1.5419E+03, 1.5520E+03,
     +  1.5620E+03, 1.5721E+03, 1.5821E+03, 1.5922E+03, 1.6022E+03,
     + 1.6123E+03, 1.6224E+03, 1.6325E+03, 1.6426E+03, 1.6526E+03,
     +  1.6627E+03, 1.6728E+03, 1.6830E+03, 1.6931E+03, 1.7032E+03,
     + 1.7133E+03, 1.7234E+03, 1.7336E+03, 1.7437E+03, 1.7539E+03,
     +  1.7640E+03, 1.7742E+03, 1.7844E+03, 1.7945E+03, 1.8047E+03,
     + 1.8149E+03, 1.8251E+03, 1.8353E+03, 1.8455E+03, 1.8557E+03,
     +  1.8659E+03, 1.8761E+03, 1.8864E+03, 1.8966E+03, 1.9068E+03,
     + 1.9171E+03, 1.9273E+03, 1.9376E+03, 1.9479E+03, 1.9581E+03,
     +  1.9684E+03, 1.9787E+03, 1.9890E+03, 1.9993E+03, 2.0096E+03,
     + 2.0199E+03, 2.0302E+03, 2.0406E+03, 2.0509E+03, 2.0612E+03,
     +  2.0716E+03, 2.0819E+03, 2.0923E+03, 2.1027E+03, 2.1130E+03,
     + 2.1234E+03, 2.1338E+03, 2.1442E+03, 2.1546E+03, 2.1650E+03,
     +  2.1754E+03, 2.1858E+03, 2.1962E+03, 2.2067E+03, 2.2171E+03,
     + 2.2276E+03, 2.2380E+03, 2.2485E+03, 2.2590E+03, 2.2694E+03,
     +  2.2799E+03, 2.2904E+03, 2.3009E+03, 2.3114E+03, 2.3220E+03,
     + 2.3325E+03, 2.3430E+03, 2.3535E+03, 2.3641E+03, 2.3747E+03,
     +  2.3852E+03, 2.3958E+03, 2.4064E+03, 2.4169E+03, 2.4276E+03,
     + 2.4382E+03, 2.4488E+03, 2.4594E+03, 2.4700E+03, 2.4807E+03,
     +  2.4913E+03, 2.5019E+03, 2.5126E+03, 2.5233E+03, 2.5339E+03,
     + 2.5446E+03, 2.5553E+03, 2.5660E+03, 2.5767E+03, 2.5875E+03,
     +  2.5982E+03, 2.6089E+03, 2.6197E+03, 2.6304E+03, 2.6412E+03,
     + 2.6519E+03, 2.6627E+03, 2.6735E+03, 2.6843E+03, 2.6951E+03,
     +  2.7059E+03, 2.7168E+03, 2.7276E+03, 2.7384E+03, 2.7493E+03,
     + 2.7602E+03, 2.7710E+03, 2.7819E+03, 2.7928E+03, 2.8037E+03,
     +  2.8146E+03, 2.8255E+03, 2.8364E+03, 2.8474E+03, 2.8583E+03,
     + 2.8693E+03, 2.8803E+03, 2.8912E+03, 2.9022E+03, 2.9132E+03,
     +  2.9242E+03, 2.9352E+03, 2.9463E+03, 2.9573E+03, 2.9683E+03,
     + 2.9794E+03, 2.9904E+03, 3.0015E+03, 3.0126E+03, 3.0237E+03,
     +  3.0348E+03, 3.0459E+03, 3.0570E+03, 3.0682E+03, 3.0793E+03,
     + 3.0904E+03, 3.1016E+03, 3.1128E+03, 3.1240E+03, 3.1352E+03,
     +  3.1464E+03, 3.1576E+03, 3.1688E+03, 3.1801E+03, 3.1913E+03,
     + 3.2026E+03, 3.2138E+03, 3.2251E+03, 3.2364E+03, 3.2477E+03,
     +  3.2590E+03, 3.2703E+03, 3.2817E+03, 3.2930E+03, 3.3044E+03,
     + 3.3157E+03, 3.3271E+03, 3.3385E+03, 3.3499E+03, 3.3613E+03,
     +  3.3727E+03, 3.3842E+03, 3.3956E+03, 3.4070E+03, 3.4185E+03,
     + 3.4300E+03, 3.4415E+03, 3.4530E+03, 3.4645E+03, 3.4760E+03,
     +  3.4875E+03, 3.4991E+03, 3.5106E+03, 3.5222E+03, 3.5338E+03,
     + 3.5454E+03, 3.5570E+03, 3.5686E+03, 3.5802E+03, 3.5918E+03,
     +  3.6035E+03, 3.6151E+03, 3.6268E+03, 3.6385E+03, 3.6502E+03,
     + 3.6619E+03, 3.6736E+03, 3.6853E+03, 3.6971E+03, 3.7088E+03,
     +  3.7206E+03, 3.7324E+03, 3.7441E+03, 3.7560E+03, 3.7678E+03,
     + 3.7796E+03, 3.7914E+03, 3.8033E+03, 3.8151E+03, 3.8270E+03,
     +  3.8389E+03, 3.8508E+03, 3.8627E+03, 3.8746E+03, 3.8866E+03,
     + 3.8985E+03, 3.9105E+03, 3.9224E+03, 3.9344E+03, 3.9464E+03,
     +  3.9584E+03, 3.9704E+03, 3.9824E+03, 3.9945E+03, 4.0065E+03,
     + 4.0186E+03, 4.0307E+03, 4.0428E+03, 4.0549E+03, 4.0670E+03,
     +  4.0792E+03, 4.0913E+03, 4.1035E+03, 4.1156E+03, 4.1278E+03,
     + 4.1400E+03, 4.1522E+03, 4.1644E+03, 4.1767E+03, 4.1889E+03,
     +  4.2012E+03, 4.2134E+03, 4.2257E+03, 4.2380E+03, 4.2503E+03,
     + 4.2626E+03, 4.2750E+03, 4.2873E+03, 4.2997E+03, 4.3121E+03,
     +  4.3244E+03, 4.3368E+03, 4.3492E+03, 4.3616E+03, 4.3741E+03,
     + 4.3866E+03, 4.3990E+03, 4.4115E+03, 4.4240E+03, 4.4365E+03,
     +  4.4490E+03, 4.4615E+03, 4.4741E+03, 4.4866E+03, 4.4992E+03,
     + 4.5118E+03, 4.5244E+03, 4.5370E+03, 4.5496E+03, 4.5622E+03,
     +  4.5749E+03, 4.5876E+03, 4.6002E+03, 4.6129E+03, 4.6256E+03,
     + 4.6383E+03, 4.6510E+03, 4.6638E+03, 4.6766E+03, 4.6893E+03,
     +  4.7021E+03, 4.7149E+03, 4.7277E+03, 4.7405E+03, 4.7534E+03,
     + 4.7662E+03, 4.7791E+03, 4.7919E+03, 4.8048E+03, 4.8177E+03,
     +  4.8306E+03, 4.8436E+03, 4.8565E+03, 4.8695E+03, 4.8825E+03,
     + 4.8954E+03, 4.9084E+03, 4.9214E+03, 4.9344E+03, 4.9475E+03,
     +  4.9605E+03, 4.9736E+03, 4.9867E+03, 4.9998E+03, 5.0129E+03,
     + 5.0260E+03, 5.0391E+03, 5.0523E+03, 5.0655E+03, 5.0786E+03,
     +  5.0918E+03, 5.1050E+03, 5.1183E+03, 5.1315E+03, 5.1447E+03,
     + 5.1580E+03, 5.1712E+03, 5.1846E+03, 5.1979E+03, 5.2112E+03,
     +  5.2245E+03, 5.2378E+03, 5.2512E+03, 5.2645E+03, 5.2779E+03,
     + 5.2914E+03, 5.3048E+03, 5.3182E+03, 5.3316E+03, 5.3451E+03,
     +  5.3585E+03, 5.3720E+03, 5.3855E+03, 5.3991E+03, 5.4125E+03,
     + 5.4261E+03, 5.4397E+03, 5.4532E+03, 5.4668E+03, 5.4804E+03,
     +  5.4940E+03, 5.5076E+03, 5.5213E+03, 5.5349E+03, 5.5486E+03,
     + 5.5623E+03, 5.5759E+03, 5.5897E+03, 5.6034E+03, 5.6171E+03,
     +  5.6309E+03, 5.6447E+03, 5.6584E+03, 5.6722E+03, 5.6860E+03,
     + 5.6999E+03, 5.7137E+03, 5.7276E+03, 5.7414E+03, 5.7553E+03,
     +  5.7692E+03, 5.7831E+03, 5.7971E+03, 5.8110E+03, 5.8249E+03,
     + 5.8389E+03, 5.8529E+03, 5.8669E+03, 5.8809E+03, 5.8949E+03,
     +  5.9090E+03, 5.9230E+03, 5.9371E+03, 5.9512E+03, 5.9653E+03,
     + 5.9794E+03, 5.9935E+03, 6.0077E+03, 6.0218E+03, 6.0360E+03,
     +  6.0501E+03, 6.0644E+03, 6.0786E+03, 6.0928E+03, 6.1070E+03,
     + 6.1213E+03, 6.1355E+03, 6.1498E+03, 6.1641E+03, 6.1785E+03,
     +  6.1928E+03, 6.2071E+03, 6.2215E+03, 6.2358E+03, 6.2503E+03,
     + 6.2647E+03, 6.2791E+03, 6.2935E+03, 6.3079E+03, 6.3224E+03,
     +  6.3369E+03, 6.3514E+03, 6.3659E+03, 6.3804E+03, 6.3949E+03,
     + 6.4095E+03, 6.4240E+03, 6.4386E+03, 6.4532E+03, 6.4678E+03,
     +  6.4824E+03, 6.4971E+03, 6.5117E+03, 6.5264E+03, 6.5411E+03,
     + 6.5558E+03, 6.5705E+03, 6.5852E+03, 6.6000E+03, 6.6147E+03,
     +  6.6295E+03, 6.6443E+03, 6.6591E+03, 6.6739E+03, 6.6887E+03,
     + 6.7036E+03, 6.7184E+03, 6.7333E+03, 6.7482E+03, 6.7631E+03,
     +  6.7780E+03, 6.7930E+03, 6.8079E+03, 6.8229E+03, 6.8379E+03,
     + 6.8528E+03, 6.8678E+03, 6.8829E+03, 6.8979E+03, 6.9130E+03,
     +  6.9280E+03, 6.9431E+03, 6.9583E+03, 6.9733E+03, 6.9884E+03,
     + 7.0036E+03, 7.0187E+03, 7.0339E+03, 7.0491E+03, 7.0644E+03,
     +  7.0795E+03, 7.0948E+03, 7.1100E+03, 7.1253E+03, 7.1406E+03,
     + 7.1559E+03, 7.1712E+03, 7.1865E+03, 7.2019E+03, 7.2172E+03,
     +  7.2325E+03, 7.2480E+03, 7.2633E+03, 7.2788E+03, 7.2942E+03,
     + 7.3096E+03, 7.3251E+03, 7.3405E+03, 7.3560E+03, 7.3715E+03,
     +  7.3871E+03, 7.4026E+03, 7.4182E+03, 7.4337E+03, 7.4493E+03,
     + 7.4649E+03, 7.4805E+03, 7.4961E+03, 7.5118E+03, 7.5274E+03,
     +  7.5431E+03, 7.5588E+03, 7.5745E+03, 7.5902E+03, 7.6059E+03,
     + 7.6217E+03, 7.6374E+03, 7.6532E+03, 7.6690E+03, 7.6848E+03,
     +  7.7006E+03, 7.7165E+03, 7.7323E+03, 7.7482E+03, 7.7641E+03,
     + 7.7799E+03, 7.7958E+03, 7.8118E+03, 7.8277E+03, 7.8437E+03,
     +  7.8597E+03, 7.8756E+03, 7.8917E+03, 7.9077E+03, 7.9237E+03,
     + 7.9397E+03, 7.9558E+03, 7.9719E+03, 7.9880E+03, 8.0041E+03,
     +  8.0202E+03, 8.0364E+03, 8.0525E+03, 8.0687E+03, 8.0849E+03,
     + 8.1011E+03, 8.1173E+03, 8.1336E+03, 8.1498E+03, 8.1660E+03,
     +  8.1823E+03, 8.1986E+03, 8.2149E+03, 8.2313E+03, 8.2476E+03,
     + 8.2639E+03, 8.2803E+03, 8.2967E+03, 8.3131E+03, 8.3295E+03,
     +  8.3459E+03, 8.3624E+03, 8.3788E+03, 8.3953E+03, 8.4118E+03,
     + 8.4283E+03, 8.4449E+03, 8.4614E+03, 8.4779E+03, 8.4945E+03,
     +  8.5111E+03, 8.5277E+03, 8.5443E+03, 8.5609E+03, 8.5776E+03,
     + 8.5942E+03, 8.6109E+03, 8.6276E+03, 8.6443E+03, 8.6610E+03,
     +  8.6778E+03, 8.6946E+03, 8.7113E+03, 8.7281E+03, 8.7449E+03,
     + 8.7617E+03, 8.7785E+03, 8.7954E+03, 8.8122E+03, 8.8291E+03,
     +  8.8460E+03, 8.8629E+03, 8.8798E+03, 8.8968E+03, 8.9138E+03,
     + 8.9307E+03, 8.9477E+03, 8.9647E+03, 8.9817E+03, 8.9987E+03,
     +  9.0158E+03, 9.0328E+03, 9.0500E+03, 9.0670E+03, 9.0841E+03,
     + 9.1012E+03, 9.1184E+03, 9.1356E+03, 9.1527E+03, 9.1699E+03,
     +  9.1871E+03, 9.2043E+03, 9.2215E+03, 9.2388E+03, 9.2560E+03,
     + 9.2733E+03, 9.2906E+03, 9.3079E+03, 9.3252E+03, 9.3426E+03,
     +  9.3600E+03, 9.3773E+03, 9.3947E+03, 9.4121E+03, 9.4296E+03,
     + 9.4470E+03, 9.4644E+03, 9.4819E+03, 9.4994E+03, 9.5168E+03,
     +  9.5344E+03, 9.5519E+03, 9.5694E+03, 9.5870E+03, 9.6045E+03,
     + 9.6222E+03, 9.6398E+03, 9.6573E+03, 9.6750E+03, 9.6927E+03,
     +  9.7103E+03, 9.7280E+03, 9.7457E+03, 9.7635E+03, 9.7811E+03,
     + 9.7989E+03, 9.8167E+03, 9.8345E+03, 9.8522E+03, 9.8700E+03,
     +  9.8878E+03, 9.9057E+03, 9.9235E+03, 9.9414E+03, 9.9593E+03,
     + 9.9772E+03, 9.9951E+03, 1.0013E+04, 1.0031E+04, 1.0049E+04,
     +  1.0067E+04, 1.0085E+04, 1.0103E+04, 1.0121E+04, 1.0139E+04,
     + 1.0157E+04, 1.0175E+04, 1.0193E+04, 1.0211E+04, 1.0229E+04,
     +  1.0248E+04, 1.0266E+04, 1.0284E+04, 1.0302E+04, 1.0320E+04,
     + 1.0338E+04, 1.0357E+04, 1.0375E+04, 1.0393E+04, 1.0411E+04,
     +  1.0430E+04, 1.0448E+04, 1.0466E+04, 1.0485E+04, 1.0503E+04,
     + 1.0521E+04, 1.0540E+04, 1.0558E+04, 1.0577E+04, 1.0595E+04,
     +  1.0613E+04, 1.0632E+04, 1.0650E+04, 1.0669E+04, 1.0687E+04,
     + 1.0706E+04, 1.0724E+04, 1.0743E+04, 1.0762E+04, 1.0780E+04,
     +  1.0799E+04, 1.0817E+04, 1.0836E+04, 1.0855E+04, 1.0873E+04,
     + 1.0892E+04, 1.0911E+04, 1.0929E+04, 1.0948E+04, 1.0967E+04,
     +  1.0986E+04, 1.1004E+04, 1.1023E+04, 1.1042E+04, 1.1061E+04,
     + 1.1080E+04, 1.1099E+04, 1.1117E+04, 1.1136E+04, 1.1155E+04,
     +  1.1174E+04, 1.1193E+04, 1.1212E+04, 1.1231E+04, 1.1250E+04,
     + 1.1269E+04, 1.1288E+04, 1.1307E+04, 1.1326E+04, 1.1345E+04,
     +  1.1364E+04, 1.1383E+04, 1.1402E+04, 1.1421E+04, 1.1441E+04,
     + 1.1460E+04, 1.1479E+04, 1.1498E+04, 1.1517E+04, 1.1537E+04,
     +  1.1556E+04, 1.1575E+04, 1.1594E+04, 1.1614E+04, 1.1633E+04,
     + 1.1652E+04, 1.1672E+04, 1.1691E+04, 1.1710E+04, 1.1730E+04,
     +  1.1749E+04, 1.1768E+04, 1.1788E+04, 1.1807E+04, 1.1827E+04,
     + 1.1846E+04, 1.1866E+04, 1.1885E+04, 1.1905E+04, 1.1924E+04,
     +  1.1944E+04, 1.1963E+04, 1.1983E+04, 1.2002E+04, 1.2022E+04,
     + 1.2042E+04, 1.2061E+04, 1.2081E+04, 1.2101E+04, 1.2120E+04,
     +  1.2140E+04, 1.2160E+04, 1.2179E+04, 1.2199E+04, 1.2219E+04,
     + 1.2239E+04, 1.2259E+04, 1.2278E+04, 1.2298E+04, 1.2318E+04,
     +  1.2338E+04, 1.2358E+04, 1.2378E+04, 1.2398E+04, 1.2417E+04,
     + 1.2437E+04, 1.2457E+04, 1.2477E+04, 1.2497E+04, 1.2517E+04,
     +  1.2537E+04, 1.2557E+04, 1.2577E+04, 1.2598E+04, 1.2618E+04,
     + 1.2638E+04, 1.2658E+04, 1.2678E+04, 1.2698E+04, 1.2718E+04,
     +  1.2738E+04, 1.2759E+04, 1.2779E+04, 1.2799E+04, 1.2819E+04,
     + 1.2840E+04, 1.2860E+04, 1.2880E+04, 1.2900E+04, 1.2921E+04,
     +  1.2941E+04, 1.2961E+04, 1.2982E+04, 1.3002E+04, 1.3023E+04,
     + 1.3043E+04, 1.3063E+04, 1.3084E+04, 1.3104E+04, 1.3125E+04,
     +  1.3145E+04, 1.3166E+04, 1.3186E+04, 1.3207E+04, 1.3227E+04,
     + 1.3248E+04, 1.3269E+04, 1.3289E+04, 1.3310E+04, 1.3330E+04,
     +  1.3351E+04, 1.3372E+04, 1.3392E+04, 1.3413E+04, 1.3434E+04,
     + 1.3455E+04, 1.3475E+04, 1.3496E+04, 1.3517E+04, 1.3538E+04,
     +  1.3558E+04, 1.3579E+04, 1.3600E+04, 1.3621E+04, 1.3642E+04,
     + 1.3663E+04, 1.3684E+04, 1.3705E+04, 1.3726E+04, 1.3747E+04,
     +  1.3767E+04, 1.3788E+04, 1.3809E+04, 1.3830E+04, 1.3851E+04,
     + 1.3873E+04, 1.3894E+04, 1.3915E+04, 1.3936E+04, 1.3957E+04,
     +  1.3978E+04, 1.3999E+04, 1.4020E+04, 1.4042E+04, 1.4063E+04,
     + 1.4084E+04, 1.4105E+04, 1.4126E+04, 1.4148E+04, 1.4169E+04,
     +  1.4190E+04, 1.4211E+04, 1.4233E+04, 1.4254E+04, 1.4275E+04,
     + 1.4297E+04, 1.4318E+04, 1.4340E+04, 1.4361E+04, 1.4382E+04,
     +  1.4404E+04, 1.4425E+04, 1.4447E+04, 1.4468E+04, 1.4490E+04,
     + 1.4511E+04, 1.4533E+04, 1.4554E+04, 1.4576E+04, 1.4598E+04,
     +  1.4619E+04, 1.4641E+04, 1.4662E+04, 1.4684E+04, 1.4706E+04,
     + 1.4728E+04, 1.4749E+04, 1.4771E+04, 1.4793E+04, 1.4814E+04,
     +  1.4836E+04, 1.4858E+04, 1.4880E+04, 1.4902E+04, 1.4923E+04,
     + 1.4945E+04, 1.4967E+04, 1.4989E+04, 1.5011E+04, 1.5033E+04,
     +  1.5055E+04, 1.5077E+04, 1.5099E+04, 1.5121E+04, 1.5142E+04,
     + 1.5164E+04, 1.5186E+04, 1.5209E+04, 1.5231E+04, 1.5253E+04,
     +  1.5275E+04, 1.5297E+04, 1.5319E+04, 1.5341E+04, 1.5363E+04,
     + 1.5385E+04, 1.5408E+04, 1.5430E+04, 1.5452E+04, 1.5474E+04,
     +  1.5496E+04, 1.5519E+04, 1.5541E+04, 1.5563E+04, 1.5586E+04,
     + 1.5608E+04, 1.5630E+04, 1.5653E+04, 1.5675E+04, 1.5697E+04,
     +  1.5720E+04, 1.5742E+04, 1.5765E+04, 1.5787E+04, 1.5809E+04,
     + 1.5832E+04, 1.5854E+04, 1.5877E+04, 1.5899E+04, 1.5922E+04,
     +  1.5945E+04, 1.5967E+04, 1.5990E+04, 1.6012E+04, 1.6035E+04,
     + 1.6058E+04, 1.6080E+04, 1.6103E+04, 1.6126E+04, 1.6148E+04,
     +  1.6171E+04, 1.6194E+04, 1.6217E+04, 1.6239E+04, 1.6262E+04,
     + 1.6285E+04, 1.6308E+04, 1.6331E+04, 1.6353E+04, 1.6376E+04,
     +  1.6399E+04, 1.6422E+04, 1.6445E+04, 1.6468E+04, 1.6491E+04,
     + 1.6514E+04, 1.6537E+04, 1.6560E+04, 1.6583E+04, 1.6606E+04,
     +  1.6629E+04, 1.6652E+04, 1.6675E+04, 1.6698E+04, 1.6721E+04,
     + 1.6744E+04, 1.6767E+04, 1.6791E+04, 1.6814E+04, 1.6837E+04,
     +  1.6860E+04, 1.6883E+04, 1.6906E+04, 1.6930E+04, 1.6953E+04,
     + 1.6976E+04, 1.7000E+04, 1.7023E+04, 1.7046E+04, 1.7070E+04,
     +  1.7093E+04, 1.7116E+04, 1.7140E+04, 1.7163E+04, 1.7186E+04,
     + 1.7210E+04, 1.7233E+04, 1.7257E+04, 1.7280E+04, 1.7304E+04,
     +  1.7327E+04, 1.7351E+04, 1.7374E+04, 1.7398E+04, 1.7422E+04,
     + 1.7445E+04, 1.7469E+04, 1.7492E+04, 1.7516E+04, 1.7540E+04,
     +  1.7563E+04, 1.7587E+04, 1.7611E+04, 1.7635E+04, 1.7658E+04,
     + 1.7682E+04, 1.7706E+04, 1.7730E+04, 1.7753E+04, 1.7777E+04,
     +  1.7801E+04, 1.7825E+04, 1.7849E+04, 1.7873E+04, 1.7897E+04,
     + 1.7921E+04, 1.7945E+04, 1.7968E+04, 1.7992E+04, 1.8016E+04,
     +  1.8040E+04, 1.8064E+04, 1.8088E+04, 1.8113E+04, 1.8136E+04,
     + 1.8161E+04, 1.8185E+04, 1.8209E+04, 1.8233E+04, 1.8257E+04,
     +  1.8281E+04, 1.8305E+04, 1.8330E+04, 1.8354E+04, 1.8378E+04,
     + 1.8402E+04, 1.8427E+04, 1.8451E+04, 1.8475E+04, 1.8499E+04,
     +  1.8524E+04, 1.8548E+04, 1.8572E+04, 1.8597E+04, 1.8621E+04,
     + 1.8646E+04, 1.8670E+04, 1.8694E+04, 1.8719E+04, 1.8743E+04,
     +  1.8768E+04, 1.8792E+04, 1.8817E+04, 1.8841E+04, 1.8866E+04,
     + 1.8890E+04, 1.8915E+04, 1.8940E+04, 1.8964E+04, 1.8989E+04,
     +  1.9013E+04, 1.9038E+04, 1.9063E+04, 1.9087E+04, 1.9112E+04,
     + 1.9137E+04, 1.9162E+04, 1.9186E+04, 1.9211E+04, 1.9236E+04,
     +  1.9261E+04, 1.9285E+04, 1.9310E+04, 1.9335E+04, 1.9360E+04,
     + 1.9385E+04, 1.9410E+04, 1.9435E+04, 1.9460E+04, 1.9485E+04,
     +  1.9510E+04, 1.9535E+04, 1.9560E+04, 1.9585E+04, 1.9610E+04,
     + 1.9635E+04, 1.9660E+04, 1.9685E+04, 1.9710E+04, 1.9735E+04,
     +  1.9760E+04, 1.9785E+04, 1.9810E+04, 1.9835E+04, 1.9861E+04,
     + 1.9886E+04, 1.9911E+04, 1.9936E+04, 1.9962E+04, 1.9987E+04,
     +  2.0012E+04, 2.0037E+04, 2.0063E+04, 2.0088E+04, 2.0113E+04,
     + 2.0139E+04, 2.0164E+04, 2.0190E+04, 2.0215E+04, 2.0240E+04,
     +  2.0266E+04, 2.0291E+04, 2.0317E+04, 2.0342E+04, 2.0368E+04,
     + 2.0393E+04, 2.0419E+04, 2.0444E+04, 2.0470E+04, 2.0495E+04,
     +  2.0521E+04, 2.0547E+04, 2.0572E+04, 2.0598E+04, 2.0624E+04,
     + 2.0649E+04, 2.0675E+04, 2.0701E+04, 2.0727E+04, 2.0752E+04,
     +  2.0778E+04, 2.0804E+04, 2.0830E+04, 2.0855E+04, 2.0881E+04,
     + 2.0907E+04, 2.0933E+04, 2.0959E+04, 2.0985E+04, 2.1011E+04,
     +  2.1036E+04, 2.1063E+04, 2.1088E+04, 2.1114E+04, 2.1140E+04,
     + 2.1166E+04, 2.1192E+04, 2.1218E+04, 2.1244E+04, 2.1271E+04,
     +  2.1297E+04, 2.1323E+04, 2.1349E+04, 2.1375E+04, 2.1401E+04,
     + 2.1427E+04, 2.1453E+04, 2.1480E+04, 2.1506E+04, 2.1532E+04,
     +  2.1558E+04, 2.1585E+04, 2.1611E+04, 2.1637E+04, 2.1664E+04,
     + 2.1690E+04, 2.1716E+04, 2.1743E+04, 2.1769E+04, 2.1795E+04,
     +  2.1822E+04, 2.1848E+04, 2.1875E+04, 2.1901E+04, 2.1928E+04,
     + 2.1954E+04, 2.1981E+04, 2.2007E+04, 2.2034E+04, 2.2060E+04,
     +  2.2087E+04, 2.2113E+04, 2.2140E+04, 2.2166E+04, 2.2193E+04,
     + 2.2220E+04, 2.2246E+04, 2.2273E+04, 2.2300E+04, 2.2327E+04,
     +  2.2353E+04, 2.2380E+04, 2.2407E+04, 2.2434E+04, 2.2460E+04,
     + 2.2487E+04, 2.2514E+04, 2.2541E+04, 2.2568E+04, 2.2595E+04,
     +  2.2621E+04, 2.2648E+04, 2.2675E+04, 2.2702E+04, 2.2729E+04,
     + 2.2756E+04, 2.2783E+04, 2.2810E+04, 2.2837E+04, 2.2864E+04,
     +  2.2891E+04, 2.2918E+04, 2.2945E+04, 2.2972E+04, 2.2999E+04,
     + 2.3027E+04, 2.3054E+04, 2.3081E+04, 2.3108E+04, 2.3135E+04,
     +  2.3162E+04, 2.3190E+04, 2.3217E+04, 2.3244E+04, 2.3272E+04,
     + 2.3299E+04, 2.3326E+04, 2.3353E+04, 2.3381E+04, 2.3408E+04,
     +  2.3436E+04, 2.3463E+04, 2.3490E+04, 2.3518E+04, 2.3545E+04,
     + 2.3573E+04, 2.3600E+04, 2.3628E+04, 2.3655E+04, 2.3683E+04,
     +  2.3710E+04, 2.3738E+04, 2.3765E+04, 2.3793E+04, 2.3820E+04,
     + 2.3848E+04, 2.3876E+04, 2.3903E+04, 2.3931E+04, 2.3959E+04,
     +  2.3986E+04, 2.4014E+04, 2.4042E+04, 2.4070E+04, 2.4097E+04,
     + 2.4125E+04, 2.4153E+04, 2.4181E+04, 2.4209E+04, 2.4236E+04,
     +  2.4264E+04, 2.4292E+04, 2.4320E+04, 2.4348E+04, 2.4376E+04,
     + 2.4404E+04, 2.4432E+04, 2.4460E+04, 2.4488E+04, 2.4516E+04,
     +  2.4544E+04, 2.4572E+04, 2.4600E+04, 2.4628E+04, 2.4656E+04,
     + 2.4684E+04, 2.4712E+04, 2.4740E+04, 2.4768E+04, 2.4796E+04,
     +  2.4825E+04, 2.4853E+04, 2.4881E+04, 2.4909E+04, 2.4938E+04,
     + 2.4966E+04, 2.4994E+04, 2.5022E+04, 2.5051E+04, 2.5079E+04,
     +  2.5107E+04, 2.5136E+04, 2.5164E+04, 2.5193E+04, 2.5221E+04,
     + 2.5249E+04, 2.5278E+04, 2.5306E+04, 2.5335E+04, 2.5363E+04,
     +  2.5392E+04, 2.5420E+04, 2.5449E+04, 2.5477E+04, 2.5506E+04,
     + 2.5534E+04, 2.5563E+04, 2.5592E+04, 2.5620E+04, 2.5649E+04,
     +  2.5678E+04, 2.5706E+04, 2.5735E+04, 2.5764E+04, 2.5792E+04,
     + 2.5821E+04, 2.5850E+04, 2.5879E+04, 2.5907E+04, 2.5936E+04,
     +  2.5965E+04, 2.5994E+04, 2.6023E+04, 2.6052E+04, 2.6081E+04,
     + 2.6110E+04, 2.6138E+04, 2.6167E+04, 2.6196E+04, 2.6225E+04,
     +  2.6254E+04, 2.6283E+04, 2.6312E+04, 2.6341E+04, 2.6370E+04,
     + 2.6399E+04, 2.6428E+04, 2.6458E+04, 2.6487E+04, 2.6516E+04,
     +  2.6545E+04, 2.6574E+04, 2.6603E+04, 2.6633E+04, 2.6662E+04,
     + 2.6691E+04, 2.6720E+04, 2.6750E+04, 2.6779E+04, 2.6808E+04,
     +  2.6837E+04, 2.6867E+04, 2.6896E+04, 2.6925E+04, 2.6955E+04,
     + 2.6984E+04, 2.7014E+04, 2.7043E+04, 2.7072E+04, 2.7102E+04,
     +  2.7131E+04, 2.7161E+04, 2.7190E+04, 2.7220E+04, 2.7249E+04,
     + 2.7279E+04, 2.7309E+04, 2.7338E+04, 2.7368E+04, 2.7397E+04,
     +  2.7427E+04, 2.7457E+04, 2.7486E+04, 2.7516E+04, 2.7546E+04,
     + 2.7575E+04, 2.7605E+04, 2.7635E+04, 2.7665E+04, 2.7694E+04,
     +  2.7724E+04, 2.7754E+04, 2.7784E+04, 2.7814E+04, 2.7844E+04,
     + 2.7873E+04, 2.7903E+04, 2.7933E+04, 2.7963E+04, 2.7993E+04,
     +  2.8023E+04, 2.8053E+04, 2.8083E+04, 2.8113E+04, 2.8143E+04,
     + 2.8173E+04, 2.8203E+04, 2.8233E+04, 2.8263E+04, 2.8293E+04,
     +  2.8323E+04, 2.8354E+04, 2.8384E+04, 2.8414E+04, 2.8444E+04,
     + 2.8474E+04, 2.8504E+04, 2.8535E+04, 2.8565E+04, 2.8595E+04,
     +  2.8626E+04, 2.8656E+04, 2.8686E+04, 2.8716E+04, 2.8747E+04,
     + 2.8777E+04, 2.8808E+04, 2.8838E+04, 2.8868E+04, 2.8899E+04,
     +  2.8929E+04, 2.8960E+04, 2.8990E+04, 2.9021E+04, 2.9051E+04,
     + 2.9082E+04, 2.9112E+04, 2.9143E+04, 2.9173E+04, 2.9204E+04,
     +  2.9235E+04, 2.9265E+04, 2.9296E+04, 2.9326E+04, 2.9357E+04,
     + 2.9388E+04, 2.9419E+04, 2.9449E+04, 2.9480E+04, 2.9511E+04,
     +  2.9542E+04, 2.9572E+04, 2.9603E+04, 2.9634E+04, 2.9665E+04,
     + 2.9696E+04, 2.9726E+04, 2.9757E+04, 2.9788E+04, 2.9819E+04,
     +  2.9850E+04, 2.9881E+04, 2.9912E+04, 2.9943E+04, 2.9974E+04,
     + 3.0005E+04, 3.0036E+04, 3.0067E+04, 3.0098E+04, 3.0129E+04,
     +  3.0160E+04, 3.0191E+04, 3.0222E+04, 3.0254E+04, 3.0285E+04,
     + 3.0316E+04, 3.0347E+04, 3.0378E+04, 3.0409E+04, 3.0441E+04,
     +  3.0472E+04, 3.0503E+04, 3.0535E+04, 3.0566E+04, 3.0597E+04,
     + 3.0628E+04, 3.0660E+04, 3.0691E+04, 3.0723E+04, 3.0754E+04,
     +  3.0785E+04, 3.0817E+04, 3.0848E+04, 3.0880E+04, 3.0911E+04,
     + 3.0943E+04, 3.0974E+04, 3.1006E+04, 3.1037E+04, 3.1069E+04,
     +  3.1100E+04, 3.1132E+04, 3.1164E+04, 3.1195E+04, 3.1227E+04,
     + 3.1259E+04, 3.1290E+04, 3.1322E+04, 3.1354E+04, 3.1385E+04,
     +  3.1417E+04, 3.1449E+04, 3.1481E+04, 3.1512E+04, 3.1544E+04,
     + 3.1576E+04, 3.1608E+04, 3.1640E+04, 3.1672E+04, 3.1703E+04,
     +  3.1735E+04, 3.1767E+04, 3.1799E+04, 3.1831E+04, 3.1863E+04,
     + 3.1895E+04, 3.1927E+04, 3.1959E+04, 3.1991E+04, 3.2023E+04,
     +  3.2055E+04, 3.2087E+04, 3.2120E+04, 3.2152E+04, 3.2184E+04,
     + 3.2216E+04, 3.2248E+04, 3.2280E+04, 3.2312E+04, 3.2344E+04,
     +  3.2377E+04, 3.2409E+04, 3.2441E+04, 3.2474E+04, 3.2506E+04,
     + 3.2538E+04, 3.2570E+04, 3.2603E+04, 3.2635E+04, 3.2668E+04,
     +  3.2700E+04, 3.2732E+04, 3.2765E+04, 3.2797E+04, 3.2830E+04,
     + 3.2862E+04, 3.2895E+04, 3.2927E+04, 3.2960E+04, 3.2992E+04,
     +  3.3025E+04, 3.3057E+04, 3.3090E+04, 3.3123E+04, 3.3155E+04,
     + 3.3188E+04, 3.3220E+04, 3.3253E+04, 3.3286E+04, 3.3318E+04,
     +  3.3351E+04, 3.3384E+04, 3.3416E+04, 3.3449E+04, 3.3482E+04,
     + 3.3515E+04, 3.3548E+04, 3.3580E+04, 3.3613E+04, 3.3646E+04,
     +  3.3679E+04, 3.3712E+04, 3.3745E+04, 3.3778E+04, 3.3811E+04,
     + 3.3844E+04, 3.3877E+04, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 3,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_SO2   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 2),QofT( 2,NT),Q(NT)
      data (Tmx(I),I=1,2)/3500.,3500./
      data xgj/ 1.,1./
c...      SO2
c...        --       626
      data (QofT( 1,J),J=1,2408)/ 1.4543E+00, 3.4580E+00, 6.1847E+00,
     +  9.4073E+00, 1.3053E+01, 1.7077E+01, 2.1447E+01, 2.6136E+01,
     + 3.1126E+01, 3.6398E+01, 4.1938E+01, 4.7734E+01, 5.3775E+01,
     +  6.0051E+01, 6.6555E+01, 7.3277E+01, 8.0213E+01, 8.7354E+01,
     + 9.4695E+01, 1.0223E+02, 1.1787E+02, 1.3424E+02, 1.5129E+02,
     +  1.6902E+02, 1.8739E+02, 2.0638E+02, 2.2598E+02, 2.4616E+02,
     + 2.6690E+02, 2.8820E+02, 3.1004E+02, 3.3240E+02, 3.5528E+02,
     +  3.7866E+02, 4.0253E+02, 4.2689E+02, 4.5172E+02, 4.7701E+02,
     + 5.0276E+02, 5.2896E+02, 5.5560E+02, 5.8267E+02, 6.1017E+02,
     +  6.3812E+02, 6.6646E+02, 6.9520E+02, 7.2436E+02, 7.5391E+02,
     + 7.8385E+02, 8.1419E+02, 8.4491E+02, 8.7601E+02, 9.0750E+02,
     +  9.3936E+02, 9.7159E+02, 1.0042E+03, 1.0372E+03, 1.0705E+03,
     + 1.1042E+03, 1.1382E+03, 1.1727E+03, 1.2074E+03, 1.2426E+03,
     +  1.2780E+03, 1.3139E+03, 1.3501E+03, 1.3866E+03, 1.4235E+03,
     + 1.4608E+03, 1.4984E+03, 1.5363E+03, 1.5746E+03, 1.6133E+03,
     +  1.6523E+03, 1.6916E+03, 1.7313E+03, 1.7714E+03, 1.8118E+03,
     + 1.8526E+03, 1.8937E+03, 1.9351E+03, 1.9770E+03, 2.0192E+03,
     +  2.0617E+03, 2.1046E+03, 2.1479E+03, 2.1915E+03, 2.2355E+03,
     + 2.2799E+03, 2.3246E+03, 2.3698E+03, 2.4152E+03, 2.4611E+03,
     +  2.5073E+03, 2.5539E+03, 2.6008E+03, 2.6482E+03, 2.6959E+03,
     + 2.7440E+03, 2.7925E+03, 2.8414E+03, 2.8907E+03, 2.9403E+03,
     +  2.9904E+03, 3.0408E+03, 3.0917E+03, 3.1429E+03, 3.1946E+03,
     + 3.2466E+03, 3.2991E+03, 3.3519E+03, 3.4052E+03, 3.4589E+03,
     +  3.5130E+03, 3.5675E+03, 3.6224E+03, 3.6777E+03, 3.7335E+03,
     + 3.7897E+03, 3.8463E+03, 3.9034E+03, 3.9608E+03, 4.0188E+03,
     +  4.0771E+03, 4.1359E+03, 4.1951E+03, 4.2548E+03, 4.3149E+03,
     + 4.3755E+03, 4.4365E+03, 4.4981E+03, 4.5600E+03, 4.6224E+03,
     +  4.6853E+03, 4.7486E+03, 4.8124E+03, 4.8767E+03, 4.9414E+03,
     + 5.0066E+03, 5.0723E+03, 5.1385E+03, 5.2052E+03, 5.2724E+03,
     +  5.3400E+03, 5.4081E+03, 5.4768E+03, 5.5459E+03, 5.6155E+03,
     + 5.6857E+03, 5.7563E+03, 5.8275E+03, 5.8992E+03, 5.9714E+03,
     +  6.0441E+03, 6.1174E+03, 6.1912E+03, 6.2654E+03, 6.3403E+03,
     + 6.4157E+03, 6.4915E+03, 6.5680E+03, 6.6450E+03, 6.7226E+03,
     +  6.8007E+03, 6.8794E+03, 6.9585E+03, 7.0383E+03, 7.1187E+03,
     + 7.1996E+03, 7.2811E+03, 7.3632E+03, 7.4458E+03, 7.5290E+03,
     +  7.6129E+03, 7.6972E+03, 7.7823E+03, 7.8678E+03, 7.9540E+03,
     + 8.0408E+03, 8.1283E+03, 8.2162E+03, 8.3049E+03, 8.3942E+03,
     +  8.4840E+03, 8.5746E+03, 8.6657E+03, 8.7574E+03, 8.8498E+03,
     + 8.9428E+03, 9.0366E+03, 9.1309E+03, 9.2258E+03, 9.3215E+03,
     +  9.4178E+03, 9.5147E+03, 9.6124E+03, 9.7107E+03, 9.8096E+03,
     + 9.9093E+03, 1.0010E+04, 1.0111E+04, 1.0212E+04, 1.0315E+04,
     +  1.0418E+04, 1.0522E+04, 1.0626E+04, 1.0731E+04, 1.0837E+04,
     + 1.0944E+04, 1.1051E+04, 1.1159E+04, 1.1268E+04, 1.1378E+04,
     +  1.1488E+04, 1.1599E+04, 1.1711E+04, 1.1824E+04, 1.1937E+04,
     + 1.2051E+04, 1.2166E+04, 1.2281E+04, 1.2398E+04, 1.2515E+04,
     +  1.2633E+04, 1.2751E+04, 1.2871E+04, 1.2991E+04, 1.3112E+04,
     + 1.3234E+04, 1.3357E+04, 1.3480E+04, 1.3605E+04, 1.3730E+04,
     +  1.3856E+04, 1.3982E+04, 1.4110E+04, 1.4239E+04, 1.4368E+04,
     + 1.4498E+04, 1.4629E+04, 1.4761E+04, 1.4894E+04, 1.5027E+04,
     +  1.5162E+04, 1.5297E+04, 1.5433E+04, 1.5570E+04, 1.5708E+04,
     + 1.5847E+04, 1.5987E+04, 1.6127E+04, 1.6269E+04, 1.6411E+04,
     +  1.6555E+04, 1.6699E+04, 1.6844E+04, 1.6990E+04, 1.7137E+04,
     + 1.7285E+04, 1.7434E+04, 1.7584E+04, 1.7735E+04, 1.7887E+04,
     +  1.8040E+04, 1.8193E+04, 1.8348E+04, 1.8504E+04, 1.8661E+04,
     + 1.8818E+04, 1.8977E+04, 1.9137E+04, 1.9297E+04, 1.9459E+04,
     +  1.9622E+04, 1.9785E+04, 1.9950E+04, 2.0116E+04, 2.0282E+04,
     + 2.0450E+04, 2.0619E+04, 2.0789E+04, 2.0960E+04, 2.1132E+04,
     +  2.1305E+04, 2.1479E+04, 2.1654E+04, 2.1830E+04, 2.2008E+04,
     + 2.2186E+04, 2.2366E+04, 2.2546E+04, 2.2728E+04, 2.2911E+04,
     +  2.3095E+04, 2.3280E+04, 2.3466E+04, 2.3654E+04, 2.3842E+04,
     + 2.4032E+04, 2.4222E+04, 2.4414E+04, 2.4607E+04, 2.4802E+04,
     +  2.4997E+04, 2.5194E+04, 2.5391E+04, 2.5590E+04, 2.5790E+04,
     + 2.5992E+04, 2.6194E+04, 2.6398E+04, 2.6603E+04, 2.6809E+04,
     +  2.7016E+04, 2.7225E+04, 2.7435E+04, 2.7646E+04, 2.7858E+04,
     + 2.8072E+04, 2.8286E+04, 2.8502E+04, 2.8720E+04, 2.8938E+04,
     +  2.9158E+04, 2.9379E+04, 2.9602E+04, 2.9826E+04, 3.0051E+04,
     + 3.0277E+04, 3.0505E+04, 3.0734E+04, 3.0964E+04, 3.1196E+04,
     +  3.1429E+04, 3.1663E+04, 3.1899E+04, 3.2136E+04, 3.2374E+04,
     + 3.2614E+04, 3.2855E+04, 3.3097E+04, 3.3341E+04, 3.3586E+04,
     +  3.3833E+04, 3.4081E+04, 3.4331E+04, 3.4582E+04, 3.4834E+04,
     + 3.5087E+04, 3.5343E+04, 3.5599E+04, 3.5857E+04, 3.6117E+04,
     +  3.6378E+04, 3.6640E+04, 3.6904E+04, 3.7170E+04, 3.7436E+04,
     + 3.7705E+04, 3.7974E+04, 3.8246E+04, 3.8519E+04, 3.8793E+04,
     +  3.9069E+04, 3.9346E+04, 3.9625E+04, 3.9905E+04, 4.0187E+04,
     + 4.0471E+04, 4.0756E+04, 4.1043E+04, 4.1331E+04, 4.1621E+04,
     +  4.1912E+04, 4.2205E+04, 4.2499E+04, 4.2796E+04, 4.3093E+04,
     + 4.3393E+04, 4.3694E+04, 4.3996E+04, 4.4301E+04, 4.4607E+04,
     +  4.4914E+04, 4.5223E+04, 4.5534E+04, 4.5847E+04, 4.6161E+04,
     + 4.6477E+04, 4.6794E+04, 4.7113E+04, 4.7435E+04, 4.7757E+04,
     +  4.8082E+04, 4.8408E+04, 4.8736E+04, 4.9065E+04, 4.9396E+04,
     + 4.9729E+04, 5.0064E+04, 5.0401E+04, 5.0739E+04, 5.1079E+04,
     +  5.1421E+04, 5.1765E+04, 5.2110E+04, 5.2458E+04, 5.2807E+04,
     + 5.3158E+04, 5.3510E+04, 5.3865E+04, 5.4221E+04, 5.4580E+04,
     +  5.4940E+04, 5.5302E+04, 5.5665E+04, 5.6031E+04, 5.6399E+04,
     + 5.6768E+04, 5.7140E+04, 5.7513E+04, 5.7888E+04, 5.8265E+04,
     +  5.8644E+04, 5.9025E+04, 5.9408E+04, 5.9793E+04, 6.0180E+04,
     + 6.0569E+04, 6.0959E+04, 6.1352E+04, 6.1747E+04, 6.2143E+04,
     +  6.2542E+04, 6.2943E+04, 6.3345E+04, 6.3750E+04, 6.4157E+04,
     + 6.4566E+04, 6.4977E+04, 6.5390E+04, 6.5805E+04, 6.6222E+04,
     +  6.6641E+04, 6.7062E+04, 6.7485E+04, 6.7910E+04, 6.8338E+04,
     + 6.8767E+04, 6.9199E+04, 6.9633E+04, 7.0069E+04, 7.0507E+04,
     +  7.0947E+04, 7.1390E+04, 7.1834E+04, 7.2281E+04, 7.2730E+04,
     + 7.3181E+04, 7.3634E+04, 7.4090E+04, 7.4547E+04, 7.5007E+04,
     +  7.5469E+04, 7.5934E+04, 7.6400E+04, 7.6869E+04, 7.7341E+04,
     + 7.7814E+04, 7.8290E+04, 7.8768E+04, 7.9248E+04, 7.9731E+04,
     +  8.0215E+04, 8.0702E+04, 8.1192E+04, 8.1684E+04, 8.2178E+04,
     + 8.2675E+04, 8.3174E+04, 8.3675E+04, 8.4178E+04, 8.4684E+04,
     +  8.5193E+04, 8.5704E+04, 8.6217E+04, 8.6733E+04, 8.7251E+04,
     + 8.7771E+04, 8.8294E+04, 8.8820E+04, 8.9347E+04, 8.9878E+04,
     +  9.0410E+04, 9.0946E+04, 9.1483E+04, 9.2024E+04, 9.2567E+04,
     + 9.3112E+04, 9.3660E+04, 9.4210E+04, 9.4763E+04, 9.5318E+04,
     +  9.5876E+04, 9.6437E+04, 9.7000E+04, 9.7566E+04, 9.8134E+04,
     + 9.8705E+04, 9.9279E+04, 9.9855E+04, 1.0043E+05, 1.0101E+05,
     +  1.0160E+05, 1.0219E+05, 1.0278E+05, 1.0337E+05, 1.0396E+05,
     + 1.0456E+05, 1.0516E+05, 1.0576E+05, 1.0637E+05, 1.0698E+05,
     +  1.0759E+05, 1.0820E+05, 1.0882E+05, 1.0944E+05, 1.1006E+05,
     + 1.1068E+05, 1.1131E+05, 1.1194E+05, 1.1258E+05, 1.1321E+05,
     +  1.1385E+05, 1.1449E+05, 1.1514E+05, 1.1578E+05, 1.1643E+05,
     + 1.1709E+05, 1.1774E+05, 1.1840E+05, 1.1906E+05, 1.1973E+05,
     +  1.2040E+05, 1.2107E+05, 1.2174E+05, 1.2242E+05, 1.2310E+05,
     + 1.2378E+05, 1.2446E+05, 1.2515E+05, 1.2584E+05, 1.2654E+05,
     +  1.2724E+05, 1.2794E+05, 1.2864E+05, 1.2935E+05, 1.3006E+05,
     + 1.3077E+05, 1.3148E+05, 1.3220E+05, 1.3292E+05, 1.3365E+05,
     +  1.3438E+05, 1.3511E+05, 1.3584E+05, 1.3658E+05, 1.3732E+05,
     + 1.3806E+05, 1.3881E+05, 1.3956E+05, 1.4031E+05, 1.4107E+05,
     +  1.4183E+05, 1.4259E+05, 1.4336E+05, 1.4413E+05, 1.4490E+05,
     + 1.4568E+05, 1.4646E+05, 1.4724E+05, 1.4802E+05, 1.4881E+05,
     +  1.4961E+05, 1.5040E+05, 1.5120E+05, 1.5200E+05, 1.5281E+05,
     + 1.5362E+05, 1.5443E+05, 1.5525E+05, 1.5606E+05, 1.5689E+05,
     +  1.5771E+05, 1.5854E+05, 1.5938E+05, 1.6021E+05, 1.6105E+05,
     + 1.6189E+05, 1.6274E+05, 1.6359E+05, 1.6444E+05, 1.6530E+05,
     +  1.6616E+05, 1.6703E+05, 1.6789E+05, 1.6877E+05, 1.6964E+05,
     + 1.7052E+05, 1.7140E+05, 1.7229E+05, 1.7318E+05, 1.7407E+05,
     +  1.7496E+05, 1.7586E+05, 1.7677E+05, 1.7768E+05, 1.7859E+05,
     + 1.7950E+05, 1.8042E+05, 1.8134E+05, 1.8227E+05, 1.8320E+05,
     +  1.8413E+05, 1.8507E+05, 1.8601E+05, 1.8695E+05, 1.8790E+05,
     + 1.8885E+05, 1.8981E+05, 1.9077E+05, 1.9173E+05, 1.9270E+05,
     +  1.9367E+05, 1.9464E+05, 1.9562E+05, 1.9660E+05, 1.9759E+05,
     + 1.9858E+05, 1.9957E+05, 2.0057E+05, 2.0157E+05, 2.0258E+05,
     +  2.0359E+05, 2.0460E+05, 2.0562E+05, 2.0664E+05, 2.0767E+05,
     + 2.0870E+05, 2.0973E+05, 2.1077E+05, 2.1181E+05, 2.1286E+05,
     +  2.1391E+05, 2.1496E+05, 2.1602E+05, 2.1708E+05, 2.1815E+05,
     + 2.1922E+05, 2.2030E+05, 2.2137E+05, 2.2246E+05, 2.2354E+05,
     +  2.2464E+05, 2.2573E+05, 2.2683E+05, 2.2793E+05, 2.2904E+05,
     + 2.3016E+05, 2.3127E+05, 2.3239E+05, 2.3352E+05, 2.3465E+05,
     +  2.3578E+05, 2.3692E+05, 2.3806E+05, 2.3921E+05, 2.4036E+05,
     + 2.4151E+05, 2.4267E+05, 2.4384E+05, 2.4501E+05, 2.4618E+05,
     +  2.4736E+05, 2.4854E+05, 2.4972E+05, 2.5091E+05, 2.5211E+05,
     + 2.5331E+05, 2.5451E+05, 2.5572E+05, 2.5694E+05, 2.5815E+05,
     +  2.5937E+05, 2.6060E+05, 2.6183E+05, 2.6307E+05, 2.6431E+05,
     + 2.6555E+05, 2.6680E+05, 2.6806E+05, 2.6932E+05, 2.7058E+05,
     +  2.7185E+05, 2.7312E+05, 2.7440E+05, 2.7568E+05, 2.7697E+05,
     + 2.7826E+05, 2.7955E+05, 2.8086E+05, 2.8216E+05, 2.8347E+05,
     +  2.8479E+05, 2.8611E+05, 2.8743E+05, 2.8876E+05, 2.9010E+05,
     + 2.9144E+05, 2.9278E+05, 2.9413E+05, 2.9548E+05, 2.9684E+05,
     +  2.9821E+05, 2.9958E+05, 3.0095E+05, 3.0233E+05, 3.0371E+05,
     + 3.0510E+05, 3.0649E+05, 3.0789E+05, 3.0930E+05, 3.1071E+05,
     +  3.1212E+05, 3.1354E+05, 3.1496E+05, 3.1639E+05, 3.1782E+05,
     + 3.1926E+05, 3.2071E+05, 3.2216E+05, 3.2361E+05, 3.2507E+05,
     +  3.2654E+05, 3.2801E+05, 3.2948E+05, 3.3096E+05, 3.3245E+05,
     + 3.3394E+05, 3.3543E+05, 3.3694E+05, 3.3844E+05, 3.3995E+05,
     +  3.4147E+05, 3.4299E+05, 3.4452E+05, 3.4606E+05, 3.4759E+05,
     + 3.4914E+05, 3.5069E+05, 3.5224E+05, 3.5380E+05, 3.5537E+05,
     +  3.5694E+05, 3.5852E+05, 3.6010E+05, 3.6169E+05, 3.6328E+05,
     + 3.6488E+05, 3.6648E+05, 3.6809E+05, 3.6971E+05, 3.7133E+05,
     +  3.7295E+05, 3.7459E+05, 3.7622E+05, 3.7787E+05, 3.7951E+05,
     + 3.8117E+05, 3.8283E+05, 3.8449E+05, 3.8617E+05, 3.8784E+05,
     +  3.8953E+05, 3.9121E+05, 3.9291E+05, 3.9461E+05, 3.9631E+05,
     + 3.9803E+05, 3.9974E+05, 4.0147E+05, 4.0320E+05, 4.0493E+05,
     +  4.0667E+05, 4.0842E+05, 4.1017E+05, 4.1193E+05, 4.1369E+05,
     + 4.1546E+05, 4.1724E+05, 4.1902E+05, 4.2081E+05, 4.2261E+05,
     +  4.2441E+05, 4.2621E+05, 4.2802E+05, 4.2984E+05, 4.3167E+05,
     + 4.3350E+05, 4.3533E+05, 4.3718E+05, 4.3903E+05, 4.4088E+05,
     +  4.4274E+05, 4.4461E+05, 4.4648E+05, 4.4836E+05, 4.5025E+05,
     + 4.5214E+05, 4.5404E+05, 4.5594E+05, 4.5785E+05, 4.5977E+05,
     +  4.6169E+05, 4.6362E+05, 4.6556E+05, 4.6750E+05, 4.6945E+05,
     + 4.7141E+05, 4.7337E+05, 4.7534E+05, 4.7731E+05, 4.7929E+05,
     +  4.8128E+05, 4.8327E+05, 4.8527E+05, 4.8728E+05, 4.8929E+05,
     + 4.9131E+05, 4.9334E+05, 4.9537E+05, 4.9741E+05, 4.9946E+05,
     +  5.0151E+05, 5.0357E+05, 5.0563E+05, 5.0771E+05, 5.0979E+05,
     + 5.1187E+05, 5.1396E+05, 5.1606E+05, 5.1817E+05, 5.2028E+05,
     +  5.2240E+05, 5.2453E+05, 5.2666E+05, 5.2880E+05, 5.3095E+05,
     + 5.3310E+05, 5.3526E+05, 5.3743E+05, 5.3960E+05, 5.4178E+05,
     +  5.4397E+05, 5.4617E+05, 5.4837E+05, 5.5058E+05, 5.5279E+05,
     + 5.5502E+05, 5.5725E+05, 5.5948E+05, 5.6173E+05, 5.6398E+05,
     +  5.6623E+05, 5.6850E+05, 5.7077E+05, 5.7305E+05, 5.7534E+05,
     + 5.7763E+05, 5.7993E+05, 5.8224E+05, 5.8456E+05, 5.8688E+05,
     +  5.8921E+05, 5.9154E+05, 5.9389E+05, 5.9624E+05, 5.9860E+05,
     + 6.0096E+05, 6.0334E+05, 6.0572E+05, 6.0811E+05, 6.1050E+05,
     +  6.1291E+05, 6.1532E+05, 6.1773E+05, 6.2016E+05, 6.2259E+05,
     + 6.2503E+05, 6.2748E+05, 6.2994E+05, 6.3240E+05, 6.3487E+05,
     +  6.3735E+05, 6.3983E+05, 6.4232E+05, 6.4483E+05, 6.4733E+05,
     + 6.4985E+05, 6.5237E+05, 6.5491E+05, 6.5744E+05, 6.5999E+05,
     +  6.6255E+05, 6.6511E+05, 6.6768E+05, 6.7026E+05, 6.7284E+05,
     + 6.7544E+05, 6.7804E+05, 6.8065E+05, 6.8326E+05, 6.8589E+05,
     +  6.8852E+05, 6.9116E+05, 6.9381E+05, 6.9647E+05, 6.9913E+05,
     + 7.0180E+05, 7.0448E+05, 7.0717E+05, 7.0987E+05, 7.1258E+05,
     +  7.1529E+05, 7.1801E+05, 7.2074E+05, 7.2348E+05, 7.2622E+05,
     + 7.2897E+05, 7.3174E+05, 7.3451E+05, 7.3728E+05, 7.4007E+05,
     +  7.4287E+05, 7.4567E+05, 7.4848E+05, 7.5130E+05, 7.5413E+05,
     + 7.5696E+05, 7.5981E+05, 7.6266E+05, 7.6552E+05, 7.6839E+05,
     +  7.7127E+05, 7.7416E+05, 7.7705E+05, 7.7996E+05, 7.8287E+05,
     + 7.8579E+05, 7.8872E+05, 7.9165E+05, 7.9460E+05, 7.9756E+05,
     +  8.0052E+05, 8.0349E+05, 8.0647E+05, 8.0946E+05, 8.1246E+05,
     + 8.1547E+05, 8.1848E+05, 8.2151E+05, 8.2454E+05, 8.2758E+05,
     +  8.3063E+05, 8.3369E+05, 8.3676E+05, 8.3984E+05, 8.4293E+05,
     + 8.4602E+05, 8.4912E+05, 8.5224E+05, 8.5536E+05, 8.5849E+05,
     +  8.6163E+05, 8.6478E+05, 8.6794E+05, 8.7110E+05, 8.7428E+05,
     + 8.7746E+05, 8.8066E+05, 8.8386E+05, 8.8707E+05, 8.9029E+05,
     +  8.9352E+05, 8.9676E+05, 9.0001E+05, 9.0327E+05, 9.0654E+05,
     + 9.0982E+05, 9.1310E+05, 9.1640E+05, 9.1970E+05, 9.2302E+05,
     +  9.2634E+05, 9.2967E+05, 9.3301E+05, 9.3637E+05, 9.3973E+05,
     + 9.4310E+05, 9.4648E+05, 9.4987E+05, 9.5327E+05, 9.5667E+05,
     +  9.6009E+05, 9.6352E+05, 9.6696E+05, 9.7040E+05, 9.7386E+05,
     + 9.7733E+05, 9.8080E+05, 9.8429E+05, 9.8778E+05, 9.9129E+05,
     +  9.9480E+05, 9.9833E+05, 1.0019E+06, 1.0054E+06, 1.0090E+06,
     + 1.0125E+06, 1.0161E+06, 1.0197E+06, 1.0233E+06, 1.0269E+06,
     +  1.0305E+06, 1.0341E+06, 1.0377E+06, 1.0414E+06, 1.0450E+06,
     + 1.0487E+06, 1.0524E+06, 1.0561E+06, 1.0598E+06, 1.0635E+06,
     +  1.0672E+06, 1.0709E+06, 1.0746E+06, 1.0784E+06, 1.0821E+06,
     + 1.0859E+06, 1.0897E+06, 1.0935E+06, 1.0972E+06, 1.1011E+06,
     +  1.1049E+06, 1.1087E+06, 1.1125E+06, 1.1164E+06, 1.1202E+06,
     + 1.1241E+06, 1.1280E+06, 1.1319E+06, 1.1358E+06, 1.1397E+06,
     +  1.1436E+06, 1.1475E+06, 1.1515E+06, 1.1554E+06, 1.1594E+06,
     + 1.1634E+06, 1.1674E+06, 1.1714E+06, 1.1754E+06, 1.1794E+06,
     +  1.1834E+06, 1.1874E+06, 1.1915E+06, 1.1955E+06, 1.1996E+06,
     + 1.2037E+06, 1.2078E+06, 1.2119E+06, 1.2160E+06, 1.2201E+06,
     +  1.2243E+06, 1.2284E+06, 1.2326E+06, 1.2367E+06, 1.2409E+06,
     + 1.2451E+06, 1.2493E+06, 1.2535E+06, 1.2578E+06, 1.2620E+06,
     +  1.2662E+06, 1.2705E+06, 1.2748E+06, 1.2791E+06, 1.2833E+06,
     + 1.2876E+06, 1.2920E+06, 1.2963E+06, 1.3006E+06, 1.3050E+06,
     +  1.3093E+06, 1.3137E+06, 1.3181E+06, 1.3225E+06, 1.3269E+06,
     + 1.3313E+06, 1.3357E+06, 1.3402E+06, 1.3446E+06, 1.3491E+06,
     +  1.3536E+06, 1.3580E+06, 1.3625E+06, 1.3671E+06, 1.3716E+06,
     + 1.3761E+06, 1.3807E+06, 1.3852E+06, 1.3898E+06, 1.3944E+06,
     +  1.3990E+06, 1.4036E+06, 1.4082E+06, 1.4128E+06, 1.4174E+06,
     + 1.4221E+06, 1.4268E+06, 1.4314E+06, 1.4361E+06, 1.4408E+06,
     +  1.4455E+06, 1.4503E+06, 1.4550E+06, 1.4597E+06, 1.4645E+06,
     + 1.4693E+06, 1.4741E+06, 1.4789E+06, 1.4837E+06, 1.4885E+06,
     +  1.4933E+06, 1.4982E+06, 1.5030E+06, 1.5079E+06, 1.5128E+06,
     + 1.5177E+06, 1.5226E+06, 1.5275E+06, 1.5324E+06, 1.5374E+06,
     +  1.5423E+06, 1.5473E+06, 1.5523E+06, 1.5573E+06, 1.5623E+06,
     + 1.5673E+06, 1.5723E+06, 1.5774E+06, 1.5825E+06, 1.5875E+06,
     +  1.5926E+06, 1.5977E+06, 1.6028E+06, 1.6079E+06, 1.6131E+06,
     + 1.6182E+06, 1.6234E+06, 1.6286E+06, 1.6337E+06, 1.6389E+06,
     +  1.6441E+06, 1.6494E+06, 1.6546E+06, 1.6599E+06, 1.6651E+06,
     + 1.6704E+06, 1.6757E+06, 1.6810E+06, 1.6863E+06, 1.6916E+06,
     +  1.6970E+06, 1.7023E+06, 1.7077E+06, 1.7131E+06, 1.7185E+06,
     + 1.7239E+06, 1.7293E+06, 1.7348E+06, 1.7402E+06, 1.7457E+06,
     +  1.7512E+06, 1.7566E+06, 1.7621E+06, 1.7677E+06, 1.7732E+06,
     + 1.7787E+06, 1.7843E+06, 1.7899E+06, 1.7955E+06, 1.8011E+06,
     +  1.8067E+06, 1.8123E+06, 1.8179E+06, 1.8236E+06, 1.8292E+06,
     + 1.8349E+06, 1.8406E+06, 1.8463E+06, 1.8521E+06, 1.8578E+06,
     +  1.8635E+06, 1.8693E+06, 1.8751E+06, 1.8809E+06, 1.8867E+06,
     + 1.8925E+06, 1.8983E+06, 1.9042E+06, 1.9100E+06, 1.9159E+06,
     +  1.9218E+06, 1.9277E+06, 1.9336E+06, 1.9396E+06, 1.9455E+06,
     + 1.9515E+06, 1.9575E+06, 1.9634E+06, 1.9695E+06, 1.9755E+06,
     +  1.9815E+06, 1.9876E+06, 1.9936E+06, 1.9997E+06, 2.0058E+06,
     + 2.0119E+06, 2.0180E+06, 2.0241E+06, 2.0303E+06, 2.0365E+06,
     +  2.0426E+06, 2.0488E+06, 2.0550E+06, 2.0613E+06, 2.0675E+06,
     + 2.0738E+06, 2.0800E+06, 2.0863E+06, 2.0926E+06, 2.0989E+06,
     +  2.1052E+06, 2.1116E+06, 2.1179E+06, 2.1243E+06, 2.1307E+06,
     + 2.1371E+06, 2.1435E+06, 2.1500E+06, 2.1564E+06, 2.1629E+06,
     +  2.1693E+06, 2.1758E+06, 2.1823E+06, 2.1889E+06, 2.1954E+06,
     + 2.2020E+06, 2.2085E+06, 2.2151E+06, 2.2217E+06, 2.2283E+06,
     +  2.2350E+06, 2.2416E+06, 2.2483E+06, 2.2549E+06, 2.2616E+06,
     + 2.2683E+06, 2.2751E+06, 2.2818E+06, 2.2886E+06, 2.2953E+06,
     +  2.3021E+06, 2.3089E+06, 2.3157E+06, 2.3226E+06, 2.3294E+06,
     + 2.3363E+06, 2.3432E+06, 2.3501E+06, 2.3570E+06, 2.3639E+06,
     +  2.3708E+06, 2.3778E+06, 2.3848E+06, 2.3918E+06, 2.3988E+06,
     + 2.4058E+06, 2.4129E+06, 2.4199E+06, 2.4270E+06, 2.4341E+06,
     +  2.4412E+06, 2.4483E+06, 2.4554E+06, 2.4626E+06, 2.4698E+06,
     + 2.4769E+06, 2.4842E+06, 2.4914E+06, 2.4986E+06, 2.5059E+06,
     +  2.5131E+06, 2.5204E+06, 2.5277E+06, 2.5350E+06, 2.5424E+06,
     + 2.5497E+06, 2.5571E+06, 2.5645E+06, 2.5719E+06, 2.5793E+06,
     +  2.5867E+06, 2.5942E+06, 2.6017E+06, 2.6091E+06, 2.6166E+06,
     + 2.6242E+06, 2.6317E+06, 2.6393E+06, 2.6468E+06, 2.6544E+06,
     +  2.6620E+06, 2.6696E+06, 2.6773E+06, 2.6849E+06, 2.6926E+06,
     + 2.7003E+06, 2.7080E+06, 2.7157E+06, 2.7235E+06, 2.7312E+06,
     +  2.7390E+06, 2.7468E+06, 2.7546E+06, 2.7624E+06, 2.7703E+06,
     + 2.7781E+06, 2.7860E+06, 2.7939E+06, 2.8018E+06, 2.8098E+06,
     +  2.8177E+06, 2.8257E+06, 2.8337E+06, 2.8417E+06, 2.8497E+06,
     + 2.8577E+06, 2.8658E+06, 2.8739E+06, 2.8820E+06, 2.8901E+06,
     +  2.8982E+06, 2.9064E+06, 2.9145E+06, 2.9227E+06, 2.9309E+06,
     + 2.9391E+06, 2.9474E+06, 2.9556E+06, 2.9639E+06, 2.9722E+06,
     +  2.9805E+06, 2.9888E+06, 2.9972E+06, 3.0055E+06, 3.0139E+06,
     + 3.0223E+06, 3.0307E+06, 3.0392E+06, 3.0476E+06, 3.0561E+06,
     +  3.0646E+06, 3.0731E+06, 3.0816E+06, 3.0902E+06, 3.0987E+06,
     + 3.1073E+06, 3.1159E+06, 3.1245E+06, 3.1332E+06, 3.1418E+06,
     +  3.1505E+06, 3.1592E+06, 3.1679E+06, 3.1767E+06, 3.1854E+06,
     + 3.1942E+06, 3.2030E+06, 3.2118E+06, 3.2206E+06, 3.2295E+06,
     +  3.2383E+06, 3.2472E+06, 3.2561E+06, 3.2650E+06, 3.2740E+06,
     + 3.2829E+06, 3.2919E+06, 3.3009E+06, 3.3100E+06, 3.3190E+06,
     +  3.3280E+06, 3.3371E+06, 3.3462E+06, 3.3553E+06, 3.3645E+06,
     + 3.3736E+06, 3.3828E+06, 3.3920E+06, 3.4012E+06, 3.4105E+06,
     +  3.4197E+06, 3.4290E+06, 3.4383E+06, 3.4476E+06, 3.4569E+06,
     + 3.4663E+06, 3.4756E+06, 3.4850E+06, 3.4944E+06, 3.5039E+06,
     +  3.5133E+06, 3.5228E+06, 3.5323E+06, 3.5418E+06, 3.5513E+06,
     + 3.5609E+06, 3.5705E+06, 3.5801E+06, 3.5897E+06, 3.5993E+06,
     +  3.6089E+06, 3.6186E+06, 3.6283E+06, 3.6380E+06, 3.6478E+06,
     + 3.6575E+06, 3.6673E+06, 3.6771E+06, 3.6869E+06, 3.6967E+06,
     +  3.7066E+06, 3.7165E+06, 3.7264E+06, 3.7363E+06, 3.7462E+06,
     + 3.7562E+06, 3.7662E+06, 3.7762E+06, 3.7862E+06, 3.7962E+06,
     +  3.8063E+06, 3.8164E+06, 3.8265E+06, 3.8366E+06, 3.8468E+06,
     + 3.8569E+06, 3.8671E+06, 3.8773E+06, 3.8876E+06, 3.8978E+06,
     +  3.9081E+06, 3.9184E+06, 3.9287E+06, 3.9390E+06, 3.9494E+06,
     + 3.9598E+06, 3.9702E+06, 3.9806E+06, 3.9910E+06, 4.0015E+06,
     +  4.0120E+06, 4.0225E+06, 4.0330E+06, 4.0436E+06, 4.0542E+06,
     + 4.0648E+06, 4.0754E+06, 4.0860E+06, 4.0967E+06, 4.1074E+06,
     +  4.1181E+06, 4.1288E+06, 4.1395E+06, 4.1503E+06, 4.1611E+06,
     + 4.1719E+06, 4.1827E+06, 4.1936E+06, 4.2045E+06, 4.2154E+06,
     +  4.2263E+06, 4.2372E+06, 4.2482E+06, 4.2592E+06, 4.2702E+06,
     + 4.2813E+06, 4.2923E+06, 4.3034E+06, 4.3145E+06, 4.3256E+06,
     +  4.3368E+06, 4.3479E+06, 4.3591E+06, 4.3703E+06, 4.3816E+06,
     + 4.3928E+06, 4.4041E+06, 4.4154E+06, 4.4268E+06, 4.4381E+06,
     +  4.4495E+06, 4.4609E+06, 4.4723E+06, 4.4837E+06, 4.4952E+06,
     + 4.5067E+06, 4.5182E+06, 4.5297E+06, 4.5413E+06, 4.5529E+06,
     +  4.5645E+06, 4.5761E+06, 4.5877E+06, 4.5994E+06, 4.6111E+06,
     + 4.6228E+06, 4.6346E+06, 4.6463E+06, 4.6581E+06, 4.6699E+06,
     +  4.6818E+06, 4.6936E+06, 4.7055E+06, 4.7174E+06, 4.7293E+06,
     + 4.7413E+06, 4.7533E+06, 4.7653E+06, 4.7773E+06, 4.7894E+06,
     +  4.8014E+06, 4.8135E+06, 4.8256E+06, 4.8378E+06, 4.8499E+06,
     + 4.8621E+06, 4.8744E+06, 4.8866E+06, 4.8989E+06, 4.9111E+06,
     +  4.9235E+06, 4.9358E+06, 4.9482E+06, 4.9605E+06, 4.9729E+06,
     + 4.9854E+06, 4.9978E+06, 5.0103E+06, 5.0228E+06, 5.0354E+06,
     +  5.0479E+06, 5.0605E+06, 5.0731E+06, 5.0857E+06, 5.0984E+06,
     + 5.1111E+06, 5.1238E+06, 5.1365E+06, 5.1492E+06, 5.1620E+06,
     +  5.1748E+06, 5.1876E+06, 5.2005E+06, 5.2134E+06, 5.2263E+06,
     + 5.2392E+06, 5.2521E+06, 5.2651E+06, 5.2781E+06, 5.2911E+06,
     +  5.3042E+06, 5.3173E+06, 5.3304E+06, 5.3435E+06, 5.3567E+06,
     + 5.3698E+06, 5.3830E+06, 5.3963E+06, 5.4095E+06, 5.4228E+06,
     +  5.4361E+06, 5.4494E+06, 5.4628E+06, 5.4762E+06, 5.4896E+06,
     + 5.5030E+06, 5.5165E+06, 5.5300E+06, 5.5435E+06, 5.5570E+06,
     +  5.5706E+06, 5.5842E+06, 5.5978E+06, 5.6114E+06, 5.6251E+06,
     + 5.6388E+06, 5.6525E+06, 5.6662E+06, 5.6800E+06, 5.6938E+06,
     +  5.7076E+06, 5.7215E+06, 5.7353E+06, 5.7492E+06, 5.7632E+06,
     + 5.7771E+06, 5.7911E+06, 5.8051E+06, 5.8191E+06, 5.8332E+06,
     +  5.8473E+06, 5.8614E+06, 5.8756E+06, 5.8897E+06, 5.9039E+06,
     + 5.9181E+06, 5.9324E+06, 5.9467E+06, 5.9610E+06, 5.9753E+06,
     +  5.9896E+06, 6.0040E+06, 6.0184E+06, 6.0329E+06, 6.0473E+06,
     + 6.0618E+06, 6.0763E+06, 6.0909E+06, 6.1055E+06, 6.1201E+06,
     +  6.1347E+06, 6.1493E+06, 6.1640E+06, 6.1787E+06, 6.1935E+06,
     + 6.2082E+06, 6.2230E+06, 6.2378E+06, 6.2527E+06, 6.2676E+06,
     +  6.2825E+06, 6.2974E+06, 6.3123E+06, 6.3273E+06, 6.3423E+06,
     + 6.3574E+06, 6.3725E+06, 6.3876E+06, 6.4027E+06, 6.4178E+06,
     +  6.4330E+06, 6.4482E+06, 6.4635E+06, 6.4787E+06, 6.4940E+06,
     + 6.5093E+06, 6.5247E+06, 6.5401E+06, 6.5555E+06, 6.5709E+06,
     +  6.5864E+06, 6.6019E+06, 6.6174E+06, 6.6329E+06, 6.6485E+06,
     + 6.6641E+06, 6.6798E+06, 6.6954E+06, 6.7111E+06, 6.7268E+06,
     +  6.7426E+06, 6.7584E+06, 6.7742E+06, 6.7900E+06, 6.8059E+06,
     + 6.8218E+06, 6.8377E+06, 6.8536E+06, 6.8696E+06, 6.8856E+06,
     +  6.9017E+06, 6.9177E+06, 6.9338E+06, 6.9500E+06, 6.9661E+06,
     + 6.9823E+06, 6.9985E+06, 7.0148E+06, 7.0311E+06, 7.0474E+06,
     +  7.0637E+06, 7.0801E+06, 7.0965E+06, 7.1129E+06, 7.1293E+06,
     + 7.1458E+06, 7.1623E+06, 7.1789E+06, 7.1954E+06, 7.2121E+06,
     +  7.2287E+06, 7.2453E+06, 7.2620E+06, 7.2788E+06, 7.2955E+06,
     + 7.3123E+06, 7.3291E+06, 7.3460E+06, 7.3628E+06, 7.3797E+06,
     +  7.3967E+06, 7.4136E+06, 7.4306E+06, 7.4477E+06, 7.4647E+06,
     + 7.4818E+06, 7.4989E+06, 7.5161E+06, 7.5333E+06, 7.5505E+06,
     +  7.5677E+06, 7.5850E+06, 7.6023E+06, 7.6196E+06, 7.6370E+06,
     + 7.6544E+06, 7.6718E+06, 7.6893E+06, 7.7067E+06, 7.7243E+06,
     +  7.7418E+06, 7.7594E+06, 7.7770E+06, 7.7947E+06, 7.8123E+06,
     + 7.8300E+06, 7.8478E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       646
      data (QofT( 2,J),J=1,2408)/ 1.3148E+00, 3.4623E+00, 6.2102E+00,
     +  9.4482E+00, 1.3111E+01, 1.7153E+01, 2.1543E+01, 2.6255E+01,
     + 3.1267E+01, 3.6564E+01, 4.2130E+01, 4.7953E+01, 5.4022E+01,
     +  6.0328E+01, 6.6861E+01, 7.3615E+01, 8.0583E+01, 8.7757E+01,
     + 9.5133E+01, 1.0270E+02, 1.1842E+02, 1.3486E+02, 1.5199E+02,
     +  1.6980E+02, 1.8826E+02, 2.0734E+02, 2.2702E+02, 2.4730E+02,
     + 2.6814E+02, 2.8954E+02, 3.1147E+02, 3.3394E+02, 3.5693E+02,
     +  3.8041E+02, 4.0440E+02, 4.2886E+02, 4.5381E+02, 4.7922E+02,
     + 5.0508E+02, 5.3140E+02, 5.5816E+02, 5.8536E+02, 6.1298E+02,
     +  6.4104E+02, 6.6950E+02, 6.9838E+02, 7.2766E+02, 7.5735E+02,
     + 7.8743E+02, 8.1790E+02, 8.4876E+02, 8.8001E+02, 9.1164E+02,
     +  9.4365E+02, 9.7602E+02, 1.0088E+03, 1.0419E+03, 1.0754E+03,
     + 1.1092E+03, 1.1434E+03, 1.1780E+03, 1.2129E+03, 1.2482E+03,
     +  1.2839E+03, 1.3199E+03, 1.3562E+03, 1.3930E+03, 1.4300E+03,
     + 1.4674E+03, 1.5052E+03, 1.5433E+03, 1.5818E+03, 1.6206E+03,
     +  1.6598E+03, 1.6993E+03, 1.7392E+03, 1.7795E+03, 1.8201E+03,
     + 1.8610E+03, 1.9023E+03, 1.9440E+03, 1.9860E+03, 2.0284E+03,
     +  2.0712E+03, 2.1143E+03, 2.1577E+03, 2.2016E+03, 2.2457E+03,
     + 2.2903E+03, 2.3353E+03, 2.3806E+03, 2.4263E+03, 2.4723E+03,
     +  2.5187E+03, 2.5655E+03, 2.6127E+03, 2.6603E+03, 2.7082E+03,
     + 2.7566E+03, 2.8053E+03, 2.8544E+03, 2.9039E+03, 2.9538E+03,
     +  3.0040E+03, 3.0547E+03, 3.1058E+03, 3.1573E+03, 3.2092E+03,
     + 3.2614E+03, 3.3141E+03, 3.3672E+03, 3.4207E+03, 3.4746E+03,
     +  3.5289E+03, 3.5837E+03, 3.6389E+03, 3.6945E+03, 3.7505E+03,
     + 3.8069E+03, 3.8638E+03, 3.9211E+03, 3.9789E+03, 4.0370E+03,
     +  4.0957E+03, 4.1547E+03, 4.2142E+03, 4.2742E+03, 4.3346E+03,
     + 4.3954E+03, 4.4567E+03, 4.5185E+03, 4.5807E+03, 4.6434E+03,
     +  4.7065E+03, 4.7702E+03, 4.8342E+03, 4.8988E+03, 4.9638E+03,
     + 5.0294E+03, 5.0953E+03, 5.1618E+03, 5.2288E+03, 5.2962E+03,
     +  5.3642E+03, 5.4327E+03, 5.5016E+03, 5.5710E+03, 5.6410E+03,
     + 5.7115E+03, 5.7824E+03, 5.8540E+03, 5.9259E+03, 5.9984E+03,
     +  6.0715E+03, 6.1450E+03, 6.2192E+03, 6.2938E+03, 6.3690E+03,
     + 6.4447E+03, 6.5210E+03, 6.5977E+03, 6.6751E+03, 6.7530E+03,
     +  6.8314E+03, 6.9104E+03, 6.9900E+03, 7.0701E+03, 7.1509E+03,
     + 7.2321E+03, 7.3140E+03, 7.3964E+03, 7.4794E+03, 7.5630E+03,
     +  7.6472E+03, 7.7320E+03, 7.8174E+03, 7.9034E+03, 7.9899E+03,
     + 8.0771E+03, 8.1649E+03, 8.2534E+03, 8.3424E+03, 8.4320E+03,
     +  8.5223E+03, 8.6132E+03, 8.7047E+03, 8.7969E+03, 8.8897E+03,
     + 8.9832E+03, 9.0772E+03, 9.1720E+03, 9.2674E+03, 9.3635E+03,
     +  9.4602E+03, 9.5576E+03, 9.6556E+03, 9.7544E+03, 9.8538E+03,
     + 9.9539E+03, 1.0055E+04, 1.0156E+04, 1.0258E+04, 1.0361E+04,
     +  1.0465E+04, 1.0569E+04, 1.0674E+04, 1.0780E+04, 1.0886E+04,
     + 1.0993E+04, 1.1101E+04, 1.1210E+04, 1.1319E+04, 1.1429E+04,
     +  1.1540E+04, 1.1651E+04, 1.1764E+04, 1.1877E+04, 1.1990E+04,
     + 1.2105E+04, 1.2220E+04, 1.2336E+04, 1.2453E+04, 1.2571E+04,
     +  1.2689E+04, 1.2808E+04, 1.2929E+04, 1.3049E+04, 1.3171E+04,
     + 1.3293E+04, 1.3417E+04, 1.3541E+04, 1.3666E+04, 1.3791E+04,
     +  1.3918E+04, 1.4045E+04, 1.4173E+04, 1.4302E+04, 1.4432E+04,
     + 1.4563E+04, 1.4694E+04, 1.4827E+04, 1.4960E+04, 1.5094E+04,
     +  1.5229E+04, 1.5365E+04, 1.5502E+04, 1.5640E+04, 1.5778E+04,
     + 1.5918E+04, 1.6058E+04, 1.6199E+04, 1.6341E+04, 1.6484E+04,
     +  1.6629E+04, 1.6773E+04, 1.6919E+04, 1.7066E+04, 1.7214E+04,
     + 1.7362E+04, 1.7512E+04, 1.7663E+04, 1.7814E+04, 1.7967E+04,
     +  1.8120E+04, 1.8274E+04, 1.8430E+04, 1.8586E+04, 1.8744E+04,
     + 1.8902E+04, 1.9061E+04, 1.9222E+04, 1.9383E+04, 1.9545E+04,
     +  1.9709E+04, 1.9873E+04, 2.0039E+04, 2.0205E+04, 2.0373E+04,
     + 2.0541E+04, 2.0711E+04, 2.0881E+04, 2.1053E+04, 2.1226E+04,
     +  2.1399E+04, 2.1574E+04, 2.1750E+04, 2.1927E+04, 2.2106E+04,
     + 2.2285E+04, 2.2465E+04, 2.2646E+04, 2.2829E+04, 2.3013E+04,
     +  2.3197E+04, 2.3383E+04, 2.3570E+04, 2.3758E+04, 2.3948E+04,
     + 2.4138E+04, 2.4330E+04, 2.4523E+04, 2.4717E+04, 2.4912E+04,
     +  2.5108E+04, 2.5305E+04, 2.5504E+04, 2.5704E+04, 2.5905E+04,
     + 2.6107E+04, 2.6310E+04, 2.6515E+04, 2.6720E+04, 2.6928E+04,
     +  2.7136E+04, 2.7345E+04, 2.7556E+04, 2.7768E+04, 2.7981E+04,
     + 2.8196E+04, 2.8411E+04, 2.8628E+04, 2.8847E+04, 2.9066E+04,
     +  2.9287E+04, 2.9509E+04, 2.9733E+04, 2.9957E+04, 3.0183E+04,
     + 3.0411E+04, 3.0639E+04, 3.0869E+04, 3.1101E+04, 3.1333E+04,
     +  3.1567E+04, 3.1803E+04, 3.2039E+04, 3.2277E+04, 3.2517E+04,
     + 3.2758E+04, 3.3000E+04, 3.3243E+04, 3.3488E+04, 3.3735E+04,
     +  3.3982E+04, 3.4231E+04, 3.4482E+04, 3.4734E+04, 3.4987E+04,
     + 3.5242E+04, 3.5498E+04, 3.5756E+04, 3.6015E+04, 3.6276E+04,
     +  3.6538E+04, 3.6801E+04, 3.7067E+04, 3.7333E+04, 3.7601E+04,
     + 3.7871E+04, 3.8141E+04, 3.8414E+04, 3.8688E+04, 3.8963E+04,
     +  3.9241E+04, 3.9519E+04, 3.9799E+04, 4.0081E+04, 4.0364E+04,
     + 4.0649E+04, 4.0935E+04, 4.1223E+04, 4.1512E+04, 4.1803E+04,
     +  4.2096E+04, 4.2390E+04, 4.2686E+04, 4.2984E+04, 4.3283E+04,
     + 4.3583E+04, 4.3886E+04, 4.4189E+04, 4.4495E+04, 4.4802E+04,
     +  4.5111E+04, 4.5422E+04, 4.5734E+04, 4.6048E+04, 4.6363E+04,
     + 4.6681E+04, 4.6999E+04, 4.7320E+04, 4.7643E+04, 4.7967E+04,
     +  4.8292E+04, 4.8620E+04, 4.8949E+04, 4.9280E+04, 4.9613E+04,
     + 4.9947E+04, 5.0284E+04, 5.0622E+04, 5.0962E+04, 5.1303E+04,
     +  5.1646E+04, 5.1992E+04, 5.2339E+04, 5.2687E+04, 5.3038E+04,
     + 5.3390E+04, 5.3745E+04, 5.4101E+04, 5.4459E+04, 5.4819E+04,
     +  5.5180E+04, 5.5544E+04, 5.5909E+04, 5.6276E+04, 5.6646E+04,
     + 5.7017E+04, 5.7390E+04, 5.7764E+04, 5.8141E+04, 5.8520E+04,
     +  5.8901E+04, 5.9283E+04, 5.9668E+04, 6.0054E+04, 6.0443E+04,
     + 6.0833E+04, 6.1226E+04, 6.1620E+04, 6.2017E+04, 6.2415E+04,
     +  6.2815E+04, 6.3218E+04, 6.3622E+04, 6.4029E+04, 6.4437E+04,
     + 6.4848E+04, 6.5260E+04, 6.5675E+04, 6.6092E+04, 6.6511E+04,
     +  6.6931E+04, 6.7354E+04, 6.7779E+04, 6.8207E+04, 6.8636E+04,
     + 6.9067E+04, 6.9501E+04, 6.9936E+04, 7.0374E+04, 7.0814E+04,
     +  7.1256E+04, 7.1700E+04, 7.2147E+04, 7.2596E+04, 7.3047E+04,
     + 7.3500E+04, 7.3955E+04, 7.4412E+04, 7.4872E+04, 7.5334E+04,
     +  7.5798E+04, 7.6264E+04, 7.6733E+04, 7.7204E+04, 7.7677E+04,
     + 7.8153E+04, 7.8630E+04, 7.9110E+04, 7.9593E+04, 8.0077E+04,
     +  8.0564E+04, 8.1054E+04, 8.1545E+04, 8.2039E+04, 8.2535E+04,
     + 8.3034E+04, 8.3535E+04, 8.4039E+04, 8.4544E+04, 8.5053E+04,
     +  8.5563E+04, 8.6076E+04, 8.6592E+04, 8.7110E+04, 8.7630E+04,
     + 8.8153E+04, 8.8678E+04, 8.9206E+04, 8.9736E+04, 9.0268E+04,
     +  9.0803E+04, 9.1341E+04, 9.1881E+04, 9.2423E+04, 9.2969E+04,
     + 9.3516E+04, 9.4067E+04, 9.4619E+04, 9.5174E+04, 9.5732E+04,
     +  9.6293E+04, 9.6855E+04, 9.7421E+04, 9.7989E+04, 9.8560E+04,
     + 9.9133E+04, 9.9710E+04, 1.0029E+05, 1.0087E+05, 1.0145E+05,
     +  1.0204E+05, 1.0263E+05, 1.0322E+05, 1.0382E+05, 1.0441E+05,
     + 1.0501E+05, 1.0562E+05, 1.0622E+05, 1.0683E+05, 1.0744E+05,
     +  1.0805E+05, 1.0867E+05, 1.0929E+05, 1.0991E+05, 1.1054E+05,
     + 1.1116E+05, 1.1179E+05, 1.1243E+05, 1.1306E+05, 1.1370E+05,
     +  1.1434E+05, 1.1499E+05, 1.1564E+05, 1.1629E+05, 1.1694E+05,
     + 1.1759E+05, 1.1825E+05, 1.1891E+05, 1.1958E+05, 1.2025E+05,
     +  1.2092E+05, 1.2159E+05, 1.2227E+05, 1.2295E+05, 1.2363E+05,
     + 1.2431E+05, 1.2500E+05, 1.2569E+05, 1.2639E+05, 1.2709E+05,
     +  1.2779E+05, 1.2849E+05, 1.2920E+05, 1.2991E+05, 1.3062E+05,
     + 1.3133E+05, 1.3205E+05, 1.3277E+05, 1.3350E+05, 1.3423E+05,
     +  1.3496E+05, 1.3569E+05, 1.3643E+05, 1.3717E+05, 1.3791E+05,
     + 1.3866E+05, 1.3941E+05, 1.4016E+05, 1.4092E+05, 1.4168E+05,
     +  1.4244E+05, 1.4321E+05, 1.4398E+05, 1.4475E+05, 1.4553E+05,
     + 1.4631E+05, 1.4709E+05, 1.4788E+05, 1.4866E+05, 1.4946E+05,
     +  1.5025E+05, 1.5105E+05, 1.5185E+05, 1.5266E+05, 1.5347E+05,
     + 1.5428E+05, 1.5510E+05, 1.5592E+05, 1.5674E+05, 1.5756E+05,
     +  1.5839E+05, 1.5923E+05, 1.6006E+05, 1.6090E+05, 1.6175E+05,
     + 1.6259E+05, 1.6344E+05, 1.6430E+05, 1.6515E+05, 1.6601E+05,
     +  1.6688E+05, 1.6775E+05, 1.6862E+05, 1.6949E+05, 1.7037E+05,
     + 1.7125E+05, 1.7214E+05, 1.7303E+05, 1.7392E+05, 1.7482E+05,
     +  1.7572E+05, 1.7662E+05, 1.7753E+05, 1.7844E+05, 1.7936E+05,
     + 1.8027E+05, 1.8120E+05, 1.8212E+05, 1.8305E+05, 1.8399E+05,
     +  1.8492E+05, 1.8586E+05, 1.8681E+05, 1.8776E+05, 1.8871E+05,
     + 1.8966E+05, 1.9062E+05, 1.9159E+05, 1.9255E+05, 1.9353E+05,
     +  1.9450E+05, 1.9548E+05, 1.9646E+05, 1.9745E+05, 1.9844E+05,
     + 1.9943E+05, 2.0043E+05, 2.0144E+05, 2.0244E+05, 2.0345E+05,
     +  2.0447E+05, 2.0548E+05, 2.0651E+05, 2.0753E+05, 2.0856E+05,
     + 2.0960E+05, 2.1064E+05, 2.1168E+05, 2.1272E+05, 2.1377E+05,
     +  2.1483E+05, 2.1589E+05, 2.1695E+05, 2.1802E+05, 2.1909E+05,
     + 2.2016E+05, 2.2124E+05, 2.2233E+05, 2.2341E+05, 2.2451E+05,
     +  2.2560E+05, 2.2670E+05, 2.2781E+05, 2.2891E+05, 2.3003E+05,
     + 2.3114E+05, 2.3227E+05, 2.3339E+05, 2.3452E+05, 2.3565E+05,
     +  2.3679E+05, 2.3794E+05, 2.3908E+05, 2.4024E+05, 2.4139E+05,
     + 2.4255E+05, 2.4372E+05, 2.4488E+05, 2.4606E+05, 2.4724E+05,
     +  2.4842E+05, 2.4961E+05, 2.5080E+05, 2.5199E+05, 2.5319E+05,
     + 2.5440E+05, 2.5561E+05, 2.5682E+05, 2.5804E+05, 2.5926E+05,
     +  2.6049E+05, 2.6172E+05, 2.6296E+05, 2.6420E+05, 2.6544E+05,
     + 2.6669E+05, 2.6795E+05, 2.6921E+05, 2.7047E+05, 2.7174E+05,
     +  2.7301E+05, 2.7429E+05, 2.7557E+05, 2.7686E+05, 2.7815E+05,
     + 2.7945E+05, 2.8075E+05, 2.8206E+05, 2.8337E+05, 2.8469E+05,
     +  2.8601E+05, 2.8733E+05, 2.8867E+05, 2.9000E+05, 2.9134E+05,
     + 2.9269E+05, 2.9404E+05, 2.9539E+05, 2.9675E+05, 2.9812E+05,
     +  2.9949E+05, 3.0086E+05, 3.0224E+05, 3.0362E+05, 3.0501E+05,
     + 3.0641E+05, 3.0781E+05, 3.0921E+05, 3.1062E+05, 3.1204E+05,
     +  3.1346E+05, 3.1488E+05, 3.1631E+05, 3.1775E+05, 3.1919E+05,
     + 3.2063E+05, 3.2208E+05, 3.2354E+05, 3.2500E+05, 3.2646E+05,
     +  3.2794E+05, 3.2941E+05, 3.3089E+05, 3.3238E+05, 3.3387E+05,
     + 3.3537E+05, 3.3687E+05, 3.3838E+05, 3.3989E+05, 3.4141E+05,
     +  3.4293E+05, 3.4446E+05, 3.4600E+05, 3.4754E+05, 3.4908E+05,
     + 3.5063E+05, 3.5219E+05, 3.5375E+05, 3.5532E+05, 3.5689E+05,
     +  3.5847E+05, 3.6005E+05, 3.6164E+05, 3.6323E+05, 3.6483E+05,
     + 3.6644E+05, 3.6805E+05, 3.6967E+05, 3.7129E+05, 3.7292E+05,
     +  3.7455E+05, 3.7619E+05, 3.7783E+05, 3.7948E+05, 3.8114E+05,
     + 3.8280E+05, 3.8447E+05, 3.8614E+05, 3.8782E+05, 3.8950E+05,
     +  3.9119E+05, 3.9289E+05, 3.9459E+05, 3.9630E+05, 3.9801E+05,
     + 3.9973E+05, 4.0145E+05, 4.0318E+05, 4.0492E+05, 4.0666E+05,
     +  4.0841E+05, 4.1016E+05, 4.1192E+05, 4.1369E+05, 4.1546E+05,
     + 4.1724E+05, 4.1902E+05, 4.2081E+05, 4.2261E+05, 4.2441E+05,
     +  4.2622E+05, 4.2803E+05, 4.2985E+05, 4.3168E+05, 4.3351E+05,
     + 4.3535E+05, 4.3719E+05, 4.3904E+05, 4.4090E+05, 4.4276E+05,
     +  4.4463E+05, 4.4651E+05, 4.4839E+05, 4.5028E+05, 4.5217E+05,
     + 4.5407E+05, 4.5598E+05, 4.5789E+05, 4.5981E+05, 4.6173E+05,
     +  4.6366E+05, 4.6560E+05, 4.6755E+05, 4.6950E+05, 4.7145E+05,
     + 4.7342E+05, 4.7539E+05, 4.7736E+05, 4.7935E+05, 4.8134E+05,
     +  4.8333E+05, 4.8533E+05, 4.8734E+05, 4.8936E+05, 4.9138E+05,
     + 4.9341E+05, 4.9544E+05, 4.9748E+05, 4.9953E+05, 5.0159E+05,
     +  5.0365E+05, 5.0572E+05, 5.0779E+05, 5.0987E+05, 5.1196E+05,
     + 5.1405E+05, 5.1616E+05, 5.1826E+05, 5.2038E+05, 5.2250E+05,
     +  5.2463E+05, 5.2676E+05, 5.2891E+05, 5.3106E+05, 5.3321E+05,
     + 5.3537E+05, 5.3754E+05, 5.3972E+05, 5.4190E+05, 5.4409E+05,
     +  5.4629E+05, 5.4849E+05, 5.5071E+05, 5.5292E+05, 5.5515E+05,
     + 5.5738E+05, 5.5962E+05, 5.6187E+05, 5.6412E+05, 5.6638E+05,
     +  5.6865E+05, 5.7092E+05, 5.7320E+05, 5.7549E+05, 5.7779E+05,
     + 5.8009E+05, 5.8240E+05, 5.8472E+05, 5.8705E+05, 5.8938E+05,
     +  5.9172E+05, 5.9406E+05, 5.9642E+05, 5.9878E+05, 6.0115E+05,
     + 6.0352E+05, 6.0591E+05, 6.0830E+05, 6.1070E+05, 6.1310E+05,
     +  6.1552E+05, 6.1794E+05, 6.2036E+05, 6.2280E+05, 6.2524E+05,
     + 6.2769E+05, 6.3015E+05, 6.3262E+05, 6.3509E+05, 6.3757E+05,
     +  6.4006E+05, 6.4256E+05, 6.4506E+05, 6.4757E+05, 6.5009E+05,
     + 6.5262E+05, 6.5515E+05, 6.5769E+05, 6.6024E+05, 6.6280E+05,
     +  6.6537E+05, 6.6794E+05, 6.7052E+05, 6.7311E+05, 6.7570E+05,
     + 6.7831E+05, 6.8092E+05, 6.8354E+05, 6.8617E+05, 6.8881E+05,
     +  6.9145E+05, 6.9410E+05, 6.9676E+05, 6.9943E+05, 7.0211E+05,
     + 7.0479E+05, 7.0748E+05, 7.1018E+05, 7.1289E+05, 7.1561E+05,
     +  7.1833E+05, 7.2106E+05, 7.2380E+05, 7.2655E+05, 7.2931E+05,
     + 7.3207E+05, 7.3485E+05, 7.3763E+05, 7.4042E+05, 7.4322E+05,
     +  7.4602E+05, 7.4884E+05, 7.5166E+05, 7.5449E+05, 7.5733E+05,
     + 7.6018E+05, 7.6304E+05, 7.6590E+05, 7.6878E+05, 7.7166E+05,
     +  7.7455E+05, 7.7745E+05, 7.8035E+05, 7.8327E+05, 7.8620E+05,
     + 7.8913E+05, 7.9207E+05, 7.9502E+05, 7.9798E+05, 8.0095E+05,
     +  8.0392E+05, 8.0691E+05, 8.0990E+05, 8.1290E+05, 8.1591E+05,
     + 8.1893E+05, 8.2196E+05, 8.2500E+05, 8.2804E+05, 8.3110E+05,
     +  8.3416E+05, 8.3723E+05, 8.4032E+05, 8.4341E+05, 8.4651E+05,
     + 8.4961E+05, 8.5273E+05, 8.5586E+05, 8.5899E+05, 8.6214E+05,
     +  8.6529E+05, 8.6845E+05, 8.7162E+05, 8.7480E+05, 8.7799E+05,
     + 8.8119E+05, 8.8440E+05, 8.8761E+05, 8.9084E+05, 8.9407E+05,
     +  8.9732E+05, 9.0057E+05, 9.0384E+05, 9.0711E+05, 9.1039E+05,
     + 9.1368E+05, 9.1698E+05, 9.2029E+05, 9.2361E+05, 9.2693E+05,
     +  9.3027E+05, 9.3362E+05, 9.3698E+05, 9.4034E+05, 9.4372E+05,
     + 9.4710E+05, 9.5049E+05, 9.5390E+05, 9.5731E+05, 9.6073E+05,
     +  9.6417E+05, 9.6761E+05, 9.7106E+05, 9.7452E+05, 9.7799E+05,
     + 9.8147E+05, 9.8496E+05, 9.8846E+05, 9.9197E+05, 9.9549E+05,
     +  9.9902E+05, 1.0026E+06, 1.0061E+06, 1.0097E+06, 1.0132E+06,
     + 1.0168E+06, 1.0204E+06, 1.0240E+06, 1.0276E+06, 1.0312E+06,
     +  1.0349E+06, 1.0385E+06, 1.0421E+06, 1.0458E+06, 1.0495E+06,
     + 1.0532E+06, 1.0568E+06, 1.0605E+06, 1.0642E+06, 1.0680E+06,
     +  1.0717E+06, 1.0754E+06, 1.0792E+06, 1.0829E+06, 1.0867E+06,
     + 1.0905E+06, 1.0943E+06, 1.0981E+06, 1.1019E+06, 1.1057E+06,
     +  1.1096E+06, 1.1134E+06, 1.1173E+06, 1.1211E+06, 1.1250E+06,
     + 1.1289E+06, 1.1328E+06, 1.1367E+06, 1.1406E+06, 1.1445E+06,
     +  1.1485E+06, 1.1524E+06, 1.1564E+06, 1.1603E+06, 1.1643E+06,
     + 1.1683E+06, 1.1723E+06, 1.1763E+06, 1.1803E+06, 1.1844E+06,
     +  1.1884E+06, 1.1925E+06, 1.1965E+06, 1.2006E+06, 1.2047E+06,
     + 1.2088E+06, 1.2129E+06, 1.2170E+06, 1.2212E+06, 1.2253E+06,
     +  1.2295E+06, 1.2336E+06, 1.2378E+06, 1.2420E+06, 1.2462E+06,
     + 1.2504E+06, 1.2546E+06, 1.2588E+06, 1.2631E+06, 1.2673E+06,
     +  1.2716E+06, 1.2759E+06, 1.2802E+06, 1.2845E+06, 1.2888E+06,
     + 1.2931E+06, 1.2974E+06, 1.3018E+06, 1.3061E+06, 1.3105E+06,
     +  1.3149E+06, 1.3193E+06, 1.3237E+06, 1.3281E+06, 1.3325E+06,
     + 1.3369E+06, 1.3414E+06, 1.3458E+06, 1.3503E+06, 1.3548E+06,
     +  1.3593E+06, 1.3638E+06, 1.3683E+06, 1.3728E+06, 1.3774E+06,
     + 1.3819E+06, 1.3865E+06, 1.3911E+06, 1.3957E+06, 1.4003E+06,
     +  1.4049E+06, 1.4095E+06, 1.4141E+06, 1.4188E+06, 1.4234E+06,
     + 1.4281E+06, 1.4328E+06, 1.4375E+06, 1.4422E+06, 1.4469E+06,
     +  1.4517E+06, 1.4564E+06, 1.4612E+06, 1.4659E+06, 1.4707E+06,
     + 1.4755E+06, 1.4803E+06, 1.4851E+06, 1.4899E+06, 1.4948E+06,
     +  1.4996E+06, 1.5045E+06, 1.5094E+06, 1.5143E+06, 1.5192E+06,
     + 1.5241E+06, 1.5290E+06, 1.5340E+06, 1.5389E+06, 1.5439E+06,
     +  1.5489E+06, 1.5539E+06, 1.5589E+06, 1.5639E+06, 1.5689E+06,
     + 1.5739E+06, 1.5790E+06, 1.5841E+06, 1.5891E+06, 1.5942E+06,
     +  1.5993E+06, 1.6045E+06, 1.6096E+06, 1.6147E+06, 1.6199E+06,
     + 1.6251E+06, 1.6302E+06, 1.6354E+06, 1.6406E+06, 1.6459E+06,
     +  1.6511E+06, 1.6563E+06, 1.6616E+06, 1.6669E+06, 1.6722E+06,
     + 1.6775E+06, 1.6828E+06, 1.6881E+06, 1.6934E+06, 1.6988E+06,
     +  1.7042E+06, 1.7095E+06, 1.7149E+06, 1.7203E+06, 1.7258E+06,
     + 1.7312E+06, 1.7366E+06, 1.7421E+06, 1.7476E+06, 1.7531E+06,
     +  1.7586E+06, 1.7641E+06, 1.7696E+06, 1.7751E+06, 1.7807E+06,
     + 1.7863E+06, 1.7918E+06, 1.7974E+06, 1.8030E+06, 1.8087E+06,
     +  1.8143E+06, 1.8199E+06, 1.8256E+06, 1.8313E+06, 1.8370E+06,
     + 1.8427E+06, 1.8484E+06, 1.8541E+06, 1.8599E+06, 1.8656E+06,
     +  1.8714E+06, 1.8772E+06, 1.8830E+06, 1.8888E+06, 1.8947E+06,
     + 1.9005E+06, 1.9064E+06, 1.9122E+06, 1.9181E+06, 1.9240E+06,
     +  1.9299E+06, 1.9359E+06, 1.9418E+06, 1.9478E+06, 1.9537E+06,
     + 1.9597E+06, 1.9657E+06, 1.9717E+06, 1.9778E+06, 1.9838E+06,
     +  1.9899E+06, 1.9959E+06, 2.0020E+06, 2.0081E+06, 2.0143E+06,
     + 2.0204E+06, 2.0265E+06, 2.0327E+06, 2.0389E+06, 2.0451E+06,
     +  2.0513E+06, 2.0575E+06, 2.0637E+06, 2.0700E+06, 2.0762E+06,
     + 2.0825E+06, 2.0888E+06, 2.0951E+06, 2.1014E+06, 2.1078E+06,
     +  2.1141E+06, 2.1205E+06, 2.1269E+06, 2.1333E+06, 2.1397E+06,
     + 2.1461E+06, 2.1526E+06, 2.1590E+06, 2.1655E+06, 2.1720E+06,
     +  2.1785E+06, 2.1850E+06, 2.1916E+06, 2.1981E+06, 2.2047E+06,
     + 2.2112E+06, 2.2178E+06, 2.2245E+06, 2.2311E+06, 2.2377E+06,
     +  2.2444E+06, 2.2511E+06, 2.2578E+06, 2.2645E+06, 2.2712E+06,
     + 2.2779E+06, 2.2847E+06, 2.2914E+06, 2.2982E+06, 2.3050E+06,
     +  2.3118E+06, 2.3187E+06, 2.3255E+06, 2.3324E+06, 2.3392E+06,
     + 2.3461E+06, 2.3531E+06, 2.3600E+06, 2.3669E+06, 2.3739E+06,
     +  2.3809E+06, 2.3878E+06, 2.3948E+06, 2.4019E+06, 2.4089E+06,
     + 2.4160E+06, 2.4230E+06, 2.4301E+06, 2.4372E+06, 2.4443E+06,
     +  2.4515E+06, 2.4586E+06, 2.4658E+06, 2.4730E+06, 2.4802E+06,
     + 2.4874E+06, 2.4946E+06, 2.5019E+06, 2.5091E+06, 2.5164E+06,
     +  2.5237E+06, 2.5310E+06, 2.5384E+06, 2.5457E+06, 2.5531E+06,
     + 2.5605E+06, 2.5679E+06, 2.5753E+06, 2.5827E+06, 2.5902E+06,
     +  2.5976E+06, 2.6051E+06, 2.6126E+06, 2.6201E+06, 2.6277E+06,
     + 2.6352E+06, 2.6428E+06, 2.6504E+06, 2.6580E+06, 2.6656E+06,
     +  2.6732E+06, 2.6809E+06, 2.6886E+06, 2.6963E+06, 2.7040E+06,
     + 2.7117E+06, 2.7194E+06, 2.7272E+06, 2.7350E+06, 2.7427E+06,
     +  2.7506E+06, 2.7584E+06, 2.7662E+06, 2.7741E+06, 2.7820E+06,
     + 2.7899E+06, 2.7978E+06, 2.8057E+06, 2.8137E+06, 2.8216E+06,
     +  2.8296E+06, 2.8376E+06, 2.8456E+06, 2.8537E+06, 2.8617E+06,
     + 2.8698E+06, 2.8779E+06, 2.8860E+06, 2.8941E+06, 2.9023E+06,
     +  2.9104E+06, 2.9186E+06, 2.9268E+06, 2.9350E+06, 2.9433E+06,
     + 2.9515E+06, 2.9598E+06, 2.9681E+06, 2.9764E+06, 2.9847E+06,
     +  2.9930E+06, 3.0014E+06, 3.0098E+06, 3.0182E+06, 3.0266E+06,
     + 3.0350E+06, 3.0435E+06, 3.0520E+06, 3.0604E+06, 3.0690E+06,
     +  3.0775E+06, 3.0860E+06, 3.0946E+06, 3.1032E+06, 3.1118E+06,
     + 3.1204E+06, 3.1290E+06, 3.1377E+06, 3.1464E+06, 3.1551E+06,
     +  3.1638E+06, 3.1725E+06, 3.1813E+06, 3.1900E+06, 3.1988E+06,
     + 3.2076E+06, 3.2165E+06, 3.2253E+06, 3.2342E+06, 3.2431E+06,
     +  3.2520E+06, 3.2609E+06, 3.2698E+06, 3.2788E+06, 3.2878E+06,
     + 3.2968E+06, 3.3058E+06, 3.3148E+06, 3.3239E+06, 3.3330E+06,
     +  3.3421E+06, 3.3512E+06, 3.3603E+06, 3.3695E+06, 3.3786E+06,
     + 3.3878E+06, 3.3971E+06, 3.4063E+06, 3.4155E+06, 3.4248E+06,
     +  3.4341E+06, 3.4434E+06, 3.4528E+06, 3.4621E+06, 3.4715E+06,
     + 3.4809E+06, 3.4903E+06, 3.4997E+06, 3.5092E+06, 3.5186E+06,
     +  3.5281E+06, 3.5376E+06, 3.5472E+06, 3.5567E+06, 3.5663E+06,
     + 3.5759E+06, 3.5855E+06, 3.5951E+06, 3.6048E+06, 3.6144E+06,
     +  3.6241E+06, 3.6339E+06, 3.6436E+06, 3.6533E+06, 3.6631E+06,
     + 3.6729E+06, 3.6827E+06, 3.6926E+06, 3.7024E+06, 3.7123E+06,
     +  3.7222E+06, 3.7321E+06, 3.7421E+06, 3.7520E+06, 3.7620E+06,
     + 3.7720E+06, 3.7820E+06, 3.7921E+06, 3.8021E+06, 3.8122E+06,
     +  3.8223E+06, 3.8324E+06, 3.8426E+06, 3.8528E+06, 3.8629E+06,
     + 3.8732E+06, 3.8834E+06, 3.8936E+06, 3.9039E+06, 3.9142E+06,
     +  3.9245E+06, 3.9349E+06, 3.9452E+06, 3.9556E+06, 3.9660E+06,
     + 3.9764E+06, 3.9869E+06, 3.9973E+06, 4.0078E+06, 4.0183E+06,
     +  4.0289E+06, 4.0394E+06, 4.0500E+06, 4.0606E+06, 4.0712E+06,
     + 4.0819E+06, 4.0925E+06, 4.1032E+06, 4.1139E+06, 4.1246E+06,
     +  4.1354E+06, 4.1462E+06, 4.1569E+06, 4.1678E+06, 4.1786E+06,
     + 4.1895E+06, 4.2003E+06, 4.2112E+06, 4.2222E+06, 4.2331E+06,
     +  4.2441E+06, 4.2551E+06, 4.2661E+06, 4.2771E+06, 4.2882E+06,
     + 4.2993E+06, 4.3104E+06, 4.3215E+06, 4.3326E+06, 4.3438E+06,
     +  4.3550E+06, 4.3662E+06, 4.3775E+06, 4.3887E+06, 4.4000E+06,
     + 4.4113E+06, 4.4226E+06, 4.4340E+06, 4.4454E+06, 4.4568E+06,
     +  4.4682E+06, 4.4796E+06, 4.4911E+06, 4.5026E+06, 4.5141E+06,
     + 4.5256E+06, 4.5372E+06, 4.5488E+06, 4.5604E+06, 4.5720E+06,
     +  4.5837E+06, 4.5953E+06, 4.6070E+06, 4.6187E+06, 4.6305E+06,
     + 4.6423E+06, 4.6541E+06, 4.6659E+06, 4.6777E+06, 4.6896E+06,
     +  4.7015E+06, 4.7134E+06, 4.7253E+06, 4.7372E+06, 4.7492E+06,
     + 4.7612E+06, 4.7733E+06, 4.7853E+06, 4.7974E+06, 4.8095E+06,
     +  4.8216E+06, 4.8338E+06, 4.8459E+06, 4.8581E+06, 4.8703E+06,
     + 4.8826E+06, 4.8948E+06, 4.9071E+06, 4.9194E+06, 4.9318E+06,
     +  4.9442E+06, 4.9565E+06, 4.9690E+06, 4.9814E+06, 4.9938E+06,
     + 5.0063E+06, 5.0188E+06, 5.0314E+06, 5.0439E+06, 5.0565E+06,
     +  5.0691E+06, 5.0818E+06, 5.0944E+06, 5.1071E+06, 5.1198E+06,
     + 5.1325E+06, 5.1453E+06, 5.1581E+06, 5.1709E+06, 5.1837E+06,
     +  5.1966E+06, 5.2094E+06, 5.2223E+06, 5.2353E+06, 5.2482E+06,
     + 5.2612E+06, 5.2742E+06, 5.2872E+06, 5.3003E+06, 5.3134E+06,
     +  5.3265E+06, 5.3396E+06, 5.3528E+06, 5.3660E+06, 5.3792E+06,
     + 5.3924E+06, 5.4057E+06, 5.4189E+06, 5.4322E+06, 5.4456E+06,
     +  5.4589E+06, 5.4723E+06, 5.4857E+06, 5.4992E+06, 5.5126E+06,
     + 5.5261E+06, 5.5396E+06, 5.5532E+06, 5.5668E+06, 5.5803E+06,
     +  5.5940E+06, 5.6076E+06, 5.6213E+06, 5.6350E+06, 5.6487E+06,
     + 5.6624E+06, 5.6762E+06, 5.6900E+06, 5.7039E+06, 5.7177E+06,
     +  5.7316E+06, 5.7455E+06, 5.7594E+06, 5.7734E+06, 5.7874E+06,
     + 5.8014E+06, 5.8154E+06, 5.8295E+06, 5.8436E+06, 5.8577E+06,
     +  5.8719E+06, 5.8860E+06, 5.9002E+06, 5.9145E+06, 5.9287E+06,
     + 5.9430E+06, 5.9573E+06, 5.9716E+06, 5.9860E+06, 6.0004E+06,
     +  6.0148E+06, 6.0292E+06, 6.0437E+06, 6.0582E+06, 6.0727E+06,
     + 6.0873E+06, 6.1019E+06, 6.1165E+06, 6.1311E+06, 6.1458E+06,
     +  6.1604E+06, 6.1752E+06, 6.1899E+06, 6.2047E+06, 6.2195E+06,
     + 6.2343E+06, 6.2491E+06, 6.2640E+06, 6.2789E+06, 6.2939E+06,
     +  6.3088E+06, 6.3238E+06, 6.3388E+06, 6.3539E+06, 6.3690E+06,
     + 6.3841E+06, 6.3992E+06, 6.4144E+06, 6.4296E+06, 6.4448E+06,
     +  6.4600E+06, 6.4753E+06, 6.4906E+06, 6.5059E+06, 6.5213E+06,
     + 6.5367E+06, 6.5521E+06, 6.5675E+06, 6.5830E+06, 6.5985E+06,
     +  6.6140E+06, 6.6296E+06, 6.6452E+06, 6.6608E+06, 6.6764E+06,
     + 6.6921E+06, 6.7078E+06, 6.7235E+06, 6.7393E+06, 6.7551E+06,
     +  6.7709E+06, 6.7867E+06, 6.8026E+06, 6.8185E+06, 6.8344E+06,
     + 6.8504E+06, 6.8664E+06, 6.8824E+06, 6.8985E+06, 6.9145E+06,
     +  6.9306E+06, 6.9468E+06, 6.9629E+06, 6.9791E+06, 6.9954E+06,
     + 7.0116E+06, 7.0279E+06, 7.0442E+06, 7.0606E+06, 7.0769E+06,
     +  7.0933E+06, 7.1098E+06, 7.1262E+06, 7.1427E+06, 7.1592E+06,
     + 7.1758E+06, 7.1924E+06, 7.2090E+06, 7.2256E+06, 7.2423E+06,
     +  7.2590E+06, 7.2757E+06, 7.2925E+06, 7.3093E+06, 7.3261E+06,
     + 7.3430E+06, 7.3599E+06, 7.3768E+06, 7.3937E+06, 7.4107E+06,
     +  7.4277E+06, 7.4447E+06, 7.4618E+06, 7.4789E+06, 7.4960E+06,
     + 7.5132E+06, 7.5304E+06, 7.5476E+06, 7.5648E+06, 7.5821E+06,
     +  7.5994E+06, 7.6168E+06, 7.6342E+06, 7.6516E+06, 7.6690E+06,
     + 7.6865E+06, 7.7040E+06, 7.7215E+06, 7.7391E+06, 7.7567E+06,
     +  7.7743E+06, 7.7919E+06, 7.8096E+06, 7.8273E+06, 7.8451E+06,
     + 7.8629E+06, 7.8807E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_NO2   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 1),QofT( 1,NT),Q(NT)
      data Tmx(1)/3500./
      data xgj/ 3./
c...      NO2
c...        --       646
      data (QofT( 1,J),J=1,2408)/ 3.0678E+00, 8.0505E+00, 1.4331E+01,
     +  2.1706E+01, 3.0037E+01, 3.9224E+01, 4.9194E+01, 5.9891E+01,
     + 7.1268E+01, 8.3287E+01, 9.5915E+01, 1.0912E+02, 1.2289E+02,
     +  1.3719E+02, 1.5201E+02, 1.6732E+02, 1.8312E+02, 1.9938E+02,
     + 2.1610E+02, 2.3327E+02, 2.6888E+02, 3.0614E+02, 3.4499E+02,
     +  3.8535E+02, 4.2717E+02, 4.7041E+02, 5.1501E+02, 5.6095E+02,
     + 6.0817E+02, 6.5665E+02, 7.0636E+02, 7.5727E+02, 8.0934E+02,
     +  8.6255E+02, 9.1689E+02, 9.7232E+02, 1.0288E+03, 1.0864E+03,
     + 1.1450E+03, 1.2046E+03, 1.2652E+03, 1.3268E+03, 1.3894E+03,
     +  1.4529E+03, 1.5174E+03, 1.5828E+03, 1.6491E+03, 1.7164E+03,
     + 1.7845E+03, 1.8535E+03, 1.9233E+03, 1.9940E+03, 2.0656E+03,
     +  2.1380E+03, 2.2113E+03, 2.2853E+03, 2.3602E+03, 2.4359E+03,
     + 2.5123E+03, 2.5896E+03, 2.6676E+03, 2.7464E+03, 2.8260E+03,
     +  2.9064E+03, 2.9875E+03, 3.0693E+03, 3.1519E+03, 3.2352E+03,
     + 3.3192E+03, 3.4040E+03, 3.4895E+03, 3.5757E+03, 3.6627E+03,
     +  3.7503E+03, 3.8386E+03, 3.9277E+03, 4.0174E+03, 4.1079E+03,
     + 4.1990E+03, 4.2908E+03, 4.3833E+03, 4.4765E+03, 4.5704E+03,
     +  4.6649E+03, 4.7601E+03, 4.8560E+03, 4.9526E+03, 5.0498E+03,
     + 5.1477E+03, 5.2463E+03, 5.3456E+03, 5.4455E+03, 5.5461E+03,
     +  5.6474E+03, 5.7493E+03, 5.8519E+03, 5.9551E+03, 6.0591E+03,
     + 6.1637E+03, 6.2689E+03, 6.3748E+03, 6.4815E+03, 6.5888E+03,
     +  6.6966E+03, 6.8053E+03, 6.9146E+03, 7.0245E+03, 7.1351E+03,
     + 7.2465E+03, 7.3585E+03, 7.4711E+03, 7.5845E+03, 7.6985E+03,
     +  7.8131E+03, 7.9286E+03, 8.0447E+03, 8.1615E+03, 8.2789E+03,
     + 8.3970E+03, 8.5159E+03, 8.6355E+03, 8.7557E+03, 8.8767E+03,
     +  8.9983E+03, 9.1207E+03, 9.2438E+03, 9.3676E+03, 9.4921E+03,
     + 9.6174E+03, 9.7433E+03, 9.8700E+03, 9.9974E+03, 1.0126E+04,
     +  1.0254E+04, 1.0384E+04, 1.0514E+04, 1.0646E+04, 1.0777E+04,
     + 1.0910E+04, 1.1043E+04, 1.1177E+04, 1.1312E+04, 1.1448E+04,
     +  1.1584E+04, 1.1722E+04, 1.1860E+04, 1.1998E+04, 1.2138E+04,
     + 1.2278E+04, 1.2419E+04, 1.2561E+04, 1.2704E+04, 1.2848E+04,
     +  1.2992E+04, 1.3137E+04, 1.3283E+04, 1.3430E+04, 1.3577E+04,
     + 1.3726E+04, 1.3875E+04, 1.4025E+04, 1.4176E+04, 1.4328E+04,
     +  1.4481E+04, 1.4634E+04, 1.4789E+04, 1.4944E+04, 1.5100E+04,
     + 1.5257E+04, 1.5415E+04, 1.5574E+04, 1.5733E+04, 1.5894E+04,
     +  1.6055E+04, 1.6218E+04, 1.6381E+04, 1.6545E+04, 1.6710E+04,
     + 1.6876E+04, 1.7043E+04, 1.7211E+04, 1.7380E+04, 1.7550E+04,
     +  1.7720E+04, 1.7892E+04, 1.8065E+04, 1.8238E+04, 1.8413E+04,
     + 1.8588E+04, 1.8765E+04, 1.8942E+04, 1.9121E+04, 1.9300E+04,
     +  1.9481E+04, 1.9662E+04, 1.9845E+04, 2.0028E+04, 2.0213E+04,
     + 2.0398E+04, 2.0585E+04, 2.0772E+04, 2.0961E+04, 2.1151E+04,
     +  2.1342E+04, 2.1533E+04, 2.1726E+04, 2.1920E+04, 2.2115E+04,
     + 2.2312E+04, 2.2509E+04, 2.2707E+04, 2.2907E+04, 2.3107E+04,
     +  2.3309E+04, 2.3511E+04, 2.3715E+04, 2.3920E+04, 2.4126E+04,
     + 2.4334E+04, 2.4542E+04, 2.4752E+04, 2.4963E+04, 2.5175E+04,
     +  2.5388E+04, 2.5602E+04, 2.5817E+04, 2.6034E+04, 2.6252E+04,
     + 2.6471E+04, 2.6691E+04, 2.6913E+04, 2.7135E+04, 2.7359E+04,
     +  2.7584E+04, 2.7811E+04, 2.8038E+04, 2.8267E+04, 2.8497E+04,
     + 2.8729E+04, 2.8961E+04, 2.9195E+04, 2.9430E+04, 2.9667E+04,
     +  2.9905E+04, 3.0144E+04, 3.0384E+04, 3.0626E+04, 3.0869E+04,
     + 3.1113E+04, 3.1359E+04, 3.1606E+04, 3.1854E+04, 3.2104E+04,
     +  3.2355E+04, 3.2608E+04, 3.2862E+04, 3.3117E+04, 3.3373E+04,
     + 3.3631E+04, 3.3890E+04, 3.4151E+04, 3.4413E+04, 3.4677E+04,
     +  3.4942E+04, 3.5209E+04, 3.5477E+04, 3.5746E+04, 3.6017E+04,
     + 3.6289E+04, 3.6563E+04, 3.6838E+04, 3.7114E+04, 3.7393E+04,
     +  3.7672E+04, 3.7954E+04, 3.8236E+04, 3.8520E+04, 3.8806E+04,
     + 3.9093E+04, 3.9382E+04, 3.9672E+04, 3.9964E+04, 4.0258E+04,
     +  4.0553E+04, 4.0849E+04, 4.1147E+04, 4.1447E+04, 4.1748E+04,
     + 4.2051E+04, 4.2355E+04, 4.2661E+04, 4.2969E+04, 4.3279E+04,
     +  4.3589E+04, 4.3902E+04, 4.4217E+04, 4.4532E+04, 4.4850E+04,
     + 4.5169E+04, 4.5490E+04, 4.5812E+04, 4.6137E+04, 4.6463E+04,
     +  4.6791E+04, 4.7120E+04, 4.7451E+04, 4.7784E+04, 4.8119E+04,
     + 4.8455E+04, 4.8793E+04, 4.9133E+04, 4.9474E+04, 4.9818E+04,
     +  5.0163E+04, 5.0510E+04, 5.0858E+04, 5.1209E+04, 5.1562E+04,
     + 5.1916E+04, 5.2272E+04, 5.2629E+04, 5.2989E+04, 5.3351E+04,
     +  5.3714E+04, 5.4079E+04, 5.4446E+04, 5.4815E+04, 5.5186E+04,
     + 5.5559E+04, 5.5934E+04, 5.6310E+04, 5.6689E+04, 5.7069E+04,
     +  5.7451E+04, 5.7836E+04, 5.8222E+04, 5.8610E+04, 5.9000E+04,
     + 5.9393E+04, 5.9787E+04, 6.0183E+04, 6.0581E+04, 6.0981E+04,
     +  6.1383E+04, 6.1787E+04, 6.2194E+04, 6.2602E+04, 6.3012E+04,
     + 6.3424E+04, 6.3839E+04, 6.4255E+04, 6.4674E+04, 6.5094E+04,
     +  6.5517E+04, 6.5942E+04, 6.6369E+04, 6.6798E+04, 6.7229E+04,
     + 6.7663E+04, 6.8098E+04, 6.8535E+04, 6.8976E+04, 6.9417E+04,
     +  6.9861E+04, 7.0308E+04, 7.0756E+04, 7.1207E+04, 7.1660E+04,
     + 7.2115E+04, 7.2572E+04, 7.3032E+04, 7.3494E+04, 7.3958E+04,
     +  7.4424E+04, 7.4893E+04, 7.5364E+04, 7.5837E+04, 7.6312E+04,
     + 7.6790E+04, 7.7270E+04, 7.7753E+04, 7.8237E+04, 7.8725E+04,
     +  7.9215E+04, 7.9706E+04, 8.0200E+04, 8.0697E+04, 8.1196E+04,
     + 8.1697E+04, 8.2201E+04, 8.2707E+04, 8.3215E+04, 8.3727E+04,
     +  8.4240E+04, 8.4756E+04, 8.5274E+04, 8.5795E+04, 8.6319E+04,
     + 8.6844E+04, 8.7372E+04, 8.7903E+04, 8.8437E+04, 8.8973E+04,
     +  8.9510E+04, 9.0051E+04, 9.0595E+04, 9.1141E+04, 9.1690E+04,
     + 9.2241E+04, 9.2794E+04, 9.3350E+04, 9.3909E+04, 9.4471E+04,
     +  9.5035E+04, 9.5602E+04, 9.6171E+04, 9.6743E+04, 9.7318E+04,
     + 9.7895E+04, 9.8475E+04, 9.9058E+04, 9.9643E+04, 1.0023E+05,
     +  1.0082E+05, 1.0142E+05, 1.0201E+05, 1.0261E+05, 1.0321E+05,
     + 1.0382E+05, 1.0442E+05, 1.0504E+05, 1.0565E+05, 1.0626E+05,
     +  1.0688E+05, 1.0750E+05, 1.0813E+05, 1.0875E+05, 1.0938E+05,
     + 1.1002E+05, 1.1065E+05, 1.1129E+05, 1.1193E+05, 1.1258E+05,
     +  1.1322E+05, 1.1387E+05, 1.1453E+05, 1.1518E+05, 1.1584E+05,
     + 1.1650E+05, 1.1717E+05, 1.1784E+05, 1.1851E+05, 1.1918E+05,
     +  1.1986E+05, 1.2054E+05, 1.2122E+05, 1.2191E+05, 1.2259E+05,
     + 1.2329E+05, 1.2398E+05, 1.2468E+05, 1.2538E+05, 1.2609E+05,
     +  1.2679E+05, 1.2750E+05, 1.2822E+05, 1.2893E+05, 1.2965E+05,
     + 1.3038E+05, 1.3110E+05, 1.3183E+05, 1.3257E+05, 1.3330E+05,
     +  1.3404E+05, 1.3478E+05, 1.3553E+05, 1.3628E+05, 1.3703E+05,
     + 1.3778E+05, 1.3854E+05, 1.3930E+05, 1.4007E+05, 1.4084E+05,
     +  1.4161E+05, 1.4239E+05, 1.4316E+05, 1.4395E+05, 1.4473E+05,
     + 1.4552E+05, 1.4631E+05, 1.4711E+05, 1.4791E+05, 1.4871E+05,
     +  1.4951E+05, 1.5032E+05, 1.5114E+05, 1.5195E+05, 1.5277E+05,
     + 1.5359E+05, 1.5442E+05, 1.5525E+05, 1.5608E+05, 1.5692E+05,
     +  1.5776E+05, 1.5861E+05, 1.5945E+05, 1.6030E+05, 1.6116E+05,
     + 1.6202E+05, 1.6288E+05, 1.6375E+05, 1.6461E+05, 1.6549E+05,
     +  1.6636E+05, 1.6724E+05, 1.6813E+05, 1.6902E+05, 1.6991E+05,
     + 1.7080E+05, 1.7170E+05, 1.7260E+05, 1.7351E+05, 1.7442E+05,
     +  1.7533E+05, 1.7625E+05, 1.7717E+05, 1.7810E+05, 1.7902E+05,
     + 1.7996E+05, 1.8089E+05, 1.8183E+05, 1.8278E+05, 1.8373E+05,
     +  1.8468E+05, 1.8563E+05, 1.8659E+05, 1.8756E+05, 1.8853E+05,
     + 1.8950E+05, 1.9047E+05, 1.9145E+05, 1.9243E+05, 1.9342E+05,
     +  1.9441E+05, 1.9541E+05, 1.9641E+05, 1.9741E+05, 1.9842E+05,
     + 1.9943E+05, 2.0045E+05, 2.0147E+05, 2.0249E+05, 2.0352E+05,
     +  2.0455E+05, 2.0559E+05, 2.0663E+05, 2.0767E+05, 2.0872E+05,
     + 2.0977E+05, 2.1083E+05, 2.1189E+05, 2.1296E+05, 2.1403E+05,
     +  2.1510E+05, 2.1618E+05, 2.1726E+05, 2.1835E+05, 2.1944E+05,
     + 2.2054E+05, 2.2164E+05, 2.2274E+05, 2.2385E+05, 2.2496E+05,
     +  2.2608E+05, 2.2720E+05, 2.2833E+05, 2.2945E+05, 2.3059E+05,
     + 2.3173E+05, 2.3287E+05, 2.3402E+05, 2.3517E+05, 2.3633E+05,
     +  2.3749E+05, 2.3866E+05, 2.3983E+05, 2.4100E+05, 2.4218E+05,
     + 2.4337E+05, 2.4456E+05, 2.4575E+05, 2.4695E+05, 2.4815E+05,
     +  2.4936E+05, 2.5057E+05, 2.5178E+05, 2.5301E+05, 2.5423E+05,
     + 2.5546E+05, 2.5670E+05, 2.5794E+05, 2.5918E+05, 2.6043E+05,
     +  2.6169E+05, 2.6295E+05, 2.6421E+05, 2.6548E+05, 2.6675E+05,
     + 2.6803E+05, 2.6931E+05, 2.7060E+05, 2.7189E+05, 2.7319E+05,
     +  2.7449E+05, 2.7580E+05, 2.7711E+05, 2.7843E+05, 2.7975E+05,
     + 2.8108E+05, 2.8241E+05, 2.8375E+05, 2.8509E+05, 2.8644E+05,
     +  2.8779E+05, 2.8915E+05, 2.9051E+05, 2.9188E+05, 2.9325E+05,
     + 2.9463E+05, 2.9601E+05, 2.9740E+05, 2.9879E+05, 3.0019E+05,
     +  3.0159E+05, 3.0300E+05, 3.0441E+05, 3.0583E+05, 3.0726E+05,
     + 3.0868E+05, 3.1012E+05, 3.1156E+05, 3.1300E+05, 3.1445E+05,
     +  3.1591E+05, 3.1737E+05, 3.1884E+05, 3.2031E+05, 3.2179E+05,
     + 3.2327E+05, 3.2476E+05, 3.2625E+05, 3.2775E+05, 3.2925E+05,
     +  3.3076E+05, 3.3228E+05, 3.3380E+05, 3.3532E+05, 3.3685E+05,
     + 3.3839E+05, 3.3993E+05, 3.4148E+05, 3.4304E+05, 3.4459E+05,
     +  3.4616E+05, 3.4773E+05, 3.4931E+05, 3.5089E+05, 3.5248E+05,
     + 3.5407E+05, 3.5567E+05, 3.5727E+05, 3.5888E+05, 3.6050E+05,
     +  3.6212E+05, 3.6375E+05, 3.6538E+05, 3.6702E+05, 3.6866E+05,
     + 3.7031E+05, 3.7197E+05, 3.7363E+05, 3.7530E+05, 3.7697E+05,
     +  3.7866E+05, 3.8034E+05, 3.8203E+05, 3.8373E+05, 3.8543E+05,
     + 3.8714E+05, 3.8886E+05, 3.9058E+05, 3.9231E+05, 3.9404E+05,
     +  3.9578E+05, 3.9753E+05, 3.9928E+05, 4.0104E+05, 4.0280E+05,
     + 4.0457E+05, 4.0635E+05, 4.0813E+05, 4.0992E+05, 4.1171E+05,
     +  4.1352E+05, 4.1532E+05, 4.1714E+05, 4.1895E+05, 4.2078E+05,
     + 4.2261E+05, 4.2445E+05, 4.2630E+05, 4.2815E+05, 4.3001E+05,
     +  4.3187E+05, 4.3374E+05, 4.3562E+05, 4.3750E+05, 4.3939E+05,
     + 4.4129E+05, 4.4319E+05, 4.4510E+05, 4.4701E+05, 4.4893E+05,
     +  4.5086E+05, 4.5280E+05, 4.5474E+05, 4.5669E+05, 4.5864E+05,
     + 4.6060E+05, 4.6257E+05, 4.6454E+05, 4.6653E+05, 4.6851E+05,
     +  4.7051E+05, 4.7251E+05, 4.7452E+05, 4.7653E+05, 4.7855E+05,
     + 4.8058E+05, 4.8262E+05, 4.8466E+05, 4.8671E+05, 4.8876E+05,
     +  4.9083E+05, 4.9289E+05, 4.9497E+05, 4.9705E+05, 4.9914E+05,
     + 5.0124E+05, 5.0334E+05, 5.0545E+05, 5.0757E+05, 5.0970E+05,
     +  5.1183E+05, 5.1397E+05, 5.1611E+05, 5.1827E+05, 5.2043E+05,
     + 5.2259E+05, 5.2477E+05, 5.2695E+05, 5.2914E+05, 5.3133E+05,
     +  5.3354E+05, 5.3574E+05, 5.3796E+05, 5.4019E+05, 5.4242E+05,
     + 5.4466E+05, 5.4690E+05, 5.4916E+05, 5.5142E+05, 5.5369E+05,
     +  5.5596E+05, 5.5824E+05, 5.6053E+05, 5.6283E+05, 5.6514E+05,
     + 5.6745E+05, 5.6977E+05, 5.7210E+05, 5.7443E+05, 5.7677E+05,
     +  5.7913E+05, 5.8148E+05, 5.8385E+05, 5.8622E+05, 5.8860E+05,
     + 5.9099E+05, 5.9339E+05, 5.9579E+05, 5.9820E+05, 6.0062E+05,
     +  6.0304E+05, 6.0548E+05, 6.0792E+05, 6.1037E+05, 6.1283E+05,
     + 6.1529E+05, 6.1777E+05, 6.2025E+05, 6.2274E+05, 6.2523E+05,
     +  6.2774E+05, 6.3025E+05, 6.3277E+05, 6.3530E+05, 6.3783E+05,
     + 6.4038E+05, 6.4293E+05, 6.4549E+05, 6.4806E+05, 6.5063E+05,
     +  6.5322E+05, 6.5581E+05, 6.5841E+05, 6.6102E+05, 6.6364E+05,
     + 6.6626E+05, 6.6890E+05, 6.7154E+05, 6.7419E+05, 6.7684E+05,
     +  6.7951E+05, 6.8219E+05, 6.8487E+05, 6.8756E+05, 6.9026E+05,
     + 6.9297E+05, 6.9568E+05, 6.9841E+05, 7.0114E+05, 7.0388E+05,
     +  7.0663E+05, 7.0939E+05, 7.1216E+05, 7.1493E+05, 7.1771E+05,
     + 7.2051E+05, 7.2331E+05, 7.2612E+05, 7.2893E+05, 7.3176E+05,
     +  7.3460E+05, 7.3744E+05, 7.4029E+05, 7.4316E+05, 7.4602E+05,
     + 7.4890E+05, 7.5179E+05, 7.5469E+05, 7.5759E+05, 7.6051E+05,
     +  7.6343E+05, 7.6636E+05, 7.6930E+05, 7.7225E+05, 7.7521E+05,
     + 7.7817E+05, 7.8115E+05, 7.8414E+05, 7.8713E+05, 7.9013E+05,
     +  7.9315E+05, 7.9617E+05, 7.9920E+05, 8.0224E+05, 8.0528E+05,
     + 8.0834E+05, 8.1141E+05, 8.1448E+05, 8.1757E+05, 8.2066E+05,
     +  8.2377E+05, 8.2688E+05, 8.3000E+05, 8.3313E+05, 8.3627E+05,
     + 8.3942E+05, 8.4258E+05, 8.4575E+05, 8.4893E+05, 8.5211E+05,
     +  8.5531E+05, 8.5852E+05, 8.6173E+05, 8.6496E+05, 8.6819E+05,
     + 8.7144E+05, 8.7469E+05, 8.7795E+05, 8.8123E+05, 8.8451E+05,
     +  8.8780E+05, 8.9110E+05, 8.9441E+05, 8.9773E+05, 9.0107E+05,
     + 9.0441E+05, 9.0776E+05, 9.1112E+05, 9.1449E+05, 9.1787E+05,
     +  9.2126E+05, 9.2466E+05, 9.2806E+05, 9.3148E+05, 9.3491E+05,
     + 9.3835E+05, 9.4180E+05, 9.4526E+05, 9.4873E+05, 9.5221E+05,
     +  9.5570E+05, 9.5919E+05, 9.6270E+05, 9.6622E+05, 9.6975E+05,
     + 9.7329E+05, 9.7684E+05, 9.8040E+05, 9.8397E+05, 9.8755E+05,
     +  9.9114E+05, 9.9474E+05, 9.9835E+05, 1.0020E+06, 1.0056E+06,
     + 1.0092E+06, 1.0129E+06, 1.0166E+06, 1.0202E+06, 1.0239E+06,
     +  1.0276E+06, 1.0313E+06, 1.0350E+06, 1.0388E+06, 1.0425E+06,
     + 1.0462E+06, 1.0500E+06, 1.0538E+06, 1.0575E+06, 1.0613E+06,
     +  1.0651E+06, 1.0689E+06, 1.0728E+06, 1.0766E+06, 1.0804E+06,
     + 1.0843E+06, 1.0882E+06, 1.0920E+06, 1.0959E+06, 1.0998E+06,
     +  1.1037E+06, 1.1076E+06, 1.1116E+06, 1.1155E+06, 1.1195E+06,
     + 1.1234E+06, 1.1274E+06, 1.1314E+06, 1.1354E+06, 1.1394E+06,
     +  1.1434E+06, 1.1474E+06, 1.1515E+06, 1.1555E+06, 1.1596E+06,
     + 1.1637E+06, 1.1678E+06, 1.1719E+06, 1.1760E+06, 1.1801E+06,
     +  1.1842E+06, 1.1884E+06, 1.1925E+06, 1.1967E+06, 1.2009E+06,
     + 1.2050E+06, 1.2092E+06, 1.2135E+06, 1.2177E+06, 1.2219E+06,
     +  1.2262E+06, 1.2304E+06, 1.2347E+06, 1.2390E+06, 1.2433E+06,
     + 1.2476E+06, 1.2519E+06, 1.2562E+06, 1.2605E+06, 1.2649E+06,
     +  1.2693E+06, 1.2736E+06, 1.2780E+06, 1.2824E+06, 1.2868E+06,
     + 1.2912E+06, 1.2957E+06, 1.3001E+06, 1.3046E+06, 1.3091E+06,
     +  1.3135E+06, 1.3180E+06, 1.3225E+06, 1.3271E+06, 1.3316E+06,
     + 1.3361E+06, 1.3407E+06, 1.3452E+06, 1.3498E+06, 1.3544E+06,
     +  1.3590E+06, 1.3636E+06, 1.3683E+06, 1.3729E+06, 1.3776E+06,
     + 1.3822E+06, 1.3869E+06, 1.3916E+06, 1.3963E+06, 1.4010E+06,
     +  1.4057E+06, 1.4105E+06, 1.4152E+06, 1.4200E+06, 1.4248E+06,
     + 1.4296E+06, 1.4344E+06, 1.4392E+06, 1.4440E+06, 1.4489E+06,
     +  1.4537E+06, 1.4586E+06, 1.4635E+06, 1.4684E+06, 1.4733E+06,
     + 1.4782E+06, 1.4831E+06, 1.4881E+06, 1.4930E+06, 1.4980E+06,
     +  1.5030E+06, 1.5080E+06, 1.5130E+06, 1.5180E+06, 1.5230E+06,
     + 1.5281E+06, 1.5331E+06, 1.5382E+06, 1.5433E+06, 1.5484E+06,
     +  1.5535E+06, 1.5587E+06, 1.5638E+06, 1.5689E+06, 1.5741E+06,
     + 1.5793E+06, 1.5845E+06, 1.5897E+06, 1.5949E+06, 1.6002E+06,
     +  1.6054E+06, 1.6107E+06, 1.6159E+06, 1.6212E+06, 1.6265E+06,
     + 1.6319E+06, 1.6372E+06, 1.6425E+06, 1.6479E+06, 1.6533E+06,
     +  1.6586E+06, 1.6640E+06, 1.6695E+06, 1.6749E+06, 1.6803E+06,
     + 1.6858E+06, 1.6912E+06, 1.6967E+06, 1.7022E+06, 1.7077E+06,
     +  1.7133E+06, 1.7188E+06, 1.7243E+06, 1.7299E+06, 1.7355E+06,
     + 1.7411E+06, 1.7467E+06, 1.7523E+06, 1.7580E+06, 1.7636E+06,
     +  1.7693E+06, 1.7750E+06, 1.7806E+06, 1.7864E+06, 1.7921E+06,
     + 1.7978E+06, 1.8036E+06, 1.8093E+06, 1.8151E+06, 1.8209E+06,
     +  1.8267E+06, 1.8326E+06, 1.8384E+06, 1.8442E+06, 1.8501E+06,
     + 1.8560E+06, 1.8619E+06, 1.8678E+06, 1.8737E+06, 1.8797E+06,
     +  1.8856E+06, 1.8916E+06, 1.8976E+06, 1.9036E+06, 1.9096E+06,
     + 1.9156E+06, 1.9217E+06, 1.9277E+06, 1.9338E+06, 1.9399E+06,
     +  1.9460E+06, 1.9521E+06, 1.9583E+06, 1.9644E+06, 1.9706E+06,
     + 1.9768E+06, 1.9830E+06, 1.9892E+06, 1.9954E+06, 2.0017E+06,
     +  2.0079E+06, 2.0142E+06, 2.0205E+06, 2.0268E+06, 2.0331E+06,
     + 2.0394E+06, 2.0458E+06, 2.0522E+06, 2.0585E+06, 2.0649E+06,
     +  2.0714E+06, 2.0778E+06, 2.0842E+06, 2.0907E+06, 2.0972E+06,
     + 2.1037E+06, 2.1102E+06, 2.1167E+06, 2.1232E+06, 2.1298E+06,
     +  2.1364E+06, 2.1429E+06, 2.1495E+06, 2.1562E+06, 2.1628E+06,
     + 2.1695E+06, 2.1761E+06, 2.1828E+06, 2.1895E+06, 2.1962E+06,
     +  2.2030E+06, 2.2097E+06, 2.2165E+06, 2.2232E+06, 2.2300E+06,
     + 2.2369E+06, 2.2437E+06, 2.2505E+06, 2.2574E+06, 2.2643E+06,
     +  2.2712E+06, 2.2781E+06, 2.2850E+06, 2.2919E+06, 2.2989E+06,
     + 2.3059E+06, 2.3129E+06, 2.3199E+06, 2.3269E+06, 2.3340E+06,
     +  2.3410E+06, 2.3481E+06, 2.3552E+06, 2.3623E+06, 2.3694E+06,
     + 2.3766E+06, 2.3837E+06, 2.3909E+06, 2.3981E+06, 2.4053E+06,
     +  2.4126E+06, 2.4198E+06, 2.4271E+06, 2.4344E+06, 2.4417E+06,
     + 2.4490E+06, 2.4563E+06, 2.4637E+06, 2.4710E+06, 2.4784E+06,
     +  2.4858E+06, 2.4932E+06, 2.5007E+06, 2.5081E+06, 2.5156E+06,
     + 2.5231E+06, 2.5306E+06, 2.5381E+06, 2.5457E+06, 2.5532E+06,
     +  2.5608E+06, 2.5684E+06, 2.5760E+06, 2.5836E+06, 2.5913E+06,
     + 2.5989E+06, 2.6066E+06, 2.6143E+06, 2.6221E+06, 2.6298E+06,
     +  2.6375E+06, 2.6453E+06, 2.6531E+06, 2.6609E+06, 2.6687E+06,
     + 2.6766E+06, 2.6845E+06, 2.6923E+06, 2.7002E+06, 2.7082E+06,
     +  2.7161E+06, 2.7240E+06, 2.7320E+06, 2.7400E+06, 2.7480E+06,
     + 2.7560E+06, 2.7641E+06, 2.7722E+06, 2.7802E+06, 2.7883E+06,
     +  2.7965E+06, 2.8046E+06, 2.8128E+06, 2.8209E+06, 2.8291E+06,
     + 2.8373E+06, 2.8456E+06, 2.8538E+06, 2.8621E+06, 2.8704E+06,
     +  2.8787E+06, 2.8870E+06, 2.8954E+06, 2.9037E+06, 2.9121E+06,
     + 2.9205E+06, 2.9289E+06, 2.9374E+06, 2.9458E+06, 2.9543E+06,
     +  2.9628E+06, 2.9713E+06, 2.9798E+06, 2.9884E+06, 2.9970E+06,
     + 3.0056E+06, 3.0142E+06, 3.0228E+06, 3.0315E+06, 3.0401E+06,
     +  3.0488E+06, 3.0575E+06, 3.0663E+06, 3.0750E+06, 3.0838E+06,
     + 3.0926E+06, 3.1014E+06, 3.1102E+06, 3.1190E+06, 3.1279E+06,
     +  3.1368E+06, 3.1457E+06, 3.1546E+06, 3.1636E+06, 3.1725E+06,
     + 3.1815E+06, 3.1905E+06, 3.1996E+06, 3.2086E+06, 3.2177E+06,
     +  3.2268E+06, 3.2359E+06, 3.2450E+06, 3.2541E+06, 3.2633E+06,
     + 3.2725E+06, 3.2817E+06, 3.2909E+06, 3.3002E+06, 3.3094E+06,
     +  3.3187E+06, 3.3280E+06, 3.3374E+06, 3.3467E+06, 3.3561E+06,
     + 3.3655E+06, 3.3749E+06, 3.3843E+06, 3.3938E+06, 3.4033E+06,
     +  3.4128E+06, 3.4223E+06, 3.4318E+06, 3.4414E+06, 3.4509E+06,
     + 3.4605E+06, 3.4702E+06, 3.4798E+06, 3.4895E+06, 3.4991E+06,
     +  3.5088E+06, 3.5186E+06, 3.5283E+06, 3.5381E+06, 3.5479E+06,
     + 3.5577E+06, 3.5675E+06, 3.5774E+06, 3.5872E+06, 3.5971E+06,
     +  3.6071E+06, 3.6170E+06, 3.6270E+06, 3.6369E+06, 3.6469E+06,
     + 3.6570E+06, 3.6670E+06, 3.6771E+06, 3.6872E+06, 3.6973E+06,
     +  3.7074E+06, 3.7176E+06, 3.7277E+06, 3.7379E+06, 3.7482E+06,
     + 3.7584E+06, 3.7687E+06, 3.7790E+06, 3.7893E+06, 3.7996E+06,
     +  3.8099E+06, 3.8203E+06, 3.8307E+06, 3.8411E+06, 3.8516E+06,
     + 3.8620E+06, 3.8725E+06, 3.8830E+06, 3.8936E+06, 3.9041E+06,
     +  3.9147E+06, 3.9253E+06, 3.9359E+06, 3.9466E+06, 3.9572E+06,
     + 3.9679E+06, 3.9786E+06, 3.9894E+06, 4.0001E+06, 4.0109E+06,
     +  4.0217E+06, 4.0325E+06, 4.0434E+06, 4.0542E+06, 4.0651E+06,
     + 4.0760E+06, 4.0870E+06, 4.0979E+06, 4.1089E+06, 4.1199E+06,
     +  4.1310E+06, 4.1420E+06, 4.1531E+06, 4.1642E+06, 4.1753E+06,
     + 4.1865E+06, 4.1976E+06, 4.2088E+06, 4.2200E+06, 4.2313E+06,
     +  4.2426E+06, 4.2538E+06, 4.2652E+06, 4.2765E+06, 4.2878E+06,
     + 4.2992E+06, 4.3106E+06, 4.3221E+06, 4.3335E+06, 4.3450E+06,
     +  4.3565E+06, 4.3680E+06, 4.3796E+06, 4.3911E+06, 4.4027E+06,
     + 4.4144E+06, 4.4260E+06, 4.4377E+06, 4.4494E+06, 4.4611E+06,
     +  4.4728E+06, 4.4846E+06, 4.4964E+06, 4.5082E+06, 4.5200E+06,
     + 4.5319E+06, 4.5438E+06, 4.5557E+06, 4.5676E+06, 4.5796E+06,
     +  4.5916E+06, 4.6036E+06, 4.6156E+06, 4.6277E+06, 4.6398E+06,
     + 4.6519E+06, 4.6640E+06, 4.6762E+06, 4.6884E+06, 4.7006E+06,
     +  4.7128E+06, 4.7251E+06, 4.7373E+06, 4.7497E+06, 4.7620E+06,
     + 4.7743E+06, 4.7867E+06, 4.7991E+06, 4.8116E+06, 4.8240E+06,
     +  4.8365E+06, 4.8490E+06, 4.8616E+06, 4.8741E+06, 4.8867E+06,
     + 4.8993E+06, 4.9120E+06, 4.9246E+06, 4.9373E+06, 4.9500E+06,
     +  4.9628E+06, 4.9755E+06, 4.9883E+06, 5.0011E+06, 5.0140E+06,
     + 5.0269E+06, 5.0397E+06, 5.0527E+06, 5.0656E+06, 5.0786E+06,
     +  5.0916E+06, 5.1046E+06, 5.1177E+06, 5.1307E+06, 5.1438E+06,
     + 5.1570E+06, 5.1701E+06, 5.1833E+06, 5.1965E+06, 5.2097E+06,
     +  5.2230E+06, 5.2363E+06, 5.2496E+06, 5.2630E+06, 5.2763E+06,
     + 5.2897E+06, 5.3031E+06, 5.3166E+06, 5.3301E+06, 5.3436E+06,
     +  5.3571E+06, 5.3706E+06, 5.3842E+06, 5.3978E+06, 5.4115E+06,
     + 5.4251E+06, 5.4388E+06, 5.4525E+06, 5.4663E+06, 5.4800E+06,
     +  5.4938E+06, 5.5077E+06, 5.5215E+06, 5.5354E+06, 5.5493E+06,
     + 5.5632E+06, 5.5772E+06, 5.5912E+06, 5.6052E+06, 5.6193E+06,
     +  5.6333E+06, 5.6474E+06, 5.6616E+06, 5.6757E+06, 5.6899E+06,
     + 5.7041E+06, 5.7183E+06, 5.7326E+06, 5.7469E+06, 5.7612E+06,
     +  5.7756E+06, 5.7900E+06, 5.8044E+06, 5.8188E+06, 5.8333E+06,
     + 5.8478E+06, 5.8623E+06, 5.8768E+06, 5.8914E+06, 5.9060E+06,
     +  5.9206E+06, 5.9353E+06, 5.9500E+06, 5.9647E+06, 5.9794E+06,
     + 5.9942E+06, 6.0090E+06, 6.0239E+06, 6.0387E+06, 6.0536E+06,
     +  6.0685E+06, 6.0835E+06, 6.0984E+06, 6.1135E+06, 6.1285E+06,
     + 6.1435E+06, 6.1586E+06, 6.1738E+06, 6.1889E+06, 6.2041E+06,
     +  6.2193E+06, 6.2345E+06, 6.2498E+06, 6.2651E+06, 6.2804E+06,
     + 6.2958E+06, 6.3112E+06, 6.3266E+06, 6.3420E+06, 6.3575E+06,
     +  6.3730E+06, 6.3885E+06, 6.4041E+06, 6.4197E+06, 6.4353E+06,
     + 6.4510E+06, 6.4666E+06, 6.4823E+06, 6.4981E+06, 6.5139E+06,
     +  6.5297E+06, 6.5455E+06, 6.5613E+06, 6.5772E+06, 6.5932E+06,
     + 6.6091E+06, 6.6251E+06, 6.6411E+06, 6.6572E+06, 6.6732E+06,
     +  6.6893E+06, 6.7055E+06, 6.7216E+06, 6.7378E+06, 6.7540E+06,
     + 6.7703E+06, 6.7866E+06, 6.8029E+06, 6.8192E+06, 6.8356E+06,
     +  6.8520E+06, 6.8685E+06, 6.8850E+06, 6.9015E+06, 6.9180E+06,
     + 6.9346E+06, 6.9511E+06, 6.9678E+06, 6.9844E+06, 7.0011E+06,
     +  7.0178E+06, 7.0346E+06, 7.0514E+06, 7.0682E+06, 7.0850E+06,
     + 7.1019E+06, 7.1188E+06, 7.1358E+06, 7.1527E+06, 7.1697E+06,
     +  7.1868E+06, 7.2038E+06, 7.2209E+06, 7.2381E+06, 7.2552E+06,
     + 7.2724E+06, 7.2897E+06, 7.3069E+06, 7.3242E+06, 7.3415E+06,
     +  7.3589E+06, 7.3763E+06, 7.3937E+06, 7.4111E+06, 7.4286E+06,
     + 7.4461E+06, 7.4637E+06, 7.4813E+06, 7.4989E+06, 7.5165E+06,
     +  7.5342E+06, 7.5519E+06, 7.5697E+06, 7.5874E+06, 7.6052E+06,
     + 7.6231E+06, 7.6410E+06, 7.6589E+06, 7.6768E+06, 7.6948E+06,
     +  7.7128E+06, 7.7308E+06, 7.7489E+06, 7.7670E+06, 7.7851E+06,
     + 7.8033E+06, 7.8215E+06, 7.8397E+06, 7.8580E+06, 7.8763E+06,
     +  7.8947E+06, 7.9130E+06, 7.9314E+06, 7.9499E+06, 7.9684E+06,
     + 7.9869E+06, 8.0054E+06, 8.0240E+06, 8.0426E+06, 8.0612E+06,
     +  8.0799E+06, 8.0986E+06, 8.1173E+06, 8.1361E+06, 8.1549E+06,
     + 8.1738E+06, 8.1927E+06, 8.2116E+06, 8.2305E+06, 8.2495E+06,
     +  8.2685E+06, 8.2876E+06, 8.3067E+06, 8.3258E+06, 8.3449E+06,
     + 8.3641E+06, 8.3833E+06, 8.4026E+06, 8.4219E+06, 8.4412E+06,
     +  8.4606E+06, 8.4800E+06, 8.4994E+06, 8.5189E+06, 8.5384E+06,
     + 8.5579E+06, 8.5775E+06, 8.5971E+06, 8.6167E+06, 8.6364E+06,
     +  8.6561E+06, 8.6759E+06, 8.6956E+06, 8.7155E+06, 8.7353E+06,
     + 8.7552E+06, 8.7751E+06, 8.7951E+06, 8.8151E+06, 8.8351E+06,
     +  8.8552E+06, 8.8753E+06, 8.8954E+06, 8.9156E+06, 8.9358E+06,
     + 8.9560E+06, 8.9763E+06, 8.9966E+06, 9.0170E+06, 9.0374E+06,
     +  9.0578E+06, 9.0783E+06, 9.0988E+06, 9.1193E+06, 9.1399E+06,
     + 9.1605E+06, 9.1811E+06, 9.2018E+06, 9.2225E+06, 9.2433E+06,
     +  9.2641E+06, 9.2849E+06, 9.3058E+06, 9.3267E+06, 9.3476E+06,
     + 9.3686E+06, 9.3896E+06, 9.4106E+06, 9.4317E+06, 9.4528E+06,
     +  9.4740E+06, 9.4952E+06, 9.5164E+06, 9.5377E+06, 9.5590E+06,
     + 9.5803E+06, 9.6017E+06, 9.6231E+06, 9.6446E+06, 9.6661E+06,
     +  9.6876E+06, 9.7092E+06, 9.7308E+06, 9.7524E+06, 9.7741E+06,
     + 9.7958E+06, 9.8176E+06, 9.8394E+06, 9.8612E+06, 9.8831E+06,
     +  9.9050E+06, 9.9270E+06, 9.9489E+06, 9.9710E+06, 9.9930E+06,
     + 1.0015E+07, 1.0037E+07, 1.0059E+07, 1.0082E+07, 1.0104E+07,
     +  1.0126E+07, 1.0149E+07, 1.0171E+07, 1.0193E+07, 1.0216E+07,
     + 1.0238E+07, 1.0261E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_NH3   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 2),QofT( 2,NT),Q(NT)
      data (Tmx(I),I=1,2)/6000.,6000./ 
      data xgj/ 3.,2./
c...      NH3
c...        --      4111
      data (QofT( 1,J),J=1,2408)/ 3.8341E+00, 6.7833E+00, 8.2192E+00,
     +  9.1440E+00, 9.9774E+00, 1.0909E+01, 1.2004E+01, 1.3266E+01,
     + 1.4676E+01, 1.6208E+01, 1.7840E+01, 1.9556E+01, 2.1344E+01,
     +  2.3195E+01, 2.5105E+01, 2.7069E+01, 2.9086E+01, 3.1156E+01,
     + 3.3277E+01, 3.5450E+01, 3.9949E+01, 4.4653E+01, 4.9562E+01,
     +  5.4672E+01, 5.9980E+01, 6.5482E+01, 7.1174E+01, 7.7049E+01,
     + 8.3103E+01, 8.9331E+01, 9.5726E+01, 1.0229E+02, 1.0900E+02,
     +  1.1587E+02, 1.2290E+02, 1.3006E+02, 1.3737E+02, 1.4482E+02,
     + 1.5241E+02, 1.6012E+02, 1.6797E+02, 1.7595E+02, 1.8405E+02,
     +  1.9227E+02, 2.0062E+02, 2.0909E+02, 2.1767E+02, 2.2637E+02,
     + 2.3518E+02, 2.4411E+02, 2.5314E+02, 2.6229E+02, 2.7155E+02,
     +  2.8091E+02, 2.9038E+02, 2.9995E+02, 3.0963E+02, 3.1941E+02,
     + 3.2929E+02, 3.3927E+02, 3.4935E+02, 3.5953E+02, 3.6980E+02,
     +  3.8018E+02, 3.9064E+02, 4.0121E+02, 4.1187E+02, 4.2262E+02,
     + 4.3346E+02, 4.4440E+02, 4.5542E+02, 4.6654E+02, 4.7775E+02,
     +  4.8905E+02, 5.0043E+02, 5.1190E+02, 5.2346E+02, 5.3511E+02,
     + 5.4684E+02, 5.5866E+02, 5.7056E+02, 5.8255E+02, 5.9462E+02,
     +  6.0678E+02, 6.1902E+02, 6.3134E+02, 6.4374E+02, 6.5623E+02,
     + 6.6880E+02, 6.8145E+02, 6.9418E+02, 7.0699E+02, 7.1988E+02,
     +  7.3285E+02, 7.4590E+02, 7.5902E+02, 7.7223E+02, 7.8552E+02,
     + 7.9888E+02, 8.1233E+02, 8.2585E+02, 8.3945E+02, 8.5313E+02,
     +  8.6688E+02, 8.8072E+02, 8.9463E+02, 9.0862E+02, 9.2268E+02,
     + 9.3682E+02, 9.5104E+02, 9.6534E+02, 9.7971E+02, 9.9416E+02,
     +  1.0087E+03, 1.0233E+03, 1.0380E+03, 1.0527E+03, 1.0676E+03,
     + 1.0825E+03, 1.0975E+03, 1.1125E+03, 1.1277E+03, 1.1429E+03,
     +  1.1582E+03, 1.1735E+03, 1.1890E+03, 1.2045E+03, 1.2201E+03,
     + 1.2358E+03, 1.2516E+03, 1.2674E+03, 1.2833E+03, 1.2993E+03,
     +  1.3154E+03, 1.3315E+03, 1.3478E+03, 1.3641E+03, 1.3805E+03,
     + 1.3969E+03, 1.4135E+03, 1.4301E+03, 1.4468E+03, 1.4636E+03,
     +  1.4805E+03, 1.4974E+03, 1.5144E+03, 1.5315E+03, 1.5487E+03,
     + 1.5660E+03, 1.5834E+03, 1.6008E+03, 1.6183E+03, 1.6359E+03,
     +  1.6536E+03, 1.6714E+03, 1.6893E+03, 1.7072E+03, 1.7252E+03,
     + 1.7433E+03, 1.7615E+03, 1.7798E+03, 1.7982E+03, 1.8166E+03,
     +  1.8352E+03, 1.8538E+03, 1.8725E+03, 1.8913E+03, 1.9102E+03,
     + 1.9292E+03, 1.9483E+03, 1.9675E+03, 1.9867E+03, 2.0061E+03,
     +  2.0255E+03, 2.0450E+03, 2.0646E+03, 2.0843E+03, 2.1042E+03,
     + 2.1241E+03, 2.1440E+03, 2.1641E+03, 2.1843E+03, 2.2046E+03,
     +  2.2250E+03, 2.2454E+03, 2.2660E+03, 2.2867E+03, 2.3074E+03,
     + 2.3283E+03, 2.3493E+03, 2.3703E+03, 2.3915E+03, 2.4127E+03,
     +  2.4341E+03, 2.4556E+03, 2.4771E+03, 2.4988E+03, 2.5206E+03,
     + 2.5424E+03, 2.5644E+03, 2.5865E+03, 2.6087E+03, 2.6310E+03,
     +  2.6534E+03, 2.6759E+03, 2.6985E+03, 2.7212E+03, 2.7441E+03,
     + 2.7670E+03, 2.7900E+03, 2.8132E+03, 2.8365E+03, 2.8599E+03,
     +  2.8834E+03, 2.9070E+03, 2.9307E+03, 2.9545E+03, 2.9785E+03,
     + 3.0025E+03, 3.0267E+03, 3.0510E+03, 3.0754E+03, 3.1000E+03,
     +  3.1246E+03, 3.1494E+03, 3.1743E+03, 3.1993E+03, 3.2244E+03,
     + 3.2497E+03, 3.2750E+03, 3.3005E+03, 3.3262E+03, 3.3519E+03,
     +  3.3778E+03, 3.4038E+03, 3.4299E+03, 3.4561E+03, 3.4825E+03,
     + 3.5090E+03, 3.5356E+03, 3.5624E+03, 3.5893E+03, 3.6163E+03,
     +  3.6435E+03, 3.6707E+03, 3.6982E+03, 3.7257E+03, 3.7534E+03,
     + 3.7812E+03, 3.8092E+03, 3.8373E+03, 3.8655E+03, 3.8938E+03,
     +  3.9223E+03, 3.9510E+03, 3.9798E+03, 4.0087E+03, 4.0377E+03,
     + 4.0670E+03, 4.0963E+03, 4.1258E+03, 4.1554E+03, 4.1852E+03,
     +  4.2151E+03, 4.2452E+03, 4.2754E+03, 4.3058E+03, 4.3363E+03,
     + 4.3669E+03, 4.3977E+03, 4.4287E+03, 4.4598E+03, 4.4911E+03,
     +  4.5225E+03, 4.5541E+03, 4.5858E+03, 4.6177E+03, 4.6497E+03,
     + 4.6819E+03, 4.7142E+03, 4.7467E+03, 4.7794E+03, 4.8122E+03,
     +  4.8452E+03, 4.8783E+03, 4.9116E+03, 4.9451E+03, 4.9787E+03,
     + 5.0125E+03, 5.0465E+03, 5.0806E+03, 5.1149E+03, 5.1494E+03,
     +  5.1840E+03, 5.2188E+03, 5.2538E+03, 5.2889E+03, 5.3242E+03,
     + 5.3597E+03, 5.3954E+03, 5.4312E+03, 5.4672E+03, 5.5034E+03,
     +  5.5397E+03, 5.5763E+03, 5.6130E+03, 5.6499E+03, 5.6869E+03,
     + 5.7242E+03, 5.7616E+03, 5.7992E+03, 5.8370E+03, 5.8750E+03,
     +  5.9132E+03, 5.9515E+03, 5.9900E+03, 6.0288E+03, 6.0677E+03,
     + 6.1068E+03, 6.1461E+03, 6.1856E+03, 6.2252E+03, 6.2651E+03,
     +  6.3052E+03, 6.3454E+03, 6.3859E+03, 6.4265E+03, 6.4674E+03,
     + 6.5084E+03, 6.5496E+03, 6.5911E+03, 6.6327E+03, 6.6746E+03,
     +  6.7166E+03, 6.7589E+03, 6.8013E+03, 6.8440E+03, 6.8868E+03,
     + 6.9299E+03, 6.9732E+03, 7.0167E+03, 7.0604E+03, 7.1043E+03,
     +  7.1484E+03, 7.1928E+03, 7.2373E+03, 7.2821E+03, 7.3270E+03,
     + 7.3722E+03, 7.4177E+03, 7.4633E+03, 7.5091E+03, 7.5552E+03,
     +  7.6015E+03, 7.6480E+03, 7.6948E+03, 7.7417E+03, 7.7889E+03,
     + 7.8364E+03, 7.8840E+03, 7.9319E+03, 7.9800E+03, 8.0283E+03,
     +  8.0769E+03, 8.1257E+03, 8.1747E+03, 8.2240E+03, 8.2735E+03,
     + 8.3232E+03, 8.3732E+03, 8.4234E+03, 8.4739E+03, 8.5246E+03,
     +  8.5755E+03, 8.6267E+03, 8.6782E+03, 8.7298E+03, 8.7818E+03,
     + 8.8339E+03, 8.8863E+03, 8.9390E+03, 8.9919E+03, 9.0451E+03,
     +  9.0985E+03, 9.1522E+03, 9.2061E+03, 9.2603E+03, 9.3148E+03,
     + 9.3695E+03, 9.4244E+03, 9.4797E+03, 9.5352E+03, 9.5909E+03,
     +  9.6469E+03, 9.7032E+03, 9.7597E+03, 9.8166E+03, 9.8736E+03,
     + 9.9310E+03, 9.9886E+03, 1.0047E+04, 1.0105E+04, 1.0163E+04,
     +  1.0222E+04, 1.0281E+04, 1.0340E+04, 1.0400E+04, 1.0460E+04,
     + 1.0520E+04, 1.0580E+04, 1.0641E+04, 1.0702E+04, 1.0763E+04,
     +  1.0824E+04, 1.0886E+04, 1.0948E+04, 1.1011E+04, 1.1074E+04,
     + 1.1137E+04, 1.1200E+04, 1.1263E+04, 1.1327E+04, 1.1392E+04,
     +  1.1456E+04, 1.1521E+04, 1.1586E+04, 1.1651E+04, 1.1717E+04,
     + 1.1783E+04, 1.1849E+04, 1.1916E+04, 1.1983E+04, 1.2050E+04,
     +  1.2118E+04, 1.2186E+04, 1.2254E+04, 1.2322E+04, 1.2391E+04,
     + 1.2460E+04, 1.2530E+04, 1.2599E+04, 1.2670E+04, 1.2740E+04,
     +  1.2811E+04, 1.2882E+04, 1.2953E+04, 1.3025E+04, 1.3097E+04,
     + 1.3170E+04, 1.3242E+04, 1.3315E+04, 1.3389E+04, 1.3463E+04,
     +  1.3537E+04, 1.3611E+04, 1.3686E+04, 1.3761E+04, 1.3837E+04,
     + 1.3912E+04, 1.3989E+04, 1.4065E+04, 1.4142E+04, 1.4219E+04,
     +  1.4297E+04, 1.4375E+04, 1.4453E+04, 1.4532E+04, 1.4611E+04,
     + 1.4690E+04, 1.4770E+04, 1.4850E+04, 1.4930E+04, 1.5011E+04,
     +  1.5092E+04, 1.5174E+04, 1.5256E+04, 1.5338E+04, 1.5421E+04,
     + 1.5504E+04, 1.5587E+04, 1.5671E+04, 1.5756E+04, 1.5840E+04,
     +  1.5925E+04, 1.6010E+04, 1.6096E+04, 1.6182E+04, 1.6269E+04,
     + 1.6356E+04, 1.6443E+04, 1.6531E+04, 1.6619E+04, 1.6708E+04,
     +  1.6796E+04, 1.6886E+04, 1.6976E+04, 1.7066E+04, 1.7156E+04,
     + 1.7247E+04, 1.7338E+04, 1.7430E+04, 1.7522E+04, 1.7615E+04,
     +  1.7708E+04, 1.7801E+04, 1.7895E+04, 1.7990E+04, 1.8084E+04,
     + 1.8179E+04, 1.8275E+04, 1.8371E+04, 1.8467E+04, 1.8564E+04,
     +  1.8661E+04, 1.8759E+04, 1.8857E+04, 1.8956E+04, 1.9055E+04,
     + 1.9154E+04, 1.9254E+04, 1.9355E+04, 1.9455E+04, 1.9557E+04,
     +  1.9658E+04, 1.9761E+04, 1.9863E+04, 1.9966E+04, 2.0070E+04,
     + 2.0174E+04, 2.0278E+04, 2.0383E+04, 2.0489E+04, 2.0594E+04,
     +  2.0701E+04, 2.0807E+04, 2.0915E+04, 2.1022E+04, 2.1131E+04,
     + 2.1239E+04, 2.1349E+04, 2.1458E+04, 2.1568E+04, 2.1679E+04,
     +  2.1790E+04, 2.1902E+04, 2.2014E+04, 2.2126E+04, 2.2239E+04,
     + 2.2353E+04, 2.2467E+04, 2.2582E+04, 2.2697E+04, 2.2812E+04,
     +  2.2928E+04, 2.3045E+04, 2.3162E+04, 2.3280E+04, 2.3398E+04,
     + 2.3517E+04, 2.3636E+04, 2.3755E+04, 2.3876E+04, 2.3996E+04,
     +  2.4118E+04, 2.4240E+04, 2.4362E+04, 2.4485E+04, 2.4608E+04,
     + 2.4732E+04, 2.4857E+04, 2.4982E+04, 2.5107E+04, 2.5233E+04,
     +  2.5360E+04, 2.5487E+04, 2.5615E+04, 2.5743E+04, 2.5872E+04,
     + 2.6002E+04, 2.6132E+04, 2.6262E+04, 2.6393E+04, 2.6525E+04,
     +  2.6657E+04, 2.6790E+04, 2.6923E+04, 2.7057E+04, 2.7192E+04,
     + 2.7327E+04, 2.7463E+04, 2.7599E+04, 2.7736E+04, 2.7873E+04,
     +  2.8012E+04, 2.8150E+04, 2.8289E+04, 2.8429E+04, 2.8570E+04,
     + 2.8711E+04, 2.8852E+04, 2.8995E+04, 2.9138E+04, 2.9281E+04,
     +  2.9425E+04, 2.9570E+04, 2.9715E+04, 2.9861E+04, 3.0008E+04,
     + 3.0155E+04, 3.0303E+04, 3.0451E+04, 3.0600E+04, 3.0750E+04,
     +  3.0900E+04, 3.1051E+04, 3.1203E+04, 3.1355E+04, 3.1508E+04,
     + 3.1662E+04, 3.1816E+04, 3.1971E+04, 3.2127E+04, 3.2283E+04,
     +  3.2440E+04, 3.2597E+04, 3.2755E+04, 3.2914E+04, 3.3074E+04,
     + 3.3234E+04, 3.3395E+04, 3.3556E+04, 3.3719E+04, 3.3881E+04,
     +  3.4045E+04, 3.4209E+04, 3.4374E+04, 3.4540E+04, 3.4706E+04,
     + 3.4873E+04, 3.5041E+04, 3.5210E+04, 3.5379E+04, 3.5549E+04,
     +  3.5719E+04, 3.5891E+04, 3.6063E+04, 3.6235E+04, 3.6409E+04,
     + 3.6583E+04, 3.6758E+04, 3.6934E+04, 3.7110E+04, 3.7287E+04,
     +  3.7465E+04, 3.7644E+04, 3.7823E+04, 3.8003E+04, 3.8184E+04,
     + 3.8365E+04, 3.8548E+04, 3.8731E+04, 3.8915E+04, 3.9099E+04,
     +  3.9285E+04, 3.9471E+04, 3.9658E+04, 3.9845E+04, 4.0034E+04,
     + 4.0223E+04, 4.0413E+04, 4.0604E+04, 4.0795E+04, 4.0988E+04,
     +  4.1181E+04, 4.1375E+04, 4.1570E+04, 4.1765E+04, 4.1961E+04,
     + 4.2159E+04, 4.2357E+04, 4.2555E+04, 4.2755E+04, 4.2955E+04,
     +  4.3157E+04, 4.3359E+04, 4.3562E+04, 4.3765E+04, 4.3970E+04,
     + 4.4175E+04, 4.4381E+04, 4.4589E+04, 4.4796E+04, 4.5005E+04,
     +  4.5215E+04, 4.5425E+04, 4.5637E+04, 4.5849E+04, 4.6062E+04,
     + 4.6276E+04, 4.6490E+04, 4.6706E+04, 4.6923E+04, 4.7140E+04,
     +  4.7358E+04, 4.7577E+04, 4.7797E+04, 4.8018E+04, 4.8240E+04,
     + 4.8463E+04, 4.8687E+04, 4.8911E+04, 4.9136E+04, 4.9363E+04,
     +  4.9590E+04, 4.9818E+04, 5.0047E+04, 5.0277E+04, 5.0508E+04,
     + 5.0740E+04, 5.0973E+04, 5.1207E+04, 5.1441E+04, 5.1677E+04,
     +  5.1913E+04, 5.2151E+04, 5.2389E+04, 5.2629E+04, 5.2869E+04,
     + 5.3110E+04, 5.3352E+04, 5.3596E+04, 5.3840E+04, 5.4085E+04,
     +  5.4331E+04, 5.4578E+04, 5.4826E+04, 5.5076E+04, 5.5326E+04,
     + 5.5577E+04, 5.5829E+04, 5.6082E+04, 5.6336E+04, 5.6591E+04,
     +  5.6847E+04, 5.7104E+04, 5.7362E+04, 5.7621E+04, 5.7882E+04,
     + 5.8143E+04, 5.8405E+04, 5.8668E+04, 5.8933E+04, 5.9198E+04,
     +  5.9464E+04, 5.9732E+04, 6.0000E+04, 6.0270E+04, 6.0540E+04,
     + 6.0812E+04, 6.1085E+04, 6.1359E+04, 6.1633E+04, 6.1909E+04,
     +  6.2186E+04, 6.2464E+04, 6.2744E+04, 6.3024E+04, 6.3305E+04,
     + 6.3588E+04, 6.3871E+04, 6.4156E+04, 6.4442E+04, 6.4729E+04,
     +  6.5017E+04, 6.5306E+04, 6.5596E+04, 6.5888E+04, 6.6180E+04,
     + 6.6474E+04, 6.6769E+04, 6.7065E+04, 6.7362E+04, 6.7660E+04,
     +  6.7959E+04, 6.8260E+04, 6.8562E+04, 6.8864E+04, 6.9168E+04,
     + 6.9474E+04, 6.9780E+04, 7.0088E+04, 7.0396E+04, 7.0706E+04,
     +  7.1017E+04, 7.1330E+04, 7.1643E+04, 7.1958E+04, 7.2274E+04,
     + 7.2591E+04, 7.2910E+04, 7.3229E+04, 7.3550E+04, 7.3872E+04,
     +  7.4195E+04, 7.4520E+04, 7.4846E+04, 7.5173E+04, 7.5501E+04,
     + 7.5830E+04, 7.6161E+04, 7.6493E+04, 7.6826E+04, 7.7161E+04,
     +  7.7497E+04, 7.7834E+04, 7.8172E+04, 7.8512E+04, 7.8853E+04,
     + 7.9195E+04, 7.9539E+04, 7.9884E+04, 8.0230E+04, 8.0577E+04,
     +  8.0926E+04, 8.1276E+04, 8.1627E+04, 8.1980E+04, 8.2334E+04,
     + 8.2690E+04, 8.3046E+04, 8.3404E+04, 8.3764E+04, 8.4125E+04,
     +  8.4487E+04, 8.4850E+04, 8.5215E+04, 8.5581E+04, 8.5949E+04,
     + 8.6318E+04, 8.6688E+04, 8.7060E+04, 8.7433E+04, 8.7808E+04,
     +  8.8184E+04, 8.8561E+04, 8.8940E+04, 8.9320E+04, 8.9701E+04,
     + 9.0084E+04, 9.0469E+04, 9.0855E+04, 9.1242E+04, 9.1631E+04,
     +  9.2021E+04, 9.2412E+04, 9.2806E+04, 9.3200E+04, 9.3596E+04,
     + 9.3993E+04, 9.4392E+04, 9.4793E+04, 9.5195E+04, 9.5598E+04,
     +  9.6003E+04, 9.6409E+04, 9.6817E+04, 9.7226E+04, 9.7637E+04,
     + 9.8050E+04, 9.8463E+04, 9.8879E+04, 9.9296E+04, 9.9714E+04,
     +  1.0013E+05, 1.0056E+05, 1.0098E+05, 1.0140E+05, 1.0183E+05,
     + 1.0226E+05, 1.0269E+05, 1.0312E+05, 1.0355E+05, 1.0398E+05,
     +  1.0442E+05, 1.0486E+05, 1.0530E+05, 1.0574E+05, 1.0618E+05,
     + 1.0662E+05, 1.0707E+05, 1.0751E+05, 1.0796E+05, 1.0841E+05,
     +  1.0886E+05, 1.0932E+05, 1.0977E+05, 1.1023E+05, 1.1069E+05,
     + 1.1115E+05, 1.1161E+05, 1.1207E+05, 1.1254E+05, 1.1300E+05,
     +  1.1347E+05, 1.1394E+05, 1.1441E+05, 1.1489E+05, 1.1536E+05,
     + 1.1584E+05, 1.1632E+05, 1.1680E+05, 1.1728E+05, 1.1776E+05,
     +  1.1825E+05, 1.1873E+05, 1.1922E+05, 1.1971E+05, 1.2021E+05,
     + 1.2070E+05, 1.2120E+05, 1.2169E+05, 1.2219E+05, 1.2269E+05,
     +  1.2320E+05, 1.2370E+05, 1.2421E+05, 1.2472E+05, 1.2523E+05,
     + 1.2574E+05, 1.2625E+05, 1.2677E+05, 1.2728E+05, 1.2780E+05,
     +  1.2832E+05, 1.2885E+05, 1.2937E+05, 1.2990E+05, 1.3043E+05,
     + 1.3096E+05, 1.3149E+05, 1.3202E+05, 1.3256E+05, 1.3310E+05,
     +  1.3364E+05, 1.3418E+05, 1.3472E+05, 1.3527E+05, 1.3581E+05,
     + 1.3636E+05, 1.3692E+05, 1.3747E+05, 1.3802E+05, 1.3858E+05,
     +  1.3914E+05, 1.3970E+05, 1.4026E+05, 1.4083E+05, 1.4139E+05,
     + 1.4196E+05, 1.4253E+05, 1.4311E+05, 1.4368E+05, 1.4426E+05,
     +  1.4484E+05, 1.4542E+05, 1.4600E+05, 1.4658E+05, 1.4717E+05,
     + 1.4776E+05, 1.4835E+05, 1.4894E+05, 1.4954E+05, 1.5013E+05,
     +  1.5073E+05, 1.5133E+05, 1.5194E+05, 1.5254E+05, 1.5315E+05,
     + 1.5376E+05, 1.5437E+05, 1.5498E+05, 1.5560E+05, 1.5622E+05,
     +  1.5684E+05, 1.5746E+05, 1.5808E+05, 1.5871E+05, 1.5934E+05,
     + 1.5997E+05, 1.6060E+05, 1.6123E+05, 1.6187E+05, 1.6251E+05,
     +  1.6315E+05, 1.6379E+05, 1.6444E+05, 1.6509E+05, 1.6574E+05,
     + 1.6639E+05, 1.6704E+05, 1.6770E+05, 1.6836E+05, 1.6902E+05,
     +  1.6968E+05, 1.7035E+05, 1.7102E+05, 1.7169E+05, 1.7236E+05,
     + 1.7303E+05, 1.7371E+05, 1.7439E+05, 1.7507E+05, 1.7575E+05,
     +  1.7644E+05, 1.7713E+05, 1.7782E+05, 1.7851E+05, 1.7921E+05,
     + 1.7990E+05, 1.8060E+05, 1.8130E+05, 1.8201E+05, 1.8272E+05,
     +  1.8342E+05, 1.8414E+05, 1.8485E+05, 1.8557E+05, 1.8629E+05,
     + 1.8701E+05, 1.8773E+05, 1.8846E+05, 1.8918E+05, 1.8991E+05,
     +  1.9065E+05, 1.9138E+05, 1.9212E+05, 1.9286E+05, 1.9360E+05,
     + 1.9435E+05, 1.9510E+05, 1.9585E+05, 1.9660E+05, 1.9735E+05,
     +  1.9811E+05, 1.9887E+05, 1.9963E+05, 2.0040E+05, 2.0117E+05,
     + 2.0194E+05, 2.0271E+05, 2.0348E+05, 2.0426E+05, 2.0504E+05,
     +  2.0582E+05, 2.0661E+05, 2.0740E+05, 2.0819E+05, 2.0898E+05,
     + 2.0978E+05, 2.1057E+05, 2.1138E+05, 2.1218E+05, 2.1298E+05,
     +  2.1379E+05, 2.1460E+05, 2.1542E+05, 2.1623E+05, 2.1705E+05,
     + 2.1788E+05, 2.1870E+05, 2.1953E+05, 2.2036E+05, 2.2119E+05,
     +  2.2202E+05, 2.2286E+05, 2.2370E+05, 2.2454E+05, 2.2539E+05,
     + 2.2624E+05, 2.2709E+05, 2.2794E+05, 2.2880E+05, 2.2966E+05,
     +  2.3052E+05, 2.3139E+05, 2.3225E+05, 2.3313E+05, 2.3400E+05,
     + 2.3487E+05, 2.3575E+05, 2.3663E+05, 2.3752E+05, 2.3841E+05,
     +  2.3930E+05, 2.4019E+05, 2.4108E+05, 2.4198E+05, 2.4288E+05,
     + 2.4379E+05, 2.4470E+05, 2.4561E+05, 2.4652E+05, 2.4743E+05,
     +  2.4835E+05, 2.4927E+05, 2.5020E+05, 2.5113E+05, 2.5206E+05,
     + 2.5299E+05, 2.5392E+05, 2.5486E+05, 2.5581E+05, 2.5675E+05,
     +  2.5770E+05, 2.5865E+05, 2.5960E+05, 2.6056E+05, 2.6152E+05,
     + 2.6248E+05, 2.6345E+05, 2.6442E+05, 2.6539E+05, 2.6636E+05,
     +  2.6734E+05, 2.6832E+05, 2.6930E+05, 2.7029E+05, 2.7128E+05,
     + 2.7227E+05, 2.7327E+05, 2.7427E+05, 2.7527E+05, 2.7628E+05,
     +  2.7728E+05, 2.7830E+05, 2.7931E+05, 2.8033E+05, 2.8135E+05,
     + 2.8237E+05, 2.8340E+05, 2.8443E+05, 2.8546E+05, 2.8650E+05,
     +  2.8754E+05, 2.8858E+05, 2.8963E+05, 2.9068E+05, 2.9173E+05,
     + 2.9279E+05, 2.9384E+05, 2.9491E+05, 2.9597E+05, 2.9704E+05,
     +  2.9811E+05, 2.9919E+05, 3.0027E+05, 3.0135E+05, 3.0243E+05,
     + 3.0352E+05, 3.0461E+05, 3.0571E+05, 3.0681E+05, 3.0791E+05,
     +  3.0901E+05, 3.1012E+05, 3.1123E+05, 3.1235E+05, 3.1346E+05,
     + 3.1458E+05, 3.1571E+05, 3.1684E+05, 3.1797E+05, 3.1910E+05,
     +  3.2024E+05, 3.2138E+05, 3.2253E+05, 3.2368E+05, 3.2483E+05,
     + 3.2599E+05, 3.2714E+05, 3.2831E+05, 3.2947E+05, 3.3064E+05,
     +  3.3181E+05, 3.3299E+05, 3.3417E+05, 3.3535E+05, 3.3654E+05,
     + 3.3773E+05, 3.3892E+05, 3.4012E+05, 3.4132E+05, 3.4253E+05,
     +  3.4373E+05, 3.4495E+05, 3.4616E+05, 3.4738E+05, 3.4860E+05,
     + 3.4983E+05, 3.5106E+05, 3.5229E+05, 3.5353E+05, 3.5477E+05,
     +  3.5601E+05, 3.5726E+05, 3.5851E+05, 3.5976E+05, 3.6102E+05,
     + 3.6228E+05, 3.6355E+05, 3.6482E+05, 3.6609E+05, 3.6737E+05,
     +  3.6865E+05, 3.6993E+05, 3.7122E+05, 3.7251E+05, 3.7381E+05,
     + 3.7511E+05, 3.7641E+05, 3.7772E+05, 3.7903E+05, 3.8034E+05,
     +  3.8166E+05, 3.8298E+05, 3.8431E+05, 3.8564E+05, 3.8697E+05,
     + 3.8831E+05, 3.8965E+05, 3.9099E+05, 3.9234E+05, 3.9370E+05,
     +  3.9505E+05, 3.9641E+05, 3.9778E+05, 3.9914E+05, 4.0052E+05,
     + 4.0189E+05, 4.0327E+05, 4.0466E+05, 4.0604E+05, 4.0743E+05,
     +  4.0883E+05, 4.1023E+05, 4.1163E+05, 4.1304E+05, 4.1445E+05,
     + 4.1587E+05, 4.1729E+05, 4.1871E+05, 4.2014E+05, 4.2157E+05,
     +  4.2300E+05, 4.2444E+05, 4.2589E+05, 4.2733E+05, 4.2879E+05,
     + 4.3024E+05, 4.3170E+05, 4.3316E+05, 4.3463E+05, 4.3610E+05,
     +  4.3758E+05, 4.3906E+05, 4.4055E+05, 4.4203E+05, 4.4353E+05,
     + 4.4502E+05, 4.4652E+05, 4.4803E+05, 4.4954E+05, 4.5105E+05,
     +  4.5257E+05, 4.5409E+05, 4.5562E+05, 4.5715E+05, 4.5868E+05,
     + 4.6022E+05, 4.6177E+05, 4.6331E+05, 4.6486E+05, 4.6642E+05,
     +  4.6798E+05, 4.6954E+05, 4.7111E+05, 4.7269E+05, 4.7426E+05,
     + 4.7585E+05, 4.7743E+05, 4.7902E+05, 4.8062E+05, 4.8222E+05,
     +  4.8382E+05, 4.8543E+05, 4.8704E+05, 4.8866E+05, 4.9028E+05,
     + 4.9190E+05, 4.9353E+05, 4.9517E+05, 4.9680E+05, 4.9845E+05,
     +  5.0009E+05, 5.0175E+05, 5.0340E+05, 5.0506E+05, 5.0673E+05,
     + 5.0840E+05, 5.1007E+05, 5.1175E+05, 5.1344E+05, 5.1512E+05,
     +  5.1682E+05, 5.1851E+05, 5.2021E+05, 5.2192E+05, 5.2363E+05,
     + 5.2535E+05, 5.2707E+05, 5.2879E+05, 5.3052E+05, 5.3225E+05,
     +  5.3399E+05, 5.3573E+05, 5.3748E+05, 5.3923E+05, 5.4099E+05,
     + 5.4275E+05, 5.4452E+05, 5.4629E+05, 5.4806E+05, 5.4984E+05,
     +  5.5163E+05, 5.5342E+05, 5.5521E+05, 5.5701E+05, 5.5881E+05,
     + 5.6062E+05, 5.6244E+05, 5.6425E+05, 5.6608E+05, 5.6790E+05,
     +  5.6974E+05, 5.7157E+05, 5.7342E+05, 5.7526E+05, 5.7711E+05,
     + 5.7897E+05, 5.8083E+05, 5.8270E+05, 5.8457E+05, 5.8645E+05,
     +  5.8833E+05, 5.9021E+05, 5.9210E+05, 5.9400E+05, 5.9590E+05,
     + 5.9780E+05, 5.9971E+05, 6.0163E+05, 6.0355E+05, 6.0547E+05,
     +  6.0740E+05, 6.0934E+05, 6.1128E+05, 6.1322E+05, 6.1517E+05,
     + 6.1713E+05, 6.1909E+05, 6.2105E+05, 6.2302E+05, 6.2500E+05,
     +  6.2698E+05, 6.2896E+05, 6.3095E+05, 6.3295E+05, 6.3495E+05,
     + 6.3695E+05, 6.3897E+05, 6.4098E+05, 6.4300E+05, 6.4503E+05,
     +  6.4706E+05, 6.4909E+05, 6.5114E+05, 6.5318E+05, 6.5523E+05,
     + 6.5729E+05, 6.5935E+05, 6.6142E+05, 6.6349E+05, 6.6557E+05,
     +  6.6765E+05, 6.6974E+05, 6.7184E+05, 6.7393E+05, 6.7604E+05,
     + 6.7815E+05, 6.8026E+05, 6.8238E+05, 6.8451E+05, 6.8664E+05,
     +  6.8877E+05, 6.9091E+05, 6.9306E+05, 6.9521E+05, 6.9737E+05,
     + 6.9953E+05, 7.0170E+05, 7.0387E+05, 7.0605E+05, 7.0824E+05,
     +  7.1042E+05, 7.1262E+05, 7.1482E+05, 7.1703E+05, 7.1924E+05,
     + 7.2145E+05, 7.2367E+05, 7.2590E+05, 7.2814E+05, 7.3037E+05,
     +  7.3262E+05, 7.3487E+05, 7.3712E+05, 7.3938E+05, 7.4165E+05,
     + 7.4392E+05, 7.4620E+05, 7.4848E+05, 7.5077E+05, 7.5306E+05,
     +  7.5536E+05, 7.5767E+05, 7.5998E+05, 7.6229E+05, 7.6461E+05,
     + 7.6694E+05, 7.6927E+05, 7.7161E+05, 7.7396E+05, 7.7631E+05,
     +  7.7866E+05, 7.8102E+05, 7.8339E+05, 7.8576E+05, 7.8814E+05,
     + 7.9053E+05, 7.9292E+05, 7.9531E+05, 7.9771E+05, 8.0012E+05,
     +  8.0253E+05, 8.0495E+05, 8.0738E+05, 8.0981E+05, 8.1224E+05,
     + 8.1469E+05, 8.1713E+05, 8.1959E+05, 8.2205E+05, 8.2451E+05,
     +  8.2698E+05, 8.2946E+05, 8.3194E+05, 8.3443E+05, 8.3692E+05,
     + 8.3942E+05, 8.4193E+05, 8.4444E+05, 8.4696E+05, 8.4949E+05,
     +  8.5202E+05, 8.5455E+05, 8.5709E+05, 8.5964E+05, 8.6219E+05,
     + 8.6475E+05, 8.6732E+05, 8.6989E+05, 8.7247E+05, 8.7505E+05,
     +  8.7764E+05, 8.8024E+05, 8.8284E+05, 8.8545E+05, 8.8806E+05,
     + 8.9068E+05, 8.9331E+05, 8.9594E+05, 8.9858E+05, 9.0123E+05,
     +  9.0388E+05, 9.0653E+05, 9.0920E+05, 9.1187E+05, 9.1454E+05,
     + 9.1722E+05, 9.1991E+05, 9.2260E+05, 9.2530E+05, 9.2801E+05,
     +  9.3072E+05, 9.3344E+05, 9.3617E+05, 9.3890E+05, 9.4163E+05,
     + 9.4438E+05, 9.4713E+05, 9.4988E+05, 9.5265E+05, 9.5542E+05,
     +  9.5819E+05, 9.6097E+05, 9.6376E+05, 9.6655E+05, 9.6935E+05,
     + 9.7216E+05, 9.7497E+05, 9.7779E+05, 9.8062E+05, 9.8345E+05,
     +  9.8629E+05, 9.8913E+05, 9.9199E+05, 9.9484E+05, 9.9771E+05,
     + 1.0006E+06, 1.0035E+06, 1.0063E+06, 1.0092E+06, 1.0121E+06,
     +  1.0150E+06, 1.0179E+06, 1.0209E+06, 1.0238E+06, 1.0267E+06,
     + 1.0296E+06, 1.0326E+06, 1.0355E+06, 1.0385E+06, 1.0414E+06,
     +  1.0444E+06, 1.0474E+06, 1.0504E+06, 1.0534E+06, 1.0564E+06,
     + 1.0594E+06, 1.0624E+06, 1.0654E+06, 1.0684E+06, 1.0714E+06,
     +  1.0745E+06, 1.0775E+06, 1.0805E+06, 1.0836E+06, 1.0867E+06,
     + 1.0897E+06, 1.0928E+06, 1.0959E+06, 1.0990E+06, 1.1021E+06,
     +  1.1052E+06, 1.1083E+06, 1.1114E+06, 1.1145E+06, 1.1176E+06,
     + 1.1208E+06, 1.1239E+06, 1.1271E+06, 1.1302E+06, 1.1334E+06,
     +  1.1366E+06, 1.1397E+06, 1.1429E+06, 1.1461E+06, 1.1493E+06,
     + 1.1525E+06, 1.1557E+06, 1.1589E+06, 1.1622E+06, 1.1654E+06,
     +  1.1686E+06, 1.1719E+06, 1.1751E+06, 1.1784E+06, 1.1817E+06,
     + 1.1849E+06, 1.1882E+06, 1.1915E+06, 1.1948E+06, 1.1981E+06,
     +  1.2014E+06, 1.2047E+06, 1.2080E+06, 1.2114E+06, 1.2147E+06,
     + 1.2180E+06, 1.2214E+06, 1.2248E+06, 1.2281E+06, 1.2315E+06,
     +  1.2349E+06, 1.2383E+06, 1.2416E+06, 1.2450E+06, 1.2485E+06,
     + 1.2519E+06, 1.2553E+06, 1.2587E+06, 1.2622E+06, 1.2656E+06,
     +  1.2690E+06, 1.2725E+06, 1.2760E+06, 1.2794E+06, 1.2829E+06,
     + 1.2864E+06, 1.2899E+06, 1.2934E+06, 1.2969E+06, 1.3004E+06,
     +  1.3040E+06, 1.3075E+06, 1.3110E+06, 1.3146E+06, 1.3181E+06,
     + 1.3217E+06, 1.3252E+06, 1.3288E+06, 1.3324E+06, 1.3360E+06,
     +  1.3396E+06, 1.3432E+06, 1.3468E+06, 1.3504E+06, 1.3540E+06,
     + 1.3577E+06, 1.3613E+06, 1.3650E+06, 1.3686E+06, 1.3723E+06,
     +  1.3759E+06, 1.3796E+06, 1.3833E+06, 1.3870E+06, 1.3907E+06,
     + 1.3944E+06, 1.3981E+06, 1.4018E+06, 1.4056E+06, 1.4093E+06,
     +  1.4131E+06, 1.4168E+06, 1.4206E+06, 1.4243E+06, 1.4281E+06,
     + 1.4319E+06, 1.4357E+06, 1.4395E+06, 1.4433E+06, 1.4471E+06,
     +  1.4509E+06, 1.4548E+06, 1.4586E+06, 1.4624E+06, 1.4663E+06,
     + 1.4702E+06, 1.4740E+06, 1.4779E+06, 1.4818E+06, 1.4857E+06,
     +  1.4896E+06, 1.4935E+06, 1.4974E+06, 1.5013E+06, 1.5052E+06,
     + 1.5092E+06, 1.5131E+06, 1.5171E+06, 1.5210E+06, 1.5250E+06,
     +  1.5290E+06, 1.5330E+06, 1.5369E+06, 1.5409E+06, 1.5450E+06,
     + 1.5490E+06, 1.5530E+06, 1.5570E+06, 1.5611E+06, 1.5651E+06,
     +  1.5692E+06, 1.5732E+06, 1.5773E+06, 1.5814E+06, 1.5855E+06,
     + 1.5895E+06, 1.5936E+06, 1.5978E+06, 1.6019E+06, 1.6060E+06,
     +  1.6101E+06, 1.6143E+06, 1.6184E+06, 1.6226E+06, 1.6267E+06,
     + 1.6309E+06, 1.6351E+06, 1.6393E+06, 1.6435E+06, 1.6477E+06,
     +  1.6519E+06, 1.6561E+06, 1.6604E+06, 1.6646E+06, 1.6688E+06,
     + 1.6731E+06, 1.6774E+06, 1.6816E+06, 1.6859E+06, 1.6902E+06,
     +  1.6945E+06, 1.6988E+06, 1.7031E+06, 1.7074E+06, 1.7118E+06,
     + 1.7161E+06, 1.7204E+06, 1.7248E+06, 1.7291E+06, 1.7335E+06,
     +  1.7379E+06, 1.7423E+06, 1.7467E+06, 1.7511E+06, 1.7555E+06,
     + 1.7599E+06, 1.7643E+06, 1.7688E+06, 1.7732E+06, 1.7776E+06,
     +  1.7821E+06, 1.7866E+06, 1.7911E+06, 1.7955E+06, 1.8000E+06,
     + 1.8045E+06, 1.8090E+06, 1.8136E+06, 1.8181E+06, 1.8226E+06,
     +  1.8272E+06, 1.8317E+06, 1.8363E+06, 1.8408E+06, 1.8454E+06,
     + 1.8500E+06, 1.8546E+06, 1.8592E+06, 1.8638E+06, 1.8684E+06,
     +  1.8731E+06, 1.8777E+06, 1.8823E+06, 1.8870E+06, 1.8916E+06,
     + 1.8963E+06, 1.9010E+06, 1.9057E+06, 1.9104E+06, 1.9151E+06,
     +  1.9198E+06, 1.9245E+06, 1.9292E+06, 1.9340E+06, 1.9387E+06,
     + 1.9435E+06, 1.9482E+06, 1.9530E+06, 1.9578E+06, 1.9626E+06,
     +  1.9674E+06, 1.9722E+06, 1.9770E+06, 1.9818E+06, 1.9867E+06,
     + 1.9915E+06, 1.9964E+06, 2.0012E+06, 2.0061E+06, 2.0110E+06,
     +  2.0159E+06, 2.0207E+06, 2.0257E+06, 2.0306E+06, 2.0355E+06,
     + 2.0404E+06, 2.0453E+06, 2.0503E+06, 2.0553E+06, 2.0602E+06,
     +  2.0652E+06, 2.0702E+06, 2.0752E+06, 2.0802E+06, 2.0852E+06,
     + 2.0902E+06, 2.0952E+06, 2.1002E+06, 2.1053E+06, 2.1103E+06,
     +  2.1154E+06, 2.1205E+06, 2.1256E+06, 2.1306E+06, 2.1357E+06,
     + 2.1408E+06, 2.1460E+06, 2.1511E+06, 2.1562E+06, 2.1614E+06,
     +  2.1665E+06, 2.1717E+06, 2.1768E+06, 2.1820E+06, 2.1872E+06,
     + 2.1924E+06, 2.1976E+06, 2.2028E+06, 2.2080E+06, 2.2133E+06,
     +  2.2185E+06, 2.2238E+06, 2.2290E+06, 2.2343E+06, 2.2396E+06,
     + 2.2448E+06, 2.2501E+06, 2.2554E+06, 2.2608E+06, 2.2661E+06,
     +  2.2714E+06, 2.2768E+06, 2.2821E+06, 2.2875E+06, 2.2928E+06,
     + 2.2982E+06, 2.3036E+06, 2.3090E+06, 2.3144E+06, 2.3198E+06,
     +  2.3252E+06, 2.3307E+06, 2.3361E+06, 2.3415E+06, 2.3470E+06,
     + 2.3525E+06, 2.3580E+06, 2.3634E+06, 2.3689E+06, 2.3744E+06,
     +  2.3800E+06, 2.3855E+06, 2.3910E+06, 2.3966E+06, 2.4021E+06,
     + 2.4077E+06, 2.4132E+06, 2.4188E+06, 2.4244E+06, 2.4300E+06,
     +  2.4356E+06, 2.4412E+06, 2.4469E+06, 2.4525E+06, 2.4581E+06,
     + 2.4638E+06, 2.4695E+06, 2.4751E+06, 2.4808E+06, 2.4865E+06,
     +  2.4922E+06, 2.4979E+06, 2.5037E+06, 2.5094E+06, 2.5151E+06,
     + 2.5209E+06, 2.5266E+06, 2.5324E+06, 2.5382E+06, 2.5440E+06,
     +  2.5498E+06, 2.5556E+06, 2.5614E+06, 2.5672E+06, 2.5730E+06,
     + 2.5789E+06, 2.5847E+06, 2.5906E+06, 2.5965E+06, 2.6024E+06,
     +  2.6082E+06, 2.6141E+06, 2.6201E+06, 2.6260E+06, 2.6319E+06,
     + 2.6378E+06, 2.6438E+06, 2.6498E+06, 2.6557E+06, 2.6617E+06,
     +  2.6677E+06, 2.6737E+06, 2.6797E+06, 2.6857E+06, 2.6917E+06,
     + 2.6978E+06, 2.7038E+06, 2.7099E+06, 2.7159E+06, 2.7220E+06,
     +  2.7281E+06, 2.7342E+06, 2.7403E+06, 2.7464E+06, 2.7525E+06,
     + 2.7587E+06, 2.7648E+06, 2.7710E+06, 2.7771E+06, 2.7833E+06,
     +  2.7895E+06, 2.7957E+06, 2.8019E+06, 2.8081E+06, 2.8143E+06,
     + 2.8205E+06, 2.8268E+06, 2.8330E+06, 2.8393E+06, 2.8455E+06,
     +  2.8518E+06, 2.8581E+06, 2.8644E+06, 2.8707E+06, 2.8770E+06,
     + 2.8834E+06, 2.8897E+06, 2.8961E+06, 2.9024E+06, 2.9088E+06,
     +  2.9152E+06, 2.9216E+06, 2.9280E+06, 2.9344E+06, 2.9408E+06,
     + 2.9472E+06, 2.9536E+06, 2.9601E+06, 2.9666E+06, 2.9730E+06,
     +  2.9795E+06, 2.9860E+06, 2.9925E+06, 2.9990E+06, 3.0055E+06,
     + 3.0120E+06, 3.0186E+06, 3.0251E+06, 3.0317E+06, 3.0382E+06,
     +  3.0448E+06, 3.0514E+06, 3.0580E+06, 3.0646E+06, 3.0712E+06,
     + 3.0779E+06, 3.0845E+06, 3.0912E+06, 3.0978E+06, 3.1045E+06,
     +  3.1112E+06, 3.1179E+06, 3.1245E+06, 3.1313E+06, 3.1380E+06,
     + 3.1447E+06, 3.1514E+06, 3.1582E+06, 3.1650E+06, 3.1717E+06,
     +  3.1785E+06, 3.1853E+06, 3.1921E+06, 3.1989E+06, 3.2057E+06,
     + 3.2126E+06, 3.2194E+06, 3.2263E+06, 3.2331E+06, 3.2400E+06,
     +  3.2469E+06, 3.2538E+06, 3.2607E+06, 3.2676E+06, 3.2745E+06,
     + 3.2814E+06, 3.2884E+06, 3.2953E+06, 3.3023E+06, 3.3093E+06,
     +  3.3163E+06, 3.3233E+06, 3.3303E+06, 3.3373E+06, 3.3443E+06,
     + 3.3513E+06, 3.3584E+06, 3.3654E+06, 3.3725E+06, 3.3796E+06,
     +  3.3867E+06, 3.3938E+06, 3.4009E+06, 3.4080E+06, 3.4151E+06,
     + 3.4223E+06, 3.4294E+06, 3.4366E+06, 3.4438E+06, 3.4509E+06,
     +  3.4581E+06, 3.4653E+06, 3.4726E+06, 3.4798E+06, 3.4870E+06,
     + 3.4943E+06, 3.5015E+06, 3.5088E+06, 3.5161E+06, 3.5233E+06,
     +  3.5306E+06, 3.5379E+06, 3.5453E+06, 3.5526E+06, 3.5599E+06,
     + 3.5673E+06, 3.5746E+06, 3.5820E+06, 3.5894E+06, 3.5968E+06,
     +  3.6042E+06, 3.6116E+06, 3.6190E+06, 3.6265E+06, 3.6339E+06,
     + 3.6414E+06, 3.6488E+06, 3.6563E+06, 3.6638E+06, 3.6713E+06,
     +  3.6788E+06, 3.6863E+06, 3.6938E+06, 3.7014E+06, 3.7089E+06,
     + 3.7165E+06, 3.7241E+06, 3.7316E+06, 3.7392E+06, 3.7468E+06,
     +  3.7545E+06, 3.7621E+06, 3.7697E+06, 3.7774E+06, 3.7850E+06,
     + 3.7927E+06, 3.8004E+06, 3.8081E+06, 3.8158E+06, 3.8235E+06,
     +  3.8312E+06, 3.8389E+06, 3.8467E+06, 3.8544E+06, 3.8622E+06,
     + 3.8700E+06, 3.8777E+06, 3.8855E+06, 3.8933E+06, 3.9012E+06,
     +  3.9090E+06, 3.9168E+06, 3.9247E+06, 3.9325E+06, 3.9404E+06,
     + 3.9483E+06, 3.9562E+06, 3.9641E+06, 3.9720E+06, 3.9799E+06,
     +  3.9879E+06, 3.9958E+06, 4.0038E+06, 4.0118E+06, 4.0197E+06,
     + 4.0277E+06, 4.0357E+06, 4.0437E+06, 4.0518E+06, 4.0598E+06,
     +  4.0678E+06, 4.0759E+06, 4.0840E+06, 4.0920E+06, 4.1001E+06,
     + 4.1082E+06, 4.1164E+06, 4.1245E+06, 4.1326E+06, 4.1407E+06,
     +  4.1489E+06, 4.1571E+06, 4.1652E+06, 4.1734E+06, 4.1816E+06,
     + 4.1898E+06, 4.1981E+06, 4.2063E+06, 4.2145E+06, 4.2228E+06,
     +  4.2311E+06, 4.2393E+06, 4.2476E+06, 4.2559E+06, 4.2642E+06,
     + 4.2725E+06, 4.2809E+06, 4.2892E+06, 4.2976E+06, 4.3059E+06,
     +  4.3143E+06, 4.3227E+06, 4.3311E+06, 4.3395E+06, 4.3479E+06,
     + 4.3563E+06, 4.3648E+06, 4.3732E+06, 4.3817E+06, 4.3902E+06,
     +  4.3987E+06, 4.4072E+06, 4.4157E+06, 4.4242E+06, 4.4327E+06,
     + 4.4413E+06, 4.4498E+06, 4.4584E+06, 4.4670E+06, 4.4755E+06,
     +  4.4841E+06, 4.4927E+06, 4.5014E+06, 4.5100E+06, 4.5186E+06,
     + 4.5273E+06, 4.5359E+06, 4.5446E+06, 4.5533E+06, 4.5620E+06,
     +  4.5707E+06, 4.5794E+06, 4.5882E+06, 4.5969E+06, 4.6057E+06,
     + 4.6144E+06, 4.6232E+06, 4.6320E+06, 4.6408E+06, 4.6496E+06,
     +  4.6584E+06, 4.6672E+06, 4.6761E+06, 4.6849E+06, 4.6938E+06,
     + 4.7027E+06, 4.7116E+06, 4.7205E+06, 4.7294E+06, 4.7383E+06,
     +  4.7472E+06, 4.7562E+06, 4.7651E+06, 4.7741E+06, 4.7831E+06,
     + 4.7921E+06, 4.8011E+06, 4.8101E+06, 4.8191E+06, 4.8282E+06,
     +  4.8372E+06, 4.8463E+06, 4.8553E+06, 4.8644E+06, 4.8735E+06,
     + 4.8826E+06, 4.8917E+06, 4.9008E+06, 4.9100E+06, 4.9191E+06,
     +  4.9283E+06, 4.9375E+06, 4.9466E+06, 4.9558E+06, 4.9650E+06,
     + 4.9743E+06, 4.9835E+06, 4.9927E+06, 5.0020E+06, 5.0112E+06,
     +  5.0205E+06, 5.0298E+06, 5.0391E+06, 5.0484E+06, 5.0577E+06,
     + 5.0671E+06, 5.0764E+06, 5.0858E+06, 5.0951E+06, 5.1045E+06,
     +  5.1139E+06, 5.1233E+06, 5.1327E+06, 5.1421E+06, 5.1516E+06,
     + 5.1610E+06, 5.1705E+06, 5.1799E+06, 5.1894E+06, 5.1989E+06,
     +  5.2084E+06, 5.2179E+06, 5.2275E+06, 5.2370E+06, 5.2465E+06,
     + 5.2561E+06, 5.2657E+06, 5.2753E+06, 5.2849E+06, 5.2945E+06,
     +  5.3041E+06, 5.3137E+06, 5.3234E+06, 5.3330E+06, 5.3427E+06,
     + 5.3524E+06, 5.3620E+06, 5.3717E+06, 5.3815E+06, 5.3912E+06,
     +  5.4009E+06, 5.4107E+06, 5.4204E+06, 5.4302E+06, 5.4400E+06,
     + 5.4498E+06, 5.4596E+06, 5.4694E+06, 5.4792E+06, 5.4891E+06,
     +  5.4989E+06, 5.5088E+06, 5.5186E+06, 5.5285E+06, 5.5384E+06,
     + 5.5483E+06, 5.5583E+06, 5.5682E+06, 5.5781E+06, 5.5881E+06,
     +  5.5981E+06, 5.6080E+06, 5.6180E+06, 5.6280E+06, 5.6380E+06,
     + 5.6481E+06, 5.6581E+06, 5.6682E+06, 5.6782E+06, 5.6883E+06,
     +  5.6984E+06, 5.7085E+06, 5.7186E+06, 5.7287E+06, 5.7388E+06,
     + 5.7490E+06, 5.7591E+06, 5.7693E+06, 5.7795E+06, 5.7897E+06,
     +  5.7999E+06, 5.8101E+06, 5.8203E+06, 5.8305E+06, 5.8408E+06,
     + 5.8510E+06, 5.8613E+06, 5.8716E+06, 5.8819E+06, 5.8922E+06,
     +  5.9025E+06, 5.9129E+06, 5.9232E+06, 5.9336E+06, 5.9439E+06,
     + 5.9543E+06, 5.9647E+06, 5.9751E+06, 5.9855E+06, 5.9959E+06,
     +  6.0064E+06, 6.0168E+06, 6.0273E+06, 6.0378E+06, 6.0482E+06,
     + 6.0587E+06, 6.0692E+06, 6.0798E+06, 6.0903E+06, 6.1008E+06,
     +  6.1114E+06, 6.1220E+06, 6.1325E+06, 6.1431E+06, 6.1537E+06,
     + 6.1643E+06, 6.1750E+06, 6.1856E+06, 6.1963E+06, 6.2069E+06,
     +  6.2176E+06, 6.2283E+06, 6.2390E+06, 6.2497E+06, 6.2604E+06,
     + 6.2712E+06, 6.2819E+06, 6.2927E+06, 6.3034E+06, 6.3142E+06,
     +  6.3250E+06, 6.3358E+06, 6.3466E+06, 6.3575E+06, 6.3683E+06,
     + 6.3791E+06, 6.3900E+06, 6.4009E+06, 6.4118E+06, 6.4227E+06,
     +  6.4336E+06, 6.4445E+06, 6.4555E+06, 6.4664E+06, 6.4774E+06,
     + 6.4883E+06, 6.4993E+06, 6.5103E+06, 6.5213E+06, 6.5323E+06,
     +  6.5434E+06, 6.5544E+06, 6.5655E+06, 6.5765E+06, 6.5876E+06,
     + 6.5987E+06, 6.6098E+06, 6.6209E+06, 6.6321E+06, 6.6432E+06,
     +  6.6544E+06, 6.6655E+06, 6.6767E+06, 6.6879E+06, 6.6991E+06,
     + 6.7103E+06, 6.7215E+06, 6.7328E+06, 6.7440E+06, 6.7553E+06,
     +  6.7665E+06, 6.7778E+06, 6.7891E+06, 6.8004E+06, 6.8118E+06,
     + 6.8231E+06, 6.8344E+06, 6.8458E+06, 6.8572E+06, 6.8685E+06,
     +  6.8799E+06, 6.8913E+06, 6.9027E+06, 6.9142E+06, 6.9256E+06/
      data (QofT( 1,J),J=2409,4510)/ 6.9371E+06, 6.9485E+06,
     +  6.9600E+06, 6.9715E+06, 6.9830E+06, 6.9945E+06, 7.0060E+06,
     +  7.0176E+06,
     + 7.0291E+06, 7.0407E+06, 7.0523E+06, 7.0639E+06, 7.0755E+06,
     +  7.0871E+06, 7.0987E+06, 7.1103E+06, 7.1220E+06, 7.1336E+06,
     + 7.1453E+06, 7.1570E+06, 7.1687E+06, 7.1804E+06, 7.1921E+06,
     +  7.2038E+06, 7.2156E+06, 7.2273E+06, 7.2391E+06, 7.2509E+06,
     + 7.2627E+06, 7.2745E+06, 7.2863E+06, 7.2981E+06, 7.3100E+06,
     +  7.3218E+06, 7.3337E+06, 7.3456E+06, 7.3575E+06, 7.3694E+06,
     + 7.3813E+06, 7.3932E+06, 7.4051E+06, 7.4171E+06, 7.4290E+06,
     +  7.4410E+06, 7.4530E+06, 7.4650E+06, 7.4770E+06, 7.4890E+06,
     + 7.5011E+06, 7.5131E+06, 7.5252E+06, 7.5373E+06, 7.5493E+06,
     +  7.5614E+06, 7.5735E+06, 7.5857E+06, 7.5978E+06, 7.6099E+06,
     + 7.6221E+06, 7.6343E+06, 7.6465E+06, 7.6586E+06, 7.6709E+06,
     +  7.6831E+06, 7.6953E+06, 7.7075E+06, 7.7198E+06, 7.7321E+06,
     + 7.7444E+06, 7.7566E+06, 7.7689E+06, 7.7813E+06, 7.7936E+06,
     +  7.8059E+06, 7.8183E+06, 7.8307E+06, 7.8430E+06, 7.8554E+06,
     + 7.8678E+06, 7.8802E+06, 7.8927E+06, 7.9051E+06, 7.9176E+06,
     +  7.9300E+06, 7.9425E+06, 7.9550E+06, 7.9675E+06, 7.9800E+06,
     + 7.9925E+06, 8.0051E+06, 8.0176E+06, 8.0302E+06, 8.0427E+06,
     +  8.0553E+06, 8.0679E+06, 8.0805E+06, 8.0932E+06, 8.1058E+06,
     + 8.1184E+06, 8.1311E+06, 8.1438E+06, 8.1565E+06, 8.1692E+06,
     +  8.1819E+06, 8.1946E+06, 8.2073E+06, 8.2201E+06, 8.2328E+06,
     + 8.2456E+06, 8.2584E+06, 8.2712E+06, 8.2840E+06, 8.2968E+06,
     +  8.3096E+06, 8.3225E+06, 8.3353E+06, 8.3482E+06, 8.3611E+06,
     + 8.3740E+06, 8.3869E+06, 8.3998E+06, 8.4127E+06, 8.4257E+06,
     +  8.4386E+06, 8.4516E+06, 8.4646E+06, 8.4776E+06, 8.4906E+06,
     + 8.5036E+06, 8.5166E+06, 8.5297E+06, 8.5427E+06, 8.5558E+06,
     +  8.5689E+06, 8.5820E+06, 8.5951E+06, 8.6082E+06, 8.6213E+06,
     + 8.6344E+06, 8.6476E+06, 8.6608E+06, 8.6739E+06, 8.6871E+06,
     +  8.7003E+06, 8.7135E+06, 8.7268E+06, 8.7400E+06, 8.7533E+06,
     + 8.7665E+06, 8.7798E+06, 8.7931E+06, 8.8064E+06, 8.8197E+06,
     +  8.8330E+06, 8.8464E+06, 8.8597E+06, 8.8731E+06, 8.8865E+06,
     + 8.8999E+06, 8.9133E+06, 8.9267E+06, 8.9401E+06, 8.9535E+06,
     +  8.9670E+06, 8.9804E+06, 8.9939E+06, 9.0074E+06, 9.0209E+06,
     + 9.0344E+06, 9.0479E+06, 9.0615E+06, 9.0750E+06, 9.0886E+06,
     +  9.1022E+06, 9.1157E+06, 9.1293E+06, 9.1430E+06, 9.1566E+06,
     + 9.1702E+06, 9.1839E+06, 9.1975E+06, 9.2112E+06, 9.2249E+06,
     +  9.2386E+06, 9.2523E+06, 9.2660E+06, 9.2797E+06, 9.2935E+06,
     + 9.3072E+06, 9.3210E+06, 9.3348E+06, 9.3486E+06, 9.3624E+06,
     +  9.3762E+06, 9.3901E+06, 9.4039E+06, 9.4178E+06, 9.4316E+06,
     + 9.4455E+06, 9.4594E+06, 9.4733E+06, 9.4873E+06, 9.5012E+06,
     +  9.5151E+06, 9.5291E+06, 9.5431E+06, 9.5570E+06, 9.5710E+06,
     + 9.5850E+06, 9.5991E+06, 9.6131E+06, 9.6271E+06, 9.6412E+06,
     +  9.6553E+06, 9.6694E+06, 9.6835E+06, 9.6976E+06, 9.7117E+06,
     + 9.7258E+06, 9.7400E+06, 9.7541E+06, 9.7683E+06, 9.7825E+06,
     +  9.7967E+06, 9.8109E+06, 9.8251E+06, 9.8393E+06, 9.8536E+06,
     + 9.8678E+06, 9.8821E+06, 9.8964E+06, 9.9107E+06, 9.9250E+06,
     +  9.9393E+06, 9.9536E+06, 9.9680E+06, 9.9823E+06, 9.9967E+06,
     + 1.0011E+07, 1.0025E+07, 1.0040E+07, 1.0054E+07, 1.0069E+07,
     +  1.0083E+07, 1.0098E+07, 1.0112E+07, 1.0127E+07, 1.0141E+07,
     + 1.0156E+07, 1.0170E+07, 1.0185E+07, 1.0199E+07, 1.0214E+07,
     +  1.0228E+07, 1.0243E+07, 1.0257E+07, 1.0272E+07, 1.0287E+07,
     + 1.0301E+07, 1.0316E+07, 1.0331E+07, 1.0345E+07, 1.0360E+07,
     +  1.0375E+07, 1.0389E+07, 1.0404E+07, 1.0419E+07, 1.0434E+07,
     + 1.0448E+07, 1.0463E+07, 1.0478E+07, 1.0493E+07, 1.0507E+07,
     +  1.0522E+07, 1.0537E+07, 1.0552E+07, 1.0567E+07, 1.0582E+07,
     + 1.0597E+07, 1.0611E+07, 1.0626E+07, 1.0641E+07, 1.0656E+07,
     +  1.0671E+07, 1.0686E+07, 1.0701E+07, 1.0716E+07, 1.0731E+07,
     + 1.0746E+07, 1.0761E+07, 1.0776E+07, 1.0791E+07, 1.0806E+07,
     +  1.0821E+07, 1.0836E+07, 1.0851E+07, 1.0867E+07, 1.0882E+07,
     + 1.0897E+07, 1.0912E+07, 1.0927E+07, 1.0942E+07, 1.0957E+07,
     +  1.0973E+07, 1.0988E+07, 1.1003E+07, 1.1018E+07, 1.1033E+07,
     + 1.1049E+07, 1.1064E+07, 1.1079E+07, 1.1095E+07, 1.1110E+07,
     +  1.1125E+07, 1.1141E+07, 1.1156E+07, 1.1171E+07, 1.1187E+07,
     + 1.1202E+07, 1.1217E+07, 1.1233E+07, 1.1248E+07, 1.1264E+07,
     +  1.1279E+07, 1.1294E+07, 1.1310E+07, 1.1325E+07, 1.1341E+07,
     + 1.1356E+07, 1.1372E+07, 1.1387E+07, 1.1403E+07, 1.1419E+07,
     +  1.1434E+07, 1.1450E+07, 1.1465E+07, 1.1481E+07, 1.1497E+07,
     + 1.1512E+07, 1.1528E+07, 1.1543E+07, 1.1559E+07, 1.1575E+07,
     +  1.1590E+07, 1.1606E+07, 1.1622E+07, 1.1638E+07, 1.1653E+07,
     + 1.1669E+07, 1.1685E+07, 1.1701E+07, 1.1716E+07, 1.1732E+07,
     +  1.1748E+07, 1.1764E+07, 1.1780E+07, 1.1796E+07, 1.1811E+07,
     + 1.1827E+07, 1.1843E+07, 1.1859E+07, 1.1875E+07, 1.1891E+07,
     +  1.1907E+07, 1.1923E+07, 1.1939E+07, 1.1955E+07, 1.1971E+07,
     + 1.1987E+07, 1.2003E+07, 1.2019E+07, 1.2035E+07, 1.2051E+07,
     +  1.2067E+07, 1.2083E+07, 1.2099E+07, 1.2115E+07, 1.2131E+07,
     + 1.2147E+07, 1.2164E+07, 1.2180E+07, 1.2196E+07, 1.2212E+07,
     +  1.2228E+07, 1.2245E+07, 1.2261E+07, 1.2277E+07, 1.2293E+07,
     + 1.2309E+07, 1.2326E+07, 1.2342E+07, 1.2358E+07, 1.2375E+07,
     +  1.2391E+07, 1.2407E+07, 1.2424E+07, 1.2440E+07, 1.2456E+07,
     + 1.2473E+07, 1.2489E+07, 1.2505E+07, 1.2522E+07, 1.2538E+07,
     +  1.2555E+07, 1.2571E+07, 1.2588E+07, 1.2604E+07, 1.2621E+07,
     + 1.2637E+07, 1.2654E+07, 1.2670E+07, 1.2687E+07, 1.2703E+07,
     +  1.2720E+07, 1.2736E+07, 1.2753E+07, 1.2770E+07, 1.2786E+07,
     + 1.2803E+07, 1.2819E+07, 1.2836E+07, 1.2853E+07, 1.2869E+07,
     +  1.2886E+07, 1.2903E+07, 1.2919E+07, 1.2936E+07, 1.2953E+07,
     + 1.2970E+07, 1.2986E+07, 1.3003E+07, 1.3020E+07, 1.3037E+07,
     +  1.3054E+07, 1.3070E+07, 1.3087E+07, 1.3104E+07, 1.3121E+07,
     + 1.3138E+07, 1.3155E+07, 1.3172E+07, 1.3189E+07, 1.3205E+07,
     +  1.3222E+07, 1.3239E+07, 1.3256E+07, 1.3273E+07, 1.3290E+07,
     + 1.3307E+07, 1.3324E+07, 1.3341E+07, 1.3358E+07, 1.3375E+07,
     +  1.3392E+07, 1.3410E+07, 1.3427E+07, 1.3444E+07, 1.3461E+07,
     + 1.3478E+07, 1.3495E+07, 1.3512E+07, 1.3529E+07, 1.3547E+07,
     +  1.3564E+07, 1.3581E+07, 1.3598E+07, 1.3615E+07, 1.3633E+07,
     + 1.3650E+07, 1.3667E+07, 1.3684E+07, 1.3702E+07, 1.3719E+07,
     +  1.3736E+07, 1.3754E+07, 1.3771E+07, 1.3788E+07, 1.3806E+07,
     + 1.3823E+07, 1.3840E+07, 1.3858E+07, 1.3875E+07, 1.3893E+07,
     +  1.3910E+07, 1.3927E+07, 1.3945E+07, 1.3962E+07, 1.3980E+07,
     + 1.3997E+07, 1.4015E+07, 1.4032E+07, 1.4050E+07, 1.4067E+07,
     +  1.4085E+07, 1.4103E+07, 1.4120E+07, 1.4138E+07, 1.4155E+07,
     + 1.4173E+07, 1.4191E+07, 1.4208E+07, 1.4226E+07, 1.4244E+07,
     +  1.4261E+07, 1.4279E+07, 1.4297E+07, 1.4314E+07, 1.4332E+07,
     + 1.4350E+07, 1.4368E+07, 1.4385E+07, 1.4403E+07, 1.4421E+07,
     +  1.4439E+07, 1.4457E+07, 1.4474E+07, 1.4492E+07, 1.4510E+07,
     + 1.4528E+07, 1.4546E+07, 1.4564E+07, 1.4582E+07, 1.4599E+07,
     +  1.4617E+07, 1.4635E+07, 1.4653E+07, 1.4671E+07, 1.4689E+07,
     + 1.4707E+07, 1.4725E+07, 1.4743E+07, 1.4761E+07, 1.4779E+07,
     +  1.4797E+07, 1.4815E+07, 1.4833E+07, 1.4852E+07, 1.4870E+07,
     + 1.4888E+07, 1.4906E+07, 1.4924E+07, 1.4942E+07, 1.4960E+07,
     +  1.4978E+07, 1.4997E+07, 1.5015E+07, 1.5033E+07, 1.5051E+07,
     + 1.5070E+07, 1.5088E+07, 1.5106E+07, 1.5124E+07, 1.5143E+07,
     +  1.5161E+07, 1.5179E+07, 1.5197E+07, 1.5216E+07, 1.5234E+07,
     + 1.5253E+07, 1.5271E+07, 1.5289E+07, 1.5308E+07, 1.5326E+07,
     +  1.5344E+07, 1.5363E+07, 1.5381E+07, 1.5400E+07, 1.5418E+07,
     + 1.5437E+07, 1.5455E+07, 1.5474E+07, 1.5492E+07, 1.5511E+07,
     +  1.5529E+07, 1.5548E+07, 1.5566E+07, 1.5585E+07, 1.5604E+07,
     + 1.5622E+07, 1.5641E+07, 1.5659E+07, 1.5678E+07, 1.5697E+07,
     +  1.5715E+07, 1.5734E+07, 1.5753E+07, 1.5771E+07, 1.5790E+07,
     + 1.5809E+07, 1.5828E+07, 1.5846E+07, 1.5865E+07, 1.5884E+07,
     +  1.5903E+07, 1.5921E+07, 1.5940E+07, 1.5959E+07, 1.5978E+07,
     + 1.5997E+07, 1.6016E+07, 1.6035E+07, 1.6053E+07, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --      5111
      data (QofT( 2,J),J=1,2408)/ 2.6894E+00, 4.6386E+00, 5.5732E+00,
     +  6.1741E+00, 6.7201E+00, 7.3357E+00, 8.0630E+00, 8.9034E+00,
     + 9.8431E+00, 1.0865E+01, 1.1955E+01, 1.3100E+01, 1.4294E+01,
     +  1.5530E+01, 1.6806E+01, 1.8118E+01, 1.9466E+01, 2.0848E+01,
     + 2.2266E+01, 2.3717E+01, 2.6724E+01, 2.9868E+01, 3.3149E+01,
     +  3.6564E+01, 4.0112E+01, 4.3790E+01, 4.7594E+01, 5.1522E+01,
     + 5.5568E+01, 5.9731E+01, 6.4006E+01, 6.8390E+01, 7.2880E+01,
     +  7.7472E+01, 8.2165E+01, 8.6956E+01, 9.1841E+01, 9.6820E+01,
     + 1.0189E+02, 1.0705E+02, 1.1229E+02, 1.1762E+02, 1.2304E+02,
     +  1.2854E+02, 1.3411E+02, 1.3977E+02, 1.4551E+02, 1.5132E+02,
     + 1.5721E+02, 1.6318E+02, 1.6922E+02, 1.7533E+02, 1.8152E+02,
     +  1.8777E+02, 1.9410E+02, 2.0050E+02, 2.0697E+02, 2.1350E+02,
     + 2.2011E+02, 2.2678E+02, 2.3351E+02, 2.4032E+02, 2.4719E+02,
     +  2.5412E+02, 2.6112E+02, 2.6818E+02, 2.7530E+02, 2.8248E+02,
     + 2.8973E+02, 2.9704E+02, 3.0441E+02, 3.1184E+02, 3.1933E+02,
     +  3.2688E+02, 3.3449E+02, 3.4216E+02, 3.4988E+02, 3.5767E+02,
     + 3.6551E+02, 3.7341E+02, 3.8136E+02, 3.8937E+02, 3.9744E+02,
     +  4.0557E+02, 4.1375E+02, 4.2198E+02, 4.3027E+02, 4.3862E+02,
     + 4.4702E+02, 4.5547E+02, 4.6398E+02, 4.7254E+02, 4.8115E+02,
     +  4.8982E+02, 4.9854E+02, 5.0732E+02, 5.1615E+02, 5.2503E+02,
     + 5.3396E+02, 5.4295E+02, 5.5199E+02, 5.6108E+02, 5.7022E+02,
     +  5.7941E+02, 5.8866E+02, 5.9796E+02, 6.0731E+02, 6.1671E+02,
     + 6.2616E+02, 6.3567E+02, 6.4522E+02, 6.5483E+02, 6.6449E+02,
     +  6.7420E+02, 6.8396E+02, 6.9377E+02, 7.0364E+02, 7.1355E+02,
     + 7.2352E+02, 7.3354E+02, 7.4361E+02, 7.5373E+02, 7.6391E+02,
     +  7.7413E+02, 7.8441E+02, 7.9474E+02, 8.0512E+02, 8.1555E+02,
     + 8.2603E+02, 8.3657E+02, 8.4716E+02, 8.5780E+02, 8.6849E+02,
     +  8.7924E+02, 8.9004E+02, 9.0089E+02, 9.1179E+02, 9.2275E+02,
     + 9.3376E+02, 9.4483E+02, 9.5594E+02, 9.6711E+02, 9.7834E+02,
     +  9.8962E+02, 1.0010E+03, 1.0123E+03, 1.0238E+03, 1.0353E+03,
     + 1.0468E+03, 1.0584E+03, 1.0701E+03, 1.0818E+03, 1.0936E+03,
     +  1.1054E+03, 1.1173E+03, 1.1292E+03, 1.1412E+03, 1.1533E+03,
     + 1.1654E+03, 1.1776E+03, 1.1898E+03, 1.2021E+03, 1.2144E+03,
     +  1.2268E+03, 1.2393E+03, 1.2518E+03, 1.2644E+03, 1.2770E+03,
     + 1.2897E+03, 1.3025E+03, 1.3153E+03, 1.3282E+03, 1.3411E+03,
     +  1.3541E+03, 1.3672E+03, 1.3803E+03, 1.3935E+03, 1.4068E+03,
     + 1.4201E+03, 1.4335E+03, 1.4469E+03, 1.4604E+03, 1.4740E+03,
     +  1.4876E+03, 1.5013E+03, 1.5150E+03, 1.5289E+03, 1.5427E+03,
     + 1.5567E+03, 1.5707E+03, 1.5848E+03, 1.5990E+03, 1.6132E+03,
     +  1.6275E+03, 1.6418E+03, 1.6562E+03, 1.6707E+03, 1.6853E+03,
     + 1.6999E+03, 1.7146E+03, 1.7294E+03, 1.7442E+03, 1.7592E+03,
     +  1.7741E+03, 1.7892E+03, 1.8043E+03, 1.8195E+03, 1.8348E+03,
     + 1.8501E+03, 1.8655E+03, 1.8810E+03, 1.8966E+03, 1.9122E+03,
     +  1.9279E+03, 1.9437E+03, 1.9596E+03, 1.9755E+03, 1.9915E+03,
     + 2.0076E+03, 2.0238E+03, 2.0400E+03, 2.0563E+03, 2.0727E+03,
     +  2.0892E+03, 2.1058E+03, 2.1224E+03, 2.1391E+03, 2.1559E+03,
     + 2.1728E+03, 2.1897E+03, 2.2068E+03, 2.2239E+03, 2.2411E+03,
     +  2.2584E+03, 2.2757E+03, 2.2932E+03, 2.3107E+03, 2.3283E+03,
     + 2.3460E+03, 2.3638E+03, 2.3816E+03, 2.3996E+03, 2.4176E+03,
     +  2.4358E+03, 2.4540E+03, 2.4723E+03, 2.4906E+03, 2.5091E+03,
     + 2.5277E+03, 2.5463E+03, 2.5651E+03, 2.5839E+03, 2.6028E+03,
     +  2.6218E+03, 2.6409E+03, 2.6601E+03, 2.6794E+03, 2.6987E+03,
     + 2.7182E+03, 2.7378E+03, 2.7574E+03, 2.7771E+03, 2.7970E+03,
     +  2.8169E+03, 2.8369E+03, 2.8571E+03, 2.8773E+03, 2.8976E+03,
     + 2.9180E+03, 2.9385E+03, 2.9591E+03, 2.9798E+03, 3.0006E+03,
     +  3.0215E+03, 3.0425E+03, 3.0636E+03, 3.0848E+03, 3.1061E+03,
     + 3.1275E+03, 3.1490E+03, 3.1706E+03, 3.1923E+03, 3.2141E+03,
     +  3.2360E+03, 3.2580E+03, 3.2801E+03, 3.3023E+03, 3.3246E+03,
     + 3.3471E+03, 3.3696E+03, 3.3922E+03, 3.4150E+03, 3.4378E+03,
     +  3.4608E+03, 3.4838E+03, 3.5070E+03, 3.5303E+03, 3.5537E+03,
     + 3.5772E+03, 3.6008E+03, 3.6245E+03, 3.6483E+03, 3.6723E+03,
     +  3.6963E+03, 3.7205E+03, 3.7448E+03, 3.7691E+03, 3.7936E+03,
     + 3.8183E+03, 3.8430E+03, 3.8678E+03, 3.8928E+03, 3.9179E+03,
     +  3.9431E+03, 3.9684E+03, 3.9938E+03, 4.0193E+03, 4.0450E+03,
     + 4.0708E+03, 4.0967E+03, 4.1227E+03, 4.1488E+03, 4.1751E+03,
     +  4.2014E+03, 4.2279E+03, 4.2546E+03, 4.2813E+03, 4.3082E+03,
     + 4.3351E+03, 4.3623E+03, 4.3895E+03, 4.4168E+03, 4.4443E+03,
     +  4.4719E+03, 4.4997E+03, 4.5275E+03, 4.5555E+03, 4.5836E+03,
     + 4.6119E+03, 4.6402E+03, 4.6687E+03, 4.6974E+03, 4.7261E+03,
     +  4.7550E+03, 4.7840E+03, 4.8132E+03, 4.8424E+03, 4.8718E+03,
     + 4.9014E+03, 4.9311E+03, 4.9609E+03, 4.9908E+03, 5.0209E+03,
     +  5.0511E+03, 5.0814E+03, 5.1119E+03, 5.1425E+03, 5.1733E+03,
     + 5.2042E+03, 5.2352E+03, 5.2664E+03, 5.2977E+03, 5.3291E+03,
     +  5.3607E+03, 5.3924E+03, 5.4243E+03, 5.4563E+03, 5.4885E+03,
     + 5.5208E+03, 5.5532E+03, 5.5858E+03, 5.6185E+03, 5.6514E+03,
     +  5.6844E+03, 5.7175E+03, 5.7508E+03, 5.7843E+03, 5.8179E+03,
     + 5.8516E+03, 5.8855E+03, 5.9195E+03, 5.9537E+03, 5.9881E+03,
     +  6.0226E+03, 6.0572E+03, 6.0920E+03, 6.1269E+03, 6.1620E+03,
     + 6.1973E+03, 6.2327E+03, 6.2682E+03, 6.3039E+03, 6.3398E+03,
     +  6.3758E+03, 6.4120E+03, 6.4483E+03, 6.4848E+03, 6.5214E+03,
     + 6.5582E+03, 6.5952E+03, 6.6323E+03, 6.6696E+03, 6.7070E+03,
     +  6.7446E+03, 6.7824E+03, 6.8203E+03, 6.8584E+03, 6.8966E+03,
     + 6.9350E+03, 6.9736E+03, 7.0123E+03, 7.0512E+03, 7.0903E+03,
     +  7.1295E+03, 7.1689E+03, 7.2085E+03, 7.2482E+03, 7.2881E+03,
     + 7.3282E+03, 7.3684E+03, 7.4089E+03, 7.4494E+03, 7.4902E+03,
     +  7.5311E+03, 7.5722E+03, 7.6135E+03, 7.6549E+03, 7.6966E+03,
     + 7.7384E+03, 7.7803E+03, 7.8225E+03, 7.8648E+03, 7.9073E+03,
     +  7.9500E+03, 7.9928E+03, 8.0359E+03, 8.0791E+03, 8.1225E+03,
     + 8.1660E+03, 8.2098E+03, 8.2537E+03, 8.2978E+03, 8.3422E+03,
     +  8.3866E+03, 8.4313E+03, 8.4762E+03, 8.5212E+03, 8.5664E+03,
     + 8.6118E+03, 8.6574E+03, 8.7032E+03, 8.7492E+03, 8.7954E+03,
     +  8.8417E+03, 8.8883E+03, 8.9350E+03, 8.9819E+03, 9.0291E+03,
     + 9.0764E+03, 9.1239E+03, 9.1716E+03, 9.2195E+03, 9.2676E+03,
     +  9.3159E+03, 9.3643E+03, 9.4130E+03, 9.4619E+03, 9.5110E+03,
     + 9.5603E+03, 9.6097E+03, 9.6594E+03, 9.7093E+03, 9.7594E+03,
     +  9.8097E+03, 9.8601E+03, 9.9108E+03, 9.9617E+03, 1.0013E+04,
     + 1.0064E+04, 1.0116E+04, 1.0167E+04, 1.0219E+04, 1.0271E+04,
     +  1.0324E+04, 1.0376E+04, 1.0429E+04, 1.0482E+04, 1.0535E+04,
     + 1.0589E+04, 1.0642E+04, 1.0696E+04, 1.0750E+04, 1.0804E+04,
     +  1.0859E+04, 1.0914E+04, 1.0968E+04, 1.1024E+04, 1.1079E+04,
     + 1.1134E+04, 1.1190E+04, 1.1246E+04, 1.1302E+04, 1.1359E+04,
     +  1.1416E+04, 1.1472E+04, 1.1530E+04, 1.1587E+04, 1.1644E+04,
     + 1.1702E+04, 1.1760E+04, 1.1818E+04, 1.1877E+04, 1.1936E+04,
     +  1.1995E+04, 1.2054E+04, 1.2113E+04, 1.2173E+04, 1.2233E+04,
     + 1.2293E+04, 1.2353E+04, 1.2414E+04, 1.2474E+04, 1.2535E+04,
     +  1.2597E+04, 1.2658E+04, 1.2720E+04, 1.2782E+04, 1.2844E+04,
     + 1.2907E+04, 1.2969E+04, 1.3032E+04, 1.3095E+04, 1.3159E+04,
     +  1.3223E+04, 1.3286E+04, 1.3351E+04, 1.3415E+04, 1.3480E+04,
     + 1.3545E+04, 1.3610E+04, 1.3675E+04, 1.3741E+04, 1.3807E+04,
     +  1.3873E+04, 1.3939E+04, 1.4006E+04, 1.4073E+04, 1.4140E+04,
     + 1.4208E+04, 1.4275E+04, 1.4343E+04, 1.4411E+04, 1.4480E+04,
     +  1.4549E+04, 1.4618E+04, 1.4687E+04, 1.4756E+04, 1.4826E+04,
     + 1.4896E+04, 1.4966E+04, 1.5037E+04, 1.5108E+04, 1.5179E+04,
     +  1.5250E+04, 1.5322E+04, 1.5394E+04, 1.5466E+04, 1.5538E+04,
     + 1.5611E+04, 1.5684E+04, 1.5757E+04, 1.5831E+04, 1.5905E+04,
     +  1.5979E+04, 1.6053E+04, 1.6128E+04, 1.6203E+04, 1.6278E+04,
     + 1.6353E+04, 1.6429E+04, 1.6505E+04, 1.6582E+04, 1.6658E+04,
     +  1.6735E+04, 1.6812E+04, 1.6890E+04, 1.6968E+04, 1.7046E+04,
     + 1.7124E+04, 1.7203E+04, 1.7281E+04, 1.7361E+04, 1.7440E+04,
     +  1.7520E+04, 1.7600E+04, 1.7680E+04, 1.7761E+04, 1.7842E+04,
     + 1.7923E+04, 1.8005E+04, 1.8087E+04, 1.8169E+04, 1.8251E+04,
     +  1.8334E+04, 1.8417E+04, 1.8501E+04, 1.8584E+04, 1.8668E+04,
     + 1.8753E+04, 1.8837E+04, 1.8922E+04, 1.9007E+04, 1.9093E+04,
     +  1.9179E+04, 1.9265E+04, 1.9351E+04, 1.9438E+04, 1.9525E+04,
     + 1.9612E+04, 1.9700E+04, 1.9788E+04, 1.9877E+04, 1.9965E+04,
     +  2.0054E+04, 2.0144E+04, 2.0233E+04, 2.0323E+04, 2.0413E+04,
     + 2.0504E+04, 2.0595E+04, 2.0686E+04, 2.0778E+04, 2.0870E+04,
     +  2.0962E+04, 2.1055E+04, 2.1147E+04, 2.1241E+04, 2.1334E+04,
     + 2.1428E+04, 2.1522E+04, 2.1617E+04, 2.1712E+04, 2.1807E+04,
     +  2.1903E+04, 2.1999E+04, 2.2095E+04, 2.2192E+04, 2.2289E+04,
     + 2.2386E+04, 2.2484E+04, 2.2581E+04, 2.2680E+04, 2.2779E+04,
     +  2.2878E+04, 2.2977E+04, 2.3077E+04, 2.3177E+04, 2.3277E+04,
     + 2.3378E+04, 2.3479E+04, 2.3581E+04, 2.3683E+04, 2.3785E+04,
     +  2.3887E+04, 2.3990E+04, 2.4094E+04, 2.4197E+04, 2.4301E+04,
     + 2.4406E+04, 2.4510E+04, 2.4616E+04, 2.4721E+04, 2.4827E+04,
     +  2.4933E+04, 2.5040E+04, 2.5147E+04, 2.5254E+04, 2.5362E+04,
     + 2.5470E+04, 2.5578E+04, 2.5687E+04, 2.5796E+04, 2.5906E+04,
     +  2.6016E+04, 2.6126E+04, 2.6237E+04, 2.6348E+04, 2.6460E+04,
     + 2.6571E+04, 2.6684E+04, 2.6796E+04, 2.6909E+04, 2.7023E+04,
     +  2.7137E+04, 2.7251E+04, 2.7366E+04, 2.7481E+04, 2.7596E+04,
     + 2.7712E+04, 2.7828E+04, 2.7945E+04, 2.8062E+04, 2.8179E+04,
     +  2.8297E+04, 2.8415E+04, 2.8534E+04, 2.8653E+04, 2.8772E+04,
     + 2.8892E+04, 2.9012E+04, 2.9133E+04, 2.9254E+04, 2.9375E+04,
     +  2.9497E+04, 2.9619E+04, 2.9742E+04, 2.9865E+04, 2.9989E+04,
     + 3.0113E+04, 3.0237E+04, 3.0362E+04, 3.0487E+04, 3.0613E+04,
     +  3.0739E+04, 3.0865E+04, 3.0992E+04, 3.1120E+04, 3.1247E+04,
     + 3.1376E+04, 3.1504E+04, 3.1633E+04, 3.1763E+04, 3.1893E+04,
     +  3.2023E+04, 3.2154E+04, 3.2285E+04, 3.2417E+04, 3.2549E+04,
     + 3.2681E+04, 3.2814E+04, 3.2948E+04, 3.3082E+04, 3.3216E+04,
     +  3.3351E+04, 3.3486E+04, 3.3622E+04, 3.3758E+04, 3.3895E+04,
     + 3.4032E+04, 3.4169E+04, 3.4307E+04, 3.4446E+04, 3.4584E+04,
     +  3.4724E+04, 3.4864E+04, 3.5004E+04, 3.5144E+04, 3.5286E+04,
     + 3.5427E+04, 3.5569E+04, 3.5712E+04, 3.5855E+04, 3.5999E+04,
     +  3.6143E+04, 3.6287E+04, 3.6432E+04, 3.6577E+04, 3.6723E+04,
     + 3.6870E+04, 3.7017E+04, 3.7164E+04, 3.7312E+04, 3.7460E+04,
     +  3.7609E+04, 3.7758E+04, 3.7908E+04, 3.8058E+04, 3.8209E+04,
     + 3.8360E+04, 3.8512E+04, 3.8664E+04, 3.8817E+04, 3.8970E+04,
     +  3.9124E+04, 3.9278E+04, 3.9433E+04, 3.9588E+04, 3.9744E+04,
     + 3.9900E+04, 4.0057E+04, 4.0214E+04, 4.0372E+04, 4.0530E+04,
     +  4.0689E+04, 4.0848E+04, 4.1008E+04, 4.1168E+04, 4.1329E+04,
     + 4.1490E+04, 4.1652E+04, 4.1814E+04, 4.1977E+04, 4.2141E+04,
     +  4.2305E+04, 4.2469E+04, 4.2634E+04, 4.2800E+04, 4.2966E+04,
     + 4.3132E+04, 4.3300E+04, 4.3467E+04, 4.3635E+04, 4.3804E+04,
     +  4.3973E+04, 4.4143E+04, 4.4314E+04, 4.4484E+04, 4.4656E+04,
     + 4.4828E+04, 4.5000E+04, 4.5173E+04, 4.5347E+04, 4.5521E+04,
     +  4.5696E+04, 4.5871E+04, 4.6047E+04, 4.6223E+04, 4.6400E+04,
     + 4.6578E+04, 4.6756E+04, 4.6935E+04, 4.7114E+04, 4.7294E+04,
     +  4.7474E+04, 4.7655E+04, 4.7836E+04, 4.8018E+04, 4.8201E+04,
     + 4.8384E+04, 4.8568E+04, 4.8752E+04, 4.8937E+04, 4.9123E+04,
     +  4.9309E+04, 4.9495E+04, 4.9683E+04, 4.9870E+04, 5.0059E+04,
     + 5.0248E+04, 5.0437E+04, 5.0628E+04, 5.0818E+04, 5.1010E+04,
     +  5.1202E+04, 5.1394E+04, 5.1587E+04, 5.1781E+04, 5.1975E+04,
     + 5.2170E+04, 5.2366E+04, 5.2562E+04, 5.2759E+04, 5.2956E+04,
     +  5.3154E+04, 5.3353E+04, 5.3552E+04, 5.3752E+04, 5.3952E+04,
     + 5.4154E+04, 5.4355E+04, 5.4558E+04, 5.4760E+04, 5.4964E+04,
     +  5.5168E+04, 5.5373E+04, 5.5578E+04, 5.5785E+04, 5.5991E+04,
     + 5.6199E+04, 5.6407E+04, 5.6615E+04, 5.6824E+04, 5.7034E+04,
     +  5.7245E+04, 5.7456E+04, 5.7668E+04, 5.7880E+04, 5.8093E+04,
     + 5.8307E+04, 5.8522E+04, 5.8737E+04, 5.8952E+04, 5.9169E+04,
     +  5.9386E+04, 5.9603E+04, 5.9822E+04, 6.0041E+04, 6.0260E+04,
     + 6.0481E+04, 6.0702E+04, 6.0923E+04, 6.1146E+04, 6.1369E+04,
     +  6.1592E+04, 6.1817E+04, 6.2042E+04, 6.2268E+04, 6.2494E+04,
     + 6.2721E+04, 6.2949E+04, 6.3177E+04, 6.3406E+04, 6.3636E+04,
     +  6.3867E+04, 6.4098E+04, 6.4330E+04, 6.4562E+04, 6.4796E+04,
     + 6.5030E+04, 6.5264E+04, 6.5500E+04, 6.5736E+04, 6.5973E+04,
     +  6.6210E+04, 6.6448E+04, 6.6687E+04, 6.6927E+04, 6.7167E+04,
     + 6.7408E+04, 6.7650E+04, 6.7892E+04, 6.8136E+04, 6.8379E+04,
     +  6.8624E+04, 6.8869E+04, 6.9116E+04, 6.9362E+04, 6.9610E+04,
     + 6.9858E+04, 7.0107E+04, 7.0357E+04, 7.0608E+04, 7.0859E+04,
     +  7.1111E+04, 7.1363E+04, 7.1617E+04, 7.1871E+04, 7.2126E+04,
     + 7.2382E+04, 7.2638E+04, 7.2895E+04, 7.3153E+04, 7.3412E+04,
     +  7.3671E+04, 7.3932E+04, 7.4193E+04, 7.4454E+04, 7.4717E+04,
     + 7.4980E+04, 7.5244E+04, 7.5509E+04, 7.5775E+04, 7.6041E+04,
     +  7.6308E+04, 7.6576E+04, 7.6845E+04, 7.7114E+04, 7.7384E+04,
     + 7.7656E+04, 7.7927E+04, 7.8200E+04, 7.8473E+04, 7.8748E+04,
     +  7.9023E+04, 7.9298E+04, 7.9575E+04, 7.9852E+04, 8.0130E+04,
     + 8.0409E+04, 8.0689E+04, 8.0970E+04, 8.1251E+04, 8.1533E+04,
     +  8.1816E+04, 8.2100E+04, 8.2385E+04, 8.2670E+04, 8.2957E+04,
     + 8.3244E+04, 8.3532E+04, 8.3820E+04, 8.4110E+04, 8.4400E+04,
     +  8.4692E+04, 8.4984E+04, 8.5277E+04, 8.5570E+04, 8.5865E+04,
     + 8.6160E+04, 8.6457E+04, 8.6754E+04, 8.7052E+04, 8.7350E+04,
     +  8.7650E+04, 8.7951E+04, 8.8252E+04, 8.8554E+04, 8.8857E+04,
     + 8.9161E+04, 8.9466E+04, 8.9771E+04, 9.0078E+04, 9.0385E+04,
     +  9.0694E+04, 9.1003E+04, 9.1313E+04, 9.1623E+04, 9.1935E+04,
     + 9.2248E+04, 9.2561E+04, 9.2876E+04, 9.3191E+04, 9.3507E+04,
     +  9.3824E+04, 9.4142E+04, 9.4461E+04, 9.4780E+04, 9.5101E+04,
     + 9.5422E+04, 9.5745E+04, 9.6068E+04, 9.6392E+04, 9.6717E+04,
     +  9.7043E+04, 9.7370E+04, 9.7698E+04, 9.8027E+04, 9.8356E+04,
     + 9.8687E+04, 9.9018E+04, 9.9351E+04, 9.9684E+04, 1.0002E+05,
     +  1.0035E+05, 1.0069E+05, 1.0103E+05, 1.0136E+05, 1.0170E+05,
     + 1.0204E+05, 1.0238E+05, 1.0273E+05, 1.0307E+05, 1.0341E+05,
     +  1.0376E+05, 1.0410E+05, 1.0445E+05, 1.0480E+05, 1.0514E+05,
     + 1.0549E+05, 1.0584E+05, 1.0619E+05, 1.0655E+05, 1.0690E+05,
     +  1.0725E+05, 1.0761E+05, 1.0797E+05, 1.0832E+05, 1.0868E+05,
     + 1.0904E+05, 1.0940E+05, 1.0976E+05, 1.1012E+05, 1.1049E+05,
     +  1.1085E+05, 1.1121E+05, 1.1158E+05, 1.1195E+05, 1.1231E+05,
     + 1.1268E+05, 1.1305E+05, 1.1342E+05, 1.1380E+05, 1.1417E+05,
     +  1.1454E+05, 1.1492E+05, 1.1529E+05, 1.1567E+05, 1.1605E+05,
     + 1.1643E+05, 1.1681E+05, 1.1719E+05, 1.1757E+05, 1.1795E+05,
     +  1.1834E+05, 1.1872E+05, 1.1911E+05, 1.1950E+05, 1.1988E+05,
     + 1.2027E+05, 1.2066E+05, 1.2105E+05, 1.2145E+05, 1.2184E+05,
     +  1.2224E+05, 1.2263E+05, 1.2303E+05, 1.2342E+05, 1.2382E+05,
     + 1.2422E+05, 1.2462E+05, 1.2503E+05, 1.2543E+05, 1.2583E+05,
     +  1.2624E+05, 1.2664E+05, 1.2705E+05, 1.2746E+05, 1.2787E+05,
     + 1.2828E+05, 1.2869E+05, 1.2910E+05, 1.2952E+05, 1.2993E+05,
     +  1.3035E+05, 1.3077E+05, 1.3118E+05, 1.3160E+05, 1.3202E+05,
     + 1.3245E+05, 1.3287E+05, 1.3329E+05, 1.3372E+05, 1.3414E+05,
     +  1.3457E+05, 1.3500E+05, 1.3543E+05, 1.3586E+05, 1.3629E+05,
     + 1.3672E+05, 1.3715E+05, 1.3759E+05, 1.3803E+05, 1.3846E+05,
     +  1.3890E+05, 1.3934E+05, 1.3978E+05, 1.4022E+05, 1.4067E+05,
     + 1.4111E+05, 1.4156E+05, 1.4200E+05, 1.4245E+05, 1.4290E+05,
     +  1.4335E+05, 1.4380E+05, 1.4425E+05, 1.4470E+05, 1.4516E+05,
     + 1.4561E+05, 1.4607E+05, 1.4653E+05, 1.4699E+05, 1.4745E+05,
     +  1.4791E+05, 1.4837E+05, 1.4884E+05, 1.4930E+05, 1.4977E+05,
     + 1.5024E+05, 1.5070E+05, 1.5117E+05, 1.5165E+05, 1.5212E+05,
     +  1.5259E+05, 1.5307E+05, 1.5354E+05, 1.5402E+05, 1.5450E+05,
     + 1.5498E+05, 1.5546E+05, 1.5594E+05, 1.5642E+05, 1.5691E+05,
     +  1.5739E+05, 1.5788E+05, 1.5837E+05, 1.5886E+05, 1.5935E+05,
     + 1.5984E+05, 1.6033E+05, 1.6083E+05, 1.6132E+05, 1.6182E+05,
     +  1.6232E+05, 1.6282E+05, 1.6332E+05, 1.6382E+05, 1.6432E+05,
     + 1.6483E+05, 1.6533E+05, 1.6584E+05, 1.6635E+05, 1.6686E+05,
     +  1.6737E+05, 1.6788E+05, 1.6839E+05, 1.6891E+05, 1.6943E+05,
     + 1.6994E+05, 1.7046E+05, 1.7098E+05, 1.7150E+05, 1.7202E+05,
     +  1.7255E+05, 1.7307E+05, 1.7360E+05, 1.7413E+05, 1.7466E+05,
     + 1.7519E+05, 1.7572E+05, 1.7625E+05, 1.7678E+05, 1.7732E+05,
     +  1.7786E+05, 1.7840E+05, 1.7893E+05, 1.7948E+05, 1.8002E+05,
     + 1.8056E+05, 1.8111E+05, 1.8165E+05, 1.8220E+05, 1.8275E+05,
     +  1.8330E+05, 1.8385E+05, 1.8440E+05, 1.8496E+05, 1.8551E+05,
     + 1.8607E+05, 1.8663E+05, 1.8719E+05, 1.8775E+05, 1.8831E+05,
     +  1.8888E+05, 1.8944E+05, 1.9001E+05, 1.9058E+05, 1.9115E+05,
     + 1.9172E+05, 1.9229E+05, 1.9286E+05, 1.9344E+05, 1.9401E+05,
     +  1.9459E+05, 1.9517E+05, 1.9575E+05, 1.9633E+05, 1.9692E+05,
     + 1.9750E+05, 1.9809E+05, 1.9867E+05, 1.9926E+05, 1.9985E+05,
     +  2.0045E+05, 2.0104E+05, 2.0163E+05, 2.0223E+05, 2.0283E+05,
     + 2.0343E+05, 2.0403E+05, 2.0463E+05, 2.0523E+05, 2.0584E+05,
     +  2.0644E+05, 2.0705E+05, 2.0766E+05, 2.0827E+05, 2.0888E+05,
     + 2.0950E+05, 2.1011E+05, 2.1073E+05, 2.1135E+05, 2.1196E+05,
     +  2.1259E+05, 2.1321E+05, 2.1383E+05, 2.1446E+05, 2.1508E+05,
     + 2.1571E+05, 2.1634E+05, 2.1697E+05, 2.1761E+05, 2.1824E+05,
     +  2.1888E+05, 2.1951E+05, 2.2015E+05, 2.2079E+05, 2.2143E+05,
     + 2.2208E+05, 2.2272E+05, 2.2337E+05, 2.2402E+05, 2.2466E+05,
     +  2.2532E+05, 2.2597E+05, 2.2662E+05, 2.2728E+05, 2.2793E+05,
     + 2.2859E+05, 2.2925E+05, 2.2991E+05, 2.3058E+05, 2.3124E+05,
     +  2.3191E+05, 2.3258E+05, 2.3324E+05, 2.3392E+05, 2.3459E+05,
     + 2.3526E+05, 2.3594E+05, 2.3661E+05, 2.3729E+05, 2.3797E+05,
     +  2.3866E+05, 2.3934E+05, 2.4002E+05, 2.4071E+05, 2.4140E+05,
     + 2.4209E+05, 2.4278E+05, 2.4347E+05, 2.4417E+05, 2.4486E+05,
     +  2.4556E+05, 2.4626E+05, 2.4696E+05, 2.4766E+05, 2.4837E+05,
     + 2.4907E+05, 2.4978E+05, 2.5049E+05, 2.5120E+05, 2.5191E+05,
     +  2.5263E+05, 2.5334E+05, 2.5406E+05, 2.5478E+05, 2.5550E+05,
     + 2.5622E+05, 2.5694E+05, 2.5767E+05, 2.5840E+05, 2.5913E+05,
     +  2.5986E+05, 2.6059E+05, 2.6132E+05, 2.6206E+05, 2.6279E+05,
     + 2.6353E+05, 2.6427E+05, 2.6501E+05, 2.6576E+05, 2.6650E+05,
     +  2.6725E+05, 2.6800E+05, 2.6875E+05, 2.6950E+05, 2.7026E+05,
     + 2.7101E+05, 2.7177E+05, 2.7253E+05, 2.7329E+05, 2.7405E+05,
     +  2.7481E+05, 2.7558E+05, 2.7635E+05, 2.7712E+05, 2.7789E+05,
     + 2.7866E+05, 2.7943E+05, 2.8021E+05, 2.8099E+05, 2.8177E+05,
     +  2.8255E+05, 2.8333E+05, 2.8411E+05, 2.8490E+05, 2.8569E+05,
     + 2.8648E+05, 2.8727E+05, 2.8806E+05, 2.8886E+05, 2.8965E+05,
     +  2.9045E+05, 2.9125E+05, 2.9206E+05, 2.9286E+05, 2.9367E+05,
     + 2.9447E+05, 2.9528E+05, 2.9609E+05, 2.9691E+05, 2.9772E+05,
     +  2.9854E+05, 2.9935E+05, 3.0017E+05, 3.0100E+05, 3.0182E+05,
     + 3.0264E+05, 3.0347E+05, 3.0430E+05, 3.0513E+05, 3.0596E+05,
     +  3.0680E+05, 3.0763E+05, 3.0847E+05, 3.0931E+05, 3.1015E+05,
     + 3.1100E+05, 3.1184E+05, 3.1269E+05, 3.1354E+05, 3.1439E+05,
     +  3.1524E+05, 3.1610E+05, 3.1695E+05, 3.1781E+05, 3.1867E+05,
     + 3.1953E+05, 3.2039E+05, 3.2126E+05, 3.2213E+05, 3.2300E+05,
     +  3.2387E+05, 3.2474E+05, 3.2562E+05, 3.2649E+05, 3.2737E+05,
     + 3.2825E+05, 3.2913E+05, 3.3002E+05, 3.3090E+05, 3.3179E+05,
     +  3.3268E+05, 3.3357E+05, 3.3447E+05, 3.3536E+05, 3.3626E+05,
     + 3.3716E+05, 3.3806E+05, 3.3896E+05, 3.3987E+05, 3.4078E+05,
     +  3.4169E+05, 3.4260E+05, 3.4351E+05, 3.4442E+05, 3.4534E+05,
     + 3.4626E+05, 3.4718E+05, 3.4810E+05, 3.4903E+05, 3.4995E+05,
     +  3.5088E+05, 3.5181E+05, 3.5274E+05, 3.5368E+05, 3.5461E+05,
     + 3.5555E+05, 3.5649E+05, 3.5743E+05, 3.5838E+05, 3.5932E+05,
     +  3.6027E+05, 3.6122E+05, 3.6217E+05, 3.6313E+05, 3.6408E+05,
     + 3.6504E+05, 3.6600E+05, 3.6696E+05, 3.6792E+05, 3.6889E+05,
     +  3.6986E+05, 3.7083E+05, 3.7180E+05, 3.7277E+05, 3.7375E+05,
     + 3.7473E+05, 3.7571E+05, 3.7669E+05, 3.7767E+05, 3.7866E+05,
     +  3.7965E+05, 3.8064E+05, 3.8163E+05, 3.8262E+05, 3.8362E+05,
     + 3.8462E+05, 3.8562E+05, 3.8662E+05, 3.8762E+05, 3.8863E+05,
     +  3.8964E+05, 3.9065E+05, 3.9166E+05, 3.9267E+05, 3.9369E+05,
     + 3.9471E+05, 3.9573E+05, 3.9675E+05, 3.9778E+05, 3.9880E+05,
     +  3.9983E+05, 4.0086E+05, 4.0190E+05, 4.0293E+05, 4.0397E+05,
     + 4.0501E+05, 4.0605E+05, 4.0709E+05, 4.0814E+05, 4.0919E+05,
     +  4.1024E+05, 4.1129E+05, 4.1234E+05, 4.1340E+05, 4.1446E+05,
     + 4.1552E+05, 4.1658E+05, 4.1765E+05, 4.1871E+05, 4.1978E+05,
     +  4.2085E+05, 4.2193E+05, 4.2300E+05, 4.2408E+05, 4.2516E+05,
     + 4.2624E+05, 4.2733E+05, 4.2841E+05, 4.2950E+05, 4.3059E+05,
     +  4.3168E+05, 4.3278E+05, 4.3387E+05, 4.3497E+05, 4.3607E+05,
     + 4.3718E+05, 4.3828E+05, 4.3939E+05, 4.4050E+05, 4.4161E+05,
     +  4.4273E+05, 4.4385E+05, 4.4496E+05, 4.4608E+05, 4.4721E+05,
     + 4.4833E+05, 4.4946E+05, 4.5059E+05, 4.5172E+05, 4.5286E+05,
     +  4.5399E+05, 4.5513E+05, 4.5627E+05, 4.5742E+05, 4.5856E+05,
     + 4.5971E+05, 4.6086E+05, 4.6201E+05, 4.6316E+05, 4.6432E+05,
     +  4.6548E+05, 4.6664E+05, 4.6780E+05, 4.6897E+05, 4.7014E+05,
     + 4.7131E+05, 4.7248E+05, 4.7365E+05, 4.7483E+05, 4.7601E+05,
     +  4.7719E+05, 4.7837E+05, 4.7956E+05, 4.8075E+05, 4.8194E+05,
     + 4.8313E+05, 4.8433E+05, 4.8552E+05, 4.8672E+05, 4.8792E+05,
     +  4.8913E+05, 4.9033E+05, 4.9154E+05, 4.9275E+05, 4.9397E+05,
     + 4.9518E+05, 4.9640E+05, 4.9762E+05, 4.9884E+05, 5.0007E+05,
     +  5.0130E+05, 5.0253E+05, 5.0376E+05, 5.0499E+05, 5.0623E+05,
     + 5.0747E+05, 5.0871E+05, 5.0995E+05, 5.1120E+05, 5.1245E+05,
     +  5.1370E+05, 5.1495E+05, 5.1620E+05, 5.1746E+05, 5.1872E+05,
     + 5.1998E+05, 5.2125E+05, 5.2252E+05, 5.2379E+05, 5.2506E+05,
     +  5.2633E+05, 5.2761E+05, 5.2889E+05, 5.3017E+05, 5.3145E+05,
     + 5.3274E+05, 5.3403E+05, 5.3532E+05, 5.3661E+05, 5.3791E+05,
     +  5.3920E+05, 5.4050E+05, 5.4181E+05, 5.4311E+05, 5.4442E+05,
     + 5.4573E+05, 5.4704E+05, 5.4836E+05, 5.4968E+05, 5.5099E+05,
     +  5.5232E+05, 5.5364E+05, 5.5497E+05, 5.5630E+05, 5.5763E+05,
     + 5.5896E+05, 5.6030E+05, 5.6164E+05, 5.6298E+05, 5.6433E+05,
     +  5.6567E+05, 5.6702E+05, 5.6837E+05, 5.6973E+05, 5.7108E+05,
     + 5.7244E+05, 5.7380E+05, 5.7517E+05, 5.7653E+05, 5.7790E+05,
     +  5.7927E+05, 5.8065E+05, 5.8202E+05, 5.8340E+05, 5.8478E+05,
     + 5.8617E+05, 5.8755E+05, 5.8894E+05, 5.9033E+05, 5.9173E+05,
     +  5.9312E+05, 5.9452E+05, 5.9592E+05, 5.9733E+05, 5.9873E+05,
     + 6.0014E+05, 6.0156E+05, 6.0297E+05, 6.0439E+05, 6.0580E+05,
     +  6.0723E+05, 6.0865E+05, 6.1008E+05, 6.1151E+05, 6.1294E+05,
     + 6.1437E+05, 6.1581E+05, 6.1725E+05, 6.1869E+05, 6.2013E+05,
     +  6.2158E+05, 6.2303E+05, 6.2448E+05, 6.2594E+05, 6.2739E+05,
     + 6.2885E+05, 6.3031E+05, 6.3178E+05, 6.3325E+05, 6.3472E+05,
     +  6.3619E+05, 6.3767E+05, 6.3914E+05, 6.4062E+05, 6.4211E+05,
     + 6.4359E+05, 6.4508E+05, 6.4657E+05, 6.4806E+05, 6.4956E+05,
     +  6.5106E+05, 6.5256E+05, 6.5406E+05, 6.5557E+05, 6.5708E+05,
     + 6.5859E+05, 6.6010E+05, 6.6162E+05, 6.6314E+05, 6.6466E+05,
     +  6.6619E+05, 6.6772E+05, 6.6925E+05, 6.7078E+05, 6.7231E+05,
     + 6.7385E+05, 6.7539E+05, 6.7694E+05, 6.7848E+05, 6.8003E+05,
     +  6.8158E+05, 6.8314E+05, 6.8469E+05, 6.8625E+05, 6.8781E+05,
     + 6.8938E+05, 6.9094E+05, 6.9251E+05, 6.9409E+05, 6.9566E+05,
     +  6.9724E+05, 6.9882E+05, 7.0040E+05, 7.0199E+05, 7.0358E+05,
     + 7.0517E+05, 7.0677E+05, 7.0836E+05, 7.0996E+05, 7.1156E+05,
     +  7.1317E+05, 7.1478E+05, 7.1639E+05, 7.1800E+05, 7.1962E+05,
     + 7.2123E+05, 7.2286E+05, 7.2448E+05, 7.2611E+05, 7.2774E+05,
     +  7.2937E+05, 7.3100E+05, 7.3264E+05, 7.3428E+05, 7.3592E+05,
     + 7.3757E+05, 7.3922E+05, 7.4087E+05, 7.4252E+05, 7.4418E+05,
     +  7.4584E+05, 7.4750E+05, 7.4917E+05, 7.5084E+05, 7.5251E+05,
     + 7.5418E+05, 7.5586E+05, 7.5754E+05, 7.5922E+05, 7.6090E+05,
     +  7.6259E+05, 7.6428E+05, 7.6597E+05, 7.6767E+05, 7.6937E+05,
     + 7.7107E+05, 7.7277E+05, 7.7448E+05, 7.7619E+05, 7.7790E+05,
     +  7.7962E+05, 7.8133E+05, 7.8305E+05, 7.8478E+05, 7.8650E+05,
     + 7.8823E+05, 7.8997E+05, 7.9170E+05, 7.9344E+05, 7.9518E+05,
     +  7.9692E+05, 7.9867E+05, 8.0042E+05, 8.0217E+05, 8.0393E+05,
     + 8.0568E+05, 8.0744E+05, 8.0921E+05, 8.1097E+05, 8.1274E+05,
     +  8.1451E+05, 8.1629E+05, 8.1807E+05, 8.1985E+05, 8.2163E+05,
     + 8.2342E+05, 8.2521E+05, 8.2700E+05, 8.2879E+05, 8.3059E+05,
     +  8.3239E+05, 8.3419E+05, 8.3600E+05, 8.3781E+05, 8.3962E+05,
     + 8.4144E+05, 8.4326E+05, 8.4508E+05, 8.4690E+05, 8.4873E+05,
     +  8.5056E+05, 8.5239E+05, 8.5422E+05, 8.5606E+05, 8.5790E+05,
     + 8.5975E+05, 8.6159E+05, 8.6344E+05, 8.6530E+05, 8.6715E+05,
     +  8.6901E+05, 8.7087E+05, 8.7274E+05, 8.7460E+05, 8.7647E+05,
     + 8.7835E+05, 8.8022E+05, 8.8210E+05, 8.8398E+05, 8.8587E+05,
     +  8.8776E+05, 8.8965E+05, 8.9154E+05, 8.9344E+05, 8.9534E+05,
     + 8.9724E+05, 8.9915E+05, 9.0106E+05, 9.0297E+05, 9.0488E+05,
     +  9.0680E+05, 9.0872E+05, 9.1064E+05, 9.1257E+05, 9.1450E+05,
     + 9.1643E+05, 9.1837E+05, 9.2031E+05, 9.2225E+05, 9.2419E+05,
     +  9.2614E+05, 9.2809E+05, 9.3004E+05, 9.3200E+05, 9.3396E+05,
     + 9.3592E+05, 9.3789E+05, 9.3985E+05, 9.4183E+05, 9.4380E+05,
     +  9.4578E+05, 9.4776E+05, 9.4974E+05, 9.5173E+05, 9.5372E+05,
     + 9.5571E+05, 9.5771E+05, 9.5970E+05, 9.6171E+05, 9.6371E+05,
     +  9.6572E+05, 9.6773E+05, 9.6974E+05, 9.7176E+05, 9.7378E+05,
     + 9.7580E+05, 9.7783E+05, 9.7986E+05, 9.8189E+05, 9.8392E+05,
     +  9.8596E+05, 9.8800E+05, 9.9005E+05, 9.9210E+05, 9.9415E+05,
     + 9.9620E+05, 9.9826E+05, 1.0003E+06, 1.0024E+06, 1.0044E+06,
     +  1.0065E+06, 1.0086E+06, 1.0107E+06, 1.0127E+06, 1.0148E+06,
     + 1.0169E+06, 1.0190E+06, 1.0211E+06, 1.0232E+06, 1.0253E+06,
     +  1.0274E+06, 1.0295E+06, 1.0316E+06, 1.0337E+06, 1.0358E+06,
     + 1.0379E+06, 1.0400E+06, 1.0422E+06, 1.0443E+06, 1.0464E+06,
     +  1.0485E+06, 1.0507E+06, 1.0528E+06, 1.0550E+06, 1.0571E+06,
     + 1.0592E+06, 1.0614E+06, 1.0635E+06, 1.0657E+06, 1.0679E+06,
     +  1.0700E+06, 1.0722E+06, 1.0744E+06, 1.0765E+06, 1.0787E+06,
     + 1.0809E+06, 1.0831E+06, 1.0852E+06, 1.0874E+06, 1.0896E+06,
     +  1.0918E+06, 1.0940E+06, 1.0962E+06, 1.0984E+06, 1.1006E+06,
     + 1.1028E+06, 1.1050E+06, 1.1073E+06, 1.1095E+06, 1.1117E+06,
     +  1.1139E+06, 1.1162E+06, 1.1184E+06, 1.1206E+06, 1.1229E+06,
     + 1.1251E+06, 1.1273E+06, 1.1296E+06, 1.1318E+06, 1.1341E+06,
     +  1.1364E+06, 1.1386E+06, 1.1409E+06, 1.1431E+06, 1.1454E+06,
     + 1.1477E+06, 1.1500E+06, 1.1522E+06, 1.1545E+06, 1.1568E+06,
     +  1.1591E+06, 1.1614E+06, 1.1637E+06, 1.1660E+06, 1.1683E+06,
     + 1.1706E+06, 1.1729E+06, 1.1752E+06, 1.1775E+06, 1.1799E+06,
     +  1.1822E+06, 1.1845E+06, 1.1868E+06, 1.1892E+06, 1.1915E+06,
     + 1.1938E+06, 1.1962E+06, 1.1985E+06, 1.2009E+06, 1.2032E+06,
     +  1.2056E+06, 1.2079E+06, 1.2103E+06, 1.2127E+06, 1.2150E+06,
     + 1.2174E+06, 1.2198E+06, 1.2222E+06, 1.2245E+06, 1.2269E+06,
     +  1.2293E+06, 1.2317E+06, 1.2341E+06, 1.2365E+06, 1.2389E+06,
     + 1.2413E+06, 1.2437E+06, 1.2461E+06, 1.2485E+06, 1.2509E+06,
     +  1.2534E+06, 1.2558E+06, 1.2582E+06, 1.2606E+06, 1.2631E+06,
     + 1.2655E+06, 1.2679E+06, 1.2704E+06, 1.2728E+06, 1.2753E+06,
     +  1.2777E+06, 1.2802E+06, 1.2827E+06, 1.2851E+06, 1.2876E+06,
     + 1.2901E+06, 1.2925E+06, 1.2950E+06, 1.2975E+06, 1.3000E+06,
     +  1.3025E+06, 1.3049E+06, 1.3074E+06, 1.3099E+06, 1.3124E+06,
     + 1.3149E+06, 1.3174E+06, 1.3199E+06, 1.3225E+06, 1.3250E+06,
     +  1.3275E+06, 1.3300E+06, 1.3325E+06, 1.3351E+06, 1.3376E+06,
     + 1.3401E+06, 1.3427E+06, 1.3452E+06, 1.3478E+06, 1.3503E+06,
     +  1.3529E+06, 1.3554E+06, 1.3580E+06, 1.3606E+06, 1.3631E+06,
     + 1.3657E+06, 1.3683E+06, 1.3709E+06, 1.3734E+06, 1.3760E+06,
     +  1.3786E+06, 1.3812E+06, 1.3838E+06, 1.3864E+06, 1.3890E+06,
     + 1.3916E+06, 1.3942E+06, 1.3968E+06, 1.3994E+06, 1.4020E+06,
     +  1.4047E+06, 1.4073E+06, 1.4099E+06, 1.4125E+06, 1.4152E+06,
     + 1.4178E+06, 1.4205E+06, 1.4231E+06, 1.4258E+06, 1.4284E+06,
     +  1.4311E+06, 1.4337E+06, 1.4364E+06, 1.4390E+06, 1.4417E+06,
     + 1.4444E+06, 1.4471E+06, 1.4497E+06, 1.4524E+06, 1.4551E+06,
     +  1.4578E+06, 1.4605E+06, 1.4632E+06, 1.4659E+06, 1.4686E+06,
     + 1.4713E+06, 1.4740E+06, 1.4767E+06, 1.4794E+06, 1.4822E+06,
     +  1.4849E+06, 1.4876E+06, 1.4903E+06, 1.4931E+06, 1.4958E+06,
     + 1.4986E+06, 1.5013E+06, 1.5041E+06, 1.5068E+06, 1.5096E+06,
     +  1.5123E+06, 1.5151E+06, 1.5179E+06, 1.5206E+06, 1.5234E+06,
     + 1.5262E+06, 1.5289E+06, 1.5317E+06, 1.5345E+06, 1.5373E+06,
     +  1.5401E+06, 1.5429E+06, 1.5457E+06, 1.5485E+06, 1.5513E+06,
     + 1.5541E+06, 1.5569E+06, 1.5598E+06, 1.5626E+06, 1.5654E+06,
     +  1.5682E+06, 1.5711E+06, 1.5739E+06, 1.5767E+06, 1.5796E+06,
     + 1.5824E+06, 1.5853E+06, 1.5881E+06, 1.5910E+06, 1.5938E+06,
     +  1.5967E+06, 1.5996E+06, 1.6024E+06, 1.6053E+06, 1.6082E+06,
     + 1.6111E+06, 1.6140E+06, 1.6169E+06, 1.6197E+06, 1.6226E+06,
     +  1.6255E+06, 1.6284E+06, 1.6313E+06, 1.6343E+06, 1.6372E+06,
     + 1.6401E+06, 1.6430E+06, 1.6459E+06, 1.6489E+06, 1.6518E+06,
     +  1.6547E+06, 1.6577E+06, 1.6606E+06, 1.6635E+06, 1.6665E+06,
     + 1.6694E+06, 1.6724E+06, 1.6754E+06, 1.6783E+06, 1.6813E+06,
     +  1.6843E+06, 1.6872E+06, 1.6902E+06, 1.6932E+06, 1.6962E+06,
     + 1.6992E+06, 1.7022E+06, 1.7051E+06, 1.7081E+06, 1.7111E+06,
     +  1.7142E+06, 1.7172E+06, 1.7202E+06, 1.7232E+06, 1.7262E+06,
     + 1.7292E+06, 1.7323E+06, 1.7353E+06, 1.7383E+06, 1.7414E+06,
     +  1.7444E+06, 1.7475E+06, 1.7505E+06, 1.7536E+06, 1.7566E+06,
     + 1.7597E+06, 1.7627E+06, 1.7658E+06, 1.7689E+06, 1.7719E+06,
     +  1.7750E+06, 1.7781E+06, 1.7812E+06, 1.7843E+06, 1.7874E+06,
     + 1.7905E+06, 1.7936E+06, 1.7967E+06, 1.7998E+06, 1.8029E+06,
     +  1.8060E+06, 1.8091E+06, 1.8122E+06, 1.8153E+06, 1.8185E+06,
     + 1.8216E+06, 1.8247E+06, 1.8279E+06, 1.8310E+06, 1.8342E+06,
     +  1.8373E+06, 1.8405E+06, 1.8436E+06, 1.8468E+06, 1.8500E+06,
     + 1.8531E+06, 1.8563E+06, 1.8595E+06, 1.8626E+06, 1.8658E+06,
     +  1.8690E+06, 1.8722E+06, 1.8754E+06, 1.8786E+06, 1.8818E+06,
     + 1.8850E+06, 1.8882E+06, 1.8914E+06, 1.8946E+06, 1.8978E+06,
     +  1.9011E+06, 1.9043E+06, 1.9075E+06, 1.9108E+06, 1.9140E+06,
     + 1.9172E+06, 1.9205E+06, 1.9237E+06, 1.9270E+06, 1.9302E+06,
     +  1.9335E+06, 1.9368E+06, 1.9400E+06, 1.9433E+06, 1.9466E+06,
     + 1.9498E+06, 1.9531E+06, 1.9564E+06, 1.9597E+06, 1.9630E+06,
     +  1.9663E+06, 1.9696E+06, 1.9729E+06, 1.9762E+06, 1.9795E+06,
     + 1.9828E+06, 1.9861E+06, 1.9895E+06, 1.9928E+06, 1.9961E+06,
     +  1.9994E+06, 2.0028E+06, 2.0061E+06, 2.0095E+06, 2.0128E+06,
     + 2.0162E+06, 2.0195E+06, 2.0229E+06, 2.0262E+06, 2.0296E+06,
     +  2.0330E+06, 2.0363E+06, 2.0397E+06, 2.0431E+06, 2.0465E+06,
     + 2.0499E+06, 2.0533E+06, 2.0567E+06, 2.0600E+06, 2.0635E+06,
     +  2.0669E+06, 2.0703E+06, 2.0737E+06, 2.0771E+06, 2.0805E+06,
     + 2.0839E+06, 2.0874E+06, 2.0908E+06, 2.0942E+06, 2.0977E+06,
     +  2.1011E+06, 2.1046E+06, 2.1080E+06, 2.1115E+06, 2.1149E+06,
     + 2.1184E+06, 2.1219E+06, 2.1253E+06, 2.1288E+06, 2.1323E+06,
     +  2.1358E+06, 2.1392E+06, 2.1427E+06, 2.1462E+06, 2.1497E+06,
     + 2.1532E+06, 2.1567E+06, 2.1602E+06, 2.1637E+06, 2.1673E+06,
     +  2.1708E+06, 2.1743E+06, 2.1778E+06, 2.1813E+06, 2.1849E+06,
     + 2.1884E+06, 2.1920E+06, 2.1955E+06, 2.1990E+06, 2.2026E+06,
     +  2.2062E+06, 2.2097E+06, 2.2133E+06, 2.2168E+06, 2.2204E+06,
     + 2.2240E+06, 2.2276E+06, 2.2311E+06, 2.2347E+06, 2.2383E+06,
     +  2.2419E+06, 2.2455E+06, 2.2491E+06, 2.2527E+06, 2.2563E+06,
     + 2.2599E+06, 2.2635E+06, 2.2672E+06, 2.2708E+06, 2.2744E+06,
     +  2.2780E+06, 2.2817E+06, 2.2853E+06, 2.2890E+06, 2.2926E+06,
     + 2.2962E+06, 2.2999E+06, 2.3036E+06, 2.3072E+06, 2.3109E+06,
     +  2.3146E+06, 2.3182E+06, 2.3219E+06, 2.3256E+06, 2.3293E+06,
     + 2.3329E+06, 2.3366E+06, 2.3403E+06, 2.3440E+06, 2.3477E+06,
     +  2.3514E+06, 2.3551E+06, 2.3589E+06, 2.3626E+06, 2.3663E+06,
     + 2.3700E+06, 2.3737E+06, 2.3775E+06, 2.3812E+06, 2.3850E+06,
     +  2.3887E+06, 2.3924E+06, 2.3962E+06, 2.4000E+06, 2.4037E+06,
     + 2.4075E+06, 2.4112E+06, 2.4150E+06, 2.4188E+06, 2.4226E+06,
     +  2.4263E+06, 2.4301E+06, 2.4339E+06, 2.4377E+06, 2.4415E+06/
      data (QofT( 2,J),J=2409,4510)/ 2.4453E+06, 2.4491E+06,
     +  2.4529E+06, 2.4567E+06, 2.4605E+06, 2.4644E+06, 2.4682E+06,
     +  2.4720E+06,
     + 2.4758E+06, 2.4797E+06, 2.4835E+06, 2.4874E+06, 2.4912E+06,
     +  2.4951E+06, 2.4989E+06, 2.5028E+06, 2.5066E+06, 2.5105E+06,
     + 2.5144E+06, 2.5182E+06, 2.5221E+06, 2.5260E+06, 2.5299E+06,
     +  2.5338E+06, 2.5377E+06, 2.5415E+06, 2.5454E+06, 2.5494E+06,
     + 2.5533E+06, 2.5572E+06, 2.5611E+06, 2.5650E+06, 2.5689E+06,
     +  2.5728E+06, 2.5768E+06, 2.5807E+06, 2.5846E+06, 2.5886E+06,
     + 2.5925E+06, 2.5965E+06, 2.6004E+06, 2.6044E+06, 2.6084E+06,
     +  2.6123E+06, 2.6163E+06, 2.6203E+06, 2.6242E+06, 2.6282E+06,
     + 2.6322E+06, 2.6362E+06, 2.6402E+06, 2.6442E+06, 2.6482E+06,
     +  2.6522E+06, 2.6562E+06, 2.6602E+06, 2.6642E+06, 2.6682E+06,
     + 2.6722E+06, 2.6763E+06, 2.6803E+06, 2.6843E+06, 2.6884E+06,
     +  2.6924E+06, 2.6965E+06, 2.7005E+06, 2.7046E+06, 2.7086E+06,
     + 2.7127E+06, 2.7167E+06, 2.7208E+06, 2.7249E+06, 2.7289E+06,
     +  2.7330E+06, 2.7371E+06, 2.7412E+06, 2.7453E+06, 2.7494E+06,
     + 2.7535E+06, 2.7576E+06, 2.7617E+06, 2.7658E+06, 2.7699E+06,
     +  2.7740E+06, 2.7782E+06, 2.7823E+06, 2.7864E+06, 2.7905E+06,
     + 2.7947E+06, 2.7988E+06, 2.8030E+06, 2.8071E+06, 2.8113E+06,
     +  2.8154E+06, 2.8196E+06, 2.8237E+06, 2.8279E+06, 2.8321E+06,
     + 2.8363E+06, 2.8404E+06, 2.8446E+06, 2.8488E+06, 2.8530E+06,
     +  2.8572E+06, 2.8614E+06, 2.8656E+06, 2.8698E+06, 2.8740E+06,
     + 2.8782E+06, 2.8824E+06, 2.8867E+06, 2.8909E+06, 2.8951E+06,
     +  2.8993E+06, 2.9036E+06, 2.9078E+06, 2.9121E+06, 2.9163E+06,
     + 2.9206E+06, 2.9248E+06, 2.9291E+06, 2.9334E+06, 2.9376E+06,
     +  2.9419E+06, 2.9462E+06, 2.9504E+06, 2.9547E+06, 2.9590E+06,
     + 2.9633E+06, 2.9676E+06, 2.9719E+06, 2.9762E+06, 2.9805E+06,
     +  2.9848E+06, 2.9891E+06, 2.9935E+06, 2.9978E+06, 3.0021E+06,
     + 3.0064E+06, 3.0108E+06, 3.0151E+06, 3.0194E+06, 3.0238E+06,
     +  3.0281E+06, 3.0325E+06, 3.0369E+06, 3.0412E+06, 3.0456E+06,
     + 3.0499E+06, 3.0543E+06, 3.0587E+06, 3.0631E+06, 3.0675E+06,
     +  3.0719E+06, 3.0762E+06, 3.0806E+06, 3.0850E+06, 3.0894E+06,
     + 3.0938E+06, 3.0983E+06, 3.1027E+06, 3.1071E+06, 3.1115E+06,
     +  3.1159E+06, 3.1204E+06, 3.1248E+06, 3.1292E+06, 3.1337E+06,
     + 3.1381E+06, 3.1426E+06, 3.1470E+06, 3.1515E+06, 3.1560E+06,
     +  3.1604E+06, 3.1649E+06, 3.1694E+06, 3.1738E+06, 3.1783E+06,
     + 3.1828E+06, 3.1873E+06, 3.1918E+06, 3.1963E+06, 3.2008E+06,
     +  3.2053E+06, 3.2098E+06, 3.2143E+06, 3.2188E+06, 3.2234E+06,
     + 3.2279E+06, 3.2324E+06, 3.2369E+06, 3.2415E+06, 3.2460E+06,
     +  3.2506E+06, 3.2551E+06, 3.2596E+06, 3.2642E+06, 3.2688E+06,
     + 3.2733E+06, 3.2779E+06, 3.2825E+06, 3.2870E+06, 3.2916E+06,
     +  3.2962E+06, 3.3008E+06, 3.3054E+06, 3.3100E+06, 3.3146E+06,
     + 3.3192E+06, 3.3238E+06, 3.3284E+06, 3.3330E+06, 3.3376E+06,
     +  3.3422E+06, 3.3469E+06, 3.3515E+06, 3.3561E+06, 3.3608E+06,
     + 3.3654E+06, 3.3700E+06, 3.3747E+06, 3.3793E+06, 3.3840E+06,
     +  3.3887E+06, 3.3933E+06, 3.3980E+06, 3.4027E+06, 3.4073E+06,
     + 3.4120E+06, 3.4167E+06, 3.4214E+06, 3.4261E+06, 3.4308E+06,
     +  3.4355E+06, 3.4402E+06, 3.4449E+06, 3.4496E+06, 3.4543E+06,
     + 3.4590E+06, 3.4637E+06, 3.4685E+06, 3.4732E+06, 3.4779E+06,
     +  3.4827E+06, 3.4874E+06, 3.4922E+06, 3.4969E+06, 3.5017E+06,
     + 3.5064E+06, 3.5112E+06, 3.5159E+06, 3.5207E+06, 3.5255E+06,
     +  3.5303E+06, 3.5350E+06, 3.5398E+06, 3.5446E+06, 3.5494E+06,
     + 3.5542E+06, 3.5590E+06, 3.5638E+06, 3.5686E+06, 3.5734E+06,
     +  3.5782E+06, 3.5831E+06, 3.5879E+06, 3.5927E+06, 3.5975E+06,
     + 3.6024E+06, 3.6072E+06, 3.6121E+06, 3.6169E+06, 3.6218E+06,
     +  3.6266E+06, 3.6315E+06, 3.6363E+06, 3.6412E+06, 3.6461E+06,
     + 3.6509E+06, 3.6558E+06, 3.6607E+06, 3.6656E+06, 3.6705E+06,
     +  3.6754E+06, 3.6803E+06, 3.6852E+06, 3.6901E+06, 3.6950E+06,
     + 3.6999E+06, 3.7048E+06, 3.7097E+06, 3.7147E+06, 3.7196E+06,
     +  3.7245E+06, 3.7295E+06, 3.7344E+06, 3.7393E+06, 3.7443E+06,
     + 3.7492E+06, 3.7542E+06, 3.7592E+06, 3.7641E+06, 3.7691E+06,
     +  3.7741E+06, 3.7790E+06, 3.7840E+06, 3.7890E+06, 3.7940E+06,
     + 3.7990E+06, 3.8040E+06, 3.8090E+06, 3.8140E+06, 3.8190E+06,
     +  3.8240E+06, 3.8290E+06, 3.8340E+06, 3.8390E+06, 3.8441E+06,
     + 3.8491E+06, 3.8541E+06, 3.8592E+06, 3.8642E+06, 3.8692E+06,
     +  3.8743E+06, 3.8793E+06, 3.8844E+06, 3.8895E+06, 3.8945E+06,
     + 3.8996E+06, 3.9047E+06, 3.9097E+06, 3.9148E+06, 3.9199E+06,
     +  3.9250E+06, 3.9301E+06, 3.9352E+06, 3.9403E+06, 3.9454E+06,
     + 3.9505E+06, 3.9556E+06, 3.9607E+06, 3.9658E+06, 3.9710E+06,
     +  3.9761E+06, 3.9812E+06, 3.9864E+06, 3.9915E+06, 3.9966E+06,
     + 4.0018E+06, 4.0069E+06, 4.0121E+06, 4.0172E+06, 4.0224E+06,
     +  4.0276E+06, 4.0327E+06, 4.0379E+06, 4.0431E+06, 4.0483E+06,
     + 4.0535E+06, 4.0586E+06, 4.0638E+06, 4.0690E+06, 4.0742E+06,
     +  4.0794E+06, 4.0847E+06, 4.0899E+06, 4.0951E+06, 4.1003E+06,
     + 4.1055E+06, 4.1108E+06, 4.1160E+06, 4.1212E+06, 4.1265E+06,
     +  4.1317E+06, 4.1369E+06, 4.1422E+06, 4.1475E+06, 4.1527E+06,
     + 4.1580E+06, 4.1632E+06, 4.1685E+06, 4.1738E+06, 4.1791E+06,
     +  4.1843E+06, 4.1896E+06, 4.1949E+06, 4.2002E+06, 4.2055E+06,
     + 4.2108E+06, 4.2161E+06, 4.2214E+06, 4.2267E+06, 4.2321E+06,
     +  4.2374E+06, 4.2427E+06, 4.2480E+06, 4.2534E+06, 4.2587E+06,
     + 4.2640E+06, 4.2694E+06, 4.2747E+06, 4.2801E+06, 4.2854E+06,
     +  4.2908E+06, 4.2962E+06, 4.3015E+06, 4.3069E+06, 4.3123E+06,
     + 4.3177E+06, 4.3230E+06, 4.3284E+06, 4.3338E+06, 4.3392E+06,
     +  4.3446E+06, 4.3500E+06, 4.3554E+06, 4.3608E+06, 4.3662E+06,
     + 4.3717E+06, 4.3771E+06, 4.3825E+06, 4.3879E+06, 4.3934E+06,
     +  4.3988E+06, 4.4042E+06, 4.4097E+06, 4.4151E+06, 4.4206E+06,
     + 4.4261E+06, 4.4315E+06, 4.4370E+06, 4.4424E+06, 4.4479E+06,
     +  4.4534E+06, 4.4589E+06, 4.4644E+06, 4.4698E+06, 4.4753E+06,
     + 4.4808E+06, 4.4863E+06, 4.4918E+06, 4.4973E+06, 4.5028E+06,
     +  4.5084E+06, 4.5139E+06, 4.5194E+06, 4.5249E+06, 4.5305E+06,
     + 4.5360E+06, 4.5415E+06, 4.5471E+06, 4.5526E+06, 4.5582E+06,
     +  4.5637E+06, 4.5693E+06, 4.5748E+06, 4.5804E+06, 4.5860E+06,
     + 4.5915E+06, 4.5971E+06, 4.6027E+06, 4.6083E+06, 4.6139E+06,
     +  4.6195E+06, 4.6251E+06, 4.6306E+06, 4.6363E+06, 4.6419E+06,
     + 4.6475E+06, 4.6531E+06, 4.6587E+06, 4.6643E+06, 4.6700E+06,
     +  4.6756E+06, 4.6812E+06, 4.6869E+06, 4.6925E+06, 4.6981E+06,
     + 4.7038E+06, 4.7094E+06, 4.7151E+06, 4.7208E+06, 4.7264E+06,
     +  4.7321E+06, 4.7378E+06, 4.7434E+06, 4.7491E+06, 4.7548E+06,
     + 4.7605E+06, 4.7662E+06, 4.7719E+06, 4.7776E+06, 4.7833E+06,
     +  4.7890E+06, 4.7947E+06, 4.8004E+06, 4.8061E+06, 4.8119E+06,
     + 4.8176E+06, 4.8233E+06, 4.8290E+06, 4.8348E+06, 4.8405E+06,
     +  4.8463E+06, 4.8520E+06, 4.8578E+06, 4.8635E+06, 4.8693E+06,
     + 4.8750E+06, 4.8808E+06, 4.8866E+06, 4.8924E+06, 4.8981E+06,
     +  4.9039E+06, 4.9097E+06, 4.9155E+06, 4.9213E+06, 4.9271E+06,
     + 4.9329E+06, 4.9387E+06, 4.9445E+06, 4.9503E+06, 4.9561E+06,
     +  4.9620E+06, 4.9678E+06, 4.9736E+06, 4.9795E+06, 4.9853E+06,
     + 4.9911E+06, 4.9970E+06, 5.0028E+06, 5.0087E+06, 5.0145E+06,
     +  5.0204E+06, 5.0263E+06, 5.0321E+06, 5.0380E+06, 5.0439E+06,
     + 5.0498E+06, 5.0556E+06, 5.0615E+06, 5.0674E+06, 5.0733E+06,
     +  5.0792E+06, 5.0851E+06, 5.0910E+06, 5.0969E+06, 5.1028E+06,
     + 5.1088E+06, 5.1147E+06, 5.1206E+06, 5.1265E+06, 5.1325E+06,
     +  5.1384E+06, 5.1443E+06, 5.1503E+06, 5.1562E+06, 5.1622E+06,
     + 5.1681E+06, 5.1741E+06, 5.1801E+06, 5.1860E+06, 5.1920E+06,
     +  5.1980E+06, 5.2039E+06, 5.2099E+06, 5.2159E+06, 5.2219E+06,
     + 5.2279E+06, 5.2339E+06, 5.2399E+06, 5.2459E+06, 5.2519E+06,
     +  5.2579E+06, 5.2639E+06, 5.2700E+06, 5.2760E+06, 5.2820E+06,
     + 5.2880E+06, 5.2941E+06, 5.3001E+06, 5.3061E+06, 5.3122E+06,
     +  5.3182E+06, 5.3243E+06, 5.3304E+06, 5.3364E+06, 5.3425E+06,
     + 5.3486E+06, 5.3546E+06, 5.3607E+06, 5.3668E+06, 5.3729E+06,
     +  5.3790E+06, 5.3850E+06, 5.3911E+06, 5.3972E+06, 5.4033E+06,
     + 5.4094E+06, 5.4156E+06, 5.4217E+06, 5.4278E+06, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_HNO3  (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 2),QofT( 2,NT),Q(NT)
      data (Tmx(I),I=1,2)/3500.,3500./
      data xgj/ 6.,4./
c...     HNO3
c...        --       146
      data (QofT( 1,J),J=1,2408)/ 3.5735E+01, 9.5900E+01, 1.7318E+02,
     +  2.6436E+02, 3.6758E+02, 4.8157E+02, 6.0539E+02, 7.3832E+02,
     + 8.7977E+02, 1.0293E+03, 1.1864E+03, 1.3507E+03, 1.5221E+03,
     +  1.7001E+03, 1.8846E+03, 2.0753E+03, 2.2721E+03, 2.4747E+03,
     + 2.6830E+03, 2.8968E+03, 3.3406E+03, 3.8049E+03, 4.2890E+03,
     +  4.7921E+03, 5.3134E+03, 5.8524E+03, 6.4085E+03, 6.9812E+03,
     + 7.5700E+03, 8.1745E+03, 8.7944E+03, 9.4291E+03, 1.0078E+04,
     +  1.0742E+04, 1.1420E+04, 1.2111E+04, 1.2816E+04, 1.3534E+04,
     + 1.4265E+04, 1.5008E+04, 1.5765E+04, 1.6533E+04, 1.7314E+04,
     +  1.8107E+04, 1.8911E+04, 1.9728E+04, 2.0556E+04, 2.1395E+04,
     + 2.2246E+04, 2.3108E+04, 2.3982E+04, 2.4866E+04, 2.5762E+04,
     +  2.6669E+04, 2.7586E+04, 2.8515E+04, 2.9454E+04, 3.0405E+04,
     + 3.1367E+04, 3.2339E+04, 3.3323E+04, 3.4317E+04, 3.5323E+04,
     +  3.6339E+04, 3.7367E+04, 3.8407E+04, 3.9457E+04, 4.0520E+04,
     + 4.1593E+04, 4.2679E+04, 4.3776E+04, 4.4885E+04, 4.6007E+04,
     +  4.7140E+04, 4.8286E+04, 4.9445E+04, 5.0616E+04, 5.1800E+04,
     + 5.2997E+04, 5.4207E+04, 5.5431E+04, 5.6669E+04, 5.7920E+04,
     +  5.9186E+04, 6.0466E+04, 6.1760E+04, 6.3069E+04, 6.4393E+04,
     + 6.5733E+04, 6.7087E+04, 6.8458E+04, 6.9845E+04, 7.1248E+04,
     +  7.2667E+04, 7.4104E+04, 7.5557E+04, 7.7028E+04, 7.8517E+04,
     + 8.0023E+04, 8.1548E+04, 8.3092E+04, 8.4654E+04, 8.6236E+04,
     +  8.7837E+04, 8.9458E+04, 9.1099E+04, 9.2760E+04, 9.4443E+04,
     + 9.6147E+04, 9.7872E+04, 9.9619E+04, 1.0139E+05, 1.0318E+05,
     +  1.0500E+05, 1.0683E+05, 1.0870E+05, 1.1058E+05, 1.1249E+05,
     + 1.1443E+05, 1.1639E+05, 1.1837E+05, 1.2039E+05, 1.2242E+05,
     +  1.2449E+05, 1.2658E+05, 1.2870E+05, 1.3085E+05, 1.3303E+05,
     + 1.3523E+05, 1.3747E+05, 1.3973E+05, 1.4203E+05, 1.4435E+05,
     +  1.4671E+05, 1.4910E+05, 1.5152E+05, 1.5397E+05, 1.5646E+05,
     + 1.5898E+05, 1.6153E+05, 1.6412E+05, 1.6675E+05, 1.6941E+05,
     +  1.7210E+05, 1.7483E+05, 1.7760E+05, 1.8041E+05, 1.8325E+05,
     + 1.8614E+05, 1.8906E+05, 1.9202E+05, 1.9503E+05, 1.9807E+05,
     +  2.0116E+05, 2.0428E+05, 2.0746E+05, 2.1067E+05, 2.1393E+05,
     + 2.1723E+05, 2.2058E+05, 2.2397E+05, 2.2741E+05, 2.3090E+05,
     +  2.3443E+05, 2.3801E+05, 2.4165E+05, 2.4533E+05, 2.4906E+05,
     + 2.5284E+05, 2.5668E+05, 2.6056E+05, 2.6450E+05, 2.6850E+05,
     +  2.7255E+05, 2.7665E+05, 2.8081E+05, 2.8503E+05, 2.8930E+05,
     + 2.9364E+05, 2.9803E+05, 3.0248E+05, 3.0699E+05, 3.1157E+05,
     +  3.1620E+05, 3.2091E+05, 3.2567E+05, 3.3050E+05, 3.3539E+05,
     + 3.4035E+05, 3.4538E+05, 3.5048E+05, 3.5565E+05, 3.6088E+05,
     +  3.6619E+05, 3.7157E+05, 3.7702E+05, 3.8255E+05, 3.8815E+05,
     + 3.9383E+05, 3.9958E+05, 4.0541E+05, 4.1132E+05, 4.1731E+05,
     +  4.2338E+05, 4.2954E+05, 4.3577E+05, 4.4209E+05, 4.4850E+05,
     + 4.5499E+05, 4.6157E+05, 4.6823E+05, 4.7499E+05, 4.8184E+05,
     +  4.8878E+05, 4.9581E+05, 5.0293E+05, 5.1015E+05, 5.1747E+05,
     + 5.2488E+05, 5.3240E+05, 5.4001E+05, 5.4773E+05, 5.5555E+05,
     +  5.6347E+05, 5.7149E+05, 5.7963E+05, 5.8787E+05, 5.9622E+05,
     + 6.0468E+05, 6.1326E+05, 6.2194E+05, 6.3074E+05, 6.3966E+05,
     +  6.4870E+05, 6.5785E+05, 6.6713E+05, 6.7652E+05, 6.8605E+05,
     + 6.9569E+05, 7.0546E+05, 7.1536E+05, 7.2539E+05, 7.3556E+05,
     +  7.4585E+05, 7.5628E+05, 7.6684E+05, 7.7755E+05, 7.8839E+05,
     + 7.9937E+05, 8.1050E+05, 8.2177E+05, 8.3318E+05, 8.4475E+05,
     +  8.5646E+05, 8.6833E+05, 8.8035E+05, 8.9253E+05, 9.0486E+05,
     + 9.1735E+05, 9.3000E+05, 9.4282E+05, 9.5580E+05, 9.6895E+05,
     +  9.8226E+05, 9.9575E+05, 1.0094E+06, 1.0232E+06, 1.0373E+06,
     + 1.0514E+06, 1.0658E+06, 1.0804E+06, 1.0951E+06, 1.1100E+06,
     +  1.1252E+06, 1.1405E+06, 1.1560E+06, 1.1717E+06, 1.1876E+06,
     + 1.2037E+06, 1.2200E+06, 1.2365E+06, 1.2532E+06, 1.2701E+06,
     +  1.2873E+06, 1.3046E+06, 1.3222E+06, 1.3400E+06, 1.3580E+06,
     + 1.3763E+06, 1.3947E+06, 1.4134E+06, 1.4324E+06, 1.4515E+06,
     +  1.4710E+06, 1.4906E+06, 1.5105E+06, 1.5307E+06, 1.5511E+06,
     + 1.5717E+06, 1.5926E+06, 1.6138E+06, 1.6352E+06, 1.6569E+06,
     +  1.6789E+06, 1.7011E+06, 1.7236E+06, 1.7464E+06, 1.7694E+06,
     + 1.7928E+06, 1.8164E+06, 1.8403E+06, 1.8645E+06, 1.8890E+06,
     +  1.9139E+06, 1.9390E+06, 1.9644E+06, 1.9901E+06, 2.0161E+06,
     + 2.0425E+06, 2.0692E+06, 2.0962E+06, 2.1235E+06, 2.1512E+06,
     +  2.1792E+06, 2.2075E+06, 2.2362E+06, 2.2652E+06, 2.2945E+06,
     + 2.3243E+06, 2.3543E+06, 2.3848E+06, 2.4156E+06, 2.4468E+06,
     +  2.4783E+06, 2.5102E+06, 2.5425E+06, 2.5752E+06, 2.6083E+06,
     + 2.6418E+06, 2.6756E+06, 2.7099E+06, 2.7446E+06, 2.7797E+06,
     +  2.8152E+06, 2.8511E+06, 2.8875E+06, 2.9243E+06, 2.9615E+06,
     + 2.9991E+06, 3.0372E+06, 3.0758E+06, 3.1148E+06, 3.1543E+06,
     +  3.1942E+06, 3.2346E+06, 3.2754E+06, 3.3168E+06, 3.3586E+06,
     + 3.4009E+06, 3.4437E+06, 3.4870E+06, 3.5308E+06, 3.5751E+06,
     +  3.6200E+06, 3.6653E+06, 3.7112E+06, 3.7576E+06, 3.8045E+06,
     + 3.8520E+06, 3.9000E+06, 3.9486E+06, 3.9978E+06, 4.0475E+06,
     +  4.0978E+06, 4.1486E+06, 4.2001E+06, 4.2521E+06, 4.3047E+06,
     + 4.3580E+06, 4.4118E+06, 4.4662E+06, 4.5213E+06, 4.5770E+06,
     +  4.6333E+06, 4.6903E+06, 4.7479E+06, 4.8062E+06, 4.8651E+06,
     + 4.9247E+06, 4.9850E+06, 5.0459E+06, 5.1076E+06, 5.1699E+06,
     +  5.2330E+06, 5.2967E+06, 5.3612E+06, 5.4263E+06, 5.4923E+06,
     + 5.5589E+06, 5.6263E+06, 5.6945E+06, 5.7634E+06, 5.8331E+06,
     +  5.9035E+06, 5.9748E+06, 6.0468E+06, 6.1196E+06, 6.1933E+06,
     + 6.2677E+06, 6.3430E+06, 6.4191E+06, 6.4961E+06, 6.5739E+06,
     +  6.6526E+06, 6.7321E+06, 6.8125E+06, 6.8938E+06, 6.9760E+06,
     + 7.0591E+06, 7.1431E+06, 7.2280E+06, 7.3139E+06, 7.4007E+06,
     +  7.4884E+06, 7.5771E+06, 7.6668E+06, 7.7574E+06, 7.8490E+06,
     + 7.9416E+06, 8.0353E+06, 8.1299E+06, 8.2256E+06, 8.3223E+06,
     +  8.4200E+06, 8.5188E+06, 8.6187E+06, 8.7196E+06, 8.8216E+06,
     + 8.9248E+06, 9.0290E+06, 9.1344E+06, 9.2409E+06, 9.3485E+06,
     +  9.4573E+06, 9.5672E+06, 9.6783E+06, 9.7906E+06, 9.9041E+06,
     + 1.0019E+07, 1.0135E+07, 1.0252E+07, 1.0370E+07, 1.0490E+07,
     +  1.0611E+07, 1.0733E+07, 1.0857E+07, 1.0981E+07, 1.1108E+07,
     + 1.1235E+07, 1.1364E+07, 1.1494E+07, 1.1626E+07, 1.1758E+07,
     +  1.1893E+07, 1.2028E+07, 1.2166E+07, 1.2304E+07, 1.2444E+07,
     + 1.2586E+07, 1.2729E+07, 1.2873E+07, 1.3019E+07, 1.3166E+07,
     +  1.3315E+07, 1.3466E+07, 1.3618E+07, 1.3772E+07, 1.3927E+07,
     + 1.4084E+07, 1.4242E+07, 1.4402E+07, 1.4564E+07, 1.4728E+07,
     +  1.4893E+07, 1.5060E+07, 1.5228E+07, 1.5399E+07, 1.5571E+07,
     + 1.5744E+07, 1.5920E+07, 1.6097E+07, 1.6276E+07, 1.6457E+07,
     +  1.6640E+07, 1.6825E+07, 1.7011E+07, 1.7200E+07, 1.7390E+07,
     + 1.7583E+07, 1.7777E+07, 1.7973E+07, 1.8171E+07, 1.8372E+07,
     +  1.8574E+07, 1.8778E+07, 1.8984E+07, 1.9193E+07, 1.9403E+07,
     + 1.9616E+07, 1.9831E+07, 2.0048E+07, 2.0267E+07, 2.0488E+07,
     +  2.0712E+07, 2.0937E+07, 2.1165E+07, 2.1395E+07, 2.1628E+07,
     + 2.1863E+07, 2.2100E+07, 2.2340E+07, 2.2582E+07, 2.2826E+07,
     +  2.3073E+07, 2.3322E+07, 2.3573E+07, 2.3828E+07, 2.4084E+07,
     + 2.4343E+07, 2.4605E+07, 2.4869E+07, 2.5136E+07, 2.5406E+07,
     +  2.5678E+07, 2.5953E+07, 2.6230E+07, 2.6510E+07, 2.6793E+07,
     + 2.7079E+07, 2.7368E+07, 2.7659E+07, 2.7953E+07, 2.8250E+07,
     +  2.8550E+07, 2.8853E+07, 2.9159E+07, 2.9467E+07, 2.9779E+07,
     + 3.0094E+07, 3.0412E+07, 3.0733E+07, 3.1056E+07, 3.1384E+07,
     +  3.1714E+07, 3.2047E+07, 3.2384E+07, 3.2724E+07, 3.3067E+07,
     + 3.3413E+07, 3.3763E+07, 3.4116E+07, 3.4472E+07, 3.4832E+07,
     +  3.5195E+07, 3.5562E+07, 3.5932E+07, 3.6306E+07, 3.6683E+07,
     + 3.7064E+07, 3.7448E+07, 3.7837E+07, 3.8228E+07, 3.8624E+07,
     +  3.9023E+07, 3.9426E+07, 3.9833E+07, 4.0243E+07, 4.0658E+07,
     + 4.1076E+07, 4.1499E+07, 4.1925E+07, 4.2355E+07, 4.2790E+07,
     +  4.3228E+07, 4.3671E+07, 4.4117E+07, 4.4568E+07, 4.5023E+07,
     + 4.5482E+07, 4.5946E+07, 4.6414E+07, 4.6886E+07, 4.7363E+07,
     +  4.7844E+07, 4.8329E+07, 4.8819E+07, 4.9314E+07, 4.9813E+07,
     + 5.0316E+07, 5.0825E+07, 5.1338E+07, 5.1856E+07, 5.2378E+07,
     +  5.2906E+07, 5.3438E+07, 5.3975E+07, 5.4517E+07, 5.5064E+07,
     + 5.5616E+07, 5.6173E+07, 5.6735E+07, 5.7302E+07, 5.7875E+07,
     +  5.8452E+07, 5.9035E+07, 5.9624E+07, 6.0217E+07, 6.0816E+07,
     + 6.1421E+07, 6.2030E+07, 6.2646E+07, 6.3267E+07, 6.3894E+07,
     +  6.4526E+07, 6.5164E+07, 6.5808E+07, 6.6457E+07, 6.7112E+07,
     + 6.7774E+07, 6.8441E+07, 6.9114E+07, 6.9793E+07, 7.0479E+07,
     +  7.1170E+07, 7.1868E+07, 7.2572E+07, 7.3282E+07, 7.3999E+07,
     + 7.4722E+07, 7.5451E+07, 7.6187E+07, 7.6929E+07, 7.7678E+07,
     +  7.8434E+07, 7.9196E+07, 7.9966E+07, 8.0742E+07, 8.1524E+07,
     + 8.2314E+07, 8.3111E+07, 8.3915E+07, 8.4725E+07, 8.5543E+07,
     +  8.6369E+07, 8.7201E+07, 8.8041E+07, 8.8888E+07, 8.9743E+07,
     + 9.0605E+07, 9.1474E+07, 9.2351E+07, 9.3236E+07, 9.4129E+07,
     +  9.5029E+07, 9.5938E+07, 9.6854E+07, 9.7778E+07, 9.8710E+07,
     + 9.9650E+07, 1.0060E+08, 1.0156E+08, 1.0252E+08, 1.0349E+08,
     +  1.0448E+08, 1.0547E+08, 1.0646E+08, 1.0747E+08, 1.0849E+08,
     + 1.0951E+08, 1.1055E+08, 1.1159E+08, 1.1264E+08, 1.1370E+08,
     +  1.1477E+08, 1.1585E+08, 1.1694E+08, 1.1804E+08, 1.1914E+08,
     + 1.2026E+08, 1.2138E+08, 1.2252E+08, 1.2366E+08, 1.2482E+08,
     +  1.2598E+08, 1.2716E+08, 1.2834E+08, 1.2954E+08, 1.3074E+08,
     + 1.3196E+08, 1.3318E+08, 1.3442E+08, 1.3566E+08, 1.3692E+08,
     +  1.3819E+08, 1.3947E+08, 1.4075E+08, 1.4205E+08, 1.4336E+08,
     + 1.4469E+08, 1.4602E+08, 1.4736E+08, 1.4872E+08, 1.5008E+08,
     +  1.5146E+08, 1.5285E+08, 1.5425E+08, 1.5566E+08, 1.5709E+08,
     + 1.5852E+08, 1.5997E+08, 1.6143E+08, 1.6290E+08, 1.6439E+08,
     +  1.6588E+08, 1.6739E+08, 1.6892E+08, 1.7045E+08, 1.7200E+08,
     + 1.7356E+08, 1.7513E+08, 1.7672E+08, 1.7831E+08, 1.7993E+08,
     +  1.8155E+08, 1.8319E+08, 1.8484E+08, 1.8651E+08, 1.8819E+08,
     + 1.8988E+08, 1.9159E+08, 1.9331E+08, 1.9504E+08, 1.9679E+08,
     +  1.9855E+08, 2.0033E+08, 2.0212E+08, 2.0393E+08, 2.0575E+08,
     + 2.0759E+08, 2.0944E+08, 2.1131E+08, 2.1319E+08, 2.1509E+08,
     +  2.1700E+08, 2.1892E+08, 2.2087E+08, 2.2283E+08, 2.2480E+08,
     + 2.2679E+08, 2.2880E+08, 2.3082E+08, 2.3286E+08, 2.3492E+08,
     +  2.3699E+08, 2.3908E+08, 2.4118E+08, 2.4330E+08, 2.4544E+08,
     + 2.4760E+08, 2.4977E+08, 2.5196E+08, 2.5417E+08, 2.5640E+08,
     +  2.5864E+08, 2.6090E+08, 2.6318E+08, 2.6548E+08, 2.6780E+08,
     + 2.7013E+08, 2.7249E+08, 2.7486E+08, 2.7725E+08, 2.7966E+08,
     +  2.8209E+08, 2.8453E+08, 2.8700E+08, 2.8949E+08, 2.9199E+08,
     + 2.9452E+08, 2.9706E+08, 2.9963E+08, 3.0221E+08, 3.0482E+08,
     +  3.0745E+08, 3.1009E+08, 3.1276E+08, 3.1545E+08, 3.1816E+08,
     + 3.2089E+08, 3.2364E+08, 3.2642E+08, 3.2921E+08, 3.3203E+08,
     +  3.3487E+08, 3.3773E+08, 3.4061E+08, 3.4352E+08, 3.4645E+08,
     + 3.4940E+08, 3.5237E+08, 3.5537E+08, 3.5839E+08, 3.6143E+08,
     +  3.6450E+08, 3.6759E+08, 3.7071E+08, 3.7385E+08, 3.7701E+08,
     + 3.8019E+08, 3.8341E+08, 3.8664E+08, 3.8990E+08, 3.9319E+08,
     +  3.9650E+08, 3.9984E+08, 4.0320E+08, 4.0659E+08, 4.1000E+08,
     + 4.1344E+08, 4.1691E+08, 4.2040E+08, 4.2392E+08, 4.2747E+08,
     +  4.3104E+08, 4.3464E+08, 4.3827E+08, 4.4192E+08, 4.4560E+08,
     + 4.4931E+08, 4.5305E+08, 4.5682E+08, 4.6062E+08, 4.6444E+08,
     +  4.6829E+08, 4.7217E+08, 4.7609E+08, 4.8003E+08, 4.8400E+08,
     + 4.8800E+08, 4.9203E+08, 4.9609E+08, 5.0018E+08, 5.0430E+08,
     +  5.0846E+08, 5.1264E+08, 5.1685E+08, 5.2110E+08, 5.2538E+08,
     + 5.2969E+08, 5.3403E+08, 5.3841E+08, 5.4281E+08, 5.4725E+08,
     +  5.5173E+08, 5.5623E+08, 5.6077E+08, 5.6535E+08, 5.6996E+08,
     + 5.7460E+08, 5.7927E+08, 5.8398E+08, 5.8873E+08, 5.9351E+08,
     +  5.9833E+08, 6.0318E+08, 6.0807E+08, 6.1299E+08, 6.1795E+08,
     + 6.2294E+08, 6.2798E+08, 6.3305E+08, 6.3815E+08, 6.4330E+08,
     +  6.4848E+08, 6.5370E+08, 6.5896E+08, 6.6425E+08, 6.6959E+08,
     + 6.7496E+08, 6.8038E+08, 6.8583E+08, 6.9132E+08, 6.9686E+08,
     +  7.0243E+08, 7.0804E+08, 7.1370E+08, 7.1939E+08, 7.2513E+08,
     + 7.3091E+08, 7.3673E+08, 7.4259E+08, 7.4849E+08, 7.5444E+08,
     +  7.6043E+08, 7.6646E+08, 7.7254E+08, 7.7866E+08, 7.8482E+08,
     + 7.9103E+08, 7.9729E+08, 8.0359E+08, 8.0993E+08, 8.1632E+08,
     +  8.2275E+08, 8.2924E+08, 8.3576E+08, 8.4234E+08, 8.4896E+08,
     + 8.5563E+08, 8.6234E+08, 8.6911E+08, 8.7592E+08, 8.8278E+08,
     +  8.8969E+08, 8.9665E+08, 9.0366E+08, 9.1072E+08, 9.1783E+08,
     + 9.2498E+08, 9.3219E+08, 9.3945E+08, 9.4677E+08, 9.5413E+08,
     +  9.6155E+08, 9.6901E+08, 9.7653E+08, 9.8411E+08, 9.9174E+08,
     + 9.9942E+08, 1.0072E+09, 1.0149E+09, 1.0228E+09, 1.0307E+09,
     +  1.0386E+09, 1.0466E+09, 1.0547E+09, 1.0628E+09, 1.0710E+09,
     + 1.0793E+09, 1.0876E+09, 1.0959E+09, 1.1043E+09, 1.1128E+09,
     +  1.1213E+09, 1.1299E+09, 1.1385E+09, 1.1473E+09, 1.1560E+09,
     + 1.1649E+09, 1.1737E+09, 1.1827E+09, 1.1917E+09, 1.2008E+09,
     +  1.2099E+09, 1.2191E+09, 1.2284E+09, 1.2377E+09, 1.2471E+09,
     + 1.2566E+09, 1.2661E+09, 1.2757E+09, 1.2854E+09, 1.2951E+09,
     +  1.3049E+09, 1.3147E+09, 1.3246E+09, 1.3346E+09, 1.3447E+09,
     + 1.3548E+09, 1.3650E+09, 1.3753E+09, 1.3856E+09, 1.3960E+09,
     +  1.4065E+09, 1.4171E+09, 1.4277E+09, 1.4384E+09, 1.4492E+09,
     + 1.4600E+09, 1.4709E+09, 1.4819E+09, 1.4930E+09, 1.5041E+09,
     +  1.5153E+09, 1.5266E+09, 1.5380E+09, 1.5494E+09, 1.5610E+09,
     + 1.5726E+09, 1.5842E+09, 1.5960E+09, 1.6078E+09, 1.6197E+09,
     +  1.6317E+09, 1.6438E+09, 1.6560E+09, 1.6682E+09, 1.6805E+09,
     + 1.6929E+09, 1.7054E+09, 1.7180E+09, 1.7307E+09, 1.7434E+09,
     +  1.7562E+09, 1.7691E+09, 1.7821E+09, 1.7952E+09, 1.8084E+09,
     + 1.8216E+09, 1.8350E+09, 1.8484E+09, 1.8620E+09, 1.8756E+09,
     +  1.8893E+09, 1.9031E+09, 1.9170E+09, 1.9309E+09, 1.9450E+09,
     + 1.9592E+09, 1.9734E+09, 1.9878E+09, 2.0022E+09, 2.0168E+09,
     +  2.0314E+09, 2.0462E+09, 2.0610E+09, 2.0759E+09, 2.0910E+09,
     + 2.1061E+09, 2.1213E+09, 2.1366E+09, 2.1521E+09, 2.1676E+09,
     +  2.1832E+09, 2.1990E+09, 2.2148E+09, 2.2307E+09, 2.2468E+09,
     + 2.2629E+09, 2.2792E+09, 2.2956E+09, 2.3120E+09, 2.3286E+09,
     +  2.3453E+09, 2.3621E+09, 2.3790E+09, 2.3960E+09, 2.4131E+09,
     + 2.4303E+09, 2.4477E+09, 2.4651E+09, 2.4827E+09, 2.5004E+09,
     +  2.5182E+09, 2.5361E+09, 2.5541E+09, 2.5723E+09, 2.5905E+09,
     + 2.6089E+09, 2.6274E+09, 2.6460E+09, 2.6648E+09, 2.6836E+09,
     +  2.7026E+09, 2.7217E+09, 2.7409E+09, 2.7603E+09, 2.7797E+09,
     + 2.7993E+09, 2.8190E+09, 2.8389E+09, 2.8589E+09, 2.8790E+09,
     +  2.8992E+09, 2.9195E+09, 2.9400E+09, 2.9606E+09, 2.9814E+09,
     + 3.0023E+09, 3.0233E+09, 3.0444E+09, 3.0657E+09, 3.0871E+09,
     +  3.1087E+09, 3.1304E+09, 3.1522E+09, 3.1742E+09, 3.1963E+09,
     + 3.2185E+09, 3.2409E+09, 3.2634E+09, 3.2861E+09, 3.3089E+09,
     +  3.3318E+09, 3.3549E+09, 3.3782E+09, 3.4016E+09, 3.4251E+09,
     + 3.4488E+09, 3.4726E+09, 3.4966E+09, 3.5208E+09, 3.5450E+09,
     +  3.5695E+09, 3.5941E+09, 3.6188E+09, 3.6437E+09, 3.6688E+09,
     + 3.6940E+09, 3.7193E+09, 3.7449E+09, 3.7705E+09, 3.7964E+09,
     +  3.8224E+09, 3.8486E+09, 3.8749E+09, 3.9014E+09, 3.9280E+09,
     + 3.9549E+09, 3.9819E+09, 4.0090E+09, 4.0363E+09, 4.0638E+09,
     +  4.0915E+09, 4.1193E+09, 4.1473E+09, 4.1755E+09, 4.2039E+09,
     + 4.2324E+09, 4.2611E+09, 4.2900E+09, 4.3190E+09, 4.3483E+09,
     +  4.3777E+09, 4.4073E+09, 4.4371E+09, 4.4670E+09, 4.4972E+09,
     + 4.5275E+09, 4.5580E+09, 4.5887E+09, 4.6196E+09, 4.6507E+09,
     +  4.6820E+09, 4.7134E+09, 4.7451E+09, 4.7769E+09, 4.8089E+09,
     + 4.8412E+09, 4.8736E+09, 4.9062E+09, 4.9391E+09, 4.9721E+09,
     +  5.0053E+09, 5.0387E+09, 5.0724E+09, 5.1062E+09, 5.1402E+09,
     + 5.1745E+09, 5.2089E+09, 5.2436E+09, 5.2784E+09, 5.3135E+09,
     +  5.3488E+09, 5.3843E+09, 5.4200E+09, 5.4560E+09, 5.4921E+09,
     + 5.5285E+09, 5.5650E+09, 5.6018E+09, 5.6389E+09, 5.6761E+09,
     +  5.7136E+09, 5.7513E+09, 5.7892E+09, 5.8273E+09, 5.8657E+09,
     + 5.9043E+09, 5.9431E+09, 5.9822E+09, 6.0215E+09, 6.0610E+09,
     +  6.1008E+09, 6.1408E+09, 6.1811E+09, 6.2215E+09, 6.2623E+09,
     + 6.3032E+09, 6.3444E+09, 6.3859E+09, 6.4276E+09, 6.4695E+09,
     +  6.5117E+09, 6.5542E+09, 6.5969E+09, 6.6398E+09, 6.6830E+09,
     + 6.7265E+09, 6.7702E+09, 6.8142E+09, 6.8584E+09, 6.9029E+09,
     +  6.9476E+09, 6.9926E+09, 7.0379E+09, 7.0835E+09, 7.1293E+09,
     + 7.1754E+09, 7.2217E+09, 7.2683E+09, 7.3152E+09, 7.3624E+09,
     +  7.4099E+09, 7.4576E+09, 7.5056E+09, 7.5539E+09, 7.6024E+09,
     + 7.6513E+09, 7.7004E+09, 7.7498E+09, 7.7995E+09, 7.8495E+09,
     +  7.8998E+09, 7.9504E+09, 8.0013E+09, 8.0525E+09, 8.1039E+09,
     + 8.1557E+09, 8.2078E+09, 8.2601E+09, 8.3128E+09, 8.3658E+09,
     +  8.4191E+09, 8.4726E+09, 8.5266E+09, 8.5808E+09, 8.6353E+09,
     + 8.6901E+09, 8.7453E+09, 8.8007E+09, 8.8565E+09, 8.9127E+09,
     +  8.9691E+09, 9.0258E+09, 9.0829E+09, 9.1403E+09, 9.1981E+09,
     + 9.2562E+09, 9.3146E+09, 9.3733E+09, 9.4324E+09, 9.4918E+09,
     +  9.5515E+09, 9.6116E+09, 9.6721E+09, 9.7329E+09, 9.7940E+09,
     + 9.8555E+09, 9.9173E+09, 9.9795E+09, 1.0042E+10, 1.0105E+10,
     +  1.0168E+10, 1.0232E+10, 1.0296E+10, 1.0360E+10, 1.0425E+10,
     + 1.0490E+10, 1.0555E+10, 1.0621E+10, 1.0687E+10, 1.0754E+10,
     +  1.0821E+10, 1.0888E+10, 1.0956E+10, 1.1024E+10, 1.1092E+10,
     + 1.1161E+10, 1.1230E+10, 1.1300E+10, 1.1370E+10, 1.1440E+10,
     +  1.1511E+10, 1.1582E+10, 1.1654E+10, 1.1726E+10, 1.1798E+10,
     + 1.1871E+10, 1.1944E+10, 1.2018E+10, 1.2092E+10, 1.2166E+10,
     +  1.2241E+10, 1.2316E+10, 1.2392E+10, 1.2468E+10, 1.2545E+10,
     + 1.2622E+10, 1.2699E+10, 1.2777E+10, 1.2855E+10, 1.2934E+10,
     +  1.3013E+10, 1.3092E+10, 1.3172E+10, 1.3253E+10, 1.3334E+10,
     + 1.3415E+10, 1.3497E+10, 1.3579E+10, 1.3662E+10, 1.3745E+10,
     +  1.3829E+10, 1.3913E+10, 1.3997E+10, 1.4082E+10, 1.4168E+10,
     + 1.4254E+10, 1.4340E+10, 1.4427E+10, 1.4514E+10, 1.4602E+10,
     +  1.4691E+10, 1.4779E+10, 1.4869E+10, 1.4958E+10, 1.5049E+10,
     + 1.5139E+10, 1.5231E+10, 1.5323E+10, 1.5415E+10, 1.5508E+10,
     +  1.5601E+10, 1.5695E+10, 1.5789E+10, 1.5884E+10, 1.5979E+10,
     + 1.6075E+10, 1.6171E+10, 1.6268E+10, 1.6366E+10, 1.6464E+10,
     +  1.6562E+10, 1.6661E+10, 1.6761E+10, 1.6861E+10, 1.6962E+10,
     + 1.7063E+10, 1.7165E+10, 1.7267E+10, 1.7370E+10, 1.7473E+10,
     +  1.7577E+10, 1.7682E+10, 1.7787E+10, 1.7892E+10, 1.7998E+10,
     + 1.8105E+10, 1.8213E+10, 1.8321E+10, 1.8429E+10, 1.8538E+10,
     +  1.8648E+10, 1.8758E+10, 1.8869E+10, 1.8981E+10, 1.9093E+10,
     + 1.9205E+10, 1.9319E+10, 1.9432E+10, 1.9547E+10, 1.9662E+10,
     +  1.9778E+10, 1.9894E+10, 2.0011E+10, 2.0129E+10, 2.0247E+10,
     + 2.0366E+10, 2.0485E+10, 2.0605E+10, 2.0726E+10, 2.0847E+10,
     +  2.0969E+10, 2.1092E+10, 2.1215E+10, 2.1339E+10, 2.1464E+10,
     + 2.1589E+10, 2.1715E+10, 2.1842E+10, 2.1969E+10, 2.2097E+10,
     +  2.2226E+10, 2.2355E+10, 2.2485E+10, 2.2616E+10, 2.2747E+10,
     + 2.2879E+10, 2.3012E+10, 2.3145E+10, 2.3279E+10, 2.3414E+10,
     +  2.3550E+10, 2.3686E+10, 2.3823E+10, 2.3961E+10, 2.4099E+10,
     + 2.4238E+10, 2.4378E+10, 2.4519E+10, 2.4660E+10, 2.4803E+10,
     +  2.4945E+10, 2.5089E+10, 2.5233E+10, 2.5378E+10, 2.5524E+10,
     + 2.5671E+10, 2.5818E+10, 2.5966E+10, 2.6115E+10, 2.6265E+10,
     +  2.6415E+10, 2.6567E+10, 2.6719E+10, 2.6872E+10, 2.7025E+10,
     + 2.7180E+10, 2.7335E+10, 2.7491E+10, 2.7648E+10, 2.7805E+10,
     +  2.7964E+10, 2.8123E+10, 2.8283E+10, 2.8444E+10, 2.8606E+10,
     + 2.8768E+10, 2.8932E+10, 2.9096E+10, 2.9261E+10, 2.9427E+10,
     +  2.9594E+10, 2.9762E+10, 2.9930E+10, 3.0100E+10, 3.0270E+10,
     + 3.0441E+10, 3.0613E+10, 3.0786E+10, 3.0960E+10, 3.1135E+10,
     +  3.1310E+10, 3.1487E+10, 3.1664E+10, 3.1842E+10, 3.2021E+10,
     + 3.2202E+10, 3.2383E+10, 3.2564E+10, 3.2747E+10, 3.2931E+10,
     +  3.3116E+10, 3.3302E+10, 3.3488E+10, 3.3676E+10, 3.3864E+10,
     + 3.4054E+10, 3.4244E+10, 3.4436E+10, 3.4628E+10, 3.4821E+10,
     +  3.5016E+10, 3.5211E+10, 3.5407E+10, 3.5605E+10, 3.5803E+10,
     + 3.6002E+10, 3.6203E+10, 3.6404E+10, 3.6606E+10, 3.6810E+10,
     +  3.7014E+10, 3.7219E+10, 3.7426E+10, 3.7633E+10, 3.7842E+10,
     + 3.8051E+10, 3.8262E+10, 3.8474E+10, 3.8686E+10, 3.8900E+10,
     +  3.9115E+10, 3.9331E+10, 3.9548E+10, 3.9766E+10, 3.9985E+10,
     + 4.0206E+10, 4.0427E+10, 4.0650E+10, 4.0873E+10, 4.1098E+10,
     +  4.1324E+10, 4.1551E+10, 4.1779E+10, 4.2008E+10, 4.2239E+10,
     + 4.2470E+10, 4.2703E+10, 4.2937E+10, 4.3172E+10, 4.3408E+10,
     +  4.3645E+10, 4.3884E+10, 4.4123E+10, 4.4364E+10, 4.4606E+10,
     + 4.4850E+10, 4.5094E+10, 4.5340E+10, 4.5587E+10, 4.5835E+10,
     +  4.6084E+10, 4.6335E+10, 4.6587E+10, 4.6840E+10, 4.7094E+10,
     + 4.7350E+10, 4.7606E+10, 4.7865E+10, 4.8124E+10, 4.8384E+10,
     +  4.8646E+10, 4.8910E+10, 4.9174E+10, 4.9440E+10, 4.9707E+10,
     + 4.9975E+10, 5.0245E+10, 5.0516E+10, 5.0788E+10, 5.1062E+10,
     +  5.1337E+10, 5.1613E+10, 5.1891E+10, 5.2170E+10, 5.2451E+10,
     + 5.2732E+10, 5.3016E+10, 5.3300E+10, 5.3586E+10, 5.3873E+10,
     +  5.4162E+10, 5.4452E+10, 5.4744E+10, 5.5037E+10, 5.5331E+10,
     + 5.5627E+10, 5.5924E+10, 5.6223E+10, 5.6523E+10, 5.6824E+10,
     +  5.7127E+10, 5.7432E+10, 5.7738E+10, 5.8045E+10, 5.8354E+10,
     + 5.8665E+10, 5.8976E+10, 5.9290E+10, 5.9605E+10, 5.9921E+10,
     +  6.0239E+10, 6.0559E+10, 6.0880E+10, 6.1202E+10, 6.1526E+10,
     + 6.1852E+10, 6.2179E+10, 6.2508E+10, 6.2838E+10, 6.3170E+10,
     +  6.3504E+10, 6.3839E+10, 6.4176E+10, 6.4514E+10, 6.4854E+10,
     + 6.5196E+10, 6.5539E+10, 6.5884E+10, 6.6231E+10, 6.6579E+10,
     +  6.6929E+10, 6.7280E+10, 6.7633E+10, 6.7988E+10, 6.8345E+10,
     + 6.8703E+10, 6.9063E+10, 6.9425E+10, 6.9788E+10, 7.0153E+10,
     +  7.0520E+10, 7.0888E+10, 7.1259E+10, 7.1631E+10, 7.2005E+10,
     + 7.2380E+10, 7.2758E+10, 7.3137E+10, 7.3518E+10, 7.3900E+10,
     +  7.4285E+10, 7.4671E+10, 7.5060E+10, 7.5450E+10, 7.5842E+10,
     + 7.6235E+10, 7.6631E+10, 7.7028E+10, 7.7428E+10, 7.7829E+10,
     +  7.8232E+10, 7.8637E+10, 7.9044E+10, 7.9452E+10, 7.9863E+10,
     + 8.0276E+10, 8.0690E+10, 8.1107E+10, 8.1525E+10, 8.1945E+10,
     +  8.2368E+10, 8.2792E+10, 8.3218E+10, 8.3647E+10, 8.4077E+10,
     + 8.4509E+10, 8.4943E+10, 8.5380E+10, 8.5818E+10, 8.6259E+10,
     +  8.6701E+10, 8.7145E+10, 8.7592E+10, 8.8041E+10, 8.8491E+10,
     + 8.8944E+10, 8.9399E+10, 8.9856E+10, 9.0315E+10, 9.0777E+10,
     +  9.1240E+10, 9.1706E+10, 9.2173E+10, 9.2643E+10, 9.3115E+10,
     + 9.3590E+10, 9.4066E+10, 9.4545E+10, 9.5025E+10, 9.5508E+10,
     +  9.5994E+10, 9.6481E+10, 9.6971E+10, 9.7463E+10, 9.7957E+10,
     + 9.8454E+10, 9.8953E+10, 9.9454E+10, 9.9957E+10, 1.0046E+11,
     +  1.0097E+11, 1.0148E+11, 1.0199E+11, 1.0251E+11, 1.0303E+11,
     + 1.0355E+11, 1.0407E+11, 1.0459E+11, 1.0512E+11, 1.0565E+11,
     +  1.0618E+11, 1.0672E+11, 1.0725E+11, 1.0779E+11, 1.0833E+11,
     + 1.0888E+11, 1.0942E+11, 1.0997E+11, 1.1052E+11, 1.1108E+11,
     +  1.1163E+11, 1.1219E+11, 1.1275E+11, 1.1332E+11, 1.1389E+11,
     + 1.1445E+11, 1.1503E+11, 1.1560E+11, 1.1618E+11, 1.1676E+11,
     +  1.1734E+11, 1.1792E+11, 1.1851E+11, 1.1910E+11, 1.1969E+11,
     + 1.2029E+11, 1.2089E+11, 1.2149E+11, 1.2209E+11, 1.2270E+11,
     +  1.2331E+11, 1.2392E+11, 1.2453E+11, 1.2515E+11, 1.2577E+11,
     + 1.2639E+11, 1.2702E+11, 1.2765E+11, 1.2828E+11, 1.2891E+11,
     +  1.2955E+11, 1.3019E+11, 1.3083E+11, 1.3148E+11, 1.3213E+11,
     + 1.3278E+11, 1.3343E+11, 1.3409E+11, 1.3475E+11, 1.3541E+11,
     +  1.3608E+11, 1.3675E+11, 1.3742E+11, 1.3810E+11, 1.3877E+11,
     + 1.3945E+11, 1.4014E+11, 1.4083E+11, 1.4152E+11, 1.4221E+11,
     +  1.4290E+11, 1.4360E+11, 1.4431E+11, 1.4501E+11, 1.4572E+11,
     + 1.4643E+11, 1.4715E+11, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       156
      data (QofT( 2,J),J=1,2408)/ 2.3826E+01, 6.3940E+01, 1.1546E+02,
     +  1.7626E+02, 2.4508E+02, 3.2108E+02, 4.0364E+02, 4.9227E+02,
     + 5.8658E+02, 6.8625E+02, 7.9100E+02, 9.0060E+02, 1.0148E+03,
     +  1.1335E+03, 1.2565E+03, 1.3837E+03, 1.5149E+03, 1.6500E+03,
     + 1.7889E+03, 1.9314E+03, 2.2273E+03, 2.5369E+03, 2.8597E+03,
     +  3.1951E+03, 3.5427E+03, 3.9021E+03, 4.2728E+03, 4.6547E+03,
     + 5.0473E+03, 5.4503E+03, 5.8636E+03, 6.2868E+03, 6.7197E+03,
     +  7.1622E+03, 7.6140E+03, 8.0749E+03, 8.5448E+03, 9.0235E+03,
     + 9.5108E+03, 1.0007E+04, 1.0511E+04, 1.1023E+04, 1.1544E+04,
     +  1.2073E+04, 1.2609E+04, 1.3153E+04, 1.3705E+04, 1.4265E+04,
     + 1.4832E+04, 1.5407E+04, 1.5990E+04, 1.6579E+04, 1.7177E+04,
     +  1.7781E+04, 1.8393E+04, 1.9012E+04, 1.9639E+04, 2.0272E+04,
     + 2.0914E+04, 2.1562E+04, 2.2218E+04, 2.2881E+04, 2.3552E+04,
     +  2.4230E+04, 2.4915E+04, 2.5608E+04, 2.6309E+04, 2.7017E+04,
     + 2.7733E+04, 2.8457E+04, 2.9189E+04, 2.9929E+04, 3.0677E+04,
     +  3.1433E+04, 3.2197E+04, 3.2970E+04, 3.3751E+04, 3.4541E+04,
     + 3.5340E+04, 3.6147E+04, 3.6964E+04, 3.7790E+04, 3.8625E+04,
     +  3.9469E+04, 4.0323E+04, 4.1187E+04, 4.2061E+04, 4.2945E+04,
     + 4.3839E+04, 4.4744E+04, 4.5659E+04, 4.6585E+04, 4.7522E+04,
     +  4.8470E+04, 4.9429E+04, 5.0400E+04, 5.1382E+04, 5.2377E+04,
     + 5.3383E+04, 5.4402E+04, 5.5434E+04, 5.6478E+04, 5.7535E+04,
     +  5.8605E+04, 5.9688E+04, 6.0785E+04, 6.1896E+04, 6.3021E+04,
     + 6.4161E+04, 6.5314E+04, 6.6483E+04, 6.7666E+04, 6.8865E+04,
     +  7.0079E+04, 7.1309E+04, 7.2555E+04, 7.3817E+04, 7.5096E+04,
     + 7.6391E+04, 7.7704E+04, 7.9033E+04, 8.0380E+04, 8.1745E+04,
     +  8.3128E+04, 8.4530E+04, 8.5950E+04, 8.7389E+04, 8.8848E+04,
     + 9.0325E+04, 9.1823E+04, 9.3341E+04, 9.4879E+04, 9.6438E+04,
     +  9.8018E+04, 9.9620E+04, 1.0124E+05, 1.0289E+05, 1.0456E+05,
     + 1.0625E+05, 1.0796E+05, 1.0969E+05, 1.1145E+05, 1.1324E+05,
     +  1.1505E+05, 1.1688E+05, 1.1874E+05, 1.2062E+05, 1.2253E+05,
     + 1.2447E+05, 1.2643E+05, 1.2842E+05, 1.3044E+05, 1.3248E+05,
     +  1.3455E+05, 1.3665E+05, 1.3878E+05, 1.4094E+05, 1.4313E+05,
     + 1.4535E+05, 1.4760E+05, 1.4988E+05, 1.5219E+05, 1.5453E+05,
     +  1.5691E+05, 1.5932E+05, 1.6176E+05, 1.6423E+05, 1.6674E+05,
     + 1.6928E+05, 1.7186E+05, 1.7448E+05, 1.7713E+05, 1.7981E+05,
     +  1.8254E+05, 1.8530E+05, 1.8810E+05, 1.9093E+05, 1.9381E+05,
     + 1.9673E+05, 1.9968E+05, 2.0268E+05, 2.0571E+05, 2.0879E+05,
     +  2.1192E+05, 2.1508E+05, 2.1829E+05, 2.2154E+05, 2.2483E+05,
     + 2.2817E+05, 2.3156E+05, 2.3499E+05, 2.3847E+05, 2.4200E+05,
     +  2.4558E+05, 2.4920E+05, 2.5287E+05, 2.5660E+05, 2.6037E+05,
     + 2.6420E+05, 2.6807E+05, 2.7200E+05, 2.7599E+05, 2.8003E+05,
     +  2.8412E+05, 2.8827E+05, 2.9247E+05, 2.9673E+05, 3.0105E+05,
     + 3.0543E+05, 3.0986E+05, 3.1436E+05, 3.1891E+05, 3.2353E+05,
     +  3.2821E+05, 3.3296E+05, 3.3776E+05, 3.4263E+05, 3.4757E+05,
     + 3.5257E+05, 3.5764E+05, 3.6278E+05, 3.6799E+05, 3.7326E+05,
     +  3.7861E+05, 3.8403E+05, 3.8952E+05, 3.9508E+05, 4.0072E+05,
     + 4.0643E+05, 4.1222E+05, 4.1809E+05, 4.2403E+05, 4.3005E+05,
     +  4.3616E+05, 4.4234E+05, 4.4860E+05, 4.5495E+05, 4.6138E+05,
     + 4.6790E+05, 4.7450E+05, 4.8119E+05, 4.8797E+05, 4.9483E+05,
     +  5.0179E+05, 5.0884E+05, 5.1598E+05, 5.2321E+05, 5.3054E+05,
     + 5.3796E+05, 5.4548E+05, 5.5310E+05, 5.6082E+05, 5.6864E+05,
     +  5.7656E+05, 5.8458E+05, 5.9271E+05, 6.0095E+05, 6.0929E+05,
     + 6.1773E+05, 6.2629E+05, 6.3496E+05, 6.4374E+05, 6.5263E+05,
     +  6.6164E+05, 6.7076E+05, 6.8001E+05, 6.8937E+05, 6.9885E+05,
     + 7.0845E+05, 7.1817E+05, 7.2802E+05, 7.3800E+05, 7.4810E+05,
     +  7.5834E+05, 7.6870E+05, 7.7919E+05, 7.8982E+05, 8.0058E+05,
     + 8.1148E+05, 8.2252E+05, 8.3370E+05, 8.4502E+05, 8.5649E+05,
     +  8.6809E+05, 8.7985E+05, 8.9175E+05, 9.0381E+05, 9.1601E+05,
     + 9.2837E+05, 9.4088E+05, 9.5355E+05, 9.6638E+05, 9.7937E+05,
     +  9.9253E+05, 1.0058E+06, 1.0193E+06, 1.0330E+06, 1.0468E+06,
     + 1.0608E+06, 1.0750E+06, 1.0893E+06, 1.1038E+06, 1.1185E+06,
     +  1.1334E+06, 1.1485E+06, 1.1637E+06, 1.1792E+06, 1.1948E+06,
     + 1.2107E+06, 1.2267E+06, 1.2429E+06, 1.2593E+06, 1.2759E+06,
     +  1.2928E+06, 1.3098E+06, 1.3270E+06, 1.3445E+06, 1.3621E+06,
     + 1.3800E+06, 1.3981E+06, 1.4164E+06, 1.4350E+06, 1.4537E+06,
     +  1.4727E+06, 1.4920E+06, 1.5114E+06, 1.5311E+06, 1.5510E+06,
     + 1.5712E+06, 1.5916E+06, 1.6123E+06, 1.6332E+06, 1.6543E+06,
     +  1.6758E+06, 1.6974E+06, 1.7194E+06, 1.7416E+06, 1.7640E+06,
     + 1.7867E+06, 1.8097E+06, 1.8330E+06, 1.8566E+06, 1.8804E+06,
     +  1.9045E+06, 1.9289E+06, 1.9536E+06, 1.9786E+06, 2.0039E+06,
     + 2.0294E+06, 2.0553E+06, 2.0815E+06, 2.1080E+06, 2.1348E+06,
     +  2.1619E+06, 2.1894E+06, 2.2171E+06, 2.2452E+06, 2.2737E+06,
     + 2.3024E+06, 2.3315E+06, 2.3609E+06, 2.3907E+06, 2.4208E+06,
     +  2.4513E+06, 2.4821E+06, 2.5133E+06, 2.5448E+06, 2.5768E+06,
     + 2.6090E+06, 2.6417E+06, 2.6747E+06, 2.7081E+06, 2.7419E+06,
     +  2.7761E+06, 2.8107E+06, 2.8457E+06, 2.8811E+06, 2.9169E+06,
     + 2.9531E+06, 2.9897E+06, 3.0267E+06, 3.0642E+06, 3.1021E+06,
     +  3.1404E+06, 3.1791E+06, 3.2183E+06, 3.2580E+06, 3.2981E+06,
     + 3.3386E+06, 3.3796E+06, 3.4211E+06, 3.4630E+06, 3.5054E+06,
     +  3.5483E+06, 3.5917E+06, 3.6356E+06, 3.6800E+06, 3.7248E+06,
     + 3.7702E+06, 3.8161E+06, 3.8624E+06, 3.9094E+06, 3.9568E+06,
     +  4.0048E+06, 4.0533E+06, 4.1023E+06, 4.1519E+06, 4.2020E+06,
     + 4.2527E+06, 4.3040E+06, 4.3558E+06, 4.4082E+06, 4.4612E+06,
     +  4.5148E+06, 4.5689E+06, 4.6237E+06, 4.6791E+06, 4.7351E+06,
     + 4.7916E+06, 4.8489E+06, 4.9067E+06, 4.9652E+06, 5.0243E+06,
     +  5.0841E+06, 5.1445E+06, 5.2056E+06, 5.2673E+06, 5.3298E+06,
     + 5.3929E+06, 5.4567E+06, 5.5212E+06, 5.5863E+06, 5.6522E+06,
     +  5.7189E+06, 5.7862E+06, 5.8542E+06, 5.9230E+06, 5.9926E+06,
     + 6.0629E+06, 6.1339E+06, 6.2057E+06, 6.2783E+06, 6.3517E+06,
     +  6.4259E+06, 6.5008E+06, 6.5766E+06, 6.6531E+06, 6.7305E+06,
     + 6.8087E+06, 6.8878E+06, 6.9677E+06, 7.0484E+06, 7.1300E+06,
     +  7.2125E+06, 7.2958E+06, 7.3800E+06, 7.4652E+06, 7.5512E+06,
     + 7.6381E+06, 7.7260E+06, 7.8147E+06, 7.9044E+06, 7.9951E+06,
     +  8.0867E+06, 8.1793E+06, 8.2728E+06, 8.3674E+06, 8.4629E+06,
     + 8.5594E+06, 8.6570E+06, 8.7555E+06, 8.8551E+06, 8.9557E+06,
     +  9.0574E+06, 9.1601E+06, 9.2639E+06, 9.3688E+06, 9.4748E+06,
     + 9.5818E+06, 9.6900E+06, 9.7993E+06, 9.9097E+06, 1.0021E+07,
     +  1.0134E+07, 1.0248E+07, 1.0363E+07, 1.0479E+07, 1.0597E+07,
     + 1.0715E+07, 1.0835E+07, 1.0956E+07, 1.1078E+07, 1.1202E+07,
     +  1.1327E+07, 1.1453E+07, 1.1580E+07, 1.1709E+07, 1.1839E+07,
     + 1.1970E+07, 1.2103E+07, 1.2237E+07, 1.2372E+07, 1.2509E+07,
     +  1.2647E+07, 1.2787E+07, 1.2928E+07, 1.3070E+07, 1.3214E+07,
     + 1.3359E+07, 1.3506E+07, 1.3654E+07, 1.3804E+07, 1.3955E+07,
     +  1.4108E+07, 1.4262E+07, 1.4418E+07, 1.4575E+07, 1.4734E+07,
     + 1.4894E+07, 1.5056E+07, 1.5220E+07, 1.5385E+07, 1.5552E+07,
     +  1.5721E+07, 1.5891E+07, 1.6063E+07, 1.6237E+07, 1.6412E+07,
     + 1.6589E+07, 1.6768E+07, 1.6949E+07, 1.7131E+07, 1.7316E+07,
     +  1.7502E+07, 1.7689E+07, 1.7879E+07, 1.8071E+07, 1.8264E+07,
     + 1.8460E+07, 1.8657E+07, 1.8856E+07, 1.9057E+07, 1.9260E+07,
     +  1.9465E+07, 1.9672E+07, 1.9881E+07, 2.0092E+07, 2.0306E+07,
     + 2.0521E+07, 2.0738E+07, 2.0958E+07, 2.1179E+07, 2.1403E+07,
     +  2.1629E+07, 2.1857E+07, 2.2087E+07, 2.2319E+07, 2.2554E+07,
     + 2.2791E+07, 2.3030E+07, 2.3272E+07, 2.3515E+07, 2.3762E+07,
     +  2.4010E+07, 2.4261E+07, 2.4514E+07, 2.4770E+07, 2.5028E+07,
     + 2.5288E+07, 2.5552E+07, 2.5817E+07, 2.6085E+07, 2.6356E+07,
     +  2.6629E+07, 2.6905E+07, 2.7183E+07, 2.7464E+07, 2.7748E+07,
     + 2.8034E+07, 2.8323E+07, 2.8615E+07, 2.8909E+07, 2.9207E+07,
     +  2.9507E+07, 2.9810E+07, 3.0115E+07, 3.0424E+07, 3.0735E+07,
     + 3.1050E+07, 3.1367E+07, 3.1687E+07, 3.2011E+07, 3.2337E+07,
     +  3.2666E+07, 3.2999E+07, 3.3334E+07, 3.3673E+07, 3.4014E+07,
     + 3.4359E+07, 3.4707E+07, 3.5059E+07, 3.5413E+07, 3.5771E+07,
     +  3.6132E+07, 3.6496E+07, 3.6864E+07, 3.7235E+07, 3.7610E+07,
     + 3.7988E+07, 3.8370E+07, 3.8755E+07, 3.9143E+07, 3.9535E+07,
     +  3.9931E+07, 4.0330E+07, 4.0733E+07, 4.1139E+07, 4.1550E+07,
     + 4.1964E+07, 4.2382E+07, 4.2803E+07, 4.3229E+07, 4.3658E+07,
     +  4.4091E+07, 4.4528E+07, 4.4969E+07, 4.5414E+07, 4.5863E+07,
     + 4.6316E+07, 4.6773E+07, 4.7235E+07, 4.7700E+07, 4.8170E+07,
     +  4.8643E+07, 4.9122E+07, 4.9604E+07, 5.0091E+07, 5.0582E+07,
     + 5.1077E+07, 5.1577E+07, 5.2081E+07, 5.2590E+07, 5.3103E+07,
     +  5.3621E+07, 5.4144E+07, 5.4671E+07, 5.5203E+07, 5.5739E+07,
     + 5.6281E+07, 5.6827E+07, 5.7378E+07, 5.7934E+07, 5.8494E+07,
     +  5.9060E+07, 5.9631E+07, 6.0206E+07, 6.0787E+07, 6.1373E+07,
     + 6.1964E+07, 6.2560E+07, 6.3161E+07, 6.3768E+07, 6.4380E+07,
     +  6.4998E+07, 6.5620E+07, 6.6248E+07, 6.6882E+07, 6.7521E+07,
     + 6.8166E+07, 6.8816E+07, 6.9473E+07, 7.0134E+07, 7.0802E+07,
     +  7.1475E+07, 7.2154E+07, 7.2839E+07, 7.3530E+07, 7.4227E+07,
     + 7.4930E+07, 7.5639E+07, 7.6354E+07, 7.7075E+07, 7.7803E+07,
     +  7.8536E+07, 7.9276E+07, 8.0023E+07, 8.0775E+07, 8.1535E+07,
     + 8.2300E+07, 8.3073E+07, 8.3852E+07, 8.4637E+07, 8.5429E+07,
     +  8.6228E+07, 8.7034E+07, 8.7847E+07, 8.8666E+07, 8.9493E+07,
     + 9.0326E+07, 9.1167E+07, 9.2015E+07, 9.2870E+07, 9.3732E+07,
     +  9.4601E+07, 9.5478E+07, 9.6362E+07, 9.7254E+07, 9.8153E+07,
     + 9.9060E+07, 9.9974E+07, 1.0090E+08, 1.0183E+08, 1.0276E+08,
     +  1.0371E+08, 1.0466E+08, 1.0562E+08, 1.0659E+08, 1.0757E+08,
     + 1.0856E+08, 1.0955E+08, 1.1055E+08, 1.1156E+08, 1.1258E+08,
     +  1.1361E+08, 1.1465E+08, 1.1569E+08, 1.1674E+08, 1.1781E+08,
     + 1.1888E+08, 1.1996E+08, 1.2104E+08, 1.2214E+08, 1.2325E+08,
     +  1.2436E+08, 1.2549E+08, 1.2662E+08, 1.2777E+08, 1.2892E+08,
     + 1.3008E+08, 1.3126E+08, 1.3244E+08, 1.3363E+08, 1.3483E+08,
     +  1.3604E+08, 1.3726E+08, 1.3849E+08, 1.3973E+08, 1.4098E+08,
     + 1.4224E+08, 1.4352E+08, 1.4480E+08, 1.4609E+08, 1.4739E+08,
     +  1.4871E+08, 1.5003E+08, 1.5136E+08, 1.5271E+08, 1.5407E+08,
     + 1.5543E+08, 1.5681E+08, 1.5820E+08, 1.5960E+08, 1.6101E+08,
     +  1.6244E+08, 1.6387E+08, 1.6532E+08, 1.6677E+08, 1.6824E+08,
     + 1.6973E+08, 1.7122E+08, 1.7272E+08, 1.7424E+08, 1.7577E+08,
     +  1.7731E+08, 1.7887E+08, 1.8043E+08, 1.8201E+08, 1.8360E+08,
     + 1.8521E+08, 1.8682E+08, 1.8845E+08, 1.9009E+08, 1.9175E+08,
     +  1.9342E+08, 1.9510E+08, 1.9680E+08, 1.9850E+08, 2.0023E+08,
     + 2.0196E+08, 2.0371E+08, 2.0547E+08, 2.0725E+08, 2.0904E+08,
     +  2.1085E+08, 2.1267E+08, 2.1450E+08, 2.1635E+08, 2.1821E+08,
     + 2.2009E+08, 2.2198E+08, 2.2389E+08, 2.2581E+08, 2.2774E+08,
     +  2.2969E+08, 2.3166E+08, 2.3364E+08, 2.3564E+08, 2.3765E+08,
     + 2.3968E+08, 2.4173E+08, 2.4379E+08, 2.4586E+08, 2.4795E+08,
     +  2.5006E+08, 2.5219E+08, 2.5433E+08, 2.5649E+08, 2.5866E+08,
     + 2.6085E+08, 2.6306E+08, 2.6528E+08, 2.6753E+08, 2.6979E+08,
     +  2.7206E+08, 2.7436E+08, 2.7667E+08, 2.7900E+08, 2.8135E+08,
     + 2.8371E+08, 2.8609E+08, 2.8850E+08, 2.9092E+08, 2.9335E+08,
     +  2.9581E+08, 2.9829E+08, 3.0078E+08, 3.0329E+08, 3.0583E+08,
     + 3.0838E+08, 3.1095E+08, 3.1354E+08, 3.1615E+08, 3.1878E+08,
     +  3.2143E+08, 3.2410E+08, 3.2679E+08, 3.2950E+08, 3.3223E+08,
     + 3.3498E+08, 3.3776E+08, 3.4055E+08, 3.4336E+08, 3.4620E+08,
     +  3.4905E+08, 3.5193E+08, 3.5483E+08, 3.5775E+08, 3.6070E+08,
     + 3.6366E+08, 3.6665E+08, 3.6966E+08, 3.7269E+08, 3.7574E+08,
     +  3.7882E+08, 3.8192E+08, 3.8505E+08, 3.8819E+08, 3.9136E+08,
     + 3.9456E+08, 3.9777E+08, 4.0101E+08, 4.0428E+08, 4.0757E+08,
     +  4.1088E+08, 4.1422E+08, 4.1758E+08, 4.2097E+08, 4.2438E+08,
     + 4.2782E+08, 4.3128E+08, 4.3477E+08, 4.3828E+08, 4.4182E+08,
     +  4.4539E+08, 4.4898E+08, 4.5260E+08, 4.5625E+08, 4.5992E+08,
     + 4.6362E+08, 4.6734E+08, 4.7109E+08, 4.7487E+08, 4.7868E+08,
     +  4.8252E+08, 4.8638E+08, 4.9027E+08, 4.9419E+08, 4.9814E+08,
     + 5.0212E+08, 5.0612E+08, 5.1016E+08, 5.1422E+08, 5.1831E+08,
     +  5.2244E+08, 5.2659E+08, 5.3077E+08, 5.3498E+08, 5.3923E+08,
     + 5.4350E+08, 5.4781E+08, 5.5214E+08, 5.5651E+08, 5.6091E+08,
     +  5.6534E+08, 5.6980E+08, 5.7429E+08, 5.7882E+08, 5.8338E+08,
     + 5.8797E+08, 5.9259E+08, 5.9725E+08, 6.0194E+08, 6.0666E+08,
     +  6.1142E+08, 6.1621E+08, 6.2104E+08, 6.2590E+08, 6.3079E+08,
     + 6.3572E+08, 6.4068E+08, 6.4568E+08, 6.5072E+08, 6.5579E+08,
     +  6.6089E+08, 6.6604E+08, 6.7121E+08, 6.7643E+08, 6.8168E+08,
     + 6.8697E+08, 6.9230E+08, 6.9766E+08, 7.0306E+08, 7.0850E+08,
     +  7.1398E+08, 7.1950E+08, 7.2505E+08, 7.3065E+08, 7.3628E+08,
     + 7.4195E+08, 7.4767E+08, 7.5342E+08, 7.5921E+08, 7.6504E+08,
     +  7.7092E+08, 7.7683E+08, 7.8279E+08, 7.8879E+08, 7.9483E+08,
     + 8.0091E+08, 8.0703E+08, 8.1320E+08, 8.1941E+08, 8.2566E+08,
     +  8.3195E+08, 8.3829E+08, 8.4467E+08, 8.5110E+08, 8.5757E+08,
     + 8.6409E+08, 8.7065E+08, 8.7726E+08, 8.8391E+08, 8.9061E+08,
     +  8.9735E+08, 9.0414E+08, 9.1098E+08, 9.1786E+08, 9.2479E+08,
     + 9.3177E+08, 9.3880E+08, 9.4587E+08, 9.5299E+08, 9.6017E+08,
     +  9.6739E+08, 9.7466E+08, 9.8198E+08, 9.8935E+08, 9.9677E+08,
     + 1.0042E+09, 1.0118E+09, 1.0193E+09, 1.0270E+09, 1.0346E+09,
     +  1.0424E+09, 1.0501E+09, 1.0580E+09, 1.0659E+09, 1.0738E+09,
     + 1.0818E+09, 1.0898E+09, 1.0979E+09, 1.1061E+09, 1.1143E+09,
     +  1.1226E+09, 1.1309E+09, 1.1393E+09, 1.1477E+09, 1.1562E+09,
     + 1.1647E+09, 1.1734E+09, 1.1820E+09, 1.1907E+09, 1.1995E+09,
     +  1.2084E+09, 1.2173E+09, 1.2262E+09, 1.2352E+09, 1.2443E+09,
     + 1.2535E+09, 1.2627E+09, 1.2719E+09, 1.2812E+09, 1.2906E+09,
     +  1.3001E+09, 1.3096E+09, 1.3191E+09, 1.3288E+09, 1.3385E+09,
     + 1.3483E+09, 1.3581E+09, 1.3680E+09, 1.3779E+09, 1.3880E+09,
     +  1.3981E+09, 1.4082E+09, 1.4184E+09, 1.4287E+09, 1.4391E+09,
     + 1.4495E+09, 1.4600E+09, 1.4706E+09, 1.4812E+09, 1.4919E+09,
     +  1.5027E+09, 1.5136E+09, 1.5245E+09, 1.5355E+09, 1.5465E+09,
     + 1.5577E+09, 1.5689E+09, 1.5801E+09, 1.5915E+09, 1.6029E+09,
     +  1.6144E+09, 1.6260E+09, 1.6377E+09, 1.6494E+09, 1.6612E+09,
     + 1.6731E+09, 1.6850E+09, 1.6971E+09, 1.7092E+09, 1.7214E+09,
     +  1.7336E+09, 1.7460E+09, 1.7584E+09, 1.7709E+09, 1.7835E+09,
     + 1.7962E+09, 1.8090E+09, 1.8218E+09, 1.8347E+09, 1.8477E+09,
     +  1.8608E+09, 1.8740E+09, 1.8872E+09, 1.9006E+09, 1.9140E+09,
     + 1.9275E+09, 1.9411E+09, 1.9548E+09, 1.9686E+09, 1.9824E+09,
     +  1.9964E+09, 2.0104E+09, 2.0246E+09, 2.0388E+09, 2.0531E+09,
     + 2.0675E+09, 2.0820E+09, 2.0966E+09, 2.1113E+09, 2.1260E+09,
     +  2.1409E+09, 2.1558E+09, 2.1709E+09, 2.1861E+09, 2.2013E+09,
     + 2.2166E+09, 2.2321E+09, 2.2476E+09, 2.2633E+09, 2.2790E+09,
     +  2.2948E+09, 2.3108E+09, 2.3268E+09, 2.3429E+09, 2.3592E+09,
     + 2.3755E+09, 2.3919E+09, 2.4085E+09, 2.4251E+09, 2.4419E+09,
     +  2.4588E+09, 2.4757E+09, 2.4928E+09, 2.5100E+09, 2.5272E+09,
     + 2.5446E+09, 2.5621E+09, 2.5798E+09, 2.5975E+09, 2.6153E+09,
     +  2.6332E+09, 2.6513E+09, 2.6695E+09, 2.6877E+09, 2.7061E+09,
     + 2.7246E+09, 2.7433E+09, 2.7620E+09, 2.7809E+09, 2.7998E+09,
     +  2.8189E+09, 2.8381E+09, 2.8574E+09, 2.8769E+09, 2.8965E+09,
     + 2.9161E+09, 2.9359E+09, 2.9559E+09, 2.9759E+09, 2.9961E+09,
     +  3.0164E+09, 3.0368E+09, 3.0574E+09, 3.0781E+09, 3.0989E+09,
     + 3.1198E+09, 3.1408E+09, 3.1620E+09, 3.1833E+09, 3.2048E+09,
     +  3.2264E+09, 3.2481E+09, 3.2699E+09, 3.2919E+09, 3.3140E+09,
     + 3.3363E+09, 3.3586E+09, 3.3812E+09, 3.4038E+09, 3.4266E+09,
     +  3.4495E+09, 3.4726E+09, 3.4958E+09, 3.5192E+09, 3.5427E+09,
     + 3.5663E+09, 3.5901E+09, 3.6140E+09, 3.6381E+09, 3.6623E+09,
     +  3.6866E+09, 3.7111E+09, 3.7358E+09, 3.7606E+09, 3.7855E+09,
     + 3.8106E+09, 3.8359E+09, 3.8613E+09, 3.8868E+09, 3.9126E+09,
     +  3.9384E+09, 3.9644E+09, 3.9906E+09, 4.0170E+09, 4.0434E+09,
     + 4.0701E+09, 4.0969E+09, 4.1239E+09, 4.1510E+09, 4.1783E+09,
     +  4.2057E+09, 4.2334E+09, 4.2611E+09, 4.2891E+09, 4.3172E+09,
     + 4.3455E+09, 4.3739E+09, 4.4026E+09, 4.4313E+09, 4.4603E+09,
     +  4.4894E+09, 4.5187E+09, 4.5482E+09, 4.5779E+09, 4.6077E+09,
     + 4.6377E+09, 4.6679E+09, 4.6982E+09, 4.7288E+09, 4.7595E+09,
     +  4.7904E+09, 4.8215E+09, 4.8527E+09, 4.8842E+09, 4.9158E+09,
     + 4.9476E+09, 4.9796E+09, 5.0118E+09, 5.0442E+09, 5.0768E+09,
     +  5.1096E+09, 5.1425E+09, 5.1757E+09, 5.2090E+09, 5.2425E+09,
     + 5.2763E+09, 5.3102E+09, 5.3443E+09, 5.3787E+09, 5.4132E+09,
     +  5.4479E+09, 5.4829E+09, 5.5180E+09, 5.5533E+09, 5.5889E+09,
     + 5.6246E+09, 5.6606E+09, 5.6967E+09, 5.7331E+09, 5.7697E+09,
     +  5.8065E+09, 5.8435E+09, 5.8807E+09, 5.9182E+09, 5.9558E+09,
     + 5.9937E+09, 6.0318E+09, 6.0701E+09, 6.1086E+09, 6.1474E+09,
     +  6.1864E+09, 6.2256E+09, 6.2650E+09, 6.3047E+09, 6.3445E+09,
     + 6.3847E+09, 6.4250E+09, 6.4656E+09, 6.5064E+09, 6.5474E+09,
     +  6.5887E+09, 6.6302E+09, 6.6719E+09, 6.7139E+09, 6.7562E+09,
     + 6.7986E+09, 6.8413E+09, 6.8843E+09, 6.9275E+09, 6.9709E+09,
     +  7.0146E+09, 7.0585E+09, 7.1027E+09, 7.1472E+09, 7.1919E+09,
     + 7.2368E+09, 7.2820E+09, 7.3275E+09, 7.3732E+09, 7.4191E+09,
     +  7.4654E+09, 7.5119E+09, 7.5586E+09, 7.6056E+09, 7.6529E+09,
     + 7.7005E+09, 7.7483E+09, 7.7964E+09, 7.8447E+09, 7.8934E+09,
     +  7.9423E+09, 7.9915E+09, 8.0409E+09, 8.0907E+09, 8.1407E+09,
     + 8.1910E+09, 8.2415E+09, 8.2924E+09, 8.3435E+09, 8.3950E+09,
     +  8.4467E+09, 8.4987E+09, 8.5510E+09, 8.6036E+09, 8.6565E+09,
     + 8.7096E+09, 8.7631E+09, 8.8169E+09, 8.8710E+09, 8.9253E+09,
     +  8.9800E+09, 9.0350E+09, 9.0903E+09, 9.1458E+09, 9.2017E+09,
     + 9.2580E+09, 9.3145E+09, 9.3713E+09, 9.4285E+09, 9.4859E+09,
     +  9.5437E+09, 9.6018E+09, 9.6602E+09, 9.7190E+09, 9.7780E+09,
     + 9.8374E+09, 9.8971E+09, 9.9572E+09, 1.0018E+10, 1.0078E+10,
     +  1.0139E+10, 1.0201E+10, 1.0262E+10, 1.0324E+10, 1.0387E+10,
     + 1.0450E+10, 1.0513E+10, 1.0576E+10, 1.0640E+10, 1.0704E+10,
     +  1.0769E+10, 1.0833E+10, 1.0899E+10, 1.0964E+10, 1.1030E+10,
     + 1.1096E+10, 1.1163E+10, 1.1230E+10, 1.1297E+10, 1.1365E+10,
     +  1.1433E+10, 1.1501E+10, 1.1570E+10, 1.1639E+10, 1.1709E+10,
     + 1.1779E+10, 1.1849E+10, 1.1920E+10, 1.1991E+10, 1.2062E+10,
     +  1.2134E+10, 1.2207E+10, 1.2279E+10, 1.2352E+10, 1.2426E+10,
     + 1.2499E+10, 1.2574E+10, 1.2648E+10, 1.2723E+10, 1.2799E+10,
     +  1.2875E+10, 1.2951E+10, 1.3027E+10, 1.3105E+10, 1.3182E+10,
     + 1.3260E+10, 1.3338E+10, 1.3417E+10, 1.3496E+10, 1.3576E+10,
     +  1.3656E+10, 1.3736E+10, 1.3817E+10, 1.3898E+10, 1.3980E+10,
     + 1.4062E+10, 1.4145E+10, 1.4228E+10, 1.4311E+10, 1.4395E+10,
     +  1.4479E+10, 1.4564E+10, 1.4649E+10, 1.4735E+10, 1.4821E+10,
     + 1.4908E+10, 1.4995E+10, 1.5083E+10, 1.5171E+10, 1.5259E+10,
     +  1.5348E+10, 1.5437E+10, 1.5527E+10, 1.5618E+10, 1.5708E+10,
     + 1.5800E+10, 1.5892E+10, 1.5984E+10, 1.6077E+10, 1.6170E+10,
     +  1.6264E+10, 1.6358E+10, 1.6453E+10, 1.6548E+10, 1.6644E+10,
     + 1.6740E+10, 1.6836E+10, 1.6934E+10, 1.7031E+10, 1.7130E+10,
     +  1.7229E+10, 1.7328E+10, 1.7428E+10, 1.7528E+10, 1.7629E+10,
     + 1.7730E+10, 1.7832E+10, 1.7935E+10, 1.8038E+10, 1.8141E+10,
     +  1.8245E+10, 1.8350E+10, 1.8455E+10, 1.8561E+10, 1.8667E+10,
     + 1.8774E+10, 1.8881E+10, 1.8989E+10, 1.9097E+10, 1.9206E+10,
     +  1.9316E+10, 1.9426E+10, 1.9537E+10, 1.9648E+10, 1.9760E+10,
     + 1.9872E+10, 1.9985E+10, 2.0099E+10, 2.0213E+10, 2.0328E+10,
     +  2.0443E+10, 2.0559E+10, 2.0676E+10, 2.0793E+10, 2.0911E+10,
     + 2.1029E+10, 2.1148E+10, 2.1268E+10, 2.1388E+10, 2.1509E+10,
     +  2.1630E+10, 2.1752E+10, 2.1875E+10, 2.1998E+10, 2.2122E+10,
     + 2.2247E+10, 2.2372E+10, 2.2498E+10, 2.2624E+10, 2.2751E+10,
     +  2.2879E+10, 2.3008E+10, 2.3137E+10, 2.3266E+10, 2.3397E+10,
     + 2.3528E+10, 2.3660E+10, 2.3792E+10, 2.3925E+10, 2.4059E+10,
     +  2.4193E+10, 2.4328E+10, 2.4464E+10, 2.4601E+10, 2.4738E+10,
     + 2.4876E+10, 2.5014E+10, 2.5154E+10, 2.5294E+10, 2.5434E+10,
     +  2.5576E+10, 2.5718E+10, 2.5860E+10, 2.6004E+10, 2.6148E+10,
     + 2.6293E+10, 2.6439E+10, 2.6585E+10, 2.6733E+10, 2.6880E+10,
     +  2.7029E+10, 2.7178E+10, 2.7329E+10, 2.7480E+10, 2.7631E+10,
     + 2.7784E+10, 2.7937E+10, 2.8091E+10, 2.8245E+10, 2.8401E+10,
     +  2.8557E+10, 2.8714E+10, 2.8872E+10, 2.9031E+10, 2.9190E+10,
     + 2.9350E+10, 2.9511E+10, 2.9673E+10, 2.9836E+10, 2.9999E+10,
     +  3.0163E+10, 3.0328E+10, 3.0494E+10, 3.0661E+10, 3.0828E+10,
     + 3.0997E+10, 3.1166E+10, 3.1336E+10, 3.1507E+10, 3.1678E+10,
     +  3.1851E+10, 3.2024E+10, 3.2198E+10, 3.2374E+10, 3.2549E+10,
     + 3.2726E+10, 3.2904E+10, 3.3083E+10, 3.3262E+10, 3.3442E+10,
     +  3.3624E+10, 3.3806E+10, 3.3989E+10, 3.4173E+10, 3.4357E+10,
     + 3.4543E+10, 3.4730E+10, 3.4917E+10, 3.5106E+10, 3.5295E+10,
     +  3.5485E+10, 3.5677E+10, 3.5869E+10, 3.6062E+10, 3.6256E+10,
     + 3.6451E+10, 3.6647E+10, 3.6844E+10, 3.7042E+10, 3.7241E+10,
     +  3.7440E+10, 3.7641E+10, 3.7843E+10, 3.8046E+10, 3.8249E+10,
     + 3.8454E+10, 3.8660E+10, 3.8866E+10, 3.9074E+10, 3.9283E+10,
     +  3.9492E+10, 3.9703E+10, 3.9915E+10, 4.0128E+10, 4.0341E+10,
     + 4.0556E+10, 4.0772E+10, 4.0989E+10, 4.1207E+10, 4.1426E+10,
     +  4.1646E+10, 4.1867E+10, 4.2089E+10, 4.2313E+10, 4.2537E+10,
     + 4.2762E+10, 4.2989E+10, 4.3216E+10, 4.3445E+10, 4.3675E+10,
     +  4.3906E+10, 4.4138E+10, 4.4371E+10, 4.4605E+10, 4.4840E+10,
     + 4.5077E+10, 4.5314E+10, 4.5553E+10, 4.5793E+10, 4.6034E+10,
     +  4.6276E+10, 4.6519E+10, 4.6764E+10, 4.7009E+10, 4.7256E+10,
     + 4.7504E+10, 4.7753E+10, 4.8004E+10, 4.8255E+10, 4.8508E+10,
     +  4.8762E+10, 4.9017E+10, 4.9273E+10, 4.9531E+10, 4.9790E+10,
     + 5.0050E+10, 5.0311E+10, 5.0573E+10, 5.0837E+10, 5.1102E+10,
     +  5.1368E+10, 5.1636E+10, 5.1904E+10, 5.2174E+10, 5.2446E+10,
     + 5.2718E+10, 5.2992E+10, 5.3267E+10, 5.3543E+10, 5.3821E+10,
     +  5.4100E+10, 5.4381E+10, 5.4662E+10, 5.4945E+10, 5.5229E+10,
     + 5.5515E+10, 5.5802E+10, 5.6090E+10, 5.6380E+10, 5.6671E+10,
     +  5.6963E+10, 5.7257E+10, 5.7552E+10, 5.7849E+10, 5.8147E+10,
     + 5.8446E+10, 5.8747E+10, 5.9049E+10, 5.9352E+10, 5.9657E+10,
     +  5.9963E+10, 6.0271E+10, 6.0580E+10, 6.0891E+10, 6.1203E+10,
     + 6.1517E+10, 6.1832E+10, 6.2148E+10, 6.2466E+10, 6.2785E+10,
     +  6.3106E+10, 6.3428E+10, 6.3752E+10, 6.4078E+10, 6.4404E+10,
     + 6.4733E+10, 6.5063E+10, 6.5394E+10, 6.5727E+10, 6.6061E+10,
     +  6.6397E+10, 6.6735E+10, 6.7074E+10, 6.7415E+10, 6.7757E+10,
     + 6.8101E+10, 6.8446E+10, 6.8793E+10, 6.9142E+10, 6.9492E+10,
     +  6.9844E+10, 7.0197E+10, 7.0552E+10, 7.0909E+10, 7.1267E+10,
     + 7.1627E+10, 7.1989E+10, 7.2352E+10, 7.2717E+10, 7.3083E+10,
     +  7.3451E+10, 7.3821E+10, 7.4193E+10, 7.4566E+10, 7.4941E+10,
     + 7.5318E+10, 7.5696E+10, 7.6076E+10, 7.6458E+10, 7.6842E+10,
     +  7.7227E+10, 7.7614E+10, 7.8003E+10, 7.8394E+10, 7.8786E+10,
     + 7.9181E+10, 7.9577E+10, 7.9974E+10, 8.0374E+10, 8.0775E+10,
     +  8.1178E+10, 8.1584E+10, 8.1990E+10, 8.2399E+10, 8.2810E+10,
     + 8.3222E+10, 8.3636E+10, 8.4052E+10, 8.4470E+10, 8.4890E+10,
     +  8.5312E+10, 8.5736E+10, 8.6161E+10, 8.6589E+10, 8.7018E+10,
     + 8.7450E+10, 8.7883E+10, 8.8318E+10, 8.8755E+10, 8.9195E+10,
     +  8.9636E+10, 9.0079E+10, 9.0524E+10, 9.0971E+10, 9.1420E+10,
     + 9.1871E+10, 9.2324E+10, 9.2780E+10, 9.3237E+10, 9.3696E+10,
     +  9.4157E+10, 9.4621E+10, 9.5086E+10, 9.5553E+10, 9.6023E+10,
     + 9.6495E+10, 9.6968E+10, 9.7444E+10, 9.7922E+10, 9.8402E+10,
     +  9.8885E+10, 9.9369E+10, 9.9855E+10, 1.0034E+11, 1.0083E+11,
     + 1.0133E+11, 1.0182E+11, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_OH    (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 3),QofT( 3,NT),Q(NT)
      data (Tmx(I),I=1,3)/9000.,3500.,3500./
      data xgj/ 2.,2.,3./
c...       OH
c...        --        61
      data (QofT( 1,J),J=1,2408)/ 1.5385E+01, 1.5686E+01, 1.5790E+01,
     +  1.5842E+01, 1.5873E+01, 1.5894E+01, 1.5909E+01, 1.5920E+01,
     + 1.5929E+01, 1.5936E+01, 1.5942E+01, 1.5948E+01, 1.5953E+01,
     +  1.5959E+01, 1.5965E+01, 1.5973E+01, 1.5983E+01, 1.5994E+01,
     + 1.6009E+01, 1.6027E+01, 1.6073E+01, 1.6136E+01, 1.6216E+01,
     +  1.6315E+01, 1.6432E+01, 1.6568E+01, 1.6723E+01, 1.6894E+01,
     + 1.7083E+01, 1.7288E+01, 1.7508E+01, 1.7742E+01, 1.7990E+01,
     +  1.8251E+01, 1.8524E+01, 1.8809E+01, 1.9105E+01, 1.9411E+01,
     + 1.9727E+01, 2.0053E+01, 2.0388E+01, 2.0731E+01, 2.1082E+01,
     +  2.1441E+01, 2.1808E+01, 2.2181E+01, 2.2561E+01, 2.2948E+01,
     + 2.3341E+01, 2.3740E+01, 2.4145E+01, 2.4555E+01, 2.4970E+01,
     +  2.5390E+01, 2.5816E+01, 2.6246E+01, 2.6680E+01, 2.7118E+01,
     + 2.7561E+01, 2.8008E+01, 2.8459E+01, 2.8913E+01, 2.9371E+01,
     +  2.9832E+01, 3.0297E+01, 3.0765E+01, 3.1236E+01, 3.1710E+01,
     + 3.2187E+01, 3.2666E+01, 3.3149E+01, 3.3634E+01, 3.4121E+01,
     +  3.4611E+01, 3.5104E+01, 3.5599E+01, 3.6096E+01, 3.6595E+01,
     + 3.7096E+01, 3.7599E+01, 3.8105E+01, 3.8612E+01, 3.9121E+01,
     +  3.9632E+01, 4.0145E+01, 4.0659E+01, 4.1175E+01, 4.1693E+01,
     + 4.2212E+01, 4.2733E+01, 4.3256E+01, 4.3779E+01, 4.4305E+01,
     +  4.4831E+01, 4.5359E+01, 4.5888E+01, 4.6419E+01, 4.6951E+01,
     + 4.7484E+01, 4.8018E+01, 4.8553E+01, 4.9090E+01, 4.9627E+01,
     +  5.0166E+01, 5.0706E+01, 5.1246E+01, 5.1788E+01, 5.2331E+01,
     + 5.2874E+01, 5.3419E+01, 5.3964E+01, 5.4511E+01, 5.5058E+01,
     +  5.5606E+01, 5.6155E+01, 5.6705E+01, 5.7255E+01, 5.7807E+01,
     + 5.8359E+01, 5.8911E+01, 5.9465E+01, 6.0019E+01, 6.0574E+01,
     +  6.1130E+01, 6.1686E+01, 6.2243E+01, 6.2801E+01, 6.3359E+01,
     + 6.3918E+01, 6.4477E+01, 6.5037E+01, 6.5598E+01, 6.6159E+01,
     +  6.6721E+01, 6.7283E+01, 6.7846E+01, 6.8410E+01, 6.8974E+01,
     + 6.9538E+01, 7.0103E+01, 7.0668E+01, 7.1234E+01, 7.1801E+01,
     +  7.2368E+01, 7.2935E+01, 7.3503E+01, 7.4071E+01, 7.4640E+01,
     + 7.5209E+01, 7.5778E+01, 7.6348E+01, 7.6918E+01, 7.7489E+01,
     +  7.8060E+01, 7.8632E+01, 7.9203E+01, 7.9776E+01, 8.0348E+01,
     + 8.0921E+01, 8.1495E+01, 8.2068E+01, 8.2642E+01, 8.3217E+01,
     +  8.3791E+01, 8.4366E+01, 8.4942E+01, 8.5517E+01, 8.6093E+01,
     + 8.6670E+01, 8.7246E+01, 8.7823E+01, 8.8400E+01, 8.8978E+01,
     +  8.9555E+01, 9.0133E+01, 9.0712E+01, 9.1290E+01, 9.1869E+01,
     + 9.2448E+01, 9.3028E+01, 9.3607E+01, 9.4187E+01, 9.4767E+01,
     +  9.5348E+01, 9.5928E+01, 9.6509E+01, 9.7090E+01, 9.7672E+01,
     + 9.8253E+01, 9.8835E+01, 9.9417E+01, 1.0000E+02, 1.0058E+02,
     +  1.0116E+02, 1.0175E+02, 1.0233E+02, 1.0291E+02, 1.0350E+02,
     + 1.0408E+02, 1.0467E+02, 1.0525E+02, 1.0583E+02, 1.0642E+02,
     +  1.0700E+02, 1.0759E+02, 1.0817E+02, 1.0876E+02, 1.0934E+02,
     + 1.0993E+02, 1.1052E+02, 1.1110E+02, 1.1169E+02, 1.1227E+02,
     +  1.1286E+02, 1.1345E+02, 1.1403E+02, 1.1462E+02, 1.1521E+02,
     + 1.1580E+02, 1.1638E+02, 1.1697E+02, 1.1756E+02, 1.1815E+02,
     +  1.1874E+02, 1.1932E+02, 1.1991E+02, 1.2050E+02, 1.2109E+02,
     + 1.2168E+02, 1.2227E+02, 1.2286E+02, 1.2345E+02, 1.2404E+02,
     +  1.2462E+02, 1.2521E+02, 1.2580E+02, 1.2639E+02, 1.2698E+02,
     + 1.2757E+02, 1.2817E+02, 1.2876E+02, 1.2935E+02, 1.2994E+02,
     +  1.3053E+02, 1.3112E+02, 1.3171E+02, 1.3230E+02, 1.3289E+02,
     + 1.3348E+02, 1.3408E+02, 1.3467E+02, 1.3526E+02, 1.3585E+02,
     +  1.3644E+02, 1.3704E+02, 1.3763E+02, 1.3822E+02, 1.3881E+02,
     + 1.3941E+02, 1.4000E+02, 1.4059E+02, 1.4119E+02, 1.4178E+02,
     +  1.4237E+02, 1.4297E+02, 1.4356E+02, 1.4415E+02, 1.4475E+02,
     + 1.4534E+02, 1.4593E+02, 1.4653E+02, 1.4712E+02, 1.4772E+02,
     +  1.4831E+02, 1.4891E+02, 1.4950E+02, 1.5010E+02, 1.5069E+02,
     + 1.5129E+02, 1.5188E+02, 1.5248E+02, 1.5307E+02, 1.5367E+02,
     +  1.5426E+02, 1.5486E+02, 1.5545E+02, 1.5605E+02, 1.5664E+02,
     + 1.5724E+02, 1.5784E+02, 1.5843E+02, 1.5903E+02, 1.5962E+02,
     +  1.6022E+02, 1.6082E+02, 1.6141E+02, 1.6201E+02, 1.6261E+02,
     + 1.6320E+02, 1.6380E+02, 1.6440E+02, 1.6500E+02, 1.6559E+02,
     +  1.6619E+02, 1.6679E+02, 1.6739E+02, 1.6798E+02, 1.6858E+02,
     + 1.6918E+02, 1.6978E+02, 1.7038E+02, 1.7097E+02, 1.7157E+02,
     +  1.7217E+02, 1.7277E+02, 1.7337E+02, 1.7397E+02, 1.7457E+02,
     + 1.7516E+02, 1.7576E+02, 1.7636E+02, 1.7696E+02, 1.7756E+02,
     +  1.7816E+02, 1.7876E+02, 1.7936E+02, 1.7996E+02, 1.8056E+02,
     + 1.8116E+02, 1.8176E+02, 1.8236E+02, 1.8296E+02, 1.8356E+02,
     +  1.8416E+02, 1.8476E+02, 1.8536E+02, 1.8596E+02, 1.8656E+02,
     + 1.8716E+02, 1.8777E+02, 1.8837E+02, 1.8897E+02, 1.8957E+02,
     +  1.9017E+02, 1.9077E+02, 1.9137E+02, 1.9198E+02, 1.9258E+02,
     + 1.9318E+02, 1.9378E+02, 1.9438E+02, 1.9499E+02, 1.9559E+02,
     +  1.9619E+02, 1.9679E+02, 1.9740E+02, 1.9800E+02, 1.9860E+02,
     + 1.9921E+02, 1.9981E+02, 2.0041E+02, 2.0102E+02, 2.0162E+02,
     +  2.0222E+02, 2.0283E+02, 2.0343E+02, 2.0404E+02, 2.0464E+02,
     + 2.0524E+02, 2.0585E+02, 2.0645E+02, 2.0706E+02, 2.0766E+02,
     +  2.0827E+02, 2.0887E+02, 2.0948E+02, 2.1008E+02, 2.1069E+02,
     + 2.1129E+02, 2.1190E+02, 2.1250E+02, 2.1311E+02, 2.1372E+02,
     +  2.1432E+02, 2.1493E+02, 2.1554E+02, 2.1614E+02, 2.1675E+02,
     + 2.1736E+02, 2.1796E+02, 2.1857E+02, 2.1918E+02, 2.1978E+02,
     +  2.2039E+02, 2.2100E+02, 2.2161E+02, 2.2221E+02, 2.2282E+02,
     + 2.2343E+02, 2.2404E+02, 2.2465E+02, 2.2526E+02, 2.2586E+02,
     +  2.2647E+02, 2.2708E+02, 2.2769E+02, 2.2830E+02, 2.2891E+02,
     + 2.2952E+02, 2.3013E+02, 2.3074E+02, 2.3135E+02, 2.3196E+02,
     +  2.3257E+02, 2.3318E+02, 2.3379E+02, 2.3440E+02, 2.3501E+02,
     + 2.3562E+02, 2.3623E+02, 2.3684E+02, 2.3746E+02, 2.3807E+02,
     +  2.3868E+02, 2.3929E+02, 2.3990E+02, 2.4052E+02, 2.4113E+02,
     + 2.4174E+02, 2.4235E+02, 2.4297E+02, 2.4358E+02, 2.4419E+02,
     +  2.4481E+02, 2.4542E+02, 2.4603E+02, 2.4665E+02, 2.4726E+02,
     + 2.4788E+02, 2.4849E+02, 2.4910E+02, 2.4972E+02, 2.5033E+02,
     +  2.5095E+02, 2.5156E+02, 2.5218E+02, 2.5280E+02, 2.5341E+02,
     + 2.5403E+02, 2.5464E+02, 2.5526E+02, 2.5588E+02, 2.5649E+02,
     +  2.5711E+02, 2.5773E+02, 2.5834E+02, 2.5896E+02, 2.5958E+02,
     + 2.6020E+02, 2.6081E+02, 2.6143E+02, 2.6205E+02, 2.6267E+02,
     +  2.6329E+02, 2.6391E+02, 2.6452E+02, 2.6514E+02, 2.6576E+02,
     + 2.6638E+02, 2.6700E+02, 2.6762E+02, 2.6824E+02, 2.6886E+02,
     +  2.6948E+02, 2.7010E+02, 2.7073E+02, 2.7135E+02, 2.7197E+02,
     + 2.7259E+02, 2.7321E+02, 2.7383E+02, 2.7446E+02, 2.7508E+02,
     +  2.7570E+02, 2.7632E+02, 2.7695E+02, 2.7757E+02, 2.7819E+02,
     + 2.7882E+02, 2.7944E+02, 2.8006E+02, 2.8069E+02, 2.8131E+02,
     +  2.8194E+02, 2.8256E+02, 2.8319E+02, 2.8381E+02, 2.8444E+02,
     + 2.8506E+02, 2.8569E+02, 2.8632E+02, 2.8694E+02, 2.8757E+02,
     +  2.8820E+02, 2.8882E+02, 2.8945E+02, 2.9008E+02, 2.9071E+02,
     + 2.9133E+02, 2.9196E+02, 2.9259E+02, 2.9322E+02, 2.9385E+02,
     +  2.9448E+02, 2.9511E+02, 2.9574E+02, 2.9637E+02, 2.9700E+02,
     + 2.9763E+02, 2.9826E+02, 2.9889E+02, 2.9952E+02, 3.0015E+02,
     +  3.0078E+02, 3.0141E+02, 3.0205E+02, 3.0268E+02, 3.0331E+02,
     + 3.0394E+02, 3.0458E+02, 3.0521E+02, 3.0584E+02, 3.0648E+02,
     +  3.0711E+02, 3.0774E+02, 3.0838E+02, 3.0901E+02, 3.0965E+02,
     + 3.1028E+02, 3.1092E+02, 3.1155E+02, 3.1219E+02, 3.1283E+02,
     +  3.1346E+02, 3.1410E+02, 3.1474E+02, 3.1537E+02, 3.1601E+02,
     + 3.1665E+02, 3.1729E+02, 3.1793E+02, 3.1856E+02, 3.1920E+02,
     +  3.1984E+02, 3.2048E+02, 3.2112E+02, 3.2176E+02, 3.2240E+02,
     + 3.2304E+02, 3.2368E+02, 3.2432E+02, 3.2496E+02, 3.2561E+02,
     +  3.2625E+02, 3.2689E+02, 3.2753E+02, 3.2817E+02, 3.2882E+02,
     + 3.2946E+02, 3.3010E+02, 3.3075E+02, 3.3139E+02, 3.3204E+02,
     +  3.3268E+02, 3.3332E+02, 3.3397E+02, 3.3461E+02, 3.3526E+02,
     + 3.3591E+02, 3.3655E+02, 3.3720E+02, 3.3785E+02, 3.3849E+02,
     +  3.3914E+02, 3.3979E+02, 3.4044E+02, 3.4108E+02, 3.4173E+02,
     + 3.4238E+02, 3.4303E+02, 3.4368E+02, 3.4433E+02, 3.4498E+02,
     +  3.4563E+02, 3.4628E+02, 3.4693E+02, 3.4758E+02, 3.4823E+02,
     + 3.4889E+02, 3.4954E+02, 3.5019E+02, 3.5084E+02, 3.5150E+02,
     +  3.5215E+02, 3.5280E+02, 3.5346E+02, 3.5411E+02, 3.5476E+02,
     + 3.5542E+02, 3.5607E+02, 3.5673E+02, 3.5739E+02, 3.5804E+02,
     +  3.5870E+02, 3.5935E+02, 3.6001E+02, 3.6067E+02, 3.6133E+02,
     + 3.6198E+02, 3.6264E+02, 3.6330E+02, 3.6396E+02, 3.6462E+02,
     +  3.6528E+02, 3.6594E+02, 3.6660E+02, 3.6726E+02, 3.6792E+02,
     + 3.6858E+02, 3.6924E+02, 3.6990E+02, 3.7057E+02, 3.7123E+02,
     +  3.7189E+02, 3.7255E+02, 3.7322E+02, 3.7388E+02, 3.7454E+02,
     + 3.7521E+02, 3.7587E+02, 3.7654E+02, 3.7720E+02, 3.7787E+02,
     +  3.7854E+02, 3.7920E+02, 3.7987E+02, 3.8054E+02, 3.8120E+02,
     + 3.8187E+02, 3.8254E+02, 3.8321E+02, 3.8387E+02, 3.8454E+02,
     +  3.8521E+02, 3.8588E+02, 3.8655E+02, 3.8722E+02, 3.8789E+02,
     + 3.8856E+02, 3.8924E+02, 3.8991E+02, 3.9058E+02, 3.9125E+02,
     +  3.9192E+02, 3.9260E+02, 3.9327E+02, 3.9394E+02, 3.9462E+02,
     + 3.9529E+02, 3.9597E+02, 3.9664E+02, 3.9732E+02, 3.9799E+02,
     +  3.9867E+02, 3.9935E+02, 4.0002E+02, 4.0070E+02, 4.0138E+02,
     + 4.0206E+02, 4.0274E+02, 4.0341E+02, 4.0409E+02, 4.0477E+02,
     +  4.0545E+02, 4.0613E+02, 4.0681E+02, 4.0749E+02, 4.0817E+02,
     + 4.0886E+02, 4.0954E+02, 4.1022E+02, 4.1090E+02, 4.1159E+02,
     +  4.1227E+02, 4.1295E+02, 4.1364E+02, 4.1432E+02, 4.1501E+02,
     + 4.1569E+02, 4.1638E+02, 4.1706E+02, 4.1775E+02, 4.1843E+02,
     +  4.1912E+02, 4.1981E+02, 4.2050E+02, 4.2118E+02, 4.2187E+02,
     + 4.2256E+02, 4.2325E+02, 4.2394E+02, 4.2463E+02, 4.2532E+02,
     +  4.2601E+02, 4.2670E+02, 4.2739E+02, 4.2809E+02, 4.2878E+02,
     + 4.2947E+02, 4.3016E+02, 4.3086E+02, 4.3155E+02, 4.3224E+02,
     +  4.3294E+02, 4.3363E+02, 4.3433E+02, 4.3502E+02, 4.3572E+02,
     + 4.3642E+02, 4.3711E+02, 4.3781E+02, 4.3851E+02, 4.3921E+02,
     +  4.3990E+02, 4.4060E+02, 4.4130E+02, 4.4200E+02, 4.4270E+02,
     + 4.4340E+02, 4.4410E+02, 4.4480E+02, 4.4550E+02, 4.4621E+02,
     +  4.4691E+02, 4.4761E+02, 4.4831E+02, 4.4902E+02, 4.4972E+02,
     + 4.5042E+02, 4.5113E+02, 4.5183E+02, 4.5254E+02, 4.5324E+02,
     +  4.5395E+02, 4.5466E+02, 4.5536E+02, 4.5607E+02, 4.5678E+02,
     + 4.5749E+02, 4.5820E+02, 4.5890E+02, 4.5961E+02, 4.6032E+02,
     +  4.6103E+02, 4.6174E+02, 4.6245E+02, 4.6317E+02, 4.6388E+02,
     + 4.6459E+02, 4.6530E+02, 4.6601E+02, 4.6673E+02, 4.6744E+02,
     +  4.6815E+02, 4.6887E+02, 4.6958E+02, 4.7030E+02, 4.7101E+02,
     + 4.7173E+02, 4.7245E+02, 4.7316E+02, 4.7388E+02, 4.7460E+02,
     +  4.7532E+02, 4.7604E+02, 4.7675E+02, 4.7747E+02, 4.7819E+02,
     + 4.7891E+02, 4.7963E+02, 4.8036E+02, 4.8108E+02, 4.8180E+02,
     +  4.8252E+02, 4.8324E+02, 4.8397E+02, 4.8469E+02, 4.8541E+02,
     + 4.8614E+02, 4.8686E+02, 4.8759E+02, 4.8831E+02, 4.8904E+02,
     +  4.8977E+02, 4.9049E+02, 4.9122E+02, 4.9195E+02, 4.9268E+02,
     + 4.9340E+02, 4.9413E+02, 4.9486E+02, 4.9559E+02, 4.9632E+02,
     +  4.9705E+02, 4.9778E+02, 4.9851E+02, 4.9925E+02, 4.9998E+02,
     + 5.0071E+02, 5.0144E+02, 5.0218E+02, 5.0291E+02, 5.0365E+02,
     +  5.0438E+02, 5.0512E+02, 5.0585E+02, 5.0659E+02, 5.0732E+02,
     + 5.0806E+02, 5.0880E+02, 5.0954E+02, 5.1027E+02, 5.1101E+02,
     +  5.1175E+02, 5.1249E+02, 5.1323E+02, 5.1397E+02, 5.1471E+02,
     + 5.1545E+02, 5.1620E+02, 5.1694E+02, 5.1768E+02, 5.1842E+02,
     +  5.1917E+02, 5.1991E+02, 5.2066E+02, 5.2140E+02, 5.2215E+02,
     + 5.2289E+02, 5.2364E+02, 5.2438E+02, 5.2513E+02, 5.2588E+02,
     +  5.2662E+02, 5.2737E+02, 5.2812E+02, 5.2887E+02, 5.2962E+02,
     + 5.3037E+02, 5.3112E+02, 5.3187E+02, 5.3262E+02, 5.3337E+02,
     +  5.3413E+02, 5.3488E+02, 5.3563E+02, 5.3639E+02, 5.3714E+02,
     + 5.3789E+02, 5.3865E+02, 5.3940E+02, 5.4016E+02, 5.4092E+02,
     +  5.4167E+02, 5.4243E+02, 5.4319E+02, 5.4395E+02, 5.4470E+02,
     + 5.4546E+02, 5.4622E+02, 5.4698E+02, 5.4774E+02, 5.4850E+02,
     +  5.4926E+02, 5.5002E+02, 5.5079E+02, 5.5155E+02, 5.5231E+02,
     + 5.5307E+02, 5.5384E+02, 5.5460E+02, 5.5537E+02, 5.5613E+02,
     +  5.5690E+02, 5.5766E+02, 5.5843E+02, 5.5920E+02, 5.5996E+02,
     + 5.6073E+02, 5.6150E+02, 5.6227E+02, 5.6304E+02, 5.6381E+02,
     +  5.6458E+02, 5.6535E+02, 5.6612E+02, 5.6689E+02, 5.6766E+02,
     + 5.6844E+02, 5.6921E+02, 5.6998E+02, 5.7076E+02, 5.7153E+02,
     +  5.7230E+02, 5.7308E+02, 5.7386E+02, 5.7463E+02, 5.7541E+02,
     + 5.7618E+02, 5.7696E+02, 5.7774E+02, 5.7852E+02, 5.7930E+02,
     +  5.8008E+02, 5.8086E+02, 5.8164E+02, 5.8242E+02, 5.8320E+02,
     + 5.8398E+02, 5.8476E+02, 5.8554E+02, 5.8633E+02, 5.8711E+02,
     +  5.8789E+02, 5.8868E+02, 5.8946E+02, 5.9025E+02, 5.9104E+02,
     + 5.9182E+02, 5.9261E+02, 5.9340E+02, 5.9418E+02, 5.9497E+02,
     +  5.9576E+02, 5.9655E+02, 5.9734E+02, 5.9813E+02, 5.9892E+02,
     + 5.9971E+02, 6.0050E+02, 6.0129E+02, 6.0208E+02, 6.0288E+02,
     +  6.0367E+02, 6.0446E+02, 6.0526E+02, 6.0605E+02, 6.0685E+02,
     + 6.0764E+02, 6.0844E+02, 6.0924E+02, 6.1003E+02, 6.1083E+02,
     +  6.1163E+02, 6.1243E+02, 6.1323E+02, 6.1403E+02, 6.1483E+02,
     + 6.1563E+02, 6.1643E+02, 6.1723E+02, 6.1803E+02, 6.1883E+02,
     +  6.1964E+02, 6.2044E+02, 6.2124E+02, 6.2205E+02, 6.2285E+02,
     + 6.2366E+02, 6.2446E+02, 6.2527E+02, 6.2607E+02, 6.2688E+02,
     +  6.2769E+02, 6.2850E+02, 6.2931E+02, 6.3011E+02, 6.3092E+02,
     + 6.3173E+02, 6.3254E+02, 6.3335E+02, 6.3417E+02, 6.3498E+02,
     +  6.3579E+02, 6.3660E+02, 6.3742E+02, 6.3823E+02, 6.3904E+02,
     + 6.3986E+02, 6.4067E+02, 6.4149E+02, 6.4231E+02, 6.4312E+02,
     +  6.4394E+02, 6.4476E+02, 6.4558E+02, 6.4639E+02, 6.4721E+02,
     + 6.4803E+02, 6.4885E+02, 6.4967E+02, 6.5049E+02, 6.5132E+02,
     +  6.5214E+02, 6.5296E+02, 6.5378E+02, 6.5461E+02, 6.5543E+02,
     + 6.5626E+02, 6.5708E+02, 6.5791E+02, 6.5873E+02, 6.5956E+02,
     +  6.6039E+02, 6.6121E+02, 6.6204E+02, 6.6287E+02, 6.6370E+02,
     + 6.6453E+02, 6.6536E+02, 6.6619E+02, 6.6702E+02, 6.6785E+02,
     +  6.6868E+02, 6.6951E+02, 6.7035E+02, 6.7118E+02, 6.7201E+02,
     + 6.7285E+02, 6.7368E+02, 6.7452E+02, 6.7535E+02, 6.7619E+02,
     +  6.7703E+02, 6.7786E+02, 6.7870E+02, 6.7954E+02, 6.8038E+02,
     + 6.8122E+02, 6.8206E+02, 6.8290E+02, 6.8374E+02, 6.8458E+02,
     +  6.8542E+02, 6.8626E+02, 6.8711E+02, 6.8795E+02, 6.8879E+02,
     + 6.8964E+02, 6.9048E+02, 6.9133E+02, 6.9217E+02, 6.9302E+02,
     +  6.9387E+02, 6.9471E+02, 6.9556E+02, 6.9641E+02, 6.9726E+02,
     + 6.9811E+02, 6.9896E+02, 6.9981E+02, 7.0066E+02, 7.0151E+02,
     +  7.0236E+02, 7.0321E+02, 7.0406E+02, 7.0492E+02, 7.0577E+02,
     + 7.0663E+02, 7.0748E+02, 7.0834E+02, 7.0919E+02, 7.1005E+02,
     +  7.1090E+02, 7.1176E+02, 7.1262E+02, 7.1348E+02, 7.1434E+02,
     + 7.1519E+02, 7.1605E+02, 7.1691E+02, 7.1778E+02, 7.1864E+02,
     +  7.1950E+02, 7.2036E+02, 7.2122E+02, 7.2209E+02, 7.2295E+02,
     + 7.2382E+02, 7.2468E+02, 7.2555E+02, 7.2641E+02, 7.2728E+02,
     +  7.2814E+02, 7.2901E+02, 7.2988E+02, 7.3075E+02, 7.3162E+02,
     + 7.3249E+02, 7.3336E+02, 7.3423E+02, 7.3510E+02, 7.3597E+02,
     +  7.3684E+02, 7.3771E+02, 7.3859E+02, 7.3946E+02, 7.4033E+02,
     + 7.4121E+02, 7.4208E+02, 7.4296E+02, 7.4383E+02, 7.4471E+02,
     +  7.4559E+02, 7.4647E+02, 7.4734E+02, 7.4822E+02, 7.4910E+02,
     + 7.4998E+02, 7.5086E+02, 7.5174E+02, 7.5262E+02, 7.5350E+02,
     +  7.5439E+02, 7.5527E+02, 7.5615E+02, 7.5704E+02, 7.5792E+02,
     + 7.5880E+02, 7.5969E+02, 7.6058E+02, 7.6146E+02, 7.6235E+02,
     +  7.6324E+02, 7.6412E+02, 7.6501E+02, 7.6590E+02, 7.6679E+02,
     + 7.6768E+02, 7.6857E+02, 7.6946E+02, 7.7035E+02, 7.7125E+02,
     +  7.7214E+02, 7.7303E+02, 7.7392E+02, 7.7482E+02, 7.7571E+02,
     + 7.7661E+02, 7.7750E+02, 7.7840E+02, 7.7930E+02, 7.8019E+02,
     +  7.8109E+02, 7.8199E+02, 7.8289E+02, 7.8379E+02, 7.8469E+02,
     + 7.8559E+02, 7.8649E+02, 7.8739E+02, 7.8829E+02, 7.8919E+02,
     +  7.9010E+02, 7.9100E+02, 7.9190E+02, 7.9281E+02, 7.9371E+02,
     + 7.9462E+02, 7.9553E+02, 7.9643E+02, 7.9734E+02, 7.9825E+02,
     +  7.9915E+02, 8.0006E+02, 8.0097E+02, 8.0188E+02, 8.0279E+02,
     + 8.0370E+02, 8.0461E+02, 8.0553E+02, 8.0644E+02, 8.0735E+02,
     +  8.0826E+02, 8.0918E+02, 8.1009E+02, 8.1101E+02, 8.1192E+02,
     + 8.1284E+02, 8.1376E+02, 8.1467E+02, 8.1559E+02, 8.1651E+02,
     +  8.1743E+02, 8.1835E+02, 8.1927E+02, 8.2019E+02, 8.2111E+02,
     + 8.2203E+02, 8.2295E+02, 8.2387E+02, 8.2480E+02, 8.2572E+02,
     +  8.2664E+02, 8.2757E+02, 8.2849E+02, 8.2942E+02, 8.3034E+02,
     + 8.3127E+02, 8.3220E+02, 8.3313E+02, 8.3405E+02, 8.3498E+02,
     +  8.3591E+02, 8.3684E+02, 8.3777E+02, 8.3870E+02, 8.3963E+02,
     + 8.4057E+02, 8.4150E+02, 8.4243E+02, 8.4336E+02, 8.4430E+02,
     +  8.4523E+02, 8.4617E+02, 8.4710E+02, 8.4804E+02, 8.4898E+02,
     + 8.4991E+02, 8.5085E+02, 8.5179E+02, 8.5273E+02, 8.5367E+02,
     +  8.5461E+02, 8.5555E+02, 8.5649E+02, 8.5743E+02, 8.5837E+02,
     + 8.5932E+02, 8.6026E+02, 8.6120E+02, 8.6215E+02, 8.6309E+02,
     +  8.6404E+02, 8.6498E+02, 8.6593E+02, 8.6688E+02, 8.6782E+02,
     + 8.6877E+02, 8.6972E+02, 8.7067E+02, 8.7162E+02, 8.7257E+02,
     +  8.7352E+02, 8.7447E+02, 8.7542E+02, 8.7637E+02, 8.7733E+02,
     + 8.7828E+02, 8.7923E+02, 8.8019E+02, 8.8114E+02, 8.8210E+02,
     +  8.8306E+02, 8.8401E+02, 8.8497E+02, 8.8593E+02, 8.8689E+02,
     + 8.8784E+02, 8.8880E+02, 8.8976E+02, 8.9072E+02, 8.9168E+02,
     +  8.9265E+02, 8.9361E+02, 8.9457E+02, 8.9553E+02, 8.9650E+02,
     + 8.9746E+02, 8.9843E+02, 8.9939E+02, 9.0036E+02, 9.0132E+02,
     +  9.0229E+02, 9.0326E+02, 9.0423E+02, 9.0519E+02, 9.0616E+02,
     + 9.0713E+02, 9.0810E+02, 9.0907E+02, 9.1004E+02, 9.1102E+02,
     +  9.1199E+02, 9.1296E+02, 9.1394E+02, 9.1491E+02, 9.1588E+02,
     + 9.1686E+02, 9.1783E+02, 9.1881E+02, 9.1979E+02, 9.2076E+02,
     +  9.2174E+02, 9.2272E+02, 9.2370E+02, 9.2468E+02, 9.2566E+02,
     + 9.2664E+02, 9.2762E+02, 9.2860E+02, 9.2958E+02, 9.3057E+02,
     +  9.3155E+02, 9.3253E+02, 9.3352E+02, 9.3450E+02, 9.3549E+02,
     + 9.3647E+02, 9.3746E+02, 9.3845E+02, 9.3944E+02, 9.4042E+02,
     +  9.4141E+02, 9.4240E+02, 9.4339E+02, 9.4438E+02, 9.4537E+02,
     + 9.4637E+02, 9.4736E+02, 9.4835E+02, 9.4934E+02, 9.5034E+02,
     +  9.5133E+02, 9.5233E+02, 9.5332E+02, 9.5432E+02, 9.5531E+02,
     + 9.5631E+02, 9.5731E+02, 9.5831E+02, 9.5931E+02, 9.6030E+02,
     +  9.6130E+02, 9.6230E+02, 9.6331E+02, 9.6431E+02, 9.6531E+02,
     + 9.6631E+02, 9.6731E+02, 9.6832E+02, 9.6932E+02, 9.7033E+02,
     +  9.7133E+02, 9.7234E+02, 9.7334E+02, 9.7435E+02, 9.7536E+02,
     + 9.7637E+02, 9.7738E+02, 9.7839E+02, 9.7940E+02, 9.8041E+02,
     +  9.8142E+02, 9.8243E+02, 9.8344E+02, 9.8445E+02, 9.8547E+02,
     + 9.8648E+02, 9.8749E+02, 9.8851E+02, 9.8952E+02, 9.9054E+02,
     +  9.9156E+02, 9.9257E+02, 9.9359E+02, 9.9461E+02, 9.9563E+02,
     + 9.9665E+02, 9.9767E+02, 9.9869E+02, 9.9971E+02, 1.0007E+03,
     +  1.0018E+03, 1.0028E+03, 1.0038E+03, 1.0048E+03, 1.0058E+03,
     + 1.0069E+03, 1.0079E+03, 1.0089E+03, 1.0099E+03, 1.0110E+03,
     +  1.0120E+03, 1.0130E+03, 1.0141E+03, 1.0151E+03, 1.0161E+03,
     + 1.0171E+03, 1.0182E+03, 1.0192E+03, 1.0202E+03, 1.0213E+03,
     +  1.0223E+03, 1.0233E+03, 1.0244E+03, 1.0254E+03, 1.0264E+03,
     + 1.0275E+03, 1.0285E+03, 1.0296E+03, 1.0306E+03, 1.0316E+03,
     +  1.0327E+03, 1.0337E+03, 1.0348E+03, 1.0358E+03, 1.0368E+03,
     + 1.0379E+03, 1.0389E+03, 1.0400E+03, 1.0410E+03, 1.0420E+03,
     +  1.0431E+03, 1.0441E+03, 1.0452E+03, 1.0462E+03, 1.0473E+03,
     + 1.0483E+03, 1.0494E+03, 1.0504E+03, 1.0515E+03, 1.0525E+03,
     +  1.0536E+03, 1.0546E+03, 1.0557E+03, 1.0567E+03, 1.0578E+03,
     + 1.0588E+03, 1.0599E+03, 1.0609E+03, 1.0620E+03, 1.0630E+03,
     +  1.0641E+03, 1.0652E+03, 1.0662E+03, 1.0673E+03, 1.0683E+03,
     + 1.0694E+03, 1.0705E+03, 1.0715E+03, 1.0726E+03, 1.0736E+03,
     +  1.0747E+03, 1.0758E+03, 1.0768E+03, 1.0779E+03, 1.0789E+03,
     + 1.0800E+03, 1.0811E+03, 1.0821E+03, 1.0832E+03, 1.0843E+03,
     +  1.0853E+03, 1.0864E+03, 1.0875E+03, 1.0885E+03, 1.0896E+03,
     + 1.0907E+03, 1.0918E+03, 1.0928E+03, 1.0939E+03, 1.0950E+03,
     +  1.0960E+03, 1.0971E+03, 1.0982E+03, 1.0993E+03, 1.1003E+03,
     + 1.1014E+03, 1.1025E+03, 1.1036E+03, 1.1046E+03, 1.1057E+03,
     +  1.1068E+03, 1.1079E+03, 1.1090E+03, 1.1100E+03, 1.1111E+03,
     + 1.1122E+03, 1.1133E+03, 1.1144E+03, 1.1155E+03, 1.1165E+03,
     +  1.1176E+03, 1.1187E+03, 1.1198E+03, 1.1209E+03, 1.1220E+03,
     + 1.1231E+03, 1.1241E+03, 1.1252E+03, 1.1263E+03, 1.1274E+03,
     +  1.1285E+03, 1.1296E+03, 1.1307E+03, 1.1318E+03, 1.1329E+03,
     + 1.1340E+03, 1.1351E+03, 1.1361E+03, 1.1372E+03, 1.1383E+03,
     +  1.1394E+03, 1.1405E+03, 1.1416E+03, 1.1427E+03, 1.1438E+03,
     + 1.1449E+03, 1.1460E+03, 1.1471E+03, 1.1482E+03, 1.1493E+03,
     +  1.1504E+03, 1.1515E+03, 1.1526E+03, 1.1537E+03, 1.1548E+03,
     + 1.1559E+03, 1.1571E+03, 1.1582E+03, 1.1593E+03, 1.1604E+03,
     +  1.1615E+03, 1.1626E+03, 1.1637E+03, 1.1648E+03, 1.1659E+03,
     + 1.1670E+03, 1.1681E+03, 1.1692E+03, 1.1704E+03, 1.1715E+03,
     +  1.1726E+03, 1.1737E+03, 1.1748E+03, 1.1759E+03, 1.1770E+03,
     + 1.1782E+03, 1.1793E+03, 1.1804E+03, 1.1815E+03, 1.1826E+03,
     +  1.1838E+03, 1.1849E+03, 1.1860E+03, 1.1871E+03, 1.1882E+03,
     + 1.1894E+03, 1.1905E+03, 1.1916E+03, 1.1927E+03, 1.1939E+03,
     +  1.1950E+03, 1.1961E+03, 1.1972E+03, 1.1984E+03, 1.1995E+03,
     + 1.2006E+03, 1.2017E+03, 1.2029E+03, 1.2040E+03, 1.2051E+03,
     +  1.2063E+03, 1.2074E+03, 1.2085E+03, 1.2097E+03, 1.2108E+03,
     + 1.2119E+03, 1.2131E+03, 1.2142E+03, 1.2153E+03, 1.2165E+03,
     +  1.2176E+03, 1.2187E+03, 1.2199E+03, 1.2210E+03, 1.2222E+03,
     + 1.2233E+03, 1.2244E+03, 1.2256E+03, 1.2267E+03, 1.2279E+03,
     +  1.2290E+03, 1.2301E+03, 1.2313E+03, 1.2324E+03, 1.2336E+03,
     + 1.2347E+03, 1.2359E+03, 1.2370E+03, 1.2382E+03, 1.2393E+03,
     +  1.2405E+03, 1.2416E+03, 1.2428E+03, 1.2439E+03, 1.2451E+03,
     + 1.2462E+03, 1.2474E+03, 1.2485E+03, 1.2497E+03, 1.2508E+03,
     +  1.2520E+03, 1.2531E+03, 1.2543E+03, 1.2555E+03, 1.2566E+03,
     + 1.2578E+03, 1.2589E+03, 1.2601E+03, 1.2612E+03, 1.2624E+03,
     +  1.2636E+03, 1.2647E+03, 1.2659E+03, 1.2671E+03, 1.2682E+03,
     + 1.2694E+03, 1.2705E+03, 1.2717E+03, 1.2729E+03, 1.2740E+03,
     +  1.2752E+03, 1.2764E+03, 1.2775E+03, 1.2787E+03, 1.2799E+03,
     + 1.2810E+03, 1.2822E+03, 1.2834E+03, 1.2846E+03, 1.2857E+03,
     +  1.2869E+03, 1.2881E+03, 1.2892E+03, 1.2904E+03, 1.2916E+03,
     + 1.2928E+03, 1.2939E+03, 1.2951E+03, 1.2963E+03, 1.2975E+03,
     +  1.2987E+03, 1.2998E+03, 1.3010E+03, 1.3022E+03, 1.3034E+03,
     + 1.3046E+03, 1.3057E+03, 1.3069E+03, 1.3081E+03, 1.3093E+03,
     +  1.3105E+03, 1.3117E+03, 1.3128E+03, 1.3140E+03, 1.3152E+03,
     + 1.3164E+03, 1.3176E+03, 1.3188E+03, 1.3200E+03, 1.3212E+03,
     +  1.3224E+03, 1.3235E+03, 1.3247E+03, 1.3259E+03, 1.3271E+03,
     + 1.3283E+03, 1.3295E+03, 1.3307E+03, 1.3319E+03, 1.3331E+03,
     +  1.3343E+03, 1.3355E+03, 1.3367E+03, 1.3379E+03, 1.3391E+03,
     + 1.3403E+03, 1.3415E+03, 1.3427E+03, 1.3439E+03, 1.3451E+03,
     +  1.3463E+03, 1.3475E+03, 1.3487E+03, 1.3499E+03, 1.3511E+03,
     + 1.3523E+03, 1.3535E+03, 1.3547E+03, 1.3559E+03, 1.3571E+03,
     +  1.3583E+03, 1.3596E+03, 1.3608E+03, 1.3620E+03, 1.3632E+03,
     + 1.3644E+03, 1.3656E+03, 1.3668E+03, 1.3680E+03, 1.3693E+03,
     +  1.3705E+03, 1.3717E+03, 1.3729E+03, 1.3741E+03, 1.3753E+03,
     + 1.3765E+03, 1.3778E+03, 1.3790E+03, 1.3802E+03, 1.3814E+03,
     +  1.3826E+03, 1.3839E+03, 1.3851E+03, 1.3863E+03, 1.3875E+03,
     + 1.3888E+03, 1.3900E+03, 1.3912E+03, 1.3924E+03, 1.3937E+03,
     +  1.3949E+03, 1.3961E+03, 1.3973E+03, 1.3986E+03, 1.3998E+03,
     + 1.4010E+03, 1.4023E+03, 1.4035E+03, 1.4047E+03, 1.4060E+03,
     +  1.4072E+03, 1.4084E+03, 1.4097E+03, 1.4109E+03, 1.4121E+03,
     + 1.4134E+03, 1.4146E+03, 1.4158E+03, 1.4171E+03, 1.4183E+03,
     +  1.4196E+03, 1.4208E+03, 1.4220E+03, 1.4233E+03, 1.4245E+03,
     + 1.4258E+03, 1.4270E+03, 1.4282E+03, 1.4295E+03, 1.4307E+03,
     +  1.4320E+03, 1.4332E+03, 1.4345E+03, 1.4357E+03, 1.4370E+03,
     + 1.4382E+03, 1.4395E+03, 1.4407E+03, 1.4420E+03, 1.4432E+03,
     +  1.4445E+03, 1.4457E+03, 1.4470E+03, 1.4482E+03, 1.4495E+03,
     + 1.4507E+03, 1.4520E+03, 1.4532E+03, 1.4545E+03, 1.4558E+03,
     +  1.4570E+03, 1.4583E+03, 1.4595E+03, 1.4608E+03, 1.4620E+03,
     + 1.4633E+03, 1.4646E+03, 1.4658E+03, 1.4671E+03, 1.4684E+03,
     +  1.4696E+03, 1.4709E+03, 1.4721E+03, 1.4734E+03, 1.4747E+03,
     + 1.4759E+03, 1.4772E+03, 1.4785E+03, 1.4797E+03, 1.4810E+03,
     +  1.4823E+03, 1.4836E+03, 1.4848E+03, 1.4861E+03, 1.4874E+03,
     + 1.4886E+03, 1.4899E+03, 1.4912E+03, 1.4925E+03, 1.4937E+03,
     +  1.4950E+03, 1.4963E+03, 1.4976E+03, 1.4989E+03, 1.5001E+03,
     + 1.5014E+03, 1.5027E+03, 1.5040E+03, 1.5053E+03, 1.5065E+03,
     +  1.5078E+03, 1.5091E+03, 1.5104E+03, 1.5117E+03, 1.5129E+03,
     + 1.5142E+03, 1.5155E+03, 1.5168E+03, 1.5181E+03, 1.5194E+03,
     +  1.5207E+03, 1.5220E+03, 1.5233E+03, 1.5245E+03, 1.5258E+03,
     + 1.5271E+03, 1.5284E+03, 1.5297E+03, 1.5310E+03, 1.5323E+03,
     +  1.5336E+03, 1.5349E+03, 1.5362E+03, 1.5375E+03, 1.5388E+03,
     + 1.5401E+03, 1.5414E+03, 1.5427E+03, 1.5440E+03, 1.5453E+03,
     +  1.5466E+03, 1.5479E+03, 1.5492E+03, 1.5505E+03, 1.5518E+03,
     + 1.5531E+03, 1.5544E+03, 1.5557E+03, 1.5570E+03, 1.5583E+03,
     +  1.5596E+03, 1.5609E+03, 1.5622E+03, 1.5635E+03, 1.5649E+03,
     + 1.5662E+03, 1.5675E+03, 1.5688E+03, 1.5701E+03, 1.5714E+03,
     +  1.5727E+03, 1.5740E+03, 1.5754E+03, 1.5767E+03, 1.5780E+03,
     + 1.5793E+03, 1.5806E+03, 1.5819E+03, 1.5833E+03, 1.5846E+03,
     +  1.5859E+03, 1.5872E+03, 1.5885E+03, 1.5899E+03, 1.5912E+03,
     + 1.5925E+03, 1.5938E+03, 1.5952E+03, 1.5965E+03, 1.5978E+03,
     +  1.5991E+03, 1.6005E+03, 1.6018E+03, 1.6031E+03, 1.6044E+03,
     + 1.6058E+03, 1.6071E+03, 1.6084E+03, 1.6098E+03, 1.6111E+03,
     +  1.6124E+03, 1.6138E+03, 1.6151E+03, 1.6164E+03, 1.6178E+03,
     + 1.6191E+03, 1.6204E+03, 1.6218E+03, 1.6231E+03, 1.6245E+03,
     +  1.6258E+03, 1.6271E+03, 1.6285E+03, 1.6298E+03, 1.6312E+03,
     + 1.6325E+03, 1.6338E+03, 1.6352E+03, 1.6365E+03, 1.6379E+03,
     +  1.6392E+03, 1.6406E+03, 1.6419E+03, 1.6433E+03, 1.6446E+03,
     + 1.6459E+03, 1.6473E+03, 1.6486E+03, 1.6500E+03, 1.6513E+03,
     +  1.6527E+03, 1.6541E+03, 1.6554E+03, 1.6568E+03, 1.6581E+03,
     + 1.6595E+03, 1.6608E+03, 1.6622E+03, 1.6635E+03, 1.6649E+03,
     +  1.6663E+03, 1.6676E+03, 1.6690E+03, 1.6703E+03, 1.6717E+03,
     + 1.6731E+03, 1.6744E+03, 1.6758E+03, 1.6771E+03, 1.6785E+03,
     +  1.6799E+03, 1.6812E+03, 1.6826E+03, 1.6840E+03, 1.6853E+03,
     + 1.6867E+03, 1.6881E+03, 1.6894E+03, 1.6908E+03, 1.6922E+03,
     +  1.6936E+03, 1.6949E+03, 1.6963E+03, 1.6977E+03, 1.6990E+03,
     + 1.7004E+03, 1.7018E+03, 1.7032E+03, 1.7045E+03, 1.7059E+03,
     +  1.7073E+03, 1.7087E+03, 1.7101E+03, 1.7114E+03, 1.7128E+03,
     + 1.7142E+03, 1.7156E+03, 1.7170E+03, 1.7183E+03, 1.7197E+03,
     +  1.7211E+03, 1.7225E+03, 1.7239E+03, 1.7253E+03, 1.7266E+03,
     + 1.7280E+03, 1.7294E+03, 1.7308E+03, 1.7322E+03, 1.7336E+03,
     +  1.7350E+03, 1.7364E+03, 1.7378E+03, 1.7392E+03, 1.7405E+03,
     + 1.7419E+03, 1.7433E+03, 1.7447E+03, 1.7461E+03, 1.7475E+03,
     +  1.7489E+03, 1.7503E+03, 1.7517E+03, 1.7531E+03, 1.7545E+03,
     + 1.7559E+03, 1.7573E+03, 1.7587E+03, 1.7601E+03, 1.7615E+03,
     +  1.7629E+03, 1.7643E+03, 1.7657E+03, 1.7671E+03, 1.7685E+03,
     + 1.7699E+03, 1.7714E+03, 1.7728E+03, 1.7742E+03, 1.7756E+03,
     +  1.7770E+03, 1.7784E+03, 1.7798E+03, 1.7812E+03, 1.7826E+03,
     + 1.7840E+03, 1.7855E+03, 1.7869E+03, 1.7883E+03, 1.7897E+03,
     +  1.7911E+03, 1.7925E+03, 1.7940E+03, 1.7954E+03, 1.7968E+03,
     + 1.7982E+03, 1.7996E+03, 1.8011E+03, 1.8025E+03, 1.8039E+03,
     +  1.8053E+03, 1.8067E+03, 1.8082E+03, 1.8096E+03, 1.8110E+03,
     + 1.8124E+03, 1.8139E+03, 1.8153E+03, 1.8167E+03, 1.8182E+03,
     +  1.8196E+03, 1.8210E+03, 1.8224E+03, 1.8239E+03, 1.8253E+03,
     + 1.8267E+03, 1.8282E+03, 1.8296E+03, 1.8310E+03, 1.8325E+03,
     +  1.8339E+03, 1.8354E+03, 1.8368E+03, 1.8382E+03, 1.8397E+03,
     + 1.8411E+03, 1.8425E+03, 1.8440E+03, 1.8454E+03, 1.8469E+03,
     +  1.8483E+03, 1.8498E+03, 1.8512E+03, 1.8526E+03, 1.8541E+03,
     + 1.8555E+03, 1.8570E+03, 1.8584E+03, 1.8599E+03, 1.8613E+03,
     +  1.8628E+03, 1.8642E+03, 1.8657E+03, 1.8671E+03, 1.8686E+03,
     + 1.8700E+03, 1.8715E+03, 1.8729E+03, 1.8744E+03, 1.8759E+03,
     +  1.8773E+03, 1.8788E+03, 1.8802E+03, 1.8817E+03, 1.8831E+03,
     + 1.8846E+03, 1.8861E+03, 1.8875E+03, 1.8890E+03, 1.8904E+03,
     +  1.8919E+03, 1.8934E+03, 1.8948E+03, 1.8963E+03, 1.8978E+03,
     + 1.8992E+03, 1.9007E+03, 1.9022E+03, 1.9036E+03, 1.9051E+03,
     +  1.9066E+03, 1.9080E+03, 1.9095E+03, 1.9110E+03, 1.9124E+03,
     + 1.9139E+03, 1.9154E+03, 1.9169E+03, 1.9183E+03, 1.9198E+03,
     +  1.9213E+03, 1.9228E+03, 1.9242E+03, 1.9257E+03, 1.9272E+03,
     + 1.9287E+03, 1.9302E+03, 1.9316E+03, 1.9331E+03, 1.9346E+03,
     +  1.9361E+03, 1.9376E+03, 1.9391E+03, 1.9405E+03, 1.9420E+03,
     + 1.9435E+03, 1.9450E+03, 1.9465E+03, 1.9480E+03, 1.9495E+03,
     +  1.9509E+03, 1.9524E+03, 1.9539E+03, 1.9554E+03, 1.9569E+03,
     + 1.9584E+03, 1.9599E+03, 1.9614E+03, 1.9629E+03, 1.9644E+03,
     +  1.9659E+03, 1.9674E+03, 1.9689E+03, 1.9704E+03, 1.9719E+03,
     + 1.9734E+03, 1.9749E+03, 1.9764E+03, 1.9779E+03, 1.9794E+03,
     +  1.9809E+03, 1.9824E+03, 1.9839E+03, 1.9854E+03, 1.9869E+03,
     + 1.9884E+03, 1.9899E+03, 1.9914E+03, 1.9929E+03, 1.9944E+03,
     +  1.9959E+03, 1.9974E+03, 1.9990E+03, 2.0005E+03, 2.0020E+03,
     + 2.0035E+03, 2.0050E+03, 2.0065E+03, 2.0080E+03, 2.0096E+03,
     +  2.0111E+03, 2.0126E+03, 2.0141E+03, 2.0156E+03, 2.0171E+03,
     + 2.0187E+03, 2.0202E+03, 2.0217E+03, 2.0232E+03, 2.0247E+03,
     +  2.0263E+03, 2.0278E+03, 2.0293E+03, 2.0308E+03, 2.0324E+03,
     + 2.0339E+03, 2.0354E+03, 2.0370E+03, 2.0385E+03, 2.0400E+03,
     +  2.0415E+03, 2.0431E+03, 2.0446E+03, 2.0461E+03, 2.0477E+03,
     + 2.0492E+03, 2.0507E+03, 2.0523E+03, 2.0538E+03, 2.0553E+03,
     +  2.0569E+03, 2.0584E+03, 2.0599E+03, 2.0615E+03, 2.0630E+03,
     + 2.0646E+03, 2.0661E+03, 2.0676E+03, 2.0692E+03, 2.0707E+03,
     +  2.0723E+03, 2.0738E+03, 2.0754E+03, 2.0769E+03, 2.0785E+03,
     + 2.0800E+03, 2.0816E+03, 2.0831E+03, 2.0846E+03, 2.0862E+03,
     +  2.0877E+03, 2.0893E+03, 2.0909E+03, 2.0924E+03, 2.0940E+03,
     + 2.0955E+03, 2.0971E+03, 2.0986E+03, 2.1002E+03, 2.1017E+03,
     +  2.1033E+03, 2.1048E+03, 2.1064E+03, 2.1080E+03, 2.1095E+03,
     + 2.1111E+03, 2.1126E+03, 2.1142E+03, 2.1158E+03, 2.1173E+03,
     +  2.1189E+03, 2.1205E+03, 2.1220E+03, 2.1236E+03, 2.1252E+03,
     + 2.1267E+03, 2.1283E+03, 2.1299E+03, 2.1314E+03, 2.1330E+03,
     +  2.1346E+03, 2.1361E+03, 2.1377E+03, 2.1393E+03, 2.1409E+03,
     + 2.1424E+03, 2.1440E+03, 2.1456E+03, 2.1472E+03, 2.1487E+03,
     +  2.1503E+03, 2.1519E+03, 2.1535E+03, 2.1551E+03, 2.1566E+03,
     + 2.1582E+03, 2.1598E+03, 2.1614E+03, 2.1630E+03, 2.1646E+03,
     +  2.1661E+03, 2.1677E+03, 2.1693E+03, 2.1709E+03, 2.1725E+03,
     + 2.1741E+03, 2.1757E+03, 2.1773E+03, 2.1788E+03, 2.1804E+03,
     +  2.1820E+03, 2.1836E+03, 2.1852E+03, 2.1868E+03, 2.1884E+03,
     + 2.1900E+03, 2.1916E+03, 2.1932E+03, 2.1948E+03, 2.1964E+03,
     +  2.1980E+03, 2.1996E+03, 2.2012E+03, 2.2028E+03, 2.2044E+03,
     + 2.2060E+03, 2.2076E+03, 2.2092E+03, 2.2108E+03, 2.2124E+03,
     +  2.2140E+03, 2.2156E+03, 2.2172E+03, 2.2188E+03, 2.2204E+03,
     + 2.2220E+03, 2.2237E+03, 2.2253E+03, 2.2269E+03, 2.2285E+03,
     +  2.2301E+03, 2.2317E+03, 2.2333E+03, 2.2349E+03, 2.2366E+03,
     + 2.2382E+03, 2.2398E+03, 2.2414E+03, 2.2430E+03, 2.2446E+03,
     +  2.2463E+03, 2.2479E+03, 2.2495E+03, 2.2511E+03, 2.2528E+03,
     + 2.2544E+03, 2.2560E+03, 2.2576E+03, 2.2592E+03, 2.2609E+03,
     +  2.2625E+03, 2.2641E+03, 2.2658E+03, 2.2674E+03, 2.2690E+03,
     + 2.2706E+03, 2.2723E+03, 2.2739E+03, 2.2755E+03, 2.2772E+03,
     +  2.2788E+03, 2.2804E+03, 2.2821E+03, 2.2837E+03, 2.2853E+03,
     + 2.2870E+03, 2.2886E+03, 2.2903E+03, 2.2919E+03, 2.2935E+03,
     +  2.2952E+03, 2.2968E+03, 2.2985E+03, 2.3001E+03, 2.3018E+03,
     + 2.3034E+03, 2.3050E+03, 2.3067E+03, 2.3083E+03, 2.3100E+03,
     +  2.3116E+03, 2.3133E+03, 2.3149E+03, 2.3166E+03, 2.3182E+03,
     + 2.3199E+03, 2.3215E+03, 2.3232E+03, 2.3248E+03, 2.3265E+03,
     +  2.3281E+03, 2.3298E+03, 2.3315E+03, 2.3331E+03, 2.3348E+03,
     + 2.3364E+03, 2.3381E+03, 2.3397E+03, 2.3414E+03, 2.3431E+03,
     +  2.3447E+03, 2.3464E+03, 2.3481E+03, 2.3497E+03, 2.3514E+03/
      data (QofT( 1,J),J=2409,4510)/ 2.3531E+03, 2.3547E+03,
     +  2.3564E+03, 2.3581E+03, 2.3597E+03, 2.3614E+03, 2.3631E+03,
     +  2.3647E+03,
     + 2.3664E+03, 2.3681E+03, 2.3697E+03, 2.3714E+03, 2.3731E+03,
     +  2.3748E+03, 2.3764E+03, 2.3781E+03, 2.3798E+03, 2.3815E+03,
     + 2.3832E+03, 2.3848E+03, 2.3865E+03, 2.3882E+03, 2.3899E+03,
     +  2.3916E+03, 2.3932E+03, 2.3949E+03, 2.3966E+03, 2.3983E+03,
     + 2.4000E+03, 2.4017E+03, 2.4034E+03, 2.4050E+03, 2.4067E+03,
     +  2.4084E+03, 2.4101E+03, 2.4118E+03, 2.4135E+03, 2.4152E+03,
     + 2.4169E+03, 2.4186E+03, 2.4203E+03, 2.4220E+03, 2.4237E+03,
     +  2.4254E+03, 2.4270E+03, 2.4287E+03, 2.4304E+03, 2.4321E+03,
     + 2.4338E+03, 2.4355E+03, 2.4372E+03, 2.4389E+03, 2.4407E+03,
     +  2.4424E+03, 2.4441E+03, 2.4458E+03, 2.4475E+03, 2.4492E+03,
     + 2.4509E+03, 2.4526E+03, 2.4543E+03, 2.4560E+03, 2.4577E+03,
     +  2.4594E+03, 2.4611E+03, 2.4629E+03, 2.4646E+03, 2.4663E+03,
     + 2.4680E+03, 2.4697E+03, 2.4714E+03, 2.4731E+03, 2.4749E+03,
     +  2.4766E+03, 2.4783E+03, 2.4800E+03, 2.4817E+03, 2.4835E+03,
     + 2.4852E+03, 2.4869E+03, 2.4886E+03, 2.4903E+03, 2.4921E+03,
     +  2.4938E+03, 2.4955E+03, 2.4973E+03, 2.4990E+03, 2.5007E+03,
     + 2.5024E+03, 2.5042E+03, 2.5059E+03, 2.5076E+03, 2.5094E+03,
     +  2.5111E+03, 2.5128E+03, 2.5146E+03, 2.5163E+03, 2.5180E+03,
     + 2.5198E+03, 2.5215E+03, 2.5232E+03, 2.5250E+03, 2.5267E+03,
     +  2.5285E+03, 2.5302E+03, 2.5319E+03, 2.5337E+03, 2.5354E+03,
     + 2.5372E+03, 2.5389E+03, 2.5407E+03, 2.5424E+03, 2.5441E+03,
     +  2.5459E+03, 2.5476E+03, 2.5494E+03, 2.5511E+03, 2.5529E+03,
     + 2.5546E+03, 2.5564E+03, 2.5581E+03, 2.5599E+03, 2.5617E+03,
     +  2.5634E+03, 2.5652E+03, 2.5669E+03, 2.5687E+03, 2.5704E+03,
     + 2.5722E+03, 2.5739E+03, 2.5757E+03, 2.5775E+03, 2.5792E+03,
     +  2.5810E+03, 2.5828E+03, 2.5845E+03, 2.5863E+03, 2.5880E+03,
     + 2.5898E+03, 2.5916E+03, 2.5933E+03, 2.5951E+03, 2.5969E+03,
     +  2.5987E+03, 2.6004E+03, 2.6022E+03, 2.6040E+03, 2.6057E+03,
     + 2.6075E+03, 2.6093E+03, 2.6111E+03, 2.6128E+03, 2.6146E+03,
     +  2.6164E+03, 2.6182E+03, 2.6199E+03, 2.6217E+03, 2.6235E+03,
     + 2.6253E+03, 2.6271E+03, 2.6288E+03, 2.6306E+03, 2.6324E+03,
     +  2.6342E+03, 2.6360E+03, 2.6378E+03, 2.6395E+03, 2.6413E+03,
     + 2.6431E+03, 2.6449E+03, 2.6467E+03, 2.6485E+03, 2.6503E+03,
     +  2.6521E+03, 2.6539E+03, 2.6557E+03, 2.6574E+03, 2.6592E+03,
     + 2.6610E+03, 2.6628E+03, 2.6646E+03, 2.6664E+03, 2.6682E+03,
     +  2.6700E+03, 2.6718E+03, 2.6736E+03, 2.6754E+03, 2.6772E+03,
     + 2.6790E+03, 2.6808E+03, 2.6826E+03, 2.6844E+03, 2.6862E+03,
     +  2.6881E+03, 2.6899E+03, 2.6917E+03, 2.6935E+03, 2.6953E+03,
     + 2.6971E+03, 2.6989E+03, 2.7007E+03, 2.7025E+03, 2.7043E+03,
     +  2.7062E+03, 2.7080E+03, 2.7098E+03, 2.7116E+03, 2.7134E+03,
     + 2.7152E+03, 2.7171E+03, 2.7189E+03, 2.7207E+03, 2.7225E+03,
     +  2.7243E+03, 2.7262E+03, 2.7280E+03, 2.7298E+03, 2.7316E+03,
     + 2.7334E+03, 2.7353E+03, 2.7371E+03, 2.7389E+03, 2.7408E+03,
     +  2.7426E+03, 2.7444E+03, 2.7462E+03, 2.7481E+03, 2.7499E+03,
     + 2.7517E+03, 2.7536E+03, 2.7554E+03, 2.7572E+03, 2.7591E+03,
     +  2.7609E+03, 2.7627E+03, 2.7646E+03, 2.7664E+03, 2.7683E+03,
     + 2.7701E+03, 2.7719E+03, 2.7738E+03, 2.7756E+03, 2.7775E+03,
     +  2.7793E+03, 2.7812E+03, 2.7830E+03, 2.7849E+03, 2.7867E+03,
     + 2.7885E+03, 2.7904E+03, 2.7922E+03, 2.7941E+03, 2.7959E+03,
     +  2.7978E+03, 2.7996E+03, 2.8015E+03, 2.8034E+03, 2.8052E+03,
     + 2.8071E+03, 2.8089E+03, 2.8108E+03, 2.8126E+03, 2.8145E+03,
     +  2.8163E+03, 2.8182E+03, 2.8201E+03, 2.8219E+03, 2.8238E+03,
     + 2.8257E+03, 2.8275E+03, 2.8294E+03, 2.8312E+03, 2.8331E+03,
     +  2.8350E+03, 2.8368E+03, 2.8387E+03, 2.8406E+03, 2.8424E+03,
     + 2.8443E+03, 2.8462E+03, 2.8481E+03, 2.8499E+03, 2.8518E+03,
     +  2.8537E+03, 2.8556E+03, 2.8574E+03, 2.8593E+03, 2.8612E+03,
     + 2.8631E+03, 2.8649E+03, 2.8668E+03, 2.8687E+03, 2.8706E+03,
     +  2.8725E+03, 2.8743E+03, 2.8762E+03, 2.8781E+03, 2.8800E+03,
     + 2.8819E+03, 2.8838E+03, 2.8857E+03, 2.8875E+03, 2.8894E+03,
     +  2.8913E+03, 2.8932E+03, 2.8951E+03, 2.8970E+03, 2.8989E+03,
     + 2.9008E+03, 2.9027E+03, 2.9046E+03, 2.9065E+03, 2.9084E+03,
     +  2.9103E+03, 2.9121E+03, 2.9140E+03, 2.9159E+03, 2.9178E+03,
     + 2.9197E+03, 2.9216E+03, 2.9235E+03, 2.9255E+03, 2.9274E+03,
     +  2.9293E+03, 2.9312E+03, 2.9331E+03, 2.9350E+03, 2.9369E+03,
     + 2.9388E+03, 2.9407E+03, 2.9426E+03, 2.9445E+03, 2.9464E+03,
     +  2.9483E+03, 2.9503E+03, 2.9522E+03, 2.9541E+03, 2.9560E+03,
     + 2.9579E+03, 2.9598E+03, 2.9618E+03, 2.9637E+03, 2.9656E+03,
     +  2.9675E+03, 2.9694E+03, 2.9714E+03, 2.9733E+03, 2.9752E+03,
     + 2.9771E+03, 2.9790E+03, 2.9810E+03, 2.9829E+03, 2.9848E+03,
     +  2.9868E+03, 2.9887E+03, 2.9906E+03, 2.9925E+03, 2.9945E+03,
     + 2.9964E+03, 2.9983E+03, 3.0003E+03, 3.0022E+03, 3.0041E+03,
     +  3.0061E+03, 3.0080E+03, 3.0099E+03, 3.0119E+03, 3.0138E+03,
     + 3.0158E+03, 3.0177E+03, 3.0196E+03, 3.0216E+03, 3.0235E+03,
     +  3.0255E+03, 3.0274E+03, 3.0294E+03, 3.0313E+03, 3.0332E+03,
     + 3.0352E+03, 3.0371E+03, 3.0391E+03, 3.0410E+03, 3.0430E+03,
     +  3.0449E+03, 3.0469E+03, 3.0488E+03, 3.0508E+03, 3.0527E+03,
     + 3.0547E+03, 3.0567E+03, 3.0586E+03, 3.0606E+03, 3.0625E+03,
     +  3.0645E+03, 3.0664E+03, 3.0684E+03, 3.0704E+03, 3.0723E+03,
     + 3.0743E+03, 3.0763E+03, 3.0782E+03, 3.0802E+03, 3.0821E+03,
     +  3.0841E+03, 3.0861E+03, 3.0880E+03, 3.0900E+03, 3.0920E+03,
     + 3.0940E+03, 3.0959E+03, 3.0979E+03, 3.0999E+03, 3.1018E+03,
     +  3.1038E+03, 3.1058E+03, 3.1078E+03, 3.1097E+03, 3.1117E+03,
     + 3.1137E+03, 3.1157E+03, 3.1177E+03, 3.1196E+03, 3.1216E+03,
     +  3.1236E+03, 3.1256E+03, 3.1276E+03, 3.1296E+03, 3.1315E+03,
     + 3.1335E+03, 3.1355E+03, 3.1375E+03, 3.1395E+03, 3.1415E+03,
     +  3.1435E+03, 3.1455E+03, 3.1474E+03, 3.1494E+03, 3.1514E+03,
     + 3.1534E+03, 3.1554E+03, 3.1574E+03, 3.1594E+03, 3.1614E+03,
     +  3.1634E+03, 3.1654E+03, 3.1674E+03, 3.1694E+03, 3.1714E+03,
     + 3.1734E+03, 3.1754E+03, 3.1774E+03, 3.1794E+03, 3.1814E+03,
     +  3.1834E+03, 3.1854E+03, 3.1874E+03, 3.1894E+03, 3.1915E+03,
     + 3.1935E+03, 3.1955E+03, 3.1975E+03, 3.1995E+03, 3.2015E+03,
     +  3.2035E+03, 3.2055E+03, 3.2075E+03, 3.2096E+03, 3.2116E+03,
     + 3.2136E+03, 3.2156E+03, 3.2176E+03, 3.2197E+03, 3.2217E+03,
     +  3.2237E+03, 3.2257E+03, 3.2277E+03, 3.2298E+03, 3.2318E+03,
     + 3.2338E+03, 3.2358E+03, 3.2379E+03, 3.2399E+03, 3.2419E+03,
     +  3.2440E+03, 3.2460E+03, 3.2480E+03, 3.2500E+03, 3.2521E+03,
     + 3.2541E+03, 3.2561E+03, 3.2582E+03, 3.2602E+03, 3.2623E+03,
     +  3.2643E+03, 3.2663E+03, 3.2684E+03, 3.2704E+03, 3.2724E+03,
     + 3.2745E+03, 3.2765E+03, 3.2786E+03, 3.2806E+03, 3.2827E+03,
     +  3.2847E+03, 3.2867E+03, 3.2888E+03, 3.2908E+03, 3.2929E+03,
     + 3.2949E+03, 3.2970E+03, 3.2990E+03, 3.3011E+03, 3.3031E+03,
     +  3.3052E+03, 3.3072E+03, 3.3093E+03, 3.3114E+03, 3.3134E+03,
     + 3.3155E+03, 3.3175E+03, 3.3196E+03, 3.3216E+03, 3.3237E+03,
     +  3.3258E+03, 3.3278E+03, 3.3299E+03, 3.3320E+03, 3.3340E+03,
     + 3.3361E+03, 3.3382E+03, 3.3402E+03, 3.3423E+03, 3.3444E+03,
     +  3.3464E+03, 3.3485E+03, 3.3506E+03, 3.3526E+03, 3.3547E+03,
     + 3.3568E+03, 3.3589E+03, 3.3609E+03, 3.3630E+03, 3.3651E+03,
     +  3.3672E+03, 3.3692E+03, 3.3713E+03, 3.3734E+03, 3.3755E+03,
     + 3.3776E+03, 3.3796E+03, 3.3817E+03, 3.3838E+03, 3.3859E+03,
     +  3.3880E+03, 3.3901E+03, 3.3921E+03, 3.3942E+03, 3.3963E+03,
     + 3.3984E+03, 3.4005E+03, 3.4026E+03, 3.4047E+03, 3.4068E+03,
     +  3.4089E+03, 3.4110E+03, 3.4131E+03, 3.4152E+03, 3.4172E+03,
     + 3.4193E+03, 3.4214E+03, 3.4235E+03, 3.4256E+03, 3.4277E+03,
     +  3.4298E+03, 3.4319E+03, 3.4340E+03, 3.4361E+03, 3.4383E+03,
     + 3.4404E+03, 3.4425E+03, 3.4446E+03, 3.4467E+03, 3.4488E+03,
     +  3.4509E+03, 3.4530E+03, 3.4551E+03, 3.4572E+03, 3.4593E+03,
     + 3.4615E+03, 3.4636E+03, 3.4657E+03, 3.4678E+03, 3.4699E+03,
     +  3.4720E+03, 3.4742E+03, 3.4763E+03, 3.4784E+03, 3.4805E+03,
     + 3.4826E+03, 3.4848E+03, 3.4869E+03, 3.4890E+03, 3.4911E+03,
     +  3.4933E+03, 3.4954E+03, 3.4975E+03, 3.4996E+03, 3.5018E+03,
     + 3.5039E+03, 3.5060E+03, 3.5082E+03, 3.5103E+03, 3.5124E+03,
     +  3.5145E+03, 3.5167E+03, 3.5188E+03, 3.5210E+03, 3.5231E+03,
     + 3.5252E+03, 3.5274E+03, 3.5295E+03, 3.5316E+03, 3.5338E+03,
     +  3.5359E+03, 3.5381E+03, 3.5402E+03, 3.5424E+03, 3.5445E+03,
     + 3.5467E+03, 3.5488E+03, 3.5509E+03, 3.5531E+03, 3.5552E+03,
     +  3.5574E+03, 3.5595E+03, 3.5617E+03, 3.5638E+03, 3.5660E+03,
     + 3.5682E+03, 3.5703E+03, 3.5725E+03, 3.5746E+03, 3.5768E+03,
     +  3.5789E+03, 3.5811E+03, 3.5833E+03, 3.5854E+03, 3.5876E+03,
     + 3.5897E+03, 3.5919E+03, 3.5941E+03, 3.5962E+03, 3.5984E+03,
     +  3.6006E+03, 3.6027E+03, 3.6049E+03, 3.6071E+03, 3.6092E+03,
     + 3.6114E+03, 3.6136E+03, 3.6157E+03, 3.6179E+03, 3.6201E+03,
     +  3.6223E+03, 3.6244E+03, 3.6266E+03, 3.6288E+03, 3.6310E+03,
     + 3.6332E+03, 3.6353E+03, 3.6375E+03, 3.6397E+03, 3.6419E+03,
     +  3.6441E+03, 3.6462E+03, 3.6484E+03, 3.6506E+03, 3.6528E+03,
     + 3.6550E+03, 3.6572E+03, 3.6594E+03, 3.6615E+03, 3.6637E+03,
     +  3.6659E+03, 3.6681E+03, 3.6703E+03, 3.6725E+03, 3.6747E+03,
     + 3.6769E+03, 3.6791E+03, 3.6813E+03, 3.6835E+03, 3.6857E+03,
     +  3.6879E+03, 3.6901E+03, 3.6923E+03, 3.6945E+03, 3.6967E+03,
     + 3.6989E+03, 3.7011E+03, 3.7033E+03, 3.7055E+03, 3.7077E+03,
     +  3.7099E+03, 3.7121E+03, 3.7143E+03, 3.7165E+03, 3.7188E+03,
     + 3.7210E+03, 3.7232E+03, 3.7254E+03, 3.7276E+03, 3.7298E+03,
     +  3.7320E+03, 3.7343E+03, 3.7365E+03, 3.7387E+03, 3.7409E+03,
     + 3.7431E+03, 3.7453E+03, 3.7476E+03, 3.7498E+03, 3.7520E+03,
     +  3.7542E+03, 3.7565E+03, 3.7587E+03, 3.7609E+03, 3.7631E+03,
     + 3.7654E+03, 3.7676E+03, 3.7698E+03, 3.7721E+03, 3.7743E+03,
     +  3.7765E+03, 3.7788E+03, 3.7810E+03, 3.7832E+03, 3.7855E+03,
     + 3.7877E+03, 3.7899E+03, 3.7922E+03, 3.7944E+03, 3.7966E+03,
     +  3.7989E+03, 3.8011E+03, 3.8034E+03, 3.8056E+03, 3.8079E+03,
     + 3.8101E+03, 3.8123E+03, 3.8146E+03, 3.8168E+03, 3.8191E+03,
     +  3.8213E+03, 3.8236E+03, 3.8258E+03, 3.8281E+03, 3.8303E+03,
     + 3.8326E+03, 3.8348E+03, 3.8371E+03, 3.8394E+03, 3.8416E+03,
     +  3.8439E+03, 3.8461E+03, 3.8484E+03, 3.8506E+03, 3.8529E+03,
     + 3.8552E+03, 3.8574E+03, 3.8597E+03, 3.8620E+03, 3.8642E+03,
     +  3.8665E+03, 3.8688E+03, 3.8710E+03, 3.8733E+03, 3.8756E+03,
     + 3.8778E+03, 3.8801E+03, 3.8824E+03, 3.8846E+03, 3.8869E+03,
     +  3.8892E+03, 3.8915E+03, 3.8937E+03, 3.8960E+03, 3.8983E+03,
     + 3.9006E+03, 3.9029E+03, 3.9051E+03, 3.9074E+03, 3.9097E+03,
     +  3.9120E+03, 3.9143E+03, 3.9165E+03, 3.9188E+03, 3.9211E+03,
     + 3.9234E+03, 3.9257E+03, 3.9280E+03, 3.9303E+03, 3.9326E+03,
     +  3.9348E+03, 3.9371E+03, 3.9394E+03, 3.9417E+03, 3.9440E+03,
     + 3.9463E+03, 3.9486E+03, 3.9509E+03, 3.9532E+03, 3.9555E+03,
     +  3.9578E+03, 3.9601E+03, 3.9624E+03, 3.9647E+03, 3.9670E+03,
     + 3.9693E+03, 3.9716E+03, 3.9739E+03, 3.9762E+03, 3.9785E+03,
     +  3.9808E+03, 3.9831E+03, 3.9855E+03, 3.9878E+03, 3.9901E+03,
     + 3.9924E+03, 3.9947E+03, 3.9970E+03, 3.9993E+03, 4.0016E+03,
     +  4.0040E+03, 4.0063E+03, 4.0086E+03, 4.0109E+03, 4.0132E+03,
     + 4.0155E+03, 4.0179E+03, 4.0202E+03, 4.0225E+03, 4.0248E+03,
     +  4.0272E+03, 4.0295E+03, 4.0318E+03, 4.0341E+03, 4.0365E+03,
     + 4.0388E+03, 4.0411E+03, 4.0435E+03, 4.0458E+03, 4.0481E+03,
     +  4.0505E+03, 4.0528E+03, 4.0551E+03, 4.0575E+03, 4.0598E+03,
     + 4.0621E+03, 4.0645E+03, 4.0668E+03, 4.0691E+03, 4.0715E+03,
     +  4.0738E+03, 4.0762E+03, 4.0785E+03, 4.0809E+03, 4.0832E+03,
     + 4.0855E+03, 4.0879E+03, 4.0902E+03, 4.0926E+03, 4.0949E+03,
     +  4.0973E+03, 4.0996E+03, 4.1020E+03, 4.1043E+03, 4.1067E+03,
     + 4.1091E+03, 4.1114E+03, 4.1138E+03, 4.1161E+03, 4.1185E+03,
     +  4.1208E+03, 4.1232E+03, 4.1256E+03, 4.1279E+03, 4.1303E+03,
     + 4.1326E+03, 4.1350E+03, 4.1374E+03, 4.1397E+03, 4.1421E+03,
     +  4.1445E+03, 4.1468E+03, 4.1492E+03, 4.1516E+03, 4.1539E+03,
     + 4.1563E+03, 4.1587E+03, 4.1611E+03, 4.1634E+03, 4.1658E+03,
     +  4.1682E+03, 4.1706E+03, 4.1729E+03, 4.1753E+03, 4.1777E+03,
     + 4.1801E+03, 4.1825E+03, 4.1848E+03, 4.1872E+03, 4.1896E+03,
     +  4.1920E+03, 4.1944E+03, 4.1968E+03, 4.1991E+03, 4.2015E+03,
     + 4.2039E+03, 4.2063E+03, 4.2087E+03, 4.2111E+03, 4.2135E+03,
     +  4.2159E+03, 4.2183E+03, 4.2207E+03, 4.2231E+03, 4.2254E+03,
     + 4.2278E+03, 4.2302E+03, 4.2326E+03, 4.2350E+03, 4.2374E+03,
     +  4.2398E+03, 4.2422E+03, 4.2446E+03, 4.2471E+03, 4.2495E+03,
     + 4.2519E+03, 4.2543E+03, 4.2567E+03, 4.2591E+03, 4.2615E+03,
     +  4.2639E+03, 4.2663E+03, 4.2687E+03, 4.2711E+03, 4.2735E+03,
     + 4.2760E+03, 4.2784E+03, 4.2808E+03, 4.2832E+03, 4.2856E+03,
     +  4.2880E+03, 4.2905E+03, 4.2929E+03, 4.2953E+03, 4.2977E+03,
     + 4.3001E+03, 4.3026E+03, 4.3050E+03, 4.3074E+03, 4.3098E+03,
     +  4.3123E+03, 4.3147E+03, 4.3171E+03, 4.3196E+03, 4.3220E+03,
     + 4.3244E+03, 4.3269E+03, 4.3293E+03, 4.3317E+03, 4.3342E+03,
     +  4.3366E+03, 4.3390E+03, 4.3415E+03, 4.3439E+03, 4.3463E+03,
     + 4.3488E+03, 4.3512E+03, 4.3537E+03, 4.3561E+03, 4.3585E+03,
     +  4.3610E+03, 4.3634E+03, 4.3659E+03, 4.3683E+03, 4.3708E+03,
     + 4.3732E+03, 4.3757E+03, 4.3781E+03, 4.3806E+03, 4.3830E+03,
     +  4.3855E+03, 4.3879E+03, 4.3904E+03, 4.3928E+03, 4.3953E+03,
     + 4.3978E+03, 4.4002E+03, 4.4027E+03, 4.4051E+03, 4.4076E+03,
     +  4.4101E+03, 4.4125E+03, 4.4150E+03, 4.4174E+03, 4.4199E+03,
     + 4.4224E+03, 4.4248E+03, 4.4273E+03, 4.4298E+03, 4.4323E+03,
     +  4.4347E+03, 4.4372E+03, 4.4397E+03, 4.4421E+03, 4.4446E+03,
     + 4.4471E+03, 4.4496E+03, 4.4520E+03, 4.4545E+03, 4.4570E+03,
     +  4.4595E+03, 4.4619E+03, 4.4644E+03, 4.4669E+03, 4.4694E+03,
     + 4.4719E+03, 4.4744E+03, 4.4768E+03, 4.4793E+03, 4.4818E+03,
     +  4.4843E+03, 4.4868E+03, 4.4893E+03, 4.4918E+03, 4.4943E+03,
     + 4.4968E+03, 4.4992E+03, 4.5017E+03, 4.5042E+03, 4.5067E+03,
     +  4.5092E+03, 4.5117E+03, 4.5142E+03, 4.5167E+03, 4.5192E+03,
     + 4.5217E+03, 4.5242E+03, 4.5267E+03, 4.5292E+03, 4.5317E+03,
     +  4.5342E+03, 4.5367E+03, 4.5393E+03, 4.5418E+03, 4.5443E+03,
     + 4.5468E+03, 4.5493E+03, 4.5518E+03, 4.5543E+03, 4.5568E+03,
     +  4.5593E+03, 4.5619E+03, 4.5644E+03, 4.5669E+03, 4.5694E+03,
     + 4.5719E+03, 4.5744E+03, 4.5770E+03, 4.5795E+03, 4.5820E+03,
     +  4.5845E+03, 4.5870E+03, 4.5896E+03, 4.5921E+03, 4.5946E+03,
     + 4.5971E+03, 4.5997E+03, 4.6022E+03, 4.6047E+03, 4.6073E+03,
     +  4.6098E+03, 4.6123E+03, 4.6149E+03, 4.6174E+03, 4.6199E+03,
     + 4.6225E+03, 4.6250E+03, 4.6275E+03, 4.6301E+03, 4.6326E+03,
     +  4.6352E+03, 4.6377E+03, 4.6402E+03, 4.6428E+03, 4.6453E+03,
     + 4.6479E+03, 4.6504E+03, 4.6530E+03, 4.6555E+03, 4.6580E+03,
     +  4.6606E+03, 4.6631E+03, 4.6657E+03, 4.6682E+03, 4.6708E+03,
     + 4.6734E+03, 4.6759E+03, 4.6785E+03, 4.6810E+03, 4.6836E+03,
     +  4.6861E+03, 4.6887E+03, 4.6912E+03, 4.6938E+03, 4.6964E+03,
     + 4.6989E+03, 4.7015E+03, 4.7041E+03, 4.7066E+03, 4.7092E+03,
     +  4.7118E+03, 4.7143E+03, 4.7169E+03, 4.7195E+03, 4.7220E+03,
     + 4.7246E+03, 4.7272E+03, 4.7297E+03, 4.7323E+03, 4.7349E+03,
     +  4.7375E+03, 4.7400E+03, 4.7426E+03, 4.7452E+03, 4.7478E+03,
     + 4.7504E+03, 4.7529E+03, 4.7555E+03, 4.7581E+03, 4.7607E+03,
     +  4.7633E+03, 4.7658E+03, 4.7684E+03, 4.7710E+03, 4.7736E+03,
     + 4.7762E+03, 4.7788E+03, 4.7814E+03, 4.7840E+03, 4.7866E+03,
     +  4.7892E+03, 4.7917E+03, 4.7943E+03, 4.7969E+03, 4.7995E+03,
     + 4.8021E+03, 4.8047E+03, 4.8073E+03, 4.8099E+03, 4.8125E+03,
     +  4.8151E+03, 4.8177E+03, 4.8203E+03, 4.8229E+03, 4.8255E+03,
     + 4.8281E+03, 4.8308E+03, 4.8334E+03, 4.8360E+03, 4.8386E+03,
     +  4.8412E+03, 4.8438E+03, 4.8464E+03, 4.8490E+03, 4.8516E+03,
     + 4.8543E+03, 4.8569E+03, 4.8595E+03, 4.8621E+03, 4.8647E+03,
     +  4.8673E+03, 4.8700E+03, 4.8726E+03, 4.8752E+03, 4.8778E+03,
     + 4.8805E+03, 4.8831E+03, 4.8857E+03, 4.8883E+03, 4.8910E+03,
     +  4.8936E+03, 4.8962E+03, 4.8988E+03, 4.9015E+03, 4.9041E+03,
     + 4.9067E+03, 4.9094E+03, 4.9120E+03, 4.9146E+03, 4.9173E+03,
     +  4.9199E+03, 4.9225E+03, 4.9252E+03, 4.9278E+03, 4.9305E+03,
     + 4.9331E+03, 4.9358E+03, 4.9384E+03, 4.9410E+03, 4.9437E+03,
     +  4.9463E+03, 4.9490E+03, 4.9516E+03, 4.9543E+03, 4.9569E+03,
     + 4.9596E+03, 4.9622E+03, 4.9649E+03, 4.9675E+03, 4.9702E+03,
     +  4.9728E+03, 4.9755E+03, 4.9781E+03, 4.9808E+03, 4.9835E+03,
     + 4.9861E+03, 4.9888E+03, 4.9914E+03, 4.9941E+03, 4.9968E+03,
     +  4.9994E+03, 5.0021E+03, 5.0048E+03, 5.0074E+03, 5.0101E+03,
     + 5.0128E+03, 5.0154E+03, 5.0181E+03, 5.0208E+03, 5.0234E+03,
     +  5.0261E+03, 5.0288E+03, 5.0315E+03, 5.0341E+03, 5.0368E+03,
     + 5.0395E+03, 5.0422E+03, 5.0448E+03, 5.0475E+03, 5.0502E+03,
     +  5.0529E+03, 5.0556E+03, 5.0583E+03, 5.0609E+03, 5.0636E+03,
     + 5.0663E+03, 5.0690E+03, 5.0717E+03, 5.0744E+03, 5.0771E+03,
     +  5.0798E+03, 5.0824E+03, 5.0851E+03, 5.0878E+03, 5.0905E+03,
     + 5.0932E+03, 5.0959E+03, 5.0986E+03, 5.1013E+03, 5.1040E+03,
     +  5.1067E+03, 5.1094E+03, 5.1121E+03, 5.1148E+03, 5.1175E+03,
     + 5.1202E+03, 5.1229E+03, 5.1256E+03, 5.1283E+03, 5.1310E+03,
     +  5.1337E+03, 5.1365E+03, 5.1392E+03, 5.1419E+03, 5.1446E+03,
     + 5.1473E+03, 5.1500E+03, 5.1527E+03, 5.1554E+03, 5.1582E+03,
     +  5.1609E+03, 5.1636E+03, 5.1663E+03, 5.1690E+03, 5.1718E+03,
     + 5.1745E+03, 5.1772E+03, 5.1799E+03, 5.1826E+03, 5.1854E+03,
     +  5.1881E+03, 5.1908E+03, 5.1935E+03, 5.1963E+03, 5.1990E+03,
     + 5.2017E+03, 5.2045E+03, 5.2072E+03, 5.2099E+03, 5.2127E+03,
     +  5.2154E+03, 5.2181E+03, 5.2209E+03, 5.2236E+03, 5.2264E+03,
     + 5.2291E+03, 5.2318E+03, 5.2346E+03, 5.2373E+03, 5.2401E+03,
     +  5.2428E+03, 5.2455E+03, 5.2483E+03, 5.2510E+03, 5.2538E+03,
     + 5.2565E+03, 5.2593E+03, 5.2620E+03, 5.2648E+03, 5.2675E+03,
     +  5.2703E+03, 5.2730E+03, 5.2758E+03, 5.2786E+03, 5.2813E+03,
     + 5.2841E+03, 5.2868E+03, 5.2896E+03, 5.2923E+03, 5.2951E+03,
     +  5.2979E+03, 5.3006E+03, 5.3034E+03, 5.3062E+03, 5.3089E+03,
     + 5.3117E+03, 5.3145E+03, 5.3172E+03, 5.3200E+03, 5.3228E+03,
     +  5.3255E+03, 5.3283E+03, 5.3311E+03, 5.3339E+03, 5.3366E+03,
     + 5.3394E+03, 5.3422E+03, 5.3450E+03, 5.3477E+03, 5.3505E+03,
     +  5.3533E+03, 5.3561E+03, 5.3589E+03, 5.3616E+03, 5.3644E+03,
     + 5.3672E+03, 5.3700E+03, 5.3728E+03, 5.3756E+03, 5.3784E+03,
     +  5.3811E+03, 5.3839E+03, 5.3867E+03, 5.3895E+03, 5.3923E+03,
     + 5.3951E+03, 5.3979E+03, 5.4007E+03, 5.4035E+03, 5.4063E+03,
     +  5.4091E+03, 5.4119E+03, 5.4147E+03, 5.4175E+03, 5.4203E+03,
     + 5.4231E+03, 5.4259E+03, 5.4287E+03, 5.4315E+03, 5.4343E+03,
     +  5.4371E+03, 5.4399E+03, 5.4427E+03, 5.4455E+03, 5.4484E+03,
     + 5.4512E+03, 5.4540E+03, 5.4568E+03, 5.4596E+03, 5.4624E+03,
     +  5.4652E+03, 5.4681E+03, 5.4709E+03, 5.4737E+03, 5.4765E+03,
     + 5.4793E+03, 5.4821E+03, 5.4850E+03, 5.4878E+03, 5.4906E+03,
     +  5.4934E+03, 5.4963E+03, 5.4991E+03, 5.5019E+03, 5.5048E+03,
     + 5.5076E+03, 5.5104E+03, 5.5132E+03, 5.5161E+03, 5.5189E+03,
     +  5.5217E+03, 5.5246E+03, 5.5274E+03, 5.5303E+03, 5.5331E+03,
     + 5.5359E+03, 5.5388E+03, 5.5416E+03, 5.5444E+03, 5.5473E+03,
     +  5.5501E+03, 5.5530E+03, 5.5558E+03, 5.5587E+03, 5.5615E+03,
     + 5.5644E+03, 5.5672E+03, 5.5701E+03, 5.5729E+03, 5.5758E+03,
     +  5.5786E+03, 5.5815E+03, 5.5843E+03, 5.5872E+03, 5.5900E+03,
     + 5.5929E+03, 5.5957E+03, 5.5986E+03, 5.6015E+03, 5.6043E+03,
     +  5.6072E+03, 5.6100E+03, 5.6129E+03, 5.6158E+03, 5.6186E+03,
     + 5.6215E+03, 5.6244E+03, 5.6272E+03, 5.6301E+03, 5.6330E+03,
     +  5.6358E+03, 5.6387E+03, 5.6416E+03, 5.6445E+03, 5.6473E+03,
     + 5.6502E+03, 5.6531E+03, 5.6560E+03, 5.6588E+03, 5.6617E+03,
     +  5.6646E+03, 5.6675E+03, 5.6704E+03, 5.6732E+03, 5.6761E+03,
     + 5.6790E+03, 5.6819E+03, 5.6848E+03, 5.6877E+03, 5.6906E+03,
     +  5.6934E+03, 5.6963E+03, 5.6992E+03, 5.7021E+03, 5.7050E+03,
     + 5.7079E+03, 5.7108E+03, 5.7137E+03, 5.7166E+03, 5.7195E+03,
     +  5.7224E+03, 5.7253E+03, 5.7282E+03, 5.7311E+03, 5.7340E+03,
     + 5.7369E+03, 5.7398E+03, 5.7427E+03, 5.7456E+03, 5.7485E+03,
     +  5.7514E+03, 5.7543E+03, 5.7572E+03, 5.7601E+03, 5.7630E+03,
     + 5.7659E+03, 5.7689E+03, 5.7718E+03, 5.7747E+03, 5.7776E+03,
     +  5.7805E+03, 5.7834E+03, 5.7864E+03, 5.7893E+03, 5.7922E+03,
     + 5.7951E+03, 5.7980E+03, 5.8010E+03, 5.8039E+03, 5.8068E+03,
     +  5.8097E+03, 5.8126E+03, 5.8156E+03, 5.8185E+03, 5.8214E+03,
     + 5.8244E+03, 5.8273E+03, 5.8302E+03, 5.8332E+03, 5.8361E+03,
     +  5.8390E+03, 5.8420E+03, 5.8449E+03, 5.8478E+03, 5.8508E+03,
     + 5.8537E+03, 5.8566E+03, 5.8596E+03, 5.8625E+03, 5.8655E+03,
     +  5.8684E+03, 5.8714E+03, 5.8743E+03, 5.8772E+03, 5.8802E+03,
     + 5.8831E+03, 5.8861E+03, 5.8890E+03, 5.8920E+03, 5.8949E+03,
     +  5.8979E+03, 5.9008E+03, 5.9038E+03, 5.9068E+03, 5.9097E+03,
     + 5.9127E+03, 5.9156E+03, 5.9186E+03, 5.9215E+03, 5.9245E+03,
     +  5.9275E+03, 5.9304E+03, 5.9334E+03, 5.9363E+03, 5.9393E+03,
     + 5.9423E+03, 5.9452E+03, 5.9482E+03, 5.9512E+03, 5.9542E+03,
     +  5.9571E+03, 5.9601E+03, 5.9631E+03, 5.9660E+03, 5.9690E+03,
     + 5.9720E+03, 5.9750E+03, 5.9779E+03, 5.9809E+03, 5.9839E+03,
     +  5.9869E+03, 5.9899E+03, 5.9928E+03, 5.9958E+03, 5.9988E+03,
     + 6.0018E+03, 6.0048E+03, 6.0078E+03, 6.0107E+03, 6.0137E+03,
     +  6.0167E+03, 6.0197E+03, 6.0227E+03, 6.0257E+03, 6.0287E+03,
     + 6.0317E+03, 6.0347E+03, 6.0377E+03, 6.0407E+03, 6.0437E+03,
     +  6.0467E+03, 6.0497E+03, 6.0527E+03, 6.0557E+03, 6.0587E+03,
     + 6.0617E+03, 6.0647E+03, 6.0677E+03, 6.0707E+03, 6.0737E+03,
     +  6.0767E+03, 6.0797E+03, 6.0827E+03, 6.0857E+03, 6.0887E+03,
     + 6.0917E+03, 6.0947E+03, 6.0978E+03, 6.1008E+03, 6.1038E+03,
     +  6.1068E+03, 6.1098E+03, 6.1128E+03, 6.1159E+03, 6.1189E+03,
     + 6.1219E+03, 6.1249E+03, 6.1279E+03, 6.1310E+03, 6.1340E+03,
     +  6.1370E+03, 6.1400E+03, 6.1431E+03, 6.1461E+03, 6.1491E+03,
     + 6.1521E+03, 6.1552E+03, 6.1582E+03, 6.1612E+03, 6.1643E+03,
     +  6.1673E+03, 6.1703E+03, 6.1734E+03, 6.1764E+03, 6.1795E+03,
     + 6.1825E+03, 6.1855E+03, 6.1886E+03, 6.1916E+03, 6.1947E+03,
     +  6.1977E+03, 6.2007E+03, 6.2038E+03, 6.2068E+03, 6.2099E+03,
     + 6.2129E+03, 6.2160E+03, 6.2190E+03, 6.2221E+03, 6.2251E+03,
     +  6.2282E+03, 6.2312E+03, 6.2343E+03, 6.2373E+03, 6.2404E+03,
     + 6.2435E+03, 6.2465E+03, 6.2496E+03, 6.2526E+03, 6.2557E+03,
     +  6.2588E+03, 6.2618E+03, 6.2649E+03, 6.2679E+03, 6.2710E+03,
     + 6.2741E+03, 6.2771E+03, 6.2802E+03, 6.2833E+03, 6.2864E+03,
     +  6.2894E+03, 6.2925E+03, 6.2956E+03, 6.2986E+03, 6.3017E+03,
     + 6.3048E+03, 6.3079E+03, 6.3109E+03, 6.3140E+03, 6.3171E+03,
     +  6.3202E+03, 6.3233E+03, 6.3263E+03, 6.3294E+03, 6.3325E+03,
     + 6.3356E+03, 6.3387E+03, 6.3418E+03, 6.3449E+03, 6.3479E+03,
     +  6.3510E+03, 6.3541E+03, 6.3572E+03, 6.3603E+03, 6.3634E+03,
     + 6.3665E+03, 6.3696E+03, 6.3727E+03, 6.3758E+03, 6.3789E+03,
     +  6.3820E+03, 6.3851E+03, 6.3882E+03, 6.3913E+03, 6.3944E+03,
     + 6.3975E+03, 6.4006E+03, 6.4037E+03, 6.4068E+03, 6.4099E+03,
     +  6.4130E+03, 6.4161E+03, 6.4192E+03, 6.4223E+03, 6.4254E+03,
     + 6.4286E+03, 6.4317E+03, 6.4348E+03, 6.4379E+03, 6.4410E+03,
     +  6.4441E+03, 6.4472E+03, 6.4504E+03, 6.4535E+03, 6.4566E+03,
     + 6.4597E+03, 6.4628E+03, 6.4660E+03, 6.4691E+03, 6.4722E+03,
     +  6.4753E+03, 6.4785E+03, 6.4816E+03, 6.4847E+03, 6.4879E+03,
     + 6.4910E+03, 6.4941E+03, 6.4973E+03, 6.5004E+03, 6.5035E+03,
     +  6.5067E+03, 6.5098E+03, 6.5129E+03, 6.5161E+03, 6.5192E+03,
     + 6.5223E+03, 6.5255E+03, 6.5286E+03, 6.5318E+03, 6.5349E+03,
     +  6.5381E+03, 6.5412E+03, 6.5443E+03, 6.5475E+03, 6.5506E+03,
     + 6.5538E+03, 6.5569E+03, 6.5601E+03, 6.5632E+03, 6.5664E+03,
     +  6.5695E+03, 6.5727E+03, 6.5759E+03, 6.5790E+03, 6.5822E+03,
     + 6.5853E+03, 6.5885E+03, 6.5916E+03, 6.5948E+03, 6.5980E+03,
     +  6.6011E+03, 6.6043E+03, 6.6075E+03, 6.6106E+03, 6.6138E+03,
     + 6.6170E+03, 6.6201E+03, 6.6233E+03, 6.6265E+03, 6.6296E+03,
     +  6.6328E+03, 6.6360E+03, 6.6391E+03, 6.6423E+03, 6.6455E+03,
     + 6.6487E+03, 6.6518E+03, 6.6550E+03, 6.6582E+03, 6.6614E+03,
     +  6.6646E+03, 6.6677E+03, 6.6709E+03, 6.6741E+03, 6.6773E+03,
     + 6.6805E+03, 6.6837E+03, 6.6869E+03, 6.6900E+03, 6.6932E+03,
     +  6.6964E+03, 6.6996E+03, 6.7028E+03, 6.7060E+03, 6.7092E+03,
     + 6.7124E+03, 6.7156E+03, 6.7188E+03, 6.7220E+03, 6.7252E+03,
     +  6.7284E+03, 6.7316E+03, 6.7348E+03, 6.7380E+03, 6.7412E+03,
     + 6.7444E+03, 6.7476E+03, 6.7508E+03, 6.7540E+03, 6.7572E+03,
     +  6.7604E+03, 6.7636E+03, 6.7668E+03, 6.7700E+03, 6.7733E+03,
     + 6.7765E+03, 6.7797E+03, 6.7829E+03, 6.7861E+03, 6.7893E+03,
     +  6.7925E+03, 6.7958E+03, 6.7990E+03, 6.8022E+03, 6.8054E+03,
     + 6.8086E+03, 6.8119E+03, 6.8151E+03, 6.8183E+03, 6.8215E+03,
     +  6.8248E+03, 6.8280E+03, 6.8312E+03, 6.8345E+03, 6.8377E+03,
     + 6.8409E+03, 6.8442E+03, 6.8474E+03, 6.8506E+03, 6.8539E+03,
     +  6.8571E+03, 6.8603E+03, 6.8636E+03, 6.8668E+03, 6.8700E+03,
     + 6.8733E+03, 6.8765E+03, 6.8798E+03, 6.8830E+03, 6.8863E+03,
     +  6.8895E+03, 6.8927E+03, 6.8960E+03, 6.8992E+03, 6.9025E+03,
     + 6.9057E+03, 6.9090E+03, 6.9122E+03, 6.9155E+03, 6.9187E+03,
     +  6.9220E+03, 6.9253E+03, 6.9285E+03, 6.9318E+03, 6.9350E+03,
     + 6.9383E+03, 6.9415E+03, 6.9448E+03, 6.9481E+03, 6.9513E+03,
     +  6.9546E+03, 6.9579E+03, 6.9611E+03, 6.9644E+03, 6.9676E+03,
     + 6.9709E+03, 6.9742E+03, 6.9775E+03, 6.9807E+03, 6.9840E+03,
     +  6.9873E+03, 6.9905E+03, 6.9938E+03, 6.9971E+03, 7.0004E+03,
     + 7.0036E+03, 7.0069E+03, 7.0102E+03, 7.0135E+03, 7.0168E+03,
     +  7.0200E+03, 7.0233E+03, 7.0266E+03, 7.0299E+03, 7.0332E+03,
     + 7.0365E+03, 7.0398E+03, 7.0430E+03, 7.0463E+03, 7.0496E+03,
     +  7.0529E+03, 7.0562E+03, 7.0595E+03, 7.0628E+03, 7.0661E+03,
     + 7.0694E+03, 7.0727E+03, 7.0760E+03, 7.0793E+03, 7.0826E+03,
     +  7.0859E+03, 7.0892E+03, 7.0925E+03, 7.0958E+03, 7.0991E+03,
     + 7.1024E+03, 7.1057E+03, 7.1090E+03, 7.1123E+03, 7.1156E+03,
     +  7.1189E+03, 7.1222E+03, 7.1255E+03, 7.1289E+03, 7.1322E+03,
     + 7.1355E+03, 7.1388E+03, 7.1421E+03, 7.1454E+03, 7.1487E+03,
     +  7.1521E+03, 7.1554E+03, 7.1587E+03, 7.1620E+03, 7.1653E+03,
     + 7.1687E+03, 7.1720E+03, 7.1753E+03, 7.1786E+03, 7.1820E+03,
     +  7.1853E+03, 7.1886E+03, 7.1920E+03, 7.1953E+03, 7.1986E+03,
     + 7.2019E+03, 7.2053E+03, 7.2086E+03, 7.2119E+03, 7.2153E+03,
     +  7.2186E+03, 7.2220E+03, 7.2253E+03, 7.2286E+03, 7.2320E+03,
     + 7.2353E+03, 7.2387E+03, 7.2420E+03, 7.2453E+03, 7.2487E+03,
     +  7.2520E+03, 7.2554E+03, 7.2587E+03, 7.2621E+03, 7.2654E+03,
     + 7.2688E+03, 7.2721E+03, 7.2755E+03, 7.2788E+03, 7.2822E+03,
     +  7.2855E+03, 7.2889E+03, 7.2923E+03, 7.2956E+03, 7.2990E+03,
     + 7.3023E+03, 7.3057E+03, 7.3091E+03, 7.3124E+03, 7.3158E+03,
     +  7.3191E+03, 7.3225E+03, 7.3259E+03, 7.3292E+03, 7.3326E+03,
     + 7.3360E+03, 7.3393E+03, 7.3427E+03, 7.3461E+03, 7.3495E+03,
     +  7.3528E+03, 7.3562E+03, 7.3596E+03, 7.3630E+03, 7.3663E+03,
     + 7.3697E+03, 7.3731E+03, 7.3765E+03, 7.3799E+03, 7.3832E+03,
     +  7.3866E+03, 7.3900E+03, 7.3934E+03, 7.3968E+03, 7.4002E+03,
     + 7.4035E+03, 7.4069E+03, 7.4103E+03, 7.4137E+03, 7.4171E+03,
     +  7.4205E+03, 7.4239E+03, 7.4273E+03, 7.4307E+03, 7.4341E+03,
     + 7.4375E+03, 7.4409E+03, 7.4443E+03, 7.4477E+03, 7.4511E+03,
     +  7.4545E+03, 7.4579E+03, 7.4613E+03, 7.4647E+03, 7.4681E+03,
     + 7.4715E+03, 7.4749E+03, 7.4783E+03, 7.4817E+03, 7.4851E+03,
     +  7.4885E+03, 7.4919E+03, 7.4953E+03, 7.4987E+03, 7.5022E+03,
     + 7.5056E+03, 7.5090E+03, 7.5124E+03, 7.5158E+03, 7.5192E+03,
     +  7.5227E+03, 7.5261E+03, 7.5295E+03, 7.5329E+03, 7.5363E+03,
     + 7.5398E+03, 7.5432E+03, 7.5466E+03, 7.5500E+03, 7.5535E+03,
     +  7.5569E+03, 7.5603E+03, 7.5638E+03, 7.5672E+03, 7.5706E+03,
     + 7.5741E+03, 7.5775E+03, 7.5809E+03, 7.5844E+03, 7.5878E+03,
     +  7.5912E+03, 7.5947E+03, 7.5981E+03, 7.6015E+03, 7.6050E+03,
     + 7.6084E+03, 7.6119E+03, 7.6153E+03, 7.6188E+03, 7.6222E+03,
     +  7.6257E+03, 7.6291E+03, 7.6325E+03, 7.6360E+03, 7.6394E+03,
     + 7.6429E+03, 7.6464E+03, 7.6498E+03, 7.6533E+03/
c...        --        81
      data (QofT( 2,J),J=1,2408)/ 1.5974E+01, 1.5987E+01, 1.5991E+01,
     +  1.5993E+01, 1.5995E+01, 1.5996E+01, 1.5996E+01, 1.5997E+01,
     + 1.5997E+01, 1.5998E+01, 1.5998E+01, 1.5999E+01, 1.6000E+01,
     +  1.6003E+01, 1.6006E+01, 1.6012E+01, 1.6020E+01, 1.6030E+01,
     + 1.6043E+01, 1.6060E+01, 1.6105E+01, 1.6167E+01, 1.6247E+01,
     +  1.6346E+01, 1.6465E+01, 1.6602E+01, 1.6758E+01, 1.6932E+01,
     + 1.7123E+01, 1.7330E+01, 1.7552E+01, 1.7789E+01, 1.8040E+01,
     +  1.8304E+01, 1.8580E+01, 1.8867E+01, 1.9166E+01, 1.9476E+01,
     + 1.9795E+01, 2.0124E+01, 2.0462E+01, 2.0808E+01, 2.1163E+01,
     +  2.1525E+01, 2.1895E+01, 2.2272E+01, 2.2656E+01, 2.3046E+01,
     + 2.3443E+01, 2.3845E+01, 2.4253E+01, 2.4667E+01, 2.5086E+01,
     +  2.5510E+01, 2.5939E+01, 2.6372E+01, 2.6810E+01, 2.7252E+01,
     + 2.7699E+01, 2.8149E+01, 2.8604E+01, 2.9062E+01, 2.9523E+01,
     +  2.9988E+01, 3.0457E+01, 3.0929E+01, 3.1403E+01, 3.1881E+01,
     + 3.2362E+01, 3.2846E+01, 3.3332E+01, 3.3821E+01, 3.4312E+01,
     +  3.4806E+01, 3.5302E+01, 3.5801E+01, 3.6302E+01, 3.6805E+01,
     + 3.7310E+01, 3.7818E+01, 3.8327E+01, 3.8838E+01, 3.9351E+01,
     +  3.9866E+01, 4.0383E+01, 4.0901E+01, 4.1421E+01, 4.1943E+01,
     + 4.2466E+01, 4.2991E+01, 4.3517E+01, 4.4045E+01, 4.4574E+01,
     +  4.5105E+01, 4.5637E+01, 4.6170E+01, 4.6705E+01, 4.7240E+01,
     + 4.7777E+01, 4.8315E+01, 4.8855E+01, 4.9395E+01, 4.9937E+01,
     +  5.0480E+01, 5.1023E+01, 5.1568E+01, 5.2114E+01, 5.2660E+01,
     + 5.3208E+01, 5.3757E+01, 5.4306E+01, 5.4857E+01, 5.5408E+01,
     +  5.5960E+01, 5.6513E+01, 5.7067E+01, 5.7621E+01, 5.8177E+01,
     + 5.8733E+01, 5.9290E+01, 5.9847E+01, 6.0406E+01, 6.0965E+01,
     +  6.1524E+01, 6.2085E+01, 6.2646E+01, 6.3207E+01, 6.3770E+01,
     + 6.4333E+01, 6.4896E+01, 6.5460E+01, 6.6025E+01, 6.6590E+01,
     +  6.7156E+01, 6.7723E+01, 6.8290E+01, 6.8857E+01, 6.9425E+01,
     + 6.9994E+01, 7.0563E+01, 7.1132E+01, 7.1702E+01, 7.2273E+01,
     +  7.2844E+01, 7.3415E+01, 7.3987E+01, 7.4559E+01, 7.5132E+01,
     + 7.5705E+01, 7.6279E+01, 7.6853E+01, 7.7427E+01, 7.8002E+01,
     +  7.8577E+01, 7.9153E+01, 7.9729E+01, 8.0305E+01, 8.0882E+01,
     + 8.1459E+01, 8.2037E+01, 8.2614E+01, 8.3193E+01, 8.3771E+01,
     +  8.4350E+01, 8.4929E+01, 8.5509E+01, 8.6089E+01, 8.6669E+01,
     + 8.7250E+01, 8.7830E+01, 8.8411E+01, 8.8993E+01, 8.9574E+01,
     +  9.0156E+01, 9.0738E+01, 9.1321E+01, 9.1904E+01, 9.2487E+01,
     + 9.3070E+01, 9.3653E+01, 9.4238E+01, 9.4822E+01, 9.5406E+01,
     +  9.5991E+01, 9.6576E+01, 9.7161E+01, 9.7746E+01, 9.8332E+01,
     + 9.8918E+01, 9.9504E+01, 1.0009E+02, 1.0068E+02, 1.0126E+02,
     +  1.0185E+02, 1.0244E+02, 1.0303E+02, 1.0361E+02, 1.0420E+02,
     + 1.0479E+02, 1.0538E+02, 1.0597E+02, 1.0656E+02, 1.0714E+02,
     +  1.0773E+02, 1.0832E+02, 1.0891E+02, 1.0950E+02, 1.1009E+02,
     + 1.1068E+02, 1.1127E+02, 1.1186E+02, 1.1246E+02, 1.1305E+02,
     +  1.1364E+02, 1.1423E+02, 1.1482E+02, 1.1541E+02, 1.1600E+02,
     + 1.1660E+02, 1.1719E+02, 1.1778E+02, 1.1837E+02, 1.1897E+02,
     +  1.1956E+02, 1.2015E+02, 1.2075E+02, 1.2134E+02, 1.2193E+02,
     + 1.2253E+02, 1.2312E+02, 1.2372E+02, 1.2431E+02, 1.2491E+02,
     +  1.2550E+02, 1.2609E+02, 1.2669E+02, 1.2729E+02, 1.2788E+02,
     + 1.2848E+02, 1.2907E+02, 1.2967E+02, 1.3027E+02, 1.3086E+02,
     +  1.3146E+02, 1.3205E+02, 1.3265E+02, 1.3325E+02, 1.3385E+02,
     + 1.3444E+02, 1.3504E+02, 1.3564E+02, 1.3624E+02, 1.3683E+02,
     +  1.3743E+02, 1.3803E+02, 1.3863E+02, 1.3923E+02, 1.3983E+02,
     + 1.4043E+02, 1.4103E+02, 1.4163E+02, 1.4223E+02, 1.4282E+02,
     +  1.4343E+02, 1.4403E+02, 1.4462E+02, 1.4523E+02, 1.4583E+02,
     + 1.4643E+02, 1.4703E+02, 1.4763E+02, 1.4823E+02, 1.4883E+02,
     +  1.4944E+02, 1.5004E+02, 1.5064E+02, 1.5124E+02, 1.5184E+02,
     + 1.5245E+02, 1.5305E+02, 1.5365E+02, 1.5426E+02, 1.5486E+02,
     +  1.5546E+02, 1.5607E+02, 1.5667E+02, 1.5728E+02, 1.5788E+02,
     + 1.5848E+02, 1.5909E+02, 1.5969E+02, 1.6030E+02, 1.6090E+02,
     +  1.6151E+02, 1.6212E+02, 1.6272E+02, 1.6333E+02, 1.6393E+02,
     + 1.6454E+02, 1.6515E+02, 1.6575E+02, 1.6636E+02, 1.6697E+02,
     +  1.6758E+02, 1.6818E+02, 1.6879E+02, 1.6940E+02, 1.7001E+02,
     + 1.7062E+02, 1.7123E+02, 1.7183E+02, 1.7244E+02, 1.7305E+02,
     +  1.7366E+02, 1.7427E+02, 1.7488E+02, 1.7549E+02, 1.7610E+02,
     + 1.7671E+02, 1.7732E+02, 1.7794E+02, 1.7855E+02, 1.7916E+02,
     +  1.7977E+02, 1.8038E+02, 1.8100E+02, 1.8161E+02, 1.8222E+02,
     + 1.8283E+02, 1.8345E+02, 1.8406E+02, 1.8467E+02, 1.8529E+02,
     +  1.8590E+02, 1.8652E+02, 1.8713E+02, 1.8775E+02, 1.8836E+02,
     + 1.8898E+02, 1.8959E+02, 1.9021E+02, 1.9083E+02, 1.9144E+02,
     +  1.9206E+02, 1.9268E+02, 1.9329E+02, 1.9391E+02, 1.9453E+02,
     + 1.9515E+02, 1.9576E+02, 1.9638E+02, 1.9700E+02, 1.9762E+02,
     +  1.9824E+02, 1.9886E+02, 1.9948E+02, 2.0010E+02, 2.0072E+02,
     + 2.0134E+02, 2.0196E+02, 2.0258E+02, 2.0321E+02, 2.0383E+02,
     +  2.0445E+02, 2.0507E+02, 2.0569E+02, 2.0632E+02, 2.0694E+02,
     + 2.0756E+02, 2.0819E+02, 2.0881E+02, 2.0943E+02, 2.1006E+02,
     +  2.1069E+02, 2.1131E+02, 2.1194E+02, 2.1256E+02, 2.1319E+02,
     + 2.1382E+02, 2.1444E+02, 2.1507E+02, 2.1570E+02, 2.1632E+02,
     +  2.1695E+02, 2.1758E+02, 2.1821E+02, 2.1884E+02, 2.1947E+02,
     + 2.2010E+02, 2.2073E+02, 2.2136E+02, 2.2199E+02, 2.2262E+02,
     +  2.2325E+02, 2.2388E+02, 2.2451E+02, 2.2514E+02, 2.2578E+02,
     + 2.2641E+02, 2.2704E+02, 2.2768E+02, 2.2831E+02, 2.2895E+02,
     +  2.2958E+02, 2.3021E+02, 2.3085E+02, 2.3148E+02, 2.3212E+02,
     + 2.3276E+02, 2.3339E+02, 2.3403E+02, 2.3467E+02, 2.3530E+02,
     +  2.3594E+02, 2.3658E+02, 2.3722E+02, 2.3786E+02, 2.3850E+02,
     + 2.3914E+02, 2.3978E+02, 2.4042E+02, 2.4106E+02, 2.4170E+02,
     +  2.4234E+02, 2.4298E+02, 2.4362E+02, 2.4427E+02, 2.4491E+02,
     + 2.4555E+02, 2.4620E+02, 2.4684E+02, 2.4748E+02, 2.4813E+02,
     +  2.4878E+02, 2.4942E+02, 2.5007E+02, 2.5071E+02, 2.5136E+02,
     + 2.5201E+02, 2.5266E+02, 2.5330E+02, 2.5395E+02, 2.5460E+02,
     +  2.5525E+02, 2.5590E+02, 2.5655E+02, 2.5720E+02, 2.5785E+02,
     + 2.5850E+02, 2.5915E+02, 2.5980E+02, 2.6045E+02, 2.6111E+02,
     +  2.6176E+02, 2.6241E+02, 2.6307E+02, 2.6372E+02, 2.6437E+02,
     + 2.6503E+02, 2.6569E+02, 2.6634E+02, 2.6700E+02, 2.6765E+02,
     +  2.6831E+02, 2.6897E+02, 2.6963E+02, 2.7028E+02, 2.7094E+02,
     + 2.7160E+02, 2.7226E+02, 2.7292E+02, 2.7358E+02, 2.7424E+02,
     +  2.7490E+02, 2.7556E+02, 2.7623E+02, 2.7689E+02, 2.7755E+02,
     + 2.7821E+02, 2.7888E+02, 2.7954E+02, 2.8021E+02, 2.8087E+02,
     +  2.8154E+02, 2.8220E+02, 2.8287E+02, 2.8354E+02, 2.8420E+02,
     + 2.8487E+02, 2.8554E+02, 2.8621E+02, 2.8687E+02, 2.8754E+02,
     +  2.8821E+02, 2.8889E+02, 2.8956E+02, 2.9023E+02, 2.9090E+02,
     + 2.9157E+02, 2.9224E+02, 2.9291E+02, 2.9359E+02, 2.9426E+02,
     +  2.9494E+02, 2.9561E+02, 2.9629E+02, 2.9696E+02, 2.9764E+02,
     + 2.9831E+02, 2.9899E+02, 2.9967E+02, 3.0035E+02, 3.0103E+02,
     +  3.0170E+02, 3.0238E+02, 3.0306E+02, 3.0374E+02, 3.0443E+02,
     + 3.0511E+02, 3.0579E+02, 3.0647E+02, 3.0715E+02, 3.0783E+02,
     +  3.0852E+02, 3.0920E+02, 3.0989E+02, 3.1057E+02, 3.1126E+02,
     + 3.1194E+02, 3.1263E+02, 3.1331E+02, 3.1400E+02, 3.1469E+02,
     +  3.1538E+02, 3.1607E+02, 3.1676E+02, 3.1745E+02, 3.1814E+02,
     + 3.1883E+02, 3.1952E+02, 3.2021E+02, 3.2090E+02, 3.2160E+02,
     +  3.2229E+02, 3.2298E+02, 3.2368E+02, 3.2437E+02, 3.2507E+02,
     + 3.2576E+02, 3.2646E+02, 3.2715E+02, 3.2785E+02, 3.2855E+02,
     +  3.2925E+02, 3.2994E+02, 3.3064E+02, 3.3135E+02, 3.3204E+02,
     + 3.3274E+02, 3.3344E+02, 3.3415E+02, 3.3485E+02, 3.3555E+02,
     +  3.3625E+02, 3.3696E+02, 3.3766E+02, 3.3837E+02, 3.3907E+02,
     + 3.3978E+02, 3.4048E+02, 3.4119E+02, 3.4190E+02, 3.4261E+02,
     +  3.4331E+02, 3.4402E+02, 3.4473E+02, 3.4544E+02, 3.4615E+02,
     + 3.4686E+02, 3.4757E+02, 3.4828E+02, 3.4900E+02, 3.4971E+02,
     +  3.5042E+02, 3.5114E+02, 3.5185E+02, 3.5257E+02, 3.5328E+02,
     + 3.5400E+02, 3.5471E+02, 3.5543E+02, 3.5615E+02, 3.5686E+02,
     +  3.5759E+02, 3.5830E+02, 3.5902E+02, 3.5974E+02, 3.6046E+02,
     + 3.6118E+02, 3.6191E+02, 3.6263E+02, 3.6335E+02, 3.6408E+02,
     +  3.6480E+02, 3.6552E+02, 3.6625E+02, 3.6697E+02, 3.6770E+02,
     + 3.6843E+02, 3.6915E+02, 3.6988E+02, 3.7061E+02, 3.7134E+02,
     +  3.7207E+02, 3.7279E+02, 3.7352E+02, 3.7426E+02, 3.7499E+02,
     + 3.7572E+02, 3.7645E+02, 3.7718E+02, 3.7792E+02, 3.7865E+02,
     +  3.7939E+02, 3.8012E+02, 3.8086E+02, 3.8159E+02, 3.8233E+02,
     + 3.8307E+02, 3.8381E+02, 3.8455E+02, 3.8528E+02, 3.8602E+02,
     +  3.8676E+02, 3.8750E+02, 3.8825E+02, 3.8899E+02, 3.8973E+02,
     + 3.9047E+02, 3.9121E+02, 3.9196E+02, 3.9270E+02, 3.9345E+02,
     +  3.9419E+02, 3.9494E+02, 3.9569E+02, 3.9643E+02, 3.9718E+02,
     + 3.9793E+02, 3.9868E+02, 3.9943E+02, 4.0018E+02, 4.0093E+02,
     +  4.0168E+02, 4.0243E+02, 4.0318E+02, 4.0394E+02, 4.0469E+02,
     + 4.0545E+02, 4.0620E+02, 4.0695E+02, 4.0771E+02, 4.0847E+02,
     +  4.0922E+02, 4.0998E+02, 4.1074E+02, 4.1150E+02, 4.1225E+02,
     + 4.1301E+02, 4.1378E+02, 4.1453E+02, 4.1530E+02, 4.1606E+02,
     +  4.1682E+02, 4.1758E+02, 4.1835E+02, 4.1911E+02, 4.1988E+02,
     + 4.2064E+02, 4.2141E+02, 4.2217E+02, 4.2294E+02, 4.2371E+02,
     +  4.2448E+02, 4.2524E+02, 4.2602E+02, 4.2678E+02, 4.2756E+02,
     + 4.2833E+02, 4.2910E+02, 4.2987E+02, 4.3064E+02, 4.3142E+02,
     +  4.3219E+02, 4.3297E+02, 4.3374E+02, 4.3452E+02, 4.3529E+02,
     + 4.3607E+02, 4.3685E+02, 4.3762E+02, 4.3840E+02, 4.3918E+02,
     +  4.3996E+02, 4.4074E+02, 4.4152E+02, 4.4230E+02, 4.4309E+02,
     + 4.4387E+02, 4.4465E+02, 4.4544E+02, 4.4622E+02, 4.4701E+02,
     +  4.4779E+02, 4.4858E+02, 4.4937E+02, 4.5015E+02, 4.5094E+02,
     + 4.5173E+02, 4.5252E+02, 4.5331E+02, 4.5410E+02, 4.5489E+02,
     +  4.5568E+02, 4.5647E+02, 4.5727E+02, 4.5806E+02, 4.5885E+02,
     + 4.5965E+02, 4.6044E+02, 4.6124E+02, 4.6203E+02, 4.6283E+02,
     +  4.6363E+02, 4.6443E+02, 4.6522E+02, 4.6602E+02, 4.6682E+02,
     + 4.6762E+02, 4.6843E+02, 4.6923E+02, 4.7003E+02, 4.7083E+02,
     +  4.7163E+02, 4.7244E+02, 4.7324E+02, 4.7405E+02, 4.7485E+02,
     + 4.7566E+02, 4.7647E+02, 4.7728E+02, 4.7808E+02, 4.7889E+02,
     +  4.7970E+02, 4.8051E+02, 4.8132E+02, 4.8213E+02, 4.8294E+02,
     + 4.8376E+02, 4.8457E+02, 4.8539E+02, 4.8620E+02, 4.8702E+02,
     +  4.8783E+02, 4.8865E+02, 4.8946E+02, 4.9028E+02, 4.9110E+02,
     + 4.9191E+02, 4.9273E+02, 4.9355E+02, 4.9437E+02, 4.9519E+02,
     +  4.9602E+02, 4.9684E+02, 4.9766E+02, 4.9848E+02, 4.9931E+02,
     + 5.0013E+02, 5.0096E+02, 5.0178E+02, 5.0261E+02, 5.0344E+02,
     +  5.0426E+02, 5.0509E+02, 5.0592E+02, 5.0675E+02, 5.0758E+02,
     + 5.0841E+02, 5.0924E+02, 5.1008E+02, 5.1091E+02, 5.1174E+02,
     +  5.1257E+02, 5.1341E+02, 5.1424E+02, 5.1508E+02, 5.1591E+02,
     + 5.1675E+02, 5.1759E+02, 5.1843E+02, 5.1926E+02, 5.2011E+02,
     +  5.2094E+02, 5.2179E+02, 5.2262E+02, 5.2347E+02, 5.2431E+02,
     + 5.2516E+02, 5.2599E+02, 5.2684E+02, 5.2769E+02, 5.2853E+02,
     +  5.2938E+02, 5.3022E+02, 5.3107E+02, 5.3192E+02, 5.3277E+02,
     + 5.3362E+02, 5.3447E+02, 5.3532E+02, 5.3617E+02, 5.3702E+02,
     +  5.3787E+02, 5.3872E+02, 5.3958E+02, 5.4043E+02, 5.4128E+02,
     + 5.4214E+02, 5.4300E+02, 5.4385E+02, 5.4471E+02, 5.4557E+02,
     +  5.4643E+02, 5.4729E+02, 5.4815E+02, 5.4901E+02, 5.4987E+02,
     + 5.5073E+02, 5.5159E+02, 5.5245E+02, 5.5332E+02, 5.5418E+02,
     +  5.5505E+02, 5.5591E+02, 5.5678E+02, 5.5765E+02, 5.5851E+02,
     + 5.5938E+02, 5.6025E+02, 5.6111E+02, 5.6198E+02, 5.6286E+02,
     +  5.6373E+02, 5.6460E+02, 5.6547E+02, 5.6634E+02, 5.6721E+02,
     + 5.6809E+02, 5.6896E+02, 5.6984E+02, 5.7072E+02, 5.7159E+02,
     +  5.7247E+02, 5.7335E+02, 5.7423E+02, 5.7511E+02, 5.7598E+02,
     + 5.7687E+02, 5.7775E+02, 5.7863E+02, 5.7951E+02, 5.8039E+02,
     +  5.8128E+02, 5.8216E+02, 5.8305E+02, 5.8393E+02, 5.8482E+02,
     + 5.8570E+02, 5.8660E+02, 5.8748E+02, 5.8837E+02, 5.8926E+02,
     +  5.9015E+02, 5.9104E+02, 5.9193E+02, 5.9282E+02, 5.9371E+02,
     + 5.9461E+02, 5.9550E+02, 5.9640E+02, 5.9729E+02, 5.9819E+02,
     +  5.9908E+02, 5.9998E+02, 6.0088E+02, 6.0178E+02, 6.0268E+02,
     + 6.0357E+02, 6.0447E+02, 6.0538E+02, 6.0627E+02, 6.0718E+02,
     +  6.0808E+02, 6.0898E+02, 6.0989E+02, 6.1079E+02, 6.1170E+02,
     + 6.1260E+02, 6.1351E+02, 6.1441E+02, 6.1533E+02, 6.1623E+02,
     +  6.1714E+02, 6.1805E+02, 6.1896E+02, 6.1987E+02, 6.2078E+02,
     + 6.2170E+02, 6.2261E+02, 6.2352E+02, 6.2443E+02, 6.2535E+02,
     +  6.2627E+02, 6.2718E+02, 6.2810E+02, 6.2901E+02, 6.2993E+02,
     + 6.3085E+02, 6.3177E+02, 6.3269E+02, 6.3361E+02, 6.3454E+02,
     +  6.3545E+02, 6.3638E+02, 6.3730E+02, 6.3823E+02, 6.3915E+02,
     + 6.4007E+02, 6.4100E+02, 6.4193E+02, 6.4286E+02, 6.4378E+02,
     +  6.4471E+02, 6.4564E+02, 6.4657E+02, 6.4750E+02, 6.4843E+02,
     + 6.4936E+02, 6.5029E+02, 6.5122E+02, 6.5216E+02, 6.5310E+02,
     +  6.5403E+02, 6.5496E+02, 6.5590E+02, 6.5684E+02, 6.5778E+02,
     + 6.5871E+02, 6.5965E+02, 6.6059E+02, 6.6153E+02, 6.6247E+02,
     +  6.6341E+02, 6.6435E+02, 6.6530E+02, 6.6624E+02, 6.6718E+02,
     + 6.6812E+02, 6.6907E+02, 6.7002E+02, 6.7096E+02, 6.7191E+02,
     +  6.7286E+02, 6.7381E+02, 6.7476E+02, 6.7571E+02, 6.7665E+02,
     + 6.7761E+02, 6.7856E+02, 6.7951E+02, 6.8047E+02, 6.8142E+02,
     +  6.8237E+02, 6.8332E+02, 6.8428E+02, 6.8524E+02, 6.8619E+02,
     + 6.8715E+02, 6.8811E+02, 6.8907E+02, 6.9003E+02, 6.9099E+02,
     +  6.9195E+02, 6.9291E+02, 6.9387E+02, 6.9483E+02, 6.9580E+02,
     + 6.9676E+02, 6.9773E+02, 6.9869E+02, 6.9966E+02, 7.0063E+02,
     +  7.0159E+02, 7.0256E+02, 7.0353E+02, 7.0450E+02, 7.0547E+02,
     + 7.0644E+02, 7.0741E+02, 7.0838E+02, 7.0935E+02, 7.1033E+02,
     +  7.1130E+02, 7.1228E+02, 7.1325E+02, 7.1423E+02, 7.1520E+02,
     + 7.1618E+02, 7.1716E+02, 7.1814E+02, 7.1912E+02, 7.2010E+02,
     +  7.2108E+02, 7.2206E+02, 7.2304E+02, 7.2402E+02, 7.2500E+02,
     + 7.2599E+02, 7.2697E+02, 7.2796E+02, 7.2894E+02, 7.2993E+02,
     +  7.3092E+02, 7.3190E+02, 7.3289E+02, 7.3388E+02, 7.3487E+02,
     + 7.3586E+02, 7.3685E+02, 7.3785E+02, 7.3884E+02, 7.3983E+02,
     +  7.4082E+02, 7.4181E+02, 7.4281E+02, 7.4381E+02, 7.4480E+02,
     + 7.4580E+02, 7.4680E+02, 7.4779E+02, 7.4880E+02, 7.4979E+02,
     +  7.5080E+02, 7.5179E+02, 7.5280E+02, 7.5380E+02, 7.5480E+02,
     + 7.5580E+02, 7.5681E+02, 7.5781E+02, 7.5882E+02, 7.5983E+02,
     +  7.6083E+02, 7.6184E+02, 7.6285E+02, 7.6385E+02, 7.6487E+02,
     + 7.6588E+02, 7.6688E+02, 7.6790E+02, 7.6891E+02, 7.6992E+02,
     +  7.7094E+02, 7.7195E+02, 7.7296E+02, 7.7398E+02, 7.7499E+02,
     + 7.7601E+02, 7.7703E+02, 7.7805E+02, 7.7907E+02, 7.8009E+02,
     +  7.8111E+02, 7.8213E+02, 7.8315E+02, 7.8417E+02, 7.8519E+02,
     + 7.8621E+02, 7.8724E+02, 7.8827E+02, 7.8929E+02, 7.9031E+02,
     +  7.9134E+02, 7.9237E+02, 7.9340E+02, 7.9443E+02, 7.9545E+02,
     + 7.9649E+02, 7.9752E+02, 7.9855E+02, 7.9958E+02, 8.0061E+02,
     +  8.0164E+02, 8.0268E+02, 8.0371E+02, 8.0475E+02, 8.0578E+02,
     + 8.0682E+02, 8.0786E+02, 8.0890E+02, 8.0993E+02, 8.1098E+02,
     +  8.1201E+02, 8.1306E+02, 8.1409E+02, 8.1514E+02, 8.1618E+02,
     + 8.1722E+02, 8.1827E+02, 8.1931E+02, 8.2035E+02, 8.2140E+02,
     +  8.2245E+02, 8.2350E+02, 8.2454E+02, 8.2559E+02, 8.2664E+02,
     + 8.2769E+02, 8.2875E+02, 8.2980E+02, 8.3085E+02, 8.3190E+02,
     +  8.3295E+02, 8.3401E+02, 8.3506E+02, 8.3612E+02, 8.3717E+02,
     + 8.3823E+02, 8.3929E+02, 8.4034E+02, 8.4140E+02, 8.4246E+02,
     +  8.4352E+02, 8.4459E+02, 8.4565E+02, 8.4671E+02, 8.4777E+02,
     + 8.4883E+02, 8.4990E+02, 8.5096E+02, 8.5203E+02, 8.5309E+02,
     +  8.5416E+02, 8.5522E+02, 8.5630E+02, 8.5736E+02, 8.5843E+02,
     + 8.5951E+02, 8.6057E+02, 8.6164E+02, 8.6272E+02, 8.6379E+02,
     +  8.6486E+02, 8.6594E+02, 8.6701E+02, 8.6809E+02, 8.6916E+02,
     + 8.7024E+02, 8.7132E+02, 8.7240E+02, 8.7347E+02, 8.7456E+02,
     +  8.7563E+02, 8.7672E+02, 8.7780E+02, 8.7888E+02, 8.7996E+02,
     + 8.8104E+02, 8.8213E+02, 8.8322E+02, 8.8430E+02, 8.8539E+02,
     +  8.8647E+02, 8.8756E+02, 8.8865E+02, 8.8974E+02, 8.9083E+02,
     + 8.9192E+02, 8.9301E+02, 8.9410E+02, 8.9519E+02, 8.9629E+02,
     +  8.9738E+02, 8.9848E+02, 8.9957E+02, 9.0067E+02, 9.0176E+02,
     + 9.0285E+02, 9.0395E+02, 9.0505E+02, 9.0615E+02, 9.0725E+02,
     +  9.0835E+02, 9.0945E+02, 9.1055E+02, 9.1165E+02, 9.1276E+02,
     + 9.1387E+02, 9.1497E+02, 9.1608E+02, 9.1718E+02, 9.1829E+02,
     +  9.1939E+02, 9.2050E+02, 9.2161E+02, 9.2272E+02, 9.2383E+02,
     + 9.2494E+02, 9.2605E+02, 9.2716E+02, 9.2828E+02, 9.2939E+02,
     +  9.3051E+02, 9.3162E+02, 9.3273E+02, 9.3384E+02, 9.3497E+02,
     + 9.3608E+02, 9.3720E+02, 9.3832E+02, 9.3943E+02, 9.4056E+02,
     +  9.4168E+02, 9.4280E+02, 9.4392E+02, 9.4505E+02, 9.4617E+02,
     + 9.4729E+02, 9.4841E+02, 9.4954E+02, 9.5067E+02, 9.5180E+02,
     +  9.5292E+02, 9.5405E+02, 9.5517E+02, 9.5631E+02, 9.5744E+02,
     + 9.5857E+02, 9.5970E+02, 9.6084E+02, 9.6196E+02, 9.6310E+02,
     +  9.6424E+02, 9.6536E+02, 9.6650E+02, 9.6764E+02, 9.6878E+02,
     + 9.6991E+02, 9.7105E+02, 9.7219E+02, 9.7333E+02, 9.7447E+02,
     +  9.7561E+02, 9.7675E+02, 9.7790E+02, 9.7904E+02, 9.8018E+02,
     + 9.8133E+02, 9.8248E+02, 9.8362E+02, 9.8477E+02, 9.8591E+02,
     +  9.8706E+02, 9.8821E+02, 9.8936E+02, 9.9050E+02, 9.9166E+02,
     + 9.9281E+02, 9.9396E+02, 9.9511E+02, 9.9627E+02, 9.9742E+02,
     +  9.9858E+02, 9.9973E+02, 1.0009E+03, 1.0020E+03, 1.0032E+03,
     + 1.0044E+03, 1.0055E+03, 1.0067E+03, 1.0078E+03, 1.0090E+03,
     +  1.0102E+03, 1.0113E+03, 1.0125E+03, 1.0136E+03, 1.0148E+03,
     + 1.0160E+03, 1.0171E+03, 1.0183E+03, 1.0195E+03, 1.0206E+03,
     +  1.0218E+03, 1.0230E+03, 1.0242E+03, 1.0253E+03, 1.0265E+03,
     + 1.0277E+03, 1.0288E+03, 1.0300E+03, 1.0312E+03, 1.0324E+03,
     +  1.0335E+03, 1.0347E+03, 1.0359E+03, 1.0371E+03, 1.0382E+03,
     + 1.0394E+03, 1.0406E+03, 1.0418E+03, 1.0430E+03, 1.0441E+03,
     +  1.0453E+03, 1.0465E+03, 1.0477E+03, 1.0489E+03, 1.0501E+03,
     + 1.0512E+03, 1.0524E+03, 1.0536E+03, 1.0548E+03, 1.0560E+03,
     +  1.0572E+03, 1.0584E+03, 1.0596E+03, 1.0607E+03, 1.0619E+03,
     + 1.0631E+03, 1.0643E+03, 1.0655E+03, 1.0667E+03, 1.0679E+03,
     +  1.0691E+03, 1.0703E+03, 1.0715E+03, 1.0727E+03, 1.0739E+03,
     + 1.0751E+03, 1.0763E+03, 1.0775E+03, 1.0787E+03, 1.0799E+03,
     +  1.0811E+03, 1.0823E+03, 1.0835E+03, 1.0847E+03, 1.0859E+03,
     + 1.0871E+03, 1.0883E+03, 1.0895E+03, 1.0907E+03, 1.0920E+03,
     +  1.0932E+03, 1.0944E+03, 1.0956E+03, 1.0968E+03, 1.0980E+03,
     + 1.0992E+03, 1.1004E+03, 1.1016E+03, 1.1029E+03, 1.1041E+03,
     +  1.1053E+03, 1.1065E+03, 1.1077E+03, 1.1089E+03, 1.1102E+03,
     + 1.1114E+03, 1.1126E+03, 1.1138E+03, 1.1151E+03, 1.1163E+03,
     +  1.1175E+03, 1.1187E+03, 1.1199E+03, 1.1212E+03, 1.1224E+03,
     + 1.1236E+03, 1.1248E+03, 1.1261E+03, 1.1273E+03, 1.1285E+03,
     +  1.1298E+03, 1.1310E+03, 1.1322E+03, 1.1335E+03, 1.1347E+03,
     + 1.1359E+03, 1.1372E+03, 1.1384E+03, 1.1396E+03, 1.1409E+03,
     +  1.1421E+03, 1.1433E+03, 1.1446E+03, 1.1458E+03, 1.1471E+03,
     + 1.1483E+03, 1.1495E+03, 1.1508E+03, 1.1520E+03, 1.1533E+03,
     +  1.1545E+03, 1.1558E+03, 1.1570E+03, 1.1582E+03, 1.1595E+03,
     + 1.1607E+03, 1.1620E+03, 1.1632E+03, 1.1645E+03, 1.1657E+03,
     +  1.1670E+03, 1.1682E+03, 1.1695E+03, 1.1707E+03, 1.1720E+03,
     + 1.1733E+03, 1.1745E+03, 1.1758E+03, 1.1770E+03, 1.1783E+03,
     +  1.1795E+03, 1.1808E+03, 1.1821E+03, 1.1833E+03, 1.1846E+03,
     + 1.1858E+03, 1.1871E+03, 1.1884E+03, 1.1896E+03, 1.1909E+03,
     +  1.1922E+03, 1.1934E+03, 1.1947E+03, 1.1959E+03, 1.1972E+03,
     + 1.1985E+03, 1.1997E+03, 1.2010E+03, 1.2023E+03, 1.2036E+03,
     +  1.2048E+03, 1.2061E+03, 1.2074E+03, 1.2086E+03, 1.2099E+03,
     + 1.2112E+03, 1.2125E+03, 1.2138E+03, 1.2150E+03, 1.2163E+03,
     +  1.2176E+03, 1.2189E+03, 1.2201E+03, 1.2214E+03, 1.2227E+03,
     + 1.2240E+03, 1.2253E+03, 1.2266E+03, 1.2278E+03, 1.2291E+03,
     +  1.2304E+03, 1.2317E+03, 1.2330E+03, 1.2343E+03, 1.2356E+03,
     + 1.2368E+03, 1.2381E+03, 1.2394E+03, 1.2407E+03, 1.2420E+03,
     +  1.2433E+03, 1.2446E+03, 1.2459E+03, 1.2472E+03, 1.2485E+03,
     + 1.2498E+03, 1.2511E+03, 1.2524E+03, 1.2537E+03, 1.2550E+03,
     +  1.2563E+03, 1.2576E+03, 1.2589E+03, 1.2602E+03, 1.2615E+03,
     + 1.2628E+03, 1.2641E+03, 1.2654E+03, 1.2667E+03, 1.2680E+03,
     +  1.2693E+03, 1.2706E+03, 1.2719E+03, 1.2732E+03, 1.2745E+03,
     + 1.2758E+03, 1.2772E+03, 1.2785E+03, 1.2798E+03, 1.2811E+03,
     +  1.2824E+03, 1.2837E+03, 1.2850E+03, 1.2863E+03, 1.2877E+03,
     + 1.2890E+03, 1.2903E+03, 1.2916E+03, 1.2929E+03, 1.2942E+03,
     +  1.2956E+03, 1.2969E+03, 1.2982E+03, 1.2995E+03, 1.3009E+03,
     + 1.3022E+03, 1.3035E+03, 1.3048E+03, 1.3062E+03, 1.3075E+03,
     +  1.3088E+03, 1.3101E+03, 1.3115E+03, 1.3128E+03, 1.3141E+03,
     + 1.3155E+03, 1.3168E+03, 1.3181E+03, 1.3194E+03, 1.3208E+03,
     +  1.3221E+03, 1.3235E+03, 1.3248E+03, 1.3261E+03, 1.3275E+03,
     + 1.3288E+03, 1.3301E+03, 1.3315E+03, 1.3328E+03, 1.3342E+03,
     +  1.3355E+03, 1.3368E+03, 1.3382E+03, 1.3395E+03, 1.3409E+03,
     + 1.3422E+03, 1.3436E+03, 1.3449E+03, 1.3462E+03, 1.3476E+03,
     +  1.3489E+03, 1.3503E+03, 1.3516E+03, 1.3530E+03, 1.3543E+03,
     + 1.3557E+03, 1.3570E+03, 1.3584E+03, 1.3598E+03, 1.3611E+03,
     +  1.3625E+03, 1.3638E+03, 1.3652E+03, 1.3665E+03, 1.3679E+03,
     + 1.3693E+03, 1.3706E+03, 1.3720E+03, 1.3733E+03, 1.3747E+03,
     +  1.3761E+03, 1.3774E+03, 1.3788E+03, 1.3801E+03, 1.3815E+03,
     + 1.3829E+03, 1.3842E+03, 1.3856E+03, 1.3870E+03, 1.3884E+03,
     +  1.3897E+03, 1.3911E+03, 1.3925E+03, 1.3938E+03, 1.3952E+03,
     + 1.3966E+03, 1.3979E+03, 1.3993E+03, 1.4007E+03, 1.4021E+03,
     +  1.4034E+03, 1.4048E+03, 1.4062E+03, 1.4076E+03, 1.4090E+03,
     + 1.4103E+03, 1.4117E+03, 1.4131E+03, 1.4145E+03, 1.4159E+03,
     +  1.4172E+03, 1.4186E+03, 1.4200E+03, 1.4214E+03, 1.4228E+03,
     + 1.4242E+03, 1.4256E+03, 1.4270E+03, 1.4283E+03, 1.4297E+03,
     +  1.4311E+03, 1.4325E+03, 1.4339E+03, 1.4353E+03, 1.4367E+03,
     + 1.4381E+03, 1.4395E+03, 1.4409E+03, 1.4423E+03, 1.4437E+03,
     +  1.4451E+03, 1.4465E+03, 1.4479E+03, 1.4493E+03, 1.4507E+03,
     + 1.4521E+03, 1.4535E+03, 1.4549E+03, 1.4563E+03, 1.4577E+03,
     +  1.4591E+03, 1.4605E+03, 1.4619E+03, 1.4633E+03, 1.4647E+03,
     + 1.4661E+03, 1.4675E+03, 1.4689E+03, 1.4703E+03, 1.4717E+03,
     +  1.4732E+03, 1.4746E+03, 1.4760E+03, 1.4774E+03, 1.4788E+03,
     + 1.4802E+03, 1.4816E+03, 1.4831E+03, 1.4845E+03, 1.4859E+03,
     +  1.4873E+03, 1.4887E+03, 1.4902E+03, 1.4916E+03, 1.4930E+03,
     + 1.4944E+03, 1.4958E+03, 1.4973E+03, 1.4987E+03, 1.5001E+03,
     +  1.5015E+03, 1.5030E+03, 1.5044E+03, 1.5058E+03, 1.5072E+03,
     + 1.5087E+03, 1.5101E+03, 1.5115E+03, 1.5130E+03, 1.5144E+03,
     +  1.5158E+03, 1.5173E+03, 1.5187E+03, 1.5201E+03, 1.5216E+03,
     + 1.5230E+03, 1.5244E+03, 1.5259E+03, 1.5273E+03, 1.5288E+03,
     +  1.5302E+03, 1.5316E+03, 1.5331E+03, 1.5345E+03, 1.5360E+03,
     + 1.5374E+03, 1.5389E+03, 1.5403E+03, 1.5417E+03, 1.5432E+03,
     +  1.5446E+03, 1.5461E+03, 1.5475E+03, 1.5490E+03, 1.5504E+03,
     + 1.5519E+03, 1.5533E+03, 1.5548E+03, 1.5562E+03, 1.5577E+03,
     +  1.5591E+03, 1.5606E+03, 1.5620E+03, 1.5635E+03, 1.5650E+03,
     + 1.5664E+03, 1.5679E+03, 1.5693E+03, 1.5708E+03, 1.5723E+03,
     +  1.5737E+03, 1.5752E+03, 1.5766E+03, 1.5781E+03, 1.5796E+03,
     + 1.5810E+03, 1.5825E+03, 1.5840E+03, 1.5854E+03, 1.5869E+03,
     +  1.5884E+03, 1.5898E+03, 1.5913E+03, 1.5928E+03, 1.5942E+03,
     + 1.5957E+03, 1.5972E+03, 1.5987E+03, 1.6001E+03, 1.6016E+03,
     +  1.6031E+03, 1.6046E+03, 1.6060E+03, 1.6075E+03, 1.6090E+03,
     + 1.6105E+03, 1.6119E+03, 1.6134E+03, 1.6149E+03, 1.6164E+03,
     +  1.6179E+03, 1.6194E+03, 1.6208E+03, 1.6223E+03, 1.6238E+03,
     + 1.6253E+03, 1.6268E+03, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        62
      data (QofT( 3,J),J=1,2408)/ 2.3986E+01, 2.3993E+01, 2.3995E+01,
     +  2.3997E+01, 2.3997E+01, 2.3998E+01, 2.4001E+01, 2.4007E+01,
     + 2.4020E+01, 2.4044E+01, 2.4082E+01, 2.4137E+01, 2.4211E+01,
     +  2.4305E+01, 2.4420E+01, 2.4556E+01, 2.4712E+01, 2.4888E+01,
     + 2.5082E+01, 2.5294E+01, 2.5766E+01, 2.6295E+01, 2.6874E+01,
     +  2.7495E+01, 2.8155E+01, 2.8848E+01, 2.9572E+01, 3.0324E+01,
     + 3.1101E+01, 3.1902E+01, 3.2727E+01, 3.3573E+01, 3.4439E+01,
     +  3.5326E+01, 3.6232E+01, 3.7157E+01, 3.8100E+01, 3.9060E+01,
     + 4.0038E+01, 4.1032E+01, 4.2043E+01, 4.3069E+01, 4.4111E+01,
     +  4.5168E+01, 4.6240E+01, 4.7326E+01, 4.8425E+01, 4.9538E+01,
     + 5.0664E+01, 5.1803E+01, 5.2955E+01, 5.4118E+01, 5.5293E+01,
     +  5.6480E+01, 5.7677E+01, 5.8886E+01, 6.0105E+01, 6.1334E+01,
     + 6.2573E+01, 6.3822E+01, 6.5080E+01, 6.6347E+01, 6.7623E+01,
     +  6.8908E+01, 7.0201E+01, 7.1502E+01, 7.2811E+01, 7.4128E+01,
     + 7.5452E+01, 7.6784E+01, 7.8123E+01, 7.9468E+01, 8.0821E+01,
     +  8.2180E+01, 8.3545E+01, 8.4917E+01, 8.6294E+01, 8.7678E+01,
     + 8.9067E+01, 9.0462E+01, 9.1862E+01, 9.3268E+01, 9.4679E+01,
     +  9.6094E+01, 9.7515E+01, 9.8941E+01, 1.0037E+02, 1.0181E+02,
     + 1.0325E+02, 1.0469E+02, 1.0614E+02, 1.0759E+02, 1.0904E+02,
     +  1.1050E+02, 1.1197E+02, 1.1344E+02, 1.1491E+02, 1.1638E+02,
     + 1.1786E+02, 1.1934E+02, 1.2083E+02, 1.2232E+02, 1.2381E+02,
     +  1.2530E+02, 1.2680E+02, 1.2830E+02, 1.2981E+02, 1.3131E+02,
     + 1.3282E+02, 1.3433E+02, 1.3585E+02, 1.3737E+02, 1.3889E+02,
     +  1.4041E+02, 1.4194E+02, 1.4346E+02, 1.4499E+02, 1.4653E+02,
     + 1.4806E+02, 1.4960E+02, 1.5114E+02, 1.5268E+02, 1.5422E+02,
     +  1.5577E+02, 1.5732E+02, 1.5887E+02, 1.6042E+02, 1.6197E+02,
     + 1.6353E+02, 1.6508E+02, 1.6664E+02, 1.6820E+02, 1.6977E+02,
     +  1.7133E+02, 1.7290E+02, 1.7446E+02, 1.7603E+02, 1.7760E+02,
     + 1.7918E+02, 1.8075E+02, 1.8232E+02, 1.8390E+02, 1.8548E+02,
     +  1.8706E+02, 1.8864E+02, 1.9022E+02, 1.9181E+02, 1.9339E+02,
     + 1.9498E+02, 1.9657E+02, 1.9816E+02, 1.9975E+02, 2.0134E+02,
     +  2.0293E+02, 2.0453E+02, 2.0612E+02, 2.0772E+02, 2.0932E+02,
     + 2.1092E+02, 2.1252E+02, 2.1412E+02, 2.1572E+02, 2.1732E+02,
     +  2.1893E+02, 2.2053E+02, 2.2214E+02, 2.2374E+02, 2.2535E+02,
     + 2.2696E+02, 2.2857E+02, 2.3018E+02, 2.3180E+02, 2.3341E+02,
     +  2.3502E+02, 2.3664E+02, 2.3825E+02, 2.3987E+02, 2.4149E+02,
     + 2.4310E+02, 2.4472E+02, 2.4634E+02, 2.4796E+02, 2.4959E+02,
     +  2.5121E+02, 2.5283E+02, 2.5445E+02, 2.5608E+02, 2.5770E+02,
     + 2.5933E+02, 2.6096E+02, 2.6258E+02, 2.6421E+02, 2.6584E+02,
     +  2.6747E+02, 2.6910E+02, 2.7073E+02, 2.7236E+02, 2.7400E+02,
     + 2.7563E+02, 2.7726E+02, 2.7890E+02, 2.8053E+02, 2.8217E+02,
     +  2.8380E+02, 2.8544E+02, 2.8707E+02, 2.8871E+02, 2.9035E+02,
     + 2.9199E+02, 2.9363E+02, 2.9527E+02, 2.9691E+02, 2.9855E+02,
     +  3.0019E+02, 3.0183E+02, 3.0348E+02, 3.0512E+02, 3.0676E+02,
     + 3.0841E+02, 3.1005E+02, 3.1170E+02, 3.1335E+02, 3.1499E+02,
     +  3.1664E+02, 3.1829E+02, 3.1993E+02, 3.2158E+02, 3.2323E+02,
     + 3.2488E+02, 3.2653E+02, 3.2818E+02, 3.2983E+02, 3.3148E+02,
     +  3.3313E+02, 3.3478E+02, 3.3644E+02, 3.3809E+02, 3.3974E+02,
     + 3.4139E+02, 3.4305E+02, 3.4470E+02, 3.4636E+02, 3.4802E+02,
     +  3.4967E+02, 3.5133E+02, 3.5298E+02, 3.5464E+02, 3.5630E+02,
     + 3.5795E+02, 3.5961E+02, 3.6127E+02, 3.6293E+02, 3.6459E+02,
     +  3.6625E+02, 3.6791E+02, 3.6957E+02, 3.7123E+02, 3.7289E+02,
     + 3.7455E+02, 3.7621E+02, 3.7788E+02, 3.7954E+02, 3.8120E+02,
     +  3.8286E+02, 3.8453E+02, 3.8619E+02, 3.8786E+02, 3.8952E+02,
     + 3.9118E+02, 3.9285E+02, 3.9451E+02, 3.9618E+02, 3.9785E+02,
     +  3.9951E+02, 4.0118E+02, 4.0285E+02, 4.0452E+02, 4.0618E+02,
     + 4.0785E+02, 4.0952E+02, 4.1119E+02, 4.1286E+02, 4.1453E+02,
     +  4.1620E+02, 4.1787E+02, 4.1954E+02, 4.2121E+02, 4.2288E+02,
     + 4.2455E+02, 4.2622E+02, 4.2790E+02, 4.2957E+02, 4.3124E+02,
     +  4.3291E+02, 4.3458E+02, 4.3626E+02, 4.3793E+02, 4.3961E+02,
     + 4.4128E+02, 4.4296E+02, 4.4463E+02, 4.4631E+02, 4.4798E+02,
     +  4.4966E+02, 4.5134E+02, 4.5301E+02, 4.5469E+02, 4.5636E+02,
     + 4.5804E+02, 4.5972E+02, 4.6140E+02, 4.6308E+02, 4.6476E+02,
     +  4.6643E+02, 4.6811E+02, 4.6979E+02, 4.7147E+02, 4.7315E+02,
     + 4.7484E+02, 4.7651E+02, 4.7820E+02, 4.7988E+02, 4.8156E+02,
     +  4.8324E+02, 4.8492E+02, 4.8661E+02, 4.8829E+02, 4.8997E+02,
     + 4.9166E+02, 4.9334E+02, 4.9502E+02, 4.9671E+02, 4.9839E+02,
     +  5.0008E+02, 5.0176E+02, 5.0345E+02, 5.0513E+02, 5.0682E+02,
     + 5.0851E+02, 5.1019E+02, 5.1188E+02, 5.1357E+02, 5.1526E+02,
     +  5.1694E+02, 5.1863E+02, 5.2032E+02, 5.2201E+02, 5.2370E+02,
     + 5.2539E+02, 5.2708E+02, 5.2877E+02, 5.3046E+02, 5.3215E+02,
     +  5.3384E+02, 5.3554E+02, 5.3723E+02, 5.3892E+02, 5.4061E+02,
     + 5.4230E+02, 5.4399E+02, 5.4569E+02, 5.4738E+02, 5.4908E+02,
     +  5.5077E+02, 5.5246E+02, 5.5416E+02, 5.5586E+02, 5.5755E+02,
     + 5.5925E+02, 5.6095E+02, 5.6265E+02, 5.6434E+02, 5.6604E+02,
     +  5.6774E+02, 5.6943E+02, 5.7113E+02, 5.7283E+02, 5.7453E+02,
     + 5.7623E+02, 5.7793E+02, 5.7963E+02, 5.8133E+02, 5.8303E+02,
     +  5.8473E+02, 5.8643E+02, 5.8813E+02, 5.8984E+02, 5.9154E+02,
     + 5.9324E+02, 5.9495E+02, 5.9666E+02, 5.9836E+02, 6.0006E+02,
     +  6.0177E+02, 6.0347E+02, 6.0518E+02, 6.0689E+02, 6.0859E+02,
     + 6.1030E+02, 6.1201E+02, 6.1371E+02, 6.1542E+02, 6.1713E+02,
     +  6.1884E+02, 6.2055E+02, 6.2226E+02, 6.2397E+02, 6.2568E+02,
     + 6.2739E+02, 6.2910E+02, 6.3081E+02, 6.3253E+02, 6.3423E+02,
     +  6.3595E+02, 6.3766E+02, 6.3938E+02, 6.4109E+02, 6.4280E+02,
     + 6.4452E+02, 6.4624E+02, 6.4795E+02, 6.4967E+02, 6.5138E+02,
     +  6.5310E+02, 6.5482E+02, 6.5654E+02, 6.5825E+02, 6.5997E+02,
     + 6.6170E+02, 6.6341E+02, 6.6513E+02, 6.6685E+02, 6.6858E+02,
     +  6.7030E+02, 6.7202E+02, 6.7374E+02, 6.7546E+02, 6.7719E+02,
     + 6.7891E+02, 6.8064E+02, 6.8236E+02, 6.8409E+02, 6.8581E+02,
     +  6.8754E+02, 6.8926E+02, 6.9099E+02, 6.9272E+02, 6.9444E+02,
     + 6.9617E+02, 6.9790E+02, 6.9963E+02, 7.0136E+02, 7.0310E+02,
     +  7.0483E+02, 7.0656E+02, 7.0829E+02, 7.1002E+02, 7.1175E+02,
     + 7.1349E+02, 7.1522E+02, 7.1696E+02, 7.1869E+02, 7.2043E+02,
     +  7.2216E+02, 7.2390E+02, 7.2564E+02, 7.2737E+02, 7.2911E+02,
     + 7.3086E+02, 7.3259E+02, 7.3433E+02, 7.3607E+02, 7.3781E+02,
     +  7.3955E+02, 7.4130E+02, 7.4303E+02, 7.4478E+02, 7.4652E+02,
     + 7.4827E+02, 7.5001E+02, 7.5176E+02, 7.5350E+02, 7.5525E+02,
     +  7.5699E+02, 7.5874E+02, 7.6049E+02, 7.6224E+02, 7.6399E+02,
     + 7.6574E+02, 7.6749E+02, 7.6924E+02, 7.7099E+02, 7.7274E+02,
     +  7.7449E+02, 7.7625E+02, 7.7800E+02, 7.7976E+02, 7.8151E+02,
     + 7.8326E+02, 7.8502E+02, 7.8678E+02, 7.8854E+02, 7.9029E+02,
     +  7.9205E+02, 7.9381E+02, 7.9557E+02, 7.9733E+02, 7.9909E+02,
     + 8.0085E+02, 8.0261E+02, 8.0438E+02, 8.0615E+02, 8.0791E+02,
     +  8.0967E+02, 8.1143E+02, 8.1320E+02, 8.1497E+02, 8.1673E+02,
     + 8.1850E+02, 8.2027E+02, 8.2204E+02, 8.2381E+02, 8.2558E+02,
     +  8.2735E+02, 8.2912E+02, 8.3089E+02, 8.3267E+02, 8.3444E+02,
     + 8.3622E+02, 8.3799E+02, 8.3976E+02, 8.4154E+02, 8.4332E+02,
     +  8.4509E+02, 8.4687E+02, 8.4866E+02, 8.5044E+02, 8.5221E+02,
     + 8.5399E+02, 8.5578E+02, 8.5756E+02, 8.5934E+02, 8.6113E+02,
     +  8.6291E+02, 8.6470E+02, 8.6648E+02, 8.6827E+02, 8.7005E+02,
     + 8.7184E+02, 8.7363E+02, 8.7542E+02, 8.7721E+02, 8.7900E+02,
     +  8.8079E+02, 8.8259E+02, 8.8438E+02, 8.8617E+02, 8.8797E+02,
     + 8.8976E+02, 8.9156E+02, 8.9335E+02, 8.9515E+02, 8.9694E+02,
     +  8.9875E+02, 9.0054E+02, 9.0234E+02, 9.0415E+02, 9.0595E+02,
     + 9.0775E+02, 9.0955E+02, 9.1136E+02, 9.1316E+02, 9.1497E+02,
     +  9.1678E+02, 9.1858E+02, 9.2039E+02, 9.2220E+02, 9.2401E+02,
     + 9.2582E+02, 9.2762E+02, 9.2943E+02, 9.3124E+02, 9.3306E+02,
     +  9.3488E+02, 9.3669E+02, 9.3850E+02, 9.4032E+02, 9.4213E+02,
     + 9.4396E+02, 9.4577E+02, 9.4759E+02, 9.4941E+02, 9.5123E+02,
     +  9.5306E+02, 9.5487E+02, 9.5670E+02, 9.5852E+02, 9.6035E+02,
     + 9.6218E+02, 9.6399E+02, 9.6582E+02, 9.6765E+02, 9.6948E+02,
     +  9.7131E+02, 9.7315E+02, 9.7498E+02, 9.7681E+02, 9.7864E+02,
     + 9.8047E+02, 9.8231E+02, 9.8414E+02, 9.8597E+02, 9.8782E+02,
     +  9.8965E+02, 9.9149E+02, 9.9333E+02, 9.9517E+02, 9.9701E+02,
     + 9.9885E+02, 1.0007E+03, 1.0025E+03, 1.0044E+03, 1.0062E+03,
     +  1.0081E+03, 1.0099E+03, 1.0118E+03, 1.0136E+03, 1.0155E+03,
     + 1.0173E+03, 1.0192E+03, 1.0210E+03, 1.0229E+03, 1.0247E+03,
     +  1.0266E+03, 1.0284E+03, 1.0303E+03, 1.0322E+03, 1.0340E+03,
     + 1.0359E+03, 1.0377E+03, 1.0396E+03, 1.0415E+03, 1.0433E+03,
     +  1.0452E+03, 1.0470E+03, 1.0489E+03, 1.0508E+03, 1.0526E+03,
     + 1.0545E+03, 1.0564E+03, 1.0582E+03, 1.0601E+03, 1.0620E+03,
     +  1.0639E+03, 1.0657E+03, 1.0676E+03, 1.0695E+03, 1.0714E+03,
     + 1.0732E+03, 1.0751E+03, 1.0770E+03, 1.0789E+03, 1.0808E+03,
     +  1.0826E+03, 1.0845E+03, 1.0864E+03, 1.0883E+03, 1.0902E+03,
     + 1.0921E+03, 1.0939E+03, 1.0958E+03, 1.0977E+03, 1.0996E+03,
     +  1.1015E+03, 1.1034E+03, 1.1053E+03, 1.1072E+03, 1.1091E+03,
     + 1.1110E+03, 1.1129E+03, 1.1148E+03, 1.1167E+03, 1.1186E+03,
     +  1.1205E+03, 1.1224E+03, 1.1243E+03, 1.1262E+03, 1.1281E+03,
     + 1.1300E+03, 1.1319E+03, 1.1338E+03, 1.1357E+03, 1.1376E+03,
     +  1.1395E+03, 1.1414E+03, 1.1433E+03, 1.1452E+03, 1.1472E+03,
     + 1.1491E+03, 1.1510E+03, 1.1529E+03, 1.1548E+03, 1.1567E+03,
     +  1.1587E+03, 1.1606E+03, 1.1625E+03, 1.1644E+03, 1.1663E+03,
     + 1.1683E+03, 1.1702E+03, 1.1721E+03, 1.1740E+03, 1.1760E+03,
     +  1.1779E+03, 1.1798E+03, 1.1818E+03, 1.1837E+03, 1.1856E+03,
     + 1.1876E+03, 1.1895E+03, 1.1914E+03, 1.1934E+03, 1.1953E+03,
     +  1.1973E+03, 1.1992E+03, 1.2011E+03, 1.2031E+03, 1.2050E+03,
     + 1.2070E+03, 1.2089E+03, 1.2109E+03, 1.2128E+03, 1.2148E+03,
     +  1.2167E+03, 1.2187E+03, 1.2206E+03, 1.2226E+03, 1.2245E+03,
     + 1.2265E+03, 1.2284E+03, 1.2304E+03, 1.2323E+03, 1.2343E+03,
     +  1.2362E+03, 1.2382E+03, 1.2402E+03, 1.2421E+03, 1.2441E+03,
     + 1.2461E+03, 1.2480E+03, 1.2500E+03, 1.2520E+03, 1.2539E+03,
     +  1.2559E+03, 1.2579E+03, 1.2598E+03, 1.2618E+03, 1.2638E+03,
     + 1.2658E+03, 1.2677E+03, 1.2697E+03, 1.2717E+03, 1.2737E+03,
     +  1.2757E+03, 1.2776E+03, 1.2796E+03, 1.2816E+03, 1.2836E+03,
     + 1.2856E+03, 1.2876E+03, 1.2895E+03, 1.2915E+03, 1.2935E+03,
     +  1.2955E+03, 1.2975E+03, 1.2995E+03, 1.3015E+03, 1.3035E+03,
     + 1.3055E+03, 1.3075E+03, 1.3095E+03, 1.3115E+03, 1.3135E+03,
     +  1.3155E+03, 1.3175E+03, 1.3195E+03, 1.3215E+03, 1.3235E+03,
     + 1.3255E+03, 1.3275E+03, 1.3295E+03, 1.3315E+03, 1.3336E+03,
     +  1.3356E+03, 1.3376E+03, 1.3396E+03, 1.3416E+03, 1.3436E+03,
     + 1.3457E+03, 1.3477E+03, 1.3497E+03, 1.3517E+03, 1.3537E+03,
     +  1.3558E+03, 1.3578E+03, 1.3598E+03, 1.3618E+03, 1.3639E+03,
     + 1.3659E+03, 1.3679E+03, 1.3700E+03, 1.3720E+03, 1.3740E+03,
     +  1.3761E+03, 1.3781E+03, 1.3801E+03, 1.3822E+03, 1.3842E+03,
     + 1.3862E+03, 1.3883E+03, 1.3903E+03, 1.3924E+03, 1.3944E+03,
     +  1.3965E+03, 1.3985E+03, 1.4006E+03, 1.4026E+03, 1.4047E+03,
     + 1.4067E+03, 1.4088E+03, 1.4108E+03, 1.4129E+03, 1.4149E+03,
     +  1.4170E+03, 1.4190E+03, 1.4211E+03, 1.4232E+03, 1.4252E+03,
     + 1.4273E+03, 1.4293E+03, 1.4314E+03, 1.4335E+03, 1.4355E+03,
     +  1.4376E+03, 1.4397E+03, 1.4417E+03, 1.4438E+03, 1.4459E+03,
     + 1.4480E+03, 1.4500E+03, 1.4521E+03, 1.4542E+03, 1.4563E+03,
     +  1.4584E+03, 1.4604E+03, 1.4625E+03, 1.4646E+03, 1.4667E+03,
     + 1.4688E+03, 1.4709E+03, 1.4730E+03, 1.4750E+03, 1.4771E+03,
     +  1.4792E+03, 1.4813E+03, 1.4834E+03, 1.4855E+03, 1.4876E+03,
     + 1.4897E+03, 1.4918E+03, 1.4939E+03, 1.4960E+03, 1.4981E+03,
     +  1.5002E+03, 1.5023E+03, 1.5044E+03, 1.5065E+03, 1.5086E+03,
     + 1.5107E+03, 1.5128E+03, 1.5149E+03, 1.5171E+03, 1.5192E+03,
     +  1.5213E+03, 1.5234E+03, 1.5255E+03, 1.5276E+03, 1.5298E+03,
     + 1.5319E+03, 1.5340E+03, 1.5361E+03, 1.5382E+03, 1.5404E+03,
     +  1.5425E+03, 1.5446E+03, 1.5468E+03, 1.5489E+03, 1.5510E+03,
     + 1.5532E+03, 1.5553E+03, 1.5574E+03, 1.5596E+03, 1.5617E+03,
     +  1.5638E+03, 1.5660E+03, 1.5681E+03, 1.5703E+03, 1.5724E+03,
     + 1.5745E+03, 1.5767E+03, 1.5788E+03, 1.5810E+03, 1.5831E+03,
     +  1.5853E+03, 1.5874E+03, 1.5896E+03, 1.5917E+03, 1.5939E+03,
     + 1.5960E+03, 1.5982E+03, 1.6004E+03, 1.6025E+03, 1.6047E+03,
     +  1.6069E+03, 1.6090E+03, 1.6112E+03, 1.6133E+03, 1.6155E+03,
     + 1.6177E+03, 1.6198E+03, 1.6220E+03, 1.6242E+03, 1.6264E+03,
     +  1.6285E+03, 1.6307E+03, 1.6329E+03, 1.6351E+03, 1.6372E+03,
     + 1.6394E+03, 1.6416E+03, 1.6438E+03, 1.6460E+03, 1.6482E+03,
     +  1.6503E+03, 1.6525E+03, 1.6547E+03, 1.6569E+03, 1.6591E+03,
     + 1.6613E+03, 1.6635E+03, 1.6657E+03, 1.6679E+03, 1.6701E+03,
     +  1.6723E+03, 1.6745E+03, 1.6767E+03, 1.6789E+03, 1.6811E+03,
     + 1.6833E+03, 1.6855E+03, 1.6877E+03, 1.6899E+03, 1.6921E+03,
     +  1.6943E+03, 1.6965E+03, 1.6987E+03, 1.7010E+03, 1.7032E+03,
     + 1.7054E+03, 1.7076E+03, 1.7098E+03, 1.7121E+03, 1.7143E+03,
     +  1.7165E+03, 1.7187E+03, 1.7209E+03, 1.7232E+03, 1.7254E+03,
     + 1.7276E+03, 1.7299E+03, 1.7321E+03, 1.7343E+03, 1.7366E+03,
     +  1.7388E+03, 1.7410E+03, 1.7433E+03, 1.7455E+03, 1.7478E+03,
     + 1.7500E+03, 1.7522E+03, 1.7545E+03, 1.7567E+03, 1.7590E+03,
     +  1.7612E+03, 1.7635E+03, 1.7657E+03, 1.7680E+03, 1.7702E+03,
     + 1.7725E+03, 1.7747E+03, 1.7770E+03, 1.7792E+03, 1.7815E+03,
     +  1.7838E+03, 1.7860E+03, 1.7883E+03, 1.7906E+03, 1.7928E+03,
     + 1.7951E+03, 1.7974E+03, 1.7996E+03, 1.8019E+03, 1.8042E+03,
     +  1.8064E+03, 1.8087E+03, 1.8110E+03, 1.8133E+03, 1.8156E+03,
     + 1.8178E+03, 1.8201E+03, 1.8224E+03, 1.8247E+03, 1.8270E+03,
     +  1.8292E+03, 1.8315E+03, 1.8338E+03, 1.8361E+03, 1.8384E+03,
     + 1.8407E+03, 1.8430E+03, 1.8453E+03, 1.8476E+03, 1.8499E+03,
     +  1.8522E+03, 1.8545E+03, 1.8568E+03, 1.8591E+03, 1.8614E+03,
     + 1.8637E+03, 1.8660E+03, 1.8683E+03, 1.8706E+03, 1.8729E+03,
     +  1.8752E+03, 1.8775E+03, 1.8799E+03, 1.8822E+03, 1.8845E+03,
     + 1.8868E+03, 1.8891E+03, 1.8914E+03, 1.8938E+03, 1.8961E+03,
     +  1.8984E+03, 1.9007E+03, 1.9031E+03, 1.9054E+03, 1.9077E+03,
     + 1.9101E+03, 1.9124E+03, 1.9147E+03, 1.9170E+03, 1.9194E+03,
     +  1.9217E+03, 1.9241E+03, 1.9264E+03, 1.9287E+03, 1.9311E+03,
     + 1.9334E+03, 1.9358E+03, 1.9381E+03, 1.9405E+03, 1.9428E+03,
     +  1.9452E+03, 1.9475E+03, 1.9499E+03, 1.9522E+03, 1.9546E+03,
     + 1.9569E+03, 1.9593E+03, 1.9616E+03, 1.9640E+03, 1.9664E+03,
     +  1.9687E+03, 1.9711E+03, 1.9734E+03, 1.9758E+03, 1.9782E+03,
     + 1.9806E+03, 1.9829E+03, 1.9853E+03, 1.9877E+03, 1.9900E+03,
     +  1.9924E+03, 1.9948E+03, 1.9972E+03, 1.9996E+03, 2.0019E+03,
     + 2.0043E+03, 2.0067E+03, 2.0091E+03, 2.0115E+03, 2.0139E+03,
     +  2.0162E+03, 2.0186E+03, 2.0210E+03, 2.0234E+03, 2.0258E+03,
     + 2.0282E+03, 2.0306E+03, 2.0330E+03, 2.0354E+03, 2.0378E+03,
     +  2.0402E+03, 2.0426E+03, 2.0450E+03, 2.0474E+03, 2.0498E+03,
     + 2.0522E+03, 2.0546E+03, 2.0570E+03, 2.0595E+03, 2.0619E+03,
     +  2.0643E+03, 2.0667E+03, 2.0691E+03, 2.0715E+03, 2.0739E+03,
     + 2.0764E+03, 2.0788E+03, 2.0812E+03, 2.0836E+03, 2.0861E+03,
     +  2.0885E+03, 2.0909E+03, 2.0933E+03, 2.0958E+03, 2.0982E+03,
     + 2.1007E+03, 2.1031E+03, 2.1055E+03, 2.1080E+03, 2.1104E+03,
     +  2.1129E+03, 2.1153E+03, 2.1177E+03, 2.1202E+03, 2.1226E+03,
     + 2.1251E+03, 2.1275E+03, 2.1300E+03, 2.1324E+03, 2.1349E+03,
     +  2.1373E+03, 2.1398E+03, 2.1422E+03, 2.1447E+03, 2.1472E+03,
     + 2.1496E+03, 2.1521E+03, 2.1545E+03, 2.1570E+03, 2.1595E+03,
     +  2.1619E+03, 2.1644E+03, 2.1669E+03, 2.1694E+03, 2.1718E+03,
     + 2.1743E+03, 2.1768E+03, 2.1793E+03, 2.1817E+03, 2.1842E+03,
     +  2.1867E+03, 2.1892E+03, 2.1917E+03, 2.1941E+03, 2.1966E+03,
     + 2.1991E+03, 2.2016E+03, 2.2041E+03, 2.2066E+03, 2.2091E+03,
     +  2.2116E+03, 2.2141E+03, 2.2165E+03, 2.2191E+03, 2.2216E+03,
     + 2.2240E+03, 2.2266E+03, 2.2290E+03, 2.2316E+03, 2.2341E+03,
     +  2.2366E+03, 2.2391E+03, 2.2416E+03, 2.2441E+03, 2.2466E+03,
     + 2.2491E+03, 2.2516E+03, 2.2542E+03, 2.2567E+03, 2.2592E+03,
     +  2.2617E+03, 2.2642E+03, 2.2668E+03, 2.2693E+03, 2.2718E+03,
     + 2.2743E+03, 2.2769E+03, 2.2794E+03, 2.2819E+03, 2.2845E+03,
     +  2.2870E+03, 2.2895E+03, 2.2921E+03, 2.2946E+03, 2.2971E+03,
     + 2.2997E+03, 2.3022E+03, 2.3048E+03, 2.3073E+03, 2.3099E+03,
     +  2.3124E+03, 2.3150E+03, 2.3175E+03, 2.3200E+03, 2.3226E+03,
     + 2.3252E+03, 2.3277E+03, 2.3303E+03, 2.3328E+03, 2.3354E+03,
     +  2.3379E+03, 2.3405E+03, 2.3431E+03, 2.3456E+03, 2.3482E+03,
     + 2.3508E+03, 2.3534E+03, 2.3559E+03, 2.3585E+03, 2.3611E+03,
     +  2.3636E+03, 2.3662E+03, 2.3688E+03, 2.3714E+03, 2.3740E+03,
     + 2.3765E+03, 2.3791E+03, 2.3817E+03, 2.3843E+03, 2.3869E+03,
     +  2.3894E+03, 2.3920E+03, 2.3946E+03, 2.3972E+03, 2.3998E+03,
     + 2.4024E+03, 2.4050E+03, 2.4076E+03, 2.4102E+03, 2.4128E+03,
     +  2.4154E+03, 2.4180E+03, 2.4206E+03, 2.4232E+03, 2.4258E+03,
     + 2.4284E+03, 2.4310E+03, 2.4337E+03, 2.4363E+03, 2.4389E+03,
     +  2.4415E+03, 2.4441E+03, 2.4467E+03, 2.4494E+03, 2.4520E+03,
     + 2.4546E+03, 2.4572E+03, 2.4598E+03, 2.4625E+03, 2.4651E+03,
     +  2.4677E+03, 2.4704E+03, 2.4730E+03, 2.4756E+03, 2.4782E+03,
     + 2.4809E+03, 2.4835E+03, 2.4861E+03, 2.4888E+03, 2.4914E+03,
     +  2.4941E+03, 2.4967E+03, 2.4994E+03, 2.5020E+03, 2.5047E+03,
     + 2.5073E+03, 2.5099E+03, 2.5126E+03, 2.5153E+03, 2.5179E+03,
     +  2.5206E+03, 2.5232E+03, 2.5259E+03, 2.5286E+03, 2.5312E+03,
     + 2.5339E+03, 2.5365E+03, 2.5392E+03, 2.5419E+03, 2.5445E+03,
     +  2.5472E+03, 2.5499E+03, 2.5526E+03, 2.5552E+03, 2.5579E+03,
     + 2.5606E+03, 2.5633E+03, 2.5659E+03, 2.5686E+03, 2.5713E+03,
     +  2.5740E+03, 2.5767E+03, 2.5793E+03, 2.5820E+03, 2.5847E+03,
     + 2.5874E+03, 2.5901E+03, 2.5928E+03, 2.5955E+03, 2.5982E+03,
     +  2.6009E+03, 2.6036E+03, 2.6063E+03, 2.6090E+03, 2.6117E+03,
     + 2.6144E+03, 2.6171E+03, 2.6198E+03, 2.6225E+03, 2.6252E+03,
     +  2.6279E+03, 2.6306E+03, 2.6334E+03, 2.6361E+03, 2.6388E+03,
     + 2.6415E+03, 2.6442E+03, 2.6469E+03, 2.6497E+03, 2.6524E+03,
     +  2.6551E+03, 2.6578E+03, 2.6606E+03, 2.6633E+03, 2.6660E+03,
     + 2.6687E+03, 2.6715E+03, 2.6742E+03, 2.6769E+03, 2.6797E+03,
     +  2.6824E+03, 2.6852E+03, 2.6879E+03, 2.6907E+03, 2.6934E+03,
     + 2.6961E+03, 2.6989E+03, 2.7016E+03, 2.7044E+03, 2.7071E+03,
     +  2.7099E+03, 2.7127E+03, 2.7154E+03, 2.7182E+03, 2.7209E+03,
     + 2.7237E+03, 2.7264E+03, 2.7292E+03, 2.7320E+03, 2.7347E+03,
     +  2.7375E+03, 2.7403E+03, 2.7430E+03, 2.7458E+03, 2.7486E+03,
     + 2.7513E+03, 2.7541E+03, 2.7569E+03, 2.7597E+03, 2.7624E+03,
     +  2.7652E+03, 2.7680E+03, 2.7708E+03, 2.7736E+03, 2.7763E+03,
     + 2.7791E+03, 2.7819E+03, 2.7847E+03, 2.7875E+03, 2.7903E+03,
     +  2.7931E+03, 2.7959E+03, 2.7987E+03, 2.8015E+03, 2.8043E+03,
     + 2.8071E+03, 2.8099E+03, 2.8127E+03, 2.8155E+03, 2.8183E+03,
     +  2.8211E+03, 2.8239E+03, 2.8267E+03, 2.8295E+03, 2.8324E+03,
     + 2.8352E+03, 2.8380E+03, 2.8408E+03, 2.8436E+03, 2.8465E+03,
     +  2.8493E+03, 2.8521E+03, 2.8549E+03, 2.8577E+03, 2.8606E+03,
     + 2.8634E+03, 2.8662E+03, 2.8690E+03, 2.8719E+03, 2.8747E+03,
     +  2.8776E+03, 2.8804E+03, 2.8832E+03, 2.8861E+03, 2.8889E+03,
     + 2.8918E+03, 2.8946E+03, 2.8975E+03, 2.9003E+03, 2.9031E+03,
     +  2.9060E+03, 2.9088E+03, 2.9117E+03, 2.9146E+03, 2.9174E+03,
     + 2.9203E+03, 2.9231E+03, 2.9260E+03, 2.9288E+03, 2.9317E+03,
     +  2.9346E+03, 2.9374E+03, 2.9403E+03, 2.9432E+03, 2.9460E+03,
     + 2.9489E+03, 2.9518E+03, 2.9547E+03, 2.9575E+03, 2.9604E+03,
     +  2.9633E+03, 2.9661E+03, 2.9690E+03, 2.9719E+03, 2.9748E+03,
     + 2.9777E+03, 2.9806E+03, 2.9835E+03, 2.9863E+03, 2.9892E+03,
     +  2.9921E+03, 2.9950E+03, 2.9979E+03, 3.0008E+03, 3.0037E+03,
     + 3.0066E+03, 3.0095E+03, 3.0124E+03, 3.0153E+03, 3.0182E+03,
     +  3.0211E+03, 3.0240E+03, 3.0269E+03, 3.0298E+03, 3.0328E+03,
     + 3.0357E+03, 3.0386E+03, 3.0415E+03, 3.0444E+03, 3.0473E+03,
     +  3.0503E+03, 3.0532E+03, 3.0561E+03, 3.0590E+03, 3.0620E+03,
     + 3.0649E+03, 3.0678E+03, 3.0707E+03, 3.0737E+03, 3.0766E+03,
     +  3.0795E+03, 3.0825E+03, 3.0854E+03, 3.0883E+03, 3.0913E+03,
     + 3.0942E+03, 3.0972E+03, 3.1001E+03, 3.1031E+03, 3.1060E+03,
     +  3.1089E+03, 3.1119E+03, 3.1149E+03, 3.1178E+03, 3.1208E+03,
     + 3.1237E+03, 3.1267E+03, 3.1296E+03, 3.1326E+03, 3.1355E+03,
     +  3.1385E+03, 3.1415E+03, 3.1444E+03, 3.1474E+03, 3.1504E+03,
     + 3.1533E+03, 3.1563E+03, 3.1593E+03, 3.1623E+03, 3.1652E+03,
     +  3.1682E+03, 3.1712E+03, 3.1742E+03, 3.1771E+03, 3.1801E+03,
     + 3.1831E+03, 3.1861E+03, 3.1891E+03, 3.1921E+03, 3.1951E+03,
     +  3.1980E+03, 3.2010E+03, 3.2040E+03, 3.2070E+03, 3.2100E+03,
     + 3.2130E+03, 3.2160E+03, 3.2190E+03, 3.2220E+03, 3.2250E+03,
     +  3.2280E+03, 3.2310E+03, 3.2341E+03, 3.2371E+03, 3.2401E+03,
     + 3.2431E+03, 3.2461E+03, 3.2491E+03, 3.2521E+03, 3.2551E+03,
     +  3.2582E+03, 3.2612E+03, 3.2642E+03, 3.2672E+03, 3.2703E+03,
     + 3.2733E+03, 3.2763E+03, 3.2793E+03, 3.2824E+03, 3.2854E+03,
     +  3.2884E+03, 3.2915E+03, 3.2945E+03, 3.2975E+03, 3.3006E+03,
     + 3.3036E+03, 3.3067E+03, 3.3097E+03, 3.3128E+03, 3.3158E+03,
     +  3.3188E+03, 3.3219E+03, 3.3249E+03, 3.3280E+03, 3.3311E+03,
     + 3.3341E+03, 3.3372E+03, 3.3402E+03, 3.3433E+03, 3.3463E+03,
     +  3.3494E+03, 3.3525E+03, 3.3555E+03, 3.3586E+03, 3.3617E+03,
     + 3.3647E+03, 3.3678E+03, 3.3709E+03, 3.3739E+03, 3.3770E+03,
     +  3.3801E+03, 3.3832E+03, 3.3862E+03, 3.3893E+03, 3.3924E+03,
     + 3.3955E+03, 3.3986E+03, 3.4017E+03, 3.4047E+03, 3.4078E+03,
     +  3.4109E+03, 3.4140E+03, 3.4171E+03, 3.4202E+03, 3.4233E+03,
     + 3.4264E+03, 3.4295E+03, 3.4326E+03, 3.4357E+03, 3.4388E+03,
     +  3.4419E+03, 3.4450E+03, 3.4481E+03, 3.4512E+03, 3.4543E+03,
     + 3.4574E+03, 3.4606E+03, 3.4637E+03, 3.4668E+03, 3.4699E+03,
     +  3.4730E+03, 3.4762E+03, 3.4793E+03, 3.4824E+03, 3.4855E+03,
     + 3.4886E+03, 3.4918E+03, 3.4949E+03, 3.4980E+03, 3.5012E+03,
     +  3.5043E+03, 3.5074E+03, 3.5106E+03, 3.5137E+03, 3.5168E+03,
     + 3.5200E+03, 3.5231E+03, 3.5263E+03, 3.5294E+03, 3.5326E+03,
     +  3.5357E+03, 3.5389E+03, 3.5420E+03, 3.5452E+03, 3.5483E+03,
     + 3.5515E+03, 3.5546E+03, 3.5578E+03, 3.5609E+03, 3.5641E+03,
     +  3.5673E+03, 3.5704E+03, 3.5736E+03, 3.5768E+03, 3.5799E+03,
     + 3.5831E+03, 3.5863E+03, 3.5894E+03, 3.5926E+03, 3.5958E+03,
     +  3.5990E+03, 3.6022E+03, 3.6053E+03, 3.6085E+03, 3.6117E+03,
     + 3.6149E+03, 3.6181E+03, 3.6212E+03, 3.6244E+03, 3.6276E+03,
     +  3.6308E+03, 3.6340E+03, 3.6372E+03, 3.6404E+03, 3.6436E+03,
     + 3.6468E+03, 3.6500E+03, 3.6532E+03, 3.6564E+03, 3.6596E+03,
     +  3.6628E+03, 3.6660E+03, 3.6692E+03, 3.6724E+03, 3.6756E+03,
     + 3.6788E+03, 3.6821E+03, 3.6853E+03, 3.6885E+03, 3.6917E+03,
     +  3.6949E+03, 3.6982E+03, 3.7014E+03, 3.7046E+03, 3.7078E+03,
     + 3.7110E+03, 3.7143E+03, 3.7175E+03, 3.7207E+03, 3.7240E+03,
     +  3.7272E+03, 3.7304E+03, 3.7337E+03, 3.7369E+03, 3.7402E+03,
     + 3.7434E+03, 3.7466E+03, 3.7499E+03, 3.7531E+03, 3.7564E+03,
     +  3.7596E+03, 3.7629E+03, 3.7661E+03, 3.7694E+03, 3.7726E+03,
     + 3.7759E+03, 3.7791E+03, 3.7824E+03, 3.7857E+03, 3.7889E+03,
     +  3.7922E+03, 3.7955E+03, 3.7987E+03, 3.8020E+03, 3.8053E+03,
     + 3.8085E+03, 3.8118E+03, 3.8151E+03, 3.8183E+03, 3.8216E+03,
     +  3.8249E+03, 3.8282E+03, 3.8315E+03, 3.8348E+03, 3.8380E+03,
     + 3.8413E+03, 3.8446E+03, 3.8479E+03, 3.8512E+03, 3.8545E+03,
     +  3.8578E+03, 3.8611E+03, 3.8643E+03, 3.8677E+03, 3.8710E+03,
     + 3.8742E+03, 3.8776E+03, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 3,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_HF    (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 2),QofT( 2,NT),Q(NT)
      data (Tmx(I),I=1,2)/6000.,6000./ 
      data xgj/ 4.,6./
c...       HF
c...        --        19
      data (QofT( 1,J),J=1,2408)/ 4.0000E+00, 4.0000E+00, 4.0000E+00,
     +  4.0000E+00, 4.0001E+00, 4.0006E+00, 4.0026E+00, 4.0074E+00,
     + 4.0168E+00, 4.0324E+00, 4.0554E+00, 4.0868E+00, 4.1268E+00,
     +  4.1756E+00, 4.2328E+00, 4.2979E+00, 4.3705E+00, 4.4499E+00,
     + 4.5353E+00, 4.6262E+00, 4.8220E+00, 5.0329E+00, 5.2554E+00,
     +  5.4868E+00, 5.7250E+00, 5.9686E+00, 6.2162E+00, 6.4671E+00,
     + 6.7207E+00, 6.9763E+00, 7.2337E+00, 7.4926E+00, 7.7527E+00,
     +  8.0138E+00, 8.2759E+00, 8.5386E+00, 8.8021E+00, 9.0662E+00,
     + 9.3308E+00, 9.5958E+00, 9.8613E+00, 1.0127E+01, 1.0393E+01,
     +  1.0660E+01, 1.0926E+01, 1.1193E+01, 1.1461E+01, 1.1728E+01,
     + 1.1996E+01, 1.2263E+01, 1.2531E+01, 1.2800E+01, 1.3068E+01,
     +  1.3336E+01, 1.3605E+01, 1.3873E+01, 1.4142E+01, 1.4411E+01,
     + 1.4680E+01, 1.4949E+01, 1.5218E+01, 1.5487E+01, 1.5756E+01,
     +  1.6026E+01, 1.6295E+01, 1.6564E+01, 1.6834E+01, 1.7103E+01,
     + 1.7373E+01, 1.7643E+01, 1.7912E+01, 1.8182E+01, 1.8452E+01,
     +  1.8722E+01, 1.8992E+01, 1.9262E+01, 1.9532E+01, 1.9802E+01,
     + 2.0072E+01, 2.0342E+01, 2.0612E+01, 2.0882E+01, 2.1152E+01,
     +  2.1422E+01, 2.1693E+01, 2.1963E+01, 2.2233E+01, 2.2504E+01,
     + 2.2774E+01, 2.3044E+01, 2.3315E+01, 2.3585E+01, 2.3856E+01,
     +  2.4126E+01, 2.4396E+01, 2.4667E+01, 2.4938E+01, 2.5208E+01,
     + 2.5479E+01, 2.5749E+01, 2.6020E+01, 2.6291E+01, 2.6561E+01,
     +  2.6832E+01, 2.7103E+01, 2.7373E+01, 2.7644E+01, 2.7915E+01,
     + 2.8185E+01, 2.8456E+01, 2.8727E+01, 2.8998E+01, 2.9269E+01,
     +  2.9539E+01, 2.9810E+01, 3.0081E+01, 3.0352E+01, 3.0623E+01,
     + 3.0894E+01, 3.1165E+01, 3.1436E+01, 3.1707E+01, 3.1978E+01,
     +  3.2249E+01, 3.2520E+01, 3.2791E+01, 3.3062E+01, 3.3333E+01,
     + 3.3604E+01, 3.3875E+01, 3.4146E+01, 3.4417E+01, 3.4688E+01,
     +  3.4959E+01, 3.5230E+01, 3.5501E+01, 3.5773E+01, 3.6044E+01,
     + 3.6315E+01, 3.6586E+01, 3.6857E+01, 3.7129E+01, 3.7400E+01,
     +  3.7671E+01, 3.7942E+01, 3.8213E+01, 3.8485E+01, 3.8756E+01,
     + 3.9027E+01, 3.9299E+01, 3.9570E+01, 3.9841E+01, 4.0113E+01,
     +  4.0384E+01, 4.0655E+01, 4.0927E+01, 4.1198E+01, 4.1469E+01,
     + 4.1741E+01, 4.2012E+01, 4.2283E+01, 4.2555E+01, 4.2826E+01,
     +  4.3098E+01, 4.3369E+01, 4.3641E+01, 4.3912E+01, 4.4184E+01,
     + 4.4455E+01, 4.4727E+01, 4.4998E+01, 4.5270E+01, 4.5541E+01,
     +  4.5813E+01, 4.6084E+01, 4.6356E+01, 4.6627E+01, 4.6899E+01,
     + 4.7170E+01, 4.7442E+01, 4.7714E+01, 4.7985E+01, 4.8257E+01,
     +  4.8528E+01, 4.8800E+01, 4.9072E+01, 4.9343E+01, 4.9615E+01,
     + 4.9887E+01, 5.0158E+01, 5.0430E+01, 5.0702E+01, 5.0974E+01,
     +  5.1245E+01, 5.1517E+01, 5.1789E+01, 5.2060E+01, 5.2332E+01,
     + 5.2604E+01, 5.2876E+01, 5.3148E+01, 5.3419E+01, 5.3691E+01,
     +  5.3963E+01, 5.4235E+01, 5.4507E+01, 5.4778E+01, 5.5050E+01,
     + 5.5322E+01, 5.5594E+01, 5.5866E+01, 5.6138E+01, 5.6410E+01,
     +  5.6682E+01, 5.6953E+01, 5.7225E+01, 5.7497E+01, 5.7769E+01,
     + 5.8041E+01, 5.8313E+01, 5.8585E+01, 5.8857E+01, 5.9129E+01,
     +  5.9401E+01, 5.9673E+01, 5.9945E+01, 6.0217E+01, 6.0489E+01,
     + 6.0761E+01, 6.1033E+01, 6.1305E+01, 6.1577E+01, 6.1849E+01,
     +  6.2121E+01, 6.2393E+01, 6.2666E+01, 6.2938E+01, 6.3210E+01,
     + 6.3482E+01, 6.3754E+01, 6.4026E+01, 6.4298E+01, 6.4570E+01,
     +  6.4843E+01, 6.5115E+01, 6.5387E+01, 6.5659E+01, 6.5931E+01,
     + 6.6204E+01, 6.6476E+01, 6.6748E+01, 6.7020E+01, 6.7293E+01,
     +  6.7565E+01, 6.7837E+01, 6.8109E+01, 6.8382E+01, 6.8654E+01,
     + 6.8926E+01, 6.9199E+01, 6.9471E+01, 6.9743E+01, 7.0016E+01,
     +  7.0288E+01, 7.0560E+01, 7.0833E+01, 7.1105E+01, 7.1378E+01,
     + 7.1650E+01, 7.1922E+01, 7.2195E+01, 7.2467E+01, 7.2740E+01,
     +  7.3012E+01, 7.3285E+01, 7.3557E+01, 7.3830E+01, 7.4102E+01,
     + 7.4375E+01, 7.4647E+01, 7.4920E+01, 7.5192E+01, 7.5465E+01,
     +  7.5737E+01, 7.6010E+01, 7.6283E+01, 7.6555E+01, 7.6828E+01,
     + 7.7100E+01, 7.7373E+01, 7.7646E+01, 7.7918E+01, 7.8191E+01,
     +  7.8464E+01, 7.8736E+01, 7.9009E+01, 7.9282E+01, 7.9555E+01,
     + 7.9827E+01, 8.0100E+01, 8.0373E+01, 8.0646E+01, 8.0919E+01,
     +  8.1191E+01, 8.1464E+01, 8.1737E+01, 8.2010E+01, 8.2283E+01,
     + 8.2556E+01, 8.2829E+01, 8.3102E+01, 8.3374E+01, 8.3647E+01,
     +  8.3920E+01, 8.4193E+01, 8.4466E+01, 8.4739E+01, 8.5012E+01,
     + 8.5285E+01, 8.5559E+01, 8.5832E+01, 8.6105E+01, 8.6378E+01,
     +  8.6651E+01, 8.6924E+01, 8.7197E+01, 8.7470E+01, 8.7744E+01,
     + 8.8017E+01, 8.8290E+01, 8.8563E+01, 8.8837E+01, 8.9110E+01,
     +  8.9383E+01, 8.9656E+01, 8.9930E+01, 9.0203E+01, 9.0476E+01,
     + 9.0750E+01, 9.1023E+01, 9.1297E+01, 9.1570E+01, 9.1844E+01,
     +  9.2117E+01, 9.2391E+01, 9.2664E+01, 9.2938E+01, 9.3211E+01,
     + 9.3485E+01, 9.3758E+01, 9.4032E+01, 9.4306E+01, 9.4579E+01,
     +  9.4853E+01, 9.5127E+01, 9.5401E+01, 9.5674E+01, 9.5948E+01,
     + 9.6222E+01, 9.6496E+01, 9.6770E+01, 9.7044E+01, 9.7317E+01,
     +  9.7591E+01, 9.7865E+01, 9.8139E+01, 9.8413E+01, 9.8687E+01,
     + 9.8961E+01, 9.9236E+01, 9.9510E+01, 9.9784E+01, 1.0006E+02,
     +  1.0033E+02, 1.0061E+02, 1.0088E+02, 1.0115E+02, 1.0143E+02,
     + 1.0170E+02, 1.0198E+02, 1.0225E+02, 1.0253E+02, 1.0280E+02,
     +  1.0308E+02, 1.0335E+02, 1.0362E+02, 1.0390E+02, 1.0417E+02,
     + 1.0445E+02, 1.0472E+02, 1.0500E+02, 1.0527E+02, 1.0555E+02,
     +  1.0582E+02, 1.0610E+02, 1.0637E+02, 1.0665E+02, 1.0692E+02,
     + 1.0720E+02, 1.0747E+02, 1.0775E+02, 1.0802E+02, 1.0830E+02,
     +  1.0857E+02, 1.0885E+02, 1.0912E+02, 1.0940E+02, 1.0967E+02,
     + 1.0995E+02, 1.1022E+02, 1.1050E+02, 1.1077E+02, 1.1105E+02,
     +  1.1132E+02, 1.1160E+02, 1.1188E+02, 1.1215E+02, 1.1243E+02,
     + 1.1270E+02, 1.1298E+02, 1.1325E+02, 1.1353E+02, 1.1381E+02,
     +  1.1408E+02, 1.1436E+02, 1.1463E+02, 1.1491E+02, 1.1519E+02,
     + 1.1546E+02, 1.1574E+02, 1.1601E+02, 1.1629E+02, 1.1657E+02,
     +  1.1684E+02, 1.1712E+02, 1.1740E+02, 1.1767E+02, 1.1795E+02,
     + 1.1823E+02, 1.1850E+02, 1.1878E+02, 1.1906E+02, 1.1933E+02,
     +  1.1961E+02, 1.1989E+02, 1.2016E+02, 1.2044E+02, 1.2072E+02,
     + 1.2099E+02, 1.2127E+02, 1.2155E+02, 1.2183E+02, 1.2210E+02,
     +  1.2238E+02, 1.2266E+02, 1.2293E+02, 1.2321E+02, 1.2349E+02,
     + 1.2377E+02, 1.2404E+02, 1.2432E+02, 1.2460E+02, 1.2488E+02,
     +  1.2516E+02, 1.2543E+02, 1.2571E+02, 1.2599E+02, 1.2627E+02,
     + 1.2655E+02, 1.2682E+02, 1.2710E+02, 1.2738E+02, 1.2766E+02,
     +  1.2794E+02, 1.2821E+02, 1.2849E+02, 1.2877E+02, 1.2905E+02,
     + 1.2933E+02, 1.2961E+02, 1.2989E+02, 1.3017E+02, 1.3044E+02,
     +  1.3072E+02, 1.3100E+02, 1.3128E+02, 1.3156E+02, 1.3184E+02,
     + 1.3212E+02, 1.3240E+02, 1.3268E+02, 1.3296E+02, 1.3324E+02,
     +  1.3352E+02, 1.3380E+02, 1.3408E+02, 1.3436E+02, 1.3464E+02,
     + 1.3492E+02, 1.3520E+02, 1.3548E+02, 1.3576E+02, 1.3604E+02,
     +  1.3632E+02, 1.3660E+02, 1.3688E+02, 1.3716E+02, 1.3744E+02,
     + 1.3772E+02, 1.3800E+02, 1.3828E+02, 1.3856E+02, 1.3884E+02,
     +  1.3912E+02, 1.3940E+02, 1.3969E+02, 1.3997E+02, 1.4025E+02,
     + 1.4053E+02, 1.4081E+02, 1.4109E+02, 1.4137E+02, 1.4166E+02,
     +  1.4194E+02, 1.4222E+02, 1.4250E+02, 1.4278E+02, 1.4306E+02,
     + 1.4335E+02, 1.4363E+02, 1.4391E+02, 1.4419E+02, 1.4448E+02,
     +  1.4476E+02, 1.4504E+02, 1.4532E+02, 1.4561E+02, 1.4589E+02,
     + 1.4617E+02, 1.4646E+02, 1.4674E+02, 1.4702E+02, 1.4730E+02,
     +  1.4759E+02, 1.4787E+02, 1.4815E+02, 1.4844E+02, 1.4872E+02,
     + 1.4901E+02, 1.4929E+02, 1.4957E+02, 1.4986E+02, 1.5014E+02,
     +  1.5043E+02, 1.5071E+02, 1.5099E+02, 1.5128E+02, 1.5156E+02,
     + 1.5185E+02, 1.5213E+02, 1.5242E+02, 1.5270E+02, 1.5299E+02,
     +  1.5327E+02, 1.5356E+02, 1.5384E+02, 1.5413E+02, 1.5441E+02,
     + 1.5470E+02, 1.5498E+02, 1.5527E+02, 1.5556E+02, 1.5584E+02,
     +  1.5613E+02, 1.5641E+02, 1.5670E+02, 1.5699E+02, 1.5727E+02,
     + 1.5756E+02, 1.5785E+02, 1.5813E+02, 1.5842E+02, 1.5871E+02,
     +  1.5899E+02, 1.5928E+02, 1.5957E+02, 1.5985E+02, 1.6014E+02,
     + 1.6043E+02, 1.6072E+02, 1.6100E+02, 1.6129E+02, 1.6158E+02,
     +  1.6187E+02, 1.6215E+02, 1.6244E+02, 1.6273E+02, 1.6302E+02,
     + 1.6331E+02, 1.6360E+02, 1.6388E+02, 1.6417E+02, 1.6446E+02,
     +  1.6475E+02, 1.6504E+02, 1.6533E+02, 1.6562E+02, 1.6591E+02,
     + 1.6620E+02, 1.6649E+02, 1.6678E+02, 1.6707E+02, 1.6736E+02,
     +  1.6765E+02, 1.6794E+02, 1.6823E+02, 1.6852E+02, 1.6881E+02,
     + 1.6910E+02, 1.6939E+02, 1.6968E+02, 1.6997E+02, 1.7026E+02,
     +  1.7055E+02, 1.7084E+02, 1.7113E+02, 1.7143E+02, 1.7172E+02,
     + 1.7201E+02, 1.7230E+02, 1.7259E+02, 1.7288E+02, 1.7318E+02,
     +  1.7347E+02, 1.7376E+02, 1.7405E+02, 1.7434E+02, 1.7464E+02,
     + 1.7493E+02, 1.7522E+02, 1.7552E+02, 1.7581E+02, 1.7610E+02,
     +  1.7640E+02, 1.7669E+02, 1.7698E+02, 1.7728E+02, 1.7757E+02,
     + 1.7786E+02, 1.7816E+02, 1.7845E+02, 1.7875E+02, 1.7904E+02,
     +  1.7933E+02, 1.7963E+02, 1.7992E+02, 1.8022E+02, 1.8051E+02,
     + 1.8081E+02, 1.8110E+02, 1.8140E+02, 1.8169E+02, 1.8199E+02,
     +  1.8229E+02, 1.8258E+02, 1.8288E+02, 1.8317E+02, 1.8347E+02,
     + 1.8377E+02, 1.8406E+02, 1.8436E+02, 1.8465E+02, 1.8495E+02,
     +  1.8525E+02, 1.8555E+02, 1.8584E+02, 1.8614E+02, 1.8644E+02,
     + 1.8673E+02, 1.8703E+02, 1.8733E+02, 1.8763E+02, 1.8793E+02,
     +  1.8822E+02, 1.8852E+02, 1.8882E+02, 1.8912E+02, 1.8942E+02,
     + 1.8972E+02, 1.9002E+02, 1.9031E+02, 1.9061E+02, 1.9091E+02,
     +  1.9121E+02, 1.9151E+02, 1.9181E+02, 1.9211E+02, 1.9241E+02,
     + 1.9271E+02, 1.9301E+02, 1.9331E+02, 1.9361E+02, 1.9391E+02,
     +  1.9421E+02, 1.9451E+02, 1.9482E+02, 1.9512E+02, 1.9542E+02,
     + 1.9572E+02, 1.9602E+02, 1.9632E+02, 1.9662E+02, 1.9693E+02,
     +  1.9723E+02, 1.9753E+02, 1.9783E+02, 1.9813E+02, 1.9844E+02,
     + 1.9874E+02, 1.9904E+02, 1.9935E+02, 1.9965E+02, 1.9995E+02,
     +  2.0026E+02, 2.0056E+02, 2.0086E+02, 2.0117E+02, 2.0147E+02,
     + 2.0177E+02, 2.0208E+02, 2.0238E+02, 2.0269E+02, 2.0299E+02,
     +  2.0330E+02, 2.0360E+02, 2.0391E+02, 2.0421E+02, 2.0452E+02,
     + 2.0482E+02, 2.0513E+02, 2.0543E+02, 2.0574E+02, 2.0605E+02,
     +  2.0635E+02, 2.0666E+02, 2.0697E+02, 2.0727E+02, 2.0758E+02,
     + 2.0789E+02, 2.0819E+02, 2.0850E+02, 2.0881E+02, 2.0912E+02,
     +  2.0942E+02, 2.0973E+02, 2.1004E+02, 2.1035E+02, 2.1065E+02,
     + 2.1096E+02, 2.1127E+02, 2.1158E+02, 2.1189E+02, 2.1220E+02,
     +  2.1251E+02, 2.1282E+02, 2.1313E+02, 2.1343E+02, 2.1374E+02,
     + 2.1405E+02, 2.1436E+02, 2.1467E+02, 2.1498E+02, 2.1530E+02,
     +  2.1561E+02, 2.1592E+02, 2.1623E+02, 2.1654E+02, 2.1685E+02,
     + 2.1716E+02, 2.1747E+02, 2.1778E+02, 2.1810E+02, 2.1841E+02,
     +  2.1872E+02, 2.1903E+02, 2.1934E+02, 2.1966E+02, 2.1997E+02,
     + 2.2028E+02, 2.2059E+02, 2.2091E+02, 2.2122E+02, 2.2153E+02,
     +  2.2185E+02, 2.2216E+02, 2.2248E+02, 2.2279E+02, 2.2310E+02,
     + 2.2342E+02, 2.2373E+02, 2.2405E+02, 2.2436E+02, 2.2468E+02,
     +  2.2499E+02, 2.2531E+02, 2.2562E+02, 2.2594E+02, 2.2625E+02,
     + 2.2657E+02, 2.2689E+02, 2.2720E+02, 2.2752E+02, 2.2784E+02,
     +  2.2815E+02, 2.2847E+02, 2.2879E+02, 2.2910E+02, 2.2942E+02,
     + 2.2974E+02, 2.3006E+02, 2.3037E+02, 2.3069E+02, 2.3101E+02,
     +  2.3133E+02, 2.3165E+02, 2.3196E+02, 2.3228E+02, 2.3260E+02,
     + 2.3292E+02, 2.3324E+02, 2.3356E+02, 2.3388E+02, 2.3420E+02,
     +  2.3452E+02, 2.3484E+02, 2.3516E+02, 2.3548E+02, 2.3580E+02,
     + 2.3612E+02, 2.3644E+02, 2.3676E+02, 2.3708E+02, 2.3740E+02,
     +  2.3773E+02, 2.3805E+02, 2.3837E+02, 2.3869E+02, 2.3901E+02,
     + 2.3934E+02, 2.3966E+02, 2.3998E+02, 2.4030E+02, 2.4063E+02,
     +  2.4095E+02, 2.4127E+02, 2.4160E+02, 2.4192E+02, 2.4224E+02,
     + 2.4257E+02, 2.4289E+02, 2.4322E+02, 2.4354E+02, 2.4386E+02,
     +  2.4419E+02, 2.4451E+02, 2.4484E+02, 2.4516E+02, 2.4549E+02,
     + 2.4582E+02, 2.4614E+02, 2.4647E+02, 2.4679E+02, 2.4712E+02,
     +  2.4745E+02, 2.4777E+02, 2.4810E+02, 2.4843E+02, 2.4875E+02,
     + 2.4908E+02, 2.4941E+02, 2.4973E+02, 2.5006E+02, 2.5039E+02,
     +  2.5072E+02, 2.5105E+02, 2.5138E+02, 2.5170E+02, 2.5203E+02,
     + 2.5236E+02, 2.5269E+02, 2.5302E+02, 2.5335E+02, 2.5368E+02,
     +  2.5401E+02, 2.5434E+02, 2.5467E+02, 2.5500E+02, 2.5533E+02,
     + 2.5566E+02, 2.5599E+02, 2.5632E+02, 2.5665E+02, 2.5698E+02,
     +  2.5732E+02, 2.5765E+02, 2.5798E+02, 2.5831E+02, 2.5864E+02,
     + 2.5898E+02, 2.5931E+02, 2.5964E+02, 2.5997E+02, 2.6031E+02,
     +  2.6064E+02, 2.6097E+02, 2.6131E+02, 2.6164E+02, 2.6198E+02,
     + 2.6231E+02, 2.6264E+02, 2.6298E+02, 2.6331E+02, 2.6365E+02,
     +  2.6398E+02, 2.6432E+02, 2.6465E+02, 2.6499E+02, 2.6532E+02,
     + 2.6566E+02, 2.6600E+02, 2.6633E+02, 2.6667E+02, 2.6701E+02,
     +  2.6734E+02, 2.6768E+02, 2.6802E+02, 2.6835E+02, 2.6869E+02,
     + 2.6903E+02, 2.6937E+02, 2.6970E+02, 2.7004E+02, 2.7038E+02,
     +  2.7072E+02, 2.7106E+02, 2.7140E+02, 2.7174E+02, 2.7207E+02,
     + 2.7241E+02, 2.7275E+02, 2.7309E+02, 2.7343E+02, 2.7377E+02,
     +  2.7411E+02, 2.7445E+02, 2.7480E+02, 2.7514E+02, 2.7548E+02,
     + 2.7582E+02, 2.7616E+02, 2.7650E+02, 2.7684E+02, 2.7718E+02,
     +  2.7753E+02, 2.7787E+02, 2.7821E+02, 2.7855E+02, 2.7890E+02,
     + 2.7924E+02, 2.7958E+02, 2.7993E+02, 2.8027E+02, 2.8061E+02,
     +  2.8096E+02, 2.8130E+02, 2.8165E+02, 2.8199E+02, 2.8234E+02,
     + 2.8268E+02, 2.8302E+02, 2.8337E+02, 2.8372E+02, 2.8406E+02,
     +  2.8441E+02, 2.8475E+02, 2.8510E+02, 2.8545E+02, 2.8579E+02,
     + 2.8614E+02, 2.8649E+02, 2.8683E+02, 2.8718E+02, 2.8753E+02,
     +  2.8787E+02, 2.8822E+02, 2.8857E+02, 2.8892E+02, 2.8927E+02,
     + 2.8961E+02, 2.8996E+02, 2.9031E+02, 2.9066E+02, 2.9101E+02,
     +  2.9136E+02, 2.9171E+02, 2.9206E+02, 2.9241E+02, 2.9276E+02,
     + 2.9311E+02, 2.9346E+02, 2.9381E+02, 2.9416E+02, 2.9451E+02,
     +  2.9486E+02, 2.9522E+02, 2.9557E+02, 2.9592E+02, 2.9627E+02,
     + 2.9662E+02, 2.9698E+02, 2.9733E+02, 2.9768E+02, 2.9803E+02,
     +  2.9839E+02, 2.9874E+02, 2.9909E+02, 2.9945E+02, 2.9980E+02,
     + 3.0016E+02, 3.0051E+02, 3.0087E+02, 3.0122E+02, 3.0157E+02,
     +  3.0193E+02, 3.0229E+02, 3.0264E+02, 3.0300E+02, 3.0335E+02,
     + 3.0371E+02, 3.0406E+02, 3.0442E+02, 3.0478E+02, 3.0513E+02,
     +  3.0549E+02, 3.0585E+02, 3.0621E+02, 3.0656E+02, 3.0692E+02,
     + 3.0728E+02, 3.0764E+02, 3.0799E+02, 3.0835E+02, 3.0871E+02,
     +  3.0907E+02, 3.0943E+02, 3.0979E+02, 3.1015E+02, 3.1051E+02,
     + 3.1087E+02, 3.1123E+02, 3.1159E+02, 3.1195E+02, 3.1231E+02,
     +  3.1267E+02, 3.1303E+02, 3.1339E+02, 3.1375E+02, 3.1412E+02,
     + 3.1448E+02, 3.1484E+02, 3.1520E+02, 3.1556E+02, 3.1593E+02,
     +  3.1629E+02, 3.1665E+02, 3.1701E+02, 3.1738E+02, 3.1774E+02,
     + 3.1810E+02, 3.1847E+02, 3.1883E+02, 3.1920E+02, 3.1956E+02,
     +  3.1993E+02, 3.2029E+02, 3.2066E+02, 3.2102E+02, 3.2139E+02,
     + 3.2175E+02, 3.2212E+02, 3.2248E+02, 3.2285E+02, 3.2322E+02,
     +  3.2358E+02, 3.2395E+02, 3.2432E+02, 3.2468E+02, 3.2505E+02,
     + 3.2542E+02, 3.2579E+02, 3.2615E+02, 3.2652E+02, 3.2689E+02,
     +  3.2726E+02, 3.2763E+02, 3.2800E+02, 3.2837E+02, 3.2874E+02,
     + 3.2911E+02, 3.2948E+02, 3.2985E+02, 3.3022E+02, 3.3059E+02,
     +  3.3096E+02, 3.3133E+02, 3.3170E+02, 3.3207E+02, 3.3244E+02,
     + 3.3281E+02, 3.3318E+02, 3.3356E+02, 3.3393E+02, 3.3430E+02,
     +  3.3467E+02, 3.3505E+02, 3.3542E+02, 3.3579E+02, 3.3616E+02,
     + 3.3654E+02, 3.3691E+02, 3.3729E+02, 3.3766E+02, 3.3803E+02,
     +  3.3841E+02, 3.3878E+02, 3.3916E+02, 3.3953E+02, 3.3991E+02,
     + 3.4028E+02, 3.4066E+02, 3.4104E+02, 3.4141E+02, 3.4179E+02,
     +  3.4216E+02, 3.4254E+02, 3.4292E+02, 3.4329E+02, 3.4367E+02,
     + 3.4405E+02, 3.4443E+02, 3.4480E+02, 3.4518E+02, 3.4556E+02,
     +  3.4594E+02, 3.4632E+02, 3.4670E+02, 3.4708E+02, 3.4746E+02,
     + 3.4784E+02, 3.4821E+02, 3.4859E+02, 3.4897E+02, 3.4936E+02,
     +  3.4974E+02, 3.5012E+02, 3.5050E+02, 3.5088E+02, 3.5126E+02,
     + 3.5164E+02, 3.5202E+02, 3.5240E+02, 3.5279E+02, 3.5317E+02,
     +  3.5355E+02, 3.5393E+02, 3.5432E+02, 3.5470E+02, 3.5508E+02,
     + 3.5547E+02, 3.5585E+02, 3.5623E+02, 3.5662E+02, 3.5700E+02,
     +  3.5739E+02, 3.5777E+02, 3.5816E+02, 3.5854E+02, 3.5893E+02,
     + 3.5931E+02, 3.5970E+02, 3.6009E+02, 3.6047E+02, 3.6086E+02,
     +  3.6124E+02, 3.6163E+02, 3.6202E+02, 3.6241E+02, 3.6279E+02,
     + 3.6318E+02, 3.6357E+02, 3.6396E+02, 3.6434E+02, 3.6473E+02,
     +  3.6512E+02, 3.6551E+02, 3.6590E+02, 3.6629E+02, 3.6668E+02,
     + 3.6707E+02, 3.6746E+02, 3.6785E+02, 3.6824E+02, 3.6863E+02,
     +  3.6902E+02, 3.6941E+02, 3.6980E+02, 3.7019E+02, 3.7058E+02,
     + 3.7098E+02, 3.7137E+02, 3.7176E+02, 3.7215E+02, 3.7254E+02,
     +  3.7294E+02, 3.7333E+02, 3.7372E+02, 3.7412E+02, 3.7451E+02,
     + 3.7490E+02, 3.7530E+02, 3.7569E+02, 3.7609E+02, 3.7648E+02,
     +  3.7688E+02, 3.7727E+02, 3.7767E+02, 3.7806E+02, 3.7846E+02,
     + 3.7885E+02, 3.7925E+02, 3.7965E+02, 3.8004E+02, 3.8044E+02,
     +  3.8084E+02, 3.8123E+02, 3.8163E+02, 3.8203E+02, 3.8242E+02,
     + 3.8282E+02, 3.8322E+02, 3.8362E+02, 3.8402E+02, 3.8442E+02,
     +  3.8482E+02, 3.8521E+02, 3.8561E+02, 3.8601E+02, 3.8641E+02,
     + 3.8681E+02, 3.8721E+02, 3.8761E+02, 3.8801E+02, 3.8842E+02,
     +  3.8882E+02, 3.8922E+02, 3.8962E+02, 3.9002E+02, 3.9042E+02,
     + 3.9082E+02, 3.9123E+02, 3.9163E+02, 3.9203E+02, 3.9244E+02,
     +  3.9284E+02, 3.9324E+02, 3.9365E+02, 3.9405E+02, 3.9445E+02,
     + 3.9486E+02, 3.9526E+02, 3.9567E+02, 3.9607E+02, 3.9648E+02,
     +  3.9688E+02, 3.9729E+02, 3.9769E+02, 3.9810E+02, 3.9850E+02,
     + 3.9891E+02, 3.9932E+02, 3.9972E+02, 4.0013E+02, 4.0054E+02,
     +  4.0095E+02, 4.0135E+02, 4.0176E+02, 4.0217E+02, 4.0258E+02,
     + 4.0299E+02, 4.0339E+02, 4.0380E+02, 4.0421E+02, 4.0462E+02,
     +  4.0503E+02, 4.0544E+02, 4.0585E+02, 4.0626E+02, 4.0667E+02,
     + 4.0708E+02, 4.0749E+02, 4.0790E+02, 4.0831E+02, 4.0873E+02,
     +  4.0914E+02, 4.0955E+02, 4.0996E+02, 4.1037E+02, 4.1079E+02,
     + 4.1120E+02, 4.1161E+02, 4.1202E+02, 4.1244E+02, 4.1285E+02,
     +  4.1327E+02, 4.1368E+02, 4.1409E+02, 4.1451E+02, 4.1492E+02,
     + 4.1534E+02, 4.1575E+02, 4.1617E+02, 4.1658E+02, 4.1700E+02,
     +  4.1741E+02, 4.1783E+02, 4.1825E+02, 4.1866E+02, 4.1908E+02,
     + 4.1950E+02, 4.1991E+02, 4.2033E+02, 4.2075E+02, 4.2117E+02,
     +  4.2158E+02, 4.2200E+02, 4.2242E+02, 4.2284E+02, 4.2326E+02,
     + 4.2368E+02, 4.2410E+02, 4.2452E+02, 4.2494E+02, 4.2536E+02,
     +  4.2578E+02, 4.2620E+02, 4.2662E+02, 4.2704E+02, 4.2746E+02,
     + 4.2788E+02, 4.2830E+02, 4.2872E+02, 4.2915E+02, 4.2957E+02,
     +  4.2999E+02, 4.3041E+02, 4.3084E+02, 4.3126E+02, 4.3168E+02,
     + 4.3211E+02, 4.3253E+02, 4.3295E+02, 4.3338E+02, 4.3380E+02,
     +  4.3423E+02, 4.3465E+02, 4.3507E+02, 4.3550E+02, 4.3593E+02,
     + 4.3635E+02, 4.3678E+02, 4.3720E+02, 4.3763E+02, 4.3806E+02,
     +  4.3848E+02, 4.3891E+02, 4.3934E+02, 4.3976E+02, 4.4019E+02,
     + 4.4062E+02, 4.4105E+02, 4.4147E+02, 4.4190E+02, 4.4233E+02,
     +  4.4276E+02, 4.4319E+02, 4.4362E+02, 4.4405E+02, 4.4448E+02,
     + 4.4491E+02, 4.4534E+02, 4.4577E+02, 4.4620E+02, 4.4663E+02,
     +  4.4706E+02, 4.4749E+02, 4.4792E+02, 4.4835E+02, 4.4879E+02,
     + 4.4922E+02, 4.4965E+02, 4.5008E+02, 4.5052E+02, 4.5095E+02,
     +  4.5138E+02, 4.5182E+02, 4.5225E+02, 4.5268E+02, 4.5312E+02,
     + 4.5355E+02, 4.5399E+02, 4.5442E+02, 4.5486E+02, 4.5529E+02,
     +  4.5573E+02, 4.5616E+02, 4.5660E+02, 4.5703E+02, 4.5747E+02,
     + 4.5791E+02, 4.5834E+02, 4.5878E+02, 4.5922E+02, 4.5965E+02,
     +  4.6009E+02, 4.6053E+02, 4.6097E+02, 4.6141E+02, 4.6184E+02,
     + 4.6228E+02, 4.6272E+02, 4.6316E+02, 4.6360E+02, 4.6404E+02,
     +  4.6448E+02, 4.6492E+02, 4.6536E+02, 4.6580E+02, 4.6624E+02,
     + 4.6668E+02, 4.6712E+02, 4.6756E+02, 4.6801E+02, 4.6845E+02,
     +  4.6889E+02, 4.6933E+02, 4.6977E+02, 4.7022E+02, 4.7066E+02,
     + 4.7110E+02, 4.7155E+02, 4.7199E+02, 4.7243E+02, 4.7288E+02,
     +  4.7332E+02, 4.7377E+02, 4.7421E+02, 4.7466E+02, 4.7510E+02,
     + 4.7555E+02, 4.7599E+02, 4.7644E+02, 4.7688E+02, 4.7733E+02,
     +  4.7778E+02, 4.7822E+02, 4.7867E+02, 4.7912E+02, 4.7956E+02,
     + 4.8001E+02, 4.8046E+02, 4.8091E+02, 4.8136E+02, 4.8180E+02,
     +  4.8225E+02, 4.8270E+02, 4.8315E+02, 4.8360E+02, 4.8405E+02,
     + 4.8450E+02, 4.8495E+02, 4.8540E+02, 4.8585E+02, 4.8630E+02,
     +  4.8675E+02, 4.8720E+02, 4.8765E+02, 4.8811E+02, 4.8856E+02,
     + 4.8901E+02, 4.8946E+02, 4.8991E+02, 4.9037E+02, 4.9082E+02,
     +  4.9127E+02, 4.9173E+02, 4.9218E+02, 4.9263E+02, 4.9309E+02,
     + 4.9354E+02, 4.9400E+02, 4.9445E+02, 4.9491E+02, 4.9536E+02,
     +  4.9582E+02, 4.9627E+02, 4.9673E+02, 4.9718E+02, 4.9764E+02,
     + 4.9810E+02, 4.9855E+02, 4.9901E+02, 4.9947E+02, 4.9992E+02,
     +  5.0038E+02, 5.0084E+02, 5.0130E+02, 5.0176E+02, 5.0221E+02,
     + 5.0267E+02, 5.0313E+02, 5.0359E+02, 5.0405E+02, 5.0451E+02,
     +  5.0497E+02, 5.0543E+02, 5.0589E+02, 5.0635E+02, 5.0681E+02,
     + 5.0727E+02, 5.0773E+02, 5.0820E+02, 5.0866E+02, 5.0912E+02,
     +  5.0958E+02, 5.1004E+02, 5.1051E+02, 5.1097E+02, 5.1143E+02,
     + 5.1189E+02, 5.1236E+02, 5.1282E+02, 5.1329E+02, 5.1375E+02,
     +  5.1421E+02, 5.1468E+02, 5.1514E+02, 5.1561E+02, 5.1607E+02,
     + 5.1654E+02, 5.1701E+02, 5.1747E+02, 5.1794E+02, 5.1840E+02,
     +  5.1887E+02, 5.1934E+02, 5.1980E+02, 5.2027E+02, 5.2074E+02,
     + 5.2121E+02, 5.2167E+02, 5.2214E+02, 5.2261E+02, 5.2308E+02,
     +  5.2355E+02, 5.2402E+02, 5.2449E+02, 5.2496E+02, 5.2543E+02,
     + 5.2590E+02, 5.2637E+02, 5.2684E+02, 5.2731E+02, 5.2778E+02,
     +  5.2825E+02, 5.2872E+02, 5.2919E+02, 5.2967E+02, 5.3014E+02,
     + 5.3061E+02, 5.3108E+02, 5.3156E+02, 5.3203E+02, 5.3250E+02,
     +  5.3297E+02, 5.3345E+02, 5.3392E+02, 5.3440E+02, 5.3487E+02,
     + 5.3535E+02, 5.3582E+02, 5.3630E+02, 5.3677E+02, 5.3725E+02,
     +  5.3772E+02, 5.3820E+02, 5.3867E+02, 5.3915E+02, 5.3963E+02,
     + 5.4010E+02, 5.4058E+02, 5.4106E+02, 5.4154E+02, 5.4201E+02,
     +  5.4249E+02, 5.4297E+02, 5.4345E+02, 5.4393E+02, 5.4441E+02,
     + 5.4489E+02, 5.4537E+02, 5.4584E+02, 5.4632E+02, 5.4680E+02,
     +  5.4729E+02, 5.4777E+02, 5.4825E+02, 5.4873E+02, 5.4921E+02,
     + 5.4969E+02, 5.5017E+02, 5.5065E+02, 5.5114E+02, 5.5162E+02,
     +  5.5210E+02, 5.5258E+02, 5.5307E+02, 5.5355E+02, 5.5403E+02,
     + 5.5452E+02, 5.5500E+02, 5.5549E+02, 5.5597E+02, 5.5645E+02,
     +  5.5694E+02, 5.5743E+02, 5.5791E+02, 5.5840E+02, 5.5888E+02,
     + 5.5937E+02, 5.5985E+02, 5.6034E+02, 5.6083E+02, 5.6131E+02,
     +  5.6180E+02, 5.6229E+02, 5.6278E+02, 5.6327E+02, 5.6375E+02,
     + 5.6424E+02, 5.6473E+02, 5.6522E+02, 5.6571E+02, 5.6620E+02,
     +  5.6669E+02, 5.6718E+02, 5.6767E+02, 5.6816E+02, 5.6865E+02,
     + 5.6914E+02, 5.6963E+02, 5.7012E+02, 5.7061E+02, 5.7110E+02,
     +  5.7160E+02, 5.7209E+02, 5.7258E+02, 5.7307E+02, 5.7357E+02,
     + 5.7406E+02, 5.7455E+02, 5.7505E+02, 5.7554E+02, 5.7603E+02,
     +  5.7653E+02, 5.7702E+02, 5.7752E+02, 5.7801E+02, 5.7851E+02,
     + 5.7900E+02, 5.7950E+02, 5.7999E+02, 5.8049E+02, 5.8099E+02,
     +  5.8148E+02, 5.8198E+02, 5.8248E+02, 5.8297E+02, 5.8347E+02,
     + 5.8397E+02, 5.8447E+02, 5.8496E+02, 5.8546E+02, 5.8596E+02,
     +  5.8646E+02, 5.8696E+02, 5.8746E+02, 5.8796E+02, 5.8846E+02,
     + 5.8896E+02, 5.8946E+02, 5.8996E+02, 5.9046E+02, 5.9096E+02,
     +  5.9146E+02, 5.9196E+02, 5.9246E+02, 5.9297E+02, 5.9347E+02,
     + 5.9397E+02, 5.9447E+02, 5.9498E+02, 5.9548E+02, 5.9598E+02,
     +  5.9649E+02, 5.9699E+02, 5.9749E+02, 5.9800E+02, 5.9850E+02,
     + 5.9901E+02, 5.9951E+02, 6.0002E+02, 6.0052E+02, 6.0103E+02,
     +  6.0153E+02, 6.0204E+02, 6.0255E+02, 6.0305E+02, 6.0356E+02,
     + 6.0407E+02, 6.0457E+02, 6.0508E+02, 6.0559E+02, 6.0610E+02,
     +  6.0661E+02, 6.0711E+02, 6.0762E+02, 6.0813E+02, 6.0864E+02,
     + 6.0915E+02, 6.0966E+02, 6.1017E+02, 6.1068E+02, 6.1119E+02,
     +  6.1170E+02, 6.1221E+02, 6.1272E+02, 6.1323E+02, 6.1375E+02,
     + 6.1426E+02, 6.1477E+02, 6.1528E+02, 6.1579E+02, 6.1631E+02,
     +  6.1682E+02, 6.1733E+02, 6.1785E+02, 6.1836E+02, 6.1887E+02,
     + 6.1939E+02, 6.1990E+02, 6.2042E+02, 6.2093E+02, 6.2145E+02,
     +  6.2196E+02, 6.2248E+02, 6.2299E+02, 6.2351E+02, 6.2403E+02,
     + 6.2454E+02, 6.2506E+02, 6.2558E+02, 6.2609E+02, 6.2661E+02,
     +  6.2713E+02, 6.2765E+02, 6.2817E+02, 6.2868E+02, 6.2920E+02,
     + 6.2972E+02, 6.3024E+02, 6.3076E+02, 6.3128E+02, 6.3180E+02,
     +  6.3232E+02, 6.3284E+02, 6.3336E+02, 6.3388E+02, 6.3440E+02,
     + 6.3492E+02, 6.3544E+02, 6.3597E+02, 6.3649E+02, 6.3701E+02,
     +  6.3753E+02, 6.3806E+02, 6.3858E+02, 6.3910E+02, 6.3962E+02,
     + 6.4015E+02, 6.4067E+02, 6.4120E+02, 6.4172E+02, 6.4225E+02,
     +  6.4277E+02, 6.4330E+02, 6.4382E+02, 6.4435E+02, 6.4487E+02,
     + 6.4540E+02, 6.4592E+02, 6.4645E+02, 6.4698E+02, 6.4750E+02,
     +  6.4803E+02, 6.4856E+02, 6.4909E+02, 6.4961E+02, 6.5014E+02,
     + 6.5067E+02, 6.5120E+02, 6.5173E+02, 6.5226E+02, 6.5279E+02,
     +  6.5332E+02, 6.5385E+02, 6.5438E+02, 6.5491E+02, 6.5544E+02,
     + 6.5597E+02, 6.5650E+02, 6.5703E+02, 6.5756E+02, 6.5809E+02,
     +  6.5863E+02, 6.5916E+02, 6.5969E+02, 6.6022E+02, 6.6076E+02,
     + 6.6129E+02, 6.6182E+02, 6.6236E+02, 6.6289E+02, 6.6342E+02,
     +  6.6396E+02, 6.6449E+02, 6.6503E+02, 6.6556E+02, 6.6610E+02,
     + 6.6663E+02, 6.6717E+02, 6.6771E+02, 6.6824E+02, 6.6878E+02,
     +  6.6932E+02, 6.6985E+02, 6.7039E+02, 6.7093E+02, 6.7146E+02,
     + 6.7200E+02, 6.7254E+02, 6.7308E+02, 6.7362E+02, 6.7416E+02,
     +  6.7470E+02, 6.7524E+02, 6.7578E+02, 6.7632E+02, 6.7686E+02,
     + 6.7740E+02, 6.7794E+02, 6.7848E+02, 6.7902E+02, 6.7956E+02,
     +  6.8010E+02, 6.8064E+02, 6.8119E+02, 6.8173E+02, 6.8227E+02,
     + 6.8281E+02, 6.8336E+02, 6.8390E+02, 6.8444E+02, 6.8499E+02,
     +  6.8553E+02, 6.8608E+02, 6.8662E+02, 6.8716E+02, 6.8771E+02,
     + 6.8825E+02, 6.8880E+02, 6.8935E+02, 6.8989E+02, 6.9044E+02,
     +  6.9098E+02, 6.9153E+02, 6.9208E+02, 6.9262E+02, 6.9317E+02,
     + 6.9372E+02, 6.9427E+02, 6.9482E+02, 6.9536E+02, 6.9591E+02,
     +  6.9646E+02, 6.9701E+02, 6.9756E+02, 6.9811E+02, 6.9866E+02,
     + 6.9921E+02, 6.9976E+02, 7.0031E+02, 7.0086E+02, 7.0141E+02,
     +  7.0196E+02, 7.0252E+02, 7.0307E+02, 7.0362E+02, 7.0417E+02,
     + 7.0472E+02, 7.0528E+02, 7.0583E+02, 7.0638E+02, 7.0694E+02,
     +  7.0749E+02, 7.0804E+02, 7.0860E+02, 7.0915E+02, 7.0971E+02,
     + 7.1026E+02, 7.1082E+02, 7.1137E+02, 7.1193E+02, 7.1248E+02,
     +  7.1304E+02, 7.1360E+02, 7.1415E+02, 7.1471E+02, 7.1527E+02,
     + 7.1582E+02, 7.1638E+02, 7.1694E+02, 7.1750E+02, 7.1806E+02,
     +  7.1861E+02, 7.1917E+02, 7.1973E+02, 7.2029E+02, 7.2085E+02,
     + 7.2141E+02, 7.2197E+02, 7.2253E+02, 7.2309E+02, 7.2365E+02,
     +  7.2421E+02, 7.2478E+02, 7.2534E+02, 7.2590E+02, 7.2646E+02,
     + 7.2702E+02, 7.2759E+02, 7.2815E+02, 7.2871E+02, 7.2927E+02,
     +  7.2984E+02, 7.3040E+02, 7.3097E+02, 7.3153E+02, 7.3209E+02,
     + 7.3266E+02, 7.3322E+02, 7.3379E+02, 7.3435E+02, 7.3492E+02,
     +  7.3549E+02, 7.3605E+02, 7.3662E+02, 7.3719E+02, 7.3775E+02,
     + 7.3832E+02, 7.3889E+02, 7.3945E+02, 7.4002E+02, 7.4059E+02,
     +  7.4116E+02, 7.4173E+02, 7.4230E+02, 7.4287E+02, 7.4343E+02,
     + 7.4400E+02, 7.4457E+02, 7.4514E+02, 7.4571E+02, 7.4628E+02,
     +  7.4686E+02, 7.4743E+02, 7.4800E+02, 7.4857E+02, 7.4914E+02,
     + 7.4971E+02, 7.5029E+02, 7.5086E+02, 7.5143E+02, 7.5200E+02,
     +  7.5258E+02, 7.5315E+02, 7.5372E+02, 7.5430E+02, 7.5487E+02,
     + 7.5545E+02, 7.5602E+02, 7.5660E+02, 7.5717E+02, 7.5775E+02,
     +  7.5832E+02, 7.5890E+02, 7.5948E+02, 7.6005E+02, 7.6063E+02,
     + 7.6121E+02, 7.6178E+02, 7.6236E+02, 7.6294E+02, 7.6352E+02,
     +  7.6409E+02, 7.6467E+02, 7.6525E+02, 7.6583E+02, 7.6641E+02,
     + 7.6699E+02, 7.6757E+02, 7.6815E+02, 7.6873E+02, 7.6931E+02,
     +  7.6989E+02, 7.7047E+02, 7.7105E+02, 7.7163E+02, 7.7222E+02,
     + 7.7280E+02, 7.7338E+02, 7.7396E+02, 7.7455E+02, 7.7513E+02,
     +  7.7571E+02, 7.7629E+02, 7.7688E+02, 7.7746E+02, 7.7805E+02,
     + 7.7863E+02, 7.7922E+02, 7.7980E+02, 7.8039E+02, 7.8097E+02,
     +  7.8156E+02, 7.8214E+02, 7.8273E+02, 7.8331E+02, 7.8390E+02,
     + 7.8449E+02, 7.8508E+02, 7.8566E+02, 7.8625E+02, 7.8684E+02,
     +  7.8743E+02, 7.8801E+02, 7.8860E+02, 7.8919E+02, 7.8978E+02,
     + 7.9037E+02, 7.9096E+02, 7.9155E+02, 7.9214E+02, 7.9273E+02,
     +  7.9332E+02, 7.9391E+02, 7.9450E+02, 7.9509E+02, 7.9569E+02,
     + 7.9628E+02, 7.9687E+02, 7.9746E+02, 7.9806E+02, 7.9865E+02,
     +  7.9924E+02, 7.9983E+02, 8.0043E+02, 8.0102E+02, 8.0162E+02,
     + 8.0221E+02, 8.0281E+02, 8.0340E+02, 8.0400E+02, 8.0459E+02,
     +  8.0519E+02, 8.0578E+02, 8.0638E+02, 8.0697E+02, 8.0757E+02,
     + 8.0817E+02, 8.0876E+02, 8.0936E+02, 8.0996E+02, 8.1056E+02,
     +  8.1116E+02, 8.1175E+02, 8.1235E+02, 8.1295E+02, 8.1355E+02,
     + 8.1415E+02, 8.1475E+02, 8.1535E+02, 8.1595E+02, 8.1655E+02,
     +  8.1715E+02, 8.1775E+02, 8.1835E+02, 8.1895E+02, 8.1956E+02,
     + 8.2016E+02, 8.2076E+02, 8.2136E+02, 8.2196E+02, 8.2257E+02,
     +  8.2317E+02, 8.2377E+02, 8.2438E+02, 8.2498E+02, 8.2559E+02,
     + 8.2619E+02, 8.2679E+02, 8.2740E+02, 8.2800E+02, 8.2861E+02,
     +  8.2921E+02, 8.2982E+02, 8.3043E+02, 8.3103E+02, 8.3164E+02,
     + 8.3225E+02, 8.3285E+02, 8.3346E+02, 8.3407E+02, 8.3468E+02,
     +  8.3528E+02, 8.3589E+02, 8.3650E+02, 8.3711E+02, 8.3772E+02,
     + 8.3833E+02, 8.3894E+02, 8.3955E+02, 8.4016E+02, 8.4077E+02,
     +  8.4138E+02, 8.4199E+02, 8.4260E+02, 8.4321E+02, 8.4382E+02,
     + 8.4444E+02, 8.4505E+02, 8.4566E+02, 8.4627E+02, 8.4689E+02,
     +  8.4750E+02, 8.4811E+02, 8.4873E+02, 8.4934E+02, 8.4996E+02,
     + 8.5057E+02, 8.5118E+02, 8.5180E+02, 8.5241E+02, 8.5303E+02,
     +  8.5365E+02, 8.5426E+02, 8.5488E+02, 8.5549E+02, 8.5611E+02,
     + 8.5673E+02, 8.5734E+02, 8.5796E+02, 8.5858E+02, 8.5920E+02,
     +  8.5982E+02, 8.6043E+02, 8.6105E+02, 8.6167E+02, 8.6229E+02,
     + 8.6291E+02, 8.6353E+02, 8.6415E+02, 8.6477E+02, 8.6539E+02,
     +  8.6601E+02, 8.6663E+02, 8.6725E+02, 8.6788E+02, 8.6850E+02,
     + 8.6912E+02, 8.6974E+02, 8.7036E+02, 8.7099E+02, 8.7161E+02,
     +  8.7223E+02, 8.7286E+02, 8.7348E+02, 8.7410E+02, 8.7473E+02,
     + 8.7535E+02, 8.7598E+02, 8.7660E+02, 8.7723E+02, 8.7785E+02,
     +  8.7848E+02, 8.7911E+02, 8.7973E+02, 8.8036E+02, 8.8099E+02,
     + 8.8161E+02, 8.8224E+02, 8.8287E+02, 8.8350E+02, 8.8412E+02,
     +  8.8475E+02, 8.8538E+02, 8.8601E+02, 8.8664E+02, 8.8727E+02,
     + 8.8790E+02, 8.8853E+02, 8.8916E+02, 8.8979E+02, 8.9042E+02,
     +  8.9105E+02, 8.9168E+02, 8.9231E+02, 8.9294E+02, 8.9357E+02,
     + 8.9421E+02, 8.9484E+02, 8.9547E+02, 8.9610E+02, 8.9674E+02,
     +  8.9737E+02, 8.9801E+02, 8.9864E+02, 8.9927E+02, 8.9991E+02,
     + 9.0054E+02, 9.0118E+02, 9.0181E+02, 9.0245E+02, 9.0308E+02,
     +  9.0372E+02, 9.0436E+02, 9.0499E+02, 9.0563E+02, 9.0627E+02,
     + 9.0690E+02, 9.0754E+02, 9.0818E+02, 9.0882E+02, 9.0945E+02,
     +  9.1009E+02, 9.1073E+02, 9.1137E+02, 9.1201E+02, 9.1265E+02,
     + 9.1329E+02, 9.1393E+02, 9.1457E+02, 9.1521E+02, 9.1585E+02,
     +  9.1649E+02, 9.1713E+02, 9.1778E+02, 9.1842E+02, 9.1906E+02,
     + 9.1970E+02, 9.2034E+02, 9.2099E+02, 9.2163E+02, 9.2227E+02,
     +  9.2292E+02, 9.2356E+02, 9.2421E+02, 9.2485E+02, 9.2549E+02,
     + 9.2614E+02, 9.2678E+02, 9.2743E+02, 9.2808E+02, 9.2872E+02,
     +  9.2937E+02, 9.3001E+02, 9.3066E+02, 9.3131E+02, 9.3196E+02,
     + 9.3260E+02, 9.3325E+02, 9.3390E+02, 9.3455E+02, 9.3519E+02,
     +  9.3584E+02, 9.3649E+02, 9.3714E+02, 9.3779E+02, 9.3844E+02,
     + 9.3909E+02, 9.3974E+02, 9.4039E+02, 9.4104E+02, 9.4169E+02,
     +  9.4235E+02, 9.4300E+02, 9.4365E+02, 9.4430E+02, 9.4495E+02,
     + 9.4561E+02, 9.4626E+02, 9.4691E+02, 9.4757E+02, 9.4822E+02,
     +  9.4887E+02, 9.4953E+02, 9.5018E+02, 9.5084E+02, 9.5149E+02,
     + 9.5215E+02, 9.5280E+02, 9.5346E+02, 9.5411E+02, 9.5477E+02,
     +  9.5543E+02, 9.5608E+02, 9.5674E+02, 9.5740E+02, 9.5805E+02,
     + 9.5871E+02, 9.5937E+02, 9.6003E+02, 9.6069E+02, 9.6135E+02,
     +  9.6200E+02, 9.6266E+02, 9.6332E+02, 9.6398E+02, 9.6464E+02,
     + 9.6530E+02, 9.6596E+02, 9.6662E+02, 9.6729E+02, 9.6795E+02,
     +  9.6861E+02, 9.6927E+02, 9.6993E+02, 9.7060E+02, 9.7126E+02,
     + 9.7192E+02, 9.7258E+02, 9.7325E+02, 9.7391E+02, 9.7458E+02,
     +  9.7524E+02, 9.7590E+02, 9.7657E+02, 9.7723E+02, 9.7790E+02,
     + 9.7856E+02, 9.7923E+02, 9.7990E+02, 9.8056E+02, 9.8123E+02,
     +  9.8190E+02, 9.8256E+02, 9.8323E+02, 9.8390E+02, 9.8457E+02,
     + 9.8523E+02, 9.8590E+02, 9.8657E+02, 9.8724E+02, 9.8791E+02,
     +  9.8858E+02, 9.8925E+02, 9.8992E+02, 9.9059E+02, 9.9126E+02,
     + 9.9193E+02, 9.9260E+02, 9.9327E+02, 9.9394E+02, 9.9461E+02,
     +  9.9529E+02, 9.9596E+02, 9.9663E+02, 9.9730E+02, 9.9798E+02/
      data (QofT( 1,J),J=2409,4510)/ 9.9865E+02, 9.9932E+02,
     +  1.0000E+03, 1.0007E+03, 1.0013E+03, 1.0020E+03, 1.0027E+03,
     +  1.0034E+03,
     + 1.0040E+03, 1.0047E+03, 1.0054E+03, 1.0061E+03, 1.0067E+03,
     +  1.0074E+03, 1.0081E+03, 1.0088E+03, 1.0095E+03, 1.0101E+03,
     + 1.0108E+03, 1.0115E+03, 1.0122E+03, 1.0128E+03, 1.0135E+03,
     +  1.0142E+03, 1.0149E+03, 1.0156E+03, 1.0162E+03, 1.0169E+03,
     + 1.0176E+03, 1.0183E+03, 1.0190E+03, 1.0196E+03, 1.0203E+03,
     +  1.0210E+03, 1.0217E+03, 1.0224E+03, 1.0231E+03, 1.0237E+03,
     + 1.0244E+03, 1.0251E+03, 1.0258E+03, 1.0265E+03, 1.0272E+03,
     +  1.0278E+03, 1.0285E+03, 1.0292E+03, 1.0299E+03, 1.0306E+03,
     + 1.0313E+03, 1.0320E+03, 1.0326E+03, 1.0333E+03, 1.0340E+03,
     +  1.0347E+03, 1.0354E+03, 1.0361E+03, 1.0368E+03, 1.0375E+03,
     + 1.0381E+03, 1.0388E+03, 1.0395E+03, 1.0402E+03, 1.0409E+03,
     +  1.0416E+03, 1.0423E+03, 1.0430E+03, 1.0437E+03, 1.0443E+03,
     + 1.0450E+03, 1.0457E+03, 1.0464E+03, 1.0471E+03, 1.0478E+03,
     +  1.0485E+03, 1.0492E+03, 1.0499E+03, 1.0506E+03, 1.0513E+03,
     + 1.0520E+03, 1.0527E+03, 1.0534E+03, 1.0540E+03, 1.0547E+03,
     +  1.0554E+03, 1.0561E+03, 1.0568E+03, 1.0575E+03, 1.0582E+03,
     + 1.0589E+03, 1.0596E+03, 1.0603E+03, 1.0610E+03, 1.0617E+03,
     +  1.0624E+03, 1.0631E+03, 1.0638E+03, 1.0645E+03, 1.0652E+03,
     + 1.0659E+03, 1.0666E+03, 1.0673E+03, 1.0680E+03, 1.0687E+03,
     +  1.0694E+03, 1.0701E+03, 1.0708E+03, 1.0715E+03, 1.0722E+03,
     + 1.0729E+03, 1.0736E+03, 1.0743E+03, 1.0750E+03, 1.0757E+03,
     +  1.0764E+03, 1.0771E+03, 1.0778E+03, 1.0785E+03, 1.0792E+03,
     + 1.0799E+03, 1.0806E+03, 1.0813E+03, 1.0820E+03, 1.0827E+03,
     +  1.0834E+03, 1.0842E+03, 1.0849E+03, 1.0856E+03, 1.0863E+03,
     + 1.0870E+03, 1.0877E+03, 1.0884E+03, 1.0891E+03, 1.0898E+03,
     +  1.0905E+03, 1.0912E+03, 1.0919E+03, 1.0926E+03, 1.0933E+03,
     + 1.0941E+03, 1.0948E+03, 1.0955E+03, 1.0962E+03, 1.0969E+03,
     +  1.0976E+03, 1.0983E+03, 1.0990E+03, 1.0997E+03, 1.1005E+03,
     + 1.1012E+03, 1.1019E+03, 1.1026E+03, 1.1033E+03, 1.1040E+03,
     +  1.1047E+03, 1.1054E+03, 1.1062E+03, 1.1069E+03, 1.1076E+03,
     + 1.1083E+03, 1.1090E+03, 1.1097E+03, 1.1104E+03, 1.1112E+03,
     +  1.1119E+03, 1.1126E+03, 1.1133E+03, 1.1140E+03, 1.1147E+03,
     + 1.1155E+03, 1.1162E+03, 1.1169E+03, 1.1176E+03, 1.1183E+03,
     +  1.1190E+03, 1.1198E+03, 1.1205E+03, 1.1212E+03, 1.1219E+03,
     + 1.1226E+03, 1.1234E+03, 1.1241E+03, 1.1248E+03, 1.1255E+03,
     +  1.1262E+03, 1.1270E+03, 1.1277E+03, 1.1284E+03, 1.1291E+03,
     + 1.1299E+03, 1.1306E+03, 1.1313E+03, 1.1320E+03, 1.1327E+03,
     +  1.1335E+03, 1.1342E+03, 1.1349E+03, 1.1356E+03, 1.1364E+03,
     + 1.1371E+03, 1.1378E+03, 1.1385E+03, 1.1393E+03, 1.1400E+03,
     +  1.1407E+03, 1.1415E+03, 1.1422E+03, 1.1429E+03, 1.1436E+03,
     + 1.1444E+03, 1.1451E+03, 1.1458E+03, 1.1465E+03, 1.1473E+03,
     +  1.1480E+03, 1.1487E+03, 1.1495E+03, 1.1502E+03, 1.1509E+03,
     + 1.1517E+03, 1.1524E+03, 1.1531E+03, 1.1538E+03, 1.1546E+03,
     +  1.1553E+03, 1.1560E+03, 1.1568E+03, 1.1575E+03, 1.1582E+03,
     + 1.1590E+03, 1.1597E+03, 1.1604E+03, 1.1612E+03, 1.1619E+03,
     +  1.1626E+03, 1.1634E+03, 1.1641E+03, 1.1648E+03, 1.1656E+03,
     + 1.1663E+03, 1.1671E+03, 1.1678E+03, 1.1685E+03, 1.1693E+03,
     +  1.1700E+03, 1.1707E+03, 1.1715E+03, 1.1722E+03, 1.1729E+03,
     + 1.1737E+03, 1.1744E+03, 1.1752E+03, 1.1759E+03, 1.1766E+03,
     +  1.1774E+03, 1.1781E+03, 1.1789E+03, 1.1796E+03, 1.1803E+03,
     + 1.1811E+03, 1.1818E+03, 1.1826E+03, 1.1833E+03, 1.1841E+03,
     +  1.1848E+03, 1.1855E+03, 1.1863E+03, 1.1870E+03, 1.1878E+03,
     + 1.1885E+03, 1.1893E+03, 1.1900E+03, 1.1907E+03, 1.1915E+03,
     +  1.1922E+03, 1.1930E+03, 1.1937E+03, 1.1945E+03, 1.1952E+03,
     + 1.1960E+03, 1.1967E+03, 1.1975E+03, 1.1982E+03, 1.1990E+03,
     +  1.1997E+03, 1.2004E+03, 1.2012E+03, 1.2019E+03, 1.2027E+03,
     + 1.2034E+03, 1.2042E+03, 1.2049E+03, 1.2057E+03, 1.2064E+03,
     +  1.2072E+03, 1.2079E+03, 1.2087E+03, 1.2094E+03, 1.2102E+03,
     + 1.2109E+03, 1.2117E+03, 1.2124E+03, 1.2132E+03, 1.2140E+03,
     +  1.2147E+03, 1.2155E+03, 1.2162E+03, 1.2170E+03, 1.2177E+03,
     + 1.2185E+03, 1.2192E+03, 1.2200E+03, 1.2207E+03, 1.2215E+03,
     +  1.2223E+03, 1.2230E+03, 1.2238E+03, 1.2245E+03, 1.2253E+03,
     + 1.2260E+03, 1.2268E+03, 1.2275E+03, 1.2283E+03, 1.2291E+03,
     +  1.2298E+03, 1.2306E+03, 1.2313E+03, 1.2321E+03, 1.2329E+03,
     + 1.2336E+03, 1.2344E+03, 1.2351E+03, 1.2359E+03, 1.2367E+03,
     +  1.2374E+03, 1.2382E+03, 1.2389E+03, 1.2397E+03, 1.2405E+03,
     + 1.2412E+03, 1.2420E+03, 1.2428E+03, 1.2435E+03, 1.2443E+03,
     +  1.2450E+03, 1.2458E+03, 1.2466E+03, 1.2473E+03, 1.2481E+03,
     + 1.2489E+03, 1.2496E+03, 1.2504E+03, 1.2512E+03, 1.2519E+03,
     +  1.2527E+03, 1.2535E+03, 1.2542E+03, 1.2550E+03, 1.2558E+03,
     + 1.2565E+03, 1.2573E+03, 1.2581E+03, 1.2588E+03, 1.2596E+03,
     +  1.2604E+03, 1.2611E+03, 1.2619E+03, 1.2627E+03, 1.2635E+03,
     + 1.2642E+03, 1.2650E+03, 1.2658E+03, 1.2665E+03, 1.2673E+03,
     +  1.2681E+03, 1.2688E+03, 1.2696E+03, 1.2704E+03, 1.2712E+03,
     + 1.2719E+03, 1.2727E+03, 1.2735E+03, 1.2743E+03, 1.2750E+03,
     +  1.2758E+03, 1.2766E+03, 1.2774E+03, 1.2781E+03, 1.2789E+03,
     + 1.2797E+03, 1.2805E+03, 1.2812E+03, 1.2820E+03, 1.2828E+03,
     +  1.2836E+03, 1.2843E+03, 1.2851E+03, 1.2859E+03, 1.2867E+03,
     + 1.2875E+03, 1.2882E+03, 1.2890E+03, 1.2898E+03, 1.2906E+03,
     +  1.2914E+03, 1.2921E+03, 1.2929E+03, 1.2937E+03, 1.2945E+03,
     + 1.2953E+03, 1.2960E+03, 1.2968E+03, 1.2976E+03, 1.2984E+03,
     +  1.2992E+03, 1.2999E+03, 1.3007E+03, 1.3015E+03, 1.3023E+03,
     + 1.3031E+03, 1.3039E+03, 1.3046E+03, 1.3054E+03, 1.3062E+03,
     +  1.3070E+03, 1.3078E+03, 1.3086E+03, 1.3094E+03, 1.3101E+03,
     + 1.3109E+03, 1.3117E+03, 1.3125E+03, 1.3133E+03, 1.3141E+03,
     +  1.3149E+03, 1.3157E+03, 1.3164E+03, 1.3172E+03, 1.3180E+03,
     + 1.3188E+03, 1.3196E+03, 1.3204E+03, 1.3212E+03, 1.3220E+03,
     +  1.3228E+03, 1.3236E+03, 1.3243E+03, 1.3251E+03, 1.3259E+03,
     + 1.3267E+03, 1.3275E+03, 1.3283E+03, 1.3291E+03, 1.3299E+03,
     +  1.3307E+03, 1.3315E+03, 1.3323E+03, 1.3331E+03, 1.3339E+03,
     + 1.3346E+03, 1.3354E+03, 1.3362E+03, 1.3370E+03, 1.3378E+03,
     +  1.3386E+03, 1.3394E+03, 1.3402E+03, 1.3410E+03, 1.3418E+03,
     + 1.3426E+03, 1.3434E+03, 1.3442E+03, 1.3450E+03, 1.3458E+03,
     +  1.3466E+03, 1.3474E+03, 1.3482E+03, 1.3490E+03, 1.3498E+03,
     + 1.3506E+03, 1.3514E+03, 1.3522E+03, 1.3530E+03, 1.3538E+03,
     +  1.3546E+03, 1.3554E+03, 1.3562E+03, 1.3570E+03, 1.3578E+03,
     + 1.3586E+03, 1.3594E+03, 1.3602E+03, 1.3610E+03, 1.3618E+03,
     +  1.3626E+03, 1.3634E+03, 1.3642E+03, 1.3650E+03, 1.3658E+03,
     + 1.3667E+03, 1.3675E+03, 1.3683E+03, 1.3691E+03, 1.3699E+03,
     +  1.3707E+03, 1.3715E+03, 1.3723E+03, 1.3731E+03, 1.3739E+03,
     + 1.3747E+03, 1.3755E+03, 1.3763E+03, 1.3771E+03, 1.3780E+03,
     +  1.3788E+03, 1.3796E+03, 1.3804E+03, 1.3812E+03, 1.3820E+03,
     + 1.3828E+03, 1.3836E+03, 1.3844E+03, 1.3852E+03, 1.3861E+03,
     +  1.3869E+03, 1.3877E+03, 1.3885E+03, 1.3893E+03, 1.3901E+03,
     + 1.3909E+03, 1.3917E+03, 1.3926E+03, 1.3934E+03, 1.3942E+03,
     +  1.3950E+03, 1.3958E+03, 1.3966E+03, 1.3974E+03, 1.3983E+03,
     + 1.3991E+03, 1.3999E+03, 1.4007E+03, 1.4015E+03, 1.4023E+03,
     +  1.4032E+03, 1.4040E+03, 1.4048E+03, 1.4056E+03, 1.4064E+03,
     + 1.4073E+03, 1.4081E+03, 1.4089E+03, 1.4097E+03, 1.4105E+03,
     +  1.4114E+03, 1.4122E+03, 1.4130E+03, 1.4138E+03, 1.4146E+03,
     + 1.4155E+03, 1.4163E+03, 1.4171E+03, 1.4179E+03, 1.4187E+03,
     +  1.4196E+03, 1.4204E+03, 1.4212E+03, 1.4220E+03, 1.4229E+03,
     + 1.4237E+03, 1.4245E+03, 1.4253E+03, 1.4262E+03, 1.4270E+03,
     +  1.4278E+03, 1.4286E+03, 1.4295E+03, 1.4303E+03, 1.4311E+03,
     + 1.4319E+03, 1.4328E+03, 1.4336E+03, 1.4344E+03, 1.4353E+03,
     +  1.4361E+03, 1.4369E+03, 1.4377E+03, 1.4386E+03, 1.4394E+03,
     + 1.4402E+03, 1.4411E+03, 1.4419E+03, 1.4427E+03, 1.4436E+03,
     +  1.4444E+03, 1.4452E+03, 1.4460E+03, 1.4469E+03, 1.4477E+03,
     + 1.4485E+03, 1.4494E+03, 1.4502E+03, 1.4510E+03, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        29
      data (QofT( 2,J),J=1,2408)/ 6.0000E+00, 6.0000E+00, 6.0005E+00,
     +  6.0073E+00, 6.0348E+00, 6.0985E+00, 6.2073E+00, 6.3624E+00,
     + 6.5599E+00, 6.7936E+00, 7.0569E+00, 7.3438E+00, 7.6491E+00,
     +  7.9689E+00, 8.2998E+00, 8.6394E+00, 8.9858E+00, 9.3377E+00,
     + 9.6939E+00, 1.0054E+01, 1.0781E+01, 1.1517E+01, 1.2259E+01,
     +  1.3005E+01, 1.3754E+01, 1.4506E+01, 1.5261E+01, 1.6017E+01,
     + 1.6774E+01, 1.7533E+01, 1.8293E+01, 1.9053E+01, 1.9814E+01,
     +  2.0576E+01, 2.1339E+01, 2.2102E+01, 2.2866E+01, 2.3629E+01,
     + 2.4394E+01, 2.5158E+01, 2.5923E+01, 2.6688E+01, 2.7454E+01,
     +  2.8219E+01, 2.8985E+01, 2.9751E+01, 3.0517E+01, 3.1284E+01,
     + 3.2050E+01, 3.2817E+01, 3.3584E+01, 3.4351E+01, 3.5118E+01,
     +  3.5885E+01, 3.6652E+01, 3.7419E+01, 3.8187E+01, 3.8954E+01,
     + 3.9722E+01, 4.0490E+01, 4.1257E+01, 4.2025E+01, 4.2793E+01,
     +  4.3561E+01, 4.4329E+01, 4.5097E+01, 4.5865E+01, 4.6633E+01,
     + 4.7402E+01, 4.8170E+01, 4.8938E+01, 4.9707E+01, 5.0475E+01,
     +  5.1244E+01, 5.2012E+01, 5.2781E+01, 5.3549E+01, 5.4318E+01,
     + 5.5087E+01, 5.5856E+01, 5.6625E+01, 5.7393E+01, 5.8162E+01,
     +  5.8931E+01, 5.9700E+01, 6.0469E+01, 6.1238E+01, 6.2007E+01,
     + 6.2776E+01, 6.3546E+01, 6.4315E+01, 6.5084E+01, 6.5853E+01,
     +  6.6623E+01, 6.7392E+01, 6.8161E+01, 6.8931E+01, 6.9700E+01,
     + 7.0469E+01, 7.1239E+01, 7.2008E+01, 7.2778E+01, 7.3548E+01,
     +  7.4317E+01, 7.5087E+01, 7.5856E+01, 7.6626E+01, 7.7396E+01,
     + 7.8166E+01, 7.8935E+01, 7.9705E+01, 8.0475E+01, 8.1245E+01,
     +  8.2015E+01, 8.2785E+01, 8.3554E+01, 8.4324E+01, 8.5094E+01,
     + 8.5864E+01, 8.6634E+01, 8.7405E+01, 8.8175E+01, 8.8945E+01,
     +  8.9715E+01, 9.0485E+01, 9.1255E+01, 9.2026E+01, 9.2796E+01,
     + 9.3566E+01, 9.4336E+01, 9.5107E+01, 9.5877E+01, 9.6647E+01,
     +  9.7418E+01, 9.8188E+01, 9.8959E+01, 9.9729E+01, 1.0050E+02,
     + 1.0127E+02, 1.0204E+02, 1.0281E+02, 1.0358E+02, 1.0435E+02,
     +  1.0512E+02, 1.0589E+02, 1.0666E+02, 1.0744E+02, 1.0821E+02,
     + 1.0898E+02, 1.0975E+02, 1.1052E+02, 1.1129E+02, 1.1206E+02,
     +  1.1283E+02, 1.1360E+02, 1.1437E+02, 1.1514E+02, 1.1591E+02,
     + 1.1669E+02, 1.1746E+02, 1.1823E+02, 1.1900E+02, 1.1977E+02,
     +  1.2054E+02, 1.2131E+02, 1.2208E+02, 1.2285E+02, 1.2363E+02,
     + 1.2440E+02, 1.2517E+02, 1.2594E+02, 1.2671E+02, 1.2748E+02,
     +  1.2825E+02, 1.2903E+02, 1.2980E+02, 1.3057E+02, 1.3134E+02,
     + 1.3211E+02, 1.3288E+02, 1.3365E+02, 1.3443E+02, 1.3520E+02,
     +  1.3597E+02, 1.3674E+02, 1.3751E+02, 1.3828E+02, 1.3906E+02,
     + 1.3983E+02, 1.4060E+02, 1.4137E+02, 1.4214E+02, 1.4292E+02,
     +  1.4369E+02, 1.4446E+02, 1.4523E+02, 1.4600E+02, 1.4677E+02,
     + 1.4755E+02, 1.4832E+02, 1.4909E+02, 1.4986E+02, 1.5064E+02,
     +  1.5141E+02, 1.5218E+02, 1.5295E+02, 1.5372E+02, 1.5450E+02,
     + 1.5527E+02, 1.5604E+02, 1.5681E+02, 1.5759E+02, 1.5836E+02,
     +  1.5913E+02, 1.5990E+02, 1.6068E+02, 1.6145E+02, 1.6222E+02,
     + 1.6300E+02, 1.6377E+02, 1.6454E+02, 1.6531E+02, 1.6609E+02,
     +  1.6686E+02, 1.6763E+02, 1.6841E+02, 1.6918E+02, 1.6995E+02,
     + 1.7073E+02, 1.7150E+02, 1.7227E+02, 1.7305E+02, 1.7382E+02,
     +  1.7459E+02, 1.7537E+02, 1.7614E+02, 1.7691E+02, 1.7769E+02,
     + 1.7846E+02, 1.7923E+02, 1.8001E+02, 1.8078E+02, 1.8155E+02,
     +  1.8233E+02, 1.8310E+02, 1.8388E+02, 1.8465E+02, 1.8543E+02,
     + 1.8620E+02, 1.8697E+02, 1.8775E+02, 1.8852E+02, 1.8930E+02,
     +  1.9007E+02, 1.9085E+02, 1.9162E+02, 1.9240E+02, 1.9317E+02,
     + 1.9395E+02, 1.9472E+02, 1.9550E+02, 1.9627E+02, 1.9705E+02,
     +  1.9782E+02, 1.9860E+02, 1.9937E+02, 2.0015E+02, 2.0092E+02,
     + 2.0170E+02, 2.0247E+02, 2.0325E+02, 2.0403E+02, 2.0480E+02,
     +  2.0558E+02, 2.0636E+02, 2.0713E+02, 2.0791E+02, 2.0868E+02,
     + 2.0946E+02, 2.1024E+02, 2.1101E+02, 2.1179E+02, 2.1257E+02,
     +  2.1335E+02, 2.1412E+02, 2.1490E+02, 2.1568E+02, 2.1646E+02,
     + 2.1723E+02, 2.1801E+02, 2.1879E+02, 2.1957E+02, 2.2034E+02,
     +  2.2112E+02, 2.2190E+02, 2.2268E+02, 2.2346E+02, 2.2424E+02,
     + 2.2502E+02, 2.2579E+02, 2.2657E+02, 2.2735E+02, 2.2813E+02,
     +  2.2891E+02, 2.2969E+02, 2.3047E+02, 2.3125E+02, 2.3203E+02,
     + 2.3281E+02, 2.3359E+02, 2.3437E+02, 2.3515E+02, 2.3593E+02,
     +  2.3671E+02, 2.3750E+02, 2.3828E+02, 2.3906E+02, 2.3984E+02,
     + 2.4062E+02, 2.4140E+02, 2.4219E+02, 2.4297E+02, 2.4375E+02,
     +  2.4453E+02, 2.4531E+02, 2.4610E+02, 2.4688E+02, 2.4766E+02,
     + 2.4845E+02, 2.4923E+02, 2.5001E+02, 2.5080E+02, 2.5158E+02,
     +  2.5237E+02, 2.5315E+02, 2.5394E+02, 2.5472E+02, 2.5551E+02,
     + 2.5629E+02, 2.5708E+02, 2.5786E+02, 2.5865E+02, 2.5943E+02,
     +  2.6022E+02, 2.6101E+02, 2.6179E+02, 2.6258E+02, 2.6337E+02,
     + 2.6415E+02, 2.6494E+02, 2.6573E+02, 2.6652E+02, 2.6730E+02,
     +  2.6809E+02, 2.6888E+02, 2.6967E+02, 2.7046E+02, 2.7125E+02,
     + 2.7204E+02, 2.7283E+02, 2.7362E+02, 2.7441E+02, 2.7520E+02,
     +  2.7599E+02, 2.7678E+02, 2.7757E+02, 2.7836E+02, 2.7915E+02,
     + 2.7994E+02, 2.8074E+02, 2.8153E+02, 2.8232E+02, 2.8311E+02,
     +  2.8391E+02, 2.8470E+02, 2.8549E+02, 2.8629E+02, 2.8708E+02,
     + 2.8788E+02, 2.8867E+02, 2.8947E+02, 2.9026E+02, 2.9106E+02,
     +  2.9185E+02, 2.9265E+02, 2.9344E+02, 2.9424E+02, 2.9504E+02,
     + 2.9583E+02, 2.9663E+02, 2.9743E+02, 2.9823E+02, 2.9903E+02,
     +  2.9982E+02, 3.0062E+02, 3.0142E+02, 3.0222E+02, 3.0302E+02,
     + 3.0382E+02, 3.0462E+02, 3.0542E+02, 3.0622E+02, 3.0702E+02,
     +  3.0783E+02, 3.0863E+02, 3.0943E+02, 3.1023E+02, 3.1104E+02,
     + 3.1184E+02, 3.1264E+02, 3.1345E+02, 3.1425E+02, 3.1505E+02,
     +  3.1586E+02, 3.1666E+02, 3.1747E+02, 3.1828E+02, 3.1908E+02,
     + 3.1989E+02, 3.2069E+02, 3.2150E+02, 3.2231E+02, 3.2312E+02,
     +  3.2392E+02, 3.2473E+02, 3.2554E+02, 3.2635E+02, 3.2716E+02,
     + 3.2797E+02, 3.2878E+02, 3.2959E+02, 3.3040E+02, 3.3121E+02,
     +  3.3203E+02, 3.3284E+02, 3.3365E+02, 3.3446E+02, 3.3528E+02,
     + 3.3609E+02, 3.3690E+02, 3.3772E+02, 3.3853E+02, 3.3935E+02,
     +  3.4016E+02, 3.4098E+02, 3.4179E+02, 3.4261E+02, 3.4343E+02,
     + 3.4425E+02, 3.4506E+02, 3.4588E+02, 3.4670E+02, 3.4752E+02,
     +  3.4834E+02, 3.4916E+02, 3.4998E+02, 3.5080E+02, 3.5162E+02,
     + 3.5244E+02, 3.5326E+02, 3.5408E+02, 3.5491E+02, 3.5573E+02,
     +  3.5655E+02, 3.5738E+02, 3.5820E+02, 3.5903E+02, 3.5985E+02,
     + 3.6068E+02, 3.6150E+02, 3.6233E+02, 3.6315E+02, 3.6398E+02,
     +  3.6481E+02, 3.6564E+02, 3.6647E+02, 3.6729E+02, 3.6812E+02,
     + 3.6895E+02, 3.6978E+02, 3.7061E+02, 3.7145E+02, 3.7228E+02,
     +  3.7311E+02, 3.7394E+02, 3.7477E+02, 3.7561E+02, 3.7644E+02,
     + 3.7727E+02, 3.7811E+02, 3.7894E+02, 3.7978E+02, 3.8062E+02,
     +  3.8145E+02, 3.8229E+02, 3.8313E+02, 3.8396E+02, 3.8480E+02,
     + 3.8564E+02, 3.8648E+02, 3.8732E+02, 3.8816E+02, 3.8900E+02,
     +  3.8984E+02, 3.9068E+02, 3.9153E+02, 3.9237E+02, 3.9321E+02,
     + 3.9405E+02, 3.9490E+02, 3.9574E+02, 3.9659E+02, 3.9743E+02,
     +  3.9828E+02, 3.9912E+02, 3.9997E+02, 4.0082E+02, 4.0167E+02,
     + 4.0251E+02, 4.0336E+02, 4.0421E+02, 4.0506E+02, 4.0591E+02,
     +  4.0676E+02, 4.0761E+02, 4.0847E+02, 4.0932E+02, 4.1017E+02,
     + 4.1102E+02, 4.1188E+02, 4.1273E+02, 4.1359E+02, 4.1444E+02,
     +  4.1530E+02, 4.1616E+02, 4.1701E+02, 4.1787E+02, 4.1873E+02,
     + 4.1959E+02, 4.2044E+02, 4.2130E+02, 4.2216E+02, 4.2302E+02,
     +  4.2389E+02, 4.2475E+02, 4.2561E+02, 4.2647E+02, 4.2733E+02,
     + 4.2820E+02, 4.2906E+02, 4.2993E+02, 4.3079E+02, 4.3166E+02,
     +  4.3252E+02, 4.3339E+02, 4.3426E+02, 4.3513E+02, 4.3599E+02,
     + 4.3686E+02, 4.3773E+02, 4.3860E+02, 4.3947E+02, 4.4035E+02,
     +  4.4122E+02, 4.4209E+02, 4.4296E+02, 4.4384E+02, 4.4471E+02,
     + 4.4558E+02, 4.4646E+02, 4.4733E+02, 4.4821E+02, 4.4909E+02,
     +  4.4996E+02, 4.5084E+02, 4.5172E+02, 4.5260E+02, 4.5348E+02,
     + 4.5436E+02, 4.5524E+02, 4.5612E+02, 4.5700E+02, 4.5788E+02,
     +  4.5877E+02, 4.5965E+02, 4.6053E+02, 4.6142E+02, 4.6230E+02,
     + 4.6319E+02, 4.6408E+02, 4.6496E+02, 4.6585E+02, 4.6674E+02,
     +  4.6763E+02, 4.6852E+02, 4.6941E+02, 4.7030E+02, 4.7119E+02,
     + 4.7208E+02, 4.7297E+02, 4.7386E+02, 4.7476E+02, 4.7565E+02,
     +  4.7655E+02, 4.7744E+02, 4.7834E+02, 4.7923E+02, 4.8013E+02,
     + 4.8103E+02, 4.8192E+02, 4.8282E+02, 4.8372E+02, 4.8462E+02,
     +  4.8552E+02, 4.8642E+02, 4.8732E+02, 4.8823E+02, 4.8913E+02,
     + 4.9003E+02, 4.9094E+02, 4.9184E+02, 4.9275E+02, 4.9365E+02,
     +  4.9456E+02, 4.9547E+02, 4.9637E+02, 4.9728E+02, 4.9819E+02,
     + 4.9910E+02, 5.0001E+02, 5.0092E+02, 5.0183E+02, 5.0274E+02,
     +  5.0366E+02, 5.0457E+02, 5.0548E+02, 5.0640E+02, 5.0731E+02,
     + 5.0823E+02, 5.0914E+02, 5.1006E+02, 5.1098E+02, 5.1190E+02,
     +  5.1282E+02, 5.1373E+02, 5.1465E+02, 5.1558E+02, 5.1650E+02,
     + 5.1742E+02, 5.1834E+02, 5.1926E+02, 5.2019E+02, 5.2111E+02,
     +  5.2204E+02, 5.2296E+02, 5.2389E+02, 5.2482E+02, 5.2574E+02,
     + 5.2667E+02, 5.2760E+02, 5.2853E+02, 5.2946E+02, 5.3039E+02,
     +  5.3132E+02, 5.3225E+02, 5.3319E+02, 5.3412E+02, 5.3505E+02,
     + 5.3599E+02, 5.3692E+02, 5.3786E+02, 5.3880E+02, 5.3973E+02,
     +  5.4067E+02, 5.4161E+02, 5.4255E+02, 5.4349E+02, 5.4443E+02,
     + 5.4537E+02, 5.4631E+02, 5.4725E+02, 5.4820E+02, 5.4914E+02,
     +  5.5008E+02, 5.5103E+02, 5.5197E+02, 5.5292E+02, 5.5387E+02,
     + 5.5482E+02, 5.5576E+02, 5.5671E+02, 5.5766E+02, 5.5861E+02,
     +  5.5956E+02, 5.6051E+02, 5.6147E+02, 5.6242E+02, 5.6337E+02,
     + 5.6433E+02, 5.6528E+02, 5.6624E+02, 5.6719E+02, 5.6815E+02,
     +  5.6911E+02, 5.7007E+02, 5.7102E+02, 5.7198E+02, 5.7294E+02,
     + 5.7391E+02, 5.7487E+02, 5.7583E+02, 5.7679E+02, 5.7776E+02,
     +  5.7872E+02, 5.7968E+02, 5.8065E+02, 5.8162E+02, 5.8258E+02,
     + 5.8355E+02, 5.8452E+02, 5.8549E+02, 5.8646E+02, 5.8743E+02,
     +  5.8840E+02, 5.8937E+02, 5.9034E+02, 5.9132E+02, 5.9229E+02,
     + 5.9326E+02, 5.9424E+02, 5.9522E+02, 5.9619E+02, 5.9717E+02,
     +  5.9815E+02, 5.9913E+02, 6.0010E+02, 6.0108E+02, 6.0207E+02,
     + 6.0305E+02, 6.0403E+02, 6.0501E+02, 6.0599E+02, 6.0698E+02,
     +  6.0796E+02, 6.0895E+02, 6.0993E+02, 6.1092E+02, 6.1191E+02,
     + 6.1290E+02, 6.1389E+02, 6.1488E+02, 6.1587E+02, 6.1686E+02,
     +  6.1785E+02, 6.1884E+02, 6.1983E+02, 6.2083E+02, 6.2182E+02,
     + 6.2282E+02, 6.2381E+02, 6.2481E+02, 6.2581E+02, 6.2681E+02,
     +  6.2780E+02, 6.2880E+02, 6.2980E+02, 6.3080E+02, 6.3181E+02,
     + 6.3281E+02, 6.3381E+02, 6.3481E+02, 6.3582E+02, 6.3682E+02,
     +  6.3783E+02, 6.3884E+02, 6.3984E+02, 6.4085E+02, 6.4186E+02,
     + 6.4287E+02, 6.4388E+02, 6.4489E+02, 6.4590E+02, 6.4691E+02,
     +  6.4793E+02, 6.4894E+02, 6.4995E+02, 6.5097E+02, 6.5198E+02,
     + 6.5300E+02, 6.5402E+02, 6.5504E+02, 6.5605E+02, 6.5707E+02,
     +  6.5809E+02, 6.5911E+02, 6.6014E+02, 6.6116E+02, 6.6218E+02,
     + 6.6320E+02, 6.6423E+02, 6.6525E+02, 6.6628E+02, 6.6731E+02,
     +  6.6833E+02, 6.6936E+02, 6.7039E+02, 6.7142E+02, 6.7245E+02,
     + 6.7348E+02, 6.7451E+02, 6.7554E+02, 6.7658E+02, 6.7761E+02,
     +  6.7865E+02, 6.7968E+02, 6.8072E+02, 6.8175E+02, 6.8279E+02,
     + 6.8383E+02, 6.8487E+02, 6.8591E+02, 6.8695E+02, 6.8799E+02,
     +  6.8903E+02, 6.9007E+02, 6.9112E+02, 6.9216E+02, 6.9320E+02,
     + 6.9425E+02, 6.9530E+02, 6.9634E+02, 6.9739E+02, 6.9844E+02,
     +  6.9949E+02, 7.0054E+02, 7.0159E+02, 7.0264E+02, 7.0369E+02,
     + 7.0474E+02, 7.0580E+02, 7.0685E+02, 7.0791E+02, 7.0896E+02,
     +  7.1002E+02, 7.1108E+02, 7.1213E+02, 7.1319E+02, 7.1425E+02,
     + 7.1531E+02, 7.1637E+02, 7.1744E+02, 7.1850E+02, 7.1956E+02,
     +  7.2062E+02, 7.2169E+02, 7.2275E+02, 7.2382E+02, 7.2489E+02,
     + 7.2596E+02, 7.2702E+02, 7.2809E+02, 7.2916E+02, 7.3023E+02,
     +  7.3130E+02, 7.3238E+02, 7.3345E+02, 7.3452E+02, 7.3560E+02,
     + 7.3667E+02, 7.3775E+02, 7.3882E+02, 7.3990E+02, 7.4098E+02,
     +  7.4206E+02, 7.4314E+02, 7.4422E+02, 7.4530E+02, 7.4638E+02,
     + 7.4746E+02, 7.4855E+02, 7.4963E+02, 7.5072E+02, 7.5180E+02,
     +  7.5289E+02, 7.5398E+02, 7.5506E+02, 7.5615E+02, 7.5724E+02,
     + 7.5833E+02, 7.5942E+02, 7.6051E+02, 7.6161E+02, 7.6270E+02,
     +  7.6379E+02, 7.6489E+02, 7.6598E+02, 7.6708E+02, 7.6818E+02,
     + 7.6928E+02, 7.7037E+02, 7.7147E+02, 7.7257E+02, 7.7367E+02,
     +  7.7478E+02, 7.7588E+02, 7.7698E+02, 7.7808E+02, 7.7919E+02,
     + 7.8029E+02, 7.8140E+02, 7.8251E+02, 7.8361E+02, 7.8472E+02,
     +  7.8583E+02, 7.8694E+02, 7.8805E+02, 7.8916E+02, 7.9028E+02,
     + 7.9139E+02, 7.9250E+02, 7.9362E+02, 7.9473E+02, 7.9585E+02,
     +  7.9697E+02, 7.9808E+02, 7.9920E+02, 8.0032E+02, 8.0144E+02,
     + 8.0256E+02, 8.0368E+02, 8.0481E+02, 8.0593E+02, 8.0705E+02,
     +  8.0818E+02, 8.0930E+02, 8.1043E+02, 8.1156E+02, 8.1268E+02,
     + 8.1381E+02, 8.1494E+02, 8.1607E+02, 8.1720E+02, 8.1833E+02,
     +  8.1947E+02, 8.2060E+02, 8.2173E+02, 8.2287E+02, 8.2400E+02,
     + 8.2514E+02, 8.2627E+02, 8.2741E+02, 8.2855E+02, 8.2969E+02,
     +  8.3083E+02, 8.3197E+02, 8.3311E+02, 8.3426E+02, 8.3540E+02,
     + 8.3654E+02, 8.3769E+02, 8.3883E+02, 8.3998E+02, 8.4113E+02,
     +  8.4227E+02, 8.4342E+02, 8.4457E+02, 8.4572E+02, 8.4687E+02,
     + 8.4802E+02, 8.4918E+02, 8.5033E+02, 8.5148E+02, 8.5264E+02,
     +  8.5379E+02, 8.5495E+02, 8.5611E+02, 8.5727E+02, 8.5843E+02,
     + 8.5958E+02, 8.6075E+02, 8.6191E+02, 8.6307E+02, 8.6423E+02,
     +  8.6539E+02, 8.6656E+02, 8.6772E+02, 8.6889E+02, 8.7006E+02,
     + 8.7122E+02, 8.7239E+02, 8.7356E+02, 8.7473E+02, 8.7590E+02,
     +  8.7707E+02, 8.7824E+02, 8.7942E+02, 8.8059E+02, 8.8177E+02,
     + 8.8294E+02, 8.8412E+02, 8.8529E+02, 8.8647E+02, 8.8765E+02,
     +  8.8883E+02, 8.9001E+02, 8.9119E+02, 8.9237E+02, 8.9355E+02,
     + 8.9474E+02, 8.9592E+02, 8.9711E+02, 8.9829E+02, 8.9948E+02,
     +  9.0067E+02, 9.0185E+02, 9.0304E+02, 9.0423E+02, 9.0542E+02,
     + 9.0661E+02, 9.0781E+02, 9.0900E+02, 9.1019E+02, 9.1139E+02,
     +  9.1258E+02, 9.1378E+02, 9.1497E+02, 9.1617E+02, 9.1737E+02,
     + 9.1857E+02, 9.1977E+02, 9.2097E+02, 9.2217E+02, 9.2337E+02,
     +  9.2458E+02, 9.2578E+02, 9.2699E+02, 9.2819E+02, 9.2940E+02,
     + 9.3061E+02, 9.3181E+02, 9.3302E+02, 9.3423E+02, 9.3544E+02,
     +  9.3665E+02, 9.3786E+02, 9.3908E+02, 9.4029E+02, 9.4151E+02,
     + 9.4272E+02, 9.4394E+02, 9.4515E+02, 9.4637E+02, 9.4759E+02,
     +  9.4881E+02, 9.5003E+02, 9.5125E+02, 9.5247E+02, 9.5369E+02,
     + 9.5492E+02, 9.5614E+02, 9.5737E+02, 9.5859E+02, 9.5982E+02,
     +  9.6104E+02, 9.6227E+02, 9.6350E+02, 9.6473E+02, 9.6596E+02,
     + 9.6719E+02, 9.6842E+02, 9.6966E+02, 9.7089E+02, 9.7213E+02,
     +  9.7336E+02, 9.7460E+02, 9.7583E+02, 9.7707E+02, 9.7831E+02,
     + 9.7955E+02, 9.8079E+02, 9.8203E+02, 9.8327E+02, 9.8452E+02,
     +  9.8576E+02, 9.8700E+02, 9.8825E+02, 9.8949E+02, 9.9074E+02,
     + 9.9199E+02, 9.9324E+02, 9.9449E+02, 9.9574E+02, 9.9699E+02,
     +  9.9824E+02, 9.9949E+02, 1.0007E+03, 1.0020E+03, 1.0033E+03,
     + 1.0045E+03, 1.0058E+03, 1.0070E+03, 1.0083E+03, 1.0095E+03,
     +  1.0108E+03, 1.0121E+03, 1.0133E+03, 1.0146E+03, 1.0158E+03,
     + 1.0171E+03, 1.0184E+03, 1.0196E+03, 1.0209E+03, 1.0222E+03,
     +  1.0234E+03, 1.0247E+03, 1.0260E+03, 1.0272E+03, 1.0285E+03,
     + 1.0298E+03, 1.0311E+03, 1.0323E+03, 1.0336E+03, 1.0349E+03,
     +  1.0362E+03, 1.0374E+03, 1.0387E+03, 1.0400E+03, 1.0413E+03,
     + 1.0426E+03, 1.0438E+03, 1.0451E+03, 1.0464E+03, 1.0477E+03,
     +  1.0490E+03, 1.0503E+03, 1.0515E+03, 1.0528E+03, 1.0541E+03,
     + 1.0554E+03, 1.0567E+03, 1.0580E+03, 1.0593E+03, 1.0606E+03,
     +  1.0619E+03, 1.0631E+03, 1.0644E+03, 1.0657E+03, 1.0670E+03,
     + 1.0683E+03, 1.0696E+03, 1.0709E+03, 1.0722E+03, 1.0735E+03,
     +  1.0748E+03, 1.0761E+03, 1.0774E+03, 1.0787E+03, 1.0800E+03,
     + 1.0813E+03, 1.0826E+03, 1.0840E+03, 1.0853E+03, 1.0866E+03,
     +  1.0879E+03, 1.0892E+03, 1.0905E+03, 1.0918E+03, 1.0931E+03,
     + 1.0944E+03, 1.0957E+03, 1.0971E+03, 1.0984E+03, 1.0997E+03,
     +  1.1010E+03, 1.1023E+03, 1.1036E+03, 1.1050E+03, 1.1063E+03,
     + 1.1076E+03, 1.1089E+03, 1.1103E+03, 1.1116E+03, 1.1129E+03,
     +  1.1142E+03, 1.1156E+03, 1.1169E+03, 1.1182E+03, 1.1195E+03,
     + 1.1209E+03, 1.1222E+03, 1.1235E+03, 1.1249E+03, 1.1262E+03,
     +  1.1275E+03, 1.1289E+03, 1.1302E+03, 1.1315E+03, 1.1329E+03,
     + 1.1342E+03, 1.1356E+03, 1.1369E+03, 1.1382E+03, 1.1396E+03,
     +  1.1409E+03, 1.1423E+03, 1.1436E+03, 1.1449E+03, 1.1463E+03,
     + 1.1476E+03, 1.1490E+03, 1.1503E+03, 1.1517E+03, 1.1530E+03,
     +  1.1544E+03, 1.1557E+03, 1.1571E+03, 1.1584E+03, 1.1598E+03,
     + 1.1612E+03, 1.1625E+03, 1.1639E+03, 1.1652E+03, 1.1666E+03,
     +  1.1679E+03, 1.1693E+03, 1.1707E+03, 1.1720E+03, 1.1734E+03,
     + 1.1747E+03, 1.1761E+03, 1.1775E+03, 1.1788E+03, 1.1802E+03,
     +  1.1816E+03, 1.1829E+03, 1.1843E+03, 1.1857E+03, 1.1871E+03,
     + 1.1884E+03, 1.1898E+03, 1.1912E+03, 1.1925E+03, 1.1939E+03,
     +  1.1953E+03, 1.1967E+03, 1.1980E+03, 1.1994E+03, 1.2008E+03,
     + 1.2022E+03, 1.2036E+03, 1.2049E+03, 1.2063E+03, 1.2077E+03,
     +  1.2091E+03, 1.2105E+03, 1.2119E+03, 1.2133E+03, 1.2146E+03,
     + 1.2160E+03, 1.2174E+03, 1.2188E+03, 1.2202E+03, 1.2216E+03,
     +  1.2230E+03, 1.2244E+03, 1.2258E+03, 1.2272E+03, 1.2286E+03,
     + 1.2300E+03, 1.2314E+03, 1.2328E+03, 1.2342E+03, 1.2356E+03,
     +  1.2370E+03, 1.2384E+03, 1.2398E+03, 1.2412E+03, 1.2426E+03,
     + 1.2440E+03, 1.2454E+03, 1.2468E+03, 1.2482E+03, 1.2496E+03,
     +  1.2510E+03, 1.2524E+03, 1.2538E+03, 1.2553E+03, 1.2567E+03,
     + 1.2581E+03, 1.2595E+03, 1.2609E+03, 1.2623E+03, 1.2637E+03,
     +  1.2652E+03, 1.2666E+03, 1.2680E+03, 1.2694E+03, 1.2708E+03,
     + 1.2723E+03, 1.2737E+03, 1.2751E+03, 1.2765E+03, 1.2780E+03,
     +  1.2794E+03, 1.2808E+03, 1.2822E+03, 1.2837E+03, 1.2851E+03,
     + 1.2865E+03, 1.2880E+03, 1.2894E+03, 1.2908E+03, 1.2923E+03,
     +  1.2937E+03, 1.2951E+03, 1.2966E+03, 1.2980E+03, 1.2995E+03,
     + 1.3009E+03, 1.3023E+03, 1.3038E+03, 1.3052E+03, 1.3067E+03,
     +  1.3081E+03, 1.3095E+03, 1.3110E+03, 1.3124E+03, 1.3139E+03,
     + 1.3153E+03, 1.3168E+03, 1.3182E+03, 1.3197E+03, 1.3211E+03,
     +  1.3226E+03, 1.3240E+03, 1.3255E+03, 1.3269E+03, 1.3284E+03,
     + 1.3299E+03, 1.3313E+03, 1.3328E+03, 1.3342E+03, 1.3357E+03,
     +  1.3372E+03, 1.3386E+03, 1.3401E+03, 1.3415E+03, 1.3430E+03,
     + 1.3445E+03, 1.3459E+03, 1.3474E+03, 1.3489E+03, 1.3503E+03,
     +  1.3518E+03, 1.3533E+03, 1.3548E+03, 1.3562E+03, 1.3577E+03,
     + 1.3592E+03, 1.3606E+03, 1.3621E+03, 1.3636E+03, 1.3651E+03,
     +  1.3666E+03, 1.3680E+03, 1.3695E+03, 1.3710E+03, 1.3725E+03,
     + 1.3740E+03, 1.3754E+03, 1.3769E+03, 1.3784E+03, 1.3799E+03,
     +  1.3814E+03, 1.3829E+03, 1.3844E+03, 1.3858E+03, 1.3873E+03,
     + 1.3888E+03, 1.3903E+03, 1.3918E+03, 1.3933E+03, 1.3948E+03,
     +  1.3963E+03, 1.3978E+03, 1.3993E+03, 1.4008E+03, 1.4023E+03,
     + 1.4038E+03, 1.4053E+03, 1.4068E+03, 1.4083E+03, 1.4098E+03,
     +  1.4113E+03, 1.4128E+03, 1.4143E+03, 1.4158E+03, 1.4173E+03,
     + 1.4188E+03, 1.4203E+03, 1.4219E+03, 1.4234E+03, 1.4249E+03,
     +  1.4264E+03, 1.4279E+03, 1.4294E+03, 1.4309E+03, 1.4324E+03,
     + 1.4340E+03, 1.4355E+03, 1.4370E+03, 1.4385E+03, 1.4400E+03,
     +  1.4416E+03, 1.4431E+03, 1.4446E+03, 1.4461E+03, 1.4477E+03,
     + 1.4492E+03, 1.4507E+03, 1.4522E+03, 1.4538E+03, 1.4553E+03,
     +  1.4568E+03, 1.4584E+03, 1.4599E+03, 1.4614E+03, 1.4630E+03,
     + 1.4645E+03, 1.4660E+03, 1.4676E+03, 1.4691E+03, 1.4706E+03,
     +  1.4722E+03, 1.4737E+03, 1.4753E+03, 1.4768E+03, 1.4783E+03,
     + 1.4799E+03, 1.4814E+03, 1.4830E+03, 1.4845E+03, 1.4861E+03,
     +  1.4876E+03, 1.4892E+03, 1.4907E+03, 1.4923E+03, 1.4938E+03,
     + 1.4954E+03, 1.4969E+03, 1.4985E+03, 1.5000E+03, 1.5016E+03,
     +  1.5031E+03, 1.5047E+03, 1.5062E+03, 1.5078E+03, 1.5094E+03,
     + 1.5109E+03, 1.5125E+03, 1.5141E+03, 1.5156E+03, 1.5172E+03,
     +  1.5187E+03, 1.5203E+03, 1.5219E+03, 1.5234E+03, 1.5250E+03,
     + 1.5266E+03, 1.5282E+03, 1.5297E+03, 1.5313E+03, 1.5329E+03,
     +  1.5344E+03, 1.5360E+03, 1.5376E+03, 1.5392E+03, 1.5407E+03,
     + 1.5423E+03, 1.5439E+03, 1.5455E+03, 1.5471E+03, 1.5486E+03,
     +  1.5502E+03, 1.5518E+03, 1.5534E+03, 1.5550E+03, 1.5566E+03,
     + 1.5582E+03, 1.5597E+03, 1.5613E+03, 1.5629E+03, 1.5645E+03,
     +  1.5661E+03, 1.5677E+03, 1.5693E+03, 1.5709E+03, 1.5725E+03,
     + 1.5741E+03, 1.5757E+03, 1.5773E+03, 1.5789E+03, 1.5805E+03,
     +  1.5821E+03, 1.5837E+03, 1.5853E+03, 1.5869E+03, 1.5885E+03,
     + 1.5901E+03, 1.5917E+03, 1.5933E+03, 1.5949E+03, 1.5965E+03,
     +  1.5981E+03, 1.5997E+03, 1.6013E+03, 1.6030E+03, 1.6046E+03,
     + 1.6062E+03, 1.6078E+03, 1.6094E+03, 1.6110E+03, 1.6126E+03,
     +  1.6143E+03, 1.6159E+03, 1.6175E+03, 1.6191E+03, 1.6207E+03,
     + 1.6224E+03, 1.6240E+03, 1.6256E+03, 1.6272E+03, 1.6289E+03,
     +  1.6305E+03, 1.6321E+03, 1.6337E+03, 1.6354E+03, 1.6370E+03,
     + 1.6386E+03, 1.6403E+03, 1.6419E+03, 1.6435E+03, 1.6452E+03,
     +  1.6468E+03, 1.6484E+03, 1.6501E+03, 1.6517E+03, 1.6534E+03,
     + 1.6550E+03, 1.6566E+03, 1.6583E+03, 1.6599E+03, 1.6616E+03,
     +  1.6632E+03, 1.6649E+03, 1.6665E+03, 1.6681E+03, 1.6698E+03,
     + 1.6714E+03, 1.6731E+03, 1.6747E+03, 1.6764E+03, 1.6780E+03,
     +  1.6797E+03, 1.6814E+03, 1.6830E+03, 1.6847E+03, 1.6863E+03,
     + 1.6880E+03, 1.6896E+03, 1.6913E+03, 1.6930E+03, 1.6946E+03,
     +  1.6963E+03, 1.6979E+03, 1.6996E+03, 1.7013E+03, 1.7029E+03,
     + 1.7046E+03, 1.7063E+03, 1.7079E+03, 1.7096E+03, 1.7113E+03,
     +  1.7130E+03, 1.7146E+03, 1.7163E+03, 1.7180E+03, 1.7196E+03,
     + 1.7213E+03, 1.7230E+03, 1.7247E+03, 1.7264E+03, 1.7280E+03,
     +  1.7297E+03, 1.7314E+03, 1.7331E+03, 1.7348E+03, 1.7364E+03,
     + 1.7381E+03, 1.7398E+03, 1.7415E+03, 1.7432E+03, 1.7449E+03,
     +  1.7466E+03, 1.7483E+03, 1.7499E+03, 1.7516E+03, 1.7533E+03,
     + 1.7550E+03, 1.7567E+03, 1.7584E+03, 1.7601E+03, 1.7618E+03,
     +  1.7635E+03, 1.7652E+03, 1.7669E+03, 1.7686E+03, 1.7703E+03,
     + 1.7720E+03, 1.7737E+03, 1.7754E+03, 1.7771E+03, 1.7788E+03,
     +  1.7805E+03, 1.7822E+03, 1.7840E+03, 1.7857E+03, 1.7874E+03,
     + 1.7891E+03, 1.7908E+03, 1.7925E+03, 1.7942E+03, 1.7959E+03,
     +  1.7977E+03, 1.7994E+03, 1.8011E+03, 1.8028E+03, 1.8045E+03,
     + 1.8062E+03, 1.8080E+03, 1.8097E+03, 1.8114E+03, 1.8131E+03,
     +  1.8149E+03, 1.8166E+03, 1.8183E+03, 1.8200E+03, 1.8218E+03,
     + 1.8235E+03, 1.8252E+03, 1.8270E+03, 1.8287E+03, 1.8304E+03,
     +  1.8322E+03, 1.8339E+03, 1.8356E+03, 1.8374E+03, 1.8391E+03,
     + 1.8408E+03, 1.8426E+03, 1.8443E+03, 1.8461E+03, 1.8478E+03,
     +  1.8495E+03, 1.8513E+03, 1.8530E+03, 1.8548E+03, 1.8565E+03,
     + 1.8583E+03, 1.8600E+03, 1.8618E+03, 1.8635E+03, 1.8653E+03,
     +  1.8670E+03, 1.8688E+03, 1.8705E+03, 1.8723E+03, 1.8740E+03,
     + 1.8758E+03, 1.8776E+03, 1.8793E+03, 1.8811E+03, 1.8828E+03,
     +  1.8846E+03, 1.8864E+03, 1.8881E+03, 1.8899E+03, 1.8916E+03,
     + 1.8934E+03, 1.8952E+03, 1.8969E+03, 1.8987E+03, 1.9005E+03,
     +  1.9023E+03, 1.9040E+03, 1.9058E+03, 1.9076E+03, 1.9093E+03,
     + 1.9111E+03, 1.9129E+03, 1.9147E+03, 1.9164E+03, 1.9182E+03,
     +  1.9200E+03, 1.9218E+03, 1.9236E+03, 1.9253E+03, 1.9271E+03,
     + 1.9289E+03, 1.9307E+03, 1.9325E+03, 1.9343E+03, 1.9361E+03,
     +  1.9378E+03, 1.9396E+03, 1.9414E+03, 1.9432E+03, 1.9450E+03,
     + 1.9468E+03, 1.9486E+03, 1.9504E+03, 1.9522E+03, 1.9540E+03,
     +  1.9558E+03, 1.9576E+03, 1.9594E+03, 1.9612E+03, 1.9630E+03,
     + 1.9648E+03, 1.9666E+03, 1.9684E+03, 1.9702E+03, 1.9720E+03,
     +  1.9738E+03, 1.9756E+03, 1.9774E+03, 1.9792E+03, 1.9810E+03,
     + 1.9828E+03, 1.9847E+03, 1.9865E+03, 1.9883E+03, 1.9901E+03,
     +  1.9919E+03, 1.9937E+03, 1.9955E+03, 1.9974E+03, 1.9992E+03,
     + 2.0010E+03, 2.0028E+03, 2.0046E+03, 2.0065E+03, 2.0083E+03,
     +  2.0101E+03, 2.0119E+03, 2.0138E+03, 2.0156E+03, 2.0174E+03,
     + 2.0192E+03, 2.0211E+03, 2.0229E+03, 2.0247E+03, 2.0266E+03,
     +  2.0284E+03, 2.0302E+03, 2.0321E+03, 2.0339E+03, 2.0357E+03,
     + 2.0376E+03, 2.0394E+03, 2.0413E+03, 2.0431E+03, 2.0449E+03,
     +  2.0468E+03, 2.0486E+03, 2.0505E+03, 2.0523E+03, 2.0542E+03,
     + 2.0560E+03, 2.0579E+03, 2.0597E+03, 2.0616E+03, 2.0634E+03,
     +  2.0653E+03, 2.0671E+03, 2.0690E+03, 2.0708E+03, 2.0727E+03,
     + 2.0745E+03, 2.0764E+03, 2.0783E+03, 2.0801E+03, 2.0820E+03,
     +  2.0838E+03, 2.0857E+03, 2.0876E+03, 2.0894E+03, 2.0913E+03,
     + 2.0932E+03, 2.0950E+03, 2.0969E+03, 2.0988E+03, 2.1006E+03,
     +  2.1025E+03, 2.1044E+03, 2.1062E+03, 2.1081E+03, 2.1100E+03,
     + 2.1119E+03, 2.1137E+03, 2.1156E+03, 2.1175E+03, 2.1194E+03,
     +  2.1212E+03, 2.1231E+03, 2.1250E+03, 2.1269E+03, 2.1288E+03,
     + 2.1307E+03, 2.1325E+03, 2.1344E+03, 2.1363E+03, 2.1382E+03,
     +  2.1401E+03, 2.1420E+03, 2.1439E+03, 2.1458E+03, 2.1476E+03,
     + 2.1495E+03, 2.1514E+03, 2.1533E+03, 2.1552E+03, 2.1571E+03,
     +  2.1590E+03, 2.1609E+03, 2.1628E+03, 2.1647E+03, 2.1666E+03,
     + 2.1685E+03, 2.1704E+03, 2.1723E+03, 2.1742E+03, 2.1761E+03,
     +  2.1780E+03, 2.1800E+03, 2.1819E+03, 2.1838E+03, 2.1857E+03,
     + 2.1876E+03, 2.1895E+03, 2.1914E+03, 2.1933E+03, 2.1952E+03,
     +  2.1972E+03, 2.1991E+03, 2.2010E+03, 2.2029E+03, 2.2048E+03,
     + 2.2068E+03, 2.2087E+03, 2.2106E+03, 2.2125E+03, 2.2144E+03,
     +  2.2164E+03, 2.2183E+03, 2.2202E+03, 2.2222E+03, 2.2241E+03,
     + 2.2260E+03, 2.2279E+03, 2.2299E+03, 2.2318E+03, 2.2337E+03,
     +  2.2357E+03, 2.2376E+03, 2.2395E+03, 2.2415E+03, 2.2434E+03,
     + 2.2454E+03, 2.2473E+03, 2.2492E+03, 2.2512E+03, 2.2531E+03,
     +  2.2551E+03, 2.2570E+03, 2.2590E+03, 2.2609E+03, 2.2628E+03,
     + 2.2648E+03, 2.2667E+03, 2.2687E+03, 2.2706E+03, 2.2726E+03,
     +  2.2746E+03, 2.2765E+03, 2.2785E+03, 2.2804E+03, 2.2824E+03,
     + 2.2843E+03, 2.2863E+03, 2.2882E+03, 2.2902E+03, 2.2922E+03,
     +  2.2941E+03, 2.2961E+03, 2.2981E+03, 2.3000E+03, 2.3020E+03,
     + 2.3040E+03, 2.3059E+03, 2.3079E+03, 2.3099E+03, 2.3118E+03,
     +  2.3138E+03, 2.3158E+03, 2.3177E+03, 2.3197E+03, 2.3217E+03,
     + 2.3237E+03, 2.3256E+03, 2.3276E+03, 2.3296E+03, 2.3316E+03,
     +  2.3336E+03, 2.3355E+03, 2.3375E+03, 2.3395E+03, 2.3415E+03,
     + 2.3435E+03, 2.3455E+03, 2.3474E+03, 2.3494E+03, 2.3514E+03,
     +  2.3534E+03, 2.3554E+03, 2.3574E+03, 2.3594E+03, 2.3614E+03,
     + 2.3634E+03, 2.3654E+03, 2.3674E+03, 2.3694E+03, 2.3714E+03,
     +  2.3734E+03, 2.3754E+03, 2.3774E+03, 2.3794E+03, 2.3814E+03,
     + 2.3834E+03, 2.3854E+03, 2.3874E+03, 2.3894E+03, 2.3914E+03,
     +  2.3934E+03, 2.3954E+03, 2.3974E+03, 2.3994E+03, 2.4014E+03,
     + 2.4035E+03, 2.4055E+03, 2.4075E+03, 2.4095E+03, 2.4115E+03,
     +  2.4135E+03, 2.4156E+03, 2.4176E+03, 2.4196E+03, 2.4216E+03,
     + 2.4236E+03, 2.4257E+03, 2.4277E+03, 2.4297E+03, 2.4317E+03,
     +  2.4338E+03, 2.4358E+03, 2.4378E+03, 2.4398E+03, 2.4419E+03,
     + 2.4439E+03, 2.4459E+03, 2.4480E+03, 2.4500E+03, 2.4520E+03,
     +  2.4541E+03, 2.4561E+03, 2.4581E+03, 2.4602E+03, 2.4622E+03,
     + 2.4643E+03, 2.4663E+03, 2.4684E+03, 2.4704E+03, 2.4724E+03,
     +  2.4745E+03, 2.4765E+03, 2.4786E+03, 2.4806E+03, 2.4827E+03,
     + 2.4847E+03, 2.4868E+03, 2.4888E+03, 2.4909E+03, 2.4929E+03,
     +  2.4950E+03, 2.4970E+03, 2.4991E+03, 2.5012E+03, 2.5032E+03,
     + 2.5053E+03, 2.5073E+03, 2.5094E+03, 2.5115E+03, 2.5135E+03,
     +  2.5156E+03, 2.5176E+03, 2.5197E+03, 2.5218E+03, 2.5238E+03,
     + 2.5259E+03, 2.5280E+03, 2.5301E+03, 2.5321E+03, 2.5342E+03,
     +  2.5363E+03, 2.5383E+03, 2.5404E+03, 2.5425E+03, 2.5446E+03,
     + 2.5466E+03, 2.5487E+03, 2.5508E+03, 2.5529E+03, 2.5550E+03,
     +  2.5571E+03, 2.5591E+03, 2.5612E+03, 2.5633E+03, 2.5654E+03,
     + 2.5675E+03, 2.5696E+03, 2.5717E+03, 2.5737E+03, 2.5758E+03,
     +  2.5779E+03, 2.5800E+03, 2.5821E+03, 2.5842E+03, 2.5863E+03,
     + 2.5884E+03, 2.5905E+03, 2.5926E+03, 2.5947E+03, 2.5968E+03,
     +  2.5989E+03, 2.6010E+03, 2.6031E+03, 2.6052E+03, 2.6073E+03,
     + 2.6094E+03, 2.6115E+03, 2.6136E+03, 2.6157E+03, 2.6178E+03,
     +  2.6200E+03, 2.6221E+03, 2.6242E+03, 2.6263E+03, 2.6284E+03,
     + 2.6305E+03, 2.6326E+03, 2.6348E+03, 2.6369E+03, 2.6390E+03,
     +  2.6411E+03, 2.6432E+03, 2.6454E+03, 2.6475E+03, 2.6496E+03,
     + 2.6517E+03, 2.6538E+03, 2.6560E+03, 2.6581E+03, 2.6602E+03,
     +  2.6624E+03, 2.6645E+03, 2.6666E+03, 2.6687E+03, 2.6709E+03,
     + 2.6730E+03, 2.6752E+03, 2.6773E+03, 2.6794E+03, 2.6816E+03,
     +  2.6837E+03, 2.6858E+03, 2.6880E+03, 2.6901E+03, 2.6923E+03,
     + 2.6944E+03, 2.6965E+03, 2.6987E+03, 2.7008E+03, 2.7030E+03,
     +  2.7051E+03, 2.7073E+03, 2.7094E+03, 2.7116E+03, 2.7137E+03,
     + 2.7159E+03, 2.7180E+03, 2.7202E+03, 2.7223E+03, 2.7245E+03,
     +  2.7267E+03, 2.7288E+03, 2.7310E+03, 2.7331E+03, 2.7353E+03,
     + 2.7375E+03, 2.7396E+03, 2.7418E+03, 2.7439E+03, 2.7461E+03,
     +  2.7483E+03, 2.7504E+03, 2.7526E+03, 2.7548E+03, 2.7570E+03,
     + 2.7591E+03, 2.7613E+03, 2.7635E+03, 2.7656E+03, 2.7678E+03,
     +  2.7700E+03, 2.7722E+03, 2.7743E+03, 2.7765E+03, 2.7787E+03,
     + 2.7809E+03, 2.7831E+03, 2.7853E+03, 2.7874E+03, 2.7896E+03,
     +  2.7918E+03, 2.7940E+03, 2.7962E+03, 2.7984E+03, 2.8006E+03,
     + 2.8027E+03, 2.8049E+03, 2.8071E+03, 2.8093E+03, 2.8115E+03,
     +  2.8137E+03, 2.8159E+03, 2.8181E+03, 2.8203E+03, 2.8225E+03,
     + 2.8247E+03, 2.8269E+03, 2.8291E+03, 2.8313E+03, 2.8335E+03,
     +  2.8357E+03, 2.8379E+03, 2.8401E+03, 2.8423E+03, 2.8445E+03,
     + 2.8467E+03, 2.8489E+03, 2.8512E+03, 2.8534E+03, 2.8556E+03,
     +  2.8578E+03, 2.8600E+03, 2.8622E+03, 2.8644E+03, 2.8666E+03,
     + 2.8689E+03, 2.8711E+03, 2.8733E+03, 2.8755E+03, 2.8777E+03,
     +  2.8800E+03, 2.8822E+03, 2.8844E+03, 2.8866E+03, 2.8889E+03,
     + 2.8911E+03, 2.8933E+03, 2.8956E+03, 2.8978E+03, 2.9000E+03,
     +  2.9022E+03, 2.9045E+03, 2.9067E+03, 2.9089E+03, 2.9112E+03,
     + 2.9134E+03, 2.9157E+03, 2.9179E+03, 2.9201E+03, 2.9224E+03,
     +  2.9246E+03, 2.9269E+03, 2.9291E+03, 2.9313E+03, 2.9336E+03,
     + 2.9358E+03, 2.9381E+03, 2.9403E+03, 2.9426E+03, 2.9448E+03,
     +  2.9471E+03, 2.9493E+03, 2.9516E+03, 2.9538E+03, 2.9561E+03,
     + 2.9583E+03, 2.9606E+03, 2.9629E+03, 2.9651E+03, 2.9674E+03,
     +  2.9696E+03, 2.9719E+03, 2.9742E+03, 2.9764E+03, 2.9787E+03,
     + 2.9809E+03, 2.9832E+03, 2.9855E+03, 2.9877E+03, 2.9900E+03,
     +  2.9923E+03, 2.9946E+03, 2.9968E+03, 2.9991E+03, 3.0014E+03,
     + 3.0036E+03, 3.0059E+03, 3.0082E+03, 3.0105E+03, 3.0128E+03,
     +  3.0150E+03, 3.0173E+03, 3.0196E+03, 3.0219E+03, 3.0242E+03,
     + 3.0264E+03, 3.0287E+03, 3.0310E+03, 3.0333E+03, 3.0356E+03,
     +  3.0379E+03, 3.0402E+03, 3.0424E+03, 3.0447E+03, 3.0470E+03,
     + 3.0493E+03, 3.0516E+03, 3.0539E+03, 3.0562E+03, 3.0585E+03,
     +  3.0608E+03, 3.0631E+03, 3.0654E+03, 3.0677E+03, 3.0700E+03,
     + 3.0723E+03, 3.0746E+03, 3.0769E+03, 3.0792E+03, 3.0815E+03,
     +  3.0838E+03, 3.0861E+03, 3.0884E+03, 3.0908E+03, 3.0931E+03,
     + 3.0954E+03, 3.0977E+03, 3.1000E+03, 3.1023E+03, 3.1046E+03,
     +  3.1069E+03, 3.1093E+03, 3.1116E+03, 3.1139E+03, 3.1162E+03,
     + 3.1185E+03, 3.1209E+03, 3.1232E+03, 3.1255E+03, 3.1278E+03,
     +  3.1302E+03, 3.1325E+03, 3.1348E+03, 3.1371E+03, 3.1395E+03,
     + 3.1418E+03, 3.1441E+03, 3.1465E+03, 3.1488E+03, 3.1511E+03,
     +  3.1535E+03, 3.1558E+03, 3.1581E+03, 3.1605E+03, 3.1628E+03,
     + 3.1652E+03, 3.1675E+03, 3.1698E+03, 3.1722E+03, 3.1745E+03,
     +  3.1769E+03, 3.1792E+03, 3.1816E+03, 3.1839E+03, 3.1863E+03,
     + 3.1886E+03, 3.1910E+03, 3.1933E+03, 3.1957E+03, 3.1980E+03,
     +  3.2004E+03, 3.2027E+03, 3.2051E+03, 3.2074E+03, 3.2098E+03,
     + 3.2122E+03, 3.2145E+03, 3.2169E+03, 3.2192E+03, 3.2216E+03,
     +  3.2240E+03, 3.2263E+03, 3.2287E+03, 3.2311E+03, 3.2334E+03,
     + 3.2358E+03, 3.2382E+03, 3.2405E+03, 3.2429E+03, 3.2453E+03,
     +  3.2476E+03, 3.2500E+03, 3.2524E+03, 3.2548E+03, 3.2571E+03,
     + 3.2595E+03, 3.2619E+03, 3.2643E+03, 3.2667E+03, 3.2690E+03,
     +  3.2714E+03, 3.2738E+03, 3.2762E+03, 3.2786E+03, 3.2810E+03,
     + 3.2833E+03, 3.2857E+03, 3.2881E+03, 3.2905E+03, 3.2929E+03,
     +  3.2953E+03, 3.2977E+03, 3.3001E+03, 3.3025E+03, 3.3049E+03,
     + 3.3073E+03, 3.3097E+03, 3.3121E+03, 3.3145E+03, 3.3169E+03,
     +  3.3193E+03, 3.3217E+03, 3.3241E+03, 3.3265E+03, 3.3289E+03,
     + 3.3313E+03, 3.3337E+03, 3.3361E+03, 3.3385E+03, 3.3409E+03,
     +  3.3433E+03, 3.3457E+03, 3.3481E+03, 3.3506E+03, 3.3530E+03,
     + 3.3554E+03, 3.3578E+03, 3.3602E+03, 3.3626E+03, 3.3651E+03,
     +  3.3675E+03, 3.3699E+03, 3.3723E+03, 3.3747E+03, 3.3772E+03,
     + 3.3796E+03, 3.3820E+03, 3.3844E+03, 3.3869E+03, 3.3893E+03,
     +  3.3917E+03, 3.3942E+03, 3.3966E+03, 3.3990E+03, 3.4014E+03/
      data (QofT( 2,J),J=2409,4510)/ 3.4039E+03, 3.4063E+03,
     +  3.4088E+03, 3.4112E+03, 3.4136E+03, 3.4161E+03, 3.4185E+03,
     +  3.4209E+03,
     + 3.4234E+03, 3.4258E+03, 3.4283E+03, 3.4307E+03, 3.4332E+03,
     +  3.4356E+03, 3.4381E+03, 3.4405E+03, 3.4430E+03, 3.4454E+03,
     + 3.4479E+03, 3.4503E+03, 3.4528E+03, 3.4552E+03, 3.4577E+03,
     +  3.4601E+03, 3.4626E+03, 3.4650E+03, 3.4675E+03, 3.4699E+03,
     + 3.4724E+03, 3.4749E+03, 3.4773E+03, 3.4798E+03, 3.4823E+03,
     +  3.4847E+03, 3.4872E+03, 3.4897E+03, 3.4921E+03, 3.4946E+03,
     + 3.4971E+03, 3.4995E+03, 3.5020E+03, 3.5045E+03, 3.5069E+03,
     +  3.5094E+03, 3.5119E+03, 3.5144E+03, 3.5169E+03, 3.5193E+03,
     + 3.5218E+03, 3.5243E+03, 3.5268E+03, 3.5293E+03, 3.5317E+03,
     +  3.5342E+03, 3.5367E+03, 3.5392E+03, 3.5417E+03, 3.5442E+03,
     + 3.5466E+03, 3.5491E+03, 3.5516E+03, 3.5541E+03, 3.5566E+03,
     +  3.5591E+03, 3.5616E+03, 3.5641E+03, 3.5666E+03, 3.5691E+03,
     + 3.5716E+03, 3.5741E+03, 3.5766E+03, 3.5791E+03, 3.5816E+03,
     +  3.5841E+03, 3.5866E+03, 3.5891E+03, 3.5916E+03, 3.5941E+03,
     + 3.5966E+03, 3.5991E+03, 3.6016E+03, 3.6041E+03, 3.6067E+03,
     +  3.6092E+03, 3.6117E+03, 3.6142E+03, 3.6167E+03, 3.6192E+03,
     + 3.6217E+03, 3.6243E+03, 3.6268E+03, 3.6293E+03, 3.6318E+03,
     +  3.6343E+03, 3.6369E+03, 3.6394E+03, 3.6419E+03, 3.6444E+03,
     + 3.6470E+03, 3.6495E+03, 3.6520E+03, 3.6545E+03, 3.6571E+03,
     +  3.6596E+03, 3.6621E+03, 3.6647E+03, 3.6672E+03, 3.6697E+03,
     + 3.6723E+03, 3.6748E+03, 3.6773E+03, 3.6799E+03, 3.6824E+03,
     +  3.6850E+03, 3.6875E+03, 3.6900E+03, 3.6926E+03, 3.6951E+03,
     + 3.6977E+03, 3.7002E+03, 3.7028E+03, 3.7053E+03, 3.7079E+03,
     +  3.7104E+03, 3.7130E+03, 3.7155E+03, 3.7181E+03, 3.7206E+03,
     + 3.7232E+03, 3.7257E+03, 3.7283E+03, 3.7308E+03, 3.7334E+03,
     +  3.7360E+03, 3.7385E+03, 3.7411E+03, 3.7436E+03, 3.7462E+03,
     + 3.7488E+03, 3.7513E+03, 3.7539E+03, 3.7565E+03, 3.7590E+03,
     +  3.7616E+03, 3.7642E+03, 3.7667E+03, 3.7693E+03, 3.7719E+03,
     + 3.7744E+03, 3.7770E+03, 3.7796E+03, 3.7822E+03, 3.7847E+03,
     +  3.7873E+03, 3.7899E+03, 3.7925E+03, 3.7951E+03, 3.7976E+03,
     + 3.8002E+03, 3.8028E+03, 3.8054E+03, 3.8080E+03, 3.8106E+03,
     +  3.8132E+03, 3.8157E+03, 3.8183E+03, 3.8209E+03, 3.8235E+03,
     + 3.8261E+03, 3.8287E+03, 3.8313E+03, 3.8339E+03, 3.8365E+03,
     +  3.8391E+03, 3.8417E+03, 3.8443E+03, 3.8469E+03, 3.8495E+03,
     + 3.8521E+03, 3.8547E+03, 3.8573E+03, 3.8599E+03, 3.8625E+03,
     +  3.8651E+03, 3.8677E+03, 3.8703E+03, 3.8729E+03, 3.8755E+03,
     + 3.8781E+03, 3.8807E+03, 3.8834E+03, 3.8860E+03, 3.8886E+03,
     +  3.8912E+03, 3.8938E+03, 3.8964E+03, 3.8990E+03, 3.9017E+03,
     + 3.9043E+03, 3.9069E+03, 3.9095E+03, 3.9122E+03, 3.9148E+03,
     +  3.9174E+03, 3.9200E+03, 3.9227E+03, 3.9253E+03, 3.9279E+03,
     + 3.9305E+03, 3.9332E+03, 3.9358E+03, 3.9384E+03, 3.9411E+03,
     +  3.9437E+03, 3.9463E+03, 3.9490E+03, 3.9516E+03, 3.9542E+03,
     + 3.9569E+03, 3.9595E+03, 3.9622E+03, 3.9648E+03, 3.9674E+03,
     +  3.9701E+03, 3.9727E+03, 3.9754E+03, 3.9780E+03, 3.9807E+03,
     + 3.9833E+03, 3.9860E+03, 3.9886E+03, 3.9913E+03, 3.9939E+03,
     +  3.9966E+03, 3.9992E+03, 4.0019E+03, 4.0045E+03, 4.0072E+03,
     + 4.0098E+03, 4.0125E+03, 4.0152E+03, 4.0178E+03, 4.0205E+03,
     +  4.0231E+03, 4.0258E+03, 4.0285E+03, 4.0311E+03, 4.0338E+03,
     + 4.0365E+03, 4.0391E+03, 4.0418E+03, 4.0445E+03, 4.0471E+03,
     +  4.0498E+03, 4.0525E+03, 4.0552E+03, 4.0578E+03, 4.0605E+03,
     + 4.0632E+03, 4.0659E+03, 4.0685E+03, 4.0712E+03, 4.0739E+03,
     +  4.0766E+03, 4.0793E+03, 4.0819E+03, 4.0846E+03, 4.0873E+03,
     + 4.0900E+03, 4.0927E+03, 4.0954E+03, 4.0981E+03, 4.1008E+03,
     +  4.1034E+03, 4.1061E+03, 4.1088E+03, 4.1115E+03, 4.1142E+03,
     + 4.1169E+03, 4.1196E+03, 4.1223E+03, 4.1250E+03, 4.1277E+03,
     +  4.1304E+03, 4.1331E+03, 4.1358E+03, 4.1385E+03, 4.1412E+03,
     + 4.1439E+03, 4.1466E+03, 4.1493E+03, 4.1520E+03, 4.1547E+03,
     +  4.1574E+03, 4.1602E+03, 4.1629E+03, 4.1656E+03, 4.1683E+03,
     + 4.1710E+03, 4.1737E+03, 4.1764E+03, 4.1791E+03, 4.1819E+03,
     +  4.1846E+03, 4.1873E+03, 4.1900E+03, 4.1927E+03, 4.1955E+03,
     + 4.1982E+03, 4.2009E+03, 4.2036E+03, 4.2064E+03, 4.2091E+03,
     +  4.2118E+03, 4.2145E+03, 4.2173E+03, 4.2200E+03, 4.2227E+03,
     + 4.2255E+03, 4.2282E+03, 4.2309E+03, 4.2337E+03, 4.2364E+03,
     +  4.2391E+03, 4.2419E+03, 4.2446E+03, 4.2474E+03, 4.2501E+03,
     + 4.2528E+03, 4.2556E+03, 4.2583E+03, 4.2611E+03, 4.2638E+03,
     +  4.2666E+03, 4.2693E+03, 4.2721E+03, 4.2748E+03, 4.2776E+03,
     + 4.2803E+03, 4.2831E+03, 4.2858E+03, 4.2886E+03, 4.2913E+03,
     +  4.2941E+03, 4.2968E+03, 4.2996E+03, 4.3023E+03, 4.3051E+03,
     + 4.3079E+03, 4.3106E+03, 4.3134E+03, 4.3162E+03, 4.3189E+03,
     +  4.3217E+03, 4.3244E+03, 4.3272E+03, 4.3300E+03, 4.3328E+03,
     + 4.3355E+03, 4.3383E+03, 4.3411E+03, 4.3438E+03, 4.3466E+03,
     +  4.3494E+03, 4.3522E+03, 4.3549E+03, 4.3577E+03, 4.3605E+03,
     + 4.3633E+03, 4.3660E+03, 4.3688E+03, 4.3716E+03, 4.3744E+03,
     +  4.3772E+03, 4.3800E+03, 4.3827E+03, 4.3855E+03, 4.3883E+03,
     + 4.3911E+03, 4.3939E+03, 4.3967E+03, 4.3995E+03, 4.4023E+03,
     +  4.4051E+03, 4.4079E+03, 4.4107E+03, 4.4134E+03, 4.4162E+03,
     + 4.4190E+03, 4.4218E+03, 4.4246E+03, 4.4274E+03, 4.4302E+03,
     +  4.4330E+03, 4.4358E+03, 4.4387E+03, 4.4415E+03, 4.4443E+03,
     + 4.4471E+03, 4.4499E+03, 4.4527E+03, 4.4555E+03, 4.4583E+03,
     +  4.4611E+03, 4.4639E+03, 4.4667E+03, 4.4696E+03, 4.4724E+03,
     + 4.4752E+03, 4.4780E+03, 4.4808E+03, 4.4836E+03, 4.4865E+03,
     +  4.4893E+03, 4.4921E+03, 4.4949E+03, 4.4978E+03, 4.5006E+03,
     + 4.5034E+03, 4.5062E+03, 4.5091E+03, 4.5119E+03, 4.5147E+03,
     +  4.5175E+03, 4.5204E+03, 4.5232E+03, 4.5260E+03, 4.5289E+03,
     + 4.5317E+03, 4.5345E+03, 4.5374E+03, 4.5402E+03, 4.5431E+03,
     +  4.5459E+03, 4.5487E+03, 4.5516E+03, 4.5544E+03, 4.5573E+03,
     + 4.5601E+03, 4.5630E+03, 4.5658E+03, 4.5686E+03, 4.5715E+03,
     +  4.5743E+03, 4.5772E+03, 4.5800E+03, 4.5829E+03, 4.5857E+03,
     + 4.5886E+03, 4.5915E+03, 4.5943E+03, 4.5972E+03, 4.6000E+03,
     +  4.6029E+03, 4.6057E+03, 4.6086E+03, 4.6115E+03, 4.6143E+03,
     + 4.6172E+03, 4.6200E+03, 4.6229E+03, 4.6258E+03, 4.6286E+03,
     +  4.6315E+03, 4.6344E+03, 4.6372E+03, 4.6401E+03, 4.6430E+03,
     + 4.6459E+03, 4.6487E+03, 4.6516E+03, 4.6545E+03, 4.6574E+03,
     +  4.6602E+03, 4.6631E+03, 4.6660E+03, 4.6689E+03, 4.6717E+03,
     + 4.6746E+03, 4.6775E+03, 4.6804E+03, 4.6833E+03, 4.6862E+03,
     +  4.6891E+03, 4.6919E+03, 4.6948E+03, 4.6977E+03, 4.7006E+03,
     + 4.7035E+03, 4.7064E+03, 4.7093E+03, 4.7122E+03, 4.7151E+03,
     +  4.7180E+03, 4.7209E+03, 4.7238E+03, 4.7267E+03, 4.7295E+03,
     + 4.7324E+03, 4.7354E+03, 4.7383E+03, 4.7412E+03, 4.7441E+03,
     +  4.7470E+03, 4.7499E+03, 4.7528E+03, 4.7557E+03, 4.7586E+03,
     + 4.7615E+03, 4.7644E+03, 4.7673E+03, 4.7702E+03, 4.7731E+03,
     +  4.7761E+03, 4.7790E+03, 4.7819E+03, 4.7848E+03, 4.7877E+03,
     + 4.7906E+03, 4.7936E+03, 4.7965E+03, 4.7994E+03, 4.8023E+03,
     +  4.8052E+03, 4.8082E+03, 4.8111E+03, 4.8140E+03, 4.8169E+03,
     + 4.8199E+03, 4.8228E+03, 4.8257E+03, 4.8287E+03, 4.8316E+03,
     +  4.8345E+03, 4.8375E+03, 4.8404E+03, 4.8433E+03, 4.8463E+03,
     + 4.8492E+03, 4.8521E+03, 4.8551E+03, 4.8580E+03, 4.8610E+03,
     +  4.8639E+03, 4.8668E+03, 4.8698E+03, 4.8727E+03, 4.8757E+03,
     + 4.8786E+03, 4.8816E+03, 4.8845E+03, 4.8875E+03, 4.8904E+03,
     +  4.8934E+03, 4.8963E+03, 4.8993E+03, 4.9022E+03, 4.9052E+03,
     + 4.9081E+03, 4.9111E+03, 4.9140E+03, 4.9170E+03, 4.9200E+03,
     +  4.9229E+03, 4.9259E+03, 4.9288E+03, 4.9318E+03, 4.9348E+03,
     + 4.9377E+03, 4.9407E+03, 4.9437E+03, 4.9466E+03, 4.9496E+03,
     +  4.9526E+03, 4.9555E+03, 4.9585E+03, 4.9615E+03, 4.9645E+03,
     + 4.9674E+03, 4.9704E+03, 4.9734E+03, 4.9764E+03, 4.9793E+03,
     +  4.9823E+03, 4.9853E+03, 4.9883E+03, 4.9913E+03, 4.9942E+03,
     + 4.9972E+03, 5.0002E+03, 5.0032E+03, 5.0062E+03, 5.0092E+03,
     +  5.0121E+03, 5.0151E+03, 5.0181E+03, 5.0211E+03, 5.0241E+03,
     + 5.0271E+03, 5.0301E+03, 5.0331E+03, 5.0361E+03, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_HCl   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 4),QofT( 4,NT),Q(NT)
      data (Tmx(I),I=1,4)/6000.,6000.,6000.,6000./
      data xgj/ 8.,8.,12.,12./
c...      HCl
c...        --        15
      data (QofT( 1,J),J=1,2408)/ 8.0000E+00, 8.0000E+00, 8.0011E+00,
     +  8.0131E+00, 8.0590E+00, 8.1607E+00, 8.3286E+00, 8.5622E+00,
     + 8.8542E+00, 9.1951E+00, 9.5751E+00, 9.9856E+00, 1.0420E+01,
     +  1.0872E+01, 1.1338E+01, 1.1816E+01, 1.2301E+01, 1.2794E+01,
     + 1.3292E+01, 1.3794E+01, 1.4808E+01, 1.5833E+01, 1.6865E+01,
     +  1.7902E+01, 1.8944E+01, 1.9989E+01, 2.1037E+01, 2.2087E+01,
     + 2.3139E+01, 2.4192E+01, 2.5247E+01, 2.6302E+01, 2.7359E+01,
     +  2.8417E+01, 2.9475E+01, 3.0534E+01, 3.1593E+01, 3.2654E+01,
     + 3.3714E+01, 3.4775E+01, 3.5836E+01, 3.6898E+01, 3.7960E+01,
     +  3.9022E+01, 4.0084E+01, 4.1147E+01, 4.2210E+01, 4.3273E+01,
     + 4.4337E+01, 4.5400E+01, 4.6464E+01, 4.7528E+01, 4.8592E+01,
     +  4.9656E+01, 5.0720E+01, 5.1785E+01, 5.2849E+01, 5.3914E+01,
     + 5.4979E+01, 5.6043E+01, 5.7108E+01, 5.8173E+01, 5.9239E+01,
     +  6.0304E+01, 6.1369E+01, 6.2434E+01, 6.3500E+01, 6.4566E+01,
     + 6.5631E+01, 6.6697E+01, 6.7763E+01, 6.8828E+01, 6.9894E+01,
     +  7.0960E+01, 7.2026E+01, 7.3092E+01, 7.4158E+01, 7.5225E+01,
     + 7.6291E+01, 7.7357E+01, 7.8424E+01, 7.9490E+01, 8.0556E+01,
     +  8.1623E+01, 8.2689E+01, 8.3756E+01, 8.4823E+01, 8.5889E+01,
     + 8.6956E+01, 8.8023E+01, 8.9090E+01, 9.0157E+01, 9.1224E+01,
     +  9.2291E+01, 9.3358E+01, 9.4425E+01, 9.5492E+01, 9.6559E+01,
     + 9.7626E+01, 9.8693E+01, 9.9761E+01, 1.0083E+02, 1.0190E+02,
     +  1.0296E+02, 1.0403E+02, 1.0510E+02, 1.0617E+02, 1.0723E+02,
     + 1.0830E+02, 1.0937E+02, 1.1044E+02, 1.1150E+02, 1.1257E+02,
     +  1.1364E+02, 1.1471E+02, 1.1577E+02, 1.1684E+02, 1.1791E+02,
     + 1.1898E+02, 1.2005E+02, 1.2111E+02, 1.2218E+02, 1.2325E+02,
     +  1.2432E+02, 1.2539E+02, 1.2645E+02, 1.2752E+02, 1.2859E+02,
     + 1.2966E+02, 1.3073E+02, 1.3180E+02, 1.3286E+02, 1.3393E+02,
     +  1.3500E+02, 1.3607E+02, 1.3714E+02, 1.3821E+02, 1.3928E+02,
     + 1.4034E+02, 1.4141E+02, 1.4248E+02, 1.4355E+02, 1.4462E+02,
     +  1.4569E+02, 1.4676E+02, 1.4783E+02, 1.4889E+02, 1.4996E+02,
     + 1.5103E+02, 1.5210E+02, 1.5317E+02, 1.5424E+02, 1.5531E+02,
     +  1.5638E+02, 1.5745E+02, 1.5852E+02, 1.5959E+02, 1.6065E+02,
     + 1.6172E+02, 1.6279E+02, 1.6386E+02, 1.6493E+02, 1.6600E+02,
     +  1.6707E+02, 1.6814E+02, 1.6921E+02, 1.7028E+02, 1.7135E+02,
     + 1.7242E+02, 1.7349E+02, 1.7456E+02, 1.7563E+02, 1.7670E+02,
     +  1.7777E+02, 1.7884E+02, 1.7991E+02, 1.8098E+02, 1.8205E+02,
     + 1.8312E+02, 1.8419E+02, 1.8526E+02, 1.8633E+02, 1.8740E+02,
     +  1.8847E+02, 1.8954E+02, 1.9061E+02, 1.9168E+02, 1.9275E+02,
     + 1.9382E+02, 1.9489E+02, 1.9596E+02, 1.9703E+02, 1.9810E+02,
     +  1.9917E+02, 2.0024E+02, 2.0131E+02, 2.0238E+02, 2.0345E+02,
     + 2.0452E+02, 2.0559E+02, 2.0666E+02, 2.0774E+02, 2.0881E+02,
     +  2.0988E+02, 2.1095E+02, 2.1202E+02, 2.1309E+02, 2.1416E+02,
     + 2.1523E+02, 2.1630E+02, 2.1738E+02, 2.1845E+02, 2.1952E+02,
     +  2.2059E+02, 2.2166E+02, 2.2273E+02, 2.2380E+02, 2.2488E+02,
     + 2.2595E+02, 2.2702E+02, 2.2809E+02, 2.2916E+02, 2.3023E+02,
     +  2.3131E+02, 2.3238E+02, 2.3345E+02, 2.3452E+02, 2.3560E+02,
     + 2.3667E+02, 2.3774E+02, 2.3881E+02, 2.3988E+02, 2.4096E+02,
     +  2.4203E+02, 2.4310E+02, 2.4418E+02, 2.4525E+02, 2.4632E+02,
     + 2.4739E+02, 2.4847E+02, 2.4954E+02, 2.5061E+02, 2.5169E+02,
     +  2.5276E+02, 2.5383E+02, 2.5491E+02, 2.5598E+02, 2.5705E+02,
     + 2.5813E+02, 2.5920E+02, 2.6028E+02, 2.6135E+02, 2.6242E+02,
     +  2.6350E+02, 2.6457E+02, 2.6565E+02, 2.6672E+02, 2.6780E+02,
     + 2.6887E+02, 2.6995E+02, 2.7102E+02, 2.7210E+02, 2.7317E+02,
     +  2.7425E+02, 2.7532E+02, 2.7640E+02, 2.7747E+02, 2.7855E+02,
     + 2.7962E+02, 2.8070E+02, 2.8178E+02, 2.8285E+02, 2.8393E+02,
     +  2.8501E+02, 2.8608E+02, 2.8716E+02, 2.8824E+02, 2.8931E+02,
     + 2.9039E+02, 2.9147E+02, 2.9255E+02, 2.9362E+02, 2.9470E+02,
     +  2.9578E+02, 2.9686E+02, 2.9793E+02, 2.9901E+02, 3.0009E+02,
     + 3.0117E+02, 3.0225E+02, 3.0333E+02, 3.0441E+02, 3.0549E+02,
     +  3.0657E+02, 3.0765E+02, 3.0873E+02, 3.0981E+02, 3.1089E+02,
     + 3.1197E+02, 3.1305E+02, 3.1413E+02, 3.1521E+02, 3.1629E+02,
     +  3.1737E+02, 3.1845E+02, 3.1953E+02, 3.2062E+02, 3.2170E+02,
     + 3.2278E+02, 3.2386E+02, 3.2494E+02, 3.2603E+02, 3.2711E+02,
     +  3.2819E+02, 3.2928E+02, 3.3036E+02, 3.3144E+02, 3.3253E+02,
     + 3.3361E+02, 3.3470E+02, 3.3578E+02, 3.3687E+02, 3.3795E+02,
     +  3.3904E+02, 3.4012E+02, 3.4121E+02, 3.4230E+02, 3.4338E+02,
     + 3.4447E+02, 3.4556E+02, 3.4664E+02, 3.4773E+02, 3.4882E+02,
     +  3.4991E+02, 3.5100E+02, 3.5208E+02, 3.5317E+02, 3.5426E+02,
     + 3.5535E+02, 3.5644E+02, 3.5753E+02, 3.5862E+02, 3.5971E+02,
     +  3.6080E+02, 3.6189E+02, 3.6299E+02, 3.6408E+02, 3.6517E+02,
     + 3.6626E+02, 3.6735E+02, 3.6845E+02, 3.6954E+02, 3.7063E+02,
     +  3.7173E+02, 3.7282E+02, 3.7392E+02, 3.7501E+02, 3.7611E+02,
     + 3.7720E+02, 3.7830E+02, 3.7939E+02, 3.8049E+02, 3.8159E+02,
     +  3.8268E+02, 3.8378E+02, 3.8488E+02, 3.8598E+02, 3.8707E+02,
     + 3.8817E+02, 3.8927E+02, 3.9037E+02, 3.9147E+02, 3.9257E+02,
     +  3.9367E+02, 3.9477E+02, 3.9587E+02, 3.9698E+02, 3.9808E+02,
     + 3.9918E+02, 4.0028E+02, 4.0139E+02, 4.0249E+02, 4.0359E+02,
     +  4.0470E+02, 4.0580E+02, 4.0691E+02, 4.0801E+02, 4.0912E+02,
     + 4.1023E+02, 4.1133E+02, 4.1244E+02, 4.1355E+02, 4.1465E+02,
     +  4.1576E+02, 4.1687E+02, 4.1798E+02, 4.1909E+02, 4.2020E+02,
     + 4.2131E+02, 4.2242E+02, 4.2353E+02, 4.2465E+02, 4.2576E+02,
     +  4.2687E+02, 4.2798E+02, 4.2910E+02, 4.3021E+02, 4.3133E+02,
     + 4.3244E+02, 4.3356E+02, 4.3467E+02, 4.3579E+02, 4.3690E+02,
     +  4.3802E+02, 4.3914E+02, 4.4026E+02, 4.4138E+02, 4.4249E+02,
     + 4.4361E+02, 4.4473E+02, 4.4586E+02, 4.4698E+02, 4.4810E+02,
     +  4.4922E+02, 4.5034E+02, 4.5146E+02, 4.5259E+02, 4.5371E+02,
     + 4.5484E+02, 4.5596E+02, 4.5709E+02, 4.5821E+02, 4.5934E+02,
     +  4.6047E+02, 4.6159E+02, 4.6272E+02, 4.6385E+02, 4.6498E+02,
     + 4.6611E+02, 4.6724E+02, 4.6837E+02, 4.6950E+02, 4.7063E+02,
     +  4.7177E+02, 4.7290E+02, 4.7403E+02, 4.7517E+02, 4.7630E+02,
     + 4.7744E+02, 4.7857E+02, 4.7971E+02, 4.8084E+02, 4.8198E+02,
     +  4.8312E+02, 4.8426E+02, 4.8540E+02, 4.8654E+02, 4.8768E+02,
     + 4.8882E+02, 4.8996E+02, 4.9110E+02, 4.9224E+02, 4.9338E+02,
     +  4.9453E+02, 4.9567E+02, 4.9682E+02, 4.9796E+02, 4.9911E+02,
     + 5.0025E+02, 5.0140E+02, 5.0255E+02, 5.0370E+02, 5.0485E+02,
     +  5.0600E+02, 5.0715E+02, 5.0830E+02, 5.0945E+02, 5.1060E+02,
     + 5.1175E+02, 5.1290E+02, 5.1406E+02, 5.1521E+02, 5.1637E+02,
     +  5.1752E+02, 5.1868E+02, 5.1984E+02, 5.2099E+02, 5.2215E+02,
     + 5.2331E+02, 5.2447E+02, 5.2563E+02, 5.2679E+02, 5.2795E+02,
     +  5.2911E+02, 5.3028E+02, 5.3144E+02, 5.3260E+02, 5.3377E+02,
     + 5.3493E+02, 5.3610E+02, 5.3727E+02, 5.3843E+02, 5.3960E+02,
     +  5.4077E+02, 5.4194E+02, 5.4311E+02, 5.4428E+02, 5.4545E+02,
     + 5.4662E+02, 5.4780E+02, 5.4897E+02, 5.5014E+02, 5.5132E+02,
     +  5.5249E+02, 5.5367E+02, 5.5485E+02, 5.5602E+02, 5.5720E+02,
     + 5.5838E+02, 5.5956E+02, 5.6074E+02, 5.6192E+02, 5.6310E+02,
     +  5.6428E+02, 5.6547E+02, 5.6665E+02, 5.6784E+02, 5.6902E+02,
     + 5.7021E+02, 5.7139E+02, 5.7258E+02, 5.7377E+02, 5.7496E+02,
     +  5.7615E+02, 5.7734E+02, 5.7853E+02, 5.7972E+02, 5.8091E+02,
     + 5.8210E+02, 5.8330E+02, 5.8449E+02, 5.8569E+02, 5.8688E+02,
     +  5.8808E+02, 5.8928E+02, 5.9048E+02, 5.9168E+02, 5.9287E+02,
     + 5.9408E+02, 5.9528E+02, 5.9648E+02, 5.9768E+02, 5.9888E+02,
     +  6.0009E+02, 6.0129E+02, 6.0250E+02, 6.0371E+02, 6.0491E+02,
     + 6.0612E+02, 6.0733E+02, 6.0854E+02, 6.0975E+02, 6.1096E+02,
     +  6.1217E+02, 6.1339E+02, 6.1460E+02, 6.1581E+02, 6.1703E+02,
     + 6.1824E+02, 6.1946E+02, 6.2068E+02, 6.2190E+02, 6.2311E+02,
     +  6.2433E+02, 6.2555E+02, 6.2678E+02, 6.2800E+02, 6.2922E+02,
     + 6.3044E+02, 6.3167E+02, 6.3289E+02, 6.3412E+02, 6.3535E+02,
     +  6.3657E+02, 6.3780E+02, 6.3903E+02, 6.4026E+02, 6.4149E+02,
     + 6.4272E+02, 6.4396E+02, 6.4519E+02, 6.4642E+02, 6.4766E+02,
     +  6.4889E+02, 6.5013E+02, 6.5137E+02, 6.5261E+02, 6.5384E+02,
     + 6.5508E+02, 6.5632E+02, 6.5757E+02, 6.5881E+02, 6.6005E+02,
     +  6.6130E+02, 6.6254E+02, 6.6379E+02, 6.6503E+02, 6.6628E+02,
     + 6.6753E+02, 6.6878E+02, 6.7003E+02, 6.7128E+02, 6.7253E+02,
     +  6.7378E+02, 6.7503E+02, 6.7629E+02, 6.7754E+02, 6.7880E+02,
     + 6.8005E+02, 6.8131E+02, 6.8257E+02, 6.8383E+02, 6.8509E+02,
     +  6.8635E+02, 6.8761E+02, 6.8887E+02, 6.9014E+02, 6.9140E+02,
     + 6.9267E+02, 6.9393E+02, 6.9520E+02, 6.9647E+02, 6.9773E+02,
     +  6.9900E+02, 7.0027E+02, 7.0155E+02, 7.0282E+02, 7.0409E+02,
     + 7.0536E+02, 7.0664E+02, 7.0791E+02, 7.0919E+02, 7.1047E+02,
     +  7.1175E+02, 7.1302E+02, 7.1430E+02, 7.1559E+02, 7.1687E+02,
     + 7.1815E+02, 7.1943E+02, 7.2072E+02, 7.2200E+02, 7.2329E+02,
     +  7.2458E+02, 7.2586E+02, 7.2715E+02, 7.2844E+02, 7.2973E+02,
     + 7.3102E+02, 7.3232E+02, 7.3361E+02, 7.3490E+02, 7.3620E+02,
     +  7.3749E+02, 7.3879E+02, 7.4009E+02, 7.4139E+02, 7.4269E+02,
     + 7.4399E+02, 7.4529E+02, 7.4659E+02, 7.4789E+02, 7.4920E+02,
     +  7.5050E+02, 7.5181E+02, 7.5312E+02, 7.5442E+02, 7.5573E+02,
     + 7.5704E+02, 7.5835E+02, 7.5966E+02, 7.6098E+02, 7.6229E+02,
     +  7.6360E+02, 7.6492E+02, 7.6623E+02, 7.6755E+02, 7.6887E+02,
     + 7.7019E+02, 7.7151E+02, 7.7283E+02, 7.7415E+02, 7.7547E+02,
     +  7.7680E+02, 7.7812E+02, 7.7944E+02, 7.8077E+02, 7.8210E+02,
     + 7.8343E+02, 7.8476E+02, 7.8609E+02, 7.8742E+02, 7.8875E+02,
     +  7.9008E+02, 7.9141E+02, 7.9275E+02, 7.9408E+02, 7.9542E+02,
     + 7.9676E+02, 7.9810E+02, 7.9944E+02, 8.0078E+02, 8.0212E+02,
     +  8.0346E+02, 8.0480E+02, 8.0615E+02, 8.0749E+02, 8.0884E+02,
     + 8.1019E+02, 8.1153E+02, 8.1288E+02, 8.1423E+02, 8.1558E+02,
     +  8.1694E+02, 8.1829E+02, 8.1964E+02, 8.2100E+02, 8.2235E+02,
     + 8.2371E+02, 8.2507E+02, 8.2642E+02, 8.2778E+02, 8.2914E+02,
     +  8.3051E+02, 8.3187E+02, 8.3323E+02, 8.3460E+02, 8.3596E+02,
     + 8.3733E+02, 8.3869E+02, 8.4006E+02, 8.4143E+02, 8.4280E+02,
     +  8.4417E+02, 8.4555E+02, 8.4692E+02, 8.4829E+02, 8.4967E+02,
     + 8.5104E+02, 8.5242E+02, 8.5380E+02, 8.5518E+02, 8.5656E+02,
     +  8.5794E+02, 8.5932E+02, 8.6070E+02, 8.6209E+02, 8.6347E+02,
     + 8.6486E+02, 8.6624E+02, 8.6763E+02, 8.6902E+02, 8.7041E+02,
     +  8.7180E+02, 8.7319E+02, 8.7459E+02, 8.7598E+02, 8.7737E+02,
     + 8.7877E+02, 8.8017E+02, 8.8156E+02, 8.8296E+02, 8.8436E+02,
     +  8.8576E+02, 8.8717E+02, 8.8857E+02, 8.8997E+02, 8.9138E+02,
     + 8.9278E+02, 8.9419E+02, 8.9560E+02, 8.9701E+02, 8.9841E+02,
     +  8.9983E+02, 9.0124E+02, 9.0265E+02, 9.0406E+02, 9.0548E+02,
     + 9.0689E+02, 9.0831E+02, 9.0973E+02, 9.1115E+02, 9.1257E+02,
     +  9.1399E+02, 9.1541E+02, 9.1683E+02, 9.1826E+02, 9.1968E+02,
     + 9.2111E+02, 9.2253E+02, 9.2396E+02, 9.2539E+02, 9.2682E+02,
     +  9.2825E+02, 9.2968E+02, 9.3112E+02, 9.3255E+02, 9.3399E+02,
     + 9.3542E+02, 9.3686E+02, 9.3830E+02, 9.3974E+02, 9.4118E+02,
     +  9.4262E+02, 9.4406E+02, 9.4550E+02, 9.4695E+02, 9.4839E+02,
     + 9.4984E+02, 9.5129E+02, 9.5273E+02, 9.5418E+02, 9.5563E+02,
     +  9.5709E+02, 9.5854E+02, 9.5999E+02, 9.6145E+02, 9.6290E+02,
     + 9.6436E+02, 9.6582E+02, 9.6727E+02, 9.6873E+02, 9.7019E+02,
     +  9.7166E+02, 9.7312E+02, 9.7458E+02, 9.7605E+02, 9.7751E+02,
     + 9.7898E+02, 9.8045E+02, 9.8192E+02, 9.8339E+02, 9.8486E+02,
     +  9.8633E+02, 9.8780E+02, 9.8928E+02, 9.9075E+02, 9.9223E+02,
     + 9.9371E+02, 9.9519E+02, 9.9667E+02, 9.9815E+02, 9.9963E+02,
     +  1.0011E+03, 1.0026E+03, 1.0041E+03, 1.0056E+03, 1.0071E+03,
     + 1.0085E+03, 1.0100E+03, 1.0115E+03, 1.0130E+03, 1.0145E+03,
     +  1.0160E+03, 1.0175E+03, 1.0190E+03, 1.0205E+03, 1.0220E+03,
     + 1.0235E+03, 1.0250E+03, 1.0265E+03, 1.0280E+03, 1.0295E+03,
     +  1.0310E+03, 1.0325E+03, 1.0340E+03, 1.0355E+03, 1.0370E+03,
     + 1.0385E+03, 1.0400E+03, 1.0415E+03, 1.0431E+03, 1.0446E+03,
     +  1.0461E+03, 1.0476E+03, 1.0491E+03, 1.0506E+03, 1.0521E+03,
     + 1.0537E+03, 1.0552E+03, 1.0567E+03, 1.0582E+03, 1.0598E+03,
     +  1.0613E+03, 1.0628E+03, 1.0643E+03, 1.0659E+03, 1.0674E+03,
     + 1.0689E+03, 1.0705E+03, 1.0720E+03, 1.0735E+03, 1.0751E+03,
     +  1.0766E+03, 1.0781E+03, 1.0797E+03, 1.0812E+03, 1.0827E+03,
     + 1.0843E+03, 1.0858E+03, 1.0874E+03, 1.0889E+03, 1.0905E+03,
     +  1.0920E+03, 1.0936E+03, 1.0951E+03, 1.0966E+03, 1.0982E+03,
     + 1.0998E+03, 1.1013E+03, 1.1029E+03, 1.1044E+03, 1.1060E+03,
     +  1.1075E+03, 1.1091E+03, 1.1106E+03, 1.1122E+03, 1.1138E+03,
     + 1.1153E+03, 1.1169E+03, 1.1185E+03, 1.1200E+03, 1.1216E+03,
     +  1.1232E+03, 1.1247E+03, 1.1263E+03, 1.1279E+03, 1.1294E+03,
     + 1.1310E+03, 1.1326E+03, 1.1342E+03, 1.1357E+03, 1.1373E+03,
     +  1.1389E+03, 1.1405E+03, 1.1421E+03, 1.1436E+03, 1.1452E+03,
     + 1.1468E+03, 1.1484E+03, 1.1500E+03, 1.1516E+03, 1.1532E+03,
     +  1.1547E+03, 1.1563E+03, 1.1579E+03, 1.1595E+03, 1.1611E+03,
     + 1.1627E+03, 1.1643E+03, 1.1659E+03, 1.1675E+03, 1.1691E+03,
     +  1.1707E+03, 1.1723E+03, 1.1739E+03, 1.1755E+03, 1.1771E+03,
     + 1.1787E+03, 1.1803E+03, 1.1819E+03, 1.1836E+03, 1.1852E+03,
     +  1.1868E+03, 1.1884E+03, 1.1900E+03, 1.1916E+03, 1.1932E+03,
     + 1.1949E+03, 1.1965E+03, 1.1981E+03, 1.1997E+03, 1.2013E+03,
     +  1.2030E+03, 1.2046E+03, 1.2062E+03, 1.2078E+03, 1.2095E+03,
     + 1.2111E+03, 1.2127E+03, 1.2144E+03, 1.2160E+03, 1.2176E+03,
     +  1.2192E+03, 1.2209E+03, 1.2225E+03, 1.2242E+03, 1.2258E+03,
     + 1.2274E+03, 1.2291E+03, 1.2307E+03, 1.2324E+03, 1.2340E+03,
     +  1.2357E+03, 1.2373E+03, 1.2389E+03, 1.2406E+03, 1.2422E+03,
     + 1.2439E+03, 1.2455E+03, 1.2472E+03, 1.2489E+03, 1.2505E+03,
     +  1.2522E+03, 1.2538E+03, 1.2555E+03, 1.2571E+03, 1.2588E+03,
     + 1.2605E+03, 1.2621E+03, 1.2638E+03, 1.2655E+03, 1.2671E+03,
     +  1.2688E+03, 1.2705E+03, 1.2721E+03, 1.2738E+03, 1.2755E+03,
     + 1.2771E+03, 1.2788E+03, 1.2805E+03, 1.2822E+03, 1.2839E+03,
     +  1.2855E+03, 1.2872E+03, 1.2889E+03, 1.2906E+03, 1.2923E+03,
     + 1.2939E+03, 1.2956E+03, 1.2973E+03, 1.2990E+03, 1.3007E+03,
     +  1.3024E+03, 1.3041E+03, 1.3058E+03, 1.3075E+03, 1.3092E+03,
     + 1.3109E+03, 1.3125E+03, 1.3142E+03, 1.3159E+03, 1.3176E+03,
     +  1.3193E+03, 1.3210E+03, 1.3228E+03, 1.3245E+03, 1.3262E+03,
     + 1.3279E+03, 1.3296E+03, 1.3313E+03, 1.3330E+03, 1.3347E+03,
     +  1.3364E+03, 1.3381E+03, 1.3399E+03, 1.3416E+03, 1.3433E+03,
     + 1.3450E+03, 1.3467E+03, 1.3484E+03, 1.3502E+03, 1.3519E+03,
     +  1.3536E+03, 1.3553E+03, 1.3571E+03, 1.3588E+03, 1.3605E+03,
     + 1.3623E+03, 1.3640E+03, 1.3657E+03, 1.3675E+03, 1.3692E+03,
     +  1.3709E+03, 1.3727E+03, 1.3744E+03, 1.3761E+03, 1.3779E+03,
     + 1.3796E+03, 1.3814E+03, 1.3831E+03, 1.3849E+03, 1.3866E+03,
     +  1.3883E+03, 1.3901E+03, 1.3918E+03, 1.3936E+03, 1.3953E+03,
     + 1.3971E+03, 1.3989E+03, 1.4006E+03, 1.4024E+03, 1.4041E+03,
     +  1.4059E+03, 1.4076E+03, 1.4094E+03, 1.4112E+03, 1.4129E+03,
     + 1.4147E+03, 1.4165E+03, 1.4182E+03, 1.4200E+03, 1.4218E+03,
     +  1.4235E+03, 1.4253E+03, 1.4271E+03, 1.4288E+03, 1.4306E+03,
     + 1.4324E+03, 1.4342E+03, 1.4360E+03, 1.4377E+03, 1.4395E+03,
     +  1.4413E+03, 1.4431E+03, 1.4449E+03, 1.4466E+03, 1.4484E+03,
     + 1.4502E+03, 1.4520E+03, 1.4538E+03, 1.4556E+03, 1.4574E+03,
     +  1.4592E+03, 1.4610E+03, 1.4628E+03, 1.4646E+03, 1.4664E+03,
     + 1.4682E+03, 1.4700E+03, 1.4718E+03, 1.4736E+03, 1.4754E+03,
     +  1.4772E+03, 1.4790E+03, 1.4808E+03, 1.4826E+03, 1.4844E+03,
     + 1.4862E+03, 1.4880E+03, 1.4898E+03, 1.4917E+03, 1.4935E+03,
     +  1.4953E+03, 1.4971E+03, 1.4989E+03, 1.5007E+03, 1.5026E+03,
     + 1.5044E+03, 1.5062E+03, 1.5080E+03, 1.5099E+03, 1.5117E+03,
     +  1.5135E+03, 1.5154E+03, 1.5172E+03, 1.5190E+03, 1.5208E+03,
     + 1.5227E+03, 1.5245E+03, 1.5264E+03, 1.5282E+03, 1.5300E+03,
     +  1.5319E+03, 1.5337E+03, 1.5356E+03, 1.5374E+03, 1.5392E+03,
     + 1.5411E+03, 1.5429E+03, 1.5448E+03, 1.5466E+03, 1.5485E+03,
     +  1.5503E+03, 1.5522E+03, 1.5540E+03, 1.5559E+03, 1.5578E+03,
     + 1.5596E+03, 1.5615E+03, 1.5633E+03, 1.5652E+03, 1.5671E+03,
     +  1.5689E+03, 1.5708E+03, 1.5727E+03, 1.5745E+03, 1.5764E+03,
     + 1.5783E+03, 1.5801E+03, 1.5820E+03, 1.5839E+03, 1.5857E+03,
     +  1.5876E+03, 1.5895E+03, 1.5914E+03, 1.5933E+03, 1.5951E+03,
     + 1.5970E+03, 1.5989E+03, 1.6008E+03, 1.6027E+03, 1.6046E+03,
     +  1.6064E+03, 1.6083E+03, 1.6102E+03, 1.6121E+03, 1.6140E+03,
     + 1.6159E+03, 1.6178E+03, 1.6197E+03, 1.6216E+03, 1.6235E+03,
     +  1.6254E+03, 1.6273E+03, 1.6292E+03, 1.6311E+03, 1.6330E+03,
     + 1.6349E+03, 1.6368E+03, 1.6387E+03, 1.6406E+03, 1.6425E+03,
     +  1.6444E+03, 1.6463E+03, 1.6483E+03, 1.6502E+03, 1.6521E+03,
     + 1.6540E+03, 1.6559E+03, 1.6578E+03, 1.6598E+03, 1.6617E+03,
     +  1.6636E+03, 1.6655E+03, 1.6675E+03, 1.6694E+03, 1.6713E+03,
     + 1.6732E+03, 1.6752E+03, 1.6771E+03, 1.6790E+03, 1.6810E+03,
     +  1.6829E+03, 1.6848E+03, 1.6868E+03, 1.6887E+03, 1.6907E+03,
     + 1.6926E+03, 1.6945E+03, 1.6965E+03, 1.6984E+03, 1.7004E+03,
     +  1.7023E+03, 1.7043E+03, 1.7062E+03, 1.7082E+03, 1.7101E+03,
     + 1.7121E+03, 1.7140E+03, 1.7160E+03, 1.7179E+03, 1.7199E+03,
     +  1.7219E+03, 1.7238E+03, 1.7258E+03, 1.7277E+03, 1.7297E+03,
     + 1.7317E+03, 1.7336E+03, 1.7356E+03, 1.7376E+03, 1.7395E+03,
     +  1.7415E+03, 1.7435E+03, 1.7455E+03, 1.7474E+03, 1.7494E+03,
     + 1.7514E+03, 1.7534E+03, 1.7553E+03, 1.7573E+03, 1.7593E+03,
     +  1.7613E+03, 1.7633E+03, 1.7653E+03, 1.7672E+03, 1.7692E+03,
     + 1.7712E+03, 1.7732E+03, 1.7752E+03, 1.7772E+03, 1.7792E+03,
     +  1.7812E+03, 1.7832E+03, 1.7852E+03, 1.7872E+03, 1.7892E+03,
     + 1.7912E+03, 1.7932E+03, 1.7952E+03, 1.7972E+03, 1.7992E+03,
     +  1.8012E+03, 1.8032E+03, 1.8052E+03, 1.8072E+03, 1.8092E+03,
     + 1.8113E+03, 1.8133E+03, 1.8153E+03, 1.8173E+03, 1.8193E+03,
     +  1.8213E+03, 1.8234E+03, 1.8254E+03, 1.8274E+03, 1.8294E+03,
     + 1.8315E+03, 1.8335E+03, 1.8355E+03, 1.8375E+03, 1.8396E+03,
     +  1.8416E+03, 1.8436E+03, 1.8457E+03, 1.8477E+03, 1.8497E+03,
     + 1.8518E+03, 1.8538E+03, 1.8558E+03, 1.8579E+03, 1.8599E+03,
     +  1.8620E+03, 1.8640E+03, 1.8661E+03, 1.8681E+03, 1.8702E+03,
     + 1.8722E+03, 1.8743E+03, 1.8763E+03, 1.8784E+03, 1.8804E+03,
     +  1.8825E+03, 1.8845E+03, 1.8866E+03, 1.8887E+03, 1.8907E+03,
     + 1.8928E+03, 1.8948E+03, 1.8969E+03, 1.8990E+03, 1.9010E+03,
     +  1.9031E+03, 1.9052E+03, 1.9072E+03, 1.9093E+03, 1.9114E+03,
     + 1.9135E+03, 1.9155E+03, 1.9176E+03, 1.9197E+03, 1.9218E+03,
     +  1.9239E+03, 1.9259E+03, 1.9280E+03, 1.9301E+03, 1.9322E+03,
     + 1.9343E+03, 1.9364E+03, 1.9384E+03, 1.9405E+03, 1.9426E+03,
     +  1.9447E+03, 1.9468E+03, 1.9489E+03, 1.9510E+03, 1.9531E+03,
     + 1.9552E+03, 1.9573E+03, 1.9594E+03, 1.9615E+03, 1.9636E+03,
     +  1.9657E+03, 1.9678E+03, 1.9699E+03, 1.9720E+03, 1.9741E+03,
     + 1.9763E+03, 1.9784E+03, 1.9805E+03, 1.9826E+03, 1.9847E+03,
     +  1.9868E+03, 1.9890E+03, 1.9911E+03, 1.9932E+03, 1.9953E+03,
     + 1.9974E+03, 1.9996E+03, 2.0017E+03, 2.0038E+03, 2.0059E+03,
     +  2.0081E+03, 2.0102E+03, 2.0123E+03, 2.0145E+03, 2.0166E+03,
     + 2.0187E+03, 2.0209E+03, 2.0230E+03, 2.0252E+03, 2.0273E+03,
     +  2.0294E+03, 2.0316E+03, 2.0337E+03, 2.0359E+03, 2.0380E+03,
     + 2.0402E+03, 2.0423E+03, 2.0445E+03, 2.0466E+03, 2.0488E+03,
     +  2.0509E+03, 2.0531E+03, 2.0552E+03, 2.0574E+03, 2.0596E+03,
     + 2.0617E+03, 2.0639E+03, 2.0660E+03, 2.0682E+03, 2.0704E+03,
     +  2.0725E+03, 2.0747E+03, 2.0769E+03, 2.0790E+03, 2.0812E+03,
     + 2.0834E+03, 2.0856E+03, 2.0877E+03, 2.0899E+03, 2.0921E+03,
     +  2.0943E+03, 2.0964E+03, 2.0986E+03, 2.1008E+03, 2.1030E+03,
     + 2.1052E+03, 2.1074E+03, 2.1096E+03, 2.1117E+03, 2.1139E+03,
     +  2.1161E+03, 2.1183E+03, 2.1205E+03, 2.1227E+03, 2.1249E+03,
     + 2.1271E+03, 2.1293E+03, 2.1315E+03, 2.1337E+03, 2.1359E+03,
     +  2.1381E+03, 2.1403E+03, 2.1425E+03, 2.1447E+03, 2.1469E+03,
     + 2.1492E+03, 2.1514E+03, 2.1536E+03, 2.1558E+03, 2.1580E+03,
     +  2.1602E+03, 2.1624E+03, 2.1647E+03, 2.1669E+03, 2.1691E+03,
     + 2.1713E+03, 2.1736E+03, 2.1758E+03, 2.1780E+03, 2.1802E+03,
     +  2.1825E+03, 2.1847E+03, 2.1869E+03, 2.1892E+03, 2.1914E+03,
     + 2.1936E+03, 2.1959E+03, 2.1981E+03, 2.2003E+03, 2.2026E+03,
     +  2.2048E+03, 2.2071E+03, 2.2093E+03, 2.2116E+03, 2.2138E+03,
     + 2.2160E+03, 2.2183E+03, 2.2205E+03, 2.2228E+03, 2.2251E+03,
     +  2.2273E+03, 2.2296E+03, 2.2318E+03, 2.2341E+03, 2.2363E+03,
     + 2.2386E+03, 2.2409E+03, 2.2431E+03, 2.2454E+03, 2.2477E+03,
     +  2.2499E+03, 2.2522E+03, 2.2545E+03, 2.2567E+03, 2.2590E+03,
     + 2.2613E+03, 2.2635E+03, 2.2658E+03, 2.2681E+03, 2.2704E+03,
     +  2.2727E+03, 2.2749E+03, 2.2772E+03, 2.2795E+03, 2.2818E+03,
     + 2.2841E+03, 2.2864E+03, 2.2887E+03, 2.2909E+03, 2.2932E+03,
     +  2.2955E+03, 2.2978E+03, 2.3001E+03, 2.3024E+03, 2.3047E+03,
     + 2.3070E+03, 2.3093E+03, 2.3116E+03, 2.3139E+03, 2.3162E+03,
     +  2.3185E+03, 2.3208E+03, 2.3231E+03, 2.3254E+03, 2.3277E+03,
     + 2.3301E+03, 2.3324E+03, 2.3347E+03, 2.3370E+03, 2.3393E+03,
     +  2.3416E+03, 2.3440E+03, 2.3463E+03, 2.3486E+03, 2.3509E+03,
     + 2.3532E+03, 2.3556E+03, 2.3579E+03, 2.3602E+03, 2.3626E+03,
     +  2.3649E+03, 2.3672E+03, 2.3695E+03, 2.3719E+03, 2.3742E+03,
     + 2.3766E+03, 2.3789E+03, 2.3812E+03, 2.3836E+03, 2.3859E+03,
     +  2.3883E+03, 2.3906E+03, 2.3929E+03, 2.3953E+03, 2.3976E+03,
     + 2.4000E+03, 2.4023E+03, 2.4047E+03, 2.4070E+03, 2.4094E+03,
     +  2.4118E+03, 2.4141E+03, 2.4165E+03, 2.4188E+03, 2.4212E+03,
     + 2.4236E+03, 2.4259E+03, 2.4283E+03, 2.4306E+03, 2.4330E+03,
     +  2.4354E+03, 2.4378E+03, 2.4401E+03, 2.4425E+03, 2.4449E+03,
     + 2.4472E+03, 2.4496E+03, 2.4520E+03, 2.4544E+03, 2.4568E+03,
     +  2.4591E+03, 2.4615E+03, 2.4639E+03, 2.4663E+03, 2.4687E+03,
     + 2.4711E+03, 2.4735E+03, 2.4758E+03, 2.4782E+03, 2.4806E+03,
     +  2.4830E+03, 2.4854E+03, 2.4878E+03, 2.4902E+03, 2.4926E+03,
     + 2.4950E+03, 2.4974E+03, 2.4998E+03, 2.5022E+03, 2.5046E+03,
     +  2.5070E+03, 2.5094E+03, 2.5119E+03, 2.5143E+03, 2.5167E+03,
     + 2.5191E+03, 2.5215E+03, 2.5239E+03, 2.5263E+03, 2.5288E+03,
     +  2.5312E+03, 2.5336E+03, 2.5360E+03, 2.5384E+03, 2.5409E+03,
     + 2.5433E+03, 2.5457E+03, 2.5482E+03, 2.5506E+03, 2.5530E+03,
     +  2.5554E+03, 2.5579E+03, 2.5603E+03, 2.5628E+03, 2.5652E+03,
     + 2.5676E+03, 2.5701E+03, 2.5725E+03, 2.5750E+03, 2.5774E+03,
     +  2.5798E+03, 2.5823E+03, 2.5847E+03, 2.5872E+03, 2.5896E+03,
     + 2.5921E+03, 2.5946E+03, 2.5970E+03, 2.5995E+03, 2.6019E+03,
     +  2.6044E+03, 2.6068E+03, 2.6093E+03, 2.6118E+03, 2.6142E+03,
     + 2.6167E+03, 2.6192E+03, 2.6216E+03, 2.6241E+03, 2.6266E+03,
     +  2.6290E+03, 2.6315E+03, 2.6340E+03, 2.6365E+03, 2.6389E+03,
     + 2.6414E+03, 2.6439E+03, 2.6464E+03, 2.6489E+03, 2.6513E+03,
     +  2.6538E+03, 2.6563E+03, 2.6588E+03, 2.6613E+03, 2.6638E+03,
     + 2.6663E+03, 2.6688E+03, 2.6713E+03, 2.6737E+03, 2.6762E+03,
     +  2.6787E+03, 2.6812E+03, 2.6837E+03, 2.6862E+03, 2.6887E+03,
     + 2.6912E+03, 2.6938E+03, 2.6963E+03, 2.6988E+03, 2.7013E+03,
     +  2.7038E+03, 2.7063E+03, 2.7088E+03, 2.7113E+03, 2.7138E+03,
     + 2.7164E+03, 2.7189E+03, 2.7214E+03, 2.7239E+03, 2.7264E+03,
     +  2.7290E+03, 2.7315E+03, 2.7340E+03, 2.7365E+03, 2.7391E+03,
     + 2.7416E+03, 2.7441E+03, 2.7467E+03, 2.7492E+03, 2.7517E+03,
     +  2.7543E+03, 2.7568E+03, 2.7593E+03, 2.7619E+03, 2.7644E+03,
     + 2.7670E+03, 2.7695E+03, 2.7721E+03, 2.7746E+03, 2.7772E+03,
     +  2.7797E+03, 2.7823E+03, 2.7848E+03, 2.7874E+03, 2.7899E+03,
     + 2.7925E+03, 2.7950E+03, 2.7976E+03, 2.8002E+03, 2.8027E+03,
     +  2.8053E+03, 2.8078E+03, 2.8104E+03, 2.8130E+03, 2.8155E+03,
     + 2.8181E+03, 2.8207E+03, 2.8233E+03, 2.8258E+03, 2.8284E+03,
     +  2.8310E+03, 2.8336E+03, 2.8361E+03, 2.8387E+03, 2.8413E+03,
     + 2.8439E+03, 2.8465E+03, 2.8490E+03, 2.8516E+03, 2.8542E+03,
     +  2.8568E+03, 2.8594E+03, 2.8620E+03, 2.8646E+03, 2.8672E+03,
     + 2.8698E+03, 2.8724E+03, 2.8750E+03, 2.8776E+03, 2.8802E+03,
     +  2.8828E+03, 2.8854E+03, 2.8880E+03, 2.8906E+03, 2.8932E+03,
     + 2.8958E+03, 2.8984E+03, 2.9010E+03, 2.9036E+03, 2.9062E+03,
     +  2.9089E+03, 2.9115E+03, 2.9141E+03, 2.9167E+03, 2.9193E+03,
     + 2.9220E+03, 2.9246E+03, 2.9272E+03, 2.9298E+03, 2.9325E+03,
     +  2.9351E+03, 2.9377E+03, 2.9404E+03, 2.9430E+03, 2.9456E+03,
     + 2.9483E+03, 2.9509E+03, 2.9535E+03, 2.9562E+03, 2.9588E+03,
     +  2.9614E+03, 2.9641E+03, 2.9667E+03, 2.9694E+03, 2.9720E+03,
     + 2.9747E+03, 2.9773E+03, 2.9800E+03, 2.9826E+03, 2.9853E+03,
     +  2.9879E+03, 2.9906E+03, 2.9932E+03, 2.9959E+03, 2.9986E+03,
     + 3.0012E+03, 3.0039E+03, 3.0066E+03, 3.0092E+03, 3.0119E+03,
     +  3.0146E+03, 3.0172E+03, 3.0199E+03, 3.0226E+03, 3.0252E+03,
     + 3.0279E+03, 3.0306E+03, 3.0333E+03, 3.0360E+03, 3.0386E+03,
     +  3.0413E+03, 3.0440E+03, 3.0467E+03, 3.0494E+03, 3.0521E+03,
     + 3.0547E+03, 3.0574E+03, 3.0601E+03, 3.0628E+03, 3.0655E+03,
     +  3.0682E+03, 3.0709E+03, 3.0736E+03, 3.0763E+03, 3.0790E+03,
     + 3.0817E+03, 3.0844E+03, 3.0871E+03, 3.0898E+03, 3.0925E+03,
     +  3.0952E+03, 3.0979E+03, 3.1006E+03, 3.1033E+03, 3.1061E+03,
     + 3.1088E+03, 3.1115E+03, 3.1142E+03, 3.1169E+03, 3.1196E+03,
     +  3.1224E+03, 3.1251E+03, 3.1278E+03, 3.1305E+03, 3.1333E+03,
     + 3.1360E+03, 3.1387E+03, 3.1415E+03, 3.1442E+03, 3.1469E+03,
     +  3.1497E+03, 3.1524E+03, 3.1551E+03, 3.1579E+03, 3.1606E+03,
     + 3.1633E+03, 3.1661E+03, 3.1688E+03, 3.1716E+03, 3.1743E+03,
     +  3.1771E+03, 3.1798E+03, 3.1826E+03, 3.1853E+03, 3.1881E+03,
     + 3.1908E+03, 3.1936E+03, 3.1963E+03, 3.1991E+03, 3.2019E+03,
     +  3.2046E+03, 3.2074E+03, 3.2102E+03, 3.2129E+03, 3.2157E+03,
     + 3.2185E+03, 3.2212E+03, 3.2240E+03, 3.2268E+03, 3.2295E+03,
     +  3.2323E+03, 3.2351E+03, 3.2379E+03, 3.2406E+03, 3.2434E+03,
     + 3.2462E+03, 3.2490E+03, 3.2518E+03, 3.2546E+03, 3.2573E+03,
     +  3.2601E+03, 3.2629E+03, 3.2657E+03, 3.2685E+03, 3.2713E+03,
     + 3.2741E+03, 3.2769E+03, 3.2797E+03, 3.2825E+03, 3.2853E+03,
     +  3.2881E+03, 3.2909E+03, 3.2937E+03, 3.2965E+03, 3.2993E+03,
     + 3.3021E+03, 3.3049E+03, 3.3077E+03, 3.3105E+03, 3.3134E+03,
     +  3.3162E+03, 3.3190E+03, 3.3218E+03, 3.3246E+03, 3.3274E+03,
     + 3.3303E+03, 3.3331E+03, 3.3359E+03, 3.3387E+03, 3.3416E+03,
     +  3.3444E+03, 3.3472E+03, 3.3501E+03, 3.3529E+03, 3.3557E+03,
     + 3.3586E+03, 3.3614E+03, 3.3642E+03, 3.3671E+03, 3.3699E+03,
     +  3.3728E+03, 3.3756E+03, 3.3784E+03, 3.3813E+03, 3.3841E+03,
     + 3.3870E+03, 3.3898E+03, 3.3927E+03, 3.3955E+03, 3.3984E+03,
     +  3.4012E+03, 3.4041E+03, 3.4070E+03, 3.4098E+03, 3.4127E+03,
     + 3.4155E+03, 3.4184E+03, 3.4213E+03, 3.4241E+03, 3.4270E+03,
     +  3.4299E+03, 3.4327E+03, 3.4356E+03, 3.4385E+03, 3.4414E+03,
     + 3.4442E+03, 3.4471E+03, 3.4500E+03, 3.4529E+03, 3.4557E+03,
     +  3.4586E+03, 3.4615E+03, 3.4644E+03, 3.4673E+03, 3.4702E+03,
     + 3.4731E+03, 3.4760E+03, 3.4788E+03, 3.4817E+03, 3.4846E+03,
     +  3.4875E+03, 3.4904E+03, 3.4933E+03, 3.4962E+03, 3.4991E+03,
     + 3.5020E+03, 3.5049E+03, 3.5078E+03, 3.5107E+03, 3.5136E+03,
     +  3.5166E+03, 3.5195E+03, 3.5224E+03, 3.5253E+03, 3.5282E+03,
     + 3.5311E+03, 3.5340E+03, 3.5370E+03, 3.5399E+03, 3.5428E+03,
     +  3.5457E+03, 3.5486E+03, 3.5516E+03, 3.5545E+03, 3.5574E+03,
     + 3.5604E+03, 3.5633E+03, 3.5662E+03, 3.5692E+03, 3.5721E+03,
     +  3.5750E+03, 3.5780E+03, 3.5809E+03, 3.5838E+03, 3.5868E+03,
     + 3.5897E+03, 3.5927E+03, 3.5956E+03, 3.5986E+03, 3.6015E+03,
     +  3.6045E+03, 3.6074E+03, 3.6104E+03, 3.6133E+03, 3.6163E+03,
     + 3.6192E+03, 3.6222E+03, 3.6252E+03, 3.6281E+03, 3.6311E+03,
     +  3.6340E+03, 3.6370E+03, 3.6400E+03, 3.6429E+03, 3.6459E+03,
     + 3.6489E+03, 3.6518E+03, 3.6548E+03, 3.6578E+03, 3.6608E+03,
     +  3.6637E+03, 3.6667E+03, 3.6697E+03, 3.6727E+03, 3.6757E+03,
     + 3.6787E+03, 3.6816E+03, 3.6846E+03, 3.6876E+03, 3.6906E+03,
     +  3.6936E+03, 3.6966E+03, 3.6996E+03, 3.7026E+03, 3.7056E+03,
     + 3.7086E+03, 3.7116E+03, 3.7146E+03, 3.7176E+03, 3.7206E+03,
     +  3.7236E+03, 3.7266E+03, 3.7296E+03, 3.7326E+03, 3.7356E+03,
     + 3.7386E+03, 3.7416E+03, 3.7446E+03, 3.7477E+03, 3.7507E+03,
     +  3.7537E+03, 3.7567E+03, 3.7597E+03, 3.7628E+03, 3.7658E+03,
     + 3.7688E+03, 3.7718E+03, 3.7749E+03, 3.7779E+03, 3.7809E+03,
     +  3.7839E+03, 3.7870E+03, 3.7900E+03, 3.7930E+03, 3.7961E+03,
     + 3.7991E+03, 3.8022E+03, 3.8052E+03, 3.8082E+03, 3.8113E+03,
     +  3.8143E+03, 3.8174E+03, 3.8204E+03, 3.8235E+03, 3.8265E+03,
     + 3.8296E+03, 3.8326E+03, 3.8357E+03, 3.8387E+03, 3.8418E+03,
     +  3.8449E+03, 3.8479E+03, 3.8510E+03, 3.8540E+03, 3.8571E+03,
     + 3.8602E+03, 3.8632E+03, 3.8663E+03, 3.8694E+03, 3.8725E+03,
     +  3.8755E+03, 3.8786E+03, 3.8817E+03, 3.8848E+03, 3.8878E+03,
     + 3.8909E+03, 3.8940E+03, 3.8971E+03, 3.9002E+03, 3.9032E+03,
     +  3.9063E+03, 3.9094E+03, 3.9125E+03, 3.9156E+03, 3.9187E+03,
     + 3.9218E+03, 3.9249E+03, 3.9280E+03, 3.9311E+03, 3.9342E+03,
     +  3.9373E+03, 3.9404E+03, 3.9435E+03, 3.9466E+03, 3.9497E+03,
     + 3.9528E+03, 3.9559E+03, 3.9590E+03, 3.9621E+03, 3.9652E+03,
     +  3.9683E+03, 3.9715E+03, 3.9746E+03, 3.9777E+03, 3.9808E+03,
     + 3.9839E+03, 3.9871E+03, 3.9902E+03, 3.9933E+03, 3.9964E+03,
     +  3.9996E+03, 4.0027E+03, 4.0058E+03, 4.0090E+03, 4.0121E+03,
     + 4.0152E+03, 4.0184E+03, 4.0215E+03, 4.0246E+03, 4.0278E+03,
     +  4.0309E+03, 4.0341E+03, 4.0372E+03, 4.0403E+03, 4.0435E+03,
     + 4.0466E+03, 4.0498E+03, 4.0529E+03, 4.0561E+03, 4.0592E+03,
     +  4.0624E+03, 4.0656E+03, 4.0687E+03, 4.0719E+03, 4.0750E+03,
     + 4.0782E+03, 4.0814E+03, 4.0845E+03, 4.0877E+03, 4.0909E+03,
     +  4.0940E+03, 4.0972E+03, 4.1004E+03, 4.1035E+03, 4.1067E+03,
     + 4.1099E+03, 4.1131E+03, 4.1162E+03, 4.1194E+03, 4.1226E+03,
     +  4.1258E+03, 4.1290E+03, 4.1322E+03, 4.1353E+03, 4.1385E+03,
     + 4.1417E+03, 4.1449E+03, 4.1481E+03, 4.1513E+03, 4.1545E+03,
     +  4.1577E+03, 4.1609E+03, 4.1641E+03, 4.1673E+03, 4.1705E+03,
     + 4.1737E+03, 4.1769E+03, 4.1801E+03, 4.1833E+03, 4.1865E+03,
     +  4.1897E+03, 4.1929E+03, 4.1962E+03, 4.1994E+03, 4.2026E+03,
     + 4.2058E+03, 4.2090E+03, 4.2122E+03, 4.2155E+03, 4.2187E+03,
     +  4.2219E+03, 4.2251E+03, 4.2284E+03, 4.2316E+03, 4.2348E+03,
     + 4.2381E+03, 4.2413E+03, 4.2445E+03, 4.2478E+03, 4.2510E+03,
     +  4.2542E+03, 4.2575E+03, 4.2607E+03, 4.2639E+03, 4.2672E+03,
     + 4.2704E+03, 4.2737E+03, 4.2769E+03, 4.2802E+03, 4.2834E+03,
     +  4.2867E+03, 4.2899E+03, 4.2932E+03, 4.2964E+03, 4.2997E+03,
     + 4.3030E+03, 4.3062E+03, 4.3095E+03, 4.3127E+03, 4.3160E+03,
     +  4.3193E+03, 4.3225E+03, 4.3258E+03, 4.3291E+03, 4.3324E+03,
     + 4.3356E+03, 4.3389E+03, 4.3422E+03, 4.3455E+03, 4.3487E+03,
     +  4.3520E+03, 4.3553E+03, 4.3586E+03, 4.3619E+03, 4.3651E+03,
     + 4.3684E+03, 4.3717E+03, 4.3750E+03, 4.3783E+03, 4.3816E+03,
     +  4.3849E+03, 4.3882E+03, 4.3915E+03, 4.3948E+03, 4.3981E+03,
     + 4.4014E+03, 4.4047E+03, 4.4080E+03, 4.4113E+03, 4.4146E+03,
     +  4.4179E+03, 4.4212E+03, 4.4245E+03, 4.4278E+03, 4.4311E+03,
     + 4.4344E+03, 4.4378E+03, 4.4411E+03, 4.4444E+03, 4.4477E+03,
     +  4.4510E+03, 4.4544E+03, 4.4577E+03, 4.4610E+03, 4.4643E+03,
     + 4.4677E+03, 4.4710E+03, 4.4743E+03, 4.4777E+03, 4.4810E+03,
     +  4.4843E+03, 4.4877E+03, 4.4910E+03, 4.4943E+03, 4.4977E+03,
     + 4.5010E+03, 4.5044E+03, 4.5077E+03, 4.5111E+03, 4.5144E+03,
     +  4.5178E+03, 4.5211E+03, 4.5245E+03, 4.5278E+03, 4.5312E+03,
     + 4.5345E+03, 4.5379E+03, 4.5412E+03, 4.5446E+03, 4.5480E+03,
     +  4.5513E+03, 4.5547E+03, 4.5580E+03, 4.5614E+03, 4.5648E+03,
     + 4.5682E+03, 4.5715E+03, 4.5749E+03, 4.5783E+03, 4.5816E+03,
     +  4.5850E+03, 4.5884E+03, 4.5918E+03, 4.5952E+03, 4.5985E+03,
     + 4.6019E+03, 4.6053E+03, 4.6087E+03, 4.6121E+03, 4.6155E+03,
     +  4.6189E+03, 4.6223E+03, 4.6257E+03, 4.6290E+03, 4.6324E+03,
     + 4.6358E+03, 4.6392E+03, 4.6426E+03, 4.6460E+03, 4.6494E+03,
     +  4.6528E+03, 4.6563E+03, 4.6597E+03, 4.6631E+03, 4.6665E+03,
     + 4.6699E+03, 4.6733E+03, 4.6767E+03, 4.6801E+03, 4.6836E+03,
     +  4.6870E+03, 4.6904E+03, 4.6938E+03, 4.6972E+03, 4.7007E+03,
     + 4.7041E+03, 4.7075E+03, 4.7109E+03, 4.7144E+03, 4.7178E+03,
     +  4.7212E+03, 4.7247E+03, 4.7281E+03, 4.7315E+03, 4.7350E+03,
     + 4.7384E+03, 4.7419E+03, 4.7453E+03, 4.7487E+03, 4.7522E+03,
     +  4.7556E+03, 4.7591E+03, 4.7625E+03, 4.7660E+03, 4.7694E+03/
      data (QofT( 1,J),J=2409,4510)/ 4.7729E+03, 4.7763E+03,
     +  4.7798E+03, 4.7833E+03, 4.7867E+03, 4.7902E+03, 4.7936E+03,
     +  4.7971E+03,
     + 4.8006E+03, 4.8040E+03, 4.8075E+03, 4.8110E+03, 4.8144E+03,
     +  4.8179E+03, 4.8214E+03, 4.8249E+03, 4.8283E+03, 4.8318E+03,
     + 4.8353E+03, 4.8388E+03, 4.8423E+03, 4.8457E+03, 4.8492E+03,
     +  4.8527E+03, 4.8562E+03, 4.8597E+03, 4.8632E+03, 4.8667E+03,
     + 4.8702E+03, 4.8737E+03, 4.8771E+03, 4.8806E+03, 4.8841E+03,
     +  4.8876E+03, 4.8911E+03, 4.8946E+03, 4.8981E+03, 4.9017E+03,
     + 4.9052E+03, 4.9087E+03, 4.9122E+03, 4.9157E+03, 4.9192E+03,
     +  4.9227E+03, 4.9262E+03, 4.9297E+03, 4.9333E+03, 4.9368E+03,
     + 4.9403E+03, 4.9438E+03, 4.9473E+03, 4.9509E+03, 4.9544E+03,
     +  4.9579E+03, 4.9615E+03, 4.9650E+03, 4.9685E+03, 4.9721E+03,
     + 4.9756E+03, 4.9791E+03, 4.9827E+03, 4.9862E+03, 4.9897E+03,
     +  4.9933E+03, 4.9968E+03, 5.0004E+03, 5.0039E+03, 5.0075E+03,
     + 5.0110E+03, 5.0146E+03, 5.0181E+03, 5.0217E+03, 5.0252E+03,
     +  5.0288E+03, 5.0323E+03, 5.0359E+03, 5.0394E+03, 5.0430E+03,
     + 5.0466E+03, 5.0501E+03, 5.0537E+03, 5.0573E+03, 5.0608E+03,
     +  5.0644E+03, 5.0680E+03, 5.0716E+03, 5.0751E+03, 5.0787E+03,
     + 5.0823E+03, 5.0859E+03, 5.0894E+03, 5.0930E+03, 5.0966E+03,
     +  5.1002E+03, 5.1038E+03, 5.1074E+03, 5.1109E+03, 5.1145E+03,
     + 5.1181E+03, 5.1217E+03, 5.1253E+03, 5.1289E+03, 5.1325E+03,
     +  5.1361E+03, 5.1397E+03, 5.1433E+03, 5.1469E+03, 5.1505E+03,
     + 5.1541E+03, 5.1577E+03, 5.1613E+03, 5.1649E+03, 5.1685E+03,
     +  5.1722E+03, 5.1758E+03, 5.1794E+03, 5.1830E+03, 5.1866E+03,
     + 5.1902E+03, 5.1939E+03, 5.1975E+03, 5.2011E+03, 5.2047E+03,
     +  5.2083E+03, 5.2120E+03, 5.2156E+03, 5.2192E+03, 5.2229E+03,
     + 5.2265E+03, 5.2301E+03, 5.2338E+03, 5.2374E+03, 5.2410E+03,
     +  5.2447E+03, 5.2483E+03, 5.2520E+03, 5.2556E+03, 5.2593E+03,
     + 5.2629E+03, 5.2666E+03, 5.2702E+03, 5.2739E+03, 5.2775E+03,
     +  5.2812E+03, 5.2848E+03, 5.2885E+03, 5.2921E+03, 5.2958E+03,
     + 5.2994E+03, 5.3031E+03, 5.3068E+03, 5.3104E+03, 5.3141E+03,
     +  5.3178E+03, 5.3214E+03, 5.3251E+03, 5.3288E+03, 5.3325E+03,
     + 5.3361E+03, 5.3398E+03, 5.3435E+03, 5.3472E+03, 5.3509E+03,
     +  5.3545E+03, 5.3582E+03, 5.3619E+03, 5.3656E+03, 5.3693E+03,
     + 5.3730E+03, 5.3767E+03, 5.3803E+03, 5.3840E+03, 5.3877E+03,
     +  5.3914E+03, 5.3951E+03, 5.3988E+03, 5.4025E+03, 5.4062E+03,
     + 5.4099E+03, 5.4136E+03, 5.4173E+03, 5.4211E+03, 5.4248E+03,
     +  5.4285E+03, 5.4322E+03, 5.4359E+03, 5.4396E+03, 5.4433E+03,
     + 5.4470E+03, 5.4508E+03, 5.4545E+03, 5.4582E+03, 5.4619E+03,
     +  5.4656E+03, 5.4694E+03, 5.4731E+03, 5.4768E+03, 5.4806E+03,
     + 5.4843E+03, 5.4880E+03, 5.4918E+03, 5.4955E+03, 5.4992E+03,
     +  5.5030E+03, 5.5067E+03, 5.5104E+03, 5.5142E+03, 5.5179E+03,
     + 5.5217E+03, 5.5254E+03, 5.5292E+03, 5.5329E+03, 5.5367E+03,
     +  5.5404E+03, 5.5442E+03, 5.5479E+03, 5.5517E+03, 5.5555E+03,
     + 5.5592E+03, 5.5630E+03, 5.5667E+03, 5.5705E+03, 5.5743E+03,
     +  5.5780E+03, 5.5818E+03, 5.5856E+03, 5.5893E+03, 5.5931E+03,
     + 5.5969E+03, 5.6007E+03, 5.6044E+03, 5.6082E+03, 5.6120E+03,
     +  5.6158E+03, 5.6196E+03, 5.6233E+03, 5.6271E+03, 5.6309E+03,
     + 5.6347E+03, 5.6385E+03, 5.6423E+03, 5.6461E+03, 5.6499E+03,
     +  5.6537E+03, 5.6575E+03, 5.6612E+03, 5.6650E+03, 5.6688E+03,
     + 5.6727E+03, 5.6765E+03, 5.6803E+03, 5.6841E+03, 5.6879E+03,
     +  5.6917E+03, 5.6955E+03, 5.6993E+03, 5.7031E+03, 5.7069E+03,
     + 5.7107E+03, 5.7146E+03, 5.7184E+03, 5.7222E+03, 5.7260E+03,
     +  5.7298E+03, 5.7337E+03, 5.7375E+03, 5.7413E+03, 5.7452E+03,
     + 5.7490E+03, 5.7528E+03, 5.7566E+03, 5.7605E+03, 5.7643E+03,
     +  5.7682E+03, 5.7720E+03, 5.7758E+03, 5.7797E+03, 5.7835E+03,
     + 5.7874E+03, 5.7912E+03, 5.7950E+03, 5.7989E+03, 5.8027E+03,
     +  5.8066E+03, 5.8105E+03, 5.8143E+03, 5.8182E+03, 5.8220E+03,
     + 5.8259E+03, 5.8297E+03, 5.8336E+03, 5.8375E+03, 5.8413E+03,
     +  5.8452E+03, 5.8491E+03, 5.8529E+03, 5.8568E+03, 5.8607E+03,
     + 5.8645E+03, 5.8684E+03, 5.8723E+03, 5.8762E+03, 5.8800E+03,
     +  5.8839E+03, 5.8878E+03, 5.8917E+03, 5.8956E+03, 5.8994E+03,
     + 5.9033E+03, 5.9072E+03, 5.9111E+03, 5.9150E+03, 5.9189E+03,
     +  5.9228E+03, 5.9267E+03, 5.9306E+03, 5.9345E+03, 5.9384E+03,
     + 5.9423E+03, 5.9462E+03, 5.9501E+03, 5.9540E+03, 5.9579E+03,
     +  5.9618E+03, 5.9657E+03, 5.9696E+03, 5.9735E+03, 5.9774E+03,
     + 5.9814E+03, 5.9853E+03, 5.9892E+03, 5.9931E+03, 5.9970E+03,
     +  6.0009E+03, 6.0049E+03, 6.0088E+03, 6.0127E+03, 6.0166E+03,
     + 6.0206E+03, 6.0245E+03, 6.0284E+03, 6.0324E+03, 6.0363E+03,
     +  6.0402E+03, 6.0442E+03, 6.0481E+03, 6.0521E+03, 6.0560E+03,
     + 6.0599E+03, 6.0639E+03, 6.0678E+03, 6.0718E+03, 6.0757E+03,
     +  6.0797E+03, 6.0836E+03, 6.0876E+03, 6.0915E+03, 6.0955E+03,
     + 6.0994E+03, 6.1034E+03, 6.1074E+03, 6.1113E+03, 6.1153E+03,
     +  6.1192E+03, 6.1232E+03, 6.1272E+03, 6.1311E+03, 6.1351E+03,
     + 6.1391E+03, 6.1431E+03, 6.1470E+03, 6.1510E+03, 6.1550E+03,
     +  6.1590E+03, 6.1629E+03, 6.1669E+03, 6.1709E+03, 6.1749E+03,
     + 6.1789E+03, 6.1829E+03, 6.1868E+03, 6.1908E+03, 6.1948E+03,
     +  6.1988E+03, 6.2028E+03, 6.2068E+03, 6.2108E+03, 6.2148E+03,
     + 6.2188E+03, 6.2228E+03, 6.2268E+03, 6.2308E+03, 6.2348E+03,
     +  6.2388E+03, 6.2428E+03, 6.2468E+03, 6.2508E+03, 6.2549E+03,
     + 6.2589E+03, 6.2629E+03, 6.2669E+03, 6.2709E+03, 6.2749E+03,
     +  6.2790E+03, 6.2830E+03, 6.2870E+03, 6.2910E+03, 6.2950E+03,
     + 6.2991E+03, 6.3031E+03, 6.3071E+03, 6.3112E+03, 6.3152E+03,
     +  6.3192E+03, 6.3233E+03, 6.3273E+03, 6.3313E+03, 6.3354E+03,
     + 6.3394E+03, 6.3435E+03, 6.3475E+03, 6.3516E+03, 6.3556E+03,
     +  6.3597E+03, 6.3637E+03, 6.3678E+03, 6.3718E+03, 6.3759E+03,
     + 6.3799E+03, 6.3840E+03, 6.3880E+03, 6.3921E+03, 6.3961E+03,
     +  6.4002E+03, 6.4043E+03, 6.4083E+03, 6.4124E+03, 6.4165E+03,
     + 6.4205E+03, 6.4246E+03, 6.4287E+03, 6.4328E+03, 6.4368E+03,
     +  6.4409E+03, 6.4450E+03, 6.4491E+03, 6.4532E+03, 6.4572E+03,
     + 6.4613E+03, 6.4654E+03, 6.4695E+03, 6.4736E+03, 6.4777E+03,
     +  6.4818E+03, 6.4858E+03, 6.4899E+03, 6.4940E+03, 6.4981E+03,
     + 6.5022E+03, 6.5063E+03, 6.5104E+03, 6.5145E+03, 6.5186E+03,
     +  6.5227E+03, 6.5268E+03, 6.5310E+03, 6.5351E+03, 6.5392E+03,
     + 6.5433E+03, 6.5474E+03, 6.5515E+03, 6.5556E+03, 6.5597E+03,
     +  6.5639E+03, 6.5680E+03, 6.5721E+03, 6.5762E+03, 6.5804E+03,
     + 6.5845E+03, 6.5886E+03, 6.5927E+03, 6.5969E+03, 6.6010E+03,
     +  6.6051E+03, 6.6093E+03, 6.6134E+03, 6.6175E+03, 6.6217E+03,
     + 6.6258E+03, 6.6300E+03, 6.6341E+03, 6.6382E+03, 6.6424E+03,
     +  6.6465E+03, 6.6507E+03, 6.6548E+03, 6.6590E+03, 6.6631E+03,
     + 6.6673E+03, 6.6714E+03, 6.6756E+03, 6.6798E+03, 6.6839E+03,
     +  6.6881E+03, 6.6922E+03, 6.6964E+03, 6.7006E+03, 6.7047E+03,
     + 6.7089E+03, 6.7131E+03, 6.7172E+03, 6.7214E+03, 6.7256E+03,
     +  6.7298E+03, 6.7339E+03, 6.7381E+03, 6.7423E+03, 6.7465E+03,
     + 6.7507E+03, 6.7548E+03, 6.7590E+03, 6.7632E+03, 6.7674E+03,
     +  6.7716E+03, 6.7758E+03, 6.7800E+03, 6.7842E+03, 6.7884E+03,
     + 6.7926E+03, 6.7968E+03, 6.8010E+03, 6.8052E+03, 6.8094E+03,
     +  6.8136E+03, 6.8178E+03, 6.8220E+03, 6.8262E+03, 6.8304E+03,
     + 6.8346E+03, 6.8388E+03, 6.8430E+03, 6.8472E+03, 6.8515E+03,
     +  6.8557E+03, 6.8599E+03, 6.8641E+03, 6.8683E+03, 6.8726E+03,
     + 6.8768E+03, 6.8810E+03, 6.8852E+03, 6.8895E+03, 6.8937E+03,
     +  6.8979E+03, 6.9022E+03, 6.9064E+03, 6.9106E+03, 6.9149E+03,
     + 6.9191E+03, 6.9233E+03, 6.9276E+03, 6.9318E+03, 6.9361E+03,
     +  6.9403E+03, 6.9446E+03, 6.9488E+03, 6.9531E+03, 6.9573E+03,
     + 6.9616E+03, 6.9658E+03, 6.9701E+03, 6.9743E+03, 6.9786E+03,
     +  6.9828E+03, 6.9871E+03, 6.9914E+03, 6.9956E+03, 6.9999E+03,
     + 7.0042E+03, 7.0084E+03, 7.0127E+03, 7.0170E+03, 7.0212E+03,
     +  7.0255E+03, 7.0298E+03, 7.0341E+03, 7.0383E+03, 7.0426E+03,
     + 7.0469E+03, 7.0512E+03, 7.0555E+03, 7.0597E+03, 7.0640E+03,
     +  7.0683E+03, 7.0726E+03, 7.0769E+03, 7.0812E+03, 7.0855E+03,
     + 7.0898E+03, 7.0941E+03, 7.0984E+03, 7.1027E+03, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        17
      data (QofT( 2,J),J=1,2408)/ 8.0000E+00, 8.0000E+00, 8.0011E+00,
     +  8.0133E+00, 8.0596E+00, 8.1619E+00, 8.3307E+00, 8.5654E+00,
     + 8.8586E+00, 9.2006E+00, 9.5816E+00, 9.9932E+00, 1.0428E+01,
     +  1.0882E+01, 1.1349E+01, 1.1827E+01, 1.2314E+01, 1.2807E+01,
     + 1.3306E+01, 1.3809E+01, 1.4825E+01, 1.5851E+01, 1.6885E+01,
     +  1.7924E+01, 1.8967E+01, 2.0014E+01, 2.1064E+01, 2.2115E+01,
     + 2.3169E+01, 2.4224E+01, 2.5280E+01, 2.6337E+01, 2.7396E+01,
     +  2.8455E+01, 2.9515E+01, 3.0575E+01, 3.1636E+01, 3.2698E+01,
     + 3.3760E+01, 3.4823E+01, 3.5886E+01, 3.6949E+01, 3.8012E+01,
     +  3.9076E+01, 4.0140E+01, 4.1205E+01, 4.2269E+01, 4.3334E+01,
     + 4.4399E+01, 4.5464E+01, 4.6530E+01, 4.7595E+01, 4.8661E+01,
     +  4.9726E+01, 5.0792E+01, 5.1858E+01, 5.2924E+01, 5.3991E+01,
     + 5.5057E+01, 5.6123E+01, 5.7190E+01, 5.8257E+01, 5.9323E+01,
     +  6.0390E+01, 6.1457E+01, 6.2524E+01, 6.3591E+01, 6.4658E+01,
     + 6.5726E+01, 6.6793E+01, 6.7860E+01, 6.8928E+01, 6.9995E+01,
     +  7.1063E+01, 7.2130E+01, 7.3198E+01, 7.4266E+01, 7.5334E+01,
     + 7.6401E+01, 7.7469E+01, 7.8537E+01, 7.9605E+01, 8.0673E+01,
     +  8.1742E+01, 8.2810E+01, 8.3878E+01, 8.4946E+01, 8.6015E+01,
     + 8.7083E+01, 8.8151E+01, 8.9220E+01, 9.0288E+01, 9.1357E+01,
     +  9.2425E+01, 9.3494E+01, 9.4563E+01, 9.5631E+01, 9.6700E+01,
     + 9.7769E+01, 9.8838E+01, 9.9907E+01, 1.0098E+02, 1.0204E+02,
     +  1.0311E+02, 1.0418E+02, 1.0525E+02, 1.0632E+02, 1.0739E+02,
     + 1.0846E+02, 1.0953E+02, 1.1060E+02, 1.1167E+02, 1.1274E+02,
     +  1.1381E+02, 1.1487E+02, 1.1594E+02, 1.1701E+02, 1.1808E+02,
     + 1.1915E+02, 1.2022E+02, 1.2129E+02, 1.2236E+02, 1.2343E+02,
     +  1.2450E+02, 1.2557E+02, 1.2664E+02, 1.2771E+02, 1.2878E+02,
     + 1.2985E+02, 1.3092E+02, 1.3199E+02, 1.3306E+02, 1.3413E+02,
     +  1.3520E+02, 1.3627E+02, 1.3734E+02, 1.3841E+02, 1.3948E+02,
     + 1.4055E+02, 1.4162E+02, 1.4269E+02, 1.4376E+02, 1.4483E+02,
     +  1.4590E+02, 1.4697E+02, 1.4804E+02, 1.4911E+02, 1.5018E+02,
     + 1.5126E+02, 1.5233E+02, 1.5340E+02, 1.5447E+02, 1.5554E+02,
     +  1.5661E+02, 1.5768E+02, 1.5875E+02, 1.5982E+02, 1.6089E+02,
     + 1.6196E+02, 1.6303E+02, 1.6410E+02, 1.6518E+02, 1.6625E+02,
     +  1.6732E+02, 1.6839E+02, 1.6946E+02, 1.7053E+02, 1.7160E+02,
     + 1.7267E+02, 1.7374E+02, 1.7482E+02, 1.7589E+02, 1.7696E+02,
     +  1.7803E+02, 1.7910E+02, 1.8017E+02, 1.8124E+02, 1.8232E+02,
     + 1.8339E+02, 1.8446E+02, 1.8553E+02, 1.8660E+02, 1.8767E+02,
     +  1.8875E+02, 1.8982E+02, 1.9089E+02, 1.9196E+02, 1.9303E+02,
     + 1.9411E+02, 1.9518E+02, 1.9625E+02, 1.9732E+02, 1.9839E+02,
     +  1.9947E+02, 2.0054E+02, 2.0161E+02, 2.0268E+02, 2.0375E+02,
     + 2.0483E+02, 2.0590E+02, 2.0697E+02, 2.0804E+02, 2.0912E+02,
     +  2.1019E+02, 2.1126E+02, 2.1233E+02, 2.1341E+02, 2.1448E+02,
     + 2.1555E+02, 2.1663E+02, 2.1770E+02, 2.1877E+02, 2.1984E+02,
     +  2.2092E+02, 2.2199E+02, 2.2306E+02, 2.2414E+02, 2.2521E+02,
     + 2.2628E+02, 2.2736E+02, 2.2843E+02, 2.2950E+02, 2.3058E+02,
     +  2.3165E+02, 2.3272E+02, 2.3380E+02, 2.3487E+02, 2.3595E+02,
     + 2.3702E+02, 2.3809E+02, 2.3917E+02, 2.4024E+02, 2.4132E+02,
     +  2.4239E+02, 2.4346E+02, 2.4454E+02, 2.4561E+02, 2.4669E+02,
     + 2.4776E+02, 2.4884E+02, 2.4991E+02, 2.5099E+02, 2.5206E+02,
     +  2.5314E+02, 2.5421E+02, 2.5529E+02, 2.5636E+02, 2.5744E+02,
     + 2.5851E+02, 2.5959E+02, 2.6066E+02, 2.6174E+02, 2.6282E+02,
     +  2.6389E+02, 2.6497E+02, 2.6604E+02, 2.6712E+02, 2.6820E+02,
     + 2.6927E+02, 2.7035E+02, 2.7143E+02, 2.7250E+02, 2.7358E+02,
     +  2.7466E+02, 2.7573E+02, 2.7681E+02, 2.7789E+02, 2.7896E+02,
     + 2.8004E+02, 2.8112E+02, 2.8220E+02, 2.8328E+02, 2.8435E+02,
     +  2.8543E+02, 2.8651E+02, 2.8759E+02, 2.8867E+02, 2.8975E+02,
     + 2.9082E+02, 2.9190E+02, 2.9298E+02, 2.9406E+02, 2.9514E+02,
     +  2.9622E+02, 2.9730E+02, 2.9838E+02, 2.9946E+02, 3.0054E+02,
     + 3.0162E+02, 3.0270E+02, 3.0378E+02, 3.0486E+02, 3.0594E+02,
     +  3.0702E+02, 3.0811E+02, 3.0919E+02, 3.1027E+02, 3.1135E+02,
     + 3.1243E+02, 3.1351E+02, 3.1460E+02, 3.1568E+02, 3.1676E+02,
     +  3.1784E+02, 3.1893E+02, 3.2001E+02, 3.2109E+02, 3.2218E+02,
     + 3.2326E+02, 3.2435E+02, 3.2543E+02, 3.2652E+02, 3.2760E+02,
     +  3.2869E+02, 3.2977E+02, 3.3086E+02, 3.3194E+02, 3.3303E+02,
     + 3.3411E+02, 3.3520E+02, 3.3629E+02, 3.3737E+02, 3.3846E+02,
     +  3.3955E+02, 3.4063E+02, 3.4172E+02, 3.4281E+02, 3.4390E+02,
     + 3.4499E+02, 3.4608E+02, 3.4716E+02, 3.4825E+02, 3.4934E+02,
     +  3.5043E+02, 3.5152E+02, 3.5261E+02, 3.5370E+02, 3.5479E+02,
     + 3.5589E+02, 3.5698E+02, 3.5807E+02, 3.5916E+02, 3.6025E+02,
     +  3.6134E+02, 3.6244E+02, 3.6353E+02, 3.6462E+02, 3.6572E+02,
     + 3.6681E+02, 3.6791E+02, 3.6900E+02, 3.7010E+02, 3.7119E+02,
     +  3.7229E+02, 3.7338E+02, 3.7448E+02, 3.7557E+02, 3.7667E+02,
     + 3.7777E+02, 3.7887E+02, 3.7996E+02, 3.8106E+02, 3.8216E+02,
     +  3.8326E+02, 3.8436E+02, 3.8546E+02, 3.8656E+02, 3.8766E+02,
     + 3.8876E+02, 3.8986E+02, 3.9096E+02, 3.9206E+02, 3.9316E+02,
     +  3.9427E+02, 3.9537E+02, 3.9647E+02, 3.9757E+02, 3.9868E+02,
     + 3.9978E+02, 4.0089E+02, 4.0199E+02, 4.0310E+02, 4.0420E+02,
     +  4.0531E+02, 4.0642E+02, 4.0752E+02, 4.0863E+02, 4.0974E+02,
     + 4.1085E+02, 4.1195E+02, 4.1306E+02, 4.1417E+02, 4.1528E+02,
     +  4.1639E+02, 4.1750E+02, 4.1861E+02, 4.1973E+02, 4.2084E+02,
     + 4.2195E+02, 4.2306E+02, 4.2418E+02, 4.2529E+02, 4.2640E+02,
     +  4.2752E+02, 4.2863E+02, 4.2975E+02, 4.3086E+02, 4.3198E+02,
     + 4.3310E+02, 4.3421E+02, 4.3533E+02, 4.3645E+02, 4.3757E+02,
     +  4.3869E+02, 4.3981E+02, 4.4093E+02, 4.4205E+02, 4.4317E+02,
     + 4.4429E+02, 4.4541E+02, 4.4653E+02, 4.4766E+02, 4.4878E+02,
     +  4.4990E+02, 4.5103E+02, 4.5215E+02, 4.5328E+02, 4.5440E+02,
     + 4.5553E+02, 4.5666E+02, 4.5778E+02, 4.5891E+02, 4.6004E+02,
     +  4.6117E+02, 4.6230E+02, 4.6343E+02, 4.6456E+02, 4.6569E+02,
     + 4.6682E+02, 4.6795E+02, 4.6908E+02, 4.7022E+02, 4.7135E+02,
     +  4.7249E+02, 4.7362E+02, 4.7476E+02, 4.7589E+02, 4.7703E+02,
     + 4.7816E+02, 4.7930E+02, 4.8044E+02, 4.8158E+02, 4.8272E+02,
     +  4.8386E+02, 4.8500E+02, 4.8614E+02, 4.8728E+02, 4.8842E+02,
     + 4.8956E+02, 4.9071E+02, 4.9185E+02, 4.9300E+02, 4.9414E+02,
     +  4.9529E+02, 4.9643E+02, 4.9758E+02, 4.9873E+02, 4.9987E+02,
     + 5.0102E+02, 5.0217E+02, 5.0332E+02, 5.0447E+02, 5.0562E+02,
     +  5.0677E+02, 5.0792E+02, 5.0908E+02, 5.1023E+02, 5.1138E+02,
     + 5.1254E+02, 5.1369E+02, 5.1485E+02, 5.1601E+02, 5.1716E+02,
     +  5.1832E+02, 5.1948E+02, 5.2064E+02, 5.2180E+02, 5.2296E+02,
     + 5.2412E+02, 5.2528E+02, 5.2644E+02, 5.2760E+02, 5.2877E+02,
     +  5.2993E+02, 5.3109E+02, 5.3226E+02, 5.3343E+02, 5.3459E+02,
     + 5.3576E+02, 5.3693E+02, 5.3810E+02, 5.3927E+02, 5.4044E+02,
     +  5.4161E+02, 5.4278E+02, 5.4395E+02, 5.4512E+02, 5.4629E+02,
     + 5.4747E+02, 5.4864E+02, 5.4982E+02, 5.5099E+02, 5.5217E+02,
     +  5.5335E+02, 5.5453E+02, 5.5571E+02, 5.5688E+02, 5.5806E+02,
     + 5.5925E+02, 5.6043E+02, 5.6161E+02, 5.6279E+02, 5.6398E+02,
     +  5.6516E+02, 5.6634E+02, 5.6753E+02, 5.6872E+02, 5.6990E+02,
     + 5.7109E+02, 5.7228E+02, 5.7347E+02, 5.7466E+02, 5.7585E+02,
     +  5.7704E+02, 5.7823E+02, 5.7943E+02, 5.8062E+02, 5.8182E+02,
     + 5.8301E+02, 5.8421E+02, 5.8540E+02, 5.8660E+02, 5.8780E+02,
     +  5.8900E+02, 5.9020E+02, 5.9140E+02, 5.9260E+02, 5.9380E+02,
     + 5.9500E+02, 5.9621E+02, 5.9741E+02, 5.9861E+02, 5.9982E+02,
     +  6.0103E+02, 6.0223E+02, 6.0344E+02, 6.0465E+02, 6.0586E+02,
     + 6.0707E+02, 6.0828E+02, 6.0949E+02, 6.1070E+02, 6.1192E+02,
     +  6.1313E+02, 6.1435E+02, 6.1556E+02, 6.1678E+02, 6.1799E+02,
     + 6.1921E+02, 6.2043E+02, 6.2165E+02, 6.2287E+02, 6.2409E+02,
     +  6.2531E+02, 6.2654E+02, 6.2776E+02, 6.2898E+02, 6.3021E+02,
     + 6.3143E+02, 6.3266E+02, 6.3389E+02, 6.3512E+02, 6.3634E+02,
     +  6.3757E+02, 6.3881E+02, 6.4004E+02, 6.4127E+02, 6.4250E+02,
     + 6.4374E+02, 6.4497E+02, 6.4621E+02, 6.4744E+02, 6.4868E+02,
     +  6.4992E+02, 6.5116E+02, 6.5239E+02, 6.5364E+02, 6.5488E+02,
     + 6.5612E+02, 6.5736E+02, 6.5860E+02, 6.5985E+02, 6.6109E+02,
     +  6.6234E+02, 6.6359E+02, 6.6484E+02, 6.6608E+02, 6.6733E+02,
     + 6.6858E+02, 6.6983E+02, 6.7109E+02, 6.7234E+02, 6.7359E+02,
     +  6.7485E+02, 6.7610E+02, 6.7736E+02, 6.7862E+02, 6.7987E+02,
     + 6.8113E+02, 6.8239E+02, 6.8365E+02, 6.8491E+02, 6.8618E+02,
     +  6.8744E+02, 6.8870E+02, 6.8997E+02, 6.9123E+02, 6.9250E+02,
     + 6.9377E+02, 6.9504E+02, 6.9630E+02, 6.9757E+02, 6.9884E+02,
     +  7.0012E+02, 7.0139E+02, 7.0266E+02, 7.0394E+02, 7.0521E+02,
     + 7.0649E+02, 7.0776E+02, 7.0904E+02, 7.1032E+02, 7.1160E+02,
     +  7.1288E+02, 7.1416E+02, 7.1544E+02, 7.1673E+02, 7.1801E+02,
     + 7.1930E+02, 7.2058E+02, 7.2187E+02, 7.2316E+02, 7.2444E+02,
     +  7.2573E+02, 7.2702E+02, 7.2832E+02, 7.2961E+02, 7.3090E+02,
     + 7.3219E+02, 7.3349E+02, 7.3478E+02, 7.3608E+02, 7.3738E+02,
     +  7.3868E+02, 7.3998E+02, 7.4128E+02, 7.4258E+02, 7.4388E+02,
     + 7.4518E+02, 7.4648E+02, 7.4779E+02, 7.4909E+02, 7.5040E+02,
     +  7.5171E+02, 7.5302E+02, 7.5433E+02, 7.5564E+02, 7.5695E+02,
     + 7.5826E+02, 7.5957E+02, 7.6089E+02, 7.6220E+02, 7.6352E+02,
     +  7.6483E+02, 7.6615E+02, 7.6747E+02, 7.6879E+02, 7.7011E+02,
     + 7.7143E+02, 7.7275E+02, 7.7407E+02, 7.7540E+02, 7.7672E+02,
     +  7.7805E+02, 7.7938E+02, 7.8070E+02, 7.8203E+02, 7.8336E+02,
     + 7.8469E+02, 7.8602E+02, 7.8736E+02, 7.8869E+02, 7.9002E+02,
     +  7.9136E+02, 7.9270E+02, 7.9403E+02, 7.9537E+02, 7.9671E+02,
     + 7.9805E+02, 7.9939E+02, 8.0073E+02, 8.0208E+02, 8.0342E+02,
     +  8.0476E+02, 8.0611E+02, 8.0746E+02, 8.0880E+02, 8.1015E+02,
     + 8.1150E+02, 8.1285E+02, 8.1420E+02, 8.1556E+02, 8.1691E+02,
     +  8.1826E+02, 8.1962E+02, 8.2097E+02, 8.2233E+02, 8.2369E+02,
     + 8.2505E+02, 8.2641E+02, 8.2777E+02, 8.2913E+02, 8.3050E+02,
     +  8.3186E+02, 8.3322E+02, 8.3459E+02, 8.3596E+02, 8.3732E+02,
     + 8.3869E+02, 8.4006E+02, 8.4143E+02, 8.4281E+02, 8.4418E+02,
     +  8.4555E+02, 8.4693E+02, 8.4830E+02, 8.4968E+02, 8.5106E+02,
     + 8.5243E+02, 8.5381E+02, 8.5520E+02, 8.5658E+02, 8.5796E+02,
     +  8.5934E+02, 8.6073E+02, 8.6211E+02, 8.6350E+02, 8.6489E+02,
     + 8.6627E+02, 8.6766E+02, 8.6905E+02, 8.7045E+02, 8.7184E+02,
     +  8.7323E+02, 8.7463E+02, 8.7602E+02, 8.7742E+02, 8.7882E+02,
     + 8.8021E+02, 8.8161E+02, 8.8301E+02, 8.8441E+02, 8.8582E+02,
     +  8.8722E+02, 8.8862E+02, 8.9003E+02, 8.9144E+02, 8.9284E+02,
     + 8.9425E+02, 8.9566E+02, 8.9707E+02, 8.9848E+02, 8.9990E+02,
     +  9.0131E+02, 9.0272E+02, 9.0414E+02, 9.0555E+02, 9.0697E+02,
     + 9.0839E+02, 9.0981E+02, 9.1123E+02, 9.1265E+02, 9.1407E+02,
     +  9.1550E+02, 9.1692E+02, 9.1835E+02, 9.1977E+02, 9.2120E+02,
     + 9.2263E+02, 9.2406E+02, 9.2549E+02, 9.2692E+02, 9.2835E+02,
     +  9.2979E+02, 9.3122E+02, 9.3266E+02, 9.3410E+02, 9.3553E+02,
     + 9.3697E+02, 9.3841E+02, 9.3985E+02, 9.4129E+02, 9.4274E+02,
     +  9.4418E+02, 9.4563E+02, 9.4707E+02, 9.4852E+02, 9.4997E+02,
     + 9.5142E+02, 9.5287E+02, 9.5432E+02, 9.5577E+02, 9.5722E+02,
     +  9.5868E+02, 9.6013E+02, 9.6159E+02, 9.6305E+02, 9.6450E+02,
     + 9.6596E+02, 9.6742E+02, 9.6888E+02, 9.7035E+02, 9.7181E+02,
     +  9.7328E+02, 9.7474E+02, 9.7621E+02, 9.7768E+02, 9.7914E+02,
     + 9.8061E+02, 9.8208E+02, 9.8356E+02, 9.8503E+02, 9.8650E+02,
     +  9.8798E+02, 9.8945E+02, 9.9093E+02, 9.9241E+02, 9.9389E+02,
     + 9.9537E+02, 9.9685E+02, 9.9833E+02, 9.9982E+02, 1.0013E+03,
     +  1.0028E+03, 1.0043E+03, 1.0058E+03, 1.0072E+03, 1.0087E+03,
     + 1.0102E+03, 1.0117E+03, 1.0132E+03, 1.0147E+03, 1.0162E+03,
     +  1.0177E+03, 1.0192E+03, 1.0207E+03, 1.0222E+03, 1.0237E+03,
     + 1.0252E+03, 1.0267E+03, 1.0282E+03, 1.0297E+03, 1.0312E+03,
     +  1.0327E+03, 1.0342E+03, 1.0357E+03, 1.0372E+03, 1.0388E+03,
     + 1.0403E+03, 1.0418E+03, 1.0433E+03, 1.0448E+03, 1.0463E+03,
     +  1.0478E+03, 1.0494E+03, 1.0509E+03, 1.0524E+03, 1.0539E+03,
     + 1.0554E+03, 1.0570E+03, 1.0585E+03, 1.0600E+03, 1.0615E+03,
     +  1.0631E+03, 1.0646E+03, 1.0661E+03, 1.0677E+03, 1.0692E+03,
     + 1.0707E+03, 1.0723E+03, 1.0738E+03, 1.0753E+03, 1.0769E+03,
     +  1.0784E+03, 1.0799E+03, 1.0815E+03, 1.0830E+03, 1.0846E+03,
     + 1.0861E+03, 1.0877E+03, 1.0892E+03, 1.0908E+03, 1.0923E+03,
     +  1.0939E+03, 1.0954E+03, 1.0970E+03, 1.0985E+03, 1.1001E+03,
     + 1.1016E+03, 1.1032E+03, 1.1047E+03, 1.1063E+03, 1.1078E+03,
     +  1.1094E+03, 1.1110E+03, 1.1125E+03, 1.1141E+03, 1.1157E+03,
     + 1.1172E+03, 1.1188E+03, 1.1204E+03, 1.1219E+03, 1.1235E+03,
     +  1.1251E+03, 1.1266E+03, 1.1282E+03, 1.1298E+03, 1.1314E+03,
     + 1.1329E+03, 1.1345E+03, 1.1361E+03, 1.1377E+03, 1.1393E+03,
     +  1.1408E+03, 1.1424E+03, 1.1440E+03, 1.1456E+03, 1.1472E+03,
     + 1.1488E+03, 1.1504E+03, 1.1519E+03, 1.1535E+03, 1.1551E+03,
     +  1.1567E+03, 1.1583E+03, 1.1599E+03, 1.1615E+03, 1.1631E+03,
     + 1.1647E+03, 1.1663E+03, 1.1679E+03, 1.1695E+03, 1.1711E+03,
     +  1.1727E+03, 1.1743E+03, 1.1759E+03, 1.1775E+03, 1.1791E+03,
     + 1.1807E+03, 1.1824E+03, 1.1840E+03, 1.1856E+03, 1.1872E+03,
     +  1.1888E+03, 1.1904E+03, 1.1920E+03, 1.1937E+03, 1.1953E+03,
     + 1.1969E+03, 1.1985E+03, 1.2001E+03, 1.2018E+03, 1.2034E+03,
     +  1.2050E+03, 1.2067E+03, 1.2083E+03, 1.2099E+03, 1.2115E+03,
     + 1.2132E+03, 1.2148E+03, 1.2164E+03, 1.2181E+03, 1.2197E+03,
     +  1.2213E+03, 1.2230E+03, 1.2246E+03, 1.2263E+03, 1.2279E+03,
     + 1.2296E+03, 1.2312E+03, 1.2328E+03, 1.2345E+03, 1.2361E+03,
     +  1.2378E+03, 1.2394E+03, 1.2411E+03, 1.2427E+03, 1.2444E+03,
     + 1.2460E+03, 1.2477E+03, 1.2494E+03, 1.2510E+03, 1.2527E+03,
     +  1.2543E+03, 1.2560E+03, 1.2577E+03, 1.2593E+03, 1.2610E+03,
     + 1.2626E+03, 1.2643E+03, 1.2660E+03, 1.2676E+03, 1.2693E+03,
     +  1.2710E+03, 1.2727E+03, 1.2743E+03, 1.2760E+03, 1.2777E+03,
     + 1.2794E+03, 1.2810E+03, 1.2827E+03, 1.2844E+03, 1.2861E+03,
     +  1.2878E+03, 1.2894E+03, 1.2911E+03, 1.2928E+03, 1.2945E+03,
     + 1.2962E+03, 1.2979E+03, 1.2996E+03, 1.3013E+03, 1.3030E+03,
     +  1.3046E+03, 1.3063E+03, 1.3080E+03, 1.3097E+03, 1.3114E+03,
     + 1.3131E+03, 1.3148E+03, 1.3165E+03, 1.3182E+03, 1.3199E+03,
     +  1.3216E+03, 1.3233E+03, 1.3251E+03, 1.3268E+03, 1.3285E+03,
     + 1.3302E+03, 1.3319E+03, 1.3336E+03, 1.3353E+03, 1.3370E+03,
     +  1.3388E+03, 1.3405E+03, 1.3422E+03, 1.3439E+03, 1.3456E+03,
     + 1.3474E+03, 1.3491E+03, 1.3508E+03, 1.3525E+03, 1.3543E+03,
     +  1.3560E+03, 1.3577E+03, 1.3594E+03, 1.3612E+03, 1.3629E+03,
     + 1.3646E+03, 1.3664E+03, 1.3681E+03, 1.3698E+03, 1.3716E+03,
     +  1.3733E+03, 1.3751E+03, 1.3768E+03, 1.3785E+03, 1.3803E+03,
     + 1.3820E+03, 1.3838E+03, 1.3855E+03, 1.3873E+03, 1.3890E+03,
     +  1.3908E+03, 1.3925E+03, 1.3943E+03, 1.3960E+03, 1.3978E+03,
     + 1.3995E+03, 1.4013E+03, 1.4031E+03, 1.4048E+03, 1.4066E+03,
     +  1.4083E+03, 1.4101E+03, 1.4119E+03, 1.4136E+03, 1.4154E+03,
     + 1.4172E+03, 1.4189E+03, 1.4207E+03, 1.4225E+03, 1.4243E+03,
     +  1.4260E+03, 1.4278E+03, 1.4296E+03, 1.4314E+03, 1.4331E+03,
     + 1.4349E+03, 1.4367E+03, 1.4385E+03, 1.4403E+03, 1.4420E+03,
     +  1.4438E+03, 1.4456E+03, 1.4474E+03, 1.4492E+03, 1.4510E+03,
     + 1.4528E+03, 1.4546E+03, 1.4564E+03, 1.4582E+03, 1.4600E+03,
     +  1.4618E+03, 1.4635E+03, 1.4653E+03, 1.4671E+03, 1.4690E+03,
     + 1.4708E+03, 1.4726E+03, 1.4744E+03, 1.4762E+03, 1.4780E+03,
     +  1.4798E+03, 1.4816E+03, 1.4834E+03, 1.4852E+03, 1.4870E+03,
     + 1.4888E+03, 1.4907E+03, 1.4925E+03, 1.4943E+03, 1.4961E+03,
     +  1.4979E+03, 1.4998E+03, 1.5016E+03, 1.5034E+03, 1.5052E+03,
     + 1.5071E+03, 1.5089E+03, 1.5107E+03, 1.5125E+03, 1.5144E+03,
     +  1.5162E+03, 1.5180E+03, 1.5199E+03, 1.5217E+03, 1.5235E+03,
     + 1.5254E+03, 1.5272E+03, 1.5291E+03, 1.5309E+03, 1.5327E+03,
     +  1.5346E+03, 1.5364E+03, 1.5383E+03, 1.5401E+03, 1.5420E+03,
     + 1.5438E+03, 1.5457E+03, 1.5475E+03, 1.5494E+03, 1.5512E+03,
     +  1.5531E+03, 1.5549E+03, 1.5568E+03, 1.5587E+03, 1.5605E+03,
     + 1.5624E+03, 1.5642E+03, 1.5661E+03, 1.5680E+03, 1.5698E+03,
     +  1.5717E+03, 1.5736E+03, 1.5755E+03, 1.5773E+03, 1.5792E+03,
     + 1.5811E+03, 1.5829E+03, 1.5848E+03, 1.5867E+03, 1.5886E+03,
     +  1.5905E+03, 1.5923E+03, 1.5942E+03, 1.5961E+03, 1.5980E+03,
     + 1.5999E+03, 1.6018E+03, 1.6036E+03, 1.6055E+03, 1.6074E+03,
     +  1.6093E+03, 1.6112E+03, 1.6131E+03, 1.6150E+03, 1.6169E+03,
     + 1.6188E+03, 1.6207E+03, 1.6226E+03, 1.6245E+03, 1.6264E+03,
     +  1.6283E+03, 1.6302E+03, 1.6321E+03, 1.6340E+03, 1.6359E+03,
     + 1.6378E+03, 1.6397E+03, 1.6416E+03, 1.6436E+03, 1.6455E+03,
     +  1.6474E+03, 1.6493E+03, 1.6512E+03, 1.6531E+03, 1.6551E+03,
     + 1.6570E+03, 1.6589E+03, 1.6608E+03, 1.6627E+03, 1.6647E+03,
     +  1.6666E+03, 1.6685E+03, 1.6705E+03, 1.6724E+03, 1.6743E+03,
     + 1.6762E+03, 1.6782E+03, 1.6801E+03, 1.6821E+03, 1.6840E+03,
     +  1.6859E+03, 1.6879E+03, 1.6898E+03, 1.6918E+03, 1.6937E+03,
     + 1.6956E+03, 1.6976E+03, 1.6995E+03, 1.7015E+03, 1.7034E+03,
     +  1.7054E+03, 1.7073E+03, 1.7093E+03, 1.7112E+03, 1.7132E+03,
     + 1.7152E+03, 1.7171E+03, 1.7191E+03, 1.7210E+03, 1.7230E+03,
     +  1.7250E+03, 1.7269E+03, 1.7289E+03, 1.7309E+03, 1.7328E+03,
     + 1.7348E+03, 1.7368E+03, 1.7387E+03, 1.7407E+03, 1.7427E+03,
     +  1.7447E+03, 1.7466E+03, 1.7486E+03, 1.7506E+03, 1.7526E+03,
     + 1.7545E+03, 1.7565E+03, 1.7585E+03, 1.7605E+03, 1.7625E+03,
     +  1.7645E+03, 1.7665E+03, 1.7684E+03, 1.7704E+03, 1.7724E+03,
     + 1.7744E+03, 1.7764E+03, 1.7784E+03, 1.7804E+03, 1.7824E+03,
     +  1.7844E+03, 1.7864E+03, 1.7884E+03, 1.7904E+03, 1.7924E+03,
     + 1.7944E+03, 1.7964E+03, 1.7984E+03, 1.8004E+03, 1.8024E+03,
     +  1.8045E+03, 1.8065E+03, 1.8085E+03, 1.8105E+03, 1.8125E+03,
     + 1.8145E+03, 1.8166E+03, 1.8186E+03, 1.8206E+03, 1.8226E+03,
     +  1.8246E+03, 1.8267E+03, 1.8287E+03, 1.8307E+03, 1.8327E+03,
     + 1.8348E+03, 1.8368E+03, 1.8388E+03, 1.8409E+03, 1.8429E+03,
     +  1.8449E+03, 1.8470E+03, 1.8490E+03, 1.8511E+03, 1.8531E+03,
     + 1.8551E+03, 1.8572E+03, 1.8592E+03, 1.8613E+03, 1.8633E+03,
     +  1.8654E+03, 1.8674E+03, 1.8695E+03, 1.8715E+03, 1.8736E+03,
     + 1.8756E+03, 1.8777E+03, 1.8797E+03, 1.8818E+03, 1.8838E+03,
     +  1.8859E+03, 1.8880E+03, 1.8900E+03, 1.8921E+03, 1.8942E+03,
     + 1.8962E+03, 1.8983E+03, 1.9004E+03, 1.9024E+03, 1.9045E+03,
     +  1.9066E+03, 1.9086E+03, 1.9107E+03, 1.9128E+03, 1.9149E+03,
     + 1.9170E+03, 1.9190E+03, 1.9211E+03, 1.9232E+03, 1.9253E+03,
     +  1.9274E+03, 1.9294E+03, 1.9315E+03, 1.9336E+03, 1.9357E+03,
     + 1.9378E+03, 1.9399E+03, 1.9420E+03, 1.9441E+03, 1.9462E+03,
     +  1.9483E+03, 1.9504E+03, 1.9525E+03, 1.9546E+03, 1.9567E+03,
     + 1.9588E+03, 1.9609E+03, 1.9630E+03, 1.9651E+03, 1.9672E+03,
     +  1.9693E+03, 1.9714E+03, 1.9735E+03, 1.9756E+03, 1.9778E+03,
     + 1.9799E+03, 1.9820E+03, 1.9841E+03, 1.9862E+03, 1.9883E+03,
     +  1.9905E+03, 1.9926E+03, 1.9947E+03, 1.9968E+03, 1.9990E+03,
     + 2.0011E+03, 2.0032E+03, 2.0054E+03, 2.0075E+03, 2.0096E+03,
     +  2.0118E+03, 2.0139E+03, 2.0160E+03, 2.0182E+03, 2.0203E+03,
     + 2.0224E+03, 2.0246E+03, 2.0267E+03, 2.0289E+03, 2.0310E+03,
     +  2.0332E+03, 2.0353E+03, 2.0375E+03, 2.0396E+03, 2.0418E+03,
     + 2.0439E+03, 2.0461E+03, 2.0482E+03, 2.0504E+03, 2.0525E+03,
     +  2.0547E+03, 2.0569E+03, 2.0590E+03, 2.0612E+03, 2.0633E+03,
     + 2.0655E+03, 2.0677E+03, 2.0698E+03, 2.0720E+03, 2.0742E+03,
     +  2.0763E+03, 2.0785E+03, 2.0807E+03, 2.0829E+03, 2.0850E+03,
     + 2.0872E+03, 2.0894E+03, 2.0916E+03, 2.0938E+03, 2.0959E+03,
     +  2.0981E+03, 2.1003E+03, 2.1025E+03, 2.1047E+03, 2.1069E+03,
     + 2.1091E+03, 2.1113E+03, 2.1134E+03, 2.1156E+03, 2.1178E+03,
     +  2.1200E+03, 2.1222E+03, 2.1244E+03, 2.1266E+03, 2.1288E+03,
     + 2.1310E+03, 2.1332E+03, 2.1354E+03, 2.1376E+03, 2.1399E+03,
     +  2.1421E+03, 2.1443E+03, 2.1465E+03, 2.1487E+03, 2.1509E+03,
     + 2.1531E+03, 2.1553E+03, 2.1576E+03, 2.1598E+03, 2.1620E+03,
     +  2.1642E+03, 2.1664E+03, 2.1687E+03, 2.1709E+03, 2.1731E+03,
     + 2.1753E+03, 2.1776E+03, 2.1798E+03, 2.1820E+03, 2.1843E+03,
     +  2.1865E+03, 2.1887E+03, 2.1910E+03, 2.1932E+03, 2.1954E+03,
     + 2.1977E+03, 2.1999E+03, 2.2022E+03, 2.2044E+03, 2.2067E+03,
     +  2.2089E+03, 2.2112E+03, 2.2134E+03, 2.2157E+03, 2.2179E+03,
     + 2.2202E+03, 2.2224E+03, 2.2247E+03, 2.2269E+03, 2.2292E+03,
     +  2.2314E+03, 2.2337E+03, 2.2360E+03, 2.2382E+03, 2.2405E+03,
     + 2.2428E+03, 2.2450E+03, 2.2473E+03, 2.2496E+03, 2.2518E+03,
     +  2.2541E+03, 2.2564E+03, 2.2586E+03, 2.2609E+03, 2.2632E+03,
     + 2.2655E+03, 2.2678E+03, 2.2700E+03, 2.2723E+03, 2.2746E+03,
     +  2.2769E+03, 2.2792E+03, 2.2815E+03, 2.2837E+03, 2.2860E+03,
     + 2.2883E+03, 2.2906E+03, 2.2929E+03, 2.2952E+03, 2.2975E+03,
     +  2.2998E+03, 2.3021E+03, 2.3044E+03, 2.3067E+03, 2.3090E+03,
     + 2.3113E+03, 2.3136E+03, 2.3159E+03, 2.3182E+03, 2.3205E+03,
     +  2.3228E+03, 2.3251E+03, 2.3275E+03, 2.3298E+03, 2.3321E+03,
     + 2.3344E+03, 2.3367E+03, 2.3390E+03, 2.3414E+03, 2.3437E+03,
     +  2.3460E+03, 2.3483E+03, 2.3506E+03, 2.3530E+03, 2.3553E+03,
     + 2.3576E+03, 2.3600E+03, 2.3623E+03, 2.3646E+03, 2.3670E+03,
     +  2.3693E+03, 2.3716E+03, 2.3740E+03, 2.3763E+03, 2.3786E+03,
     + 2.3810E+03, 2.3833E+03, 2.3857E+03, 2.3880E+03, 2.3904E+03,
     +  2.3927E+03, 2.3951E+03, 2.3974E+03, 2.3998E+03, 2.4021E+03,
     + 2.4045E+03, 2.4068E+03, 2.4092E+03, 2.4115E+03, 2.4139E+03,
     +  2.4163E+03, 2.4186E+03, 2.4210E+03, 2.4234E+03, 2.4257E+03,
     + 2.4281E+03, 2.4305E+03, 2.4328E+03, 2.4352E+03, 2.4376E+03,
     +  2.4399E+03, 2.4423E+03, 2.4447E+03, 2.4471E+03, 2.4494E+03,
     + 2.4518E+03, 2.4542E+03, 2.4566E+03, 2.4590E+03, 2.4614E+03,
     +  2.4637E+03, 2.4661E+03, 2.4685E+03, 2.4709E+03, 2.4733E+03,
     + 2.4757E+03, 2.4781E+03, 2.4805E+03, 2.4829E+03, 2.4853E+03,
     +  2.4877E+03, 2.4901E+03, 2.4925E+03, 2.4949E+03, 2.4973E+03,
     + 2.4997E+03, 2.5021E+03, 2.5045E+03, 2.5069E+03, 2.5093E+03,
     +  2.5117E+03, 2.5142E+03, 2.5166E+03, 2.5190E+03, 2.5214E+03,
     + 2.5238E+03, 2.5262E+03, 2.5287E+03, 2.5311E+03, 2.5335E+03,
     +  2.5359E+03, 2.5384E+03, 2.5408E+03, 2.5432E+03, 2.5456E+03,
     + 2.5481E+03, 2.5505E+03, 2.5529E+03, 2.5554E+03, 2.5578E+03,
     +  2.5603E+03, 2.5627E+03, 2.5651E+03, 2.5676E+03, 2.5700E+03,
     + 2.5725E+03, 2.5749E+03, 2.5774E+03, 2.5798E+03, 2.5823E+03,
     +  2.5847E+03, 2.5872E+03, 2.5896E+03, 2.5921E+03, 2.5945E+03,
     + 2.5970E+03, 2.5994E+03, 2.6019E+03, 2.6044E+03, 2.6068E+03,
     +  2.6093E+03, 2.6117E+03, 2.6142E+03, 2.6167E+03, 2.6192E+03,
     + 2.6216E+03, 2.6241E+03, 2.6266E+03, 2.6290E+03, 2.6315E+03,
     +  2.6340E+03, 2.6365E+03, 2.6389E+03, 2.6414E+03, 2.6439E+03,
     + 2.6464E+03, 2.6489E+03, 2.6514E+03, 2.6539E+03, 2.6563E+03,
     +  2.6588E+03, 2.6613E+03, 2.6638E+03, 2.6663E+03, 2.6688E+03,
     + 2.6713E+03, 2.6738E+03, 2.6763E+03, 2.6788E+03, 2.6813E+03,
     +  2.6838E+03, 2.6863E+03, 2.6888E+03, 2.6913E+03, 2.6938E+03,
     + 2.6963E+03, 2.6988E+03, 2.7014E+03, 2.7039E+03, 2.7064E+03,
     +  2.7089E+03, 2.7114E+03, 2.7139E+03, 2.7165E+03, 2.7190E+03,
     + 2.7215E+03, 2.7240E+03, 2.7265E+03, 2.7291E+03, 2.7316E+03,
     +  2.7341E+03, 2.7367E+03, 2.7392E+03, 2.7417E+03, 2.7443E+03,
     + 2.7468E+03, 2.7493E+03, 2.7519E+03, 2.7544E+03, 2.7569E+03,
     +  2.7595E+03, 2.7620E+03, 2.7646E+03, 2.7671E+03, 2.7697E+03,
     + 2.7722E+03, 2.7748E+03, 2.7773E+03, 2.7799E+03, 2.7824E+03,
     +  2.7850E+03, 2.7875E+03, 2.7901E+03, 2.7927E+03, 2.7952E+03,
     + 2.7978E+03, 2.8003E+03, 2.8029E+03, 2.8055E+03, 2.8080E+03,
     +  2.8106E+03, 2.8132E+03, 2.8157E+03, 2.8183E+03, 2.8209E+03,
     + 2.8235E+03, 2.8260E+03, 2.8286E+03, 2.8312E+03, 2.8338E+03,
     +  2.8364E+03, 2.8389E+03, 2.8415E+03, 2.8441E+03, 2.8467E+03,
     + 2.8493E+03, 2.8519E+03, 2.8545E+03, 2.8571E+03, 2.8596E+03,
     +  2.8622E+03, 2.8648E+03, 2.8674E+03, 2.8700E+03, 2.8726E+03,
     + 2.8752E+03, 2.8778E+03, 2.8804E+03, 2.8830E+03, 2.8856E+03,
     +  2.8883E+03, 2.8909E+03, 2.8935E+03, 2.8961E+03, 2.8987E+03,
     + 2.9013E+03, 2.9039E+03, 2.9065E+03, 2.9092E+03, 2.9118E+03,
     +  2.9144E+03, 2.9170E+03, 2.9196E+03, 2.9223E+03, 2.9249E+03,
     + 2.9275E+03, 2.9302E+03, 2.9328E+03, 2.9354E+03, 2.9380E+03,
     +  2.9407E+03, 2.9433E+03, 2.9460E+03, 2.9486E+03, 2.9512E+03,
     + 2.9539E+03, 2.9565E+03, 2.9592E+03, 2.9618E+03, 2.9644E+03,
     +  2.9671E+03, 2.9697E+03, 2.9724E+03, 2.9750E+03, 2.9777E+03,
     + 2.9803E+03, 2.9830E+03, 2.9857E+03, 2.9883E+03, 2.9910E+03,
     +  2.9936E+03, 2.9963E+03, 2.9990E+03, 3.0016E+03, 3.0043E+03,
     + 3.0070E+03, 3.0096E+03, 3.0123E+03, 3.0150E+03, 3.0176E+03,
     +  3.0203E+03, 3.0230E+03, 3.0257E+03, 3.0283E+03, 3.0310E+03,
     + 3.0337E+03, 3.0364E+03, 3.0391E+03, 3.0418E+03, 3.0444E+03,
     +  3.0471E+03, 3.0498E+03, 3.0525E+03, 3.0552E+03, 3.0579E+03,
     + 3.0606E+03, 3.0633E+03, 3.0660E+03, 3.0687E+03, 3.0714E+03,
     +  3.0741E+03, 3.0768E+03, 3.0795E+03, 3.0822E+03, 3.0849E+03,
     + 3.0876E+03, 3.0903E+03, 3.0930E+03, 3.0957E+03, 3.0984E+03,
     +  3.1011E+03, 3.1039E+03, 3.1066E+03, 3.1093E+03, 3.1120E+03,
     + 3.1147E+03, 3.1175E+03, 3.1202E+03, 3.1229E+03, 3.1256E+03,
     +  3.1284E+03, 3.1311E+03, 3.1338E+03, 3.1365E+03, 3.1393E+03,
     + 3.1420E+03, 3.1447E+03, 3.1475E+03, 3.1502E+03, 3.1530E+03,
     +  3.1557E+03, 3.1584E+03, 3.1612E+03, 3.1639E+03, 3.1667E+03,
     + 3.1694E+03, 3.1722E+03, 3.1749E+03, 3.1777E+03, 3.1804E+03,
     +  3.1832E+03, 3.1859E+03, 3.1887E+03, 3.1914E+03, 3.1942E+03,
     + 3.1970E+03, 3.1997E+03, 3.2025E+03, 3.2052E+03, 3.2080E+03,
     +  3.2108E+03, 3.2135E+03, 3.2163E+03, 3.2191E+03, 3.2219E+03,
     + 3.2246E+03, 3.2274E+03, 3.2302E+03, 3.2330E+03, 3.2357E+03,
     +  3.2385E+03, 3.2413E+03, 3.2441E+03, 3.2469E+03, 3.2497E+03,
     + 3.2524E+03, 3.2552E+03, 3.2580E+03, 3.2608E+03, 3.2636E+03,
     +  3.2664E+03, 3.2692E+03, 3.2720E+03, 3.2748E+03, 3.2776E+03,
     + 3.2804E+03, 3.2832E+03, 3.2860E+03, 3.2888E+03, 3.2916E+03,
     +  3.2944E+03, 3.2972E+03, 3.3000E+03, 3.3028E+03, 3.3057E+03,
     + 3.3085E+03, 3.3113E+03, 3.3141E+03, 3.3169E+03, 3.3197E+03,
     +  3.3226E+03, 3.3254E+03, 3.3282E+03, 3.3310E+03, 3.3339E+03,
     + 3.3367E+03, 3.3395E+03, 3.3423E+03, 3.3452E+03, 3.3480E+03,
     +  3.3508E+03, 3.3537E+03, 3.3565E+03, 3.3593E+03, 3.3622E+03,
     + 3.3650E+03, 3.3679E+03, 3.3707E+03, 3.3736E+03, 3.3764E+03,
     +  3.3793E+03, 3.3821E+03, 3.3850E+03, 3.3878E+03, 3.3907E+03,
     + 3.3935E+03, 3.3964E+03, 3.3992E+03, 3.4021E+03, 3.4049E+03,
     +  3.4078E+03, 3.4107E+03, 3.4135E+03, 3.4164E+03, 3.4193E+03,
     + 3.4221E+03, 3.4250E+03, 3.4279E+03, 3.4307E+03, 3.4336E+03,
     +  3.4365E+03, 3.4394E+03, 3.4422E+03, 3.4451E+03, 3.4480E+03,
     + 3.4509E+03, 3.4538E+03, 3.4567E+03, 3.4595E+03, 3.4624E+03,
     +  3.4653E+03, 3.4682E+03, 3.4711E+03, 3.4740E+03, 3.4769E+03,
     + 3.4798E+03, 3.4827E+03, 3.4856E+03, 3.4885E+03, 3.4914E+03,
     +  3.4943E+03, 3.4972E+03, 3.5001E+03, 3.5030E+03, 3.5059E+03,
     + 3.5088E+03, 3.5117E+03, 3.5146E+03, 3.5175E+03, 3.5204E+03,
     +  3.5234E+03, 3.5263E+03, 3.5292E+03, 3.5321E+03, 3.5350E+03,
     + 3.5380E+03, 3.5409E+03, 3.5438E+03, 3.5467E+03, 3.5497E+03,
     +  3.5526E+03, 3.5555E+03, 3.5584E+03, 3.5614E+03, 3.5643E+03,
     + 3.5673E+03, 3.5702E+03, 3.5731E+03, 3.5761E+03, 3.5790E+03,
     +  3.5819E+03, 3.5849E+03, 3.5878E+03, 3.5908E+03, 3.5937E+03,
     + 3.5967E+03, 3.5996E+03, 3.6026E+03, 3.6055E+03, 3.6085E+03,
     +  3.6114E+03, 3.6144E+03, 3.6174E+03, 3.6203E+03, 3.6233E+03,
     + 3.6262E+03, 3.6292E+03, 3.6322E+03, 3.6351E+03, 3.6381E+03,
     +  3.6411E+03, 3.6441E+03, 3.6470E+03, 3.6500E+03, 3.6530E+03,
     + 3.6560E+03, 3.6589E+03, 3.6619E+03, 3.6649E+03, 3.6679E+03,
     +  3.6709E+03, 3.6738E+03, 3.6768E+03, 3.6798E+03, 3.6828E+03,
     + 3.6858E+03, 3.6888E+03, 3.6918E+03, 3.6948E+03, 3.6978E+03,
     +  3.7008E+03, 3.7038E+03, 3.7068E+03, 3.7098E+03, 3.7128E+03,
     + 3.7158E+03, 3.7188E+03, 3.7218E+03, 3.7248E+03, 3.7278E+03,
     +  3.7308E+03, 3.7338E+03, 3.7368E+03, 3.7398E+03, 3.7429E+03,
     + 3.7459E+03, 3.7489E+03, 3.7519E+03, 3.7549E+03, 3.7580E+03,
     +  3.7610E+03, 3.7640E+03, 3.7670E+03, 3.7701E+03, 3.7731E+03,
     + 3.7761E+03, 3.7792E+03, 3.7822E+03, 3.7852E+03, 3.7883E+03,
     +  3.7913E+03, 3.7943E+03, 3.7974E+03, 3.8004E+03, 3.8035E+03,
     + 3.8065E+03, 3.8096E+03, 3.8126E+03, 3.8157E+03, 3.8187E+03,
     +  3.8218E+03, 3.8248E+03, 3.8279E+03, 3.8309E+03, 3.8340E+03,
     + 3.8370E+03, 3.8401E+03, 3.8432E+03, 3.8462E+03, 3.8493E+03,
     +  3.8524E+03, 3.8554E+03, 3.8585E+03, 3.8616E+03, 3.8646E+03,
     + 3.8677E+03, 3.8708E+03, 3.8738E+03, 3.8769E+03, 3.8800E+03,
     +  3.8831E+03, 3.8862E+03, 3.8892E+03, 3.8923E+03, 3.8954E+03,
     + 3.8985E+03, 3.9016E+03, 3.9047E+03, 3.9078E+03, 3.9109E+03,
     +  3.9139E+03, 3.9170E+03, 3.9201E+03, 3.9232E+03, 3.9263E+03,
     + 3.9294E+03, 3.9325E+03, 3.9356E+03, 3.9387E+03, 3.9418E+03,
     +  3.9449E+03, 3.9481E+03, 3.9512E+03, 3.9543E+03, 3.9574E+03,
     + 3.9605E+03, 3.9636E+03, 3.9667E+03, 3.9699E+03, 3.9730E+03,
     +  3.9761E+03, 3.9792E+03, 3.9823E+03, 3.9855E+03, 3.9886E+03,
     + 3.9917E+03, 3.9948E+03, 3.9980E+03, 4.0011E+03, 4.0042E+03,
     +  4.0074E+03, 4.0105E+03, 4.0136E+03, 4.0168E+03, 4.0199E+03,
     + 4.0231E+03, 4.0262E+03, 4.0293E+03, 4.0325E+03, 4.0356E+03,
     +  4.0388E+03, 4.0419E+03, 4.0451E+03, 4.0482E+03, 4.0514E+03,
     + 4.0545E+03, 4.0577E+03, 4.0609E+03, 4.0640E+03, 4.0672E+03,
     +  4.0703E+03, 4.0735E+03, 4.0767E+03, 4.0798E+03, 4.0830E+03,
     + 4.0862E+03, 4.0893E+03, 4.0925E+03, 4.0957E+03, 4.0989E+03,
     +  4.1020E+03, 4.1052E+03, 4.1084E+03, 4.1116E+03, 4.1147E+03,
     + 4.1179E+03, 4.1211E+03, 4.1243E+03, 4.1275E+03, 4.1307E+03,
     +  4.1339E+03, 4.1371E+03, 4.1402E+03, 4.1434E+03, 4.1466E+03,
     + 4.1498E+03, 4.1530E+03, 4.1562E+03, 4.1594E+03, 4.1626E+03,
     +  4.1658E+03, 4.1690E+03, 4.1722E+03, 4.1754E+03, 4.1787E+03,
     + 4.1819E+03, 4.1851E+03, 4.1883E+03, 4.1915E+03, 4.1947E+03,
     +  4.1979E+03, 4.2012E+03, 4.2044E+03, 4.2076E+03, 4.2108E+03,
     + 4.2140E+03, 4.2173E+03, 4.2205E+03, 4.2237E+03, 4.2270E+03,
     +  4.2302E+03, 4.2334E+03, 4.2366E+03, 4.2399E+03, 4.2431E+03,
     + 4.2464E+03, 4.2496E+03, 4.2528E+03, 4.2561E+03, 4.2593E+03,
     +  4.2626E+03, 4.2658E+03, 4.2691E+03, 4.2723E+03, 4.2756E+03,
     + 4.2788E+03, 4.2821E+03, 4.2853E+03, 4.2886E+03, 4.2918E+03,
     +  4.2951E+03, 4.2984E+03, 4.3016E+03, 4.3049E+03, 4.3081E+03,
     + 4.3114E+03, 4.3147E+03, 4.3179E+03, 4.3212E+03, 4.3245E+03,
     +  4.3278E+03, 4.3310E+03, 4.3343E+03, 4.3376E+03, 4.3409E+03,
     + 4.3441E+03, 4.3474E+03, 4.3507E+03, 4.3540E+03, 4.3573E+03,
     +  4.3606E+03, 4.3638E+03, 4.3671E+03, 4.3704E+03, 4.3737E+03,
     + 4.3770E+03, 4.3803E+03, 4.3836E+03, 4.3869E+03, 4.3902E+03,
     +  4.3935E+03, 4.3968E+03, 4.4001E+03, 4.4034E+03, 4.4067E+03,
     + 4.4100E+03, 4.4133E+03, 4.4166E+03, 4.4199E+03, 4.4233E+03,
     +  4.4266E+03, 4.4299E+03, 4.4332E+03, 4.4365E+03, 4.4398E+03,
     + 4.4432E+03, 4.4465E+03, 4.4498E+03, 4.4531E+03, 4.4565E+03,
     +  4.4598E+03, 4.4631E+03, 4.4664E+03, 4.4698E+03, 4.4731E+03,
     + 4.4764E+03, 4.4798E+03, 4.4831E+03, 4.4865E+03, 4.4898E+03,
     +  4.4931E+03, 4.4965E+03, 4.4998E+03, 4.5032E+03, 4.5065E+03,
     + 4.5099E+03, 4.5132E+03, 4.5166E+03, 4.5199E+03, 4.5233E+03,
     +  4.5266E+03, 4.5300E+03, 4.5334E+03, 4.5367E+03, 4.5401E+03,
     + 4.5434E+03, 4.5468E+03, 4.5502E+03, 4.5535E+03, 4.5569E+03,
     +  4.5603E+03, 4.5636E+03, 4.5670E+03, 4.5704E+03, 4.5738E+03,
     + 4.5771E+03, 4.5805E+03, 4.5839E+03, 4.5873E+03, 4.5907E+03,
     +  4.5941E+03, 4.5974E+03, 4.6008E+03, 4.6042E+03, 4.6076E+03,
     + 4.6110E+03, 4.6144E+03, 4.6178E+03, 4.6212E+03, 4.6246E+03,
     +  4.6280E+03, 4.6314E+03, 4.6348E+03, 4.6382E+03, 4.6416E+03,
     + 4.6450E+03, 4.6484E+03, 4.6518E+03, 4.6552E+03, 4.6586E+03,
     +  4.6620E+03, 4.6654E+03, 4.6688E+03, 4.6723E+03, 4.6757E+03,
     + 4.6791E+03, 4.6825E+03, 4.6859E+03, 4.6894E+03, 4.6928E+03,
     +  4.6962E+03, 4.6996E+03, 4.7031E+03, 4.7065E+03, 4.7099E+03,
     + 4.7134E+03, 4.7168E+03, 4.7202E+03, 4.7237E+03, 4.7271E+03,
     +  4.7306E+03, 4.7340E+03, 4.7374E+03, 4.7409E+03, 4.7443E+03,
     + 4.7478E+03, 4.7512E+03, 4.7547E+03, 4.7581E+03, 4.7616E+03,
     +  4.7650E+03, 4.7685E+03, 4.7719E+03, 4.7754E+03, 4.7789E+03/
      data (QofT( 2,J),J=2409,4510)/ 4.7823E+03, 4.7858E+03,
     +  4.7892E+03, 4.7927E+03, 4.7962E+03, 4.7996E+03, 4.8031E+03,
     +  4.8066E+03,
     + 4.8100E+03, 4.8135E+03, 4.8170E+03, 4.8205E+03, 4.8240E+03,
     +  4.8274E+03, 4.8309E+03, 4.8344E+03, 4.8379E+03, 4.8414E+03,
     + 4.8448E+03, 4.8483E+03, 4.8518E+03, 4.8553E+03, 4.8588E+03,
     +  4.8623E+03, 4.8658E+03, 4.8693E+03, 4.8728E+03, 4.8763E+03,
     + 4.8798E+03, 4.8833E+03, 4.8868E+03, 4.8903E+03, 4.8938E+03,
     +  4.8973E+03, 4.9008E+03, 4.9043E+03, 4.9078E+03, 4.9113E+03,
     + 4.9149E+03, 4.9184E+03, 4.9219E+03, 4.9254E+03, 4.9289E+03,
     +  4.9324E+03, 4.9360E+03, 4.9395E+03, 4.9430E+03, 4.9465E+03,
     + 4.9501E+03, 4.9536E+03, 4.9571E+03, 4.9607E+03, 4.9642E+03,
     +  4.9677E+03, 4.9713E+03, 4.9748E+03, 4.9783E+03, 4.9819E+03,
     + 4.9854E+03, 4.9890E+03, 4.9925E+03, 4.9961E+03, 4.9996E+03,
     +  5.0032E+03, 5.0067E+03, 5.0103E+03, 5.0138E+03, 5.0174E+03,
     + 5.0209E+03, 5.0245E+03, 5.0280E+03, 5.0316E+03, 5.0352E+03,
     +  5.0387E+03, 5.0423E+03, 5.0459E+03, 5.0494E+03, 5.0530E+03,
     + 5.0566E+03, 5.0601E+03, 5.0637E+03, 5.0673E+03, 5.0709E+03,
     +  5.0744E+03, 5.0780E+03, 5.0816E+03, 5.0852E+03, 5.0888E+03,
     + 5.0923E+03, 5.0959E+03, 5.0995E+03, 5.1031E+03, 5.1067E+03,
     +  5.1103E+03, 5.1139E+03, 5.1175E+03, 5.1211E+03, 5.1247E+03,
     + 5.1283E+03, 5.1319E+03, 5.1355E+03, 5.1391E+03, 5.1427E+03,
     +  5.1463E+03, 5.1499E+03, 5.1535E+03, 5.1571E+03, 5.1607E+03,
     + 5.1643E+03, 5.1679E+03, 5.1716E+03, 5.1752E+03, 5.1788E+03,
     +  5.1824E+03, 5.1860E+03, 5.1896E+03, 5.1933E+03, 5.1969E+03,
     + 5.2005E+03, 5.2042E+03, 5.2078E+03, 5.2114E+03, 5.2150E+03,
     +  5.2187E+03, 5.2223E+03, 5.2259E+03, 5.2296E+03, 5.2332E+03,
     + 5.2369E+03, 5.2405E+03, 5.2442E+03, 5.2478E+03, 5.2514E+03,
     +  5.2551E+03, 5.2587E+03, 5.2624E+03, 5.2660E+03, 5.2697E+03,
     + 5.2733E+03, 5.2770E+03, 5.2807E+03, 5.2843E+03, 5.2880E+03,
     +  5.2916E+03, 5.2953E+03, 5.2990E+03, 5.3026E+03, 5.3063E+03,
     + 5.3100E+03, 5.3136E+03, 5.3173E+03, 5.3210E+03, 5.3247E+03,
     +  5.3283E+03, 5.3320E+03, 5.3357E+03, 5.3394E+03, 5.3431E+03,
     + 5.3467E+03, 5.3504E+03, 5.3541E+03, 5.3578E+03, 5.3615E+03,
     +  5.3652E+03, 5.3689E+03, 5.3726E+03, 5.3762E+03, 5.3799E+03,
     + 5.3836E+03, 5.3873E+03, 5.3910E+03, 5.3947E+03, 5.3984E+03,
     +  5.4021E+03, 5.4058E+03, 5.4096E+03, 5.4133E+03, 5.4170E+03,
     + 5.4207E+03, 5.4244E+03, 5.4281E+03, 5.4318E+03, 5.4355E+03,
     +  5.4393E+03, 5.4430E+03, 5.4467E+03, 5.4504E+03, 5.4541E+03,
     + 5.4579E+03, 5.4616E+03, 5.4653E+03, 5.4691E+03, 5.4728E+03,
     +  5.4765E+03, 5.4802E+03, 5.4840E+03, 5.4877E+03, 5.4915E+03,
     + 5.4952E+03, 5.4989E+03, 5.5027E+03, 5.5064E+03, 5.5102E+03,
     +  5.5139E+03, 5.5177E+03, 5.5214E+03, 5.5252E+03, 5.5289E+03,
     + 5.5327E+03, 5.5364E+03, 5.5402E+03, 5.5439E+03, 5.5477E+03,
     +  5.5515E+03, 5.5552E+03, 5.5590E+03, 5.5627E+03, 5.5665E+03,
     + 5.5703E+03, 5.5740E+03, 5.5778E+03, 5.5816E+03, 5.5854E+03,
     +  5.5891E+03, 5.5929E+03, 5.5967E+03, 5.6005E+03, 5.6042E+03,
     + 5.6080E+03, 5.6118E+03, 5.6156E+03, 5.6194E+03, 5.6232E+03,
     +  5.6270E+03, 5.6307E+03, 5.6345E+03, 5.6383E+03, 5.6421E+03,
     + 5.6459E+03, 5.6497E+03, 5.6535E+03, 5.6573E+03, 5.6611E+03,
     +  5.6649E+03, 5.6687E+03, 5.6725E+03, 5.6763E+03, 5.6801E+03,
     + 5.6839E+03, 5.6878E+03, 5.6916E+03, 5.6954E+03, 5.6992E+03,
     +  5.7030E+03, 5.7068E+03, 5.7107E+03, 5.7145E+03, 5.7183E+03,
     + 5.7221E+03, 5.7259E+03, 5.7298E+03, 5.7336E+03, 5.7374E+03,
     +  5.7413E+03, 5.7451E+03, 5.7489E+03, 5.7528E+03, 5.7566E+03,
     + 5.7604E+03, 5.7643E+03, 5.7681E+03, 5.7720E+03, 5.7758E+03,
     +  5.7796E+03, 5.7835E+03, 5.7873E+03, 5.7912E+03, 5.7950E+03,
     + 5.7989E+03, 5.8027E+03, 5.8066E+03, 5.8105E+03, 5.8143E+03,
     +  5.8182E+03, 5.8220E+03, 5.8259E+03, 5.8298E+03, 5.8336E+03,
     + 5.8375E+03, 5.8414E+03, 5.8452E+03, 5.8491E+03, 5.8530E+03,
     +  5.8568E+03, 5.8607E+03, 5.8646E+03, 5.8685E+03, 5.8723E+03,
     + 5.8762E+03, 5.8801E+03, 5.8840E+03, 5.8879E+03, 5.8918E+03,
     +  5.8956E+03, 5.8995E+03, 5.9034E+03, 5.9073E+03, 5.9112E+03,
     + 5.9151E+03, 5.9190E+03, 5.9229E+03, 5.9268E+03, 5.9307E+03,
     +  5.9346E+03, 5.9385E+03, 5.9424E+03, 5.9463E+03, 5.9502E+03,
     + 5.9541E+03, 5.9580E+03, 5.9619E+03, 5.9659E+03, 5.9698E+03,
     +  5.9737E+03, 5.9776E+03, 5.9815E+03, 5.9854E+03, 5.9894E+03,
     + 5.9933E+03, 5.9972E+03, 6.0011E+03, 6.0051E+03, 6.0090E+03,
     +  6.0129E+03, 6.0169E+03, 6.0208E+03, 6.0247E+03, 6.0287E+03,
     + 6.0326E+03, 6.0365E+03, 6.0405E+03, 6.0444E+03, 6.0484E+03,
     +  6.0523E+03, 6.0562E+03, 6.0602E+03, 6.0641E+03, 6.0681E+03,
     + 6.0720E+03, 6.0760E+03, 6.0799E+03, 6.0839E+03, 6.0879E+03,
     +  6.0918E+03, 6.0958E+03, 6.0997E+03, 6.1037E+03, 6.1077E+03,
     + 6.1116E+03, 6.1156E+03, 6.1196E+03, 6.1235E+03, 6.1275E+03,
     +  6.1315E+03, 6.1354E+03, 6.1394E+03, 6.1434E+03, 6.1474E+03,
     + 6.1514E+03, 6.1553E+03, 6.1593E+03, 6.1633E+03, 6.1673E+03,
     +  6.1713E+03, 6.1753E+03, 6.1792E+03, 6.1832E+03, 6.1872E+03,
     + 6.1912E+03, 6.1952E+03, 6.1992E+03, 6.2032E+03, 6.2072E+03,
     +  6.2112E+03, 6.2152E+03, 6.2192E+03, 6.2232E+03, 6.2272E+03,
     + 6.2312E+03, 6.2352E+03, 6.2392E+03, 6.2433E+03, 6.2473E+03,
     +  6.2513E+03, 6.2553E+03, 6.2593E+03, 6.2633E+03, 6.2674E+03,
     + 6.2714E+03, 6.2754E+03, 6.2794E+03, 6.2835E+03, 6.2875E+03,
     +  6.2915E+03, 6.2955E+03, 6.2996E+03, 6.3036E+03, 6.3076E+03,
     + 6.3117E+03, 6.3157E+03, 6.3197E+03, 6.3238E+03, 6.3278E+03,
     +  6.3319E+03, 6.3359E+03, 6.3400E+03, 6.3440E+03, 6.3481E+03,
     + 6.3521E+03, 6.3562E+03, 6.3602E+03, 6.3643E+03, 6.3683E+03,
     +  6.3724E+03, 6.3764E+03, 6.3805E+03, 6.3846E+03, 6.3886E+03,
     + 6.3927E+03, 6.3967E+03, 6.4008E+03, 6.4049E+03, 6.4089E+03,
     +  6.4130E+03, 6.4171E+03, 6.4212E+03, 6.4252E+03, 6.4293E+03,
     + 6.4334E+03, 6.4375E+03, 6.4416E+03, 6.4456E+03, 6.4497E+03,
     +  6.4538E+03, 6.4579E+03, 6.4620E+03, 6.4661E+03, 6.4702E+03,
     + 6.4742E+03, 6.4783E+03, 6.4824E+03, 6.4865E+03, 6.4906E+03,
     +  6.4947E+03, 6.4988E+03, 6.5029E+03, 6.5070E+03, 6.5111E+03,
     + 6.5152E+03, 6.5194E+03, 6.5235E+03, 6.5276E+03, 6.5317E+03,
     +  6.5358E+03, 6.5399E+03, 6.5440E+03, 6.5481E+03, 6.5523E+03,
     + 6.5564E+03, 6.5605E+03, 6.5646E+03, 6.5688E+03, 6.5729E+03,
     +  6.5770E+03, 6.5811E+03, 6.5853E+03, 6.5894E+03, 6.5935E+03,
     + 6.5977E+03, 6.6018E+03, 6.6059E+03, 6.6101E+03, 6.6142E+03,
     +  6.6184E+03, 6.6225E+03, 6.6266E+03, 6.6308E+03, 6.6349E+03,
     + 6.6391E+03, 6.6432E+03, 6.6474E+03, 6.6515E+03, 6.6557E+03,
     +  6.6598E+03, 6.6640E+03, 6.6682E+03, 6.6723E+03, 6.6765E+03,
     + 6.6806E+03, 6.6848E+03, 6.6890E+03, 6.6931E+03, 6.6973E+03,
     +  6.7015E+03, 6.7057E+03, 6.7098E+03, 6.7140E+03, 6.7182E+03,
     + 6.7224E+03, 6.7265E+03, 6.7307E+03, 6.7349E+03, 6.7391E+03,
     +  6.7433E+03, 6.7474E+03, 6.7516E+03, 6.7558E+03, 6.7600E+03,
     + 6.7642E+03, 6.7684E+03, 6.7726E+03, 6.7768E+03, 6.7810E+03,
     +  6.7852E+03, 6.7894E+03, 6.7936E+03, 6.7978E+03, 6.8020E+03,
     + 6.8062E+03, 6.8104E+03, 6.8146E+03, 6.8188E+03, 6.8230E+03,
     +  6.8272E+03, 6.8314E+03, 6.8357E+03, 6.8399E+03, 6.8441E+03,
     + 6.8483E+03, 6.8525E+03, 6.8567E+03, 6.8610E+03, 6.8652E+03,
     +  6.8694E+03, 6.8736E+03, 6.8779E+03, 6.8821E+03, 6.8863E+03,
     + 6.8906E+03, 6.8948E+03, 6.8990E+03, 6.9033E+03, 6.9075E+03,
     +  6.9118E+03, 6.9160E+03, 6.9202E+03, 6.9245E+03, 6.9287E+03,
     + 6.9330E+03, 6.9372E+03, 6.9415E+03, 6.9457E+03, 6.9500E+03,
     +  6.9542E+03, 6.9585E+03, 6.9627E+03, 6.9670E+03, 6.9713E+03,
     + 6.9755E+03, 6.9798E+03, 6.9841E+03, 6.9883E+03, 6.9926E+03,
     +  6.9969E+03, 7.0011E+03, 7.0054E+03, 7.0097E+03, 7.0139E+03,
     + 7.0182E+03, 7.0225E+03, 7.0268E+03, 7.0310E+03, 7.0353E+03,
     +  7.0396E+03, 7.0439E+03, 7.0482E+03, 7.0525E+03, 7.0568E+03,
     + 7.0610E+03, 7.0653E+03, 7.0696E+03, 7.0739E+03, 7.0782E+03,
     +  7.0825E+03, 7.0868E+03, 7.0911E+03, 7.0954E+03, 7.0997E+03,
     + 7.1040E+03, 7.1083E+03, 7.1126E+03, 7.1169E+03, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        25
      data (QofT( 3,J),J=1,2408)/ 1.2000E+01, 1.2015E+01, 1.2204E+01,
     +  1.2745E+01, 1.3622E+01, 1.4737E+01, 1.6001E+01, 1.7355E+01,
     + 1.8764E+01, 2.0208E+01, 2.1675E+01, 2.3157E+01, 2.4652E+01,
     +  2.6155E+01, 2.7665E+01, 2.9180E+01, 3.0699E+01, 3.2221E+01,
     + 3.3747E+01, 3.5275E+01, 3.8336E+01, 4.1404E+01, 4.4476E+01,
     +  4.7551E+01, 5.0629E+01, 5.3710E+01, 5.6792E+01, 5.9876E+01,
     + 6.2961E+01, 6.6047E+01, 6.9134E+01, 7.2222E+01, 7.5311E+01,
     +  7.8401E+01, 8.1491E+01, 8.4581E+01, 8.7672E+01, 9.0763E+01,
     + 9.3855E+01, 9.6947E+01, 1.0004E+02, 1.0313E+02, 1.0623E+02,
     +  1.0932E+02, 1.1241E+02, 1.1551E+02, 1.1860E+02, 1.2169E+02,
     + 1.2479E+02, 1.2788E+02, 1.3098E+02, 1.3407E+02, 1.3717E+02,
     +  1.4026E+02, 1.4336E+02, 1.4646E+02, 1.4955E+02, 1.5265E+02,
     + 1.5574E+02, 1.5884E+02, 1.6194E+02, 1.6503E+02, 1.6813E+02,
     +  1.7123E+02, 1.7432E+02, 1.7742E+02, 1.8052E+02, 1.8361E+02,
     + 1.8671E+02, 1.8981E+02, 1.9291E+02, 1.9601E+02, 1.9910E+02,
     +  2.0220E+02, 2.0530E+02, 2.0840E+02, 2.1150E+02, 2.1459E+02,
     + 2.1769E+02, 2.2079E+02, 2.2389E+02, 2.2699E+02, 2.3009E+02,
     +  2.3319E+02, 2.3629E+02, 2.3938E+02, 2.4248E+02, 2.4558E+02,
     + 2.4868E+02, 2.5178E+02, 2.5488E+02, 2.5798E+02, 2.6108E+02,
     +  2.6418E+02, 2.6728E+02, 2.7038E+02, 2.7348E+02, 2.7658E+02,
     + 2.7968E+02, 2.8278E+02, 2.8588E+02, 2.8898E+02, 2.9209E+02,
     +  2.9519E+02, 2.9829E+02, 3.0139E+02, 3.0449E+02, 3.0759E+02,
     + 3.1069E+02, 3.1379E+02, 3.1689E+02, 3.2000E+02, 3.2310E+02,
     +  3.2620E+02, 3.2930E+02, 3.3240E+02, 3.3551E+02, 3.3861E+02,
     + 3.4171E+02, 3.4481E+02, 3.4792E+02, 3.5102E+02, 3.5412E+02,
     +  3.5722E+02, 3.6033E+02, 3.6343E+02, 3.6653E+02, 3.6964E+02,
     + 3.7274E+02, 3.7584E+02, 3.7895E+02, 3.8205E+02, 3.8515E+02,
     +  3.8826E+02, 3.9136E+02, 3.9446E+02, 3.9757E+02, 4.0067E+02,
     + 4.0378E+02, 4.0688E+02, 4.0999E+02, 4.1309E+02, 4.1620E+02,
     +  4.1930E+02, 4.2241E+02, 4.2551E+02, 4.2862E+02, 4.3172E+02,
     + 4.3483E+02, 4.3793E+02, 4.4104E+02, 4.4414E+02, 4.4725E+02,
     +  4.5036E+02, 4.5346E+02, 4.5657E+02, 4.5968E+02, 4.6278E+02,
     + 4.6589E+02, 4.6900E+02, 4.7211E+02, 4.7521E+02, 4.7832E+02,
     +  4.8143E+02, 4.8454E+02, 4.8765E+02, 4.9075E+02, 4.9386E+02,
     + 4.9697E+02, 5.0008E+02, 5.0319E+02, 5.0630E+02, 5.0941E+02,
     +  5.1252E+02, 5.1563E+02, 5.1874E+02, 5.2185E+02, 5.2496E+02,
     + 5.2808E+02, 5.3119E+02, 5.3430E+02, 5.3741E+02, 5.4053E+02,
     +  5.4364E+02, 5.4675E+02, 5.4987E+02, 5.5298E+02, 5.5609E+02,
     + 5.5921E+02, 5.6232E+02, 5.6544E+02, 5.6856E+02, 5.7167E+02,
     +  5.7479E+02, 5.7791E+02, 5.8102E+02, 5.8414E+02, 5.8726E+02,
     + 5.9038E+02, 5.9350E+02, 5.9662E+02, 5.9974E+02, 6.0286E+02,
     +  6.0598E+02, 6.0910E+02, 6.1223E+02, 6.1535E+02, 6.1847E+02,
     + 6.2160E+02, 6.2472E+02, 6.2785E+02, 6.3097E+02, 6.3410E+02,
     +  6.3722E+02, 6.4035E+02, 6.4348E+02, 6.4661E+02, 6.4974E+02,
     + 6.5287E+02, 6.5600E+02, 6.5913E+02, 6.6226E+02, 6.6540E+02,
     +  6.6853E+02, 6.7166E+02, 6.7480E+02, 6.7793E+02, 6.8107E+02,
     + 6.8421E+02, 6.8735E+02, 6.9049E+02, 6.9363E+02, 6.9677E+02,
     +  6.9991E+02, 7.0305E+02, 7.0619E+02, 7.0934E+02, 7.1248E+02,
     + 7.1563E+02, 7.1878E+02, 7.2192E+02, 7.2507E+02, 7.2822E+02,
     +  7.3137E+02, 7.3452E+02, 7.3768E+02, 7.4083E+02, 7.4399E+02,
     + 7.4714E+02, 7.5030E+02, 7.5346E+02, 7.5662E+02, 7.5978E+02,
     +  7.6294E+02, 7.6610E+02, 7.6927E+02, 7.7243E+02, 7.7560E+02,
     + 7.7876E+02, 7.8193E+02, 7.8510E+02, 7.8827E+02, 7.9144E+02,
     +  7.9462E+02, 7.9779E+02, 8.0097E+02, 8.0415E+02, 8.0732E+02,
     + 8.1050E+02, 8.1369E+02, 8.1687E+02, 8.2005E+02, 8.2324E+02,
     +  8.2643E+02, 8.2961E+02, 8.3280E+02, 8.3599E+02, 8.3919E+02,
     + 8.4238E+02, 8.4558E+02, 8.4877E+02, 8.5197E+02, 8.5517E+02,
     +  8.5837E+02, 8.6158E+02, 8.6478E+02, 8.6799E+02, 8.7120E+02,
     + 8.7441E+02, 8.7762E+02, 8.8083E+02, 8.8405E+02, 8.8726E+02,
     +  8.9048E+02, 8.9370E+02, 8.9692E+02, 9.0014E+02, 9.0337E+02,
     + 9.0659E+02, 9.0982E+02, 9.1305E+02, 9.1629E+02, 9.1952E+02,
     +  9.2275E+02, 9.2599E+02, 9.2923E+02, 9.3247E+02, 9.3571E+02,
     + 9.3896E+02, 9.4221E+02, 9.4545E+02, 9.4871E+02, 9.5196E+02,
     +  9.5521E+02, 9.5847E+02, 9.6173E+02, 9.6499E+02, 9.6825E+02,
     + 9.7151E+02, 9.7478E+02, 9.7805E+02, 9.8132E+02, 9.8459E+02,
     +  9.8787E+02, 9.9114E+02, 9.9442E+02, 9.9770E+02, 1.0010E+03,
     + 1.0043E+03, 1.0076E+03, 1.0108E+03, 1.0141E+03, 1.0174E+03,
     +  1.0207E+03, 1.0240E+03, 1.0273E+03, 1.0306E+03, 1.0339E+03,
     + 1.0372E+03, 1.0406E+03, 1.0439E+03, 1.0472E+03, 1.0505E+03,
     +  1.0538E+03, 1.0571E+03, 1.0605E+03, 1.0638E+03, 1.0671E+03,
     + 1.0704E+03, 1.0738E+03, 1.0771E+03, 1.0805E+03, 1.0838E+03,
     +  1.0871E+03, 1.0905E+03, 1.0938E+03, 1.0972E+03, 1.1005E+03,
     + 1.1039E+03, 1.1072E+03, 1.1106E+03, 1.1140E+03, 1.1173E+03,
     +  1.1207E+03, 1.1241E+03, 1.1274E+03, 1.1308E+03, 1.1342E+03,
     + 1.1376E+03, 1.1410E+03, 1.1443E+03, 1.1477E+03, 1.1511E+03,
     +  1.1545E+03, 1.1579E+03, 1.1613E+03, 1.1647E+03, 1.1681E+03,
     + 1.1715E+03, 1.1749E+03, 1.1784E+03, 1.1818E+03, 1.1852E+03,
     +  1.1886E+03, 1.1920E+03, 1.1955E+03, 1.1989E+03, 1.2023E+03,
     + 1.2058E+03, 1.2092E+03, 1.2126E+03, 1.2161E+03, 1.2195E+03,
     +  1.2230E+03, 1.2264E+03, 1.2299E+03, 1.2333E+03, 1.2368E+03,
     + 1.2403E+03, 1.2437E+03, 1.2472E+03, 1.2507E+03, 1.2541E+03,
     +  1.2576E+03, 1.2611E+03, 1.2646E+03, 1.2681E+03, 1.2716E+03,
     + 1.2751E+03, 1.2786E+03, 1.2821E+03, 1.2856E+03, 1.2891E+03,
     +  1.2926E+03, 1.2961E+03, 1.2996E+03, 1.3031E+03, 1.3066E+03,
     + 1.3102E+03, 1.3137E+03, 1.3172E+03, 1.3207E+03, 1.3243E+03,
     +  1.3278E+03, 1.3314E+03, 1.3349E+03, 1.3385E+03, 1.3420E+03,
     + 1.3456E+03, 1.3491E+03, 1.3527E+03, 1.3562E+03, 1.3598E+03,
     +  1.3634E+03, 1.3669E+03, 1.3705E+03, 1.3741E+03, 1.3777E+03,
     + 1.3813E+03, 1.3849E+03, 1.3885E+03, 1.3921E+03, 1.3956E+03,
     +  1.3993E+03, 1.4029E+03, 1.4065E+03, 1.4101E+03, 1.4137E+03,
     + 1.4173E+03, 1.4209E+03, 1.4246E+03, 1.4282E+03, 1.4318E+03,
     +  1.4355E+03, 1.4391E+03, 1.4427E+03, 1.4464E+03, 1.4500E+03,
     + 1.4537E+03, 1.4573E+03, 1.4610E+03, 1.4647E+03, 1.4683E+03,
     +  1.4720E+03, 1.4757E+03, 1.4793E+03, 1.4830E+03, 1.4867E+03,
     + 1.4904E+03, 1.4941E+03, 1.4978E+03, 1.5015E+03, 1.5052E+03,
     +  1.5089E+03, 1.5126E+03, 1.5163E+03, 1.5200E+03, 1.5237E+03,
     + 1.5274E+03, 1.5312E+03, 1.5349E+03, 1.5386E+03, 1.5423E+03,
     +  1.5461E+03, 1.5498E+03, 1.5536E+03, 1.5573E+03, 1.5611E+03,
     + 1.5648E+03, 1.5686E+03, 1.5724E+03, 1.5761E+03, 1.5799E+03,
     +  1.5837E+03, 1.5874E+03, 1.5912E+03, 1.5950E+03, 1.5988E+03,
     + 1.6026E+03, 1.6064E+03, 1.6102E+03, 1.6140E+03, 1.6178E+03,
     +  1.6216E+03, 1.6254E+03, 1.6292E+03, 1.6330E+03, 1.6369E+03,
     + 1.6407E+03, 1.6445E+03, 1.6484E+03, 1.6522E+03, 1.6560E+03,
     +  1.6599E+03, 1.6637E+03, 1.6676E+03, 1.6714E+03, 1.6753E+03,
     + 1.6792E+03, 1.6830E+03, 1.6869E+03, 1.6908E+03, 1.6946E+03,
     +  1.6985E+03, 1.7024E+03, 1.7063E+03, 1.7102E+03, 1.7141E+03,
     + 1.7180E+03, 1.7219E+03, 1.7258E+03, 1.7297E+03, 1.7336E+03,
     +  1.7376E+03, 1.7415E+03, 1.7454E+03, 1.7493E+03, 1.7533E+03,
     + 1.7572E+03, 1.7612E+03, 1.7651E+03, 1.7690E+03, 1.7730E+03,
     +  1.7770E+03, 1.7809E+03, 1.7849E+03, 1.7889E+03, 1.7928E+03,
     + 1.7968E+03, 1.8008E+03, 1.8048E+03, 1.8088E+03, 1.8127E+03,
     +  1.8167E+03, 1.8207E+03, 1.8247E+03, 1.8287E+03, 1.8328E+03,
     + 1.8368E+03, 1.8408E+03, 1.8448E+03, 1.8488E+03, 1.8529E+03,
     +  1.8569E+03, 1.8609E+03, 1.8650E+03, 1.8690E+03, 1.8731E+03,
     + 1.8771E+03, 1.8812E+03, 1.8852E+03, 1.8893E+03, 1.8934E+03,
     +  1.8975E+03, 1.9015E+03, 1.9056E+03, 1.9097E+03, 1.9138E+03,
     + 1.9179E+03, 1.9220E+03, 1.9261E+03, 1.9302E+03, 1.9343E+03,
     +  1.9384E+03, 1.9425E+03, 1.9466E+03, 1.9508E+03, 1.9549E+03,
     + 1.9590E+03, 1.9632E+03, 1.9673E+03, 1.9714E+03, 1.9756E+03,
     +  1.9797E+03, 1.9839E+03, 1.9880E+03, 1.9922E+03, 1.9964E+03,
     + 2.0005E+03, 2.0047E+03, 2.0089E+03, 2.0131E+03, 2.0173E+03,
     +  2.0215E+03, 2.0257E+03, 2.0299E+03, 2.0341E+03, 2.0383E+03,
     + 2.0425E+03, 2.0467E+03, 2.0509E+03, 2.0551E+03, 2.0594E+03,
     +  2.0636E+03, 2.0678E+03, 2.0721E+03, 2.0763E+03, 2.0806E+03,
     + 2.0848E+03, 2.0891E+03, 2.0933E+03, 2.0976E+03, 2.1019E+03,
     +  2.1061E+03, 2.1104E+03, 2.1147E+03, 2.1190E+03, 2.1233E+03,
     + 2.1276E+03, 2.1319E+03, 2.1362E+03, 2.1405E+03, 2.1448E+03,
     +  2.1491E+03, 2.1534E+03, 2.1577E+03, 2.1620E+03, 2.1664E+03,
     + 2.1707E+03, 2.1750E+03, 2.1794E+03, 2.1837E+03, 2.1881E+03,
     +  2.1924E+03, 2.1968E+03, 2.2011E+03, 2.2055E+03, 2.2099E+03,
     + 2.2143E+03, 2.2186E+03, 2.2230E+03, 2.2274E+03, 2.2318E+03,
     +  2.2362E+03, 2.2406E+03, 2.2450E+03, 2.2494E+03, 2.2538E+03,
     + 2.2582E+03, 2.2626E+03, 2.2671E+03, 2.2715E+03, 2.2759E+03,
     +  2.2804E+03, 2.2848E+03, 2.2892E+03, 2.2937E+03, 2.2981E+03,
     + 2.3026E+03, 2.3071E+03, 2.3115E+03, 2.3160E+03, 2.3205E+03,
     +  2.3250E+03, 2.3294E+03, 2.3339E+03, 2.3384E+03, 2.3429E+03,
     + 2.3474E+03, 2.3519E+03, 2.3564E+03, 2.3609E+03, 2.3654E+03,
     +  2.3700E+03, 2.3745E+03, 2.3790E+03, 2.3835E+03, 2.3881E+03,
     + 2.3926E+03, 2.3972E+03, 2.4017E+03, 2.4063E+03, 2.4108E+03,
     +  2.4154E+03, 2.4200E+03, 2.4245E+03, 2.4291E+03, 2.4337E+03,
     + 2.4383E+03, 2.4429E+03, 2.4474E+03, 2.4520E+03, 2.4566E+03,
     +  2.4612E+03, 2.4659E+03, 2.4705E+03, 2.4751E+03, 2.4797E+03,
     + 2.4843E+03, 2.4890E+03, 2.4936E+03, 2.4982E+03, 2.5029E+03,
     +  2.5075E+03, 2.5122E+03, 2.5168E+03, 2.5215E+03, 2.5261E+03,
     + 2.5308E+03, 2.5355E+03, 2.5402E+03, 2.5448E+03, 2.5495E+03,
     +  2.5542E+03, 2.5589E+03, 2.5636E+03, 2.5683E+03, 2.5730E+03,
     + 2.5777E+03, 2.5824E+03, 2.5872E+03, 2.5919E+03, 2.5966E+03,
     +  2.6013E+03, 2.6061E+03, 2.6108E+03, 2.6156E+03, 2.6203E+03,
     + 2.6251E+03, 2.6298E+03, 2.6346E+03, 2.6394E+03, 2.6441E+03,
     +  2.6489E+03, 2.6537E+03, 2.6585E+03, 2.6633E+03, 2.6681E+03,
     + 2.6728E+03, 2.6776E+03, 2.6825E+03, 2.6873E+03, 2.6921E+03,
     +  2.6969E+03, 2.7017E+03, 2.7065E+03, 2.7114E+03, 2.7162E+03,
     + 2.7211E+03, 2.7259E+03, 2.7307E+03, 2.7356E+03, 2.7405E+03,
     +  2.7453E+03, 2.7502E+03, 2.7551E+03, 2.7599E+03, 2.7648E+03,
     + 2.7697E+03, 2.7746E+03, 2.7795E+03, 2.7844E+03, 2.7893E+03,
     +  2.7942E+03, 2.7991E+03, 2.8040E+03, 2.8089E+03, 2.8138E+03,
     + 2.8188E+03, 2.8237E+03, 2.8286E+03, 2.8336E+03, 2.8385E+03,
     +  2.8435E+03, 2.8484E+03, 2.8534E+03, 2.8583E+03, 2.8633E+03,
     + 2.8683E+03, 2.8733E+03, 2.8782E+03, 2.8832E+03, 2.8882E+03,
     +  2.8932E+03, 2.8982E+03, 2.9032E+03, 2.9082E+03, 2.9132E+03,
     + 2.9182E+03, 2.9233E+03, 2.9283E+03, 2.9333E+03, 2.9383E+03,
     +  2.9434E+03, 2.9484E+03, 2.9535E+03, 2.9585E+03, 2.9636E+03,
     + 2.9686E+03, 2.9737E+03, 2.9788E+03, 2.9838E+03, 2.9889E+03,
     +  2.9940E+03, 2.9991E+03, 3.0042E+03, 3.0093E+03, 3.0144E+03,
     + 3.0195E+03, 3.0246E+03, 3.0297E+03, 3.0348E+03, 3.0399E+03,
     +  3.0450E+03, 3.0502E+03, 3.0553E+03, 3.0604E+03, 3.0656E+03,
     + 3.0707E+03, 3.0759E+03, 3.0810E+03, 3.0862E+03, 3.0914E+03,
     +  3.0965E+03, 3.1017E+03, 3.1069E+03, 3.1121E+03, 3.1173E+03,
     + 3.1225E+03, 3.1277E+03, 3.1329E+03, 3.1381E+03, 3.1433E+03,
     +  3.1485E+03, 3.1537E+03, 3.1589E+03, 3.1642E+03, 3.1694E+03,
     + 3.1746E+03, 3.1799E+03, 3.1851E+03, 3.1904E+03, 3.1956E+03,
     +  3.2009E+03, 3.2061E+03, 3.2114E+03, 3.2167E+03, 3.2220E+03,
     + 3.2272E+03, 3.2325E+03, 3.2378E+03, 3.2431E+03, 3.2484E+03,
     +  3.2537E+03, 3.2590E+03, 3.2643E+03, 3.2696E+03, 3.2750E+03,
     + 3.2803E+03, 3.2856E+03, 3.2910E+03, 3.2963E+03, 3.3016E+03,
     +  3.3070E+03, 3.3123E+03, 3.3177E+03, 3.3231E+03, 3.3284E+03,
     + 3.3338E+03, 3.3392E+03, 3.3446E+03, 3.3499E+03, 3.3553E+03,
     +  3.3607E+03, 3.3661E+03, 3.3715E+03, 3.3769E+03, 3.3823E+03,
     + 3.3878E+03, 3.3932E+03, 3.3986E+03, 3.4040E+03, 3.4095E+03,
     +  3.4149E+03, 3.4204E+03, 3.4258E+03, 3.4313E+03, 3.4367E+03,
     + 3.4422E+03, 3.4476E+03, 3.4531E+03, 3.4586E+03, 3.4641E+03,
     +  3.4696E+03, 3.4750E+03, 3.4805E+03, 3.4860E+03, 3.4915E+03,
     + 3.4970E+03, 3.5026E+03, 3.5081E+03, 3.5136E+03, 3.5191E+03,
     +  3.5246E+03, 3.5302E+03, 3.5357E+03, 3.5413E+03, 3.5468E+03,
     + 3.5524E+03, 3.5579E+03, 3.5635E+03, 3.5690E+03, 3.5746E+03,
     +  3.5802E+03, 3.5858E+03, 3.5914E+03, 3.5969E+03, 3.6025E+03,
     + 3.6081E+03, 3.6137E+03, 3.6193E+03, 3.6250E+03, 3.6306E+03,
     +  3.6362E+03, 3.6418E+03, 3.6474E+03, 3.6531E+03, 3.6587E+03,
     + 3.6644E+03, 3.6700E+03, 3.6757E+03, 3.6813E+03, 3.6870E+03,
     +  3.6926E+03, 3.6983E+03, 3.7040E+03, 3.7097E+03, 3.7154E+03,
     + 3.7210E+03, 3.7267E+03, 3.7324E+03, 3.7381E+03, 3.7438E+03,
     +  3.7496E+03, 3.7553E+03, 3.7610E+03, 3.7667E+03, 3.7725E+03,
     + 3.7782E+03, 3.7839E+03, 3.7897E+03, 3.7954E+03, 3.8012E+03,
     +  3.8069E+03, 3.8127E+03, 3.8185E+03, 3.8242E+03, 3.8300E+03,
     + 3.8358E+03, 3.8416E+03, 3.8474E+03, 3.8532E+03, 3.8590E+03,
     +  3.8648E+03, 3.8706E+03, 3.8764E+03, 3.8822E+03, 3.8880E+03,
     + 3.8939E+03, 3.8997E+03, 3.9055E+03, 3.9114E+03, 3.9172E+03,
     +  3.9231E+03, 3.9289E+03, 3.9348E+03, 3.9406E+03, 3.9465E+03,
     + 3.9524E+03, 3.9583E+03, 3.9641E+03, 3.9700E+03, 3.9759E+03,
     +  3.9818E+03, 3.9877E+03, 3.9936E+03, 3.9995E+03, 4.0055E+03,
     + 4.0114E+03, 4.0173E+03, 4.0232E+03, 4.0292E+03, 4.0351E+03,
     +  4.0410E+03, 4.0470E+03, 4.0529E+03, 4.0589E+03, 4.0649E+03,
     + 4.0708E+03, 4.0768E+03, 4.0828E+03, 4.0887E+03, 4.0947E+03,
     +  4.1007E+03, 4.1067E+03, 4.1127E+03, 4.1187E+03, 4.1247E+03,
     + 4.1307E+03, 4.1368E+03, 4.1428E+03, 4.1488E+03, 4.1548E+03,
     +  4.1609E+03, 4.1669E+03, 4.1730E+03, 4.1790E+03, 4.1851E+03,
     + 4.1911E+03, 4.1972E+03, 4.2032E+03, 4.2093E+03, 4.2154E+03,
     +  4.2215E+03, 4.2276E+03, 4.2337E+03, 4.2398E+03, 4.2459E+03,
     + 4.2520E+03, 4.2581E+03, 4.2642E+03, 4.2703E+03, 4.2764E+03,
     +  4.2826E+03, 4.2887E+03, 4.2948E+03, 4.3010E+03, 4.3071E+03,
     + 4.3133E+03, 4.3194E+03, 4.3256E+03, 4.3318E+03, 4.3379E+03,
     +  4.3441E+03, 4.3503E+03, 4.3565E+03, 4.3627E+03, 4.3689E+03,
     + 4.3751E+03, 4.3813E+03, 4.3875E+03, 4.3937E+03, 4.3999E+03,
     +  4.4061E+03, 4.4123E+03, 4.4186E+03, 4.4248E+03, 4.4311E+03,
     + 4.4373E+03, 4.4436E+03, 4.4498E+03, 4.4561E+03, 4.4623E+03,
     +  4.4686E+03, 4.4749E+03, 4.4812E+03, 4.4874E+03, 4.4937E+03,
     + 4.5000E+03, 4.5063E+03, 4.5126E+03, 4.5189E+03, 4.5252E+03,
     +  4.5316E+03, 4.5379E+03, 4.5442E+03, 4.5505E+03, 4.5569E+03,
     + 4.5632E+03, 4.5696E+03, 4.5759E+03, 4.5823E+03, 4.5886E+03,
     +  4.5950E+03, 4.6014E+03, 4.6077E+03, 4.6141E+03, 4.6205E+03,
     + 4.6269E+03, 4.6333E+03, 4.6397E+03, 4.6461E+03, 4.6525E+03,
     +  4.6589E+03, 4.6653E+03, 4.6717E+03, 4.6781E+03, 4.6846E+03,
     + 4.6910E+03, 4.6974E+03, 4.7039E+03, 4.7103E+03, 4.7168E+03,
     +  4.7232E+03, 4.7297E+03, 4.7362E+03, 4.7427E+03, 4.7491E+03,
     + 4.7556E+03, 4.7621E+03, 4.7686E+03, 4.7751E+03, 4.7816E+03,
     +  4.7881E+03, 4.7946E+03, 4.8011E+03, 4.8076E+03, 4.8142E+03,
     + 4.8207E+03, 4.8272E+03, 4.8338E+03, 4.8403E+03, 4.8469E+03,
     +  4.8534E+03, 4.8600E+03, 4.8665E+03, 4.8731E+03, 4.8797E+03,
     + 4.8862E+03, 4.8928E+03, 4.8994E+03, 4.9060E+03, 4.9126E+03,
     +  4.9192E+03, 4.9258E+03, 4.9324E+03, 4.9390E+03, 4.9457E+03,
     + 4.9523E+03, 4.9589E+03, 4.9655E+03, 4.9722E+03, 4.9788E+03,
     +  4.9855E+03, 4.9921E+03, 4.9988E+03, 5.0054E+03, 5.0121E+03,
     + 5.0188E+03, 5.0255E+03, 5.0321E+03, 5.0388E+03, 5.0455E+03,
     +  5.0522E+03, 5.0589E+03, 5.0656E+03, 5.0723E+03, 5.0790E+03,
     + 5.0858E+03, 5.0925E+03, 5.0992E+03, 5.1060E+03, 5.1127E+03,
     +  5.1194E+03, 5.1262E+03, 5.1329E+03, 5.1397E+03, 5.1465E+03,
     + 5.1532E+03, 5.1600E+03, 5.1668E+03, 5.1736E+03, 5.1803E+03,
     +  5.1871E+03, 5.1939E+03, 5.2007E+03, 5.2075E+03, 5.2143E+03,
     + 5.2212E+03, 5.2280E+03, 5.2348E+03, 5.2416E+03, 5.2485E+03,
     +  5.2553E+03, 5.2622E+03, 5.2690E+03, 5.2759E+03, 5.2827E+03,
     + 5.2896E+03, 5.2965E+03, 5.3033E+03, 5.3102E+03, 5.3171E+03,
     +  5.3240E+03, 5.3309E+03, 5.3378E+03, 5.3447E+03, 5.3516E+03,
     + 5.3585E+03, 5.3654E+03, 5.3723E+03, 5.3792E+03, 5.3862E+03,
     +  5.3931E+03, 5.4001E+03, 5.4070E+03, 5.4139E+03, 5.4209E+03,
     + 5.4279E+03, 5.4348E+03, 5.4418E+03, 5.4488E+03, 5.4557E+03,
     +  5.4627E+03, 5.4697E+03, 5.4767E+03, 5.4837E+03, 5.4907E+03,
     + 5.4977E+03, 5.5047E+03, 5.5118E+03, 5.5188E+03, 5.5258E+03,
     +  5.5328E+03, 5.5399E+03, 5.5469E+03, 5.5540E+03, 5.5610E+03,
     + 5.5681E+03, 5.5751E+03, 5.5822E+03, 5.5893E+03, 5.5963E+03,
     +  5.6034E+03, 5.6105E+03, 5.6176E+03, 5.6247E+03, 5.6318E+03,
     + 5.6389E+03, 5.6460E+03, 5.6531E+03, 5.6602E+03, 5.6674E+03,
     +  5.6745E+03, 5.6816E+03, 5.6888E+03, 5.6959E+03, 5.7030E+03,
     + 5.7102E+03, 5.7174E+03, 5.7245E+03, 5.7317E+03, 5.7389E+03,
     +  5.7460E+03, 5.7532E+03, 5.7604E+03, 5.7676E+03, 5.7748E+03,
     + 5.7820E+03, 5.7892E+03, 5.7964E+03, 5.8036E+03, 5.8108E+03,
     +  5.8181E+03, 5.8253E+03, 5.8325E+03, 5.8398E+03, 5.8470E+03,
     + 5.8543E+03, 5.8615E+03, 5.8688E+03, 5.8761E+03, 5.8833E+03,
     +  5.8906E+03, 5.8979E+03, 5.9052E+03, 5.9124E+03, 5.9197E+03,
     + 5.9270E+03, 5.9343E+03, 5.9416E+03, 5.9490E+03, 5.9563E+03,
     +  5.9636E+03, 5.9709E+03, 5.9783E+03, 5.9856E+03, 5.9929E+03,
     + 6.0003E+03, 6.0076E+03, 6.0150E+03, 6.0224E+03, 6.0297E+03,
     +  6.0371E+03, 6.0445E+03, 6.0519E+03, 6.0592E+03, 6.0666E+03,
     + 6.0740E+03, 6.0814E+03, 6.0888E+03, 6.0962E+03, 6.1037E+03,
     +  6.1111E+03, 6.1185E+03, 6.1259E+03, 6.1334E+03, 6.1408E+03,
     + 6.1483E+03, 6.1557E+03, 6.1632E+03, 6.1706E+03, 6.1781E+03,
     +  6.1855E+03, 6.1930E+03, 6.2005E+03, 6.2080E+03, 6.2155E+03,
     + 6.2230E+03, 6.2305E+03, 6.2380E+03, 6.2455E+03, 6.2530E+03,
     +  6.2605E+03, 6.2680E+03, 6.2756E+03, 6.2831E+03, 6.2906E+03,
     + 6.2982E+03, 6.3057E+03, 6.3133E+03, 6.3208E+03, 6.3284E+03,
     +  6.3360E+03, 6.3435E+03, 6.3511E+03, 6.3587E+03, 6.3663E+03,
     + 6.3739E+03, 6.3815E+03, 6.3891E+03, 6.3967E+03, 6.4043E+03,
     +  6.4119E+03, 6.4195E+03, 6.4271E+03, 6.4348E+03, 6.4424E+03,
     + 6.4500E+03, 6.4577E+03, 6.4653E+03, 6.4730E+03, 6.4806E+03,
     +  6.4883E+03, 6.4960E+03, 6.5037E+03, 6.5113E+03, 6.5190E+03,
     + 6.5267E+03, 6.5344E+03, 6.5421E+03, 6.5498E+03, 6.5575E+03,
     +  6.5652E+03, 6.5729E+03, 6.5807E+03, 6.5884E+03, 6.5961E+03,
     + 6.6039E+03, 6.6116E+03, 6.6194E+03, 6.6271E+03, 6.6349E+03,
     +  6.6426E+03, 6.6504E+03, 6.6582E+03, 6.6660E+03, 6.6737E+03,
     + 6.6815E+03, 6.6893E+03, 6.6971E+03, 6.7049E+03, 6.7127E+03,
     +  6.7205E+03, 6.7284E+03, 6.7362E+03, 6.7440E+03, 6.7518E+03,
     + 6.7597E+03, 6.7675E+03, 6.7754E+03, 6.7832E+03, 6.7911E+03,
     +  6.7989E+03, 6.8068E+03, 6.8147E+03, 6.8225E+03, 6.8304E+03,
     + 6.8383E+03, 6.8462E+03, 6.8541E+03, 6.8620E+03, 6.8699E+03,
     +  6.8778E+03, 6.8857E+03, 6.8937E+03, 6.9016E+03, 6.9095E+03,
     + 6.9174E+03, 6.9254E+03, 6.9333E+03, 6.9413E+03, 6.9492E+03,
     +  6.9572E+03, 6.9652E+03, 6.9731E+03, 6.9811E+03, 6.9891E+03,
     + 6.9971E+03, 7.0051E+03, 7.0131E+03, 7.0211E+03, 7.0291E+03,
     +  7.0371E+03, 7.0451E+03, 7.0531E+03, 7.0611E+03, 7.0692E+03,
     + 7.0772E+03, 7.0852E+03, 7.0933E+03, 7.1013E+03, 7.1094E+03,
     +  7.1174E+03, 7.1255E+03, 7.1336E+03, 7.1416E+03, 7.1497E+03,
     + 7.1578E+03, 7.1659E+03, 7.1740E+03, 7.1821E+03, 7.1902E+03,
     +  7.1983E+03, 7.2064E+03, 7.2145E+03, 7.2227E+03, 7.2308E+03,
     + 7.2389E+03, 7.2471E+03, 7.2552E+03, 7.2633E+03, 7.2715E+03,
     +  7.2797E+03, 7.2878E+03, 7.2960E+03, 7.3042E+03, 7.3123E+03,
     + 7.3205E+03, 7.3287E+03, 7.3369E+03, 7.3451E+03, 7.3533E+03,
     +  7.3615E+03, 7.3697E+03, 7.3779E+03, 7.3862E+03, 7.3944E+03,
     + 7.4026E+03, 7.4109E+03, 7.4191E+03, 7.4274E+03, 7.4356E+03,
     +  7.4439E+03, 7.4521E+03, 7.4604E+03, 7.4687E+03, 7.4769E+03,
     + 7.4852E+03, 7.4935E+03, 7.5018E+03, 7.5101E+03, 7.5184E+03,
     +  7.5267E+03, 7.5350E+03, 7.5433E+03, 7.5517E+03, 7.5600E+03,
     + 7.5683E+03, 7.5767E+03, 7.5850E+03, 7.5933E+03, 7.6017E+03,
     +  7.6101E+03, 7.6184E+03, 7.6268E+03, 7.6352E+03, 7.6435E+03,
     + 7.6519E+03, 7.6603E+03, 7.6687E+03, 7.6771E+03, 7.6855E+03,
     +  7.6939E+03, 7.7023E+03, 7.7107E+03, 7.7191E+03, 7.7276E+03,
     + 7.7360E+03, 7.7444E+03, 7.7529E+03, 7.7613E+03, 7.7698E+03,
     +  7.7782E+03, 7.7867E+03, 7.7952E+03, 7.8036E+03, 7.8121E+03,
     + 7.8206E+03, 7.8291E+03, 7.8376E+03, 7.8461E+03, 7.8546E+03,
     +  7.8631E+03, 7.8716E+03, 7.8801E+03, 7.8886E+03, 7.8972E+03,
     + 7.9057E+03, 7.9142E+03, 7.9228E+03, 7.9313E+03, 7.9399E+03,
     +  7.9484E+03, 7.9570E+03, 7.9655E+03, 7.9741E+03, 7.9827E+03,
     + 7.9913E+03, 7.9999E+03, 8.0084E+03, 8.0170E+03, 8.0256E+03,
     +  8.0343E+03, 8.0429E+03, 8.0515E+03, 8.0601E+03, 8.0687E+03,
     + 8.0774E+03, 8.0860E+03, 8.0946E+03, 8.1033E+03, 8.1119E+03,
     +  8.1206E+03, 8.1292E+03, 8.1379E+03, 8.1466E+03, 8.1553E+03,
     + 8.1639E+03, 8.1726E+03, 8.1813E+03, 8.1900E+03, 8.1987E+03,
     +  8.2074E+03, 8.2161E+03, 8.2249E+03, 8.2336E+03, 8.2423E+03,
     + 8.2510E+03, 8.2598E+03, 8.2685E+03, 8.2773E+03, 8.2860E+03,
     +  8.2948E+03, 8.3035E+03, 8.3123E+03, 8.3211E+03, 8.3298E+03,
     + 8.3386E+03, 8.3474E+03, 8.3562E+03, 8.3650E+03, 8.3738E+03,
     +  8.3826E+03, 8.3914E+03, 8.4002E+03, 8.4090E+03, 8.4179E+03,
     + 8.4267E+03, 8.4355E+03, 8.4444E+03, 8.4532E+03, 8.4621E+03,
     +  8.4709E+03, 8.4798E+03, 8.4887E+03, 8.4975E+03, 8.5064E+03,
     + 8.5153E+03, 8.5242E+03, 8.5331E+03, 8.5420E+03, 8.5509E+03,
     +  8.5598E+03, 8.5687E+03, 8.5776E+03, 8.5865E+03, 8.5954E+03,
     + 8.6044E+03, 8.6133E+03, 8.6223E+03, 8.6312E+03, 8.6402E+03,
     +  8.6491E+03, 8.6581E+03, 8.6670E+03, 8.6760E+03, 8.6850E+03,
     + 8.6940E+03, 8.7030E+03, 8.7119E+03, 8.7209E+03, 8.7299E+03,
     +  8.7390E+03, 8.7480E+03, 8.7570E+03, 8.7660E+03, 8.7750E+03,
     + 8.7841E+03, 8.7931E+03, 8.8021E+03, 8.8112E+03, 8.8202E+03,
     +  8.8293E+03, 8.8384E+03, 8.8474E+03, 8.8565E+03, 8.8656E+03,
     + 8.8747E+03, 8.8837E+03, 8.8928E+03, 8.9019E+03, 8.9110E+03,
     +  8.9201E+03, 8.9293E+03, 8.9384E+03, 8.9475E+03, 8.9566E+03,
     + 8.9658E+03, 8.9749E+03, 8.9840E+03, 8.9932E+03, 9.0023E+03,
     +  9.0115E+03, 9.0207E+03, 9.0298E+03, 9.0390E+03, 9.0482E+03,
     + 9.0574E+03, 9.0665E+03, 9.0757E+03, 9.0849E+03, 9.0941E+03,
     +  9.1033E+03, 9.1126E+03, 9.1218E+03, 9.1310E+03, 9.1402E+03,
     + 9.1495E+03, 9.1587E+03, 9.1679E+03, 9.1772E+03, 9.1864E+03,
     +  9.1957E+03, 9.2050E+03, 9.2142E+03, 9.2235E+03, 9.2328E+03,
     + 9.2421E+03, 9.2514E+03, 9.2606E+03, 9.2699E+03, 9.2792E+03,
     +  9.2886E+03, 9.2979E+03, 9.3072E+03, 9.3165E+03, 9.3258E+03,
     + 9.3352E+03, 9.3445E+03, 9.3539E+03, 9.3632E+03, 9.3726E+03,
     +  9.3819E+03, 9.3913E+03, 9.4006E+03, 9.4100E+03, 9.4194E+03,
     + 9.4288E+03, 9.4382E+03, 9.4476E+03, 9.4570E+03, 9.4664E+03,
     +  9.4758E+03, 9.4852E+03, 9.4946E+03, 9.5040E+03, 9.5135E+03,
     + 9.5229E+03, 9.5323E+03, 9.5418E+03, 9.5512E+03, 9.5607E+03,
     +  9.5701E+03, 9.5796E+03, 9.5891E+03, 9.5986E+03, 9.6080E+03,
     + 9.6175E+03, 9.6270E+03, 9.6365E+03, 9.6460E+03, 9.6555E+03,
     +  9.6650E+03, 9.6745E+03, 9.6840E+03, 9.6936E+03, 9.7031E+03,
     + 9.7126E+03, 9.7222E+03, 9.7317E+03, 9.7413E+03, 9.7508E+03,
     +  9.7604E+03, 9.7700E+03, 9.7795E+03, 9.7891E+03, 9.7987E+03,
     + 9.8083E+03, 9.8179E+03, 9.8274E+03, 9.8370E+03, 9.8467E+03,
     +  9.8563E+03, 9.8659E+03, 9.8755E+03, 9.8851E+03, 9.8948E+03,
     + 9.9044E+03, 9.9140E+03, 9.9237E+03, 9.9333E+03, 9.9430E+03,
     +  9.9526E+03, 9.9623E+03, 9.9720E+03, 9.9817E+03, 9.9913E+03,
     + 1.0001E+04, 1.0011E+04, 1.0020E+04, 1.0030E+04, 1.0040E+04,
     +  1.0050E+04, 1.0059E+04, 1.0069E+04, 1.0079E+04, 1.0088E+04,
     + 1.0098E+04, 1.0108E+04, 1.0118E+04, 1.0127E+04, 1.0137E+04,
     +  1.0147E+04, 1.0157E+04, 1.0166E+04, 1.0176E+04, 1.0186E+04,
     + 1.0196E+04, 1.0206E+04, 1.0215E+04, 1.0225E+04, 1.0235E+04,
     +  1.0245E+04, 1.0255E+04, 1.0264E+04, 1.0274E+04, 1.0284E+04,
     + 1.0294E+04, 1.0304E+04, 1.0314E+04, 1.0323E+04, 1.0333E+04,
     +  1.0343E+04, 1.0353E+04, 1.0363E+04, 1.0373E+04, 1.0383E+04,
     + 1.0393E+04, 1.0402E+04, 1.0412E+04, 1.0422E+04, 1.0432E+04,
     +  1.0442E+04, 1.0452E+04, 1.0462E+04, 1.0472E+04, 1.0482E+04,
     + 1.0492E+04, 1.0502E+04, 1.0512E+04, 1.0522E+04, 1.0532E+04,
     +  1.0542E+04, 1.0551E+04, 1.0561E+04, 1.0571E+04, 1.0581E+04,
     + 1.0591E+04, 1.0601E+04, 1.0611E+04, 1.0621E+04, 1.0631E+04,
     +  1.0641E+04, 1.0651E+04, 1.0661E+04, 1.0672E+04, 1.0682E+04,
     + 1.0692E+04, 1.0702E+04, 1.0712E+04, 1.0722E+04, 1.0732E+04,
     +  1.0742E+04, 1.0752E+04, 1.0762E+04, 1.0772E+04, 1.0782E+04,
     + 1.0792E+04, 1.0802E+04, 1.0812E+04, 1.0823E+04, 1.0833E+04,
     +  1.0843E+04, 1.0853E+04, 1.0863E+04, 1.0873E+04, 1.0883E+04,
     + 1.0893E+04, 1.0904E+04, 1.0914E+04, 1.0924E+04, 1.0934E+04,
     +  1.0944E+04, 1.0954E+04, 1.0965E+04, 1.0975E+04, 1.0985E+04,
     + 1.0995E+04, 1.1005E+04, 1.1016E+04, 1.1026E+04, 1.1036E+04,
     +  1.1046E+04, 1.1056E+04, 1.1067E+04, 1.1077E+04, 1.1087E+04,
     + 1.1097E+04, 1.1108E+04, 1.1118E+04, 1.1128E+04, 1.1138E+04,
     +  1.1149E+04, 1.1159E+04, 1.1169E+04, 1.1179E+04, 1.1190E+04,
     + 1.1200E+04, 1.1210E+04, 1.1221E+04, 1.1231E+04, 1.1241E+04,
     +  1.1252E+04, 1.1262E+04, 1.1272E+04, 1.1282E+04, 1.1293E+04,
     + 1.1303E+04, 1.1314E+04, 1.1324E+04, 1.1334E+04, 1.1345E+04,
     +  1.1355E+04, 1.1365E+04, 1.1376E+04, 1.1386E+04, 1.1396E+04,
     + 1.1407E+04, 1.1417E+04, 1.1428E+04, 1.1438E+04, 1.1448E+04,
     +  1.1459E+04, 1.1469E+04, 1.1480E+04, 1.1490E+04, 1.1501E+04,
     + 1.1511E+04, 1.1522E+04, 1.1532E+04, 1.1542E+04, 1.1553E+04,
     +  1.1563E+04, 1.1574E+04, 1.1584E+04, 1.1595E+04, 1.1605E+04,
     + 1.1616E+04, 1.1626E+04, 1.1637E+04, 1.1647E+04, 1.1658E+04,
     +  1.1668E+04, 1.1679E+04, 1.1689E+04, 1.1700E+04, 1.1710E+04,
     + 1.1721E+04, 1.1732E+04, 1.1742E+04, 1.1753E+04, 1.1763E+04,
     +  1.1774E+04, 1.1784E+04, 1.1795E+04, 1.1806E+04, 1.1816E+04,
     + 1.1827E+04, 1.1837E+04, 1.1848E+04, 1.1858E+04, 1.1869E+04,
     +  1.1880E+04, 1.1890E+04, 1.1901E+04, 1.1912E+04, 1.1922E+04,
     + 1.1933E+04, 1.1944E+04, 1.1954E+04, 1.1965E+04, 1.1976E+04,
     +  1.1986E+04, 1.1997E+04, 1.2008E+04, 1.2018E+04, 1.2029E+04,
     + 1.2040E+04, 1.2050E+04, 1.2061E+04, 1.2072E+04, 1.2082E+04,
     +  1.2093E+04, 1.2104E+04, 1.2115E+04, 1.2125E+04, 1.2136E+04,
     + 1.2147E+04, 1.2158E+04, 1.2168E+04, 1.2179E+04, 1.2190E+04,
     +  1.2201E+04, 1.2211E+04, 1.2222E+04, 1.2233E+04, 1.2244E+04,
     + 1.2255E+04, 1.2265E+04, 1.2276E+04, 1.2287E+04, 1.2298E+04,
     +  1.2309E+04, 1.2319E+04, 1.2330E+04, 1.2341E+04, 1.2352E+04,
     + 1.2363E+04, 1.2374E+04, 1.2384E+04, 1.2395E+04, 1.2406E+04,
     +  1.2417E+04, 1.2428E+04, 1.2439E+04, 1.2450E+04, 1.2461E+04,
     + 1.2472E+04, 1.2482E+04, 1.2493E+04, 1.2504E+04, 1.2515E+04,
     +  1.2526E+04, 1.2537E+04, 1.2548E+04, 1.2559E+04, 1.2570E+04,
     + 1.2581E+04, 1.2592E+04, 1.2603E+04, 1.2614E+04, 1.2625E+04,
     +  1.2636E+04, 1.2647E+04, 1.2658E+04, 1.2669E+04, 1.2679E+04,
     + 1.2690E+04, 1.2701E+04, 1.2713E+04, 1.2724E+04, 1.2735E+04,
     +  1.2746E+04, 1.2757E+04, 1.2768E+04, 1.2779E+04, 1.2790E+04,
     + 1.2801E+04, 1.2812E+04, 1.2823E+04, 1.2834E+04, 1.2845E+04,
     +  1.2856E+04, 1.2867E+04, 1.2878E+04, 1.2889E+04, 1.2900E+04,
     + 1.2911E+04, 1.2923E+04, 1.2934E+04, 1.2945E+04, 1.2956E+04,
     +  1.2967E+04, 1.2978E+04, 1.2989E+04, 1.3000E+04, 1.3012E+04,
     + 1.3023E+04, 1.3034E+04, 1.3045E+04, 1.3056E+04, 1.3067E+04,
     +  1.3079E+04, 1.3090E+04, 1.3101E+04, 1.3112E+04, 1.3123E+04,
     + 1.3134E+04, 1.3146E+04, 1.3157E+04, 1.3168E+04, 1.3179E+04,
     +  1.3191E+04, 1.3202E+04, 1.3213E+04, 1.3224E+04, 1.3235E+04,
     + 1.3247E+04, 1.3258E+04, 1.3269E+04, 1.3280E+04, 1.3292E+04,
     +  1.3303E+04, 1.3314E+04, 1.3326E+04, 1.3337E+04, 1.3348E+04,
     + 1.3359E+04, 1.3371E+04, 1.3382E+04, 1.3393E+04, 1.3405E+04,
     +  1.3416E+04, 1.3427E+04, 1.3439E+04, 1.3450E+04, 1.3461E+04,
     + 1.3473E+04, 1.3484E+04, 1.3495E+04, 1.3507E+04, 1.3518E+04,
     +  1.3530E+04, 1.3541E+04, 1.3552E+04, 1.3564E+04, 1.3575E+04,
     + 1.3586E+04, 1.3598E+04, 1.3609E+04, 1.3621E+04, 1.3632E+04,
     +  1.3643E+04, 1.3655E+04, 1.3666E+04, 1.3678E+04, 1.3689E+04,
     + 1.3701E+04, 1.3712E+04, 1.3724E+04, 1.3735E+04, 1.3747E+04,
     +  1.3758E+04, 1.3769E+04, 1.3781E+04, 1.3792E+04, 1.3804E+04,
     + 1.3815E+04, 1.3827E+04, 1.3838E+04, 1.3850E+04, 1.3861E+04,
     +  1.3873E+04, 1.3885E+04, 1.3896E+04, 1.3908E+04, 1.3919E+04,
     + 1.3931E+04, 1.3942E+04, 1.3954E+04, 1.3965E+04, 1.3977E+04,
     +  1.3988E+04, 1.4000E+04, 1.4012E+04, 1.4023E+04, 1.4035E+04,
     + 1.4046E+04, 1.4058E+04, 1.4070E+04, 1.4081E+04, 1.4093E+04,
     +  1.4104E+04, 1.4116E+04, 1.4128E+04, 1.4139E+04, 1.4151E+04,
     + 1.4163E+04, 1.4174E+04, 1.4186E+04, 1.4198E+04, 1.4209E+04,
     +  1.4221E+04, 1.4233E+04, 1.4244E+04, 1.4256E+04, 1.4268E+04,
     + 1.4279E+04, 1.4291E+04, 1.4303E+04, 1.4315E+04, 1.4326E+04,
     +  1.4338E+04, 1.4350E+04, 1.4361E+04, 1.4373E+04, 1.4385E+04,
     + 1.4397E+04, 1.4408E+04, 1.4420E+04, 1.4432E+04, 1.4444E+04,
     +  1.4455E+04, 1.4467E+04, 1.4479E+04, 1.4491E+04, 1.4503E+04,
     + 1.4514E+04, 1.4526E+04, 1.4538E+04, 1.4550E+04, 1.4562E+04,
     +  1.4573E+04, 1.4585E+04, 1.4597E+04, 1.4609E+04, 1.4621E+04,
     + 1.4633E+04, 1.4644E+04, 1.4656E+04, 1.4668E+04, 1.4680E+04,
     +  1.4692E+04, 1.4704E+04, 1.4716E+04, 1.4728E+04, 1.4739E+04,
     + 1.4751E+04, 1.4763E+04, 1.4775E+04, 1.4787E+04, 1.4799E+04,
     +  1.4811E+04, 1.4823E+04, 1.4835E+04, 1.4847E+04, 1.4859E+04,
     + 1.4871E+04, 1.4883E+04, 1.4894E+04, 1.4906E+04, 1.4918E+04,
     +  1.4930E+04, 1.4942E+04, 1.4954E+04, 1.4966E+04, 1.4978E+04,
     + 1.4990E+04, 1.5002E+04, 1.5014E+04, 1.5026E+04, 1.5038E+04,
     +  1.5050E+04, 1.5062E+04, 1.5074E+04, 1.5086E+04, 1.5098E+04,
     + 1.5110E+04, 1.5123E+04, 1.5135E+04, 1.5147E+04, 1.5159E+04,
     +  1.5171E+04, 1.5183E+04, 1.5195E+04, 1.5207E+04, 1.5219E+04,
     + 1.5231E+04, 1.5243E+04, 1.5255E+04, 1.5268E+04, 1.5280E+04,
     +  1.5292E+04, 1.5304E+04, 1.5316E+04, 1.5328E+04, 1.5340E+04,
     + 1.5352E+04, 1.5365E+04, 1.5377E+04, 1.5389E+04, 1.5401E+04,
     +  1.5413E+04, 1.5425E+04, 1.5438E+04, 1.5450E+04, 1.5462E+04,
     + 1.5474E+04, 1.5486E+04, 1.5498E+04, 1.5511E+04, 1.5523E+04,
     +  1.5535E+04, 1.5547E+04, 1.5560E+04, 1.5572E+04, 1.5584E+04,
     + 1.5596E+04, 1.5609E+04, 1.5621E+04, 1.5633E+04, 1.5645E+04,
     +  1.5658E+04, 1.5670E+04, 1.5682E+04, 1.5694E+04, 1.5707E+04,
     + 1.5719E+04, 1.5731E+04, 1.5744E+04, 1.5756E+04, 1.5768E+04,
     +  1.5780E+04, 1.5793E+04, 1.5805E+04, 1.5817E+04, 1.5830E+04,
     + 1.5842E+04, 1.5854E+04, 1.5867E+04, 1.5879E+04, 1.5892E+04,
     +  1.5904E+04, 1.5916E+04, 1.5929E+04, 1.5941E+04, 1.5953E+04,
     + 1.5966E+04, 1.5978E+04, 1.5991E+04, 1.6003E+04, 1.6015E+04,
     +  1.6028E+04, 1.6040E+04, 1.6053E+04, 1.6065E+04, 1.6078E+04,
     + 1.6090E+04, 1.6102E+04, 1.6115E+04, 1.6127E+04, 1.6140E+04,
     +  1.6152E+04, 1.6165E+04, 1.6177E+04, 1.6190E+04, 1.6202E+04,
     + 1.6215E+04, 1.6227E+04, 1.6240E+04, 1.6252E+04, 1.6265E+04,
     +  1.6277E+04, 1.6290E+04, 1.6302E+04, 1.6315E+04, 1.6327E+04,
     + 1.6340E+04, 1.6352E+04, 1.6365E+04, 1.6377E+04, 1.6390E+04,
     +  1.6403E+04, 1.6415E+04, 1.6428E+04, 1.6440E+04, 1.6453E+04,
     + 1.6465E+04, 1.6478E+04, 1.6491E+04, 1.6503E+04, 1.6516E+04,
     +  1.6528E+04, 1.6541E+04, 1.6554E+04, 1.6566E+04, 1.6579E+04,
     + 1.6592E+04, 1.6604E+04, 1.6617E+04, 1.6629E+04, 1.6642E+04,
     +  1.6655E+04, 1.6667E+04, 1.6680E+04, 1.6693E+04, 1.6705E+04,
     + 1.6718E+04, 1.6731E+04, 1.6743E+04, 1.6756E+04, 1.6769E+04,
     +  1.6782E+04, 1.6794E+04, 1.6807E+04, 1.6820E+04, 1.6832E+04,
     + 1.6845E+04, 1.6858E+04, 1.6871E+04, 1.6883E+04, 1.6896E+04,
     +  1.6909E+04, 1.6922E+04, 1.6934E+04, 1.6947E+04, 1.6960E+04,
     + 1.6973E+04, 1.6986E+04, 1.6998E+04, 1.7011E+04, 1.7024E+04,
     +  1.7037E+04, 1.7050E+04, 1.7062E+04, 1.7075E+04, 1.7088E+04,
     + 1.7101E+04, 1.7114E+04, 1.7127E+04, 1.7139E+04, 1.7152E+04,
     +  1.7165E+04, 1.7178E+04, 1.7191E+04, 1.7204E+04, 1.7217E+04/
      data (QofT( 3,J),J=2409,4510)/ 1.7229E+04, 1.7242E+04,
     +  1.7255E+04, 1.7268E+04, 1.7281E+04, 1.7294E+04, 1.7307E+04,
     +  1.7320E+04,
     + 1.7333E+04, 1.7346E+04, 1.7358E+04, 1.7371E+04, 1.7384E+04,
     +  1.7397E+04, 1.7410E+04, 1.7423E+04, 1.7436E+04, 1.7449E+04,
     + 1.7462E+04, 1.7475E+04, 1.7488E+04, 1.7501E+04, 1.7514E+04,
     +  1.7527E+04, 1.7540E+04, 1.7553E+04, 1.7566E+04, 1.7579E+04,
     + 1.7592E+04, 1.7605E+04, 1.7618E+04, 1.7631E+04, 1.7644E+04,
     +  1.7657E+04, 1.7670E+04, 1.7683E+04, 1.7696E+04, 1.7709E+04,
     + 1.7722E+04, 1.7735E+04, 1.7748E+04, 1.7762E+04, 1.7775E+04,
     +  1.7788E+04, 1.7801E+04, 1.7814E+04, 1.7827E+04, 1.7840E+04,
     + 1.7853E+04, 1.7866E+04, 1.7879E+04, 1.7893E+04, 1.7906E+04,
     +  1.7919E+04, 1.7932E+04, 1.7945E+04, 1.7958E+04, 1.7971E+04,
     + 1.7985E+04, 1.7998E+04, 1.8011E+04, 1.8024E+04, 1.8037E+04,
     +  1.8050E+04, 1.8064E+04, 1.8077E+04, 1.8090E+04, 1.8103E+04,
     + 1.8116E+04, 1.8130E+04, 1.8143E+04, 1.8156E+04, 1.8169E+04,
     +  1.8182E+04, 1.8196E+04, 1.8209E+04, 1.8222E+04, 1.8235E+04,
     + 1.8249E+04, 1.8262E+04, 1.8275E+04, 1.8288E+04, 1.8302E+04,
     +  1.8315E+04, 1.8328E+04, 1.8342E+04, 1.8355E+04, 1.8368E+04,
     + 1.8381E+04, 1.8395E+04, 1.8408E+04, 1.8421E+04, 1.8435E+04,
     +  1.8448E+04, 1.8461E+04, 1.8475E+04, 1.8488E+04, 1.8501E+04,
     + 1.8515E+04, 1.8528E+04, 1.8541E+04, 1.8555E+04, 1.8568E+04,
     +  1.8582E+04, 1.8595E+04, 1.8608E+04, 1.8622E+04, 1.8635E+04,
     + 1.8649E+04, 1.8662E+04, 1.8675E+04, 1.8689E+04, 1.8702E+04,
     +  1.8716E+04, 1.8729E+04, 1.8742E+04, 1.8756E+04, 1.8769E+04,
     + 1.8783E+04, 1.8796E+04, 1.8810E+04, 1.8823E+04, 1.8837E+04,
     +  1.8850E+04, 1.8864E+04, 1.8877E+04, 1.8890E+04, 1.8904E+04,
     + 1.8917E+04, 1.8931E+04, 1.8944E+04, 1.8958E+04, 1.8971E+04,
     +  1.8985E+04, 1.8999E+04, 1.9012E+04, 1.9026E+04, 1.9039E+04,
     + 1.9053E+04, 1.9066E+04, 1.9080E+04, 1.9093E+04, 1.9107E+04,
     +  1.9120E+04, 1.9134E+04, 1.9148E+04, 1.9161E+04, 1.9175E+04,
     + 1.9188E+04, 1.9202E+04, 1.9215E+04, 1.9229E+04, 1.9243E+04,
     +  1.9256E+04, 1.9270E+04, 1.9284E+04, 1.9297E+04, 1.9311E+04,
     + 1.9324E+04, 1.9338E+04, 1.9352E+04, 1.9365E+04, 1.9379E+04,
     +  1.9393E+04, 1.9406E+04, 1.9420E+04, 1.9434E+04, 1.9447E+04,
     + 1.9461E+04, 1.9475E+04, 1.9488E+04, 1.9502E+04, 1.9516E+04,
     +  1.9530E+04, 1.9543E+04, 1.9557E+04, 1.9571E+04, 1.9584E+04,
     + 1.9598E+04, 1.9612E+04, 1.9626E+04, 1.9639E+04, 1.9653E+04,
     +  1.9667E+04, 1.9681E+04, 1.9694E+04, 1.9708E+04, 1.9722E+04,
     + 1.9736E+04, 1.9749E+04, 1.9763E+04, 1.9777E+04, 1.9791E+04,
     +  1.9805E+04, 1.9818E+04, 1.9832E+04, 1.9846E+04, 1.9860E+04,
     + 1.9874E+04, 1.9888E+04, 1.9901E+04, 1.9915E+04, 1.9929E+04,
     +  1.9943E+04, 1.9957E+04, 1.9971E+04, 1.9984E+04, 1.9998E+04,
     + 2.0012E+04, 2.0026E+04, 2.0040E+04, 2.0054E+04, 2.0068E+04,
     +  2.0082E+04, 2.0096E+04, 2.0109E+04, 2.0123E+04, 2.0137E+04,
     + 2.0151E+04, 2.0165E+04, 2.0179E+04, 2.0193E+04, 2.0207E+04,
     +  2.0221E+04, 2.0235E+04, 2.0249E+04, 2.0263E+04, 2.0277E+04,
     + 2.0291E+04, 2.0305E+04, 2.0319E+04, 2.0332E+04, 2.0346E+04,
     +  2.0360E+04, 2.0374E+04, 2.0388E+04, 2.0402E+04, 2.0416E+04,
     + 2.0430E+04, 2.0444E+04, 2.0459E+04, 2.0473E+04, 2.0487E+04,
     +  2.0501E+04, 2.0515E+04, 2.0529E+04, 2.0543E+04, 2.0557E+04,
     + 2.0571E+04, 2.0585E+04, 2.0599E+04, 2.0613E+04, 2.0627E+04,
     +  2.0641E+04, 2.0655E+04, 2.0669E+04, 2.0683E+04, 2.0698E+04,
     + 2.0712E+04, 2.0726E+04, 2.0740E+04, 2.0754E+04, 2.0768E+04,
     +  2.0782E+04, 2.0796E+04, 2.0811E+04, 2.0825E+04, 2.0839E+04,
     + 2.0853E+04, 2.0867E+04, 2.0881E+04, 2.0895E+04, 2.0910E+04,
     +  2.0924E+04, 2.0938E+04, 2.0952E+04, 2.0966E+04, 2.0981E+04,
     + 2.0995E+04, 2.1009E+04, 2.1023E+04, 2.1037E+04, 2.1052E+04,
     +  2.1066E+04, 2.1080E+04, 2.1094E+04, 2.1108E+04, 2.1123E+04,
     + 2.1137E+04, 2.1151E+04, 2.1165E+04, 2.1180E+04, 2.1194E+04,
     +  2.1208E+04, 2.1223E+04, 2.1237E+04, 2.1251E+04, 2.1265E+04,
     + 2.1280E+04, 2.1294E+04, 2.1308E+04, 2.1323E+04, 2.1337E+04,
     +  2.1351E+04, 2.1365E+04, 2.1380E+04, 2.1394E+04, 2.1408E+04,
     + 2.1423E+04, 2.1437E+04, 2.1451E+04, 2.1466E+04, 2.1480E+04,
     +  2.1495E+04, 2.1509E+04, 2.1523E+04, 2.1538E+04, 2.1552E+04,
     + 2.1566E+04, 2.1581E+04, 2.1595E+04, 2.1610E+04, 2.1624E+04,
     +  2.1638E+04, 2.1653E+04, 2.1667E+04, 2.1682E+04, 2.1696E+04,
     + 2.1710E+04, 2.1725E+04, 2.1739E+04, 2.1754E+04, 2.1768E+04,
     +  2.1783E+04, 2.1797E+04, 2.1812E+04, 2.1826E+04, 2.1841E+04,
     + 2.1855E+04, 2.1869E+04, 2.1884E+04, 2.1898E+04, 2.1913E+04,
     +  2.1927E+04, 2.1942E+04, 2.1956E+04, 2.1971E+04, 2.1985E+04,
     + 2.2000E+04, 2.2015E+04, 2.2029E+04, 2.2044E+04, 2.2058E+04,
     +  2.2073E+04, 2.2087E+04, 2.2102E+04, 2.2116E+04, 2.2131E+04,
     + 2.2145E+04, 2.2160E+04, 2.2175E+04, 2.2189E+04, 2.2204E+04,
     +  2.2218E+04, 2.2233E+04, 2.2248E+04, 2.2262E+04, 2.2277E+04,
     + 2.2291E+04, 2.2306E+04, 2.2321E+04, 2.2335E+04, 2.2350E+04,
     +  2.2364E+04, 2.2379E+04, 2.2394E+04, 2.2408E+04, 2.2423E+04,
     + 2.2438E+04, 2.2452E+04, 2.2467E+04, 2.2482E+04, 2.2496E+04,
     +  2.2511E+04, 2.2526E+04, 2.2540E+04, 2.2555E+04, 2.2570E+04,
     + 2.2584E+04, 2.2599E+04, 2.2614E+04, 2.2629E+04, 2.2643E+04,
     +  2.2658E+04, 2.2673E+04, 2.2688E+04, 2.2702E+04, 2.2717E+04,
     + 2.2732E+04, 2.2746E+04, 2.2761E+04, 2.2776E+04, 2.2791E+04,
     +  2.2806E+04, 2.2820E+04, 2.2835E+04, 2.2850E+04, 2.2865E+04,
     + 2.2879E+04, 2.2894E+04, 2.2909E+04, 2.2924E+04, 2.2939E+04,
     +  2.2953E+04, 2.2968E+04, 2.2983E+04, 2.2998E+04, 2.3013E+04,
     + 2.3028E+04, 2.3042E+04, 2.3057E+04, 2.3072E+04, 2.3087E+04,
     +  2.3102E+04, 2.3117E+04, 2.3132E+04, 2.3146E+04, 2.3161E+04,
     + 2.3176E+04, 2.3191E+04, 2.3206E+04, 2.3221E+04, 2.3236E+04,
     +  2.3251E+04, 2.3266E+04, 2.3281E+04, 2.3295E+04, 2.3310E+04,
     + 2.3325E+04, 2.3340E+04, 2.3355E+04, 2.3370E+04, 2.3385E+04,
     +  2.3400E+04, 2.3415E+04, 2.3430E+04, 2.3445E+04, 2.3460E+04,
     + 2.3475E+04, 2.3490E+04, 2.3505E+04, 2.3520E+04, 2.3535E+04,
     +  2.3550E+04, 2.3565E+04, 2.3580E+04, 2.3595E+04, 2.3610E+04,
     + 2.3625E+04, 2.3640E+04, 2.3655E+04, 2.3670E+04, 2.3685E+04,
     +  2.3700E+04, 2.3715E+04, 2.3730E+04, 2.3745E+04, 2.3760E+04,
     + 2.3775E+04, 2.3790E+04, 2.3805E+04, 2.3820E+04, 2.3835E+04,
     +  2.3851E+04, 2.3866E+04, 2.3881E+04, 2.3896E+04, 2.3911E+04,
     + 2.3926E+04, 2.3941E+04, 2.3956E+04, 2.3971E+04, 2.3986E+04,
     +  2.4002E+04, 2.4017E+04, 2.4032E+04, 2.4047E+04, 2.4062E+04,
     + 2.4077E+04, 2.4092E+04, 2.4108E+04, 2.4123E+04, 2.4138E+04,
     +  2.4153E+04, 2.4168E+04, 2.4183E+04, 2.4199E+04, 2.4214E+04,
     + 2.4229E+04, 2.4244E+04, 2.4259E+04, 2.4275E+04, 2.4290E+04,
     +  2.4305E+04, 2.4320E+04, 2.4336E+04, 2.4351E+04, 2.4366E+04,
     + 2.4381E+04, 2.4396E+04, 2.4412E+04, 2.4427E+04, 2.4442E+04,
     +  2.4457E+04, 2.4473E+04, 2.4488E+04, 2.4503E+04, 2.4519E+04,
     + 2.4534E+04, 2.4549E+04, 2.4564E+04, 2.4580E+04, 2.4595E+04,
     +  2.4610E+04, 2.4626E+04, 2.4641E+04, 2.4656E+04, 2.4672E+04,
     + 2.4687E+04, 2.4702E+04, 2.4718E+04, 2.4733E+04, 2.4748E+04,
     +  2.4764E+04, 2.4779E+04, 2.4794E+04, 2.4810E+04, 2.4825E+04,
     + 2.4840E+04, 2.4856E+04, 2.4871E+04, 2.4887E+04, 2.4902E+04,
     +  2.4917E+04, 2.4933E+04, 2.4948E+04, 2.4963E+04, 2.4979E+04,
     + 2.4994E+04, 2.5010E+04, 2.5025E+04, 2.5041E+04, 2.5056E+04,
     +  2.5071E+04, 2.5087E+04, 2.5102E+04, 2.5118E+04, 2.5133E+04,
     + 2.5149E+04, 2.5164E+04, 2.5180E+04, 2.5195E+04, 2.5211E+04,
     +  2.5226E+04, 2.5242E+04, 2.5257E+04, 2.5272E+04, 2.5288E+04,
     + 2.5303E+04, 2.5319E+04, 2.5334E+04, 2.5350E+04, 2.5366E+04,
     +  2.5381E+04, 2.5397E+04, 2.5412E+04, 2.5428E+04, 2.5443E+04,
     + 2.5459E+04, 2.5474E+04, 2.5490E+04, 2.5505E+04, 2.5521E+04,
     +  2.5536E+04, 2.5552E+04, 2.5568E+04, 2.5583E+04, 2.5599E+04,
     + 2.5614E+04, 2.5630E+04, 2.5646E+04, 2.5661E+04, 2.5677E+04,
     +  2.5692E+04, 2.5708E+04, 2.5724E+04, 2.5739E+04, 2.5755E+04,
     + 2.5770E+04, 2.5786E+04, 2.5802E+04, 2.5817E+04, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        27
      data (QofT( 4,J),J=1,2408)/ 1.2000E+01, 1.2016E+01, 1.2207E+01,
     +  1.2753E+01, 1.3637E+01, 1.4759E+01, 1.6028E+01, 1.7388E+01,
     + 1.8802E+01, 2.0251E+01, 2.1722E+01, 2.3210E+01, 2.4709E+01,
     +  2.6217E+01, 2.7731E+01, 2.9251E+01, 3.0775E+01, 3.2302E+01,
     + 3.3832E+01, 3.5364E+01, 3.8435E+01, 4.1512E+01, 4.4593E+01,
     +  4.7677E+01, 5.0765E+01, 5.3854E+01, 5.6946E+01, 6.0039E+01,
     + 6.3133E+01, 6.6228E+01, 6.9325E+01, 7.2422E+01, 7.5520E+01,
     +  7.8618E+01, 8.1717E+01, 8.4817E+01, 8.7917E+01, 9.1017E+01,
     + 9.4118E+01, 9.7220E+01, 1.0032E+02, 1.0342E+02, 1.0653E+02,
     +  1.0963E+02, 1.1273E+02, 1.1583E+02, 1.1894E+02, 1.2204E+02,
     + 1.2514E+02, 1.2825E+02, 1.3135E+02, 1.3446E+02, 1.3756E+02,
     +  1.4066E+02, 1.4377E+02, 1.4687E+02, 1.4998E+02, 1.5308E+02,
     + 1.5619E+02, 1.5929E+02, 1.6240E+02, 1.6551E+02, 1.6861E+02,
     +  1.7172E+02, 1.7482E+02, 1.7793E+02, 1.8104E+02, 1.8414E+02,
     + 1.8725E+02, 1.9035E+02, 1.9346E+02, 1.9657E+02, 1.9968E+02,
     +  2.0278E+02, 2.0589E+02, 2.0900E+02, 2.1210E+02, 2.1521E+02,
     + 2.1832E+02, 2.2143E+02, 2.2453E+02, 2.2764E+02, 2.3075E+02,
     +  2.3386E+02, 2.3697E+02, 2.4008E+02, 2.4318E+02, 2.4629E+02,
     + 2.4940E+02, 2.5251E+02, 2.5562E+02, 2.5873E+02, 2.6184E+02,
     +  2.6494E+02, 2.6805E+02, 2.7116E+02, 2.7427E+02, 2.7738E+02,
     + 2.8049E+02, 2.8360E+02, 2.8671E+02, 2.8982E+02, 2.9293E+02,
     +  2.9604E+02, 2.9915E+02, 3.0226E+02, 3.0537E+02, 3.0848E+02,
     + 3.1159E+02, 3.1470E+02, 3.1781E+02, 3.2092E+02, 3.2403E+02,
     +  3.2715E+02, 3.3026E+02, 3.3337E+02, 3.3648E+02, 3.3959E+02,
     + 3.4270E+02, 3.4581E+02, 3.4892E+02, 3.5204E+02, 3.5515E+02,
     +  3.5826E+02, 3.6137E+02, 3.6448E+02, 3.6760E+02, 3.7071E+02,
     + 3.7382E+02, 3.7693E+02, 3.8005E+02, 3.8316E+02, 3.8627E+02,
     +  3.8938E+02, 3.9250E+02, 3.9561E+02, 3.9872E+02, 4.0184E+02,
     + 4.0495E+02, 4.0806E+02, 4.1118E+02, 4.1429E+02, 4.1740E+02,
     +  4.2052E+02, 4.2363E+02, 4.2675E+02, 4.2986E+02, 4.3298E+02,
     + 4.3609E+02, 4.3921E+02, 4.4232E+02, 4.4544E+02, 4.4855E+02,
     +  4.5167E+02, 4.5478E+02, 4.5790E+02, 4.6101E+02, 4.6413E+02,
     + 4.6725E+02, 4.7036E+02, 4.7348E+02, 4.7660E+02, 4.7971E+02,
     +  4.8283E+02, 4.8595E+02, 4.8907E+02, 4.9218E+02, 4.9530E+02,
     + 4.9842E+02, 5.0154E+02, 5.0466E+02, 5.0778E+02, 5.1089E+02,
     +  5.1401E+02, 5.1713E+02, 5.2025E+02, 5.2337E+02, 5.2649E+02,
     + 5.2962E+02, 5.3274E+02, 5.3586E+02, 5.3898E+02, 5.4210E+02,
     +  5.4522E+02, 5.4835E+02, 5.5147E+02, 5.5459E+02, 5.5772E+02,
     + 5.6084E+02, 5.6396E+02, 5.6709E+02, 5.7021E+02, 5.7334E+02,
     +  5.7647E+02, 5.7959E+02, 5.8272E+02, 5.8585E+02, 5.8897E+02,
     + 5.9210E+02, 5.9523E+02, 5.9836E+02, 6.0149E+02, 6.0462E+02,
     +  6.0775E+02, 6.1088E+02, 6.1401E+02, 6.1715E+02, 6.2028E+02,
     + 6.2341E+02, 6.2655E+02, 6.2968E+02, 6.3282E+02, 6.3595E+02,
     +  6.3909E+02, 6.4222E+02, 6.4536E+02, 6.4850E+02, 6.5164E+02,
     + 6.5478E+02, 6.5792E+02, 6.6106E+02, 6.6420E+02, 6.6734E+02,
     +  6.7049E+02, 6.7363E+02, 6.7677E+02, 6.7992E+02, 6.8306E+02,
     + 6.8621E+02, 6.8936E+02, 6.9251E+02, 6.9566E+02, 6.9881E+02,
     +  7.0196E+02, 7.0511E+02, 7.0826E+02, 7.1142E+02, 7.1457E+02,
     + 7.1773E+02, 7.2088E+02, 7.2404E+02, 7.2720E+02, 7.3036E+02,
     +  7.3352E+02, 7.3668E+02, 7.3984E+02, 7.4301E+02, 7.4617E+02,
     + 7.4934E+02, 7.5250E+02, 7.5567E+02, 7.5884E+02, 7.6201E+02,
     +  7.6518E+02, 7.6835E+02, 7.7153E+02, 7.7470E+02, 7.7788E+02,
     + 7.8105E+02, 7.8423E+02, 7.8741E+02, 7.9059E+02, 7.9377E+02,
     +  7.9696E+02, 8.0014E+02, 8.0333E+02, 8.0651E+02, 8.0970E+02,
     + 8.1289E+02, 8.1608E+02, 8.1928E+02, 8.2247E+02, 8.2567E+02,
     +  8.2886E+02, 8.3206E+02, 8.3526E+02, 8.3846E+02, 8.4166E+02,
     + 8.4487E+02, 8.4807E+02, 8.5128E+02, 8.5449E+02, 8.5770E+02,
     +  8.6091E+02, 8.6412E+02, 8.6734E+02, 8.7056E+02, 8.7377E+02,
     + 8.7699E+02, 8.8021E+02, 8.8344E+02, 8.8666E+02, 8.8989E+02,
     +  8.9312E+02, 8.9635E+02, 8.9958E+02, 9.0281E+02, 9.0605E+02,
     + 9.0928E+02, 9.1252E+02, 9.1576E+02, 9.1900E+02, 9.2225E+02,
     +  9.2549E+02, 9.2874E+02, 9.3199E+02, 9.3524E+02, 9.3849E+02,
     + 9.4175E+02, 9.4501E+02, 9.4827E+02, 9.5153E+02, 9.5479E+02,
     +  9.5805E+02, 9.6132E+02, 9.6459E+02, 9.6786E+02, 9.7113E+02,
     + 9.7441E+02, 9.7768E+02, 9.8096E+02, 9.8424E+02, 9.8753E+02,
     +  9.9081E+02, 9.9410E+02, 9.9739E+02, 1.0007E+03, 1.0040E+03,
     + 1.0073E+03, 1.0106E+03, 1.0139E+03, 1.0172E+03, 1.0205E+03,
     +  1.0238E+03, 1.0271E+03, 1.0304E+03, 1.0337E+03, 1.0370E+03,
     + 1.0404E+03, 1.0437E+03, 1.0470E+03, 1.0503E+03, 1.0536E+03,
     +  1.0570E+03, 1.0603E+03, 1.0636E+03, 1.0670E+03, 1.0703E+03,
     + 1.0737E+03, 1.0770E+03, 1.0804E+03, 1.0837E+03, 1.0871E+03,
     +  1.0904E+03, 1.0938E+03, 1.0971E+03, 1.1005E+03, 1.1038E+03,
     + 1.1072E+03, 1.1106E+03, 1.1140E+03, 1.1173E+03, 1.1207E+03,
     +  1.1241E+03, 1.1275E+03, 1.1308E+03, 1.1342E+03, 1.1376E+03,
     + 1.1410E+03, 1.1444E+03, 1.1478E+03, 1.1512E+03, 1.1546E+03,
     +  1.1580E+03, 1.1614E+03, 1.1648E+03, 1.1682E+03, 1.1717E+03,
     + 1.1751E+03, 1.1785E+03, 1.1819E+03, 1.1854E+03, 1.1888E+03,
     +  1.1922E+03, 1.1957E+03, 1.1991E+03, 1.2025E+03, 1.2060E+03,
     + 1.2094E+03, 1.2129E+03, 1.2163E+03, 1.2198E+03, 1.2232E+03,
     +  1.2267E+03, 1.2302E+03, 1.2336E+03, 1.2371E+03, 1.2406E+03,
     + 1.2440E+03, 1.2475E+03, 1.2510E+03, 1.2545E+03, 1.2580E+03,
     +  1.2615E+03, 1.2650E+03, 1.2685E+03, 1.2719E+03, 1.2755E+03,
     + 1.2790E+03, 1.2825E+03, 1.2860E+03, 1.2895E+03, 1.2930E+03,
     +  1.2965E+03, 1.3000E+03, 1.3036E+03, 1.3071E+03, 1.3106E+03,
     + 1.3142E+03, 1.3177E+03, 1.3213E+03, 1.3248E+03, 1.3283E+03,
     +  1.3319E+03, 1.3354E+03, 1.3390E+03, 1.3426E+03, 1.3461E+03,
     + 1.3497E+03, 1.3533E+03, 1.3568E+03, 1.3604E+03, 1.3640E+03,
     +  1.3676E+03, 1.3712E+03, 1.3747E+03, 1.3783E+03, 1.3819E+03,
     + 1.3855E+03, 1.3891E+03, 1.3927E+03, 1.3964E+03, 1.4000E+03,
     +  1.4036E+03, 1.4072E+03, 1.4108E+03, 1.4144E+03, 1.4181E+03,
     + 1.4217E+03, 1.4253E+03, 1.4290E+03, 1.4326E+03, 1.4363E+03,
     +  1.4399E+03, 1.4436E+03, 1.4472E+03, 1.4509E+03, 1.4545E+03,
     + 1.4582E+03, 1.4619E+03, 1.4655E+03, 1.4692E+03, 1.4729E+03,
     +  1.4766E+03, 1.4803E+03, 1.4839E+03, 1.4876E+03, 1.4913E+03,
     + 1.4950E+03, 1.4987E+03, 1.5024E+03, 1.5061E+03, 1.5099E+03,
     +  1.5136E+03, 1.5173E+03, 1.5210E+03, 1.5247E+03, 1.5285E+03,
     + 1.5322E+03, 1.5359E+03, 1.5397E+03, 1.5434E+03, 1.5472E+03,
     +  1.5509E+03, 1.5547E+03, 1.5584E+03, 1.5622E+03, 1.5660E+03,
     + 1.5697E+03, 1.5735E+03, 1.5773E+03, 1.5811E+03, 1.5848E+03,
     +  1.5886E+03, 1.5924E+03, 1.5962E+03, 1.6000E+03, 1.6038E+03,
     + 1.6076E+03, 1.6114E+03, 1.6152E+03, 1.6191E+03, 1.6229E+03,
     +  1.6267E+03, 1.6305E+03, 1.6343E+03, 1.6382E+03, 1.6420E+03,
     + 1.6459E+03, 1.6497E+03, 1.6536E+03, 1.6574E+03, 1.6613E+03,
     +  1.6651E+03, 1.6690E+03, 1.6728E+03, 1.6767E+03, 1.6806E+03,
     + 1.6845E+03, 1.6884E+03, 1.6922E+03, 1.6961E+03, 1.7000E+03,
     +  1.7039E+03, 1.7078E+03, 1.7117E+03, 1.7156E+03, 1.7195E+03,
     + 1.7234E+03, 1.7274E+03, 1.7313E+03, 1.7352E+03, 1.7391E+03,
     +  1.7431E+03, 1.7470E+03, 1.7510E+03, 1.7549E+03, 1.7589E+03,
     + 1.7628E+03, 1.7668E+03, 1.7707E+03, 1.7747E+03, 1.7786E+03,
     +  1.7826E+03, 1.7866E+03, 1.7906E+03, 1.7946E+03, 1.7985E+03,
     + 1.8025E+03, 1.8065E+03, 1.8105E+03, 1.8145E+03, 1.8185E+03,
     +  1.8225E+03, 1.8266E+03, 1.8306E+03, 1.8346E+03, 1.8386E+03,
     + 1.8427E+03, 1.8467E+03, 1.8507E+03, 1.8548E+03, 1.8588E+03,
     +  1.8629E+03, 1.8669E+03, 1.8710E+03, 1.8750E+03, 1.8791E+03,
     + 1.8832E+03, 1.8872E+03, 1.8913E+03, 1.8954E+03, 1.8995E+03,
     +  1.9035E+03, 1.9076E+03, 1.9117E+03, 1.9158E+03, 1.9199E+03,
     + 1.9240E+03, 1.9282E+03, 1.9323E+03, 1.9364E+03, 1.9405E+03,
     +  1.9446E+03, 1.9488E+03, 1.9529E+03, 1.9570E+03, 1.9612E+03,
     + 1.9653E+03, 1.9695E+03, 1.9736E+03, 1.9778E+03, 1.9820E+03,
     +  1.9861E+03, 1.9903E+03, 1.9945E+03, 1.9986E+03, 2.0028E+03,
     + 2.0070E+03, 2.0112E+03, 2.0154E+03, 2.0196E+03, 2.0238E+03,
     +  2.0280E+03, 2.0322E+03, 2.0364E+03, 2.0406E+03, 2.0449E+03,
     + 2.0491E+03, 2.0533E+03, 2.0576E+03, 2.0618E+03, 2.0660E+03,
     +  2.0703E+03, 2.0745E+03, 2.0788E+03, 2.0830E+03, 2.0873E+03,
     + 2.0916E+03, 2.0958E+03, 2.1001E+03, 2.1044E+03, 2.1087E+03,
     +  2.1130E+03, 2.1173E+03, 2.1216E+03, 2.1258E+03, 2.1302E+03,
     + 2.1345E+03, 2.1388E+03, 2.1431E+03, 2.1474E+03, 2.1517E+03,
     +  2.1561E+03, 2.1604E+03, 2.1647E+03, 2.1691E+03, 2.1734E+03,
     + 2.1778E+03, 2.1821E+03, 2.1865E+03, 2.1908E+03, 2.1952E+03,
     +  2.1996E+03, 2.2039E+03, 2.2083E+03, 2.2127E+03, 2.2171E+03,
     + 2.2215E+03, 2.2259E+03, 2.2302E+03, 2.2346E+03, 2.2391E+03,
     +  2.2435E+03, 2.2479E+03, 2.2523E+03, 2.2567E+03, 2.2611E+03,
     + 2.2656E+03, 2.2700E+03, 2.2744E+03, 2.2789E+03, 2.2833E+03,
     +  2.2878E+03, 2.2922E+03, 2.2967E+03, 2.3012E+03, 2.3056E+03,
     + 2.3101E+03, 2.3146E+03, 2.3191E+03, 2.3235E+03, 2.3280E+03,
     +  2.3325E+03, 2.3370E+03, 2.3415E+03, 2.3460E+03, 2.3505E+03,
     + 2.3550E+03, 2.3596E+03, 2.3641E+03, 2.3686E+03, 2.3731E+03,
     +  2.3777E+03, 2.3822E+03, 2.3868E+03, 2.3913E+03, 2.3959E+03,
     + 2.4004E+03, 2.4050E+03, 2.4095E+03, 2.4141E+03, 2.4187E+03,
     +  2.4232E+03, 2.4278E+03, 2.4324E+03, 2.4370E+03, 2.4416E+03,
     + 2.4462E+03, 2.4508E+03, 2.4554E+03, 2.4600E+03, 2.4646E+03,
     +  2.4692E+03, 2.4739E+03, 2.4785E+03, 2.4831E+03, 2.4878E+03,
     + 2.4924E+03, 2.4970E+03, 2.5017E+03, 2.5063E+03, 2.5110E+03,
     +  2.5157E+03, 2.5203E+03, 2.5250E+03, 2.5297E+03, 2.5343E+03,
     + 2.5390E+03, 2.5437E+03, 2.5484E+03, 2.5531E+03, 2.5578E+03,
     +  2.5625E+03, 2.5672E+03, 2.5719E+03, 2.5766E+03, 2.5813E+03,
     + 2.5861E+03, 2.5908E+03, 2.5955E+03, 2.6003E+03, 2.6050E+03,
     +  2.6098E+03, 2.6145E+03, 2.6193E+03, 2.6240E+03, 2.6288E+03,
     + 2.6335E+03, 2.6383E+03, 2.6431E+03, 2.6479E+03, 2.6526E+03,
     +  2.6574E+03, 2.6622E+03, 2.6670E+03, 2.6718E+03, 2.6766E+03,
     + 2.6814E+03, 2.6863E+03, 2.6911E+03, 2.6959E+03, 2.7007E+03,
     +  2.7056E+03, 2.7104E+03, 2.7152E+03, 2.7201E+03, 2.7249E+03,
     + 2.7298E+03, 2.7346E+03, 2.7395E+03, 2.7443E+03, 2.7492E+03,
     +  2.7541E+03, 2.7590E+03, 2.7639E+03, 2.7687E+03, 2.7736E+03,
     + 2.7785E+03, 2.7834E+03, 2.7883E+03, 2.7932E+03, 2.7981E+03,
     +  2.8031E+03, 2.8080E+03, 2.8129E+03, 2.8178E+03, 2.8228E+03,
     + 2.8277E+03, 2.8326E+03, 2.8376E+03, 2.8425E+03, 2.8475E+03,
     +  2.8525E+03, 2.8574E+03, 2.8624E+03, 2.8674E+03, 2.8723E+03,
     + 2.8773E+03, 2.8823E+03, 2.8873E+03, 2.8923E+03, 2.8973E+03,
     +  2.9023E+03, 2.9073E+03, 2.9123E+03, 2.9173E+03, 2.9223E+03,
     + 2.9274E+03, 2.9324E+03, 2.9374E+03, 2.9425E+03, 2.9475E+03,
     +  2.9526E+03, 2.9576E+03, 2.9627E+03, 2.9677E+03, 2.9728E+03,
     + 2.9778E+03, 2.9829E+03, 2.9880E+03, 2.9931E+03, 2.9982E+03,
     +  3.0032E+03, 3.0083E+03, 3.0134E+03, 3.0185E+03, 3.0236E+03,
     + 3.0288E+03, 3.0339E+03, 3.0390E+03, 3.0441E+03, 3.0492E+03,
     +  3.0544E+03, 3.0595E+03, 3.0646E+03, 3.0698E+03, 3.0749E+03,
     + 3.0801E+03, 3.0853E+03, 3.0904E+03, 3.0956E+03, 3.1008E+03,
     +  3.1059E+03, 3.1111E+03, 3.1163E+03, 3.1215E+03, 3.1267E+03,
     + 3.1319E+03, 3.1371E+03, 3.1423E+03, 3.1475E+03, 3.1527E+03,
     +  3.1579E+03, 3.1631E+03, 3.1684E+03, 3.1736E+03, 3.1788E+03,
     + 3.1841E+03, 3.1893E+03, 3.1946E+03, 3.1998E+03, 3.2051E+03,
     +  3.2103E+03, 3.2156E+03, 3.2209E+03, 3.2262E+03, 3.2314E+03,
     + 3.2367E+03, 3.2420E+03, 3.2473E+03, 3.2526E+03, 3.2579E+03,
     +  3.2632E+03, 3.2685E+03, 3.2738E+03, 3.2791E+03, 3.2845E+03,
     + 3.2898E+03, 3.2951E+03, 3.3005E+03, 3.3058E+03, 3.3112E+03,
     +  3.3165E+03, 3.3219E+03, 3.3272E+03, 3.3326E+03, 3.3379E+03,
     + 3.3433E+03, 3.3487E+03, 3.3541E+03, 3.3594E+03, 3.3648E+03,
     +  3.3702E+03, 3.3756E+03, 3.3810E+03, 3.3864E+03, 3.3918E+03,
     + 3.3973E+03, 3.4027E+03, 3.4081E+03, 3.4135E+03, 3.4190E+03,
     +  3.4244E+03, 3.4298E+03, 3.4353E+03, 3.4407E+03, 3.4462E+03,
     + 3.4516E+03, 3.4571E+03, 3.4626E+03, 3.4680E+03, 3.4735E+03,
     +  3.4790E+03, 3.4845E+03, 3.4900E+03, 3.4955E+03, 3.5010E+03,
     + 3.5065E+03, 3.5120E+03, 3.5175E+03, 3.5230E+03, 3.5285E+03,
     +  3.5340E+03, 3.5396E+03, 3.5451E+03, 3.5506E+03, 3.5562E+03,
     + 3.5617E+03, 3.5673E+03, 3.5728E+03, 3.5784E+03, 3.5839E+03,
     +  3.5895E+03, 3.5951E+03, 3.6006E+03, 3.6062E+03, 3.6118E+03,
     + 3.6174E+03, 3.6230E+03, 3.6286E+03, 3.6342E+03, 3.6398E+03,
     +  3.6454E+03, 3.6510E+03, 3.6566E+03, 3.6622E+03, 3.6679E+03,
     + 3.6735E+03, 3.6791E+03, 3.6848E+03, 3.6904E+03, 3.6961E+03,
     +  3.7017E+03, 3.7074E+03, 3.7130E+03, 3.7187E+03, 3.7244E+03,
     + 3.7301E+03, 3.7357E+03, 3.7414E+03, 3.7471E+03, 3.7528E+03,
     +  3.7585E+03, 3.7642E+03, 3.7699E+03, 3.7756E+03, 3.7813E+03,
     + 3.7870E+03, 3.7928E+03, 3.7985E+03, 3.8042E+03, 3.8100E+03,
     +  3.8157E+03, 3.8214E+03, 3.8272E+03, 3.8330E+03, 3.8387E+03,
     + 3.8445E+03, 3.8502E+03, 3.8560E+03, 3.8618E+03, 3.8676E+03,
     +  3.8733E+03, 3.8791E+03, 3.8849E+03, 3.8907E+03, 3.8965E+03,
     + 3.9023E+03, 3.9081E+03, 3.9139E+03, 3.9198E+03, 3.9256E+03,
     +  3.9314E+03, 3.9372E+03, 3.9431E+03, 3.9489E+03, 3.9548E+03,
     + 3.9606E+03, 3.9665E+03, 3.9723E+03, 3.9782E+03, 3.9840E+03,
     +  3.9899E+03, 3.9958E+03, 4.0017E+03, 4.0076E+03, 4.0134E+03,
     + 4.0193E+03, 4.0252E+03, 4.0311E+03, 4.0370E+03, 4.0429E+03,
     +  4.0489E+03, 4.0548E+03, 4.0607E+03, 4.0666E+03, 4.0725E+03,
     + 4.0785E+03, 4.0844E+03, 4.0904E+03, 4.0963E+03, 4.1023E+03,
     +  4.1082E+03, 4.1142E+03, 4.1201E+03, 4.1261E+03, 4.1321E+03,
     + 4.1381E+03, 4.1440E+03, 4.1500E+03, 4.1560E+03, 4.1620E+03,
     +  4.1680E+03, 4.1740E+03, 4.1800E+03, 4.1860E+03, 4.1921E+03,
     + 4.1981E+03, 4.2041E+03, 4.2101E+03, 4.2162E+03, 4.2222E+03,
     +  4.2282E+03, 4.2343E+03, 4.2403E+03, 4.2464E+03, 4.2525E+03,
     + 4.2585E+03, 4.2646E+03, 4.2707E+03, 4.2767E+03, 4.2828E+03,
     +  4.2889E+03, 4.2950E+03, 4.3011E+03, 4.3072E+03, 4.3133E+03,
     + 4.3194E+03, 4.3255E+03, 4.3316E+03, 4.3377E+03, 4.3439E+03,
     +  4.3500E+03, 4.3561E+03, 4.3623E+03, 4.3684E+03, 4.3746E+03,
     + 4.3807E+03, 4.3869E+03, 4.3930E+03, 4.3992E+03, 4.4053E+03,
     +  4.4115E+03, 4.4177E+03, 4.4239E+03, 4.4301E+03, 4.4362E+03,
     + 4.4424E+03, 4.4486E+03, 4.4548E+03, 4.4610E+03, 4.4672E+03,
     +  4.4735E+03, 4.4797E+03, 4.4859E+03, 4.4921E+03, 4.4984E+03,
     + 4.5046E+03, 4.5108E+03, 4.5171E+03, 4.5233E+03, 4.5296E+03,
     +  4.5358E+03, 4.5421E+03, 4.5484E+03, 4.5546E+03, 4.5609E+03,
     + 4.5672E+03, 4.5735E+03, 4.5797E+03, 4.5860E+03, 4.5923E+03,
     +  4.5986E+03, 4.6049E+03, 4.6112E+03, 4.6175E+03, 4.6239E+03,
     + 4.6302E+03, 4.6365E+03, 4.6428E+03, 4.6492E+03, 4.6555E+03,
     +  4.6618E+03, 4.6682E+03, 4.6745E+03, 4.6809E+03, 4.6873E+03,
     + 4.6936E+03, 4.7000E+03, 4.7064E+03, 4.7127E+03, 4.7191E+03,
     +  4.7255E+03, 4.7319E+03, 4.7383E+03, 4.7447E+03, 4.7511E+03,
     + 4.7575E+03, 4.7639E+03, 4.7703E+03, 4.7767E+03, 4.7831E+03,
     +  4.7896E+03, 4.7960E+03, 4.8024E+03, 4.8089E+03, 4.8153E+03,
     + 4.8218E+03, 4.8282E+03, 4.8347E+03, 4.8411E+03, 4.8476E+03,
     +  4.8541E+03, 4.8605E+03, 4.8670E+03, 4.8735E+03, 4.8800E+03,
     + 4.8865E+03, 4.8930E+03, 4.8994E+03, 4.9060E+03, 4.9125E+03,
     +  4.9190E+03, 4.9255E+03, 4.9320E+03, 4.9385E+03, 4.9450E+03,
     + 4.9516E+03, 4.9581E+03, 4.9647E+03, 4.9712E+03, 4.9777E+03,
     +  4.9843E+03, 4.9909E+03, 4.9974E+03, 5.0040E+03, 5.0106E+03,
     + 5.0171E+03, 5.0237E+03, 5.0303E+03, 5.0369E+03, 5.0435E+03,
     +  5.0501E+03, 5.0566E+03, 5.0633E+03, 5.0699E+03, 5.0765E+03,
     + 5.0831E+03, 5.0897E+03, 5.0963E+03, 5.1030E+03, 5.1096E+03,
     +  5.1162E+03, 5.1229E+03, 5.1295E+03, 5.1362E+03, 5.1428E+03,
     + 5.1495E+03, 5.1561E+03, 5.1628E+03, 5.1695E+03, 5.1761E+03,
     +  5.1828E+03, 5.1895E+03, 5.1962E+03, 5.2029E+03, 5.2096E+03,
     + 5.2163E+03, 5.2230E+03, 5.2297E+03, 5.2364E+03, 5.2431E+03,
     +  5.2498E+03, 5.2565E+03, 5.2633E+03, 5.2700E+03, 5.2767E+03,
     + 5.2835E+03, 5.2902E+03, 5.2970E+03, 5.3037E+03, 5.3105E+03,
     +  5.3172E+03, 5.3240E+03, 5.3308E+03, 5.3375E+03, 5.3443E+03,
     + 5.3511E+03, 5.3579E+03, 5.3647E+03, 5.3715E+03, 5.3783E+03,
     +  5.3851E+03, 5.3919E+03, 5.3987E+03, 5.4055E+03, 5.4123E+03,
     + 5.4191E+03, 5.4259E+03, 5.4328E+03, 5.4396E+03, 5.4465E+03,
     +  5.4533E+03, 5.4601E+03, 5.4670E+03, 5.4738E+03, 5.4807E+03,
     + 5.4876E+03, 5.4944E+03, 5.5013E+03, 5.5082E+03, 5.5151E+03,
     +  5.5219E+03, 5.5288E+03, 5.5357E+03, 5.5426E+03, 5.5495E+03,
     + 5.5564E+03, 5.5633E+03, 5.5702E+03, 5.5772E+03, 5.5841E+03,
     +  5.5910E+03, 5.5979E+03, 5.6049E+03, 5.6118E+03, 5.6187E+03,
     + 5.6257E+03, 5.6326E+03, 5.6396E+03, 5.6465E+03, 5.6535E+03,
     +  5.6605E+03, 5.6674E+03, 5.6744E+03, 5.6814E+03, 5.6884E+03,
     + 5.6954E+03, 5.7023E+03, 5.7093E+03, 5.7163E+03, 5.7233E+03,
     +  5.7303E+03, 5.7374E+03, 5.7444E+03, 5.7514E+03, 5.7584E+03,
     + 5.7654E+03, 5.7725E+03, 5.7795E+03, 5.7865E+03, 5.7936E+03,
     +  5.8006E+03, 5.8077E+03, 5.8147E+03, 5.8218E+03, 5.8289E+03,
     + 5.8359E+03, 5.8430E+03, 5.8501E+03, 5.8571E+03, 5.8642E+03,
     +  5.8713E+03, 5.8784E+03, 5.8855E+03, 5.8926E+03, 5.8997E+03,
     + 5.9068E+03, 5.9139E+03, 5.9210E+03, 5.9281E+03, 5.9353E+03,
     +  5.9424E+03, 5.9495E+03, 5.9567E+03, 5.9638E+03, 5.9709E+03,
     + 5.9781E+03, 5.9852E+03, 5.9924E+03, 5.9995E+03, 6.0067E+03,
     +  6.0139E+03, 6.0210E+03, 6.0282E+03, 6.0354E+03, 6.0426E+03,
     + 6.0498E+03, 6.0570E+03, 6.0642E+03, 6.0713E+03, 6.0786E+03,
     +  6.0858E+03, 6.0930E+03, 6.1002E+03, 6.1074E+03, 6.1146E+03,
     + 6.1218E+03, 6.1291E+03, 6.1363E+03, 6.1436E+03, 6.1508E+03,
     +  6.1580E+03, 6.1653E+03, 6.1725E+03, 6.1798E+03, 6.1871E+03,
     + 6.1943E+03, 6.2016E+03, 6.2089E+03, 6.2161E+03, 6.2234E+03,
     +  6.2307E+03, 6.2380E+03, 6.2453E+03, 6.2526E+03, 6.2599E+03,
     + 6.2672E+03, 6.2745E+03, 6.2818E+03, 6.2891E+03, 6.2965E+03,
     +  6.3038E+03, 6.3111E+03, 6.3184E+03, 6.3258E+03, 6.3331E+03,
     + 6.3405E+03, 6.3478E+03, 6.3552E+03, 6.3625E+03, 6.3699E+03,
     +  6.3772E+03, 6.3846E+03, 6.3920E+03, 6.3994E+03, 6.4067E+03,
     + 6.4141E+03, 6.4215E+03, 6.4289E+03, 6.4363E+03, 6.4437E+03,
     +  6.4511E+03, 6.4585E+03, 6.4659E+03, 6.4733E+03, 6.4808E+03,
     + 6.4882E+03, 6.4956E+03, 6.5030E+03, 6.5105E+03, 6.5179E+03,
     +  6.5253E+03, 6.5328E+03, 6.5402E+03, 6.5477E+03, 6.5551E+03,
     + 6.5626E+03, 6.5701E+03, 6.5775E+03, 6.5850E+03, 6.5925E+03,
     +  6.6000E+03, 6.6075E+03, 6.6149E+03, 6.6224E+03, 6.6299E+03,
     + 6.6374E+03, 6.6449E+03, 6.6524E+03, 6.6600E+03, 6.6675E+03,
     +  6.6750E+03, 6.6825E+03, 6.6900E+03, 6.6976E+03, 6.7051E+03,
     + 6.7126E+03, 6.7202E+03, 6.7277E+03, 6.7353E+03, 6.7428E+03,
     +  6.7504E+03, 6.7579E+03, 6.7655E+03, 6.7731E+03, 6.7807E+03,
     + 6.7882E+03, 6.7958E+03, 6.8034E+03, 6.8110E+03, 6.8186E+03,
     +  6.8262E+03, 6.8338E+03, 6.8414E+03, 6.8490E+03, 6.8566E+03,
     + 6.8642E+03, 6.8718E+03, 6.8794E+03, 6.8871E+03, 6.8947E+03,
     +  6.9023E+03, 6.9100E+03, 6.9176E+03, 6.9253E+03, 6.9329E+03,
     + 6.9406E+03, 6.9482E+03, 6.9559E+03, 6.9635E+03, 6.9712E+03,
     +  6.9789E+03, 6.9865E+03, 6.9942E+03, 7.0019E+03, 7.0096E+03,
     + 7.0173E+03, 7.0250E+03, 7.0327E+03, 7.0404E+03, 7.0481E+03,
     +  7.0558E+03, 7.0635E+03, 7.0712E+03, 7.0789E+03, 7.0867E+03,
     + 7.0944E+03, 7.1021E+03, 7.1099E+03, 7.1176E+03, 7.1253E+03,
     +  7.1331E+03, 7.1408E+03, 7.1486E+03, 7.1563E+03, 7.1641E+03,
     + 7.1719E+03, 7.1796E+03, 7.1874E+03, 7.1952E+03, 7.2030E+03,
     +  7.2108E+03, 7.2185E+03, 7.2263E+03, 7.2341E+03, 7.2419E+03,
     + 7.2497E+03, 7.2575E+03, 7.2653E+03, 7.2732E+03, 7.2810E+03,
     +  7.2888E+03, 7.2966E+03, 7.3044E+03, 7.3123E+03, 7.3201E+03,
     + 7.3280E+03, 7.3358E+03, 7.3436E+03, 7.3515E+03, 7.3593E+03,
     +  7.3672E+03, 7.3751E+03, 7.3829E+03, 7.3908E+03, 7.3987E+03,
     + 7.4065E+03, 7.4144E+03, 7.4223E+03, 7.4302E+03, 7.4381E+03,
     +  7.4460E+03, 7.4539E+03, 7.4618E+03, 7.4697E+03, 7.4776E+03,
     + 7.4855E+03, 7.4934E+03, 7.5013E+03, 7.5093E+03, 7.5172E+03,
     +  7.5251E+03, 7.5331E+03, 7.5410E+03, 7.5489E+03, 7.5569E+03,
     + 7.5648E+03, 7.5728E+03, 7.5807E+03, 7.5887E+03, 7.5967E+03,
     +  7.6046E+03, 7.6126E+03, 7.6206E+03, 7.6286E+03, 7.6365E+03,
     + 7.6445E+03, 7.6525E+03, 7.6605E+03, 7.6685E+03, 7.6765E+03,
     +  7.6845E+03, 7.6925E+03, 7.7005E+03, 7.7085E+03, 7.7165E+03,
     + 7.7246E+03, 7.7326E+03, 7.7406E+03, 7.7487E+03, 7.7567E+03,
     +  7.7647E+03, 7.7728E+03, 7.7808E+03, 7.7889E+03, 7.7969E+03,
     + 7.8050E+03, 7.8130E+03, 7.8211E+03, 7.8292E+03, 7.8372E+03,
     +  7.8453E+03, 7.8534E+03, 7.8615E+03, 7.8696E+03, 7.8777E+03,
     + 7.8857E+03, 7.8938E+03, 7.9019E+03, 7.9100E+03, 7.9182E+03,
     +  7.9263E+03, 7.9344E+03, 7.9425E+03, 7.9506E+03, 7.9587E+03,
     + 7.9669E+03, 7.9750E+03, 7.9831E+03, 7.9913E+03, 7.9994E+03,
     +  8.0076E+03, 8.0157E+03, 8.0239E+03, 8.0320E+03, 8.0402E+03,
     + 8.0483E+03, 8.0565E+03, 8.0647E+03, 8.0729E+03, 8.0810E+03,
     +  8.0892E+03, 8.0974E+03, 8.1056E+03, 8.1138E+03, 8.1220E+03,
     + 8.1302E+03, 8.1384E+03, 8.1466E+03, 8.1548E+03, 8.1630E+03,
     +  8.1712E+03, 8.1794E+03, 8.1877E+03, 8.1959E+03, 8.2041E+03,
     + 8.2123E+03, 8.2206E+03, 8.2288E+03, 8.2371E+03, 8.2453E+03,
     +  8.2536E+03, 8.2618E+03, 8.2701E+03, 8.2783E+03, 8.2866E+03,
     + 8.2949E+03, 8.3031E+03, 8.3114E+03, 8.3197E+03, 8.3280E+03,
     +  8.3363E+03, 8.3445E+03, 8.3528E+03, 8.3611E+03, 8.3694E+03,
     + 8.3777E+03, 8.3860E+03, 8.3944E+03, 8.4027E+03, 8.4110E+03,
     +  8.4193E+03, 8.4276E+03, 8.4360E+03, 8.4443E+03, 8.4526E+03,
     + 8.4610E+03, 8.4693E+03, 8.4776E+03, 8.4860E+03, 8.4943E+03,
     +  8.5027E+03, 8.5110E+03, 8.5194E+03, 8.5278E+03, 8.5361E+03,
     + 8.5445E+03, 8.5529E+03, 8.5613E+03, 8.5696E+03, 8.5780E+03,
     +  8.5864E+03, 8.5948E+03, 8.6032E+03, 8.6116E+03, 8.6200E+03,
     + 8.6284E+03, 8.6368E+03, 8.6452E+03, 8.6536E+03, 8.6621E+03,
     +  8.6705E+03, 8.6789E+03, 8.6873E+03, 8.6958E+03, 8.7042E+03,
     + 8.7126E+03, 8.7211E+03, 8.7295E+03, 8.7380E+03, 8.7464E+03,
     +  8.7549E+03, 8.7633E+03, 8.7718E+03, 8.7803E+03, 8.7887E+03,
     + 8.7972E+03, 8.8057E+03, 8.8142E+03, 8.8226E+03, 8.8311E+03,
     +  8.8396E+03, 8.8481E+03, 8.8566E+03, 8.8651E+03, 8.8736E+03,
     + 8.8821E+03, 8.8906E+03, 8.8991E+03, 8.9076E+03, 8.9162E+03,
     +  8.9247E+03, 8.9332E+03, 8.9417E+03, 8.9503E+03, 8.9588E+03,
     + 8.9673E+03, 8.9759E+03, 8.9844E+03, 8.9930E+03, 9.0015E+03,
     +  9.0101E+03, 9.0186E+03, 9.0272E+03, 9.0358E+03, 9.0443E+03,
     + 9.0529E+03, 9.0615E+03, 9.0700E+03, 9.0786E+03, 9.0872E+03,
     +  9.0958E+03, 9.1044E+03, 9.1130E+03, 9.1216E+03, 9.1302E+03,
     + 9.1388E+03, 9.1474E+03, 9.1560E+03, 9.1646E+03, 9.1732E+03,
     +  9.1818E+03, 9.1905E+03, 9.1991E+03, 9.2077E+03, 9.2164E+03,
     + 9.2250E+03, 9.2336E+03, 9.2423E+03, 9.2509E+03, 9.2596E+03,
     +  9.2682E+03, 9.2769E+03, 9.2855E+03, 9.2942E+03, 9.3029E+03,
     + 9.3115E+03, 9.3202E+03, 9.3289E+03, 9.3375E+03, 9.3462E+03,
     +  9.3549E+03, 9.3636E+03, 9.3723E+03, 9.3810E+03, 9.3897E+03,
     + 9.3984E+03, 9.4071E+03, 9.4158E+03, 9.4245E+03, 9.4332E+03,
     +  9.4419E+03, 9.4506E+03, 9.4594E+03, 9.4681E+03, 9.4768E+03,
     + 9.4856E+03, 9.4943E+03, 9.5030E+03, 9.5118E+03, 9.5205E+03,
     +  9.5293E+03, 9.5380E+03, 9.5468E+03, 9.5555E+03, 9.5643E+03,
     + 9.5730E+03, 9.5818E+03, 9.5906E+03, 9.5994E+03, 9.6081E+03,
     +  9.6169E+03, 9.6257E+03, 9.6345E+03, 9.6433E+03, 9.6521E+03,
     + 9.6608E+03, 9.6696E+03, 9.6784E+03, 9.6873E+03, 9.6961E+03,
     +  9.7049E+03, 9.7137E+03, 9.7225E+03, 9.7313E+03, 9.7401E+03,
     + 9.7490E+03, 9.7578E+03, 9.7666E+03, 9.7755E+03, 9.7843E+03,
     +  9.7931E+03, 9.8020E+03, 9.8108E+03, 9.8197E+03, 9.8285E+03,
     + 9.8374E+03, 9.8463E+03, 9.8551E+03, 9.8640E+03, 9.8729E+03,
     +  9.8817E+03, 9.8906E+03, 9.8995E+03, 9.9084E+03, 9.9172E+03,
     + 9.9261E+03, 9.9350E+03, 9.9439E+03, 9.9528E+03, 9.9617E+03,
     +  9.9706E+03, 9.9795E+03, 9.9884E+03, 9.9973E+03, 1.0006E+04,
     + 1.0015E+04, 1.0024E+04, 1.0033E+04, 1.0042E+04, 1.0051E+04,
     +  1.0060E+04, 1.0069E+04, 1.0078E+04, 1.0087E+04, 1.0096E+04,
     + 1.0105E+04, 1.0113E+04, 1.0122E+04, 1.0131E+04, 1.0140E+04,
     +  1.0149E+04, 1.0158E+04, 1.0167E+04, 1.0176E+04, 1.0185E+04,
     + 1.0194E+04, 1.0203E+04, 1.0212E+04, 1.0221E+04, 1.0230E+04,
     +  1.0239E+04, 1.0248E+04, 1.0257E+04, 1.0266E+04, 1.0275E+04,
     + 1.0284E+04, 1.0293E+04, 1.0302E+04, 1.0311E+04, 1.0320E+04,
     +  1.0329E+04, 1.0338E+04, 1.0347E+04, 1.0356E+04, 1.0365E+04,
     + 1.0374E+04, 1.0383E+04, 1.0392E+04, 1.0401E+04, 1.0411E+04,
     +  1.0420E+04, 1.0429E+04, 1.0438E+04, 1.0447E+04, 1.0456E+04,
     + 1.0465E+04, 1.0474E+04, 1.0483E+04, 1.0492E+04, 1.0501E+04,
     +  1.0510E+04, 1.0519E+04, 1.0528E+04, 1.0538E+04, 1.0547E+04,
     + 1.0556E+04, 1.0565E+04, 1.0574E+04, 1.0583E+04, 1.0592E+04,
     +  1.0601E+04, 1.0610E+04, 1.0619E+04, 1.0629E+04, 1.0638E+04,
     + 1.0647E+04, 1.0656E+04, 1.0665E+04, 1.0674E+04, 1.0683E+04,
     +  1.0693E+04, 1.0702E+04, 1.0711E+04, 1.0720E+04, 1.0729E+04,
     + 1.0738E+04, 1.0747E+04, 1.0757E+04, 1.0766E+04, 1.0775E+04,
     +  1.0784E+04, 1.0793E+04, 1.0802E+04, 1.0812E+04, 1.0821E+04,
     + 1.0830E+04, 1.0839E+04, 1.0848E+04, 1.0858E+04, 1.0867E+04,
     +  1.0876E+04, 1.0885E+04, 1.0894E+04, 1.0904E+04, 1.0913E+04,
     + 1.0922E+04, 1.0931E+04, 1.0940E+04, 1.0950E+04, 1.0959E+04,
     +  1.0968E+04, 1.0977E+04, 1.0986E+04, 1.0996E+04, 1.1005E+04,
     + 1.1014E+04, 1.1023E+04, 1.1033E+04, 1.1042E+04, 1.1051E+04,
     +  1.1060E+04, 1.1070E+04, 1.1079E+04, 1.1088E+04, 1.1097E+04,
     + 1.1107E+04, 1.1116E+04, 1.1125E+04, 1.1135E+04, 1.1144E+04,
     +  1.1153E+04, 1.1162E+04, 1.1172E+04, 1.1181E+04, 1.1190E+04,
     + 1.1200E+04, 1.1209E+04, 1.1218E+04, 1.1227E+04, 1.1237E+04,
     +  1.1246E+04, 1.1255E+04, 1.1265E+04, 1.1274E+04, 1.1283E+04,
     + 1.1293E+04, 1.1302E+04, 1.1311E+04, 1.1321E+04, 1.1330E+04,
     +  1.1339E+04, 1.1349E+04, 1.1358E+04, 1.1367E+04, 1.1377E+04,
     + 1.1386E+04, 1.1395E+04, 1.1405E+04, 1.1414E+04, 1.1423E+04,
     +  1.1433E+04, 1.1442E+04, 1.1452E+04, 1.1461E+04, 1.1470E+04,
     + 1.1480E+04, 1.1489E+04, 1.1498E+04, 1.1508E+04, 1.1517E+04,
     +  1.1527E+04, 1.1536E+04, 1.1545E+04, 1.1555E+04, 1.1564E+04,
     + 1.1574E+04, 1.1583E+04, 1.1592E+04, 1.1602E+04, 1.1611E+04,
     +  1.1621E+04, 1.1630E+04, 1.1639E+04, 1.1649E+04, 1.1658E+04,
     + 1.1668E+04, 1.1677E+04, 1.1687E+04, 1.1696E+04, 1.1705E+04,
     +  1.1715E+04, 1.1724E+04, 1.1734E+04, 1.1743E+04, 1.1753E+04,
     + 1.1762E+04, 1.1772E+04, 1.1781E+04, 1.1791E+04, 1.1800E+04,
     +  1.1809E+04, 1.1819E+04, 1.1828E+04, 1.1838E+04, 1.1847E+04,
     + 1.1857E+04, 1.1866E+04, 1.1876E+04, 1.1885E+04, 1.1895E+04,
     +  1.1904E+04, 1.1914E+04, 1.1923E+04, 1.1933E+04, 1.1942E+04,
     + 1.1952E+04, 1.1961E+04, 1.1971E+04, 1.1980E+04, 1.1990E+04,
     +  1.1999E+04, 1.2009E+04, 1.2018E+04, 1.2028E+04, 1.2038E+04,
     + 1.2047E+04, 1.2057E+04, 1.2066E+04, 1.2076E+04, 1.2085E+04,
     +  1.2095E+04, 1.2104E+04, 1.2114E+04, 1.2123E+04, 1.2133E+04,
     + 1.2143E+04, 1.2152E+04, 1.2162E+04, 1.2171E+04, 1.2181E+04,
     +  1.2190E+04, 1.2200E+04, 1.2210E+04, 1.2219E+04, 1.2229E+04,
     + 1.2238E+04, 1.2248E+04, 1.2257E+04, 1.2267E+04, 1.2277E+04,
     +  1.2286E+04, 1.2296E+04, 1.2305E+04, 1.2315E+04, 1.2325E+04,
     + 1.2334E+04, 1.2344E+04, 1.2353E+04, 1.2363E+04, 1.2373E+04,
     +  1.2382E+04, 1.2392E+04, 1.2402E+04, 1.2411E+04, 1.2421E+04,
     + 1.2430E+04, 1.2440E+04, 1.2450E+04, 1.2459E+04, 1.2469E+04,
     +  1.2479E+04, 1.2488E+04, 1.2498E+04, 1.2508E+04, 1.2517E+04,
     + 1.2527E+04, 1.2537E+04, 1.2546E+04, 1.2556E+04, 1.2566E+04,
     +  1.2575E+04, 1.2585E+04, 1.2595E+04, 1.2604E+04, 1.2614E+04,
     + 1.2624E+04, 1.2633E+04, 1.2643E+04, 1.2653E+04, 1.2662E+04,
     +  1.2672E+04, 1.2682E+04, 1.2692E+04, 1.2701E+04, 1.2711E+04,
     + 1.2721E+04, 1.2730E+04, 1.2740E+04, 1.2750E+04, 1.2759E+04,
     +  1.2769E+04, 1.2779E+04, 1.2789E+04, 1.2798E+04, 1.2808E+04,
     + 1.2818E+04, 1.2828E+04, 1.2837E+04, 1.2847E+04, 1.2857E+04,
     +  1.2867E+04, 1.2876E+04, 1.2886E+04, 1.2896E+04, 1.2906E+04,
     + 1.2915E+04, 1.2925E+04, 1.2935E+04, 1.2945E+04, 1.2954E+04,
     +  1.2964E+04, 1.2974E+04, 1.2984E+04, 1.2993E+04, 1.3003E+04,
     + 1.3013E+04, 1.3023E+04, 1.3033E+04, 1.3042E+04, 1.3052E+04,
     +  1.3062E+04, 1.3072E+04, 1.3081E+04, 1.3091E+04, 1.3101E+04,
     + 1.3111E+04, 1.3121E+04, 1.3130E+04, 1.3140E+04, 1.3150E+04,
     +  1.3160E+04, 1.3170E+04, 1.3180E+04, 1.3189E+04, 1.3199E+04,
     + 1.3209E+04, 1.3219E+04, 1.3229E+04, 1.3238E+04, 1.3248E+04,
     +  1.3258E+04, 1.3268E+04, 1.3278E+04, 1.3288E+04, 1.3298E+04,
     + 1.3307E+04, 1.3317E+04, 1.3327E+04, 1.3337E+04, 1.3347E+04,
     +  1.3357E+04, 1.3367E+04, 1.3376E+04, 1.3386E+04, 1.3396E+04,
     + 1.3406E+04, 1.3416E+04, 1.3426E+04, 1.3436E+04, 1.3445E+04,
     +  1.3455E+04, 1.3465E+04, 1.3475E+04, 1.3485E+04, 1.3495E+04,
     + 1.3505E+04, 1.3515E+04, 1.3525E+04, 1.3534E+04, 1.3544E+04,
     +  1.3554E+04, 1.3564E+04, 1.3574E+04, 1.3584E+04, 1.3594E+04,
     + 1.3604E+04, 1.3614E+04, 1.3624E+04, 1.3634E+04, 1.3644E+04,
     +  1.3653E+04, 1.3663E+04, 1.3673E+04, 1.3683E+04, 1.3693E+04,
     + 1.3703E+04, 1.3713E+04, 1.3723E+04, 1.3733E+04, 1.3743E+04,
     +  1.3753E+04, 1.3763E+04, 1.3773E+04, 1.3783E+04, 1.3793E+04,
     + 1.3803E+04, 1.3813E+04, 1.3823E+04, 1.3832E+04, 1.3842E+04,
     +  1.3852E+04, 1.3862E+04, 1.3872E+04, 1.3882E+04, 1.3892E+04,
     + 1.3902E+04, 1.3912E+04, 1.3922E+04, 1.3932E+04, 1.3942E+04,
     +  1.3952E+04, 1.3962E+04, 1.3972E+04, 1.3982E+04, 1.3992E+04,
     + 1.4002E+04, 1.4012E+04, 1.4022E+04, 1.4032E+04, 1.4042E+04,
     +  1.4052E+04, 1.4062E+04, 1.4072E+04, 1.4082E+04, 1.4092E+04,
     + 1.4102E+04, 1.4112E+04, 1.4122E+04, 1.4132E+04, 1.4142E+04,
     +  1.4153E+04, 1.4163E+04, 1.4173E+04, 1.4183E+04, 1.4193E+04,
     + 1.4203E+04, 1.4213E+04, 1.4223E+04, 1.4233E+04, 1.4243E+04,
     +  1.4253E+04, 1.4263E+04, 1.4273E+04, 1.4283E+04, 1.4293E+04,
     + 1.4303E+04, 1.4313E+04, 1.4323E+04, 1.4333E+04, 1.4344E+04,
     +  1.4354E+04, 1.4364E+04, 1.4374E+04, 1.4384E+04, 1.4394E+04,
     + 1.4404E+04, 1.4414E+04, 1.4424E+04, 1.4434E+04, 1.4444E+04,
     +  1.4454E+04, 1.4465E+04, 1.4475E+04, 1.4485E+04, 1.4495E+04,
     + 1.4505E+04, 1.4515E+04, 1.4525E+04, 1.4535E+04, 1.4545E+04,
     +  1.4556E+04, 1.4566E+04, 1.4576E+04, 1.4586E+04, 1.4596E+04,
     + 1.4606E+04, 1.4616E+04, 1.4626E+04, 1.4637E+04, 1.4647E+04,
     +  1.4657E+04, 1.4667E+04, 1.4677E+04, 1.4687E+04, 1.4697E+04,
     + 1.4708E+04, 1.4718E+04, 1.4728E+04, 1.4738E+04, 1.4748E+04,
     +  1.4758E+04, 1.4768E+04, 1.4779E+04, 1.4789E+04, 1.4799E+04,
     + 1.4809E+04, 1.4819E+04, 1.4829E+04, 1.4840E+04, 1.4850E+04,
     +  1.4860E+04, 1.4870E+04, 1.4880E+04, 1.4890E+04, 1.4901E+04,
     + 1.4911E+04, 1.4921E+04, 1.4931E+04, 1.4941E+04, 1.4952E+04,
     +  1.4962E+04, 1.4972E+04, 1.4982E+04, 1.4992E+04, 1.5003E+04,
     + 1.5013E+04, 1.5023E+04, 1.5033E+04, 1.5043E+04, 1.5054E+04,
     +  1.5064E+04, 1.5074E+04, 1.5084E+04, 1.5094E+04, 1.5105E+04,
     + 1.5115E+04, 1.5125E+04, 1.5135E+04, 1.5146E+04, 1.5156E+04,
     +  1.5166E+04, 1.5176E+04, 1.5187E+04, 1.5197E+04, 1.5207E+04,
     + 1.5217E+04, 1.5228E+04, 1.5238E+04, 1.5248E+04, 1.5258E+04,
     +  1.5269E+04, 1.5279E+04, 1.5289E+04, 1.5299E+04, 1.5310E+04,
     + 1.5320E+04, 1.5330E+04, 1.5340E+04, 1.5351E+04, 1.5361E+04,
     +  1.5371E+04, 1.5381E+04, 1.5392E+04, 1.5402E+04, 1.5412E+04,
     + 1.5422E+04, 1.5433E+04, 1.5443E+04, 1.5453E+04, 1.5464E+04,
     +  1.5474E+04, 1.5484E+04, 1.5494E+04, 1.5505E+04, 1.5515E+04,
     + 1.5525E+04, 1.5536E+04, 1.5546E+04, 1.5556E+04, 1.5567E+04,
     +  1.5577E+04, 1.5587E+04, 1.5598E+04, 1.5608E+04, 1.5618E+04,
     + 1.5628E+04, 1.5639E+04, 1.5649E+04, 1.5659E+04, 1.5670E+04,
     +  1.5680E+04, 1.5690E+04, 1.5701E+04, 1.5711E+04, 1.5721E+04,
     + 1.5732E+04, 1.5742E+04, 1.5752E+04, 1.5763E+04, 1.5773E+04,
     +  1.5783E+04, 1.5794E+04, 1.5804E+04, 1.5814E+04, 1.5825E+04,
     + 1.5835E+04, 1.5845E+04, 1.5856E+04, 1.5866E+04, 1.5877E+04,
     +  1.5887E+04, 1.5897E+04, 1.5908E+04, 1.5918E+04, 1.5928E+04/
      data (QofT( 4,J),J=2409,4510)/ 1.5939E+04, 1.5949E+04,
     +  1.5959E+04, 1.5970E+04, 1.5980E+04, 1.5991E+04, 1.6001E+04,
     +  1.6011E+04,
     + 1.6022E+04, 1.6032E+04, 1.6043E+04, 1.6053E+04, 1.6063E+04,
     +  1.6074E+04, 1.6084E+04, 1.6094E+04, 1.6105E+04, 1.6115E+04,
     + 1.6126E+04, 1.6136E+04, 1.6146E+04, 1.6157E+04, 1.6167E+04,
     +  1.6178E+04, 1.6188E+04, 1.6198E+04, 1.6209E+04, 1.6219E+04,
     + 1.6230E+04, 1.6240E+04, 1.6251E+04, 1.6261E+04, 1.6271E+04,
     +  1.6282E+04, 1.6292E+04, 1.6303E+04, 1.6313E+04, 1.6324E+04,
     + 1.6334E+04, 1.6344E+04, 1.6355E+04, 1.6365E+04, 1.6376E+04,
     +  1.6386E+04, 1.6397E+04, 1.6407E+04, 1.6417E+04, 1.6428E+04,
     + 1.6438E+04, 1.6449E+04, 1.6459E+04, 1.6470E+04, 1.6480E+04,
     +  1.6491E+04, 1.6501E+04, 1.6512E+04, 1.6522E+04, 1.6533E+04,
     + 1.6543E+04, 1.6553E+04, 1.6564E+04, 1.6574E+04, 1.6585E+04,
     +  1.6595E+04, 1.6606E+04, 1.6616E+04, 1.6627E+04, 1.6637E+04,
     + 1.6648E+04, 1.6658E+04, 1.6669E+04, 1.6679E+04, 1.6690E+04,
     +  1.6700E+04, 1.6711E+04, 1.6721E+04, 1.6732E+04, 1.6742E+04,
     + 1.6753E+04, 1.6763E+04, 1.6774E+04, 1.6784E+04, 1.6795E+04,
     +  1.6805E+04, 1.6816E+04, 1.6826E+04, 1.6837E+04, 1.6847E+04,
     + 1.6858E+04, 1.6868E+04, 1.6879E+04, 1.6889E+04, 1.6900E+04,
     +  1.6910E+04, 1.6921E+04, 1.6931E+04, 1.6942E+04, 1.6952E+04,
     + 1.6963E+04, 1.6973E+04, 1.6984E+04, 1.6994E+04, 1.7005E+04,
     +  1.7016E+04, 1.7026E+04, 1.7037E+04, 1.7047E+04, 1.7058E+04,
     + 1.7068E+04, 1.7079E+04, 1.7089E+04, 1.7100E+04, 1.7110E+04,
     +  1.7121E+04, 1.7132E+04, 1.7142E+04, 1.7153E+04, 1.7163E+04,
     + 1.7174E+04, 1.7184E+04, 1.7195E+04, 1.7205E+04, 1.7216E+04,
     +  1.7227E+04, 1.7237E+04, 1.7248E+04, 1.7258E+04, 1.7269E+04,
     + 1.7279E+04, 1.7290E+04, 1.7301E+04, 1.7311E+04, 1.7322E+04,
     +  1.7332E+04, 1.7343E+04, 1.7353E+04, 1.7364E+04, 1.7375E+04,
     + 1.7385E+04, 1.7396E+04, 1.7406E+04, 1.7417E+04, 1.7428E+04,
     +  1.7438E+04, 1.7449E+04, 1.7459E+04, 1.7470E+04, 1.7481E+04,
     + 1.7491E+04, 1.7502E+04, 1.7512E+04, 1.7523E+04, 1.7534E+04,
     +  1.7544E+04, 1.7555E+04, 1.7565E+04, 1.7576E+04, 1.7587E+04,
     + 1.7597E+04, 1.7608E+04, 1.7619E+04, 1.7629E+04, 1.7640E+04,
     +  1.7650E+04, 1.7661E+04, 1.7672E+04, 1.7682E+04, 1.7693E+04,
     + 1.7704E+04, 1.7714E+04, 1.7725E+04, 1.7735E+04, 1.7746E+04,
     +  1.7757E+04, 1.7767E+04, 1.7778E+04, 1.7789E+04, 1.7799E+04,
     + 1.7810E+04, 1.7821E+04, 1.7831E+04, 1.7842E+04, 1.7853E+04,
     +  1.7863E+04, 1.7874E+04, 1.7885E+04, 1.7895E+04, 1.7906E+04,
     + 1.7916E+04, 1.7927E+04, 1.7938E+04, 1.7948E+04, 1.7959E+04,
     +  1.7970E+04, 1.7980E+04, 1.7991E+04, 1.8002E+04, 1.8012E+04,
     + 1.8023E+04, 1.8034E+04, 1.8045E+04, 1.8055E+04, 1.8066E+04,
     +  1.8077E+04, 1.8087E+04, 1.8098E+04, 1.8109E+04, 1.8119E+04,
     + 1.8130E+04, 1.8141E+04, 1.8151E+04, 1.8162E+04, 1.8173E+04,
     +  1.8183E+04, 1.8194E+04, 1.8205E+04, 1.8216E+04, 1.8226E+04,
     + 1.8237E+04, 1.8248E+04, 1.8258E+04, 1.8269E+04, 1.8280E+04,
     +  1.8290E+04, 1.8301E+04, 1.8312E+04, 1.8323E+04, 1.8333E+04,
     + 1.8344E+04, 1.8355E+04, 1.8365E+04, 1.8376E+04, 1.8387E+04,
     +  1.8398E+04, 1.8408E+04, 1.8419E+04, 1.8430E+04, 1.8440E+04,
     + 1.8451E+04, 1.8462E+04, 1.8473E+04, 1.8483E+04, 1.8494E+04,
     +  1.8505E+04, 1.8516E+04, 1.8526E+04, 1.8537E+04, 1.8548E+04,
     + 1.8558E+04, 1.8569E+04, 1.8580E+04, 1.8591E+04, 1.8601E+04,
     +  1.8612E+04, 1.8623E+04, 1.8634E+04, 1.8644E+04, 1.8655E+04,
     + 1.8666E+04, 1.8677E+04, 1.8687E+04, 1.8698E+04, 1.8709E+04,
     +  1.8720E+04, 1.8730E+04, 1.8741E+04, 1.8752E+04, 1.8763E+04,
     + 1.8773E+04, 1.8784E+04, 1.8795E+04, 1.8806E+04, 1.8817E+04,
     +  1.8827E+04, 1.8838E+04, 1.8849E+04, 1.8860E+04, 1.8870E+04,
     + 1.8881E+04, 1.8892E+04, 1.8903E+04, 1.8914E+04, 1.8924E+04,
     +  1.8935E+04, 1.8946E+04, 1.8957E+04, 1.8967E+04, 1.8978E+04,
     + 1.8989E+04, 1.9000E+04, 1.9011E+04, 1.9021E+04, 1.9032E+04,
     +  1.9043E+04, 1.9054E+04, 1.9065E+04, 1.9075E+04, 1.9086E+04,
     + 1.9097E+04, 1.9108E+04, 1.9119E+04, 1.9129E+04, 1.9140E+04,
     +  1.9151E+04, 1.9162E+04, 1.9173E+04, 1.9183E+04, 1.9194E+04,
     + 1.9205E+04, 1.9216E+04, 1.9227E+04, 1.9237E+04, 1.9248E+04,
     +  1.9259E+04, 1.9270E+04, 1.9281E+04, 1.9292E+04, 1.9302E+04,
     + 1.9313E+04, 1.9324E+04, 1.9335E+04, 1.9346E+04, 1.9356E+04,
     +  1.9367E+04, 1.9378E+04, 1.9389E+04, 1.9400E+04, 1.9411E+04,
     + 1.9421E+04, 1.9432E+04, 1.9443E+04, 1.9454E+04, 1.9465E+04,
     +  1.9476E+04, 1.9486E+04, 1.9497E+04, 1.9508E+04, 1.9519E+04,
     + 1.9530E+04, 1.9541E+04, 1.9552E+04, 1.9562E+04, 1.9573E+04,
     +  1.9584E+04, 1.9595E+04, 1.9606E+04, 1.9617E+04, 1.9627E+04,
     + 1.9638E+04, 1.9649E+04, 1.9660E+04, 1.9671E+04, 1.9682E+04,
     +  1.9693E+04, 1.9703E+04, 1.9714E+04, 1.9725E+04, 1.9736E+04,
     + 1.9747E+04, 1.9758E+04, 1.9769E+04, 1.9780E+04, 1.9790E+04,
     +  1.9801E+04, 1.9812E+04, 1.9823E+04, 1.9834E+04, 1.9845E+04,
     + 1.9856E+04, 1.9867E+04, 1.9877E+04, 1.9888E+04, 1.9899E+04,
     +  1.9910E+04, 1.9921E+04, 1.9932E+04, 1.9943E+04, 1.9954E+04,
     + 1.9964E+04, 1.9975E+04, 1.9986E+04, 1.9997E+04, 2.0008E+04,
     +  2.0019E+04, 2.0030E+04, 2.0041E+04, 2.0052E+04, 2.0062E+04,
     + 2.0073E+04, 2.0084E+04, 2.0095E+04, 2.0106E+04, 2.0117E+04,
     +  2.0128E+04, 2.0139E+04, 2.0150E+04, 2.0161E+04, 2.0171E+04,
     + 2.0182E+04, 2.0193E+04, 2.0204E+04, 2.0215E+04, 2.0226E+04,
     +  2.0237E+04, 2.0248E+04, 2.0259E+04, 2.0270E+04, 2.0281E+04,
     + 2.0292E+04, 2.0302E+04, 2.0313E+04, 2.0324E+04, 2.0335E+04,
     +  2.0346E+04, 2.0357E+04, 2.0368E+04, 2.0379E+04, 2.0390E+04,
     + 2.0401E+04, 2.0412E+04, 2.0423E+04, 2.0434E+04, 2.0444E+04,
     +  2.0455E+04, 2.0466E+04, 2.0477E+04, 2.0488E+04, 2.0499E+04,
     + 2.0510E+04, 2.0521E+04, 2.0532E+04, 2.0543E+04, 2.0554E+04,
     +  2.0565E+04, 2.0576E+04, 2.0587E+04, 2.0598E+04, 2.0608E+04,
     + 2.0619E+04, 2.0630E+04, 2.0641E+04, 2.0652E+04, 2.0663E+04,
     +  2.0674E+04, 2.0685E+04, 2.0696E+04, 2.0707E+04, 2.0718E+04,
     + 2.0729E+04, 2.0740E+04, 2.0751E+04, 2.0762E+04, 2.0773E+04,
     +  2.0784E+04, 2.0795E+04, 2.0806E+04, 2.0817E+04, 2.0828E+04,
     + 2.0839E+04, 2.0849E+04, 2.0860E+04, 2.0871E+04, 2.0882E+04,
     +  2.0893E+04, 2.0904E+04, 2.0915E+04, 2.0926E+04, 2.0937E+04,
     + 2.0948E+04, 2.0959E+04, 2.0970E+04, 2.0981E+04, 2.0992E+04,
     +  2.1003E+04, 2.1014E+04, 2.1025E+04, 2.1036E+04, 2.1047E+04,
     + 2.1058E+04, 2.1069E+04, 2.1080E+04, 2.1091E+04, 2.1102E+04,
     +  2.1113E+04, 2.1124E+04, 2.1135E+04, 2.1146E+04, 2.1157E+04,
     + 2.1168E+04, 2.1179E+04, 2.1190E+04, 2.1201E+04, 2.1212E+04,
     +  2.1223E+04, 2.1234E+04, 2.1245E+04, 2.1256E+04, 2.1267E+04,
     + 2.1278E+04, 2.1289E+04, 2.1300E+04, 2.1311E+04, 2.1322E+04,
     +  2.1333E+04, 2.1344E+04, 2.1355E+04, 2.1366E+04, 2.1377E+04,
     + 2.1388E+04, 2.1399E+04, 2.1410E+04, 2.1421E+04, 2.1432E+04,
     +  2.1443E+04, 2.1454E+04, 2.1465E+04, 2.1476E+04, 2.1487E+04,
     + 2.1498E+04, 2.1509E+04, 2.1520E+04, 2.1531E+04, 2.1542E+04,
     +  2.1553E+04, 2.1564E+04, 2.1575E+04, 2.1586E+04, 2.1597E+04,
     + 2.1608E+04, 2.1619E+04, 2.1630E+04, 2.1641E+04, 2.1652E+04,
     +  2.1663E+04, 2.1674E+04, 2.1685E+04, 2.1696E+04, 2.1707E+04,
     + 2.1718E+04, 2.1729E+04, 2.1740E+04, 2.1751E+04, 2.1762E+04,
     +  2.1774E+04, 2.1785E+04, 2.1796E+04, 2.1807E+04, 2.1818E+04,
     + 2.1829E+04, 2.1840E+04, 2.1851E+04, 2.1862E+04, 2.1873E+04,
     +  2.1884E+04, 2.1895E+04, 2.1906E+04, 2.1917E+04, 2.1928E+04,
     + 2.1939E+04, 2.1950E+04, 2.1961E+04, 2.1972E+04, 2.1983E+04,
     +  2.1994E+04, 2.2005E+04, 2.2016E+04, 2.2027E+04, 2.2039E+04,
     + 2.2050E+04, 2.2061E+04, 2.2072E+04, 2.2083E+04, 2.2094E+04,
     +  2.2105E+04, 2.2116E+04, 2.2127E+04, 2.2138E+04, 2.2149E+04,
     + 2.2160E+04, 2.2171E+04, 2.2182E+04, 2.2193E+04, 2.2204E+04,
     +  2.2215E+04, 2.2227E+04, 2.2238E+04, 2.2249E+04, 2.2260E+04,
     + 2.2271E+04, 2.2282E+04, 2.2293E+04, 2.2304E+04, 2.2315E+04,
     +  2.2326E+04, 2.2337E+04, 2.2348E+04, 2.2359E+04, 2.2370E+04,
     + 2.2381E+04, 2.2393E+04, 2.2404E+04, 2.2415E+04, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_HBr   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 4),QofT( 4,NT),Q(NT)
      data (Tmx(I),I=1,4)/6000.,6000.,6000.,6000./
      data xgj/ 8.,8.,12.,12./
c...      HBr
c...        --        19
      data (QofT( 1,J),J=1,2408)/ 8.0000E+00, 8.0001E+00, 8.0080E+00,
     +  8.0591E+00, 8.1964E+00, 8.4377E+00, 8.7766E+00, 9.1955E+00,
     + 9.6756E+00, 1.0201E+01, 1.0758E+01, 1.1339E+01, 1.1937E+01,
     +  1.2548E+01, 1.3168E+01, 1.3795E+01, 1.4428E+01, 1.5065E+01,
     + 1.5706E+01, 1.6349E+01, 1.7644E+01, 1.8945E+01, 2.0252E+01,
     +  2.1563E+01, 2.2877E+01, 2.4194E+01, 2.5512E+01, 2.6832E+01,
     + 2.8154E+01, 2.9477E+01, 3.0801E+01, 3.2125E+01, 3.3451E+01,
     +  3.4777E+01, 3.6103E+01, 3.7431E+01, 3.8758E+01, 4.0086E+01,
     + 4.1415E+01, 4.2743E+01, 4.4072E+01, 4.5402E+01, 4.6731E+01,
     +  4.8061E+01, 4.9391E+01, 5.0721E+01, 5.2052E+01, 5.3382E+01,
     + 5.4713E+01, 5.6044E+01, 5.7375E+01, 5.8706E+01, 6.0038E+01,
     +  6.1369E+01, 6.2701E+01, 6.4032E+01, 6.5364E+01, 6.6696E+01,
     + 6.8028E+01, 6.9360E+01, 7.0692E+01, 7.2024E+01, 7.3357E+01,
     +  7.4689E+01, 7.6022E+01, 7.7354E+01, 7.8687E+01, 8.0020E+01,
     + 8.1352E+01, 8.2685E+01, 8.4018E+01, 8.5351E+01, 8.6684E+01,
     +  8.8018E+01, 8.9351E+01, 9.0684E+01, 9.2017E+01, 9.3351E+01,
     + 9.4684E+01, 9.6018E+01, 9.7351E+01, 9.8685E+01, 1.0002E+02,
     +  1.0135E+02, 1.0269E+02, 1.0402E+02, 1.0535E+02, 1.0669E+02,
     + 1.0802E+02, 1.0936E+02, 1.1069E+02, 1.1202E+02, 1.1336E+02,
     +  1.1469E+02, 1.1603E+02, 1.1736E+02, 1.1870E+02, 1.2003E+02,
     + 1.2136E+02, 1.2270E+02, 1.2403E+02, 1.2537E+02, 1.2670E+02,
     +  1.2804E+02, 1.2937E+02, 1.3071E+02, 1.3204E+02, 1.3338E+02,
     + 1.3471E+02, 1.3605E+02, 1.3738E+02, 1.3872E+02, 1.4005E+02,
     +  1.4139E+02, 1.4272E+02, 1.4406E+02, 1.4539E+02, 1.4673E+02,
     + 1.4806E+02, 1.4940E+02, 1.5073E+02, 1.5207E+02, 1.5340E+02,
     +  1.5474E+02, 1.5608E+02, 1.5741E+02, 1.5875E+02, 1.6008E+02,
     + 1.6142E+02, 1.6275E+02, 1.6409E+02, 1.6543E+02, 1.6676E+02,
     +  1.6810E+02, 1.6943E+02, 1.7077E+02, 1.7211E+02, 1.7344E+02,
     + 1.7478E+02, 1.7611E+02, 1.7745E+02, 1.7879E+02, 1.8012E+02,
     +  1.8146E+02, 1.8280E+02, 1.8413E+02, 1.8547E+02, 1.8681E+02,
     + 1.8814E+02, 1.8948E+02, 1.9082E+02, 1.9215E+02, 1.9349E+02,
     +  1.9483E+02, 1.9616E+02, 1.9750E+02, 1.9884E+02, 2.0017E+02,
     + 2.0151E+02, 2.0285E+02, 2.0419E+02, 2.0552E+02, 2.0686E+02,
     +  2.0820E+02, 2.0953E+02, 2.1087E+02, 2.1221E+02, 2.1355E+02,
     + 2.1488E+02, 2.1622E+02, 2.1756E+02, 2.1890E+02, 2.2023E+02,
     +  2.2157E+02, 2.2291E+02, 2.2425E+02, 2.2559E+02, 2.2692E+02,
     + 2.2826E+02, 2.2960E+02, 2.3094E+02, 2.3228E+02, 2.3361E+02,
     +  2.3495E+02, 2.3629E+02, 2.3763E+02, 2.3897E+02, 2.4031E+02,
     + 2.4165E+02, 2.4298E+02, 2.4432E+02, 2.4566E+02, 2.4700E+02,
     +  2.4834E+02, 2.4968E+02, 2.5102E+02, 2.5236E+02, 2.5370E+02,
     + 2.5504E+02, 2.5637E+02, 2.5771E+02, 2.5905E+02, 2.6039E+02,
     +  2.6173E+02, 2.6307E+02, 2.6441E+02, 2.6575E+02, 2.6709E+02,
     + 2.6843E+02, 2.6977E+02, 2.7111E+02, 2.7245E+02, 2.7380E+02,
     +  2.7514E+02, 2.7648E+02, 2.7782E+02, 2.7916E+02, 2.8050E+02,
     + 2.8184E+02, 2.8318E+02, 2.8452E+02, 2.8587E+02, 2.8721E+02,
     +  2.8855E+02, 2.8989E+02, 2.9123E+02, 2.9258E+02, 2.9392E+02,
     + 2.9526E+02, 2.9660E+02, 2.9795E+02, 2.9929E+02, 3.0063E+02,
     +  3.0198E+02, 3.0332E+02, 3.0466E+02, 3.0601E+02, 3.0735E+02,
     + 3.0869E+02, 3.1004E+02, 3.1138E+02, 3.1273E+02, 3.1407E+02,
     +  3.1542E+02, 3.1676E+02, 3.1811E+02, 3.1945E+02, 3.2080E+02,
     + 3.2214E+02, 3.2349E+02, 3.2484E+02, 3.2618E+02, 3.2753E+02,
     +  3.2887E+02, 3.3022E+02, 3.3157E+02, 3.3292E+02, 3.3426E+02,
     + 3.3561E+02, 3.3696E+02, 3.3831E+02, 3.3966E+02, 3.4101E+02,
     +  3.4235E+02, 3.4370E+02, 3.4505E+02, 3.4640E+02, 3.4775E+02,
     + 3.4910E+02, 3.5045E+02, 3.5180E+02, 3.5316E+02, 3.5451E+02,
     +  3.5586E+02, 3.5721E+02, 3.5856E+02, 3.5991E+02, 3.6127E+02,
     + 3.6262E+02, 3.6397E+02, 3.6533E+02, 3.6668E+02, 3.6804E+02,
     +  3.6939E+02, 3.7074E+02, 3.7210E+02, 3.7346E+02, 3.7481E+02,
     + 3.7617E+02, 3.7752E+02, 3.7888E+02, 3.8024E+02, 3.8160E+02,
     +  3.8295E+02, 3.8431E+02, 3.8567E+02, 3.8703E+02, 3.8839E+02,
     + 3.8975E+02, 3.9111E+02, 3.9247E+02, 3.9383E+02, 3.9519E+02,
     +  3.9655E+02, 3.9791E+02, 3.9928E+02, 4.0064E+02, 4.0200E+02,
     + 4.0337E+02, 4.0473E+02, 4.0609E+02, 4.0746E+02, 4.0882E+02,
     +  4.1019E+02, 4.1156E+02, 4.1292E+02, 4.1429E+02, 4.1566E+02,
     + 4.1702E+02, 4.1839E+02, 4.1976E+02, 4.2113E+02, 4.2250E+02,
     +  4.2387E+02, 4.2524E+02, 4.2661E+02, 4.2798E+02, 4.2936E+02,
     + 4.3073E+02, 4.3210E+02, 4.3348E+02, 4.3485E+02, 4.3622E+02,
     +  4.3760E+02, 4.3898E+02, 4.4035E+02, 4.4173E+02, 4.4311E+02,
     + 4.4448E+02, 4.4586E+02, 4.4724E+02, 4.4862E+02, 4.5000E+02,
     +  4.5138E+02, 4.5276E+02, 4.5414E+02, 4.5553E+02, 4.5691E+02,
     + 4.5829E+02, 4.5968E+02, 4.6106E+02, 4.6244E+02, 4.6383E+02,
     +  4.6522E+02, 4.6660E+02, 4.6799E+02, 4.6938E+02, 4.7077E+02,
     + 4.7216E+02, 4.7355E+02, 4.7494E+02, 4.7633E+02, 4.7772E+02,
     +  4.7911E+02, 4.8051E+02, 4.8190E+02, 4.8329E+02, 4.8469E+02,
     + 4.8608E+02, 4.8748E+02, 4.8888E+02, 4.9027E+02, 4.9167E+02,
     +  4.9307E+02, 4.9447E+02, 4.9587E+02, 4.9727E+02, 4.9867E+02,
     + 5.0008E+02, 5.0148E+02, 5.0288E+02, 5.0429E+02, 5.0569E+02,
     +  5.0710E+02, 5.0851E+02, 5.0991E+02, 5.1132E+02, 5.1273E+02,
     + 5.1414E+02, 5.1555E+02, 5.1696E+02, 5.1837E+02, 5.1978E+02,
     +  5.2120E+02, 5.2261E+02, 5.2403E+02, 5.2544E+02, 5.2686E+02,
     + 5.2827E+02, 5.2969E+02, 5.3111E+02, 5.3253E+02, 5.3395E+02,
     +  5.3537E+02, 5.3679E+02, 5.3821E+02, 5.3964E+02, 5.4106E+02,
     + 5.4249E+02, 5.4391E+02, 5.4534E+02, 5.4677E+02, 5.4819E+02,
     +  5.4962E+02, 5.5105E+02, 5.5248E+02, 5.5392E+02, 5.5535E+02,
     + 5.5678E+02, 5.5821E+02, 5.5965E+02, 5.6108E+02, 5.6252E+02,
     +  5.6396E+02, 5.6540E+02, 5.6684E+02, 5.6828E+02, 5.6972E+02,
     + 5.7116E+02, 5.7260E+02, 5.7404E+02, 5.7549E+02, 5.7693E+02,
     +  5.7838E+02, 5.7983E+02, 5.8127E+02, 5.8272E+02, 5.8417E+02,
     + 5.8562E+02, 5.8707E+02, 5.8853E+02, 5.8998E+02, 5.9143E+02,
     +  5.9289E+02, 5.9434E+02, 5.9580E+02, 5.9726E+02, 5.9872E+02,
     + 6.0018E+02, 6.0164E+02, 6.0310E+02, 6.0456E+02, 6.0603E+02,
     +  6.0749E+02, 6.0896E+02, 6.1042E+02, 6.1189E+02, 6.1336E+02,
     + 6.1483E+02, 6.1630E+02, 6.1777E+02, 6.1924E+02, 6.2071E+02,
     +  6.2219E+02, 6.2366E+02, 6.2514E+02, 6.2662E+02, 6.2809E+02,
     + 6.2957E+02, 6.3105E+02, 6.3253E+02, 6.3402E+02, 6.3550E+02,
     +  6.3698E+02, 6.3847E+02, 6.3995E+02, 6.4144E+02, 6.4293E+02,
     + 6.4442E+02, 6.4591E+02, 6.4740E+02, 6.4889E+02, 6.5039E+02,
     +  6.5188E+02, 6.5337E+02, 6.5487E+02, 6.5637E+02, 6.5787E+02,
     + 6.5937E+02, 6.6087E+02, 6.6237E+02, 6.6387E+02, 6.6537E+02,
     +  6.6688E+02, 6.6839E+02, 6.6989E+02, 6.7140E+02, 6.7291E+02,
     + 6.7442E+02, 6.7593E+02, 6.7744E+02, 6.7896E+02, 6.8047E+02,
     +  6.8199E+02, 6.8350E+02, 6.8502E+02, 6.8654E+02, 6.8806E+02,
     + 6.8958E+02, 6.9110E+02, 6.9263E+02, 6.9415E+02, 6.9568E+02,
     +  6.9720E+02, 6.9873E+02, 7.0026E+02, 7.0179E+02, 7.0332E+02,
     + 7.0485E+02, 7.0638E+02, 7.0792E+02, 7.0945E+02, 7.1099E+02,
     +  7.1253E+02, 7.1407E+02, 7.1561E+02, 7.1715E+02, 7.1869E+02,
     + 7.2023E+02, 7.2178E+02, 7.2332E+02, 7.2487E+02, 7.2642E+02,
     +  7.2797E+02, 7.2952E+02, 7.3107E+02, 7.3262E+02, 7.3418E+02,
     + 7.3573E+02, 7.3729E+02, 7.3885E+02, 7.4040E+02, 7.4196E+02,
     +  7.4352E+02, 7.4509E+02, 7.4665E+02, 7.4821E+02, 7.4978E+02,
     + 7.5135E+02, 7.5291E+02, 7.5448E+02, 7.5605E+02, 7.5763E+02,
     +  7.5920E+02, 7.6077E+02, 7.6235E+02, 7.6392E+02, 7.6550E+02,
     + 7.6708E+02, 7.6866E+02, 7.7024E+02, 7.7182E+02, 7.7341E+02,
     +  7.7499E+02, 7.7658E+02, 7.7816E+02, 7.7975E+02, 7.8134E+02,
     + 7.8293E+02, 7.8453E+02, 7.8612E+02, 7.8771E+02, 7.8931E+02,
     +  7.9091E+02, 7.9250E+02, 7.9410E+02, 7.9571E+02, 7.9731E+02,
     + 7.9891E+02, 8.0051E+02, 8.0212E+02, 8.0373E+02, 8.0534E+02,
     +  8.0695E+02, 8.0856E+02, 8.1017E+02, 8.1178E+02, 8.1340E+02,
     + 8.1501E+02, 8.1663E+02, 8.1825E+02, 8.1987E+02, 8.2149E+02,
     +  8.2311E+02, 8.2473E+02, 8.2636E+02, 8.2798E+02, 8.2961E+02,
     + 8.3124E+02, 8.3287E+02, 8.3450E+02, 8.3613E+02, 8.3777E+02,
     +  8.3940E+02, 8.4104E+02, 8.4267E+02, 8.4431E+02, 8.4595E+02,
     + 8.4759E+02, 8.4924E+02, 8.5088E+02, 8.5253E+02, 8.5417E+02,
     +  8.5582E+02, 8.5747E+02, 8.5912E+02, 8.6077E+02, 8.6242E+02,
     + 8.6408E+02, 8.6573E+02, 8.6739E+02, 8.6905E+02, 8.7071E+02,
     +  8.7237E+02, 8.7403E+02, 8.7570E+02, 8.7736E+02, 8.7903E+02,
     + 8.8069E+02, 8.8236E+02, 8.8403E+02, 8.8570E+02, 8.8738E+02,
     +  8.8905E+02, 8.9073E+02, 8.9240E+02, 8.9408E+02, 8.9576E+02,
     + 8.9744E+02, 8.9912E+02, 9.0081E+02, 9.0249E+02, 9.0418E+02,
     +  9.0587E+02, 9.0755E+02, 9.0924E+02, 9.1094E+02, 9.1263E+02,
     + 9.1432E+02, 9.1602E+02, 9.1771E+02, 9.1941E+02, 9.2111E+02,
     +  9.2281E+02, 9.2452E+02, 9.2622E+02, 9.2792E+02, 9.2963E+02,
     + 9.3134E+02, 9.3305E+02, 9.3476E+02, 9.3647E+02, 9.3818E+02,
     +  9.3990E+02, 9.4161E+02, 9.4333E+02, 9.4505E+02, 9.4677E+02,
     + 9.4849E+02, 9.5021E+02, 9.5194E+02, 9.5366E+02, 9.5539E+02,
     +  9.5712E+02, 9.5885E+02, 9.6058E+02, 9.6231E+02, 9.6405E+02,
     + 9.6578E+02, 9.6752E+02, 9.6926E+02, 9.7100E+02, 9.7274E+02,
     +  9.7448E+02, 9.7622E+02, 9.7797E+02, 9.7971E+02, 9.8146E+02,
     + 9.8321E+02, 9.8496E+02, 9.8671E+02, 9.8847E+02, 9.9022E+02,
     +  9.9198E+02, 9.9374E+02, 9.9549E+02, 9.9725E+02, 9.9902E+02,
     + 1.0008E+03, 1.0025E+03, 1.0043E+03, 1.0061E+03, 1.0078E+03,
     +  1.0096E+03, 1.0114E+03, 1.0132E+03, 1.0149E+03, 1.0167E+03,
     + 1.0185E+03, 1.0203E+03, 1.0220E+03, 1.0238E+03, 1.0256E+03,
     +  1.0274E+03, 1.0292E+03, 1.0310E+03, 1.0328E+03, 1.0345E+03,
     + 1.0363E+03, 1.0381E+03, 1.0399E+03, 1.0417E+03, 1.0435E+03,
     +  1.0453E+03, 1.0471E+03, 1.0489E+03, 1.0507E+03, 1.0525E+03,
     + 1.0543E+03, 1.0561E+03, 1.0580E+03, 1.0598E+03, 1.0616E+03,
     +  1.0634E+03, 1.0652E+03, 1.0670E+03, 1.0688E+03, 1.0707E+03,
     + 1.0725E+03, 1.0743E+03, 1.0761E+03, 1.0779E+03, 1.0798E+03,
     +  1.0816E+03, 1.0834E+03, 1.0853E+03, 1.0871E+03, 1.0889E+03,
     + 1.0908E+03, 1.0926E+03, 1.0944E+03, 1.0963E+03, 1.0981E+03,
     +  1.1000E+03, 1.1018E+03, 1.1036E+03, 1.1055E+03, 1.1073E+03,
     + 1.1092E+03, 1.1110E+03, 1.1129E+03, 1.1147E+03, 1.1166E+03,
     +  1.1185E+03, 1.1203E+03, 1.1222E+03, 1.1240E+03, 1.1259E+03,
     + 1.1278E+03, 1.1296E+03, 1.1315E+03, 1.1334E+03, 1.1352E+03,
     +  1.1371E+03, 1.1390E+03, 1.1409E+03, 1.1427E+03, 1.1446E+03,
     + 1.1465E+03, 1.1484E+03, 1.1503E+03, 1.1521E+03, 1.1540E+03,
     +  1.1559E+03, 1.1578E+03, 1.1597E+03, 1.1616E+03, 1.1635E+03,
     + 1.1654E+03, 1.1673E+03, 1.1692E+03, 1.1711E+03, 1.1730E+03,
     +  1.1749E+03, 1.1768E+03, 1.1787E+03, 1.1806E+03, 1.1825E+03,
     + 1.1844E+03, 1.1863E+03, 1.1882E+03, 1.1901E+03, 1.1920E+03,
     +  1.1939E+03, 1.1959E+03, 1.1978E+03, 1.1997E+03, 1.2016E+03,
     + 1.2036E+03, 1.2055E+03, 1.2074E+03, 1.2093E+03, 1.2113E+03,
     +  1.2132E+03, 1.2151E+03, 1.2171E+03, 1.2190E+03, 1.2209E+03,
     + 1.2229E+03, 1.2248E+03, 1.2268E+03, 1.2287E+03, 1.2306E+03,
     +  1.2326E+03, 1.2345E+03, 1.2365E+03, 1.2384E+03, 1.2404E+03,
     + 1.2423E+03, 1.2443E+03, 1.2463E+03, 1.2482E+03, 1.2502E+03,
     +  1.2521E+03, 1.2541E+03, 1.2561E+03, 1.2580E+03, 1.2600E+03,
     + 1.2620E+03, 1.2639E+03, 1.2659E+03, 1.2679E+03, 1.2699E+03,
     +  1.2718E+03, 1.2738E+03, 1.2758E+03, 1.2778E+03, 1.2798E+03,
     + 1.2818E+03, 1.2837E+03, 1.2857E+03, 1.2877E+03, 1.2897E+03,
     +  1.2917E+03, 1.2937E+03, 1.2957E+03, 1.2977E+03, 1.2997E+03,
     + 1.3017E+03, 1.3037E+03, 1.3057E+03, 1.3077E+03, 1.3097E+03,
     +  1.3117E+03, 1.3137E+03, 1.3157E+03, 1.3177E+03, 1.3198E+03,
     + 1.3218E+03, 1.3238E+03, 1.3258E+03, 1.3278E+03, 1.3298E+03,
     +  1.3319E+03, 1.3339E+03, 1.3359E+03, 1.3380E+03, 1.3400E+03,
     + 1.3420E+03, 1.3440E+03, 1.3461E+03, 1.3481E+03, 1.3502E+03,
     +  1.3522E+03, 1.3542E+03, 1.3563E+03, 1.3583E+03, 1.3604E+03,
     + 1.3624E+03, 1.3645E+03, 1.3665E+03, 1.3686E+03, 1.3706E+03,
     +  1.3727E+03, 1.3747E+03, 1.3768E+03, 1.3788E+03, 1.3809E+03,
     + 1.3830E+03, 1.3850E+03, 1.3871E+03, 1.3892E+03, 1.3912E+03,
     +  1.3933E+03, 1.3954E+03, 1.3974E+03, 1.3995E+03, 1.4016E+03,
     + 1.4037E+03, 1.4057E+03, 1.4078E+03, 1.4099E+03, 1.4120E+03,
     +  1.4141E+03, 1.4162E+03, 1.4183E+03, 1.4203E+03, 1.4224E+03,
     + 1.4245E+03, 1.4266E+03, 1.4287E+03, 1.4308E+03, 1.4329E+03,
     +  1.4350E+03, 1.4371E+03, 1.4392E+03, 1.4413E+03, 1.4434E+03,
     + 1.4456E+03, 1.4477E+03, 1.4498E+03, 1.4519E+03, 1.4540E+03,
     +  1.4561E+03, 1.4582E+03, 1.4604E+03, 1.4625E+03, 1.4646E+03,
     + 1.4667E+03, 1.4689E+03, 1.4710E+03, 1.4731E+03, 1.4753E+03,
     +  1.4774E+03, 1.4795E+03, 1.4817E+03, 1.4838E+03, 1.4859E+03,
     + 1.4881E+03, 1.4902E+03, 1.4924E+03, 1.4945E+03, 1.4967E+03,
     +  1.4988E+03, 1.5010E+03, 1.5031E+03, 1.5053E+03, 1.5074E+03,
     + 1.5096E+03, 1.5117E+03, 1.5139E+03, 1.5161E+03, 1.5182E+03,
     +  1.5204E+03, 1.5226E+03, 1.5247E+03, 1.5269E+03, 1.5291E+03,
     + 1.5312E+03, 1.5334E+03, 1.5356E+03, 1.5378E+03, 1.5400E+03,
     +  1.5421E+03, 1.5443E+03, 1.5465E+03, 1.5487E+03, 1.5509E+03,
     + 1.5531E+03, 1.5553E+03, 1.5574E+03, 1.5596E+03, 1.5618E+03,
     +  1.5640E+03, 1.5662E+03, 1.5684E+03, 1.5706E+03, 1.5728E+03,
     + 1.5750E+03, 1.5773E+03, 1.5795E+03, 1.5817E+03, 1.5839E+03,
     +  1.5861E+03, 1.5883E+03, 1.5905E+03, 1.5927E+03, 1.5950E+03,
     + 1.5972E+03, 1.5994E+03, 1.6016E+03, 1.6039E+03, 1.6061E+03,
     +  1.6083E+03, 1.6106E+03, 1.6128E+03, 1.6150E+03, 1.6173E+03,
     + 1.6195E+03, 1.6217E+03, 1.6240E+03, 1.6262E+03, 1.6285E+03,
     +  1.6307E+03, 1.6330E+03, 1.6352E+03, 1.6375E+03, 1.6397E+03,
     + 1.6420E+03, 1.6442E+03, 1.6465E+03, 1.6487E+03, 1.6510E+03,
     +  1.6533E+03, 1.6555E+03, 1.6578E+03, 1.6601E+03, 1.6623E+03,
     + 1.6646E+03, 1.6669E+03, 1.6691E+03, 1.6714E+03, 1.6737E+03,
     +  1.6760E+03, 1.6782E+03, 1.6805E+03, 1.6828E+03, 1.6851E+03,
     + 1.6874E+03, 1.6897E+03, 1.6920E+03, 1.6943E+03, 1.6965E+03,
     +  1.6988E+03, 1.7011E+03, 1.7034E+03, 1.7057E+03, 1.7080E+03,
     + 1.7103E+03, 1.7126E+03, 1.7150E+03, 1.7173E+03, 1.7196E+03,
     +  1.7219E+03, 1.7242E+03, 1.7265E+03, 1.7288E+03, 1.7311E+03,
     + 1.7335E+03, 1.7358E+03, 1.7381E+03, 1.7404E+03, 1.7428E+03,
     +  1.7451E+03, 1.7474E+03, 1.7497E+03, 1.7521E+03, 1.7544E+03,
     + 1.7568E+03, 1.7591E+03, 1.7614E+03, 1.7638E+03, 1.7661E+03,
     +  1.7685E+03, 1.7708E+03, 1.7732E+03, 1.7755E+03, 1.7779E+03,
     + 1.7802E+03, 1.7826E+03, 1.7849E+03, 1.7873E+03, 1.7896E+03,
     +  1.7920E+03, 1.7944E+03, 1.7967E+03, 1.7991E+03, 1.8015E+03,
     + 1.8038E+03, 1.8062E+03, 1.8086E+03, 1.8109E+03, 1.8133E+03,
     +  1.8157E+03, 1.8181E+03, 1.8205E+03, 1.8228E+03, 1.8252E+03,
     + 1.8276E+03, 1.8300E+03, 1.8324E+03, 1.8348E+03, 1.8372E+03,
     +  1.8396E+03, 1.8420E+03, 1.8444E+03, 1.8468E+03, 1.8492E+03,
     + 1.8516E+03, 1.8540E+03, 1.8564E+03, 1.8588E+03, 1.8612E+03,
     +  1.8636E+03, 1.8660E+03, 1.8684E+03, 1.8708E+03, 1.8733E+03,
     + 1.8757E+03, 1.8781E+03, 1.8805E+03, 1.8829E+03, 1.8854E+03,
     +  1.8878E+03, 1.8902E+03, 1.8927E+03, 1.8951E+03, 1.8975E+03,
     + 1.9000E+03, 1.9024E+03, 1.9048E+03, 1.9073E+03, 1.9097E+03,
     +  1.9122E+03, 1.9146E+03, 1.9171E+03, 1.9195E+03, 1.9220E+03,
     + 1.9244E+03, 1.9269E+03, 1.9293E+03, 1.9318E+03, 1.9342E+03,
     +  1.9367E+03, 1.9392E+03, 1.9416E+03, 1.9441E+03, 1.9466E+03,
     + 1.9490E+03, 1.9515E+03, 1.9540E+03, 1.9564E+03, 1.9589E+03,
     +  1.9614E+03, 1.9639E+03, 1.9664E+03, 1.9688E+03, 1.9713E+03,
     + 1.9738E+03, 1.9763E+03, 1.9788E+03, 1.9813E+03, 1.9838E+03,
     +  1.9863E+03, 1.9888E+03, 1.9913E+03, 1.9938E+03, 1.9963E+03,
     + 1.9988E+03, 2.0013E+03, 2.0038E+03, 2.0063E+03, 2.0088E+03,
     +  2.0113E+03, 2.0138E+03, 2.0163E+03, 2.0189E+03, 2.0214E+03,
     + 2.0239E+03, 2.0264E+03, 2.0289E+03, 2.0315E+03, 2.0340E+03,
     +  2.0365E+03, 2.0391E+03, 2.0416E+03, 2.0441E+03, 2.0467E+03,
     + 2.0492E+03, 2.0517E+03, 2.0543E+03, 2.0568E+03, 2.0594E+03,
     +  2.0619E+03, 2.0645E+03, 2.0670E+03, 2.0696E+03, 2.0721E+03,
     + 2.0747E+03, 2.0772E+03, 2.0798E+03, 2.0823E+03, 2.0849E+03,
     +  2.0875E+03, 2.0900E+03, 2.0926E+03, 2.0952E+03, 2.0977E+03,
     + 2.1003E+03, 2.1029E+03, 2.1055E+03, 2.1080E+03, 2.1106E+03,
     +  2.1132E+03, 2.1158E+03, 2.1184E+03, 2.1209E+03, 2.1235E+03,
     + 2.1261E+03, 2.1287E+03, 2.1313E+03, 2.1339E+03, 2.1365E+03,
     +  2.1391E+03, 2.1417E+03, 2.1443E+03, 2.1469E+03, 2.1495E+03,
     + 2.1521E+03, 2.1547E+03, 2.1573E+03, 2.1599E+03, 2.1625E+03,
     +  2.1651E+03, 2.1678E+03, 2.1704E+03, 2.1730E+03, 2.1756E+03,
     + 2.1782E+03, 2.1809E+03, 2.1835E+03, 2.1861E+03, 2.1888E+03,
     +  2.1914E+03, 2.1940E+03, 2.1967E+03, 2.1993E+03, 2.2019E+03,
     + 2.2046E+03, 2.2072E+03, 2.2099E+03, 2.2125E+03, 2.2151E+03,
     +  2.2178E+03, 2.2204E+03, 2.2231E+03, 2.2257E+03, 2.2284E+03,
     + 2.2311E+03, 2.2337E+03, 2.2364E+03, 2.2390E+03, 2.2417E+03,
     +  2.2444E+03, 2.2470E+03, 2.2497E+03, 2.2524E+03, 2.2551E+03,
     + 2.2577E+03, 2.2604E+03, 2.2631E+03, 2.2658E+03, 2.2684E+03,
     +  2.2711E+03, 2.2738E+03, 2.2765E+03, 2.2792E+03, 2.2819E+03,
     + 2.2846E+03, 2.2873E+03, 2.2900E+03, 2.2927E+03, 2.2954E+03,
     +  2.2981E+03, 2.3008E+03, 2.3035E+03, 2.3062E+03, 2.3089E+03,
     + 2.3116E+03, 2.3143E+03, 2.3170E+03, 2.3197E+03, 2.3224E+03,
     +  2.3252E+03, 2.3279E+03, 2.3306E+03, 2.3333E+03, 2.3361E+03,
     + 2.3388E+03, 2.3415E+03, 2.3442E+03, 2.3470E+03, 2.3497E+03,
     +  2.3524E+03, 2.3552E+03, 2.3579E+03, 2.3607E+03, 2.3634E+03,
     + 2.3661E+03, 2.3689E+03, 2.3716E+03, 2.3744E+03, 2.3771E+03,
     +  2.3799E+03, 2.3826E+03, 2.3854E+03, 2.3882E+03, 2.3909E+03,
     + 2.3937E+03, 2.3965E+03, 2.3992E+03, 2.4020E+03, 2.4048E+03,
     +  2.4075E+03, 2.4103E+03, 2.4131E+03, 2.4158E+03, 2.4186E+03,
     + 2.4214E+03, 2.4242E+03, 2.4270E+03, 2.4298E+03, 2.4325E+03,
     +  2.4353E+03, 2.4381E+03, 2.4409E+03, 2.4437E+03, 2.4465E+03,
     + 2.4493E+03, 2.4521E+03, 2.4549E+03, 2.4577E+03, 2.4605E+03,
     +  2.4633E+03, 2.4661E+03, 2.4689E+03, 2.4717E+03, 2.4745E+03,
     + 2.4774E+03, 2.4802E+03, 2.4830E+03, 2.4858E+03, 2.4886E+03,
     +  2.4915E+03, 2.4943E+03, 2.4971E+03, 2.4999E+03, 2.5028E+03,
     + 2.5056E+03, 2.5084E+03, 2.5113E+03, 2.5141E+03, 2.5169E+03,
     +  2.5198E+03, 2.5226E+03, 2.5255E+03, 2.5283E+03, 2.5312E+03,
     + 2.5340E+03, 2.5369E+03, 2.5397E+03, 2.5426E+03, 2.5454E+03,
     +  2.5483E+03, 2.5512E+03, 2.5540E+03, 2.5569E+03, 2.5598E+03,
     + 2.5626E+03, 2.5655E+03, 2.5684E+03, 2.5712E+03, 2.5741E+03,
     +  2.5770E+03, 2.5799E+03, 2.5827E+03, 2.5856E+03, 2.5885E+03,
     + 2.5914E+03, 2.5943E+03, 2.5972E+03, 2.6001E+03, 2.6030E+03,
     +  2.6058E+03, 2.6087E+03, 2.6116E+03, 2.6145E+03, 2.6174E+03,
     + 2.6203E+03, 2.6232E+03, 2.6262E+03, 2.6291E+03, 2.6320E+03,
     +  2.6349E+03, 2.6378E+03, 2.6407E+03, 2.6436E+03, 2.6466E+03,
     + 2.6495E+03, 2.6524E+03, 2.6553E+03, 2.6582E+03, 2.6612E+03,
     +  2.6641E+03, 2.6670E+03, 2.6700E+03, 2.6729E+03, 2.6758E+03,
     + 2.6788E+03, 2.6817E+03, 2.6847E+03, 2.6876E+03, 2.6906E+03,
     +  2.6935E+03, 2.6965E+03, 2.6994E+03, 2.7024E+03, 2.7053E+03,
     + 2.7083E+03, 2.7112E+03, 2.7142E+03, 2.7172E+03, 2.7201E+03,
     +  2.7231E+03, 2.7261E+03, 2.7290E+03, 2.7320E+03, 2.7350E+03,
     + 2.7379E+03, 2.7409E+03, 2.7439E+03, 2.7469E+03, 2.7499E+03,
     +  2.7528E+03, 2.7558E+03, 2.7588E+03, 2.7618E+03, 2.7648E+03,
     + 2.7678E+03, 2.7708E+03, 2.7738E+03, 2.7768E+03, 2.7798E+03,
     +  2.7828E+03, 2.7858E+03, 2.7888E+03, 2.7918E+03, 2.7948E+03,
     + 2.7978E+03, 2.8008E+03, 2.8038E+03, 2.8069E+03, 2.8099E+03,
     +  2.8129E+03, 2.8159E+03, 2.8189E+03, 2.8220E+03, 2.8250E+03,
     + 2.8280E+03, 2.8311E+03, 2.8341E+03, 2.8371E+03, 2.8402E+03,
     +  2.8432E+03, 2.8462E+03, 2.8493E+03, 2.8523E+03, 2.8554E+03,
     + 2.8584E+03, 2.8615E+03, 2.8645E+03, 2.8676E+03, 2.8706E+03,
     +  2.8737E+03, 2.8767E+03, 2.8798E+03, 2.8829E+03, 2.8859E+03,
     + 2.8890E+03, 2.8920E+03, 2.8951E+03, 2.8982E+03, 2.9013E+03,
     +  2.9043E+03, 2.9074E+03, 2.9105E+03, 2.9136E+03, 2.9166E+03,
     + 2.9197E+03, 2.9228E+03, 2.9259E+03, 2.9290E+03, 2.9321E+03,
     +  2.9352E+03, 2.9383E+03, 2.9414E+03, 2.9445E+03, 2.9476E+03,
     + 2.9507E+03, 2.9538E+03, 2.9569E+03, 2.9600E+03, 2.9631E+03,
     +  2.9662E+03, 2.9693E+03, 2.9724E+03, 2.9755E+03, 2.9787E+03,
     + 2.9818E+03, 2.9849E+03, 2.9880E+03, 2.9911E+03, 2.9943E+03,
     +  2.9974E+03, 3.0005E+03, 3.0037E+03, 3.0068E+03, 3.0099E+03,
     + 3.0131E+03, 3.0162E+03, 3.0194E+03, 3.0225E+03, 3.0256E+03,
     +  3.0288E+03, 3.0319E+03, 3.0351E+03, 3.0382E+03, 3.0414E+03,
     + 3.0446E+03, 3.0477E+03, 3.0509E+03, 3.0540E+03, 3.0572E+03,
     +  3.0604E+03, 3.0635E+03, 3.0667E+03, 3.0699E+03, 3.0730E+03,
     + 3.0762E+03, 3.0794E+03, 3.0826E+03, 3.0858E+03, 3.0889E+03,
     +  3.0921E+03, 3.0953E+03, 3.0985E+03, 3.1017E+03, 3.1049E+03,
     + 3.1081E+03, 3.1113E+03, 3.1145E+03, 3.1177E+03, 3.1209E+03,
     +  3.1241E+03, 3.1273E+03, 3.1305E+03, 3.1337E+03, 3.1369E+03,
     + 3.1401E+03, 3.1433E+03, 3.1465E+03, 3.1497E+03, 3.1530E+03,
     +  3.1562E+03, 3.1594E+03, 3.1626E+03, 3.1659E+03, 3.1691E+03,
     + 3.1723E+03, 3.1755E+03, 3.1788E+03, 3.1820E+03, 3.1853E+03,
     +  3.1885E+03, 3.1917E+03, 3.1950E+03, 3.1982E+03, 3.2015E+03,
     + 3.2047E+03, 3.2080E+03, 3.2112E+03, 3.2145E+03, 3.2177E+03,
     +  3.2210E+03, 3.2242E+03, 3.2275E+03, 3.2308E+03, 3.2340E+03,
     + 3.2373E+03, 3.2406E+03, 3.2438E+03, 3.2471E+03, 3.2504E+03,
     +  3.2537E+03, 3.2569E+03, 3.2602E+03, 3.2635E+03, 3.2668E+03,
     + 3.2701E+03, 3.2734E+03, 3.2766E+03, 3.2799E+03, 3.2832E+03,
     +  3.2865E+03, 3.2898E+03, 3.2931E+03, 3.2964E+03, 3.2997E+03,
     + 3.3030E+03, 3.3063E+03, 3.3096E+03, 3.3130E+03, 3.3163E+03,
     +  3.3196E+03, 3.3229E+03, 3.3262E+03, 3.3295E+03, 3.3328E+03,
     + 3.3362E+03, 3.3395E+03, 3.3428E+03, 3.3461E+03, 3.3495E+03,
     +  3.3528E+03, 3.3561E+03, 3.3595E+03, 3.3628E+03, 3.3662E+03,
     + 3.3695E+03, 3.3728E+03, 3.3762E+03, 3.3795E+03, 3.3829E+03,
     +  3.3862E+03, 3.3896E+03, 3.3929E+03, 3.3963E+03, 3.3997E+03,
     + 3.4030E+03, 3.4064E+03, 3.4097E+03, 3.4131E+03, 3.4165E+03,
     +  3.4198E+03, 3.4232E+03, 3.4266E+03, 3.4300E+03, 3.4333E+03,
     + 3.4367E+03, 3.4401E+03, 3.4435E+03, 3.4469E+03, 3.4503E+03,
     +  3.4536E+03, 3.4570E+03, 3.4604E+03, 3.4638E+03, 3.4672E+03,
     + 3.4706E+03, 3.4740E+03, 3.4774E+03, 3.4808E+03, 3.4842E+03,
     +  3.4876E+03, 3.4910E+03, 3.4944E+03, 3.4979E+03, 3.5013E+03,
     + 3.5047E+03, 3.5081E+03, 3.5115E+03, 3.5150E+03, 3.5184E+03,
     +  3.5218E+03, 3.5252E+03, 3.5287E+03, 3.5321E+03, 3.5355E+03,
     + 3.5390E+03, 3.5424E+03, 3.5458E+03, 3.5493E+03, 3.5527E+03,
     +  3.5562E+03, 3.5596E+03, 3.5631E+03, 3.5665E+03, 3.5700E+03,
     + 3.5734E+03, 3.5769E+03, 3.5803E+03, 3.5838E+03, 3.5872E+03,
     +  3.5907E+03, 3.5942E+03, 3.5976E+03, 3.6011E+03, 3.6046E+03,
     + 3.6081E+03, 3.6115E+03, 3.6150E+03, 3.6185E+03, 3.6220E+03,
     +  3.6255E+03, 3.6289E+03, 3.6324E+03, 3.6359E+03, 3.6394E+03,
     + 3.6429E+03, 3.6464E+03, 3.6499E+03, 3.6534E+03, 3.6569E+03,
     +  3.6604E+03, 3.6639E+03, 3.6674E+03, 3.6709E+03, 3.6744E+03,
     + 3.6779E+03, 3.6814E+03, 3.6849E+03, 3.6885E+03, 3.6920E+03,
     +  3.6955E+03, 3.6990E+03, 3.7025E+03, 3.7061E+03, 3.7096E+03,
     + 3.7131E+03, 3.7167E+03, 3.7202E+03, 3.7237E+03, 3.7273E+03,
     +  3.7308E+03, 3.7343E+03, 3.7379E+03, 3.7414E+03, 3.7450E+03,
     + 3.7485E+03, 3.7521E+03, 3.7556E+03, 3.7592E+03, 3.7627E+03,
     +  3.7663E+03, 3.7699E+03, 3.7734E+03, 3.7770E+03, 3.7806E+03,
     + 3.7841E+03, 3.7877E+03, 3.7913E+03, 3.7948E+03, 3.7984E+03,
     +  3.8020E+03, 3.8056E+03, 3.8092E+03, 3.8127E+03, 3.8163E+03,
     + 3.8199E+03, 3.8235E+03, 3.8271E+03, 3.8307E+03, 3.8343E+03,
     +  3.8379E+03, 3.8415E+03, 3.8451E+03, 3.8487E+03, 3.8523E+03,
     + 3.8559E+03, 3.8595E+03, 3.8631E+03, 3.8667E+03, 3.8703E+03,
     +  3.8739E+03, 3.8776E+03, 3.8812E+03, 3.8848E+03, 3.8884E+03,
     + 3.8921E+03, 3.8957E+03, 3.8993E+03, 3.9029E+03, 3.9066E+03,
     +  3.9102E+03, 3.9138E+03, 3.9175E+03, 3.9211E+03, 3.9248E+03,
     + 3.9284E+03, 3.9321E+03, 3.9357E+03, 3.9394E+03, 3.9430E+03,
     +  3.9467E+03, 3.9503E+03, 3.9540E+03, 3.9576E+03, 3.9613E+03,
     + 3.9650E+03, 3.9686E+03, 3.9723E+03, 3.9760E+03, 3.9796E+03,
     +  3.9833E+03, 3.9870E+03, 3.9907E+03, 3.9943E+03, 3.9980E+03,
     + 4.0017E+03, 4.0054E+03, 4.0091E+03, 4.0128E+03, 4.0165E+03,
     +  4.0202E+03, 4.0239E+03, 4.0275E+03, 4.0312E+03, 4.0349E+03,
     + 4.0387E+03, 4.0424E+03, 4.0461E+03, 4.0498E+03, 4.0535E+03,
     +  4.0572E+03, 4.0609E+03, 4.0646E+03, 4.0683E+03, 4.0721E+03,
     + 4.0758E+03, 4.0795E+03, 4.0832E+03, 4.0870E+03, 4.0907E+03,
     +  4.0944E+03, 4.0982E+03, 4.1019E+03, 4.1056E+03, 4.1094E+03,
     + 4.1131E+03, 4.1168E+03, 4.1206E+03, 4.1243E+03, 4.1281E+03,
     +  4.1318E+03, 4.1356E+03, 4.1393E+03, 4.1431E+03, 4.1469E+03,
     + 4.1506E+03, 4.1544E+03, 4.1582E+03, 4.1619E+03, 4.1657E+03,
     +  4.1695E+03, 4.1732E+03, 4.1770E+03, 4.1808E+03, 4.1846E+03,
     + 4.1883E+03, 4.1921E+03, 4.1959E+03, 4.1997E+03, 4.2035E+03,
     +  4.2073E+03, 4.2111E+03, 4.2149E+03, 4.2186E+03, 4.2224E+03,
     + 4.2262E+03, 4.2300E+03, 4.2339E+03, 4.2377E+03, 4.2415E+03,
     +  4.2453E+03, 4.2491E+03, 4.2529E+03, 4.2567E+03, 4.2605E+03,
     + 4.2643E+03, 4.2682E+03, 4.2720E+03, 4.2758E+03, 4.2796E+03,
     +  4.2835E+03, 4.2873E+03, 4.2911E+03, 4.2950E+03, 4.2988E+03,
     + 4.3026E+03, 4.3065E+03, 4.3103E+03, 4.3142E+03, 4.3180E+03,
     +  4.3219E+03, 4.3257E+03, 4.3296E+03, 4.3334E+03, 4.3373E+03,
     + 4.3411E+03, 4.3450E+03, 4.3489E+03, 4.3527E+03, 4.3566E+03,
     +  4.3605E+03, 4.3643E+03, 4.3682E+03, 4.3721E+03, 4.3759E+03,
     + 4.3798E+03, 4.3837E+03, 4.3876E+03, 4.3915E+03, 4.3954E+03,
     +  4.3992E+03, 4.4031E+03, 4.4070E+03, 4.4109E+03, 4.4148E+03,
     + 4.4187E+03, 4.4226E+03, 4.4265E+03, 4.4304E+03, 4.4343E+03,
     +  4.4382E+03, 4.4421E+03, 4.4460E+03, 4.4500E+03, 4.4539E+03,
     + 4.4578E+03, 4.4617E+03, 4.4656E+03, 4.4696E+03, 4.4735E+03,
     +  4.4774E+03, 4.4813E+03, 4.4853E+03, 4.4892E+03, 4.4931E+03,
     + 4.4971E+03, 4.5010E+03, 4.5049E+03, 4.5089E+03, 4.5128E+03,
     +  4.5168E+03, 4.5207E+03, 4.5247E+03, 4.5286E+03, 4.5326E+03,
     + 4.5365E+03, 4.5405E+03, 4.5445E+03, 4.5484E+03, 4.5524E+03,
     +  4.5564E+03, 4.5603E+03, 4.5643E+03, 4.5683E+03, 4.5722E+03,
     + 4.5762E+03, 4.5802E+03, 4.5842E+03, 4.5882E+03, 4.5921E+03,
     +  4.5961E+03, 4.6001E+03, 4.6041E+03, 4.6081E+03, 4.6121E+03,
     + 4.6161E+03, 4.6201E+03, 4.6241E+03, 4.6281E+03, 4.6321E+03,
     +  4.6361E+03, 4.6401E+03, 4.6441E+03, 4.6481E+03, 4.6521E+03,
     + 4.6562E+03, 4.6602E+03, 4.6642E+03, 4.6682E+03, 4.6722E+03,
     +  4.6763E+03, 4.6803E+03, 4.6843E+03, 4.6883E+03, 4.6924E+03,
     + 4.6964E+03, 4.7005E+03, 4.7045E+03, 4.7085E+03, 4.7126E+03,
     +  4.7166E+03, 4.7207E+03, 4.7247E+03, 4.7288E+03, 4.7328E+03,
     + 4.7369E+03, 4.7409E+03, 4.7450E+03, 4.7491E+03, 4.7531E+03,
     +  4.7572E+03, 4.7613E+03, 4.7653E+03, 4.7694E+03, 4.7735E+03,
     + 4.7775E+03, 4.7816E+03, 4.7857E+03, 4.7898E+03, 4.7939E+03,
     +  4.7979E+03, 4.8020E+03, 4.8061E+03, 4.8102E+03, 4.8143E+03,
     + 4.8184E+03, 4.8225E+03, 4.8266E+03, 4.8307E+03, 4.8348E+03,
     +  4.8389E+03, 4.8430E+03, 4.8471E+03, 4.8512E+03, 4.8553E+03,
     + 4.8595E+03, 4.8636E+03, 4.8677E+03, 4.8718E+03, 4.8759E+03,
     +  4.8801E+03, 4.8842E+03, 4.8883E+03, 4.8925E+03, 4.8966E+03,
     + 4.9007E+03, 4.9049E+03, 4.9090E+03, 4.9131E+03, 4.9173E+03,
     +  4.9214E+03, 4.9256E+03, 4.9297E+03, 4.9339E+03, 4.9380E+03,
     + 4.9422E+03, 4.9463E+03, 4.9505E+03, 4.9547E+03, 4.9588E+03,
     +  4.9630E+03, 4.9672E+03, 4.9713E+03, 4.9755E+03, 4.9797E+03,
     + 4.9838E+03, 4.9880E+03, 4.9922E+03, 4.9964E+03, 5.0006E+03,
     +  5.0048E+03, 5.0089E+03, 5.0131E+03, 5.0173E+03, 5.0215E+03,
     + 5.0257E+03, 5.0299E+03, 5.0341E+03, 5.0383E+03, 5.0425E+03,
     +  5.0467E+03, 5.0509E+03, 5.0551E+03, 5.0593E+03, 5.0636E+03,
     + 5.0678E+03, 5.0720E+03, 5.0762E+03, 5.0804E+03, 5.0846E+03,
     +  5.0889E+03, 5.0931E+03, 5.0973E+03, 5.1016E+03, 5.1058E+03,
     + 5.1100E+03, 5.1143E+03, 5.1185E+03, 5.1227E+03, 5.1270E+03,
     +  5.1312E+03, 5.1355E+03, 5.1397E+03, 5.1440E+03, 5.1482E+03,
     + 5.1525E+03, 5.1568E+03, 5.1610E+03, 5.1653E+03, 5.1695E+03,
     +  5.1738E+03, 5.1781E+03, 5.1823E+03, 5.1866E+03, 5.1909E+03,
     + 5.1952E+03, 5.1994E+03, 5.2037E+03, 5.2080E+03, 5.2123E+03,
     +  5.2166E+03, 5.2209E+03, 5.2251E+03, 5.2294E+03, 5.2337E+03,
     + 5.2380E+03, 5.2423E+03, 5.2466E+03, 5.2509E+03, 5.2552E+03,
     +  5.2595E+03, 5.2638E+03, 5.2682E+03, 5.2725E+03, 5.2768E+03,
     + 5.2811E+03, 5.2854E+03, 5.2897E+03, 5.2941E+03, 5.2984E+03,
     +  5.3027E+03, 5.3070E+03, 5.3114E+03, 5.3157E+03, 5.3200E+03,
     + 5.3244E+03, 5.3287E+03, 5.3330E+03, 5.3374E+03, 5.3417E+03,
     +  5.3461E+03, 5.3504E+03, 5.3548E+03, 5.3591E+03, 5.3635E+03,
     + 5.3678E+03, 5.3722E+03, 5.3766E+03, 5.3809E+03, 5.3853E+03,
     +  5.3896E+03, 5.3940E+03, 5.3984E+03, 5.4028E+03, 5.4071E+03,
     + 5.4115E+03, 5.4159E+03, 5.4203E+03, 5.4247E+03, 5.4290E+03,
     +  5.4334E+03, 5.4378E+03, 5.4422E+03, 5.4466E+03, 5.4510E+03,
     + 5.4554E+03, 5.4598E+03, 5.4642E+03, 5.4686E+03, 5.4730E+03,
     +  5.4774E+03, 5.4818E+03, 5.4862E+03, 5.4906E+03, 5.4951E+03,
     + 5.4995E+03, 5.5039E+03, 5.5083E+03, 5.5127E+03, 5.5172E+03,
     +  5.5216E+03, 5.5260E+03, 5.5305E+03, 5.5349E+03, 5.5393E+03,
     + 5.5438E+03, 5.5482E+03, 5.5526E+03, 5.5571E+03, 5.5615E+03,
     +  5.5660E+03, 5.5704E+03, 5.5749E+03, 5.5793E+03, 5.5838E+03,
     + 5.5882E+03, 5.5927E+03, 5.5972E+03, 5.6016E+03, 5.6061E+03,
     +  5.6106E+03, 5.6150E+03, 5.6195E+03, 5.6240E+03, 5.6285E+03,
     + 5.6329E+03, 5.6374E+03, 5.6419E+03, 5.6464E+03, 5.6509E+03,
     +  5.6554E+03, 5.6599E+03, 5.6643E+03, 5.6688E+03, 5.6733E+03,
     + 5.6778E+03, 5.6823E+03, 5.6868E+03, 5.6913E+03, 5.6959E+03,
     +  5.7004E+03, 5.7049E+03, 5.7094E+03, 5.7139E+03, 5.7184E+03,
     + 5.7229E+03, 5.7275E+03, 5.7320E+03, 5.7365E+03, 5.7410E+03,
     +  5.7456E+03, 5.7501E+03, 5.7546E+03, 5.7592E+03, 5.7637E+03,
     + 5.7682E+03, 5.7728E+03, 5.7773E+03, 5.7819E+03, 5.7864E+03,
     +  5.7910E+03, 5.7955E+03, 5.8001E+03, 5.8046E+03, 5.8092E+03,
     + 5.8138E+03, 5.8183E+03, 5.8229E+03, 5.8274E+03, 5.8320E+03,
     +  5.8366E+03, 5.8412E+03, 5.8457E+03, 5.8503E+03, 5.8549E+03,
     + 5.8595E+03, 5.8641E+03, 5.8686E+03, 5.8732E+03, 5.8778E+03,
     +  5.8824E+03, 5.8870E+03, 5.8916E+03, 5.8962E+03, 5.9008E+03,
     + 5.9054E+03, 5.9100E+03, 5.9146E+03, 5.9192E+03, 5.9238E+03,
     +  5.9284E+03, 5.9330E+03, 5.9377E+03, 5.9423E+03, 5.9469E+03,
     + 5.9515E+03, 5.9561E+03, 5.9608E+03, 5.9654E+03, 5.9700E+03,
     +  5.9747E+03, 5.9793E+03, 5.9839E+03, 5.9886E+03, 5.9932E+03,
     + 5.9978E+03, 6.0025E+03, 6.0071E+03, 6.0118E+03, 6.0164E+03,
     +  6.0211E+03, 6.0257E+03, 6.0304E+03, 6.0351E+03, 6.0397E+03,
     + 6.0444E+03, 6.0491E+03, 6.0537E+03, 6.0584E+03, 6.0631E+03,
     +  6.0677E+03, 6.0724E+03, 6.0771E+03, 6.0818E+03, 6.0864E+03,
     + 6.0911E+03, 6.0958E+03, 6.1005E+03, 6.1052E+03, 6.1099E+03,
     +  6.1146E+03, 6.1193E+03, 6.1240E+03, 6.1287E+03, 6.1334E+03,
     + 6.1381E+03, 6.1428E+03, 6.1475E+03, 6.1522E+03, 6.1569E+03,
     +  6.1616E+03, 6.1663E+03, 6.1711E+03, 6.1758E+03, 6.1805E+03,
     + 6.1852E+03, 6.1900E+03, 6.1947E+03, 6.1994E+03, 6.2042E+03,
     +  6.2089E+03, 6.2136E+03, 6.2184E+03, 6.2231E+03, 6.2279E+03,
     + 6.2326E+03, 6.2373E+03, 6.2421E+03, 6.2468E+03, 6.2516E+03,
     +  6.2564E+03, 6.2611E+03, 6.2659E+03, 6.2706E+03, 6.2754E+03,
     + 6.2802E+03, 6.2849E+03, 6.2897E+03, 6.2945E+03, 6.2992E+03,
     +  6.3040E+03, 6.3088E+03, 6.3136E+03, 6.3184E+03, 6.3232E+03,
     + 6.3279E+03, 6.3327E+03, 6.3375E+03, 6.3423E+03, 6.3471E+03,
     +  6.3519E+03, 6.3567E+03, 6.3615E+03, 6.3663E+03, 6.3711E+03,
     + 6.3759E+03, 6.3807E+03, 6.3855E+03, 6.3904E+03, 6.3952E+03,
     +  6.4000E+03, 6.4048E+03, 6.4096E+03, 6.4145E+03, 6.4193E+03,
     + 6.4241E+03, 6.4289E+03, 6.4338E+03, 6.4386E+03, 6.4434E+03,
     +  6.4483E+03, 6.4531E+03, 6.4580E+03, 6.4628E+03, 6.4677E+03,
     + 6.4725E+03, 6.4774E+03, 6.4822E+03, 6.4871E+03, 6.4919E+03,
     +  6.4968E+03, 6.5016E+03, 6.5065E+03, 6.5114E+03, 6.5162E+03/
      data (QofT( 1,J),J=2409,4510)/ 6.5211E+03, 6.5260E+03,
     +  6.5309E+03, 6.5357E+03, 6.5406E+03, 6.5455E+03, 6.5504E+03,
     +  6.5553E+03,
     + 6.5601E+03, 6.5650E+03, 6.5699E+03, 6.5748E+03, 6.5797E+03,
     +  6.5846E+03, 6.5895E+03, 6.5944E+03, 6.5993E+03, 6.6042E+03,
     + 6.6091E+03, 6.6140E+03, 6.6189E+03, 6.6239E+03, 6.6288E+03,
     +  6.6337E+03, 6.6386E+03, 6.6435E+03, 6.6485E+03, 6.6534E+03,
     + 6.6583E+03, 6.6632E+03, 6.6682E+03, 6.6731E+03, 6.6780E+03,
     +  6.6830E+03, 6.6879E+03, 6.6929E+03, 6.6978E+03, 6.7028E+03,
     + 6.7077E+03, 6.7127E+03, 6.7176E+03, 6.7226E+03, 6.7275E+03,
     +  6.7325E+03, 6.7374E+03, 6.7424E+03, 6.7474E+03, 6.7523E+03,
     + 6.7573E+03, 6.7623E+03, 6.7673E+03, 6.7722E+03, 6.7772E+03,
     +  6.7822E+03, 6.7872E+03, 6.7922E+03, 6.7971E+03, 6.8021E+03,
     + 6.8071E+03, 6.8121E+03, 6.8171E+03, 6.8221E+03, 6.8271E+03,
     +  6.8321E+03, 6.8371E+03, 6.8421E+03, 6.8471E+03, 6.8521E+03,
     + 6.8571E+03, 6.8622E+03, 6.8672E+03, 6.8722E+03, 6.8772E+03,
     +  6.8822E+03, 6.8872E+03, 6.8923E+03, 6.8973E+03, 6.9023E+03,
     + 6.9074E+03, 6.9124E+03, 6.9174E+03, 6.9225E+03, 6.9275E+03,
     +  6.9326E+03, 6.9376E+03, 6.9426E+03, 6.9477E+03, 6.9527E+03,
     + 6.9578E+03, 6.9629E+03, 6.9679E+03, 6.9730E+03, 6.9780E+03,
     +  6.9831E+03, 6.9882E+03, 6.9932E+03, 6.9983E+03, 7.0034E+03,
     + 7.0084E+03, 7.0135E+03, 7.0186E+03, 7.0237E+03, 7.0288E+03,
     +  7.0338E+03, 7.0389E+03, 7.0440E+03, 7.0491E+03, 7.0542E+03,
     + 7.0593E+03, 7.0644E+03, 7.0695E+03, 7.0746E+03, 7.0797E+03,
     +  7.0848E+03, 7.0899E+03, 7.0950E+03, 7.1001E+03, 7.1052E+03,
     + 7.1104E+03, 7.1155E+03, 7.1206E+03, 7.1257E+03, 7.1308E+03,
     +  7.1360E+03, 7.1411E+03, 7.1462E+03, 7.1514E+03, 7.1565E+03,
     + 7.1616E+03, 7.1668E+03, 7.1719E+03, 7.1771E+03, 7.1822E+03,
     +  7.1874E+03, 7.1925E+03, 7.1977E+03, 7.2028E+03, 7.2080E+03,
     + 7.2131E+03, 7.2183E+03, 7.2234E+03, 7.2286E+03, 7.2338E+03,
     +  7.2389E+03, 7.2441E+03, 7.2493E+03, 7.2545E+03, 7.2596E+03,
     + 7.2648E+03, 7.2700E+03, 7.2752E+03, 7.2804E+03, 7.2855E+03,
     +  7.2907E+03, 7.2959E+03, 7.3011E+03, 7.3063E+03, 7.3115E+03,
     + 7.3167E+03, 7.3219E+03, 7.3271E+03, 7.3323E+03, 7.3375E+03,
     +  7.3427E+03, 7.3480E+03, 7.3532E+03, 7.3584E+03, 7.3636E+03,
     + 7.3688E+03, 7.3740E+03, 7.3793E+03, 7.3845E+03, 7.3897E+03,
     +  7.3950E+03, 7.4002E+03, 7.4054E+03, 7.4107E+03, 7.4159E+03,
     + 7.4211E+03, 7.4264E+03, 7.4316E+03, 7.4369E+03, 7.4421E+03,
     +  7.4474E+03, 7.4526E+03, 7.4579E+03, 7.4631E+03, 7.4684E+03,
     + 7.4737E+03, 7.4789E+03, 7.4842E+03, 7.4895E+03, 7.4947E+03,
     +  7.5000E+03, 7.5053E+03, 7.5106E+03, 7.5158E+03, 7.5211E+03,
     + 7.5264E+03, 7.5317E+03, 7.5370E+03, 7.5423E+03, 7.5476E+03,
     +  7.5529E+03, 7.5582E+03, 7.5635E+03, 7.5688E+03, 7.5741E+03,
     + 7.5794E+03, 7.5847E+03, 7.5900E+03, 7.5953E+03, 7.6006E+03,
     +  7.6059E+03, 7.6112E+03, 7.6165E+03, 7.6219E+03, 7.6272E+03,
     + 7.6325E+03, 7.6378E+03, 7.6432E+03, 7.6485E+03, 7.6538E+03,
     +  7.6592E+03, 7.6645E+03, 7.6699E+03, 7.6752E+03, 7.6805E+03,
     + 7.6859E+03, 7.6912E+03, 7.6966E+03, 7.7019E+03, 7.7073E+03,
     +  7.7126E+03, 7.7180E+03, 7.7234E+03, 7.7287E+03, 7.7341E+03,
     + 7.7395E+03, 7.7448E+03, 7.7502E+03, 7.7556E+03, 7.7610E+03,
     +  7.7663E+03, 7.7717E+03, 7.7771E+03, 7.7825E+03, 7.7879E+03,
     + 7.7933E+03, 7.7986E+03, 7.8040E+03, 7.8094E+03, 7.8148E+03,
     +  7.8202E+03, 7.8256E+03, 7.8310E+03, 7.8364E+03, 7.8418E+03,
     + 7.8473E+03, 7.8527E+03, 7.8581E+03, 7.8635E+03, 7.8689E+03,
     +  7.8743E+03, 7.8798E+03, 7.8852E+03, 7.8906E+03, 7.8960E+03,
     + 7.9015E+03, 7.9069E+03, 7.9123E+03, 7.9178E+03, 7.9232E+03,
     +  7.9286E+03, 7.9341E+03, 7.9395E+03, 7.9450E+03, 7.9504E+03,
     + 7.9559E+03, 7.9613E+03, 7.9668E+03, 7.9722E+03, 7.9777E+03,
     +  7.9832E+03, 7.9886E+03, 7.9941E+03, 7.9996E+03, 8.0050E+03,
     + 8.0105E+03, 8.0160E+03, 8.0215E+03, 8.0269E+03, 8.0324E+03,
     +  8.0379E+03, 8.0434E+03, 8.0489E+03, 8.0544E+03, 8.0599E+03,
     + 8.0653E+03, 8.0708E+03, 8.0763E+03, 8.0818E+03, 8.0873E+03,
     +  8.0928E+03, 8.0984E+03, 8.1039E+03, 8.1094E+03, 8.1149E+03,
     + 8.1204E+03, 8.1259E+03, 8.1314E+03, 8.1370E+03, 8.1425E+03,
     +  8.1480E+03, 8.1535E+03, 8.1591E+03, 8.1646E+03, 8.1701E+03,
     + 8.1757E+03, 8.1812E+03, 8.1867E+03, 8.1923E+03, 8.1978E+03,
     +  8.2034E+03, 8.2089E+03, 8.2145E+03, 8.2200E+03, 8.2256E+03,
     + 8.2311E+03, 8.2367E+03, 8.2422E+03, 8.2478E+03, 8.2534E+03,
     +  8.2589E+03, 8.2645E+03, 8.2701E+03, 8.2757E+03, 8.2812E+03,
     + 8.2868E+03, 8.2924E+03, 8.2980E+03, 8.3036E+03, 8.3091E+03,
     +  8.3147E+03, 8.3203E+03, 8.3259E+03, 8.3315E+03, 8.3371E+03,
     + 8.3427E+03, 8.3483E+03, 8.3539E+03, 8.3595E+03, 8.3651E+03,
     +  8.3707E+03, 8.3763E+03, 8.3820E+03, 8.3876E+03, 8.3932E+03,
     + 8.3988E+03, 8.4044E+03, 8.4101E+03, 8.4157E+03, 8.4213E+03,
     +  8.4269E+03, 8.4326E+03, 8.4382E+03, 8.4438E+03, 8.4495E+03,
     + 8.4551E+03, 8.4608E+03, 8.4664E+03, 8.4721E+03, 8.4777E+03,
     +  8.4834E+03, 8.4890E+03, 8.4947E+03, 8.5003E+03, 8.5060E+03,
     + 8.5116E+03, 8.5173E+03, 8.5230E+03, 8.5286E+03, 8.5343E+03,
     +  8.5400E+03, 8.5457E+03, 8.5513E+03, 8.5570E+03, 8.5627E+03,
     + 8.5684E+03, 8.5741E+03, 8.5797E+03, 8.5854E+03, 8.5911E+03,
     +  8.5968E+03, 8.6025E+03, 8.6082E+03, 8.6139E+03, 8.6196E+03,
     + 8.6253E+03, 8.6310E+03, 8.6367E+03, 8.6424E+03, 8.6482E+03,
     +  8.6539E+03, 8.6596E+03, 8.6653E+03, 8.6710E+03, 8.6767E+03,
     + 8.6825E+03, 8.6882E+03, 8.6939E+03, 8.6997E+03, 8.7054E+03,
     +  8.7111E+03, 8.7169E+03, 8.7226E+03, 8.7283E+03, 8.7341E+03,
     + 8.7398E+03, 8.7456E+03, 8.7513E+03, 8.7571E+03, 8.7628E+03,
     +  8.7686E+03, 8.7744E+03, 8.7801E+03, 8.7859E+03, 8.7916E+03,
     + 8.7974E+03, 8.8032E+03, 8.8089E+03, 8.8147E+03, 8.8205E+03,
     +  8.8263E+03, 8.8321E+03, 8.8378E+03, 8.8436E+03, 8.8494E+03,
     + 8.8552E+03, 8.8610E+03, 8.8668E+03, 8.8726E+03, 8.8784E+03,
     +  8.8842E+03, 8.8900E+03, 8.8958E+03, 8.9016E+03, 8.9074E+03,
     + 8.9132E+03, 8.9190E+03, 8.9248E+03, 8.9306E+03, 8.9364E+03,
     +  8.9423E+03, 8.9481E+03, 8.9539E+03, 8.9597E+03, 8.9656E+03,
     + 8.9714E+03, 8.9772E+03, 8.9831E+03, 8.9889E+03, 8.9947E+03,
     +  9.0006E+03, 9.0064E+03, 9.0123E+03, 9.0181E+03, 9.0240E+03,
     + 9.0298E+03, 9.0357E+03, 9.0415E+03, 9.0474E+03, 9.0532E+03,
     +  9.0591E+03, 9.0650E+03, 9.0708E+03, 9.0767E+03, 9.0826E+03,
     + 9.0884E+03, 9.0943E+03, 9.1002E+03, 9.1061E+03, 9.1119E+03,
     +  9.1178E+03, 9.1237E+03, 9.1296E+03, 9.1355E+03, 9.1414E+03,
     + 9.1473E+03, 9.1532E+03, 9.1591E+03, 9.1650E+03, 9.1709E+03,
     +  9.1768E+03, 9.1827E+03, 9.1886E+03, 9.1945E+03, 9.2004E+03,
     + 9.2063E+03, 9.2122E+03, 9.2181E+03, 9.2241E+03, 9.2300E+03,
     +  9.2359E+03, 9.2418E+03, 9.2478E+03, 9.2537E+03, 9.2596E+03,
     + 9.2656E+03, 9.2715E+03, 9.2774E+03, 9.2834E+03, 9.2893E+03,
     +  9.2953E+03, 9.3012E+03, 9.3072E+03, 9.3131E+03, 9.3191E+03,
     + 9.3250E+03, 9.3310E+03, 9.3369E+03, 9.3429E+03, 9.3489E+03,
     +  9.3548E+03, 9.3608E+03, 9.3668E+03, 9.3727E+03, 9.3787E+03,
     + 9.3847E+03, 9.3907E+03, 9.3967E+03, 9.4026E+03, 9.4086E+03,
     +  9.4146E+03, 9.4206E+03, 9.4266E+03, 9.4326E+03, 9.4386E+03,
     + 9.4446E+03, 9.4506E+03, 9.4566E+03, 9.4626E+03, 9.4686E+03,
     +  9.4746E+03, 9.4806E+03, 9.4866E+03, 9.4926E+03, 9.4987E+03,
     + 9.5047E+03, 9.5107E+03, 9.5167E+03, 9.5227E+03, 9.5288E+03,
     +  9.5348E+03, 9.5408E+03, 9.5469E+03, 9.5529E+03, 9.5589E+03,
     + 9.5650E+03, 9.5710E+03, 9.5771E+03, 9.5831E+03, 9.5892E+03,
     +  9.5952E+03, 9.6013E+03, 9.6073E+03, 9.6134E+03, 9.6194E+03,
     + 9.6255E+03, 9.6316E+03, 9.6376E+03, 9.6437E+03, 9.6498E+03,
     +  9.6558E+03, 9.6619E+03, 9.6680E+03, 9.6740E+03, 9.6801E+03,
     + 9.6862E+03, 9.6923E+03, 9.6984E+03, 9.7045E+03, 9.7106E+03,
     +  9.7167E+03, 9.7227E+03, 9.7288E+03, 9.7349E+03, 9.7410E+03,
     + 9.7471E+03, 9.7532E+03, 9.7594E+03, 9.7655E+03, 9.7716E+03,
     +  9.7777E+03, 9.7838E+03, 9.7899E+03, 9.7960E+03, 9.8022E+03,
     + 9.8083E+03, 9.8144E+03, 9.8205E+03, 9.8267E+03, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        11
      data (QofT( 2,J),J=1,2408)/ 8.0000E+00, 8.0001E+00, 8.0080E+00,
     +  8.0592E+00, 8.1967E+00, 8.4383E+00, 8.7774E+00, 9.1966E+00,
     + 9.6770E+00, 1.0202E+01, 1.0760E+01, 1.1341E+01, 1.1940E+01,
     +  1.2550E+01, 1.3171E+01, 1.3798E+01, 1.4431E+01, 1.5068E+01,
     + 1.5709E+01, 1.6353E+01, 1.7648E+01, 1.8950E+01, 2.0258E+01,
     +  2.1569E+01, 2.2883E+01, 2.4200E+01, 2.5519E+01, 2.6840E+01,
     + 2.8162E+01, 2.9485E+01, 3.0809E+01, 3.2134E+01, 3.3460E+01,
     +  3.4787E+01, 3.6114E+01, 3.7441E+01, 3.8769E+01, 4.0098E+01,
     + 4.1427E+01, 4.2756E+01, 4.4085E+01, 4.5415E+01, 4.6745E+01,
     +  4.8075E+01, 4.9405E+01, 5.0736E+01, 5.2067E+01, 5.3398E+01,
     + 5.4729E+01, 5.6060E+01, 5.7392E+01, 5.8723E+01, 6.0055E+01,
     +  6.1387E+01, 6.2719E+01, 6.4051E+01, 6.5383E+01, 6.6716E+01,
     + 6.8048E+01, 6.9381E+01, 7.0713E+01, 7.2046E+01, 7.3379E+01,
     +  7.4711E+01, 7.6044E+01, 7.7377E+01, 7.8710E+01, 8.0044E+01,
     + 8.1377E+01, 8.2710E+01, 8.4043E+01, 8.5377E+01, 8.6710E+01,
     +  8.8044E+01, 8.9377E+01, 9.0711E+01, 9.2045E+01, 9.3379E+01,
     + 9.4713E+01, 9.6046E+01, 9.7380E+01, 9.8714E+01, 1.0005E+02,
     +  1.0138E+02, 1.0272E+02, 1.0405E+02, 1.0539E+02, 1.0672E+02,
     + 1.0805E+02, 1.0939E+02, 1.1072E+02, 1.1206E+02, 1.1339E+02,
     +  1.1473E+02, 1.1606E+02, 1.1740E+02, 1.1873E+02, 1.2007E+02,
     + 1.2140E+02, 1.2274E+02, 1.2407E+02, 1.2541E+02, 1.2674E+02,
     +  1.2808E+02, 1.2941E+02, 1.3075E+02, 1.3208E+02, 1.3342E+02,
     + 1.3475E+02, 1.3609E+02, 1.3742E+02, 1.3876E+02, 1.4009E+02,
     +  1.4143E+02, 1.4276E+02, 1.4410E+02, 1.4544E+02, 1.4677E+02,
     + 1.4811E+02, 1.4944E+02, 1.5078E+02, 1.5212E+02, 1.5345E+02,
     +  1.5479E+02, 1.5612E+02, 1.5746E+02, 1.5880E+02, 1.6013E+02,
     + 1.6147E+02, 1.6280E+02, 1.6414E+02, 1.6548E+02, 1.6681E+02,
     +  1.6815E+02, 1.6949E+02, 1.7082E+02, 1.7216E+02, 1.7350E+02,
     + 1.7483E+02, 1.7617E+02, 1.7751E+02, 1.7884E+02, 1.8018E+02,
     +  1.8152E+02, 1.8285E+02, 1.8419E+02, 1.8553E+02, 1.8686E+02,
     + 1.8820E+02, 1.8954E+02, 1.9087E+02, 1.9221E+02, 1.9355E+02,
     +  1.9489E+02, 1.9622E+02, 1.9756E+02, 1.9890E+02, 2.0023E+02,
     + 2.0157E+02, 2.0291E+02, 2.0425E+02, 2.0559E+02, 2.0692E+02,
     +  2.0826E+02, 2.0960E+02, 2.1094E+02, 2.1227E+02, 2.1361E+02,
     + 2.1495E+02, 2.1629E+02, 2.1763E+02, 2.1896E+02, 2.2030E+02,
     +  2.2164E+02, 2.2298E+02, 2.2432E+02, 2.2565E+02, 2.2699E+02,
     + 2.2833E+02, 2.2967E+02, 2.3101E+02, 2.3235E+02, 2.3369E+02,
     +  2.3502E+02, 2.3636E+02, 2.3770E+02, 2.3904E+02, 2.4038E+02,
     + 2.4172E+02, 2.4306E+02, 2.4440E+02, 2.4574E+02, 2.4708E+02,
     +  2.4842E+02, 2.4975E+02, 2.5109E+02, 2.5243E+02, 2.5377E+02,
     + 2.5511E+02, 2.5645E+02, 2.5779E+02, 2.5913E+02, 2.6047E+02,
     +  2.6181E+02, 2.6315E+02, 2.6449E+02, 2.6583E+02, 2.6718E+02,
     + 2.6852E+02, 2.6986E+02, 2.7120E+02, 2.7254E+02, 2.7388E+02,
     +  2.7522E+02, 2.7656E+02, 2.7790E+02, 2.7924E+02, 2.8059E+02,
     + 2.8193E+02, 2.8327E+02, 2.8461E+02, 2.8595E+02, 2.8730E+02,
     +  2.8864E+02, 2.8998E+02, 2.9132E+02, 2.9267E+02, 2.9401E+02,
     + 2.9535E+02, 2.9669E+02, 2.9804E+02, 2.9938E+02, 3.0072E+02,
     +  3.0207E+02, 3.0341E+02, 3.0476E+02, 3.0610E+02, 3.0744E+02,
     + 3.0879E+02, 3.1013E+02, 3.1148E+02, 3.1282E+02, 3.1417E+02,
     +  3.1551E+02, 3.1686E+02, 3.1820E+02, 3.1955E+02, 3.2090E+02,
     + 3.2224E+02, 3.2359E+02, 3.2493E+02, 3.2628E+02, 3.2763E+02,
     +  3.2898E+02, 3.3032E+02, 3.3167E+02, 3.3302E+02, 3.3437E+02,
     + 3.3571E+02, 3.3706E+02, 3.3841E+02, 3.3976E+02, 3.4111E+02,
     +  3.4246E+02, 3.4381E+02, 3.4516E+02, 3.4651E+02, 3.4786E+02,
     + 3.4921E+02, 3.5056E+02, 3.5191E+02, 3.5326E+02, 3.5462E+02,
     +  3.5597E+02, 3.5732E+02, 3.5867E+02, 3.6003E+02, 3.6138E+02,
     + 3.6273E+02, 3.6409E+02, 3.6544E+02, 3.6679E+02, 3.6815E+02,
     +  3.6950E+02, 3.7086E+02, 3.7221E+02, 3.7357E+02, 3.7493E+02,
     + 3.7628E+02, 3.7764E+02, 3.7900E+02, 3.8036E+02, 3.8171E+02,
     +  3.8307E+02, 3.8443E+02, 3.8579E+02, 3.8715E+02, 3.8851E+02,
     + 3.8987E+02, 3.9123E+02, 3.9259E+02, 3.9395E+02, 3.9531E+02,
     +  3.9667E+02, 3.9804E+02, 3.9940E+02, 4.0076E+02, 4.0213E+02,
     + 4.0349E+02, 4.0485E+02, 4.0622E+02, 4.0759E+02, 4.0895E+02,
     +  4.1032E+02, 4.1168E+02, 4.1305E+02, 4.1442E+02, 4.1579E+02,
     + 4.1715E+02, 4.1852E+02, 4.1989E+02, 4.2126E+02, 4.2263E+02,
     +  4.2400E+02, 4.2537E+02, 4.2675E+02, 4.2812E+02, 4.2949E+02,
     + 4.3086E+02, 4.3224E+02, 4.3361E+02, 4.3499E+02, 4.3636E+02,
     +  4.3774E+02, 4.3911E+02, 4.4049E+02, 4.4187E+02, 4.4324E+02,
     + 4.4462E+02, 4.4600E+02, 4.4738E+02, 4.4876E+02, 4.5014E+02,
     +  4.5152E+02, 4.5290E+02, 4.5428E+02, 4.5567E+02, 4.5705E+02,
     + 4.5843E+02, 4.5982E+02, 4.6120E+02, 4.6259E+02, 4.6397E+02,
     +  4.6536E+02, 4.6675E+02, 4.6814E+02, 4.6953E+02, 4.7091E+02,
     + 4.7230E+02, 4.7369E+02, 4.7509E+02, 4.7648E+02, 4.7787E+02,
     +  4.7926E+02, 4.8066E+02, 4.8205E+02, 4.8344E+02, 4.8484E+02,
     + 4.8624E+02, 4.8763E+02, 4.8903E+02, 4.9043E+02, 4.9183E+02,
     +  4.9323E+02, 4.9463E+02, 4.9603E+02, 4.9743E+02, 4.9883E+02,
     + 5.0023E+02, 5.0164E+02, 5.0304E+02, 5.0445E+02, 5.0585E+02,
     +  5.0726E+02, 5.0866E+02, 5.1007E+02, 5.1148E+02, 5.1289E+02,
     + 5.1430E+02, 5.1571E+02, 5.1712E+02, 5.1853E+02, 5.1995E+02,
     +  5.2136E+02, 5.2278E+02, 5.2419E+02, 5.2561E+02, 5.2702E+02,
     + 5.2844E+02, 5.2986E+02, 5.3128E+02, 5.3270E+02, 5.3412E+02,
     +  5.3554E+02, 5.3696E+02, 5.3838E+02, 5.3981E+02, 5.4123E+02,
     + 5.4266E+02, 5.4408E+02, 5.4551E+02, 5.4694E+02, 5.4837E+02,
     +  5.4980E+02, 5.5123E+02, 5.5266E+02, 5.5409E+02, 5.5552E+02,
     + 5.5696E+02, 5.5839E+02, 5.5983E+02, 5.6126E+02, 5.6270E+02,
     +  5.6414E+02, 5.6558E+02, 5.6701E+02, 5.6845E+02, 5.6990E+02,
     + 5.7134E+02, 5.7278E+02, 5.7423E+02, 5.7567E+02, 5.7712E+02,
     +  5.7856E+02, 5.8001E+02, 5.8146E+02, 5.8291E+02, 5.8436E+02,
     + 5.8581E+02, 5.8726E+02, 5.8871E+02, 5.9017E+02, 5.9162E+02,
     +  5.9308E+02, 5.9453E+02, 5.9599E+02, 5.9745E+02, 5.9891E+02,
     + 6.0037E+02, 6.0183E+02, 6.0329E+02, 6.0475E+02, 6.0622E+02,
     +  6.0768E+02, 6.0915E+02, 6.1062E+02, 6.1208E+02, 6.1355E+02,
     + 6.1502E+02, 6.1649E+02, 6.1797E+02, 6.1944E+02, 6.2091E+02,
     +  6.2239E+02, 6.2386E+02, 6.2534E+02, 6.2682E+02, 6.2829E+02,
     + 6.2977E+02, 6.3125E+02, 6.3274E+02, 6.3422E+02, 6.3570E+02,
     +  6.3719E+02, 6.3867E+02, 6.4016E+02, 6.4165E+02, 6.4314E+02,
     + 6.4462E+02, 6.4612E+02, 6.4761E+02, 6.4910E+02, 6.5059E+02,
     +  6.5209E+02, 6.5358E+02, 6.5508E+02, 6.5658E+02, 6.5808E+02,
     + 6.5958E+02, 6.6108E+02, 6.6258E+02, 6.6408E+02, 6.6559E+02,
     +  6.6709E+02, 6.6860E+02, 6.7011E+02, 6.7162E+02, 6.7313E+02,
     + 6.7464E+02, 6.7615E+02, 6.7766E+02, 6.7917E+02, 6.8069E+02,
     +  6.8221E+02, 6.8372E+02, 6.8524E+02, 6.8676E+02, 6.8828E+02,
     + 6.8980E+02, 6.9133E+02, 6.9285E+02, 6.9437E+02, 6.9590E+02,
     +  6.9743E+02, 6.9895E+02, 7.0048E+02, 7.0201E+02, 7.0355E+02,
     + 7.0508E+02, 7.0661E+02, 7.0815E+02, 7.0968E+02, 7.1122E+02,
     +  7.1276E+02, 7.1430E+02, 7.1584E+02, 7.1738E+02, 7.1892E+02,
     + 7.2047E+02, 7.2201E+02, 7.2356E+02, 7.2511E+02, 7.2666E+02,
     +  7.2820E+02, 7.2976E+02, 7.3131E+02, 7.3286E+02, 7.3442E+02,
     + 7.3597E+02, 7.3753E+02, 7.3909E+02, 7.4064E+02, 7.4220E+02,
     +  7.4377E+02, 7.4533E+02, 7.4689E+02, 7.4846E+02, 7.5002E+02,
     + 7.5159E+02, 7.5316E+02, 7.5473E+02, 7.5630E+02, 7.5787E+02,
     +  7.5945E+02, 7.6102E+02, 7.6260E+02, 7.6417E+02, 7.6575E+02,
     + 7.6733E+02, 7.6891E+02, 7.7049E+02, 7.7208E+02, 7.7366E+02,
     +  7.7525E+02, 7.7683E+02, 7.7842E+02, 7.8001E+02, 7.8160E+02,
     + 7.8319E+02, 7.8478E+02, 7.8638E+02, 7.8797E+02, 7.8957E+02,
     +  7.9117E+02, 7.9277E+02, 7.9437E+02, 7.9597E+02, 7.9757E+02,
     + 7.9917E+02, 8.0078E+02, 8.0238E+02, 8.0399E+02, 8.0560E+02,
     +  8.0721E+02, 8.0882E+02, 8.1043E+02, 8.1205E+02, 8.1366E+02,
     + 8.1528E+02, 8.1690E+02, 8.1852E+02, 8.2014E+02, 8.2176E+02,
     +  8.2338E+02, 8.2500E+02, 8.2663E+02, 8.2826E+02, 8.2988E+02,
     + 8.3151E+02, 8.3314E+02, 8.3478E+02, 8.3641E+02, 8.3804E+02,
     +  8.3968E+02, 8.4131E+02, 8.4295E+02, 8.4459E+02, 8.4623E+02,
     + 8.4787E+02, 8.4952E+02, 8.5116E+02, 8.5281E+02, 8.5446E+02,
     +  8.5610E+02, 8.5775E+02, 8.5940E+02, 8.6106E+02, 8.6271E+02,
     + 8.6437E+02, 8.6602E+02, 8.6768E+02, 8.6934E+02, 8.7100E+02,
     +  8.7266E+02, 8.7432E+02, 8.7599E+02, 8.7765E+02, 8.7932E+02,
     + 8.8099E+02, 8.8266E+02, 8.8433E+02, 8.8600E+02, 8.8767E+02,
     +  8.8935E+02, 8.9102E+02, 8.9270E+02, 8.9438E+02, 8.9606E+02,
     + 8.9774E+02, 8.9942E+02, 9.0111E+02, 9.0279E+02, 9.0448E+02,
     +  9.0617E+02, 9.0786E+02, 9.0955E+02, 9.1124E+02, 9.1293E+02,
     + 9.1463E+02, 9.1632E+02, 9.1802E+02, 9.1972E+02, 9.2142E+02,
     +  9.2312E+02, 9.2483E+02, 9.2653E+02, 9.2824E+02, 9.2994E+02,
     + 9.3165E+02, 9.3336E+02, 9.3507E+02, 9.3678E+02, 9.3850E+02,
     +  9.4021E+02, 9.4193E+02, 9.4365E+02, 9.4537E+02, 9.4709E+02,
     + 9.4881E+02, 9.5053E+02, 9.5226E+02, 9.5399E+02, 9.5571E+02,
     +  9.5744E+02, 9.5917E+02, 9.6090E+02, 9.6264E+02, 9.6437E+02,
     + 9.6611E+02, 9.6784E+02, 9.6958E+02, 9.7132E+02, 9.7307E+02,
     +  9.7481E+02, 9.7655E+02, 9.7830E+02, 9.8004E+02, 9.8179E+02,
     + 9.8354E+02, 9.8529E+02, 9.8705E+02, 9.8880E+02, 9.9056E+02,
     +  9.9231E+02, 9.9407E+02, 9.9583E+02, 9.9759E+02, 9.9935E+02,
     + 1.0011E+03, 1.0029E+03, 1.0047E+03, 1.0064E+03, 1.0082E+03,
     +  1.0100E+03, 1.0117E+03, 1.0135E+03, 1.0153E+03, 1.0171E+03,
     + 1.0188E+03, 1.0206E+03, 1.0224E+03, 1.0242E+03, 1.0260E+03,
     +  1.0277E+03, 1.0295E+03, 1.0313E+03, 1.0331E+03, 1.0349E+03,
     + 1.0367E+03, 1.0385E+03, 1.0403E+03, 1.0421E+03, 1.0439E+03,
     +  1.0457E+03, 1.0475E+03, 1.0493E+03, 1.0511E+03, 1.0529E+03,
     + 1.0547E+03, 1.0565E+03, 1.0583E+03, 1.0601E+03, 1.0619E+03,
     +  1.0638E+03, 1.0656E+03, 1.0674E+03, 1.0692E+03, 1.0710E+03,
     + 1.0728E+03, 1.0747E+03, 1.0765E+03, 1.0783E+03, 1.0801E+03,
     +  1.0820E+03, 1.0838E+03, 1.0856E+03, 1.0875E+03, 1.0893E+03,
     + 1.0911E+03, 1.0930E+03, 1.0948E+03, 1.0966E+03, 1.0985E+03,
     +  1.1003E+03, 1.1022E+03, 1.1040E+03, 1.1059E+03, 1.1077E+03,
     + 1.1096E+03, 1.1114E+03, 1.1133E+03, 1.1151E+03, 1.1170E+03,
     +  1.1188E+03, 1.1207E+03, 1.1226E+03, 1.1244E+03, 1.1263E+03,
     + 1.1282E+03, 1.1300E+03, 1.1319E+03, 1.1338E+03, 1.1356E+03,
     +  1.1375E+03, 1.1394E+03, 1.1412E+03, 1.1431E+03, 1.1450E+03,
     + 1.1469E+03, 1.1488E+03, 1.1506E+03, 1.1525E+03, 1.1544E+03,
     +  1.1563E+03, 1.1582E+03, 1.1601E+03, 1.1620E+03, 1.1639E+03,
     + 1.1658E+03, 1.1677E+03, 1.1696E+03, 1.1715E+03, 1.1734E+03,
     +  1.1753E+03, 1.1772E+03, 1.1791E+03, 1.1810E+03, 1.1829E+03,
     + 1.1848E+03, 1.1867E+03, 1.1886E+03, 1.1905E+03, 1.1924E+03,
     +  1.1944E+03, 1.1963E+03, 1.1982E+03, 1.2001E+03, 1.2020E+03,
     + 1.2040E+03, 1.2059E+03, 1.2078E+03, 1.2098E+03, 1.2117E+03,
     +  1.2136E+03, 1.2156E+03, 1.2175E+03, 1.2194E+03, 1.2214E+03,
     + 1.2233E+03, 1.2252E+03, 1.2272E+03, 1.2291E+03, 1.2311E+03,
     +  1.2330E+03, 1.2350E+03, 1.2369E+03, 1.2389E+03, 1.2408E+03,
     + 1.2428E+03, 1.2447E+03, 1.2467E+03, 1.2487E+03, 1.2506E+03,
     +  1.2526E+03, 1.2545E+03, 1.2565E+03, 1.2585E+03, 1.2604E+03,
     + 1.2624E+03, 1.2644E+03, 1.2664E+03, 1.2683E+03, 1.2703E+03,
     +  1.2723E+03, 1.2743E+03, 1.2763E+03, 1.2782E+03, 1.2802E+03,
     + 1.2822E+03, 1.2842E+03, 1.2862E+03, 1.2882E+03, 1.2902E+03,
     +  1.2922E+03, 1.2941E+03, 1.2961E+03, 1.2981E+03, 1.3001E+03,
     + 1.3021E+03, 1.3041E+03, 1.3061E+03, 1.3082E+03, 1.3102E+03,
     +  1.3122E+03, 1.3142E+03, 1.3162E+03, 1.3182E+03, 1.3202E+03,
     + 1.3222E+03, 1.3243E+03, 1.3263E+03, 1.3283E+03, 1.3303E+03,
     +  1.3323E+03, 1.3344E+03, 1.3364E+03, 1.3384E+03, 1.3405E+03,
     + 1.3425E+03, 1.3445E+03, 1.3466E+03, 1.3486E+03, 1.3506E+03,
     +  1.3527E+03, 1.3547E+03, 1.3568E+03, 1.3588E+03, 1.3608E+03,
     + 1.3629E+03, 1.3649E+03, 1.3670E+03, 1.3690E+03, 1.3711E+03,
     +  1.3732E+03, 1.3752E+03, 1.3773E+03, 1.3793E+03, 1.3814E+03,
     + 1.3835E+03, 1.3855E+03, 1.3876E+03, 1.3896E+03, 1.3917E+03,
     +  1.3938E+03, 1.3959E+03, 1.3979E+03, 1.4000E+03, 1.4021E+03,
     + 1.4042E+03, 1.4062E+03, 1.4083E+03, 1.4104E+03, 1.4125E+03,
     +  1.4146E+03, 1.4167E+03, 1.4188E+03, 1.4209E+03, 1.4229E+03,
     + 1.4250E+03, 1.4271E+03, 1.4292E+03, 1.4313E+03, 1.4334E+03,
     +  1.4355E+03, 1.4376E+03, 1.4397E+03, 1.4419E+03, 1.4440E+03,
     + 1.4461E+03, 1.4482E+03, 1.4503E+03, 1.4524E+03, 1.4545E+03,
     +  1.4567E+03, 1.4588E+03, 1.4609E+03, 1.4630E+03, 1.4651E+03,
     + 1.4673E+03, 1.4694E+03, 1.4715E+03, 1.4737E+03, 1.4758E+03,
     +  1.4779E+03, 1.4801E+03, 1.4822E+03, 1.4843E+03, 1.4865E+03,
     + 1.4886E+03, 1.4908E+03, 1.4929E+03, 1.4951E+03, 1.4972E+03,
     +  1.4993E+03, 1.5015E+03, 1.5037E+03, 1.5058E+03, 1.5080E+03,
     + 1.5101E+03, 1.5123E+03, 1.5144E+03, 1.5166E+03, 1.5188E+03,
     +  1.5209E+03, 1.5231E+03, 1.5253E+03, 1.5274E+03, 1.5296E+03,
     + 1.5318E+03, 1.5340E+03, 1.5361E+03, 1.5383E+03, 1.5405E+03,
     +  1.5427E+03, 1.5449E+03, 1.5471E+03, 1.5492E+03, 1.5514E+03,
     + 1.5536E+03, 1.5558E+03, 1.5580E+03, 1.5602E+03, 1.5624E+03,
     +  1.5646E+03, 1.5668E+03, 1.5690E+03, 1.5712E+03, 1.5734E+03,
     + 1.5756E+03, 1.5778E+03, 1.5800E+03, 1.5822E+03, 1.5845E+03,
     +  1.5867E+03, 1.5889E+03, 1.5911E+03, 1.5933E+03, 1.5955E+03,
     + 1.5978E+03, 1.6000E+03, 1.6022E+03, 1.6044E+03, 1.6067E+03,
     +  1.6089E+03, 1.6111E+03, 1.6134E+03, 1.6156E+03, 1.6178E+03,
     + 1.6201E+03, 1.6223E+03, 1.6246E+03, 1.6268E+03, 1.6291E+03,
     +  1.6313E+03, 1.6335E+03, 1.6358E+03, 1.6380E+03, 1.6403E+03,
     + 1.6426E+03, 1.6448E+03, 1.6471E+03, 1.6493E+03, 1.6516E+03,
     +  1.6539E+03, 1.6561E+03, 1.6584E+03, 1.6607E+03, 1.6629E+03,
     + 1.6652E+03, 1.6675E+03, 1.6697E+03, 1.6720E+03, 1.6743E+03,
     +  1.6766E+03, 1.6789E+03, 1.6811E+03, 1.6834E+03, 1.6857E+03,
     + 1.6880E+03, 1.6903E+03, 1.6926E+03, 1.6949E+03, 1.6972E+03,
     +  1.6995E+03, 1.7018E+03, 1.7041E+03, 1.7064E+03, 1.7087E+03,
     + 1.7110E+03, 1.7133E+03, 1.7156E+03, 1.7179E+03, 1.7202E+03,
     +  1.7225E+03, 1.7248E+03, 1.7271E+03, 1.7295E+03, 1.7318E+03,
     + 1.7341E+03, 1.7364E+03, 1.7387E+03, 1.7411E+03, 1.7434E+03,
     +  1.7457E+03, 1.7481E+03, 1.7504E+03, 1.7527E+03, 1.7551E+03,
     + 1.7574E+03, 1.7597E+03, 1.7621E+03, 1.7644E+03, 1.7668E+03,
     +  1.7691E+03, 1.7714E+03, 1.7738E+03, 1.7761E+03, 1.7785E+03,
     + 1.7809E+03, 1.7832E+03, 1.7856E+03, 1.7879E+03, 1.7903E+03,
     +  1.7926E+03, 1.7950E+03, 1.7974E+03, 1.7997E+03, 1.8021E+03,
     + 1.8045E+03, 1.8069E+03, 1.8092E+03, 1.8116E+03, 1.8140E+03,
     +  1.8164E+03, 1.8187E+03, 1.8211E+03, 1.8235E+03, 1.8259E+03,
     + 1.8283E+03, 1.8307E+03, 1.8331E+03, 1.8354E+03, 1.8378E+03,
     +  1.8402E+03, 1.8426E+03, 1.8450E+03, 1.8474E+03, 1.8498E+03,
     + 1.8522E+03, 1.8546E+03, 1.8570E+03, 1.8595E+03, 1.8619E+03,
     +  1.8643E+03, 1.8667E+03, 1.8691E+03, 1.8715E+03, 1.8739E+03,
     + 1.8764E+03, 1.8788E+03, 1.8812E+03, 1.8836E+03, 1.8861E+03,
     +  1.8885E+03, 1.8909E+03, 1.8933E+03, 1.8958E+03, 1.8982E+03,
     + 1.9007E+03, 1.9031E+03, 1.9055E+03, 1.9080E+03, 1.9104E+03,
     +  1.9129E+03, 1.9153E+03, 1.9178E+03, 1.9202E+03, 1.9227E+03,
     + 1.9251E+03, 1.9276E+03, 1.9300E+03, 1.9325E+03, 1.9349E+03,
     +  1.9374E+03, 1.9399E+03, 1.9423E+03, 1.9448E+03, 1.9473E+03,
     + 1.9497E+03, 1.9522E+03, 1.9547E+03, 1.9572E+03, 1.9596E+03,
     +  1.9621E+03, 1.9646E+03, 1.9671E+03, 1.9696E+03, 1.9721E+03,
     + 1.9745E+03, 1.9770E+03, 1.9795E+03, 1.9820E+03, 1.9845E+03,
     +  1.9870E+03, 1.9895E+03, 1.9920E+03, 1.9945E+03, 1.9970E+03,
     + 1.9995E+03, 2.0020E+03, 2.0045E+03, 2.0070E+03, 2.0095E+03,
     +  2.0121E+03, 2.0146E+03, 2.0171E+03, 2.0196E+03, 2.0221E+03,
     + 2.0246E+03, 2.0272E+03, 2.0297E+03, 2.0322E+03, 2.0347E+03,
     +  2.0373E+03, 2.0398E+03, 2.0423E+03, 2.0449E+03, 2.0474E+03,
     + 2.0499E+03, 2.0525E+03, 2.0550E+03, 2.0576E+03, 2.0601E+03,
     +  2.0627E+03, 2.0652E+03, 2.0678E+03, 2.0703E+03, 2.0729E+03,
     + 2.0754E+03, 2.0780E+03, 2.0805E+03, 2.0831E+03, 2.0857E+03,
     +  2.0882E+03, 2.0908E+03, 2.0934E+03, 2.0959E+03, 2.0985E+03,
     + 2.1011E+03, 2.1036E+03, 2.1062E+03, 2.1088E+03, 2.1114E+03,
     +  2.1140E+03, 2.1165E+03, 2.1191E+03, 2.1217E+03, 2.1243E+03,
     + 2.1269E+03, 2.1295E+03, 2.1321E+03, 2.1347E+03, 2.1373E+03,
     +  2.1399E+03, 2.1425E+03, 2.1451E+03, 2.1477E+03, 2.1503E+03,
     + 2.1529E+03, 2.1555E+03, 2.1581E+03, 2.1607E+03, 2.1633E+03,
     +  2.1659E+03, 2.1685E+03, 2.1712E+03, 2.1738E+03, 2.1764E+03,
     + 2.1790E+03, 2.1817E+03, 2.1843E+03, 2.1869E+03, 2.1895E+03,
     +  2.1922E+03, 2.1948E+03, 2.1974E+03, 2.2001E+03, 2.2027E+03,
     + 2.2054E+03, 2.2080E+03, 2.2106E+03, 2.2133E+03, 2.2159E+03,
     +  2.2186E+03, 2.2212E+03, 2.2239E+03, 2.2266E+03, 2.2292E+03,
     + 2.2319E+03, 2.2345E+03, 2.2372E+03, 2.2399E+03, 2.2425E+03,
     +  2.2452E+03, 2.2479E+03, 2.2505E+03, 2.2532E+03, 2.2559E+03,
     + 2.2585E+03, 2.2612E+03, 2.2639E+03, 2.2666E+03, 2.2693E+03,
     +  2.2719E+03, 2.2746E+03, 2.2773E+03, 2.2800E+03, 2.2827E+03,
     + 2.2854E+03, 2.2881E+03, 2.2908E+03, 2.2935E+03, 2.2962E+03,
     +  2.2989E+03, 2.3016E+03, 2.3043E+03, 2.3070E+03, 2.3097E+03,
     + 2.3124E+03, 2.3151E+03, 2.3178E+03, 2.3206E+03, 2.3233E+03,
     +  2.3260E+03, 2.3287E+03, 2.3314E+03, 2.3342E+03, 2.3369E+03,
     + 2.3396E+03, 2.3423E+03, 2.3451E+03, 2.3478E+03, 2.3505E+03,
     +  2.3533E+03, 2.3560E+03, 2.3588E+03, 2.3615E+03, 2.3642E+03,
     + 2.3670E+03, 2.3697E+03, 2.3725E+03, 2.3752E+03, 2.3780E+03,
     +  2.3807E+03, 2.3835E+03, 2.3862E+03, 2.3890E+03, 2.3918E+03,
     + 2.3945E+03, 2.3973E+03, 2.4001E+03, 2.4028E+03, 2.4056E+03,
     +  2.4084E+03, 2.4111E+03, 2.4139E+03, 2.4167E+03, 2.4195E+03,
     + 2.4222E+03, 2.4250E+03, 2.4278E+03, 2.4306E+03, 2.4334E+03,
     +  2.4362E+03, 2.4390E+03, 2.4418E+03, 2.4446E+03, 2.4473E+03,
     + 2.4501E+03, 2.4529E+03, 2.4557E+03, 2.4585E+03, 2.4614E+03,
     +  2.4642E+03, 2.4670E+03, 2.4698E+03, 2.4726E+03, 2.4754E+03,
     + 2.4782E+03, 2.4810E+03, 2.4839E+03, 2.4867E+03, 2.4895E+03,
     +  2.4923E+03, 2.4951E+03, 2.4980E+03, 2.5008E+03, 2.5036E+03,
     + 2.5065E+03, 2.5093E+03, 2.5121E+03, 2.5150E+03, 2.5178E+03,
     +  2.5207E+03, 2.5235E+03, 2.5263E+03, 2.5292E+03, 2.5320E+03,
     + 2.5349E+03, 2.5377E+03, 2.5406E+03, 2.5434E+03, 2.5463E+03,
     +  2.5492E+03, 2.5520E+03, 2.5549E+03, 2.5578E+03, 2.5606E+03,
     + 2.5635E+03, 2.5664E+03, 2.5692E+03, 2.5721E+03, 2.5750E+03,
     +  2.5779E+03, 2.5807E+03, 2.5836E+03, 2.5865E+03, 2.5894E+03,
     + 2.5923E+03, 2.5951E+03, 2.5980E+03, 2.6009E+03, 2.6038E+03,
     +  2.6067E+03, 2.6096E+03, 2.6125E+03, 2.6154E+03, 2.6183E+03,
     + 2.6212E+03, 2.6241E+03, 2.6270E+03, 2.6299E+03, 2.6328E+03,
     +  2.6358E+03, 2.6387E+03, 2.6416E+03, 2.6445E+03, 2.6474E+03,
     + 2.6503E+03, 2.6533E+03, 2.6562E+03, 2.6591E+03, 2.6620E+03,
     +  2.6650E+03, 2.6679E+03, 2.6708E+03, 2.6738E+03, 2.6767E+03,
     + 2.6797E+03, 2.6826E+03, 2.6855E+03, 2.6885E+03, 2.6914E+03,
     +  2.6944E+03, 2.6973E+03, 2.7003E+03, 2.7032E+03, 2.7062E+03,
     + 2.7091E+03, 2.7121E+03, 2.7151E+03, 2.7180E+03, 2.7210E+03,
     +  2.7240E+03, 2.7269E+03, 2.7299E+03, 2.7329E+03, 2.7358E+03,
     + 2.7388E+03, 2.7418E+03, 2.7448E+03, 2.7477E+03, 2.7507E+03,
     +  2.7537E+03, 2.7567E+03, 2.7597E+03, 2.7627E+03, 2.7657E+03,
     + 2.7687E+03, 2.7716E+03, 2.7746E+03, 2.7776E+03, 2.7806E+03,
     +  2.7836E+03, 2.7866E+03, 2.7896E+03, 2.7927E+03, 2.7957E+03,
     + 2.7987E+03, 2.8017E+03, 2.8047E+03, 2.8077E+03, 2.8107E+03,
     +  2.8138E+03, 2.8168E+03, 2.8198E+03, 2.8228E+03, 2.8258E+03,
     + 2.8289E+03, 2.8319E+03, 2.8349E+03, 2.8380E+03, 2.8410E+03,
     +  2.8440E+03, 2.8471E+03, 2.8501E+03, 2.8532E+03, 2.8562E+03,
     + 2.8593E+03, 2.8623E+03, 2.8654E+03, 2.8684E+03, 2.8715E+03,
     +  2.8745E+03, 2.8776E+03, 2.8806E+03, 2.8837E+03, 2.8868E+03,
     + 2.8898E+03, 2.8929E+03, 2.8960E+03, 2.8990E+03, 2.9021E+03,
     +  2.9052E+03, 2.9082E+03, 2.9113E+03, 2.9144E+03, 2.9175E+03,
     + 2.9206E+03, 2.9237E+03, 2.9267E+03, 2.9298E+03, 2.9329E+03,
     +  2.9360E+03, 2.9391E+03, 2.9422E+03, 2.9453E+03, 2.9484E+03,
     + 2.9515E+03, 2.9546E+03, 2.9577E+03, 2.9608E+03, 2.9639E+03,
     +  2.9670E+03, 2.9701E+03, 2.9732E+03, 2.9764E+03, 2.9795E+03,
     + 2.9826E+03, 2.9857E+03, 2.9888E+03, 2.9920E+03, 2.9951E+03,
     +  2.9982E+03, 3.0013E+03, 3.0045E+03, 3.0076E+03, 3.0107E+03,
     + 3.0139E+03, 3.0170E+03, 3.0202E+03, 3.0233E+03, 3.0264E+03,
     +  3.0296E+03, 3.0327E+03, 3.0359E+03, 3.0390E+03, 3.0422E+03,
     + 3.0453E+03, 3.0485E+03, 3.0517E+03, 3.0548E+03, 3.0580E+03,
     +  3.0611E+03, 3.0643E+03, 3.0675E+03, 3.0707E+03, 3.0738E+03,
     + 3.0770E+03, 3.0802E+03, 3.0833E+03, 3.0865E+03, 3.0897E+03,
     +  3.0929E+03, 3.0961E+03, 3.0993E+03, 3.1024E+03, 3.1056E+03,
     + 3.1088E+03, 3.1120E+03, 3.1152E+03, 3.1184E+03, 3.1216E+03,
     +  3.1248E+03, 3.1280E+03, 3.1312E+03, 3.1344E+03, 3.1376E+03,
     + 3.1408E+03, 3.1441E+03, 3.1473E+03, 3.1505E+03, 3.1537E+03,
     +  3.1569E+03, 3.1601E+03, 3.1634E+03, 3.1666E+03, 3.1698E+03,
     + 3.1730E+03, 3.1763E+03, 3.1795E+03, 3.1827E+03, 3.1860E+03,
     +  3.1892E+03, 3.1924E+03, 3.1957E+03, 3.1989E+03, 3.2022E+03,
     + 3.2054E+03, 3.2087E+03, 3.2119E+03, 3.2152E+03, 3.2184E+03,
     +  3.2217E+03, 3.2249E+03, 3.2282E+03, 3.2315E+03, 3.2347E+03,
     + 3.2380E+03, 3.2412E+03, 3.2445E+03, 3.2478E+03, 3.2511E+03,
     +  3.2543E+03, 3.2576E+03, 3.2609E+03, 3.2642E+03, 3.2674E+03,
     + 3.2707E+03, 3.2740E+03, 3.2773E+03, 3.2806E+03, 3.2839E+03,
     +  3.2872E+03, 3.2905E+03, 3.2938E+03, 3.2971E+03, 3.3004E+03,
     + 3.3037E+03, 3.3070E+03, 3.3103E+03, 3.3136E+03, 3.3169E+03,
     +  3.3202E+03, 3.3235E+03, 3.3268E+03, 3.3301E+03, 3.3335E+03,
     + 3.3368E+03, 3.3401E+03, 3.3434E+03, 3.3467E+03, 3.3501E+03,
     +  3.3534E+03, 3.3567E+03, 3.3601E+03, 3.3634E+03, 3.3667E+03,
     + 3.3701E+03, 3.3734E+03, 3.3768E+03, 3.3801E+03, 3.3834E+03,
     +  3.3868E+03, 3.3901E+03, 3.3935E+03, 3.3968E+03, 3.4002E+03,
     + 3.4036E+03, 3.4069E+03, 3.4103E+03, 3.4136E+03, 3.4170E+03,
     +  3.4204E+03, 3.4237E+03, 3.4271E+03, 3.4305E+03, 3.4338E+03,
     + 3.4372E+03, 3.4406E+03, 3.4440E+03, 3.4474E+03, 3.4507E+03,
     +  3.4541E+03, 3.4575E+03, 3.4609E+03, 3.4643E+03, 3.4677E+03,
     + 3.4711E+03, 3.4745E+03, 3.4779E+03, 3.4813E+03, 3.4847E+03,
     +  3.4881E+03, 3.4915E+03, 3.4949E+03, 3.4983E+03, 3.5017E+03,
     + 3.5051E+03, 3.5085E+03, 3.5119E+03, 3.5154E+03, 3.5188E+03,
     +  3.5222E+03, 3.5256E+03, 3.5291E+03, 3.5325E+03, 3.5359E+03,
     + 3.5393E+03, 3.5428E+03, 3.5462E+03, 3.5496E+03, 3.5531E+03,
     +  3.5565E+03, 3.5600E+03, 3.5634E+03, 3.5669E+03, 3.5703E+03,
     + 3.5738E+03, 3.5772E+03, 3.5807E+03, 3.5841E+03, 3.5876E+03,
     +  3.5910E+03, 3.5945E+03, 3.5979E+03, 3.6014E+03, 3.6049E+03,
     + 3.6083E+03, 3.6118E+03, 3.6153E+03, 3.6188E+03, 3.6222E+03,
     +  3.6257E+03, 3.6292E+03, 3.6327E+03, 3.6362E+03, 3.6396E+03,
     + 3.6431E+03, 3.6466E+03, 3.6501E+03, 3.6536E+03, 3.6571E+03,
     +  3.6606E+03, 3.6641E+03, 3.6676E+03, 3.6711E+03, 3.6746E+03,
     + 3.6781E+03, 3.6816E+03, 3.6851E+03, 3.6886E+03, 3.6921E+03,
     +  3.6957E+03, 3.6992E+03, 3.7027E+03, 3.7062E+03, 3.7097E+03,
     + 3.7133E+03, 3.7168E+03, 3.7203E+03, 3.7238E+03, 3.7274E+03,
     +  3.7309E+03, 3.7344E+03, 3.7380E+03, 3.7415E+03, 3.7451E+03,
     + 3.7486E+03, 3.7521E+03, 3.7557E+03, 3.7592E+03, 3.7628E+03,
     +  3.7663E+03, 3.7699E+03, 3.7735E+03, 3.7770E+03, 3.7806E+03,
     + 3.7841E+03, 3.7877E+03, 3.7913E+03, 3.7948E+03, 3.7984E+03,
     +  3.8020E+03, 3.8055E+03, 3.8091E+03, 3.8127E+03, 3.8163E+03,
     + 3.8198E+03, 3.8234E+03, 3.8270E+03, 3.8306E+03, 3.8342E+03,
     +  3.8378E+03, 3.8414E+03, 3.8450E+03, 3.8486E+03, 3.8522E+03,
     + 3.8558E+03, 3.8594E+03, 3.8630E+03, 3.8666E+03, 3.8702E+03,
     +  3.8738E+03, 3.8774E+03, 3.8810E+03, 3.8846E+03, 3.8882E+03,
     + 3.8918E+03, 3.8955E+03, 3.8991E+03, 3.9027E+03, 3.9063E+03,
     +  3.9100E+03, 3.9136E+03, 3.9172E+03, 3.9209E+03, 3.9245E+03,
     + 3.9281E+03, 3.9318E+03, 3.9354E+03, 3.9391E+03, 3.9427E+03,
     +  3.9463E+03, 3.9500E+03, 3.9536E+03, 3.9573E+03, 3.9609E+03,
     + 3.9646E+03, 3.9683E+03, 3.9719E+03, 3.9756E+03, 3.9792E+03,
     +  3.9829E+03, 3.9866E+03, 3.9902E+03, 3.9939E+03, 3.9976E+03,
     + 4.0013E+03, 4.0049E+03, 4.0086E+03, 4.0123E+03, 4.0160E+03,
     +  4.0197E+03, 4.0233E+03, 4.0270E+03, 4.0307E+03, 4.0344E+03,
     + 4.0381E+03, 4.0418E+03, 4.0455E+03, 4.0492E+03, 4.0529E+03,
     +  4.0566E+03, 4.0603E+03, 4.0640E+03, 4.0677E+03, 4.0714E+03,
     + 4.0751E+03, 4.0789E+03, 4.0826E+03, 4.0863E+03, 4.0900E+03,
     +  4.0937E+03, 4.0975E+03, 4.1012E+03, 4.1049E+03, 4.1086E+03,
     + 4.1124E+03, 4.1161E+03, 4.1198E+03, 4.1236E+03, 4.1273E+03,
     +  4.1311E+03, 4.1348E+03, 4.1385E+03, 4.1423E+03, 4.1460E+03,
     + 4.1498E+03, 4.1535E+03, 4.1573E+03, 4.1611E+03, 4.1648E+03,
     +  4.1686E+03, 4.1723E+03, 4.1761E+03, 4.1799E+03, 4.1836E+03,
     + 4.1874E+03, 4.1912E+03, 4.1949E+03, 4.1987E+03, 4.2025E+03,
     +  4.2063E+03, 4.2101E+03, 4.2138E+03, 4.2176E+03, 4.2214E+03,
     + 4.2252E+03, 4.2290E+03, 4.2328E+03, 4.2366E+03, 4.2404E+03,
     +  4.2442E+03, 4.2480E+03, 4.2518E+03, 4.2556E+03, 4.2594E+03,
     + 4.2632E+03, 4.2670E+03, 4.2708E+03, 4.2746E+03, 4.2784E+03,
     +  4.2822E+03, 4.2861E+03, 4.2899E+03, 4.2937E+03, 4.2975E+03,
     + 4.3014E+03, 4.3052E+03, 4.3090E+03, 4.3128E+03, 4.3167E+03,
     +  4.3205E+03, 4.3244E+03, 4.3282E+03, 4.3320E+03, 4.3359E+03,
     + 4.3397E+03, 4.3436E+03, 4.3474E+03, 4.3513E+03, 4.3551E+03,
     +  4.3590E+03, 4.3628E+03, 4.3667E+03, 4.3706E+03, 4.3744E+03,
     + 4.3783E+03, 4.3821E+03, 4.3860E+03, 4.3899E+03, 4.3938E+03,
     +  4.3976E+03, 4.4015E+03, 4.4054E+03, 4.4093E+03, 4.4131E+03,
     + 4.4170E+03, 4.4209E+03, 4.4248E+03, 4.4287E+03, 4.4326E+03,
     +  4.4365E+03, 4.4404E+03, 4.4443E+03, 4.4482E+03, 4.4521E+03,
     + 4.4560E+03, 4.4599E+03, 4.4638E+03, 4.4677E+03, 4.4716E+03,
     +  4.4755E+03, 4.4794E+03, 4.4833E+03, 4.4873E+03, 4.4912E+03,
     + 4.4951E+03, 4.4990E+03, 4.5029E+03, 4.5069E+03, 4.5108E+03,
     +  4.5147E+03, 4.5187E+03, 4.5226E+03, 4.5265E+03, 4.5305E+03,
     + 4.5344E+03, 4.5384E+03, 4.5423E+03, 4.5463E+03, 4.5502E+03,
     +  4.5542E+03, 4.5581E+03, 4.5621E+03, 4.5660E+03, 4.5700E+03,
     + 4.5739E+03, 4.5779E+03, 4.5819E+03, 4.5858E+03, 4.5898E+03,
     +  4.5938E+03, 4.5977E+03, 4.6017E+03, 4.6057E+03, 4.6097E+03,
     + 4.6136E+03, 4.6176E+03, 4.6216E+03, 4.6256E+03, 4.6296E+03,
     +  4.6336E+03, 4.6376E+03, 4.6415E+03, 4.6455E+03, 4.6495E+03,
     + 4.6535E+03, 4.6575E+03, 4.6615E+03, 4.6655E+03, 4.6695E+03,
     +  4.6736E+03, 4.6776E+03, 4.6816E+03, 4.6856E+03, 4.6896E+03,
     + 4.6936E+03, 4.6976E+03, 4.7017E+03, 4.7057E+03, 4.7097E+03,
     +  4.7137E+03, 4.7178E+03, 4.7218E+03, 4.7258E+03, 4.7299E+03,
     + 4.7339E+03, 4.7379E+03, 4.7420E+03, 4.7460E+03, 4.7501E+03,
     +  4.7541E+03, 4.7582E+03, 4.7622E+03, 4.7663E+03, 4.7703E+03,
     + 4.7744E+03, 4.7784E+03, 4.7825E+03, 4.7866E+03, 4.7906E+03,
     +  4.7947E+03, 4.7988E+03, 4.8028E+03, 4.8069E+03, 4.8110E+03,
     + 4.8150E+03, 4.8191E+03, 4.8232E+03, 4.8273E+03, 4.8314E+03,
     +  4.8354E+03, 4.8395E+03, 4.8436E+03, 4.8477E+03, 4.8518E+03,
     + 4.8559E+03, 4.8600E+03, 4.8641E+03, 4.8682E+03, 4.8723E+03,
     +  4.8764E+03, 4.8805E+03, 4.8846E+03, 4.8887E+03, 4.8928E+03,
     + 4.8969E+03, 4.9011E+03, 4.9052E+03, 4.9093E+03, 4.9134E+03,
     +  4.9175E+03, 4.9217E+03, 4.9258E+03, 4.9299E+03, 4.9341E+03,
     + 4.9382E+03, 4.9423E+03, 4.9465E+03, 4.9506E+03, 4.9547E+03,
     +  4.9589E+03, 4.9630E+03, 4.9672E+03, 4.9713E+03, 4.9755E+03,
     + 4.9796E+03, 4.9838E+03, 4.9879E+03, 4.9921E+03, 4.9963E+03,
     +  5.0004E+03, 5.0046E+03, 5.0087E+03, 5.0129E+03, 5.0171E+03,
     + 5.0213E+03, 5.0254E+03, 5.0296E+03, 5.0338E+03, 5.0380E+03,
     +  5.0421E+03, 5.0463E+03, 5.0505E+03, 5.0547E+03, 5.0589E+03,
     + 5.0631E+03, 5.0673E+03, 5.0715E+03, 5.0757E+03, 5.0799E+03,
     +  5.0841E+03, 5.0883E+03, 5.0925E+03, 5.0967E+03, 5.1009E+03,
     + 5.1051E+03, 5.1093E+03, 5.1135E+03, 5.1177E+03, 5.1220E+03,
     +  5.1262E+03, 5.1304E+03, 5.1346E+03, 5.1388E+03, 5.1431E+03,
     + 5.1473E+03, 5.1515E+03, 5.1558E+03, 5.1600E+03, 5.1642E+03,
     +  5.1685E+03, 5.1727E+03, 5.1770E+03, 5.1812E+03, 5.1855E+03,
     + 5.1897E+03, 5.1940E+03, 5.1982E+03, 5.2025E+03, 5.2067E+03,
     +  5.2110E+03, 5.2152E+03, 5.2195E+03, 5.2238E+03, 5.2280E+03,
     + 5.2323E+03, 5.2366E+03, 5.2408E+03, 5.2451E+03, 5.2494E+03,
     +  5.2537E+03, 5.2579E+03, 5.2622E+03, 5.2665E+03, 5.2708E+03,
     + 5.2751E+03, 5.2794E+03, 5.2837E+03, 5.2880E+03, 5.2923E+03,
     +  5.2966E+03, 5.3009E+03, 5.3052E+03, 5.3095E+03, 5.3138E+03,
     + 5.3181E+03, 5.3224E+03, 5.3267E+03, 5.3310E+03, 5.3353E+03,
     +  5.3396E+03, 5.3440E+03, 5.3483E+03, 5.3526E+03, 5.3569E+03,
     + 5.3612E+03, 5.3656E+03, 5.3699E+03, 5.3742E+03, 5.3786E+03,
     +  5.3829E+03, 5.3872E+03, 5.3916E+03, 5.3959E+03, 5.4003E+03,
     + 5.4046E+03, 5.4090E+03, 5.4133E+03, 5.4177E+03, 5.4220E+03,
     +  5.4264E+03, 5.4307E+03, 5.4351E+03, 5.4395E+03, 5.4438E+03,
     + 5.4482E+03, 5.4526E+03, 5.4569E+03, 5.4613E+03, 5.4657E+03,
     +  5.4700E+03, 5.4744E+03, 5.4788E+03, 5.4832E+03, 5.4876E+03,
     + 5.4919E+03, 5.4963E+03, 5.5007E+03, 5.5051E+03, 5.5095E+03,
     +  5.5139E+03, 5.5183E+03, 5.5227E+03, 5.5271E+03, 5.5315E+03,
     + 5.5359E+03, 5.5403E+03, 5.5447E+03, 5.5491E+03, 5.5535E+03,
     +  5.5579E+03, 5.5624E+03, 5.5668E+03, 5.5712E+03, 5.5756E+03,
     + 5.5800E+03, 5.5845E+03, 5.5889E+03, 5.5933E+03, 5.5978E+03,
     +  5.6022E+03, 5.6066E+03, 5.6111E+03, 5.6155E+03, 5.6199E+03,
     + 5.6244E+03, 5.6288E+03, 5.6333E+03, 5.6377E+03, 5.6422E+03,
     +  5.6466E+03, 5.6511E+03, 5.6555E+03, 5.6600E+03, 5.6645E+03,
     + 5.6689E+03, 5.6734E+03, 5.6779E+03, 5.6823E+03, 5.6868E+03,
     +  5.6913E+03, 5.6957E+03, 5.7002E+03, 5.7047E+03, 5.7092E+03,
     + 5.7137E+03, 5.7181E+03, 5.7226E+03, 5.7271E+03, 5.7316E+03,
     +  5.7361E+03, 5.7406E+03, 5.7451E+03, 5.7496E+03, 5.7541E+03,
     + 5.7586E+03, 5.7631E+03, 5.7676E+03, 5.7721E+03, 5.7766E+03,
     +  5.7811E+03, 5.7856E+03, 5.7901E+03, 5.7947E+03, 5.7992E+03,
     + 5.8037E+03, 5.8082E+03, 5.8127E+03, 5.8173E+03, 5.8218E+03,
     +  5.8263E+03, 5.8309E+03, 5.8354E+03, 5.8399E+03, 5.8445E+03,
     + 5.8490E+03, 5.8536E+03, 5.8581E+03, 5.8626E+03, 5.8672E+03,
     +  5.8717E+03, 5.8763E+03, 5.8808E+03, 5.8854E+03, 5.8900E+03,
     + 5.8945E+03, 5.8991E+03, 5.9036E+03, 5.9082E+03, 5.9128E+03,
     +  5.9173E+03, 5.9219E+03, 5.9265E+03, 5.9311E+03, 5.9356E+03,
     + 5.9402E+03, 5.9448E+03, 5.9494E+03, 5.9540E+03, 5.9586E+03,
     +  5.9631E+03, 5.9677E+03, 5.9723E+03, 5.9769E+03, 5.9815E+03,
     + 5.9861E+03, 5.9907E+03, 5.9953E+03, 5.9999E+03, 6.0045E+03,
     +  6.0091E+03, 6.0138E+03, 6.0184E+03, 6.0230E+03, 6.0276E+03,
     + 6.0322E+03, 6.0368E+03, 6.0415E+03, 6.0461E+03, 6.0507E+03,
     +  6.0553E+03, 6.0600E+03, 6.0646E+03, 6.0692E+03, 6.0739E+03,
     + 6.0785E+03, 6.0831E+03, 6.0878E+03, 6.0924E+03, 6.0971E+03,
     +  6.1017E+03, 6.1064E+03, 6.1110E+03, 6.1157E+03, 6.1203E+03,
     + 6.1250E+03, 6.1296E+03, 6.1343E+03, 6.1390E+03, 6.1436E+03,
     +  6.1483E+03, 6.1530E+03, 6.1576E+03, 6.1623E+03, 6.1670E+03,
     + 6.1717E+03, 6.1763E+03, 6.1810E+03, 6.1857E+03, 6.1904E+03,
     +  6.1951E+03, 6.1998E+03, 6.2045E+03, 6.2091E+03, 6.2138E+03,
     + 6.2185E+03, 6.2232E+03, 6.2279E+03, 6.2326E+03, 6.2373E+03,
     +  6.2420E+03, 6.2468E+03, 6.2515E+03, 6.2562E+03, 6.2609E+03,
     + 6.2656E+03, 6.2703E+03, 6.2750E+03, 6.2798E+03, 6.2845E+03,
     +  6.2892E+03, 6.2939E+03, 6.2987E+03, 6.3034E+03, 6.3081E+03,
     + 6.3129E+03, 6.3176E+03, 6.3223E+03, 6.3271E+03, 6.3318E+03,
     +  6.3366E+03, 6.3413E+03, 6.3461E+03, 6.3508E+03, 6.3556E+03,
     + 6.3603E+03, 6.3651E+03, 6.3698E+03, 6.3746E+03, 6.3794E+03,
     +  6.3841E+03, 6.3889E+03, 6.3937E+03, 6.3984E+03, 6.4032E+03,
     + 6.4080E+03, 6.4127E+03, 6.4175E+03, 6.4223E+03, 6.4271E+03,
     +  6.4319E+03, 6.4367E+03, 6.4414E+03, 6.4462E+03, 6.4510E+03,
     + 6.4558E+03, 6.4606E+03, 6.4654E+03, 6.4702E+03, 6.4750E+03,
     +  6.4798E+03, 6.4846E+03, 6.4894E+03, 6.4942E+03, 6.4990E+03/
      data (QofT( 2,J),J=2409,4510)/ 6.5039E+03, 6.5087E+03,
     +  6.5135E+03, 6.5183E+03, 6.5231E+03, 6.5280E+03, 6.5328E+03,
     +  6.5376E+03,
     + 6.5424E+03, 6.5473E+03, 6.5521E+03, 6.5569E+03, 6.5618E+03,
     +  6.5666E+03, 6.5714E+03, 6.5763E+03, 6.5811E+03, 6.5860E+03,
     + 6.5908E+03, 6.5957E+03, 6.6005E+03, 6.6054E+03, 6.6102E+03,
     +  6.6151E+03, 6.6200E+03, 6.6248E+03, 6.6297E+03, 6.6345E+03,
     + 6.6394E+03, 6.6443E+03, 6.6492E+03, 6.6540E+03, 6.6589E+03,
     +  6.6638E+03, 6.6687E+03, 6.6735E+03, 6.6784E+03, 6.6833E+03,
     + 6.6882E+03, 6.6931E+03, 6.6980E+03, 6.7029E+03, 6.7078E+03,
     +  6.7127E+03, 6.7176E+03, 6.7225E+03, 6.7274E+03, 6.7323E+03,
     + 6.7372E+03, 6.7421E+03, 6.7470E+03, 6.7519E+03, 6.7568E+03,
     +  6.7617E+03, 6.7667E+03, 6.7716E+03, 6.7765E+03, 6.7814E+03,
     + 6.7863E+03, 6.7913E+03, 6.7962E+03, 6.8011E+03, 6.8061E+03,
     +  6.8110E+03, 6.8159E+03, 6.8209E+03, 6.8258E+03, 6.8308E+03,
     + 6.8357E+03, 6.8407E+03, 6.8456E+03, 6.8506E+03, 6.8555E+03,
     +  6.8605E+03, 6.8654E+03, 6.8704E+03, 6.8753E+03, 6.8803E+03,
     + 6.8853E+03, 6.8902E+03, 6.8952E+03, 6.9002E+03, 6.9052E+03,
     +  6.9101E+03, 6.9151E+03, 6.9201E+03, 6.9251E+03, 6.9300E+03,
     + 6.9350E+03, 6.9400E+03, 6.9450E+03, 6.9500E+03, 6.9550E+03,
     +  6.9600E+03, 6.9650E+03, 6.9700E+03, 6.9750E+03, 6.9800E+03,
     + 6.9850E+03, 6.9900E+03, 6.9950E+03, 7.0000E+03, 7.0050E+03,
     +  7.0100E+03, 7.0150E+03, 7.0201E+03, 7.0251E+03, 7.0301E+03,
     + 7.0351E+03, 7.0402E+03, 7.0452E+03, 7.0502E+03, 7.0552E+03,
     +  7.0603E+03, 7.0653E+03, 7.0703E+03, 7.0754E+03, 7.0804E+03,
     + 7.0855E+03, 7.0905E+03, 7.0956E+03, 7.1006E+03, 7.1057E+03,
     +  7.1107E+03, 7.1158E+03, 7.1208E+03, 7.1259E+03, 7.1309E+03,
     + 7.1360E+03, 7.1411E+03, 7.1461E+03, 7.1512E+03, 7.1563E+03,
     +  7.1613E+03, 7.1664E+03, 7.1715E+03, 7.1766E+03, 7.1817E+03,
     + 7.1867E+03, 7.1918E+03, 7.1969E+03, 7.2020E+03, 7.2071E+03,
     +  7.2122E+03, 7.2173E+03, 7.2224E+03, 7.2275E+03, 7.2326E+03,
     + 7.2377E+03, 7.2428E+03, 7.2479E+03, 7.2530E+03, 7.2581E+03,
     +  7.2632E+03, 7.2683E+03, 7.2734E+03, 7.2785E+03, 7.2837E+03,
     + 7.2888E+03, 7.2939E+03, 7.2990E+03, 7.3041E+03, 7.3093E+03,
     +  7.3144E+03, 7.3195E+03, 7.3247E+03, 7.3298E+03, 7.3349E+03,
     + 7.3401E+03, 7.3452E+03, 7.3504E+03, 7.3555E+03, 7.3607E+03,
     +  7.3658E+03, 7.3710E+03, 7.3761E+03, 7.3813E+03, 7.3864E+03,
     + 7.3916E+03, 7.3968E+03, 7.4019E+03, 7.4071E+03, 7.4122E+03,
     +  7.4174E+03, 7.4226E+03, 7.4278E+03, 7.4329E+03, 7.4381E+03,
     + 7.4433E+03, 7.4485E+03, 7.4537E+03, 7.4588E+03, 7.4640E+03,
     +  7.4692E+03, 7.4744E+03, 7.4796E+03, 7.4848E+03, 7.4900E+03,
     + 7.4952E+03, 7.5004E+03, 7.5056E+03, 7.5108E+03, 7.5160E+03,
     +  7.5212E+03, 7.5264E+03, 7.5316E+03, 7.5368E+03, 7.5421E+03,
     + 7.5473E+03, 7.5525E+03, 7.5577E+03, 7.5629E+03, 7.5682E+03,
     +  7.5734E+03, 7.5786E+03, 7.5839E+03, 7.5891E+03, 7.5943E+03,
     + 7.5996E+03, 7.6048E+03, 7.6100E+03, 7.6153E+03, 7.6205E+03,
     +  7.6258E+03, 7.6310E+03, 7.6363E+03, 7.6415E+03, 7.6468E+03,
     + 7.6520E+03, 7.6573E+03, 7.6626E+03, 7.6678E+03, 7.6731E+03,
     +  7.6783E+03, 7.6836E+03, 7.6889E+03, 7.6942E+03, 7.6994E+03,
     + 7.7047E+03, 7.7100E+03, 7.7153E+03, 7.7205E+03, 7.7258E+03,
     +  7.7311E+03, 7.7364E+03, 7.7417E+03, 7.7470E+03, 7.7523E+03,
     + 7.7576E+03, 7.7629E+03, 7.7682E+03, 7.7735E+03, 7.7788E+03,
     +  7.7841E+03, 7.7894E+03, 7.7947E+03, 7.8000E+03, 7.8053E+03,
     + 7.8106E+03, 7.8159E+03, 7.8213E+03, 7.8266E+03, 7.8319E+03,
     +  7.8372E+03, 7.8426E+03, 7.8479E+03, 7.8532E+03, 7.8585E+03,
     + 7.8639E+03, 7.8692E+03, 7.8745E+03, 7.8799E+03, 7.8852E+03,
     +  7.8906E+03, 7.8959E+03, 7.9013E+03, 7.9066E+03, 7.9120E+03,
     + 7.9173E+03, 7.9227E+03, 7.9280E+03, 7.9334E+03, 7.9388E+03,
     +  7.9441E+03, 7.9495E+03, 7.9548E+03, 7.9602E+03, 7.9656E+03,
     + 7.9710E+03, 7.9763E+03, 7.9817E+03, 7.9871E+03, 7.9925E+03,
     +  7.9978E+03, 8.0032E+03, 8.0086E+03, 8.0140E+03, 8.0194E+03,
     + 8.0248E+03, 8.0302E+03, 8.0356E+03, 8.0410E+03, 8.0464E+03,
     +  8.0518E+03, 8.0572E+03, 8.0626E+03, 8.0680E+03, 8.0734E+03,
     + 8.0788E+03, 8.0842E+03, 8.0896E+03, 8.0951E+03, 8.1005E+03,
     +  8.1059E+03, 8.1113E+03, 8.1167E+03, 8.1222E+03, 8.1276E+03,
     + 8.1330E+03, 8.1385E+03, 8.1439E+03, 8.1493E+03, 8.1548E+03,
     +  8.1602E+03, 8.1656E+03, 8.1711E+03, 8.1765E+03, 8.1820E+03,
     + 8.1874E+03, 8.1929E+03, 8.1983E+03, 8.2038E+03, 8.2092E+03,
     +  8.2147E+03, 8.2202E+03, 8.2256E+03, 8.2311E+03, 8.2366E+03,
     + 8.2420E+03, 8.2475E+03, 8.2530E+03, 8.2585E+03, 8.2639E+03,
     +  8.2694E+03, 8.2749E+03, 8.2804E+03, 8.2859E+03, 8.2913E+03,
     + 8.2968E+03, 8.3023E+03, 8.3078E+03, 8.3133E+03, 8.3188E+03,
     +  8.3243E+03, 8.3298E+03, 8.3353E+03, 8.3408E+03, 8.3463E+03,
     + 8.3518E+03, 8.3573E+03, 8.3628E+03, 8.3683E+03, 8.3739E+03,
     +  8.3794E+03, 8.3849E+03, 8.3904E+03, 8.3959E+03, 8.4015E+03,
     + 8.4070E+03, 8.4125E+03, 8.4180E+03, 8.4236E+03, 8.4291E+03,
     +  8.4347E+03, 8.4402E+03, 8.4457E+03, 8.4513E+03, 8.4568E+03,
     + 8.4624E+03, 8.4679E+03, 8.4735E+03, 8.4790E+03, 8.4846E+03,
     +  8.4901E+03, 8.4957E+03, 8.5012E+03, 8.5068E+03, 8.5124E+03,
     + 8.5179E+03, 8.5235E+03, 8.5291E+03, 8.5346E+03, 8.5402E+03,
     +  8.5458E+03, 8.5514E+03, 8.5569E+03, 8.5625E+03, 8.5681E+03,
     + 8.5737E+03, 8.5793E+03, 8.5848E+03, 8.5904E+03, 8.5960E+03,
     +  8.6016E+03, 8.6072E+03, 8.6128E+03, 8.6184E+03, 8.6240E+03,
     + 8.6296E+03, 8.6352E+03, 8.6408E+03, 8.6464E+03, 8.6521E+03,
     +  8.6577E+03, 8.6633E+03, 8.6689E+03, 8.6745E+03, 8.6801E+03,
     + 8.6858E+03, 8.6914E+03, 8.6970E+03, 8.7026E+03, 8.7083E+03,
     +  8.7139E+03, 8.7195E+03, 8.7252E+03, 8.7308E+03, 8.7364E+03,
     + 8.7421E+03, 8.7477E+03, 8.7534E+03, 8.7590E+03, 8.7647E+03,
     +  8.7703E+03, 8.7760E+03, 8.7816E+03, 8.7873E+03, 8.7929E+03,
     + 8.7986E+03, 8.8043E+03, 8.8099E+03, 8.8156E+03, 8.8213E+03,
     +  8.8269E+03, 8.8326E+03, 8.8383E+03, 8.8440E+03, 8.8496E+03,
     + 8.8553E+03, 8.8610E+03, 8.8667E+03, 8.8724E+03, 8.8780E+03,
     +  8.8837E+03, 8.8894E+03, 8.8951E+03, 8.9008E+03, 8.9065E+03,
     + 8.9122E+03, 8.9179E+03, 8.9236E+03, 8.9293E+03, 8.9350E+03,
     +  8.9407E+03, 8.9464E+03, 8.9522E+03, 8.9579E+03, 8.9636E+03,
     + 8.9693E+03, 8.9750E+03, 8.9807E+03, 8.9865E+03, 8.9922E+03,
     +  8.9979E+03, 9.0036E+03, 9.0094E+03, 9.0151E+03, 9.0208E+03,
     + 9.0266E+03, 9.0323E+03, 9.0381E+03, 9.0438E+03, 9.0495E+03,
     +  9.0553E+03, 9.0610E+03, 9.0668E+03, 9.0725E+03, 9.0783E+03,
     + 9.0841E+03, 9.0898E+03, 9.0956E+03, 9.1013E+03, 9.1071E+03,
     +  9.1129E+03, 9.1186E+03, 9.1244E+03, 9.1302E+03, 9.1359E+03,
     + 9.1417E+03, 9.1475E+03, 9.1533E+03, 9.1590E+03, 9.1648E+03,
     +  9.1706E+03, 9.1764E+03, 9.1822E+03, 9.1880E+03, 9.1938E+03,
     + 9.1996E+03, 9.2054E+03, 9.2112E+03, 9.2170E+03, 9.2228E+03,
     +  9.2286E+03, 9.2344E+03, 9.2402E+03, 9.2460E+03, 9.2518E+03,
     + 9.2576E+03, 9.2634E+03, 9.2692E+03, 9.2751E+03, 9.2809E+03,
     +  9.2867E+03, 9.2925E+03, 9.2983E+03, 9.3042E+03, 9.3100E+03,
     + 9.3158E+03, 9.3217E+03, 9.3275E+03, 9.3333E+03, 9.3392E+03,
     +  9.3450E+03, 9.3509E+03, 9.3567E+03, 9.3626E+03, 9.3684E+03,
     + 9.3743E+03, 9.3801E+03, 9.3860E+03, 9.3918E+03, 9.3977E+03,
     +  9.4035E+03, 9.4094E+03, 9.4153E+03, 9.4211E+03, 9.4270E+03,
     + 9.4329E+03, 9.4387E+03, 9.4446E+03, 9.4505E+03, 9.4564E+03,
     +  9.4622E+03, 9.4681E+03, 9.4740E+03, 9.4799E+03, 9.4858E+03,
     + 9.4917E+03, 9.4975E+03, 9.5034E+03, 9.5093E+03, 9.5152E+03,
     +  9.5211E+03, 9.5270E+03, 9.5329E+03, 9.5388E+03, 9.5447E+03,
     + 9.5506E+03, 9.5566E+03, 9.5625E+03, 9.5684E+03, 9.5743E+03,
     +  9.5802E+03, 9.5861E+03, 9.5920E+03, 9.5980E+03, 9.6039E+03,
     + 9.6098E+03, 9.6157E+03, 9.6217E+03, 9.6276E+03, 9.6335E+03,
     +  9.6395E+03, 9.6454E+03, 9.6514E+03, 9.6573E+03, 9.6632E+03,
     + 9.6692E+03, 9.6751E+03, 9.6811E+03, 9.6870E+03, 9.6930E+03,
     +  9.6989E+03, 9.7049E+03, 9.7108E+03, 9.7168E+03, 9.7228E+03,
     + 9.7287E+03, 9.7347E+03, 9.7407E+03, 9.7466E+03, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        29
      data (QofT( 3,J),J=1,2408)/ 1.2000E+01, 1.2080E+01, 1.2612E+01,
     +  1.3701E+01, 1.5162E+01, 1.6827E+01, 1.8600E+01, 2.0433E+01,
     + 2.2301E+01, 2.4192E+01, 2.6098E+01, 2.8015E+01, 2.9940E+01,
     +  3.1871E+01, 3.3807E+01, 3.5746E+01, 3.7689E+01, 3.9634E+01,
     + 4.1582E+01, 4.3531E+01, 4.7434E+01, 5.1341E+01, 5.5251E+01,
     +  5.9165E+01, 6.3080E+01, 6.6997E+01, 7.0916E+01, 7.4836E+01,
     + 7.8757E+01, 8.2679E+01, 8.6601E+01, 9.0525E+01, 9.4449E+01,
     +  9.8373E+01, 1.0230E+02, 1.0622E+02, 1.1015E+02, 1.1408E+02,
     + 1.1800E+02, 1.2193E+02, 1.2586E+02, 1.2978E+02, 1.3371E+02,
     +  1.3764E+02, 1.4157E+02, 1.4549E+02, 1.4942E+02, 1.5335E+02,
     + 1.5728E+02, 1.6121E+02, 1.6514E+02, 1.6907E+02, 1.7300E+02,
     +  1.7693E+02, 1.8086E+02, 1.8479E+02, 1.8872E+02, 1.9265E+02,
     + 1.9658E+02, 2.0051E+02, 2.0444E+02, 2.0837E+02, 2.1231E+02,
     +  2.1624E+02, 2.2017E+02, 2.2410E+02, 2.2803E+02, 2.3196E+02,
     + 2.3590E+02, 2.3983E+02, 2.4376E+02, 2.4769E+02, 2.5163E+02,
     +  2.5556E+02, 2.5949E+02, 2.6342E+02, 2.6736E+02, 2.7129E+02,
     + 2.7522E+02, 2.7916E+02, 2.8309E+02, 2.8703E+02, 2.9096E+02,
     +  2.9489E+02, 2.9883E+02, 3.0276E+02, 3.0670E+02, 3.1063E+02,
     + 3.1457E+02, 3.1850E+02, 3.2243E+02, 3.2637E+02, 3.3030E+02,
     +  3.3424E+02, 3.3818E+02, 3.4211E+02, 3.4605E+02, 3.4998E+02,
     + 3.5392E+02, 3.5785E+02, 3.6179E+02, 3.6573E+02, 3.6966E+02,
     +  3.7360E+02, 3.7753E+02, 3.8147E+02, 3.8541E+02, 3.8935E+02,
     + 3.9328E+02, 3.9722E+02, 4.0116E+02, 4.0509E+02, 4.0903E+02,
     +  4.1297E+02, 4.1691E+02, 4.2084E+02, 4.2478E+02, 4.2872E+02,
     + 4.3266E+02, 4.3660E+02, 4.4054E+02, 4.4447E+02, 4.4841E+02,
     +  4.5235E+02, 4.5629E+02, 4.6023E+02, 4.6417E+02, 4.6811E+02,
     + 4.7205E+02, 4.7599E+02, 4.7993E+02, 4.8387E+02, 4.8781E+02,
     +  4.9175E+02, 4.9569E+02, 4.9963E+02, 5.0358E+02, 5.0752E+02,
     + 5.1146E+02, 5.1540E+02, 5.1934E+02, 5.2329E+02, 5.2723E+02,
     +  5.3117E+02, 5.3511E+02, 5.3906E+02, 5.4300E+02, 5.4695E+02,
     + 5.5089E+02, 5.5484E+02, 5.5878E+02, 5.6273E+02, 5.6667E+02,
     +  5.7062E+02, 5.7456E+02, 5.7851E+02, 5.8246E+02, 5.8640E+02,
     + 5.9035E+02, 5.9430E+02, 5.9825E+02, 6.0220E+02, 6.0615E+02,
     +  6.1010E+02, 6.1405E+02, 6.1800E+02, 6.2195E+02, 6.2590E+02,
     + 6.2986E+02, 6.3381E+02, 6.3776E+02, 6.4172E+02, 6.4567E+02,
     +  6.4963E+02, 6.5358E+02, 6.5754E+02, 6.6150E+02, 6.6545E+02,
     + 6.6941E+02, 6.7337E+02, 6.7733E+02, 6.8129E+02, 6.8525E+02,
     +  6.8922E+02, 6.9318E+02, 6.9714E+02, 7.0111E+02, 7.0507E+02,
     + 7.0904E+02, 7.1301E+02, 7.1697E+02, 7.2094E+02, 7.2491E+02,
     +  7.2888E+02, 7.3285E+02, 7.3683E+02, 7.4080E+02, 7.4478E+02,
     + 7.4875E+02, 7.5273E+02, 7.5671E+02, 7.6069E+02, 7.6467E+02,
     +  7.6865E+02, 7.7263E+02, 7.7661E+02, 7.8060E+02, 7.8458E+02,
     + 7.8857E+02, 7.9256E+02, 7.9655E+02, 8.0054E+02, 8.0453E+02,
     +  8.0853E+02, 8.1252E+02, 8.1652E+02, 8.2052E+02, 8.2452E+02,
     + 8.2852E+02, 8.3252E+02, 8.3653E+02, 8.4053E+02, 8.4454E+02,
     +  8.4855E+02, 8.5256E+02, 8.5657E+02, 8.6058E+02, 8.6460E+02,
     + 8.6862E+02, 8.7263E+02, 8.7666E+02, 8.8068E+02, 8.8470E+02,
     +  8.8873E+02, 8.9276E+02, 8.9678E+02, 9.0082E+02, 9.0485E+02,
     + 9.0889E+02, 9.1292E+02, 9.1696E+02, 9.2100E+02, 9.2505E+02,
     +  9.2909E+02, 9.3314E+02, 9.3719E+02, 9.4124E+02, 9.4529E+02,
     + 9.4935E+02, 9.5341E+02, 9.5747E+02, 9.6153E+02, 9.6560E+02,
     +  9.6966E+02, 9.7373E+02, 9.7780E+02, 9.8188E+02, 9.8595E+02,
     + 9.9003E+02, 9.9411E+02, 9.9820E+02, 1.0023E+03, 1.0064E+03,
     +  1.0105E+03, 1.0146E+03, 1.0187E+03, 1.0228E+03, 1.0269E+03,
     + 1.0310E+03, 1.0351E+03, 1.0392E+03, 1.0433E+03, 1.0474E+03,
     +  1.0515E+03, 1.0556E+03, 1.0598E+03, 1.0639E+03, 1.0680E+03,
     + 1.0721E+03, 1.0763E+03, 1.0804E+03, 1.0846E+03, 1.0887E+03,
     +  1.0928E+03, 1.0970E+03, 1.1011E+03, 1.1053E+03, 1.1095E+03,
     + 1.1136E+03, 1.1178E+03, 1.1219E+03, 1.1261E+03, 1.1303E+03,
     +  1.1345E+03, 1.1386E+03, 1.1428E+03, 1.1470E+03, 1.1512E+03,
     + 1.1554E+03, 1.1596E+03, 1.1638E+03, 1.1680E+03, 1.1722E+03,
     +  1.1764E+03, 1.1806E+03, 1.1848E+03, 1.1891E+03, 1.1933E+03,
     + 1.1975E+03, 1.2017E+03, 1.2060E+03, 1.2102E+03, 1.2144E+03,
     +  1.2187E+03, 1.2229E+03, 1.2272E+03, 1.2314E+03, 1.2357E+03,
     + 1.2399E+03, 1.2442E+03, 1.2485E+03, 1.2528E+03, 1.2570E+03,
     +  1.2613E+03, 1.2656E+03, 1.2699E+03, 1.2742E+03, 1.2785E+03,
     + 1.2828E+03, 1.2871E+03, 1.2914E+03, 1.2957E+03, 1.3000E+03,
     +  1.3043E+03, 1.3086E+03, 1.3129E+03, 1.3173E+03, 1.3216E+03,
     + 1.3259E+03, 1.3303E+03, 1.3346E+03, 1.3390E+03, 1.3433E+03,
     +  1.3477E+03, 1.3521E+03, 1.3564E+03, 1.3608E+03, 1.3652E+03,
     + 1.3695E+03, 1.3739E+03, 1.3783E+03, 1.3827E+03, 1.3871E+03,
     +  1.3915E+03, 1.3959E+03, 1.4003E+03, 1.4047E+03, 1.4091E+03,
     + 1.4135E+03, 1.4180E+03, 1.4224E+03, 1.4268E+03, 1.4313E+03,
     +  1.4357E+03, 1.4401E+03, 1.4446E+03, 1.4490E+03, 1.4535E+03,
     + 1.4580E+03, 1.4624E+03, 1.4669E+03, 1.4714E+03, 1.4759E+03,
     +  1.4803E+03, 1.4848E+03, 1.4893E+03, 1.4938E+03, 1.4983E+03,
     + 1.5028E+03, 1.5073E+03, 1.5118E+03, 1.5164E+03, 1.5209E+03,
     +  1.5254E+03, 1.5300E+03, 1.5345E+03, 1.5390E+03, 1.5436E+03,
     + 1.5481E+03, 1.5527E+03, 1.5573E+03, 1.5618E+03, 1.5664E+03,
     +  1.5710E+03, 1.5756E+03, 1.5801E+03, 1.5847E+03, 1.5893E+03,
     + 1.5939E+03, 1.5985E+03, 1.6031E+03, 1.6077E+03, 1.6124E+03,
     +  1.6170E+03, 1.6216E+03, 1.6262E+03, 1.6309E+03, 1.6355E+03,
     + 1.6402E+03, 1.6448E+03, 1.6495E+03, 1.6541E+03, 1.6588E+03,
     +  1.6635E+03, 1.6682E+03, 1.6728E+03, 1.6775E+03, 1.6822E+03,
     + 1.6869E+03, 1.6916E+03, 1.6963E+03, 1.7010E+03, 1.7058E+03,
     +  1.7105E+03, 1.7152E+03, 1.7199E+03, 1.7247E+03, 1.7294E+03,
     + 1.7342E+03, 1.7389E+03, 1.7437E+03, 1.7484E+03, 1.7532E+03,
     +  1.7580E+03, 1.7627E+03, 1.7675E+03, 1.7723E+03, 1.7771E+03,
     + 1.7819E+03, 1.7867E+03, 1.7915E+03, 1.7963E+03, 1.8011E+03,
     +  1.8060E+03, 1.8108E+03, 1.8156E+03, 1.8205E+03, 1.8253E+03,
     + 1.8302E+03, 1.8350E+03, 1.8399E+03, 1.8447E+03, 1.8496E+03,
     +  1.8545E+03, 1.8594E+03, 1.8642E+03, 1.8691E+03, 1.8740E+03,
     + 1.8789E+03, 1.8838E+03, 1.8888E+03, 1.8937E+03, 1.8986E+03,
     +  1.9035E+03, 1.9085E+03, 1.9134E+03, 1.9183E+03, 1.9233E+03,
     + 1.9282E+03, 1.9332E+03, 1.9382E+03, 1.9431E+03, 1.9481E+03,
     +  1.9531E+03, 1.9581E+03, 1.9631E+03, 1.9681E+03, 1.9731E+03,
     + 1.9781E+03, 1.9831E+03, 1.9881E+03, 1.9931E+03, 1.9982E+03,
     +  2.0032E+03, 2.0083E+03, 2.0133E+03, 2.0184E+03, 2.0234E+03,
     + 2.0285E+03, 2.0335E+03, 2.0386E+03, 2.0437E+03, 2.0488E+03,
     +  2.0539E+03, 2.0590E+03, 2.0641E+03, 2.0692E+03, 2.0743E+03,
     + 2.0794E+03, 2.0845E+03, 2.0897E+03, 2.0948E+03, 2.0999E+03,
     +  2.1051E+03, 2.1102E+03, 2.1154E+03, 2.1206E+03, 2.1257E+03,
     + 2.1309E+03, 2.1361E+03, 2.1413E+03, 2.1465E+03, 2.1517E+03,
     +  2.1569E+03, 2.1621E+03, 2.1673E+03, 2.1725E+03, 2.1777E+03,
     + 2.1830E+03, 2.1882E+03, 2.1934E+03, 2.1987E+03, 2.2039E+03,
     +  2.2092E+03, 2.2145E+03, 2.2197E+03, 2.2250E+03, 2.2303E+03,
     + 2.2356E+03, 2.2409E+03, 2.2462E+03, 2.2515E+03, 2.2568E+03,
     +  2.2621E+03, 2.2674E+03, 2.2728E+03, 2.2781E+03, 2.2834E+03,
     + 2.2888E+03, 2.2941E+03, 2.2995E+03, 2.3048E+03, 2.3102E+03,
     +  2.3156E+03, 2.3210E+03, 2.3264E+03, 2.3317E+03, 2.3371E+03,
     + 2.3425E+03, 2.3480E+03, 2.3534E+03, 2.3588E+03, 2.3642E+03,
     +  2.3696E+03, 2.3751E+03, 2.3805E+03, 2.3860E+03, 2.3914E+03,
     + 2.3969E+03, 2.4024E+03, 2.4078E+03, 2.4133E+03, 2.4188E+03,
     +  2.4243E+03, 2.4298E+03, 2.4353E+03, 2.4408E+03, 2.4463E+03,
     + 2.4518E+03, 2.4574E+03, 2.4629E+03, 2.4684E+03, 2.4740E+03,
     +  2.4795E+03, 2.4851E+03, 2.4906E+03, 2.4962E+03, 2.5018E+03,
     + 2.5074E+03, 2.5129E+03, 2.5185E+03, 2.5241E+03, 2.5297E+03,
     +  2.5353E+03, 2.5410E+03, 2.5466E+03, 2.5522E+03, 2.5578E+03,
     + 2.5635E+03, 2.5691E+03, 2.5748E+03, 2.5804E+03, 2.5861E+03,
     +  2.5918E+03, 2.5974E+03, 2.6031E+03, 2.6088E+03, 2.6145E+03,
     + 2.6202E+03, 2.6259E+03, 2.6316E+03, 2.6373E+03, 2.6431E+03,
     +  2.6488E+03, 2.6545E+03, 2.6603E+03, 2.6660E+03, 2.6718E+03,
     + 2.6775E+03, 2.6833E+03, 2.6891E+03, 2.6948E+03, 2.7006E+03,
     +  2.7064E+03, 2.7122E+03, 2.7180E+03, 2.7238E+03, 2.7296E+03,
     + 2.7354E+03, 2.7413E+03, 2.7471E+03, 2.7529E+03, 2.7588E+03,
     +  2.7646E+03, 2.7705E+03, 2.7763E+03, 2.7822E+03, 2.7881E+03,
     + 2.7940E+03, 2.7999E+03, 2.8057E+03, 2.8116E+03, 2.8176E+03,
     +  2.8235E+03, 2.8294E+03, 2.8353E+03, 2.8412E+03, 2.8472E+03,
     + 2.8531E+03, 2.8591E+03, 2.8650E+03, 2.8710E+03, 2.8769E+03,
     +  2.8829E+03, 2.8889E+03, 2.8949E+03, 2.9009E+03, 2.9069E+03,
     + 2.9129E+03, 2.9189E+03, 2.9249E+03, 2.9309E+03, 2.9369E+03,
     +  2.9430E+03, 2.9490E+03, 2.9551E+03, 2.9611E+03, 2.9672E+03,
     + 2.9732E+03, 2.9793E+03, 2.9854E+03, 2.9915E+03, 2.9976E+03,
     +  3.0036E+03, 3.0097E+03, 3.0159E+03, 3.0220E+03, 3.0281E+03,
     + 3.0342E+03, 3.0404E+03, 3.0465E+03, 3.0526E+03, 3.0588E+03,
     +  3.0649E+03, 3.0711E+03, 3.0773E+03, 3.0835E+03, 3.0896E+03,
     + 3.0958E+03, 3.1020E+03, 3.1082E+03, 3.1144E+03, 3.1206E+03,
     +  3.1269E+03, 3.1331E+03, 3.1393E+03, 3.1456E+03, 3.1518E+03,
     + 3.1581E+03, 3.1643E+03, 3.1706E+03, 3.1769E+03, 3.1831E+03,
     +  3.1894E+03, 3.1957E+03, 3.2020E+03, 3.2083E+03, 3.2146E+03,
     + 3.2209E+03, 3.2272E+03, 3.2336E+03, 3.2399E+03, 3.2462E+03,
     +  3.2526E+03, 3.2589E+03, 3.2653E+03, 3.2717E+03, 3.2780E+03,
     + 3.2844E+03, 3.2908E+03, 3.2972E+03, 3.3036E+03, 3.3100E+03,
     +  3.3164E+03, 3.3228E+03, 3.3292E+03, 3.3357E+03, 3.3421E+03,
     + 3.3485E+03, 3.3550E+03, 3.3614E+03, 3.3679E+03, 3.3744E+03,
     +  3.3808E+03, 3.3873E+03, 3.3938E+03, 3.4003E+03, 3.4068E+03,
     + 3.4133E+03, 3.4198E+03, 3.4263E+03, 3.4328E+03, 3.4394E+03,
     +  3.4459E+03, 3.4524E+03, 3.4590E+03, 3.4655E+03, 3.4721E+03,
     + 3.4787E+03, 3.4852E+03, 3.4918E+03, 3.4984E+03, 3.5050E+03,
     +  3.5116E+03, 3.5182E+03, 3.5248E+03, 3.5314E+03, 3.5381E+03,
     + 3.5447E+03, 3.5513E+03, 3.5580E+03, 3.5646E+03, 3.5713E+03,
     +  3.5779E+03, 3.5846E+03, 3.5913E+03, 3.5980E+03, 3.6047E+03,
     + 3.6114E+03, 3.6181E+03, 3.6248E+03, 3.6315E+03, 3.6382E+03,
     +  3.6449E+03, 3.6517E+03, 3.6584E+03, 3.6652E+03, 3.6719E+03,
     + 3.6787E+03, 3.6854E+03, 3.6922E+03, 3.6990E+03, 3.7058E+03,
     +  3.7126E+03, 3.7194E+03, 3.7262E+03, 3.7330E+03, 3.7398E+03,
     + 3.7466E+03, 3.7534E+03, 3.7603E+03, 3.7671E+03, 3.7740E+03,
     +  3.7808E+03, 3.7877E+03, 3.7946E+03, 3.8014E+03, 3.8083E+03,
     + 3.8152E+03, 3.8221E+03, 3.8290E+03, 3.8359E+03, 3.8428E+03,
     +  3.8498E+03, 3.8567E+03, 3.8636E+03, 3.8706E+03, 3.8775E+03,
     + 3.8845E+03, 3.8914E+03, 3.8984E+03, 3.9054E+03, 3.9123E+03,
     +  3.9193E+03, 3.9263E+03, 3.9333E+03, 3.9403E+03, 3.9473E+03,
     + 3.9543E+03, 3.9614E+03, 3.9684E+03, 3.9754E+03, 3.9825E+03,
     +  3.9895E+03, 3.9966E+03, 4.0037E+03, 4.0107E+03, 4.0178E+03,
     + 4.0249E+03, 4.0320E+03, 4.0391E+03, 4.0462E+03, 4.0533E+03,
     +  4.0604E+03, 4.0675E+03, 4.0747E+03, 4.0818E+03, 4.0889E+03,
     + 4.0961E+03, 4.1032E+03, 4.1104E+03, 4.1176E+03, 4.1247E+03,
     +  4.1319E+03, 4.1391E+03, 4.1463E+03, 4.1535E+03, 4.1607E+03,
     + 4.1679E+03, 4.1752E+03, 4.1824E+03, 4.1896E+03, 4.1969E+03,
     +  4.2041E+03, 4.2114E+03, 4.2186E+03, 4.2259E+03, 4.2332E+03,
     + 4.2404E+03, 4.2477E+03, 4.2550E+03, 4.2623E+03, 4.2696E+03,
     +  4.2769E+03, 4.2843E+03, 4.2916E+03, 4.2989E+03, 4.3063E+03,
     + 4.3136E+03, 4.3210E+03, 4.3283E+03, 4.3357E+03, 4.3431E+03,
     +  4.3504E+03, 4.3578E+03, 4.3652E+03, 4.3726E+03, 4.3800E+03,
     + 4.3874E+03, 4.3948E+03, 4.4023E+03, 4.4097E+03, 4.4171E+03,
     +  4.4246E+03, 4.4320E+03, 4.4395E+03, 4.4470E+03, 4.4544E+03,
     + 4.4619E+03, 4.4694E+03, 4.4769E+03, 4.4844E+03, 4.4919E+03,
     +  4.4994E+03, 4.5069E+03, 4.5144E+03, 4.5220E+03, 4.5295E+03,
     + 4.5370E+03, 4.5446E+03, 4.5522E+03, 4.5597E+03, 4.5673E+03,
     +  4.5749E+03, 4.5825E+03, 4.5900E+03, 4.5976E+03, 4.6052E+03,
     + 4.6129E+03, 4.6205E+03, 4.6281E+03, 4.6357E+03, 4.6434E+03,
     +  4.6510E+03, 4.6587E+03, 4.6663E+03, 4.6740E+03, 4.6817E+03,
     + 4.6893E+03, 4.6970E+03, 4.7047E+03, 4.7124E+03, 4.7201E+03,
     +  4.7278E+03, 4.7355E+03, 4.7433E+03, 4.7510E+03, 4.7587E+03,
     + 4.7665E+03, 4.7742E+03, 4.7820E+03, 4.7897E+03, 4.7975E+03,
     +  4.8053E+03, 4.8131E+03, 4.8209E+03, 4.8287E+03, 4.8365E+03,
     + 4.8443E+03, 4.8521E+03, 4.8599E+03, 4.8677E+03, 4.8756E+03,
     +  4.8834E+03, 4.8913E+03, 4.8991E+03, 4.9070E+03, 4.9149E+03,
     + 4.9227E+03, 4.9306E+03, 4.9385E+03, 4.9464E+03, 4.9543E+03,
     +  4.9622E+03, 4.9702E+03, 4.9781E+03, 4.9860E+03, 4.9939E+03,
     + 5.0019E+03, 5.0098E+03, 5.0178E+03, 5.0258E+03, 5.0337E+03,
     +  5.0417E+03, 5.0497E+03, 5.0577E+03, 5.0657E+03, 5.0737E+03,
     + 5.0817E+03, 5.0897E+03, 5.0978E+03, 5.1058E+03, 5.1138E+03,
     +  5.1219E+03, 5.1299E+03, 5.1380E+03, 5.1461E+03, 5.1541E+03,
     + 5.1622E+03, 5.1703E+03, 5.1784E+03, 5.1865E+03, 5.1946E+03,
     +  5.2027E+03, 5.2108E+03, 5.2189E+03, 5.2271E+03, 5.2352E+03,
     + 5.2434E+03, 5.2515E+03, 5.2597E+03, 5.2679E+03, 5.2760E+03,
     +  5.2842E+03, 5.2924E+03, 5.3006E+03, 5.3088E+03, 5.3170E+03,
     + 5.3252E+03, 5.3334E+03, 5.3417E+03, 5.3499E+03, 5.3581E+03,
     +  5.3664E+03, 5.3746E+03, 5.3829E+03, 5.3912E+03, 5.3994E+03,
     + 5.4077E+03, 5.4160E+03, 5.4243E+03, 5.4326E+03, 5.4409E+03,
     +  5.4492E+03, 5.4576E+03, 5.4659E+03, 5.4742E+03, 5.4826E+03,
     + 5.4909E+03, 5.4993E+03, 5.5077E+03, 5.5160E+03, 5.5244E+03,
     +  5.5328E+03, 5.5412E+03, 5.5496E+03, 5.5580E+03, 5.5664E+03,
     + 5.5748E+03, 5.5832E+03, 5.5917E+03, 5.6001E+03, 5.6085E+03,
     +  5.6170E+03, 5.6255E+03, 5.6339E+03, 5.6424E+03, 5.6509E+03,
     + 5.6594E+03, 5.6679E+03, 5.6764E+03, 5.6849E+03, 5.6934E+03,
     +  5.7019E+03, 5.7104E+03, 5.7190E+03, 5.7275E+03, 5.7361E+03,
     + 5.7446E+03, 5.7532E+03, 5.7617E+03, 5.7703E+03, 5.7789E+03,
     +  5.7875E+03, 5.7961E+03, 5.8047E+03, 5.8133E+03, 5.8219E+03,
     + 5.8305E+03, 5.8392E+03, 5.8478E+03, 5.8564E+03, 5.8651E+03,
     +  5.8737E+03, 5.8824E+03, 5.8911E+03, 5.8998E+03, 5.9084E+03,
     + 5.9171E+03, 5.9258E+03, 5.9345E+03, 5.9433E+03, 5.9520E+03,
     +  5.9607E+03, 5.9694E+03, 5.9782E+03, 5.9869E+03, 5.9957E+03,
     + 6.0044E+03, 6.0132E+03, 6.0220E+03, 6.0308E+03, 6.0395E+03,
     +  6.0483E+03, 6.0571E+03, 6.0659E+03, 6.0748E+03, 6.0836E+03,
     + 6.0924E+03, 6.1012E+03, 6.1101E+03, 6.1189E+03, 6.1278E+03,
     +  6.1367E+03, 6.1455E+03, 6.1544E+03, 6.1633E+03, 6.1722E+03,
     + 6.1811E+03, 6.1900E+03, 6.1989E+03, 6.2078E+03, 6.2167E+03,
     +  6.2257E+03, 6.2346E+03, 6.2436E+03, 6.2525E+03, 6.2615E+03,
     + 6.2704E+03, 6.2794E+03, 6.2884E+03, 6.2974E+03, 6.3064E+03,
     +  6.3154E+03, 6.3244E+03, 6.3334E+03, 6.3424E+03, 6.3514E+03,
     + 6.3605E+03, 6.3695E+03, 6.3786E+03, 6.3876E+03, 6.3967E+03,
     +  6.4058E+03, 6.4148E+03, 6.4239E+03, 6.4330E+03, 6.4421E+03,
     + 6.4512E+03, 6.4603E+03, 6.4695E+03, 6.4786E+03, 6.4877E+03,
     +  6.4968E+03, 6.5060E+03, 6.5151E+03, 6.5243E+03, 6.5335E+03,
     + 6.5427E+03, 6.5518E+03, 6.5610E+03, 6.5702E+03, 6.5794E+03,
     +  6.5886E+03, 6.5978E+03, 6.6071E+03, 6.6163E+03, 6.6255E+03,
     + 6.6348E+03, 6.6440E+03, 6.6533E+03, 6.6625E+03, 6.6718E+03,
     +  6.6811E+03, 6.6904E+03, 6.6997E+03, 6.7090E+03, 6.7183E+03,
     + 6.7276E+03, 6.7369E+03, 6.7462E+03, 6.7556E+03, 6.7649E+03,
     +  6.7743E+03, 6.7836E+03, 6.7930E+03, 6.8023E+03, 6.8117E+03,
     + 6.8211E+03, 6.8305E+03, 6.8399E+03, 6.8493E+03, 6.8587E+03,
     +  6.8681E+03, 6.8775E+03, 6.8870E+03, 6.8964E+03, 6.9059E+03,
     + 6.9153E+03, 6.9248E+03, 6.9342E+03, 6.9437E+03, 6.9532E+03,
     +  6.9627E+03, 6.9722E+03, 6.9817E+03, 6.9912E+03, 7.0007E+03,
     + 7.0102E+03, 7.0197E+03, 7.0293E+03, 7.0388E+03, 7.0484E+03,
     +  7.0579E+03, 7.0675E+03, 7.0771E+03, 7.0866E+03, 7.0962E+03,
     + 7.1058E+03, 7.1154E+03, 7.1250E+03, 7.1346E+03, 7.1443E+03,
     +  7.1539E+03, 7.1635E+03, 7.1732E+03, 7.1828E+03, 7.1925E+03,
     + 7.2021E+03, 7.2118E+03, 7.2215E+03, 7.2311E+03, 7.2408E+03,
     +  7.2505E+03, 7.2602E+03, 7.2699E+03, 7.2797E+03, 7.2894E+03,
     + 7.2991E+03, 7.3089E+03, 7.3186E+03, 7.3284E+03, 7.3381E+03,
     +  7.3479E+03, 7.3577E+03, 7.3674E+03, 7.3772E+03, 7.3870E+03,
     + 7.3968E+03, 7.4066E+03, 7.4164E+03, 7.4263E+03, 7.4361E+03,
     +  7.4459E+03, 7.4558E+03, 7.4656E+03, 7.4755E+03, 7.4854E+03,
     + 7.4952E+03, 7.5051E+03, 7.5150E+03, 7.5249E+03, 7.5348E+03,
     +  7.5447E+03, 7.5546E+03, 7.5645E+03, 7.5744E+03, 7.5844E+03,
     + 7.5943E+03, 7.6043E+03, 7.6142E+03, 7.6242E+03, 7.6342E+03,
     +  7.6441E+03, 7.6541E+03, 7.6641E+03, 7.6741E+03, 7.6841E+03,
     + 7.6941E+03, 7.7042E+03, 7.7142E+03, 7.7242E+03, 7.7343E+03,
     +  7.7443E+03, 7.7544E+03, 7.7644E+03, 7.7745E+03, 7.7846E+03,
     + 7.7946E+03, 7.8047E+03, 7.8148E+03, 7.8249E+03, 7.8350E+03,
     +  7.8452E+03, 7.8553E+03, 7.8654E+03, 7.8756E+03, 7.8857E+03,
     + 7.8959E+03, 7.9060E+03, 7.9162E+03, 7.9264E+03, 7.9365E+03,
     +  7.9467E+03, 7.9569E+03, 7.9671E+03, 7.9773E+03, 7.9876E+03,
     + 7.9978E+03, 8.0080E+03, 8.0182E+03, 8.0285E+03, 8.0387E+03,
     +  8.0490E+03, 8.0593E+03, 8.0695E+03, 8.0798E+03, 8.0901E+03,
     + 8.1004E+03, 8.1107E+03, 8.1210E+03, 8.1313E+03, 8.1417E+03,
     +  8.1520E+03, 8.1623E+03, 8.1727E+03, 8.1830E+03, 8.1934E+03,
     + 8.2037E+03, 8.2141E+03, 8.2245E+03, 8.2349E+03, 8.2453E+03,
     +  8.2557E+03, 8.2661E+03, 8.2765E+03, 8.2869E+03, 8.2973E+03,
     + 8.3078E+03, 8.3182E+03, 8.3287E+03, 8.3391E+03, 8.3496E+03,
     +  8.3601E+03, 8.3705E+03, 8.3810E+03, 8.3915E+03, 8.4020E+03,
     + 8.4125E+03, 8.4230E+03, 8.4336E+03, 8.4441E+03, 8.4546E+03,
     +  8.4652E+03, 8.4757E+03, 8.4863E+03, 8.4968E+03, 8.5074E+03,
     + 8.5180E+03, 8.5286E+03, 8.5392E+03, 8.5498E+03, 8.5604E+03,
     +  8.5710E+03, 8.5816E+03, 8.5922E+03, 8.6029E+03, 8.6135E+03,
     + 8.6241E+03, 8.6348E+03, 8.6455E+03, 8.6561E+03, 8.6668E+03,
     +  8.6775E+03, 8.6882E+03, 8.6989E+03, 8.7096E+03, 8.7203E+03,
     + 8.7310E+03, 8.7417E+03, 8.7525E+03, 8.7632E+03, 8.7740E+03,
     +  8.7847E+03, 8.7955E+03, 8.8063E+03, 8.8170E+03, 8.8278E+03,
     + 8.8386E+03, 8.8494E+03, 8.8602E+03, 8.8710E+03, 8.8818E+03,
     +  8.8927E+03, 8.9035E+03, 8.9143E+03, 8.9252E+03, 8.9360E+03,
     + 8.9469E+03, 8.9578E+03, 8.9686E+03, 8.9795E+03, 8.9904E+03,
     +  9.0013E+03, 9.0122E+03, 9.0231E+03, 9.0341E+03, 9.0450E+03,
     + 9.0559E+03, 9.0669E+03, 9.0778E+03, 9.0888E+03, 9.0997E+03,
     +  9.1107E+03, 9.1217E+03, 9.1326E+03, 9.1436E+03, 9.1546E+03,
     + 9.1656E+03, 9.1766E+03, 9.1877E+03, 9.1987E+03, 9.2097E+03,
     +  9.2208E+03, 9.2318E+03, 9.2429E+03, 9.2539E+03, 9.2650E+03,
     + 9.2761E+03, 9.2871E+03, 9.2982E+03, 9.3093E+03, 9.3204E+03,
     +  9.3315E+03, 9.3427E+03, 9.3538E+03, 9.3649E+03, 9.3761E+03,
     + 9.3872E+03, 9.3984E+03, 9.4095E+03, 9.4207E+03, 9.4319E+03,
     +  9.4430E+03, 9.4542E+03, 9.4654E+03, 9.4766E+03, 9.4878E+03,
     + 9.4991E+03, 9.5103E+03, 9.5215E+03, 9.5328E+03, 9.5440E+03,
     +  9.5553E+03, 9.5665E+03, 9.5778E+03, 9.5891E+03, 9.6003E+03,
     + 9.6116E+03, 9.6229E+03, 9.6342E+03, 9.6455E+03, 9.6569E+03,
     +  9.6682E+03, 9.6795E+03, 9.6909E+03, 9.7022E+03, 9.7136E+03,
     + 9.7249E+03, 9.7363E+03, 9.7477E+03, 9.7590E+03, 9.7704E+03,
     +  9.7818E+03, 9.7932E+03, 9.8046E+03, 9.8161E+03, 9.8275E+03,
     + 9.8389E+03, 9.8504E+03, 9.8618E+03, 9.8733E+03, 9.8847E+03,
     +  9.8962E+03, 9.9077E+03, 9.9191E+03, 9.9306E+03, 9.9421E+03,
     + 9.9536E+03, 9.9651E+03, 9.9767E+03, 9.9882E+03, 9.9997E+03,
     +  1.0011E+04, 1.0023E+04, 1.0034E+04, 1.0046E+04, 1.0057E+04,
     + 1.0069E+04, 1.0081E+04, 1.0092E+04, 1.0104E+04, 1.0115E+04,
     +  1.0127E+04, 1.0139E+04, 1.0150E+04, 1.0162E+04, 1.0174E+04,
     + 1.0185E+04, 1.0197E+04, 1.0209E+04, 1.0220E+04, 1.0232E+04,
     +  1.0244E+04, 1.0255E+04, 1.0267E+04, 1.0279E+04, 1.0290E+04,
     + 1.0302E+04, 1.0314E+04, 1.0326E+04, 1.0337E+04, 1.0349E+04,
     +  1.0361E+04, 1.0373E+04, 1.0384E+04, 1.0396E+04, 1.0408E+04,
     + 1.0420E+04, 1.0431E+04, 1.0443E+04, 1.0455E+04, 1.0467E+04,
     +  1.0479E+04, 1.0491E+04, 1.0502E+04, 1.0514E+04, 1.0526E+04,
     + 1.0538E+04, 1.0550E+04, 1.0562E+04, 1.0574E+04, 1.0585E+04,
     +  1.0597E+04, 1.0609E+04, 1.0621E+04, 1.0633E+04, 1.0645E+04,
     + 1.0657E+04, 1.0669E+04, 1.0681E+04, 1.0693E+04, 1.0705E+04,
     +  1.0717E+04, 1.0729E+04, 1.0741E+04, 1.0753E+04, 1.0765E+04,
     + 1.0777E+04, 1.0789E+04, 1.0801E+04, 1.0813E+04, 1.0825E+04,
     +  1.0837E+04, 1.0849E+04, 1.0861E+04, 1.0873E+04, 1.0885E+04,
     + 1.0897E+04, 1.0909E+04, 1.0921E+04, 1.0933E+04, 1.0946E+04,
     +  1.0958E+04, 1.0970E+04, 1.0982E+04, 1.0994E+04, 1.1006E+04,
     + 1.1018E+04, 1.1030E+04, 1.1043E+04, 1.1055E+04, 1.1067E+04,
     +  1.1079E+04, 1.1091E+04, 1.1104E+04, 1.1116E+04, 1.1128E+04,
     + 1.1140E+04, 1.1152E+04, 1.1165E+04, 1.1177E+04, 1.1189E+04,
     +  1.1201E+04, 1.1214E+04, 1.1226E+04, 1.1238E+04, 1.1250E+04,
     + 1.1263E+04, 1.1275E+04, 1.1287E+04, 1.1300E+04, 1.1312E+04,
     +  1.1324E+04, 1.1337E+04, 1.1349E+04, 1.1361E+04, 1.1374E+04,
     + 1.1386E+04, 1.1398E+04, 1.1411E+04, 1.1423E+04, 1.1436E+04,
     +  1.1448E+04, 1.1460E+04, 1.1473E+04, 1.1485E+04, 1.1498E+04,
     + 1.1510E+04, 1.1523E+04, 1.1535E+04, 1.1547E+04, 1.1560E+04,
     +  1.1572E+04, 1.1585E+04, 1.1597E+04, 1.1610E+04, 1.1622E+04,
     + 1.1635E+04, 1.1647E+04, 1.1660E+04, 1.1672E+04, 1.1685E+04,
     +  1.1697E+04, 1.1710E+04, 1.1723E+04, 1.1735E+04, 1.1748E+04,
     + 1.1760E+04, 1.1773E+04, 1.1785E+04, 1.1798E+04, 1.1811E+04,
     +  1.1823E+04, 1.1836E+04, 1.1849E+04, 1.1861E+04, 1.1874E+04,
     + 1.1886E+04, 1.1899E+04, 1.1912E+04, 1.1924E+04, 1.1937E+04,
     +  1.1950E+04, 1.1963E+04, 1.1975E+04, 1.1988E+04, 1.2001E+04,
     + 1.2013E+04, 1.2026E+04, 1.2039E+04, 1.2052E+04, 1.2064E+04,
     +  1.2077E+04, 1.2090E+04, 1.2103E+04, 1.2115E+04, 1.2128E+04,
     + 1.2141E+04, 1.2154E+04, 1.2167E+04, 1.2179E+04, 1.2192E+04,
     +  1.2205E+04, 1.2218E+04, 1.2231E+04, 1.2244E+04, 1.2256E+04,
     + 1.2269E+04, 1.2282E+04, 1.2295E+04, 1.2308E+04, 1.2321E+04,
     +  1.2334E+04, 1.2347E+04, 1.2360E+04, 1.2373E+04, 1.2385E+04,
     + 1.2398E+04, 1.2411E+04, 1.2424E+04, 1.2437E+04, 1.2450E+04,
     +  1.2463E+04, 1.2476E+04, 1.2489E+04, 1.2502E+04, 1.2515E+04,
     + 1.2528E+04, 1.2541E+04, 1.2554E+04, 1.2567E+04, 1.2580E+04,
     +  1.2593E+04, 1.2606E+04, 1.2619E+04, 1.2633E+04, 1.2646E+04,
     + 1.2659E+04, 1.2672E+04, 1.2685E+04, 1.2698E+04, 1.2711E+04,
     +  1.2724E+04, 1.2737E+04, 1.2750E+04, 1.2764E+04, 1.2777E+04,
     + 1.2790E+04, 1.2803E+04, 1.2816E+04, 1.2829E+04, 1.2843E+04,
     +  1.2856E+04, 1.2869E+04, 1.2882E+04, 1.2895E+04, 1.2909E+04,
     + 1.2922E+04, 1.2935E+04, 1.2948E+04, 1.2962E+04, 1.2975E+04,
     +  1.2988E+04, 1.3001E+04, 1.3015E+04, 1.3028E+04, 1.3041E+04,
     + 1.3054E+04, 1.3068E+04, 1.3081E+04, 1.3094E+04, 1.3108E+04,
     +  1.3121E+04, 1.3134E+04, 1.3148E+04, 1.3161E+04, 1.3174E+04,
     + 1.3188E+04, 1.3201E+04, 1.3215E+04, 1.3228E+04, 1.3241E+04,
     +  1.3255E+04, 1.3268E+04, 1.3282E+04, 1.3295E+04, 1.3308E+04,
     + 1.3322E+04, 1.3335E+04, 1.3349E+04, 1.3362E+04, 1.3376E+04,
     +  1.3389E+04, 1.3403E+04, 1.3416E+04, 1.3430E+04, 1.3443E+04,
     + 1.3457E+04, 1.3470E+04, 1.3484E+04, 1.3497E+04, 1.3511E+04,
     +  1.3524E+04, 1.3538E+04, 1.3551E+04, 1.3565E+04, 1.3579E+04,
     + 1.3592E+04, 1.3606E+04, 1.3619E+04, 1.3633E+04, 1.3647E+04,
     +  1.3660E+04, 1.3674E+04, 1.3687E+04, 1.3701E+04, 1.3715E+04,
     + 1.3728E+04, 1.3742E+04, 1.3756E+04, 1.3769E+04, 1.3783E+04,
     +  1.3797E+04, 1.3810E+04, 1.3824E+04, 1.3838E+04, 1.3852E+04,
     + 1.3865E+04, 1.3879E+04, 1.3893E+04, 1.3907E+04, 1.3920E+04,
     +  1.3934E+04, 1.3948E+04, 1.3962E+04, 1.3975E+04, 1.3989E+04,
     + 1.4003E+04, 1.4017E+04, 1.4031E+04, 1.4044E+04, 1.4058E+04,
     +  1.4072E+04, 1.4086E+04, 1.4100E+04, 1.4114E+04, 1.4127E+04,
     + 1.4141E+04, 1.4155E+04, 1.4169E+04, 1.4183E+04, 1.4197E+04,
     +  1.4211E+04, 1.4225E+04, 1.4239E+04, 1.4253E+04, 1.4266E+04,
     + 1.4280E+04, 1.4294E+04, 1.4308E+04, 1.4322E+04, 1.4336E+04,
     +  1.4350E+04, 1.4364E+04, 1.4378E+04, 1.4392E+04, 1.4406E+04,
     + 1.4420E+04, 1.4434E+04, 1.4448E+04, 1.4462E+04, 1.4476E+04,
     +  1.4490E+04, 1.4504E+04, 1.4519E+04, 1.4533E+04, 1.4547E+04,
     + 1.4561E+04, 1.4575E+04, 1.4589E+04, 1.4603E+04, 1.4617E+04,
     +  1.4631E+04, 1.4645E+04, 1.4660E+04, 1.4674E+04, 1.4688E+04,
     + 1.4702E+04, 1.4716E+04, 1.4730E+04, 1.4744E+04, 1.4759E+04,
     +  1.4773E+04, 1.4787E+04, 1.4801E+04, 1.4815E+04, 1.4830E+04,
     + 1.4844E+04, 1.4858E+04, 1.4872E+04, 1.4887E+04, 1.4901E+04,
     +  1.4915E+04, 1.4929E+04, 1.4944E+04, 1.4958E+04, 1.4972E+04,
     + 1.4987E+04, 1.5001E+04, 1.5015E+04, 1.5029E+04, 1.5044E+04,
     +  1.5058E+04, 1.5072E+04, 1.5087E+04, 1.5101E+04, 1.5116E+04,
     + 1.5130E+04, 1.5144E+04, 1.5159E+04, 1.5173E+04, 1.5187E+04,
     +  1.5202E+04, 1.5216E+04, 1.5231E+04, 1.5245E+04, 1.5260E+04,
     + 1.5274E+04, 1.5288E+04, 1.5303E+04, 1.5317E+04, 1.5332E+04,
     +  1.5346E+04, 1.5361E+04, 1.5375E+04, 1.5390E+04, 1.5404E+04,
     + 1.5419E+04, 1.5433E+04, 1.5448E+04, 1.5462E+04, 1.5477E+04,
     +  1.5491E+04, 1.5506E+04, 1.5521E+04, 1.5535E+04, 1.5550E+04,
     + 1.5564E+04, 1.5579E+04, 1.5593E+04, 1.5608E+04, 1.5623E+04,
     +  1.5637E+04, 1.5652E+04, 1.5667E+04, 1.5681E+04, 1.5696E+04,
     + 1.5710E+04, 1.5725E+04, 1.5740E+04, 1.5754E+04, 1.5769E+04,
     +  1.5784E+04, 1.5799E+04, 1.5813E+04, 1.5828E+04, 1.5843E+04,
     + 1.5857E+04, 1.5872E+04, 1.5887E+04, 1.5902E+04, 1.5916E+04,
     +  1.5931E+04, 1.5946E+04, 1.5961E+04, 1.5975E+04, 1.5990E+04,
     + 1.6005E+04, 1.6020E+04, 1.6035E+04, 1.6049E+04, 1.6064E+04,
     +  1.6079E+04, 1.6094E+04, 1.6109E+04, 1.6124E+04, 1.6138E+04,
     + 1.6153E+04, 1.6168E+04, 1.6183E+04, 1.6198E+04, 1.6213E+04,
     +  1.6228E+04, 1.6243E+04, 1.6258E+04, 1.6273E+04, 1.6287E+04,
     + 1.6302E+04, 1.6317E+04, 1.6332E+04, 1.6347E+04, 1.6362E+04,
     +  1.6377E+04, 1.6392E+04, 1.6407E+04, 1.6422E+04, 1.6437E+04,
     + 1.6452E+04, 1.6467E+04, 1.6482E+04, 1.6497E+04, 1.6512E+04,
     +  1.6527E+04, 1.6542E+04, 1.6557E+04, 1.6572E+04, 1.6588E+04,
     + 1.6603E+04, 1.6618E+04, 1.6633E+04, 1.6648E+04, 1.6663E+04,
     +  1.6678E+04, 1.6693E+04, 1.6708E+04, 1.6723E+04, 1.6739E+04,
     + 1.6754E+04, 1.6769E+04, 1.6784E+04, 1.6799E+04, 1.6814E+04,
     +  1.6830E+04, 1.6845E+04, 1.6860E+04, 1.6875E+04, 1.6890E+04,
     + 1.6906E+04, 1.6921E+04, 1.6936E+04, 1.6951E+04, 1.6967E+04,
     +  1.6982E+04, 1.6997E+04, 1.7012E+04, 1.7028E+04, 1.7043E+04,
     + 1.7058E+04, 1.7074E+04, 1.7089E+04, 1.7104E+04, 1.7119E+04,
     +  1.7135E+04, 1.7150E+04, 1.7165E+04, 1.7181E+04, 1.7196E+04,
     + 1.7212E+04, 1.7227E+04, 1.7242E+04, 1.7258E+04, 1.7273E+04,
     +  1.7288E+04, 1.7304E+04, 1.7319E+04, 1.7335E+04, 1.7350E+04,
     + 1.7366E+04, 1.7381E+04, 1.7396E+04, 1.7412E+04, 1.7427E+04,
     +  1.7443E+04, 1.7458E+04, 1.7474E+04, 1.7489E+04, 1.7505E+04,
     + 1.7520E+04, 1.7536E+04, 1.7551E+04, 1.7567E+04, 1.7582E+04,
     +  1.7598E+04, 1.7613E+04, 1.7629E+04, 1.7644E+04, 1.7660E+04,
     + 1.7676E+04, 1.7691E+04, 1.7707E+04, 1.7722E+04, 1.7738E+04,
     +  1.7754E+04, 1.7769E+04, 1.7785E+04, 1.7800E+04, 1.7816E+04,
     + 1.7832E+04, 1.7847E+04, 1.7863E+04, 1.7879E+04, 1.7894E+04,
     +  1.7910E+04, 1.7926E+04, 1.7941E+04, 1.7957E+04, 1.7973E+04,
     + 1.7988E+04, 1.8004E+04, 1.8020E+04, 1.8036E+04, 1.8051E+04,
     +  1.8067E+04, 1.8083E+04, 1.8099E+04, 1.8114E+04, 1.8130E+04,
     + 1.8146E+04, 1.8162E+04, 1.8178E+04, 1.8193E+04, 1.8209E+04,
     +  1.8225E+04, 1.8241E+04, 1.8257E+04, 1.8272E+04, 1.8288E+04,
     + 1.8304E+04, 1.8320E+04, 1.8336E+04, 1.8352E+04, 1.8368E+04,
     +  1.8384E+04, 1.8399E+04, 1.8415E+04, 1.8431E+04, 1.8447E+04,
     + 1.8463E+04, 1.8479E+04, 1.8495E+04, 1.8511E+04, 1.8527E+04,
     +  1.8543E+04, 1.8559E+04, 1.8575E+04, 1.8591E+04, 1.8607E+04,
     + 1.8623E+04, 1.8639E+04, 1.8655E+04, 1.8671E+04, 1.8687E+04,
     +  1.8703E+04, 1.8719E+04, 1.8735E+04, 1.8751E+04, 1.8767E+04,
     + 1.8783E+04, 1.8799E+04, 1.8815E+04, 1.8831E+04, 1.8847E+04,
     +  1.8863E+04, 1.8879E+04, 1.8896E+04, 1.8912E+04, 1.8928E+04,
     + 1.8944E+04, 1.8960E+04, 1.8976E+04, 1.8992E+04, 1.9009E+04,
     +  1.9025E+04, 1.9041E+04, 1.9057E+04, 1.9073E+04, 1.9089E+04,
     + 1.9106E+04, 1.9122E+04, 1.9138E+04, 1.9154E+04, 1.9170E+04,
     +  1.9187E+04, 1.9203E+04, 1.9219E+04, 1.9235E+04, 1.9252E+04,
     + 1.9268E+04, 1.9284E+04, 1.9301E+04, 1.9317E+04, 1.9333E+04,
     +  1.9349E+04, 1.9366E+04, 1.9382E+04, 1.9398E+04, 1.9415E+04,
     + 1.9431E+04, 1.9447E+04, 1.9464E+04, 1.9480E+04, 1.9496E+04,
     +  1.9513E+04, 1.9529E+04, 1.9546E+04, 1.9562E+04, 1.9578E+04,
     + 1.9595E+04, 1.9611E+04, 1.9628E+04, 1.9644E+04, 1.9661E+04,
     +  1.9677E+04, 1.9693E+04, 1.9710E+04, 1.9726E+04, 1.9743E+04,
     + 1.9759E+04, 1.9776E+04, 1.9792E+04, 1.9809E+04, 1.9825E+04,
     +  1.9842E+04, 1.9858E+04, 1.9875E+04, 1.9891E+04, 1.9908E+04,
     + 1.9924E+04, 1.9941E+04, 1.9958E+04, 1.9974E+04, 1.9991E+04,
     +  2.0007E+04, 2.0024E+04, 2.0040E+04, 2.0057E+04, 2.0074E+04,
     + 2.0090E+04, 2.0107E+04, 2.0124E+04, 2.0140E+04, 2.0157E+04,
     +  2.0173E+04, 2.0190E+04, 2.0207E+04, 2.0223E+04, 2.0240E+04,
     + 2.0257E+04, 2.0273E+04, 2.0290E+04, 2.0307E+04, 2.0324E+04,
     +  2.0340E+04, 2.0357E+04, 2.0374E+04, 2.0390E+04, 2.0407E+04,
     + 2.0424E+04, 2.0441E+04, 2.0458E+04, 2.0474E+04, 2.0491E+04,
     +  2.0508E+04, 2.0525E+04, 2.0541E+04, 2.0558E+04, 2.0575E+04,
     + 2.0592E+04, 2.0609E+04, 2.0626E+04, 2.0642E+04, 2.0659E+04,
     +  2.0676E+04, 2.0693E+04, 2.0710E+04, 2.0727E+04, 2.0744E+04,
     + 2.0760E+04, 2.0777E+04, 2.0794E+04, 2.0811E+04, 2.0828E+04,
     +  2.0845E+04, 2.0862E+04, 2.0879E+04, 2.0896E+04, 2.0913E+04,
     + 2.0930E+04, 2.0947E+04, 2.0964E+04, 2.0981E+04, 2.0998E+04,
     +  2.1015E+04, 2.1032E+04, 2.1049E+04, 2.1066E+04, 2.1083E+04,
     + 2.1100E+04, 2.1117E+04, 2.1134E+04, 2.1151E+04, 2.1168E+04,
     +  2.1185E+04, 2.1202E+04, 2.1219E+04, 2.1236E+04, 2.1253E+04,
     + 2.1270E+04, 2.1287E+04, 2.1305E+04, 2.1322E+04, 2.1339E+04,
     +  2.1356E+04, 2.1373E+04, 2.1390E+04, 2.1407E+04, 2.1425E+04,
     + 2.1442E+04, 2.1459E+04, 2.1476E+04, 2.1493E+04, 2.1510E+04,
     +  2.1528E+04, 2.1545E+04, 2.1562E+04, 2.1579E+04, 2.1596E+04,
     + 2.1614E+04, 2.1631E+04, 2.1648E+04, 2.1665E+04, 2.1683E+04,
     +  2.1700E+04, 2.1717E+04, 2.1735E+04, 2.1752E+04, 2.1769E+04,
     + 2.1786E+04, 2.1804E+04, 2.1821E+04, 2.1838E+04, 2.1856E+04,
     +  2.1873E+04, 2.1890E+04, 2.1908E+04, 2.1925E+04, 2.1942E+04,
     + 2.1960E+04, 2.1977E+04, 2.1995E+04, 2.2012E+04, 2.2029E+04,
     +  2.2047E+04, 2.2064E+04, 2.2082E+04, 2.2099E+04, 2.2116E+04,
     + 2.2134E+04, 2.2151E+04, 2.2169E+04, 2.2186E+04, 2.2204E+04,
     +  2.2221E+04, 2.2239E+04, 2.2256E+04, 2.2274E+04, 2.2291E+04,
     + 2.2309E+04, 2.2326E+04, 2.2344E+04, 2.2361E+04, 2.2379E+04,
     +  2.2396E+04, 2.2414E+04, 2.2431E+04, 2.2449E+04, 2.2466E+04,
     + 2.2484E+04, 2.2502E+04, 2.2519E+04, 2.2537E+04, 2.2554E+04,
     +  2.2572E+04, 2.2590E+04, 2.2607E+04, 2.2625E+04, 2.2642E+04,
     + 2.2660E+04, 2.2678E+04, 2.2695E+04, 2.2713E+04, 2.2731E+04,
     +  2.2748E+04, 2.2766E+04, 2.2784E+04, 2.2801E+04, 2.2819E+04,
     + 2.2837E+04, 2.2854E+04, 2.2872E+04, 2.2890E+04, 2.2908E+04,
     +  2.2925E+04, 2.2943E+04, 2.2961E+04, 2.2979E+04, 2.2996E+04,
     + 2.3014E+04, 2.3032E+04, 2.3050E+04, 2.3068E+04, 2.3085E+04,
     +  2.3103E+04, 2.3121E+04, 2.3139E+04, 2.3157E+04, 2.3174E+04,
     + 2.3192E+04, 2.3210E+04, 2.3228E+04, 2.3246E+04, 2.3264E+04,
     +  2.3282E+04, 2.3299E+04, 2.3317E+04, 2.3335E+04, 2.3353E+04,
     + 2.3371E+04, 2.3389E+04, 2.3407E+04, 2.3425E+04, 2.3443E+04,
     +  2.3461E+04, 2.3479E+04, 2.3497E+04, 2.3515E+04, 2.3533E+04,
     + 2.3551E+04, 2.3569E+04, 2.3586E+04, 2.3604E+04, 2.3622E+04,
     +  2.3640E+04, 2.3659E+04, 2.3677E+04, 2.3695E+04, 2.3713E+04,
     + 2.3731E+04, 2.3749E+04, 2.3767E+04, 2.3785E+04, 2.3803E+04,
     +  2.3821E+04, 2.3839E+04, 2.3857E+04, 2.3875E+04, 2.3893E+04,
     + 2.3911E+04, 2.3930E+04, 2.3948E+04, 2.3966E+04, 2.3984E+04,
     +  2.4002E+04, 2.4020E+04, 2.4038E+04, 2.4057E+04, 2.4075E+04/
      data (QofT( 3,J),J=2409,4510)/ 2.4093E+04, 2.4111E+04,
     +  2.4129E+04, 2.4147E+04, 2.4166E+04, 2.4184E+04, 2.4202E+04,
     +  2.4220E+04,
     + 2.4238E+04, 2.4257E+04, 2.4275E+04, 2.4293E+04, 2.4311E+04,
     +  2.4330E+04, 2.4348E+04, 2.4366E+04, 2.4385E+04, 2.4403E+04,
     + 2.4421E+04, 2.4439E+04, 2.4458E+04, 2.4476E+04, 2.4494E+04,
     +  2.4513E+04, 2.4531E+04, 2.4549E+04, 2.4568E+04, 2.4586E+04,
     + 2.4604E+04, 2.4623E+04, 2.4641E+04, 2.4660E+04, 2.4678E+04,
     +  2.4696E+04, 2.4715E+04, 2.4733E+04, 2.4752E+04, 2.4770E+04,
     + 2.4788E+04, 2.4807E+04, 2.4825E+04, 2.4844E+04, 2.4862E+04,
     +  2.4881E+04, 2.4899E+04, 2.4918E+04, 2.4936E+04, 2.4955E+04,
     + 2.4973E+04, 2.4992E+04, 2.5010E+04, 2.5029E+04, 2.5047E+04,
     +  2.5066E+04, 2.5084E+04, 2.5103E+04, 2.5121E+04, 2.5140E+04,
     + 2.5158E+04, 2.5177E+04, 2.5195E+04, 2.5214E+04, 2.5233E+04,
     +  2.5251E+04, 2.5270E+04, 2.5288E+04, 2.5307E+04, 2.5326E+04,
     + 2.5344E+04, 2.5363E+04, 2.5382E+04, 2.5400E+04, 2.5419E+04,
     +  2.5437E+04, 2.5456E+04, 2.5475E+04, 2.5494E+04, 2.5512E+04,
     + 2.5531E+04, 2.5550E+04, 2.5568E+04, 2.5587E+04, 2.5606E+04,
     +  2.5624E+04, 2.5643E+04, 2.5662E+04, 2.5681E+04, 2.5699E+04,
     + 2.5718E+04, 2.5737E+04, 2.5756E+04, 2.5774E+04, 2.5793E+04,
     +  2.5812E+04, 2.5831E+04, 2.5850E+04, 2.5868E+04, 2.5887E+04,
     + 2.5906E+04, 2.5925E+04, 2.5944E+04, 2.5963E+04, 2.5981E+04,
     +  2.6000E+04, 2.6019E+04, 2.6038E+04, 2.6057E+04, 2.6076E+04,
     + 2.6095E+04, 2.6114E+04, 2.6132E+04, 2.6151E+04, 2.6170E+04,
     +  2.6189E+04, 2.6208E+04, 2.6227E+04, 2.6246E+04, 2.6265E+04,
     + 2.6284E+04, 2.6303E+04, 2.6322E+04, 2.6341E+04, 2.6360E+04,
     +  2.6379E+04, 2.6398E+04, 2.6417E+04, 2.6436E+04, 2.6455E+04,
     + 2.6474E+04, 2.6493E+04, 2.6512E+04, 2.6531E+04, 2.6550E+04,
     +  2.6569E+04, 2.6588E+04, 2.6607E+04, 2.6626E+04, 2.6645E+04,
     + 2.6664E+04, 2.6683E+04, 2.6702E+04, 2.6722E+04, 2.6741E+04,
     +  2.6760E+04, 2.6779E+04, 2.6798E+04, 2.6817E+04, 2.6836E+04,
     + 2.6855E+04, 2.6875E+04, 2.6894E+04, 2.6913E+04, 2.6932E+04,
     +  2.6951E+04, 2.6970E+04, 2.6990E+04, 2.7009E+04, 2.7028E+04,
     + 2.7047E+04, 2.7066E+04, 2.7086E+04, 2.7105E+04, 2.7124E+04,
     +  2.7143E+04, 2.7163E+04, 2.7182E+04, 2.7201E+04, 2.7220E+04,
     + 2.7240E+04, 2.7259E+04, 2.7278E+04, 2.7298E+04, 2.7317E+04,
     +  2.7336E+04, 2.7356E+04, 2.7375E+04, 2.7394E+04, 2.7414E+04,
     + 2.7433E+04, 2.7452E+04, 2.7472E+04, 2.7491E+04, 2.7510E+04,
     +  2.7530E+04, 2.7549E+04, 2.7568E+04, 2.7588E+04, 2.7607E+04,
     + 2.7627E+04, 2.7646E+04, 2.7665E+04, 2.7685E+04, 2.7704E+04,
     +  2.7724E+04, 2.7743E+04, 2.7763E+04, 2.7782E+04, 2.7802E+04,
     + 2.7821E+04, 2.7840E+04, 2.7860E+04, 2.7879E+04, 2.7899E+04,
     +  2.7918E+04, 2.7938E+04, 2.7957E+04, 2.7977E+04, 2.7997E+04,
     + 2.8016E+04, 2.8036E+04, 2.8055E+04, 2.8075E+04, 2.8094E+04,
     +  2.8114E+04, 2.8133E+04, 2.8153E+04, 2.8173E+04, 2.8192E+04,
     + 2.8212E+04, 2.8231E+04, 2.8251E+04, 2.8271E+04, 2.8290E+04,
     +  2.8310E+04, 2.8329E+04, 2.8349E+04, 2.8369E+04, 2.8388E+04,
     + 2.8408E+04, 2.8428E+04, 2.8447E+04, 2.8467E+04, 2.8487E+04,
     +  2.8506E+04, 2.8526E+04, 2.8546E+04, 2.8566E+04, 2.8585E+04,
     + 2.8605E+04, 2.8625E+04, 2.8644E+04, 2.8664E+04, 2.8684E+04,
     +  2.8704E+04, 2.8723E+04, 2.8743E+04, 2.8763E+04, 2.8783E+04,
     + 2.8803E+04, 2.8822E+04, 2.8842E+04, 2.8862E+04, 2.8882E+04,
     +  2.8902E+04, 2.8921E+04, 2.8941E+04, 2.8961E+04, 2.8981E+04,
     + 2.9001E+04, 2.9021E+04, 2.9041E+04, 2.9060E+04, 2.9080E+04,
     +  2.9100E+04, 2.9120E+04, 2.9140E+04, 2.9160E+04, 2.9180E+04,
     + 2.9200E+04, 2.9220E+04, 2.9240E+04, 2.9259E+04, 2.9279E+04,
     +  2.9299E+04, 2.9319E+04, 2.9339E+04, 2.9359E+04, 2.9379E+04,
     + 2.9399E+04, 2.9419E+04, 2.9439E+04, 2.9459E+04, 2.9479E+04,
     +  2.9499E+04, 2.9519E+04, 2.9539E+04, 2.9559E+04, 2.9579E+04,
     + 2.9599E+04, 2.9619E+04, 2.9639E+04, 2.9659E+04, 2.9680E+04,
     +  2.9700E+04, 2.9720E+04, 2.9740E+04, 2.9760E+04, 2.9780E+04,
     + 2.9800E+04, 2.9820E+04, 2.9840E+04, 2.9860E+04, 2.9881E+04,
     +  2.9901E+04, 2.9921E+04, 2.9941E+04, 2.9961E+04, 2.9981E+04,
     + 3.0001E+04, 3.0022E+04, 3.0042E+04, 3.0062E+04, 3.0082E+04,
     +  3.0102E+04, 3.0123E+04, 3.0143E+04, 3.0163E+04, 3.0183E+04,
     + 3.0203E+04, 3.0224E+04, 3.0244E+04, 3.0264E+04, 3.0284E+04,
     +  3.0305E+04, 3.0325E+04, 3.0345E+04, 3.0366E+04, 3.0386E+04,
     + 3.0406E+04, 3.0426E+04, 3.0447E+04, 3.0467E+04, 3.0487E+04,
     +  3.0508E+04, 3.0528E+04, 3.0548E+04, 3.0569E+04, 3.0589E+04,
     + 3.0609E+04, 3.0630E+04, 3.0650E+04, 3.0670E+04, 3.0691E+04,
     +  3.0711E+04, 3.0732E+04, 3.0752E+04, 3.0772E+04, 3.0793E+04,
     + 3.0813E+04, 3.0834E+04, 3.0854E+04, 3.0875E+04, 3.0895E+04,
     +  3.0915E+04, 3.0936E+04, 3.0956E+04, 3.0977E+04, 3.0997E+04,
     + 3.1018E+04, 3.1038E+04, 3.1059E+04, 3.1079E+04, 3.1100E+04,
     +  3.1120E+04, 3.1141E+04, 3.1161E+04, 3.1182E+04, 3.1202E+04,
     + 3.1223E+04, 3.1243E+04, 3.1264E+04, 3.1285E+04, 3.1305E+04,
     +  3.1326E+04, 3.1346E+04, 3.1367E+04, 3.1387E+04, 3.1408E+04,
     + 3.1429E+04, 3.1449E+04, 3.1470E+04, 3.1490E+04, 3.1511E+04,
     +  3.1532E+04, 3.1552E+04, 3.1573E+04, 3.1594E+04, 3.1614E+04,
     + 3.1635E+04, 3.1656E+04, 3.1676E+04, 3.1697E+04, 3.1718E+04,
     +  3.1738E+04, 3.1759E+04, 3.1780E+04, 3.1800E+04, 3.1821E+04,
     + 3.1842E+04, 3.1863E+04, 3.1883E+04, 3.1904E+04, 3.1925E+04,
     +  3.1946E+04, 3.1966E+04, 3.1987E+04, 3.2008E+04, 3.2029E+04,
     + 3.2049E+04, 3.2070E+04, 3.2091E+04, 3.2112E+04, 3.2133E+04,
     +  3.2154E+04, 3.2174E+04, 3.2195E+04, 3.2216E+04, 3.2237E+04,
     + 3.2258E+04, 3.2279E+04, 3.2299E+04, 3.2320E+04, 3.2341E+04,
     +  3.2362E+04, 3.2383E+04, 3.2404E+04, 3.2425E+04, 3.2446E+04,
     + 3.2466E+04, 3.2487E+04, 3.2508E+04, 3.2529E+04, 3.2550E+04,
     +  3.2571E+04, 3.2592E+04, 3.2613E+04, 3.2634E+04, 3.2655E+04,
     + 3.2676E+04, 3.2697E+04, 3.2718E+04, 3.2739E+04, 3.2760E+04,
     +  3.2781E+04, 3.2802E+04, 3.2823E+04, 3.2844E+04, 3.2865E+04,
     + 3.2886E+04, 3.2907E+04, 3.2928E+04, 3.2949E+04, 3.2970E+04,
     +  3.2991E+04, 3.3012E+04, 3.3033E+04, 3.3054E+04, 3.3075E+04,
     + 3.3097E+04, 3.3118E+04, 3.3139E+04, 3.3160E+04, 3.3181E+04,
     +  3.3202E+04, 3.3223E+04, 3.3244E+04, 3.3266E+04, 3.3287E+04,
     + 3.3308E+04, 3.3329E+04, 3.3350E+04, 3.3371E+04, 3.3392E+04,
     +  3.3414E+04, 3.3435E+04, 3.3456E+04, 3.3477E+04, 3.3498E+04,
     + 3.3520E+04, 3.3541E+04, 3.3562E+04, 3.3583E+04, 3.3605E+04,
     +  3.3626E+04, 3.3647E+04, 3.3668E+04, 3.3690E+04, 3.3711E+04,
     + 3.3732E+04, 3.3753E+04, 3.3775E+04, 3.3796E+04, 3.3817E+04,
     +  3.3839E+04, 3.3860E+04, 3.3881E+04, 3.3902E+04, 3.3924E+04,
     + 3.3945E+04, 3.3966E+04, 3.3988E+04, 3.4009E+04, 3.4031E+04,
     +  3.4052E+04, 3.4073E+04, 3.4095E+04, 3.4116E+04, 3.4137E+04,
     + 3.4159E+04, 3.4180E+04, 3.4202E+04, 3.4223E+04, 3.4244E+04,
     +  3.4266E+04, 3.4287E+04, 3.4309E+04, 3.4330E+04, 3.4352E+04,
     + 3.4373E+04, 3.4394E+04, 3.4416E+04, 3.4437E+04, 3.4459E+04,
     +  3.4480E+04, 3.4502E+04, 3.4523E+04, 3.4545E+04, 3.4566E+04,
     + 3.4588E+04, 3.4609E+04, 3.4631E+04, 3.4652E+04, 3.4674E+04,
     +  3.4695E+04, 3.4717E+04, 3.4739E+04, 3.4760E+04, 3.4782E+04,
     + 3.4803E+04, 3.4825E+04, 3.4846E+04, 3.4868E+04, 3.4890E+04,
     +  3.4911E+04, 3.4933E+04, 3.4954E+04, 3.4976E+04, 3.4998E+04,
     + 3.5019E+04, 3.5041E+04, 3.5063E+04, 3.5084E+04, 3.5106E+04,
     +  3.5128E+04, 3.5149E+04, 3.5171E+04, 3.5193E+04, 3.5214E+04,
     + 3.5236E+04, 3.5258E+04, 3.5279E+04, 3.5301E+04, 3.5323E+04,
     +  3.5344E+04, 3.5366E+04, 3.5388E+04, 3.5410E+04, 3.5431E+04,
     + 3.5453E+04, 3.5475E+04, 3.5497E+04, 3.5518E+04, 3.5540E+04,
     +  3.5562E+04, 3.5584E+04, 3.5606E+04, 3.5627E+04, 3.5649E+04,
     + 3.5671E+04, 3.5693E+04, 3.5715E+04, 3.5736E+04, 3.5758E+04,
     +  3.5780E+04, 3.5802E+04, 3.5824E+04, 3.5846E+04, 3.5867E+04,
     + 3.5889E+04, 3.5911E+04, 3.5933E+04, 3.5955E+04, 3.5977E+04,
     +  3.5999E+04, 3.6021E+04, 3.6042E+04, 3.6064E+04, 3.6086E+04,
     + 3.6108E+04, 3.6130E+04, 3.6152E+04, 3.6174E+04, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        21
      data (QofT( 4,J),J=1,2408)/ 1.2000E+01, 1.2080E+01, 1.2614E+01,
     +  1.3704E+01, 1.5167E+01, 1.6833E+01, 1.8608E+01, 2.0442E+01,
     + 2.2311E+01, 2.4203E+01, 2.6111E+01, 2.8029E+01, 2.9955E+01,
     +  3.1887E+01, 3.3824E+01, 3.5765E+01, 3.7709E+01, 3.9656E+01,
     + 4.1604E+01, 4.3555E+01, 4.7460E+01, 5.1370E+01, 5.5283E+01,
     +  5.9198E+01, 6.3116E+01, 6.7036E+01, 7.0957E+01, 7.4879E+01,
     + 7.8803E+01, 8.2727E+01, 8.6652E+01, 9.0577E+01, 9.4504E+01,
     +  9.8431E+01, 1.0236E+02, 1.0629E+02, 1.1021E+02, 1.1414E+02,
     + 1.1807E+02, 1.2200E+02, 1.2593E+02, 1.2986E+02, 1.3379E+02,
     +  1.3772E+02, 1.4165E+02, 1.4558E+02, 1.4951E+02, 1.5344E+02,
     + 1.5738E+02, 1.6131E+02, 1.6524E+02, 1.6917E+02, 1.7310E+02,
     +  1.7703E+02, 1.8097E+02, 1.8490E+02, 1.8883E+02, 1.9277E+02,
     + 1.9670E+02, 2.0063E+02, 2.0457E+02, 2.0850E+02, 2.1243E+02,
     +  2.1637E+02, 2.2030E+02, 2.2423E+02, 2.2817E+02, 2.3210E+02,
     + 2.3604E+02, 2.3997E+02, 2.4391E+02, 2.4784E+02, 2.5178E+02,
     +  2.5571E+02, 2.5965E+02, 2.6358E+02, 2.6752E+02, 2.7145E+02,
     + 2.7539E+02, 2.7933E+02, 2.8326E+02, 2.8720E+02, 2.9113E+02,
     +  2.9507E+02, 2.9901E+02, 3.0294E+02, 3.0688E+02, 3.1082E+02,
     + 3.1476E+02, 3.1869E+02, 3.2263E+02, 3.2657E+02, 3.3050E+02,
     +  3.3444E+02, 3.3838E+02, 3.4232E+02, 3.4626E+02, 3.5019E+02,
     + 3.5413E+02, 3.5807E+02, 3.6201E+02, 3.6595E+02, 3.6989E+02,
     +  3.7382E+02, 3.7776E+02, 3.8170E+02, 3.8564E+02, 3.8958E+02,
     + 3.9352E+02, 3.9746E+02, 4.0140E+02, 4.0534E+02, 4.0928E+02,
     +  4.1322E+02, 4.1716E+02, 4.2110E+02, 4.2504E+02, 4.2898E+02,
     + 4.3292E+02, 4.3686E+02, 4.4080E+02, 4.4474E+02, 4.4869E+02,
     +  4.5263E+02, 4.5657E+02, 4.6051E+02, 4.6445E+02, 4.6839E+02,
     + 4.7234E+02, 4.7628E+02, 4.8022E+02, 4.8416E+02, 4.8811E+02,
     +  4.9205E+02, 4.9599E+02, 4.9994E+02, 5.0388E+02, 5.0783E+02,
     + 5.1177E+02, 5.1571E+02, 5.1966E+02, 5.2360E+02, 5.2755E+02,
     +  5.3149E+02, 5.3544E+02, 5.3939E+02, 5.4333E+02, 5.4728E+02,
     + 5.5123E+02, 5.5517E+02, 5.5912E+02, 5.6307E+02, 5.6702E+02,
     +  5.7096E+02, 5.7491E+02, 5.7886E+02, 5.8281E+02, 5.8676E+02,
     + 5.9071E+02, 5.9466E+02, 5.9861E+02, 6.0256E+02, 6.0652E+02,
     +  6.1047E+02, 6.1442E+02, 6.1838E+02, 6.2233E+02, 6.2628E+02,
     + 6.3024E+02, 6.3419E+02, 6.3815E+02, 6.4211E+02, 6.4606E+02,
     +  6.5002E+02, 6.5398E+02, 6.5794E+02, 6.6190E+02, 6.6586E+02,
     + 6.6982E+02, 6.7378E+02, 6.7774E+02, 6.8171E+02, 6.8567E+02,
     +  6.8964E+02, 6.9360E+02, 6.9757E+02, 7.0153E+02, 7.0550E+02,
     + 7.0947E+02, 7.1344E+02, 7.1741E+02, 7.2138E+02, 7.2535E+02,
     +  7.2933E+02, 7.3330E+02, 7.3728E+02, 7.4125E+02, 7.4523E+02,
     + 7.4921E+02, 7.5319E+02, 7.5717E+02, 7.6115E+02, 7.6513E+02,
     +  7.6912E+02, 7.7310E+02, 7.7709E+02, 7.8107E+02, 7.8506E+02,
     + 7.8905E+02, 7.9304E+02, 7.9704E+02, 8.0103E+02, 8.0503E+02,
     +  8.0902E+02, 8.1302E+02, 8.1702E+02, 8.2102E+02, 8.2502E+02,
     + 8.2903E+02, 8.3303E+02, 8.3704E+02, 8.4105E+02, 8.4506E+02,
     +  8.4907E+02, 8.5308E+02, 8.5709E+02, 8.6111E+02, 8.6513E+02,
     + 8.6915E+02, 8.7317E+02, 8.7719E+02, 8.8122E+02, 8.8524E+02,
     +  8.8927E+02, 8.9330E+02, 8.9734E+02, 9.0137E+02, 9.0541E+02,
     + 9.0944E+02, 9.1348E+02, 9.1753E+02, 9.2157E+02, 9.2562E+02,
     +  9.2966E+02, 9.3371E+02, 9.3777E+02, 9.4182E+02, 9.4588E+02,
     + 9.4993E+02, 9.5400E+02, 9.5806E+02, 9.6212E+02, 9.6619E+02,
     +  9.7026E+02, 9.7433E+02, 9.7841E+02, 9.8248E+02, 9.8656E+02,
     + 9.9064E+02, 9.9473E+02, 9.9881E+02, 1.0029E+03, 1.0070E+03,
     +  1.0111E+03, 1.0152E+03, 1.0193E+03, 1.0234E+03, 1.0275E+03,
     + 1.0316E+03, 1.0357E+03, 1.0398E+03, 1.0439E+03, 1.0480E+03,
     +  1.0522E+03, 1.0563E+03, 1.0604E+03, 1.0645E+03, 1.0687E+03,
     + 1.0728E+03, 1.0769E+03, 1.0811E+03, 1.0852E+03, 1.0894E+03,
     +  1.0935E+03, 1.0977E+03, 1.1018E+03, 1.1060E+03, 1.1101E+03,
     + 1.1143E+03, 1.1185E+03, 1.1226E+03, 1.1268E+03, 1.1310E+03,
     +  1.1352E+03, 1.1394E+03, 1.1435E+03, 1.1477E+03, 1.1519E+03,
     + 1.1561E+03, 1.1603E+03, 1.1645E+03, 1.1687E+03, 1.1729E+03,
     +  1.1771E+03, 1.1814E+03, 1.1856E+03, 1.1898E+03, 1.1940E+03,
     + 1.1982E+03, 1.2025E+03, 1.2067E+03, 1.2110E+03, 1.2152E+03,
     +  1.2194E+03, 1.2237E+03, 1.2279E+03, 1.2322E+03, 1.2365E+03,
     + 1.2407E+03, 1.2450E+03, 1.2493E+03, 1.2535E+03, 1.2578E+03,
     +  1.2621E+03, 1.2664E+03, 1.2707E+03, 1.2750E+03, 1.2793E+03,
     + 1.2836E+03, 1.2879E+03, 1.2922E+03, 1.2965E+03, 1.3008E+03,
     +  1.3051E+03, 1.3095E+03, 1.3138E+03, 1.3181E+03, 1.3224E+03,
     + 1.3268E+03, 1.3311E+03, 1.3355E+03, 1.3398E+03, 1.3442E+03,
     +  1.3485E+03, 1.3529E+03, 1.3573E+03, 1.3616E+03, 1.3660E+03,
     + 1.3704E+03, 1.3748E+03, 1.3792E+03, 1.3836E+03, 1.3880E+03,
     +  1.3924E+03, 1.3968E+03, 1.4012E+03, 1.4056E+03, 1.4100E+03,
     + 1.4144E+03, 1.4189E+03, 1.4233E+03, 1.4277E+03, 1.4322E+03,
     +  1.4366E+03, 1.4411E+03, 1.4455E+03, 1.4500E+03, 1.4544E+03,
     + 1.4589E+03, 1.4634E+03, 1.4678E+03, 1.4723E+03, 1.4768E+03,
     +  1.4813E+03, 1.4858E+03, 1.4903E+03, 1.4948E+03, 1.4993E+03,
     + 1.5038E+03, 1.5083E+03, 1.5128E+03, 1.5173E+03, 1.5219E+03,
     +  1.5264E+03, 1.5309E+03, 1.5355E+03, 1.5400E+03, 1.5446E+03,
     + 1.5491E+03, 1.5537E+03, 1.5583E+03, 1.5628E+03, 1.5674E+03,
     +  1.5720E+03, 1.5766E+03, 1.5812E+03, 1.5857E+03, 1.5903E+03,
     + 1.5949E+03, 1.5996E+03, 1.6042E+03, 1.6088E+03, 1.6134E+03,
     +  1.6180E+03, 1.6227E+03, 1.6273E+03, 1.6319E+03, 1.6366E+03,
     + 1.6412E+03, 1.6459E+03, 1.6506E+03, 1.6552E+03, 1.6599E+03,
     +  1.6646E+03, 1.6692E+03, 1.6739E+03, 1.6786E+03, 1.6833E+03,
     + 1.6880E+03, 1.6927E+03, 1.6974E+03, 1.7021E+03, 1.7069E+03,
     +  1.7116E+03, 1.7163E+03, 1.7211E+03, 1.7258E+03, 1.7305E+03,
     + 1.7353E+03, 1.7400E+03, 1.7448E+03, 1.7496E+03, 1.7543E+03,
     +  1.7591E+03, 1.7639E+03, 1.7687E+03, 1.7735E+03, 1.7783E+03,
     + 1.7831E+03, 1.7879E+03, 1.7927E+03, 1.7975E+03, 1.8023E+03,
     +  1.8071E+03, 1.8120E+03, 1.8168E+03, 1.8217E+03, 1.8265E+03,
     + 1.8314E+03, 1.8362E+03, 1.8411E+03, 1.8459E+03, 1.8508E+03,
     +  1.8557E+03, 1.8606E+03, 1.8655E+03, 1.8704E+03, 1.8753E+03,
     + 1.8802E+03, 1.8851E+03, 1.8900E+03, 1.8949E+03, 1.8998E+03,
     +  1.9048E+03, 1.9097E+03, 1.9147E+03, 1.9196E+03, 1.9246E+03,
     + 1.9295E+03, 1.9345E+03, 1.9395E+03, 1.9444E+03, 1.9494E+03,
     +  1.9544E+03, 1.9594E+03, 1.9644E+03, 1.9694E+03, 1.9744E+03,
     + 1.9794E+03, 1.9844E+03, 1.9894E+03, 1.9945E+03, 1.9995E+03,
     +  2.0045E+03, 2.0096E+03, 2.0146E+03, 2.0197E+03, 2.0248E+03,
     + 2.0298E+03, 2.0349E+03, 2.0400E+03, 2.0451E+03, 2.0502E+03,
     +  2.0552E+03, 2.0603E+03, 2.0655E+03, 2.0706E+03, 2.0757E+03,
     + 2.0808E+03, 2.0859E+03, 2.0911E+03, 2.0962E+03, 2.1014E+03,
     +  2.1065E+03, 2.1117E+03, 2.1168E+03, 2.1220E+03, 2.1272E+03,
     + 2.1323E+03, 2.1375E+03, 2.1427E+03, 2.1479E+03, 2.1531E+03,
     +  2.1583E+03, 2.1635E+03, 2.1687E+03, 2.1740E+03, 2.1792E+03,
     + 2.1844E+03, 2.1897E+03, 2.1949E+03, 2.2002E+03, 2.2054E+03,
     +  2.2107E+03, 2.2160E+03, 2.2212E+03, 2.2265E+03, 2.2318E+03,
     + 2.2371E+03, 2.2424E+03, 2.2477E+03, 2.2530E+03, 2.2583E+03,
     +  2.2636E+03, 2.2690E+03, 2.2743E+03, 2.2796E+03, 2.2850E+03,
     + 2.2903E+03, 2.2957E+03, 2.3010E+03, 2.3064E+03, 2.3118E+03,
     +  2.3172E+03, 2.3225E+03, 2.3279E+03, 2.3333E+03, 2.3387E+03,
     + 2.3441E+03, 2.3496E+03, 2.3550E+03, 2.3604E+03, 2.3658E+03,
     +  2.3713E+03, 2.3767E+03, 2.3822E+03, 2.3876E+03, 2.3931E+03,
     + 2.3985E+03, 2.4040E+03, 2.4095E+03, 2.4150E+03, 2.4205E+03,
     +  2.4260E+03, 2.4315E+03, 2.4370E+03, 2.4425E+03, 2.4480E+03,
     + 2.4535E+03, 2.4590E+03, 2.4646E+03, 2.4701E+03, 2.4757E+03,
     +  2.4812E+03, 2.4868E+03, 2.4924E+03, 2.4979E+03, 2.5035E+03,
     + 2.5091E+03, 2.5147E+03, 2.5203E+03, 2.5259E+03, 2.5315E+03,
     +  2.5371E+03, 2.5427E+03, 2.5483E+03, 2.5540E+03, 2.5596E+03,
     + 2.5653E+03, 2.5709E+03, 2.5766E+03, 2.5822E+03, 2.5879E+03,
     +  2.5936E+03, 2.5992E+03, 2.6049E+03, 2.6106E+03, 2.6163E+03,
     + 2.6220E+03, 2.6277E+03, 2.6334E+03, 2.6392E+03, 2.6449E+03,
     +  2.6506E+03, 2.6564E+03, 2.6621E+03, 2.6679E+03, 2.6736E+03,
     + 2.6794E+03, 2.6851E+03, 2.6909E+03, 2.6967E+03, 2.7025E+03,
     +  2.7083E+03, 2.7141E+03, 2.7199E+03, 2.7257E+03, 2.7315E+03,
     + 2.7373E+03, 2.7432E+03, 2.7490E+03, 2.7549E+03, 2.7607E+03,
     +  2.7666E+03, 2.7724E+03, 2.7783E+03, 2.7842E+03, 2.7900E+03,
     + 2.7959E+03, 2.8018E+03, 2.8077E+03, 2.8136E+03, 2.8195E+03,
     +  2.8254E+03, 2.8314E+03, 2.8373E+03, 2.8432E+03, 2.8492E+03,
     + 2.8551E+03, 2.8611E+03, 2.8670E+03, 2.8730E+03, 2.8790E+03,
     +  2.8849E+03, 2.8909E+03, 2.8969E+03, 2.9029E+03, 2.9089E+03,
     + 2.9149E+03, 2.9209E+03, 2.9269E+03, 2.9330E+03, 2.9390E+03,
     +  2.9450E+03, 2.9511E+03, 2.9571E+03, 2.9632E+03, 2.9693E+03,
     + 2.9753E+03, 2.9814E+03, 2.9875E+03, 2.9936E+03, 2.9997E+03,
     +  3.0058E+03, 3.0119E+03, 3.0180E+03, 3.0241E+03, 3.0302E+03,
     + 3.0364E+03, 3.0425E+03, 3.0486E+03, 3.0548E+03, 3.0610E+03,
     +  3.0671E+03, 3.0733E+03, 3.0795E+03, 3.0856E+03, 3.0918E+03,
     + 3.0980E+03, 3.1042E+03, 3.1104E+03, 3.1166E+03, 3.1229E+03,
     +  3.1291E+03, 3.1353E+03, 3.1416E+03, 3.1478E+03, 3.1540E+03,
     + 3.1603E+03, 3.1666E+03, 3.1728E+03, 3.1791E+03, 3.1854E+03,
     +  3.1917E+03, 3.1980E+03, 3.2043E+03, 3.2106E+03, 3.2169E+03,
     + 3.2232E+03, 3.2295E+03, 3.2359E+03, 3.2422E+03, 3.2486E+03,
     +  3.2549E+03, 3.2613E+03, 3.2676E+03, 3.2740E+03, 3.2804E+03,
     + 3.2868E+03, 3.2931E+03, 3.2995E+03, 3.3059E+03, 3.3123E+03,
     +  3.3188E+03, 3.3252E+03, 3.3316E+03, 3.3380E+03, 3.3445E+03,
     + 3.3509E+03, 3.3574E+03, 3.3638E+03, 3.3703E+03, 3.3768E+03,
     +  3.3832E+03, 3.3897E+03, 3.3962E+03, 3.4027E+03, 3.4092E+03,
     + 3.4157E+03, 3.4222E+03, 3.4288E+03, 3.4353E+03, 3.4418E+03,
     +  3.4484E+03, 3.4549E+03, 3.4615E+03, 3.4680E+03, 3.4746E+03,
     + 3.4812E+03, 3.4878E+03, 3.4943E+03, 3.5009E+03, 3.5075E+03,
     +  3.5141E+03, 3.5207E+03, 3.5274E+03, 3.5340E+03, 3.5406E+03,
     + 3.5473E+03, 3.5539E+03, 3.5605E+03, 3.5672E+03, 3.5739E+03,
     +  3.5805E+03, 3.5872E+03, 3.5939E+03, 3.6006E+03, 3.6073E+03,
     + 3.6140E+03, 3.6207E+03, 3.6274E+03, 3.6341E+03, 3.6408E+03,
     +  3.6476E+03, 3.6543E+03, 3.6611E+03, 3.6678E+03, 3.6746E+03,
     + 3.6813E+03, 3.6881E+03, 3.6949E+03, 3.7017E+03, 3.7085E+03,
     +  3.7153E+03, 3.7221E+03, 3.7289E+03, 3.7357E+03, 3.7425E+03,
     + 3.7493E+03, 3.7562E+03, 3.7630E+03, 3.7699E+03, 3.7767E+03,
     +  3.7836E+03, 3.7905E+03, 3.7973E+03, 3.8042E+03, 3.8111E+03,
     + 3.8180E+03, 3.8249E+03, 3.8318E+03, 3.8387E+03, 3.8456E+03,
     +  3.8526E+03, 3.8595E+03, 3.8664E+03, 3.8734E+03, 3.8803E+03,
     + 3.8873E+03, 3.8943E+03, 3.9012E+03, 3.9082E+03, 3.9152E+03,
     +  3.9222E+03, 3.9292E+03, 3.9362E+03, 3.9432E+03, 3.9502E+03,
     + 3.9572E+03, 3.9643E+03, 3.9713E+03, 3.9784E+03, 3.9854E+03,
     +  3.9925E+03, 3.9995E+03, 4.0066E+03, 4.0137E+03, 4.0208E+03,
     + 4.0278E+03, 4.0349E+03, 4.0420E+03, 4.0491E+03, 4.0563E+03,
     +  4.0634E+03, 4.0705E+03, 4.0776E+03, 4.0848E+03, 4.0919E+03,
     + 4.0991E+03, 4.1063E+03, 4.1134E+03, 4.1206E+03, 4.1278E+03,
     +  4.1350E+03, 4.1422E+03, 4.1494E+03, 4.1566E+03, 4.1638E+03,
     + 4.1710E+03, 4.1782E+03, 4.1855E+03, 4.1927E+03, 4.1999E+03,
     +  4.2072E+03, 4.2145E+03, 4.2217E+03, 4.2290E+03, 4.2363E+03,
     + 4.2436E+03, 4.2509E+03, 4.2582E+03, 4.2655E+03, 4.2728E+03,
     +  4.2801E+03, 4.2874E+03, 4.2948E+03, 4.3021E+03, 4.3094E+03,
     + 4.3168E+03, 4.3241E+03, 4.3315E+03, 4.3389E+03, 4.3463E+03,
     +  4.3536E+03, 4.3610E+03, 4.3684E+03, 4.3758E+03, 4.3833E+03,
     + 4.3907E+03, 4.3981E+03, 4.4055E+03, 4.4130E+03, 4.4204E+03,
     +  4.4279E+03, 4.4353E+03, 4.4428E+03, 4.4503E+03, 4.4577E+03,
     + 4.4652E+03, 4.4727E+03, 4.4802E+03, 4.4877E+03, 4.4952E+03,
     +  4.5027E+03, 4.5103E+03, 4.5178E+03, 4.5253E+03, 4.5329E+03,
     + 4.5404E+03, 4.5480E+03, 4.5555E+03, 4.5631E+03, 4.5707E+03,
     +  4.5783E+03, 4.5859E+03, 4.5935E+03, 4.6011E+03, 4.6087E+03,
     + 4.6163E+03, 4.6239E+03, 4.6315E+03, 4.6392E+03, 4.6468E+03,
     +  4.6545E+03, 4.6621E+03, 4.6698E+03, 4.6775E+03, 4.6851E+03,
     + 4.6928E+03, 4.7005E+03, 4.7082E+03, 4.7159E+03, 4.7236E+03,
     +  4.7313E+03, 4.7391E+03, 4.7468E+03, 4.7545E+03, 4.7623E+03,
     + 4.7700E+03, 4.7778E+03, 4.7855E+03, 4.7933E+03, 4.8011E+03,
     +  4.8089E+03, 4.8167E+03, 4.8245E+03, 4.8323E+03, 4.8401E+03,
     + 4.8479E+03, 4.8557E+03, 4.8636E+03, 4.8714E+03, 4.8792E+03,
     +  4.8871E+03, 4.8949E+03, 4.9028E+03, 4.9107E+03, 4.9186E+03,
     + 4.9264E+03, 4.9343E+03, 4.9422E+03, 4.9501E+03, 4.9580E+03,
     +  4.9660E+03, 4.9739E+03, 4.9818E+03, 4.9898E+03, 4.9977E+03,
     + 5.0057E+03, 5.0136E+03, 5.0216E+03, 5.0295E+03, 5.0375E+03,
     +  5.0455E+03, 5.0535E+03, 5.0615E+03, 5.0695E+03, 5.0775E+03,
     + 5.0855E+03, 5.0936E+03, 5.1016E+03, 5.1096E+03, 5.1177E+03,
     +  5.1257E+03, 5.1338E+03, 5.1419E+03, 5.1499E+03, 5.1580E+03,
     + 5.1661E+03, 5.1742E+03, 5.1823E+03, 5.1904E+03, 5.1985E+03,
     +  5.2066E+03, 5.2148E+03, 5.2229E+03, 5.2310E+03, 5.2392E+03,
     + 5.2473E+03, 5.2555E+03, 5.2637E+03, 5.2718E+03, 5.2800E+03,
     +  5.2882E+03, 5.2964E+03, 5.3046E+03, 5.3128E+03, 5.3210E+03,
     + 5.3292E+03, 5.3375E+03, 5.3457E+03, 5.3539E+03, 5.3622E+03,
     +  5.3704E+03, 5.3787E+03, 5.3870E+03, 5.3953E+03, 5.4035E+03,
     + 5.4118E+03, 5.4201E+03, 5.4284E+03, 5.4367E+03, 5.4451E+03,
     +  5.4534E+03, 5.4617E+03, 5.4700E+03, 5.4784E+03, 5.4867E+03,
     + 5.4951E+03, 5.5035E+03, 5.5118E+03, 5.5202E+03, 5.5286E+03,
     +  5.5370E+03, 5.5454E+03, 5.5538E+03, 5.5622E+03, 5.5706E+03,
     + 5.5790E+03, 5.5875E+03, 5.5959E+03, 5.6044E+03, 5.6128E+03,
     +  5.6213E+03, 5.6297E+03, 5.6382E+03, 5.6467E+03, 5.6552E+03,
     + 5.6637E+03, 5.6722E+03, 5.6807E+03, 5.6892E+03, 5.6977E+03,
     +  5.7062E+03, 5.7148E+03, 5.7233E+03, 5.7319E+03, 5.7404E+03,
     + 5.7490E+03, 5.7576E+03, 5.7661E+03, 5.7747E+03, 5.7833E+03,
     +  5.7919E+03, 5.8005E+03, 5.8091E+03, 5.8177E+03, 5.8264E+03,
     + 5.8350E+03, 5.8436E+03, 5.8523E+03, 5.8609E+03, 5.8696E+03,
     +  5.8782E+03, 5.8869E+03, 5.8956E+03, 5.9043E+03, 5.9130E+03,
     + 5.9217E+03, 5.9304E+03, 5.9391E+03, 5.9478E+03, 5.9565E+03,
     +  5.9653E+03, 5.9740E+03, 5.9827E+03, 5.9915E+03, 6.0003E+03,
     + 6.0090E+03, 6.0178E+03, 6.0266E+03, 6.0354E+03, 6.0442E+03,
     +  6.0530E+03, 6.0618E+03, 6.0706E+03, 6.0794E+03, 6.0882E+03,
     + 6.0971E+03, 6.1059E+03, 6.1148E+03, 6.1236E+03, 6.1325E+03,
     +  6.1414E+03, 6.1502E+03, 6.1591E+03, 6.1680E+03, 6.1769E+03,
     + 6.1858E+03, 6.1947E+03, 6.2037E+03, 6.2126E+03, 6.2215E+03,
     +  6.2305E+03, 6.2394E+03, 6.2484E+03, 6.2573E+03, 6.2663E+03,
     + 6.2753E+03, 6.2842E+03, 6.2932E+03, 6.3022E+03, 6.3112E+03,
     +  6.3202E+03, 6.3292E+03, 6.3383E+03, 6.3473E+03, 6.3563E+03,
     + 6.3654E+03, 6.3744E+03, 6.3835E+03, 6.3926E+03, 6.4016E+03,
     +  6.4107E+03, 6.4198E+03, 6.4289E+03, 6.4380E+03, 6.4471E+03,
     + 6.4562E+03, 6.4653E+03, 6.4744E+03, 6.4836E+03, 6.4927E+03,
     +  6.5019E+03, 6.5110E+03, 6.5202E+03, 6.5293E+03, 6.5385E+03,
     + 6.5477E+03, 6.5569E+03, 6.5661E+03, 6.5753E+03, 6.5845E+03,
     +  6.5937E+03, 6.6029E+03, 6.6122E+03, 6.6214E+03, 6.6307E+03,
     + 6.6399E+03, 6.6492E+03, 6.6584E+03, 6.6677E+03, 6.6770E+03,
     +  6.6863E+03, 6.6956E+03, 6.7049E+03, 6.7142E+03, 6.7235E+03,
     + 6.7328E+03, 6.7421E+03, 6.7515E+03, 6.7608E+03, 6.7702E+03,
     +  6.7795E+03, 6.7889E+03, 6.7982E+03, 6.8076E+03, 6.8170E+03,
     + 6.8264E+03, 6.8358E+03, 6.8452E+03, 6.8546E+03, 6.8640E+03,
     +  6.8734E+03, 6.8829E+03, 6.8923E+03, 6.9018E+03, 6.9112E+03,
     + 6.9207E+03, 6.9301E+03, 6.9396E+03, 6.9491E+03, 6.9586E+03,
     +  6.9681E+03, 6.9776E+03, 6.9871E+03, 6.9966E+03, 7.0061E+03,
     + 7.0157E+03, 7.0252E+03, 7.0347E+03, 7.0443E+03, 7.0539E+03,
     +  7.0634E+03, 7.0730E+03, 7.0826E+03, 7.0922E+03, 7.1017E+03,
     + 7.1113E+03, 7.1210E+03, 7.1306E+03, 7.1402E+03, 7.1498E+03,
     +  7.1594E+03, 7.1691E+03, 7.1787E+03, 7.1884E+03, 7.1981E+03,
     + 7.2077E+03, 7.2174E+03, 7.2271E+03, 7.2368E+03, 7.2465E+03,
     +  7.2562E+03, 7.2659E+03, 7.2756E+03, 7.2853E+03, 7.2951E+03,
     + 7.3048E+03, 7.3146E+03, 7.3243E+03, 7.3341E+03, 7.3438E+03,
     +  7.3536E+03, 7.3634E+03, 7.3732E+03, 7.3830E+03, 7.3928E+03,
     + 7.4026E+03, 7.4124E+03, 7.4222E+03, 7.4321E+03, 7.4419E+03,
     +  7.4517E+03, 7.4616E+03, 7.4715E+03, 7.4813E+03, 7.4912E+03,
     + 7.5011E+03, 7.5110E+03, 7.5209E+03, 7.5308E+03, 7.5407E+03,
     +  7.5506E+03, 7.5605E+03, 7.5704E+03, 7.5804E+03, 7.5903E+03,
     + 7.6003E+03, 7.6102E+03, 7.6202E+03, 7.6302E+03, 7.6401E+03,
     +  7.6501E+03, 7.6601E+03, 7.6701E+03, 7.6801E+03, 7.6901E+03,
     + 7.7002E+03, 7.7102E+03, 7.7202E+03, 7.7303E+03, 7.7403E+03,
     +  7.7504E+03, 7.7604E+03, 7.7705E+03, 7.7806E+03, 7.7907E+03,
     + 7.8008E+03, 7.8109E+03, 7.8210E+03, 7.8311E+03, 7.8412E+03,
     +  7.8513E+03, 7.8615E+03, 7.8716E+03, 7.8817E+03, 7.8919E+03,
     + 7.9021E+03, 7.9122E+03, 7.9224E+03, 7.9326E+03, 7.9428E+03,
     +  7.9530E+03, 7.9632E+03, 7.9734E+03, 7.9836E+03, 7.9938E+03,
     + 8.0041E+03, 8.0143E+03, 8.0245E+03, 8.0348E+03, 8.0451E+03,
     +  8.0553E+03, 8.0656E+03, 8.0759E+03, 8.0862E+03, 8.0965E+03,
     + 8.1068E+03, 8.1171E+03, 8.1274E+03, 8.1377E+03, 8.1481E+03,
     +  8.1584E+03, 8.1687E+03, 8.1791E+03, 8.1895E+03, 8.1998E+03,
     + 8.2102E+03, 8.2206E+03, 8.2310E+03, 8.2414E+03, 8.2518E+03,
     +  8.2622E+03, 8.2726E+03, 8.2830E+03, 8.2934E+03, 8.3039E+03,
     + 8.3143E+03, 8.3248E+03, 8.3352E+03, 8.3457E+03, 8.3562E+03,
     +  8.3667E+03, 8.3771E+03, 8.3876E+03, 8.3981E+03, 8.4086E+03,
     + 8.4192E+03, 8.4297E+03, 8.4402E+03, 8.4508E+03, 8.4613E+03,
     +  8.4718E+03, 8.4824E+03, 8.4930E+03, 8.5035E+03, 8.5141E+03,
     + 8.5247E+03, 8.5353E+03, 8.5459E+03, 8.5565E+03, 8.5671E+03,
     +  8.5777E+03, 8.5884E+03, 8.5990E+03, 8.6097E+03, 8.6203E+03,
     + 8.6310E+03, 8.6416E+03, 8.6523E+03, 8.6630E+03, 8.6737E+03,
     +  8.6844E+03, 8.6951E+03, 8.7058E+03, 8.7165E+03, 8.7272E+03,
     + 8.7379E+03, 8.7487E+03, 8.7594E+03, 8.7702E+03, 8.7809E+03,
     +  8.7917E+03, 8.8025E+03, 8.8132E+03, 8.8240E+03, 8.8348E+03,
     + 8.8456E+03, 8.8564E+03, 8.8672E+03, 8.8780E+03, 8.8889E+03,
     +  8.8997E+03, 8.9106E+03, 8.9214E+03, 8.9323E+03, 8.9431E+03,
     + 8.9540E+03, 8.9649E+03, 8.9758E+03, 8.9867E+03, 8.9975E+03,
     +  9.0085E+03, 9.0194E+03, 9.0303E+03, 9.0412E+03, 9.0522E+03,
     + 9.0631E+03, 9.0740E+03, 9.0850E+03, 9.0960E+03, 9.1069E+03,
     +  9.1179E+03, 9.1289E+03, 9.1399E+03, 9.1509E+03, 9.1619E+03,
     + 9.1729E+03, 9.1839E+03, 9.1950E+03, 9.2060E+03, 9.2170E+03,
     +  9.2281E+03, 9.2391E+03, 9.2502E+03, 9.2613E+03, 9.2723E+03,
     + 9.2834E+03, 9.2945E+03, 9.3056E+03, 9.3167E+03, 9.3278E+03,
     +  9.3390E+03, 9.3501E+03, 9.3612E+03, 9.3724E+03, 9.3835E+03,
     + 9.3947E+03, 9.4058E+03, 9.4170E+03, 9.4282E+03, 9.4394E+03,
     +  9.4506E+03, 9.4618E+03, 9.4730E+03, 9.4842E+03, 9.4954E+03,
     + 9.5066E+03, 9.5179E+03, 9.5291E+03, 9.5404E+03, 9.5516E+03,
     +  9.5629E+03, 9.5741E+03, 9.5854E+03, 9.5967E+03, 9.6080E+03,
     + 9.6193E+03, 9.6306E+03, 9.6419E+03, 9.6532E+03, 9.6646E+03,
     +  9.6759E+03, 9.6872E+03, 9.6986E+03, 9.7099E+03, 9.7213E+03,
     + 9.7327E+03, 9.7441E+03, 9.7554E+03, 9.7668E+03, 9.7782E+03,
     +  9.7896E+03, 9.8010E+03, 9.8125E+03, 9.8239E+03, 9.8353E+03,
     + 9.8468E+03, 9.8582E+03, 9.8697E+03, 9.8811E+03, 9.8926E+03,
     +  9.9041E+03, 9.9156E+03, 9.9271E+03, 9.9386E+03, 9.9501E+03,
     + 9.9616E+03, 9.9731E+03, 9.9846E+03, 9.9962E+03, 1.0008E+04,
     +  1.0019E+04, 1.0031E+04, 1.0042E+04, 1.0054E+04, 1.0066E+04,
     + 1.0077E+04, 1.0089E+04, 1.0100E+04, 1.0112E+04, 1.0124E+04,
     +  1.0135E+04, 1.0147E+04, 1.0158E+04, 1.0170E+04, 1.0182E+04,
     + 1.0193E+04, 1.0205E+04, 1.0217E+04, 1.0228E+04, 1.0240E+04,
     +  1.0252E+04, 1.0263E+04, 1.0275E+04, 1.0287E+04, 1.0299E+04,
     + 1.0310E+04, 1.0322E+04, 1.0334E+04, 1.0346E+04, 1.0357E+04,
     +  1.0369E+04, 1.0381E+04, 1.0393E+04, 1.0404E+04, 1.0416E+04,
     + 1.0428E+04, 1.0440E+04, 1.0452E+04, 1.0463E+04, 1.0475E+04,
     +  1.0487E+04, 1.0499E+04, 1.0511E+04, 1.0523E+04, 1.0535E+04,
     + 1.0546E+04, 1.0558E+04, 1.0570E+04, 1.0582E+04, 1.0594E+04,
     +  1.0606E+04, 1.0618E+04, 1.0630E+04, 1.0642E+04, 1.0654E+04,
     + 1.0666E+04, 1.0677E+04, 1.0689E+04, 1.0701E+04, 1.0713E+04,
     +  1.0725E+04, 1.0737E+04, 1.0749E+04, 1.0761E+04, 1.0773E+04,
     + 1.0785E+04, 1.0797E+04, 1.0809E+04, 1.0821E+04, 1.0833E+04,
     +  1.0846E+04, 1.0858E+04, 1.0870E+04, 1.0882E+04, 1.0894E+04,
     + 1.0906E+04, 1.0918E+04, 1.0930E+04, 1.0942E+04, 1.0954E+04,
     +  1.0966E+04, 1.0979E+04, 1.0991E+04, 1.1003E+04, 1.1015E+04,
     + 1.1027E+04, 1.1039E+04, 1.1051E+04, 1.1064E+04, 1.1076E+04,
     +  1.1088E+04, 1.1100E+04, 1.1112E+04, 1.1125E+04, 1.1137E+04,
     + 1.1149E+04, 1.1161E+04, 1.1174E+04, 1.1186E+04, 1.1198E+04,
     +  1.1210E+04, 1.1223E+04, 1.1235E+04, 1.1247E+04, 1.1260E+04,
     + 1.1272E+04, 1.1284E+04, 1.1296E+04, 1.1309E+04, 1.1321E+04,
     +  1.1333E+04, 1.1346E+04, 1.1358E+04, 1.1371E+04, 1.1383E+04,
     + 1.1395E+04, 1.1408E+04, 1.1420E+04, 1.1432E+04, 1.1445E+04,
     +  1.1457E+04, 1.1470E+04, 1.1482E+04, 1.1494E+04, 1.1507E+04,
     + 1.1519E+04, 1.1532E+04, 1.1544E+04, 1.1557E+04, 1.1569E+04,
     +  1.1582E+04, 1.1594E+04, 1.1607E+04, 1.1619E+04, 1.1632E+04,
     + 1.1644E+04, 1.1657E+04, 1.1669E+04, 1.1682E+04, 1.1694E+04,
     +  1.1707E+04, 1.1719E+04, 1.1732E+04, 1.1745E+04, 1.1757E+04,
     + 1.1770E+04, 1.1782E+04, 1.1795E+04, 1.1808E+04, 1.1820E+04,
     +  1.1833E+04, 1.1845E+04, 1.1858E+04, 1.1871E+04, 1.1883E+04,
     + 1.1896E+04, 1.1909E+04, 1.1921E+04, 1.1934E+04, 1.1947E+04,
     +  1.1960E+04, 1.1972E+04, 1.1985E+04, 1.1998E+04, 1.2010E+04,
     + 1.2023E+04, 1.2036E+04, 1.2049E+04, 1.2061E+04, 1.2074E+04,
     +  1.2087E+04, 1.2100E+04, 1.2112E+04, 1.2125E+04, 1.2138E+04,
     + 1.2151E+04, 1.2164E+04, 1.2176E+04, 1.2189E+04, 1.2202E+04,
     +  1.2215E+04, 1.2228E+04, 1.2241E+04, 1.2254E+04, 1.2266E+04,
     + 1.2279E+04, 1.2292E+04, 1.2305E+04, 1.2318E+04, 1.2331E+04,
     +  1.2344E+04, 1.2357E+04, 1.2370E+04, 1.2383E+04, 1.2395E+04,
     + 1.2408E+04, 1.2421E+04, 1.2434E+04, 1.2447E+04, 1.2460E+04,
     +  1.2473E+04, 1.2486E+04, 1.2499E+04, 1.2512E+04, 1.2525E+04,
     + 1.2538E+04, 1.2551E+04, 1.2564E+04, 1.2577E+04, 1.2590E+04,
     +  1.2604E+04, 1.2617E+04, 1.2630E+04, 1.2643E+04, 1.2656E+04,
     + 1.2669E+04, 1.2682E+04, 1.2695E+04, 1.2708E+04, 1.2721E+04,
     +  1.2734E+04, 1.2748E+04, 1.2761E+04, 1.2774E+04, 1.2787E+04,
     + 1.2800E+04, 1.2813E+04, 1.2827E+04, 1.2840E+04, 1.2853E+04,
     +  1.2866E+04, 1.2879E+04, 1.2893E+04, 1.2906E+04, 1.2919E+04,
     + 1.2932E+04, 1.2946E+04, 1.2959E+04, 1.2972E+04, 1.2985E+04,
     +  1.2999E+04, 1.3012E+04, 1.3025E+04, 1.3038E+04, 1.3052E+04,
     + 1.3065E+04, 1.3078E+04, 1.3092E+04, 1.3105E+04, 1.3118E+04,
     +  1.3132E+04, 1.3145E+04, 1.3158E+04, 1.3172E+04, 1.3185E+04,
     + 1.3199E+04, 1.3212E+04, 1.3225E+04, 1.3239E+04, 1.3252E+04,
     +  1.3266E+04, 1.3279E+04, 1.3292E+04, 1.3306E+04, 1.3319E+04,
     + 1.3333E+04, 1.3346E+04, 1.3360E+04, 1.3373E+04, 1.3387E+04,
     +  1.3400E+04, 1.3414E+04, 1.3427E+04, 1.3441E+04, 1.3454E+04,
     + 1.3468E+04, 1.3481E+04, 1.3495E+04, 1.3508E+04, 1.3522E+04,
     +  1.3535E+04, 1.3549E+04, 1.3562E+04, 1.3576E+04, 1.3590E+04,
     + 1.3603E+04, 1.3617E+04, 1.3630E+04, 1.3644E+04, 1.3658E+04,
     +  1.3671E+04, 1.3685E+04, 1.3699E+04, 1.3712E+04, 1.3726E+04,
     + 1.3740E+04, 1.3753E+04, 1.3767E+04, 1.3781E+04, 1.3794E+04,
     +  1.3808E+04, 1.3822E+04, 1.3835E+04, 1.3849E+04, 1.3863E+04,
     + 1.3877E+04, 1.3890E+04, 1.3904E+04, 1.3918E+04, 1.3932E+04,
     +  1.3945E+04, 1.3959E+04, 1.3973E+04, 1.3987E+04, 1.4001E+04,
     + 1.4014E+04, 1.4028E+04, 1.4042E+04, 1.4056E+04, 1.4070E+04,
     +  1.4084E+04, 1.4097E+04, 1.4111E+04, 1.4125E+04, 1.4139E+04,
     + 1.4153E+04, 1.4167E+04, 1.4181E+04, 1.4195E+04, 1.4208E+04,
     +  1.4222E+04, 1.4236E+04, 1.4250E+04, 1.4264E+04, 1.4278E+04,
     + 1.4292E+04, 1.4306E+04, 1.4320E+04, 1.4334E+04, 1.4348E+04,
     +  1.4362E+04, 1.4376E+04, 1.4390E+04, 1.4404E+04, 1.4418E+04,
     + 1.4432E+04, 1.4446E+04, 1.4460E+04, 1.4474E+04, 1.4488E+04,
     +  1.4502E+04, 1.4516E+04, 1.4530E+04, 1.4544E+04, 1.4559E+04,
     + 1.4573E+04, 1.4587E+04, 1.4601E+04, 1.4615E+04, 1.4629E+04,
     +  1.4643E+04, 1.4657E+04, 1.4671E+04, 1.4686E+04, 1.4700E+04,
     + 1.4714E+04, 1.4728E+04, 1.4742E+04, 1.4757E+04, 1.4771E+04,
     +  1.4785E+04, 1.4799E+04, 1.4813E+04, 1.4828E+04, 1.4842E+04,
     + 1.4856E+04, 1.4870E+04, 1.4885E+04, 1.4899E+04, 1.4913E+04,
     +  1.4927E+04, 1.4942E+04, 1.4956E+04, 1.4970E+04, 1.4984E+04,
     + 1.4999E+04, 1.5013E+04, 1.5027E+04, 1.5042E+04, 1.5056E+04,
     +  1.5070E+04, 1.5085E+04, 1.5099E+04, 1.5114E+04, 1.5128E+04,
     + 1.5142E+04, 1.5157E+04, 1.5171E+04, 1.5185E+04, 1.5200E+04,
     +  1.5214E+04, 1.5229E+04, 1.5243E+04, 1.5258E+04, 1.5272E+04,
     + 1.5286E+04, 1.5301E+04, 1.5315E+04, 1.5330E+04, 1.5344E+04,
     +  1.5359E+04, 1.5373E+04, 1.5388E+04, 1.5402E+04, 1.5417E+04,
     + 1.5431E+04, 1.5446E+04, 1.5460E+04, 1.5475E+04, 1.5490E+04,
     +  1.5504E+04, 1.5519E+04, 1.5533E+04, 1.5548E+04, 1.5562E+04,
     + 1.5577E+04, 1.5592E+04, 1.5606E+04, 1.5621E+04, 1.5635E+04,
     +  1.5650E+04, 1.5665E+04, 1.5679E+04, 1.5694E+04, 1.5709E+04,
     + 1.5723E+04, 1.5738E+04, 1.5753E+04, 1.5767E+04, 1.5782E+04,
     +  1.5797E+04, 1.5812E+04, 1.5826E+04, 1.5841E+04, 1.5856E+04,
     + 1.5870E+04, 1.5885E+04, 1.5900E+04, 1.5915E+04, 1.5929E+04,
     +  1.5944E+04, 1.5959E+04, 1.5974E+04, 1.5989E+04, 1.6003E+04,
     + 1.6018E+04, 1.6033E+04, 1.6048E+04, 1.6063E+04, 1.6077E+04,
     +  1.6092E+04, 1.6107E+04, 1.6122E+04, 1.6137E+04, 1.6152E+04,
     + 1.6167E+04, 1.6182E+04, 1.6196E+04, 1.6211E+04, 1.6226E+04,
     +  1.6241E+04, 1.6256E+04, 1.6271E+04, 1.6286E+04, 1.6301E+04,
     + 1.6316E+04, 1.6331E+04, 1.6346E+04, 1.6361E+04, 1.6376E+04,
     +  1.6391E+04, 1.6406E+04, 1.6421E+04, 1.6436E+04, 1.6451E+04,
     + 1.6466E+04, 1.6481E+04, 1.6496E+04, 1.6511E+04, 1.6526E+04,
     +  1.6541E+04, 1.6556E+04, 1.6571E+04, 1.6586E+04, 1.6601E+04,
     + 1.6616E+04, 1.6631E+04, 1.6646E+04, 1.6662E+04, 1.6677E+04,
     +  1.6692E+04, 1.6707E+04, 1.6722E+04, 1.6737E+04, 1.6752E+04,
     + 1.6768E+04, 1.6783E+04, 1.6798E+04, 1.6813E+04, 1.6828E+04,
     +  1.6844E+04, 1.6859E+04, 1.6874E+04, 1.6889E+04, 1.6904E+04,
     + 1.6920E+04, 1.6935E+04, 1.6950E+04, 1.6965E+04, 1.6981E+04,
     +  1.6996E+04, 1.7011E+04, 1.7026E+04, 1.7042E+04, 1.7057E+04,
     + 1.7072E+04, 1.7088E+04, 1.7103E+04, 1.7118E+04, 1.7134E+04,
     +  1.7149E+04, 1.7164E+04, 1.7180E+04, 1.7195E+04, 1.7210E+04,
     + 1.7226E+04, 1.7241E+04, 1.7256E+04, 1.7272E+04, 1.7287E+04,
     +  1.7303E+04, 1.7318E+04, 1.7334E+04, 1.7349E+04, 1.7364E+04,
     + 1.7380E+04, 1.7395E+04, 1.7411E+04, 1.7426E+04, 1.7442E+04,
     +  1.7457E+04, 1.7473E+04, 1.7488E+04, 1.7504E+04, 1.7519E+04,
     + 1.7535E+04, 1.7550E+04, 1.7566E+04, 1.7581E+04, 1.7597E+04,
     +  1.7612E+04, 1.7628E+04, 1.7643E+04, 1.7659E+04, 1.7675E+04,
     + 1.7690E+04, 1.7706E+04, 1.7721E+04, 1.7737E+04, 1.7753E+04,
     +  1.7768E+04, 1.7784E+04, 1.7799E+04, 1.7815E+04, 1.7831E+04,
     + 1.7846E+04, 1.7862E+04, 1.7878E+04, 1.7893E+04, 1.7909E+04,
     +  1.7925E+04, 1.7941E+04, 1.7956E+04, 1.7972E+04, 1.7988E+04,
     + 1.8003E+04, 1.8019E+04, 1.8035E+04, 1.8051E+04, 1.8066E+04,
     +  1.8082E+04, 1.8098E+04, 1.8114E+04, 1.8129E+04, 1.8145E+04,
     + 1.8161E+04, 1.8177E+04, 1.8193E+04, 1.8208E+04, 1.8224E+04,
     +  1.8240E+04, 1.8256E+04, 1.8272E+04, 1.8288E+04, 1.8303E+04,
     + 1.8319E+04, 1.8335E+04, 1.8351E+04, 1.8367E+04, 1.8383E+04,
     +  1.8399E+04, 1.8415E+04, 1.8431E+04, 1.8447E+04, 1.8462E+04,
     + 1.8478E+04, 1.8494E+04, 1.8510E+04, 1.8526E+04, 1.8542E+04,
     +  1.8558E+04, 1.8574E+04, 1.8590E+04, 1.8606E+04, 1.8622E+04,
     + 1.8638E+04, 1.8654E+04, 1.8670E+04, 1.8686E+04, 1.8702E+04,
     +  1.8718E+04, 1.8734E+04, 1.8750E+04, 1.8766E+04, 1.8782E+04,
     + 1.8799E+04, 1.8815E+04, 1.8831E+04, 1.8847E+04, 1.8863E+04,
     +  1.8879E+04, 1.8895E+04, 1.8911E+04, 1.8927E+04, 1.8944E+04,
     + 1.8960E+04, 1.8976E+04, 1.8992E+04, 1.9008E+04, 1.9024E+04,
     +  1.9041E+04, 1.9057E+04, 1.9073E+04, 1.9089E+04, 1.9105E+04,
     + 1.9122E+04, 1.9138E+04, 1.9154E+04, 1.9170E+04, 1.9186E+04,
     +  1.9203E+04, 1.9219E+04, 1.9235E+04, 1.9251E+04, 1.9268E+04,
     + 1.9284E+04, 1.9300E+04, 1.9317E+04, 1.9333E+04, 1.9349E+04,
     +  1.9366E+04, 1.9382E+04, 1.9398E+04, 1.9415E+04, 1.9431E+04,
     + 1.9447E+04, 1.9464E+04, 1.9480E+04, 1.9496E+04, 1.9513E+04,
     +  1.9529E+04, 1.9546E+04, 1.9562E+04, 1.9578E+04, 1.9595E+04,
     + 1.9611E+04, 1.9628E+04, 1.9644E+04, 1.9660E+04, 1.9677E+04,
     +  1.9693E+04, 1.9710E+04, 1.9726E+04, 1.9743E+04, 1.9759E+04,
     + 1.9776E+04, 1.9792E+04, 1.9809E+04, 1.9825E+04, 1.9842E+04,
     +  1.9858E+04, 1.9875E+04, 1.9891E+04, 1.9908E+04, 1.9925E+04,
     + 1.9941E+04, 1.9958E+04, 1.9974E+04, 1.9991E+04, 2.0007E+04,
     +  2.0024E+04, 2.0041E+04, 2.0057E+04, 2.0074E+04, 2.0090E+04,
     + 2.0107E+04, 2.0124E+04, 2.0140E+04, 2.0157E+04, 2.0174E+04,
     +  2.0190E+04, 2.0207E+04, 2.0224E+04, 2.0240E+04, 2.0257E+04,
     + 2.0274E+04, 2.0290E+04, 2.0307E+04, 2.0324E+04, 2.0341E+04,
     +  2.0357E+04, 2.0374E+04, 2.0391E+04, 2.0408E+04, 2.0424E+04,
     + 2.0441E+04, 2.0458E+04, 2.0475E+04, 2.0491E+04, 2.0508E+04,
     +  2.0525E+04, 2.0542E+04, 2.0559E+04, 2.0576E+04, 2.0592E+04,
     + 2.0609E+04, 2.0626E+04, 2.0643E+04, 2.0660E+04, 2.0677E+04,
     +  2.0693E+04, 2.0710E+04, 2.0727E+04, 2.0744E+04, 2.0761E+04,
     + 2.0778E+04, 2.0795E+04, 2.0812E+04, 2.0829E+04, 2.0846E+04,
     +  2.0863E+04, 2.0880E+04, 2.0896E+04, 2.0913E+04, 2.0930E+04,
     + 2.0947E+04, 2.0964E+04, 2.0981E+04, 2.0998E+04, 2.1015E+04,
     +  2.1032E+04, 2.1049E+04, 2.1066E+04, 2.1083E+04, 2.1100E+04,
     + 2.1118E+04, 2.1135E+04, 2.1152E+04, 2.1169E+04, 2.1186E+04,
     +  2.1203E+04, 2.1220E+04, 2.1237E+04, 2.1254E+04, 2.1271E+04,
     + 2.1288E+04, 2.1305E+04, 2.1323E+04, 2.1340E+04, 2.1357E+04,
     +  2.1374E+04, 2.1391E+04, 2.1408E+04, 2.1425E+04, 2.1443E+04,
     + 2.1460E+04, 2.1477E+04, 2.1494E+04, 2.1511E+04, 2.1529E+04,
     +  2.1546E+04, 2.1563E+04, 2.1580E+04, 2.1598E+04, 2.1615E+04,
     + 2.1632E+04, 2.1649E+04, 2.1667E+04, 2.1684E+04, 2.1701E+04,
     +  2.1718E+04, 2.1736E+04, 2.1753E+04, 2.1770E+04, 2.1788E+04,
     + 2.1805E+04, 2.1822E+04, 2.1840E+04, 2.1857E+04, 2.1874E+04,
     +  2.1892E+04, 2.1909E+04, 2.1926E+04, 2.1944E+04, 2.1961E+04,
     + 2.1978E+04, 2.1996E+04, 2.2013E+04, 2.2031E+04, 2.2048E+04,
     +  2.2065E+04, 2.2083E+04, 2.2100E+04, 2.2118E+04, 2.2135E+04,
     + 2.2153E+04, 2.2170E+04, 2.2188E+04, 2.2205E+04, 2.2223E+04,
     +  2.2240E+04, 2.2257E+04, 2.2275E+04, 2.2293E+04, 2.2310E+04,
     + 2.2328E+04, 2.2345E+04, 2.2363E+04, 2.2380E+04, 2.2398E+04,
     +  2.2415E+04, 2.2433E+04, 2.2450E+04, 2.2468E+04, 2.2486E+04,
     + 2.2503E+04, 2.2521E+04, 2.2538E+04, 2.2556E+04, 2.2574E+04,
     +  2.2591E+04, 2.2609E+04, 2.2626E+04, 2.2644E+04, 2.2662E+04,
     + 2.2679E+04, 2.2697E+04, 2.2715E+04, 2.2732E+04, 2.2750E+04,
     +  2.2768E+04, 2.2785E+04, 2.2803E+04, 2.2821E+04, 2.2839E+04,
     + 2.2856E+04, 2.2874E+04, 2.2892E+04, 2.2910E+04, 2.2927E+04,
     +  2.2945E+04, 2.2963E+04, 2.2981E+04, 2.2998E+04, 2.3016E+04,
     + 2.3034E+04, 2.3052E+04, 2.3070E+04, 2.3087E+04, 2.3105E+04,
     +  2.3123E+04, 2.3141E+04, 2.3159E+04, 2.3177E+04, 2.3194E+04,
     + 2.3212E+04, 2.3230E+04, 2.3248E+04, 2.3266E+04, 2.3284E+04,
     +  2.3302E+04, 2.3320E+04, 2.3337E+04, 2.3355E+04, 2.3373E+04,
     + 2.3391E+04, 2.3409E+04, 2.3427E+04, 2.3445E+04, 2.3463E+04,
     +  2.3481E+04, 2.3499E+04, 2.3517E+04, 2.3535E+04, 2.3553E+04,
     + 2.3571E+04, 2.3589E+04, 2.3607E+04, 2.3625E+04, 2.3643E+04,
     +  2.3661E+04, 2.3679E+04, 2.3697E+04, 2.3715E+04, 2.3733E+04,
     + 2.3751E+04, 2.3769E+04, 2.3787E+04, 2.3805E+04, 2.3823E+04,
     +  2.3842E+04, 2.3860E+04, 2.3878E+04, 2.3896E+04, 2.3914E+04,
     + 2.3932E+04, 2.3950E+04, 2.3968E+04, 2.3987E+04, 2.4005E+04,
     +  2.4023E+04, 2.4041E+04, 2.4059E+04, 2.4077E+04, 2.4096E+04/
      data (QofT( 4,J),J=2409,4510)/ 2.4114E+04, 2.4132E+04,
     +  2.4150E+04, 2.4168E+04, 2.4187E+04, 2.4205E+04, 2.4223E+04,
     +  2.4241E+04,
     + 2.4260E+04, 2.4278E+04, 2.4296E+04, 2.4314E+04, 2.4333E+04,
     +  2.4351E+04, 2.4369E+04, 2.4387E+04, 2.4406E+04, 2.4424E+04,
     + 2.4442E+04, 2.4461E+04, 2.4479E+04, 2.4497E+04, 2.4516E+04,
     +  2.4534E+04, 2.4552E+04, 2.4571E+04, 2.4589E+04, 2.4608E+04,
     + 2.4626E+04, 2.4644E+04, 2.4663E+04, 2.4681E+04, 2.4699E+04,
     +  2.4718E+04, 2.4736E+04, 2.4755E+04, 2.4773E+04, 2.4792E+04,
     + 2.4810E+04, 2.4829E+04, 2.4847E+04, 2.4865E+04, 2.4884E+04,
     +  2.4902E+04, 2.4921E+04, 2.4939E+04, 2.4958E+04, 2.4976E+04,
     + 2.4995E+04, 2.5013E+04, 2.5032E+04, 2.5050E+04, 2.5069E+04,
     +  2.5088E+04, 2.5106E+04, 2.5125E+04, 2.5143E+04, 2.5162E+04,
     + 2.5180E+04, 2.5199E+04, 2.5218E+04, 2.5236E+04, 2.5255E+04,
     +  2.5273E+04, 2.5292E+04, 2.5311E+04, 2.5329E+04, 2.5348E+04,
     + 2.5367E+04, 2.5385E+04, 2.5404E+04, 2.5423E+04, 2.5441E+04,
     +  2.5460E+04, 2.5479E+04, 2.5497E+04, 2.5516E+04, 2.5535E+04,
     + 2.5553E+04, 2.5572E+04, 2.5591E+04, 2.5610E+04, 2.5628E+04,
     +  2.5647E+04, 2.5666E+04, 2.5685E+04, 2.5703E+04, 2.5722E+04,
     + 2.5741E+04, 2.5760E+04, 2.5778E+04, 2.5797E+04, 2.5816E+04,
     +  2.5835E+04, 2.5854E+04, 2.5873E+04, 2.5891E+04, 2.5910E+04,
     + 2.5929E+04, 2.5948E+04, 2.5967E+04, 2.5986E+04, 2.6004E+04,
     +  2.6023E+04, 2.6042E+04, 2.6061E+04, 2.6080E+04, 2.6099E+04,
     + 2.6118E+04, 2.6137E+04, 2.6156E+04, 2.6175E+04, 2.6194E+04,
     +  2.6212E+04, 2.6231E+04, 2.6250E+04, 2.6269E+04, 2.6288E+04,
     + 2.6307E+04, 2.6326E+04, 2.6345E+04, 2.6364E+04, 2.6383E+04,
     +  2.6402E+04, 2.6421E+04, 2.6440E+04, 2.6459E+04, 2.6478E+04,
     + 2.6497E+04, 2.6516E+04, 2.6535E+04, 2.6555E+04, 2.6574E+04,
     +  2.6593E+04, 2.6612E+04, 2.6631E+04, 2.6650E+04, 2.6669E+04,
     + 2.6688E+04, 2.6707E+04, 2.6726E+04, 2.6746E+04, 2.6765E+04,
     +  2.6784E+04, 2.6803E+04, 2.6822E+04, 2.6841E+04, 2.6860E+04,
     + 2.6880E+04, 2.6899E+04, 2.6918E+04, 2.6937E+04, 2.6956E+04,
     +  2.6976E+04, 2.6995E+04, 2.7014E+04, 2.7033E+04, 2.7052E+04,
     + 2.7072E+04, 2.7091E+04, 2.7110E+04, 2.7129E+04, 2.7149E+04,
     +  2.7168E+04, 2.7187E+04, 2.7206E+04, 2.7226E+04, 2.7245E+04,
     + 2.7264E+04, 2.7284E+04, 2.7303E+04, 2.7322E+04, 2.7342E+04,
     +  2.7361E+04, 2.7380E+04, 2.7400E+04, 2.7419E+04, 2.7438E+04,
     + 2.7458E+04, 2.7477E+04, 2.7496E+04, 2.7516E+04, 2.7535E+04,
     +  2.7555E+04, 2.7574E+04, 2.7593E+04, 2.7613E+04, 2.7632E+04,
     + 2.7652E+04, 2.7671E+04, 2.7691E+04, 2.7710E+04, 2.7730E+04,
     +  2.7749E+04, 2.7768E+04, 2.7788E+04, 2.7807E+04, 2.7827E+04,
     + 2.7846E+04, 2.7866E+04, 2.7885E+04, 2.7905E+04, 2.7924E+04,
     +  2.7944E+04, 2.7964E+04, 2.7983E+04, 2.8003E+04, 2.8022E+04,
     + 2.8042E+04, 2.8061E+04, 2.8081E+04, 2.8100E+04, 2.8120E+04,
     +  2.8140E+04, 2.8159E+04, 2.8179E+04, 2.8198E+04, 2.8218E+04,
     + 2.8238E+04, 2.8257E+04, 2.8277E+04, 2.8297E+04, 2.8316E+04,
     +  2.8336E+04, 2.8356E+04, 2.8375E+04, 2.8395E+04, 2.8415E+04,
     + 2.8434E+04, 2.8454E+04, 2.8474E+04, 2.8493E+04, 2.8513E+04,
     +  2.8533E+04, 2.8553E+04, 2.8572E+04, 2.8592E+04, 2.8612E+04,
     + 2.8632E+04, 2.8651E+04, 2.8671E+04, 2.8691E+04, 2.8711E+04,
     +  2.8730E+04, 2.8750E+04, 2.8770E+04, 2.8790E+04, 2.8810E+04,
     + 2.8829E+04, 2.8849E+04, 2.8869E+04, 2.8889E+04, 2.8909E+04,
     +  2.8929E+04, 2.8948E+04, 2.8968E+04, 2.8988E+04, 2.9008E+04,
     + 2.9028E+04, 2.9048E+04, 2.9068E+04, 2.9088E+04, 2.9107E+04,
     +  2.9127E+04, 2.9147E+04, 2.9167E+04, 2.9187E+04, 2.9207E+04,
     + 2.9227E+04, 2.9247E+04, 2.9267E+04, 2.9287E+04, 2.9307E+04,
     +  2.9327E+04, 2.9347E+04, 2.9367E+04, 2.9387E+04, 2.9407E+04,
     + 2.9427E+04, 2.9447E+04, 2.9467E+04, 2.9487E+04, 2.9507E+04,
     +  2.9527E+04, 2.9547E+04, 2.9567E+04, 2.9587E+04, 2.9607E+04,
     + 2.9627E+04, 2.9647E+04, 2.9667E+04, 2.9688E+04, 2.9708E+04,
     +  2.9728E+04, 2.9748E+04, 2.9768E+04, 2.9788E+04, 2.9808E+04,
     + 2.9828E+04, 2.9848E+04, 2.9869E+04, 2.9889E+04, 2.9909E+04,
     +  2.9929E+04, 2.9949E+04, 2.9969E+04, 2.9990E+04, 3.0010E+04,
     + 3.0030E+04, 3.0050E+04, 3.0070E+04, 3.0091E+04, 3.0111E+04,
     +  3.0131E+04, 3.0151E+04, 3.0172E+04, 3.0192E+04, 3.0212E+04,
     + 3.0232E+04, 3.0253E+04, 3.0273E+04, 3.0293E+04, 3.0313E+04,
     +  3.0334E+04, 3.0354E+04, 3.0374E+04, 3.0395E+04, 3.0415E+04,
     + 3.0435E+04, 3.0456E+04, 3.0476E+04, 3.0496E+04, 3.0517E+04,
     +  3.0537E+04, 3.0557E+04, 3.0578E+04, 3.0598E+04, 3.0618E+04,
     + 3.0639E+04, 3.0659E+04, 3.0680E+04, 3.0700E+04, 3.0720E+04,
     +  3.0741E+04, 3.0761E+04, 3.0782E+04, 3.0802E+04, 3.0823E+04,
     + 3.0843E+04, 3.0864E+04, 3.0884E+04, 3.0904E+04, 3.0925E+04,
     +  3.0945E+04, 3.0966E+04, 3.0986E+04, 3.1007E+04, 3.1027E+04,
     + 3.1048E+04, 3.1068E+04, 3.1089E+04, 3.1109E+04, 3.1130E+04,
     +  3.1151E+04, 3.1171E+04, 3.1192E+04, 3.1212E+04, 3.1233E+04,
     + 3.1253E+04, 3.1274E+04, 3.1295E+04, 3.1315E+04, 3.1336E+04,
     +  3.1356E+04, 3.1377E+04, 3.1398E+04, 3.1418E+04, 3.1439E+04,
     + 3.1459E+04, 3.1480E+04, 3.1501E+04, 3.1521E+04, 3.1542E+04,
     +  3.1563E+04, 3.1583E+04, 3.1604E+04, 3.1625E+04, 3.1645E+04,
     + 3.1666E+04, 3.1687E+04, 3.1708E+04, 3.1728E+04, 3.1749E+04,
     +  3.1770E+04, 3.1790E+04, 3.1811E+04, 3.1832E+04, 3.1853E+04,
     + 3.1873E+04, 3.1894E+04, 3.1915E+04, 3.1936E+04, 3.1957E+04,
     +  3.1977E+04, 3.1998E+04, 3.2019E+04, 3.2040E+04, 3.2061E+04,
     + 3.2081E+04, 3.2102E+04, 3.2123E+04, 3.2144E+04, 3.2165E+04,
     +  3.2186E+04, 3.2207E+04, 3.2227E+04, 3.2248E+04, 3.2269E+04,
     + 3.2290E+04, 3.2311E+04, 3.2332E+04, 3.2353E+04, 3.2374E+04,
     +  3.2395E+04, 3.2415E+04, 3.2436E+04, 3.2457E+04, 3.2478E+04,
     + 3.2499E+04, 3.2520E+04, 3.2541E+04, 3.2562E+04, 3.2583E+04,
     +  3.2604E+04, 3.2625E+04, 3.2646E+04, 3.2667E+04, 3.2688E+04,
     + 3.2709E+04, 3.2730E+04, 3.2751E+04, 3.2772E+04, 3.2793E+04,
     +  3.2814E+04, 3.2835E+04, 3.2856E+04, 3.2877E+04, 3.2898E+04,
     + 3.2919E+04, 3.2940E+04, 3.2962E+04, 3.2983E+04, 3.3004E+04,
     +  3.3025E+04, 3.3046E+04, 3.3067E+04, 3.3088E+04, 3.3109E+04,
     + 3.3130E+04, 3.3152E+04, 3.3173E+04, 3.3194E+04, 3.3215E+04,
     +  3.3236E+04, 3.3257E+04, 3.3279E+04, 3.3300E+04, 3.3321E+04,
     + 3.3342E+04, 3.3363E+04, 3.3384E+04, 3.3406E+04, 3.3427E+04,
     +  3.3448E+04, 3.3469E+04, 3.3491E+04, 3.3512E+04, 3.3533E+04,
     + 3.3554E+04, 3.3576E+04, 3.3597E+04, 3.3618E+04, 3.3639E+04,
     +  3.3661E+04, 3.3682E+04, 3.3703E+04, 3.3725E+04, 3.3746E+04,
     + 3.3767E+04, 3.3788E+04, 3.3810E+04, 3.3831E+04, 3.3852E+04,
     +  3.3874E+04, 3.3895E+04, 3.3917E+04, 3.3938E+04, 3.3959E+04,
     + 3.3981E+04, 3.4002E+04, 3.4023E+04, 3.4045E+04, 3.4066E+04,
     +  3.4088E+04, 3.4109E+04, 3.4130E+04, 3.4152E+04, 3.4173E+04,
     + 3.4195E+04, 3.4216E+04, 3.4238E+04, 3.4259E+04, 3.4281E+04,
     +  3.4302E+04, 3.4323E+04, 3.4345E+04, 3.4366E+04, 3.4388E+04,
     + 3.4409E+04, 3.4431E+04, 3.4452E+04, 3.4474E+04, 3.4495E+04,
     +  3.4517E+04, 3.4538E+04, 3.4560E+04, 3.4582E+04, 3.4603E+04,
     + 3.4625E+04, 3.4646E+04, 3.4668E+04, 3.4689E+04, 3.4711E+04,
     +  3.4733E+04, 3.4754E+04, 3.4776E+04, 3.4797E+04, 3.4819E+04,
     + 3.4841E+04, 3.4862E+04, 3.4884E+04, 3.4905E+04, 3.4927E+04,
     +  3.4949E+04, 3.4970E+04, 3.4992E+04, 3.5014E+04, 3.5035E+04,
     + 3.5057E+04, 3.5079E+04, 3.5100E+04, 3.5122E+04, 3.5144E+04,
     +  3.5166E+04, 3.5187E+04, 3.5209E+04, 3.5231E+04, 3.5252E+04,
     + 3.5274E+04, 3.5296E+04, 3.5318E+04, 3.5339E+04, 3.5361E+04,
     +  3.5383E+04, 3.5405E+04, 3.5426E+04, 3.5448E+04, 3.5470E+04,
     + 3.5492E+04, 3.5514E+04, 3.5535E+04, 3.5557E+04, 3.5579E+04,
     +  3.5601E+04, 3.5623E+04, 3.5645E+04, 3.5666E+04, 3.5688E+04,
     + 3.5710E+04, 3.5732E+04, 3.5754E+04, 3.5776E+04, 3.5798E+04,
     +  3.5819E+04, 3.5841E+04, 3.5863E+04, 3.5885E+04, 3.5907E+04,
     + 3.5929E+04, 3.5951E+04, 3.5973E+04, 3.5995E+04, 3.6017E+04,
     +  3.6039E+04, 3.6061E+04, 3.6083E+04, 3.6104E+04, 3.6126E+04,
     + 3.6148E+04, 3.6170E+04, 3.6192E+04, 3.6214E+04, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_HI    (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 2),QofT( 2,NT),Q(NT)
      data (Tmx(I),I=1,2)/6000.,6000./
      data xgj/ 12.,18./
c...       HI
c...        --        17
      data (QofT( 1,J),J=1,2408)/ 1.2000E+01, 1.2003E+01, 1.2076E+01,
     +  1.2354E+01, 1.2893E+01, 1.3657E+01, 1.4587E+01, 1.5627E+01,
     + 1.6740E+01, 1.7901E+01, 1.9094E+01, 2.0309E+01, 2.1539E+01,
     +  2.2782E+01, 2.4033E+01, 2.5291E+01, 2.6554E+01, 2.7822E+01,
     + 2.9093E+01, 3.0368E+01, 3.2923E+01, 3.5487E+01, 3.8056E+01,
     +  4.0629E+01, 4.3206E+01, 4.5785E+01, 4.8367E+01, 5.0950E+01,
     + 5.3535E+01, 5.6122E+01, 5.8709E+01, 6.1298E+01, 6.3887E+01,
     +  6.6477E+01, 6.9068E+01, 7.1660E+01, 7.4252E+01, 7.6844E+01,
     + 7.9437E+01, 8.2030E+01, 8.4624E+01, 8.7218E+01, 8.9812E+01,
     +  9.2407E+01, 9.5002E+01, 9.7597E+01, 1.0019E+02, 1.0279E+02,
     + 1.0538E+02, 1.0798E+02, 1.1058E+02, 1.1317E+02, 1.1577E+02,
     +  1.1837E+02, 1.2096E+02, 1.2356E+02, 1.2616E+02, 1.2875E+02,
     + 1.3135E+02, 1.3395E+02, 1.3655E+02, 1.3915E+02, 1.4174E+02,
     +  1.4434E+02, 1.4694E+02, 1.4954E+02, 1.5214E+02, 1.5474E+02,
     + 1.5734E+02, 1.5993E+02, 1.6253E+02, 1.6513E+02, 1.6773E+02,
     +  1.7033E+02, 1.7293E+02, 1.7553E+02, 1.7813E+02, 1.8073E+02,
     + 1.8333E+02, 1.8593E+02, 1.8853E+02, 1.9113E+02, 1.9373E+02,
     +  1.9633E+02, 1.9893E+02, 2.0153E+02, 2.0413E+02, 2.0673E+02,
     + 2.0933E+02, 2.1193E+02, 2.1453E+02, 2.1714E+02, 2.1974E+02,
     +  2.2234E+02, 2.2494E+02, 2.2754E+02, 2.3014E+02, 2.3274E+02,
     + 2.3535E+02, 2.3795E+02, 2.4055E+02, 2.4315E+02, 2.4575E+02,
     +  2.4835E+02, 2.5096E+02, 2.5356E+02, 2.5616E+02, 2.5876E+02,
     + 2.6137E+02, 2.6397E+02, 2.6657E+02, 2.6917E+02, 2.7178E+02,
     +  2.7438E+02, 2.7698E+02, 2.7959E+02, 2.8219E+02, 2.8479E+02,
     + 2.8740E+02, 2.9000E+02, 2.9260E+02, 2.9521E+02, 2.9781E+02,
     +  3.0041E+02, 3.0302E+02, 3.0562E+02, 3.0822E+02, 3.1083E+02,
     + 3.1343E+02, 3.1604E+02, 3.1864E+02, 3.2125E+02, 3.2385E+02,
     +  3.2645E+02, 3.2906E+02, 3.3166E+02, 3.3427E+02, 3.3687E+02,
     + 3.3948E+02, 3.4208E+02, 3.4469E+02, 3.4729E+02, 3.4990E+02,
     +  3.5250E+02, 3.5511E+02, 3.5772E+02, 3.6032E+02, 3.6293E+02,
     + 3.6553E+02, 3.6814E+02, 3.7074E+02, 3.7335E+02, 3.7596E+02,
     +  3.7856E+02, 3.8117E+02, 3.8378E+02, 3.8638E+02, 3.8899E+02,
     + 3.9160E+02, 3.9420E+02, 3.9681E+02, 3.9942E+02, 4.0203E+02,
     +  4.0463E+02, 4.0724E+02, 4.0985E+02, 4.1246E+02, 4.1507E+02,
     + 4.1767E+02, 4.2028E+02, 4.2289E+02, 4.2550E+02, 4.2811E+02,
     +  4.3072E+02, 4.3333E+02, 4.3594E+02, 4.3855E+02, 4.4116E+02,
     + 4.4377E+02, 4.4638E+02, 4.4899E+02, 4.5160E+02, 4.5421E+02,
     +  4.5682E+02, 4.5943E+02, 4.6204E+02, 4.6465E+02, 4.6726E+02,
     + 4.6988E+02, 4.7249E+02, 4.7510E+02, 4.7771E+02, 4.8033E+02,
     +  4.8294E+02, 4.8555E+02, 4.8817E+02, 4.9078E+02, 4.9339E+02,
     + 4.9601E+02, 4.9862E+02, 5.0124E+02, 5.0385E+02, 5.0647E+02,
     +  5.0909E+02, 5.1170E+02, 5.1432E+02, 5.1694E+02, 5.1955E+02,
     + 5.2217E+02, 5.2479E+02, 5.2741E+02, 5.3003E+02, 5.3265E+02,
     +  5.3526E+02, 5.3788E+02, 5.4050E+02, 5.4313E+02, 5.4575E+02,
     + 5.4837E+02, 5.5099E+02, 5.5361E+02, 5.5624E+02, 5.5886E+02,
     +  5.6148E+02, 5.6411E+02, 5.6673E+02, 5.6936E+02, 5.7198E+02,
     + 5.7461E+02, 5.7723E+02, 5.7986E+02, 5.8249E+02, 5.8512E+02,
     +  5.8775E+02, 5.9038E+02, 5.9301E+02, 5.9564E+02, 5.9827E+02,
     + 6.0090E+02, 6.0353E+02, 6.0616E+02, 6.0880E+02, 6.1143E+02,
     +  6.1407E+02, 6.1670E+02, 6.1934E+02, 6.2198E+02, 6.2461E+02,
     + 6.2725E+02, 6.2989E+02, 6.3253E+02, 6.3517E+02, 6.3781E+02,
     +  6.4045E+02, 6.4309E+02, 6.4574E+02, 6.4838E+02, 6.5103E+02,
     + 6.5367E+02, 6.5632E+02, 6.5897E+02, 6.6161E+02, 6.6426E+02,
     +  6.6691E+02, 6.6956E+02, 6.7221E+02, 6.7487E+02, 6.7752E+02,
     + 6.8017E+02, 6.8283E+02, 6.8548E+02, 6.8814E+02, 6.9080E+02,
     +  6.9346E+02, 6.9612E+02, 6.9878E+02, 7.0144E+02, 7.0410E+02,
     + 7.0676E+02, 7.0943E+02, 7.1209E+02, 7.1476E+02, 7.1743E+02,
     +  7.2009E+02, 7.2276E+02, 7.2543E+02, 7.2811E+02, 7.3078E+02,
     + 7.3345E+02, 7.3613E+02, 7.3880E+02, 7.4148E+02, 7.4416E+02,
     +  7.4684E+02, 7.4952E+02, 7.5220E+02, 7.5488E+02, 7.5757E+02,
     + 7.6025E+02, 7.6294E+02, 7.6563E+02, 7.6832E+02, 7.7101E+02,
     +  7.7370E+02, 7.7639E+02, 7.7909E+02, 7.8178E+02, 7.8448E+02,
     + 7.8718E+02, 7.8988E+02, 7.9258E+02, 7.9528E+02, 7.9798E+02,
     +  8.0069E+02, 8.0339E+02, 8.0610E+02, 8.0881E+02, 8.1152E+02,
     + 8.1423E+02, 8.1694E+02, 8.1966E+02, 8.2237E+02, 8.2509E+02,
     +  8.2781E+02, 8.3053E+02, 8.3325E+02, 8.3597E+02, 8.3870E+02,
     + 8.4143E+02, 8.4415E+02, 8.4688E+02, 8.4961E+02, 8.5235E+02,
     +  8.5508E+02, 8.5782E+02, 8.6055E+02, 8.6329E+02, 8.6603E+02,
     + 8.6877E+02, 8.7152E+02, 8.7426E+02, 8.7701E+02, 8.7976E+02,
     +  8.8251E+02, 8.8526E+02, 8.8801E+02, 8.9077E+02, 8.9352E+02,
     + 8.9628E+02, 8.9904E+02, 9.0180E+02, 9.0457E+02, 9.0733E+02,
     +  9.1010E+02, 9.1287E+02, 9.1564E+02, 9.1841E+02, 9.2119E+02,
     + 9.2396E+02, 9.2674E+02, 9.2952E+02, 9.3230E+02, 9.3508E+02,
     +  9.3787E+02, 9.4066E+02, 9.4344E+02, 9.4623E+02, 9.4903E+02,
     + 9.5182E+02, 9.5462E+02, 9.5742E+02, 9.6022E+02, 9.6302E+02,
     +  9.6582E+02, 9.6863E+02, 9.7143E+02, 9.7424E+02, 9.7706E+02,
     + 9.7987E+02, 9.8269E+02, 9.8550E+02, 9.8832E+02, 9.9114E+02,
     +  9.9397E+02, 9.9679E+02, 9.9962E+02, 1.0025E+03, 1.0053E+03,
     + 1.0081E+03, 1.0110E+03, 1.0138E+03, 1.0166E+03, 1.0195E+03,
     +  1.0223E+03, 1.0252E+03, 1.0280E+03, 1.0309E+03, 1.0337E+03,
     + 1.0366E+03, 1.0394E+03, 1.0423E+03, 1.0451E+03, 1.0480E+03,
     +  1.0509E+03, 1.0537E+03, 1.0566E+03, 1.0595E+03, 1.0624E+03,
     + 1.0652E+03, 1.0681E+03, 1.0710E+03, 1.0739E+03, 1.0768E+03,
     +  1.0796E+03, 1.0825E+03, 1.0854E+03, 1.0883E+03, 1.0912E+03,
     + 1.0941E+03, 1.0970E+03, 1.0999E+03, 1.1028E+03, 1.1057E+03,
     +  1.1086E+03, 1.1116E+03, 1.1145E+03, 1.1174E+03, 1.1203E+03,
     + 1.1232E+03, 1.1262E+03, 1.1291E+03, 1.1320E+03, 1.1349E+03,
     +  1.1379E+03, 1.1408E+03, 1.1438E+03, 1.1467E+03, 1.1496E+03,
     + 1.1526E+03, 1.1555E+03, 1.1585E+03, 1.1614E+03, 1.1644E+03,
     +  1.1673E+03, 1.1703E+03, 1.1733E+03, 1.1762E+03, 1.1792E+03,
     + 1.1822E+03, 1.1851E+03, 1.1881E+03, 1.1911E+03, 1.1941E+03,
     +  1.1971E+03, 1.2000E+03, 1.2030E+03, 1.2060E+03, 1.2090E+03,
     + 1.2120E+03, 1.2150E+03, 1.2180E+03, 1.2210E+03, 1.2240E+03,
     +  1.2270E+03, 1.2300E+03, 1.2330E+03, 1.2361E+03, 1.2391E+03,
     + 1.2421E+03, 1.2451E+03, 1.2481E+03, 1.2512E+03, 1.2542E+03,
     +  1.2572E+03, 1.2603E+03, 1.2633E+03, 1.2664E+03, 1.2694E+03,
     + 1.2724E+03, 1.2755E+03, 1.2785E+03, 1.2816E+03, 1.2847E+03,
     +  1.2877E+03, 1.2908E+03, 1.2938E+03, 1.2969E+03, 1.3000E+03,
     + 1.3031E+03, 1.3061E+03, 1.3092E+03, 1.3123E+03, 1.3154E+03,
     +  1.3185E+03, 1.3215E+03, 1.3246E+03, 1.3277E+03, 1.3308E+03,
     + 1.3339E+03, 1.3370E+03, 1.3401E+03, 1.3432E+03, 1.3464E+03,
     +  1.3495E+03, 1.3526E+03, 1.3557E+03, 1.3588E+03, 1.3620E+03,
     + 1.3651E+03, 1.3682E+03, 1.3713E+03, 1.3745E+03, 1.3776E+03,
     +  1.3808E+03, 1.3839E+03, 1.3871E+03, 1.3902E+03, 1.3934E+03,
     + 1.3965E+03, 1.3997E+03, 1.4028E+03, 1.4060E+03, 1.4092E+03,
     +  1.4123E+03, 1.4155E+03, 1.4187E+03, 1.4219E+03, 1.4250E+03,
     + 1.4282E+03, 1.4314E+03, 1.4346E+03, 1.4378E+03, 1.4410E+03,
     +  1.4442E+03, 1.4474E+03, 1.4506E+03, 1.4538E+03, 1.4570E+03,
     + 1.4602E+03, 1.4634E+03, 1.4667E+03, 1.4699E+03, 1.4731E+03,
     +  1.4763E+03, 1.4796E+03, 1.4828E+03, 1.4860E+03, 1.4893E+03,
     + 1.4925E+03, 1.4958E+03, 1.4990E+03, 1.5023E+03, 1.5055E+03,
     +  1.5088E+03, 1.5120E+03, 1.5153E+03, 1.5186E+03, 1.5218E+03,
     + 1.5251E+03, 1.5284E+03, 1.5317E+03, 1.5349E+03, 1.5382E+03,
     +  1.5415E+03, 1.5448E+03, 1.5481E+03, 1.5514E+03, 1.5547E+03,
     + 1.5580E+03, 1.5613E+03, 1.5646E+03, 1.5679E+03, 1.5712E+03,
     +  1.5745E+03, 1.5779E+03, 1.5812E+03, 1.5845E+03, 1.5878E+03,
     + 1.5912E+03, 1.5945E+03, 1.5978E+03, 1.6012E+03, 1.6045E+03,
     +  1.6079E+03, 1.6112E+03, 1.6146E+03, 1.6179E+03, 1.6213E+03,
     + 1.6247E+03, 1.6280E+03, 1.6314E+03, 1.6348E+03, 1.6381E+03,
     +  1.6415E+03, 1.6449E+03, 1.6483E+03, 1.6517E+03, 1.6551E+03,
     + 1.6585E+03, 1.6618E+03, 1.6652E+03, 1.6686E+03, 1.6721E+03,
     +  1.6755E+03, 1.6789E+03, 1.6823E+03, 1.6857E+03, 1.6891E+03,
     + 1.6926E+03, 1.6960E+03, 1.6994E+03, 1.7028E+03, 1.7063E+03,
     +  1.7097E+03, 1.7132E+03, 1.7166E+03, 1.7201E+03, 1.7235E+03,
     + 1.7270E+03, 1.7304E+03, 1.7339E+03, 1.7374E+03, 1.7408E+03,
     +  1.7443E+03, 1.7478E+03, 1.7513E+03, 1.7547E+03, 1.7582E+03,
     + 1.7617E+03, 1.7652E+03, 1.7687E+03, 1.7722E+03, 1.7757E+03,
     +  1.7792E+03, 1.7827E+03, 1.7862E+03, 1.7897E+03, 1.7932E+03,
     + 1.7968E+03, 1.8003E+03, 1.8038E+03, 1.8073E+03, 1.8109E+03,
     +  1.8144E+03, 1.8179E+03, 1.8215E+03, 1.8250E+03, 1.8286E+03,
     + 1.8321E+03, 1.8357E+03, 1.8392E+03, 1.8428E+03, 1.8464E+03,
     +  1.8499E+03, 1.8535E+03, 1.8571E+03, 1.8607E+03, 1.8642E+03,
     + 1.8678E+03, 1.8714E+03, 1.8750E+03, 1.8786E+03, 1.8822E+03,
     +  1.8858E+03, 1.8894E+03, 1.8930E+03, 1.8966E+03, 1.9002E+03,
     + 1.9038E+03, 1.9075E+03, 1.9111E+03, 1.9147E+03, 1.9183E+03,
     +  1.9220E+03, 1.9256E+03, 1.9292E+03, 1.9329E+03, 1.9365E+03,
     + 1.9402E+03, 1.9438E+03, 1.9475E+03, 1.9512E+03, 1.9548E+03,
     +  1.9585E+03, 1.9622E+03, 1.9658E+03, 1.9695E+03, 1.9732E+03,
     + 1.9769E+03, 1.9806E+03, 1.9842E+03, 1.9879E+03, 1.9916E+03,
     +  1.9953E+03, 1.9990E+03, 2.0027E+03, 2.0064E+03, 2.0102E+03,
     + 2.0139E+03, 2.0176E+03, 2.0213E+03, 2.0250E+03, 2.0288E+03,
     +  2.0325E+03, 2.0362E+03, 2.0400E+03, 2.0437E+03, 2.0475E+03,
     + 2.0512E+03, 2.0550E+03, 2.0587E+03, 2.0625E+03, 2.0662E+03,
     +  2.0700E+03, 2.0738E+03, 2.0776E+03, 2.0813E+03, 2.0851E+03,
     + 2.0889E+03, 2.0927E+03, 2.0965E+03, 2.1003E+03, 2.1041E+03,
     +  2.1079E+03, 2.1117E+03, 2.1155E+03, 2.1193E+03, 2.1231E+03,
     + 2.1269E+03, 2.1307E+03, 2.1345E+03, 2.1384E+03, 2.1422E+03,
     +  2.1460E+03, 2.1499E+03, 2.1537E+03, 2.1576E+03, 2.1614E+03,
     + 2.1653E+03, 2.1691E+03, 2.1730E+03, 2.1768E+03, 2.1807E+03,
     +  2.1846E+03, 2.1884E+03, 2.1923E+03, 2.1962E+03, 2.2001E+03,
     + 2.2040E+03, 2.2078E+03, 2.2117E+03, 2.2156E+03, 2.2195E+03,
     +  2.2234E+03, 2.2273E+03, 2.2312E+03, 2.2352E+03, 2.2391E+03,
     + 2.2430E+03, 2.2469E+03, 2.2508E+03, 2.2548E+03, 2.2587E+03,
     +  2.2626E+03, 2.2666E+03, 2.2705E+03, 2.2745E+03, 2.2784E+03,
     + 2.2824E+03, 2.2863E+03, 2.2903E+03, 2.2942E+03, 2.2982E+03,
     +  2.3022E+03, 2.3062E+03, 2.3101E+03, 2.3141E+03, 2.3181E+03,
     + 2.3221E+03, 2.3261E+03, 2.3301E+03, 2.3341E+03, 2.3381E+03,
     +  2.3421E+03, 2.3461E+03, 2.3501E+03, 2.3541E+03, 2.3581E+03,
     + 2.3622E+03, 2.3662E+03, 2.3702E+03, 2.3743E+03, 2.3783E+03,
     +  2.3823E+03, 2.3864E+03, 2.3904E+03, 2.3945E+03, 2.3985E+03,
     + 2.4026E+03, 2.4066E+03, 2.4107E+03, 2.4148E+03, 2.4189E+03,
     +  2.4229E+03, 2.4270E+03, 2.4311E+03, 2.4352E+03, 2.4393E+03,
     + 2.4434E+03, 2.4475E+03, 2.4516E+03, 2.4557E+03, 2.4598E+03,
     +  2.4639E+03, 2.4680E+03, 2.4721E+03, 2.4762E+03, 2.4803E+03,
     + 2.4845E+03, 2.4886E+03, 2.4927E+03, 2.4969E+03, 2.5010E+03,
     +  2.5052E+03, 2.5093E+03, 2.5135E+03, 2.5176E+03, 2.5218E+03,
     + 2.5260E+03, 2.5301E+03, 2.5343E+03, 2.5385E+03, 2.5426E+03,
     +  2.5468E+03, 2.5510E+03, 2.5552E+03, 2.5594E+03, 2.5636E+03,
     + 2.5678E+03, 2.5720E+03, 2.5762E+03, 2.5804E+03, 2.5846E+03,
     +  2.5888E+03, 2.5930E+03, 2.5973E+03, 2.6015E+03, 2.6057E+03,
     + 2.6100E+03, 2.6142E+03, 2.6184E+03, 2.6227E+03, 2.6269E+03,
     +  2.6312E+03, 2.6354E+03, 2.6397E+03, 2.6440E+03, 2.6482E+03,
     + 2.6525E+03, 2.6568E+03, 2.6611E+03, 2.6653E+03, 2.6696E+03,
     +  2.6739E+03, 2.6782E+03, 2.6825E+03, 2.6868E+03, 2.6911E+03,
     + 2.6954E+03, 2.6997E+03, 2.7040E+03, 2.7083E+03, 2.7127E+03,
     +  2.7170E+03, 2.7213E+03, 2.7256E+03, 2.7300E+03, 2.7343E+03,
     + 2.7387E+03, 2.7430E+03, 2.7473E+03, 2.7517E+03, 2.7561E+03,
     +  2.7604E+03, 2.7648E+03, 2.7691E+03, 2.7735E+03, 2.7779E+03,
     + 2.7823E+03, 2.7866E+03, 2.7910E+03, 2.7954E+03, 2.7998E+03,
     +  2.8042E+03, 2.8086E+03, 2.8130E+03, 2.8174E+03, 2.8218E+03,
     + 2.8262E+03, 2.8307E+03, 2.8351E+03, 2.8395E+03, 2.8439E+03,
     +  2.8484E+03, 2.8528E+03, 2.8572E+03, 2.8617E+03, 2.8661E+03,
     + 2.8706E+03, 2.8750E+03, 2.8795E+03, 2.8840E+03, 2.8884E+03,
     +  2.8929E+03, 2.8974E+03, 2.9018E+03, 2.9063E+03, 2.9108E+03,
     + 2.9153E+03, 2.9198E+03, 2.9243E+03, 2.9288E+03, 2.9333E+03,
     +  2.9378E+03, 2.9423E+03, 2.9468E+03, 2.9513E+03, 2.9558E+03,
     + 2.9603E+03, 2.9649E+03, 2.9694E+03, 2.9739E+03, 2.9785E+03,
     +  2.9830E+03, 2.9876E+03, 2.9921E+03, 2.9967E+03, 3.0012E+03,
     + 3.0058E+03, 3.0103E+03, 3.0149E+03, 3.0195E+03, 3.0240E+03,
     +  3.0286E+03, 3.0332E+03, 3.0378E+03, 3.0424E+03, 3.0470E+03,
     + 3.0516E+03, 3.0562E+03, 3.0608E+03, 3.0654E+03, 3.0700E+03,
     +  3.0746E+03, 3.0792E+03, 3.0838E+03, 3.0885E+03, 3.0931E+03,
     + 3.0977E+03, 3.1024E+03, 3.1070E+03, 3.1117E+03, 3.1163E+03,
     +  3.1210E+03, 3.1256E+03, 3.1303E+03, 3.1349E+03, 3.1396E+03,
     + 3.1443E+03, 3.1489E+03, 3.1536E+03, 3.1583E+03, 3.1630E+03,
     +  3.1677E+03, 3.1724E+03, 3.1771E+03, 3.1818E+03, 3.1865E+03,
     + 3.1912E+03, 3.1959E+03, 3.2006E+03, 3.2053E+03, 3.2100E+03,
     +  3.2148E+03, 3.2195E+03, 3.2242E+03, 3.2290E+03, 3.2337E+03,
     + 3.2384E+03, 3.2432E+03, 3.2479E+03, 3.2527E+03, 3.2574E+03,
     +  3.2622E+03, 3.2670E+03, 3.2717E+03, 3.2765E+03, 3.2813E+03,
     + 3.2861E+03, 3.2909E+03, 3.2956E+03, 3.3004E+03, 3.3052E+03,
     +  3.3100E+03, 3.3148E+03, 3.3196E+03, 3.3245E+03, 3.3293E+03,
     + 3.3341E+03, 3.3389E+03, 3.3437E+03, 3.3486E+03, 3.3534E+03,
     +  3.3582E+03, 3.3631E+03, 3.3679E+03, 3.3728E+03, 3.3776E+03,
     + 3.3825E+03, 3.3873E+03, 3.3922E+03, 3.3971E+03, 3.4019E+03,
     +  3.4068E+03, 3.4117E+03, 3.4166E+03, 3.4214E+03, 3.4263E+03,
     + 3.4312E+03, 3.4361E+03, 3.4410E+03, 3.4459E+03, 3.4508E+03,
     +  3.4558E+03, 3.4607E+03, 3.4656E+03, 3.4705E+03, 3.4754E+03,
     + 3.4804E+03, 3.4853E+03, 3.4902E+03, 3.4952E+03, 3.5001E+03,
     +  3.5051E+03, 3.5100E+03, 3.5150E+03, 3.5199E+03, 3.5249E+03,
     + 3.5299E+03, 3.5348E+03, 3.5398E+03, 3.5448E+03, 3.5498E+03,
     +  3.5548E+03, 3.5598E+03, 3.5648E+03, 3.5697E+03, 3.5747E+03,
     + 3.5798E+03, 3.5848E+03, 3.5898E+03, 3.5948E+03, 3.5998E+03,
     +  3.6048E+03, 3.6099E+03, 3.6149E+03, 3.6199E+03, 3.6250E+03,
     + 3.6300E+03, 3.6351E+03, 3.6401E+03, 3.6452E+03, 3.6502E+03,
     +  3.6553E+03, 3.6604E+03, 3.6654E+03, 3.6705E+03, 3.6756E+03,
     + 3.6807E+03, 3.6857E+03, 3.6908E+03, 3.6959E+03, 3.7010E+03,
     +  3.7061E+03, 3.7112E+03, 3.7163E+03, 3.7214E+03, 3.7265E+03,
     + 3.7317E+03, 3.7368E+03, 3.7419E+03, 3.7470E+03, 3.7522E+03,
     +  3.7573E+03, 3.7625E+03, 3.7676E+03, 3.7728E+03, 3.7779E+03,
     + 3.7831E+03, 3.7882E+03, 3.7934E+03, 3.7986E+03, 3.8037E+03,
     +  3.8089E+03, 3.8141E+03, 3.8193E+03, 3.8244E+03, 3.8296E+03,
     + 3.8348E+03, 3.8400E+03, 3.8452E+03, 3.8504E+03, 3.8556E+03,
     +  3.8609E+03, 3.8661E+03, 3.8713E+03, 3.8765E+03, 3.8818E+03,
     + 3.8870E+03, 3.8922E+03, 3.8975E+03, 3.9027E+03, 3.9080E+03,
     +  3.9132E+03, 3.9185E+03, 3.9237E+03, 3.9290E+03, 3.9342E+03,
     + 3.9395E+03, 3.9448E+03, 3.9501E+03, 3.9554E+03, 3.9606E+03,
     +  3.9659E+03, 3.9712E+03, 3.9765E+03, 3.9818E+03, 3.9871E+03,
     + 3.9924E+03, 3.9977E+03, 4.0031E+03, 4.0084E+03, 4.0137E+03,
     +  4.0190E+03, 4.0244E+03, 4.0297E+03, 4.0350E+03, 4.0404E+03,
     + 4.0457E+03, 4.0511E+03, 4.0564E+03, 4.0618E+03, 4.0672E+03,
     +  4.0725E+03, 4.0779E+03, 4.0833E+03, 4.0886E+03, 4.0940E+03,
     + 4.0994E+03, 4.1048E+03, 4.1102E+03, 4.1156E+03, 4.1210E+03,
     +  4.1264E+03, 4.1318E+03, 4.1372E+03, 4.1426E+03, 4.1480E+03,
     + 4.1535E+03, 4.1589E+03, 4.1643E+03, 4.1698E+03, 4.1752E+03,
     +  4.1806E+03, 4.1861E+03, 4.1915E+03, 4.1970E+03, 4.2025E+03,
     + 4.2079E+03, 4.2134E+03, 4.2189E+03, 4.2243E+03, 4.2298E+03,
     +  4.2353E+03, 4.2408E+03, 4.2463E+03, 4.2518E+03, 4.2572E+03,
     + 4.2627E+03, 4.2683E+03, 4.2738E+03, 4.2793E+03, 4.2848E+03,
     +  4.2903E+03, 4.2958E+03, 4.3014E+03, 4.3069E+03, 4.3124E+03,
     + 4.3180E+03, 4.3235E+03, 4.3291E+03, 4.3346E+03, 4.3402E+03,
     +  4.3457E+03, 4.3513E+03, 4.3568E+03, 4.3624E+03, 4.3680E+03,
     + 4.3736E+03, 4.3791E+03, 4.3847E+03, 4.3903E+03, 4.3959E+03,
     +  4.4015E+03, 4.4071E+03, 4.4127E+03, 4.4183E+03, 4.4239E+03,
     + 4.4296E+03, 4.4352E+03, 4.4408E+03, 4.4464E+03, 4.4521E+03,
     +  4.4577E+03, 4.4633E+03, 4.4690E+03, 4.4746E+03, 4.4803E+03,
     + 4.4859E+03, 4.4916E+03, 4.4972E+03, 4.5029E+03, 4.5086E+03,
     +  4.5143E+03, 4.5199E+03, 4.5256E+03, 4.5313E+03, 4.5370E+03,
     + 4.5427E+03, 4.5484E+03, 4.5541E+03, 4.5598E+03, 4.5655E+03,
     +  4.5712E+03, 4.5769E+03, 4.5826E+03, 4.5884E+03, 4.5941E+03,
     + 4.5998E+03, 4.6056E+03, 4.6113E+03, 4.6170E+03, 4.6228E+03,
     +  4.6285E+03, 4.6343E+03, 4.6401E+03, 4.6458E+03, 4.6516E+03,
     + 4.6574E+03, 4.6631E+03, 4.6689E+03, 4.6747E+03, 4.6805E+03,
     +  4.6863E+03, 4.6921E+03, 4.6979E+03, 4.7037E+03, 4.7095E+03,
     + 4.7153E+03, 4.7211E+03, 4.7269E+03, 4.7327E+03, 4.7386E+03,
     +  4.7444E+03, 4.7502E+03, 4.7561E+03, 4.7619E+03, 4.7677E+03,
     + 4.7736E+03, 4.7795E+03, 4.7853E+03, 4.7912E+03, 4.7970E+03,
     +  4.8029E+03, 4.8088E+03, 4.8146E+03, 4.8205E+03, 4.8264E+03,
     + 4.8323E+03, 4.8382E+03, 4.8441E+03, 4.8500E+03, 4.8559E+03,
     +  4.8618E+03, 4.8677E+03, 4.8736E+03, 4.8795E+03, 4.8855E+03,
     + 4.8914E+03, 4.8973E+03, 4.9033E+03, 4.9092E+03, 4.9151E+03,
     +  4.9211E+03, 4.9270E+03, 4.9330E+03, 4.9389E+03, 4.9449E+03,
     + 4.9509E+03, 4.9568E+03, 4.9628E+03, 4.9688E+03, 4.9748E+03,
     +  4.9808E+03, 4.9867E+03, 4.9927E+03, 4.9987E+03, 5.0047E+03,
     + 5.0107E+03, 5.0167E+03, 5.0228E+03, 5.0288E+03, 5.0348E+03,
     +  5.0408E+03, 5.0468E+03, 5.0529E+03, 5.0589E+03, 5.0650E+03,
     + 5.0710E+03, 5.0770E+03, 5.0831E+03, 5.0891E+03, 5.0952E+03,
     +  5.1013E+03, 5.1073E+03, 5.1134E+03, 5.1195E+03, 5.1256E+03,
     + 5.1316E+03, 5.1377E+03, 5.1438E+03, 5.1499E+03, 5.1560E+03,
     +  5.1621E+03, 5.1682E+03, 5.1743E+03, 5.1804E+03, 5.1866E+03,
     + 5.1927E+03, 5.1988E+03, 5.2049E+03, 5.2111E+03, 5.2172E+03,
     +  5.2234E+03, 5.2295E+03, 5.2356E+03, 5.2418E+03, 5.2480E+03,
     + 5.2541E+03, 5.2603E+03, 5.2665E+03, 5.2726E+03, 5.2788E+03,
     +  5.2850E+03, 5.2912E+03, 5.2974E+03, 5.3035E+03, 5.3097E+03,
     + 5.3159E+03, 5.3221E+03, 5.3284E+03, 5.3346E+03, 5.3408E+03,
     +  5.3470E+03, 5.3532E+03, 5.3595E+03, 5.3657E+03, 5.3719E+03,
     + 5.3782E+03, 5.3844E+03, 5.3906E+03, 5.3969E+03, 5.4032E+03,
     +  5.4094E+03, 5.4157E+03, 5.4219E+03, 5.4282E+03, 5.4345E+03,
     + 5.4408E+03, 5.4470E+03, 5.4533E+03, 5.4596E+03, 5.4659E+03,
     +  5.4722E+03, 5.4785E+03, 5.4848E+03, 5.4911E+03, 5.4974E+03,
     + 5.5038E+03, 5.5101E+03, 5.5164E+03, 5.5227E+03, 5.5291E+03,
     +  5.5354E+03, 5.5418E+03, 5.5481E+03, 5.5544E+03, 5.5608E+03,
     + 5.5672E+03, 5.5735E+03, 5.5799E+03, 5.5862E+03, 5.5926E+03,
     +  5.5990E+03, 5.6054E+03, 5.6118E+03, 5.6182E+03, 5.6245E+03,
     + 5.6309E+03, 5.6373E+03, 5.6437E+03, 5.6502E+03, 5.6566E+03,
     +  5.6630E+03, 5.6694E+03, 5.6758E+03, 5.6822E+03, 5.6887E+03,
     + 5.6951E+03, 5.7016E+03, 5.7080E+03, 5.7144E+03, 5.7209E+03,
     +  5.7274E+03, 5.7338E+03, 5.7403E+03, 5.7467E+03, 5.7532E+03,
     + 5.7597E+03, 5.7662E+03, 5.7726E+03, 5.7791E+03, 5.7856E+03,
     +  5.7921E+03, 5.7986E+03, 5.8051E+03, 5.8116E+03, 5.8181E+03,
     + 5.8246E+03, 5.8312E+03, 5.8377E+03, 5.8442E+03, 5.8507E+03,
     +  5.8573E+03, 5.8638E+03, 5.8704E+03, 5.8769E+03, 5.8835E+03,
     + 5.8900E+03, 5.8966E+03, 5.9031E+03, 5.9097E+03, 5.9163E+03,
     +  5.9228E+03, 5.9294E+03, 5.9360E+03, 5.9426E+03, 5.9492E+03,
     + 5.9558E+03, 5.9624E+03, 5.9690E+03, 5.9756E+03, 5.9822E+03,
     +  5.9888E+03, 5.9954E+03, 6.0020E+03, 6.0086E+03, 6.0153E+03,
     + 6.0219E+03, 6.0285E+03, 6.0352E+03, 6.0418E+03, 6.0485E+03,
     +  6.0551E+03, 6.0618E+03, 6.0684E+03, 6.0751E+03, 6.0818E+03,
     + 6.0884E+03, 6.0951E+03, 6.1018E+03, 6.1085E+03, 6.1152E+03,
     +  6.1219E+03, 6.1285E+03, 6.1352E+03, 6.1419E+03, 6.1487E+03,
     + 6.1554E+03, 6.1621E+03, 6.1688E+03, 6.1755E+03, 6.1822E+03,
     +  6.1890E+03, 6.1957E+03, 6.2025E+03, 6.2092E+03, 6.2159E+03,
     + 6.2227E+03, 6.2294E+03, 6.2362E+03, 6.2430E+03, 6.2497E+03,
     +  6.2565E+03, 6.2633E+03, 6.2701E+03, 6.2768E+03, 6.2836E+03,
     + 6.2904E+03, 6.2972E+03, 6.3040E+03, 6.3108E+03, 6.3176E+03,
     +  6.3244E+03, 6.3312E+03, 6.3380E+03, 6.3449E+03, 6.3517E+03,
     + 6.3585E+03, 6.3654E+03, 6.3722E+03, 6.3790E+03, 6.3859E+03,
     +  6.3927E+03, 6.3996E+03, 6.4064E+03, 6.4133E+03, 6.4202E+03,
     + 6.4270E+03, 6.4339E+03, 6.4408E+03, 6.4477E+03, 6.4545E+03,
     +  6.4614E+03, 6.4683E+03, 6.4752E+03, 6.4821E+03, 6.4890E+03,
     + 6.4959E+03, 6.5028E+03, 6.5097E+03, 6.5167E+03, 6.5236E+03,
     +  6.5305E+03, 6.5375E+03, 6.5444E+03, 6.5513E+03, 6.5583E+03,
     + 6.5652E+03, 6.5722E+03, 6.5791E+03, 6.5861E+03, 6.5930E+03,
     +  6.6000E+03, 6.6070E+03, 6.6140E+03, 6.6209E+03, 6.6279E+03,
     + 6.6349E+03, 6.6419E+03, 6.6489E+03, 6.6559E+03, 6.6629E+03,
     +  6.6699E+03, 6.6769E+03, 6.6839E+03, 6.6909E+03, 6.6980E+03,
     + 6.7050E+03, 6.7120E+03, 6.7190E+03, 6.7261E+03, 6.7331E+03,
     +  6.7402E+03, 6.7472E+03, 6.7543E+03, 6.7613E+03, 6.7684E+03,
     + 6.7755E+03, 6.7825E+03, 6.7896E+03, 6.7967E+03, 6.8038E+03,
     +  6.8108E+03, 6.8179E+03, 6.8250E+03, 6.8321E+03, 6.8392E+03,
     + 6.8463E+03, 6.8534E+03, 6.8605E+03, 6.8677E+03, 6.8748E+03,
     +  6.8819E+03, 6.8890E+03, 6.8962E+03, 6.9033E+03, 6.9104E+03,
     + 6.9176E+03, 6.9247E+03, 6.9319E+03, 6.9390E+03, 6.9462E+03,
     +  6.9534E+03, 6.9605E+03, 6.9677E+03, 6.9749E+03, 6.9821E+03,
     + 6.9892E+03, 6.9964E+03, 7.0036E+03, 7.0108E+03, 7.0180E+03,
     +  7.0252E+03, 7.0324E+03, 7.0396E+03, 7.0468E+03, 7.0541E+03,
     + 7.0613E+03, 7.0685E+03, 7.0757E+03, 7.0830E+03, 7.0902E+03,
     +  7.0975E+03, 7.1047E+03, 7.1119E+03, 7.1192E+03, 7.1265E+03,
     + 7.1337E+03, 7.1410E+03, 7.1483E+03, 7.1555E+03, 7.1628E+03,
     +  7.1701E+03, 7.1774E+03, 7.1847E+03, 7.1920E+03, 7.1993E+03,
     + 7.2066E+03, 7.2139E+03, 7.2212E+03, 7.2285E+03, 7.2358E+03,
     +  7.2431E+03, 7.2504E+03, 7.2578E+03, 7.2651E+03, 7.2724E+03,
     + 7.2798E+03, 7.2871E+03, 7.2945E+03, 7.3018E+03, 7.3092E+03,
     +  7.3165E+03, 7.3239E+03, 7.3313E+03, 7.3386E+03, 7.3460E+03,
     + 7.3534E+03, 7.3608E+03, 7.3682E+03, 7.3756E+03, 7.3830E+03,
     +  7.3904E+03, 7.3978E+03, 7.4052E+03, 7.4126E+03, 7.4200E+03,
     + 7.4274E+03, 7.4348E+03, 7.4423E+03, 7.4497E+03, 7.4571E+03,
     +  7.4646E+03, 7.4720E+03, 7.4794E+03, 7.4869E+03, 7.4944E+03,
     + 7.5018E+03, 7.5093E+03, 7.5167E+03, 7.5242E+03, 7.5317E+03,
     +  7.5392E+03, 7.5466E+03, 7.5541E+03, 7.5616E+03, 7.5691E+03,
     + 7.5766E+03, 7.5841E+03, 7.5916E+03, 7.5991E+03, 7.6066E+03,
     +  7.6141E+03, 7.6217E+03, 7.6292E+03, 7.6367E+03, 7.6443E+03,
     + 7.6518E+03, 7.6593E+03, 7.6669E+03, 7.6744E+03, 7.6820E+03,
     +  7.6895E+03, 7.6971E+03, 7.7047E+03, 7.7122E+03, 7.7198E+03,
     + 7.7274E+03, 7.7349E+03, 7.7425E+03, 7.7501E+03, 7.7577E+03,
     +  7.7653E+03, 7.7729E+03, 7.7805E+03, 7.7881E+03, 7.7957E+03,
     + 7.8033E+03, 7.8110E+03, 7.8186E+03, 7.8262E+03, 7.8338E+03,
     +  7.8415E+03, 7.8491E+03, 7.8568E+03, 7.8644E+03, 7.8720E+03,
     + 7.8797E+03, 7.8874E+03, 7.8950E+03, 7.9027E+03, 7.9104E+03,
     +  7.9180E+03, 7.9257E+03, 7.9334E+03, 7.9411E+03, 7.9488E+03,
     + 7.9565E+03, 7.9641E+03, 7.9718E+03, 7.9796E+03, 7.9873E+03,
     +  7.9950E+03, 8.0027E+03, 8.0104E+03, 8.0181E+03, 8.0259E+03,
     + 8.0336E+03, 8.0413E+03, 8.0491E+03, 8.0568E+03, 8.0646E+03,
     +  8.0723E+03, 8.0801E+03, 8.0878E+03, 8.0956E+03, 8.1034E+03,
     + 8.1111E+03, 8.1189E+03, 8.1267E+03, 8.1345E+03, 8.1422E+03,
     +  8.1500E+03, 8.1578E+03, 8.1656E+03, 8.1734E+03, 8.1812E+03,
     + 8.1890E+03, 8.1969E+03, 8.2047E+03, 8.2125E+03, 8.2203E+03,
     +  8.2282E+03, 8.2360E+03, 8.2438E+03, 8.2517E+03, 8.2595E+03,
     + 8.2674E+03, 8.2752E+03, 8.2831E+03, 8.2909E+03, 8.2988E+03,
     +  8.3067E+03, 8.3145E+03, 8.3224E+03, 8.3303E+03, 8.3382E+03,
     + 8.3461E+03, 8.3540E+03, 8.3618E+03, 8.3697E+03, 8.3777E+03,
     +  8.3856E+03, 8.3935E+03, 8.4014E+03, 8.4093E+03, 8.4172E+03,
     + 8.4252E+03, 8.4331E+03, 8.4410E+03, 8.4490E+03, 8.4569E+03,
     +  8.4648E+03, 8.4728E+03, 8.4807E+03, 8.4887E+03, 8.4967E+03,
     + 8.5046E+03, 8.5126E+03, 8.5206E+03, 8.5286E+03, 8.5365E+03,
     +  8.5445E+03, 8.5525E+03, 8.5605E+03, 8.5685E+03, 8.5765E+03,
     + 8.5845E+03, 8.5925E+03, 8.6005E+03, 8.6085E+03, 8.6166E+03,
     +  8.6246E+03, 8.6326E+03, 8.6406E+03, 8.6487E+03, 8.6567E+03,
     + 8.6648E+03, 8.6728E+03, 8.6809E+03, 8.6889E+03, 8.6970E+03,
     +  8.7050E+03, 8.7131E+03, 8.7212E+03, 8.7292E+03, 8.7373E+03,
     + 8.7454E+03, 8.7535E+03, 8.7616E+03, 8.7697E+03, 8.7778E+03,
     +  8.7859E+03, 8.7940E+03, 8.8021E+03, 8.8102E+03, 8.8183E+03,
     + 8.8264E+03, 8.8346E+03, 8.8427E+03, 8.8508E+03, 8.8590E+03,
     +  8.8671E+03, 8.8752E+03, 8.8834E+03, 8.8915E+03, 8.8997E+03,
     + 8.9078E+03, 8.9160E+03, 8.9242E+03, 8.9324E+03, 8.9405E+03,
     +  8.9487E+03, 8.9569E+03, 8.9651E+03, 8.9733E+03, 8.9815E+03,
     + 8.9897E+03, 8.9979E+03, 9.0061E+03, 9.0143E+03, 9.0225E+03,
     +  9.0307E+03, 9.0389E+03, 9.0471E+03, 9.0554E+03, 9.0636E+03,
     + 9.0718E+03, 9.0801E+03, 9.0883E+03, 9.0966E+03, 9.1048E+03,
     +  9.1131E+03, 9.1213E+03, 9.1296E+03, 9.1379E+03, 9.1461E+03,
     + 9.1544E+03, 9.1627E+03, 9.1710E+03, 9.1793E+03, 9.1876E+03,
     +  9.1959E+03, 9.2042E+03, 9.2125E+03, 9.2208E+03, 9.2291E+03,
     + 9.2374E+03, 9.2457E+03, 9.2540E+03, 9.2623E+03, 9.2707E+03,
     +  9.2790E+03, 9.2873E+03, 9.2957E+03, 9.3040E+03, 9.3124E+03,
     + 9.3207E+03, 9.3291E+03, 9.3374E+03, 9.3458E+03, 9.3542E+03,
     +  9.3625E+03, 9.3709E+03, 9.3793E+03, 9.3877E+03, 9.3961E+03,
     + 9.4045E+03, 9.4129E+03, 9.4212E+03, 9.4297E+03, 9.4381E+03,
     +  9.4465E+03, 9.4549E+03, 9.4633E+03, 9.4717E+03, 9.4801E+03,
     + 9.4886E+03, 9.4970E+03, 9.5054E+03, 9.5139E+03, 9.5223E+03,
     +  9.5308E+03, 9.5392E+03, 9.5477E+03, 9.5561E+03, 9.5646E+03,
     + 9.5731E+03, 9.5815E+03, 9.5900E+03, 9.5985E+03, 9.6070E+03,
     +  9.6155E+03, 9.6240E+03, 9.6324E+03, 9.6409E+03, 9.6494E+03,
     + 9.6579E+03, 9.6665E+03, 9.6750E+03, 9.6835E+03, 9.6920E+03,
     +  9.7005E+03, 9.7091E+03, 9.7176E+03, 9.7261E+03, 9.7347E+03,
     + 9.7432E+03, 9.7518E+03, 9.7603E+03, 9.7689E+03, 9.7774E+03,
     +  9.7860E+03, 9.7946E+03, 9.8031E+03, 9.8117E+03, 9.8203E+03,
     + 9.8289E+03, 9.8374E+03, 9.8460E+03, 9.8546E+03, 9.8632E+03,
     +  9.8718E+03, 9.8804E+03, 9.8890E+03, 9.8976E+03, 9.9063E+03,
     + 9.9149E+03, 9.9235E+03, 9.9321E+03, 9.9408E+03, 9.9494E+03,
     +  9.9580E+03, 9.9667E+03, 9.9753E+03, 9.9840E+03, 9.9926E+03,
     + 1.0001E+04, 1.0010E+04, 1.0019E+04, 1.0027E+04, 1.0036E+04,
     +  1.0045E+04, 1.0053E+04, 1.0062E+04, 1.0071E+04, 1.0079E+04,
     + 1.0088E+04, 1.0097E+04, 1.0105E+04, 1.0114E+04, 1.0123E+04,
     +  1.0132E+04, 1.0140E+04, 1.0149E+04, 1.0158E+04, 1.0167E+04,
     + 1.0175E+04, 1.0184E+04, 1.0193E+04, 1.0201E+04, 1.0210E+04,
     +  1.0219E+04, 1.0228E+04, 1.0236E+04, 1.0245E+04, 1.0254E+04,
     + 1.0263E+04, 1.0272E+04, 1.0280E+04, 1.0289E+04, 1.0298E+04,
     +  1.0307E+04, 1.0315E+04, 1.0324E+04, 1.0333E+04, 1.0342E+04,
     + 1.0351E+04, 1.0360E+04, 1.0368E+04, 1.0377E+04, 1.0386E+04,
     +  1.0395E+04, 1.0404E+04, 1.0412E+04, 1.0421E+04, 1.0430E+04,
     + 1.0439E+04, 1.0448E+04, 1.0457E+04, 1.0466E+04, 1.0474E+04,
     +  1.0483E+04, 1.0492E+04, 1.0501E+04, 1.0510E+04, 1.0519E+04,
     + 1.0528E+04, 1.0537E+04, 1.0545E+04, 1.0554E+04, 1.0563E+04,
     +  1.0572E+04, 1.0581E+04, 1.0590E+04, 1.0599E+04, 1.0608E+04,
     + 1.0617E+04, 1.0626E+04, 1.0635E+04, 1.0644E+04, 1.0652E+04,
     +  1.0661E+04, 1.0670E+04, 1.0679E+04, 1.0688E+04, 1.0697E+04,
     + 1.0706E+04, 1.0715E+04, 1.0724E+04, 1.0733E+04, 1.0742E+04,
     +  1.0751E+04, 1.0760E+04, 1.0769E+04, 1.0778E+04, 1.0787E+04,
     + 1.0796E+04, 1.0805E+04, 1.0814E+04, 1.0823E+04, 1.0832E+04,
     +  1.0841E+04, 1.0850E+04, 1.0859E+04, 1.0868E+04, 1.0877E+04,
     + 1.0886E+04, 1.0895E+04, 1.0904E+04, 1.0913E+04, 1.0922E+04,
     +  1.0931E+04, 1.0940E+04, 1.0950E+04, 1.0959E+04, 1.0968E+04,
     + 1.0977E+04, 1.0986E+04, 1.0995E+04, 1.1004E+04, 1.1013E+04,
     +  1.1022E+04, 1.1031E+04, 1.1040E+04, 1.1049E+04, 1.1059E+04,
     + 1.1068E+04, 1.1077E+04, 1.1086E+04, 1.1095E+04, 1.1104E+04,
     +  1.1113E+04, 1.1122E+04, 1.1132E+04, 1.1141E+04, 1.1150E+04,
     + 1.1159E+04, 1.1168E+04, 1.1177E+04, 1.1186E+04, 1.1196E+04,
     +  1.1205E+04, 1.1214E+04, 1.1223E+04, 1.1232E+04, 1.1241E+04,
     + 1.1251E+04, 1.1260E+04, 1.1269E+04, 1.1278E+04, 1.1287E+04,
     +  1.1297E+04, 1.1306E+04, 1.1315E+04, 1.1324E+04, 1.1333E+04,
     + 1.1343E+04, 1.1352E+04, 1.1361E+04, 1.1370E+04, 1.1380E+04,
     +  1.1389E+04, 1.1398E+04, 1.1407E+04, 1.1417E+04, 1.1426E+04,
     + 1.1435E+04, 1.1444E+04, 1.1454E+04, 1.1463E+04, 1.1472E+04,
     +  1.1481E+04, 1.1491E+04, 1.1500E+04, 1.1509E+04, 1.1519E+04,
     + 1.1528E+04, 1.1537E+04, 1.1547E+04, 1.1556E+04, 1.1565E+04,
     +  1.1574E+04, 1.1584E+04, 1.1593E+04, 1.1602E+04, 1.1612E+04,
     + 1.1621E+04, 1.1630E+04, 1.1640E+04, 1.1649E+04, 1.1658E+04,
     +  1.1668E+04, 1.1677E+04, 1.1687E+04, 1.1696E+04, 1.1705E+04,
     + 1.1715E+04, 1.1724E+04, 1.1733E+04, 1.1743E+04, 1.1752E+04,
     +  1.1762E+04, 1.1771E+04, 1.1780E+04, 1.1790E+04, 1.1799E+04,
     + 1.1809E+04, 1.1818E+04, 1.1827E+04, 1.1837E+04, 1.1846E+04,
     +  1.1856E+04, 1.1865E+04, 1.1874E+04, 1.1884E+04, 1.1893E+04,
     + 1.1903E+04, 1.1912E+04, 1.1922E+04, 1.1931E+04, 1.1941E+04,
     +  1.1950E+04, 1.1960E+04, 1.1969E+04, 1.1978E+04, 1.1988E+04,
     + 1.1997E+04, 1.2007E+04, 1.2016E+04, 1.2026E+04, 1.2035E+04,
     +  1.2045E+04, 1.2054E+04, 1.2064E+04, 1.2073E+04, 1.2083E+04,
     + 1.2092E+04, 1.2102E+04, 1.2111E+04, 1.2121E+04, 1.2131E+04,
     +  1.2140E+04, 1.2150E+04, 1.2159E+04, 1.2169E+04, 1.2178E+04,
     + 1.2188E+04, 1.2197E+04, 1.2207E+04, 1.2216E+04, 1.2226E+04,
     +  1.2236E+04, 1.2245E+04, 1.2255E+04, 1.2264E+04, 1.2274E+04,
     + 1.2284E+04, 1.2293E+04, 1.2303E+04, 1.2312E+04, 1.2322E+04,
     +  1.2332E+04, 1.2341E+04, 1.2351E+04, 1.2360E+04, 1.2370E+04,
     + 1.2380E+04, 1.2389E+04, 1.2399E+04, 1.2409E+04, 1.2418E+04,
     +  1.2428E+04, 1.2437E+04, 1.2447E+04, 1.2457E+04, 1.2466E+04,
     + 1.2476E+04, 1.2486E+04, 1.2495E+04, 1.2505E+04, 1.2515E+04,
     +  1.2524E+04, 1.2534E+04, 1.2544E+04, 1.2553E+04, 1.2563E+04,
     + 1.2573E+04, 1.2583E+04, 1.2592E+04, 1.2602E+04, 1.2612E+04,
     +  1.2621E+04, 1.2631E+04, 1.2641E+04, 1.2651E+04, 1.2660E+04,
     + 1.2670E+04, 1.2680E+04, 1.2690E+04, 1.2699E+04, 1.2709E+04,
     +  1.2719E+04, 1.2729E+04, 1.2738E+04, 1.2748E+04, 1.2758E+04,
     + 1.2768E+04, 1.2777E+04, 1.2787E+04, 1.2797E+04, 1.2807E+04,
     +  1.2817E+04, 1.2826E+04, 1.2836E+04, 1.2846E+04, 1.2856E+04,
     + 1.2866E+04, 1.2875E+04, 1.2885E+04, 1.2895E+04, 1.2905E+04,
     +  1.2915E+04, 1.2924E+04, 1.2934E+04, 1.2944E+04, 1.2954E+04,
     + 1.2964E+04, 1.2974E+04, 1.2983E+04, 1.2993E+04, 1.3003E+04,
     +  1.3013E+04, 1.3023E+04, 1.3033E+04, 1.3043E+04, 1.3053E+04,
     + 1.3062E+04, 1.3072E+04, 1.3082E+04, 1.3092E+04, 1.3102E+04,
     +  1.3112E+04, 1.3122E+04, 1.3132E+04, 1.3142E+04, 1.3151E+04,
     + 1.3161E+04, 1.3171E+04, 1.3181E+04, 1.3191E+04, 1.3201E+04,
     +  1.3211E+04, 1.3221E+04, 1.3231E+04, 1.3241E+04, 1.3251E+04,
     + 1.3261E+04, 1.3271E+04, 1.3281E+04, 1.3291E+04, 1.3301E+04,
     +  1.3311E+04, 1.3321E+04, 1.3330E+04, 1.3340E+04, 1.3350E+04,
     + 1.3360E+04, 1.3370E+04, 1.3380E+04, 1.3390E+04, 1.3400E+04,
     +  1.3410E+04, 1.3420E+04, 1.3430E+04, 1.3440E+04, 1.3450E+04,
     + 1.3460E+04, 1.3471E+04, 1.3481E+04, 1.3491E+04, 1.3501E+04,
     +  1.3511E+04, 1.3521E+04, 1.3531E+04, 1.3541E+04, 1.3551E+04,
     + 1.3561E+04, 1.3571E+04, 1.3581E+04, 1.3591E+04, 1.3601E+04,
     +  1.3611E+04, 1.3621E+04, 1.3631E+04, 1.3641E+04, 1.3652E+04,
     + 1.3662E+04, 1.3672E+04, 1.3682E+04, 1.3692E+04, 1.3702E+04,
     +  1.3712E+04, 1.3722E+04, 1.3732E+04, 1.3743E+04, 1.3753E+04,
     + 1.3763E+04, 1.3773E+04, 1.3783E+04, 1.3793E+04, 1.3803E+04,
     +  1.3813E+04, 1.3824E+04, 1.3834E+04, 1.3844E+04, 1.3854E+04/
      data (QofT( 1,J),J=2409,4510)/ 1.3864E+04, 1.3874E+04,
     +  1.3885E+04, 1.3895E+04, 1.3905E+04, 1.3915E+04, 1.3925E+04,
     +  1.3935E+04,
     + 1.3946E+04, 1.3956E+04, 1.3966E+04, 1.3976E+04, 1.3986E+04,
     +  1.3997E+04, 1.4007E+04, 1.4017E+04, 1.4027E+04, 1.4038E+04,
     + 1.4048E+04, 1.4058E+04, 1.4068E+04, 1.4078E+04, 1.4089E+04,
     +  1.4099E+04, 1.4109E+04, 1.4119E+04, 1.4130E+04, 1.4140E+04,
     + 1.4150E+04, 1.4160E+04, 1.4171E+04, 1.4181E+04, 1.4191E+04,
     +  1.4202E+04, 1.4212E+04, 1.4222E+04, 1.4232E+04, 1.4243E+04,
     + 1.4253E+04, 1.4263E+04, 1.4274E+04, 1.4284E+04, 1.4294E+04,
     +  1.4305E+04, 1.4315E+04, 1.4325E+04, 1.4336E+04, 1.4346E+04,
     + 1.4356E+04, 1.4367E+04, 1.4377E+04, 1.4387E+04, 1.4398E+04,
     +  1.4408E+04, 1.4418E+04, 1.4429E+04, 1.4439E+04, 1.4449E+04,
     + 1.4460E+04, 1.4470E+04, 1.4480E+04, 1.4491E+04, 1.4501E+04,
     +  1.4512E+04, 1.4522E+04, 1.4532E+04, 1.4543E+04, 1.4553E+04,
     + 1.4564E+04, 1.4574E+04, 1.4584E+04, 1.4595E+04, 1.4605E+04,
     +  1.4616E+04, 1.4626E+04, 1.4636E+04, 1.4647E+04, 1.4657E+04,
     + 1.4668E+04, 1.4678E+04, 1.4689E+04, 1.4699E+04, 1.4709E+04,
     +  1.4720E+04, 1.4730E+04, 1.4741E+04, 1.4751E+04, 1.4762E+04,
     + 1.4772E+04, 1.4783E+04, 1.4793E+04, 1.4804E+04, 1.4814E+04,
     +  1.4825E+04, 1.4835E+04, 1.4846E+04, 1.4856E+04, 1.4867E+04,
     + 1.4877E+04, 1.4888E+04, 1.4898E+04, 1.4909E+04, 1.4919E+04,
     +  1.4930E+04, 1.4940E+04, 1.4951E+04, 1.4961E+04, 1.4972E+04,
     + 1.4982E+04, 1.4993E+04, 1.5003E+04, 1.5014E+04, 1.5025E+04,
     +  1.5035E+04, 1.5046E+04, 1.5056E+04, 1.5067E+04, 1.5077E+04,
     + 1.5088E+04, 1.5098E+04, 1.5109E+04, 1.5120E+04, 1.5130E+04,
     +  1.5141E+04, 1.5151E+04, 1.5162E+04, 1.5173E+04, 1.5183E+04,
     + 1.5194E+04, 1.5204E+04, 1.5215E+04, 1.5226E+04, 1.5236E+04,
     +  1.5247E+04, 1.5258E+04, 1.5268E+04, 1.5279E+04, 1.5289E+04,
     + 1.5300E+04, 1.5311E+04, 1.5321E+04, 1.5332E+04, 1.5343E+04,
     +  1.5353E+04, 1.5364E+04, 1.5375E+04, 1.5385E+04, 1.5396E+04,
     + 1.5407E+04, 1.5417E+04, 1.5428E+04, 1.5439E+04, 1.5449E+04,
     +  1.5460E+04, 1.5471E+04, 1.5481E+04, 1.5492E+04, 1.5503E+04,
     + 1.5514E+04, 1.5524E+04, 1.5535E+04, 1.5546E+04, 1.5556E+04,
     +  1.5567E+04, 1.5578E+04, 1.5589E+04, 1.5599E+04, 1.5610E+04,
     + 1.5621E+04, 1.5632E+04, 1.5642E+04, 1.5653E+04, 1.5664E+04,
     +  1.5675E+04, 1.5685E+04, 1.5696E+04, 1.5707E+04, 1.5718E+04,
     + 1.5728E+04, 1.5739E+04, 1.5750E+04, 1.5761E+04, 1.5772E+04,
     +  1.5782E+04, 1.5793E+04, 1.5804E+04, 1.5815E+04, 1.5826E+04,
     + 1.5836E+04, 1.5847E+04, 1.5858E+04, 1.5869E+04, 1.5880E+04,
     +  1.5890E+04, 1.5901E+04, 1.5912E+04, 1.5923E+04, 1.5934E+04,
     + 1.5945E+04, 1.5956E+04, 1.5966E+04, 1.5977E+04, 1.5988E+04,
     +  1.5999E+04, 1.6010E+04, 1.6021E+04, 1.6032E+04, 1.6042E+04,
     + 1.6053E+04, 1.6064E+04, 1.6075E+04, 1.6086E+04, 1.6097E+04,
     +  1.6108E+04, 1.6119E+04, 1.6130E+04, 1.6140E+04, 1.6151E+04,
     + 1.6162E+04, 1.6173E+04, 1.6184E+04, 1.6195E+04, 1.6206E+04,
     +  1.6217E+04, 1.6228E+04, 1.6239E+04, 1.6250E+04, 1.6261E+04,
     + 1.6271E+04, 1.6282E+04, 1.6293E+04, 1.6304E+04, 1.6315E+04,
     +  1.6326E+04, 1.6337E+04, 1.6348E+04, 1.6359E+04, 1.6370E+04,
     + 1.6381E+04, 1.6392E+04, 1.6403E+04, 1.6414E+04, 1.6425E+04,
     +  1.6436E+04, 1.6447E+04, 1.6458E+04, 1.6469E+04, 1.6480E+04,
     + 1.6491E+04, 1.6502E+04, 1.6513E+04, 1.6524E+04, 1.6535E+04,
     +  1.6546E+04, 1.6557E+04, 1.6568E+04, 1.6579E+04, 1.6590E+04,
     + 1.6601E+04, 1.6612E+04, 1.6623E+04, 1.6634E+04, 1.6645E+04,
     +  1.6657E+04, 1.6668E+04, 1.6679E+04, 1.6690E+04, 1.6701E+04,
     + 1.6712E+04, 1.6723E+04, 1.6734E+04, 1.6745E+04, 1.6756E+04,
     +  1.6767E+04, 1.6778E+04, 1.6790E+04, 1.6801E+04, 1.6812E+04,
     + 1.6823E+04, 1.6834E+04, 1.6845E+04, 1.6856E+04, 1.6867E+04,
     +  1.6878E+04, 1.6890E+04, 1.6901E+04, 1.6912E+04, 1.6923E+04,
     + 1.6934E+04, 1.6945E+04, 1.6956E+04, 1.6968E+04, 1.6979E+04,
     +  1.6990E+04, 1.7001E+04, 1.7012E+04, 1.7023E+04, 1.7034E+04,
     + 1.7046E+04, 1.7057E+04, 1.7068E+04, 1.7079E+04, 1.7090E+04,
     +  1.7102E+04, 1.7113E+04, 1.7124E+04, 1.7135E+04, 1.7146E+04,
     + 1.7158E+04, 1.7169E+04, 1.7180E+04, 1.7191E+04, 1.7202E+04,
     +  1.7214E+04, 1.7225E+04, 1.7236E+04, 1.7247E+04, 1.7259E+04,
     + 1.7270E+04, 1.7281E+04, 1.7292E+04, 1.7303E+04, 1.7315E+04,
     +  1.7326E+04, 1.7337E+04, 1.7349E+04, 1.7360E+04, 1.7371E+04,
     + 1.7382E+04, 1.7394E+04, 1.7405E+04, 1.7416E+04, 1.7427E+04,
     +  1.7439E+04, 1.7450E+04, 1.7461E+04, 1.7473E+04, 1.7484E+04,
     + 1.7495E+04, 1.7506E+04, 1.7518E+04, 1.7529E+04, 1.7540E+04,
     +  1.7552E+04, 1.7563E+04, 1.7574E+04, 1.7586E+04, 1.7597E+04,
     + 1.7608E+04, 1.7620E+04, 1.7631E+04, 1.7642E+04, 1.7654E+04,
     +  1.7665E+04, 1.7676E+04, 1.7688E+04, 1.7699E+04, 1.7711E+04,
     + 1.7722E+04, 1.7733E+04, 1.7745E+04, 1.7756E+04, 1.7767E+04,
     +  1.7779E+04, 1.7790E+04, 1.7802E+04, 1.7813E+04, 1.7824E+04,
     + 1.7836E+04, 1.7847E+04, 1.7858E+04, 1.7870E+04, 1.7881E+04,
     +  1.7893E+04, 1.7904E+04, 1.7916E+04, 1.7927E+04, 1.7938E+04,
     + 1.7950E+04, 1.7961E+04, 1.7973E+04, 1.7984E+04, 1.7996E+04,
     +  1.8007E+04, 1.8018E+04, 1.8030E+04, 1.8041E+04, 1.8053E+04,
     + 1.8064E+04, 1.8076E+04, 1.8087E+04, 1.8099E+04, 1.8110E+04,
     +  1.8122E+04, 1.8133E+04, 1.8145E+04, 1.8156E+04, 1.8168E+04,
     + 1.8179E+04, 1.8191E+04, 1.8202E+04, 1.8214E+04, 1.8225E+04,
     +  1.8237E+04, 1.8248E+04, 1.8260E+04, 1.8271E+04, 1.8283E+04,
     + 1.8294E+04, 1.8306E+04, 1.8317E+04, 1.8329E+04, 1.8340E+04,
     +  1.8352E+04, 1.8363E+04, 1.8375E+04, 1.8386E+04, 1.8398E+04,
     + 1.8410E+04, 1.8421E+04, 1.8433E+04, 1.8444E+04, 1.8456E+04,
     +  1.8467E+04, 1.8479E+04, 1.8491E+04, 1.8502E+04, 1.8514E+04,
     + 1.8525E+04, 1.8537E+04, 1.8548E+04, 1.8560E+04, 1.8572E+04,
     +  1.8583E+04, 1.8595E+04, 1.8606E+04, 1.8618E+04, 1.8630E+04,
     + 1.8641E+04, 1.8653E+04, 1.8665E+04, 1.8676E+04, 1.8688E+04,
     +  1.8699E+04, 1.8711E+04, 1.8723E+04, 1.8734E+04, 1.8746E+04,
     + 1.8758E+04, 1.8769E+04, 1.8781E+04, 1.8793E+04, 1.8804E+04,
     +  1.8816E+04, 1.8828E+04, 1.8839E+04, 1.8851E+04, 1.8863E+04,
     + 1.8874E+04, 1.8886E+04, 1.8898E+04, 1.8909E+04, 1.8921E+04,
     +  1.8933E+04, 1.8944E+04, 1.8956E+04, 1.8968E+04, 1.8980E+04,
     + 1.8991E+04, 1.9003E+04, 1.9015E+04, 1.9026E+04, 1.9038E+04,
     +  1.9050E+04, 1.9062E+04, 1.9073E+04, 1.9085E+04, 1.9097E+04,
     + 1.9109E+04, 1.9120E+04, 1.9132E+04, 1.9144E+04, 1.9156E+04,
     +  1.9167E+04, 1.9179E+04, 1.9191E+04, 1.9203E+04, 1.9214E+04,
     + 1.9226E+04, 1.9238E+04, 1.9250E+04, 1.9261E+04, 1.9273E+04,
     +  1.9285E+04, 1.9297E+04, 1.9309E+04, 1.9320E+04, 1.9332E+04,
     + 1.9344E+04, 1.9356E+04, 1.9368E+04, 1.9379E+04, 1.9391E+04,
     +  1.9403E+04, 1.9415E+04, 1.9427E+04, 1.9439E+04, 1.9450E+04,
     + 1.9462E+04, 1.9474E+04, 1.9486E+04, 1.9498E+04, 1.9510E+04,
     +  1.9521E+04, 1.9533E+04, 1.9545E+04, 1.9557E+04, 1.9569E+04,
     + 1.9581E+04, 1.9593E+04, 1.9604E+04, 1.9616E+04, 1.9628E+04,
     +  1.9640E+04, 1.9652E+04, 1.9664E+04, 1.9676E+04, 1.9688E+04,
     + 1.9699E+04, 1.9711E+04, 1.9723E+04, 1.9735E+04, 1.9747E+04,
     +  1.9759E+04, 1.9771E+04, 1.9783E+04, 1.9795E+04, 1.9807E+04,
     + 1.9819E+04, 1.9831E+04, 1.9842E+04, 1.9854E+04, 1.9866E+04,
     +  1.9878E+04, 1.9890E+04, 1.9902E+04, 1.9914E+04, 1.9926E+04,
     + 1.9938E+04, 1.9950E+04, 1.9962E+04, 1.9974E+04, 1.9986E+04,
     +  1.9998E+04, 2.0010E+04, 2.0022E+04, 2.0034E+04, 2.0046E+04,
     + 2.0058E+04, 2.0070E+04, 2.0082E+04, 2.0094E+04, 2.0106E+04,
     +  2.0118E+04, 2.0130E+04, 2.0142E+04, 2.0154E+04, 2.0166E+04,
     + 2.0178E+04, 2.0190E+04, 2.0202E+04, 2.0214E+04, 2.0226E+04,
     +  2.0238E+04, 2.0250E+04, 2.0262E+04, 2.0274E+04, 2.0286E+04,
     + 2.0298E+04, 2.0310E+04, 2.0322E+04, 2.0334E+04, 2.0346E+04,
     +  2.0358E+04, 2.0370E+04, 2.0382E+04, 2.0394E+04, 2.0406E+04,
     + 2.0419E+04, 2.0431E+04, 2.0443E+04, 2.0455E+04, 2.0467E+04,
     +  2.0479E+04, 2.0491E+04, 2.0503E+04, 2.0515E+04, 2.0527E+04,
     + 2.0539E+04, 2.0552E+04, 2.0564E+04, 2.0576E+04, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        27
      data (QofT( 2,J),J=1,2408)/ 1.8005E+01, 1.8501E+01, 2.0391E+01,
     +  2.3280E+01, 2.6632E+01, 3.0189E+01, 3.3847E+01, 3.7560E+01,
     + 4.1306E+01, 4.5075E+01, 4.8859E+01, 5.2654E+01, 5.6458E+01,
     +  6.0269E+01, 6.4084E+01, 6.7904E+01, 7.1727E+01, 7.5553E+01,
     + 7.9381E+01, 8.3211E+01, 9.0876E+01, 9.8546E+01, 1.0622E+02,
     +  1.1390E+02, 1.2158E+02, 1.2926E+02, 1.3694E+02, 1.4463E+02,
     + 1.5231E+02, 1.6000E+02, 1.6769E+02, 1.7537E+02, 1.8306E+02,
     +  1.9075E+02, 1.9845E+02, 2.0614E+02, 2.1383E+02, 2.2152E+02,
     + 2.2922E+02, 2.3691E+02, 2.4460E+02, 2.5230E+02, 2.5999E+02,
     +  2.6769E+02, 2.7538E+02, 2.8308E+02, 2.9077E+02, 2.9847E+02,
     + 3.0617E+02, 3.1387E+02, 3.2156E+02, 3.2926E+02, 3.3696E+02,
     +  3.4466E+02, 3.5236E+02, 3.6005E+02, 3.6775E+02, 3.7545E+02,
     + 3.8315E+02, 3.9085E+02, 3.9855E+02, 4.0625E+02, 4.1395E+02,
     +  4.2166E+02, 4.2936E+02, 4.3706E+02, 4.4476E+02, 4.5246E+02,
     + 4.6016E+02, 4.6787E+02, 4.7557E+02, 4.8327E+02, 4.9098E+02,
     +  4.9868E+02, 5.0638E+02, 5.1409E+02, 5.2179E+02, 5.2949E+02,
     + 5.3720E+02, 5.4490E+02, 5.5261E+02, 5.6031E+02, 5.6802E+02,
     +  5.7573E+02, 5.8343E+02, 5.9114E+02, 5.9884E+02, 6.0655E+02,
     + 6.1426E+02, 6.2197E+02, 6.2967E+02, 6.3738E+02, 6.4509E+02,
     +  6.5280E+02, 6.6050E+02, 6.6821E+02, 6.7592E+02, 6.8363E+02,
     + 6.9134E+02, 6.9905E+02, 7.0676E+02, 7.1447E+02, 7.2218E+02,
     +  7.2989E+02, 7.3760E+02, 7.4531E+02, 7.5303E+02, 7.6074E+02,
     + 7.6845E+02, 7.7616E+02, 7.8387E+02, 7.9159E+02, 7.9930E+02,
     +  8.0702E+02, 8.1473E+02, 8.2244E+02, 8.3016E+02, 8.3787E+02,
     + 8.4559E+02, 8.5331E+02, 8.6102E+02, 8.6874E+02, 8.7646E+02,
     +  8.8418E+02, 8.9189E+02, 8.9961E+02, 9.0733E+02, 9.1505E+02,
     + 9.2277E+02, 9.3050E+02, 9.3822E+02, 9.4594E+02, 9.5366E+02,
     +  9.6139E+02, 9.6911E+02, 9.7684E+02, 9.8457E+02, 9.9229E+02,
     + 1.0000E+03, 1.0078E+03, 1.0155E+03, 1.0232E+03, 1.0309E+03,
     +  1.0387E+03, 1.0464E+03, 1.0541E+03, 1.0619E+03, 1.0696E+03,
     + 1.0774E+03, 1.0851E+03, 1.0928E+03, 1.1006E+03, 1.1083E+03,
     +  1.1161E+03, 1.1238E+03, 1.1316E+03, 1.1393E+03, 1.1471E+03,
     + 1.1548E+03, 1.1626E+03, 1.1703E+03, 1.1781E+03, 1.1858E+03,
     +  1.1936E+03, 1.2014E+03, 1.2091E+03, 1.2169E+03, 1.2247E+03,
     + 1.2324E+03, 1.2402E+03, 1.2480E+03, 1.2558E+03, 1.2635E+03,
     +  1.2713E+03, 1.2791E+03, 1.2869E+03, 1.2947E+03, 1.3025E+03,
     + 1.3103E+03, 1.3180E+03, 1.3258E+03, 1.3336E+03, 1.3415E+03,
     +  1.3493E+03, 1.3571E+03, 1.3649E+03, 1.3727E+03, 1.3805E+03,
     + 1.3883E+03, 1.3962E+03, 1.4040E+03, 1.4118E+03, 1.4197E+03,
     +  1.4275E+03, 1.4353E+03, 1.4432E+03, 1.4510E+03, 1.4589E+03,
     + 1.4667E+03, 1.4746E+03, 1.4825E+03, 1.4903E+03, 1.4982E+03,
     +  1.5061E+03, 1.5140E+03, 1.5219E+03, 1.5297E+03, 1.5376E+03,
     + 1.5455E+03, 1.5534E+03, 1.5613E+03, 1.5692E+03, 1.5772E+03,
     +  1.5851E+03, 1.5930E+03, 1.6009E+03, 1.6089E+03, 1.6168E+03,
     + 1.6248E+03, 1.6327E+03, 1.6407E+03, 1.6486E+03, 1.6566E+03,
     +  1.6645E+03, 1.6725E+03, 1.6805E+03, 1.6885E+03, 1.6965E+03,
     + 1.7045E+03, 1.7125E+03, 1.7205E+03, 1.7285E+03, 1.7365E+03,
     +  1.7445E+03, 1.7526E+03, 1.7606E+03, 1.7686E+03, 1.7767E+03,
     + 1.7847E+03, 1.7928E+03, 1.8009E+03, 1.8089E+03, 1.8170E+03,
     +  1.8251E+03, 1.8332E+03, 1.8413E+03, 1.8494E+03, 1.8575E+03,
     + 1.8656E+03, 1.8737E+03, 1.8819E+03, 1.8900E+03, 1.8981E+03,
     +  1.9063E+03, 1.9145E+03, 1.9226E+03, 1.9308E+03, 1.9390E+03,
     + 1.9472E+03, 1.9553E+03, 1.9635E+03, 1.9718E+03, 1.9800E+03,
     +  1.9882E+03, 1.9964E+03, 2.0047E+03, 2.0129E+03, 2.0212E+03,
     + 2.0294E+03, 2.0377E+03, 2.0460E+03, 2.0542E+03, 2.0625E+03,
     +  2.0708E+03, 2.0791E+03, 2.0875E+03, 2.0958E+03, 2.1041E+03,
     + 2.1124E+03, 2.1208E+03, 2.1291E+03, 2.1375E+03, 2.1459E+03,
     +  2.1543E+03, 2.1627E+03, 2.1710E+03, 2.1795E+03, 2.1879E+03,
     + 2.1963E+03, 2.2047E+03, 2.2132E+03, 2.2216E+03, 2.2301E+03,
     +  2.2385E+03, 2.2470E+03, 2.2555E+03, 2.2640E+03, 2.2725E+03,
     + 2.2810E+03, 2.2895E+03, 2.2980E+03, 2.3066E+03, 2.3151E+03,
     +  2.3237E+03, 2.3322E+03, 2.3408E+03, 2.3494E+03, 2.3580E+03,
     + 2.3666E+03, 2.3752E+03, 2.3838E+03, 2.3925E+03, 2.4011E+03,
     +  2.4097E+03, 2.4184E+03, 2.4271E+03, 2.4358E+03, 2.4444E+03,
     + 2.4531E+03, 2.4618E+03, 2.4706E+03, 2.4793E+03, 2.4880E+03,
     +  2.4968E+03, 2.5055E+03, 2.5143E+03, 2.5231E+03, 2.5319E+03,
     + 2.5407E+03, 2.5495E+03, 2.5583E+03, 2.5671E+03, 2.5759E+03,
     +  2.5848E+03, 2.5936E+03, 2.6025E+03, 2.6114E+03, 2.6203E+03,
     + 2.6292E+03, 2.6381E+03, 2.6470E+03, 2.6559E+03, 2.6649E+03,
     +  2.6738E+03, 2.6828E+03, 2.6918E+03, 2.7007E+03, 2.7097E+03,
     + 2.7187E+03, 2.7277E+03, 2.7368E+03, 2.7458E+03, 2.7549E+03,
     +  2.7639E+03, 2.7730E+03, 2.7821E+03, 2.7912E+03, 2.8003E+03,
     + 2.8094E+03, 2.8185E+03, 2.8276E+03, 2.8368E+03, 2.8459E+03,
     +  2.8551E+03, 2.8643E+03, 2.8735E+03, 2.8827E+03, 2.8919E+03,
     + 2.9011E+03, 2.9103E+03, 2.9196E+03, 2.9288E+03, 2.9381E+03,
     +  2.9474E+03, 2.9567E+03, 2.9660E+03, 2.9753E+03, 2.9846E+03,
     + 2.9939E+03, 3.0033E+03, 3.0126E+03, 3.0220E+03, 3.0314E+03,
     +  3.0408E+03, 3.0502E+03, 3.0596E+03, 3.0690E+03, 3.0785E+03,
     + 3.0879E+03, 3.0974E+03, 3.1069E+03, 3.1163E+03, 3.1258E+03,
     +  3.1353E+03, 3.1449E+03, 3.1544E+03, 3.1639E+03, 3.1735E+03,
     + 3.1831E+03, 3.1927E+03, 3.2022E+03, 3.2118E+03, 3.2215E+03,
     +  3.2311E+03, 3.2407E+03, 3.2504E+03, 3.2600E+03, 3.2697E+03,
     + 3.2794E+03, 3.2891E+03, 3.2988E+03, 3.3085E+03, 3.3183E+03,
     +  3.3280E+03, 3.3378E+03, 3.3476E+03, 3.3573E+03, 3.3671E+03,
     + 3.3769E+03, 3.3868E+03, 3.3966E+03, 3.4064E+03, 3.4163E+03,
     +  3.4262E+03, 3.4361E+03, 3.4459E+03, 3.4559E+03, 3.4658E+03,
     + 3.4757E+03, 3.4856E+03, 3.4956E+03, 3.5056E+03, 3.5156E+03,
     +  3.5255E+03, 3.5356E+03, 3.5456E+03, 3.5556E+03, 3.5656E+03,
     + 3.5757E+03, 3.5858E+03, 3.5959E+03, 3.6059E+03, 3.6160E+03,
     +  3.6262E+03, 3.6363E+03, 3.6464E+03, 3.6566E+03, 3.6668E+03,
     + 3.6769E+03, 3.6871E+03, 3.6974E+03, 3.7076E+03, 3.7178E+03,
     +  3.7280E+03, 3.7383E+03, 3.7486E+03, 3.7589E+03, 3.7692E+03,
     + 3.7795E+03, 3.7898E+03, 3.8001E+03, 3.8105E+03, 3.8208E+03,
     +  3.8312E+03, 3.8416E+03, 3.8520E+03, 3.8624E+03, 3.8728E+03,
     + 3.8833E+03, 3.8937E+03, 3.9042E+03, 3.9146E+03, 3.9251E+03,
     +  3.9356E+03, 3.9461E+03, 3.9567E+03, 3.9672E+03, 3.9778E+03,
     + 3.9883E+03, 3.9989E+03, 4.0095E+03, 4.0201E+03, 4.0307E+03,
     +  4.0414E+03, 4.0520E+03, 4.0627E+03, 4.0733E+03, 4.0840E+03,
     + 4.0947E+03, 4.1054E+03, 4.1162E+03, 4.1269E+03, 4.1376E+03,
     +  4.1484E+03, 4.1592E+03, 4.1700E+03, 4.1808E+03, 4.1916E+03,
     + 4.2024E+03, 4.2133E+03, 4.2241E+03, 4.2350E+03, 4.2459E+03,
     +  4.2568E+03, 4.2677E+03, 4.2786E+03, 4.2895E+03, 4.3005E+03,
     + 4.3114E+03, 4.3224E+03, 4.3334E+03, 4.3444E+03, 4.3554E+03,
     +  4.3665E+03, 4.3775E+03, 4.3886E+03, 4.3996E+03, 4.4107E+03,
     + 4.4218E+03, 4.4329E+03, 4.4440E+03, 4.4552E+03, 4.4663E+03,
     +  4.4775E+03, 4.4887E+03, 4.4999E+03, 4.5111E+03, 4.5223E+03,
     + 4.5335E+03, 4.5448E+03, 4.5560E+03, 4.5673E+03, 4.5786E+03,
     +  4.5899E+03, 4.6012E+03, 4.6125E+03, 4.6238E+03, 4.6352E+03,
     + 4.6466E+03, 4.6579E+03, 4.6693E+03, 4.6807E+03, 4.6922E+03,
     +  4.7036E+03, 4.7150E+03, 4.7265E+03, 4.7380E+03, 4.7495E+03,
     + 4.7610E+03, 4.7725E+03, 4.7840E+03, 4.7956E+03, 4.8071E+03,
     +  4.8187E+03, 4.8303E+03, 4.8419E+03, 4.8535E+03, 4.8651E+03,
     + 4.8768E+03, 4.8884E+03, 4.9001E+03, 4.9118E+03, 4.9235E+03,
     +  4.9352E+03, 4.9469E+03, 4.9586E+03, 4.9704E+03, 4.9821E+03,
     + 4.9939E+03, 5.0057E+03, 5.0175E+03, 5.0293E+03, 5.0412E+03,
     +  5.0530E+03, 5.0649E+03, 5.0767E+03, 5.0886E+03, 5.1005E+03,
     + 5.1124E+03, 5.1244E+03, 5.1363E+03, 5.1483E+03, 5.1603E+03,
     +  5.1722E+03, 5.1842E+03, 5.1963E+03, 5.2083E+03, 5.2203E+03,
     + 5.2324E+03, 5.2444E+03, 5.2565E+03, 5.2686E+03, 5.2807E+03,
     +  5.2929E+03, 5.3050E+03, 5.3172E+03, 5.3293E+03, 5.3415E+03,
     + 5.3537E+03, 5.3659E+03, 5.3781E+03, 5.3904E+03, 5.4026E+03,
     +  5.4149E+03, 5.4272E+03, 5.4395E+03, 5.4518E+03, 5.4641E+03,
     + 5.4764E+03, 5.4888E+03, 5.5011E+03, 5.5135E+03, 5.5259E+03,
     +  5.5383E+03, 5.5507E+03, 5.5632E+03, 5.5756E+03, 5.5881E+03,
     + 5.6006E+03, 5.6130E+03, 5.6255E+03, 5.6381E+03, 5.6506E+03,
     +  5.6631E+03, 5.6757E+03, 5.6883E+03, 5.7009E+03, 5.7135E+03,
     + 5.7261E+03, 5.7387E+03, 5.7514E+03, 5.7640E+03, 5.7767E+03,
     +  5.7894E+03, 5.8021E+03, 5.8148E+03, 5.8276E+03, 5.8403E+03,
     + 5.8531E+03, 5.8658E+03, 5.8786E+03, 5.8914E+03, 5.9042E+03,
     +  5.9171E+03, 5.9299E+03, 5.9428E+03, 5.9557E+03, 5.9685E+03,
     + 5.9814E+03, 5.9944E+03, 6.0073E+03, 6.0202E+03, 6.0332E+03,
     +  6.0462E+03, 6.0592E+03, 6.0722E+03, 6.0852E+03, 6.0982E+03,
     + 6.1113E+03, 6.1243E+03, 6.1374E+03, 6.1505E+03, 6.1636E+03,
     +  6.1767E+03, 6.1898E+03, 6.2030E+03, 6.2161E+03, 6.2293E+03,
     + 6.2425E+03, 6.2557E+03, 6.2689E+03, 6.2822E+03, 6.2954E+03,
     +  6.3087E+03, 6.3219E+03, 6.3352E+03, 6.3485E+03, 6.3619E+03,
     + 6.3752E+03, 6.3885E+03, 6.4019E+03, 6.4153E+03, 6.4287E+03,
     +  6.4421E+03, 6.4555E+03, 6.4689E+03, 6.4824E+03, 6.4958E+03,
     + 6.5093E+03, 6.5228E+03, 6.5363E+03, 6.5498E+03, 6.5634E+03,
     +  6.5769E+03, 6.5905E+03, 6.6041E+03, 6.6177E+03, 6.6313E+03,
     + 6.6449E+03, 6.6585E+03, 6.6722E+03, 6.6858E+03, 6.6995E+03,
     +  6.7132E+03, 6.7269E+03, 6.7406E+03, 6.7544E+03, 6.7681E+03,
     + 6.7819E+03, 6.7957E+03, 6.8095E+03, 6.8233E+03, 6.8371E+03,
     +  6.8510E+03, 6.8648E+03, 6.8787E+03, 6.8926E+03, 6.9065E+03,
     + 6.9204E+03, 6.9343E+03, 6.9483E+03, 6.9622E+03, 6.9762E+03,
     +  6.9902E+03, 7.0042E+03, 7.0182E+03, 7.0322E+03, 7.0463E+03,
     + 7.0603E+03, 7.0744E+03, 7.0885E+03, 7.1026E+03, 7.1167E+03,
     +  7.1308E+03, 7.1450E+03, 7.1591E+03, 7.1733E+03, 7.1875E+03,
     + 7.2017E+03, 7.2159E+03, 7.2302E+03, 7.2444E+03, 7.2587E+03,
     +  7.2730E+03, 7.2873E+03, 7.3016E+03, 7.3159E+03, 7.3302E+03,
     + 7.3446E+03, 7.3589E+03, 7.3733E+03, 7.3877E+03, 7.4021E+03,
     +  7.4166E+03, 7.4310E+03, 7.4454E+03, 7.4599E+03, 7.4744E+03,
     + 7.4889E+03, 7.5034E+03, 7.5179E+03, 7.5325E+03, 7.5470E+03,
     +  7.5616E+03, 7.5762E+03, 7.5908E+03, 7.6054E+03, 7.6201E+03,
     + 7.6347E+03, 7.6494E+03, 7.6640E+03, 7.6787E+03, 7.6934E+03,
     +  7.7082E+03, 7.7229E+03, 7.7376E+03, 7.7524E+03, 7.7672E+03,
     + 7.7820E+03, 7.7968E+03, 7.8116E+03, 7.8264E+03, 7.8413E+03,
     +  7.8562E+03, 7.8710E+03, 7.8859E+03, 7.9009E+03, 7.9158E+03,
     + 7.9307E+03, 7.9457E+03, 7.9607E+03, 7.9756E+03, 7.9906E+03,
     +  8.0057E+03, 8.0207E+03, 8.0357E+03, 8.0508E+03, 8.0659E+03,
     + 8.0810E+03, 8.0961E+03, 8.1112E+03, 8.1263E+03, 8.1415E+03,
     +  8.1566E+03, 8.1718E+03, 8.1870E+03, 8.2022E+03, 8.2174E+03,
     + 8.2327E+03, 8.2479E+03, 8.2632E+03, 8.2785E+03, 8.2938E+03,
     +  8.3091E+03, 8.3244E+03, 8.3398E+03, 8.3551E+03, 8.3705E+03,
     + 8.3859E+03, 8.4013E+03, 8.4167E+03, 8.4321E+03, 8.4476E+03,
     +  8.4630E+03, 8.4785E+03, 8.4940E+03, 8.5095E+03, 8.5250E+03,
     + 8.5406E+03, 8.5561E+03, 8.5717E+03, 8.5872E+03, 8.6028E+03,
     +  8.6185E+03, 8.6341E+03, 8.6497E+03, 8.6654E+03, 8.6810E+03,
     + 8.6967E+03, 8.7124E+03, 8.7281E+03, 8.7439E+03, 8.7596E+03,
     +  8.7754E+03, 8.7911E+03, 8.8069E+03, 8.8227E+03, 8.8386E+03,
     + 8.8544E+03, 8.8702E+03, 8.8861E+03, 8.9020E+03, 8.9179E+03,
     +  8.9338E+03, 8.9497E+03, 8.9656E+03, 8.9816E+03, 8.9976E+03,
     + 9.0136E+03, 9.0296E+03, 9.0456E+03, 9.0616E+03, 9.0776E+03,
     +  9.0937E+03, 9.1098E+03, 9.1259E+03, 9.1420E+03, 9.1581E+03,
     + 9.1742E+03, 9.1904E+03, 9.2065E+03, 9.2227E+03, 9.2389E+03,
     +  9.2551E+03, 9.2713E+03, 9.2876E+03, 9.3038E+03, 9.3201E+03,
     + 9.3364E+03, 9.3527E+03, 9.3690E+03, 9.3853E+03, 9.4016E+03,
     +  9.4180E+03, 9.4344E+03, 9.4508E+03, 9.4672E+03, 9.4836E+03,
     + 9.5000E+03, 9.5165E+03, 9.5329E+03, 9.5494E+03, 9.5659E+03,
     +  9.5824E+03, 9.5989E+03, 9.6155E+03, 9.6320E+03, 9.6486E+03,
     + 9.6652E+03, 9.6818E+03, 9.6984E+03, 9.7150E+03, 9.7317E+03,
     +  9.7483E+03, 9.7650E+03, 9.7817E+03, 9.7984E+03, 9.8151E+03,
     + 9.8318E+03, 9.8486E+03, 9.8653E+03, 9.8821E+03, 9.8989E+03,
     +  9.9157E+03, 9.9325E+03, 9.9494E+03, 9.9662E+03, 9.9831E+03,
     + 1.0000E+04, 1.0017E+04, 1.0034E+04, 1.0051E+04, 1.0068E+04,
     +  1.0085E+04, 1.0102E+04, 1.0119E+04, 1.0136E+04, 1.0153E+04,
     + 1.0170E+04, 1.0187E+04, 1.0204E+04, 1.0221E+04, 1.0238E+04,
     +  1.0255E+04, 1.0272E+04, 1.0289E+04, 1.0306E+04, 1.0324E+04,
     + 1.0341E+04, 1.0358E+04, 1.0375E+04, 1.0392E+04, 1.0410E+04,
     +  1.0427E+04, 1.0444E+04, 1.0462E+04, 1.0479E+04, 1.0496E+04,
     + 1.0514E+04, 1.0531E+04, 1.0548E+04, 1.0566E+04, 1.0583E+04,
     +  1.0600E+04, 1.0618E+04, 1.0635E+04, 1.0653E+04, 1.0670E+04,
     + 1.0688E+04, 1.0705E+04, 1.0723E+04, 1.0740E+04, 1.0758E+04,
     +  1.0775E+04, 1.0793E+04, 1.0811E+04, 1.0828E+04, 1.0846E+04,
     + 1.0863E+04, 1.0881E+04, 1.0899E+04, 1.0916E+04, 1.0934E+04,
     +  1.0952E+04, 1.0970E+04, 1.0987E+04, 1.1005E+04, 1.1023E+04,
     + 1.1041E+04, 1.1058E+04, 1.1076E+04, 1.1094E+04, 1.1112E+04,
     +  1.1130E+04, 1.1148E+04, 1.1166E+04, 1.1183E+04, 1.1201E+04,
     + 1.1219E+04, 1.1237E+04, 1.1255E+04, 1.1273E+04, 1.1291E+04,
     +  1.1309E+04, 1.1327E+04, 1.1345E+04, 1.1363E+04, 1.1381E+04,
     + 1.1400E+04, 1.1418E+04, 1.1436E+04, 1.1454E+04, 1.1472E+04,
     +  1.1490E+04, 1.1508E+04, 1.1527E+04, 1.1545E+04, 1.1563E+04,
     + 1.1581E+04, 1.1600E+04, 1.1618E+04, 1.1636E+04, 1.1654E+04,
     +  1.1673E+04, 1.1691E+04, 1.1709E+04, 1.1728E+04, 1.1746E+04,
     + 1.1764E+04, 1.1783E+04, 1.1801E+04, 1.1820E+04, 1.1838E+04,
     +  1.1857E+04, 1.1875E+04, 1.1894E+04, 1.1912E+04, 1.1931E+04,
     + 1.1949E+04, 1.1968E+04, 1.1986E+04, 1.2005E+04, 1.2024E+04,
     +  1.2042E+04, 1.2061E+04, 1.2079E+04, 1.2098E+04, 1.2117E+04,
     + 1.2136E+04, 1.2154E+04, 1.2173E+04, 1.2192E+04, 1.2210E+04,
     +  1.2229E+04, 1.2248E+04, 1.2267E+04, 1.2286E+04, 1.2304E+04,
     + 1.2323E+04, 1.2342E+04, 1.2361E+04, 1.2380E+04, 1.2399E+04,
     +  1.2418E+04, 1.2437E+04, 1.2456E+04, 1.2475E+04, 1.2494E+04,
     + 1.2513E+04, 1.2532E+04, 1.2551E+04, 1.2570E+04, 1.2589E+04,
     +  1.2608E+04, 1.2627E+04, 1.2646E+04, 1.2665E+04, 1.2684E+04,
     + 1.2703E+04, 1.2723E+04, 1.2742E+04, 1.2761E+04, 1.2780E+04,
     +  1.2799E+04, 1.2819E+04, 1.2838E+04, 1.2857E+04, 1.2876E+04,
     + 1.2896E+04, 1.2915E+04, 1.2934E+04, 1.2954E+04, 1.2973E+04,
     +  1.2992E+04, 1.3012E+04, 1.3031E+04, 1.3051E+04, 1.3070E+04,
     + 1.3090E+04, 1.3109E+04, 1.3128E+04, 1.3148E+04, 1.3167E+04,
     +  1.3187E+04, 1.3207E+04, 1.3226E+04, 1.3246E+04, 1.3265E+04,
     + 1.3285E+04, 1.3304E+04, 1.3324E+04, 1.3344E+04, 1.3363E+04,
     +  1.3383E+04, 1.3403E+04, 1.3422E+04, 1.3442E+04, 1.3462E+04,
     + 1.3482E+04, 1.3501E+04, 1.3521E+04, 1.3541E+04, 1.3561E+04,
     +  1.3581E+04, 1.3601E+04, 1.3620E+04, 1.3640E+04, 1.3660E+04,
     + 1.3680E+04, 1.3700E+04, 1.3720E+04, 1.3740E+04, 1.3760E+04,
     +  1.3780E+04, 1.3800E+04, 1.3820E+04, 1.3840E+04, 1.3860E+04,
     + 1.3880E+04, 1.3900E+04, 1.3920E+04, 1.3940E+04, 1.3960E+04,
     +  1.3980E+04, 1.4001E+04, 1.4021E+04, 1.4041E+04, 1.4061E+04,
     + 1.4081E+04, 1.4101E+04, 1.4122E+04, 1.4142E+04, 1.4162E+04,
     +  1.4182E+04, 1.4203E+04, 1.4223E+04, 1.4243E+04, 1.4264E+04,
     + 1.4284E+04, 1.4304E+04, 1.4325E+04, 1.4345E+04, 1.4366E+04,
     +  1.4386E+04, 1.4407E+04, 1.4427E+04, 1.4447E+04, 1.4468E+04,
     + 1.4488E+04, 1.4509E+04, 1.4530E+04, 1.4550E+04, 1.4571E+04,
     +  1.4591E+04, 1.4612E+04, 1.4632E+04, 1.4653E+04, 1.4674E+04,
     + 1.4694E+04, 1.4715E+04, 1.4736E+04, 1.4756E+04, 1.4777E+04,
     +  1.4798E+04, 1.4819E+04, 1.4839E+04, 1.4860E+04, 1.4881E+04,
     + 1.4902E+04, 1.4923E+04, 1.4943E+04, 1.4964E+04, 1.4985E+04,
     +  1.5006E+04, 1.5027E+04, 1.5048E+04, 1.5069E+04, 1.5090E+04,
     + 1.5111E+04, 1.5132E+04, 1.5153E+04, 1.5174E+04, 1.5195E+04,
     +  1.5216E+04, 1.5237E+04, 1.5258E+04, 1.5279E+04, 1.5300E+04,
     + 1.5321E+04, 1.5342E+04, 1.5363E+04, 1.5384E+04, 1.5406E+04,
     +  1.5427E+04, 1.5448E+04, 1.5469E+04, 1.5490E+04, 1.5512E+04,
     + 1.5533E+04, 1.5554E+04, 1.5575E+04, 1.5597E+04, 1.5618E+04,
     +  1.5639E+04, 1.5661E+04, 1.5682E+04, 1.5703E+04, 1.5725E+04,
     + 1.5746E+04, 1.5768E+04, 1.5789E+04, 1.5810E+04, 1.5832E+04,
     +  1.5853E+04, 1.5875E+04, 1.5896E+04, 1.5918E+04, 1.5939E+04,
     + 1.5961E+04, 1.5983E+04, 1.6004E+04, 1.6026E+04, 1.6047E+04,
     +  1.6069E+04, 1.6091E+04, 1.6112E+04, 1.6134E+04, 1.6156E+04,
     + 1.6177E+04, 1.6199E+04, 1.6221E+04, 1.6243E+04, 1.6264E+04,
     +  1.6286E+04, 1.6308E+04, 1.6330E+04, 1.6352E+04, 1.6373E+04,
     + 1.6395E+04, 1.6417E+04, 1.6439E+04, 1.6461E+04, 1.6483E+04,
     +  1.6505E+04, 1.6527E+04, 1.6549E+04, 1.6571E+04, 1.6593E+04,
     + 1.6615E+04, 1.6637E+04, 1.6659E+04, 1.6681E+04, 1.6703E+04,
     +  1.6725E+04, 1.6747E+04, 1.6769E+04, 1.6791E+04, 1.6813E+04,
     + 1.6835E+04, 1.6858E+04, 1.6880E+04, 1.6902E+04, 1.6924E+04,
     +  1.6946E+04, 1.6969E+04, 1.6991E+04, 1.7013E+04, 1.7035E+04,
     + 1.7058E+04, 1.7080E+04, 1.7102E+04, 1.7125E+04, 1.7147E+04,
     +  1.7169E+04, 1.7192E+04, 1.7214E+04, 1.7237E+04, 1.7259E+04,
     + 1.7281E+04, 1.7304E+04, 1.7326E+04, 1.7349E+04, 1.7371E+04,
     +  1.7394E+04, 1.7416E+04, 1.7439E+04, 1.7462E+04, 1.7484E+04,
     + 1.7507E+04, 1.7529E+04, 1.7552E+04, 1.7575E+04, 1.7597E+04,
     +  1.7620E+04, 1.7643E+04, 1.7665E+04, 1.7688E+04, 1.7711E+04,
     + 1.7733E+04, 1.7756E+04, 1.7779E+04, 1.7802E+04, 1.7825E+04,
     +  1.7847E+04, 1.7870E+04, 1.7893E+04, 1.7916E+04, 1.7939E+04,
     + 1.7962E+04, 1.7985E+04, 1.8007E+04, 1.8030E+04, 1.8053E+04,
     +  1.8076E+04, 1.8099E+04, 1.8122E+04, 1.8145E+04, 1.8168E+04,
     + 1.8191E+04, 1.8214E+04, 1.8237E+04, 1.8260E+04, 1.8284E+04,
     +  1.8307E+04, 1.8330E+04, 1.8353E+04, 1.8376E+04, 1.8399E+04,
     + 1.8422E+04, 1.8446E+04, 1.8469E+04, 1.8492E+04, 1.8515E+04,
     +  1.8539E+04, 1.8562E+04, 1.8585E+04, 1.8608E+04, 1.8632E+04,
     + 1.8655E+04, 1.8678E+04, 1.8702E+04, 1.8725E+04, 1.8749E+04,
     +  1.8772E+04, 1.8795E+04, 1.8819E+04, 1.8842E+04, 1.8866E+04,
     + 1.8889E+04, 1.8913E+04, 1.8936E+04, 1.8960E+04, 1.8983E+04,
     +  1.9007E+04, 1.9030E+04, 1.9054E+04, 1.9077E+04, 1.9101E+04,
     + 1.9125E+04, 1.9148E+04, 1.9172E+04, 1.9196E+04, 1.9219E+04,
     +  1.9243E+04, 1.9267E+04, 1.9290E+04, 1.9314E+04, 1.9338E+04,
     + 1.9362E+04, 1.9385E+04, 1.9409E+04, 1.9433E+04, 1.9457E+04,
     +  1.9481E+04, 1.9505E+04, 1.9528E+04, 1.9552E+04, 1.9576E+04,
     + 1.9600E+04, 1.9624E+04, 1.9648E+04, 1.9672E+04, 1.9696E+04,
     +  1.9720E+04, 1.9744E+04, 1.9768E+04, 1.9792E+04, 1.9816E+04,
     + 1.9840E+04, 1.9864E+04, 1.9888E+04, 1.9912E+04, 1.9936E+04,
     +  1.9961E+04, 1.9985E+04, 2.0009E+04, 2.0033E+04, 2.0057E+04,
     + 2.0081E+04, 2.0106E+04, 2.0130E+04, 2.0154E+04, 2.0178E+04,
     +  2.0203E+04, 2.0227E+04, 2.0251E+04, 2.0276E+04, 2.0300E+04,
     + 2.0324E+04, 2.0349E+04, 2.0373E+04, 2.0397E+04, 2.0422E+04,
     +  2.0446E+04, 2.0471E+04, 2.0495E+04, 2.0519E+04, 2.0544E+04,
     + 2.0568E+04, 2.0593E+04, 2.0617E+04, 2.0642E+04, 2.0667E+04,
     +  2.0691E+04, 2.0716E+04, 2.0740E+04, 2.0765E+04, 2.0790E+04,
     + 2.0814E+04, 2.0839E+04, 2.0863E+04, 2.0888E+04, 2.0913E+04,
     +  2.0938E+04, 2.0962E+04, 2.0987E+04, 2.1012E+04, 2.1036E+04,
     + 2.1061E+04, 2.1086E+04, 2.1111E+04, 2.1136E+04, 2.1161E+04,
     +  2.1185E+04, 2.1210E+04, 2.1235E+04, 2.1260E+04, 2.1285E+04,
     + 2.1310E+04, 2.1335E+04, 2.1360E+04, 2.1385E+04, 2.1410E+04,
     +  2.1435E+04, 2.1460E+04, 2.1485E+04, 2.1510E+04, 2.1535E+04,
     + 2.1560E+04, 2.1585E+04, 2.1610E+04, 2.1635E+04, 2.1660E+04,
     +  2.1685E+04, 2.1710E+04, 2.1736E+04, 2.1761E+04, 2.1786E+04,
     + 2.1811E+04, 2.1836E+04, 2.1862E+04, 2.1887E+04, 2.1912E+04,
     +  2.1937E+04, 2.1963E+04, 2.1988E+04, 2.2013E+04, 2.2039E+04,
     + 2.2064E+04, 2.2089E+04, 2.2115E+04, 2.2140E+04, 2.2166E+04,
     +  2.2191E+04, 2.2216E+04, 2.2242E+04, 2.2267E+04, 2.2293E+04,
     + 2.2318E+04, 2.2344E+04, 2.2369E+04, 2.2395E+04, 2.2420E+04,
     +  2.2446E+04, 2.2472E+04, 2.2497E+04, 2.2523E+04, 2.2548E+04,
     + 2.2574E+04, 2.2600E+04, 2.2625E+04, 2.2651E+04, 2.2677E+04,
     +  2.2702E+04, 2.2728E+04, 2.2754E+04, 2.2780E+04, 2.2805E+04,
     + 2.2831E+04, 2.2857E+04, 2.2883E+04, 2.2909E+04, 2.2934E+04,
     +  2.2960E+04, 2.2986E+04, 2.3012E+04, 2.3038E+04, 2.3064E+04,
     + 2.3090E+04, 2.3116E+04, 2.3142E+04, 2.3167E+04, 2.3193E+04,
     +  2.3219E+04, 2.3245E+04, 2.3271E+04, 2.3297E+04, 2.3324E+04,
     + 2.3350E+04, 2.3376E+04, 2.3402E+04, 2.3428E+04, 2.3454E+04,
     +  2.3480E+04, 2.3506E+04, 2.3532E+04, 2.3559E+04, 2.3585E+04,
     + 2.3611E+04, 2.3637E+04, 2.3663E+04, 2.3690E+04, 2.3716E+04,
     +  2.3742E+04, 2.3768E+04, 2.3795E+04, 2.3821E+04, 2.3847E+04,
     + 2.3874E+04, 2.3900E+04, 2.3926E+04, 2.3953E+04, 2.3979E+04,
     +  2.4005E+04, 2.4032E+04, 2.4058E+04, 2.4085E+04, 2.4111E+04,
     + 2.4138E+04, 2.4164E+04, 2.4191E+04, 2.4217E+04, 2.4244E+04,
     +  2.4270E+04, 2.4297E+04, 2.4323E+04, 2.4350E+04, 2.4377E+04,
     + 2.4403E+04, 2.4430E+04, 2.4457E+04, 2.4483E+04, 2.4510E+04,
     +  2.4537E+04, 2.4563E+04, 2.4590E+04, 2.4617E+04, 2.4643E+04,
     + 2.4670E+04, 2.4697E+04, 2.4724E+04, 2.4750E+04, 2.4777E+04,
     +  2.4804E+04, 2.4831E+04, 2.4858E+04, 2.4885E+04, 2.4912E+04,
     + 2.4938E+04, 2.4965E+04, 2.4992E+04, 2.5019E+04, 2.5046E+04,
     +  2.5073E+04, 2.5100E+04, 2.5127E+04, 2.5154E+04, 2.5181E+04,
     + 2.5208E+04, 2.5235E+04, 2.5262E+04, 2.5289E+04, 2.5316E+04,
     +  2.5343E+04, 2.5371E+04, 2.5398E+04, 2.5425E+04, 2.5452E+04,
     + 2.5479E+04, 2.5506E+04, 2.5534E+04, 2.5561E+04, 2.5588E+04,
     +  2.5615E+04, 2.5642E+04, 2.5670E+04, 2.5697E+04, 2.5724E+04,
     + 2.5752E+04, 2.5779E+04, 2.5806E+04, 2.5834E+04, 2.5861E+04,
     +  2.5888E+04, 2.5916E+04, 2.5943E+04, 2.5970E+04, 2.5998E+04,
     + 2.6025E+04, 2.6053E+04, 2.6080E+04, 2.6108E+04, 2.6135E+04,
     +  2.6163E+04, 2.6190E+04, 2.6218E+04, 2.6245E+04, 2.6273E+04,
     + 2.6301E+04, 2.6328E+04, 2.6356E+04, 2.6383E+04, 2.6411E+04,
     +  2.6439E+04, 2.6466E+04, 2.6494E+04, 2.6522E+04, 2.6549E+04,
     + 2.6577E+04, 2.6605E+04, 2.6632E+04, 2.6660E+04, 2.6688E+04,
     +  2.6716E+04, 2.6744E+04, 2.6771E+04, 2.6799E+04, 2.6827E+04,
     + 2.6855E+04, 2.6883E+04, 2.6911E+04, 2.6939E+04, 2.6966E+04,
     +  2.6994E+04, 2.7022E+04, 2.7050E+04, 2.7078E+04, 2.7106E+04,
     + 2.7134E+04, 2.7162E+04, 2.7190E+04, 2.7218E+04, 2.7246E+04,
     +  2.7274E+04, 2.7302E+04, 2.7330E+04, 2.7358E+04, 2.7387E+04,
     + 2.7415E+04, 2.7443E+04, 2.7471E+04, 2.7499E+04, 2.7527E+04,
     +  2.7556E+04, 2.7584E+04, 2.7612E+04, 2.7640E+04, 2.7668E+04,
     + 2.7697E+04, 2.7725E+04, 2.7753E+04, 2.7781E+04, 2.7810E+04,
     +  2.7838E+04, 2.7866E+04, 2.7895E+04, 2.7923E+04, 2.7952E+04,
     + 2.7980E+04, 2.8008E+04, 2.8037E+04, 2.8065E+04, 2.8094E+04,
     +  2.8122E+04, 2.8151E+04, 2.8179E+04, 2.8207E+04, 2.8236E+04,
     + 2.8265E+04, 2.8293E+04, 2.8322E+04, 2.8350E+04, 2.8379E+04,
     +  2.8407E+04, 2.8436E+04, 2.8465E+04, 2.8493E+04, 2.8522E+04,
     + 2.8550E+04, 2.8579E+04, 2.8608E+04, 2.8637E+04, 2.8665E+04,
     +  2.8694E+04, 2.8723E+04, 2.8751E+04, 2.8780E+04, 2.8809E+04,
     + 2.8838E+04, 2.8867E+04, 2.8895E+04, 2.8924E+04, 2.8953E+04,
     +  2.8982E+04, 2.9011E+04, 2.9040E+04, 2.9069E+04, 2.9097E+04,
     + 2.9126E+04, 2.9155E+04, 2.9184E+04, 2.9213E+04, 2.9242E+04,
     +  2.9271E+04, 2.9300E+04, 2.9329E+04, 2.9358E+04, 2.9387E+04,
     + 2.9416E+04, 2.9445E+04, 2.9474E+04, 2.9504E+04, 2.9533E+04,
     +  2.9562E+04, 2.9591E+04, 2.9620E+04, 2.9649E+04, 2.9678E+04,
     + 2.9708E+04, 2.9737E+04, 2.9766E+04, 2.9795E+04, 2.9824E+04,
     +  2.9854E+04, 2.9883E+04, 2.9912E+04, 2.9941E+04, 2.9971E+04,
     + 3.0000E+04, 3.0029E+04, 3.0059E+04, 3.0088E+04, 3.0117E+04,
     +  3.0147E+04, 3.0176E+04, 3.0206E+04, 3.0235E+04, 3.0264E+04,
     + 3.0294E+04, 3.0323E+04, 3.0353E+04, 3.0382E+04, 3.0412E+04,
     +  3.0441E+04, 3.0471E+04, 3.0500E+04, 3.0530E+04, 3.0560E+04,
     + 3.0589E+04, 3.0619E+04, 3.0648E+04, 3.0678E+04, 3.0708E+04,
     +  3.0737E+04, 3.0767E+04, 3.0797E+04, 3.0826E+04, 3.0856E+04,
     + 3.0886E+04, 3.0915E+04, 3.0945E+04, 3.0975E+04, 3.1005E+04,
     +  3.1034E+04, 3.1064E+04, 3.1094E+04, 3.1124E+04, 3.1154E+04,
     + 3.1183E+04, 3.1213E+04, 3.1243E+04, 3.1273E+04, 3.1303E+04,
     +  3.1333E+04, 3.1363E+04, 3.1393E+04, 3.1423E+04, 3.1452E+04,
     + 3.1482E+04, 3.1512E+04, 3.1542E+04, 3.1572E+04, 3.1602E+04,
     +  3.1632E+04, 3.1662E+04, 3.1693E+04, 3.1723E+04, 3.1753E+04,
     + 3.1783E+04, 3.1813E+04, 3.1843E+04, 3.1873E+04, 3.1903E+04,
     +  3.1933E+04, 3.1964E+04, 3.1994E+04, 3.2024E+04, 3.2054E+04,
     + 3.2084E+04, 3.2115E+04, 3.2145E+04, 3.2175E+04, 3.2205E+04,
     +  3.2236E+04, 3.2266E+04, 3.2296E+04, 3.2327E+04, 3.2357E+04,
     + 3.2387E+04, 3.2418E+04, 3.2448E+04, 3.2478E+04, 3.2509E+04,
     +  3.2539E+04, 3.2570E+04, 3.2600E+04, 3.2631E+04, 3.2661E+04,
     + 3.2692E+04, 3.2722E+04, 3.2753E+04, 3.2783E+04, 3.2814E+04,
     +  3.2844E+04, 3.2875E+04, 3.2905E+04, 3.2936E+04, 3.2966E+04,
     + 3.2997E+04, 3.3028E+04, 3.3058E+04, 3.3089E+04, 3.3120E+04,
     +  3.3150E+04, 3.3181E+04, 3.3212E+04, 3.3242E+04, 3.3273E+04,
     + 3.3304E+04, 3.3334E+04, 3.3365E+04, 3.3396E+04, 3.3427E+04,
     +  3.3458E+04, 3.3488E+04, 3.3519E+04, 3.3550E+04, 3.3581E+04,
     + 3.3612E+04, 3.3643E+04, 3.3674E+04, 3.3704E+04, 3.3735E+04,
     +  3.3766E+04, 3.3797E+04, 3.3828E+04, 3.3859E+04, 3.3890E+04,
     + 3.3921E+04, 3.3952E+04, 3.3983E+04, 3.4014E+04, 3.4045E+04,
     +  3.4076E+04, 3.4107E+04, 3.4138E+04, 3.4169E+04, 3.4200E+04,
     + 3.4232E+04, 3.4263E+04, 3.4294E+04, 3.4325E+04, 3.4356E+04,
     +  3.4387E+04, 3.4418E+04, 3.4450E+04, 3.4481E+04, 3.4512E+04,
     + 3.4543E+04, 3.4575E+04, 3.4606E+04, 3.4637E+04, 3.4668E+04,
     +  3.4700E+04, 3.4731E+04, 3.4762E+04, 3.4794E+04, 3.4825E+04,
     + 3.4856E+04, 3.4888E+04, 3.4919E+04, 3.4950E+04, 3.4982E+04,
     +  3.5013E+04, 3.5045E+04, 3.5076E+04, 3.5108E+04, 3.5139E+04,
     + 3.5171E+04, 3.5202E+04, 3.5234E+04, 3.5265E+04, 3.5297E+04,
     +  3.5328E+04, 3.5360E+04, 3.5391E+04, 3.5423E+04, 3.5454E+04,
     + 3.5486E+04, 3.5518E+04, 3.5549E+04, 3.5581E+04, 3.5613E+04,
     +  3.5644E+04, 3.5676E+04, 3.5708E+04, 3.5739E+04, 3.5771E+04,
     + 3.5803E+04, 3.5835E+04, 3.5866E+04, 3.5898E+04, 3.5930E+04,
     +  3.5962E+04, 3.5993E+04, 3.6025E+04, 3.6057E+04, 3.6089E+04,
     + 3.6121E+04, 3.6153E+04, 3.6184E+04, 3.6216E+04, 3.6248E+04,
     +  3.6280E+04, 3.6312E+04, 3.6344E+04, 3.6376E+04, 3.6408E+04,
     + 3.6440E+04, 3.6472E+04, 3.6504E+04, 3.6536E+04, 3.6568E+04,
     +  3.6600E+04, 3.6632E+04, 3.6664E+04, 3.6696E+04, 3.6728E+04,
     + 3.6760E+04, 3.6792E+04, 3.6825E+04, 3.6857E+04, 3.6889E+04,
     +  3.6921E+04, 3.6953E+04, 3.6985E+04, 3.7018E+04, 3.7050E+04,
     + 3.7082E+04, 3.7114E+04, 3.7146E+04, 3.7179E+04, 3.7211E+04,
     +  3.7243E+04, 3.7276E+04, 3.7308E+04, 3.7340E+04, 3.7372E+04,
     + 3.7405E+04, 3.7437E+04, 3.7470E+04, 3.7502E+04, 3.7534E+04,
     +  3.7567E+04, 3.7599E+04, 3.7632E+04, 3.7664E+04, 3.7696E+04,
     + 3.7729E+04, 3.7761E+04, 3.7794E+04, 3.7826E+04, 3.7859E+04,
     +  3.7891E+04, 3.7924E+04, 3.7956E+04, 3.7989E+04, 3.8022E+04,
     + 3.8054E+04, 3.8087E+04, 3.8119E+04, 3.8152E+04, 3.8185E+04,
     +  3.8217E+04, 3.8250E+04, 3.8283E+04, 3.8315E+04, 3.8348E+04,
     + 3.8381E+04, 3.8413E+04, 3.8446E+04, 3.8479E+04, 3.8511E+04,
     +  3.8544E+04, 3.8577E+04, 3.8610E+04, 3.8643E+04, 3.8675E+04,
     + 3.8708E+04, 3.8741E+04, 3.8774E+04, 3.8807E+04, 3.8840E+04,
     +  3.8873E+04, 3.8905E+04, 3.8938E+04, 3.8971E+04, 3.9004E+04,
     + 3.9037E+04, 3.9070E+04, 3.9103E+04, 3.9136E+04, 3.9169E+04,
     +  3.9202E+04, 3.9235E+04, 3.9268E+04, 3.9301E+04, 3.9334E+04,
     + 3.9367E+04, 3.9400E+04, 3.9433E+04, 3.9466E+04, 3.9499E+04,
     +  3.9533E+04, 3.9566E+04, 3.9599E+04, 3.9632E+04, 3.9665E+04,
     + 3.9698E+04, 3.9732E+04, 3.9765E+04, 3.9798E+04, 3.9831E+04,
     +  3.9864E+04, 3.9898E+04, 3.9931E+04, 3.9964E+04, 3.9997E+04,
     + 4.0031E+04, 4.0064E+04, 4.0097E+04, 4.0131E+04, 4.0164E+04,
     +  4.0197E+04, 4.0231E+04, 4.0264E+04, 4.0297E+04, 4.0331E+04,
     + 4.0364E+04, 4.0398E+04, 4.0431E+04, 4.0465E+04, 4.0498E+04,
     +  4.0532E+04, 4.0565E+04, 4.0598E+04, 4.0632E+04, 4.0665E+04,
     + 4.0699E+04, 4.0733E+04, 4.0766E+04, 4.0800E+04, 4.0833E+04,
     +  4.0867E+04, 4.0900E+04, 4.0934E+04, 4.0968E+04, 4.1001E+04,
     + 4.1035E+04, 4.1069E+04, 4.1102E+04, 4.1136E+04, 4.1170E+04,
     +  4.1203E+04, 4.1237E+04, 4.1271E+04, 4.1304E+04, 4.1338E+04,
     + 4.1372E+04, 4.1406E+04, 4.1440E+04, 4.1473E+04, 4.1507E+04,
     +  4.1541E+04, 4.1575E+04, 4.1609E+04, 4.1642E+04, 4.1676E+04,
     + 4.1710E+04, 4.1744E+04, 4.1778E+04, 4.1812E+04, 4.1846E+04,
     +  4.1880E+04, 4.1914E+04, 4.1948E+04, 4.1982E+04, 4.2016E+04,
     + 4.2050E+04, 4.2084E+04, 4.2118E+04, 4.2152E+04, 4.2186E+04,
     +  4.2220E+04, 4.2254E+04, 4.2288E+04, 4.2322E+04, 4.2356E+04,
     + 4.2390E+04, 4.2424E+04, 4.2458E+04, 4.2492E+04, 4.2527E+04,
     +  4.2561E+04, 4.2595E+04, 4.2629E+04, 4.2663E+04, 4.2698E+04,
     + 4.2732E+04, 4.2766E+04, 4.2800E+04, 4.2835E+04, 4.2869E+04,
     +  4.2903E+04, 4.2937E+04, 4.2972E+04, 4.3006E+04, 4.3040E+04,
     + 4.3075E+04, 4.3109E+04, 4.3143E+04, 4.3178E+04, 4.3212E+04,
     +  4.3246E+04, 4.3281E+04, 4.3315E+04, 4.3350E+04, 4.3384E+04,
     + 4.3419E+04, 4.3453E+04, 4.3487E+04, 4.3522E+04, 4.3556E+04,
     +  4.3591E+04, 4.3625E+04, 4.3660E+04, 4.3695E+04, 4.3729E+04,
     + 4.3764E+04, 4.3798E+04, 4.3833E+04, 4.3867E+04, 4.3902E+04,
     +  4.3937E+04, 4.3971E+04, 4.4006E+04, 4.4041E+04, 4.4075E+04,
     + 4.4110E+04, 4.4145E+04, 4.4179E+04, 4.4214E+04, 4.4249E+04,
     +  4.4283E+04, 4.4318E+04, 4.4353E+04, 4.4388E+04, 4.4422E+04,
     + 4.4457E+04, 4.4492E+04, 4.4527E+04, 4.4562E+04, 4.4596E+04,
     +  4.4631E+04, 4.4666E+04, 4.4701E+04, 4.4736E+04, 4.4771E+04,
     + 4.4806E+04, 4.4841E+04, 4.4875E+04, 4.4910E+04, 4.4945E+04,
     +  4.4980E+04, 4.5015E+04, 4.5050E+04, 4.5085E+04, 4.5120E+04,
     + 4.5155E+04, 4.5190E+04, 4.5225E+04, 4.5260E+04, 4.5295E+04,
     +  4.5330E+04, 4.5366E+04, 4.5401E+04, 4.5436E+04, 4.5471E+04,
     + 4.5506E+04, 4.5541E+04, 4.5576E+04, 4.5611E+04, 4.5647E+04,
     +  4.5682E+04, 4.5717E+04, 4.5752E+04, 4.5787E+04, 4.5822E+04,
     + 4.5858E+04, 4.5893E+04, 4.5928E+04, 4.5963E+04, 4.5999E+04,
     +  4.6034E+04, 4.6069E+04, 4.6105E+04, 4.6140E+04, 4.6175E+04,
     + 4.6211E+04, 4.6246E+04, 4.6281E+04, 4.6317E+04, 4.6352E+04,
     +  4.6387E+04, 4.6423E+04, 4.6458E+04, 4.6494E+04, 4.6529E+04,
     + 4.6565E+04, 4.6600E+04, 4.6636E+04, 4.6671E+04, 4.6707E+04,
     +  4.6742E+04, 4.6778E+04, 4.6813E+04, 4.6849E+04, 4.6884E+04,
     + 4.6920E+04, 4.6955E+04, 4.6991E+04, 4.7026E+04, 4.7062E+04,
     +  4.7098E+04, 4.7133E+04, 4.7169E+04, 4.7205E+04, 4.7240E+04,
     + 4.7276E+04, 4.7312E+04, 4.7347E+04, 4.7383E+04, 4.7419E+04,
     +  4.7454E+04, 4.7490E+04, 4.7526E+04, 4.7562E+04, 4.7597E+04,
     + 4.7633E+04, 4.7669E+04, 4.7705E+04, 4.7741E+04, 4.7776E+04,
     +  4.7812E+04, 4.7848E+04, 4.7884E+04, 4.7920E+04, 4.7956E+04,
     + 4.7992E+04, 4.8027E+04, 4.8063E+04, 4.8099E+04, 4.8135E+04,
     +  4.8171E+04, 4.8207E+04, 4.8243E+04, 4.8279E+04, 4.8315E+04,
     + 4.8351E+04, 4.8387E+04, 4.8423E+04, 4.8459E+04, 4.8495E+04,
     +  4.8531E+04, 4.8567E+04, 4.8603E+04, 4.8639E+04, 4.8675E+04,
     + 4.8711E+04, 4.8748E+04, 4.8784E+04, 4.8820E+04, 4.8856E+04,
     +  4.8892E+04, 4.8928E+04, 4.8964E+04, 4.9001E+04, 4.9037E+04,
     + 4.9073E+04, 4.9109E+04, 4.9145E+04, 4.9182E+04, 4.9218E+04,
     +  4.9254E+04, 4.9291E+04, 4.9327E+04, 4.9363E+04, 4.9399E+04,
     + 4.9436E+04, 4.9472E+04, 4.9508E+04, 4.9545E+04, 4.9581E+04,
     +  4.9617E+04, 4.9654E+04, 4.9690E+04, 4.9727E+04, 4.9763E+04,
     + 4.9799E+04, 4.9836E+04, 4.9872E+04, 4.9909E+04, 4.9945E+04,
     +  4.9982E+04, 5.0018E+04, 5.0055E+04, 5.0091E+04, 5.0128E+04,
     + 5.0164E+04, 5.0201E+04, 5.0237E+04, 5.0274E+04, 5.0310E+04,
     +  5.0347E+04, 5.0384E+04, 5.0420E+04, 5.0457E+04, 5.0493E+04,
     + 5.0530E+04, 5.0567E+04, 5.0603E+04, 5.0640E+04, 5.0677E+04,
     +  5.0713E+04, 5.0750E+04, 5.0787E+04, 5.0823E+04, 5.0860E+04/
      data (QofT( 2,J),J=2409,4510)/ 5.0897E+04, 5.0934E+04,
     +  5.0970E+04, 5.1007E+04, 5.1044E+04, 5.1081E+04, 5.1118E+04,
     +  5.1154E+04,
     + 5.1191E+04, 5.1228E+04, 5.1265E+04, 5.1302E+04, 5.1339E+04,
     +  5.1375E+04, 5.1412E+04, 5.1449E+04, 5.1486E+04, 5.1523E+04,
     + 5.1560E+04, 5.1597E+04, 5.1634E+04, 5.1671E+04, 5.1708E+04,
     +  5.1745E+04, 5.1782E+04, 5.1819E+04, 5.1856E+04, 5.1893E+04,
     + 5.1930E+04, 5.1967E+04, 5.2004E+04, 5.2041E+04, 5.2078E+04,
     +  5.2115E+04, 5.2152E+04, 5.2189E+04, 5.2226E+04, 5.2263E+04,
     + 5.2300E+04, 5.2338E+04, 5.2375E+04, 5.2412E+04, 5.2449E+04,
     +  5.2486E+04, 5.2523E+04, 5.2561E+04, 5.2598E+04, 5.2635E+04,
     + 5.2672E+04, 5.2710E+04, 5.2747E+04, 5.2784E+04, 5.2821E+04,
     +  5.2859E+04, 5.2896E+04, 5.2933E+04, 5.2971E+04, 5.3008E+04,
     + 5.3045E+04, 5.3083E+04, 5.3120E+04, 5.3157E+04, 5.3195E+04,
     +  5.3232E+04, 5.3269E+04, 5.3307E+04, 5.3344E+04, 5.3382E+04,
     + 5.3419E+04, 5.3457E+04, 5.3494E+04, 5.3532E+04, 5.3569E+04,
     +  5.3606E+04, 5.3644E+04, 5.3681E+04, 5.3719E+04, 5.3757E+04,
     + 5.3794E+04, 5.3832E+04, 5.3869E+04, 5.3907E+04, 5.3944E+04,
     +  5.3982E+04, 5.4020E+04, 5.4057E+04, 5.4095E+04, 5.4132E+04,
     + 5.4170E+04, 5.4208E+04, 5.4245E+04, 5.4283E+04, 5.4321E+04,
     +  5.4358E+04, 5.4396E+04, 5.4434E+04, 5.4472E+04, 5.4509E+04,
     + 5.4547E+04, 5.4585E+04, 5.4622E+04, 5.4660E+04, 5.4698E+04,
     +  5.4736E+04, 5.4774E+04, 5.4811E+04, 5.4849E+04, 5.4887E+04,
     + 5.4925E+04, 5.4963E+04, 5.5001E+04, 5.5039E+04, 5.5076E+04,
     +  5.5114E+04, 5.5152E+04, 5.5190E+04, 5.5228E+04, 5.5266E+04,
     + 5.5304E+04, 5.5342E+04, 5.5380E+04, 5.5418E+04, 5.5456E+04,
     +  5.5494E+04, 5.5532E+04, 5.5570E+04, 5.5608E+04, 5.5646E+04,
     + 5.5684E+04, 5.5722E+04, 5.5760E+04, 5.5798E+04, 5.5836E+04,
     +  5.5874E+04, 5.5912E+04, 5.5950E+04, 5.5989E+04, 5.6027E+04,
     + 5.6065E+04, 5.6103E+04, 5.6141E+04, 5.6179E+04, 5.6217E+04,
     +  5.6256E+04, 5.6294E+04, 5.6332E+04, 5.6370E+04, 5.6409E+04,
     + 5.6447E+04, 5.6485E+04, 5.6523E+04, 5.6562E+04, 5.6600E+04,
     +  5.6638E+04, 5.6676E+04, 5.6715E+04, 5.6753E+04, 5.6791E+04,
     + 5.6830E+04, 5.6868E+04, 5.6906E+04, 5.6945E+04, 5.6983E+04,
     +  5.7022E+04, 5.7060E+04, 5.7098E+04, 5.7137E+04, 5.7175E+04,
     + 5.7214E+04, 5.7252E+04, 5.7291E+04, 5.7329E+04, 5.7367E+04,
     +  5.7406E+04, 5.7444E+04, 5.7483E+04, 5.7521E+04, 5.7560E+04,
     + 5.7599E+04, 5.7637E+04, 5.7676E+04, 5.7714E+04, 5.7753E+04,
     +  5.7791E+04, 5.7830E+04, 5.7869E+04, 5.7907E+04, 5.7946E+04,
     + 5.7984E+04, 5.8023E+04, 5.8062E+04, 5.8100E+04, 5.8139E+04,
     +  5.8178E+04, 5.8216E+04, 5.8255E+04, 5.8294E+04, 5.8333E+04,
     + 5.8371E+04, 5.8410E+04, 5.8449E+04, 5.8487E+04, 5.8526E+04,
     +  5.8565E+04, 5.8604E+04, 5.8643E+04, 5.8681E+04, 5.8720E+04,
     + 5.8759E+04, 5.8798E+04, 5.8837E+04, 5.8876E+04, 5.8914E+04,
     +  5.8953E+04, 5.8992E+04, 5.9031E+04, 5.9070E+04, 5.9109E+04,
     + 5.9148E+04, 5.9187E+04, 5.9226E+04, 5.9265E+04, 5.9304E+04,
     +  5.9343E+04, 5.9382E+04, 5.9421E+04, 5.9460E+04, 5.9499E+04,
     + 5.9538E+04, 5.9577E+04, 5.9616E+04, 5.9655E+04, 5.9694E+04,
     +  5.9733E+04, 5.9772E+04, 5.9811E+04, 5.9850E+04, 5.9889E+04,
     + 5.9928E+04, 5.9967E+04, 6.0007E+04, 6.0046E+04, 6.0085E+04,
     +  6.0124E+04, 6.0163E+04, 6.0202E+04, 6.0241E+04, 6.0281E+04,
     + 6.0320E+04, 6.0359E+04, 6.0398E+04, 6.0438E+04, 6.0477E+04,
     +  6.0516E+04, 6.0555E+04, 6.0595E+04, 6.0634E+04, 6.0673E+04,
     + 6.0712E+04, 6.0752E+04, 6.0791E+04, 6.0830E+04, 6.0870E+04,
     +  6.0909E+04, 6.0948E+04, 6.0988E+04, 6.1027E+04, 6.1067E+04,
     + 6.1106E+04, 6.1145E+04, 6.1185E+04, 6.1224E+04, 6.1264E+04,
     +  6.1303E+04, 6.1343E+04, 6.1382E+04, 6.1422E+04, 6.1461E+04,
     + 6.1500E+04, 6.1540E+04, 6.1580E+04, 6.1619E+04, 6.1659E+04,
     +  6.1698E+04, 6.1738E+04, 6.1777E+04, 6.1817E+04, 6.1856E+04,
     + 6.1896E+04, 6.1935E+04, 6.1975E+04, 6.2015E+04, 6.2054E+04,
     +  6.2094E+04, 6.2134E+04, 6.2173E+04, 6.2213E+04, 6.2253E+04,
     + 6.2292E+04, 6.2332E+04, 6.2372E+04, 6.2411E+04, 6.2451E+04,
     +  6.2491E+04, 6.2530E+04, 6.2570E+04, 6.2610E+04, 6.2650E+04,
     + 6.2690E+04, 6.2729E+04, 6.2769E+04, 6.2809E+04, 6.2849E+04,
     +  6.2889E+04, 6.2928E+04, 6.2968E+04, 6.3008E+04, 6.3048E+04,
     + 6.3088E+04, 6.3128E+04, 6.3167E+04, 6.3207E+04, 6.3247E+04,
     +  6.3287E+04, 6.3327E+04, 6.3367E+04, 6.3407E+04, 6.3447E+04,
     + 6.3487E+04, 6.3527E+04, 6.3567E+04, 6.3607E+04, 6.3647E+04,
     +  6.3687E+04, 6.3727E+04, 6.3767E+04, 6.3807E+04, 6.3847E+04,
     + 6.3887E+04, 6.3927E+04, 6.3967E+04, 6.4007E+04, 6.4047E+04,
     +  6.4087E+04, 6.4127E+04, 6.4167E+04, 6.4208E+04, 6.4248E+04,
     + 6.4288E+04, 6.4328E+04, 6.4368E+04, 6.4408E+04, 6.4448E+04,
     +  6.4489E+04, 6.4529E+04, 6.4569E+04, 6.4609E+04, 6.4649E+04,
     + 6.4690E+04, 6.4730E+04, 6.4770E+04, 6.4810E+04, 6.4851E+04,
     +  6.4891E+04, 6.4931E+04, 6.4972E+04, 6.5012E+04, 6.5052E+04,
     + 6.5092E+04, 6.5133E+04, 6.5173E+04, 6.5213E+04, 6.5254E+04,
     +  6.5294E+04, 6.5335E+04, 6.5375E+04, 6.5415E+04, 6.5456E+04,
     + 6.5496E+04, 6.5537E+04, 6.5577E+04, 6.5617E+04, 6.5658E+04,
     +  6.5698E+04, 6.5739E+04, 6.5779E+04, 6.5820E+04, 6.5860E+04,
     + 6.5901E+04, 6.5941E+04, 6.5982E+04, 6.6022E+04, 6.6063E+04,
     +  6.6103E+04, 6.6144E+04, 6.6184E+04, 6.6225E+04, 6.6266E+04,
     + 6.6306E+04, 6.6347E+04, 6.6387E+04, 6.6428E+04, 6.6469E+04,
     +  6.6509E+04, 6.6550E+04, 6.6590E+04, 6.6631E+04, 6.6672E+04,
     + 6.6712E+04, 6.6753E+04, 6.6794E+04, 6.6835E+04, 6.6875E+04,
     +  6.6916E+04, 6.6957E+04, 6.6997E+04, 6.7038E+04, 6.7079E+04,
     + 6.7120E+04, 6.7160E+04, 6.7201E+04, 6.7242E+04, 6.7283E+04,
     +  6.7324E+04, 6.7364E+04, 6.7405E+04, 6.7446E+04, 6.7487E+04,
     + 6.7528E+04, 6.7569E+04, 6.7610E+04, 6.7650E+04, 6.7691E+04,
     +  6.7732E+04, 6.7773E+04, 6.7814E+04, 6.7855E+04, 6.7896E+04,
     + 6.7937E+04, 6.7978E+04, 6.8019E+04, 6.8060E+04, 6.8101E+04,
     +  6.8142E+04, 6.8183E+04, 6.8224E+04, 6.8265E+04, 6.8306E+04,
     + 6.8347E+04, 6.8388E+04, 6.8429E+04, 6.8470E+04, 6.8511E+04,
     +  6.8552E+04, 6.8593E+04, 6.8634E+04, 6.8675E+04, 6.8716E+04,
     + 6.8757E+04, 6.8799E+04, 6.8840E+04, 6.8881E+04, 6.8922E+04,
     +  6.8963E+04, 6.9004E+04, 6.9046E+04, 6.9087E+04, 6.9128E+04,
     + 6.9169E+04, 6.9210E+04, 6.9252E+04, 6.9293E+04, 6.9334E+04,
     +  6.9375E+04, 6.9416E+04, 6.9458E+04, 6.9499E+04, 6.9540E+04,
     + 6.9582E+04, 6.9623E+04, 6.9664E+04, 6.9705E+04, 6.9747E+04,
     +  6.9788E+04, 6.9829E+04, 6.9871E+04, 6.9912E+04, 6.9954E+04,
     + 6.9995E+04, 7.0036E+04, 7.0078E+04, 7.0119E+04, 7.0160E+04,
     +  7.0202E+04, 7.0243E+04, 7.0285E+04, 7.0326E+04, 7.0368E+04,
     + 7.0409E+04, 7.0451E+04, 7.0492E+04, 7.0534E+04, 7.0575E+04,
     +  7.0617E+04, 7.0658E+04, 7.0700E+04, 7.0741E+04, 7.0783E+04,
     + 7.0824E+04, 7.0866E+04, 7.0907E+04, 7.0949E+04, 7.0990E+04,
     +  7.1032E+04, 7.1074E+04, 7.1115E+04, 7.1157E+04, 7.1198E+04,
     + 7.1240E+04, 7.1282E+04, 7.1323E+04, 7.1365E+04, 7.1407E+04,
     +  7.1448E+04, 7.1490E+04, 7.1532E+04, 7.1573E+04, 7.1615E+04,
     + 7.1657E+04, 7.1698E+04, 7.1740E+04, 7.1782E+04, 7.1824E+04,
     +  7.1865E+04, 7.1907E+04, 7.1949E+04, 7.1991E+04, 7.2033E+04,
     + 7.2074E+04, 7.2116E+04, 7.2158E+04, 7.2200E+04, 7.2242E+04,
     +  7.2283E+04, 7.2325E+04, 7.2367E+04, 7.2409E+04, 7.2451E+04,
     + 7.2493E+04, 7.2535E+04, 7.2577E+04, 7.2618E+04, 7.2660E+04,
     +  7.2702E+04, 7.2744E+04, 7.2786E+04, 7.2828E+04, 7.2870E+04,
     + 7.2912E+04, 7.2954E+04, 7.2996E+04, 7.3038E+04, 7.3080E+04,
     +  7.3122E+04, 7.3164E+04, 7.3206E+04, 7.3248E+04, 7.3290E+04,
     + 7.3332E+04, 7.3374E+04, 7.3416E+04, 7.3458E+04, 7.3500E+04,
     +  7.3542E+04, 7.3585E+04, 7.3627E+04, 7.3669E+04, 7.3711E+04,
     + 7.3753E+04, 7.3795E+04, 7.3837E+04, 7.3879E+04, 7.3922E+04,
     +  7.3964E+04, 7.4006E+04, 7.4048E+04, 7.4090E+04, 7.4133E+04,
     + 7.4175E+04, 7.4217E+04, 7.4259E+04, 7.4301E+04, 7.4344E+04,
     +  7.4386E+04, 7.4428E+04, 7.4471E+04, 7.4513E+04, 7.4555E+04,
     + 7.4597E+04, 7.4640E+04, 7.4682E+04, 7.4724E+04, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_ClO   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 2),QofT( 2,NT),Q(NT)
      data (Tmx(I),I=1,2)/3500.,3500./
      data xgj/ 4.,4./
c...      ClO
c...        --        56
      data (QofT( 1,J),J=1,2408)/ 9.7616E+00, 1.8630E+01, 2.7560E+01,
     +  3.6502E+01, 4.5449E+01, 5.4398E+01, 6.3349E+01, 7.2300E+01,
     + 8.1252E+01, 9.0205E+01, 9.9158E+01, 1.0811E+02, 1.1707E+02,
     +  1.2602E+02, 1.3498E+02, 1.4393E+02, 1.5289E+02, 1.6184E+02,
     + 1.7080E+02, 1.7975E+02, 1.9767E+02, 2.1558E+02, 2.3350E+02,
     +  2.5142E+02, 2.6933E+02, 2.8725E+02, 3.0517E+02, 3.2309E+02,
     + 3.4102E+02, 3.5894E+02, 3.7687E+02, 3.9480E+02, 4.1273E+02,
     +  4.3067E+02, 4.4862E+02, 4.6657E+02, 4.8453E+02, 5.0250E+02,
     + 5.2048E+02, 5.3847E+02, 5.5649E+02, 5.7452E+02, 5.9258E+02,
     +  6.1066E+02, 6.2877E+02, 6.4691E+02, 6.6508E+02, 6.8330E+02,
     + 7.0155E+02, 7.1985E+02, 7.3819E+02, 7.5659E+02, 7.7503E+02,
     +  7.9354E+02, 8.1211E+02, 8.3073E+02, 8.4943E+02, 8.6819E+02,
     + 8.8702E+02, 9.0594E+02, 9.2492E+02, 9.4399E+02, 9.6313E+02,
     +  9.8236E+02, 1.0017E+03, 1.0211E+03, 1.0406E+03, 1.0602E+03,
     + 1.0798E+03, 1.0996E+03, 1.1195E+03, 1.1394E+03, 1.1595E+03,
     +  1.1797E+03, 1.1999E+03, 1.2203E+03, 1.2408E+03, 1.2614E+03,
     + 1.2821E+03, 1.3029E+03, 1.3238E+03, 1.3448E+03, 1.3659E+03,
     +  1.3871E+03, 1.4085E+03, 1.4299E+03, 1.4515E+03, 1.4732E+03,
     + 1.4949E+03, 1.5168E+03, 1.5388E+03, 1.5610E+03, 1.5832E+03,
     +  1.6055E+03, 1.6280E+03, 1.6506E+03, 1.6733E+03, 1.6961E+03,
     + 1.7190E+03, 1.7420E+03, 1.7652E+03, 1.7885E+03, 1.8118E+03,
     +  1.8353E+03, 1.8589E+03, 1.8827E+03, 1.9065E+03, 1.9305E+03,
     + 1.9546E+03, 1.9787E+03, 2.0030E+03, 2.0274E+03, 2.0520E+03,
     +  2.0766E+03, 2.1014E+03, 2.1262E+03, 2.1512E+03, 2.1763E+03,
     + 2.2016E+03, 2.2269E+03, 2.2524E+03, 2.2779E+03, 2.3036E+03,
     +  2.3294E+03, 2.3553E+03, 2.3813E+03, 2.4075E+03, 2.4337E+03,
     + 2.4601E+03, 2.4866E+03, 2.5132E+03, 2.5399E+03, 2.5667E+03,
     +  2.5936E+03, 2.6207E+03, 2.6479E+03, 2.6752E+03, 2.7026E+03,
     + 2.7301E+03, 2.7577E+03, 2.7854E+03, 2.8133E+03, 2.8413E+03,
     +  2.8694E+03, 2.8976E+03, 2.9259E+03, 2.9543E+03, 2.9829E+03,
     + 3.0115E+03, 3.0403E+03, 3.0692E+03, 3.0982E+03, 3.1273E+03,
     +  3.1566E+03, 3.1859E+03, 3.2154E+03, 3.2449E+03, 3.2746E+03,
     + 3.3044E+03, 3.3343E+03, 3.3644E+03, 3.3945E+03, 3.4248E+03,
     +  3.4552E+03, 3.4857E+03, 3.5163E+03, 3.5470E+03, 3.5778E+03,
     + 3.6088E+03, 3.6399E+03, 3.6710E+03, 3.7024E+03, 3.7338E+03,
     +  3.7653E+03, 3.7969E+03, 3.8287E+03, 3.8606E+03, 3.8925E+03,
     + 3.9246E+03, 3.9569E+03, 3.9892E+03, 4.0216E+03, 4.0542E+03,
     +  4.0869E+03, 4.1197E+03, 4.1526E+03, 4.1856E+03, 4.2188E+03,
     + 4.2521E+03, 4.2854E+03, 4.3189E+03, 4.3525E+03, 4.3862E+03,
     +  4.4201E+03, 4.4540E+03, 4.4881E+03, 4.5223E+03, 4.5566E+03,
     + 4.5910E+03, 4.6255E+03, 4.6602E+03, 4.6949E+03, 4.7298E+03,
     +  4.7648E+03, 4.7999E+03, 4.8351E+03, 4.8705E+03, 4.9060E+03,
     + 4.9415E+03, 4.9772E+03, 5.0130E+03, 5.0490E+03, 5.0850E+03,
     +  5.1212E+03, 5.1575E+03, 5.1939E+03, 5.2304E+03, 5.2670E+03,
     + 5.3038E+03, 5.3406E+03, 5.3776E+03, 5.4147E+03, 5.4519E+03,
     +  5.4892E+03, 5.5267E+03, 5.5642E+03, 5.6019E+03, 5.6397E+03,
     + 5.6777E+03, 5.7157E+03, 5.7538E+03, 5.7921E+03, 5.8305E+03,
     +  5.8690E+03, 5.9076E+03, 5.9463E+03, 5.9852E+03, 6.0242E+03,
     + 6.0633E+03, 6.1025E+03, 6.1418E+03, 6.1812E+03, 6.2208E+03,
     +  6.2605E+03, 6.3003E+03, 6.3402E+03, 6.3803E+03, 6.4204E+03,
     + 6.4607E+03, 6.5010E+03, 6.5415E+03, 6.5822E+03, 6.6229E+03,
     +  6.6638E+03, 6.7048E+03, 6.7459E+03, 6.7871E+03, 6.8284E+03,
     + 6.8699E+03, 6.9114E+03, 6.9531E+03, 6.9949E+03, 7.0369E+03,
     +  7.0789E+03, 7.1211E+03, 7.1633E+03, 7.2058E+03, 7.2483E+03,
     + 7.2909E+03, 7.3337E+03, 7.3765E+03, 7.4196E+03, 7.4627E+03,
     +  7.5059E+03, 7.5493E+03, 7.5927E+03, 7.6363E+03, 7.6800E+03,
     + 7.7239E+03, 7.7678E+03, 7.8119E+03, 7.8561E+03, 7.9004E+03,
     +  7.9448E+03, 7.9893E+03, 8.0340E+03, 8.0788E+03, 8.1237E+03,
     + 8.1687E+03, 8.2138E+03, 8.2591E+03, 8.3045E+03, 8.3500E+03,
     +  8.3957E+03, 8.4413E+03, 8.4872E+03, 8.5332E+03, 8.5793E+03,
     + 8.6255E+03, 8.6719E+03, 8.7183E+03, 8.7649E+03, 8.8116E+03,
     +  8.8584E+03, 8.9054E+03, 8.9524E+03, 8.9996E+03, 9.0469E+03,
     + 9.0943E+03, 9.1418E+03, 9.1895E+03, 9.2373E+03, 9.2852E+03,
     +  9.3332E+03, 9.3813E+03, 9.4296E+03, 9.4780E+03, 9.5265E+03,
     + 9.5751E+03, 9.6238E+03, 9.6726E+03, 9.7216E+03, 9.7708E+03,
     +  9.8200E+03, 9.8693E+03, 9.9188E+03, 9.9683E+03, 1.0018E+04,
     + 1.0068E+04, 1.0118E+04, 1.0168E+04, 1.0218E+04, 1.0268E+04,
     +  1.0319E+04, 1.0369E+04, 1.0420E+04, 1.0471E+04, 1.0522E+04,
     + 1.0573E+04, 1.0624E+04, 1.0675E+04, 1.0727E+04, 1.0778E+04,
     +  1.0830E+04, 1.0881E+04, 1.0933E+04, 1.0985E+04, 1.1038E+04,
     + 1.1090E+04, 1.1142E+04, 1.1195E+04, 1.1247E+04, 1.1300E+04,
     +  1.1353E+04, 1.1406E+04, 1.1459E+04, 1.1512E+04, 1.1565E+04,
     + 1.1619E+04, 1.1672E+04, 1.1726E+04, 1.1780E+04, 1.1834E+04,
     +  1.1888E+04, 1.1942E+04, 1.1996E+04, 1.2051E+04, 1.2105E+04,
     + 1.2160E+04, 1.2215E+04, 1.2270E+04, 1.2325E+04, 1.2380E+04,
     +  1.2435E+04, 1.2491E+04, 1.2546E+04, 1.2602E+04, 1.2657E+04,
     + 1.2713E+04, 1.2769E+04, 1.2825E+04, 1.2882E+04, 1.2938E+04,
     +  1.2994E+04, 1.3051E+04, 1.3108E+04, 1.3165E+04, 1.3222E+04,
     + 1.3279E+04, 1.3336E+04, 1.3393E+04, 1.3451E+04, 1.3508E+04,
     +  1.3566E+04, 1.3624E+04, 1.3682E+04, 1.3740E+04, 1.3798E+04,
     + 1.3856E+04, 1.3915E+04, 1.3973E+04, 1.4032E+04, 1.4091E+04,
     +  1.4150E+04, 1.4209E+04, 1.4268E+04, 1.4327E+04, 1.4387E+04,
     + 1.4446E+04, 1.4506E+04, 1.4566E+04, 1.4626E+04, 1.4686E+04,
     +  1.4746E+04, 1.4806E+04, 1.4866E+04, 1.4927E+04, 1.4988E+04,
     + 1.5048E+04, 1.5109E+04, 1.5170E+04, 1.5231E+04, 1.5293E+04,
     +  1.5354E+04, 1.5415E+04, 1.5477E+04, 1.5539E+04, 1.5601E+04,
     + 1.5663E+04, 1.5725E+04, 1.5787E+04, 1.5849E+04, 1.5912E+04,
     +  1.5974E+04, 1.6037E+04, 1.6100E+04, 1.6163E+04, 1.6226E+04,
     + 1.6289E+04, 1.6353E+04, 1.6416E+04, 1.6480E+04, 1.6543E+04,
     +  1.6607E+04, 1.6671E+04, 1.6735E+04, 1.6799E+04, 1.6864E+04,
     + 1.6928E+04, 1.6993E+04, 1.7057E+04, 1.7122E+04, 1.7187E+04,
     +  1.7252E+04, 1.7317E+04, 1.7383E+04, 1.7448E+04, 1.7514E+04,
     + 1.7579E+04, 1.7645E+04, 1.7711E+04, 1.7777E+04, 1.7843E+04,
     +  1.7909E+04, 1.7976E+04, 1.8042E+04, 1.8109E+04, 1.8176E+04,
     + 1.8243E+04, 1.8310E+04, 1.8377E+04, 1.8444E+04, 1.8511E+04,
     +  1.8579E+04, 1.8647E+04, 1.8714E+04, 1.8782E+04, 1.8850E+04,
     + 1.8918E+04, 1.8987E+04, 1.9055E+04, 1.9124E+04, 1.9192E+04,
     +  1.9261E+04, 1.9330E+04, 1.9399E+04, 1.9468E+04, 1.9537E+04,
     + 1.9606E+04, 1.9676E+04, 1.9746E+04, 1.9815E+04, 1.9885E+04,
     +  1.9955E+04, 2.0025E+04, 2.0095E+04, 2.0166E+04, 2.0236E+04,
     + 2.0307E+04, 2.0378E+04, 2.0448E+04, 2.0519E+04, 2.0591E+04,
     +  2.0662E+04, 2.0733E+04, 2.0805E+04, 2.0876E+04, 2.0948E+04,
     + 2.1020E+04, 2.1092E+04, 2.1164E+04, 2.1236E+04, 2.1308E+04,
     +  2.1381E+04, 2.1453E+04, 2.1526E+04, 2.1599E+04, 2.1672E+04,
     + 2.1745E+04, 2.1818E+04, 2.1891E+04, 2.1965E+04, 2.2038E+04,
     +  2.2112E+04, 2.2186E+04, 2.2260E+04, 2.2334E+04, 2.2408E+04,
     + 2.2482E+04, 2.2557E+04, 2.2631E+04, 2.2706E+04, 2.2781E+04,
     +  2.2856E+04, 2.2931E+04, 2.3006E+04, 2.3081E+04, 2.3157E+04,
     + 2.3232E+04, 2.3308E+04, 2.3384E+04, 2.3460E+04, 2.3536E+04,
     +  2.3612E+04, 2.3688E+04, 2.3765E+04, 2.3841E+04, 2.3918E+04,
     + 2.3995E+04, 2.4072E+04, 2.4149E+04, 2.4226E+04, 2.4303E+04,
     +  2.4380E+04, 2.4458E+04, 2.4536E+04, 2.4614E+04, 2.4691E+04,
     + 2.4769E+04, 2.4848E+04, 2.4926E+04, 2.5004E+04, 2.5083E+04,
     +  2.5161E+04, 2.5240E+04, 2.5319E+04, 2.5398E+04, 2.5477E+04,
     + 2.5557E+04, 2.5636E+04, 2.5715E+04, 2.5795E+04, 2.5875E+04,
     +  2.5955E+04, 2.6035E+04, 2.6115E+04, 2.6195E+04, 2.6276E+04,
     + 2.6356E+04, 2.6437E+04, 2.6518E+04, 2.6599E+04, 2.6680E+04,
     +  2.6761E+04, 2.6842E+04, 2.6923E+04, 2.7005E+04, 2.7087E+04,
     + 2.7168E+04, 2.7250E+04, 2.7332E+04, 2.7414E+04, 2.7497E+04,
     +  2.7579E+04, 2.7662E+04, 2.7744E+04, 2.7827E+04, 2.7910E+04,
     + 2.7993E+04, 2.8076E+04, 2.8159E+04, 2.8243E+04, 2.8326E+04,
     +  2.8410E+04, 2.8494E+04, 2.8578E+04, 2.8662E+04, 2.8746E+04,
     + 2.8830E+04, 2.8914E+04, 2.8999E+04, 2.9084E+04, 2.9168E+04,
     +  2.9253E+04, 2.9338E+04, 2.9423E+04, 2.9509E+04, 2.9594E+04,
     + 2.9680E+04, 2.9765E+04, 2.9851E+04, 2.9937E+04, 3.0023E+04,
     +  3.0109E+04, 3.0195E+04, 3.0282E+04, 3.0368E+04, 3.0455E+04,
     + 3.0542E+04, 3.0629E+04, 3.0716E+04, 3.0803E+04, 3.0890E+04,
     +  3.0978E+04, 3.1065E+04, 3.1153E+04, 3.1241E+04, 3.1328E+04,
     + 3.1416E+04, 3.1505E+04, 3.1593E+04, 3.1681E+04, 3.1770E+04,
     +  3.1858E+04, 3.1947E+04, 3.2036E+04, 3.2125E+04, 3.2214E+04,
     + 3.2304E+04, 3.2393E+04, 3.2482E+04, 3.2572E+04, 3.2662E+04,
     +  3.2752E+04, 3.2842E+04, 3.2932E+04, 3.3022E+04, 3.3113E+04,
     + 3.3203E+04, 3.3294E+04, 3.3385E+04, 3.3476E+04, 3.3567E+04,
     +  3.3658E+04, 3.3749E+04, 3.3841E+04, 3.3932E+04, 3.4024E+04,
     + 3.4116E+04, 3.4208E+04, 3.4300E+04, 3.4392E+04, 3.4484E+04,
     +  3.4577E+04, 3.4669E+04, 3.4762E+04, 3.4855E+04, 3.4947E+04,
     + 3.5040E+04, 3.5134E+04, 3.5227E+04, 3.5320E+04, 3.5414E+04,
     +  3.5508E+04, 3.5601E+04, 3.5695E+04, 3.5789E+04, 3.5884E+04,
     + 3.5978E+04, 3.6072E+04, 3.6167E+04, 3.6262E+04, 3.6356E+04,
     +  3.6451E+04, 3.6546E+04, 3.6642E+04, 3.6737E+04, 3.6832E+04,
     + 3.6928E+04, 3.7024E+04, 3.7120E+04, 3.7216E+04, 3.7312E+04,
     +  3.7408E+04, 3.7504E+04, 3.7600E+04, 3.7697E+04, 3.7794E+04,
     + 3.7891E+04, 3.7988E+04, 3.8085E+04, 3.8182E+04, 3.8279E+04,
     +  3.8377E+04, 3.8474E+04, 3.8572E+04, 3.8670E+04, 3.8768E+04,
     + 3.8866E+04, 3.8964E+04, 3.9063E+04, 3.9161E+04, 3.9260E+04,
     +  3.9358E+04, 3.9457E+04, 3.9556E+04, 3.9655E+04, 3.9755E+04,
     + 3.9854E+04, 3.9953E+04, 4.0053E+04, 4.0153E+04, 4.0253E+04,
     +  4.0352E+04, 4.0453E+04, 4.0553E+04, 4.0653E+04, 4.0754E+04,
     + 4.0854E+04, 4.0955E+04, 4.1056E+04, 4.1157E+04, 4.1258E+04,
     +  4.1359E+04, 4.1461E+04, 4.1562E+04, 4.1664E+04, 4.1766E+04,
     + 4.1868E+04, 4.1970E+04, 4.2072E+04, 4.2174E+04, 4.2276E+04,
     +  4.2379E+04, 4.2482E+04, 4.2584E+04, 4.2687E+04, 4.2790E+04,
     + 4.2894E+04, 4.2997E+04, 4.3100E+04, 4.3204E+04, 4.3307E+04,
     +  4.3411E+04, 4.3515E+04, 4.3619E+04, 4.3723E+04, 4.3828E+04,
     + 4.3932E+04, 4.4037E+04, 4.4141E+04, 4.4246E+04, 4.4351E+04,
     +  4.4456E+04, 4.4561E+04, 4.4667E+04, 4.4772E+04, 4.4878E+04,
     + 4.4983E+04, 4.5089E+04, 4.5195E+04, 4.5301E+04, 4.5407E+04,
     +  4.5514E+04, 4.5620E+04, 4.5727E+04, 4.5833E+04, 4.5940E+04,
     + 4.6047E+04, 4.6154E+04, 4.6261E+04, 4.6369E+04, 4.6476E+04,
     +  4.6584E+04, 4.6692E+04, 4.6800E+04, 4.6908E+04, 4.7016E+04,
     + 4.7124E+04, 4.7232E+04, 4.7341E+04, 4.7450E+04, 4.7558E+04,
     +  4.7667E+04, 4.7776E+04, 4.7885E+04, 4.7994E+04, 4.8104E+04,
     + 4.8213E+04, 4.8323E+04, 4.8433E+04, 4.8543E+04, 4.8653E+04,
     +  4.8763E+04, 4.8873E+04, 4.8983E+04, 4.9094E+04, 4.9205E+04,
     + 4.9315E+04, 4.9426E+04, 4.9537E+04, 4.9649E+04, 4.9760E+04,
     +  4.9871E+04, 4.9983E+04, 5.0094E+04, 5.0206E+04, 5.0318E+04,
     + 5.0430E+04, 5.0543E+04, 5.0655E+04, 5.0767E+04, 5.0880E+04,
     +  5.0992E+04, 5.1105E+04, 5.1218E+04, 5.1331E+04, 5.1444E+04,
     + 5.1558E+04, 5.1672E+04, 5.1785E+04, 5.1899E+04, 5.2013E+04,
     +  5.2127E+04, 5.2241E+04, 5.2355E+04, 5.2469E+04, 5.2584E+04,
     + 5.2698E+04, 5.2813E+04, 5.2928E+04, 5.3043E+04, 5.3158E+04,
     +  5.3273E+04, 5.3389E+04, 5.3504E+04, 5.3620E+04, 5.3736E+04,
     + 5.3852E+04, 5.3967E+04, 5.4084E+04, 5.4200E+04, 5.4316E+04,
     +  5.4433E+04, 5.4549E+04, 5.4666E+04, 5.4783E+04, 5.4900E+04,
     + 5.5017E+04, 5.5135E+04, 5.5252E+04, 5.5370E+04, 5.5488E+04,
     +  5.5605E+04, 5.5723E+04, 5.5841E+04, 5.5959E+04, 5.6078E+04,
     + 5.6196E+04, 5.6315E+04, 5.6434E+04, 5.6552E+04, 5.6671E+04,
     +  5.6790E+04, 5.6910E+04, 5.7029E+04, 5.7148E+04, 5.7268E+04,
     + 5.7388E+04, 5.7508E+04, 5.7628E+04, 5.7748E+04, 5.7868E+04,
     +  5.7988E+04, 5.8109E+04, 5.8230E+04, 5.8350E+04, 5.8471E+04,
     + 5.8592E+04, 5.8713E+04, 5.8835E+04, 5.8956E+04, 5.9078E+04,
     +  5.9199E+04, 5.9321E+04, 5.9443E+04, 5.9565E+04, 5.9687E+04,
     + 5.9809E+04, 5.9932E+04, 6.0054E+04, 6.0177E+04, 6.0300E+04,
     +  6.0423E+04, 6.0546E+04, 6.0669E+04, 6.0792E+04, 6.0916E+04,
     + 6.1040E+04, 6.1163E+04, 6.1287E+04, 6.1411E+04, 6.1535E+04,
     +  6.1659E+04, 6.1784E+04, 6.1908E+04, 6.2033E+04, 6.2158E+04,
     + 6.2283E+04, 6.2408E+04, 6.2533E+04, 6.2658E+04, 6.2783E+04,
     +  6.2909E+04, 6.3035E+04, 6.3160E+04, 6.3286E+04, 6.3412E+04,
     + 6.3538E+04, 6.3665E+04, 6.3791E+04, 6.3918E+04, 6.4044E+04,
     +  6.4171E+04, 6.4298E+04, 6.4425E+04, 6.4552E+04, 6.4680E+04,
     + 6.4807E+04, 6.4935E+04, 6.5062E+04, 6.5190E+04, 6.5318E+04,
     +  6.5446E+04, 6.5575E+04, 6.5703E+04, 6.5831E+04, 6.5960E+04,
     + 6.6089E+04, 6.6217E+04, 6.6347E+04, 6.6476E+04, 6.6605E+04,
     +  6.6734E+04, 6.6864E+04, 6.6994E+04, 6.7123E+04, 6.7253E+04,
     + 6.7383E+04, 6.7513E+04, 6.7644E+04, 6.7774E+04, 6.7905E+04,
     +  6.8035E+04, 6.8166E+04, 6.8297E+04, 6.8428E+04, 6.8559E+04,
     + 6.8691E+04, 6.8822E+04, 6.8954E+04, 6.9085E+04, 6.9217E+04,
     +  6.9349E+04, 6.9481E+04, 6.9614E+04, 6.9746E+04, 6.9878E+04,
     + 7.0011E+04, 7.0144E+04, 7.0277E+04, 7.0410E+04, 7.0543E+04,
     +  7.0676E+04, 7.0809E+04, 7.0943E+04, 7.1077E+04, 7.1211E+04,
     + 7.1344E+04, 7.1478E+04, 7.1612E+04, 7.1747E+04, 7.1881E+04,
     +  7.2016E+04, 7.2151E+04, 7.2285E+04, 7.2420E+04, 7.2556E+04,
     + 7.2691E+04, 7.2826E+04, 7.2961E+04, 7.3097E+04, 7.3233E+04,
     +  7.3369E+04, 7.3504E+04, 7.3641E+04, 7.3777E+04, 7.3913E+04,
     + 7.4050E+04, 7.4186E+04, 7.4323E+04, 7.4460E+04, 7.4597E+04,
     +  7.4734E+04, 7.4872E+04, 7.5009E+04, 7.5147E+04, 7.5284E+04,
     + 7.5422E+04, 7.5560E+04, 7.5698E+04, 7.5836E+04, 7.5974E+04,
     +  7.6113E+04, 7.6251E+04, 7.6390E+04, 7.6529E+04, 7.6668E+04,
     + 7.6807E+04, 7.6946E+04, 7.7086E+04, 7.7225E+04, 7.7365E+04,
     +  7.7505E+04, 7.7644E+04, 7.7784E+04, 7.7925E+04, 7.8065E+04,
     + 7.8205E+04, 7.8346E+04, 7.8486E+04, 7.8627E+04, 7.8768E+04,
     +  7.8909E+04, 7.9051E+04, 7.9192E+04, 7.9333E+04, 7.9475E+04,
     + 7.9616E+04, 7.9758E+04, 7.9900E+04, 8.0042E+04, 8.0184E+04,
     +  8.0327E+04, 8.0470E+04, 8.0612E+04, 8.0755E+04, 8.0898E+04,
     + 8.1041E+04, 8.1184E+04, 8.1327E+04, 8.1470E+04, 8.1614E+04,
     +  8.1758E+04, 8.1901E+04, 8.2045E+04, 8.2189E+04, 8.2334E+04,
     + 8.2478E+04, 8.2622E+04, 8.2767E+04, 8.2911E+04, 8.3056E+04,
     +  8.3201E+04, 8.3346E+04, 8.3492E+04, 8.3637E+04, 8.3782E+04,
     + 8.3928E+04, 8.4074E+04, 8.4220E+04, 8.4366E+04, 8.4512E+04,
     +  8.4658E+04, 8.4804E+04, 8.4951E+04, 8.5097E+04, 8.5244E+04,
     + 8.5391E+04, 8.5538E+04, 8.5686E+04, 8.5833E+04, 8.5980E+04,
     +  8.6128E+04, 8.6276E+04, 8.6423E+04, 8.6571E+04, 8.6719E+04,
     + 8.6868E+04, 8.7016E+04, 8.7164E+04, 8.7313E+04, 8.7462E+04,
     +  8.7611E+04, 8.7760E+04, 8.7909E+04, 8.8058E+04, 8.8208E+04,
     + 8.8357E+04, 8.8507E+04, 8.8657E+04, 8.8807E+04, 8.8957E+04,
     +  8.9107E+04, 8.9257E+04, 8.9407E+04, 8.9558E+04, 8.9709E+04,
     + 8.9860E+04, 9.0010E+04, 9.0162E+04, 9.0313E+04, 9.0464E+04,
     +  9.0616E+04, 9.0767E+04, 9.0919E+04, 9.1071E+04, 9.1223E+04,
     + 9.1375E+04, 9.1527E+04, 9.1680E+04, 9.1832E+04, 9.1985E+04,
     +  9.2138E+04, 9.2291E+04, 9.2444E+04, 9.2597E+04, 9.2750E+04,
     + 9.2904E+04, 9.3057E+04, 9.3211E+04, 9.3365E+04, 9.3519E+04,
     +  9.3673E+04, 9.3827E+04, 9.3982E+04, 9.4136E+04, 9.4291E+04,
     + 9.4446E+04, 9.4601E+04, 9.4756E+04, 9.4911E+04, 9.5066E+04,
     +  9.5221E+04, 9.5377E+04, 9.5533E+04, 9.5689E+04, 9.5844E+04,
     + 9.6001E+04, 9.6157E+04, 9.6313E+04, 9.6470E+04, 9.6626E+04,
     +  9.6783E+04, 9.6940E+04, 9.7097E+04, 9.7254E+04, 9.7411E+04,
     + 9.7569E+04, 9.7726E+04, 9.7884E+04, 9.8042E+04, 9.8199E+04,
     +  9.8358E+04, 9.8516E+04, 9.8674E+04, 9.8833E+04, 9.8991E+04,
     + 9.9150E+04, 9.9309E+04, 9.9468E+04, 9.9627E+04, 9.9786E+04,
     +  9.9945E+04, 1.0011E+05, 1.0026E+05, 1.0042E+05, 1.0058E+05,
     + 1.0074E+05, 1.0090E+05, 1.0106E+05, 1.0122E+05, 1.0139E+05,
     +  1.0155E+05, 1.0171E+05, 1.0187E+05, 1.0203E+05, 1.0219E+05,
     + 1.0235E+05, 1.0251E+05, 1.0267E+05, 1.0284E+05, 1.0300E+05,
     +  1.0316E+05, 1.0332E+05, 1.0349E+05, 1.0365E+05, 1.0381E+05,
     + 1.0397E+05, 1.0414E+05, 1.0430E+05, 1.0446E+05, 1.0462E+05,
     +  1.0479E+05, 1.0495E+05, 1.0512E+05, 1.0528E+05, 1.0544E+05,
     + 1.0561E+05, 1.0577E+05, 1.0593E+05, 1.0610E+05, 1.0626E+05,
     +  1.0643E+05, 1.0659E+05, 1.0676E+05, 1.0692E+05, 1.0709E+05,
     + 1.0725E+05, 1.0742E+05, 1.0758E+05, 1.0775E+05, 1.0792E+05,
     +  1.0808E+05, 1.0825E+05, 1.0841E+05, 1.0858E+05, 1.0875E+05,
     + 1.0891E+05, 1.0908E+05, 1.0925E+05, 1.0941E+05, 1.0958E+05,
     +  1.0975E+05, 1.0992E+05, 1.1008E+05, 1.1025E+05, 1.1042E+05,
     + 1.1059E+05, 1.1076E+05, 1.1092E+05, 1.1109E+05, 1.1126E+05,
     +  1.1143E+05, 1.1160E+05, 1.1177E+05, 1.1194E+05, 1.1210E+05,
     + 1.1227E+05, 1.1244E+05, 1.1261E+05, 1.1278E+05, 1.1295E+05,
     +  1.1312E+05, 1.1329E+05, 1.1346E+05, 1.1363E+05, 1.1380E+05,
     + 1.1397E+05, 1.1414E+05, 1.1432E+05, 1.1449E+05, 1.1466E+05,
     +  1.1483E+05, 1.1500E+05, 1.1517E+05, 1.1534E+05, 1.1552E+05,
     + 1.1569E+05, 1.1586E+05, 1.1603E+05, 1.1620E+05, 1.1638E+05,
     +  1.1655E+05, 1.1672E+05, 1.1690E+05, 1.1707E+05, 1.1724E+05,
     + 1.1741E+05, 1.1759E+05, 1.1776E+05, 1.1793E+05, 1.1811E+05,
     +  1.1828E+05, 1.1846E+05, 1.1863E+05, 1.1881E+05, 1.1898E+05,
     + 1.1915E+05, 1.1933E+05, 1.1950E+05, 1.1968E+05, 1.1985E+05,
     +  1.2003E+05, 1.2020E+05, 1.2038E+05, 1.2056E+05, 1.2073E+05,
     + 1.2091E+05, 1.2108E+05, 1.2126E+05, 1.2144E+05, 1.2161E+05,
     +  1.2179E+05, 1.2197E+05, 1.2214E+05, 1.2232E+05, 1.2250E+05,
     + 1.2267E+05, 1.2285E+05, 1.2303E+05, 1.2321E+05, 1.2338E+05,
     +  1.2356E+05, 1.2374E+05, 1.2392E+05, 1.2410E+05, 1.2427E+05,
     + 1.2445E+05, 1.2463E+05, 1.2481E+05, 1.2499E+05, 1.2517E+05,
     +  1.2535E+05, 1.2553E+05, 1.2571E+05, 1.2589E+05, 1.2607E+05,
     + 1.2625E+05, 1.2643E+05, 1.2661E+05, 1.2679E+05, 1.2697E+05,
     +  1.2715E+05, 1.2733E+05, 1.2751E+05, 1.2769E+05, 1.2787E+05,
     + 1.2805E+05, 1.2823E+05, 1.2842E+05, 1.2860E+05, 1.2878E+05,
     +  1.2896E+05, 1.2914E+05, 1.2933E+05, 1.2951E+05, 1.2969E+05,
     + 1.2987E+05, 1.3006E+05, 1.3024E+05, 1.3042E+05, 1.3060E+05,
     +  1.3079E+05, 1.3097E+05, 1.3115E+05, 1.3134E+05, 1.3152E+05,
     + 1.3171E+05, 1.3189E+05, 1.3207E+05, 1.3226E+05, 1.3244E+05,
     +  1.3263E+05, 1.3281E+05, 1.3300E+05, 1.3318E+05, 1.3337E+05,
     + 1.3355E+05, 1.3374E+05, 1.3392E+05, 1.3411E+05, 1.3430E+05,
     +  1.3448E+05, 1.3467E+05, 1.3485E+05, 1.3504E+05, 1.3523E+05,
     + 1.3541E+05, 1.3560E+05, 1.3579E+05, 1.3597E+05, 1.3616E+05,
     +  1.3635E+05, 1.3653E+05, 1.3672E+05, 1.3691E+05, 1.3710E+05,
     + 1.3729E+05, 1.3747E+05, 1.3766E+05, 1.3785E+05, 1.3804E+05,
     +  1.3823E+05, 1.3842E+05, 1.3860E+05, 1.3879E+05, 1.3898E+05,
     + 1.3917E+05, 1.3936E+05, 1.3955E+05, 1.3974E+05, 1.3993E+05,
     +  1.4012E+05, 1.4031E+05, 1.4050E+05, 1.4069E+05, 1.4088E+05,
     + 1.4107E+05, 1.4126E+05, 1.4145E+05, 1.4165E+05, 1.4184E+05,
     +  1.4203E+05, 1.4222E+05, 1.4241E+05, 1.4260E+05, 1.4279E+05,
     + 1.4299E+05, 1.4318E+05, 1.4337E+05, 1.4356E+05, 1.4376E+05,
     +  1.4395E+05, 1.4414E+05, 1.4433E+05, 1.4453E+05, 1.4472E+05,
     + 1.4491E+05, 1.4511E+05, 1.4530E+05, 1.4549E+05, 1.4569E+05,
     +  1.4588E+05, 1.4608E+05, 1.4627E+05, 1.4646E+05, 1.4666E+05,
     + 1.4685E+05, 1.4705E+05, 1.4724E+05, 1.4744E+05, 1.4763E+05,
     +  1.4783E+05, 1.4802E+05, 1.4822E+05, 1.4842E+05, 1.4861E+05,
     + 1.4881E+05, 1.4900E+05, 1.4920E+05, 1.4940E+05, 1.4959E+05,
     +  1.4979E+05, 1.4999E+05, 1.5018E+05, 1.5038E+05, 1.5058E+05,
     + 1.5077E+05, 1.5097E+05, 1.5117E+05, 1.5137E+05, 1.5157E+05,
     +  1.5176E+05, 1.5196E+05, 1.5216E+05, 1.5236E+05, 1.5256E+05,
     + 1.5276E+05, 1.5295E+05, 1.5315E+05, 1.5335E+05, 1.5355E+05,
     +  1.5375E+05, 1.5395E+05, 1.5415E+05, 1.5435E+05, 1.5455E+05,
     + 1.5475E+05, 1.5495E+05, 1.5515E+05, 1.5535E+05, 1.5555E+05,
     +  1.5575E+05, 1.5595E+05, 1.5615E+05, 1.5636E+05, 1.5656E+05,
     + 1.5676E+05, 1.5696E+05, 1.5716E+05, 1.5736E+05, 1.5756E+05,
     +  1.5777E+05, 1.5797E+05, 1.5817E+05, 1.5837E+05, 1.5858E+05,
     + 1.5878E+05, 1.5898E+05, 1.5919E+05, 1.5939E+05, 1.5959E+05,
     +  1.5979E+05, 1.6000E+05, 1.6020E+05, 1.6041E+05, 1.6061E+05,
     + 1.6081E+05, 1.6102E+05, 1.6122E+05, 1.6143E+05, 1.6163E+05,
     +  1.6184E+05, 1.6204E+05, 1.6225E+05, 1.6245E+05, 1.6266E+05,
     + 1.6286E+05, 1.6307E+05, 1.6327E+05, 1.6348E+05, 1.6369E+05,
     +  1.6389E+05, 1.6410E+05, 1.6430E+05, 1.6451E+05, 1.6472E+05,
     + 1.6492E+05, 1.6513E+05, 1.6534E+05, 1.6555E+05, 1.6575E+05,
     +  1.6596E+05, 1.6617E+05, 1.6638E+05, 1.6658E+05, 1.6679E+05,
     + 1.6700E+05, 1.6721E+05, 1.6742E+05, 1.6763E+05, 1.6783E+05,
     +  1.6804E+05, 1.6825E+05, 1.6846E+05, 1.6867E+05, 1.6888E+05,
     + 1.6909E+05, 1.6930E+05, 1.6951E+05, 1.6972E+05, 1.6993E+05,
     +  1.7014E+05, 1.7035E+05, 1.7056E+05, 1.7077E+05, 1.7098E+05,
     + 1.7119E+05, 1.7140E+05, 1.7161E+05, 1.7183E+05, 1.7204E+05,
     +  1.7225E+05, 1.7246E+05, 1.7267E+05, 1.7288E+05, 1.7310E+05,
     + 1.7331E+05, 1.7352E+05, 1.7373E+05, 1.7395E+05, 1.7416E+05,
     +  1.7437E+05, 1.7458E+05, 1.7480E+05, 1.7501E+05, 1.7522E+05,
     + 1.7544E+05, 1.7565E+05, 1.7587E+05, 1.7608E+05, 1.7629E+05,
     +  1.7651E+05, 1.7672E+05, 1.7694E+05, 1.7715E+05, 1.7737E+05,
     + 1.7758E+05, 1.7780E+05, 1.7801E+05, 1.7823E+05, 1.7844E+05,
     +  1.7866E+05, 1.7887E+05, 1.7909E+05, 1.7931E+05, 1.7952E+05,
     + 1.7974E+05, 1.7996E+05, 1.8017E+05, 1.8039E+05, 1.8061E+05,
     +  1.8082E+05, 1.8104E+05, 1.8126E+05, 1.8147E+05, 1.8169E+05,
     + 1.8191E+05, 1.8213E+05, 1.8235E+05, 1.8256E+05, 1.8278E+05,
     +  1.8300E+05, 1.8322E+05, 1.8344E+05, 1.8366E+05, 1.8387E+05,
     + 1.8409E+05, 1.8431E+05, 1.8453E+05, 1.8475E+05, 1.8497E+05,
     +  1.8519E+05, 1.8541E+05, 1.8563E+05, 1.8585E+05, 1.8607E+05,
     + 1.8629E+05, 1.8651E+05, 1.8673E+05, 1.8695E+05, 1.8717E+05,
     +  1.8740E+05, 1.8762E+05, 1.8784E+05, 1.8806E+05, 1.8828E+05,
     + 1.8850E+05, 1.8873E+05, 1.8895E+05, 1.8917E+05, 1.8939E+05,
     +  1.8961E+05, 1.8984E+05, 1.9006E+05, 1.9028E+05, 1.9051E+05,
     + 1.9073E+05, 1.9095E+05, 1.9118E+05, 1.9140E+05, 1.9162E+05,
     +  1.9185E+05, 1.9207E+05, 1.9230E+05, 1.9252E+05, 1.9274E+05,
     + 1.9297E+05, 1.9319E+05, 1.9342E+05, 1.9364E+05, 1.9387E+05,
     +  1.9409E+05, 1.9432E+05, 1.9454E+05, 1.9477E+05, 1.9500E+05,
     + 1.9522E+05, 1.9545E+05, 1.9567E+05, 1.9590E+05, 1.9613E+05,
     +  1.9635E+05, 1.9658E+05, 1.9681E+05, 1.9703E+05, 1.9726E+05,
     + 1.9749E+05, 1.9772E+05, 1.9794E+05, 1.9817E+05, 1.9840E+05,
     +  1.9863E+05, 1.9885E+05, 1.9908E+05, 1.9931E+05, 1.9954E+05,
     + 1.9977E+05, 2.0000E+05, 2.0023E+05, 2.0045E+05, 2.0068E+05,
     +  2.0091E+05, 2.0114E+05, 2.0137E+05, 2.0160E+05, 2.0183E+05,
     + 2.0206E+05, 2.0229E+05, 2.0252E+05, 2.0275E+05, 2.0298E+05,
     +  2.0321E+05, 2.0345E+05, 2.0368E+05, 2.0391E+05, 2.0414E+05,
     + 2.0437E+05, 2.0460E+05, 2.0483E+05, 2.0506E+05, 2.0530E+05,
     +  2.0553E+05, 2.0576E+05, 2.0599E+05, 2.0623E+05, 2.0646E+05,
     + 2.0669E+05, 2.0692E+05, 2.0716E+05, 2.0739E+05, 2.0762E+05,
     +  2.0786E+05, 2.0809E+05, 2.0832E+05, 2.0856E+05, 2.0879E+05,
     + 2.0903E+05, 2.0926E+05, 2.0950E+05, 2.0973E+05, 2.0996E+05,
     +  2.1020E+05, 2.1043E+05, 2.1067E+05, 2.1090E+05, 2.1114E+05,
     + 2.1138E+05, 2.1161E+05, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        76
      data (QofT( 2,J),J=1,2408)/ 9.9124E+00, 1.8937E+01, 2.8022E+01,
     +  3.7119E+01, 4.6220E+01, 5.5324E+01, 6.4429E+01, 7.3535E+01,
     + 8.2641E+01, 9.1748E+01, 1.0086E+02, 1.0996E+02, 1.1907E+02,
     +  1.2818E+02, 1.3729E+02, 1.4640E+02, 1.5551E+02, 1.6462E+02,
     + 1.7373E+02, 1.8284E+02, 2.0106E+02, 2.1929E+02, 2.3751E+02,
     +  2.5574E+02, 2.7396E+02, 2.9219E+02, 3.1042E+02, 3.2865E+02,
     + 3.4688E+02, 3.6512E+02, 3.8336E+02, 4.0159E+02, 4.1984E+02,
     +  4.3808E+02, 4.5634E+02, 4.7460E+02, 4.9287E+02, 5.1115E+02,
     + 5.2944E+02, 5.4775E+02, 5.6607E+02, 5.8441E+02, 6.0278E+02,
     +  6.2117E+02, 6.3959E+02, 6.5805E+02, 6.7654E+02, 6.9506E+02,
     + 7.1363E+02, 7.3224E+02, 7.5090E+02, 7.6962E+02, 7.8838E+02,
     +  8.0721E+02, 8.2609E+02, 8.4504E+02, 8.6406E+02, 8.8314E+02,
     + 9.0230E+02, 9.2154E+02, 9.4086E+02, 9.6025E+02, 9.7972E+02,
     +  9.9928E+02, 1.0189E+03, 1.0387E+03, 1.0585E+03, 1.0784E+03,
     + 1.0984E+03, 1.1185E+03, 1.1388E+03, 1.1591E+03, 1.1795E+03,
     +  1.2000E+03, 1.2206E+03, 1.2414E+03, 1.2622E+03, 1.2831E+03,
     + 1.3042E+03, 1.3253E+03, 1.3466E+03, 1.3680E+03, 1.3894E+03,
     +  1.4110E+03, 1.4328E+03, 1.4546E+03, 1.4765E+03, 1.4985E+03,
     + 1.5207E+03, 1.5430E+03, 1.5654E+03, 1.5879E+03, 1.6105E+03,
     +  1.6333E+03, 1.6561E+03, 1.6791E+03, 1.7022E+03, 1.7254E+03,
     + 1.7487E+03, 1.7721E+03, 1.7957E+03, 1.8194E+03, 1.8432E+03,
     +  1.8671E+03, 1.8911E+03, 1.9152E+03, 1.9395E+03, 1.9639E+03,
     + 1.9884E+03, 2.0130E+03, 2.0377E+03, 2.0626E+03, 2.0875E+03,
     +  2.1126E+03, 2.1378E+03, 2.1631E+03, 2.1885E+03, 2.2141E+03,
     + 2.2398E+03, 2.2655E+03, 2.2915E+03, 2.3175E+03, 2.3436E+03,
     +  2.3699E+03, 2.3962E+03, 2.4227E+03, 2.4493E+03, 2.4761E+03,
     + 2.5029E+03, 2.5299E+03, 2.5569E+03, 2.5841E+03, 2.6115E+03,
     +  2.6389E+03, 2.6664E+03, 2.6941E+03, 2.7218E+03, 2.7498E+03,
     + 2.7778E+03, 2.8059E+03, 2.8341E+03, 2.8625E+03, 2.8910E+03,
     +  2.9196E+03, 2.9483E+03, 2.9771E+03, 3.0061E+03, 3.0352E+03,
     + 3.0643E+03, 3.0936E+03, 3.1231E+03, 3.1526E+03, 3.1823E+03,
     +  3.2120E+03, 3.2419E+03, 3.2719E+03, 3.3020E+03, 3.3323E+03,
     + 3.3626E+03, 3.3931E+03, 3.4237E+03, 3.4544E+03, 3.4852E+03,
     +  3.5162E+03, 3.5472E+03, 3.5784E+03, 3.6097E+03, 3.6411E+03,
     + 3.6726E+03, 3.7043E+03, 3.7361E+03, 3.7680E+03, 3.8000E+03,
     +  3.8321E+03, 3.8643E+03, 3.8967E+03, 3.9292E+03, 3.9617E+03,
     + 3.9945E+03, 4.0273E+03, 4.0602E+03, 4.0933E+03, 4.1265E+03,
     +  4.1598E+03, 4.1932E+03, 4.2267E+03, 4.2604E+03, 4.2941E+03,
     + 4.3281E+03, 4.3621E+03, 4.3962E+03, 4.4304E+03, 4.4648E+03,
     +  4.4993E+03, 4.5339E+03, 4.5686E+03, 4.6034E+03, 4.6384E+03,
     + 4.6735E+03, 4.7087E+03, 4.7440E+03, 4.7794E+03, 4.8150E+03,
     +  4.8506E+03, 4.8864E+03, 4.9223E+03, 4.9584E+03, 4.9945E+03,
     + 5.0308E+03, 5.0671E+03, 5.1037E+03, 5.1402E+03, 5.1770E+03,
     +  5.2139E+03, 5.2508E+03, 5.2879E+03, 5.3252E+03, 5.3625E+03,
     + 5.4000E+03, 5.4375E+03, 5.4753E+03, 5.5131E+03, 5.5510E+03,
     +  5.5891E+03, 5.6273E+03, 5.6656E+03, 5.7040E+03, 5.7425E+03,
     + 5.7811E+03, 5.8199E+03, 5.8589E+03, 5.8978E+03, 5.9370E+03,
     +  5.9762E+03, 6.0156E+03, 6.0551E+03, 6.0947E+03, 6.1345E+03,
     + 6.1744E+03, 6.2143E+03, 6.2544E+03, 6.2946E+03, 6.3350E+03,
     +  6.3755E+03, 6.4160E+03, 6.4567E+03, 6.4976E+03, 6.5385E+03,
     + 6.5796E+03, 6.6207E+03, 6.6621E+03, 6.7035E+03, 6.7450E+03,
     +  6.7868E+03, 6.8285E+03, 6.8704E+03, 6.9124E+03, 6.9546E+03,
     + 6.9969E+03, 7.0393E+03, 7.0818E+03, 7.1244E+03, 7.1672E+03,
     +  7.2101E+03, 7.2531E+03, 7.2963E+03, 7.3395E+03, 7.3829E+03,
     + 7.4264E+03, 7.4700E+03, 7.5137E+03, 7.5576E+03, 7.6015E+03,
     +  7.6456E+03, 7.6899E+03, 7.7342E+03, 7.7786E+03, 7.8232E+03,
     + 7.8680E+03, 7.9128E+03, 7.9577E+03, 8.0028E+03, 8.0480E+03,
     +  8.0933E+03, 8.1388E+03, 8.1843E+03, 8.2300E+03, 8.2758E+03,
     + 8.3217E+03, 8.3678E+03, 8.4140E+03, 8.4603E+03, 8.5067E+03,
     +  8.5532E+03, 8.5999E+03, 8.6467E+03, 8.6936E+03, 8.7406E+03,
     + 8.7878E+03, 8.8350E+03, 8.8825E+03, 8.9299E+03, 8.9776E+03,
     +  9.0254E+03, 9.0732E+03, 9.1213E+03, 9.1694E+03, 9.2177E+03,
     + 9.2660E+03, 9.3146E+03, 9.3632E+03, 9.4119E+03, 9.4608E+03,
     +  9.5098E+03, 9.5589E+03, 9.6081E+03, 9.6575E+03, 9.7070E+03,
     + 9.7566E+03, 9.8063E+03, 9.8561E+03, 9.9061E+03, 9.9562E+03,
     +  1.0006E+04, 1.0057E+04, 1.0107E+04, 1.0158E+04, 1.0209E+04,
     + 1.0259E+04, 1.0310E+04, 1.0361E+04, 1.0413E+04, 1.0464E+04,
     +  1.0516E+04, 1.0567E+04, 1.0619E+04, 1.0671E+04, 1.0723E+04,
     + 1.0775E+04, 1.0827E+04, 1.0879E+04, 1.0932E+04, 1.0984E+04,
     +  1.1037E+04, 1.1090E+04, 1.1143E+04, 1.1196E+04, 1.1249E+04,
     + 1.1302E+04, 1.1356E+04, 1.1409E+04, 1.1463E+04, 1.1517E+04,
     +  1.1571E+04, 1.1625E+04, 1.1679E+04, 1.1733E+04, 1.1788E+04,
     + 1.1842E+04, 1.1897E+04, 1.1952E+04, 1.2007E+04, 1.2062E+04,
     +  1.2117E+04, 1.2172E+04, 1.2227E+04, 1.2283E+04, 1.2339E+04,
     + 1.2394E+04, 1.2450E+04, 1.2506E+04, 1.2563E+04, 1.2619E+04,
     +  1.2675E+04, 1.2732E+04, 1.2788E+04, 1.2845E+04, 1.2902E+04,
     + 1.2959E+04, 1.3016E+04, 1.3074E+04, 1.3131E+04, 1.3189E+04,
     +  1.3246E+04, 1.3304E+04, 1.3362E+04, 1.3420E+04, 1.3478E+04,
     + 1.3536E+04, 1.3595E+04, 1.3653E+04, 1.3712E+04, 1.3771E+04,
     +  1.3830E+04, 1.3889E+04, 1.3948E+04, 1.4007E+04, 1.4066E+04,
     + 1.4126E+04, 1.4186E+04, 1.4245E+04, 1.4305E+04, 1.4365E+04,
     +  1.4426E+04, 1.4486E+04, 1.4546E+04, 1.4607E+04, 1.4667E+04,
     + 1.4728E+04, 1.4789E+04, 1.4850E+04, 1.4911E+04, 1.4973E+04,
     +  1.5034E+04, 1.5095E+04, 1.5157E+04, 1.5219E+04, 1.5281E+04,
     + 1.5343E+04, 1.5405E+04, 1.5467E+04, 1.5530E+04, 1.5592E+04,
     +  1.5655E+04, 1.5718E+04, 1.5781E+04, 1.5844E+04, 1.5907E+04,
     + 1.5970E+04, 1.6033E+04, 1.6097E+04, 1.6161E+04, 1.6224E+04,
     +  1.6288E+04, 1.6352E+04, 1.6416E+04, 1.6481E+04, 1.6545E+04,
     + 1.6610E+04, 1.6674E+04, 1.6739E+04, 1.6804E+04, 1.6869E+04,
     +  1.6934E+04, 1.6999E+04, 1.7065E+04, 1.7130E+04, 1.7196E+04,
     + 1.7262E+04, 1.7328E+04, 1.7394E+04, 1.7460E+04, 1.7526E+04,
     +  1.7593E+04, 1.7659E+04, 1.7726E+04, 1.7793E+04, 1.7860E+04,
     + 1.7927E+04, 1.7994E+04, 1.8061E+04, 1.8129E+04, 1.8196E+04,
     +  1.8264E+04, 1.8332E+04, 1.8400E+04, 1.8468E+04, 1.8536E+04,
     + 1.8604E+04, 1.8673E+04, 1.8741E+04, 1.8810E+04, 1.8879E+04,
     +  1.8948E+04, 1.9016E+04, 1.9086E+04, 1.9155E+04, 1.9225E+04,
     + 1.9294E+04, 1.9364E+04, 1.9434E+04, 1.9503E+04, 1.9573E+04,
     +  1.9644E+04, 1.9714E+04, 1.9785E+04, 1.9855E+04, 1.9926E+04,
     + 1.9997E+04, 2.0068E+04, 2.0139E+04, 2.0210E+04, 2.0281E+04,
     +  2.0353E+04, 2.0424E+04, 2.0496E+04, 2.0568E+04, 2.0640E+04,
     + 2.0712E+04, 2.0784E+04, 2.0856E+04, 2.0929E+04, 2.1001E+04,
     +  2.1074E+04, 2.1147E+04, 2.1220E+04, 2.1293E+04, 2.1366E+04,
     + 2.1440E+04, 2.1513E+04, 2.1587E+04, 2.1661E+04, 2.1734E+04,
     +  2.1808E+04, 2.1882E+04, 2.1957E+04, 2.2031E+04, 2.2106E+04,
     + 2.2180E+04, 2.2255E+04, 2.2330E+04, 2.2405E+04, 2.2480E+04,
     +  2.2555E+04, 2.2631E+04, 2.2706E+04, 2.2782E+04, 2.2857E+04,
     + 2.2933E+04, 2.3009E+04, 2.3085E+04, 2.3162E+04, 2.3238E+04,
     +  2.3315E+04, 2.3391E+04, 2.3468E+04, 2.3545E+04, 2.3622E+04,
     + 2.3699E+04, 2.3776E+04, 2.3854E+04, 2.3931E+04, 2.4009E+04,
     +  2.4087E+04, 2.4165E+04, 2.4243E+04, 2.4321E+04, 2.4399E+04,
     + 2.4478E+04, 2.4556E+04, 2.4635E+04, 2.4714E+04, 2.4793E+04,
     +  2.4872E+04, 2.4951E+04, 2.5030E+04, 2.5110E+04, 2.5189E+04,
     + 2.5269E+04, 2.5349E+04, 2.5429E+04, 2.5509E+04, 2.5589E+04,
     +  2.5669E+04, 2.5750E+04, 2.5830E+04, 2.5911E+04, 2.5992E+04,
     + 2.6073E+04, 2.6154E+04, 2.6235E+04, 2.6317E+04, 2.6398E+04,
     +  2.6480E+04, 2.6561E+04, 2.6643E+04, 2.6725E+04, 2.6807E+04,
     + 2.6890E+04, 2.6972E+04, 2.7054E+04, 2.7137E+04, 2.7220E+04,
     +  2.7303E+04, 2.7386E+04, 2.7469E+04, 2.7552E+04, 2.7635E+04,
     + 2.7719E+04, 2.7803E+04, 2.7886E+04, 2.7970E+04, 2.8054E+04,
     +  2.8139E+04, 2.8223E+04, 2.8307E+04, 2.8392E+04, 2.8476E+04,
     + 2.8561E+04, 2.8646E+04, 2.8731E+04, 2.8816E+04, 2.8902E+04,
     +  2.8987E+04, 2.9073E+04, 2.9158E+04, 2.9244E+04, 2.9330E+04,
     + 2.9416E+04, 2.9502E+04, 2.9589E+04, 2.9675E+04, 2.9762E+04,
     +  2.9849E+04, 2.9935E+04, 3.0022E+04, 3.0109E+04, 3.0197E+04,
     + 3.0284E+04, 3.0371E+04, 3.0459E+04, 3.0547E+04, 3.0635E+04,
     +  3.0723E+04, 3.0811E+04, 3.0899E+04, 3.0987E+04, 3.1076E+04,
     + 3.1165E+04, 3.1253E+04, 3.1342E+04, 3.1431E+04, 3.1520E+04,
     +  3.1610E+04, 3.1699E+04, 3.1789E+04, 3.1878E+04, 3.1968E+04,
     + 3.2058E+04, 3.2148E+04, 3.2238E+04, 3.2328E+04, 3.2419E+04,
     +  3.2509E+04, 3.2600E+04, 3.2691E+04, 3.2782E+04, 3.2873E+04,
     + 3.2964E+04, 3.3055E+04, 3.3147E+04, 3.3238E+04, 3.3330E+04,
     +  3.3422E+04, 3.3514E+04, 3.3606E+04, 3.3698E+04, 3.3791E+04,
     + 3.3883E+04, 3.3976E+04, 3.4068E+04, 3.4161E+04, 3.4254E+04,
     +  3.4347E+04, 3.4441E+04, 3.4534E+04, 3.4628E+04, 3.4721E+04,
     + 3.4815E+04, 3.4909E+04, 3.5003E+04, 3.5097E+04, 3.5191E+04,
     +  3.5286E+04, 3.5380E+04, 3.5475E+04, 3.5570E+04, 3.5665E+04,
     + 3.5760E+04, 3.5855E+04, 3.5950E+04, 3.6046E+04, 3.6141E+04,
     +  3.6237E+04, 3.6333E+04, 3.6429E+04, 3.6525E+04, 3.6621E+04,
     + 3.6717E+04, 3.6814E+04, 3.6910E+04, 3.7007E+04, 3.7104E+04,
     +  3.7201E+04, 3.7298E+04, 3.7395E+04, 3.7493E+04, 3.7590E+04,
     + 3.7688E+04, 3.7786E+04, 3.7883E+04, 3.7981E+04, 3.8080E+04,
     +  3.8178E+04, 3.8276E+04, 3.8375E+04, 3.8473E+04, 3.8572E+04,
     + 3.8671E+04, 3.8770E+04, 3.8869E+04, 3.8969E+04, 3.9068E+04,
     +  3.9168E+04, 3.9267E+04, 3.9367E+04, 3.9467E+04, 3.9567E+04,
     + 3.9667E+04, 3.9768E+04, 3.9868E+04, 3.9969E+04, 4.0069E+04,
     +  4.0170E+04, 4.0272E+04, 4.0372E+04, 4.0474E+04, 4.0575E+04,
     + 4.0677E+04, 4.0778E+04, 4.0880E+04, 4.0982E+04, 4.1084E+04,
     +  4.1186E+04, 4.1288E+04, 4.1391E+04, 4.1493E+04, 4.1596E+04,
     + 4.1699E+04, 4.1802E+04, 4.1905E+04, 4.2008E+04, 4.2111E+04,
     +  4.2215E+04, 4.2318E+04, 4.2422E+04, 4.2526E+04, 4.2630E+04,
     + 4.2734E+04, 4.2838E+04, 4.2942E+04, 4.3047E+04, 4.3152E+04,
     +  4.3256E+04, 4.3361E+04, 4.3466E+04, 4.3571E+04, 4.3676E+04,
     + 4.3782E+04, 4.3887E+04, 4.3993E+04, 4.4099E+04, 4.4205E+04,
     +  4.4311E+04, 4.4417E+04, 4.4523E+04, 4.4630E+04, 4.4736E+04,
     + 4.4843E+04, 4.4950E+04, 4.5057E+04, 4.5164E+04, 4.5271E+04,
     +  4.5378E+04, 4.5486E+04, 4.5593E+04, 4.5701E+04, 4.5809E+04,
     + 4.5917E+04, 4.6025E+04, 4.6133E+04, 4.6242E+04, 4.6350E+04,
     +  4.6459E+04, 4.6568E+04, 4.6676E+04, 4.6785E+04, 4.6895E+04,
     + 4.7004E+04, 4.7113E+04, 4.7223E+04, 4.7332E+04, 4.7442E+04,
     +  4.7552E+04, 4.7662E+04, 4.7772E+04, 4.7883E+04, 4.7993E+04,
     + 4.8104E+04, 4.8214E+04, 4.8325E+04, 4.8436E+04, 4.8547E+04,
     +  4.8658E+04, 4.8770E+04, 4.8881E+04, 4.8993E+04, 4.9105E+04,
     + 4.9217E+04, 4.9329E+04, 4.9441E+04, 4.9553E+04, 4.9665E+04,
     +  4.9778E+04, 4.9891E+04, 5.0004E+04, 5.0116E+04, 5.0230E+04,
     + 5.0343E+04, 5.0456E+04, 5.0570E+04, 5.0683E+04, 5.0797E+04,
     +  5.0911E+04, 5.1025E+04, 5.1139E+04, 5.1253E+04, 5.1367E+04,
     + 5.1482E+04, 5.1596E+04, 5.1711E+04, 5.1826E+04, 5.1941E+04,
     +  5.2056E+04, 5.2172E+04, 5.2287E+04, 5.2402E+04, 5.2518E+04,
     + 5.2634E+04, 5.2750E+04, 5.2866E+04, 5.2982E+04, 5.3098E+04,
     +  5.3215E+04, 5.3331E+04, 5.3448E+04, 5.3565E+04, 5.3682E+04,
     + 5.3799E+04, 5.3916E+04, 5.4034E+04, 5.4151E+04, 5.4269E+04,
     +  5.4386E+04, 5.4504E+04, 5.4622E+04, 5.4740E+04, 5.4859E+04,
     + 5.4977E+04, 5.5096E+04, 5.5214E+04, 5.5333E+04, 5.5452E+04,
     +  5.5571E+04, 5.5690E+04, 5.5810E+04, 5.5929E+04, 5.6049E+04,
     + 5.6168E+04, 5.6288E+04, 5.6408E+04, 5.6528E+04, 5.6649E+04,
     +  5.6769E+04, 5.6889E+04, 5.7010E+04, 5.7131E+04, 5.7252E+04,
     + 5.7373E+04, 5.7494E+04, 5.7615E+04, 5.7737E+04, 5.7858E+04,
     +  5.7980E+04, 5.8102E+04, 5.8224E+04, 5.8346E+04, 5.8468E+04,
     + 5.8590E+04, 5.8713E+04, 5.8835E+04, 5.8958E+04, 5.9081E+04,
     +  5.9204E+04, 5.9327E+04, 5.9450E+04, 5.9574E+04, 5.9697E+04,
     + 5.9821E+04, 5.9945E+04, 6.0068E+04, 6.0193E+04, 6.0317E+04,
     +  6.0441E+04, 6.0565E+04, 6.0690E+04, 6.0815E+04, 6.0940E+04,
     + 6.1065E+04, 6.1190E+04, 6.1315E+04, 6.1440E+04, 6.1566E+04,
     +  6.1691E+04, 6.1817E+04, 6.1943E+04, 6.2069E+04, 6.2195E+04,
     + 6.2322E+04, 6.2448E+04, 6.2574E+04, 6.2701E+04, 6.2828E+04,
     +  6.2955E+04, 6.3082E+04, 6.3209E+04, 6.3336E+04, 6.3464E+04,
     + 6.3591E+04, 6.3719E+04, 6.3847E+04, 6.3975E+04, 6.4103E+04,
     +  6.4231E+04, 6.4360E+04, 6.4488E+04, 6.4617E+04, 6.4746E+04,
     + 6.4875E+04, 6.5004E+04, 6.5133E+04, 6.5262E+04, 6.5392E+04,
     +  6.5521E+04, 6.5651E+04, 6.5781E+04, 6.5911E+04, 6.6041E+04,
     + 6.6171E+04, 6.6301E+04, 6.6432E+04, 6.6562E+04, 6.6693E+04,
     +  6.6824E+04, 6.6955E+04, 6.7086E+04, 6.7217E+04, 6.7349E+04,
     + 6.7480E+04, 6.7612E+04, 6.7744E+04, 6.7876E+04, 6.8008E+04,
     +  6.8140E+04, 6.8272E+04, 6.8405E+04, 6.8538E+04, 6.8670E+04,
     + 6.8803E+04, 6.8936E+04, 6.9069E+04, 6.9202E+04, 6.9336E+04,
     +  6.9470E+04, 6.9603E+04, 6.9737E+04, 6.9871E+04, 7.0005E+04,
     + 7.0139E+04, 7.0274E+04, 7.0408E+04, 7.0542E+04, 7.0677E+04,
     +  7.0812E+04, 7.0947E+04, 7.1082E+04, 7.1217E+04, 7.1353E+04,
     + 7.1488E+04, 7.1624E+04, 7.1760E+04, 7.1895E+04, 7.2032E+04,
     +  7.2168E+04, 7.2304E+04, 7.2440E+04, 7.2577E+04, 7.2714E+04,
     + 7.2851E+04, 7.2987E+04, 7.3125E+04, 7.3262E+04, 7.3399E+04,
     +  7.3537E+04, 7.3675E+04, 7.3812E+04, 7.3950E+04, 7.4088E+04,
     + 7.4226E+04, 7.4364E+04, 7.4503E+04, 7.4642E+04, 7.4780E+04,
     +  7.4919E+04, 7.5058E+04, 7.5197E+04, 7.5336E+04, 7.5476E+04,
     + 7.5615E+04, 7.5755E+04, 7.5894E+04, 7.6034E+04, 7.6174E+04,
     +  7.6314E+04, 7.6455E+04, 7.6595E+04, 7.6736E+04, 7.6876E+04,
     + 7.7017E+04, 7.7158E+04, 7.7299E+04, 7.7441E+04, 7.7582E+04,
     +  7.7723E+04, 7.7865E+04, 7.8007E+04, 7.8149E+04, 7.8291E+04,
     + 7.8433E+04, 7.8575E+04, 7.8717E+04, 7.8860E+04, 7.9003E+04,
     +  7.9146E+04, 7.9288E+04, 7.9431E+04, 7.9575E+04, 7.9718E+04,
     + 7.9861E+04, 8.0005E+04, 8.0149E+04, 8.0292E+04, 8.0437E+04,
     +  8.0581E+04, 8.0725E+04, 8.0870E+04, 8.1014E+04, 8.1159E+04,
     + 8.1303E+04, 8.1448E+04, 8.1594E+04, 8.1739E+04, 8.1884E+04,
     +  8.2029E+04, 8.2175E+04, 8.2321E+04, 8.2466E+04, 8.2613E+04,
     + 8.2759E+04, 8.2905E+04, 8.3051E+04, 8.3198E+04, 8.3344E+04,
     +  8.3491E+04, 8.3638E+04, 8.3785E+04, 8.3932E+04, 8.4080E+04,
     + 8.4227E+04, 8.4375E+04, 8.4523E+04, 8.4671E+04, 8.4819E+04,
     +  8.4966E+04, 8.5115E+04, 8.5263E+04, 8.5412E+04, 8.5561E+04,
     + 8.5709E+04, 8.5858E+04, 8.6007E+04, 8.6156E+04, 8.6306E+04,
     +  8.6455E+04, 8.6605E+04, 8.6754E+04, 8.6904E+04, 8.7054E+04,
     + 8.7205E+04, 8.7355E+04, 8.7505E+04, 8.7656E+04, 8.7806E+04,
     +  8.7957E+04, 8.8108E+04, 8.8259E+04, 8.8411E+04, 8.8562E+04,
     + 8.8713E+04, 8.8865E+04, 8.9016E+04, 8.9168E+04, 8.9320E+04,
     +  8.9472E+04, 8.9625E+04, 8.9777E+04, 8.9930E+04, 9.0083E+04,
     + 9.0235E+04, 9.0388E+04, 9.0541E+04, 9.0695E+04, 9.0848E+04,
     +  9.1001E+04, 9.1155E+04, 9.1309E+04, 9.1462E+04, 9.1616E+04,
     + 9.1770E+04, 9.1925E+04, 9.2079E+04, 9.2234E+04, 9.2389E+04,
     +  9.2543E+04, 9.2698E+04, 9.2853E+04, 9.3009E+04, 9.3164E+04,
     + 9.3319E+04, 9.3475E+04, 9.3631E+04, 9.3787E+04, 9.3943E+04,
     +  9.4099E+04, 9.4255E+04, 9.4411E+04, 9.4568E+04, 9.4725E+04,
     + 9.4882E+04, 9.5038E+04, 9.5196E+04, 9.5353E+04, 9.5510E+04,
     +  9.5668E+04, 9.5825E+04, 9.5983E+04, 9.6141E+04, 9.6299E+04,
     + 9.6457E+04, 9.6615E+04, 9.6774E+04, 9.6932E+04, 9.7091E+04,
     +  9.7250E+04, 9.7409E+04, 9.7568E+04, 9.7727E+04, 9.7887E+04,
     + 9.8046E+04, 9.8206E+04, 9.8365E+04, 9.8526E+04, 9.8686E+04,
     +  9.8846E+04, 9.9006E+04, 9.9166E+04, 9.9327E+04, 9.9488E+04,
     + 9.9649E+04, 9.9810E+04, 9.9971E+04, 1.0013E+05, 1.0029E+05,
     +  1.0045E+05, 1.0062E+05, 1.0078E+05, 1.0094E+05, 1.0110E+05,
     + 1.0126E+05, 1.0143E+05, 1.0159E+05, 1.0175E+05, 1.0191E+05,
     +  1.0208E+05, 1.0224E+05, 1.0240E+05, 1.0257E+05, 1.0273E+05,
     + 1.0289E+05, 1.0306E+05, 1.0322E+05, 1.0339E+05, 1.0355E+05,
     +  1.0371E+05, 1.0388E+05, 1.0404E+05, 1.0421E+05, 1.0437E+05,
     + 1.0454E+05, 1.0470E+05, 1.0487E+05, 1.0503E+05, 1.0520E+05,
     +  1.0536E+05, 1.0553E+05, 1.0569E+05, 1.0586E+05, 1.0603E+05,
     + 1.0619E+05, 1.0636E+05, 1.0653E+05, 1.0669E+05, 1.0686E+05,
     +  1.0703E+05, 1.0719E+05, 1.0736E+05, 1.0753E+05, 1.0770E+05,
     + 1.0786E+05, 1.0803E+05, 1.0820E+05, 1.0837E+05, 1.0853E+05,
     +  1.0870E+05, 1.0887E+05, 1.0904E+05, 1.0921E+05, 1.0938E+05,
     + 1.0955E+05, 1.0972E+05, 1.0988E+05, 1.1005E+05, 1.1022E+05,
     +  1.1039E+05, 1.1056E+05, 1.1073E+05, 1.1090E+05, 1.1107E+05,
     + 1.1124E+05, 1.1141E+05, 1.1158E+05, 1.1175E+05, 1.1193E+05,
     +  1.1210E+05, 1.1227E+05, 1.1244E+05, 1.1261E+05, 1.1278E+05,
     + 1.1295E+05, 1.1312E+05, 1.1330E+05, 1.1347E+05, 1.1364E+05,
     +  1.1381E+05, 1.1399E+05, 1.1416E+05, 1.1433E+05, 1.1450E+05,
     + 1.1468E+05, 1.1485E+05, 1.1502E+05, 1.1520E+05, 1.1537E+05,
     +  1.1554E+05, 1.1572E+05, 1.1589E+05, 1.1607E+05, 1.1624E+05,
     + 1.1641E+05, 1.1659E+05, 1.1676E+05, 1.1694E+05, 1.1711E+05,
     +  1.1729E+05, 1.1746E+05, 1.1764E+05, 1.1781E+05, 1.1799E+05,
     + 1.1817E+05, 1.1834E+05, 1.1852E+05, 1.1869E+05, 1.1887E+05,
     +  1.1905E+05, 1.1922E+05, 1.1940E+05, 1.1958E+05, 1.1975E+05,
     + 1.1993E+05, 1.2011E+05, 1.2028E+05, 1.2046E+05, 1.2064E+05,
     +  1.2082E+05, 1.2100E+05, 1.2117E+05, 1.2135E+05, 1.2153E+05,
     + 1.2171E+05, 1.2189E+05, 1.2207E+05, 1.2224E+05, 1.2242E+05,
     +  1.2260E+05, 1.2278E+05, 1.2296E+05, 1.2314E+05, 1.2332E+05,
     + 1.2350E+05, 1.2368E+05, 1.2386E+05, 1.2404E+05, 1.2422E+05,
     +  1.2440E+05, 1.2458E+05, 1.2476E+05, 1.2494E+05, 1.2512E+05,
     + 1.2531E+05, 1.2549E+05, 1.2567E+05, 1.2585E+05, 1.2603E+05,
     +  1.2621E+05, 1.2639E+05, 1.2658E+05, 1.2676E+05, 1.2694E+05,
     + 1.2712E+05, 1.2731E+05, 1.2749E+05, 1.2767E+05, 1.2786E+05,
     +  1.2804E+05, 1.2822E+05, 1.2841E+05, 1.2859E+05, 1.2877E+05,
     + 1.2896E+05, 1.2914E+05, 1.2932E+05, 1.2951E+05, 1.2969E+05,
     +  1.2988E+05, 1.3006E+05, 1.3025E+05, 1.3043E+05, 1.3062E+05,
     + 1.3080E+05, 1.3099E+05, 1.3117E+05, 1.3136E+05, 1.3155E+05,
     +  1.3173E+05, 1.3192E+05, 1.3210E+05, 1.3229E+05, 1.3248E+05,
     + 1.3266E+05, 1.3285E+05, 1.3304E+05, 1.3322E+05, 1.3341E+05,
     +  1.3360E+05, 1.3379E+05, 1.3397E+05, 1.3416E+05, 1.3435E+05,
     + 1.3454E+05, 1.3472E+05, 1.3491E+05, 1.3510E+05, 1.3529E+05,
     +  1.3548E+05, 1.3567E+05, 1.3586E+05, 1.3604E+05, 1.3623E+05,
     + 1.3642E+05, 1.3661E+05, 1.3680E+05, 1.3699E+05, 1.3718E+05,
     +  1.3737E+05, 1.3756E+05, 1.3775E+05, 1.3794E+05, 1.3813E+05,
     + 1.3832E+05, 1.3852E+05, 1.3871E+05, 1.3890E+05, 1.3909E+05,
     +  1.3928E+05, 1.3947E+05, 1.3966E+05, 1.3986E+05, 1.4005E+05,
     + 1.4024E+05, 1.4043E+05, 1.4062E+05, 1.4082E+05, 1.4101E+05,
     +  1.4120E+05, 1.4139E+05, 1.4159E+05, 1.4178E+05, 1.4197E+05,
     + 1.4217E+05, 1.4236E+05, 1.4255E+05, 1.4275E+05, 1.4294E+05,
     +  1.4314E+05, 1.4333E+05, 1.4352E+05, 1.4372E+05, 1.4391E+05,
     + 1.4411E+05, 1.4430E+05, 1.4450E+05, 1.4469E+05, 1.4489E+05,
     +  1.4509E+05, 1.4528E+05, 1.4548E+05, 1.4567E+05, 1.4587E+05,
     + 1.4606E+05, 1.4626E+05, 1.4646E+05, 1.4665E+05, 1.4685E+05,
     +  1.4705E+05, 1.4724E+05, 1.4744E+05, 1.4764E+05, 1.4784E+05,
     + 1.4803E+05, 1.4823E+05, 1.4843E+05, 1.4863E+05, 1.4883E+05,
     +  1.4902E+05, 1.4922E+05, 1.4942E+05, 1.4962E+05, 1.4982E+05,
     + 1.5002E+05, 1.5022E+05, 1.5042E+05, 1.5061E+05, 1.5081E+05,
     +  1.5101E+05, 1.5121E+05, 1.5141E+05, 1.5161E+05, 1.5181E+05,
     + 1.5201E+05, 1.5221E+05, 1.5242E+05, 1.5262E+05, 1.5282E+05,
     +  1.5302E+05, 1.5322E+05, 1.5342E+05, 1.5362E+05, 1.5382E+05,
     + 1.5403E+05, 1.5423E+05, 1.5443E+05, 1.5463E+05, 1.5483E+05,
     +  1.5504E+05, 1.5524E+05, 1.5544E+05, 1.5564E+05, 1.5585E+05,
     + 1.5605E+05, 1.5625E+05, 1.5646E+05, 1.5666E+05, 1.5686E+05,
     +  1.5707E+05, 1.5727E+05, 1.5747E+05, 1.5768E+05, 1.5788E+05,
     + 1.5809E+05, 1.5829E+05, 1.5850E+05, 1.5870E+05, 1.5891E+05,
     +  1.5911E+05, 1.5932E+05, 1.5952E+05, 1.5973E+05, 1.5993E+05,
     + 1.6014E+05, 1.6035E+05, 1.6055E+05, 1.6076E+05, 1.6097E+05,
     +  1.6117E+05, 1.6138E+05, 1.6158E+05, 1.6179E+05, 1.6200E+05,
     + 1.6221E+05, 1.6241E+05, 1.6262E+05, 1.6283E+05, 1.6304E+05,
     +  1.6324E+05, 1.6345E+05, 1.6366E+05, 1.6387E+05, 1.6408E+05,
     + 1.6429E+05, 1.6449E+05, 1.6470E+05, 1.6491E+05, 1.6512E+05,
     +  1.6533E+05, 1.6554E+05, 1.6575E+05, 1.6596E+05, 1.6617E+05,
     + 1.6638E+05, 1.6659E+05, 1.6680E+05, 1.6701E+05, 1.6722E+05,
     +  1.6743E+05, 1.6764E+05, 1.6785E+05, 1.6806E+05, 1.6828E+05,
     + 1.6849E+05, 1.6870E+05, 1.6891E+05, 1.6912E+05, 1.6933E+05,
     +  1.6955E+05, 1.6976E+05, 1.6997E+05, 1.7018E+05, 1.7040E+05,
     + 1.7061E+05, 1.7082E+05, 1.7103E+05, 1.7125E+05, 1.7146E+05,
     +  1.7167E+05, 1.7189E+05, 1.7210E+05, 1.7232E+05, 1.7253E+05,
     + 1.7274E+05, 1.7296E+05, 1.7317E+05, 1.7339E+05, 1.7360E+05,
     +  1.7382E+05, 1.7403E+05, 1.7425E+05, 1.7446E+05, 1.7468E+05,
     + 1.7489E+05, 1.7511E+05, 1.7532E+05, 1.7554E+05, 1.7576E+05,
     +  1.7597E+05, 1.7619E+05, 1.7641E+05, 1.7662E+05, 1.7684E+05,
     + 1.7706E+05, 1.7727E+05, 1.7749E+05, 1.7771E+05, 1.7793E+05,
     +  1.7814E+05, 1.7836E+05, 1.7858E+05, 1.7880E+05, 1.7901E+05,
     + 1.7923E+05, 1.7945E+05, 1.7967E+05, 1.7989E+05, 1.8011E+05,
     +  1.8033E+05, 1.8055E+05, 1.8076E+05, 1.8098E+05, 1.8120E+05,
     + 1.8142E+05, 1.8164E+05, 1.8186E+05, 1.8208E+05, 1.8230E+05,
     +  1.8252E+05, 1.8274E+05, 1.8297E+05, 1.8319E+05, 1.8341E+05,
     + 1.8363E+05, 1.8385E+05, 1.8407E+05, 1.8429E+05, 1.8451E+05,
     +  1.8474E+05, 1.8496E+05, 1.8518E+05, 1.8540E+05, 1.8562E+05,
     + 1.8585E+05, 1.8607E+05, 1.8629E+05, 1.8652E+05, 1.8674E+05,
     +  1.8696E+05, 1.8718E+05, 1.8741E+05, 1.8763E+05, 1.8786E+05,
     + 1.8808E+05, 1.8830E+05, 1.8853E+05, 1.8875E+05, 1.8898E+05,
     +  1.8920E+05, 1.8943E+05, 1.8965E+05, 1.8988E+05, 1.9010E+05,
     + 1.9033E+05, 1.9055E+05, 1.9078E+05, 1.9100E+05, 1.9123E+05,
     +  1.9145E+05, 1.9168E+05, 1.9191E+05, 1.9213E+05, 1.9236E+05,
     + 1.9259E+05, 1.9281E+05, 1.9304E+05, 1.9327E+05, 1.9350E+05,
     +  1.9372E+05, 1.9395E+05, 1.9418E+05, 1.9441E+05, 1.9463E+05,
     + 1.9486E+05, 1.9509E+05, 1.9532E+05, 1.9555E+05, 1.9578E+05,
     +  1.9600E+05, 1.9623E+05, 1.9646E+05, 1.9669E+05, 1.9692E+05,
     + 1.9715E+05, 1.9738E+05, 1.9761E+05, 1.9784E+05, 1.9807E+05,
     +  1.9830E+05, 1.9853E+05, 1.9876E+05, 1.9899E+05, 1.9922E+05,
     + 1.9945E+05, 1.9968E+05, 1.9992E+05, 2.0015E+05, 2.0038E+05,
     +  2.0061E+05, 2.0084E+05, 2.0107E+05, 2.0131E+05, 2.0154E+05,
     + 2.0177E+05, 2.0200E+05, 2.0223E+05, 2.0247E+05, 2.0270E+05,
     +  2.0293E+05, 2.0317E+05, 2.0340E+05, 2.0363E+05, 2.0387E+05,
     + 2.0410E+05, 2.0433E+05, 2.0457E+05, 2.0480E+05, 2.0504E+05,
     +  2.0527E+05, 2.0551E+05, 2.0574E+05, 2.0597E+05, 2.0621E+05,
     + 2.0645E+05, 2.0668E+05, 2.0692E+05, 2.0715E+05, 2.0739E+05,
     +  2.0762E+05, 2.0786E+05, 2.0810E+05, 2.0833E+05, 2.0857E+05,
     + 2.0880E+05, 2.0904E+05, 2.0928E+05, 2.0951E+05, 2.0975E+05,
     +  2.0999E+05, 2.1023E+05, 2.1046E+05, 2.1070E+05, 2.1094E+05,
     + 2.1118E+05, 2.1142E+05, 2.1165E+05, 2.1189E+05, 2.1213E+05,
     +  2.1237E+05, 2.1261E+05, 2.1285E+05, 2.1309E+05, 2.1332E+05,
     + 2.1356E+05, 2.1380E+05, 2.1404E+05, 2.1428E+05, 2.1452E+05,
     +  2.1476E+05, 2.1500E+05, 2.1524E+05, 2.1548E+05, 2.1572E+05,
     + 2.1597E+05, 2.1621E+05, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_OCS   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 5),QofT( 5,NT),Q(NT)
      data (Tmx(I),I=1,5)/3500.,3500.,3500.,3500.,3500./
      data xgj/ 1.,1.,2.,4.,1./
c...      OCS
c...        --       622
      data (QofT( 1,J),J=1,2408)/ 3.7809E+00, 7.1972E+00, 1.0621E+01,
     +  1.4046E+01, 1.7473E+01, 2.0899E+01, 2.4326E+01, 2.7753E+01,
     + 3.1180E+01, 3.4608E+01, 3.8035E+01, 4.1463E+01, 4.4891E+01,
     +  4.8319E+01, 5.1746E+01, 5.5174E+01, 5.8603E+01, 6.2031E+01,
     + 6.5459E+01, 6.8887E+01, 7.5744E+01, 8.2601E+01, 8.9459E+01,
     +  9.6317E+01, 1.0318E+02, 1.1003E+02, 1.1689E+02, 1.2375E+02,
     + 1.3061E+02, 1.3747E+02, 1.4433E+02, 1.5119E+02, 1.5805E+02,
     +  1.6491E+02, 1.7178E+02, 1.7864E+02, 1.8550E+02, 1.9236E+02,
     + 1.9922E+02, 2.0609E+02, 2.1295E+02, 2.1982E+02, 2.2668E+02,
     +  2.3355E+02, 2.4042E+02, 2.4729E+02, 2.5416E+02, 2.6103E+02,
     + 2.6791E+02, 2.7478E+02, 2.8166E+02, 2.8855E+02, 2.9543E+02,
     +  3.0232E+02, 3.0922E+02, 3.1612E+02, 3.2303E+02, 3.2994E+02,
     + 3.3687E+02, 3.4380E+02, 3.5074E+02, 3.5769E+02, 3.6465E+02,
     +  3.7162E+02, 3.7861E+02, 3.8560E+02, 3.9262E+02, 3.9965E+02,
     + 4.0669E+02, 4.1375E+02, 4.2083E+02, 4.2793E+02, 4.3505E+02,
     +  4.4219E+02, 4.4936E+02, 4.5654E+02, 4.6376E+02, 4.7099E+02,
     + 4.7825E+02, 4.8555E+02, 4.9287E+02, 5.0022E+02, 5.0760E+02,
     +  5.1501E+02, 5.2247E+02, 5.2995E+02, 5.3747E+02, 5.4502E+02,
     + 5.5261E+02, 5.6024E+02, 5.6791E+02, 5.7562E+02, 5.8338E+02,
     +  5.9117E+02, 5.9900E+02, 6.0689E+02, 6.1482E+02, 6.2280E+02,
     + 6.3082E+02, 6.3889E+02, 6.4702E+02, 6.5519E+02, 6.6341E+02,
     +  6.7170E+02, 6.8002E+02, 6.8841E+02, 6.9686E+02, 7.0535E+02,
     + 7.1390E+02, 7.2252E+02, 7.3119E+02, 7.3992E+02, 7.4872E+02,
     +  7.5758E+02, 7.6649E+02, 7.7547E+02, 7.8452E+02, 7.9363E+02,
     + 8.0281E+02, 8.1206E+02, 8.2137E+02, 8.3075E+02, 8.4020E+02,
     +  8.4972E+02, 8.5931E+02, 8.6898E+02, 8.7871E+02, 8.8853E+02,
     + 8.9841E+02, 9.0837E+02, 9.1841E+02, 9.2852E+02, 9.3872E+02,
     +  9.4898E+02, 9.5934E+02, 9.6976E+02, 9.8027E+02, 9.9086E+02,
     + 1.0015E+03, 1.0123E+03, 1.0231E+03, 1.0341E+03, 1.0451E+03,
     +  1.0562E+03, 1.0674E+03, 1.0786E+03, 1.0900E+03, 1.1014E+03,
     + 1.1130E+03, 1.1246E+03, 1.1363E+03, 1.1482E+03, 1.1601E+03,
     +  1.1721E+03, 1.1841E+03, 1.1963E+03, 1.2086E+03, 1.2210E+03,
     + 1.2335E+03, 1.2461E+03, 1.2588E+03, 1.2715E+03, 1.2844E+03,
     +  1.2974E+03, 1.3105E+03, 1.3237E+03, 1.3369E+03, 1.3503E+03,
     + 1.3638E+03, 1.3774E+03, 1.3912E+03, 1.4050E+03, 1.4189E+03,
     +  1.4329E+03, 1.4471E+03, 1.4613E+03, 1.4757E+03, 1.4902E+03,
     + 1.5047E+03, 1.5194E+03, 1.5343E+03, 1.5492E+03, 1.5642E+03,
     +  1.5794E+03, 1.5947E+03, 1.6101E+03, 1.6256E+03, 1.6412E+03,
     + 1.6570E+03, 1.6728E+03, 1.6888E+03, 1.7050E+03, 1.7212E+03,
     +  1.7376E+03, 1.7541E+03, 1.7707E+03, 1.7874E+03, 1.8043E+03,
     + 1.8213E+03, 1.8384E+03, 1.8557E+03, 1.8731E+03, 1.8906E+03,
     +  1.9082E+03, 1.9260E+03, 1.9439E+03, 1.9620E+03, 1.9802E+03,
     + 1.9985E+03, 2.0169E+03, 2.0355E+03, 2.0543E+03, 2.0731E+03,
     +  2.0921E+03, 2.1113E+03, 2.1306E+03, 2.1500E+03, 2.1696E+03,
     + 2.1893E+03, 2.2092E+03, 2.2292E+03, 2.2494E+03, 2.2697E+03,
     +  2.2901E+03, 2.3107E+03, 2.3315E+03, 2.3524E+03, 2.3734E+03,
     + 2.3947E+03, 2.4160E+03, 2.4375E+03, 2.4592E+03, 2.4810E+03,
     +  2.5030E+03, 2.5252E+03, 2.5475E+03, 2.5699E+03, 2.5925E+03,
     + 2.6153E+03, 2.6383E+03, 2.6614E+03, 2.6847E+03, 2.7081E+03,
     +  2.7317E+03, 2.7555E+03, 2.7794E+03, 2.8035E+03, 2.8278E+03,
     + 2.8522E+03, 2.8768E+03, 2.9016E+03, 2.9266E+03, 2.9517E+03,
     +  2.9770E+03, 3.0025E+03, 3.0282E+03, 3.0540E+03, 3.0800E+03,
     + 3.1062E+03, 3.1326E+03, 3.1592E+03, 3.1859E+03, 3.2128E+03,
     +  3.2399E+03, 3.2672E+03, 3.2947E+03, 3.3224E+03, 3.3502E+03,
     + 3.3783E+03, 3.4065E+03, 3.4350E+03, 3.4636E+03, 3.4924E+03,
     +  3.5214E+03, 3.5506E+03, 3.5800E+03, 3.6096E+03, 3.6394E+03,
     + 3.6694E+03, 3.6996E+03, 3.7300E+03, 3.7606E+03, 3.7914E+03,
     +  3.8224E+03, 3.8537E+03, 3.8851E+03, 3.9167E+03, 3.9485E+03,
     + 3.9806E+03, 4.0128E+03, 4.0453E+03, 4.0780E+03, 4.1109E+03,
     +  4.1440E+03, 4.1774E+03, 4.2109E+03, 4.2447E+03, 4.2786E+03,
     + 4.3129E+03, 4.3473E+03, 4.3819E+03, 4.4168E+03, 4.4519E+03,
     +  4.4872E+03, 4.5228E+03, 4.5586E+03, 4.5946E+03, 4.6308E+03,
     + 4.6673E+03, 4.7040E+03, 4.7409E+03, 4.7781E+03, 4.8155E+03,
     +  4.8532E+03, 4.8911E+03, 4.9292E+03, 4.9676E+03, 5.0062E+03,
     + 5.0450E+03, 5.0841E+03, 5.1235E+03, 5.1631E+03, 5.2029E+03,
     +  5.2430E+03, 5.2833E+03, 5.3239E+03, 5.3648E+03, 5.4059E+03,
     + 5.4472E+03, 5.4888E+03, 5.5307E+03, 5.5728E+03, 5.6152E+03,
     +  5.6579E+03, 5.7008E+03, 5.7440E+03, 5.7874E+03, 5.8311E+03,
     + 5.8751E+03, 5.9193E+03, 5.9638E+03, 6.0086E+03, 6.0537E+03,
     +  6.0990E+03, 6.1446E+03, 6.1905E+03, 6.2367E+03, 6.2831E+03,
     + 6.3298E+03, 6.3768E+03, 6.4241E+03, 6.4717E+03, 6.5195E+03,
     +  6.5677E+03, 6.6161E+03, 6.6648E+03, 6.7138E+03, 6.7631E+03,
     + 6.8127E+03, 6.8626E+03, 6.9128E+03, 6.9633E+03, 7.0141E+03,
     +  7.0652E+03, 7.1166E+03, 7.1682E+03, 7.2202E+03, 7.2725E+03,
     + 7.3252E+03, 7.3781E+03, 7.4313E+03, 7.4848E+03, 7.5387E+03,
     +  7.5928E+03, 7.6473E+03, 7.7021E+03, 7.7572E+03, 7.8127E+03,
     + 7.8684E+03, 7.9245E+03, 7.9809E+03, 8.0376E+03, 8.0947E+03,
     +  8.1520E+03, 8.2098E+03, 8.2678E+03, 8.3262E+03, 8.3849E+03,
     + 8.4439E+03, 8.5033E+03, 8.5630E+03, 8.6231E+03, 8.6835E+03,
     +  8.7442E+03, 8.8053E+03, 8.8668E+03, 8.9285E+03, 8.9907E+03,
     + 9.0532E+03, 9.1160E+03, 9.1792E+03, 9.2427E+03, 9.3066E+03,
     +  9.3709E+03, 9.4355E+03, 9.5005E+03, 9.5658E+03, 9.6315E+03,
     + 9.6976E+03, 9.7640E+03, 9.8309E+03, 9.8980E+03, 9.9656E+03,
     +  1.0033E+04, 1.0102E+04, 1.0170E+04, 1.0240E+04, 1.0309E+04,
     + 1.0379E+04, 1.0449E+04, 1.0520E+04, 1.0591E+04, 1.0662E+04,
     +  1.0734E+04, 1.0806E+04, 1.0878E+04, 1.0951E+04, 1.1025E+04,
     + 1.1098E+04, 1.1172E+04, 1.1247E+04, 1.1322E+04, 1.1397E+04,
     +  1.1473E+04, 1.1549E+04, 1.1626E+04, 1.1703E+04, 1.1780E+04,
     + 1.1858E+04, 1.1936E+04, 1.2015E+04, 1.2094E+04, 1.2173E+04,
     +  1.2253E+04, 1.2333E+04, 1.2414E+04, 1.2495E+04, 1.2577E+04,
     + 1.2659E+04, 1.2741E+04, 1.2824E+04, 1.2908E+04, 1.2991E+04,
     +  1.3076E+04, 1.3160E+04, 1.3245E+04, 1.3331E+04, 1.3417E+04,
     + 1.3503E+04, 1.3590E+04, 1.3678E+04, 1.3765E+04, 1.3854E+04,
     +  1.3942E+04, 1.4032E+04, 1.4121E+04, 1.4211E+04, 1.4302E+04,
     + 1.4393E+04, 1.4484E+04, 1.4576E+04, 1.4669E+04, 1.4762E+04,
     +  1.4855E+04, 1.4949E+04, 1.5044E+04, 1.5138E+04, 1.5234E+04,
     + 1.5330E+04, 1.5426E+04, 1.5523E+04, 1.5620E+04, 1.5718E+04,
     +  1.5816E+04, 1.5915E+04, 1.6014E+04, 1.6114E+04, 1.6214E+04,
     + 1.6315E+04, 1.6416E+04, 1.6518E+04, 1.6620E+04, 1.6723E+04,
     +  1.6827E+04, 1.6931E+04, 1.7035E+04, 1.7140E+04, 1.7245E+04,
     + 1.7351E+04, 1.7458E+04, 1.7565E+04, 1.7672E+04, 1.7780E+04,
     +  1.7889E+04, 1.7998E+04, 1.8108E+04, 1.8218E+04, 1.8329E+04,
     + 1.8440E+04, 1.8552E+04, 1.8664E+04, 1.8777E+04, 1.8891E+04,
     +  1.9005E+04, 1.9120E+04, 1.9235E+04, 1.9351E+04, 1.9467E+04,
     + 1.9584E+04, 1.9701E+04, 1.9819E+04, 1.9938E+04, 2.0057E+04,
     +  2.0177E+04, 2.0297E+04, 2.0418E+04, 2.0540E+04, 2.0662E+04,
     + 2.0784E+04, 2.0908E+04, 2.1032E+04, 2.1156E+04, 2.1281E+04,
     +  2.1407E+04, 2.1533E+04, 2.1660E+04, 2.1787E+04, 2.1915E+04,
     + 2.2044E+04, 2.2173E+04, 2.2303E+04, 2.2434E+04, 2.2565E+04,
     +  2.2697E+04, 2.2829E+04, 2.2962E+04, 2.3095E+04, 2.3230E+04,
     + 2.3365E+04, 2.3500E+04, 2.3636E+04, 2.3773E+04, 2.3911E+04,
     +  2.4049E+04, 2.4187E+04, 2.4327E+04, 2.4467E+04, 2.4607E+04,
     + 2.4749E+04, 2.4891E+04, 2.5033E+04, 2.5177E+04, 2.5321E+04,
     +  2.5465E+04, 2.5611E+04, 2.5757E+04, 2.5903E+04, 2.6051E+04,
     + 2.6199E+04, 2.6347E+04, 2.6497E+04, 2.6647E+04, 2.6798E+04,
     +  2.6949E+04, 2.7101E+04, 2.7254E+04, 2.7407E+04, 2.7562E+04,
     + 2.7717E+04, 2.7872E+04, 2.8029E+04, 2.8186E+04, 2.8343E+04,
     +  2.8502E+04, 2.8661E+04, 2.8821E+04, 2.8982E+04, 2.9143E+04,
     + 2.9305E+04, 2.9468E+04, 2.9631E+04, 2.9796E+04, 2.9961E+04,
     +  3.0126E+04, 3.0293E+04, 3.0460E+04, 3.0628E+04, 3.0797E+04,
     + 3.0966E+04, 3.1137E+04, 3.1308E+04, 3.1479E+04, 3.1652E+04,
     +  3.1825E+04, 3.1999E+04, 3.2174E+04, 3.2350E+04, 3.2526E+04,
     + 3.2703E+04, 3.2881E+04, 3.3060E+04, 3.3239E+04, 3.3420E+04,
     +  3.3601E+04, 3.3783E+04, 3.3965E+04, 3.4149E+04, 3.4333E+04,
     + 3.4518E+04, 3.4704E+04, 3.4891E+04, 3.5078E+04, 3.5266E+04,
     +  3.5456E+04, 3.5646E+04, 3.5836E+04, 3.6028E+04, 3.6220E+04,
     + 3.6414E+04, 3.6608E+04, 3.6803E+04, 3.6999E+04, 3.7195E+04,
     +  3.7393E+04, 3.7591E+04, 3.7790E+04, 3.7990E+04, 3.8191E+04,
     + 3.8393E+04, 3.8595E+04, 3.8799E+04, 3.9003E+04, 3.9208E+04,
     +  3.9414E+04, 3.9621E+04, 3.9829E+04, 4.0038E+04, 4.0248E+04,
     + 4.0458E+04, 4.0669E+04, 4.0882E+04, 4.1095E+04, 4.1309E+04,
     +  4.1524E+04, 4.1740E+04, 4.1956E+04, 4.2174E+04, 4.2393E+04,
     + 4.2612E+04, 4.2833E+04, 4.3054E+04, 4.3276E+04, 4.3500E+04,
     +  4.3724E+04, 4.3949E+04, 4.4175E+04, 4.4402E+04, 4.4630E+04,
     + 4.4858E+04, 4.5088E+04, 4.5319E+04, 4.5551E+04, 4.5783E+04,
     +  4.6017E+04, 4.6251E+04, 4.6487E+04, 4.6724E+04, 4.6961E+04,
     + 4.7199E+04, 4.7439E+04, 4.7679E+04, 4.7921E+04, 4.8163E+04,
     +  4.8406E+04, 4.8651E+04, 4.8896E+04, 4.9143E+04, 4.9390E+04,
     + 4.9638E+04, 4.9888E+04, 5.0138E+04, 5.0390E+04, 5.0642E+04,
     +  5.0896E+04, 5.1150E+04, 5.1406E+04, 5.1662E+04, 5.1920E+04,
     + 5.2178E+04, 5.2438E+04, 5.2699E+04, 5.2960E+04, 5.3223E+04,
     +  5.3487E+04, 5.3752E+04, 5.4018E+04, 5.4285E+04, 5.4553E+04,
     + 5.4822E+04, 5.5093E+04, 5.5364E+04, 5.5636E+04, 5.5910E+04,
     +  5.6184E+04, 5.6460E+04, 5.6737E+04, 5.7015E+04, 5.7294E+04,
     + 5.7574E+04, 5.7855E+04, 5.8137E+04, 5.8421E+04, 5.8705E+04,
     +  5.8991E+04, 5.9277E+04, 5.9565E+04, 5.9854E+04, 6.0144E+04,
     + 6.0436E+04, 6.0728E+04, 6.1022E+04, 6.1316E+04, 6.1612E+04,
     +  6.1909E+04, 6.2207E+04, 6.2507E+04, 6.2807E+04, 6.3109E+04,
     + 6.3412E+04, 6.3716E+04, 6.4021E+04, 6.4327E+04, 6.4635E+04,
     +  6.4944E+04, 6.5254E+04, 6.5565E+04, 6.5877E+04, 6.6190E+04,
     + 6.6505E+04, 6.6821E+04, 6.7138E+04, 6.7457E+04, 6.7776E+04,
     +  6.8097E+04, 6.8419E+04, 6.8742E+04, 6.9067E+04, 6.9392E+04,
     + 6.9719E+04, 7.0048E+04, 7.0377E+04, 7.0708E+04, 7.1040E+04,
     +  7.1373E+04, 7.1708E+04, 7.2043E+04, 7.2380E+04, 7.2719E+04,
     + 7.3058E+04, 7.3399E+04, 7.3741E+04, 7.4085E+04, 7.4429E+04,
     +  7.4775E+04, 7.5123E+04, 7.5471E+04, 7.5821E+04, 7.6172E+04,
     + 7.6525E+04, 7.6879E+04, 7.7234E+04, 7.7591E+04, 7.7948E+04,
     +  7.8308E+04, 7.8668E+04, 7.9030E+04, 7.9393E+04, 7.9758E+04,
     + 8.0124E+04, 8.0491E+04, 8.0859E+04, 8.1229E+04, 8.1601E+04,
     +  8.1973E+04, 8.2348E+04, 8.2723E+04, 8.3100E+04, 8.3478E+04,
     + 8.3858E+04, 8.4239E+04, 8.4621E+04, 8.5005E+04, 8.5390E+04,
     +  8.5777E+04, 8.6165E+04, 8.6554E+04, 8.6945E+04, 8.7338E+04,
     + 8.7731E+04, 8.8126E+04, 8.8523E+04, 8.8921E+04, 8.9321E+04,
     +  8.9722E+04, 9.0124E+04, 9.0528E+04, 9.0933E+04, 9.1340E+04,
     + 9.1748E+04, 9.2158E+04, 9.2569E+04, 9.2982E+04, 9.3396E+04,
     +  9.3812E+04, 9.4229E+04, 9.4647E+04, 9.5067E+04, 9.5489E+04,
     + 9.5912E+04, 9.6337E+04, 9.6763E+04, 9.7191E+04, 9.7620E+04,
     +  9.8051E+04, 9.8483E+04, 9.8917E+04, 9.9352E+04, 9.9789E+04,
     + 1.0023E+05, 1.0067E+05, 1.0111E+05, 1.0155E+05, 1.0200E+05,
     +  1.0244E+05, 1.0289E+05, 1.0334E+05, 1.0379E+05, 1.0424E+05,
     + 1.0470E+05, 1.0515E+05, 1.0561E+05, 1.0607E+05, 1.0653E+05,
     +  1.0699E+05, 1.0746E+05, 1.0792E+05, 1.0839E+05, 1.0886E+05,
     + 1.0933E+05, 1.0980E+05, 1.1028E+05, 1.1075E+05, 1.1123E+05,
     +  1.1171E+05, 1.1219E+05, 1.1267E+05, 1.1315E+05, 1.1364E+05,
     + 1.1412E+05, 1.1461E+05, 1.1510E+05, 1.1560E+05, 1.1609E+05,
     +  1.1658E+05, 1.1708E+05, 1.1758E+05, 1.1808E+05, 1.1858E+05,
     + 1.1909E+05, 1.1959E+05, 1.2010E+05, 1.2061E+05, 1.2112E+05,
     +  1.2163E+05, 1.2215E+05, 1.2266E+05, 1.2318E+05, 1.2370E+05,
     + 1.2422E+05, 1.2475E+05, 1.2527E+05, 1.2580E+05, 1.2633E+05,
     +  1.2686E+05, 1.2739E+05, 1.2792E+05, 1.2846E+05, 1.2900E+05,
     + 1.2954E+05, 1.3008E+05, 1.3062E+05, 1.3117E+05, 1.3171E+05,
     +  1.3226E+05, 1.3281E+05, 1.3336E+05, 1.3392E+05, 1.3448E+05,
     + 1.3503E+05, 1.3559E+05, 1.3615E+05, 1.3672E+05, 1.3728E+05,
     +  1.3785E+05, 1.3842E+05, 1.3899E+05, 1.3956E+05, 1.4014E+05,
     + 1.4072E+05, 1.4129E+05, 1.4187E+05, 1.4246E+05, 1.4304E+05,
     +  1.4363E+05, 1.4422E+05, 1.4481E+05, 1.4540E+05, 1.4599E+05,
     + 1.4659E+05, 1.4719E+05, 1.4779E+05, 1.4839E+05, 1.4899E+05,
     +  1.4960E+05, 1.5021E+05, 1.5082E+05, 1.5143E+05, 1.5204E+05,
     + 1.5266E+05, 1.5328E+05, 1.5390E+05, 1.5452E+05, 1.5514E+05,
     +  1.5577E+05, 1.5640E+05, 1.5703E+05, 1.5766E+05, 1.5829E+05,
     + 1.5893E+05, 1.5957E+05, 1.6021E+05, 1.6085E+05, 1.6150E+05,
     +  1.6214E+05, 1.6279E+05, 1.6344E+05, 1.6410E+05, 1.6475E+05,
     + 1.6541E+05, 1.6607E+05, 1.6673E+05, 1.6739E+05, 1.6806E+05,
     +  1.6873E+05, 1.6940E+05, 1.7007E+05, 1.7074E+05, 1.7142E+05,
     + 1.7210E+05, 1.7278E+05, 1.7346E+05, 1.7415E+05, 1.7483E+05,
     +  1.7552E+05, 1.7621E+05, 1.7691E+05, 1.7760E+05, 1.7830E+05,
     + 1.7900E+05, 1.7971E+05, 1.8041E+05, 1.8112E+05, 1.8183E+05,
     +  1.8254E+05, 1.8325E+05, 1.8397E+05, 1.8469E+05, 1.8541E+05,
     + 1.8613E+05, 1.8686E+05, 1.8758E+05, 1.8831E+05, 1.8904E+05,
     +  1.8978E+05, 1.9052E+05, 1.9125E+05, 1.9200E+05, 1.9274E+05,
     + 1.9349E+05, 1.9423E+05, 1.9498E+05, 1.9574E+05, 1.9649E+05,
     +  1.9725E+05, 1.9801E+05, 1.9877E+05, 1.9954E+05, 2.0030E+05,
     + 2.0107E+05, 2.0184E+05, 2.0262E+05, 2.0340E+05, 2.0417E+05,
     +  2.0496E+05, 2.0574E+05, 2.0653E+05, 2.0731E+05, 2.0811E+05,
     + 2.0890E+05, 2.0969E+05, 2.1049E+05, 2.1129E+05, 2.1210E+05,
     +  2.1290E+05, 2.1371E+05, 2.1452E+05, 2.1533E+05, 2.1615E+05,
     + 2.1697E+05, 2.1779E+05, 2.1861E+05, 2.1944E+05, 2.2027E+05,
     +  2.2110E+05, 2.2193E+05, 2.2277E+05, 2.2360E+05, 2.2444E+05,
     + 2.2529E+05, 2.2613E+05, 2.2698E+05, 2.2783E+05, 2.2869E+05,
     +  2.2954E+05, 2.3040E+05, 2.3126E+05, 2.3213E+05, 2.3299E+05,
     + 2.3386E+05, 2.3473E+05, 2.3561E+05, 2.3649E+05, 2.3737E+05,
     +  2.3825E+05, 2.3913E+05, 2.4002E+05, 2.4091E+05, 2.4180E+05,
     + 2.4270E+05, 2.4360E+05, 2.4450E+05, 2.4540E+05, 2.4631E+05,
     +  2.4722E+05, 2.4813E+05, 2.4904E+05, 2.4996E+05, 2.5088E+05,
     + 2.5180E+05, 2.5273E+05, 2.5366E+05, 2.5459E+05, 2.5552E+05,
     +  2.5646E+05, 2.5739E+05, 2.5834E+05, 2.5928E+05, 2.6023E+05,
     + 2.6118E+05, 2.6213E+05, 2.6309E+05, 2.6405E+05, 2.6501E+05,
     +  2.6597E+05, 2.6694E+05, 2.6791E+05, 2.6888E+05, 2.6986E+05,
     + 2.7084E+05, 2.7182E+05, 2.7280E+05, 2.7379E+05, 2.7478E+05,
     +  2.7577E+05, 2.7677E+05, 2.7776E+05, 2.7877E+05, 2.7977E+05,
     + 2.8078E+05, 2.8179E+05, 2.8280E+05, 2.8382E+05, 2.8484E+05,
     +  2.8586E+05, 2.8688E+05, 2.8791E+05, 2.8894E+05, 2.8998E+05,
     + 2.9101E+05, 2.9205E+05, 2.9309E+05, 2.9414E+05, 2.9519E+05,
     +  2.9624E+05, 2.9730E+05, 2.9835E+05, 2.9941E+05, 3.0048E+05,
     + 3.0155E+05, 3.0262E+05, 3.0369E+05, 3.0476E+05, 3.0584E+05,
     +  3.0693E+05, 3.0801E+05, 3.0910E+05, 3.1019E+05, 3.1129E+05,
     + 3.1238E+05, 3.1348E+05, 3.1459E+05, 3.1570E+05, 3.1681E+05,
     +  3.1792E+05, 3.1904E+05, 3.2016E+05, 3.2128E+05, 3.2241E+05,
     + 3.2353E+05, 3.2467E+05, 3.2580E+05, 3.2694E+05, 3.2808E+05,
     +  3.2923E+05, 3.3038E+05, 3.3153E+05, 3.3268E+05, 3.3384E+05,
     + 3.3500E+05, 3.3617E+05, 3.3734E+05, 3.3851E+05, 3.3968E+05,
     +  3.4086E+05, 3.4204E+05, 3.4323E+05, 3.4441E+05, 3.4560E+05,
     + 3.4680E+05, 3.4800E+05, 3.4920E+05, 3.5040E+05, 3.5161E+05,
     +  3.5282E+05, 3.5403E+05, 3.5525E+05, 3.5647E+05, 3.5770E+05,
     + 3.5893E+05, 3.6016E+05, 3.6139E+05, 3.6263E+05, 3.6387E+05,
     +  3.6512E+05, 3.6636E+05, 3.6762E+05, 3.6887E+05, 3.7013E+05,
     + 3.7139E+05, 3.7266E+05, 3.7393E+05, 3.7520E+05, 3.7648E+05,
     +  3.7776E+05, 3.7904E+05, 3.8033E+05, 3.8162E+05, 3.8291E+05,
     + 3.8421E+05, 3.8551E+05, 3.8681E+05, 3.8812E+05, 3.8943E+05,
     +  3.9075E+05, 3.9206E+05, 3.9339E+05, 3.9471E+05, 3.9604E+05,
     + 3.9737E+05, 3.9871E+05, 4.0005E+05, 4.0139E+05, 4.0274E+05,
     +  4.0409E+05, 4.0545E+05, 4.0681E+05, 4.0817E+05, 4.0953E+05,
     + 4.1090E+05, 4.1228E+05, 4.1365E+05, 4.1503E+05, 4.1642E+05,
     +  4.1781E+05, 4.1920E+05, 4.2059E+05, 4.2199E+05, 4.2339E+05,
     + 4.2480E+05, 4.2621E+05, 4.2763E+05, 4.2904E+05, 4.3046E+05,
     +  4.3189E+05, 4.3332E+05, 4.3475E+05, 4.3619E+05, 4.3763E+05,
     + 4.3908E+05, 4.4052E+05, 4.4198E+05, 4.4343E+05, 4.4489E+05,
     +  4.4636E+05, 4.4782E+05, 4.4930E+05, 4.5077E+05, 4.5225E+05,
     + 4.5373E+05, 4.5522E+05, 4.5671E+05, 4.5821E+05, 4.5971E+05,
     +  4.6121E+05, 4.6272E+05, 4.6423E+05, 4.6574E+05, 4.6726E+05,
     + 4.6878E+05, 4.7031E+05, 4.7184E+05, 4.7338E+05, 4.7492E+05,
     +  4.7646E+05, 4.7801E+05, 4.7956E+05, 4.8111E+05, 4.8267E+05,
     + 4.8423E+05, 4.8580E+05, 4.8737E+05, 4.8895E+05, 4.9053E+05,
     +  4.9211E+05, 4.9370E+05, 4.9529E+05, 4.9689E+05, 4.9849E+05,
     + 5.0009E+05, 5.0170E+05, 5.0331E+05, 5.0493E+05, 5.0655E+05,
     +  5.0818E+05, 5.0980E+05, 5.1144E+05, 5.1308E+05, 5.1472E+05,
     + 5.1636E+05, 5.1802E+05, 5.1967E+05, 5.2133E+05, 5.2299E+05,
     +  5.2466E+05, 5.2633E+05, 5.2801E+05, 5.2969E+05, 5.3137E+05,
     + 5.3306E+05, 5.3476E+05, 5.3645E+05, 5.3816E+05, 5.3986E+05,
     +  5.4157E+05, 5.4329E+05, 5.4501E+05, 5.4673E+05, 5.4846E+05,
     + 5.5019E+05, 5.5193E+05, 5.5367E+05, 5.5542E+05, 5.5717E+05,
     +  5.5892E+05, 5.6068E+05, 5.6245E+05, 5.6421E+05, 5.6599E+05,
     + 5.6776E+05, 5.6955E+05, 5.7133E+05, 5.7312E+05, 5.7492E+05,
     +  5.7672E+05, 5.7852E+05, 5.8033E+05, 5.8214E+05, 5.8396E+05,
     + 5.8578E+05, 5.8761E+05, 5.8944E+05, 5.9128E+05, 5.9312E+05,
     +  5.9496E+05, 5.9682E+05, 5.9867E+05, 6.0053E+05, 6.0239E+05,
     + 6.0426E+05, 6.0614E+05, 6.0801E+05, 6.0990E+05, 6.1178E+05,
     +  6.1368E+05, 6.1557E+05, 6.1747E+05, 6.1938E+05, 6.2129E+05,
     + 6.2321E+05, 6.2513E+05, 6.2705E+05, 6.2898E+05, 6.3092E+05,
     +  6.3286E+05, 6.3480E+05, 6.3675E+05, 6.3871E+05, 6.4066E+05,
     + 6.4263E+05, 6.4460E+05, 6.4657E+05, 6.4855E+05, 6.5053E+05,
     +  6.5252E+05, 6.5451E+05, 6.5651E+05, 6.5851E+05, 6.6052E+05,
     + 6.6253E+05, 6.6455E+05, 6.6657E+05, 6.6860E+05, 6.7063E+05,
     +  6.7267E+05, 6.7471E+05, 6.7676E+05, 6.7881E+05, 6.8087E+05,
     + 6.8293E+05, 6.8500E+05, 6.8707E+05, 6.8915E+05, 6.9123E+05,
     +  6.9332E+05, 6.9541E+05, 6.9751E+05, 6.9961E+05, 7.0172E+05,
     + 7.0384E+05, 7.0595E+05, 7.0808E+05, 7.1021E+05, 7.1234E+05,
     +  7.1448E+05, 7.1662E+05, 7.1877E+05, 7.2093E+05, 7.2308E+05,
     + 7.2525E+05, 7.2742E+05, 7.2959E+05, 7.3178E+05, 7.3396E+05,
     +  7.3615E+05, 7.3835E+05, 7.4055E+05, 7.4276E+05, 7.4497E+05,
     + 7.4719E+05, 7.4941E+05, 7.5164E+05, 7.5387E+05, 7.5611E+05,
     +  7.5835E+05, 7.6060E+05, 7.6286E+05, 7.6512E+05, 7.6738E+05,
     + 7.6965E+05, 7.7193E+05, 7.7421E+05, 7.7650E+05, 7.7879E+05,
     +  7.8109E+05, 7.8339E+05, 7.8570E+05, 7.8802E+05, 7.9034E+05,
     + 7.9266E+05, 7.9499E+05, 7.9733E+05, 7.9967E+05, 8.0202E+05,
     +  8.0437E+05, 8.0673E+05, 8.0910E+05, 8.1147E+05, 8.1384E+05,
     + 8.1622E+05, 8.1861E+05, 8.2100E+05, 8.2340E+05, 8.2580E+05,
     +  8.2821E+05, 8.3063E+05, 8.3305E+05, 8.3547E+05, 8.3791E+05,
     + 8.4034E+05, 8.4279E+05, 8.4524E+05, 8.4769E+05, 8.5015E+05,
     +  8.5262E+05, 8.5509E+05, 8.5757E+05, 8.6005E+05, 8.6254E+05,
     + 8.6504E+05, 8.6754E+05, 8.7004E+05, 8.7256E+05, 8.7507E+05,
     +  8.7760E+05, 8.8013E+05, 8.8267E+05, 8.8521E+05, 8.8776E+05,
     + 8.9031E+05, 8.9287E+05, 8.9543E+05, 8.9801E+05, 9.0058E+05,
     +  9.0317E+05, 9.0576E+05, 9.0835E+05, 9.1095E+05, 9.1356E+05,
     + 9.1617E+05, 9.1879E+05, 9.2142E+05, 9.2405E+05, 9.2669E+05,
     +  9.2933E+05, 9.3198E+05, 9.3464E+05, 9.3730E+05, 9.3997E+05,
     + 9.4264E+05, 9.4532E+05, 9.4801E+05, 9.5070E+05, 9.5340E+05,
     +  9.5610E+05, 9.5881E+05, 9.6153E+05, 9.6425E+05, 9.6698E+05,
     + 9.6972E+05, 9.7246E+05, 9.7521E+05, 9.7796E+05, 9.8072E+05,
     +  9.8349E+05, 9.8626E+05, 9.8904E+05, 9.9183E+05, 9.9462E+05,
     + 9.9742E+05, 1.0002E+06, 1.0030E+06, 1.0059E+06, 1.0087E+06,
     +  1.0115E+06, 1.0143E+06, 1.0172E+06, 1.0200E+06, 1.0229E+06,
     + 1.0258E+06, 1.0286E+06, 1.0315E+06, 1.0344E+06, 1.0373E+06,
     +  1.0402E+06, 1.0431E+06, 1.0460E+06, 1.0489E+06, 1.0518E+06,
     + 1.0547E+06, 1.0577E+06, 1.0606E+06, 1.0636E+06, 1.0665E+06,
     +  1.0695E+06, 1.0724E+06, 1.0754E+06, 1.0784E+06, 1.0814E+06,
     + 1.0844E+06, 1.0874E+06, 1.0904E+06, 1.0934E+06, 1.0964E+06,
     +  1.0994E+06, 1.1025E+06, 1.1055E+06, 1.1086E+06, 1.1116E+06,
     + 1.1147E+06, 1.1177E+06, 1.1208E+06, 1.1239E+06, 1.1270E+06,
     +  1.1301E+06, 1.1332E+06, 1.1363E+06, 1.1394E+06, 1.1425E+06,
     + 1.1457E+06, 1.1488E+06, 1.1519E+06, 1.1551E+06, 1.1582E+06,
     +  1.1614E+06, 1.1646E+06, 1.1678E+06, 1.1709E+06, 1.1741E+06,
     + 1.1773E+06, 1.1805E+06, 1.1837E+06, 1.1870E+06, 1.1902E+06,
     +  1.1934E+06, 1.1967E+06, 1.1999E+06, 1.2032E+06, 1.2064E+06,
     + 1.2097E+06, 1.2130E+06, 1.2163E+06, 1.2196E+06, 1.2229E+06,
     +  1.2262E+06, 1.2295E+06, 1.2328E+06, 1.2361E+06, 1.2394E+06,
     + 1.2428E+06, 1.2461E+06, 1.2495E+06, 1.2529E+06, 1.2562E+06,
     +  1.2596E+06, 1.2630E+06, 1.2664E+06, 1.2698E+06, 1.2732E+06,
     + 1.2766E+06, 1.2800E+06, 1.2835E+06, 1.2869E+06, 1.2903E+06,
     +  1.2938E+06, 1.2972E+06, 1.3007E+06, 1.3042E+06, 1.3077E+06,
     + 1.3111E+06, 1.3146E+06, 1.3181E+06, 1.3217E+06, 1.3252E+06,
     +  1.3287E+06, 1.3322E+06, 1.3358E+06, 1.3393E+06, 1.3429E+06,
     + 1.3464E+06, 1.3500E+06, 1.3536E+06, 1.3572E+06, 1.3608E+06,
     +  1.3644E+06, 1.3680E+06, 1.3716E+06, 1.3752E+06, 1.3789E+06,
     + 1.3825E+06, 1.3862E+06, 1.3898E+06, 1.3935E+06, 1.3971E+06,
     +  1.4008E+06, 1.4045E+06, 1.4082E+06, 1.4119E+06, 1.4156E+06,
     + 1.4193E+06, 1.4231E+06, 1.4268E+06, 1.4305E+06, 1.4343E+06,
     +  1.4380E+06, 1.4418E+06, 1.4456E+06, 1.4494E+06, 1.4532E+06,
     + 1.4570E+06, 1.4608E+06, 1.4646E+06, 1.4684E+06, 1.4722E+06,
     +  1.4761E+06, 1.4799E+06, 1.4838E+06, 1.4876E+06, 1.4915E+06,
     + 1.4954E+06, 1.4993E+06, 1.5032E+06, 1.5071E+06, 1.5110E+06,
     +  1.5149E+06, 1.5188E+06, 1.5227E+06, 1.5267E+06, 1.5306E+06,
     + 1.5346E+06, 1.5386E+06, 1.5425E+06, 1.5465E+06, 1.5505E+06,
     +  1.5545E+06, 1.5585E+06, 1.5626E+06, 1.5666E+06, 1.5706E+06,
     + 1.5747E+06, 1.5787E+06, 1.5828E+06, 1.5868E+06, 1.5909E+06,
     +  1.5950E+06, 1.5991E+06, 1.6032E+06, 1.6073E+06, 1.6114E+06,
     + 1.6155E+06, 1.6197E+06, 1.6238E+06, 1.6280E+06, 1.6321E+06,
     +  1.6363E+06, 1.6405E+06, 1.6447E+06, 1.6489E+06, 1.6531E+06,
     + 1.6573E+06, 1.6615E+06, 1.6657E+06, 1.6700E+06, 1.6742E+06,
     +  1.6785E+06, 1.6828E+06, 1.6870E+06, 1.6913E+06, 1.6956E+06,
     + 1.6999E+06, 1.7042E+06, 1.7085E+06, 1.7129E+06, 1.7172E+06,
     +  1.7215E+06, 1.7259E+06, 1.7302E+06, 1.7346E+06, 1.7390E+06,
     + 1.7434E+06, 1.7478E+06, 1.7522E+06, 1.7566E+06, 1.7610E+06,
     +  1.7655E+06, 1.7699E+06, 1.7744E+06, 1.7788E+06, 1.7833E+06,
     + 1.7878E+06, 1.7922E+06, 1.7967E+06, 1.8012E+06, 1.8058E+06,
     +  1.8103E+06, 1.8148E+06, 1.8194E+06, 1.8239E+06, 1.8285E+06,
     + 1.8330E+06, 1.8376E+06, 1.8422E+06, 1.8468E+06, 1.8514E+06,
     +  1.8560E+06, 1.8606E+06, 1.8653E+06, 1.8699E+06, 1.8746E+06,
     + 1.8792E+06, 1.8839E+06, 1.8886E+06, 1.8933E+06, 1.8980E+06,
     +  1.9027E+06, 1.9074E+06, 1.9121E+06, 1.9169E+06, 1.9216E+06,
     + 1.9264E+06, 1.9311E+06, 1.9359E+06, 1.9407E+06, 1.9455E+06,
     +  1.9503E+06, 1.9551E+06, 1.9599E+06, 1.9648E+06, 1.9696E+06,
     + 1.9745E+06, 1.9793E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       624
      data (QofT( 2,J),J=1,2408)/ 3.8662E+00, 7.3687E+00, 1.0878E+01,
     +  1.4390E+01, 1.7902E+01, 2.1414E+01, 2.4927E+01, 2.8440E+01,
     + 3.1953E+01, 3.5467E+01, 3.8980E+01, 4.2494E+01, 4.6007E+01,
     +  4.9521E+01, 5.3035E+01, 5.6549E+01, 6.0063E+01, 6.3577E+01,
     + 6.7091E+01, 7.0605E+01, 7.7634E+01, 8.4663E+01, 9.1692E+01,
     +  9.8722E+01, 1.0575E+02, 1.1278E+02, 1.1981E+02, 1.2684E+02,
     + 1.3388E+02, 1.4091E+02, 1.4794E+02, 1.5497E+02, 1.6201E+02,
     +  1.6904E+02, 1.7607E+02, 1.8311E+02, 1.9014E+02, 1.9717E+02,
     + 2.0421E+02, 2.1125E+02, 2.1828E+02, 2.2532E+02, 2.3235E+02,
     +  2.3939E+02, 2.4644E+02, 2.5348E+02, 2.6052E+02, 2.6756E+02,
     + 2.7461E+02, 2.8166E+02, 2.8871E+02, 2.9577E+02, 3.0283E+02,
     +  3.0989E+02, 3.1697E+02, 3.2404E+02, 3.3112E+02, 3.3821E+02,
     + 3.4531E+02, 3.5241E+02, 3.5953E+02, 3.6665E+02, 3.7379E+02,
     +  3.8094E+02, 3.8809E+02, 3.9527E+02, 4.0246E+02, 4.0967E+02,
     + 4.1689E+02, 4.2413E+02, 4.3139E+02, 4.3867E+02, 4.4597E+02,
     +  4.5329E+02, 4.6064E+02, 4.6800E+02, 4.7540E+02, 4.8283E+02,
     + 4.9027E+02, 4.9775E+02, 5.0526E+02, 5.1280E+02, 5.2037E+02,
     +  5.2798E+02, 5.3562E+02, 5.4330E+02, 5.5100E+02, 5.5876E+02,
     + 5.6654E+02, 5.7437E+02, 5.8224E+02, 5.9015E+02, 5.9811E+02,
     +  6.0611E+02, 6.1415E+02, 6.2224E+02, 6.3038E+02, 6.3856E+02,
     + 6.4680E+02, 6.5508E+02, 6.6342E+02, 6.7182E+02, 6.8026E+02,
     +  6.8875E+02, 6.9731E+02, 7.0592E+02, 7.1458E+02, 7.2331E+02,
     + 7.3209E+02, 7.4094E+02, 7.4984E+02, 7.5881E+02, 7.6784E+02,
     +  7.7694E+02, 7.8610E+02, 7.9533E+02, 8.0462E+02, 8.1398E+02,
     + 8.2341E+02, 8.3291E+02, 8.4247E+02, 8.5211E+02, 8.6183E+02,
     +  8.7161E+02, 8.8146E+02, 8.9139E+02, 9.0141E+02, 9.1149E+02,
     + 9.2165E+02, 9.3189E+02, 9.4220E+02, 9.5260E+02, 9.6308E+02,
     +  9.7364E+02, 9.8428E+02, 9.9500E+02, 1.0058E+03, 1.0167E+03,
     + 1.0277E+03, 1.0387E+03, 1.0499E+03, 1.0611E+03, 1.0725E+03,
     +  1.0839E+03, 1.0954E+03, 1.1070E+03, 1.1187E+03, 1.1304E+03,
     + 1.1423E+03, 1.1543E+03, 1.1663E+03, 1.1785E+03, 1.1907E+03,
     +  1.2031E+03, 1.2155E+03, 1.2281E+03, 1.2407E+03, 1.2535E+03,
     + 1.2663E+03, 1.2793E+03, 1.2923E+03, 1.3055E+03, 1.3187E+03,
     +  1.3321E+03, 1.3456E+03, 1.3591E+03, 1.3728E+03, 1.3866E+03,
     + 1.4005E+03, 1.4145E+03, 1.4286E+03, 1.4429E+03, 1.4572E+03,
     +  1.4717E+03, 1.4862E+03, 1.5009E+03, 1.5157E+03, 1.5306E+03,
     + 1.5456E+03, 1.5608E+03, 1.5760E+03, 1.5914E+03, 1.6069E+03,
     +  1.6225E+03, 1.6383E+03, 1.6541E+03, 1.6701E+03, 1.6862E+03,
     + 1.7025E+03, 1.7188E+03, 1.7353E+03, 1.7519E+03, 1.7687E+03,
     +  1.7855E+03, 1.8025E+03, 1.8197E+03, 1.8369E+03, 1.8543E+03,
     + 1.8718E+03, 1.8895E+03, 1.9073E+03, 1.9252E+03, 1.9433E+03,
     +  1.9615E+03, 1.9798E+03, 1.9983E+03, 2.0169E+03, 2.0356E+03,
     + 2.0545E+03, 2.0735E+03, 2.0927E+03, 2.1120E+03, 2.1315E+03,
     +  2.1511E+03, 2.1709E+03, 2.1907E+03, 2.2108E+03, 2.2310E+03,
     + 2.2513E+03, 2.2718E+03, 2.2925E+03, 2.3133E+03, 2.3342E+03,
     +  2.3553E+03, 2.3766E+03, 2.3980E+03, 2.4195E+03, 2.4413E+03,
     + 2.4631E+03, 2.4852E+03, 2.5074E+03, 2.5297E+03, 2.5523E+03,
     +  2.5750E+03, 2.5978E+03, 2.6208E+03, 2.6440E+03, 2.6673E+03,
     + 2.6908E+03, 2.7145E+03, 2.7384E+03, 2.7624E+03, 2.7866E+03,
     +  2.8109E+03, 2.8355E+03, 2.8602E+03, 2.8851E+03, 2.9101E+03,
     + 2.9353E+03, 2.9607E+03, 2.9863E+03, 3.0121E+03, 3.0381E+03,
     +  3.0642E+03, 3.0905E+03, 3.1170E+03, 3.1437E+03, 3.1705E+03,
     + 3.1976E+03, 3.2248E+03, 3.2522E+03, 3.2798E+03, 3.3076E+03,
     +  3.3356E+03, 3.3638E+03, 3.3922E+03, 3.4208E+03, 3.4495E+03,
     + 3.4785E+03, 3.5077E+03, 3.5370E+03, 3.5666E+03, 3.5964E+03,
     +  3.6263E+03, 3.6565E+03, 3.6868E+03, 3.7174E+03, 3.7482E+03,
     + 3.7792E+03, 3.8104E+03, 3.8418E+03, 3.8734E+03, 3.9052E+03,
     +  3.9372E+03, 3.9695E+03, 4.0019E+03, 4.0346E+03, 4.0675E+03,
     + 4.1006E+03, 4.1339E+03, 4.1675E+03, 4.2013E+03, 4.2352E+03,
     +  4.2695E+03, 4.3039E+03, 4.3386E+03, 4.3735E+03, 4.4086E+03,
     + 4.4439E+03, 4.4795E+03, 4.5153E+03, 4.5513E+03, 4.5876E+03,
     +  4.6241E+03, 4.6609E+03, 4.6978E+03, 4.7350E+03, 4.7725E+03,
     + 4.8102E+03, 4.8481E+03, 4.8863E+03, 4.9247E+03, 4.9634E+03,
     +  5.0023E+03, 5.0415E+03, 5.0809E+03, 5.1205E+03, 5.1605E+03,
     + 5.2006E+03, 5.2410E+03, 5.2817E+03, 5.3226E+03, 5.3638E+03,
     +  5.4052E+03, 5.4470E+03, 5.4889E+03, 5.5311E+03, 5.5736E+03,
     + 5.6164E+03, 5.6594E+03, 5.7027E+03, 5.7462E+03, 5.7901E+03,
     +  5.8342E+03, 5.8785E+03, 5.9232E+03, 5.9681E+03, 6.0133E+03,
     + 6.0587E+03, 6.1045E+03, 6.1505E+03, 6.1968E+03, 6.2434E+03,
     +  6.2903E+03, 6.3374E+03, 6.3849E+03, 6.4326E+03, 6.4806E+03,
     + 6.5290E+03, 6.5776E+03, 6.6265E+03, 6.6757E+03, 6.7252E+03,
     +  6.7750E+03, 6.8251E+03, 6.8754E+03, 6.9261E+03, 6.9771E+03,
     + 7.0284E+03, 7.0800E+03, 7.1319E+03, 7.1841E+03, 7.2367E+03,
     +  7.2895E+03, 7.3427E+03, 7.3961E+03, 7.4499E+03, 7.5040E+03,
     + 7.5584E+03, 7.6132E+03, 7.6682E+03, 7.7236E+03, 7.7793E+03,
     +  7.8353E+03, 7.8917E+03, 7.9484E+03, 8.0054E+03, 8.0627E+03,
     + 8.1204E+03, 8.1785E+03, 8.2368E+03, 8.2955E+03, 8.3545E+03,
     +  8.4139E+03, 8.4736E+03, 8.5336E+03, 8.5940E+03, 8.6548E+03,
     + 8.7159E+03, 8.7773E+03, 8.8391E+03, 8.9013E+03, 8.9638E+03,
     +  9.0266E+03, 9.0899E+03, 9.1534E+03, 9.2173E+03, 9.2816E+03,
     + 9.3463E+03, 9.4113E+03, 9.4767E+03, 9.5425E+03, 9.6086E+03,
     +  9.6751E+03, 9.7420E+03, 9.8092E+03, 9.8769E+03, 9.9448E+03,
     + 1.0013E+04, 1.0082E+04, 1.0151E+04, 1.0221E+04, 1.0291E+04,
     +  1.0361E+04, 1.0432E+04, 1.0503E+04, 1.0574E+04, 1.0646E+04,
     + 1.0718E+04, 1.0791E+04, 1.0864E+04, 1.0937E+04, 1.1011E+04,
     +  1.1086E+04, 1.1160E+04, 1.1235E+04, 1.1311E+04, 1.1387E+04,
     + 1.1463E+04, 1.1540E+04, 1.1617E+04, 1.1694E+04, 1.1772E+04,
     +  1.1851E+04, 1.1930E+04, 1.2009E+04, 1.2089E+04, 1.2169E+04,
     + 1.2249E+04, 1.2330E+04, 1.2412E+04, 1.2493E+04, 1.2576E+04,
     +  1.2658E+04, 1.2742E+04, 1.2825E+04, 1.2909E+04, 1.2994E+04,
     + 1.3079E+04, 1.3164E+04, 1.3250E+04, 1.3336E+04, 1.3423E+04,
     +  1.3510E+04, 1.3598E+04, 1.3686E+04, 1.3775E+04, 1.3864E+04,
     + 1.3953E+04, 1.4043E+04, 1.4134E+04, 1.4224E+04, 1.4316E+04,
     +  1.4408E+04, 1.4500E+04, 1.4593E+04, 1.4686E+04, 1.4780E+04,
     + 1.4874E+04, 1.4969E+04, 1.5064E+04, 1.5160E+04, 1.5256E+04,
     +  1.5353E+04, 1.5450E+04, 1.5548E+04, 1.5646E+04, 1.5745E+04,
     + 1.5844E+04, 1.5944E+04, 1.6044E+04, 1.6145E+04, 1.6246E+04,
     +  1.6348E+04, 1.6451E+04, 1.6553E+04, 1.6657E+04, 1.6761E+04,
     + 1.6865E+04, 1.6970E+04, 1.7075E+04, 1.7181E+04, 1.7288E+04,
     +  1.7395E+04, 1.7503E+04, 1.7611E+04, 1.7719E+04, 1.7829E+04,
     + 1.7938E+04, 1.8049E+04, 1.8160E+04, 1.8271E+04, 1.8383E+04,
     +  1.8495E+04, 1.8608E+04, 1.8722E+04, 1.8836E+04, 1.8951E+04,
     + 1.9066E+04, 1.9182E+04, 1.9299E+04, 1.9416E+04, 1.9533E+04,
     +  1.9652E+04, 1.9770E+04, 1.9890E+04, 2.0010E+04, 2.0130E+04,
     + 2.0251E+04, 2.0373E+04, 2.0495E+04, 2.0618E+04, 2.0742E+04,
     +  2.0866E+04, 2.0990E+04, 2.1116E+04, 2.1242E+04, 2.1368E+04,
     + 2.1495E+04, 2.1623E+04, 2.1751E+04, 2.1880E+04, 2.2010E+04,
     +  2.2140E+04, 2.2271E+04, 2.2402E+04, 2.2534E+04, 2.2667E+04,
     + 2.2800E+04, 2.2934E+04, 2.3069E+04, 2.3204E+04, 2.3340E+04,
     +  2.3477E+04, 2.3614E+04, 2.3752E+04, 2.3890E+04, 2.4029E+04,
     + 2.4169E+04, 2.4309E+04, 2.4451E+04, 2.4592E+04, 2.4735E+04,
     +  2.4878E+04, 2.5022E+04, 2.5166E+04, 2.5311E+04, 2.5457E+04,
     + 2.5603E+04, 2.5751E+04, 2.5898E+04, 2.6047E+04, 2.6196E+04,
     +  2.6346E+04, 2.6497E+04, 2.6648E+04, 2.6800E+04, 2.6953E+04,
     + 2.7106E+04, 2.7260E+04, 2.7415E+04, 2.7571E+04, 2.7727E+04,
     +  2.7884E+04, 2.8041E+04, 2.8200E+04, 2.8359E+04, 2.8519E+04,
     + 2.8679E+04, 2.8841E+04, 2.9003E+04, 2.9166E+04, 2.9329E+04,
     +  2.9493E+04, 2.9658E+04, 2.9824E+04, 2.9991E+04, 3.0158E+04,
     + 3.0326E+04, 3.0495E+04, 3.0664E+04, 3.0834E+04, 3.1006E+04,
     +  3.1177E+04, 3.1350E+04, 3.1523E+04, 3.1698E+04, 3.1872E+04,
     + 3.2048E+04, 3.2225E+04, 3.2402E+04, 3.2580E+04, 3.2759E+04,
     +  3.2938E+04, 3.3119E+04, 3.3300E+04, 3.3482E+04, 3.3665E+04,
     + 3.3849E+04, 3.4033E+04, 3.4218E+04, 3.4405E+04, 3.4592E+04,
     +  3.4779E+04, 3.4968E+04, 3.5157E+04, 3.5347E+04, 3.5539E+04,
     + 3.5730E+04, 3.5923E+04, 3.6117E+04, 3.6311E+04, 3.6506E+04,
     +  3.6702E+04, 3.6899E+04, 3.7097E+04, 3.7296E+04, 3.7495E+04,
     + 3.7696E+04, 3.7897E+04, 3.8099E+04, 3.8302E+04, 3.8506E+04,
     +  3.8711E+04, 3.8917E+04, 3.9123E+04, 3.9330E+04, 3.9539E+04,
     + 3.9748E+04, 3.9958E+04, 4.0169E+04, 4.0381E+04, 4.0594E+04,
     +  4.0807E+04, 4.1022E+04, 4.1238E+04, 4.1454E+04, 4.1671E+04,
     + 4.1890E+04, 4.2109E+04, 4.2329E+04, 4.2550E+04, 4.2772E+04,
     +  4.2995E+04, 4.3219E+04, 4.3443E+04, 4.3669E+04, 4.3896E+04,
     + 4.4124E+04, 4.4352E+04, 4.4582E+04, 4.4812E+04, 4.5044E+04,
     +  4.5276E+04, 4.5510E+04, 4.5744E+04, 4.5979E+04, 4.6216E+04,
     + 4.6453E+04, 4.6691E+04, 4.6931E+04, 4.7171E+04, 4.7412E+04,
     +  4.7655E+04, 4.7898E+04, 4.8142E+04, 4.8387E+04, 4.8634E+04,
     + 4.8881E+04, 4.9129E+04, 4.9379E+04, 4.9629E+04, 4.9881E+04,
     +  5.0133E+04, 5.0387E+04, 5.0641E+04, 5.0897E+04, 5.1153E+04,
     + 5.1411E+04, 5.1669E+04, 5.1929E+04, 5.2190E+04, 5.2452E+04,
     +  5.2715E+04, 5.2979E+04, 5.3244E+04, 5.3510E+04, 5.3777E+04,
     + 5.4045E+04, 5.4315E+04, 5.4585E+04, 5.4857E+04, 5.5129E+04,
     +  5.5403E+04, 5.5678E+04, 5.5954E+04, 5.6231E+04, 5.6509E+04,
     + 5.6788E+04, 5.7068E+04, 5.7350E+04, 5.7632E+04, 5.7916E+04,
     +  5.8201E+04, 5.8487E+04, 5.8774E+04, 5.9062E+04, 5.9352E+04,
     + 5.9642E+04, 5.9934E+04, 6.0227E+04, 6.0521E+04, 6.0816E+04,
     +  6.1112E+04, 6.1410E+04, 6.1708E+04, 6.2008E+04, 6.2309E+04,
     + 6.2611E+04, 6.2915E+04, 6.3219E+04, 6.3525E+04, 6.3832E+04,
     +  6.4140E+04, 6.4450E+04, 6.4760E+04, 6.5072E+04, 6.5385E+04,
     + 6.5699E+04, 6.6014E+04, 6.6331E+04, 6.6649E+04, 6.6968E+04,
     +  6.7288E+04, 6.7610E+04, 6.7933E+04, 6.8257E+04, 6.8582E+04,
     + 6.8908E+04, 6.9236E+04, 6.9565E+04, 6.9896E+04, 7.0227E+04,
     +  7.0560E+04, 7.0894E+04, 7.1230E+04, 7.1566E+04, 7.1904E+04,
     + 7.2243E+04, 7.2584E+04, 7.2926E+04, 7.3269E+04, 7.3614E+04,
     +  7.3959E+04, 7.4306E+04, 7.4655E+04, 7.5004E+04, 7.5355E+04,
     + 7.5708E+04, 7.6061E+04, 7.6417E+04, 7.6773E+04, 7.7131E+04,
     +  7.7490E+04, 7.7850E+04, 7.8212E+04, 7.8575E+04, 7.8939E+04,
     + 7.9305E+04, 7.9672E+04, 8.0041E+04, 8.0411E+04, 8.0782E+04,
     +  8.1155E+04, 8.1529E+04, 8.1905E+04, 8.2282E+04, 8.2660E+04,
     + 8.3040E+04, 8.3421E+04, 8.3803E+04, 8.4187E+04, 8.4572E+04,
     +  8.4959E+04, 8.5347E+04, 8.5737E+04, 8.6128E+04, 8.6521E+04,
     + 8.6915E+04, 8.7310E+04, 8.7707E+04, 8.8105E+04, 8.8505E+04,
     +  8.8906E+04, 8.9309E+04, 8.9713E+04, 9.0119E+04, 9.0526E+04,
     + 9.0935E+04, 9.1345E+04, 9.1756E+04, 9.2169E+04, 9.2584E+04,
     +  9.3000E+04, 9.3418E+04, 9.3837E+04, 9.4258E+04, 9.4680E+04,
     + 9.5104E+04, 9.5529E+04, 9.5956E+04, 9.6384E+04, 9.6814E+04,
     +  9.7245E+04, 9.7678E+04, 9.8113E+04, 9.8549E+04, 9.8986E+04,
     + 9.9425E+04, 9.9866E+04, 1.0031E+05, 1.0075E+05, 1.0120E+05,
     +  1.0165E+05, 1.0209E+05, 1.0254E+05, 1.0300E+05, 1.0345E+05,
     + 1.0390E+05, 1.0436E+05, 1.0482E+05, 1.0528E+05, 1.0574E+05,
     +  1.0620E+05, 1.0667E+05, 1.0714E+05, 1.0760E+05, 1.0807E+05,
     + 1.0855E+05, 1.0902E+05, 1.0949E+05, 1.0997E+05, 1.1045E+05,
     +  1.1093E+05, 1.1141E+05, 1.1189E+05, 1.1238E+05, 1.1286E+05,
     + 1.1335E+05, 1.1384E+05, 1.1433E+05, 1.1483E+05, 1.1532E+05,
     +  1.1582E+05, 1.1632E+05, 1.1682E+05, 1.1732E+05, 1.1782E+05,
     + 1.1833E+05, 1.1884E+05, 1.1935E+05, 1.1986E+05, 1.2037E+05,
     +  1.2088E+05, 1.2140E+05, 1.2192E+05, 1.2244E+05, 1.2296E+05,
     + 1.2348E+05, 1.2401E+05, 1.2453E+05, 1.2506E+05, 1.2559E+05,
     +  1.2613E+05, 1.2666E+05, 1.2720E+05, 1.2773E+05, 1.2827E+05,
     + 1.2881E+05, 1.2936E+05, 1.2990E+05, 1.3045E+05, 1.3100E+05,
     +  1.3155E+05, 1.3210E+05, 1.3266E+05, 1.3321E+05, 1.3377E+05,
     + 1.3433E+05, 1.3489E+05, 1.3546E+05, 1.3602E+05, 1.3659E+05,
     +  1.3716E+05, 1.3773E+05, 1.3830E+05, 1.3888E+05, 1.3946E+05,
     + 1.4004E+05, 1.4062E+05, 1.4120E+05, 1.4178E+05, 1.4237E+05,
     +  1.4296E+05, 1.4355E+05, 1.4414E+05, 1.4474E+05, 1.4534E+05,
     + 1.4594E+05, 1.4654E+05, 1.4714E+05, 1.4774E+05, 1.4835E+05,
     +  1.4896E+05, 1.4957E+05, 1.5018E+05, 1.5080E+05, 1.5141E+05,
     + 1.5203E+05, 1.5265E+05, 1.5328E+05, 1.5390E+05, 1.5453E+05,
     +  1.5516E+05, 1.5579E+05, 1.5642E+05, 1.5706E+05, 1.5770E+05,
     + 1.5834E+05, 1.5898E+05, 1.5962E+05, 1.6027E+05, 1.6092E+05,
     +  1.6157E+05, 1.6222E+05, 1.6287E+05, 1.6353E+05, 1.6419E+05,
     + 1.6485E+05, 1.6551E+05, 1.6618E+05, 1.6684E+05, 1.6751E+05,
     +  1.6818E+05, 1.6886E+05, 1.6953E+05, 1.7021E+05, 1.7089E+05,
     + 1.7157E+05, 1.7226E+05, 1.7295E+05, 1.7363E+05, 1.7433E+05,
     +  1.7502E+05, 1.7572E+05, 1.7641E+05, 1.7711E+05, 1.7782E+05,
     + 1.7852E+05, 1.7923E+05, 1.7994E+05, 1.8065E+05, 1.8136E+05,
     +  1.8208E+05, 1.8280E+05, 1.8352E+05, 1.8424E+05, 1.8496E+05,
     + 1.8569E+05, 1.8642E+05, 1.8715E+05, 1.8789E+05, 1.8862E+05,
     +  1.8936E+05, 1.9010E+05, 1.9085E+05, 1.9159E+05, 1.9234E+05,
     + 1.9309E+05, 1.9385E+05, 1.9460E+05, 1.9536E+05, 1.9612E+05,
     +  1.9688E+05, 1.9765E+05, 1.9841E+05, 1.9918E+05, 1.9996E+05,
     + 2.0073E+05, 2.0151E+05, 2.0229E+05, 2.0307E+05, 2.0385E+05,
     +  2.0464E+05, 2.0543E+05, 2.0622E+05, 2.0702E+05, 2.0781E+05,
     + 2.0861E+05, 2.0941E+05, 2.1022E+05, 2.1102E+05, 2.1183E+05,
     +  2.1264E+05, 2.1346E+05, 2.1427E+05, 2.1509E+05, 2.1591E+05,
     + 2.1674E+05, 2.1757E+05, 2.1839E+05, 2.1923E+05, 2.2006E+05,
     +  2.2090E+05, 2.2174E+05, 2.2258E+05, 2.2342E+05, 2.2427E+05,
     + 2.2512E+05, 2.2597E+05, 2.2683E+05, 2.2768E+05, 2.2854E+05,
     +  2.2941E+05, 2.3027E+05, 2.3114E+05, 2.3201E+05, 2.3288E+05,
     + 2.3376E+05, 2.3464E+05, 2.3552E+05, 2.3640E+05, 2.3729E+05,
     +  2.3818E+05, 2.3907E+05, 2.3997E+05, 2.4086E+05, 2.4176E+05,
     + 2.4267E+05, 2.4357E+05, 2.4448E+05, 2.4539E+05, 2.4630E+05,
     +  2.4722E+05, 2.4814E+05, 2.4906E+05, 2.4999E+05, 2.5091E+05,
     + 2.5184E+05, 2.5278E+05, 2.5371E+05, 2.5465E+05, 2.5559E+05,
     +  2.5654E+05, 2.5748E+05, 2.5843E+05, 2.5939E+05, 2.6034E+05,
     + 2.6130E+05, 2.6226E+05, 2.6323E+05, 2.6419E+05, 2.6516E+05,
     +  2.6613E+05, 2.6711E+05, 2.6809E+05, 2.6907E+05, 2.7005E+05,
     + 2.7104E+05, 2.7203E+05, 2.7302E+05, 2.7402E+05, 2.7502E+05,
     +  2.7602E+05, 2.7702E+05, 2.7803E+05, 2.7904E+05, 2.8005E+05,
     + 2.8107E+05, 2.8209E+05, 2.8311E+05, 2.8414E+05, 2.8517E+05,
     +  2.8620E+05, 2.8723E+05, 2.8827E+05, 2.8931E+05, 2.9035E+05,
     + 2.9140E+05, 2.9245E+05, 2.9350E+05, 2.9456E+05, 2.9561E+05,
     +  2.9668E+05, 2.9774E+05, 2.9881E+05, 2.9988E+05, 3.0095E+05,
     + 3.0203E+05, 3.0311E+05, 3.0419E+05, 3.0528E+05, 3.0637E+05,
     +  3.0746E+05, 3.0856E+05, 3.0966E+05, 3.1076E+05, 3.1186E+05,
     + 3.1297E+05, 3.1408E+05, 3.1520E+05, 3.1632E+05, 3.1744E+05,
     +  3.1856E+05, 3.1969E+05, 3.2082E+05, 3.2196E+05, 3.2309E+05,
     + 3.2423E+05, 3.2538E+05, 3.2652E+05, 3.2767E+05, 3.2883E+05,
     +  3.2998E+05, 3.3114E+05, 3.3231E+05, 3.3347E+05, 3.3464E+05,
     + 3.3582E+05, 3.3699E+05, 3.3817E+05, 3.3936E+05, 3.4054E+05,
     +  3.4173E+05, 3.4293E+05, 3.4412E+05, 3.4532E+05, 3.4653E+05,
     + 3.4773E+05, 3.4894E+05, 3.5016E+05, 3.5137E+05, 3.5259E+05,
     +  3.5382E+05, 3.5504E+05, 3.5627E+05, 3.5751E+05, 3.5875E+05,
     + 3.5999E+05, 3.6123E+05, 3.6248E+05, 3.6373E+05, 3.6498E+05,
     +  3.6624E+05, 3.6750E+05, 3.6877E+05, 3.7004E+05, 3.7131E+05,
     + 3.7259E+05, 3.7387E+05, 3.7515E+05, 3.7644E+05, 3.7772E+05,
     +  3.7902E+05, 3.8032E+05, 3.8162E+05, 3.8292E+05, 3.8423E+05,
     + 3.8554E+05, 3.8685E+05, 3.8817E+05, 3.8950E+05, 3.9082E+05,
     +  3.9215E+05, 3.9348E+05, 3.9482E+05, 3.9616E+05, 3.9751E+05,
     + 3.9885E+05, 4.0020E+05, 4.0156E+05, 4.0292E+05, 4.0428E+05,
     +  4.0565E+05, 4.0702E+05, 4.0839E+05, 4.0977E+05, 4.1115E+05,
     + 4.1253E+05, 4.1392E+05, 4.1532E+05, 4.1671E+05, 4.1811E+05,
     +  4.1952E+05, 4.2092E+05, 4.2233E+05, 4.2375E+05, 4.2517E+05,
     + 4.2659E+05, 4.2802E+05, 4.2945E+05, 4.3088E+05, 4.3232E+05,
     +  4.3376E+05, 4.3521E+05, 4.3666E+05, 4.3811E+05, 4.3957E+05,
     + 4.4103E+05, 4.4250E+05, 4.4397E+05, 4.4544E+05, 4.4692E+05,
     +  4.4840E+05, 4.4988E+05, 4.5137E+05, 4.5287E+05, 4.5436E+05,
     + 4.5587E+05, 4.5737E+05, 4.5888E+05, 4.6039E+05, 4.6191E+05,
     +  4.6343E+05, 4.6496E+05, 4.6649E+05, 4.6802E+05, 4.6956E+05,
     + 4.7110E+05, 4.7264E+05, 4.7419E+05, 4.7575E+05, 4.7730E+05,
     +  4.7887E+05, 4.8043E+05, 4.8200E+05, 4.8358E+05, 4.8515E+05,
     + 4.8674E+05, 4.8832E+05, 4.8991E+05, 4.9151E+05, 4.9311E+05,
     +  4.9471E+05, 4.9632E+05, 4.9793E+05, 4.9955E+05, 5.0117E+05,
     + 5.0279E+05, 5.0442E+05, 5.0605E+05, 5.0769E+05, 5.0933E+05,
     +  5.1098E+05, 5.1263E+05, 5.1428E+05, 5.1594E+05, 5.1760E+05,
     + 5.1927E+05, 5.2094E+05, 5.2262E+05, 5.2430E+05, 5.2598E+05,
     +  5.2767E+05, 5.2936E+05, 5.3106E+05, 5.3276E+05, 5.3447E+05,
     + 5.3618E+05, 5.3790E+05, 5.3962E+05, 5.4134E+05, 5.4307E+05,
     +  5.4480E+05, 5.4654E+05, 5.4828E+05, 5.5003E+05, 5.5178E+05,
     + 5.5353E+05, 5.5529E+05, 5.5706E+05, 5.5883E+05, 5.6060E+05,
     +  5.6238E+05, 5.6416E+05, 5.6595E+05, 5.6774E+05, 5.6954E+05,
     + 5.7134E+05, 5.7314E+05, 5.7495E+05, 5.7677E+05, 5.7859E+05,
     +  5.8041E+05, 5.8224E+05, 5.8407E+05, 5.8591E+05, 5.8775E+05,
     + 5.8960E+05, 5.9145E+05, 5.9331E+05, 5.9517E+05, 5.9703E+05,
     +  5.9890E+05, 6.0078E+05, 6.0266E+05, 6.0454E+05, 6.0643E+05,
     + 6.0833E+05, 6.1022E+05, 6.1213E+05, 6.1404E+05, 6.1595E+05,
     +  6.1787E+05, 6.1979E+05, 6.2172E+05, 6.2365E+05, 6.2559E+05,
     + 6.2753E+05, 6.2948E+05, 6.3143E+05, 6.3339E+05, 6.3535E+05,
     +  6.3731E+05, 6.3929E+05, 6.4126E+05, 6.4324E+05, 6.4523E+05,
     + 6.4722E+05, 6.4922E+05, 6.5122E+05, 6.5322E+05, 6.5523E+05,
     +  6.5725E+05, 6.5927E+05, 6.6130E+05, 6.6333E+05, 6.6536E+05,
     + 6.6741E+05, 6.6945E+05, 6.7150E+05, 6.7356E+05, 6.7562E+05,
     +  6.7769E+05, 6.7976E+05, 6.8183E+05, 6.8392E+05, 6.8600E+05,
     + 6.8809E+05, 6.9019E+05, 6.9229E+05, 6.9440E+05, 6.9651E+05,
     +  6.9863E+05, 7.0075E+05, 7.0288E+05, 7.0501E+05, 7.0715E+05,
     + 7.0930E+05, 7.1145E+05, 7.1360E+05, 7.1576E+05, 7.1792E+05,
     +  7.2009E+05, 7.2227E+05, 7.2445E+05, 7.2664E+05, 7.2883E+05,
     + 7.3102E+05, 7.3322E+05, 7.3543E+05, 7.3764E+05, 7.3986E+05,
     +  7.4208E+05, 7.4431E+05, 7.4655E+05, 7.4879E+05, 7.5103E+05,
     + 7.5328E+05, 7.5554E+05, 7.5780E+05, 7.6006E+05, 7.6234E+05,
     +  7.6461E+05, 7.6690E+05, 7.6918E+05, 7.7148E+05, 7.7378E+05,
     + 7.7608E+05, 7.7839E+05, 7.8071E+05, 7.8303E+05, 7.8536E+05,
     +  7.8769E+05, 7.9003E+05, 7.9237E+05, 7.9472E+05, 7.9708E+05,
     + 7.9944E+05, 8.0180E+05, 8.0418E+05, 8.0655E+05, 8.0894E+05,
     +  8.1132E+05, 8.1372E+05, 8.1612E+05, 8.1853E+05, 8.2094E+05,
     + 8.2335E+05, 8.2578E+05, 8.2821E+05, 8.3064E+05, 8.3308E+05,
     +  8.3553E+05, 8.3798E+05, 8.4044E+05, 8.4290E+05, 8.4537E+05,
     + 8.4784E+05, 8.5033E+05, 8.5281E+05, 8.5531E+05, 8.5780E+05,
     +  8.6031E+05, 8.6282E+05, 8.6533E+05, 8.6786E+05, 8.7038E+05,
     + 8.7292E+05, 8.7546E+05, 8.7800E+05, 8.8056E+05, 8.8311E+05,
     +  8.8568E+05, 8.8825E+05, 8.9082E+05, 8.9340E+05, 8.9599E+05,
     + 8.9859E+05, 9.0119E+05, 9.0379E+05, 9.0640E+05, 9.0902E+05,
     +  9.1165E+05, 9.1428E+05, 9.1691E+05, 9.1955E+05, 9.2220E+05,
     + 9.2486E+05, 9.2752E+05, 9.3019E+05, 9.3286E+05, 9.3554E+05,
     +  9.3822E+05, 9.4092E+05, 9.4361E+05, 9.4632E+05, 9.4903E+05,
     + 9.5175E+05, 9.5447E+05, 9.5720E+05, 9.5993E+05, 9.6268E+05,
     +  9.6542E+05, 9.6818E+05, 9.7094E+05, 9.7371E+05, 9.7648E+05,
     + 9.7926E+05, 9.8205E+05, 9.8484E+05, 9.8764E+05, 9.9044E+05,
     +  9.9325E+05, 9.9607E+05, 9.9890E+05, 1.0017E+06, 1.0046E+06,
     + 1.0074E+06, 1.0103E+06, 1.0131E+06, 1.0160E+06, 1.0189E+06,
     +  1.0217E+06, 1.0246E+06, 1.0275E+06, 1.0304E+06, 1.0333E+06,
     + 1.0362E+06, 1.0391E+06, 1.0420E+06, 1.0450E+06, 1.0479E+06,
     +  1.0509E+06, 1.0538E+06, 1.0568E+06, 1.0597E+06, 1.0627E+06,
     + 1.0657E+06, 1.0686E+06, 1.0716E+06, 1.0746E+06, 1.0776E+06,
     +  1.0806E+06, 1.0837E+06, 1.0867E+06, 1.0897E+06, 1.0928E+06,
     + 1.0958E+06, 1.0988E+06, 1.1019E+06, 1.1050E+06, 1.1080E+06,
     +  1.1111E+06, 1.1142E+06, 1.1173E+06, 1.1204E+06, 1.1235E+06,
     + 1.1266E+06, 1.1297E+06, 1.1329E+06, 1.1360E+06, 1.1391E+06,
     +  1.1423E+06, 1.1454E+06, 1.1486E+06, 1.1518E+06, 1.1549E+06,
     + 1.1581E+06, 1.1613E+06, 1.1645E+06, 1.1677E+06, 1.1709E+06,
     +  1.1741E+06, 1.1774E+06, 1.1806E+06, 1.1838E+06, 1.1871E+06,
     + 1.1903E+06, 1.1936E+06, 1.1969E+06, 1.2001E+06, 1.2034E+06,
     +  1.2067E+06, 1.2100E+06, 1.2133E+06, 1.2166E+06, 1.2199E+06,
     + 1.2233E+06, 1.2266E+06, 1.2299E+06, 1.2333E+06, 1.2366E+06,
     +  1.2400E+06, 1.2434E+06, 1.2467E+06, 1.2501E+06, 1.2535E+06,
     + 1.2569E+06, 1.2603E+06, 1.2637E+06, 1.2672E+06, 1.2706E+06,
     +  1.2740E+06, 1.2775E+06, 1.2809E+06, 1.2844E+06, 1.2878E+06,
     + 1.2913E+06, 1.2948E+06, 1.2983E+06, 1.3018E+06, 1.3053E+06,
     +  1.3088E+06, 1.3123E+06, 1.3158E+06, 1.3194E+06, 1.3229E+06,
     + 1.3265E+06, 1.3300E+06, 1.3336E+06, 1.3372E+06, 1.3407E+06,
     +  1.3443E+06, 1.3479E+06, 1.3515E+06, 1.3551E+06, 1.3588E+06,
     + 1.3624E+06, 1.3660E+06, 1.3697E+06, 1.3733E+06, 1.3770E+06,
     +  1.3806E+06, 1.3843E+06, 1.3880E+06, 1.3917E+06, 1.3954E+06,
     + 1.3991E+06, 1.4028E+06, 1.4065E+06, 1.4103E+06, 1.4140E+06,
     +  1.4177E+06, 1.4215E+06, 1.4252E+06, 1.4290E+06, 1.4328E+06,
     + 1.4366E+06, 1.4404E+06, 1.4442E+06, 1.4480E+06, 1.4518E+06,
     +  1.4556E+06, 1.4595E+06, 1.4633E+06, 1.4672E+06, 1.4710E+06,
     + 1.4749E+06, 1.4788E+06, 1.4826E+06, 1.4865E+06, 1.4904E+06,
     +  1.4943E+06, 1.4982E+06, 1.5022E+06, 1.5061E+06, 1.5100E+06,
     + 1.5140E+06, 1.5180E+06, 1.5219E+06, 1.5259E+06, 1.5299E+06,
     +  1.5339E+06, 1.5379E+06, 1.5419E+06, 1.5459E+06, 1.5499E+06,
     + 1.5539E+06, 1.5580E+06, 1.5620E+06, 1.5661E+06, 1.5702E+06,
     +  1.5742E+06, 1.5783E+06, 1.5824E+06, 1.5865E+06, 1.5906E+06,
     + 1.5947E+06, 1.5989E+06, 1.6030E+06, 1.6071E+06, 1.6113E+06,
     +  1.6155E+06, 1.6196E+06, 1.6238E+06, 1.6280E+06, 1.6322E+06,
     + 1.6364E+06, 1.6406E+06, 1.6448E+06, 1.6490E+06, 1.6533E+06,
     +  1.6575E+06, 1.6618E+06, 1.6661E+06, 1.6703E+06, 1.6746E+06,
     + 1.6789E+06, 1.6832E+06, 1.6875E+06, 1.6918E+06, 1.6962E+06,
     +  1.7005E+06, 1.7048E+06, 1.7092E+06, 1.7136E+06, 1.7179E+06,
     + 1.7223E+06, 1.7267E+06, 1.7311E+06, 1.7355E+06, 1.7399E+06,
     +  1.7444E+06, 1.7488E+06, 1.7532E+06, 1.7577E+06, 1.7621E+06,
     + 1.7666E+06, 1.7711E+06, 1.7756E+06, 1.7801E+06, 1.7846E+06,
     +  1.7891E+06, 1.7936E+06, 1.7982E+06, 1.8027E+06, 1.8073E+06,
     + 1.8118E+06, 1.8164E+06, 1.8210E+06, 1.8256E+06, 1.8302E+06,
     +  1.8348E+06, 1.8394E+06, 1.8440E+06, 1.8487E+06, 1.8533E+06,
     + 1.8580E+06, 1.8626E+06, 1.8673E+06, 1.8720E+06, 1.8767E+06,
     +  1.8814E+06, 1.8861E+06, 1.8908E+06, 1.8956E+06, 1.9003E+06,
     + 1.9051E+06, 1.9098E+06, 1.9146E+06, 1.9194E+06, 1.9242E+06,
     +  1.9290E+06, 1.9338E+06, 1.9386E+06, 1.9434E+06, 1.9483E+06,
     + 1.9531E+06, 1.9580E+06, 1.9628E+06, 1.9677E+06, 1.9726E+06,
     +  1.9775E+06, 1.9824E+06, 1.9873E+06, 1.9923E+06, 1.9972E+06,
     + 2.0021E+06, 2.0071E+06, 2.0121E+06, 2.0170E+06, 2.0220E+06,
     +  2.0270E+06, 2.0320E+06, 2.0370E+06, 2.0421E+06, 2.0471E+06,
     + 2.0521E+06, 2.0572E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       632
      data (QofT( 3,J),J=1,2408)/ 7.5837E+00, 1.4439E+01, 2.1308E+01,
     +  2.8181E+01, 3.5056E+01, 4.1931E+01, 4.8807E+01, 5.5683E+01,
     + 6.2560E+01, 6.9437E+01, 7.6314E+01, 8.3192E+01, 9.0069E+01,
     +  9.6947E+01, 1.0382E+02, 1.1070E+02, 1.1758E+02, 1.2446E+02,
     + 1.3134E+02, 1.3822E+02, 1.5198E+02, 1.6573E+02, 1.7949E+02,
     +  1.9325E+02, 2.0701E+02, 2.2078E+02, 2.3454E+02, 2.4830E+02,
     + 2.6206E+02, 2.7583E+02, 2.8959E+02, 3.0336E+02, 3.1712E+02,
     +  3.3089E+02, 3.4466E+02, 3.5843E+02, 3.7219E+02, 3.8596E+02,
     + 3.9974E+02, 4.1351E+02, 4.2728E+02, 4.4105E+02, 4.5483E+02,
     +  4.6861E+02, 4.8239E+02, 4.9618E+02, 5.0997E+02, 5.2376E+02,
     + 5.3756E+02, 5.5137E+02, 5.6518E+02, 5.7900E+02, 5.9283E+02,
     +  6.0667E+02, 6.2052E+02, 6.3439E+02, 6.4826E+02, 6.6216E+02,
     + 6.7608E+02, 6.9001E+02, 7.0396E+02, 7.1794E+02, 7.3194E+02,
     +  7.4597E+02, 7.6004E+02, 7.7412E+02, 7.8825E+02, 8.0241E+02,
     + 8.1660E+02, 8.3083E+02, 8.4511E+02, 8.5943E+02, 8.7380E+02,
     +  8.8821E+02, 9.0267E+02, 9.1720E+02, 9.3177E+02, 9.4640E+02,
     + 9.6108E+02, 9.7583E+02, 9.9066E+02, 1.0055E+03, 1.0205E+03,
     +  1.0355E+03, 1.0506E+03, 1.0658E+03, 1.0810E+03, 1.0963E+03,
     + 1.1118E+03, 1.1273E+03, 1.1428E+03, 1.1585E+03, 1.1743E+03,
     +  1.1901E+03, 1.2061E+03, 1.2221E+03, 1.2383E+03, 1.2545E+03,
     + 1.2709E+03, 1.2873E+03, 1.3039E+03, 1.3205E+03, 1.3373E+03,
     +  1.3542E+03, 1.3712E+03, 1.3884E+03, 1.4056E+03, 1.4230E+03,
     + 1.4405E+03, 1.4581E+03, 1.4758E+03, 1.4937E+03, 1.5117E+03,
     +  1.5298E+03, 1.5481E+03, 1.5665E+03, 1.5850E+03, 1.6037E+03,
     + 1.6225E+03, 1.6415E+03, 1.6606E+03, 1.6799E+03, 1.6992E+03,
     +  1.7188E+03, 1.7385E+03, 1.7583E+03, 1.7784E+03, 1.7985E+03,
     + 1.8189E+03, 1.8393E+03, 1.8600E+03, 1.8808E+03, 1.9018E+03,
     +  1.9229E+03, 1.9442E+03, 1.9657E+03, 1.9873E+03, 2.0091E+03,
     + 2.0311E+03, 2.0533E+03, 2.0756E+03, 2.0982E+03, 2.1208E+03,
     +  2.1437E+03, 2.1668E+03, 2.1901E+03, 2.2135E+03, 2.2371E+03,
     + 2.2610E+03, 2.2850E+03, 2.3092E+03, 2.3336E+03, 2.3582E+03,
     +  2.3830E+03, 2.4080E+03, 2.4331E+03, 2.4585E+03, 2.4841E+03,
     + 2.5100E+03, 2.5360E+03, 2.5622E+03, 2.5886E+03, 2.6153E+03,
     +  2.6421E+03, 2.6692E+03, 2.6965E+03, 2.7240E+03, 2.7517E+03,
     + 2.7797E+03, 2.8079E+03, 2.8363E+03, 2.8649E+03, 2.8937E+03,
     +  2.9228E+03, 2.9521E+03, 2.9817E+03, 3.0114E+03, 3.0414E+03,
     + 3.0717E+03, 3.1022E+03, 3.1329E+03, 3.1639E+03, 3.1951E+03,
     +  3.2266E+03, 3.2583E+03, 3.2902E+03, 3.3224E+03, 3.3549E+03,
     + 3.3876E+03, 3.4205E+03, 3.4537E+03, 3.4872E+03, 3.5210E+03,
     +  3.5550E+03, 3.5892E+03, 3.6238E+03, 3.6585E+03, 3.6936E+03,
     + 3.7290E+03, 3.7645E+03, 3.8004E+03, 3.8366E+03, 3.8730E+03,
     +  3.9097E+03, 3.9467E+03, 3.9839E+03, 4.0215E+03, 4.0593E+03,
     + 4.0974E+03, 4.1358E+03, 4.1746E+03, 4.2135E+03, 4.2528E+03,
     +  4.2924E+03, 4.3323E+03, 4.3724E+03, 4.4129E+03, 4.4537E+03,
     + 4.4948E+03, 4.5362E+03, 4.5779E+03, 4.6198E+03, 4.6622E+03,
     +  4.7048E+03, 4.7477E+03, 4.7910E+03, 4.8346E+03, 4.8785E+03,
     + 4.9227E+03, 4.9672E+03, 5.0121E+03, 5.0573E+03, 5.1028E+03,
     +  5.1487E+03, 5.1949E+03, 5.2414E+03, 5.2883E+03, 5.3355E+03,
     + 5.3830E+03, 5.4309E+03, 5.4791E+03, 5.5277E+03, 5.5766E+03,
     +  5.6259E+03, 5.6755E+03, 5.7255E+03, 5.7758E+03, 5.8265E+03,
     + 5.8776E+03, 5.9290E+03, 5.9808E+03, 6.0330E+03, 6.0855E+03,
     +  6.1384E+03, 6.1916E+03, 6.2453E+03, 6.2993E+03, 6.3537E+03,
     + 6.4085E+03, 6.4636E+03, 6.5191E+03, 6.5751E+03, 6.6314E+03,
     +  6.6881E+03, 6.7451E+03, 6.8026E+03, 6.8605E+03, 6.9188E+03,
     + 6.9775E+03, 7.0366E+03, 7.0961E+03, 7.1560E+03, 7.2163E+03,
     +  7.2770E+03, 7.3382E+03, 7.3997E+03, 7.4617E+03, 7.5241E+03,
     + 7.5869E+03, 7.6502E+03, 7.7138E+03, 7.7779E+03, 7.8425E+03,
     +  7.9074E+03, 7.9728E+03, 8.0386E+03, 8.1049E+03, 8.1716E+03,
     + 8.2388E+03, 8.3064E+03, 8.3745E+03, 8.4430E+03, 8.5120E+03,
     +  8.5814E+03, 8.6513E+03, 8.7216E+03, 8.7924E+03, 8.8637E+03,
     + 8.9354E+03, 9.0077E+03, 9.0804E+03, 9.1535E+03, 9.2272E+03,
     +  9.3013E+03, 9.3759E+03, 9.4510E+03, 9.5265E+03, 9.6026E+03,
     + 9.6791E+03, 9.7562E+03, 9.8338E+03, 9.9118E+03, 9.9904E+03,
     +  1.0069E+04, 1.0149E+04, 1.0229E+04, 1.0310E+04, 1.0391E+04,
     + 1.0472E+04, 1.0554E+04, 1.0637E+04, 1.0720E+04, 1.0804E+04,
     +  1.0888E+04, 1.0973E+04, 1.1058E+04, 1.1144E+04, 1.1231E+04,
     + 1.1318E+04, 1.1405E+04, 1.1493E+04, 1.1582E+04, 1.1671E+04,
     +  1.1760E+04, 1.1851E+04, 1.1941E+04, 1.2033E+04, 1.2125E+04,
     + 1.2217E+04, 1.2310E+04, 1.2404E+04, 1.2498E+04, 1.2593E+04,
     +  1.2688E+04, 1.2784E+04, 1.2881E+04, 1.2978E+04, 1.3076E+04,
     + 1.3174E+04, 1.3273E+04, 1.3373E+04, 1.3473E+04, 1.3573E+04,
     +  1.3675E+04, 1.3777E+04, 1.3879E+04, 1.3983E+04, 1.4086E+04,
     + 1.4191E+04, 1.4296E+04, 1.4402E+04, 1.4508E+04, 1.4615E+04,
     +  1.4723E+04, 1.4831E+04, 1.4940E+04, 1.5050E+04, 1.5160E+04,
     + 1.5271E+04, 1.5382E+04, 1.5494E+04, 1.5607E+04, 1.5721E+04,
     +  1.5835E+04, 1.5950E+04, 1.6065E+04, 1.6182E+04, 1.6298E+04,
     + 1.6416E+04, 1.6534E+04, 1.6653E+04, 1.6773E+04, 1.6893E+04,
     +  1.7014E+04, 1.7136E+04, 1.7259E+04, 1.7382E+04, 1.7506E+04,
     + 1.7630E+04, 1.7756E+04, 1.7882E+04, 1.8008E+04, 1.8136E+04,
     +  1.8264E+04, 1.8393E+04, 1.8523E+04, 1.8653E+04, 1.8784E+04,
     + 1.8916E+04, 1.9049E+04, 1.9182E+04, 1.9317E+04, 1.9452E+04,
     +  1.9587E+04, 1.9724E+04, 1.9861E+04, 1.9999E+04, 2.0138E+04,
     + 2.0277E+04, 2.0418E+04, 2.0559E+04, 2.0701E+04, 2.0844E+04,
     +  2.0987E+04, 2.1132E+04, 2.1277E+04, 2.1423E+04, 2.1570E+04,
     + 2.1717E+04, 2.1866E+04, 2.2015E+04, 2.2165E+04, 2.2316E+04,
     +  2.2467E+04, 2.2620E+04, 2.2773E+04, 2.2928E+04, 2.3083E+04,
     + 2.3239E+04, 2.3395E+04, 2.3553E+04, 2.3712E+04, 2.3871E+04,
     +  2.4031E+04, 2.4192E+04, 2.4354E+04, 2.4517E+04, 2.4681E+04,
     + 2.4846E+04, 2.5011E+04, 2.5177E+04, 2.5345E+04, 2.5513E+04,
     +  2.5682E+04, 2.5852E+04, 2.6023E+04, 2.6195E+04, 2.6368E+04,
     + 2.6541E+04, 2.6716E+04, 2.6892E+04, 2.7068E+04, 2.7246E+04,
     +  2.7424E+04, 2.7603E+04, 2.7783E+04, 2.7965E+04, 2.8147E+04,
     + 2.8330E+04, 2.8514E+04, 2.8699E+04, 2.8885E+04, 2.9072E+04,
     +  2.9260E+04, 2.9449E+04, 2.9639E+04, 2.9830E+04, 3.0022E+04,
     + 3.0215E+04, 3.0409E+04, 3.0604E+04, 3.0800E+04, 3.0997E+04,
     +  3.1195E+04, 3.1394E+04, 3.1594E+04, 3.1796E+04, 3.1998E+04,
     + 3.2201E+04, 3.2405E+04, 3.2610E+04, 3.2817E+04, 3.3024E+04,
     +  3.3233E+04, 3.3442E+04, 3.3653E+04, 3.3865E+04, 3.4077E+04,
     + 3.4291E+04, 3.4506E+04, 3.4722E+04, 3.4939E+04, 3.5157E+04,
     +  3.5377E+04, 3.5597E+04, 3.5819E+04, 3.6041E+04, 3.6265E+04,
     + 3.6490E+04, 3.6716E+04, 3.6943E+04, 3.7172E+04, 3.7401E+04,
     +  3.7632E+04, 3.7863E+04, 3.8096E+04, 3.8330E+04, 3.8566E+04,
     + 3.8802E+04, 3.9040E+04, 3.9278E+04, 3.9518E+04, 3.9759E+04,
     +  4.0002E+04, 4.0245E+04, 4.0490E+04, 4.0736E+04, 4.0983E+04,
     + 4.1231E+04, 4.1481E+04, 4.1732E+04, 4.1984E+04, 4.2237E+04,
     +  4.2491E+04, 4.2747E+04, 4.3004E+04, 4.3262E+04, 4.3522E+04,
     + 4.3783E+04, 4.4045E+04, 4.4308E+04, 4.4572E+04, 4.4838E+04,
     +  4.5105E+04, 4.5374E+04, 4.5643E+04, 4.5914E+04, 4.6187E+04,
     + 4.6460E+04, 4.6735E+04, 4.7011E+04, 4.7289E+04, 4.7568E+04,
     +  4.7848E+04, 4.8130E+04, 4.8412E+04, 4.8697E+04, 4.8982E+04,
     + 4.9269E+04, 4.9557E+04, 4.9847E+04, 5.0138E+04, 5.0431E+04,
     +  5.0724E+04, 5.1019E+04, 5.1316E+04, 5.1614E+04, 5.1913E+04,
     + 5.2214E+04, 5.2516E+04, 5.2820E+04, 5.3125E+04, 5.3431E+04,
     +  5.3739E+04, 5.4048E+04, 5.4359E+04, 5.4671E+04, 5.4985E+04,
     + 5.5300E+04, 5.5617E+04, 5.5935E+04, 5.6254E+04, 5.6575E+04,
     +  5.6898E+04, 5.7222E+04, 5.7547E+04, 5.7874E+04, 5.8202E+04,
     + 5.8532E+04, 5.8864E+04, 5.9197E+04, 5.9531E+04, 5.9867E+04,
     +  6.0205E+04, 6.0544E+04, 6.0885E+04, 6.1227E+04, 6.1571E+04,
     + 6.1916E+04, 6.2263E+04, 6.2611E+04, 6.2961E+04, 6.3313E+04,
     +  6.3666E+04, 6.4021E+04, 6.4377E+04, 6.4735E+04, 6.5095E+04,
     + 6.5456E+04, 6.5819E+04, 6.6183E+04, 6.6550E+04, 6.6917E+04,
     +  6.7287E+04, 6.7658E+04, 6.8030E+04, 6.8405E+04, 6.8781E+04,
     + 6.9159E+04, 6.9538E+04, 6.9919E+04, 7.0302E+04, 7.0686E+04,
     +  7.1072E+04, 7.1460E+04, 7.1850E+04, 7.2241E+04, 7.2634E+04,
     + 7.3029E+04, 7.3425E+04, 7.3824E+04, 7.4224E+04, 7.4625E+04,
     +  7.5029E+04, 7.5434E+04, 7.5841E+04, 7.6250E+04, 7.6660E+04,
     + 7.7073E+04, 7.7487E+04, 7.7903E+04, 7.8321E+04, 7.8740E+04,
     +  7.9161E+04, 7.9585E+04, 8.0010E+04, 8.0437E+04, 8.0865E+04,
     + 8.1296E+04, 8.1728E+04, 8.2163E+04, 8.2599E+04, 8.3037E+04,
     +  8.3477E+04, 8.3918E+04, 8.4362E+04, 8.4808E+04, 8.5255E+04,
     + 8.5704E+04, 8.6156E+04, 8.6609E+04, 8.7064E+04, 8.7521E+04,
     +  8.7980E+04, 8.8441E+04, 8.8904E+04, 8.9369E+04, 8.9836E+04,
     + 9.0305E+04, 9.0775E+04, 9.1248E+04, 9.1723E+04, 9.2200E+04,
     +  9.2678E+04, 9.3159E+04, 9.3642E+04, 9.4127E+04, 9.4614E+04,
     + 9.5102E+04, 9.5593E+04, 9.6086E+04, 9.6581E+04, 9.7078E+04,
     +  9.7578E+04, 9.8079E+04, 9.8582E+04, 9.9087E+04, 9.9595E+04,
     + 1.0010E+05, 1.0062E+05, 1.0113E+05, 1.0165E+05, 1.0216E+05,
     +  1.0268E+05, 1.0321E+05, 1.0373E+05, 1.0426E+05, 1.0479E+05,
     + 1.0532E+05, 1.0585E+05, 1.0639E+05, 1.0692E+05, 1.0746E+05,
     +  1.0801E+05, 1.0855E+05, 1.0910E+05, 1.0964E+05, 1.1019E+05,
     + 1.1075E+05, 1.1130E+05, 1.1186E+05, 1.1242E+05, 1.1298E+05,
     +  1.1355E+05, 1.1411E+05, 1.1468E+05, 1.1525E+05, 1.1583E+05,
     + 1.1640E+05, 1.1698E+05, 1.1756E+05, 1.1814E+05, 1.1873E+05,
     +  1.1932E+05, 1.1991E+05, 1.2050E+05, 1.2109E+05, 1.2169E+05,
     + 1.2229E+05, 1.2289E+05, 1.2349E+05, 1.2410E+05, 1.2471E+05,
     +  1.2532E+05, 1.2593E+05, 1.2655E+05, 1.2717E+05, 1.2779E+05,
     + 1.2841E+05, 1.2904E+05, 1.2967E+05, 1.3030E+05, 1.3093E+05,
     +  1.3157E+05, 1.3220E+05, 1.3285E+05, 1.3349E+05, 1.3413E+05,
     + 1.3478E+05, 1.3543E+05, 1.3609E+05, 1.3674E+05, 1.3740E+05,
     +  1.3806E+05, 1.3873E+05, 1.3939E+05, 1.4006E+05, 1.4073E+05,
     + 1.4141E+05, 1.4208E+05, 1.4276E+05, 1.4344E+05, 1.4413E+05,
     +  1.4481E+05, 1.4550E+05, 1.4620E+05, 1.4689E+05, 1.4759E+05,
     + 1.4829E+05, 1.4899E+05, 1.4970E+05, 1.5041E+05, 1.5112E+05,
     +  1.5183E+05, 1.5255E+05, 1.5327E+05, 1.5399E+05, 1.5471E+05,
     + 1.5544E+05, 1.5617E+05, 1.5690E+05, 1.5764E+05, 1.5838E+05,
     +  1.5912E+05, 1.5986E+05, 1.6061E+05, 1.6136E+05, 1.6211E+05,
     + 1.6287E+05, 1.6363E+05, 1.6439E+05, 1.6515E+05, 1.6592E+05,
     +  1.6669E+05, 1.6746E+05, 1.6824E+05, 1.6902E+05, 1.6980E+05,
     + 1.7058E+05, 1.7137E+05, 1.7216E+05, 1.7295E+05, 1.7375E+05,
     +  1.7455E+05, 1.7535E+05, 1.7615E+05, 1.7696E+05, 1.7777E+05,
     + 1.7859E+05, 1.7940E+05, 1.8022E+05, 1.8105E+05, 1.8187E+05,
     +  1.8270E+05, 1.8353E+05, 1.8437E+05, 1.8521E+05, 1.8605E+05,
     + 1.8689E+05, 1.8774E+05, 1.8859E+05, 1.8944E+05, 1.9030E+05,
     +  1.9116E+05, 1.9202E+05, 1.9289E+05, 1.9376E+05, 1.9463E+05,
     + 1.9551E+05, 1.9639E+05, 1.9727E+05, 1.9815E+05, 1.9904E+05,
     +  1.9993E+05, 2.0083E+05, 2.0173E+05, 2.0263E+05, 2.0353E+05,
     + 2.0444E+05, 2.0535E+05, 2.0627E+05, 2.0719E+05, 2.0811E+05,
     +  2.0903E+05, 2.0996E+05, 2.1089E+05, 2.1182E+05, 2.1276E+05,
     + 2.1370E+05, 2.1465E+05, 2.1559E+05, 2.1655E+05, 2.1750E+05,
     +  2.1846E+05, 2.1942E+05, 2.2038E+05, 2.2135E+05, 2.2232E+05,
     + 2.2330E+05, 2.2428E+05, 2.2526E+05, 2.2624E+05, 2.2723E+05,
     +  2.2822E+05, 2.2922E+05, 2.3022E+05, 2.3122E+05, 2.3223E+05,
     + 2.3324E+05, 2.3425E+05, 2.3527E+05, 2.3629E+05, 2.3731E+05,
     +  2.3834E+05, 2.3937E+05, 2.4041E+05, 2.4145E+05, 2.4249E+05,
     + 2.4353E+05, 2.4458E+05, 2.4564E+05, 2.4669E+05, 2.4775E+05,
     +  2.4882E+05, 2.4989E+05, 2.5096E+05, 2.5203E+05, 2.5311E+05,
     + 2.5419E+05, 2.5528E+05, 2.5637E+05, 2.5746E+05, 2.5856E+05,
     +  2.5966E+05, 2.6077E+05, 2.6188E+05, 2.6299E+05, 2.6411E+05,
     + 2.6523E+05, 2.6635E+05, 2.6748E+05, 2.6861E+05, 2.6975E+05,
     +  2.7089E+05, 2.7203E+05, 2.7318E+05, 2.7433E+05, 2.7548E+05,
     + 2.7664E+05, 2.7781E+05, 2.7897E+05, 2.8014E+05, 2.8132E+05,
     +  2.8250E+05, 2.8368E+05, 2.8487E+05, 2.8606E+05, 2.8725E+05,
     + 2.8845E+05, 2.8966E+05, 2.9086E+05, 2.9207E+05, 2.9329E+05,
     +  2.9451E+05, 2.9573E+05, 2.9696E+05, 2.9819E+05, 2.9943E+05,
     + 3.0067E+05, 3.0191E+05, 3.0316E+05, 3.0441E+05, 3.0567E+05,
     +  3.0693E+05, 3.0819E+05, 3.0946E+05, 3.1073E+05, 3.1201E+05,
     + 3.1329E+05, 3.1458E+05, 3.1587E+05, 3.1716E+05, 3.1846E+05,
     +  3.1977E+05, 3.2107E+05, 3.2238E+05, 3.2370E+05, 3.2502E+05,
     + 3.2635E+05, 3.2767E+05, 3.2901E+05, 3.3035E+05, 3.3169E+05,
     +  3.3303E+05, 3.3439E+05, 3.3574E+05, 3.3710E+05, 3.3846E+05,
     + 3.3983E+05, 3.4121E+05, 3.4258E+05, 3.4397E+05, 3.4535E+05,
     +  3.4674E+05, 3.4814E+05, 3.4954E+05, 3.5094E+05, 3.5235E+05,
     + 3.5377E+05, 3.5519E+05, 3.5661E+05, 3.5804E+05, 3.5947E+05,
     +  3.6090E+05, 3.6235E+05, 3.6379E+05, 3.6524E+05, 3.6670E+05,
     + 3.6816E+05, 3.6962E+05, 3.7109E+05, 3.7257E+05, 3.7404E+05,
     +  3.7553E+05, 3.7702E+05, 3.7851E+05, 3.8001E+05, 3.8151E+05,
     + 3.8302E+05, 3.8453E+05, 3.8605E+05, 3.8757E+05, 3.8910E+05,
     +  3.9063E+05, 3.9216E+05, 3.9370E+05, 3.9525E+05, 3.9680E+05,
     + 3.9836E+05, 3.9992E+05, 4.0148E+05, 4.0305E+05, 4.0463E+05,
     +  4.0621E+05, 4.0780E+05, 4.0939E+05, 4.1098E+05, 4.1258E+05,
     + 4.1419E+05, 4.1580E+05, 4.1742E+05, 4.1904E+05, 4.2066E+05,
     +  4.2229E+05, 4.2393E+05, 4.2557E+05, 4.2722E+05, 4.2887E+05,
     + 4.3052E+05, 4.3219E+05, 4.3385E+05, 4.3553E+05, 4.3720E+05,
     +  4.3889E+05, 4.4057E+05, 4.4227E+05, 4.4396E+05, 4.4567E+05,
     + 4.4738E+05, 4.4909E+05, 4.5081E+05, 4.5253E+05, 4.5426E+05,
     +  4.5600E+05, 4.5774E+05, 4.5949E+05, 4.6124E+05, 4.6299E+05,
     + 4.6476E+05, 4.6652E+05, 4.6830E+05, 4.7007E+05, 4.7186E+05,
     +  4.7365E+05, 4.7544E+05, 4.7724E+05, 4.7905E+05, 4.8086E+05,
     + 4.8268E+05, 4.8450E+05, 4.8633E+05, 4.8816E+05, 4.9000E+05,
     +  4.9184E+05, 4.9369E+05, 4.9555E+05, 4.9741E+05, 4.9928E+05,
     + 5.0115E+05, 5.0303E+05, 5.0491E+05, 5.0680E+05, 5.0870E+05,
     +  5.1060E+05, 5.1250E+05, 5.1442E+05, 5.1633E+05, 5.1826E+05,
     + 5.2019E+05, 5.2212E+05, 5.2406E+05, 5.2601E+05, 5.2796E+05,
     +  5.2992E+05, 5.3189E+05, 5.3386E+05, 5.3583E+05, 5.3782E+05,
     + 5.3981E+05, 5.4180E+05, 5.4380E+05, 5.4581E+05, 5.4782E+05,
     +  5.4984E+05, 5.5186E+05, 5.5389E+05, 5.5593E+05, 5.5797E+05,
     + 5.6002E+05, 5.6207E+05, 5.6413E+05, 5.6620E+05, 5.6827E+05,
     +  5.7035E+05, 5.7243E+05, 5.7452E+05, 5.7662E+05, 5.7872E+05,
     + 5.8083E+05, 5.8295E+05, 5.8507E+05, 5.8720E+05, 5.8933E+05,
     +  5.9147E+05, 5.9362E+05, 5.9577E+05, 5.9793E+05, 6.0010E+05,
     + 6.0227E+05, 6.0445E+05, 6.0663E+05, 6.0882E+05, 6.1102E+05,
     +  6.1322E+05, 6.1543E+05, 6.1765E+05, 6.1987E+05, 6.2210E+05,
     + 6.2434E+05, 6.2658E+05, 6.2883E+05, 6.3108E+05, 6.3334E+05,
     +  6.3561E+05, 6.3789E+05, 6.4017E+05, 6.4246E+05, 6.4475E+05,
     + 6.4705E+05, 6.4936E+05, 6.5168E+05, 6.5400E+05, 6.5632E+05,
     +  6.5866E+05, 6.6100E+05, 6.6335E+05, 6.6570E+05, 6.6806E+05,
     + 6.7043E+05, 6.7281E+05, 6.7519E+05, 6.7758E+05, 6.7997E+05,
     +  6.8237E+05, 6.8478E+05, 6.8720E+05, 6.8962E+05, 6.9205E+05,
     + 6.9449E+05, 6.9693E+05, 6.9938E+05, 7.0184E+05, 7.0430E+05,
     +  7.0677E+05, 7.0925E+05, 7.1173E+05, 7.1423E+05, 7.1672E+05,
     + 7.1923E+05, 7.2174E+05, 7.2426E+05, 7.2679E+05, 7.2933E+05,
     +  7.3187E+05, 7.3442E+05, 7.3697E+05, 7.3953E+05, 7.4210E+05,
     + 7.4468E+05, 7.4727E+05, 7.4986E+05, 7.5246E+05, 7.5506E+05,
     +  7.5768E+05, 7.6030E+05, 7.6293E+05, 7.6556E+05, 7.6820E+05,
     + 7.7085E+05, 7.7351E+05, 7.7618E+05, 7.7885E+05, 7.8153E+05,
     +  7.8422E+05, 7.8691E+05, 7.8961E+05, 7.9232E+05, 7.9504E+05,
     + 7.9776E+05, 8.0049E+05, 8.0323E+05, 8.0598E+05, 8.0874E+05,
     +  8.1150E+05, 8.1427E+05, 8.1705E+05, 8.1983E+05, 8.2262E+05,
     + 8.2542E+05, 8.2823E+05, 8.3105E+05, 8.3387E+05, 8.3670E+05,
     +  8.3954E+05, 8.4239E+05, 8.4524E+05, 8.4810E+05, 8.5097E+05,
     + 8.5385E+05, 8.5674E+05, 8.5963E+05, 8.6253E+05, 8.6544E+05,
     +  8.6836E+05, 8.7128E+05, 8.7422E+05, 8.7716E+05, 8.8011E+05,
     + 8.8306E+05, 8.8603E+05, 8.8900E+05, 8.9198E+05, 8.9497E+05,
     +  8.9797E+05, 9.0097E+05, 9.0399E+05, 9.0701E+05, 9.1004E+05,
     + 9.1308E+05, 9.1612E+05, 9.1918E+05, 9.2224E+05, 9.2531E+05,
     +  9.2839E+05, 9.3147E+05, 9.3457E+05, 9.3767E+05, 9.4079E+05,
     + 9.4391E+05, 9.4703E+05, 9.5017E+05, 9.5332E+05, 9.5647E+05,
     +  9.5963E+05, 9.6280E+05, 9.6598E+05, 9.6917E+05, 9.7236E+05,
     + 9.7557E+05, 9.7878E+05, 9.8200E+05, 9.8523E+05, 9.8847E+05,
     +  9.9172E+05, 9.9497E+05, 9.9824E+05, 1.0015E+06, 1.0048E+06,
     + 1.0081E+06, 1.0114E+06, 1.0147E+06, 1.0180E+06, 1.0213E+06,
     +  1.0247E+06, 1.0280E+06, 1.0314E+06, 1.0347E+06, 1.0381E+06,
     + 1.0415E+06, 1.0448E+06, 1.0482E+06, 1.0516E+06, 1.0551E+06,
     +  1.0585E+06, 1.0619E+06, 1.0654E+06, 1.0688E+06, 1.0723E+06,
     + 1.0757E+06, 1.0792E+06, 1.0827E+06, 1.0862E+06, 1.0897E+06,
     +  1.0932E+06, 1.0967E+06, 1.1002E+06, 1.1038E+06, 1.1073E+06,
     + 1.1109E+06, 1.1145E+06, 1.1180E+06, 1.1216E+06, 1.1252E+06,
     +  1.1288E+06, 1.1324E+06, 1.1361E+06, 1.1397E+06, 1.1433E+06,
     + 1.1470E+06, 1.1506E+06, 1.1543E+06, 1.1580E+06, 1.1617E+06,
     +  1.1654E+06, 1.1691E+06, 1.1728E+06, 1.1765E+06, 1.1803E+06,
     + 1.1840E+06, 1.1878E+06, 1.1915E+06, 1.1953E+06, 1.1991E+06,
     +  1.2029E+06, 1.2067E+06, 1.2105E+06, 1.2143E+06, 1.2182E+06,
     + 1.2220E+06, 1.2258E+06, 1.2297E+06, 1.2336E+06, 1.2375E+06,
     +  1.2414E+06, 1.2453E+06, 1.2492E+06, 1.2531E+06, 1.2570E+06,
     + 1.2610E+06, 1.2649E+06, 1.2689E+06, 1.2728E+06, 1.2768E+06,
     +  1.2808E+06, 1.2848E+06, 1.2888E+06, 1.2928E+06, 1.2969E+06,
     + 1.3009E+06, 1.3050E+06, 1.3090E+06, 1.3131E+06, 1.3172E+06,
     +  1.3213E+06, 1.3254E+06, 1.3295E+06, 1.3336E+06, 1.3377E+06,
     + 1.3419E+06, 1.3460E+06, 1.3502E+06, 1.3544E+06, 1.3586E+06,
     +  1.3627E+06, 1.3670E+06, 1.3712E+06, 1.3754E+06, 1.3796E+06,
     + 1.3839E+06, 1.3881E+06, 1.3924E+06, 1.3967E+06, 1.4010E+06,
     +  1.4053E+06, 1.4096E+06, 1.4139E+06, 1.4182E+06, 1.4226E+06,
     + 1.4269E+06, 1.4313E+06, 1.4357E+06, 1.4401E+06, 1.4445E+06,
     +  1.4489E+06, 1.4533E+06, 1.4577E+06, 1.4621E+06, 1.4666E+06,
     + 1.4711E+06, 1.4755E+06, 1.4800E+06, 1.4845E+06, 1.4890E+06,
     +  1.4935E+06, 1.4981E+06, 1.5026E+06, 1.5071E+06, 1.5117E+06,
     + 1.5163E+06, 1.5209E+06, 1.5255E+06, 1.5301E+06, 1.5347E+06,
     +  1.5393E+06, 1.5439E+06, 1.5486E+06, 1.5533E+06, 1.5579E+06,
     + 1.5626E+06, 1.5673E+06, 1.5720E+06, 1.5767E+06, 1.5815E+06,
     +  1.5862E+06, 1.5910E+06, 1.5957E+06, 1.6005E+06, 1.6053E+06,
     + 1.6101E+06, 1.6149E+06, 1.6197E+06, 1.6245E+06, 1.6294E+06,
     +  1.6342E+06, 1.6391E+06, 1.6440E+06, 1.6489E+06, 1.6538E+06,
     + 1.6587E+06, 1.6636E+06, 1.6686E+06, 1.6735E+06, 1.6785E+06,
     +  1.6835E+06, 1.6884E+06, 1.6934E+06, 1.6984E+06, 1.7035E+06,
     + 1.7085E+06, 1.7135E+06, 1.7186E+06, 1.7237E+06, 1.7288E+06,
     +  1.7338E+06, 1.7390E+06, 1.7441E+06, 1.7492E+06, 1.7543E+06,
     + 1.7595E+06, 1.7647E+06, 1.7698E+06, 1.7750E+06, 1.7802E+06,
     +  1.7854E+06, 1.7907E+06, 1.7959E+06, 1.8012E+06, 1.8064E+06,
     + 1.8117E+06, 1.8170E+06, 1.8223E+06, 1.8276E+06, 1.8329E+06,
     +  1.8383E+06, 1.8436E+06, 1.8490E+06, 1.8544E+06, 1.8598E+06,
     + 1.8652E+06, 1.8706E+06, 1.8760E+06, 1.8814E+06, 1.8869E+06,
     +  1.8924E+06, 1.8978E+06, 1.9033E+06, 1.9088E+06, 1.9143E+06,
     + 1.9199E+06, 1.9254E+06, 1.9310E+06, 1.9365E+06, 1.9421E+06,
     +  1.9477E+06, 1.9533E+06, 1.9589E+06, 1.9646E+06, 1.9702E+06,
     + 1.9759E+06, 1.9816E+06, 1.9872E+06, 1.9929E+06, 1.9986E+06,
     +  2.0044E+06, 2.0101E+06, 2.0159E+06, 2.0216E+06, 2.0274E+06,
     + 2.0332E+06, 2.0390E+06, 2.0448E+06, 2.0506E+06, 2.0565E+06,
     +  2.0623E+06, 2.0682E+06, 2.0741E+06, 2.0800E+06, 2.0859E+06,
     + 2.0918E+06, 2.0978E+06, 2.1037E+06, 2.1097E+06, 2.1157E+06,
     +  2.1217E+06, 2.1277E+06, 2.1337E+06, 2.1397E+06, 2.1458E+06,
     + 2.1518E+06, 2.1579E+06, 2.1640E+06, 2.1701E+06, 2.1762E+06,
     +  2.1823E+06, 2.1885E+06, 2.1946E+06, 2.2008E+06, 2.2070E+06,
     + 2.2132E+06, 2.2194E+06, 2.2256E+06, 2.2319E+06, 2.2381E+06,
     +  2.2444E+06, 2.2507E+06, 2.2570E+06, 2.2633E+06, 2.2696E+06,
     + 2.2760E+06, 2.2823E+06, 2.2887E+06, 2.2951E+06, 2.3015E+06,
     +  2.3079E+06, 2.3143E+06, 2.3208E+06, 2.3272E+06, 2.3337E+06,
     + 2.3402E+06, 2.3467E+06, 2.3532E+06, 2.3597E+06, 2.3663E+06,
     +  2.3728E+06, 2.3794E+06, 2.3860E+06, 2.3926E+06, 2.3992E+06,
     + 2.4058E+06, 2.4125E+06, 2.4191E+06, 2.4258E+06, 2.4325E+06,
     +  2.4392E+06, 2.4459E+06, 2.4527E+06, 2.4594E+06, 2.4662E+06,
     + 2.4730E+06, 2.4798E+06, 2.4866E+06, 2.4934E+06, 2.5002E+06,
     +  2.5071E+06, 2.5140E+06, 2.5209E+06, 2.5278E+06, 2.5347E+06,
     + 2.5416E+06, 2.5485E+06, 2.5555E+06, 2.5625E+06, 2.5695E+06,
     +  2.5765E+06, 2.5835E+06, 2.5905E+06, 2.5976E+06, 2.6047E+06,
     + 2.6118E+06, 2.6189E+06, 2.6260E+06, 2.6331E+06, 2.6403E+06,
     +  2.6474E+06, 2.6546E+06, 2.6618E+06, 2.6690E+06, 2.6762E+06,
     + 2.6835E+06, 2.6907E+06, 2.6980E+06, 2.7053E+06, 2.7126E+06,
     +  2.7199E+06, 2.7272E+06, 2.7346E+06, 2.7420E+06, 2.7493E+06,
     + 2.7567E+06, 2.7642E+06, 2.7716E+06, 2.7790E+06, 2.7865E+06,
     +  2.7940E+06, 2.8015E+06, 2.8090E+06, 2.8165E+06, 2.8241E+06,
     + 2.8316E+06, 2.8392E+06, 2.8468E+06, 2.8544E+06, 2.8620E+06,
     +  2.8697E+06, 2.8773E+06, 2.8850E+06, 2.8927E+06, 2.9004E+06,
     + 2.9082E+06, 2.9159E+06, 2.9237E+06, 2.9314E+06, 2.9392E+06,
     +  2.9470E+06, 2.9549E+06, 2.9627E+06, 2.9706E+06, 2.9784E+06,
     + 2.9863E+06, 2.9942E+06, 3.0022E+06, 3.0101E+06, 3.0181E+06,
     +  3.0260E+06, 3.0340E+06, 3.0420E+06, 3.0501E+06, 3.0581E+06,
     + 3.0662E+06, 3.0743E+06, 3.0824E+06, 3.0905E+06, 3.0986E+06,
     +  3.1068E+06, 3.1149E+06, 3.1231E+06, 3.1313E+06, 3.1395E+06,
     + 3.1478E+06, 3.1560E+06, 3.1643E+06, 3.1726E+06, 3.1809E+06,
     +  3.1892E+06, 3.1975E+06, 3.2059E+06, 3.2143E+06, 3.2226E+06,
     + 3.2311E+06, 3.2395E+06, 3.2479E+06, 3.2564E+06, 3.2649E+06,
     +  3.2734E+06, 3.2819E+06, 3.2904E+06, 3.2990E+06, 3.3075E+06,
     + 3.3161E+06, 3.3247E+06, 3.3334E+06, 3.3420E+06, 3.3507E+06,
     +  3.3593E+06, 3.3680E+06, 3.3767E+06, 3.3855E+06, 3.3942E+06,
     + 3.4030E+06, 3.4118E+06, 3.4206E+06, 3.4294E+06, 3.4383E+06,
     +  3.4471E+06, 3.4560E+06, 3.4649E+06, 3.4738E+06, 3.4827E+06,
     + 3.4917E+06, 3.5007E+06, 3.5096E+06, 3.5187E+06, 3.5277E+06,
     +  3.5367E+06, 3.5458E+06, 3.5549E+06, 3.5640E+06, 3.5731E+06,
     + 3.5822E+06, 3.5914E+06, 3.6006E+06, 3.6098E+06, 3.6190E+06,
     +  3.6282E+06, 3.6375E+06, 3.6467E+06, 3.6560E+06, 3.6653E+06,
     + 3.6747E+06, 3.6840E+06, 3.6934E+06, 3.7028E+06, 3.7122E+06,
     +  3.7216E+06, 3.7310E+06, 3.7405E+06, 3.7500E+06, 3.7595E+06,
     + 3.7690E+06, 3.7785E+06, 3.7881E+06, 3.7977E+06, 3.8073E+06,
     +  3.8169E+06, 3.8265E+06, 3.8362E+06, 3.8458E+06, 3.8555E+06,
     + 3.8653E+06, 3.8750E+06, 3.8848E+06, 3.8945E+06, 3.9043E+06,
     +  3.9141E+06, 3.9240E+06, 3.9338E+06, 3.9437E+06, 3.9536E+06,
     + 3.9635E+06, 3.9734E+06, 3.9834E+06, 3.9934E+06, 4.0034E+06,
     +  4.0134E+06, 4.0234E+06, 4.0335E+06, 4.0435E+06, 4.0536E+06,
     + 4.0638E+06, 4.0739E+06, 4.0840E+06, 4.0942E+06, 4.1044E+06,
     +  4.1146E+06, 4.1249E+06, 4.1351E+06, 4.1454E+06, 4.1557E+06,
     + 4.1660E+06, 4.1764E+06, 4.1867E+06, 4.1971E+06, 4.2075E+06,
     +  4.2179E+06, 4.2284E+06, 4.2388E+06, 4.2493E+06, 4.2598E+06,
     + 4.2704E+06, 4.2809E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 3,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       623
      data (QofT( 4,J),J=1,2408)/ 1.5297E+01, 2.9138E+01, 4.3008E+01,
     +  5.6884E+01, 7.0764E+01, 8.4646E+01, 9.8528E+01, 1.1241E+02,
     + 1.2630E+02, 1.4018E+02, 1.5406E+02, 1.6795E+02, 1.8184E+02,
     +  1.9572E+02, 2.0961E+02, 2.2350E+02, 2.3738E+02, 2.5127E+02,
     + 2.6516E+02, 2.7905E+02, 3.0682E+02, 3.3460E+02, 3.6238E+02,
     +  3.9016E+02, 4.1795E+02, 4.4573E+02, 4.7352E+02, 5.0130E+02,
     + 5.2909E+02, 5.5688E+02, 5.8467E+02, 6.1246E+02, 6.4026E+02,
     +  6.6805E+02, 6.9585E+02, 7.2364E+02, 7.5144E+02, 7.7924E+02,
     + 8.0705E+02, 8.3485E+02, 8.6266E+02, 8.9047E+02, 9.1828E+02,
     +  9.4609E+02, 9.7392E+02, 1.0017E+03, 1.0296E+03, 1.0574E+03,
     + 1.0853E+03, 1.1131E+03, 1.1410E+03, 1.1689E+03, 1.1968E+03,
     +  1.2247E+03, 1.2526E+03, 1.2806E+03, 1.3086E+03, 1.3366E+03,
     + 1.3646E+03, 1.3927E+03, 1.4208E+03, 1.4490E+03, 1.4772E+03,
     +  1.5055E+03, 1.5337E+03, 1.5621E+03, 1.5905E+03, 1.6190E+03,
     + 1.6475E+03, 1.6761E+03, 1.7048E+03, 1.7336E+03, 1.7624E+03,
     +  1.7914E+03, 1.8204E+03, 1.8495E+03, 1.8787E+03, 1.9080E+03,
     + 1.9375E+03, 1.9670E+03, 1.9967E+03, 2.0265E+03, 2.0564E+03,
     +  2.0865E+03, 2.1166E+03, 2.1470E+03, 2.1774E+03, 2.2080E+03,
     + 2.2388E+03, 2.2697E+03, 2.3008E+03, 2.3321E+03, 2.3635E+03,
     +  2.3951E+03, 2.4268E+03, 2.4588E+03, 2.4910E+03, 2.5233E+03,
     + 2.5558E+03, 2.5885E+03, 2.6215E+03, 2.6546E+03, 2.6879E+03,
     +  2.7215E+03, 2.7553E+03, 2.7893E+03, 2.8235E+03, 2.8579E+03,
     + 2.8926E+03, 2.9276E+03, 2.9627E+03, 2.9981E+03, 3.0338E+03,
     +  3.0697E+03, 3.1059E+03, 3.1423E+03, 3.1790E+03, 3.2159E+03,
     + 3.2532E+03, 3.2906E+03, 3.3284E+03, 3.3665E+03, 3.4048E+03,
     +  3.4434E+03, 3.4823E+03, 3.5215E+03, 3.5610E+03, 3.6008E+03,
     + 3.6409E+03, 3.6814E+03, 3.7221E+03, 3.7631E+03, 3.8044E+03,
     +  3.8461E+03, 3.8881E+03, 3.9304E+03, 3.9731E+03, 4.0160E+03,
     + 4.0594E+03, 4.1030E+03, 4.1470E+03, 4.1913E+03, 4.2360E+03,
     +  4.2810E+03, 4.3264E+03, 4.3722E+03, 4.4183E+03, 4.4647E+03,
     + 4.5116E+03, 4.5588E+03, 4.6064E+03, 4.6543E+03, 4.7027E+03,
     +  4.7514E+03, 4.8005E+03, 4.8500E+03, 4.8998E+03, 4.9501E+03,
     + 5.0008E+03, 5.0519E+03, 5.1033E+03, 5.1552E+03, 5.2075E+03,
     +  5.2602E+03, 5.3133E+03, 5.3668E+03, 5.4208E+03, 5.4751E+03,
     + 5.5299E+03, 5.5852E+03, 5.6408E+03, 5.6970E+03, 5.7535E+03,
     +  5.8105E+03, 5.8679E+03, 5.9258E+03, 5.9841E+03, 6.0429E+03,
     + 6.1022E+03, 6.1619E+03, 6.2220E+03, 6.2827E+03, 6.3438E+03,
     +  6.4054E+03, 6.4674E+03, 6.5299E+03, 6.5930E+03, 6.6565E+03,
     + 6.7205E+03, 6.7850E+03, 6.8499E+03, 6.9154E+03, 6.9814E+03,
     +  7.0479E+03, 7.1149E+03, 7.1824E+03, 7.2504E+03, 7.3189E+03,
     + 7.3880E+03, 7.4576E+03, 7.5277E+03, 7.5983E+03, 7.6695E+03,
     +  7.7412E+03, 7.8135E+03, 7.8862E+03, 7.9596E+03, 8.0334E+03,
     + 8.1079E+03, 8.1829E+03, 8.2584E+03, 8.3346E+03, 8.4112E+03,
     +  8.4885E+03, 8.5663E+03, 8.6447E+03, 8.7237E+03, 8.8033E+03,
     + 8.8834E+03, 8.9641E+03, 9.0455E+03, 9.1274E+03, 9.2100E+03,
     +  9.2931E+03, 9.3768E+03, 9.4611E+03, 9.5462E+03, 9.6317E+03,
     + 9.7179E+03, 9.8048E+03, 9.8922E+03, 9.9803E+03, 1.0069E+04,
     +  1.0158E+04, 1.0248E+04, 1.0339E+04, 1.0430E+04, 1.0522E+04,
     + 1.0615E+04, 1.0708E+04, 1.0802E+04, 1.0897E+04, 1.0992E+04,
     +  1.1088E+04, 1.1185E+04, 1.1282E+04, 1.1380E+04, 1.1479E+04,
     + 1.1578E+04, 1.1678E+04, 1.1779E+04, 1.1880E+04, 1.1982E+04,
     +  1.2085E+04, 1.2189E+04, 1.2293E+04, 1.2398E+04, 1.2504E+04,
     + 1.2611E+04, 1.2718E+04, 1.2826E+04, 1.2935E+04, 1.3044E+04,
     +  1.3154E+04, 1.3265E+04, 1.3377E+04, 1.3490E+04, 1.3603E+04,
     + 1.3717E+04, 1.3832E+04, 1.3947E+04, 1.4064E+04, 1.4181E+04,
     +  1.4299E+04, 1.4418E+04, 1.4537E+04, 1.4658E+04, 1.4779E+04,
     + 1.4901E+04, 1.5024E+04, 1.5147E+04, 1.5272E+04, 1.5397E+04,
     +  1.5523E+04, 1.5650E+04, 1.5778E+04, 1.5907E+04, 1.6036E+04,
     + 1.6166E+04, 1.6298E+04, 1.6430E+04, 1.6563E+04, 1.6696E+04,
     +  1.6831E+04, 1.6967E+04, 1.7103E+04, 1.7241E+04, 1.7379E+04,
     + 1.7518E+04, 1.7658E+04, 1.7799E+04, 1.7941E+04, 1.8084E+04,
     +  1.8227E+04, 1.8372E+04, 1.8517E+04, 1.8664E+04, 1.8811E+04,
     + 1.8960E+04, 1.9109E+04, 1.9259E+04, 1.9411E+04, 1.9563E+04,
     +  1.9716E+04, 1.9870E+04, 2.0025E+04, 2.0181E+04, 2.0338E+04,
     + 2.0496E+04, 2.0655E+04, 2.0816E+04, 2.0977E+04, 2.1139E+04,
     +  2.1302E+04, 2.1466E+04, 2.1631E+04, 2.1797E+04, 2.1965E+04,
     + 2.2133E+04, 2.2302E+04, 2.2472E+04, 2.2644E+04, 2.2816E+04,
     +  2.2990E+04, 2.3164E+04, 2.3340E+04, 2.3517E+04, 2.3695E+04,
     + 2.3874E+04, 2.4054E+04, 2.4235E+04, 2.4417E+04, 2.4600E+04,
     +  2.4785E+04, 2.4970E+04, 2.5157E+04, 2.5345E+04, 2.5534E+04,
     + 2.5724E+04, 2.5915E+04, 2.6108E+04, 2.6301E+04, 2.6496E+04,
     +  2.6692E+04, 2.6889E+04, 2.7087E+04, 2.7287E+04, 2.7488E+04,
     + 2.7689E+04, 2.7892E+04, 2.8097E+04, 2.8302E+04, 2.8509E+04,
     +  2.8717E+04, 2.8926E+04, 2.9136E+04, 2.9348E+04, 2.9561E+04,
     + 2.9775E+04, 2.9990E+04, 3.0207E+04, 3.0425E+04, 3.0644E+04,
     +  3.0864E+04, 3.1086E+04, 3.1309E+04, 3.1533E+04, 3.1759E+04,
     + 3.1986E+04, 3.2214E+04, 3.2444E+04, 3.2675E+04, 3.2907E+04,
     +  3.3140E+04, 3.3375E+04, 3.3612E+04, 3.3849E+04, 3.4088E+04,
     + 3.4329E+04, 3.4570E+04, 3.4813E+04, 3.5058E+04, 3.5304E+04,
     +  3.5551E+04, 3.5800E+04, 3.6050E+04, 3.6301E+04, 3.6554E+04,
     + 3.6808E+04, 3.7064E+04, 3.7321E+04, 3.7580E+04, 3.7840E+04,
     +  3.8102E+04, 3.8365E+04, 3.8629E+04, 3.8895E+04, 3.9163E+04,
     + 3.9432E+04, 3.9702E+04, 3.9974E+04, 4.0248E+04, 4.0523E+04,
     +  4.0799E+04, 4.1077E+04, 4.1357E+04, 4.1638E+04, 4.1921E+04,
     + 4.2205E+04, 4.2491E+04, 4.2778E+04, 4.3067E+04, 4.3358E+04,
     +  4.3650E+04, 4.3943E+04, 4.4239E+04, 4.4536E+04, 4.4834E+04,
     + 4.5134E+04, 4.5436E+04, 4.5740E+04, 4.6045E+04, 4.6351E+04,
     +  4.6660E+04, 4.6970E+04, 4.7282E+04, 4.7595E+04, 4.7910E+04,
     + 4.8227E+04, 4.8545E+04, 4.8865E+04, 4.9187E+04, 4.9511E+04,
     +  4.9836E+04, 5.0163E+04, 5.0492E+04, 5.0822E+04, 5.1155E+04,
     + 5.1489E+04, 5.1824E+04, 5.2162E+04, 5.2501E+04, 5.2842E+04,
     +  5.3185E+04, 5.3530E+04, 5.3877E+04, 5.4225E+04, 5.4575E+04,
     + 5.4927E+04, 5.5281E+04, 5.5637E+04, 5.5994E+04, 5.6354E+04,
     +  5.6715E+04, 5.7078E+04, 5.7444E+04, 5.7810E+04, 5.8179E+04,
     + 5.8550E+04, 5.8923E+04, 5.9297E+04, 5.9674E+04, 6.0052E+04,
     +  6.0433E+04, 6.0815E+04, 6.1200E+04, 6.1586E+04, 6.1974E+04,
     + 6.2364E+04, 6.2757E+04, 6.3151E+04, 6.3547E+04, 6.3945E+04,
     +  6.4345E+04, 6.4748E+04, 6.5152E+04, 6.5559E+04, 6.5967E+04,
     + 6.6378E+04, 6.6790E+04, 6.7205E+04, 6.7621E+04, 6.8040E+04,
     +  6.8461E+04, 6.8884E+04, 6.9309E+04, 6.9737E+04, 7.0166E+04,
     + 7.0598E+04, 7.1031E+04, 7.1467E+04, 7.1905E+04, 7.2346E+04,
     +  7.2788E+04, 7.3233E+04, 7.3679E+04, 7.4128E+04, 7.4579E+04,
     + 7.5033E+04, 7.5489E+04, 7.5946E+04, 7.6407E+04, 7.6869E+04,
     +  7.7334E+04, 7.7801E+04, 7.8270E+04, 7.8741E+04, 7.9215E+04,
     + 7.9691E+04, 8.0170E+04, 8.0651E+04, 8.1134E+04, 8.1619E+04,
     +  8.2107E+04, 8.2597E+04, 8.3090E+04, 8.3585E+04, 8.4082E+04,
     + 8.4582E+04, 8.5084E+04, 8.5588E+04, 8.6095E+04, 8.6605E+04,
     +  8.7116E+04, 8.7631E+04, 8.8147E+04, 8.8667E+04, 8.9188E+04,
     + 8.9712E+04, 9.0239E+04, 9.0768E+04, 9.1300E+04, 9.1834E+04,
     +  9.2371E+04, 9.2910E+04, 9.3452E+04, 9.3996E+04, 9.4543E+04,
     + 9.5093E+04, 9.5645E+04, 9.6199E+04, 9.6757E+04, 9.7316E+04,
     +  9.7879E+04, 9.8444E+04, 9.9012E+04, 9.9582E+04, 1.0016E+05,
     + 1.0073E+05, 1.0131E+05, 1.0189E+05, 1.0247E+05, 1.0306E+05,
     +  1.0365E+05, 1.0424E+05, 1.0484E+05, 1.0543E+05, 1.0603E+05,
     + 1.0664E+05, 1.0724E+05, 1.0785E+05, 1.0846E+05, 1.0908E+05,
     +  1.0969E+05, 1.1031E+05, 1.1094E+05, 1.1156E+05, 1.1219E+05,
     + 1.1282E+05, 1.1346E+05, 1.1409E+05, 1.1473E+05, 1.1538E+05,
     +  1.1602E+05, 1.1667E+05, 1.1732E+05, 1.1798E+05, 1.1863E+05,
     + 1.1929E+05, 1.1996E+05, 1.2062E+05, 1.2129E+05, 1.2196E+05,
     +  1.2264E+05, 1.2332E+05, 1.2400E+05, 1.2468E+05, 1.2537E+05,
     + 1.2606E+05, 1.2676E+05, 1.2745E+05, 1.2815E+05, 1.2886E+05,
     +  1.2956E+05, 1.3027E+05, 1.3098E+05, 1.3170E+05, 1.3242E+05,
     + 1.3314E+05, 1.3386E+05, 1.3459E+05, 1.3532E+05, 1.3606E+05,
     +  1.3680E+05, 1.3754E+05, 1.3828E+05, 1.3903E+05, 1.3978E+05,
     + 1.4053E+05, 1.4129E+05, 1.4205E+05, 1.4282E+05, 1.4358E+05,
     +  1.4435E+05, 1.4513E+05, 1.4591E+05, 1.4669E+05, 1.4747E+05,
     + 1.4826E+05, 1.4905E+05, 1.4984E+05, 1.5064E+05, 1.5144E+05,
     +  1.5225E+05, 1.5306E+05, 1.5387E+05, 1.5468E+05, 1.5550E+05,
     + 1.5632E+05, 1.5715E+05, 1.5798E+05, 1.5881E+05, 1.5965E+05,
     +  1.6049E+05, 1.6133E+05, 1.6218E+05, 1.6303E+05, 1.6388E+05,
     + 1.6474E+05, 1.6560E+05, 1.6647E+05, 1.6733E+05, 1.6821E+05,
     +  1.6908E+05, 1.6996E+05, 1.7085E+05, 1.7173E+05, 1.7262E+05,
     + 1.7352E+05, 1.7442E+05, 1.7532E+05, 1.7622E+05, 1.7713E+05,
     +  1.7805E+05, 1.7896E+05, 1.7989E+05, 1.8081E+05, 1.8174E+05,
     + 1.8267E+05, 1.8361E+05, 1.8455E+05, 1.8549E+05, 1.8644E+05,
     +  1.8739E+05, 1.8835E+05, 1.8931E+05, 1.9027E+05, 1.9124E+05,
     + 1.9221E+05, 1.9319E+05, 1.9417E+05, 1.9515E+05, 1.9614E+05,
     +  1.9713E+05, 1.9813E+05, 1.9913E+05, 2.0013E+05, 2.0114E+05,
     + 2.0215E+05, 2.0317E+05, 2.0419E+05, 2.0521E+05, 2.0624E+05,
     +  2.0728E+05, 2.0831E+05, 2.0936E+05, 2.1040E+05, 2.1145E+05,
     + 2.1250E+05, 2.1356E+05, 2.1463E+05, 2.1569E+05, 2.1676E+05,
     +  2.1784E+05, 2.1892E+05, 2.2000E+05, 2.2109E+05, 2.2218E+05,
     + 2.2328E+05, 2.2438E+05, 2.2549E+05, 2.2660E+05, 2.2771E+05,
     +  2.2883E+05, 2.2996E+05, 2.3108E+05, 2.3222E+05, 2.3335E+05,
     + 2.3450E+05, 2.3564E+05, 2.3679E+05, 2.3795E+05, 2.3911E+05,
     +  2.4027E+05, 2.4144E+05, 2.4261E+05, 2.4379E+05, 2.4497E+05,
     + 2.4616E+05, 2.4735E+05, 2.4855E+05, 2.4975E+05, 2.5096E+05,
     +  2.5217E+05, 2.5338E+05, 2.5460E+05, 2.5583E+05, 2.5706E+05,
     + 2.5829E+05, 2.5953E+05, 2.6077E+05, 2.6202E+05, 2.6328E+05,
     +  2.6454E+05, 2.6580E+05, 2.6707E+05, 2.6834E+05, 2.6962E+05,
     + 2.7090E+05, 2.7219E+05, 2.7348E+05, 2.7478E+05, 2.7608E+05,
     +  2.7739E+05, 2.7870E+05, 2.8002E+05, 2.8134E+05, 2.8267E+05,
     + 2.8400E+05, 2.8534E+05, 2.8668E+05, 2.8803E+05, 2.8938E+05,
     +  2.9074E+05, 2.9211E+05, 2.9348E+05, 2.9485E+05, 2.9623E+05,
     + 2.9761E+05, 2.9900E+05, 3.0040E+05, 3.0180E+05, 3.0320E+05,
     +  3.0461E+05, 3.0603E+05, 3.0745E+05, 3.0888E+05, 3.1031E+05,
     + 3.1174E+05, 3.1319E+05, 3.1464E+05, 3.1609E+05, 3.1755E+05,
     +  3.1901E+05, 3.2048E+05, 3.2196E+05, 3.2344E+05, 3.2492E+05,
     + 3.2641E+05, 3.2791E+05, 3.2941E+05, 3.3092E+05, 3.3244E+05,
     +  3.3396E+05, 3.3548E+05, 3.3701E+05, 3.3855E+05, 3.4009E+05,
     + 3.4164E+05, 3.4319E+05, 3.4475E+05, 3.4631E+05, 3.4788E+05,
     +  3.4946E+05, 3.5104E+05, 3.5263E+05, 3.5422E+05, 3.5582E+05,
     + 3.5743E+05, 3.5904E+05, 3.6066E+05, 3.6228E+05, 3.6391E+05,
     +  3.6554E+05, 3.6718E+05, 3.6883E+05, 3.7048E+05, 3.7214E+05,
     + 3.7380E+05, 3.7547E+05, 3.7715E+05, 3.7883E+05, 3.8052E+05,
     +  3.8222E+05, 3.8392E+05, 3.8562E+05, 3.8734E+05, 3.8906E+05,
     + 3.9078E+05, 3.9251E+05, 3.9425E+05, 3.9599E+05, 3.9774E+05,
     +  3.9950E+05, 4.0126E+05, 4.0303E+05, 4.0481E+05, 4.0659E+05,
     + 4.0838E+05, 4.1017E+05, 4.1197E+05, 4.1378E+05, 4.1559E+05,
     +  4.1741E+05, 4.1924E+05, 4.2107E+05, 4.2291E+05, 4.2475E+05,
     + 4.2660E+05, 4.2846E+05, 4.3033E+05, 4.3220E+05, 4.3408E+05,
     +  4.3596E+05, 4.3785E+05, 4.3975E+05, 4.4166E+05, 4.4357E+05,
     + 4.4548E+05, 4.4741E+05, 4.4934E+05, 4.5128E+05, 4.5322E+05,
     +  4.5517E+05, 4.5713E+05, 4.5910E+05, 4.6107E+05, 4.6305E+05,
     + 4.6503E+05, 4.6703E+05, 4.6903E+05, 4.7103E+05, 4.7305E+05,
     +  4.7507E+05, 4.7709E+05, 4.7913E+05, 4.8117E+05, 4.8322E+05,
     + 4.8527E+05, 4.8733E+05, 4.8940E+05, 4.9148E+05, 4.9356E+05,
     +  4.9565E+05, 4.9775E+05, 4.9985E+05, 5.0197E+05, 5.0409E+05,
     + 5.0621E+05, 5.0835E+05, 5.1049E+05, 5.1264E+05, 5.1479E+05,
     +  5.1695E+05, 5.1912E+05, 5.2130E+05, 5.2349E+05, 5.2568E+05,
     + 5.2788E+05, 5.3008E+05, 5.3230E+05, 5.3452E+05, 5.3675E+05,
     +  5.3899E+05, 5.4123E+05, 5.4348E+05, 5.4574E+05, 5.4801E+05,
     + 5.5029E+05, 5.5257E+05, 5.5486E+05, 5.5716E+05, 5.5946E+05,
     +  5.6177E+05, 5.6409E+05, 5.6642E+05, 5.6876E+05, 5.7110E+05,
     + 5.7346E+05, 5.7582E+05, 5.7818E+05, 5.8056E+05, 5.8294E+05,
     +  5.8533E+05, 5.8773E+05, 5.9014E+05, 5.9255E+05, 5.9498E+05,
     + 5.9741E+05, 5.9985E+05, 6.0229E+05, 6.0475E+05, 6.0721E+05,
     +  6.0968E+05, 6.1216E+05, 6.1465E+05, 6.1714E+05, 6.1965E+05,
     + 6.2216E+05, 6.2468E+05, 6.2721E+05, 6.2974E+05, 6.3229E+05,
     +  6.3484E+05, 6.3740E+05, 6.3997E+05, 6.4255E+05, 6.4514E+05,
     + 6.4773E+05, 6.5033E+05, 6.5295E+05, 6.5557E+05, 6.5820E+05,
     +  6.6083E+05, 6.6348E+05, 6.6613E+05, 6.6880E+05, 6.7147E+05,
     + 6.7415E+05, 6.7684E+05, 6.7953E+05, 6.8224E+05, 6.8495E+05,
     +  6.8768E+05, 6.9041E+05, 6.9315E+05, 6.9590E+05, 6.9866E+05,
     + 7.0142E+05, 7.0420E+05, 7.0699E+05, 7.0978E+05, 7.1258E+05,
     +  7.1539E+05, 7.1821E+05, 7.2104E+05, 7.2388E+05, 7.2673E+05,
     + 7.2959E+05, 7.3245E+05, 7.3533E+05, 7.3821E+05, 7.4110E+05,
     +  7.4400E+05, 7.4692E+05, 7.4984E+05, 7.5277E+05, 7.5570E+05,
     + 7.5865E+05, 7.6161E+05, 7.6458E+05, 7.6755E+05, 7.7054E+05,
     +  7.7353E+05, 7.7654E+05, 7.7955E+05, 7.8257E+05, 7.8560E+05,
     + 7.8865E+05, 7.9170E+05, 7.9476E+05, 7.9783E+05, 8.0091E+05,
     +  8.0400E+05, 8.0710E+05, 8.1021E+05, 8.1332E+05, 8.1645E+05,
     + 8.1959E+05, 8.2274E+05, 8.2589E+05, 8.2906E+05, 8.3224E+05,
     +  8.3542E+05, 8.3862E+05, 8.4183E+05, 8.4504E+05, 8.4827E+05,
     + 8.5150E+05, 8.5475E+05, 8.5801E+05, 8.6127E+05, 8.6455E+05,
     +  8.6783E+05, 8.7113E+05, 8.7444E+05, 8.7775E+05, 8.8108E+05,
     + 8.8441E+05, 8.8776E+05, 8.9112E+05, 8.9449E+05, 8.9786E+05,
     +  9.0125E+05, 9.0465E+05, 9.0806E+05, 9.1148E+05, 9.1490E+05,
     + 9.1834E+05, 9.2179E+05, 9.2525E+05, 9.2872E+05, 9.3221E+05,
     +  9.3570E+05, 9.3920E+05, 9.4271E+05, 9.4624E+05, 9.4977E+05,
     + 9.5332E+05, 9.5687E+05, 9.6044E+05, 9.6401E+05, 9.6760E+05,
     +  9.7120E+05, 9.7481E+05, 9.7843E+05, 9.8206E+05, 9.8570E+05,
     + 9.8935E+05, 9.9302E+05, 9.9669E+05, 1.0004E+06, 1.0041E+06,
     +  1.0078E+06, 1.0115E+06, 1.0152E+06, 1.0190E+06, 1.0227E+06,
     + 1.0265E+06, 1.0303E+06, 1.0340E+06, 1.0378E+06, 1.0416E+06,
     +  1.0455E+06, 1.0493E+06, 1.0531E+06, 1.0570E+06, 1.0608E+06,
     + 1.0647E+06, 1.0686E+06, 1.0725E+06, 1.0764E+06, 1.0803E+06,
     +  1.0843E+06, 1.0882E+06, 1.0922E+06, 1.0961E+06, 1.1001E+06,
     + 1.1041E+06, 1.1081E+06, 1.1121E+06, 1.1162E+06, 1.1202E+06,
     +  1.1242E+06, 1.1283E+06, 1.1324E+06, 1.1365E+06, 1.1406E+06,
     + 1.1447E+06, 1.1488E+06, 1.1529E+06, 1.1571E+06, 1.1612E+06,
     +  1.1654E+06, 1.1696E+06, 1.1738E+06, 1.1780E+06, 1.1822E+06,
     + 1.1864E+06, 1.1906E+06, 1.1949E+06, 1.1992E+06, 1.2034E+06,
     +  1.2077E+06, 1.2120E+06, 1.2164E+06, 1.2207E+06, 1.2250E+06,
     + 1.2294E+06, 1.2337E+06, 1.2381E+06, 1.2425E+06, 1.2469E+06,
     +  1.2513E+06, 1.2557E+06, 1.2602E+06, 1.2646E+06, 1.2691E+06,
     + 1.2736E+06, 1.2781E+06, 1.2826E+06, 1.2871E+06, 1.2916E+06,
     +  1.2962E+06, 1.3007E+06, 1.3053E+06, 1.3099E+06, 1.3145E+06,
     + 1.3191E+06, 1.3237E+06, 1.3283E+06, 1.3330E+06, 1.3376E+06,
     +  1.3423E+06, 1.3470E+06, 1.3517E+06, 1.3564E+06, 1.3611E+06,
     + 1.3658E+06, 1.3706E+06, 1.3754E+06, 1.3801E+06, 1.3849E+06,
     +  1.3897E+06, 1.3945E+06, 1.3994E+06, 1.4042E+06, 1.4091E+06,
     + 1.4140E+06, 1.4188E+06, 1.4237E+06, 1.4286E+06, 1.4336E+06,
     +  1.4385E+06, 1.4435E+06, 1.4484E+06, 1.4534E+06, 1.4584E+06,
     + 1.4634E+06, 1.4684E+06, 1.4735E+06, 1.4785E+06, 1.4836E+06,
     +  1.4887E+06, 1.4938E+06, 1.4989E+06, 1.5040E+06, 1.5091E+06,
     + 1.5143E+06, 1.5194E+06, 1.5246E+06, 1.5298E+06, 1.5350E+06,
     +  1.5402E+06, 1.5455E+06, 1.5507E+06, 1.5560E+06, 1.5612E+06,
     + 1.5665E+06, 1.5718E+06, 1.5772E+06, 1.5825E+06, 1.5878E+06,
     +  1.5932E+06, 1.5986E+06, 1.6040E+06, 1.6094E+06, 1.6148E+06,
     + 1.6203E+06, 1.6257E+06, 1.6312E+06, 1.6367E+06, 1.6421E+06,
     +  1.6477E+06, 1.6532E+06, 1.6587E+06, 1.6643E+06, 1.6699E+06,
     + 1.6754E+06, 1.6810E+06, 1.6867E+06, 1.6923E+06, 1.6979E+06,
     +  1.7036E+06, 1.7093E+06, 1.7150E+06, 1.7207E+06, 1.7264E+06,
     + 1.7321E+06, 1.7379E+06, 1.7437E+06, 1.7494E+06, 1.7552E+06,
     +  1.7611E+06, 1.7669E+06, 1.7727E+06, 1.7786E+06, 1.7845E+06,
     + 1.7904E+06, 1.7963E+06, 1.8022E+06, 1.8081E+06, 1.8141E+06,
     +  1.8201E+06, 1.8260E+06, 1.8321E+06, 1.8381E+06, 1.8441E+06,
     + 1.8502E+06, 1.8562E+06, 1.8623E+06, 1.8684E+06, 1.8745E+06,
     +  1.8807E+06, 1.8868E+06, 1.8930E+06, 1.8991E+06, 1.9053E+06,
     + 1.9116E+06, 1.9178E+06, 1.9240E+06, 1.9303E+06, 1.9366E+06,
     +  1.9429E+06, 1.9492E+06, 1.9555E+06, 1.9618E+06, 1.9682E+06,
     + 1.9746E+06, 1.9810E+06, 1.9874E+06, 1.9938E+06, 2.0003E+06,
     +  2.0067E+06, 2.0132E+06, 2.0197E+06, 2.0262E+06, 2.0327E+06,
     + 2.0393E+06, 2.0458E+06, 2.0524E+06, 2.0590E+06, 2.0656E+06,
     +  2.0722E+06, 2.0789E+06, 2.0856E+06, 2.0922E+06, 2.0989E+06,
     + 2.1057E+06, 2.1124E+06, 2.1191E+06, 2.1259E+06, 2.1327E+06,
     +  2.1395E+06, 2.1463E+06, 2.1532E+06, 2.1600E+06, 2.1669E+06,
     + 2.1738E+06, 2.1807E+06, 2.1876E+06, 2.1946E+06, 2.2015E+06,
     +  2.2085E+06, 2.2155E+06, 2.2225E+06, 2.2295E+06, 2.2366E+06,
     + 2.2437E+06, 2.2507E+06, 2.2579E+06, 2.2650E+06, 2.2721E+06,
     +  2.2793E+06, 2.2865E+06, 2.2936E+06, 2.3009E+06, 2.3081E+06,
     + 2.3153E+06, 2.3226E+06, 2.3299E+06, 2.3372E+06, 2.3445E+06,
     +  2.3519E+06, 2.3592E+06, 2.3666E+06, 2.3740E+06, 2.3814E+06,
     + 2.3889E+06, 2.3963E+06, 2.4038E+06, 2.4113E+06, 2.4188E+06,
     +  2.4263E+06, 2.4339E+06, 2.4414E+06, 2.4490E+06, 2.4566E+06,
     + 2.4642E+06, 2.4719E+06, 2.4795E+06, 2.4872E+06, 2.4949E+06,
     +  2.5026E+06, 2.5104E+06, 2.5181E+06, 2.5259E+06, 2.5337E+06,
     + 2.5415E+06, 2.5494E+06, 2.5572E+06, 2.5651E+06, 2.5730E+06,
     +  2.5809E+06, 2.5888E+06, 2.5968E+06, 2.6048E+06, 2.6128E+06,
     + 2.6208E+06, 2.6288E+06, 2.6368E+06, 2.6449E+06, 2.6530E+06,
     +  2.6611E+06, 2.6693E+06, 2.6774E+06, 2.6856E+06, 2.6938E+06,
     + 2.7020E+06, 2.7102E+06, 2.7185E+06, 2.7267E+06, 2.7350E+06,
     +  2.7433E+06, 2.7517E+06, 2.7600E+06, 2.7684E+06, 2.7768E+06,
     + 2.7852E+06, 2.7936E+06, 2.8021E+06, 2.8106E+06, 2.8191E+06,
     +  2.8276E+06, 2.8361E+06, 2.8447E+06, 2.8533E+06, 2.8619E+06,
     + 2.8705E+06, 2.8791E+06, 2.8878E+06, 2.8965E+06, 2.9052E+06,
     +  2.9139E+06, 2.9226E+06, 2.9314E+06, 2.9402E+06, 2.9490E+06,
     + 2.9579E+06, 2.9667E+06, 2.9756E+06, 2.9845E+06, 2.9934E+06,
     +  3.0023E+06, 3.0113E+06, 3.0203E+06, 3.0293E+06, 3.0383E+06,
     + 3.0474E+06, 3.0564E+06, 3.0655E+06, 3.0746E+06, 3.0838E+06,
     +  3.0929E+06, 3.1021E+06, 3.1113E+06, 3.1205E+06, 3.1298E+06,
     + 3.1390E+06, 3.1483E+06, 3.1576E+06, 3.1669E+06, 3.1763E+06,
     +  3.1857E+06, 3.1951E+06, 3.2045E+06, 3.2139E+06, 3.2234E+06,
     + 3.2329E+06, 3.2424E+06, 3.2519E+06, 3.2615E+06, 3.2711E+06,
     +  3.2807E+06, 3.2903E+06, 3.2999E+06, 3.3096E+06, 3.3193E+06,
     + 3.3290E+06, 3.3388E+06, 3.3485E+06, 3.3583E+06, 3.3681E+06,
     +  3.3779E+06, 3.3878E+06, 3.3977E+06, 3.4076E+06, 3.4175E+06,
     + 3.4274E+06, 3.4374E+06, 3.4474E+06, 3.4574E+06, 3.4675E+06,
     +  3.4775E+06, 3.4876E+06, 3.4977E+06, 3.5078E+06, 3.5180E+06,
     + 3.5282E+06, 3.5384E+06, 3.5486E+06, 3.5589E+06, 3.5691E+06,
     +  3.5794E+06, 3.5898E+06, 3.6001E+06, 3.6105E+06, 3.6209E+06,
     + 3.6313E+06, 3.6417E+06, 3.6522E+06, 3.6627E+06, 3.6732E+06,
     +  3.6838E+06, 3.6943E+06, 3.7049E+06, 3.7155E+06, 3.7262E+06,
     + 3.7368E+06, 3.7475E+06, 3.7582E+06, 3.7690E+06, 3.7797E+06,
     +  3.7905E+06, 3.8013E+06, 3.8122E+06, 3.8230E+06, 3.8339E+06,
     + 3.8448E+06, 3.8558E+06, 3.8667E+06, 3.8777E+06, 3.8887E+06,
     +  3.8997E+06, 3.9108E+06, 3.9219E+06, 3.9330E+06, 3.9441E+06,
     + 3.9553E+06, 3.9665E+06, 3.9777E+06, 3.9889E+06, 4.0002E+06,
     +  4.0115E+06, 4.0228E+06, 4.0342E+06, 4.0455E+06, 4.0569E+06,
     + 4.0683E+06, 4.0798E+06, 4.0913E+06, 4.1027E+06, 4.1143E+06,
     +  4.1258E+06, 4.1374E+06, 4.1490E+06, 4.1606E+06, 4.1723E+06,
     + 4.1840E+06, 4.1957E+06, 4.2074E+06, 4.2191E+06, 4.2309E+06,
     +  4.2427E+06, 4.2546E+06, 4.2664E+06, 4.2783E+06, 4.2903E+06,
     + 4.3022E+06, 4.3142E+06, 4.3262E+06, 4.3382E+06, 4.3502E+06,
     +  4.3623E+06, 4.3744E+06, 4.3866E+06, 4.3987E+06, 4.4109E+06,
     + 4.4231E+06, 4.4354E+06, 4.4476E+06, 4.4599E+06, 4.4723E+06,
     +  4.4846E+06, 4.4970E+06, 4.5094E+06, 4.5218E+06, 4.5343E+06,
     + 4.5468E+06, 4.5593E+06, 4.5718E+06, 4.5844E+06, 4.5970E+06,
     +  4.6096E+06, 4.6223E+06, 4.6350E+06, 4.6477E+06, 4.6604E+06,
     + 4.6732E+06, 4.6860E+06, 4.6988E+06, 4.7117E+06, 4.7246E+06,
     +  4.7375E+06, 4.7504E+06, 4.7634E+06, 4.7764E+06, 4.7894E+06,
     + 4.8024E+06, 4.8155E+06, 4.8286E+06, 4.8418E+06, 4.8549E+06,
     +  4.8681E+06, 4.8813E+06, 4.8946E+06, 4.9079E+06, 4.9212E+06,
     + 4.9345E+06, 4.9479E+06, 4.9613E+06, 4.9747E+06, 4.9882E+06,
     +  5.0017E+06, 5.0152E+06, 5.0287E+06, 5.0423E+06, 5.0559E+06,
     + 5.0695E+06, 5.0832E+06, 5.0969E+06, 5.1106E+06, 5.1244E+06,
     +  5.1382E+06, 5.1520E+06, 5.1658E+06, 5.1797E+06, 5.1936E+06,
     + 5.2075E+06, 5.2215E+06, 5.2355E+06, 5.2495E+06, 5.2635E+06,
     +  5.2776E+06, 5.2917E+06, 5.3059E+06, 5.3200E+06, 5.3342E+06,
     + 5.3485E+06, 5.3627E+06, 5.3770E+06, 5.3914E+06, 5.4057E+06,
     +  5.4201E+06, 5.4345E+06, 5.4490E+06, 5.4635E+06, 5.4780E+06,
     + 5.4925E+06, 5.5071E+06, 5.5217E+06, 5.5363E+06, 5.5510E+06,
     +  5.5657E+06, 5.5804E+06, 5.5952E+06, 5.6100E+06, 5.6248E+06,
     + 5.6397E+06, 5.6545E+06, 5.6695E+06, 5.6844E+06, 5.6994E+06,
     +  5.7144E+06, 5.7295E+06, 5.7445E+06, 5.7596E+06, 5.7748E+06,
     + 5.7900E+06, 5.8052E+06, 5.8204E+06, 5.8357E+06, 5.8510E+06,
     +  5.8663E+06, 5.8817E+06, 5.8971E+06, 5.9125E+06, 5.9280E+06,
     + 5.9435E+06, 5.9590E+06, 5.9746E+06, 5.9902E+06, 6.0058E+06,
     +  6.0214E+06, 6.0371E+06, 6.0529E+06, 6.0686E+06, 6.0844E+06,
     + 6.1002E+06, 6.1161E+06, 6.1320E+06, 6.1479E+06, 6.1639E+06,
     +  6.1799E+06, 6.1959E+06, 6.2119E+06, 6.2280E+06, 6.2442E+06,
     + 6.2603E+06, 6.2765E+06, 6.2927E+06, 6.3090E+06, 6.3253E+06,
     +  6.3416E+06, 6.3580E+06, 6.3744E+06, 6.3908E+06, 6.4073E+06,
     + 6.4238E+06, 6.4403E+06, 6.4569E+06, 6.4735E+06, 6.4901E+06,
     +  6.5068E+06, 6.5235E+06, 6.5402E+06, 6.5570E+06, 6.5738E+06,
     + 6.5906E+06, 6.6075E+06, 6.6244E+06, 6.6414E+06, 6.6584E+06,
     +  6.6754E+06, 6.6924E+06, 6.7095E+06, 6.7266E+06, 6.7438E+06,
     + 6.7610E+06, 6.7782E+06, 6.7955E+06, 6.8128E+06, 6.8301E+06,
     +  6.8475E+06, 6.8649E+06, 6.8823E+06, 6.8998E+06, 6.9173E+06,
     + 6.9349E+06, 6.9524E+06, 6.9701E+06, 6.9877E+06, 7.0054E+06,
     +  7.0231E+06, 7.0409E+06, 7.0587E+06, 7.0765E+06, 7.0944E+06,
     + 7.1123E+06, 7.1302E+06, 7.1482E+06, 7.1662E+06, 7.1843E+06,
     +  7.2024E+06, 7.2205E+06, 7.2387E+06, 7.2569E+06, 7.2751E+06,
     + 7.2934E+06, 7.3117E+06, 7.3301E+06, 7.3484E+06, 7.3669E+06,
     +  7.3853E+06, 7.4038E+06, 7.4224E+06, 7.4409E+06, 7.4595E+06,
     + 7.4782E+06, 7.4969E+06, 7.5156E+06, 7.5343E+06, 7.5531E+06,
     +  7.5720E+06, 7.5909E+06, 7.6098E+06, 7.6287E+06, 7.6477E+06,
     + 7.6667E+06, 7.6858E+06, 7.7049E+06, 7.7240E+06, 7.7432E+06,
     +  7.7624E+06, 7.7817E+06, 7.8010E+06, 7.8203E+06, 7.8397E+06,
     + 7.8591E+06, 7.8785E+06, 7.8980E+06, 7.9175E+06, 7.9371E+06,
     +  7.9567E+06, 7.9763E+06, 7.9960E+06, 8.0157E+06, 8.0355E+06,
     + 8.0553E+06, 8.0751E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 4,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       822
      data (QofT( 5,J),J=1,2408)/ 4.0058E+00, 7.6491E+00, 1.1299E+01,
     +  1.4951E+01, 1.8604E+01, 2.2257E+01, 2.5910E+01, 2.9563E+01,
     + 3.3217E+01, 3.6870E+01, 4.0524E+01, 4.4178E+01, 4.7832E+01,
     +  5.1486E+01, 5.5140E+01, 5.8795E+01, 6.2449E+01, 6.6104E+01,
     + 6.9758E+01, 7.3413E+01, 8.0723E+01, 8.8033E+01, 9.5343E+01,
     +  1.0265E+02, 1.0996E+02, 1.1728E+02, 1.2459E+02, 1.3190E+02,
     + 1.3921E+02, 1.4652E+02, 1.5384E+02, 1.6115E+02, 1.6846E+02,
     +  1.7578E+02, 1.8309E+02, 1.9041E+02, 1.9772E+02, 2.0504E+02,
     + 2.1236E+02, 2.1967E+02, 2.2699E+02, 2.3431E+02, 2.4163E+02,
     +  2.4895E+02, 2.5627E+02, 2.6359E+02, 2.7092E+02, 2.7825E+02,
     + 2.8558E+02, 2.9291E+02, 3.0025E+02, 3.0759E+02, 3.1493E+02,
     +  3.2228E+02, 3.2964E+02, 3.3700E+02, 3.4437E+02, 3.5175E+02,
     + 3.5914E+02, 3.6653E+02, 3.7394E+02, 3.8136E+02, 3.8879E+02,
     +  3.9624E+02, 4.0370E+02, 4.1117E+02, 4.1867E+02, 4.2618E+02,
     + 4.3371E+02, 4.4125E+02, 4.4882E+02, 4.5641E+02, 4.6403E+02,
     +  4.7167E+02, 4.7934E+02, 4.8703E+02, 4.9475E+02, 5.0250E+02,
     + 5.1028E+02, 5.1810E+02, 5.2594E+02, 5.3382E+02, 5.4174E+02,
     +  5.4969E+02, 5.5768E+02, 5.6570E+02, 5.7378E+02, 5.8188E+02,
     + 5.9004E+02, 5.9823E+02, 6.0647E+02, 6.1476E+02, 6.2309E+02,
     +  6.3147E+02, 6.3990E+02, 6.4839E+02, 6.5692E+02, 6.6550E+02,
     + 6.7414E+02, 6.8284E+02, 6.9158E+02, 7.0039E+02, 7.0925E+02,
     +  7.1817E+02, 7.2715E+02, 7.3620E+02, 7.4530E+02, 7.5447E+02,
     + 7.6371E+02, 7.7300E+02, 7.8237E+02, 7.9179E+02, 8.0129E+02,
     +  8.1086E+02, 8.2050E+02, 8.3021E+02, 8.3998E+02, 8.4984E+02,
     + 8.5977E+02, 8.6977E+02, 8.7984E+02, 8.8999E+02, 9.0022E+02,
     +  9.1053E+02, 9.2092E+02, 9.3138E+02, 9.4194E+02, 9.5256E+02,
     + 9.6327E+02, 9.7407E+02, 9.8495E+02, 9.9592E+02, 1.0070E+03,
     +  1.0181E+03, 1.0293E+03, 1.0407E+03, 1.0521E+03, 1.0636E+03,
     + 1.0751E+03, 1.0868E+03, 1.0986E+03, 1.1105E+03, 1.1224E+03,
     +  1.1345E+03, 1.1466E+03, 1.1589E+03, 1.1712E+03, 1.1837E+03,
     + 1.1962E+03, 1.2089E+03, 1.2216E+03, 1.2345E+03, 1.2474E+03,
     +  1.2605E+03, 1.2737E+03, 1.2869E+03, 1.3003E+03, 1.3138E+03,
     + 1.3274E+03, 1.3411E+03, 1.3549E+03, 1.3688E+03, 1.3828E+03,
     +  1.3970E+03, 1.4112E+03, 1.4256E+03, 1.4401E+03, 1.4547E+03,
     + 1.4694E+03, 1.4842E+03, 1.4992E+03, 1.5142E+03, 1.5294E+03,
     +  1.5447E+03, 1.5602E+03, 1.5757E+03, 1.5914E+03, 1.6072E+03,
     + 1.6231E+03, 1.6392E+03, 1.6554E+03, 1.6717E+03, 1.6881E+03,
     +  1.7047E+03, 1.7213E+03, 1.7381E+03, 1.7551E+03, 1.7722E+03,
     + 1.7894E+03, 1.8067E+03, 1.8242E+03, 1.8419E+03, 1.8596E+03,
     +  1.8775E+03, 1.8955E+03, 1.9137E+03, 1.9320E+03, 1.9505E+03,
     + 1.9691E+03, 1.9878E+03, 2.0067E+03, 2.0257E+03, 2.0449E+03,
     +  2.0642E+03, 2.0836E+03, 2.1033E+03, 2.1230E+03, 2.1429E+03,
     + 2.1630E+03, 2.1832E+03, 2.2036E+03, 2.2241E+03, 2.2447E+03,
     +  2.2656E+03, 2.2866E+03, 2.3077E+03, 2.3290E+03, 2.3504E+03,
     + 2.3721E+03, 2.3938E+03, 2.4158E+03, 2.4379E+03, 2.4601E+03,
     +  2.4826E+03, 2.5051E+03, 2.5279E+03, 2.5508E+03, 2.5739E+03,
     + 2.5972E+03, 2.6206E+03, 2.6442E+03, 2.6680E+03, 2.6920E+03,
     +  2.7161E+03, 2.7404E+03, 2.7649E+03, 2.7895E+03, 2.8143E+03,
     + 2.8393E+03, 2.8645E+03, 2.8899E+03, 2.9155E+03, 2.9412E+03,
     +  2.9671E+03, 2.9932E+03, 3.0195E+03, 3.0460E+03, 3.0727E+03,
     + 3.0995E+03, 3.1266E+03, 3.1538E+03, 3.1812E+03, 3.2088E+03,
     +  3.2367E+03, 3.2647E+03, 3.2929E+03, 3.3213E+03, 3.3499E+03,
     + 3.3787E+03, 3.4077E+03, 3.4369E+03, 3.4663E+03, 3.4959E+03,
     +  3.5258E+03, 3.5558E+03, 3.5860E+03, 3.6165E+03, 3.6471E+03,
     + 3.6780E+03, 3.7090E+03, 3.7403E+03, 3.7718E+03, 3.8035E+03,
     +  3.8355E+03, 3.8676E+03, 3.9000E+03, 3.9326E+03, 3.9654E+03,
     + 3.9984E+03, 4.0317E+03, 4.0651E+03, 4.0988E+03, 4.1328E+03,
     +  4.1669E+03, 4.2013E+03, 4.2359E+03, 4.2708E+03, 4.3058E+03,
     + 4.3411E+03, 4.3767E+03, 4.4125E+03, 4.4485E+03, 4.4848E+03,
     +  4.5213E+03, 4.5580E+03, 4.5950E+03, 4.6322E+03, 4.6697E+03,
     + 4.7074E+03, 4.7453E+03, 4.7835E+03, 4.8220E+03, 4.8607E+03,
     +  4.8997E+03, 4.9389E+03, 4.9784E+03, 5.0181E+03, 5.0581E+03,
     + 5.0983E+03, 5.1388E+03, 5.1796E+03, 5.2206E+03, 5.2619E+03,
     +  5.3034E+03, 5.3452E+03, 5.3873E+03, 5.4297E+03, 5.4723E+03,
     + 5.5152E+03, 5.5584E+03, 5.6018E+03, 5.6455E+03, 5.6895E+03,
     +  5.7338E+03, 5.7783E+03, 5.8232E+03, 5.8683E+03, 5.9137E+03,
     + 5.9593E+03, 6.0053E+03, 6.0516E+03, 6.0981E+03, 6.1449E+03,
     +  6.1921E+03, 6.2395E+03, 6.2872E+03, 6.3352E+03, 6.3835E+03,
     + 6.4321E+03, 6.4810E+03, 6.5302E+03, 6.5797E+03, 6.6295E+03,
     +  6.6797E+03, 6.7301E+03, 6.7808E+03, 6.8318E+03, 6.8832E+03,
     + 6.9349E+03, 6.9868E+03, 7.0391E+03, 7.0918E+03, 7.1447E+03,
     +  7.1979E+03, 7.2515E+03, 7.3054E+03, 7.3597E+03, 7.4142E+03,
     + 7.4691E+03, 7.5243E+03, 7.5798E+03, 7.6357E+03, 7.6919E+03,
     +  7.7485E+03, 7.8053E+03, 7.8625E+03, 7.9201E+03, 7.9780E+03,
     + 8.0362E+03, 8.0948E+03, 8.1537E+03, 8.2130E+03, 8.2727E+03,
     +  8.3326E+03, 8.3930E+03, 8.4537E+03, 8.5147E+03, 8.5761E+03,
     + 8.6378E+03, 8.7000E+03, 8.7624E+03, 8.8253E+03, 8.8885E+03,
     +  8.9521E+03, 9.0160E+03, 9.0803E+03, 9.1450E+03, 9.2101E+03,
     + 9.2755E+03, 9.3413E+03, 9.4075E+03, 9.4741E+03, 9.5411E+03,
     +  9.6084E+03, 9.6761E+03, 9.7442E+03, 9.8127E+03, 9.8816E+03,
     + 9.9509E+03, 1.0021E+04, 1.0091E+04, 1.0161E+04, 1.0232E+04,
     +  1.0303E+04, 1.0375E+04, 1.0447E+04, 1.0519E+04, 1.0592E+04,
     + 1.0666E+04, 1.0739E+04, 1.0813E+04, 1.0888E+04, 1.0963E+04,
     +  1.1038E+04, 1.1114E+04, 1.1190E+04, 1.1267E+04, 1.1344E+04,
     + 1.1421E+04, 1.1499E+04, 1.1578E+04, 1.1657E+04, 1.1736E+04,
     +  1.1815E+04, 1.1895E+04, 1.1976E+04, 1.2057E+04, 1.2138E+04,
     + 1.2220E+04, 1.2303E+04, 1.2385E+04, 1.2468E+04, 1.2552E+04,
     +  1.2636E+04, 1.2721E+04, 1.2806E+04, 1.2891E+04, 1.2977E+04,
     + 1.3064E+04, 1.3151E+04, 1.3238E+04, 1.3326E+04, 1.3414E+04,
     +  1.3503E+04, 1.3592E+04, 1.3682E+04, 1.3772E+04, 1.3863E+04,
     + 1.3954E+04, 1.4045E+04, 1.4138E+04, 1.4230E+04, 1.4323E+04,
     +  1.4417E+04, 1.4511E+04, 1.4606E+04, 1.4701E+04, 1.4796E+04,
     + 1.4893E+04, 1.4989E+04, 1.5086E+04, 1.5184E+04, 1.5282E+04,
     +  1.5381E+04, 1.5480E+04, 1.5580E+04, 1.5680E+04, 1.5781E+04,
     + 1.5882E+04, 1.5984E+04, 1.6086E+04, 1.6189E+04, 1.6292E+04,
     +  1.6396E+04, 1.6501E+04, 1.6605E+04, 1.6711E+04, 1.6817E+04,
     + 1.6924E+04, 1.7031E+04, 1.7139E+04, 1.7247E+04, 1.7356E+04,
     +  1.7465E+04, 1.7575E+04, 1.7686E+04, 1.7797E+04, 1.7908E+04,
     + 1.8020E+04, 1.8133E+04, 1.8247E+04, 1.8360E+04, 1.8475E+04,
     +  1.8590E+04, 1.8706E+04, 1.8822E+04, 1.8939E+04, 1.9056E+04,
     + 1.9174E+04, 1.9293E+04, 1.9412E+04, 1.9532E+04, 1.9652E+04,
     +  1.9773E+04, 1.9895E+04, 2.0017E+04, 2.0139E+04, 2.0263E+04,
     + 2.0387E+04, 2.0512E+04, 2.0637E+04, 2.0763E+04, 2.0889E+04,
     +  2.1016E+04, 2.1144E+04, 2.1272E+04, 2.1401E+04, 2.1531E+04,
     + 2.1661E+04, 2.1792E+04, 2.1924E+04, 2.2056E+04, 2.2189E+04,
     +  2.2322E+04, 2.2456E+04, 2.2591E+04, 2.2726E+04, 2.2863E+04,
     + 2.2999E+04, 2.3137E+04, 2.3275E+04, 2.3414E+04, 2.3553E+04,
     +  2.3693E+04, 2.3834E+04, 2.3975E+04, 2.4117E+04, 2.4260E+04,
     + 2.4404E+04, 2.4548E+04, 2.4693E+04, 2.4838E+04, 2.4984E+04,
     +  2.5131E+04, 2.5279E+04, 2.5427E+04, 2.5576E+04, 2.5726E+04,
     + 2.5877E+04, 2.6028E+04, 2.6180E+04, 2.6332E+04, 2.6486E+04,
     +  2.6640E+04, 2.6794E+04, 2.6950E+04, 2.7106E+04, 2.7263E+04,
     + 2.7421E+04, 2.7579E+04, 2.7739E+04, 2.7899E+04, 2.8059E+04,
     +  2.8221E+04, 2.8383E+04, 2.8546E+04, 2.8709E+04, 2.8874E+04,
     + 2.9039E+04, 2.9205E+04, 2.9372E+04, 2.9539E+04, 2.9708E+04,
     +  2.9877E+04, 3.0046E+04, 3.0217E+04, 3.0389E+04, 3.0561E+04,
     + 3.0734E+04, 3.0907E+04, 3.1082E+04, 3.1257E+04, 3.1434E+04,
     +  3.1611E+04, 3.1788E+04, 3.1967E+04, 3.2146E+04, 3.2327E+04,
     + 3.2508E+04, 3.2689E+04, 3.2872E+04, 3.3056E+04, 3.3240E+04,
     +  3.3425E+04, 3.3611E+04, 3.3798E+04, 3.3986E+04, 3.4174E+04,
     + 3.4363E+04, 3.4554E+04, 3.4745E+04, 3.4937E+04, 3.5129E+04,
     +  3.5323E+04, 3.5518E+04, 3.5713E+04, 3.5909E+04, 3.6106E+04,
     + 3.6304E+04, 3.6503E+04, 3.6703E+04, 3.6903E+04, 3.7105E+04,
     +  3.7307E+04, 3.7511E+04, 3.7715E+04, 3.7920E+04, 3.8126E+04,
     + 3.8333E+04, 3.8541E+04, 3.8749E+04, 3.8959E+04, 3.9170E+04,
     +  3.9381E+04, 3.9593E+04, 3.9807E+04, 4.0021E+04, 4.0236E+04,
     + 4.0452E+04, 4.0669E+04, 4.0887E+04, 4.1106E+04, 4.1326E+04,
     +  4.1547E+04, 4.1769E+04, 4.1991E+04, 4.2215E+04, 4.2440E+04,
     + 4.2666E+04, 4.2892E+04, 4.3120E+04, 4.3348E+04, 4.3578E+04,
     +  4.3808E+04, 4.4040E+04, 4.4272E+04, 4.4506E+04, 4.4740E+04,
     + 4.4976E+04, 4.5212E+04, 4.5450E+04, 4.5688E+04, 4.5928E+04,
     +  4.6168E+04, 4.6410E+04, 4.6653E+04, 4.6896E+04, 4.7141E+04,
     + 4.7386E+04, 4.7633E+04, 4.7881E+04, 4.8130E+04, 4.8379E+04,
     +  4.8630E+04, 4.8882E+04, 4.9135E+04, 4.9389E+04, 4.9644E+04,
     + 4.9900E+04, 5.0158E+04, 5.0416E+04, 5.0675E+04, 5.0936E+04,
     +  5.1197E+04, 5.1460E+04, 5.1724E+04, 5.1988E+04, 5.2254E+04,
     + 5.2521E+04, 5.2789E+04, 5.3059E+04, 5.3329E+04, 5.3600E+04,
     +  5.3873E+04, 5.4146E+04, 5.4421E+04, 5.4697E+04, 5.4974E+04,
     + 5.5252E+04, 5.5532E+04, 5.5812E+04, 5.6094E+04, 5.6377E+04,
     +  5.6660E+04, 5.6945E+04, 5.7232E+04, 5.7519E+04, 5.7808E+04,
     + 5.8097E+04, 5.8388E+04, 5.8680E+04, 5.8973E+04, 5.9268E+04,
     +  5.9564E+04, 5.9860E+04, 6.0158E+04, 6.0458E+04, 6.0758E+04,
     + 6.1060E+04, 6.1362E+04, 6.1666E+04, 6.1972E+04, 6.2278E+04,
     +  6.2586E+04, 6.2895E+04, 6.3205E+04, 6.3516E+04, 6.3829E+04,
     + 6.4143E+04, 6.4458E+04, 6.4774E+04, 6.5092E+04, 6.5411E+04,
     +  6.5731E+04, 6.6052E+04, 6.6375E+04, 6.6699E+04, 6.7024E+04,
     + 6.7351E+04, 6.7678E+04, 6.8008E+04, 6.8338E+04, 6.8670E+04,
     +  6.9003E+04, 6.9337E+04, 6.9672E+04, 7.0009E+04, 7.0348E+04,
     + 7.0687E+04, 7.1028E+04, 7.1370E+04, 7.1714E+04, 7.2058E+04,
     +  7.2405E+04, 7.2752E+04, 7.3101E+04, 7.3451E+04, 7.3803E+04,
     + 7.4156E+04, 7.4510E+04, 7.4866E+04, 7.5223E+04, 7.5581E+04,
     +  7.5941E+04, 7.6302E+04, 7.6665E+04, 7.7029E+04, 7.7394E+04,
     + 7.7761E+04, 7.8129E+04, 7.8499E+04, 7.8870E+04, 7.9242E+04,
     +  7.9616E+04, 7.9991E+04, 8.0368E+04, 8.0746E+04, 8.1125E+04,
     + 8.1506E+04, 8.1889E+04, 8.2273E+04, 8.2658E+04, 8.3045E+04,
     +  8.3433E+04, 8.3823E+04, 8.4214E+04, 8.4607E+04, 8.5001E+04,
     + 8.5396E+04, 8.5793E+04, 8.6192E+04, 8.6592E+04, 8.6994E+04,
     +  8.7397E+04, 8.7801E+04, 8.8207E+04, 8.8615E+04, 8.9024E+04,
     + 8.9435E+04, 8.9847E+04, 9.0261E+04, 9.0676E+04, 9.1093E+04,
     +  9.1511E+04, 9.1931E+04, 9.2353E+04, 9.2776E+04, 9.3200E+04,
     + 9.3627E+04, 9.4054E+04, 9.4484E+04, 9.4914E+04, 9.5347E+04,
     +  9.5781E+04, 9.6217E+04, 9.6654E+04, 9.7093E+04, 9.7533E+04,
     + 9.7975E+04, 9.8419E+04, 9.8864E+04, 9.9311E+04, 9.9760E+04,
     +  1.0021E+05, 1.0066E+05, 1.0112E+05, 1.0157E+05, 1.0203E+05,
     + 1.0249E+05, 1.0295E+05, 1.0341E+05, 1.0387E+05, 1.0434E+05,
     +  1.0480E+05, 1.0527E+05, 1.0574E+05, 1.0621E+05, 1.0669E+05,
     + 1.0716E+05, 1.0764E+05, 1.0812E+05, 1.0860E+05, 1.0908E+05,
     +  1.0957E+05, 1.1005E+05, 1.1054E+05, 1.1103E+05, 1.1152E+05,
     + 1.1201E+05, 1.1251E+05, 1.1300E+05, 1.1350E+05, 1.1400E+05,
     +  1.1450E+05, 1.1500E+05, 1.1551E+05, 1.1602E+05, 1.1652E+05,
     + 1.1703E+05, 1.1755E+05, 1.1806E+05, 1.1858E+05, 1.1909E+05,
     +  1.1961E+05, 1.2013E+05, 1.2066E+05, 1.2118E+05, 1.2171E+05,
     + 1.2224E+05, 1.2277E+05, 1.2330E+05, 1.2384E+05, 1.2437E+05,
     +  1.2491E+05, 1.2545E+05, 1.2599E+05, 1.2653E+05, 1.2708E+05,
     + 1.2763E+05, 1.2818E+05, 1.2873E+05, 1.2928E+05, 1.2984E+05,
     +  1.3039E+05, 1.3095E+05, 1.3151E+05, 1.3208E+05, 1.3264E+05,
     + 1.3321E+05, 1.3378E+05, 1.3435E+05, 1.3492E+05, 1.3549E+05,
     +  1.3607E+05, 1.3665E+05, 1.3723E+05, 1.3781E+05, 1.3839E+05,
     + 1.3898E+05, 1.3957E+05, 1.4016E+05, 1.4075E+05, 1.4135E+05,
     +  1.4194E+05, 1.4254E+05, 1.4314E+05, 1.4374E+05, 1.4435E+05,
     + 1.4496E+05, 1.4557E+05, 1.4618E+05, 1.4679E+05, 1.4740E+05,
     +  1.4802E+05, 1.4864E+05, 1.4926E+05, 1.4988E+05, 1.5051E+05,
     + 1.5114E+05, 1.5177E+05, 1.5240E+05, 1.5303E+05, 1.5367E+05,
     +  1.5431E+05, 1.5495E+05, 1.5559E+05, 1.5623E+05, 1.5688E+05,
     + 1.5753E+05, 1.5818E+05, 1.5883E+05, 1.5949E+05, 1.6015E+05,
     +  1.6081E+05, 1.6147E+05, 1.6213E+05, 1.6280E+05, 1.6347E+05,
     + 1.6414E+05, 1.6481E+05, 1.6549E+05, 1.6616E+05, 1.6684E+05,
     +  1.6752E+05, 1.6821E+05, 1.6889E+05, 1.6958E+05, 1.7027E+05,
     + 1.7097E+05, 1.7166E+05, 1.7236E+05, 1.7306E+05, 1.7376E+05,
     +  1.7447E+05, 1.7517E+05, 1.7588E+05, 1.7659E+05, 1.7731E+05,
     + 1.7802E+05, 1.7874E+05, 1.7946E+05, 1.8019E+05, 1.8091E+05,
     +  1.8164E+05, 1.8237E+05, 1.8310E+05, 1.8384E+05, 1.8458E+05,
     + 1.8532E+05, 1.8606E+05, 1.8680E+05, 1.8755E+05, 1.8830E+05,
     +  1.8905E+05, 1.8980E+05, 1.9056E+05, 1.9132E+05, 1.9208E+05,
     + 1.9285E+05, 1.9361E+05, 1.9438E+05, 1.9515E+05, 1.9593E+05,
     +  1.9670E+05, 1.9748E+05, 1.9826E+05, 1.9905E+05, 1.9983E+05,
     + 2.0062E+05, 2.0141E+05, 2.0221E+05, 2.0300E+05, 2.0380E+05,
     +  2.0460E+05, 2.0541E+05, 2.0621E+05, 2.0702E+05, 2.0783E+05,
     + 2.0865E+05, 2.0946E+05, 2.1028E+05, 2.1110E+05, 2.1193E+05,
     +  2.1276E+05, 2.1359E+05, 2.1442E+05, 2.1525E+05, 2.1609E+05,
     + 2.1693E+05, 2.1777E+05, 2.1862E+05, 2.1947E+05, 2.2032E+05,
     +  2.2117E+05, 2.2203E+05, 2.2288E+05, 2.2375E+05, 2.2461E+05,
     + 2.2548E+05, 2.2635E+05, 2.2722E+05, 2.2809E+05, 2.2897E+05,
     +  2.2985E+05, 2.3073E+05, 2.3162E+05, 2.3251E+05, 2.3340E+05,
     + 2.3429E+05, 2.3519E+05, 2.3609E+05, 2.3699E+05, 2.3789E+05,
     +  2.3880E+05, 2.3971E+05, 2.4063E+05, 2.4154E+05, 2.4246E+05,
     + 2.4338E+05, 2.4431E+05, 2.4524E+05, 2.4617E+05, 2.4710E+05,
     +  2.4803E+05, 2.4897E+05, 2.4991E+05, 2.5086E+05, 2.5181E+05,
     + 2.5276E+05, 2.5371E+05, 2.5467E+05, 2.5563E+05, 2.5659E+05,
     +  2.5755E+05, 2.5852E+05, 2.5949E+05, 2.6046E+05, 2.6144E+05,
     + 2.6242E+05, 2.6340E+05, 2.6439E+05, 2.6538E+05, 2.6637E+05,
     +  2.6736E+05, 2.6836E+05, 2.6936E+05, 2.7036E+05, 2.7137E+05,
     + 2.7238E+05, 2.7339E+05, 2.7441E+05, 2.7543E+05, 2.7645E+05,
     +  2.7747E+05, 2.7850E+05, 2.7953E+05, 2.8056E+05, 2.8160E+05,
     + 2.8264E+05, 2.8368E+05, 2.8473E+05, 2.8578E+05, 2.8683E+05,
     +  2.8789E+05, 2.8895E+05, 2.9001E+05, 2.9107E+05, 2.9214E+05,
     + 2.9321E+05, 2.9429E+05, 2.9536E+05, 2.9644E+05, 2.9753E+05,
     +  2.9862E+05, 2.9971E+05, 3.0080E+05, 3.0190E+05, 3.0300E+05,
     + 3.0410E+05, 3.0521E+05, 3.0632E+05, 3.0743E+05, 3.0854E+05,
     +  3.0966E+05, 3.1079E+05, 3.1191E+05, 3.1304E+05, 3.1417E+05,
     + 3.1531E+05, 3.1645E+05, 3.1759E+05, 3.1874E+05, 3.1989E+05,
     +  3.2104E+05, 3.2220E+05, 3.2336E+05, 3.2452E+05, 3.2568E+05,
     + 3.2685E+05, 3.2803E+05, 3.2920E+05, 3.3038E+05, 3.3156E+05,
     +  3.3275E+05, 3.3394E+05, 3.3513E+05, 3.3633E+05, 3.3753E+05,
     + 3.3873E+05, 3.3994E+05, 3.4115E+05, 3.4236E+05, 3.4358E+05,
     +  3.4480E+05, 3.4603E+05, 3.4725E+05, 3.4849E+05, 3.4972E+05,
     + 3.5096E+05, 3.5220E+05, 3.5345E+05, 3.5470E+05, 3.5595E+05,
     +  3.5720E+05, 3.5846E+05, 3.5973E+05, 3.6099E+05, 3.6226E+05,
     + 3.6354E+05, 3.6482E+05, 3.6610E+05, 3.6738E+05, 3.6867E+05,
     +  3.6996E+05, 3.7126E+05, 3.7256E+05, 3.7386E+05, 3.7517E+05,
     + 3.7648E+05, 3.7779E+05, 3.7911E+05, 3.8043E+05, 3.8176E+05,
     +  3.8309E+05, 3.8442E+05, 3.8576E+05, 3.8710E+05, 3.8844E+05,
     + 3.8979E+05, 3.9114E+05, 3.9249E+05, 3.9385E+05, 3.9522E+05,
     +  3.9658E+05, 3.9795E+05, 3.9933E+05, 4.0071E+05, 4.0209E+05,
     + 4.0347E+05, 4.0486E+05, 4.0626E+05, 4.0765E+05, 4.0906E+05,
     +  4.1046E+05, 4.1187E+05, 4.1328E+05, 4.1470E+05, 4.1612E+05,
     + 4.1754E+05, 4.1897E+05, 4.2041E+05, 4.2184E+05, 4.2328E+05,
     +  4.2473E+05, 4.2617E+05, 4.2763E+05, 4.2908E+05, 4.3054E+05,
     + 4.3201E+05, 4.3348E+05, 4.3495E+05, 4.3642E+05, 4.3790E+05,
     +  4.3939E+05, 4.4088E+05, 4.4237E+05, 4.4387E+05, 4.4537E+05,
     + 4.4687E+05, 4.4838E+05, 4.4989E+05, 4.5141E+05, 4.5293E+05,
     +  4.5446E+05, 4.5599E+05, 4.5752E+05, 4.5906E+05, 4.6060E+05,
     + 4.6215E+05, 4.6370E+05, 4.6525E+05, 4.6681E+05, 4.6837E+05,
     +  4.6994E+05, 4.7151E+05, 4.7309E+05, 4.7467E+05, 4.7625E+05,
     + 4.7784E+05, 4.7943E+05, 4.8103E+05, 4.8263E+05, 4.8423E+05,
     +  4.8584E+05, 4.8746E+05, 4.8908E+05, 4.9070E+05, 4.9232E+05,
     + 4.9396E+05, 4.9559E+05, 4.9723E+05, 4.9888E+05, 5.0052E+05,
     +  5.0218E+05, 5.0384E+05, 5.0550E+05, 5.0716E+05, 5.0883E+05,
     + 5.1051E+05, 5.1219E+05, 5.1387E+05, 5.1556E+05, 5.1726E+05,
     +  5.1895E+05, 5.2065E+05, 5.2236E+05, 5.2407E+05, 5.2579E+05,
     + 5.2751E+05, 5.2923E+05, 5.3096E+05, 5.3270E+05, 5.3443E+05,
     +  5.3618E+05, 5.3792E+05, 5.3968E+05, 5.4143E+05, 5.4319E+05,
     + 5.4496E+05, 5.4673E+05, 5.4850E+05, 5.5028E+05, 5.5207E+05,
     +  5.5386E+05, 5.5565E+05, 5.5745E+05, 5.5925E+05, 5.6106E+05,
     + 5.6287E+05, 5.6469E+05, 5.6651E+05, 5.6834E+05, 5.7017E+05,
     +  5.7200E+05, 5.7385E+05, 5.7569E+05, 5.7754E+05, 5.7940E+05,
     + 5.8126E+05, 5.8312E+05, 5.8499E+05, 5.8687E+05, 5.8874E+05,
     +  5.9063E+05, 5.9252E+05, 5.9441E+05, 5.9631E+05, 5.9821E+05,
     + 6.0012E+05, 6.0204E+05, 6.0395E+05, 6.0588E+05, 6.0781E+05,
     +  6.0974E+05, 6.1168E+05, 6.1362E+05, 6.1557E+05, 6.1752E+05,
     + 6.1948E+05, 6.2144E+05, 6.2341E+05, 6.2538E+05, 6.2736E+05,
     +  6.2934E+05, 6.3133E+05, 6.3332E+05, 6.3532E+05, 6.3733E+05,
     + 6.3933E+05, 6.4135E+05, 6.4337E+05, 6.4539E+05, 6.4742E+05,
     +  6.4945E+05, 6.5149E+05, 6.5354E+05, 6.5558E+05, 6.5764E+05,
     + 6.5970E+05, 6.6176E+05, 6.6383E+05, 6.6591E+05, 6.6799E+05,
     +  6.7008E+05, 6.7217E+05, 6.7426E+05, 6.7636E+05, 6.7847E+05,
     + 6.8058E+05, 6.8270E+05, 6.8482E+05, 6.8695E+05, 6.8909E+05,
     +  6.9122E+05, 6.9337E+05, 6.9552E+05, 6.9767E+05, 6.9983E+05,
     + 7.0200E+05, 7.0417E+05, 7.0635E+05, 7.0853E+05, 7.1071E+05,
     +  7.1291E+05, 7.1510E+05, 7.1731E+05, 7.1952E+05, 7.2173E+05,
     + 7.2395E+05, 7.2618E+05, 7.2841E+05, 7.3064E+05, 7.3289E+05,
     +  7.3513E+05, 7.3739E+05, 7.3965E+05, 7.4191E+05, 7.4418E+05,
     + 7.4645E+05, 7.4874E+05, 7.5102E+05, 7.5331E+05, 7.5561E+05,
     +  7.5792E+05, 7.6023E+05, 7.6254E+05, 7.6486E+05, 7.6719E+05,
     + 7.6952E+05, 7.7186E+05, 7.7420E+05, 7.7655E+05, 7.7890E+05,
     +  7.8126E+05, 7.8363E+05, 7.8600E+05, 7.8838E+05, 7.9076E+05,
     + 7.9315E+05, 7.9555E+05, 7.9795E+05, 8.0036E+05, 8.0277E+05,
     +  8.0519E+05, 8.0761E+05, 8.1004E+05, 8.1248E+05, 8.1492E+05,
     + 8.1737E+05, 8.1982E+05, 8.2228E+05, 8.2475E+05, 8.2722E+05,
     +  8.2970E+05, 8.3218E+05, 8.3467E+05, 8.3717E+05, 8.3967E+05,
     + 8.4218E+05, 8.4469E+05, 8.4721E+05, 8.4973E+05, 8.5227E+05,
     +  8.5480E+05, 8.5735E+05, 8.5990E+05, 8.6246E+05, 8.6502E+05,
     + 8.6759E+05, 8.7016E+05, 8.7274E+05, 8.7533E+05, 8.7792E+05,
     +  8.8052E+05, 8.8313E+05, 8.8574E+05, 8.8836E+05, 8.9098E+05,
     + 8.9361E+05, 8.9625E+05, 8.9889E+05, 9.0154E+05, 9.0419E+05,
     +  9.0686E+05, 9.0952E+05, 9.1220E+05, 9.1488E+05, 9.1757E+05,
     + 9.2026E+05, 9.2296E+05, 9.2567E+05, 9.2838E+05, 9.3110E+05,
     +  9.3382E+05, 9.3655E+05, 9.3929E+05, 9.4204E+05, 9.4479E+05,
     + 9.4755E+05, 9.5031E+05, 9.5308E+05, 9.5586E+05, 9.5864E+05,
     +  9.6143E+05, 9.6423E+05, 9.6703E+05, 9.6984E+05, 9.7266E+05,
     + 9.7548E+05, 9.7831E+05, 9.8115E+05, 9.8399E+05, 9.8684E+05,
     +  9.8969E+05, 9.9256E+05, 9.9543E+05, 9.9830E+05, 1.0012E+06,
     + 1.0041E+06, 1.0070E+06, 1.0099E+06, 1.0128E+06, 1.0157E+06,
     +  1.0186E+06, 1.0216E+06, 1.0245E+06, 1.0274E+06, 1.0304E+06,
     + 1.0333E+06, 1.0363E+06, 1.0393E+06, 1.0422E+06, 1.0452E+06,
     +  1.0482E+06, 1.0512E+06, 1.0542E+06, 1.0572E+06, 1.0603E+06,
     + 1.0633E+06, 1.0663E+06, 1.0694E+06, 1.0724E+06, 1.0755E+06,
     +  1.0785E+06, 1.0816E+06, 1.0847E+06, 1.0877E+06, 1.0908E+06,
     + 1.0939E+06, 1.0970E+06, 1.1001E+06, 1.1033E+06, 1.1064E+06,
     +  1.1095E+06, 1.1127E+06, 1.1158E+06, 1.1190E+06, 1.1221E+06,
     + 1.1253E+06, 1.1285E+06, 1.1316E+06, 1.1348E+06, 1.1380E+06,
     +  1.1412E+06, 1.1444E+06, 1.1477E+06, 1.1509E+06, 1.1541E+06,
     + 1.1573E+06, 1.1606E+06, 1.1639E+06, 1.1671E+06, 1.1704E+06,
     +  1.1737E+06, 1.1769E+06, 1.1802E+06, 1.1835E+06, 1.1868E+06,
     + 1.1902E+06, 1.1935E+06, 1.1968E+06, 1.2001E+06, 1.2035E+06,
     +  1.2068E+06, 1.2102E+06, 1.2136E+06, 1.2169E+06, 1.2203E+06,
     + 1.2237E+06, 1.2271E+06, 1.2305E+06, 1.2339E+06, 1.2373E+06,
     +  1.2408E+06, 1.2442E+06, 1.2476E+06, 1.2511E+06, 1.2545E+06,
     + 1.2580E+06, 1.2615E+06, 1.2650E+06, 1.2684E+06, 1.2719E+06,
     +  1.2754E+06, 1.2790E+06, 1.2825E+06, 1.2860E+06, 1.2895E+06,
     + 1.2931E+06, 1.2966E+06, 1.3002E+06, 1.3038E+06, 1.3073E+06,
     +  1.3109E+06, 1.3145E+06, 1.3181E+06, 1.3217E+06, 1.3253E+06,
     + 1.3289E+06, 1.3326E+06, 1.3362E+06, 1.3398E+06, 1.3435E+06,
     +  1.3472E+06, 1.3508E+06, 1.3545E+06, 1.3582E+06, 1.3619E+06,
     + 1.3656E+06, 1.3693E+06, 1.3730E+06, 1.3767E+06, 1.3805E+06,
     +  1.3842E+06, 1.3880E+06, 1.3917E+06, 1.3955E+06, 1.3993E+06,
     + 1.4031E+06, 1.4068E+06, 1.4106E+06, 1.4145E+06, 1.4183E+06,
     +  1.4221E+06, 1.4259E+06, 1.4298E+06, 1.4336E+06, 1.4375E+06,
     + 1.4413E+06, 1.4452E+06, 1.4491E+06, 1.4530E+06, 1.4569E+06,
     +  1.4608E+06, 1.4647E+06, 1.4686E+06, 1.4726E+06, 1.4765E+06,
     + 1.4805E+06, 1.4844E+06, 1.4884E+06, 1.4924E+06, 1.4964E+06,
     +  1.5003E+06, 1.5043E+06, 1.5084E+06, 1.5124E+06, 1.5164E+06,
     + 1.5204E+06, 1.5245E+06, 1.5285E+06, 1.5326E+06, 1.5367E+06,
     +  1.5407E+06, 1.5448E+06, 1.5489E+06, 1.5530E+06, 1.5572E+06,
     + 1.5613E+06, 1.5654E+06, 1.5696E+06, 1.5737E+06, 1.5779E+06,
     +  1.5820E+06, 1.5862E+06, 1.5904E+06, 1.5946E+06, 1.5988E+06,
     + 1.6030E+06, 1.6072E+06, 1.6114E+06, 1.6157E+06, 1.6199E+06,
     +  1.6242E+06, 1.6285E+06, 1.6327E+06, 1.6370E+06, 1.6413E+06,
     + 1.6456E+06, 1.6499E+06, 1.6542E+06, 1.6586E+06, 1.6629E+06,
     +  1.6673E+06, 1.6716E+06, 1.6760E+06, 1.6803E+06, 1.6847E+06,
     + 1.6891E+06, 1.6935E+06, 1.6979E+06, 1.7024E+06, 1.7068E+06,
     +  1.7112E+06, 1.7157E+06, 1.7201E+06, 1.7246E+06, 1.7291E+06,
     + 1.7336E+06, 1.7381E+06, 1.7426E+06, 1.7471E+06, 1.7516E+06,
     +  1.7561E+06, 1.7607E+06, 1.7652E+06, 1.7698E+06, 1.7744E+06,
     + 1.7789E+06, 1.7835E+06, 1.7881E+06, 1.7927E+06, 1.7974E+06,
     +  1.8020E+06, 1.8066E+06, 1.8113E+06, 1.8159E+06, 1.8206E+06,
     + 1.8253E+06, 1.8300E+06, 1.8347E+06, 1.8394E+06, 1.8441E+06,
     +  1.8488E+06, 1.8535E+06, 1.8583E+06, 1.8630E+06, 1.8678E+06,
     + 1.8726E+06, 1.8774E+06, 1.8821E+06, 1.8870E+06, 1.8918E+06,
     +  1.8966E+06, 1.9014E+06, 1.9063E+06, 1.9111E+06, 1.9160E+06,
     + 1.9209E+06, 1.9257E+06, 1.9306E+06, 1.9355E+06, 1.9404E+06,
     +  1.9454E+06, 1.9503E+06, 1.9552E+06, 1.9602E+06, 1.9652E+06,
     + 1.9701E+06, 1.9751E+06, 1.9801E+06, 1.9851E+06, 1.9901E+06,
     +  1.9951E+06, 2.0002E+06, 2.0052E+06, 2.0103E+06, 2.0153E+06,
     + 2.0204E+06, 2.0255E+06, 2.0306E+06, 2.0357E+06, 2.0408E+06,
     +  2.0459E+06, 2.0511E+06, 2.0562E+06, 2.0614E+06, 2.0666E+06,
     + 2.0717E+06, 2.0769E+06, 2.0821E+06, 2.0873E+06, 2.0925E+06,
     +  2.0978E+06, 2.1030E+06, 2.1083E+06, 2.1135E+06, 2.1188E+06,
     + 2.1241E+06, 2.1294E+06, 2.1347E+06, 2.1400E+06, 2.1453E+06,
     +  2.1507E+06, 2.1560E+06, 2.1614E+06, 2.1667E+06, 2.1721E+06,
     + 2.1775E+06, 2.1829E+06, 2.1883E+06, 2.1938E+06, 2.1992E+06,
     +  2.2046E+06, 2.2101E+06, 2.2156E+06, 2.2210E+06, 2.2265E+06,
     + 2.2320E+06, 2.2375E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 5,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_H2CO  (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 3),QofT( 3,NT),Q(NT)
      data (Tmx(I),I=1,3)/6000.,3500.,3500./
      data xgj/ 1.,2.,1./
c...     H2CO
c...        --       126
      data (QofT( 1,J),J=1,2408)/ 1.0911E+00, 1.5582E+00, 2.2232E+00,
     +  3.1764E+00, 4.4829E+00, 6.1380E+00, 8.1062E+00, 1.0346E+01,
     + 1.2819E+01, 1.5494E+01, 1.8343E+01, 2.1347E+01, 2.4488E+01,
     +  2.7753E+01, 3.1131E+01, 3.4614E+01, 3.8194E+01, 4.1867E+01,
     + 4.5627E+01, 4.9471E+01, 5.7400E+01, 6.5636E+01, 7.4169E+01,
     +  8.2989E+01, 9.2090E+01, 1.0147E+02, 1.1112E+02, 1.2103E+02,
     + 1.3121E+02, 1.4165E+02, 1.5234E+02, 1.6328E+02, 1.7446E+02,
     +  1.8589E+02, 1.9755E+02, 2.0944E+02, 2.2156E+02, 2.3390E+02,
     + 2.4647E+02, 2.5925E+02, 2.7225E+02, 2.8545E+02, 2.9886E+02,
     +  3.1248E+02, 3.2630E+02, 3.4031E+02, 3.5452E+02, 3.6893E+02,
     + 3.8352E+02, 3.9830E+02, 4.1327E+02, 4.2842E+02, 4.4375E+02,
     +  4.5926E+02, 4.7495E+02, 4.9081E+02, 5.0684E+02, 5.2305E+02,
     + 5.3943E+02, 5.5597E+02, 5.7268E+02, 5.8955E+02, 6.0659E+02,
     +  6.2379E+02, 6.4115E+02, 6.5867E+02, 6.7634E+02, 6.9417E+02,
     + 7.1215E+02, 7.3029E+02, 7.4858E+02, 7.6702E+02, 7.8561E+02,
     +  8.0435E+02, 8.2323E+02, 8.4226E+02, 8.6144E+02, 8.8076E+02,
     + 9.0022E+02, 9.1983E+02, 9.3957E+02, 9.5946E+02, 9.7948E+02,
     +  9.9964E+02, 1.0199E+03, 1.0404E+03, 1.0610E+03, 1.0817E+03,
     + 1.1025E+03, 1.1235E+03, 1.1446E+03, 1.1658E+03, 1.1872E+03,
     +  1.2087E+03, 1.2303E+03, 1.2521E+03, 1.2740E+03, 1.2960E+03,
     + 1.3181E+03, 1.3404E+03, 1.3628E+03, 1.3853E+03, 1.4080E+03,
     +  1.4308E+03, 1.4537E+03, 1.4767E+03, 1.4999E+03, 1.5231E+03,
     + 1.5465E+03, 1.5701E+03, 1.5937E+03, 1.6175E+03, 1.6414E+03,
     +  1.6654E+03, 1.6895E+03, 1.7138E+03, 1.7382E+03, 1.7627E+03,
     + 1.7873E+03, 1.8120E+03, 1.8369E+03, 1.8619E+03, 1.8870E+03,
     +  1.9122E+03, 1.9376E+03, 1.9631E+03, 1.9887E+03, 2.0144E+03,
     + 2.0403E+03, 2.0662E+03, 2.0923E+03, 2.1185E+03, 2.1449E+03,
     +  2.1713E+03, 2.1979E+03, 2.2246E+03, 2.2515E+03, 2.2784E+03,
     + 2.3055E+03, 2.3327E+03, 2.3601E+03, 2.3875E+03, 2.4151E+03,
     +  2.4428E+03, 2.4707E+03, 2.4987E+03, 2.5268E+03, 2.5550E+03,
     + 2.5833E+03, 2.6118E+03, 2.6405E+03, 2.6692E+03, 2.6981E+03,
     +  2.7271E+03, 2.7563E+03, 2.7856E+03, 2.8150E+03, 2.8445E+03,
     + 2.8742E+03, 2.9041E+03, 2.9340E+03, 2.9641E+03, 2.9944E+03,
     +  3.0248E+03, 3.0553E+03, 3.0859E+03, 3.1168E+03, 3.1477E+03,
     + 3.1788E+03, 3.2100E+03, 3.2414E+03, 3.2730E+03, 3.3046E+03,
     +  3.3365E+03, 3.3685E+03, 3.4006E+03, 3.4329E+03, 3.4653E+03,
     + 3.4979E+03, 3.5306E+03, 3.5635E+03, 3.5966E+03, 3.6298E+03,
     +  3.6632E+03, 3.6967E+03, 3.7304E+03, 3.7642E+03, 3.7983E+03,
     + 3.8324E+03, 3.8668E+03, 3.9013E+03, 3.9360E+03, 3.9708E+03,
     +  4.0058E+03, 4.0410E+03, 4.0764E+03, 4.1119E+03, 4.1476E+03,
     + 4.1835E+03, 4.2196E+03, 4.2558E+03, 4.2922E+03, 4.3288E+03,
     +  4.3656E+03, 4.4026E+03, 4.4397E+03, 4.4770E+03, 4.5145E+03,
     + 4.5523E+03, 4.5901E+03, 4.6282E+03, 4.6665E+03, 4.7050E+03,
     +  4.7436E+03, 4.7825E+03, 4.8216E+03, 4.8608E+03, 4.9003E+03,
     + 4.9399E+03, 4.9798E+03, 5.0199E+03, 5.0601E+03, 5.1006E+03,
     +  5.1413E+03, 5.1822E+03, 5.2233E+03, 5.2646E+03, 5.3061E+03,
     + 5.3479E+03, 5.3899E+03, 5.4320E+03, 5.4744E+03, 5.5171E+03,
     +  5.5599E+03, 5.6030E+03, 5.6463E+03, 5.6898E+03, 5.7336E+03,
     + 5.7776E+03, 5.8218E+03, 5.8662E+03, 5.9109E+03, 5.9558E+03,
     +  6.0010E+03, 6.0464E+03, 6.0920E+03, 6.1379E+03, 6.1840E+03,
     + 6.2304E+03, 6.2770E+03, 6.3239E+03, 6.3710E+03, 6.4184E+03,
     +  6.4660E+03, 6.5139E+03, 6.5621E+03, 6.6105E+03, 6.6591E+03,
     + 6.7081E+03, 6.7573E+03, 6.8067E+03, 6.8564E+03, 6.9064E+03,
     +  6.9567E+03, 7.0072E+03, 7.0580E+03, 7.1091E+03, 7.1605E+03,
     + 7.2121E+03, 7.2640E+03, 7.3163E+03, 7.3687E+03, 7.4215E+03,
     +  7.4746E+03, 7.5280E+03, 7.5816E+03, 7.6355E+03, 7.6898E+03,
     + 7.7443E+03, 7.7992E+03, 7.8543E+03, 7.9097E+03, 7.9655E+03,
     +  8.0215E+03, 8.0779E+03, 8.1346E+03, 8.1916E+03, 8.2489E+03,
     + 8.3065E+03, 8.3644E+03, 8.4227E+03, 8.4812E+03, 8.5401E+03,
     +  8.5994E+03, 8.6589E+03, 8.7188E+03, 8.7790E+03, 8.8396E+03,
     + 8.9004E+03, 8.9617E+03, 9.0232E+03, 9.0851E+03, 9.1474E+03,
     +  9.2100E+03, 9.2729E+03, 9.3362E+03, 9.3998E+03, 9.4638E+03,
     + 9.5282E+03, 9.5929E+03, 9.6580E+03, 9.7234E+03, 9.7892E+03,
     +  9.8554E+03, 9.9219E+03, 9.9888E+03, 1.0056E+04, 1.0124E+04,
     + 1.0192E+04, 1.0260E+04, 1.0329E+04, 1.0398E+04, 1.0468E+04,
     +  1.0538E+04, 1.0608E+04, 1.0679E+04, 1.0750E+04, 1.0821E+04,
     + 1.0893E+04, 1.0965E+04, 1.1038E+04, 1.1111E+04, 1.1184E+04,
     +  1.1258E+04, 1.1332E+04, 1.1407E+04, 1.1482E+04, 1.1557E+04,
     + 1.1633E+04, 1.1710E+04, 1.1786E+04, 1.1864E+04, 1.1941E+04,
     +  1.2019E+04, 1.2098E+04, 1.2177E+04, 1.2256E+04, 1.2336E+04,
     + 1.2416E+04, 1.2497E+04, 1.2578E+04, 1.2659E+04, 1.2741E+04,
     +  1.2824E+04, 1.2907E+04, 1.2990E+04, 1.3074E+04, 1.3158E+04,
     + 1.3243E+04, 1.3328E+04, 1.3414E+04, 1.3500E+04, 1.3587E+04,
     +  1.3674E+04, 1.3762E+04, 1.3850E+04, 1.3939E+04, 1.4028E+04,
     + 1.4117E+04, 1.4208E+04, 1.4298E+04, 1.4389E+04, 1.4481E+04,
     +  1.4573E+04, 1.4666E+04, 1.4759E+04, 1.4853E+04, 1.4947E+04,
     + 1.5041E+04, 1.5137E+04, 1.5232E+04, 1.5329E+04, 1.5426E+04,
     +  1.5523E+04, 1.5621E+04, 1.5719E+04, 1.5818E+04, 1.5918E+04,
     + 1.6018E+04, 1.6119E+04, 1.6220E+04, 1.6322E+04, 1.6424E+04,
     +  1.6527E+04, 1.6630E+04, 1.6734E+04, 1.6839E+04, 1.6944E+04,
     + 1.7050E+04, 1.7156E+04, 1.7263E+04, 1.7371E+04, 1.7479E+04,
     +  1.7587E+04, 1.7697E+04, 1.7806E+04, 1.7917E+04, 1.8028E+04,
     + 1.8140E+04, 1.8252E+04, 1.8365E+04, 1.8479E+04, 1.8593E+04,
     +  1.8708E+04, 1.8823E+04, 1.8939E+04, 1.9056E+04, 1.9173E+04,
     + 1.9291E+04, 1.9410E+04, 1.9529E+04, 1.9649E+04, 1.9770E+04,
     +  1.9891E+04, 2.0013E+04, 2.0135E+04, 2.0259E+04, 2.0382E+04,
     + 2.0507E+04, 2.0632E+04, 2.0758E+04, 2.0885E+04, 2.1012E+04,
     +  2.1140E+04, 2.1269E+04, 2.1398E+04, 2.1528E+04, 2.1659E+04,
     + 2.1791E+04, 2.1923E+04, 2.2056E+04, 2.2190E+04, 2.2324E+04,
     +  2.2459E+04, 2.2595E+04, 2.2731E+04, 2.2869E+04, 2.3007E+04,
     + 2.3146E+04, 2.3285E+04, 2.3426E+04, 2.3567E+04, 2.3708E+04,
     +  2.3851E+04, 2.3994E+04, 2.4139E+04, 2.4283E+04, 2.4429E+04,
     + 2.4576E+04, 2.4723E+04, 2.4871E+04, 2.5020E+04, 2.5169E+04,
     +  2.5320E+04, 2.5471E+04, 2.5623E+04, 2.5776E+04, 2.5930E+04,
     + 2.6084E+04, 2.6240E+04, 2.6396E+04, 2.6553E+04, 2.6711E+04,
     +  2.6870E+04, 2.7029E+04, 2.7190E+04, 2.7351E+04, 2.7513E+04,
     + 2.7676E+04, 2.7840E+04, 2.8005E+04, 2.8170E+04, 2.8337E+04,
     +  2.8504E+04, 2.8672E+04, 2.8841E+04, 2.9011E+04, 2.9182E+04,
     + 2.9354E+04, 2.9527E+04, 2.9701E+04, 2.9875E+04, 3.0051E+04,
     +  3.0227E+04, 3.0405E+04, 3.0583E+04, 3.0762E+04, 3.0943E+04,
     + 3.1124E+04, 3.1306E+04, 3.1489E+04, 3.1673E+04, 3.1858E+04,
     +  3.2044E+04, 3.2231E+04, 3.2419E+04, 3.2608E+04, 3.2798E+04,
     + 3.2989E+04, 3.3180E+04, 3.3373E+04, 3.3567E+04, 3.3762E+04,
     +  3.3958E+04, 3.4155E+04, 3.4353E+04, 3.4552E+04, 3.4752E+04,
     + 3.4954E+04, 3.5156E+04, 3.5359E+04, 3.5563E+04, 3.5769E+04,
     +  3.5975E+04, 3.6183E+04, 3.6391E+04, 3.6601E+04, 3.6812E+04,
     + 3.7024E+04, 3.7237E+04, 3.7451E+04, 3.7666E+04, 3.7882E+04,
     +  3.8099E+04, 3.8318E+04, 3.8538E+04, 3.8759E+04, 3.8980E+04,
     + 3.9204E+04, 3.9428E+04, 3.9653E+04, 3.9880E+04, 4.0107E+04,
     +  4.0336E+04, 4.0567E+04, 4.0798E+04, 4.1030E+04, 4.1264E+04,
     + 4.1499E+04, 4.1735E+04, 4.1972E+04, 4.2211E+04, 4.2450E+04,
     +  4.2691E+04, 4.2933E+04, 4.3177E+04, 4.3421E+04, 4.3667E+04,
     + 4.3914E+04, 4.4163E+04, 4.4412E+04, 4.4663E+04, 4.4916E+04,
     +  4.5169E+04, 4.5424E+04, 4.5680E+04, 4.5938E+04, 4.6196E+04,
     + 4.6456E+04, 4.6718E+04, 4.6980E+04, 4.7244E+04, 4.7510E+04,
     +  4.7776E+04, 4.8044E+04, 4.8314E+04, 4.8585E+04, 4.8857E+04,
     + 4.9130E+04, 4.9405E+04, 4.9681E+04, 4.9959E+04, 5.0238E+04,
     +  5.0519E+04, 5.0800E+04, 5.1084E+04, 5.1368E+04, 5.1655E+04,
     + 5.1942E+04, 5.2231E+04, 5.2522E+04, 5.2814E+04, 5.3107E+04,
     +  5.3402E+04, 5.3698E+04, 5.3996E+04, 5.4295E+04, 5.4596E+04,
     + 5.4899E+04, 5.5202E+04, 5.5508E+04, 5.5815E+04, 5.6123E+04,
     +  5.6433E+04, 5.6744E+04, 5.7057E+04, 5.7372E+04, 5.7688E+04,
     + 5.8006E+04, 5.8325E+04, 5.8646E+04, 5.8968E+04, 5.9292E+04,
     +  5.9618E+04, 5.9945E+04, 6.0274E+04, 6.0605E+04, 6.0937E+04,
     + 6.1271E+04, 6.1606E+04, 6.1943E+04, 6.2282E+04, 6.2622E+04,
     +  6.2964E+04, 6.3308E+04, 6.3654E+04, 6.4001E+04, 6.4350E+04,
     + 6.4700E+04, 6.5052E+04, 6.5407E+04, 6.5762E+04, 6.6120E+04,
     +  6.6479E+04, 6.6840E+04, 6.7203E+04, 6.7567E+04, 6.7934E+04,
     + 6.8302E+04, 6.8672E+04, 6.9043E+04, 6.9417E+04, 6.9792E+04,
     +  7.0169E+04, 7.0548E+04, 7.0929E+04, 7.1312E+04, 7.1697E+04,
     + 7.2083E+04, 7.2471E+04, 7.2861E+04, 7.3253E+04, 7.3647E+04,
     +  7.4043E+04, 7.4441E+04, 7.4841E+04, 7.5242E+04, 7.5646E+04,
     + 7.6052E+04, 7.6459E+04, 7.6868E+04, 7.7280E+04, 7.7693E+04,
     +  7.8108E+04, 7.8526E+04, 7.8945E+04, 7.9367E+04, 7.9790E+04,
     + 8.0215E+04, 8.0643E+04, 8.1072E+04, 8.1504E+04, 8.1938E+04,
     +  8.2373E+04, 8.2811E+04, 8.3251E+04, 8.3693E+04, 8.4137E+04,
     + 8.4583E+04, 8.5032E+04, 8.5482E+04, 8.5935E+04, 8.6390E+04,
     +  8.6846E+04, 8.7306E+04, 8.7767E+04, 8.8230E+04, 8.8696E+04,
     + 8.9164E+04, 8.9634E+04, 9.0106E+04, 9.0581E+04, 9.1058E+04,
     +  9.1537E+04, 9.2018E+04, 9.2502E+04, 9.2987E+04, 9.3475E+04,
     + 9.3966E+04, 9.4459E+04, 9.4954E+04, 9.5451E+04, 9.5951E+04,
     +  9.6453E+04, 9.6958E+04, 9.7465E+04, 9.7974E+04, 9.8485E+04,
     + 9.8999E+04, 9.9516E+04, 1.0003E+05, 1.0056E+05, 1.0108E+05,
     +  1.0161E+05, 1.0213E+05, 1.0267E+05, 1.0320E+05, 1.0373E+05,
     + 1.0427E+05, 1.0481E+05, 1.0536E+05, 1.0590E+05, 1.0645E+05,
     +  1.0700E+05, 1.0756E+05, 1.0811E+05, 1.0867E+05, 1.0923E+05,
     + 1.0980E+05, 1.1036E+05, 1.1093E+05, 1.1151E+05, 1.1208E+05,
     +  1.1266E+05, 1.1324E+05, 1.1382E+05, 1.1440E+05, 1.1499E+05,
     + 1.1558E+05, 1.1618E+05, 1.1677E+05, 1.1737E+05, 1.1797E+05,
     +  1.1858E+05, 1.1918E+05, 1.1979E+05, 1.2041E+05, 1.2102E+05,
     + 1.2164E+05, 1.2226E+05, 1.2289E+05, 1.2351E+05, 1.2414E+05,
     +  1.2477E+05, 1.2541E+05, 1.2605E+05, 1.2669E+05, 1.2733E+05,
     + 1.2798E+05, 1.2863E+05, 1.2928E+05, 1.2994E+05, 1.3060E+05,
     +  1.3126E+05, 1.3193E+05, 1.3259E+05, 1.3326E+05, 1.3394E+05,
     + 1.3462E+05, 1.3530E+05, 1.3598E+05, 1.3666E+05, 1.3735E+05,
     +  1.3805E+05, 1.3874E+05, 1.3944E+05, 1.4014E+05, 1.4085E+05,
     + 1.4156E+05, 1.4227E+05, 1.4298E+05, 1.4370E+05, 1.4442E+05,
     +  1.4514E+05, 1.4587E+05, 1.4660E+05, 1.4734E+05, 1.4807E+05,
     + 1.4881E+05, 1.4956E+05, 1.5031E+05, 1.5106E+05, 1.5181E+05,
     +  1.5257E+05, 1.5333E+05, 1.5409E+05, 1.5486E+05, 1.5563E+05,
     + 1.5641E+05, 1.5718E+05, 1.5796E+05, 1.5875E+05, 1.5954E+05,
     +  1.6033E+05, 1.6112E+05, 1.6192E+05, 1.6272E+05, 1.6353E+05,
     + 1.6434E+05, 1.6515E+05, 1.6597E+05, 1.6679E+05, 1.6761E+05,
     +  1.6844E+05, 1.6927E+05, 1.7011E+05, 1.7094E+05, 1.7179E+05,
     + 1.7263E+05, 1.7348E+05, 1.7433E+05, 1.7519E+05, 1.7605E+05,
     +  1.7691E+05, 1.7778E+05, 1.7865E+05, 1.7953E+05, 1.8041E+05,
     + 1.8129E+05, 1.8218E+05, 1.8307E+05, 1.8397E+05, 1.8487E+05,
     +  1.8577E+05, 1.8668E+05, 1.8759E+05, 1.8850E+05, 1.8942E+05,
     + 1.9034E+05, 1.9127E+05, 1.9220E+05, 1.9314E+05, 1.9407E+05,
     +  1.9502E+05, 1.9596E+05, 1.9691E+05, 1.9787E+05, 1.9883E+05,
     + 1.9979E+05, 2.0076E+05, 2.0173E+05, 2.0271E+05, 2.0369E+05,
     +  2.0467E+05, 2.0566E+05, 2.0665E+05, 2.0765E+05, 2.0865E+05,
     + 2.0966E+05, 2.1067E+05, 2.1168E+05, 2.1270E+05, 2.1373E+05,
     +  2.1475E+05, 2.1578E+05, 2.1682E+05, 2.1786E+05, 2.1891E+05,
     + 2.1996E+05, 2.2101E+05, 2.2207E+05, 2.2313E+05, 2.2420E+05,
     +  2.2527E+05, 2.2635E+05, 2.2743E+05, 2.2852E+05, 2.2961E+05,
     + 2.3070E+05, 2.3180E+05, 2.3291E+05, 2.3402E+05, 2.3513E+05,
     +  2.3625E+05, 2.3737E+05, 2.3850E+05, 2.3963E+05, 2.4077E+05,
     + 2.4191E+05, 2.4306E+05, 2.4421E+05, 2.4537E+05, 2.4653E+05,
     +  2.4770E+05, 2.4887E+05, 2.5005E+05, 2.5123E+05, 2.5242E+05,
     + 2.5361E+05, 2.5480E+05, 2.5601E+05, 2.5721E+05, 2.5842E+05,
     +  2.5964E+05, 2.6086E+05, 2.6209E+05, 2.6332E+05, 2.6456E+05,
     + 2.6580E+05, 2.6705E+05, 2.6830E+05, 2.6956E+05, 2.7082E+05,
     +  2.7209E+05, 2.7336E+05, 2.7464E+05, 2.7593E+05, 2.7722E+05,
     + 2.7851E+05, 2.7981E+05, 2.8112E+05, 2.8243E+05, 2.8375E+05,
     +  2.8507E+05, 2.8640E+05, 2.8773E+05, 2.8907E+05, 2.9041E+05,
     + 2.9176E+05, 2.9312E+05, 2.9448E+05, 2.9585E+05, 2.9722E+05,
     +  2.9860E+05, 2.9998E+05, 3.0137E+05, 3.0276E+05, 3.0416E+05,
     + 3.0557E+05, 3.0698E+05, 3.0840E+05, 3.0982E+05, 3.1125E+05,
     +  3.1269E+05, 3.1413E+05, 3.1558E+05, 3.1703E+05, 3.1849E+05,
     + 3.1995E+05, 3.2143E+05, 3.2290E+05, 3.2439E+05, 3.2587E+05,
     +  3.2737E+05, 3.2887E+05, 3.3038E+05, 3.3189E+05, 3.3341E+05,
     + 3.3494E+05, 3.3647E+05, 3.3801E+05, 3.3955E+05, 3.4110E+05,
     +  3.4266E+05, 3.4422E+05, 3.4579E+05, 3.4737E+05, 3.4895E+05,
     + 3.5054E+05, 3.5213E+05, 3.5374E+05, 3.5534E+05, 3.5696E+05,
     +  3.5858E+05, 3.6021E+05, 3.6184E+05, 3.6348E+05, 3.6513E+05,
     + 3.6678E+05, 3.6844E+05, 3.7011E+05, 3.7178E+05, 3.7347E+05,
     +  3.7515E+05, 3.7685E+05, 3.7855E+05, 3.8026E+05, 3.8197E+05,
     + 3.8369E+05, 3.8542E+05, 3.8715E+05, 3.8890E+05, 3.9065E+05,
     +  3.9240E+05, 3.9416E+05, 3.9593E+05, 3.9771E+05, 3.9950E+05,
     + 4.0129E+05, 4.0309E+05, 4.0489E+05, 4.0670E+05, 4.0852E+05,
     +  4.1035E+05, 4.1218E+05, 4.1403E+05, 4.1587E+05, 4.1773E+05,
     + 4.1959E+05, 4.2146E+05, 4.2334E+05, 4.2523E+05, 4.2712E+05,
     +  4.2902E+05, 4.3093E+05, 4.3285E+05, 4.3477E+05, 4.3670E+05,
     + 4.3864E+05, 4.4058E+05, 4.4254E+05, 4.4450E+05, 4.4647E+05,
     +  4.4844E+05, 4.5043E+05, 4.5242E+05, 4.5442E+05, 4.5643E+05,
     + 4.5844E+05, 4.6047E+05, 4.6250E+05, 4.6454E+05, 4.6658E+05,
     +  4.6864E+05, 4.7070E+05, 4.7277E+05, 4.7485E+05, 4.7694E+05,
     + 4.7904E+05, 4.8114E+05, 4.8325E+05, 4.8537E+05, 4.8750E+05,
     +  4.8964E+05, 4.9178E+05, 4.9393E+05, 4.9610E+05, 4.9827E+05,
     + 5.0044E+05, 5.0263E+05, 5.0483E+05, 5.0703E+05, 5.0924E+05,
     +  5.1146E+05, 5.1369E+05, 5.1593E+05, 5.1817E+05, 5.2043E+05,
     + 5.2269E+05, 5.2497E+05, 5.2725E+05, 5.2954E+05, 5.3184E+05,
     +  5.3414E+05, 5.3646E+05, 5.3878E+05, 5.4112E+05, 5.4346E+05,
     + 5.4581E+05, 5.4818E+05, 5.5055E+05, 5.5292E+05, 5.5531E+05,
     +  5.5771E+05, 5.6012E+05, 5.6253E+05, 5.6496E+05, 5.6739E+05,
     + 5.6983E+05, 5.7229E+05, 5.7475E+05, 5.7722E+05, 5.7970E+05,
     +  5.8219E+05, 5.8469E+05, 5.8720E+05, 5.8972E+05, 5.9225E+05,
     + 5.9479E+05, 5.9733E+05, 5.9989E+05, 6.0246E+05, 6.0503E+05,
     +  6.0762E+05, 6.1022E+05, 6.1282E+05, 6.1544E+05, 6.1806E+05,
     + 6.2070E+05, 6.2334E+05, 6.2600E+05, 6.2866E+05, 6.3134E+05,
     +  6.3402E+05, 6.3672E+05, 6.3942E+05, 6.4214E+05, 6.4487E+05,
     + 6.4760E+05, 6.5035E+05, 6.5310E+05, 6.5587E+05, 6.5865E+05,
     +  6.6144E+05, 6.6423E+05, 6.6704E+05, 6.6986E+05, 6.7269E+05,
     + 6.7553E+05, 6.7838E+05, 6.8124E+05, 6.8412E+05, 6.8700E+05,
     +  6.8989E+05, 6.9279E+05, 6.9571E+05, 6.9864E+05, 7.0157E+05,
     + 7.0452E+05, 7.0748E+05, 7.1045E+05, 7.1343E+05, 7.1642E+05,
     +  7.1942E+05, 7.2243E+05, 7.2546E+05, 7.2849E+05, 7.3154E+05,
     + 7.3460E+05, 7.3767E+05, 7.4075E+05, 7.4384E+05, 7.4695E+05,
     +  7.5006E+05, 7.5319E+05, 7.5633E+05, 7.5948E+05, 7.6264E+05,
     + 7.6581E+05, 7.6899E+05, 7.7219E+05, 7.7540E+05, 7.7862E+05,
     +  7.8185E+05, 7.8509E+05, 7.8835E+05, 7.9161E+05, 7.9489E+05,
     + 7.9818E+05, 8.0149E+05, 8.0480E+05, 8.0813E+05, 8.1147E+05,
     +  8.1482E+05, 8.1818E+05, 8.2156E+05, 8.2494E+05, 8.2835E+05,
     + 8.3176E+05, 8.3518E+05, 8.3862E+05, 8.4207E+05, 8.4553E+05,
     +  8.4901E+05, 8.5250E+05, 8.5600E+05, 8.5951E+05, 8.6304E+05,
     + 8.6657E+05, 8.7012E+05, 8.7369E+05, 8.7726E+05, 8.8085E+05,
     +  8.8446E+05, 8.8807E+05, 8.9170E+05, 8.9534E+05, 8.9900E+05,
     + 9.0267E+05, 9.0635E+05, 9.1004E+05, 9.1375E+05, 9.1747E+05,
     +  9.2121E+05, 9.2495E+05, 9.2872E+05, 9.3249E+05, 9.3628E+05,
     + 9.4008E+05, 9.4390E+05, 9.4772E+05, 9.5157E+05, 9.5542E+05,
     +  9.5929E+05, 9.6318E+05, 9.6708E+05, 9.7099E+05, 9.7491E+05,
     + 9.7885E+05, 9.8281E+05, 9.8678E+05, 9.9076E+05, 9.9475E+05,
     +  9.9876E+05, 1.0028E+06, 1.0068E+06, 1.0109E+06, 1.0149E+06,
     + 1.0190E+06, 1.0231E+06, 1.0272E+06, 1.0314E+06, 1.0355E+06,
     +  1.0397E+06, 1.0438E+06, 1.0480E+06, 1.0522E+06, 1.0564E+06,
     + 1.0607E+06, 1.0649E+06, 1.0692E+06, 1.0734E+06, 1.0777E+06,
     +  1.0820E+06, 1.0863E+06, 1.0907E+06, 1.0950E+06, 1.0994E+06,
     + 1.1038E+06, 1.1082E+06, 1.1126E+06, 1.1170E+06, 1.1214E+06,
     +  1.1259E+06, 1.1304E+06, 1.1349E+06, 1.1394E+06, 1.1439E+06,
     + 1.1484E+06, 1.1530E+06, 1.1575E+06, 1.1621E+06, 1.1667E+06,
     +  1.1713E+06, 1.1760E+06, 1.1806E+06, 1.1853E+06, 1.1899E+06,
     + 1.1946E+06, 1.1994E+06, 1.2041E+06, 1.2088E+06, 1.2136E+06,
     +  1.2184E+06, 1.2232E+06, 1.2280E+06, 1.2328E+06, 1.2376E+06,
     + 1.2425E+06, 1.2474E+06, 1.2523E+06, 1.2572E+06, 1.2621E+06,
     +  1.2671E+06, 1.2720E+06, 1.2770E+06, 1.2820E+06, 1.2870E+06,
     + 1.2920E+06, 1.2971E+06, 1.3022E+06, 1.3072E+06, 1.3123E+06,
     +  1.3175E+06, 1.3226E+06, 1.3278E+06, 1.3329E+06, 1.3381E+06,
     + 1.3433E+06, 1.3485E+06, 1.3538E+06, 1.3590E+06, 1.3643E+06,
     +  1.3696E+06, 1.3749E+06, 1.3803E+06, 1.3856E+06, 1.3910E+06,
     + 1.3964E+06, 1.4018E+06, 1.4072E+06, 1.4126E+06, 1.4181E+06,
     +  1.4236E+06, 1.4291E+06, 1.4346E+06, 1.4401E+06, 1.4457E+06,
     + 1.4513E+06, 1.4569E+06, 1.4625E+06, 1.4681E+06, 1.4737E+06,
     +  1.4794E+06, 1.4851E+06, 1.4908E+06, 1.4965E+06, 1.5023E+06,
     + 1.5080E+06, 1.5138E+06, 1.5196E+06, 1.5254E+06, 1.5313E+06,
     +  1.5372E+06, 1.5430E+06, 1.5489E+06, 1.5549E+06, 1.5608E+06,
     + 1.5668E+06, 1.5727E+06, 1.5787E+06, 1.5848E+06, 1.5908E+06,
     +  1.5969E+06, 1.6030E+06, 1.6091E+06, 1.6152E+06, 1.6213E+06,
     + 1.6275E+06, 1.6337E+06, 1.6399E+06, 1.6461E+06, 1.6523E+06,
     +  1.6586E+06, 1.6649E+06, 1.6712E+06, 1.6775E+06, 1.6839E+06,
     + 1.6903E+06, 1.6967E+06, 1.7031E+06, 1.7095E+06, 1.7160E+06,
     +  1.7224E+06, 1.7289E+06, 1.7355E+06, 1.7420E+06, 1.7486E+06,
     + 1.7552E+06, 1.7618E+06, 1.7684E+06, 1.7751E+06, 1.7817E+06,
     +  1.7884E+06, 1.7952E+06, 1.8019E+06, 1.8087E+06, 1.8154E+06,
     + 1.8223E+06, 1.8291E+06, 1.8359E+06, 1.8428E+06, 1.8497E+06,
     +  1.8566E+06, 1.8636E+06, 1.8705E+06, 1.8775E+06, 1.8845E+06,
     + 1.8916E+06, 1.8986E+06, 1.9057E+06, 1.9128E+06, 1.9199E+06,
     +  1.9271E+06, 1.9343E+06, 1.9415E+06, 1.9487E+06, 1.9559E+06,
     + 1.9632E+06, 1.9705E+06, 1.9778E+06, 1.9852E+06, 1.9925E+06,
     +  1.9999E+06, 2.0073E+06, 2.0148E+06, 2.0222E+06, 2.0297E+06,
     + 2.0372E+06, 2.0447E+06, 2.0523E+06, 2.0599E+06, 2.0675E+06,
     +  2.0751E+06, 2.0828E+06, 2.0905E+06, 2.0982E+06, 2.1059E+06,
     + 2.1136E+06, 2.1214E+06, 2.1292E+06, 2.1371E+06, 2.1449E+06,
     +  2.1528E+06, 2.1607E+06, 2.1686E+06, 2.1766E+06, 2.1846E+06,
     + 2.1926E+06, 2.2006E+06, 2.2087E+06, 2.2168E+06, 2.2249E+06,
     +  2.2330E+06, 2.2412E+06, 2.2494E+06, 2.2576E+06, 2.2659E+06,
     + 2.2741E+06, 2.2824E+06, 2.2907E+06, 2.2991E+06, 2.3075E+06,
     +  2.3159E+06, 2.3243E+06, 2.3328E+06, 2.3413E+06, 2.3498E+06,
     + 2.3583E+06, 2.3669E+06, 2.3755E+06, 2.3841E+06, 2.3927E+06,
     +  2.4014E+06, 2.4101E+06, 2.4188E+06, 2.4276E+06, 2.4364E+06,
     + 2.4452E+06, 2.4540E+06, 2.4629E+06, 2.4718E+06, 2.4807E+06,
     +  2.4897E+06, 2.4987E+06, 2.5077E+06, 2.5167E+06, 2.5258E+06,
     + 2.5349E+06, 2.5440E+06, 2.5532E+06, 2.5624E+06, 2.5716E+06,
     +  2.5808E+06, 2.5901E+06, 2.5994E+06, 2.6087E+06, 2.6181E+06,
     + 2.6275E+06, 2.6369E+06, 2.6463E+06, 2.6558E+06, 2.6653E+06,
     +  2.6749E+06, 2.6844E+06, 2.6940E+06, 2.7036E+06, 2.7133E+06,
     + 2.7230E+06, 2.7327E+06, 2.7425E+06, 2.7522E+06, 2.7620E+06,
     +  2.7719E+06, 2.7818E+06, 2.7917E+06, 2.8016E+06, 2.8116E+06,
     + 2.8216E+06, 2.8316E+06, 2.8416E+06, 2.8517E+06, 2.8618E+06,
     +  2.8720E+06, 2.8822E+06, 2.8924E+06, 2.9026E+06, 2.9129E+06,
     + 2.9232E+06, 2.9336E+06, 2.9439E+06, 2.9543E+06, 2.9648E+06,
     +  2.9753E+06, 2.9858E+06, 2.9963E+06, 3.0069E+06, 3.0175E+06,
     + 3.0281E+06, 3.0388E+06, 3.0495E+06, 3.0602E+06, 3.0710E+06,
     +  3.0818E+06, 3.0926E+06, 3.1035E+06, 3.1144E+06, 3.1253E+06,
     + 3.1363E+06, 3.1473E+06, 3.1583E+06, 3.1694E+06, 3.1805E+06,
     +  3.1916E+06, 3.2028E+06, 3.2140E+06, 3.2252E+06, 3.2365E+06,
     + 3.2478E+06, 3.2591E+06, 3.2705E+06, 3.2819E+06, 3.2934E+06,
     +  3.3048E+06, 3.3164E+06, 3.3279E+06, 3.3395E+06, 3.3511E+06,
     + 3.3628E+06, 3.3745E+06, 3.3862E+06, 3.3980E+06, 3.4098E+06,
     +  3.4216E+06, 3.4335E+06, 3.4454E+06, 3.4573E+06, 3.4693E+06,
     + 3.4813E+06, 3.4934E+06, 3.5055E+06, 3.5176E+06, 3.5298E+06,
     +  3.5420E+06, 3.5542E+06, 3.5665E+06, 3.5788E+06, 3.5911E+06,
     + 3.6035E+06, 3.6159E+06, 3.6284E+06, 3.6409E+06, 3.6534E+06,
     +  3.6660E+06, 3.6786E+06, 3.6913E+06, 3.7040E+06, 3.7167E+06,
     + 3.7295E+06, 3.7423E+06, 3.7551E+06, 3.7680E+06, 3.7809E+06,
     +  3.7939E+06, 3.8069E+06, 3.8199E+06, 3.8330E+06, 3.8461E+06,
     + 3.8592E+06, 3.8724E+06, 3.8857E+06, 3.8989E+06, 3.9122E+06,
     +  3.9256E+06, 3.9390E+06, 3.9524E+06, 3.9659E+06, 3.9794E+06,
     + 3.9930E+06, 4.0066E+06, 4.0202E+06, 4.0339E+06, 4.0476E+06,
     +  4.0613E+06, 4.0751E+06, 4.0890E+06, 4.1028E+06, 4.1168E+06,
     + 4.1307E+06, 4.1447E+06, 4.1588E+06, 4.1729E+06, 4.1870E+06,
     +  4.2012E+06, 4.2154E+06, 4.2296E+06, 4.2439E+06, 4.2583E+06,
     + 4.2726E+06, 4.2871E+06, 4.3015E+06, 4.3160E+06, 4.3306E+06,
     +  4.3452E+06, 4.3598E+06, 4.3745E+06, 4.3892E+06, 4.4040E+06,
     + 4.4188E+06, 4.4336E+06, 4.4485E+06, 4.4635E+06, 4.4785E+06,
     +  4.4935E+06, 4.5086E+06, 4.5237E+06, 4.5389E+06, 4.5541E+06,
     + 4.5693E+06, 4.5846E+06, 4.6000E+06, 4.6153E+06, 4.6308E+06,
     +  4.6463E+06, 4.6618E+06, 4.6773E+06, 4.6930E+06, 4.7086E+06,
     + 4.7243E+06, 4.7401E+06, 4.7559E+06, 4.7717E+06, 4.7876E+06,
     +  4.8035E+06, 4.8195E+06, 4.8355E+06, 4.8516E+06, 4.8677E+06,
     + 4.8839E+06, 4.9001E+06, 4.9164E+06, 4.9327E+06, 4.9490E+06,
     +  4.9654E+06, 4.9819E+06, 4.9984E+06, 5.0149E+06, 5.0315E+06,
     + 5.0482E+06, 5.0649E+06, 5.0816E+06, 5.0984E+06, 5.1152E+06,
     +  5.1321E+06, 5.1491E+06, 5.1660E+06, 5.1831E+06, 5.2001E+06,
     + 5.2173E+06, 5.2345E+06, 5.2517E+06, 5.2690E+06, 5.2863E+06,
     +  5.3037E+06, 5.3211E+06, 5.3386E+06, 5.3561E+06, 5.3737E+06,
     + 5.3913E+06, 5.4090E+06, 5.4267E+06, 5.4445E+06, 5.4623E+06,
     +  5.4802E+06, 5.4982E+06, 5.5162E+06, 5.5342E+06, 5.5523E+06,
     + 5.5704E+06, 5.5886E+06, 5.6069E+06, 5.6252E+06, 5.6435E+06,
     +  5.6619E+06, 5.6804E+06, 5.6989E+06, 5.7174E+06, 5.7360E+06,
     + 5.7547E+06, 5.7734E+06, 5.7922E+06, 5.8110E+06, 5.8299E+06,
     +  5.8488E+06, 5.8678E+06, 5.8869E+06, 5.9060E+06, 5.9251E+06,
     + 5.9443E+06, 5.9636E+06, 5.9829E+06, 6.0022E+06, 6.0217E+06,
     +  6.0411E+06, 6.0607E+06, 6.0803E+06, 6.0999E+06, 6.1196E+06,
     + 6.1394E+06, 6.1592E+06, 6.1790E+06, 6.1990E+06, 6.2189E+06,
     +  6.2390E+06, 6.2591E+06, 6.2792E+06, 6.2994E+06, 6.3197E+06,
     + 6.3400E+06, 6.3604E+06, 6.3808E+06, 6.4013E+06, 6.4218E+06,
     +  6.4424E+06, 6.4631E+06, 6.4838E+06, 6.5046E+06, 6.5254E+06,
     + 6.5463E+06, 6.5673E+06, 6.5883E+06, 6.6094E+06, 6.6305E+06,
     +  6.6517E+06, 6.6729E+06, 6.6942E+06, 6.7156E+06, 6.7370E+06,
     + 6.7585E+06, 6.7801E+06, 6.8017E+06, 6.8233E+06, 6.8451E+06,
     +  6.8669E+06, 6.8887E+06, 6.9106E+06, 6.9326E+06, 6.9546E+06,
     + 6.9767E+06, 6.9989E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       136
      data (QofT( 2,J),J=1,2408)/ 1.5198E+00, 4.1325E+00, 7.0369E+00,
     +  1.0378E+01, 1.4122E+01, 1.8232E+01, 2.2679E+01, 2.7440E+01,
     + 3.2496E+01, 3.7830E+01, 4.3429E+01, 4.9281E+01, 5.5374E+01,
     +  6.1701E+01, 6.8253E+01, 7.5021E+01, 8.2001E+01, 8.9184E+01,
     + 9.6567E+01, 1.0414E+02, 1.1986E+02, 1.3629E+02, 1.5341E+02,
     +  1.7120E+02, 1.8963E+02, 2.0867E+02, 2.2832E+02, 2.4854E+02,
     + 2.6934E+02, 2.9068E+02, 3.1255E+02, 3.3496E+02, 3.5787E+02,
     +  3.8128E+02, 4.0519E+02, 4.2957E+02, 4.5443E+02, 4.7974E+02,
     + 5.0552E+02, 5.3173E+02, 5.5839E+02, 5.8548E+02, 6.1300E+02,
     +  6.4093E+02, 6.6928E+02, 6.9803E+02, 7.2718E+02, 7.5673E+02,
     + 7.8667E+02, 8.1700E+02, 8.4771E+02, 8.7879E+02, 9.1025E+02,
     +  9.4207E+02, 9.7425E+02, 1.0068E+03, 1.0397E+03, 1.0730E+03,
     + 1.1066E+03, 1.1405E+03, 1.1748E+03, 1.2094E+03, 1.2444E+03,
     +  1.2797E+03, 1.3153E+03, 1.3512E+03, 1.3875E+03, 1.4241E+03,
     + 1.4610E+03, 1.4982E+03, 1.5357E+03, 1.5736E+03, 1.6117E+03,
     +  1.6502E+03, 1.6889E+03, 1.7280E+03, 1.7673E+03, 1.8070E+03,
     + 1.8469E+03, 1.8871E+03, 1.9277E+03, 1.9685E+03, 2.0096E+03,
     +  2.0509E+03, 2.0926E+03, 2.1345E+03, 2.1768E+03, 2.2192E+03,
     + 2.2620E+03, 2.3051E+03, 2.3484E+03, 2.3920E+03, 2.4358E+03,
     +  2.4800E+03, 2.5244E+03, 2.5690E+03, 2.6140E+03, 2.6591E+03,
     + 2.7046E+03, 2.7503E+03, 2.7963E+03, 2.8425E+03, 2.8890E+03,
     +  2.9358E+03, 2.9828E+03, 3.0300E+03, 3.0776E+03, 3.1253E+03,
     + 3.1734E+03, 3.2217E+03, 3.2702E+03, 3.3190E+03, 3.3680E+03,
     +  3.4173E+03, 3.4669E+03, 3.5167E+03, 3.5667E+03, 3.6170E+03,
     + 3.6676E+03, 3.7184E+03, 3.7694E+03, 3.8207E+03, 3.8722E+03,
     +  3.9240E+03, 3.9761E+03, 4.0284E+03, 4.0809E+03, 4.1337E+03,
     + 4.1868E+03, 4.2401E+03, 4.2937E+03, 4.3475E+03, 4.4016E+03,
     +  4.4559E+03, 4.5104E+03, 4.5653E+03, 4.6203E+03, 4.6757E+03,
     + 4.7313E+03, 4.7872E+03, 4.8432E+03, 4.8996E+03, 4.9563E+03,
     +  5.0131E+03, 5.0703E+03, 5.1277E+03, 5.1854E+03, 5.2433E+03,
     + 5.3016E+03, 5.3601E+03, 5.4188E+03, 5.4778E+03, 5.5371E+03,
     +  5.5967E+03, 5.6565E+03, 5.7166E+03, 5.7770E+03, 5.8377E+03,
     + 5.8986E+03, 5.9598E+03, 6.0214E+03, 6.0831E+03, 6.1452E+03,
     +  6.2076E+03, 6.2702E+03, 6.3332E+03, 6.3964E+03, 6.4599E+03,
     + 6.5237E+03, 6.5879E+03, 6.6523E+03, 6.7170E+03, 6.7820E+03,
     +  6.8474E+03, 6.9130E+03, 6.9790E+03, 7.0452E+03, 7.1118E+03,
     + 7.1786E+03, 7.2459E+03, 7.3134E+03, 7.3812E+03, 7.4494E+03,
     +  7.5179E+03, 7.5867E+03, 7.6559E+03, 7.7253E+03, 7.7951E+03,
     + 7.8652E+03, 7.9357E+03, 8.0065E+03, 8.0777E+03, 8.1492E+03,
     +  8.2211E+03, 8.2932E+03, 8.3658E+03, 8.4388E+03, 8.5120E+03,
     + 8.5856E+03, 8.6596E+03, 8.7340E+03, 8.8087E+03, 8.8838E+03,
     +  8.9592E+03, 9.0351E+03, 9.1113E+03, 9.1879E+03, 9.2649E+03,
     + 9.3423E+03, 9.4200E+03, 9.4981E+03, 9.5767E+03, 9.6556E+03,
     +  9.7350E+03, 9.8147E+03, 9.8947E+03, 9.9754E+03, 1.0056E+04,
     + 1.0138E+04, 1.0219E+04, 1.0302E+04, 1.0384E+04, 1.0467E+04,
     +  1.0551E+04, 1.0635E+04, 1.0719E+04, 1.0804E+04, 1.0889E+04,
     + 1.0974E+04, 1.1061E+04, 1.1147E+04, 1.1234E+04, 1.1322E+04,
     +  1.1409E+04, 1.1498E+04, 1.1587E+04, 1.1676E+04, 1.1766E+04,
     + 1.1856E+04, 1.1946E+04, 1.2038E+04, 1.2129E+04, 1.2221E+04,
     +  1.2314E+04, 1.2407E+04, 1.2501E+04, 1.2595E+04, 1.2689E+04,
     + 1.2785E+04, 1.2880E+04, 1.2976E+04, 1.3073E+04, 1.3170E+04,
     +  1.3268E+04, 1.3366E+04, 1.3465E+04, 1.3564E+04, 1.3664E+04,
     + 1.3764E+04, 1.3865E+04, 1.3966E+04, 1.4068E+04, 1.4171E+04,
     +  1.4274E+04, 1.4377E+04, 1.4481E+04, 1.4586E+04, 1.4691E+04,
     + 1.4797E+04, 1.4904E+04, 1.5011E+04, 1.5118E+04, 1.5227E+04,
     +  1.5335E+04, 1.5445E+04, 1.5555E+04, 1.5665E+04, 1.5776E+04,
     + 1.5888E+04, 1.6001E+04, 1.6114E+04, 1.6227E+04, 1.6341E+04,
     +  1.6456E+04, 1.6572E+04, 1.6688E+04, 1.6805E+04, 1.6922E+04,
     + 1.7040E+04, 1.7159E+04, 1.7278E+04, 1.7398E+04, 1.7519E+04,
     +  1.7640E+04, 1.7762E+04, 1.7885E+04, 1.8008E+04, 1.8132E+04,
     + 1.8257E+04, 1.8383E+04, 1.8509E+04, 1.8636E+04, 1.8763E+04,
     +  1.8891E+04, 1.9020E+04, 1.9150E+04, 1.9280E+04, 1.9411E+04,
     + 1.9543E+04, 1.9676E+04, 1.9809E+04, 1.9943E+04, 2.0078E+04,
     +  2.0213E+04, 2.0349E+04, 2.0486E+04, 2.0624E+04, 2.0763E+04,
     + 2.0902E+04, 2.1042E+04, 2.1183E+04, 2.1325E+04, 2.1467E+04,
     +  2.1610E+04, 2.1754E+04, 2.1899E+04, 2.2045E+04, 2.2191E+04,
     + 2.2338E+04, 2.2486E+04, 2.2635E+04, 2.2785E+04, 2.2935E+04,
     +  2.3087E+04, 2.3239E+04, 2.3392E+04, 2.3546E+04, 2.3701E+04,
     + 2.3857E+04, 2.4013E+04, 2.4170E+04, 2.4329E+04, 2.4488E+04,
     +  2.4648E+04, 2.4809E+04, 2.4971E+04, 2.5133E+04, 2.5297E+04,
     + 2.5462E+04, 2.5627E+04, 2.5793E+04, 2.5961E+04, 2.6129E+04,
     +  2.6298E+04, 2.6468E+04, 2.6639E+04, 2.6811E+04, 2.6984E+04,
     + 2.7158E+04, 2.7333E+04, 2.7509E+04, 2.7685E+04, 2.7863E+04,
     +  2.8042E+04, 2.8222E+04, 2.8403E+04, 2.8584E+04, 2.8767E+04,
     + 2.8951E+04, 2.9136E+04, 2.9322E+04, 2.9509E+04, 2.9696E+04,
     +  2.9885E+04, 3.0075E+04, 3.0266E+04, 3.0458E+04, 3.0652E+04,
     + 3.0846E+04, 3.1041E+04, 3.1238E+04, 3.1435E+04, 3.1634E+04,
     +  3.1833E+04, 3.2034E+04, 3.2236E+04, 3.2439E+04, 3.2643E+04,
     + 3.2848E+04, 3.3055E+04, 3.3262E+04, 3.3471E+04, 3.3681E+04,
     +  3.3892E+04, 3.4104E+04, 3.4317E+04, 3.4532E+04, 3.4747E+04,
     + 3.4964E+04, 3.5183E+04, 3.5402E+04, 3.5622E+04, 3.5844E+04,
     +  3.6067E+04, 3.6291E+04, 3.6516E+04, 3.6743E+04, 3.6971E+04,
     + 3.7200E+04, 3.7430E+04, 3.7662E+04, 3.7895E+04, 3.8129E+04,
     +  3.8364E+04, 3.8601E+04, 3.8839E+04, 3.9078E+04, 3.9319E+04,
     + 3.9561E+04, 3.9804E+04, 4.0049E+04, 4.0295E+04, 4.0542E+04,
     +  4.0791E+04, 4.1040E+04, 4.1292E+04, 4.1545E+04, 4.1799E+04,
     + 4.2054E+04, 4.2311E+04, 4.2569E+04, 4.2829E+04, 4.3090E+04,
     +  4.3353E+04, 4.3617E+04, 4.3882E+04, 4.4148E+04, 4.4417E+04,
     + 4.4687E+04, 4.4958E+04, 4.5230E+04, 4.5505E+04, 4.5780E+04,
     +  4.6057E+04, 4.6336E+04, 4.6616E+04, 4.6898E+04, 4.7181E+04,
     + 4.7465E+04, 4.7752E+04, 4.8039E+04, 4.8329E+04, 4.8619E+04,
     +  4.8912E+04, 4.9206E+04, 4.9501E+04, 4.9799E+04, 5.0098E+04,
     + 5.0398E+04, 5.0700E+04, 5.1004E+04, 5.1309E+04, 5.1616E+04,
     +  5.1924E+04, 5.2235E+04, 5.2546E+04, 5.2860E+04, 5.3175E+04,
     + 5.3492E+04, 5.3811E+04, 5.4131E+04, 5.4453E+04, 5.4777E+04,
     +  5.5102E+04, 5.5429E+04, 5.5758E+04, 5.6089E+04, 5.6422E+04,
     + 5.6756E+04, 5.7092E+04, 5.7430E+04, 5.7769E+04, 5.8111E+04,
     +  5.8454E+04, 5.8799E+04, 5.9146E+04, 5.9495E+04, 5.9845E+04,
     + 6.0198E+04, 6.0552E+04, 6.0908E+04, 6.1266E+04, 6.1626E+04,
     +  6.1988E+04, 6.2352E+04, 6.2717E+04, 6.3085E+04, 6.3455E+04,
     + 6.3826E+04, 6.4200E+04, 6.4575E+04, 6.4952E+04, 6.5332E+04,
     +  6.5713E+04, 6.6096E+04, 6.6482E+04, 6.6870E+04, 6.7259E+04,
     + 6.7651E+04, 6.8044E+04, 6.8440E+04, 6.8837E+04, 6.9237E+04,
     +  6.9639E+04, 7.0043E+04, 7.0449E+04, 7.0857E+04, 7.1268E+04,
     + 7.1680E+04, 7.2095E+04, 7.2512E+04, 7.2931E+04, 7.3352E+04,
     +  7.3775E+04, 7.4201E+04, 7.4629E+04, 7.5059E+04, 7.5491E+04,
     + 7.5925E+04, 7.6362E+04, 7.6801E+04, 7.7242E+04, 7.7686E+04,
     +  7.8132E+04, 7.8580E+04, 7.9030E+04, 7.9483E+04, 7.9938E+04,
     + 8.0396E+04, 8.0856E+04, 8.1318E+04, 8.1783E+04, 8.2250E+04,
     +  8.2719E+04, 8.3191E+04, 8.3665E+04, 8.4142E+04, 8.4621E+04,
     + 8.5103E+04, 8.5587E+04, 8.6073E+04, 8.6562E+04, 8.7054E+04,
     +  8.7548E+04, 8.8045E+04, 8.8544E+04, 8.9046E+04, 8.9550E+04,
     + 9.0057E+04, 9.0566E+04, 9.1078E+04, 9.1593E+04, 9.2110E+04,
     +  9.2630E+04, 9.3152E+04, 9.3678E+04, 9.4206E+04, 9.4736E+04,
     + 9.5270E+04, 9.5806E+04, 9.6344E+04, 9.6885E+04, 9.7430E+04,
     +  9.7977E+04, 9.8526E+04, 9.9079E+04, 9.9634E+04, 1.0019E+05,
     + 1.0075E+05, 1.0132E+05, 1.0188E+05, 1.0245E+05, 1.0302E+05,
     +  1.0360E+05, 1.0418E+05, 1.0476E+05, 1.0534E+05, 1.0593E+05,
     + 1.0652E+05, 1.0711E+05, 1.0771E+05, 1.0831E+05, 1.0891E+05,
     +  1.0951E+05, 1.1012E+05, 1.1073E+05, 1.1135E+05, 1.1196E+05,
     + 1.1258E+05, 1.1321E+05, 1.1383E+05, 1.1446E+05, 1.1509E+05,
     +  1.1573E+05, 1.1637E+05, 1.1701E+05, 1.1765E+05, 1.1830E+05,
     + 1.1895E+05, 1.1961E+05, 1.2027E+05, 1.2093E+05, 1.2159E+05,
     +  1.2226E+05, 1.2293E+05, 1.2361E+05, 1.2428E+05, 1.2497E+05,
     + 1.2565E+05, 1.2634E+05, 1.2703E+05, 1.2772E+05, 1.2842E+05,
     +  1.2912E+05, 1.2983E+05, 1.3054E+05, 1.3125E+05, 1.3196E+05,
     + 1.3268E+05, 1.3341E+05, 1.3413E+05, 1.3486E+05, 1.3559E+05,
     +  1.3633E+05, 1.3707E+05, 1.3782E+05, 1.3856E+05, 1.3931E+05,
     + 1.4007E+05, 1.4083E+05, 1.4159E+05, 1.4236E+05, 1.4313E+05,
     +  1.4390E+05, 1.4468E+05, 1.4546E+05, 1.4624E+05, 1.4703E+05,
     + 1.4782E+05, 1.4862E+05, 1.4942E+05, 1.5022E+05, 1.5103E+05,
     +  1.5184E+05, 1.5266E+05, 1.5348E+05, 1.5430E+05, 1.5513E+05,
     + 1.5596E+05, 1.5680E+05, 1.5764E+05, 1.5848E+05, 1.5933E+05,
     +  1.6018E+05, 1.6104E+05, 1.6190E+05, 1.6276E+05, 1.6363E+05,
     + 1.6450E+05, 1.6538E+05, 1.6626E+05, 1.6714E+05, 1.6803E+05,
     +  1.6893E+05, 1.6982E+05, 1.7073E+05, 1.7163E+05, 1.7254E+05,
     + 1.7346E+05, 1.7438E+05, 1.7530E+05, 1.7623E+05, 1.7716E+05,
     +  1.7810E+05, 1.7904E+05, 1.7999E+05, 1.8094E+05, 1.8189E+05,
     + 1.8285E+05, 1.8382E+05, 1.8478E+05, 1.8576E+05, 1.8674E+05,
     +  1.8772E+05, 1.8870E+05, 1.8970E+05, 1.9069E+05, 1.9169E+05,
     + 1.9270E+05, 1.9371E+05, 1.9473E+05, 1.9575E+05, 1.9677E+05,
     +  1.9780E+05, 1.9883E+05, 1.9987E+05, 2.0092E+05, 2.0197E+05,
     + 2.0302E+05, 2.0408E+05, 2.0514E+05, 2.0621E+05, 2.0729E+05,
     +  2.0837E+05, 2.0945E+05, 2.1054E+05, 2.1163E+05, 2.1273E+05,
     + 2.1384E+05, 2.1495E+05, 2.1606E+05, 2.1718E+05, 2.1831E+05,
     +  2.1944E+05, 2.2057E+05, 2.2171E+05, 2.2286E+05, 2.2401E+05,
     + 2.2517E+05, 2.2633E+05, 2.2750E+05, 2.2867E+05, 2.2985E+05,
     +  2.3103E+05, 2.3222E+05, 2.3341E+05, 2.3461E+05, 2.3582E+05,
     + 2.3703E+05, 2.3825E+05, 2.3947E+05, 2.4070E+05, 2.4193E+05,
     +  2.4317E+05, 2.4442E+05, 2.4567E+05, 2.4692E+05, 2.4818E+05,
     + 2.4945E+05, 2.5073E+05, 2.5201E+05, 2.5329E+05, 2.5458E+05,
     +  2.5588E+05, 2.5718E+05, 2.5849E+05, 2.5981E+05, 2.6113E+05,
     + 2.6246E+05, 2.6379E+05, 2.6513E+05, 2.6647E+05, 2.6782E+05,
     +  2.6918E+05, 2.7054E+05, 2.7191E+05, 2.7329E+05, 2.7467E+05,
     + 2.7606E+05, 2.7746E+05, 2.7886E+05, 2.8026E+05, 2.8168E+05,
     +  2.8310E+05, 2.8452E+05, 2.8596E+05, 2.8740E+05, 2.8884E+05,
     + 2.9029E+05, 2.9175E+05, 2.9322E+05, 2.9469E+05, 2.9617E+05,
     +  2.9765E+05, 2.9915E+05, 3.0064E+05, 3.0215E+05, 3.0366E+05,
     + 3.0518E+05, 3.0671E+05, 3.0824E+05, 3.0978E+05, 3.1132E+05,
     +  3.1288E+05, 3.1444E+05, 3.1600E+05, 3.1758E+05, 3.1916E+05,
     + 3.2075E+05, 3.2234E+05, 3.2394E+05, 3.2555E+05, 3.2717E+05,
     +  3.2879E+05, 3.3042E+05, 3.3206E+05, 3.3370E+05, 3.3536E+05,
     + 3.3702E+05, 3.3868E+05, 3.4036E+05, 3.4204E+05, 3.4373E+05,
     +  3.4543E+05, 3.4713E+05, 3.4884E+05, 3.5056E+05, 3.5229E+05,
     + 3.5402E+05, 3.5576E+05, 3.5751E+05, 3.5927E+05, 3.6103E+05,
     +  3.6281E+05, 3.6459E+05, 3.6637E+05, 3.6817E+05, 3.6997E+05,
     + 3.7179E+05, 3.7361E+05, 3.7543E+05, 3.7727E+05, 3.7911E+05,
     +  3.8096E+05, 3.8282E+05, 3.8469E+05, 3.8657E+05, 3.8845E+05,
     + 3.9034E+05, 3.9224E+05, 3.9415E+05, 3.9607E+05, 3.9800E+05,
     +  3.9993E+05, 4.0187E+05, 4.0382E+05, 4.0578E+05, 4.0775E+05,
     + 4.0972E+05, 4.1171E+05, 4.1370E+05, 4.1570E+05, 4.1771E+05,
     +  4.1973E+05, 4.2176E+05, 4.2379E+05, 4.2584E+05, 4.2789E+05,
     + 4.2996E+05, 4.3203E+05, 4.3411E+05, 4.3620E+05, 4.3830E+05,
     +  4.4040E+05, 4.4252E+05, 4.4464E+05, 4.4678E+05, 4.4892E+05,
     + 4.5107E+05, 4.5324E+05, 4.5541E+05, 4.5759E+05, 4.5978E+05,
     +  4.6198E+05, 4.6418E+05, 4.6640E+05, 4.6863E+05, 4.7087E+05,
     + 4.7311E+05, 4.7537E+05, 4.7763E+05, 4.7991E+05, 4.8219E+05,
     +  4.8449E+05, 4.8679E+05, 4.8910E+05, 4.9143E+05, 4.9376E+05,
     + 4.9610E+05, 4.9845E+05, 5.0082E+05, 5.0319E+05, 5.0557E+05,
     +  5.0797E+05, 5.1037E+05, 5.1278E+05, 5.1521E+05, 5.1764E+05,
     + 5.2008E+05, 5.2254E+05, 5.2500E+05, 5.2747E+05, 5.2996E+05,
     +  5.3245E+05, 5.3496E+05, 5.3748E+05, 5.4000E+05, 5.4254E+05,
     + 5.4509E+05, 5.4765E+05, 5.5022E+05, 5.5280E+05, 5.5539E+05,
     +  5.5799E+05, 5.6060E+05, 5.6322E+05, 5.6586E+05, 5.6850E+05,
     + 5.7116E+05, 5.7383E+05, 5.7650E+05, 5.7919E+05, 5.8189E+05,
     +  5.8460E+05, 5.8733E+05, 5.9006E+05, 5.9281E+05, 5.9556E+05,
     + 5.9833E+05, 6.0111E+05, 6.0390E+05, 6.0670E+05, 6.0952E+05,
     +  6.1234E+05, 6.1518E+05, 6.1803E+05, 6.2089E+05, 6.2376E+05,
     + 6.2664E+05, 6.2954E+05, 6.3245E+05, 6.3537E+05, 6.3830E+05,
     +  6.4124E+05, 6.4420E+05, 6.4717E+05, 6.5015E+05, 6.5314E+05,
     + 6.5614E+05, 6.5916E+05, 6.6219E+05, 6.6523E+05, 6.6828E+05,
     +  6.7135E+05, 6.7443E+05, 6.7752E+05, 6.8062E+05, 6.8374E+05,
     + 6.8687E+05, 6.9001E+05, 6.9317E+05, 6.9633E+05, 6.9951E+05,
     +  7.0271E+05, 7.0591E+05, 7.0913E+05, 7.1236E+05, 7.1561E+05,
     + 7.1887E+05, 7.2214E+05, 7.2542E+05, 7.2872E+05, 7.3203E+05,
     +  7.3535E+05, 7.3869E+05, 7.4204E+05, 7.4541E+05, 7.4879E+05,
     + 7.5218E+05, 7.5558E+05, 7.5900E+05, 7.6243E+05, 7.6588E+05,
     +  7.6934E+05, 7.7282E+05, 7.7630E+05, 7.7981E+05, 7.8332E+05,
     + 7.8685E+05, 7.9040E+05, 7.9395E+05, 7.9753E+05, 8.0111E+05,
     +  8.0471E+05, 8.0833E+05, 8.1196E+05, 8.1560E+05, 8.1926E+05,
     + 8.2294E+05, 8.2662E+05, 8.3033E+05, 8.3404E+05, 8.3778E+05,
     +  8.4152E+05, 8.4528E+05, 8.4906E+05, 8.5285E+05, 8.5666E+05,
     + 8.6048E+05, 8.6432E+05, 8.6817E+05, 8.7204E+05, 8.7592E+05,
     +  8.7981E+05, 8.8373E+05, 8.8765E+05, 8.9160E+05, 8.9556E+05,
     + 8.9953E+05, 9.0352E+05, 9.0753E+05, 9.1155E+05, 9.1559E+05,
     +  9.1964E+05, 9.2371E+05, 9.2780E+05, 9.3190E+05, 9.3601E+05,
     + 9.4015E+05, 9.4430E+05, 9.4846E+05, 9.5265E+05, 9.5684E+05,
     +  9.6106E+05, 9.6529E+05, 9.6954E+05, 9.7380E+05, 9.7808E+05,
     + 9.8238E+05, 9.8669E+05, 9.9102E+05, 9.9537E+05, 9.9974E+05,
     +  1.0041E+06, 1.0085E+06, 1.0129E+06, 1.0174E+06, 1.0218E+06,
     + 1.0263E+06, 1.0308E+06, 1.0353E+06, 1.0398E+06, 1.0443E+06,
     +  1.0489E+06, 1.0534E+06, 1.0580E+06, 1.0626E+06, 1.0673E+06,
     + 1.0719E+06, 1.0766E+06, 1.0812E+06, 1.0859E+06, 1.0907E+06,
     +  1.0954E+06, 1.1001E+06, 1.1049E+06, 1.1097E+06, 1.1145E+06,
     + 1.1193E+06, 1.1242E+06, 1.1290E+06, 1.1339E+06, 1.1388E+06,
     +  1.1437E+06, 1.1487E+06, 1.1536E+06, 1.1586E+06, 1.1636E+06,
     + 1.1686E+06, 1.1736E+06, 1.1787E+06, 1.1837E+06, 1.1888E+06,
     +  1.1939E+06, 1.1991E+06, 1.2042E+06, 1.2094E+06, 1.2145E+06,
     + 1.2198E+06, 1.2250E+06, 1.2302E+06, 1.2355E+06, 1.2408E+06,
     +  1.2461E+06, 1.2514E+06, 1.2567E+06, 1.2621E+06, 1.2675E+06,
     + 1.2729E+06, 1.2783E+06, 1.2838E+06, 1.2892E+06, 1.2947E+06,
     +  1.3002E+06, 1.3057E+06, 1.3113E+06, 1.3169E+06, 1.3225E+06,
     + 1.3281E+06, 1.3337E+06, 1.3394E+06, 1.3450E+06, 1.3507E+06,
     +  1.3564E+06, 1.3622E+06, 1.3679E+06, 1.3737E+06, 1.3795E+06,
     + 1.3853E+06, 1.3912E+06, 1.3971E+06, 1.4029E+06, 1.4089E+06,
     +  1.4148E+06, 1.4207E+06, 1.4267E+06, 1.4327E+06, 1.4387E+06,
     + 1.4448E+06, 1.4509E+06, 1.4569E+06, 1.4631E+06, 1.4692E+06,
     +  1.4753E+06, 1.4815E+06, 1.4877E+06, 1.4940E+06, 1.5002E+06,
     + 1.5065E+06, 1.5128E+06, 1.5191E+06, 1.5254E+06, 1.5318E+06,
     +  1.5382E+06, 1.5446E+06, 1.5510E+06, 1.5575E+06, 1.5640E+06,
     + 1.5705E+06, 1.5770E+06, 1.5836E+06, 1.5901E+06, 1.5967E+06,
     +  1.6034E+06, 1.6100E+06, 1.6167E+06, 1.6234E+06, 1.6301E+06,
     + 1.6369E+06, 1.6436E+06, 1.6504E+06, 1.6573E+06, 1.6641E+06,
     +  1.6710E+06, 1.6779E+06, 1.6848E+06, 1.6918E+06, 1.6987E+06,
     + 1.7057E+06, 1.7127E+06, 1.7198E+06, 1.7269E+06, 1.7340E+06,
     +  1.7411E+06, 1.7483E+06, 1.7554E+06, 1.7626E+06, 1.7699E+06,
     + 1.7771E+06, 1.7844E+06, 1.7917E+06, 1.7990E+06, 1.8064E+06,
     +  1.8138E+06, 1.8212E+06, 1.8287E+06, 1.8361E+06, 1.8436E+06,
     + 1.8511E+06, 1.8587E+06, 1.8663E+06, 1.8739E+06, 1.8815E+06,
     +  1.8892E+06, 1.8968E+06, 1.9046E+06, 1.9123E+06, 1.9201E+06,
     + 1.9279E+06, 1.9357E+06, 1.9435E+06, 1.9514E+06, 1.9593E+06,
     +  1.9673E+06, 1.9752E+06, 1.9832E+06, 1.9912E+06, 1.9993E+06,
     + 2.0074E+06, 2.0155E+06, 2.0236E+06, 2.0318E+06, 2.0400E+06,
     +  2.0482E+06, 2.0565E+06, 2.0647E+06, 2.0731E+06, 2.0814E+06,
     + 2.0898E+06, 2.0982E+06, 2.1066E+06, 2.1151E+06, 2.1235E+06,
     +  2.1321E+06, 2.1406E+06, 2.1492E+06, 2.1578E+06, 2.1665E+06,
     + 2.1751E+06, 2.1838E+06, 2.1926E+06, 2.2013E+06, 2.2101E+06,
     +  2.2189E+06, 2.2278E+06, 2.2367E+06, 2.2456E+06, 2.2546E+06,
     + 2.2635E+06, 2.2725E+06, 2.2816E+06, 2.2907E+06, 2.2998E+06,
     +  2.3089E+06, 2.3181E+06, 2.3273E+06, 2.3365E+06, 2.3458E+06,
     + 2.3551E+06, 2.3644E+06, 2.3738E+06, 2.3832E+06, 2.3926E+06,
     +  2.4021E+06, 2.4116E+06, 2.4211E+06, 2.4307E+06, 2.4403E+06,
     + 2.4499E+06, 2.4596E+06, 2.4693E+06, 2.4790E+06, 2.4888E+06,
     +  2.4986E+06, 2.5084E+06, 2.5183E+06, 2.5282E+06, 2.5381E+06,
     + 2.5481E+06, 2.5581E+06, 2.5681E+06, 2.5782E+06, 2.5883E+06,
     +  2.5984E+06, 2.6086E+06, 2.6188E+06, 2.6291E+06, 2.6393E+06,
     + 2.6497E+06, 2.6600E+06, 2.6704E+06, 2.6808E+06, 2.6913E+06,
     +  2.7018E+06, 2.7123E+06, 2.7229E+06, 2.7335E+06, 2.7441E+06,
     + 2.7548E+06, 2.7655E+06, 2.7763E+06, 2.7871E+06, 2.7979E+06,
     +  2.8087E+06, 2.8196E+06, 2.8306E+06, 2.8415E+06, 2.8526E+06,
     + 2.8636E+06, 2.8747E+06, 2.8858E+06, 2.8970E+06, 2.9082E+06,
     +  2.9194E+06, 2.9307E+06, 2.9420E+06, 2.9533E+06, 2.9647E+06,
     + 2.9762E+06, 2.9876E+06, 2.9991E+06, 3.0107E+06, 3.0223E+06,
     +  3.0339E+06, 3.0456E+06, 3.0573E+06, 3.0690E+06, 3.0808E+06,
     + 3.0926E+06, 3.1045E+06, 3.1164E+06, 3.1283E+06, 3.1403E+06,
     +  3.1523E+06, 3.1644E+06, 3.1765E+06, 3.1886E+06, 3.2008E+06,
     + 3.2130E+06, 3.2253E+06, 3.2376E+06, 3.2499E+06, 3.2623E+06,
     +  3.2748E+06, 3.2872E+06, 3.2998E+06, 3.3123E+06, 3.3249E+06,
     + 3.3376E+06, 3.3502E+06, 3.3630E+06, 3.3757E+06, 3.3885E+06,
     +  3.4014E+06, 3.4143E+06, 3.4272E+06, 3.4402E+06, 3.4532E+06,
     + 3.4663E+06, 3.4794E+06, 3.4926E+06, 3.5058E+06, 3.5190E+06,
     +  3.5323E+06, 3.5456E+06, 3.5590E+06, 3.5724E+06, 3.5859E+06,
     + 3.5994E+06, 3.6130E+06, 3.6266E+06, 3.6402E+06, 3.6539E+06,
     +  3.6676E+06, 3.6814E+06, 3.6952E+06, 3.7091E+06, 3.7230E+06,
     + 3.7370E+06, 3.7510E+06, 3.7650E+06, 3.7791E+06, 3.7933E+06,
     +  3.8075E+06, 3.8217E+06, 3.8360E+06, 3.8503E+06, 3.8647E+06,
     + 3.8791E+06, 3.8936E+06, 3.9081E+06, 3.9227E+06, 3.9373E+06,
     +  3.9520E+06, 3.9667E+06, 3.9815E+06, 3.9963E+06, 4.0111E+06,
     + 4.0260E+06, 4.0410E+06, 4.0560E+06, 4.0710E+06, 4.0861E+06,
     +  4.1013E+06, 4.1165E+06, 4.1317E+06, 4.1470E+06, 4.1624E+06,
     + 4.1778E+06, 4.1932E+06, 4.2087E+06, 4.2243E+06, 4.2399E+06,
     +  4.2555E+06, 4.2712E+06, 4.2870E+06, 4.3028E+06, 4.3187E+06,
     + 4.3346E+06, 4.3505E+06, 4.3665E+06, 4.3826E+06, 4.3987E+06,
     +  4.4149E+06, 4.4311E+06, 4.4473E+06, 4.4637E+06, 4.4800E+06,
     + 4.4965E+06, 4.5129E+06, 4.5295E+06, 4.5461E+06, 4.5627E+06,
     +  4.5794E+06, 4.5961E+06, 4.6129E+06, 4.6298E+06, 4.6467E+06,
     + 4.6637E+06, 4.6807E+06, 4.6977E+06, 4.7149E+06, 4.7320E+06,
     +  4.7493E+06, 4.7666E+06, 4.7839E+06, 4.8013E+06, 4.8188E+06,
     + 4.8363E+06, 4.8538E+06, 4.8715E+06, 4.8891E+06, 4.9069E+06,
     +  4.9247E+06, 4.9425E+06, 4.9604E+06, 4.9784E+06, 4.9964E+06,
     + 5.0145E+06, 5.0326E+06, 5.0508E+06, 5.0691E+06, 5.0874E+06,
     +  5.1057E+06, 5.1242E+06, 5.1426E+06, 5.1612E+06, 5.1798E+06,
     + 5.1984E+06, 5.2171E+06, 5.2359E+06, 5.2548E+06, 5.2736E+06,
     +  5.2926E+06, 5.3116E+06, 5.3307E+06, 5.3498E+06, 5.3690E+06,
     + 5.3883E+06, 5.4076E+06, 5.4270E+06, 5.4464E+06, 5.4659E+06,
     +  5.4854E+06, 5.5051E+06, 5.5248E+06, 5.5445E+06, 5.5643E+06,
     + 5.5842E+06, 5.6041E+06, 5.6241E+06, 5.6441E+06, 5.6643E+06,
     +  5.6844E+06, 5.7047E+06, 5.7250E+06, 5.7454E+06, 5.7658E+06,
     + 5.7863E+06, 5.8068E+06, 5.8275E+06, 5.8482E+06, 5.8689E+06,
     +  5.8897E+06, 5.9106E+06, 5.9316E+06, 5.9526E+06, 5.9736E+06,
     + 5.9948E+06, 6.0160E+06, 6.0373E+06, 6.0586E+06, 6.0800E+06,
     +  6.1015E+06, 6.1230E+06, 6.1446E+06, 6.1663E+06, 6.1881E+06,
     + 6.2099E+06, 6.2317E+06, 6.2537E+06, 6.2757E+06, 6.2978E+06,
     +  6.3199E+06, 6.3421E+06, 6.3644E+06, 6.3868E+06, 6.4092E+06,
     + 6.4317E+06, 6.4542E+06, 6.4768E+06, 6.4995E+06, 6.5223E+06,
     +  6.5451E+06, 6.5681E+06, 6.5910E+06, 6.6141E+06, 6.6372E+06,
     + 6.6604E+06, 6.6836E+06, 6.7070E+06, 6.7304E+06, 6.7539E+06,
     +  6.7774E+06, 6.8010E+06, 6.8247E+06, 6.8485E+06, 6.8723E+06,
     + 6.8962E+06, 6.9202E+06, 6.9442E+06, 6.9684E+06, 6.9926E+06,
     +  7.0168E+06, 7.0412E+06, 7.0656E+06, 7.0901E+06, 7.1147E+06,
     + 7.1393E+06, 7.1640E+06, 7.1888E+06, 7.2137E+06, 7.2386E+06,
     +  7.2637E+06, 7.2888E+06, 7.3139E+06, 7.3392E+06, 7.3645E+06,
     + 7.3899E+06, 7.4154E+06, 7.4409E+06, 7.4666E+06, 7.4923E+06,
     +  7.5181E+06, 7.5439E+06, 7.5699E+06, 7.5959E+06, 7.6220E+06,
     + 7.6482E+06, 7.6744E+06, 7.7008E+06, 7.7272E+06, 7.7537E+06,
     +  7.7803E+06, 7.8069E+06, 7.8336E+06, 7.8605E+06, 7.8874E+06,
     + 7.9143E+06, 7.9414E+06, 7.9685E+06, 7.9957E+06, 8.0230E+06,
     +  8.0504E+06, 8.0779E+06, 8.1054E+06, 8.1331E+06, 8.1608E+06,
     + 8.1886E+06, 8.2164E+06, 8.2444E+06, 8.2724E+06, 8.3006E+06,
     +  8.3288E+06, 8.3571E+06, 8.3854E+06, 8.4139E+06, 8.4424E+06,
     + 8.4711E+06, 8.4998E+06, 8.5286E+06, 8.5575E+06, 8.5864E+06,
     +  8.6155E+06, 8.6446E+06, 8.6739E+06, 8.7032E+06, 8.7326E+06,
     + 8.7621E+06, 8.7917E+06, 8.8213E+06, 8.8511E+06, 8.8809E+06,
     +  8.9109E+06, 8.9409E+06, 8.9710E+06, 9.0012E+06, 9.0315E+06,
     + 9.0618E+06, 9.0923E+06, 9.1228E+06, 9.1535E+06, 9.1842E+06,
     +  9.2150E+06, 9.2460E+06, 9.2770E+06, 9.3081E+06, 9.3392E+06,
     + 9.3705E+06, 9.4019E+06, 9.4333E+06, 9.4649E+06, 9.4965E+06,
     +  9.5283E+06, 9.5601E+06, 9.5920E+06, 9.6241E+06, 9.6562E+06,
     + 9.6884E+06, 9.7207E+06, 9.7531E+06, 9.7856E+06, 9.8181E+06,
     +  9.8508E+06, 9.8836E+06, 9.9165E+06, 9.9494E+06, 9.9825E+06,
     + 1.0016E+07, 1.0049E+07, 1.0082E+07, 1.0116E+07, 1.0149E+07,
     +  1.0183E+07, 1.0217E+07, 1.0250E+07, 1.0284E+07, 1.0318E+07,
     + 1.0353E+07, 1.0387E+07, 1.0421E+07, 1.0456E+07, 1.0490E+07,
     +  1.0525E+07, 1.0559E+07, 1.0594E+07, 1.0629E+07, 1.0664E+07,
     + 1.0699E+07, 1.0735E+07, 1.0770E+07, 1.0805E+07, 1.0841E+07,
     +  1.0876E+07, 1.0912E+07, 1.0948E+07, 1.0984E+07, 1.1020E+07,
     + 1.1056E+07, 1.1092E+07, 1.1129E+07, 1.1165E+07, 1.1202E+07,
     +  1.1239E+07, 1.1275E+07, 1.1312E+07, 1.1349E+07, 1.1386E+07,
     + 1.1424E+07, 1.1461E+07, 1.1498E+07, 1.1536E+07, 1.1573E+07,
     +  1.1611E+07, 1.1649E+07, 1.1687E+07, 1.1725E+07, 1.1763E+07,
     + 1.1801E+07, 1.1840E+07, 1.1878E+07, 1.1917E+07, 1.1956E+07,
     +  1.1994E+07, 1.2033E+07, 1.2072E+07, 1.2112E+07, 1.2151E+07,
     + 1.2190E+07, 1.2230E+07, 1.2269E+07, 1.2309E+07, 1.2349E+07,
     +  1.2389E+07, 1.2429E+07, 1.2469E+07, 1.2509E+07, 1.2550E+07,
     + 1.2590E+07, 1.2631E+07, 1.2672E+07, 1.2712E+07, 1.2753E+07,
     +  1.2795E+07, 1.2836E+07, 1.2877E+07, 1.2918E+07, 1.2960E+07,
     + 1.3002E+07, 1.3043E+07, 1.3085E+07, 1.3127E+07, 1.3170E+07,
     +  1.3212E+07, 1.3254E+07, 1.3297E+07, 1.3339E+07, 1.3382E+07,
     + 1.3425E+07, 1.3468E+07, 1.3511E+07, 1.3554E+07, 1.3597E+07,
     +  1.3641E+07, 1.3685E+07, 1.3728E+07, 1.3772E+07, 1.3816E+07,
     + 1.3860E+07, 1.3904E+07, 1.3949E+07, 1.3993E+07, 1.4038E+07,
     +  1.4082E+07, 1.4127E+07, 1.4172E+07, 1.4217E+07, 1.4262E+07,
     + 1.4307E+07, 1.4353E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       128
      data (QofT( 3,J),J=1,2408)/ 7.7730E-01, 2.1033E+00, 3.5864E+00,
     +  5.2941E+00, 7.2078E+00, 9.3092E+00, 1.1583E+01, 1.4018E+01,
     + 1.6603E+01, 1.9331E+01, 2.2195E+01, 2.5188E+01, 2.8304E+01,
     +  3.1540E+01, 3.4891E+01, 3.8353E+01, 4.1923E+01, 4.5598E+01,
     + 4.9374E+01, 5.3249E+01, 6.1286E+01, 6.9692E+01, 7.8452E+01,
     +  8.7551E+01, 9.6977E+01, 1.0672E+02, 1.1677E+02, 1.2711E+02,
     + 1.3775E+02, 1.4867E+02, 1.5986E+02, 1.7132E+02, 1.8304E+02,
     +  1.9502E+02, 2.0725E+02, 2.1972E+02, 2.3243E+02, 2.4539E+02,
     + 2.5857E+02, 2.7198E+02, 2.8562E+02, 2.9948E+02, 3.1355E+02,
     +  3.2784E+02, 3.4235E+02, 3.5705E+02, 3.7197E+02, 3.8709E+02,
     + 4.0240E+02, 4.1792E+02, 4.3363E+02, 4.4953E+02, 4.6562E+02,
     +  4.8190E+02, 4.9837E+02, 5.1501E+02, 5.3185E+02, 5.4886E+02,
     + 5.6605E+02, 5.8341E+02, 6.0095E+02, 6.1866E+02, 6.3655E+02,
     +  6.5460E+02, 6.7282E+02, 6.9121E+02, 7.0977E+02, 7.2848E+02,
     + 7.4736E+02, 7.6640E+02, 7.8560E+02, 8.0496E+02, 8.2447E+02,
     +  8.4414E+02, 8.6397E+02, 8.8395E+02, 9.0408E+02, 9.2437E+02,
     + 9.4480E+02, 9.6538E+02, 9.8611E+02, 1.0070E+03, 1.0280E+03,
     +  1.0492E+03, 1.0705E+03, 1.0919E+03, 1.1135E+03, 1.1353E+03,
     + 1.1572E+03, 1.1792E+03, 1.2014E+03, 1.2237E+03, 1.2461E+03,
     +  1.2687E+03, 1.2914E+03, 1.3142E+03, 1.3372E+03, 1.3603E+03,
     + 1.3836E+03, 1.4070E+03, 1.4305E+03, 1.4542E+03, 1.4779E+03,
     +  1.5019E+03, 1.5259E+03, 1.5501E+03, 1.5744E+03, 1.5988E+03,
     + 1.6234E+03, 1.6481E+03, 1.6729E+03, 1.6979E+03, 1.7230E+03,
     +  1.7482E+03, 1.7735E+03, 1.7990E+03, 1.8246E+03, 1.8504E+03,
     + 1.8762E+03, 1.9022E+03, 1.9283E+03, 1.9546E+03, 1.9809E+03,
     +  2.0074E+03, 2.0341E+03, 2.0608E+03, 2.0877E+03, 2.1147E+03,
     + 2.1419E+03, 2.1691E+03, 2.1965E+03, 2.2241E+03, 2.2517E+03,
     +  2.2795E+03, 2.3074E+03, 2.3355E+03, 2.3636E+03, 2.3920E+03,
     + 2.4204E+03, 2.4490E+03, 2.4777E+03, 2.5065E+03, 2.5355E+03,
     +  2.5646E+03, 2.5938E+03, 2.6232E+03, 2.6527E+03, 2.6824E+03,
     + 2.7122E+03, 2.7421E+03, 2.7721E+03, 2.8023E+03, 2.8327E+03,
     +  2.8631E+03, 2.8938E+03, 2.9245E+03, 2.9554E+03, 2.9864E+03,
     + 3.0176E+03, 3.0489E+03, 3.0804E+03, 3.1120E+03, 3.1438E+03,
     +  3.1757E+03, 3.2077E+03, 3.2399E+03, 3.2723E+03, 3.3048E+03,
     + 3.3374E+03, 3.3702E+03, 3.4032E+03, 3.4363E+03, 3.4696E+03,
     +  3.5030E+03, 3.5366E+03, 3.5703E+03, 3.6042E+03, 3.6383E+03,
     + 3.6725E+03, 3.7069E+03, 3.7414E+03, 3.7761E+03, 3.8110E+03,
     +  3.8460E+03, 3.8812E+03, 3.9166E+03, 3.9521E+03, 3.9878E+03,
     + 4.0237E+03, 4.0598E+03, 4.0960E+03, 4.1324E+03, 4.1690E+03,
     +  4.2058E+03, 4.2427E+03, 4.2798E+03, 4.3171E+03, 4.3546E+03,
     + 4.3923E+03, 4.4301E+03, 4.4682E+03, 4.5064E+03, 4.5448E+03,
     +  4.5834E+03, 4.6222E+03, 4.6612E+03, 4.7004E+03, 4.7398E+03,
     + 4.7794E+03, 4.8191E+03, 4.8591E+03, 4.8993E+03, 4.9397E+03,
     +  4.9803E+03, 5.0211E+03, 5.0620E+03, 5.1032E+03, 5.1446E+03,
     + 5.1862E+03, 5.2281E+03, 5.2702E+03, 5.3124E+03, 5.3549E+03,
     +  5.3976E+03, 5.4405E+03, 5.4836E+03, 5.5270E+03, 5.5706E+03,
     + 5.6144E+03, 5.6585E+03, 5.7027E+03, 5.7472E+03, 5.7919E+03,
     +  5.8369E+03, 5.8821E+03, 5.9275E+03, 5.9732E+03, 6.0191E+03,
     + 6.0653E+03, 6.1116E+03, 6.1583E+03, 6.2052E+03, 6.2523E+03,
     +  6.2997E+03, 6.3473E+03, 6.3952E+03, 6.4433E+03, 6.4918E+03,
     + 6.5404E+03, 6.5893E+03, 6.6385E+03, 6.6879E+03, 6.7376E+03,
     +  6.7876E+03, 6.8378E+03, 6.8883E+03, 6.9391E+03, 6.9901E+03,
     + 7.0415E+03, 7.0930E+03, 7.1449E+03, 7.1971E+03, 7.2495E+03,
     +  7.3022E+03, 7.3552E+03, 7.4085E+03, 7.4621E+03, 7.5159E+03,
     + 7.5701E+03, 7.6246E+03, 7.6793E+03, 7.7344E+03, 7.7897E+03,
     +  7.8454E+03, 7.9013E+03, 7.9576E+03, 8.0141E+03, 8.0710E+03,
     + 8.1282E+03, 8.1857E+03, 8.2435E+03, 8.3016E+03, 8.3601E+03,
     +  8.4188E+03, 8.4779E+03, 8.5373E+03, 8.5971E+03, 8.6571E+03,
     + 8.7176E+03, 8.7783E+03, 8.8394E+03, 8.9007E+03, 8.9625E+03,
     +  9.0246E+03, 9.0870E+03, 9.1498E+03, 9.2129E+03, 9.2763E+03,
     + 9.3401E+03, 9.4043E+03, 9.4688E+03, 9.5337E+03, 9.5990E+03,
     +  9.6645E+03, 9.7305E+03, 9.7968E+03, 9.8635E+03, 9.9306E+03,
     + 9.9980E+03, 1.0066E+04, 1.0134E+04, 1.0203E+04, 1.0271E+04,
     +  1.0341E+04, 1.0410E+04, 1.0481E+04, 1.0551E+04, 1.0622E+04,
     + 1.0693E+04, 1.0765E+04, 1.0837E+04, 1.0909E+04, 1.0982E+04,
     +  1.1056E+04, 1.1129E+04, 1.1203E+04, 1.1278E+04, 1.1353E+04,
     + 1.1428E+04, 1.1504E+04, 1.1580E+04, 1.1657E+04, 1.1734E+04,
     +  1.1811E+04, 1.1889E+04, 1.1967E+04, 1.2046E+04, 1.2125E+04,
     + 1.2205E+04, 1.2285E+04, 1.2365E+04, 1.2446E+04, 1.2528E+04,
     +  1.2610E+04, 1.2692E+04, 1.2775E+04, 1.2858E+04, 1.2942E+04,
     + 1.3026E+04, 1.3111E+04, 1.3196E+04, 1.3281E+04, 1.3367E+04,
     +  1.3454E+04, 1.3541E+04, 1.3628E+04, 1.3716E+04, 1.3805E+04,
     + 1.3894E+04, 1.3983E+04, 1.4073E+04, 1.4164E+04, 1.4255E+04,
     +  1.4346E+04, 1.4438E+04, 1.4530E+04, 1.4623E+04, 1.4717E+04,
     + 1.4811E+04, 1.4906E+04, 1.5001E+04, 1.5096E+04, 1.5192E+04,
     +  1.5289E+04, 1.5386E+04, 1.5484E+04, 1.5582E+04, 1.5681E+04,
     + 1.5780E+04, 1.5880E+04, 1.5981E+04, 1.6082E+04, 1.6183E+04,
     +  1.6286E+04, 1.6388E+04, 1.6492E+04, 1.6596E+04, 1.6700E+04,
     + 1.6805E+04, 1.6911E+04, 1.7017E+04, 1.7123E+04, 1.7231E+04,
     +  1.7339E+04, 1.7447E+04, 1.7556E+04, 1.7666E+04, 1.7776E+04,
     + 1.7887E+04, 1.7999E+04, 1.8111E+04, 1.8224E+04, 1.8337E+04,
     +  1.8451E+04, 1.8566E+04, 1.8681E+04, 1.8797E+04, 1.8914E+04,
     + 1.9031E+04, 1.9149E+04, 1.9267E+04, 1.9387E+04, 1.9506E+04,
     +  1.9627E+04, 1.9748E+04, 1.9870E+04, 1.9992E+04, 2.0115E+04,
     + 2.0239E+04, 2.0363E+04, 2.0489E+04, 2.0614E+04, 2.0741E+04,
     +  2.0868E+04, 2.0996E+04, 2.1125E+04, 2.1254E+04, 2.1384E+04,
     + 2.1514E+04, 2.1646E+04, 2.1778E+04, 2.1911E+04, 2.2044E+04,
     +  2.2179E+04, 2.2314E+04, 2.2450E+04, 2.2586E+04, 2.2723E+04,
     + 2.2861E+04, 2.3000E+04, 2.3139E+04, 2.3280E+04, 2.3421E+04,
     +  2.3563E+04, 2.3705E+04, 2.3848E+04, 2.3992E+04, 2.4137E+04,
     + 2.4283E+04, 2.4429E+04, 2.4577E+04, 2.4725E+04, 2.4873E+04,
     +  2.5023E+04, 2.5173E+04, 2.5325E+04, 2.5477E+04, 2.5630E+04,
     + 2.5783E+04, 2.5938E+04, 2.6093E+04, 2.6249E+04, 2.6406E+04,
     +  2.6564E+04, 2.6723E+04, 2.6882E+04, 2.7043E+04, 2.7204E+04,
     + 2.7366E+04, 2.7529E+04, 2.7693E+04, 2.7858E+04, 2.8023E+04,
     +  2.8190E+04, 2.8357E+04, 2.8526E+04, 2.8695E+04, 2.8865E+04,
     + 2.9036E+04, 2.9208E+04, 2.9381E+04, 2.9554E+04, 2.9729E+04,
     +  2.9905E+04, 3.0081E+04, 3.0259E+04, 3.0437E+04, 3.0616E+04,
     + 3.0797E+04, 3.0978E+04, 3.1160E+04, 3.1343E+04, 3.1527E+04,
     +  3.1713E+04, 3.1899E+04, 3.2086E+04, 3.2274E+04, 3.2463E+04,
     + 3.2653E+04, 3.2844E+04, 3.3036E+04, 3.3229E+04, 3.3423E+04,
     +  3.3618E+04, 3.3815E+04, 3.4012E+04, 3.4210E+04, 3.4409E+04,
     + 3.4610E+04, 3.4811E+04, 3.5013E+04, 3.5217E+04, 3.5421E+04,
     +  3.5627E+04, 3.5834E+04, 3.6041E+04, 3.6250E+04, 3.6460E+04,
     + 3.6671E+04, 3.6883E+04, 3.7096E+04, 3.7311E+04, 3.7526E+04,
     +  3.7743E+04, 3.7961E+04, 3.8180E+04, 3.8399E+04, 3.8621E+04,
     + 3.8843E+04, 3.9066E+04, 3.9291E+04, 3.9517E+04, 3.9744E+04,
     +  3.9972E+04, 4.0201E+04, 4.0431E+04, 4.0663E+04, 4.0896E+04,
     + 4.1130E+04, 4.1365E+04, 4.1602E+04, 4.1839E+04, 4.2078E+04,
     +  4.2319E+04, 4.2560E+04, 4.2803E+04, 4.3047E+04, 4.3292E+04,
     + 4.3538E+04, 4.3786E+04, 4.4035E+04, 4.4285E+04, 4.4536E+04,
     +  4.4789E+04, 4.5043E+04, 4.5298E+04, 4.5555E+04, 4.5813E+04,
     + 4.6072E+04, 4.6333E+04, 4.6595E+04, 4.6858E+04, 4.7123E+04,
     +  4.7389E+04, 4.7656E+04, 4.7925E+04, 4.8195E+04, 4.8466E+04,
     + 4.8739E+04, 4.9014E+04, 4.9289E+04, 4.9566E+04, 4.9844E+04,
     +  5.0124E+04, 5.0405E+04, 5.0688E+04, 5.0972E+04, 5.1258E+04,
     + 5.1545E+04, 5.1833E+04, 5.2123E+04, 5.2414E+04, 5.2707E+04,
     +  5.3001E+04, 5.3297E+04, 5.3594E+04, 5.3893E+04, 5.4193E+04,
     + 5.4495E+04, 5.4798E+04, 5.5103E+04, 5.5409E+04, 5.5717E+04,
     +  5.6026E+04, 5.6337E+04, 5.6650E+04, 5.6964E+04, 5.7279E+04,
     + 5.7596E+04, 5.7915E+04, 5.8235E+04, 5.8558E+04, 5.8881E+04,
     +  5.9206E+04, 5.9533E+04, 5.9861E+04, 6.0191E+04, 6.0523E+04,
     + 6.0856E+04, 6.1191E+04, 6.1528E+04, 6.1866E+04, 6.2206E+04,
     +  6.2548E+04, 6.2891E+04, 6.3236E+04, 6.3583E+04, 6.3931E+04,
     + 6.4282E+04, 6.4634E+04, 6.4987E+04, 6.5343E+04, 6.5700E+04,
     +  6.6059E+04, 6.6419E+04, 6.6782E+04, 6.7146E+04, 6.7512E+04,
     + 6.7880E+04, 6.8249E+04, 6.8621E+04, 6.8994E+04, 6.9369E+04,
     +  6.9746E+04, 7.0125E+04, 7.0506E+04, 7.0888E+04, 7.1272E+04,
     + 7.1659E+04, 7.2047E+04, 7.2437E+04, 7.2828E+04, 7.3222E+04,
     +  7.3618E+04, 7.4015E+04, 7.4415E+04, 7.4817E+04, 7.5220E+04,
     + 7.5626E+04, 7.6033E+04, 7.6442E+04, 7.6853E+04, 7.7267E+04,
     +  7.7682E+04, 7.8099E+04, 7.8519E+04, 7.8940E+04, 7.9364E+04,
     + 7.9789E+04, 8.0217E+04, 8.0646E+04, 8.1078E+04, 8.1511E+04,
     +  8.1947E+04, 8.2385E+04, 8.2825E+04, 8.3267E+04, 8.3711E+04,
     + 8.4158E+04, 8.4606E+04, 8.5057E+04, 8.5509E+04, 8.5965E+04,
     +  8.6422E+04, 8.6881E+04, 8.7343E+04, 8.7806E+04, 8.8272E+04,
     + 8.8740E+04, 8.9210E+04, 8.9683E+04, 9.0158E+04, 9.0635E+04,
     +  9.1115E+04, 9.1596E+04, 9.2080E+04, 9.2567E+04, 9.3055E+04,
     + 9.3546E+04, 9.4039E+04, 9.4535E+04, 9.5033E+04, 9.5533E+04,
     +  9.6035E+04, 9.6540E+04, 9.7048E+04, 9.7557E+04, 9.8069E+04,
     + 9.8584E+04, 9.9101E+04, 9.9621E+04, 1.0014E+05, 1.0067E+05,
     +  1.0119E+05, 1.0172E+05, 1.0225E+05, 1.0279E+05, 1.0333E+05,
     + 1.0386E+05, 1.0441E+05, 1.0495E+05, 1.0550E+05, 1.0605E+05,
     +  1.0660E+05, 1.0715E+05, 1.0771E+05, 1.0827E+05, 1.0883E+05,
     + 1.0940E+05, 1.0997E+05, 1.1054E+05, 1.1111E+05, 1.1168E+05,
     +  1.1226E+05, 1.1284E+05, 1.1343E+05, 1.1401E+05, 1.1460E+05,
     + 1.1519E+05, 1.1579E+05, 1.1639E+05, 1.1699E+05, 1.1759E+05,
     +  1.1819E+05, 1.1880E+05, 1.1941E+05, 1.2003E+05, 1.2064E+05,
     + 1.2126E+05, 1.2189E+05, 1.2251E+05, 1.2314E+05, 1.2377E+05,
     +  1.2440E+05, 1.2504E+05, 1.2568E+05, 1.2632E+05, 1.2697E+05,
     + 1.2762E+05, 1.2827E+05, 1.2893E+05, 1.2958E+05, 1.3024E+05,
     +  1.3091E+05, 1.3157E+05, 1.3224E+05, 1.3292E+05, 1.3359E+05,
     + 1.3427E+05, 1.3495E+05, 1.3564E+05, 1.3633E+05, 1.3702E+05,
     +  1.3771E+05, 1.3841E+05, 1.3911E+05, 1.3981E+05, 1.4052E+05,
     + 1.4123E+05, 1.4194E+05, 1.4266E+05, 1.4338E+05, 1.4410E+05,
     +  1.4483E+05, 1.4556E+05, 1.4629E+05, 1.4703E+05, 1.4777E+05,
     + 1.4851E+05, 1.4926E+05, 1.5001E+05, 1.5076E+05, 1.5152E+05,
     +  1.5228E+05, 1.5304E+05, 1.5381E+05, 1.5458E+05, 1.5535E+05,
     + 1.5613E+05, 1.5691E+05, 1.5769E+05, 1.5848E+05, 1.5927E+05,
     +  1.6007E+05, 1.6086E+05, 1.6167E+05, 1.6247E+05, 1.6328E+05,
     + 1.6409E+05, 1.6491E+05, 1.6573E+05, 1.6655E+05, 1.6738E+05,
     +  1.6821E+05, 1.6904E+05, 1.6988E+05, 1.7072E+05, 1.7157E+05,
     + 1.7242E+05, 1.7327E+05, 1.7413E+05, 1.7499E+05, 1.7585E+05,
     +  1.7672E+05, 1.7759E+05, 1.7847E+05, 1.7934E+05, 1.8023E+05,
     + 1.8112E+05, 1.8201E+05, 1.8290E+05, 1.8380E+05, 1.8470E+05,
     +  1.8561E+05, 1.8652E+05, 1.8744E+05, 1.8835E+05, 1.8928E+05,
     + 1.9020E+05, 1.9114E+05, 1.9207E+05, 1.9301E+05, 1.9395E+05,
     +  1.9490E+05, 1.9585E+05, 1.9681E+05, 1.9777E+05, 1.9873E+05,
     + 1.9970E+05, 2.0067E+05, 2.0165E+05, 2.0263E+05, 2.0361E+05,
     +  2.0460E+05, 2.0560E+05, 2.0659E+05, 2.0760E+05, 2.0860E+05,
     + 2.0961E+05, 2.1063E+05, 2.1165E+05, 2.1267E+05, 2.1370E+05,
     +  2.1473E+05, 2.1577E+05, 2.1681E+05, 2.1786E+05, 2.1891E+05,
     + 2.1996E+05, 2.2102E+05, 2.2209E+05, 2.2316E+05, 2.2423E+05,
     +  2.2531E+05, 2.2639E+05, 2.2748E+05, 2.2857E+05, 2.2967E+05,
     + 2.3077E+05, 2.3187E+05, 2.3298E+05, 2.3410E+05, 2.3522E+05,
     +  2.3634E+05, 2.3747E+05, 2.3861E+05, 2.3975E+05, 2.4089E+05,
     + 2.4204E+05, 2.4320E+05, 2.4435E+05, 2.4552E+05, 2.4669E+05,
     +  2.4786E+05, 2.4904E+05, 2.5022E+05, 2.5141E+05, 2.5261E+05,
     + 2.5380E+05, 2.5501E+05, 2.5622E+05, 2.5743E+05, 2.5865E+05,
     +  2.5987E+05, 2.6110E+05, 2.6234E+05, 2.6358E+05, 2.6482E+05,
     + 2.6607E+05, 2.6733E+05, 2.6859E+05, 2.6985E+05, 2.7113E+05,
     +  2.7240E+05, 2.7368E+05, 2.7497E+05, 2.7626E+05, 2.7756E+05,
     + 2.7886E+05, 2.8017E+05, 2.8149E+05, 2.8281E+05, 2.8413E+05,
     +  2.8546E+05, 2.8680E+05, 2.8814E+05, 2.8949E+05, 2.9084E+05,
     + 2.9220E+05, 2.9357E+05, 2.9494E+05, 2.9631E+05, 2.9769E+05,
     +  2.9908E+05, 3.0047E+05, 3.0187E+05, 3.0328E+05, 3.0469E+05,
     + 3.0610E+05, 3.0753E+05, 3.0895E+05, 3.1039E+05, 3.1183E+05,
     +  3.1327E+05, 3.1472E+05, 3.1618E+05, 3.1764E+05, 3.1911E+05,
     + 3.2059E+05, 3.2207E+05, 3.2356E+05, 3.2505E+05, 3.2655E+05,
     +  3.2806E+05, 3.2957E+05, 3.3109E+05, 3.3261E+05, 3.3414E+05,
     + 3.3568E+05, 3.3722E+05, 3.3877E+05, 3.4033E+05, 3.4189E+05,
     +  3.4346E+05, 3.4504E+05, 3.4662E+05, 3.4820E+05, 3.4980E+05,
     + 3.5140E+05, 3.5301E+05, 3.5462E+05, 3.5624E+05, 3.5787E+05,
     +  3.5950E+05, 3.6114E+05, 3.6279E+05, 3.6444E+05, 3.6610E+05,
     + 3.6777E+05, 3.6944E+05, 3.7112E+05, 3.7281E+05, 3.7450E+05,
     +  3.7620E+05, 3.7791E+05, 3.7963E+05, 3.8135E+05, 3.8308E+05,
     + 3.8481E+05, 3.8655E+05, 3.8830E+05, 3.9006E+05, 3.9182E+05,
     +  3.9359E+05, 3.9537E+05, 3.9715E+05, 3.9895E+05, 4.0074E+05,
     + 4.0255E+05, 4.0436E+05, 4.0618E+05, 4.0801E+05, 4.0985E+05,
     +  4.1169E+05, 4.1354E+05, 4.1540E+05, 4.1726E+05, 4.1913E+05,
     + 4.2101E+05, 4.2290E+05, 4.2479E+05, 4.2669E+05, 4.2860E+05,
     +  4.3052E+05, 4.3244E+05, 4.3438E+05, 4.3632E+05, 4.3826E+05,
     + 4.4022E+05, 4.4218E+05, 4.4415E+05, 4.4613E+05, 4.4812E+05,
     +  4.5011E+05, 4.5211E+05, 4.5412E+05, 4.5614E+05, 4.5816E+05,
     + 4.6020E+05, 4.6224E+05, 4.6429E+05, 4.6635E+05, 4.6841E+05,
     +  4.7049E+05, 4.7257E+05, 4.7466E+05, 4.7676E+05, 4.7886E+05,
     + 4.8098E+05, 4.8310E+05, 4.8523E+05, 4.8737E+05, 4.8952E+05,
     +  4.9167E+05, 4.9384E+05, 4.9601E+05, 4.9819E+05, 5.0038E+05,
     + 5.0258E+05, 5.0479E+05, 5.0700E+05, 5.0923E+05, 5.1146E+05,
     +  5.1370E+05, 5.1595E+05, 5.1821E+05, 5.2048E+05, 5.2276E+05,
     + 5.2504E+05, 5.2734E+05, 5.2964E+05, 5.3195E+05, 5.3427E+05,
     +  5.3660E+05, 5.3894E+05, 5.4129E+05, 5.4364E+05, 5.4601E+05,
     + 5.4839E+05, 5.5077E+05, 5.5316E+05, 5.5556E+05, 5.5798E+05,
     +  5.6040E+05, 5.6283E+05, 5.6527E+05, 5.6772E+05, 5.7017E+05,
     + 5.7264E+05, 5.7512E+05, 5.7761E+05, 5.8010E+05, 5.8261E+05,
     +  5.8512E+05, 5.8765E+05, 5.9018E+05, 5.9273E+05, 5.9528E+05,
     + 5.9784E+05, 6.0042E+05, 6.0300E+05, 6.0559E+05, 6.0820E+05,
     +  6.1081E+05, 6.1343E+05, 6.1606E+05, 6.1871E+05, 6.2136E+05,
     + 6.2402E+05, 6.2669E+05, 6.2938E+05, 6.3207E+05, 6.3477E+05,
     +  6.3749E+05, 6.4021E+05, 6.4294E+05, 6.4569E+05, 6.4844E+05,
     + 6.5121E+05, 6.5398E+05, 6.5677E+05, 6.5956E+05, 6.6237E+05,
     +  6.6519E+05, 6.6802E+05, 6.7085E+05, 6.7370E+05, 6.7656E+05,
     + 6.7943E+05, 6.8231E+05, 6.8521E+05, 6.8811E+05, 6.9102E+05,
     +  6.9395E+05, 6.9688E+05, 6.9983E+05, 7.0279E+05, 7.0576E+05,
     + 7.0874E+05, 7.1173E+05, 7.1473E+05, 7.1774E+05, 7.2077E+05,
     +  7.2380E+05, 7.2685E+05, 7.2991E+05, 7.3298E+05, 7.3606E+05,
     + 7.3915E+05, 7.4225E+05, 7.4537E+05, 7.4849E+05, 7.5163E+05,
     +  7.5478E+05, 7.5794E+05, 7.6112E+05, 7.6430E+05, 7.6750E+05,
     + 7.7071E+05, 7.7393E+05, 7.7716E+05, 7.8041E+05, 7.8366E+05,
     +  7.8693E+05, 7.9021E+05, 7.9350E+05, 7.9681E+05, 8.0012E+05,
     + 8.0345E+05, 8.0679E+05, 8.1015E+05, 8.1351E+05, 8.1689E+05,
     +  8.2028E+05, 8.2368E+05, 8.2710E+05, 8.3052E+05, 8.3396E+05,
     + 8.3742E+05, 8.4088E+05, 8.4436E+05, 8.4785E+05, 8.5135E+05,
     +  8.5487E+05, 8.5840E+05, 8.6194E+05, 8.6549E+05, 8.6906E+05,
     + 8.7264E+05, 8.7624E+05, 8.7984E+05, 8.8346E+05, 8.8709E+05,
     +  8.9074E+05, 8.9440E+05, 8.9807E+05, 9.0176E+05, 9.0546E+05,
     + 9.0917E+05, 9.1289E+05, 9.1663E+05, 9.2039E+05, 9.2415E+05,
     +  9.2793E+05, 9.3173E+05, 9.3553E+05, 9.3935E+05, 9.4319E+05,
     + 9.4704E+05, 9.5090E+05, 9.5477E+05, 9.5866E+05, 9.6257E+05,
     +  9.6649E+05, 9.7042E+05, 9.7436E+05, 9.7833E+05, 9.8230E+05,
     + 9.8629E+05, 9.9029E+05, 9.9431E+05, 9.9834E+05, 1.0024E+06,
     +  1.0064E+06, 1.0105E+06, 1.0146E+06, 1.0187E+06, 1.0228E+06,
     + 1.0270E+06, 1.0311E+06, 1.0353E+06, 1.0395E+06, 1.0436E+06,
     +  1.0479E+06, 1.0521E+06, 1.0563E+06, 1.0606E+06, 1.0648E+06,
     + 1.0691E+06, 1.0734E+06, 1.0777E+06, 1.0821E+06, 1.0864E+06,
     +  1.0908E+06, 1.0951E+06, 1.0995E+06, 1.1039E+06, 1.1083E+06,
     + 1.1128E+06, 1.1172E+06, 1.1217E+06, 1.1262E+06, 1.1307E+06,
     +  1.1352E+06, 1.1397E+06, 1.1443E+06, 1.1488E+06, 1.1534E+06,
     + 1.1580E+06, 1.1626E+06, 1.1673E+06, 1.1719E+06, 1.1766E+06,
     +  1.1812E+06, 1.1859E+06, 1.1906E+06, 1.1954E+06, 1.2001E+06,
     + 1.2049E+06, 1.2096E+06, 1.2144E+06, 1.2192E+06, 1.2241E+06,
     +  1.2289E+06, 1.2338E+06, 1.2386E+06, 1.2435E+06, 1.2484E+06,
     + 1.2534E+06, 1.2583E+06, 1.2633E+06, 1.2682E+06, 1.2732E+06,
     +  1.2783E+06, 1.2833E+06, 1.2883E+06, 1.2934E+06, 1.2985E+06,
     + 1.3036E+06, 1.3087E+06, 1.3138E+06, 1.3190E+06, 1.3242E+06,
     +  1.3293E+06, 1.3346E+06, 1.3398E+06, 1.3450E+06, 1.3503E+06,
     + 1.3556E+06, 1.3609E+06, 1.3662E+06, 1.3715E+06, 1.3769E+06,
     +  1.3822E+06, 1.3876E+06, 1.3930E+06, 1.3984E+06, 1.4039E+06,
     + 1.4093E+06, 1.4148E+06, 1.4203E+06, 1.4258E+06, 1.4314E+06,
     +  1.4369E+06, 1.4425E+06, 1.4481E+06, 1.4537E+06, 1.4594E+06,
     + 1.4650E+06, 1.4707E+06, 1.4764E+06, 1.4821E+06, 1.4878E+06,
     +  1.4936E+06, 1.4993E+06, 1.5051E+06, 1.5109E+06, 1.5168E+06,
     + 1.5226E+06, 1.5285E+06, 1.5344E+06, 1.5403E+06, 1.5462E+06,
     +  1.5521E+06, 1.5581E+06, 1.5641E+06, 1.5701E+06, 1.5761E+06,
     + 1.5822E+06, 1.5882E+06, 1.5943E+06, 1.6004E+06, 1.6066E+06,
     +  1.6127E+06, 1.6189E+06, 1.6251E+06, 1.6313E+06, 1.6375E+06,
     + 1.6438E+06, 1.6500E+06, 1.6563E+06, 1.6627E+06, 1.6690E+06,
     +  1.6754E+06, 1.6817E+06, 1.6881E+06, 1.6946E+06, 1.7010E+06,
     + 1.7075E+06, 1.7140E+06, 1.7205E+06, 1.7270E+06, 1.7336E+06,
     +  1.7401E+06, 1.7467E+06, 1.7534E+06, 1.7600E+06, 1.7667E+06,
     + 1.7734E+06, 1.7801E+06, 1.7868E+06, 1.7935E+06, 1.8003E+06,
     +  1.8071E+06, 1.8139E+06, 1.8208E+06, 1.8276E+06, 1.8345E+06,
     + 1.8414E+06, 1.8484E+06, 1.8553E+06, 1.8623E+06, 1.8693E+06,
     +  1.8763E+06, 1.8834E+06, 1.8905E+06, 1.8976E+06, 1.9047E+06,
     + 1.9118E+06, 1.9190E+06, 1.9262E+06, 1.9334E+06, 1.9406E+06,
     +  1.9479E+06, 1.9552E+06, 1.9625E+06, 1.9698E+06, 1.9772E+06,
     + 1.9846E+06, 1.9920E+06, 1.9994E+06, 2.0068E+06, 2.0143E+06,
     +  2.0218E+06, 2.0294E+06, 2.0369E+06, 2.0445E+06, 2.0521E+06,
     + 2.0597E+06, 2.0674E+06, 2.0750E+06, 2.0827E+06, 2.0905E+06,
     +  2.0982E+06, 2.1060E+06, 2.1138E+06, 2.1216E+06, 2.1295E+06,
     + 2.1373E+06, 2.1453E+06, 2.1532E+06, 2.1611E+06, 2.1691E+06,
     +  2.1771E+06, 2.1852E+06, 2.1932E+06, 2.2013E+06, 2.2094E+06,
     + 2.2175E+06, 2.2257E+06, 2.2339E+06, 2.2421E+06, 2.2504E+06,
     +  2.2586E+06, 2.2669E+06, 2.2752E+06, 2.2836E+06, 2.2920E+06,
     + 2.3004E+06, 2.3088E+06, 2.3173E+06, 2.3257E+06, 2.3343E+06,
     +  2.3428E+06, 2.3514E+06, 2.3600E+06, 2.3686E+06, 2.3772E+06,
     + 2.3859E+06, 2.3946E+06, 2.4033E+06, 2.4121E+06, 2.4209E+06,
     +  2.4297E+06, 2.4386E+06, 2.4474E+06, 2.4563E+06, 2.4653E+06,
     + 2.4742E+06, 2.4832E+06, 2.4922E+06, 2.5013E+06, 2.5103E+06,
     +  2.5194E+06, 2.5286E+06, 2.5377E+06, 2.5469E+06, 2.5561E+06,
     + 2.5654E+06, 2.5747E+06, 2.5840E+06, 2.5933E+06, 2.6027E+06,
     +  2.6121E+06, 2.6215E+06, 2.6310E+06, 2.6404E+06, 2.6500E+06,
     + 2.6595E+06, 2.6691E+06, 2.6787E+06, 2.6883E+06, 2.6980E+06,
     +  2.7077E+06, 2.7174E+06, 2.7272E+06, 2.7370E+06, 2.7468E+06,
     + 2.7566E+06, 2.7665E+06, 2.7764E+06, 2.7864E+06, 2.7963E+06,
     +  2.8063E+06, 2.8164E+06, 2.8264E+06, 2.8365E+06, 2.8467E+06,
     + 2.8568E+06, 2.8670E+06, 2.8773E+06, 2.8875E+06, 2.8978E+06,
     +  2.9081E+06, 2.9185E+06, 2.9289E+06, 2.9393E+06, 2.9498E+06,
     + 2.9602E+06, 2.9708E+06, 2.9813E+06, 2.9919E+06, 3.0025E+06,
     +  3.0132E+06, 3.0239E+06, 3.0346E+06, 3.0453E+06, 3.0561E+06,
     + 3.0669E+06, 3.0778E+06, 3.0887E+06, 3.0996E+06, 3.1105E+06,
     +  3.1215E+06, 3.1325E+06, 3.1436E+06, 3.1547E+06, 3.1658E+06,
     + 3.1769E+06, 3.1881E+06, 3.1994E+06, 3.2106E+06, 3.2219E+06,
     +  3.2332E+06, 3.2446E+06, 3.2560E+06, 3.2674E+06, 3.2789E+06,
     + 3.2904E+06, 3.3020E+06, 3.3135E+06, 3.3251E+06, 3.3368E+06,
     +  3.3485E+06, 3.3602E+06, 3.3720E+06, 3.3837E+06, 3.3956E+06,
     + 3.4074E+06, 3.4193E+06, 3.4313E+06, 3.4432E+06, 3.4553E+06,
     +  3.4673E+06, 3.4794E+06, 3.4915E+06, 3.5037E+06, 3.5158E+06,
     + 3.5281E+06, 3.5403E+06, 3.5527E+06, 3.5650E+06, 3.5774E+06,
     +  3.5898E+06, 3.6023E+06, 3.6147E+06, 3.6273E+06, 3.6398E+06,
     + 3.6525E+06, 3.6651E+06, 3.6778E+06, 3.6905E+06, 3.7033E+06,
     +  3.7161E+06, 3.7289E+06, 3.7418E+06, 3.7547E+06, 3.7677E+06,
     + 3.7807E+06, 3.7937E+06, 3.8068E+06, 3.8199E+06, 3.8330E+06,
     +  3.8462E+06, 3.8595E+06, 3.8727E+06, 3.8860E+06, 3.8994E+06,
     + 3.9128E+06, 3.9262E+06, 3.9397E+06, 3.9532E+06, 3.9668E+06,
     +  3.9804E+06, 3.9940E+06, 4.0077E+06, 4.0214E+06, 4.0351E+06,
     + 4.0489E+06, 4.0628E+06, 4.0767E+06, 4.0906E+06, 4.1046E+06,
     +  4.1186E+06, 4.1326E+06, 4.1467E+06, 4.1608E+06, 4.1750E+06,
     + 4.1892E+06, 4.2035E+06, 4.2178E+06, 4.2322E+06, 4.2465E+06,
     +  4.2610E+06, 4.2754E+06, 4.2900E+06, 4.3045E+06, 4.3191E+06,
     + 4.3338E+06, 4.3485E+06, 4.3632E+06, 4.3780E+06, 4.3928E+06,
     +  4.4077E+06, 4.4226E+06, 4.4375E+06, 4.4525E+06, 4.4676E+06,
     + 4.4827E+06, 4.4978E+06, 4.5130E+06, 4.5282E+06, 4.5435E+06,
     +  4.5588E+06, 4.5741E+06, 4.5895E+06, 4.6050E+06, 4.6205E+06,
     + 4.6360E+06, 4.6516E+06, 4.6672E+06, 4.6829E+06, 4.6986E+06,
     +  4.7144E+06, 4.7302E+06, 4.7461E+06, 4.7620E+06, 4.7779E+06,
     + 4.7939E+06, 4.8100E+06, 4.8261E+06, 4.8422E+06, 4.8584E+06,
     +  4.8746E+06, 4.8909E+06, 4.9073E+06, 4.9236E+06, 4.9401E+06,
     + 4.9565E+06, 4.9731E+06, 4.9896E+06, 5.0063E+06, 5.0229E+06,
     +  5.0396E+06, 5.0564E+06, 5.0732E+06, 5.0901E+06, 5.1070E+06,
     + 5.1240E+06, 5.1410E+06, 5.1580E+06, 5.1752E+06, 5.1923E+06,
     +  5.2095E+06, 5.2268E+06, 5.2441E+06, 5.2615E+06, 5.2789E+06,
     + 5.2963E+06, 5.3138E+06, 5.3314E+06, 5.3490E+06, 5.3667E+06,
     +  5.3844E+06, 5.4022E+06, 5.4200E+06, 5.4379E+06, 5.4558E+06,
     + 5.4737E+06, 5.4918E+06, 5.5098E+06, 5.5280E+06, 5.5462E+06,
     +  5.5644E+06, 5.5827E+06, 5.6010E+06, 5.6194E+06, 5.6378E+06,
     + 5.6563E+06, 5.6749E+06, 5.6935E+06, 5.7121E+06, 5.7309E+06,
     +  5.7496E+06, 5.7684E+06, 5.7873E+06, 5.8062E+06, 5.8252E+06,
     + 5.8442E+06, 5.8633E+06, 5.8825E+06, 5.9017E+06, 5.9209E+06,
     +  5.9402E+06, 5.9596E+06, 5.9790E+06, 5.9985E+06, 6.0180E+06,
     + 6.0376E+06, 6.0572E+06, 6.0769E+06, 6.0967E+06, 6.1165E+06,
     +  6.1363E+06, 6.1563E+06, 6.1762E+06, 6.1963E+06, 6.2164E+06,
     + 6.2365E+06, 6.2567E+06, 6.2770E+06, 6.2973E+06, 6.3177E+06,
     +  6.3381E+06, 6.3586E+06, 6.3792E+06, 6.3998E+06, 6.4204E+06,
     + 6.4412E+06, 6.4619E+06, 6.4828E+06, 6.5037E+06, 6.5246E+06,
     +  6.5457E+06, 6.5667E+06, 6.5879E+06, 6.6091E+06, 6.6303E+06,
     + 6.6516E+06, 6.6730E+06, 6.6945E+06, 6.7160E+06, 6.7375E+06,
     +  6.7591E+06, 6.7808E+06, 6.8025E+06, 6.8243E+06, 6.8462E+06,
     + 6.8681E+06, 6.8901E+06, 6.9122E+06, 6.9343E+06, 6.9564E+06,
     +  6.9787E+06, 7.0010E+06, 7.0233E+06, 7.0457E+06, 7.0682E+06,
     + 7.0908E+06, 7.1134E+06, 7.1360E+06, 7.1588E+06, 7.1816E+06,
     +  7.2044E+06, 7.2273E+06, 7.2503E+06, 7.2734E+06, 7.2965E+06,
     + 7.3197E+06, 7.3429E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 3,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_HOCl  (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 2),QofT( 2,NT),Q(NT)
      data (Tmx(I),I=1,2)/3500.,3500./
      data xgj/ 8.,8./
c...     HOCl
c...        --       165
      data (QofT( 1,J),J=1,2408)/ 3.9815E+00, 1.1450E+01, 2.0449E+01,
     +  3.0942E+01, 4.2766E+01, 5.5792E+01, 6.9920E+01, 8.5072E+01,
     + 1.0118E+02, 1.1820E+02, 1.3607E+02, 1.5477E+02, 1.7425E+02,
     +  1.9449E+02, 2.1545E+02, 2.3712E+02, 2.5947E+02, 2.8248E+02,
     + 3.0614E+02, 3.3042E+02, 3.8080E+02, 4.3350E+02, 4.8845E+02,
     +  5.4553E+02, 6.0468E+02, 6.6583E+02, 7.2892E+02, 7.9388E+02,
     + 8.6067E+02, 9.2923E+02, 9.9953E+02, 1.0715E+03, 1.1451E+03,
     +  1.2204E+03, 1.2972E+03, 1.3756E+03, 1.4555E+03, 1.5369E+03,
     + 1.6198E+03, 1.7041E+03, 1.7898E+03, 1.8769E+03, 1.9654E+03,
     +  2.0552E+03, 2.1464E+03, 2.2389E+03, 2.3327E+03, 2.4277E+03,
     + 2.5240E+03, 2.6216E+03, 2.7203E+03, 2.8203E+03, 2.9216E+03,
     +  3.0239E+03, 3.1275E+03, 3.2322E+03, 3.3381E+03, 3.4451E+03,
     + 3.5532E+03, 3.6624E+03, 3.7728E+03, 3.8842E+03, 3.9967E+03,
     +  4.1103E+03, 4.2250E+03, 4.3407E+03, 4.4575E+03, 4.5754E+03,
     + 4.6942E+03, 4.8141E+03, 4.9351E+03, 5.0570E+03, 5.1799E+03,
     +  5.3039E+03, 5.4288E+03, 5.5548E+03, 5.6818E+03, 5.8097E+03,
     + 5.9386E+03, 6.0685E+03, 6.1995E+03, 6.3313E+03, 6.4641E+03,
     +  6.5980E+03, 6.7327E+03, 6.8685E+03, 7.0052E+03, 7.1428E+03,
     + 7.2815E+03, 7.4211E+03, 7.5616E+03, 7.7031E+03, 7.8456E+03,
     +  7.9890E+03, 8.1334E+03, 8.2787E+03, 8.4251E+03, 8.5723E+03,
     + 8.7206E+03, 8.8698E+03, 9.0200E+03, 9.1711E+03, 9.3231E+03,
     +  9.4762E+03, 9.6303E+03, 9.7853E+03, 9.9413E+03, 1.0098E+04,
     + 1.0256E+04, 1.0415E+04, 1.0575E+04, 1.0736E+04, 1.0898E+04,
     +  1.1061E+04, 1.1225E+04, 1.1389E+04, 1.1555E+04, 1.1722E+04,
     + 1.1890E+04, 1.2059E+04, 1.2229E+04, 1.2400E+04, 1.2572E+04,
     +  1.2745E+04, 1.2919E+04, 1.3094E+04, 1.3270E+04, 1.3448E+04,
     + 1.3626E+04, 1.3805E+04, 1.3986E+04, 1.4167E+04, 1.4350E+04,
     +  1.4533E+04, 1.4718E+04, 1.4903E+04, 1.5090E+04, 1.5278E+04,
     + 1.5467E+04, 1.5657E+04, 1.5849E+04, 1.6041E+04, 1.6235E+04,
     +  1.6429E+04, 1.6625E+04, 1.6822E+04, 1.7020E+04, 1.7219E+04,
     + 1.7419E+04, 1.7621E+04, 1.7823E+04, 1.8027E+04, 1.8232E+04,
     +  1.8438E+04, 1.8645E+04, 1.8854E+04, 1.9064E+04, 1.9275E+04,
     + 1.9487E+04, 1.9700E+04, 1.9915E+04, 2.0131E+04, 2.0348E+04,
     +  2.0566E+04, 2.0785E+04, 2.1006E+04, 2.1228E+04, 2.1451E+04,
     + 2.1676E+04, 2.1902E+04, 2.2129E+04, 2.2357E+04, 2.2587E+04,
     +  2.2818E+04, 2.3050E+04, 2.3284E+04, 2.3519E+04, 2.3755E+04,
     + 2.3993E+04, 2.4232E+04, 2.4472E+04, 2.4713E+04, 2.4956E+04,
     +  2.5201E+04, 2.5446E+04, 2.5694E+04, 2.5942E+04, 2.6192E+04,
     + 2.6443E+04, 2.6696E+04, 2.6950E+04, 2.7205E+04, 2.7462E+04,
     +  2.7720E+04, 2.7980E+04, 2.8241E+04, 2.8504E+04, 2.8768E+04,
     + 2.9033E+04, 2.9300E+04, 2.9568E+04, 2.9838E+04, 3.0110E+04,
     +  3.0383E+04, 3.0657E+04, 3.0933E+04, 3.1210E+04, 3.1489E+04,
     + 3.1769E+04, 3.2051E+04, 3.2335E+04, 3.2620E+04, 3.2906E+04,
     +  3.3194E+04, 3.3484E+04, 3.3775E+04, 3.4068E+04, 3.4363E+04,
     + 3.4658E+04, 3.4956E+04, 3.5255E+04, 3.5556E+04, 3.5858E+04,
     +  3.6162E+04, 3.6468E+04, 3.6775E+04, 3.7084E+04, 3.7394E+04,
     + 3.7707E+04, 3.8021E+04, 3.8336E+04, 3.8653E+04, 3.8972E+04,
     +  3.9293E+04, 3.9615E+04, 3.9939E+04, 4.0265E+04, 4.0592E+04,
     + 4.0921E+04, 4.1252E+04, 4.1585E+04, 4.1919E+04, 4.2255E+04,
     +  4.2593E+04, 4.2933E+04, 4.3274E+04, 4.3617E+04, 4.3962E+04,
     + 4.4309E+04, 4.4657E+04, 4.5008E+04, 4.5360E+04, 4.5714E+04,
     +  4.6069E+04, 4.6427E+04, 4.6786E+04, 4.7148E+04, 4.7511E+04,
     + 4.7876E+04, 4.8243E+04, 4.8611E+04, 4.8982E+04, 4.9355E+04,
     +  4.9729E+04, 5.0105E+04, 5.0483E+04, 5.0864E+04, 5.1246E+04,
     + 5.1630E+04, 5.2016E+04, 5.2404E+04, 5.2793E+04, 5.3185E+04,
     +  5.3579E+04, 5.3975E+04, 5.4373E+04, 5.4772E+04, 5.5174E+04,
     + 5.5577E+04, 5.5983E+04, 5.6391E+04, 5.6801E+04, 5.7212E+04,
     +  5.7626E+04, 5.8042E+04, 5.8460E+04, 5.8880E+04, 5.9302E+04,
     + 5.9726E+04, 6.0152E+04, 6.0581E+04, 6.1011E+04, 6.1443E+04,
     +  6.1878E+04, 6.2315E+04, 6.2754E+04, 6.3195E+04, 6.3638E+04,
     + 6.4083E+04, 6.4531E+04, 6.4980E+04, 6.5432E+04, 6.5886E+04,
     +  6.6342E+04, 6.6800E+04, 6.7261E+04, 6.7724E+04, 6.8189E+04,
     + 6.8656E+04, 6.9126E+04, 6.9598E+04, 7.0071E+04, 7.0548E+04,
     +  7.1026E+04, 7.1507E+04, 7.1990E+04, 7.2475E+04, 7.2963E+04,
     + 7.3452E+04, 7.3944E+04, 7.4439E+04, 7.4936E+04, 7.5435E+04,
     +  7.5937E+04, 7.6441E+04, 7.6947E+04, 7.7456E+04, 7.7966E+04,
     + 7.8480E+04, 7.8995E+04, 7.9514E+04, 8.0034E+04, 8.0557E+04,
     +  8.1082E+04, 8.1611E+04, 8.2141E+04, 8.2673E+04, 8.3209E+04,
     + 8.3746E+04, 8.4286E+04, 8.4829E+04, 8.5374E+04, 8.5921E+04,
     +  8.6471E+04, 8.7024E+04, 8.7579E+04, 8.8137E+04, 8.8697E+04,
     + 8.9259E+04, 8.9824E+04, 9.0392E+04, 9.0963E+04, 9.1536E+04,
     +  9.2111E+04, 9.2689E+04, 9.3270E+04, 9.3853E+04, 9.4439E+04,
     + 9.5028E+04, 9.5619E+04, 9.6212E+04, 9.6809E+04, 9.7408E+04,
     +  9.8010E+04, 9.8614E+04, 9.9222E+04, 9.9831E+04, 1.0044E+05,
     + 1.0106E+05, 1.0168E+05, 1.0230E+05, 1.0292E+05, 1.0355E+05,
     +  1.0418E+05, 1.0481E+05, 1.0544E+05, 1.0608E+05, 1.0672E+05,
     + 1.0736E+05, 1.0801E+05, 1.0866E+05, 1.0931E+05, 1.0996E+05,
     +  1.1062E+05, 1.1128E+05, 1.1194E+05, 1.1261E+05, 1.1328E+05,
     + 1.1395E+05, 1.1462E+05, 1.1530E+05, 1.1598E+05, 1.1666E+05,
     +  1.1735E+05, 1.1804E+05, 1.1873E+05, 1.1942E+05, 1.2012E+05,
     + 1.2082E+05, 1.2152E+05, 1.2223E+05, 1.2294E+05, 1.2365E+05,
     +  1.2437E+05, 1.2509E+05, 1.2581E+05, 1.2653E+05, 1.2726E+05,
     + 1.2799E+05, 1.2873E+05, 1.2946E+05, 1.3020E+05, 1.3095E+05,
     +  1.3169E+05, 1.3244E+05, 1.3319E+05, 1.3395E+05, 1.3471E+05,
     + 1.3547E+05, 1.3624E+05, 1.3700E+05, 1.3778E+05, 1.3855E+05,
     +  1.3933E+05, 1.4011E+05, 1.4090E+05, 1.4168E+05, 1.4247E+05,
     + 1.4327E+05, 1.4407E+05, 1.4487E+05, 1.4567E+05, 1.4648E+05,
     +  1.4729E+05, 1.4810E+05, 1.4892E+05, 1.4974E+05, 1.5056E+05,
     + 1.5139E+05, 1.5222E+05, 1.5305E+05, 1.5389E+05, 1.5473E+05,
     +  1.5558E+05, 1.5642E+05, 1.5728E+05, 1.5813E+05, 1.5899E+05,
     + 1.5985E+05, 1.6071E+05, 1.6158E+05, 1.6245E+05, 1.6333E+05,
     +  1.6421E+05, 1.6509E+05, 1.6597E+05, 1.6686E+05, 1.6776E+05,
     + 1.6865E+05, 1.6955E+05, 1.7046E+05, 1.7136E+05, 1.7227E+05,
     +  1.7319E+05, 1.7411E+05, 1.7503E+05, 1.7595E+05, 1.7688E+05,
     + 1.7781E+05, 1.7875E+05, 1.7969E+05, 1.8063E+05, 1.8158E+05,
     +  1.8253E+05, 1.8348E+05, 1.8444E+05, 1.8540E+05, 1.8637E+05,
     + 1.8734E+05, 1.8831E+05, 1.8929E+05, 1.9027E+05, 1.9125E+05,
     +  1.9224E+05, 1.9323E+05, 1.9423E+05, 1.9523E+05, 1.9623E+05,
     + 1.9724E+05, 1.9825E+05, 1.9927E+05, 2.0029E+05, 2.0131E+05,
     +  2.0234E+05, 2.0337E+05, 2.0440E+05, 2.0544E+05, 2.0649E+05,
     + 2.0753E+05, 2.0858E+05, 2.0964E+05, 2.1070E+05, 2.1176E+05,
     +  2.1283E+05, 2.1390E+05, 2.1497E+05, 2.1605E+05, 2.1713E+05,
     + 2.1822E+05, 2.1931E+05, 2.2041E+05, 2.2151E+05, 2.2261E+05,
     +  2.2372E+05, 2.2483E+05, 2.2595E+05, 2.2707E+05, 2.2819E+05,
     + 2.2932E+05, 2.3045E+05, 2.3159E+05, 2.3273E+05, 2.3387E+05,
     +  2.3502E+05, 2.3618E+05, 2.3734E+05, 2.3850E+05, 2.3966E+05,
     + 2.4084E+05, 2.4201E+05, 2.4319E+05, 2.4437E+05, 2.4556E+05,
     +  2.4675E+05, 2.4795E+05, 2.4915E+05, 2.5036E+05, 2.5157E+05,
     + 2.5279E+05, 2.5400E+05, 2.5523E+05, 2.5646E+05, 2.5769E+05,
     +  2.5893E+05, 2.6017E+05, 2.6141E+05, 2.6266E+05, 2.6392E+05,
     + 2.6518E+05, 2.6644E+05, 2.6771E+05, 2.6898E+05, 2.7026E+05,
     +  2.7154E+05, 2.7283E+05, 2.7412E+05, 2.7542E+05, 2.7672E+05,
     + 2.7802E+05, 2.7933E+05, 2.8065E+05, 2.8197E+05, 2.8329E+05,
     +  2.8462E+05, 2.8596E+05, 2.8729E+05, 2.8864E+05, 2.8999E+05,
     + 2.9134E+05, 2.9270E+05, 2.9406E+05, 2.9543E+05, 2.9680E+05,
     +  2.9817E+05, 2.9956E+05, 3.0094E+05, 3.0233E+05, 3.0373E+05,
     + 3.0513E+05, 3.0654E+05, 3.0795E+05, 3.0936E+05, 3.1079E+05,
     +  3.1221E+05, 3.1364E+05, 3.1508E+05, 3.1652E+05, 3.1796E+05,
     + 3.1942E+05, 3.2087E+05, 3.2233E+05, 3.2380E+05, 3.2527E+05,
     +  3.2675E+05, 3.2823E+05, 3.2972E+05, 3.3121E+05, 3.3270E+05,
     + 3.3421E+05, 3.3571E+05, 3.3722E+05, 3.3874E+05, 3.4026E+05,
     +  3.4179E+05, 3.4333E+05, 3.4486E+05, 3.4641E+05, 3.4796E+05,
     + 3.4951E+05, 3.5107E+05, 3.5264E+05, 3.5421E+05, 3.5578E+05,
     +  3.5736E+05, 3.5895E+05, 3.6054E+05, 3.6214E+05, 3.6374E+05,
     + 3.6535E+05, 3.6696E+05, 3.6858E+05, 3.7020E+05, 3.7183E+05,
     +  3.7347E+05, 3.7511E+05, 3.7675E+05, 3.7840E+05, 3.8006E+05,
     + 3.8173E+05, 3.8339E+05, 3.8507E+05, 3.8675E+05, 3.8843E+05,
     +  3.9012E+05, 3.9182E+05, 3.9352E+05, 3.9523E+05, 3.9694E+05,
     + 3.9866E+05, 4.0038E+05, 4.0211E+05, 4.0385E+05, 4.0559E+05,
     +  4.0734E+05, 4.0909E+05, 4.1085E+05, 4.1262E+05, 4.1439E+05,
     + 4.1616E+05, 4.1795E+05, 4.1973E+05, 4.2153E+05, 4.2333E+05,
     +  4.2513E+05, 4.2695E+05, 4.2876E+05, 4.3059E+05, 4.3242E+05,
     + 4.3425E+05, 4.3609E+05, 4.3794E+05, 4.3979E+05, 4.4165E+05,
     +  4.4352E+05, 4.4539E+05, 4.4727E+05, 4.4915E+05, 4.5104E+05,
     + 4.5294E+05, 4.5484E+05, 4.5675E+05, 4.5866E+05, 4.6058E+05,
     +  4.6251E+05, 4.6444E+05, 4.6638E+05, 4.6833E+05, 4.7028E+05,
     + 4.7223E+05, 4.7420E+05, 4.7617E+05, 4.7815E+05, 4.8013E+05,
     +  4.8212E+05, 4.8411E+05, 4.8612E+05, 4.8812E+05, 4.9014E+05,
     + 4.9216E+05, 4.9419E+05, 4.9622E+05, 4.9826E+05, 5.0031E+05,
     +  5.0236E+05, 5.0442E+05, 5.0649E+05, 5.0856E+05, 5.1064E+05,
     + 5.1272E+05, 5.1482E+05, 5.1692E+05, 5.1902E+05, 5.2113E+05,
     +  5.2325E+05, 5.2538E+05, 5.2751E+05, 5.2965E+05, 5.3179E+05,
     + 5.3395E+05, 5.3610E+05, 5.3827E+05, 5.4044E+05, 5.4262E+05,
     +  5.4481E+05, 5.4700E+05, 5.4920E+05, 5.5140E+05, 5.5362E+05,
     + 5.5584E+05, 5.5806E+05, 5.6030E+05, 5.6254E+05, 5.6478E+05,
     +  5.6704E+05, 5.6930E+05, 5.7157E+05, 5.7384E+05, 5.7613E+05,
     + 5.7842E+05, 5.8071E+05, 5.8301E+05, 5.8532E+05, 5.8764E+05,
     +  5.8996E+05, 5.9230E+05, 5.9463E+05, 5.9698E+05, 5.9933E+05,
     + 6.0169E+05, 6.0406E+05, 6.0644E+05, 6.0882E+05, 6.1121E+05,
     +  6.1360E+05, 6.1600E+05, 6.1841E+05, 6.2083E+05, 6.2326E+05,
     + 6.2569E+05, 6.2813E+05, 6.3058E+05, 6.3303E+05, 6.3549E+05,
     +  6.3796E+05, 6.4044E+05, 6.4292E+05, 6.4541E+05, 6.4791E+05,
     + 6.5042E+05, 6.5293E+05, 6.5545E+05, 6.5798E+05, 6.6052E+05,
     +  6.6306E+05, 6.6562E+05, 6.6817E+05, 6.7074E+05, 6.7332E+05,
     + 6.7590E+05, 6.7849E+05, 6.8109E+05, 6.8369E+05, 6.8630E+05,
     +  6.8892E+05, 6.9155E+05, 6.9419E+05, 6.9683E+05, 6.9948E+05,
     + 7.0214E+05, 7.0481E+05, 7.0748E+05, 7.1017E+05, 7.1286E+05,
     +  7.1556E+05, 7.1826E+05, 7.2098E+05, 7.2370E+05, 7.2643E+05,
     + 7.2917E+05, 7.3192E+05, 7.3467E+05, 7.3743E+05, 7.4020E+05,
     +  7.4298E+05, 7.4577E+05, 7.4856E+05, 7.5136E+05, 7.5418E+05,
     + 7.5699E+05, 7.5982E+05, 7.6266E+05, 7.6550E+05, 7.6835E+05,
     +  7.7121E+05, 7.7408E+05, 7.7696E+05, 7.7984E+05, 7.8273E+05,
     + 7.8564E+05, 7.8855E+05, 7.9146E+05, 7.9439E+05, 7.9732E+05,
     +  8.0027E+05, 8.0322E+05, 8.0618E+05, 8.0915E+05, 8.1212E+05,
     + 8.1511E+05, 8.1810E+05, 8.2110E+05, 8.2412E+05, 8.2714E+05,
     +  8.3016E+05, 8.3320E+05, 8.3625E+05, 8.3930E+05, 8.4236E+05,
     + 8.4543E+05, 8.4851E+05, 8.5160E+05, 8.5470E+05, 8.5781E+05,
     +  8.6092E+05, 8.6404E+05, 8.6718E+05, 8.7032E+05, 8.7347E+05,
     + 8.7663E+05, 8.7979E+05, 8.8297E+05, 8.8616E+05, 8.8935E+05,
     +  8.9256E+05, 8.9577E+05, 8.9899E+05, 9.0222E+05, 9.0546E+05,
     + 9.0871E+05, 9.1196E+05, 9.1523E+05, 9.1850E+05, 9.2179E+05,
     +  9.2508E+05, 9.2839E+05, 9.3170E+05, 9.3502E+05, 9.3835E+05,
     + 9.4169E+05, 9.4504E+05, 9.4839E+05, 9.5176E+05, 9.5514E+05,
     +  9.5853E+05, 9.6192E+05, 9.6532E+05, 9.6874E+05, 9.7216E+05,
     + 9.7559E+05, 9.7904E+05, 9.8249E+05, 9.8595E+05, 9.8942E+05,
     +  9.9290E+05, 9.9639E+05, 9.9989E+05, 1.0034E+06, 1.0069E+06,
     + 1.0104E+06, 1.0140E+06, 1.0175E+06, 1.0211E+06, 1.0246E+06,
     +  1.0282E+06, 1.0318E+06, 1.0354E+06, 1.0390E+06, 1.0426E+06,
     + 1.0462E+06, 1.0499E+06, 1.0535E+06, 1.0572E+06, 1.0608E+06,
     +  1.0645E+06, 1.0682E+06, 1.0719E+06, 1.0756E+06, 1.0793E+06,
     + 1.0830E+06, 1.0868E+06, 1.0905E+06, 1.0943E+06, 1.0980E+06,
     +  1.1018E+06, 1.1056E+06, 1.1094E+06, 1.1132E+06, 1.1170E+06,
     + 1.1208E+06, 1.1247E+06, 1.1285E+06, 1.1324E+06, 1.1362E+06,
     +  1.1401E+06, 1.1440E+06, 1.1479E+06, 1.1518E+06, 1.1557E+06,
     + 1.1596E+06, 1.1636E+06, 1.1675E+06, 1.1715E+06, 1.1754E+06,
     +  1.1794E+06, 1.1834E+06, 1.1874E+06, 1.1914E+06, 1.1954E+06,
     + 1.1995E+06, 1.2035E+06, 1.2076E+06, 1.2116E+06, 1.2157E+06,
     +  1.2198E+06, 1.2239E+06, 1.2280E+06, 1.2321E+06, 1.2362E+06,
     + 1.2404E+06, 1.2445E+06, 1.2487E+06, 1.2528E+06, 1.2570E+06,
     +  1.2612E+06, 1.2654E+06, 1.2696E+06, 1.2738E+06, 1.2781E+06,
     + 1.2823E+06, 1.2866E+06, 1.2909E+06, 1.2951E+06, 1.2994E+06,
     +  1.3037E+06, 1.3080E+06, 1.3124E+06, 1.3167E+06, 1.3210E+06,
     + 1.3254E+06, 1.3298E+06, 1.3342E+06, 1.3385E+06, 1.3429E+06,
     +  1.3474E+06, 1.3518E+06, 1.3562E+06, 1.3607E+06, 1.3651E+06,
     + 1.3696E+06, 1.3741E+06, 1.3786E+06, 1.3831E+06, 1.3876E+06,
     +  1.3921E+06, 1.3967E+06, 1.4012E+06, 1.4058E+06, 1.4103E+06,
     + 1.4149E+06, 1.4195E+06, 1.4241E+06, 1.4287E+06, 1.4334E+06,
     +  1.4380E+06, 1.4427E+06, 1.4473E+06, 1.4520E+06, 1.4567E+06,
     + 1.4614E+06, 1.4661E+06, 1.4709E+06, 1.4756E+06, 1.4803E+06,
     +  1.4851E+06, 1.4899E+06, 1.4947E+06, 1.4995E+06, 1.5043E+06,
     + 1.5091E+06, 1.5139E+06, 1.5188E+06, 1.5236E+06, 1.5285E+06,
     +  1.5334E+06, 1.5383E+06, 1.5432E+06, 1.5481E+06, 1.5530E+06,
     + 1.5580E+06, 1.5629E+06, 1.5679E+06, 1.5729E+06, 1.5779E+06,
     +  1.5829E+06, 1.5879E+06, 1.5929E+06, 1.5979E+06, 1.6030E+06,
     + 1.6081E+06, 1.6131E+06, 1.6182E+06, 1.6233E+06, 1.6285E+06,
     +  1.6336E+06, 1.6387E+06, 1.6439E+06, 1.6490E+06, 1.6542E+06,
     + 1.6594E+06, 1.6646E+06, 1.6698E+06, 1.6751E+06, 1.6803E+06,
     +  1.6856E+06, 1.6908E+06, 1.6961E+06, 1.7014E+06, 1.7067E+06,
     + 1.7120E+06, 1.7174E+06, 1.7227E+06, 1.7281E+06, 1.7335E+06,
     +  1.7388E+06, 1.7442E+06, 1.7496E+06, 1.7551E+06, 1.7605E+06,
     + 1.7660E+06, 1.7714E+06, 1.7769E+06, 1.7824E+06, 1.7879E+06,
     +  1.7934E+06, 1.7989E+06, 1.8045E+06, 1.8100E+06, 1.8156E+06,
     + 1.8212E+06, 1.8268E+06, 1.8324E+06, 1.8380E+06, 1.8437E+06,
     +  1.8493E+06, 1.8550E+06, 1.8606E+06, 1.8663E+06, 1.8720E+06,
     + 1.8778E+06, 1.8835E+06, 1.8892E+06, 1.8950E+06, 1.9008E+06,
     +  1.9065E+06, 1.9123E+06, 1.9182E+06, 1.9240E+06, 1.9298E+06,
     + 1.9357E+06, 1.9416E+06, 1.9474E+06, 1.9533E+06, 1.9592E+06,
     +  1.9652E+06, 1.9711E+06, 1.9771E+06, 1.9830E+06, 1.9890E+06,
     + 1.9950E+06, 2.0010E+06, 2.0070E+06, 2.0131E+06, 2.0191E+06,
     +  2.0252E+06, 2.0313E+06, 2.0374E+06, 2.0435E+06, 2.0496E+06,
     + 2.0557E+06, 2.0619E+06, 2.0680E+06, 2.0742E+06, 2.0804E+06,
     +  2.0866E+06, 2.0928E+06, 2.0991E+06, 2.1053E+06, 2.1116E+06,
     + 2.1179E+06, 2.1242E+06, 2.1305E+06, 2.1368E+06, 2.1432E+06,
     +  2.1495E+06, 2.1559E+06, 2.1623E+06, 2.1687E+06, 2.1751E+06,
     + 2.1815E+06, 2.1879E+06, 2.1944E+06, 2.2009E+06, 2.2074E+06,
     +  2.2139E+06, 2.2204E+06, 2.2269E+06, 2.2335E+06, 2.2400E+06,
     + 2.2466E+06, 2.2532E+06, 2.2598E+06, 2.2664E+06, 2.2731E+06,
     +  2.2797E+06, 2.2864E+06, 2.2931E+06, 2.2998E+06, 2.3065E+06,
     + 2.3132E+06, 2.3200E+06, 2.3267E+06, 2.3335E+06, 2.3403E+06,
     +  2.3471E+06, 2.3539E+06, 2.3608E+06, 2.3676E+06, 2.3745E+06,
     + 2.3814E+06, 2.3883E+06, 2.3952E+06, 2.4021E+06, 2.4090E+06,
     +  2.4160E+06, 2.4230E+06, 2.4300E+06, 2.4370E+06, 2.4440E+06,
     + 2.4511E+06, 2.4581E+06, 2.4652E+06, 2.4723E+06, 2.4794E+06,
     +  2.4865E+06, 2.4936E+06, 2.5008E+06, 2.5080E+06, 2.5151E+06,
     + 2.5223E+06, 2.5295E+06, 2.5368E+06, 2.5440E+06, 2.5513E+06,
     +  2.5586E+06, 2.5659E+06, 2.5732E+06, 2.5805E+06, 2.5879E+06,
     + 2.5952E+06, 2.6026E+06, 2.6100E+06, 2.6174E+06, 2.6248E+06,
     +  2.6323E+06, 2.6397E+06, 2.6472E+06, 2.6547E+06, 2.6622E+06,
     + 2.6698E+06, 2.6773E+06, 2.6849E+06, 2.6924E+06, 2.7000E+06,
     +  2.7076E+06, 2.7153E+06, 2.7229E+06, 2.7306E+06, 2.7382E+06,
     + 2.7459E+06, 2.7537E+06, 2.7614E+06, 2.7691E+06, 2.7769E+06,
     +  2.7847E+06, 2.7925E+06, 2.8003E+06, 2.8081E+06, 2.8160E+06,
     + 2.8238E+06, 2.8317E+06, 2.8396E+06, 2.8475E+06, 2.8555E+06,
     +  2.8634E+06, 2.8714E+06, 2.8794E+06, 2.8874E+06, 2.8954E+06,
     + 2.9034E+06, 2.9115E+06, 2.9195E+06, 2.9276E+06, 2.9357E+06,
     +  2.9439E+06, 2.9520E+06, 2.9602E+06, 2.9683E+06, 2.9765E+06,
     + 2.9848E+06, 2.9930E+06, 3.0012E+06, 3.0095E+06, 3.0178E+06,
     +  3.0261E+06, 3.0344E+06, 3.0427E+06, 3.0511E+06, 3.0595E+06,
     + 3.0679E+06, 3.0763E+06, 3.0847E+06, 3.0931E+06, 3.1016E+06,
     +  3.1101E+06, 3.1186E+06, 3.1271E+06, 3.1356E+06, 3.1442E+06,
     + 3.1528E+06, 3.1614E+06, 3.1700E+06, 3.1786E+06, 3.1872E+06,
     +  3.1959E+06, 3.2046E+06, 3.2133E+06, 3.2220E+06, 3.2307E+06,
     + 3.2395E+06, 3.2483E+06, 3.2571E+06, 3.2659E+06, 3.2747E+06,
     +  3.2836E+06, 3.2924E+06, 3.3013E+06, 3.3102E+06, 3.3191E+06,
     + 3.3281E+06, 3.3371E+06, 3.3460E+06, 3.3550E+06, 3.3641E+06,
     +  3.3731E+06, 3.3821E+06, 3.3912E+06, 3.4003E+06, 3.4094E+06,
     + 3.4186E+06, 3.4277E+06, 3.4369E+06, 3.4461E+06, 3.4553E+06,
     +  3.4645E+06, 3.4738E+06, 3.4830E+06, 3.4923E+06, 3.5016E+06,
     + 3.5109E+06, 3.5203E+06, 3.5297E+06, 3.5390E+06, 3.5484E+06,
     +  3.5579E+06, 3.5673E+06, 3.5768E+06, 3.5862E+06, 3.5957E+06,
     + 3.6053E+06, 3.6148E+06, 3.6244E+06, 3.6339E+06, 3.6435E+06,
     +  3.6532E+06, 3.6628E+06, 3.6725E+06, 3.6821E+06, 3.6918E+06,
     + 3.7016E+06, 3.7113E+06, 3.7211E+06, 3.7308E+06, 3.7406E+06,
     +  3.7505E+06, 3.7603E+06, 3.7702E+06, 3.7800E+06, 3.7899E+06,
     + 3.7999E+06, 3.8098E+06, 3.8198E+06, 3.8297E+06, 3.8397E+06,
     +  3.8498E+06, 3.8598E+06, 3.8699E+06, 3.8800E+06, 3.8901E+06,
     + 3.9002E+06, 3.9103E+06, 3.9205E+06, 3.9307E+06, 3.9409E+06,
     +  3.9511E+06, 3.9614E+06, 3.9716E+06, 3.9819E+06, 3.9922E+06,
     + 4.0026E+06, 4.0129E+06, 4.0233E+06, 4.0337E+06, 4.0441E+06,
     +  4.0545E+06, 4.0650E+06, 4.0755E+06, 4.0860E+06, 4.0965E+06,
     + 4.1070E+06, 4.1176E+06, 4.1282E+06, 4.1388E+06, 4.1494E+06,
     +  4.1601E+06, 4.1707E+06, 4.1814E+06, 4.1921E+06, 4.2029E+06,
     + 4.2136E+06, 4.2244E+06, 4.2352E+06, 4.2460E+06, 4.2569E+06,
     +  4.2677E+06, 4.2786E+06, 4.2895E+06, 4.3005E+06, 4.3114E+06,
     + 4.3224E+06, 4.3334E+06, 4.3444E+06, 4.3554E+06, 4.3665E+06,
     +  4.3776E+06, 4.3887E+06, 4.3998E+06, 4.4110E+06, 4.4221E+06,
     + 4.4333E+06, 4.4445E+06, 4.4558E+06, 4.4670E+06, 4.4783E+06,
     +  4.4896E+06, 4.5009E+06, 4.5123E+06, 4.5237E+06, 4.5350E+06,
     + 4.5465E+06, 4.5579E+06, 4.5694E+06, 4.5808E+06, 4.5923E+06,
     +  4.6039E+06, 4.6154E+06, 4.6270E+06, 4.6386E+06, 4.6502E+06,
     + 4.6619E+06, 4.6735E+06, 4.6852E+06, 4.6969E+06, 4.7086E+06,
     +  4.7204E+06, 4.7322E+06, 4.7440E+06, 4.7558E+06, 4.7677E+06,
     + 4.7795E+06, 4.7914E+06, 4.8033E+06, 4.8153E+06, 4.8272E+06,
     +  4.8392E+06, 4.8512E+06, 4.8633E+06, 4.8753E+06, 4.8874E+06,
     + 4.8995E+06, 4.9116E+06, 4.9238E+06, 4.9360E+06, 4.9482E+06,
     +  4.9604E+06, 4.9726E+06, 4.9849E+06, 4.9972E+06, 5.0095E+06,
     + 5.0218E+06, 5.0342E+06, 5.0466E+06, 5.0590E+06, 5.0714E+06,
     +  5.0839E+06, 5.0964E+06, 5.1089E+06, 5.1214E+06, 5.1340E+06,
     + 5.1465E+06, 5.1592E+06, 5.1718E+06, 5.1844E+06, 5.1971E+06,
     +  5.2098E+06, 5.2225E+06, 5.2353E+06, 5.2480E+06, 5.2608E+06,
     + 5.2737E+06, 5.2865E+06, 5.2994E+06, 5.3123E+06, 5.3252E+06,
     +  5.3381E+06, 5.3511E+06, 5.3641E+06, 5.3771E+06, 5.3902E+06,
     + 5.4032E+06, 5.4163E+06, 5.4294E+06, 5.4426E+06, 5.4557E+06,
     +  5.4689E+06, 5.4822E+06, 5.4954E+06, 5.5087E+06, 5.5220E+06,
     + 5.5353E+06, 5.5486E+06, 5.5620E+06, 5.5754E+06, 5.5888E+06,
     +  5.6022E+06, 5.6157E+06, 5.6292E+06, 5.6427E+06, 5.6562E+06,
     + 5.6698E+06, 5.6834E+06, 5.6970E+06, 5.7107E+06, 5.7244E+06,
     +  5.7380E+06, 5.7518E+06, 5.7655E+06, 5.7793E+06, 5.7931E+06,
     + 5.8069E+06, 5.8208E+06, 5.8346E+06, 5.8485E+06, 5.8625E+06,
     +  5.8764E+06, 5.8904E+06, 5.9044E+06, 5.9185E+06, 5.9325E+06,
     + 5.9466E+06, 5.9607E+06, 5.9748E+06, 5.9890E+06, 6.0032E+06,
     +  6.0174E+06, 6.0317E+06, 6.0459E+06, 6.0602E+06, 6.0745E+06,
     + 6.0889E+06, 6.1033E+06, 6.1177E+06, 6.1321E+06, 6.1465E+06,
     +  6.1610E+06, 6.1755E+06, 6.1901E+06, 6.2046E+06, 6.2192E+06,
     + 6.2338E+06, 6.2485E+06, 6.2632E+06, 6.2778E+06, 6.2926E+06,
     +  6.3073E+06, 6.3221E+06, 6.3369E+06, 6.3517E+06, 6.3666E+06,
     + 6.3815E+06, 6.3964E+06, 6.4113E+06, 6.4263E+06, 6.4413E+06,
     +  6.4563E+06, 6.4714E+06, 6.4864E+06, 6.5015E+06, 6.5167E+06,
     + 6.5318E+06, 6.5470E+06, 6.5622E+06, 6.5775E+06, 6.5927E+06,
     +  6.6080E+06, 6.6234E+06, 6.6387E+06, 6.6541E+06, 6.6695E+06,
     + 6.6849E+06, 6.7004E+06, 6.7159E+06, 6.7314E+06, 6.7470E+06,
     +  6.7626E+06, 6.7782E+06, 6.7938E+06, 6.8095E+06, 6.8251E+06,
     + 6.8409E+06, 6.8566E+06, 6.8724E+06, 6.8882E+06, 6.9040E+06,
     +  6.9199E+06, 6.9358E+06, 6.9517E+06, 6.9676E+06, 6.9836E+06,
     + 6.9996E+06, 7.0156E+06, 7.0317E+06, 7.0478E+06, 7.0639E+06,
     +  7.0800E+06, 7.0962E+06, 7.1124E+06, 7.1287E+06, 7.1449E+06,
     + 7.1612E+06, 7.1775E+06, 7.1939E+06, 7.2103E+06, 7.2267E+06,
     +  7.2431E+06, 7.2596E+06, 7.2761E+06, 7.2926E+06, 7.3092E+06,
     + 7.3257E+06, 7.3424E+06, 7.3590E+06, 7.3757E+06, 7.3924E+06,
     +  7.4091E+06, 7.4259E+06, 7.4426E+06, 7.4595E+06, 7.4763E+06,
     + 7.4932E+06, 7.5101E+06, 7.5271E+06, 7.5440E+06, 7.5610E+06,
     +  7.5780E+06, 7.5951E+06, 7.6122E+06, 7.6293E+06, 7.6465E+06,
     + 7.6637E+06, 7.6809E+06, 7.6981E+06, 7.7154E+06, 7.7327E+06,
     +  7.7500E+06, 7.7674E+06, 7.7848E+06, 7.8022E+06, 7.8196E+06,
     + 7.8371E+06, 7.8546E+06, 7.8722E+06, 7.8898E+06, 7.9074E+06,
     +  7.9250E+06, 7.9427E+06, 7.9604E+06, 7.9781E+06, 7.9959E+06,
     + 8.0136E+06, 8.0315E+06, 8.0493E+06, 8.0672E+06, 8.0851E+06,
     +  8.1031E+06, 8.1211E+06, 8.1391E+06, 8.1571E+06, 8.1752E+06,
     + 8.1933E+06, 8.2114E+06, 8.2296E+06, 8.2478E+06, 8.2660E+06,
     +  8.2843E+06, 8.3026E+06, 8.3209E+06, 8.3392E+06, 8.3576E+06,
     + 8.3760E+06, 8.3945E+06, 8.4130E+06, 8.4315E+06, 8.4500E+06,
     +  8.4686E+06, 8.4872E+06, 8.5059E+06, 8.5245E+06, 8.5432E+06,
     + 8.5620E+06, 8.5808E+06, 8.5996E+06, 8.6184E+06, 8.6373E+06,
     +  8.6562E+06, 8.6751E+06, 8.6940E+06, 8.7131E+06, 8.7321E+06,
     + 8.7511E+06, 8.7702E+06, 8.7894E+06, 8.8085E+06, 8.8277E+06,
     +  8.8469E+06, 8.8662E+06, 8.8855E+06, 8.9048E+06, 8.9242E+06,
     + 8.9436E+06, 8.9630E+06, 8.9824E+06, 9.0019E+06, 9.0214E+06,
     +  9.0410E+06, 9.0606E+06, 9.0802E+06, 9.0999E+06, 9.1196E+06,
     + 9.1393E+06, 9.1590E+06, 9.1788E+06, 9.1986E+06, 9.2185E+06,
     +  9.2384E+06, 9.2583E+06, 9.2783E+06, 9.2982E+06, 9.3183E+06,
     + 9.3383E+06, 9.3584E+06, 9.3785E+06, 9.3987E+06, 9.4189E+06,
     +  9.4391E+06, 9.4594E+06, 9.4797E+06, 9.5000E+06, 9.5203E+06,
     + 9.5407E+06, 9.5612E+06, 9.5816E+06, 9.6021E+06, 9.6227E+06,
     +  9.6432E+06, 9.6638E+06, 9.6845E+06, 9.7052E+06, 9.7259E+06,
     + 9.7466E+06, 9.7674E+06, 9.7882E+06, 9.8090E+06, 9.8299E+06,
     +  9.8508E+06, 9.8718E+06, 9.8928E+06, 9.9138E+06, 9.9348E+06,
     + 9.9559E+06, 9.9770E+06, 9.9982E+06, 1.0019E+07, 1.0041E+07,
     +  1.0062E+07, 1.0083E+07, 1.0105E+07, 1.0126E+07, 1.0147E+07,
     + 1.0169E+07, 1.0190E+07, 1.0212E+07, 1.0233E+07, 1.0255E+07,
     +  1.0276E+07, 1.0298E+07, 1.0320E+07, 1.0342E+07, 1.0363E+07,
     + 1.0385E+07, 1.0407E+07, 1.0429E+07, 1.0451E+07, 1.0473E+07,
     +  1.0495E+07, 1.0517E+07, 1.0539E+07, 1.0561E+07, 1.0583E+07,
     + 1.0605E+07, 1.0627E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       167
      data (QofT( 2,J),J=1,2408)/ 4.0581E+00, 1.1636E+01, 2.0786E+01,
     +  3.1460E+01, 4.3490E+01, 5.6743E+01, 7.1119E+01, 8.6536E+01,
     + 1.0293E+02, 1.2024E+02, 1.3843E+02, 1.5746E+02, 1.7728E+02,
     +  1.9788E+02, 2.1921E+02, 2.4126E+02, 2.6400E+02, 2.8742E+02,
     + 3.1150E+02, 3.3621E+02, 3.8747E+02, 4.4111E+02, 4.9703E+02,
     +  5.5512E+02, 6.1532E+02, 6.7755E+02, 7.4175E+02, 8.0786E+02,
     + 8.7583E+02, 9.4561E+02, 1.0171E+03, 1.0904E+03, 1.1653E+03,
     +  1.2419E+03, 1.3201E+03, 1.3999E+03, 1.4812E+03, 1.5641E+03,
     + 1.6484E+03, 1.7342E+03, 1.8214E+03, 1.9101E+03, 2.0001E+03,
     +  2.0915E+03, 2.1843E+03, 2.2784E+03, 2.3739E+03, 2.4706E+03,
     + 2.5686E+03, 2.6679E+03, 2.7684E+03, 2.8702E+03, 2.9732E+03,
     +  3.0774E+03, 3.1828E+03, 3.2893E+03, 3.3971E+03, 3.5060E+03,
     + 3.6160E+03, 3.7272E+03, 3.8395E+03, 3.9529E+03, 4.0674E+03,
     +  4.1830E+03, 4.2998E+03, 4.4175E+03, 4.5364E+03, 4.6563E+03,
     + 4.7773E+03, 4.8993E+03, 5.0224E+03, 5.1465E+03, 5.2716E+03,
     +  5.3977E+03, 5.5249E+03, 5.6531E+03, 5.7823E+03, 5.9125E+03,
     + 6.0437E+03, 6.1759E+03, 6.3092E+03, 6.4434E+03, 6.5785E+03,
     +  6.7147E+03, 6.8519E+03, 6.9900E+03, 7.1292E+03, 7.2692E+03,
     + 7.4104E+03, 7.5524E+03, 7.6954E+03, 7.8395E+03, 7.9845E+03,
     +  8.1304E+03, 8.2774E+03, 8.4253E+03, 8.5742E+03, 8.7241E+03,
     + 8.8750E+03, 9.0268E+03, 9.1796E+03, 9.3335E+03, 9.4882E+03,
     +  9.6440E+03, 9.8008E+03, 9.9585E+03, 1.0117E+04, 1.0277E+04,
     + 1.0438E+04, 1.0600E+04, 1.0762E+04, 1.0926E+04, 1.1091E+04,
     +  1.1257E+04, 1.1423E+04, 1.1591E+04, 1.1760E+04, 1.1930E+04,
     + 1.2101E+04, 1.2273E+04, 1.2446E+04, 1.2620E+04, 1.2795E+04,
     +  1.2971E+04, 1.3148E+04, 1.3326E+04, 1.3505E+04, 1.3686E+04,
     + 1.3867E+04, 1.4050E+04, 1.4233E+04, 1.4418E+04, 1.4604E+04,
     +  1.4790E+04, 1.4978E+04, 1.5167E+04, 1.5358E+04, 1.5549E+04,
     + 1.5741E+04, 1.5935E+04, 1.6129E+04, 1.6325E+04, 1.6522E+04,
     +  1.6720E+04, 1.6919E+04, 1.7120E+04, 1.7321E+04, 1.7524E+04,
     + 1.7728E+04, 1.7933E+04, 1.8139E+04, 1.8346E+04, 1.8555E+04,
     +  1.8765E+04, 1.8976E+04, 1.9188E+04, 1.9401E+04, 1.9616E+04,
     + 1.9832E+04, 2.0049E+04, 2.0268E+04, 2.0487E+04, 2.0708E+04,
     +  2.0930E+04, 2.1154E+04, 2.1378E+04, 2.1604E+04, 2.1831E+04,
     + 2.2060E+04, 2.2290E+04, 2.2521E+04, 2.2753E+04, 2.2987E+04,
     +  2.3222E+04, 2.3459E+04, 2.3696E+04, 2.3936E+04, 2.4176E+04,
     + 2.4418E+04, 2.4661E+04, 2.4905E+04, 2.5151E+04, 2.5398E+04,
     +  2.5647E+04, 2.5897E+04, 2.6149E+04, 2.6402E+04, 2.6656E+04,
     + 2.6911E+04, 2.7168E+04, 2.7427E+04, 2.7687E+04, 2.7948E+04,
     +  2.8211E+04, 2.8476E+04, 2.8741E+04, 2.9009E+04, 2.9277E+04,
     + 2.9547E+04, 2.9819E+04, 3.0092E+04, 3.0367E+04, 3.0643E+04,
     +  3.0921E+04, 3.1200E+04, 3.1481E+04, 3.1763E+04, 3.2047E+04,
     + 3.2332E+04, 3.2619E+04, 3.2907E+04, 3.3197E+04, 3.3489E+04,
     +  3.3782E+04, 3.4077E+04, 3.4374E+04, 3.4672E+04, 3.4971E+04,
     + 3.5272E+04, 3.5575E+04, 3.5880E+04, 3.6186E+04, 3.6493E+04,
     +  3.6803E+04, 3.7114E+04, 3.7427E+04, 3.7741E+04, 3.8057E+04,
     + 3.8375E+04, 3.8694E+04, 3.9016E+04, 3.9338E+04, 3.9663E+04,
     +  3.9989E+04, 4.0317E+04, 4.0647E+04, 4.0978E+04, 4.1311E+04,
     + 4.1646E+04, 4.1983E+04, 4.2321E+04, 4.2662E+04, 4.3004E+04,
     +  4.3347E+04, 4.3693E+04, 4.4041E+04, 4.4390E+04, 4.4741E+04,
     + 4.5094E+04, 4.5448E+04, 4.5805E+04, 4.6163E+04, 4.6524E+04,
     +  4.6886E+04, 4.7250E+04, 4.7615E+04, 4.7983E+04, 4.8352E+04,
     + 4.8724E+04, 4.9097E+04, 4.9473E+04, 4.9850E+04, 5.0229E+04,
     +  5.0610E+04, 5.0993E+04, 5.1378E+04, 5.1765E+04, 5.2154E+04,
     + 5.2544E+04, 5.2937E+04, 5.3332E+04, 5.3729E+04, 5.4127E+04,
     +  5.4528E+04, 5.4931E+04, 5.5336E+04, 5.5742E+04, 5.6151E+04,
     + 5.6562E+04, 5.6975E+04, 5.7390E+04, 5.7807E+04, 5.8226E+04,
     +  5.8647E+04, 5.9071E+04, 5.9496E+04, 5.9923E+04, 6.0353E+04,
     + 6.0784E+04, 6.1218E+04, 6.1654E+04, 6.2092E+04, 6.2532E+04,
     +  6.2975E+04, 6.3419E+04, 6.3866E+04, 6.4314E+04, 6.4766E+04,
     + 6.5219E+04, 6.5674E+04, 6.6132E+04, 6.6591E+04, 6.7054E+04,
     +  6.7518E+04, 6.7984E+04, 6.8453E+04, 6.8924E+04, 6.9397E+04,
     + 6.9872E+04, 7.0350E+04, 7.0831E+04, 7.1313E+04, 7.1798E+04,
     +  7.2284E+04, 7.2774E+04, 7.3265E+04, 7.3759E+04, 7.4256E+04,
     + 7.4754E+04, 7.5255E+04, 7.5758E+04, 7.6264E+04, 7.6772E+04,
     +  7.7282E+04, 7.7795E+04, 7.8310E+04, 7.8828E+04, 7.9348E+04,
     + 7.9871E+04, 8.0395E+04, 8.0923E+04, 8.1452E+04, 8.1985E+04,
     +  8.2519E+04, 8.3057E+04, 8.3596E+04, 8.4138E+04, 8.4683E+04,
     + 8.5230E+04, 8.5780E+04, 8.6332E+04, 8.6887E+04, 8.7444E+04,
     +  8.8004E+04, 8.8566E+04, 8.9131E+04, 8.9698E+04, 9.0269E+04,
     + 9.0841E+04, 9.1416E+04, 9.1994E+04, 9.2575E+04, 9.3158E+04,
     +  9.3743E+04, 9.4332E+04, 9.4922E+04, 9.5516E+04, 9.6112E+04,
     + 9.6712E+04, 9.7313E+04, 9.7917E+04, 9.8525E+04, 9.9134E+04,
     +  9.9747E+04, 1.0036E+05, 1.0098E+05, 1.0160E+05, 1.0222E+05,
     + 1.0285E+05, 1.0348E+05, 1.0411E+05, 1.0475E+05, 1.0538E+05,
     +  1.0602E+05, 1.0667E+05, 1.0731E+05, 1.0796E+05, 1.0861E+05,
     + 1.0927E+05, 1.0992E+05, 1.1058E+05, 1.1125E+05, 1.1191E+05,
     +  1.1258E+05, 1.1325E+05, 1.1393E+05, 1.1460E+05, 1.1528E+05,
     + 1.1597E+05, 1.1665E+05, 1.1734E+05, 1.1803E+05, 1.1873E+05,
     +  1.1943E+05, 1.2013E+05, 1.2083E+05, 1.2154E+05, 1.2225E+05,
     + 1.2296E+05, 1.2368E+05, 1.2440E+05, 1.2512E+05, 1.2584E+05,
     +  1.2657E+05, 1.2730E+05, 1.2804E+05, 1.2878E+05, 1.2952E+05,
     + 1.3026E+05, 1.3101E+05, 1.3176E+05, 1.3251E+05, 1.3327E+05,
     +  1.3403E+05, 1.3479E+05, 1.3556E+05, 1.3632E+05, 1.3710E+05,
     + 1.3787E+05, 1.3865E+05, 1.3943E+05, 1.4022E+05, 1.4101E+05,
     +  1.4180E+05, 1.4259E+05, 1.4339E+05, 1.4419E+05, 1.4500E+05,
     + 1.4581E+05, 1.4662E+05, 1.4743E+05, 1.4825E+05, 1.4907E+05,
     +  1.4990E+05, 1.5073E+05, 1.5156E+05, 1.5239E+05, 1.5323E+05,
     + 1.5407E+05, 1.5492E+05, 1.5577E+05, 1.5662E+05, 1.5747E+05,
     +  1.5833E+05, 1.5920E+05, 1.6006E+05, 1.6093E+05, 1.6180E+05,
     + 1.6268E+05, 1.6356E+05, 1.6444E+05, 1.6533E+05, 1.6622E+05,
     +  1.6712E+05, 1.6801E+05, 1.6892E+05, 1.6982E+05, 1.7073E+05,
     + 1.7164E+05, 1.7256E+05, 1.7348E+05, 1.7440E+05, 1.7533E+05,
     +  1.7626E+05, 1.7719E+05, 1.7813E+05, 1.7907E+05, 1.8002E+05,
     + 1.8096E+05, 1.8192E+05, 1.8287E+05, 1.8383E+05, 1.8480E+05,
     +  1.8576E+05, 1.8674E+05, 1.8771E+05, 1.8869E+05, 1.8967E+05,
     + 1.9066E+05, 1.9165E+05, 1.9264E+05, 1.9364E+05, 1.9464E+05,
     +  1.9565E+05, 1.9666E+05, 1.9767E+05, 1.9869E+05, 1.9971E+05,
     + 2.0074E+05, 2.0177E+05, 2.0280E+05, 2.0384E+05, 2.0488E+05,
     +  2.0592E+05, 2.0697E+05, 2.0803E+05, 2.0908E+05, 2.1015E+05,
     + 2.1121E+05, 2.1228E+05, 2.1335E+05, 2.1443E+05, 2.1551E+05,
     +  2.1660E+05, 2.1769E+05, 2.1878E+05, 2.1988E+05, 2.2098E+05,
     + 2.2209E+05, 2.2320E+05, 2.2431E+05, 2.2543E+05, 2.2655E+05,
     +  2.2768E+05, 2.2881E+05, 2.2995E+05, 2.3109E+05, 2.3223E+05,
     + 2.3338E+05, 2.3454E+05, 2.3569E+05, 2.3685E+05, 2.3802E+05,
     +  2.3919E+05, 2.4036E+05, 2.4154E+05, 2.4272E+05, 2.4391E+05,
     + 2.4510E+05, 2.4630E+05, 2.4750E+05, 2.4871E+05, 2.4991E+05,
     +  2.5113E+05, 2.5235E+05, 2.5357E+05, 2.5480E+05, 2.5603E+05,
     + 2.5727E+05, 2.5851E+05, 2.5975E+05, 2.6100E+05, 2.6225E+05,
     +  2.6351E+05, 2.6478E+05, 2.6604E+05, 2.6732E+05, 2.6859E+05,
     + 2.6988E+05, 2.7116E+05, 2.7245E+05, 2.7375E+05, 2.7505E+05,
     +  2.7636E+05, 2.7767E+05, 2.7898E+05, 2.8030E+05, 2.8162E+05,
     + 2.8295E+05, 2.8429E+05, 2.8562E+05, 2.8697E+05, 2.8831E+05,
     +  2.8967E+05, 2.9102E+05, 2.9239E+05, 2.9375E+05, 2.9512E+05,
     + 2.9650E+05, 2.9788E+05, 2.9927E+05, 3.0066E+05, 3.0206E+05,
     +  3.0346E+05, 3.0487E+05, 3.0628E+05, 3.0769E+05, 3.0911E+05,
     + 3.1054E+05, 3.1197E+05, 3.1341E+05, 3.1485E+05, 3.1629E+05,
     +  3.1775E+05, 3.1920E+05, 3.2066E+05, 3.2213E+05, 3.2360E+05,
     + 3.2508E+05, 3.2656E+05, 3.2805E+05, 3.2954E+05, 3.3104E+05,
     +  3.3254E+05, 3.3405E+05, 3.3556E+05, 3.3708E+05, 3.3860E+05,
     + 3.4013E+05, 3.4166E+05, 3.4320E+05, 3.4475E+05, 3.4629E+05,
     +  3.4785E+05, 3.4941E+05, 3.5098E+05, 3.5255E+05, 3.5412E+05,
     + 3.5571E+05, 3.5729E+05, 3.5889E+05, 3.6048E+05, 3.6209E+05,
     +  3.6370E+05, 3.6531E+05, 3.6693E+05, 3.6855E+05, 3.7018E+05,
     + 3.7182E+05, 3.7346E+05, 3.7511E+05, 3.7676E+05, 3.7842E+05,
     +  3.8008E+05, 3.8175E+05, 3.8343E+05, 3.8511E+05, 3.8680E+05,
     + 3.8849E+05, 3.9019E+05, 3.9189E+05, 3.9360E+05, 3.9532E+05,
     +  3.9704E+05, 3.9876E+05, 4.0049E+05, 4.0223E+05, 4.0397E+05,
     + 4.0572E+05, 4.0748E+05, 4.0924E+05, 4.1101E+05, 4.1278E+05,
     +  4.1456E+05, 4.1634E+05, 4.1813E+05, 4.1993E+05, 4.2173E+05,
     + 4.2354E+05, 4.2535E+05, 4.2717E+05, 4.2900E+05, 4.3083E+05,
     +  4.3267E+05, 4.3451E+05, 4.3636E+05, 4.3822E+05, 4.4008E+05,
     + 4.4195E+05, 4.4382E+05, 4.4570E+05, 4.4759E+05, 4.4948E+05,
     +  4.5138E+05, 4.5328E+05, 4.5520E+05, 4.5711E+05, 4.5904E+05,
     + 4.6097E+05, 4.6290E+05, 4.6484E+05, 4.6679E+05, 4.6875E+05,
     +  4.7071E+05, 4.7267E+05, 4.7465E+05, 4.7663E+05, 4.7861E+05,
     + 4.8060E+05, 4.8260E+05, 4.8461E+05, 4.8662E+05, 4.8864E+05,
     +  4.9066E+05, 4.9269E+05, 4.9473E+05, 4.9678E+05, 4.9883E+05,
     + 5.0088E+05, 5.0295E+05, 5.0502E+05, 5.0709E+05, 5.0917E+05,
     +  5.1127E+05, 5.1336E+05, 5.1546E+05, 5.1757E+05, 5.1969E+05,
     + 5.2181E+05, 5.2394E+05, 5.2608E+05, 5.2822E+05, 5.3037E+05,
     +  5.3253E+05, 5.3469E+05, 5.3686E+05, 5.3903E+05, 5.4122E+05,
     + 5.4341E+05, 5.4560E+05, 5.4781E+05, 5.5002E+05, 5.5224E+05,
     +  5.5446E+05, 5.5669E+05, 5.5893E+05, 5.6118E+05, 5.6343E+05,
     + 5.6569E+05, 5.6795E+05, 5.7023E+05, 5.7251E+05, 5.7479E+05,
     +  5.7709E+05, 5.7939E+05, 5.8170E+05, 5.8401E+05, 5.8634E+05,
     + 5.8867E+05, 5.9100E+05, 5.9335E+05, 5.9570E+05, 5.9806E+05,
     +  6.0042E+05, 6.0279E+05, 6.0517E+05, 6.0756E+05, 6.0996E+05,
     + 6.1236E+05, 6.1477E+05, 6.1718E+05, 6.1961E+05, 6.2204E+05,
     +  6.2448E+05, 6.2692E+05, 6.2937E+05, 6.3184E+05, 6.3430E+05,
     + 6.3678E+05, 6.3926E+05, 6.4175E+05, 6.4425E+05, 6.4676E+05,
     +  6.4927E+05, 6.5179E+05, 6.5432E+05, 6.5685E+05, 6.5940E+05,
     + 6.6195E+05, 6.6450E+05, 6.6707E+05, 6.6965E+05, 6.7223E+05,
     +  6.7482E+05, 6.7741E+05, 6.8002E+05, 6.8263E+05, 6.8525E+05,
     + 6.8788E+05, 6.9051E+05, 6.9316E+05, 6.9581E+05, 6.9847E+05,
     +  7.0113E+05, 7.0381E+05, 7.0649E+05, 7.0918E+05, 7.1188E+05,
     + 7.1459E+05, 7.1730E+05, 7.2002E+05, 7.2275E+05, 7.2549E+05,
     +  7.2824E+05, 7.3099E+05, 7.3376E+05, 7.3653E+05, 7.3930E+05,
     + 7.4209E+05, 7.4489E+05, 7.4769E+05, 7.5050E+05, 7.5332E+05,
     +  7.5615E+05, 7.5899E+05, 7.6183E+05, 7.6468E+05, 7.6754E+05,
     + 7.7041E+05, 7.7329E+05, 7.7617E+05, 7.7907E+05, 7.8197E+05,
     +  7.8488E+05, 7.8780E+05, 7.9073E+05, 7.9366E+05, 7.9661E+05,
     + 7.9956E+05, 8.0252E+05, 8.0549E+05, 8.0847E+05, 8.1146E+05,
     +  8.1445E+05, 8.1745E+05, 8.2047E+05, 8.2349E+05, 8.2652E+05,
     + 8.2956E+05, 8.3260E+05, 8.3566E+05, 8.3872E+05, 8.4180E+05,
     +  8.4488E+05, 8.4797E+05, 8.5107E+05, 8.5418E+05, 8.5729E+05,
     + 8.6042E+05, 8.6355E+05, 8.6669E+05, 8.6985E+05, 8.7301E+05,
     +  8.7618E+05, 8.7936E+05, 8.8255E+05, 8.8574E+05, 8.8895E+05,
     + 8.9216E+05, 8.9539E+05, 8.9862E+05, 9.0186E+05, 9.0511E+05,
     +  9.0837E+05, 9.1164E+05, 9.1492E+05, 9.1821E+05, 9.2151E+05,
     + 9.2481E+05, 9.2813E+05, 9.3145E+05, 9.3478E+05, 9.3813E+05,
     +  9.4148E+05, 9.4484E+05, 9.4821E+05, 9.5159E+05, 9.5498E+05,
     + 9.5838E+05, 9.6179E+05, 9.6520E+05, 9.6863E+05, 9.7207E+05,
     +  9.7551E+05, 9.7897E+05, 9.8243E+05, 9.8591E+05, 9.8939E+05,
     + 9.9288E+05, 9.9639E+05, 9.9990E+05, 1.0034E+06, 1.0070E+06,
     +  1.0105E+06, 1.0140E+06, 1.0176E+06, 1.0212E+06, 1.0248E+06,
     + 1.0284E+06, 1.0319E+06, 1.0356E+06, 1.0392E+06, 1.0428E+06,
     +  1.0464E+06, 1.0501E+06, 1.0538E+06, 1.0574E+06, 1.0611E+06,
     + 1.0648E+06, 1.0685E+06, 1.0722E+06, 1.0759E+06, 1.0796E+06,
     +  1.0834E+06, 1.0871E+06, 1.0909E+06, 1.0947E+06, 1.0984E+06,
     + 1.1022E+06, 1.1060E+06, 1.1098E+06, 1.1137E+06, 1.1175E+06,
     +  1.1213E+06, 1.1252E+06, 1.1290E+06, 1.1329E+06, 1.1368E+06,
     + 1.1407E+06, 1.1446E+06, 1.1485E+06, 1.1524E+06, 1.1564E+06,
     +  1.1603E+06, 1.1643E+06, 1.1682E+06, 1.1722E+06, 1.1762E+06,
     + 1.1802E+06, 1.1842E+06, 1.1882E+06, 1.1922E+06, 1.1963E+06,
     +  1.2003E+06, 1.2044E+06, 1.2084E+06, 1.2125E+06, 1.2166E+06,
     + 1.2207E+06, 1.2248E+06, 1.2290E+06, 1.2331E+06, 1.2372E+06,
     +  1.2414E+06, 1.2456E+06, 1.2497E+06, 1.2539E+06, 1.2581E+06,
     + 1.2623E+06, 1.2666E+06, 1.2708E+06, 1.2750E+06, 1.2793E+06,
     +  1.2836E+06, 1.2878E+06, 1.2921E+06, 1.2964E+06, 1.3007E+06,
     + 1.3051E+06, 1.3094E+06, 1.3137E+06, 1.3181E+06, 1.3225E+06,
     +  1.3268E+06, 1.3312E+06, 1.3356E+06, 1.3400E+06, 1.3445E+06,
     + 1.3489E+06, 1.3533E+06, 1.3578E+06, 1.3623E+06, 1.3667E+06,
     +  1.3712E+06, 1.3757E+06, 1.3803E+06, 1.3848E+06, 1.3893E+06,
     + 1.3939E+06, 1.3984E+06, 1.4030E+06, 1.4076E+06, 1.4122E+06,
     +  1.4168E+06, 1.4214E+06, 1.4260E+06, 1.4307E+06, 1.4353E+06,
     + 1.4400E+06, 1.4447E+06, 1.4494E+06, 1.4541E+06, 1.4588E+06,
     +  1.4635E+06, 1.4682E+06, 1.4730E+06, 1.4778E+06, 1.4825E+06,
     + 1.4873E+06, 1.4921E+06, 1.4969E+06, 1.5017E+06, 1.5066E+06,
     +  1.5114E+06, 1.5163E+06, 1.5212E+06, 1.5260E+06, 1.5309E+06,
     + 1.5358E+06, 1.5408E+06, 1.5457E+06, 1.5506E+06, 1.5556E+06,
     +  1.5606E+06, 1.5655E+06, 1.5705E+06, 1.5755E+06, 1.5806E+06,
     + 1.5856E+06, 1.5906E+06, 1.5957E+06, 1.6007E+06, 1.6058E+06,
     +  1.6109E+06, 1.6160E+06, 1.6211E+06, 1.6263E+06, 1.6314E+06,
     + 1.6366E+06, 1.6417E+06, 1.6469E+06, 1.6521E+06, 1.6573E+06,
     +  1.6625E+06, 1.6678E+06, 1.6730E+06, 1.6783E+06, 1.6836E+06,
     + 1.6888E+06, 1.6941E+06, 1.6994E+06, 1.7048E+06, 1.7101E+06,
     +  1.7154E+06, 1.7208E+06, 1.7262E+06, 1.7316E+06, 1.7370E+06,
     + 1.7424E+06, 1.7478E+06, 1.7533E+06, 1.7587E+06, 1.7642E+06,
     +  1.7697E+06, 1.7752E+06, 1.7807E+06, 1.7862E+06, 1.7917E+06,
     + 1.7973E+06, 1.8028E+06, 1.8084E+06, 1.8140E+06, 1.8196E+06,
     +  1.8252E+06, 1.8308E+06, 1.8365E+06, 1.8421E+06, 1.8478E+06,
     + 1.8535E+06, 1.8592E+06, 1.8649E+06, 1.8706E+06, 1.8763E+06,
     +  1.8821E+06, 1.8878E+06, 1.8936E+06, 1.8994E+06, 1.9052E+06,
     + 1.9110E+06, 1.9169E+06, 1.9227E+06, 1.9286E+06, 1.9345E+06,
     +  1.9403E+06, 1.9462E+06, 1.9522E+06, 1.9581E+06, 1.9640E+06,
     + 1.9700E+06, 1.9760E+06, 1.9820E+06, 1.9880E+06, 1.9940E+06,
     +  2.0000E+06, 2.0060E+06, 2.0121E+06, 2.0182E+06, 2.0243E+06,
     + 2.0304E+06, 2.0365E+06, 2.0426E+06, 2.0487E+06, 2.0549E+06,
     +  2.0611E+06, 2.0673E+06, 2.0735E+06, 2.0797E+06, 2.0859E+06,
     + 2.0922E+06, 2.0984E+06, 2.1047E+06, 2.1110E+06, 2.1173E+06,
     +  2.1236E+06, 2.1299E+06, 2.1363E+06, 2.1427E+06, 2.1490E+06,
     + 2.1554E+06, 2.1618E+06, 2.1683E+06, 2.1747E+06, 2.1811E+06,
     +  2.1876E+06, 2.1941E+06, 2.2006E+06, 2.2071E+06, 2.2136E+06,
     + 2.2202E+06, 2.2267E+06, 2.2333E+06, 2.2399E+06, 2.2465E+06,
     +  2.2531E+06, 2.2597E+06, 2.2664E+06, 2.2731E+06, 2.2797E+06,
     + 2.2864E+06, 2.2931E+06, 2.2999E+06, 2.3066E+06, 2.3134E+06,
     +  2.3201E+06, 2.3269E+06, 2.3337E+06, 2.3405E+06, 2.3474E+06,
     + 2.3542E+06, 2.3611E+06, 2.3680E+06, 2.3749E+06, 2.3818E+06,
     +  2.3887E+06, 2.3956E+06, 2.4026E+06, 2.4096E+06, 2.4166E+06,
     + 2.4236E+06, 2.4306E+06, 2.4376E+06, 2.4447E+06, 2.4517E+06,
     +  2.4588E+06, 2.4659E+06, 2.4730E+06, 2.4802E+06, 2.4873E+06,
     + 2.4945E+06, 2.5017E+06, 2.5089E+06, 2.5161E+06, 2.5233E+06,
     +  2.5306E+06, 2.5378E+06, 2.5451E+06, 2.5524E+06, 2.5597E+06,
     + 2.5670E+06, 2.5744E+06, 2.5817E+06, 2.5891E+06, 2.5965E+06,
     +  2.6039E+06, 2.6114E+06, 2.6188E+06, 2.6263E+06, 2.6337E+06,
     + 2.6412E+06, 2.6487E+06, 2.6563E+06, 2.6638E+06, 2.6714E+06,
     +  2.6789E+06, 2.6865E+06, 2.6941E+06, 2.7018E+06, 2.7094E+06,
     + 2.7171E+06, 2.7248E+06, 2.7324E+06, 2.7402E+06, 2.7479E+06,
     +  2.7556E+06, 2.7634E+06, 2.7712E+06, 2.7790E+06, 2.7868E+06,
     + 2.7946E+06, 2.8025E+06, 2.8103E+06, 2.8182E+06, 2.8261E+06,
     +  2.8340E+06, 2.8420E+06, 2.8499E+06, 2.8579E+06, 2.8659E+06,
     + 2.8739E+06, 2.8819E+06, 2.8899E+06, 2.8980E+06, 2.9061E+06,
     +  2.9142E+06, 2.9223E+06, 2.9304E+06, 2.9385E+06, 2.9467E+06,
     + 2.9549E+06, 2.9631E+06, 2.9713E+06, 2.9795E+06, 2.9878E+06,
     +  2.9960E+06, 3.0043E+06, 3.0126E+06, 3.0210E+06, 3.0293E+06,
     + 3.0377E+06, 3.0460E+06, 3.0544E+06, 3.0628E+06, 3.0713E+06,
     +  3.0797E+06, 3.0882E+06, 3.0967E+06, 3.1052E+06, 3.1137E+06,
     + 3.1222E+06, 3.1308E+06, 3.1394E+06, 3.1480E+06, 3.1566E+06,
     +  3.1652E+06, 3.1739E+06, 3.1825E+06, 3.1912E+06, 3.1999E+06,
     + 3.2087E+06, 3.2174E+06, 3.2262E+06, 3.2349E+06, 3.2437E+06,
     +  3.2526E+06, 3.2614E+06, 3.2702E+06, 3.2791E+06, 3.2880E+06,
     + 3.2969E+06, 3.3059E+06, 3.3148E+06, 3.3238E+06, 3.3328E+06,
     +  3.3418E+06, 3.3508E+06, 3.3598E+06, 3.3689E+06, 3.3780E+06,
     + 3.3871E+06, 3.3962E+06, 3.4053E+06, 3.4145E+06, 3.4237E+06,
     +  3.4329E+06, 3.4421E+06, 3.4513E+06, 3.4606E+06, 3.4699E+06,
     + 3.4792E+06, 3.4885E+06, 3.4978E+06, 3.5072E+06, 3.5165E+06,
     +  3.5259E+06, 3.5353E+06, 3.5448E+06, 3.5542E+06, 3.5637E+06,
     + 3.5732E+06, 3.5827E+06, 3.5922E+06, 3.6018E+06, 3.6113E+06,
     +  3.6209E+06, 3.6305E+06, 3.6402E+06, 3.6498E+06, 3.6595E+06,
     + 3.6692E+06, 3.6789E+06, 3.6886E+06, 3.6984E+06, 3.7081E+06,
     +  3.7179E+06, 3.7277E+06, 3.7376E+06, 3.7474E+06, 3.7573E+06,
     + 3.7672E+06, 3.7771E+06, 3.7870E+06, 3.7970E+06, 3.8069E+06,
     +  3.8169E+06, 3.8269E+06, 3.8370E+06, 3.8470E+06, 3.8571E+06,
     + 3.8672E+06, 3.8773E+06, 3.8875E+06, 3.8976E+06, 3.9078E+06,
     +  3.9180E+06, 3.9282E+06, 3.9385E+06, 3.9487E+06, 3.9590E+06,
     + 3.9693E+06, 3.9796E+06, 3.9900E+06, 4.0003E+06, 4.0107E+06,
     +  4.0211E+06, 4.0316E+06, 4.0420E+06, 4.0525E+06, 4.0630E+06,
     + 4.0735E+06, 4.0840E+06, 4.0946E+06, 4.1052E+06, 4.1158E+06,
     +  4.1264E+06, 4.1371E+06, 4.1477E+06, 4.1584E+06, 4.1691E+06,
     + 4.1798E+06, 4.1906E+06, 4.2014E+06, 4.2122E+06, 4.2230E+06,
     +  4.2338E+06, 4.2447E+06, 4.2556E+06, 4.2665E+06, 4.2774E+06,
     + 4.2883E+06, 4.2993E+06, 4.3103E+06, 4.3213E+06, 4.3323E+06,
     +  4.3434E+06, 4.3545E+06, 4.3656E+06, 4.3767E+06, 4.3878E+06,
     + 4.3990E+06, 4.4102E+06, 4.4214E+06, 4.4326E+06, 4.4439E+06,
     +  4.4552E+06, 4.4665E+06, 4.4778E+06, 4.4891E+06, 4.5005E+06,
     + 4.5119E+06, 4.5233E+06, 4.5347E+06, 4.5462E+06, 4.5577E+06,
     +  4.5692E+06, 4.5807E+06, 4.5923E+06, 4.6038E+06, 4.6154E+06,
     + 4.6270E+06, 4.6387E+06, 4.6504E+06, 4.6620E+06, 4.6737E+06,
     +  4.6855E+06, 4.6972E+06, 4.7090E+06, 4.7208E+06, 4.7326E+06,
     + 4.7445E+06, 4.7564E+06, 4.7683E+06, 4.7802E+06, 4.7921E+06,
     +  4.8041E+06, 4.8161E+06, 4.8281E+06, 4.8401E+06, 4.8522E+06,
     + 4.8642E+06, 4.8764E+06, 4.8885E+06, 4.9006E+06, 4.9128E+06,
     +  4.9250E+06, 4.9372E+06, 4.9495E+06, 4.9617E+06, 4.9740E+06,
     + 4.9864E+06, 4.9987E+06, 5.0111E+06, 5.0235E+06, 5.0359E+06,
     +  5.0483E+06, 5.0608E+06, 5.0733E+06, 5.0858E+06, 5.0983E+06,
     + 5.1109E+06, 5.1234E+06, 5.1360E+06, 5.1487E+06, 5.1613E+06,
     +  5.1740E+06, 5.1867E+06, 5.1994E+06, 5.2122E+06, 5.2250E+06,
     + 5.2378E+06, 5.2506E+06, 5.2634E+06, 5.2763E+06, 5.2892E+06,
     +  5.3022E+06, 5.3151E+06, 5.3281E+06, 5.3411E+06, 5.3541E+06,
     + 5.3671E+06, 5.3802E+06, 5.3933E+06, 5.4064E+06, 5.4196E+06,
     +  5.4328E+06, 5.4460E+06, 5.4592E+06, 5.4724E+06, 5.4857E+06,
     + 5.4990E+06, 5.5123E+06, 5.5257E+06, 5.5390E+06, 5.5524E+06,
     +  5.5659E+06, 5.5793E+06, 5.5928E+06, 5.6063E+06, 5.6198E+06,
     + 5.6334E+06, 5.6470E+06, 5.6606E+06, 5.6742E+06, 5.6878E+06,
     +  5.7015E+06, 5.7152E+06, 5.7290E+06, 5.7427E+06, 5.7565E+06,
     + 5.7703E+06, 5.7842E+06, 5.7980E+06, 5.8119E+06, 5.8258E+06,
     +  5.8398E+06, 5.8537E+06, 5.8677E+06, 5.8817E+06, 5.8958E+06,
     + 5.9098E+06, 5.9239E+06, 5.9381E+06, 5.9522E+06, 5.9664E+06,
     +  5.9806E+06, 5.9948E+06, 6.0091E+06, 6.0234E+06, 6.0377E+06,
     + 6.0520E+06, 6.0664E+06, 6.0807E+06, 6.0952E+06, 6.1096E+06,
     +  6.1241E+06, 6.1386E+06, 6.1531E+06, 6.1676E+06, 6.1822E+06,
     + 6.1968E+06, 6.2114E+06, 6.2261E+06, 6.2408E+06, 6.2555E+06,
     +  6.2702E+06, 6.2850E+06, 6.2998E+06, 6.3146E+06, 6.3295E+06,
     + 6.3443E+06, 6.3592E+06, 6.3742E+06, 6.3891E+06, 6.4041E+06,
     +  6.4191E+06, 6.4342E+06, 6.4492E+06, 6.4643E+06, 6.4794E+06,
     + 6.4946E+06, 6.5098E+06, 6.5250E+06, 6.5402E+06, 6.5555E+06,
     +  6.5707E+06, 6.5861E+06, 6.6014E+06, 6.6168E+06, 6.6322E+06,
     + 6.6476E+06, 6.6631E+06, 6.6786E+06, 6.6941E+06, 6.7096E+06,
     +  6.7252E+06, 6.7408E+06, 6.7564E+06, 6.7720E+06, 6.7877E+06,
     + 6.8034E+06, 6.8192E+06, 6.8349E+06, 6.8507E+06, 6.8666E+06,
     +  6.8824E+06, 6.8983E+06, 6.9142E+06, 6.9302E+06, 6.9461E+06,
     + 6.9621E+06, 6.9781E+06, 6.9942E+06, 7.0103E+06, 7.0264E+06,
     +  7.0425E+06, 7.0587E+06, 7.0749E+06, 7.0911E+06, 7.1074E+06,
     + 7.1237E+06, 7.1400E+06, 7.1563E+06, 7.1727E+06, 7.1891E+06,
     +  7.2055E+06, 7.2220E+06, 7.2385E+06, 7.2550E+06, 7.2716E+06,
     + 7.2882E+06, 7.3048E+06, 7.3214E+06, 7.3381E+06, 7.3548E+06,
     +  7.3715E+06, 7.3883E+06, 7.4050E+06, 7.4219E+06, 7.4387E+06,
     + 7.4556E+06, 7.4725E+06, 7.4894E+06, 7.5064E+06, 7.5234E+06,
     +  7.5404E+06, 7.5575E+06, 7.5746E+06, 7.5917E+06, 7.6088E+06,
     + 7.6260E+06, 7.6432E+06, 7.6605E+06, 7.6777E+06, 7.6950E+06,
     +  7.7124E+06, 7.7297E+06, 7.7471E+06, 7.7646E+06, 7.7820E+06,
     + 7.7995E+06, 7.8170E+06, 7.8346E+06, 7.8521E+06, 7.8697E+06,
     +  7.8874E+06, 7.9050E+06, 7.9227E+06, 7.9405E+06, 7.9582E+06,
     + 7.9760E+06, 7.9939E+06, 8.0117E+06, 8.0296E+06, 8.0475E+06,
     +  8.0655E+06, 8.0835E+06, 8.1015E+06, 8.1195E+06, 8.1376E+06,
     + 8.1557E+06, 8.1738E+06, 8.1920E+06, 8.2102E+06, 8.2284E+06,
     +  8.2467E+06, 8.2650E+06, 8.2833E+06, 8.3017E+06, 8.3201E+06,
     + 8.3385E+06, 8.3570E+06, 8.3754E+06, 8.3940E+06, 8.4125E+06,
     +  8.4311E+06, 8.4497E+06, 8.4684E+06, 8.4870E+06, 8.5058E+06,
     + 8.5245E+06, 8.5433E+06, 8.5621E+06, 8.5809E+06, 8.5998E+06,
     +  8.6187E+06, 8.6377E+06, 8.6566E+06, 8.6756E+06, 8.6947E+06,
     + 8.7138E+06, 8.7329E+06, 8.7520E+06, 8.7712E+06, 8.7904E+06,
     +  8.8096E+06, 8.8289E+06, 8.8482E+06, 8.8675E+06, 8.8869E+06,
     + 8.9063E+06, 8.9257E+06, 8.9452E+06, 8.9647E+06, 8.9842E+06,
     +  9.0038E+06, 9.0234E+06, 9.0430E+06, 9.0626E+06, 9.0823E+06,
     + 9.1021E+06, 9.1218E+06, 9.1416E+06, 9.1615E+06, 9.1814E+06,
     +  9.2013E+06, 9.2212E+06, 9.2412E+06, 9.2612E+06, 9.2812E+06,
     + 9.3013E+06, 9.3214E+06, 9.3415E+06, 9.3617E+06, 9.3819E+06,
     +  9.4021E+06, 9.4224E+06, 9.4427E+06, 9.4631E+06, 9.4834E+06,
     + 9.5038E+06, 9.5243E+06, 9.5448E+06, 9.5653E+06, 9.5858E+06,
     +  9.6064E+06, 9.6270E+06, 9.6477E+06, 9.6684E+06, 9.6891E+06,
     + 9.7099E+06, 9.7307E+06, 9.7515E+06, 9.7723E+06, 9.7932E+06,
     +  9.8142E+06, 9.8351E+06, 9.8561E+06, 9.8772E+06, 9.8983E+06,
     + 9.9194E+06, 9.9405E+06, 9.9617E+06, 9.9829E+06, 1.0004E+07,
     +  1.0025E+07, 1.0047E+07, 1.0068E+07, 1.0089E+07, 1.0111E+07,
     + 1.0132E+07, 1.0154E+07, 1.0175E+07, 1.0197E+07, 1.0219E+07,
     +  1.0240E+07, 1.0262E+07, 1.0284E+07, 1.0305E+07, 1.0327E+07,
     + 1.0349E+07, 1.0371E+07, 1.0393E+07, 1.0415E+07, 1.0437E+07,
     +  1.0459E+07, 1.0481E+07, 1.0503E+07, 1.0525E+07, 1.0547E+07,
     + 1.0569E+07, 1.0591E+07, 1.0614E+07, 1.0636E+07, 1.0658E+07,
     +  1.0681E+07, 1.0703E+07, 1.0725E+07, 1.0748E+07, 1.0770E+07,
     + 1.0793E+07, 1.0816E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_N2    (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 3),QofT( 3,NT),Q(NT)
      data (Tmx(I),I=1,3)/9000.,9000.,9000./
      data xgj/ 1.,6.,1./
c...       N2
c...        --        44
      data (QofT( 1,J),J=1,2408)/ 6.0294E+00, 6.5197E+00, 7.4330E+00,
     +  8.5646E+00, 9.8531E+00, 1.1252E+01, 1.2722E+01, 1.4235E+01,
     + 1.5772E+01, 1.7322E+01, 1.8880E+01, 2.0442E+01, 2.2007E+01,
     +  2.3573E+01, 2.5141E+01, 2.6709E+01, 2.8277E+01, 2.9846E+01,
     + 3.1416E+01, 3.2986E+01, 3.6126E+01, 3.9267E+01, 4.2408E+01,
     +  4.5550E+01, 4.8692E+01, 5.1835E+01, 5.4978E+01, 5.8121E+01,
     + 6.1264E+01, 6.4407E+01, 6.7551E+01, 7.0694E+01, 7.3838E+01,
     +  7.6982E+01, 8.0126E+01, 8.3270E+01, 8.6414E+01, 8.9558E+01,
     + 9.2702E+01, 9.5846E+01, 9.8991E+01, 1.0214E+02, 1.0528E+02,
     +  1.0842E+02, 1.1157E+02, 1.1471E+02, 1.1786E+02, 1.2100E+02,
     + 1.2415E+02, 1.2729E+02, 1.3044E+02, 1.3358E+02, 1.3673E+02,
     +  1.3987E+02, 1.4302E+02, 1.4616E+02, 1.4931E+02, 1.5245E+02,
     + 1.5560E+02, 1.5874E+02, 1.6189E+02, 1.6503E+02, 1.6818E+02,
     +  1.7132E+02, 1.7447E+02, 1.7761E+02, 1.8076E+02, 1.8390E+02,
     + 1.8705E+02, 1.9019E+02, 1.9334E+02, 1.9648E+02, 1.9963E+02,
     +  2.0278E+02, 2.0592E+02, 2.0907E+02, 2.1221E+02, 2.1536E+02,
     + 2.1850E+02, 2.2165E+02, 2.2480E+02, 2.2794E+02, 2.3109E+02,
     +  2.3423E+02, 2.3738E+02, 2.4052E+02, 2.4367E+02, 2.4682E+02,
     + 2.4996E+02, 2.5311E+02, 2.5625E+02, 2.5940E+02, 2.6255E+02,
     +  2.6569E+02, 2.6884E+02, 2.7198E+02, 2.7513E+02, 2.7828E+02,
     + 2.8142E+02, 2.8457E+02, 2.8772E+02, 2.9086E+02, 2.9401E+02,
     +  2.9715E+02, 3.0030E+02, 3.0345E+02, 3.0659E+02, 3.0974E+02,
     + 3.1289E+02, 3.1603E+02, 3.1918E+02, 3.2233E+02, 3.2547E+02,
     +  3.2862E+02, 3.3177E+02, 3.3491E+02, 3.3806E+02, 3.4121E+02,
     + 3.4435E+02, 3.4750E+02, 3.5065E+02, 3.5379E+02, 3.5694E+02,
     +  3.6009E+02, 3.6323E+02, 3.6638E+02, 3.6953E+02, 3.7268E+02,
     + 3.7582E+02, 3.7897E+02, 3.8212E+02, 3.8526E+02, 3.8841E+02,
     +  3.9156E+02, 3.9471E+02, 3.9785E+02, 4.0100E+02, 4.0415E+02,
     + 4.0729E+02, 4.1044E+02, 4.1359E+02, 4.1674E+02, 4.1988E+02,
     +  4.2303E+02, 4.2618E+02, 4.2933E+02, 4.3247E+02, 4.3562E+02,
     + 4.3877E+02, 4.4192E+02, 4.4507E+02, 4.4821E+02, 4.5136E+02,
     +  4.5451E+02, 4.5766E+02, 4.6081E+02, 4.6395E+02, 4.6710E+02,
     + 4.7025E+02, 4.7340E+02, 4.7655E+02, 4.7970E+02, 4.8284E+02,
     +  4.8599E+02, 4.8914E+02, 4.9229E+02, 4.9544E+02, 4.9859E+02,
     + 5.0174E+02, 5.0489E+02, 5.0804E+02, 5.1118E+02, 5.1433E+02,
     +  5.1748E+02, 5.2063E+02, 5.2378E+02, 5.2693E+02, 5.3008E+02,
     + 5.3323E+02, 5.3638E+02, 5.3953E+02, 5.4268E+02, 5.4583E+02,
     +  5.4898E+02, 5.5213E+02, 5.5529E+02, 5.5844E+02, 5.6159E+02,
     + 5.6474E+02, 5.6789E+02, 5.7104E+02, 5.7419E+02, 5.7735E+02,
     +  5.8050E+02, 5.8365E+02, 5.8680E+02, 5.8996E+02, 5.9311E+02,
     + 5.9626E+02, 5.9941E+02, 6.0257E+02, 6.0572E+02, 6.0888E+02,
     +  6.1203E+02, 6.1518E+02, 6.1834E+02, 6.2149E+02, 6.2465E+02,
     + 6.2781E+02, 6.3096E+02, 6.3412E+02, 6.3727E+02, 6.4043E+02,
     +  6.4359E+02, 6.4674E+02, 6.4990E+02, 6.5306E+02, 6.5622E+02,
     + 6.5938E+02, 6.6254E+02, 6.6570E+02, 6.6886E+02, 6.7202E+02,
     +  6.7518E+02, 6.7834E+02, 6.8150E+02, 6.8466E+02, 6.8782E+02,
     + 6.9098E+02, 6.9415E+02, 6.9731E+02, 7.0047E+02, 7.0364E+02,
     +  7.0680E+02, 7.0997E+02, 7.1313E+02, 7.1630E+02, 7.1947E+02,
     + 7.2263E+02, 7.2580E+02, 7.2897E+02, 7.3214E+02, 7.3531E+02,
     +  7.3848E+02, 7.4165E+02, 7.4482E+02, 7.4799E+02, 7.5116E+02,
     + 7.5434E+02, 7.5751E+02, 7.6068E+02, 7.6386E+02, 7.6703E+02,
     +  7.7021E+02, 7.7339E+02, 7.7656E+02, 7.7974E+02, 7.8292E+02,
     + 7.8610E+02, 7.8928E+02, 7.9246E+02, 7.9564E+02, 7.9882E+02,
     +  8.0201E+02, 8.0519E+02, 8.0838E+02, 8.1156E+02, 8.1475E+02,
     + 8.1793E+02, 8.2112E+02, 8.2431E+02, 8.2750E+02, 8.3069E+02,
     +  8.3388E+02, 8.3707E+02, 8.4027E+02, 8.4346E+02, 8.4666E+02,
     + 8.4985E+02, 8.5305E+02, 8.5625E+02, 8.5945E+02, 8.6265E+02,
     +  8.6585E+02, 8.6905E+02, 8.7225E+02, 8.7545E+02, 8.7866E+02,
     + 8.8186E+02, 8.8507E+02, 8.8828E+02, 8.9149E+02, 8.9470E+02,
     +  8.9791E+02, 9.0112E+02, 9.0433E+02, 9.0755E+02, 9.1076E+02,
     + 9.1398E+02, 9.1720E+02, 9.2042E+02, 9.2364E+02, 9.2686E+02,
     +  9.3008E+02, 9.3331E+02, 9.3653E+02, 9.3976E+02, 9.4298E+02,
     + 9.4621E+02, 9.4944E+02, 9.5268E+02, 9.5591E+02, 9.5914E+02,
     +  9.6238E+02, 9.6561E+02, 9.6885E+02, 9.7209E+02, 9.7533E+02,
     + 9.7857E+02, 9.8182E+02, 9.8506E+02, 9.8831E+02, 9.9156E+02,
     +  9.9481E+02, 9.9806E+02, 1.0013E+03, 1.0046E+03, 1.0078E+03,
     + 1.0111E+03, 1.0143E+03, 1.0176E+03, 1.0209E+03, 1.0241E+03,
     +  1.0274E+03, 1.0306E+03, 1.0339E+03, 1.0372E+03, 1.0404E+03,
     + 1.0437E+03, 1.0470E+03, 1.0503E+03, 1.0535E+03, 1.0568E+03,
     +  1.0601E+03, 1.0634E+03, 1.0667E+03, 1.0700E+03, 1.0732E+03,
     + 1.0765E+03, 1.0798E+03, 1.0831E+03, 1.0864E+03, 1.0897E+03,
     +  1.0930E+03, 1.0963E+03, 1.0996E+03, 1.1029E+03, 1.1062E+03,
     + 1.1095E+03, 1.1128E+03, 1.1161E+03, 1.1194E+03, 1.1228E+03,
     +  1.1261E+03, 1.1294E+03, 1.1327E+03, 1.1360E+03, 1.1393E+03,
     + 1.1427E+03, 1.1460E+03, 1.1493E+03, 1.1527E+03, 1.1560E+03,
     +  1.1593E+03, 1.1627E+03, 1.1660E+03, 1.1693E+03, 1.1727E+03,
     + 1.1760E+03, 1.1794E+03, 1.1827E+03, 1.1861E+03, 1.1894E+03,
     +  1.1928E+03, 1.1961E+03, 1.1995E+03, 1.2028E+03, 1.2062E+03,
     + 1.2096E+03, 1.2129E+03, 1.2163E+03, 1.2197E+03, 1.2231E+03,
     +  1.2264E+03, 1.2298E+03, 1.2332E+03, 1.2366E+03, 1.2400E+03,
     + 1.2433E+03, 1.2467E+03, 1.2501E+03, 1.2535E+03, 1.2569E+03,
     +  1.2603E+03, 1.2637E+03, 1.2671E+03, 1.2705E+03, 1.2739E+03,
     + 1.2773E+03, 1.2807E+03, 1.2841E+03, 1.2876E+03, 1.2910E+03,
     +  1.2944E+03, 1.2978E+03, 1.3012E+03, 1.3047E+03, 1.3081E+03,
     + 1.3115E+03, 1.3150E+03, 1.3184E+03, 1.3218E+03, 1.3253E+03,
     +  1.3287E+03, 1.3322E+03, 1.3356E+03, 1.3391E+03, 1.3425E+03,
     + 1.3460E+03, 1.3494E+03, 1.3529E+03, 1.3564E+03, 1.3598E+03,
     +  1.3633E+03, 1.3668E+03, 1.3702E+03, 1.3737E+03, 1.3772E+03,
     + 1.3807E+03, 1.3841E+03, 1.3876E+03, 1.3911E+03, 1.3946E+03,
     +  1.3981E+03, 1.4016E+03, 1.4051E+03, 1.4086E+03, 1.4121E+03,
     + 1.4156E+03, 1.4191E+03, 1.4226E+03, 1.4261E+03, 1.4296E+03,
     +  1.4332E+03, 1.4367E+03, 1.4402E+03, 1.4437E+03, 1.4473E+03,
     + 1.4508E+03, 1.4543E+03, 1.4579E+03, 1.4614E+03, 1.4649E+03,
     +  1.4685E+03, 1.4720E+03, 1.4756E+03, 1.4791E+03, 1.4827E+03,
     + 1.4863E+03, 1.4898E+03, 1.4934E+03, 1.4969E+03, 1.5005E+03,
     +  1.5041E+03, 1.5077E+03, 1.5112E+03, 1.5148E+03, 1.5184E+03,
     + 1.5220E+03, 1.5256E+03, 1.5292E+03, 1.5328E+03, 1.5363E+03,
     +  1.5399E+03, 1.5436E+03, 1.5472E+03, 1.5508E+03, 1.5544E+03,
     + 1.5580E+03, 1.5616E+03, 1.5652E+03, 1.5688E+03, 1.5725E+03,
     +  1.5761E+03, 1.5797E+03, 1.5834E+03, 1.5870E+03, 1.5906E+03,
     + 1.5943E+03, 1.5979E+03, 1.6016E+03, 1.6052E+03, 1.6089E+03,
     +  1.6125E+03, 1.6162E+03, 1.6198E+03, 1.6235E+03, 1.6272E+03,
     + 1.6308E+03, 1.6345E+03, 1.6382E+03, 1.6419E+03, 1.6456E+03,
     +  1.6492E+03, 1.6529E+03, 1.6566E+03, 1.6603E+03, 1.6640E+03,
     + 1.6677E+03, 1.6714E+03, 1.6751E+03, 1.6788E+03, 1.6825E+03,
     +  1.6863E+03, 1.6900E+03, 1.6937E+03, 1.6974E+03, 1.7011E+03,
     + 1.7049E+03, 1.7086E+03, 1.7123E+03, 1.7161E+03, 1.7198E+03,
     +  1.7236E+03, 1.7273E+03, 1.7311E+03, 1.7348E+03, 1.7386E+03,
     + 1.7423E+03, 1.7461E+03, 1.7499E+03, 1.7536E+03, 1.7574E+03,
     +  1.7612E+03, 1.7650E+03, 1.7688E+03, 1.7725E+03, 1.7763E+03,
     + 1.7801E+03, 1.7839E+03, 1.7877E+03, 1.7915E+03, 1.7953E+03,
     +  1.7991E+03, 1.8029E+03, 1.8067E+03, 1.8106E+03, 1.8144E+03,
     + 1.8182E+03, 1.8220E+03, 1.8259E+03, 1.8297E+03, 1.8335E+03,
     +  1.8374E+03, 1.8412E+03, 1.8450E+03, 1.8489E+03, 1.8527E+03,
     + 1.8566E+03, 1.8604E+03, 1.8643E+03, 1.8682E+03, 1.8720E+03,
     +  1.8759E+03, 1.8798E+03, 1.8837E+03, 1.8875E+03, 1.8914E+03,
     + 1.8953E+03, 1.8992E+03, 1.9031E+03, 1.9070E+03, 1.9109E+03,
     +  1.9148E+03, 1.9187E+03, 1.9226E+03, 1.9265E+03, 1.9304E+03,
     + 1.9343E+03, 1.9383E+03, 1.9422E+03, 1.9461E+03, 1.9500E+03,
     +  1.9540E+03, 1.9579E+03, 1.9619E+03, 1.9658E+03, 1.9697E+03,
     + 1.9737E+03, 1.9777E+03, 1.9816E+03, 1.9856E+03, 1.9895E+03,
     +  1.9935E+03, 1.9975E+03, 2.0014E+03, 2.0054E+03, 2.0094E+03,
     + 2.0134E+03, 2.0174E+03, 2.0214E+03, 2.0254E+03, 2.0294E+03,
     +  2.0334E+03, 2.0374E+03, 2.0414E+03, 2.0454E+03, 2.0494E+03,
     + 2.0534E+03, 2.0574E+03, 2.0614E+03, 2.0655E+03, 2.0695E+03,
     +  2.0735E+03, 2.0776E+03, 2.0816E+03, 2.0857E+03, 2.0897E+03,
     + 2.0938E+03, 2.0978E+03, 2.1019E+03, 2.1059E+03, 2.1100E+03,
     +  2.1141E+03, 2.1181E+03, 2.1222E+03, 2.1263E+03, 2.1304E+03,
     + 2.1344E+03, 2.1385E+03, 2.1426E+03, 2.1467E+03, 2.1508E+03,
     +  2.1549E+03, 2.1590E+03, 2.1631E+03, 2.1672E+03, 2.1714E+03,
     + 2.1755E+03, 2.1796E+03, 2.1837E+03, 2.1878E+03, 2.1920E+03,
     +  2.1961E+03, 2.2002E+03, 2.2044E+03, 2.2085E+03, 2.2127E+03,
     + 2.2168E+03, 2.2210E+03, 2.2252E+03, 2.2293E+03, 2.2335E+03,
     +  2.2376E+03, 2.2418E+03, 2.2460E+03, 2.2502E+03, 2.2544E+03,
     + 2.2585E+03, 2.2627E+03, 2.2669E+03, 2.2711E+03, 2.2753E+03,
     +  2.2795E+03, 2.2837E+03, 2.2879E+03, 2.2922E+03, 2.2964E+03,
     + 2.3006E+03, 2.3048E+03, 2.3091E+03, 2.3133E+03, 2.3175E+03,
     +  2.3218E+03, 2.3260E+03, 2.3302E+03, 2.3345E+03, 2.3387E+03,
     + 2.3430E+03, 2.3473E+03, 2.3515E+03, 2.3558E+03, 2.3601E+03,
     +  2.3643E+03, 2.3686E+03, 2.3729E+03, 2.3772E+03, 2.3815E+03,
     + 2.3858E+03, 2.3901E+03, 2.3943E+03, 2.3987E+03, 2.4030E+03,
     +  2.4073E+03, 2.4116E+03, 2.4159E+03, 2.4202E+03, 2.4245E+03,
     + 2.4289E+03, 2.4332E+03, 2.4375E+03, 2.4419E+03, 2.4462E+03,
     +  2.4505E+03, 2.4549E+03, 2.4592E+03, 2.4636E+03, 2.4680E+03,
     + 2.4723E+03, 2.4767E+03, 2.4810E+03, 2.4854E+03, 2.4898E+03,
     +  2.4942E+03, 2.4986E+03, 2.5029E+03, 2.5073E+03, 2.5117E+03,
     + 2.5161E+03, 2.5205E+03, 2.5249E+03, 2.5293E+03, 2.5338E+03,
     +  2.5382E+03, 2.5426E+03, 2.5470E+03, 2.5514E+03, 2.5559E+03,
     + 2.5603E+03, 2.5647E+03, 2.5692E+03, 2.5736E+03, 2.5781E+03,
     +  2.5825E+03, 2.5870E+03, 2.5914E+03, 2.5959E+03, 2.6003E+03,
     + 2.6048E+03, 2.6093E+03, 2.6138E+03, 2.6182E+03, 2.6227E+03,
     +  2.6272E+03, 2.6317E+03, 2.6362E+03, 2.6407E+03, 2.6452E+03,
     + 2.6497E+03, 2.6542E+03, 2.6587E+03, 2.6632E+03, 2.6678E+03,
     +  2.6723E+03, 2.6768E+03, 2.6813E+03, 2.6859E+03, 2.6904E+03,
     + 2.6949E+03, 2.6995E+03, 2.7040E+03, 2.7086E+03, 2.7131E+03,
     +  2.7177E+03, 2.7223E+03, 2.7268E+03, 2.7314E+03, 2.7360E+03,
     + 2.7405E+03, 2.7451E+03, 2.7497E+03, 2.7543E+03, 2.7589E+03,
     +  2.7635E+03, 2.7681E+03, 2.7727E+03, 2.7773E+03, 2.7819E+03,
     + 2.7865E+03, 2.7911E+03, 2.7958E+03, 2.8004E+03, 2.8050E+03,
     +  2.8096E+03, 2.8143E+03, 2.8189E+03, 2.8235E+03, 2.8282E+03,
     + 2.8328E+03, 2.8375E+03, 2.8422E+03, 2.8468E+03, 2.8515E+03,
     +  2.8561E+03, 2.8608E+03, 2.8655E+03, 2.8702E+03, 2.8749E+03,
     + 2.8795E+03, 2.8842E+03, 2.8889E+03, 2.8936E+03, 2.8983E+03,
     +  2.9030E+03, 2.9077E+03, 2.9124E+03, 2.9172E+03, 2.9219E+03,
     + 2.9266E+03, 2.9313E+03, 2.9361E+03, 2.9408E+03, 2.9455E+03,
     +  2.9503E+03, 2.9550E+03, 2.9598E+03, 2.9645E+03, 2.9693E+03,
     + 2.9740E+03, 2.9788E+03, 2.9836E+03, 2.9883E+03, 2.9931E+03,
     +  2.9979E+03, 3.0027E+03, 3.0074E+03, 3.0122E+03, 3.0170E+03,
     + 3.0218E+03, 3.0266E+03, 3.0314E+03, 3.0362E+03, 3.0410E+03,
     +  3.0459E+03, 3.0507E+03, 3.0555E+03, 3.0603E+03, 3.0652E+03,
     + 3.0700E+03, 3.0748E+03, 3.0797E+03, 3.0845E+03, 3.0894E+03,
     +  3.0942E+03, 3.0991E+03, 3.1039E+03, 3.1088E+03, 3.1136E+03,
     + 3.1185E+03, 3.1234E+03, 3.1283E+03, 3.1331E+03, 3.1380E+03,
     +  3.1429E+03, 3.1478E+03, 3.1527E+03, 3.1576E+03, 3.1625E+03,
     + 3.1674E+03, 3.1723E+03, 3.1772E+03, 3.1822E+03, 3.1871E+03,
     +  3.1920E+03, 3.1969E+03, 3.2019E+03, 3.2068E+03, 3.2117E+03,
     + 3.2167E+03, 3.2216E+03, 3.2266E+03, 3.2315E+03, 3.2365E+03,
     +  3.2415E+03, 3.2464E+03, 3.2514E+03, 3.2564E+03, 3.2613E+03,
     + 3.2663E+03, 3.2713E+03, 3.2763E+03, 3.2813E+03, 3.2863E+03,
     +  3.2913E+03, 3.2963E+03, 3.3013E+03, 3.3063E+03, 3.3113E+03,
     + 3.3163E+03, 3.3214E+03, 3.3264E+03, 3.3314E+03, 3.3365E+03,
     +  3.3415E+03, 3.3465E+03, 3.3516E+03, 3.3566E+03, 3.3617E+03,
     + 3.3667E+03, 3.3718E+03, 3.3769E+03, 3.3819E+03, 3.3870E+03,
     +  3.3921E+03, 3.3971E+03, 3.4022E+03, 3.4073E+03, 3.4124E+03,
     + 3.4175E+03, 3.4226E+03, 3.4277E+03, 3.4328E+03, 3.4379E+03,
     +  3.4430E+03, 3.4481E+03, 3.4532E+03, 3.4584E+03, 3.4635E+03,
     + 3.4686E+03, 3.4737E+03, 3.4789E+03, 3.4840E+03, 3.4892E+03,
     +  3.4943E+03, 3.4995E+03, 3.5046E+03, 3.5098E+03, 3.5150E+03,
     + 3.5201E+03, 3.5253E+03, 3.5305E+03, 3.5356E+03, 3.5408E+03,
     +  3.5460E+03, 3.5512E+03, 3.5564E+03, 3.5616E+03, 3.5668E+03,
     + 3.5720E+03, 3.5772E+03, 3.5824E+03, 3.5876E+03, 3.5929E+03,
     +  3.5981E+03, 3.6033E+03, 3.6085E+03, 3.6138E+03, 3.6190E+03,
     + 3.6242E+03, 3.6295E+03, 3.6347E+03, 3.6400E+03, 3.6452E+03,
     +  3.6505E+03, 3.6558E+03, 3.6610E+03, 3.6663E+03, 3.6716E+03,
     + 3.6769E+03, 3.6822E+03, 3.6874E+03, 3.6927E+03, 3.6980E+03,
     +  3.7033E+03, 3.7086E+03, 3.7139E+03, 3.7192E+03, 3.7245E+03,
     + 3.7299E+03, 3.7352E+03, 3.7405E+03, 3.7458E+03, 3.7512E+03,
     +  3.7565E+03, 3.7618E+03, 3.7672E+03, 3.7725E+03, 3.7779E+03,
     + 3.7832E+03, 3.7886E+03, 3.7940E+03, 3.7993E+03, 3.8047E+03,
     +  3.8101E+03, 3.8154E+03, 3.8208E+03, 3.8262E+03, 3.8316E+03,
     + 3.8370E+03, 3.8424E+03, 3.8478E+03, 3.8532E+03, 3.8586E+03,
     +  3.8640E+03, 3.8694E+03, 3.8748E+03, 3.8803E+03, 3.8857E+03,
     + 3.8911E+03, 3.8965E+03, 3.9020E+03, 3.9074E+03, 3.9129E+03,
     +  3.9183E+03, 3.9238E+03, 3.9292E+03, 3.9347E+03, 3.9401E+03,
     + 3.9456E+03, 3.9511E+03, 3.9565E+03, 3.9620E+03, 3.9675E+03,
     +  3.9730E+03, 3.9785E+03, 3.9840E+03, 3.9895E+03, 3.9950E+03,
     + 4.0005E+03, 4.0060E+03, 4.0115E+03, 4.0170E+03, 4.0225E+03,
     +  4.0280E+03, 4.0336E+03, 4.0391E+03, 4.0446E+03, 4.0502E+03,
     + 4.0557E+03, 4.0613E+03, 4.0668E+03, 4.0724E+03, 4.0779E+03,
     +  4.0835E+03, 4.0890E+03, 4.0946E+03, 4.1002E+03, 4.1058E+03,
     + 4.1113E+03, 4.1169E+03, 4.1225E+03, 4.1281E+03, 4.1337E+03,
     +  4.1393E+03, 4.1449E+03, 4.1505E+03, 4.1561E+03, 4.1617E+03,
     + 4.1673E+03, 4.1729E+03, 4.1786E+03, 4.1842E+03, 4.1898E+03,
     +  4.1955E+03, 4.2011E+03, 4.2067E+03, 4.2124E+03, 4.2180E+03,
     + 4.2237E+03, 4.2294E+03, 4.2350E+03, 4.2407E+03, 4.2463E+03,
     +  4.2520E+03, 4.2577E+03, 4.2634E+03, 4.2691E+03, 4.2747E+03,
     + 4.2804E+03, 4.2861E+03, 4.2918E+03, 4.2975E+03, 4.3032E+03,
     +  4.3089E+03, 4.3147E+03, 4.3204E+03, 4.3261E+03, 4.3318E+03,
     + 4.3376E+03, 4.3433E+03, 4.3490E+03, 4.3548E+03, 4.3605E+03,
     +  4.3662E+03, 4.3720E+03, 4.3778E+03, 4.3835E+03, 4.3893E+03,
     + 4.3950E+03, 4.4008E+03, 4.4066E+03, 4.4124E+03, 4.4181E+03,
     +  4.4239E+03, 4.4297E+03, 4.4355E+03, 4.4413E+03, 4.4471E+03,
     + 4.4529E+03, 4.4587E+03, 4.4645E+03, 4.4703E+03, 4.4762E+03,
     +  4.4820E+03, 4.4878E+03, 4.4936E+03, 4.4995E+03, 4.5053E+03,
     + 4.5111E+03, 4.5170E+03, 4.5228E+03, 4.5287E+03, 4.5345E+03,
     +  4.5404E+03, 4.5463E+03, 4.5521E+03, 4.5580E+03, 4.5639E+03,
     + 4.5698E+03, 4.5756E+03, 4.5815E+03, 4.5874E+03, 4.5933E+03,
     +  4.5992E+03, 4.6051E+03, 4.6110E+03, 4.6169E+03, 4.6228E+03,
     + 4.6287E+03, 4.6347E+03, 4.6406E+03, 4.6465E+03, 4.6524E+03,
     +  4.6584E+03, 4.6643E+03, 4.6702E+03, 4.6762E+03, 4.6821E+03,
     + 4.6881E+03, 4.6940E+03, 4.7000E+03, 4.7060E+03, 4.7119E+03,
     +  4.7179E+03, 4.7239E+03, 4.7299E+03, 4.7358E+03, 4.7418E+03,
     + 4.7478E+03, 4.7538E+03, 4.7598E+03, 4.7658E+03, 4.7718E+03,
     +  4.7778E+03, 4.7838E+03, 4.7898E+03, 4.7959E+03, 4.8019E+03,
     + 4.8079E+03, 4.8139E+03, 4.8200E+03, 4.8260E+03, 4.8321E+03,
     +  4.8381E+03, 4.8442E+03, 4.8502E+03, 4.8563E+03, 4.8623E+03,
     + 4.8684E+03, 4.8745E+03, 4.8805E+03, 4.8866E+03, 4.8927E+03,
     +  4.8988E+03, 4.9048E+03, 4.9109E+03, 4.9170E+03, 4.9231E+03,
     + 4.9292E+03, 4.9353E+03, 4.9414E+03, 4.9475E+03, 4.9537E+03,
     +  4.9598E+03, 4.9659E+03, 4.9720E+03, 4.9782E+03, 4.9843E+03,
     + 4.9904E+03, 4.9966E+03, 5.0027E+03, 5.0089E+03, 5.0150E+03,
     +  5.0212E+03, 5.0273E+03, 5.0335E+03, 5.0397E+03, 5.0458E+03,
     + 5.0520E+03, 5.0582E+03, 5.0644E+03, 5.0706E+03, 5.0768E+03,
     +  5.0829E+03, 5.0891E+03, 5.0953E+03, 5.1015E+03, 5.1078E+03,
     + 5.1140E+03, 5.1202E+03, 5.1264E+03, 5.1326E+03, 5.1389E+03,
     +  5.1451E+03, 5.1513E+03, 5.1576E+03, 5.1638E+03, 5.1700E+03,
     + 5.1763E+03, 5.1825E+03, 5.1888E+03, 5.1951E+03, 5.2013E+03,
     +  5.2076E+03, 5.2139E+03, 5.2201E+03, 5.2264E+03, 5.2327E+03,
     + 5.2390E+03, 5.2453E+03, 5.2516E+03, 5.2579E+03, 5.2642E+03,
     +  5.2705E+03, 5.2768E+03, 5.2831E+03, 5.2894E+03, 5.2957E+03,
     + 5.3020E+03, 5.3084E+03, 5.3147E+03, 5.3210E+03, 5.3274E+03,
     +  5.3337E+03, 5.3400E+03, 5.3464E+03, 5.3527E+03, 5.3591E+03,
     + 5.3655E+03, 5.3718E+03, 5.3782E+03, 5.3846E+03, 5.3909E+03,
     +  5.3973E+03, 5.4037E+03, 5.4101E+03, 5.4165E+03, 5.4229E+03,
     + 5.4292E+03, 5.4356E+03, 5.4421E+03, 5.4485E+03, 5.4549E+03,
     +  5.4613E+03, 5.4677E+03, 5.4741E+03, 5.4805E+03, 5.4870E+03,
     + 5.4934E+03, 5.4998E+03, 5.5063E+03, 5.5127E+03, 5.5192E+03,
     +  5.5256E+03, 5.5321E+03, 5.5385E+03, 5.5450E+03, 5.5515E+03,
     + 5.5579E+03, 5.5644E+03, 5.5709E+03, 5.5774E+03, 5.5838E+03,
     +  5.5903E+03, 5.5968E+03, 5.6033E+03, 5.6098E+03, 5.6163E+03,
     + 5.6228E+03, 5.6293E+03, 5.6358E+03, 5.6424E+03, 5.6489E+03,
     +  5.6554E+03, 5.6619E+03, 5.6685E+03, 5.6750E+03, 5.6815E+03,
     + 5.6881E+03, 5.6946E+03, 5.7012E+03, 5.7077E+03, 5.7143E+03,
     +  5.7208E+03, 5.7274E+03, 5.7340E+03, 5.7405E+03, 5.7471E+03,
     + 5.7537E+03, 5.7603E+03, 5.7668E+03, 5.7734E+03, 5.7800E+03,
     +  5.7866E+03, 5.7932E+03, 5.7998E+03, 5.8064E+03, 5.8130E+03,
     + 5.8197E+03, 5.8263E+03, 5.8329E+03, 5.8395E+03, 5.8461E+03,
     +  5.8528E+03, 5.8594E+03, 5.8661E+03, 5.8727E+03, 5.8793E+03,
     + 5.8860E+03, 5.8926E+03, 5.8993E+03, 5.9060E+03, 5.9126E+03,
     +  5.9193E+03, 5.9260E+03, 5.9326E+03, 5.9393E+03, 5.9460E+03,
     + 5.9527E+03, 5.9594E+03, 5.9661E+03, 5.9728E+03, 5.9795E+03,
     +  5.9862E+03, 5.9929E+03, 5.9996E+03, 6.0063E+03, 6.0130E+03,
     + 6.0198E+03, 6.0265E+03, 6.0332E+03, 6.0399E+03, 6.0467E+03,
     +  6.0534E+03, 6.0602E+03, 6.0669E+03, 6.0737E+03, 6.0804E+03,
     + 6.0872E+03, 6.0939E+03, 6.1007E+03, 6.1075E+03, 6.1142E+03,
     +  6.1210E+03, 6.1278E+03, 6.1346E+03, 6.1414E+03, 6.1482E+03,
     + 6.1550E+03, 6.1618E+03, 6.1686E+03, 6.1754E+03, 6.1822E+03,
     +  6.1890E+03, 6.1958E+03, 6.2026E+03, 6.2094E+03, 6.2163E+03,
     + 6.2231E+03, 6.2299E+03, 6.2368E+03, 6.2436E+03, 6.2504E+03,
     +  6.2573E+03, 6.2641E+03, 6.2710E+03, 6.2779E+03, 6.2847E+03,
     + 6.2916E+03, 6.2985E+03, 6.3053E+03, 6.3122E+03, 6.3191E+03,
     +  6.3260E+03, 6.3329E+03, 6.3397E+03, 6.3466E+03, 6.3535E+03,
     + 6.3604E+03, 6.3673E+03, 6.3743E+03, 6.3812E+03, 6.3881E+03,
     +  6.3950E+03, 6.4019E+03, 6.4088E+03, 6.4158E+03, 6.4227E+03,
     + 6.4296E+03, 6.4366E+03, 6.4435E+03, 6.4505E+03, 6.4574E+03,
     +  6.4644E+03, 6.4713E+03, 6.4783E+03, 6.4853E+03, 6.4922E+03,
     + 6.4992E+03, 6.5062E+03, 6.5132E+03, 6.5201E+03, 6.5271E+03,
     +  6.5341E+03, 6.5411E+03, 6.5481E+03, 6.5551E+03, 6.5621E+03,
     + 6.5691E+03, 6.5761E+03, 6.5831E+03, 6.5902E+03, 6.5972E+03,
     +  6.6042E+03, 6.6112E+03, 6.6183E+03, 6.6253E+03, 6.6323E+03,
     + 6.6394E+03, 6.6464E+03, 6.6535E+03, 6.6605E+03, 6.6676E+03,
     +  6.6747E+03, 6.6817E+03, 6.6888E+03, 6.6959E+03, 6.7029E+03,
     + 6.7100E+03, 6.7171E+03, 6.7242E+03, 6.7313E+03, 6.7384E+03,
     +  6.7454E+03, 6.7525E+03, 6.7597E+03, 6.7668E+03, 6.7739E+03,
     + 6.7810E+03, 6.7881E+03, 6.7952E+03, 6.8023E+03, 6.8095E+03,
     +  6.8166E+03, 6.8237E+03, 6.8309E+03, 6.8380E+03, 6.8452E+03,
     + 6.8523E+03, 6.8595E+03, 6.8666E+03, 6.8738E+03, 6.8809E+03,
     +  6.8881E+03, 6.8953E+03, 6.9024E+03, 6.9096E+03, 6.9168E+03,
     + 6.9240E+03, 6.9312E+03, 6.9383E+03, 6.9455E+03, 6.9527E+03,
     +  6.9599E+03, 6.9671E+03, 6.9743E+03, 6.9816E+03, 6.9888E+03,
     + 6.9960E+03, 7.0032E+03, 7.0104E+03, 7.0177E+03, 7.0249E+03,
     +  7.0321E+03, 7.0394E+03, 7.0466E+03, 7.0539E+03, 7.0611E+03,
     + 7.0684E+03, 7.0756E+03, 7.0829E+03, 7.0901E+03, 7.0974E+03,
     +  7.1047E+03, 7.1119E+03, 7.1192E+03, 7.1265E+03, 7.1338E+03,
     + 7.1411E+03, 7.1484E+03, 7.1556E+03, 7.1629E+03, 7.1702E+03,
     +  7.1775E+03, 7.1849E+03, 7.1922E+03, 7.1995E+03, 7.2068E+03,
     + 7.2141E+03, 7.2214E+03, 7.2288E+03, 7.2361E+03, 7.2434E+03,
     +  7.2508E+03, 7.2581E+03, 7.2655E+03, 7.2728E+03, 7.2802E+03,
     + 7.2875E+03, 7.2949E+03, 7.3022E+03, 7.3096E+03, 7.3170E+03,
     +  7.3243E+03, 7.3317E+03, 7.3391E+03, 7.3465E+03, 7.3539E+03,
     + 7.3613E+03, 7.3687E+03, 7.3761E+03, 7.3835E+03, 7.3909E+03,
     +  7.3983E+03, 7.4057E+03, 7.4131E+03, 7.4205E+03, 7.4279E+03,
     + 7.4354E+03, 7.4428E+03, 7.4502E+03, 7.4576E+03, 7.4651E+03,
     +  7.4725E+03, 7.4800E+03, 7.4874E+03, 7.4949E+03, 7.5023E+03,
     + 7.5098E+03, 7.5172E+03, 7.5247E+03, 7.5322E+03, 7.5396E+03,
     +  7.5471E+03, 7.5546E+03, 7.5621E+03, 7.5696E+03, 7.5771E+03,
     + 7.5845E+03, 7.5920E+03, 7.5995E+03, 7.6070E+03, 7.6145E+03,
     +  7.6221E+03, 7.6296E+03, 7.6371E+03, 7.6446E+03, 7.6521E+03,
     + 7.6597E+03, 7.6672E+03, 7.6747E+03, 7.6822E+03, 7.6898E+03,
     +  7.6973E+03, 7.7049E+03, 7.7124E+03, 7.7200E+03, 7.7275E+03,
     + 7.7351E+03, 7.7427E+03, 7.7502E+03, 7.7578E+03, 7.7654E+03,
     +  7.7729E+03, 7.7805E+03, 7.7881E+03, 7.7957E+03, 7.8033E+03,
     + 7.8109E+03, 7.8185E+03, 7.8261E+03, 7.8337E+03, 7.8413E+03,
     +  7.8489E+03, 7.8565E+03, 7.8641E+03, 7.8717E+03, 7.8794E+03,
     + 7.8870E+03, 7.8946E+03, 7.9023E+03, 7.9099E+03, 7.9175E+03,
     +  7.9252E+03, 7.9328E+03, 7.9405E+03, 7.9481E+03, 7.9558E+03,
     + 7.9634E+03, 7.9711E+03, 7.9788E+03, 7.9864E+03, 7.9941E+03,
     +  8.0018E+03, 8.0095E+03, 8.0172E+03, 8.0248E+03, 8.0325E+03,
     + 8.0402E+03, 8.0479E+03, 8.0556E+03, 8.0633E+03, 8.0710E+03,
     +  8.0787E+03, 8.0865E+03, 8.0942E+03, 8.1019E+03, 8.1096E+03,
     + 8.1173E+03, 8.1251E+03, 8.1328E+03, 8.1405E+03, 8.1483E+03,
     +  8.1560E+03, 8.1638E+03, 8.1715E+03, 8.1793E+03, 8.1870E+03,
     + 8.1948E+03, 8.2026E+03, 8.2103E+03, 8.2181E+03, 8.2259E+03,
     +  8.2336E+03, 8.2414E+03, 8.2492E+03, 8.2570E+03, 8.2648E+03,
     + 8.2726E+03, 8.2804E+03, 8.2882E+03, 8.2960E+03, 8.3038E+03,
     +  8.3116E+03, 8.3194E+03, 8.3272E+03, 8.3350E+03, 8.3429E+03,
     + 8.3507E+03, 8.3585E+03, 8.3663E+03, 8.3742E+03, 8.3820E+03,
     +  8.3899E+03, 8.3977E+03, 8.4056E+03, 8.4134E+03, 8.4213E+03,
     + 8.4291E+03, 8.4370E+03, 8.4448E+03, 8.4527E+03, 8.4606E+03,
     +  8.4685E+03, 8.4763E+03, 8.4842E+03, 8.4921E+03, 8.5000E+03,
     + 8.5079E+03, 8.5158E+03, 8.5237E+03, 8.5316E+03, 8.5395E+03,
     +  8.5474E+03, 8.5553E+03, 8.5632E+03, 8.5711E+03, 8.5790E+03,
     + 8.5870E+03, 8.5949E+03, 8.6028E+03, 8.6108E+03, 8.6187E+03,
     +  8.6266E+03, 8.6346E+03, 8.6425E+03, 8.6505E+03, 8.6584E+03,
     + 8.6664E+03, 8.6743E+03, 8.6823E+03, 8.6903E+03, 8.6982E+03,
     +  8.7062E+03, 8.7142E+03, 8.7222E+03, 8.7302E+03, 8.7381E+03,
     + 8.7461E+03, 8.7541E+03, 8.7621E+03, 8.7701E+03, 8.7781E+03,
     +  8.7861E+03, 8.7941E+03, 8.8021E+03, 8.8101E+03, 8.8182E+03,
     + 8.8262E+03, 8.8342E+03, 8.8422E+03, 8.8503E+03, 8.8583E+03,
     +  8.8663E+03, 8.8744E+03, 8.8824E+03, 8.8905E+03, 8.8985E+03,
     + 8.9066E+03, 8.9146E+03, 8.9227E+03, 8.9307E+03, 8.9388E+03,
     +  8.9469E+03, 8.9549E+03, 8.9630E+03, 8.9711E+03, 8.9792E+03,
     + 8.9873E+03, 8.9954E+03, 9.0034E+03, 9.0115E+03, 9.0196E+03,
     +  9.0277E+03, 9.0358E+03, 9.0439E+03, 9.0521E+03, 9.0602E+03,
     + 9.0683E+03, 9.0764E+03, 9.0845E+03, 9.0927E+03, 9.1008E+03,
     +  9.1089E+03, 9.1170E+03, 9.1252E+03, 9.1333E+03, 9.1415E+03,
     + 9.1496E+03, 9.1578E+03, 9.1659E+03, 9.1741E+03, 9.1822E+03,
     +  9.1904E+03, 9.1986E+03, 9.2067E+03, 9.2149E+03, 9.2231E+03,
     + 9.2313E+03, 9.2395E+03, 9.2476E+03, 9.2558E+03, 9.2640E+03,
     +  9.2722E+03, 9.2804E+03, 9.2886E+03, 9.2968E+03, 9.3050E+03,
     + 9.3132E+03, 9.3215E+03, 9.3297E+03, 9.3379E+03, 9.3461E+03,
     +  9.3543E+03, 9.3626E+03, 9.3708E+03, 9.3790E+03, 9.3873E+03,
     + 9.3955E+03, 9.4038E+03, 9.4120E+03, 9.4203E+03, 9.4285E+03,
     +  9.4368E+03, 9.4450E+03, 9.4533E+03, 9.4616E+03, 9.4698E+03,
     + 9.4781E+03, 9.4864E+03, 9.4947E+03, 9.5030E+03, 9.5112E+03,
     +  9.5195E+03, 9.5278E+03, 9.5361E+03, 9.5444E+03, 9.5527E+03,
     + 9.5610E+03, 9.5693E+03, 9.5776E+03, 9.5860E+03, 9.5943E+03,
     +  9.6026E+03, 9.6109E+03, 9.6192E+03, 9.6276E+03, 9.6359E+03,
     + 9.6442E+03, 9.6526E+03, 9.6609E+03, 9.6693E+03, 9.6776E+03,
     +  9.6860E+03, 9.6943E+03, 9.7027E+03, 9.7110E+03, 9.7194E+03,
     + 9.7278E+03, 9.7361E+03, 9.7445E+03, 9.7529E+03, 9.7613E+03,
     +  9.7696E+03, 9.7780E+03, 9.7864E+03, 9.7948E+03, 9.8032E+03,
     + 9.8116E+03, 9.8200E+03, 9.8284E+03, 9.8368E+03, 9.8452E+03,
     +  9.8536E+03, 9.8621E+03, 9.8705E+03, 9.8789E+03, 9.8873E+03,
     + 9.8957E+03, 9.9042E+03, 9.9126E+03, 9.9210E+03, 9.9295E+03,
     +  9.9379E+03, 9.9464E+03, 9.9548E+03, 9.9633E+03, 9.9717E+03,
     + 9.9802E+03, 9.9887E+03, 9.9971E+03, 1.0006E+04, 1.0014E+04,
     +  1.0023E+04, 1.0031E+04, 1.0039E+04, 1.0048E+04, 1.0056E+04,
     + 1.0065E+04, 1.0073E+04, 1.0082E+04, 1.0090E+04, 1.0099E+04,
     +  1.0107E+04, 1.0116E+04, 1.0124E+04, 1.0133E+04, 1.0141E+04,
     + 1.0150E+04, 1.0159E+04, 1.0167E+04, 1.0176E+04, 1.0184E+04,
     +  1.0193E+04, 1.0201E+04, 1.0210E+04, 1.0218E+04, 1.0227E+04,
     + 1.0235E+04, 1.0244E+04, 1.0252E+04, 1.0261E+04, 1.0270E+04,
     +  1.0278E+04, 1.0287E+04, 1.0295E+04, 1.0304E+04, 1.0312E+04,
     + 1.0321E+04, 1.0330E+04, 1.0338E+04, 1.0347E+04, 1.0355E+04,
     +  1.0364E+04, 1.0373E+04, 1.0381E+04, 1.0390E+04, 1.0398E+04,
     + 1.0407E+04, 1.0416E+04, 1.0424E+04, 1.0433E+04, 1.0441E+04,
     +  1.0450E+04, 1.0459E+04, 1.0467E+04, 1.0476E+04, 1.0485E+04,
     + 1.0493E+04, 1.0502E+04, 1.0511E+04, 1.0519E+04, 1.0528E+04,
     +  1.0536E+04, 1.0545E+04, 1.0554E+04, 1.0562E+04, 1.0571E+04,
     + 1.0580E+04, 1.0588E+04, 1.0597E+04, 1.0606E+04, 1.0614E+04,
     +  1.0623E+04, 1.0632E+04, 1.0641E+04, 1.0649E+04, 1.0658E+04,
     + 1.0667E+04, 1.0675E+04, 1.0684E+04, 1.0693E+04, 1.0701E+04,
     +  1.0710E+04, 1.0719E+04, 1.0728E+04, 1.0736E+04, 1.0745E+04,
     + 1.0754E+04, 1.0762E+04, 1.0771E+04, 1.0780E+04, 1.0789E+04,
     +  1.0797E+04, 1.0806E+04, 1.0815E+04, 1.0824E+04, 1.0832E+04,
     + 1.0841E+04, 1.0850E+04, 1.0859E+04, 1.0867E+04, 1.0876E+04,
     +  1.0885E+04, 1.0894E+04, 1.0903E+04, 1.0911E+04, 1.0920E+04,
     + 1.0929E+04, 1.0938E+04, 1.0946E+04, 1.0955E+04, 1.0964E+04,
     +  1.0973E+04, 1.0982E+04, 1.0990E+04, 1.0999E+04, 1.1008E+04,
     + 1.1017E+04, 1.1026E+04, 1.1035E+04, 1.1043E+04, 1.1052E+04,
     +  1.1061E+04, 1.1070E+04, 1.1079E+04, 1.1088E+04, 1.1096E+04,
     + 1.1105E+04, 1.1114E+04, 1.1123E+04, 1.1132E+04, 1.1141E+04,
     +  1.1149E+04, 1.1158E+04, 1.1167E+04, 1.1176E+04, 1.1185E+04,
     + 1.1194E+04, 1.1203E+04, 1.1212E+04, 1.1220E+04, 1.1229E+04,
     +  1.1238E+04, 1.1247E+04, 1.1256E+04, 1.1265E+04, 1.1274E+04,
     + 1.1283E+04, 1.1292E+04, 1.1301E+04, 1.1309E+04, 1.1318E+04,
     +  1.1327E+04, 1.1336E+04, 1.1345E+04, 1.1354E+04, 1.1363E+04,
     + 1.1372E+04, 1.1381E+04, 1.1390E+04, 1.1399E+04, 1.1408E+04,
     +  1.1417E+04, 1.1426E+04, 1.1434E+04, 1.1443E+04, 1.1452E+04,
     + 1.1461E+04, 1.1470E+04, 1.1479E+04, 1.1488E+04, 1.1497E+04,
     +  1.1506E+04, 1.1515E+04, 1.1524E+04, 1.1533E+04, 1.1542E+04,
     + 1.1551E+04, 1.1560E+04, 1.1569E+04, 1.1578E+04, 1.1587E+04,
     +  1.1596E+04, 1.1605E+04, 1.1614E+04, 1.1623E+04, 1.1632E+04,
     + 1.1641E+04, 1.1650E+04, 1.1659E+04, 1.1668E+04, 1.1677E+04,
     +  1.1686E+04, 1.1695E+04, 1.1704E+04, 1.1713E+04, 1.1722E+04,
     + 1.1731E+04, 1.1740E+04, 1.1749E+04, 1.1759E+04, 1.1768E+04,
     +  1.1777E+04, 1.1786E+04, 1.1795E+04, 1.1804E+04, 1.1813E+04,
     + 1.1822E+04, 1.1831E+04, 1.1840E+04, 1.1849E+04, 1.1858E+04,
     +  1.1867E+04, 1.1876E+04, 1.1886E+04, 1.1895E+04, 1.1904E+04,
     + 1.1913E+04, 1.1922E+04, 1.1931E+04, 1.1940E+04, 1.1949E+04,
     +  1.1958E+04, 1.1967E+04, 1.1977E+04, 1.1986E+04, 1.1995E+04,
     + 1.2004E+04, 1.2013E+04, 1.2022E+04, 1.2031E+04, 1.2040E+04,
     +  1.2050E+04, 1.2059E+04, 1.2068E+04, 1.2077E+04, 1.2086E+04,
     + 1.2095E+04, 1.2104E+04, 1.2114E+04, 1.2123E+04, 1.2132E+04,
     +  1.2141E+04, 1.2150E+04, 1.2159E+04, 1.2169E+04, 1.2178E+04,
     + 1.2187E+04, 1.2196E+04, 1.2205E+04, 1.2215E+04, 1.2224E+04,
     +  1.2233E+04, 1.2242E+04, 1.2251E+04, 1.2261E+04, 1.2270E+04,
     + 1.2279E+04, 1.2288E+04, 1.2297E+04, 1.2307E+04, 1.2316E+04,
     +  1.2325E+04, 1.2334E+04, 1.2343E+04, 1.2353E+04, 1.2362E+04,
     + 1.2371E+04, 1.2380E+04, 1.2390E+04, 1.2399E+04, 1.2408E+04,
     +  1.2417E+04, 1.2427E+04, 1.2436E+04, 1.2445E+04, 1.2454E+04,
     + 1.2464E+04, 1.2473E+04, 1.2482E+04, 1.2491E+04, 1.2501E+04,
     +  1.2510E+04, 1.2519E+04, 1.2529E+04, 1.2538E+04, 1.2547E+04,
     + 1.2556E+04, 1.2566E+04, 1.2575E+04, 1.2584E+04, 1.2594E+04,
     +  1.2603E+04, 1.2612E+04, 1.2622E+04, 1.2631E+04, 1.2640E+04,
     + 1.2649E+04, 1.2659E+04, 1.2668E+04, 1.2677E+04, 1.2687E+04,
     +  1.2696E+04, 1.2705E+04, 1.2715E+04, 1.2724E+04, 1.2733E+04,
     + 1.2743E+04, 1.2752E+04, 1.2761E+04, 1.2771E+04, 1.2780E+04,
     +  1.2790E+04, 1.2799E+04, 1.2808E+04, 1.2818E+04, 1.2827E+04,
     + 1.2836E+04, 1.2846E+04, 1.2855E+04, 1.2864E+04, 1.2874E+04,
     +  1.2883E+04, 1.2893E+04, 1.2902E+04, 1.2911E+04, 1.2921E+04,
     + 1.2930E+04, 1.2940E+04, 1.2949E+04, 1.2958E+04, 1.2968E+04,
     +  1.2977E+04, 1.2987E+04, 1.2996E+04, 1.3005E+04, 1.3015E+04,
     + 1.3024E+04, 1.3034E+04, 1.3043E+04, 1.3053E+04, 1.3062E+04,
     +  1.3071E+04, 1.3081E+04, 1.3090E+04, 1.3100E+04, 1.3109E+04,
     + 1.3119E+04, 1.3128E+04, 1.3138E+04, 1.3147E+04, 1.3156E+04,
     +  1.3166E+04, 1.3175E+04, 1.3185E+04, 1.3194E+04, 1.3204E+04,
     + 1.3213E+04, 1.3223E+04, 1.3232E+04, 1.3242E+04, 1.3251E+04,
     +  1.3261E+04, 1.3270E+04, 1.3280E+04, 1.3289E+04, 1.3299E+04,
     + 1.3308E+04, 1.3318E+04, 1.3327E+04, 1.3337E+04, 1.3346E+04,
     +  1.3356E+04, 1.3365E+04, 1.3375E+04, 1.3384E+04, 1.3394E+04,
     + 1.3403E+04, 1.3413E+04, 1.3422E+04, 1.3432E+04, 1.3441E+04,
     +  1.3451E+04, 1.3460E+04, 1.3470E+04, 1.3480E+04, 1.3489E+04,
     + 1.3499E+04, 1.3508E+04, 1.3518E+04, 1.3527E+04, 1.3537E+04,
     +  1.3546E+04, 1.3556E+04, 1.3566E+04, 1.3575E+04, 1.3585E+04,
     + 1.3594E+04, 1.3604E+04, 1.3613E+04, 1.3623E+04, 1.3633E+04,
     +  1.3642E+04, 1.3652E+04, 1.3661E+04, 1.3671E+04, 1.3681E+04,
     + 1.3690E+04, 1.3700E+04, 1.3709E+04, 1.3719E+04, 1.3729E+04,
     +  1.3738E+04, 1.3748E+04, 1.3757E+04, 1.3767E+04, 1.3777E+04,
     + 1.3786E+04, 1.3796E+04, 1.3806E+04, 1.3815E+04, 1.3825E+04,
     +  1.3834E+04, 1.3844E+04, 1.3854E+04, 1.3863E+04, 1.3873E+04,
     + 1.3883E+04, 1.3892E+04, 1.3902E+04, 1.3912E+04, 1.3921E+04,
     +  1.3931E+04, 1.3941E+04, 1.3950E+04, 1.3960E+04, 1.3970E+04,
     + 1.3979E+04, 1.3989E+04, 1.3999E+04, 1.4008E+04, 1.4018E+04,
     +  1.4028E+04, 1.4037E+04, 1.4047E+04, 1.4057E+04, 1.4067E+04,
     + 1.4076E+04, 1.4086E+04, 1.4096E+04, 1.4105E+04, 1.4115E+04,
     +  1.4125E+04, 1.4135E+04, 1.4144E+04, 1.4154E+04, 1.4164E+04,
     + 1.4173E+04, 1.4183E+04, 1.4193E+04, 1.4203E+04, 1.4212E+04,
     +  1.4222E+04, 1.4232E+04, 1.4242E+04, 1.4251E+04, 1.4261E+04,
     + 1.4271E+04, 1.4281E+04, 1.4290E+04, 1.4300E+04, 1.4310E+04,
     +  1.4320E+04, 1.4329E+04, 1.4339E+04, 1.4349E+04, 1.4359E+04,
     + 1.4368E+04, 1.4378E+04, 1.4388E+04, 1.4398E+04, 1.4408E+04,
     +  1.4417E+04, 1.4427E+04, 1.4437E+04, 1.4447E+04, 1.4457E+04,
     + 1.4466E+04, 1.4476E+04, 1.4486E+04, 1.4496E+04, 1.4506E+04,
     +  1.4515E+04, 1.4525E+04, 1.4535E+04, 1.4545E+04, 1.4555E+04,
     + 1.4564E+04, 1.4574E+04, 1.4584E+04, 1.4594E+04, 1.4604E+04,
     +  1.4614E+04, 1.4623E+04, 1.4633E+04, 1.4643E+04, 1.4653E+04,
     + 1.4663E+04, 1.4673E+04, 1.4683E+04, 1.4692E+04, 1.4702E+04,
     +  1.4712E+04, 1.4722E+04, 1.4732E+04, 1.4742E+04, 1.4752E+04,
     + 1.4761E+04, 1.4771E+04, 1.4781E+04, 1.4791E+04, 1.4801E+04,
     +  1.4811E+04, 1.4821E+04, 1.4831E+04, 1.4841E+04, 1.4850E+04/
      data (QofT( 1,J),J=2409,4510)/ 1.4860E+04, 1.4870E+04,
     +  1.4880E+04, 1.4890E+04, 1.4900E+04, 1.4910E+04, 1.4920E+04,
     +  1.4930E+04,
     + 1.4940E+04, 1.4949E+04, 1.4959E+04, 1.4969E+04, 1.4979E+04,
     +  1.4989E+04, 1.4999E+04, 1.5009E+04, 1.5019E+04, 1.5029E+04,
     + 1.5039E+04, 1.5049E+04, 1.5059E+04, 1.5069E+04, 1.5079E+04,
     +  1.5089E+04, 1.5099E+04, 1.5108E+04, 1.5118E+04, 1.5128E+04,
     + 1.5138E+04, 1.5148E+04, 1.5158E+04, 1.5168E+04, 1.5178E+04,
     +  1.5188E+04, 1.5198E+04, 1.5208E+04, 1.5218E+04, 1.5228E+04,
     + 1.5238E+04, 1.5248E+04, 1.5258E+04, 1.5268E+04, 1.5278E+04,
     +  1.5288E+04, 1.5298E+04, 1.5308E+04, 1.5318E+04, 1.5328E+04,
     + 1.5338E+04, 1.5348E+04, 1.5358E+04, 1.5368E+04, 1.5378E+04,
     +  1.5388E+04, 1.5398E+04, 1.5408E+04, 1.5418E+04, 1.5428E+04,
     + 1.5438E+04, 1.5448E+04, 1.5458E+04, 1.5468E+04, 1.5478E+04,
     +  1.5488E+04, 1.5499E+04, 1.5509E+04, 1.5519E+04, 1.5529E+04,
     + 1.5539E+04, 1.5549E+04, 1.5559E+04, 1.5569E+04, 1.5579E+04,
     +  1.5589E+04, 1.5599E+04, 1.5609E+04, 1.5619E+04, 1.5629E+04,
     + 1.5639E+04, 1.5649E+04, 1.5660E+04, 1.5670E+04, 1.5680E+04,
     +  1.5690E+04, 1.5700E+04, 1.5710E+04, 1.5720E+04, 1.5730E+04,
     + 1.5740E+04, 1.5750E+04, 1.5760E+04, 1.5771E+04, 1.5781E+04,
     +  1.5791E+04, 1.5801E+04, 1.5811E+04, 1.5821E+04, 1.5831E+04,
     + 1.5841E+04, 1.5852E+04, 1.5862E+04, 1.5872E+04, 1.5882E+04,
     +  1.5892E+04, 1.5902E+04, 1.5912E+04, 1.5922E+04, 1.5933E+04,
     + 1.5943E+04, 1.5953E+04, 1.5963E+04, 1.5973E+04, 1.5983E+04,
     +  1.5994E+04, 1.6004E+04, 1.6014E+04, 1.6024E+04, 1.6034E+04,
     + 1.6044E+04, 1.6055E+04, 1.6065E+04, 1.6075E+04, 1.6085E+04,
     +  1.6095E+04, 1.6105E+04, 1.6116E+04, 1.6126E+04, 1.6136E+04,
     + 1.6146E+04, 1.6156E+04, 1.6167E+04, 1.6177E+04, 1.6187E+04,
     +  1.6197E+04, 1.6207E+04, 1.6218E+04, 1.6228E+04, 1.6238E+04,
     + 1.6248E+04, 1.6258E+04, 1.6269E+04, 1.6279E+04, 1.6289E+04,
     +  1.6299E+04, 1.6310E+04, 1.6320E+04, 1.6330E+04, 1.6340E+04,
     + 1.6350E+04, 1.6361E+04, 1.6371E+04, 1.6381E+04, 1.6391E+04,
     +  1.6402E+04, 1.6412E+04, 1.6422E+04, 1.6432E+04, 1.6443E+04,
     + 1.6453E+04, 1.6463E+04, 1.6473E+04, 1.6484E+04, 1.6494E+04,
     +  1.6504E+04, 1.6514E+04, 1.6525E+04, 1.6535E+04, 1.6545E+04,
     + 1.6556E+04, 1.6566E+04, 1.6576E+04, 1.6586E+04, 1.6597E+04,
     +  1.6607E+04, 1.6617E+04, 1.6628E+04, 1.6638E+04, 1.6648E+04,
     + 1.6658E+04, 1.6669E+04, 1.6679E+04, 1.6689E+04, 1.6700E+04,
     +  1.6710E+04, 1.6720E+04, 1.6731E+04, 1.6741E+04, 1.6751E+04,
     + 1.6762E+04, 1.6772E+04, 1.6782E+04, 1.6793E+04, 1.6803E+04,
     +  1.6813E+04, 1.6824E+04, 1.6834E+04, 1.6844E+04, 1.6855E+04,
     + 1.6865E+04, 1.6875E+04, 1.6886E+04, 1.6896E+04, 1.6906E+04,
     +  1.6917E+04, 1.6927E+04, 1.6937E+04, 1.6948E+04, 1.6958E+04,
     + 1.6968E+04, 1.6979E+04, 1.6989E+04, 1.7000E+04, 1.7010E+04,
     +  1.7020E+04, 1.7031E+04, 1.7041E+04, 1.7051E+04, 1.7062E+04,
     + 1.7072E+04, 1.7083E+04, 1.7093E+04, 1.7103E+04, 1.7114E+04,
     +  1.7124E+04, 1.7135E+04, 1.7145E+04, 1.7155E+04, 1.7166E+04,
     + 1.7176E+04, 1.7187E+04, 1.7197E+04, 1.7207E+04, 1.7218E+04,
     +  1.7228E+04, 1.7239E+04, 1.7249E+04, 1.7260E+04, 1.7270E+04,
     + 1.7280E+04, 1.7291E+04, 1.7301E+04, 1.7312E+04, 1.7322E+04,
     +  1.7333E+04, 1.7343E+04, 1.7353E+04, 1.7364E+04, 1.7374E+04,
     + 1.7385E+04, 1.7395E+04, 1.7406E+04, 1.7416E+04, 1.7427E+04,
     +  1.7437E+04, 1.7448E+04, 1.7458E+04, 1.7468E+04, 1.7479E+04,
     + 1.7489E+04, 1.7500E+04, 1.7510E+04, 1.7521E+04, 1.7531E+04,
     +  1.7542E+04, 1.7552E+04, 1.7563E+04, 1.7573E+04, 1.7584E+04,
     + 1.7594E+04, 1.7605E+04, 1.7615E+04, 1.7626E+04, 1.7636E+04,
     +  1.7647E+04, 1.7657E+04, 1.7668E+04, 1.7678E+04, 1.7689E+04,
     + 1.7699E+04, 1.7710E+04, 1.7720E+04, 1.7731E+04, 1.7741E+04,
     +  1.7752E+04, 1.7762E+04, 1.7773E+04, 1.7783E+04, 1.7794E+04,
     + 1.7804E+04, 1.7815E+04, 1.7825E+04, 1.7836E+04, 1.7847E+04,
     +  1.7857E+04, 1.7868E+04, 1.7878E+04, 1.7889E+04, 1.7899E+04,
     + 1.7910E+04, 1.7920E+04, 1.7931E+04, 1.7941E+04, 1.7952E+04,
     +  1.7963E+04, 1.7973E+04, 1.7984E+04, 1.7994E+04, 1.8005E+04,
     + 1.8015E+04, 1.8026E+04, 1.8037E+04, 1.8047E+04, 1.8058E+04,
     +  1.8068E+04, 1.8079E+04, 1.8089E+04, 1.8100E+04, 1.8111E+04,
     + 1.8121E+04, 1.8132E+04, 1.8142E+04, 1.8153E+04, 1.8164E+04,
     +  1.8174E+04, 1.8185E+04, 1.8195E+04, 1.8206E+04, 1.8217E+04,
     + 1.8227E+04, 1.8238E+04, 1.8249E+04, 1.8259E+04, 1.8270E+04,
     +  1.8280E+04, 1.8291E+04, 1.8302E+04, 1.8312E+04, 1.8323E+04,
     + 1.8334E+04, 1.8344E+04, 1.8355E+04, 1.8365E+04, 1.8376E+04,
     +  1.8387E+04, 1.8397E+04, 1.8408E+04, 1.8419E+04, 1.8429E+04,
     + 1.8440E+04, 1.8451E+04, 1.8461E+04, 1.8472E+04, 1.8483E+04,
     +  1.8493E+04, 1.8504E+04, 1.8515E+04, 1.8525E+04, 1.8536E+04,
     + 1.8547E+04, 1.8557E+04, 1.8568E+04, 1.8579E+04, 1.8589E+04,
     +  1.8600E+04, 1.8611E+04, 1.8621E+04, 1.8632E+04, 1.8643E+04,
     + 1.8653E+04, 1.8664E+04, 1.8675E+04, 1.8685E+04, 1.8696E+04,
     +  1.8707E+04, 1.8718E+04, 1.8728E+04, 1.8739E+04, 1.8750E+04,
     + 1.8760E+04, 1.8771E+04, 1.8782E+04, 1.8792E+04, 1.8803E+04,
     +  1.8814E+04, 1.8825E+04, 1.8835E+04, 1.8846E+04, 1.8857E+04,
     + 1.8868E+04, 1.8878E+04, 1.8889E+04, 1.8900E+04, 1.8910E+04,
     +  1.8921E+04, 1.8932E+04, 1.8943E+04, 1.8953E+04, 1.8964E+04,
     + 1.8975E+04, 1.8986E+04, 1.8996E+04, 1.9007E+04, 1.9018E+04,
     +  1.9029E+04, 1.9039E+04, 1.9050E+04, 1.9061E+04, 1.9072E+04,
     + 1.9083E+04, 1.9093E+04, 1.9104E+04, 1.9115E+04, 1.9126E+04,
     +  1.9136E+04, 1.9147E+04, 1.9158E+04, 1.9169E+04, 1.9179E+04,
     + 1.9190E+04, 1.9201E+04, 1.9212E+04, 1.9223E+04, 1.9233E+04,
     +  1.9244E+04, 1.9255E+04, 1.9266E+04, 1.9277E+04, 1.9287E+04,
     + 1.9298E+04, 1.9309E+04, 1.9320E+04, 1.9331E+04, 1.9341E+04,
     +  1.9352E+04, 1.9363E+04, 1.9374E+04, 1.9385E+04, 1.9396E+04,
     + 1.9406E+04, 1.9417E+04, 1.9428E+04, 1.9439E+04, 1.9450E+04,
     +  1.9461E+04, 1.9471E+04, 1.9482E+04, 1.9493E+04, 1.9504E+04,
     + 1.9515E+04, 1.9526E+04, 1.9536E+04, 1.9547E+04, 1.9558E+04,
     +  1.9569E+04, 1.9580E+04, 1.9591E+04, 1.9601E+04, 1.9612E+04,
     + 1.9623E+04, 1.9634E+04, 1.9645E+04, 1.9656E+04, 1.9667E+04,
     +  1.9678E+04, 1.9688E+04, 1.9699E+04, 1.9710E+04, 1.9721E+04,
     + 1.9732E+04, 1.9743E+04, 1.9754E+04, 1.9765E+04, 1.9775E+04,
     +  1.9786E+04, 1.9797E+04, 1.9808E+04, 1.9819E+04, 1.9830E+04,
     + 1.9841E+04, 1.9852E+04, 1.9863E+04, 1.9873E+04, 1.9884E+04,
     +  1.9895E+04, 1.9906E+04, 1.9917E+04, 1.9928E+04, 1.9939E+04,
     + 1.9950E+04, 1.9961E+04, 1.9972E+04, 1.9983E+04, 1.9994E+04,
     +  2.0004E+04, 2.0015E+04, 2.0026E+04, 2.0037E+04, 2.0048E+04,
     + 2.0059E+04, 2.0070E+04, 2.0081E+04, 2.0092E+04, 2.0103E+04,
     +  2.0114E+04, 2.0125E+04, 2.0136E+04, 2.0147E+04, 2.0158E+04,
     + 2.0169E+04, 2.0179E+04, 2.0190E+04, 2.0201E+04, 2.0212E+04,
     +  2.0223E+04, 2.0234E+04, 2.0245E+04, 2.0256E+04, 2.0267E+04,
     + 2.0278E+04, 2.0289E+04, 2.0300E+04, 2.0311E+04, 2.0322E+04,
     +  2.0333E+04, 2.0344E+04, 2.0355E+04, 2.0366E+04, 2.0377E+04,
     + 2.0388E+04, 2.0399E+04, 2.0410E+04, 2.0421E+04, 2.0432E+04,
     +  2.0443E+04, 2.0454E+04, 2.0465E+04, 2.0476E+04, 2.0487E+04,
     + 2.0498E+04, 2.0509E+04, 2.0520E+04, 2.0531E+04, 2.0542E+04,
     +  2.0553E+04, 2.0564E+04, 2.0575E+04, 2.0586E+04, 2.0597E+04,
     + 2.0608E+04, 2.0619E+04, 2.0630E+04, 2.0641E+04, 2.0652E+04,
     +  2.0663E+04, 2.0674E+04, 2.0685E+04, 2.0696E+04, 2.0707E+04,
     + 2.0718E+04, 2.0729E+04, 2.0740E+04, 2.0751E+04, 2.0762E+04,
     +  2.0773E+04, 2.0785E+04, 2.0796E+04, 2.0807E+04, 2.0818E+04,
     + 2.0829E+04, 2.0840E+04, 2.0851E+04, 2.0862E+04, 2.0873E+04,
     +  2.0884E+04, 2.0895E+04, 2.0906E+04, 2.0917E+04, 2.0928E+04,
     + 2.0939E+04, 2.0950E+04, 2.0962E+04, 2.0973E+04, 2.0984E+04,
     +  2.0995E+04, 2.1006E+04, 2.1017E+04, 2.1028E+04, 2.1039E+04,
     + 2.1050E+04, 2.1061E+04, 2.1072E+04, 2.1083E+04, 2.1095E+04,
     +  2.1106E+04, 2.1117E+04, 2.1128E+04, 2.1139E+04, 2.1150E+04,
     + 2.1161E+04, 2.1172E+04, 2.1183E+04, 2.1195E+04, 2.1206E+04,
     +  2.1217E+04, 2.1228E+04, 2.1239E+04, 2.1250E+04, 2.1261E+04,
     + 2.1272E+04, 2.1283E+04, 2.1295E+04, 2.1306E+04, 2.1317E+04,
     +  2.1328E+04, 2.1339E+04, 2.1350E+04, 2.1361E+04, 2.1373E+04,
     + 2.1384E+04, 2.1395E+04, 2.1406E+04, 2.1417E+04, 2.1428E+04,
     +  2.1439E+04, 2.1451E+04, 2.1462E+04, 2.1473E+04, 2.1484E+04,
     + 2.1495E+04, 2.1506E+04, 2.1517E+04, 2.1529E+04, 2.1540E+04,
     +  2.1551E+04, 2.1562E+04, 2.1573E+04, 2.1584E+04, 2.1596E+04,
     + 2.1607E+04, 2.1618E+04, 2.1629E+04, 2.1640E+04, 2.1651E+04,
     +  2.1663E+04, 2.1674E+04, 2.1685E+04, 2.1696E+04, 2.1707E+04,
     + 2.1719E+04, 2.1730E+04, 2.1741E+04, 2.1752E+04, 2.1763E+04,
     +  2.1775E+04, 2.1786E+04, 2.1797E+04, 2.1808E+04, 2.1819E+04,
     + 2.1831E+04, 2.1842E+04, 2.1853E+04, 2.1864E+04, 2.1875E+04,
     +  2.1887E+04, 2.1898E+04, 2.1909E+04, 2.1920E+04, 2.1932E+04,
     + 2.1943E+04, 2.1954E+04, 2.1965E+04, 2.1976E+04, 2.1988E+04,
     +  2.1999E+04, 2.2010E+04, 2.2021E+04, 2.2033E+04, 2.2044E+04,
     + 2.2055E+04, 2.2066E+04, 2.2078E+04, 2.2089E+04, 2.2100E+04,
     +  2.2111E+04, 2.2122E+04, 2.2134E+04, 2.2145E+04, 2.2156E+04,
     + 2.2167E+04, 2.2179E+04, 2.2190E+04, 2.2201E+04, 2.2213E+04,
     +  2.2224E+04, 2.2235E+04, 2.2246E+04, 2.2258E+04, 2.2269E+04,
     + 2.2280E+04, 2.2291E+04, 2.2303E+04, 2.2314E+04, 2.2325E+04,
     +  2.2336E+04, 2.2348E+04, 2.2359E+04, 2.2370E+04, 2.2382E+04,
     + 2.2393E+04, 2.2404E+04, 2.2415E+04, 2.2427E+04, 2.2438E+04,
     +  2.2449E+04, 2.2461E+04, 2.2472E+04, 2.2483E+04, 2.2495E+04,
     + 2.2506E+04, 2.2517E+04, 2.2528E+04, 2.2540E+04, 2.2551E+04,
     +  2.2562E+04, 2.2574E+04, 2.2585E+04, 2.2596E+04, 2.2608E+04,
     + 2.2619E+04, 2.2630E+04, 2.2642E+04, 2.2653E+04, 2.2664E+04,
     +  2.2676E+04, 2.2687E+04, 2.2698E+04, 2.2710E+04, 2.2721E+04,
     + 2.2732E+04, 2.2743E+04, 2.2755E+04, 2.2766E+04, 2.2778E+04,
     +  2.2789E+04, 2.2800E+04, 2.2812E+04, 2.2823E+04, 2.2834E+04,
     + 2.2846E+04, 2.2857E+04, 2.2868E+04, 2.2880E+04, 2.2891E+04,
     +  2.2902E+04, 2.2914E+04, 2.2925E+04, 2.2936E+04, 2.2948E+04,
     + 2.2959E+04, 2.2970E+04, 2.2982E+04, 2.2993E+04, 2.3005E+04,
     +  2.3016E+04, 2.3027E+04, 2.3039E+04, 2.3050E+04, 2.3061E+04,
     + 2.3073E+04, 2.3084E+04, 2.3096E+04, 2.3107E+04, 2.3118E+04,
     +  2.3130E+04, 2.3141E+04, 2.3153E+04, 2.3164E+04, 2.3175E+04,
     + 2.3187E+04, 2.3198E+04, 2.3209E+04, 2.3221E+04, 2.3232E+04,
     +  2.3244E+04, 2.3255E+04, 2.3266E+04, 2.3278E+04, 2.3289E+04,
     + 2.3301E+04, 2.3312E+04, 2.3323E+04, 2.3335E+04, 2.3346E+04,
     +  2.3358E+04, 2.3369E+04, 2.3381E+04, 2.3392E+04, 2.3403E+04,
     + 2.3415E+04, 2.3426E+04, 2.3438E+04, 2.3449E+04, 2.3461E+04,
     +  2.3472E+04, 2.3483E+04, 2.3495E+04, 2.3506E+04, 2.3518E+04,
     + 2.3529E+04, 2.3541E+04, 2.3552E+04, 2.3563E+04, 2.3575E+04,
     +  2.3586E+04, 2.3598E+04, 2.3609E+04, 2.3621E+04, 2.3632E+04,
     + 2.3644E+04, 2.3655E+04, 2.3666E+04, 2.3678E+04, 2.3689E+04,
     +  2.3701E+04, 2.3712E+04, 2.3724E+04, 2.3735E+04, 2.3747E+04,
     + 2.3758E+04, 2.3770E+04, 2.3781E+04, 2.3793E+04, 2.3804E+04,
     +  2.3816E+04, 2.3827E+04, 2.3838E+04, 2.3850E+04, 2.3861E+04,
     + 2.3873E+04, 2.3884E+04, 2.3896E+04, 2.3907E+04, 2.3919E+04,
     +  2.3930E+04, 2.3942E+04, 2.3953E+04, 2.3965E+04, 2.3976E+04,
     + 2.3988E+04, 2.3999E+04, 2.4011E+04, 2.4022E+04, 2.4034E+04,
     +  2.4045E+04, 2.4057E+04, 2.4068E+04, 2.4080E+04, 2.4091E+04,
     + 2.4103E+04, 2.4114E+04, 2.4126E+04, 2.4137E+04, 2.4149E+04,
     +  2.4160E+04, 2.4172E+04, 2.4183E+04, 2.4195E+04, 2.4206E+04,
     + 2.4218E+04, 2.4230E+04, 2.4241E+04, 2.4253E+04, 2.4264E+04,
     +  2.4276E+04, 2.4287E+04, 2.4299E+04, 2.4310E+04, 2.4322E+04,
     + 2.4333E+04, 2.4345E+04, 2.4356E+04, 2.4368E+04, 2.4379E+04,
     +  2.4391E+04, 2.4403E+04, 2.4414E+04, 2.4426E+04, 2.4437E+04,
     + 2.4449E+04, 2.4460E+04, 2.4472E+04, 2.4483E+04, 2.4495E+04,
     +  2.4507E+04, 2.4518E+04, 2.4530E+04, 2.4541E+04, 2.4553E+04,
     + 2.4564E+04, 2.4576E+04, 2.4587E+04, 2.4599E+04, 2.4611E+04,
     +  2.4622E+04, 2.4634E+04, 2.4645E+04, 2.4657E+04, 2.4668E+04,
     + 2.4680E+04, 2.4692E+04, 2.4703E+04, 2.4715E+04, 2.4726E+04,
     +  2.4738E+04, 2.4750E+04, 2.4761E+04, 2.4773E+04, 2.4784E+04,
     + 2.4796E+04, 2.4807E+04, 2.4819E+04, 2.4831E+04, 2.4842E+04,
     +  2.4854E+04, 2.4865E+04, 2.4877E+04, 2.4889E+04, 2.4900E+04,
     + 2.4912E+04, 2.4924E+04, 2.4935E+04, 2.4947E+04, 2.4958E+04,
     +  2.4970E+04, 2.4982E+04, 2.4993E+04, 2.5005E+04, 2.5016E+04,
     + 2.5028E+04, 2.5040E+04, 2.5051E+04, 2.5063E+04, 2.5075E+04,
     +  2.5086E+04, 2.5098E+04, 2.5109E+04, 2.5121E+04, 2.5133E+04,
     + 2.5144E+04, 2.5156E+04, 2.5168E+04, 2.5179E+04, 2.5191E+04,
     +  2.5202E+04, 2.5214E+04, 2.5226E+04, 2.5237E+04, 2.5249E+04,
     + 2.5261E+04, 2.5272E+04, 2.5284E+04, 2.5296E+04, 2.5307E+04,
     +  2.5319E+04, 2.5331E+04, 2.5342E+04, 2.5354E+04, 2.5366E+04,
     + 2.5377E+04, 2.5389E+04, 2.5401E+04, 2.5412E+04, 2.5424E+04,
     +  2.5436E+04, 2.5447E+04, 2.5459E+04, 2.5471E+04, 2.5482E+04,
     + 2.5494E+04, 2.5506E+04, 2.5517E+04, 2.5529E+04, 2.5541E+04,
     +  2.5552E+04, 2.5564E+04, 2.5576E+04, 2.5587E+04, 2.5599E+04,
     + 2.5611E+04, 2.5622E+04, 2.5634E+04, 2.5646E+04, 2.5657E+04,
     +  2.5669E+04, 2.5681E+04, 2.5692E+04, 2.5704E+04, 2.5716E+04,
     + 2.5728E+04, 2.5739E+04, 2.5751E+04, 2.5763E+04, 2.5774E+04,
     +  2.5786E+04, 2.5798E+04, 2.5810E+04, 2.5821E+04, 2.5833E+04,
     + 2.5845E+04, 2.5856E+04, 2.5868E+04, 2.5880E+04, 2.5891E+04,
     +  2.5903E+04, 2.5915E+04, 2.5927E+04, 2.5938E+04, 2.5950E+04,
     + 2.5962E+04, 2.5974E+04, 2.5985E+04, 2.5997E+04, 2.6009E+04,
     +  2.6020E+04, 2.6032E+04, 2.6044E+04, 2.6056E+04, 2.6067E+04,
     + 2.6079E+04, 2.6091E+04, 2.6103E+04, 2.6114E+04, 2.6126E+04,
     +  2.6138E+04, 2.6150E+04, 2.6161E+04, 2.6173E+04, 2.6185E+04,
     + 2.6197E+04, 2.6208E+04, 2.6220E+04, 2.6232E+04, 2.6244E+04,
     +  2.6255E+04, 2.6267E+04, 2.6279E+04, 2.6291E+04, 2.6302E+04,
     + 2.6314E+04, 2.6326E+04, 2.6338E+04, 2.6349E+04, 2.6361E+04,
     +  2.6373E+04, 2.6385E+04, 2.6396E+04, 2.6408E+04, 2.6420E+04,
     + 2.6432E+04, 2.6444E+04, 2.6455E+04, 2.6467E+04, 2.6479E+04,
     +  2.6491E+04, 2.6502E+04, 2.6514E+04, 2.6526E+04, 2.6538E+04,
     + 2.6550E+04, 2.6561E+04, 2.6573E+04, 2.6585E+04, 2.6597E+04,
     +  2.6609E+04, 2.6620E+04, 2.6632E+04, 2.6644E+04, 2.6656E+04,
     + 2.6667E+04, 2.6679E+04, 2.6691E+04, 2.6703E+04, 2.6715E+04,
     +  2.6726E+04, 2.6738E+04, 2.6750E+04, 2.6762E+04, 2.6774E+04,
     + 2.6786E+04, 2.6797E+04, 2.6809E+04, 2.6821E+04, 2.6833E+04,
     +  2.6845E+04, 2.6856E+04, 2.6868E+04, 2.6880E+04, 2.6892E+04,
     + 2.6904E+04, 2.6916E+04, 2.6927E+04, 2.6939E+04, 2.6951E+04,
     +  2.6963E+04, 2.6975E+04, 2.6986E+04, 2.6998E+04, 2.7010E+04,
     + 2.7022E+04, 2.7034E+04, 2.7046E+04, 2.7057E+04, 2.7069E+04,
     +  2.7081E+04, 2.7093E+04, 2.7105E+04, 2.7117E+04, 2.7129E+04,
     + 2.7140E+04, 2.7152E+04, 2.7164E+04, 2.7176E+04, 2.7188E+04,
     +  2.7200E+04, 2.7211E+04, 2.7223E+04, 2.7235E+04, 2.7247E+04,
     + 2.7259E+04, 2.7271E+04, 2.7283E+04, 2.7294E+04, 2.7306E+04,
     +  2.7318E+04, 2.7330E+04, 2.7342E+04, 2.7354E+04, 2.7366E+04,
     + 2.7378E+04, 2.7389E+04, 2.7401E+04, 2.7413E+04, 2.7425E+04,
     +  2.7437E+04, 2.7449E+04, 2.7461E+04, 2.7473E+04, 2.7484E+04,
     + 2.7496E+04, 2.7508E+04, 2.7520E+04, 2.7532E+04, 2.7544E+04,
     +  2.7556E+04, 2.7568E+04, 2.7579E+04, 2.7591E+04, 2.7603E+04,
     + 2.7615E+04, 2.7627E+04, 2.7639E+04, 2.7651E+04, 2.7663E+04,
     +  2.7675E+04, 2.7687E+04, 2.7698E+04, 2.7710E+04, 2.7722E+04,
     + 2.7734E+04, 2.7746E+04, 2.7758E+04, 2.7770E+04, 2.7782E+04,
     +  2.7794E+04, 2.7806E+04, 2.7817E+04, 2.7829E+04, 2.7841E+04,
     + 2.7853E+04, 2.7865E+04, 2.7877E+04, 2.7889E+04, 2.7901E+04,
     +  2.7913E+04, 2.7925E+04, 2.7937E+04, 2.7949E+04, 2.7960E+04,
     + 2.7972E+04, 2.7984E+04, 2.7996E+04, 2.8008E+04, 2.8020E+04,
     +  2.8032E+04, 2.8044E+04, 2.8056E+04, 2.8068E+04, 2.8080E+04,
     + 2.8092E+04, 2.8104E+04, 2.8116E+04, 2.8128E+04, 2.8139E+04,
     +  2.8151E+04, 2.8163E+04, 2.8175E+04, 2.8187E+04, 2.8199E+04,
     + 2.8211E+04, 2.8223E+04, 2.8235E+04, 2.8247E+04, 2.8259E+04,
     +  2.8271E+04, 2.8283E+04, 2.8295E+04, 2.8307E+04, 2.8319E+04,
     + 2.8331E+04, 2.8343E+04, 2.8355E+04, 2.8367E+04, 2.8379E+04,
     +  2.8390E+04, 2.8402E+04, 2.8414E+04, 2.8426E+04, 2.8438E+04,
     + 2.8450E+04, 2.8462E+04, 2.8474E+04, 2.8486E+04, 2.8498E+04,
     +  2.8510E+04, 2.8522E+04, 2.8534E+04, 2.8546E+04, 2.8558E+04,
     + 2.8570E+04, 2.8582E+04, 2.8594E+04, 2.8606E+04, 2.8618E+04,
     +  2.8630E+04, 2.8642E+04, 2.8654E+04, 2.8666E+04, 2.8678E+04,
     + 2.8690E+04, 2.8702E+04, 2.8714E+04, 2.8726E+04, 2.8738E+04,
     +  2.8750E+04, 2.8762E+04, 2.8774E+04, 2.8786E+04, 2.8798E+04,
     + 2.8810E+04, 2.8822E+04, 2.8834E+04, 2.8846E+04, 2.8858E+04,
     +  2.8870E+04, 2.8882E+04, 2.8894E+04, 2.8906E+04, 2.8918E+04,
     + 2.8930E+04, 2.8942E+04, 2.8954E+04, 2.8966E+04, 2.8978E+04,
     +  2.8990E+04, 2.9002E+04, 2.9014E+04, 2.9026E+04, 2.9038E+04,
     + 2.9050E+04, 2.9062E+04, 2.9074E+04, 2.9086E+04, 2.9098E+04,
     +  2.9110E+04, 2.9122E+04, 2.9134E+04, 2.9146E+04, 2.9158E+04,
     + 2.9170E+04, 2.9182E+04, 2.9195E+04, 2.9207E+04, 2.9219E+04,
     +  2.9231E+04, 2.9243E+04, 2.9255E+04, 2.9267E+04, 2.9279E+04,
     + 2.9291E+04, 2.9303E+04, 2.9315E+04, 2.9327E+04, 2.9339E+04,
     +  2.9351E+04, 2.9363E+04, 2.9375E+04, 2.9387E+04, 2.9399E+04,
     + 2.9411E+04, 2.9423E+04, 2.9435E+04, 2.9447E+04, 2.9460E+04,
     +  2.9472E+04, 2.9484E+04, 2.9496E+04, 2.9508E+04, 2.9520E+04,
     + 2.9532E+04, 2.9544E+04, 2.9556E+04, 2.9568E+04, 2.9580E+04,
     +  2.9592E+04, 2.9604E+04, 2.9616E+04, 2.9628E+04, 2.9640E+04,
     + 2.9653E+04, 2.9665E+04, 2.9677E+04, 2.9689E+04, 2.9701E+04,
     +  2.9713E+04, 2.9725E+04, 2.9737E+04, 2.9749E+04, 2.9761E+04,
     + 2.9773E+04, 2.9785E+04, 2.9798E+04, 2.9810E+04, 2.9822E+04,
     +  2.9834E+04, 2.9846E+04, 2.9858E+04, 2.9870E+04, 2.9882E+04,
     + 2.9894E+04, 2.9906E+04, 2.9918E+04, 2.9931E+04, 2.9943E+04,
     +  2.9955E+04, 2.9967E+04, 2.9979E+04, 2.9991E+04, 3.0003E+04,
     + 3.0015E+04, 3.0027E+04, 3.0039E+04, 3.0052E+04, 3.0064E+04,
     +  3.0076E+04, 3.0088E+04, 3.0100E+04, 3.0112E+04, 3.0124E+04,
     + 3.0136E+04, 3.0148E+04, 3.0161E+04, 3.0173E+04, 3.0185E+04,
     +  3.0197E+04, 3.0209E+04, 3.0221E+04, 3.0233E+04, 3.0245E+04,
     + 3.0257E+04, 3.0270E+04, 3.0282E+04, 3.0294E+04, 3.0306E+04,
     +  3.0318E+04, 3.0330E+04, 3.0342E+04, 3.0354E+04, 3.0367E+04,
     + 3.0379E+04, 3.0391E+04, 3.0403E+04, 3.0415E+04, 3.0427E+04,
     +  3.0439E+04, 3.0452E+04, 3.0464E+04, 3.0476E+04, 3.0488E+04,
     + 3.0500E+04, 3.0512E+04, 3.0524E+04, 3.0537E+04, 3.0549E+04,
     +  3.0561E+04, 3.0573E+04, 3.0585E+04, 3.0597E+04, 3.0609E+04,
     + 3.0622E+04, 3.0634E+04, 3.0646E+04, 3.0658E+04, 3.0670E+04,
     +  3.0682E+04, 3.0695E+04, 3.0707E+04, 3.0719E+04, 3.0731E+04,
     + 3.0743E+04, 3.0755E+04, 3.0767E+04, 3.0780E+04, 3.0792E+04,
     +  3.0804E+04, 3.0816E+04, 3.0828E+04, 3.0840E+04, 3.0853E+04,
     + 3.0865E+04, 3.0877E+04, 3.0889E+04, 3.0901E+04, 3.0913E+04,
     +  3.0926E+04, 3.0938E+04, 3.0950E+04, 3.0962E+04, 3.0974E+04,
     + 3.0987E+04, 3.0999E+04, 3.1011E+04, 3.1023E+04, 3.1035E+04,
     +  3.1047E+04, 3.1060E+04, 3.1072E+04, 3.1084E+04, 3.1096E+04,
     + 3.1108E+04, 3.1121E+04, 3.1133E+04, 3.1145E+04, 3.1157E+04,
     +  3.1169E+04, 3.1181E+04, 3.1194E+04, 3.1206E+04, 3.1218E+04,
     + 3.1230E+04, 3.1242E+04, 3.1255E+04, 3.1267E+04, 3.1279E+04,
     +  3.1291E+04, 3.1303E+04, 3.1316E+04, 3.1328E+04, 3.1340E+04,
     + 3.1352E+04, 3.1364E+04, 3.1377E+04, 3.1389E+04, 3.1401E+04,
     +  3.1413E+04, 3.1426E+04, 3.1438E+04, 3.1450E+04, 3.1462E+04,
     + 3.1474E+04, 3.1487E+04, 3.1499E+04, 3.1511E+04, 3.1523E+04,
     +  3.1535E+04, 3.1548E+04, 3.1560E+04, 3.1572E+04, 3.1584E+04,
     + 3.1597E+04, 3.1609E+04, 3.1621E+04, 3.1633E+04, 3.1645E+04,
     +  3.1658E+04, 3.1670E+04, 3.1682E+04, 3.1694E+04, 3.1707E+04,
     + 3.1719E+04, 3.1731E+04, 3.1743E+04, 3.1756E+04, 3.1768E+04,
     +  3.1780E+04, 3.1792E+04, 3.1804E+04, 3.1817E+04, 3.1829E+04,
     + 3.1841E+04, 3.1853E+04, 3.1866E+04, 3.1878E+04, 3.1890E+04,
     +  3.1902E+04, 3.1915E+04, 3.1927E+04, 3.1939E+04, 3.1951E+04,
     + 3.1964E+04, 3.1976E+04, 3.1988E+04, 3.2000E+04, 3.2013E+04,
     +  3.2025E+04, 3.2037E+04, 3.2049E+04, 3.2062E+04, 3.2074E+04,
     + 3.2086E+04, 3.2098E+04, 3.2111E+04, 3.2123E+04, 3.2135E+04,
     +  3.2147E+04, 3.2160E+04, 3.2172E+04, 3.2184E+04, 3.2197E+04,
     + 3.2209E+04, 3.2221E+04, 3.2233E+04, 3.2246E+04, 3.2258E+04,
     +  3.2270E+04, 3.2282E+04, 3.2295E+04, 3.2307E+04, 3.2319E+04,
     + 3.2331E+04, 3.2344E+04, 3.2356E+04, 3.2368E+04, 3.2381E+04,
     +  3.2393E+04, 3.2405E+04, 3.2417E+04, 3.2430E+04, 3.2442E+04,
     + 3.2454E+04, 3.2467E+04, 3.2479E+04, 3.2491E+04, 3.2503E+04,
     +  3.2516E+04, 3.2528E+04, 3.2540E+04, 3.2553E+04, 3.2565E+04,
     + 3.2577E+04, 3.2589E+04, 3.2602E+04, 3.2614E+04, 3.2626E+04,
     +  3.2639E+04, 3.2651E+04, 3.2663E+04, 3.2675E+04, 3.2688E+04,
     + 3.2700E+04, 3.2712E+04, 3.2725E+04, 3.2737E+04, 3.2749E+04,
     +  3.2762E+04, 3.2774E+04, 3.2786E+04, 3.2798E+04, 3.2811E+04,
     + 3.2823E+04, 3.2835E+04, 3.2848E+04, 3.2860E+04, 3.2872E+04,
     +  3.2885E+04, 3.2897E+04, 3.2909E+04, 3.2922E+04, 3.2934E+04,
     + 3.2946E+04, 3.2959E+04, 3.2971E+04, 3.2983E+04, 3.2995E+04,
     +  3.3008E+04, 3.3020E+04, 3.3032E+04, 3.3045E+04, 3.3057E+04,
     + 3.3069E+04, 3.3082E+04, 3.3094E+04, 3.3106E+04, 3.3119E+04,
     +  3.3131E+04, 3.3143E+04, 3.3156E+04, 3.3168E+04, 3.3180E+04,
     + 3.3193E+04, 3.3205E+04, 3.3217E+04, 3.3230E+04, 3.3242E+04,
     +  3.3254E+04, 3.3267E+04, 3.3279E+04, 3.3291E+04, 3.3304E+04,
     + 3.3316E+04, 3.3328E+04, 3.3341E+04, 3.3353E+04, 3.3365E+04,
     +  3.3378E+04, 3.3390E+04, 3.3402E+04, 3.3415E+04, 3.3427E+04,
     + 3.3439E+04, 3.3452E+04, 3.3464E+04, 3.3476E+04, 3.3489E+04,
     +  3.3501E+04, 3.3514E+04, 3.3526E+04, 3.3538E+04, 3.3551E+04,
     + 3.3563E+04, 3.3575E+04, 3.3588E+04, 3.3600E+04, 3.3612E+04,
     +  3.3625E+04, 3.3637E+04, 3.3649E+04, 3.3662E+04, 3.3674E+04,
     + 3.3686E+04, 3.3699E+04, 3.3711E+04, 3.3724E+04, 3.3736E+04,
     +  3.3748E+04, 3.3761E+04, 3.3773E+04, 3.3785E+04, 3.3798E+04,
     + 3.3810E+04, 3.3823E+04, 3.3835E+04, 3.3847E+04, 3.3860E+04,
     +  3.3872E+04, 3.3884E+04, 3.3897E+04, 3.3909E+04, 3.3921E+04,
     + 3.3934E+04, 3.3946E+04, 3.3959E+04, 3.3971E+04, 3.3983E+04,
     +  3.3996E+04, 3.4008E+04, 3.4021E+04, 3.4033E+04, 3.4045E+04,
     + 3.4058E+04, 3.4070E+04, 3.4082E+04, 3.4095E+04, 3.4107E+04,
     +  3.4120E+04, 3.4132E+04, 3.4144E+04, 3.4157E+04, 3.4169E+04,
     + 3.4182E+04, 3.4194E+04, 3.4206E+04, 3.4219E+04, 3.4231E+04,
     +  3.4243E+04, 3.4256E+04, 3.4268E+04, 3.4281E+04, 3.4293E+04,
     + 3.4305E+04, 3.4318E+04, 3.4330E+04, 3.4343E+04, 3.4355E+04,
     +  3.4367E+04, 3.4380E+04, 3.4392E+04, 3.4405E+04, 3.4417E+04,
     + 3.4429E+04, 3.4442E+04, 3.4454E+04, 3.4467E+04, 3.4479E+04,
     +  3.4492E+04, 3.4504E+04, 3.4516E+04, 3.4529E+04, 3.4541E+04,
     + 3.4554E+04, 3.4566E+04, 3.4578E+04, 3.4591E+04, 3.4603E+04,
     +  3.4616E+04, 3.4628E+04, 3.4640E+04, 3.4653E+04, 3.4665E+04,
     + 3.4678E+04, 3.4690E+04, 3.4703E+04, 3.4715E+04, 3.4727E+04,
     +  3.4740E+04, 3.4752E+04, 3.4765E+04, 3.4777E+04, 3.4790E+04,
     + 3.4802E+04, 3.4814E+04, 3.4827E+04, 3.4839E+04, 3.4852E+04,
     +  3.4864E+04, 3.4876E+04, 3.4889E+04, 3.4901E+04, 3.4914E+04,
     + 3.4926E+04, 3.4939E+04, 3.4951E+04, 3.4964E+04, 3.4976E+04,
     +  3.4988E+04, 3.5001E+04, 3.5013E+04, 3.5026E+04, 3.5038E+04,
     + 3.5051E+04, 3.5063E+04, 3.5075E+04, 3.5088E+04, 3.5100E+04,
     +  3.5113E+04, 3.5125E+04, 3.5138E+04, 3.5150E+04, 3.5163E+04,
     + 3.5175E+04, 3.5187E+04, 3.5200E+04, 3.5212E+04, 3.5225E+04,
     +  3.5237E+04, 3.5250E+04, 3.5262E+04, 3.5275E+04, 3.5287E+04,
     + 3.5300E+04, 3.5312E+04, 3.5324E+04, 3.5337E+04, 3.5349E+04,
     +  3.5362E+04, 3.5374E+04, 3.5387E+04, 3.5399E+04, 3.5412E+04,
     + 3.5424E+04, 3.5437E+04, 3.5449E+04, 3.5461E+04, 3.5474E+04,
     +  3.5486E+04, 3.5499E+04, 3.5511E+04, 3.5524E+04, 3.5536E+04,
     + 3.5549E+04, 3.5561E+04, 3.5574E+04, 3.5586E+04, 3.5599E+04,
     +  3.5611E+04, 3.5623E+04, 3.5636E+04, 3.5648E+04, 3.5661E+04,
     + 3.5673E+04, 3.5686E+04, 3.5698E+04, 3.5711E+04, 3.5723E+04,
     +  3.5736E+04, 3.5748E+04, 3.5761E+04, 3.5773E+04, 3.5786E+04,
     + 3.5798E+04, 3.5811E+04, 3.5823E+04, 3.5836E+04, 3.5848E+04,
     +  3.5861E+04, 3.5873E+04, 3.5886E+04, 3.5898E+04, 3.5910E+04,
     + 3.5923E+04, 3.5935E+04, 3.5948E+04, 3.5960E+04, 3.5973E+04,
     +  3.5985E+04, 3.5998E+04, 3.6010E+04, 3.6023E+04, 3.6035E+04,
     + 3.6048E+04, 3.6060E+04, 3.6073E+04, 3.6085E+04, 3.6098E+04,
     +  3.6110E+04, 3.6123E+04, 3.6135E+04, 3.6148E+04, 3.6160E+04,
     + 3.6173E+04, 3.6185E+04, 3.6198E+04, 3.6210E+04, 3.6223E+04,
     +  3.6235E+04, 3.6248E+04, 3.6260E+04, 3.6273E+04, 3.6285E+04,
     + 3.6298E+04, 3.6310E+04, 3.6323E+04, 3.6335E+04, 3.6348E+04,
     +  3.6360E+04, 3.6373E+04, 3.6385E+04, 3.6398E+04, 3.6410E+04,
     + 3.6423E+04, 3.6435E+04, 3.6448E+04, 3.6460E+04, 3.6473E+04,
     +  3.6485E+04, 3.6498E+04, 3.6510E+04, 3.6523E+04, 3.6535E+04,
     + 3.6548E+04, 3.6561E+04, 3.6573E+04, 3.6586E+04, 3.6598E+04,
     +  3.6611E+04, 3.6623E+04, 3.6636E+04, 3.6648E+04, 3.6661E+04,
     + 3.6673E+04, 3.6686E+04, 3.6698E+04, 3.6711E+04, 3.6723E+04,
     +  3.6736E+04, 3.6748E+04, 3.6761E+04, 3.6773E+04, 3.6786E+04,
     + 3.6798E+04, 3.6811E+04, 3.6823E+04, 3.6836E+04, 3.6849E+04,
     +  3.6861E+04, 3.6874E+04, 3.6886E+04, 3.6899E+04, 3.6911E+04,
     + 3.6924E+04, 3.6936E+04, 3.6949E+04, 3.6961E+04, 3.6974E+04,
     +  3.6986E+04, 3.6999E+04, 3.7011E+04, 3.7024E+04, 3.7037E+04,
     + 3.7049E+04, 3.7062E+04, 3.7074E+04, 3.7087E+04, 3.7099E+04,
     +  3.7112E+04, 3.7124E+04, 3.7137E+04, 3.7149E+04, 3.7162E+04,
     + 3.7175E+04, 3.7187E+04, 3.7200E+04, 3.7212E+04, 3.7225E+04,
     +  3.7237E+04, 3.7250E+04, 3.7262E+04, 3.7275E+04, 3.7287E+04,
     + 3.7300E+04, 3.7313E+04, 3.7325E+04, 3.7338E+04, 3.7350E+04,
     +  3.7363E+04, 3.7375E+04, 3.7388E+04, 3.7400E+04, 3.7413E+04,
     + 3.7426E+04, 3.7438E+04, 3.7451E+04, 3.7463E+04, 3.7476E+04,
     +  3.7488E+04, 3.7501E+04, 3.7513E+04, 3.7526E+04, 3.7539E+04,
     + 3.7551E+04, 3.7564E+04, 3.7576E+04, 3.7589E+04, 3.7601E+04,
     +  3.7614E+04, 3.7627E+04, 3.7639E+04, 3.7652E+04, 3.7664E+04,
     + 3.7677E+04, 3.7689E+04, 3.7702E+04, 3.7714E+04, 3.7727E+04,
     +  3.7740E+04, 3.7752E+04, 3.7765E+04, 3.7777E+04, 3.7790E+04,
     + 3.7802E+04, 3.7815E+04, 3.7828E+04, 3.7840E+04, 3.7853E+04,
     +  3.7865E+04, 3.7878E+04, 3.7891E+04, 3.7903E+04, 3.7916E+04,
     + 3.7928E+04, 3.7941E+04, 3.7953E+04, 3.7966E+04, 3.7979E+04,
     +  3.7991E+04, 3.8004E+04, 3.8016E+04, 3.8029E+04, 3.8041E+04,
     + 3.8054E+04, 3.8067E+04, 3.8079E+04, 3.8092E+04, 3.8104E+04,
     +  3.8117E+04, 3.8130E+04, 3.8142E+04, 3.8155E+04, 3.8167E+04,
     + 3.8180E+04, 3.8193E+04, 3.8205E+04, 3.8218E+04, 3.8230E+04,
     +  3.8243E+04, 3.8255E+04, 3.8268E+04, 3.8281E+04, 3.8293E+04,
     + 3.8306E+04, 3.8318E+04, 3.8331E+04, 3.8344E+04, 3.8356E+04,
     +  3.8369E+04, 3.8381E+04, 3.8394E+04, 3.8407E+04, 3.8419E+04,
     + 3.8432E+04, 3.8444E+04, 3.8457E+04, 3.8470E+04, 3.8482E+04,
     +  3.8495E+04, 3.8507E+04, 3.8520E+04, 3.8533E+04, 3.8545E+04,
     + 3.8558E+04, 3.8570E+04, 3.8583E+04, 3.8596E+04, 3.8608E+04,
     +  3.8621E+04, 3.8633E+04, 3.8646E+04, 3.8659E+04, 3.8671E+04,
     + 3.8684E+04, 3.8696E+04, 3.8709E+04, 3.8722E+04, 3.8734E+04,
     +  3.8747E+04, 3.8760E+04, 3.8772E+04, 3.8785E+04, 3.8797E+04,
     + 3.8810E+04, 3.8823E+04, 3.8835E+04, 3.8848E+04, 3.8860E+04,
     +  3.8873E+04, 3.8886E+04, 3.8898E+04, 3.8911E+04, 3.8924E+04,
     + 3.8936E+04, 3.8949E+04, 3.8961E+04, 3.8974E+04, 3.8987E+04,
     +  3.8999E+04, 3.9012E+04, 3.9025E+04, 3.9037E+04, 3.9050E+04,
     + 3.9062E+04, 3.9075E+04, 3.9088E+04, 3.9100E+04, 3.9113E+04,
     +  3.9126E+04, 3.9138E+04, 3.9151E+04, 3.9163E+04, 3.9176E+04,
     + 3.9189E+04, 3.9201E+04, 3.9214E+04, 3.9227E+04/
c...        --        45
      data (QofT( 2,J),J=1,2408)/ 6.0710E+00, 7.1380E+00, 8.9632E+00,
     +  1.0994E+01, 1.3089E+01, 1.5211E+01, 1.7347E+01, 1.9492E+01,
     + 2.1642E+01, 2.3797E+01, 2.5953E+01, 2.8112E+01, 3.0272E+01,
     +  3.2434E+01, 3.4596E+01, 3.6759E+01, 3.8923E+01, 4.1087E+01,
     + 4.3252E+01, 4.5417E+01, 4.9748E+01, 5.4079E+01, 5.8412E+01,
     +  6.2745E+01, 6.7079E+01, 7.1413E+01, 7.5747E+01, 8.0082E+01,
     + 8.4416E+01, 8.8751E+01, 9.3087E+01, 9.7422E+01, 1.0176E+02,
     +  1.0609E+02, 1.1043E+02, 1.1476E+02, 1.1910E+02, 1.2344E+02,
     + 1.2777E+02, 1.3211E+02, 1.3645E+02, 1.4078E+02, 1.4512E+02,
     +  1.4945E+02, 1.5379E+02, 1.5813E+02, 1.6246E+02, 1.6680E+02,
     + 1.7114E+02, 1.7547E+02, 1.7981E+02, 1.8415E+02, 1.8849E+02,
     +  1.9282E+02, 1.9716E+02, 2.0150E+02, 2.0583E+02, 2.1017E+02,
     + 2.1451E+02, 2.1885E+02, 2.2318E+02, 2.2752E+02, 2.3186E+02,
     +  2.3620E+02, 2.4053E+02, 2.4487E+02, 2.4921E+02, 2.5355E+02,
     + 2.5788E+02, 2.6222E+02, 2.6656E+02, 2.7090E+02, 2.7523E+02,
     +  2.7957E+02, 2.8391E+02, 2.8825E+02, 2.9259E+02, 2.9692E+02,
     + 3.0126E+02, 3.0560E+02, 3.0994E+02, 3.1428E+02, 3.1862E+02,
     +  3.2295E+02, 3.2729E+02, 3.3163E+02, 3.3597E+02, 3.4031E+02,
     + 3.4465E+02, 3.4898E+02, 3.5332E+02, 3.5766E+02, 3.6200E+02,
     +  3.6634E+02, 3.7068E+02, 3.7502E+02, 3.7936E+02, 3.8369E+02,
     + 3.8803E+02, 3.9237E+02, 3.9671E+02, 4.0105E+02, 4.0539E+02,
     +  4.0973E+02, 4.1407E+02, 4.1841E+02, 4.2274E+02, 4.2708E+02,
     + 4.3142E+02, 4.3576E+02, 4.4010E+02, 4.4444E+02, 4.4878E+02,
     +  4.5312E+02, 4.5746E+02, 4.6180E+02, 4.6614E+02, 4.7048E+02,
     + 4.7482E+02, 4.7916E+02, 4.8350E+02, 4.8784E+02, 4.9218E+02,
     +  4.9652E+02, 5.0086E+02, 5.0520E+02, 5.0954E+02, 5.1388E+02,
     + 5.1822E+02, 5.2256E+02, 5.2690E+02, 5.3124E+02, 5.3558E+02,
     +  5.3992E+02, 5.4426E+02, 5.4860E+02, 5.5294E+02, 5.5728E+02,
     + 5.6162E+02, 5.6596E+02, 5.7030E+02, 5.7464E+02, 5.7898E+02,
     +  5.8332E+02, 5.8766E+02, 5.9200E+02, 5.9634E+02, 6.0068E+02,
     + 6.0503E+02, 6.0937E+02, 6.1371E+02, 6.1805E+02, 6.2239E+02,
     +  6.2673E+02, 6.3107E+02, 6.3542E+02, 6.3976E+02, 6.4410E+02,
     + 6.4844E+02, 6.5278E+02, 6.5712E+02, 6.6147E+02, 6.6581E+02,
     +  6.7015E+02, 6.7449E+02, 6.7884E+02, 6.8318E+02, 6.8752E+02,
     + 6.9186E+02, 6.9621E+02, 7.0055E+02, 7.0489E+02, 7.0924E+02,
     +  7.1358E+02, 7.1792E+02, 7.2227E+02, 7.2661E+02, 7.3096E+02,
     + 7.3530E+02, 7.3965E+02, 7.4399E+02, 7.4833E+02, 7.5268E+02,
     +  7.5703E+02, 7.6137E+02, 7.6572E+02, 7.7006E+02, 7.7441E+02,
     + 7.7876E+02, 7.8310E+02, 7.8745E+02, 7.9180E+02, 7.9614E+02,
     +  8.0049E+02, 8.0484E+02, 8.0919E+02, 8.1354E+02, 8.1789E+02,
     + 8.2223E+02, 8.2658E+02, 8.3093E+02, 8.3528E+02, 8.3964E+02,
     +  8.4399E+02, 8.4834E+02, 8.5269E+02, 8.5704E+02, 8.6139E+02,
     + 8.6575E+02, 8.7010E+02, 8.7445E+02, 8.7881E+02, 8.8316E+02,
     +  8.8752E+02, 8.9187E+02, 8.9623E+02, 9.0059E+02, 9.0494E+02,
     + 9.0930E+02, 9.1366E+02, 9.1802E+02, 9.2238E+02, 9.2674E+02,
     +  9.3110E+02, 9.3546E+02, 9.3982E+02, 9.4418E+02, 9.4854E+02,
     + 9.5291E+02, 9.5727E+02, 9.6164E+02, 9.6600E+02, 9.7037E+02,
     +  9.7473E+02, 9.7910E+02, 9.8347E+02, 9.8784E+02, 9.9221E+02,
     + 9.9658E+02, 1.0009E+03, 1.0053E+03, 1.0097E+03, 1.0141E+03,
     +  1.0184E+03, 1.0228E+03, 1.0272E+03, 1.0316E+03, 1.0359E+03,
     + 1.0403E+03, 1.0447E+03, 1.0491E+03, 1.0535E+03, 1.0579E+03,
     +  1.0622E+03, 1.0666E+03, 1.0710E+03, 1.0754E+03, 1.0798E+03,
     + 1.0842E+03, 1.0886E+03, 1.0929E+03, 1.0973E+03, 1.1017E+03,
     +  1.1061E+03, 1.1105E+03, 1.1149E+03, 1.1193E+03, 1.1237E+03,
     + 1.1281E+03, 1.1325E+03, 1.1369E+03, 1.1413E+03, 1.1457E+03,
     +  1.1501E+03, 1.1545E+03, 1.1590E+03, 1.1634E+03, 1.1678E+03,
     + 1.1722E+03, 1.1766E+03, 1.1810E+03, 1.1854E+03, 1.1899E+03,
     +  1.1943E+03, 1.1987E+03, 1.2031E+03, 1.2075E+03, 1.2120E+03,
     + 1.2164E+03, 1.2208E+03, 1.2253E+03, 1.2297E+03, 1.2341E+03,
     +  1.2386E+03, 1.2430E+03, 1.2474E+03, 1.2519E+03, 1.2563E+03,
     + 1.2608E+03, 1.2652E+03, 1.2697E+03, 1.2741E+03, 1.2786E+03,
     +  1.2830E+03, 1.2875E+03, 1.2919E+03, 1.2964E+03, 1.3008E+03,
     + 1.3053E+03, 1.3098E+03, 1.3142E+03, 1.3187E+03, 1.3232E+03,
     +  1.3276E+03, 1.3321E+03, 1.3366E+03, 1.3411E+03, 1.3455E+03,
     + 1.3500E+03, 1.3545E+03, 1.3590E+03, 1.3635E+03, 1.3680E+03,
     +  1.3725E+03, 1.3769E+03, 1.3814E+03, 1.3859E+03, 1.3904E+03,
     + 1.3949E+03, 1.3994E+03, 1.4040E+03, 1.4085E+03, 1.4130E+03,
     +  1.4175E+03, 1.4220E+03, 1.4265E+03, 1.4310E+03, 1.4356E+03,
     + 1.4401E+03, 1.4446E+03, 1.4491E+03, 1.4537E+03, 1.4582E+03,
     +  1.4627E+03, 1.4673E+03, 1.4718E+03, 1.4764E+03, 1.4809E+03,
     + 1.4855E+03, 1.4900E+03, 1.4946E+03, 1.4991E+03, 1.5037E+03,
     +  1.5082E+03, 1.5128E+03, 1.5174E+03, 1.5219E+03, 1.5265E+03,
     + 1.5311E+03, 1.5357E+03, 1.5402E+03, 1.5448E+03, 1.5494E+03,
     +  1.5540E+03, 1.5586E+03, 1.5632E+03, 1.5678E+03, 1.5724E+03,
     + 1.5770E+03, 1.5816E+03, 1.5862E+03, 1.5908E+03, 1.5954E+03,
     +  1.6000E+03, 1.6046E+03, 1.6092E+03, 1.6139E+03, 1.6185E+03,
     + 1.6231E+03, 1.6278E+03, 1.6324E+03, 1.6370E+03, 1.6417E+03,
     +  1.6463E+03, 1.6510E+03, 1.6556E+03, 1.6603E+03, 1.6649E+03,
     + 1.6696E+03, 1.6742E+03, 1.6789E+03, 1.6836E+03, 1.6882E+03,
     +  1.6929E+03, 1.6976E+03, 1.7023E+03, 1.7069E+03, 1.7116E+03,
     + 1.7163E+03, 1.7210E+03, 1.7257E+03, 1.7304E+03, 1.7351E+03,
     +  1.7398E+03, 1.7445E+03, 1.7492E+03, 1.7540E+03, 1.7587E+03,
     + 1.7634E+03, 1.7681E+03, 1.7728E+03, 1.7776E+03, 1.7823E+03,
     +  1.7870E+03, 1.7918E+03, 1.7965E+03, 1.8013E+03, 1.8060E+03,
     + 1.8108E+03, 1.8155E+03, 1.8203E+03, 1.8251E+03, 1.8298E+03,
     +  1.8346E+03, 1.8394E+03, 1.8442E+03, 1.8489E+03, 1.8537E+03,
     + 1.8585E+03, 1.8633E+03, 1.8681E+03, 1.8729E+03, 1.8777E+03,
     +  1.8825E+03, 1.8873E+03, 1.8921E+03, 1.8969E+03, 1.9018E+03,
     + 1.9066E+03, 1.9114E+03, 1.9162E+03, 1.9211E+03, 1.9259E+03,
     +  1.9308E+03, 1.9356E+03, 1.9405E+03, 1.9453E+03, 1.9502E+03,
     + 1.9550E+03, 1.9599E+03, 1.9648E+03, 1.9696E+03, 1.9745E+03,
     +  1.9794E+03, 1.9843E+03, 1.9891E+03, 1.9940E+03, 1.9989E+03,
     + 2.0038E+03, 2.0087E+03, 2.0136E+03, 2.0185E+03, 2.0235E+03,
     +  2.0284E+03, 2.0333E+03, 2.0382E+03, 2.0431E+03, 2.0481E+03,
     + 2.0530E+03, 2.0579E+03, 2.0629E+03, 2.0678E+03, 2.0728E+03,
     +  2.0777E+03, 2.0827E+03, 2.0877E+03, 2.0926E+03, 2.0976E+03,
     + 2.1026E+03, 2.1076E+03, 2.1125E+03, 2.1175E+03, 2.1225E+03,
     +  2.1275E+03, 2.1325E+03, 2.1375E+03, 2.1425E+03, 2.1475E+03,
     + 2.1525E+03, 2.1576E+03, 2.1626E+03, 2.1676E+03, 2.1726E+03,
     +  2.1777E+03, 2.1827E+03, 2.1877E+03, 2.1928E+03, 2.1978E+03,
     + 2.2029E+03, 2.2080E+03, 2.2130E+03, 2.2181E+03, 2.2231E+03,
     +  2.2282E+03, 2.2333E+03, 2.2384E+03, 2.2435E+03, 2.2486E+03,
     + 2.2537E+03, 2.2588E+03, 2.2639E+03, 2.2690E+03, 2.2741E+03,
     +  2.2792E+03, 2.2843E+03, 2.2894E+03, 2.2946E+03, 2.2997E+03,
     + 2.3048E+03, 2.3100E+03, 2.3151E+03, 2.3203E+03, 2.3254E+03,
     +  2.3306E+03, 2.3358E+03, 2.3409E+03, 2.3461E+03, 2.3513E+03,
     + 2.3564E+03, 2.3616E+03, 2.3668E+03, 2.3720E+03, 2.3772E+03,
     +  2.3824E+03, 2.3876E+03, 2.3928E+03, 2.3980E+03, 2.4033E+03,
     + 2.4085E+03, 2.4137E+03, 2.4189E+03, 2.4242E+03, 2.4294E+03,
     +  2.4347E+03, 2.4399E+03, 2.4452E+03, 2.4504E+03, 2.4557E+03,
     + 2.4610E+03, 2.4662E+03, 2.4715E+03, 2.4768E+03, 2.4821E+03,
     +  2.4874E+03, 2.4926E+03, 2.4979E+03, 2.5032E+03, 2.5085E+03,
     + 2.5139E+03, 2.5192E+03, 2.5245E+03, 2.5298E+03, 2.5351E+03,
     +  2.5405E+03, 2.5458E+03, 2.5512E+03, 2.5565E+03, 2.5619E+03,
     + 2.5672E+03, 2.5726E+03, 2.5779E+03, 2.5833E+03, 2.5887E+03,
     +  2.5941E+03, 2.5994E+03, 2.6048E+03, 2.6102E+03, 2.6156E+03,
     + 2.6210E+03, 2.6264E+03, 2.6318E+03, 2.6372E+03, 2.6427E+03,
     +  2.6481E+03, 2.6535E+03, 2.6589E+03, 2.6644E+03, 2.6698E+03,
     + 2.6753E+03, 2.6807E+03, 2.6862E+03, 2.6916E+03, 2.6971E+03,
     +  2.7026E+03, 2.7080E+03, 2.7135E+03, 2.7190E+03, 2.7245E+03,
     + 2.7300E+03, 2.7355E+03, 2.7410E+03, 2.7465E+03, 2.7520E+03,
     +  2.7575E+03, 2.7630E+03, 2.7686E+03, 2.7741E+03, 2.7796E+03,
     + 2.7852E+03, 2.7907E+03, 2.7963E+03, 2.8018E+03, 2.8074E+03,
     +  2.8129E+03, 2.8185E+03, 2.8241E+03, 2.8296E+03, 2.8352E+03,
     + 2.8408E+03, 2.8464E+03, 2.8520E+03, 2.8576E+03, 2.8632E+03,
     +  2.8688E+03, 2.8744E+03, 2.8801E+03, 2.8857E+03, 2.8913E+03,
     + 2.8969E+03, 2.9026E+03, 2.9082E+03, 2.9139E+03, 2.9195E+03,
     +  2.9252E+03, 2.9308E+03, 2.9365E+03, 2.9422E+03, 2.9478E+03,
     + 2.9535E+03, 2.9592E+03, 2.9649E+03, 2.9706E+03, 2.9763E+03,
     +  2.9820E+03, 2.9877E+03, 2.9934E+03, 2.9991E+03, 3.0049E+03,
     + 3.0106E+03, 3.0163E+03, 3.0221E+03, 3.0278E+03, 3.0336E+03,
     +  3.0393E+03, 3.0451E+03, 3.0508E+03, 3.0566E+03, 3.0624E+03,
     + 3.0682E+03, 3.0739E+03, 3.0797E+03, 3.0855E+03, 3.0913E+03,
     +  3.0971E+03, 3.1029E+03, 3.1087E+03, 3.1145E+03, 3.1204E+03,
     + 3.1262E+03, 3.1320E+03, 3.1379E+03, 3.1437E+03, 3.1495E+03,
     +  3.1554E+03, 3.1612E+03, 3.1671E+03, 3.1730E+03, 3.1788E+03,
     + 3.1847E+03, 3.1906E+03, 3.1965E+03, 3.2024E+03, 3.2083E+03,
     +  3.2142E+03, 3.2201E+03, 3.2260E+03, 3.2319E+03, 3.2378E+03,
     + 3.2437E+03, 3.2497E+03, 3.2556E+03, 3.2615E+03, 3.2675E+03,
     +  3.2734E+03, 3.2794E+03, 3.2853E+03, 3.2913E+03, 3.2973E+03,
     + 3.3032E+03, 3.3092E+03, 3.3152E+03, 3.3212E+03, 3.3272E+03,
     +  3.3332E+03, 3.3392E+03, 3.3452E+03, 3.3512E+03, 3.3572E+03,
     + 3.3632E+03, 3.3693E+03, 3.3753E+03, 3.3813E+03, 3.3874E+03,
     +  3.3934E+03, 3.3995E+03, 3.4055E+03, 3.4116E+03, 3.4177E+03,
     + 3.4237E+03, 3.4298E+03, 3.4359E+03, 3.4420E+03, 3.4481E+03,
     +  3.4542E+03, 3.4603E+03, 3.4664E+03, 3.4725E+03, 3.4786E+03,
     + 3.4847E+03, 3.4909E+03, 3.4970E+03, 3.5031E+03, 3.5093E+03,
     +  3.5154E+03, 3.5216E+03, 3.5277E+03, 3.5339E+03, 3.5400E+03,
     + 3.5462E+03, 3.5524E+03, 3.5586E+03, 3.5648E+03, 3.5710E+03,
     +  3.5772E+03, 3.5834E+03, 3.5896E+03, 3.5958E+03, 3.6020E+03,
     + 3.6082E+03, 3.6145E+03, 3.6207E+03, 3.6269E+03, 3.6332E+03,
     +  3.6394E+03, 3.6457E+03, 3.6519E+03, 3.6582E+03, 3.6645E+03,
     + 3.6707E+03, 3.6770E+03, 3.6833E+03, 3.6896E+03, 3.6959E+03,
     +  3.7022E+03, 3.7085E+03, 3.7148E+03, 3.7211E+03, 3.7274E+03,
     + 3.7337E+03, 3.7401E+03, 3.7464E+03, 3.7527E+03, 3.7591E+03,
     +  3.7654E+03, 3.7718E+03, 3.7782E+03, 3.7845E+03, 3.7909E+03,
     + 3.7973E+03, 3.8036E+03, 3.8100E+03, 3.8164E+03, 3.8228E+03,
     +  3.8292E+03, 3.8356E+03, 3.8420E+03, 3.8484E+03, 3.8549E+03,
     + 3.8613E+03, 3.8677E+03, 3.8742E+03, 3.8806E+03, 3.8870E+03,
     +  3.8935E+03, 3.9000E+03, 3.9064E+03, 3.9129E+03, 3.9194E+03,
     + 3.9258E+03, 3.9323E+03, 3.9388E+03, 3.9453E+03, 3.9518E+03,
     +  3.9583E+03, 3.9648E+03, 3.9713E+03, 3.9778E+03, 3.9844E+03,
     + 3.9909E+03, 3.9974E+03, 4.0040E+03, 4.0105E+03, 4.0171E+03,
     +  4.0236E+03, 4.0302E+03, 4.0367E+03, 4.0433E+03, 4.0499E+03,
     + 4.0565E+03, 4.0630E+03, 4.0696E+03, 4.0762E+03, 4.0828E+03,
     +  4.0894E+03, 4.0960E+03, 4.1027E+03, 4.1093E+03, 4.1159E+03,
     + 4.1225E+03, 4.1292E+03, 4.1358E+03, 4.1425E+03, 4.1491E+03,
     +  4.1558E+03, 4.1624E+03, 4.1691E+03, 4.1758E+03, 4.1825E+03,
     + 4.1891E+03, 4.1958E+03, 4.2025E+03, 4.2092E+03, 4.2159E+03,
     +  4.2226E+03, 4.2293E+03, 4.2361E+03, 4.2428E+03, 4.2495E+03,
     + 4.2562E+03, 4.2630E+03, 4.2697E+03, 4.2765E+03, 4.2832E+03,
     +  4.2900E+03, 4.2968E+03, 4.3035E+03, 4.3103E+03, 4.3171E+03,
     + 4.3239E+03, 4.3307E+03, 4.3375E+03, 4.3443E+03, 4.3511E+03,
     +  4.3579E+03, 4.3647E+03, 4.3715E+03, 4.3784E+03, 4.3852E+03,
     + 4.3920E+03, 4.3989E+03, 4.4057E+03, 4.4126E+03, 4.4194E+03,
     +  4.4263E+03, 4.4332E+03, 4.4400E+03, 4.4469E+03, 4.4538E+03,
     + 4.4607E+03, 4.4676E+03, 4.4745E+03, 4.4814E+03, 4.4883E+03,
     +  4.4952E+03, 4.5022E+03, 4.5091E+03, 4.5160E+03, 4.5230E+03,
     + 4.5299E+03, 4.5368E+03, 4.5438E+03, 4.5508E+03, 4.5577E+03,
     +  4.5647E+03, 4.5717E+03, 4.5786E+03, 4.5856E+03, 4.5926E+03,
     + 4.5996E+03, 4.6066E+03, 4.6136E+03, 4.6206E+03, 4.6276E+03,
     +  4.6347E+03, 4.6417E+03, 4.6487E+03, 4.6558E+03, 4.6628E+03,
     + 4.6698E+03, 4.6769E+03, 4.6839E+03, 4.6910E+03, 4.6981E+03,
     +  4.7052E+03, 4.7122E+03, 4.7193E+03, 4.7264E+03, 4.7335E+03,
     + 4.7406E+03, 4.7477E+03, 4.7548E+03, 4.7619E+03, 4.7690E+03,
     +  4.7762E+03, 4.7833E+03, 4.7904E+03, 4.7976E+03, 4.8047E+03,
     + 4.8119E+03, 4.8190E+03, 4.8262E+03, 4.8334E+03, 4.8405E+03,
     +  4.8477E+03, 4.8549E+03, 4.8621E+03, 4.8693E+03, 4.8765E+03,
     + 4.8837E+03, 4.8909E+03, 4.8981E+03, 4.9053E+03, 4.9125E+03,
     +  4.9198E+03, 4.9270E+03, 4.9342E+03, 4.9415E+03, 4.9487E+03,
     + 4.9560E+03, 4.9633E+03, 4.9705E+03, 4.9778E+03, 4.9851E+03,
     +  4.9923E+03, 4.9996E+03, 5.0069E+03, 5.0142E+03, 5.0215E+03,
     + 5.0288E+03, 5.0361E+03, 5.0435E+03, 5.0508E+03, 5.0581E+03,
     +  5.0655E+03, 5.0728E+03, 5.0801E+03, 5.0875E+03, 5.0948E+03,
     + 5.1022E+03, 5.1096E+03, 5.1169E+03, 5.1243E+03, 5.1317E+03,
     +  5.1391E+03, 5.1465E+03, 5.1539E+03, 5.1613E+03, 5.1687E+03,
     + 5.1761E+03, 5.1835E+03, 5.1909E+03, 5.1984E+03, 5.2058E+03,
     +  5.2132E+03, 5.2207E+03, 5.2281E+03, 5.2356E+03, 5.2430E+03,
     + 5.2505E+03, 5.2580E+03, 5.2655E+03, 5.2729E+03, 5.2804E+03,
     +  5.2879E+03, 5.2954E+03, 5.3029E+03, 5.3104E+03, 5.3179E+03,
     + 5.3255E+03, 5.3330E+03, 5.3405E+03, 5.3480E+03, 5.3556E+03,
     +  5.3631E+03, 5.3707E+03, 5.3782E+03, 5.3858E+03, 5.3933E+03,
     + 5.4009E+03, 5.4085E+03, 5.4161E+03, 5.4237E+03, 5.4312E+03,
     +  5.4388E+03, 5.4464E+03, 5.4541E+03, 5.4617E+03, 5.4693E+03,
     + 5.4769E+03, 5.4845E+03, 5.4922E+03, 5.4998E+03, 5.5074E+03,
     +  5.5151E+03, 5.5227E+03, 5.5304E+03, 5.5381E+03, 5.5457E+03,
     + 5.5534E+03, 5.5611E+03, 5.5688E+03, 5.5765E+03, 5.5842E+03,
     +  5.5919E+03, 5.5996E+03, 5.6073E+03, 5.6150E+03, 5.6227E+03,
     + 5.6304E+03, 5.6382E+03, 5.6459E+03, 5.6537E+03, 5.6614E+03,
     +  5.6692E+03, 5.6769E+03, 5.6847E+03, 5.6925E+03, 5.7002E+03,
     + 5.7080E+03, 5.7158E+03, 5.7236E+03, 5.7314E+03, 5.7392E+03,
     +  5.7470E+03, 5.7548E+03, 5.7626E+03, 5.7704E+03, 5.7783E+03,
     + 5.7861E+03, 5.7939E+03, 5.8018E+03, 5.8096E+03, 5.8175E+03,
     +  5.8253E+03, 5.8332E+03, 5.8411E+03, 5.8489E+03, 5.8568E+03,
     + 5.8647E+03, 5.8726E+03, 5.8805E+03, 5.8884E+03, 5.8963E+03,
     +  5.9042E+03, 5.9121E+03, 5.9200E+03, 5.9279E+03, 5.9359E+03,
     + 5.9438E+03, 5.9518E+03, 5.9597E+03, 5.9677E+03, 5.9756E+03,
     +  5.9836E+03, 5.9915E+03, 5.9995E+03, 6.0075E+03, 6.0155E+03,
     + 6.0235E+03, 6.0315E+03, 6.0395E+03, 6.0475E+03, 6.0555E+03,
     +  6.0635E+03, 6.0715E+03, 6.0795E+03, 6.0876E+03, 6.0956E+03,
     + 6.1036E+03, 6.1117E+03, 6.1197E+03, 6.1278E+03, 6.1358E+03,
     +  6.1439E+03, 6.1520E+03, 6.1601E+03, 6.1681E+03, 6.1762E+03,
     + 6.1843E+03, 6.1924E+03, 6.2005E+03, 6.2086E+03, 6.2167E+03,
     +  6.2249E+03, 6.2330E+03, 6.2411E+03, 6.2492E+03, 6.2574E+03,
     + 6.2655E+03, 6.2737E+03, 6.2818E+03, 6.2900E+03, 6.2982E+03,
     +  6.3063E+03, 6.3145E+03, 6.3227E+03, 6.3309E+03, 6.3391E+03,
     + 6.3473E+03, 6.3555E+03, 6.3637E+03, 6.3719E+03, 6.3801E+03,
     +  6.3883E+03, 6.3966E+03, 6.4048E+03, 6.4130E+03, 6.4213E+03,
     + 6.4295E+03, 6.4378E+03, 6.4460E+03, 6.4543E+03, 6.4626E+03,
     +  6.4708E+03, 6.4791E+03, 6.4874E+03, 6.4957E+03, 6.5040E+03,
     + 6.5123E+03, 6.5206E+03, 6.5289E+03, 6.5372E+03, 6.5455E+03,
     +  6.5539E+03, 6.5622E+03, 6.5705E+03, 6.5789E+03, 6.5872E+03,
     + 6.5956E+03, 6.6039E+03, 6.6123E+03, 6.6207E+03, 6.6290E+03,
     +  6.6374E+03, 6.6458E+03, 6.6542E+03, 6.6626E+03, 6.6710E+03,
     + 6.6794E+03, 6.6878E+03, 6.6962E+03, 6.7046E+03, 6.7130E+03,
     +  6.7215E+03, 6.7299E+03, 6.7384E+03, 6.7468E+03, 6.7553E+03,
     + 6.7637E+03, 6.7722E+03, 6.7806E+03, 6.7891E+03, 6.7976E+03,
     +  6.8061E+03, 6.8145E+03, 6.8230E+03, 6.8315E+03, 6.8400E+03,
     + 6.8485E+03, 6.8571E+03, 6.8656E+03, 6.8741E+03, 6.8826E+03,
     +  6.8912E+03, 6.8997E+03, 6.9082E+03, 6.9168E+03, 6.9253E+03,
     + 6.9339E+03, 6.9425E+03, 6.9510E+03, 6.9596E+03, 6.9682E+03,
     +  6.9768E+03, 6.9854E+03, 6.9940E+03, 7.0026E+03, 7.0112E+03,
     + 7.0198E+03, 7.0284E+03, 7.0370E+03, 7.0456E+03, 7.0543E+03,
     +  7.0629E+03, 7.0715E+03, 7.0802E+03, 7.0888E+03, 7.0975E+03,
     + 7.1062E+03, 7.1148E+03, 7.1235E+03, 7.1322E+03, 7.1409E+03,
     +  7.1495E+03, 7.1582E+03, 7.1669E+03, 7.1756E+03, 7.1843E+03,
     + 7.1931E+03, 7.2018E+03, 7.2105E+03, 7.2192E+03, 7.2280E+03,
     +  7.2367E+03, 7.2454E+03, 7.2542E+03, 7.2629E+03, 7.2717E+03,
     + 7.2805E+03, 7.2892E+03, 7.2980E+03, 7.3068E+03, 7.3156E+03,
     +  7.3244E+03, 7.3332E+03, 7.3420E+03, 7.3508E+03, 7.3596E+03,
     + 7.3684E+03, 7.3772E+03, 7.3860E+03, 7.3949E+03, 7.4037E+03,
     +  7.4125E+03, 7.4214E+03, 7.4302E+03, 7.4391E+03, 7.4480E+03,
     + 7.4568E+03, 7.4657E+03, 7.4746E+03, 7.4835E+03, 7.4923E+03,
     +  7.5012E+03, 7.5101E+03, 7.5190E+03, 7.5279E+03, 7.5368E+03,
     + 7.5458E+03, 7.5547E+03, 7.5636E+03, 7.5725E+03, 7.5815E+03,
     +  7.5904E+03, 7.5994E+03, 7.6083E+03, 7.6173E+03, 7.6262E+03,
     + 7.6352E+03, 7.6442E+03, 7.6532E+03, 7.6622E+03, 7.6711E+03,
     +  7.6801E+03, 7.6891E+03, 7.6981E+03, 7.7071E+03, 7.7162E+03,
     + 7.7252E+03, 7.7342E+03, 7.7432E+03, 7.7523E+03, 7.7613E+03,
     +  7.7703E+03, 7.7794E+03, 7.7884E+03, 7.7975E+03, 7.8066E+03,
     + 7.8156E+03, 7.8247E+03, 7.8338E+03, 7.8429E+03, 7.8520E+03,
     +  7.8611E+03, 7.8702E+03, 7.8793E+03, 7.8884E+03, 7.8975E+03,
     + 7.9066E+03, 7.9157E+03, 7.9249E+03, 7.9340E+03, 7.9431E+03,
     +  7.9523E+03, 7.9614E+03, 7.9706E+03, 7.9797E+03, 7.9889E+03,
     + 7.9981E+03, 8.0073E+03, 8.0164E+03, 8.0256E+03, 8.0348E+03,
     +  8.0440E+03, 8.0532E+03, 8.0624E+03, 8.0716E+03, 8.0808E+03,
     + 8.0901E+03, 8.0993E+03, 8.1085E+03, 8.1178E+03, 8.1270E+03,
     +  8.1362E+03, 8.1455E+03, 8.1547E+03, 8.1640E+03, 8.1733E+03,
     + 8.1825E+03, 8.1918E+03, 8.2011E+03, 8.2104E+03, 8.2197E+03,
     +  8.2290E+03, 8.2383E+03, 8.2476E+03, 8.2569E+03, 8.2662E+03,
     + 8.2755E+03, 8.2848E+03, 8.2942E+03, 8.3035E+03, 8.3129E+03,
     +  8.3222E+03, 8.3315E+03, 8.3409E+03, 8.3503E+03, 8.3596E+03,
     + 8.3690E+03, 8.3784E+03, 8.3878E+03, 8.3971E+03, 8.4065E+03,
     +  8.4159E+03, 8.4253E+03, 8.4347E+03, 8.4441E+03, 8.4536E+03,
     + 8.4630E+03, 8.4724E+03, 8.4818E+03, 8.4913E+03, 8.5007E+03,
     +  8.5102E+03, 8.5196E+03, 8.5291E+03, 8.5385E+03, 8.5480E+03,
     + 8.5575E+03, 8.5669E+03, 8.5764E+03, 8.5859E+03, 8.5954E+03,
     +  8.6049E+03, 8.6144E+03, 8.6239E+03, 8.6334E+03, 8.6429E+03,
     + 8.6524E+03, 8.6619E+03, 8.6715E+03, 8.6810E+03, 8.6905E+03,
     +  8.7001E+03, 8.7096E+03, 8.7192E+03, 8.7288E+03, 8.7383E+03,
     + 8.7479E+03, 8.7575E+03, 8.7670E+03, 8.7766E+03, 8.7862E+03,
     +  8.7958E+03, 8.8054E+03, 8.8150E+03, 8.8246E+03, 8.8342E+03,
     + 8.8438E+03, 8.8535E+03, 8.8631E+03, 8.8727E+03, 8.8824E+03,
     +  8.8920E+03, 8.9017E+03, 8.9113E+03, 8.9210E+03, 8.9306E+03,
     + 8.9403E+03, 8.9500E+03, 8.9596E+03, 8.9693E+03, 8.9790E+03,
     +  8.9887E+03, 8.9984E+03, 9.0081E+03, 9.0178E+03, 9.0275E+03,
     + 9.0372E+03, 9.0470E+03, 9.0567E+03, 9.0664E+03, 9.0761E+03,
     +  9.0859E+03, 9.0956E+03, 9.1054E+03, 9.1151E+03, 9.1249E+03,
     + 9.1347E+03, 9.1444E+03, 9.1542E+03, 9.1640E+03, 9.1738E+03,
     +  9.1836E+03, 9.1934E+03, 9.2032E+03, 9.2130E+03, 9.2228E+03,
     + 9.2326E+03, 9.2424E+03, 9.2522E+03, 9.2621E+03, 9.2719E+03,
     +  9.2817E+03, 9.2916E+03, 9.3014E+03, 9.3113E+03, 9.3211E+03,
     + 9.3310E+03, 9.3409E+03, 9.3507E+03, 9.3606E+03, 9.3705E+03,
     +  9.3804E+03, 9.3903E+03, 9.4002E+03, 9.4101E+03, 9.4200E+03,
     + 9.4299E+03, 9.4398E+03, 9.4497E+03, 9.4596E+03, 9.4696E+03,
     +  9.4795E+03, 9.4894E+03, 9.4994E+03, 9.5093E+03, 9.5193E+03,
     + 9.5292E+03, 9.5392E+03, 9.5492E+03, 9.5592E+03, 9.5691E+03,
     +  9.5791E+03, 9.5891E+03, 9.5991E+03, 9.6091E+03, 9.6191E+03,
     + 9.6291E+03, 9.6391E+03, 9.6491E+03, 9.6592E+03, 9.6692E+03,
     +  9.6792E+03, 9.6892E+03, 9.6993E+03, 9.7093E+03, 9.7194E+03,
     + 9.7294E+03, 9.7395E+03, 9.7496E+03, 9.7596E+03, 9.7697E+03,
     +  9.7798E+03, 9.7899E+03, 9.8000E+03, 9.8101E+03, 9.8202E+03,
     + 9.8303E+03, 9.8404E+03, 9.8505E+03, 9.8606E+03, 9.8707E+03,
     +  9.8808E+03, 9.8910E+03, 9.9011E+03, 9.9113E+03, 9.9214E+03,
     + 9.9315E+03, 9.9417E+03, 9.9519E+03, 9.9620E+03, 9.9722E+03,
     +  9.9824E+03, 9.9926E+03, 1.0003E+04, 1.0013E+04, 1.0023E+04,
     + 1.0033E+04, 1.0044E+04, 1.0054E+04, 1.0064E+04, 1.0074E+04,
     +  1.0084E+04, 1.0095E+04, 1.0105E+04, 1.0115E+04, 1.0125E+04,
     + 1.0136E+04, 1.0146E+04, 1.0156E+04, 1.0166E+04, 1.0177E+04,
     +  1.0187E+04, 1.0197E+04, 1.0207E+04, 1.0218E+04, 1.0228E+04,
     + 1.0238E+04, 1.0249E+04, 1.0259E+04, 1.0269E+04, 1.0279E+04,
     +  1.0290E+04, 1.0300E+04, 1.0310E+04, 1.0321E+04, 1.0331E+04,
     + 1.0341E+04, 1.0352E+04, 1.0362E+04, 1.0373E+04, 1.0383E+04,
     +  1.0393E+04, 1.0404E+04, 1.0414E+04, 1.0424E+04, 1.0435E+04,
     + 1.0445E+04, 1.0456E+04, 1.0466E+04, 1.0476E+04, 1.0487E+04,
     +  1.0497E+04, 1.0508E+04, 1.0518E+04, 1.0528E+04, 1.0539E+04,
     + 1.0549E+04, 1.0560E+04, 1.0570E+04, 1.0581E+04, 1.0591E+04,
     +  1.0601E+04, 1.0612E+04, 1.0622E+04, 1.0633E+04, 1.0643E+04,
     + 1.0654E+04, 1.0664E+04, 1.0675E+04, 1.0685E+04, 1.0696E+04,
     +  1.0706E+04, 1.0717E+04, 1.0727E+04, 1.0738E+04, 1.0748E+04,
     + 1.0759E+04, 1.0769E+04, 1.0780E+04, 1.0791E+04, 1.0801E+04,
     +  1.0812E+04, 1.0822E+04, 1.0833E+04, 1.0843E+04, 1.0854E+04,
     + 1.0864E+04, 1.0875E+04, 1.0886E+04, 1.0896E+04, 1.0907E+04,
     +  1.0917E+04, 1.0928E+04, 1.0939E+04, 1.0949E+04, 1.0960E+04,
     + 1.0970E+04, 1.0981E+04, 1.0992E+04, 1.1002E+04, 1.1013E+04,
     +  1.1024E+04, 1.1034E+04, 1.1045E+04, 1.1056E+04, 1.1066E+04,
     + 1.1077E+04, 1.1088E+04, 1.1098E+04, 1.1109E+04, 1.1120E+04,
     +  1.1130E+04, 1.1141E+04, 1.1152E+04, 1.1162E+04, 1.1173E+04,
     + 1.1184E+04, 1.1194E+04, 1.1205E+04, 1.1216E+04, 1.1227E+04,
     +  1.1237E+04, 1.1248E+04, 1.1259E+04, 1.1270E+04, 1.1280E+04,
     + 1.1291E+04, 1.1302E+04, 1.1313E+04, 1.1323E+04, 1.1334E+04,
     +  1.1345E+04, 1.1356E+04, 1.1367E+04, 1.1377E+04, 1.1388E+04,
     + 1.1399E+04, 1.1410E+04, 1.1421E+04, 1.1431E+04, 1.1442E+04,
     +  1.1453E+04, 1.1464E+04, 1.1475E+04, 1.1486E+04, 1.1496E+04,
     + 1.1507E+04, 1.1518E+04, 1.1529E+04, 1.1540E+04, 1.1551E+04,
     +  1.1562E+04, 1.1572E+04, 1.1583E+04, 1.1594E+04, 1.1605E+04,
     + 1.1616E+04, 1.1627E+04, 1.1638E+04, 1.1649E+04, 1.1660E+04,
     +  1.1670E+04, 1.1681E+04, 1.1692E+04, 1.1703E+04, 1.1714E+04,
     + 1.1725E+04, 1.1736E+04, 1.1747E+04, 1.1758E+04, 1.1769E+04,
     +  1.1780E+04, 1.1791E+04, 1.1802E+04, 1.1813E+04, 1.1824E+04,
     + 1.1835E+04, 1.1846E+04, 1.1857E+04, 1.1868E+04, 1.1879E+04,
     +  1.1890E+04, 1.1901E+04, 1.1912E+04, 1.1923E+04, 1.1934E+04,
     + 1.1945E+04, 1.1956E+04, 1.1967E+04, 1.1978E+04, 1.1989E+04,
     +  1.2000E+04, 1.2011E+04, 1.2022E+04, 1.2033E+04, 1.2044E+04,
     + 1.2055E+04, 1.2066E+04, 1.2077E+04, 1.2089E+04, 1.2100E+04,
     +  1.2111E+04, 1.2122E+04, 1.2133E+04, 1.2144E+04, 1.2155E+04,
     + 1.2166E+04, 1.2177E+04, 1.2189E+04, 1.2200E+04, 1.2211E+04,
     +  1.2222E+04, 1.2233E+04, 1.2244E+04, 1.2255E+04, 1.2267E+04,
     + 1.2278E+04, 1.2289E+04, 1.2300E+04, 1.2311E+04, 1.2322E+04,
     +  1.2334E+04, 1.2345E+04, 1.2356E+04, 1.2367E+04, 1.2378E+04,
     + 1.2390E+04, 1.2401E+04, 1.2412E+04, 1.2423E+04, 1.2434E+04,
     +  1.2446E+04, 1.2457E+04, 1.2468E+04, 1.2479E+04, 1.2491E+04,
     + 1.2502E+04, 1.2513E+04, 1.2524E+04, 1.2536E+04, 1.2547E+04,
     +  1.2558E+04, 1.2569E+04, 1.2581E+04, 1.2592E+04, 1.2603E+04,
     + 1.2615E+04, 1.2626E+04, 1.2637E+04, 1.2648E+04, 1.2660E+04,
     +  1.2671E+04, 1.2682E+04, 1.2694E+04, 1.2705E+04, 1.2716E+04,
     + 1.2728E+04, 1.2739E+04, 1.2750E+04, 1.2762E+04, 1.2773E+04,
     +  1.2784E+04, 1.2796E+04, 1.2807E+04, 1.2819E+04, 1.2830E+04,
     + 1.2841E+04, 1.2853E+04, 1.2864E+04, 1.2875E+04, 1.2887E+04,
     +  1.2898E+04, 1.2910E+04, 1.2921E+04, 1.2932E+04, 1.2944E+04,
     + 1.2955E+04, 1.2967E+04, 1.2978E+04, 1.2990E+04, 1.3001E+04,
     +  1.3012E+04, 1.3024E+04, 1.3035E+04, 1.3047E+04, 1.3058E+04,
     + 1.3070E+04, 1.3081E+04, 1.3093E+04, 1.3104E+04, 1.3116E+04,
     +  1.3127E+04, 1.3139E+04, 1.3150E+04, 1.3162E+04, 1.3173E+04,
     + 1.3185E+04, 1.3196E+04, 1.3208E+04, 1.3219E+04, 1.3231E+04,
     +  1.3242E+04, 1.3254E+04, 1.3265E+04, 1.3277E+04, 1.3288E+04,
     + 1.3300E+04, 1.3311E+04, 1.3323E+04, 1.3335E+04, 1.3346E+04,
     +  1.3358E+04, 1.3369E+04, 1.3381E+04, 1.3392E+04, 1.3404E+04,
     + 1.3416E+04, 1.3427E+04, 1.3439E+04, 1.3450E+04, 1.3462E+04,
     +  1.3474E+04, 1.3485E+04, 1.3497E+04, 1.3509E+04, 1.3520E+04,
     + 1.3532E+04, 1.3543E+04, 1.3555E+04, 1.3567E+04, 1.3578E+04,
     +  1.3590E+04, 1.3602E+04, 1.3613E+04, 1.3625E+04, 1.3637E+04,
     + 1.3648E+04, 1.3660E+04, 1.3672E+04, 1.3683E+04, 1.3695E+04,
     +  1.3707E+04, 1.3719E+04, 1.3730E+04, 1.3742E+04, 1.3754E+04,
     + 1.3765E+04, 1.3777E+04, 1.3789E+04, 1.3801E+04, 1.3812E+04,
     +  1.3824E+04, 1.3836E+04, 1.3848E+04, 1.3859E+04, 1.3871E+04,
     + 1.3883E+04, 1.3895E+04, 1.3906E+04, 1.3918E+04, 1.3930E+04,
     +  1.3942E+04, 1.3953E+04, 1.3965E+04, 1.3977E+04, 1.3989E+04,
     + 1.4001E+04, 1.4012E+04, 1.4024E+04, 1.4036E+04, 1.4048E+04,
     +  1.4060E+04, 1.4071E+04, 1.4083E+04, 1.4095E+04, 1.4107E+04,
     + 1.4119E+04, 1.4131E+04, 1.4143E+04, 1.4154E+04, 1.4166E+04,
     +  1.4178E+04, 1.4190E+04, 1.4202E+04, 1.4214E+04, 1.4226E+04,
     + 1.4237E+04, 1.4249E+04, 1.4261E+04, 1.4273E+04, 1.4285E+04,
     +  1.4297E+04, 1.4309E+04, 1.4321E+04, 1.4333E+04, 1.4345E+04,
     + 1.4357E+04, 1.4368E+04, 1.4380E+04, 1.4392E+04, 1.4404E+04,
     +  1.4416E+04, 1.4428E+04, 1.4440E+04, 1.4452E+04, 1.4464E+04,
     + 1.4476E+04, 1.4488E+04, 1.4500E+04, 1.4512E+04, 1.4524E+04,
     +  1.4536E+04, 1.4548E+04, 1.4560E+04, 1.4572E+04, 1.4584E+04,
     + 1.4596E+04, 1.4608E+04, 1.4620E+04, 1.4632E+04, 1.4644E+04,
     +  1.4656E+04, 1.4668E+04, 1.4680E+04, 1.4692E+04, 1.4704E+04,
     + 1.4716E+04, 1.4728E+04, 1.4740E+04, 1.4752E+04, 1.4764E+04,
     +  1.4776E+04, 1.4789E+04, 1.4801E+04, 1.4813E+04, 1.4825E+04,
     + 1.4837E+04, 1.4849E+04, 1.4861E+04, 1.4873E+04, 1.4885E+04,
     +  1.4897E+04, 1.4909E+04, 1.4922E+04, 1.4934E+04, 1.4946E+04,
     + 1.4958E+04, 1.4970E+04, 1.4982E+04, 1.4994E+04, 1.5007E+04,
     +  1.5019E+04, 1.5031E+04, 1.5043E+04, 1.5055E+04, 1.5067E+04,
     + 1.5079E+04, 1.5092E+04, 1.5104E+04, 1.5116E+04, 1.5128E+04,
     +  1.5140E+04, 1.5153E+04, 1.5165E+04, 1.5177E+04, 1.5189E+04,
     + 1.5201E+04, 1.5214E+04, 1.5226E+04, 1.5238E+04, 1.5250E+04,
     +  1.5262E+04, 1.5275E+04, 1.5287E+04, 1.5299E+04, 1.5311E+04,
     + 1.5324E+04, 1.5336E+04, 1.5348E+04, 1.5360E+04, 1.5373E+04,
     +  1.5385E+04, 1.5397E+04, 1.5409E+04, 1.5422E+04, 1.5434E+04,
     + 1.5446E+04, 1.5459E+04, 1.5471E+04, 1.5483E+04, 1.5495E+04,
     +  1.5508E+04, 1.5520E+04, 1.5532E+04, 1.5545E+04, 1.5557E+04,
     + 1.5569E+04, 1.5582E+04, 1.5594E+04, 1.5606E+04, 1.5619E+04,
     +  1.5631E+04, 1.5643E+04, 1.5656E+04, 1.5668E+04, 1.5680E+04,
     + 1.5693E+04, 1.5705E+04, 1.5718E+04, 1.5730E+04, 1.5742E+04,
     +  1.5755E+04, 1.5767E+04, 1.5779E+04, 1.5792E+04, 1.5804E+04,
     + 1.5817E+04, 1.5829E+04, 1.5841E+04, 1.5854E+04, 1.5866E+04,
     +  1.5879E+04, 1.5891E+04, 1.5904E+04, 1.5916E+04, 1.5928E+04,
     + 1.5941E+04, 1.5953E+04, 1.5966E+04, 1.5978E+04, 1.5991E+04,
     +  1.6003E+04, 1.6016E+04, 1.6028E+04, 1.6041E+04, 1.6053E+04,
     + 1.6066E+04, 1.6078E+04, 1.6090E+04, 1.6103E+04, 1.6115E+04,
     +  1.6128E+04, 1.6140E+04, 1.6153E+04, 1.6165E+04, 1.6178E+04,
     + 1.6191E+04, 1.6203E+04, 1.6216E+04, 1.6228E+04, 1.6241E+04,
     +  1.6253E+04, 1.6266E+04, 1.6278E+04, 1.6291E+04, 1.6303E+04,
     + 1.6316E+04, 1.6328E+04, 1.6341E+04, 1.6354E+04, 1.6366E+04,
     +  1.6379E+04, 1.6391E+04, 1.6404E+04, 1.6416E+04, 1.6429E+04,
     + 1.6442E+04, 1.6454E+04, 1.6467E+04, 1.6479E+04, 1.6492E+04,
     +  1.6505E+04, 1.6517E+04, 1.6530E+04, 1.6543E+04, 1.6555E+04,
     + 1.6568E+04, 1.6580E+04, 1.6593E+04, 1.6606E+04, 1.6618E+04,
     +  1.6631E+04, 1.6644E+04, 1.6656E+04, 1.6669E+04, 1.6682E+04,
     + 1.6694E+04, 1.6707E+04, 1.6720E+04, 1.6732E+04, 1.6745E+04,
     +  1.6758E+04, 1.6770E+04, 1.6783E+04, 1.6796E+04, 1.6809E+04,
     + 1.6821E+04, 1.6834E+04, 1.6847E+04, 1.6859E+04, 1.6872E+04,
     +  1.6885E+04, 1.6898E+04, 1.6910E+04, 1.6923E+04, 1.6936E+04,
     + 1.6948E+04, 1.6961E+04, 1.6974E+04, 1.6987E+04, 1.6999E+04,
     +  1.7012E+04, 1.7025E+04, 1.7038E+04, 1.7051E+04, 1.7063E+04,
     + 1.7076E+04, 1.7089E+04, 1.7102E+04, 1.7114E+04, 1.7127E+04,
     +  1.7140E+04, 1.7153E+04, 1.7166E+04, 1.7178E+04, 1.7191E+04,
     + 1.7204E+04, 1.7217E+04, 1.7230E+04, 1.7243E+04, 1.7255E+04,
     +  1.7268E+04, 1.7281E+04, 1.7294E+04, 1.7307E+04, 1.7320E+04,
     + 1.7332E+04, 1.7345E+04, 1.7358E+04, 1.7371E+04, 1.7384E+04,
     +  1.7397E+04, 1.7410E+04, 1.7423E+04, 1.7435E+04, 1.7448E+04,
     + 1.7461E+04, 1.7474E+04, 1.7487E+04, 1.7500E+04, 1.7513E+04,
     +  1.7526E+04, 1.7539E+04, 1.7552E+04, 1.7564E+04, 1.7577E+04,
     + 1.7590E+04, 1.7603E+04, 1.7616E+04, 1.7629E+04, 1.7642E+04,
     +  1.7655E+04, 1.7668E+04, 1.7681E+04, 1.7694E+04, 1.7707E+04,
     + 1.7720E+04, 1.7733E+04, 1.7746E+04, 1.7759E+04, 1.7772E+04,
     +  1.7785E+04, 1.7798E+04, 1.7811E+04, 1.7824E+04, 1.7837E+04,
     + 1.7850E+04, 1.7863E+04, 1.7876E+04, 1.7889E+04, 1.7902E+04,
     +  1.7915E+04, 1.7928E+04, 1.7941E+04, 1.7954E+04, 1.7967E+04,
     + 1.7980E+04, 1.7993E+04, 1.8006E+04, 1.8019E+04, 1.8032E+04,
     +  1.8045E+04, 1.8058E+04, 1.8071E+04, 1.8084E+04, 1.8097E+04,
     + 1.8110E+04, 1.8123E+04, 1.8136E+04, 1.8150E+04, 1.8163E+04,
     +  1.8176E+04, 1.8189E+04, 1.8202E+04, 1.8215E+04, 1.8228E+04,
     + 1.8241E+04, 1.8254E+04, 1.8267E+04, 1.8281E+04, 1.8294E+04,
     +  1.8307E+04, 1.8320E+04, 1.8333E+04, 1.8346E+04, 1.8359E+04,
     + 1.8372E+04, 1.8386E+04, 1.8399E+04, 1.8412E+04, 1.8425E+04,
     +  1.8438E+04, 1.8451E+04, 1.8465E+04, 1.8478E+04, 1.8491E+04,
     + 1.8504E+04, 1.8517E+04, 1.8530E+04, 1.8544E+04, 1.8557E+04,
     +  1.8570E+04, 1.8583E+04, 1.8596E+04, 1.8610E+04, 1.8623E+04,
     + 1.8636E+04, 1.8649E+04, 1.8662E+04, 1.8676E+04, 1.8689E+04,
     +  1.8702E+04, 1.8715E+04, 1.8729E+04, 1.8742E+04, 1.8755E+04,
     + 1.8768E+04, 1.8782E+04, 1.8795E+04, 1.8808E+04, 1.8821E+04,
     +  1.8835E+04, 1.8848E+04, 1.8861E+04, 1.8874E+04, 1.8888E+04,
     + 1.8901E+04, 1.8914E+04, 1.8928E+04, 1.8941E+04, 1.8954E+04,
     +  1.8967E+04, 1.8981E+04, 1.8994E+04, 1.9007E+04, 1.9021E+04,
     + 1.9034E+04, 1.9047E+04, 1.9061E+04, 1.9074E+04, 1.9087E+04,
     +  1.9101E+04, 1.9114E+04, 1.9127E+04, 1.9141E+04, 1.9154E+04,
     + 1.9167E+04, 1.9181E+04, 1.9194E+04, 1.9207E+04, 1.9221E+04,
     +  1.9234E+04, 1.9247E+04, 1.9261E+04, 1.9274E+04, 1.9288E+04,
     + 1.9301E+04, 1.9314E+04, 1.9328E+04, 1.9341E+04, 1.9354E+04,
     +  1.9368E+04, 1.9381E+04, 1.9395E+04, 1.9408E+04, 1.9422E+04,
     + 1.9435E+04, 1.9448E+04, 1.9462E+04, 1.9475E+04, 1.9489E+04,
     +  1.9502E+04, 1.9515E+04, 1.9529E+04, 1.9542E+04, 1.9556E+04,
     + 1.9569E+04, 1.9583E+04, 1.9596E+04, 1.9610E+04, 1.9623E+04,
     +  1.9637E+04, 1.9650E+04, 1.9663E+04, 1.9677E+04, 1.9690E+04,
     + 1.9704E+04, 1.9717E+04, 1.9731E+04, 1.9744E+04, 1.9758E+04,
     +  1.9771E+04, 1.9785E+04, 1.9798E+04, 1.9812E+04, 1.9825E+04,
     + 1.9839E+04, 1.9852E+04, 1.9866E+04, 1.9880E+04, 1.9893E+04,
     +  1.9907E+04, 1.9920E+04, 1.9934E+04, 1.9947E+04, 1.9961E+04,
     + 1.9974E+04, 1.9988E+04, 2.0001E+04, 2.0015E+04, 2.0029E+04,
     +  2.0042E+04, 2.0056E+04, 2.0069E+04, 2.0083E+04, 2.0096E+04,
     + 2.0110E+04, 2.0124E+04, 2.0137E+04, 2.0151E+04, 2.0164E+04,
     +  2.0178E+04, 2.0192E+04, 2.0205E+04, 2.0219E+04, 2.0232E+04,
     + 2.0246E+04, 2.0260E+04, 2.0273E+04, 2.0287E+04, 2.0300E+04,
     +  2.0314E+04, 2.0328E+04, 2.0341E+04, 2.0355E+04, 2.0369E+04,
     + 2.0382E+04, 2.0396E+04, 2.0410E+04, 2.0423E+04, 2.0437E+04,
     +  2.0451E+04, 2.0464E+04, 2.0478E+04, 2.0492E+04, 2.0505E+04,
     + 2.0519E+04, 2.0533E+04, 2.0546E+04, 2.0560E+04, 2.0574E+04,
     +  2.0587E+04, 2.0601E+04, 2.0615E+04, 2.0628E+04, 2.0642E+04/
      data (QofT( 2,J),J=2409,4510)/ 2.0656E+04, 2.0670E+04,
     +  2.0683E+04, 2.0697E+04, 2.0711E+04, 2.0724E+04, 2.0738E+04,
     +  2.0752E+04,
     + 2.0766E+04, 2.0779E+04, 2.0793E+04, 2.0807E+04, 2.0821E+04,
     +  2.0834E+04, 2.0848E+04, 2.0862E+04, 2.0876E+04, 2.0889E+04,
     + 2.0903E+04, 2.0917E+04, 2.0931E+04, 2.0945E+04, 2.0958E+04,
     +  2.0972E+04, 2.0986E+04, 2.1000E+04, 2.1014E+04, 2.1027E+04,
     + 2.1041E+04, 2.1055E+04, 2.1069E+04, 2.1083E+04, 2.1096E+04,
     +  2.1110E+04, 2.1124E+04, 2.1138E+04, 2.1152E+04, 2.1165E+04,
     + 2.1179E+04, 2.1193E+04, 2.1207E+04, 2.1221E+04, 2.1235E+04,
     +  2.1248E+04, 2.1262E+04, 2.1276E+04, 2.1290E+04, 2.1304E+04,
     + 2.1318E+04, 2.1332E+04, 2.1346E+04, 2.1359E+04, 2.1373E+04,
     +  2.1387E+04, 2.1401E+04, 2.1415E+04, 2.1429E+04, 2.1443E+04,
     + 2.1457E+04, 2.1470E+04, 2.1484E+04, 2.1498E+04, 2.1512E+04,
     +  2.1526E+04, 2.1540E+04, 2.1554E+04, 2.1568E+04, 2.1582E+04,
     + 2.1596E+04, 2.1610E+04, 2.1624E+04, 2.1637E+04, 2.1651E+04,
     +  2.1665E+04, 2.1679E+04, 2.1693E+04, 2.1707E+04, 2.1721E+04,
     + 2.1735E+04, 2.1749E+04, 2.1763E+04, 2.1777E+04, 2.1791E+04,
     +  2.1805E+04, 2.1819E+04, 2.1833E+04, 2.1847E+04, 2.1861E+04,
     + 2.1875E+04, 2.1889E+04, 2.1903E+04, 2.1917E+04, 2.1931E+04,
     +  2.1945E+04, 2.1959E+04, 2.1973E+04, 2.1987E+04, 2.2001E+04,
     + 2.2015E+04, 2.2029E+04, 2.2043E+04, 2.2057E+04, 2.2071E+04,
     +  2.2085E+04, 2.2099E+04, 2.2113E+04, 2.2127E+04, 2.2141E+04,
     + 2.2155E+04, 2.2169E+04, 2.2183E+04, 2.2197E+04, 2.2211E+04,
     +  2.2226E+04, 2.2240E+04, 2.2254E+04, 2.2268E+04, 2.2282E+04,
     + 2.2296E+04, 2.2310E+04, 2.2324E+04, 2.2338E+04, 2.2352E+04,
     +  2.2366E+04, 2.2380E+04, 2.2395E+04, 2.2409E+04, 2.2423E+04,
     + 2.2437E+04, 2.2451E+04, 2.2465E+04, 2.2479E+04, 2.2493E+04,
     +  2.2507E+04, 2.2522E+04, 2.2536E+04, 2.2550E+04, 2.2564E+04,
     + 2.2578E+04, 2.2592E+04, 2.2606E+04, 2.2621E+04, 2.2635E+04,
     +  2.2649E+04, 2.2663E+04, 2.2677E+04, 2.2691E+04, 2.2705E+04,
     + 2.2720E+04, 2.2734E+04, 2.2748E+04, 2.2762E+04, 2.2776E+04,
     +  2.2790E+04, 2.2805E+04, 2.2819E+04, 2.2833E+04, 2.2847E+04,
     + 2.2861E+04, 2.2876E+04, 2.2890E+04, 2.2904E+04, 2.2918E+04,
     +  2.2932E+04, 2.2947E+04, 2.2961E+04, 2.2975E+04, 2.2989E+04,
     + 2.3004E+04, 2.3018E+04, 2.3032E+04, 2.3046E+04, 2.3061E+04,
     +  2.3075E+04, 2.3089E+04, 2.3103E+04, 2.3117E+04, 2.3132E+04,
     + 2.3146E+04, 2.3160E+04, 2.3175E+04, 2.3189E+04, 2.3203E+04,
     +  2.3217E+04, 2.3232E+04, 2.3246E+04, 2.3260E+04, 2.3274E+04,
     + 2.3289E+04, 2.3303E+04, 2.3317E+04, 2.3332E+04, 2.3346E+04,
     +  2.3360E+04, 2.3374E+04, 2.3389E+04, 2.3403E+04, 2.3417E+04,
     + 2.3432E+04, 2.3446E+04, 2.3460E+04, 2.3475E+04, 2.3489E+04,
     +  2.3503E+04, 2.3518E+04, 2.3532E+04, 2.3546E+04, 2.3561E+04,
     + 2.3575E+04, 2.3589E+04, 2.3604E+04, 2.3618E+04, 2.3632E+04,
     +  2.3647E+04, 2.3661E+04, 2.3675E+04, 2.3690E+04, 2.3704E+04,
     + 2.3719E+04, 2.3733E+04, 2.3747E+04, 2.3762E+04, 2.3776E+04,
     +  2.3790E+04, 2.3805E+04, 2.3819E+04, 2.3834E+04, 2.3848E+04,
     + 2.3862E+04, 2.3877E+04, 2.3891E+04, 2.3906E+04, 2.3920E+04,
     +  2.3934E+04, 2.3949E+04, 2.3963E+04, 2.3978E+04, 2.3992E+04,
     + 2.4007E+04, 2.4021E+04, 2.4035E+04, 2.4050E+04, 2.4064E+04,
     +  2.4079E+04, 2.4093E+04, 2.4108E+04, 2.4122E+04, 2.4137E+04,
     + 2.4151E+04, 2.4165E+04, 2.4180E+04, 2.4194E+04, 2.4209E+04,
     +  2.4223E+04, 2.4238E+04, 2.4252E+04, 2.4267E+04, 2.4281E+04,
     + 2.4296E+04, 2.4310E+04, 2.4325E+04, 2.4339E+04, 2.4354E+04,
     +  2.4368E+04, 2.4383E+04, 2.4397E+04, 2.4412E+04, 2.4426E+04,
     + 2.4441E+04, 2.4455E+04, 2.4470E+04, 2.4484E+04, 2.4499E+04,
     +  2.4513E+04, 2.4528E+04, 2.4542E+04, 2.4557E+04, 2.4571E+04,
     + 2.4586E+04, 2.4600E+04, 2.4615E+04, 2.4630E+04, 2.4644E+04,
     +  2.4659E+04, 2.4673E+04, 2.4688E+04, 2.4702E+04, 2.4717E+04,
     + 2.4731E+04, 2.4746E+04, 2.4761E+04, 2.4775E+04, 2.4790E+04,
     +  2.4804E+04, 2.4819E+04, 2.4833E+04, 2.4848E+04, 2.4863E+04,
     + 2.4877E+04, 2.4892E+04, 2.4906E+04, 2.4921E+04, 2.4936E+04,
     +  2.4950E+04, 2.4965E+04, 2.4979E+04, 2.4994E+04, 2.5009E+04,
     + 2.5023E+04, 2.5038E+04, 2.5053E+04, 2.5067E+04, 2.5082E+04,
     +  2.5096E+04, 2.5111E+04, 2.5126E+04, 2.5140E+04, 2.5155E+04,
     + 2.5170E+04, 2.5184E+04, 2.5199E+04, 2.5214E+04, 2.5228E+04,
     +  2.5243E+04, 2.5258E+04, 2.5272E+04, 2.5287E+04, 2.5302E+04,
     + 2.5316E+04, 2.5331E+04, 2.5346E+04, 2.5360E+04, 2.5375E+04,
     +  2.5390E+04, 2.5404E+04, 2.5419E+04, 2.5434E+04, 2.5448E+04,
     + 2.5463E+04, 2.5478E+04, 2.5493E+04, 2.5507E+04, 2.5522E+04,
     +  2.5537E+04, 2.5551E+04, 2.5566E+04, 2.5581E+04, 2.5596E+04,
     + 2.5610E+04, 2.5625E+04, 2.5640E+04, 2.5654E+04, 2.5669E+04,
     +  2.5684E+04, 2.5699E+04, 2.5713E+04, 2.5728E+04, 2.5743E+04,
     + 2.5758E+04, 2.5772E+04, 2.5787E+04, 2.5802E+04, 2.5817E+04,
     +  2.5831E+04, 2.5846E+04, 2.5861E+04, 2.5876E+04, 2.5891E+04,
     + 2.5905E+04, 2.5920E+04, 2.5935E+04, 2.5950E+04, 2.5964E+04,
     +  2.5979E+04, 2.5994E+04, 2.6009E+04, 2.6024E+04, 2.6038E+04,
     + 2.6053E+04, 2.6068E+04, 2.6083E+04, 2.6098E+04, 2.6112E+04,
     +  2.6127E+04, 2.6142E+04, 2.6157E+04, 2.6172E+04, 2.6187E+04,
     + 2.6201E+04, 2.6216E+04, 2.6231E+04, 2.6246E+04, 2.6261E+04,
     +  2.6276E+04, 2.6290E+04, 2.6305E+04, 2.6320E+04, 2.6335E+04,
     + 2.6350E+04, 2.6365E+04, 2.6380E+04, 2.6394E+04, 2.6409E+04,
     +  2.6424E+04, 2.6439E+04, 2.6454E+04, 2.6469E+04, 2.6484E+04,
     + 2.6498E+04, 2.6513E+04, 2.6528E+04, 2.6543E+04, 2.6558E+04,
     +  2.6573E+04, 2.6588E+04, 2.6603E+04, 2.6618E+04, 2.6633E+04,
     + 2.6647E+04, 2.6662E+04, 2.6677E+04, 2.6692E+04, 2.6707E+04,
     +  2.6722E+04, 2.6737E+04, 2.6752E+04, 2.6767E+04, 2.6782E+04,
     + 2.6797E+04, 2.6812E+04, 2.6826E+04, 2.6841E+04, 2.6856E+04,
     +  2.6871E+04, 2.6886E+04, 2.6901E+04, 2.6916E+04, 2.6931E+04,
     + 2.6946E+04, 2.6961E+04, 2.6976E+04, 2.6991E+04, 2.7006E+04,
     +  2.7021E+04, 2.7036E+04, 2.7051E+04, 2.7066E+04, 2.7081E+04,
     + 2.7096E+04, 2.7111E+04, 2.7126E+04, 2.7141E+04, 2.7156E+04,
     +  2.7171E+04, 2.7186E+04, 2.7201E+04, 2.7216E+04, 2.7231E+04,
     + 2.7246E+04, 2.7261E+04, 2.7276E+04, 2.7291E+04, 2.7306E+04,
     +  2.7321E+04, 2.7336E+04, 2.7351E+04, 2.7366E+04, 2.7381E+04,
     + 2.7396E+04, 2.7411E+04, 2.7426E+04, 2.7441E+04, 2.7456E+04,
     +  2.7471E+04, 2.7486E+04, 2.7501E+04, 2.7516E+04, 2.7531E+04,
     + 2.7546E+04, 2.7561E+04, 2.7576E+04, 2.7591E+04, 2.7606E+04,
     +  2.7622E+04, 2.7637E+04, 2.7652E+04, 2.7667E+04, 2.7682E+04,
     + 2.7697E+04, 2.7712E+04, 2.7727E+04, 2.7742E+04, 2.7757E+04,
     +  2.7772E+04, 2.7787E+04, 2.7803E+04, 2.7818E+04, 2.7833E+04,
     + 2.7848E+04, 2.7863E+04, 2.7878E+04, 2.7893E+04, 2.7908E+04,
     +  2.7923E+04, 2.7938E+04, 2.7954E+04, 2.7969E+04, 2.7984E+04,
     + 2.7999E+04, 2.8014E+04, 2.8029E+04, 2.8044E+04, 2.8059E+04,
     +  2.8075E+04, 2.8090E+04, 2.8105E+04, 2.8120E+04, 2.8135E+04,
     + 2.8150E+04, 2.8165E+04, 2.8181E+04, 2.8196E+04, 2.8211E+04,
     +  2.8226E+04, 2.8241E+04, 2.8256E+04, 2.8272E+04, 2.8287E+04,
     + 2.8302E+04, 2.8317E+04, 2.8332E+04, 2.8347E+04, 2.8363E+04,
     +  2.8378E+04, 2.8393E+04, 2.8408E+04, 2.8423E+04, 2.8439E+04,
     + 2.8454E+04, 2.8469E+04, 2.8484E+04, 2.8499E+04, 2.8515E+04,
     +  2.8530E+04, 2.8545E+04, 2.8560E+04, 2.8575E+04, 2.8591E+04,
     + 2.8606E+04, 2.8621E+04, 2.8636E+04, 2.8652E+04, 2.8667E+04,
     +  2.8682E+04, 2.8697E+04, 2.8712E+04, 2.8728E+04, 2.8743E+04,
     + 2.8758E+04, 2.8773E+04, 2.8789E+04, 2.8804E+04, 2.8819E+04,
     +  2.8834E+04, 2.8850E+04, 2.8865E+04, 2.8880E+04, 2.8895E+04,
     + 2.8911E+04, 2.8926E+04, 2.8941E+04, 2.8957E+04, 2.8972E+04,
     +  2.8987E+04, 2.9002E+04, 2.9018E+04, 2.9033E+04, 2.9048E+04,
     + 2.9063E+04, 2.9079E+04, 2.9094E+04, 2.9109E+04, 2.9125E+04,
     +  2.9140E+04, 2.9155E+04, 2.9171E+04, 2.9186E+04, 2.9201E+04,
     + 2.9216E+04, 2.9232E+04, 2.9247E+04, 2.9262E+04, 2.9278E+04,
     +  2.9293E+04, 2.9308E+04, 2.9324E+04, 2.9339E+04, 2.9354E+04,
     + 2.9370E+04, 2.9385E+04, 2.9400E+04, 2.9416E+04, 2.9431E+04,
     +  2.9446E+04, 2.9462E+04, 2.9477E+04, 2.9492E+04, 2.9508E+04,
     + 2.9523E+04, 2.9538E+04, 2.9554E+04, 2.9569E+04, 2.9585E+04,
     +  2.9600E+04, 2.9615E+04, 2.9631E+04, 2.9646E+04, 2.9661E+04,
     + 2.9677E+04, 2.9692E+04, 2.9708E+04, 2.9723E+04, 2.9738E+04,
     +  2.9754E+04, 2.9769E+04, 2.9784E+04, 2.9800E+04, 2.9815E+04,
     + 2.9831E+04, 2.9846E+04, 2.9861E+04, 2.9877E+04, 2.9892E+04,
     +  2.9908E+04, 2.9923E+04, 2.9939E+04, 2.9954E+04, 2.9969E+04,
     + 2.9985E+04, 3.0000E+04, 3.0016E+04, 3.0031E+04, 3.0046E+04,
     +  3.0062E+04, 3.0077E+04, 3.0093E+04, 3.0108E+04, 3.0124E+04,
     + 3.0139E+04, 3.0155E+04, 3.0170E+04, 3.0185E+04, 3.0201E+04,
     +  3.0216E+04, 3.0232E+04, 3.0247E+04, 3.0263E+04, 3.0278E+04,
     + 3.0294E+04, 3.0309E+04, 3.0325E+04, 3.0340E+04, 3.0355E+04,
     +  3.0371E+04, 3.0386E+04, 3.0402E+04, 3.0417E+04, 3.0433E+04,
     + 3.0448E+04, 3.0464E+04, 3.0479E+04, 3.0495E+04, 3.0510E+04,
     +  3.0526E+04, 3.0541E+04, 3.0557E+04, 3.0572E+04, 3.0588E+04,
     + 3.0603E+04, 3.0619E+04, 3.0634E+04, 3.0650E+04, 3.0665E+04,
     +  3.0681E+04, 3.0696E+04, 3.0712E+04, 3.0727E+04, 3.0743E+04,
     + 3.0759E+04, 3.0774E+04, 3.0790E+04, 3.0805E+04, 3.0821E+04,
     +  3.0836E+04, 3.0852E+04, 3.0867E+04, 3.0883E+04, 3.0898E+04,
     + 3.0914E+04, 3.0929E+04, 3.0945E+04, 3.0961E+04, 3.0976E+04,
     +  3.0992E+04, 3.1007E+04, 3.1023E+04, 3.1038E+04, 3.1054E+04,
     + 3.1069E+04, 3.1085E+04, 3.1101E+04, 3.1116E+04, 3.1132E+04,
     +  3.1147E+04, 3.1163E+04, 3.1179E+04, 3.1194E+04, 3.1210E+04,
     + 3.1225E+04, 3.1241E+04, 3.1256E+04, 3.1272E+04, 3.1288E+04,
     +  3.1303E+04, 3.1319E+04, 3.1334E+04, 3.1350E+04, 3.1366E+04,
     + 3.1381E+04, 3.1397E+04, 3.1413E+04, 3.1428E+04, 3.1444E+04,
     +  3.1459E+04, 3.1475E+04, 3.1491E+04, 3.1506E+04, 3.1522E+04,
     + 3.1538E+04, 3.1553E+04, 3.1569E+04, 3.1584E+04, 3.1600E+04,
     +  3.1616E+04, 3.1631E+04, 3.1647E+04, 3.1663E+04, 3.1678E+04,
     + 3.1694E+04, 3.1710E+04, 3.1725E+04, 3.1741E+04, 3.1757E+04,
     +  3.1772E+04, 3.1788E+04, 3.1804E+04, 3.1819E+04, 3.1835E+04,
     + 3.1851E+04, 3.1866E+04, 3.1882E+04, 3.1898E+04, 3.1913E+04,
     +  3.1929E+04, 3.1945E+04, 3.1960E+04, 3.1976E+04, 3.1992E+04,
     + 3.2007E+04, 3.2023E+04, 3.2039E+04, 3.2054E+04, 3.2070E+04,
     +  3.2086E+04, 3.2102E+04, 3.2117E+04, 3.2133E+04, 3.2149E+04,
     + 3.2164E+04, 3.2180E+04, 3.2196E+04, 3.2212E+04, 3.2227E+04,
     +  3.2243E+04, 3.2259E+04, 3.2274E+04, 3.2290E+04, 3.2306E+04,
     + 3.2322E+04, 3.2337E+04, 3.2353E+04, 3.2369E+04, 3.2385E+04,
     +  3.2400E+04, 3.2416E+04, 3.2432E+04, 3.2448E+04, 3.2463E+04,
     + 3.2479E+04, 3.2495E+04, 3.2511E+04, 3.2526E+04, 3.2542E+04,
     +  3.2558E+04, 3.2574E+04, 3.2589E+04, 3.2605E+04, 3.2621E+04,
     + 3.2637E+04, 3.2652E+04, 3.2668E+04, 3.2684E+04, 3.2700E+04,
     +  3.2716E+04, 3.2731E+04, 3.2747E+04, 3.2763E+04, 3.2779E+04,
     + 3.2794E+04, 3.2810E+04, 3.2826E+04, 3.2842E+04, 3.2858E+04,
     +  3.2873E+04, 3.2889E+04, 3.2905E+04, 3.2921E+04, 3.2937E+04,
     + 3.2952E+04, 3.2968E+04, 3.2984E+04, 3.3000E+04, 3.3016E+04,
     +  3.3031E+04, 3.3047E+04, 3.3063E+04, 3.3079E+04, 3.3095E+04,
     + 3.3111E+04, 3.3126E+04, 3.3142E+04, 3.3158E+04, 3.3174E+04,
     +  3.3190E+04, 3.3206E+04, 3.3221E+04, 3.3237E+04, 3.3253E+04,
     + 3.3269E+04, 3.3285E+04, 3.3301E+04, 3.3317E+04, 3.3332E+04,
     +  3.3348E+04, 3.3364E+04, 3.3380E+04, 3.3396E+04, 3.3412E+04,
     + 3.3428E+04, 3.3443E+04, 3.3459E+04, 3.3475E+04, 3.3491E+04,
     +  3.3507E+04, 3.3523E+04, 3.3539E+04, 3.3555E+04, 3.3570E+04,
     + 3.3586E+04, 3.3602E+04, 3.3618E+04, 3.3634E+04, 3.3650E+04,
     +  3.3666E+04, 3.3682E+04, 3.3698E+04, 3.3713E+04, 3.3729E+04,
     + 3.3745E+04, 3.3761E+04, 3.3777E+04, 3.3793E+04, 3.3809E+04,
     +  3.3825E+04, 3.3841E+04, 3.3857E+04, 3.3872E+04, 3.3888E+04,
     + 3.3904E+04, 3.3920E+04, 3.3936E+04, 3.3952E+04, 3.3968E+04,
     +  3.3984E+04, 3.4000E+04, 3.4016E+04, 3.4032E+04, 3.4048E+04,
     + 3.4064E+04, 3.4080E+04, 3.4096E+04, 3.4111E+04, 3.4127E+04,
     +  3.4143E+04, 3.4159E+04, 3.4175E+04, 3.4191E+04, 3.4207E+04,
     + 3.4223E+04, 3.4239E+04, 3.4255E+04, 3.4271E+04, 3.4287E+04,
     +  3.4303E+04, 3.4319E+04, 3.4335E+04, 3.4351E+04, 3.4367E+04,
     + 3.4383E+04, 3.4399E+04, 3.4415E+04, 3.4431E+04, 3.4447E+04,
     +  3.4463E+04, 3.4479E+04, 3.4495E+04, 3.4511E+04, 3.4527E+04,
     + 3.4543E+04, 3.4559E+04, 3.4575E+04, 3.4591E+04, 3.4607E+04,
     +  3.4623E+04, 3.4639E+04, 3.4655E+04, 3.4671E+04, 3.4687E+04,
     + 3.4703E+04, 3.4719E+04, 3.4735E+04, 3.4751E+04, 3.4767E+04,
     +  3.4783E+04, 3.4799E+04, 3.4815E+04, 3.4831E+04, 3.4847E+04,
     + 3.4863E+04, 3.4879E+04, 3.4895E+04, 3.4911E+04, 3.4927E+04,
     +  3.4943E+04, 3.4959E+04, 3.4975E+04, 3.4991E+04, 3.5007E+04,
     + 3.5023E+04, 3.5039E+04, 3.5055E+04, 3.5071E+04, 3.5087E+04,
     +  3.5103E+04, 3.5119E+04, 3.5136E+04, 3.5152E+04, 3.5168E+04,
     + 3.5184E+04, 3.5200E+04, 3.5216E+04, 3.5232E+04, 3.5248E+04,
     +  3.5264E+04, 3.5280E+04, 3.5296E+04, 3.5312E+04, 3.5328E+04,
     + 3.5344E+04, 3.5361E+04, 3.5377E+04, 3.5393E+04, 3.5409E+04,
     +  3.5425E+04, 3.5441E+04, 3.5457E+04, 3.5473E+04, 3.5489E+04,
     + 3.5505E+04, 3.5521E+04, 3.5538E+04, 3.5554E+04, 3.5570E+04,
     +  3.5586E+04, 3.5602E+04, 3.5618E+04, 3.5634E+04, 3.5650E+04,
     + 3.5666E+04, 3.5682E+04, 3.5699E+04, 3.5715E+04, 3.5731E+04,
     +  3.5747E+04, 3.5763E+04, 3.5779E+04, 3.5795E+04, 3.5811E+04,
     + 3.5828E+04, 3.5844E+04, 3.5860E+04, 3.5876E+04, 3.5892E+04,
     +  3.5908E+04, 3.5924E+04, 3.5941E+04, 3.5957E+04, 3.5973E+04,
     + 3.5989E+04, 3.6005E+04, 3.6021E+04, 3.6037E+04, 3.6054E+04,
     +  3.6070E+04, 3.6086E+04, 3.6102E+04, 3.6118E+04, 3.6134E+04,
     + 3.6151E+04, 3.6167E+04, 3.6183E+04, 3.6199E+04, 3.6215E+04,
     +  3.6231E+04, 3.6248E+04, 3.6264E+04, 3.6280E+04, 3.6296E+04,
     + 3.6312E+04, 3.6328E+04, 3.6345E+04, 3.6361E+04, 3.6377E+04,
     +  3.6393E+04, 3.6409E+04, 3.6426E+04, 3.6442E+04, 3.6458E+04,
     + 3.6474E+04, 3.6490E+04, 3.6507E+04, 3.6523E+04, 3.6539E+04,
     +  3.6555E+04, 3.6571E+04, 3.6588E+04, 3.6604E+04, 3.6620E+04,
     + 3.6636E+04, 3.6652E+04, 3.6669E+04, 3.6685E+04, 3.6701E+04,
     +  3.6717E+04, 3.6733E+04, 3.6750E+04, 3.6766E+04, 3.6782E+04,
     + 3.6798E+04, 3.6815E+04, 3.6831E+04, 3.6847E+04, 3.6863E+04,
     +  3.6880E+04, 3.6896E+04, 3.6912E+04, 3.6928E+04, 3.6944E+04,
     + 3.6961E+04, 3.6977E+04, 3.6993E+04, 3.7009E+04, 3.7026E+04,
     +  3.7042E+04, 3.7058E+04, 3.7074E+04, 3.7091E+04, 3.7107E+04,
     + 3.7123E+04, 3.7140E+04, 3.7156E+04, 3.7172E+04, 3.7188E+04,
     +  3.7205E+04, 3.7221E+04, 3.7237E+04, 3.7253E+04, 3.7270E+04,
     + 3.7286E+04, 3.7302E+04, 3.7318E+04, 3.7335E+04, 3.7351E+04,
     +  3.7367E+04, 3.7384E+04, 3.7400E+04, 3.7416E+04, 3.7432E+04,
     + 3.7449E+04, 3.7465E+04, 3.7481E+04, 3.7498E+04, 3.7514E+04,
     +  3.7530E+04, 3.7547E+04, 3.7563E+04, 3.7579E+04, 3.7595E+04,
     + 3.7612E+04, 3.7628E+04, 3.7644E+04, 3.7661E+04, 3.7677E+04,
     +  3.7693E+04, 3.7710E+04, 3.7726E+04, 3.7742E+04, 3.7759E+04,
     + 3.7775E+04, 3.7791E+04, 3.7808E+04, 3.7824E+04, 3.7840E+04,
     +  3.7857E+04, 3.7873E+04, 3.7889E+04, 3.7906E+04, 3.7922E+04,
     + 3.7938E+04, 3.7955E+04, 3.7971E+04, 3.7987E+04, 3.8004E+04,
     +  3.8020E+04, 3.8036E+04, 3.8053E+04, 3.8069E+04, 3.8085E+04,
     + 3.8102E+04, 3.8118E+04, 3.8134E+04, 3.8151E+04, 3.8167E+04,
     +  3.8183E+04, 3.8200E+04, 3.8216E+04, 3.8233E+04, 3.8249E+04,
     + 3.8265E+04, 3.8282E+04, 3.8298E+04, 3.8314E+04, 3.8331E+04,
     +  3.8347E+04, 3.8363E+04, 3.8380E+04, 3.8396E+04, 3.8413E+04,
     + 3.8429E+04, 3.8445E+04, 3.8462E+04, 3.8478E+04, 3.8495E+04,
     +  3.8511E+04, 3.8527E+04, 3.8544E+04, 3.8560E+04, 3.8576E+04,
     + 3.8593E+04, 3.8609E+04, 3.8626E+04, 3.8642E+04, 3.8658E+04,
     +  3.8675E+04, 3.8691E+04, 3.8708E+04, 3.8724E+04, 3.8740E+04,
     + 3.8757E+04, 3.8773E+04, 3.8790E+04, 3.8806E+04, 3.8823E+04,
     +  3.8839E+04, 3.8855E+04, 3.8872E+04, 3.8888E+04, 3.8905E+04,
     + 3.8921E+04, 3.8938E+04, 3.8954E+04, 3.8970E+04, 3.8987E+04,
     +  3.9003E+04, 3.9020E+04, 3.9036E+04, 3.9053E+04, 3.9069E+04,
     + 3.9085E+04, 3.9102E+04, 3.9118E+04, 3.9135E+04, 3.9151E+04,
     +  3.9168E+04, 3.9184E+04, 3.9201E+04, 3.9217E+04, 3.9233E+04,
     + 3.9250E+04, 3.9266E+04, 3.9283E+04, 3.9299E+04, 3.9316E+04,
     +  3.9332E+04, 3.9349E+04, 3.9365E+04, 3.9382E+04, 3.9398E+04,
     + 3.9414E+04, 3.9431E+04, 3.9447E+04, 3.9464E+04, 3.9480E+04,
     +  3.9497E+04, 3.9513E+04, 3.9530E+04, 3.9546E+04, 3.9563E+04,
     + 3.9579E+04, 3.9596E+04, 3.9612E+04, 3.9629E+04, 3.9645E+04,
     +  3.9662E+04, 3.9678E+04, 3.9695E+04, 3.9711E+04, 3.9728E+04,
     + 3.9744E+04, 3.9761E+04, 3.9777E+04, 3.9794E+04, 3.9810E+04,
     +  3.9827E+04, 3.9843E+04, 3.9860E+04, 3.9876E+04, 3.9893E+04,
     + 3.9909E+04, 3.9926E+04, 3.9942E+04, 3.9959E+04, 3.9975E+04,
     +  3.9992E+04, 4.0008E+04, 4.0025E+04, 4.0041E+04, 4.0058E+04,
     + 4.0074E+04, 4.0091E+04, 4.0107E+04, 4.0124E+04, 4.0140E+04,
     +  4.0157E+04, 4.0174E+04, 4.0190E+04, 4.0207E+04, 4.0223E+04,
     + 4.0240E+04, 4.0256E+04, 4.0273E+04, 4.0289E+04, 4.0306E+04,
     +  4.0322E+04, 4.0339E+04, 4.0355E+04, 4.0372E+04, 4.0389E+04,
     + 4.0405E+04, 4.0422E+04, 4.0438E+04, 4.0455E+04, 4.0471E+04,
     +  4.0488E+04, 4.0504E+04, 4.0521E+04, 4.0538E+04, 4.0554E+04,
     + 4.0571E+04, 4.0587E+04, 4.0604E+04, 4.0620E+04, 4.0637E+04,
     +  4.0653E+04, 4.0670E+04, 4.0687E+04, 4.0703E+04, 4.0720E+04,
     + 4.0736E+04, 4.0753E+04, 4.0770E+04, 4.0786E+04, 4.0803E+04,
     +  4.0819E+04, 4.0836E+04, 4.0852E+04, 4.0869E+04, 4.0886E+04,
     + 4.0902E+04, 4.0919E+04, 4.0935E+04, 4.0952E+04, 4.0969E+04,
     +  4.0985E+04, 4.1002E+04, 4.1018E+04, 4.1035E+04, 4.1052E+04,
     + 4.1068E+04, 4.1085E+04, 4.1101E+04, 4.1118E+04, 4.1135E+04,
     +  4.1151E+04, 4.1168E+04, 4.1184E+04, 4.1201E+04, 4.1218E+04,
     + 4.1234E+04, 4.1251E+04, 4.1267E+04, 4.1284E+04, 4.1301E+04,
     +  4.1317E+04, 4.1334E+04, 4.1351E+04, 4.1367E+04, 4.1384E+04,
     + 4.1400E+04, 4.1417E+04, 4.1434E+04, 4.1450E+04, 4.1467E+04,
     +  4.1484E+04, 4.1500E+04, 4.1517E+04, 4.1534E+04, 4.1550E+04,
     + 4.1567E+04, 4.1583E+04, 4.1600E+04, 4.1617E+04, 4.1633E+04,
     +  4.1650E+04, 4.1667E+04, 4.1683E+04, 4.1700E+04, 4.1717E+04,
     + 4.1733E+04, 4.1750E+04, 4.1767E+04, 4.1783E+04, 4.1800E+04,
     +  4.1817E+04, 4.1833E+04, 4.1850E+04, 4.1867E+04, 4.1883E+04,
     + 4.1900E+04, 4.1917E+04, 4.1933E+04, 4.1950E+04, 4.1967E+04,
     +  4.1983E+04, 4.2000E+04, 4.2017E+04, 4.2033E+04, 4.2050E+04,
     + 4.2067E+04, 4.2083E+04, 4.2100E+04, 4.2117E+04, 4.2133E+04,
     +  4.2150E+04, 4.2167E+04, 4.2183E+04, 4.2200E+04, 4.2217E+04,
     + 4.2233E+04, 4.2250E+04, 4.2267E+04, 4.2284E+04, 4.2300E+04,
     +  4.2317E+04, 4.2334E+04, 4.2350E+04, 4.2367E+04, 4.2384E+04,
     + 4.2400E+04, 4.2417E+04, 4.2434E+04, 4.2451E+04, 4.2467E+04,
     +  4.2484E+04, 4.2501E+04, 4.2517E+04, 4.2534E+04, 4.2551E+04,
     + 4.2568E+04, 4.2584E+04, 4.2601E+04, 4.2618E+04, 4.2634E+04,
     +  4.2651E+04, 4.2668E+04, 4.2685E+04, 4.2701E+04, 4.2718E+04,
     + 4.2735E+04, 4.2751E+04, 4.2768E+04, 4.2785E+04, 4.2802E+04,
     +  4.2818E+04, 4.2835E+04, 4.2852E+04, 4.2869E+04, 4.2885E+04,
     + 4.2902E+04, 4.2919E+04, 4.2936E+04, 4.2952E+04, 4.2969E+04,
     +  4.2986E+04, 4.3002E+04, 4.3019E+04, 4.3036E+04, 4.3053E+04,
     + 4.3069E+04, 4.3086E+04, 4.3103E+04, 4.3120E+04, 4.3136E+04,
     +  4.3153E+04, 4.3170E+04, 4.3187E+04, 4.3204E+04, 4.3220E+04,
     + 4.3237E+04, 4.3254E+04, 4.3271E+04, 4.3287E+04, 4.3304E+04,
     +  4.3321E+04, 4.3338E+04, 4.3354E+04, 4.3371E+04, 4.3388E+04,
     + 4.3405E+04, 4.3421E+04, 4.3438E+04, 4.3455E+04, 4.3472E+04,
     +  4.3489E+04, 4.3505E+04, 4.3522E+04, 4.3539E+04, 4.3556E+04,
     + 4.3572E+04, 4.3589E+04, 4.3606E+04, 4.3623E+04, 4.3640E+04,
     +  4.3656E+04, 4.3673E+04, 4.3690E+04, 4.3707E+04, 4.3724E+04,
     + 4.3740E+04, 4.3757E+04, 4.3774E+04, 4.3791E+04, 4.3808E+04,
     +  4.3824E+04, 4.3841E+04, 4.3858E+04, 4.3875E+04, 4.3892E+04,
     + 4.3908E+04, 4.3925E+04, 4.3942E+04, 4.3959E+04, 4.3976E+04,
     +  4.3992E+04, 4.4009E+04, 4.4026E+04, 4.4043E+04, 4.4060E+04,
     + 4.4077E+04, 4.4093E+04, 4.4110E+04, 4.4127E+04, 4.4144E+04,
     +  4.4161E+04, 4.4177E+04, 4.4194E+04, 4.4211E+04, 4.4228E+04,
     + 4.4245E+04, 4.4262E+04, 4.4278E+04, 4.4295E+04, 4.4312E+04,
     +  4.4329E+04, 4.4346E+04, 4.4363E+04, 4.4379E+04, 4.4396E+04,
     + 4.4413E+04, 4.4430E+04, 4.4447E+04, 4.4464E+04, 4.4480E+04,
     +  4.4497E+04, 4.4514E+04, 4.4531E+04, 4.4548E+04, 4.4565E+04,
     + 4.4582E+04, 4.4598E+04, 4.4615E+04, 4.4632E+04, 4.4649E+04,
     +  4.4666E+04, 4.4683E+04, 4.4700E+04, 4.4716E+04, 4.4733E+04,
     + 4.4750E+04, 4.4767E+04, 4.4784E+04, 4.4801E+04, 4.4818E+04,
     +  4.4834E+04, 4.4851E+04, 4.4868E+04, 4.4885E+04, 4.4902E+04,
     + 4.4919E+04, 4.4936E+04, 4.4953E+04, 4.4969E+04, 4.4986E+04,
     +  4.5003E+04, 4.5020E+04, 4.5037E+04, 4.5054E+04, 4.5071E+04,
     + 4.5088E+04, 4.5104E+04, 4.5121E+04, 4.5138E+04, 4.5155E+04,
     +  4.5172E+04, 4.5189E+04, 4.5206E+04, 4.5223E+04, 4.5240E+04,
     + 4.5256E+04, 4.5273E+04, 4.5290E+04, 4.5307E+04, 4.5324E+04,
     +  4.5341E+04, 4.5358E+04, 4.5375E+04, 4.5392E+04, 4.5409E+04,
     + 4.5425E+04, 4.5442E+04, 4.5459E+04, 4.5476E+04, 4.5493E+04,
     +  4.5510E+04, 4.5527E+04, 4.5544E+04, 4.5561E+04, 4.5578E+04,
     + 4.5595E+04, 4.5611E+04, 4.5628E+04, 4.5645E+04, 4.5662E+04,
     +  4.5679E+04, 4.5696E+04, 4.5713E+04, 4.5730E+04, 4.5747E+04,
     + 4.5764E+04, 4.5781E+04, 4.5798E+04, 4.5814E+04, 4.5831E+04,
     +  4.5848E+04, 4.5865E+04, 4.5882E+04, 4.5899E+04, 4.5916E+04,
     + 4.5933E+04, 4.5950E+04, 4.5967E+04, 4.5984E+04, 4.6001E+04,
     +  4.6018E+04, 4.6035E+04, 4.6052E+04, 4.6068E+04, 4.6085E+04,
     + 4.6102E+04, 4.6119E+04, 4.6136E+04, 4.6153E+04, 4.6170E+04,
     +  4.6187E+04, 4.6204E+04, 4.6221E+04, 4.6238E+04, 4.6255E+04,
     + 4.6272E+04, 4.6289E+04, 4.6306E+04, 4.6323E+04, 4.6340E+04,
     +  4.6357E+04, 4.6374E+04, 4.6391E+04, 4.6408E+04, 4.6424E+04,
     + 4.6441E+04, 4.6458E+04, 4.6475E+04, 4.6492E+04, 4.6509E+04,
     +  4.6526E+04, 4.6543E+04, 4.6560E+04, 4.6577E+04, 4.6594E+04,
     + 4.6611E+04, 4.6628E+04, 4.6645E+04, 4.6662E+04, 4.6679E+04,
     +  4.6696E+04, 4.6713E+04, 4.6730E+04, 4.6747E+04, 4.6764E+04,
     + 4.6781E+04, 4.6798E+04, 4.6815E+04, 4.6832E+04, 4.6849E+04,
     +  4.6866E+04, 4.6883E+04, 4.6900E+04, 4.6917E+04, 4.6934E+04,
     + 4.6951E+04, 4.6968E+04, 4.6985E+04, 4.7002E+04, 4.7019E+04,
     +  4.7036E+04, 4.7053E+04, 4.7070E+04, 4.7087E+04, 4.7104E+04,
     + 4.7121E+04, 4.7138E+04, 4.7155E+04, 4.7172E+04, 4.7189E+04,
     +  4.7206E+04, 4.7223E+04, 4.7240E+04, 4.7257E+04, 4.7274E+04,
     + 4.7291E+04, 4.7308E+04, 4.7325E+04, 4.7342E+04, 4.7359E+04,
     +  4.7376E+04, 4.7393E+04, 4.7410E+04, 4.7427E+04, 4.7444E+04,
     + 4.7461E+04, 4.7478E+04, 4.7495E+04, 4.7512E+04, 4.7529E+04,
     +  4.7546E+04, 4.7563E+04, 4.7580E+04, 4.7597E+04, 4.7614E+04,
     + 4.7631E+04, 4.7648E+04, 4.7665E+04, 4.7682E+04, 4.7699E+04,
     +  4.7716E+04, 4.7733E+04, 4.7750E+04, 4.7767E+04, 4.7784E+04,
     + 4.7801E+04, 4.7819E+04, 4.7836E+04, 4.7853E+04, 4.7870E+04,
     +  4.7887E+04, 4.7904E+04, 4.7921E+04, 4.7938E+04, 4.7955E+04,
     + 4.7972E+04, 4.7989E+04, 4.8006E+04, 4.8023E+04, 4.8040E+04,
     +  4.8057E+04, 4.8074E+04, 4.8091E+04, 4.8108E+04, 4.8125E+04,
     + 4.8142E+04, 4.8159E+04, 4.8177E+04, 4.8194E+04, 4.8211E+04,
     +  4.8228E+04, 4.8245E+04, 4.8262E+04, 4.8279E+04, 4.8296E+04,
     + 4.8313E+04, 4.8330E+04, 4.8347E+04, 4.8364E+04, 4.8381E+04,
     +  4.8398E+04, 4.8415E+04, 4.8432E+04, 4.8450E+04, 4.8467E+04,
     + 4.8484E+04, 4.8501E+04, 4.8518E+04, 4.8535E+04, 4.8552E+04,
     +  4.8569E+04, 4.8586E+04, 4.8603E+04, 4.8620E+04, 4.8637E+04,
     + 4.8654E+04, 4.8672E+04, 4.8689E+04, 4.8706E+04, 4.8723E+04,
     +  4.8740E+04, 4.8757E+04, 4.8774E+04, 4.8791E+04, 4.8808E+04,
     + 4.8825E+04, 4.8842E+04, 4.8859E+04, 4.8877E+04, 4.8894E+04,
     +  4.8911E+04, 4.8928E+04, 4.8945E+04, 4.8962E+04, 4.8979E+04,
     + 4.8996E+04, 4.9013E+04, 4.9030E+04, 4.9048E+04, 4.9065E+04,
     +  4.9082E+04, 4.9099E+04, 4.9116E+04, 4.9133E+04, 4.9150E+04,
     + 4.9167E+04, 4.9184E+04, 4.9201E+04, 4.9219E+04, 4.9236E+04,
     +  4.9253E+04, 4.9270E+04, 4.9287E+04, 4.9304E+04, 4.9321E+04,
     + 4.9338E+04, 4.9355E+04, 4.9373E+04, 4.9390E+04, 4.9407E+04,
     +  4.9424E+04, 4.9441E+04, 4.9458E+04, 4.9475E+04, 4.9492E+04,
     + 4.9510E+04, 4.9527E+04, 4.9544E+04, 4.9561E+04, 4.9578E+04,
     +  4.9595E+04, 4.9612E+04, 4.9629E+04, 4.9647E+04, 4.9664E+04,
     + 4.9681E+04, 4.9698E+04, 4.9715E+04, 4.9732E+04, 4.9749E+04,
     +  4.9766E+04, 4.9784E+04, 4.9801E+04, 4.9818E+04, 4.9835E+04,
     + 4.9852E+04, 4.9869E+04, 4.9886E+04, 4.9904E+04, 4.9921E+04,
     +  4.9938E+04, 4.9955E+04, 4.9972E+04, 4.9989E+04, 5.0006E+04,
     + 5.0024E+04, 5.0041E+04, 5.0058E+04, 5.0075E+04, 5.0092E+04,
     +  5.0109E+04, 5.0126E+04, 5.0144E+04, 5.0161E+04, 5.0178E+04,
     + 5.0195E+04, 5.0212E+04, 5.0229E+04, 5.0247E+04, 5.0264E+04,
     +  5.0281E+04, 5.0298E+04, 5.0315E+04, 5.0332E+04, 5.0349E+04,
     + 5.0367E+04, 5.0384E+04, 5.0401E+04, 5.0418E+04, 5.0435E+04,
     +  5.0452E+04, 5.0470E+04, 5.0487E+04, 5.0504E+04, 5.0521E+04,
     + 5.0538E+04, 5.0555E+04, 5.0573E+04, 5.0590E+04, 5.0607E+04,
     +  5.0624E+04, 5.0641E+04, 5.0658E+04, 5.0676E+04, 5.0693E+04,
     + 5.0710E+04, 5.0727E+04, 5.0744E+04, 5.0762E+04, 5.0779E+04,
     +  5.0796E+04, 5.0813E+04, 5.0830E+04, 5.0847E+04, 5.0865E+04,
     + 5.0882E+04, 5.0899E+04, 5.0916E+04, 5.0933E+04, 5.0951E+04,
     +  5.0968E+04, 5.0985E+04, 5.1002E+04, 5.1019E+04, 5.1036E+04,
     + 5.1054E+04, 5.1071E+04, 5.1088E+04, 5.1105E+04, 5.1122E+04,
     +  5.1140E+04, 5.1157E+04, 5.1174E+04, 5.1191E+04, 5.1208E+04,
     + 5.1226E+04, 5.1243E+04, 5.1260E+04, 5.1277E+04, 5.1294E+04,
     +  5.1312E+04, 5.1329E+04, 5.1346E+04, 5.1363E+04, 5.1380E+04,
     + 5.1398E+04, 5.1415E+04, 5.1432E+04, 5.1449E+04, 5.1466E+04,
     +  5.1484E+04, 5.1501E+04, 5.1518E+04, 5.1535E+04, 5.1553E+04,
     + 5.1570E+04, 5.1587E+04, 5.1604E+04, 5.1621E+04, 5.1639E+04,
     +  5.1656E+04, 5.1673E+04, 5.1690E+04, 5.1707E+04, 5.1725E+04,
     + 5.1742E+04, 5.1759E+04, 5.1776E+04, 5.1794E+04, 5.1811E+04,
     +  5.1828E+04, 5.1845E+04, 5.1862E+04, 5.1880E+04, 5.1897E+04,
     + 5.1914E+04, 5.1931E+04, 5.1949E+04, 5.1966E+04, 5.1983E+04,
     +  5.2000E+04, 5.2018E+04, 5.2035E+04, 5.2052E+04, 5.2069E+04,
     + 5.2086E+04, 5.2104E+04, 5.2121E+04, 5.2138E+04, 5.2155E+04,
     +  5.2173E+04, 5.2190E+04, 5.2207E+04, 5.2224E+04, 5.2242E+04,
     + 5.2259E+04, 5.2276E+04, 5.2293E+04, 5.2311E+04, 5.2328E+04,
     +  5.2345E+04, 5.2362E+04, 5.2380E+04, 5.2397E+04, 5.2414E+04,
     + 5.2431E+04, 5.2449E+04, 5.2466E+04, 5.2483E+04, 5.2500E+04,
     +  5.2518E+04, 5.2535E+04, 5.2552E+04, 5.2569E+04, 5.2587E+04,
     + 5.2604E+04, 5.2621E+04, 5.2638E+04, 5.2656E+04, 5.2673E+04,
     +  5.2690E+04, 5.2707E+04, 5.2725E+04, 5.2742E+04, 5.2759E+04,
     + 5.2776E+04, 5.2794E+04, 5.2811E+04, 5.2828E+04, 5.2845E+04,
     +  5.2863E+04, 5.2880E+04, 5.2897E+04, 5.2914E+04, 5.2932E+04,
     + 5.2949E+04, 5.2966E+04, 5.2984E+04, 5.3001E+04, 5.3018E+04,
     +  5.3035E+04, 5.3053E+04, 5.3070E+04, 5.3087E+04, 5.3104E+04,
     + 5.3122E+04, 5.3139E+04, 5.3156E+04, 5.3173E+04, 5.3191E+04,
     +  5.3208E+04, 5.3225E+04, 5.3243E+04, 5.3260E+04, 5.3277E+04,
     + 5.3294E+04, 5.3312E+04, 5.3329E+04, 5.3346E+04, 5.3364E+04,
     +  5.3381E+04, 5.3398E+04, 5.3415E+04, 5.3433E+04, 5.3450E+04,
     + 5.3467E+04, 5.3485E+04, 5.3502E+04, 5.3519E+04, 5.3536E+04,
     +  5.3554E+04, 5.3571E+04, 5.3588E+04, 5.3606E+04, 5.3623E+04,
     + 5.3640E+04, 5.3657E+04, 5.3675E+04, 5.3692E+04, 5.3709E+04,
     +  5.3727E+04, 5.3744E+04, 5.3761E+04, 5.3779E+04, 5.3796E+04,
     + 5.3813E+04, 5.3830E+04, 5.3848E+04, 5.3865E+04, 5.3882E+04,
     +  5.3900E+04, 5.3917E+04, 5.3934E+04, 5.3952E+04, 5.3969E+04,
     + 5.3986E+04, 5.4003E+04, 5.4021E+04, 5.4038E+04, 5.4055E+04,
     +  5.4073E+04, 5.4090E+04, 5.4107E+04, 5.4125E+04, 5.4142E+04,
     + 5.4159E+04, 5.4177E+04, 5.4194E+04, 5.4211E+04/
c...        --        55
      data (QofT( 3,J),J=1,2408)/ 1.0429E+00, 1.6233E+00, 2.5394E+00,
     +  3.4629E+00, 4.3269E+00, 5.1394E+00, 5.9194E+00, 6.6814E+00,
     + 7.4344E+00, 8.1833E+00, 8.9303E+00, 9.6767E+00, 1.0423E+01,
     +  1.1170E+01, 1.1916E+01, 1.2663E+01, 1.3410E+01, 1.4157E+01,
     + 1.4904E+01, 1.5652E+01, 1.7147E+01, 1.8642E+01, 2.0138E+01,
     +  2.1633E+01, 2.3129E+01, 2.4625E+01, 2.6121E+01, 2.7618E+01,
     + 2.9114E+01, 3.0610E+01, 3.2107E+01, 3.3603E+01, 3.5100E+01,
     +  3.6596E+01, 3.8093E+01, 3.9590E+01, 4.1086E+01, 4.2583E+01,
     + 4.4080E+01, 4.5576E+01, 4.7073E+01, 4.8570E+01, 5.0067E+01,
     +  5.1564E+01, 5.3061E+01, 5.4558E+01, 5.6054E+01, 5.7551E+01,
     + 5.9048E+01, 6.0545E+01, 6.2042E+01, 6.3539E+01, 6.5036E+01,
     +  6.6533E+01, 6.8031E+01, 6.9528E+01, 7.1025E+01, 7.2522E+01,
     + 7.4019E+01, 7.5516E+01, 7.7013E+01, 7.8510E+01, 8.0008E+01,
     +  8.1505E+01, 8.3002E+01, 8.4499E+01, 8.5996E+01, 8.7494E+01,
     + 8.8991E+01, 9.0488E+01, 9.1986E+01, 9.3483E+01, 9.4980E+01,
     +  9.6478E+01, 9.7975E+01, 9.9472E+01, 1.0097E+02, 1.0247E+02,
     + 1.0396E+02, 1.0546E+02, 1.0696E+02, 1.0846E+02, 1.0995E+02,
     +  1.1145E+02, 1.1295E+02, 1.1445E+02, 1.1594E+02, 1.1744E+02,
     + 1.1894E+02, 1.2044E+02, 1.2193E+02, 1.2343E+02, 1.2493E+02,
     +  1.2643E+02, 1.2792E+02, 1.2942E+02, 1.3092E+02, 1.3242E+02,
     + 1.3392E+02, 1.3541E+02, 1.3691E+02, 1.3841E+02, 1.3991E+02,
     +  1.4140E+02, 1.4290E+02, 1.4440E+02, 1.4590E+02, 1.4739E+02,
     + 1.4889E+02, 1.5039E+02, 1.5189E+02, 1.5339E+02, 1.5488E+02,
     +  1.5638E+02, 1.5788E+02, 1.5938E+02, 1.6088E+02, 1.6237E+02,
     + 1.6387E+02, 1.6537E+02, 1.6687E+02, 1.6836E+02, 1.6986E+02,
     +  1.7136E+02, 1.7286E+02, 1.7436E+02, 1.7586E+02, 1.7735E+02,
     + 1.7885E+02, 1.8035E+02, 1.8185E+02, 1.8335E+02, 1.8484E+02,
     +  1.8634E+02, 1.8784E+02, 1.8934E+02, 1.9084E+02, 1.9233E+02,
     + 1.9383E+02, 1.9533E+02, 1.9683E+02, 1.9833E+02, 1.9983E+02,
     +  2.0132E+02, 2.0282E+02, 2.0432E+02, 2.0582E+02, 2.0732E+02,
     + 2.0882E+02, 2.1031E+02, 2.1181E+02, 2.1331E+02, 2.1481E+02,
     +  2.1631E+02, 2.1781E+02, 2.1931E+02, 2.2080E+02, 2.2230E+02,
     + 2.2380E+02, 2.2530E+02, 2.2680E+02, 2.2830E+02, 2.2980E+02,
     +  2.3130E+02, 2.3280E+02, 2.3429E+02, 2.3579E+02, 2.3729E+02,
     + 2.3879E+02, 2.4029E+02, 2.4179E+02, 2.4329E+02, 2.4479E+02,
     +  2.4629E+02, 2.4779E+02, 2.4929E+02, 2.5079E+02, 2.5229E+02,
     + 2.5379E+02, 2.5529E+02, 2.5679E+02, 2.5829E+02, 2.5979E+02,
     +  2.6129E+02, 2.6279E+02, 2.6429E+02, 2.6579E+02, 2.6729E+02,
     + 2.6879E+02, 2.7029E+02, 2.7179E+02, 2.7329E+02, 2.7479E+02,
     +  2.7629E+02, 2.7779E+02, 2.7929E+02, 2.8080E+02, 2.8230E+02,
     + 2.8380E+02, 2.8530E+02, 2.8680E+02, 2.8830E+02, 2.8981E+02,
     +  2.9131E+02, 2.9281E+02, 2.9431E+02, 2.9582E+02, 2.9732E+02,
     + 2.9882E+02, 3.0032E+02, 3.0183E+02, 3.0333E+02, 3.0483E+02,
     +  3.0634E+02, 3.0784E+02, 3.0935E+02, 3.1085E+02, 3.1236E+02,
     + 3.1386E+02, 3.1537E+02, 3.1687E+02, 3.1838E+02, 3.1988E+02,
     +  3.2139E+02, 3.2289E+02, 3.2440E+02, 3.2591E+02, 3.2741E+02,
     + 3.2892E+02, 3.3043E+02, 3.3193E+02, 3.3344E+02, 3.3495E+02,
     +  3.3646E+02, 3.3797E+02, 3.3948E+02, 3.4098E+02, 3.4249E+02,
     + 3.4400E+02, 3.4551E+02, 3.4702E+02, 3.4853E+02, 3.5004E+02,
     +  3.5156E+02, 3.5307E+02, 3.5458E+02, 3.5609E+02, 3.5760E+02,
     + 3.5912E+02, 3.6063E+02, 3.6214E+02, 3.6366E+02, 3.6517E+02,
     +  3.6669E+02, 3.6820E+02, 3.6972E+02, 3.7123E+02, 3.7275E+02,
     + 3.7426E+02, 3.7578E+02, 3.7730E+02, 3.7882E+02, 3.8033E+02,
     +  3.8185E+02, 3.8337E+02, 3.8489E+02, 3.8641E+02, 3.8793E+02,
     + 3.8945E+02, 3.9097E+02, 3.9249E+02, 3.9402E+02, 3.9554E+02,
     +  3.9706E+02, 3.9859E+02, 4.0011E+02, 4.0163E+02, 4.0316E+02,
     + 4.0468E+02, 4.0621E+02, 4.0774E+02, 4.0926E+02, 4.1079E+02,
     +  4.1232E+02, 4.1385E+02, 4.1538E+02, 4.1691E+02, 4.1844E+02,
     + 4.1997E+02, 4.2150E+02, 4.2303E+02, 4.2456E+02, 4.2610E+02,
     +  4.2763E+02, 4.2917E+02, 4.3070E+02, 4.3224E+02, 4.3377E+02,
     + 4.3531E+02, 4.3685E+02, 4.3839E+02, 4.3992E+02, 4.4146E+02,
     +  4.4300E+02, 4.4454E+02, 4.4609E+02, 4.4763E+02, 4.4917E+02,
     + 4.5071E+02, 4.5226E+02, 4.5380E+02, 4.5535E+02, 4.5689E+02,
     +  4.5844E+02, 4.5999E+02, 4.6154E+02, 4.6309E+02, 4.6463E+02,
     + 4.6619E+02, 4.6774E+02, 4.6929E+02, 4.7084E+02, 4.7239E+02,
     +  4.7395E+02, 4.7550E+02, 4.7706E+02, 4.7862E+02, 4.8017E+02,
     + 4.8173E+02, 4.8329E+02, 4.8485E+02, 4.8641E+02, 4.8797E+02,
     +  4.8953E+02, 4.9110E+02, 4.9266E+02, 4.9423E+02, 4.9579E+02,
     + 4.9736E+02, 4.9892E+02, 5.0049E+02, 5.0206E+02, 5.0363E+02,
     +  5.0520E+02, 5.0677E+02, 5.0835E+02, 5.0992E+02, 5.1149E+02,
     + 5.1307E+02, 5.1464E+02, 5.1622E+02, 5.1780E+02, 5.1938E+02,
     +  5.2096E+02, 5.2254E+02, 5.2412E+02, 5.2570E+02, 5.2728E+02,
     + 5.2887E+02, 5.3045E+02, 5.3204E+02, 5.3363E+02, 5.3521E+02,
     +  5.3680E+02, 5.3839E+02, 5.3998E+02, 5.4158E+02, 5.4317E+02,
     + 5.4476E+02, 5.4636E+02, 5.4795E+02, 5.4955E+02, 5.5115E+02,
     +  5.5275E+02, 5.5435E+02, 5.5595E+02, 5.5755E+02, 5.5915E+02,
     + 5.6076E+02, 5.6236E+02, 5.6397E+02, 5.6558E+02, 5.6719E+02,
     +  5.6879E+02, 5.7041E+02, 5.7202E+02, 5.7363E+02, 5.7524E+02,
     + 5.7686E+02, 5.7847E+02, 5.8009E+02, 5.8171E+02, 5.8333E+02,
     +  5.8495E+02, 5.8657E+02, 5.8819E+02, 5.8982E+02, 5.9144E+02,
     + 5.9307E+02, 5.9469E+02, 5.9632E+02, 5.9795E+02, 5.9958E+02,
     +  6.0121E+02, 6.0285E+02, 6.0448E+02, 6.0612E+02, 6.0775E+02,
     + 6.0939E+02, 6.1103E+02, 6.1267E+02, 6.1431E+02, 6.1595E+02,
     +  6.1759E+02, 6.1924E+02, 6.2088E+02, 6.2253E+02, 6.2418E+02,
     + 6.2583E+02, 6.2748E+02, 6.2913E+02, 6.3078E+02, 6.3244E+02,
     +  6.3409E+02, 6.3575E+02, 6.3741E+02, 6.3907E+02, 6.4073E+02,
     + 6.4239E+02, 6.4405E+02, 6.4571E+02, 6.4738E+02, 6.4905E+02,
     +  6.5071E+02, 6.5238E+02, 6.5405E+02, 6.5573E+02, 6.5740E+02,
     + 6.5907E+02, 6.6075E+02, 6.6242E+02, 6.6410E+02, 6.6578E+02,
     +  6.6746E+02, 6.6915E+02, 6.7083E+02, 6.7251E+02, 6.7420E+02,
     + 6.7589E+02, 6.7757E+02, 6.7926E+02, 6.8096E+02, 6.8265E+02,
     +  6.8434E+02, 6.8604E+02, 6.8773E+02, 6.8943E+02, 6.9113E+02,
     + 6.9283E+02, 6.9453E+02, 6.9624E+02, 6.9794E+02, 6.9965E+02,
     +  7.0135E+02, 7.0306E+02, 7.0477E+02, 7.0648E+02, 7.0820E+02,
     + 7.0991E+02, 7.1162E+02, 7.1334E+02, 7.1506E+02, 7.1678E+02,
     +  7.1850E+02, 7.2022E+02, 7.2195E+02, 7.2367E+02, 7.2540E+02,
     + 7.2713E+02, 7.2886E+02, 7.3059E+02, 7.3232E+02, 7.3405E+02,
     +  7.3579E+02, 7.3752E+02, 7.3926E+02, 7.4100E+02, 7.4274E+02,
     + 7.4448E+02, 7.4623E+02, 7.4797E+02, 7.4972E+02, 7.5147E+02,
     +  7.5322E+02, 7.5497E+02, 7.5672E+02, 7.5847E+02, 7.6023E+02,
     + 7.6198E+02, 7.6374E+02, 7.6550E+02, 7.6726E+02, 7.6902E+02,
     +  7.7079E+02, 7.7255E+02, 7.7432E+02, 7.7609E+02, 7.7786E+02,
     + 7.7963E+02, 7.8140E+02, 7.8318E+02, 7.8495E+02, 7.8673E+02,
     +  7.8851E+02, 7.9029E+02, 7.9207E+02, 7.9385E+02, 7.9564E+02,
     + 7.9742E+02, 7.9921E+02, 8.0100E+02, 8.0279E+02, 8.0458E+02,
     +  8.0637E+02, 8.0817E+02, 8.0997E+02, 8.1176E+02, 8.1356E+02,
     + 8.1537E+02, 8.1717E+02, 8.1897E+02, 8.2078E+02, 8.2259E+02,
     +  8.2439E+02, 8.2620E+02, 8.2802E+02, 8.2983E+02, 8.3164E+02,
     + 8.3346E+02, 8.3528E+02, 8.3710E+02, 8.3892E+02, 8.4074E+02,
     +  8.4257E+02, 8.4439E+02, 8.4622E+02, 8.4805E+02, 8.4988E+02,
     + 8.5171E+02, 8.5354E+02, 8.5538E+02, 8.5722E+02, 8.5905E+02,
     +  8.6089E+02, 8.6274E+02, 8.6458E+02, 8.6642E+02, 8.6827E+02,
     + 8.7012E+02, 8.7197E+02, 8.7382E+02, 8.7567E+02, 8.7752E+02,
     +  8.7938E+02, 8.8124E+02, 8.8309E+02, 8.8496E+02, 8.8682E+02,
     + 8.8868E+02, 8.9055E+02, 8.9241E+02, 8.9428E+02, 8.9615E+02,
     +  8.9802E+02, 8.9990E+02, 9.0177E+02, 9.0365E+02, 9.0552E+02,
     + 9.0740E+02, 9.0928E+02, 9.1117E+02, 9.1305E+02, 9.1494E+02,
     +  9.1683E+02, 9.1871E+02, 9.2061E+02, 9.2250E+02, 9.2439E+02,
     + 9.2629E+02, 9.2819E+02, 9.3008E+02, 9.3199E+02, 9.3389E+02,
     +  9.3579E+02, 9.3770E+02, 9.3960E+02, 9.4151E+02, 9.4342E+02,
     + 9.4534E+02, 9.4725E+02, 9.4916E+02, 9.5108E+02, 9.5300E+02,
     +  9.5492E+02, 9.5684E+02, 9.5877E+02, 9.6069E+02, 9.6262E+02,
     + 9.6455E+02, 9.6648E+02, 9.6841E+02, 9.7034E+02, 9.7228E+02,
     +  9.7421E+02, 9.7615E+02, 9.7809E+02, 9.8003E+02, 9.8198E+02,
     + 9.8392E+02, 9.8587E+02, 9.8782E+02, 9.8977E+02, 9.9172E+02,
     +  9.9367E+02, 9.9563E+02, 9.9759E+02, 9.9954E+02, 1.0015E+03,
     + 1.0035E+03, 1.0054E+03, 1.0074E+03, 1.0094E+03, 1.0113E+03,
     +  1.0133E+03, 1.0153E+03, 1.0172E+03, 1.0192E+03, 1.0212E+03,
     + 1.0232E+03, 1.0252E+03, 1.0271E+03, 1.0291E+03, 1.0311E+03,
     +  1.0331E+03, 1.0351E+03, 1.0371E+03, 1.0391E+03, 1.0411E+03,
     + 1.0431E+03, 1.0451E+03, 1.0471E+03, 1.0491E+03, 1.0511E+03,
     +  1.0531E+03, 1.0551E+03, 1.0571E+03, 1.0591E+03, 1.0611E+03,
     + 1.0631E+03, 1.0651E+03, 1.0671E+03, 1.0692E+03, 1.0712E+03,
     +  1.0732E+03, 1.0752E+03, 1.0772E+03, 1.0793E+03, 1.0813E+03,
     + 1.0833E+03, 1.0854E+03, 1.0874E+03, 1.0894E+03, 1.0915E+03,
     +  1.0935E+03, 1.0955E+03, 1.0976E+03, 1.0996E+03, 1.1017E+03,
     + 1.1037E+03, 1.1058E+03, 1.1078E+03, 1.1099E+03, 1.1119E+03,
     +  1.1140E+03, 1.1160E+03, 1.1181E+03, 1.1202E+03, 1.1222E+03,
     + 1.1243E+03, 1.1264E+03, 1.1284E+03, 1.1305E+03, 1.1326E+03,
     +  1.1346E+03, 1.1367E+03, 1.1388E+03, 1.1409E+03, 1.1429E+03,
     + 1.1450E+03, 1.1471E+03, 1.1492E+03, 1.1513E+03, 1.1534E+03,
     +  1.1555E+03, 1.1576E+03, 1.1596E+03, 1.1617E+03, 1.1638E+03,
     + 1.1659E+03, 1.1680E+03, 1.1701E+03, 1.1722E+03, 1.1744E+03,
     +  1.1765E+03, 1.1786E+03, 1.1807E+03, 1.1828E+03, 1.1849E+03,
     + 1.1870E+03, 1.1891E+03, 1.1913E+03, 1.1934E+03, 1.1955E+03,
     +  1.1976E+03, 1.1998E+03, 1.2019E+03, 1.2040E+03, 1.2062E+03,
     + 1.2083E+03, 1.2104E+03, 1.2126E+03, 1.2147E+03, 1.2169E+03,
     +  1.2190E+03, 1.2211E+03, 1.2233E+03, 1.2254E+03, 1.2276E+03,
     + 1.2297E+03, 1.2319E+03, 1.2340E+03, 1.2362E+03, 1.2384E+03,
     +  1.2405E+03, 1.2427E+03, 1.2449E+03, 1.2470E+03, 1.2492E+03,
     + 1.2514E+03, 1.2535E+03, 1.2557E+03, 1.2579E+03, 1.2601E+03,
     +  1.2622E+03, 1.2644E+03, 1.2666E+03, 1.2688E+03, 1.2710E+03,
     + 1.2732E+03, 1.2753E+03, 1.2775E+03, 1.2797E+03, 1.2819E+03,
     +  1.2841E+03, 1.2863E+03, 1.2885E+03, 1.2907E+03, 1.2929E+03,
     + 1.2951E+03, 1.2973E+03, 1.2995E+03, 1.3018E+03, 1.3040E+03,
     +  1.3062E+03, 1.3084E+03, 1.3106E+03, 1.3128E+03, 1.3151E+03,
     + 1.3173E+03, 1.3195E+03, 1.3217E+03, 1.3240E+03, 1.3262E+03,
     +  1.3284E+03, 1.3307E+03, 1.3329E+03, 1.3351E+03, 1.3374E+03,
     + 1.3396E+03, 1.3419E+03, 1.3441E+03, 1.3464E+03, 1.3486E+03,
     +  1.3509E+03, 1.3531E+03, 1.3554E+03, 1.3576E+03, 1.3599E+03,
     + 1.3621E+03, 1.3644E+03, 1.3667E+03, 1.3689E+03, 1.3712E+03,
     +  1.3735E+03, 1.3757E+03, 1.3780E+03, 1.3803E+03, 1.3826E+03,
     + 1.3848E+03, 1.3871E+03, 1.3894E+03, 1.3917E+03, 1.3940E+03,
     +  1.3962E+03, 1.3985E+03, 1.4008E+03, 1.4031E+03, 1.4054E+03,
     + 1.4077E+03, 1.4100E+03, 1.4123E+03, 1.4146E+03, 1.4169E+03,
     +  1.4192E+03, 1.4215E+03, 1.4238E+03, 1.4261E+03, 1.4284E+03,
     + 1.4308E+03, 1.4331E+03, 1.4354E+03, 1.4377E+03, 1.4400E+03,
     +  1.4424E+03, 1.4447E+03, 1.4470E+03, 1.4493E+03, 1.4517E+03,
     + 1.4540E+03, 1.4563E+03, 1.4587E+03, 1.4610E+03, 1.4633E+03,
     +  1.4657E+03, 1.4680E+03, 1.4704E+03, 1.4727E+03, 1.4751E+03,
     + 1.4774E+03, 1.4798E+03, 1.4821E+03, 1.4845E+03, 1.4868E+03,
     +  1.4892E+03, 1.4916E+03, 1.4939E+03, 1.4963E+03, 1.4987E+03,
     + 1.5010E+03, 1.5034E+03, 1.5058E+03, 1.5081E+03, 1.5105E+03,
     +  1.5129E+03, 1.5153E+03, 1.5176E+03, 1.5200E+03, 1.5224E+03,
     + 1.5248E+03, 1.5272E+03, 1.5296E+03, 1.5320E+03, 1.5344E+03,
     +  1.5368E+03, 1.5392E+03, 1.5416E+03, 1.5440E+03, 1.5464E+03,
     + 1.5488E+03, 1.5512E+03, 1.5536E+03, 1.5560E+03, 1.5584E+03,
     +  1.5608E+03, 1.5632E+03, 1.5657E+03, 1.5681E+03, 1.5705E+03,
     + 1.5729E+03, 1.5753E+03, 1.5778E+03, 1.5802E+03, 1.5826E+03,
     +  1.5851E+03, 1.5875E+03, 1.5899E+03, 1.5924E+03, 1.5948E+03,
     + 1.5972E+03, 1.5997E+03, 1.6021E+03, 1.6046E+03, 1.6070E+03,
     +  1.6095E+03, 1.6119E+03, 1.6144E+03, 1.6168E+03, 1.6193E+03,
     + 1.6218E+03, 1.6242E+03, 1.6267E+03, 1.6292E+03, 1.6316E+03,
     +  1.6341E+03, 1.6366E+03, 1.6390E+03, 1.6415E+03, 1.6440E+03,
     + 1.6465E+03, 1.6489E+03, 1.6514E+03, 1.6539E+03, 1.6564E+03,
     +  1.6589E+03, 1.6614E+03, 1.6639E+03, 1.6664E+03, 1.6688E+03,
     + 1.6713E+03, 1.6738E+03, 1.6763E+03, 1.6788E+03, 1.6813E+03,
     +  1.6839E+03, 1.6864E+03, 1.6889E+03, 1.6914E+03, 1.6939E+03,
     + 1.6964E+03, 1.6989E+03, 1.7014E+03, 1.7040E+03, 1.7065E+03,
     +  1.7090E+03, 1.7115E+03, 1.7141E+03, 1.7166E+03, 1.7191E+03,
     + 1.7217E+03, 1.7242E+03, 1.7267E+03, 1.7293E+03, 1.7318E+03,
     +  1.7343E+03, 1.7369E+03, 1.7394E+03, 1.7420E+03, 1.7445E+03,
     + 1.7471E+03, 1.7496E+03, 1.7522E+03, 1.7547E+03, 1.7573E+03,
     +  1.7599E+03, 1.7624E+03, 1.7650E+03, 1.7676E+03, 1.7701E+03,
     + 1.7727E+03, 1.7753E+03, 1.7778E+03, 1.7804E+03, 1.7830E+03,
     +  1.7856E+03, 1.7882E+03, 1.7907E+03, 1.7933E+03, 1.7959E+03,
     + 1.7985E+03, 1.8011E+03, 1.8037E+03, 1.8063E+03, 1.8089E+03,
     +  1.8115E+03, 1.8141E+03, 1.8167E+03, 1.8193E+03, 1.8219E+03,
     + 1.8245E+03, 1.8271E+03, 1.8297E+03, 1.8323E+03, 1.8349E+03,
     +  1.8375E+03, 1.8402E+03, 1.8428E+03, 1.8454E+03, 1.8480E+03,
     + 1.8506E+03, 1.8533E+03, 1.8559E+03, 1.8585E+03, 1.8612E+03,
     +  1.8638E+03, 1.8664E+03, 1.8691E+03, 1.8717E+03, 1.8743E+03,
     + 1.8770E+03, 1.8796E+03, 1.8823E+03, 1.8849E+03, 1.8876E+03,
     +  1.8902E+03, 1.8929E+03, 1.8955E+03, 1.8982E+03, 1.9009E+03,
     + 1.9035E+03, 1.9062E+03, 1.9088E+03, 1.9115E+03, 1.9142E+03,
     +  1.9169E+03, 1.9195E+03, 1.9222E+03, 1.9249E+03, 1.9276E+03,
     + 1.9302E+03, 1.9329E+03, 1.9356E+03, 1.9383E+03, 1.9410E+03,
     +  1.9437E+03, 1.9464E+03, 1.9490E+03, 1.9517E+03, 1.9544E+03,
     + 1.9571E+03, 1.9598E+03, 1.9625E+03, 1.9652E+03, 1.9679E+03,
     +  1.9706E+03, 1.9734E+03, 1.9761E+03, 1.9788E+03, 1.9815E+03,
     + 1.9842E+03, 1.9869E+03, 1.9897E+03, 1.9924E+03, 1.9951E+03,
     +  1.9978E+03, 2.0005E+03, 2.0033E+03, 2.0060E+03, 2.0087E+03,
     + 2.0115E+03, 2.0142E+03, 2.0170E+03, 2.0197E+03, 2.0224E+03,
     +  2.0252E+03, 2.0279E+03, 2.0307E+03, 2.0334E+03, 2.0362E+03,
     + 2.0389E+03, 2.0417E+03, 2.0444E+03, 2.0472E+03, 2.0500E+03,
     +  2.0527E+03, 2.0555E+03, 2.0582E+03, 2.0610E+03, 2.0638E+03,
     + 2.0666E+03, 2.0693E+03, 2.0721E+03, 2.0749E+03, 2.0777E+03,
     +  2.0804E+03, 2.0832E+03, 2.0860E+03, 2.0888E+03, 2.0916E+03,
     + 2.0944E+03, 2.0972E+03, 2.0999E+03, 2.1027E+03, 2.1055E+03,
     +  2.1083E+03, 2.1111E+03, 2.1139E+03, 2.1167E+03, 2.1195E+03,
     + 2.1224E+03, 2.1252E+03, 2.1280E+03, 2.1308E+03, 2.1336E+03,
     +  2.1364E+03, 2.1392E+03, 2.1421E+03, 2.1449E+03, 2.1477E+03,
     + 2.1505E+03, 2.1534E+03, 2.1562E+03, 2.1590E+03, 2.1619E+03,
     +  2.1647E+03, 2.1675E+03, 2.1704E+03, 2.1732E+03, 2.1760E+03,
     + 2.1789E+03, 2.1817E+03, 2.1846E+03, 2.1874E+03, 2.1903E+03,
     +  2.1931E+03, 2.1960E+03, 2.1988E+03, 2.2017E+03, 2.2046E+03,
     + 2.2074E+03, 2.2103E+03, 2.2132E+03, 2.2160E+03, 2.2189E+03,
     +  2.2218E+03, 2.2246E+03, 2.2275E+03, 2.2304E+03, 2.2333E+03,
     + 2.2361E+03, 2.2390E+03, 2.2419E+03, 2.2448E+03, 2.2477E+03,
     +  2.2506E+03, 2.2535E+03, 2.2564E+03, 2.2593E+03, 2.2622E+03,
     + 2.2651E+03, 2.2680E+03, 2.2709E+03, 2.2738E+03, 2.2767E+03,
     +  2.2796E+03, 2.2825E+03, 2.2854E+03, 2.2883E+03, 2.2912E+03,
     + 2.2941E+03, 2.2971E+03, 2.3000E+03, 2.3029E+03, 2.3058E+03,
     +  2.3087E+03, 2.3117E+03, 2.3146E+03, 2.3175E+03, 2.3205E+03,
     + 2.3234E+03, 2.3263E+03, 2.3293E+03, 2.3322E+03, 2.3352E+03,
     +  2.3381E+03, 2.3410E+03, 2.3440E+03, 2.3469E+03, 2.3499E+03,
     + 2.3528E+03, 2.3558E+03, 2.3587E+03, 2.3617E+03, 2.3647E+03,
     +  2.3676E+03, 2.3706E+03, 2.3736E+03, 2.3765E+03, 2.3795E+03,
     + 2.3825E+03, 2.3854E+03, 2.3884E+03, 2.3914E+03, 2.3944E+03,
     +  2.3973E+03, 2.4003E+03, 2.4033E+03, 2.4063E+03, 2.4093E+03,
     + 2.4123E+03, 2.4153E+03, 2.4182E+03, 2.4212E+03, 2.4242E+03,
     +  2.4272E+03, 2.4302E+03, 2.4332E+03, 2.4362E+03, 2.4392E+03,
     + 2.4422E+03, 2.4453E+03, 2.4483E+03, 2.4513E+03, 2.4543E+03,
     +  2.4573E+03, 2.4603E+03, 2.4633E+03, 2.4664E+03, 2.4694E+03,
     + 2.4724E+03, 2.4754E+03, 2.4785E+03, 2.4815E+03, 2.4845E+03,
     +  2.4876E+03, 2.4906E+03, 2.4936E+03, 2.4967E+03, 2.4997E+03,
     + 2.5027E+03, 2.5058E+03, 2.5088E+03, 2.5119E+03, 2.5149E+03,
     +  2.5180E+03, 2.5210E+03, 2.5241E+03, 2.5272E+03, 2.5302E+03,
     + 2.5333E+03, 2.5363E+03, 2.5394E+03, 2.5425E+03, 2.5455E+03,
     +  2.5486E+03, 2.5517E+03, 2.5547E+03, 2.5578E+03, 2.5609E+03,
     + 2.5640E+03, 2.5670E+03, 2.5701E+03, 2.5732E+03, 2.5763E+03,
     +  2.5794E+03, 2.5825E+03, 2.5856E+03, 2.5886E+03, 2.5917E+03,
     + 2.5948E+03, 2.5979E+03, 2.6010E+03, 2.6041E+03, 2.6072E+03,
     +  2.6103E+03, 2.6134E+03, 2.6166E+03, 2.6197E+03, 2.6228E+03,
     + 2.6259E+03, 2.6290E+03, 2.6321E+03, 2.6352E+03, 2.6384E+03,
     +  2.6415E+03, 2.6446E+03, 2.6477E+03, 2.6509E+03, 2.6540E+03,
     + 2.6571E+03, 2.6602E+03, 2.6634E+03, 2.6665E+03, 2.6697E+03,
     +  2.6728E+03, 2.6759E+03, 2.6791E+03, 2.6822E+03, 2.6854E+03,
     + 2.6885E+03, 2.6917E+03, 2.6948E+03, 2.6980E+03, 2.7011E+03,
     +  2.7043E+03, 2.7074E+03, 2.7106E+03, 2.7138E+03, 2.7169E+03,
     + 2.7201E+03, 2.7233E+03, 2.7264E+03, 2.7296E+03, 2.7328E+03,
     +  2.7360E+03, 2.7391E+03, 2.7423E+03, 2.7455E+03, 2.7487E+03,
     + 2.7519E+03, 2.7550E+03, 2.7582E+03, 2.7614E+03, 2.7646E+03,
     +  2.7678E+03, 2.7710E+03, 2.7742E+03, 2.7774E+03, 2.7806E+03,
     + 2.7838E+03, 2.7870E+03, 2.7902E+03, 2.7934E+03, 2.7966E+03,
     +  2.7998E+03, 2.8030E+03, 2.8062E+03, 2.8095E+03, 2.8127E+03,
     + 2.8159E+03, 2.8191E+03, 2.8223E+03, 2.8256E+03, 2.8288E+03,
     +  2.8320E+03, 2.8352E+03, 2.8385E+03, 2.8417E+03, 2.8449E+03,
     + 2.8482E+03, 2.8514E+03, 2.8547E+03, 2.8579E+03, 2.8611E+03,
     +  2.8644E+03, 2.8676E+03, 2.8709E+03, 2.8741E+03, 2.8774E+03,
     + 2.8806E+03, 2.8839E+03, 2.8871E+03, 2.8904E+03, 2.8937E+03,
     +  2.8969E+03, 2.9002E+03, 2.9034E+03, 2.9067E+03, 2.9100E+03,
     + 2.9133E+03, 2.9165E+03, 2.9198E+03, 2.9231E+03, 2.9264E+03,
     +  2.9296E+03, 2.9329E+03, 2.9362E+03, 2.9395E+03, 2.9428E+03,
     + 2.9461E+03, 2.9493E+03, 2.9526E+03, 2.9559E+03, 2.9592E+03,
     +  2.9625E+03, 2.9658E+03, 2.9691E+03, 2.9724E+03, 2.9757E+03,
     + 2.9790E+03, 2.9823E+03, 2.9856E+03, 2.9890E+03, 2.9923E+03,
     +  2.9956E+03, 2.9989E+03, 3.0022E+03, 3.0055E+03, 3.0088E+03,
     + 3.0122E+03, 3.0155E+03, 3.0188E+03, 3.0221E+03, 3.0255E+03,
     +  3.0288E+03, 3.0321E+03, 3.0355E+03, 3.0388E+03, 3.0421E+03,
     + 3.0455E+03, 3.0488E+03, 3.0522E+03, 3.0555E+03, 3.0589E+03,
     +  3.0622E+03, 3.0656E+03, 3.0689E+03, 3.0723E+03, 3.0756E+03,
     + 3.0790E+03, 3.0823E+03, 3.0857E+03, 3.0891E+03, 3.0924E+03,
     +  3.0958E+03, 3.0991E+03, 3.1025E+03, 3.1059E+03, 3.1093E+03,
     + 3.1126E+03, 3.1160E+03, 3.1194E+03, 3.1228E+03, 3.1261E+03,
     +  3.1295E+03, 3.1329E+03, 3.1363E+03, 3.1397E+03, 3.1431E+03,
     + 3.1465E+03, 3.1498E+03, 3.1532E+03, 3.1566E+03, 3.1600E+03,
     +  3.1634E+03, 3.1668E+03, 3.1702E+03, 3.1736E+03, 3.1770E+03,
     + 3.1804E+03, 3.1839E+03, 3.1873E+03, 3.1907E+03, 3.1941E+03,
     +  3.1975E+03, 3.2009E+03, 3.2043E+03, 3.2078E+03, 3.2112E+03,
     + 3.2146E+03, 3.2180E+03, 3.2215E+03, 3.2249E+03, 3.2283E+03,
     +  3.2318E+03, 3.2352E+03, 3.2386E+03, 3.2421E+03, 3.2455E+03,
     + 3.2489E+03, 3.2524E+03, 3.2558E+03, 3.2593E+03, 3.2627E+03,
     +  3.2662E+03, 3.2696E+03, 3.2731E+03, 3.2765E+03, 3.2800E+03,
     + 3.2834E+03, 3.2869E+03, 3.2904E+03, 3.2938E+03, 3.2973E+03,
     +  3.3008E+03, 3.3042E+03, 3.3077E+03, 3.3112E+03, 3.3146E+03,
     + 3.3181E+03, 3.3216E+03, 3.3251E+03, 3.3285E+03, 3.3320E+03,
     +  3.3355E+03, 3.3390E+03, 3.3425E+03, 3.3460E+03, 3.3495E+03,
     + 3.3530E+03, 3.3564E+03, 3.3599E+03, 3.3634E+03, 3.3669E+03,
     +  3.3704E+03, 3.3739E+03, 3.3774E+03, 3.3809E+03, 3.3844E+03,
     + 3.3880E+03, 3.3915E+03, 3.3950E+03, 3.3985E+03, 3.4020E+03,
     +  3.4055E+03, 3.4090E+03, 3.4126E+03, 3.4161E+03, 3.4196E+03,
     + 3.4231E+03, 3.4266E+03, 3.4302E+03, 3.4337E+03, 3.4372E+03,
     +  3.4408E+03, 3.4443E+03, 3.4478E+03, 3.4514E+03, 3.4549E+03,
     + 3.4584E+03, 3.4620E+03, 3.4655E+03, 3.4691E+03, 3.4726E+03,
     +  3.4762E+03, 3.4797E+03, 3.4833E+03, 3.4868E+03, 3.4904E+03,
     + 3.4939E+03, 3.4975E+03, 3.5011E+03, 3.5046E+03, 3.5082E+03,
     +  3.5118E+03, 3.5153E+03, 3.5189E+03, 3.5225E+03, 3.5260E+03,
     + 3.5296E+03, 3.5332E+03, 3.5368E+03, 3.5403E+03, 3.5439E+03,
     +  3.5475E+03, 3.5511E+03, 3.5547E+03, 3.5583E+03, 3.5618E+03,
     + 3.5654E+03, 3.5690E+03, 3.5726E+03, 3.5762E+03, 3.5798E+03,
     +  3.5834E+03, 3.5870E+03, 3.5906E+03, 3.5942E+03, 3.5978E+03,
     + 3.6014E+03, 3.6050E+03, 3.6086E+03, 3.6122E+03, 3.6159E+03,
     +  3.6195E+03, 3.6231E+03, 3.6267E+03, 3.6303E+03, 3.6339E+03,
     + 3.6376E+03, 3.6412E+03, 3.6448E+03, 3.6484E+03, 3.6521E+03,
     +  3.6557E+03, 3.6593E+03, 3.6630E+03, 3.6666E+03, 3.6702E+03,
     + 3.6739E+03, 3.6775E+03, 3.6812E+03, 3.6848E+03, 3.6884E+03,
     +  3.6921E+03, 3.6957E+03, 3.6994E+03, 3.7030E+03, 3.7067E+03,
     + 3.7103E+03, 3.7140E+03, 3.7177E+03, 3.7213E+03, 3.7250E+03,
     +  3.7286E+03, 3.7323E+03, 3.7360E+03, 3.7396E+03, 3.7433E+03,
     + 3.7470E+03, 3.7506E+03, 3.7543E+03, 3.7580E+03, 3.7617E+03,
     +  3.7653E+03, 3.7690E+03, 3.7727E+03, 3.7764E+03, 3.7801E+03,
     + 3.7838E+03, 3.7875E+03, 3.7911E+03, 3.7948E+03, 3.7985E+03,
     +  3.8022E+03, 3.8059E+03, 3.8096E+03, 3.8133E+03, 3.8170E+03,
     + 3.8207E+03, 3.8244E+03, 3.8281E+03, 3.8318E+03, 3.8355E+03,
     +  3.8392E+03, 3.8430E+03, 3.8467E+03, 3.8504E+03, 3.8541E+03,
     + 3.8578E+03, 3.8615E+03, 3.8653E+03, 3.8690E+03, 3.8727E+03,
     +  3.8764E+03, 3.8802E+03, 3.8839E+03, 3.8876E+03, 3.8914E+03,
     + 3.8951E+03, 3.8988E+03, 3.9026E+03, 3.9063E+03, 3.9100E+03,
     +  3.9138E+03, 3.9175E+03, 3.9213E+03, 3.9250E+03, 3.9288E+03,
     + 3.9325E+03, 3.9363E+03, 3.9400E+03, 3.9438E+03, 3.9475E+03,
     +  3.9513E+03, 3.9550E+03, 3.9588E+03, 3.9626E+03, 3.9663E+03,
     + 3.9701E+03, 3.9739E+03, 3.9776E+03, 3.9814E+03, 3.9852E+03,
     +  3.9889E+03, 3.9927E+03, 3.9965E+03, 4.0003E+03, 4.0041E+03,
     + 4.0078E+03, 4.0116E+03, 4.0154E+03, 4.0192E+03, 4.0230E+03,
     +  4.0268E+03, 4.0305E+03, 4.0343E+03, 4.0381E+03, 4.0419E+03,
     + 4.0457E+03, 4.0495E+03, 4.0533E+03, 4.0571E+03, 4.0609E+03,
     +  4.0647E+03, 4.0685E+03, 4.0723E+03, 4.0762E+03, 4.0800E+03,
     + 4.0838E+03, 4.0876E+03, 4.0914E+03, 4.0952E+03, 4.0990E+03,
     +  4.1029E+03, 4.1067E+03, 4.1105E+03, 4.1143E+03, 4.1181E+03,
     + 4.1220E+03, 4.1258E+03, 4.1296E+03, 4.1335E+03, 4.1373E+03,
     +  4.1411E+03, 4.1450E+03, 4.1488E+03, 4.1526E+03, 4.1565E+03,
     + 4.1603E+03, 4.1642E+03, 4.1680E+03, 4.1719E+03, 4.1757E+03,
     +  4.1796E+03, 4.1834E+03, 4.1873E+03, 4.1911E+03, 4.1950E+03,
     + 4.1988E+03, 4.2027E+03, 4.2066E+03, 4.2104E+03, 4.2143E+03,
     +  4.2182E+03, 4.2220E+03, 4.2259E+03, 4.2298E+03, 4.2336E+03,
     + 4.2375E+03, 4.2414E+03, 4.2453E+03, 4.2491E+03, 4.2530E+03,
     +  4.2569E+03, 4.2608E+03, 4.2647E+03, 4.2685E+03, 4.2724E+03,
     + 4.2763E+03, 4.2802E+03, 4.2841E+03, 4.2880E+03, 4.2919E+03,
     +  4.2958E+03, 4.2997E+03, 4.3036E+03, 4.3075E+03, 4.3114E+03,
     + 4.3153E+03, 4.3192E+03, 4.3231E+03, 4.3270E+03, 4.3309E+03,
     +  4.3348E+03, 4.3387E+03, 4.3426E+03, 4.3466E+03, 4.3505E+03,
     + 4.3544E+03, 4.3583E+03, 4.3622E+03, 4.3662E+03, 4.3701E+03,
     +  4.3740E+03, 4.3779E+03, 4.3819E+03, 4.3858E+03, 4.3897E+03,
     + 4.3937E+03, 4.3976E+03, 4.4015E+03, 4.4055E+03, 4.4094E+03,
     +  4.4133E+03, 4.4173E+03, 4.4212E+03, 4.4252E+03, 4.4291E+03,
     + 4.4331E+03, 4.4370E+03, 4.4410E+03, 4.4449E+03, 4.4489E+03,
     +  4.4528E+03, 4.4568E+03, 4.4608E+03, 4.4647E+03, 4.4687E+03,
     + 4.4726E+03, 4.4766E+03, 4.4806E+03, 4.4845E+03, 4.4885E+03,
     +  4.4925E+03, 4.4964E+03, 4.5004E+03, 4.5044E+03, 4.5084E+03,
     + 4.5123E+03, 4.5163E+03, 4.5203E+03, 4.5243E+03, 4.5283E+03,
     +  4.5323E+03, 4.5362E+03, 4.5402E+03, 4.5442E+03, 4.5482E+03,
     + 4.5522E+03, 4.5562E+03, 4.5602E+03, 4.5642E+03, 4.5682E+03,
     +  4.5722E+03, 4.5762E+03, 4.5802E+03, 4.5842E+03, 4.5882E+03,
     + 4.5922E+03, 4.5962E+03, 4.6002E+03, 4.6042E+03, 4.6083E+03,
     +  4.6123E+03, 4.6163E+03, 4.6203E+03, 4.6243E+03, 4.6283E+03,
     + 4.6324E+03, 4.6364E+03, 4.6404E+03, 4.6444E+03, 4.6485E+03,
     +  4.6525E+03, 4.6565E+03, 4.6606E+03, 4.6646E+03, 4.6686E+03,
     + 4.6727E+03, 4.6767E+03, 4.6807E+03, 4.6848E+03, 4.6888E+03,
     +  4.6929E+03, 4.6969E+03, 4.7010E+03, 4.7050E+03, 4.7090E+03,
     + 4.7131E+03, 4.7172E+03, 4.7212E+03, 4.7253E+03, 4.7293E+03,
     +  4.7334E+03, 4.7374E+03, 4.7415E+03, 4.7456E+03, 4.7496E+03,
     + 4.7537E+03, 4.7578E+03, 4.7618E+03, 4.7659E+03, 4.7700E+03,
     +  4.7740E+03, 4.7781E+03, 4.7822E+03, 4.7863E+03, 4.7903E+03,
     + 4.7944E+03, 4.7985E+03, 4.8026E+03, 4.8067E+03, 4.8107E+03,
     +  4.8148E+03, 4.8189E+03, 4.8230E+03, 4.8271E+03, 4.8312E+03,
     + 4.8353E+03, 4.8394E+03, 4.8435E+03, 4.8476E+03, 4.8517E+03,
     +  4.8558E+03, 4.8599E+03, 4.8640E+03, 4.8681E+03, 4.8722E+03,
     + 4.8763E+03, 4.8804E+03, 4.8845E+03, 4.8886E+03, 4.8927E+03,
     +  4.8969E+03, 4.9010E+03, 4.9051E+03, 4.9092E+03, 4.9133E+03,
     + 4.9175E+03, 4.9216E+03, 4.9257E+03, 4.9298E+03, 4.9340E+03,
     +  4.9381E+03, 4.9422E+03, 4.9464E+03, 4.9505E+03, 4.9546E+03,
     + 4.9588E+03, 4.9629E+03, 4.9670E+03, 4.9712E+03, 4.9753E+03,
     +  4.9795E+03, 4.9836E+03, 4.9877E+03, 4.9919E+03, 4.9960E+03,
     + 5.0002E+03, 5.0043E+03, 5.0085E+03, 5.0127E+03, 5.0168E+03,
     +  5.0210E+03, 5.0251E+03, 5.0293E+03, 5.0334E+03, 5.0376E+03,
     + 5.0418E+03, 5.0459E+03, 5.0501E+03, 5.0543E+03, 5.0584E+03,
     +  5.0626E+03, 5.0668E+03, 5.0710E+03, 5.0751E+03, 5.0793E+03,
     + 5.0835E+03, 5.0877E+03, 5.0918E+03, 5.0960E+03, 5.1002E+03,
     +  5.1044E+03, 5.1086E+03, 5.1128E+03, 5.1170E+03, 5.1211E+03,
     + 5.1253E+03, 5.1295E+03, 5.1337E+03, 5.1379E+03, 5.1421E+03,
     +  5.1463E+03, 5.1505E+03, 5.1547E+03, 5.1589E+03, 5.1631E+03,
     + 5.1673E+03, 5.1715E+03, 5.1757E+03, 5.1800E+03, 5.1842E+03,
     +  5.1884E+03, 5.1926E+03, 5.1968E+03, 5.2010E+03, 5.2052E+03,
     + 5.2095E+03, 5.2137E+03, 5.2179E+03, 5.2221E+03, 5.2264E+03,
     +  5.2306E+03, 5.2348E+03, 5.2390E+03, 5.2433E+03, 5.2475E+03,
     + 5.2517E+03, 5.2560E+03, 5.2602E+03, 5.2644E+03, 5.2687E+03,
     +  5.2729E+03, 5.2772E+03, 5.2814E+03, 5.2856E+03, 5.2899E+03,
     + 5.2941E+03, 5.2984E+03, 5.3026E+03, 5.3069E+03, 5.3111E+03,
     +  5.3154E+03, 5.3196E+03, 5.3239E+03, 5.3281E+03, 5.3324E+03,
     + 5.3367E+03, 5.3409E+03, 5.3452E+03, 5.3495E+03, 5.3537E+03,
     +  5.3580E+03, 5.3623E+03, 5.3665E+03, 5.3708E+03, 5.3751E+03,
     + 5.3793E+03, 5.3836E+03, 5.3879E+03, 5.3922E+03, 5.3964E+03,
     +  5.4007E+03, 5.4050E+03, 5.4093E+03, 5.4136E+03, 5.4179E+03,
     + 5.4221E+03, 5.4264E+03, 5.4307E+03, 5.4350E+03, 5.4393E+03,
     +  5.4436E+03, 5.4479E+03, 5.4522E+03, 5.4565E+03, 5.4608E+03,
     + 5.4651E+03, 5.4694E+03, 5.4737E+03, 5.4780E+03, 5.4823E+03,
     +  5.4866E+03, 5.4909E+03, 5.4952E+03, 5.4995E+03, 5.5038E+03,
     + 5.5082E+03, 5.5125E+03, 5.5168E+03, 5.5211E+03, 5.5254E+03,
     +  5.5297E+03, 5.5341E+03, 5.5384E+03, 5.5427E+03, 5.5470E+03,
     + 5.5514E+03, 5.5557E+03, 5.5600E+03, 5.5643E+03, 5.5687E+03,
     +  5.5730E+03, 5.5773E+03, 5.5817E+03, 5.5860E+03, 5.5904E+03,
     + 5.5947E+03, 5.5990E+03, 5.6034E+03, 5.6077E+03, 5.6121E+03,
     +  5.6164E+03, 5.6208E+03, 5.6251E+03, 5.6295E+03, 5.6338E+03,
     + 5.6382E+03, 5.6425E+03, 5.6469E+03, 5.6512E+03, 5.6556E+03,
     +  5.6599E+03, 5.6643E+03, 5.6687E+03, 5.6730E+03, 5.6774E+03,
     + 5.6817E+03, 5.6861E+03, 5.6905E+03, 5.6949E+03, 5.6992E+03,
     +  5.7036E+03, 5.7080E+03, 5.7123E+03, 5.7167E+03, 5.7211E+03,
     + 5.7255E+03, 5.7299E+03, 5.7342E+03, 5.7386E+03, 5.7430E+03,
     +  5.7474E+03, 5.7518E+03, 5.7562E+03, 5.7605E+03, 5.7649E+03,
     + 5.7693E+03, 5.7737E+03, 5.7781E+03, 5.7825E+03, 5.7869E+03,
     +  5.7913E+03, 5.7957E+03, 5.8001E+03, 5.8045E+03, 5.8089E+03,
     + 5.8133E+03, 5.8177E+03, 5.8221E+03, 5.8265E+03, 5.8309E+03,
     +  5.8353E+03, 5.8398E+03, 5.8442E+03, 5.8486E+03, 5.8530E+03,
     + 5.8574E+03, 5.8618E+03, 5.8663E+03, 5.8707E+03, 5.8751E+03,
     +  5.8795E+03, 5.8839E+03, 5.8884E+03, 5.8928E+03, 5.8972E+03,
     + 5.9016E+03, 5.9061E+03, 5.9105E+03, 5.9149E+03, 5.9194E+03,
     +  5.9238E+03, 5.9283E+03, 5.9327E+03, 5.9371E+03, 5.9416E+03,
     + 5.9460E+03, 5.9505E+03, 5.9549E+03, 5.9593E+03, 5.9638E+03,
     +  5.9682E+03, 5.9727E+03, 5.9771E+03, 5.9816E+03, 5.9860E+03,
     + 5.9905E+03, 5.9950E+03, 5.9994E+03, 6.0039E+03, 6.0083E+03,
     +  6.0128E+03, 6.0172E+03, 6.0217E+03, 6.0262E+03, 6.0306E+03,
     + 6.0351E+03, 6.0396E+03, 6.0440E+03, 6.0485E+03, 6.0530E+03,
     +  6.0575E+03, 6.0619E+03, 6.0664E+03, 6.0709E+03, 6.0754E+03,
     + 6.0798E+03, 6.0843E+03, 6.0888E+03, 6.0933E+03, 6.0978E+03,
     +  6.1023E+03, 6.1067E+03, 6.1112E+03, 6.1157E+03, 6.1202E+03,
     + 6.1247E+03, 6.1292E+03, 6.1337E+03, 6.1382E+03, 6.1427E+03,
     +  6.1472E+03, 6.1517E+03, 6.1562E+03, 6.1607E+03, 6.1652E+03,
     + 6.1697E+03, 6.1742E+03, 6.1787E+03, 6.1832E+03, 6.1877E+03,
     +  6.1922E+03, 6.1967E+03, 6.2012E+03, 6.2058E+03, 6.2103E+03,
     + 6.2148E+03, 6.2193E+03, 6.2238E+03, 6.2283E+03, 6.2329E+03,
     +  6.2374E+03, 6.2419E+03, 6.2464E+03, 6.2510E+03, 6.2555E+03,
     + 6.2600E+03, 6.2645E+03, 6.2691E+03, 6.2736E+03, 6.2781E+03,
     +  6.2827E+03, 6.2872E+03, 6.2917E+03, 6.2963E+03, 6.3008E+03,
     + 6.3054E+03, 6.3099E+03, 6.3144E+03, 6.3190E+03, 6.3235E+03,
     +  6.3281E+03, 6.3326E+03, 6.3372E+03, 6.3417E+03, 6.3463E+03,
     + 6.3508E+03, 6.3554E+03, 6.3599E+03, 6.3645E+03, 6.3690E+03,
     +  6.3736E+03, 6.3782E+03, 6.3827E+03, 6.3873E+03, 6.3918E+03,
     + 6.3964E+03, 6.4010E+03, 6.4055E+03, 6.4101E+03, 6.4147E+03,
     +  6.4192E+03, 6.4238E+03, 6.4284E+03, 6.4330E+03, 6.4375E+03,
     + 6.4421E+03, 6.4467E+03, 6.4513E+03, 6.4559E+03, 6.4604E+03,
     +  6.4650E+03, 6.4696E+03, 6.4742E+03, 6.4788E+03, 6.4834E+03,
     + 6.4879E+03, 6.4925E+03, 6.4971E+03, 6.5017E+03, 6.5063E+03,
     +  6.5109E+03, 6.5155E+03, 6.5201E+03, 6.5247E+03, 6.5293E+03,
     + 6.5339E+03, 6.5385E+03, 6.5431E+03, 6.5477E+03, 6.5523E+03,
     +  6.5569E+03, 6.5615E+03, 6.5661E+03, 6.5707E+03, 6.5753E+03,
     + 6.5799E+03, 6.5846E+03, 6.5892E+03, 6.5938E+03, 6.5984E+03,
     +  6.6030E+03, 6.6076E+03, 6.6122E+03, 6.6169E+03, 6.6215E+03,
     + 6.6261E+03, 6.6307E+03, 6.6354E+03, 6.6400E+03, 6.6446E+03,
     +  6.6492E+03, 6.6539E+03, 6.6585E+03, 6.6631E+03, 6.6678E+03,
     + 6.6724E+03, 6.6770E+03, 6.6817E+03, 6.6863E+03, 6.6910E+03,
     +  6.6956E+03, 6.7002E+03, 6.7049E+03, 6.7095E+03, 6.7142E+03,
     + 6.7188E+03, 6.7235E+03, 6.7281E+03, 6.7328E+03, 6.7374E+03,
     +  6.7421E+03, 6.7467E+03, 6.7514E+03, 6.7560E+03, 6.7607E+03,
     + 6.7653E+03, 6.7700E+03, 6.7746E+03, 6.7793E+03, 6.7840E+03,
     +  6.7886E+03, 6.7933E+03, 6.7980E+03, 6.8026E+03, 6.8073E+03,
     + 6.8120E+03, 6.8166E+03, 6.8213E+03, 6.8260E+03, 6.8307E+03,
     +  6.8353E+03, 6.8400E+03, 6.8447E+03, 6.8494E+03, 6.8540E+03,
     + 6.8587E+03, 6.8634E+03, 6.8681E+03, 6.8728E+03, 6.8774E+03,
     +  6.8821E+03, 6.8868E+03, 6.8915E+03, 6.8962E+03, 6.9009E+03,
     + 6.9056E+03, 6.9103E+03, 6.9150E+03, 6.9197E+03, 6.9243E+03,
     +  6.9290E+03, 6.9337E+03, 6.9384E+03, 6.9431E+03, 6.9478E+03,
     + 6.9525E+03, 6.9572E+03, 6.9620E+03, 6.9667E+03, 6.9714E+03,
     +  6.9761E+03, 6.9808E+03, 6.9855E+03, 6.9902E+03, 6.9949E+03,
     + 6.9996E+03, 7.0043E+03, 7.0091E+03, 7.0138E+03, 7.0185E+03,
     +  7.0232E+03, 7.0279E+03, 7.0327E+03, 7.0374E+03, 7.0421E+03,
     + 7.0468E+03, 7.0516E+03, 7.0563E+03, 7.0610E+03, 7.0657E+03,
     +  7.0705E+03, 7.0752E+03, 7.0799E+03, 7.0847E+03, 7.0894E+03,
     + 7.0941E+03, 7.0989E+03, 7.1036E+03, 7.1083E+03, 7.1131E+03,
     +  7.1178E+03, 7.1226E+03, 7.1273E+03, 7.1321E+03, 7.1368E+03,
     + 7.1415E+03, 7.1463E+03, 7.1510E+03, 7.1558E+03, 7.1605E+03,
     +  7.1653E+03, 7.1700E+03, 7.1748E+03, 7.1796E+03, 7.1843E+03/
      data (QofT( 3,J),J=2409,4510)/ 7.1891E+03, 7.1938E+03,
     +  7.1986E+03, 7.2033E+03, 7.2081E+03, 7.2129E+03, 7.2176E+03,
     +  7.2224E+03,
     + 7.2272E+03, 7.2319E+03, 7.2367E+03, 7.2415E+03, 7.2462E+03,
     +  7.2510E+03, 7.2558E+03, 7.2606E+03, 7.2653E+03, 7.2701E+03,
     + 7.2749E+03, 7.2797E+03, 7.2844E+03, 7.2892E+03, 7.2940E+03,
     +  7.2988E+03, 7.3036E+03, 7.3084E+03, 7.3131E+03, 7.3179E+03,
     + 7.3227E+03, 7.3275E+03, 7.3323E+03, 7.3371E+03, 7.3419E+03,
     +  7.3467E+03, 7.3515E+03, 7.3562E+03, 7.3610E+03, 7.3658E+03,
     + 7.3706E+03, 7.3754E+03, 7.3802E+03, 7.3850E+03, 7.3898E+03,
     +  7.3946E+03, 7.3995E+03, 7.4043E+03, 7.4091E+03, 7.4139E+03,
     + 7.4187E+03, 7.4235E+03, 7.4283E+03, 7.4331E+03, 7.4379E+03,
     +  7.4427E+03, 7.4476E+03, 7.4524E+03, 7.4572E+03, 7.4620E+03,
     + 7.4668E+03, 7.4716E+03, 7.4765E+03, 7.4813E+03, 7.4861E+03,
     +  7.4909E+03, 7.4958E+03, 7.5006E+03, 7.5054E+03, 7.5102E+03,
     + 7.5151E+03, 7.5199E+03, 7.5247E+03, 7.5296E+03, 7.5344E+03,
     +  7.5392E+03, 7.5441E+03, 7.5489E+03, 7.5537E+03, 7.5586E+03,
     + 7.5634E+03, 7.5683E+03, 7.5731E+03, 7.5779E+03, 7.5828E+03,
     +  7.5876E+03, 7.5925E+03, 7.5973E+03, 7.6022E+03, 7.6070E+03,
     + 7.6119E+03, 7.6167E+03, 7.6216E+03, 7.6264E+03, 7.6313E+03,
     +  7.6361E+03, 7.6410E+03, 7.6459E+03, 7.6507E+03, 7.6556E+03,
     + 7.6604E+03, 7.6653E+03, 7.6702E+03, 7.6750E+03, 7.6799E+03,
     +  7.6848E+03, 7.6896E+03, 7.6945E+03, 7.6994E+03, 7.7042E+03,
     + 7.7091E+03, 7.7140E+03, 7.7188E+03, 7.7237E+03, 7.7286E+03,
     +  7.7335E+03, 7.7383E+03, 7.7432E+03, 7.7481E+03, 7.7530E+03,
     + 7.7579E+03, 7.7627E+03, 7.7676E+03, 7.7725E+03, 7.7774E+03,
     +  7.7823E+03, 7.7872E+03, 7.7921E+03, 7.7969E+03, 7.8018E+03,
     + 7.8067E+03, 7.8116E+03, 7.8165E+03, 7.8214E+03, 7.8263E+03,
     +  7.8312E+03, 7.8361E+03, 7.8410E+03, 7.8459E+03, 7.8508E+03,
     + 7.8557E+03, 7.8606E+03, 7.8655E+03, 7.8704E+03, 7.8753E+03,
     +  7.8802E+03, 7.8851E+03, 7.8900E+03, 7.8949E+03, 7.8998E+03,
     + 7.9048E+03, 7.9097E+03, 7.9146E+03, 7.9195E+03, 7.9244E+03,
     +  7.9293E+03, 7.9342E+03, 7.9392E+03, 7.9441E+03, 7.9490E+03,
     + 7.9539E+03, 7.9588E+03, 7.9638E+03, 7.9687E+03, 7.9736E+03,
     +  7.9785E+03, 7.9835E+03, 7.9884E+03, 7.9933E+03, 7.9983E+03,
     + 8.0032E+03, 8.0081E+03, 8.0130E+03, 8.0180E+03, 8.0229E+03,
     +  8.0279E+03, 8.0328E+03, 8.0377E+03, 8.0427E+03, 8.0476E+03,
     + 8.0525E+03, 8.0575E+03, 8.0624E+03, 8.0674E+03, 8.0723E+03,
     +  8.0773E+03, 8.0822E+03, 8.0872E+03, 8.0921E+03, 8.0971E+03,
     + 8.1020E+03, 8.1070E+03, 8.1119E+03, 8.1169E+03, 8.1218E+03,
     +  8.1268E+03, 8.1317E+03, 8.1367E+03, 8.1416E+03, 8.1466E+03,
     + 8.1516E+03, 8.1565E+03, 8.1615E+03, 8.1664E+03, 8.1714E+03,
     +  8.1764E+03, 8.1813E+03, 8.1863E+03, 8.1913E+03, 8.1962E+03,
     + 8.2012E+03, 8.2062E+03, 8.2112E+03, 8.2161E+03, 8.2211E+03,
     +  8.2261E+03, 8.2310E+03, 8.2360E+03, 8.2410E+03, 8.2460E+03,
     + 8.2510E+03, 8.2559E+03, 8.2609E+03, 8.2659E+03, 8.2709E+03,
     +  8.2759E+03, 8.2809E+03, 8.2858E+03, 8.2908E+03, 8.2958E+03,
     + 8.3008E+03, 8.3058E+03, 8.3108E+03, 8.3158E+03, 8.3208E+03,
     +  8.3258E+03, 8.3308E+03, 8.3357E+03, 8.3407E+03, 8.3457E+03,
     + 8.3507E+03, 8.3557E+03, 8.3607E+03, 8.3657E+03, 8.3707E+03,
     +  8.3757E+03, 8.3807E+03, 8.3858E+03, 8.3908E+03, 8.3958E+03,
     + 8.4008E+03, 8.4058E+03, 8.4108E+03, 8.4158E+03, 8.4208E+03,
     +  8.4258E+03, 8.4308E+03, 8.4358E+03, 8.4409E+03, 8.4459E+03,
     + 8.4509E+03, 8.4559E+03, 8.4609E+03, 8.4660E+03, 8.4710E+03,
     +  8.4760E+03, 8.4810E+03, 8.4860E+03, 8.4911E+03, 8.4961E+03,
     + 8.5011E+03, 8.5061E+03, 8.5112E+03, 8.5162E+03, 8.5212E+03,
     +  8.5263E+03, 8.5313E+03, 8.5363E+03, 8.5414E+03, 8.5464E+03,
     + 8.5514E+03, 8.5565E+03, 8.5615E+03, 8.5665E+03, 8.5716E+03,
     +  8.5766E+03, 8.5817E+03, 8.5867E+03, 8.5917E+03, 8.5968E+03,
     + 8.6018E+03, 8.6069E+03, 8.6119E+03, 8.6170E+03, 8.6220E+03,
     +  8.6271E+03, 8.6321E+03, 8.6372E+03, 8.6422E+03, 8.6473E+03,
     + 8.6523E+03, 8.6574E+03, 8.6624E+03, 8.6675E+03, 8.6725E+03,
     +  8.6776E+03, 8.6827E+03, 8.6877E+03, 8.6928E+03, 8.6978E+03,
     + 8.7029E+03, 8.7080E+03, 8.7130E+03, 8.7181E+03, 8.7232E+03,
     +  8.7282E+03, 8.7333E+03, 8.7384E+03, 8.7434E+03, 8.7485E+03,
     + 8.7536E+03, 8.7587E+03, 8.7637E+03, 8.7688E+03, 8.7739E+03,
     +  8.7789E+03, 8.7840E+03, 8.7891E+03, 8.7942E+03, 8.7993E+03,
     + 8.8043E+03, 8.8094E+03, 8.8145E+03, 8.8196E+03, 8.8247E+03,
     +  8.8298E+03, 8.8348E+03, 8.8399E+03, 8.8450E+03, 8.8501E+03,
     + 8.8552E+03, 8.8603E+03, 8.8654E+03, 8.8705E+03, 8.8756E+03,
     +  8.8807E+03, 8.8857E+03, 8.8908E+03, 8.8959E+03, 8.9010E+03,
     + 8.9061E+03, 8.9112E+03, 8.9163E+03, 8.9214E+03, 8.9265E+03,
     +  8.9316E+03, 8.9367E+03, 8.9418E+03, 8.9470E+03, 8.9521E+03,
     + 8.9572E+03, 8.9623E+03, 8.9674E+03, 8.9725E+03, 8.9776E+03,
     +  8.9827E+03, 8.9878E+03, 8.9929E+03, 8.9980E+03, 9.0032E+03,
     + 9.0083E+03, 9.0134E+03, 9.0185E+03, 9.0236E+03, 9.0288E+03,
     +  9.0339E+03, 9.0390E+03, 9.0441E+03, 9.0492E+03, 9.0544E+03,
     + 9.0595E+03, 9.0646E+03, 9.0697E+03, 9.0749E+03, 9.0800E+03,
     +  9.0851E+03, 9.0902E+03, 9.0954E+03, 9.1005E+03, 9.1056E+03,
     + 9.1108E+03, 9.1159E+03, 9.1210E+03, 9.1262E+03, 9.1313E+03,
     +  9.1364E+03, 9.1416E+03, 9.1467E+03, 9.1519E+03, 9.1570E+03,
     + 9.1621E+03, 9.1673E+03, 9.1724E+03, 9.1776E+03, 9.1827E+03,
     +  9.1879E+03, 9.1930E+03, 9.1982E+03, 9.2033E+03, 9.2085E+03,
     + 9.2136E+03, 9.2188E+03, 9.2239E+03, 9.2291E+03, 9.2342E+03,
     +  9.2394E+03, 9.2445E+03, 9.2497E+03, 9.2548E+03, 9.2600E+03,
     + 9.2651E+03, 9.2703E+03, 9.2755E+03, 9.2806E+03, 9.2858E+03,
     +  9.2910E+03, 9.2961E+03, 9.3013E+03, 9.3064E+03, 9.3116E+03,
     + 9.3168E+03, 9.3219E+03, 9.3271E+03, 9.3323E+03, 9.3374E+03,
     +  9.3426E+03, 9.3478E+03, 9.3530E+03, 9.3581E+03, 9.3633E+03,
     + 9.3685E+03, 9.3737E+03, 9.3788E+03, 9.3840E+03, 9.3892E+03,
     +  9.3944E+03, 9.3996E+03, 9.4047E+03, 9.4099E+03, 9.4151E+03,
     + 9.4203E+03, 9.4255E+03, 9.4306E+03, 9.4358E+03, 9.4410E+03,
     +  9.4462E+03, 9.4514E+03, 9.4566E+03, 9.4618E+03, 9.4670E+03,
     + 9.4722E+03, 9.4773E+03, 9.4825E+03, 9.4877E+03, 9.4929E+03,
     +  9.4981E+03, 9.5033E+03, 9.5085E+03, 9.5137E+03, 9.5189E+03,
     + 9.5241E+03, 9.5293E+03, 9.5345E+03, 9.5397E+03, 9.5449E+03,
     +  9.5501E+03, 9.5553E+03, 9.5605E+03, 9.5657E+03, 9.5709E+03,
     + 9.5762E+03, 9.5814E+03, 9.5866E+03, 9.5918E+03, 9.5970E+03,
     +  9.6022E+03, 9.6074E+03, 9.6126E+03, 9.6178E+03, 9.6231E+03,
     + 9.6283E+03, 9.6335E+03, 9.6387E+03, 9.6439E+03, 9.6491E+03,
     +  9.6544E+03, 9.6596E+03, 9.6648E+03, 9.6700E+03, 9.6752E+03,
     + 9.6805E+03, 9.6857E+03, 9.6909E+03, 9.6961E+03, 9.7014E+03,
     +  9.7066E+03, 9.7118E+03, 9.7171E+03, 9.7223E+03, 9.7275E+03,
     + 9.7327E+03, 9.7380E+03, 9.7432E+03, 9.7484E+03, 9.7537E+03,
     +  9.7589E+03, 9.7642E+03, 9.7694E+03, 9.7746E+03, 9.7799E+03,
     + 9.7851E+03, 9.7903E+03, 9.7956E+03, 9.8008E+03, 9.8061E+03,
     +  9.8113E+03, 9.8166E+03, 9.8218E+03, 9.8270E+03, 9.8323E+03,
     + 9.8375E+03, 9.8428E+03, 9.8480E+03, 9.8533E+03, 9.8585E+03,
     +  9.8638E+03, 9.8690E+03, 9.8743E+03, 9.8795E+03, 9.8848E+03,
     + 9.8901E+03, 9.8953E+03, 9.9006E+03, 9.9058E+03, 9.9111E+03,
     +  9.9163E+03, 9.9216E+03, 9.9269E+03, 9.9321E+03, 9.9374E+03,
     + 9.9426E+03, 9.9479E+03, 9.9532E+03, 9.9584E+03, 9.9637E+03,
     +  9.9690E+03, 9.9742E+03, 9.9795E+03, 9.9848E+03, 9.9900E+03,
     + 9.9953E+03, 1.0001E+04, 1.0006E+04, 1.0011E+04, 1.0016E+04,
     +  1.0022E+04, 1.0027E+04, 1.0032E+04, 1.0038E+04, 1.0043E+04,
     + 1.0048E+04, 1.0053E+04, 1.0059E+04, 1.0064E+04, 1.0069E+04,
     +  1.0074E+04, 1.0080E+04, 1.0085E+04, 1.0090E+04, 1.0096E+04,
     + 1.0101E+04, 1.0106E+04, 1.0111E+04, 1.0117E+04, 1.0122E+04,
     +  1.0127E+04, 1.0133E+04, 1.0138E+04, 1.0143E+04, 1.0148E+04,
     + 1.0154E+04, 1.0159E+04, 1.0164E+04, 1.0170E+04, 1.0175E+04,
     +  1.0180E+04, 1.0186E+04, 1.0191E+04, 1.0196E+04, 1.0201E+04,
     + 1.0207E+04, 1.0212E+04, 1.0217E+04, 1.0223E+04, 1.0228E+04,
     +  1.0233E+04, 1.0239E+04, 1.0244E+04, 1.0249E+04, 1.0254E+04,
     + 1.0260E+04, 1.0265E+04, 1.0270E+04, 1.0276E+04, 1.0281E+04,
     +  1.0286E+04, 1.0292E+04, 1.0297E+04, 1.0302E+04, 1.0308E+04,
     + 1.0313E+04, 1.0318E+04, 1.0324E+04, 1.0329E+04, 1.0334E+04,
     +  1.0340E+04, 1.0345E+04, 1.0350E+04, 1.0355E+04, 1.0361E+04,
     + 1.0366E+04, 1.0371E+04, 1.0377E+04, 1.0382E+04, 1.0387E+04,
     +  1.0393E+04, 1.0398E+04, 1.0403E+04, 1.0409E+04, 1.0414E+04,
     + 1.0419E+04, 1.0425E+04, 1.0430E+04, 1.0435E+04, 1.0441E+04,
     +  1.0446E+04, 1.0451E+04, 1.0457E+04, 1.0462E+04, 1.0467E+04,
     + 1.0473E+04, 1.0478E+04, 1.0483E+04, 1.0489E+04, 1.0494E+04,
     +  1.0499E+04, 1.0505E+04, 1.0510E+04, 1.0515E+04, 1.0521E+04,
     + 1.0526E+04, 1.0531E+04, 1.0537E+04, 1.0542E+04, 1.0548E+04,
     +  1.0553E+04, 1.0558E+04, 1.0564E+04, 1.0569E+04, 1.0574E+04,
     + 1.0580E+04, 1.0585E+04, 1.0590E+04, 1.0596E+04, 1.0601E+04,
     +  1.0606E+04, 1.0612E+04, 1.0617E+04, 1.0622E+04, 1.0628E+04,
     + 1.0633E+04, 1.0639E+04, 1.0644E+04, 1.0649E+04, 1.0655E+04,
     +  1.0660E+04, 1.0665E+04, 1.0671E+04, 1.0676E+04, 1.0681E+04,
     + 1.0687E+04, 1.0692E+04, 1.0698E+04, 1.0703E+04, 1.0708E+04,
     +  1.0714E+04, 1.0719E+04, 1.0724E+04, 1.0730E+04, 1.0735E+04,
     + 1.0740E+04, 1.0746E+04, 1.0751E+04, 1.0757E+04, 1.0762E+04,
     +  1.0767E+04, 1.0773E+04, 1.0778E+04, 1.0783E+04, 1.0789E+04,
     + 1.0794E+04, 1.0800E+04, 1.0805E+04, 1.0810E+04, 1.0816E+04,
     +  1.0821E+04, 1.0827E+04, 1.0832E+04, 1.0837E+04, 1.0843E+04,
     + 1.0848E+04, 1.0853E+04, 1.0859E+04, 1.0864E+04, 1.0870E+04,
     +  1.0875E+04, 1.0880E+04, 1.0886E+04, 1.0891E+04, 1.0897E+04,
     + 1.0902E+04, 1.0907E+04, 1.0913E+04, 1.0918E+04, 1.0924E+04,
     +  1.0929E+04, 1.0934E+04, 1.0940E+04, 1.0945E+04, 1.0951E+04,
     + 1.0956E+04, 1.0961E+04, 1.0967E+04, 1.0972E+04, 1.0978E+04,
     +  1.0983E+04, 1.0988E+04, 1.0994E+04, 1.0999E+04, 1.1005E+04,
     + 1.1010E+04, 1.1015E+04, 1.1021E+04, 1.1026E+04, 1.1032E+04,
     +  1.1037E+04, 1.1042E+04, 1.1048E+04, 1.1053E+04, 1.1059E+04,
     + 1.1064E+04, 1.1069E+04, 1.1075E+04, 1.1080E+04, 1.1086E+04,
     +  1.1091E+04, 1.1097E+04, 1.1102E+04, 1.1107E+04, 1.1113E+04,
     + 1.1118E+04, 1.1124E+04, 1.1129E+04, 1.1134E+04, 1.1140E+04,
     +  1.1145E+04, 1.1151E+04, 1.1156E+04, 1.1162E+04, 1.1167E+04,
     + 1.1172E+04, 1.1178E+04, 1.1183E+04, 1.1189E+04, 1.1194E+04,
     +  1.1200E+04, 1.1205E+04, 1.1210E+04, 1.1216E+04, 1.1221E+04,
     + 1.1227E+04, 1.1232E+04, 1.1238E+04, 1.1243E+04, 1.1248E+04,
     +  1.1254E+04, 1.1259E+04, 1.1265E+04, 1.1270E+04, 1.1276E+04,
     + 1.1281E+04, 1.1287E+04, 1.1292E+04, 1.1297E+04, 1.1303E+04,
     +  1.1308E+04, 1.1314E+04, 1.1319E+04, 1.1325E+04, 1.1330E+04,
     + 1.1336E+04, 1.1341E+04, 1.1346E+04, 1.1352E+04, 1.1357E+04,
     +  1.1363E+04, 1.1368E+04, 1.1374E+04, 1.1379E+04, 1.1385E+04,
     + 1.1390E+04, 1.1395E+04, 1.1401E+04, 1.1406E+04, 1.1412E+04,
     +  1.1417E+04, 1.1423E+04, 1.1428E+04, 1.1434E+04, 1.1439E+04,
     + 1.1445E+04, 1.1450E+04, 1.1455E+04, 1.1461E+04, 1.1466E+04,
     +  1.1472E+04, 1.1477E+04, 1.1483E+04, 1.1488E+04, 1.1494E+04,
     + 1.1499E+04, 1.1505E+04, 1.1510E+04, 1.1516E+04, 1.1521E+04,
     +  1.1527E+04, 1.1532E+04, 1.1537E+04, 1.1543E+04, 1.1548E+04,
     + 1.1554E+04, 1.1559E+04, 1.1565E+04, 1.1570E+04, 1.1576E+04,
     +  1.1581E+04, 1.1587E+04, 1.1592E+04, 1.1598E+04, 1.1603E+04,
     + 1.1609E+04, 1.1614E+04, 1.1620E+04, 1.1625E+04, 1.1631E+04,
     +  1.1636E+04, 1.1641E+04, 1.1647E+04, 1.1652E+04, 1.1658E+04,
     + 1.1663E+04, 1.1669E+04, 1.1674E+04, 1.1680E+04, 1.1685E+04,
     +  1.1691E+04, 1.1696E+04, 1.1702E+04, 1.1707E+04, 1.1713E+04,
     + 1.1718E+04, 1.1724E+04, 1.1729E+04, 1.1735E+04, 1.1740E+04,
     +  1.1746E+04, 1.1751E+04, 1.1757E+04, 1.1762E+04, 1.1768E+04,
     + 1.1773E+04, 1.1779E+04, 1.1784E+04, 1.1790E+04, 1.1795E+04,
     +  1.1801E+04, 1.1806E+04, 1.1812E+04, 1.1817E+04, 1.1823E+04,
     + 1.1828E+04, 1.1834E+04, 1.1839E+04, 1.1845E+04, 1.1850E+04,
     +  1.1856E+04, 1.1861E+04, 1.1867E+04, 1.1872E+04, 1.1878E+04,
     + 1.1883E+04, 1.1889E+04, 1.1894E+04, 1.1900E+04, 1.1905E+04,
     +  1.1911E+04, 1.1916E+04, 1.1922E+04, 1.1927E+04, 1.1933E+04,
     + 1.1938E+04, 1.1944E+04, 1.1949E+04, 1.1955E+04, 1.1960E+04,
     +  1.1966E+04, 1.1971E+04, 1.1977E+04, 1.1982E+04, 1.1988E+04,
     + 1.1993E+04, 1.1999E+04, 1.2005E+04, 1.2010E+04, 1.2016E+04,
     +  1.2021E+04, 1.2027E+04, 1.2032E+04, 1.2038E+04, 1.2043E+04,
     + 1.2049E+04, 1.2054E+04, 1.2060E+04, 1.2065E+04, 1.2071E+04,
     +  1.2076E+04, 1.2082E+04, 1.2087E+04, 1.2093E+04, 1.2098E+04,
     + 1.2104E+04, 1.2110E+04, 1.2115E+04, 1.2121E+04, 1.2126E+04,
     +  1.2132E+04, 1.2137E+04, 1.2143E+04, 1.2148E+04, 1.2154E+04,
     + 1.2159E+04, 1.2165E+04, 1.2170E+04, 1.2176E+04, 1.2181E+04,
     +  1.2187E+04, 1.2193E+04, 1.2198E+04, 1.2204E+04, 1.2209E+04,
     + 1.2215E+04, 1.2220E+04, 1.2226E+04, 1.2231E+04, 1.2237E+04,
     +  1.2242E+04, 1.2248E+04, 1.2254E+04, 1.2259E+04, 1.2265E+04,
     + 1.2270E+04, 1.2276E+04, 1.2281E+04, 1.2287E+04, 1.2292E+04,
     +  1.2298E+04, 1.2303E+04, 1.2309E+04, 1.2315E+04, 1.2320E+04,
     + 1.2326E+04, 1.2331E+04, 1.2337E+04, 1.2342E+04, 1.2348E+04,
     +  1.2353E+04, 1.2359E+04, 1.2365E+04, 1.2370E+04, 1.2376E+04,
     + 1.2381E+04, 1.2387E+04, 1.2392E+04, 1.2398E+04, 1.2403E+04,
     +  1.2409E+04, 1.2415E+04, 1.2420E+04, 1.2426E+04, 1.2431E+04,
     + 1.2437E+04, 1.2442E+04, 1.2448E+04, 1.2454E+04, 1.2459E+04,
     +  1.2465E+04, 1.2470E+04, 1.2476E+04, 1.2481E+04, 1.2487E+04,
     + 1.2493E+04, 1.2498E+04, 1.2504E+04, 1.2509E+04, 1.2515E+04,
     +  1.2520E+04, 1.2526E+04, 1.2532E+04, 1.2537E+04, 1.2543E+04,
     + 1.2548E+04, 1.2554E+04, 1.2559E+04, 1.2565E+04, 1.2571E+04,
     +  1.2576E+04, 1.2582E+04, 1.2587E+04, 1.2593E+04, 1.2598E+04,
     + 1.2604E+04, 1.2610E+04, 1.2615E+04, 1.2621E+04, 1.2626E+04,
     +  1.2632E+04, 1.2638E+04, 1.2643E+04, 1.2649E+04, 1.2654E+04,
     + 1.2660E+04, 1.2665E+04, 1.2671E+04, 1.2677E+04, 1.2682E+04,
     +  1.2688E+04, 1.2693E+04, 1.2699E+04, 1.2705E+04, 1.2710E+04,
     + 1.2716E+04, 1.2721E+04, 1.2727E+04, 1.2733E+04, 1.2738E+04,
     +  1.2744E+04, 1.2749E+04, 1.2755E+04, 1.2760E+04, 1.2766E+04,
     + 1.2772E+04, 1.2777E+04, 1.2783E+04, 1.2788E+04, 1.2794E+04,
     +  1.2800E+04, 1.2805E+04, 1.2811E+04, 1.2816E+04, 1.2822E+04,
     + 1.2828E+04, 1.2833E+04, 1.2839E+04, 1.2844E+04, 1.2850E+04,
     +  1.2856E+04, 1.2861E+04, 1.2867E+04, 1.2873E+04, 1.2878E+04,
     + 1.2884E+04, 1.2889E+04, 1.2895E+04, 1.2901E+04, 1.2906E+04,
     +  1.2912E+04, 1.2917E+04, 1.2923E+04, 1.2929E+04, 1.2934E+04,
     + 1.2940E+04, 1.2945E+04, 1.2951E+04, 1.2957E+04, 1.2962E+04,
     +  1.2968E+04, 1.2973E+04, 1.2979E+04, 1.2985E+04, 1.2990E+04,
     + 1.2996E+04, 1.3002E+04, 1.3007E+04, 1.3013E+04, 1.3018E+04,
     +  1.3024E+04, 1.3030E+04, 1.3035E+04, 1.3041E+04, 1.3047E+04,
     + 1.3052E+04, 1.3058E+04, 1.3063E+04, 1.3069E+04, 1.3075E+04,
     +  1.3080E+04, 1.3086E+04, 1.3092E+04, 1.3097E+04, 1.3103E+04,
     + 1.3108E+04, 1.3114E+04, 1.3120E+04, 1.3125E+04, 1.3131E+04,
     +  1.3137E+04, 1.3142E+04, 1.3148E+04, 1.3153E+04, 1.3159E+04,
     + 1.3165E+04, 1.3170E+04, 1.3176E+04, 1.3182E+04, 1.3187E+04,
     +  1.3193E+04, 1.3198E+04, 1.3204E+04, 1.3210E+04, 1.3215E+04,
     + 1.3221E+04, 1.3227E+04, 1.3232E+04, 1.3238E+04, 1.3244E+04,
     +  1.3249E+04, 1.3255E+04, 1.3261E+04, 1.3266E+04, 1.3272E+04,
     + 1.3277E+04, 1.3283E+04, 1.3289E+04, 1.3294E+04, 1.3300E+04,
     +  1.3306E+04, 1.3311E+04, 1.3317E+04, 1.3323E+04, 1.3328E+04,
     + 1.3334E+04, 1.3340E+04, 1.3345E+04, 1.3351E+04, 1.3356E+04,
     +  1.3362E+04, 1.3368E+04, 1.3373E+04, 1.3379E+04, 1.3385E+04,
     + 1.3390E+04, 1.3396E+04, 1.3402E+04, 1.3407E+04, 1.3413E+04,
     +  1.3419E+04, 1.3424E+04, 1.3430E+04, 1.3436E+04, 1.3441E+04,
     + 1.3447E+04, 1.3453E+04, 1.3458E+04, 1.3464E+04, 1.3470E+04,
     +  1.3475E+04, 1.3481E+04, 1.3487E+04, 1.3492E+04, 1.3498E+04,
     + 1.3503E+04, 1.3509E+04, 1.3515E+04, 1.3520E+04, 1.3526E+04,
     +  1.3532E+04, 1.3537E+04, 1.3543E+04, 1.3549E+04, 1.3554E+04,
     + 1.3560E+04, 1.3566E+04, 1.3571E+04, 1.3577E+04, 1.3583E+04,
     +  1.3588E+04, 1.3594E+04, 1.3600E+04, 1.3605E+04, 1.3611E+04,
     + 1.3617E+04, 1.3622E+04, 1.3628E+04, 1.3634E+04, 1.3639E+04,
     +  1.3645E+04, 1.3651E+04, 1.3657E+04, 1.3662E+04, 1.3668E+04,
     + 1.3674E+04, 1.3679E+04, 1.3685E+04, 1.3691E+04, 1.3696E+04,
     +  1.3702E+04, 1.3708E+04, 1.3713E+04, 1.3719E+04, 1.3725E+04,
     + 1.3730E+04, 1.3736E+04, 1.3742E+04, 1.3747E+04, 1.3753E+04,
     +  1.3759E+04, 1.3764E+04, 1.3770E+04, 1.3776E+04, 1.3781E+04,
     + 1.3787E+04, 1.3793E+04, 1.3798E+04, 1.3804E+04, 1.3810E+04,
     +  1.3816E+04, 1.3821E+04, 1.3827E+04, 1.3833E+04, 1.3838E+04,
     + 1.3844E+04, 1.3850E+04, 1.3855E+04, 1.3861E+04, 1.3867E+04,
     +  1.3872E+04, 1.3878E+04, 1.3884E+04, 1.3889E+04, 1.3895E+04,
     + 1.3901E+04, 1.3907E+04, 1.3912E+04, 1.3918E+04, 1.3924E+04,
     +  1.3929E+04, 1.3935E+04, 1.3941E+04, 1.3946E+04, 1.3952E+04,
     + 1.3958E+04, 1.3964E+04, 1.3969E+04, 1.3975E+04, 1.3981E+04,
     +  1.3986E+04, 1.3992E+04, 1.3998E+04, 1.4003E+04, 1.4009E+04,
     + 1.4015E+04, 1.4021E+04, 1.4026E+04, 1.4032E+04, 1.4038E+04,
     +  1.4043E+04, 1.4049E+04, 1.4055E+04, 1.4060E+04, 1.4066E+04,
     + 1.4072E+04, 1.4078E+04, 1.4083E+04, 1.4089E+04, 1.4095E+04,
     +  1.4100E+04, 1.4106E+04, 1.4112E+04, 1.4118E+04, 1.4123E+04,
     + 1.4129E+04, 1.4135E+04, 1.4140E+04, 1.4146E+04, 1.4152E+04,
     +  1.4157E+04, 1.4163E+04, 1.4169E+04, 1.4175E+04, 1.4180E+04,
     + 1.4186E+04, 1.4192E+04, 1.4197E+04, 1.4203E+04, 1.4209E+04,
     +  1.4215E+04, 1.4220E+04, 1.4226E+04, 1.4232E+04, 1.4238E+04,
     + 1.4243E+04, 1.4249E+04, 1.4255E+04, 1.4260E+04, 1.4266E+04,
     +  1.4272E+04, 1.4278E+04, 1.4283E+04, 1.4289E+04, 1.4295E+04,
     + 1.4300E+04, 1.4306E+04, 1.4312E+04, 1.4318E+04, 1.4323E+04,
     +  1.4329E+04, 1.4335E+04, 1.4341E+04, 1.4346E+04, 1.4352E+04,
     + 1.4358E+04, 1.4363E+04, 1.4369E+04, 1.4375E+04, 1.4381E+04,
     +  1.4386E+04, 1.4392E+04, 1.4398E+04, 1.4404E+04, 1.4409E+04,
     + 1.4415E+04, 1.4421E+04, 1.4426E+04, 1.4432E+04, 1.4438E+04,
     +  1.4444E+04, 1.4449E+04, 1.4455E+04, 1.4461E+04, 1.4467E+04,
     + 1.4472E+04, 1.4478E+04, 1.4484E+04, 1.4490E+04, 1.4495E+04,
     +  1.4501E+04, 1.4507E+04, 1.4512E+04, 1.4518E+04, 1.4524E+04,
     + 1.4530E+04, 1.4535E+04, 1.4541E+04, 1.4547E+04, 1.4553E+04,
     +  1.4558E+04, 1.4564E+04, 1.4570E+04, 1.4576E+04, 1.4581E+04,
     + 1.4587E+04, 1.4593E+04, 1.4599E+04, 1.4604E+04, 1.4610E+04,
     +  1.4616E+04, 1.4622E+04, 1.4627E+04, 1.4633E+04, 1.4639E+04,
     + 1.4645E+04, 1.4650E+04, 1.4656E+04, 1.4662E+04, 1.4668E+04,
     +  1.4673E+04, 1.4679E+04, 1.4685E+04, 1.4691E+04, 1.4696E+04,
     + 1.4702E+04, 1.4708E+04, 1.4714E+04, 1.4719E+04, 1.4725E+04,
     +  1.4731E+04, 1.4737E+04, 1.4742E+04, 1.4748E+04, 1.4754E+04,
     + 1.4760E+04, 1.4765E+04, 1.4771E+04, 1.4777E+04, 1.4783E+04,
     +  1.4788E+04, 1.4794E+04, 1.4800E+04, 1.4806E+04, 1.4811E+04,
     + 1.4817E+04, 1.4823E+04, 1.4829E+04, 1.4834E+04, 1.4840E+04,
     +  1.4846E+04, 1.4852E+04, 1.4857E+04, 1.4863E+04, 1.4869E+04,
     + 1.4875E+04, 1.4880E+04, 1.4886E+04, 1.4892E+04, 1.4898E+04,
     +  1.4904E+04, 1.4909E+04, 1.4915E+04, 1.4921E+04, 1.4927E+04,
     + 1.4932E+04, 1.4938E+04, 1.4944E+04, 1.4950E+04, 1.4955E+04,
     +  1.4961E+04, 1.4967E+04, 1.4973E+04, 1.4979E+04, 1.4984E+04,
     + 1.4990E+04, 1.4996E+04, 1.5002E+04, 1.5007E+04, 1.5013E+04,
     +  1.5019E+04, 1.5025E+04, 1.5030E+04, 1.5036E+04, 1.5042E+04,
     + 1.5048E+04, 1.5054E+04, 1.5059E+04, 1.5065E+04, 1.5071E+04,
     +  1.5077E+04, 1.5082E+04, 1.5088E+04, 1.5094E+04, 1.5100E+04,
     + 1.5106E+04, 1.5111E+04, 1.5117E+04, 1.5123E+04, 1.5129E+04,
     +  1.5134E+04, 1.5140E+04, 1.5146E+04, 1.5152E+04, 1.5158E+04,
     + 1.5163E+04, 1.5169E+04, 1.5175E+04, 1.5181E+04, 1.5186E+04,
     +  1.5192E+04, 1.5198E+04, 1.5204E+04, 1.5210E+04, 1.5215E+04,
     + 1.5221E+04, 1.5227E+04, 1.5233E+04, 1.5239E+04, 1.5244E+04,
     +  1.5250E+04, 1.5256E+04, 1.5262E+04, 1.5267E+04, 1.5273E+04,
     + 1.5279E+04, 1.5285E+04, 1.5291E+04, 1.5296E+04, 1.5302E+04,
     +  1.5308E+04, 1.5314E+04, 1.5320E+04, 1.5325E+04, 1.5331E+04,
     + 1.5337E+04, 1.5343E+04, 1.5348E+04, 1.5354E+04, 1.5360E+04,
     +  1.5366E+04, 1.5372E+04, 1.5377E+04, 1.5383E+04, 1.5389E+04,
     + 1.5395E+04, 1.5401E+04, 1.5406E+04, 1.5412E+04, 1.5418E+04,
     +  1.5424E+04, 1.5430E+04, 1.5435E+04, 1.5441E+04, 1.5447E+04,
     + 1.5453E+04, 1.5459E+04, 1.5464E+04, 1.5470E+04, 1.5476E+04,
     +  1.5482E+04, 1.5488E+04, 1.5493E+04, 1.5499E+04, 1.5505E+04,
     + 1.5511E+04, 1.5517E+04, 1.5522E+04, 1.5528E+04, 1.5534E+04,
     +  1.5540E+04, 1.5546E+04, 1.5551E+04, 1.5557E+04, 1.5563E+04,
     + 1.5569E+04, 1.5575E+04, 1.5580E+04, 1.5586E+04, 1.5592E+04,
     +  1.5598E+04, 1.5604E+04, 1.5610E+04, 1.5615E+04, 1.5621E+04,
     + 1.5627E+04, 1.5633E+04, 1.5639E+04, 1.5644E+04, 1.5650E+04,
     +  1.5656E+04, 1.5662E+04, 1.5668E+04, 1.5673E+04, 1.5679E+04,
     + 1.5685E+04, 1.5691E+04, 1.5697E+04, 1.5702E+04, 1.5708E+04,
     +  1.5714E+04, 1.5720E+04, 1.5726E+04, 1.5732E+04, 1.5737E+04,
     + 1.5743E+04, 1.5749E+04, 1.5755E+04, 1.5761E+04, 1.5766E+04,
     +  1.5772E+04, 1.5778E+04, 1.5784E+04, 1.5790E+04, 1.5796E+04,
     + 1.5801E+04, 1.5807E+04, 1.5813E+04, 1.5819E+04, 1.5825E+04,
     +  1.5830E+04, 1.5836E+04, 1.5842E+04, 1.5848E+04, 1.5854E+04,
     + 1.5860E+04, 1.5865E+04, 1.5871E+04, 1.5877E+04, 1.5883E+04,
     +  1.5889E+04, 1.5894E+04, 1.5900E+04, 1.5906E+04, 1.5912E+04,
     + 1.5918E+04, 1.5924E+04, 1.5929E+04, 1.5935E+04, 1.5941E+04,
     +  1.5947E+04, 1.5953E+04, 1.5959E+04, 1.5964E+04, 1.5970E+04,
     + 1.5976E+04, 1.5982E+04, 1.5988E+04, 1.5994E+04, 1.5999E+04,
     +  1.6005E+04, 1.6011E+04, 1.6017E+04, 1.6023E+04, 1.6029E+04,
     + 1.6034E+04, 1.6040E+04, 1.6046E+04, 1.6052E+04, 1.6058E+04,
     +  1.6064E+04, 1.6069E+04, 1.6075E+04, 1.6081E+04, 1.6087E+04,
     + 1.6093E+04, 1.6099E+04, 1.6104E+04, 1.6110E+04, 1.6116E+04,
     +  1.6122E+04, 1.6128E+04, 1.6134E+04, 1.6139E+04, 1.6145E+04,
     + 1.6151E+04, 1.6157E+04, 1.6163E+04, 1.6169E+04, 1.6174E+04,
     +  1.6180E+04, 1.6186E+04, 1.6192E+04, 1.6198E+04, 1.6204E+04,
     + 1.6209E+04, 1.6215E+04, 1.6221E+04, 1.6227E+04, 1.6233E+04,
     +  1.6239E+04, 1.6244E+04, 1.6250E+04, 1.6256E+04, 1.6262E+04,
     + 1.6268E+04, 1.6274E+04, 1.6280E+04, 1.6285E+04, 1.6291E+04,
     +  1.6297E+04, 1.6303E+04, 1.6309E+04, 1.6315E+04, 1.6320E+04,
     + 1.6326E+04, 1.6332E+04, 1.6338E+04, 1.6344E+04, 1.6350E+04,
     +  1.6356E+04, 1.6361E+04, 1.6367E+04, 1.6373E+04, 1.6379E+04,
     + 1.6385E+04, 1.6391E+04, 1.6396E+04, 1.6402E+04, 1.6408E+04,
     +  1.6414E+04, 1.6420E+04, 1.6426E+04, 1.6432E+04, 1.6437E+04,
     + 1.6443E+04, 1.6449E+04, 1.6455E+04, 1.6461E+04, 1.6467E+04,
     +  1.6473E+04, 1.6478E+04, 1.6484E+04, 1.6490E+04, 1.6496E+04,
     + 1.6502E+04, 1.6508E+04, 1.6514E+04, 1.6519E+04, 1.6525E+04,
     +  1.6531E+04, 1.6537E+04, 1.6543E+04, 1.6549E+04, 1.6555E+04,
     + 1.6560E+04, 1.6566E+04, 1.6572E+04, 1.6578E+04, 1.6584E+04,
     +  1.6590E+04, 1.6596E+04, 1.6601E+04, 1.6607E+04, 1.6613E+04,
     + 1.6619E+04, 1.6625E+04, 1.6631E+04, 1.6637E+04, 1.6642E+04,
     +  1.6648E+04, 1.6654E+04, 1.6660E+04, 1.6666E+04, 1.6672E+04,
     + 1.6678E+04, 1.6684E+04, 1.6689E+04, 1.6695E+04, 1.6701E+04,
     +  1.6707E+04, 1.6713E+04, 1.6719E+04, 1.6725E+04, 1.6730E+04,
     + 1.6736E+04, 1.6742E+04, 1.6748E+04, 1.6754E+04, 1.6760E+04,
     +  1.6766E+04, 1.6772E+04, 1.6777E+04, 1.6783E+04, 1.6789E+04,
     + 1.6795E+04, 1.6801E+04, 1.6807E+04, 1.6813E+04, 1.6818E+04,
     +  1.6824E+04, 1.6830E+04, 1.6836E+04, 1.6842E+04, 1.6848E+04,
     + 1.6854E+04, 1.6860E+04, 1.6865E+04, 1.6871E+04, 1.6877E+04,
     +  1.6883E+04, 1.6889E+04, 1.6895E+04, 1.6901E+04, 1.6907E+04,
     + 1.6912E+04, 1.6918E+04, 1.6924E+04, 1.6930E+04, 1.6936E+04,
     +  1.6942E+04, 1.6948E+04, 1.6954E+04, 1.6959E+04, 1.6965E+04,
     + 1.6971E+04, 1.6977E+04, 1.6983E+04, 1.6989E+04, 1.6995E+04,
     +  1.7001E+04, 1.7006E+04, 1.7012E+04, 1.7018E+04, 1.7024E+04,
     + 1.7030E+04, 1.7036E+04, 1.7042E+04, 1.7048E+04, 1.7053E+04,
     +  1.7059E+04, 1.7065E+04, 1.7071E+04, 1.7077E+04, 1.7083E+04,
     + 1.7089E+04, 1.7095E+04, 1.7101E+04, 1.7106E+04, 1.7112E+04,
     +  1.7118E+04, 1.7124E+04, 1.7130E+04, 1.7136E+04, 1.7142E+04,
     + 1.7148E+04, 1.7154E+04, 1.7159E+04, 1.7165E+04, 1.7171E+04,
     +  1.7177E+04, 1.7183E+04, 1.7189E+04, 1.7195E+04, 1.7201E+04,
     + 1.7206E+04, 1.7212E+04, 1.7218E+04, 1.7224E+04, 1.7230E+04,
     +  1.7236E+04, 1.7242E+04, 1.7248E+04, 1.7254E+04, 1.7259E+04,
     + 1.7265E+04, 1.7271E+04, 1.7277E+04, 1.7283E+04, 1.7289E+04,
     +  1.7295E+04, 1.7301E+04, 1.7307E+04, 1.7313E+04, 1.7318E+04,
     + 1.7324E+04, 1.7330E+04, 1.7336E+04, 1.7342E+04, 1.7348E+04,
     +  1.7354E+04, 1.7360E+04, 1.7366E+04, 1.7371E+04, 1.7377E+04,
     + 1.7383E+04, 1.7389E+04, 1.7395E+04, 1.7401E+04, 1.7407E+04,
     +  1.7413E+04, 1.7419E+04, 1.7425E+04, 1.7430E+04, 1.7436E+04,
     + 1.7442E+04, 1.7448E+04, 1.7454E+04, 1.7460E+04, 1.7466E+04,
     +  1.7472E+04, 1.7478E+04, 1.7483E+04, 1.7489E+04, 1.7495E+04,
     + 1.7501E+04, 1.7507E+04, 1.7513E+04, 1.7519E+04, 1.7525E+04,
     +  1.7531E+04, 1.7537E+04, 1.7542E+04, 1.7548E+04, 1.7554E+04,
     + 1.7560E+04, 1.7566E+04, 1.7572E+04, 1.7578E+04, 1.7584E+04,
     +  1.7590E+04, 1.7596E+04, 1.7602E+04, 1.7607E+04, 1.7613E+04,
     + 1.7619E+04, 1.7625E+04, 1.7631E+04, 1.7637E+04, 1.7643E+04,
     +  1.7649E+04, 1.7655E+04, 1.7661E+04, 1.7666E+04, 1.7672E+04,
     + 1.7678E+04, 1.7684E+04, 1.7690E+04, 1.7696E+04, 1.7702E+04,
     +  1.7708E+04, 1.7714E+04, 1.7720E+04, 1.7726E+04, 1.7731E+04,
     + 1.7737E+04, 1.7743E+04, 1.7749E+04, 1.7755E+04, 1.7761E+04,
     +  1.7767E+04, 1.7773E+04, 1.7779E+04, 1.7785E+04, 1.7791E+04,
     + 1.7796E+04, 1.7802E+04, 1.7808E+04, 1.7814E+04, 1.7820E+04,
     +  1.7826E+04, 1.7832E+04, 1.7838E+04, 1.7844E+04, 1.7850E+04,
     + 1.7856E+04, 1.7862E+04, 1.7867E+04, 1.7873E+04, 1.7879E+04,
     +  1.7885E+04, 1.7891E+04, 1.7897E+04, 1.7903E+04, 1.7909E+04,
     + 1.7915E+04, 1.7921E+04, 1.7927E+04, 1.7933E+04, 1.7938E+04,
     +  1.7944E+04, 1.7950E+04, 1.7956E+04, 1.7962E+04, 1.7968E+04,
     + 1.7974E+04, 1.7980E+04, 1.7986E+04, 1.7992E+04, 1.7998E+04,
     +  1.8004E+04, 1.8009E+04, 1.8015E+04, 1.8021E+04, 1.8027E+04,
     + 1.8033E+04, 1.8039E+04, 1.8045E+04, 1.8051E+04, 1.8057E+04,
     +  1.8063E+04, 1.8069E+04, 1.8075E+04, 1.8080E+04, 1.8086E+04,
     + 1.8092E+04, 1.8098E+04, 1.8104E+04, 1.8110E+04, 1.8116E+04,
     +  1.8122E+04, 1.8128E+04, 1.8134E+04, 1.8140E+04, 1.8146E+04,
     + 1.8152E+04, 1.8157E+04, 1.8163E+04, 1.8169E+04, 1.8175E+04,
     +  1.8181E+04, 1.8187E+04, 1.8193E+04, 1.8199E+04, 1.8205E+04,
     + 1.8211E+04, 1.8217E+04, 1.8223E+04, 1.8229E+04, 1.8235E+04,
     +  1.8240E+04, 1.8246E+04, 1.8252E+04, 1.8258E+04, 1.8264E+04,
     + 1.8270E+04, 1.8276E+04, 1.8282E+04, 1.8288E+04, 1.8294E+04,
     +  1.8300E+04, 1.8306E+04, 1.8312E+04, 1.8318E+04, 1.8323E+04,
     + 1.8329E+04, 1.8335E+04, 1.8341E+04, 1.8347E+04, 1.8353E+04,
     +  1.8359E+04, 1.8365E+04, 1.8371E+04, 1.8377E+04, 1.8383E+04,
     + 1.8389E+04, 1.8395E+04, 1.8401E+04, 1.8407E+04, 1.8412E+04,
     +  1.8418E+04, 1.8424E+04, 1.8430E+04, 1.8436E+04, 1.8442E+04,
     + 1.8448E+04, 1.8454E+04, 1.8460E+04, 1.8466E+04, 1.8472E+04,
     +  1.8478E+04, 1.8484E+04, 1.8490E+04, 1.8496E+04, 1.8501E+04,
     + 1.8507E+04, 1.8513E+04, 1.8519E+04, 1.8525E+04, 1.8531E+04,
     +  1.8537E+04, 1.8543E+04, 1.8549E+04, 1.8555E+04, 1.8561E+04,
     + 1.8567E+04, 1.8573E+04, 1.8579E+04, 1.8585E+04, 1.8591E+04,
     +  1.8596E+04, 1.8602E+04, 1.8608E+04, 1.8614E+04, 1.8620E+04,
     + 1.8626E+04, 1.8632E+04, 1.8638E+04, 1.8644E+04, 1.8650E+04,
     +  1.8656E+04, 1.8662E+04, 1.8668E+04, 1.8674E+04, 1.8680E+04,
     + 1.8686E+04, 1.8692E+04, 1.8697E+04, 1.8703E+04, 1.8709E+04,
     +  1.8715E+04, 1.8721E+04, 1.8727E+04, 1.8733E+04, 1.8739E+04,
     + 1.8745E+04, 1.8751E+04, 1.8757E+04, 1.8763E+04/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_HCN   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 3),QofT( 3,NT),Q(NT)
      data (Tmx(I),I=1,3)/3500.,3500.,3500./
      data xgj/ 6.,12.,4./
c...      HCN
c...        --       124
      data (QofT( 1,J),J=1,2408)/ 6.2559E+00, 8.1968E+00, 1.0795E+01,
     +  1.3522E+01, 1.6291E+01, 1.9079E+01, 2.1877E+01, 2.4681E+01,
     + 2.7489E+01, 3.0300E+01, 3.3113E+01, 3.5927E+01, 3.8743E+01,
     +  4.1559E+01, 4.4376E+01, 4.7193E+01, 5.0011E+01, 5.2830E+01,
     + 5.5648E+01, 5.8467E+01, 6.4106E+01, 6.9745E+01, 7.5385E+01,
     +  8.1026E+01, 8.6666E+01, 9.2307E+01, 9.7949E+01, 1.0359E+02,
     + 1.0923E+02, 1.1487E+02, 1.2052E+02, 1.2616E+02, 1.3180E+02,
     +  1.3744E+02, 1.4308E+02, 1.4873E+02, 1.5437E+02, 1.6001E+02,
     + 1.6566E+02, 1.7130E+02, 1.7694E+02, 1.8258E+02, 1.8823E+02,
     +  1.9387E+02, 1.9951E+02, 2.0516E+02, 2.1080E+02, 2.1645E+02,
     + 2.2209E+02, 2.2773E+02, 2.3338E+02, 2.3902E+02, 2.4467E+02,
     +  2.5031E+02, 2.5596E+02, 2.6160E+02, 2.6725E+02, 2.7289E+02,
     + 2.7854E+02, 2.8419E+02, 2.8984E+02, 2.9549E+02, 3.0114E+02,
     +  3.0679E+02, 3.1245E+02, 3.1810E+02, 3.2376E+02, 3.2942E+02,
     + 3.3508E+02, 3.4074E+02, 3.4641E+02, 3.5208E+02, 3.5775E+02,
     +  3.6343E+02, 3.6911E+02, 3.7480E+02, 3.8048E+02, 3.8618E+02,
     + 3.9188E+02, 3.9758E+02, 4.0329E+02, 4.0901E+02, 4.1473E+02,
     +  4.2046E+02, 4.2619E+02, 4.3194E+02, 4.3769E+02, 4.4345E+02,
     + 4.4922E+02, 4.5500E+02, 4.6079E+02, 4.6658E+02, 4.7239E+02,
     +  4.7821E+02, 4.8404E+02, 4.8989E+02, 4.9574E+02, 5.0161E+02,
     + 5.0749E+02, 5.1338E+02, 5.1929E+02, 5.2521E+02, 5.3115E+02,
     +  5.3710E+02, 5.4307E+02, 5.4906E+02, 5.5506E+02, 5.6107E+02,
     + 5.6711E+02, 5.7316E+02, 5.7923E+02, 5.8532E+02, 5.9143E+02,
     +  5.9756E+02, 6.0371E+02, 6.0988E+02, 6.1607E+02, 6.2228E+02,
     + 6.2851E+02, 6.3476E+02, 6.4104E+02, 6.4734E+02, 6.5366E+02,
     +  6.6000E+02, 6.6637E+02, 6.7277E+02, 6.7919E+02, 6.8563E+02,
     + 6.9210E+02, 6.9859E+02, 7.0511E+02, 7.1166E+02, 7.1823E+02,
     +  7.2483E+02, 7.3146E+02, 7.3811E+02, 7.4480E+02, 7.5151E+02,
     + 7.5825E+02, 7.6502E+02, 7.7182E+02, 7.7865E+02, 7.8551E+02,
     +  7.9240E+02, 7.9932E+02, 8.0627E+02, 8.1325E+02, 8.2026E+02,
     + 8.2731E+02, 8.3439E+02, 8.4150E+02, 8.4864E+02, 8.5582E+02,
     +  8.6303E+02, 8.7027E+02, 8.7755E+02, 8.8486E+02, 8.9220E+02,
     + 8.9958E+02, 9.0700E+02, 9.1445E+02, 9.2193E+02, 9.2946E+02,
     +  9.3701E+02, 9.4461E+02, 9.5224E+02, 9.5991E+02, 9.6761E+02,
     + 9.7535E+02, 9.8313E+02, 9.9095E+02, 9.9880E+02, 1.0067E+03,
     +  1.0146E+03, 1.0226E+03, 1.0306E+03, 1.0387E+03, 1.0467E+03,
     + 1.0549E+03, 1.0630E+03, 1.0713E+03, 1.0795E+03, 1.0878E+03,
     +  1.0961E+03, 1.1045E+03, 1.1129E+03, 1.1214E+03, 1.1299E+03,
     + 1.1384E+03, 1.1470E+03, 1.1556E+03, 1.1643E+03, 1.1730E+03,
     +  1.1817E+03, 1.1905E+03, 1.1993E+03, 1.2082E+03, 1.2171E+03,
     + 1.2261E+03, 1.2351E+03, 1.2442E+03, 1.2533E+03, 1.2624E+03,
     +  1.2716E+03, 1.2809E+03, 1.2902E+03, 1.2995E+03, 1.3089E+03,
     + 1.3183E+03, 1.3277E+03, 1.3373E+03, 1.3468E+03, 1.3564E+03,
     +  1.3661E+03, 1.3758E+03, 1.3855E+03, 1.3953E+03, 1.4052E+03,
     + 1.4151E+03, 1.4250E+03, 1.4350E+03, 1.4451E+03, 1.4551E+03,
     +  1.4653E+03, 1.4755E+03, 1.4857E+03, 1.4960E+03, 1.5063E+03,
     + 1.5167E+03, 1.5272E+03, 1.5377E+03, 1.5482E+03, 1.5588E+03,
     +  1.5694E+03, 1.5801E+03, 1.5909E+03, 1.6017E+03, 1.6125E+03,
     + 1.6234E+03, 1.6344E+03, 1.6454E+03, 1.6564E+03, 1.6676E+03,
     +  1.6787E+03, 1.6899E+03, 1.7012E+03, 1.7125E+03, 1.7239E+03,
     + 1.7354E+03, 1.7469E+03, 1.7584E+03, 1.7700E+03, 1.7817E+03,
     +  1.7934E+03, 1.8051E+03, 1.8170E+03, 1.8288E+03, 1.8408E+03,
     + 1.8528E+03, 1.8648E+03, 1.8769E+03, 1.8891E+03, 1.9013E+03,
     +  1.9136E+03, 1.9259E+03, 1.9383E+03, 1.9507E+03, 1.9633E+03,
     + 1.9758E+03, 1.9884E+03, 2.0011E+03, 2.0139E+03, 2.0267E+03,
     +  2.0395E+03, 2.0525E+03, 2.0654E+03, 2.0785E+03, 2.0916E+03,
     + 2.1047E+03, 2.1180E+03, 2.1313E+03, 2.1446E+03, 2.1580E+03,
     +  2.1715E+03, 2.1850E+03, 2.1986E+03, 2.2123E+03, 2.2260E+03,
     + 2.2398E+03, 2.2536E+03, 2.2675E+03, 2.2815E+03, 2.2955E+03,
     +  2.3096E+03, 2.3238E+03, 2.3380E+03, 2.3523E+03, 2.3667E+03,
     + 2.3811E+03, 2.3956E+03, 2.4101E+03, 2.4247E+03, 2.4394E+03,
     +  2.4542E+03, 2.4690E+03, 2.4839E+03, 2.4988E+03, 2.5138E+03,
     + 2.5289E+03, 2.5441E+03, 2.5593E+03, 2.5746E+03, 2.5899E+03,
     +  2.6054E+03, 2.6209E+03, 2.6364E+03, 2.6521E+03, 2.6678E+03,
     + 2.6835E+03, 2.6994E+03, 2.7153E+03, 2.7313E+03, 2.7473E+03,
     +  2.7634E+03, 2.7796E+03, 2.7959E+03, 2.8122E+03, 2.8286E+03,
     + 2.8451E+03, 2.8617E+03, 2.8783E+03, 2.8950E+03, 2.9118E+03,
     +  2.9286E+03, 2.9455E+03, 2.9625E+03, 2.9796E+03, 2.9967E+03,
     + 3.0139E+03, 3.0312E+03, 3.0486E+03, 3.0660E+03, 3.0835E+03,
     +  3.1011E+03, 3.1188E+03, 3.1365E+03, 3.1543E+03, 3.1722E+03,
     + 3.1902E+03, 3.2082E+03, 3.2264E+03, 3.2446E+03, 3.2629E+03,
     +  3.2812E+03, 3.2997E+03, 3.3182E+03, 3.3368E+03, 3.3554E+03,
     + 3.3742E+03, 3.3930E+03, 3.4120E+03, 3.4309E+03, 3.4500E+03,
     +  3.4692E+03, 3.4884E+03, 3.5077E+03, 3.5271E+03, 3.5466E+03,
     + 3.5662E+03, 3.5858E+03, 3.6056E+03, 3.6254E+03, 3.6453E+03,
     +  3.6653E+03, 3.6853E+03, 3.7055E+03, 3.7257E+03, 3.7460E+03,
     + 3.7664E+03, 3.7869E+03, 3.8075E+03, 3.8281E+03, 3.8489E+03,
     +  3.8697E+03, 3.8906E+03, 3.9116E+03, 3.9327E+03, 3.9539E+03,
     + 3.9752E+03, 3.9965E+03, 4.0179E+03, 4.0395E+03, 4.0611E+03,
     +  4.0828E+03, 4.1046E+03, 4.1265E+03, 4.1485E+03, 4.1705E+03,
     + 4.1927E+03, 4.2149E+03, 4.2373E+03, 4.2597E+03, 4.2822E+03,
     +  4.3048E+03, 4.3275E+03, 4.3503E+03, 4.3732E+03, 4.3962E+03,
     + 4.4193E+03, 4.4424E+03, 4.4657E+03, 4.4891E+03, 4.5125E+03,
     +  4.5361E+03, 4.5597E+03, 4.5834E+03, 4.6073E+03, 4.6312E+03,
     + 4.6552E+03, 4.6794E+03, 4.7036E+03, 4.7279E+03, 4.7523E+03,
     +  4.7768E+03, 4.8014E+03, 4.8261E+03, 4.8509E+03, 4.8758E+03,
     + 4.9008E+03, 4.9259E+03, 4.9511E+03, 4.9765E+03, 5.0019E+03,
     +  5.0274E+03, 5.0530E+03, 5.0787E+03, 5.1045E+03, 5.1304E+03,
     + 5.1564E+03, 5.1825E+03, 5.2087E+03, 5.2351E+03, 5.2615E+03,
     +  5.2880E+03, 5.3147E+03, 5.3414E+03, 5.3682E+03, 5.3952E+03,
     + 5.4222E+03, 5.4494E+03, 5.4767E+03, 5.5040E+03, 5.5315E+03,
     +  5.5591E+03, 5.5868E+03, 5.6146E+03, 5.6425E+03, 5.6705E+03,
     + 5.6987E+03, 5.7269E+03, 5.7552E+03, 5.7837E+03, 5.8123E+03,
     +  5.8409E+03, 5.8697E+03, 5.8986E+03, 5.9276E+03, 5.9568E+03,
     + 5.9860E+03, 6.0153E+03, 6.0448E+03, 6.0744E+03, 6.1040E+03,
     +  6.1338E+03, 6.1637E+03, 6.1938E+03, 6.2239E+03, 6.2542E+03,
     + 6.2845E+03, 6.3150E+03, 6.3456E+03, 6.3763E+03, 6.4072E+03,
     +  6.4381E+03, 6.4692E+03, 6.5004E+03, 6.5317E+03, 6.5631E+03,
     + 6.5947E+03, 6.6263E+03, 6.6581E+03, 6.6900E+03, 6.7220E+03,
     +  6.7542E+03, 6.7864E+03, 6.8188E+03, 6.8513E+03, 6.8839E+03,
     + 6.9167E+03, 6.9496E+03, 6.9825E+03, 7.0157E+03, 7.0489E+03,
     +  7.0823E+03, 7.1158E+03, 7.1494E+03, 7.1831E+03, 7.2170E+03,
     + 7.2510E+03, 7.2851E+03, 7.3202E+03, 7.3557E+03, 7.3913E+03,
     +  7.4270E+03, 7.4629E+03, 7.4990E+03, 7.5351E+03, 7.5714E+03,
     + 7.6079E+03, 7.6444E+03, 7.6811E+03, 7.7180E+03, 7.7550E+03,
     +  7.7921E+03, 7.8294E+03, 7.8668E+03, 7.9043E+03, 7.9420E+03,
     + 7.9798E+03, 8.0178E+03, 8.0559E+03, 8.0941E+03, 8.1325E+03,
     +  8.1710E+03, 8.2097E+03, 8.2485E+03, 8.2875E+03, 8.3266E+03,
     + 8.3658E+03, 8.4052E+03, 8.4447E+03, 8.4844E+03, 8.5242E+03,
     +  8.5642E+03, 8.6043E+03, 8.6446E+03, 8.6850E+03, 8.7255E+03,
     + 8.7663E+03, 8.8071E+03, 8.8481E+03, 8.8893E+03, 8.9306E+03,
     +  8.9720E+03, 9.0136E+03, 9.0554E+03, 9.0973E+03, 9.1393E+03,
     + 9.1815E+03, 9.2239E+03, 9.2664E+03, 9.3091E+03, 9.3519E+03,
     +  9.3949E+03, 9.4380E+03, 9.4813E+03, 9.5247E+03, 9.5683E+03,
     + 9.6121E+03, 9.6560E+03, 9.7000E+03, 9.7443E+03, 9.7886E+03,
     +  9.8332E+03, 9.8779E+03, 9.9227E+03, 9.9677E+03, 1.0013E+04,
     + 1.0058E+04, 1.0104E+04, 1.0149E+04, 1.0195E+04, 1.0241E+04,
     +  1.0287E+04, 1.0334E+04, 1.0380E+04, 1.0427E+04, 1.0474E+04,
     + 1.0521E+04, 1.0568E+04, 1.0615E+04, 1.0662E+04, 1.0710E+04,
     +  1.0758E+04, 1.0806E+04, 1.0854E+04, 1.0902E+04, 1.0951E+04,
     + 1.0999E+04, 1.1048E+04, 1.1097E+04, 1.1146E+04, 1.1196E+04,
     +  1.1245E+04, 1.1295E+04, 1.1344E+04, 1.1394E+04, 1.1445E+04,
     + 1.1495E+04, 1.1546E+04, 1.1596E+04, 1.1647E+04, 1.1698E+04,
     +  1.1749E+04, 1.1801E+04, 1.1852E+04, 1.1904E+04, 1.1956E+04,
     + 1.2008E+04, 1.2060E+04, 1.2113E+04, 1.2165E+04, 1.2218E+04,
     +  1.2271E+04, 1.2324E+04, 1.2378E+04, 1.2431E+04, 1.2485E+04,
     + 1.2539E+04, 1.2593E+04, 1.2647E+04, 1.2702E+04, 1.2757E+04,
     +  1.2811E+04, 1.2866E+04, 1.2922E+04, 1.2977E+04, 1.3033E+04,
     + 1.3088E+04, 1.3144E+04, 1.3200E+04, 1.3257E+04, 1.3313E+04,
     +  1.3370E+04, 1.3427E+04, 1.3484E+04, 1.3541E+04, 1.3599E+04,
     + 1.3656E+04, 1.3714E+04, 1.3772E+04, 1.3830E+04, 1.3889E+04,
     +  1.3947E+04, 1.4006E+04, 1.4065E+04, 1.4124E+04, 1.4184E+04,
     + 1.4244E+04, 1.4303E+04, 1.4363E+04, 1.4423E+04, 1.4484E+04,
     +  1.4544E+04, 1.4605E+04, 1.4666E+04, 1.4727E+04, 1.4789E+04,
     + 1.4850E+04, 1.4912E+04, 1.4974E+04, 1.5036E+04, 1.5099E+04,
     +  1.5161E+04, 1.5224E+04, 1.5287E+04, 1.5351E+04, 1.5414E+04,
     + 1.5478E+04, 1.5541E+04, 1.5606E+04, 1.5670E+04, 1.5734E+04,
     +  1.5799E+04, 1.5864E+04, 1.5929E+04, 1.5994E+04, 1.6060E+04,
     + 1.6125E+04, 1.6191E+04, 1.6258E+04, 1.6324E+04, 1.6391E+04,
     +  1.6457E+04, 1.6524E+04, 1.6592E+04, 1.6659E+04, 1.6727E+04,
     + 1.6795E+04, 1.6863E+04, 1.6931E+04, 1.7000E+04, 1.7068E+04,
     +  1.7137E+04, 1.7206E+04, 1.7276E+04, 1.7345E+04, 1.7415E+04,
     + 1.7485E+04, 1.7556E+04, 1.7626E+04, 1.7697E+04, 1.7768E+04,
     +  1.7839E+04, 1.7911E+04, 1.7982E+04, 1.8054E+04, 1.8126E+04,
     + 1.8199E+04, 1.8271E+04, 1.8344E+04, 1.8417E+04, 1.8490E+04,
     +  1.8564E+04, 1.8637E+04, 1.8711E+04, 1.8785E+04, 1.8860E+04,
     + 1.8934E+04, 1.9009E+04, 1.9084E+04, 1.9160E+04, 1.9235E+04,
     +  1.9311E+04, 1.9387E+04, 1.9463E+04, 1.9540E+04, 1.9617E+04,
     + 1.9694E+04, 1.9771E+04, 1.9848E+04, 1.9926E+04, 2.0004E+04,
     +  2.0082E+04, 2.0161E+04, 2.0239E+04, 2.0318E+04, 2.0397E+04,
     + 2.0477E+04, 2.0556E+04, 2.0636E+04, 2.0716E+04, 2.0797E+04,
     +  2.0877E+04, 2.0958E+04, 2.1039E+04, 2.1121E+04, 2.1202E+04,
     + 2.1284E+04, 2.1366E+04, 2.1449E+04, 2.1531E+04, 2.1614E+04,
     +  2.1697E+04, 2.1781E+04, 2.1864E+04, 2.1948E+04, 2.2032E+04,
     + 2.2117E+04, 2.2201E+04, 2.2286E+04, 2.2371E+04, 2.2457E+04,
     +  2.2543E+04, 2.2628E+04, 2.2715E+04, 2.2801E+04, 2.2888E+04,
     + 2.2975E+04, 2.3062E+04, 2.3150E+04, 2.3237E+04, 2.3325E+04,
     +  2.3414E+04, 2.3502E+04, 2.3591E+04, 2.3680E+04, 2.3769E+04,
     + 2.3859E+04, 2.3949E+04, 2.4039E+04, 2.4130E+04, 2.4220E+04,
     +  2.4311E+04, 2.4402E+04, 2.4494E+04, 2.4586E+04, 2.4678E+04,
     + 2.4770E+04, 2.4863E+04, 2.4956E+04, 2.5049E+04, 2.5142E+04,
     +  2.5236E+04, 2.5330E+04, 2.5424E+04, 2.5519E+04, 2.5613E+04,
     + 2.5708E+04, 2.5804E+04, 2.5899E+04, 2.5995E+04, 2.6092E+04,
     +  2.6188E+04, 2.6285E+04, 2.6382E+04, 2.6479E+04, 2.6577E+04,
     + 2.6675E+04, 2.6773E+04, 2.6871E+04, 2.6970E+04, 2.7069E+04,
     +  2.7169E+04, 2.7268E+04, 2.7368E+04, 2.7468E+04, 2.7569E+04,
     + 2.7670E+04, 2.7771E+04, 2.7872E+04, 2.7974E+04, 2.8076E+04,
     +  2.8178E+04, 2.8281E+04, 2.8383E+04, 2.8487E+04, 2.8590E+04,
     + 2.8694E+04, 2.8798E+04, 2.8902E+04, 2.9007E+04, 2.9112E+04,
     +  2.9217E+04, 2.9322E+04, 2.9428E+04, 2.9534E+04, 2.9641E+04,
     + 2.9748E+04, 2.9855E+04, 2.9962E+04, 3.0070E+04, 3.0178E+04,
     +  3.0286E+04, 3.0395E+04, 3.0503E+04, 3.0613E+04, 3.0722E+04,
     + 3.0832E+04, 3.0942E+04, 3.1053E+04, 3.1163E+04, 3.1274E+04,
     +  3.1386E+04, 3.1498E+04, 3.1610E+04, 3.1722E+04, 3.1835E+04,
     + 3.1948E+04, 3.2061E+04, 3.2174E+04, 3.2288E+04, 3.2403E+04,
     +  3.2517E+04, 3.2632E+04, 3.2747E+04, 3.2863E+04, 3.2979E+04,
     + 3.3095E+04, 3.3211E+04, 3.3328E+04, 3.3445E+04, 3.3563E+04,
     +  3.3681E+04, 3.3799E+04, 3.3917E+04, 3.4036E+04, 3.4155E+04,
     + 3.4275E+04, 3.4394E+04, 3.4515E+04, 3.4635E+04, 3.4756E+04,
     +  3.4877E+04, 3.4998E+04, 3.5120E+04, 3.5242E+04, 3.5365E+04,
     + 3.5488E+04, 3.5611E+04, 3.5734E+04, 3.5858E+04, 3.5982E+04,
     +  3.6107E+04, 3.6232E+04, 3.6357E+04, 3.6482E+04, 3.6608E+04,
     + 3.6734E+04, 3.6861E+04, 3.6988E+04, 3.7115E+04, 3.7243E+04,
     +  3.7371E+04, 3.7499E+04, 3.7628E+04, 3.7757E+04, 3.7886E+04,
     + 3.8016E+04, 3.8146E+04, 3.8276E+04, 3.8407E+04, 3.8538E+04,
     +  3.8669E+04, 3.8801E+04, 3.8934E+04, 3.9066E+04, 3.9199E+04,
     + 3.9332E+04, 3.9466E+04, 3.9600E+04, 3.9734E+04, 3.9869E+04,
     +  4.0004E+04, 4.0139E+04, 4.0275E+04, 4.0411E+04, 4.0548E+04,
     + 4.0685E+04, 4.0822E+04, 4.0960E+04, 4.1098E+04, 4.1236E+04,
     +  4.1375E+04, 4.1514E+04, 4.1653E+04, 4.1793E+04, 4.1933E+04,
     + 4.2074E+04, 4.2215E+04, 4.2356E+04, 4.2498E+04, 4.2640E+04,
     +  4.2782E+04, 4.2925E+04, 4.3068E+04, 4.3212E+04, 4.3356E+04,
     + 4.3500E+04, 4.3645E+04, 4.3790E+04, 4.3936E+04, 4.4082E+04,
     +  4.4228E+04, 4.4375E+04, 4.4522E+04, 4.4669E+04, 4.4817E+04,
     + 4.4965E+04, 4.5114E+04, 4.5263E+04, 4.5412E+04, 4.5562E+04,
     +  4.5712E+04, 4.5863E+04, 4.6014E+04, 4.6165E+04, 4.6317E+04,
     + 4.6469E+04, 4.6621E+04, 4.6774E+04, 4.6928E+04, 4.7081E+04,
     +  4.7235E+04, 4.7390E+04, 4.7545E+04, 4.7700E+04, 4.7856E+04,
     + 4.8012E+04, 4.8169E+04, 4.8326E+04, 4.8483E+04, 4.8641E+04,
     +  4.8799E+04, 4.8957E+04, 4.9116E+04, 4.9276E+04, 4.9436E+04,
     + 4.9596E+04, 4.9756E+04, 4.9917E+04, 5.0079E+04, 5.0241E+04,
     +  5.0403E+04, 5.0566E+04, 5.0729E+04, 5.0892E+04, 5.1056E+04,
     + 5.1221E+04, 5.1386E+04, 5.1551E+04, 5.1716E+04, 5.1882E+04,
     +  5.2049E+04, 5.2216E+04, 5.2383E+04, 5.2551E+04, 5.2719E+04,
     + 5.2888E+04, 5.3057E+04, 5.3226E+04, 5.3396E+04, 5.3566E+04,
     +  5.3737E+04, 5.3908E+04, 5.4080E+04, 5.4252E+04, 5.4425E+04,
     + 5.4598E+04, 5.4771E+04, 5.4945E+04, 5.5119E+04, 5.5294E+04,
     +  5.5469E+04, 5.5644E+04, 5.5820E+04, 5.5997E+04, 5.6174E+04,
     + 5.6351E+04, 5.6529E+04, 5.6707E+04, 5.6886E+04, 5.7065E+04,
     +  5.7244E+04, 5.7424E+04, 5.7605E+04, 5.7786E+04, 5.7967E+04,
     + 5.8149E+04, 5.8331E+04, 5.8514E+04, 5.8697E+04, 5.8881E+04,
     +  5.9065E+04, 5.9249E+04, 5.9434E+04, 5.9620E+04, 5.9806E+04,
     + 5.9992E+04, 6.0179E+04, 6.0366E+04, 6.0554E+04, 6.0742E+04,
     +  6.0931E+04, 6.1120E+04, 6.1310E+04, 6.1500E+04, 6.1690E+04,
     + 6.1881E+04, 6.2073E+04, 6.2265E+04, 6.2457E+04, 6.2650E+04,
     +  6.2843E+04, 6.3037E+04, 6.3231E+04, 6.3426E+04, 6.3622E+04,
     + 6.3817E+04, 6.4014E+04, 6.4210E+04, 6.4407E+04, 6.4605E+04,
     +  6.4803E+04, 6.5002E+04, 6.5201E+04, 6.5401E+04, 6.5601E+04,
     + 6.5801E+04, 6.6002E+04, 6.6204E+04, 6.6406E+04, 6.6608E+04,
     +  6.6812E+04, 6.7015E+04, 6.7219E+04, 6.7424E+04, 6.7628E+04,
     + 6.7834E+04, 6.8040E+04, 6.8246E+04, 6.8453E+04, 6.8661E+04,
     +  6.8869E+04, 6.9077E+04, 6.9286E+04, 6.9496E+04, 6.9706E+04,
     + 6.9916E+04, 7.0127E+04, 7.0339E+04, 7.0551E+04, 7.0763E+04,
     +  7.0976E+04, 7.1190E+04, 7.1404E+04, 7.1618E+04, 7.1833E+04,
     + 7.2049E+04, 7.2265E+04, 7.2482E+04, 7.2699E+04, 7.2916E+04,
     +  7.3134E+04, 7.3353E+04, 7.3572E+04, 7.3792E+04, 7.4012E+04,
     + 7.4233E+04, 7.4454E+04, 7.4676E+04, 7.4898E+04, 7.5121E+04,
     +  7.5344E+04, 7.5568E+04, 7.5793E+04, 7.6018E+04, 7.6243E+04,
     + 7.6469E+04, 7.6696E+04, 7.6923E+04, 7.7150E+04, 7.7378E+04,
     +  7.7607E+04, 7.7836E+04, 7.8066E+04, 7.8296E+04, 7.8527E+04,
     + 7.8758E+04, 7.8990E+04, 7.9223E+04, 7.9456E+04, 7.9689E+04,
     +  7.9923E+04, 8.0158E+04, 8.0393E+04, 8.0629E+04, 8.0865E+04,
     + 8.1102E+04, 8.1339E+04, 8.1577E+04, 8.1816E+04, 8.2055E+04,
     +  8.2294E+04, 8.2534E+04, 8.2775E+04, 8.3016E+04, 8.3258E+04,
     + 8.3500E+04, 8.3743E+04, 8.3987E+04, 8.4231E+04, 8.4475E+04,
     +  8.4721E+04, 8.4966E+04, 8.5213E+04, 8.5460E+04, 8.5707E+04,
     + 8.5955E+04, 8.6203E+04, 8.6453E+04, 8.6702E+04, 8.6953E+04,
     +  8.7203E+04, 8.7455E+04, 8.7707E+04, 8.7959E+04, 8.8213E+04,
     + 8.8466E+04, 8.8721E+04, 8.8975E+04, 8.9231E+04, 8.9487E+04,
     +  8.9744E+04, 9.0001E+04, 9.0259E+04, 9.0517E+04, 9.0776E+04,
     + 9.1036E+04, 9.1296E+04, 9.1556E+04, 9.1818E+04, 9.2080E+04,
     +  9.2342E+04, 9.2605E+04, 9.2869E+04, 9.3133E+04, 9.3398E+04,
     + 9.3664E+04, 9.3930E+04, 9.4196E+04, 9.4464E+04, 9.4732E+04,
     +  9.5000E+04, 9.5269E+04, 9.5539E+04, 9.5809E+04, 9.6080E+04,
     + 9.6352E+04, 9.6624E+04, 9.6896E+04, 9.7170E+04, 9.7444E+04,
     +  9.7718E+04, 9.7994E+04, 9.8269E+04, 9.8546E+04, 9.8823E+04,
     + 9.9100E+04, 9.9379E+04, 9.9658E+04, 9.9937E+04, 1.0022E+05,
     +  1.0050E+05, 1.0078E+05, 1.0106E+05, 1.0134E+05, 1.0163E+05,
     + 1.0191E+05, 1.0220E+05, 1.0248E+05, 1.0277E+05, 1.0305E+05,
     +  1.0334E+05, 1.0363E+05, 1.0392E+05, 1.0420E+05, 1.0449E+05,
     + 1.0478E+05, 1.0508E+05, 1.0537E+05, 1.0566E+05, 1.0595E+05,
     +  1.0625E+05, 1.0654E+05, 1.0683E+05, 1.0713E+05, 1.0743E+05,
     + 1.0772E+05, 1.0802E+05, 1.0832E+05, 1.0862E+05, 1.0891E+05,
     +  1.0921E+05, 1.0952E+05, 1.0982E+05, 1.1012E+05, 1.1042E+05,
     + 1.1072E+05, 1.1103E+05, 1.1133E+05, 1.1164E+05, 1.1194E+05,
     +  1.1225E+05, 1.1256E+05, 1.1286E+05, 1.1317E+05, 1.1348E+05,
     + 1.1379E+05, 1.1410E+05, 1.1441E+05, 1.1473E+05, 1.1504E+05,
     +  1.1535E+05, 1.1567E+05, 1.1598E+05, 1.1630E+05, 1.1661E+05,
     + 1.1693E+05, 1.1724E+05, 1.1756E+05, 1.1788E+05, 1.1820E+05,
     +  1.1852E+05, 1.1884E+05, 1.1916E+05, 1.1948E+05, 1.1981E+05,
     + 1.2013E+05, 1.2045E+05, 1.2078E+05, 1.2111E+05, 1.2143E+05,
     +  1.2176E+05, 1.2209E+05, 1.2241E+05, 1.2274E+05, 1.2307E+05,
     + 1.2340E+05, 1.2373E+05, 1.2407E+05, 1.2440E+05, 1.2473E+05,
     +  1.2507E+05, 1.2540E+05, 1.2574E+05, 1.2607E+05, 1.2641E+05,
     + 1.2675E+05, 1.2708E+05, 1.2742E+05, 1.2776E+05, 1.2810E+05,
     +  1.2844E+05, 1.2879E+05, 1.2913E+05, 1.2947E+05, 1.2982E+05,
     + 1.3016E+05, 1.3051E+05, 1.3085E+05, 1.3120E+05, 1.3155E+05,
     +  1.3189E+05, 1.3224E+05, 1.3259E+05, 1.3294E+05, 1.3329E+05,
     + 1.3365E+05, 1.3400E+05, 1.3435E+05, 1.3471E+05, 1.3506E+05,
     +  1.3542E+05, 1.3577E+05, 1.3613E+05, 1.3649E+05, 1.3685E+05,
     + 1.3721E+05, 1.3757E+05, 1.3793E+05, 1.3829E+05, 1.3865E+05,
     +  1.3901E+05, 1.3938E+05, 1.3974E+05, 1.4011E+05, 1.4047E+05,
     + 1.4084E+05, 1.4121E+05, 1.4158E+05, 1.4195E+05, 1.4232E+05,
     +  1.4269E+05, 1.4306E+05, 1.4343E+05, 1.4380E+05, 1.4418E+05,
     + 1.4455E+05, 1.4493E+05, 1.4530E+05, 1.4568E+05, 1.4606E+05,
     +  1.4644E+05, 1.4681E+05, 1.4719E+05, 1.4758E+05, 1.4796E+05,
     + 1.4834E+05, 1.4872E+05, 1.4911E+05, 1.4949E+05, 1.4988E+05,
     +  1.5026E+05, 1.5065E+05, 1.5104E+05, 1.5143E+05, 1.5181E+05,
     + 1.5220E+05, 1.5260E+05, 1.5299E+05, 1.5338E+05, 1.5377E+05,
     +  1.5417E+05, 1.5456E+05, 1.5496E+05, 1.5535E+05, 1.5575E+05,
     + 1.5615E+05, 1.5655E+05, 1.5695E+05, 1.5735E+05, 1.5775E+05,
     +  1.5815E+05, 1.5855E+05, 1.5896E+05, 1.5936E+05, 1.5977E+05,
     + 1.6017E+05, 1.6058E+05, 1.6099E+05, 1.6140E+05, 1.6181E+05,
     +  1.6222E+05, 1.6263E+05, 1.6304E+05, 1.6345E+05, 1.6387E+05,
     + 1.6428E+05, 1.6470E+05, 1.6511E+05, 1.6553E+05, 1.6595E+05,
     +  1.6637E+05, 1.6678E+05, 1.6721E+05, 1.6763E+05, 1.6805E+05,
     + 1.6847E+05, 1.6889E+05, 1.6932E+05, 1.6974E+05, 1.7017E+05,
     +  1.7060E+05, 1.7102E+05, 1.7145E+05, 1.7188E+05, 1.7231E+05,
     + 1.7274E+05, 1.7318E+05, 1.7361E+05, 1.7404E+05, 1.7448E+05,
     +  1.7491E+05, 1.7535E+05, 1.7579E+05, 1.7622E+05, 1.7666E+05,
     + 1.7710E+05, 1.7754E+05, 1.7799E+05, 1.7843E+05, 1.7887E+05,
     +  1.7932E+05, 1.7976E+05, 1.8021E+05, 1.8065E+05, 1.8110E+05,
     + 1.8155E+05, 1.8200E+05, 1.8245E+05, 1.8290E+05, 1.8335E+05,
     +  1.8380E+05, 1.8426E+05, 1.8471E+05, 1.8517E+05, 1.8562E+05,
     + 1.8608E+05, 1.8654E+05, 1.8700E+05, 1.8746E+05, 1.8792E+05,
     +  1.8838E+05, 1.8884E+05, 1.8931E+05, 1.8977E+05, 1.9024E+05,
     + 1.9070E+05, 1.9117E+05, 1.9164E+05, 1.9211E+05, 1.9258E+05,
     +  1.9305E+05, 1.9352E+05, 1.9399E+05, 1.9447E+05, 1.9494E+05,
     + 1.9542E+05, 1.9589E+05, 1.9637E+05, 1.9685E+05, 1.9733E+05,
     +  1.9781E+05, 1.9829E+05, 1.9877E+05, 1.9925E+05, 1.9973E+05,
     + 2.0022E+05, 2.0070E+05, 2.0119E+05, 2.0168E+05, 2.0217E+05,
     +  2.0266E+05, 2.0315E+05, 2.0364E+05, 2.0413E+05, 2.0462E+05,
     + 2.0512E+05, 2.0561E+05, 2.0611E+05, 2.0660E+05, 2.0710E+05,
     +  2.0760E+05, 2.0810E+05, 2.0860E+05, 2.0910E+05, 2.0960E+05,
     + 2.1011E+05, 2.1061E+05, 2.1112E+05, 2.1162E+05, 2.1213E+05,
     +  2.1264E+05, 2.1315E+05, 2.1366E+05, 2.1417E+05, 2.1468E+05,
     + 2.1519E+05, 2.1571E+05, 2.1622E+05, 2.1674E+05, 2.1725E+05,
     +  2.1777E+05, 2.1829E+05, 2.1881E+05, 2.1933E+05, 2.1985E+05,
     + 2.2037E+05, 2.2090E+05, 2.2142E+05, 2.2195E+05, 2.2247E+05,
     +  2.2300E+05, 2.2353E+05, 2.2406E+05, 2.2459E+05, 2.2512E+05,
     + 2.2565E+05, 2.2619E+05, 2.2672E+05, 2.2726E+05, 2.2779E+05,
     +  2.2833E+05, 2.2887E+05, 2.2941E+05, 2.2995E+05, 2.3049E+05,
     + 2.3103E+05, 2.3158E+05, 2.3212E+05, 2.3267E+05, 2.3321E+05,
     +  2.3376E+05, 2.3431E+05, 2.3486E+05, 2.3541E+05, 2.3596E+05,
     + 2.3651E+05, 2.3707E+05, 2.3762E+05, 2.3818E+05, 2.3873E+05,
     +  2.3929E+05, 2.3985E+05, 2.4041E+05, 2.4097E+05, 2.4153E+05,
     + 2.4209E+05, 2.4266E+05, 2.4322E+05, 2.4379E+05, 2.4436E+05,
     +  2.4492E+05, 2.4549E+05, 2.4606E+05, 2.4663E+05, 2.4721E+05,
     + 2.4778E+05, 2.4835E+05, 2.4893E+05, 2.4951E+05, 2.5008E+05,
     +  2.5066E+05, 2.5124E+05, 2.5182E+05, 2.5240E+05, 2.5299E+05,
     + 2.5357E+05, 2.5415E+05, 2.5474E+05, 2.5533E+05, 2.5591E+05,
     +  2.5650E+05, 2.5709E+05, 2.5768E+05, 2.5828E+05, 2.5887E+05,
     + 2.5946E+05, 2.6006E+05, 2.6066E+05, 2.6125E+05, 2.6185E+05,
     +  2.6245E+05, 2.6305E+05, 2.6366E+05, 2.6426E+05, 2.6486E+05,
     + 2.6547E+05, 2.6607E+05, 2.6668E+05, 2.6729E+05, 2.6790E+05,
     +  2.6851E+05, 2.6912E+05, 2.6973E+05, 2.7035E+05, 2.7096E+05,
     + 2.7158E+05, 2.7220E+05, 2.7281E+05, 2.7343E+05, 2.7405E+05,
     +  2.7468E+05, 2.7530E+05, 2.7592E+05, 2.7655E+05, 2.7717E+05,
     + 2.7780E+05, 2.7843E+05, 2.7906E+05, 2.7969E+05, 2.8032E+05,
     +  2.8095E+05, 2.8159E+05, 2.8222E+05, 2.8286E+05, 2.8350E+05,
     + 2.8413E+05, 2.8477E+05, 2.8541E+05, 2.8606E+05, 2.8670E+05,
     +  2.8734E+05, 2.8799E+05, 2.8863E+05, 2.8928E+05, 2.8993E+05,
     + 2.9058E+05, 2.9123E+05, 2.9188E+05, 2.9254E+05, 2.9319E+05,
     +  2.9385E+05, 2.9450E+05, 2.9516E+05, 2.9582E+05, 2.9648E+05,
     + 2.9714E+05, 2.9780E+05, 2.9847E+05, 2.9913E+05, 2.9980E+05,
     +  3.0046E+05, 3.0113E+05, 3.0180E+05, 3.0247E+05, 3.0314E+05,
     + 3.0382E+05, 3.0449E+05, 3.0517E+05, 3.0584E+05, 3.0652E+05,
     +  3.0720E+05, 3.0788E+05, 3.0856E+05, 3.0924E+05, 3.0993E+05,
     + 3.1061E+05, 3.1130E+05, 3.1198E+05, 3.1267E+05, 3.1336E+05,
     +  3.1405E+05, 3.1474E+05, 3.1544E+05, 3.1613E+05, 3.1683E+05,
     + 3.1752E+05, 3.1822E+05, 3.1892E+05, 3.1962E+05, 3.2032E+05,
     +  3.2103E+05, 3.2173E+05, 3.2243E+05, 3.2314E+05, 3.2385E+05,
     + 3.2456E+05, 3.2527E+05, 3.2598E+05, 3.2669E+05, 3.2740E+05,
     +  3.2812E+05, 3.2883E+05, 3.2955E+05, 3.3027E+05, 3.3099E+05,
     + 3.3171E+05, 3.3243E+05, 3.3316E+05, 3.3388E+05, 3.3461E+05,
     +  3.3533E+05, 3.3606E+05, 3.3679E+05, 3.3752E+05, 3.3826E+05,
     + 3.3899E+05, 3.3972E+05, 3.4046E+05, 3.4120E+05, 3.4193E+05,
     +  3.4267E+05, 3.4342E+05, 3.4416E+05, 3.4490E+05, 3.4565E+05,
     + 3.4639E+05, 3.4714E+05, 3.4789E+05, 3.4864E+05, 3.4939E+05,
     +  3.5014E+05, 3.5089E+05, 3.5165E+05, 3.5240E+05, 3.5316E+05,
     + 3.5392E+05, 3.5468E+05, 3.5544E+05, 3.5620E+05, 3.5697E+05,
     +  3.5773E+05, 3.5850E+05, 3.5926E+05, 3.6003E+05, 3.6080E+05,
     + 3.6158E+05, 3.6235E+05, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       134
      data (QofT( 2,J),J=1,2408)/ 1.2569E+01, 1.6645E+01, 2.2003E+01,
     +  2.7608E+01, 3.3294E+01, 3.9016E+01, 4.4757E+01, 5.0511E+01,
     + 5.6272E+01, 6.2038E+01, 6.7809E+01, 7.3582E+01, 7.9357E+01,
     +  8.5134E+01, 9.0913E+01, 9.6692E+01, 1.0247E+02, 1.0825E+02,
     + 1.1404E+02, 1.1982E+02, 1.3138E+02, 1.4295E+02, 1.5452E+02,
     +  1.6609E+02, 1.7766E+02, 1.8923E+02, 2.0080E+02, 2.1237E+02,
     + 2.2394E+02, 2.3552E+02, 2.4709E+02, 2.5866E+02, 2.7024E+02,
     +  2.8181E+02, 2.9338E+02, 3.0496E+02, 3.1653E+02, 3.2810E+02,
     + 3.3968E+02, 3.5125E+02, 3.6283E+02, 3.7440E+02, 3.8598E+02,
     +  3.9755E+02, 4.0913E+02, 4.2070E+02, 4.3228E+02, 4.4385E+02,
     + 4.5543E+02, 4.6701E+02, 4.7858E+02, 4.9016E+02, 5.0174E+02,
     +  5.1332E+02, 5.2490E+02, 5.3648E+02, 5.4806E+02, 5.5964E+02,
     + 5.7122E+02, 5.8281E+02, 5.9440E+02, 6.0599E+02, 6.1758E+02,
     +  6.2917E+02, 6.4077E+02, 6.5237E+02, 6.6397E+02, 6.7558E+02,
     + 6.8720E+02, 6.9882E+02, 7.1044E+02, 7.2207E+02, 7.3371E+02,
     +  7.4535E+02, 7.5700E+02, 7.6866E+02, 7.8033E+02, 7.9201E+02,
     + 8.0370E+02, 8.1541E+02, 8.2712E+02, 8.3885E+02, 8.5059E+02,
     +  8.6234E+02, 8.7411E+02, 8.8589E+02, 8.9769E+02, 9.0951E+02,
     + 9.2135E+02, 9.3320E+02, 9.4508E+02, 9.5698E+02, 9.6890E+02,
     +  9.8084E+02, 9.9280E+02, 1.0048E+03, 1.0168E+03, 1.0288E+03,
     + 1.0409E+03, 1.0530E+03, 1.0651E+03, 1.0773E+03, 1.0895E+03,
     +  1.1017E+03, 1.1139E+03, 1.1262E+03, 1.1385E+03, 1.1509E+03,
     + 1.1633E+03, 1.1757E+03, 1.1882E+03, 1.2007E+03, 1.2132E+03,
     +  1.2258E+03, 1.2384E+03, 1.2511E+03, 1.2638E+03, 1.2765E+03,
     + 1.2893E+03, 1.3022E+03, 1.3150E+03, 1.3280E+03, 1.3410E+03,
     +  1.3540E+03, 1.3671E+03, 1.3802E+03, 1.3934E+03, 1.4066E+03,
     + 1.4199E+03, 1.4332E+03, 1.4466E+03, 1.4601E+03, 1.4736E+03,
     +  1.4871E+03, 1.5007E+03, 1.5144E+03, 1.5281E+03, 1.5419E+03,
     + 1.5557E+03, 1.5697E+03, 1.5836E+03, 1.5977E+03, 1.6117E+03,
     +  1.6259E+03, 1.6401E+03, 1.6544E+03, 1.6687E+03, 1.6831E+03,
     + 1.6976E+03, 1.7122E+03, 1.7268E+03, 1.7415E+03, 1.7562E+03,
     +  1.7710E+03, 1.7859E+03, 1.8009E+03, 1.8159E+03, 1.8310E+03,
     + 1.8461E+03, 1.8614E+03, 1.8767E+03, 1.8921E+03, 1.9075E+03,
     +  1.9231E+03, 1.9387E+03, 1.9544E+03, 1.9701E+03, 1.9860E+03,
     + 2.0019E+03, 2.0179E+03, 2.0339E+03, 2.0501E+03, 2.0663E+03,
     +  2.0826E+03, 2.0990E+03, 2.1155E+03, 2.1320E+03, 2.1487E+03,
     + 2.1654E+03, 2.1822E+03, 2.1991E+03, 2.2160E+03, 2.2331E+03,
     +  2.2502E+03, 2.2674E+03, 2.2847E+03, 2.3021E+03, 2.3196E+03,
     + 2.3371E+03, 2.3548E+03, 2.3725E+03, 2.3904E+03, 2.4083E+03,
     +  2.4263E+03, 2.4444E+03, 2.4625E+03, 2.4808E+03, 2.4992E+03,
     + 2.5176E+03, 2.5362E+03, 2.5548E+03, 2.5735E+03, 2.5924E+03,
     +  2.6113E+03, 2.6303E+03, 2.6494E+03, 2.6686E+03, 2.6879E+03,
     + 2.7073E+03, 2.7268E+03, 2.7464E+03, 2.7660E+03, 2.7858E+03,
     +  2.8057E+03, 2.8257E+03, 2.8458E+03, 2.8659E+03, 2.8862E+03,
     + 2.9066E+03, 2.9270E+03, 2.9476E+03, 2.9683E+03, 2.9891E+03,
     +  3.0099E+03, 3.0309E+03, 3.0520E+03, 3.0732E+03, 3.0945E+03,
     + 3.1159E+03, 3.1374E+03, 3.1590E+03, 3.1807E+03, 3.2025E+03,
     +  3.2244E+03, 3.2465E+03, 3.2686E+03, 3.2909E+03, 3.3132E+03,
     + 3.3357E+03, 3.3582E+03, 3.3809E+03, 3.4037E+03, 3.4266E+03,
     +  3.4496E+03, 3.4728E+03, 3.4960E+03, 3.5193E+03, 3.5428E+03,
     + 3.5664E+03, 3.5901E+03, 3.6139E+03, 3.6378E+03, 3.6618E+03,
     +  3.6859E+03, 3.7102E+03, 3.7346E+03, 3.7591E+03, 3.7837E+03,
     + 3.8084E+03, 3.8332E+03, 3.8582E+03, 3.8833E+03, 3.9085E+03,
     +  3.9338E+03, 3.9592E+03, 3.9848E+03, 4.0105E+03, 4.0362E+03,
     + 4.0622E+03, 4.0882E+03, 4.1144E+03, 4.1407E+03, 4.1671E+03,
     +  4.1936E+03, 4.2203E+03, 4.2471E+03, 4.2740E+03, 4.3010E+03,
     + 4.3282E+03, 4.3555E+03, 4.3829E+03, 4.4104E+03, 4.4381E+03,
     +  4.4659E+03, 4.4938E+03, 4.5219E+03, 4.5500E+03, 4.5784E+03,
     + 4.6068E+03, 4.6354E+03, 4.6641E+03, 4.6929E+03, 4.7219E+03,
     +  4.7510E+03, 4.7803E+03, 4.8096E+03, 4.8392E+03, 4.8688E+03,
     + 4.8986E+03, 4.9285E+03, 4.9586E+03, 4.9887E+03, 5.0191E+03,
     +  5.0495E+03, 5.0801E+03, 5.1109E+03, 5.1418E+03, 5.1728E+03,
     + 5.2039E+03, 5.2352E+03, 5.2667E+03, 5.2983E+03, 5.3300E+03,
     +  5.3619E+03, 5.3939E+03, 5.4260E+03, 5.4583E+03, 5.4908E+03,
     + 5.5234E+03, 5.5561E+03, 5.5890E+03, 5.6220E+03, 5.6552E+03,
     +  5.6885E+03, 5.7220E+03, 5.7556E+03, 5.7894E+03, 5.8233E+03,
     + 5.8574E+03, 5.8916E+03, 5.9260E+03, 5.9605E+03, 5.9952E+03,
     +  6.0300E+03, 6.0650E+03, 6.1001E+03, 6.1354E+03, 6.1708E+03,
     + 6.2064E+03, 6.2422E+03, 6.2781E+03, 6.3142E+03, 6.3504E+03,
     +  6.3868E+03, 6.4233E+03, 6.4600E+03, 6.4969E+03, 6.5339E+03,
     + 6.5711E+03, 6.6084E+03, 6.6459E+03, 6.6836E+03, 6.7214E+03,
     +  6.7594E+03, 6.7976E+03, 6.8359E+03, 6.8744E+03, 6.9130E+03,
     + 6.9518E+03, 6.9908E+03, 7.0300E+03, 7.0693E+03, 7.1088E+03,
     +  7.1484E+03, 7.1882E+03, 7.2282E+03, 7.2684E+03, 7.3087E+03,
     + 7.3492E+03, 7.3899E+03, 7.4308E+03, 7.4718E+03, 7.5130E+03,
     +  7.5544E+03, 7.5959E+03, 7.6376E+03, 7.6795E+03, 7.7216E+03,
     + 7.7638E+03, 7.8063E+03, 7.8489E+03, 7.8917E+03, 7.9346E+03,
     +  7.9778E+03, 8.0211E+03, 8.0646E+03, 8.1083E+03, 8.1522E+03,
     + 8.1962E+03, 8.2405E+03, 8.2849E+03, 8.3295E+03, 8.3743E+03,
     +  8.4193E+03, 8.4644E+03, 8.5098E+03, 8.5553E+03, 8.6011E+03,
     + 8.6470E+03, 8.6931E+03, 8.7394E+03, 8.7859E+03, 8.8326E+03,
     +  8.8794E+03, 8.9265E+03, 8.9738E+03, 9.0212E+03, 9.0688E+03,
     + 9.1167E+03, 9.1647E+03, 9.2130E+03, 9.2614E+03, 9.3100E+03,
     +  9.3588E+03, 9.4079E+03, 9.4571E+03, 9.5065E+03, 9.5561E+03,
     + 9.6059E+03, 9.6560E+03, 9.7062E+03, 9.7566E+03, 9.8073E+03,
     +  9.8581E+03, 9.9091E+03, 9.9604E+03, 1.0012E+04, 1.0064E+04,
     + 1.0115E+04, 1.0167E+04, 1.0220E+04, 1.0272E+04, 1.0325E+04,
     +  1.0378E+04, 1.0431E+04, 1.0484E+04, 1.0538E+04, 1.0592E+04,
     + 1.0646E+04, 1.0700E+04, 1.0754E+04, 1.0809E+04, 1.0864E+04,
     +  1.0919E+04, 1.0974E+04, 1.1030E+04, 1.1085E+04, 1.1141E+04,
     + 1.1197E+04, 1.1254E+04, 1.1310E+04, 1.1367E+04, 1.1424E+04,
     +  1.1481E+04, 1.1539E+04, 1.1597E+04, 1.1655E+04, 1.1713E+04,
     + 1.1771E+04, 1.1830E+04, 1.1889E+04, 1.1948E+04, 1.2007E+04,
     +  1.2067E+04, 1.2126E+04, 1.2186E+04, 1.2247E+04, 1.2307E+04,
     + 1.2368E+04, 1.2429E+04, 1.2490E+04, 1.2551E+04, 1.2613E+04,
     +  1.2675E+04, 1.2737E+04, 1.2799E+04, 1.2862E+04, 1.2925E+04,
     + 1.2988E+04, 1.3051E+04, 1.3115E+04, 1.3179E+04, 1.3243E+04,
     +  1.3307E+04, 1.3372E+04, 1.3436E+04, 1.3501E+04, 1.3567E+04,
     + 1.3632E+04, 1.3698E+04, 1.3764E+04, 1.3830E+04, 1.3897E+04,
     +  1.3964E+04, 1.4031E+04, 1.4098E+04, 1.4166E+04, 1.4233E+04,
     + 1.4301E+04, 1.4370E+04, 1.4438E+04, 1.4507E+04, 1.4576E+04,
     +  1.4646E+04, 1.4715E+04, 1.4785E+04, 1.4855E+04, 1.4926E+04,
     + 1.4996E+04, 1.5067E+04, 1.5144E+04, 1.5218E+04, 1.5292E+04,
     +  1.5366E+04, 1.5441E+04, 1.5516E+04, 1.5591E+04, 1.5667E+04,
     + 1.5743E+04, 1.5819E+04, 1.5895E+04, 1.5972E+04, 1.6049E+04,
     +  1.6126E+04, 1.6203E+04, 1.6281E+04, 1.6359E+04, 1.6438E+04,
     + 1.6516E+04, 1.6595E+04, 1.6675E+04, 1.6754E+04, 1.6834E+04,
     +  1.6914E+04, 1.6995E+04, 1.7075E+04, 1.7156E+04, 1.7238E+04,
     + 1.7320E+04, 1.7401E+04, 1.7484E+04, 1.7566E+04, 1.7649E+04,
     +  1.7732E+04, 1.7816E+04, 1.7900E+04, 1.7984E+04, 1.8068E+04,
     + 1.8153E+04, 1.8238E+04, 1.8323E+04, 1.8409E+04, 1.8495E+04,
     +  1.8581E+04, 1.8668E+04, 1.8755E+04, 1.8842E+04, 1.8930E+04,
     + 1.9018E+04, 1.9106E+04, 1.9194E+04, 1.9283E+04, 1.9372E+04,
     +  1.9462E+04, 1.9552E+04, 1.9642E+04, 1.9732E+04, 1.9823E+04,
     + 1.9914E+04, 2.0006E+04, 2.0097E+04, 2.0189E+04, 2.0282E+04,
     +  2.0375E+04, 2.0468E+04, 2.0561E+04, 2.0655E+04, 2.0749E+04,
     + 2.0844E+04, 2.0938E+04, 2.1033E+04, 2.1129E+04, 2.1225E+04,
     +  2.1321E+04, 2.1417E+04, 2.1514E+04, 2.1611E+04, 2.1709E+04,
     + 2.1807E+04, 2.1905E+04, 2.2003E+04, 2.2102E+04, 2.2202E+04,
     +  2.2301E+04, 2.2401E+04, 2.2501E+04, 2.2602E+04, 2.2703E+04,
     + 2.2804E+04, 2.2906E+04, 2.3008E+04, 2.3111E+04, 2.3213E+04,
     +  2.3316E+04, 2.3420E+04, 2.3524E+04, 2.3628E+04, 2.3733E+04,
     + 2.3838E+04, 2.3943E+04, 2.4049E+04, 2.4155E+04, 2.4261E+04,
     +  2.4368E+04, 2.4475E+04, 2.4583E+04, 2.4690E+04, 2.4799E+04,
     + 2.4907E+04, 2.5016E+04, 2.5126E+04, 2.5236E+04, 2.5346E+04,
     +  2.5456E+04, 2.5567E+04, 2.5678E+04, 2.5790E+04, 2.5902E+04,
     + 2.6015E+04, 2.6127E+04, 2.6241E+04, 2.6354E+04, 2.6468E+04,
     +  2.6583E+04, 2.6697E+04, 2.6812E+04, 2.6928E+04, 2.7044E+04,
     + 2.7160E+04, 2.7277E+04, 2.7394E+04, 2.7512E+04, 2.7629E+04,
     +  2.7748E+04, 2.7866E+04, 2.7986E+04, 2.8105E+04, 2.8225E+04,
     + 2.8345E+04, 2.8466E+04, 2.8587E+04, 2.8709E+04, 2.8831E+04,
     +  2.8953E+04, 2.9076E+04, 2.9199E+04, 2.9322E+04, 2.9446E+04,
     + 2.9571E+04, 2.9696E+04, 2.9821E+04, 2.9946E+04, 3.0072E+04,
     +  3.0199E+04, 3.0326E+04, 3.0453E+04, 3.0581E+04, 3.0709E+04,
     + 3.0838E+04, 3.0967E+04, 3.1096E+04, 3.1226E+04, 3.1356E+04,
     +  3.1487E+04, 3.1618E+04, 3.1750E+04, 3.1882E+04, 3.2014E+04,
     + 3.2147E+04, 3.2280E+04, 3.2414E+04, 3.2548E+04, 3.2683E+04,
     +  3.2818E+04, 3.2953E+04, 3.3089E+04, 3.3226E+04, 3.3363E+04,
     + 3.3500E+04, 3.3638E+04, 3.3776E+04, 3.3914E+04, 3.4053E+04,
     +  3.4193E+04, 3.4333E+04, 3.4473E+04, 3.4614E+04, 3.4756E+04,
     + 3.4897E+04, 3.5040E+04, 3.5182E+04, 3.5326E+04, 3.5469E+04,
     +  3.5613E+04, 3.5758E+04, 3.5903E+04, 3.6048E+04, 3.6194E+04,
     + 3.6341E+04, 3.6488E+04, 3.6635E+04, 3.6783E+04, 3.6931E+04,
     +  3.7080E+04, 3.7229E+04, 3.7379E+04, 3.7529E+04, 3.7680E+04,
     + 3.7831E+04, 3.7982E+04, 3.8134E+04, 3.8287E+04, 3.8440E+04,
     +  3.8594E+04, 3.8748E+04, 3.8902E+04, 3.9057E+04, 3.9213E+04,
     + 3.9369E+04, 3.9525E+04, 3.9682E+04, 3.9840E+04, 3.9998E+04,
     +  4.0156E+04, 4.0315E+04, 4.0475E+04, 4.0635E+04, 4.0795E+04,
     + 4.0956E+04, 4.1117E+04, 4.1279E+04, 4.1442E+04, 4.1605E+04,
     +  4.1768E+04, 4.1932E+04, 4.2097E+04, 4.2262E+04, 4.2427E+04,
     + 4.2593E+04, 4.2760E+04, 4.2927E+04, 4.3095E+04, 4.3263E+04,
     +  4.3431E+04, 4.3600E+04, 4.3770E+04, 4.3940E+04, 4.4111E+04,
     + 4.4282E+04, 4.4454E+04, 4.4626E+04, 4.4799E+04, 4.4972E+04,
     +  4.5146E+04, 4.5321E+04, 4.5496E+04, 4.5671E+04, 4.5847E+04,
     + 4.6024E+04, 4.6201E+04, 4.6378E+04, 4.6557E+04, 4.6735E+04,
     +  4.6915E+04, 4.7094E+04, 4.7275E+04, 4.7456E+04, 4.7637E+04,
     + 4.7819E+04, 4.8002E+04, 4.8185E+04, 4.8368E+04, 4.8553E+04,
     +  4.8737E+04, 4.8923E+04, 4.9109E+04, 4.9295E+04, 4.9482E+04,
     + 4.9670E+04, 4.9858E+04, 5.0047E+04, 5.0236E+04, 5.0426E+04,
     +  5.0616E+04, 5.0807E+04, 5.0999E+04, 5.1191E+04, 5.1383E+04,
     + 5.1577E+04, 5.1770E+04, 5.1965E+04, 5.2160E+04, 5.2356E+04,
     +  5.2552E+04, 5.2748E+04, 5.2946E+04, 5.3144E+04, 5.3342E+04,
     + 5.3541E+04, 5.3741E+04, 5.3941E+04, 5.4142E+04, 5.4344E+04,
     +  5.4546E+04, 5.4748E+04, 5.4951E+04, 5.5155E+04, 5.5360E+04,
     + 5.5565E+04, 5.5770E+04, 5.5977E+04, 5.6184E+04, 5.6391E+04,
     +  5.6599E+04, 5.6808E+04, 5.7017E+04, 5.7227E+04, 5.7437E+04,
     + 5.7649E+04, 5.7860E+04, 5.8073E+04, 5.8286E+04, 5.8499E+04,
     +  5.8714E+04, 5.8928E+04, 5.9144E+04, 5.9360E+04, 5.9577E+04,
     + 5.9794E+04, 6.0012E+04, 6.0231E+04, 6.0450E+04, 6.0670E+04,
     +  6.0890E+04, 6.1111E+04, 6.1333E+04, 6.1556E+04, 6.1779E+04,
     + 6.2002E+04, 6.2227E+04, 6.2452E+04, 6.2677E+04, 6.2904E+04,
     +  6.3131E+04, 6.3358E+04, 6.3587E+04, 6.3815E+04, 6.4045E+04,
     + 6.4275E+04, 6.4506E+04, 6.4738E+04, 6.4970E+04, 6.5203E+04,
     +  6.5436E+04, 6.5670E+04, 6.5905E+04, 6.6141E+04, 6.6377E+04,
     + 6.6614E+04, 6.6851E+04, 6.7089E+04, 6.7328E+04, 6.7568E+04,
     +  6.7808E+04, 6.8049E+04, 6.8291E+04, 6.8533E+04, 6.8776E+04,
     + 6.9019E+04, 6.9264E+04, 6.9509E+04, 6.9754E+04, 7.0001E+04,
     +  7.0248E+04, 7.0496E+04, 7.0744E+04, 7.0993E+04, 7.1243E+04,
     + 7.1494E+04, 7.1745E+04, 7.1997E+04, 7.2249E+04, 7.2503E+04,
     +  7.2757E+04, 7.3012E+04, 7.3267E+04, 7.3523E+04, 7.3780E+04,
     + 7.4038E+04, 7.4296E+04, 7.4555E+04, 7.4815E+04, 7.5075E+04,
     +  7.5337E+04, 7.5599E+04, 7.5861E+04, 7.6125E+04, 7.6389E+04,
     + 7.6653E+04, 7.6919E+04, 7.7185E+04, 7.7452E+04, 7.7720E+04,
     +  7.7989E+04, 7.8258E+04, 7.8528E+04, 7.8799E+04, 7.9070E+04,
     + 7.9342E+04, 7.9615E+04, 7.9889E+04, 8.0163E+04, 8.0438E+04,
     +  8.0714E+04, 8.0991E+04, 8.1269E+04, 8.1547E+04, 8.1826E+04,
     + 8.2105E+04, 8.2386E+04, 8.2667E+04, 8.2949E+04, 8.3232E+04,
     +  8.3515E+04, 8.3800E+04, 8.4085E+04, 8.4371E+04, 8.4657E+04,
     + 8.4945E+04, 8.5233E+04, 8.5522E+04, 8.5811E+04, 8.6102E+04,
     +  8.6393E+04, 8.6685E+04, 8.6978E+04, 8.7272E+04, 8.7566E+04,
     + 8.7861E+04, 8.8157E+04, 8.8454E+04, 8.8752E+04, 8.9050E+04,
     +  8.9349E+04, 8.9649E+04, 8.9950E+04, 9.0252E+04, 9.0554E+04,
     + 9.0857E+04, 9.1161E+04, 9.1466E+04, 9.1772E+04, 9.2078E+04,
     +  9.2386E+04, 9.2694E+04, 9.3003E+04, 9.3312E+04, 9.3623E+04,
     + 9.3934E+04, 9.4246E+04, 9.4559E+04, 9.4873E+04, 9.5188E+04,
     +  9.5503E+04, 9.5820E+04, 9.6137E+04, 9.6455E+04, 9.6774E+04,
     + 9.7093E+04, 9.7414E+04, 9.7735E+04, 9.8057E+04, 9.8380E+04,
     +  9.8704E+04, 9.9029E+04, 9.9354E+04, 9.9681E+04, 1.0001E+05,
     + 1.0034E+05, 1.0067E+05, 1.0100E+05, 1.0133E+05, 1.0166E+05,
     +  1.0199E+05, 1.0232E+05, 1.0266E+05, 1.0299E+05, 1.0333E+05,
     + 1.0367E+05, 1.0400E+05, 1.0434E+05, 1.0468E+05, 1.0502E+05,
     +  1.0536E+05, 1.0570E+05, 1.0605E+05, 1.0639E+05, 1.0674E+05,
     + 1.0708E+05, 1.0743E+05, 1.0777E+05, 1.0812E+05, 1.0847E+05,
     +  1.0882E+05, 1.0917E+05, 1.0953E+05, 1.0988E+05, 1.1023E+05,
     + 1.1059E+05, 1.1094E+05, 1.1130E+05, 1.1166E+05, 1.1201E+05,
     +  1.1237E+05, 1.1273E+05, 1.1309E+05, 1.1346E+05, 1.1382E+05,
     + 1.1418E+05, 1.1455E+05, 1.1491E+05, 1.1528E+05, 1.1565E+05,
     +  1.1601E+05, 1.1638E+05, 1.1675E+05, 1.1712E+05, 1.1750E+05,
     + 1.1787E+05, 1.1824E+05, 1.1862E+05, 1.1899E+05, 1.1937E+05,
     +  1.1975E+05, 1.2013E+05, 1.2051E+05, 1.2089E+05, 1.2127E+05,
     + 1.2165E+05, 1.2204E+05, 1.2242E+05, 1.2281E+05, 1.2319E+05,
     +  1.2358E+05, 1.2397E+05, 1.2436E+05, 1.2475E+05, 1.2514E+05,
     + 1.2553E+05, 1.2592E+05, 1.2632E+05, 1.2671E+05, 1.2711E+05,
     +  1.2751E+05, 1.2790E+05, 1.2830E+05, 1.2870E+05, 1.2911E+05,
     + 1.2951E+05, 1.2991E+05, 1.3031E+05, 1.3072E+05, 1.3113E+05,
     +  1.3153E+05, 1.3194E+05, 1.3235E+05, 1.3276E+05, 1.3317E+05,
     + 1.3358E+05, 1.3400E+05, 1.3441E+05, 1.3483E+05, 1.3524E+05,
     +  1.3566E+05, 1.3608E+05, 1.3650E+05, 1.3692E+05, 1.3734E+05,
     + 1.3776E+05, 1.3818E+05, 1.3861E+05, 1.3903E+05, 1.3946E+05,
     +  1.3989E+05, 1.4032E+05, 1.4075E+05, 1.4118E+05, 1.4161E+05,
     + 1.4204E+05, 1.4247E+05, 1.4291E+05, 1.4335E+05, 1.4378E+05,
     +  1.4422E+05, 1.4466E+05, 1.4510E+05, 1.4554E+05, 1.4598E+05,
     + 1.4643E+05, 1.4687E+05, 1.4732E+05, 1.4776E+05, 1.4821E+05,
     +  1.4866E+05, 1.4911E+05, 1.4956E+05, 1.5001E+05, 1.5047E+05,
     + 1.5092E+05, 1.5137E+05, 1.5183E+05, 1.5229E+05, 1.5275E+05,
     +  1.5321E+05, 1.5367E+05, 1.5413E+05, 1.5459E+05, 1.5506E+05,
     + 1.5552E+05, 1.5599E+05, 1.5645E+05, 1.5692E+05, 1.5739E+05,
     +  1.5786E+05, 1.5834E+05, 1.5881E+05, 1.5928E+05, 1.5976E+05,
     + 1.6023E+05, 1.6071E+05, 1.6119E+05, 1.6167E+05, 1.6215E+05,
     +  1.6263E+05, 1.6312E+05, 1.6360E+05, 1.6409E+05, 1.6457E+05,
     + 1.6506E+05, 1.6555E+05, 1.6604E+05, 1.6653E+05, 1.6702E+05,
     +  1.6752E+05, 1.6801E+05, 1.6851E+05, 1.6900E+05, 1.6950E+05,
     + 1.7000E+05, 1.7050E+05, 1.7100E+05, 1.7151E+05, 1.7201E+05,
     +  1.7251E+05, 1.7302E+05, 1.7353E+05, 1.7404E+05, 1.7455E+05,
     + 1.7506E+05, 1.7557E+05, 1.7608E+05, 1.7660E+05, 1.7711E+05,
     +  1.7763E+05, 1.7815E+05, 1.7867E+05, 1.7919E+05, 1.7971E+05,
     + 1.8024E+05, 1.8076E+05, 1.8128E+05, 1.8181E+05, 1.8234E+05,
     +  1.8287E+05, 1.8340E+05, 1.8393E+05, 1.8446E+05, 1.8500E+05,
     + 1.8553E+05, 1.8607E+05, 1.8661E+05, 1.8715E+05, 1.8769E+05,
     +  1.8823E+05, 1.8877E+05, 1.8931E+05, 1.8986E+05, 1.9041E+05,
     + 1.9095E+05, 1.9150E+05, 1.9205E+05, 1.9260E+05, 1.9316E+05,
     +  1.9371E+05, 1.9427E+05, 1.9482E+05, 1.9538E+05, 1.9594E+05,
     + 1.9650E+05, 1.9706E+05, 1.9762E+05, 1.9819E+05, 1.9875E+05,
     +  1.9932E+05, 1.9989E+05, 2.0046E+05, 2.0103E+05, 2.0160E+05,
     + 2.0217E+05, 2.0275E+05, 2.0332E+05, 2.0390E+05, 2.0448E+05,
     +  2.0506E+05, 2.0564E+05, 2.0622E+05, 2.0680E+05, 2.0739E+05,
     + 2.0798E+05, 2.0856E+05, 2.0915E+05, 2.0974E+05, 2.1033E+05,
     +  2.1093E+05, 2.1152E+05, 2.1212E+05, 2.1271E+05, 2.1331E+05,
     + 2.1391E+05, 2.1451E+05, 2.1511E+05, 2.1572E+05, 2.1632E+05,
     +  2.1693E+05, 2.1753E+05, 2.1814E+05, 2.1875E+05, 2.1936E+05,
     + 2.1998E+05, 2.2059E+05, 2.2121E+05, 2.2182E+05, 2.2244E+05,
     +  2.2306E+05, 2.2368E+05, 2.2431E+05, 2.2493E+05, 2.2556E+05,
     + 2.2618E+05, 2.2681E+05, 2.2744E+05, 2.2807E+05, 2.2870E+05,
     +  2.2934E+05, 2.2997E+05, 2.3061E+05, 2.3124E+05, 2.3188E+05,
     + 2.3252E+05, 2.3317E+05, 2.3381E+05, 2.3445E+05, 2.3510E+05,
     +  2.3575E+05, 2.3640E+05, 2.3705E+05, 2.3770E+05, 2.3835E+05,
     + 2.3901E+05, 2.3966E+05, 2.4032E+05, 2.4098E+05, 2.4164E+05,
     +  2.4230E+05, 2.4297E+05, 2.4363E+05, 2.4430E+05, 2.4496E+05,
     + 2.4563E+05, 2.4630E+05, 2.4698E+05, 2.4765E+05, 2.4832E+05,
     +  2.4900E+05, 2.4968E+05, 2.5036E+05, 2.5104E+05, 2.5172E+05,
     + 2.5240E+05, 2.5309E+05, 2.5378E+05, 2.5446E+05, 2.5515E+05,
     +  2.5584E+05, 2.5654E+05, 2.5723E+05, 2.5793E+05, 2.5862E+05,
     + 2.5932E+05, 2.6002E+05, 2.6072E+05, 2.6143E+05, 2.6213E+05,
     +  2.6284E+05, 2.6354E+05, 2.6425E+05, 2.6496E+05, 2.6568E+05,
     + 2.6639E+05, 2.6710E+05, 2.6782E+05, 2.6854E+05, 2.6926E+05,
     +  2.6998E+05, 2.7070E+05, 2.7143E+05, 2.7215E+05, 2.7288E+05,
     + 2.7361E+05, 2.7434E+05, 2.7507E+05, 2.7580E+05, 2.7654E+05,
     +  2.7728E+05, 2.7801E+05, 2.7875E+05, 2.7950E+05, 2.8024E+05,
     + 2.8098E+05, 2.8173E+05, 2.8248E+05, 2.8323E+05, 2.8398E+05,
     +  2.8473E+05, 2.8548E+05, 2.8624E+05, 2.8700E+05, 2.8775E+05,
     + 2.8851E+05, 2.8928E+05, 2.9004E+05, 2.9080E+05, 2.9157E+05,
     +  2.9234E+05, 2.9311E+05, 2.9388E+05, 2.9465E+05, 2.9543E+05,
     + 2.9620E+05, 2.9698E+05, 2.9776E+05, 2.9854E+05, 2.9933E+05,
     +  3.0011E+05, 3.0090E+05, 3.0168E+05, 3.0247E+05, 3.0326E+05,
     + 3.0406E+05, 3.0485E+05, 3.0565E+05, 3.0644E+05, 3.0724E+05,
     +  3.0804E+05, 3.0885E+05, 3.0965E+05, 3.1046E+05, 3.1126E+05,
     + 3.1207E+05, 3.1288E+05, 3.1370E+05, 3.1451E+05, 3.1533E+05,
     +  3.1614E+05, 3.1696E+05, 3.1778E+05, 3.1860E+05, 3.1943E+05,
     + 3.2025E+05, 3.2108E+05, 3.2191E+05, 3.2274E+05, 3.2358E+05,
     +  3.2441E+05, 3.2525E+05, 3.2608E+05, 3.2692E+05, 3.2776E+05,
     + 3.2861E+05, 3.2945E+05, 3.3030E+05, 3.3115E+05, 3.3200E+05,
     +  3.3285E+05, 3.3370E+05, 3.3456E+05, 3.3541E+05, 3.3627E+05,
     + 3.3713E+05, 3.3799E+05, 3.3886E+05, 3.3972E+05, 3.4059E+05,
     +  3.4146E+05, 3.4233E+05, 3.4320E+05, 3.4407E+05, 3.4495E+05,
     + 3.4583E+05, 3.4671E+05, 3.4759E+05, 3.4847E+05, 3.4936E+05,
     +  3.5024E+05, 3.5113E+05, 3.5202E+05, 3.5291E+05, 3.5381E+05,
     + 3.5470E+05, 3.5560E+05, 3.5650E+05, 3.5740E+05, 3.5830E+05,
     +  3.5921E+05, 3.6011E+05, 3.6102E+05, 3.6193E+05, 3.6284E+05,
     + 3.6376E+05, 3.6467E+05, 3.6559E+05, 3.6651E+05, 3.6743E+05,
     +  3.6835E+05, 3.6928E+05, 3.7020E+05, 3.7113E+05, 3.7206E+05,
     + 3.7299E+05, 3.7393E+05, 3.7486E+05, 3.7580E+05, 3.7674E+05,
     +  3.7768E+05, 3.7862E+05, 3.7957E+05, 3.8051E+05, 3.8146E+05,
     + 3.8241E+05, 3.8337E+05, 3.8432E+05, 3.8528E+05, 3.8623E+05,
     +  3.8719E+05, 3.8816E+05, 3.8912E+05, 3.9008E+05, 3.9105E+05,
     + 3.9202E+05, 3.9299E+05, 3.9397E+05, 3.9494E+05, 3.9592E+05,
     +  3.9690E+05, 3.9788E+05, 3.9886E+05, 3.9984E+05, 4.0083E+05,
     + 4.0182E+05, 4.0281E+05, 4.0380E+05, 4.0480E+05, 4.0579E+05,
     +  4.0679E+05, 4.0779E+05, 4.0879E+05, 4.0980E+05, 4.1080E+05,
     + 4.1181E+05, 4.1282E+05, 4.1383E+05, 4.1485E+05, 4.1586E+05,
     +  4.1688E+05, 4.1790E+05, 4.1892E+05, 4.1994E+05, 4.2097E+05,
     + 4.2200E+05, 4.2303E+05, 4.2406E+05, 4.2509E+05, 4.2613E+05,
     +  4.2716E+05, 4.2820E+05, 4.2924E+05, 4.3029E+05, 4.3133E+05,
     + 4.3238E+05, 4.3343E+05, 4.3448E+05, 4.3553E+05, 4.3659E+05,
     +  4.3765E+05, 4.3871E+05, 4.3977E+05, 4.4083E+05, 4.4190E+05,
     + 4.4297E+05, 4.4404E+05, 4.4511E+05, 4.4618E+05, 4.4726E+05,
     +  4.4833E+05, 4.4941E+05, 4.5050E+05, 4.5158E+05, 4.5267E+05,
     + 4.5375E+05, 4.5484E+05, 4.5594E+05, 4.5703E+05, 4.5813E+05,
     +  4.5923E+05, 4.6033E+05, 4.6143E+05, 4.6253E+05, 4.6364E+05,
     + 4.6475E+05, 4.6586E+05, 4.6697E+05, 4.6809E+05, 4.6921E+05,
     +  4.7033E+05, 4.7145E+05, 4.7257E+05, 4.7370E+05, 4.7483E+05,
     + 4.7596E+05, 4.7709E+05, 4.7822E+05, 4.7936E+05, 4.8050E+05,
     +  4.8164E+05, 4.8278E+05, 4.8392E+05, 4.8507E+05, 4.8622E+05,
     + 4.8737E+05, 4.8853E+05, 4.8968E+05, 4.9084E+05, 4.9200E+05,
     +  4.9316E+05, 4.9433E+05, 4.9549E+05, 4.9666E+05, 4.9783E+05,
     + 4.9900E+05, 5.0018E+05, 5.0136E+05, 5.0254E+05, 5.0372E+05,
     +  5.0490E+05, 5.0609E+05, 5.0728E+05, 5.0847E+05, 5.0966E+05,
     + 5.1085E+05, 5.1205E+05, 5.1325E+05, 5.1445E+05, 5.1566E+05,
     +  5.1686E+05, 5.1807E+05, 5.1928E+05, 5.2049E+05, 5.2171E+05,
     + 5.2293E+05, 5.2414E+05, 5.2537E+05, 5.2659E+05, 5.2782E+05,
     +  5.2904E+05, 5.3027E+05, 5.3151E+05, 5.3274E+05, 5.3398E+05,
     + 5.3522E+05, 5.3646E+05, 5.3771E+05, 5.3895E+05, 5.4020E+05,
     +  5.4145E+05, 5.4271E+05, 5.4396E+05, 5.4522E+05, 5.4648E+05,
     + 5.4774E+05, 5.4901E+05, 5.5027E+05, 5.5154E+05, 5.5281E+05,
     +  5.5409E+05, 5.5536E+05, 5.5664E+05, 5.5792E+05, 5.5921E+05,
     + 5.6049E+05, 5.6178E+05, 5.6307E+05, 5.6436E+05, 5.6566E+05,
     +  5.6696E+05, 5.6825E+05, 5.6956E+05, 5.7086E+05, 5.7217E+05,
     + 5.7348E+05, 5.7479E+05, 5.7610E+05, 5.7742E+05, 5.7874E+05,
     +  5.8006E+05, 5.8138E+05, 5.8271E+05, 5.8403E+05, 5.8536E+05,
     + 5.8670E+05, 5.8803E+05, 5.8937E+05, 5.9071E+05, 5.9205E+05,
     +  5.9340E+05, 5.9474E+05, 5.9609E+05, 5.9744E+05, 5.9880E+05,
     + 6.0015E+05, 6.0151E+05, 6.0287E+05, 6.0424E+05, 6.0561E+05,
     +  6.0697E+05, 6.0835E+05, 6.0972E+05, 6.1109E+05, 6.1247E+05,
     + 6.1385E+05, 6.1524E+05, 6.1662E+05, 6.1801E+05, 6.1940E+05,
     +  6.2080E+05, 6.2219E+05, 6.2359E+05, 6.2499E+05, 6.2639E+05,
     + 6.2780E+05, 6.2921E+05, 6.3062E+05, 6.3203E+05, 6.3345E+05,
     +  6.3486E+05, 6.3629E+05, 6.3771E+05, 6.3913E+05, 6.4056E+05,
     + 6.4199E+05, 6.4343E+05, 6.4486E+05, 6.4630E+05, 6.4774E+05,
     +  6.4918E+05, 6.5063E+05, 6.5208E+05, 6.5353E+05, 6.5498E+05,
     + 6.5644E+05, 6.5790E+05, 6.5936E+05, 6.6082E+05, 6.6229E+05,
     +  6.6375E+05, 6.6523E+05, 6.6670E+05, 6.6818E+05, 6.6965E+05,
     + 6.7114E+05, 6.7262E+05, 6.7411E+05, 6.7559E+05, 6.7709E+05,
     +  6.7858E+05, 6.8008E+05, 6.8158E+05, 6.8308E+05, 6.8458E+05,
     + 6.8609E+05, 6.8760E+05, 6.8911E+05, 6.9063E+05, 6.9215E+05,
     +  6.9367E+05, 6.9519E+05, 6.9671E+05, 6.9824E+05, 6.9977E+05,
     + 7.0131E+05, 7.0284E+05, 7.0438E+05, 7.0592E+05, 7.0747E+05,
     +  7.0901E+05, 7.1056E+05, 7.1212E+05, 7.1367E+05, 7.1523E+05,
     + 7.1679E+05, 7.1835E+05, 7.1992E+05, 7.2148E+05, 7.2305E+05,
     +  7.2463E+05, 7.2620E+05, 7.2778E+05, 7.2936E+05, 7.3095E+05,
     + 7.3253E+05, 7.3412E+05, 7.3572E+05, 7.3731E+05, 7.3891E+05,
     +  7.4051E+05, 7.4211E+05, 7.4372E+05, 7.4533E+05, 7.4694E+05,
     + 7.4855E+05, 7.5017E+05, 7.5179E+05, 7.5341E+05, 7.5503E+05,
     +  7.5666E+05, 7.5829E+05, 7.5993E+05, 7.6156E+05, 7.6320E+05,
     + 7.6484E+05, 7.6649E+05, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       125
      data (QofT( 3,J),J=1,2408)/ 4.7055E+00, 5.5622E+00, 7.3581E+00,
     +  9.2353E+00, 1.1139E+01, 1.3055E+01, 1.4977E+01, 1.6904E+01,
     + 1.8833E+01, 2.0764E+01, 2.2696E+01, 2.4629E+01, 2.6563E+01,
     +  2.8497E+01, 3.0432E+01, 3.2367E+01, 3.4303E+01, 3.6239E+01,
     + 3.8175E+01, 4.0112E+01, 4.3985E+01, 4.7859E+01, 5.1734E+01,
     +  5.5609E+01, 5.9484E+01, 6.3360E+01, 6.7236E+01, 7.1113E+01,
     + 7.4989E+01, 7.8866E+01, 8.2743E+01, 8.6621E+01, 9.0499E+01,
     +  9.4377E+01, 9.8255E+01, 1.0213E+02, 1.0601E+02, 1.0989E+02,
     + 1.1377E+02, 1.1765E+02, 1.2153E+02, 1.2541E+02, 1.2929E+02,
     +  1.3317E+02, 1.3705E+02, 1.4093E+02, 1.4481E+02, 1.4869E+02,
     + 1.5257E+02, 1.5645E+02, 1.6033E+02, 1.6422E+02, 1.6810E+02,
     +  1.7198E+02, 1.7586E+02, 1.7975E+02, 1.8363E+02, 1.8752E+02,
     + 1.9140E+02, 1.9529E+02, 1.9917E+02, 2.0306E+02, 2.0695E+02,
     +  2.1084E+02, 2.1473E+02, 2.1862E+02, 2.2252E+02, 2.2641E+02,
     + 2.3031E+02, 2.3421E+02, 2.3811E+02, 2.4201E+02, 2.4592E+02,
     +  2.4982E+02, 2.5373E+02, 2.5765E+02, 2.6156E+02, 2.6548E+02,
     + 2.6941E+02, 2.7334E+02, 2.7727E+02, 2.8121E+02, 2.8515E+02,
     +  2.8909E+02, 2.9304E+02, 2.9700E+02, 3.0096E+02, 3.0493E+02,
     + 3.0891E+02, 3.1289E+02, 3.1688E+02, 3.2087E+02, 3.2487E+02,
     +  3.2888E+02, 3.3290E+02, 3.3693E+02, 3.4096E+02, 3.4501E+02,
     + 3.4906E+02, 3.5313E+02, 3.5720E+02, 3.6128E+02, 3.6538E+02,
     +  3.6948E+02, 3.7360E+02, 3.7773E+02, 3.8187E+02, 3.8602E+02,
     + 3.9018E+02, 3.9436E+02, 3.9855E+02, 4.0275E+02, 4.0696E+02,
     +  4.1119E+02, 4.1544E+02, 4.1970E+02, 4.2397E+02, 4.2826E+02,
     + 4.3256E+02, 4.3688E+02, 4.4122E+02, 4.4557E+02, 4.4994E+02,
     +  4.5432E+02, 4.5872E+02, 4.6314E+02, 4.6758E+02, 4.7203E+02,
     + 4.7650E+02, 4.8100E+02, 4.8551E+02, 4.9003E+02, 4.9458E+02,
     +  4.9915E+02, 5.0374E+02, 5.0834E+02, 5.1297E+02, 5.1762E+02,
     + 5.2229E+02, 5.2698E+02, 5.3169E+02, 5.3642E+02, 5.4118E+02,
     +  5.4595E+02, 5.5075E+02, 5.5557E+02, 5.6042E+02, 5.6529E+02,
     + 5.7018E+02, 5.7509E+02, 5.8003E+02, 5.8499E+02, 5.8998E+02,
     +  5.9499E+02, 6.0002E+02, 6.0508E+02, 6.1017E+02, 6.1528E+02,
     + 6.2041E+02, 6.2558E+02, 6.3076E+02, 6.3598E+02, 6.4122E+02,
     +  6.4648E+02, 6.5178E+02, 6.5710E+02, 6.6245E+02, 6.6782E+02,
     + 6.7323E+02, 6.7866E+02, 6.8412E+02, 6.8961E+02, 6.9512E+02,
     +  7.0067E+02, 7.0624E+02, 7.1184E+02, 7.1748E+02, 7.2314E+02,
     + 7.2883E+02, 7.3455E+02, 7.4031E+02, 7.4609E+02, 7.5190E+02,
     +  7.5775E+02, 7.6362E+02, 7.6953E+02, 7.7546E+02, 7.8143E+02,
     + 7.8743E+02, 7.9347E+02, 7.9953E+02, 8.0563E+02, 8.1176E+02,
     +  8.1792E+02, 8.2411E+02, 8.3034E+02, 8.3660E+02, 8.4290E+02,
     + 8.4923E+02, 8.5559E+02, 8.6199E+02, 8.6842E+02, 8.7488E+02,
     +  8.8138E+02, 8.8791E+02, 8.9448E+02, 9.0109E+02, 9.0773E+02,
     + 9.1440E+02, 9.2111E+02, 9.2786E+02, 9.3464E+02, 9.4146E+02,
     +  9.4832E+02, 9.5521E+02, 9.6214E+02, 9.6910E+02, 9.7610E+02,
     + 9.8314E+02, 9.9022E+02, 9.9734E+02, 1.0045E+03, 1.0117E+03,
     +  1.0189E+03, 1.0262E+03, 1.0335E+03, 1.0408E+03, 1.0482E+03,
     + 1.0556E+03, 1.0631E+03, 1.0706E+03, 1.0781E+03, 1.0857E+03,
     +  1.0933E+03, 1.1010E+03, 1.1087E+03, 1.1164E+03, 1.1242E+03,
     + 1.1320E+03, 1.1399E+03, 1.1478E+03, 1.1558E+03, 1.1637E+03,
     +  1.1718E+03, 1.1798E+03, 1.1880E+03, 1.1961E+03, 1.2043E+03,
     + 1.2126E+03, 1.2208E+03, 1.2292E+03, 1.2375E+03, 1.2460E+03,
     +  1.2544E+03, 1.2629E+03, 1.2715E+03, 1.2800E+03, 1.2887E+03,
     + 1.2973E+03, 1.3061E+03, 1.3148E+03, 1.3236E+03, 1.3325E+03,
     +  1.3414E+03, 1.3503E+03, 1.3593E+03, 1.3684E+03, 1.3775E+03,
     + 1.3866E+03, 1.3958E+03, 1.4050E+03, 1.4143E+03, 1.4236E+03,
     +  1.4329E+03, 1.4423E+03, 1.4518E+03, 1.4613E+03, 1.4709E+03,
     + 1.4805E+03, 1.4901E+03, 1.4998E+03, 1.5095E+03, 1.5193E+03,
     +  1.5292E+03, 1.5391E+03, 1.5490E+03, 1.5590E+03, 1.5690E+03,
     + 1.5791E+03, 1.5893E+03, 1.5995E+03, 1.6097E+03, 1.6200E+03,
     +  1.6303E+03, 1.6407E+03, 1.6512E+03, 1.6617E+03, 1.6722E+03,
     + 1.6828E+03, 1.6934E+03, 1.7041E+03, 1.7149E+03, 1.7257E+03,
     +  1.7366E+03, 1.7475E+03, 1.7584E+03, 1.7694E+03, 1.7805E+03,
     + 1.7916E+03, 1.8028E+03, 1.8140E+03, 1.8253E+03, 1.8366E+03,
     +  1.8480E+03, 1.8595E+03, 1.8710E+03, 1.8825E+03, 1.8941E+03,
     + 1.9058E+03, 1.9175E+03, 1.9293E+03, 1.9411E+03, 1.9530E+03,
     +  1.9650E+03, 1.9770E+03, 1.9890E+03, 2.0011E+03, 2.0133E+03,
     + 2.0255E+03, 2.0378E+03, 2.0501E+03, 2.0625E+03, 2.0750E+03,
     +  2.0875E+03, 2.1001E+03, 2.1127E+03, 2.1254E+03, 2.1381E+03,
     + 2.1510E+03, 2.1638E+03, 2.1767E+03, 2.1897E+03, 2.2028E+03,
     +  2.2159E+03, 2.2290E+03, 2.2423E+03, 2.2556E+03, 2.2689E+03,
     + 2.2823E+03, 2.2958E+03, 2.3093E+03, 2.3229E+03, 2.3366E+03,
     +  2.3503E+03, 2.3641E+03, 2.3779E+03, 2.3918E+03, 2.4058E+03,
     + 2.4198E+03, 2.4339E+03, 2.4480E+03, 2.4623E+03, 2.4766E+03,
     +  2.4909E+03, 2.5053E+03, 2.5198E+03, 2.5343E+03, 2.5489E+03,
     + 2.5636E+03, 2.5783E+03, 2.5931E+03, 2.6080E+03, 2.6229E+03,
     +  2.6379E+03, 2.6530E+03, 2.6681E+03, 2.6833E+03, 2.6986E+03,
     + 2.7139E+03, 2.7293E+03, 2.7448E+03, 2.7603E+03, 2.7759E+03,
     +  2.7916E+03, 2.8073E+03, 2.8232E+03, 2.8390E+03, 2.8550E+03,
     + 2.8710E+03, 2.8871E+03, 2.9032E+03, 2.9195E+03, 2.9358E+03,
     +  2.9521E+03, 2.9686E+03, 2.9851E+03, 3.0016E+03, 3.0183E+03,
     + 3.0350E+03, 3.0518E+03, 3.0687E+03, 3.0856E+03, 3.1026E+03,
     +  3.1197E+03, 3.1368E+03, 3.1541E+03, 3.1714E+03, 3.1887E+03,
     + 3.2062E+03, 3.2237E+03, 3.2413E+03, 3.2590E+03, 3.2767E+03,
     +  3.2945E+03, 3.3124E+03, 3.3304E+03, 3.3484E+03, 3.3666E+03,
     + 3.3848E+03, 3.4030E+03, 3.4214E+03, 3.4398E+03, 3.4583E+03,
     +  3.4769E+03, 3.4955E+03, 3.5143E+03, 3.5331E+03, 3.5520E+03,
     + 3.5710E+03, 3.5900E+03, 3.6091E+03, 3.6283E+03, 3.6476E+03,
     +  3.6670E+03, 3.6864E+03, 3.7060E+03, 3.7256E+03, 3.7452E+03,
     + 3.7650E+03, 3.7849E+03, 3.8048E+03, 3.8248E+03, 3.8449E+03,
     +  3.8650E+03, 3.8853E+03, 3.9056E+03, 3.9261E+03, 3.9466E+03,
     + 3.9671E+03, 3.9878E+03, 4.0086E+03, 4.0294E+03, 4.0503E+03,
     +  4.0713E+03, 4.0924E+03, 4.1136E+03, 4.1349E+03, 4.1562E+03,
     + 4.1776E+03, 4.1991E+03, 4.2207E+03, 4.2424E+03, 4.2642E+03,
     +  4.2861E+03, 4.3080E+03, 4.3301E+03, 4.3522E+03, 4.3744E+03,
     + 4.3967E+03, 4.4191E+03, 4.4416E+03, 4.4641E+03, 4.4868E+03,
     +  4.5095E+03, 4.5324E+03, 4.5553E+03, 4.5783E+03, 4.6014E+03,
     + 4.6246E+03, 4.6479E+03, 4.6713E+03, 4.6947E+03, 4.7183E+03,
     +  4.7419E+03, 4.7657E+03, 4.7895E+03, 4.8134E+03, 4.8375E+03,
     + 4.8616E+03, 4.8858E+03, 4.9101E+03, 4.9345E+03, 4.9590E+03,
     +  4.9836E+03, 5.0083E+03, 5.0330E+03, 5.0579E+03, 5.0829E+03,
     + 5.1079E+03, 5.1331E+03, 5.1584E+03, 5.1837E+03, 5.2092E+03,
     +  5.2347E+03, 5.2604E+03, 5.2861E+03, 5.3120E+03, 5.3379E+03,
     + 5.3639E+03, 5.3901E+03, 5.4163E+03, 5.4427E+03, 5.4691E+03,
     +  5.4956E+03, 5.5223E+03, 5.5490E+03, 5.5758E+03, 5.6028E+03,
     + 5.6298E+03, 5.6570E+03, 5.6842E+03, 5.7116E+03, 5.7390E+03,
     +  5.7666E+03, 5.7942E+03, 5.8220E+03, 5.8499E+03, 5.8778E+03,
     + 5.9059E+03, 5.9341E+03, 5.9624E+03, 5.9908E+03, 6.0193E+03,
     +  6.0479E+03, 6.0766E+03, 6.1054E+03, 6.1343E+03, 6.1633E+03,
     + 6.1925E+03, 6.2217E+03, 6.2511E+03, 6.2805E+03, 6.3101E+03,
     +  6.3398E+03, 6.3696E+03, 6.3995E+03, 6.4295E+03, 6.4596E+03,
     + 6.4898E+03, 6.5201E+03, 6.5506E+03, 6.5811E+03, 6.6118E+03,
     +  6.6426E+03, 6.6735E+03, 6.7045E+03, 6.7356E+03, 6.7668E+03,
     + 6.7982E+03, 6.8296E+03, 6.8612E+03, 6.8929E+03, 6.9247E+03,
     +  6.9566E+03, 6.9887E+03, 7.0208E+03, 7.0531E+03, 7.0854E+03,
     + 7.1179E+03, 7.1505E+03, 7.1833E+03, 7.2161E+03, 7.2491E+03,
     +  7.2822E+03, 7.3154E+03, 7.3487E+03, 7.3821E+03, 7.4157E+03,
     + 7.4494E+03, 7.4832E+03, 7.5171E+03, 7.5511E+03, 7.5853E+03,
     +  7.6195E+03, 7.6540E+03, 7.6885E+03, 7.7231E+03, 7.7579E+03,
     + 7.7928E+03, 7.8278E+03, 7.8629E+03, 7.8982E+03, 7.9336E+03,
     +  7.9691E+03, 8.0047E+03, 8.0405E+03, 8.0763E+03, 8.1124E+03,
     + 8.1485E+03, 8.1848E+03, 8.2211E+03, 8.2577E+03, 8.2943E+03,
     +  8.3311E+03, 8.3680E+03, 8.4050E+03, 8.4421E+03, 8.4794E+03,
     + 8.5168E+03, 8.5544E+03, 8.5921E+03, 8.6299E+03, 8.6678E+03,
     +  8.7059E+03, 8.7441E+03, 8.7824E+03, 8.8209E+03, 8.8594E+03,
     + 8.8982E+03, 8.9370E+03, 8.9760E+03, 9.0151E+03, 9.0544E+03,
     +  9.0938E+03, 9.1333E+03, 9.1730E+03, 9.2128E+03, 9.2527E+03,
     + 9.2928E+03, 9.3330E+03, 9.3733E+03, 9.4138E+03, 9.4544E+03,
     +  9.4952E+03, 9.5361E+03, 9.5771E+03, 9.6183E+03, 9.6596E+03,
     + 9.7011E+03, 9.7427E+03, 9.7844E+03, 9.8263E+03, 9.8683E+03,
     +  9.9104E+03, 9.9527E+03, 9.9952E+03, 1.0038E+04, 1.0080E+04,
     + 1.0123E+04, 1.0166E+04, 1.0209E+04, 1.0253E+04, 1.0296E+04,
     +  1.0340E+04, 1.0384E+04, 1.0427E+04, 1.0471E+04, 1.0516E+04,
     + 1.0560E+04, 1.0604E+04, 1.0649E+04, 1.0694E+04, 1.0739E+04,
     +  1.0784E+04, 1.0829E+04, 1.0874E+04, 1.0920E+04, 1.0965E+04,
     + 1.1011E+04, 1.1057E+04, 1.1103E+04, 1.1150E+04, 1.1196E+04,
     +  1.1243E+04, 1.1289E+04, 1.1336E+04, 1.1383E+04, 1.1430E+04,
     + 1.1478E+04, 1.1525E+04, 1.1573E+04, 1.1621E+04, 1.1669E+04,
     +  1.1717E+04, 1.1765E+04, 1.1813E+04, 1.1862E+04, 1.1911E+04,
     + 1.1960E+04, 1.2009E+04, 1.2058E+04, 1.2107E+04, 1.2157E+04,
     +  1.2206E+04, 1.2256E+04, 1.2306E+04, 1.2356E+04, 1.2407E+04,
     + 1.2457E+04, 1.2508E+04, 1.2559E+04, 1.2610E+04, 1.2661E+04,
     +  1.2712E+04, 1.2764E+04, 1.2815E+04, 1.2867E+04, 1.2919E+04,
     + 1.2971E+04, 1.3023E+04, 1.3076E+04, 1.3129E+04, 1.3181E+04,
     +  1.3234E+04, 1.3287E+04, 1.3341E+04, 1.3394E+04, 1.3448E+04,
     + 1.3502E+04, 1.3556E+04, 1.3610E+04, 1.3664E+04, 1.3719E+04,
     +  1.3773E+04, 1.3828E+04, 1.3883E+04, 1.3938E+04, 1.3994E+04,
     + 1.4049E+04, 1.4105E+04, 1.4161E+04, 1.4217E+04, 1.4273E+04,
     +  1.4329E+04, 1.4386E+04, 1.4443E+04, 1.4500E+04, 1.4557E+04,
     + 1.4614E+04, 1.4672E+04, 1.4729E+04, 1.4787E+04, 1.4845E+04,
     +  1.4903E+04, 1.4962E+04, 1.5020E+04, 1.5079E+04, 1.5138E+04,
     + 1.5197E+04, 1.5256E+04, 1.5316E+04, 1.5375E+04, 1.5435E+04,
     +  1.5495E+04, 1.5555E+04, 1.5616E+04, 1.5676E+04, 1.5737E+04,
     + 1.5798E+04, 1.5859E+04, 1.5920E+04, 1.5982E+04, 1.6044E+04,
     +  1.6105E+04, 1.6167E+04, 1.6230E+04, 1.6292E+04, 1.6355E+04,
     + 1.6418E+04, 1.6481E+04, 1.6544E+04, 1.6607E+04, 1.6671E+04,
     +  1.6735E+04, 1.6799E+04, 1.6863E+04, 1.6927E+04, 1.6992E+04,
     + 1.7056E+04, 1.7121E+04, 1.7186E+04, 1.7252E+04, 1.7317E+04,
     +  1.7383E+04, 1.7449E+04, 1.7515E+04, 1.7581E+04, 1.7648E+04,
     + 1.7715E+04, 1.7782E+04, 1.7849E+04, 1.7916E+04, 1.7984E+04,
     +  1.8051E+04, 1.8119E+04, 1.8187E+04, 1.8256E+04, 1.8324E+04,
     + 1.8393E+04, 1.8462E+04, 1.8531E+04, 1.8600E+04, 1.8670E+04,
     +  1.8740E+04, 1.8810E+04, 1.8880E+04, 1.8950E+04, 1.9021E+04,
     + 1.9092E+04, 1.9163E+04, 1.9234E+04, 1.9305E+04, 1.9377E+04,
     +  1.9449E+04, 1.9521E+04, 1.9593E+04, 1.9666E+04, 1.9738E+04,
     + 1.9811E+04, 1.9884E+04, 1.9958E+04, 2.0031E+04, 2.0105E+04,
     +  2.0179E+04, 2.0253E+04, 2.0328E+04, 2.0402E+04, 2.0477E+04,
     + 2.0552E+04, 2.0627E+04, 2.0703E+04, 2.0779E+04, 2.0855E+04,
     +  2.0931E+04, 2.1007E+04, 2.1084E+04, 2.1161E+04, 2.1238E+04,
     + 2.1315E+04, 2.1392E+04, 2.1470E+04, 2.1548E+04, 2.1626E+04,
     +  2.1705E+04, 2.1783E+04, 2.1862E+04, 2.1941E+04, 2.2020E+04,
     + 2.2100E+04, 2.2180E+04, 2.2260E+04, 2.2340E+04, 2.2420E+04,
     +  2.2501E+04, 2.2582E+04, 2.2663E+04, 2.2744E+04, 2.2826E+04,
     + 2.2908E+04, 2.2990E+04, 2.3072E+04, 2.3154E+04, 2.3237E+04,
     +  2.3320E+04, 2.3403E+04, 2.3487E+04, 2.3571E+04, 2.3654E+04,
     + 2.3739E+04, 2.3823E+04, 2.3908E+04, 2.3992E+04, 2.4078E+04,
     +  2.4163E+04, 2.4249E+04, 2.4334E+04, 2.4420E+04, 2.4507E+04,
     + 2.4593E+04, 2.4680E+04, 2.4767E+04, 2.4854E+04, 2.4942E+04,
     +  2.5030E+04, 2.5118E+04, 2.5206E+04, 2.5295E+04, 2.5383E+04,
     + 2.5472E+04, 2.5562E+04, 2.5651E+04, 2.5741E+04, 2.5831E+04,
     +  2.5921E+04, 2.6012E+04, 2.6102E+04, 2.6193E+04, 2.6285E+04,
     + 2.6376E+04, 2.6468E+04, 2.6560E+04, 2.6652E+04, 2.6745E+04,
     +  2.6837E+04, 2.6930E+04, 2.7024E+04, 2.7117E+04, 2.7211E+04,
     + 2.7305E+04, 2.7399E+04, 2.7494E+04, 2.7589E+04, 2.7684E+04,
     +  2.7779E+04, 2.7875E+04, 2.7971E+04, 2.8067E+04, 2.8163E+04,
     + 2.8260E+04, 2.8357E+04, 2.8454E+04, 2.8552E+04, 2.8649E+04,
     +  2.8747E+04, 2.8845E+04, 2.8944E+04, 2.9043E+04, 2.9142E+04,
     + 2.9241E+04, 2.9341E+04, 2.9441E+04, 2.9541E+04, 2.9641E+04,
     +  2.9742E+04, 2.9843E+04, 2.9944E+04, 3.0045E+04, 3.0147E+04,
     + 3.0249E+04, 3.0351E+04, 3.0454E+04, 3.0557E+04, 3.0660E+04,
     +  3.0763E+04, 3.0867E+04, 3.0971E+04, 3.1075E+04, 3.1180E+04,
     + 3.1285E+04, 3.1390E+04, 3.1495E+04, 3.1601E+04, 3.1707E+04,
     +  3.1813E+04, 3.1919E+04, 3.2026E+04, 3.2133E+04, 3.2240E+04,
     + 3.2348E+04, 3.2456E+04, 3.2564E+04, 3.2672E+04, 3.2781E+04,
     +  3.2890E+04, 3.3000E+04, 3.3109E+04, 3.3219E+04, 3.3329E+04,
     + 3.3440E+04, 3.3551E+04, 3.3662E+04, 3.3773E+04, 3.3885E+04,
     +  3.3997E+04, 3.4109E+04, 3.4221E+04, 3.4334E+04, 3.4447E+04,
     + 3.4561E+04, 3.4674E+04, 3.4788E+04, 3.4903E+04, 3.5017E+04,
     +  3.5132E+04, 3.5248E+04, 3.5363E+04, 3.5479E+04, 3.5595E+04,
     + 3.5711E+04, 3.5828E+04, 3.5945E+04, 3.6062E+04, 3.6180E+04,
     +  3.6298E+04, 3.6416E+04, 3.6535E+04, 3.6654E+04, 3.6773E+04,
     + 3.6892E+04, 3.7012E+04, 3.7132E+04, 3.7252E+04, 3.7373E+04,
     +  3.7494E+04, 3.7615E+04, 3.7737E+04, 3.7859E+04, 3.7981E+04,
     + 3.8104E+04, 3.8227E+04, 3.8350E+04, 3.8473E+04, 3.8597E+04,
     +  3.8721E+04, 3.8846E+04, 3.8971E+04, 3.9096E+04, 3.9221E+04,
     + 3.9347E+04, 3.9473E+04, 3.9599E+04, 3.9726E+04, 3.9853E+04,
     +  3.9980E+04, 4.0108E+04, 4.0236E+04, 4.0364E+04, 4.0493E+04,
     + 4.0622E+04, 4.0751E+04, 4.0881E+04, 4.1011E+04, 4.1141E+04,
     +  4.1272E+04, 4.1402E+04, 4.1534E+04, 4.1665E+04, 4.1797E+04,
     + 4.1929E+04, 4.2062E+04, 4.2195E+04, 4.2328E+04, 4.2462E+04,
     +  4.2596E+04, 4.2730E+04, 4.2865E+04, 4.2999E+04, 4.3135E+04,
     + 4.3270E+04, 4.3406E+04, 4.3543E+04, 4.3679E+04, 4.3816E+04,
     +  4.3953E+04, 4.4091E+04, 4.4229E+04, 4.4367E+04, 4.4506E+04,
     + 4.4645E+04, 4.4784E+04, 4.4924E+04, 4.5064E+04, 4.5204E+04,
     +  4.5345E+04, 4.5486E+04, 4.5628E+04, 4.5770E+04, 4.5912E+04,
     + 4.6054E+04, 4.6197E+04, 4.6340E+04, 4.6484E+04, 4.6628E+04,
     +  4.6772E+04, 4.6916E+04, 4.7061E+04, 4.7207E+04, 4.7352E+04,
     + 4.7498E+04, 4.7645E+04, 4.7791E+04, 4.7939E+04, 4.8086E+04,
     +  4.8234E+04, 4.8382E+04, 4.8531E+04, 4.8679E+04, 4.8829E+04,
     + 4.8978E+04, 4.9128E+04, 4.9279E+04, 4.9429E+04, 4.9581E+04,
     +  4.9732E+04, 4.9884E+04, 5.0036E+04, 5.0189E+04, 5.0341E+04,
     + 5.0495E+04, 5.0648E+04, 5.0802E+04, 5.0957E+04, 5.1112E+04,
     +  5.1267E+04, 5.1422E+04, 5.1578E+04, 5.1734E+04, 5.1891E+04,
     + 5.2048E+04, 5.2206E+04, 5.2363E+04, 5.2522E+04, 5.2680E+04,
     +  5.2839E+04, 5.2998E+04, 5.3158E+04, 5.3318E+04, 5.3478E+04,
     + 5.3639E+04, 5.3801E+04, 5.3962E+04, 5.4124E+04, 5.4286E+04,
     +  5.4449E+04, 5.4612E+04, 5.4776E+04, 5.4940E+04, 5.5104E+04,
     + 5.5269E+04, 5.5434E+04, 5.5599E+04, 5.5765E+04, 5.5932E+04,
     +  5.6098E+04, 5.6265E+04, 5.6433E+04, 5.6601E+04, 5.6769E+04,
     + 5.6937E+04, 5.7106E+04, 5.7276E+04, 5.7446E+04, 5.7616E+04,
     +  5.7787E+04, 5.7958E+04, 5.8129E+04, 5.8301E+04, 5.8473E+04,
     + 5.8646E+04, 5.8819E+04, 5.8992E+04, 5.9166E+04, 5.9340E+04,
     +  5.9515E+04, 5.9690E+04, 5.9866E+04, 6.0041E+04, 6.0218E+04,
     + 6.0395E+04, 6.0572E+04, 6.0749E+04, 6.0927E+04, 6.1105E+04,
     +  6.1284E+04, 6.1463E+04, 6.1643E+04, 6.1823E+04, 6.2004E+04,
     + 6.2184E+04, 6.2366E+04, 6.2547E+04, 6.2730E+04, 6.2912E+04,
     +  6.3095E+04, 6.3278E+04, 6.3462E+04, 6.3646E+04, 6.3831E+04,
     + 6.4016E+04, 6.4202E+04, 6.4388E+04, 6.4574E+04, 6.4761E+04,
     +  6.4948E+04, 6.5136E+04, 6.5324E+04, 6.5512E+04, 6.5701E+04,
     + 6.5891E+04, 6.6080E+04, 6.6271E+04, 6.6461E+04, 6.6652E+04,
     +  6.6844E+04, 6.7036E+04, 6.7228E+04, 6.7421E+04, 6.7614E+04,
     + 6.7808E+04, 6.8002E+04, 6.8197E+04, 6.8392E+04, 6.8587E+04,
     +  6.8783E+04, 6.8980E+04, 6.9177E+04, 6.9374E+04, 6.9572E+04,
     + 6.9770E+04, 6.9968E+04, 7.0167E+04, 7.0367E+04, 7.0567E+04,
     +  7.0767E+04, 7.0968E+04, 7.1170E+04, 7.1371E+04, 7.1574E+04,
     + 7.1776E+04, 7.1979E+04, 7.2183E+04, 7.2387E+04, 7.2591E+04,
     +  7.2796E+04, 7.3002E+04, 7.3208E+04, 7.3414E+04, 7.3621E+04,
     + 7.3828E+04, 7.4036E+04, 7.4244E+04, 7.4453E+04, 7.4662E+04,
     +  7.4871E+04, 7.5081E+04, 7.5292E+04, 7.5503E+04, 7.5714E+04,
     + 7.5926E+04, 7.6139E+04, 7.6352E+04, 7.6565E+04, 7.6779E+04,
     +  7.6993E+04, 7.7208E+04, 7.7423E+04, 7.7639E+04, 7.7855E+04,
     + 7.8071E+04, 7.8289E+04, 7.8506E+04, 7.8724E+04, 7.8943E+04,
     +  7.9162E+04, 7.9381E+04, 7.9601E+04, 7.9822E+04, 8.0043E+04,
     + 8.0264E+04, 8.0486E+04, 8.0709E+04, 8.0932E+04, 8.1155E+04,
     +  8.1379E+04, 8.1603E+04, 8.1828E+04, 8.2054E+04, 8.2280E+04,
     + 8.2506E+04, 8.2733E+04, 8.2960E+04, 8.3188E+04, 8.3416E+04,
     +  8.3645E+04, 8.3874E+04, 8.4104E+04, 8.4335E+04, 8.4565E+04,
     + 8.4797E+04, 8.5029E+04, 8.5261E+04, 8.5494E+04, 8.5727E+04,
     +  8.5961E+04, 8.6195E+04, 8.6430E+04, 8.6666E+04, 8.6901E+04,
     + 8.7138E+04, 8.7375E+04, 8.7612E+04, 8.7850E+04, 8.8088E+04,
     +  8.8327E+04, 8.8567E+04, 8.8807E+04, 8.9047E+04, 8.9288E+04,
     + 8.9530E+04, 8.9772E+04, 9.0014E+04, 9.0257E+04, 9.0501E+04,
     +  9.0745E+04, 9.0990E+04, 9.1235E+04, 9.1481E+04, 9.1727E+04,
     + 9.1973E+04, 9.2221E+04, 9.2468E+04, 9.2717E+04, 9.2966E+04,
     +  9.3215E+04, 9.3465E+04, 9.3715E+04, 9.3966E+04, 9.4218E+04,
     + 9.4470E+04, 9.4722E+04, 9.4975E+04, 9.5229E+04, 9.5483E+04,
     +  9.5738E+04, 9.5993E+04, 9.6249E+04, 9.6505E+04, 9.6762E+04,
     + 9.7019E+04, 9.7277E+04, 9.7536E+04, 9.7795E+04, 9.8054E+04,
     +  9.8314E+04, 9.8575E+04, 9.8836E+04, 9.9098E+04, 9.9360E+04,
     + 9.9623E+04, 9.9886E+04, 1.0015E+05, 1.0041E+05, 1.0068E+05,
     +  1.0095E+05, 1.0121E+05, 1.0148E+05, 1.0175E+05, 1.0201E+05,
     + 1.0228E+05, 1.0255E+05, 1.0282E+05, 1.0309E+05, 1.0336E+05,
     +  1.0363E+05, 1.0390E+05, 1.0418E+05, 1.0445E+05, 1.0472E+05,
     + 1.0500E+05, 1.0527E+05, 1.0555E+05, 1.0582E+05, 1.0610E+05,
     +  1.0638E+05, 1.0665E+05, 1.0693E+05, 1.0721E+05, 1.0749E+05,
     + 1.0777E+05, 1.0805E+05, 1.0833E+05, 1.0861E+05, 1.0889E+05,
     +  1.0918E+05, 1.0946E+05, 1.0974E+05, 1.1003E+05, 1.1031E+05,
     + 1.1060E+05, 1.1088E+05, 1.1117E+05, 1.1146E+05, 1.1175E+05,
     +  1.1203E+05, 1.1232E+05, 1.1261E+05, 1.1290E+05, 1.1319E+05,
     + 1.1348E+05, 1.1378E+05, 1.1407E+05, 1.1436E+05, 1.1466E+05,
     +  1.1495E+05, 1.1525E+05, 1.1554E+05, 1.1584E+05, 1.1613E+05,
     + 1.1643E+05, 1.1673E+05, 1.1703E+05, 1.1733E+05, 1.1763E+05,
     +  1.1793E+05, 1.1823E+05, 1.1853E+05, 1.1883E+05, 1.1914E+05,
     + 1.1944E+05, 1.1974E+05, 1.2005E+05, 1.2035E+05, 1.2066E+05,
     +  1.2096E+05, 1.2127E+05, 1.2158E+05, 1.2189E+05, 1.2220E+05,
     + 1.2251E+05, 1.2282E+05, 1.2313E+05, 1.2344E+05, 1.2375E+05,
     +  1.2406E+05, 1.2438E+05, 1.2469E+05, 1.2501E+05, 1.2532E+05,
     + 1.2564E+05, 1.2595E+05, 1.2627E+05, 1.2659E+05, 1.2691E+05,
     +  1.2723E+05, 1.2754E+05, 1.2787E+05, 1.2819E+05, 1.2851E+05,
     + 1.2883E+05, 1.2915E+05, 1.2948E+05, 1.2980E+05, 1.3012E+05,
     +  1.3045E+05, 1.3078E+05, 1.3110E+05, 1.3143E+05, 1.3176E+05,
     + 1.3209E+05, 1.3242E+05, 1.3275E+05, 1.3308E+05, 1.3341E+05,
     +  1.3374E+05, 1.3407E+05, 1.3440E+05, 1.3474E+05, 1.3507E+05,
     + 1.3541E+05, 1.3574E+05, 1.3608E+05, 1.3642E+05, 1.3676E+05,
     +  1.3709E+05, 1.3743E+05, 1.3777E+05, 1.3811E+05, 1.3845E+05,
     + 1.3880E+05, 1.3914E+05, 1.3948E+05, 1.3982E+05, 1.4017E+05,
     +  1.4051E+05, 1.4086E+05, 1.4121E+05, 1.4155E+05, 1.4190E+05,
     + 1.4225E+05, 1.4260E+05, 1.4295E+05, 1.4330E+05, 1.4365E+05,
     +  1.4400E+05, 1.4435E+05, 1.4471E+05, 1.4506E+05, 1.4542E+05,
     + 1.4577E+05, 1.4613E+05, 1.4648E+05, 1.4684E+05, 1.4720E+05,
     +  1.4756E+05, 1.4792E+05, 1.4828E+05, 1.4864E+05, 1.4900E+05,
     + 1.4936E+05, 1.4972E+05, 1.5009E+05, 1.5045E+05, 1.5082E+05,
     +  1.5118E+05, 1.5155E+05, 1.5191E+05, 1.5228E+05, 1.5265E+05,
     + 1.5302E+05, 1.5339E+05, 1.5376E+05, 1.5413E+05, 1.5450E+05,
     +  1.5488E+05, 1.5525E+05, 1.5562E+05, 1.5600E+05, 1.5637E+05,
     + 1.5675E+05, 1.5713E+05, 1.5750E+05, 1.5788E+05, 1.5826E+05,
     +  1.5864E+05, 1.5902E+05, 1.5940E+05, 1.5978E+05, 1.6017E+05,
     + 1.6055E+05, 1.6093E+05, 1.6132E+05, 1.6170E+05, 1.6209E+05,
     +  1.6248E+05, 1.6287E+05, 1.6325E+05, 1.6364E+05, 1.6403E+05,
     + 1.6442E+05, 1.6481E+05, 1.6521E+05, 1.6560E+05, 1.6599E+05,
     +  1.6639E+05, 1.6678E+05, 1.6718E+05, 1.6757E+05, 1.6797E+05,
     + 1.6837E+05, 1.6877E+05, 1.6917E+05, 1.6957E+05, 1.6997E+05,
     +  1.7037E+05, 1.7077E+05, 1.7118E+05, 1.7158E+05, 1.7199E+05,
     + 1.7239E+05, 1.7280E+05, 1.7320E+05, 1.7361E+05, 1.7402E+05,
     +  1.7443E+05, 1.7484E+05, 1.7525E+05, 1.7566E+05, 1.7607E+05,
     + 1.7649E+05, 1.7690E+05, 1.7732E+05, 1.7773E+05, 1.7815E+05,
     +  1.7856E+05, 1.7898E+05, 1.7940E+05, 1.7982E+05, 1.8024E+05,
     + 1.8066E+05, 1.8108E+05, 1.8150E+05, 1.8193E+05, 1.8235E+05,
     +  1.8278E+05, 1.8320E+05, 1.8363E+05, 1.8405E+05, 1.8448E+05,
     + 1.8491E+05, 1.8534E+05, 1.8577E+05, 1.8620E+05, 1.8663E+05,
     +  1.8706E+05, 1.8750E+05, 1.8793E+05, 1.8837E+05, 1.8880E+05,
     + 1.8924E+05, 1.8968E+05, 1.9011E+05, 1.9055E+05, 1.9099E+05,
     +  1.9143E+05, 1.9187E+05, 1.9232E+05, 1.9276E+05, 1.9320E+05,
     + 1.9365E+05, 1.9409E+05, 1.9454E+05, 1.9498E+05, 1.9543E+05,
     +  1.9588E+05, 1.9633E+05, 1.9678E+05, 1.9723E+05, 1.9768E+05,
     + 1.9813E+05, 1.9859E+05, 1.9904E+05, 1.9950E+05, 1.9995E+05,
     +  2.0041E+05, 2.0087E+05, 2.0132E+05, 2.0178E+05, 2.0224E+05,
     + 2.0270E+05, 2.0317E+05, 2.0363E+05, 2.0409E+05, 2.0455E+05,
     +  2.0502E+05, 2.0549E+05, 2.0595E+05, 2.0642E+05, 2.0689E+05,
     + 2.0736E+05, 2.0783E+05, 2.0830E+05, 2.0877E+05, 2.0924E+05,
     +  2.0971E+05, 2.1019E+05, 2.1066E+05, 2.1114E+05, 2.1161E+05,
     + 2.1209E+05, 2.1257E+05, 2.1305E+05, 2.1353E+05, 2.1401E+05,
     +  2.1449E+05, 2.1497E+05, 2.1546E+05, 2.1594E+05, 2.1643E+05,
     + 2.1691E+05, 2.1740E+05, 2.1789E+05, 2.1838E+05, 2.1886E+05,
     +  2.1935E+05, 2.1985E+05, 2.2034E+05, 2.2083E+05, 2.2132E+05,
     + 2.2182E+05, 2.2231E+05, 2.2281E+05, 2.2331E+05, 2.2381E+05,
     +  2.2430E+05, 2.2480E+05, 2.2530E+05, 2.2581E+05, 2.2631E+05,
     + 2.2681E+05, 2.2732E+05, 2.2782E+05, 2.2833E+05, 2.2883E+05,
     +  2.2934E+05, 2.2985E+05, 2.3036E+05, 2.3087E+05, 2.3138E+05,
     + 2.3189E+05, 2.3241E+05, 2.3292E+05, 2.3343E+05, 2.3395E+05,
     +  2.3447E+05, 2.3498E+05, 2.3550E+05, 2.3602E+05, 2.3654E+05,
     + 2.3706E+05, 2.3758E+05, 2.3811E+05, 2.3863E+05, 2.3916E+05,
     +  2.3968E+05, 2.4021E+05, 2.4073E+05, 2.4126E+05, 2.4179E+05,
     + 2.4232E+05, 2.4285E+05, 2.4339E+05, 2.4392E+05, 2.4445E+05,
     +  2.4499E+05, 2.4552E+05, 2.4606E+05, 2.4660E+05, 2.4713E+05,
     + 2.4767E+05, 2.4821E+05, 2.4875E+05, 2.4930E+05, 2.4984E+05,
     +  2.5038E+05, 2.5093E+05, 2.5147E+05, 2.5202E+05, 2.5257E+05,
     + 2.5312E+05, 2.5367E+05, 2.5422E+05, 2.5477E+05, 2.5532E+05,
     +  2.5587E+05, 2.5643E+05, 2.5698E+05, 2.5754E+05, 2.5810E+05,
     + 2.5865E+05, 2.5921E+05, 2.5977E+05, 2.6033E+05, 2.6090E+05,
     +  2.6146E+05, 2.6202E+05, 2.6259E+05, 2.6315E+05, 2.6372E+05,
     + 2.6429E+05, 2.6485E+05, 2.6542E+05, 2.6599E+05, 2.6656E+05,
     +  2.6714E+05, 2.6771E+05, 2.6828E+05, 2.6886E+05, 2.6944E+05,
     + 2.7001E+05, 2.7059E+05, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 3,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_CH3Cl (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 2),QofT( 2,NT),Q(NT)
      data (Tmx(I),I=1,2)/3500.,3500./
      data xgj/ 4.,4./
c...    CH3Cl
c...        --       215
      data (QofT( 1,J),J=1,2408)/ 2.2920E+01, 5.2031E+01, 8.5042E+01,
     +  1.2113E+02, 1.6003E+02, 2.0153E+02, 2.4548E+02, 2.9171E+02,
     + 3.4014E+02, 3.9067E+02, 4.4325E+02, 4.9785E+02, 5.5445E+02,
     +  6.1301E+02, 6.7354E+02, 7.3602E+02, 8.0043E+02, 8.6675E+02,
     + 9.3497E+02, 1.0051E+03, 1.1508E+03, 1.3037E+03, 1.4636E+03,
     +  1.6302E+03, 1.8033E+03, 1.9826E+03, 2.1680E+03, 2.3593E+03,
     + 2.5562E+03, 2.7586E+03, 2.9663E+03, 3.1791E+03, 3.3970E+03,
     +  3.6197E+03, 3.8472E+03, 4.0794E+03, 4.3162E+03, 4.5574E+03,
     + 4.8030E+03, 5.0529E+03, 5.3070E+03, 5.5653E+03, 5.8276E+03,
     +  6.0940E+03, 6.3643E+03, 6.6385E+03, 6.9166E+03, 7.1984E+03,
     + 7.4840E+03, 7.7733E+03, 8.0662E+03, 8.3627E+03, 8.6628E+03,
     +  8.9664E+03, 9.2734E+03, 9.5839E+03, 9.8978E+03, 1.0215E+04,
     + 1.0536E+04, 1.0860E+04, 1.1187E+04, 1.1517E+04, 1.1851E+04,
     +  1.2188E+04, 1.2528E+04, 1.2871E+04, 1.3217E+04, 1.3566E+04,
     + 1.3919E+04, 1.4274E+04, 1.4633E+04, 1.4994E+04, 1.5359E+04,
     +  1.5727E+04, 1.6097E+04, 1.6470E+04, 1.6847E+04, 1.7226E+04,
     + 1.7609E+04, 1.7994E+04, 1.8382E+04, 1.8773E+04, 1.9167E+04,
     +  1.9564E+04, 1.9963E+04, 2.0366E+04, 2.0771E+04, 2.1180E+04,
     + 2.1591E+04, 2.2005E+04, 2.2422E+04, 2.2842E+04, 2.3265E+04,
     +  2.3691E+04, 2.4120E+04, 2.4551E+04, 2.4986E+04, 2.5423E+04,
     + 2.5864E+04, 2.6307E+04, 2.6754E+04, 2.7203E+04, 2.7655E+04,
     +  2.8111E+04, 2.8569E+04, 2.9031E+04, 2.9495E+04, 2.9963E+04,
     + 3.0434E+04, 3.0908E+04, 3.1385E+04, 3.1865E+04, 3.2349E+04,
     +  3.2835E+04, 3.3325E+04, 3.3818E+04, 3.4315E+04, 3.4815E+04,
     + 3.5318E+04, 3.5824E+04, 3.6334E+04, 3.6848E+04, 3.7365E+04,
     +  3.7885E+04, 3.8409E+04, 3.8936E+04, 3.9467E+04, 4.0002E+04,
     + 4.0540E+04, 4.1082E+04, 4.1628E+04, 4.2177E+04, 4.2731E+04,
     +  4.3288E+04, 4.3849E+04, 4.4414E+04, 4.4982E+04, 4.5555E+04,
     + 4.6132E+04, 4.6713E+04, 4.7298E+04, 4.7887E+04, 4.8481E+04,
     +  4.9078E+04, 4.9680E+04, 5.0287E+04, 5.0897E+04, 5.1512E+04,
     + 5.2132E+04, 5.2756E+04, 5.3384E+04, 5.4018E+04, 5.4655E+04,
     +  5.5298E+04, 5.5945E+04, 5.6597E+04, 5.7254E+04, 5.7916E+04,
     + 5.8583E+04, 5.9255E+04, 5.9931E+04, 6.0613E+04, 6.1301E+04,
     +  6.1993E+04, 6.2690E+04, 6.3393E+04, 6.4101E+04, 6.4815E+04,
     + 6.5534E+04, 6.6259E+04, 6.6989E+04, 6.7725E+04, 6.8467E+04,
     +  6.9214E+04, 6.9967E+04, 7.0727E+04, 7.1492E+04, 7.2263E+04,
     + 7.3040E+04, 7.3823E+04, 7.4613E+04, 7.5408E+04, 7.6210E+04,
     +  7.7019E+04, 7.7833E+04, 7.8655E+04, 7.9483E+04, 8.0317E+04,
     + 8.1158E+04, 8.2006E+04, 8.2861E+04, 8.3723E+04, 8.4591E+04,
     +  8.5467E+04, 8.6350E+04, 8.7240E+04, 8.8137E+04, 8.9041E+04,
     + 8.9953E+04, 9.0872E+04, 9.1799E+04, 9.2734E+04, 9.3676E+04,
     +  9.4625E+04, 9.5583E+04, 9.6548E+04, 9.7522E+04, 9.8503E+04,
     + 9.9493E+04, 1.0049E+05, 1.0150E+05, 1.0251E+05, 1.0353E+05,
     +  1.0456E+05, 1.0560E+05, 1.0665E+05, 1.0771E+05, 1.0878E+05,
     + 1.0985E+05, 1.1093E+05, 1.1203E+05, 1.1313E+05, 1.1424E+05,
     +  1.1536E+05, 1.1649E+05, 1.1763E+05, 1.1878E+05, 1.1994E+05,
     + 1.2111E+05, 1.2228E+05, 1.2347E+05, 1.2467E+05, 1.2588E+05,
     +  1.2710E+05, 1.2833E+05, 1.2956E+05, 1.3081E+05, 1.3207E+05,
     + 1.3334E+05, 1.3463E+05, 1.3592E+05, 1.3722E+05, 1.3853E+05,
     +  1.3986E+05, 1.4120E+05, 1.4254E+05, 1.4390E+05, 1.4527E+05,
     + 1.4665E+05, 1.4805E+05, 1.4945E+05, 1.5087E+05, 1.5230E+05,
     +  1.5374E+05, 1.5520E+05, 1.5666E+05, 1.5814E+05, 1.5963E+05,
     + 1.6114E+05, 1.6265E+05, 1.6418E+05, 1.6572E+05, 1.6728E+05,
     +  1.6885E+05, 1.7043E+05, 1.7202E+05, 1.7363E+05, 1.7526E+05,
     + 1.7689E+05, 1.7854E+05, 1.8021E+05, 1.8188E+05, 1.8358E+05,
     +  1.8528E+05, 1.8700E+05, 1.8874E+05, 1.9049E+05, 1.9226E+05,
     + 1.9404E+05, 1.9583E+05, 1.9764E+05, 1.9947E+05, 2.0131E+05,
     +  2.0316E+05, 2.0504E+05, 2.0693E+05, 2.0883E+05, 2.1075E+05,
     + 2.1269E+05, 2.1464E+05, 2.1661E+05, 2.1860E+05, 2.2060E+05,
     +  2.2262E+05, 2.2466E+05, 2.2671E+05, 2.2878E+05, 2.3087E+05,
     + 2.3298E+05, 2.3510E+05, 2.3724E+05, 2.3940E+05, 2.4158E+05,
     +  2.4378E+05, 2.4599E+05, 2.4823E+05, 2.5048E+05, 2.5275E+05,
     + 2.5504E+05, 2.5735E+05, 2.5968E+05, 2.6203E+05, 2.6440E+05,
     +  2.6679E+05, 2.6920E+05, 2.7163E+05, 2.7408E+05, 2.7655E+05,
     + 2.7904E+05, 2.8155E+05, 2.8408E+05, 2.8663E+05, 2.8921E+05,
     +  2.9181E+05, 2.9442E+05, 2.9706E+05, 2.9973E+05, 3.0241E+05,
     + 3.0512E+05, 3.0785E+05, 3.1060E+05, 3.1337E+05, 3.1617E+05,
     +  3.1899E+05, 3.2184E+05, 3.2471E+05, 3.2760E+05, 3.3051E+05,
     + 3.3346E+05, 3.3642E+05, 3.3941E+05, 3.4242E+05, 3.4546E+05,
     +  3.4853E+05, 3.5162E+05, 3.5474E+05, 3.5788E+05, 3.6104E+05,
     + 3.6424E+05, 3.6746E+05, 3.7070E+05, 3.7398E+05, 3.7728E+05,
     +  3.8061E+05, 3.8396E+05, 3.8734E+05, 3.9076E+05, 3.9419E+05,
     + 3.9766E+05, 4.0116E+05, 4.0468E+05, 4.0824E+05, 4.1182E+05,
     +  4.1543E+05, 4.1907E+05, 4.2274E+05, 4.2645E+05, 4.3018E+05,
     + 4.3394E+05, 4.3773E+05, 4.4156E+05, 4.4541E+05, 4.4930E+05,
     +  4.5322E+05, 4.5717E+05, 4.6116E+05, 4.6517E+05, 4.6922E+05,
     + 4.7330E+05, 4.7742E+05, 4.8156E+05, 4.8575E+05, 4.8996E+05,
     +  4.9421E+05, 4.9850E+05, 5.0282E+05, 5.0717E+05, 5.1156E+05,
     + 5.1599E+05, 5.2045E+05, 5.2495E+05, 5.2948E+05, 5.3405E+05,
     +  5.3866E+05, 5.4331E+05, 5.4799E+05, 5.5271E+05, 5.5747E+05,
     + 5.6227E+05, 5.6710E+05, 5.7198E+05, 5.7689E+05, 5.8184E+05,
     +  5.8684E+05, 5.9187E+05, 5.9694E+05, 6.0206E+05, 6.0721E+05,
     + 6.1241E+05, 6.1765E+05, 6.2293E+05, 6.2825E+05, 6.3362E+05,
     +  6.3903E+05, 6.4448E+05, 6.4998E+05, 6.5552E+05, 6.6110E+05,
     + 6.6673E+05, 6.7240E+05, 6.7812E+05, 6.8388E+05, 6.8969E+05,
     +  6.9555E+05, 7.0145E+05, 7.0740E+05, 7.1340E+05, 7.1944E+05,
     + 7.2553E+05, 7.3167E+05, 7.3786E+05, 7.4410E+05, 7.5039E+05,
     +  7.5673E+05, 7.6311E+05, 7.6955E+05, 7.7604E+05, 7.8258E+05,
     + 7.8917E+05, 7.9582E+05, 8.0251E+05, 8.0926E+05, 8.1606E+05,
     +  8.2292E+05, 8.2983E+05, 8.3679E+05, 8.4381E+05, 8.5088E+05,
     + 8.5801E+05, 8.6520E+05, 8.7244E+05, 8.7974E+05, 8.8709E+05,
     +  8.9451E+05, 9.0198E+05, 9.0951E+05, 9.1710E+05, 9.2474E+05,
     + 9.3245E+05, 9.4022E+05, 9.4804E+05, 9.5593E+05, 9.6388E+05,
     +  9.7190E+05, 9.7997E+05, 9.8811E+05, 9.9631E+05, 1.0046E+06,
     + 1.0129E+06, 1.0213E+06, 1.0297E+06, 1.0383E+06, 1.0469E+06,
     +  1.0555E+06, 1.0642E+06, 1.0730E+06, 1.0819E+06, 1.0908E+06,
     + 1.0998E+06, 1.1089E+06, 1.1180E+06, 1.1272E+06, 1.1365E+06,
     +  1.1458E+06, 1.1553E+06, 1.1647E+06, 1.1743E+06, 1.1839E+06,
     + 1.1937E+06, 1.2034E+06, 1.2133E+06, 1.2232E+06, 1.2333E+06,
     +  1.2433E+06, 1.2535E+06, 1.2638E+06, 1.2741E+06, 1.2845E+06,
     + 1.2950E+06, 1.3055E+06, 1.3162E+06, 1.3269E+06, 1.3377E+06,
     +  1.3486E+06, 1.3595E+06, 1.3706E+06, 1.3817E+06, 1.3929E+06,
     + 1.4042E+06, 1.4156E+06, 1.4271E+06, 1.4387E+06, 1.4503E+06,
     +  1.4621E+06, 1.4739E+06, 1.4858E+06, 1.4978E+06, 1.5099E+06,
     + 1.5221E+06, 1.5344E+06, 1.5468E+06, 1.5592E+06, 1.5718E+06,
     +  1.5844E+06, 1.5972E+06, 1.6100E+06, 1.6230E+06, 1.6360E+06,
     + 1.6492E+06, 1.6624E+06, 1.6757E+06, 1.6892E+06, 1.7027E+06,
     +  1.7163E+06, 1.7301E+06, 1.7439E+06, 1.7579E+06, 1.7719E+06,
     + 1.7861E+06, 1.8003E+06, 1.8147E+06, 1.8291E+06, 1.8437E+06,
     +  1.8584E+06, 1.8732E+06, 1.8881E+06, 1.9031E+06, 1.9183E+06,
     + 1.9335E+06, 1.9488E+06, 1.9643E+06, 1.9799E+06, 1.9956E+06,
     +  2.0114E+06, 2.0273E+06, 2.0434E+06, 2.0595E+06, 2.0758E+06,
     + 2.0922E+06, 2.1087E+06, 2.1254E+06, 2.1422E+06, 2.1590E+06,
     +  2.1761E+06, 2.1932E+06, 2.2105E+06, 2.2279E+06, 2.2454E+06,
     + 2.2630E+06, 2.2808E+06, 2.2987E+06, 2.3167E+06, 2.3349E+06,
     +  2.3532E+06, 2.3717E+06, 2.3902E+06, 2.4089E+06, 2.4278E+06,
     + 2.4468E+06, 2.4659E+06, 2.4851E+06, 2.5045E+06, 2.5241E+06,
     +  2.5437E+06, 2.5636E+06, 2.5835E+06, 2.6036E+06, 2.6239E+06,
     + 2.6443E+06, 2.6649E+06, 2.6856E+06, 2.7064E+06, 2.7274E+06,
     +  2.7486E+06, 2.7699E+06, 2.7913E+06, 2.8129E+06, 2.8347E+06,
     + 2.8566E+06, 2.8787E+06, 2.9009E+06, 2.9233E+06, 2.9459E+06,
     +  2.9686E+06, 2.9915E+06, 3.0146E+06, 3.0378E+06, 3.0611E+06,
     + 3.0847E+06, 3.1084E+06, 3.1323E+06, 3.1564E+06, 3.1806E+06,
     +  3.2050E+06, 3.2296E+06, 3.2543E+06, 3.2792E+06, 3.3043E+06,
     + 3.3296E+06, 3.3551E+06, 3.3807E+06, 3.4066E+06, 3.4326E+06,
     +  3.4588E+06, 3.4851E+06, 3.5117E+06, 3.5384E+06, 3.5654E+06,
     + 3.5925E+06, 3.6198E+06, 3.6474E+06, 3.6751E+06, 3.7030E+06,
     +  3.7311E+06, 3.7594E+06, 3.7879E+06, 3.8166E+06, 3.8455E+06,
     + 3.8746E+06, 3.9039E+06, 3.9334E+06, 3.9632E+06, 3.9931E+06,
     +  4.0232E+06, 4.0536E+06, 4.0842E+06, 4.1149E+06, 4.1459E+06,
     + 4.1771E+06, 4.2086E+06, 4.2402E+06, 4.2721E+06, 4.3042E+06,
     +  4.3365E+06, 4.3691E+06, 4.4018E+06, 4.4348E+06, 4.4680E+06,
     + 4.5015E+06, 4.5352E+06, 4.5691E+06, 4.6033E+06, 4.6377E+06,
     +  4.6723E+06, 4.7072E+06, 4.7423E+06, 4.7776E+06, 4.8132E+06,
     + 4.8491E+06, 4.8852E+06, 4.9215E+06, 4.9581E+06, 4.9950E+06,
     +  5.0321E+06, 5.0694E+06, 5.1070E+06, 5.1449E+06, 5.1830E+06,
     + 5.2214E+06, 5.2601E+06, 5.2990E+06, 5.3382E+06, 5.3777E+06,
     +  5.4174E+06, 5.4574E+06, 5.4977E+06, 5.5382E+06, 5.5790E+06,
     + 5.6201E+06, 5.6615E+06, 5.7032E+06, 5.7451E+06, 5.7874E+06,
     +  5.8299E+06, 5.8727E+06, 5.9158E+06, 5.9592E+06, 6.0029E+06,
     + 6.0469E+06, 6.0912E+06, 6.1358E+06, 6.1807E+06, 6.2259E+06,
     +  6.2714E+06, 6.3172E+06, 6.3633E+06, 6.4097E+06, 6.4565E+06,
     + 6.5035E+06, 6.5509E+06, 6.5986E+06, 6.6466E+06, 6.6949E+06,
     +  6.7436E+06, 6.7926E+06, 6.8419E+06, 6.8916E+06, 6.9416E+06,
     + 6.9919E+06, 7.0425E+06, 7.0935E+06, 7.1449E+06, 7.1966E+06,
     +  7.2486E+06, 7.3010E+06, 7.3537E+06, 7.4068E+06, 7.4602E+06,
     + 7.5140E+06, 7.5682E+06, 7.6227E+06, 7.6775E+06, 7.7328E+06,
     +  7.7884E+06, 7.8444E+06, 7.9007E+06, 7.9574E+06, 8.0145E+06,
     + 8.0720E+06, 8.1299E+06, 8.1881E+06, 8.2467E+06, 8.3058E+06,
     +  8.3652E+06, 8.4250E+06, 8.4852E+06, 8.5457E+06, 8.6067E+06,
     + 8.6681E+06, 8.7299E+06, 8.7921E+06, 8.8547E+06, 8.9178E+06,
     +  8.9812E+06, 9.0451E+06, 9.1093E+06, 9.1740E+06, 9.2392E+06,
     + 9.3047E+06, 9.3707E+06, 9.4371E+06, 9.5040E+06, 9.5712E+06,
     +  9.6390E+06, 9.7071E+06, 9.7758E+06, 9.8448E+06, 9.9143E+06,
     + 9.9843E+06, 1.0055E+07, 1.0126E+07, 1.0197E+07, 1.0269E+07,
     +  1.0341E+07, 1.0414E+07, 1.0487E+07, 1.0561E+07, 1.0635E+07,
     + 1.0709E+07, 1.0785E+07, 1.0860E+07, 1.0936E+07, 1.1013E+07,
     +  1.1090E+07, 1.1168E+07, 1.1246E+07, 1.1324E+07, 1.1403E+07,
     + 1.1483E+07, 1.1563E+07, 1.1644E+07, 1.1725E+07, 1.1807E+07,
     +  1.1889E+07, 1.1972E+07, 1.2055E+07, 1.2139E+07, 1.2223E+07,
     + 1.2308E+07, 1.2393E+07, 1.2479E+07, 1.2566E+07, 1.2653E+07,
     +  1.2740E+07, 1.2829E+07, 1.2917E+07, 1.3007E+07, 1.3097E+07,
     + 1.3187E+07, 1.3278E+07, 1.3370E+07, 1.3462E+07, 1.3555E+07,
     +  1.3648E+07, 1.3742E+07, 1.3837E+07, 1.3932E+07, 1.4028E+07,
     + 1.4124E+07, 1.4221E+07, 1.4318E+07, 1.4417E+07, 1.4516E+07,
     +  1.4615E+07, 1.4715E+07, 1.4816E+07, 1.4917E+07, 1.5019E+07,
     + 1.5122E+07, 1.5225E+07, 1.5329E+07, 1.5434E+07, 1.5539E+07,
     +  1.5645E+07, 1.5751E+07, 1.5859E+07, 1.5966E+07, 1.6075E+07,
     + 1.6184E+07, 1.6294E+07, 1.6405E+07, 1.6516E+07, 1.6628E+07,
     +  1.6741E+07, 1.6854E+07, 1.6969E+07, 1.7083E+07, 1.7199E+07,
     + 1.7315E+07, 1.7432E+07, 1.7550E+07, 1.7669E+07, 1.7788E+07,
     +  1.7908E+07, 1.8028E+07, 1.8150E+07, 1.8272E+07, 1.8395E+07,
     + 1.8519E+07, 1.8643E+07, 1.8769E+07, 1.8895E+07, 1.9021E+07,
     +  1.9149E+07, 1.9277E+07, 1.9407E+07, 1.9536E+07, 1.9667E+07,
     + 1.9799E+07, 1.9931E+07, 2.0064E+07, 2.0198E+07, 2.0333E+07,
     +  2.0469E+07, 2.0605E+07, 2.0743E+07, 2.0881E+07, 2.1020E+07,
     + 2.1160E+07, 2.1301E+07, 2.1442E+07, 2.1585E+07, 2.1728E+07,
     +  2.1872E+07, 2.2017E+07, 2.2163E+07, 2.2310E+07, 2.2458E+07,
     + 2.2607E+07, 2.2756E+07, 2.2907E+07, 2.3058E+07, 2.3210E+07,
     +  2.3364E+07, 2.3518E+07, 2.3673E+07, 2.3829E+07, 2.3986E+07,
     + 2.4144E+07, 2.4303E+07, 2.4463E+07, 2.4624E+07, 2.4785E+07,
     +  2.4948E+07, 2.5112E+07, 2.5277E+07, 2.5443E+07, 2.5609E+07,
     + 2.5777E+07, 2.5946E+07, 2.6116E+07, 2.6286E+07, 2.6458E+07,
     +  2.6631E+07, 2.6805E+07, 2.6980E+07, 2.7156E+07, 2.7333E+07,
     + 2.7511E+07, 2.7690E+07, 2.7871E+07, 2.8052E+07, 2.8234E+07,
     +  2.8418E+07, 2.8602E+07, 2.8788E+07, 2.8975E+07, 2.9163E+07,
     + 2.9352E+07, 2.9542E+07, 2.9733E+07, 2.9926E+07, 3.0119E+07,
     +  3.0314E+07, 3.0510E+07, 3.0707E+07, 3.0905E+07, 3.1105E+07,
     + 3.1305E+07, 3.1507E+07, 3.1710E+07, 3.1914E+07, 3.2120E+07,
     +  3.2326E+07, 3.2534E+07, 3.2743E+07, 3.2953E+07, 3.3165E+07,
     + 3.3377E+07, 3.3592E+07, 3.3807E+07, 3.4023E+07, 3.4241E+07,
     +  3.4460E+07, 3.4681E+07, 3.4902E+07, 3.5125E+07, 3.5350E+07,
     + 3.5575E+07, 3.5802E+07, 3.6030E+07, 3.6260E+07, 3.6491E+07,
     +  3.6723E+07, 3.6957E+07, 3.7192E+07, 3.7428E+07, 3.7666E+07,
     + 3.7905E+07, 3.8146E+07, 3.8387E+07, 3.8631E+07, 3.8876E+07,
     +  3.9122E+07, 3.9369E+07, 3.9618E+07, 3.9869E+07, 4.0121E+07,
     + 4.0374E+07, 4.0629E+07, 4.0885E+07, 4.1143E+07, 4.1403E+07,
     +  4.1663E+07, 4.1926E+07, 4.2190E+07, 4.2455E+07, 4.2722E+07,
     + 4.2990E+07, 4.3260E+07, 4.3532E+07, 4.3805E+07, 4.4080E+07,
     +  4.4356E+07, 4.4634E+07, 4.4913E+07, 4.5194E+07, 4.5477E+07,
     + 4.5761E+07, 4.6047E+07, 4.6335E+07, 4.6624E+07, 4.6915E+07,
     +  4.7208E+07, 4.7502E+07, 4.7798E+07, 4.8095E+07, 4.8395E+07,
     + 4.8696E+07, 4.8998E+07, 4.9303E+07, 4.9609E+07, 4.9917E+07,
     +  5.0226E+07, 5.0538E+07, 5.0851E+07, 5.1166E+07, 5.1483E+07,
     + 5.1801E+07, 5.2122E+07, 5.2444E+07, 5.2768E+07, 5.3094E+07,
     +  5.3421E+07, 5.3751E+07, 5.4082E+07, 5.4416E+07, 5.4751E+07,
     + 5.5088E+07, 5.5427E+07, 5.5768E+07, 5.6111E+07, 5.6455E+07,
     +  5.6802E+07, 5.7151E+07, 5.7501E+07, 5.7854E+07, 5.8208E+07,
     + 5.8565E+07, 5.8923E+07, 5.9284E+07, 5.9647E+07, 6.0011E+07,
     +  6.0378E+07, 6.0746E+07, 6.1117E+07, 6.1490E+07, 6.1865E+07,
     + 6.2242E+07, 6.2621E+07, 6.3002E+07, 6.3386E+07, 6.3771E+07,
     +  6.4159E+07, 6.4549E+07, 6.4941E+07, 6.5335E+07, 6.5731E+07,
     + 6.6130E+07, 6.6530E+07, 6.6933E+07, 6.7339E+07, 6.7746E+07,
     +  6.8156E+07, 6.8568E+07, 6.8982E+07, 6.9399E+07, 6.9818E+07,
     + 7.0239E+07, 7.0662E+07, 7.1088E+07, 7.1517E+07, 7.1947E+07,
     +  7.2380E+07, 7.2815E+07, 7.3253E+07, 7.3693E+07, 7.4136E+07,
     + 7.4581E+07, 7.5028E+07, 7.5478E+07, 7.5931E+07, 7.6386E+07,
     +  7.6843E+07, 7.7303E+07, 7.7765E+07, 7.8230E+07, 7.8698E+07,
     + 7.9168E+07, 7.9640E+07, 8.0115E+07, 8.0593E+07, 8.1074E+07,
     +  8.1557E+07, 8.2042E+07, 8.2531E+07, 8.3021E+07, 8.3515E+07,
     + 8.4011E+07, 8.4510E+07, 8.5012E+07, 8.5517E+07, 8.6024E+07,
     +  8.6534E+07, 8.7047E+07, 8.7562E+07, 8.8080E+07, 8.8601E+07,
     + 8.9125E+07, 8.9652E+07, 9.0182E+07, 9.0714E+07, 9.1250E+07,
     +  9.1788E+07, 9.2329E+07, 9.2873E+07, 9.3420E+07, 9.3970E+07,
     + 9.4523E+07, 9.5079E+07, 9.5638E+07, 9.6200E+07, 9.6765E+07,
     +  9.7333E+07, 9.7904E+07, 9.8478E+07, 9.9055E+07, 9.9635E+07,
     + 1.0022E+08, 1.0080E+08, 1.0139E+08, 1.0199E+08, 1.0258E+08,
     +  1.0318E+08, 1.0378E+08, 1.0439E+08, 1.0500E+08, 1.0561E+08,
     + 1.0623E+08, 1.0685E+08, 1.0747E+08, 1.0809E+08, 1.0872E+08,
     +  1.0935E+08, 1.0999E+08, 1.1063E+08, 1.1127E+08, 1.1191E+08,
     + 1.1256E+08, 1.1321E+08, 1.1387E+08, 1.1453E+08, 1.1519E+08,
     +  1.1586E+08, 1.1653E+08, 1.1720E+08, 1.1788E+08, 1.1856E+08,
     + 1.1924E+08, 1.1993E+08, 1.2062E+08, 1.2132E+08, 1.2202E+08,
     +  1.2272E+08, 1.2342E+08, 1.2413E+08, 1.2485E+08, 1.2556E+08,
     + 1.2628E+08, 1.2701E+08, 1.2774E+08, 1.2847E+08, 1.2921E+08,
     +  1.2995E+08, 1.3069E+08, 1.3144E+08, 1.3219E+08, 1.3294E+08,
     + 1.3370E+08, 1.3447E+08, 1.3524E+08, 1.3601E+08, 1.3678E+08,
     +  1.3756E+08, 1.3835E+08, 1.3913E+08, 1.3992E+08, 1.4072E+08,
     + 1.4152E+08, 1.4233E+08, 1.4313E+08, 1.4395E+08, 1.4476E+08,
     +  1.4558E+08, 1.4641E+08, 1.4724E+08, 1.4807E+08, 1.4891E+08,
     + 1.4975E+08, 1.5060E+08, 1.5145E+08, 1.5231E+08, 1.5317E+08,
     +  1.5403E+08, 1.5490E+08, 1.5577E+08, 1.5665E+08, 1.5754E+08,
     + 1.5842E+08, 1.5931E+08, 1.6021E+08, 1.6111E+08, 1.6202E+08,
     +  1.6293E+08, 1.6384E+08, 1.6476E+08, 1.6568E+08, 1.6661E+08,
     + 1.6755E+08, 1.6849E+08, 1.6943E+08, 1.7038E+08, 1.7133E+08,
     +  1.7229E+08, 1.7325E+08, 1.7422E+08, 1.7519E+08, 1.7617E+08,
     + 1.7715E+08, 1.7814E+08, 1.7913E+08, 1.8013E+08, 1.8113E+08,
     +  1.8214E+08, 1.8315E+08, 1.8417E+08, 1.8519E+08, 1.8622E+08,
     + 1.8725E+08, 1.8829E+08, 1.8934E+08, 1.9039E+08, 1.9144E+08,
     +  1.9250E+08, 1.9357E+08, 1.9464E+08, 1.9571E+08, 1.9680E+08,
     + 1.9788E+08, 1.9898E+08, 2.0007E+08, 2.0118E+08, 2.0229E+08,
     +  2.0340E+08, 2.0452E+08, 2.0565E+08, 2.0678E+08, 2.0792E+08,
     + 2.0906E+08, 2.1021E+08, 2.1136E+08, 2.1252E+08, 2.1369E+08,
     +  2.1486E+08, 2.1604E+08, 2.1722E+08, 2.1841E+08, 2.1961E+08,
     + 2.2081E+08, 2.2202E+08, 2.2323E+08, 2.2445E+08, 2.2568E+08,
     +  2.2691E+08, 2.2815E+08, 2.2939E+08, 2.3064E+08, 2.3190E+08,
     + 2.3316E+08, 2.3443E+08, 2.3571E+08, 2.3699E+08, 2.3828E+08,
     +  2.3957E+08, 2.4087E+08, 2.4218E+08, 2.4350E+08, 2.4482E+08,
     + 2.4614E+08, 2.4748E+08, 2.4882E+08, 2.5016E+08, 2.5152E+08,
     +  2.5288E+08, 2.5424E+08, 2.5562E+08, 2.5700E+08, 2.5839E+08,
     + 2.5978E+08, 2.6118E+08, 2.6259E+08, 2.6400E+08, 2.6543E+08,
     +  2.6685E+08, 2.6829E+08, 2.6973E+08, 2.7118E+08, 2.7264E+08,
     + 2.7410E+08, 2.7558E+08, 2.7705E+08, 2.7854E+08, 2.8003E+08,
     +  2.8153E+08, 2.8304E+08, 2.8456E+08, 2.8608E+08, 2.8761E+08,
     + 2.8915E+08, 2.9069E+08, 2.9224E+08, 2.9380E+08, 2.9537E+08,
     +  2.9695E+08, 2.9853E+08, 3.0012E+08, 3.0172E+08, 3.0332E+08,
     + 3.0494E+08, 3.0656E+08, 3.0819E+08, 3.0983E+08, 3.1147E+08,
     +  3.1313E+08, 3.1479E+08, 3.1646E+08, 3.1813E+08, 3.1982E+08,
     + 3.2151E+08, 3.2322E+08, 3.2493E+08, 3.2664E+08, 3.2837E+08,
     +  3.3011E+08, 3.3185E+08, 3.3360E+08, 3.3536E+08, 3.3713E+08,
     + 3.3891E+08, 3.4069E+08, 3.4249E+08, 3.4429E+08, 3.4610E+08,
     +  3.4792E+08, 3.4975E+08, 3.5159E+08, 3.5344E+08, 3.5529E+08,
     + 3.5716E+08, 3.5903E+08, 3.6091E+08, 3.6280E+08, 3.6471E+08,
     +  3.6661E+08, 3.6853E+08, 3.7046E+08, 3.7240E+08, 3.7434E+08,
     + 3.7630E+08, 3.7826E+08, 3.8024E+08, 3.8222E+08, 3.8422E+08,
     +  3.8622E+08, 3.8823E+08, 3.9025E+08, 3.9228E+08, 3.9432E+08,
     + 3.9637E+08, 3.9843E+08, 4.0050E+08, 4.0258E+08, 4.0467E+08,
     +  4.0677E+08, 4.0888E+08, 4.1100E+08, 4.1313E+08, 4.1527E+08,
     + 4.1742E+08, 4.1958E+08, 4.2175E+08, 4.2393E+08, 4.2612E+08,
     +  4.2832E+08, 4.3053E+08, 4.3276E+08, 4.3499E+08, 4.3723E+08,
     + 4.3948E+08, 4.4175E+08, 4.4402E+08, 4.4631E+08, 4.4860E+08,
     +  4.5091E+08, 4.5323E+08, 4.5555E+08, 4.5789E+08, 4.6024E+08,
     + 4.6261E+08, 4.6498E+08, 4.6736E+08, 4.6975E+08, 4.7216E+08,
     +  4.7458E+08, 4.7701E+08, 4.7944E+08, 4.8190E+08, 4.8436E+08,
     + 4.8683E+08, 4.8932E+08, 4.9181E+08, 4.9432E+08, 4.9684E+08,
     +  4.9937E+08, 5.0192E+08, 5.0447E+08, 5.0704E+08, 5.0962E+08,
     + 5.1221E+08, 5.1481E+08, 5.1743E+08, 5.2005E+08, 5.2269E+08,
     +  5.2535E+08, 5.2801E+08, 5.3069E+08, 5.3337E+08, 5.3608E+08,
     + 5.3879E+08, 5.4151E+08, 5.4425E+08, 5.4700E+08, 5.4977E+08,
     +  5.5254E+08, 5.5533E+08, 5.5814E+08, 5.6095E+08, 5.6378E+08,
     + 5.6662E+08, 5.6947E+08, 5.7234E+08, 5.7522E+08, 5.7812E+08,
     +  5.8102E+08, 5.8394E+08, 5.8688E+08, 5.8982E+08, 5.9278E+08,
     + 5.9576E+08, 5.9875E+08, 6.0175E+08, 6.0476E+08, 6.0779E+08,
     +  6.1083E+08, 6.1389E+08, 6.1696E+08, 6.2004E+08, 6.2314E+08,
     + 6.2625E+08, 6.2938E+08, 6.3252E+08, 6.3568E+08, 6.3885E+08,
     +  6.4203E+08, 6.4523E+08, 6.4844E+08, 6.5167E+08, 6.5491E+08,
     + 6.5817E+08, 6.6144E+08, 6.6473E+08, 6.6803E+08, 6.7134E+08,
     +  6.7468E+08, 6.7802E+08, 6.8138E+08, 6.8476E+08, 6.8815E+08,
     + 6.9156E+08, 6.9498E+08, 6.9842E+08, 7.0187E+08, 7.0534E+08,
     +  7.0883E+08, 7.1233E+08, 7.1584E+08, 7.1938E+08, 7.2292E+08,
     + 7.2649E+08, 7.3007E+08, 7.3366E+08, 7.3728E+08, 7.4090E+08,
     +  7.4455E+08, 7.4821E+08, 7.5189E+08, 7.5558E+08, 7.5929E+08,
     + 7.6302E+08, 7.6676E+08, 7.7052E+08, 7.7430E+08, 7.7809E+08,
     +  7.8190E+08, 7.8573E+08, 7.8957E+08, 7.9344E+08, 7.9732E+08,
     + 8.0121E+08, 8.0513E+08, 8.0906E+08, 8.1301E+08, 8.1697E+08,
     +  8.2096E+08, 8.2496E+08, 8.2898E+08, 8.3302E+08, 8.3707E+08,
     + 8.4114E+08, 8.4524E+08, 8.4934E+08, 8.5347E+08, 8.5762E+08,
     +  8.6178E+08, 8.6596E+08, 8.7017E+08, 8.7439E+08, 8.7862E+08,
     + 8.8288E+08, 8.8716E+08, 8.9145E+08, 8.9576E+08, 9.0010E+08,
     +  9.0445E+08, 9.0882E+08, 9.1321E+08, 9.1762E+08, 9.2205E+08,
     + 9.2650E+08, 9.3096E+08, 9.3545E+08, 9.3996E+08, 9.4449E+08,
     +  9.4903E+08, 9.5360E+08, 9.5819E+08, 9.6279E+08, 9.6742E+08,
     + 9.7207E+08, 9.7674E+08, 9.8142E+08, 9.8613E+08, 9.9086E+08,
     +  9.9561E+08, 1.0004E+09, 1.0052E+09, 1.0100E+09, 1.0148E+09,
     + 1.0197E+09, 1.0245E+09, 1.0294E+09, 1.0344E+09, 1.0393E+09,
     +  1.0443E+09, 1.0492E+09, 1.0542E+09, 1.0593E+09, 1.0643E+09,
     + 1.0694E+09, 1.0745E+09, 1.0796E+09, 1.0847E+09, 1.0899E+09,
     +  1.0951E+09, 1.1003E+09, 1.1055E+09, 1.1107E+09, 1.1160E+09,
     + 1.1213E+09, 1.1266E+09, 1.1320E+09, 1.1373E+09, 1.1427E+09,
     +  1.1481E+09, 1.1536E+09, 1.1590E+09, 1.1645E+09, 1.1700E+09,
     + 1.1755E+09, 1.1811E+09, 1.1866E+09, 1.1922E+09, 1.1979E+09,
     +  1.2035E+09, 1.2092E+09, 1.2149E+09, 1.2206E+09, 1.2263E+09,
     + 1.2321E+09, 1.2379E+09, 1.2437E+09, 1.2495E+09, 1.2554E+09,
     +  1.2613E+09, 1.2672E+09, 1.2732E+09, 1.2791E+09, 1.2851E+09,
     + 1.2911E+09, 1.2972E+09, 1.3033E+09, 1.3094E+09, 1.3155E+09,
     +  1.3216E+09, 1.3278E+09, 1.3340E+09, 1.3402E+09, 1.3465E+09,
     + 1.3528E+09, 1.3591E+09, 1.3654E+09, 1.3718E+09, 1.3782E+09,
     +  1.3846E+09, 1.3910E+09, 1.3975E+09, 1.4040E+09, 1.4105E+09,
     + 1.4171E+09, 1.4236E+09, 1.4302E+09, 1.4369E+09, 1.4435E+09,
     +  1.4502E+09, 1.4569E+09, 1.4637E+09, 1.4705E+09, 1.4773E+09,
     + 1.4841E+09, 1.4910E+09, 1.4979E+09, 1.5048E+09, 1.5117E+09,
     +  1.5187E+09, 1.5257E+09, 1.5328E+09, 1.5398E+09, 1.5469E+09,
     + 1.5540E+09, 1.5612E+09, 1.5684E+09, 1.5756E+09, 1.5828E+09,
     +  1.5901E+09, 1.5974E+09, 1.6048E+09, 1.6121E+09, 1.6195E+09,
     + 1.6270E+09, 1.6344E+09, 1.6419E+09, 1.6494E+09, 1.6570E+09,
     +  1.6646E+09, 1.6722E+09, 1.6799E+09, 1.6875E+09, 1.6952E+09,
     + 1.7030E+09, 1.7108E+09, 1.7186E+09, 1.7264E+09, 1.7343E+09,
     +  1.7422E+09, 1.7501E+09, 1.7581E+09, 1.7661E+09, 1.7742E+09,
     + 1.7822E+09, 1.7903E+09, 1.7985E+09, 1.8067E+09, 1.8149E+09,
     +  1.8231E+09, 1.8314E+09, 1.8397E+09, 1.8480E+09, 1.8564E+09,
     + 1.8648E+09, 1.8733E+09, 1.8818E+09, 1.8903E+09, 1.8988E+09,
     +  1.9074E+09, 1.9160E+09, 1.9247E+09, 1.9334E+09, 1.9421E+09,
     + 1.9509E+09, 1.9597E+09, 1.9685E+09, 1.9774E+09, 1.9863E+09,
     +  1.9953E+09, 2.0042E+09, 2.0133E+09, 2.0223E+09, 2.0314E+09,
     + 2.0405E+09, 2.0497E+09, 2.0589E+09, 2.0682E+09, 2.0774E+09,
     +  2.0868E+09, 2.0961E+09, 2.1055E+09, 2.1150E+09, 2.1244E+09,
     + 2.1339E+09, 2.1435E+09, 2.1531E+09, 2.1627E+09, 2.1724E+09,
     +  2.1821E+09, 2.1918E+09, 2.2016E+09, 2.2114E+09, 2.2213E+09,
     + 2.2312E+09, 2.2412E+09, 2.2511E+09, 2.2612E+09, 2.2712E+09,
     +  2.2813E+09, 2.2915E+09, 2.3017E+09, 2.3119E+09, 2.3222E+09,
     + 2.3325E+09, 2.3429E+09, 2.3533E+09, 2.3637E+09, 2.3742E+09,
     +  2.3847E+09, 2.3953E+09, 2.4059E+09, 2.4165E+09, 2.4272E+09,
     + 2.4380E+09, 2.4487E+09, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       217
      data (QofT( 2,J),J=1,2408)/ 2.3206E+01, 5.2766E+01, 8.6290E+01,
     +  1.2294E+02, 1.6245E+02, 2.0460E+02, 2.4923E+02, 2.9620E+02,
     + 3.4538E+02, 3.9671E+02, 4.5012E+02, 5.0558E+02, 5.6307E+02,
     +  6.2256E+02, 6.8404E+02, 7.4751E+02, 8.1293E+02, 8.8030E+02,
     + 9.4960E+02, 1.0208E+03, 1.1688E+03, 1.3241E+03, 1.4866E+03,
     +  1.6558E+03, 1.8316E+03, 2.0138E+03, 2.2022E+03, 2.3965E+03,
     + 2.5965E+03, 2.8021E+03, 3.0130E+03, 3.2293E+03, 3.4506E+03,
     +  3.6769E+03, 3.9080E+03, 4.1438E+03, 4.3843E+03, 4.6294E+03,
     + 4.8789E+03, 5.1327E+03, 5.3909E+03, 5.6532E+03, 5.9197E+03,
     +  6.1903E+03, 6.4649E+03, 6.7435E+03, 7.0259E+03, 7.3122E+03,
     + 7.6023E+03, 7.8962E+03, 8.1937E+03, 8.4950E+03, 8.7998E+03,
     +  9.1082E+03, 9.4201E+03, 9.7355E+03, 1.0054E+04, 1.0377E+04,
     + 1.0702E+04, 1.1031E+04, 1.1364E+04, 1.1699E+04, 1.2038E+04,
     +  1.2381E+04, 1.2726E+04, 1.3075E+04, 1.3426E+04, 1.3781E+04,
     + 1.4139E+04, 1.4500E+04, 1.4865E+04, 1.5232E+04, 1.5602E+04,
     +  1.5975E+04, 1.6352E+04, 1.6731E+04, 1.7114E+04, 1.7499E+04,
     + 1.7887E+04, 1.8279E+04, 1.8673E+04, 1.9070E+04, 1.9470E+04,
     +  1.9873E+04, 2.0279E+04, 2.0688E+04, 2.1100E+04, 2.1515E+04,
     + 2.1933E+04, 2.2354E+04, 2.2777E+04, 2.3204E+04, 2.3634E+04,
     +  2.4066E+04, 2.4502E+04, 2.4940E+04, 2.5381E+04, 2.5826E+04,
     + 2.6273E+04, 2.6724E+04, 2.7177E+04, 2.7634E+04, 2.8093E+04,
     +  2.8556E+04, 2.9022E+04, 2.9491E+04, 2.9963E+04, 3.0438E+04,
     + 3.0916E+04, 3.1397E+04, 3.1882E+04, 3.2370E+04, 3.2861E+04,
     +  3.3356E+04, 3.3853E+04, 3.4354E+04, 3.4859E+04, 3.5366E+04,
     + 3.5877E+04, 3.6392E+04, 3.6910E+04, 3.7432E+04, 3.7957E+04,
     +  3.8485E+04, 3.9017E+04, 3.9553E+04, 4.0092E+04, 4.0636E+04,
     + 4.1182E+04, 4.1733E+04, 4.2287E+04, 4.2846E+04, 4.3408E+04,
     +  4.3974E+04, 4.4543E+04, 4.5117E+04, 4.5695E+04, 4.6277E+04,
     + 4.6863E+04, 4.7453E+04, 4.8048E+04, 4.8646E+04, 4.9249E+04,
     +  4.9856E+04, 5.0468E+04, 5.1083E+04, 5.1704E+04, 5.2329E+04,
     + 5.2958E+04, 5.3592E+04, 5.4230E+04, 5.4874E+04, 5.5522E+04,
     +  5.6174E+04, 5.6832E+04, 5.7494E+04, 5.8162E+04, 5.8834E+04,
     + 5.9511E+04, 6.0194E+04, 6.0881E+04, 6.1574E+04, 6.2272E+04,
     +  6.2975E+04, 6.3684E+04, 6.4398E+04, 6.5117E+04, 6.5842E+04,
     + 6.6573E+04, 6.7309E+04, 6.8051E+04, 6.8799E+04, 6.9552E+04,
     +  7.0311E+04, 7.1076E+04, 7.1848E+04, 7.2625E+04, 7.3408E+04,
     + 7.4197E+04, 7.4993E+04, 7.5795E+04, 7.6603E+04, 7.7418E+04,
     +  7.8239E+04, 7.9067E+04, 7.9901E+04, 8.0742E+04, 8.1590E+04,
     + 8.2445E+04, 8.3306E+04, 8.4174E+04, 8.5050E+04, 8.5932E+04,
     +  8.6822E+04, 8.7718E+04, 8.8623E+04, 8.9534E+04, 9.0453E+04,
     + 9.1379E+04, 9.2313E+04, 9.3254E+04, 9.4204E+04, 9.5161E+04,
     +  9.6125E+04, 9.7098E+04, 9.8079E+04, 9.9068E+04, 1.0006E+05,
     + 1.0107E+05, 1.0208E+05, 1.0311E+05, 1.0414E+05, 1.0517E+05,
     +  1.0622E+05, 1.0728E+05, 1.0834E+05, 1.0942E+05, 1.1050E+05,
     + 1.1159E+05, 1.1269E+05, 1.1380E+05, 1.1492E+05, 1.1605E+05,
     +  1.1719E+05, 1.1834E+05, 1.1950E+05, 1.2066E+05, 1.2184E+05,
     + 1.2303E+05, 1.2422E+05, 1.2543E+05, 1.2665E+05, 1.2787E+05,
     +  1.2911E+05, 1.3036E+05, 1.3162E+05, 1.3289E+05, 1.3417E+05,
     + 1.3546E+05, 1.3676E+05, 1.3807E+05, 1.3940E+05, 1.4073E+05,
     +  1.4208E+05, 1.4343E+05, 1.4480E+05, 1.4618E+05, 1.4758E+05,
     + 1.4898E+05, 1.5040E+05, 1.5182E+05, 1.5326E+05, 1.5472E+05,
     +  1.5618E+05, 1.5766E+05, 1.5915E+05, 1.6065E+05, 1.6216E+05,
     + 1.6369E+05, 1.6523E+05, 1.6678E+05, 1.6835E+05, 1.6993E+05,
     +  1.7152E+05, 1.7313E+05, 1.7475E+05, 1.7639E+05, 1.7803E+05,
     + 1.7970E+05, 1.8137E+05, 1.8306E+05, 1.8477E+05, 1.8649E+05,
     +  1.8822E+05, 1.8997E+05, 1.9173E+05, 1.9351E+05, 1.9530E+05,
     + 1.9711E+05, 1.9894E+05, 2.0078E+05, 2.0263E+05, 2.0450E+05,
     +  2.0639E+05, 2.0829E+05, 2.1021E+05, 2.1214E+05, 2.1409E+05,
     + 2.1606E+05, 2.1804E+05, 2.2004E+05, 2.2206E+05, 2.2410E+05,
     +  2.2615E+05, 2.2822E+05, 2.3031E+05, 2.3241E+05, 2.3453E+05,
     + 2.3667E+05, 2.3883E+05, 2.4101E+05, 2.4320E+05, 2.4541E+05,
     +  2.4765E+05, 2.4990E+05, 2.5217E+05, 2.5445E+05, 2.5676E+05,
     + 2.5909E+05, 2.6144E+05, 2.6380E+05, 2.6619E+05, 2.6859E+05,
     +  2.7102E+05, 2.7347E+05, 2.7594E+05, 2.7842E+05, 2.8093E+05,
     + 2.8346E+05, 2.8601E+05, 2.8859E+05, 2.9118E+05, 2.9380E+05,
     +  2.9643E+05, 2.9909E+05, 3.0178E+05, 3.0448E+05, 3.0721E+05,
     + 3.0996E+05, 3.1273E+05, 3.1552E+05, 3.1834E+05, 3.2119E+05,
     +  3.2405E+05, 3.2694E+05, 3.2986E+05, 3.3279E+05, 3.3576E+05,
     + 3.3874E+05, 3.4176E+05, 3.4479E+05, 3.4786E+05, 3.5094E+05,
     +  3.5406E+05, 3.5720E+05, 3.6036E+05, 3.6355E+05, 3.6677E+05,
     + 3.7002E+05, 3.7329E+05, 3.7658E+05, 3.7991E+05, 3.8326E+05,
     +  3.8664E+05, 3.9005E+05, 3.9349E+05, 3.9695E+05, 4.0045E+05,
     + 4.0397E+05, 4.0752E+05, 4.1110E+05, 4.1471E+05, 4.1835E+05,
     +  4.2202E+05, 4.2572E+05, 4.2945E+05, 4.3321E+05, 4.3700E+05,
     + 4.4082E+05, 4.4468E+05, 4.4856E+05, 4.5248E+05, 4.5643E+05,
     +  4.6041E+05, 4.6442E+05, 4.6847E+05, 4.7255E+05, 4.7666E+05,
     + 4.8081E+05, 4.8499E+05, 4.8920E+05, 4.9345E+05, 4.9774E+05,
     +  5.0205E+05, 5.0641E+05, 5.1079E+05, 5.1522E+05, 5.1968E+05,
     + 5.2417E+05, 5.2871E+05, 5.3328E+05, 5.3788E+05, 5.4253E+05,
     +  5.4721E+05, 5.5193E+05, 5.5668E+05, 5.6148E+05, 5.6631E+05,
     + 5.7119E+05, 5.7610E+05, 5.8105E+05, 5.8604E+05, 5.9107E+05,
     +  5.9615E+05, 6.0126E+05, 6.0641E+05, 6.1161E+05, 6.1685E+05,
     + 6.2213E+05, 6.2745E+05, 6.3281E+05, 6.3822E+05, 6.4367E+05,
     +  6.4917E+05, 6.5471E+05, 6.6029E+05, 6.6592E+05, 6.7159E+05,
     + 6.7730E+05, 6.8307E+05, 6.8888E+05, 6.9473E+05, 7.0063E+05,
     +  7.0658E+05, 7.1258E+05, 7.1862E+05, 7.2471E+05, 7.3085E+05,
     + 7.3704E+05, 7.4328E+05, 7.4957E+05, 7.5591E+05, 7.6229E+05,
     +  7.6873E+05, 7.7522E+05, 7.8176E+05, 7.8835E+05, 7.9500E+05,
     + 8.0169E+05, 8.0844E+05, 8.1525E+05, 8.2210E+05, 8.2901E+05,
     +  8.3598E+05, 8.4300E+05, 8.5007E+05, 8.5720E+05, 8.6439E+05,
     + 8.7163E+05, 8.7893E+05, 8.8628E+05, 8.9370E+05, 9.0117E+05,
     +  9.0870E+05, 9.1629E+05, 9.2394E+05, 9.3165E+05, 9.3942E+05,
     + 9.4725E+05, 9.5514E+05, 9.6309E+05, 9.7110E+05, 9.7918E+05,
     +  9.8732E+05, 9.9552E+05, 1.0038E+06, 1.0121E+06, 1.0205E+06,
     + 1.0290E+06, 1.0375E+06, 1.0461E+06, 1.0547E+06, 1.0635E+06,
     +  1.0723E+06, 1.0811E+06, 1.0901E+06, 1.0991E+06, 1.1081E+06,
     + 1.1173E+06, 1.1265E+06, 1.1357E+06, 1.1451E+06, 1.1545E+06,
     +  1.1640E+06, 1.1736E+06, 1.1832E+06, 1.1929E+06, 1.2027E+06,
     + 1.2126E+06, 1.2225E+06, 1.2326E+06, 1.2427E+06, 1.2528E+06,
     +  1.2631E+06, 1.2734E+06, 1.2838E+06, 1.2943E+06, 1.3049E+06,
     + 1.3155E+06, 1.3262E+06, 1.3370E+06, 1.3479E+06, 1.3589E+06,
     +  1.3700E+06, 1.3811E+06, 1.3923E+06, 1.4036E+06, 1.4150E+06,
     + 1.4265E+06, 1.4381E+06, 1.4498E+06, 1.4615E+06, 1.4733E+06,
     +  1.4853E+06, 1.4973E+06, 1.5094E+06, 1.5216E+06, 1.5339E+06,
     + 1.5463E+06, 1.5587E+06, 1.5713E+06, 1.5840E+06, 1.5967E+06,
     +  1.6096E+06, 1.6225E+06, 1.6356E+06, 1.6487E+06, 1.6620E+06,
     + 1.6753E+06, 1.6888E+06, 1.7023E+06, 1.7160E+06, 1.7297E+06,
     +  1.7436E+06, 1.7575E+06, 1.7716E+06, 1.7858E+06, 1.8000E+06,
     + 1.8144E+06, 1.8289E+06, 1.8435E+06, 1.8582E+06, 1.8730E+06,
     +  1.8879E+06, 1.9029E+06, 1.9181E+06, 1.9333E+06, 1.9487E+06,
     + 1.9642E+06, 1.9798E+06, 1.9955E+06, 2.0113E+06, 2.0273E+06,
     +  2.0433E+06, 2.0595E+06, 2.0758E+06, 2.0922E+06, 2.1088E+06,
     + 2.1254E+06, 2.1422E+06, 2.1591E+06, 2.1762E+06, 2.1933E+06,
     +  2.2106E+06, 2.2280E+06, 2.2456E+06, 2.2632E+06, 2.2810E+06,
     + 2.2989E+06, 2.3170E+06, 2.3352E+06, 2.3535E+06, 2.3720E+06,
     +  2.3906E+06, 2.4093E+06, 2.4282E+06, 2.4472E+06, 2.4663E+06,
     + 2.4856E+06, 2.5050E+06, 2.5246E+06, 2.5443E+06, 2.5641E+06,
     +  2.5841E+06, 2.6043E+06, 2.6245E+06, 2.6450E+06, 2.6656E+06,
     + 2.6863E+06, 2.7072E+06, 2.7282E+06, 2.7494E+06, 2.7707E+06,
     +  2.7922E+06, 2.8138E+06, 2.8356E+06, 2.8576E+06, 2.8797E+06,
     + 2.9020E+06, 2.9244E+06, 2.9470E+06, 2.9697E+06, 2.9927E+06,
     +  3.0157E+06, 3.0390E+06, 3.0624E+06, 3.0860E+06, 3.1097E+06,
     + 3.1337E+06, 3.1578E+06, 3.1820E+06, 3.2065E+06, 3.2311E+06,
     +  3.2559E+06, 3.2808E+06, 3.3060E+06, 3.3313E+06, 3.3568E+06,
     + 3.3825E+06, 3.4083E+06, 3.4344E+06, 3.4606E+06, 3.4871E+06,
     +  3.5137E+06, 3.5405E+06, 3.5674E+06, 3.5946E+06, 3.6220E+06,
     + 3.6496E+06, 3.6773E+06, 3.7053E+06, 3.7334E+06, 3.7618E+06,
     +  3.7903E+06, 3.8191E+06, 3.8480E+06, 3.8772E+06, 3.9066E+06,
     + 3.9361E+06, 3.9659E+06, 3.9959E+06, 4.0261E+06, 4.0565E+06,
     +  4.0871E+06, 4.1180E+06, 4.1490E+06, 4.1803E+06, 4.2118E+06,
     + 4.2435E+06, 4.2754E+06, 4.3076E+06, 4.3399E+06, 4.3725E+06,
     +  4.4054E+06, 4.4384E+06, 4.4717E+06, 4.5052E+06, 4.5390E+06,
     + 4.5730E+06, 4.6072E+06, 4.6417E+06, 4.6764E+06, 4.7113E+06,
     +  4.7465E+06, 4.7819E+06, 4.8176E+06, 4.8535E+06, 4.8897E+06,
     + 4.9261E+06, 4.9627E+06, 4.9997E+06, 5.0368E+06, 5.0743E+06,
     +  5.1120E+06, 5.1499E+06, 5.1881E+06, 5.2266E+06, 5.2653E+06,
     + 5.3043E+06, 5.3436E+06, 5.3832E+06, 5.4230E+06, 5.4630E+06,
     +  5.5034E+06, 5.5440E+06, 5.5850E+06, 5.6261E+06, 5.6676E+06,
     + 5.7094E+06, 5.7514E+06, 5.7938E+06, 5.8364E+06, 5.8793E+06,
     +  5.9225E+06, 5.9660E+06, 6.0098E+06, 6.0538E+06, 6.0982E+06,
     + 6.1429E+06, 6.1879E+06, 6.2332E+06, 6.2788E+06, 6.3247E+06,
     +  6.3709E+06, 6.4175E+06, 6.4643E+06, 6.5115E+06, 6.5590E+06,
     + 6.6068E+06, 6.6549E+06, 6.7034E+06, 6.7522E+06, 6.8013E+06,
     +  6.8507E+06, 6.9005E+06, 6.9506E+06, 7.0010E+06, 7.0518E+06,
     + 7.1029E+06, 7.1544E+06, 7.2062E+06, 7.2584E+06, 7.3109E+06,
     +  7.3637E+06, 7.4169E+06, 7.4705E+06, 7.5244E+06, 7.5787E+06,
     + 7.6333E+06, 7.6884E+06, 7.7437E+06, 7.7995E+06, 7.8556E+06,
     +  7.9121E+06, 7.9689E+06, 8.0262E+06, 8.0838E+06, 8.1418E+06,
     + 8.2002E+06, 8.2590E+06, 8.3182E+06, 8.3777E+06, 8.4377E+06,
     +  8.4980E+06, 8.5588E+06, 8.6199E+06, 8.6815E+06, 8.7434E+06,
     + 8.8058E+06, 8.8686E+06, 8.9318E+06, 8.9954E+06, 9.0594E+06,
     +  9.1239E+06, 9.1887E+06, 9.2540E+06, 9.3198E+06, 9.3859E+06,
     + 9.4525E+06, 9.5195E+06, 9.5870E+06, 9.6549E+06, 9.7233E+06,
     +  9.7921E+06, 9.8613E+06, 9.9310E+06, 1.0001E+07, 1.0072E+07,
     + 1.0143E+07, 1.0214E+07, 1.0286E+07, 1.0359E+07, 1.0432E+07,
     +  1.0505E+07, 1.0579E+07, 1.0654E+07, 1.0728E+07, 1.0804E+07,
     + 1.0880E+07, 1.0956E+07, 1.1033E+07, 1.1110E+07, 1.1188E+07,
     +  1.1266E+07, 1.1345E+07, 1.1424E+07, 1.1504E+07, 1.1585E+07,
     + 1.1665E+07, 1.1747E+07, 1.1829E+07, 1.1911E+07, 1.1994E+07,
     +  1.2078E+07, 1.2162E+07, 1.2246E+07, 1.2331E+07, 1.2417E+07,
     + 1.2503E+07, 1.2590E+07, 1.2677E+07, 1.2765E+07, 1.2854E+07,
     +  1.2943E+07, 1.3032E+07, 1.3123E+07, 1.3213E+07, 1.3305E+07,
     + 1.3397E+07, 1.3489E+07, 1.3582E+07, 1.3676E+07, 1.3770E+07,
     +  1.3865E+07, 1.3960E+07, 1.4056E+07, 1.4153E+07, 1.4250E+07,
     + 1.4348E+07, 1.4447E+07, 1.4546E+07, 1.4646E+07, 1.4746E+07,
     +  1.4847E+07, 1.4949E+07, 1.5051E+07, 1.5154E+07, 1.5258E+07,
     + 1.5362E+07, 1.5467E+07, 1.5573E+07, 1.5679E+07, 1.5786E+07,
     +  1.5893E+07, 1.6002E+07, 1.6111E+07, 1.6220E+07, 1.6330E+07,
     + 1.6441E+07, 1.6553E+07, 1.6666E+07, 1.6779E+07, 1.6892E+07,
     +  1.7007E+07, 1.7122E+07, 1.7238E+07, 1.7355E+07, 1.7472E+07,
     + 1.7590E+07, 1.7709E+07, 1.7829E+07, 1.7949E+07, 1.8070E+07,
     +  1.8192E+07, 1.8315E+07, 1.8438E+07, 1.8562E+07, 1.8687E+07,
     + 1.8813E+07, 1.8939E+07, 1.9067E+07, 1.9195E+07, 1.9324E+07,
     +  1.9453E+07, 1.9584E+07, 1.9715E+07, 1.9847E+07, 1.9980E+07,
     + 2.0113E+07, 2.0248E+07, 2.0383E+07, 2.0519E+07, 2.0656E+07,
     +  2.0794E+07, 2.0933E+07, 2.1072E+07, 2.1213E+07, 2.1354E+07,
     + 2.1496E+07, 2.1639E+07, 2.1783E+07, 2.1928E+07, 2.2073E+07,
     +  2.2220E+07, 2.2367E+07, 2.2516E+07, 2.2665E+07, 2.2815E+07,
     + 2.2966E+07, 2.3118E+07, 2.3271E+07, 2.3425E+07, 2.3579E+07,
     +  2.3735E+07, 2.3892E+07, 2.4049E+07, 2.4208E+07, 2.4367E+07,
     + 2.4528E+07, 2.4689E+07, 2.4852E+07, 2.5015E+07, 2.5179E+07,
     +  2.5345E+07, 2.5511E+07, 2.5678E+07, 2.5847E+07, 2.6016E+07,
     + 2.6187E+07, 2.6358E+07, 2.6531E+07, 2.6704E+07, 2.6879E+07,
     +  2.7054E+07, 2.7231E+07, 2.7409E+07, 2.7587E+07, 2.7767E+07,
     + 2.7948E+07, 2.8130E+07, 2.8313E+07, 2.8498E+07, 2.8683E+07,
     +  2.8869E+07, 2.9057E+07, 2.9246E+07, 2.9435E+07, 2.9626E+07,
     + 2.9818E+07, 3.0011E+07, 3.0206E+07, 3.0401E+07, 3.0598E+07,
     +  3.0796E+07, 3.0995E+07, 3.1195E+07, 3.1396E+07, 3.1599E+07,
     + 3.1803E+07, 3.2008E+07, 3.2214E+07, 3.2421E+07, 3.2630E+07,
     +  3.2840E+07, 3.3051E+07, 3.3263E+07, 3.3477E+07, 3.3692E+07,
     + 3.3908E+07, 3.4125E+07, 3.4344E+07, 3.4564E+07, 3.4785E+07,
     +  3.5008E+07, 3.5232E+07, 3.5457E+07, 3.5683E+07, 3.5911E+07,
     + 3.6140E+07, 3.6371E+07, 3.6603E+07, 3.6836E+07, 3.7071E+07,
     +  3.7307E+07, 3.7544E+07, 3.7783E+07, 3.8023E+07, 3.8264E+07,
     + 3.8507E+07, 3.8752E+07, 3.8998E+07, 3.9245E+07, 3.9493E+07,
     +  3.9743E+07, 3.9995E+07, 4.0248E+07, 4.0503E+07, 4.0758E+07,
     + 4.1016E+07, 4.1275E+07, 4.1535E+07, 4.1797E+07, 4.2061E+07,
     +  4.2326E+07, 4.2592E+07, 4.2860E+07, 4.3130E+07, 4.3401E+07,
     + 4.3674E+07, 4.3948E+07, 4.4224E+07, 4.4501E+07, 4.4780E+07,
     +  4.5061E+07, 4.5343E+07, 4.5627E+07, 4.5913E+07, 4.6200E+07,
     + 4.6489E+07, 4.6779E+07, 4.7071E+07, 4.7365E+07, 4.7661E+07,
     +  4.7958E+07, 4.8257E+07, 4.8557E+07, 4.8860E+07, 4.9164E+07,
     + 4.9470E+07, 4.9777E+07, 5.0086E+07, 5.0397E+07, 5.0710E+07,
     +  5.1025E+07, 5.1341E+07, 5.1659E+07, 5.1979E+07, 5.2301E+07,
     + 5.2625E+07, 5.2950E+07, 5.3278E+07, 5.3607E+07, 5.3938E+07,
     +  5.4271E+07, 5.4605E+07, 5.4942E+07, 5.5281E+07, 5.5621E+07,
     + 5.5964E+07, 5.6308E+07, 5.6654E+07, 5.7003E+07, 5.7353E+07,
     +  5.7705E+07, 5.8059E+07, 5.8415E+07, 5.8773E+07, 5.9134E+07,
     + 5.9496E+07, 5.9860E+07, 6.0226E+07, 6.0595E+07, 6.0965E+07,
     +  6.1338E+07, 6.1712E+07, 6.2089E+07, 6.2468E+07, 6.2848E+07,
     + 6.3231E+07, 6.3617E+07, 6.4004E+07, 6.4393E+07, 6.4785E+07,
     +  6.5179E+07, 6.5575E+07, 6.5973E+07, 6.6373E+07, 6.6776E+07,
     + 6.7181E+07, 6.7588E+07, 6.7998E+07, 6.8409E+07, 6.8823E+07,
     +  6.9239E+07, 6.9658E+07, 7.0079E+07, 7.0502E+07, 7.0928E+07,
     + 7.1356E+07, 7.1786E+07, 7.2218E+07, 7.2653E+07, 7.3091E+07,
     +  7.3531E+07, 7.3973E+07, 7.4418E+07, 7.4865E+07, 7.5314E+07,
     + 7.5767E+07, 7.6221E+07, 7.6678E+07, 7.7138E+07, 7.7600E+07,
     +  7.8065E+07, 7.8532E+07, 7.9002E+07, 7.9474E+07, 7.9949E+07,
     + 8.0426E+07, 8.0906E+07, 8.1389E+07, 8.1875E+07, 8.2363E+07,
     +  8.2853E+07, 8.3347E+07, 8.3843E+07, 8.4341E+07, 8.4843E+07,
     + 8.5347E+07, 8.5854E+07, 8.6364E+07, 8.6876E+07, 8.7392E+07,
     +  8.7910E+07, 8.8431E+07, 8.8954E+07, 8.9481E+07, 9.0010E+07,
     + 9.0542E+07, 9.1078E+07, 9.1616E+07, 9.2157E+07, 9.2700E+07,
     +  9.3247E+07, 9.3797E+07, 9.4350E+07, 9.4905E+07, 9.5464E+07,
     + 9.6026E+07, 9.6591E+07, 9.7158E+07, 9.7729E+07, 9.8303E+07,
     +  9.8880E+07, 9.9460E+07, 1.0004E+08, 1.0063E+08, 1.0122E+08,
     + 1.0181E+08, 1.0241E+08, 1.0301E+08, 1.0361E+08, 1.0421E+08,
     +  1.0482E+08, 1.0544E+08, 1.0605E+08, 1.0667E+08, 1.0729E+08,
     + 1.0792E+08, 1.0854E+08, 1.0918E+08, 1.0981E+08, 1.1045E+08,
     +  1.1109E+08, 1.1174E+08, 1.1239E+08, 1.1304E+08, 1.1369E+08,
     + 1.1435E+08, 1.1502E+08, 1.1568E+08, 1.1635E+08, 1.1702E+08,
     +  1.1770E+08, 1.1838E+08, 1.1907E+08, 1.1975E+08, 1.2044E+08,
     + 1.2114E+08, 1.2184E+08, 1.2254E+08, 1.2325E+08, 1.2396E+08,
     +  1.2467E+08, 1.2539E+08, 1.2611E+08, 1.2683E+08, 1.2756E+08,
     + 1.2829E+08, 1.2903E+08, 1.2977E+08, 1.3051E+08, 1.3126E+08,
     +  1.3201E+08, 1.3277E+08, 1.3353E+08, 1.3429E+08, 1.3506E+08,
     + 1.3583E+08, 1.3661E+08, 1.3739E+08, 1.3817E+08, 1.3896E+08,
     +  1.3975E+08, 1.4055E+08, 1.4135E+08, 1.4215E+08, 1.4296E+08,
     + 1.4377E+08, 1.4459E+08, 1.4541E+08, 1.4624E+08, 1.4707E+08,
     +  1.4790E+08, 1.4874E+08, 1.4958E+08, 1.5043E+08, 1.5128E+08,
     + 1.5214E+08, 1.5300E+08, 1.5386E+08, 1.5473E+08, 1.5560E+08,
     +  1.5648E+08, 1.5736E+08, 1.5825E+08, 1.5914E+08, 1.6004E+08,
     + 1.6094E+08, 1.6185E+08, 1.6276E+08, 1.6367E+08, 1.6459E+08,
     +  1.6552E+08, 1.6645E+08, 1.6738E+08, 1.6832E+08, 1.6926E+08,
     + 1.7021E+08, 1.7117E+08, 1.7212E+08, 1.7309E+08, 1.7406E+08,
     +  1.7503E+08, 1.7601E+08, 1.7699E+08, 1.7798E+08, 1.7897E+08,
     + 1.7997E+08, 1.8097E+08, 1.8198E+08, 1.8299E+08, 1.8401E+08,
     +  1.8504E+08, 1.8606E+08, 1.8710E+08, 1.8814E+08, 1.8918E+08,
     + 1.9023E+08, 1.9129E+08, 1.9235E+08, 1.9342E+08, 1.9449E+08,
     +  1.9556E+08, 1.9665E+08, 1.9773E+08, 1.9883E+08, 1.9993E+08,
     + 2.0103E+08, 2.0214E+08, 2.0326E+08, 2.0438E+08, 2.0550E+08,
     +  2.0664E+08, 2.0778E+08, 2.0892E+08, 2.1007E+08, 2.1122E+08,
     + 2.1239E+08, 2.1355E+08, 2.1473E+08, 2.1590E+08, 2.1709E+08,
     +  2.1828E+08, 2.1948E+08, 2.2068E+08, 2.2189E+08, 2.2310E+08,
     + 2.2432E+08, 2.2555E+08, 2.2678E+08, 2.2802E+08, 2.2927E+08,
     +  2.3052E+08, 2.3178E+08, 2.3304E+08, 2.3431E+08, 2.3559E+08,
     + 2.3687E+08, 2.3816E+08, 2.3946E+08, 2.4076E+08, 2.4207E+08,
     +  2.4338E+08, 2.4471E+08, 2.4603E+08, 2.4737E+08, 2.4871E+08,
     + 2.5006E+08, 2.5141E+08, 2.5278E+08, 2.5414E+08, 2.5552E+08,
     +  2.5690E+08, 2.5829E+08, 2.5969E+08, 2.6109E+08, 2.6250E+08,
     + 2.6391E+08, 2.6534E+08, 2.6677E+08, 2.6820E+08, 2.6965E+08,
     +  2.7110E+08, 2.7256E+08, 2.7402E+08, 2.7550E+08, 2.7698E+08,
     + 2.7847E+08, 2.7996E+08, 2.8146E+08, 2.8297E+08, 2.8449E+08,
     +  2.8601E+08, 2.8754E+08, 2.8908E+08, 2.9063E+08, 2.9218E+08,
     + 2.9375E+08, 2.9532E+08, 2.9689E+08, 2.9848E+08, 3.0007E+08,
     +  3.0167E+08, 3.0328E+08, 3.0489E+08, 3.0652E+08, 3.0815E+08,
     + 3.0979E+08, 3.1144E+08, 3.1309E+08, 3.1476E+08, 3.1643E+08,
     +  3.1811E+08, 3.1980E+08, 3.2149E+08, 3.2320E+08, 3.2491E+08,
     + 3.2663E+08, 3.2836E+08, 3.3010E+08, 3.3184E+08, 3.3360E+08,
     +  3.3536E+08, 3.3713E+08, 3.3891E+08, 3.4070E+08, 3.4249E+08,
     + 3.4430E+08, 3.4611E+08, 3.4794E+08, 3.4977E+08, 3.5161E+08,
     +  3.5346E+08, 3.5532E+08, 3.5719E+08, 3.5906E+08, 3.6095E+08,
     + 3.6284E+08, 3.6474E+08, 3.6666E+08, 3.6858E+08, 3.7051E+08,
     +  3.7245E+08, 3.7440E+08, 3.7636E+08, 3.7832E+08, 3.8030E+08,
     + 3.8229E+08, 3.8428E+08, 3.8629E+08, 3.8830E+08, 3.9033E+08,
     +  3.9236E+08, 3.9441E+08, 3.9646E+08, 3.9853E+08, 4.0060E+08,
     + 4.0268E+08, 4.0477E+08, 4.0688E+08, 4.0899E+08, 4.1111E+08,
     +  4.1325E+08, 4.1539E+08, 4.1754E+08, 4.1971E+08, 4.2188E+08,
     + 4.2406E+08, 4.2626E+08, 4.2846E+08, 4.3068E+08, 4.3290E+08,
     +  4.3514E+08, 4.3739E+08, 4.3964E+08, 4.4191E+08, 4.4419E+08,
     + 4.4648E+08, 4.4878E+08, 4.5109E+08, 4.5341E+08, 4.5574E+08,
     +  4.5809E+08, 4.6044E+08, 4.6280E+08, 4.6518E+08, 4.6757E+08,
     + 4.6997E+08, 4.7238E+08, 4.7480E+08, 4.7723E+08, 4.7967E+08,
     +  4.8213E+08, 4.8460E+08, 4.8708E+08, 4.8956E+08, 4.9207E+08,
     + 4.9458E+08, 4.9710E+08, 4.9964E+08, 5.0219E+08, 5.0475E+08,
     +  5.0732E+08, 5.0991E+08, 5.1250E+08, 5.1511E+08, 5.1773E+08,
     + 5.2036E+08, 5.2301E+08, 5.2566E+08, 5.2833E+08, 5.3101E+08,
     +  5.3371E+08, 5.3641E+08, 5.3913E+08, 5.4186E+08, 5.4461E+08,
     + 5.4736E+08, 5.5013E+08, 5.5292E+08, 5.5571E+08, 5.5852E+08,
     +  5.6134E+08, 5.6417E+08, 5.6702E+08, 5.6988E+08, 5.7275E+08,
     + 5.7564E+08, 5.7854E+08, 5.8145E+08, 5.8438E+08, 5.8732E+08,
     +  5.9027E+08, 5.9324E+08, 5.9622E+08, 5.9921E+08, 6.0222E+08,
     + 6.0524E+08, 6.0828E+08, 6.1132E+08, 6.1439E+08, 6.1746E+08,
     +  6.2056E+08, 6.2366E+08, 6.2678E+08, 6.2991E+08, 6.3306E+08,
     + 6.3622E+08, 6.3940E+08, 6.4259E+08, 6.4580E+08, 6.4902E+08,
     +  6.5225E+08, 6.5550E+08, 6.5876E+08, 6.6204E+08, 6.6534E+08,
     + 6.6865E+08, 6.7197E+08, 6.7531E+08, 6.7866E+08, 6.8203E+08,
     +  6.8542E+08, 6.8882E+08, 6.9223E+08, 6.9566E+08, 6.9911E+08,
     + 7.0257E+08, 7.0605E+08, 7.0954E+08, 7.1305E+08, 7.1657E+08,
     +  7.2011E+08, 7.2367E+08, 7.2724E+08, 7.3083E+08, 7.3443E+08,
     + 7.3805E+08, 7.4169E+08, 7.4534E+08, 7.4901E+08, 7.5270E+08,
     +  7.5640E+08, 7.6012E+08, 7.6386E+08, 7.6761E+08, 7.7138E+08,
     + 7.7516E+08, 7.7897E+08, 7.8279E+08, 7.8662E+08, 7.9048E+08,
     +  7.9435E+08, 7.9824E+08, 8.0215E+08, 8.0607E+08, 8.1001E+08,
     + 8.1397E+08, 8.1794E+08, 8.2194E+08, 8.2595E+08, 8.2998E+08,
     +  8.3403E+08, 8.3809E+08, 8.4218E+08, 8.4628E+08, 8.5040E+08,
     + 8.5454E+08, 8.5869E+08, 8.6287E+08, 8.6706E+08, 8.7127E+08,
     +  8.7550E+08, 8.7975E+08, 8.8402E+08, 8.8831E+08, 8.9261E+08,
     + 8.9694E+08, 9.0128E+08, 9.0564E+08, 9.1003E+08, 9.1443E+08,
     +  9.1885E+08, 9.2329E+08, 9.2775E+08, 9.3223E+08, 9.3673E+08,
     + 9.4125E+08, 9.4579E+08, 9.5035E+08, 9.5493E+08, 9.5952E+08,
     +  9.6414E+08, 9.6878E+08, 9.7344E+08, 9.7812E+08, 9.8282E+08,
     + 9.8755E+08, 9.9229E+08, 9.9705E+08, 1.0018E+09, 1.0066E+09,
     +  1.0115E+09, 1.0163E+09, 1.0212E+09, 1.0261E+09, 1.0310E+09,
     + 1.0359E+09, 1.0409E+09, 1.0458E+09, 1.0508E+09, 1.0559E+09,
     +  1.0609E+09, 1.0660E+09, 1.0710E+09, 1.0761E+09, 1.0813E+09,
     + 1.0864E+09, 1.0916E+09, 1.0968E+09, 1.1020E+09, 1.1072E+09,
     +  1.1125E+09, 1.1178E+09, 1.1231E+09, 1.1284E+09, 1.1338E+09,
     + 1.1392E+09, 1.1446E+09, 1.1500E+09, 1.1554E+09, 1.1609E+09,
     +  1.1664E+09, 1.1719E+09, 1.1775E+09, 1.1830E+09, 1.1886E+09,
     + 1.1942E+09, 1.1999E+09, 1.2055E+09, 1.2112E+09, 1.2169E+09,
     +  1.2227E+09, 1.2284E+09, 1.2342E+09, 1.2400E+09, 1.2459E+09,
     + 1.2517E+09, 1.2576E+09, 1.2635E+09, 1.2694E+09, 1.2754E+09,
     +  1.2814E+09, 1.2874E+09, 1.2934E+09, 1.2995E+09, 1.3056E+09,
     + 1.3117E+09, 1.3179E+09, 1.3240E+09, 1.3302E+09, 1.3364E+09,
     +  1.3427E+09, 1.3490E+09, 1.3553E+09, 1.3616E+09, 1.3679E+09,
     + 1.3743E+09, 1.3807E+09, 1.3872E+09, 1.3936E+09, 1.4001E+09,
     +  1.4066E+09, 1.4132E+09, 1.4197E+09, 1.4263E+09, 1.4330E+09,
     + 1.4396E+09, 1.4463E+09, 1.4530E+09, 1.4598E+09, 1.4665E+09,
     +  1.4733E+09, 1.4802E+09, 1.4870E+09, 1.4939E+09, 1.5008E+09,
     + 1.5077E+09, 1.5147E+09, 1.5217E+09, 1.5288E+09, 1.5358E+09,
     +  1.5429E+09, 1.5500E+09, 1.5572E+09, 1.5643E+09, 1.5716E+09,
     + 1.5788E+09, 1.5861E+09, 1.5934E+09, 1.6007E+09, 1.6081E+09,
     +  1.6155E+09, 1.6229E+09, 1.6303E+09, 1.6378E+09, 1.6453E+09,
     + 1.6529E+09, 1.6605E+09, 1.6681E+09, 1.6757E+09, 1.6834E+09,
     +  1.6911E+09, 1.6988E+09, 1.7066E+09, 1.7144E+09, 1.7222E+09,
     + 1.7301E+09, 1.7380E+09, 1.7460E+09, 1.7539E+09, 1.7619E+09,
     +  1.7700E+09, 1.7780E+09, 1.7861E+09, 1.7943E+09, 1.8024E+09,
     + 1.8106E+09, 1.8189E+09, 1.8271E+09, 1.8354E+09, 1.8438E+09,
     +  1.8521E+09, 1.8606E+09, 1.8690E+09, 1.8775E+09, 1.8860E+09,
     + 1.8945E+09, 1.9031E+09, 1.9117E+09, 1.9204E+09, 1.9291E+09,
     +  1.9378E+09, 1.9466E+09, 1.9554E+09, 1.9642E+09, 1.9731E+09,
     + 1.9820E+09, 1.9909E+09, 1.9999E+09, 2.0089E+09, 2.0180E+09,
     +  2.0270E+09, 2.0362E+09, 2.0453E+09, 2.0545E+09, 2.0638E+09,
     + 2.0731E+09, 2.0824E+09, 2.0917E+09, 2.1011E+09, 2.1105E+09,
     +  2.1200E+09, 2.1295E+09, 2.1391E+09, 2.1487E+09, 2.1583E+09,
     + 2.1679E+09, 2.1776E+09, 2.1874E+09, 2.1972E+09, 2.2070E+09,
     +  2.2168E+09, 2.2268E+09, 2.2367E+09, 2.2467E+09, 2.2567E+09,
     + 2.2668E+09, 2.2769E+09, 2.2870E+09, 2.2972E+09, 2.3074E+09,
     +  2.3177E+09, 2.3280E+09, 2.3384E+09, 2.3488E+09, 2.3592E+09,
     + 2.3697E+09, 2.3802E+09, 2.3907E+09, 2.4014E+09, 2.4120E+09,
     +  2.4227E+09, 2.4334E+09, 2.4442E+09, 2.4550E+09, 2.4659E+09,
     + 2.4768E+09, 2.4878E+09, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_H2O2  (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 1),QofT( 1,NT),Q(NT)
      data Tmx(1)/6000./
      data xgj/ 1./
c...     H2O2
c...        --      1661
      data (QofT( 1,J),J=1,2408)/ 1.7701E+00, 3.7711E+00, 5.6521E+00,
     +  7.6991E+00, 1.0149E+01, 1.3088E+01, 1.6536E+01, 2.0481E+01,
     + 2.4903E+01, 2.9779E+01, 3.5084E+01, 4.0800E+01, 4.6906E+01,
     +  5.3387E+01, 6.0227E+01, 6.7412E+01, 7.4931E+01, 8.2771E+01,
     + 9.0923E+01, 9.9377E+01, 1.1716E+02, 1.3604E+02, 1.5598E+02,
     +  1.7693E+02, 1.9883E+02, 2.2166E+02, 2.4538E+02, 2.6995E+02,
     + 2.9536E+02, 3.2157E+02, 3.4857E+02, 3.7633E+02, 4.0484E+02,
     +  4.3408E+02, 4.6404E+02, 4.9471E+02, 5.2606E+02, 5.5811E+02,
     + 5.9083E+02, 6.2422E+02, 6.5827E+02, 6.9298E+02, 7.2835E+02,
     +  7.6436E+02, 8.0103E+02, 8.3834E+02, 8.7630E+02, 9.1490E+02,
     + 9.5415E+02, 9.9405E+02, 1.0346E+03, 1.0758E+03, 1.1176E+03,
     +  1.1601E+03, 1.2032E+03, 1.2470E+03, 1.2915E+03, 1.3366E+03,
     + 1.3823E+03, 1.4288E+03, 1.4759E+03, 1.5237E+03, 1.5721E+03,
     +  1.6212E+03, 1.6710E+03, 1.7215E+03, 1.7727E+03, 1.8246E+03,
     + 1.8772E+03, 1.9304E+03, 1.9844E+03, 2.0391E+03, 2.0945E+03,
     +  2.1506E+03, 2.2074E+03, 2.2649E+03, 2.3232E+03, 2.3822E+03,
     + 2.4419E+03, 2.5024E+03, 2.5636E+03, 2.6256E+03, 2.6883E+03,
     +  2.7517E+03, 2.8159E+03, 2.8809E+03, 2.9466E+03, 3.0131E+03,
     + 3.0804E+03, 3.1485E+03, 3.2173E+03, 3.2869E+03, 3.3573E+03,
     +  3.4285E+03, 3.5005E+03, 3.5733E+03, 3.6469E+03, 3.7213E+03,
     + 3.7965E+03, 3.8726E+03, 3.9494E+03, 4.0271E+03, 4.1056E+03,
     +  4.1850E+03, 4.2652E+03, 4.3462E+03, 4.4281E+03, 4.5108E+03,
     + 4.5944E+03, 4.6789E+03, 4.7642E+03, 4.8504E+03, 4.9375E+03,
     +  5.0254E+03, 5.1143E+03, 5.2040E+03, 5.2946E+03, 5.3862E+03,
     + 5.4786E+03, 5.5719E+03, 5.6662E+03, 5.7614E+03, 5.8575E+03,
     +  5.9545E+03, 6.0525E+03, 6.1514E+03, 6.2513E+03, 6.3521E+03,
     + 6.4538E+03, 6.5566E+03, 6.6603E+03, 6.7650E+03, 6.8706E+03,
     +  6.9773E+03, 7.0849E+03, 7.1935E+03, 7.3032E+03, 7.4138E+03,
     + 7.5255E+03, 7.6382E+03, 7.7519E+03, 7.8666E+03, 7.9824E+03,
     +  8.0992E+03, 8.2171E+03, 8.3360E+03, 8.4560E+03, 8.5771E+03,
     + 8.6992E+03, 8.8225E+03, 8.9468E+03, 9.0722E+03, 9.1987E+03,
     +  9.3263E+03, 9.4550E+03, 9.5849E+03, 9.7159E+03, 9.8480E+03,
     + 9.9813E+03, 1.0116E+04, 1.0251E+04, 1.0388E+04, 1.0526E+04,
     +  1.0665E+04, 1.0805E+04, 1.0947E+04, 1.1089E+04, 1.1233E+04,
     + 1.1378E+04, 1.1524E+04, 1.1672E+04, 1.1821E+04, 1.1971E+04,
     +  1.2122E+04, 1.2274E+04, 1.2428E+04, 1.2583E+04, 1.2739E+04,
     + 1.2897E+04, 1.3056E+04, 1.3216E+04, 1.3377E+04, 1.3540E+04,
     +  1.3704E+04, 1.3870E+04, 1.4036E+04, 1.4205E+04, 1.4374E+04,
     + 1.4545E+04, 1.4717E+04, 1.4891E+04, 1.5066E+04, 1.5242E+04,
     +  1.5420E+04, 1.5599E+04, 1.5779E+04, 1.5962E+04, 1.6145E+04,
     + 1.6330E+04, 1.6516E+04, 1.6704E+04, 1.6893E+04, 1.7084E+04,
     +  1.7276E+04, 1.7470E+04, 1.7665E+04, 1.7862E+04, 1.8060E+04,
     + 1.8260E+04, 1.8461E+04, 1.8664E+04, 1.8869E+04, 1.9075E+04,
     +  1.9282E+04, 1.9491E+04, 1.9702E+04, 1.9914E+04, 2.0128E+04,
     + 2.0344E+04, 2.0561E+04, 2.0780E+04, 2.1000E+04, 2.1222E+04,
     +  2.1446E+04, 2.1671E+04, 2.1899E+04, 2.2127E+04, 2.2358E+04,
     + 2.2590E+04, 2.2824E+04, 2.3060E+04, 2.3297E+04, 2.3536E+04,
     +  2.3777E+04, 2.4019E+04, 2.4264E+04, 2.4510E+04, 2.4758E+04,
     + 2.5008E+04, 2.5259E+04, 2.5513E+04, 2.5768E+04, 2.6025E+04,
     +  2.6284E+04, 2.6545E+04, 2.6807E+04, 2.7072E+04, 2.7338E+04,
     + 2.7607E+04, 2.7877E+04, 2.8149E+04, 2.8423E+04, 2.8699E+04,
     +  2.8977E+04, 2.9257E+04, 2.9539E+04, 2.9823E+04, 3.0109E+04,
     + 3.0397E+04, 3.0687E+04, 3.0979E+04, 3.1272E+04, 3.1568E+04,
     +  3.1866E+04, 3.2167E+04, 3.2469E+04, 3.2773E+04, 3.3079E+04,
     + 3.3388E+04, 3.3698E+04, 3.4011E+04, 3.4326E+04, 3.4643E+04,
     +  3.4962E+04, 3.5283E+04, 3.5607E+04, 3.5933E+04, 3.6261E+04,
     + 3.6591E+04, 3.6923E+04, 3.7258E+04, 3.7594E+04, 3.7934E+04,
     +  3.8275E+04, 3.8619E+04, 3.8965E+04, 3.9313E+04, 3.9663E+04,
     + 4.0016E+04, 4.0371E+04, 4.0729E+04, 4.1089E+04, 4.1451E+04,
     +  4.1816E+04, 4.2183E+04, 4.2553E+04, 4.2924E+04, 4.3299E+04,
     + 4.3676E+04, 4.4055E+04, 4.4437E+04, 4.4821E+04, 4.5207E+04,
     +  4.5597E+04, 4.5988E+04, 4.6383E+04, 4.6779E+04, 4.7179E+04,
     + 4.7581E+04, 4.7985E+04, 4.8392E+04, 4.8802E+04, 4.9214E+04,
     +  4.9629E+04, 5.0046E+04, 5.0467E+04, 5.0889E+04, 5.1315E+04,
     + 5.1743E+04, 5.2174E+04, 5.2608E+04, 5.3044E+04, 5.3483E+04,
     +  5.3925E+04, 5.4369E+04, 5.4817E+04, 5.5267E+04, 5.5720E+04,
     + 5.6176E+04, 5.6634E+04, 5.7096E+04, 5.7560E+04, 5.8027E+04,
     +  5.8497E+04, 5.8970E+04, 5.9446E+04, 5.9924E+04, 6.0406E+04,
     + 6.0891E+04, 6.1378E+04, 6.1869E+04, 6.2362E+04, 6.2859E+04,
     +  6.3358E+04, 6.3861E+04, 6.4367E+04, 6.4875E+04, 6.5387E+04,
     + 6.5902E+04, 6.6420E+04, 6.6941E+04, 6.7465E+04, 6.7992E+04,
     +  6.8523E+04, 6.9056E+04, 6.9593E+04, 7.0133E+04, 7.0676E+04,
     + 7.1222E+04, 7.1772E+04, 7.2325E+04, 7.2881E+04, 7.3440E+04,
     +  7.4003E+04, 7.4569E+04, 7.5138E+04, 7.5711E+04, 7.6287E+04,
     + 7.6866E+04, 7.7449E+04, 7.8035E+04, 7.8625E+04, 7.9218E+04,
     +  7.9814E+04, 8.0414E+04, 8.1017E+04, 8.1624E+04, 8.2234E+04,
     + 8.2848E+04, 8.3466E+04, 8.4086E+04, 8.4711E+04, 8.5339E+04,
     +  8.5970E+04, 8.6606E+04, 8.7244E+04, 8.7887E+04, 8.8533E+04,
     + 8.9183E+04, 8.9836E+04, 9.0493E+04, 9.1154E+04, 9.1819E+04,
     +  9.2487E+04, 9.3159E+04, 9.3835E+04, 9.4514E+04, 9.5198E+04,
     + 9.5885E+04, 9.6576E+04, 9.7271E+04, 9.7970E+04, 9.8673E+04,
     +  9.9379E+04, 1.0009E+05, 1.0080E+05, 1.0152E+05, 1.0224E+05,
     + 1.0297E+05, 1.0370E+05, 1.0444E+05, 1.0517E+05, 1.0592E+05,
     +  1.0666E+05, 1.0741E+05, 1.0817E+05, 1.0893E+05, 1.0969E+05,
     + 1.1046E+05, 1.1123E+05, 1.1200E+05, 1.1278E+05, 1.1357E+05,
     +  1.1435E+05, 1.1515E+05, 1.1594E+05, 1.1674E+05, 1.1755E+05,
     + 1.1836E+05, 1.1917E+05, 1.1999E+05, 1.2081E+05, 1.2164E+05,
     +  1.2247E+05, 1.2330E+05, 1.2414E+05, 1.2499E+05, 1.2584E+05,
     + 1.2669E+05, 1.2755E+05, 1.2841E+05, 1.2928E+05, 1.3015E+05,
     +  1.3103E+05, 1.3191E+05, 1.3279E+05, 1.3368E+05, 1.3458E+05,
     + 1.3547E+05, 1.3638E+05, 1.3729E+05, 1.3820E+05, 1.3912E+05,
     +  1.4004E+05, 1.4097E+05, 1.4190E+05, 1.4284E+05, 1.4378E+05,
     + 1.4473E+05, 1.4568E+05, 1.4663E+05, 1.4759E+05, 1.4856E+05,
     +  1.4953E+05, 1.5051E+05, 1.5149E+05, 1.5248E+05, 1.5347E+05,
     + 1.5446E+05, 1.5546E+05, 1.5647E+05, 1.5748E+05, 1.5850E+05,
     +  1.5952E+05, 1.6054E+05, 1.6158E+05, 1.6261E+05, 1.6365E+05,
     + 1.6470E+05, 1.6575E+05, 1.6681E+05, 1.6787E+05, 1.6894E+05,
     +  1.7002E+05, 1.7110E+05, 1.7218E+05, 1.7327E+05, 1.7436E+05,
     + 1.7546E+05, 1.7657E+05, 1.7768E+05, 1.7880E+05, 1.7992E+05,
     +  1.8105E+05, 1.8218E+05, 1.8332E+05, 1.8446E+05, 1.8561E+05,
     + 1.8677E+05, 1.8793E+05, 1.8909E+05, 1.9027E+05, 1.9144E+05,
     +  1.9263E+05, 1.9382E+05, 1.9501E+05, 1.9621E+05, 1.9742E+05,
     + 1.9863E+05, 1.9985E+05, 2.0107E+05, 2.0230E+05, 2.0354E+05,
     +  2.0478E+05, 2.0602E+05, 2.0728E+05, 2.0854E+05, 2.0980E+05,
     + 2.1107E+05, 2.1235E+05, 2.1363E+05, 2.1492E+05, 2.1622E+05,
     +  2.1752E+05, 2.1882E+05, 2.2014E+05, 2.2146E+05, 2.2278E+05,
     + 2.2411E+05, 2.2545E+05, 2.2680E+05, 2.2815E+05, 2.2950E+05,
     +  2.3087E+05, 2.3223E+05, 2.3361E+05, 2.3499E+05, 2.3638E+05,
     + 2.3777E+05, 2.3918E+05, 2.4058E+05, 2.4200E+05, 2.4342E+05,
     +  2.4484E+05, 2.4628E+05, 2.4772E+05, 2.4916E+05, 2.5062E+05,
     + 2.5207E+05, 2.5354E+05, 2.5501E+05, 2.5649E+05, 2.5798E+05,
     +  2.5947E+05, 2.6097E+05, 2.6248E+05, 2.6399E+05, 2.6551E+05,
     + 2.6703E+05, 2.6857E+05, 2.7011E+05, 2.7165E+05, 2.7321E+05,
     +  2.7477E+05, 2.7633E+05, 2.7791E+05, 2.7949E+05, 2.8108E+05,
     + 2.8267E+05, 2.8427E+05, 2.8588E+05, 2.8750E+05, 2.8912E+05,
     +  2.9075E+05, 2.9239E+05, 2.9403E+05, 2.9569E+05, 2.9735E+05,
     + 2.9901E+05, 3.0068E+05, 3.0236E+05, 3.0405E+05, 3.0575E+05,
     +  3.0745E+05, 3.0916E+05, 3.1088E+05, 3.1260E+05, 3.1433E+05,
     + 3.1607E+05, 3.1782E+05, 3.1957E+05, 3.2133E+05, 3.2310E+05,
     +  3.2488E+05, 3.2666E+05, 3.2845E+05, 3.3025E+05, 3.3206E+05,
     + 3.3387E+05, 3.3569E+05, 3.3752E+05, 3.3936E+05, 3.4120E+05,
     +  3.4305E+05, 3.4491E+05, 3.4678E+05, 3.4866E+05, 3.5054E+05,
     + 3.5243E+05, 3.5433E+05, 3.5624E+05, 3.5815E+05, 3.6007E+05,
     +  3.6200E+05, 3.6394E+05, 3.6589E+05, 3.6784E+05, 3.6981E+05,
     + 3.7178E+05, 3.7375E+05, 3.7574E+05, 3.7773E+05, 3.7974E+05,
     +  3.8175E+05, 3.8377E+05, 3.8579E+05, 3.8783E+05, 3.8987E+05,
     + 3.9192E+05, 3.9398E+05, 3.9605E+05, 3.9813E+05, 4.0021E+05,
     +  4.0230E+05, 4.0440E+05, 4.0651E+05, 4.0863E+05, 4.1076E+05,
     + 4.1289E+05, 4.1503E+05, 4.1719E+05, 4.1935E+05, 4.2151E+05,
     +  4.2369E+05, 4.2588E+05, 4.2807E+05, 4.3027E+05, 4.3248E+05,
     + 4.3470E+05, 4.3693E+05, 4.3917E+05, 4.4141E+05, 4.4367E+05,
     +  4.4593E+05, 4.4820E+05, 4.5048E+05, 4.5277E+05, 4.5507E+05,
     + 4.5738E+05, 4.5969E+05, 4.6202E+05, 4.6435E+05, 4.6670E+05,
     +  4.6905E+05, 4.7141E+05, 4.7378E+05, 4.7615E+05, 4.7854E+05,
     + 4.8094E+05, 4.8334E+05, 4.8576E+05, 4.8818E+05, 4.9061E+05,
     +  4.9305E+05, 4.9550E+05, 4.9796E+05, 5.0043E+05, 5.0291E+05,
     + 5.0540E+05, 5.0790E+05, 5.1040E+05, 5.1292E+05, 5.1544E+05,
     +  5.1797E+05, 5.2052E+05, 5.2307E+05, 5.2563E+05, 5.2820E+05,
     + 5.3078E+05, 5.3337E+05, 5.3597E+05, 5.3858E+05, 5.4120E+05,
     +  5.4383E+05, 5.4646E+05, 5.4911E+05, 5.5177E+05, 5.5443E+05,
     + 5.5711E+05, 5.5979E+05, 5.6249E+05, 5.6519E+05, 5.6791E+05,
     +  5.7063E+05, 5.7336E+05, 5.7611E+05, 5.7886E+05, 5.8162E+05,
     + 5.8440E+05, 5.8718E+05, 5.8997E+05, 5.9277E+05, 5.9558E+05,
     +  5.9840E+05, 6.0124E+05, 6.0408E+05, 6.0693E+05, 6.0979E+05,
     + 6.1266E+05, 6.1554E+05, 6.1843E+05, 6.2133E+05, 6.2425E+05,
     +  6.2717E+05, 6.3010E+05, 6.3304E+05, 6.3599E+05, 6.3895E+05,
     + 6.4193E+05, 6.4491E+05, 6.4790E+05, 6.5090E+05, 6.5391E+05,
     +  6.5694E+05, 6.5997E+05, 6.6301E+05, 6.6607E+05, 6.6913E+05,
     + 6.7221E+05, 6.7529E+05, 6.7838E+05, 6.8149E+05, 6.8461E+05,
     +  6.8773E+05, 6.9087E+05, 6.9401E+05, 6.9717E+05, 7.0034E+05,
     + 7.0352E+05, 7.0671E+05, 7.0991E+05, 7.1312E+05, 7.1634E+05,
     +  7.1957E+05, 7.2281E+05, 7.2606E+05, 7.2932E+05, 7.3260E+05,
     + 7.3588E+05, 7.3918E+05, 7.4248E+05, 7.4580E+05, 7.4913E+05,
     +  7.5246E+05, 7.5581E+05, 7.5917E+05, 7.6254E+05, 7.6592E+05,
     + 7.6931E+05, 7.7272E+05, 7.7613E+05, 7.7956E+05, 7.8299E+05,
     +  7.8644E+05, 7.8989E+05, 7.9336E+05, 7.9684E+05, 8.0033E+05,
     + 8.0383E+05, 8.0734E+05, 8.1087E+05, 8.1440E+05, 8.1794E+05,
     +  8.2150E+05, 8.2507E+05, 8.2865E+05, 8.3224E+05, 8.3584E+05,
     + 8.3945E+05, 8.4307E+05, 8.4670E+05, 8.5035E+05, 8.5401E+05,
     +  8.5767E+05, 8.6135E+05, 8.6504E+05, 8.6874E+05, 8.7246E+05,
     + 8.7618E+05, 8.7992E+05, 8.8366E+05, 8.8742E+05, 8.9119E+05,
     +  8.9497E+05, 8.9876E+05, 9.0257E+05, 9.0638E+05, 9.1021E+05,
     + 9.1404E+05, 9.1789E+05, 9.2175E+05, 9.2563E+05, 9.2951E+05,
     +  9.3341E+05, 9.3731E+05, 9.4123E+05, 9.4516E+05, 9.4910E+05,
     + 9.5305E+05, 9.5702E+05, 9.6100E+05, 9.6498E+05, 9.6898E+05,
     +  9.7299E+05, 9.7702E+05, 9.8105E+05, 9.8510E+05, 9.8916E+05,
     + 9.9322E+05, 9.9731E+05, 1.0014E+06, 1.0055E+06, 1.0096E+06,
     +  1.0137E+06, 1.0179E+06, 1.0220E+06, 1.0262E+06, 1.0304E+06,
     + 1.0346E+06, 1.0388E+06, 1.0430E+06, 1.0472E+06, 1.0514E+06,
     +  1.0557E+06, 1.0599E+06, 1.0642E+06, 1.0685E+06, 1.0728E+06,
     + 1.0771E+06, 1.0814E+06, 1.0857E+06, 1.0901E+06, 1.0944E+06,
     +  1.0988E+06, 1.1032E+06, 1.1076E+06, 1.1120E+06, 1.1164E+06,
     + 1.1208E+06, 1.1253E+06, 1.1297E+06, 1.1342E+06, 1.1387E+06,
     +  1.1431E+06, 1.1476E+06, 1.1522E+06, 1.1567E+06, 1.1612E+06,
     + 1.1658E+06, 1.1703E+06, 1.1749E+06, 1.1795E+06, 1.1841E+06,
     +  1.1887E+06, 1.1933E+06, 1.1980E+06, 1.2026E+06, 1.2073E+06,
     + 1.2119E+06, 1.2166E+06, 1.2213E+06, 1.2260E+06, 1.2308E+06,
     +  1.2355E+06, 1.2402E+06, 1.2450E+06, 1.2498E+06, 1.2545E+06,
     + 1.2593E+06, 1.2642E+06, 1.2690E+06, 1.2738E+06, 1.2787E+06,
     +  1.2835E+06, 1.2884E+06, 1.2933E+06, 1.2982E+06, 1.3031E+06,
     + 1.3080E+06, 1.3129E+06, 1.3179E+06, 1.3228E+06, 1.3278E+06,
     +  1.3328E+06, 1.3378E+06, 1.3428E+06, 1.3478E+06, 1.3528E+06,
     + 1.3579E+06, 1.3630E+06, 1.3680E+06, 1.3731E+06, 1.3782E+06,
     +  1.3833E+06, 1.3884E+06, 1.3936E+06, 1.3987E+06, 1.4039E+06,
     + 1.4091E+06, 1.4142E+06, 1.4194E+06, 1.4247E+06, 1.4299E+06,
     +  1.4351E+06, 1.4404E+06, 1.4456E+06, 1.4509E+06, 1.4562E+06,
     + 1.4615E+06, 1.4668E+06, 1.4721E+06, 1.4775E+06, 1.4828E+06,
     +  1.4882E+06, 1.4936E+06, 1.4990E+06, 1.5044E+06, 1.5098E+06,
     + 1.5152E+06, 1.5207E+06, 1.5261E+06, 1.5316E+06, 1.5371E+06,
     +  1.5426E+06, 1.5481E+06, 1.5536E+06, 1.5591E+06, 1.5647E+06,
     + 1.5703E+06, 1.5758E+06, 1.5814E+06, 1.5870E+06, 1.5926E+06,
     +  1.5982E+06, 1.6039E+06, 1.6095E+06, 1.6152E+06, 1.6209E+06,
     + 1.6266E+06, 1.6323E+06, 1.6380E+06, 1.6437E+06, 1.6495E+06,
     +  1.6552E+06, 1.6610E+06, 1.6668E+06, 1.6726E+06, 1.6784E+06,
     + 1.6842E+06, 1.6900E+06, 1.6959E+06, 1.7017E+06, 1.7076E+06,
     +  1.7135E+06, 1.7194E+06, 1.7253E+06, 1.7312E+06, 1.7372E+06,
     + 1.7431E+06, 1.7491E+06, 1.7551E+06, 1.7611E+06, 1.7671E+06,
     +  1.7731E+06, 1.7791E+06, 1.7852E+06, 1.7912E+06, 1.7973E+06,
     + 1.8034E+06, 1.8095E+06, 1.8156E+06, 1.8217E+06, 1.8279E+06,
     +  1.8340E+06, 1.8402E+06, 1.8464E+06, 1.8525E+06, 1.8588E+06,
     + 1.8650E+06, 1.8712E+06, 1.8774E+06, 1.8837E+06, 1.8900E+06,
     +  1.8963E+06, 1.9026E+06, 1.9089E+06, 1.9152E+06, 1.9215E+06,
     + 1.9279E+06, 1.9342E+06, 1.9406E+06, 1.9470E+06, 1.9534E+06,
     +  1.9598E+06, 1.9663E+06, 1.9727E+06, 1.9792E+06, 1.9856E+06,
     + 1.9921E+06, 1.9986E+06, 2.0051E+06, 2.0117E+06, 2.0182E+06,
     +  2.0247E+06, 2.0313E+06, 2.0379E+06, 2.0445E+06, 2.0511E+06,
     + 2.0577E+06, 2.0643E+06, 2.0710E+06, 2.0776E+06, 2.0843E+06,
     +  2.0910E+06, 2.0977E+06, 2.1044E+06, 2.1111E+06, 2.1179E+06,
     + 2.1246E+06, 2.1314E+06, 2.1382E+06, 2.1450E+06, 2.1518E+06,
     +  2.1586E+06, 2.1654E+06, 2.1723E+06, 2.1791E+06, 2.1860E+06,
     + 2.1929E+06, 2.1998E+06, 2.2067E+06, 2.2136E+06, 2.2206E+06,
     +  2.2275E+06, 2.2345E+06, 2.2415E+06, 2.2485E+06, 2.2555E+06,
     + 2.2625E+06, 2.2696E+06, 2.2766E+06, 2.2837E+06, 2.2907E+06,
     +  2.2978E+06, 2.3049E+06, 2.3121E+06, 2.3192E+06, 2.3263E+06,
     + 2.3335E+06, 2.3407E+06, 2.3478E+06, 2.3550E+06, 2.3622E+06,
     +  2.3695E+06, 2.3767E+06, 2.3840E+06, 2.3912E+06, 2.3985E+06,
     + 2.4058E+06, 2.4131E+06, 2.4204E+06, 2.4278E+06, 2.4351E+06,
     +  2.4425E+06, 2.4498E+06, 2.4572E+06, 2.4646E+06, 2.4720E+06,
     + 2.4795E+06, 2.4869E+06, 2.4944E+06, 2.5018E+06, 2.5093E+06,
     +  2.5168E+06, 2.5243E+06, 2.5319E+06, 2.5394E+06, 2.5469E+06,
     + 2.5545E+06, 2.5621E+06, 2.5697E+06, 2.5773E+06, 2.5849E+06,
     +  2.5925E+06, 2.6002E+06, 2.6078E+06, 2.6155E+06, 2.6232E+06,
     + 2.6309E+06, 2.6386E+06, 2.6463E+06, 2.6541E+06, 2.6618E+06,
     +  2.6696E+06, 2.6774E+06, 2.6852E+06, 2.6930E+06, 2.7008E+06,
     + 2.7086E+06, 2.7165E+06, 2.7243E+06, 2.7322E+06, 2.7401E+06,
     +  2.7480E+06, 2.7559E+06, 2.7638E+06, 2.7718E+06, 2.7797E+06,
     + 2.7877E+06, 2.7957E+06, 2.8037E+06, 2.8117E+06, 2.8197E+06,
     +  2.8278E+06, 2.8358E+06, 2.8439E+06, 2.8520E+06, 2.8600E+06,
     + 2.8682E+06, 2.8763E+06, 2.8844E+06, 2.8926E+06, 2.9007E+06,
     +  2.9089E+06, 2.9171E+06, 2.9253E+06, 2.9335E+06, 2.9417E+06,
     + 2.9500E+06, 2.9582E+06, 2.9665E+06, 2.9748E+06, 2.9830E+06,
     +  2.9914E+06, 2.9997E+06, 3.0080E+06, 3.0164E+06, 3.0247E+06,
     + 3.0331E+06, 3.0415E+06, 3.0499E+06, 3.0583E+06, 3.0667E+06,
     +  3.0752E+06, 3.0836E+06, 3.0921E+06, 3.1006E+06, 3.1091E+06,
     + 3.1176E+06, 3.1261E+06, 3.1347E+06, 3.1432E+06, 3.1518E+06,
     +  3.1604E+06, 3.1689E+06, 3.1775E+06, 3.1862E+06, 3.1948E+06,
     + 3.2034E+06, 3.2121E+06, 3.2208E+06, 3.2295E+06, 3.2382E+06,
     +  3.2469E+06, 3.2556E+06, 3.2643E+06, 3.2731E+06, 3.2819E+06,
     + 3.2906E+06, 3.2994E+06, 3.3082E+06, 3.3171E+06, 3.3259E+06,
     +  3.3347E+06, 3.3436E+06, 3.3525E+06, 3.3614E+06, 3.3703E+06,
     + 3.3792E+06, 3.3881E+06, 3.3970E+06, 3.4060E+06, 3.4150E+06,
     +  3.4239E+06, 3.4329E+06, 3.4419E+06, 3.4510E+06, 3.4600E+06,
     + 3.4690E+06, 3.4781E+06, 3.4872E+06, 3.4963E+06, 3.5054E+06,
     +  3.5145E+06, 3.5236E+06, 3.5327E+06, 3.5419E+06, 3.5511E+06,
     + 3.5602E+06, 3.5694E+06, 3.5786E+06, 3.5879E+06, 3.5971E+06,
     +  3.6064E+06, 3.6156E+06, 3.6249E+06, 3.6342E+06, 3.6435E+06,
     + 3.6528E+06, 3.6621E+06, 3.6715E+06, 3.6808E+06, 3.6902E+06,
     +  3.6996E+06, 3.7089E+06, 3.7184E+06, 3.7278E+06, 3.7372E+06,
     + 3.7467E+06, 3.7561E+06, 3.7656E+06, 3.7751E+06, 3.7846E+06,
     +  3.7941E+06, 3.8036E+06, 3.8131E+06, 3.8227E+06, 3.8323E+06,
     + 3.8418E+06, 3.8514E+06, 3.8610E+06, 3.8707E+06, 3.8803E+06,
     +  3.8899E+06, 3.8996E+06, 3.9093E+06, 3.9189E+06, 3.9286E+06,
     + 3.9384E+06, 3.9481E+06, 3.9578E+06, 3.9676E+06, 3.9773E+06,
     +  3.9871E+06, 3.9969E+06, 4.0067E+06, 4.0165E+06, 4.0263E+06,
     + 4.0362E+06, 4.0460E+06, 4.0559E+06, 4.0658E+06, 4.0757E+06,
     +  4.0856E+06, 4.0955E+06, 4.1054E+06, 4.1154E+06, 4.1253E+06,
     + 4.1353E+06, 4.1453E+06, 4.1553E+06, 4.1653E+06, 4.1753E+06,
     +  4.1854E+06, 4.1954E+06, 4.2055E+06, 4.2155E+06, 4.2256E+06,
     + 4.2357E+06, 4.2459E+06, 4.2560E+06, 4.2661E+06, 4.2763E+06,
     +  4.2864E+06, 4.2966E+06, 4.3068E+06, 4.3170E+06, 4.3272E+06,
     + 4.3375E+06, 4.3477E+06, 4.3580E+06, 4.3682E+06, 4.3785E+06,
     +  4.3888E+06, 4.3991E+06, 4.4095E+06, 4.4198E+06, 4.4301E+06,
     + 4.4405E+06, 4.4509E+06, 4.4613E+06, 4.4716E+06, 4.4821E+06,
     +  4.4925E+06, 4.5029E+06, 4.5134E+06, 4.5238E+06, 4.5343E+06,
     + 4.5448E+06, 4.5553E+06, 4.5658E+06, 4.5763E+06, 4.5869E+06,
     +  4.5974E+06, 4.6080E+06, 4.6186E+06, 4.6292E+06, 4.6398E+06,
     + 4.6504E+06, 4.6610E+06, 4.6717E+06, 4.6823E+06, 4.6930E+06,
     +  4.7037E+06, 4.7143E+06, 4.7251E+06, 4.7358E+06, 4.7465E+06,
     + 4.7572E+06, 4.7680E+06, 4.7788E+06, 4.7895E+06, 4.8003E+06,
     +  4.8111E+06, 4.8220E+06, 4.8328E+06, 4.8436E+06, 4.8545E+06,
     + 4.8654E+06, 4.8762E+06, 4.8871E+06, 4.8980E+06, 4.9090E+06,
     +  4.9199E+06, 4.9308E+06, 4.9418E+06, 4.9528E+06, 4.9638E+06,
     + 4.9747E+06, 4.9858E+06, 4.9968E+06, 5.0078E+06, 5.0189E+06,
     +  5.0299E+06, 5.0410E+06, 5.0521E+06, 5.0632E+06, 5.0743E+06,
     + 5.0854E+06, 5.0965E+06, 5.1077E+06, 5.1188E+06, 5.1300E+06,
     +  5.1412E+06, 5.1524E+06, 5.1636E+06, 5.1748E+06, 5.1860E+06,
     + 5.1973E+06, 5.2085E+06, 5.2198E+06, 5.2311E+06, 5.2424E+06,
     +  5.2537E+06, 5.2650E+06, 5.2763E+06, 5.2876E+06, 5.2990E+06,
     + 5.3104E+06, 5.3217E+06, 5.3331E+06, 5.3445E+06, 5.3560E+06,
     +  5.3674E+06, 5.3788E+06, 5.3903E+06, 5.4017E+06, 5.4132E+06,
     + 5.4247E+06, 5.4362E+06, 5.4477E+06, 5.4593E+06, 5.4708E+06,
     +  5.4823E+06, 5.4939E+06, 5.5055E+06, 5.5171E+06, 5.5287E+06,
     + 5.5403E+06, 5.5519E+06, 5.5635E+06, 5.5752E+06, 5.5868E+06,
     +  5.5985E+06, 5.6102E+06, 5.6219E+06, 5.6336E+06, 5.6453E+06,
     + 5.6570E+06, 5.6688E+06, 5.6805E+06, 5.6923E+06, 5.7041E+06,
     +  5.7159E+06, 5.7277E+06, 5.7395E+06, 5.7513E+06, 5.7632E+06,
     + 5.7750E+06, 5.7869E+06, 5.7988E+06, 5.8107E+06, 5.8226E+06,
     +  5.8345E+06, 5.8464E+06, 5.8583E+06, 5.8703E+06, 5.8822E+06,
     + 5.8942E+06, 5.9062E+06, 5.9182E+06, 5.9302E+06, 5.9422E+06,
     +  5.9542E+06, 5.9663E+06, 5.9783E+06, 5.9904E+06, 6.0025E+06,
     + 6.0146E+06, 6.0267E+06, 6.0388E+06, 6.0509E+06, 6.0630E+06,
     +  6.0752E+06, 6.0873E+06, 6.0995E+06, 6.1117E+06, 6.1239E+06,
     + 6.1361E+06, 6.1483E+06, 6.1606E+06, 6.1728E+06, 6.1850E+06,
     +  6.1973E+06, 6.2096E+06, 6.2219E+06, 6.2342E+06, 6.2465E+06,
     + 6.2588E+06, 6.2712E+06, 6.2835E+06, 6.2959E+06, 6.3082E+06,
     +  6.3206E+06, 6.3330E+06, 6.3454E+06, 6.3578E+06, 6.3702E+06,
     + 6.3827E+06, 6.3951E+06, 6.4076E+06, 6.4201E+06, 6.4326E+06,
     +  6.4451E+06, 6.4576E+06, 6.4701E+06, 6.4826E+06, 6.4952E+06,
     + 6.5077E+06, 6.5203E+06, 6.5329E+06, 6.5454E+06, 6.5580E+06,
     +  6.5707E+06, 6.5833E+06, 6.5959E+06, 6.6086E+06, 6.6212E+06,
     + 6.6339E+06, 6.6466E+06, 6.6593E+06, 6.6720E+06, 6.6847E+06,
     +  6.6974E+06, 6.7101E+06, 6.7229E+06, 6.7356E+06, 6.7484E+06,
     + 6.7612E+06, 6.7740E+06, 6.7868E+06, 6.7996E+06, 6.8124E+06,
     +  6.8253E+06, 6.8381E+06, 6.8510E+06, 6.8638E+06, 6.8767E+06,
     + 6.8896E+06, 6.9025E+06, 6.9154E+06, 6.9284E+06, 6.9413E+06,
     +  6.9542E+06, 6.9672E+06, 6.9802E+06, 6.9932E+06, 7.0062E+06,
     + 7.0192E+06, 7.0322E+06, 7.0452E+06, 7.0582E+06, 7.0713E+06,
     +  7.0843E+06, 7.0974E+06, 7.1105E+06, 7.1236E+06, 7.1367E+06,
     + 7.1498E+06, 7.1629E+06, 7.1761E+06, 7.1892E+06, 7.2024E+06,
     +  7.2155E+06, 7.2287E+06, 7.2419E+06, 7.2551E+06, 7.2683E+06,
     + 7.2816E+06, 7.2948E+06, 7.3080E+06, 7.3213E+06, 7.3346E+06,
     +  7.3478E+06, 7.3611E+06, 7.3744E+06, 7.3877E+06, 7.4011E+06,
     + 7.4144E+06, 7.4277E+06, 7.4411E+06, 7.4545E+06, 7.4678E+06,
     +  7.4812E+06, 7.4946E+06, 7.5080E+06, 7.5214E+06, 7.5349E+06,
     + 7.5483E+06, 7.5618E+06, 7.5752E+06, 7.5887E+06, 7.6022E+06,
     +  7.6157E+06, 7.6292E+06, 7.6427E+06, 7.6562E+06, 7.6697E+06,
     + 7.6833E+06, 7.6969E+06, 7.7104E+06, 7.7240E+06, 7.7376E+06,
     +  7.7512E+06, 7.7648E+06, 7.7784E+06, 7.7920E+06, 7.8057E+06,
     + 7.8193E+06, 7.8330E+06, 7.8467E+06, 7.8604E+06, 7.8740E+06,
     +  7.8878E+06, 7.9015E+06, 7.9152E+06, 7.9289E+06, 7.9427E+06,
     + 7.9564E+06, 7.9702E+06, 7.9840E+06, 7.9978E+06, 8.0116E+06,
     +  8.0254E+06, 8.0392E+06, 8.0530E+06, 8.0669E+06, 8.0807E+06,
     + 8.0946E+06, 8.1084E+06, 8.1223E+06, 8.1362E+06, 8.1501E+06,
     +  8.1640E+06, 8.1779E+06, 8.1919E+06, 8.2058E+06, 8.2198E+06,
     + 8.2337E+06, 8.2477E+06, 8.2617E+06, 8.2757E+06, 8.2897E+06,
     +  8.3037E+06, 8.3177E+06, 8.3317E+06, 8.3458E+06, 8.3598E+06,
     + 8.3739E+06, 8.3880E+06, 8.4021E+06, 8.4162E+06, 8.4303E+06,
     +  8.4444E+06, 8.4585E+06, 8.4726E+06, 8.4868E+06, 8.5009E+06,
     + 8.5151E+06, 8.5293E+06, 8.5434E+06, 8.5576E+06, 8.5718E+06,
     +  8.5861E+06, 8.6003E+06, 8.6145E+06, 8.6288E+06, 8.6430E+06,
     + 8.6573E+06, 8.6716E+06, 8.6858E+06, 8.7001E+06, 8.7144E+06,
     +  8.7287E+06, 8.7431E+06, 8.7574E+06, 8.7717E+06, 8.7861E+06,
     + 8.8005E+06, 8.8148E+06, 8.8292E+06, 8.8436E+06, 8.8580E+06,
     +  8.8724E+06, 8.8868E+06, 8.9013E+06, 8.9157E+06, 8.9302E+06,
     + 8.9446E+06, 8.9591E+06, 8.9736E+06, 8.9881E+06, 9.0026E+06,
     +  9.0171E+06, 9.0316E+06, 9.0461E+06, 9.0607E+06, 9.0752E+06,
     + 9.0898E+06, 9.1043E+06, 9.1189E+06, 9.1335E+06, 9.1481E+06,
     +  9.1627E+06, 9.1773E+06, 9.1919E+06, 9.2066E+06, 9.2212E+06,
     + 9.2358E+06, 9.2505E+06, 9.2652E+06, 9.2799E+06, 9.2945E+06,
     +  9.3092E+06, 9.3240E+06, 9.3387E+06, 9.3534E+06, 9.3681E+06,
     + 9.3829E+06, 9.3976E+06, 9.4124E+06, 9.4272E+06, 9.4420E+06,
     +  9.4568E+06, 9.4716E+06, 9.4864E+06, 9.5012E+06, 9.5160E+06,
     + 9.5309E+06, 9.5457E+06, 9.5606E+06, 9.5754E+06, 9.5903E+06,
     +  9.6052E+06, 9.6201E+06, 9.6350E+06, 9.6499E+06, 9.6648E+06,
     + 9.6798E+06, 9.6947E+06, 9.7097E+06, 9.7246E+06, 9.7396E+06,
     +  9.7546E+06, 9.7696E+06, 9.7846E+06, 9.7996E+06, 9.8146E+06,
     + 9.8296E+06, 9.8446E+06, 9.8597E+06, 9.8747E+06, 9.8898E+06,
     +  9.9049E+06, 9.9200E+06, 9.9350E+06, 9.9501E+06, 9.9652E+06,
     + 9.9804E+06, 9.9955E+06, 1.0011E+07, 1.0026E+07, 1.0041E+07,
     +  1.0056E+07, 1.0071E+07, 1.0086E+07, 1.0102E+07, 1.0117E+07,
     + 1.0132E+07, 1.0147E+07, 1.0162E+07, 1.0178E+07, 1.0193E+07,
     +  1.0208E+07, 1.0223E+07, 1.0239E+07, 1.0254E+07, 1.0269E+07,
     + 1.0285E+07, 1.0300E+07, 1.0315E+07, 1.0330E+07, 1.0346E+07,
     +  1.0361E+07, 1.0376E+07, 1.0392E+07, 1.0407E+07, 1.0423E+07,
     + 1.0438E+07, 1.0453E+07, 1.0469E+07, 1.0484E+07, 1.0500E+07,
     +  1.0515E+07, 1.0530E+07, 1.0546E+07, 1.0561E+07, 1.0577E+07,
     + 1.0592E+07, 1.0608E+07, 1.0623E+07, 1.0639E+07, 1.0654E+07,
     +  1.0670E+07, 1.0685E+07, 1.0701E+07, 1.0716E+07, 1.0732E+07,
     + 1.0747E+07, 1.0763E+07, 1.0779E+07, 1.0794E+07, 1.0810E+07,
     +  1.0825E+07, 1.0841E+07, 1.0856E+07, 1.0872E+07, 1.0888E+07,
     + 1.0903E+07, 1.0919E+07, 1.0935E+07, 1.0950E+07, 1.0966E+07,
     +  1.0982E+07, 1.0997E+07, 1.1013E+07, 1.1029E+07, 1.1044E+07,
     + 1.1060E+07, 1.1076E+07, 1.1092E+07, 1.1107E+07, 1.1123E+07,
     +  1.1139E+07, 1.1155E+07, 1.1171E+07, 1.1186E+07, 1.1202E+07,
     + 1.1218E+07, 1.1234E+07, 1.1250E+07, 1.1265E+07, 1.1281E+07,
     +  1.1297E+07, 1.1313E+07, 1.1329E+07, 1.1345E+07, 1.1360E+07,
     + 1.1376E+07, 1.1392E+07, 1.1408E+07, 1.1424E+07, 1.1440E+07,
     +  1.1456E+07, 1.1472E+07, 1.1488E+07, 1.1504E+07, 1.1520E+07,
     + 1.1536E+07, 1.1552E+07, 1.1568E+07, 1.1584E+07, 1.1600E+07,
     +  1.1616E+07, 1.1632E+07, 1.1648E+07, 1.1664E+07, 1.1680E+07,
     + 1.1696E+07, 1.1712E+07, 1.1728E+07, 1.1744E+07, 1.1760E+07,
     +  1.1776E+07, 1.1792E+07, 1.1808E+07, 1.1824E+07, 1.1841E+07,
     + 1.1857E+07, 1.1873E+07, 1.1889E+07, 1.1905E+07, 1.1921E+07,
     +  1.1937E+07, 1.1954E+07, 1.1970E+07, 1.1986E+07, 1.2002E+07,
     + 1.2018E+07, 1.2035E+07, 1.2051E+07, 1.2067E+07, 1.2083E+07,
     +  1.2099E+07, 1.2116E+07, 1.2132E+07, 1.2148E+07, 1.2165E+07,
     + 1.2181E+07, 1.2197E+07, 1.2213E+07, 1.2230E+07, 1.2246E+07,
     +  1.2262E+07, 1.2279E+07, 1.2295E+07, 1.2311E+07, 1.2328E+07,
     + 1.2344E+07, 1.2360E+07, 1.2377E+07, 1.2393E+07, 1.2410E+07,
     +  1.2426E+07, 1.2442E+07, 1.2459E+07, 1.2475E+07, 1.2492E+07,
     + 1.2508E+07, 1.2524E+07, 1.2541E+07, 1.2557E+07, 1.2574E+07,
     +  1.2590E+07, 1.2607E+07, 1.2623E+07, 1.2640E+07, 1.2656E+07,
     + 1.2673E+07, 1.2689E+07, 1.2706E+07, 1.2722E+07, 1.2739E+07,
     +  1.2755E+07, 1.2772E+07, 1.2788E+07, 1.2805E+07, 1.2822E+07,
     + 1.2838E+07, 1.2855E+07, 1.2871E+07, 1.2888E+07, 1.2905E+07,
     +  1.2921E+07, 1.2938E+07, 1.2954E+07, 1.2971E+07, 1.2988E+07,
     + 1.3004E+07, 1.3021E+07, 1.3038E+07, 1.3054E+07, 1.3071E+07,
     +  1.3088E+07, 1.3104E+07, 1.3121E+07, 1.3138E+07, 1.3154E+07,
     + 1.3171E+07, 1.3188E+07, 1.3205E+07, 1.3221E+07, 1.3238E+07,
     +  1.3255E+07, 1.3272E+07, 1.3288E+07, 1.3305E+07, 1.3322E+07,
     + 1.3339E+07, 1.3356E+07, 1.3372E+07, 1.3389E+07, 1.3406E+07,
     +  1.3423E+07, 1.3440E+07, 1.3457E+07, 1.3473E+07, 1.3490E+07,
     + 1.3507E+07, 1.3524E+07, 1.3541E+07, 1.3558E+07, 1.3575E+07,
     +  1.3591E+07, 1.3608E+07, 1.3625E+07, 1.3642E+07, 1.3659E+07,
     + 1.3676E+07, 1.3693E+07, 1.3710E+07, 1.3727E+07, 1.3744E+07,
     +  1.3761E+07, 1.3778E+07, 1.3795E+07, 1.3812E+07, 1.3829E+07,
     + 1.3846E+07, 1.3863E+07, 1.3880E+07, 1.3897E+07, 1.3914E+07,
     +  1.3931E+07, 1.3948E+07, 1.3965E+07, 1.3982E+07, 1.3999E+07,
     + 1.4016E+07, 1.4033E+07, 1.4050E+07, 1.4067E+07, 1.4084E+07,
     +  1.4101E+07, 1.4119E+07, 1.4136E+07, 1.4153E+07, 1.4170E+07,
     + 1.4187E+07, 1.4204E+07, 1.4221E+07, 1.4238E+07, 1.4256E+07,
     +  1.4273E+07, 1.4290E+07, 1.4307E+07, 1.4324E+07, 1.4341E+07,
     + 1.4359E+07, 1.4376E+07, 1.4393E+07, 1.4410E+07, 1.4427E+07,
     +  1.4445E+07, 1.4462E+07, 1.4479E+07, 1.4496E+07, 1.4514E+07,
     + 1.4531E+07, 1.4548E+07, 1.4565E+07, 1.4583E+07, 1.4600E+07,
     +  1.4617E+07, 1.4635E+07, 1.4652E+07, 1.4669E+07, 1.4686E+07,
     + 1.4704E+07, 1.4721E+07, 1.4738E+07, 1.4756E+07, 1.4773E+07,
     +  1.4790E+07, 1.4808E+07, 1.4825E+07, 1.4843E+07, 1.4860E+07,
     + 1.4877E+07, 1.4895E+07, 1.4912E+07, 1.4929E+07, 1.4947E+07,
     +  1.4964E+07, 1.4982E+07, 1.4999E+07, 1.5017E+07, 1.5034E+07,
     + 1.5051E+07, 1.5069E+07, 1.5086E+07, 1.5104E+07, 1.5121E+07,
     +  1.5139E+07, 1.5156E+07, 1.5174E+07, 1.5191E+07, 1.5209E+07,
     + 1.5226E+07, 1.5244E+07, 1.5261E+07, 1.5279E+07, 1.5296E+07,
     +  1.5314E+07, 1.5331E+07, 1.5349E+07, 1.5366E+07, 1.5384E+07,
     + 1.5402E+07, 1.5419E+07, 1.5437E+07, 1.5454E+07, 1.5472E+07,
     +  1.5489E+07, 1.5507E+07, 1.5525E+07, 1.5542E+07, 1.5560E+07,
     + 1.5577E+07, 1.5595E+07, 1.5613E+07, 1.5630E+07, 1.5648E+07,
     +  1.5666E+07, 1.5683E+07, 1.5701E+07, 1.5719E+07, 1.5736E+07,
     + 1.5754E+07, 1.5772E+07, 1.5789E+07, 1.5807E+07, 1.5825E+07,
     +  1.5842E+07, 1.5860E+07, 1.5878E+07, 1.5896E+07, 1.5913E+07,
     + 1.5931E+07, 1.5949E+07, 1.5967E+07, 1.5984E+07, 1.6002E+07,
     +  1.6020E+07, 1.6038E+07, 1.6055E+07, 1.6073E+07, 1.6091E+07,
     + 1.6109E+07, 1.6126E+07, 1.6144E+07, 1.6162E+07, 1.6180E+07,
     +  1.6198E+07, 1.6216E+07, 1.6233E+07, 1.6251E+07, 1.6269E+07,
     + 1.6287E+07, 1.6305E+07, 1.6323E+07, 1.6340E+07, 1.6358E+07,
     +  1.6376E+07, 1.6394E+07, 1.6412E+07, 1.6430E+07, 1.6448E+07,
     + 1.6466E+07, 1.6484E+07, 1.6501E+07, 1.6519E+07, 1.6537E+07,
     +  1.6555E+07, 1.6573E+07, 1.6591E+07, 1.6609E+07, 1.6627E+07,
     + 1.6645E+07, 1.6663E+07, 1.6681E+07, 1.6699E+07, 1.6717E+07,
     +  1.6735E+07, 1.6753E+07, 1.6771E+07, 1.6789E+07, 1.6807E+07,
     + 1.6825E+07, 1.6843E+07, 1.6861E+07, 1.6879E+07, 1.6897E+07,
     +  1.6915E+07, 1.6933E+07, 1.6951E+07, 1.6969E+07, 1.6987E+07,
     + 1.7005E+07, 1.7023E+07, 1.7041E+07, 1.7059E+07, 1.7077E+07,
     +  1.7095E+07, 1.7113E+07, 1.7132E+07, 1.7150E+07, 1.7168E+07,
     + 1.7186E+07, 1.7204E+07, 1.7222E+07, 1.7240E+07, 1.7258E+07,
     +  1.7276E+07, 1.7295E+07, 1.7313E+07, 1.7331E+07, 1.7349E+07,
     + 1.7367E+07, 1.7385E+07, 1.7404E+07, 1.7422E+07, 1.7440E+07,
     +  1.7458E+07, 1.7476E+07, 1.7494E+07, 1.7513E+07, 1.7531E+07,
     + 1.7549E+07, 1.7567E+07, 1.7585E+07, 1.7604E+07, 1.7622E+07,
     +  1.7640E+07, 1.7658E+07, 1.7677E+07, 1.7695E+07, 1.7713E+07,
     + 1.7731E+07, 1.7750E+07, 1.7768E+07, 1.7786E+07, 1.7804E+07,
     +  1.7823E+07, 1.7841E+07, 1.7859E+07, 1.7878E+07, 1.7896E+07,
     + 1.7914E+07, 1.7933E+07, 1.7951E+07, 1.7969E+07, 1.7987E+07,
     +  1.8006E+07, 1.8024E+07, 1.8042E+07, 1.8061E+07, 1.8079E+07,
     + 1.8097E+07, 1.8116E+07, 1.8134E+07, 1.8153E+07, 1.8171E+07,
     +  1.8189E+07, 1.8208E+07, 1.8226E+07, 1.8244E+07, 1.8263E+07,
     + 1.8281E+07, 1.8300E+07, 1.8318E+07, 1.8336E+07, 1.8355E+07,
     +  1.8373E+07, 1.8392E+07, 1.8410E+07, 1.8429E+07, 1.8447E+07,
     + 1.8465E+07, 1.8484E+07, 1.8502E+07, 1.8521E+07, 1.8539E+07,
     +  1.8558E+07, 1.8576E+07, 1.8595E+07, 1.8613E+07, 1.8632E+07,
     + 1.8650E+07, 1.8669E+07, 1.8687E+07, 1.8706E+07, 1.8724E+07,
     +  1.8743E+07, 1.8761E+07, 1.8780E+07, 1.8798E+07, 1.8817E+07,
     + 1.8835E+07, 1.8854E+07, 1.8872E+07, 1.8891E+07, 1.8909E+07,
     +  1.8928E+07, 1.8946E+07, 1.8965E+07, 1.8984E+07, 1.9002E+07,
     + 1.9021E+07, 1.9039E+07, 1.9058E+07, 1.9076E+07, 1.9095E+07,
     +  1.9114E+07, 1.9132E+07, 1.9151E+07, 1.9169E+07, 1.9188E+07,
     + 1.9207E+07, 1.9225E+07, 1.9244E+07, 1.9262E+07, 1.9281E+07,
     +  1.9300E+07, 1.9318E+07, 1.9337E+07, 1.9356E+07, 1.9374E+07,
     + 1.9393E+07, 1.9412E+07, 1.9430E+07, 1.9449E+07, 1.9468E+07,
     +  1.9486E+07, 1.9505E+07, 1.9524E+07, 1.9542E+07, 1.9561E+07,
     + 1.9580E+07, 1.9598E+07, 1.9617E+07, 1.9636E+07, 1.9655E+07,
     +  1.9673E+07, 1.9692E+07, 1.9711E+07, 1.9729E+07, 1.9748E+07,
     + 1.9767E+07, 1.9786E+07, 1.9804E+07, 1.9823E+07, 1.9842E+07,
     +  1.9861E+07, 1.9879E+07, 1.9898E+07, 1.9917E+07, 1.9936E+07,
     + 1.9955E+07, 1.9973E+07, 1.9992E+07, 2.0011E+07, 2.0030E+07,
     +  2.0048E+07, 2.0067E+07, 2.0086E+07, 2.0105E+07, 2.0124E+07,
     + 2.0142E+07, 2.0161E+07, 2.0180E+07, 2.0199E+07, 2.0218E+07,
     +  2.0237E+07, 2.0255E+07, 2.0274E+07, 2.0293E+07, 2.0312E+07,
     + 2.0331E+07, 2.0350E+07, 2.0369E+07, 2.0387E+07, 2.0406E+07,
     +  2.0425E+07, 2.0444E+07, 2.0463E+07, 2.0482E+07, 2.0501E+07,
     + 2.0520E+07, 2.0538E+07, 2.0557E+07, 2.0576E+07, 2.0595E+07,
     +  2.0614E+07, 2.0633E+07, 2.0652E+07, 2.0671E+07, 2.0690E+07,
     + 2.0709E+07, 2.0728E+07, 2.0747E+07, 2.0765E+07, 2.0784E+07,
     +  2.0803E+07, 2.0822E+07, 2.0841E+07, 2.0860E+07, 2.0879E+07,
     + 2.0898E+07, 2.0917E+07, 2.0936E+07, 2.0955E+07, 2.0974E+07,
     +  2.0993E+07, 2.1012E+07, 2.1031E+07, 2.1050E+07, 2.1069E+07,
     + 2.1088E+07, 2.1107E+07, 2.1126E+07, 2.1145E+07, 2.1164E+07,
     +  2.1183E+07, 2.1202E+07, 2.1221E+07, 2.1240E+07, 2.1259E+07,
     + 2.1278E+07, 2.1297E+07, 2.1316E+07, 2.1335E+07, 2.1354E+07,
     +  2.1373E+07, 2.1392E+07, 2.1411E+07, 2.1430E+07, 2.1449E+07,
     + 2.1469E+07, 2.1488E+07, 2.1507E+07, 2.1526E+07, 2.1545E+07,
     +  2.1564E+07, 2.1583E+07, 2.1602E+07, 2.1621E+07, 2.1640E+07/
      data (QofT( 1,J),J=2409,4510)/ 2.1659E+07, 2.1678E+07,
     +  2.1698E+07, 2.1717E+07, 2.1736E+07, 2.1755E+07, 2.1774E+07,
     +  2.1793E+07,
     + 2.1812E+07, 2.1831E+07, 2.1851E+07, 2.1870E+07, 2.1889E+07,
     +  2.1908E+07, 2.1927E+07, 2.1946E+07, 2.1965E+07, 2.1985E+07,
     + 2.2004E+07, 2.2023E+07, 2.2042E+07, 2.2061E+07, 2.2080E+07,
     +  2.2100E+07, 2.2119E+07, 2.2138E+07, 2.2157E+07, 2.2176E+07,
     + 2.2195E+07, 2.2215E+07, 2.2234E+07, 2.2253E+07, 2.2272E+07,
     +  2.2291E+07, 2.2311E+07, 2.2330E+07, 2.2349E+07, 2.2368E+07,
     + 2.2387E+07, 2.2407E+07, 2.2426E+07, 2.2445E+07, 2.2464E+07,
     +  2.2484E+07, 2.2503E+07, 2.2522E+07, 2.2541E+07, 2.2561E+07,
     + 2.2580E+07, 2.2599E+07, 2.2618E+07, 2.2638E+07, 2.2657E+07,
     +  2.2676E+07, 2.2695E+07, 2.2715E+07, 2.2734E+07, 2.2753E+07,
     + 2.2772E+07, 2.2792E+07, 2.2811E+07, 2.2830E+07, 2.2850E+07,
     +  2.2869E+07, 2.2888E+07, 2.2907E+07, 2.2927E+07, 2.2946E+07,
     + 2.2965E+07, 2.2985E+07, 2.3004E+07, 2.3023E+07, 2.3043E+07,
     +  2.3062E+07, 2.3081E+07, 2.3101E+07, 2.3120E+07, 2.3139E+07,
     + 2.3159E+07, 2.3178E+07, 2.3197E+07, 2.3217E+07, 2.3236E+07,
     +  2.3255E+07, 2.3275E+07, 2.3294E+07, 2.3313E+07, 2.3333E+07,
     + 2.3352E+07, 2.3371E+07, 2.3391E+07, 2.3410E+07, 2.3429E+07,
     +  2.3449E+07, 2.3468E+07, 2.3488E+07, 2.3507E+07, 2.3526E+07,
     + 2.3546E+07, 2.3565E+07, 2.3585E+07, 2.3604E+07, 2.3623E+07,
     +  2.3643E+07, 2.3662E+07, 2.3682E+07, 2.3701E+07, 2.3720E+07,
     + 2.3740E+07, 2.3759E+07, 2.3779E+07, 2.3798E+07, 2.3817E+07,
     +  2.3837E+07, 2.3856E+07, 2.3876E+07, 2.3895E+07, 2.3915E+07,
     + 2.3934E+07, 2.3954E+07, 2.3973E+07, 2.3992E+07, 2.4012E+07,
     +  2.4031E+07, 2.4051E+07, 2.4070E+07, 2.4090E+07, 2.4109E+07,
     + 2.4129E+07, 2.4148E+07, 2.4168E+07, 2.4187E+07, 2.4206E+07,
     +  2.4226E+07, 2.4245E+07, 2.4265E+07, 2.4284E+07, 2.4304E+07,
     + 2.4323E+07, 2.4343E+07, 2.4362E+07, 2.4382E+07, 2.4401E+07,
     +  2.4421E+07, 2.4440E+07, 2.4460E+07, 2.4479E+07, 2.4499E+07,
     + 2.4518E+07, 2.4538E+07, 2.4557E+07, 2.4577E+07, 2.4596E+07,
     +  2.4616E+07, 2.4636E+07, 2.4655E+07, 2.4675E+07, 2.4694E+07,
     + 2.4714E+07, 2.4733E+07, 2.4753E+07, 2.4772E+07, 2.4792E+07,
     +  2.4811E+07, 2.4831E+07, 2.4850E+07, 2.4870E+07, 2.4890E+07,
     + 2.4909E+07, 2.4929E+07, 2.4948E+07, 2.4968E+07, 2.4987E+07,
     +  2.5007E+07, 2.5027E+07, 2.5046E+07, 2.5066E+07, 2.5085E+07,
     + 2.5105E+07, 2.5124E+07, 2.5144E+07, 2.5164E+07, 2.5183E+07,
     +  2.5203E+07, 2.5222E+07, 2.5242E+07, 2.5262E+07, 2.5281E+07,
     + 2.5301E+07, 2.5320E+07, 2.5340E+07, 2.5360E+07, 2.5379E+07,
     +  2.5399E+07, 2.5418E+07, 2.5438E+07, 2.5458E+07, 2.5477E+07,
     + 2.5497E+07, 2.5517E+07, 2.5536E+07, 2.5556E+07, 2.5575E+07,
     +  2.5595E+07, 2.5615E+07, 2.5634E+07, 2.5654E+07, 2.5674E+07,
     + 2.5693E+07, 2.5713E+07, 2.5733E+07, 2.5752E+07, 2.5772E+07,
     +  2.5791E+07, 2.5811E+07, 2.5831E+07, 2.5850E+07, 2.5870E+07,
     + 2.5890E+07, 2.5909E+07, 2.5929E+07, 2.5949E+07, 2.5968E+07,
     +  2.5988E+07, 2.6008E+07, 2.6027E+07, 2.6047E+07, 2.6067E+07,
     + 2.6087E+07, 2.6106E+07, 2.6126E+07, 2.6146E+07, 2.6165E+07,
     +  2.6185E+07, 2.6205E+07, 2.6224E+07, 2.6244E+07, 2.6264E+07,
     + 2.6283E+07, 2.6303E+07, 2.6323E+07, 2.6343E+07, 2.6362E+07,
     +  2.6382E+07, 2.6402E+07, 2.6421E+07, 2.6441E+07, 2.6461E+07,
     + 2.6481E+07, 2.6500E+07, 2.6520E+07, 2.6540E+07, 2.6559E+07,
     +  2.6579E+07, 2.6599E+07, 2.6619E+07, 2.6638E+07, 2.6658E+07,
     + 2.6678E+07, 2.6698E+07, 2.6717E+07, 2.6737E+07, 2.6757E+07,
     +  2.6777E+07, 2.6796E+07, 2.6816E+07, 2.6836E+07, 2.6856E+07,
     + 2.6875E+07, 2.6895E+07, 2.6915E+07, 2.6935E+07, 2.6954E+07,
     +  2.6974E+07, 2.6994E+07, 2.7014E+07, 2.7033E+07, 2.7053E+07,
     + 2.7073E+07, 2.7093E+07, 2.7113E+07, 2.7132E+07, 2.7152E+07,
     +  2.7172E+07, 2.7192E+07, 2.7211E+07, 2.7231E+07, 2.7251E+07,
     + 2.7271E+07, 2.7291E+07, 2.7310E+07, 2.7330E+07, 2.7350E+07,
     +  2.7370E+07, 2.7390E+07, 2.7409E+07, 2.7429E+07, 2.7449E+07,
     + 2.7469E+07, 2.7489E+07, 2.7508E+07, 2.7528E+07, 2.7548E+07,
     +  2.7568E+07, 2.7588E+07, 2.7607E+07, 2.7627E+07, 2.7647E+07,
     + 2.7667E+07, 2.7687E+07, 2.7706E+07, 2.7726E+07, 2.7746E+07,
     +  2.7766E+07, 2.7786E+07, 2.7806E+07, 2.7825E+07, 2.7845E+07,
     + 2.7865E+07, 2.7885E+07, 2.7905E+07, 2.7925E+07, 2.7944E+07,
     +  2.7964E+07, 2.7984E+07, 2.8004E+07, 2.8024E+07, 2.8044E+07,
     + 2.8064E+07, 2.8083E+07, 2.8103E+07, 2.8123E+07, 2.8143E+07,
     +  2.8163E+07, 2.8183E+07, 2.8203E+07, 2.8222E+07, 2.8242E+07,
     + 2.8262E+07, 2.8282E+07, 2.8302E+07, 2.8322E+07, 2.8342E+07,
     +  2.8361E+07, 2.8381E+07, 2.8401E+07, 2.8421E+07, 2.8441E+07,
     + 2.8461E+07, 2.8481E+07, 2.8501E+07, 2.8520E+07, 2.8540E+07,
     +  2.8560E+07, 2.8580E+07, 2.8600E+07, 2.8620E+07, 2.8640E+07,
     + 2.8660E+07, 2.8679E+07, 2.8699E+07, 2.8719E+07, 2.8739E+07,
     +  2.8759E+07, 2.8779E+07, 2.8799E+07, 2.8819E+07, 2.8839E+07,
     + 2.8859E+07, 2.8878E+07, 2.8898E+07, 2.8918E+07, 2.8938E+07,
     +  2.8958E+07, 2.8978E+07, 2.8998E+07, 2.9018E+07, 2.9038E+07,
     + 2.9058E+07, 2.9078E+07, 2.9097E+07, 2.9117E+07, 2.9137E+07,
     +  2.9157E+07, 2.9177E+07, 2.9197E+07, 2.9217E+07, 2.9237E+07,
     + 2.9257E+07, 2.9277E+07, 2.9297E+07, 2.9317E+07, 2.9337E+07,
     +  2.9356E+07, 2.9376E+07, 2.9396E+07, 2.9416E+07, 2.9436E+07,
     + 2.9456E+07, 2.9476E+07, 2.9496E+07, 2.9516E+07, 2.9536E+07,
     +  2.9556E+07, 2.9576E+07, 2.9596E+07, 2.9616E+07, 2.9636E+07,
     + 2.9656E+07, 2.9675E+07, 2.9695E+07, 2.9715E+07, 2.9735E+07,
     +  2.9755E+07, 2.9775E+07, 2.9795E+07, 2.9815E+07, 2.9835E+07,
     + 2.9855E+07, 2.9875E+07, 2.9895E+07, 2.9915E+07, 2.9935E+07,
     +  2.9955E+07, 2.9975E+07, 2.9995E+07, 3.0015E+07, 3.0035E+07,
     + 3.0055E+07, 3.0075E+07, 3.0095E+07, 3.0115E+07, 3.0134E+07,
     +  3.0154E+07, 3.0174E+07, 3.0194E+07, 3.0214E+07, 3.0234E+07,
     + 3.0254E+07, 3.0274E+07, 3.0294E+07, 3.0314E+07, 3.0334E+07,
     +  3.0354E+07, 3.0374E+07, 3.0394E+07, 3.0414E+07, 3.0434E+07,
     + 3.0454E+07, 3.0474E+07, 3.0494E+07, 3.0514E+07, 3.0534E+07,
     +  3.0554E+07, 3.0574E+07, 3.0594E+07, 3.0614E+07, 3.0634E+07,
     + 3.0654E+07, 3.0674E+07, 3.0694E+07, 3.0714E+07, 3.0734E+07,
     +  3.0754E+07, 3.0774E+07, 3.0794E+07, 3.0814E+07, 3.0834E+07,
     + 3.0854E+07, 3.0874E+07, 3.0894E+07, 3.0914E+07, 3.0934E+07,
     +  3.0954E+07, 3.0974E+07, 3.0994E+07, 3.1014E+07, 3.1034E+07,
     + 3.1054E+07, 3.1074E+07, 3.1094E+07, 3.1114E+07, 3.1134E+07,
     +  3.1154E+07, 3.1174E+07, 3.1194E+07, 3.1214E+07, 3.1234E+07,
     + 3.1254E+07, 3.1274E+07, 3.1294E+07, 3.1314E+07, 3.1334E+07,
     +  3.1354E+07, 3.1374E+07, 3.1394E+07, 3.1414E+07, 3.1434E+07,
     + 3.1454E+07, 3.1474E+07, 3.1494E+07, 3.1514E+07, 3.1534E+07,
     +  3.1554E+07, 3.1574E+07, 3.1594E+07, 3.1614E+07, 3.1635E+07,
     + 3.1655E+07, 3.1675E+07, 3.1695E+07, 3.1715E+07, 3.1735E+07,
     +  3.1755E+07, 3.1775E+07, 3.1795E+07, 3.1815E+07, 3.1835E+07,
     + 3.1855E+07, 3.1875E+07, 3.1895E+07, 3.1915E+07, 3.1935E+07,
     +  3.1955E+07, 3.1975E+07, 3.1995E+07, 3.2015E+07, 3.2035E+07,
     + 3.2055E+07, 3.2075E+07, 3.2095E+07, 3.2115E+07, 3.2135E+07,
     +  3.2155E+07, 3.2175E+07, 3.2196E+07, 3.2216E+07, 3.2236E+07,
     + 3.2256E+07, 3.2276E+07, 3.2296E+07, 3.2316E+07, 3.2336E+07,
     +  3.2356E+07, 3.2376E+07, 3.2396E+07, 3.2416E+07, 3.2436E+07,
     + 3.2456E+07, 3.2476E+07, 3.2496E+07, 3.2516E+07, 3.2536E+07,
     +  3.2556E+07, 3.2576E+07, 3.2596E+07, 3.2617E+07, 3.2637E+07,
     + 3.2657E+07, 3.2677E+07, 3.2697E+07, 3.2717E+07, 3.2737E+07,
     +  3.2757E+07, 3.2777E+07, 3.2797E+07, 3.2817E+07, 3.2837E+07,
     + 3.2857E+07, 3.2877E+07, 3.2897E+07, 3.2917E+07, 3.2937E+07,
     +  3.2957E+07, 3.2978E+07, 3.2998E+07, 3.3018E+07, 3.3038E+07,
     + 3.3058E+07, 3.3078E+07, 3.3098E+07, 3.3118E+07, 3.3138E+07,
     +  3.3158E+07, 3.3178E+07, 3.3198E+07, 3.3218E+07, 3.3238E+07,
     + 3.3258E+07, 3.3278E+07, 3.3299E+07, 3.3319E+07, 3.3339E+07,
     +  3.3359E+07, 3.3379E+07, 3.3399E+07, 3.3419E+07, 3.3439E+07,
     + 3.3459E+07, 3.3479E+07, 3.3499E+07, 3.3519E+07, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_C2H2  (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 3),QofT( 3,NT),Q(NT)
      data (Tmx(I),I=1,3)/5000.,3500.,3500./ 
      data xgj/ 1.,8.,6./
c...     C2H2
c...        --      1221
      data (QofT( 1,J),J=1,2408)/ 1.3048E+00, 2.6878E+00, 4.1047E+00,
     +  5.3877E+00, 6.6009E+00, 7.7885E+00, 8.9684E+00, 1.0147E+01,
     + 1.1325E+01, 1.2504E+01, 1.3683E+01, 1.4863E+01, 1.6043E+01,
     +  1.7223E+01, 1.8403E+01, 1.9584E+01, 2.0764E+01, 2.1945E+01,
     + 2.3126E+01, 2.4307E+01, 2.6668E+01, 2.9030E+01, 3.1393E+01,
     +  3.3755E+01, 3.6117E+01, 3.8480E+01, 4.0843E+01, 4.3205E+01,
     + 4.5568E+01, 4.7931E+01, 5.0294E+01, 5.2656E+01, 5.5019E+01,
     +  5.7382E+01, 5.9745E+01, 6.2108E+01, 6.4471E+01, 6.6834E+01,
     + 6.9197E+01, 7.1560E+01, 7.3924E+01, 7.6287E+01, 7.8650E+01,
     +  8.1013E+01, 8.3377E+01, 8.5740E+01, 8.8104E+01, 9.0468E+01,
     + 9.2832E+01, 9.5196E+01, 9.7560E+01, 9.9925E+01, 1.0229E+02,
     +  1.0466E+02, 1.0702E+02, 1.0939E+02, 1.1176E+02, 1.1413E+02,
     + 1.1650E+02, 1.1887E+02, 1.2124E+02, 1.2362E+02, 1.2599E+02,
     +  1.2837E+02, 1.3075E+02, 1.3313E+02, 1.3552E+02, 1.3791E+02,
     + 1.4030E+02, 1.4269E+02, 1.4509E+02, 1.4749E+02, 1.4990E+02,
     +  1.5231E+02, 1.5473E+02, 1.5715E+02, 1.5958E+02, 1.6201E+02,
     + 1.6445E+02, 1.6690E+02, 1.6935E+02, 1.7182E+02, 1.7428E+02,
     +  1.7676E+02, 1.7925E+02, 1.8175E+02, 1.8425E+02, 1.8677E+02,
     + 1.8929E+02, 1.9183E+02, 1.9438E+02, 1.9694E+02, 1.9951E+02,
     +  2.0209E+02, 2.0469E+02, 2.0730E+02, 2.0992E+02, 2.1256E+02,
     + 2.1521E+02, 2.1788E+02, 2.2056E+02, 2.2326E+02, 2.2598E+02,
     +  2.2871E+02, 2.3146E+02, 2.3422E+02, 2.3701E+02, 2.3981E+02,
     + 2.4263E+02, 2.4548E+02, 2.4834E+02, 2.5122E+02, 2.5412E+02,
     +  2.5704E+02, 2.5999E+02, 2.6295E+02, 2.6594E+02, 2.6895E+02,
     + 2.7198E+02, 2.7504E+02, 2.7812E+02, 2.8123E+02, 2.8436E+02,
     +  2.8751E+02, 2.9069E+02, 2.9390E+02, 2.9713E+02, 3.0039E+02,
     + 3.0367E+02, 3.0699E+02, 3.1033E+02, 3.1370E+02, 3.1709E+02,
     +  3.2052E+02, 3.2398E+02, 3.2746E+02, 3.3098E+02, 3.3453E+02,
     + 3.3810E+02, 3.4171E+02, 3.4535E+02, 3.4902E+02, 3.5273E+02,
     +  3.5647E+02, 3.6024E+02, 3.6404E+02, 3.6788E+02, 3.7175E+02,
     + 3.7566E+02, 3.7960E+02, 3.8358E+02, 3.8759E+02, 3.9164E+02,
     +  3.9573E+02, 3.9985E+02, 4.0401E+02, 4.0821E+02, 4.1245E+02,
     + 4.1673E+02, 4.2104E+02, 4.2539E+02, 4.2979E+02, 4.3422E+02,
     +  4.3870E+02, 4.4321E+02, 4.4777E+02, 4.5236E+02, 4.5700E+02,
     + 4.6169E+02, 4.6641E+02, 4.7118E+02, 4.7599E+02, 4.8085E+02,
     +  4.8575E+02, 4.9069E+02, 4.9568E+02, 5.0072E+02, 5.0580E+02,
     + 5.1093E+02, 5.1610E+02, 5.2133E+02, 5.2660E+02, 5.3191E+02,
     +  5.3728E+02, 5.4270E+02, 5.4816E+02, 5.5368E+02, 5.5924E+02,
     + 5.6485E+02, 5.7052E+02, 5.7624E+02, 5.8201E+02, 5.8783E+02,
     +  5.9370E+02, 5.9963E+02, 6.0561E+02, 6.1164E+02, 6.1773E+02,
     + 6.2387E+02, 6.3007E+02, 6.3633E+02, 6.4264E+02, 6.4900E+02,
     +  6.5543E+02, 6.6191E+02, 6.6845E+02, 6.7505E+02, 6.8170E+02,
     + 6.8842E+02, 6.9519E+02, 7.0203E+02, 7.0892E+02, 7.1588E+02,
     +  7.2290E+02, 7.2998E+02, 7.3713E+02, 7.4433E+02, 7.5160E+02,
     + 7.5894E+02, 7.6633E+02, 7.7380E+02, 7.8133E+02, 7.8892E+02,
     +  7.9658E+02, 8.0431E+02, 8.1211E+02, 8.1997E+02, 8.2790E+02,
     + 8.3590E+02, 8.4397E+02, 8.5211E+02, 8.6032E+02, 8.6861E+02,
     +  8.7696E+02, 8.8538E+02, 8.9388E+02, 9.0245E+02, 9.1110E+02,
     + 9.1982E+02, 9.2861E+02, 9.3748E+02, 9.4642E+02, 9.5544E+02,
     +  9.6454E+02, 9.7372E+02, 9.8297E+02, 9.9230E+02, 1.0017E+03,
     + 1.0112E+03, 1.0208E+03, 1.0304E+03, 1.0402E+03, 1.0500E+03,
     +  1.0599E+03, 1.0699E+03, 1.0799E+03, 1.0901E+03, 1.1003E+03,
     + 1.1106E+03, 1.1210E+03, 1.1315E+03, 1.1421E+03, 1.1528E+03,
     +  1.1635E+03, 1.1744E+03, 1.1853E+03, 1.1963E+03, 1.2074E+03,
     + 1.2186E+03, 1.2299E+03, 1.2413E+03, 1.2528E+03, 1.2644E+03,
     +  1.2761E+03, 1.2879E+03, 1.2997E+03, 1.3117E+03, 1.3237E+03,
     + 1.3359E+03, 1.3482E+03, 1.3605E+03, 1.3730E+03, 1.3855E+03,
     +  1.3982E+03, 1.4110E+03, 1.4238E+03, 1.4368E+03, 1.4499E+03,
     + 1.4631E+03, 1.4764E+03, 1.4897E+03, 1.5032E+03, 1.5169E+03,
     +  1.5306E+03, 1.5444E+03, 1.5583E+03, 1.5724E+03, 1.5865E+03,
     + 1.6008E+03, 1.6152E+03, 1.6297E+03, 1.6443E+03, 1.6590E+03,
     +  1.6739E+03, 1.6888E+03, 1.7039E+03, 1.7191E+03, 1.7344E+03,
     + 1.7499E+03, 1.7654E+03, 1.7811E+03, 1.7969E+03, 1.8128E+03,
     +  1.8289E+03, 1.8450E+03, 1.8613E+03, 1.8778E+03, 1.8943E+03,
     + 1.9110E+03, 1.9278E+03, 1.9448E+03, 1.9618E+03, 1.9790E+03,
     +  1.9964E+03, 2.0138E+03, 2.0314E+03, 2.0492E+03, 2.0671E+03,
     + 2.0851E+03, 2.1032E+03, 2.1215E+03, 2.1399E+03, 2.1585E+03,
     +  2.1772E+03, 2.1961E+03, 2.2151E+03, 2.2342E+03, 2.2535E+03,
     + 2.2729E+03, 2.2925E+03, 2.3122E+03, 2.3321E+03, 2.3521E+03,
     +  2.3723E+03, 2.3926E+03, 2.4131E+03, 2.4337E+03, 2.4545E+03,
     + 2.4754E+03, 2.4965E+03, 2.5178E+03, 2.5392E+03, 2.5608E+03,
     +  2.5825E+03, 2.6044E+03, 2.6264E+03, 2.6487E+03, 2.6711E+03,
     + 2.6936E+03, 2.7163E+03, 2.7392E+03, 2.7623E+03, 2.7855E+03,
     +  2.8089E+03, 2.8324E+03, 2.8562E+03, 2.8801E+03, 2.9042E+03,
     + 2.9284E+03, 2.9529E+03, 2.9775E+03, 3.0023E+03, 3.0273E+03,
     +  3.0524E+03, 3.0778E+03, 3.1033E+03, 3.1290E+03, 3.1549E+03,
     + 3.1810E+03, 3.2073E+03, 3.2338E+03, 3.2604E+03, 3.2873E+03,
     +  3.3143E+03, 3.3416E+03, 3.3690E+03, 3.3966E+03, 3.4245E+03,
     + 3.4525E+03, 3.4807E+03, 3.5091E+03, 3.5378E+03, 3.5666E+03,
     +  3.5956E+03, 3.6249E+03, 3.6544E+03, 3.6840E+03, 3.7139E+03,
     + 3.7440E+03, 3.7743E+03, 3.8048E+03, 3.8355E+03, 3.8665E+03,
     +  3.8977E+03, 3.9290E+03, 3.9607E+03, 3.9925E+03, 4.0245E+03,
     + 4.0568E+03, 4.0893E+03, 4.1221E+03, 4.1550E+03, 4.1882E+03,
     +  4.2216E+03, 4.2553E+03, 4.2892E+03, 4.3233E+03, 4.3577E+03,
     + 4.3923E+03, 4.4271E+03, 4.4622E+03, 4.4975E+03, 4.5331E+03,
     +  4.5689E+03, 4.6050E+03, 4.6413E+03, 4.6779E+03, 4.7147E+03,
     + 4.7518E+03, 4.7891E+03, 4.8267E+03, 4.8645E+03, 4.9026E+03,
     +  4.9410E+03, 4.9796E+03, 5.0185E+03, 5.0577E+03, 5.0971E+03,
     + 5.1368E+03, 5.1767E+03, 5.2170E+03, 5.2575E+03, 5.2982E+03,
     +  5.3393E+03, 5.3806E+03, 5.4222E+03, 5.4641E+03, 5.5063E+03,
     + 5.5488E+03, 5.5915E+03, 5.6346E+03, 5.6779E+03, 5.7215E+03,
     +  5.7654E+03, 5.8096E+03, 5.8541E+03, 5.8989E+03, 5.9440E+03,
     + 5.9894E+03, 6.0351E+03, 6.0812E+03, 6.1275E+03, 6.1741E+03,
     +  6.2210E+03, 6.2683E+03, 6.3159E+03, 6.3637E+03, 6.4119E+03,
     + 6.4604E+03, 6.5093E+03, 6.5584E+03, 6.6079E+03, 6.6577E+03,
     +  6.7079E+03, 6.7583E+03, 6.8092E+03, 6.8603E+03, 6.9118E+03,
     + 6.9636E+03, 7.0157E+03, 7.0682E+03, 7.1211E+03, 7.1742E+03,
     +  7.2278E+03, 7.2817E+03, 7.3359E+03, 7.3905E+03, 7.4454E+03,
     + 7.5007E+03, 7.5564E+03, 7.6124E+03, 7.6688E+03, 7.7255E+03,
     +  7.7826E+03, 7.8401E+03, 7.8980E+03, 7.9562E+03, 8.0148E+03,
     + 8.0738E+03, 8.1332E+03, 8.1929E+03, 8.2531E+03, 8.3136E+03,
     +  8.3745E+03, 8.4358E+03, 8.4975E+03, 8.5596E+03, 8.6221E+03,
     + 8.6850E+03, 8.7482E+03, 8.8119E+03, 8.8760E+03, 8.9405E+03,
     +  9.0055E+03, 9.0708E+03, 9.1365E+03, 9.2027E+03, 9.2693E+03,
     + 9.3363E+03, 9.4037E+03, 9.4716E+03, 9.5398E+03, 9.6086E+03,
     +  9.6777E+03, 9.7473E+03, 9.8173E+03, 9.8878E+03, 9.9587E+03,
     + 1.0030E+04, 1.0102E+04, 1.0174E+04, 1.0247E+04, 1.0320E+04,
     +  1.0394E+04, 1.0468E+04, 1.0542E+04, 1.0617E+04, 1.0693E+04,
     + 1.0769E+04, 1.0845E+04, 1.0922E+04, 1.0999E+04, 1.1077E+04,
     +  1.1156E+04, 1.1234E+04, 1.1314E+04, 1.1393E+04, 1.1474E+04,
     + 1.1555E+04, 1.1636E+04, 1.1718E+04, 1.1800E+04, 1.1883E+04,
     +  1.1966E+04, 1.2050E+04, 1.2134E+04, 1.2219E+04, 1.2304E+04,
     + 1.2390E+04, 1.2477E+04, 1.2564E+04, 1.2651E+04, 1.2739E+04,
     +  1.2828E+04, 1.2917E+04, 1.3007E+04, 1.3097E+04, 1.3188E+04,
     + 1.3279E+04, 1.3371E+04, 1.3463E+04, 1.3556E+04, 1.3650E+04,
     +  1.3744E+04, 1.3838E+04, 1.3934E+04, 1.4029E+04, 1.4126E+04,
     + 1.4223E+04, 1.4320E+04, 1.4418E+04, 1.4517E+04, 1.4617E+04,
     +  1.4716E+04, 1.4817E+04, 1.4918E+04, 1.5020E+04, 1.5122E+04,
     + 1.5225E+04, 1.5329E+04, 1.5433E+04, 1.5538E+04, 1.5643E+04,
     +  1.5749E+04, 1.5856E+04, 1.5963E+04, 1.6071E+04, 1.6180E+04,
     + 1.6289E+04, 1.6399E+04, 1.6509E+04, 1.6621E+04, 1.6732E+04,
     +  1.6845E+04, 1.6958E+04, 1.7072E+04, 1.7186E+04, 1.7302E+04,
     + 1.7418E+04, 1.7534E+04, 1.7651E+04, 1.7769E+04, 1.7888E+04,
     +  1.8007E+04, 1.8127E+04, 1.8248E+04, 1.8369E+04, 1.8491E+04,
     + 1.8614E+04, 1.8738E+04, 1.8862E+04, 1.8987E+04, 1.9113E+04,
     +  1.9239E+04, 1.9366E+04, 1.9494E+04, 1.9623E+04, 1.9753E+04,
     + 1.9883E+04, 2.0014E+04, 2.0145E+04, 2.0278E+04, 2.0411E+04,
     +  2.0545E+04, 2.0680E+04, 2.0815E+04, 2.0951E+04, 2.1088E+04,
     + 2.1226E+04, 2.1365E+04, 2.1504E+04, 2.1645E+04, 2.1786E+04,
     +  2.1928E+04, 2.2070E+04, 2.2214E+04, 2.2358E+04, 2.2503E+04,
     + 2.2649E+04, 2.2796E+04, 2.2943E+04, 2.3092E+04, 2.3241E+04,
     +  2.3391E+04, 2.3542E+04, 2.3694E+04, 2.3847E+04, 2.4000E+04,
     + 2.4155E+04, 2.4310E+04, 2.4466E+04, 2.4623E+04, 2.4781E+04,
     +  2.4940E+04, 2.5100E+04, 2.5261E+04, 2.5422E+04, 2.5585E+04,
     + 2.5748E+04, 2.5912E+04, 2.6077E+04, 2.6243E+04, 2.6410E+04,
     +  2.6578E+04, 2.6747E+04, 2.6917E+04, 2.7088E+04, 2.7260E+04,
     + 2.7432E+04, 2.7606E+04, 2.7781E+04, 2.7956E+04, 2.8133E+04,
     +  2.8310E+04, 2.8489E+04, 2.8668E+04, 2.8849E+04, 2.9030E+04,
     + 2.9213E+04, 2.9396E+04, 2.9581E+04, 2.9766E+04, 2.9953E+04,
     +  3.0140E+04, 3.0329E+04, 3.0518E+04, 3.0709E+04, 3.0901E+04,
     + 3.1093E+04, 3.1287E+04, 3.1482E+04, 3.1678E+04, 3.1875E+04,
     +  3.2073E+04, 3.2272E+04, 3.2472E+04, 3.2674E+04, 3.2876E+04,
     + 3.3080E+04, 3.3284E+04, 3.3490E+04, 3.3697E+04, 3.3905E+04,
     +  3.4114E+04, 3.4324E+04, 3.4535E+04, 3.4748E+04, 3.4961E+04,
     + 3.5176E+04, 3.5392E+04, 3.5609E+04, 3.5827E+04, 3.6046E+04,
     +  3.6267E+04, 3.6489E+04, 3.6712E+04, 3.6936E+04, 3.7161E+04,
     + 3.7387E+04, 3.7615E+04, 3.7844E+04, 3.8074E+04, 3.8305E+04,
     +  3.8538E+04, 3.8772E+04, 3.9007E+04, 3.9243E+04, 3.9481E+04,
     + 3.9720E+04, 3.9960E+04, 4.0201E+04, 4.0444E+04, 4.0687E+04,
     +  4.0933E+04, 4.1179E+04, 4.1427E+04, 4.1676E+04, 4.1926E+04,
     + 4.2178E+04, 4.2431E+04, 4.2685E+04, 4.2941E+04, 4.3198E+04,
     +  4.3456E+04, 4.3716E+04, 4.3977E+04, 4.4239E+04, 4.4503E+04,
     + 4.4768E+04, 4.5034E+04, 4.5302E+04, 4.5571E+04, 4.5842E+04,
     +  4.6114E+04, 4.6387E+04, 4.6662E+04, 4.6938E+04, 4.7216E+04,
     + 4.7495E+04, 4.7776E+04, 4.8058E+04, 4.8341E+04, 4.8626E+04,
     +  4.8912E+04, 4.9200E+04, 4.9490E+04, 4.9780E+04, 5.0073E+04,
     + 5.0366E+04, 5.0661E+04, 5.0958E+04, 5.1256E+04, 5.1556E+04,
     +  5.1858E+04, 5.2160E+04, 5.2465E+04, 5.2771E+04, 5.3078E+04,
     + 5.3387E+04, 5.3698E+04, 5.4010E+04, 5.4323E+04, 5.4639E+04,
     +  5.4956E+04, 5.5274E+04, 5.5594E+04, 5.5916E+04, 5.6239E+04,
     + 5.6564E+04, 5.6890E+04, 5.7219E+04, 5.7548E+04, 5.7880E+04,
     +  5.8213E+04, 5.8548E+04, 5.8884E+04, 5.9222E+04, 5.9562E+04,
     + 5.9903E+04, 6.0246E+04, 6.0591E+04, 6.0938E+04, 6.1286E+04,
     +  6.1636E+04, 6.1988E+04, 6.2341E+04, 6.2697E+04, 6.3053E+04,
     + 6.3412E+04, 6.3773E+04, 6.4135E+04, 6.4499E+04, 6.4865E+04,
     +  6.5232E+04, 6.5602E+04, 6.5973E+04, 6.6346E+04, 6.6721E+04,
     + 6.7097E+04, 6.7476E+04, 6.7856E+04, 6.8238E+04, 6.8622E+04,
     +  6.9008E+04, 6.9396E+04, 6.9786E+04, 7.0177E+04, 7.0571E+04,
     + 7.0966E+04, 7.1363E+04, 7.1762E+04, 7.2163E+04, 7.2566E+04,
     +  7.2971E+04, 7.3378E+04, 7.3787E+04, 7.4198E+04, 7.4610E+04,
     + 7.5025E+04, 7.5442E+04, 7.5860E+04, 7.6281E+04, 7.6704E+04,
     +  7.7129E+04, 7.7555E+04, 7.7984E+04, 7.8415E+04, 7.8848E+04,
     + 7.9282E+04, 7.9719E+04, 8.0158E+04, 8.0599E+04, 8.1043E+04,
     +  8.1488E+04, 8.1935E+04, 8.2385E+04, 8.2836E+04, 8.3290E+04,
     + 8.3746E+04, 8.4203E+04, 8.4664E+04, 8.5126E+04, 8.5590E+04,
     +  8.6057E+04, 8.6525E+04, 8.6996E+04, 8.7469E+04, 8.7945E+04,
     + 8.8422E+04, 8.8902E+04, 8.9384E+04, 8.9868E+04, 9.0354E+04,
     +  9.0843E+04, 9.1334E+04, 9.1827E+04, 9.2323E+04, 9.2820E+04,
     + 9.3320E+04, 9.3823E+04, 9.4327E+04, 9.4834E+04, 9.5344E+04,
     +  9.5855E+04, 9.6369E+04, 9.6885E+04, 9.7404E+04, 9.7925E+04,
     + 9.8448E+04, 9.8974E+04, 9.9502E+04, 1.0003E+05, 1.0057E+05,
     +  1.0110E+05, 1.0164E+05, 1.0218E+05, 1.0272E+05, 1.0327E+05,
     + 1.0381E+05, 1.0436E+05, 1.0492E+05, 1.0547E+05, 1.0603E+05,
     +  1.0659E+05, 1.0715E+05, 1.0772E+05, 1.0828E+05, 1.0885E+05,
     + 1.0943E+05, 1.1000E+05, 1.1058E+05, 1.1116E+05, 1.1174E+05,
     +  1.1233E+05, 1.1292E+05, 1.1351E+05, 1.1410E+05, 1.1470E+05,
     + 1.1530E+05, 1.1590E+05, 1.1650E+05, 1.1711E+05, 1.1772E+05,
     +  1.1833E+05, 1.1894E+05, 1.1956E+05, 1.2018E+05, 1.2080E+05,
     + 1.2143E+05, 1.2206E+05, 1.2269E+05, 1.2332E+05, 1.2396E+05,
     +  1.2459E+05, 1.2524E+05, 1.2588E+05, 1.2653E+05, 1.2718E+05,
     + 1.2783E+05, 1.2849E+05, 1.2915E+05, 1.2981E+05, 1.3047E+05,
     +  1.3114E+05, 1.3181E+05, 1.3248E+05, 1.3316E+05, 1.3384E+05,
     + 1.3452E+05, 1.3520E+05, 1.3589E+05, 1.3658E+05, 1.3728E+05,
     +  1.3797E+05, 1.3867E+05, 1.3937E+05, 1.4008E+05, 1.4079E+05,
     + 1.4150E+05, 1.4221E+05, 1.4293E+05, 1.4365E+05, 1.4437E+05,
     +  1.4510E+05, 1.4583E+05, 1.4656E+05, 1.4730E+05, 1.4804E+05,
     + 1.4878E+05, 1.4952E+05, 1.5027E+05, 1.5102E+05, 1.5178E+05,
     +  1.5253E+05, 1.5329E+05, 1.5406E+05, 1.5482E+05, 1.5559E+05,
     + 1.5637E+05, 1.5714E+05, 1.5792E+05, 1.5871E+05, 1.5949E+05,
     +  1.6028E+05, 1.6107E+05, 1.6187E+05, 1.6267E+05, 1.6347E+05,
     + 1.6428E+05, 1.6509E+05, 1.6590E+05, 1.6671E+05, 1.6753E+05,
     +  1.6836E+05, 1.6918E+05, 1.7001E+05, 1.7084E+05, 1.7168E+05,
     + 1.7252E+05, 1.7336E+05, 1.7421E+05, 1.7506E+05, 1.7591E+05,
     +  1.7677E+05, 1.7763E+05, 1.7849E+05, 1.7935E+05, 1.8023E+05,
     + 1.8110E+05, 1.8198E+05, 1.8286E+05, 1.8374E+05, 1.8463E+05,
     +  1.8552E+05, 1.8641E+05, 1.8731E+05, 1.8821E+05, 1.8912E+05,
     + 1.9003E+05, 1.9094E+05, 1.9186E+05, 1.9278E+05, 1.9370E+05,
     +  1.9463E+05, 1.9556E+05, 1.9649E+05, 1.9743E+05, 1.9837E+05,
     + 1.9932E+05, 2.0027E+05, 2.0122E+05, 2.0218E+05, 2.0314E+05,
     +  2.0410E+05, 2.0507E+05, 2.0604E+05, 2.0702E+05, 2.0800E+05,
     + 2.0898E+05, 2.0997E+05, 2.1096E+05, 2.1195E+05, 2.1295E+05,
     +  2.1396E+05, 2.1496E+05, 2.1597E+05, 2.1699E+05, 2.1800E+05,
     + 2.1903E+05, 2.2005E+05, 2.2108E+05, 2.2211E+05, 2.2315E+05,
     +  2.2419E+05, 2.2524E+05, 2.2629E+05, 2.2734E+05, 2.2840E+05,
     + 2.2946E+05, 2.3053E+05, 2.3160E+05, 2.3267E+05, 2.3375E+05,
     +  2.3483E+05, 2.3591E+05, 2.3700E+05, 2.3810E+05, 2.3920E+05,
     + 2.4030E+05, 2.4140E+05, 2.4251E+05, 2.4363E+05, 2.4475E+05,
     +  2.4587E+05, 2.4700E+05, 2.4813E+05, 2.4927E+05, 2.5040E+05,
     + 2.5155E+05, 2.5270E+05, 2.5385E+05, 2.5501E+05, 2.5617E+05,
     +  2.5733E+05, 2.5850E+05, 2.5968E+05, 2.6085E+05, 2.6204E+05,
     + 2.6322E+05, 2.6441E+05, 2.6561E+05, 2.6681E+05, 2.6801E+05,
     +  2.6922E+05, 2.7044E+05, 2.7165E+05, 2.7288E+05, 2.7410E+05,
     + 2.7533E+05, 2.7657E+05, 2.7781E+05, 2.7905E+05, 2.8030E+05,
     +  2.8155E+05, 2.8281E+05, 2.8408E+05, 2.8534E+05, 2.8661E+05,
     + 2.8789E+05, 2.8917E+05, 2.9046E+05, 2.9174E+05, 2.9304E+05,
     +  2.9434E+05, 2.9564E+05, 2.9695E+05, 2.9826E+05, 2.9958E+05,
     + 3.0090E+05, 3.0223E+05, 3.0356E+05, 3.0490E+05, 3.0624E+05,
     +  3.0758E+05, 3.0893E+05, 3.1029E+05, 3.1165E+05, 3.1301E+05,
     + 3.1438E+05, 3.1576E+05, 3.1714E+05, 3.1852E+05, 3.1991E+05,
     +  3.2130E+05, 3.2270E+05, 3.2411E+05, 3.2551E+05, 3.2693E+05,
     + 3.2834E+05, 3.2977E+05, 3.3120E+05, 3.3263E+05, 3.3407E+05,
     +  3.3551E+05, 3.3696E+05, 3.3841E+05, 3.3987E+05, 3.4133E+05,
     + 3.4280E+05, 3.4427E+05, 3.4575E+05, 3.4723E+05, 3.4872E+05,
     +  3.5021E+05, 3.5171E+05, 3.5321E+05, 3.5472E+05, 3.5623E+05,
     + 3.5775E+05, 3.5927E+05, 3.6080E+05, 3.6234E+05, 3.6387E+05,
     +  3.6542E+05, 3.6697E+05, 3.6852E+05, 3.7008E+05, 3.7165E+05,
     + 3.7322E+05, 3.7479E+05, 3.7637E+05, 3.7796E+05, 3.7955E+05,
     +  3.8115E+05, 3.8275E+05, 3.8436E+05, 3.8597E+05, 3.8759E+05,
     + 3.8921E+05, 3.9084E+05, 3.9247E+05, 3.9411E+05, 3.9576E+05,
     +  3.9741E+05, 3.9906E+05, 4.0072E+05, 4.0239E+05, 4.0406E+05,
     + 4.0574E+05, 4.0742E+05, 4.0911E+05, 4.1080E+05, 4.1250E+05,
     +  4.1421E+05, 4.1592E+05, 4.1763E+05, 4.1935E+05, 4.2108E+05,
     + 4.2281E+05, 4.2455E+05, 4.2630E+05, 4.2805E+05, 4.2980E+05,
     +  4.3156E+05, 4.3333E+05, 4.3510E+05, 4.3688E+05, 4.3866E+05,
     + 4.4045E+05, 4.4224E+05, 4.4405E+05, 4.4585E+05, 4.4766E+05,
     +  4.4948E+05, 4.5130E+05, 4.5313E+05, 4.5497E+05, 4.5681E+05,
     + 4.5866E+05, 4.6051E+05, 4.6237E+05, 4.6423E+05, 4.6610E+05,
     +  4.6798E+05, 4.6986E+05, 4.7175E+05, 4.7364E+05, 4.7554E+05,
     + 4.7745E+05, 4.7936E+05, 4.8127E+05, 4.8320E+05, 4.8513E+05,
     +  4.8706E+05, 4.8900E+05, 4.9095E+05, 4.9290E+05, 4.9486E+05,
     + 4.9683E+05, 4.9880E+05, 5.0078E+05, 5.0276E+05, 5.0475E+05,
     +  5.0675E+05, 5.0875E+05, 5.1076E+05, 5.1277E+05, 5.1479E+05,
     + 5.1682E+05, 5.1885E+05, 5.2089E+05, 5.2293E+05, 5.2498E+05,
     +  5.2704E+05, 5.2910E+05, 5.3117E+05, 5.3325E+05, 5.3533E+05,
     + 5.3742E+05, 5.3952E+05, 5.4162E+05, 5.4373E+05, 5.4584E+05,
     +  5.4796E+05, 5.5009E+05, 5.5222E+05, 5.5436E+05, 5.5650E+05,
     + 5.5866E+05, 5.6081E+05, 5.6298E+05, 5.6515E+05, 5.6733E+05,
     +  5.6951E+05, 5.7170E+05, 5.7390E+05, 5.7610E+05, 5.7831E+05,
     + 5.8053E+05, 5.8275E+05, 5.8498E+05, 5.8722E+05, 5.8946E+05,
     +  5.9171E+05, 5.9397E+05, 5.9623E+05, 5.9850E+05, 6.0077E+05,
     + 6.0305E+05, 6.0534E+05, 6.0764E+05, 6.0994E+05, 6.1225E+05,
     +  6.1456E+05, 6.1689E+05, 6.1922E+05, 6.2155E+05, 6.2389E+05,
     + 6.2624E+05, 6.2860E+05, 6.3096E+05, 6.3333E+05, 6.3571E+05,
     +  6.3809E+05, 6.4048E+05, 6.4287E+05, 6.4528E+05, 6.4769E+05,
     + 6.5010E+05, 6.5253E+05, 6.5496E+05, 6.5740E+05, 6.5984E+05,
     +  6.6229E+05, 6.6475E+05, 6.6722E+05, 6.6969E+05, 6.7217E+05,
     + 6.7465E+05, 6.7715E+05, 6.7965E+05, 6.8215E+05, 6.8467E+05,
     +  6.8719E+05, 6.8972E+05, 6.9225E+05, 6.9479E+05, 6.9734E+05,
     + 6.9990E+05, 7.0246E+05, 7.0503E+05, 7.0761E+05, 7.1020E+05,
     +  7.1279E+05, 7.1539E+05, 7.1799E+05, 7.2061E+05, 7.2323E+05,
     + 7.2585E+05, 7.2849E+05, 7.3113E+05, 7.3378E+05, 7.3644E+05,
     +  7.3910E+05, 7.4177E+05, 7.4445E+05, 7.4713E+05, 7.4983E+05,
     + 7.5253E+05, 7.5523E+05, 7.5795E+05, 7.6067E+05, 7.6340E+05,
     +  7.6614E+05, 7.6888E+05, 7.7163E+05, 7.7439E+05, 7.7716E+05,
     + 7.7993E+05, 7.8271E+05, 7.8550E+05, 7.8830E+05, 7.9110E+05,
     +  7.9391E+05, 7.9673E+05, 7.9956E+05, 8.0239E+05, 8.0523E+05,
     + 8.0808E+05, 8.1094E+05, 8.1380E+05, 8.1667E+05, 8.1955E+05,
     +  8.2243E+05, 8.2533E+05, 8.2823E+05, 8.3114E+05, 8.3406E+05,
     + 8.3698E+05, 8.3991E+05, 8.4285E+05, 8.4580E+05, 8.4875E+05,
     +  8.5172E+05, 8.5469E+05, 8.5766E+05, 8.6065E+05, 8.6364E+05,
     + 8.6664E+05, 8.6965E+05, 8.7267E+05, 8.7569E+05, 8.7873E+05,
     +  8.8177E+05, 8.8481E+05, 8.8787E+05, 8.9093E+05, 8.9401E+05,
     + 8.9708E+05, 9.0017E+05, 9.0327E+05, 9.0637E+05, 9.0948E+05,
     +  9.1260E+05, 9.1573E+05, 9.1886E+05, 9.2200E+05, 9.2515E+05,
     + 9.2831E+05, 9.3148E+05, 9.3465E+05, 9.3783E+05, 9.4102E+05,
     +  9.4422E+05, 9.4743E+05, 9.5064E+05, 9.5387E+05, 9.5710E+05,
     + 9.6033E+05, 9.6358E+05, 9.6684E+05, 9.7010E+05, 9.7337E+05,
     +  9.7665E+05, 9.7994E+05, 9.8323E+05, 9.8653E+05, 9.8985E+05,
     + 9.9317E+05, 9.9649E+05, 9.9983E+05, 1.0032E+06, 1.0065E+06,
     +  1.0099E+06, 1.0133E+06, 1.0166E+06, 1.0200E+06, 1.0234E+06,
     + 1.0268E+06, 1.0302E+06, 1.0336E+06, 1.0371E+06, 1.0405E+06,
     +  1.0440E+06, 1.0474E+06, 1.0509E+06, 1.0543E+06, 1.0578E+06,
     + 1.0613E+06, 1.0648E+06, 1.0683E+06, 1.0718E+06, 1.0753E+06,
     +  1.0788E+06, 1.0824E+06, 1.0859E+06, 1.0895E+06, 1.0930E+06,
     + 1.0966E+06, 1.1002E+06, 1.1038E+06, 1.1074E+06, 1.1110E+06,
     +  1.1146E+06, 1.1182E+06, 1.1218E+06, 1.1255E+06, 1.1291E+06,
     + 1.1328E+06, 1.1364E+06, 1.1401E+06, 1.1438E+06, 1.1475E+06,
     +  1.1512E+06, 1.1549E+06, 1.1586E+06, 1.1623E+06, 1.1661E+06,
     + 1.1698E+06, 1.1736E+06, 1.1773E+06, 1.1811E+06, 1.1849E+06,
     +  1.1887E+06, 1.1924E+06, 1.1962E+06, 1.2001E+06, 1.2039E+06,
     + 1.2077E+06, 1.2115E+06, 1.2154E+06, 1.2193E+06, 1.2231E+06,
     +  1.2270E+06, 1.2309E+06, 1.2348E+06, 1.2387E+06, 1.2426E+06,
     + 1.2465E+06, 1.2504E+06, 1.2544E+06, 1.2583E+06, 1.2622E+06,
     +  1.2662E+06, 1.2702E+06, 1.2742E+06, 1.2782E+06, 1.2822E+06,
     + 1.2862E+06, 1.2902E+06, 1.2942E+06, 1.2982E+06, 1.3023E+06,
     +  1.3063E+06, 1.3104E+06, 1.3145E+06, 1.3185E+06, 1.3226E+06,
     + 1.3267E+06, 1.3308E+06, 1.3350E+06, 1.3391E+06, 1.3432E+06,
     +  1.3474E+06, 1.3515E+06, 1.3557E+06, 1.3598E+06, 1.3640E+06,
     + 1.3682E+06, 1.3724E+06, 1.3766E+06, 1.3808E+06, 1.3851E+06,
     +  1.3893E+06, 1.3935E+06, 1.3978E+06, 1.4021E+06, 1.4063E+06,
     + 1.4106E+06, 1.4149E+06, 1.4192E+06, 1.4235E+06, 1.4278E+06,
     +  1.4322E+06, 1.4365E+06, 1.4408E+06, 1.4452E+06, 1.4496E+06,
     + 1.4539E+06, 1.4583E+06, 1.4627E+06, 1.4671E+06, 1.4715E+06,
     +  1.4760E+06, 1.4804E+06, 1.4848E+06, 1.4893E+06, 1.4937E+06,
     + 1.4982E+06, 1.5027E+06, 1.5072E+06, 1.5117E+06, 1.5162E+06,
     +  1.5207E+06, 1.5252E+06, 1.5298E+06, 1.5343E+06, 1.5389E+06,
     + 1.5434E+06, 1.5480E+06, 1.5526E+06, 1.5572E+06, 1.5618E+06,
     +  1.5664E+06, 1.5710E+06, 1.5757E+06, 1.5803E+06, 1.5849E+06,
     + 1.5896E+06, 1.5943E+06, 1.5990E+06, 1.6036E+06, 1.6083E+06,
     +  1.6131E+06, 1.6178E+06, 1.6225E+06, 1.6272E+06, 1.6320E+06,
     + 1.6367E+06, 1.6415E+06, 1.6463E+06, 1.6511E+06, 1.6559E+06,
     +  1.6607E+06, 1.6655E+06, 1.6703E+06, 1.6752E+06, 1.6800E+06,
     + 1.6849E+06, 1.6897E+06, 1.6946E+06, 1.6995E+06, 1.7044E+06,
     +  1.7093E+06, 1.7142E+06, 1.7191E+06, 1.7241E+06, 1.7290E+06,
     + 1.7340E+06, 1.7389E+06, 1.7439E+06, 1.7489E+06, 1.7539E+06,
     +  1.7589E+06, 1.7639E+06, 1.7689E+06, 1.7740E+06, 1.7790E+06,
     + 1.7841E+06, 1.7891E+06, 1.7942E+06, 1.7993E+06, 1.8044E+06,
     +  1.8095E+06, 1.8146E+06, 1.8197E+06, 1.8249E+06, 1.8300E+06,
     + 1.8352E+06, 1.8403E+06, 1.8455E+06, 1.8507E+06, 1.8559E+06,
     +  1.8611E+06, 1.8663E+06, 1.8715E+06, 1.8768E+06, 1.8820E+06,
     + 1.8873E+06, 1.8925E+06, 1.8978E+06, 1.9031E+06, 1.9084E+06,
     +  1.9137E+06, 1.9190E+06, 1.9243E+06, 1.9297E+06, 1.9350E+06,
     + 1.9404E+06, 1.9458E+06, 1.9511E+06, 1.9565E+06, 1.9619E+06,
     +  1.9673E+06, 1.9728E+06, 1.9782E+06, 1.9836E+06, 1.9891E+06,
     + 1.9945E+06, 2.0000E+06, 2.0055E+06, 2.0110E+06, 2.0165E+06,
     +  2.0220E+06, 2.0275E+06, 2.0331E+06, 2.0386E+06, 2.0442E+06,
     + 2.0497E+06, 2.0553E+06, 2.0609E+06, 2.0665E+06, 2.0721E+06,
     +  2.0777E+06, 2.0833E+06, 2.0890E+06, 2.0946E+06, 2.1003E+06,
     + 2.1059E+06, 2.1116E+06, 2.1173E+06, 2.1230E+06, 2.1287E+06,
     +  2.1345E+06, 2.1402E+06, 2.1459E+06, 2.1517E+06, 2.1574E+06,
     + 2.1632E+06, 2.1690E+06, 2.1748E+06, 2.1806E+06, 2.1864E+06,
     +  2.1923E+06, 2.1981E+06, 2.2039E+06, 2.2098E+06, 2.2157E+06,
     + 2.2216E+06, 2.2275E+06, 2.2334E+06, 2.2393E+06, 2.2452E+06,
     +  2.2511E+06, 2.2571E+06, 2.2630E+06, 2.2690E+06, 2.2750E+06,
     + 2.2810E+06, 2.2870E+06, 2.2930E+06, 2.2990E+06, 2.3050E+06,
     +  2.3111E+06, 2.3171E+06, 2.3232E+06, 2.3292E+06, 2.3353E+06,
     + 2.3414E+06, 2.3475E+06, 2.3537E+06, 2.3598E+06, 2.3659E+06,
     +  2.3721E+06, 2.3782E+06, 2.3844E+06, 2.3906E+06, 2.3968E+06,
     + 2.4030E+06, 2.4092E+06, 2.4154E+06, 2.4217E+06, 2.4279E+06,
     +  2.4342E+06, 2.4404E+06, 2.4467E+06, 2.4530E+06, 2.4593E+06,
     + 2.4656E+06, 2.4719E+06, 2.4783E+06, 2.4846E+06, 2.4910E+06,
     +  2.4973E+06, 2.5037E+06, 2.5101E+06, 2.5165E+06, 2.5229E+06,
     + 2.5294E+06, 2.5358E+06, 2.5422E+06, 2.5487E+06, 2.5552E+06,
     +  2.5616E+06, 2.5681E+06, 2.5746E+06, 2.5811E+06, 2.5877E+06,
     + 2.5942E+06, 2.6007E+06, 2.6073E+06, 2.6138E+06, 2.6204E+06,
     +  2.6270E+06, 2.6336E+06, 2.6402E+06, 2.6468E+06, 2.6535E+06,
     + 2.6601E+06, 2.6668E+06, 2.6734E+06, 2.6801E+06, 2.6868E+06,
     +  2.6935E+06, 2.7002E+06, 2.7069E+06, 2.7137E+06, 2.7204E+06,
     + 2.7272E+06, 2.7339E+06, 2.7407E+06, 2.7475E+06, 2.7543E+06,
     +  2.7611E+06, 2.7679E+06, 2.7748E+06, 2.7816E+06, 2.7885E+06,
     + 2.7953E+06, 2.8022E+06, 2.8091E+06, 2.8160E+06, 2.8229E+06,
     +  2.8299E+06, 2.8368E+06, 2.8437E+06, 2.8507E+06, 2.8577E+06,
     + 2.8646E+06, 2.8716E+06, 2.8786E+06, 2.8857E+06, 2.8927E+06,
     +  2.8997E+06, 2.9068E+06, 2.9138E+06, 2.9209E+06, 2.9280E+06,
     + 2.9351E+06, 2.9422E+06, 2.9493E+06, 2.9564E+06, 2.9636E+06,
     +  2.9707E+06, 2.9779E+06, 2.9851E+06, 2.9922E+06, 2.9994E+06,
     + 3.0066E+06, 3.0139E+06, 3.0211E+06, 3.0283E+06, 3.0356E+06,
     +  3.0429E+06, 3.0501E+06, 3.0574E+06, 3.0647E+06, 3.0720E+06,
     + 3.0794E+06, 3.0867E+06, 3.0940E+06, 3.1014E+06, 3.1088E+06,
     +  3.1161E+06, 3.1235E+06, 3.1309E+06, 3.1384E+06, 3.1458E+06,
     + 3.1532E+06, 3.1607E+06, 3.1681E+06, 3.1756E+06, 3.1831E+06,
     +  3.1906E+06, 3.1981E+06, 3.2056E+06, 3.2131E+06, 3.2207E+06,
     + 3.2282E+06, 3.2358E+06, 3.2434E+06, 3.2510E+06, 3.2586E+06,
     +  3.2662E+06, 3.2738E+06, 3.2814E+06, 3.2891E+06, 3.2967E+06,
     + 3.3044E+06, 3.3121E+06, 3.3198E+06, 3.3275E+06, 3.3352E+06,
     +  3.3429E+06, 3.3507E+06, 3.3584E+06, 3.3662E+06, 3.3740E+06,
     + 3.3817E+06, 3.3895E+06, 3.3974E+06, 3.4052E+06, 3.4130E+06,
     +  3.4209E+06, 3.4287E+06, 3.4366E+06, 3.4445E+06, 3.4524E+06,
     + 3.4603E+06, 3.4682E+06, 3.4761E+06, 3.4840E+06, 3.4920E+06,
     +  3.4999E+06, 3.5079E+06, 3.5159E+06, 3.5239E+06, 3.5319E+06,
     + 3.5399E+06, 3.5480E+06, 3.5560E+06, 3.5641E+06, 3.5721E+06,
     +  3.5802E+06, 3.5883E+06, 3.5964E+06, 3.6045E+06, 3.6127E+06,
     + 3.6208E+06, 3.6290E+06, 3.6371E+06, 3.6453E+06, 3.6535E+06,
     +  3.6617E+06, 3.6699E+06, 3.6781E+06, 3.6863E+06, 3.6946E+06,
     + 3.7029E+06, 3.7111E+06, 3.7194E+06, 3.7277E+06, 3.7360E+06,
     +  3.7443E+06, 3.7527E+06, 3.7610E+06, 3.7693E+06, 3.7777E+06,
     + 3.7861E+06, 3.7945E+06, 3.8029E+06, 3.8113E+06, 3.8197E+06,
     +  3.8282E+06, 3.8366E+06, 3.8451E+06, 3.8535E+06, 3.8620E+06,
     + 3.8705E+06, 3.8790E+06, 3.8875E+06, 3.8961E+06, 3.9046E+06,
     +  3.9132E+06, 3.9218E+06, 3.9303E+06, 3.9389E+06, 3.9475E+06,
     + 3.9561E+06, 3.9648E+06, 3.9734E+06, 3.9821E+06, 3.9907E+06,
     +  3.9994E+06, 4.0081E+06, 4.0168E+06, 4.0255E+06, 4.0342E+06,
     + 4.0430E+06, 4.0517E+06, 4.0605E+06, 4.0693E+06, 4.0780E+06,
     +  4.0868E+06, 4.0957E+06, 4.1045E+06, 4.1133E+06, 4.1222E+06,
     + 4.1310E+06, 4.1399E+06, 4.1488E+06, 4.1577E+06, 4.1666E+06,
     +  4.1755E+06, 4.1844E+06, 4.1934E+06, 4.2023E+06, 4.2113E+06,
     + 4.2203E+06, 4.2292E+06, 4.2382E+06, 4.2473E+06, 4.2563E+06,
     +  4.2653E+06, 4.2744E+06, 4.2834E+06, 4.2925E+06, 4.3016E+06,
     + 4.3107E+06, 4.3198E+06, 4.3289E+06, 4.3381E+06, 4.3472E+06,
     +  4.3564E+06, 4.3656E+06, 4.3747E+06, 4.3839E+06, 4.3932E+06,
     + 4.4024E+06, 4.4116E+06, 4.4209E+06, 4.4301E+06, 4.4394E+06,
     +  4.4487E+06, 4.4580E+06, 4.4673E+06, 4.4766E+06, 4.4859E+06,
     + 4.4953E+06, 4.5046E+06, 4.5140E+06, 4.5234E+06, 4.5328E+06,
     +  4.5422E+06, 4.5516E+06, 4.5610E+06, 4.5705E+06, 4.5799E+06,
     + 4.5894E+06, 4.5989E+06, 4.6084E+06, 4.6179E+06, 4.6274E+06,
     +  4.6369E+06, 4.6464E+06, 4.6560E+06, 4.6656E+06, 4.6751E+06,
     + 4.6847E+06, 4.6943E+06, 4.7039E+06, 4.7136E+06, 4.7232E+06,
     +  4.7328E+06, 4.7425E+06, 4.7522E+06, 4.7619E+06, 4.7716E+06,
     + 4.7813E+06, 4.7910E+06, 4.8007E+06, 4.8105E+06, 4.8203E+06,
     +  4.8300E+06, 4.8398E+06, 4.8496E+06, 4.8594E+06, 4.8692E+06,
     + 4.8791E+06, 4.8889E+06, 4.8988E+06, 4.9087E+06, 4.9185E+06,
     +  4.9284E+06, 4.9384E+06, 4.9483E+06, 4.9582E+06, 4.9682E+06,
     + 4.9781E+06, 4.9881E+06, 4.9981E+06, 5.0081E+06, 5.0181E+06,
     +  5.0281E+06, 5.0381E+06, 5.0482E+06, 5.0582E+06, 5.0683E+06,
     + 5.0784E+06, 5.0885E+06, 5.0986E+06, 5.1087E+06, 5.1188E+06,
     +  5.1290E+06, 5.1391E+06, 5.1493E+06, 5.1595E+06, 5.1697E+06,
     + 5.1799E+06, 5.1901E+06, 5.2003E+06, 5.2106E+06, 5.2208E+06,
     +  5.2311E+06, 5.2414E+06, 5.2517E+06, 5.2620E+06, 5.2723E+06,
     + 5.2826E+06, 5.2930E+06, 5.3033E+06, 5.3137E+06, 5.3241E+06,
     +  5.3345E+06, 5.3449E+06, 5.3553E+06, 5.3657E+06, 5.3762E+06,
     + 5.3866E+06, 5.3971E+06, 5.4076E+06, 5.4181E+06, 5.4286E+06,
     +  5.4391E+06, 5.4496E+06, 5.4602E+06, 5.4707E+06, 5.4813E+06,
     + 5.4919E+06, 5.5024E+06, 5.5131E+06, 5.5237E+06, 5.5343E+06,
     +  5.5449E+06, 5.5556E+06, 5.5663E+06, 5.5769E+06, 5.5876E+06,
     + 5.5983E+06, 5.6091E+06, 5.6198E+06, 5.6305E+06, 5.6413E+06,
     +  5.6520E+06, 5.6628E+06, 5.6736E+06, 5.6844E+06, 5.6952E+06,
     + 5.7061E+06, 5.7169E+06, 5.7278E+06, 5.7386E+06, 5.7495E+06,
     +  5.7604E+06, 5.7713E+06, 5.7822E+06, 5.7932E+06, 5.8041E+06,
     + 5.8151E+06, 5.8260E+06, 5.8370E+06, 5.8480E+06, 5.8590E+06,
     +  5.8700E+06, 5.8810E+06, 5.8921E+06, 5.9031E+06, 5.9142E+06,
     + 5.9253E+06, 5.9364E+06, 5.9475E+06, 5.9586E+06, 5.9697E+06,
     +  5.9809E+06, 5.9920E+06, 6.0032E+06, 6.0144E+06, 6.0256E+06,
     + 6.0368E+06, 6.0480E+06, 6.0592E+06, 6.0705E+06, 6.0817E+06,
     +  6.0930E+06, 6.1043E+06, 6.1156E+06, 6.1269E+06, 6.1382E+06,
     + 6.1495E+06, 6.1609E+06, 6.1722E+06, 6.1836E+06, 6.1950E+06,
     +  6.2064E+06, 6.2178E+06, 6.2292E+06, 6.2406E+06, 6.2520E+06,
     + 6.2635E+06, 6.2750E+06, 6.2865E+06, 6.2979E+06, 6.3095E+06,
     +  6.3210E+06, 6.3325E+06, 6.3440E+06, 6.3556E+06, 6.3672E+06,
     + 6.3788E+06, 6.3904E+06, 6.4020E+06, 6.4136E+06, 6.4252E+06,
     +  6.4369E+06, 6.4485E+06, 6.4602E+06, 6.4719E+06, 6.4836E+06,
     + 6.4953E+06, 6.5070E+06, 6.5187E+06, 6.5305E+06, 6.5422E+06,
     +  6.5540E+06, 6.5658E+06, 6.5776E+06, 6.5894E+06, 6.6012E+06,
     + 6.6130E+06, 6.6249E+06, 6.6367E+06, 6.6486E+06, 6.6605E+06,
     +  6.6724E+06, 6.6843E+06, 6.6962E+06, 6.7082E+06, 6.7201E+06,
     + 6.7321E+06, 6.7440E+06, 6.7560E+06, 6.7680E+06, 6.7800E+06,
     +  6.7920E+06, 6.8041E+06, 6.8161E+06, 6.8282E+06, 6.8403E+06,
     + 6.8523E+06, 6.8644E+06, 6.8766E+06, 6.8887E+06, 6.9008E+06,
     +  6.9130E+06, 6.9251E+06, 6.9373E+06, 6.9495E+06, 6.9617E+06,
     + 6.9739E+06, 6.9861E+06, 6.9984E+06, 7.0106E+06, 7.0229E+06,
     +  7.0351E+06, 7.0474E+06, 7.0597E+06, 7.0720E+06, 7.0844E+06,
     + 7.0967E+06, 7.1090E+06, 7.1214E+06, 7.1338E+06, 7.1462E+06,
     +  7.1586E+06, 7.1710E+06, 7.1834E+06, 7.1958E+06, 7.2083E+06,
     + 7.2208E+06, 7.2332E+06, 7.2457E+06, 7.2582E+06, 7.2707E+06,
     +  7.2833E+06, 7.2958E+06, 7.3084E+06, 7.3209E+06, 7.3335E+06,
     + 7.3461E+06, 7.3587E+06, 7.3713E+06, 7.3839E+06, 7.3966E+06,
     +  7.4092E+06, 7.4219E+06, 7.4346E+06, 7.4473E+06, 7.4600E+06,
     + 7.4727E+06, 7.4854E+06, 7.4981E+06, 7.5109E+06, 7.5237E+06,
     +  7.5364E+06, 7.5492E+06, 7.5620E+06, 7.5749E+06, 7.5877E+06,
     + 7.6005E+06, 7.6134E+06, 7.6262E+06, 7.6391E+06, 7.6520E+06,
     +  7.6649E+06, 7.6778E+06, 7.6908E+06, 7.7037E+06, 7.7167E+06,
     + 7.7296E+06, 7.7426E+06, 7.7556E+06, 7.7686E+06, 7.7816E+06,
     +  7.7947E+06, 7.8077E+06, 7.8208E+06, 7.8338E+06, 7.8469E+06,
     + 7.8600E+06, 7.8731E+06, 7.8862E+06, 7.8994E+06, 7.9125E+06,
     +  7.9257E+06, 7.9388E+06, 7.9520E+06, 7.9652E+06, 7.9784E+06,
     + 7.9916E+06, 8.0049E+06, 8.0181E+06, 8.0314E+06, 8.0446E+06,
     +  8.0579E+06, 8.0712E+06, 8.0845E+06, 8.0978E+06, 8.1112E+06,
     + 8.1245E+06, 8.1379E+06, 8.1512E+06, 8.1646E+06, 8.1780E+06,
     +  8.1914E+06, 8.2048E+06, 8.2183E+06, 8.2317E+06, 8.2452E+06,
     + 8.2587E+06, 8.2721E+06, 8.2856E+06, 8.2991E+06, 8.3127E+06,
     +  8.3262E+06, 8.3397E+06, 8.3533E+06, 8.3669E+06, 8.3805E+06,
     + 8.3941E+06, 8.4077E+06, 8.4213E+06, 8.4349E+06, 8.4486E+06,
     +  8.4622E+06, 8.4759E+06, 8.4896E+06, 8.5033E+06, 8.5170E+06,
     + 8.5307E+06, 8.5444E+06, 8.5582E+06, 8.5720E+06, 8.5857E+06,
     +  8.5995E+06, 8.6133E+06, 8.6271E+06, 8.6409E+06, 8.6548E+06,
     + 8.6686E+06, 8.6825E+06, 8.6964E+06, 8.7102E+06, 8.7241E+06,
     +  8.7381E+06, 8.7520E+06, 8.7659E+06, 8.7799E+06, 8.7938E+06,
     + 8.8078E+06, 8.8218E+06, 8.8358E+06, 8.8498E+06, 8.8638E+06,
     +  8.8778E+06, 8.8919E+06, 8.9060E+06, 8.9200E+06, 8.9341E+06,
     + 8.9482E+06, 8.9623E+06, 8.9764E+06, 8.9906E+06, 9.0047E+06,
     +  9.0189E+06, 9.0331E+06, 9.0473E+06, 9.0615E+06, 9.0757E+06,
     + 9.0899E+06, 9.1041E+06, 9.1184E+06, 9.1326E+06, 9.1469E+06,
     +  9.1612E+06, 9.1755E+06, 9.1898E+06, 9.2041E+06, 9.2185E+06/
      data (QofT( 1,J),J=2409,4510)/ 9.2328E+06, 9.2472E+06,
     +  9.2615E+06, 9.2759E+06, 9.2903E+06, 9.3047E+06, 9.3192E+06,
     +  9.3336E+06,
     + 9.3481E+06, 9.3625E+06, 9.3770E+06, 9.3915E+06, 9.4060E+06,
     +  9.4205E+06, 9.4350E+06, 9.4495E+06, 9.4641E+06, 9.4787E+06,
     + 9.4932E+06, 9.5078E+06, 9.5224E+06, 9.5370E+06, 9.5516E+06,
     +  9.5663E+06, 9.5809E+06, 9.5956E+06, 9.6103E+06, 9.6249E+06,
     + 9.6396E+06, 9.6544E+06, 9.6691E+06, 9.6838E+06, 9.6986E+06,
     +  9.7133E+06, 9.7281E+06, 9.7429E+06, 9.7577E+06, 9.7725E+06,
     + 9.7873E+06, 9.8021E+06, 9.8170E+06, 9.8319E+06, 9.8467E+06,
     +  9.8616E+06, 9.8765E+06, 9.8914E+06, 9.9063E+06, 9.9213E+06,
     + 9.9362E+06, 9.9512E+06, 9.9662E+06, 9.9811E+06, 9.9961E+06,
     +  1.0011E+07, 1.0026E+07, 1.0041E+07, 1.0056E+07, 1.0071E+07,
     + 1.0086E+07, 1.0101E+07, 1.0117E+07, 1.0132E+07, 1.0147E+07,
     +  1.0162E+07, 1.0177E+07, 1.0192E+07, 1.0207E+07, 1.0223E+07,
     + 1.0238E+07, 1.0253E+07, 1.0268E+07, 1.0283E+07, 1.0299E+07,
     +  1.0314E+07, 1.0329E+07, 1.0344E+07, 1.0360E+07, 1.0375E+07,
     + 1.0390E+07, 1.0406E+07, 1.0421E+07, 1.0436E+07, 1.0452E+07,
     +  1.0467E+07, 1.0483E+07, 1.0498E+07, 1.0513E+07, 1.0529E+07,
     + 1.0544E+07, 1.0560E+07, 1.0575E+07, 1.0591E+07, 1.0606E+07,
     +  1.0622E+07, 1.0637E+07, 1.0653E+07, 1.0668E+07, 1.0684E+07,
     + 1.0699E+07, 1.0715E+07, 1.0731E+07, 1.0746E+07, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --      1231
      data (QofT( 2,J),J=1,2408)/ 8.8145E+00, 1.2669E+01, 1.7185E+01,
     +  2.1819E+01, 2.6493E+01, 3.1185E+01, 3.5888E+01, 4.0597E+01,
     + 4.5310E+01, 5.0026E+01, 5.4744E+01, 5.9464E+01, 6.4184E+01,
     +  6.8906E+01, 7.3628E+01, 7.8351E+01, 8.3074E+01, 8.7798E+01,
     + 9.2522E+01, 9.7246E+01, 1.0670E+02, 1.1615E+02, 1.2560E+02,
     +  1.3505E+02, 1.4450E+02, 1.5395E+02, 1.6340E+02, 1.7286E+02,
     + 1.8231E+02, 1.9176E+02, 2.0122E+02, 2.1067E+02, 2.2012E+02,
     +  2.2958E+02, 2.3903E+02, 2.4861E+02, 2.5808E+02, 2.6754E+02,
     + 2.7701E+02, 2.8647E+02, 2.9594E+02, 3.0541E+02, 3.1487E+02,
     +  3.2434E+02, 3.3381E+02, 3.4328E+02, 3.5275E+02, 3.6222E+02,
     + 3.7170E+02, 3.8117E+02, 3.9064E+02, 4.0012E+02, 4.0960E+02,
     +  4.1908E+02, 4.2857E+02, 4.3806E+02, 4.4755E+02, 4.5704E+02,
     + 4.6655E+02, 4.7605E+02, 4.8556E+02, 4.9508E+02, 5.0461E+02,
     +  5.1414E+02, 5.2369E+02, 5.3324E+02, 5.4281E+02, 5.5238E+02,
     + 5.6197E+02, 5.7158E+02, 5.8120E+02, 5.9083E+02, 6.0049E+02,
     +  6.1016E+02, 6.1986E+02, 6.2958E+02, 6.3931E+02, 6.4908E+02,
     + 6.5887E+02, 6.6869E+02, 6.7854E+02, 6.8841E+02, 6.9832E+02,
     +  7.0827E+02, 7.1825E+02, 7.2827E+02, 7.3832E+02, 7.4842E+02,
     + 7.5855E+02, 7.6874E+02, 7.7896E+02, 7.8923E+02, 7.9956E+02,
     +  8.0993E+02, 8.2036E+02, 8.3084E+02, 8.4137E+02, 8.5197E+02,
     + 8.6262E+02, 8.7334E+02, 8.8411E+02, 8.9496E+02, 9.0586E+02,
     +  9.1683E+02, 9.2788E+02, 9.3900E+02, 9.5018E+02, 9.6144E+02,
     + 9.7278E+02, 9.8420E+02, 9.9570E+02, 1.0073E+03, 1.0189E+03,
     +  1.0307E+03, 1.0425E+03, 1.0544E+03, 1.0664E+03, 1.0786E+03,
     + 1.0908E+03, 1.1030E+03, 1.1154E+03, 1.1279E+03, 1.1405E+03,
     +  1.1532E+03, 1.1660E+03, 1.1789E+03, 1.1919E+03, 1.2050E+03,
     + 1.2182E+03, 1.2315E+03, 1.2450E+03, 1.2585E+03, 1.2722E+03,
     +  1.2860E+03, 1.2999E+03, 1.3139E+03, 1.3281E+03, 1.3423E+03,
     + 1.3567E+03, 1.3713E+03, 1.3859E+03, 1.4007E+03, 1.4156E+03,
     +  1.4307E+03, 1.4458E+03, 1.4612E+03, 1.4766E+03, 1.4922E+03,
     + 1.5080E+03, 1.5238E+03, 1.5398E+03, 1.5560E+03, 1.5723E+03,
     +  1.5888E+03, 1.6054E+03, 1.6222E+03, 1.6391E+03, 1.6562E+03,
     + 1.6734E+03, 1.6908E+03, 1.7083E+03, 1.7261E+03, 1.7439E+03,
     +  1.7620E+03, 1.7802E+03, 1.7985E+03, 1.8171E+03, 1.8358E+03,
     + 1.8547E+03, 1.8737E+03, 1.8929E+03, 1.9124E+03, 1.9320E+03,
     +  1.9517E+03, 1.9717E+03, 1.9918E+03, 2.0121E+03, 2.0326E+03,
     + 2.0533E+03, 2.0742E+03, 2.0953E+03, 2.1166E+03, 2.1380E+03,
     +  2.1597E+03, 2.1815E+03, 2.2036E+03, 2.2259E+03, 2.2483E+03,
     + 2.2710E+03, 2.2939E+03, 2.3170E+03, 2.3403E+03, 2.3638E+03,
     +  2.3875E+03, 2.4114E+03, 2.4356E+03, 2.4600E+03, 2.4846E+03,
     + 2.5094E+03, 2.5344E+03, 2.5597E+03, 2.5852E+03, 2.6110E+03,
     +  2.6369E+03, 2.6631E+03, 2.6895E+03, 2.7162E+03, 2.7431E+03,
     + 2.7703E+03, 2.7977E+03, 2.8253E+03, 2.8532E+03, 2.8813E+03,
     +  2.9097E+03, 2.9384E+03, 2.9673E+03, 2.9964E+03, 3.0258E+03,
     + 3.0555E+03, 3.0854E+03, 3.1156E+03, 3.1461E+03, 3.1768E+03,
     +  3.2078E+03, 3.2391E+03, 3.2706E+03, 3.3025E+03, 3.3346E+03,
     + 3.3670E+03, 3.3996E+03, 3.4326E+03, 3.4658E+03, 3.4993E+03,
     +  3.5332E+03, 3.5673E+03, 3.6017E+03, 3.6364E+03, 3.6714E+03,
     + 3.7068E+03, 3.7424E+03, 3.7783E+03, 3.8145E+03, 3.8511E+03,
     +  3.8879E+03, 3.9251E+03, 3.9626E+03, 4.0004E+03, 4.0386E+03,
     + 4.0770E+03, 4.1158E+03, 4.1549E+03, 4.1944E+03, 4.2342E+03,
     +  4.2743E+03, 4.3148E+03, 4.3556E+03, 4.3967E+03, 4.4382E+03,
     + 4.4801E+03, 4.5223E+03, 4.5648E+03, 4.6077E+03, 4.6510E+03,
     +  4.6946E+03, 4.7386E+03, 4.7830E+03, 4.8277E+03, 4.8728E+03,
     + 4.9183E+03, 4.9641E+03, 5.0104E+03, 5.0570E+03, 5.1040E+03,
     +  5.1514E+03, 5.1992E+03, 5.2473E+03, 5.2959E+03, 5.3449E+03,
     + 5.3942E+03, 5.4440E+03, 5.4942E+03, 5.5448E+03, 5.5958E+03,
     +  5.6472E+03, 5.6990E+03, 5.7513E+03, 5.8040E+03, 5.8571E+03,
     + 5.9106E+03, 5.9646E+03, 6.0190E+03, 6.0738E+03, 6.1291E+03,
     +  6.1849E+03, 6.2410E+03, 6.2977E+03, 6.3547E+03, 6.4123E+03,
     + 6.4703E+03, 6.5287E+03, 6.5877E+03, 6.6470E+03, 6.7069E+03,
     +  6.7673E+03, 6.8281E+03, 6.8894E+03, 6.9512E+03, 7.0135E+03,
     + 7.0762E+03, 7.1395E+03, 7.2033E+03, 7.2675E+03, 7.3323E+03,
     +  7.3976E+03, 7.4633E+03, 7.5297E+03, 7.5965E+03, 7.6638E+03,
     + 7.7317E+03, 7.8001E+03, 7.8690E+03, 7.9385E+03, 8.0085E+03,
     +  8.0790E+03, 8.1501E+03, 8.2218E+03, 8.2940E+03, 8.3667E+03,
     + 8.4400E+03, 8.5139E+03, 8.5883E+03, 8.6634E+03, 8.7389E+03,
     +  8.8151E+03, 8.8919E+03, 8.9692E+03, 9.0471E+03, 9.1257E+03,
     + 9.2048E+03, 9.2845E+03, 9.3648E+03, 9.4458E+03, 9.5273E+03,
     +  9.6095E+03, 9.6923E+03, 9.7757E+03, 9.8597E+03, 9.9444E+03,
     + 1.0030E+04, 1.0116E+04, 1.0202E+04, 1.0290E+04, 1.0377E+04,
     +  1.0466E+04, 1.0555E+04, 1.0645E+04, 1.0736E+04, 1.0827E+04,
     + 1.0919E+04, 1.1011E+04, 1.1105E+04, 1.1199E+04, 1.1293E+04,
     +  1.1389E+04, 1.1485E+04, 1.1582E+04, 1.1679E+04, 1.1777E+04,
     + 1.1876E+04, 1.1976E+04, 1.2076E+04, 1.2177E+04, 1.2279E+04,
     +  1.2382E+04, 1.2485E+04, 1.2590E+04, 1.2694E+04, 1.2800E+04,
     + 1.2906E+04, 1.3014E+04, 1.3122E+04, 1.3230E+04, 1.3340E+04,
     +  1.3450E+04, 1.3561E+04, 1.3673E+04, 1.3786E+04, 1.3900E+04,
     + 1.4014E+04, 1.4129E+04, 1.4245E+04, 1.4362E+04, 1.4480E+04,
     +  1.4599E+04, 1.4718E+04, 1.4838E+04, 1.4959E+04, 1.5081E+04,
     + 1.5204E+04, 1.5328E+04, 1.5452E+04, 1.5578E+04, 1.5704E+04,
     +  1.5832E+04, 1.5960E+04, 1.6089E+04, 1.6219E+04, 1.6350E+04,
     + 1.6482E+04, 1.6614E+04, 1.6748E+04, 1.6883E+04, 1.7018E+04,
     +  1.7155E+04, 1.7292E+04, 1.7431E+04, 1.7570E+04, 1.7711E+04,
     + 1.7852E+04, 1.7994E+04, 1.8138E+04, 1.8282E+04, 1.8427E+04,
     +  1.8574E+04, 1.8721E+04, 1.8870E+04, 1.9019E+04, 1.9170E+04,
     + 1.9321E+04, 1.9474E+04, 1.9627E+04, 1.9782E+04, 1.9938E+04,
     +  2.0095E+04, 2.0252E+04, 2.0411E+04, 2.0571E+04, 2.0733E+04,
     + 2.0895E+04, 2.1058E+04, 2.1223E+04, 2.1388E+04, 2.1555E+04,
     +  2.1723E+04, 2.1892E+04, 2.2062E+04, 2.2234E+04, 2.2406E+04,
     + 2.2580E+04, 2.2755E+04, 2.2931E+04, 2.3108E+04, 2.3286E+04,
     +  2.3466E+04, 2.3647E+04, 2.3829E+04, 2.4012E+04, 2.4197E+04,
     + 2.4382E+04, 2.4569E+04, 2.4758E+04, 2.4947E+04, 2.5138E+04,
     +  2.5330E+04, 2.5523E+04, 2.5718E+04, 2.5914E+04, 2.6111E+04,
     + 2.6310E+04, 2.6509E+04, 2.6711E+04, 2.6913E+04, 2.7117E+04,
     +  2.7322E+04, 2.7529E+04, 2.7737E+04, 2.7946E+04, 2.8157E+04,
     + 2.8369E+04, 2.8582E+04, 2.8797E+04, 2.9014E+04, 2.9231E+04,
     +  2.9450E+04, 2.9671E+04, 2.9893E+04, 3.0116E+04, 3.0341E+04,
     + 3.0568E+04, 3.0796E+04, 3.1025E+04, 3.1256E+04, 3.1488E+04,
     +  3.1722E+04, 3.1958E+04, 3.2195E+04, 3.2433E+04, 3.2673E+04,
     + 3.2915E+04, 3.3158E+04, 3.3403E+04, 3.3649E+04, 3.3897E+04,
     +  3.4146E+04, 3.4397E+04, 3.4650E+04, 3.4904E+04, 3.5160E+04,
     + 3.5418E+04, 3.5677E+04, 3.5938E+04, 3.6201E+04, 3.6465E+04,
     +  3.6731E+04, 3.6999E+04, 3.7268E+04, 3.7539E+04, 3.7812E+04,
     + 3.8086E+04, 3.8363E+04, 3.8641E+04, 3.8921E+04, 3.9202E+04,
     +  3.9486E+04, 3.9771E+04, 4.0058E+04, 4.0346E+04, 4.0637E+04,
     + 4.0929E+04, 4.1224E+04, 4.1520E+04, 4.1818E+04, 4.2118E+04,
     +  4.2419E+04, 4.2723E+04, 4.3029E+04, 4.3336E+04, 4.3645E+04,
     + 4.3957E+04, 4.4270E+04, 4.4585E+04, 4.4902E+04, 4.5222E+04,
     +  4.5543E+04, 4.5866E+04, 4.6191E+04, 4.6518E+04, 4.6847E+04,
     + 4.7179E+04, 4.7512E+04, 4.7847E+04, 4.8185E+04, 4.8524E+04,
     +  4.8866E+04, 4.9210E+04, 4.9555E+04, 4.9903E+04, 5.0254E+04,
     + 5.0606E+04, 5.0960E+04, 5.1317E+04, 5.1676E+04, 5.2037E+04,
     +  5.2400E+04, 5.2765E+04, 5.3133E+04, 5.3503E+04, 5.3875E+04,
     + 5.4249E+04, 5.4626E+04, 5.5005E+04, 5.5386E+04, 5.5770E+04,
     +  5.6156E+04, 5.6544E+04, 5.6935E+04, 5.7328E+04, 5.7723E+04,
     + 5.8121E+04, 5.8521E+04, 5.8924E+04, 5.9329E+04, 5.9737E+04,
     +  6.0146E+04, 6.0559E+04, 6.0974E+04, 6.1391E+04, 6.1811E+04,
     + 6.2233E+04, 6.2658E+04, 6.3086E+04, 6.3516E+04, 6.3949E+04,
     +  6.4384E+04, 6.4821E+04, 6.5262E+04, 6.5705E+04, 6.6151E+04,
     + 6.6599E+04, 6.7050E+04, 6.7504E+04, 6.7960E+04, 6.8419E+04,
     +  6.8881E+04, 6.9345E+04, 6.9813E+04, 7.0283E+04, 7.0755E+04,
     + 7.1231E+04, 7.1709E+04, 7.2191E+04, 7.2675E+04, 7.3162E+04,
     +  7.3651E+04, 7.4144E+04, 7.4640E+04, 7.5138E+04, 7.5639E+04,
     + 7.6144E+04, 7.6651E+04, 7.7161E+04, 7.7674E+04, 7.8191E+04,
     +  7.8710E+04, 7.9232E+04, 7.9757E+04, 8.0286E+04, 8.0817E+04,
     + 8.1352E+04, 8.1889E+04, 8.2430E+04, 8.2974E+04, 8.3521E+04,
     +  8.4071E+04, 8.4624E+04, 8.5181E+04, 8.5741E+04, 8.6304E+04,
     + 8.6870E+04, 8.7439E+04, 8.8012E+04, 8.8588E+04, 8.9168E+04,
     +  8.9750E+04, 9.0336E+04, 9.0926E+04, 9.1519E+04, 9.2115E+04,
     + 9.2714E+04, 9.3317E+04, 9.3924E+04, 9.4534E+04, 9.5147E+04,
     +  9.5764E+04, 9.6385E+04, 9.7008E+04, 9.7636E+04, 9.8267E+04,
     + 9.8902E+04, 9.9540E+04, 1.0018E+05, 1.0083E+05, 1.0148E+05,
     +  1.0213E+05, 1.0279E+05, 1.0345E+05, 1.0411E+05, 1.0478E+05,
     + 1.0545E+05, 1.0612E+05, 1.0680E+05, 1.0749E+05, 1.0817E+05,
     +  1.0886E+05, 1.0956E+05, 1.1026E+05, 1.1096E+05, 1.1166E+05,
     + 1.1237E+05, 1.1309E+05, 1.1381E+05, 1.1453E+05, 1.1526E+05,
     +  1.1599E+05, 1.1672E+05, 1.1746E+05, 1.1820E+05, 1.1895E+05,
     + 1.1970E+05, 1.2045E+05, 1.2121E+05, 1.2197E+05, 1.2274E+05,
     +  1.2351E+05, 1.2429E+05, 1.2507E+05, 1.2585E+05, 1.2664E+05,
     + 1.2744E+05, 1.2823E+05, 1.2904E+05, 1.2984E+05, 1.3065E+05,
     +  1.3147E+05, 1.3229E+05, 1.3311E+05, 1.3394E+05, 1.3477E+05,
     + 1.3561E+05, 1.3645E+05, 1.3730E+05, 1.3815E+05, 1.3901E+05,
     +  1.3987E+05, 1.4074E+05, 1.4161E+05, 1.4248E+05, 1.4336E+05,
     + 1.4425E+05, 1.4514E+05, 1.4603E+05, 1.4693E+05, 1.4783E+05,
     +  1.4874E+05, 1.4966E+05, 1.5057E+05, 1.5150E+05, 1.5243E+05,
     + 1.5336E+05, 1.5430E+05, 1.5524E+05, 1.5619E+05, 1.5715E+05,
     +  1.5810E+05, 1.5907E+05, 1.6004E+05, 1.6101E+05, 1.6199E+05,
     + 1.6298E+05, 1.6397E+05, 1.6496E+05, 1.6596E+05, 1.6697E+05,
     +  1.6798E+05, 1.6900E+05, 1.7002E+05, 1.7105E+05, 1.7208E+05,
     + 1.7312E+05, 1.7416E+05, 1.7521E+05, 1.7627E+05, 1.7733E+05,
     +  1.7840E+05, 1.7947E+05, 1.8055E+05, 1.8163E+05, 1.8272E+05,
     + 1.8382E+05, 1.8492E+05, 1.8602E+05, 1.8714E+05, 1.8825E+05,
     +  1.8938E+05, 1.9051E+05, 1.9164E+05, 1.9279E+05, 1.9393E+05,
     + 1.9509E+05, 1.9625E+05, 1.9741E+05, 1.9859E+05, 1.9977E+05,
     +  2.0095E+05, 2.0214E+05, 2.0334E+05, 2.0454E+05, 2.0575E+05,
     + 2.0697E+05, 2.0819E+05, 2.0942E+05, 2.1065E+05, 2.1189E+05,
     +  2.1314E+05, 2.1439E+05, 2.1565E+05, 2.1692E+05, 2.1819E+05,
     + 2.1947E+05, 2.2076E+05, 2.2205E+05, 2.2336E+05, 2.2466E+05,
     +  2.2598E+05, 2.2730E+05, 2.2862E+05, 2.2996E+05, 2.3130E+05,
     + 2.3264E+05, 2.3400E+05, 2.3536E+05, 2.3673E+05, 2.3810E+05,
     +  2.3949E+05, 2.4088E+05, 2.4227E+05, 2.4368E+05, 2.4509E+05,
     + 2.4651E+05, 2.4793E+05, 2.4936E+05, 2.5080E+05, 2.5225E+05,
     +  2.5371E+05, 2.5517E+05, 2.5664E+05, 2.5811E+05, 2.5960E+05,
     + 2.6109E+05, 2.6259E+05, 2.6410E+05, 2.6561E+05, 2.6713E+05,
     +  2.6866E+05, 2.7020E+05, 2.7174E+05, 2.7330E+05, 2.7486E+05,
     + 2.7643E+05, 2.7800E+05, 2.7959E+05, 2.8118E+05, 2.8278E+05,
     +  2.8439E+05, 2.8601E+05, 2.8763E+05, 2.8926E+05, 2.9090E+05,
     + 2.9255E+05, 2.9421E+05, 2.9588E+05, 2.9755E+05, 2.9923E+05,
     +  3.0092E+05, 3.0262E+05, 3.0433E+05, 3.0605E+05, 3.0777E+05,
     + 3.0950E+05, 3.1124E+05, 3.1299E+05, 3.1475E+05, 3.1652E+05,
     +  3.1830E+05, 3.2008E+05, 3.2188E+05, 3.2368E+05, 3.2549E+05,
     + 3.2731E+05, 3.2914E+05, 3.3098E+05, 3.3283E+05, 3.3468E+05,
     +  3.3655E+05, 3.3842E+05, 3.4031E+05, 3.4220E+05, 3.4410E+05,
     + 3.4602E+05, 3.4794E+05, 3.4987E+05, 3.5181E+05, 3.5376E+05,
     +  3.5572E+05, 3.5769E+05, 3.5966E+05, 3.6165E+05, 3.6365E+05,
     + 3.6566E+05, 3.6767E+05, 3.6970E+05, 3.7174E+05, 3.7378E+05,
     +  3.7584E+05, 3.7791E+05, 3.7998E+05, 3.8207E+05, 3.8416E+05,
     + 3.8627E+05, 3.8839E+05, 3.9052E+05, 3.9265E+05, 3.9480E+05,
     +  3.9696E+05, 3.9913E+05, 4.0131E+05, 4.0349E+05, 4.0569E+05,
     + 4.0790E+05, 4.1013E+05, 4.1236E+05, 4.1460E+05, 4.1685E+05,
     +  4.1912E+05, 4.2139E+05, 4.2368E+05, 4.2597E+05, 4.2828E+05,
     + 4.3060E+05, 4.3293E+05, 4.3527E+05, 4.3762E+05, 4.3998E+05,
     +  4.4236E+05, 4.4474E+05, 4.4714E+05, 4.4955E+05, 4.5196E+05,
     + 4.5440E+05, 4.5684E+05, 4.5929E+05, 4.6176E+05, 4.6423E+05,
     +  4.6672E+05, 4.6922E+05, 4.7174E+05, 4.7426E+05, 4.7680E+05,
     + 4.7935E+05, 4.8191E+05, 4.8448E+05, 4.8706E+05, 4.8966E+05,
     +  4.9227E+05, 4.9489E+05, 4.9752E+05, 5.0017E+05, 5.0282E+05,
     + 5.0549E+05, 5.0818E+05, 5.1087E+05, 5.1358E+05, 5.1630E+05,
     +  5.1903E+05, 5.2178E+05, 5.2454E+05, 5.2731E+05, 5.3009E+05,
     + 5.3289E+05, 5.3570E+05, 5.3853E+05, 5.4136E+05, 5.4421E+05,
     +  5.4707E+05, 5.4995E+05, 5.5284E+05, 5.5574E+05, 5.5866E+05,
     + 5.6159E+05, 5.6453E+05, 5.6749E+05, 5.7046E+05, 5.7344E+05,
     +  5.7644E+05, 5.7945E+05, 5.8248E+05, 5.8552E+05, 5.8857E+05,
     + 5.9164E+05, 5.9472E+05, 5.9782E+05, 6.0093E+05, 6.0405E+05,
     +  6.0719E+05, 6.1034E+05, 6.1351E+05, 6.1669E+05, 6.1989E+05,
     + 6.2310E+05, 6.2632E+05, 6.2956E+05, 6.3282E+05, 6.3609E+05,
     +  6.3937E+05, 6.4267E+05, 6.4599E+05, 6.4932E+05, 6.5266E+05,
     + 6.5602E+05, 6.5940E+05, 6.6279E+05, 6.6619E+05, 6.6961E+05,
     +  6.7305E+05, 6.7650E+05, 6.7997E+05, 6.8345E+05, 6.8695E+05,
     + 6.9047E+05, 6.9400E+05, 6.9754E+05, 7.0111E+05, 7.0468E+05,
     +  7.0828E+05, 7.1189E+05, 7.1552E+05, 7.1916E+05, 7.2282E+05,
     + 7.2650E+05, 7.3019E+05, 7.3390E+05, 7.3762E+05, 7.4137E+05,
     +  7.4512E+05, 7.4890E+05, 7.5269E+05, 7.5650E+05, 7.6033E+05,
     + 7.6417E+05, 7.6803E+05, 7.7191E+05, 7.7581E+05, 7.7972E+05,
     +  7.8365E+05, 7.8759E+05, 7.9156E+05, 7.9554E+05, 7.9954E+05,
     + 8.0356E+05, 8.0759E+05, 8.1165E+05, 8.1572E+05, 8.1981E+05,
     +  8.2391E+05, 8.2804E+05, 8.3218E+05, 8.3635E+05, 8.4053E+05,
     + 8.4472E+05, 8.4894E+05, 8.5318E+05, 8.5743E+05, 8.6170E+05,
     +  8.6599E+05, 8.7030E+05, 8.7463E+05, 8.7898E+05, 8.8335E+05,
     + 8.8774E+05, 8.9214E+05, 8.9657E+05, 9.0101E+05, 9.0547E+05,
     +  9.0996E+05, 9.1446E+05, 9.1898E+05, 9.2352E+05, 9.2808E+05,
     + 9.3266E+05, 9.3727E+05, 9.4189E+05, 9.4653E+05, 9.5119E+05,
     +  9.5587E+05, 9.6057E+05, 9.6530E+05, 9.7004E+05, 9.7480E+05,
     + 9.7959E+05, 9.8439E+05, 9.8922E+05, 9.9406E+05, 9.9893E+05,
     +  1.0038E+06, 1.0087E+06, 1.0137E+06, 1.0186E+06, 1.0236E+06,
     + 1.0286E+06, 1.0336E+06, 1.0386E+06, 1.0437E+06, 1.0488E+06,
     +  1.0539E+06, 1.0590E+06, 1.0641E+06, 1.0693E+06, 1.0745E+06,
     + 1.0797E+06, 1.0849E+06, 1.0902E+06, 1.0955E+06, 1.1008E+06,
     +  1.1061E+06, 1.1115E+06, 1.1168E+06, 1.1222E+06, 1.1276E+06,
     + 1.1331E+06, 1.1385E+06, 1.1440E+06, 1.1495E+06, 1.1551E+06,
     +  1.1606E+06, 1.1662E+06, 1.1718E+06, 1.1774E+06, 1.1831E+06,
     + 1.1887E+06, 1.1944E+06, 1.2002E+06, 1.2059E+06, 1.2117E+06,
     +  1.2175E+06, 1.2233E+06, 1.2291E+06, 1.2350E+06, 1.2409E+06,
     + 1.2468E+06, 1.2527E+06, 1.2587E+06, 1.2647E+06, 1.2707E+06,
     +  1.2768E+06, 1.2828E+06, 1.2889E+06, 1.2950E+06, 1.3012E+06,
     + 1.3074E+06, 1.3135E+06, 1.3198E+06, 1.3260E+06, 1.3323E+06,
     +  1.3386E+06, 1.3449E+06, 1.3513E+06, 1.3576E+06, 1.3640E+06,
     + 1.3705E+06, 1.3769E+06, 1.3834E+06, 1.3899E+06, 1.3965E+06,
     +  1.4030E+06, 1.4096E+06, 1.4162E+06, 1.4229E+06, 1.4296E+06,
     + 1.4363E+06, 1.4430E+06, 1.4498E+06, 1.4565E+06, 1.4634E+06,
     +  1.4702E+06, 1.4771E+06, 1.4840E+06, 1.4909E+06, 1.4978E+06,
     + 1.5048E+06, 1.5118E+06, 1.5189E+06, 1.5260E+06, 1.5331E+06,
     +  1.5402E+06, 1.5473E+06, 1.5545E+06, 1.5617E+06, 1.5690E+06,
     + 1.5763E+06, 1.5836E+06, 1.5909E+06, 1.5983E+06, 1.6057E+06,
     +  1.6131E+06, 1.6205E+06, 1.6280E+06, 1.6355E+06, 1.6431E+06,
     + 1.6507E+06, 1.6583E+06, 1.6659E+06, 1.6736E+06, 1.6813E+06,
     +  1.6890E+06, 1.6968E+06, 1.7046E+06, 1.7124E+06, 1.7202E+06,
     + 1.7281E+06, 1.7361E+06, 1.7440E+06, 1.7520E+06, 1.7600E+06,
     +  1.7681E+06, 1.7761E+06, 1.7843E+06, 1.7924E+06, 1.8006E+06,
     + 1.8088E+06, 1.8170E+06, 1.8253E+06, 1.8336E+06, 1.8420E+06,
     +  1.8503E+06, 1.8588E+06, 1.8672E+06, 1.8757E+06, 1.8842E+06,
     + 1.8927E+06, 1.9013E+06, 1.9099E+06, 1.9186E+06, 1.9273E+06,
     +  1.9360E+06, 1.9447E+06, 1.9535E+06, 1.9624E+06, 1.9712E+06,
     + 1.9801E+06, 1.9890E+06, 1.9980E+06, 2.0070E+06, 2.0160E+06,
     +  2.0251E+06, 2.0342E+06, 2.0434E+06, 2.0525E+06, 2.0618E+06,
     + 2.0710E+06, 2.0803E+06, 2.0896E+06, 2.0990E+06, 2.1084E+06,
     +  2.1178E+06, 2.1273E+06, 2.1368E+06, 2.1464E+06, 2.1559E+06,
     + 2.1656E+06, 2.1752E+06, 2.1849E+06, 2.1947E+06, 2.2044E+06,
     +  2.2143E+06, 2.2241E+06, 2.2340E+06, 2.2439E+06, 2.2539E+06,
     + 2.2639E+06, 2.2740E+06, 2.2840E+06, 2.2942E+06, 2.3043E+06,
     +  2.3145E+06, 2.3248E+06, 2.3351E+06, 2.3454E+06, 2.3558E+06,
     + 2.3662E+06, 2.3766E+06, 2.3871E+06, 2.3976E+06, 2.4082E+06,
     +  2.4188E+06, 2.4294E+06, 2.4401E+06, 2.4509E+06, 2.4616E+06,
     + 2.4724E+06, 2.4833E+06, 2.4942E+06, 2.5051E+06, 2.5161E+06,
     +  2.5272E+06, 2.5382E+06, 2.5493E+06, 2.5605E+06, 2.5717E+06,
     + 2.5829E+06, 2.5942E+06, 2.6055E+06, 2.6169E+06, 2.6283E+06,
     +  2.6398E+06, 2.6513E+06, 2.6628E+06, 2.6744E+06, 2.6860E+06,
     + 2.6977E+06, 2.7094E+06, 2.7212E+06, 2.7330E+06, 2.7449E+06,
     +  2.7568E+06, 2.7687E+06, 2.7807E+06, 2.7928E+06, 2.8049E+06,
     + 2.8170E+06, 2.8292E+06, 2.8414E+06, 2.8537E+06, 2.8660E+06,
     +  2.8783E+06, 2.8908E+06, 2.9032E+06, 2.9157E+06, 2.9283E+06,
     + 2.9409E+06, 2.9535E+06, 2.9662E+06, 2.9790E+06, 2.9918E+06,
     +  3.0046E+06, 3.0175E+06, 3.0304E+06, 3.0434E+06, 3.0565E+06,
     + 3.0695E+06, 3.0827E+06, 3.0959E+06, 3.1091E+06, 3.1224E+06,
     +  3.1357E+06, 3.1491E+06, 3.1625E+06, 3.1760E+06, 3.1895E+06,
     + 3.2031E+06, 3.2168E+06, 3.2305E+06, 3.2442E+06, 3.2580E+06,
     +  3.2718E+06, 3.2857E+06, 3.2997E+06, 3.3137E+06, 3.3277E+06,
     + 3.3418E+06, 3.3560E+06, 3.3702E+06, 3.3844E+06, 3.3987E+06,
     +  3.4131E+06, 3.4275E+06, 3.4420E+06, 3.4565E+06, 3.4711E+06,
     + 3.4858E+06, 3.5004E+06, 3.5152E+06, 3.5300E+06, 3.5448E+06,
     +  3.5597E+06, 3.5747E+06, 3.5897E+06, 3.6048E+06, 3.6199E+06,
     + 3.6351E+06, 3.6504E+06, 3.6657E+06, 3.6810E+06, 3.6964E+06,
     +  3.7119E+06, 3.7274E+06, 3.7430E+06, 3.7586E+06, 3.7743E+06,
     + 3.7901E+06, 3.8059E+06, 3.8218E+06, 3.8377E+06, 3.8537E+06,
     +  3.8697E+06, 3.8859E+06, 3.9020E+06, 3.9182E+06, 3.9345E+06,
     + 3.9509E+06, 3.9673E+06, 3.9837E+06, 4.0003E+06, 4.0168E+06,
     +  4.0335E+06, 4.0502E+06, 4.0669E+06, 4.0838E+06, 4.1007E+06,
     + 4.1176E+06, 4.1346E+06, 4.1517E+06, 4.1688E+06, 4.1860E+06,
     +  4.2033E+06, 4.2206E+06, 4.2380E+06, 4.2554E+06, 4.2729E+06,
     + 4.2905E+06, 4.3081E+06, 4.3258E+06, 4.3436E+06, 4.3614E+06,
     +  4.3793E+06, 4.3973E+06, 4.4153E+06, 4.4334E+06, 4.4515E+06,
     + 4.4698E+06, 4.4880E+06, 4.5064E+06, 4.5248E+06, 4.5433E+06,
     +  4.5618E+06, 4.5805E+06, 4.5991E+06, 4.6179E+06, 4.6367E+06,
     + 4.6556E+06, 4.6745E+06, 4.6936E+06, 4.7126E+06, 4.7318E+06,
     +  4.7510E+06, 4.7703E+06, 4.7897E+06, 4.8091E+06, 4.8286E+06,
     + 4.8482E+06, 4.8678E+06, 4.8875E+06, 4.9073E+06, 4.9271E+06,
     +  4.9471E+06, 4.9671E+06, 4.9871E+06, 5.0073E+06, 5.0275E+06,
     + 5.0477E+06, 5.0681E+06, 5.0885E+06, 5.1090E+06, 5.1296E+06,
     +  5.1502E+06, 5.1709E+06, 5.1917E+06, 5.2126E+06, 5.2335E+06,
     + 5.2545E+06, 5.2756E+06, 5.2967E+06, 5.3180E+06, 5.3393E+06,
     +  5.3606E+06, 5.3821E+06, 5.4036E+06, 5.4252E+06, 5.4469E+06,
     + 5.4687E+06, 5.4905E+06, 5.5124E+06, 5.5344E+06, 5.5565E+06,
     +  5.5786E+06, 5.6008E+06, 5.6231E+06, 5.6455E+06, 5.6680E+06,
     + 5.6905E+06, 5.7131E+06, 5.7358E+06, 5.7586E+06, 5.7814E+06,
     +  5.8043E+06, 5.8274E+06, 5.8504E+06, 5.8736E+06, 5.8969E+06,
     + 5.9202E+06, 5.9436E+06, 5.9671E+06, 5.9907E+06, 6.0143E+06,
     +  6.0381E+06, 6.0619E+06, 6.0858E+06, 6.1098E+06, 6.1339E+06,
     + 6.1580E+06, 6.1823E+06, 6.2066E+06, 6.2310E+06, 6.2555E+06,
     +  6.2801E+06, 6.3047E+06, 6.3295E+06, 6.3543E+06, 6.3792E+06,
     + 6.4042E+06, 6.4293E+06, 6.4545E+06, 6.4797E+06, 6.5051E+06,
     +  6.5305E+06, 6.5561E+06, 6.5817E+06, 6.6074E+06, 6.6332E+06,
     + 6.6590E+06, 6.6850E+06, 6.7110E+06, 6.7372E+06, 6.7634E+06,
     +  6.7897E+06, 6.8162E+06, 6.8427E+06, 6.8693E+06, 6.8959E+06,
     + 6.9227E+06, 6.9496E+06, 6.9765E+06, 7.0036E+06, 7.0307E+06,
     +  7.0580E+06, 7.0853E+06, 7.1127E+06, 7.1402E+06, 7.1679E+06,
     + 7.1956E+06, 7.2234E+06, 7.2512E+06, 7.2792E+06, 7.3073E+06,
     +  7.3355E+06, 7.3638E+06, 7.3921E+06, 7.4206E+06, 7.4492E+06,
     + 7.4778E+06, 7.5066E+06, 7.5354E+06, 7.5644E+06, 7.5934E+06,
     +  7.6226E+06, 7.6518E+06, 7.6811E+06, 7.7106E+06, 7.7401E+06,
     + 7.7698E+06, 7.7995E+06, 7.8293E+06, 7.8593E+06, 7.8893E+06,
     +  7.9195E+06, 7.9497E+06, 7.9801E+06, 8.0105E+06, 8.0411E+06,
     + 8.0717E+06, 8.1025E+06, 8.1333E+06, 8.1643E+06, 8.1953E+06,
     +  8.2265E+06, 8.2578E+06, 8.2892E+06, 8.3206E+06, 8.3522E+06,
     + 8.3839E+06, 8.4157E+06, 8.4476E+06, 8.4796E+06, 8.5117E+06,
     +  8.5440E+06, 8.5763E+06, 8.6087E+06, 8.6413E+06, 8.6739E+06,
     + 8.7067E+06, 8.7396E+06, 8.7726E+06, 8.8056E+06, 8.8388E+06,
     +  8.8722E+06, 8.9056E+06, 8.9391E+06, 8.9728E+06, 9.0065E+06,
     + 9.0404E+06, 9.0744E+06, 9.1084E+06, 9.1426E+06, 9.1770E+06,
     +  9.2114E+06, 9.2459E+06, 9.2806E+06, 9.3154E+06, 9.3502E+06,
     + 9.3852E+06, 9.4204E+06, 9.4556E+06, 9.4909E+06, 9.5264E+06,
     +  9.5620E+06, 9.5977E+06, 9.6335E+06, 9.6694E+06, 9.7055E+06,
     + 9.7416E+06, 9.7779E+06, 9.8143E+06, 9.8509E+06, 9.8875E+06,
     +  9.9243E+06, 9.9611E+06, 9.9982E+06, 1.0035E+07, 1.0073E+07,
     + 1.0110E+07, 1.0147E+07, 1.0185E+07, 1.0223E+07, 1.0261E+07,
     +  1.0299E+07, 1.0337E+07, 1.0375E+07, 1.0413E+07, 1.0452E+07,
     + 1.0490E+07, 1.0529E+07, 1.0568E+07, 1.0607E+07, 1.0646E+07,
     +  1.0685E+07, 1.0725E+07, 1.0764E+07, 1.0804E+07, 1.0843E+07,
     + 1.0883E+07, 1.0923E+07, 1.0963E+07, 1.1004E+07, 1.1044E+07,
     +  1.1085E+07, 1.1125E+07, 1.1166E+07, 1.1207E+07, 1.1248E+07,
     + 1.1289E+07, 1.1330E+07, 1.1372E+07, 1.1414E+07, 1.1455E+07,
     +  1.1497E+07, 1.1539E+07, 1.1581E+07, 1.1623E+07, 1.1666E+07,
     + 1.1708E+07, 1.1751E+07, 1.1794E+07, 1.1837E+07, 1.1880E+07,
     +  1.1923E+07, 1.1966E+07, 1.2010E+07, 1.2053E+07, 1.2097E+07,
     + 1.2141E+07, 1.2185E+07, 1.2229E+07, 1.2274E+07, 1.2318E+07,
     +  1.2363E+07, 1.2407E+07, 1.2452E+07, 1.2497E+07, 1.2542E+07,
     + 1.2588E+07, 1.2633E+07, 1.2679E+07, 1.2725E+07, 1.2771E+07,
     +  1.2817E+07, 1.2863E+07, 1.2909E+07, 1.2956E+07, 1.3002E+07,
     + 1.3049E+07, 1.3096E+07, 1.3143E+07, 1.3190E+07, 1.3238E+07,
     +  1.3285E+07, 1.3333E+07, 1.3381E+07, 1.3429E+07, 1.3477E+07,
     + 1.3525E+07, 1.3573E+07, 1.3622E+07, 1.3671E+07, 1.3720E+07,
     +  1.3769E+07, 1.3818E+07, 1.3867E+07, 1.3917E+07, 1.3966E+07,
     + 1.4016E+07, 1.4066E+07, 1.4116E+07, 1.4166E+07, 1.4217E+07,
     +  1.4267E+07, 1.4318E+07, 1.4369E+07, 1.4420E+07, 1.4471E+07,
     + 1.4523E+07, 1.4574E+07, 1.4626E+07, 1.4678E+07, 1.4730E+07,
     +  1.4782E+07, 1.4835E+07, 1.4887E+07, 1.4940E+07, 1.4993E+07,
     + 1.5046E+07, 1.5099E+07, 1.5152E+07, 1.5206E+07, 1.5259E+07,
     +  1.5313E+07, 1.5367E+07, 1.5421E+07, 1.5476E+07, 1.5530E+07,
     + 1.5585E+07, 1.5640E+07, 1.5695E+07, 1.5750E+07, 1.5805E+07,
     +  1.5861E+07, 1.5917E+07, 1.5973E+07, 1.6029E+07, 1.6085E+07,
     + 1.6141E+07, 1.6198E+07, 1.6255E+07, 1.6311E+07, 1.6369E+07,
     +  1.6426E+07, 1.6483E+07, 1.6541E+07, 1.6599E+07, 1.6657E+07,
     + 1.6715E+07, 1.6773E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --      1222
      data (QofT( 3,J),J=1,2408)/ 1.0441E+00, 4.7471E+00, 8.8297E+00,
     +  1.2980E+01, 1.7154E+01, 2.1340E+01, 2.5531E+01, 2.9727E+01,
     + 3.3925E+01, 3.8125E+01, 4.2327E+01, 4.6529E+01, 5.0732E+01,
     +  5.4935E+01, 5.9139E+01, 6.3344E+01, 6.7548E+01, 7.1753E+01,
     + 7.5959E+01, 8.0164E+01, 8.8575E+01, 9.6987E+01, 1.0540E+02,
     +  1.1381E+02, 1.2222E+02, 1.3064E+02, 1.3905E+02, 1.4746E+02,
     + 1.5588E+02, 1.6429E+02, 1.7270E+02, 1.8112E+02, 1.8953E+02,
     +  1.9794E+02, 2.0636E+02, 2.1477E+02, 2.2319E+02, 2.3160E+02,
     + 2.4002E+02, 2.4843E+02, 2.5685E+02, 2.6527E+02, 2.7368E+02,
     +  2.8210E+02, 2.9052E+02, 2.9894E+02, 3.0736E+02, 3.1578E+02,
     + 3.2421E+02, 3.3264E+02, 3.4107E+02, 3.4951E+02, 3.5795E+02,
     +  3.6640E+02, 3.7485E+02, 3.8332E+02, 3.9178E+02, 4.0027E+02,
     + 4.0876E+02, 4.1726E+02, 4.2577E+02, 4.3429E+02, 4.4284E+02,
     +  4.5139E+02, 4.5997E+02, 4.6856E+02, 4.7718E+02, 4.8582E+02,
     + 4.9448E+02, 5.0316E+02, 5.1188E+02, 5.2062E+02, 5.2940E+02,
     +  5.3820E+02, 5.4705E+02, 5.5592E+02, 5.6483E+02, 5.7379E+02,
     + 5.8278E+02, 5.9181E+02, 6.0090E+02, 6.1003E+02, 6.1921E+02,
     +  6.2843E+02, 6.3772E+02, 6.4705E+02, 6.5644E+02, 6.6589E+02,
     + 6.7540E+02, 6.8497E+02, 6.9461E+02, 7.0431E+02, 7.1409E+02,
     +  7.2392E+02, 7.3384E+02, 7.4382E+02, 7.5388E+02, 7.6402E+02,
     + 7.7424E+02, 7.8454E+02, 7.9492E+02, 8.0539E+02, 8.1593E+02,
     +  8.2658E+02, 8.3731E+02, 8.4814E+02, 8.5905E+02, 8.7006E+02,
     + 8.8117E+02, 8.9238E+02, 9.0370E+02, 9.1511E+02, 9.2663E+02,
     +  9.3825E+02, 9.4999E+02, 9.6184E+02, 9.7379E+02, 9.8587E+02,
     + 9.9805E+02, 1.0103E+03, 1.0228E+03, 1.0353E+03, 1.0480E+03,
     +  1.0608E+03, 1.0737E+03, 1.0867E+03, 1.0999E+03, 1.1132E+03,
     + 1.1266E+03, 1.1402E+03, 1.1539E+03, 1.1678E+03, 1.1817E+03,
     +  1.1959E+03, 1.2101E+03, 1.2245E+03, 1.2391E+03, 1.2538E+03,
     + 1.2687E+03, 1.2837E+03, 1.2989E+03, 1.3142E+03, 1.3296E+03,
     +  1.3453E+03, 1.3611E+03, 1.3770E+03, 1.3932E+03, 1.4094E+03,
     + 1.4259E+03, 1.4425E+03, 1.4593E+03, 1.4763E+03, 1.4934E+03,
     +  1.5108E+03, 1.5282E+03, 1.5459E+03, 1.5638E+03, 1.5818E+03,
     + 1.6001E+03, 1.6185E+03, 1.6371E+03, 1.6559E+03, 1.6749E+03,
     +  1.6941E+03, 1.7135E+03, 1.7330E+03, 1.7528E+03, 1.7728E+03,
     + 1.7930E+03, 1.8134E+03, 1.8340E+03, 1.8548E+03, 1.8758E+03,
     +  1.8970E+03, 1.9185E+03, 1.9402E+03, 1.9621E+03, 1.9842E+03,
     + 2.0065E+03, 2.0291E+03, 2.0519E+03, 2.0749E+03, 2.0981E+03,
     +  2.1216E+03, 2.1453E+03, 2.1693E+03, 2.1935E+03, 2.2179E+03,
     + 2.2426E+03, 2.2675E+03, 2.2927E+03, 2.3181E+03, 2.3438E+03,
     +  2.3698E+03, 2.3960E+03, 2.4224E+03, 2.4491E+03, 2.4761E+03,
     + 2.5034E+03, 2.5500E+03, 2.5780E+03, 2.6063E+03, 2.6348E+03,
     +  2.6637E+03, 2.6928E+03, 2.7222E+03, 2.7519E+03, 2.7818E+03,
     + 2.8121E+03, 2.8426E+03, 2.8735E+03, 2.9046E+03, 2.9361E+03,
     +  2.9678E+03, 2.9999E+03, 3.0323E+03, 3.0650E+03, 3.0980E+03,
     + 3.1313E+03, 3.1649E+03, 3.1989E+03, 3.2331E+03, 3.2677E+03,
     +  3.3027E+03, 3.3379E+03, 3.3735E+03, 3.4095E+03, 3.4458E+03,
     + 3.4824E+03, 3.5193E+03, 3.5567E+03, 3.5943E+03, 3.6323E+03,
     +  3.6707E+03, 3.7095E+03, 3.7486E+03, 3.7880E+03, 3.8278E+03,
     + 3.8681E+03, 3.9086E+03, 3.9496E+03, 3.9909E+03, 4.0326E+03,
     +  4.0747E+03, 4.1172E+03, 4.1601E+03, 4.2034E+03, 4.2470E+03,
     + 4.2911E+03, 4.3356E+03, 4.3805E+03, 4.4258E+03, 4.4715E+03,
     +  4.5176E+03, 4.5641E+03, 4.6111E+03, 4.6585E+03, 4.7063E+03,
     + 4.7545E+03, 4.8033E+03, 4.8524E+03, 4.9019E+03, 4.9520E+03,
     +  5.0024E+03, 5.0533E+03, 5.1047E+03, 5.1565E+03, 5.2088E+03,
     + 5.2616E+03, 5.3148E+03, 5.3685E+03, 5.4227E+03, 5.4774E+03,
     +  5.5325E+03, 5.5881E+03, 5.6443E+03, 5.7009E+03, 5.7580E+03,
     + 5.8156E+03, 5.8737E+03, 5.9324E+03, 5.9915E+03, 6.0512E+03,
     +  6.1114E+03, 6.1721E+03, 6.2333E+03, 6.2951E+03, 6.3574E+03,
     + 6.4203E+03, 6.4837E+03, 6.5476E+03, 6.6121E+03, 6.6772E+03,
     +  6.7428E+03, 6.8089E+03, 6.8757E+03, 6.9430E+03, 7.0109E+03,
     + 7.0794E+03, 7.1485E+03, 7.2181E+03, 7.2883E+03, 7.3592E+03,
     +  7.4306E+03, 7.5027E+03, 7.5754E+03, 7.6486E+03, 7.7225E+03,
     + 7.7971E+03, 7.8722E+03, 7.9480E+03, 8.0244E+03, 8.1015E+03,
     +  8.1792E+03, 8.2576E+03, 8.3366E+03, 8.4163E+03, 8.4967E+03,
     + 8.5777E+03, 8.6594E+03, 8.7417E+03, 8.8248E+03, 8.9086E+03,
     +  8.9930E+03, 9.0782E+03, 9.1640E+03, 9.2506E+03, 9.3379E+03,
     + 9.4259E+03, 9.5145E+03, 9.6040E+03, 9.6942E+03, 9.7851E+03,
     +  9.8768E+03, 9.9692E+03, 1.0062E+04, 1.0156E+04, 1.0251E+04,
     + 1.0346E+04, 1.0443E+04, 1.0540E+04, 1.0638E+04, 1.0736E+04,
     +  1.0836E+04, 1.0936E+04, 1.1037E+04, 1.1139E+04, 1.1241E+04,
     + 1.1345E+04, 1.1449E+04, 1.1554E+04, 1.1660E+04, 1.1767E+04,
     +  1.1875E+04, 1.1983E+04, 1.2093E+04, 1.2203E+04, 1.2314E+04,
     + 1.2426E+04, 1.2539E+04, 1.2653E+04, 1.2768E+04, 1.2883E+04,
     +  1.3000E+04, 1.3117E+04, 1.3236E+04, 1.3355E+04, 1.3476E+04,
     + 1.3597E+04, 1.3719E+04, 1.3842E+04, 1.3966E+04, 1.4091E+04,
     +  1.4217E+04, 1.4344E+04, 1.4472E+04, 1.4601E+04, 1.4732E+04,
     + 1.4863E+04, 1.4995E+04, 1.5128E+04, 1.5262E+04, 1.5397E+04,
     +  1.5533E+04, 1.5670E+04, 1.5809E+04, 1.5948E+04, 1.6089E+04,
     + 1.6230E+04, 1.6373E+04, 1.6517E+04, 1.6661E+04, 1.6807E+04,
     +  1.6954E+04, 1.7102E+04, 1.7252E+04, 1.7402E+04, 1.7554E+04,
     + 1.7707E+04, 1.7860E+04, 1.8016E+04, 1.8172E+04, 1.8329E+04,
     +  1.8488E+04, 1.8648E+04, 1.8809E+04, 1.8971E+04, 1.9134E+04,
     + 1.9299E+04, 1.9465E+04, 1.9632E+04, 1.9801E+04, 1.9970E+04,
     +  2.0141E+04, 2.0313E+04, 2.0487E+04, 2.0662E+04, 2.0838E+04,
     + 2.1015E+04, 2.1194E+04, 2.1374E+04, 2.1556E+04, 2.1738E+04,
     +  2.1922E+04, 2.2108E+04, 2.2295E+04, 2.2483E+04, 2.2673E+04,
     + 2.2864E+04, 2.3056E+04, 2.3250E+04, 2.3445E+04, 2.3642E+04,
     +  2.3840E+04, 2.4040E+04, 2.4241E+04, 2.4443E+04, 2.4647E+04,
     + 2.4853E+04, 2.5060E+04, 2.5268E+04, 2.5478E+04, 2.5690E+04,
     +  2.5903E+04, 2.6118E+04, 2.6334E+04, 2.6551E+04, 2.6771E+04,
     + 2.6992E+04, 2.7214E+04, 2.7438E+04, 2.7664E+04, 2.7891E+04,
     +  2.8120E+04, 2.8351E+04, 2.8583E+04, 2.8817E+04, 2.9053E+04,
     + 2.9290E+04, 2.9529E+04, 2.9770E+04, 3.0012E+04, 3.0256E+04,
     +  3.0502E+04, 3.0750E+04, 3.0999E+04, 3.1250E+04, 3.1503E+04,
     + 3.1758E+04, 3.2015E+04, 3.2273E+04, 3.2533E+04, 3.2795E+04,
     +  3.3059E+04, 3.3325E+04, 3.3592E+04, 3.3862E+04, 3.4133E+04,
     + 3.4406E+04, 3.4682E+04, 3.4959E+04, 3.5238E+04, 3.5519E+04,
     +  3.5802E+04, 3.6087E+04, 3.6374E+04, 3.6663E+04, 3.6953E+04,
     + 3.7246E+04, 3.7541E+04, 3.7838E+04, 3.8137E+04, 3.8439E+04,
     +  3.8742E+04, 3.9047E+04, 3.9355E+04, 3.9664E+04, 3.9976E+04,
     + 4.0290E+04, 4.0606E+04, 4.0924E+04, 4.1244E+04, 4.1567E+04,
     +  4.1892E+04, 4.2219E+04, 4.2548E+04, 4.2879E+04, 4.3213E+04,
     + 4.3549E+04, 4.3887E+04, 4.4228E+04, 4.4571E+04, 4.4916E+04,
     +  4.5264E+04, 4.5614E+04, 4.5966E+04, 4.6321E+04, 4.6678E+04,
     + 4.7037E+04, 4.7399E+04, 4.7763E+04, 4.8130E+04, 4.8500E+04,
     +  4.8871E+04, 4.9246E+04, 4.9623E+04, 5.0002E+04, 5.0384E+04,
     + 5.0768E+04, 5.1155E+04, 5.1545E+04, 5.1937E+04, 5.2332E+04,
     +  5.2729E+04, 5.3129E+04, 5.3532E+04, 5.3937E+04, 5.4346E+04,
     + 5.4756E+04, 5.5170E+04, 5.5586E+04, 5.6005E+04, 5.6427E+04,
     +  5.6852E+04, 5.7279E+04, 5.7709E+04, 5.8142E+04, 5.8578E+04,
     + 5.9017E+04, 5.9459E+04, 5.9903E+04, 6.0351E+04, 6.0801E+04,
     +  6.1255E+04, 6.1711E+04, 6.2170E+04, 6.2632E+04, 6.3098E+04,
     + 6.3566E+04, 6.4038E+04, 6.4512E+04, 6.4990E+04, 6.5470E+04,
     +  6.5954E+04, 6.6441E+04, 6.6931E+04, 6.7424E+04, 6.7921E+04,
     + 6.8420E+04, 6.8923E+04, 6.9429E+04, 6.9939E+04, 7.0451E+04,
     +  7.0967E+04, 7.1487E+04, 7.2009E+04, 7.2535E+04, 7.3064E+04,
     + 7.3597E+04, 7.4133E+04, 7.4673E+04, 7.5216E+04, 7.5762E+04,
     +  7.6312E+04, 7.6866E+04, 7.7423E+04, 7.7983E+04, 7.8547E+04,
     + 7.9115E+04, 7.9686E+04, 8.0261E+04, 8.0840E+04, 8.1422E+04,
     +  8.2008E+04, 8.2598E+04, 8.3191E+04, 8.3788E+04, 8.4389E+04,
     + 8.4993E+04, 8.5602E+04, 8.6214E+04, 8.6830E+04, 8.7450E+04,
     +  8.8074E+04, 8.8702E+04, 8.9333E+04, 8.9969E+04, 9.0609E+04,
     + 9.1252E+04, 9.1900E+04, 9.2551E+04, 9.3207E+04, 9.3867E+04,
     +  9.4531E+04, 9.5199E+04, 9.5871E+04, 9.6547E+04, 9.7228E+04,
     + 9.7913E+04, 9.8602E+04, 9.9295E+04, 9.9992E+04, 1.0069E+05,
     +  1.0140E+05, 1.0211E+05, 1.0283E+05, 1.0355E+05, 1.0427E+05,
     + 1.0500E+05, 1.0573E+05, 1.0647E+05, 1.0721E+05, 1.0795E+05,
     +  1.0871E+05, 1.0946E+05, 1.1022E+05, 1.1098E+05, 1.1175E+05,
     + 1.1253E+05, 1.1331E+05, 1.1409E+05, 1.1488E+05, 1.1567E+05,
     +  1.1647E+05, 1.1727E+05, 1.1808E+05, 1.1889E+05, 1.1971E+05,
     + 1.2053E+05, 1.2136E+05, 1.2219E+05, 1.2303E+05, 1.2387E+05,
     +  1.2472E+05, 1.2557E+05, 1.2643E+05, 1.2729E+05, 1.2816E+05,
     + 1.2903E+05, 1.2991E+05, 1.3079E+05, 1.3168E+05, 1.3257E+05,
     +  1.3347E+05, 1.3438E+05, 1.3529E+05, 1.3620E+05, 1.3712E+05,
     + 1.3805E+05, 1.3898E+05, 1.3992E+05, 1.4086E+05, 1.4181E+05,
     +  1.4277E+05, 1.4373E+05, 1.4469E+05, 1.4566E+05, 1.4664E+05,
     + 1.4762E+05, 1.4861E+05, 1.4961E+05, 1.5061E+05, 1.5161E+05,
     +  1.5262E+05, 1.5364E+05, 1.5467E+05, 1.5570E+05, 1.5673E+05,
     + 1.5777E+05, 1.5882E+05, 1.5988E+05, 1.6094E+05, 1.6200E+05,
     +  1.6308E+05, 1.6416E+05, 1.6524E+05, 1.6633E+05, 1.6743E+05,
     + 1.6854E+05, 1.6965E+05, 1.7076E+05, 1.7189E+05, 1.7302E+05,
     +  1.7415E+05, 1.7530E+05, 1.7645E+05, 1.7760E+05, 1.7877E+05,
     + 1.7994E+05, 1.8111E+05, 1.8230E+05, 1.8349E+05, 1.8468E+05,
     +  1.8589E+05, 1.8710E+05, 1.8832E+05, 1.8954E+05, 1.9077E+05,
     + 1.9201E+05, 1.9326E+05, 1.9451E+05, 1.9577E+05, 1.9704E+05,
     +  1.9831E+05, 1.9959E+05, 2.0088E+05, 2.0218E+05, 2.0348E+05,
     + 2.0480E+05, 2.0611E+05, 2.0744E+05, 2.0877E+05, 2.1011E+05,
     +  2.1146E+05, 2.1282E+05, 2.1418E+05, 2.1555E+05, 2.1693E+05,
     + 2.1832E+05, 2.1972E+05, 2.2112E+05, 2.2253E+05, 2.2395E+05,
     +  2.2538E+05, 2.2681E+05, 2.2825E+05, 2.2970E+05, 2.3116E+05,
     + 2.3263E+05, 2.3410E+05, 2.3559E+05, 2.3708E+05, 2.3858E+05,
     +  2.4009E+05, 2.4160E+05, 2.4313E+05, 2.4466E+05, 2.4620E+05,
     + 2.4775E+05, 2.4931E+05, 2.5088E+05, 2.5246E+05, 2.5404E+05,
     +  2.5564E+05, 2.5724E+05, 2.5885E+05, 2.6047E+05, 2.6210E+05,
     + 2.6374E+05, 2.6539E+05, 2.6704E+05, 2.6871E+05, 2.7038E+05,
     +  2.7207E+05, 2.7376E+05, 2.7546E+05, 2.7718E+05, 2.7890E+05,
     + 2.8063E+05, 2.8237E+05, 2.8412E+05, 2.8588E+05, 2.8764E+05,
     +  2.8942E+05, 2.9121E+05, 2.9301E+05, 2.9482E+05, 2.9663E+05,
     + 2.9846E+05, 3.0030E+05, 3.0214E+05, 3.0400E+05, 3.0587E+05,
     +  3.0774E+05, 3.0963E+05, 3.1153E+05, 3.1344E+05, 3.1535E+05,
     + 3.1728E+05, 3.1922E+05, 3.2117E+05, 3.2313E+05, 3.2510E+05,
     +  3.2708E+05, 3.2907E+05, 3.3107E+05, 3.3308E+05, 3.3511E+05,
     + 3.3714E+05, 3.3919E+05, 3.4124E+05, 3.4331E+05, 3.4539E+05,
     +  3.4748E+05, 3.4958E+05, 3.5169E+05, 3.5381E+05, 3.5594E+05,
     + 3.5809E+05, 3.6025E+05, 3.6241E+05, 3.6459E+05, 3.6678E+05,
     +  3.6899E+05, 3.7120E+05, 3.7343E+05, 3.7566E+05, 3.7791E+05,
     + 3.8017E+05, 3.8245E+05, 3.8473E+05, 3.8703E+05, 3.8934E+05,
     +  3.9166E+05, 3.9399E+05, 3.9634E+05, 3.9870E+05, 4.0107E+05,
     + 4.0345E+05, 4.0585E+05, 4.0825E+05, 4.1067E+05, 4.1311E+05,
     +  4.1555E+05, 4.1801E+05, 4.2048E+05, 4.2297E+05, 4.2546E+05,
     + 4.2797E+05, 4.3050E+05, 4.3303E+05, 4.3558E+05, 4.3814E+05,
     +  4.4072E+05, 4.4331E+05, 4.4591E+05, 4.4853E+05, 4.5116E+05,
     + 4.5380E+05, 4.5645E+05, 4.5912E+05, 4.6181E+05, 4.6451E+05,
     +  4.6722E+05, 4.6994E+05, 4.7268E+05, 4.7544E+05, 4.7820E+05,
     + 4.8099E+05, 4.8378E+05, 4.8659E+05, 4.8942E+05, 4.9226E+05,
     +  4.9511E+05, 4.9798E+05, 5.0086E+05, 5.0376E+05, 5.0667E+05,
     + 5.0960E+05, 5.1254E+05, 5.1550E+05, 5.1847E+05, 5.2146E+05,
     +  5.2446E+05, 5.2748E+05, 5.3051E+05, 5.3356E+05, 5.3662E+05,
     + 5.3970E+05, 5.4280E+05, 5.4591E+05, 5.4903E+05, 5.5217E+05,
     +  5.5533E+05, 5.5850E+05, 5.6169E+05, 5.6490E+05, 5.6812E+05,
     + 5.7136E+05, 5.7461E+05, 5.7788E+05, 5.8117E+05, 5.8447E+05,
     +  5.8779E+05, 5.9113E+05, 5.9448E+05, 5.9785E+05, 6.0124E+05,
     + 6.0464E+05, 6.0806E+05, 6.1150E+05, 6.1495E+05, 6.1842E+05,
     +  6.2191E+05, 6.2542E+05, 6.2894E+05, 6.3248E+05, 6.3604E+05,
     + 6.3962E+05, 6.4321E+05, 6.4682E+05, 6.5045E+05, 6.5410E+05,
     +  6.5777E+05, 6.6145E+05, 6.6515E+05, 6.6887E+05, 6.7261E+05,
     + 6.7636E+05, 6.8014E+05, 6.8393E+05, 6.8775E+05, 6.9158E+05,
     +  6.9543E+05, 6.9930E+05, 7.0318E+05, 7.0709E+05, 7.1101E+05,
     + 7.1496E+05, 7.1892E+05, 7.2291E+05, 7.2691E+05, 7.3093E+05,
     +  7.3497E+05, 7.3904E+05, 7.4312E+05, 7.4722E+05, 7.5134E+05,
     + 7.5548E+05, 7.5964E+05, 7.6382E+05, 7.6802E+05, 7.7225E+05,
     +  7.7649E+05, 7.8075E+05, 7.8504E+05, 7.8934E+05, 7.9366E+05,
     + 7.9801E+05, 8.0238E+05, 8.0676E+05, 8.1117E+05, 8.1560E+05,
     +  8.2006E+05, 8.2453E+05, 8.2902E+05, 8.3354E+05, 8.3808E+05,
     + 8.4264E+05, 8.4722E+05, 8.5182E+05, 8.5644E+05, 8.6109E+05,
     +  8.6576E+05, 8.7045E+05, 8.7517E+05, 8.7990E+05, 8.8466E+05,
     + 8.8944E+05, 8.9425E+05, 8.9907E+05, 9.0392E+05, 9.0880E+05,
     +  9.1369E+05, 9.1861E+05, 9.2356E+05, 9.2852E+05, 9.3351E+05,
     + 9.3853E+05, 9.4356E+05, 9.4862E+05, 9.5371E+05, 9.5882E+05,
     +  9.6395E+05, 9.6911E+05, 9.7429E+05, 9.7949E+05, 9.8472E+05,
     + 9.8998E+05, 9.9526E+05, 1.0006E+06, 1.0059E+06, 1.0112E+06,
     +  1.0166E+06, 1.0220E+06, 1.0275E+06, 1.0329E+06, 1.0384E+06,
     + 1.0439E+06, 1.0494E+06, 1.0550E+06, 1.0606E+06, 1.0662E+06,
     +  1.0718E+06, 1.0775E+06, 1.0832E+06, 1.0889E+06, 1.0946E+06,
     + 1.1004E+06, 1.1062E+06, 1.1120E+06, 1.1179E+06, 1.1237E+06,
     +  1.1296E+06, 1.1356E+06, 1.1415E+06, 1.1475E+06, 1.1535E+06,
     + 1.1596E+06, 1.1656E+06, 1.1717E+06, 1.1778E+06, 1.1840E+06,
     +  1.1902E+06, 1.1964E+06, 1.2026E+06, 1.2089E+06, 1.2152E+06,
     + 1.2215E+06, 1.2279E+06, 1.2342E+06, 1.2406E+06, 1.2471E+06,
     +  1.2536E+06, 1.2601E+06, 1.2666E+06, 1.2731E+06, 1.2797E+06,
     + 1.2864E+06, 1.2930E+06, 1.2997E+06, 1.3064E+06, 1.3131E+06,
     +  1.3199E+06, 1.3267E+06, 1.3335E+06, 1.3404E+06, 1.3473E+06,
     + 1.3542E+06, 1.3612E+06, 1.3682E+06, 1.3752E+06, 1.3822E+06,
     +  1.3893E+06, 1.3964E+06, 1.4036E+06, 1.4107E+06, 1.4180E+06,
     + 1.4252E+06, 1.4325E+06, 1.4398E+06, 1.4471E+06, 1.4545E+06,
     +  1.4619E+06, 1.4693E+06, 1.4768E+06, 1.4843E+06, 1.4919E+06,
     + 1.4994E+06, 1.5071E+06, 1.5147E+06, 1.5224E+06, 1.5301E+06,
     +  1.5378E+06, 1.5456E+06, 1.5534E+06, 1.5613E+06, 1.5692E+06,
     + 1.5771E+06, 1.5850E+06, 1.5930E+06, 1.6010E+06, 1.6091E+06,
     +  1.6172E+06, 1.6253E+06, 1.6335E+06, 1.6417E+06, 1.6499E+06,
     + 1.6582E+06, 1.6665E+06, 1.6749E+06, 1.6833E+06, 1.6917E+06,
     +  1.7002E+06, 1.7087E+06, 1.7172E+06, 1.7258E+06, 1.7344E+06,
     + 1.7430E+06, 1.7517E+06, 1.7604E+06, 1.7692E+06, 1.7780E+06,
     +  1.7868E+06, 1.7957E+06, 1.8046E+06, 1.8136E+06, 1.8226E+06,
     + 1.8316E+06, 1.8407E+06, 1.8498E+06, 1.8590E+06, 1.8682E+06,
     +  1.8774E+06, 1.8867E+06, 1.8960E+06, 1.9053E+06, 1.9147E+06,
     + 1.9242E+06, 1.9336E+06, 1.9432E+06, 1.9527E+06, 1.9623E+06,
     +  1.9720E+06, 1.9816E+06, 1.9914E+06, 2.0011E+06, 2.0109E+06,
     + 2.0208E+06, 2.0307E+06, 2.0406E+06, 2.0506E+06, 2.0606E+06,
     +  2.0707E+06, 2.0808E+06, 2.0909E+06, 2.1011E+06, 2.1114E+06,
     + 2.1217E+06, 2.1320E+06, 2.1424E+06, 2.1528E+06, 2.1632E+06,
     +  2.1737E+06, 2.1843E+06, 2.1949E+06, 2.2055E+06, 2.2162E+06,
     + 2.2269E+06, 2.2377E+06, 2.2485E+06, 2.2594E+06, 2.2703E+06,
     +  2.2813E+06, 2.2923E+06, 2.3033E+06, 2.3145E+06, 2.3256E+06,
     + 2.3368E+06, 2.3480E+06, 2.3593E+06, 2.3707E+06, 2.3821E+06,
     +  2.3935E+06, 2.4050E+06, 2.4165E+06, 2.4281E+06, 2.4397E+06,
     + 2.4514E+06, 2.4631E+06, 2.4749E+06, 2.4867E+06, 2.4986E+06,
     +  2.5105E+06, 2.5225E+06, 2.5345E+06, 2.5466E+06, 2.5587E+06,
     + 2.5709E+06, 2.5831E+06, 2.5954E+06, 2.6077E+06, 2.6201E+06,
     +  2.6326E+06, 2.6450E+06, 2.6576E+06, 2.6702E+06, 2.6828E+06,
     + 2.6955E+06, 2.7083E+06, 2.7211E+06, 2.7339E+06, 2.7468E+06,
     +  2.7598E+06, 2.7728E+06, 2.7859E+06, 2.7990E+06, 2.8122E+06,
     + 2.8254E+06, 2.8387E+06, 2.8520E+06, 2.8654E+06, 2.8789E+06,
     +  2.8924E+06, 2.9060E+06, 2.9196E+06, 2.9333E+06, 2.9470E+06,
     + 2.9608E+06, 2.9746E+06, 2.9885E+06, 3.0025E+06, 3.0165E+06,
     +  3.0306E+06, 3.0447E+06, 3.0589E+06, 3.0732E+06, 3.0875E+06,
     + 3.1018E+06, 3.1163E+06, 3.1308E+06, 3.1453E+06, 3.1599E+06,
     +  3.1746E+06, 3.1893E+06, 3.2041E+06, 3.2189E+06, 3.2338E+06,
     + 3.2488E+06, 3.2638E+06, 3.2789E+06, 3.2940E+06, 3.3092E+06,
     +  3.3245E+06, 3.3398E+06, 3.3552E+06, 3.3707E+06, 3.3862E+06,
     + 3.4018E+06, 3.4174E+06, 3.4332E+06, 3.4489E+06, 3.4648E+06,
     +  3.4807E+06, 3.4966E+06, 3.5127E+06, 3.5287E+06, 3.5449E+06,
     + 3.5611E+06, 3.5774E+06, 3.5938E+06, 3.6102E+06, 3.6267E+06,
     +  3.6432E+06, 3.6598E+06, 3.6765E+06, 3.6933E+06, 3.7101E+06,
     + 3.7270E+06, 3.7439E+06, 3.7610E+06, 3.7781E+06, 3.7952E+06,
     +  3.8124E+06, 3.8297E+06, 3.8471E+06, 3.8645E+06, 3.8820E+06,
     + 3.8996E+06, 3.9173E+06, 3.9350E+06, 3.9528E+06, 3.9706E+06,
     +  3.9885E+06, 4.0065E+06, 4.0246E+06, 4.0427E+06, 4.0609E+06,
     + 4.0792E+06, 4.0976E+06, 4.1160E+06, 4.1345E+06, 4.1531E+06,
     +  4.1717E+06, 4.1905E+06, 4.2093E+06, 4.2281E+06, 4.2471E+06,
     + 4.2661E+06, 4.2852E+06, 4.3043E+06, 4.3236E+06, 4.3429E+06,
     +  4.3623E+06, 4.3818E+06, 4.4013E+06, 4.4210E+06, 4.4407E+06,
     + 4.4604E+06, 4.4803E+06, 4.5002E+06, 4.5202E+06, 4.5403E+06,
     +  4.5605E+06, 4.5807E+06, 4.6011E+06, 4.6215E+06, 4.6420E+06,
     + 4.6625E+06, 4.6832E+06, 4.7039E+06, 4.7247E+06, 4.7456E+06,
     +  4.7666E+06, 4.7876E+06, 4.8088E+06, 4.8300E+06, 4.8513E+06,
     + 4.8726E+06, 4.8941E+06, 4.9157E+06, 4.9373E+06, 4.9590E+06,
     +  4.9808E+06, 5.0027E+06, 5.0246E+06, 5.0467E+06, 5.0688E+06,
     + 5.0910E+06, 5.1133E+06, 5.1357E+06, 5.1582E+06, 5.1808E+06,
     +  5.2034E+06, 5.2262E+06, 5.2490E+06, 5.2719E+06, 5.2949E+06,
     + 5.3180E+06, 5.3412E+06, 5.3644E+06, 5.3878E+06, 5.4112E+06,
     +  5.4348E+06, 5.4584E+06, 5.4821E+06, 5.5059E+06, 5.5298E+06,
     + 5.5538E+06, 5.5779E+06, 5.6021E+06, 5.6263E+06, 5.6507E+06,
     +  5.6751E+06, 5.6997E+06, 5.7243E+06, 5.7490E+06, 5.7739E+06,
     + 5.7988E+06, 5.8238E+06, 5.8489E+06, 5.8741E+06, 5.8994E+06,
     +  5.9248E+06, 5.9503E+06, 5.9759E+06, 6.0016E+06, 6.0273E+06,
     + 6.0532E+06, 6.0792E+06, 6.1053E+06, 6.1314E+06, 6.1577E+06,
     +  6.1841E+06, 6.2105E+06, 6.2371E+06, 6.2638E+06, 6.2905E+06,
     + 6.3174E+06, 6.3444E+06, 6.3714E+06, 6.3986E+06, 6.4259E+06,
     +  6.4533E+06, 6.4807E+06, 6.5083E+06, 6.5360E+06, 6.5638E+06,
     + 6.5917E+06, 6.6197E+06, 6.6478E+06, 6.6760E+06, 6.7043E+06,
     +  6.7327E+06, 6.7612E+06, 6.7899E+06, 6.8186E+06, 6.8474E+06,
     + 6.8764E+06, 6.9054E+06, 6.9346E+06, 6.9639E+06, 6.9933E+06,
     +  7.0228E+06, 7.0524E+06, 7.0821E+06, 7.1119E+06, 7.1418E+06,
     + 7.1718E+06, 7.2020E+06, 7.2323E+06, 7.2626E+06, 7.2931E+06,
     +  7.3237E+06, 7.3544E+06, 7.3853E+06, 7.4162E+06, 7.4473E+06,
     + 7.4784E+06, 7.5097E+06, 7.5411E+06, 7.5726E+06, 7.6042E+06,
     +  7.6360E+06, 7.6679E+06, 7.6998E+06, 7.7319E+06, 7.7641E+06,
     + 7.7965E+06, 7.8289E+06, 7.8615E+06, 7.8942E+06, 7.9270E+06,
     +  7.9599E+06, 7.9929E+06, 8.0261E+06, 8.0594E+06, 8.0928E+06,
     + 8.1263E+06, 8.1600E+06, 8.1938E+06, 8.2277E+06, 8.2617E+06,
     +  8.2958E+06, 8.3301E+06, 8.3645E+06, 8.3990E+06, 8.4337E+06,
     + 8.4684E+06, 8.5033E+06, 8.5384E+06, 8.5735E+06, 8.6088E+06,
     +  8.6442E+06, 8.6797E+06, 8.7154E+06, 8.7512E+06, 8.7871E+06,
     + 8.8231E+06, 8.8593E+06, 8.8956E+06, 8.9321E+06, 8.9686E+06,
     +  9.0053E+06, 9.0422E+06, 9.0791E+06, 9.1162E+06, 9.1535E+06,
     + 9.1908E+06, 9.2283E+06, 9.2660E+06, 9.3037E+06, 9.3416E+06,
     +  9.3797E+06, 9.4179E+06, 9.4562E+06, 9.4946E+06, 9.5332E+06,
     + 9.5719E+06, 9.6108E+06, 9.6498E+06, 9.6889E+06, 9.7282E+06,
     +  9.7676E+06, 9.8072E+06, 9.8469E+06, 9.8868E+06, 9.9267E+06,
     + 9.9669E+06, 1.0007E+07, 1.0048E+07, 1.0088E+07, 1.0129E+07,
     +  1.0170E+07, 1.0211E+07, 1.0252E+07, 1.0293E+07, 1.0334E+07,
     + 1.0376E+07, 1.0418E+07, 1.0460E+07, 1.0502E+07, 1.0544E+07,
     +  1.0586E+07, 1.0629E+07, 1.0671E+07, 1.0714E+07, 1.0757E+07,
     + 1.0800E+07, 1.0843E+07, 1.0887E+07, 1.0930E+07, 1.0974E+07,
     +  1.1018E+07, 1.1062E+07, 1.1106E+07, 1.1150E+07, 1.1194E+07,
     + 1.1239E+07, 1.1284E+07, 1.1329E+07, 1.1374E+07, 1.1419E+07,
     +  1.1464E+07, 1.1510E+07, 1.1556E+07, 1.1601E+07, 1.1647E+07,
     + 1.1694E+07, 1.1740E+07, 1.1787E+07, 1.1833E+07, 1.1880E+07,
     +  1.1927E+07, 1.1974E+07, 1.2021E+07, 1.2069E+07, 1.2117E+07,
     + 1.2164E+07, 1.2212E+07, 1.2261E+07, 1.2309E+07, 1.2357E+07,
     +  1.2406E+07, 1.2455E+07, 1.2504E+07, 1.2553E+07, 1.2602E+07,
     + 1.2652E+07, 1.2701E+07, 1.2751E+07, 1.2801E+07, 1.2851E+07,
     +  1.2902E+07, 1.2952E+07, 1.3003E+07, 1.3054E+07, 1.3105E+07,
     + 1.3156E+07, 1.3208E+07, 1.3259E+07, 1.3311E+07, 1.3363E+07,
     +  1.3415E+07, 1.3467E+07, 1.3520E+07, 1.3572E+07, 1.3625E+07,
     + 1.3678E+07, 1.3731E+07, 1.3785E+07, 1.3838E+07, 1.3892E+07,
     +  1.3946E+07, 1.4000E+07, 1.4054E+07, 1.4109E+07, 1.4164E+07,
     + 1.4218E+07, 1.4273E+07, 1.4329E+07, 1.4384E+07, 1.4440E+07,
     +  1.4495E+07, 1.4551E+07, 1.4608E+07, 1.4664E+07, 1.4721E+07,
     + 1.4777E+07, 1.4834E+07, 1.4891E+07, 1.4949E+07, 1.5006E+07,
     +  1.5064E+07, 1.5122E+07, 1.5180E+07, 1.5238E+07, 1.5297E+07,
     + 1.5355E+07, 1.5414E+07, 1.5473E+07, 1.5533E+07, 1.5592E+07,
     +  1.5652E+07, 1.5712E+07, 1.5772E+07, 1.5832E+07, 1.5892E+07,
     + 1.5953E+07, 1.6014E+07, 1.6075E+07, 1.6136E+07, 1.6198E+07,
     +  1.6260E+07, 1.6322E+07, 1.6384E+07, 1.6446E+07, 1.6509E+07,
     + 1.6571E+07, 1.6634E+07, 1.6697E+07, 1.6761E+07, 1.6824E+07,
     +  1.6888E+07, 1.6952E+07, 1.7017E+07, 1.7081E+07, 1.7146E+07,
     + 1.7211E+07, 1.7276E+07, 1.7341E+07, 1.7406E+07, 1.7472E+07,
     +  1.7538E+07, 1.7604E+07, 1.7671E+07, 1.7737E+07, 1.7804E+07,
     + 1.7871E+07, 1.7939E+07, 1.8006E+07, 1.8074E+07, 1.8142E+07,
     +  1.8210E+07, 1.8278E+07, 1.8347E+07, 1.8416E+07, 1.8485E+07,
     + 1.8554E+07, 1.8624E+07, 1.8694E+07, 1.8764E+07, 1.8834E+07,
     +  1.8904E+07, 1.8975E+07, 1.9046E+07, 1.9117E+07, 1.9188E+07,
     + 1.9260E+07, 1.9332E+07, 1.9404E+07, 1.9476E+07, 1.9549E+07,
     +  1.9622E+07, 1.9695E+07, 1.9768E+07, 1.9842E+07, 1.9915E+07,
     + 1.9989E+07, 2.0064E+07, 2.0138E+07, 2.0213E+07, 2.0288E+07,
     +  2.0363E+07, 2.0438E+07, 2.0514E+07, 2.0590E+07, 2.0666E+07,
     + 2.0743E+07, 2.0819E+07, 2.0896E+07, 2.0974E+07, 2.1051E+07,
     +  2.1129E+07, 2.1207E+07, 2.1285E+07, 2.1363E+07, 2.1442E+07,
     + 2.1521E+07, 2.1600E+07, 2.1680E+07, 2.1759E+07, 2.1839E+07,
     +  2.1920E+07, 2.2000E+07, 2.2081E+07, 2.2162E+07, 2.2243E+07,
     + 2.2325E+07, 2.2407E+07, 2.2489E+07, 2.2571E+07, 2.2654E+07,
     +  2.2737E+07, 2.2820E+07, 2.2903E+07, 2.2987E+07, 2.3071E+07,
     + 2.3155E+07, 2.3239E+07, 2.3324E+07, 2.3409E+07, 2.3495E+07,
     +  2.3580E+07, 2.3666E+07, 2.3752E+07, 2.3839E+07, 2.3925E+07,
     + 2.4012E+07, 2.4099E+07, 2.4187E+07, 2.4275E+07, 2.4363E+07,
     +  2.4451E+07, 2.4540E+07, 2.4629E+07, 2.4718E+07, 2.4807E+07,
     + 2.4897E+07, 2.4987E+07, 2.5077E+07, 2.5168E+07, 2.5259E+07,
     +  2.5350E+07, 2.5442E+07, 2.5533E+07, 2.5625E+07, 2.5718E+07,
     + 2.5810E+07, 2.5903E+07, 2.5997E+07, 2.6090E+07, 2.6184E+07,
     +  2.6278E+07, 2.6373E+07, 2.6467E+07, 2.6562E+07, 2.6658E+07,
     + 2.6753E+07, 2.6849E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 3,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_C2H6  (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 2),QofT( 2,NT),Q(NT)
      data (Tmx(I),I=1,2)/3500.,3500./
      data xgj/ 1.,2./
c...     C2H6
c...        --      1221
      data (QofT( 1,J),J=1,2408)/ 1.5003E+01, 3.5616E+01, 6.1180E+01,
     +  9.0693E+01, 1.2365E+02, 1.5975E+02, 1.9877E+02, 2.4054E+02,
     + 2.8493E+02, 3.3182E+02, 3.8109E+02, 4.3265E+02, 4.8642E+02,
     +  5.4230E+02, 6.0023E+02, 6.6014E+02, 7.2196E+02, 7.8563E+02,
     + 8.5110E+02, 9.1833E+02, 1.0578E+03, 1.2038E+03, 1.3561E+03,
     +  1.5142E+03, 1.6781E+03, 1.8475E+03, 2.0223E+03, 2.2023E+03,
     + 2.3873E+03, 2.5773E+03, 2.7721E+03, 2.9715E+03, 3.1756E+03,
     +  3.3842E+03, 3.5972E+03, 3.8146E+03, 4.0362E+03, 4.2622E+03,
     + 4.4924E+03, 4.7267E+03, 4.9651E+03, 5.2077E+03, 5.4544E+03,
     +  5.7051E+03, 5.9599E+03, 6.2188E+03, 6.4817E+03, 6.7488E+03,
     + 7.0199E+03, 7.2951E+03, 7.5744E+03, 7.8578E+03, 8.1454E+03,
     +  8.4372E+03, 8.7330E+03, 9.0331E+03, 9.3375E+03, 9.6461E+03,
     + 9.9591E+03, 1.0276E+04, 1.0598E+04, 1.0924E+04, 1.1254E+04,
     +  1.1589E+04, 1.1928E+04, 1.2272E+04, 1.2621E+04, 1.2974E+04,
     + 1.3332E+04, 1.3694E+04, 1.4061E+04, 1.4433E+04, 1.4809E+04,
     +  1.5191E+04, 1.5577E+04, 1.5968E+04, 1.6364E+04, 1.6765E+04,
     + 1.7172E+04, 1.7583E+04, 1.7999E+04, 1.8421E+04, 1.8848E+04,
     +  1.9280E+04, 1.9717E+04, 2.0160E+04, 2.0608E+04, 2.1062E+04,
     + 2.1521E+04, 2.1985E+04, 2.2456E+04, 2.2932E+04, 2.3413E+04,
     +  2.3901E+04, 2.4394E+04, 2.4893E+04, 2.5399E+04, 2.5910E+04,
     + 2.6427E+04, 2.6951E+04, 2.7480E+04, 2.8017E+04, 2.8559E+04,
     +  2.9108E+04, 2.9663E+04, 3.0225E+04, 3.0793E+04, 3.1368E+04,
     + 3.1950E+04, 3.2539E+04, 3.3135E+04, 3.3738E+04, 3.4348E+04,
     +  3.4965E+04, 3.5590E+04, 3.6221E+04, 3.6861E+04, 3.7507E+04,
     + 3.8162E+04, 3.8824E+04, 3.9494E+04, 4.0173E+04, 4.0859E+04,
     +  4.1553E+04, 4.2256E+04, 4.2967E+04, 4.3686E+04, 4.4414E+04,
     + 4.5151E+04, 4.5896E+04, 4.6650E+04, 4.7414E+04, 4.8186E+04,
     +  4.8968E+04, 4.9760E+04, 5.0561E+04, 5.1371E+04, 5.2191E+04,
     + 5.3022E+04, 5.3862E+04, 5.4713E+04, 5.5574E+04, 5.6445E+04,
     +  5.7327E+04, 5.8220E+04, 5.9124E+04, 6.0039E+04, 6.0965E+04,
     + 6.1903E+04, 6.2852E+04, 6.3813E+04, 6.4786E+04, 6.5771E+04,
     +  6.6768E+04, 6.7778E+04, 6.8800E+04, 6.9835E+04, 7.0883E+04,
     + 7.1944E+04, 7.3018E+04, 7.4106E+04, 7.5208E+04, 7.6323E+04,
     +  7.7453E+04, 7.8597E+04, 7.9755E+04, 8.0928E+04, 8.2116E+04,
     + 8.3319E+04, 8.4538E+04, 8.5772E+04, 8.7023E+04, 8.8289E+04,
     +  8.9571E+04, 9.0870E+04, 9.2186E+04, 9.3518E+04, 9.4868E+04,
     + 9.6235E+04, 9.7620E+04, 9.9024E+04, 1.0044E+05, 1.0188E+05,
     +  1.0334E+05, 1.0482E+05, 1.0632E+05, 1.0783E+05, 1.0937E+05,
     + 1.1093E+05, 1.1250E+05, 1.1410E+05, 1.1572E+05, 1.1736E+05,
     +  1.1902E+05, 1.2071E+05, 1.2241E+05, 1.2414E+05, 1.2589E+05,
     + 1.2767E+05, 1.2947E+05, 1.3129E+05, 1.3314E+05, 1.3501E+05,
     +  1.3691E+05, 1.3883E+05, 1.4077E+05, 1.4275E+05, 1.4475E+05,
     + 1.4678E+05, 1.4883E+05, 1.5091E+05, 1.5302E+05, 1.5516E+05,
     +  1.5733E+05, 1.5952E+05, 1.6175E+05, 1.6400E+05, 1.6629E+05,
     + 1.6861E+05, 1.7096E+05, 1.7334E+05, 1.7575E+05, 1.7820E+05,
     +  1.8068E+05, 1.8319E+05, 1.8574E+05, 1.8832E+05, 1.9093E+05,
     + 1.9358E+05, 1.9627E+05, 1.9900E+05, 2.0176E+05, 2.0456E+05,
     +  2.0740E+05, 2.1027E+05, 2.1319E+05, 2.1615E+05, 2.1914E+05,
     + 2.2218E+05, 2.2526E+05, 2.2838E+05, 2.3154E+05, 2.3475E+05,
     +  2.3800E+05, 2.4130E+05, 2.4464E+05, 2.4803E+05, 2.5146E+05,
     + 2.5494E+05, 2.5847E+05, 2.6205E+05, 2.6568E+05, 2.6935E+05,
     +  2.7308E+05, 2.7686E+05, 2.8069E+05, 2.8457E+05, 2.8851E+05,
     + 2.9250E+05, 2.9655E+05, 3.0065E+05, 3.0481E+05, 3.0902E+05,
     +  3.1330E+05, 3.1763E+05, 3.2203E+05, 3.2648E+05, 3.3100E+05,
     + 3.3558E+05, 3.4022E+05, 3.4492E+05, 3.4969E+05, 3.5453E+05,
     +  3.5943E+05, 3.6441E+05, 3.6945E+05, 3.7456E+05, 3.7974E+05,
     + 3.8499E+05, 3.9032E+05, 3.9572E+05, 4.0119E+05, 4.0674E+05,
     +  4.1237E+05, 4.1807E+05, 4.2385E+05, 4.2972E+05, 4.3566E+05,
     + 4.4169E+05, 4.4780E+05, 4.5400E+05, 4.6028E+05, 4.6665E+05,
     +  4.7311E+05, 4.7966E+05, 4.8629E+05, 4.9302E+05, 4.9985E+05,
     + 5.0676E+05, 5.1378E+05, 5.2089E+05, 5.2810E+05, 5.3541E+05,
     +  5.4282E+05, 5.5033E+05, 5.5795E+05, 5.6568E+05, 5.7351E+05,
     + 5.8145E+05, 5.8950E+05, 5.9766E+05, 6.0593E+05, 6.1432E+05,
     +  6.2283E+05, 6.3145E+05, 6.4019E+05, 6.4906E+05, 6.5805E+05,
     + 6.6716E+05, 6.7640E+05, 6.8576E+05, 6.9526E+05, 7.0489E+05,
     +  7.1465E+05, 7.2455E+05, 7.3458E+05, 7.4475E+05, 7.5507E+05,
     + 7.6552E+05, 7.7612E+05, 7.8687E+05, 7.9777E+05, 8.0882E+05,
     +  8.2002E+05, 8.3138E+05, 8.4289E+05, 8.5456E+05, 8.6640E+05,
     + 8.7840E+05, 8.9056E+05, 9.0290E+05, 9.1540E+05, 9.2808E+05,
     +  9.4093E+05, 9.5396E+05, 9.6717E+05, 9.8056E+05, 9.9414E+05,
     + 1.0079E+06, 1.0219E+06, 1.0360E+06, 1.0504E+06, 1.0649E+06,
     +  1.0796E+06, 1.0946E+06, 1.1097E+06, 1.1251E+06, 1.1407E+06,
     + 1.1565E+06, 1.1725E+06, 1.1887E+06, 1.2052E+06, 1.2218E+06,
     +  1.2388E+06, 1.2559E+06, 1.2733E+06, 1.2909E+06, 1.3088E+06,
     + 1.3269E+06, 1.3452E+06, 1.3638E+06, 1.3827E+06, 1.4018E+06,
     +  1.4212E+06, 1.4409E+06, 1.4608E+06, 1.4810E+06, 1.5015E+06,
     + 1.5223E+06, 1.5433E+06, 1.5646E+06, 1.5863E+06, 1.6082E+06,
     +  1.6304E+06, 1.6530E+06, 1.6758E+06, 1.6990E+06, 1.7224E+06,
     + 1.7462E+06, 1.7704E+06, 1.7948E+06, 1.8196E+06, 1.8447E+06,
     +  1.8702E+06, 1.8960E+06, 1.9222E+06, 1.9488E+06, 1.9757E+06,
     + 2.0029E+06, 2.0306E+06, 2.0586E+06, 2.0870E+06, 2.1158E+06,
     +  2.1450E+06, 2.1746E+06, 2.2046E+06, 2.2350E+06, 2.2658E+06,
     + 2.2970E+06, 2.3287E+06, 2.3608E+06, 2.3933E+06, 2.4263E+06,
     +  2.4597E+06, 2.4936E+06, 2.5279E+06, 2.5628E+06, 2.5981E+06,
     + 2.6338E+06, 2.6701E+06, 2.7068E+06, 2.7441E+06, 2.7819E+06,
     +  2.8201E+06, 2.8589E+06, 2.8982E+06, 2.9381E+06, 2.9785E+06,
     + 3.0195E+06, 3.0610E+06, 3.1030E+06, 3.1457E+06, 3.1889E+06,
     +  3.2327E+06, 3.2771E+06, 3.3221E+06, 3.3677E+06, 3.4139E+06,
     + 3.4608E+06, 3.5083E+06, 3.5564E+06, 3.6052E+06, 3.6547E+06,
     +  3.7048E+06, 3.7556E+06, 3.8071E+06, 3.8592E+06, 3.9121E+06,
     + 3.9657E+06, 4.0200E+06, 4.0751E+06, 4.1309E+06, 4.1875E+06,
     +  4.2448E+06, 4.3029E+06, 4.3617E+06, 4.4214E+06, 4.4819E+06,
     + 4.5432E+06, 4.6053E+06, 4.6682E+06, 4.7320E+06, 4.7967E+06,
     +  4.8622E+06, 4.9286E+06, 4.9959E+06, 5.0641E+06, 5.1332E+06,
     + 5.2033E+06, 5.2742E+06, 5.3462E+06, 5.4191E+06, 5.4930E+06,
     +  5.5678E+06, 5.6437E+06, 5.7206E+06, 5.7985E+06, 5.8775E+06,
     + 5.9575E+06, 6.0386E+06, 6.1208E+06, 6.2041E+06, 6.2885E+06,
     +  6.3740E+06, 6.4606E+06, 6.5485E+06, 6.6375E+06, 6.7276E+06,
     + 6.8190E+06, 6.9116E+06, 7.0055E+06, 7.1005E+06, 7.1969E+06,
     +  7.2945E+06, 7.3935E+06, 7.4937E+06, 7.5953E+06, 7.6982E+06,
     + 7.8026E+06, 7.9082E+06, 8.0153E+06, 8.1239E+06, 8.2338E+06,
     +  8.3452E+06, 8.4581E+06, 8.5725E+06, 8.6885E+06, 8.8059E+06,
     + 8.9249E+06, 9.0455E+06, 9.1677E+06, 9.2915E+06, 9.4169E+06,
     +  9.5440E+06, 9.6728E+06, 9.8032E+06, 9.9354E+06, 1.0069E+07,
     + 1.0205E+07, 1.0343E+07, 1.0482E+07, 1.0623E+07, 1.0766E+07,
     +  1.0911E+07, 1.1058E+07, 1.1206E+07, 1.1357E+07, 1.1510E+07,
     + 1.1665E+07, 1.1821E+07, 1.1980E+07, 1.2141E+07, 1.2304E+07,
     +  1.2469E+07, 1.2636E+07, 1.2806E+07, 1.2977E+07, 1.3151E+07,
     + 1.3328E+07, 1.3506E+07, 1.3687E+07, 1.3870E+07, 1.4056E+07,
     +  1.4244E+07, 1.4435E+07, 1.4628E+07, 1.4823E+07, 1.5021E+07,
     + 1.5222E+07, 1.5425E+07, 1.5631E+07, 1.5840E+07, 1.6051E+07,
     +  1.6265E+07, 1.6482E+07, 1.6702E+07, 1.6924E+07, 1.7150E+07,
     + 1.7378E+07, 1.7610E+07, 1.7844E+07, 1.8081E+07, 1.8322E+07,
     +  1.8566E+07, 1.8812E+07, 1.9062E+07, 1.9316E+07, 1.9572E+07,
     + 1.9832E+07, 2.0095E+07, 2.0362E+07, 2.0632E+07, 2.0905E+07,
     +  2.1182E+07, 2.1463E+07, 2.1747E+07, 2.2035E+07, 2.2327E+07,
     + 2.2622E+07, 2.2921E+07, 2.3224E+07, 2.3531E+07, 2.3842E+07,
     +  2.4157E+07, 2.4476E+07, 2.4799E+07, 2.5127E+07, 2.5458E+07,
     + 2.5794E+07, 2.6134E+07, 2.6478E+07, 2.6827E+07, 2.7180E+07,
     +  2.7538E+07, 2.7901E+07, 2.8268E+07, 2.8639E+07, 2.9016E+07,
     + 2.9397E+07, 2.9783E+07, 3.0174E+07, 3.0570E+07, 3.0972E+07,
     +  3.1378E+07, 3.1789E+07, 3.2206E+07, 3.2628E+07, 3.3056E+07,
     + 3.3488E+07, 3.3927E+07, 3.4371E+07, 3.4820E+07, 3.5276E+07,
     +  3.5737E+07, 3.6204E+07, 3.6677E+07, 3.7156E+07, 3.7640E+07,
     + 3.8132E+07, 3.8629E+07, 3.9133E+07, 3.9643E+07, 4.0159E+07,
     +  4.0682E+07, 4.1212E+07, 4.1748E+07, 4.2291E+07, 4.2841E+07,
     + 4.3398E+07, 4.3962E+07, 4.4533E+07, 4.5112E+07, 4.5697E+07,
     +  4.6290E+07, 4.6891E+07, 4.7499E+07, 4.8114E+07, 4.8738E+07,
     + 4.9369E+07, 5.0008E+07, 5.0655E+07, 5.1310E+07, 5.1974E+07,
     +  5.2646E+07, 5.3326E+07, 5.4015E+07, 5.4712E+07, 5.5419E+07,
     + 5.6134E+07, 5.6858E+07, 5.7591E+07, 5.8333E+07, 5.9084E+07,
     +  5.9845E+07, 6.0616E+07, 6.1396E+07, 6.2185E+07, 6.2985E+07,
     + 6.3795E+07, 6.4614E+07, 6.5444E+07, 6.6284E+07, 6.7135E+07,
     +  6.7996E+07, 6.8868E+07, 6.9751E+07, 7.0644E+07, 7.1549E+07,
     + 7.2465E+07, 7.3393E+07, 7.4332E+07, 7.5282E+07, 7.6245E+07,
     +  7.7219E+07, 7.8205E+07, 7.9204E+07, 8.0214E+07, 8.1238E+07,
     + 8.2274E+07, 8.3323E+07, 8.4384E+07, 8.5459E+07, 8.6547E+07,
     +  8.7649E+07, 8.8764E+07, 8.9893E+07, 9.1035E+07, 9.2192E+07,
     + 9.3363E+07, 9.4548E+07, 9.5748E+07, 9.6963E+07, 9.8193E+07,
     +  9.9437E+07, 1.0070E+08, 1.0197E+08, 1.0326E+08, 1.0457E+08,
     + 1.0589E+08, 1.0723E+08, 1.0859E+08, 1.0996E+08, 1.1135E+08,
     +  1.1275E+08, 1.1418E+08, 1.1562E+08, 1.1708E+08, 1.1855E+08,
     + 1.2004E+08, 1.2156E+08, 1.2309E+08, 1.2464E+08, 1.2620E+08,
     +  1.2779E+08, 1.2940E+08, 1.3102E+08, 1.3267E+08, 1.3433E+08,
     + 1.3602E+08, 1.3772E+08, 1.3945E+08, 1.4120E+08, 1.4297E+08,
     +  1.4476E+08, 1.4657E+08, 1.4840E+08, 1.5026E+08, 1.5214E+08,
     + 1.5404E+08, 1.5596E+08, 1.5791E+08, 1.5988E+08, 1.6187E+08,
     +  1.6389E+08, 1.6593E+08, 1.6800E+08, 1.7009E+08, 1.7221E+08,
     + 1.7435E+08, 1.7652E+08, 1.7872E+08, 1.8094E+08, 1.8319E+08,
     +  1.8546E+08, 1.8776E+08, 1.9009E+08, 1.9245E+08, 1.9483E+08,
     + 1.9725E+08, 1.9969E+08, 2.0216E+08, 2.0466E+08, 2.0720E+08,
     +  2.0976E+08, 2.1235E+08, 2.1497E+08, 2.1763E+08, 2.2031E+08,
     + 2.2303E+08, 2.2578E+08, 2.2857E+08, 2.3138E+08, 2.3423E+08,
     +  2.3712E+08, 2.4003E+08, 2.4299E+08, 2.4597E+08, 2.4900E+08,
     + 2.5206E+08, 2.5515E+08, 2.5828E+08, 2.6145E+08, 2.6466E+08,
     +  2.6790E+08, 2.7118E+08, 2.7451E+08, 2.7787E+08, 2.8127E+08,
     + 2.8471E+08, 2.8819E+08, 2.9171E+08, 2.9527E+08, 2.9888E+08,
     +  3.0253E+08, 3.0622E+08, 3.0995E+08, 3.1373E+08, 3.1755E+08,
     + 3.2142E+08, 3.2533E+08, 3.2929E+08, 3.3330E+08, 3.3735E+08,
     +  3.4145E+08, 3.4560E+08, 3.4979E+08, 3.5404E+08, 3.5833E+08,
     + 3.6268E+08, 3.6707E+08, 3.7152E+08, 3.7602E+08, 3.8057E+08,
     +  3.8518E+08, 3.8983E+08, 3.9455E+08, 3.9931E+08, 4.0414E+08,
     + 4.0902E+08, 4.1395E+08, 4.1895E+08, 4.2400E+08, 4.2911E+08,
     +  4.3428E+08, 4.3951E+08, 4.4479E+08, 4.5015E+08, 4.5556E+08,
     + 4.6103E+08, 4.6657E+08, 4.7218E+08, 4.7784E+08, 4.8358E+08,
     +  4.8938E+08, 4.9524E+08, 5.0118E+08, 5.0718E+08, 5.1325E+08,
     + 5.1940E+08, 5.2561E+08, 5.3189E+08, 5.3825E+08, 5.4468E+08,
     +  5.5118E+08, 5.5776E+08, 5.6442E+08, 5.7115E+08, 5.7795E+08,
     + 5.8484E+08, 5.9180E+08, 5.9885E+08, 6.0597E+08, 6.1318E+08,
     +  6.2047E+08, 6.2784E+08, 6.3530E+08, 6.4284E+08, 6.5047E+08,
     + 6.5818E+08, 6.6599E+08, 6.7388E+08, 6.8186E+08, 6.8993E+08,
     +  6.9810E+08, 7.0636E+08, 7.1471E+08, 7.2315E+08, 7.3170E+08,
     + 7.4034E+08, 7.4907E+08, 7.5791E+08, 7.6685E+08, 7.7589E+08,
     +  7.8503E+08, 7.9427E+08, 8.0362E+08, 8.1308E+08, 8.2264E+08,
     + 8.3231E+08, 8.4209E+08, 8.5198E+08, 8.6198E+08, 8.7210E+08,
     +  8.8232E+08, 8.9267E+08, 9.0313E+08, 9.1371E+08, 9.2440E+08,
     + 9.3522E+08, 9.4616E+08, 9.5722E+08, 9.6841E+08, 9.7972E+08,
     +  9.9116E+08, 1.0027E+09, 1.0144E+09, 1.0263E+09, 1.0382E+09,
     + 1.0503E+09, 1.0625E+09, 1.0749E+09, 1.0874E+09, 1.1001E+09,
     +  1.1128E+09, 1.1258E+09, 1.1389E+09, 1.1521E+09, 1.1654E+09,
     + 1.1790E+09, 1.1926E+09, 1.2064E+09, 1.2204E+09, 1.2345E+09,
     +  1.2488E+09, 1.2633E+09, 1.2779E+09, 1.2926E+09, 1.3076E+09,
     + 1.3227E+09, 1.3379E+09, 1.3534E+09, 1.3690E+09, 1.3847E+09,
     +  1.4007E+09, 1.4168E+09, 1.4331E+09, 1.4496E+09, 1.4663E+09,
     + 1.4831E+09, 1.5002E+09, 1.5174E+09, 1.5348E+09, 1.5524E+09,
     +  1.5702E+09, 1.5882E+09, 1.6064E+09, 1.6248E+09, 1.6434E+09,
     + 1.6622E+09, 1.6812E+09, 1.7004E+09, 1.7199E+09, 1.7395E+09,
     +  1.7593E+09, 1.7794E+09, 1.7997E+09, 1.8202E+09, 1.8410E+09,
     + 1.8619E+09, 1.8831E+09, 1.9046E+09, 1.9262E+09, 1.9481E+09,
     +  1.9703E+09, 1.9926E+09, 2.0153E+09, 2.0381E+09, 2.0612E+09,
     + 2.0846E+09, 2.1082E+09, 2.1321E+09, 2.1563E+09, 2.1807E+09,
     +  2.2053E+09, 2.2303E+09, 2.2555E+09, 2.2809E+09, 2.3067E+09,
     + 2.3327E+09, 2.3590E+09, 2.3856E+09, 2.4125E+09, 2.4397E+09,
     +  2.4672E+09, 2.4949E+09, 2.5230E+09, 2.5514E+09, 2.5801E+09,
     + 2.6090E+09, 2.6383E+09, 2.6680E+09, 2.6979E+09, 2.7281E+09,
     +  2.7587E+09, 2.7896E+09, 2.8209E+09, 2.8524E+09, 2.8844E+09,
     + 2.9166E+09, 2.9492E+09, 2.9822E+09, 3.0155E+09, 3.0491E+09,
     +  3.0832E+09, 3.1175E+09, 3.1523E+09, 3.1874E+09, 3.2229E+09,
     + 3.2588E+09, 3.2951E+09, 3.3317E+09, 3.3688E+09, 3.4062E+09,
     +  3.4440E+09, 3.4823E+09, 3.5209E+09, 3.5600E+09, 3.5994E+09,
     + 3.6393E+09, 3.6796E+09, 3.7204E+09, 3.7615E+09, 3.8031E+09,
     +  3.8452E+09, 3.8877E+09, 3.9306E+09, 3.9740E+09, 4.0179E+09,
     + 4.0622E+09, 4.1070E+09, 4.1523E+09, 4.1980E+09, 4.2443E+09,
     +  4.2910E+09, 4.3382E+09, 4.3859E+09, 4.4341E+09, 4.4828E+09,
     + 4.5320E+09, 4.5818E+09, 4.6320E+09, 4.6828E+09, 4.7342E+09,
     +  4.7860E+09, 4.8384E+09, 4.8914E+09, 4.9449E+09, 4.9990E+09,
     + 5.0537E+09, 5.1089E+09, 5.1647E+09, 5.2210E+09, 5.2780E+09,
     +  5.3356E+09, 5.3937E+09, 5.4525E+09, 5.5119E+09, 5.5719E+09,
     + 5.6325E+09, 5.6938E+09, 5.7557E+09, 5.8183E+09, 5.8815E+09,
     +  5.9453E+09, 6.0098E+09, 6.0750E+09, 6.1409E+09, 6.2074E+09,
     + 6.2747E+09, 6.3426E+09, 6.4113E+09, 6.4806E+09, 6.5507E+09,
     +  6.6215E+09, 6.6930E+09, 6.7653E+09, 6.8383E+09, 6.9121E+09,
     + 6.9866E+09, 7.0619E+09, 7.1380E+09, 7.2148E+09, 7.2925E+09,
     +  7.3710E+09, 7.4502E+09, 7.5303E+09, 7.6112E+09, 7.6929E+09,
     + 7.7755E+09, 7.8590E+09, 7.9432E+09, 8.0284E+09, 8.1144E+09,
     +  8.2013E+09, 8.2891E+09, 8.3778E+09, 8.4674E+09, 8.5579E+09,
     + 8.6494E+09, 8.7417E+09, 8.8351E+09, 8.9293E+09, 9.0246E+09,
     +  9.1208E+09, 9.2180E+09, 9.3162E+09, 9.4154E+09, 9.5156E+09,
     + 9.6168E+09, 9.7190E+09, 9.8223E+09, 9.9267E+09, 1.0032E+10,
     +  1.0139E+10, 1.0246E+10, 1.0355E+10, 1.0464E+10, 1.0575E+10,
     + 1.0687E+10, 1.0800E+10, 1.0915E+10, 1.1030E+10, 1.1147E+10,
     +  1.1264E+10, 1.1383E+10, 1.1504E+10, 1.1625E+10, 1.1748E+10,
     + 1.1871E+10, 1.1996E+10, 1.2123E+10, 1.2250E+10, 1.2379E+10,
     +  1.2509E+10, 1.2641E+10, 1.2774E+10, 1.2908E+10, 1.3044E+10,
     + 1.3180E+10, 1.3319E+10, 1.3458E+10, 1.3599E+10, 1.3742E+10,
     +  1.3886E+10, 1.4031E+10, 1.4178E+10, 1.4326E+10, 1.4476E+10,
     + 1.4627E+10, 1.4780E+10, 1.4934E+10, 1.5090E+10, 1.5247E+10,
     +  1.5406E+10, 1.5566E+10, 1.5728E+10, 1.5892E+10, 1.6058E+10,
     + 1.6225E+10, 1.6393E+10, 1.6563E+10, 1.6736E+10, 1.6909E+10,
     +  1.7085E+10, 1.7262E+10, 1.7441E+10, 1.7622E+10, 1.7804E+10,
     + 1.7988E+10, 1.8174E+10, 1.8362E+10, 1.8552E+10, 1.8744E+10,
     +  1.8938E+10, 1.9133E+10, 1.9331E+10, 1.9530E+10, 1.9731E+10,
     + 1.9935E+10, 2.0140E+10, 2.0347E+10, 2.0557E+10, 2.0768E+10,
     +  2.0982E+10, 2.1198E+10, 2.1415E+10, 2.1635E+10, 2.1857E+10,
     + 2.2082E+10, 2.2308E+10, 2.2537E+10, 2.2768E+10, 2.3001E+10,
     +  2.3236E+10, 2.3474E+10, 2.3714E+10, 2.3957E+10, 2.4201E+10,
     + 2.4449E+10, 2.4698E+10, 2.4950E+10, 2.5205E+10, 2.5462E+10,
     +  2.5721E+10, 2.5983E+10, 2.6248E+10, 2.6515E+10, 2.6785E+10,
     + 2.7057E+10, 2.7332E+10, 2.7610E+10, 2.7890E+10, 2.8173E+10,
     +  2.8459E+10, 2.8747E+10, 2.9039E+10, 2.9333E+10, 2.9630E+10,
     + 2.9930E+10, 3.0233E+10, 3.0539E+10, 3.0847E+10, 3.1159E+10,
     +  3.1474E+10, 3.1792E+10, 3.2112E+10, 3.2436E+10, 3.2763E+10,
     + 3.3094E+10, 3.3427E+10, 3.3763E+10, 3.4103E+10, 3.4446E+10,
     +  3.4793E+10, 3.5142E+10, 3.5496E+10, 3.5852E+10, 3.6212E+10,
     + 3.6575E+10, 3.6942E+10, 3.7312E+10, 3.7686E+10, 3.8064E+10,
     +  3.8445E+10, 3.8829E+10, 3.9218E+10, 3.9610E+10, 4.0006E+10,
     + 4.0405E+10, 4.0809E+10, 4.1216E+10, 4.1627E+10, 4.2042E+10,
     +  4.2461E+10, 4.2884E+10, 4.3311E+10, 4.3742E+10, 4.4177E+10,
     + 4.4617E+10, 4.5060E+10, 4.5508E+10, 4.5960E+10, 4.6416E+10,
     +  4.6877E+10, 4.7342E+10, 4.7811E+10, 4.8285E+10, 4.8763E+10,
     + 4.9246E+10, 4.9733E+10, 5.0225E+10, 5.0722E+10, 5.1223E+10,
     +  5.1729E+10, 5.2240E+10, 5.2755E+10, 5.3276E+10, 5.3801E+10,
     + 5.4331E+10, 5.4867E+10, 5.5407E+10, 5.5953E+10, 5.6503E+10,
     +  5.7059E+10, 5.7620E+10, 5.8186E+10, 5.8757E+10, 5.9334E+10,
     + 5.9917E+10, 6.0504E+10, 6.1098E+10, 6.1696E+10, 6.2301E+10,
     +  6.2911E+10, 6.3527E+10, 6.4148E+10, 6.4775E+10, 6.5409E+10,
     + 6.6048E+10, 6.6693E+10, 6.7344E+10, 6.8001E+10, 6.8664E+10,
     +  6.9334E+10, 7.0009E+10, 7.0691E+10, 7.1379E+10, 7.2074E+10,
     + 7.2775E+10, 7.3483E+10, 7.4197E+10, 7.4918E+10, 7.5645E+10,
     +  7.6380E+10, 7.7121E+10, 7.7869E+10, 7.8624E+10, 7.9385E+10,
     + 8.0154E+10, 8.0930E+10, 8.1713E+10, 8.2504E+10, 8.3301E+10,
     +  8.4107E+10, 8.4919E+10, 8.5739E+10, 8.6566E+10, 8.7402E+10,
     + 8.8244E+10, 8.9095E+10, 8.9953E+10, 9.0820E+10, 9.1694E+10,
     +  9.2576E+10, 9.3467E+10, 9.4365E+10, 9.5272E+10, 9.6187E+10,
     + 9.7111E+10, 9.8043E+10, 9.8983E+10, 9.9932E+10, 1.0089E+11,
     +  1.0186E+11, 1.0283E+11, 1.0382E+11, 1.0481E+11, 1.0581E+11,
     + 1.0682E+11, 1.0784E+11, 1.0887E+11, 1.0991E+11, 1.1096E+11,
     +  1.1202E+11, 1.1309E+11, 1.1417E+11, 1.1525E+11, 1.1635E+11,
     + 1.1746E+11, 1.1858E+11, 1.1971E+11, 1.2084E+11, 1.2199E+11,
     +  1.2315E+11, 1.2432E+11, 1.2550E+11, 1.2669E+11, 1.2789E+11,
     + 1.2910E+11, 1.3033E+11, 1.3156E+11, 1.3280E+11, 1.3406E+11,
     +  1.3533E+11, 1.3661E+11, 1.3790E+11, 1.3920E+11, 1.4052E+11,
     + 1.4184E+11, 1.4318E+11, 1.4453E+11, 1.4589E+11, 1.4727E+11,
     +  1.4865E+11, 1.5005E+11, 1.5146E+11, 1.5289E+11, 1.5433E+11,
     + 1.5578E+11, 1.5724E+11, 1.5871E+11, 1.6020E+11, 1.6171E+11,
     +  1.6322E+11, 1.6475E+11, 1.6630E+11, 1.6785E+11, 1.6942E+11,
     + 1.7101E+11, 1.7261E+11, 1.7422E+11, 1.7585E+11, 1.7749E+11,
     +  1.7915E+11, 1.8082E+11, 1.8251E+11, 1.8421E+11, 1.8593E+11,
     + 1.8766E+11, 1.8941E+11, 1.9117E+11, 1.9295E+11, 1.9474E+11,
     +  1.9655E+11, 1.9838E+11, 2.0022E+11, 2.0208E+11, 2.0395E+11,
     + 2.0585E+11, 2.0776E+11, 2.0968E+11, 2.1162E+11, 2.1358E+11,
     +  2.1556E+11, 2.1755E+11, 2.1957E+11, 2.2160E+11, 2.2364E+11,
     + 2.2571E+11, 2.2779E+11, 2.2990E+11, 2.3202E+11, 2.3416E+11,
     +  2.3631E+11, 2.3849E+11, 2.4069E+11, 2.4290E+11, 2.4514E+11,
     + 2.4739E+11, 2.4967E+11, 2.5196E+11, 2.5428E+11, 2.5661E+11,
     +  2.5897E+11, 2.6134E+11, 2.6374E+11, 2.6616E+11, 2.6860E+11,
     + 2.7106E+11, 2.7354E+11, 2.7604E+11, 2.7857E+11, 2.8111E+11,
     +  2.8368E+11, 2.8627E+11, 2.8889E+11, 2.9153E+11, 2.9419E+11,
     + 2.9687E+11, 2.9957E+11, 3.0230E+11, 3.0506E+11, 3.0784E+11,
     +  3.1064E+11, 3.1346E+11, 3.1631E+11, 3.1919E+11, 3.2209E+11,
     + 3.2501E+11, 3.2796E+11, 3.3094E+11, 3.3394E+11, 3.3697E+11,
     +  3.4002E+11, 3.4310E+11, 3.4621E+11, 3.4934E+11, 3.5251E+11,
     + 3.5569E+11, 3.5891E+11, 3.6215E+11, 3.6542E+11, 3.6872E+11,
     +  3.7205E+11, 3.7541E+11, 3.7879E+11, 3.8221E+11, 3.8565E+11,
     + 3.8912E+11, 3.9263E+11, 3.9616E+11, 3.9972E+11, 4.0332E+11,
     +  4.0694E+11, 4.1060E+11, 4.1428E+11, 4.1800E+11, 4.2175E+11,
     + 4.2554E+11, 4.2935E+11, 4.3320E+11, 4.3708E+11, 4.4099E+11,
     +  4.4494E+11, 4.4892E+11, 4.5293E+11, 4.5698E+11, 4.6107E+11,
     + 4.6518E+11, 4.6934E+11, 4.7352E+11, 4.7775E+11, 4.8201E+11,
     +  4.8630E+11, 4.9064E+11, 4.9501E+11, 4.9941E+11, 5.0386E+11,
     + 5.0834E+11, 5.1286E+11, 5.1742E+11, 5.2201E+11, 5.2665E+11,
     +  5.3132E+11, 5.3604E+11, 5.4079E+11, 5.4558E+11, 5.5042E+11,
     + 5.5529E+11, 5.6021E+11, 5.6517E+11, 5.7017E+11, 5.7521E+11,
     +  5.8029E+11, 5.8542E+11, 5.9059E+11, 5.9581E+11, 6.0106E+11,
     + 6.0637E+11, 6.1171E+11, 6.1710E+11, 6.2254E+11, 6.2802E+11,
     +  6.3355E+11, 6.3913E+11, 6.4475E+11, 6.5042E+11, 6.5613E+11,
     + 6.6190E+11, 6.6771E+11, 6.7357E+11, 6.7948E+11, 6.8544E+11,
     +  6.9145E+11, 6.9751E+11, 7.0362E+11, 7.0978E+11, 7.1599E+11,
     + 7.2226E+11, 7.2857E+11, 7.3494E+11, 7.4136E+11, 7.4784E+11,
     +  7.5437E+11, 7.6095E+11, 7.6759E+11, 7.7428E+11, 7.8103E+11,
     + 7.8784E+11, 7.9470E+11, 8.0162E+11, 8.0859E+11, 8.1563E+11,
     +  8.2272E+11, 8.2987E+11, 8.3708E+11, 8.4435E+11, 8.5168E+11,
     + 8.5907E+11, 8.6652E+11, 8.7404E+11, 8.8161E+11, 8.8925E+11,
     +  8.9695E+11, 9.0472E+11, 9.1255E+11, 9.2044E+11, 9.2840E+11,
     + 9.3642E+11, 9.4451E+11, 9.5267E+11, 9.6089E+11, 9.6918E+11,
     +  9.7754E+11, 9.8597E+11, 9.9446E+11, 1.0030E+12, 1.0117E+12,
     + 1.0204E+12, 1.0292E+12, 1.0380E+12, 1.0469E+12, 1.0559E+12,
     +  1.0650E+12, 1.0741E+12, 1.0834E+12, 1.0927E+12, 1.1020E+12,
     + 1.1115E+12, 1.1210E+12, 1.1306E+12, 1.1403E+12, 1.1500E+12,
     +  1.1599E+12, 1.1698E+12, 1.1798E+12, 1.1899E+12, 1.2000E+12,
     + 1.2103E+12, 1.2206E+12, 1.2310E+12, 1.2415E+12, 1.2521E+12,
     +  1.2628E+12, 1.2735E+12, 1.2844E+12, 1.2953E+12, 1.3063E+12,
     + 1.3174E+12, 1.3286E+12, 1.3399E+12, 1.3513E+12, 1.3628E+12,
     +  1.3744E+12, 1.3860E+12, 1.3978E+12, 1.4096E+12, 1.4216E+12,
     + 1.4336E+12, 1.4457E+12, 1.4580E+12, 1.4703E+12, 1.4827E+12,
     +  1.4953E+12, 1.5079E+12, 1.5206E+12, 1.5335E+12, 1.5464E+12,
     + 1.5595E+12, 1.5726E+12, 1.5859E+12, 1.5992E+12, 1.6127E+12,
     +  1.6263E+12, 1.6400E+12, 1.6538E+12, 1.6677E+12, 1.6817E+12,
     + 1.6958E+12, 1.7101E+12, 1.7244E+12, 1.7389E+12, 1.7535E+12,
     +  1.7682E+12, 1.7830E+12, 1.7980E+12, 1.8130E+12, 1.8282E+12,
     + 1.8435E+12, 1.8589E+12, 1.8745E+12, 1.8901E+12, 1.9059E+12,
     +  1.9219E+12, 1.9379E+12, 1.9541E+12, 1.9704E+12, 1.9868E+12,
     + 2.0034E+12, 2.0201E+12, 2.0369E+12, 2.0539E+12, 2.0709E+12,
     +  2.0882E+12, 2.1055E+12, 2.1230E+12, 2.1407E+12, 2.1585E+12,
     + 2.1764E+12, 2.1945E+12, 2.2127E+12, 2.2310E+12, 2.2495E+12,
     +  2.2682E+12, 2.2869E+12, 2.3059E+12, 2.3250E+12, 2.3442E+12,
     + 2.3636E+12, 2.3831E+12, 2.4028E+12, 2.4227E+12, 2.4427E+12,
     +  2.4629E+12, 2.4832E+12, 2.5037E+12, 2.5243E+12, 2.5451E+12,
     + 2.5661E+12, 2.5872E+12, 2.6085E+12, 2.6300E+12, 2.6516E+12,
     +  2.6734E+12, 2.6954E+12, 2.7175E+12, 2.7398E+12, 2.7623E+12,
     + 2.7850E+12, 2.8078E+12, 2.8309E+12, 2.8541E+12, 2.8775E+12,
     +  2.9010E+12, 2.9248E+12, 2.9487E+12, 2.9728E+12, 2.9971E+12,
     + 3.0216E+12, 3.0463E+12, 3.0712E+12, 3.0963E+12, 3.1216E+12,
     +  3.1470E+12, 3.1727E+12, 3.1986E+12, 3.2246E+12, 3.2509E+12,
     + 3.2774E+12, 3.3040E+12, 3.3309E+12, 3.3580E+12, 3.3853E+12,
     +  3.4128E+12, 3.4405E+12, 3.4685E+12, 3.4966E+12, 3.5250E+12,
     + 3.5536E+12, 3.5824E+12, 3.6114E+12, 3.6407E+12, 3.6702E+12,
     +  3.6999E+12, 3.7298E+12, 3.7600E+12, 3.7904E+12, 3.8211E+12,
     + 3.8519E+12, 3.8830E+12, 3.9144E+12, 3.9460E+12, 3.9778E+12,
     +  4.0099E+12, 4.0422E+12, 4.0748E+12, 4.1076E+12, 4.1407E+12,
     + 4.1740E+12, 4.2076E+12, 4.2414E+12, 4.2755E+12, 4.3099E+12,
     +  4.3445E+12, 4.3794E+12, 4.4145E+12, 4.4500E+12, 4.4857E+12,
     + 4.5216E+12, 4.5579E+12, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --      1231
      data (QofT( 2,J),J=1,2408)/ 7.6145E+00, 1.8129E+01, 3.1171E+01,
     +  4.6230E+01, 6.3049E+01, 8.1471E+01, 1.0138E+02, 1.2270E+02,
     + 1.4536E+02, 1.6929E+02, 1.9444E+02, 2.2076E+02, 2.4820E+02,
     +  2.7673E+02, 3.0630E+02, 3.3688E+02, 3.6843E+02, 4.0094E+02,
     + 4.3436E+02, 4.6867E+02, 5.3989E+02, 6.1442E+02, 6.9212E+02,
     +  7.7285E+02, 8.5652E+02, 9.4301E+02, 1.0322E+03, 1.1241E+03,
     + 1.2186E+03, 1.3155E+03, 1.4150E+03, 1.5168E+03, 1.6210E+03,
     +  1.7274E+03, 1.8362E+03, 1.9472E+03, 2.0603E+03, 2.1757E+03,
     + 2.2932E+03, 2.4128E+03, 2.5345E+03, 2.6584E+03, 2.7843E+03,
     +  2.9123E+03, 3.0424E+03, 3.1746E+03, 3.3088E+03, 3.4451E+03,
     + 3.5836E+03, 3.7240E+03, 3.8666E+03, 4.0113E+03, 4.1582E+03,
     +  4.3071E+03, 4.4581E+03, 4.6114E+03, 4.7668E+03, 4.9243E+03,
     + 5.0841E+03, 5.2461E+03, 5.4102E+03, 5.5767E+03, 5.7454E+03,
     +  5.9163E+03, 6.0896E+03, 6.2651E+03, 6.4430E+03, 6.6233E+03,
     + 6.8059E+03, 6.9909E+03, 7.1783E+03, 7.3681E+03, 7.5603E+03,
     +  7.7551E+03, 7.9523E+03, 8.1521E+03, 8.3543E+03, 8.5591E+03,
     + 8.7665E+03, 8.9765E+03, 9.1891E+03, 9.4044E+03, 9.6222E+03,
     +  9.8428E+03, 1.0066E+04, 1.0292E+04, 1.0521E+04, 1.0753E+04,
     + 1.0987E+04, 1.1224E+04, 1.1464E+04, 1.1707E+04, 1.1953E+04,
     +  1.2202E+04, 1.2454E+04, 1.2709E+04, 1.2967E+04, 1.3228E+04,
     + 1.3492E+04, 1.3759E+04, 1.4030E+04, 1.4304E+04, 1.4580E+04,
     +  1.4861E+04, 1.5144E+04, 1.5431E+04, 1.5721E+04, 1.6015E+04,
     + 1.6312E+04, 1.6613E+04, 1.6917E+04, 1.7225E+04, 1.7536E+04,
     +  1.7852E+04, 1.8170E+04, 1.8493E+04, 1.8819E+04, 1.9150E+04,
     + 1.9484E+04, 1.9822E+04, 2.0164E+04, 2.0511E+04, 2.0861E+04,
     +  2.1215E+04, 2.1574E+04, 2.1937E+04, 2.2304E+04, 2.2676E+04,
     + 2.3052E+04, 2.3433E+04, 2.3818E+04, 2.4208E+04, 2.4602E+04,
     +  2.5002E+04, 2.5406E+04, 2.5815E+04, 2.6229E+04, 2.6647E+04,
     + 2.7071E+04, 2.7501E+04, 2.7935E+04, 2.8374E+04, 2.8819E+04,
     +  2.9270E+04, 2.9726E+04, 3.0187E+04, 3.0655E+04, 3.1128E+04,
     + 3.1606E+04, 3.2091E+04, 3.2582E+04, 3.3079E+04, 3.3581E+04,
     +  3.4091E+04, 3.4606E+04, 3.5128E+04, 3.5657E+04, 3.6192E+04,
     + 3.6734E+04, 3.7282E+04, 3.7838E+04, 3.8400E+04, 3.8970E+04,
     +  3.9547E+04, 4.0131E+04, 4.0722E+04, 4.1321E+04, 4.1928E+04,
     + 4.2543E+04, 4.3165E+04, 4.3795E+04, 4.4434E+04, 4.5080E+04,
     +  4.5735E+04, 4.6398E+04, 4.7070E+04, 4.7751E+04, 4.8440E+04,
     + 4.9138E+04, 4.9845E+04, 5.0562E+04, 5.1288E+04, 5.2023E+04,
     +  5.2768E+04, 5.3522E+04, 5.4287E+04, 5.5062E+04, 5.5846E+04,
     + 5.6641E+04, 5.7446E+04, 5.8262E+04, 5.9089E+04, 5.9927E+04,
     +  6.0775E+04, 6.1635E+04, 6.2507E+04, 6.3390E+04, 6.4284E+04,
     + 6.5190E+04, 6.6109E+04, 6.7040E+04, 6.7983E+04, 6.8939E+04,
     +  6.9907E+04, 7.0889E+04, 7.1883E+04, 7.2891E+04, 7.3913E+04,
     + 7.4948E+04, 7.5997E+04, 7.7060E+04, 7.8138E+04, 7.9229E+04,
     +  8.0336E+04, 8.1458E+04, 8.2594E+04, 8.3746E+04, 8.4914E+04,
     + 8.6098E+04, 8.7297E+04, 8.8513E+04, 8.9745E+04, 9.0994E+04,
     +  9.2260E+04, 9.3543E+04, 9.4844E+04, 9.6162E+04, 9.7498E+04,
     + 9.8852E+04, 1.0023E+05, 1.0162E+05, 1.0303E+05, 1.0446E+05,
     +  1.0591E+05, 1.0738E+05, 1.0887E+05, 1.1037E+05, 1.1190E+05,
     + 1.1346E+05, 1.1503E+05, 1.1662E+05, 1.1824E+05, 1.1988E+05,
     +  1.2154E+05, 1.2322E+05, 1.2493E+05, 1.2666E+05, 1.2841E+05,
     + 1.3019E+05, 1.3199E+05, 1.3382E+05, 1.3567E+05, 1.3755E+05,
     +  1.3945E+05, 1.4138E+05, 1.4334E+05, 1.4532E+05, 1.4733E+05,
     + 1.4937E+05, 1.5144E+05, 1.5353E+05, 1.5566E+05, 1.5781E+05,
     +  1.5999E+05, 1.6221E+05, 1.6445E+05, 1.6672E+05, 1.6903E+05,
     + 1.7137E+05, 1.7374E+05, 1.7614E+05, 1.7858E+05, 1.8105E+05,
     +  1.8355E+05, 1.8609E+05, 1.8867E+05, 1.9128E+05, 1.9392E+05,
     + 1.9661E+05, 1.9933E+05, 2.0208E+05, 2.0488E+05, 2.0771E+05,
     +  2.1059E+05, 2.1350E+05, 2.1646E+05, 2.1945E+05, 2.2249E+05,
     + 2.2557E+05, 2.2869E+05, 2.3185E+05, 2.3506E+05, 2.3831E+05,
     +  2.4161E+05, 2.4496E+05, 2.4835E+05, 2.5178E+05, 2.5527E+05,
     + 2.5880E+05, 2.6238E+05, 2.6602E+05, 2.6970E+05, 2.7343E+05,
     +  2.7722E+05, 2.8105E+05, 2.8495E+05, 2.8889E+05, 2.9289E+05,
     + 2.9695E+05, 3.0106E+05, 3.0523E+05, 3.0945E+05, 3.1374E+05,
     +  3.1808E+05, 3.2249E+05, 3.2695E+05, 3.3148E+05, 3.3607E+05,
     + 3.4072E+05, 3.4544E+05, 3.5023E+05, 3.5508E+05, 3.5999E+05,
     +  3.6498E+05, 3.7004E+05, 3.7516E+05, 3.8036E+05, 3.8562E+05,
     + 3.9097E+05, 3.9638E+05, 4.0187E+05, 4.0744E+05, 4.1308E+05,
     +  4.1880E+05, 4.2460E+05, 4.3048E+05, 4.3645E+05, 4.4249E+05,
     + 4.4862E+05, 4.5483E+05, 4.6113E+05, 4.6752E+05, 4.7399E+05,
     +  4.8056E+05, 4.8721E+05, 4.9396E+05, 5.0080E+05, 5.0774E+05,
     + 5.1477E+05, 5.2190E+05, 5.2913E+05, 5.3645E+05, 5.4388E+05,
     +  5.5141E+05, 5.5905E+05, 5.6679E+05, 5.7463E+05, 5.8259E+05,
     + 5.9066E+05, 5.9883E+05, 6.0712E+05, 6.1553E+05, 6.2405E+05,
     +  6.3269E+05, 6.4144E+05, 6.5032E+05, 6.5932E+05, 6.6844E+05,
     + 6.7769E+05, 6.8707E+05, 6.9658E+05, 7.0622E+05, 7.1599E+05,
     +  7.2589E+05, 7.3593E+05, 7.4611E+05, 7.5643E+05, 7.6689E+05,
     + 7.7750E+05, 7.8825E+05, 7.9915E+05, 8.1020E+05, 8.2140E+05,
     +  8.3275E+05, 8.4426E+05, 8.5593E+05, 8.6776E+05, 8.7975E+05,
     + 8.9191E+05, 9.0423E+05, 9.1672E+05, 9.2939E+05, 9.4223E+05,
     +  9.5524E+05, 9.6843E+05, 9.8180E+05, 9.9536E+05, 1.0091E+06,
     + 1.0230E+06, 1.0372E+06, 1.0515E+06, 1.0660E+06, 1.0807E+06,
     +  1.0956E+06, 1.1107E+06, 1.1260E+06, 1.1416E+06, 1.1573E+06,
     + 1.1733E+06, 1.1894E+06, 1.2058E+06, 1.2224E+06, 1.2393E+06,
     +  1.2564E+06, 1.2737E+06, 1.2912E+06, 1.3090E+06, 1.3270E+06,
     + 1.3453E+06, 1.3638E+06, 1.3826E+06, 1.4016E+06, 1.4209E+06,
     +  1.4405E+06, 1.4603E+06, 1.4804E+06, 1.5008E+06, 1.5214E+06,
     + 1.5423E+06, 1.5635E+06, 1.5850E+06, 1.6068E+06, 1.6289E+06,
     +  1.6512E+06, 1.6739E+06, 1.6969E+06, 1.7202E+06, 1.7438E+06,
     + 1.7678E+06, 1.7920E+06, 1.8166E+06, 1.8415E+06, 1.8668E+06,
     +  1.8924E+06, 1.9184E+06, 1.9447E+06, 1.9713E+06, 1.9983E+06,
     + 2.0257E+06, 2.0535E+06, 2.0816E+06, 2.1101E+06, 2.1390E+06,
     +  2.1683E+06, 2.1979E+06, 2.2280E+06, 2.2585E+06, 2.2894E+06,
     + 2.3207E+06, 2.3524E+06, 2.3846E+06, 2.4172E+06, 2.4502E+06,
     +  2.4837E+06, 2.5176E+06, 2.5520E+06, 2.5868E+06, 2.6221E+06,
     + 2.6579E+06, 2.6942E+06, 2.7309E+06, 2.7682E+06, 2.8059E+06,
     +  2.8442E+06, 2.8829E+06, 2.9222E+06, 2.9620E+06, 3.0024E+06,
     + 3.0433E+06, 3.0847E+06, 3.1267E+06, 3.1692E+06, 3.2123E+06,
     +  3.2560E+06, 3.3003E+06, 3.3452E+06, 3.3906E+06, 3.4367E+06,
     + 3.4834E+06, 3.5307E+06, 3.5786E+06, 3.6272E+06, 3.6764E+06,
     +  3.7263E+06, 3.7769E+06, 3.8281E+06, 3.8800E+06, 3.9326E+06,
     + 3.9859E+06, 4.0399E+06, 4.0946E+06, 4.1500E+06, 4.2062E+06,
     +  4.2631E+06, 4.3208E+06, 4.3793E+06, 4.4385E+06, 4.4985E+06,
     + 4.5593E+06, 4.6209E+06, 4.6833E+06, 4.7466E+06, 4.8106E+06,
     +  4.8756E+06, 4.9414E+06, 5.0080E+06, 5.0756E+06, 5.1440E+06,
     + 5.2133E+06, 5.2836E+06, 5.3548E+06, 5.4269E+06, 5.4999E+06,
     +  5.5740E+06, 5.6490E+06, 5.7250E+06, 5.8020E+06, 5.8800E+06,
     + 5.9590E+06, 6.0391E+06, 6.1202E+06, 6.2024E+06, 6.2857E+06,
     +  6.3700E+06, 6.4555E+06, 6.5421E+06, 6.6298E+06, 6.7187E+06,
     + 6.8087E+06, 6.8999E+06, 6.9924E+06, 7.0860E+06, 7.1808E+06,
     +  7.2769E+06, 7.3743E+06, 7.4729E+06, 7.5728E+06, 7.6740E+06,
     + 7.7765E+06, 7.8804E+06, 7.9856E+06, 8.0922E+06, 8.2002E+06,
     +  8.3096E+06, 8.4204E+06, 8.5327E+06, 8.6464E+06, 8.7616E+06,
     + 8.8783E+06, 8.9965E+06, 9.1163E+06, 9.2376E+06, 9.3605E+06,
     +  9.4850E+06, 9.6110E+06, 9.7388E+06, 9.8682E+06, 9.9992E+06,
     + 1.0132E+07, 1.0266E+07, 1.0403E+07, 1.0541E+07, 1.0680E+07,
     +  1.0822E+07, 1.0965E+07, 1.1111E+07, 1.1258E+07, 1.1407E+07,
     + 1.1558E+07, 1.1711E+07, 1.1865E+07, 1.2022E+07, 1.2181E+07,
     +  1.2342E+07, 1.2505E+07, 1.2670E+07, 1.2837E+07, 1.3007E+07,
     + 1.3178E+07, 1.3352E+07, 1.3528E+07, 1.3706E+07, 1.3887E+07,
     +  1.4070E+07, 1.4255E+07, 1.4442E+07, 1.4632E+07, 1.4825E+07,
     + 1.5019E+07, 1.5217E+07, 1.5417E+07, 1.5619E+07, 1.5824E+07,
     +  1.6032E+07, 1.6242E+07, 1.6455E+07, 1.6670E+07, 1.6889E+07,
     + 1.7110E+07, 1.7334E+07, 1.7561E+07, 1.7791E+07, 1.8023E+07,
     +  1.8259E+07, 1.8497E+07, 1.8739E+07, 1.8984E+07, 1.9232E+07,
     + 1.9483E+07, 1.9737E+07, 1.9994E+07, 2.0255E+07, 2.0519E+07,
     +  2.0786E+07, 2.1057E+07, 2.1331E+07, 2.1608E+07, 2.1889E+07,
     + 2.2174E+07, 2.2462E+07, 2.2754E+07, 2.3049E+07, 2.3349E+07,
     +  2.3652E+07, 2.3958E+07, 2.4269E+07, 2.4584E+07, 2.4902E+07,
     + 2.5225E+07, 2.5551E+07, 2.5882E+07, 2.6217E+07, 2.6556E+07,
     +  2.6899E+07, 2.7247E+07, 2.7599E+07, 2.7955E+07, 2.8316E+07,
     + 2.8682E+07, 2.9052E+07, 2.9426E+07, 2.9805E+07, 3.0190E+07,
     +  3.0578E+07, 3.0972E+07, 3.1371E+07, 3.1774E+07, 3.2183E+07,
     + 3.2596E+07, 3.3015E+07, 3.3439E+07, 3.3869E+07, 3.4303E+07,
     +  3.4744E+07, 3.5189E+07, 3.5640E+07, 3.6097E+07, 3.6559E+07,
     + 3.7027E+07, 3.7501E+07, 3.7981E+07, 3.8467E+07, 3.8959E+07,
     +  3.9457E+07, 3.9961E+07, 4.0471E+07, 4.0987E+07, 4.1510E+07,
     + 4.2040E+07, 4.2576E+07, 4.3118E+07, 4.3668E+07, 4.4224E+07,
     +  4.4787E+07, 4.5356E+07, 4.5933E+07, 4.6517E+07, 4.7108E+07,
     + 4.7707E+07, 4.8313E+07, 4.8926E+07, 4.9547E+07, 5.0175E+07,
     +  5.0811E+07, 5.1455E+07, 5.2107E+07, 5.2767E+07, 5.3434E+07,
     + 5.4110E+07, 5.4795E+07, 5.5487E+07, 5.6189E+07, 5.6898E+07,
     +  5.7617E+07, 5.8344E+07, 5.9080E+07, 5.9825E+07, 6.0579E+07,
     + 6.1343E+07, 6.2115E+07, 6.2897E+07, 6.3689E+07, 6.4490E+07,
     +  6.5301E+07, 6.6122E+07, 6.6952E+07, 6.7793E+07, 6.8644E+07,
     + 6.9506E+07, 7.0377E+07, 7.1260E+07, 7.2153E+07, 7.3057E+07,
     +  7.3972E+07, 7.4898E+07, 7.5835E+07, 7.6783E+07, 7.7743E+07,
     + 7.8715E+07, 7.9698E+07, 8.0693E+07, 8.1700E+07, 8.2719E+07,
     +  8.3751E+07, 8.4795E+07, 8.5851E+07, 8.6921E+07, 8.8003E+07,
     + 8.9098E+07, 9.0206E+07, 9.1328E+07, 9.2463E+07, 9.3612E+07,
     +  9.4774E+07, 9.5950E+07, 9.7141E+07, 9.8346E+07, 9.9565E+07,
     + 1.0080E+08, 1.0205E+08, 1.0331E+08, 1.0459E+08, 1.0588E+08,
     +  1.0719E+08, 1.0852E+08, 1.0986E+08, 1.1121E+08, 1.1259E+08,
     + 1.1398E+08, 1.1538E+08, 1.1680E+08, 1.1824E+08, 1.1970E+08,
     +  1.2117E+08, 1.2267E+08, 1.2418E+08, 1.2570E+08, 1.2725E+08,
     + 1.2881E+08, 1.3039E+08, 1.3199E+08, 1.3361E+08, 1.3525E+08,
     +  1.3691E+08, 1.3859E+08, 1.4028E+08, 1.4200E+08, 1.4374E+08,
     + 1.4550E+08, 1.4728E+08, 1.4908E+08, 1.5090E+08, 1.5274E+08,
     +  1.5461E+08, 1.5649E+08, 1.5840E+08, 1.6033E+08, 1.6229E+08,
     + 1.6426E+08, 1.6626E+08, 1.6829E+08, 1.7033E+08, 1.7241E+08,
     +  1.7450E+08, 1.7662E+08, 1.7877E+08, 1.8094E+08, 1.8313E+08,
     + 1.8535E+08, 1.8760E+08, 1.8987E+08, 1.9217E+08, 1.9450E+08,
     +  1.9685E+08, 1.9923E+08, 2.0164E+08, 2.0408E+08, 2.0654E+08,
     + 2.0904E+08, 2.1156E+08, 2.1411E+08, 2.1669E+08, 2.1931E+08,
     +  2.2195E+08, 2.2462E+08, 2.2732E+08, 2.3006E+08, 2.3283E+08,
     + 2.3562E+08, 2.3846E+08, 2.4132E+08, 2.4422E+08, 2.4715E+08,
     +  2.5011E+08, 2.5311E+08, 2.5614E+08, 2.5921E+08, 2.6232E+08,
     + 2.6546E+08, 2.6863E+08, 2.7184E+08, 2.7509E+08, 2.7838E+08,
     +  2.8170E+08, 2.8507E+08, 2.8847E+08, 2.9191E+08, 2.9539E+08,
     + 2.9891E+08, 3.0247E+08, 3.0607E+08, 3.0971E+08, 3.1339E+08,
     +  3.1712E+08, 3.2089E+08, 3.2470E+08, 3.2855E+08, 3.3245E+08,
     + 3.3640E+08, 3.4039E+08, 3.4442E+08, 3.4850E+08, 3.5263E+08,
     +  3.5680E+08, 3.6102E+08, 3.6529E+08, 3.6961E+08, 3.7398E+08,
     + 3.7839E+08, 3.8286E+08, 3.8738E+08, 3.9194E+08, 3.9656E+08,
     +  4.0124E+08, 4.0596E+08, 4.1074E+08, 4.1558E+08, 4.2046E+08,
     + 4.2541E+08, 4.3041E+08, 4.3546E+08, 4.4057E+08, 4.4574E+08,
     +  4.5097E+08, 4.5626E+08, 4.6161E+08, 4.6701E+08, 4.7248E+08,
     + 4.7801E+08, 4.8361E+08, 4.8926E+08, 4.9498E+08, 5.0076E+08,
     +  5.0661E+08, 5.1252E+08, 5.1850E+08, 5.2455E+08, 5.3066E+08,
     + 5.3685E+08, 5.4310E+08, 5.4942E+08, 5.5581E+08, 5.6228E+08,
     +  5.6881E+08, 5.7542E+08, 5.8211E+08, 5.8886E+08, 5.9570E+08,
     + 6.0261E+08, 6.0959E+08, 6.1666E+08, 6.2380E+08, 6.3102E+08,
     +  6.3832E+08, 6.4571E+08, 6.5317E+08, 6.6072E+08, 6.6836E+08,
     + 6.7607E+08, 6.8388E+08, 6.9176E+08, 6.9974E+08, 7.0781E+08,
     +  7.1596E+08, 7.2420E+08, 7.3254E+08, 7.4097E+08, 7.4949E+08,
     + 7.5810E+08, 7.6681E+08, 7.7562E+08, 7.8452E+08, 7.9352E+08,
     +  8.0262E+08, 8.1182E+08, 8.2113E+08, 8.3053E+08, 8.4004E+08,
     + 8.4965E+08, 8.5937E+08, 8.6919E+08, 8.7912E+08, 8.8916E+08,
     +  8.9932E+08, 9.0958E+08, 9.1995E+08, 9.3044E+08, 9.4104E+08,
     + 9.5176E+08, 9.6260E+08, 9.7355E+08, 9.8463E+08, 9.9582E+08,
     +  1.0071E+09, 1.0186E+09, 1.0301E+09, 1.0418E+09, 1.0537E+09,
     + 1.0656E+09, 1.0777E+09, 1.0899E+09, 1.1022E+09, 1.1147E+09,
     +  1.1273E+09, 1.1401E+09, 1.1529E+09, 1.1660E+09, 1.1791E+09,
     + 1.1924E+09, 1.2059E+09, 1.2195E+09, 1.2332E+09, 1.2471E+09,
     +  1.2612E+09, 1.2754E+09, 1.2897E+09, 1.3042E+09, 1.3189E+09,
     + 1.3337E+09, 1.3487E+09, 1.3638E+09, 1.3791E+09, 1.3946E+09,
     +  1.4102E+09, 1.4260E+09, 1.4420E+09, 1.4582E+09, 1.4745E+09,
     + 1.4910E+09, 1.5076E+09, 1.5245E+09, 1.5415E+09, 1.5587E+09,
     +  1.5761E+09, 1.5937E+09, 1.6115E+09, 1.6294E+09, 1.6476E+09,
     + 1.6659E+09, 1.6845E+09, 1.7032E+09, 1.7221E+09, 1.7413E+09,
     +  1.7606E+09, 1.7802E+09, 1.7999E+09, 1.8199E+09, 1.8401E+09,
     + 1.8605E+09, 1.8811E+09, 1.9019E+09, 1.9230E+09, 1.9442E+09,
     +  1.9657E+09, 1.9875E+09, 2.0094E+09, 2.0316E+09, 2.0540E+09,
     + 2.0767E+09, 2.0996E+09, 2.1227E+09, 2.1461E+09, 2.1698E+09,
     +  2.1936E+09, 2.2178E+09, 2.2422E+09, 2.2668E+09, 2.2917E+09,
     + 2.3169E+09, 2.3423E+09, 2.3680E+09, 2.3940E+09, 2.4202E+09,
     +  2.4468E+09, 2.4736E+09, 2.5006E+09, 2.5280E+09, 2.5557E+09,
     + 2.5836E+09, 2.6118E+09, 2.6404E+09, 2.6692E+09, 2.6983E+09,
     +  2.7278E+09, 2.7575E+09, 2.7875E+09, 2.8179E+09, 2.8486E+09,
     + 2.8796E+09, 2.9109E+09, 2.9426E+09, 2.9745E+09, 3.0069E+09,
     +  3.0395E+09, 3.0725E+09, 3.1058E+09, 3.1395E+09, 3.1735E+09,
     + 3.2079E+09, 3.2427E+09, 3.2778E+09, 3.3132E+09, 3.3491E+09,
     +  3.3853E+09, 3.4218E+09, 3.4588E+09, 3.4961E+09, 3.5338E+09,
     + 3.5719E+09, 3.6105E+09, 3.6494E+09, 3.6887E+09, 3.7284E+09,
     +  3.7685E+09, 3.8090E+09, 3.8500E+09, 3.8913E+09, 3.9331E+09,
     + 3.9753E+09, 4.0180E+09, 4.0611E+09, 4.1046E+09, 4.1486E+09,
     +  4.1931E+09, 4.2380E+09, 4.2833E+09, 4.3291E+09, 4.3754E+09,
     + 4.4222E+09, 4.4694E+09, 4.5171E+09, 4.5653E+09, 4.6140E+09,
     +  4.6632E+09, 4.7129E+09, 4.7631E+09, 4.8139E+09, 4.8651E+09,
     + 4.9169E+09, 4.9691E+09, 5.0220E+09, 5.0753E+09, 5.1292E+09,
     +  5.1837E+09, 5.2387E+09, 5.2942E+09, 5.3503E+09, 5.4070E+09,
     + 5.4643E+09, 5.5221E+09, 5.5806E+09, 5.6396E+09, 5.6992E+09,
     +  5.7594E+09, 5.8202E+09, 5.8817E+09, 5.9437E+09, 6.0064E+09,
     + 6.0698E+09, 6.1337E+09, 6.1983E+09, 6.2636E+09, 6.3295E+09,
     +  6.3961E+09, 6.4633E+09, 6.5313E+09, 6.5999E+09, 6.6692E+09,
     + 6.7392E+09, 6.8098E+09, 6.8813E+09, 6.9534E+09, 7.0262E+09,
     +  7.0998E+09, 7.1741E+09, 7.2492E+09, 7.3250E+09, 7.4015E+09,
     + 7.4789E+09, 7.5570E+09, 7.6358E+09, 7.7155E+09, 7.7960E+09,
     +  7.8772E+09, 7.9593E+09, 8.0422E+09, 8.1259E+09, 8.2105E+09,
     + 8.2959E+09, 8.3821E+09, 8.4692E+09, 8.5572E+09, 8.6460E+09,
     +  8.7357E+09, 8.8264E+09, 8.9179E+09, 9.0103E+09, 9.1036E+09,
     + 9.1979E+09, 9.2931E+09, 9.3892E+09, 9.4863E+09, 9.5843E+09,
     +  9.6833E+09, 9.7833E+09, 9.8843E+09, 9.9863E+09, 1.0089E+10,
     + 1.0193E+10, 1.0298E+10, 1.0404E+10, 1.0511E+10, 1.0620E+10,
     +  1.0729E+10, 1.0839E+10, 1.0950E+10, 1.1063E+10, 1.1176E+10,
     + 1.1291E+10, 1.1407E+10, 1.1524E+10, 1.1642E+10, 1.1761E+10,
     +  1.1882E+10, 1.2003E+10, 1.2126E+10, 1.2250E+10, 1.2375E+10,
     + 1.2502E+10, 1.2629E+10, 1.2758E+10, 1.2888E+10, 1.3020E+10,
     +  1.3152E+10, 1.3286E+10, 1.3422E+10, 1.3558E+10, 1.3696E+10,
     + 1.3836E+10, 1.3976E+10, 1.4118E+10, 1.4262E+10, 1.4406E+10,
     +  1.4553E+10, 1.4700E+10, 1.4849E+10, 1.5000E+10, 1.5152E+10,
     + 1.5305E+10, 1.5460E+10, 1.5616E+10, 1.5774E+10, 1.5934E+10,
     +  1.6095E+10, 1.6257E+10, 1.6421E+10, 1.6587E+10, 1.6754E+10,
     + 1.6923E+10, 1.7094E+10, 1.7266E+10, 1.7439E+10, 1.7615E+10,
     +  1.7792E+10, 1.7971E+10, 1.8152E+10, 1.8334E+10, 1.8518E+10,
     + 1.8704E+10, 1.8891E+10, 1.9081E+10, 1.9272E+10, 1.9465E+10,
     +  1.9660E+10, 1.9857E+10, 2.0055E+10, 2.0256E+10, 2.0458E+10,
     + 2.0663E+10, 2.0869E+10, 2.1077E+10, 2.1287E+10, 2.1500E+10,
     +  2.1714E+10, 2.1930E+10, 2.2149E+10, 2.2369E+10, 2.2592E+10,
     + 2.2817E+10, 2.3043E+10, 2.3272E+10, 2.3504E+10, 2.3737E+10,
     +  2.3973E+10, 2.4210E+10, 2.4450E+10, 2.4693E+10, 2.4937E+10,
     + 2.5184E+10, 2.5433E+10, 2.5685E+10, 2.5939E+10, 2.6196E+10,
     +  2.6454E+10, 2.6716E+10, 2.6979E+10, 2.7246E+10, 2.7514E+10,
     + 2.7785E+10, 2.8059E+10, 2.8336E+10, 2.8615E+10, 2.8896E+10,
     +  2.9180E+10, 2.9467E+10, 2.9757E+10, 3.0049E+10, 3.0344E+10,
     + 3.0642E+10, 3.0943E+10, 3.1246E+10, 3.1553E+10, 3.1862E+10,
     +  3.2174E+10, 3.2489E+10, 3.2807E+10, 3.3128E+10, 3.3451E+10,
     + 3.3778E+10, 3.4108E+10, 3.4441E+10, 3.4777E+10, 3.5117E+10,
     +  3.5459E+10, 3.5805E+10, 3.6153E+10, 3.6505E+10, 3.6861E+10,
     + 3.7219E+10, 3.7581E+10, 3.7947E+10, 3.8315E+10, 3.8688E+10,
     +  3.9063E+10, 3.9442E+10, 3.9825E+10, 4.0211E+10, 4.0601E+10,
     + 4.0994E+10, 4.1391E+10, 4.1791E+10, 4.2196E+10, 4.2604E+10,
     +  4.3016E+10, 4.3431E+10, 4.3851E+10, 4.4274E+10, 4.4701E+10,
     + 4.5132E+10, 4.5567E+10, 4.6006E+10, 4.6450E+10, 4.6897E+10,
     +  4.7348E+10, 4.7804E+10, 4.8263E+10, 4.8727E+10, 4.9195E+10,
     + 4.9668E+10, 5.0144E+10, 5.0625E+10, 5.1111E+10, 5.1601E+10,
     +  5.2095E+10, 5.2594E+10, 5.3098E+10, 5.3606E+10, 5.4119E+10,
     + 5.4636E+10, 5.5158E+10, 5.5685E+10, 5.6217E+10, 5.6753E+10,
     +  5.7295E+10, 5.7841E+10, 5.8393E+10, 5.8949E+10, 5.9510E+10,
     + 6.0077E+10, 6.0649E+10, 6.1226E+10, 6.1808E+10, 6.2395E+10,
     +  6.2988E+10, 6.3586E+10, 6.4189E+10, 6.4798E+10, 6.5413E+10,
     + 6.6033E+10, 6.6658E+10, 6.7290E+10, 6.7927E+10, 6.8569E+10,
     +  6.9218E+10, 6.9872E+10, 7.0533E+10, 7.1199E+10, 7.1871E+10,
     + 7.2550E+10, 7.3234E+10, 7.3925E+10, 7.4621E+10, 7.5324E+10,
     +  7.6034E+10, 7.6750E+10, 7.7472E+10, 7.8200E+10, 7.8936E+10,
     + 7.9677E+10, 8.0426E+10, 8.1181E+10, 8.1943E+10, 8.2712E+10,
     +  8.3487E+10, 8.4270E+10, 8.5059E+10, 8.5856E+10, 8.6660E+10,
     + 8.7471E+10, 8.8289E+10, 8.9114E+10, 8.9947E+10, 9.0787E+10,
     +  9.1635E+10, 9.2490E+10, 9.3353E+10, 9.4224E+10, 9.5102E+10,
     + 9.5988E+10, 9.6882E+10, 9.7784E+10, 9.8694E+10, 9.9612E+10,
     +  1.0054E+11, 1.0147E+11, 1.0242E+11, 1.0337E+11, 1.0433E+11,
     + 1.0529E+11, 1.0627E+11, 1.0725E+11, 1.0825E+11, 1.0925E+11,
     +  1.1026E+11, 1.1128E+11, 1.1231E+11, 1.1335E+11, 1.1440E+11,
     + 1.1546E+11, 1.1652E+11, 1.1760E+11, 1.1868E+11, 1.1978E+11,
     +  1.2088E+11, 1.2199E+11, 1.2312E+11, 1.2425E+11, 1.2540E+11,
     + 1.2655E+11, 1.2771E+11, 1.2889E+11, 1.3007E+11, 1.3127E+11,
     +  1.3247E+11, 1.3369E+11, 1.3491E+11, 1.3615E+11, 1.3740E+11,
     + 1.3866E+11, 1.3992E+11, 1.4121E+11, 1.4250E+11, 1.4380E+11,
     +  1.4511E+11, 1.4644E+11, 1.4778E+11, 1.4913E+11, 1.5049E+11,
     + 1.5186E+11, 1.5325E+11, 1.5464E+11, 1.5605E+11, 1.5747E+11,
     +  1.5891E+11, 1.6035E+11, 1.6181E+11, 1.6328E+11, 1.6476E+11,
     + 1.6626E+11, 1.6777E+11, 1.6929E+11, 1.7083E+11, 1.7238E+11,
     +  1.7394E+11, 1.7552E+11, 1.7711E+11, 1.7871E+11, 1.8033E+11,
     + 1.8196E+11, 1.8360E+11, 1.8526E+11, 1.8694E+11, 1.8862E+11,
     +  1.9033E+11, 1.9204E+11, 1.9378E+11, 1.9552E+11, 1.9729E+11,
     + 1.9906E+11, 2.0085E+11, 2.0266E+11, 2.0449E+11, 2.0632E+11,
     +  2.0818E+11, 2.1005E+11, 2.1194E+11, 2.1384E+11, 2.1576E+11,
     + 2.1769E+11, 2.1965E+11, 2.2161E+11, 2.2360E+11, 2.2560E+11,
     +  2.2762E+11, 2.2966E+11, 2.3171E+11, 2.3378E+11, 2.3587E+11,
     + 2.3798E+11, 2.4010E+11, 2.4225E+11, 2.4441E+11, 2.4659E+11,
     +  2.4879E+11, 2.5100E+11, 2.5324E+11, 2.5549E+11, 2.5777E+11,
     + 2.6006E+11, 2.6237E+11, 2.6471E+11, 2.6706E+11, 2.6943E+11,
     +  2.7182E+11, 2.7423E+11, 2.7667E+11, 2.7912E+11, 2.8159E+11,
     + 2.8409E+11, 2.8660E+11, 2.8914E+11, 2.9170E+11, 2.9428E+11,
     +  2.9688E+11, 2.9950E+11, 3.0215E+11, 3.0482E+11, 3.0751E+11,
     + 3.1022E+11, 3.1296E+11, 3.1571E+11, 3.1850E+11, 3.2130E+11,
     +  3.2413E+11, 3.2698E+11, 3.2986E+11, 3.3276E+11, 3.3569E+11,
     + 3.3863E+11, 3.4161E+11, 3.4461E+11, 3.4763E+11, 3.5068E+11,
     +  3.5376E+11, 3.5686E+11, 3.5998E+11, 3.6314E+11, 3.6632E+11,
     + 3.6952E+11, 3.7275E+11, 3.7601E+11, 3.7930E+11, 3.8261E+11,
     +  3.8595E+11, 3.8932E+11, 3.9272E+11, 3.9614E+11, 3.9960E+11,
     + 4.0308E+11, 4.0659E+11, 4.1013E+11, 4.1370E+11, 4.1730E+11,
     +  4.2093E+11, 4.2459E+11, 4.2828E+11, 4.3200E+11, 4.3575E+11,
     + 4.3953E+11, 4.4335E+11, 4.4719E+11, 4.5107E+11, 4.5497E+11,
     +  4.5892E+11, 4.6289E+11, 4.6690E+11, 4.7093E+11, 4.7501E+11,
     + 4.7911E+11, 4.8325E+11, 4.8743E+11, 4.9163E+11, 4.9588E+11,
     +  5.0015E+11, 5.0447E+11, 5.0882E+11, 5.1320E+11, 5.1762E+11,
     + 5.2208E+11, 5.2657E+11, 5.3110E+11, 5.3567E+11, 5.4027E+11,
     +  5.4491E+11, 5.4959E+11, 5.5431E+11, 5.5907E+11, 5.6386E+11,
     + 5.6870E+11, 5.7357E+11, 5.7848E+11, 5.8344E+11, 5.8843E+11,
     +  5.9347E+11, 5.9854E+11, 6.0366E+11, 6.0882E+11, 6.1402E+11,
     + 6.1927E+11, 6.2455E+11, 6.2988E+11, 6.3525E+11, 6.4067E+11,
     +  6.4613E+11, 6.5163E+11, 6.5718E+11, 6.6278E+11, 6.6842E+11,
     + 6.7410E+11, 6.7983E+11, 6.8561E+11, 6.9144E+11, 6.9731E+11,
     +  7.0323E+11, 7.0919E+11, 7.1521E+11, 7.2127E+11, 7.2739E+11,
     + 7.3355E+11, 7.3976E+11, 7.4602E+11, 7.5233E+11, 7.5870E+11,
     +  7.6511E+11, 7.7158E+11, 7.7810E+11, 7.8467E+11, 7.9129E+11,
     + 7.9797E+11, 8.0470E+11, 8.1148E+11, 8.1832E+11, 8.2521E+11,
     +  8.3216E+11, 8.3917E+11, 8.4623E+11, 8.5335E+11, 8.6052E+11,
     + 8.6775E+11, 8.7504E+11, 8.8239E+11, 8.8980E+11, 8.9726E+11,
     +  9.0479E+11, 9.1238E+11, 9.2002E+11, 9.2773E+11, 9.3550E+11,
     + 9.4333E+11, 9.5122E+11, 9.5918E+11, 9.6720E+11, 9.7528E+11,
     +  9.8343E+11, 9.9164E+11, 9.9992E+11, 1.0083E+12, 1.0167E+12,
     + 1.0252E+12, 1.0337E+12, 1.0423E+12, 1.0510E+12, 1.0597E+12,
     +  1.0686E+12, 1.0774E+12, 1.0864E+12, 1.0954E+12, 1.1045E+12,
     + 1.1137E+12, 1.1230E+12, 1.1323E+12, 1.1417E+12, 1.1511E+12,
     +  1.1607E+12, 1.1703E+12, 1.1800E+12, 1.1897E+12, 1.1996E+12,
     + 1.2095E+12, 1.2195E+12, 1.2296E+12, 1.2398E+12, 1.2500E+12,
     +  1.2603E+12, 1.2707E+12, 1.2812E+12, 1.2918E+12, 1.3024E+12,
     + 1.3131E+12, 1.3240E+12, 1.3349E+12, 1.3459E+12, 1.3569E+12,
     +  1.3681E+12, 1.3793E+12, 1.3907E+12, 1.4021E+12, 1.4136E+12,
     + 1.4252E+12, 1.4369E+12, 1.4487E+12, 1.4606E+12, 1.4725E+12,
     +  1.4846E+12, 1.4967E+12, 1.5090E+12, 1.5213E+12, 1.5338E+12,
     + 1.5463E+12, 1.5590E+12, 1.5717E+12, 1.5845E+12, 1.5975E+12,
     +  1.6105E+12, 1.6236E+12, 1.6369E+12, 1.6502E+12, 1.6637E+12,
     + 1.6772E+12, 1.6909E+12, 1.7046E+12, 1.7185E+12, 1.7325E+12,
     +  1.7465E+12, 1.7607E+12, 1.7750E+12, 1.7895E+12, 1.8040E+12,
     + 1.8186E+12, 1.8334E+12, 1.8482E+12, 1.8632E+12, 1.8783E+12,
     +  1.8935E+12, 1.9088E+12, 1.9243E+12, 1.9398E+12, 1.9555E+12,
     + 1.9713E+12, 1.9872E+12, 2.0033E+12, 2.0195E+12, 2.0358E+12,
     +  2.0522E+12, 2.0687E+12, 2.0854E+12, 2.1022E+12, 2.1191E+12,
     + 2.1362E+12, 2.1534E+12, 2.1707E+12, 2.1881E+12, 2.2057E+12,
     +  2.2235E+12, 2.2413E+12, 2.2593E+12, 2.2774E+12, 2.2957E+12,
     + 2.3141E+12, 2.3327E+12, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_PH3   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 1),QofT( 1,NT),Q(NT)
      data Tmx(1)/6000./
      data xgj/ 2./
c...      PH3
c...        --      1111
      data (QofT( 1,J),J=1,2408)/ 8.0002E+00, 8.0978E+00, 8.7691E+00,
     +  1.0169E+01, 1.2083E+01, 1.4315E+01, 1.6758E+01, 1.9366E+01,
     + 2.2117E+01, 2.5003E+01, 2.8020E+01, 3.1166E+01, 3.4440E+01,
     +  3.7840E+01, 4.1366E+01, 4.5015E+01, 4.8785E+01, 5.2675E+01,
     + 5.6681E+01, 6.0801E+01, 6.9372E+01, 7.8367E+01, 8.7766E+01,
     +  9.7549E+01, 1.0770E+02, 1.1820E+02, 1.2904E+02, 1.4021E+02,
     + 1.5169E+02, 1.6348E+02, 1.7556E+02, 1.8793E+02, 2.0058E+02,
     +  2.1351E+02, 2.2670E+02, 2.4016E+02, 2.5388E+02, 2.6785E+02,
     + 2.8207E+02, 2.9653E+02, 3.1123E+02, 3.2617E+02, 3.4134E+02,
     +  3.5674E+02, 3.7236E+02, 3.8821E+02, 4.0428E+02, 4.2056E+02,
     + 4.3706E+02, 4.5376E+02, 4.7068E+02, 4.8780E+02, 5.0513E+02,
     +  5.2265E+02, 5.4038E+02, 5.5830E+02, 5.7642E+02, 5.9473E+02,
     + 6.1323E+02, 6.3192E+02, 6.5079E+02, 6.6985E+02, 6.8910E+02,
     +  7.0852E+02, 7.2813E+02, 7.4791E+02, 7.6787E+02, 7.8801E+02,
     + 8.0832E+02, 8.2881E+02, 8.4946E+02, 8.7029E+02, 8.9128E+02,
     +  9.1244E+02, 9.3377E+02, 9.5526E+02, 9.7692E+02, 9.9874E+02,
     + 1.0207E+03, 1.0429E+03, 1.0652E+03, 1.0876E+03, 1.1103E+03,
     +  1.1330E+03, 1.1560E+03, 1.1791E+03, 1.2023E+03, 1.2257E+03,
     + 1.2493E+03, 1.2730E+03, 1.2968E+03, 1.3209E+03, 1.3450E+03,
     +  1.3693E+03, 1.3938E+03, 1.4184E+03, 1.4432E+03, 1.4681E+03,
     + 1.4932E+03, 1.5184E+03, 1.5438E+03, 1.5693E+03, 1.5950E+03,
     +  1.6208E+03, 1.6468E+03, 1.6729E+03, 1.6991E+03, 1.7256E+03,
     + 1.7521E+03, 1.7788E+03, 1.8057E+03, 1.8327E+03, 1.8599E+03,
     +  1.8872E+03, 1.9147E+03, 1.9423E+03, 1.9701E+03, 1.9980E+03,
     + 2.0261E+03, 2.0543E+03, 2.0827E+03, 2.1113E+03, 2.1400E+03,
     +  2.1688E+03, 2.1979E+03, 2.2270E+03, 2.2564E+03, 2.2859E+03,
     + 2.3155E+03, 2.3453E+03, 2.3753E+03, 2.4054E+03, 2.4357E+03,
     +  2.4662E+03, 2.4968E+03, 2.5276E+03, 2.5586E+03, 2.5897E+03,
     + 2.6210E+03, 2.6525E+03, 2.6841E+03, 2.7159E+03, 2.7479E+03,
     +  2.7801E+03, 2.8124E+03, 2.8449E+03, 2.8776E+03, 2.9105E+03,
     + 2.9436E+03, 2.9768E+03, 3.0102E+03, 3.0438E+03, 3.0776E+03,
     +  3.1116E+03, 3.1458E+03, 3.1801E+03, 3.2147E+03, 3.2494E+03,
     + 3.2844E+03, 3.3195E+03, 3.3548E+03, 3.3904E+03, 3.4261E+03,
     +  3.4620E+03, 3.4982E+03, 3.5345E+03, 3.5711E+03, 3.6078E+03,
     + 3.6448E+03, 3.6820E+03, 3.7193E+03, 3.7569E+03, 3.7948E+03,
     +  3.8328E+03, 3.8711E+03, 3.9095E+03, 3.9482E+03, 3.9871E+03,
     + 4.0263E+03, 4.0657E+03, 4.1053E+03, 4.1451E+03, 4.1852E+03,
     +  4.2255E+03, 4.2660E+03, 4.3068E+03, 4.3478E+03, 4.3891E+03,
     + 4.4305E+03, 4.4723E+03, 4.5143E+03, 4.5565E+03, 4.5990E+03,
     +  4.6417E+03, 4.6847E+03, 4.7280E+03, 4.7715E+03, 4.8153E+03,
     + 4.8593E+03, 4.9036E+03, 4.9481E+03, 4.9930E+03, 5.0381E+03,
     +  5.0834E+03, 5.1291E+03, 5.1750E+03, 5.2212E+03, 5.2676E+03,
     + 5.3144E+03, 5.3614E+03, 5.4087E+03, 5.4563E+03, 5.5042E+03,
     +  5.5524E+03, 5.6009E+03, 5.6497E+03, 5.6987E+03, 5.7481E+03,
     + 5.7978E+03, 5.8477E+03, 5.8980E+03, 5.9486E+03, 5.9995E+03,
     +  6.0507E+03, 6.1022E+03, 6.1541E+03, 6.2062E+03, 6.2587E+03,
     + 6.3115E+03, 6.3647E+03, 6.4181E+03, 6.4719E+03, 6.5260E+03,
     +  6.5805E+03, 6.6352E+03, 6.6904E+03, 6.7458E+03, 6.8016E+03,
     + 6.8578E+03, 6.9143E+03, 6.9712E+03, 7.0284E+03, 7.0859E+03,
     +  7.1438E+03, 7.2021E+03, 7.2608E+03, 7.3198E+03, 7.3791E+03,
     + 7.4389E+03, 7.4990E+03, 7.5595E+03, 7.6203E+03, 7.6816E+03,
     +  7.7432E+03, 7.8052E+03, 7.8676E+03, 7.9304E+03, 7.9936E+03,
     + 8.0571E+03, 8.1211E+03, 8.1855E+03, 8.2503E+03, 8.3154E+03,
     +  8.3810E+03, 8.4470E+03, 8.5134E+03, 8.5802E+03, 8.6475E+03,
     + 8.7151E+03, 8.7832E+03, 8.8517E+03, 8.9207E+03, 8.9900E+03,
     +  9.0598E+03, 9.1301E+03, 9.2007E+03, 9.2719E+03, 9.3434E+03,
     + 9.4155E+03, 9.4879E+03, 9.5608E+03, 9.6342E+03, 9.7081E+03,
     +  9.7824E+03, 9.8571E+03, 9.9324E+03, 1.0008E+04, 1.0084E+04,
     + 1.0161E+04, 1.0238E+04, 1.0316E+04, 1.0394E+04, 1.0472E+04,
     +  1.0551E+04, 1.0631E+04, 1.0711E+04, 1.0792E+04, 1.0873E+04,
     + 1.0954E+04, 1.1037E+04, 1.1119E+04, 1.1202E+04, 1.1286E+04,
     +  1.1370E+04, 1.1455E+04, 1.1540E+04, 1.1626E+04, 1.1712E+04,
     + 1.1799E+04, 1.1887E+04, 1.1974E+04, 1.2063E+04, 1.2152E+04,
     +  1.2242E+04, 1.2332E+04, 1.2423E+04, 1.2514E+04, 1.2606E+04,
     + 1.2698E+04, 1.2791E+04, 1.2885E+04, 1.2979E+04, 1.3074E+04,
     +  1.3169E+04, 1.3265E+04, 1.3362E+04, 1.3459E+04, 1.3557E+04,
     + 1.3655E+04, 1.3754E+04, 1.3854E+04, 1.3954E+04, 1.4055E+04,
     +  1.4156E+04, 1.4258E+04, 1.4361E+04, 1.4464E+04, 1.4568E+04,
     + 1.4673E+04, 1.4778E+04, 1.4884E+04, 1.4991E+04, 1.5098E+04,
     +  1.5206E+04, 1.5315E+04, 1.5424E+04, 1.5534E+04, 1.5645E+04,
     + 1.5756E+04, 1.5868E+04, 1.5981E+04, 1.6095E+04, 1.6209E+04,
     +  1.6324E+04, 1.6439E+04, 1.6556E+04, 1.6673E+04, 1.6790E+04,
     + 1.6909E+04, 1.7028E+04, 1.7148E+04, 1.7269E+04, 1.7390E+04,
     +  1.7512E+04, 1.7635E+04, 1.7759E+04, 1.7883E+04, 1.8008E+04,
     + 1.8134E+04, 1.8261E+04, 1.8389E+04, 1.8517E+04, 1.8646E+04,
     +  1.8776E+04, 1.8907E+04, 1.9038E+04, 1.9170E+04, 1.9304E+04,
     + 1.9437E+04, 1.9572E+04, 1.9708E+04, 1.9844E+04, 1.9981E+04,
     +  2.0120E+04, 2.0259E+04, 2.0398E+04, 2.0539E+04, 2.0680E+04,
     + 2.0823E+04, 2.0966E+04, 2.1110E+04, 2.1255E+04, 2.1401E+04,
     +  2.1548E+04, 2.1695E+04, 2.1844E+04, 2.1993E+04, 2.2144E+04,
     + 2.2295E+04, 2.2447E+04, 2.2600E+04, 2.2755E+04, 2.2910E+04,
     +  2.3065E+04, 2.3222E+04, 2.3380E+04, 2.3539E+04, 2.3699E+04,
     + 2.3859E+04, 2.4021E+04, 2.4184E+04, 2.4347E+04, 2.4512E+04,
     +  2.4678E+04, 2.4844E+04, 2.5012E+04, 2.5181E+04, 2.5350E+04,
     + 2.5521E+04, 2.5693E+04, 2.5866E+04, 2.6039E+04, 2.6214E+04,
     +  2.6390E+04, 2.6567E+04, 2.6745E+04, 2.6924E+04, 2.7104E+04,
     + 2.7286E+04, 2.7468E+04, 2.7651E+04, 2.7836E+04, 2.8022E+04,
     +  2.8208E+04, 2.8396E+04, 2.8585E+04, 2.8775E+04, 2.8966E+04,
     + 2.9159E+04, 2.9352E+04, 2.9547E+04, 2.9743E+04, 2.9940E+04,
     +  3.0138E+04, 3.0338E+04, 3.0538E+04, 3.0740E+04, 3.0943E+04,
     + 3.1147E+04, 3.1352E+04, 3.1559E+04, 3.1767E+04, 3.1976E+04,
     +  3.2186E+04, 3.2398E+04, 3.2610E+04, 3.2824E+04, 3.3040E+04,
     + 3.3256E+04, 3.3474E+04, 3.3693E+04, 3.3914E+04, 3.4135E+04,
     +  3.4358E+04, 3.4583E+04, 3.4808E+04, 3.5035E+04, 3.5264E+04,
     + 3.5493E+04, 3.5724E+04, 3.5957E+04, 3.6190E+04, 3.6426E+04,
     +  3.6662E+04, 3.6900E+04, 3.7139E+04, 3.7380E+04, 3.7622E+04,
     + 3.7865E+04, 3.8110E+04, 3.8357E+04, 3.8604E+04, 3.8854E+04,
     +  3.9104E+04, 3.9356E+04, 3.9610E+04, 3.9865E+04, 4.0122E+04,
     + 4.0380E+04, 4.0639E+04, 4.0900E+04, 4.1163E+04, 4.1427E+04,
     +  4.1692E+04, 4.1960E+04, 4.2228E+04, 4.2499E+04, 4.2770E+04,
     + 4.3044E+04, 4.3319E+04, 4.3595E+04, 4.3873E+04, 4.4153E+04,
     +  4.4435E+04, 4.4718E+04, 4.5002E+04, 4.5288E+04, 4.5576E+04,
     + 4.5866E+04, 4.6157E+04, 4.6450E+04, 4.6744E+04, 4.7041E+04,
     +  4.7339E+04, 4.7638E+04, 4.7940E+04, 4.8243E+04, 4.8548E+04,
     + 4.8854E+04, 4.9163E+04, 4.9473E+04, 4.9785E+04, 5.0098E+04,
     +  5.0414E+04, 5.0731E+04, 5.1050E+04, 5.1371E+04, 5.1693E+04,
     + 5.2018E+04, 5.2344E+04, 5.2672E+04, 5.3002E+04, 5.3334E+04,
     +  5.3668E+04, 5.4004E+04, 5.4341E+04, 5.4681E+04, 5.5022E+04,
     + 5.5366E+04, 5.5711E+04, 5.6058E+04, 5.6407E+04, 5.6758E+04,
     +  5.7111E+04, 5.7467E+04, 5.7824E+04, 5.8183E+04, 5.8544E+04,
     + 5.8907E+04, 5.9272E+04, 5.9640E+04, 6.0009E+04, 6.0380E+04,
     +  6.0754E+04, 6.1129E+04, 6.1507E+04, 6.1887E+04, 6.2268E+04,
     + 6.2652E+04, 6.3039E+04, 6.3427E+04, 6.3817E+04, 6.4210E+04,
     +  6.4605E+04, 6.5002E+04, 6.5401E+04, 6.5803E+04, 6.6206E+04,
     + 6.6612E+04, 6.7020E+04, 6.7431E+04, 6.7843E+04, 6.8258E+04,
     +  6.8676E+04, 6.9095E+04, 6.9517E+04, 6.9941E+04, 7.0368E+04,
     + 7.0797E+04, 7.1228E+04, 7.1662E+04, 7.2098E+04, 7.2536E+04,
     +  7.2977E+04, 7.3420E+04, 7.3866E+04, 7.4314E+04, 7.4765E+04,
     + 7.5218E+04, 7.5674E+04, 7.6132E+04, 7.6592E+04, 7.7055E+04,
     +  7.7521E+04, 7.7989E+04, 7.8460E+04, 7.8933E+04, 7.9409E+04,
     + 7.9887E+04, 8.0368E+04, 8.0852E+04, 8.1338E+04, 8.1827E+04,
     +  8.2319E+04, 8.2813E+04, 8.3310E+04, 8.3810E+04, 8.4312E+04,
     + 8.4817E+04, 8.5325E+04, 8.5836E+04, 8.6349E+04, 8.6865E+04,
     +  8.7384E+04, 8.7906E+04, 8.8430E+04, 8.8957E+04, 8.9488E+04,
     + 9.0021E+04, 9.0557E+04, 9.1095E+04, 9.1637E+04, 9.2182E+04,
     +  9.2729E+04, 9.3279E+04, 9.3833E+04, 9.4389E+04, 9.4948E+04,
     + 9.5511E+04, 9.6076E+04, 9.6644E+04, 9.7216E+04, 9.7790E+04,
     +  9.8368E+04, 9.8948E+04, 9.9532E+04, 1.0012E+05, 1.0071E+05,
     + 1.0130E+05, 1.0190E+05, 1.0250E+05, 1.0310E+05, 1.0371E+05,
     +  1.0431E+05, 1.0493E+05, 1.0554E+05, 1.0616E+05, 1.0678E+05,
     + 1.0741E+05, 1.0804E+05, 1.0867E+05, 1.0930E+05, 1.0994E+05,
     +  1.1058E+05, 1.1123E+05, 1.1188E+05, 1.1253E+05, 1.1318E+05,
     + 1.1384E+05, 1.1451E+05, 1.1517E+05, 1.1584E+05, 1.1651E+05,
     +  1.1719E+05, 1.1787E+05, 1.1855E+05, 1.1924E+05, 1.1993E+05,
     + 1.2063E+05, 1.2132E+05, 1.2202E+05, 1.2273E+05, 1.2344E+05,
     +  1.2415E+05, 1.2487E+05, 1.2559E+05, 1.2631E+05, 1.2704E+05,
     + 1.2777E+05, 1.2850E+05, 1.2924E+05, 1.2999E+05, 1.3073E+05,
     +  1.3148E+05, 1.3224E+05, 1.3299E+05, 1.3376E+05, 1.3452E+05,
     + 1.3529E+05, 1.3607E+05, 1.3684E+05, 1.3762E+05, 1.3841E+05,
     +  1.3920E+05, 1.3999E+05, 1.4079E+05, 1.4159E+05, 1.4240E+05,
     + 1.4321E+05, 1.4403E+05, 1.4484E+05, 1.4567E+05, 1.4649E+05,
     +  1.4732E+05, 1.4816E+05, 1.4900E+05, 1.4984E+05, 1.5069E+05,
     + 1.5154E+05, 1.5240E+05, 1.5326E+05, 1.5413E+05, 1.5500E+05,
     +  1.5587E+05, 1.5675E+05, 1.5763E+05, 1.5852E+05, 1.5941E+05,
     + 1.6031E+05, 1.6121E+05, 1.6212E+05, 1.6303E+05, 1.6394E+05,
     +  1.6486E+05, 1.6579E+05, 1.6672E+05, 1.6765E+05, 1.6859E+05,
     + 1.6953E+05, 1.7048E+05, 1.7143E+05, 1.7239E+05, 1.7335E+05,
     +  1.7432E+05, 1.7529E+05, 1.7627E+05, 1.7725E+05, 1.7823E+05,
     + 1.7922E+05, 1.8022E+05, 1.8122E+05, 1.8223E+05, 1.8324E+05,
     +  1.8426E+05, 1.8528E+05, 1.8630E+05, 1.8733E+05, 1.8837E+05,
     + 1.8941E+05, 1.9046E+05, 1.9151E+05, 1.9257E+05, 1.9363E+05,
     +  1.9470E+05, 1.9577E+05, 1.9685E+05, 1.9793E+05, 1.9902E+05,
     + 2.0012E+05, 2.0122E+05, 2.0232E+05, 2.0343E+05, 2.0455E+05,
     +  2.0567E+05, 2.0679E+05, 2.0793E+05, 2.0906E+05, 2.1021E+05,
     + 2.1136E+05, 2.1251E+05, 2.1367E+05, 2.1484E+05, 2.1601E+05,
     +  2.1719E+05, 2.1837E+05, 2.1956E+05, 2.2075E+05, 2.2195E+05,
     + 2.2316E+05, 2.2437E+05, 2.2559E+05, 2.2681E+05, 2.2804E+05,
     +  2.2928E+05, 2.3052E+05, 2.3177E+05, 2.3302E+05, 2.3428E+05,
     + 2.3555E+05, 2.3682E+05, 2.3810E+05, 2.3938E+05, 2.4067E+05,
     +  2.4197E+05, 2.4327E+05, 2.4458E+05, 2.4590E+05, 2.4722E+05,
     + 2.4855E+05, 2.4988E+05, 2.5122E+05, 2.5257E+05, 2.5392E+05,
     +  2.5528E+05, 2.5665E+05, 2.5802E+05, 2.5940E+05, 2.6079E+05,
     + 2.6218E+05, 2.6358E+05, 2.6499E+05, 2.6640E+05, 2.6782E+05,
     +  2.6925E+05, 2.7068E+05, 2.7212E+05, 2.7357E+05, 2.7502E+05,
     + 2.7648E+05, 2.7795E+05, 2.7942E+05, 2.8090E+05, 2.8239E+05,
     +  2.8388E+05, 2.8539E+05, 2.8690E+05, 2.8841E+05, 2.8994E+05,
     + 2.9147E+05, 2.9300E+05, 2.9455E+05, 2.9610E+05, 2.9766E+05,
     +  2.9923E+05, 3.0080E+05, 3.0238E+05, 3.0397E+05, 3.0557E+05,
     + 3.0717E+05, 3.0878E+05, 3.1040E+05, 3.1203E+05, 3.1366E+05,
     +  3.1530E+05, 3.1695E+05, 3.1861E+05, 3.2027E+05, 3.2194E+05,
     + 3.2362E+05, 3.2531E+05, 3.2700E+05, 3.2871E+05, 3.3042E+05,
     +  3.3214E+05, 3.3386E+05, 3.3560E+05, 3.3734E+05, 3.3909E+05,
     + 3.4085E+05, 3.4261E+05, 3.4439E+05, 3.4617E+05, 3.4796E+05,
     +  3.4976E+05, 3.5157E+05, 3.5338E+05, 3.5521E+05, 3.5704E+05,
     + 3.5888E+05, 3.6073E+05, 3.6259E+05, 3.6445E+05, 3.6633E+05,
     +  3.6821E+05, 3.7010E+05, 3.7200E+05, 3.7391E+05, 3.7583E+05,
     + 3.7775E+05, 3.7969E+05, 3.8163E+05, 3.8358E+05, 3.8554E+05,
     +  3.8751E+05, 3.8949E+05, 3.9148E+05, 3.9347E+05, 3.9548E+05,
     + 3.9749E+05, 3.9952E+05, 4.0155E+05, 4.0359E+05, 4.0564E+05,
     +  4.0770E+05, 4.0977E+05, 4.1185E+05, 4.1394E+05, 4.1603E+05,
     + 4.1814E+05, 4.2026E+05, 4.2238E+05, 4.2452E+05, 4.2666E+05,
     +  4.2881E+05, 4.3098E+05, 4.3315E+05, 4.3533E+05, 4.3752E+05,
     + 4.3972E+05, 4.4193E+05, 4.4416E+05, 4.4639E+05, 4.4863E+05,
     +  4.5088E+05, 4.5314E+05, 4.5541E+05, 4.5769E+05, 4.5998E+05,
     + 4.6228E+05, 4.6459E+05, 4.6691E+05, 4.6924E+05, 4.7158E+05,
     +  4.7393E+05, 4.7629E+05, 4.7867E+05, 4.8105E+05, 4.8344E+05,
     + 4.8584E+05, 4.8826E+05, 4.9068E+05, 4.9312E+05, 4.9556E+05,
     +  4.9802E+05, 5.0048E+05, 5.0296E+05, 5.0545E+05, 5.0794E+05,
     + 5.1045E+05, 5.1297E+05, 5.1550E+05, 5.1805E+05, 5.2060E+05,
     +  5.2316E+05, 5.2574E+05, 5.2832E+05, 5.3092E+05, 5.3353E+05,
     + 5.3615E+05, 5.3878E+05, 5.4142E+05, 5.4407E+05, 5.4673E+05,
     +  5.4941E+05, 5.5210E+05, 5.5479E+05, 5.5750E+05, 5.6023E+05,
     + 5.6296E+05, 5.6570E+05, 5.6846E+05, 5.7123E+05, 5.7401E+05,
     +  5.7680E+05, 5.7960E+05, 5.8242E+05, 5.8524E+05, 5.8808E+05,
     + 5.9093E+05, 5.9379E+05, 5.9667E+05, 5.9956E+05, 6.0245E+05,
     +  6.0537E+05, 6.0829E+05, 6.1122E+05, 6.1417E+05, 6.1713E+05,
     + 6.2010E+05, 6.2309E+05, 6.2609E+05, 6.2910E+05, 6.3212E+05,
     +  6.3515E+05, 6.3820E+05, 6.4126E+05, 6.4434E+05, 6.4742E+05,
     + 6.5052E+05, 6.5363E+05, 6.5676E+05, 6.5989E+05, 6.6304E+05,
     +  6.6621E+05, 6.6938E+05, 6.7257E+05, 6.7577E+05, 6.7899E+05,
     + 6.8222E+05, 6.8546E+05, 6.8871E+05, 6.9198E+05, 6.9526E+05,
     +  6.9856E+05, 7.0187E+05, 7.0519E+05, 7.0853E+05, 7.1188E+05,
     + 7.1524E+05, 7.1862E+05, 7.2201E+05, 7.2541E+05, 7.2883E+05,
     +  7.3226E+05, 7.3571E+05, 7.3917E+05, 7.4264E+05, 7.4613E+05,
     + 7.4963E+05, 7.5315E+05, 7.5668E+05, 7.6022E+05, 7.6378E+05,
     +  7.6736E+05, 7.7094E+05, 7.7455E+05, 7.7816E+05, 7.8179E+05,
     + 7.8544E+05, 7.8910E+05, 7.9277E+05, 7.9646E+05, 8.0017E+05,
     +  8.0388E+05, 8.0762E+05, 8.1137E+05, 8.1513E+05, 8.1891E+05,
     + 8.2270E+05, 8.2651E+05, 8.3033E+05, 8.3417E+05, 8.3803E+05,
     +  8.4189E+05, 8.4578E+05, 8.4968E+05, 8.5359E+05, 8.5752E+05,
     + 8.6147E+05, 8.6543E+05, 8.6941E+05, 8.7340E+05, 8.7741E+05,
     +  8.8143E+05, 8.8547E+05, 8.8953E+05, 8.9360E+05, 8.9768E+05,
     + 9.0179E+05, 9.0591E+05, 9.1004E+05, 9.1419E+05, 9.1836E+05,
     +  9.2254E+05, 9.2674E+05, 9.3096E+05, 9.3519E+05, 9.3944E+05,
     + 9.4370E+05, 9.4799E+05, 9.5228E+05, 9.5660E+05, 9.6093E+05,
     +  9.6528E+05, 9.6964E+05, 9.7402E+05, 9.7842E+05, 9.8283E+05,
     + 9.8727E+05, 9.9172E+05, 9.9618E+05, 1.0007E+06, 1.0052E+06,
     +  1.0097E+06, 1.0142E+06, 1.0188E+06, 1.0233E+06, 1.0279E+06,
     + 1.0325E+06, 1.0371E+06, 1.0418E+06, 1.0464E+06, 1.0511E+06,
     +  1.0558E+06, 1.0605E+06, 1.0652E+06, 1.0700E+06, 1.0747E+06,
     + 1.0795E+06, 1.0843E+06, 1.0891E+06, 1.0940E+06, 1.0988E+06,
     +  1.1037E+06, 1.1086E+06, 1.1135E+06, 1.1184E+06, 1.1234E+06,
     + 1.1283E+06, 1.1333E+06, 1.1383E+06, 1.1433E+06, 1.1484E+06,
     +  1.1534E+06, 1.1585E+06, 1.1636E+06, 1.1687E+06, 1.1738E+06,
     + 1.1790E+06, 1.1841E+06, 1.1893E+06, 1.1945E+06, 1.1997E+06,
     +  1.2050E+06, 1.2103E+06, 1.2155E+06, 1.2208E+06, 1.2262E+06,
     + 1.2315E+06, 1.2369E+06, 1.2422E+06, 1.2476E+06, 1.2531E+06,
     +  1.2585E+06, 1.2640E+06, 1.2694E+06, 1.2749E+06, 1.2804E+06,
     + 1.2860E+06, 1.2915E+06, 1.2971E+06, 1.3027E+06, 1.3083E+06,
     +  1.3140E+06, 1.3196E+06, 1.3253E+06, 1.3310E+06, 1.3367E+06,
     + 1.3425E+06, 1.3482E+06, 1.3540E+06, 1.3598E+06, 1.3656E+06,
     +  1.3715E+06, 1.3773E+06, 1.3832E+06, 1.3891E+06, 1.3951E+06,
     + 1.4010E+06, 1.4070E+06, 1.4130E+06, 1.4190E+06, 1.4250E+06,
     +  1.4311E+06, 1.4371E+06, 1.4432E+06, 1.4494E+06, 1.4555E+06,
     + 1.4617E+06, 1.4678E+06, 1.4740E+06, 1.4803E+06, 1.4865E+06,
     +  1.4928E+06, 1.4991E+06, 1.5054E+06, 1.5117E+06, 1.5181E+06,
     + 1.5245E+06, 1.5309E+06, 1.5373E+06, 1.5437E+06, 1.5502E+06,
     +  1.5567E+06, 1.5632E+06, 1.5698E+06, 1.5763E+06, 1.5829E+06,
     + 1.5895E+06, 1.5961E+06, 1.6028E+06, 1.6094E+06, 1.6161E+06,
     +  1.6229E+06, 1.6296E+06, 1.6364E+06, 1.6432E+06, 1.6500E+06,
     + 1.6568E+06, 1.6637E+06, 1.6705E+06, 1.6774E+06, 1.6844E+06,
     +  1.6913E+06, 1.6983E+06, 1.7053E+06, 1.7123E+06, 1.7194E+06,
     + 1.7264E+06, 1.7335E+06, 1.7406E+06, 1.7478E+06, 1.7549E+06,
     +  1.7621E+06, 1.7693E+06, 1.7766E+06, 1.7838E+06, 1.7911E+06,
     + 1.7984E+06, 1.8058E+06, 1.8131E+06, 1.8205E+06, 1.8279E+06,
     +  1.8354E+06, 1.8428E+06, 1.8503E+06, 1.8578E+06, 1.8653E+06,
     + 1.8729E+06, 1.8805E+06, 1.8881E+06, 1.8957E+06, 1.9034E+06,
     +  1.9110E+06, 1.9188E+06, 1.9265E+06, 1.9342E+06, 1.9420E+06,
     + 1.9498E+06, 1.9577E+06, 1.9655E+06, 1.9734E+06, 1.9813E+06,
     +  1.9893E+06, 1.9972E+06, 2.0052E+06, 2.0132E+06, 2.0213E+06,
     + 2.0293E+06, 2.0374E+06, 2.0455E+06, 2.0537E+06, 2.0619E+06,
     +  2.0701E+06, 2.0783E+06, 2.0865E+06, 2.0948E+06, 2.1031E+06,
     + 2.1114E+06, 2.1198E+06, 2.1282E+06, 2.1366E+06, 2.1450E+06,
     +  2.1535E+06, 2.1620E+06, 2.1705E+06, 2.1791E+06, 2.1876E+06,
     + 2.1962E+06, 2.2049E+06, 2.2135E+06, 2.2222E+06, 2.2309E+06,
     +  2.2396E+06, 2.2484E+06, 2.2572E+06, 2.2660E+06, 2.2749E+06,
     + 2.2837E+06, 2.2926E+06, 2.3016E+06, 2.3105E+06, 2.3195E+06,
     +  2.3285E+06, 2.3376E+06, 2.3467E+06, 2.3558E+06, 2.3649E+06,
     + 2.3740E+06, 2.3832E+06, 2.3924E+06, 2.4017E+06, 2.4109E+06,
     +  2.4202E+06, 2.4296E+06, 2.4389E+06, 2.4483E+06, 2.4577E+06,
     + 2.4672E+06, 2.4766E+06, 2.4861E+06, 2.4957E+06, 2.5052E+06,
     +  2.5148E+06, 2.5244E+06, 2.5341E+06, 2.5438E+06, 2.5535E+06,
     + 2.5632E+06, 2.5730E+06, 2.5828E+06, 2.5926E+06, 2.6025E+06,
     +  2.6123E+06, 2.6223E+06, 2.6322E+06, 2.6422E+06, 2.6522E+06,
     + 2.6622E+06, 2.6723E+06, 2.6824E+06, 2.6925E+06, 2.7027E+06,
     +  2.7128E+06, 2.7231E+06, 2.7333E+06, 2.7436E+06, 2.7539E+06,
     + 2.7642E+06, 2.7746E+06, 2.7850E+06, 2.7954E+06, 2.8059E+06,
     +  2.8164E+06, 2.8269E+06, 2.8375E+06, 2.8481E+06, 2.8587E+06,
     + 2.8693E+06, 2.8800E+06, 2.8907E+06, 2.9015E+06, 2.9122E+06,
     +  2.9231E+06, 2.9339E+06, 2.9448E+06, 2.9557E+06, 2.9666E+06,
     + 2.9776E+06, 2.9886E+06, 2.9996E+06, 3.0107E+06, 3.0218E+06,
     +  3.0329E+06, 3.0440E+06, 3.0552E+06, 3.0665E+06, 3.0777E+06,
     + 3.0890E+06, 3.1003E+06, 3.1117E+06, 3.1231E+06, 3.1345E+06,
     +  3.1459E+06, 3.1574E+06, 3.1689E+06, 3.1805E+06, 3.1921E+06,
     + 3.2037E+06, 3.2153E+06, 3.2270E+06, 3.2387E+06, 3.2505E+06,
     +  3.2623E+06, 3.2741E+06, 3.2859E+06, 3.2978E+06, 3.3097E+06,
     + 3.3217E+06, 3.3337E+06, 3.3457E+06, 3.3577E+06, 3.3698E+06,
     +  3.3820E+06, 3.3941E+06, 3.4063E+06, 3.4185E+06, 3.4308E+06,
     + 3.4431E+06, 3.4554E+06, 3.4677E+06, 3.4801E+06, 3.4926E+06,
     +  3.5050E+06, 3.5175E+06, 3.5301E+06, 3.5426E+06, 3.5552E+06,
     + 3.5679E+06, 3.5805E+06, 3.5932E+06, 3.6060E+06, 3.6188E+06,
     +  3.6316E+06, 3.6444E+06, 3.6573E+06, 3.6702E+06, 3.6832E+06,
     + 3.6962E+06, 3.7092E+06, 3.7222E+06, 3.7353E+06, 3.7485E+06,
     +  3.7616E+06, 3.7748E+06, 3.7881E+06, 3.8014E+06, 3.8147E+06,
     + 3.8280E+06, 3.8414E+06, 3.8548E+06, 3.8683E+06, 3.8818E+06,
     +  3.8953E+06, 3.9088E+06, 3.9224E+06, 3.9361E+06, 3.9498E+06,
     + 3.9635E+06, 3.9772E+06, 3.9910E+06, 4.0048E+06, 4.0187E+06,
     +  4.0326E+06, 4.0465E+06, 4.0605E+06, 4.0745E+06, 4.0885E+06,
     + 4.1026E+06, 4.1167E+06, 4.1308E+06, 4.1450E+06, 4.1593E+06,
     +  4.1735E+06, 4.1878E+06, 4.2022E+06, 4.2165E+06, 4.2310E+06,
     + 4.2454E+06, 4.2599E+06, 4.2744E+06, 4.2890E+06, 4.3036E+06,
     +  4.3183E+06, 4.3329E+06, 4.3477E+06, 4.3624E+06, 4.3772E+06,
     + 4.3920E+06, 4.4069E+06, 4.4218E+06, 4.4368E+06, 4.4518E+06,
     +  4.4668E+06, 4.4818E+06, 4.4970E+06, 4.5121E+06, 4.5273E+06,
     + 4.5425E+06, 4.5578E+06, 4.5731E+06, 4.5884E+06, 4.6038E+06,
     +  4.6192E+06, 4.6346E+06, 4.6501E+06, 4.6657E+06, 4.6812E+06,
     + 4.6968E+06, 4.7125E+06, 4.7282E+06, 4.7439E+06, 4.7597E+06,
     +  4.7755E+06, 4.7914E+06, 4.8072E+06, 4.8232E+06, 4.8391E+06,
     + 4.8552E+06, 4.8712E+06, 4.8873E+06, 4.9034E+06, 4.9196E+06,
     +  4.9358E+06, 4.9521E+06, 4.9684E+06, 4.9847E+06, 5.0011E+06,
     + 5.0175E+06, 5.0339E+06, 5.0504E+06, 5.0670E+06, 5.0835E+06,
     +  5.1001E+06, 5.1168E+06, 5.1335E+06, 5.1502E+06, 5.1670E+06,
     + 5.1838E+06, 5.2007E+06, 5.2176E+06, 5.2346E+06, 5.2516E+06,
     +  5.2686E+06, 5.2857E+06, 5.3028E+06, 5.3199E+06, 5.3371E+06,
     + 5.3544E+06, 5.3716E+06, 5.3890E+06, 5.4063E+06, 5.4237E+06,
     +  5.4412E+06, 5.4587E+06, 5.4762E+06, 5.4938E+06, 5.5114E+06,
     + 5.5290E+06, 5.5467E+06, 5.5645E+06, 5.5823E+06, 5.6001E+06,
     +  5.6179E+06, 5.6359E+06, 5.6538E+06, 5.6718E+06, 5.6899E+06,
     + 5.7079E+06, 5.7261E+06, 5.7442E+06, 5.7624E+06, 5.7807E+06,
     +  5.7990E+06, 5.8173E+06, 5.8357E+06, 5.8541E+06, 5.8726E+06,
     + 5.8911E+06, 5.9097E+06, 5.9283E+06, 5.9469E+06, 5.9656E+06,
     +  5.9844E+06, 6.0031E+06, 6.0219E+06, 6.0408E+06, 6.0597E+06,
     + 6.0787E+06, 6.0977E+06, 6.1167E+06, 6.1358E+06, 6.1549E+06,
     +  6.1741E+06, 6.1933E+06, 6.2125E+06, 6.2319E+06, 6.2512E+06,
     + 6.2706E+06, 6.2900E+06, 6.3095E+06, 6.3290E+06, 6.3486E+06,
     +  6.3682E+06, 6.3879E+06, 6.4076E+06, 6.4273E+06, 6.4471E+06,
     + 6.4670E+06, 6.4869E+06, 6.5068E+06, 6.5268E+06, 6.5468E+06,
     +  6.5668E+06, 6.5869E+06, 6.6071E+06, 6.6273E+06, 6.6475E+06,
     + 6.6678E+06, 6.6882E+06, 6.7086E+06, 6.7290E+06, 6.7495E+06,
     +  6.7700E+06, 6.7905E+06, 6.8112E+06, 6.8318E+06, 6.8525E+06,
     + 6.8733E+06, 6.8941E+06, 6.9149E+06, 6.9358E+06, 6.9567E+06,
     +  6.9777E+06, 6.9987E+06, 7.0198E+06, 7.0409E+06, 7.0621E+06,
     + 7.0833E+06, 7.1046E+06, 7.1259E+06, 7.1472E+06, 7.1686E+06,
     +  7.1901E+06, 7.2115E+06, 7.2331E+06, 7.2547E+06, 7.2763E+06,
     + 7.2980E+06, 7.3197E+06, 7.3415E+06, 7.3633E+06, 7.3852E+06,
     +  7.4071E+06, 7.4290E+06, 7.4511E+06, 7.4731E+06, 7.4952E+06,
     + 7.5174E+06, 7.5396E+06, 7.5618E+06, 7.5841E+06, 7.6064E+06,
     +  7.6288E+06, 7.6513E+06, 7.6738E+06, 7.6963E+06, 7.7189E+06,
     + 7.7415E+06, 7.7642E+06, 7.7869E+06, 7.8097E+06, 7.8325E+06,
     +  7.8554E+06, 7.8783E+06, 7.9013E+06, 7.9243E+06, 7.9473E+06,
     + 7.9705E+06, 7.9936E+06, 8.0168E+06, 8.0401E+06, 8.0634E+06,
     +  8.0867E+06, 8.1101E+06, 8.1336E+06, 8.1571E+06, 8.1807E+06,
     + 8.2043E+06, 8.2279E+06, 8.2516E+06, 8.2753E+06, 8.2991E+06,
     +  8.3230E+06, 8.3469E+06, 8.3708E+06, 8.3948E+06, 8.4189E+06,
     + 8.4430E+06, 8.4671E+06, 8.4913E+06, 8.5155E+06, 8.5398E+06,
     +  8.5642E+06, 8.5885E+06, 8.6130E+06, 8.6375E+06, 8.6620E+06,
     + 8.6866E+06, 8.7112E+06, 8.7359E+06, 8.7607E+06, 8.7855E+06,
     +  8.8103E+06, 8.8352E+06, 8.8601E+06, 8.8851E+06, 8.9102E+06,
     + 8.9352E+06, 8.9604E+06, 8.9856E+06, 9.0108E+06, 9.0361E+06,
     +  9.0615E+06, 9.0868E+06, 9.1123E+06, 9.1378E+06, 9.1633E+06,
     + 9.1889E+06, 9.2146E+06, 9.2403E+06, 9.2660E+06, 9.2918E+06,
     +  9.3177E+06, 9.3436E+06, 9.3695E+06, 9.3955E+06, 9.4216E+06,
     + 9.4477E+06, 9.4739E+06, 9.5001E+06, 9.5263E+06, 9.5526E+06,
     +  9.5790E+06, 9.6054E+06, 9.6319E+06, 9.6584E+06, 9.6850E+06,
     + 9.7116E+06, 9.7383E+06, 9.7650E+06, 9.7918E+06, 9.8186E+06,
     +  9.8455E+06, 9.8724E+06, 9.8994E+06, 9.9264E+06, 9.9535E+06,
     + 9.9807E+06, 1.0008E+07, 1.0035E+07, 1.0062E+07, 1.0090E+07,
     +  1.0117E+07, 1.0145E+07, 1.0172E+07, 1.0200E+07, 1.0227E+07,
     + 1.0255E+07, 1.0283E+07, 1.0310E+07, 1.0338E+07, 1.0366E+07,
     +  1.0394E+07, 1.0422E+07, 1.0450E+07, 1.0478E+07, 1.0506E+07,
     + 1.0535E+07, 1.0563E+07, 1.0591E+07, 1.0619E+07, 1.0648E+07,
     +  1.0676E+07, 1.0705E+07, 1.0733E+07, 1.0762E+07, 1.0791E+07,
     + 1.0819E+07, 1.0848E+07, 1.0877E+07, 1.0906E+07, 1.0935E+07,
     +  1.0964E+07, 1.0993E+07, 1.1022E+07, 1.1051E+07, 1.1080E+07,
     + 1.1110E+07, 1.1139E+07, 1.1168E+07, 1.1198E+07, 1.1227E+07,
     +  1.1257E+07, 1.1286E+07, 1.1316E+07, 1.1346E+07, 1.1375E+07,
     + 1.1405E+07, 1.1435E+07, 1.1465E+07, 1.1495E+07, 1.1525E+07,
     +  1.1555E+07, 1.1585E+07, 1.1615E+07, 1.1646E+07, 1.1676E+07,
     + 1.1706E+07, 1.1737E+07, 1.1767E+07, 1.1798E+07, 1.1828E+07,
     +  1.1859E+07, 1.1889E+07, 1.1920E+07, 1.1951E+07, 1.1982E+07,
     + 1.2013E+07, 1.2044E+07, 1.2075E+07, 1.2106E+07, 1.2137E+07,
     +  1.2168E+07, 1.2199E+07, 1.2230E+07, 1.2262E+07, 1.2293E+07,
     + 1.2325E+07, 1.2356E+07, 1.2388E+07, 1.2419E+07, 1.2451E+07,
     +  1.2483E+07, 1.2514E+07, 1.2546E+07, 1.2578E+07, 1.2610E+07,
     + 1.2642E+07, 1.2674E+07, 1.2706E+07, 1.2739E+07, 1.2771E+07,
     +  1.2803E+07, 1.2835E+07, 1.2868E+07, 1.2900E+07, 1.2933E+07,
     + 1.2965E+07, 1.2998E+07, 1.3031E+07, 1.3063E+07, 1.3096E+07,
     +  1.3129E+07, 1.3162E+07, 1.3195E+07, 1.3228E+07, 1.3261E+07,
     + 1.3294E+07, 1.3327E+07, 1.3360E+07, 1.3394E+07, 1.3427E+07,
     +  1.3460E+07, 1.3494E+07, 1.3527E+07, 1.3561E+07, 1.3595E+07,
     + 1.3628E+07, 1.3662E+07, 1.3696E+07, 1.3730E+07, 1.3764E+07,
     +  1.3798E+07, 1.3832E+07, 1.3866E+07, 1.3900E+07, 1.3934E+07,
     + 1.3968E+07, 1.4003E+07, 1.4037E+07, 1.4071E+07, 1.4106E+07,
     +  1.4141E+07, 1.4175E+07, 1.4210E+07, 1.4244E+07, 1.4279E+07,
     + 1.4314E+07, 1.4349E+07, 1.4384E+07, 1.4419E+07, 1.4454E+07,
     +  1.4489E+07, 1.4524E+07, 1.4560E+07, 1.4595E+07, 1.4630E+07,
     + 1.4666E+07, 1.4701E+07, 1.4737E+07, 1.4772E+07, 1.4808E+07,
     +  1.4844E+07, 1.4879E+07, 1.4915E+07, 1.4951E+07, 1.4987E+07,
     + 1.5023E+07, 1.5059E+07, 1.5095E+07, 1.5131E+07, 1.5167E+07,
     +  1.5204E+07, 1.5240E+07, 1.5276E+07, 1.5313E+07, 1.5349E+07,
     + 1.5386E+07, 1.5423E+07, 1.5459E+07, 1.5496E+07, 1.5533E+07,
     +  1.5570E+07, 1.5607E+07, 1.5644E+07, 1.5681E+07, 1.5718E+07,
     + 1.5755E+07, 1.5792E+07, 1.5829E+07, 1.5867E+07, 1.5904E+07,
     +  1.5942E+07, 1.5979E+07, 1.6017E+07, 1.6054E+07, 1.6092E+07,
     + 1.6130E+07, 1.6167E+07, 1.6205E+07, 1.6243E+07, 1.6281E+07,
     +  1.6319E+07, 1.6357E+07, 1.6396E+07, 1.6434E+07, 1.6472E+07,
     + 1.6510E+07, 1.6549E+07, 1.6587E+07, 1.6626E+07, 1.6664E+07,
     +  1.6703E+07, 1.6742E+07, 1.6780E+07, 1.6819E+07, 1.6858E+07,
     + 1.6897E+07, 1.6936E+07, 1.6975E+07, 1.7014E+07, 1.7053E+07,
     +  1.7093E+07, 1.7132E+07, 1.7171E+07, 1.7211E+07, 1.7250E+07,
     + 1.7290E+07, 1.7329E+07, 1.7369E+07, 1.7409E+07, 1.7448E+07,
     +  1.7488E+07, 1.7528E+07, 1.7568E+07, 1.7608E+07, 1.7648E+07,
     + 1.7688E+07, 1.7728E+07, 1.7769E+07, 1.7809E+07, 1.7849E+07,
     +  1.7890E+07, 1.7930E+07, 1.7971E+07, 1.8011E+07, 1.8052E+07,
     + 1.8093E+07, 1.8134E+07, 1.8174E+07, 1.8215E+07, 1.8256E+07,
     +  1.8297E+07, 1.8338E+07, 1.8380E+07, 1.8421E+07, 1.8462E+07,
     + 1.8503E+07, 1.8545E+07, 1.8586E+07, 1.8628E+07, 1.8669E+07,
     +  1.8711E+07, 1.8753E+07, 1.8794E+07, 1.8836E+07, 1.8878E+07,
     + 1.8920E+07, 1.8962E+07, 1.9004E+07, 1.9046E+07, 1.9088E+07,
     +  1.9131E+07, 1.9173E+07, 1.9215E+07, 1.9258E+07, 1.9300E+07,
     + 1.9343E+07, 1.9385E+07, 1.9428E+07, 1.9471E+07, 1.9514E+07,
     +  1.9556E+07, 1.9599E+07, 1.9642E+07, 1.9685E+07, 1.9728E+07,
     + 1.9772E+07, 1.9815E+07, 1.9858E+07, 1.9901E+07, 1.9945E+07,
     +  1.9988E+07, 2.0032E+07, 2.0075E+07, 2.0119E+07, 2.0163E+07,
     + 2.0207E+07, 2.0250E+07, 2.0294E+07, 2.0338E+07, 2.0382E+07,
     +  2.0426E+07, 2.0470E+07, 2.0515E+07, 2.0559E+07, 2.0603E+07,
     + 2.0648E+07, 2.0692E+07, 2.0737E+07, 2.0781E+07, 2.0826E+07,
     +  2.0871E+07, 2.0915E+07, 2.0960E+07, 2.1005E+07, 2.1050E+07,
     + 2.1095E+07, 2.1140E+07, 2.1185E+07, 2.1230E+07, 2.1276E+07,
     +  2.1321E+07, 2.1366E+07, 2.1412E+07, 2.1457E+07, 2.1503E+07,
     + 2.1548E+07, 2.1594E+07, 2.1640E+07, 2.1686E+07, 2.1732E+07,
     +  2.1777E+07, 2.1823E+07, 2.1870E+07, 2.1916E+07, 2.1962E+07,
     + 2.2008E+07, 2.2054E+07, 2.2101E+07, 2.2147E+07, 2.2194E+07,
     +  2.2240E+07, 2.2287E+07, 2.2334E+07, 2.2380E+07, 2.2427E+07,
     + 2.2474E+07, 2.2521E+07, 2.2568E+07, 2.2615E+07, 2.2662E+07,
     +  2.2709E+07, 2.2757E+07, 2.2804E+07, 2.2851E+07, 2.2899E+07,
     + 2.2946E+07, 2.2994E+07, 2.3041E+07, 2.3089E+07, 2.3137E+07,
     +  2.3185E+07, 2.3232E+07, 2.3280E+07, 2.3328E+07, 2.3376E+07,
     + 2.3425E+07, 2.3473E+07, 2.3521E+07, 2.3569E+07, 2.3618E+07,
     +  2.3666E+07, 2.3715E+07, 2.3763E+07, 2.3812E+07, 2.3860E+07,
     + 2.3909E+07, 2.3958E+07, 2.4007E+07, 2.4056E+07, 2.4105E+07,
     +  2.4154E+07, 2.4203E+07, 2.4252E+07, 2.4301E+07, 2.4351E+07,
     + 2.4400E+07, 2.4450E+07, 2.4499E+07, 2.4549E+07, 2.4598E+07,
     +  2.4648E+07, 2.4698E+07, 2.4748E+07, 2.4797E+07, 2.4847E+07,
     + 2.4897E+07, 2.4947E+07, 2.4998E+07, 2.5048E+07, 2.5098E+07,
     +  2.5148E+07, 2.5199E+07, 2.5249E+07, 2.5300E+07, 2.5350E+07,
     + 2.5401E+07, 2.5452E+07, 2.5502E+07, 2.5553E+07, 2.5604E+07,
     +  2.5655E+07, 2.5706E+07, 2.5757E+07, 2.5808E+07, 2.5860E+07,
     + 2.5911E+07, 2.5962E+07, 2.6014E+07, 2.6065E+07, 2.6117E+07,
     +  2.6168E+07, 2.6220E+07, 2.6272E+07, 2.6323E+07, 2.6375E+07,
     + 2.6427E+07, 2.6479E+07, 2.6531E+07, 2.6583E+07, 2.6635E+07,
     +  2.6688E+07, 2.6740E+07, 2.6792E+07, 2.6845E+07, 2.6897E+07,
     + 2.6950E+07, 2.7002E+07, 2.7055E+07, 2.7108E+07, 2.7161E+07,
     +  2.7213E+07, 2.7266E+07, 2.7319E+07, 2.7372E+07, 2.7425E+07,
     + 2.7479E+07, 2.7532E+07, 2.7585E+07, 2.7639E+07, 2.7692E+07,
     +  2.7746E+07, 2.7799E+07, 2.7853E+07, 2.7906E+07, 2.7960E+07,
     + 2.8014E+07, 2.8068E+07, 2.8122E+07, 2.8176E+07, 2.8230E+07,
     +  2.8284E+07, 2.8338E+07, 2.8392E+07, 2.8447E+07, 2.8501E+07,
     + 2.8556E+07, 2.8610E+07, 2.8665E+07, 2.8719E+07, 2.8774E+07,
     +  2.8829E+07, 2.8884E+07, 2.8939E+07, 2.8994E+07, 2.9049E+07,
     + 2.9104E+07, 2.9159E+07, 2.9214E+07, 2.9269E+07, 2.9325E+07,
     +  2.9380E+07, 2.9436E+07, 2.9491E+07, 2.9547E+07, 2.9603E+07,
     + 2.9658E+07, 2.9714E+07, 2.9770E+07, 2.9826E+07, 2.9882E+07,
     +  2.9938E+07, 2.9994E+07, 3.0050E+07, 3.0106E+07, 3.0163E+07,
     + 3.0219E+07, 3.0276E+07, 3.0332E+07, 3.0389E+07, 3.0445E+07,
     +  3.0502E+07, 3.0559E+07, 3.0616E+07, 3.0672E+07, 3.0729E+07,
     + 3.0786E+07, 3.0844E+07, 3.0901E+07, 3.0958E+07, 3.1015E+07,
     +  3.1073E+07, 3.1130E+07, 3.1187E+07, 3.1245E+07, 3.1303E+07,
     + 3.1360E+07, 3.1418E+07, 3.1476E+07, 3.1534E+07, 3.1591E+07,
     +  3.1649E+07, 3.1707E+07, 3.1766E+07, 3.1824E+07, 3.1882E+07,
     + 3.1940E+07, 3.1999E+07, 3.2057E+07, 3.2116E+07, 3.2174E+07,
     +  3.2233E+07, 3.2291E+07, 3.2350E+07, 3.2409E+07, 3.2468E+07,
     + 3.2527E+07, 3.2586E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_COF2  (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 2),QofT( 2,NT),Q(NT)
      data (Tmx(I),I=1,2)/3500.,3500./
      data xgj/ 1.,2./
c...     COF2
c...        --       269
      data (QofT( 1,J),J=1,2408)/ 1.3003E+01, 3.5011E+01, 6.3292E+01,
     +  9.6670E+01, 1.3446E+02, 1.7619E+02, 2.2153E+02, 2.7021E+02,
     + 3.2200E+02, 3.7674E+02, 4.3428E+02, 4.9448E+02, 5.5724E+02,
     +  6.2244E+02, 6.9001E+02, 7.5987E+02, 8.3194E+02, 9.0615E+02,
     + 9.8245E+02, 1.0608E+03, 1.2233E+03, 1.3934E+03, 1.5708E+03,
     +  1.7551E+03, 1.9461E+03, 2.1436E+03, 2.3474E+03, 2.5572E+03,
     + 2.7730E+03, 2.9945E+03, 3.2216E+03, 3.4542E+03, 3.6922E+03,
     +  3.9354E+03, 4.1838E+03, 4.4372E+03, 4.6955E+03, 4.9586E+03,
     + 5.2265E+03, 5.4991E+03, 5.7763E+03, 6.0580E+03, 6.3442E+03,
     +  6.6347E+03, 6.9296E+03, 7.2287E+03, 7.5321E+03, 7.8396E+03,
     + 8.1512E+03, 8.4669E+03, 8.7866E+03, 9.1103E+03, 9.4379E+03,
     +  9.7694E+03, 1.0105E+04, 1.0444E+04, 1.0787E+04, 1.1134E+04,
     + 1.1485E+04, 1.1839E+04, 1.2197E+04, 1.2559E+04, 1.2924E+04,
     +  1.3293E+04, 1.3666E+04, 1.4043E+04, 1.4423E+04, 1.4807E+04,
     + 1.5195E+04, 1.5586E+04, 1.5981E+04, 1.6380E+04, 1.6782E+04,
     +  1.7188E+04, 1.7598E+04, 1.8012E+04, 1.8430E+04, 1.8851E+04,
     + 1.9277E+04, 1.9706E+04, 2.0139E+04, 2.0576E+04, 2.1017E+04,
     +  2.1462E+04, 2.1911E+04, 2.2365E+04, 2.2822E+04, 2.3284E+04,
     + 2.3750E+04, 2.4220E+04, 2.4695E+04, 2.5174E+04, 2.5657E+04,
     +  2.6145E+04, 2.6637E+04, 2.7135E+04, 2.7636E+04, 2.8143E+04,
     + 2.8654E+04, 2.9170E+04, 2.9691E+04, 3.0217E+04, 3.0748E+04,
     +  3.1285E+04, 3.1826E+04, 3.2373E+04, 3.2925E+04, 3.3482E+04,
     + 3.4045E+04, 3.4613E+04, 3.5187E+04, 3.5766E+04, 3.6352E+04,
     +  3.6943E+04, 3.7540E+04, 3.8143E+04, 3.8752E+04, 3.9367E+04,
     + 3.9989E+04, 4.0617E+04, 4.1251E+04, 4.1891E+04, 4.2539E+04,
     +  4.3192E+04, 4.3853E+04, 4.4520E+04, 4.5195E+04, 4.5876E+04,
     + 4.6564E+04, 4.7260E+04, 4.7962E+04, 4.8672E+04, 4.9390E+04,
     +  5.0115E+04, 5.0847E+04, 5.1588E+04, 5.2336E+04, 5.3092E+04,
     + 5.3856E+04, 5.4628E+04, 5.5409E+04, 5.6197E+04, 5.6994E+04,
     +  5.7800E+04, 5.8614E+04, 5.9437E+04, 6.0269E+04, 6.1109E+04,
     + 6.1959E+04, 6.2817E+04, 6.3685E+04, 6.4563E+04, 6.5449E+04,
     +  6.6346E+04, 6.7252E+04, 6.8167E+04, 6.9093E+04, 7.0028E+04,
     + 7.0974E+04, 7.1930E+04, 7.2896E+04, 7.3873E+04, 7.4860E+04,
     +  7.5858E+04, 7.6867E+04, 7.7887E+04, 7.8918E+04, 7.9959E+04,
     + 8.1013E+04, 8.2077E+04, 8.3154E+04, 8.4241E+04, 8.5341E+04,
     +  8.6453E+04, 8.7576E+04, 8.8712E+04, 8.9860E+04, 9.1021E+04,
     + 9.2194E+04, 9.3380E+04, 9.4579E+04, 9.5790E+04, 9.7015E+04,
     +  9.8253E+04, 9.9505E+04, 1.0077E+05, 1.0205E+05, 1.0334E+05,
     + 1.0465E+05, 1.0597E+05, 1.0730E+05, 1.0865E+05, 1.1002E+05,
     +  1.1139E+05, 1.1279E+05, 1.1420E+05, 1.1562E+05, 1.1706E+05,
     + 1.1851E+05, 1.1998E+05, 1.2147E+05, 1.2297E+05, 1.2449E+05,
     +  1.2602E+05, 1.2757E+05, 1.2914E+05, 1.3072E+05, 1.3233E+05,
     + 1.3394E+05, 1.3558E+05, 1.3723E+05, 1.3890E+05, 1.4059E+05,
     +  1.4230E+05, 1.4402E+05, 1.4576E+05, 1.4752E+05, 1.4930E+05,
     + 1.5110E+05, 1.5292E+05, 1.5476E+05, 1.5661E+05, 1.5849E+05,
     +  1.6038E+05, 1.6230E+05, 1.6423E+05, 1.6619E+05, 1.6817E+05,
     + 1.7017E+05, 1.7218E+05, 1.7422E+05, 1.7628E+05, 1.7837E+05,
     +  1.8047E+05, 1.8260E+05, 1.8475E+05, 1.8692E+05, 1.8911E+05,
     + 1.9133E+05, 1.9357E+05, 1.9583E+05, 1.9811E+05, 2.0042E+05,
     +  2.0276E+05, 2.0511E+05, 2.0750E+05, 2.0990E+05, 2.1234E+05,
     + 2.1479E+05, 2.1727E+05, 2.1978E+05, 2.2231E+05, 2.2487E+05,
     +  2.2746E+05, 2.3007E+05, 2.3271E+05, 2.3537E+05, 2.3807E+05,
     + 2.4079E+05, 2.4353E+05, 2.4631E+05, 2.4911E+05, 2.5195E+05,
     +  2.5481E+05, 2.5770E+05, 2.6062E+05, 2.6356E+05, 2.6654E+05,
     + 2.6955E+05, 2.7259E+05, 2.7566E+05, 2.7876E+05, 2.8189E+05,
     +  2.8505E+05, 2.8825E+05, 2.9147E+05, 2.9473E+05, 2.9802E+05,
     + 3.0135E+05, 3.0470E+05, 3.0810E+05, 3.1152E+05, 3.1498E+05,
     +  3.1847E+05, 3.2200E+05, 3.2556E+05, 3.2916E+05, 3.3279E+05,
     + 3.3646E+05, 3.4016E+05, 3.4390E+05, 3.4768E+05, 3.5149E+05,
     +  3.5535E+05, 3.5924E+05, 3.6316E+05, 3.6713E+05, 3.7113E+05,
     + 3.7518E+05, 3.7926E+05, 3.8338E+05, 3.8755E+05, 3.9175E+05,
     +  3.9600E+05, 4.0028E+05, 4.0461E+05, 4.0898E+05, 4.1339E+05,
     + 4.1784E+05, 4.2234E+05, 4.2687E+05, 4.3146E+05, 4.3608E+05,
     +  4.4076E+05, 4.4547E+05, 4.5023E+05, 4.5504E+05, 4.5989E+05,
     + 4.6479E+05, 4.6973E+05, 4.7473E+05, 4.7976E+05, 4.8485E+05,
     +  4.8999E+05, 4.9517E+05, 5.0040E+05, 5.0569E+05, 5.1102E+05,
     + 5.1640E+05, 5.2183E+05, 5.2732E+05, 5.3285E+05, 5.3844E+05,
     +  5.4408E+05, 5.4977E+05, 5.5552E+05, 5.6132E+05, 5.6717E+05,
     + 5.7308E+05, 5.7904E+05, 5.8506E+05, 5.9113E+05, 5.9726E+05,
     +  6.0345E+05, 6.0969E+05, 6.1599E+05, 6.2235E+05, 6.2877E+05,
     + 6.3525E+05, 6.4178E+05, 6.4838E+05, 6.5504E+05, 6.6175E+05,
     +  6.6853E+05, 6.7538E+05, 6.8228E+05, 6.8925E+05, 6.9628E+05,
     + 7.0337E+05, 7.1053E+05, 7.1775E+05, 7.2504E+05, 7.3240E+05,
     +  7.3982E+05, 7.4731E+05, 7.5486E+05, 7.6249E+05, 7.7018E+05,
     + 7.7794E+05, 7.8577E+05, 7.9368E+05, 8.0165E+05, 8.0969E+05,
     +  8.1781E+05, 8.2600E+05, 8.3426E+05, 8.4259E+05, 8.5100E+05,
     + 8.5949E+05, 8.6805E+05, 8.7668E+05, 8.8539E+05, 8.9418E+05,
     +  9.0305E+05, 9.1199E+05, 9.2102E+05, 9.3012E+05, 9.3930E+05,
     + 9.4857E+05, 9.5791E+05, 9.6734E+05, 9.7685E+05, 9.8644E+05,
     +  9.9612E+05, 1.0059E+06, 1.0157E+06, 1.0257E+06, 1.0357E+06,
     + 1.0458E+06, 1.0560E+06, 1.0663E+06, 1.0766E+06, 1.0871E+06,
     +  1.0976E+06, 1.1083E+06, 1.1190E+06, 1.1298E+06, 1.1408E+06,
     + 1.1518E+06, 1.1629E+06, 1.1741E+06, 1.1854E+06, 1.1968E+06,
     +  1.2083E+06, 1.2198E+06, 1.2315E+06, 1.2433E+06, 1.2552E+06,
     + 1.2672E+06, 1.2793E+06, 1.2915E+06, 1.3038E+06, 1.3162E+06,
     +  1.3287E+06, 1.3413E+06, 1.3540E+06, 1.3668E+06, 1.3798E+06,
     + 1.3928E+06, 1.4060E+06, 1.4192E+06, 1.4326E+06, 1.4461E+06,
     +  1.4597E+06, 1.4734E+06, 1.4872E+06, 1.5011E+06, 1.5152E+06,
     + 1.5294E+06, 1.5437E+06, 1.5581E+06, 1.5726E+06, 1.5872E+06,
     +  1.6020E+06, 1.6169E+06, 1.6319E+06, 1.6470E+06, 1.6623E+06,
     + 1.6777E+06, 1.6932E+06, 1.7088E+06, 1.7246E+06, 1.7405E+06,
     +  1.7565E+06, 1.7727E+06, 1.7890E+06, 1.8054E+06, 1.8219E+06,
     + 1.8386E+06, 1.8555E+06, 1.8724E+06, 1.8895E+06, 1.9068E+06,
     +  1.9241E+06, 1.9416E+06, 1.9593E+06, 1.9771E+06, 1.9950E+06,
     + 2.0131E+06, 2.0314E+06, 2.0497E+06, 2.0683E+06, 2.0869E+06,
     +  2.1058E+06, 2.1247E+06, 2.1439E+06, 2.1631E+06, 2.1826E+06,
     + 2.2021E+06, 2.2219E+06, 2.2418E+06, 2.2618E+06, 2.2820E+06,
     +  2.3024E+06, 2.3229E+06, 2.3436E+06, 2.3645E+06, 2.3855E+06,
     + 2.4067E+06, 2.4281E+06, 2.4496E+06, 2.4713E+06, 2.4931E+06,
     +  2.5152E+06, 2.5374E+06, 2.5598E+06, 2.5823E+06, 2.6050E+06,
     + 2.6279E+06, 2.6510E+06, 2.6743E+06, 2.6977E+06, 2.7213E+06,
     +  2.7451E+06, 2.7691E+06, 2.7933E+06, 2.8177E+06, 2.8422E+06,
     + 2.8670E+06, 2.8919E+06, 2.9170E+06, 2.9423E+06, 2.9678E+06,
     +  2.9935E+06, 3.0194E+06, 3.0455E+06, 3.0718E+06, 3.0983E+06,
     + 3.1250E+06, 3.1519E+06, 3.1790E+06, 3.2063E+06, 3.2339E+06,
     +  3.2616E+06, 3.2895E+06, 3.3177E+06, 3.3460E+06, 3.3746E+06,
     + 3.4034E+06, 3.4324E+06, 3.4616E+06, 3.4911E+06, 3.5208E+06,
     +  3.5506E+06, 3.5808E+06, 3.6111E+06, 3.6417E+06, 3.6724E+06,
     + 3.7035E+06, 3.7347E+06, 3.7662E+06, 3.7979E+06, 3.8299E+06,
     +  3.8621E+06, 3.8945E+06, 3.9272E+06, 3.9601E+06, 3.9933E+06,
     + 4.0267E+06, 4.0603E+06, 4.0942E+06, 4.1284E+06, 4.1628E+06,
     +  4.1974E+06, 4.2323E+06, 4.2675E+06, 4.3029E+06, 4.3386E+06,
     + 4.3745E+06, 4.4107E+06, 4.4472E+06, 4.4839E+06, 4.5209E+06,
     +  4.5582E+06, 4.5957E+06, 4.6335E+06, 4.6716E+06, 4.7100E+06,
     + 4.7486E+06, 4.7875E+06, 4.8267E+06, 4.8662E+06, 4.9060E+06,
     +  4.9460E+06, 4.9864E+06, 5.0270E+06, 5.0679E+06, 5.1091E+06,
     + 5.1506E+06, 5.1924E+06, 5.2345E+06, 5.2770E+06, 5.3197E+06,
     +  5.3627E+06, 5.4060E+06, 5.4496E+06, 5.4936E+06, 5.5378E+06,
     + 5.5824E+06, 5.6273E+06, 5.6724E+06, 5.7180E+06, 5.7638E+06,
     +  5.8100E+06, 5.8564E+06, 5.9033E+06, 5.9504E+06, 5.9979E+06,
     + 6.0457E+06, 6.0938E+06, 6.1423E+06, 6.1911E+06, 6.2403E+06,
     +  6.2898E+06, 6.3397E+06, 6.3898E+06, 6.4404E+06, 6.4913E+06,
     + 6.5426E+06, 6.5942E+06, 6.6461E+06, 6.6985E+06, 6.7511E+06,
     +  6.8042E+06, 6.8576E+06, 6.9114E+06, 6.9656E+06, 7.0201E+06,
     + 7.0750E+06, 7.1303E+06, 7.1860E+06, 7.2420E+06, 7.2985E+06,
     +  7.3553E+06, 7.4125E+06, 7.4701E+06, 7.5281E+06, 7.5865E+06,
     + 7.6453E+06, 7.7045E+06, 7.7641E+06, 7.8241E+06, 7.8845E+06,
     +  7.9453E+06, 8.0065E+06, 8.0682E+06, 8.1302E+06, 8.1927E+06,
     + 8.2556E+06, 8.3189E+06, 8.3827E+06, 8.4469E+06, 8.5115E+06,
     +  8.5765E+06, 8.6420E+06, 8.7079E+06, 8.7743E+06, 8.8411E+06,
     + 8.9084E+06, 8.9761E+06, 9.0442E+06, 9.1128E+06, 9.1819E+06,
     +  9.2514E+06, 9.3214E+06, 9.3919E+06, 9.4628E+06, 9.5342E+06,
     + 9.6061E+06, 9.6784E+06, 9.7513E+06, 9.8246E+06, 9.8984E+06,
     +  9.9726E+06, 1.0047E+07, 1.0123E+07, 1.0198E+07, 1.0275E+07,
     + 1.0351E+07, 1.0429E+07, 1.0506E+07, 1.0585E+07, 1.0663E+07,
     +  1.0743E+07, 1.0823E+07, 1.0903E+07, 1.0984E+07, 1.1065E+07,
     + 1.1147E+07, 1.1230E+07, 1.1313E+07, 1.1396E+07, 1.1480E+07,
     +  1.1565E+07, 1.1650E+07, 1.1736E+07, 1.1822E+07, 1.1909E+07,
     + 1.1996E+07, 1.2084E+07, 1.2172E+07, 1.2261E+07, 1.2351E+07,
     +  1.2441E+07, 1.2532E+07, 1.2624E+07, 1.2715E+07, 1.2808E+07,
     + 1.2901E+07, 1.2995E+07, 1.3089E+07, 1.3184E+07, 1.3280E+07,
     +  1.3376E+07, 1.3473E+07, 1.3570E+07, 1.3668E+07, 1.3767E+07,
     + 1.3866E+07, 1.3966E+07, 1.4066E+07, 1.4167E+07, 1.4269E+07,
     +  1.4371E+07, 1.4474E+07, 1.4578E+07, 1.4682E+07, 1.4787E+07,
     + 1.4893E+07, 1.4999E+07, 1.5106E+07, 1.5214E+07, 1.5322E+07,
     +  1.5431E+07, 1.5541E+07, 1.5651E+07, 1.5763E+07, 1.5874E+07,
     + 1.5987E+07, 1.6100E+07, 1.6214E+07, 1.6328E+07, 1.6443E+07,
     +  1.6559E+07, 1.6676E+07, 1.6794E+07, 1.6912E+07, 1.7031E+07,
     + 1.7150E+07, 1.7270E+07, 1.7391E+07, 1.7513E+07, 1.7636E+07,
     +  1.7759E+07, 1.7883E+07, 1.8008E+07, 1.8134E+07, 1.8260E+07,
     + 1.8387E+07, 1.8515E+07, 1.8644E+07, 1.8773E+07, 1.8903E+07,
     +  1.9034E+07, 1.9166E+07, 1.9299E+07, 1.9432E+07, 1.9567E+07,
     + 1.9702E+07, 1.9838E+07, 1.9974E+07, 2.0112E+07, 2.0250E+07,
     +  2.0389E+07, 2.0529E+07, 2.0670E+07, 2.0812E+07, 2.0955E+07,
     + 2.1098E+07, 2.1242E+07, 2.1387E+07, 2.1533E+07, 2.1680E+07,
     +  2.1828E+07, 2.1977E+07, 2.2126E+07, 2.2277E+07, 2.2428E+07,
     + 2.2580E+07, 2.2733E+07, 2.2887E+07, 2.3042E+07, 2.3198E+07,
     +  2.3355E+07, 2.3513E+07, 2.3671E+07, 2.3831E+07, 2.3992E+07,
     + 2.4153E+07, 2.4315E+07, 2.4479E+07, 2.4643E+07, 2.4809E+07,
     +  2.4975E+07, 2.5142E+07, 2.5310E+07, 2.5480E+07, 2.5650E+07,
     + 2.5821E+07, 2.5993E+07, 2.6166E+07, 2.6341E+07, 2.6516E+07,
     +  2.6692E+07, 2.6869E+07, 2.7048E+07, 2.7227E+07, 2.7408E+07,
     + 2.7589E+07, 2.7772E+07, 2.7955E+07, 2.8140E+07, 2.8325E+07,
     +  2.8512E+07, 2.8700E+07, 2.8889E+07, 2.9079E+07, 2.9270E+07,
     + 2.9462E+07, 2.9656E+07, 2.9850E+07, 3.0045E+07, 3.0242E+07,
     +  3.0440E+07, 3.0639E+07, 3.0839E+07, 3.1040E+07, 3.1242E+07,
     + 3.1446E+07, 3.1650E+07, 3.1856E+07, 3.2063E+07, 3.2271E+07,
     +  3.2481E+07, 3.2691E+07, 3.2903E+07, 3.3116E+07, 3.3330E+07,
     + 3.3545E+07, 3.3762E+07, 3.3980E+07, 3.4199E+07, 3.4419E+07,
     +  3.4641E+07, 3.4863E+07, 3.5087E+07, 3.5313E+07, 3.5539E+07,
     + 3.5767E+07, 3.5996E+07, 3.6226E+07, 3.6458E+07, 3.6691E+07,
     +  3.6925E+07, 3.7161E+07, 3.7397E+07, 3.7636E+07, 3.7875E+07,
     + 3.8116E+07, 3.8358E+07, 3.8602E+07, 3.8846E+07, 3.9093E+07,
     +  3.9340E+07, 3.9589E+07, 3.9840E+07, 4.0091E+07, 4.0344E+07,
     + 4.0599E+07, 4.0855E+07, 4.1112E+07, 4.1371E+07, 4.1631E+07,
     +  4.1893E+07, 4.2156E+07, 4.2420E+07, 4.2686E+07, 4.2953E+07,
     + 4.3222E+07, 4.3493E+07, 4.3764E+07, 4.4038E+07, 4.4313E+07,
     +  4.4589E+07, 4.4867E+07, 4.5146E+07, 4.5427E+07, 4.5709E+07,
     + 4.5993E+07, 4.6278E+07, 4.6565E+07, 4.6854E+07, 4.7144E+07,
     +  4.7436E+07, 4.7729E+07, 4.8024E+07, 4.8320E+07, 4.8618E+07,
     + 4.8918E+07, 4.9219E+07, 4.9522E+07, 4.9826E+07, 5.0132E+07,
     +  5.0440E+07, 5.0749E+07, 5.1060E+07, 5.1373E+07, 5.1688E+07,
     + 5.2004E+07, 5.2321E+07, 5.2641E+07, 5.2962E+07, 5.3285E+07,
     +  5.3610E+07, 5.3936E+07, 5.4264E+07, 5.4594E+07, 5.4925E+07,
     + 5.5259E+07, 5.5594E+07, 5.5931E+07, 5.6269E+07, 5.6610E+07,
     +  5.6952E+07, 5.7296E+07, 5.7642E+07, 5.7990E+07, 5.8339E+07,
     + 5.8691E+07, 5.9044E+07, 5.9399E+07, 5.9756E+07, 6.0115E+07,
     +  6.0475E+07, 6.0838E+07, 6.1202E+07, 6.1569E+07, 6.1937E+07,
     + 6.2307E+07, 6.2680E+07, 6.3054E+07, 6.3430E+07, 6.3808E+07,
     +  6.4188E+07, 6.4570E+07, 6.4954E+07, 6.5340E+07, 6.5728E+07,
     + 6.6118E+07, 6.6510E+07, 6.6904E+07, 6.7300E+07, 6.7698E+07,
     +  6.8098E+07, 6.8500E+07, 6.8904E+07, 6.9311E+07, 6.9719E+07,
     + 7.0130E+07, 7.0543E+07, 7.0957E+07, 7.1374E+07, 7.1793E+07,
     +  7.2215E+07, 7.2638E+07, 7.3064E+07, 7.3491E+07, 7.3921E+07,
     + 7.4353E+07, 7.4788E+07, 7.5224E+07, 7.5663E+07, 7.6104E+07,
     +  7.6547E+07, 7.6993E+07, 7.7440E+07, 7.7891E+07, 7.8343E+07,
     + 7.8797E+07, 7.9254E+07, 7.9714E+07, 8.0175E+07, 8.0639E+07,
     +  8.1105E+07, 8.1574E+07, 8.2045E+07, 8.2518E+07, 8.2994E+07,
     + 8.3472E+07, 8.3953E+07, 8.4435E+07, 8.4921E+07, 8.5409E+07,
     +  8.5899E+07, 8.6392E+07, 8.6887E+07, 8.7384E+07, 8.7884E+07,
     + 8.8387E+07, 8.8892E+07, 8.9400E+07, 8.9910E+07, 9.0423E+07,
     +  9.0938E+07, 9.1456E+07, 9.1976E+07, 9.2499E+07, 9.3025E+07,
     + 9.3553E+07, 9.4084E+07, 9.4617E+07, 9.5154E+07, 9.5692E+07,
     +  9.6234E+07, 9.6778E+07, 9.7325E+07, 9.7874E+07, 9.8426E+07,
     + 9.8981E+07, 9.9539E+07, 1.0010E+08, 1.0066E+08, 1.0123E+08,
     +  1.0180E+08, 1.0237E+08, 1.0294E+08, 1.0352E+08, 1.0410E+08,
     + 1.0468E+08, 1.0527E+08, 1.0586E+08, 1.0645E+08, 1.0704E+08,
     +  1.0764E+08, 1.0824E+08, 1.0884E+08, 1.0945E+08, 1.1006E+08,
     + 1.1067E+08, 1.1128E+08, 1.1190E+08, 1.1252E+08, 1.1314E+08,
     +  1.1377E+08, 1.1440E+08, 1.1503E+08, 1.1567E+08, 1.1631E+08,
     + 1.1695E+08, 1.1760E+08, 1.1824E+08, 1.1889E+08, 1.1955E+08,
     +  1.2021E+08, 1.2087E+08, 1.2153E+08, 1.2220E+08, 1.2287E+08,
     + 1.2354E+08, 1.2422E+08, 1.2490E+08, 1.2558E+08, 1.2627E+08,
     +  1.2696E+08, 1.2765E+08, 1.2835E+08, 1.2905E+08, 1.2975E+08,
     + 1.3046E+08, 1.3117E+08, 1.3188E+08, 1.3260E+08, 1.3331E+08,
     +  1.3404E+08, 1.3476E+08, 1.3549E+08, 1.3623E+08, 1.3697E+08,
     + 1.3771E+08, 1.3845E+08, 1.3920E+08, 1.3995E+08, 1.4070E+08,
     +  1.4146E+08, 1.4222E+08, 1.4299E+08, 1.4376E+08, 1.4453E+08,
     + 1.4530E+08, 1.4608E+08, 1.4687E+08, 1.4765E+08, 1.4844E+08,
     +  1.4924E+08, 1.5004E+08, 1.5084E+08, 1.5164E+08, 1.5245E+08,
     + 1.5327E+08, 1.5408E+08, 1.5490E+08, 1.5573E+08, 1.5656E+08,
     +  1.5739E+08, 1.5822E+08, 1.5906E+08, 1.5991E+08, 1.6076E+08,
     + 1.6161E+08, 1.6246E+08, 1.6332E+08, 1.6419E+08, 1.6505E+08,
     +  1.6592E+08, 1.6680E+08, 1.6768E+08, 1.6856E+08, 1.6945E+08,
     + 1.7034E+08, 1.7124E+08, 1.7214E+08, 1.7304E+08, 1.7395E+08,
     +  1.7486E+08, 1.7578E+08, 1.7670E+08, 1.7762E+08, 1.7855E+08,
     + 1.7948E+08, 1.8042E+08, 1.8136E+08, 1.8231E+08, 1.8326E+08,
     +  1.8421E+08, 1.8517E+08, 1.8614E+08, 1.8710E+08, 1.8808E+08,
     + 1.8905E+08, 1.9003E+08, 1.9102E+08, 1.9201E+08, 1.9300E+08,
     +  1.9400E+08, 1.9501E+08, 1.9601E+08, 1.9703E+08, 1.9804E+08,
     + 1.9906E+08, 2.0009E+08, 2.0112E+08, 2.0216E+08, 2.0320E+08,
     +  2.0424E+08, 2.0529E+08, 2.0634E+08, 2.0740E+08, 2.0847E+08,
     + 2.0953E+08, 2.1061E+08, 2.1168E+08, 2.1277E+08, 2.1385E+08,
     +  2.1495E+08, 2.1604E+08, 2.1715E+08, 2.1825E+08, 2.1936E+08,
     + 2.2048E+08, 2.2160E+08, 2.2273E+08, 2.2386E+08, 2.2500E+08,
     +  2.2614E+08, 2.2729E+08, 2.2844E+08, 2.2960E+08, 2.3076E+08,
     + 2.3193E+08, 2.3310E+08, 2.3428E+08, 2.3546E+08, 2.3665E+08,
     +  2.3784E+08, 2.3904E+08, 2.4024E+08, 2.4145E+08, 2.4266E+08,
     + 2.4388E+08, 2.4511E+08, 2.4634E+08, 2.4757E+08, 2.4882E+08,
     +  2.5006E+08, 2.5131E+08, 2.5257E+08, 2.5383E+08, 2.5510E+08,
     + 2.5638E+08, 2.5766E+08, 2.5894E+08, 2.6023E+08, 2.6153E+08,
     +  2.6283E+08, 2.6414E+08, 2.6545E+08, 2.6677E+08, 2.6809E+08,
     + 2.6942E+08, 2.7076E+08, 2.7210E+08, 2.7345E+08, 2.7480E+08,
     +  2.7616E+08, 2.7753E+08, 2.7890E+08, 2.8028E+08, 2.8166E+08,
     + 2.8305E+08, 2.8444E+08, 2.8584E+08, 2.8725E+08, 2.8866E+08,
     +  2.9008E+08, 2.9151E+08, 2.9294E+08, 2.9437E+08, 2.9582E+08,
     + 2.9727E+08, 2.9872E+08, 3.0018E+08, 3.0165E+08, 3.0313E+08,
     +  3.0461E+08, 3.0609E+08, 3.0759E+08, 3.0909E+08, 3.1059E+08,
     + 3.1211E+08, 3.1362E+08, 3.1515E+08, 3.1668E+08, 3.1822E+08,
     +  3.1976E+08, 3.2132E+08, 3.2287E+08, 3.2444E+08, 3.2601E+08,
     + 3.2759E+08, 3.2917E+08, 3.3076E+08, 3.3236E+08, 3.3396E+08,
     +  3.3558E+08, 3.3719E+08, 3.3882E+08, 3.4045E+08, 3.4209E+08,
     + 3.4373E+08, 3.4539E+08, 3.4705E+08, 3.4871E+08, 3.5038E+08,
     +  3.5206E+08, 3.5375E+08, 3.5545E+08, 3.5715E+08, 3.5885E+08,
     + 3.6057E+08, 3.6229E+08, 3.6402E+08, 3.6576E+08, 3.6750E+08,
     +  3.6925E+08, 3.7101E+08, 3.7278E+08, 3.7455E+08, 3.7633E+08,
     + 3.7812E+08, 3.7992E+08, 3.8172E+08, 3.8353E+08, 3.8535E+08,
     +  3.8717E+08, 3.8901E+08, 3.9085E+08, 3.9269E+08, 3.9455E+08,
     + 3.9641E+08, 3.9828E+08, 4.0016E+08, 4.0205E+08, 4.0394E+08,
     +  4.0584E+08, 4.0775E+08, 4.0967E+08, 4.1160E+08, 4.1353E+08,
     + 4.1547E+08, 4.1742E+08, 4.1938E+08, 4.2134E+08, 4.2332E+08,
     +  4.2530E+08, 4.2729E+08, 4.2928E+08, 4.3129E+08, 4.3330E+08,
     + 4.3532E+08, 4.3735E+08, 4.3939E+08, 4.4144E+08, 4.4349E+08,
     +  4.4556E+08, 4.4763E+08, 4.4971E+08, 4.5180E+08, 4.5389E+08,
     + 4.5600E+08, 4.5811E+08, 4.6024E+08, 4.6237E+08, 4.6451E+08,
     +  4.6665E+08, 4.6881E+08, 4.7098E+08, 4.7315E+08, 4.7533E+08,
     + 4.7752E+08, 4.7973E+08, 4.8193E+08, 4.8415E+08, 4.8638E+08,
     +  4.8862E+08, 4.9086E+08, 4.9312E+08, 4.9538E+08, 4.9765E+08,
     + 4.9993E+08, 5.0222E+08, 5.0452E+08, 5.0683E+08, 5.0915E+08,
     +  5.1147E+08, 5.1381E+08, 5.1616E+08, 5.1851E+08, 5.2088E+08,
     + 5.2325E+08, 5.2563E+08, 5.2802E+08, 5.3043E+08, 5.3284E+08,
     +  5.3526E+08, 5.3769E+08, 5.4013E+08, 5.4258E+08, 5.4504E+08,
     + 5.4751E+08, 5.4999E+08, 5.5248E+08, 5.5498E+08, 5.5748E+08,
     +  5.6000E+08, 5.6253E+08, 5.6507E+08, 5.6762E+08, 5.7018E+08,
     + 5.7274E+08, 5.7532E+08, 5.7791E+08, 5.8051E+08, 5.8312E+08,
     +  5.8574E+08, 5.8837E+08, 5.9101E+08, 5.9366E+08, 5.9632E+08,
     + 5.9899E+08, 6.0167E+08, 6.0436E+08, 6.0706E+08, 6.0977E+08,
     +  6.1250E+08, 6.1523E+08, 6.1797E+08, 6.2073E+08, 6.2350E+08,
     + 6.2627E+08, 6.2906E+08, 6.3186E+08, 6.3467E+08, 6.3749E+08,
     +  6.4032E+08, 6.4316E+08, 6.4601E+08, 6.4887E+08, 6.5175E+08,
     + 6.5463E+08, 6.5753E+08, 6.6044E+08, 6.6336E+08, 6.6629E+08,
     +  6.6923E+08, 6.7218E+08, 6.7515E+08, 6.7812E+08, 6.8111E+08,
     + 6.8411E+08, 6.8712E+08, 6.9014E+08, 6.9317E+08, 6.9622E+08,
     +  6.9928E+08, 7.0234E+08, 7.0542E+08, 7.0852E+08, 7.1162E+08,
     + 7.1473E+08, 7.1786E+08, 7.2100E+08, 7.2415E+08, 7.2732E+08,
     +  7.3049E+08, 7.3368E+08, 7.3688E+08, 7.4009E+08, 7.4331E+08,
     + 7.4655E+08, 7.4980E+08, 7.5306E+08, 7.5633E+08, 7.5962E+08,
     +  7.6291E+08, 7.6622E+08, 7.6955E+08, 7.7288E+08, 7.7623E+08,
     + 7.7959E+08, 7.8296E+08, 7.8635E+08, 7.8975E+08, 7.9316E+08,
     +  7.9658E+08, 8.0002E+08, 8.0347E+08, 8.0693E+08, 8.1041E+08,
     + 8.1390E+08, 8.1740E+08, 8.2091E+08, 8.2444E+08, 8.2798E+08,
     +  8.3154E+08, 8.3511E+08, 8.3869E+08, 8.4228E+08, 8.4589E+08,
     + 8.4951E+08, 8.5315E+08, 8.5680E+08, 8.6046E+08, 8.6414E+08,
     +  8.6783E+08, 8.7153E+08, 8.7525E+08, 8.7898E+08, 8.8272E+08,
     + 8.8648E+08, 8.9025E+08, 8.9404E+08, 8.9784E+08, 9.0166E+08,
     +  9.0549E+08, 9.0933E+08, 9.1319E+08, 9.1706E+08, 9.2094E+08,
     + 9.2484E+08, 9.2876E+08, 9.3269E+08, 9.3663E+08, 9.4059E+08,
     +  9.4456E+08, 9.4855E+08, 9.5255E+08, 9.5657E+08, 9.6060E+08,
     + 9.6464E+08, 9.6870E+08, 9.7278E+08, 9.7687E+08, 9.8098E+08,
     +  9.8510E+08, 9.8923E+08, 9.9338E+08, 9.9755E+08, 1.0017E+09,
     + 1.0059E+09, 1.0101E+09, 1.0144E+09, 1.0186E+09, 1.0229E+09,
     +  1.0271E+09, 1.0314E+09, 1.0357E+09, 1.0401E+09, 1.0444E+09,
     + 1.0487E+09, 1.0531E+09, 1.0575E+09, 1.0619E+09, 1.0663E+09,
     +  1.0707E+09, 1.0752E+09, 1.0797E+09, 1.0841E+09, 1.0886E+09,
     + 1.0931E+09, 1.0977E+09, 1.1022E+09, 1.1068E+09, 1.1114E+09,
     +  1.1159E+09, 1.1206E+09, 1.1252E+09, 1.1298E+09, 1.1345E+09,
     + 1.1392E+09, 1.1439E+09, 1.1486E+09, 1.1533E+09, 1.1580E+09,
     +  1.1628E+09, 1.1676E+09, 1.1724E+09, 1.1772E+09, 1.1820E+09,
     + 1.1869E+09, 1.1917E+09, 1.1966E+09, 1.2015E+09, 1.2064E+09,
     +  1.2114E+09, 1.2163E+09, 1.2213E+09, 1.2263E+09, 1.2313E+09,
     + 1.2363E+09, 1.2413E+09, 1.2464E+09, 1.2515E+09, 1.2566E+09,
     +  1.2617E+09, 1.2668E+09, 1.2720E+09, 1.2771E+09, 1.2823E+09,
     + 1.2875E+09, 1.2927E+09, 1.2980E+09, 1.3032E+09, 1.3085E+09,
     +  1.3138E+09, 1.3191E+09, 1.3245E+09, 1.3298E+09, 1.3352E+09,
     + 1.3406E+09, 1.3460E+09, 1.3514E+09, 1.3569E+09, 1.3623E+09,
     +  1.3678E+09, 1.3733E+09, 1.3788E+09, 1.3844E+09, 1.3900E+09,
     + 1.3955E+09, 1.4011E+09, 1.4068E+09, 1.4124E+09, 1.4181E+09,
     +  1.4237E+09, 1.4294E+09, 1.4352E+09, 1.4409E+09, 1.4467E+09,
     + 1.4524E+09, 1.4582E+09, 1.4641E+09, 1.4699E+09, 1.4758E+09,
     +  1.4816E+09, 1.4876E+09, 1.4935E+09, 1.4994E+09, 1.5054E+09,
     + 1.5114E+09, 1.5174E+09, 1.5234E+09, 1.5294E+09, 1.5355E+09,
     +  1.5416E+09, 1.5477E+09, 1.5538E+09, 1.5600E+09, 1.5662E+09,
     + 1.5724E+09, 1.5786E+09, 1.5848E+09, 1.5911E+09, 1.5973E+09,
     +  1.6036E+09, 1.6100E+09, 1.6163E+09, 1.6227E+09, 1.6291E+09,
     + 1.6355E+09, 1.6419E+09, 1.6484E+09, 1.6548E+09, 1.6613E+09,
     +  1.6679E+09, 1.6744E+09, 1.6810E+09, 1.6876E+09, 1.6942E+09,
     + 1.7008E+09, 1.7074E+09, 1.7141E+09, 1.7208E+09, 1.7276E+09,
     +  1.7343E+09, 1.7411E+09, 1.7479E+09, 1.7547E+09, 1.7615E+09,
     + 1.7684E+09, 1.7753E+09, 1.7822E+09, 1.7891E+09, 1.7961E+09,
     +  1.8030E+09, 1.8100E+09, 1.8171E+09, 1.8241E+09, 1.8312E+09,
     + 1.8383E+09, 1.8454E+09, 1.8525E+09, 1.8597E+09, 1.8669E+09,
     +  1.8741E+09, 1.8814E+09, 1.8886E+09, 1.8959E+09, 1.9032E+09,
     + 1.9106E+09, 1.9180E+09, 1.9253E+09, 1.9328E+09, 1.9402E+09,
     +  1.9477E+09, 1.9551E+09, 1.9627E+09, 1.9702E+09, 1.9778E+09,
     + 1.9854E+09, 1.9930E+09, 2.0006E+09, 2.0083E+09, 2.0160E+09,
     +  2.0237E+09, 2.0314E+09, 2.0392E+09, 2.0470E+09, 2.0548E+09,
     + 2.0627E+09, 2.0705E+09, 2.0784E+09, 2.0864E+09, 2.0943E+09,
     +  2.1023E+09, 2.1103E+09, 2.1183E+09, 2.1264E+09, 2.1345E+09,
     + 2.1426E+09, 2.1507E+09, 2.1589E+09, 2.1671E+09, 2.1753E+09,
     +  2.1835E+09, 2.1918E+09, 2.2001E+09, 2.2084E+09, 2.2168E+09,
     + 2.2252E+09, 2.2336E+09, 2.2420E+09, 2.2505E+09, 2.2590E+09,
     +  2.2675E+09, 2.2760E+09, 2.2846E+09, 2.2932E+09, 2.3019E+09,
     + 2.3105E+09, 2.3192E+09, 2.3279E+09, 2.3367E+09, 2.3455E+09,
     +  2.3543E+09, 2.3631E+09, 2.3720E+09, 2.3809E+09, 2.3898E+09,
     + 2.3987E+09, 2.4077E+09, 2.4167E+09, 2.4257E+09, 2.4348E+09,
     +  2.4439E+09, 2.4530E+09, 2.4622E+09, 2.4714E+09, 2.4806E+09,
     + 2.4898E+09, 2.4991E+09, 2.5084E+09, 2.5178E+09, 2.5271E+09,
     +  2.5365E+09, 2.5459E+09, 2.5554E+09, 2.5649E+09, 2.5744E+09,
     + 2.5840E+09, 2.5935E+09, 2.6031E+09, 2.6128E+09, 2.6225E+09,
     +  2.6322E+09, 2.6419E+09, 2.6517E+09, 2.6615E+09, 2.6713E+09,
     + 2.6812E+09, 2.6910E+09, 2.7010E+09, 2.7109E+09, 2.7209E+09,
     +  2.7309E+09, 2.7410E+09, 2.7511E+09, 2.7612E+09, 2.7713E+09,
     + 2.7815E+09, 2.7917E+09, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       369
      data (QofT( 2,J),J=1,2408)/ 2.6007E+01, 7.0023E+01, 1.2658E+02,
     +  1.9334E+02, 2.6892E+02, 3.5239E+02, 4.4306E+02, 5.4041E+02,
     + 6.4401E+02, 7.5349E+02, 8.6856E+02, 9.8897E+02, 1.1145E+03,
     +  1.2449E+03, 1.3800E+03, 1.5197E+03, 1.6639E+03, 1.8123E+03,
     + 1.9649E+03, 2.1216E+03, 2.4467E+03, 2.7869E+03, 3.1416E+03,
     +  3.5102E+03, 3.8922E+03, 4.2872E+03, 4.6947E+03, 5.1144E+03,
     + 5.5459E+03, 5.9890E+03, 6.4432E+03, 6.9085E+03, 7.3844E+03,
     +  7.8709E+03, 8.3676E+03, 8.8743E+03, 9.3910E+03, 9.9173E+03,
     + 1.0453E+04, 1.0998E+04, 1.1553E+04, 1.2116E+04, 1.2688E+04,
     +  1.3269E+04, 1.3859E+04, 1.4457E+04, 1.5064E+04, 1.5679E+04,
     + 1.6302E+04, 1.6934E+04, 1.7573E+04, 1.8221E+04, 1.8876E+04,
     +  1.9539E+04, 2.0210E+04, 2.0888E+04, 2.1574E+04, 2.2268E+04,
     + 2.2969E+04, 2.3678E+04, 2.4394E+04, 2.5117E+04, 2.5848E+04,
     +  2.6587E+04, 2.7332E+04, 2.8086E+04, 2.8846E+04, 2.9614E+04,
     + 3.0389E+04, 3.1172E+04, 3.1962E+04, 3.2759E+04, 3.3564E+04,
     +  3.4377E+04, 3.5197E+04, 3.6025E+04, 3.6860E+04, 3.7703E+04,
     + 3.8553E+04, 3.9412E+04, 4.0278E+04, 4.1152E+04, 4.2034E+04,
     +  4.2925E+04, 4.3823E+04, 4.4730E+04, 4.5645E+04, 4.6568E+04,
     + 4.7500E+04, 4.8441E+04, 4.9390E+04, 5.0348E+04, 5.1315E+04,
     +  5.2291E+04, 5.3276E+04, 5.4270E+04, 5.5273E+04, 5.6286E+04,
     + 5.7309E+04, 5.8341E+04, 5.9384E+04, 6.0436E+04, 6.1498E+04,
     +  6.2570E+04, 6.3653E+04, 6.4746E+04, 6.5850E+04, 6.6965E+04,
     + 6.8091E+04, 6.9227E+04, 7.0375E+04, 7.1534E+04, 7.2705E+04,
     +  7.3888E+04, 7.5082E+04, 7.6288E+04, 7.7506E+04, 7.8737E+04,
     + 7.9980E+04, 8.1235E+04, 8.2504E+04, 8.3785E+04, 8.5080E+04,
     +  8.6387E+04, 8.7708E+04, 8.9043E+04, 9.0392E+04, 9.1754E+04,
     + 9.3131E+04, 9.4522E+04, 9.5928E+04, 9.7348E+04, 9.8783E+04,
     +  1.0023E+05, 1.0170E+05, 1.0318E+05, 1.0468E+05, 1.0619E+05,
     + 1.0772E+05, 1.0926E+05, 1.1082E+05, 1.1240E+05, 1.1399E+05,
     +  1.1560E+05, 1.1723E+05, 1.1888E+05, 1.2054E+05, 1.2222E+05,
     + 1.2392E+05, 1.2564E+05, 1.2738E+05, 1.2913E+05, 1.3090E+05,
     +  1.3270E+05, 1.3451E+05, 1.3634E+05, 1.3819E+05, 1.4006E+05,
     + 1.4195E+05, 1.4387E+05, 1.4580E+05, 1.4775E+05, 1.4973E+05,
     +  1.5172E+05, 1.5374E+05, 1.5578E+05, 1.5784E+05, 1.5993E+05,
     + 1.6203E+05, 1.6416E+05, 1.6632E+05, 1.6849E+05, 1.7069E+05,
     +  1.7291E+05, 1.7516E+05, 1.7743E+05, 1.7973E+05, 1.8205E+05,
     + 1.8440E+05, 1.8677E+05, 1.8917E+05, 1.9159E+05, 1.9404E+05,
     +  1.9652E+05, 1.9902E+05, 2.0155E+05, 2.0411E+05, 2.0669E+05,
     + 2.0931E+05, 2.1195E+05, 2.1462E+05, 2.1732E+05, 2.2004E+05,
     +  2.2280E+05, 2.2559E+05, 2.2841E+05, 2.3125E+05, 2.3413E+05,
     + 2.3704E+05, 2.3998E+05, 2.4295E+05, 2.4596E+05, 2.4899E+05,
     +  2.5206E+05, 2.5516E+05, 2.5830E+05, 2.6146E+05, 2.6467E+05,
     + 2.6790E+05, 2.7117E+05, 2.7448E+05, 2.7782E+05, 2.8120E+05,
     +  2.8461E+05, 2.8806E+05, 2.9154E+05, 2.9506E+05, 2.9862E+05,
     + 3.0222E+05, 3.0586E+05, 3.0953E+05, 3.1324E+05, 3.1699E+05,
     +  3.2079E+05, 3.2462E+05, 3.2849E+05, 3.3240E+05, 3.3636E+05,
     + 3.4035E+05, 3.4439E+05, 3.4847E+05, 3.5259E+05, 3.5675E+05,
     +  3.6096E+05, 3.6522E+05, 3.6951E+05, 3.7386E+05, 3.7824E+05,
     + 3.8268E+05, 3.8715E+05, 3.9168E+05, 3.9625E+05, 4.0087E+05,
     +  4.0554E+05, 4.1026E+05, 4.1502E+05, 4.1983E+05, 4.2470E+05,
     + 4.2961E+05, 4.3457E+05, 4.3959E+05, 4.4466E+05, 4.4977E+05,
     +  4.5494E+05, 4.6017E+05, 4.6545E+05, 4.7078E+05, 4.7616E+05,
     + 4.8160E+05, 4.8710E+05, 4.9265E+05, 4.9826E+05, 5.0392E+05,
     +  5.0965E+05, 5.1543E+05, 5.2126E+05, 5.2716E+05, 5.3312E+05,
     + 5.3914E+05, 5.4521E+05, 5.5135E+05, 5.5755E+05, 5.6382E+05,
     +  5.7014E+05, 5.7653E+05, 5.8299E+05, 5.8950E+05, 5.9609E+05,
     + 6.0273E+05, 6.0945E+05, 6.1623E+05, 6.2308E+05, 6.2999E+05,
     +  6.3698E+05, 6.4403E+05, 6.5116E+05, 6.5835E+05, 6.6562E+05,
     + 6.7295E+05, 6.8036E+05, 6.8785E+05, 6.9540E+05, 7.0303E+05,
     +  7.1074E+05, 7.1852E+05, 7.2637E+05, 7.3430E+05, 7.4231E+05,
     + 7.5040E+05, 7.5857E+05, 7.6682E+05, 7.7514E+05, 7.8355E+05,
     +  7.9204E+05, 8.0061E+05, 8.0926E+05, 8.1800E+05, 8.2682E+05,
     + 8.3573E+05, 8.4472E+05, 8.5380E+05, 8.6297E+05, 8.7222E+05,
     +  8.8157E+05, 8.9100E+05, 9.0052E+05, 9.1013E+05, 9.1984E+05,
     + 9.2963E+05, 9.3953E+05, 9.4951E+05, 9.5959E+05, 9.6976E+05,
     +  9.8003E+05, 9.9040E+05, 1.0009E+06, 1.0114E+06, 1.0221E+06,
     + 1.0329E+06, 1.0437E+06, 1.0547E+06, 1.0658E+06, 1.0769E+06,
     +  1.0882E+06, 1.0996E+06, 1.1111E+06, 1.1227E+06, 1.1344E+06,
     + 1.1462E+06, 1.1581E+06, 1.1702E+06, 1.1823E+06, 1.1946E+06,
     +  1.2070E+06, 1.2195E+06, 1.2321E+06, 1.2448E+06, 1.2576E+06,
     + 1.2706E+06, 1.2836E+06, 1.2968E+06, 1.3102E+06, 1.3236E+06,
     +  1.3371E+06, 1.3508E+06, 1.3646E+06, 1.3786E+06, 1.3926E+06,
     + 1.4068E+06, 1.4211E+06, 1.4356E+06, 1.4502E+06, 1.4649E+06,
     +  1.4797E+06, 1.4947E+06, 1.5098E+06, 1.5251E+06, 1.5405E+06,
     + 1.5560E+06, 1.5716E+06, 1.5874E+06, 1.6034E+06, 1.6195E+06,
     +  1.6357E+06, 1.6521E+06, 1.6686E+06, 1.6853E+06, 1.7021E+06,
     + 1.7191E+06, 1.7362E+06, 1.7535E+06, 1.7709E+06, 1.7885E+06,
     +  1.8062E+06, 1.8241E+06, 1.8421E+06, 1.8603E+06, 1.8787E+06,
     + 1.8972E+06, 1.9159E+06, 1.9348E+06, 1.9538E+06, 1.9730E+06,
     +  1.9924E+06, 2.0119E+06, 2.0316E+06, 2.0514E+06, 2.0715E+06,
     + 2.0917E+06, 2.1121E+06, 2.1326E+06, 2.1534E+06, 2.1743E+06,
     +  2.1954E+06, 2.2167E+06, 2.2381E+06, 2.2598E+06, 2.2816E+06,
     + 2.3037E+06, 2.3259E+06, 2.3483E+06, 2.3709E+06, 2.3937E+06,
     +  2.4167E+06, 2.4398E+06, 2.4632E+06, 2.4868E+06, 2.5106E+06,
     + 2.5345E+06, 2.5587E+06, 2.5831E+06, 2.6077E+06, 2.6325E+06,
     +  2.6575E+06, 2.6827E+06, 2.7082E+06, 2.7338E+06, 2.7597E+06,
     + 2.7858E+06, 2.8121E+06, 2.8386E+06, 2.8653E+06, 2.8923E+06,
     +  2.9195E+06, 2.9469E+06, 2.9746E+06, 3.0025E+06, 3.0306E+06,
     + 3.0589E+06, 3.0875E+06, 3.1163E+06, 3.1454E+06, 3.1746E+06,
     +  3.2042E+06, 3.2340E+06, 3.2640E+06, 3.2943E+06, 3.3248E+06,
     + 3.3556E+06, 3.3866E+06, 3.4179E+06, 3.4494E+06, 3.4812E+06,
     +  3.5132E+06, 3.5455E+06, 3.5781E+06, 3.6110E+06, 3.6441E+06,
     + 3.6775E+06, 3.7111E+06, 3.7450E+06, 3.7792E+06, 3.8137E+06,
     +  3.8484E+06, 3.8835E+06, 3.9188E+06, 3.9544E+06, 3.9903E+06,
     + 4.0265E+06, 4.0629E+06, 4.0997E+06, 4.1367E+06, 4.1741E+06,
     +  4.2117E+06, 4.2497E+06, 4.2879E+06, 4.3265E+06, 4.3653E+06,
     + 4.4045E+06, 4.4440E+06, 4.4838E+06, 4.5239E+06, 4.5643E+06,
     +  4.6051E+06, 4.6461E+06, 4.6875E+06, 4.7292E+06, 4.7713E+06,
     + 4.8137E+06, 4.8564E+06, 4.8994E+06, 4.9428E+06, 4.9865E+06,
     +  5.0306E+06, 5.0750E+06, 5.1198E+06, 5.1649E+06, 5.2103E+06,
     + 5.2561E+06, 5.3023E+06, 5.3488E+06, 5.3957E+06, 5.4429E+06,
     +  5.4906E+06, 5.5385E+06, 5.5869E+06, 5.6356E+06, 5.6847E+06,
     + 5.7342E+06, 5.7841E+06, 5.8343E+06, 5.8849E+06, 5.9359E+06,
     +  5.9874E+06, 6.0392E+06, 6.0913E+06, 6.1439E+06, 6.1969E+06,
     + 6.2503E+06, 6.3041E+06, 6.3584E+06, 6.4130E+06, 6.4680E+06,
     +  6.5235E+06, 6.5794E+06, 6.6357E+06, 6.6924E+06, 6.7495E+06,
     + 6.8071E+06, 6.8651E+06, 6.9236E+06, 6.9825E+06, 7.0418E+06,
     +  7.1016E+06, 7.1618E+06, 7.2225E+06, 7.2836E+06, 7.3452E+06,
     + 7.4073E+06, 7.4698E+06, 7.5328E+06, 7.5962E+06, 7.6601E+06,
     +  7.7245E+06, 7.7894E+06, 7.8547E+06, 7.9206E+06, 7.9869E+06,
     + 8.0537E+06, 8.1210E+06, 8.1888E+06, 8.2571E+06, 8.3259E+06,
     +  8.3952E+06, 8.4650E+06, 8.5353E+06, 8.6062E+06, 8.6775E+06,
     + 8.7494E+06, 8.8218E+06, 8.8948E+06, 8.9682E+06, 9.0422E+06,
     +  9.1168E+06, 9.1918E+06, 9.2675E+06, 9.3436E+06, 9.4203E+06,
     + 9.4976E+06, 9.5755E+06, 9.6539E+06, 9.7328E+06, 9.8124E+06,
     +  9.8925E+06, 9.9731E+06, 1.0054E+07, 1.0136E+07, 1.0219E+07,
     + 1.0302E+07, 1.0385E+07, 1.0470E+07, 1.0554E+07, 1.0640E+07,
     +  1.0726E+07, 1.0812E+07, 1.0900E+07, 1.0988E+07, 1.1076E+07,
     + 1.1165E+07, 1.1255E+07, 1.1345E+07, 1.1436E+07, 1.1528E+07,
     +  1.1620E+07, 1.1713E+07, 1.1807E+07, 1.1901E+07, 1.1996E+07,
     + 1.2092E+07, 1.2188E+07, 1.2285E+07, 1.2383E+07, 1.2481E+07,
     +  1.2580E+07, 1.2680E+07, 1.2780E+07, 1.2881E+07, 1.2983E+07,
     + 1.3086E+07, 1.3189E+07, 1.3293E+07, 1.3397E+07, 1.3503E+07,
     +  1.3609E+07, 1.3716E+07, 1.3823E+07, 1.3932E+07, 1.4041E+07,
     + 1.4151E+07, 1.4261E+07, 1.4373E+07, 1.4485E+07, 1.4598E+07,
     +  1.4711E+07, 1.4826E+07, 1.4941E+07, 1.5057E+07, 1.5174E+07,
     + 1.5291E+07, 1.5410E+07, 1.5529E+07, 1.5649E+07, 1.5770E+07,
     +  1.5891E+07, 1.6014E+07, 1.6137E+07, 1.6261E+07, 1.6386E+07,
     + 1.6512E+07, 1.6638E+07, 1.6766E+07, 1.6894E+07, 1.7024E+07,
     +  1.7154E+07, 1.7285E+07, 1.7417E+07, 1.7549E+07, 1.7683E+07,
     + 1.7817E+07, 1.7953E+07, 1.8089E+07, 1.8226E+07, 1.8365E+07,
     +  1.8504E+07, 1.8644E+07, 1.8784E+07, 1.8926E+07, 1.9069E+07,
     + 1.9213E+07, 1.9358E+07, 1.9503E+07, 1.9650E+07, 1.9797E+07,
     +  1.9946E+07, 2.0096E+07, 2.0246E+07, 2.0398E+07, 2.0550E+07,
     + 2.0704E+07, 2.0858E+07, 2.1014E+07, 2.1170E+07, 2.1328E+07,
     +  2.1486E+07, 2.1646E+07, 2.1807E+07, 2.1968E+07, 2.2131E+07,
     + 2.2295E+07, 2.2460E+07, 2.2626E+07, 2.2793E+07, 2.2961E+07,
     +  2.3130E+07, 2.3301E+07, 2.3472E+07, 2.3644E+07, 2.3818E+07,
     + 2.3993E+07, 2.4169E+07, 2.4346E+07, 2.4524E+07, 2.4703E+07,
     +  2.4884E+07, 2.5065E+07, 2.5248E+07, 2.5432E+07, 2.5617E+07,
     + 2.5803E+07, 2.5991E+07, 2.6179E+07, 2.6369E+07, 2.6560E+07,
     +  2.6753E+07, 2.6946E+07, 2.7141E+07, 2.7337E+07, 2.7534E+07,
     + 2.7733E+07, 2.7932E+07, 2.8133E+07, 2.8335E+07, 2.8539E+07,
     +  2.8744E+07, 2.8950E+07, 2.9157E+07, 2.9366E+07, 2.9576E+07,
     + 2.9787E+07, 3.0000E+07, 3.0214E+07, 3.0429E+07, 3.0646E+07,
     +  3.0864E+07, 3.1083E+07, 3.1304E+07, 3.1526E+07, 3.1750E+07,
     + 3.1974E+07, 3.2201E+07, 3.2428E+07, 3.2658E+07, 3.2888E+07,
     +  3.3120E+07, 3.3353E+07, 3.3588E+07, 3.3824E+07, 3.4062E+07,
     + 3.4301E+07, 3.4542E+07, 3.4784E+07, 3.5028E+07, 3.5273E+07,
     +  3.5520E+07, 3.5768E+07, 3.6017E+07, 3.6268E+07, 3.6521E+07,
     + 3.6775E+07, 3.7031E+07, 3.7289E+07, 3.7548E+07, 3.7808E+07,
     +  3.8070E+07, 3.8334E+07, 3.8599E+07, 3.8866E+07, 3.9135E+07,
     + 3.9405E+07, 3.9676E+07, 3.9950E+07, 4.0225E+07, 4.0502E+07,
     +  4.0780E+07, 4.1060E+07, 4.1342E+07, 4.1625E+07, 4.1910E+07,
     + 4.2197E+07, 4.2486E+07, 4.2776E+07, 4.3068E+07, 4.3362E+07,
     +  4.3658E+07, 4.3955E+07, 4.4254E+07, 4.4555E+07, 4.4857E+07,
     + 4.5162E+07, 4.5468E+07, 4.5776E+07, 4.6086E+07, 4.6398E+07,
     +  4.6712E+07, 4.7027E+07, 4.7344E+07, 4.7663E+07, 4.7985E+07,
     + 4.8308E+07, 4.8632E+07, 4.8959E+07, 4.9288E+07, 4.9619E+07,
     +  4.9951E+07, 5.0286E+07, 5.0622E+07, 5.0961E+07, 5.1301E+07,
     + 5.1644E+07, 5.1988E+07, 5.2334E+07, 5.2683E+07, 5.3033E+07,
     +  5.3386E+07, 5.3741E+07, 5.4097E+07, 5.4456E+07, 5.4817E+07,
     + 5.5180E+07, 5.5545E+07, 5.5912E+07, 5.6281E+07, 5.6652E+07,
     +  5.7026E+07, 5.7402E+07, 5.7779E+07, 5.8159E+07, 5.8542E+07,
     + 5.8926E+07, 5.9313E+07, 5.9702E+07, 6.0093E+07, 6.0486E+07,
     +  6.0881E+07, 6.1279E+07, 6.1679E+07, 6.2082E+07, 6.2486E+07,
     + 6.2893E+07, 6.3303E+07, 6.3714E+07, 6.4128E+07, 6.4545E+07,
     +  6.4963E+07, 6.5384E+07, 6.5808E+07, 6.6234E+07, 6.6662E+07,
     + 6.7093E+07, 6.7526E+07, 6.7962E+07, 6.8400E+07, 6.8840E+07,
     +  6.9283E+07, 6.9729E+07, 7.0176E+07, 7.0627E+07, 7.1080E+07,
     + 7.1536E+07, 7.1994E+07, 7.2454E+07, 7.2918E+07, 7.3384E+07,
     +  7.3852E+07, 7.4323E+07, 7.4797E+07, 7.5273E+07, 7.5752E+07,
     + 7.6234E+07, 7.6718E+07, 7.7205E+07, 7.7695E+07, 7.8187E+07,
     +  7.8683E+07, 7.9181E+07, 7.9681E+07, 8.0185E+07, 8.0691E+07,
     + 8.1200E+07, 8.1712E+07, 8.2226E+07, 8.2744E+07, 8.3264E+07,
     +  8.3788E+07, 8.4314E+07, 8.4843E+07, 8.5374E+07, 8.5909E+07,
     + 8.6447E+07, 8.6988E+07, 8.7531E+07, 8.8078E+07, 8.8627E+07,
     +  8.9180E+07, 8.9735E+07, 9.0294E+07, 9.0856E+07, 9.1420E+07,
     + 9.1988E+07, 9.2559E+07, 9.3133E+07, 9.3710E+07, 9.4290E+07,
     +  9.4873E+07, 9.5460E+07, 9.6050E+07, 9.6642E+07, 9.7238E+07,
     + 9.7838E+07, 9.8440E+07, 9.9046E+07, 9.9655E+07, 1.0027E+08,
     +  1.0088E+08, 1.0150E+08, 1.0212E+08, 1.0275E+08, 1.0338E+08,
     + 1.0401E+08, 1.0465E+08, 1.0528E+08, 1.0593E+08, 1.0657E+08,
     +  1.0722E+08, 1.0787E+08, 1.0853E+08, 1.0919E+08, 1.0985E+08,
     + 1.1052E+08, 1.1119E+08, 1.1186E+08, 1.1254E+08, 1.1322E+08,
     +  1.1391E+08, 1.1460E+08, 1.1529E+08, 1.1598E+08, 1.1668E+08,
     + 1.1738E+08, 1.1809E+08, 1.1880E+08, 1.1951E+08, 1.2023E+08,
     +  1.2095E+08, 1.2168E+08, 1.2241E+08, 1.2314E+08, 1.2388E+08,
     + 1.2462E+08, 1.2536E+08, 1.2611E+08, 1.2686E+08, 1.2762E+08,
     +  1.2838E+08, 1.2914E+08, 1.2991E+08, 1.3068E+08, 1.3146E+08,
     + 1.3224E+08, 1.3302E+08, 1.3381E+08, 1.3460E+08, 1.3540E+08,
     +  1.3620E+08, 1.3700E+08, 1.3781E+08, 1.3862E+08, 1.3944E+08,
     + 1.4026E+08, 1.4109E+08, 1.4192E+08, 1.4275E+08, 1.4359E+08,
     +  1.4443E+08, 1.4528E+08, 1.4613E+08, 1.4699E+08, 1.4785E+08,
     + 1.4871E+08, 1.4958E+08, 1.5045E+08, 1.5133E+08, 1.5221E+08,
     +  1.5310E+08, 1.5399E+08, 1.5488E+08, 1.5578E+08, 1.5669E+08,
     + 1.5760E+08, 1.5851E+08, 1.5943E+08, 1.6035E+08, 1.6128E+08,
     +  1.6221E+08, 1.6315E+08, 1.6409E+08, 1.6504E+08, 1.6599E+08,
     + 1.6695E+08, 1.6791E+08, 1.6887E+08, 1.6985E+08, 1.7082E+08,
     +  1.7180E+08, 1.7279E+08, 1.7378E+08, 1.7477E+08, 1.7577E+08,
     + 1.7678E+08, 1.7779E+08, 1.7880E+08, 1.7982E+08, 1.8085E+08,
     +  1.8188E+08, 1.8292E+08, 1.8396E+08, 1.8500E+08, 1.8605E+08,
     + 1.8711E+08, 1.8817E+08, 1.8924E+08, 1.9031E+08, 1.9139E+08,
     +  1.9247E+08, 1.9356E+08, 1.9465E+08, 1.9575E+08, 1.9686E+08,
     + 1.9797E+08, 1.9908E+08, 2.0020E+08, 2.0133E+08, 2.0246E+08,
     +  2.0360E+08, 2.0474E+08, 2.0589E+08, 2.0704E+08, 2.0820E+08,
     + 2.0937E+08, 2.1054E+08, 2.1172E+08, 2.1290E+08, 2.1409E+08,
     +  2.1528E+08, 2.1648E+08, 2.1769E+08, 2.1890E+08, 2.2012E+08,
     + 2.2134E+08, 2.2257E+08, 2.2380E+08, 2.2505E+08, 2.2629E+08,
     +  2.2755E+08, 2.2881E+08, 2.3007E+08, 2.3134E+08, 2.3262E+08,
     + 2.3391E+08, 2.3520E+08, 2.3649E+08, 2.3779E+08, 2.3910E+08,
     +  2.4042E+08, 2.4174E+08, 2.4307E+08, 2.4440E+08, 2.4574E+08,
     + 2.4709E+08, 2.4844E+08, 2.4980E+08, 2.5117E+08, 2.5254E+08,
     +  2.5392E+08, 2.5531E+08, 2.5670E+08, 2.5810E+08, 2.5950E+08,
     + 2.6092E+08, 2.6234E+08, 2.6376E+08, 2.6520E+08, 2.6663E+08,
     +  2.6808E+08, 2.6953E+08, 2.7099E+08, 2.7246E+08, 2.7394E+08,
     + 2.7542E+08, 2.7690E+08, 2.7840E+08, 2.7990E+08, 2.8141E+08,
     +  2.8293E+08, 2.8445E+08, 2.8598E+08, 2.8752E+08, 2.8906E+08,
     + 2.9061E+08, 2.9217E+08, 2.9374E+08, 2.9531E+08, 2.9690E+08,
     +  2.9848E+08, 3.0008E+08, 3.0168E+08, 3.0329E+08, 3.0491E+08,
     + 3.0654E+08, 3.0817E+08, 3.0981E+08, 3.1146E+08, 3.1312E+08,
     +  3.1478E+08, 3.1645E+08, 3.1813E+08, 3.1982E+08, 3.2152E+08,
     + 3.2322E+08, 3.2493E+08, 3.2665E+08, 3.2838E+08, 3.3011E+08,
     +  3.3185E+08, 3.3360E+08, 3.3536E+08, 3.3713E+08, 3.3891E+08,
     + 3.4069E+08, 3.4248E+08, 3.4428E+08, 3.4609E+08, 3.4790E+08,
     +  3.4973E+08, 3.5156E+08, 3.5340E+08, 3.5525E+08, 3.5711E+08,
     + 3.5897E+08, 3.6085E+08, 3.6273E+08, 3.6463E+08, 3.6653E+08,
     +  3.6844E+08, 3.7035E+08, 3.7228E+08, 3.7422E+08, 3.7616E+08,
     + 3.7811E+08, 3.8007E+08, 3.8204E+08, 3.8402E+08, 3.8601E+08,
     +  3.8801E+08, 3.9002E+08, 3.9203E+08, 3.9406E+08, 3.9609E+08,
     + 3.9813E+08, 4.0019E+08, 4.0225E+08, 4.0432E+08, 4.0640E+08,
     +  4.0849E+08, 4.1059E+08, 4.1269E+08, 4.1481E+08, 4.1694E+08,
     + 4.1907E+08, 4.2122E+08, 4.2338E+08, 4.2554E+08, 4.2772E+08,
     +  4.2990E+08, 4.3210E+08, 4.3430E+08, 4.3651E+08, 4.3874E+08,
     + 4.4097E+08, 4.4321E+08, 4.4547E+08, 4.4773E+08, 4.5000E+08,
     +  4.5229E+08, 4.5458E+08, 4.5689E+08, 4.5920E+08, 4.6152E+08,
     + 4.6386E+08, 4.6620E+08, 4.6856E+08, 4.7092E+08, 4.7330E+08,
     +  4.7569E+08, 4.7808E+08, 4.8049E+08, 4.8291E+08, 4.8534E+08,
     + 4.8777E+08, 4.9022E+08, 4.9269E+08, 4.9516E+08, 4.9764E+08,
     +  5.0013E+08, 5.0264E+08, 5.0515E+08, 5.0768E+08, 5.1021E+08,
     + 5.1276E+08, 5.1532E+08, 5.1789E+08, 5.2047E+08, 5.2306E+08,
     +  5.2567E+08, 5.2828E+08, 5.3091E+08, 5.3355E+08, 5.3620E+08,
     + 5.3886E+08, 5.4153E+08, 5.4421E+08, 5.4691E+08, 5.4961E+08,
     +  5.5233E+08, 5.5506E+08, 5.5780E+08, 5.6056E+08, 5.6332E+08,
     + 5.6610E+08, 5.6889E+08, 5.7169E+08, 5.7451E+08, 5.7733E+08,
     +  5.8017E+08, 5.8302E+08, 5.8588E+08, 5.8876E+08, 5.9164E+08,
     + 5.9454E+08, 5.9745E+08, 6.0038E+08, 6.0331E+08, 6.0626E+08,
     +  6.0922E+08, 6.1220E+08, 6.1518E+08, 6.1818E+08, 6.2120E+08,
     + 6.2422E+08, 6.2726E+08, 6.3031E+08, 6.3337E+08, 6.3645E+08,
     +  6.3954E+08, 6.4264E+08, 6.4576E+08, 6.4889E+08, 6.5203E+08,
     + 6.5518E+08, 6.5835E+08, 6.6153E+08, 6.6473E+08, 6.6794E+08,
     +  6.7116E+08, 6.7440E+08, 6.7765E+08, 6.8091E+08, 6.8419E+08,
     + 6.8748E+08, 6.9078E+08, 6.9410E+08, 6.9743E+08, 7.0078E+08,
     +  7.0414E+08, 7.0751E+08, 7.1090E+08, 7.1430E+08, 7.1772E+08,
     + 7.2115E+08, 7.2460E+08, 7.2805E+08, 7.3153E+08, 7.3502E+08,
     +  7.3852E+08, 7.4204E+08, 7.4557E+08, 7.4912E+08, 7.5268E+08,
     + 7.5625E+08, 7.5984E+08, 7.6345E+08, 7.6707E+08, 7.7071E+08,
     +  7.7436E+08, 7.7802E+08, 7.8170E+08, 7.8540E+08, 7.8911E+08,
     + 7.9284E+08, 7.9658E+08, 8.0034E+08, 8.0411E+08, 8.0790E+08,
     +  8.1170E+08, 8.1552E+08, 8.1935E+08, 8.2321E+08, 8.2707E+08,
     + 8.3095E+08, 8.3485E+08, 8.3877E+08, 8.4270E+08, 8.4664E+08,
     +  8.5060E+08, 8.5458E+08, 8.5858E+08, 8.6259E+08, 8.6662E+08,
     + 8.7066E+08, 8.7472E+08, 8.7880E+08, 8.8289E+08, 8.8700E+08,
     +  8.9113E+08, 8.9527E+08, 8.9943E+08, 9.0361E+08, 9.0780E+08,
     + 9.1201E+08, 9.1624E+08, 9.2048E+08, 9.2474E+08, 9.2902E+08,
     +  9.3332E+08, 9.3763E+08, 9.4196E+08, 9.4631E+08, 9.5068E+08,
     + 9.5506E+08, 9.5946E+08, 9.6388E+08, 9.6832E+08, 9.7277E+08,
     +  9.7724E+08, 9.8173E+08, 9.8624E+08, 9.9077E+08, 9.9531E+08,
     + 9.9987E+08, 1.0045E+09, 1.0091E+09, 1.0137E+09, 1.0183E+09,
     +  1.0230E+09, 1.0276E+09, 1.0323E+09, 1.0370E+09, 1.0418E+09,
     + 1.0465E+09, 1.0513E+09, 1.0561E+09, 1.0609E+09, 1.0657E+09,
     +  1.0705E+09, 1.0754E+09, 1.0803E+09, 1.0852E+09, 1.0901E+09,
     + 1.0950E+09, 1.1000E+09, 1.1050E+09, 1.1100E+09, 1.1150E+09,
     +  1.1200E+09, 1.1251E+09, 1.1302E+09, 1.1352E+09, 1.1404E+09,
     + 1.1455E+09, 1.1507E+09, 1.1558E+09, 1.1610E+09, 1.1662E+09,
     +  1.1715E+09, 1.1767E+09, 1.1820E+09, 1.1873E+09, 1.1926E+09,
     + 1.1980E+09, 1.2033E+09, 1.2087E+09, 1.2141E+09, 1.2196E+09,
     +  1.2250E+09, 1.2305E+09, 1.2360E+09, 1.2415E+09, 1.2470E+09,
     + 1.2526E+09, 1.2581E+09, 1.2637E+09, 1.2693E+09, 1.2750E+09,
     +  1.2806E+09, 1.2863E+09, 1.2920E+09, 1.2978E+09, 1.3035E+09,
     + 1.3093E+09, 1.3151E+09, 1.3209E+09, 1.3267E+09, 1.3326E+09,
     +  1.3385E+09, 1.3444E+09, 1.3503E+09, 1.3563E+09, 1.3622E+09,
     + 1.3682E+09, 1.3743E+09, 1.3803E+09, 1.3864E+09, 1.3925E+09,
     +  1.3986E+09, 1.4047E+09, 1.4109E+09, 1.4170E+09, 1.4233E+09,
     + 1.4295E+09, 1.4357E+09, 1.4420E+09, 1.4483E+09, 1.4546E+09,
     +  1.4610E+09, 1.4674E+09, 1.4738E+09, 1.4802E+09, 1.4866E+09,
     + 1.4931E+09, 1.4996E+09, 1.5061E+09, 1.5127E+09, 1.5192E+09,
     +  1.5258E+09, 1.5325E+09, 1.5391E+09, 1.5458E+09, 1.5525E+09,
     + 1.5592E+09, 1.5659E+09, 1.5727E+09, 1.5795E+09, 1.5863E+09,
     +  1.5932E+09, 1.6001E+09, 1.6070E+09, 1.6139E+09, 1.6208E+09,
     + 1.6278E+09, 1.6348E+09, 1.6418E+09, 1.6489E+09, 1.6560E+09,
     +  1.6631E+09, 1.6702E+09, 1.6774E+09, 1.6846E+09, 1.6918E+09,
     + 1.6990E+09, 1.7063E+09, 1.7136E+09, 1.7209E+09, 1.7283E+09,
     +  1.7357E+09, 1.7431E+09, 1.7505E+09, 1.7580E+09, 1.7655E+09,
     + 1.7730E+09, 1.7805E+09, 1.7881E+09, 1.7957E+09, 1.8033E+09,
     +  1.8110E+09, 1.8187E+09, 1.8264E+09, 1.8341E+09, 1.8419E+09,
     + 1.8497E+09, 1.8575E+09, 1.8654E+09, 1.8733E+09, 1.8812E+09,
     +  1.8891E+09, 1.8971E+09, 1.9051E+09, 1.9131E+09, 1.9212E+09,
     + 1.9293E+09, 1.9374E+09, 1.9456E+09, 1.9538E+09, 1.9620E+09,
     +  1.9702E+09, 1.9785E+09, 1.9868E+09, 1.9951E+09, 2.0035E+09,
     + 2.0119E+09, 2.0203E+09, 2.0288E+09, 2.0372E+09, 2.0458E+09,
     +  2.0543E+09, 2.0629E+09, 2.0715E+09, 2.0801E+09, 2.0888E+09,
     + 2.0975E+09, 2.1062E+09, 2.1150E+09, 2.1238E+09, 2.1326E+09,
     +  2.1415E+09, 2.1504E+09, 2.1593E+09, 2.1683E+09, 2.1773E+09,
     + 2.1863E+09, 2.1954E+09, 2.2044E+09, 2.2136E+09, 2.2227E+09,
     +  2.2319E+09, 2.2411E+09, 2.2504E+09, 2.2597E+09, 2.2690E+09,
     + 2.2783E+09, 2.2877E+09, 2.2972E+09, 2.3066E+09, 2.3161E+09,
     +  2.3256E+09, 2.3352E+09, 2.3448E+09, 2.3544E+09, 2.3641E+09,
     + 2.3738E+09, 2.3835E+09, 2.3933E+09, 2.4030E+09, 2.4129E+09,
     +  2.4228E+09, 2.4327E+09, 2.4426E+09, 2.4526E+09, 2.4626E+09,
     + 2.4726E+09, 2.4827E+09, 2.4928E+09, 2.5030E+09, 2.5132E+09,
     +  2.5234E+09, 2.5337E+09, 2.5440E+09, 2.5543E+09, 2.5647E+09,
     + 2.5751E+09, 2.5855E+09, 2.5960E+09, 2.6065E+09, 2.6171E+09,
     +  2.6277E+09, 2.6383E+09, 2.6490E+09, 2.6597E+09, 2.6704E+09,
     + 2.6812E+09, 2.6920E+09, 2.7029E+09, 2.7138E+09, 2.7247E+09,
     +  2.7357E+09, 2.7467E+09, 2.7577E+09, 2.7688E+09, 2.7799E+09,
     + 2.7911E+09, 2.8023E+09, 2.8135E+09, 2.8248E+09, 2.8361E+09,
     +  2.8475E+09, 2.8589E+09, 2.8703E+09, 2.8818E+09, 2.8933E+09,
     + 2.9049E+09, 2.9165E+09, 2.9282E+09, 2.9398E+09, 2.9516E+09,
     +  2.9633E+09, 2.9751E+09, 2.9870E+09, 2.9989E+09, 3.0108E+09,
     + 3.0227E+09, 3.0348E+09, 3.0468E+09, 3.0589E+09, 3.0710E+09,
     +  3.0832E+09, 3.0954E+09, 3.1077E+09, 3.1200E+09, 3.1323E+09,
     + 3.1447E+09, 3.1572E+09, 3.1696E+09, 3.1821E+09, 3.1947E+09,
     +  3.2073E+09, 3.2200E+09, 3.2326E+09, 3.2454E+09, 3.2582E+09,
     + 3.2710E+09, 3.2838E+09, 3.2967E+09, 3.3097E+09, 3.3227E+09,
     +  3.3357E+09, 3.3488E+09, 3.3619E+09, 3.3751E+09, 3.3883E+09,
     + 3.4016E+09, 3.4149E+09, 3.4283E+09, 3.4417E+09, 3.4551E+09,
     +  3.4686E+09, 3.4822E+09, 3.4957E+09, 3.5094E+09, 3.5230E+09,
     + 3.5368E+09, 3.5505E+09, 3.5644E+09, 3.5782E+09, 3.5921E+09,
     +  3.6061E+09, 3.6201E+09, 3.6341E+09, 3.6482E+09, 3.6624E+09,
     + 3.6766E+09, 3.6908E+09, 3.7051E+09, 3.7195E+09, 3.7338E+09,
     +  3.7483E+09, 3.7628E+09, 3.7773E+09, 3.7919E+09, 3.8065E+09,
     + 3.8212E+09, 3.8359E+09, 3.8507E+09, 3.8655E+09, 3.8804E+09,
     +  3.8953E+09, 3.9103E+09, 3.9253E+09, 3.9404E+09, 3.9556E+09,
     + 3.9707E+09, 3.9860E+09, 4.0012E+09, 4.0166E+09, 4.0320E+09,
     +  4.0474E+09, 4.0629E+09, 4.0784E+09, 4.0940E+09, 4.1096E+09,
     + 4.1253E+09, 4.1411E+09, 4.1569E+09, 4.1727E+09, 4.1886E+09,
     +  4.2046E+09, 4.2206E+09, 4.2367E+09, 4.2528E+09, 4.2689E+09,
     + 4.2852E+09, 4.3014E+09, 4.3178E+09, 4.3342E+09, 4.3506E+09,
     +  4.3671E+09, 4.3836E+09, 4.4002E+09, 4.4169E+09, 4.4336E+09,
     + 4.4504E+09, 4.4672E+09, 4.4841E+09, 4.5010E+09, 4.5180E+09,
     +  4.5350E+09, 4.5521E+09, 4.5693E+09, 4.5865E+09, 4.6038E+09,
     + 4.6211E+09, 4.6385E+09, 4.6559E+09, 4.6734E+09, 4.6909E+09,
     +  4.7086E+09, 4.7262E+09, 4.7439E+09, 4.7617E+09, 4.7796E+09,
     + 4.7975E+09, 4.8154E+09, 4.8334E+09, 4.8515E+09, 4.8697E+09,
     +  4.8878E+09, 4.9061E+09, 4.9244E+09, 4.9428E+09, 4.9612E+09,
     + 4.9797E+09, 4.9983E+09, 5.0169E+09, 5.0355E+09, 5.0543E+09,
     +  5.0731E+09, 5.0919E+09, 5.1108E+09, 5.1298E+09, 5.1488E+09,
     + 5.1679E+09, 5.1871E+09, 5.2063E+09, 5.2256E+09, 5.2449E+09,
     +  5.2643E+09, 5.2838E+09, 5.3033E+09, 5.3229E+09, 5.3426E+09,
     + 5.3623E+09, 5.3821E+09, 5.4020E+09, 5.4219E+09, 5.4419E+09,
     +  5.4619E+09, 5.4820E+09, 5.5022E+09, 5.5224E+09, 5.5427E+09,
     + 5.5631E+09, 5.5835E+09, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_SF6   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 1),QofT( 1,NT),Q(NT)
      data Tmx(1)/3500./
      data xgj/ 1./
c...      SF6
c...        --        29
      data (QofT( 1,J),J=1,2408)/ 1.0220E+02, 2.8646E+02, 5.2343E+02,
     +  8.0367E+02, 1.1213E+03, 1.4724E+03, 1.8540E+03, 2.2639E+03,
     + 2.7001E+03, 3.1612E+03, 3.6460E+03, 4.1533E+03, 4.6822E+03,
     +  5.2318E+03, 5.8013E+03, 6.3901E+03, 6.9976E+03, 7.6233E+03,
     + 8.2665E+03, 8.9269E+03, 1.0297E+04, 1.1732E+04, 1.3227E+04,
     +  1.4781E+04, 1.6391E+04, 1.8056E+04, 1.9774E+04, 2.1543E+04,
     + 2.3362E+04, 2.5230E+04, 2.7145E+04, 2.9106E+04, 3.1113E+04,
     +  3.3164E+04, 3.5259E+04, 3.7397E+04, 3.9578E+04, 4.1801E+04,
     + 4.4066E+04, 4.6373E+04, 4.8722E+04, 5.1113E+04, 5.3545E+04,
     +  5.6020E+04, 5.8537E+04, 6.1097E+04, 6.3702E+04, 6.6350E+04,
     + 6.9044E+04, 7.1784E+04, 7.4572E+04, 7.7408E+04, 8.0293E+04,
     +  8.3230E+04, 8.6219E+04, 8.9262E+04, 9.2361E+04, 9.5517E+04,
     + 9.8733E+04, 1.0201E+05, 1.0535E+05, 1.0876E+05, 1.1223E+05,
     +  1.1577E+05, 1.1939E+05, 1.2308E+05, 1.2685E+05, 1.3069E+05,
     + 1.3463E+05, 1.3864E+05, 1.4275E+05, 1.4695E+05, 1.5124E+05,
     +  1.5563E+05, 1.6013E+05, 1.6473E+05, 1.6944E+05, 1.7426E+05,
     + 1.7920E+05, 1.8426E+05, 1.8944E+05, 1.9475E+05, 2.0020E+05,
     +  2.0579E+05, 2.1151E+05, 2.1739E+05, 2.2342E+05, 2.2960E+05,
     + 2.3595E+05, 2.4247E+05, 2.4916E+05, 2.5603E+05, 2.6308E+05,
     +  2.7033E+05, 2.7777E+05, 2.8542E+05, 2.9328E+05, 3.0136E+05,
     + 3.0966E+05, 3.1819E+05, 3.2696E+05, 3.3598E+05, 3.4525E+05,
     +  3.5479E+05, 3.6460E+05, 3.7469E+05, 3.8506E+05, 3.9574E+05,
     + 4.0672E+05, 4.1802E+05, 4.2964E+05, 4.4160E+05, 4.5392E+05,
     +  4.6659E+05, 4.7963E+05, 4.9305E+05, 5.0687E+05, 5.2109E+05,
     + 5.3573E+05, 5.5081E+05, 5.6633E+05, 5.8230E+05, 5.9876E+05,
     +  6.1570E+05, 6.3314E+05, 6.5111E+05, 6.6961E+05, 6.8866E+05,
     + 7.0828E+05, 7.2849E+05, 7.4930E+05, 7.7074E+05, 7.9282E+05,
     +  8.1556E+05, 8.3899E+05, 8.6312E+05, 8.8798E+05, 9.1358E+05,
     + 9.3996E+05, 9.6713E+05, 9.9512E+05, 1.0240E+06, 1.0537E+06,
     +  1.0843E+06, 1.1158E+06, 1.1483E+06, 1.1817E+06, 1.2162E+06,
     + 1.2517E+06, 1.2883E+06, 1.3260E+06, 1.3649E+06, 1.4049E+06,
     +  1.4461E+06, 1.4886E+06, 1.5324E+06, 1.5775E+06, 1.6240E+06,
     + 1.6718E+06, 1.7212E+06, 1.7720E+06, 1.8243E+06, 1.8783E+06,
     +  1.9338E+06, 1.9911E+06, 2.0501E+06, 2.1109E+06, 2.1735E+06,
     + 2.2380E+06, 2.3045E+06, 2.3730E+06, 2.4435E+06, 2.5162E+06,
     +  2.5911E+06, 2.6682E+06, 2.7477E+06, 2.8296E+06, 2.9139E+06,
     + 3.0008E+06, 3.0903E+06, 3.1826E+06, 3.2776E+06, 3.3754E+06,
     +  3.4762E+06, 3.5801E+06, 3.6870E+06, 3.7972E+06, 3.9107E+06,
     + 4.0277E+06, 4.1481E+06, 4.2721E+06, 4.3999E+06, 4.5315E+06,
     +  4.6670E+06, 4.8066E+06, 4.9504E+06, 5.0984E+06, 5.2509E+06,
     + 5.4080E+06, 5.5698E+06, 5.7363E+06, 5.9079E+06, 6.0845E+06,
     +  6.2664E+06, 6.4538E+06, 6.6467E+06, 6.8453E+06, 7.0498E+06,
     + 7.2604E+06, 7.4773E+06, 7.7006E+06, 7.9305E+06, 8.1671E+06,
     +  8.4108E+06, 8.6617E+06, 8.9200E+06, 9.1859E+06, 9.4596E+06,
     + 9.7414E+06, 1.0032E+07, 1.0330E+07, 1.0638E+07, 1.0954E+07,
     +  1.1280E+07, 1.1615E+07, 1.1960E+07, 1.2315E+07, 1.2681E+07,
     + 1.3057E+07, 1.3444E+07, 1.3842E+07, 1.4252E+07, 1.4674E+07,
     +  1.5109E+07, 1.5555E+07, 1.6015E+07, 1.6488E+07, 1.6975E+07,
     + 1.7476E+07, 1.7991E+07, 1.8521E+07, 1.9067E+07, 1.9628E+07,
     +  2.0206E+07, 2.0800E+07, 2.1411E+07, 2.2039E+07, 2.2686E+07,
     + 2.3351E+07, 2.4035E+07, 2.4739E+07, 2.5463E+07, 2.6207E+07,
     +  2.6973E+07, 2.7760E+07, 2.8570E+07, 2.9403E+07, 3.0259E+07,
     + 3.1140E+07, 3.2045E+07, 3.2976E+07, 3.3934E+07, 3.4918E+07,
     +  3.5931E+07, 3.6971E+07, 3.8042E+07, 3.9142E+07, 4.0273E+07,
     + 4.1435E+07, 4.2630E+07, 4.3859E+07, 4.5122E+07, 4.6420E+07,
     +  4.7755E+07, 4.9126E+07, 5.0536E+07, 5.1985E+07, 5.3474E+07,
     + 5.5004E+07, 5.6577E+07, 5.8193E+07, 5.9854E+07, 6.1560E+07,
     +  6.3314E+07, 6.5116E+07, 6.6967E+07, 6.8870E+07, 7.0824E+07,
     + 7.2832E+07, 7.4895E+07, 7.7014E+07, 7.9192E+07, 8.1428E+07,
     +  8.3726E+07, 8.6086E+07, 8.8510E+07, 9.0999E+07, 9.3557E+07,
     + 9.6183E+07, 9.8881E+07, 1.0165E+08, 1.0450E+08, 1.0742E+08,
     +  1.1042E+08, 1.1350E+08, 1.1666E+08, 1.1991E+08, 1.2325E+08,
     + 1.2667E+08, 1.3019E+08, 1.3380E+08, 1.3751E+08, 1.4131E+08,
     +  1.4522E+08, 1.4923E+08, 1.5334E+08, 1.5757E+08, 1.6191E+08,
     + 1.6636E+08, 1.7093E+08, 1.7562E+08, 1.8043E+08, 1.8537E+08,
     +  1.9044E+08, 1.9564E+08, 2.0098E+08, 2.0646E+08, 2.1208E+08,
     + 2.1784E+08, 2.2376E+08, 2.2983E+08, 2.3607E+08, 2.4246E+08,
     +  2.4901E+08, 2.5574E+08, 2.6264E+08, 2.6972E+08, 2.7699E+08,
     + 2.8444E+08, 2.9208E+08, 2.9991E+08, 3.0795E+08, 3.1620E+08,
     +  3.2465E+08, 3.3333E+08, 3.4222E+08, 3.5134E+08, 3.6070E+08,
     + 3.7029E+08, 3.8012E+08, 3.9021E+08, 4.0055E+08, 4.1115E+08,
     +  4.2202E+08, 4.3317E+08, 4.4459E+08, 4.5631E+08, 4.6831E+08,
     + 4.8062E+08, 4.9324E+08, 5.0618E+08, 5.1944E+08, 5.3303E+08,
     +  5.4696E+08, 5.6124E+08, 5.7587E+08, 5.9087E+08, 6.0624E+08,
     + 6.2199E+08, 6.3813E+08, 6.5467E+08, 6.7162E+08, 6.8899E+08,
     +  7.0679E+08, 7.2502E+08, 7.4370E+08, 7.6284E+08, 7.8246E+08,
     + 8.0255E+08, 8.2313E+08, 8.4422E+08, 8.6582E+08, 8.8794E+08,
     +  9.1061E+08, 9.3383E+08, 9.5761E+08, 9.8197E+08, 1.0069E+09,
     + 1.0325E+09, 1.0586E+09, 1.0854E+09, 1.1129E+09, 1.1410E+09,
     +  1.1698E+09, 1.1992E+09, 1.2294E+09, 1.2603E+09, 1.2919E+09,
     + 1.3243E+09, 1.3575E+09, 1.3915E+09, 1.4262E+09, 1.4618E+09,
     +  1.4983E+09, 1.5356E+09, 1.5737E+09, 1.6128E+09, 1.6528E+09,
     + 1.6938E+09, 1.7357E+09, 1.7786E+09, 1.8225E+09, 1.8675E+09,
     +  1.9135E+09, 1.9605E+09, 2.0087E+09, 2.0580E+09, 2.1084E+09,
     + 2.1600E+09, 2.2129E+09, 2.2669E+09, 2.3222E+09, 2.3788E+09,
     +  2.4367E+09, 2.4959E+09, 2.5565E+09, 2.6185E+09, 2.6819E+09,
     + 2.7468E+09, 2.8131E+09, 2.8810E+09, 2.9505E+09, 3.0215E+09,
     +  3.0941E+09, 3.1685E+09, 3.2445E+09, 3.3222E+09, 3.4017E+09,
     + 3.4830E+09, 3.5662E+09, 3.6512E+09, 3.7381E+09, 3.8271E+09,
     +  3.9180E+09, 4.0110E+09, 4.1060E+09, 4.2032E+09, 4.3026E+09,
     + 4.4042E+09, 4.5081E+09, 4.6144E+09, 4.7229E+09, 4.8340E+09,
     +  4.9474E+09, 5.0634E+09, 5.1820E+09, 5.3033E+09, 5.4272E+09,
     + 5.5538E+09, 5.6833E+09, 5.8156E+09, 5.9508E+09, 6.0890E+09,
     +  6.2303E+09, 6.3747E+09, 6.5222E+09, 6.6729E+09, 6.8270E+09,
     + 6.9845E+09, 7.1453E+09, 7.3097E+09, 7.4777E+09, 7.6493E+09,
     +  7.8247E+09, 8.0038E+09, 8.1869E+09, 8.3739E+09, 8.5649E+09,
     + 8.7601E+09, 8.9595E+09, 9.1632E+09, 9.3713E+09, 9.5838E+09,
     +  9.8009E+09, 1.0023E+10, 1.0249E+10, 1.0481E+10, 1.0717E+10,
     + 1.0958E+10, 1.1205E+10, 1.1457E+10, 1.1714E+10, 1.1976E+10,
     +  1.2244E+10, 1.2518E+10, 1.2798E+10, 1.3083E+10, 1.3375E+10,
     + 1.3673E+10, 1.3976E+10, 1.4287E+10, 1.4604E+10, 1.4927E+10,
     +  1.5258E+10, 1.5595E+10, 1.5939E+10, 1.6290E+10, 1.6649E+10,
     + 1.7016E+10, 1.7389E+10, 1.7771E+10, 1.8161E+10, 1.8558E+10,
     +  1.8964E+10, 1.9378E+10, 1.9801E+10, 2.0233E+10, 2.0673E+10,
     + 2.1123E+10, 2.1582E+10, 2.2050E+10, 2.2528E+10, 2.3015E+10,
     +  2.3513E+10, 2.4020E+10, 2.4539E+10, 2.5067E+10, 2.5607E+10,
     + 2.6157E+10, 2.6719E+10, 2.7292E+10, 2.7876E+10, 2.8473E+10,
     +  2.9081E+10, 2.9702E+10, 3.0335E+10, 3.0981E+10, 3.1640E+10,
     + 3.2313E+10, 3.2998E+10, 3.3698E+10, 3.4411E+10, 3.5139E+10,
     +  3.5882E+10, 3.6639E+10, 3.7411E+10, 3.8199E+10, 3.9002E+10,
     + 3.9822E+10, 4.0657E+10, 4.1509E+10, 4.2378E+10, 4.3265E+10,
     +  4.4168E+10, 4.5090E+10, 4.6029E+10, 4.6988E+10, 4.7965E+10,
     + 4.8961E+10, 4.9977E+10, 5.1012E+10, 5.2068E+10, 5.3145E+10,
     +  5.4242E+10, 5.5361E+10, 5.6502E+10, 5.7665E+10, 5.8851E+10,
     + 6.0060E+10, 6.1292E+10, 6.2547E+10, 6.3828E+10, 6.5133E+10,
     +  6.6463E+10, 6.7819E+10, 6.9201E+10, 7.0609E+10, 7.2045E+10,
     + 7.3508E+10, 7.4999E+10, 7.6519E+10, 7.8067E+10, 7.9646E+10,
     +  8.1254E+10, 8.2893E+10, 8.4564E+10, 8.6266E+10, 8.8000E+10,
     + 8.9768E+10, 9.1569E+10, 9.3404E+10, 9.5274E+10, 9.7179E+10,
     +  9.9120E+10, 1.0110E+11, 1.0311E+11, 1.0516E+11, 1.0726E+11,
     + 1.0939E+11, 1.1156E+11, 1.1377E+11, 1.1602E+11, 1.1831E+11,
     +  1.2065E+11, 1.2303E+11, 1.2546E+11, 1.2793E+11, 1.3044E+11,
     + 1.3301E+11, 1.3562E+11, 1.3828E+11, 1.4098E+11, 1.4374E+11,
     +  1.4655E+11, 1.4941E+11, 1.5233E+11, 1.5529E+11, 1.5831E+11,
     + 1.6139E+11, 1.6452E+11, 1.6771E+11, 1.7096E+11, 1.7427E+11,
     +  1.7764E+11, 1.8107E+11, 1.8456E+11, 1.8812E+11, 1.9174E+11,
     + 1.9543E+11, 1.9918E+11, 2.0300E+11, 2.0689E+11, 2.1085E+11,
     +  2.1489E+11, 2.1899E+11, 2.2317E+11, 2.2742E+11, 2.3175E+11,
     + 2.3616E+11, 2.4065E+11, 2.4522E+11, 2.4987E+11, 2.5460E+11,
     +  2.5942E+11, 2.6432E+11, 2.6931E+11, 2.7439E+11, 2.7956E+11,
     + 2.8482E+11, 2.9018E+11, 2.9563E+11, 3.0117E+11, 3.0682E+11,
     +  3.1256E+11, 3.1841E+11, 3.2436E+11, 3.3041E+11, 3.3657E+11,
     + 3.4284E+11, 3.4922E+11, 3.5571E+11, 3.6231E+11, 3.6903E+11,
     +  3.7587E+11, 3.8283E+11, 3.8990E+11, 3.9710E+11, 4.0443E+11,
     + 4.1188E+11, 4.1947E+11, 4.2718E+11, 4.3503E+11, 4.4302E+11,
     +  4.5114E+11, 4.5940E+11, 4.6781E+11, 4.7636E+11, 4.8506E+11,
     + 4.9390E+11, 5.0290E+11, 5.1206E+11, 5.2137E+11, 5.3084E+11,
     +  5.4047E+11, 5.5027E+11, 5.6024E+11, 5.7037E+11, 5.8068E+11,
     + 5.9116E+11, 6.0182E+11, 6.1267E+11, 6.2369E+11, 6.3491E+11,
     +  6.4631E+11, 6.5791E+11, 6.6970E+11, 6.8170E+11, 6.9389E+11,
     + 7.0629E+11, 7.1890E+11, 7.3172E+11, 7.4476E+11, 7.5801E+11,
     +  7.7149E+11, 7.8520E+11, 7.9913E+11, 8.1329E+11, 8.2770E+11,
     + 8.4234E+11, 8.5723E+11, 8.7236E+11, 8.8775E+11, 9.0339E+11,
     +  9.1929E+11, 9.3546E+11, 9.5189E+11, 9.6859E+11, 9.8558E+11,
     + 1.0028E+12, 1.0204E+12, 1.0382E+12, 1.0563E+12, 1.0748E+12,
     +  1.0935E+12, 1.1125E+12, 1.1319E+12, 1.1516E+12, 1.1715E+12,
     + 1.1919E+12, 1.2125E+12, 1.2335E+12, 1.2548E+12, 1.2765E+12,
     +  1.2985E+12, 1.3209E+12, 1.3437E+12, 1.3668E+12, 1.3903E+12,
     + 1.4141E+12, 1.4384E+12, 1.4630E+12, 1.4881E+12, 1.5135E+12,
     +  1.5394E+12, 1.5657E+12, 1.5924E+12, 1.6195E+12, 1.6471E+12,
     + 1.6751E+12, 1.7035E+12, 1.7324E+12, 1.7618E+12, 1.7917E+12,
     +  1.8220E+12, 1.8528E+12, 1.8841E+12, 1.9158E+12, 1.9481E+12,
     + 1.9809E+12, 2.0143E+12, 2.0481E+12, 2.0825E+12, 2.1174E+12,
     +  2.1529E+12, 2.1890E+12, 2.2256E+12, 2.2628E+12, 2.3005E+12,
     + 2.3389E+12, 2.3779E+12, 2.4175E+12, 2.4577E+12, 2.4985E+12,
     +  2.5400E+12, 2.5821E+12, 2.6248E+12, 2.6683E+12, 2.7124E+12,
     + 2.7572E+12, 2.8027E+12, 2.8489E+12, 2.8958E+12, 2.9435E+12,
     +  2.9919E+12, 3.0410E+12, 3.0909E+12, 3.1416E+12, 3.1931E+12,
     + 3.2453E+12, 3.2984E+12, 3.3522E+12, 3.4069E+12, 3.4625E+12,
     +  3.5189E+12, 3.5761E+12, 3.6342E+12, 3.6933E+12, 3.7532E+12,
     + 3.8140E+12, 3.8758E+12, 3.9385E+12, 4.0022E+12, 4.0668E+12,
     +  4.1324E+12, 4.1990E+12, 4.2667E+12, 4.3353E+12, 4.4050E+12,
     + 4.4757E+12, 4.5475E+12, 4.6204E+12, 4.6944E+12, 4.7695E+12,
     +  4.8458E+12, 4.9232E+12, 5.0017E+12, 5.0814E+12, 5.1624E+12,
     + 5.2445E+12, 5.3279E+12, 5.4125E+12, 5.4984E+12, 5.5856E+12,
     +  5.6740E+12, 5.7638E+12, 5.8549E+12, 5.9474E+12, 6.0413E+12,
     + 6.1365E+12, 6.2332E+12, 6.3313E+12, 6.4308E+12, 6.5318E+12,
     +  6.6343E+12, 6.7384E+12, 6.8439E+12, 6.9510E+12, 7.0597E+12,
     + 7.1700E+12, 7.2819E+12, 7.3955E+12, 7.5107E+12, 7.6276E+12,
     +  7.7463E+12, 7.8666E+12, 7.9888E+12, 8.1127E+12, 8.2384E+12,
     + 8.3660E+12, 8.4954E+12, 8.6267E+12, 8.7599E+12, 8.8950E+12,
     +  9.0321E+12, 9.1712E+12, 9.3123E+12, 9.4554E+12, 9.6007E+12,
     + 9.7480E+12, 9.8974E+12, 1.0049E+13, 1.0203E+13, 1.0359E+13,
     +  1.0517E+13, 1.0678E+13, 1.0840E+13, 1.1006E+13, 1.1173E+13,
     + 1.1343E+13, 1.1515E+13, 1.1690E+13, 1.1868E+13, 1.2047E+13,
     +  1.2230E+13, 1.2415E+13, 1.2602E+13, 1.2793E+13, 1.2986E+13,
     + 1.3182E+13, 1.3380E+13, 1.3581E+13, 1.3786E+13, 1.3993E+13,
     +  1.4203E+13, 1.4416E+13, 1.4632E+13, 1.4851E+13, 1.5073E+13,
     + 1.5298E+13, 1.5527E+13, 1.5758E+13, 1.5993E+13, 1.6231E+13,
     +  1.6473E+13, 1.6718E+13, 1.6966E+13, 1.7218E+13, 1.7473E+13,
     + 1.7732E+13, 1.7994E+13, 1.8260E+13, 1.8530E+13, 1.8804E+13,
     +  1.9081E+13, 1.9362E+13, 1.9647E+13, 1.9936E+13, 2.0230E+13,
     + 2.0527E+13, 2.0828E+13, 2.1133E+13, 2.1443E+13, 2.1757E+13,
     +  2.2075E+13, 2.2397E+13, 2.2724E+13, 2.3056E+13, 2.3392E+13,
     + 2.3733E+13, 2.4078E+13, 2.4428E+13, 2.4783E+13, 2.5142E+13,
     +  2.5507E+13, 2.5877E+13, 2.6251E+13, 2.6631E+13, 2.7016E+13,
     + 2.7406E+13, 2.7801E+13, 2.8202E+13, 2.8608E+13, 2.9020E+13,
     +  2.9437E+13, 2.9860E+13, 3.0288E+13, 3.0723E+13, 3.1163E+13,
     + 3.1609E+13, 3.2062E+13, 3.2520E+13, 3.2984E+13, 3.3455E+13,
     +  3.3932E+13, 3.4415E+13, 3.4905E+13, 3.5402E+13, 3.5905E+13,
     + 3.6415E+13, 3.6931E+13, 3.7455E+13, 3.7985E+13, 3.8523E+13,
     +  3.9068E+13, 3.9619E+13, 4.0179E+13, 4.0745E+13, 4.1320E+13,
     + 4.1902E+13, 4.2491E+13, 4.3088E+13, 4.3694E+13, 4.4307E+13,
     +  4.4928E+13, 4.5558E+13, 4.6196E+13, 4.6842E+13, 4.7497E+13,
     + 4.8160E+13, 4.8832E+13, 4.9513E+13, 5.0203E+13, 5.0902E+13,
     +  5.1610E+13, 5.2327E+13, 5.3054E+13, 5.3790E+13, 5.4535E+13,
     + 5.5291E+13, 5.6056E+13, 5.6831E+13, 5.7617E+13, 5.8412E+13,
     +  5.9218E+13, 6.0034E+13, 6.0861E+13, 6.1699E+13, 6.2547E+13,
     + 6.3407E+13, 6.4277E+13, 6.5159E+13, 6.6052E+13, 6.6957E+13,
     +  6.7873E+13, 6.8802E+13, 6.9742E+13, 7.0694E+13, 7.1658E+13,
     + 7.2635E+13, 7.3624E+13, 7.4626E+13, 7.5641E+13, 7.6668E+13,
     +  7.7709E+13, 7.8763E+13, 7.9831E+13, 8.0912E+13, 8.2007E+13,
     + 8.3116E+13, 8.4239E+13, 8.5376E+13, 8.6528E+13, 8.7694E+13,
     +  8.8876E+13, 9.0072E+13, 9.1283E+13, 9.2509E+13, 9.3751E+13,
     + 9.5009E+13, 9.6283E+13, 9.7572E+13, 9.8878E+13, 1.0020E+14,
     +  1.0154E+14, 1.0289E+14, 1.0427E+14, 1.0566E+14, 1.0707E+14,
     + 1.0849E+14, 1.0993E+14, 1.1139E+14, 1.1287E+14, 1.1437E+14,
     +  1.1589E+14, 1.1742E+14, 1.1898E+14, 1.2055E+14, 1.2214E+14,
     + 1.2376E+14, 1.2539E+14, 1.2704E+14, 1.2872E+14, 1.3041E+14,
     +  1.3213E+14, 1.3386E+14, 1.3562E+14, 1.3740E+14, 1.3920E+14,
     + 1.4103E+14, 1.4287E+14, 1.4474E+14, 1.4664E+14, 1.4855E+14,
     +  1.5049E+14, 1.5245E+14, 1.5444E+14, 1.5645E+14, 1.5849E+14,
     + 1.6055E+14, 1.6264E+14, 1.6475E+14, 1.6688E+14, 1.6905E+14,
     +  1.7124E+14, 1.7345E+14, 1.7570E+14, 1.7797E+14, 1.8027E+14,
     + 1.8259E+14, 1.8495E+14, 1.8733E+14, 1.8974E+14, 1.9218E+14,
     +  1.9465E+14, 1.9715E+14, 1.9969E+14, 2.0225E+14, 2.0484E+14,
     + 2.0746E+14, 2.1012E+14, 2.1280E+14, 2.1552E+14, 2.1828E+14,
     +  2.2106E+14, 2.2388E+14, 2.2673E+14, 2.2962E+14, 2.3254E+14,
     + 2.3549E+14, 2.3848E+14, 2.4151E+14, 2.4457E+14, 2.4767E+14,
     +  2.5081E+14, 2.5398E+14, 2.5719E+14, 2.6044E+14, 2.6373E+14,
     + 2.6706E+14, 2.7043E+14, 2.7383E+14, 2.7728E+14, 2.8077E+14,
     +  2.8430E+14, 2.8787E+14, 2.9148E+14, 2.9513E+14, 2.9883E+14,
     + 3.0258E+14, 3.0636E+14, 3.1019E+14, 3.1407E+14, 3.1799E+14,
     +  3.2196E+14, 3.2597E+14, 3.3003E+14, 3.3414E+14, 3.3830E+14,
     + 3.4250E+14, 3.4676E+14, 3.5106E+14, 3.5541E+14, 3.5982E+14,
     +  3.6427E+14, 3.6878E+14, 3.7334E+14, 3.7796E+14, 3.8262E+14,
     + 3.8735E+14, 3.9212E+14, 3.9695E+14, 4.0184E+14, 4.0678E+14,
     +  4.1179E+14, 4.1684E+14, 4.2196E+14, 4.2714E+14, 4.3237E+14,
     + 4.3767E+14, 4.4303E+14, 4.4845E+14, 4.5393E+14, 4.5947E+14,
     +  4.6508E+14, 4.7075E+14, 4.7649E+14, 4.8229E+14, 4.8816E+14,
     + 4.9410E+14, 5.0011E+14, 5.0618E+14, 5.1232E+14, 5.1853E+14,
     +  5.2482E+14, 5.3117E+14, 5.3760E+14, 5.4410E+14, 5.5068E+14,
     + 5.5732E+14, 5.6405E+14, 5.7085E+14, 5.7773E+14, 5.8468E+14,
     +  5.9172E+14, 5.9883E+14, 6.0603E+14, 6.1330E+14, 6.2066E+14,
     + 6.2810E+14, 6.3563E+14, 6.4324E+14, 6.5093E+14, 6.5871E+14,
     +  6.6658E+14, 6.7454E+14, 6.8259E+14, 6.9072E+14, 6.9895E+14,
     + 7.0727E+14, 7.1568E+14, 7.2419E+14, 7.3279E+14, 7.4149E+14,
     +  7.5029E+14, 7.5918E+14, 7.6817E+14, 7.7727E+14, 7.8646E+14,
     + 7.9576E+14, 8.0516E+14, 8.1466E+14, 8.2427E+14, 8.3399E+14,
     +  8.4381E+14, 8.5374E+14, 8.6378E+14, 8.7394E+14, 8.8420E+14,
     + 8.9458E+14, 9.0507E+14, 9.1568E+14, 9.2641E+14, 9.3725E+14,
     +  9.4822E+14, 9.5930E+14, 9.7051E+14, 9.8183E+14, 9.9329E+14,
     + 1.0049E+15, 1.0166E+15, 1.0284E+15, 1.0404E+15, 1.0525E+15,
     +  1.0647E+15, 1.0770E+15, 1.0895E+15, 1.1022E+15, 1.1149E+15,
     + 1.1278E+15, 1.1409E+15, 1.1541E+15, 1.1674E+15, 1.1809E+15,
     +  1.1945E+15, 1.2083E+15, 1.2222E+15, 1.2363E+15, 1.2505E+15,
     + 1.2649E+15, 1.2794E+15, 1.2941E+15, 1.3090E+15, 1.3240E+15,
     +  1.3391E+15, 1.3545E+15, 1.3700E+15, 1.3856E+15, 1.4015E+15,
     + 1.4175E+15, 1.4336E+15, 1.4500E+15, 1.4665E+15, 1.4832E+15,
     +  1.5001E+15, 1.5172E+15, 1.5344E+15, 1.5518E+15, 1.5694E+15,
     + 1.5872E+15, 1.6052E+15, 1.6234E+15, 1.6418E+15, 1.6604E+15,
     +  1.6791E+15, 1.6981E+15, 1.7173E+15, 1.7366E+15, 1.7562E+15,
     + 1.7760E+15, 1.7960E+15, 1.8162E+15, 1.8366E+15, 1.8573E+15,
     +  1.8781E+15, 1.8992E+15, 1.9205E+15, 1.9420E+15, 1.9638E+15,
     + 1.9857E+15, 2.0079E+15, 2.0304E+15, 2.0531E+15, 2.0760E+15,
     +  2.0991E+15, 2.1225E+15, 2.1462E+15, 2.1701E+15, 2.1942E+15,
     + 2.2186E+15, 2.2433E+15, 2.2682E+15, 2.2933E+15, 2.3188E+15,
     +  2.3444E+15, 2.3704E+15, 2.3966E+15, 2.4231E+15, 2.4499E+15,
     + 2.4770E+15, 2.5043E+15, 2.5319E+15, 2.5598E+15, 2.5880E+15,
     +  2.6165E+15, 2.6453E+15, 2.6744E+15, 2.7038E+15, 2.7335E+15,
     + 2.7635E+15, 2.7938E+15, 2.8244E+15, 2.8553E+15, 2.8865E+15,
     +  2.9181E+15, 2.9500E+15, 2.9822E+15, 3.0148E+15, 3.0477E+15,
     + 3.0809E+15, 3.1144E+15, 3.1483E+15, 3.1826E+15, 3.2172E+15,
     +  3.2522E+15, 3.2875E+15, 3.3231E+15, 3.3592E+15, 3.3956E+15,
     + 3.4324E+15, 3.4695E+15, 3.5070E+15, 3.5449E+15, 3.5832E+15,
     +  3.6219E+15, 3.6610E+15, 3.7005E+15, 3.7403E+15, 3.7806E+15,
     + 3.8213E+15, 3.8624E+15, 3.9039E+15, 3.9458E+15, 3.9882E+15,
     +  4.0309E+15, 4.0742E+15, 4.1178E+15, 4.1619E+15, 4.2064E+15,
     + 4.2514E+15, 4.2968E+15, 4.3427E+15, 4.3890E+15, 4.4359E+15,
     +  4.4831E+15, 4.5309E+15, 4.5791E+15, 4.6278E+15, 4.6770E+15,
     + 4.7267E+15, 4.7769E+15, 4.8276E+15, 4.8788E+15, 4.9305E+15,
     +  4.9827E+15, 5.0354E+15, 5.0887E+15, 5.1425E+15, 5.1968E+15,
     + 5.2517E+15, 5.3071E+15, 5.3631E+15, 5.4196E+15, 5.4766E+15,
     +  5.5343E+15, 5.5925E+15, 5.6513E+15, 5.7106E+15, 5.7706E+15,
     + 5.8311E+15, 5.8923E+15, 5.9540E+15, 6.0164E+15, 6.0793E+15,
     +  6.1429E+15, 6.2071E+15, 6.2720E+15, 6.3374E+15, 6.4035E+15,
     + 6.4703E+15, 6.5377E+15, 6.6058E+15, 6.6745E+15, 6.7440E+15,
     +  6.8141E+15, 6.8848E+15, 6.9563E+15, 7.0285E+15, 7.1013E+15,
     + 7.1749E+15, 7.2492E+15, 7.3242E+15, 7.4000E+15, 7.4765E+15,
     +  7.5537E+15, 7.6317E+15, 7.7104E+15, 7.7899E+15, 7.8702E+15,
     + 7.9512E+15, 8.0331E+15, 8.1157E+15, 8.1991E+15, 8.2834E+15,
     +  8.3684E+15, 8.4543E+15, 8.5409E+15, 8.6285E+15, 8.7168E+15,
     + 8.8060E+15, 8.8961E+15, 8.9871E+15, 9.0789E+15, 9.1716E+15,
     +  9.2652E+15, 9.3596E+15, 9.4550E+15, 9.5513E+15, 9.6485E+15,
     + 9.7467E+15, 9.8458E+15, 9.9458E+15, 1.0047E+16, 1.0149E+16,
     +  1.0252E+16, 1.0356E+16, 1.0461E+16, 1.0566E+16, 1.0673E+16,
     + 1.0781E+16, 1.0890E+16, 1.1000E+16, 1.1111E+16, 1.1223E+16,
     +  1.1336E+16, 1.1451E+16, 1.1566E+16, 1.1682E+16, 1.1800E+16,
     + 1.1918E+16, 1.2038E+16, 1.2159E+16, 1.2281E+16, 1.2404E+16,
     +  1.2528E+16, 1.2654E+16, 1.2780E+16, 1.2908E+16, 1.3037E+16,
     + 1.3168E+16, 1.3299E+16, 1.3432E+16, 1.3566E+16, 1.3701E+16,
     +  1.3837E+16, 1.3975E+16, 1.4114E+16, 1.4255E+16, 1.4396E+16,
     + 1.4539E+16, 1.4683E+16, 1.4829E+16, 1.4976E+16, 1.5125E+16,
     +  1.5274E+16, 1.5425E+16, 1.5578E+16, 1.5732E+16, 1.5887E+16,
     + 1.6044E+16, 1.6203E+16, 1.6362E+16, 1.6524E+16, 1.6686E+16,
     +  1.6851E+16, 1.7016E+16, 1.7184E+16, 1.7352E+16, 1.7523E+16,
     + 1.7695E+16, 1.7868E+16, 1.8044E+16, 1.8220E+16, 1.8399E+16,
     +  1.8579E+16, 1.8761E+16, 1.8944E+16, 1.9129E+16, 1.9316E+16,
     + 1.9504E+16, 1.9694E+16, 1.9886E+16, 2.0080E+16, 2.0276E+16,
     +  2.0473E+16, 2.0672E+16, 2.0873E+16, 2.1075E+16, 2.1280E+16,
     + 2.1486E+16, 2.1695E+16, 2.1905E+16, 2.2117E+16, 2.2331E+16,
     +  2.2547E+16, 2.2765E+16, 2.2985E+16, 2.3207E+16, 2.3431E+16,
     + 2.3657E+16, 2.3885E+16, 2.4115E+16, 2.4347E+16, 2.4582E+16,
     +  2.4818E+16, 2.5057E+16, 2.5297E+16, 2.5540E+16, 2.5785E+16,
     + 2.6032E+16, 2.6282E+16, 2.6534E+16, 2.6788E+16, 2.7044E+16,
     +  2.7303E+16, 2.7564E+16, 2.7827E+16, 2.8092E+16, 2.8360E+16,
     + 2.8631E+16, 2.8904E+16, 2.9179E+16, 2.9457E+16, 2.9737E+16,
     +  3.0020E+16, 3.0305E+16, 3.0593E+16, 3.0883E+16, 3.1176E+16,
     + 3.1472E+16, 3.1770E+16, 3.2071E+16, 3.2374E+16, 3.2680E+16,
     +  3.2989E+16, 3.3301E+16, 3.3615E+16, 3.3933E+16, 3.4253E+16,
     + 3.4576E+16, 3.4901E+16, 3.5230E+16, 3.5562E+16, 3.5896E+16,
     +  3.6234E+16, 3.6574E+16, 3.6918E+16, 3.7264E+16, 3.7613E+16,
     + 3.7966E+16, 3.8322E+16, 3.8681E+16, 3.9043E+16, 3.9408E+16,
     +  3.9776E+16, 4.0148E+16, 4.0523E+16, 4.0901E+16, 4.1282E+16,
     + 4.1667E+16, 4.2055E+16, 4.2447E+16, 4.2842E+16, 4.3240E+16,
     +  4.3642E+16, 4.4048E+16, 4.4457E+16, 4.4869E+16, 4.5286E+16,
     + 4.5705E+16, 4.6129E+16, 4.6556E+16, 4.6987E+16, 4.7422E+16,
     +  4.7860E+16, 4.8302E+16, 4.8748E+16, 4.9198E+16, 4.9652E+16,
     + 5.0110E+16, 5.0571E+16, 5.1037E+16, 5.1507E+16, 5.1981E+16,
     +  5.2458E+16, 5.2940E+16, 5.3427E+16, 5.3917E+16, 5.4412E+16,
     + 5.4910E+16, 5.5414E+16, 5.5921E+16, 5.6433E+16, 5.6949E+16,
     +  5.7470E+16, 5.7995E+16, 5.8525E+16, 5.9059E+16, 5.9598E+16,
     + 6.0141E+16, 6.0689E+16, 6.1242E+16, 6.1800E+16, 6.2362E+16,
     +  6.2929E+16, 6.3501E+16, 6.4078E+16, 6.4659E+16, 6.5246E+16,
     + 6.5838E+16, 6.6435E+16, 6.7036E+16, 6.7643E+16, 6.8255E+16,
     +  6.8873E+16, 6.9495E+16, 7.0123E+16, 7.0756E+16, 7.1395E+16,
     + 7.2039E+16, 7.2688E+16, 7.3343E+16, 7.4004E+16, 7.4670E+16,
     +  7.5341E+16, 7.6019E+16, 7.6702E+16, 7.7391E+16, 7.8085E+16,
     + 7.8786E+16, 7.9492E+16, 8.0204E+16, 8.0923E+16, 8.1647E+16,
     +  8.2378E+16, 8.3114E+16, 8.3857E+16, 8.4606E+16, 8.5362E+16,
     + 8.6123E+16, 8.6891E+16, 8.7666E+16, 8.8447E+16, 8.9234E+16,
     +  9.0029E+16, 9.0829E+16, 9.1637E+16, 9.2451E+16, 9.3272E+16,
     + 9.4100E+16, 9.4935E+16, 9.5776E+16, 9.6625E+16, 9.7481E+16,
     +  9.8344E+16, 9.9214E+16, 1.0009E+17, 1.0098E+17, 1.0187E+17,
     + 1.0277E+17, 1.0367E+17, 1.0459E+17, 1.0551E+17, 1.0644E+17,
     +  1.0738E+17, 1.0832E+17, 1.0927E+17, 1.1024E+17, 1.1120E+17,
     + 1.1218E+17, 1.1317E+17, 1.1416E+17, 1.1516E+17, 1.1617E+17,
     +  1.1719E+17, 1.1821E+17, 1.1925E+17, 1.2029E+17, 1.2134E+17,
     + 1.2240E+17, 1.2347E+17, 1.2455E+17, 1.2563E+17, 1.2673E+17,
     +  1.2783E+17, 1.2895E+17, 1.3007E+17, 1.3120E+17, 1.3234E+17,
     + 1.3349E+17, 1.3465E+17, 1.3582E+17, 1.3700E+17, 1.3819E+17,
     +  1.3938E+17, 1.4059E+17, 1.4181E+17, 1.4304E+17, 1.4427E+17,
     + 1.4552E+17, 1.4678E+17, 1.4805E+17, 1.4932E+17, 1.5061E+17,
     +  1.5191E+17, 1.5322E+17, 1.5454E+17, 1.5587E+17, 1.5721E+17,
     + 1.5856E+17, 1.5993E+17, 1.6130E+17, 1.6269E+17, 1.6408E+17,
     +  1.6549E+17, 1.6691E+17, 1.6834E+17, 1.6978E+17, 1.7123E+17,
     + 1.7270E+17, 1.7418E+17, 1.7567E+17, 1.7717E+17, 1.7868E+17,
     +  1.8020E+17, 1.8174E+17, 1.8329E+17, 1.8485E+17, 1.8643E+17,
     + 1.8801E+17, 1.8961E+17, 1.9123E+17, 1.9285E+17, 1.9449E+17,
     +  1.9614E+17, 1.9781E+17, 1.9948E+17, 2.0118E+17, 2.0288E+17,
     + 2.0460E+17, 2.0633E+17, 2.0808E+17, 2.0984E+17, 2.1161E+17,
     +  2.1340E+17, 2.1520E+17, 2.1702E+17, 2.1885E+17, 2.2069E+17,
     + 2.2255E+17, 2.2443E+17, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_H2S   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 3),QofT( 3,NT),Q(NT)
      data (Tmx(I),I=1,3)/6000.,3500.,3500./
      data xgj/ 1.,1.,4./
c...      H2S
c...        --       121
      data (QofT( 1,J),J=1,2408)/ 1.0000E+00, 1.0005E+00, 1.0153E+00,
     +  1.0858E+00, 1.2458E+00, 1.5022E+00, 1.8443E+00, 2.2565E+00,
     + 2.7237E+00, 3.2340E+00, 3.7784E+00, 4.3504E+00, 4.9455E+00,
     +  5.5606E+00, 6.1932E+00, 6.8417E+00, 7.5052E+00, 8.1827E+00,
     + 8.8736E+00, 9.5776E+00, 1.1024E+01, 1.2519E+01, 1.4064E+01,
     +  1.5656E+01, 1.7296E+01, 1.8984E+01, 2.0719E+01, 2.2500E+01,
     + 2.4326E+01, 2.6198E+01, 2.8114E+01, 3.0074E+01, 3.2077E+01,
     +  3.4122E+01, 3.6208E+01, 3.8335E+01, 4.0503E+01, 4.2709E+01,
     + 4.4955E+01, 4.7239E+01, 4.9560E+01, 5.1918E+01, 5.4313E+01,
     +  5.6744E+01, 5.9210E+01, 6.1711E+01, 6.4246E+01, 6.6815E+01,
     + 6.9418E+01, 7.2054E+01, 7.4723E+01, 7.7424E+01, 8.0158E+01,
     +  8.2922E+01, 8.5718E+01, 8.8545E+01, 9.1403E+01, 9.4290E+01,
     + 9.7208E+01, 1.0016E+02, 1.0313E+02, 1.0614E+02, 1.0917E+02,
     +  1.1224E+02, 1.1533E+02, 1.1845E+02, 1.2160E+02, 1.2477E+02,
     + 1.2797E+02, 1.3120E+02, 1.3446E+02, 1.3774E+02, 1.4105E+02,
     +  1.4439E+02, 1.4775E+02, 1.5114E+02, 1.5455E+02, 1.5799E+02,
     + 1.6145E+02, 1.6494E+02, 1.6846E+02, 1.7200E+02, 1.7556E+02,
     +  1.7915E+02, 1.8276E+02, 1.8640E+02, 1.9006E+02, 1.9375E+02,
     + 1.9746E+02, 2.0119E+02, 2.0495E+02, 2.0873E+02, 2.1253E+02,
     +  2.1635E+02, 2.2020E+02, 2.2408E+02, 2.2797E+02, 2.3189E+02,
     + 2.3583E+02, 2.3979E+02, 2.4377E+02, 2.4778E+02, 2.5181E+02,
     +  2.5586E+02, 2.5993E+02, 2.6403E+02, 2.6815E+02, 2.7229E+02,
     + 2.7645E+02, 2.8063E+02, 2.8483E+02, 2.8906E+02, 2.9330E+02,
     +  2.9757E+02, 3.0186E+02, 3.0617E+02, 3.1050E+02, 3.1485E+02,
     + 3.1922E+02, 3.2362E+02, 3.2803E+02, 3.3247E+02, 3.3692E+02,
     +  3.4140E+02, 3.4590E+02, 3.5042E+02, 3.5496E+02, 3.5952E+02,
     + 3.6410E+02, 3.6870E+02, 3.7332E+02, 3.7796E+02, 3.8262E+02,
     +  3.8731E+02, 3.9201E+02, 3.9673E+02, 4.0148E+02, 4.0624E+02,
     + 4.1103E+02, 4.1583E+02, 4.2066E+02, 4.2550E+02, 4.3037E+02,
     +  4.3525E+02, 4.4016E+02, 4.4509E+02, 4.5003E+02, 4.5500E+02,
     + 4.5999E+02, 4.6500E+02, 4.7003E+02, 4.7507E+02, 4.8014E+02,
     +  4.8523E+02, 4.9034E+02, 4.9547E+02, 5.0062E+02, 5.0579E+02,
     + 5.1098E+02, 5.1619E+02, 5.2143E+02, 5.2668E+02, 5.3195E+02,
     +  5.3724E+02, 5.4256E+02, 5.4789E+02, 5.5325E+02, 5.5862E+02,
     + 5.6402E+02, 5.6943E+02, 5.7487E+02, 5.8033E+02, 5.8580E+02,
     +  5.9130E+02, 5.9682E+02, 6.0236E+02, 6.0792E+02, 6.1351E+02,
     + 6.1911E+02, 6.2473E+02, 6.3038E+02, 6.3604E+02, 6.4173E+02,
     +  6.4744E+02, 6.5316E+02, 6.5891E+02, 6.6468E+02, 6.7048E+02,
     + 6.7629E+02, 6.8212E+02, 6.8798E+02, 6.9386E+02, 6.9976E+02,
     +  7.0567E+02, 7.1162E+02, 7.1758E+02, 7.2356E+02, 7.2957E+02,
     + 7.3560E+02, 7.4165E+02, 7.4772E+02, 7.5381E+02, 7.5992E+02,
     +  7.6606E+02, 7.7222E+02, 7.7840E+02, 7.8460E+02, 7.9083E+02,
     + 7.9707E+02, 8.0334E+02, 8.0963E+02, 8.1595E+02, 8.2228E+02,
     +  8.2864E+02, 8.3502E+02, 8.4142E+02, 8.4785E+02, 8.5430E+02,
     + 8.6077E+02, 8.6726E+02, 8.7378E+02, 8.8031E+02, 8.8688E+02,
     +  8.9346E+02, 9.0007E+02, 9.0670E+02, 9.1335E+02, 9.2003E+02,
     + 9.2673E+02, 9.3345E+02, 9.4020E+02, 9.4697E+02, 9.5376E+02,
     +  9.6058E+02, 9.6742E+02, 9.7429E+02, 9.8117E+02, 9.8809E+02,
     + 9.9502E+02, 1.0020E+03, 1.0090E+03, 1.0160E+03, 1.0230E+03,
     +  1.0301E+03, 1.0371E+03, 1.0442E+03, 1.0514E+03, 1.0585E+03,
     + 1.0657E+03, 1.0729E+03, 1.0801E+03, 1.0874E+03, 1.0947E+03,
     +  1.1020E+03, 1.1093E+03, 1.1166E+03, 1.1240E+03, 1.1314E+03,
     + 1.1388E+03, 1.1463E+03, 1.1538E+03, 1.1613E+03, 1.1688E+03,
     +  1.1763E+03, 1.1839E+03, 1.1915E+03, 1.1992E+03, 1.2068E+03,
     + 1.2145E+03, 1.2222E+03, 1.2299E+03, 1.2377E+03, 1.2455E+03,
     +  1.2533E+03, 1.2611E+03, 1.2690E+03, 1.2769E+03, 1.2848E+03,
     + 1.2927E+03, 1.3007E+03, 1.3087E+03, 1.3167E+03, 1.3248E+03,
     +  1.3329E+03, 1.3410E+03, 1.3491E+03, 1.3573E+03, 1.3655E+03,
     + 1.3737E+03, 1.3819E+03, 1.3902E+03, 1.3985E+03, 1.4068E+03,
     +  1.4152E+03, 1.4235E+03, 1.4319E+03, 1.4404E+03, 1.4488E+03,
     + 1.4573E+03, 1.4659E+03, 1.4744E+03, 1.4830E+03, 1.4916E+03,
     +  1.5002E+03, 1.5089E+03, 1.5176E+03, 1.5263E+03, 1.5351E+03,
     + 1.5438E+03, 1.5527E+03, 1.5615E+03, 1.5704E+03, 1.5793E+03,
     +  1.5882E+03, 1.5971E+03, 1.6061E+03, 1.6151E+03, 1.6242E+03,
     + 1.6333E+03, 1.6424E+03, 1.6515E+03, 1.6607E+03, 1.6699E+03,
     +  1.6791E+03, 1.6883E+03, 1.6976E+03, 1.7069E+03, 1.7163E+03,
     + 1.7257E+03, 1.7351E+03, 1.7445E+03, 1.7540E+03, 1.7635E+03,
     +  1.7730E+03, 1.7826E+03, 1.7922E+03, 1.8018E+03, 1.8115E+03,
     + 1.8212E+03, 1.8309E+03, 1.8407E+03, 1.8504E+03, 1.8603E+03,
     +  1.8701E+03, 1.8800E+03, 1.8899E+03, 1.8999E+03, 1.9099E+03,
     + 1.9199E+03, 1.9299E+03, 1.9400E+03, 1.9501E+03, 1.9603E+03,
     +  1.9704E+03, 1.9806E+03, 1.9909E+03, 2.0012E+03, 2.0115E+03,
     + 2.0218E+03, 2.0322E+03, 2.0426E+03, 2.0531E+03, 2.0636E+03,
     +  2.0741E+03, 2.0846E+03, 2.0952E+03, 2.1058E+03, 2.1165E+03,
     + 2.1272E+03, 2.1379E+03, 2.1487E+03, 2.1595E+03, 2.1703E+03,
     +  2.1812E+03, 2.1921E+03, 2.2030E+03, 2.2140E+03, 2.2250E+03,
     + 2.2360E+03, 2.2471E+03, 2.2582E+03, 2.2693E+03, 2.2805E+03,
     +  2.2918E+03, 2.3030E+03, 2.3143E+03, 2.3256E+03, 2.3370E+03,
     + 2.3484E+03, 2.3599E+03, 2.3713E+03, 2.3828E+03, 2.3944E+03,
     +  2.4060E+03, 2.4176E+03, 2.4293E+03, 2.4410E+03, 2.4527E+03,
     + 2.4645E+03, 2.4763E+03, 2.4882E+03, 2.5001E+03, 2.5120E+03,
     +  2.5240E+03, 2.5360E+03, 2.5480E+03, 2.5601E+03, 2.5722E+03,
     + 2.5844E+03, 2.5966E+03, 2.6088E+03, 2.6211E+03, 2.6334E+03,
     +  2.6458E+03, 2.6582E+03, 2.6706E+03, 2.6831E+03, 2.6956E+03,
     + 2.7082E+03, 2.7208E+03, 2.7334E+03, 2.7461E+03, 2.7588E+03,
     +  2.7716E+03, 2.7844E+03, 2.7972E+03, 2.8101E+03, 2.8230E+03,
     + 2.8360E+03, 2.8490E+03, 2.8620E+03, 2.8751E+03, 2.8883E+03,
     +  2.9014E+03, 2.9146E+03, 2.9279E+03, 2.9412E+03, 2.9545E+03,
     + 2.9679E+03, 2.9813E+03, 2.9948E+03, 3.0083E+03, 3.0219E+03,
     +  3.0355E+03, 3.0491E+03, 3.0628E+03, 3.0765E+03, 3.0903E+03,
     + 3.1041E+03, 3.1179E+03, 3.1318E+03, 3.1458E+03, 3.1598E+03,
     +  3.1738E+03, 3.1879E+03, 3.2020E+03, 3.2162E+03, 3.2304E+03,
     + 3.2446E+03, 3.2589E+03, 3.2733E+03, 3.2877E+03, 3.3021E+03,
     +  3.3166E+03, 3.3311E+03, 3.3457E+03, 3.3603E+03, 3.3750E+03,
     + 3.3897E+03, 3.4044E+03, 3.4192E+03, 3.4341E+03, 3.4490E+03,
     +  3.4639E+03, 3.4789E+03, 3.4939E+03, 3.5090E+03, 3.5241E+03,
     + 3.5393E+03, 3.5545E+03, 3.5698E+03, 3.5851E+03, 3.6005E+03,
     +  3.6159E+03, 3.6314E+03, 3.6469E+03, 3.6624E+03, 3.6780E+03,
     + 3.6937E+03, 3.7094E+03, 3.7251E+03, 3.7409E+03, 3.7568E+03,
     +  3.7727E+03, 3.7886E+03, 3.8046E+03, 3.8207E+03, 3.8368E+03,
     + 3.8529E+03, 3.8691E+03, 3.8854E+03, 3.9017E+03, 3.9180E+03,
     +  3.9344E+03, 3.9508E+03, 3.9673E+03, 3.9839E+03, 4.0005E+03,
     + 4.0171E+03, 4.0338E+03, 4.0506E+03, 4.0674E+03, 4.0843E+03,
     +  4.1012E+03, 4.1181E+03, 4.1351E+03, 4.1522E+03, 4.1693E+03,
     + 4.1865E+03, 4.2037E+03, 4.2210E+03, 4.2383E+03, 4.2557E+03,
     +  4.2731E+03, 4.2906E+03, 4.3081E+03, 4.3257E+03, 4.3434E+03,
     + 4.3611E+03, 4.3788E+03, 4.3966E+03, 4.4145E+03, 4.4324E+03,
     +  4.4504E+03, 4.4684E+03, 4.4865E+03, 4.5046E+03, 4.5228E+03,
     + 4.5411E+03, 4.5594E+03, 4.5777E+03, 4.5961E+03, 4.6146E+03,
     +  4.6331E+03, 4.6517E+03, 4.6703E+03, 4.6890E+03, 4.7078E+03,
     + 4.7266E+03, 4.7454E+03, 4.7644E+03, 4.7833E+03, 4.8024E+03,
     +  4.8215E+03, 4.8406E+03, 4.8598E+03, 4.8791E+03, 4.8984E+03,
     + 4.9178E+03, 4.9372E+03, 4.9567E+03, 4.9762E+03, 4.9959E+03,
     +  5.0155E+03, 5.0353E+03, 5.0550E+03, 5.0749E+03, 5.0948E+03,
     + 5.1148E+03, 5.1348E+03, 5.1549E+03, 5.1750E+03, 5.1952E+03,
     +  5.2155E+03, 5.2358E+03, 5.2562E+03, 5.2766E+03, 5.2971E+03,
     + 5.3177E+03, 5.3383E+03, 5.3590E+03, 5.3798E+03, 5.4006E+03,
     +  5.4214E+03, 5.4424E+03, 5.4634E+03, 5.4844E+03, 5.5056E+03,
     + 5.5267E+03, 5.5480E+03, 5.5693E+03, 5.5907E+03, 5.6121E+03,
     +  5.6336E+03, 5.6551E+03, 5.6768E+03, 5.6985E+03, 5.7202E+03,
     + 5.7420E+03, 5.7639E+03, 5.7858E+03, 5.8078E+03, 5.8299E+03,
     +  5.8521E+03, 5.8743E+03, 5.8965E+03, 5.9188E+03, 5.9412E+03,
     + 5.9637E+03, 5.9862E+03, 6.0088E+03, 6.0315E+03, 6.0542E+03,
     +  6.0770E+03, 6.0998E+03, 6.1228E+03, 6.1458E+03, 6.1688E+03,
     + 6.1919E+03, 6.2151E+03, 6.2384E+03, 6.2617E+03, 6.2851E+03,
     +  6.3085E+03, 6.3321E+03, 6.3557E+03, 6.3793E+03, 6.4031E+03,
     + 6.4269E+03, 6.4507E+03, 6.4747E+03, 6.4987E+03, 6.5227E+03,
     +  6.5469E+03, 6.5711E+03, 6.5954E+03, 6.6197E+03, 6.6442E+03,
     + 6.6687E+03, 6.6932E+03, 6.7178E+03, 6.7425E+03, 6.7673E+03,
     +  6.7922E+03, 6.8171E+03, 6.8421E+03, 6.8671E+03, 6.8923E+03,
     + 6.9175E+03, 6.9427E+03, 6.9681E+03, 6.9935E+03, 7.0190E+03,
     +  7.0445E+03, 7.0702E+03, 7.0959E+03, 7.1217E+03, 7.1475E+03,
     + 7.1734E+03, 7.1994E+03, 7.2255E+03, 7.2517E+03, 7.2779E+03,
     +  7.3042E+03, 7.3305E+03, 7.3570E+03, 7.3835E+03, 7.4101E+03,
     + 7.4368E+03, 7.4635E+03, 7.4903E+03, 7.5172E+03, 7.5442E+03,
     +  7.5712E+03, 7.5983E+03, 7.6255E+03, 7.6528E+03, 7.6802E+03,
     + 7.7076E+03, 7.7351E+03, 7.7627E+03, 7.7903E+03, 7.8181E+03,
     +  7.8459E+03, 7.8738E+03, 7.9017E+03, 7.9298E+03, 7.9579E+03,
     + 7.9861E+03, 8.0144E+03, 8.0427E+03, 8.0712E+03, 8.0997E+03,
     +  8.1283E+03, 8.1569E+03, 8.1857E+03, 8.2145E+03, 8.2434E+03,
     + 8.2724E+03, 8.3015E+03, 8.3307E+03, 8.3599E+03, 8.3892E+03,
     +  8.4186E+03, 8.4481E+03, 8.4776E+03, 8.5073E+03, 8.5370E+03,
     + 8.5668E+03, 8.5967E+03, 8.6266E+03, 8.6567E+03, 8.6868E+03,
     +  8.7170E+03, 8.7473E+03, 8.7777E+03, 8.8082E+03, 8.8387E+03,
     + 8.8693E+03, 8.9000E+03, 8.9308E+03, 8.9617E+03, 8.9927E+03,
     +  9.0237E+03, 9.0549E+03, 9.0861E+03, 9.1174E+03, 9.1488E+03,
     + 9.1802E+03, 9.2118E+03, 9.2434E+03, 9.2752E+03, 9.3070E+03,
     +  9.3389E+03, 9.3709E+03, 9.4029E+03, 9.4351E+03, 9.4673E+03,
     + 9.4997E+03, 9.5321E+03, 9.5646E+03, 9.5972E+03, 9.6299E+03,
     +  9.6627E+03, 9.6955E+03, 9.7285E+03, 9.7615E+03, 9.7946E+03,
     + 9.8278E+03, 9.8611E+03, 9.8945E+03, 9.9280E+03, 9.9616E+03,
     +  9.9952E+03, 1.0029E+04, 1.0063E+04, 1.0097E+04, 1.0131E+04,
     + 1.0165E+04, 1.0199E+04, 1.0233E+04, 1.0268E+04, 1.0302E+04,
     +  1.0337E+04, 1.0371E+04, 1.0406E+04, 1.0441E+04, 1.0476E+04,
     + 1.0511E+04, 1.0546E+04, 1.0581E+04, 1.0617E+04, 1.0652E+04,
     +  1.0688E+04, 1.0723E+04, 1.0759E+04, 1.0795E+04, 1.0831E+04,
     + 1.0866E+04, 1.0903E+04, 1.0939E+04, 1.0975E+04, 1.1011E+04,
     +  1.1048E+04, 1.1084E+04, 1.1121E+04, 1.1158E+04, 1.1194E+04,
     + 1.1231E+04, 1.1268E+04, 1.1306E+04, 1.1343E+04, 1.1380E+04,
     +  1.1417E+04, 1.1455E+04, 1.1493E+04, 1.1530E+04, 1.1568E+04,
     + 1.1606E+04, 1.1644E+04, 1.1682E+04, 1.1720E+04, 1.1758E+04,
     +  1.1797E+04, 1.1835E+04, 1.1874E+04, 1.1913E+04, 1.1951E+04,
     + 1.1990E+04, 1.2029E+04, 1.2068E+04, 1.2108E+04, 1.2147E+04,
     +  1.2186E+04, 1.2226E+04, 1.2265E+04, 1.2305E+04, 1.2345E+04,
     + 1.2385E+04, 1.2425E+04, 1.2465E+04, 1.2505E+04, 1.2545E+04,
     +  1.2586E+04, 1.2626E+04, 1.2667E+04, 1.2708E+04, 1.2748E+04,
     + 1.2789E+04, 1.2830E+04, 1.2872E+04, 1.2913E+04, 1.2954E+04,
     +  1.2996E+04, 1.3037E+04, 1.3079E+04, 1.3121E+04, 1.3163E+04,
     + 1.3205E+04, 1.3247E+04, 1.3289E+04, 1.3331E+04, 1.3374E+04,
     +  1.3416E+04, 1.3459E+04, 1.3501E+04, 1.3544E+04, 1.3587E+04,
     + 1.3630E+04, 1.3673E+04, 1.3717E+04, 1.3760E+04, 1.3804E+04,
     +  1.3847E+04, 1.3891E+04, 1.3935E+04, 1.3979E+04, 1.4023E+04,
     + 1.4067E+04, 1.4111E+04, 1.4156E+04, 1.4200E+04, 1.4245E+04,
     +  1.4289E+04, 1.4334E+04, 1.4379E+04, 1.4424E+04, 1.4469E+04,
     + 1.4515E+04, 1.4560E+04, 1.4605E+04, 1.4651E+04, 1.4697E+04,
     +  1.4743E+04, 1.4789E+04, 1.4835E+04, 1.4881E+04, 1.4927E+04,
     + 1.4973E+04, 1.5020E+04, 1.5067E+04, 1.5113E+04, 1.5160E+04,
     +  1.5207E+04, 1.5254E+04, 1.5302E+04, 1.5349E+04, 1.5396E+04,
     + 1.5444E+04, 1.5492E+04, 1.5539E+04, 1.5587E+04, 1.5635E+04,
     +  1.5683E+04, 1.5732E+04, 1.5780E+04, 1.5829E+04, 1.5877E+04,
     + 1.5926E+04, 1.5975E+04, 1.6024E+04, 1.6073E+04, 1.6122E+04,
     +  1.6171E+04, 1.6221E+04, 1.6270E+04, 1.6320E+04, 1.6370E+04,
     + 1.6420E+04, 1.6470E+04, 1.6520E+04, 1.6570E+04, 1.6621E+04,
     +  1.6671E+04, 1.6722E+04, 1.6773E+04, 1.6824E+04, 1.6875E+04,
     + 1.6926E+04, 1.6977E+04, 1.7029E+04, 1.7080E+04, 1.7132E+04,
     +  1.7184E+04, 1.7236E+04, 1.7288E+04, 1.7340E+04, 1.7392E+04,
     + 1.7444E+04, 1.7497E+04, 1.7550E+04, 1.7602E+04, 1.7655E+04,
     +  1.7708E+04, 1.7761E+04, 1.7815E+04, 1.7868E+04, 1.7922E+04,
     + 1.7975E+04, 1.8029E+04, 1.8083E+04, 1.8137E+04, 1.8191E+04,
     +  1.8246E+04, 1.8300E+04, 1.8355E+04, 1.8409E+04, 1.8464E+04,
     + 1.8519E+04, 1.8574E+04, 1.8629E+04, 1.8685E+04, 1.8740E+04,
     +  1.8796E+04, 1.8851E+04, 1.8907E+04, 1.8963E+04, 1.9019E+04,
     + 1.9076E+04, 1.9132E+04, 1.9189E+04, 1.9245E+04, 1.9302E+04,
     +  1.9359E+04, 1.9416E+04, 1.9473E+04, 1.9531E+04, 1.9588E+04,
     + 1.9646E+04, 1.9703E+04, 1.9761E+04, 1.9819E+04, 1.9877E+04,
     +  1.9936E+04, 1.9994E+04, 2.0053E+04, 2.0111E+04, 2.0170E+04,
     + 2.0229E+04, 2.0288E+04, 2.0347E+04, 2.0407E+04, 2.0466E+04,
     +  2.0526E+04, 2.0586E+04, 2.0646E+04, 2.0706E+04, 2.0766E+04,
     + 2.0826E+04, 2.0887E+04, 2.0947E+04, 2.1008E+04, 2.1069E+04,
     +  2.1130E+04, 2.1191E+04, 2.1252E+04, 2.1314E+04, 2.1375E+04,
     + 2.1437E+04, 2.1499E+04, 2.1561E+04, 2.1623E+04, 2.1685E+04,
     +  2.1748E+04, 2.1810E+04, 2.1873E+04, 2.1936E+04, 2.1999E+04,
     + 2.2062E+04, 2.2125E+04, 2.2189E+04, 2.2252E+04, 2.2316E+04,
     +  2.2380E+04, 2.2444E+04, 2.2508E+04, 2.2572E+04, 2.2637E+04,
     + 2.2702E+04, 2.2766E+04, 2.2831E+04, 2.2896E+04, 2.2961E+04,
     +  2.3027E+04, 2.3092E+04, 2.3158E+04, 2.3224E+04, 2.3290E+04,
     + 2.3356E+04, 2.3422E+04, 2.3488E+04, 2.3555E+04, 2.3621E+04,
     +  2.3688E+04, 2.3755E+04, 2.3822E+04, 2.3889E+04, 2.3957E+04,
     + 2.4024E+04, 2.4092E+04, 2.4160E+04, 2.4228E+04, 2.4296E+04,
     +  2.4365E+04, 2.4433E+04, 2.4502E+04, 2.4570E+04, 2.4639E+04,
     + 2.4708E+04, 2.4778E+04, 2.4847E+04, 2.4917E+04, 2.4986E+04,
     +  2.5056E+04, 2.5126E+04, 2.5196E+04, 2.5267E+04, 2.5337E+04,
     + 2.5408E+04, 2.5478E+04, 2.5549E+04, 2.5621E+04, 2.5692E+04,
     +  2.5763E+04, 2.5835E+04, 2.5906E+04, 2.5978E+04, 2.6050E+04,
     + 2.6123E+04, 2.6195E+04, 2.6267E+04, 2.6340E+04, 2.6413E+04,
     +  2.6486E+04, 2.6559E+04, 2.6632E+04, 2.6706E+04, 2.6779E+04,
     + 2.6853E+04, 2.6927E+04, 2.7001E+04, 2.7075E+04, 2.7150E+04,
     +  2.7224E+04, 2.7299E+04, 2.7374E+04, 2.7449E+04, 2.7524E+04,
     + 2.7600E+04, 2.7675E+04, 2.7751E+04, 2.7827E+04, 2.7903E+04,
     +  2.7979E+04, 2.8056E+04, 2.8132E+04, 2.8209E+04, 2.8286E+04,
     + 2.8363E+04, 2.8440E+04, 2.8517E+04, 2.8595E+04, 2.8673E+04,
     +  2.8750E+04, 2.8828E+04, 2.8907E+04, 2.8985E+04, 2.9063E+04,
     + 2.9142E+04, 2.9221E+04, 2.9300E+04, 2.9379E+04, 2.9459E+04,
     +  2.9538E+04, 2.9618E+04, 2.9698E+04, 2.9778E+04, 2.9858E+04,
     + 2.9938E+04, 3.0019E+04, 3.0100E+04, 3.0181E+04, 3.0262E+04,
     +  3.0343E+04, 3.0424E+04, 3.0506E+04, 3.0588E+04, 3.0670E+04,
     + 3.0752E+04, 3.0834E+04, 3.0916E+04, 3.0999E+04, 3.1082E+04,
     +  3.1165E+04, 3.1248E+04, 3.1331E+04, 3.1415E+04, 3.1498E+04,
     + 3.1582E+04, 3.1666E+04, 3.1750E+04, 3.1835E+04, 3.1919E+04,
     +  3.2004E+04, 3.2089E+04, 3.2174E+04, 3.2259E+04, 3.2344E+04,
     + 3.2430E+04, 3.2516E+04, 3.2602E+04, 3.2688E+04, 3.2774E+04,
     +  3.2861E+04, 3.2947E+04, 3.3034E+04, 3.3121E+04, 3.3208E+04,
     + 3.3296E+04, 3.3383E+04, 3.3471E+04, 3.3559E+04, 3.3647E+04,
     +  3.3735E+04, 3.3824E+04, 3.3913E+04, 3.4001E+04, 3.4090E+04,
     + 3.4180E+04, 3.4269E+04, 3.4358E+04, 3.4448E+04, 3.4538E+04,
     +  3.4628E+04, 3.4719E+04, 3.4809E+04, 3.4900E+04, 3.4990E+04,
     + 3.5082E+04, 3.5173E+04, 3.5264E+04, 3.5356E+04, 3.5447E+04,
     +  3.5539E+04, 3.5632E+04, 3.5724E+04, 3.5816E+04, 3.5909E+04,
     + 3.6002E+04, 3.6095E+04, 3.6188E+04, 3.6282E+04, 3.6375E+04,
     +  3.6469E+04, 3.6563E+04, 3.6657E+04, 3.6752E+04, 3.6846E+04,
     + 3.6941E+04, 3.7036E+04, 3.7131E+04, 3.7227E+04, 3.7322E+04,
     +  3.7418E+04, 3.7514E+04, 3.7610E+04, 3.7706E+04, 3.7803E+04,
     + 3.7899E+04, 3.7996E+04, 3.8093E+04, 3.8191E+04, 3.8288E+04,
     +  3.8386E+04, 3.8484E+04, 3.8582E+04, 3.8680E+04, 3.8778E+04,
     + 3.8877E+04, 3.8976E+04, 3.9075E+04, 3.9174E+04, 3.9273E+04,
     +  3.9373E+04, 3.9473E+04, 3.9573E+04, 3.9673E+04, 3.9773E+04,
     + 3.9874E+04, 3.9974E+04, 4.0075E+04, 4.0177E+04, 4.0278E+04,
     +  4.0380E+04, 4.0481E+04, 4.0583E+04, 4.0685E+04, 4.0788E+04,
     + 4.0890E+04, 4.0993E+04, 4.1096E+04, 4.1199E+04, 4.1303E+04,
     +  4.1406E+04, 4.1510E+04, 4.1614E+04, 4.1718E+04, 4.1822E+04,
     + 4.1927E+04, 4.2032E+04, 4.2137E+04, 4.2242E+04, 4.2347E+04,
     +  4.2453E+04, 4.2559E+04, 4.2665E+04, 4.2771E+04, 4.2877E+04,
     + 4.2984E+04, 4.3091E+04, 4.3198E+04, 4.3305E+04, 4.3412E+04,
     +  4.3520E+04, 4.3628E+04, 4.3736E+04, 4.3844E+04, 4.3953E+04,
     + 4.4061E+04, 4.4170E+04, 4.4279E+04, 4.4388E+04, 4.4498E+04,
     +  4.4608E+04, 4.4718E+04, 4.4828E+04, 4.4938E+04, 4.5048E+04,
     + 4.5159E+04, 4.5270E+04, 4.5381E+04, 4.5493E+04, 4.5604E+04,
     +  4.5716E+04, 4.5828E+04, 4.5940E+04, 4.6053E+04, 4.6165E+04,
     + 4.6278E+04, 4.6391E+04, 4.6505E+04, 4.6618E+04, 4.6732E+04,
     +  4.6846E+04, 4.6960E+04, 4.7074E+04, 4.7189E+04, 4.7303E+04,
     + 4.7418E+04, 4.7534E+04, 4.7649E+04, 4.7765E+04, 4.7881E+04,
     +  4.7997E+04, 4.8113E+04, 4.8229E+04, 4.8346E+04, 4.8463E+04,
     + 4.8580E+04, 4.8698E+04, 4.8815E+04, 4.8933E+04, 4.9051E+04,
     +  4.9169E+04, 4.9288E+04, 4.9406E+04, 4.9525E+04, 4.9644E+04,
     + 4.9764E+04, 4.9883E+04, 5.0003E+04, 5.0123E+04, 5.0243E+04,
     +  5.0363E+04, 5.0484E+04, 5.0605E+04, 5.0726E+04, 5.0847E+04,
     + 5.0969E+04, 5.1091E+04, 5.1213E+04, 5.1335E+04, 5.1457E+04,
     +  5.1580E+04, 5.1703E+04, 5.1826E+04, 5.1949E+04, 5.2073E+04,
     + 5.2196E+04, 5.2320E+04, 5.2445E+04, 5.2569E+04, 5.2694E+04,
     +  5.2819E+04, 5.2944E+04, 5.3069E+04, 5.3195E+04, 5.3320E+04,
     + 5.3446E+04, 5.3573E+04, 5.3699E+04, 5.3826E+04, 5.3953E+04,
     +  5.4080E+04, 5.4207E+04, 5.4335E+04, 5.4463E+04, 5.4591E+04,
     + 5.4719E+04, 5.4847E+04, 5.4976E+04, 5.5105E+04, 5.5234E+04,
     +  5.5364E+04, 5.5493E+04, 5.5623E+04, 5.5753E+04, 5.5884E+04,
     + 5.6014E+04, 5.6145E+04, 5.6276E+04, 5.6408E+04, 5.6539E+04,
     +  5.6671E+04, 5.6803E+04, 5.6935E+04, 5.7067E+04, 5.7200E+04,
     + 5.7333E+04, 5.7466E+04, 5.7600E+04, 5.7733E+04, 5.7867E+04,
     +  5.8001E+04, 5.8135E+04, 5.8270E+04, 5.8405E+04, 5.8540E+04,
     + 5.8675E+04, 5.8810E+04, 5.8946E+04, 5.9082E+04, 5.9218E+04,
     +  5.9355E+04, 5.9492E+04, 5.9628E+04, 5.9766E+04, 5.9903E+04,
     + 6.0041E+04, 6.0178E+04, 6.0317E+04, 6.0455E+04, 6.0594E+04,
     +  6.0732E+04, 6.0871E+04, 6.1011E+04, 6.1150E+04, 6.1290E+04,
     + 6.1430E+04, 6.1570E+04, 6.1711E+04, 6.1852E+04, 6.1993E+04,
     +  6.2134E+04, 6.2275E+04, 6.2417E+04, 6.2559E+04, 6.2701E+04,
     + 6.2844E+04, 6.2986E+04, 6.3129E+04, 6.3272E+04, 6.3416E+04,
     +  6.3559E+04, 6.3703E+04, 6.3847E+04, 6.3992E+04, 6.4136E+04,
     + 6.4281E+04, 6.4426E+04, 6.4572E+04, 6.4717E+04, 6.4863E+04,
     +  6.5009E+04, 6.5156E+04, 6.5302E+04, 6.5449E+04, 6.5596E+04,
     + 6.5744E+04, 6.5891E+04, 6.6039E+04, 6.6187E+04, 6.6336E+04,
     +  6.6484E+04, 6.6633E+04, 6.6782E+04, 6.6931E+04, 6.7081E+04,
     + 6.7231E+04, 6.7381E+04, 6.7531E+04, 6.7682E+04, 6.7833E+04,
     +  6.7984E+04, 6.8135E+04, 6.8287E+04, 6.8438E+04, 6.8590E+04,
     + 6.8743E+04, 6.8895E+04, 6.9048E+04, 6.9201E+04, 6.9355E+04,
     +  6.9508E+04, 6.9662E+04, 6.9816E+04, 6.9970E+04, 7.0125E+04,
     + 7.0280E+04, 7.0435E+04, 7.0590E+04, 7.0746E+04, 7.0902E+04,
     +  7.1058E+04, 7.1214E+04, 7.1371E+04, 7.1528E+04, 7.1685E+04,
     + 7.1843E+04, 7.2000E+04, 7.2158E+04, 7.2316E+04, 7.2475E+04,
     +  7.2633E+04, 7.2792E+04, 7.2952E+04, 7.3111E+04, 7.3271E+04,
     + 7.3431E+04, 7.3591E+04, 7.3752E+04, 7.3912E+04, 7.4073E+04,
     +  7.4235E+04, 7.4396E+04, 7.4558E+04, 7.4720E+04, 7.4882E+04,
     + 7.5045E+04, 7.5208E+04, 7.5371E+04, 7.5534E+04, 7.5698E+04,
     +  7.5862E+04, 7.6026E+04, 7.6190E+04, 7.6355E+04, 7.6520E+04,
     + 7.6685E+04, 7.6850E+04, 7.7016E+04, 7.7182E+04, 7.7348E+04,
     +  7.7515E+04, 7.7682E+04, 7.7849E+04, 7.8016E+04, 7.8184E+04,
     + 7.8351E+04, 7.8520E+04, 7.8688E+04, 7.8856E+04, 7.9025E+04,
     +  7.9195E+04, 7.9364E+04, 7.9534E+04, 7.9704E+04, 7.9874E+04,
     + 8.0044E+04, 8.0215E+04, 8.0386E+04, 8.0557E+04, 8.0729E+04,
     +  8.0901E+04, 8.1073E+04, 8.1245E+04, 8.1418E+04, 8.1591E+04,
     + 8.1764E+04, 8.1937E+04, 8.2111E+04, 8.2285E+04, 8.2459E+04,
     +  8.2634E+04, 8.2808E+04, 8.2983E+04, 8.3159E+04, 8.3334E+04,
     + 8.3510E+04, 8.3686E+04, 8.3863E+04, 8.4039E+04, 8.4216E+04,
     +  8.4393E+04, 8.4571E+04, 8.4749E+04, 8.4927E+04, 8.5105E+04,
     + 8.5284E+04, 8.5462E+04, 8.5642E+04, 8.5821E+04, 8.6001E+04,
     +  8.6180E+04, 8.6361E+04, 8.6541E+04, 8.6722E+04, 8.6903E+04,
     + 8.7084E+04, 8.7266E+04, 8.7448E+04, 8.7630E+04, 8.7812E+04,
     +  8.7995E+04, 8.8178E+04, 8.8361E+04, 8.8545E+04, 8.8728E+04,
     + 8.8912E+04, 8.9097E+04, 8.9281E+04, 8.9466E+04, 8.9651E+04,
     +  8.9837E+04, 9.0023E+04, 9.0209E+04, 9.0395E+04, 9.0581E+04,
     + 9.0768E+04, 9.0955E+04, 9.1143E+04, 9.1330E+04, 9.1518E+04,
     +  9.1707E+04, 9.1895E+04, 9.2084E+04, 9.2273E+04, 9.2462E+04,
     + 9.2652E+04, 9.2842E+04, 9.3032E+04, 9.3222E+04, 9.3413E+04,
     +  9.3604E+04, 9.3795E+04, 9.3987E+04, 9.4179E+04, 9.4371E+04,
     + 9.4564E+04, 9.4756E+04, 9.4949E+04, 9.5142E+04, 9.5336E+04,
     +  9.5530E+04, 9.5724E+04, 9.5918E+04, 9.6113E+04, 9.6308E+04,
     + 9.6503E+04, 9.6699E+04, 9.6895E+04, 9.7091E+04, 9.7287E+04,
     +  9.7484E+04, 9.7681E+04, 9.7878E+04, 9.8076E+04, 9.8273E+04,
     + 9.8472E+04, 9.8670E+04, 9.8869E+04, 9.9068E+04, 9.9267E+04,
     +  9.9466E+04, 9.9666E+04, 9.9866E+04, 1.0007E+05, 1.0027E+05,
     + 1.0047E+05, 1.0067E+05, 1.0087E+05, 1.0107E+05, 1.0128E+05,
     +  1.0148E+05, 1.0168E+05, 1.0188E+05, 1.0209E+05, 1.0229E+05,
     + 1.0249E+05, 1.0270E+05, 1.0290E+05, 1.0311E+05, 1.0331E+05,
     +  1.0352E+05, 1.0372E+05, 1.0393E+05, 1.0414E+05, 1.0434E+05,
     + 1.0455E+05, 1.0476E+05, 1.0496E+05, 1.0517E+05, 1.0538E+05,
     +  1.0559E+05, 1.0579E+05, 1.0600E+05, 1.0621E+05, 1.0642E+05,
     + 1.0663E+05, 1.0684E+05, 1.0705E+05, 1.0726E+05, 1.0747E+05,
     +  1.0768E+05, 1.0790E+05, 1.0811E+05, 1.0832E+05, 1.0853E+05,
     + 1.0874E+05, 1.0896E+05, 1.0917E+05, 1.0938E+05, 1.0960E+05,
     +  1.0981E+05, 1.1003E+05, 1.1024E+05, 1.1046E+05, 1.1067E+05,
     + 1.1089E+05, 1.1110E+05, 1.1132E+05, 1.1153E+05, 1.1175E+05,
     +  1.1197E+05, 1.1219E+05, 1.1240E+05, 1.1262E+05, 1.1284E+05,
     + 1.1306E+05, 1.1328E+05, 1.1350E+05, 1.1372E+05, 1.1394E+05,
     +  1.1416E+05, 1.1438E+05, 1.1460E+05, 1.1482E+05, 1.1504E+05,
     + 1.1526E+05, 1.1548E+05, 1.1570E+05, 1.1593E+05, 1.1615E+05,
     +  1.1637E+05, 1.1659E+05, 1.1682E+05, 1.1704E+05, 1.1727E+05,
     + 1.1749E+05, 1.1772E+05, 1.1794E+05, 1.1817E+05, 1.1839E+05,
     +  1.1862E+05, 1.1884E+05, 1.1907E+05, 1.1930E+05, 1.1952E+05,
     + 1.1975E+05, 1.1998E+05, 1.2021E+05, 1.2044E+05, 1.2066E+05,
     +  1.2089E+05, 1.2112E+05, 1.2135E+05, 1.2158E+05, 1.2181E+05,
     + 1.2204E+05, 1.2227E+05, 1.2251E+05, 1.2274E+05, 1.2297E+05,
     +  1.2320E+05, 1.2343E+05, 1.2367E+05, 1.2390E+05, 1.2413E+05,
     + 1.2436E+05, 1.2460E+05, 1.2483E+05, 1.2507E+05, 1.2530E+05,
     +  1.2554E+05, 1.2577E+05, 1.2601E+05, 1.2624E+05, 1.2648E+05,
     + 1.2672E+05, 1.2695E+05, 1.2719E+05, 1.2743E+05, 1.2767E+05,
     +  1.2790E+05, 1.2814E+05, 1.2838E+05, 1.2862E+05, 1.2886E+05,
     + 1.2910E+05, 1.2934E+05, 1.2958E+05, 1.2982E+05, 1.3006E+05,
     +  1.3030E+05, 1.3054E+05, 1.3079E+05, 1.3103E+05, 1.3127E+05,
     + 1.3151E+05, 1.3176E+05, 1.3200E+05, 1.3224E+05, 1.3249E+05,
     +  1.3273E+05, 1.3298E+05, 1.3322E+05, 1.3347E+05, 1.3371E+05,
     + 1.3396E+05, 1.3420E+05, 1.3445E+05, 1.3470E+05, 1.3494E+05,
     +  1.3519E+05, 1.3544E+05, 1.3569E+05, 1.3593E+05, 1.3618E+05,
     + 1.3643E+05, 1.3668E+05, 1.3693E+05, 1.3718E+05, 1.3743E+05,
     +  1.3768E+05, 1.3793E+05, 1.3818E+05, 1.3843E+05, 1.3869E+05,
     + 1.3894E+05, 1.3919E+05, 1.3944E+05, 1.3970E+05, 1.3995E+05,
     +  1.4020E+05, 1.4046E+05, 1.4071E+05, 1.4097E+05, 1.4122E+05,
     + 1.4147E+05, 1.4173E+05, 1.4199E+05, 1.4224E+05, 1.4250E+05,
     +  1.4276E+05, 1.4301E+05, 1.4327E+05, 1.4353E+05, 1.4379E+05,
     + 1.4404E+05, 1.4430E+05, 1.4456E+05, 1.4482E+05, 1.4508E+05,
     +  1.4534E+05, 1.4560E+05, 1.4586E+05, 1.4612E+05, 1.4638E+05,
     + 1.4664E+05, 1.4691E+05, 1.4717E+05, 1.4743E+05, 1.4769E+05,
     +  1.4796E+05, 1.4822E+05, 1.4848E+05, 1.4875E+05, 1.4901E+05,
     + 1.4927E+05, 1.4954E+05, 1.4980E+05, 1.5007E+05, 1.5034E+05,
     +  1.5060E+05, 1.5087E+05, 1.5114E+05, 1.5140E+05, 1.5167E+05,
     + 1.5194E+05, 1.5221E+05, 1.5247E+05, 1.5274E+05, 1.5301E+05,
     +  1.5328E+05, 1.5355E+05, 1.5382E+05, 1.5409E+05, 1.5436E+05,
     + 1.5463E+05, 1.5490E+05, 1.5518E+05, 1.5545E+05, 1.5572E+05,
     +  1.5599E+05, 1.5626E+05, 1.5654E+05, 1.5681E+05, 1.5708E+05,
     + 1.5736E+05, 1.5763E+05, 1.5791E+05, 1.5818E+05, 1.5846E+05,
     +  1.5873E+05, 1.5901E+05, 1.5929E+05, 1.5956E+05, 1.5984E+05,
     + 1.6012E+05, 1.6040E+05, 1.6067E+05, 1.6095E+05, 1.6123E+05,
     +  1.6151E+05, 1.6179E+05, 1.6207E+05, 1.6235E+05, 1.6263E+05,
     + 1.6291E+05, 1.6319E+05, 1.6347E+05, 1.6375E+05, 1.6403E+05,
     +  1.6432E+05, 1.6460E+05, 1.6488E+05, 1.6516E+05, 1.6545E+05,
     + 1.6573E+05, 1.6601E+05, 1.6630E+05, 1.6658E+05, 1.6687E+05,
     +  1.6715E+05, 1.6744E+05, 1.6773E+05, 1.6801E+05, 1.6830E+05,
     + 1.6858E+05, 1.6887E+05, 1.6916E+05, 1.6945E+05, 1.6974E+05,
     +  1.7002E+05, 1.7031E+05, 1.7060E+05, 1.7089E+05, 1.7118E+05,
     + 1.7147E+05, 1.7176E+05, 1.7205E+05, 1.7234E+05, 1.7264E+05,
     +  1.7293E+05, 1.7322E+05, 1.7351E+05, 1.7380E+05, 1.7410E+05,
     + 1.7439E+05, 1.7468E+05, 1.7498E+05, 1.7527E+05, 1.7557E+05,
     +  1.7586E+05, 1.7616E+05, 1.7645E+05, 1.7675E+05, 1.7705E+05,
     + 1.7734E+05, 1.7764E+05, 1.7794E+05, 1.7823E+05, 1.7853E+05,
     +  1.7883E+05, 1.7913E+05, 1.7943E+05, 1.7973E+05, 1.8003E+05,
     + 1.8033E+05, 1.8063E+05, 1.8093E+05, 1.8123E+05, 1.8153E+05,
     +  1.8183E+05, 1.8213E+05, 1.8243E+05, 1.8274E+05, 1.8304E+05,
     + 1.8334E+05, 1.8365E+05, 1.8395E+05, 1.8425E+05, 1.8456E+05,
     +  1.8486E+05, 1.8517E+05, 1.8547E+05, 1.8578E+05, 1.8609E+05,
     + 1.8639E+05, 1.8670E+05, 1.8701E+05, 1.8731E+05, 1.8762E+05,
     +  1.8793E+05, 1.8824E+05, 1.8855E+05, 1.8885E+05, 1.8916E+05,
     + 1.8947E+05, 1.8978E+05, 1.9009E+05, 1.9040E+05, 1.9072E+05,
     +  1.9103E+05, 1.9134E+05, 1.9165E+05, 1.9196E+05, 1.9228E+05,
     + 1.9259E+05, 1.9290E+05, 1.9321E+05, 1.9353E+05, 1.9384E+05,
     +  1.9416E+05, 1.9447E+05, 1.9479E+05, 1.9510E+05, 1.9542E+05,
     + 1.9574E+05, 1.9605E+05, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       141
      data (QofT( 2,J),J=1,2408)/ 1.0000E+00, 1.0007E+00, 1.0198E+00,
     +  1.1130E+00, 1.3260E+00, 1.6661E+00, 2.1162E+00, 2.6514E+00,
     + 3.2478E+00, 3.8862E+00, 4.4464E+00, 4.9724E+00, 5.5188E+00,
     +  6.0848E+00, 6.6696E+00, 7.2729E+00, 7.8939E+00, 8.5321E+00,
     + 9.1872E+00, 9.8587E+00, 1.1249E+01, 1.2701E+01, 1.4211E+01,
     +  1.5777E+01, 1.7398E+01, 1.9072E+01, 2.0796E+01, 2.2570E+01,
     + 2.4392E+01, 2.6261E+01, 2.8176E+01, 3.0135E+01, 3.2138E+01,
     +  3.4184E+01, 3.6272E+01, 3.8400E+01, 4.0569E+01, 4.2777E+01,
     + 4.5025E+01, 4.7310E+01, 4.9633E+01, 5.1992E+01, 5.4389E+01,
     +  5.6820E+01, 5.9288E+01, 6.1790E+01, 6.4326E+01, 6.6896E+01,
     + 6.9499E+01, 7.2136E+01, 7.4805E+01, 7.7506E+01, 8.0239E+01,
     +  8.3004E+01, 8.5799E+01, 8.8626E+01, 9.1482E+01, 9.4369E+01,
     + 9.7286E+01, 1.0023E+02, 1.0321E+02, 1.0621E+02, 1.0924E+02,
     +  1.1231E+02, 1.1539E+02, 1.1851E+02, 1.2166E+02, 1.2483E+02,
     + 1.2803E+02, 1.3125E+02, 1.3451E+02, 1.3779E+02, 1.4109E+02,
     +  1.4442E+02, 1.4778E+02, 1.5116E+02, 1.5457E+02, 1.5800E+02,
     + 1.6146E+02, 1.6495E+02, 1.6846E+02, 1.7199E+02, 1.7555E+02,
     +  1.7913E+02, 1.8273E+02, 1.8636E+02, 1.9002E+02, 1.9369E+02,
     + 1.9739E+02, 2.0112E+02, 2.0487E+02, 2.0864E+02, 2.1243E+02,
     +  2.1624E+02, 2.2008E+02, 2.2394E+02, 2.2783E+02, 2.3173E+02,
     + 2.3566E+02, 2.3961E+02, 2.4359E+02, 2.4758E+02, 2.5160E+02,
     +  2.5564E+02, 2.5970E+02, 2.6378E+02, 2.6788E+02, 2.7201E+02,
     + 2.7615E+02, 2.8032E+02, 2.8451E+02, 2.8872E+02, 2.9295E+02,
     +  2.9720E+02, 3.0147E+02, 3.0576E+02, 3.1007E+02, 3.1441E+02,
     + 3.1877E+02, 3.2314E+02, 3.2754E+02, 3.3195E+02, 3.3639E+02,
     +  3.4085E+02, 3.4532E+02, 3.4982E+02, 3.5434E+02, 3.5888E+02,
     + 3.6344E+02, 3.6802E+02, 3.7262E+02, 3.7724E+02, 3.8188E+02,
     +  3.8654E+02, 3.9122E+02, 3.9592E+02, 4.0063E+02, 4.0537E+02,
     + 4.1014E+02, 4.1491E+02, 4.1971E+02, 4.2453E+02, 4.2937E+02,
     +  4.3423E+02, 4.3911E+02, 4.4401E+02, 4.4893E+02, 4.5387E+02,
     + 4.5883E+02, 4.6381E+02, 4.6880E+02, 4.7382E+02, 4.7886E+02,
     +  4.8392E+02, 4.8900E+02, 4.9410E+02, 4.9922E+02, 5.0435E+02,
     + 5.0951E+02, 5.1469E+02, 5.1989E+02, 5.2511E+02, 5.3035E+02,
     +  5.3560E+02, 5.4088E+02, 5.4618E+02, 5.5150E+02, 5.5684E+02,
     + 5.6220E+02, 5.6758E+02, 5.7298E+02, 5.7840E+02, 5.8384E+02,
     +  5.8930E+02, 5.9478E+02, 6.0028E+02, 6.0580E+02, 6.1134E+02,
     + 6.1690E+02, 6.2249E+02, 6.2809E+02, 6.3371E+02, 6.3936E+02,
     +  6.4502E+02, 6.5071E+02, 6.5642E+02, 6.6214E+02, 6.6789E+02,
     + 6.7366E+02, 6.7945E+02, 6.8526E+02, 6.9109E+02, 6.9694E+02,
     +  7.0281E+02, 7.0871E+02, 7.1462E+02, 7.2056E+02, 7.2652E+02,
     + 7.3250E+02, 7.3849E+02, 7.4451E+02, 7.5056E+02, 7.5663E+02,
     +  7.6271E+02, 7.6882E+02, 7.7495E+02, 7.8109E+02, 7.8727E+02,
     + 7.9346E+02, 7.9967E+02, 8.0591E+02, 8.1217E+02, 8.1845E+02,
     +  8.2476E+02, 8.3107E+02, 8.3742E+02, 8.4379E+02, 8.5019E+02,
     + 8.5660E+02, 8.6303E+02, 8.6949E+02, 8.7597E+02, 8.8247E+02,
     +  8.8899E+02, 8.9554E+02, 9.0211E+02, 9.0870E+02, 9.1532E+02,
     + 9.2196E+02, 9.2862E+02, 9.3530E+02, 9.4201E+02, 9.4874E+02,
     +  9.5549E+02, 9.6227E+02, 9.6907E+02, 9.7588E+02, 9.8273E+02,
     + 9.8960E+02, 9.9649E+02, 1.0034E+03, 1.0103E+03, 1.0173E+03,
     +  1.0243E+03, 1.0313E+03, 1.0383E+03, 1.0454E+03, 1.0525E+03,
     + 1.0596E+03, 1.0667E+03, 1.0739E+03, 1.0810E+03, 1.0882E+03,
     +  1.0955E+03, 1.1027E+03, 1.1100E+03, 1.1173E+03, 1.1246E+03,
     + 1.1320E+03, 1.1393E+03, 1.1467E+03, 1.1542E+03, 1.1616E+03,
     +  1.1691E+03, 1.1766E+03, 1.1841E+03, 1.1916E+03, 1.1992E+03,
     + 1.2068E+03, 1.2144E+03, 1.2221E+03, 1.2298E+03, 1.2375E+03,
     +  1.2452E+03, 1.2529E+03, 1.2607E+03, 1.2685E+03, 1.2763E+03,
     + 1.2842E+03, 1.2921E+03, 1.3000E+03, 1.3079E+03, 1.3159E+03,
     +  1.3239E+03, 1.3319E+03, 1.3399E+03, 1.3480E+03, 1.3561E+03,
     + 1.3642E+03, 1.3724E+03, 1.3805E+03, 1.3887E+03, 1.3970E+03,
     +  1.4052E+03, 1.4135E+03, 1.4218E+03, 1.4301E+03, 1.4385E+03,
     + 1.4469E+03, 1.4553E+03, 1.4638E+03, 1.4722E+03, 1.4807E+03,
     +  1.4893E+03, 1.4978E+03, 1.5064E+03, 1.5150E+03, 1.5237E+03,
     + 1.5323E+03, 1.5410E+03, 1.5498E+03, 1.5585E+03, 1.5673E+03,
     +  1.5761E+03, 1.5850E+03, 1.5939E+03, 1.6027E+03, 1.6117E+03,
     + 1.6206E+03, 1.6296E+03, 1.6387E+03, 1.6477E+03, 1.6568E+03,
     +  1.6659E+03, 1.6750E+03, 1.6842E+03, 1.6934E+03, 1.7026E+03,
     + 1.7119E+03, 1.7212E+03, 1.7305E+03, 1.7398E+03, 1.7492E+03,
     +  1.7586E+03, 1.7680E+03, 1.7775E+03, 1.7870E+03, 1.7965E+03,
     + 1.8061E+03, 1.8157E+03, 1.8253E+03, 1.8350E+03, 1.8446E+03,
     +  1.8544E+03, 1.8641E+03, 1.8739E+03, 1.8837E+03, 1.8935E+03,
     + 1.9034E+03, 1.9133E+03, 1.9233E+03, 1.9332E+03, 1.9433E+03,
     +  1.9533E+03, 1.9634E+03, 1.9735E+03, 1.9836E+03, 1.9937E+03,
     + 2.0040E+03, 2.0142E+03, 2.0244E+03, 2.0348E+03, 2.0451E+03,
     +  2.0554E+03, 2.0658E+03, 2.0763E+03, 2.0867E+03, 2.0972E+03,
     + 2.1078E+03, 2.1183E+03, 2.1289E+03, 2.1396E+03, 2.1502E+03,
     +  2.1609E+03, 2.1717E+03, 2.1824E+03, 2.1932E+03, 2.2041E+03,
     + 2.2149E+03, 2.2259E+03, 2.2368E+03, 2.2478E+03, 2.2588E+03,
     +  2.2698E+03, 2.2809E+03, 2.2920E+03, 2.3032E+03, 2.3144E+03,
     + 2.3256E+03, 2.3369E+03, 2.3482E+03, 2.3595E+03, 2.3709E+03,
     +  2.3823E+03, 2.3937E+03, 2.4052E+03, 2.4167E+03, 2.4283E+03,
     + 2.4398E+03, 2.4515E+03, 2.4631E+03, 2.4748E+03, 2.4866E+03,
     +  2.4983E+03, 2.5101E+03, 2.5220E+03, 2.5339E+03, 2.5458E+03,
     + 2.5578E+03, 2.5698E+03, 2.5818E+03, 2.5939E+03, 2.6060E+03,
     +  2.6181E+03, 2.6303E+03, 2.6425E+03, 2.6548E+03, 2.6671E+03,
     + 2.6794E+03, 2.6918E+03, 2.7043E+03, 2.7167E+03, 2.7292E+03,
     +  2.7417E+03, 2.7543E+03, 2.7669E+03, 2.7796E+03, 2.7923E+03,
     + 2.8050E+03, 2.8178E+03, 2.8306E+03, 2.8435E+03, 2.8564E+03,
     +  2.8693E+03, 2.8823E+03, 2.8953E+03, 2.9084E+03, 2.9214E+03,
     + 2.9346E+03, 2.9478E+03, 2.9610E+03, 2.9742E+03, 2.9876E+03,
     +  3.0009E+03, 3.0143E+03, 3.0277E+03, 3.0412E+03, 3.0547E+03,
     + 3.0683E+03, 3.0818E+03, 3.0955E+03, 3.1092E+03, 3.1229E+03,
     +  3.1366E+03, 3.1505E+03, 3.1643E+03, 3.1782E+03, 3.1921E+03,
     + 3.2061E+03, 3.2201E+03, 3.2342E+03, 3.2483E+03, 3.2625E+03,
     +  3.2767E+03, 3.2909E+03, 3.3052E+03, 3.3195E+03, 3.3339E+03,
     + 3.3483E+03, 3.3628E+03, 3.3773E+03, 3.3918E+03, 3.4064E+03,
     +  3.4211E+03, 3.4358E+03, 3.4505E+03, 3.4653E+03, 3.4801E+03,
     + 3.4949E+03, 3.5099E+03, 3.5248E+03, 3.5398E+03, 3.5549E+03,
     +  3.5700E+03, 3.5851E+03, 3.6003E+03, 3.6155E+03, 3.6308E+03,
     + 3.6461E+03, 3.6615E+03, 3.6769E+03, 3.6924E+03, 3.7079E+03,
     +  3.7235E+03, 3.7390E+03, 3.7547E+03, 3.7704E+03, 3.7862E+03,
     + 3.8020E+03, 3.8178E+03, 3.8337E+03, 3.8497E+03, 3.8656E+03,
     +  3.8817E+03, 3.8978E+03, 3.9139E+03, 3.9301E+03, 3.9463E+03,
     + 3.9626E+03, 3.9789E+03, 3.9953E+03, 4.0117E+03, 4.0282E+03,
     +  4.0447E+03, 4.0613E+03, 4.0779E+03, 4.0946E+03, 4.1114E+03,
     + 4.1281E+03, 4.1450E+03, 4.1618E+03, 4.1788E+03, 4.1957E+03,
     +  4.2128E+03, 4.2299E+03, 4.2470E+03, 4.2642E+03, 4.2814E+03,
     + 4.2987E+03, 4.3160E+03, 4.3334E+03, 4.3509E+03, 4.3683E+03,
     +  4.3859E+03, 4.4035E+03, 4.4211E+03, 4.4388E+03, 4.4566E+03,
     + 4.4744E+03, 4.4923E+03, 4.5102E+03, 4.5281E+03, 4.5462E+03,
     +  4.5642E+03, 4.5824E+03, 4.6006E+03, 4.6188E+03, 4.6371E+03,
     + 4.6554E+03, 4.6738E+03, 4.6922E+03, 4.7107E+03, 4.7293E+03,
     +  4.7479E+03, 4.7666E+03, 4.7853E+03, 4.8041E+03, 4.8229E+03,
     + 4.8418E+03, 4.8608E+03, 4.8798E+03, 4.8988E+03, 4.9179E+03,
     +  4.9371E+03, 4.9563E+03, 4.9756E+03, 4.9949E+03, 5.0143E+03,
     + 5.0337E+03, 5.0532E+03, 5.0728E+03, 5.0924E+03, 5.1121E+03,
     +  5.1318E+03, 5.1516E+03, 5.1715E+03, 5.1913E+03, 5.2113E+03,
     + 5.2313E+03, 5.2514E+03, 5.2715E+03, 5.2917E+03, 5.3120E+03,
     +  5.3323E+03, 5.3527E+03, 5.3731E+03, 5.3936E+03, 5.4141E+03,
     + 5.4347E+03, 5.4554E+03, 5.4761E+03, 5.4969E+03, 5.5177E+03,
     +  5.5387E+03, 5.5596E+03, 5.5806E+03, 5.6017E+03, 5.6229E+03,
     + 5.6440E+03, 5.6653E+03, 5.6866E+03, 5.7080E+03, 5.7295E+03,
     +  5.7510E+03, 5.7725E+03, 5.7942E+03, 5.8159E+03, 5.8376E+03,
     + 5.8594E+03, 5.8813E+03, 5.9032E+03, 5.9253E+03, 5.9473E+03,
     +  5.9694E+03, 5.9917E+03, 6.0139E+03, 6.0362E+03, 6.0586E+03,
     + 6.0811E+03, 6.1036E+03, 6.1261E+03, 6.1487E+03, 6.1715E+03,
     +  6.1942E+03, 6.2170E+03, 6.2399E+03, 6.2629E+03, 6.2859E+03,
     + 6.3090E+03, 6.3321E+03, 6.3553E+03, 6.3786E+03, 6.4020E+03,
     +  6.4254E+03, 6.4488E+03, 6.4724E+03, 6.4960E+03, 6.5196E+03,
     + 6.5434E+03, 6.5672E+03, 6.5910E+03, 6.6150E+03, 6.6390E+03,
     +  6.6631E+03, 6.6872E+03, 6.7114E+03, 6.7357E+03, 6.7600E+03,
     + 6.7844E+03, 6.8089E+03, 6.8334E+03, 6.8581E+03, 6.8827E+03,
     +  6.9075E+03, 6.9323E+03, 6.9572E+03, 6.9822E+03, 7.0072E+03,
     + 7.0322E+03, 7.0574E+03, 7.0826E+03, 7.1079E+03, 7.1333E+03,
     +  7.1587E+03, 7.1843E+03, 7.2098E+03, 7.2355E+03, 7.2612E+03,
     + 7.2870E+03, 7.3128E+03, 7.3387E+03, 7.3648E+03, 7.3909E+03,
     +  7.4170E+03, 7.4432E+03, 7.4695E+03, 7.4958E+03, 7.5223E+03,
     + 7.5488E+03, 7.5753E+03, 7.6020E+03, 7.6287E+03, 7.6555E+03,
     +  7.6824E+03, 7.7093E+03, 7.7363E+03, 7.7634E+03, 7.7905E+03,
     + 7.8178E+03, 7.8451E+03, 7.8725E+03, 7.8999E+03, 7.9274E+03,
     +  7.9550E+03, 7.9827E+03, 8.0105E+03, 8.0383E+03, 8.0662E+03,
     + 8.0941E+03, 8.1222E+03, 8.1503E+03, 8.1785E+03, 8.2068E+03,
     +  8.2351E+03, 8.2636E+03, 8.2921E+03, 8.3206E+03, 8.3493E+03,
     + 8.3780E+03, 8.4068E+03, 8.4357E+03, 8.4647E+03, 8.4937E+03,
     +  8.5228E+03, 8.5520E+03, 8.5813E+03, 8.6107E+03, 8.6401E+03,
     + 8.6696E+03, 8.6992E+03, 8.7288E+03, 8.7586E+03, 8.7884E+03,
     +  8.8183E+03, 8.8483E+03, 8.8783E+03, 8.9085E+03, 8.9387E+03,
     + 8.9690E+03, 8.9993E+03, 9.0298E+03, 9.0603E+03, 9.0910E+03,
     +  9.1216E+03, 9.1524E+03, 9.1833E+03, 9.2142E+03, 9.2452E+03,
     + 9.2763E+03, 9.3075E+03, 9.3388E+03, 9.3702E+03, 9.4016E+03,
     +  9.4330E+03, 9.4646E+03, 9.4964E+03, 9.5281E+03, 9.5599E+03,
     + 9.5919E+03, 9.6239E+03, 9.6559E+03, 9.6881E+03, 9.7204E+03,
     +  9.7527E+03, 9.7852E+03, 9.8176E+03, 9.8502E+03, 9.8829E+03,
     + 9.9157E+03, 9.9485E+03, 9.9815E+03, 1.0014E+04, 1.0048E+04,
     +  1.0081E+04, 1.0114E+04, 1.0147E+04, 1.0181E+04, 1.0214E+04,
     + 1.0248E+04, 1.0282E+04, 1.0315E+04, 1.0349E+04, 1.0383E+04,
     +  1.0417E+04, 1.0451E+04, 1.0486E+04, 1.0520E+04, 1.0554E+04,
     + 1.0589E+04, 1.0623E+04, 1.0658E+04, 1.0693E+04, 1.0728E+04,
     +  1.0763E+04, 1.0798E+04, 1.0833E+04, 1.0868E+04, 1.0903E+04,
     + 1.0939E+04, 1.0974E+04, 1.1010E+04, 1.1045E+04, 1.1081E+04,
     +  1.1117E+04, 1.1153E+04, 1.1189E+04, 1.1225E+04, 1.1261E+04,
     + 1.1297E+04, 1.1334E+04, 1.1370E+04, 1.1407E+04, 1.1443E+04,
     +  1.1480E+04, 1.1517E+04, 1.1554E+04, 1.1591E+04, 1.1628E+04,
     + 1.1665E+04, 1.1703E+04, 1.1740E+04, 1.1777E+04, 1.1815E+04,
     +  1.1853E+04, 1.1890E+04, 1.1928E+04, 1.1966E+04, 1.2004E+04,
     + 1.2042E+04, 1.2081E+04, 1.2119E+04, 1.2157E+04, 1.2196E+04,
     +  1.2235E+04, 1.2273E+04, 1.2312E+04, 1.2351E+04, 1.2390E+04,
     + 1.2429E+04, 1.2468E+04, 1.2508E+04, 1.2547E+04, 1.2586E+04,
     +  1.2626E+04, 1.2666E+04, 1.2706E+04, 1.2745E+04, 1.2785E+04,
     + 1.2825E+04, 1.2866E+04, 1.2906E+04, 1.2946E+04, 1.2987E+04,
     +  1.3027E+04, 1.3068E+04, 1.3109E+04, 1.3150E+04, 1.3191E+04,
     + 1.3232E+04, 1.3273E+04, 1.3314E+04, 1.3355E+04, 1.3397E+04,
     +  1.3439E+04, 1.3480E+04, 1.3522E+04, 1.3564E+04, 1.3606E+04,
     + 1.3648E+04, 1.3690E+04, 1.3732E+04, 1.3775E+04, 1.3817E+04,
     +  1.3860E+04, 1.3903E+04, 1.3945E+04, 1.3988E+04, 1.4031E+04,
     + 1.4074E+04, 1.4118E+04, 1.4161E+04, 1.4204E+04, 1.4248E+04,
     +  1.4291E+04, 1.4335E+04, 1.4379E+04, 1.4423E+04, 1.4467E+04,
     + 1.4511E+04, 1.4555E+04, 1.4600E+04, 1.4644E+04, 1.4689E+04,
     +  1.4734E+04, 1.4778E+04, 1.4823E+04, 1.4868E+04, 1.4913E+04,
     + 1.4959E+04, 1.5004E+04, 1.5049E+04, 1.5095E+04, 1.5141E+04,
     +  1.5186E+04, 1.5232E+04, 1.5278E+04, 1.5324E+04, 1.5371E+04,
     + 1.5417E+04, 1.5463E+04, 1.5510E+04, 1.5556E+04, 1.5603E+04,
     +  1.5650E+04, 1.5697E+04, 1.5744E+04, 1.5791E+04, 1.5839E+04,
     + 1.5886E+04, 1.5934E+04, 1.5981E+04, 1.6029E+04, 1.6077E+04,
     +  1.6125E+04, 1.6173E+04, 1.6221E+04, 1.6269E+04, 1.6318E+04,
     + 1.6366E+04, 1.6415E+04, 1.6464E+04, 1.6513E+04, 1.6562E+04,
     +  1.6611E+04, 1.6660E+04, 1.6709E+04, 1.6759E+04, 1.6808E+04,
     + 1.6858E+04, 1.6908E+04, 1.6958E+04, 1.7008E+04, 1.7058E+04,
     +  1.7108E+04, 1.7159E+04, 1.7209E+04, 1.7260E+04, 1.7310E+04,
     + 1.7361E+04, 1.7412E+04, 1.7463E+04, 1.7514E+04, 1.7566E+04,
     +  1.7617E+04, 1.7669E+04, 1.7720E+04, 1.7772E+04, 1.7824E+04,
     + 1.7876E+04, 1.7928E+04, 1.7981E+04, 1.8033E+04, 1.8086E+04,
     +  1.8138E+04, 1.8191E+04, 1.8244E+04, 1.8297E+04, 1.8350E+04,
     + 1.8403E+04, 1.8457E+04, 1.8510E+04, 1.8564E+04, 1.8617E+04,
     +  1.8671E+04, 1.8725E+04, 1.8779E+04, 1.8833E+04, 1.8888E+04,
     + 1.8942E+04, 1.8997E+04, 1.9052E+04, 1.9106E+04, 1.9161E+04,
     +  1.9216E+04, 1.9272E+04, 1.9327E+04, 1.9382E+04, 1.9438E+04,
     + 1.9494E+04, 1.9550E+04, 1.9606E+04, 1.9662E+04, 1.9718E+04,
     +  1.9774E+04, 1.9831E+04, 1.9887E+04, 1.9944E+04, 2.0001E+04,
     + 2.0058E+04, 2.0115E+04, 2.0172E+04, 2.0229E+04, 2.0287E+04,
     +  2.0344E+04, 2.0402E+04, 2.0460E+04, 2.0518E+04, 2.0576E+04,
     + 2.0634E+04, 2.0693E+04, 2.0751E+04, 2.0810E+04, 2.0869E+04,
     +  2.0928E+04, 2.0987E+04, 2.1046E+04, 2.1105E+04, 2.1165E+04,
     + 2.1224E+04, 2.1284E+04, 2.1344E+04, 2.1404E+04, 2.1464E+04,
     +  2.1524E+04, 2.1584E+04, 2.1645E+04, 2.1706E+04, 2.1766E+04,
     + 2.1827E+04, 2.1888E+04, 2.1949E+04, 2.2011E+04, 2.2072E+04,
     +  2.2134E+04, 2.2195E+04, 2.2257E+04, 2.2319E+04, 2.2381E+04,
     + 2.2444E+04, 2.2506E+04, 2.2568E+04, 2.2631E+04, 2.2694E+04,
     +  2.2757E+04, 2.2820E+04, 2.2883E+04, 2.2946E+04, 2.3010E+04,
     + 2.3074E+04, 2.3137E+04, 2.3201E+04, 2.3265E+04, 2.3329E+04,
     +  2.3394E+04, 2.3458E+04, 2.3523E+04, 2.3588E+04, 2.3652E+04,
     + 2.3717E+04, 2.3783E+04, 2.3848E+04, 2.3913E+04, 2.3979E+04,
     +  2.4045E+04, 2.4111E+04, 2.4177E+04, 2.4243E+04, 2.4309E+04,
     + 2.4375E+04, 2.4442E+04, 2.4509E+04, 2.4576E+04, 2.4643E+04,
     +  2.4710E+04, 2.4777E+04, 2.4844E+04, 2.4912E+04, 2.4980E+04,
     + 2.5048E+04, 2.5116E+04, 2.5184E+04, 2.5252E+04, 2.5321E+04,
     +  2.5389E+04, 2.5458E+04, 2.5527E+04, 2.5596E+04, 2.5665E+04,
     + 2.5734E+04, 2.5804E+04, 2.5873E+04, 2.5943E+04, 2.6013E+04,
     +  2.6083E+04, 2.6153E+04, 2.6224E+04, 2.6294E+04, 2.6365E+04,
     + 2.6436E+04, 2.6507E+04, 2.6578E+04, 2.6649E+04, 2.6721E+04,
     +  2.6792E+04, 2.6864E+04, 2.6936E+04, 2.7008E+04, 2.7080E+04,
     + 2.7152E+04, 2.7225E+04, 2.7297E+04, 2.7370E+04, 2.7443E+04,
     +  2.7516E+04, 2.7590E+04, 2.7663E+04, 2.7736E+04, 2.7810E+04,
     + 2.7884E+04, 2.7958E+04, 2.8032E+04, 2.8106E+04, 2.8181E+04,
     +  2.8256E+04, 2.8330E+04, 2.8405E+04, 2.8480E+04, 2.8556E+04,
     + 2.8631E+04, 2.8707E+04, 2.8782E+04, 2.8858E+04, 2.8934E+04,
     +  2.9011E+04, 2.9087E+04, 2.9163E+04, 2.9240E+04, 2.9317E+04,
     + 2.9394E+04, 2.9471E+04, 2.9548E+04, 2.9626E+04, 2.9703E+04,
     +  2.9781E+04, 2.9859E+04, 2.9937E+04, 3.0016E+04, 3.0094E+04,
     + 3.0173E+04, 3.0251E+04, 3.0330E+04, 3.0409E+04, 3.0489E+04,
     +  3.0568E+04, 3.0648E+04, 3.0727E+04, 3.0807E+04, 3.0887E+04,
     + 3.0967E+04, 3.1048E+04, 3.1128E+04, 3.1209E+04, 3.1290E+04,
     +  3.1371E+04, 3.1452E+04, 3.1534E+04, 3.1615E+04, 3.1697E+04,
     + 3.1779E+04, 3.1861E+04, 3.1943E+04, 3.2025E+04, 3.2108E+04,
     +  3.2191E+04, 3.2273E+04, 3.2357E+04, 3.2440E+04, 3.2523E+04,
     + 3.2607E+04, 3.2690E+04, 3.2774E+04, 3.2858E+04, 3.2943E+04,
     +  3.3027E+04, 3.3112E+04, 3.3196E+04, 3.3281E+04, 3.3366E+04,
     + 3.3452E+04, 3.3537E+04, 3.3622E+04, 3.3708E+04, 3.3794E+04,
     +  3.3880E+04, 3.3967E+04, 3.4053E+04, 3.4140E+04, 3.4226E+04,
     + 3.4313E+04, 3.4401E+04, 3.4488E+04, 3.4575E+04, 3.4663E+04,
     +  3.4751E+04, 3.4839E+04, 3.4927E+04, 3.5015E+04, 3.5104E+04,
     + 3.5193E+04, 3.5282E+04, 3.5371E+04, 3.5460E+04, 3.5549E+04,
     +  3.5639E+04, 3.5729E+04, 3.5819E+04, 3.5909E+04, 3.5999E+04,
     + 3.6090E+04, 3.6180E+04, 3.6271E+04, 3.6362E+04, 3.6454E+04,
     +  3.6545E+04, 3.6637E+04, 3.6728E+04, 3.6820E+04, 3.6912E+04,
     + 3.7005E+04, 3.7097E+04, 3.7190E+04, 3.7283E+04, 3.7376E+04,
     +  3.7469E+04, 3.7562E+04, 3.7656E+04, 3.7750E+04, 3.7844E+04,
     + 3.7938E+04, 3.8032E+04, 3.8127E+04, 3.8221E+04, 3.8316E+04,
     +  3.8411E+04, 3.8506E+04, 3.8602E+04, 3.8698E+04, 3.8793E+04,
     + 3.8889E+04, 3.8985E+04, 3.9082E+04, 3.9178E+04, 3.9275E+04,
     +  3.9372E+04, 3.9469E+04, 3.9566E+04, 3.9664E+04, 3.9762E+04,
     + 3.9859E+04, 3.9958E+04, 4.0056E+04, 4.0154E+04, 4.0253E+04,
     +  4.0352E+04, 4.0451E+04, 4.0550E+04, 4.0649E+04, 4.0749E+04,
     + 4.0849E+04, 4.0949E+04, 4.1049E+04, 4.1149E+04, 4.1250E+04,
     +  4.1350E+04, 4.1451E+04, 4.1552E+04, 4.1654E+04, 4.1755E+04,
     + 4.1857E+04, 4.1959E+04, 4.2061E+04, 4.2163E+04, 4.2266E+04,
     +  4.2368E+04, 4.2471E+04, 4.2574E+04, 4.2678E+04, 4.2781E+04,
     + 4.2885E+04, 4.2989E+04, 4.3093E+04, 4.3197E+04, 4.3301E+04,
     +  4.3406E+04, 4.3511E+04, 4.3616E+04, 4.3721E+04, 4.3827E+04,
     + 4.3932E+04, 4.4038E+04, 4.4144E+04, 4.4251E+04, 4.4357E+04,
     +  4.4464E+04, 4.4571E+04, 4.4678E+04, 4.4785E+04, 4.4892E+04,
     + 4.5000E+04, 4.5108E+04, 4.5216E+04, 4.5324E+04, 4.5433E+04,
     +  4.5541E+04, 4.5650E+04, 4.5759E+04, 4.5869E+04, 4.5978E+04,
     + 4.6088E+04, 4.6198E+04, 4.6308E+04, 4.6418E+04, 4.6529E+04,
     +  4.6640E+04, 4.6751E+04, 4.6862E+04, 4.6973E+04, 4.7085E+04,
     + 4.7197E+04, 4.7309E+04, 4.7421E+04, 4.7533E+04, 4.7646E+04,
     +  4.7759E+04, 4.7872E+04, 4.7985E+04, 4.8098E+04, 4.8212E+04,
     + 4.8326E+04, 4.8440E+04, 4.8554E+04, 4.8669E+04, 4.8784E+04,
     +  4.8899E+04, 4.9014E+04, 4.9129E+04, 4.9245E+04, 4.9361E+04,
     + 4.9477E+04, 4.9593E+04, 4.9709E+04, 4.9826E+04, 4.9943E+04,
     +  5.0060E+04, 5.0177E+04, 5.0295E+04, 5.0412E+04, 5.0530E+04,
     + 5.0649E+04, 5.0767E+04, 5.0886E+04, 5.1004E+04, 5.1123E+04,
     +  5.1243E+04, 5.1362E+04, 5.1482E+04, 5.1602E+04, 5.1722E+04,
     + 5.1842E+04, 5.1963E+04, 5.2084E+04, 5.2205E+04, 5.2326E+04,
     +  5.2447E+04, 5.2569E+04, 5.2691E+04, 5.2813E+04, 5.2936E+04,
     + 5.3058E+04, 5.3181E+04, 5.3304E+04, 5.3427E+04, 5.3551E+04,
     +  5.3674E+04, 5.3798E+04, 5.3922E+04, 5.4047E+04, 5.4171E+04,
     + 5.4296E+04, 5.4421E+04, 5.4547E+04, 5.4672E+04, 5.4798E+04,
     +  5.4924E+04, 5.5050E+04, 5.5176E+04, 5.5303E+04, 5.5430E+04,
     + 5.5557E+04, 5.5684E+04, 5.5812E+04, 5.5940E+04, 5.6068E+04,
     +  5.6196E+04, 5.6324E+04, 5.6453E+04, 5.6582E+04, 5.6711E+04,
     + 5.6840E+04, 5.6970E+04, 5.7100E+04, 5.7230E+04, 5.7360E+04,
     +  5.7491E+04, 5.7622E+04, 5.7753E+04, 5.7884E+04, 5.8015E+04,
     + 5.8147E+04, 5.8279E+04, 5.8411E+04, 5.8544E+04, 5.8676E+04,
     +  5.8809E+04, 5.8943E+04, 5.9076E+04, 5.9209E+04, 5.9343E+04,
     + 5.9478E+04, 5.9612E+04, 5.9746E+04, 5.9881E+04, 6.0016E+04,
     +  6.0151E+04, 6.0287E+04, 6.0423E+04, 6.0559E+04, 6.0695E+04,
     + 6.0832E+04, 6.0968E+04, 6.1105E+04, 6.1242E+04, 6.1380E+04,
     +  6.1518E+04, 6.1656E+04, 6.1794E+04, 6.1932E+04, 6.2071E+04,
     + 6.2210E+04, 6.2349E+04, 6.2488E+04, 6.2628E+04, 6.2768E+04,
     +  6.2908E+04, 6.3048E+04, 6.3189E+04, 6.3330E+04, 6.3471E+04,
     + 6.3612E+04, 6.3754E+04, 6.3896E+04, 6.4038E+04, 6.4180E+04,
     +  6.4323E+04, 6.4466E+04, 6.4609E+04, 6.4752E+04, 6.4896E+04,
     + 6.5040E+04, 6.5184E+04, 6.5328E+04, 6.5473E+04, 6.5618E+04,
     +  6.5763E+04, 6.5908E+04, 6.6054E+04, 6.6200E+04, 6.6346E+04,
     + 6.6492E+04, 6.6639E+04, 6.6786E+04, 6.6933E+04, 6.7080E+04,
     +  6.7228E+04, 6.7376E+04, 6.7524E+04, 6.7672E+04, 6.7821E+04,
     + 6.7970E+04, 6.8119E+04, 6.8269E+04, 6.8418E+04, 6.8568E+04,
     +  6.8718E+04, 6.8869E+04, 6.9020E+04, 6.9171E+04, 6.9322E+04,
     + 6.9473E+04, 6.9625E+04, 6.9777E+04, 6.9929E+04, 7.0082E+04,
     +  7.0235E+04, 7.0388E+04, 7.0541E+04, 7.0695E+04, 7.0849E+04,
     + 7.1003E+04, 7.1157E+04, 7.1312E+04, 7.1467E+04, 7.1622E+04,
     +  7.1778E+04, 7.1933E+04, 7.2089E+04, 7.2245E+04, 7.2402E+04,
     + 7.2559E+04, 7.2716E+04, 7.2873E+04, 7.3031E+04, 7.3188E+04,
     +  7.3347E+04, 7.3505E+04, 7.3664E+04, 7.3822E+04, 7.3982E+04,
     + 7.4141E+04, 7.4301E+04, 7.4461E+04, 7.4621E+04, 7.4782E+04,
     +  7.4942E+04, 7.5103E+04, 7.5265E+04, 7.5426E+04, 7.5588E+04,
     + 7.5751E+04, 7.5913E+04, 7.6076E+04, 7.6239E+04, 7.6402E+04,
     +  7.6566E+04, 7.6729E+04, 7.6893E+04, 7.7058E+04, 7.7222E+04,
     + 7.7387E+04, 7.7552E+04, 7.7718E+04, 7.7884E+04, 7.8050E+04,
     +  7.8216E+04, 7.8383E+04, 7.8549E+04, 7.8717E+04, 7.8884E+04,
     + 7.9052E+04, 7.9220E+04, 7.9388E+04, 7.9556E+04, 7.9725E+04,
     +  7.9894E+04, 8.0064E+04, 8.0233E+04, 8.0403E+04, 8.0574E+04,
     + 8.0744E+04, 8.0915E+04, 8.1086E+04, 8.1257E+04, 8.1429E+04,
     +  8.1601E+04, 8.1773E+04, 8.1946E+04, 8.2118E+04, 8.2291E+04,
     + 8.2465E+04, 8.2638E+04, 8.2812E+04, 8.2986E+04, 8.3161E+04,
     +  8.3336E+04, 8.3511E+04, 8.3686E+04, 8.3862E+04, 8.4038E+04,
     + 8.4214E+04, 8.4391E+04, 8.4568E+04, 8.4745E+04, 8.4922E+04,
     +  8.5100E+04, 8.5278E+04, 8.5456E+04, 8.5634E+04, 8.5813E+04,
     + 8.5993E+04, 8.6172E+04, 8.6352E+04, 8.6532E+04, 8.6712E+04,
     +  8.6893E+04, 8.7074E+04, 8.7255E+04, 8.7436E+04, 8.7618E+04,
     + 8.7800E+04, 8.7983E+04, 8.8165E+04, 8.8348E+04, 8.8532E+04,
     +  8.8715E+04, 8.8899E+04, 8.9083E+04, 8.9268E+04, 8.9453E+04,
     + 8.9638E+04, 8.9823E+04, 9.0009E+04, 9.0195E+04, 9.0381E+04,
     +  9.0568E+04, 9.0754E+04, 9.0942E+04, 9.1129E+04, 9.1317E+04,
     + 9.1505E+04, 9.1693E+04, 9.1882E+04, 9.2071E+04, 9.2261E+04,
     +  9.2450E+04, 9.2640E+04, 9.2830E+04, 9.3021E+04, 9.3212E+04,
     + 9.3403E+04, 9.3594E+04, 9.3786E+04, 9.3978E+04, 9.4171E+04,
     +  9.4363E+04, 9.4556E+04, 9.4750E+04, 9.4943E+04, 9.5137E+04,
     + 9.5331E+04, 9.5526E+04, 9.5721E+04, 9.5916E+04, 9.6112E+04,
     +  9.6308E+04, 9.6504E+04, 9.6700E+04, 9.6897E+04, 9.7094E+04,
     + 9.7291E+04, 9.7489E+04, 9.7687E+04, 9.7885E+04, 9.8084E+04,
     +  9.8283E+04, 9.8482E+04, 9.8682E+04, 9.8882E+04, 9.9082E+04,
     + 9.9282E+04, 9.9483E+04, 9.9684E+04, 9.9886E+04, 1.0009E+05,
     +  1.0029E+05, 1.0049E+05, 1.0070E+05, 1.0090E+05, 1.0110E+05,
     + 1.0131E+05, 1.0151E+05, 1.0171E+05, 1.0192E+05, 1.0212E+05,
     +  1.0233E+05, 1.0253E+05, 1.0274E+05, 1.0295E+05, 1.0315E+05,
     + 1.0336E+05, 1.0357E+05, 1.0378E+05, 1.0398E+05, 1.0419E+05,
     +  1.0440E+05, 1.0461E+05, 1.0482E+05, 1.0503E+05, 1.0524E+05,
     + 1.0545E+05, 1.0566E+05, 1.0587E+05, 1.0608E+05, 1.0629E+05,
     +  1.0650E+05, 1.0672E+05, 1.0693E+05, 1.0714E+05, 1.0736E+05,
     + 1.0757E+05, 1.0778E+05, 1.0800E+05, 1.0821E+05, 1.0843E+05,
     +  1.0864E+05, 1.0886E+05, 1.0907E+05, 1.0929E+05, 1.0951E+05,
     + 1.0972E+05, 1.0994E+05, 1.1016E+05, 1.1038E+05, 1.1059E+05,
     +  1.1081E+05, 1.1103E+05, 1.1125E+05, 1.1147E+05, 1.1169E+05,
     + 1.1191E+05, 1.1213E+05, 1.1235E+05, 1.1257E+05, 1.1280E+05,
     +  1.1302E+05, 1.1324E+05, 1.1346E+05, 1.1369E+05, 1.1391E+05,
     + 1.1413E+05, 1.1436E+05, 1.1458E+05, 1.1481E+05, 1.1503E+05,
     +  1.1526E+05, 1.1548E+05, 1.1571E+05, 1.1594E+05, 1.1616E+05,
     + 1.1639E+05, 1.1662E+05, 1.1684E+05, 1.1707E+05, 1.1730E+05,
     +  1.1753E+05, 1.1776E+05, 1.1799E+05, 1.1822E+05, 1.1845E+05,
     + 1.1868E+05, 1.1891E+05, 1.1914E+05, 1.1937E+05, 1.1961E+05,
     +  1.1984E+05, 1.2007E+05, 1.2030E+05, 1.2054E+05, 1.2077E+05,
     + 1.2101E+05, 1.2124E+05, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       131
      data (QofT( 3,J),J=1,2408)/ 4.0000E+00, 4.0026E+00, 4.0790E+00,
     +  4.4504E+00, 5.2994E+00, 6.6565E+00, 8.4534E+00, 1.0590E+01,
     + 1.2972E+01, 1.5523E+01, 1.7767E+01, 1.9868E+01, 2.2051E+01,
     +  2.4312E+01, 2.6649E+01, 2.9059E+01, 3.1540E+01, 3.4090E+01,
     + 3.6707E+01, 3.9390E+01, 4.4945E+01, 5.0744E+01, 5.6778E+01,
     +  6.3036E+01, 6.9512E+01, 7.6198E+01, 8.3087E+01, 9.0174E+01,
     + 9.7454E+01, 1.0492E+02, 1.1257E+02, 1.2040E+02, 1.2840E+02,
     +  1.3657E+02, 1.4491E+02, 1.5342E+02, 1.6208E+02, 1.7090E+02,
     + 1.7988E+02, 1.8901E+02, 1.9829E+02, 2.0772E+02, 2.1729E+02,
     +  2.2701E+02, 2.3686E+02, 2.4686E+02, 2.5699E+02, 2.6726E+02,
     + 2.7766E+02, 2.8819E+02, 2.9886E+02, 3.0965E+02, 3.2057E+02,
     +  3.3161E+02, 3.4278E+02, 3.5407E+02, 3.6549E+02, 3.7702E+02,
     + 3.8867E+02, 4.0044E+02, 4.1233E+02, 4.2433E+02, 4.3644E+02,
     +  4.4867E+02, 4.6102E+02, 4.7347E+02, 4.8603E+02, 4.9870E+02,
     + 5.1149E+02, 5.2437E+02, 5.3737E+02, 5.5047E+02, 5.6367E+02,
     +  5.7698E+02, 5.9040E+02, 6.0391E+02, 6.1753E+02, 6.3124E+02,
     + 6.4506E+02, 6.5899E+02, 6.7300E+02, 6.8712E+02, 7.0133E+02,
     +  7.1563E+02, 7.3004E+02, 7.4454E+02, 7.5914E+02, 7.7383E+02,
     + 7.8861E+02, 8.0349E+02, 8.1846E+02, 8.3352E+02, 8.4868E+02,
     +  8.6392E+02, 8.7925E+02, 8.9468E+02, 9.1020E+02, 9.2580E+02,
     + 9.4150E+02, 9.5729E+02, 9.7316E+02, 9.8912E+02, 1.0052E+03,
     +  1.0213E+03, 1.0375E+03, 1.0538E+03, 1.0702E+03, 1.0867E+03,
     + 1.1033E+03, 1.1199E+03, 1.1366E+03, 1.1534E+03, 1.1703E+03,
     +  1.1873E+03, 1.2044E+03, 1.2216E+03, 1.2388E+03, 1.2561E+03,
     + 1.2735E+03, 1.2910E+03, 1.3085E+03, 1.3262E+03, 1.3439E+03,
     +  1.3617E+03, 1.3796E+03, 1.3976E+03, 1.4156E+03, 1.4338E+03,
     + 1.4520E+03, 1.4703E+03, 1.4886E+03, 1.5071E+03, 1.5256E+03,
     +  1.5443E+03, 1.5629E+03, 1.5817E+03, 1.6006E+03, 1.6195E+03,
     + 1.6385E+03, 1.6576E+03, 1.6768E+03, 1.6960E+03, 1.7154E+03,
     +  1.7348E+03, 1.7543E+03, 1.7739E+03, 1.7935E+03, 1.8133E+03,
     + 1.8331E+03, 1.8529E+03, 1.8729E+03, 1.8930E+03, 1.9131E+03,
     +  1.9333E+03, 1.9536E+03, 1.9740E+03, 1.9944E+03, 2.0149E+03,
     + 2.0356E+03, 2.0562E+03, 2.0770E+03, 2.0979E+03, 2.1188E+03,
     +  2.1398E+03, 2.1609E+03, 2.1821E+03, 2.2033E+03, 2.2246E+03,
     + 2.2460E+03, 2.2675E+03, 2.2891E+03, 2.3107E+03, 2.3325E+03,
     +  2.3543E+03, 2.3762E+03, 2.3982E+03, 2.4202E+03, 2.4424E+03,
     + 2.4646E+03, 2.4869E+03, 2.5093E+03, 2.5317E+03, 2.5543E+03,
     +  2.5769E+03, 2.5996E+03, 2.6224E+03, 2.6453E+03, 2.6683E+03,
     + 2.6913E+03, 2.7145E+03, 2.7377E+03, 2.7610E+03, 2.7843E+03,
     +  2.8078E+03, 2.8314E+03, 2.8550E+03, 2.8787E+03, 2.9025E+03,
     + 2.9264E+03, 2.9503E+03, 2.9744E+03, 2.9986E+03, 3.0228E+03,
     +  3.0471E+03, 3.0715E+03, 3.0960E+03, 3.1205E+03, 3.1452E+03,
     + 3.1699E+03, 3.1948E+03, 3.2197E+03, 3.2447E+03, 3.2698E+03,
     +  3.2950E+03, 3.3202E+03, 3.3456E+03, 3.3710E+03, 3.3966E+03,
     + 3.4222E+03, 3.4479E+03, 3.4737E+03, 3.4996E+03, 3.5255E+03,
     +  3.5516E+03, 3.5778E+03, 3.6040E+03, 3.6303E+03, 3.6568E+03,
     + 3.6833E+03, 3.7099E+03, 3.7366E+03, 3.7634E+03, 3.7903E+03,
     +  3.8173E+03, 3.8443E+03, 3.8715E+03, 3.8987E+03, 3.9261E+03,
     + 3.9535E+03, 3.9811E+03, 4.0087E+03, 4.0364E+03, 4.0642E+03,
     +  4.0922E+03, 4.1201E+03, 4.1482E+03, 4.1764E+03, 4.2047E+03,
     + 4.2331E+03, 4.2616E+03, 4.2902E+03, 4.3188E+03, 4.3476E+03,
     +  4.3765E+03, 4.4055E+03, 4.4345E+03, 4.4637E+03, 4.4929E+03,
     + 4.5223E+03, 4.5517E+03, 4.5813E+03, 4.6110E+03, 4.6407E+03,
     +  4.6706E+03, 4.7005E+03, 4.7306E+03, 4.7607E+03, 4.7910E+03,
     + 4.8213E+03, 4.8518E+03, 4.8824E+03, 4.9130E+03, 4.9438E+03,
     +  4.9746E+03, 5.0056E+03, 5.0367E+03, 5.0679E+03, 5.0991E+03,
     + 5.1305E+03, 5.1620E+03, 5.1936E+03, 5.2253E+03, 5.2571E+03,
     +  5.2890E+03, 5.3210E+03, 5.3531E+03, 5.3854E+03, 5.4177E+03,
     + 5.4502E+03, 5.4827E+03, 5.5154E+03, 5.5481E+03, 5.5810E+03,
     +  5.6139E+03, 5.6470E+03, 5.6802E+03, 5.7135E+03, 5.7469E+03,
     + 5.7805E+03, 5.8141E+03, 5.8479E+03, 5.8817E+03, 5.9157E+03,
     +  5.9498E+03, 5.9840E+03, 6.0182E+03, 6.0527E+03, 6.0872E+03,
     + 6.1219E+03, 6.1566E+03, 6.1915E+03, 6.2265E+03, 6.2616E+03,
     +  6.2968E+03, 6.3321E+03, 6.3676E+03, 6.4031E+03, 6.4388E+03,
     + 6.4746E+03, 6.5105E+03, 6.5466E+03, 6.5827E+03, 6.6189E+03,
     +  6.6553E+03, 6.6918E+03, 6.7285E+03, 6.7652E+03, 6.8020E+03,
     + 6.8390E+03, 6.8761E+03, 6.9133E+03, 6.9507E+03, 6.9881E+03,
     +  7.0257E+03, 7.0635E+03, 7.1012E+03, 7.1392E+03, 7.1773E+03,
     + 7.2155E+03, 7.2538E+03, 7.2923E+03, 7.3308E+03, 7.3695E+03,
     +  7.4084E+03, 7.4473E+03, 7.4863E+03, 7.5256E+03, 7.5649E+03,
     + 7.6043E+03, 7.6439E+03, 7.6836E+03, 7.7235E+03, 7.7634E+03,
     +  7.8035E+03, 7.8438E+03, 7.8841E+03, 7.9246E+03, 7.9652E+03,
     + 8.0060E+03, 8.0468E+03, 8.0878E+03, 8.1290E+03, 8.1702E+03,
     +  8.2116E+03, 8.2532E+03, 8.2949E+03, 8.3367E+03, 8.3786E+03,
     + 8.4206E+03, 8.4628E+03, 8.5052E+03, 8.5477E+03, 8.5903E+03,
     +  8.6330E+03, 8.6759E+03, 8.7190E+03, 8.7622E+03, 8.8055E+03,
     + 8.8489E+03, 8.8924E+03, 8.9361E+03, 8.9800E+03, 9.0240E+03,
     +  9.0681E+03, 9.1124E+03, 9.1569E+03, 9.2015E+03, 9.2461E+03,
     + 9.2909E+03, 9.3360E+03, 9.3811E+03, 9.4264E+03, 9.4718E+03,
     +  9.5174E+03, 9.5631E+03, 9.6090E+03, 9.6550E+03, 9.7011E+03,
     + 9.7474E+03, 9.7938E+03, 9.8404E+03, 9.8871E+03, 9.9340E+03,
     +  9.9810E+03, 1.0028E+04, 1.0076E+04, 1.0123E+04, 1.0171E+04,
     + 1.0218E+04, 1.0266E+04, 1.0314E+04, 1.0363E+04, 1.0411E+04,
     +  1.0460E+04, 1.0508E+04, 1.0557E+04, 1.0606E+04, 1.0655E+04,
     + 1.0705E+04, 1.0754E+04, 1.0804E+04, 1.0853E+04, 1.0903E+04,
     +  1.0953E+04, 1.1004E+04, 1.1054E+04, 1.1105E+04, 1.1155E+04,
     + 1.1206E+04, 1.1257E+04, 1.1308E+04, 1.1360E+04, 1.1411E+04,
     +  1.1463E+04, 1.1515E+04, 1.1567E+04, 1.1619E+04, 1.1671E+04,
     + 1.1724E+04, 1.1777E+04, 1.1829E+04, 1.1882E+04, 1.1936E+04,
     +  1.1989E+04, 1.2042E+04, 1.2096E+04, 1.2150E+04, 1.2204E+04,
     + 1.2258E+04, 1.2312E+04, 1.2367E+04, 1.2421E+04, 1.2476E+04,
     +  1.2531E+04, 1.2586E+04, 1.2642E+04, 1.2697E+04, 1.2753E+04,
     + 1.2809E+04, 1.2865E+04, 1.2921E+04, 1.2977E+04, 1.3034E+04,
     +  1.3091E+04, 1.3147E+04, 1.3205E+04, 1.3262E+04, 1.3319E+04,
     + 1.3377E+04, 1.3435E+04, 1.3492E+04, 1.3551E+04, 1.3609E+04,
     +  1.3667E+04, 1.3726E+04, 1.3785E+04, 1.3844E+04, 1.3903E+04,
     + 1.3963E+04, 1.4022E+04, 1.4082E+04, 1.4142E+04, 1.4202E+04,
     +  1.4262E+04, 1.4323E+04, 1.4383E+04, 1.4444E+04, 1.4505E+04,
     + 1.4566E+04, 1.4628E+04, 1.4689E+04, 1.4751E+04, 1.4813E+04,
     +  1.4875E+04, 1.4938E+04, 1.5000E+04, 1.5063E+04, 1.5126E+04,
     + 1.5189E+04, 1.5252E+04, 1.5316E+04, 1.5380E+04, 1.5443E+04,
     +  1.5508E+04, 1.5572E+04, 1.5636E+04, 1.5701E+04, 1.5766E+04,
     + 1.5831E+04, 1.5896E+04, 1.5962E+04, 1.6027E+04, 1.6093E+04,
     +  1.6159E+04, 1.6225E+04, 1.6292E+04, 1.6358E+04, 1.6425E+04,
     + 1.6492E+04, 1.6559E+04, 1.6627E+04, 1.6694E+04, 1.6762E+04,
     +  1.6830E+04, 1.6899E+04, 1.6967E+04, 1.7036E+04, 1.7105E+04,
     + 1.7174E+04, 1.7243E+04, 1.7312E+04, 1.7382E+04, 1.7452E+04,
     +  1.7522E+04, 1.7592E+04, 1.7663E+04, 1.7733E+04, 1.7805E+04,
     + 1.7876E+04, 1.7947E+04, 1.8019E+04, 1.8090E+04, 1.8162E+04,
     +  1.8235E+04, 1.8307E+04, 1.8380E+04, 1.8452E+04, 1.8525E+04,
     + 1.8599E+04, 1.8672E+04, 1.8746E+04, 1.8820E+04, 1.8894E+04,
     +  1.8968E+04, 1.9043E+04, 1.9118E+04, 1.9193E+04, 1.9268E+04,
     + 1.9343E+04, 1.9419E+04, 1.9495E+04, 1.9571E+04, 1.9647E+04,
     +  1.9724E+04, 1.9801E+04, 1.9878E+04, 1.9955E+04, 2.0032E+04,
     + 2.0110E+04, 2.0188E+04, 2.0266E+04, 2.0345E+04, 2.0423E+04,
     +  2.0502E+04, 2.0581E+04, 2.0660E+04, 2.0740E+04, 2.0820E+04,
     + 2.0900E+04, 2.0980E+04, 2.1060E+04, 2.1141E+04, 2.1222E+04,
     +  2.1303E+04, 2.1384E+04, 2.1466E+04, 2.1548E+04, 2.1630E+04,
     + 2.1712E+04, 2.1795E+04, 2.1877E+04, 2.1960E+04, 2.2044E+04,
     +  2.2127E+04, 2.2211E+04, 2.2295E+04, 2.2379E+04, 2.2464E+04,
     + 2.2548E+04, 2.2633E+04, 2.2718E+04, 2.2804E+04, 2.2890E+04,
     +  2.2976E+04, 2.3062E+04, 2.3148E+04, 2.3235E+04, 2.3322E+04,
     + 2.3409E+04, 2.3496E+04, 2.3584E+04, 2.3672E+04, 2.3760E+04,
     +  2.3848E+04, 2.3937E+04, 2.4026E+04, 2.4115E+04, 2.4205E+04,
     + 2.4294E+04, 2.4384E+04, 2.4474E+04, 2.4565E+04, 2.4655E+04,
     +  2.4746E+04, 2.4837E+04, 2.4929E+04, 2.5021E+04, 2.5113E+04,
     + 2.5205E+04, 2.5297E+04, 2.5390E+04, 2.5483E+04, 2.5576E+04,
     +  2.5670E+04, 2.5764E+04, 2.5858E+04, 2.5952E+04, 2.6046E+04,
     + 2.6141E+04, 2.6236E+04, 2.6332E+04, 2.6427E+04, 2.6523E+04,
     +  2.6620E+04, 2.6716E+04, 2.6813E+04, 2.6910E+04, 2.7007E+04,
     + 2.7104E+04, 2.7202E+04, 2.7300E+04, 2.7399E+04, 2.7497E+04,
     +  2.7596E+04, 2.7695E+04, 2.7795E+04, 2.7894E+04, 2.7994E+04,
     + 2.8094E+04, 2.8195E+04, 2.8296E+04, 2.8397E+04, 2.8498E+04,
     +  2.8600E+04, 2.8702E+04, 2.8804E+04, 2.8906E+04, 2.9009E+04,
     + 2.9112E+04, 2.9215E+04, 2.9319E+04, 2.9423E+04, 2.9527E+04,
     +  2.9631E+04, 2.9736E+04, 2.9841E+04, 2.9946E+04, 3.0052E+04,
     + 3.0158E+04, 3.0264E+04, 3.0370E+04, 3.0477E+04, 3.0584E+04,
     +  3.0692E+04, 3.0799E+04, 3.0907E+04, 3.1015E+04, 3.1124E+04,
     + 3.1233E+04, 3.1342E+04, 3.1451E+04, 3.1561E+04, 3.1671E+04,
     +  3.1781E+04, 3.1892E+04, 3.2002E+04, 3.2113E+04, 3.2225E+04,
     + 3.2337E+04, 3.2449E+04, 3.2561E+04, 3.2674E+04, 3.2787E+04,
     +  3.2900E+04, 3.3014E+04, 3.3127E+04, 3.3242E+04, 3.3356E+04,
     + 3.3471E+04, 3.3586E+04, 3.3701E+04, 3.3817E+04, 3.3933E+04,
     +  3.4049E+04, 3.4166E+04, 3.4283E+04, 3.4400E+04, 3.4518E+04,
     + 3.4636E+04, 3.4754E+04, 3.4872E+04, 3.4991E+04, 3.5110E+04,
     +  3.5230E+04, 3.5349E+04, 3.5469E+04, 3.5590E+04, 3.5711E+04,
     + 3.5832E+04, 3.5953E+04, 3.6075E+04, 3.6197E+04, 3.6319E+04,
     +  3.6442E+04, 3.6564E+04, 3.6688E+04, 3.6812E+04, 3.6935E+04,
     + 3.7060E+04, 3.7184E+04, 3.7309E+04, 3.7434E+04, 3.7560E+04,
     +  3.7686E+04, 3.7812E+04, 3.7939E+04, 3.8065E+04, 3.8193E+04,
     + 3.8320E+04, 3.8448E+04, 3.8576E+04, 3.8705E+04, 3.8834E+04,
     +  3.8963E+04, 3.9092E+04, 3.9222E+04, 3.9352E+04, 3.9483E+04,
     + 3.9614E+04, 3.9745E+04, 3.9877E+04, 4.0009E+04, 4.0141E+04,
     +  4.0273E+04, 4.0406E+04, 4.0539E+04, 4.0673E+04, 4.0807E+04,
     + 4.0941E+04, 4.1076E+04, 4.1211E+04, 4.1346E+04, 4.1482E+04,
     +  4.1618E+04, 4.1754E+04, 4.1891E+04, 4.2028E+04, 4.2166E+04,
     + 4.2303E+04, 4.2442E+04, 4.2580E+04, 4.2719E+04, 4.2858E+04,
     +  4.2998E+04, 4.3137E+04, 4.3278E+04, 4.3418E+04, 4.3559E+04,
     + 4.3701E+04, 4.3842E+04, 4.3984E+04, 4.4127E+04, 4.4270E+04,
     +  4.4413E+04, 4.4556E+04, 4.4700E+04, 4.4844E+04, 4.4989E+04,
     + 4.5134E+04, 4.5279E+04, 4.5425E+04, 4.5571E+04, 4.5717E+04,
     +  4.5864E+04, 4.6011E+04, 4.6159E+04, 4.6306E+04, 4.6455E+04,
     + 4.6603E+04, 4.6752E+04, 4.6902E+04, 4.7052E+04, 4.7202E+04,
     +  4.7352E+04, 4.7503E+04, 4.7654E+04, 4.7806E+04, 4.7958E+04,
     + 4.8110E+04, 4.8263E+04, 4.8416E+04, 4.8570E+04, 4.8724E+04,
     +  4.8878E+04, 4.9033E+04, 4.9188E+04, 4.9343E+04, 4.9499E+04,
     + 4.9655E+04, 4.9812E+04, 4.9969E+04, 5.0126E+04, 5.0284E+04,
     +  5.0442E+04, 5.0600E+04, 5.0759E+04, 5.0919E+04, 5.1078E+04,
     + 5.1238E+04, 5.1399E+04, 5.1560E+04, 5.1721E+04, 5.1883E+04,
     +  5.2045E+04, 5.2207E+04, 5.2370E+04, 5.2534E+04, 5.2697E+04,
     + 5.2861E+04, 5.3026E+04, 5.3191E+04, 5.3356E+04, 5.3522E+04,
     +  5.3688E+04, 5.3854E+04, 5.4021E+04, 5.4188E+04, 5.4356E+04,
     + 5.4524E+04, 5.4693E+04, 5.4862E+04, 5.5031E+04, 5.5201E+04,
     +  5.5371E+04, 5.5542E+04, 5.5713E+04, 5.5884E+04, 5.6056E+04,
     + 5.6228E+04, 5.6401E+04, 5.6574E+04, 5.6747E+04, 5.6921E+04,
     +  5.7095E+04, 5.7270E+04, 5.7445E+04, 5.7621E+04, 5.7797E+04,
     + 5.7973E+04, 5.8150E+04, 5.8327E+04, 5.8505E+04, 5.8683E+04,
     +  5.8862E+04, 5.9041E+04, 5.9220E+04, 5.9400E+04, 5.9580E+04,
     + 5.9761E+04, 5.9942E+04, 6.0123E+04, 6.0305E+04, 6.0488E+04,
     +  6.0671E+04, 6.0854E+04, 6.1038E+04, 6.1222E+04, 6.1406E+04,
     + 6.1591E+04, 6.1777E+04, 6.1963E+04, 6.2149E+04, 6.2336E+04,
     +  6.2523E+04, 6.2710E+04, 6.2899E+04, 6.3087E+04, 6.3276E+04,
     + 6.3466E+04, 6.3656E+04, 6.3846E+04, 6.4037E+04, 6.4228E+04,
     +  6.4420E+04, 6.4611E+04, 6.4804E+04, 6.4997E+04, 6.5191E+04,
     + 6.5385E+04, 6.5579E+04, 6.5774E+04, 6.5969E+04, 6.6165E+04,
     +  6.6361E+04, 6.6558E+04, 6.6755E+04, 6.6952E+04, 6.7151E+04,
     + 6.7349E+04, 6.7548E+04, 6.7747E+04, 6.7947E+04, 6.8148E+04,
     +  6.8348E+04, 6.8550E+04, 6.8751E+04, 6.8954E+04, 6.9156E+04,
     + 6.9359E+04, 6.9563E+04, 6.9767E+04, 6.9972E+04, 7.0177E+04,
     +  7.0382E+04, 7.0588E+04, 7.0794E+04, 7.1002E+04, 7.1209E+04,
     + 7.1417E+04, 7.1625E+04, 7.1834E+04, 7.2043E+04, 7.2253E+04,
     +  7.2463E+04, 7.2674E+04, 7.2885E+04, 7.3097E+04, 7.3309E+04,
     + 7.3522E+04, 7.3735E+04, 7.3949E+04, 7.4163E+04, 7.4378E+04,
     +  7.4593E+04, 7.4808E+04, 7.5024E+04, 7.5241E+04, 7.5458E+04,
     + 7.5676E+04, 7.5894E+04, 7.6112E+04, 7.6331E+04, 7.6551E+04,
     +  7.6771E+04, 7.6991E+04, 7.7213E+04, 7.7434E+04, 7.7656E+04,
     + 7.7879E+04, 7.8102E+04, 7.8325E+04, 7.8549E+04, 7.8774E+04,
     +  7.8999E+04, 7.9224E+04, 7.9451E+04, 7.9677E+04, 7.9904E+04,
     + 8.0132E+04, 8.0360E+04, 8.0589E+04, 8.0818E+04, 8.1048E+04,
     +  8.1278E+04, 8.1508E+04, 8.1740E+04, 8.1971E+04, 8.2203E+04,
     + 8.2436E+04, 8.2669E+04, 8.2903E+04, 8.3138E+04, 8.3373E+04,
     +  8.3608E+04, 8.3844E+04, 8.4080E+04, 8.4317E+04, 8.4554E+04,
     + 8.4792E+04, 8.5031E+04, 8.5270E+04, 8.5510E+04, 8.5749E+04,
     +  8.5990E+04, 8.6232E+04, 8.6473E+04, 8.6715E+04, 8.6958E+04,
     + 8.7201E+04, 8.7445E+04, 8.7689E+04, 8.7934E+04, 8.8180E+04,
     +  8.8426E+04, 8.8672E+04, 8.8919E+04, 8.9167E+04, 8.9415E+04,
     + 8.9664E+04, 8.9913E+04, 9.0163E+04, 9.0413E+04, 9.0664E+04,
     +  9.0916E+04, 9.1167E+04, 9.1420E+04, 9.1673E+04, 9.1926E+04,
     + 9.2181E+04, 9.2435E+04, 9.2691E+04, 9.2947E+04, 9.3203E+04,
     +  9.3460E+04, 9.3717E+04, 9.3975E+04, 9.4234E+04, 9.4493E+04,
     + 9.4753E+04, 9.5013E+04, 9.5274E+04, 9.5536E+04, 9.5798E+04,
     +  9.6060E+04, 9.6323E+04, 9.6587E+04, 9.6851E+04, 9.7116E+04,
     + 9.7381E+04, 9.7647E+04, 9.7914E+04, 9.8181E+04, 9.8449E+04,
     +  9.8717E+04, 9.8986E+04, 9.9255E+04, 9.9525E+04, 9.9796E+04,
     + 1.0007E+05, 1.0034E+05, 1.0061E+05, 1.0088E+05, 1.0116E+05,
     +  1.0143E+05, 1.0171E+05, 1.0198E+05, 1.0226E+05, 1.0253E+05,
     + 1.0281E+05, 1.0309E+05, 1.0337E+05, 1.0365E+05, 1.0392E+05,
     +  1.0420E+05, 1.0448E+05, 1.0477E+05, 1.0505E+05, 1.0533E+05,
     + 1.0561E+05, 1.0590E+05, 1.0618E+05, 1.0647E+05, 1.0675E+05,
     +  1.0704E+05, 1.0732E+05, 1.0761E+05, 1.0790E+05, 1.0819E+05,
     + 1.0848E+05, 1.0877E+05, 1.0906E+05, 1.0935E+05, 1.0964E+05,
     +  1.0993E+05, 1.1022E+05, 1.1051E+05, 1.1081E+05, 1.1110E+05,
     + 1.1140E+05, 1.1169E+05, 1.1199E+05, 1.1229E+05, 1.1258E+05,
     +  1.1288E+05, 1.1318E+05, 1.1348E+05, 1.1378E+05, 1.1408E+05,
     + 1.1438E+05, 1.1469E+05, 1.1499E+05, 1.1529E+05, 1.1559E+05,
     +  1.1590E+05, 1.1620E+05, 1.1651E+05, 1.1682E+05, 1.1712E+05,
     + 1.1743E+05, 1.1774E+05, 1.1805E+05, 1.1836E+05, 1.1867E+05,
     +  1.1898E+05, 1.1929E+05, 1.1960E+05, 1.1991E+05, 1.2023E+05,
     + 1.2054E+05, 1.2086E+05, 1.2117E+05, 1.2149E+05, 1.2180E+05,
     +  1.2212E+05, 1.2244E+05, 1.2276E+05, 1.2308E+05, 1.2340E+05,
     + 1.2372E+05, 1.2404E+05, 1.2436E+05, 1.2468E+05, 1.2501E+05,
     +  1.2533E+05, 1.2565E+05, 1.2598E+05, 1.2630E+05, 1.2663E+05,
     + 1.2696E+05, 1.2729E+05, 1.2761E+05, 1.2794E+05, 1.2827E+05,
     +  1.2860E+05, 1.2893E+05, 1.2927E+05, 1.2960E+05, 1.2993E+05,
     + 1.3027E+05, 1.3060E+05, 1.3094E+05, 1.3127E+05, 1.3161E+05,
     +  1.3195E+05, 1.3228E+05, 1.3262E+05, 1.3296E+05, 1.3330E+05,
     + 1.3364E+05, 1.3398E+05, 1.3432E+05, 1.3467E+05, 1.3501E+05,
     +  1.3535E+05, 1.3570E+05, 1.3604E+05, 1.3639E+05, 1.3674E+05,
     + 1.3708E+05, 1.3743E+05, 1.3778E+05, 1.3813E+05, 1.3848E+05,
     +  1.3883E+05, 1.3918E+05, 1.3954E+05, 1.3989E+05, 1.4024E+05,
     + 1.4060E+05, 1.4095E+05, 1.4131E+05, 1.4167E+05, 1.4202E+05,
     +  1.4238E+05, 1.4274E+05, 1.4310E+05, 1.4346E+05, 1.4382E+05,
     + 1.4418E+05, 1.4454E+05, 1.4491E+05, 1.4527E+05, 1.4563E+05,
     +  1.4600E+05, 1.4637E+05, 1.4673E+05, 1.4710E+05, 1.4747E+05,
     + 1.4784E+05, 1.4821E+05, 1.4858E+05, 1.4895E+05, 1.4932E+05,
     +  1.4969E+05, 1.5006E+05, 1.5044E+05, 1.5081E+05, 1.5119E+05,
     + 1.5156E+05, 1.5194E+05, 1.5232E+05, 1.5270E+05, 1.5308E+05,
     +  1.5346E+05, 1.5384E+05, 1.5422E+05, 1.5460E+05, 1.5498E+05,
     + 1.5537E+05, 1.5575E+05, 1.5613E+05, 1.5652E+05, 1.5691E+05,
     +  1.5729E+05, 1.5768E+05, 1.5807E+05, 1.5846E+05, 1.5885E+05,
     + 1.5924E+05, 1.5963E+05, 1.6003E+05, 1.6042E+05, 1.6081E+05,
     +  1.6121E+05, 1.6160E+05, 1.6200E+05, 1.6240E+05, 1.6279E+05,
     + 1.6319E+05, 1.6359E+05, 1.6399E+05, 1.6439E+05, 1.6480E+05,
     +  1.6520E+05, 1.6560E+05, 1.6600E+05, 1.6641E+05, 1.6682E+05,
     + 1.6722E+05, 1.6763E+05, 1.6804E+05, 1.6845E+05, 1.6885E+05,
     +  1.6926E+05, 1.6968E+05, 1.7009E+05, 1.7050E+05, 1.7091E+05,
     + 1.7133E+05, 1.7174E+05, 1.7216E+05, 1.7257E+05, 1.7299E+05,
     +  1.7341E+05, 1.7383E+05, 1.7425E+05, 1.7467E+05, 1.7509E+05,
     + 1.7551E+05, 1.7594E+05, 1.7636E+05, 1.7678E+05, 1.7721E+05,
     +  1.7764E+05, 1.7806E+05, 1.7849E+05, 1.7892E+05, 1.7935E+05,
     + 1.7978E+05, 1.8021E+05, 1.8064E+05, 1.8107E+05, 1.8151E+05,
     +  1.8194E+05, 1.8238E+05, 1.8281E+05, 1.8325E+05, 1.8369E+05,
     + 1.8413E+05, 1.8456E+05, 1.8500E+05, 1.8544E+05, 1.8589E+05,
     +  1.8633E+05, 1.8677E+05, 1.8722E+05, 1.8766E+05, 1.8811E+05,
     + 1.8855E+05, 1.8900E+05, 1.8945E+05, 1.8990E+05, 1.9035E+05,
     +  1.9080E+05, 1.9125E+05, 1.9170E+05, 1.9216E+05, 1.9261E+05,
     + 1.9307E+05, 1.9352E+05, 1.9398E+05, 1.9444E+05, 1.9489E+05,
     +  1.9535E+05, 1.9581E+05, 1.9627E+05, 1.9674E+05, 1.9720E+05,
     + 1.9766E+05, 1.9813E+05, 1.9859E+05, 1.9906E+05, 1.9952E+05,
     +  1.9999E+05, 2.0046E+05, 2.0093E+05, 2.0140E+05, 2.0187E+05,
     + 2.0234E+05, 2.0282E+05, 2.0329E+05, 2.0377E+05, 2.0424E+05,
     +  2.0472E+05, 2.0520E+05, 2.0567E+05, 2.0615E+05, 2.0663E+05,
     + 2.0711E+05, 2.0760E+05, 2.0808E+05, 2.0856E+05, 2.0905E+05,
     +  2.0953E+05, 2.1002E+05, 2.1050E+05, 2.1099E+05, 2.1148E+05,
     + 2.1197E+05, 2.1246E+05, 2.1295E+05, 2.1345E+05, 2.1394E+05,
     +  2.1443E+05, 2.1493E+05, 2.1542E+05, 2.1592E+05, 2.1642E+05,
     + 2.1692E+05, 2.1742E+05, 2.1792E+05, 2.1842E+05, 2.1892E+05,
     +  2.1942E+05, 2.1993E+05, 2.2043E+05, 2.2094E+05, 2.2145E+05,
     + 2.2195E+05, 2.2246E+05, 2.2297E+05, 2.2348E+05, 2.2399E+05,
     +  2.2451E+05, 2.2502E+05, 2.2553E+05, 2.2605E+05, 2.2656E+05,
     + 2.2708E+05, 2.2760E+05, 2.2812E+05, 2.2864E+05, 2.2916E+05,
     +  2.2968E+05, 2.3020E+05, 2.3073E+05, 2.3125E+05, 2.3178E+05,
     + 2.3230E+05, 2.3283E+05, 2.3336E+05, 2.3389E+05, 2.3442E+05,
     +  2.3495E+05, 2.3548E+05, 2.3601E+05, 2.3655E+05, 2.3708E+05,
     + 2.3762E+05, 2.3815E+05, 2.3869E+05, 2.3923E+05, 2.3977E+05,
     +  2.4031E+05, 2.4085E+05, 2.4139E+05, 2.4194E+05, 2.4248E+05,
     + 2.4303E+05, 2.4357E+05, 2.4412E+05, 2.4467E+05, 2.4522E+05,
     +  2.4577E+05, 2.4632E+05, 2.4687E+05, 2.4742E+05, 2.4798E+05,
     + 2.4853E+05, 2.4909E+05, 2.4965E+05, 2.5020E+05, 2.5076E+05,
     +  2.5132E+05, 2.5188E+05, 2.5244E+05, 2.5301E+05, 2.5357E+05,
     + 2.5414E+05, 2.5470E+05, 2.5527E+05, 2.5584E+05, 2.5640E+05,
     +  2.5697E+05, 2.5755E+05, 2.5812E+05, 2.5869E+05, 2.5926E+05,
     + 2.5984E+05, 2.6041E+05, 2.6099E+05, 2.6157E+05, 2.6215E+05,
     +  2.6273E+05, 2.6331E+05, 2.6389E+05, 2.6447E+05, 2.6506E+05,
     + 2.6564E+05, 2.6623E+05, 2.6681E+05, 2.6740E+05, 2.6799E+05,
     +  2.6858E+05, 2.6917E+05, 2.6976E+05, 2.7036E+05, 2.7095E+05,
     + 2.7154E+05, 2.7214E+05, 2.7274E+05, 2.7334E+05, 2.7393E+05,
     +  2.7454E+05, 2.7514E+05, 2.7574E+05, 2.7634E+05, 2.7695E+05,
     + 2.7755E+05, 2.7816E+05, 2.7876E+05, 2.7937E+05, 2.7998E+05,
     +  2.8059E+05, 2.8120E+05, 2.8182E+05, 2.8243E+05, 2.8305E+05,
     + 2.8366E+05, 2.8428E+05, 2.8490E+05, 2.8551E+05, 2.8613E+05,
     +  2.8676E+05, 2.8738E+05, 2.8800E+05, 2.8863E+05, 2.8925E+05,
     + 2.8988E+05, 2.9050E+05, 2.9113E+05, 2.9176E+05, 2.9239E+05,
     +  2.9302E+05, 2.9366E+05, 2.9429E+05, 2.9493E+05, 2.9556E+05,
     + 2.9620E+05, 2.9684E+05, 2.9748E+05, 2.9812E+05, 2.9876E+05,
     +  2.9940E+05, 3.0004E+05, 3.0069E+05, 3.0133E+05, 3.0198E+05,
     + 3.0263E+05, 3.0328E+05, 3.0393E+05, 3.0458E+05, 3.0523E+05,
     +  3.0588E+05, 3.0654E+05, 3.0719E+05, 3.0785E+05, 3.0851E+05,
     + 3.0917E+05, 3.0983E+05, 3.1049E+05, 3.1115E+05, 3.1181E+05,
     +  3.1248E+05, 3.1314E+05, 3.1381E+05, 3.1448E+05, 3.1515E+05,
     + 3.1582E+05, 3.1649E+05, 3.1716E+05, 3.1783E+05, 3.1851E+05,
     +  3.1918E+05, 3.1986E+05, 3.2054E+05, 3.2122E+05, 3.2190E+05,
     + 3.2258E+05, 3.2326E+05, 3.2394E+05, 3.2463E+05, 3.2531E+05,
     +  3.2600E+05, 3.2669E+05, 3.2738E+05, 3.2807E+05, 3.2876E+05,
     + 3.2945E+05, 3.3015E+05, 3.3084E+05, 3.3154E+05, 3.3223E+05,
     +  3.3293E+05, 3.3363E+05, 3.3433E+05, 3.3503E+05, 3.3574E+05,
     + 3.3644E+05, 3.3715E+05, 3.3785E+05, 3.3856E+05, 3.3927E+05,
     +  3.3998E+05, 3.4069E+05, 3.4140E+05, 3.4212E+05, 3.4283E+05,
     + 3.4355E+05, 3.4426E+05, 3.4498E+05, 3.4570E+05, 3.4642E+05,
     +  3.4714E+05, 3.4786E+05, 3.4859E+05, 3.4931E+05, 3.5004E+05,
     + 3.5077E+05, 3.5150E+05, 3.5223E+05, 3.5296E+05, 3.5369E+05,
     +  3.5442E+05, 3.5516E+05, 3.5589E+05, 3.5663E+05, 3.5737E+05,
     + 3.5811E+05, 3.5885E+05, 3.5959E+05, 3.6033E+05, 3.6108E+05,
     +  3.6182E+05, 3.6257E+05, 3.6332E+05, 3.6407E+05, 3.6482E+05,
     + 3.6557E+05, 3.6632E+05, 3.6708E+05, 3.6783E+05, 3.6859E+05,
     +  3.6934E+05, 3.7010E+05, 3.7086E+05, 3.7163E+05, 3.7239E+05,
     + 3.7315E+05, 3.7392E+05, 3.7468E+05, 3.7545E+05, 3.7622E+05,
     +  3.7699E+05, 3.7776E+05, 3.7853E+05, 3.7931E+05, 3.8008E+05,
     + 3.8086E+05, 3.8163E+05, 3.8241E+05, 3.8319E+05, 3.8397E+05,
     +  3.8476E+05, 3.8554E+05, 3.8632E+05, 3.8711E+05, 3.8790E+05,
     + 3.8869E+05, 3.8948E+05, 3.9027E+05, 3.9106E+05, 3.9185E+05,
     +  3.9265E+05, 3.9344E+05, 3.9424E+05, 3.9504E+05, 3.9584E+05,
     + 3.9664E+05, 3.9744E+05, 3.9825E+05, 3.9905E+05, 3.9986E+05,
     +  4.0067E+05, 4.0147E+05, 4.0228E+05, 4.0310E+05, 4.0391E+05,
     + 4.0472E+05, 4.0554E+05, 4.0635E+05, 4.0717E+05, 4.0799E+05,
     +  4.0881E+05, 4.0963E+05, 4.1046E+05, 4.1128E+05, 4.1211E+05,
     + 4.1293E+05, 4.1376E+05, 4.1459E+05, 4.1542E+05, 4.1625E+05,
     +  4.1709E+05, 4.1792E+05, 4.1876E+05, 4.1960E+05, 4.2043E+05,
     + 4.2127E+05, 4.2212E+05, 4.2296E+05, 4.2380E+05, 4.2465E+05,
     +  4.2549E+05, 4.2634E+05, 4.2719E+05, 4.2804E+05, 4.2889E+05,
     + 4.2975E+05, 4.3060E+05, 4.3146E+05, 4.3231E+05, 4.3317E+05,
     +  4.3403E+05, 4.3489E+05, 4.3576E+05, 4.3662E+05, 4.3749E+05,
     + 4.3835E+05, 4.3922E+05, 4.4009E+05, 4.4096E+05, 4.4183E+05,
     +  4.4271E+05, 4.4358E+05, 4.4446E+05, 4.4533E+05, 4.4621E+05,
     + 4.4709E+05, 4.4797E+05, 4.4886E+05, 4.4974E+05, 4.5063E+05,
     +  4.5151E+05, 4.5240E+05, 4.5329E+05, 4.5418E+05, 4.5507E+05,
     + 4.5597E+05, 4.5686E+05, 4.5776E+05, 4.5866E+05, 4.5956E+05,
     +  4.6046E+05, 4.6136E+05, 4.6226E+05, 4.6317E+05, 4.6408E+05,
     + 4.6498E+05, 4.6589E+05, 4.6680E+05, 4.6771E+05, 4.6863E+05,
     +  4.6954E+05, 4.7046E+05, 4.7137E+05, 4.7229E+05, 4.7321E+05,
     + 4.7414E+05, 4.7506E+05, 4.7598E+05, 4.7691E+05, 4.7784E+05,
     +  4.7876E+05, 4.7969E+05, 4.8063E+05, 4.8156E+05, 4.8249E+05,
     + 4.8343E+05, 4.8437E+05, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 3,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_HCOOH (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 1),QofT( 1,NT),Q(NT)
      data Tmx(1)/3500./
      data xgj/ 4./
c...    HCOOH
c...        --       126
      data (QofT( 1,J),J=1,2408)/ 8.0985E+00, 2.1053E+01, 3.7588E+01,
     +  5.7053E+01, 7.9059E+01, 1.0334E+02, 1.2970E+02, 1.5799E+02,
     + 1.8808E+02, 2.1988E+02, 2.5329E+02, 2.8824E+02, 3.2467E+02,
     +  3.6251E+02, 4.0172E+02, 4.4226E+02, 4.8407E+02, 5.2713E+02,
     + 5.7139E+02, 6.1683E+02, 7.1111E+02, 8.0977E+02, 9.1262E+02,
     +  1.0195E+03, 1.1303E+03, 1.2448E+03, 1.3629E+03, 1.4846E+03,
     + 1.6096E+03, 1.7381E+03, 1.8697E+03, 2.0046E+03, 2.1425E+03,
     +  2.2835E+03, 2.4275E+03, 2.5744E+03, 2.7241E+03, 2.8766E+03,
     + 3.0319E+03, 3.1899E+03, 3.3505E+03, 3.5138E+03, 3.6796E+03,
     +  3.8480E+03, 4.0189E+03, 4.1923E+03, 4.3681E+03, 4.5463E+03,
     + 4.7269E+03, 4.9098E+03, 5.0951E+03, 5.2826E+03, 5.4724E+03,
     +  5.6645E+03, 5.8588E+03, 6.0552E+03, 6.2540E+03, 6.4548E+03,
     + 6.6578E+03, 6.8630E+03, 7.0702E+03, 7.2796E+03, 7.4912E+03,
     +  7.7048E+03, 7.9205E+03, 8.1383E+03, 8.3581E+03, 8.5800E+03,
     + 8.8041E+03, 9.0301E+03, 9.2582E+03, 9.4884E+03, 9.7207E+03,
     +  9.9551E+03, 1.0191E+04, 1.0430E+04, 1.0671E+04, 1.0913E+04,
     + 1.1158E+04, 1.1405E+04, 1.1654E+04, 1.1905E+04, 1.2158E+04,
     +  1.2414E+04, 1.2671E+04, 1.2931E+04, 1.3193E+04, 1.3457E+04,
     + 1.3724E+04, 1.3993E+04, 1.4264E+04, 1.4537E+04, 1.4812E+04,
     +  1.5090E+04, 1.5371E+04, 1.5653E+04, 1.5938E+04, 1.6226E+04,
     + 1.6516E+04, 1.6808E+04, 1.7103E+04, 1.7400E+04, 1.7700E+04,
     +  1.8003E+04, 1.8308E+04, 1.8616E+04, 1.8926E+04, 1.9239E+04,
     + 1.9555E+04, 1.9874E+04, 2.0195E+04, 2.0519E+04, 2.0846E+04,
     +  2.1176E+04, 2.1509E+04, 2.1845E+04, 2.2184E+04, 2.2525E+04,
     + 2.2870E+04, 2.3218E+04, 2.3569E+04, 2.3923E+04, 2.4281E+04,
     +  2.4642E+04, 2.5005E+04, 2.5373E+04, 2.5743E+04, 2.6117E+04,
     + 2.6495E+04, 2.6876E+04, 2.7260E+04, 2.7648E+04, 2.8040E+04,
     +  2.8435E+04, 2.8834E+04, 2.9236E+04, 2.9643E+04, 3.0053E+04,
     + 3.0467E+04, 3.0885E+04, 3.1307E+04, 3.1733E+04, 3.2163E+04,
     +  3.2597E+04, 3.3036E+04, 3.3478E+04, 3.3925E+04, 3.4375E+04,
     + 3.4831E+04, 3.5291E+04, 3.5755E+04, 3.6223E+04, 3.6697E+04,
     +  3.7174E+04, 3.7657E+04, 3.8144E+04, 3.8636E+04, 3.9133E+04,
     + 3.9634E+04, 4.0141E+04, 4.0652E+04, 4.1169E+04, 4.1691E+04,
     +  4.2217E+04, 4.2750E+04, 4.3287E+04, 4.3829E+04, 4.4377E+04,
     + 4.4931E+04, 4.5490E+04, 4.6054E+04, 4.6624E+04, 4.7200E+04,
     +  4.7782E+04, 4.8370E+04, 4.8963E+04, 4.9562E+04, 5.0168E+04,
     + 5.0779E+04, 5.1396E+04, 5.2020E+04, 5.2650E+04, 5.3287E+04,
     +  5.3930E+04, 5.4579E+04, 5.5235E+04, 5.5898E+04, 5.6567E+04,
     + 5.7243E+04, 5.7926E+04, 5.8616E+04, 5.9312E+04, 6.0016E+04,
     +  6.0727E+04, 6.1445E+04, 6.2171E+04, 6.2904E+04, 6.3645E+04,
     + 6.4392E+04, 6.5148E+04, 6.5912E+04, 6.6682E+04, 6.7462E+04,
     +  6.8248E+04, 6.9043E+04, 6.9846E+04, 7.0658E+04, 7.1477E+04,
     + 7.2305E+04, 7.3141E+04, 7.3986E+04, 7.4840E+04, 7.5702E+04,
     +  7.6573E+04, 7.7453E+04, 7.8342E+04, 7.9240E+04, 8.0148E+04,
     + 8.1064E+04, 8.1990E+04, 8.2926E+04, 8.3870E+04, 8.4825E+04,
     +  8.5789E+04, 8.6764E+04, 8.7748E+04, 8.8742E+04, 8.9746E+04,
     + 9.0761E+04, 9.1786E+04, 9.2822E+04, 9.3867E+04, 9.4924E+04,
     +  9.5992E+04, 9.7070E+04, 9.8160E+04, 9.9260E+04, 1.0037E+05,
     + 1.0149E+05, 1.0263E+05, 1.0378E+05, 1.0493E+05, 1.0610E+05,
     +  1.0729E+05, 1.0848E+05, 1.0968E+05, 1.1090E+05, 1.1213E+05,
     + 1.1338E+05, 1.1463E+05, 1.1590E+05, 1.1718E+05, 1.1848E+05,
     +  1.1978E+05, 1.2110E+05, 1.2244E+05, 1.2378E+05, 1.2515E+05,
     + 1.2652E+05, 1.2791E+05, 1.2931E+05, 1.3073E+05, 1.3216E+05,
     +  1.3361E+05, 1.3507E+05, 1.3654E+05, 1.3803E+05, 1.3954E+05,
     + 1.4106E+05, 1.4259E+05, 1.4415E+05, 1.4571E+05, 1.4729E+05,
     +  1.4889E+05, 1.5051E+05, 1.5214E+05, 1.5379E+05, 1.5545E+05,
     + 1.5713E+05, 1.5883E+05, 1.6054E+05, 1.6227E+05, 1.6402E+05,
     +  1.6579E+05, 1.6757E+05, 1.6937E+05, 1.7119E+05, 1.7303E+05,
     + 1.7489E+05, 1.7676E+05, 1.7866E+05, 1.8057E+05, 1.8250E+05,
     +  1.8445E+05, 1.8642E+05, 1.8841E+05, 1.9042E+05, 1.9245E+05,
     + 1.9450E+05, 1.9657E+05, 1.9866E+05, 2.0077E+05, 2.0290E+05,
     +  2.0505E+05, 2.0723E+05, 2.0942E+05, 2.1164E+05, 2.1388E+05,
     + 2.1614E+05, 2.1842E+05, 2.2073E+05, 2.2306E+05, 2.2541E+05,
     +  2.2778E+05, 2.3018E+05, 2.3260E+05, 2.3505E+05, 2.3752E+05,
     + 2.4001E+05, 2.4253E+05, 2.4507E+05, 2.4764E+05, 2.5023E+05,
     +  2.5285E+05, 2.5549E+05, 2.5816E+05, 2.6085E+05, 2.6357E+05,
     + 2.6632E+05, 2.6909E+05, 2.7190E+05, 2.7472E+05, 2.7758E+05,
     +  2.8046E+05, 2.8337E+05, 2.8631E+05, 2.8928E+05, 2.9228E+05,
     + 2.9530E+05, 2.9836E+05, 3.0144E+05, 3.0455E+05, 3.0770E+05,
     +  3.1087E+05, 3.1408E+05, 3.1731E+05, 3.2058E+05, 3.2387E+05,
     + 3.2720E+05, 3.3057E+05, 3.3396E+05, 3.3738E+05, 3.4084E+05,
     +  3.4433E+05, 3.4786E+05, 3.5142E+05, 3.5501E+05, 3.5864E+05,
     + 3.6230E+05, 3.6599E+05, 3.6972E+05, 3.7349E+05, 3.7729E+05,
     +  3.8113E+05, 3.8501E+05, 3.8892E+05, 3.9287E+05, 3.9685E+05,
     + 4.0087E+05, 4.0494E+05, 4.0904E+05, 4.1317E+05, 4.1735E+05,
     +  4.2157E+05, 4.2582E+05, 4.3012E+05, 4.3446E+05, 4.3883E+05,
     + 4.4325E+05, 4.4771E+05, 4.5221E+05, 4.5676E+05, 4.6134E+05,
     +  4.6597E+05, 4.7065E+05, 4.7536E+05, 4.8012E+05, 4.8493E+05,
     + 4.8978E+05, 4.9467E+05, 4.9961E+05, 5.0459E+05, 5.0963E+05,
     +  5.1470E+05, 5.1983E+05, 5.2500E+05, 5.3022E+05, 5.3549E+05,
     + 5.4081E+05, 5.4618E+05, 5.5159E+05, 5.5706E+05, 5.6258E+05,
     +  5.6814E+05, 5.7376E+05, 5.7943E+05, 5.8516E+05, 5.9093E+05,
     + 5.9676E+05, 6.0264E+05, 6.0858E+05, 6.1456E+05, 6.2061E+05,
     +  6.2671E+05, 6.3286E+05, 6.3908E+05, 6.4534E+05, 6.5167E+05,
     + 6.5805E+05, 6.6449E+05, 6.7099E+05, 6.7755E+05, 6.8417E+05,
     +  6.9085E+05, 6.9759E+05, 7.0438E+05, 7.1125E+05, 7.1817E+05,
     + 7.2516E+05, 7.3220E+05, 7.3932E+05, 7.4649E+05, 7.5373E+05,
     +  7.6104E+05, 7.6841E+05, 7.7585E+05, 7.8336E+05, 7.9093E+05,
     + 7.9857E+05, 8.0628E+05, 8.1406E+05, 8.2190E+05, 8.2982E+05,
     +  8.3781E+05, 8.4587E+05, 8.5400E+05, 8.6221E+05, 8.7048E+05,
     + 8.7883E+05, 8.8726E+05, 8.9576E+05, 9.0434E+05, 9.1299E+05,
     +  9.2172E+05, 9.3052E+05, 9.3941E+05, 9.4837E+05, 9.5741E+05,
     + 9.6653E+05, 9.7573E+05, 9.8502E+05, 9.9438E+05, 1.0038E+06,
     +  1.0134E+06, 1.0230E+06, 1.0327E+06, 1.0425E+06, 1.0523E+06,
     + 1.0623E+06, 1.0723E+06, 1.0825E+06, 1.0927E+06, 1.1030E+06,
     +  1.1134E+06, 1.1239E+06, 1.1345E+06, 1.1451E+06, 1.1559E+06,
     + 1.1668E+06, 1.1777E+06, 1.1888E+06, 1.1999E+06, 1.2112E+06,
     +  1.2225E+06, 1.2340E+06, 1.2455E+06, 1.2571E+06, 1.2689E+06,
     + 1.2807E+06, 1.2927E+06, 1.3047E+06, 1.3168E+06, 1.3291E+06,
     +  1.3415E+06, 1.3539E+06, 1.3665E+06, 1.3792E+06, 1.3920E+06,
     + 1.4049E+06, 1.4179E+06, 1.4310E+06, 1.4442E+06, 1.4576E+06,
     +  1.4711E+06, 1.4846E+06, 1.4983E+06, 1.5121E+06, 1.5261E+06,
     + 1.5401E+06, 1.5543E+06, 1.5686E+06, 1.5830E+06, 1.5975E+06,
     +  1.6121E+06, 1.6269E+06, 1.6418E+06, 1.6568E+06, 1.6720E+06,
     + 1.6873E+06, 1.7027E+06, 1.7182E+06, 1.7339E+06, 1.7497E+06,
     +  1.7656E+06, 1.7817E+06, 1.7979E+06, 1.8143E+06, 1.8308E+06,
     + 1.8474E+06, 1.8641E+06, 1.8810E+06, 1.8981E+06, 1.9152E+06,
     +  1.9326E+06, 1.9500E+06, 1.9677E+06, 1.9854E+06, 2.0033E+06,
     + 2.0214E+06, 2.0396E+06, 2.0580E+06, 2.0765E+06, 2.0951E+06,
     +  2.1140E+06, 2.1330E+06, 2.1521E+06, 2.1714E+06, 2.1908E+06,
     + 2.2105E+06, 2.2302E+06, 2.2502E+06, 2.2703E+06, 2.2905E+06,
     +  2.3110E+06, 2.3316E+06, 2.3524E+06, 2.3733E+06, 2.3944E+06,
     + 2.4157E+06, 2.4372E+06, 2.4588E+06, 2.4807E+06, 2.5027E+06,
     +  2.5248E+06, 2.5472E+06, 2.5697E+06, 2.5925E+06, 2.6154E+06,
     + 2.6385E+06, 2.6617E+06, 2.6852E+06, 2.7089E+06, 2.7327E+06,
     +  2.7568E+06, 2.7810E+06, 2.8055E+06, 2.8301E+06, 2.8550E+06,
     + 2.8800E+06, 2.9053E+06, 2.9307E+06, 2.9564E+06, 2.9822E+06,
     +  3.0083E+06, 3.0346E+06, 3.0611E+06, 3.0878E+06, 3.1147E+06,
     + 3.1418E+06, 3.1692E+06, 3.1967E+06, 3.2245E+06, 3.2525E+06,
     +  3.2808E+06, 3.3092E+06, 3.3379E+06, 3.3669E+06, 3.3960E+06,
     + 3.4254E+06, 3.4550E+06, 3.4849E+06, 3.5150E+06, 3.5453E+06,
     +  3.5759E+06, 3.6067E+06, 3.6377E+06, 3.6690E+06, 3.7006E+06,
     + 3.7324E+06, 3.7644E+06, 3.7967E+06, 3.8293E+06, 3.8621E+06,
     +  3.8952E+06, 3.9285E+06, 3.9622E+06, 3.9960E+06, 4.0301E+06,
     + 4.0646E+06, 4.0992E+06, 4.1342E+06, 4.1694E+06, 4.2049E+06,
     +  4.2406E+06, 4.2767E+06, 4.3130E+06, 4.3497E+06, 4.3866E+06,
     + 4.4238E+06, 4.4612E+06, 4.4990E+06, 4.5371E+06, 4.5754E+06,
     +  4.6141E+06, 4.6531E+06, 4.6923E+06, 4.7319E+06, 4.7718E+06,
     + 4.8120E+06, 4.8525E+06, 4.8933E+06, 4.9344E+06, 4.9758E+06,
     +  5.0176E+06, 5.0597E+06, 5.1021E+06, 5.1448E+06, 5.1879E+06,
     + 5.2313E+06, 5.2750E+06, 5.3191E+06, 5.3635E+06, 5.4083E+06,
     +  5.4534E+06, 5.4988E+06, 5.5446E+06, 5.5907E+06, 5.6372E+06,
     + 5.6840E+06, 5.7312E+06, 5.7788E+06, 5.8267E+06, 5.8750E+06,
     +  5.9237E+06, 5.9727E+06, 6.0221E+06, 6.0719E+06, 6.1220E+06,
     + 6.1726E+06, 6.2235E+06, 6.2748E+06, 6.3265E+06, 6.3785E+06,
     +  6.4310E+06, 6.4839E+06, 6.5372E+06, 6.5908E+06, 6.6449E+06,
     + 6.6994E+06, 6.7543E+06, 6.8096E+06, 6.8653E+06, 6.9215E+06,
     +  6.9780E+06, 7.0350E+06, 7.0924E+06, 7.1503E+06, 7.2086E+06,
     + 7.2673E+06, 7.3265E+06, 7.3861E+06, 7.4461E+06, 7.5066E+06,
     +  7.5675E+06, 7.6289E+06, 7.6908E+06, 7.7531E+06, 7.8159E+06,
     + 7.8791E+06, 7.9428E+06, 8.0070E+06, 8.0717E+06, 8.1368E+06,
     +  8.2025E+06, 8.2686E+06, 8.3352E+06, 8.4023E+06, 8.4699E+06,
     + 8.5379E+06, 8.6065E+06, 8.6756E+06, 8.7452E+06, 8.8154E+06,
     +  8.8860E+06, 8.9571E+06, 9.0288E+06, 9.1010E+06, 9.1738E+06,
     + 9.2470E+06, 9.3208E+06, 9.3952E+06, 9.4700E+06, 9.5455E+06,
     +  9.6215E+06, 9.6980E+06, 9.7751E+06, 9.8528E+06, 9.9310E+06,
     + 1.0010E+07, 1.0089E+07, 1.0169E+07, 1.0250E+07, 1.0331E+07,
     +  1.0412E+07, 1.0495E+07, 1.0578E+07, 1.0661E+07, 1.0745E+07,
     + 1.0830E+07, 1.0915E+07, 1.1001E+07, 1.1088E+07, 1.1175E+07,
     +  1.1263E+07, 1.1351E+07, 1.1440E+07, 1.1530E+07, 1.1620E+07,
     + 1.1711E+07, 1.1803E+07, 1.1895E+07, 1.1988E+07, 1.2082E+07,
     +  1.2176E+07, 1.2271E+07, 1.2367E+07, 1.2463E+07, 1.2560E+07,
     + 1.2658E+07, 1.2757E+07, 1.2856E+07, 1.2955E+07, 1.3056E+07,
     +  1.3157E+07, 1.3259E+07, 1.3362E+07, 1.3465E+07, 1.3570E+07,
     + 1.3675E+07, 1.3780E+07, 1.3887E+07, 1.3994E+07, 1.4102E+07,
     +  1.4210E+07, 1.4320E+07, 1.4430E+07, 1.4541E+07, 1.4653E+07,
     + 1.4765E+07, 1.4879E+07, 1.4993E+07, 1.5108E+07, 1.5224E+07,
     +  1.5340E+07, 1.5458E+07, 1.5576E+07, 1.5695E+07, 1.5815E+07,
     + 1.5936E+07, 1.6057E+07, 1.6180E+07, 1.6303E+07, 1.6427E+07,
     +  1.6552E+07, 1.6678E+07, 1.6805E+07, 1.6932E+07, 1.7061E+07,
     + 1.7190E+07, 1.7320E+07, 1.7452E+07, 1.7584E+07, 1.7717E+07,
     +  1.7851E+07, 1.7986E+07, 1.8121E+07, 1.8258E+07, 1.8396E+07,
     + 1.8535E+07, 1.8674E+07, 1.8815E+07, 1.8956E+07, 1.9099E+07,
     +  1.9242E+07, 1.9387E+07, 1.9532E+07, 1.9679E+07, 1.9826E+07,
     + 1.9975E+07, 2.0124E+07, 2.0275E+07, 2.0426E+07, 2.0579E+07,
     +  2.0733E+07, 2.0887E+07, 2.1043E+07, 2.1200E+07, 2.1358E+07,
     + 2.1517E+07, 2.1677E+07, 2.1838E+07, 2.2000E+07, 2.2164E+07,
     +  2.2328E+07, 2.2494E+07, 2.2660E+07, 2.2828E+07, 2.2997E+07,
     + 2.3167E+07, 2.3339E+07, 2.3511E+07, 2.3685E+07, 2.3859E+07,
     +  2.4035E+07, 2.4212E+07, 2.4391E+07, 2.4570E+07, 2.4751E+07,
     + 2.4933E+07, 2.5116E+07, 2.5300E+07, 2.5486E+07, 2.5673E+07,
     +  2.5861E+07, 2.6051E+07, 2.6241E+07, 2.6433E+07, 2.6626E+07,
     + 2.6821E+07, 2.7017E+07, 2.7214E+07, 2.7412E+07, 2.7612E+07,
     +  2.7813E+07, 2.8016E+07, 2.8219E+07, 2.8425E+07, 2.8631E+07,
     + 2.8839E+07, 2.9048E+07, 2.9259E+07, 2.9471E+07, 2.9684E+07,
     +  2.9899E+07, 3.0116E+07, 3.0333E+07, 3.0552E+07, 3.0773E+07,
     + 3.0995E+07, 3.1219E+07, 3.1444E+07, 3.1670E+07, 3.1898E+07,
     +  3.2128E+07, 3.2359E+07, 3.2591E+07, 3.2825E+07, 3.3061E+07,
     + 3.3298E+07, 3.3536E+07, 3.3777E+07, 3.4019E+07, 3.4262E+07,
     +  3.4507E+07, 3.4754E+07, 3.5002E+07, 3.5252E+07, 3.5503E+07,
     + 3.5756E+07, 3.6011E+07, 3.6267E+07, 3.6525E+07, 3.6785E+07,
     +  3.7047E+07, 3.7310E+07, 3.7575E+07, 3.7841E+07, 3.8109E+07,
     + 3.8379E+07, 3.8651E+07, 3.8925E+07, 3.9200E+07, 3.9477E+07,
     +  3.9756E+07, 4.0037E+07, 4.0319E+07, 4.0603E+07, 4.0890E+07,
     + 4.1178E+07, 4.1467E+07, 4.1759E+07, 4.2053E+07, 4.2348E+07,
     +  4.2646E+07, 4.2945E+07, 4.3246E+07, 4.3549E+07, 4.3854E+07,
     + 4.4161E+07, 4.4470E+07, 4.4781E+07, 4.5094E+07, 4.5409E+07,
     +  4.5726E+07, 4.6045E+07, 4.6366E+07, 4.6689E+07, 4.7014E+07,
     + 4.7341E+07, 4.7671E+07, 4.8002E+07, 4.8335E+07, 4.8671E+07,
     +  4.9009E+07, 4.9349E+07, 4.9691E+07, 5.0035E+07, 5.0381E+07,
     + 5.0730E+07, 5.1080E+07, 5.1433E+07, 5.1788E+07, 5.2146E+07,
     +  5.2506E+07, 5.2868E+07, 5.3232E+07, 5.3598E+07, 5.3967E+07,
     + 5.4338E+07, 5.4712E+07, 5.5087E+07, 5.5466E+07, 5.5846E+07,
     +  5.6229E+07, 5.6614E+07, 5.7002E+07, 5.7392E+07, 5.7785E+07,
     + 5.8180E+07, 5.8577E+07, 5.8977E+07, 5.9380E+07, 5.9785E+07,
     +  6.0192E+07, 6.0602E+07, 6.1015E+07, 6.1430E+07, 6.1848E+07,
     + 6.2268E+07, 6.2691E+07, 6.3117E+07, 6.3545E+07, 6.3976E+07,
     +  6.4410E+07, 6.4846E+07, 6.5285E+07, 6.5726E+07, 6.6171E+07,
     + 6.6618E+07, 6.7068E+07, 6.7520E+07, 6.7976E+07, 6.8434E+07,
     +  6.8895E+07, 6.9359E+07, 6.9826E+07, 7.0295E+07, 7.0768E+07,
     + 7.1243E+07, 7.1721E+07, 7.2203E+07, 7.2687E+07, 7.3174E+07,
     +  7.3664E+07, 7.4157E+07, 7.4653E+07, 7.5153E+07, 7.5655E+07,
     + 7.6160E+07, 7.6668E+07, 7.7180E+07, 7.7694E+07, 7.8212E+07,
     +  7.8733E+07, 7.9257E+07, 7.9784E+07, 8.0315E+07, 8.0848E+07,
     + 8.1385E+07, 8.1925E+07, 8.2469E+07, 8.3015E+07, 8.3565E+07,
     +  8.4119E+07, 8.4675E+07, 8.5235E+07, 8.5799E+07, 8.6365E+07,
     + 8.6936E+07, 8.7509E+07, 8.8086E+07, 8.8667E+07, 8.9251E+07,
     +  8.9839E+07, 9.0430E+07, 9.1024E+07, 9.1623E+07, 9.2224E+07,
     + 9.2830E+07, 9.3439E+07, 9.4052E+07, 9.4668E+07, 9.5288E+07,
     +  9.5912E+07, 9.6539E+07, 9.7170E+07, 9.7805E+07, 9.8444E+07,
     + 9.9087E+07, 9.9733E+07, 1.0038E+08, 1.0104E+08, 1.0170E+08,
     +  1.0236E+08, 1.0302E+08, 1.0369E+08, 1.0437E+08, 1.0504E+08,
     + 1.0573E+08, 1.0641E+08, 1.0710E+08, 1.0780E+08, 1.0849E+08,
     +  1.0920E+08, 1.0990E+08, 1.1061E+08, 1.1133E+08, 1.1205E+08,
     + 1.1277E+08, 1.1350E+08, 1.1423E+08, 1.1497E+08, 1.1571E+08,
     +  1.1645E+08, 1.1720E+08, 1.1795E+08, 1.1871E+08, 1.1947E+08,
     + 1.2024E+08, 1.2101E+08, 1.2179E+08, 1.2257E+08, 1.2335E+08,
     +  1.2414E+08, 1.2494E+08, 1.2573E+08, 1.2654E+08, 1.2735E+08,
     + 1.2816E+08, 1.2898E+08, 1.2980E+08, 1.3062E+08, 1.3146E+08,
     +  1.3229E+08, 1.3313E+08, 1.3398E+08, 1.3483E+08, 1.3569E+08,
     + 1.3655E+08, 1.3742E+08, 1.3829E+08, 1.3916E+08, 1.4004E+08,
     +  1.4093E+08, 1.4182E+08, 1.4272E+08, 1.4362E+08, 1.4453E+08,
     + 1.4544E+08, 1.4636E+08, 1.4728E+08, 1.4821E+08, 1.4914E+08,
     +  1.5008E+08, 1.5102E+08, 1.5197E+08, 1.5293E+08, 1.5389E+08,
     + 1.5486E+08, 1.5583E+08, 1.5680E+08, 1.5779E+08, 1.5877E+08,
     +  1.5977E+08, 1.6077E+08, 1.6177E+08, 1.6278E+08, 1.6380E+08,
     + 1.6482E+08, 1.6585E+08, 1.6689E+08, 1.6793E+08, 1.6897E+08,
     +  1.7002E+08, 1.7108E+08, 1.7215E+08, 1.7322E+08, 1.7429E+08,
     + 1.7537E+08, 1.7646E+08, 1.7756E+08, 1.7866E+08, 1.7976E+08,
     +  1.8088E+08, 1.8200E+08, 1.8312E+08, 1.8425E+08, 1.8539E+08,
     + 1.8654E+08, 1.8769E+08, 1.8885E+08, 1.9001E+08, 1.9118E+08,
     +  1.9236E+08, 1.9354E+08, 1.9473E+08, 1.9593E+08, 1.9713E+08,
     + 1.9834E+08, 1.9956E+08, 2.0079E+08, 2.0202E+08, 2.0325E+08,
     +  2.0450E+08, 2.0575E+08, 2.0701E+08, 2.0828E+08, 2.0955E+08,
     + 2.1083E+08, 2.1212E+08, 2.1341E+08, 2.1471E+08, 2.1602E+08,
     +  2.1734E+08, 2.1866E+08, 2.1999E+08, 2.2133E+08, 2.2267E+08,
     + 2.2403E+08, 2.2539E+08, 2.2676E+08, 2.2813E+08, 2.2951E+08,
     +  2.3090E+08, 2.3230E+08, 2.3371E+08, 2.3512E+08, 2.3655E+08,
     + 2.3797E+08, 2.3941E+08, 2.4086E+08, 2.4231E+08, 2.4377E+08,
     +  2.4524E+08, 2.4672E+08, 2.4820E+08, 2.4970E+08, 2.5120E+08,
     + 2.5271E+08, 2.5423E+08, 2.5575E+08, 2.5729E+08, 2.5883E+08,
     +  2.6038E+08, 2.6194E+08, 2.6351E+08, 2.6509E+08, 2.6668E+08,
     + 2.6827E+08, 2.6987E+08, 2.7149E+08, 2.7311E+08, 2.7474E+08,
     +  2.7638E+08, 2.7802E+08, 2.7968E+08, 2.8134E+08, 2.8302E+08,
     + 2.8470E+08, 2.8639E+08, 2.8810E+08, 2.8981E+08, 2.9153E+08,
     +  2.9326E+08, 2.9500E+08, 2.9674E+08, 2.9850E+08, 3.0027E+08,
     + 3.0205E+08, 3.0383E+08, 3.0563E+08, 3.0743E+08, 3.0925E+08,
     +  3.1107E+08, 3.1291E+08, 3.1475E+08, 3.1661E+08, 3.1847E+08,
     + 3.2035E+08, 3.2223E+08, 3.2413E+08, 3.2603E+08, 3.2795E+08,
     +  3.2987E+08, 3.3181E+08, 3.3375E+08, 3.3571E+08, 3.3768E+08,
     + 3.3965E+08, 3.4164E+08, 3.4364E+08, 3.4565E+08, 3.4767E+08,
     +  3.4970E+08, 3.5174E+08, 3.5379E+08, 3.5585E+08, 3.5793E+08,
     + 3.6001E+08, 3.6211E+08, 3.6422E+08, 3.6634E+08, 3.6846E+08,
     +  3.7061E+08, 3.7276E+08, 3.7492E+08, 3.7710E+08, 3.7928E+08,
     + 3.8148E+08, 3.8369E+08, 3.8591E+08, 3.8814E+08, 3.9039E+08,
     +  3.9265E+08, 3.9491E+08, 3.9720E+08, 3.9949E+08, 4.0179E+08,
     + 4.0411E+08, 4.0644E+08, 4.0878E+08, 4.1113E+08, 4.1350E+08,
     +  4.1588E+08, 4.1827E+08, 4.2067E+08, 4.2308E+08, 4.2551E+08,
     + 4.2795E+08, 4.3041E+08, 4.3287E+08, 4.3535E+08, 4.3784E+08,
     +  4.4035E+08, 4.4287E+08, 4.4540E+08, 4.4794E+08, 4.5050E+08,
     + 4.5307E+08, 4.5566E+08, 4.5825E+08, 4.6087E+08, 4.6349E+08,
     +  4.6613E+08, 4.6878E+08, 4.7145E+08, 4.7413E+08, 4.7682E+08,
     + 4.7953E+08, 4.8225E+08, 4.8499E+08, 4.8774E+08, 4.9050E+08,
     +  4.9328E+08, 4.9607E+08, 4.9888E+08, 5.0170E+08, 5.0453E+08,
     + 5.0738E+08, 5.1025E+08, 5.1313E+08, 5.1602E+08, 5.1893E+08,
     +  5.2186E+08, 5.2480E+08, 5.2775E+08, 5.3072E+08, 5.3371E+08,
     + 5.3671E+08, 5.3972E+08, 5.4275E+08, 5.4580E+08, 5.4886E+08,
     +  5.5194E+08, 5.5503E+08, 5.5814E+08, 5.6127E+08, 5.6441E+08,
     + 5.6756E+08, 5.7074E+08, 5.7393E+08, 5.7713E+08, 5.8035E+08,
     +  5.8359E+08, 5.8685E+08, 5.9012E+08, 5.9340E+08, 5.9671E+08,
     + 6.0003E+08, 6.0337E+08, 6.0672E+08, 6.1009E+08, 6.1348E+08,
     +  6.1689E+08, 6.2031E+08, 6.2375E+08, 6.2721E+08, 6.3069E+08,
     + 6.3418E+08, 6.3769E+08, 6.4122E+08, 6.4476E+08, 6.4833E+08,
     +  6.5191E+08, 6.5551E+08, 6.5912E+08, 6.6276E+08, 6.6641E+08,
     + 6.7009E+08, 6.7378E+08, 6.7749E+08, 6.8122E+08, 6.8496E+08,
     +  6.8873E+08, 6.9251E+08, 6.9632E+08, 7.0014E+08, 7.0398E+08,
     + 7.0784E+08, 7.1172E+08, 7.1562E+08, 7.1954E+08, 7.2347E+08,
     +  7.2743E+08, 7.3141E+08, 7.3541E+08, 7.3942E+08, 7.4346E+08,
     + 7.4752E+08, 7.5159E+08, 7.5569E+08, 7.5981E+08, 7.6395E+08,
     +  7.6811E+08, 7.7229E+08, 7.7649E+08, 7.8071E+08, 7.8495E+08,
     + 7.8921E+08, 7.9350E+08, 7.9780E+08, 8.0213E+08, 8.0647E+08,
     +  8.1084E+08, 8.1523E+08, 8.1965E+08, 8.2408E+08, 8.2853E+08,
     + 8.3301E+08, 8.3751E+08, 8.4203E+08, 8.4658E+08, 8.5114E+08,
     +  8.5573E+08, 8.6034E+08, 8.6498E+08, 8.6963E+08, 8.7431E+08,
     + 8.7902E+08, 8.8374E+08, 8.8849E+08, 8.9326E+08, 8.9806E+08,
     +  9.0287E+08, 9.0772E+08, 9.1258E+08, 9.1747E+08, 9.2238E+08,
     + 9.2732E+08, 9.3228E+08, 9.3727E+08, 9.4228E+08, 9.4731E+08,
     +  9.5237E+08, 9.5745E+08, 9.6256E+08, 9.6769E+08, 9.7285E+08,
     + 9.7803E+08, 9.8324E+08, 9.8847E+08, 9.9373E+08, 9.9901E+08,
     +  1.0043E+09, 1.0097E+09, 1.0150E+09, 1.0204E+09, 1.0258E+09,
     + 1.0312E+09, 1.0367E+09, 1.0422E+09, 1.0477E+09, 1.0533E+09,
     +  1.0588E+09, 1.0644E+09, 1.0701E+09, 1.0757E+09, 1.0814E+09,
     + 1.0871E+09, 1.0928E+09, 1.0986E+09, 1.1044E+09, 1.1102E+09,
     +  1.1160E+09, 1.1219E+09, 1.1278E+09, 1.1337E+09, 1.1397E+09,
     + 1.1457E+09, 1.1517E+09, 1.1577E+09, 1.1638E+09, 1.1699E+09,
     +  1.1760E+09, 1.1822E+09, 1.1884E+09, 1.1946E+09, 1.2008E+09,
     + 1.2071E+09, 1.2134E+09, 1.2198E+09, 1.2261E+09, 1.2325E+09,
     +  1.2390E+09, 1.2454E+09, 1.2519E+09, 1.2584E+09, 1.2650E+09,
     + 1.2716E+09, 1.2782E+09, 1.2848E+09, 1.2915E+09, 1.2982E+09,
     +  1.3049E+09, 1.3117E+09, 1.3185E+09, 1.3253E+09, 1.3322E+09,
     + 1.3391E+09, 1.3460E+09, 1.3530E+09, 1.3600E+09, 1.3670E+09,
     +  1.3741E+09, 1.3812E+09, 1.3883E+09, 1.3955E+09, 1.4027E+09,
     + 1.4099E+09, 1.4171E+09, 1.4244E+09, 1.4318E+09, 1.4391E+09,
     +  1.4465E+09, 1.4540E+09, 1.4614E+09, 1.4690E+09, 1.4765E+09,
     + 1.4841E+09, 1.4917E+09, 1.4993E+09, 1.5070E+09, 1.5147E+09,
     +  1.5225E+09, 1.5303E+09, 1.5381E+09, 1.5460E+09, 1.5539E+09,
     + 1.5618E+09, 1.5698E+09, 1.5778E+09, 1.5858E+09, 1.5939E+09,
     +  1.6020E+09, 1.6102E+09, 1.6184E+09, 1.6266E+09, 1.6349E+09,
     + 1.6432E+09, 1.6515E+09, 1.6599E+09, 1.6684E+09, 1.6768E+09,
     +  1.6853E+09, 1.6939E+09, 1.7024E+09, 1.7111E+09, 1.7197E+09,
     + 1.7284E+09, 1.7372E+09, 1.7460E+09, 1.7548E+09, 1.7636E+09,
     +  1.7725E+09, 1.7815E+09, 1.7905E+09, 1.7995E+09, 1.8086E+09,
     + 1.8177E+09, 1.8268E+09, 1.8360E+09, 1.8453E+09, 1.8545E+09,
     +  1.8639E+09, 1.8732E+09, 1.8826E+09, 1.8921E+09, 1.9016E+09,
     + 1.9111E+09, 1.9207E+09, 1.9303E+09, 1.9400E+09, 1.9497E+09,
     +  1.9594E+09, 1.9692E+09, 1.9791E+09, 1.9890E+09, 1.9989E+09,
     + 2.0089E+09, 2.0189E+09, 2.0290E+09, 2.0391E+09, 2.0493E+09,
     +  2.0595E+09, 2.0697E+09, 2.0800E+09, 2.0904E+09, 2.1007E+09,
     + 2.1112E+09, 2.1217E+09, 2.1322E+09, 2.1428E+09, 2.1534E+09,
     +  2.1641E+09, 2.1748E+09, 2.1856E+09, 2.1964E+09, 2.2073E+09,
     + 2.2182E+09, 2.2292E+09, 2.2402E+09, 2.2513E+09, 2.2624E+09,
     +  2.2735E+09, 2.2848E+09, 2.2960E+09, 2.3074E+09, 2.3187E+09,
     + 2.3301E+09, 2.3416E+09, 2.3531E+09, 2.3647E+09, 2.3763E+09,
     +  2.3880E+09, 2.3997E+09, 2.4115E+09, 2.4234E+09, 2.4352E+09,
     + 2.4472E+09, 2.4592E+09, 2.4712E+09, 2.4833E+09, 2.4955E+09,
     +  2.5077E+09, 2.5200E+09, 2.5323E+09, 2.5446E+09, 2.5571E+09,
     + 2.5696E+09, 2.5821E+09, 2.5947E+09, 2.6073E+09, 2.6200E+09,
     +  2.6328E+09, 2.6456E+09, 2.6585E+09, 2.6714E+09, 2.6844E+09,
     + 2.6975E+09, 2.7106E+09, 2.7237E+09, 2.7370E+09, 2.7502E+09,
     +  2.7636E+09, 2.7770E+09, 2.7904E+09, 2.8039E+09, 2.8175E+09,
     + 2.8312E+09, 2.8448E+09, 2.8586E+09, 2.8724E+09, 2.8863E+09,
     +  2.9002E+09, 2.9142E+09, 2.9283E+09, 2.9424E+09, 2.9566E+09,
     + 2.9708E+09, 2.9851E+09, 2.9995E+09, 3.0139E+09, 3.0284E+09,
     +  3.0430E+09, 3.0576E+09, 3.0723E+09, 3.0871E+09, 3.1019E+09,
     + 3.1167E+09, 3.1317E+09, 3.1467E+09, 3.1618E+09, 3.1769E+09,
     +  3.1921E+09, 3.2074E+09, 3.2227E+09, 3.2381E+09, 3.2536E+09,
     + 3.2692E+09, 3.2848E+09, 3.3004E+09, 3.3162E+09, 3.3320E+09,
     +  3.3479E+09, 3.3638E+09, 3.3798E+09, 3.3959E+09, 3.4121E+09,
     + 3.4283E+09, 3.4446E+09, 3.4610E+09, 3.4774E+09, 3.4939E+09,
     +  3.5105E+09, 3.5271E+09, 3.5439E+09, 3.5607E+09, 3.5775E+09,
     + 3.5945E+09, 3.6115E+09, 3.6286E+09, 3.6457E+09, 3.6630E+09,
     +  3.6803E+09, 3.6977E+09, 3.7151E+09, 3.7326E+09, 3.7503E+09,
     + 3.7679E+09, 3.7857E+09, 3.8035E+09, 3.8214E+09, 3.8394E+09,
     +  3.8575E+09, 3.8756E+09, 3.8939E+09, 3.9121E+09, 3.9305E+09,
     + 3.9490E+09, 3.9675E+09, 3.9861E+09, 4.0048E+09, 4.0236E+09,
     +  4.0424E+09, 4.0614E+09, 4.0804E+09, 4.0995E+09, 4.1186E+09,
     + 4.1379E+09, 4.1572E+09, 4.1766E+09, 4.1961E+09, 4.2157E+09,
     +  4.2354E+09, 4.2551E+09, 4.2750E+09, 4.2949E+09, 4.3149E+09,
     + 4.3350E+09, 4.3552E+09, 4.3754E+09, 4.3958E+09, 4.4162E+09,
     +  4.4367E+09, 4.4573E+09, 4.4780E+09, 4.4988E+09, 4.5196E+09,
     + 4.5406E+09, 4.5616E+09, 4.5828E+09, 4.6040E+09, 4.6253E+09,
     +  4.6467E+09, 4.6682E+09, 4.6897E+09, 4.7114E+09, 4.7332E+09,
     + 4.7550E+09, 4.7770E+09, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_HO2   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 1),QofT( 1,NT),Q(NT)
      data Tmx(1)/3500./
      data xgj/ 2./
c...      HO2
c...        --       166
      data (QofT( 1,J),J=1,2408)/ 5.5942E-01, 2.8684E+00, 5.0843E+00,
     +  7.5781E+00, 1.0356E+01, 1.3399E+01, 1.6688E+01, 2.0208E+01,
     + 2.3945E+01, 2.7888E+01, 3.2025E+01, 3.6349E+01, 4.0852E+01,
     +  4.5527E+01, 5.0368E+01, 5.5370E+01, 6.0527E+01, 6.5835E+01,
     + 7.1290E+01, 7.6888E+01, 8.8498E+01, 1.0064E+02, 1.1329E+02,
     +  1.2644E+02, 1.4005E+02, 1.5412E+02, 1.6864E+02, 1.8358E+02,
     + 1.9894E+02, 2.1471E+02, 2.3087E+02, 2.4742E+02, 2.6434E+02,
     +  2.8164E+02, 2.9930E+02, 3.1731E+02, 3.3567E+02, 3.5437E+02,
     + 3.7340E+02, 3.9277E+02, 4.1245E+02, 4.3246E+02, 4.5278E+02,
     +  4.7341E+02, 4.9434E+02, 5.1557E+02, 5.3710E+02, 5.5892E+02,
     + 5.8103E+02, 6.0342E+02, 6.2609E+02, 6.4904E+02, 6.7227E+02,
     +  6.9576E+02, 7.1952E+02, 7.4354E+02, 7.6783E+02, 7.9238E+02,
     + 8.1718E+02, 8.4224E+02, 8.6754E+02, 8.9310E+02, 9.1890E+02,
     +  9.4494E+02, 9.7123E+02, 9.9775E+02, 1.0245E+03, 1.0515E+03,
     + 1.0787E+03, 1.1062E+03, 1.1339E+03, 1.1618E+03, 1.1900E+03,
     +  1.2183E+03, 1.2469E+03, 1.2757E+03, 1.3048E+03, 1.3340E+03,
     + 1.3635E+03, 1.3932E+03, 1.4231E+03, 1.4532E+03, 1.4835E+03,
     +  1.5140E+03, 1.5447E+03, 1.5756E+03, 1.6068E+03, 1.6381E+03,
     + 1.6697E+03, 1.7014E+03, 1.7334E+03, 1.7655E+03, 1.7978E+03,
     +  1.8304E+03, 1.8631E+03, 1.8961E+03, 1.9292E+03, 1.9625E+03,
     + 1.9960E+03, 2.0297E+03, 2.0636E+03, 2.0977E+03, 2.1320E+03,
     +  2.1664E+03, 2.2011E+03, 2.2359E+03, 2.2709E+03, 2.3062E+03,
     + 2.3416E+03, 2.3772E+03, 2.4129E+03, 2.4489E+03, 2.4850E+03,
     +  2.5213E+03, 2.5578E+03, 2.5945E+03, 2.6314E+03, 2.6684E+03,
     + 2.7057E+03, 2.7431E+03, 2.7807E+03, 2.8185E+03, 2.8565E+03,
     +  2.8946E+03, 2.9329E+03, 2.9714E+03, 3.0101E+03, 3.0490E+03,
     + 3.0881E+03, 3.1273E+03, 3.1667E+03, 3.2063E+03, 3.2461E+03,
     +  3.2861E+03, 3.3262E+03, 3.3665E+03, 3.4070E+03, 3.4477E+03,
     + 3.4886E+03, 3.5297E+03, 3.5709E+03, 3.6123E+03, 3.6539E+03,
     +  3.6957E+03, 3.7377E+03, 3.7798E+03, 3.8222E+03, 3.8647E+03,
     + 3.9075E+03, 3.9504E+03, 3.9935E+03, 4.0367E+03, 4.0802E+03,
     +  4.1239E+03, 4.1677E+03, 4.2117E+03, 4.2560E+03, 4.3004E+03,
     + 4.3450E+03, 4.3898E+03, 4.4348E+03, 4.4800E+03, 4.5253E+03,
     +  4.5709E+03, 4.6167E+03, 4.6627E+03, 4.7088E+03, 4.7552E+03,
     + 4.8017E+03, 4.8485E+03, 4.8954E+03, 4.9426E+03, 4.9899E+03,
     +  5.0375E+03, 5.0852E+03, 5.1332E+03, 5.1814E+03, 5.2297E+03,
     + 5.2783E+03, 5.3271E+03, 5.3761E+03, 5.4253E+03, 5.4747E+03,
     +  5.5243E+03, 5.5741E+03, 5.6242E+03, 5.6743E+03, 5.7248E+03,
     + 5.7755E+03, 5.8264E+03, 5.8775E+03, 5.9288E+03, 5.9803E+03,
     +  6.0321E+03, 6.0840E+03, 6.1362E+03, 6.1886E+03, 6.2413E+03,
     + 6.2941E+03, 6.3472E+03, 6.4005E+03, 6.4540E+03, 6.5078E+03,
     +  6.5617E+03, 6.6160E+03, 6.6704E+03, 6.7250E+03, 6.7799E+03,
     + 6.8350E+03, 6.8904E+03, 6.9460E+03, 7.0018E+03, 7.0579E+03,
     +  7.1142E+03, 7.1707E+03, 7.2275E+03, 7.2845E+03, 7.3418E+03,
     + 7.3992E+03, 7.4570E+03, 7.5149E+03, 7.5732E+03, 7.6316E+03,
     +  7.6903E+03, 7.7493E+03, 7.8085E+03, 7.8680E+03, 7.9277E+03,
     + 7.9877E+03, 8.0479E+03, 8.1083E+03, 8.1690E+03, 8.2300E+03,
     +  8.2913E+03, 8.3528E+03, 8.4145E+03, 8.4765E+03, 8.5388E+03,
     + 8.6013E+03, 8.6641E+03, 8.7271E+03, 8.7905E+03, 8.8541E+03,
     +  8.9179E+03, 8.9820E+03, 9.0464E+03, 9.1111E+03, 9.1760E+03,
     + 9.2412E+03, 9.3067E+03, 9.3724E+03, 9.4384E+03, 9.5048E+03,
     +  9.5713E+03, 9.6382E+03, 9.7053E+03, 9.7727E+03, 9.8404E+03,
     + 9.9084E+03, 9.9766E+03, 1.0045E+04, 1.0114E+04, 1.0183E+04,
     +  1.0252E+04, 1.0322E+04, 1.0392E+04, 1.0462E+04, 1.0533E+04,
     + 1.0604E+04, 1.0675E+04, 1.0746E+04, 1.0818E+04, 1.0890E+04,
     +  1.0962E+04, 1.1035E+04, 1.1108E+04, 1.1181E+04, 1.1255E+04,
     + 1.1329E+04, 1.1403E+04, 1.1477E+04, 1.1552E+04, 1.1627E+04,
     +  1.1702E+04, 1.1778E+04, 1.1854E+04, 1.1930E+04, 1.2007E+04,
     + 1.2084E+04, 1.2161E+04, 1.2239E+04, 1.2317E+04, 1.2395E+04,
     +  1.2473E+04, 1.2552E+04, 1.2631E+04, 1.2711E+04, 1.2791E+04,
     + 1.2871E+04, 1.2951E+04, 1.3032E+04, 1.3113E+04, 1.3194E+04,
     +  1.3276E+04, 1.3358E+04, 1.3441E+04, 1.3524E+04, 1.3607E+04,
     + 1.3690E+04, 1.3774E+04, 1.3858E+04, 1.3942E+04, 1.4027E+04,
     +  1.4112E+04, 1.4198E+04, 1.4284E+04, 1.4370E+04, 1.4456E+04,
     + 1.4543E+04, 1.4630E+04, 1.4718E+04, 1.4806E+04, 1.4894E+04,
     +  1.4983E+04, 1.5072E+04, 1.5161E+04, 1.5250E+04, 1.5341E+04,
     + 1.5431E+04, 1.5522E+04, 1.5613E+04, 1.5704E+04, 1.5796E+04,
     +  1.5888E+04, 1.5981E+04, 1.6074E+04, 1.6167E+04, 1.6261E+04,
     + 1.6355E+04, 1.6449E+04, 1.6544E+04, 1.6639E+04, 1.6734E+04,
     +  1.6830E+04, 1.6927E+04, 1.7023E+04, 1.7120E+04, 1.7218E+04,
     + 1.7315E+04, 1.7414E+04, 1.7512E+04, 1.7611E+04, 1.7710E+04,
     +  1.7810E+04, 1.7910E+04, 1.8010E+04, 1.8111E+04, 1.8213E+04,
     + 1.8314E+04, 1.8416E+04, 1.8519E+04, 1.8622E+04, 1.8725E+04,
     +  1.8828E+04, 1.8933E+04, 1.9037E+04, 1.9142E+04, 1.9247E+04,
     + 1.9353E+04, 1.9459E+04, 1.9565E+04, 1.9672E+04, 1.9779E+04,
     +  1.9887E+04, 1.9995E+04, 2.0103E+04, 2.0212E+04, 2.0322E+04,
     + 2.0431E+04, 2.0542E+04, 2.0652E+04, 2.0763E+04, 2.0874E+04,
     +  2.0986E+04, 2.1099E+04, 2.1211E+04, 2.1324E+04, 2.1438E+04,
     + 2.1552E+04, 2.1666E+04, 2.1781E+04, 2.1896E+04, 2.2012E+04,
     +  2.2128E+04, 2.2245E+04, 2.2362E+04, 2.2479E+04, 2.2597E+04,
     + 2.2715E+04, 2.2834E+04, 2.2953E+04, 2.3073E+04, 2.3193E+04,
     +  2.3313E+04, 2.3434E+04, 2.3556E+04, 2.3678E+04, 2.3800E+04,
     + 2.3923E+04, 2.4046E+04, 2.4170E+04, 2.4294E+04, 2.4418E+04,
     +  2.4543E+04, 2.4669E+04, 2.4795E+04, 2.4921E+04, 2.5048E+04,
     + 2.5176E+04, 2.5304E+04, 2.5432E+04, 2.5560E+04, 2.5690E+04,
     +  2.5819E+04, 2.5950E+04, 2.6080E+04, 2.6212E+04, 2.6343E+04,
     + 2.6475E+04, 2.6608E+04, 2.6741E+04, 2.6874E+04, 2.7008E+04,
     +  2.7143E+04, 2.7278E+04, 2.7413E+04, 2.7549E+04, 2.7686E+04,
     + 2.7823E+04, 2.7960E+04, 2.8098E+04, 2.8237E+04, 2.8375E+04,
     +  2.8515E+04, 2.8655E+04, 2.8795E+04, 2.8936E+04, 2.9077E+04,
     + 2.9219E+04, 2.9362E+04, 2.9505E+04, 2.9648E+04, 2.9792E+04,
     +  2.9937E+04, 3.0082E+04, 3.0227E+04, 3.0373E+04, 3.0519E+04,
     + 3.0667E+04, 3.0814E+04, 3.0962E+04, 3.1111E+04, 3.1260E+04,
     +  3.1410E+04, 3.1560E+04, 3.1710E+04, 3.1862E+04, 3.2013E+04,
     + 3.2166E+04, 3.2319E+04, 3.2472E+04, 3.2626E+04, 3.2780E+04,
     +  3.2935E+04, 3.3091E+04, 3.3247E+04, 3.3403E+04, 3.3561E+04,
     + 3.3718E+04, 3.3876E+04, 3.4035E+04, 3.4194E+04, 3.4354E+04,
     +  3.4515E+04, 3.4676E+04, 3.4837E+04, 3.4999E+04, 3.5162E+04,
     + 3.5325E+04, 3.5489E+04, 3.5653E+04, 3.5818E+04, 3.5984E+04,
     +  3.6150E+04, 3.6316E+04, 3.6483E+04, 3.6651E+04, 3.6819E+04,
     + 3.6988E+04, 3.7158E+04, 3.7328E+04, 3.7498E+04, 3.7669E+04,
     +  3.7841E+04, 3.8013E+04, 3.8187E+04, 3.8360E+04, 3.8534E+04,
     + 3.8709E+04, 3.8884E+04, 3.9060E+04, 3.9236E+04, 3.9413E+04,
     +  3.9591E+04, 3.9769E+04, 3.9948E+04, 4.0127E+04, 4.0308E+04,
     + 4.0488E+04, 4.0670E+04, 4.0851E+04, 4.1034E+04, 4.1217E+04,
     +  4.1400E+04, 4.1585E+04, 4.1770E+04, 4.1955E+04, 4.2141E+04,
     + 4.2328E+04, 4.2515E+04, 4.2703E+04, 4.2892E+04, 4.3081E+04,
     +  4.3271E+04, 4.3462E+04, 4.3653E+04, 4.3844E+04, 4.4037E+04,
     + 4.4230E+04, 4.4423E+04, 4.4618E+04, 4.4813E+04, 4.5008E+04,
     +  4.5204E+04, 4.5401E+04, 4.5599E+04, 4.5797E+04, 4.5995E+04,
     + 4.6195E+04, 4.6395E+04, 4.6596E+04, 4.6797E+04, 4.6999E+04,
     +  4.7202E+04, 4.7405E+04, 4.7609E+04, 4.7814E+04, 4.8019E+04,
     + 4.8225E+04, 4.8431E+04, 4.8639E+04, 4.8847E+04, 4.9055E+04,
     +  4.9265E+04, 4.9475E+04, 4.9685E+04, 4.9897E+04, 5.0109E+04,
     + 5.0321E+04, 5.0535E+04, 5.0749E+04, 5.0964E+04, 5.1179E+04,
     +  5.1395E+04, 5.1612E+04, 5.1830E+04, 5.2048E+04, 5.2267E+04,
     + 5.2486E+04, 5.2706E+04, 5.2927E+04, 5.3149E+04, 5.3371E+04,
     +  5.3595E+04, 5.3818E+04, 5.4043E+04, 5.4268E+04, 5.4494E+04,
     + 5.4720E+04, 5.4948E+04, 5.5176E+04, 5.5405E+04, 5.5634E+04,
     +  5.5864E+04, 5.6095E+04, 5.6327E+04, 5.6559E+04, 5.6792E+04,
     + 5.7026E+04, 5.7261E+04, 5.7496E+04, 5.7732E+04, 5.7969E+04,
     +  5.8206E+04, 5.8444E+04, 5.8684E+04, 5.8923E+04, 5.9164E+04,
     + 5.9405E+04, 5.9647E+04, 5.9889E+04, 6.0133E+04, 6.0377E+04,
     +  6.0622E+04, 6.0868E+04, 6.1114E+04, 6.1361E+04, 6.1609E+04,
     + 6.1858E+04, 6.2108E+04, 6.2358E+04, 6.2609E+04, 6.2861E+04,
     +  6.3113E+04, 6.3367E+04, 6.3621E+04, 6.3876E+04, 6.4131E+04,
     + 6.4388E+04, 6.4645E+04, 6.4903E+04, 6.5162E+04, 6.5421E+04,
     +  6.5682E+04, 6.5943E+04, 6.6205E+04, 6.6468E+04, 6.6731E+04,
     + 6.6996E+04, 6.7261E+04, 6.7527E+04, 6.7793E+04, 6.8061E+04,
     +  6.8329E+04, 6.8598E+04, 6.8868E+04, 6.9139E+04, 6.9411E+04,
     + 6.9683E+04, 6.9957E+04, 7.0231E+04, 7.0505E+04, 7.0781E+04,
     +  7.1057E+04, 7.1335E+04, 7.1613E+04, 7.1892E+04, 7.2172E+04,
     + 7.2452E+04, 7.2734E+04, 7.3016E+04, 7.3299E+04, 7.3583E+04,
     +  7.3868E+04, 7.4154E+04, 7.4440E+04, 7.4728E+04, 7.5016E+04,
     + 7.5305E+04, 7.5595E+04, 7.5886E+04, 7.6177E+04, 7.6470E+04,
     +  7.6763E+04, 7.7057E+04, 7.7352E+04, 7.7648E+04, 7.7945E+04,
     + 7.8243E+04, 7.8541E+04, 7.8841E+04, 7.9141E+04, 7.9442E+04,
     +  7.9744E+04, 8.0047E+04, 8.0351E+04, 8.0655E+04, 8.0961E+04,
     + 8.1268E+04, 8.1575E+04, 8.1883E+04, 8.2192E+04, 8.2502E+04,
     +  8.2813E+04, 8.3125E+04, 8.3438E+04, 8.3751E+04, 8.4066E+04,
     + 8.4381E+04, 8.4698E+04, 8.5015E+04, 8.5333E+04, 8.5652E+04,
     +  8.5972E+04, 8.6293E+04, 8.6615E+04, 8.6938E+04, 8.7261E+04,
     + 8.7586E+04, 8.7911E+04, 8.8238E+04, 8.8565E+04, 8.8894E+04,
     +  8.9223E+04, 8.9553E+04, 8.9884E+04, 9.0216E+04, 9.0549E+04,
     + 9.0883E+04, 9.1218E+04, 9.1554E+04, 9.1891E+04, 9.2228E+04,
     +  9.2567E+04, 9.2907E+04, 9.3248E+04, 9.3589E+04, 9.3932E+04,
     + 9.4275E+04, 9.4620E+04, 9.4965E+04, 9.5311E+04, 9.5659E+04,
     +  9.6007E+04, 9.6357E+04, 9.6707E+04, 9.7058E+04, 9.7411E+04,
     + 9.7764E+04, 9.8118E+04, 9.8474E+04, 9.8830E+04, 9.9187E+04,
     +  9.9545E+04, 9.9904E+04, 1.0026E+05, 1.0063E+05, 1.0099E+05,
     + 1.0135E+05, 1.0172E+05, 1.0208E+05, 1.0245E+05, 1.0281E+05,
     +  1.0318E+05, 1.0355E+05, 1.0392E+05, 1.0429E+05, 1.0467E+05,
     + 1.0504E+05, 1.0541E+05, 1.0579E+05, 1.0617E+05, 1.0654E+05,
     +  1.0692E+05, 1.0730E+05, 1.0768E+05, 1.0806E+05, 1.0845E+05,
     + 1.0883E+05, 1.0921E+05, 1.0960E+05, 1.0999E+05, 1.1038E+05,
     +  1.1076E+05, 1.1115E+05, 1.1155E+05, 1.1194E+05, 1.1233E+05,
     + 1.1273E+05, 1.1312E+05, 1.1352E+05, 1.1391E+05, 1.1431E+05,
     +  1.1471E+05, 1.1511E+05, 1.1552E+05, 1.1592E+05, 1.1632E+05,
     + 1.1673E+05, 1.1713E+05, 1.1754E+05, 1.1795E+05, 1.1836E+05,
     +  1.1877E+05, 1.1918E+05, 1.1959E+05, 1.2001E+05, 1.2042E+05,
     + 1.2084E+05, 1.2126E+05, 1.2167E+05, 1.2209E+05, 1.2251E+05,
     +  1.2294E+05, 1.2336E+05, 1.2378E+05, 1.2421E+05, 1.2463E+05,
     + 1.2506E+05, 1.2549E+05, 1.2592E+05, 1.2635E+05, 1.2678E+05,
     +  1.2721E+05, 1.2765E+05, 1.2808E+05, 1.2852E+05, 1.2896E+05,
     + 1.2940E+05, 1.2984E+05, 1.3028E+05, 1.3072E+05, 1.3116E+05,
     +  1.3161E+05, 1.3205E+05, 1.3250E+05, 1.3295E+05, 1.3340E+05,
     + 1.3385E+05, 1.3430E+05, 1.3475E+05, 1.3521E+05, 1.3566E+05,
     +  1.3612E+05, 1.3658E+05, 1.3703E+05, 1.3749E+05, 1.3795E+05,
     + 1.3842E+05, 1.3888E+05, 1.3935E+05, 1.3981E+05, 1.4028E+05,
     +  1.4075E+05, 1.4122E+05, 1.4169E+05, 1.4216E+05, 1.4263E+05,
     + 1.4311E+05, 1.4358E+05, 1.4406E+05, 1.4454E+05, 1.4502E+05,
     +  1.4550E+05, 1.4598E+05, 1.4646E+05, 1.4694E+05, 1.4743E+05,
     + 1.4792E+05, 1.4840E+05, 1.4889E+05, 1.4938E+05, 1.4988E+05,
     +  1.5037E+05, 1.5086E+05, 1.5136E+05, 1.5186E+05, 1.5235E+05,
     + 1.5285E+05, 1.5335E+05, 1.5385E+05, 1.5436E+05, 1.5486E+05,
     +  1.5537E+05, 1.5587E+05, 1.5638E+05, 1.5689E+05, 1.5740E+05,
     + 1.5791E+05, 1.5843E+05, 1.5894E+05, 1.5946E+05, 1.5998E+05,
     +  1.6049E+05, 1.6101E+05, 1.6153E+05, 1.6206E+05, 1.6258E+05,
     + 1.6311E+05, 1.6363E+05, 1.6416E+05, 1.6469E+05, 1.6522E+05,
     +  1.6575E+05, 1.6628E+05, 1.6682E+05, 1.6735E+05, 1.6789E+05,
     + 1.6843E+05, 1.6897E+05, 1.6951E+05, 1.7005E+05, 1.7060E+05,
     +  1.7114E+05, 1.7169E+05, 1.7223E+05, 1.7278E+05, 1.7333E+05,
     + 1.7389E+05, 1.7444E+05, 1.7499E+05, 1.7555E+05, 1.7611E+05,
     +  1.7667E+05, 1.7723E+05, 1.7779E+05, 1.7835E+05, 1.7891E+05,
     + 1.7948E+05, 1.8005E+05, 1.8062E+05, 1.8118E+05, 1.8176E+05,
     +  1.8233E+05, 1.8290E+05, 1.8348E+05, 1.8405E+05, 1.8463E+05,
     + 1.8521E+05, 1.8579E+05, 1.8638E+05, 1.8696E+05, 1.8754E+05,
     +  1.8813E+05, 1.8872E+05, 1.8931E+05, 1.8990E+05, 1.9049E+05,
     + 1.9109E+05, 1.9168E+05, 1.9228E+05, 1.9288E+05, 1.9348E+05,
     +  1.9408E+05, 1.9468E+05, 1.9528E+05, 1.9589E+05, 1.9649E+05,
     + 1.9710E+05, 1.9771E+05, 1.9832E+05, 1.9894E+05, 1.9955E+05,
     +  2.0017E+05, 2.0078E+05, 2.0140E+05, 2.0202E+05, 2.0264E+05,
     + 2.0327E+05, 2.0389E+05, 2.0452E+05, 2.0514E+05, 2.0577E+05,
     +  2.0640E+05, 2.0704E+05, 2.0767E+05, 2.0830E+05, 2.0894E+05,
     + 2.0958E+05, 2.1022E+05, 2.1086E+05, 2.1150E+05, 2.1215E+05,
     +  2.1279E+05, 2.1344E+05, 2.1409E+05, 2.1474E+05, 2.1539E+05,
     + 2.1604E+05, 2.1670E+05, 2.1735E+05, 2.1801E+05, 2.1867E+05,
     +  2.1933E+05, 2.2000E+05, 2.2066E+05, 2.2133E+05, 2.2199E+05,
     + 2.2266E+05, 2.2333E+05, 2.2400E+05, 2.2468E+05, 2.2535E+05,
     +  2.2603E+05, 2.2671E+05, 2.2739E+05, 2.2807E+05, 2.2875E+05,
     + 2.2944E+05, 2.3012E+05, 2.3081E+05, 2.3150E+05, 2.3219E+05,
     +  2.3288E+05, 2.3358E+05, 2.3427E+05, 2.3497E+05, 2.3567E+05,
     + 2.3637E+05, 2.3707E+05, 2.3778E+05, 2.3848E+05, 2.3919E+05,
     +  2.3990E+05, 2.4061E+05, 2.4132E+05, 2.4204E+05, 2.4275E+05,
     + 2.4347E+05, 2.4419E+05, 2.4491E+05, 2.4563E+05, 2.4635E+05,
     +  2.4708E+05, 2.4781E+05, 2.4853E+05, 2.4926E+05, 2.5000E+05,
     + 2.5073E+05, 2.5146E+05, 2.5220E+05, 2.5294E+05, 2.5368E+05,
     +  2.5442E+05, 2.5517E+05, 2.5591E+05, 2.5666E+05, 2.5741E+05,
     + 2.5816E+05, 2.5891E+05, 2.5966E+05, 2.6042E+05, 2.6118E+05,
     +  2.6194E+05, 2.6270E+05, 2.6346E+05, 2.6422E+05, 2.6499E+05,
     + 2.6576E+05, 2.6653E+05, 2.6730E+05, 2.6807E+05, 2.6885E+05,
     +  2.6962E+05, 2.7040E+05, 2.7118E+05, 2.7196E+05, 2.7275E+05,
     + 2.7353E+05, 2.7432E+05, 2.7511E+05, 2.7590E+05, 2.7669E+05,
     +  2.7748E+05, 2.7828E+05, 2.7908E+05, 2.7988E+05, 2.8068E+05,
     + 2.8148E+05, 2.8228E+05, 2.8309E+05, 2.8390E+05, 2.8471E+05,
     +  2.8552E+05, 2.8633E+05, 2.8715E+05, 2.8797E+05, 2.8879E+05,
     + 2.8961E+05, 2.9043E+05, 2.9125E+05, 2.9208E+05, 2.9291E+05,
     +  2.9374E+05, 2.9457E+05, 2.9540E+05, 2.9624E+05, 2.9708E+05,
     + 2.9792E+05, 2.9876E+05, 2.9960E+05, 3.0045E+05, 3.0129E+05,
     +  3.0214E+05, 3.0299E+05, 3.0384E+05, 3.0470E+05, 3.0555E+05,
     + 3.0641E+05, 3.0727E+05, 3.0813E+05, 3.0900E+05, 3.0986E+05,
     +  3.1073E+05, 3.1160E+05, 3.1247E+05, 3.1334E+05, 3.1422E+05,
     + 3.1509E+05, 3.1597E+05, 3.1685E+05, 3.1773E+05, 3.1862E+05,
     +  3.1951E+05, 3.2039E+05, 3.2128E+05, 3.2218E+05, 3.2307E+05,
     + 3.2396E+05, 3.2486E+05, 3.2576E+05, 3.2666E+05, 3.2757E+05,
     +  3.2847E+05, 3.2938E+05, 3.3029E+05, 3.3120E+05, 3.3212E+05,
     + 3.3303E+05, 3.3395E+05, 3.3487E+05, 3.3579E+05, 3.3671E+05,
     +  3.3764E+05, 3.3857E+05, 3.3949E+05, 3.4043E+05, 3.4136E+05,
     + 3.4229E+05, 3.4323E+05, 3.4417E+05, 3.4511E+05, 3.4605E+05,
     +  3.4700E+05, 3.4795E+05, 3.4890E+05, 3.4985E+05, 3.5080E+05,
     + 3.5176E+05, 3.5271E+05, 3.5367E+05, 3.5463E+05, 3.5560E+05,
     +  3.5656E+05, 3.5753E+05, 3.5850E+05, 3.5947E+05, 3.6045E+05,
     + 3.6142E+05, 3.6240E+05, 3.6338E+05, 3.6436E+05, 3.6535E+05,
     +  3.6633E+05, 3.6732E+05, 3.6831E+05, 3.6930E+05, 3.7030E+05,
     + 3.7129E+05, 3.7229E+05, 3.7329E+05, 3.7430E+05, 3.7530E+05,
     +  3.7631E+05, 3.7732E+05, 3.7833E+05, 3.7934E+05, 3.8036E+05,
     + 3.8137E+05, 3.8239E+05, 3.8342E+05, 3.8444E+05, 3.8546E+05,
     +  3.8649E+05, 3.8752E+05, 3.8856E+05, 3.8959E+05, 3.9063E+05,
     + 3.9167E+05, 3.9271E+05, 3.9375E+05, 3.9480E+05, 3.9584E+05,
     +  3.9689E+05, 3.9794E+05, 3.9900E+05, 4.0005E+05, 4.0111E+05,
     + 4.0217E+05, 4.0324E+05, 4.0430E+05, 4.0537E+05, 4.0644E+05,
     +  4.0751E+05, 4.0858E+05, 4.0966E+05, 4.1074E+05, 4.1182E+05,
     + 4.1290E+05, 4.1399E+05, 4.1507E+05, 4.1616E+05, 4.1725E+05,
     +  4.1835E+05, 4.1944E+05, 4.2054E+05, 4.2164E+05, 4.2274E+05,
     + 4.2385E+05, 4.2496E+05, 4.2607E+05, 4.2718E+05, 4.2829E+05,
     +  4.2941E+05, 4.3053E+05, 4.3165E+05, 4.3277E+05, 4.3390E+05,
     + 4.3502E+05, 4.3615E+05, 4.3729E+05, 4.3842E+05, 4.3956E+05,
     +  4.4070E+05, 4.4184E+05, 4.4298E+05, 4.4413E+05, 4.4528E+05,
     + 4.4643E+05, 4.4758E+05, 4.4873E+05, 4.4989E+05, 4.5105E+05,
     +  4.5221E+05, 4.5338E+05, 4.5455E+05, 4.5571E+05, 4.5689E+05,
     + 4.5806E+05, 4.5924E+05, 4.6042E+05, 4.6160E+05, 4.6278E+05,
     +  4.6397E+05, 4.6515E+05, 4.6634E+05, 4.6754E+05, 4.6873E+05,
     + 4.6993E+05, 4.7113E+05, 4.7233E+05, 4.7354E+05, 4.7475E+05,
     +  4.7595E+05, 4.7717E+05, 4.7838E+05, 4.7960E+05, 4.8082E+05,
     + 4.8204E+05, 4.8326E+05, 4.8449E+05, 4.8572E+05, 4.8695E+05,
     +  4.8818E+05, 4.8942E+05, 4.9066E+05, 4.9190E+05, 4.9314E+05,
     + 4.9439E+05, 4.9564E+05, 4.9689E+05, 4.9814E+05, 4.9940E+05,
     +  5.0066E+05, 5.0192E+05, 5.0318E+05, 5.0445E+05, 5.0571E+05,
     + 5.0699E+05, 5.0826E+05, 5.0953E+05, 5.1081E+05, 5.1209E+05,
     +  5.1338E+05, 5.1466E+05, 5.1595E+05, 5.1724E+05, 5.1853E+05,
     + 5.1983E+05, 5.2113E+05, 5.2243E+05, 5.2373E+05, 5.2504E+05,
     +  5.2635E+05, 5.2766E+05, 5.2897E+05, 5.3029E+05, 5.3161E+05,
     + 5.3293E+05, 5.3425E+05, 5.3558E+05, 5.3691E+05, 5.3824E+05,
     +  5.3957E+05, 5.4091E+05, 5.4225E+05, 5.4359E+05, 5.4494E+05,
     + 5.4628E+05, 5.4763E+05, 5.4899E+05, 5.5034E+05, 5.5170E+05,
     +  5.5306E+05, 5.5442E+05, 5.5579E+05, 5.5715E+05, 5.5852E+05,
     + 5.5990E+05, 5.6127E+05, 5.6265E+05, 5.6403E+05, 5.6542E+05,
     +  5.6680E+05, 5.6819E+05, 5.6958E+05, 5.7098E+05, 5.7237E+05,
     + 5.7377E+05, 5.7518E+05, 5.7658E+05, 5.7799E+05, 5.7940E+05,
     +  5.8081E+05, 5.8223E+05, 5.8365E+05, 5.8507E+05, 5.8649E+05,
     + 5.8792E+05, 5.8935E+05, 5.9078E+05, 5.9222E+05, 5.9365E+05,
     +  5.9509E+05, 5.9654E+05, 5.9798E+05, 5.9943E+05, 6.0088E+05,
     + 6.0234E+05, 6.0379E+05, 6.0525E+05, 6.0671E+05, 6.0818E+05,
     +  6.0964E+05, 6.1112E+05, 6.1259E+05, 6.1406E+05, 6.1554E+05,
     + 6.1702E+05, 6.1851E+05, 6.2000E+05, 6.2149E+05, 6.2298E+05,
     +  6.2447E+05, 6.2597E+05, 6.2747E+05, 6.2898E+05, 6.3048E+05,
     + 6.3199E+05, 6.3350E+05, 6.3502E+05, 6.3654E+05, 6.3806E+05,
     +  6.3958E+05, 6.4111E+05, 6.4264E+05, 6.4417E+05, 6.4571E+05,
     + 6.4724E+05, 6.4878E+05, 6.5033E+05, 6.5187E+05, 6.5342E+05,
     +  6.5497E+05, 6.5653E+05, 6.5809E+05, 6.5965E+05, 6.6121E+05,
     + 6.6278E+05, 6.6435E+05, 6.6592E+05, 6.6750E+05, 6.6907E+05,
     +  6.7066E+05, 6.7224E+05, 6.7383E+05, 6.7542E+05, 6.7701E+05,
     + 6.7860E+05, 6.8020E+05, 6.8180E+05, 6.8341E+05, 6.8502E+05,
     +  6.8663E+05, 6.8824E+05, 6.8986E+05, 6.9148E+05, 6.9310E+05,
     + 6.9472E+05, 6.9635E+05, 6.9798E+05, 6.9962E+05, 7.0125E+05,
     +  7.0289E+05, 7.0454E+05, 7.0618E+05, 7.0783E+05, 7.0948E+05,
     + 7.1114E+05, 7.1280E+05, 7.1446E+05, 7.1612E+05, 7.1779E+05,
     +  7.1946E+05, 7.2113E+05, 7.2281E+05, 7.2449E+05, 7.2617E+05,
     + 7.2786E+05, 7.2955E+05, 7.3124E+05, 7.3293E+05, 7.3463E+05,
     +  7.3633E+05, 7.3803E+05, 7.3974E+05, 7.4145E+05, 7.4316E+05,
     + 7.4488E+05, 7.4660E+05, 7.4832E+05, 7.5005E+05, 7.5178E+05,
     +  7.5351E+05, 7.5524E+05, 7.5698E+05, 7.5872E+05, 7.6046E+05,
     + 7.6221E+05, 7.6396E+05, 7.6572E+05, 7.6747E+05, 7.6923E+05,
     +  7.7100E+05, 7.7276E+05, 7.7453E+05, 7.7630E+05, 7.7808E+05,
     + 7.7986E+05, 7.8164E+05, 7.8342E+05, 7.8521E+05, 7.8700E+05,
     +  7.8880E+05, 7.9060E+05, 7.9240E+05, 7.9420E+05, 7.9601E+05,
     + 7.9782E+05, 7.9963E+05, 8.0145E+05, 8.0327E+05, 8.0510E+05,
     +  8.0692E+05, 8.0875E+05, 8.1059E+05, 8.1242E+05, 8.1426E+05,
     + 8.1610E+05, 8.1795E+05, 8.1980E+05, 8.2165E+05, 8.2351E+05,
     +  8.2537E+05, 8.2723E+05, 8.2910E+05, 8.3097E+05, 8.3284E+05,
     + 8.3471E+05, 8.3659E+05, 8.3848E+05, 8.4036E+05, 8.4225E+05,
     +  8.4414E+05, 8.4604E+05, 8.4794E+05, 8.4984E+05, 8.5174E+05,
     + 8.5365E+05, 8.5557E+05, 8.5748E+05, 8.5940E+05, 8.6132E+05,
     +  8.6325E+05, 8.6518E+05, 8.6711E+05, 8.6905E+05, 8.7098E+05,
     + 8.7293E+05, 8.7487E+05, 8.7682E+05, 8.7877E+05, 8.8073E+05,
     +  8.8269E+05, 8.8465E+05, 8.8662E+05, 8.8859E+05, 8.9056E+05,
     + 8.9254E+05, 8.9452E+05, 8.9650E+05, 8.9849E+05, 9.0048E+05,
     +  9.0247E+05, 9.0447E+05, 9.0647E+05, 9.0847E+05, 9.1048E+05,
     + 9.1249E+05, 9.1451E+05, 9.1652E+05, 9.1854E+05, 9.2057E+05,
     +  9.2260E+05, 9.2463E+05, 9.2666E+05, 9.2870E+05, 9.3075E+05,
     + 9.3279E+05, 9.3484E+05, 9.3689E+05, 9.3895E+05, 9.4101E+05,
     +  9.4307E+05, 9.4514E+05, 9.4721E+05, 9.4928E+05, 9.5136E+05,
     + 9.5344E+05, 9.5553E+05, 9.5762E+05, 9.5971E+05, 9.6180E+05,
     +  9.6390E+05, 9.6600E+05, 9.6811E+05, 9.7022E+05, 9.7233E+05,
     + 9.7445E+05, 9.7657E+05, 9.7869E+05, 9.8082E+05, 9.8295E+05,
     +  9.8509E+05, 9.8723E+05, 9.8937E+05, 9.9151E+05, 9.9366E+05,
     + 9.9582E+05, 9.9797E+05, 1.0001E+06, 1.0023E+06, 1.0045E+06,
     +  1.0066E+06, 1.0088E+06, 1.0110E+06, 1.0132E+06, 1.0154E+06,
     + 1.0175E+06, 1.0197E+06, 1.0219E+06, 1.0241E+06, 1.0263E+06,
     +  1.0285E+06, 1.0308E+06, 1.0330E+06, 1.0352E+06, 1.0374E+06,
     + 1.0396E+06, 1.0419E+06, 1.0441E+06, 1.0463E+06, 1.0486E+06,
     +  1.0508E+06, 1.0531E+06, 1.0553E+06, 1.0576E+06, 1.0599E+06,
     + 1.0621E+06, 1.0644E+06, 1.0667E+06, 1.0689E+06, 1.0712E+06,
     +  1.0735E+06, 1.0758E+06, 1.0781E+06, 1.0804E+06, 1.0827E+06,
     + 1.0850E+06, 1.0873E+06, 1.0896E+06, 1.0919E+06, 1.0942E+06,
     +  1.0965E+06, 1.0989E+06, 1.1012E+06, 1.1035E+06, 1.1059E+06,
     + 1.1082E+06, 1.1105E+06, 1.1129E+06, 1.1152E+06, 1.1176E+06,
     +  1.1200E+06, 1.1223E+06, 1.1247E+06, 1.1271E+06, 1.1294E+06,
     + 1.1318E+06, 1.1342E+06, 1.1366E+06, 1.1390E+06, 1.1414E+06,
     +  1.1438E+06, 1.1462E+06, 1.1486E+06, 1.1510E+06, 1.1534E+06,
     + 1.1558E+06, 1.1582E+06, 1.1607E+06, 1.1631E+06, 1.1655E+06,
     +  1.1680E+06, 1.1704E+06, 1.1729E+06, 1.1753E+06, 1.1778E+06,
     + 1.1802E+06, 1.1827E+06, 1.1852E+06, 1.1876E+06, 1.1901E+06,
     +  1.1926E+06, 1.1951E+06, 1.1975E+06, 1.2000E+06, 1.2025E+06,
     + 1.2050E+06, 1.2075E+06, 1.2100E+06, 1.2125E+06, 1.2151E+06,
     +  1.2176E+06, 1.2201E+06, 1.2226E+06, 1.2252E+06, 1.2277E+06,
     + 1.2302E+06, 1.2328E+06, 1.2353E+06, 1.2379E+06, 1.2404E+06,
     +  1.2430E+06, 1.2456E+06, 1.2481E+06, 1.2507E+06, 1.2533E+06,
     + 1.2559E+06, 1.2584E+06, 1.2610E+06, 1.2636E+06, 1.2662E+06,
     +  1.2688E+06, 1.2714E+06, 1.2740E+06, 1.2766E+06, 1.2793E+06,
     + 1.2819E+06, 1.2845E+06, 1.2871E+06, 1.2898E+06, 1.2924E+06,
     +  1.2951E+06, 1.2977E+06, 1.3004E+06, 1.3030E+06, 1.3057E+06,
     + 1.3083E+06, 1.3110E+06, 1.3137E+06, 1.3164E+06, 1.3190E+06,
     +  1.3217E+06, 1.3244E+06, 1.3271E+06, 1.3298E+06, 1.3325E+06,
     + 1.3352E+06, 1.3379E+06, 1.3406E+06, 1.3434E+06, 1.3461E+06,
     +  1.3488E+06, 1.3516E+06, 1.3543E+06, 1.3570E+06, 1.3598E+06,
     + 1.3625E+06, 1.3653E+06, 1.3680E+06, 1.3708E+06, 1.3736E+06,
     +  1.3763E+06, 1.3791E+06, 1.3819E+06, 1.3847E+06, 1.3875E+06,
     + 1.3903E+06, 1.3931E+06, 1.3959E+06, 1.3987E+06, 1.4015E+06,
     +  1.4043E+06, 1.4071E+06, 1.4099E+06, 1.4128E+06, 1.4156E+06,
     + 1.4184E+06, 1.4213E+06, 1.4241E+06, 1.4270E+06, 1.4298E+06,
     +  1.4327E+06, 1.4356E+06, 1.4384E+06, 1.4413E+06, 1.4442E+06,
     + 1.4471E+06, 1.4500E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_O     (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 1),QofT( 1,NT),Q(NT)
      data Tmx(1)/10./
      data xgj/ 0./
c...        O
c..no data for these species yet      
   99 return
      end
c
c     *****************
      Subroutine QT_ClONO2(                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 2),QofT( 2,NT),Q(NT)
      data (Tmx(I),I=1,2)/3500.,3500./
      data xgj/ 12.,12./
c...   ClONO2
c...        --      5646
      data (QofT( 1,J),J=1,2408)/ 2.4088E+02, 6.6891E+02, 1.2214E+03,
     +  1.8749E+03, 2.6155E+03, 3.4341E+03, 4.3238E+03, 5.2793E+03,
     + 6.2965E+03, 7.3717E+03, 8.5020E+03, 9.6849E+03, 1.0918E+04,
     +  1.2200E+04, 1.3528E+04, 1.4901E+04, 1.6318E+04, 1.7778E+04,
     + 1.9279E+04, 2.0820E+04, 2.4021E+04, 2.7377E+04, 3.0882E+04,
     +  3.4536E+04, 3.8338E+04, 4.2287E+04, 4.6386E+04, 5.0635E+04,
     + 5.5036E+04, 5.9594E+04, 6.4309E+04, 6.9186E+04, 7.4226E+04,
     +  7.9436E+04, 8.4817E+04, 9.0374E+04, 9.6110E+04, 1.0203E+05,
     + 1.0814E+05, 1.1444E+05, 1.2093E+05, 1.2763E+05, 1.3453E+05,
     +  1.4164E+05, 1.4896E+05, 1.5650E+05, 1.6427E+05, 1.7227E+05,
     + 1.8049E+05, 1.8896E+05, 1.9768E+05, 2.0664E+05, 2.1586E+05,
     +  2.2533E+05, 2.3508E+05, 2.4510E+05, 2.5539E+05, 2.6598E+05,
     + 2.7685E+05, 2.8802E+05, 2.9950E+05, 3.1128E+05, 3.2339E+05,
     +  3.3581E+05, 3.4858E+05, 3.6168E+05, 3.7512E+05, 3.8893E+05,
     + 4.0309E+05, 4.1763E+05, 4.3255E+05, 4.4785E+05, 4.6355E+05,
     +  4.7966E+05, 4.9617E+05, 5.1311E+05, 5.3049E+05, 5.4831E+05,
     + 5.6658E+05, 5.8531E+05, 6.0451E+05, 6.2420E+05, 6.4438E+05,
     +  6.6506E+05, 6.8626E+05, 7.0799E+05, 7.3025E+05, 7.5307E+05,
     + 7.7645E+05, 8.0041E+05, 8.2495E+05, 8.5009E+05, 8.7585E+05,
     +  9.0224E+05, 9.2927E+05, 9.5695E+05, 9.8530E+05, 1.0143E+06,
     + 1.0441E+06, 1.0745E+06, 1.1057E+06, 1.1376E+06, 1.1703E+06,
     +  1.2038E+06, 1.2381E+06, 1.2732E+06, 1.3091E+06, 1.3459E+06,
     + 1.3835E+06, 1.4220E+06, 1.4614E+06, 1.5018E+06, 1.5431E+06,
     +  1.5853E+06, 1.6286E+06, 1.6728E+06, 1.7181E+06, 1.7644E+06,
     + 1.8118E+06, 1.8603E+06, 1.9100E+06, 1.9607E+06, 2.0126E+06,
     +  2.0657E+06, 2.1201E+06, 2.1756E+06, 2.2325E+06, 2.2906E+06,
     + 2.3500E+06, 2.4108E+06, 2.4730E+06, 2.5366E+06, 2.6016E+06,
     +  2.6680E+06, 2.7360E+06, 2.8055E+06, 2.8765E+06, 2.9491E+06,
     + 3.0233E+06, 3.0992E+06, 3.1768E+06, 3.2561E+06, 3.3371E+06,
     +  3.4200E+06, 3.5046E+06, 3.5911E+06, 3.6795E+06, 3.7699E+06,
     + 3.8622E+06, 3.9566E+06, 4.0529E+06, 4.1514E+06, 4.2521E+06,
     +  4.3549E+06, 4.4599E+06, 4.5672E+06, 4.6768E+06, 4.7888E+06,
     + 4.9032E+06, 5.0200E+06, 5.1393E+06, 5.2612E+06, 5.3856E+06,
     +  5.5127E+06, 5.6426E+06, 5.7751E+06, 5.9105E+06, 6.0487E+06,
     + 6.1898E+06, 6.3339E+06, 6.4811E+06, 6.6313E+06, 6.7846E+06,
     +  6.9412E+06, 7.1010E+06, 7.2641E+06, 7.4307E+06, 7.6006E+06,
     + 7.7741E+06, 7.9512E+06, 8.1319E+06, 8.3164E+06, 8.5046E+06,
     +  8.6967E+06, 8.8927E+06, 9.0928E+06, 9.2969E+06, 9.5051E+06,
     + 9.7176E+06, 9.9344E+06, 1.0156E+07, 1.0381E+07, 1.0612E+07,
     +  1.0846E+07, 1.1086E+07, 1.1330E+07, 1.1580E+07, 1.1834E+07,
     + 1.2093E+07, 1.2358E+07, 1.2627E+07, 1.2903E+07, 1.3183E+07,
     +  1.3469E+07, 1.3761E+07, 1.4058E+07, 1.4362E+07, 1.4671E+07,
     + 1.4987E+07, 1.5308E+07, 1.5636E+07, 1.5970E+07, 1.6311E+07,
     +  1.6658E+07, 1.7012E+07, 1.7373E+07, 1.7741E+07, 1.8116E+07,
     + 1.8499E+07, 1.8888E+07, 1.9285E+07, 1.9690E+07, 2.0102E+07,
     +  2.0523E+07, 2.0951E+07, 2.1388E+07, 2.1832E+07, 2.2285E+07,
     + 2.2747E+07, 2.3218E+07, 2.3697E+07, 2.4186E+07, 2.4683E+07,
     +  2.5190E+07, 2.5707E+07, 2.6233E+07, 2.6769E+07, 2.7315E+07,
     + 2.7871E+07, 2.8437E+07, 2.9014E+07, 2.9602E+07, 3.0200E+07,
     +  3.0810E+07, 3.1430E+07, 3.2063E+07, 3.2706E+07, 3.3362E+07,
     + 3.4029E+07, 3.4709E+07, 3.5401E+07, 3.6106E+07, 3.6824E+07,
     +  3.7554E+07, 3.8298E+07, 3.9055E+07, 3.9826E+07, 4.0611E+07,
     + 4.1410E+07, 4.2224E+07, 4.3051E+07, 4.3894E+07, 4.4752E+07,
     +  4.5625E+07, 4.6514E+07, 4.7418E+07, 4.8339E+07, 4.9276E+07,
     + 5.0229E+07, 5.1199E+07, 5.2187E+07, 5.3192E+07, 5.4214E+07,
     +  5.5255E+07, 5.6313E+07, 5.7390E+07, 5.8486E+07, 5.9602E+07,
     + 6.0736E+07, 6.1890E+07, 6.3065E+07, 6.4259E+07, 6.5475E+07,
     +  6.6711E+07, 6.7969E+07, 6.9248E+07, 7.0549E+07, 7.1873E+07,
     + 7.3219E+07, 7.4589E+07, 7.5982E+07, 7.7398E+07, 7.8839E+07,
     +  8.0304E+07, 8.1794E+07, 8.3309E+07, 8.4850E+07, 8.6417E+07,
     + 8.8010E+07, 8.9630E+07, 9.1278E+07, 9.2953E+07, 9.4656E+07,
     +  9.6388E+07, 9.8148E+07, 9.9938E+07, 1.0176E+08, 1.0361E+08,
     + 1.0549E+08, 1.0740E+08, 1.0934E+08, 1.1132E+08, 1.1333E+08,
     +  1.1537E+08, 1.1744E+08, 1.1955E+08, 1.2170E+08, 1.2388E+08,
     + 1.2609E+08, 1.2834E+08, 1.3063E+08, 1.3295E+08, 1.3531E+08,
     +  1.3771E+08, 1.4015E+08, 1.4263E+08, 1.4515E+08, 1.4771E+08,
     + 1.5031E+08, 1.5295E+08, 1.5564E+08, 1.5837E+08, 1.6114E+08,
     +  1.6395E+08, 1.6681E+08, 1.6972E+08, 1.7267E+08, 1.7567E+08,
     + 1.7872E+08, 1.8181E+08, 1.8496E+08, 1.8815E+08, 1.9139E+08,
     +  1.9469E+08, 1.9804E+08, 2.0144E+08, 2.0489E+08, 2.0840E+08,
     + 2.1196E+08, 2.1557E+08, 2.1925E+08, 2.2298E+08, 2.2677E+08,
     +  2.3061E+08, 2.3452E+08, 2.3849E+08, 2.4251E+08, 2.4660E+08,
     + 2.5076E+08, 2.5497E+08, 2.5926E+08, 2.6360E+08, 2.6802E+08,
     +  2.7250E+08, 2.7705E+08, 2.8167E+08, 2.8636E+08, 2.9112E+08,
     + 2.9595E+08, 3.0086E+08, 3.0584E+08, 3.1089E+08, 3.1602E+08,
     +  3.2123E+08, 3.2652E+08, 3.3189E+08, 3.3733E+08, 3.4286E+08,
     + 3.4847E+08, 3.5417E+08, 3.5995E+08, 3.6581E+08, 3.7177E+08,
     +  3.7781E+08, 3.8394E+08, 3.9016E+08, 3.9647E+08, 4.0288E+08,
     + 4.0938E+08, 4.1598E+08, 4.2267E+08, 4.2946E+08, 4.3636E+08,
     +  4.4335E+08, 4.5044E+08, 4.5764E+08, 4.6494E+08, 4.7235E+08,
     + 4.7987E+08, 4.8750E+08, 4.9523E+08, 5.0308E+08, 5.1104E+08,
     +  5.1912E+08, 5.2731E+08, 5.3562E+08, 5.4405E+08, 5.5261E+08,
     + 5.6128E+08, 5.7008E+08, 5.7900E+08, 5.8805E+08, 5.9723E+08,
     +  6.0654E+08, 6.1599E+08, 6.2556E+08, 6.3528E+08, 6.4513E+08,
     + 6.5511E+08, 6.6525E+08, 6.7552E+08, 6.8594E+08, 6.9650E+08,
     +  7.0721E+08, 7.1807E+08, 7.2909E+08, 7.4026E+08, 7.5158E+08,
     + 7.6306E+08, 7.7471E+08, 7.8651E+08, 7.9848E+08, 8.1061E+08,
     +  8.2291E+08, 8.3539E+08, 8.4803E+08, 8.6085E+08, 8.7384E+08,
     + 8.8702E+08, 9.0037E+08, 9.1391E+08, 9.2763E+08, 9.4155E+08,
     +  9.5565E+08, 9.6994E+08, 9.8443E+08, 9.9911E+08, 1.0140E+09,
     + 1.0291E+09, 1.0444E+09, 1.0599E+09, 1.0756E+09, 1.0915E+09,
     +  1.1076E+09, 1.1240E+09, 1.1406E+09, 1.1574E+09, 1.1744E+09,
     + 1.1916E+09, 1.2091E+09, 1.2268E+09, 1.2448E+09, 1.2630E+09,
     +  1.2814E+09, 1.3001E+09, 1.3190E+09, 1.3381E+09, 1.3576E+09,
     + 1.3773E+09, 1.3972E+09, 1.4174E+09, 1.4379E+09, 1.4586E+09,
     +  1.4796E+09, 1.5009E+09, 1.5225E+09, 1.5443E+09, 1.5665E+09,
     + 1.5889E+09, 1.6116E+09, 1.6346E+09, 1.6580E+09, 1.6816E+09,
     +  1.7055E+09, 1.7297E+09, 1.7543E+09, 1.7791E+09, 1.8043E+09,
     + 1.8298E+09, 1.8556E+09, 1.8818E+09, 1.9083E+09, 1.9352E+09,
     +  1.9623E+09, 1.9899E+09, 2.0178E+09, 2.0460E+09, 2.0746E+09,
     + 2.1036E+09, 2.1329E+09, 2.1626E+09, 2.1927E+09, 2.2232E+09,
     +  2.2540E+09, 2.2852E+09, 2.3169E+09, 2.3489E+09, 2.3813E+09,
     + 2.4142E+09, 2.4474E+09, 2.4811E+09, 2.5152E+09, 2.5497E+09,
     +  2.5847E+09, 2.6200E+09, 2.6559E+09, 2.6921E+09, 2.7289E+09,
     + 2.7660E+09, 2.8037E+09, 2.8418E+09, 2.8804E+09, 2.9194E+09,
     +  2.9589E+09, 2.9990E+09, 3.0395E+09, 3.0805E+09, 3.1220E+09,
     + 3.1640E+09, 3.2066E+09, 3.2496E+09, 3.2932E+09, 3.3373E+09,
     +  3.3820E+09, 3.4272E+09, 3.4729E+09, 3.5192E+09, 3.5661E+09,
     + 3.6135E+09, 3.6615E+09, 3.7101E+09, 3.7593E+09, 3.8090E+09,
     +  3.8594E+09, 3.9104E+09, 3.9619E+09, 4.0141E+09, 4.0670E+09,
     + 4.1204E+09, 4.1745E+09, 4.2293E+09, 4.2847E+09, 4.3407E+09,
     +  4.3974E+09, 4.4548E+09, 4.5129E+09, 4.5717E+09, 4.6312E+09,
     + 4.6913E+09, 4.7522E+09, 4.8138E+09, 4.8761E+09, 4.9392E+09,
     +  5.0030E+09, 5.0675E+09, 5.1328E+09, 5.1989E+09, 5.2657E+09,
     + 5.3333E+09, 5.4018E+09, 5.4710E+09, 5.5410E+09, 5.6118E+09,
     +  5.6835E+09, 5.7560E+09, 5.8293E+09, 5.9035E+09, 5.9785E+09,
     + 6.0544E+09, 6.1312E+09, 6.2088E+09, 6.2874E+09, 6.3669E+09,
     +  6.4472E+09, 6.5285E+09, 6.6108E+09, 6.6939E+09, 6.7781E+09,
     + 6.8631E+09, 6.9492E+09, 7.0362E+09, 7.1243E+09, 7.2133E+09,
     +  7.3033E+09, 7.3944E+09, 7.4865E+09, 7.5796E+09, 7.6738E+09,
     + 7.7690E+09, 7.8654E+09, 7.9628E+09, 8.0613E+09, 8.1609E+09,
     +  8.2616E+09, 8.3635E+09, 8.4665E+09, 8.5707E+09, 8.6760E+09,
     + 8.7825E+09, 8.8902E+09, 8.9991E+09, 9.1092E+09, 9.2205E+09,
     +  9.3331E+09, 9.4469E+09, 9.5619E+09, 9.6783E+09, 9.7959E+09,
     + 9.9148E+09, 1.0035E+10, 1.0157E+10, 1.0280E+10, 1.0404E+10,
     +  1.0529E+10, 1.0656E+10, 1.0785E+10, 1.0915E+10, 1.1046E+10,
     + 1.1179E+10, 1.1313E+10, 1.1448E+10, 1.1585E+10, 1.1724E+10,
     +  1.1864E+10, 1.2005E+10, 1.2148E+10, 1.2293E+10, 1.2439E+10,
     + 1.2587E+10, 1.2736E+10, 1.2887E+10, 1.3040E+10, 1.3194E+10,
     +  1.3350E+10, 1.3508E+10, 1.3667E+10, 1.3828E+10, 1.3991E+10,
     + 1.4155E+10, 1.4321E+10, 1.4489E+10, 1.4659E+10, 1.4831E+10,
     +  1.5004E+10, 1.5179E+10, 1.5356E+10, 1.5535E+10, 1.5716E+10,
     + 1.5899E+10, 1.6084E+10, 1.6270E+10, 1.6459E+10, 1.6650E+10,
     +  1.6842E+10, 1.7037E+10, 1.7234E+10, 1.7432E+10, 1.7633E+10,
     + 1.7836E+10, 1.8041E+10, 1.8248E+10, 1.8458E+10, 1.8669E+10,
     +  1.8883E+10, 1.9099E+10, 1.9317E+10, 1.9538E+10, 1.9761E+10,
     + 1.9986E+10, 2.0213E+10, 2.0443E+10, 2.0675E+10, 2.0909E+10,
     +  2.1146E+10, 2.1386E+10, 2.1628E+10, 2.1872E+10, 2.2119E+10,
     + 2.2368E+10, 2.2620E+10, 2.2874E+10, 2.3131E+10, 2.3391E+10,
     +  2.3654E+10, 2.3919E+10, 2.4186E+10, 2.4457E+10, 2.4730E+10,
     + 2.5006E+10, 2.5285E+10, 2.5566E+10, 2.5851E+10, 2.6138E+10,
     +  2.6428E+10, 2.6721E+10, 2.7017E+10, 2.7316E+10, 2.7618E+10,
     + 2.7924E+10, 2.8232E+10, 2.8543E+10, 2.8857E+10, 2.9175E+10,
     +  2.9495E+10, 2.9819E+10, 3.0146E+10, 3.0477E+10, 3.0810E+10,
     + 3.1147E+10, 3.1488E+10, 3.1831E+10, 3.2179E+10, 3.2529E+10,
     +  3.2883E+10, 3.3241E+10, 3.3602E+10, 3.3966E+10, 3.4334E+10,
     + 3.4706E+10, 3.5082E+10, 3.5461E+10, 3.5844E+10, 3.6230E+10,
     +  3.6621E+10, 3.7015E+10, 3.7413E+10, 3.7815E+10, 3.8221E+10,
     + 3.8631E+10, 3.9044E+10, 3.9462E+10, 3.9884E+10, 4.0310E+10,
     +  4.0740E+10, 4.1175E+10, 4.1613E+10, 4.2056E+10, 4.2503E+10,
     + 4.2954E+10, 4.3410E+10, 4.3870E+10, 4.4335E+10, 4.4803E+10,
     +  4.5277E+10, 4.5755E+10, 4.6238E+10, 4.6725E+10, 4.7217E+10,
     + 4.7713E+10, 4.8215E+10, 4.8721E+10, 4.9232E+10, 4.9747E+10,
     +  5.0268E+10, 5.0794E+10, 5.1324E+10, 5.1860E+10, 5.2401E+10,
     + 5.2947E+10, 5.3498E+10, 5.4054E+10, 5.4615E+10, 5.5182E+10,
     +  5.5754E+10, 5.6332E+10, 5.6914E+10, 5.7503E+10, 5.8097E+10,
     + 5.8696E+10, 5.9301E+10, 5.9912E+10, 6.0529E+10, 6.1151E+10,
     +  6.1779E+10, 6.2413E+10, 6.3052E+10, 6.3698E+10, 6.4350E+10,
     + 6.5008E+10, 6.5672E+10, 6.6342E+10, 6.7018E+10, 6.7700E+10,
     +  6.8389E+10, 6.9084E+10, 6.9786E+10, 7.0494E+10, 7.1209E+10,
     + 7.1930E+10, 7.2657E+10, 7.3392E+10, 7.4133E+10, 7.4881E+10,
     +  7.5636E+10, 7.6398E+10, 7.7166E+10, 7.7942E+10, 7.8725E+10,
     + 7.9515E+10, 8.0312E+10, 8.1116E+10, 8.1928E+10, 8.2747E+10,
     +  8.3574E+10, 8.4408E+10, 8.5249E+10, 8.6098E+10, 8.6955E+10,
     + 8.7820E+10, 8.8692E+10, 8.9572E+10, 9.0460E+10, 9.1357E+10,
     +  9.2261E+10, 9.3173E+10, 9.4094E+10, 9.5022E+10, 9.5959E+10,
     + 9.6905E+10, 9.7859E+10, 9.8821E+10, 9.9792E+10, 1.0077E+11,
     +  1.0176E+11, 1.0276E+11, 1.0376E+11, 1.0478E+11, 1.0580E+11,
     + 1.0684E+11, 1.0788E+11, 1.0893E+11, 1.0999E+11, 1.1106E+11,
     +  1.1214E+11, 1.1323E+11, 1.1433E+11, 1.1544E+11, 1.1655E+11,
     + 1.1768E+11, 1.1882E+11, 1.1997E+11, 1.2113E+11, 1.2229E+11,
     +  1.2347E+11, 1.2466E+11, 1.2586E+11, 1.2707E+11, 1.2829E+11,
     + 1.2952E+11, 1.3076E+11, 1.3201E+11, 1.3328E+11, 1.3455E+11,
     +  1.3584E+11, 1.3713E+11, 1.3844E+11, 1.3976E+11, 1.4109E+11,
     + 1.4243E+11, 1.4378E+11, 1.4515E+11, 1.4652E+11, 1.4791E+11,
     +  1.4931E+11, 1.5072E+11, 1.5215E+11, 1.5359E+11, 1.5503E+11,
     + 1.5650E+11, 1.5797E+11, 1.5946E+11, 1.6096E+11, 1.6247E+11,
     +  1.6399E+11, 1.6553E+11, 1.6708E+11, 1.6864E+11, 1.7022E+11,
     + 1.7181E+11, 1.7342E+11, 1.7503E+11, 1.7667E+11, 1.7831E+11,
     +  1.7997E+11, 1.8164E+11, 1.8333E+11, 1.8503E+11, 1.8675E+11,
     + 1.8848E+11, 1.9022E+11, 1.9198E+11, 1.9376E+11, 1.9555E+11,
     +  1.9735E+11, 1.9917E+11, 2.0100E+11, 2.0285E+11, 2.0472E+11,
     + 2.0660E+11, 2.0850E+11, 2.1041E+11, 2.1234E+11, 2.1428E+11,
     +  2.1624E+11, 2.1822E+11, 2.2021E+11, 2.2222E+11, 2.2425E+11,
     + 2.2629E+11, 2.2835E+11, 2.3043E+11, 2.3252E+11, 2.3463E+11,
     +  2.3676E+11, 2.3890E+11, 2.4107E+11, 2.4325E+11, 2.4545E+11,
     + 2.4767E+11, 2.4990E+11, 2.5216E+11, 2.5443E+11, 2.5672E+11,
     +  2.5903E+11, 2.6136E+11, 2.6370E+11, 2.6607E+11, 2.6846E+11,
     + 2.7086E+11, 2.7329E+11, 2.7573E+11, 2.7820E+11, 2.8068E+11,
     +  2.8318E+11, 2.8571E+11, 2.8825E+11, 2.9082E+11, 2.9341E+11,
     + 2.9601E+11, 2.9864E+11, 3.0129E+11, 3.0396E+11, 3.0665E+11,
     +  3.0937E+11, 3.1210E+11, 3.1486E+11, 3.1764E+11, 3.2044E+11,
     + 3.2327E+11, 3.2611E+11, 3.2898E+11, 3.3187E+11, 3.3479E+11,
     +  3.3773E+11, 3.4069E+11, 3.4368E+11, 3.4669E+11, 3.4972E+11,
     + 3.5278E+11, 3.5586E+11, 3.5896E+11, 3.6209E+11, 3.6525E+11,
     +  3.6843E+11, 3.7164E+11, 3.7487E+11, 3.7812E+11, 3.8141E+11,
     + 3.8471E+11, 3.8805E+11, 3.9141E+11, 3.9479E+11, 3.9821E+11,
     +  4.0165E+11, 4.0511E+11, 4.0861E+11, 4.1213E+11, 4.1568E+11,
     + 4.1925E+11, 4.2286E+11, 4.2649E+11, 4.3015E+11, 4.3384E+11,
     +  4.3756E+11, 4.4130E+11, 4.4508E+11, 4.4888E+11, 4.5272E+11,
     + 4.5658E+11, 4.6048E+11, 4.6440E+11, 4.6835E+11, 4.7234E+11,
     +  4.7636E+11, 4.8040E+11, 4.8448E+11, 4.8859E+11, 4.9273E+11,
     + 4.9690E+11, 5.0111E+11, 5.0535E+11, 5.0962E+11, 5.1392E+11,
     +  5.1825E+11, 5.2262E+11, 5.2702E+11, 5.3146E+11, 5.3593E+11,
     + 5.4043E+11, 5.4497E+11, 5.4954E+11, 5.5415E+11, 5.5879E+11,
     +  5.6347E+11, 5.6818E+11, 5.7293E+11, 5.7771E+11, 5.8254E+11,
     + 5.8739E+11, 5.9229E+11, 5.9722E+11, 6.0219E+11, 6.0719E+11,
     +  6.1223E+11, 6.1732E+11, 6.2244E+11, 6.2759E+11, 6.3279E+11,
     + 6.3803E+11, 6.4330E+11, 6.4862E+11, 6.5397E+11, 6.5937E+11,
     +  6.6480E+11, 6.7028E+11, 6.7579E+11, 6.8135E+11, 6.8695E+11,
     + 6.9259E+11, 6.9827E+11, 7.0400E+11, 7.0977E+11, 7.1558E+11,
     +  7.2143E+11, 7.2733E+11, 7.3327E+11, 7.3925E+11, 7.4528E+11,
     + 7.5136E+11, 7.5748E+11, 7.6364E+11, 7.6985E+11, 7.7610E+11,
     +  7.8241E+11, 7.8875E+11, 7.9515E+11, 8.0159E+11, 8.0808E+11,
     + 8.1461E+11, 8.2120E+11, 8.2783E+11, 8.3451E+11, 8.4124E+11,
     +  8.4802E+11, 8.5485E+11, 8.6173E+11, 8.6865E+11, 8.7563E+11,
     + 8.8266E+11, 8.8974E+11, 8.9688E+11, 9.0406E+11, 9.1130E+11,
     +  9.1859E+11, 9.2593E+11, 9.3332E+11, 9.4077E+11, 9.4828E+11,
     + 9.5583E+11, 9.6344E+11, 9.7111E+11, 9.7883E+11, 9.8661E+11,
     +  9.9444E+11, 1.0023E+12, 1.0103E+12, 1.0183E+12, 1.0263E+12,
     + 1.0345E+12, 1.0426E+12, 1.0509E+12, 1.0592E+12, 1.0675E+12,
     +  1.0759E+12, 1.0844E+12, 1.0929E+12, 1.1015E+12, 1.1102E+12,
     + 1.1189E+12, 1.1277E+12, 1.1365E+12, 1.1455E+12, 1.1544E+12,
     +  1.1635E+12, 1.1725E+12, 1.1817E+12, 1.1909E+12, 1.2002E+12,
     + 1.2096E+12, 1.2190E+12, 1.2285E+12, 1.2381E+12, 1.2477E+12,
     +  1.2574E+12, 1.2671E+12, 1.2770E+12, 1.2869E+12, 1.2968E+12,
     + 1.3069E+12, 1.3170E+12, 1.3272E+12, 1.3374E+12, 1.3477E+12,
     +  1.3581E+12, 1.3686E+12, 1.3791E+12, 1.3897E+12, 1.4004E+12,
     + 1.4112E+12, 1.4220E+12, 1.4329E+12, 1.4439E+12, 1.4550E+12,
     +  1.4661E+12, 1.4774E+12, 1.4887E+12, 1.5000E+12, 1.5115E+12,
     + 1.5230E+12, 1.5346E+12, 1.5463E+12, 1.5581E+12, 1.5700E+12,
     +  1.5819E+12, 1.5939E+12, 1.6060E+12, 1.6182E+12, 1.6305E+12,
     + 1.6428E+12, 1.6553E+12, 1.6678E+12, 1.6804E+12, 1.6931E+12,
     +  1.7059E+12, 1.7188E+12, 1.7317E+12, 1.7448E+12, 1.7579E+12,
     + 1.7711E+12, 1.7845E+12, 1.7979E+12, 1.8114E+12, 1.8250E+12,
     +  1.8386E+12, 1.8524E+12, 1.8663E+12, 1.8802E+12, 1.8943E+12,
     + 1.9085E+12, 1.9227E+12, 1.9371E+12, 1.9515E+12, 1.9660E+12,
     +  1.9807E+12, 1.9954E+12, 2.0103E+12, 2.0252E+12, 2.0402E+12,
     + 2.0554E+12, 2.0706E+12, 2.0860E+12, 2.1014E+12, 2.1170E+12,
     +  2.1326E+12, 2.1484E+12, 2.1642E+12, 2.1802E+12, 2.1963E+12,
     + 2.2125E+12, 2.2288E+12, 2.2452E+12, 2.2617E+12, 2.2783E+12,
     +  2.2950E+12, 2.3119E+12, 2.3288E+12, 2.3459E+12, 2.3631E+12,
     + 2.3804E+12, 2.3978E+12, 2.4153E+12, 2.4330E+12, 2.4507E+12,
     +  2.4686E+12, 2.4866E+12, 2.5047E+12, 2.5230E+12, 2.5413E+12,
     + 2.5598E+12, 2.5784E+12, 2.5971E+12, 2.6160E+12, 2.6349E+12,
     +  2.6540E+12, 2.6732E+12, 2.6926E+12, 2.7120E+12, 2.7316E+12,
     + 2.7514E+12, 2.7712E+12, 2.7912E+12, 2.8113E+12, 2.8316E+12,
     +  2.8520E+12, 2.8725E+12, 2.8931E+12, 2.9139E+12, 2.9348E+12,
     + 2.9559E+12, 2.9771E+12, 2.9984E+12, 3.0198E+12, 3.0415E+12,
     +  3.0632E+12, 3.0851E+12, 3.1071E+12, 3.1293E+12, 3.1516E+12,
     + 3.1741E+12, 3.1967E+12, 3.2194E+12, 3.2423E+12, 3.2653E+12,
     +  3.2885E+12, 3.3119E+12, 3.3354E+12, 3.3590E+12, 3.3828E+12,
     + 3.4068E+12, 3.4309E+12, 3.4551E+12, 3.4795E+12, 3.5041E+12,
     +  3.5288E+12, 3.5537E+12, 3.5787E+12, 3.6039E+12, 3.6293E+12,
     + 3.6548E+12, 3.6805E+12, 3.7064E+12, 3.7324E+12, 3.7586E+12,
     +  3.7849E+12, 3.8114E+12, 3.8381E+12, 3.8649E+12, 3.8920E+12,
     + 3.9192E+12, 3.9465E+12, 3.9741E+12, 4.0018E+12, 4.0297E+12,
     +  4.0577E+12, 4.0860E+12, 4.1144E+12, 4.1430E+12, 4.1718E+12,
     + 4.2007E+12, 4.2299E+12, 4.2592E+12, 4.2887E+12, 4.3184E+12,
     +  4.3483E+12, 4.3784E+12, 4.4086E+12, 4.4391E+12, 4.4697E+12,
     + 4.5005E+12, 4.5316E+12, 4.5628E+12, 4.5942E+12, 4.6258E+12,
     +  4.6576E+12, 4.6896E+12, 4.7218E+12, 4.7542E+12, 4.7868E+12,
     + 4.8196E+12, 4.8526E+12, 4.8858E+12, 4.9193E+12, 4.9529E+12,
     +  4.9867E+12, 5.0208E+12, 5.0550E+12, 5.0895E+12, 5.1242E+12,
     + 5.1591E+12, 5.1942E+12, 5.2295E+12, 5.2651E+12, 5.3008E+12,
     +  5.3368E+12, 5.3730E+12, 5.4094E+12, 5.4461E+12, 5.4830E+12,
     + 5.5201E+12, 5.5574E+12, 5.5950E+12, 5.6328E+12, 5.6708E+12,
     +  5.7090E+12, 5.7475E+12, 5.7862E+12, 5.8252E+12, 5.8644E+12,
     + 5.9038E+12, 5.9435E+12, 5.9834E+12, 6.0236E+12, 6.0640E+12,
     +  6.1047E+12, 6.1456E+12, 6.1867E+12, 6.2281E+12, 6.2698E+12,
     + 6.3117E+12, 6.3538E+12, 6.3962E+12, 6.4389E+12, 6.4818E+12,
     +  6.5250E+12, 6.5684E+12, 6.6122E+12, 6.6561E+12, 6.7004E+12,
     + 6.7449E+12, 6.7896E+12, 6.8347E+12, 6.8800E+12, 6.9256E+12,
     +  6.9714E+12, 7.0175E+12, 7.0640E+12, 7.1106E+12, 7.1576E+12,
     + 7.2049E+12, 7.2524E+12, 7.3002E+12, 7.3483E+12, 7.3967E+12,
     +  7.4454E+12, 7.4943E+12, 7.5436E+12, 7.5931E+12, 7.6430E+12,
     + 7.6931E+12, 7.7436E+12, 7.7943E+12, 7.8453E+12, 7.8967E+12,
     +  7.9483E+12, 8.0003E+12, 8.0526E+12, 8.1051E+12, 8.1580E+12,
     + 8.2112E+12, 8.2647E+12, 8.3186E+12, 8.3727E+12, 8.4272E+12,
     +  8.4820E+12, 8.5371E+12, 8.5925E+12, 8.6483E+12, 8.7043E+12,
     + 8.7608E+12, 8.8175E+12, 8.8746E+12, 8.9320E+12, 8.9898E+12,
     +  9.0478E+12, 9.1063E+12, 9.1651E+12, 9.2242E+12, 9.2836E+12,
     + 9.3434E+12, 9.4036E+12, 9.4641E+12, 9.5250E+12, 9.5862E+12,
     +  9.6478E+12, 9.7097E+12, 9.7720E+12, 9.8347E+12, 9.8977E+12,
     + 9.9611E+12, 1.0025E+13, 1.0089E+13, 1.0153E+13, 1.0218E+13,
     +  1.0284E+13, 1.0349E+13, 1.0415E+13, 1.0482E+13, 1.0548E+13,
     + 1.0615E+13, 1.0683E+13, 1.0751E+13, 1.0819E+13, 1.0888E+13,
     +  1.0957E+13, 1.1027E+13, 1.1097E+13, 1.1167E+13, 1.1238E+13,
     + 1.1309E+13, 1.1380E+13, 1.1452E+13, 1.1524E+13, 1.1597E+13,
     +  1.1670E+13, 1.1744E+13, 1.1818E+13, 1.1892E+13, 1.1967E+13,
     + 1.2043E+13, 1.2118E+13, 1.2195E+13, 1.2271E+13, 1.2348E+13,
     +  1.2426E+13, 1.2503E+13, 1.2582E+13, 1.2661E+13, 1.2740E+13,
     + 1.2820E+13, 1.2900E+13, 1.2980E+13, 1.3061E+13, 1.3143E+13,
     +  1.3225E+13, 1.3307E+13, 1.3390E+13, 1.3474E+13, 1.3557E+13,
     + 1.3642E+13, 1.3726E+13, 1.3812E+13, 1.3897E+13, 1.3984E+13,
     +  1.4070E+13, 1.4158E+13, 1.4245E+13, 1.4333E+13, 1.4422E+13,
     + 1.4511E+13, 1.4601E+13, 1.4691E+13, 1.4782E+13, 1.4873E+13,
     +  1.4965E+13, 1.5057E+13, 1.5150E+13, 1.5243E+13, 1.5337E+13,
     + 1.5431E+13, 1.5526E+13, 1.5621E+13, 1.5717E+13, 1.5813E+13,
     +  1.5910E+13, 1.6008E+13, 1.6106E+13, 1.6204E+13, 1.6303E+13,
     + 1.6403E+13, 1.6503E+13, 1.6604E+13, 1.6705E+13, 1.6807E+13,
     +  1.6910E+13, 1.7013E+13, 1.7116E+13, 1.7220E+13, 1.7325E+13,
     + 1.7430E+13, 1.7536E+13, 1.7643E+13, 1.7750E+13, 1.7857E+13,
     +  1.7966E+13, 1.8075E+13, 1.8184E+13, 1.8294E+13, 1.8405E+13,
     + 1.8516E+13, 1.8628E+13, 1.8740E+13, 1.8853E+13, 1.8967E+13,
     +  1.9081E+13, 1.9196E+13, 1.9312E+13, 1.9428E+13, 1.9545E+13,
     + 1.9662E+13, 1.9780E+13, 1.9899E+13, 2.0018E+13, 2.0138E+13,
     +  2.0259E+13, 2.0380E+13, 2.0502E+13, 2.0625E+13, 2.0748E+13,
     + 2.0872E+13, 2.0997E+13, 2.1122E+13, 2.1248E+13, 2.1375E+13,
     +  2.1503E+13, 2.1631E+13, 2.1759E+13, 2.1889E+13, 2.2019E+13,
     + 2.2150E+13, 2.2281E+13, 2.2414E+13, 2.2547E+13, 2.2680E+13,
     +  2.2815E+13, 2.2950E+13, 2.3086E+13, 2.3223E+13, 2.3360E+13,
     + 2.3498E+13, 2.3637E+13, 2.3776E+13, 2.3917E+13, 2.4058E+13,
     +  2.4199E+13, 2.4342E+13, 2.4485E+13, 2.4629E+13, 2.4774E+13,
     + 2.4920E+13, 2.5066E+13, 2.5213E+13, 2.5361E+13, 2.5510E+13,
     +  2.5660E+13, 2.5810E+13, 2.5961E+13, 2.6113E+13, 2.6266E+13,
     + 2.6419E+13, 2.6574E+13, 2.6729E+13, 2.6885E+13, 2.7042E+13,
     +  2.7199E+13, 2.7358E+13, 2.7517E+13, 2.7677E+13, 2.7838E+13,
     + 2.8000E+13, 2.8163E+13, 2.8327E+13, 2.8491E+13, 2.8656E+13,
     +  2.8823E+13, 2.8990E+13, 2.9158E+13, 2.9326E+13, 2.9496E+13,
     + 2.9667E+13, 2.9838E+13, 3.0010E+13, 3.0184E+13, 3.0358E+13,
     +  3.0533E+13, 3.0709E+13, 3.0886E+13, 3.1064E+13, 3.1243E+13,
     + 3.1422E+13, 3.1603E+13, 3.1784E+13, 3.1967E+13, 3.2150E+13,
     +  3.2335E+13, 3.2520E+13, 3.2707E+13, 3.2894E+13, 3.3082E+13,
     + 3.3272E+13, 3.3462E+13, 3.3653E+13, 3.3845E+13, 3.4039E+13,
     +  3.4233E+13, 3.4428E+13, 3.4624E+13, 3.4822E+13, 3.5020E+13,
     + 3.5219E+13, 3.5419E+13, 3.5621E+13, 3.5823E+13, 3.6027E+13,
     +  3.6231E+13, 3.6437E+13, 3.6643E+13, 3.6851E+13, 3.7060E+13,
     + 3.7269E+13, 3.7480E+13, 3.7692E+13, 3.7905E+13, 3.8119E+13,
     +  3.8335E+13, 3.8551E+13, 3.8768E+13, 3.8987E+13, 3.9207E+13,
     + 3.9427E+13, 3.9649E+13, 3.9872E+13, 4.0097E+13, 4.0322E+13,
     +  4.0548E+13, 4.0776E+13, 4.1005E+13, 4.1235E+13, 4.1466E+13,
     + 4.1698E+13, 4.1932E+13, 4.2166E+13, 4.2402E+13, 4.2639E+13,
     +  4.2877E+13, 4.3117E+13, 4.3357E+13, 4.3599E+13, 4.3842E+13,
     + 4.4087E+13, 4.4332E+13, 4.4579E+13, 4.4827E+13, 4.5076E+13,
     +  4.5327E+13, 4.5579E+13, 4.5832E+13, 4.6086E+13, 4.6342E+13,
     + 4.6599E+13, 4.6857E+13, 4.7116E+13, 4.7377E+13, 4.7639E+13,
     +  4.7903E+13, 4.8168E+13, 4.8434E+13, 4.8701E+13, 4.8970E+13,
     + 4.9240E+13, 4.9511E+13, 4.9784E+13, 5.0058E+13, 5.0334E+13,
     +  5.0611E+13, 5.0889E+13, 5.1169E+13, 5.1450E+13, 5.1732E+13,
     + 5.2016E+13, 5.2301E+13, 5.2588E+13, 5.2876E+13, 5.3166E+13,
     +  5.3457E+13, 5.3749E+13, 5.4043E+13, 5.4338E+13, 5.4635E+13,
     + 5.4933E+13, 5.5233E+13, 5.5534E+13, 5.5837E+13, 5.6141E+13,
     +  5.6447E+13, 5.6754E+13, 5.7063E+13, 5.7373E+13, 5.7685E+13,
     + 5.7998E+13, 5.8313E+13, 5.8630E+13, 5.8948E+13, 5.9267E+13,
     +  5.9588E+13, 5.9911E+13, 6.0235E+13, 6.0561E+13, 6.0889E+13,
     + 6.1218E+13, 6.1548E+13, 6.1881E+13, 6.2215E+13, 6.2550E+13,
     +  6.2887E+13, 6.3226E+13, 6.3567E+13, 6.3909E+13, 6.4253E+13,
     + 6.4598E+13, 6.4945E+13, 6.5294E+13, 6.5645E+13, 6.5997E+13,
     +  6.6351E+13, 6.6707E+13, 6.7064E+13, 6.7423E+13, 6.7784E+13,
     + 6.8147E+13, 6.8511E+13, 6.8877E+13, 6.9245E+13, 6.9615E+13,
     +  6.9987E+13, 7.0360E+13, 7.0735E+13, 7.1112E+13, 7.1491E+13,
     + 7.1872E+13, 7.2254E+13, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --      7646
      data (QofT( 2,J),J=1,2408)/ 2.4681E+02, 6.8566E+02, 1.2522E+03,
     +  1.9222E+03, 2.6816E+03, 3.5210E+03, 4.4333E+03, 5.4132E+03,
     + 6.4562E+03, 7.5587E+03, 8.7178E+03, 9.9307E+03, 1.1195E+04,
     +  1.2509E+04, 1.3871E+04, 1.5280E+04, 1.6733E+04, 1.8229E+04,
     + 1.9768E+04, 2.1349E+04, 2.4632E+04, 2.8073E+04, 3.1667E+04,
     +  3.5414E+04, 3.9313E+04, 4.3363E+04, 4.7566E+04, 5.1923E+04,
     + 5.6436E+04, 6.1109E+04, 6.5945E+04, 7.0946E+04, 7.6115E+04,
     +  8.1456E+04, 8.6975E+04, 9.2673E+04, 9.8555E+04, 1.0463E+05,
     + 1.1089E+05, 1.1735E+05, 1.2401E+05, 1.3087E+05, 1.3795E+05,
     +  1.4524E+05, 1.5275E+05, 1.6049E+05, 1.6845E+05, 1.7665E+05,
     + 1.8509E+05, 1.9377E+05, 2.0271E+05, 2.1190E+05, 2.2135E+05,
     +  2.3107E+05, 2.4106E+05, 2.5133E+05, 2.6189E+05, 2.7274E+05,
     + 2.8389E+05, 2.9535E+05, 3.0712E+05, 3.1920E+05, 3.3162E+05,
     +  3.4436E+05, 3.5745E+05, 3.7088E+05, 3.8467E+05, 3.9882E+05,
     + 4.1335E+05, 4.2826E+05, 4.4355E+05, 4.5925E+05, 4.7535E+05,
     +  4.9186E+05, 5.0880E+05, 5.2617E+05, 5.4399E+05, 5.6226E+05,
     + 5.8100E+05, 6.0020E+05, 6.1990E+05, 6.4008E+05, 6.6077E+05,
     +  6.8199E+05, 7.0373E+05, 7.2600E+05, 7.4884E+05, 7.7223E+05,
     + 7.9621E+05, 8.2078E+05, 8.4594E+05, 8.7173E+05, 8.9814E+05,
     +  9.2520E+05, 9.5292E+05, 9.8130E+05, 1.0104E+06, 1.0402E+06,
     + 1.0707E+06, 1.1019E+06, 1.1339E+06, 1.1666E+06, 1.2001E+06,
     +  1.2345E+06, 1.2696E+06, 1.3056E+06, 1.3424E+06, 1.3801E+06,
     + 1.4187E+06, 1.4582E+06, 1.4986E+06, 1.5400E+06, 1.5824E+06,
     +  1.6257E+06, 1.6700E+06, 1.7154E+06, 1.7618E+06, 1.8094E+06,
     + 1.8580E+06, 1.9077E+06, 1.9586E+06, 2.0106E+06, 2.0639E+06,
     +  2.1183E+06, 2.1740E+06, 2.2310E+06, 2.2893E+06, 2.3489E+06,
     + 2.4098E+06, 2.4722E+06, 2.5359E+06, 2.6011E+06, 2.6678E+06,
     +  2.7359E+06, 2.8056E+06, 2.8769E+06, 2.9497E+06, 3.0242E+06,
     + 3.1003E+06, 3.1781E+06, 3.2577E+06, 3.3390E+06, 3.4221E+06,
     +  3.5070E+06, 3.5938E+06, 3.6825E+06, 3.7732E+06, 3.8658E+06,
     + 3.9605E+06, 4.0573E+06, 4.1561E+06, 4.2571E+06, 4.3603E+06,
     +  4.4657E+06, 4.5734E+06, 4.6835E+06, 4.7959E+06, 4.9107E+06,
     + 5.0280E+06, 5.1478E+06, 5.2701E+06, 5.3951E+06, 5.5227E+06,
     +  5.6531E+06, 5.7862E+06, 5.9221E+06, 6.0609E+06, 6.2027E+06,
     + 6.3474E+06, 6.4952E+06, 6.6460E+06, 6.8001E+06, 6.9573E+06,
     +  7.1179E+06, 7.2817E+06, 7.4490E+06, 7.6198E+06, 7.7941E+06,
     + 7.9720E+06, 8.1536E+06, 8.3389E+06, 8.5281E+06, 8.7211E+06,
     +  8.9181E+06, 9.1191E+06, 9.3242E+06, 9.5335E+06, 9.7471E+06,
     + 9.9650E+06, 1.0187E+07, 1.0414E+07, 1.0646E+07, 1.0882E+07,
     +  1.1122E+07, 1.1368E+07, 1.1619E+07, 1.1874E+07, 1.2135E+07,
     + 1.2401E+07, 1.2672E+07, 1.2949E+07, 1.3231E+07, 1.3519E+07,
     +  1.3812E+07, 1.4111E+07, 1.4416E+07, 1.4727E+07, 1.5045E+07,
     + 1.5368E+07, 1.5698E+07, 1.6034E+07, 1.6377E+07, 1.6726E+07,
     +  1.7082E+07, 1.7445E+07, 1.7816E+07, 1.8193E+07, 1.8577E+07,
     + 1.8969E+07, 1.9369E+07, 1.9776E+07, 2.0191E+07, 2.0614E+07,
     +  2.1045E+07, 2.1484E+07, 2.1932E+07, 2.2388E+07, 2.2853E+07,
     + 2.3326E+07, 2.3809E+07, 2.4300E+07, 2.4801E+07, 2.5312E+07,
     +  2.5831E+07, 2.6361E+07, 2.6900E+07, 2.7450E+07, 2.8010E+07,
     + 2.8580E+07, 2.9161E+07, 2.9753E+07, 3.0355E+07, 3.0969E+07,
     +  3.1594E+07, 3.2230E+07, 3.2879E+07, 3.3539E+07, 3.4211E+07,
     + 3.4896E+07, 3.5593E+07, 3.6302E+07, 3.7025E+07, 3.7761E+07,
     +  3.8510E+07, 3.9273E+07, 4.0050E+07, 4.0840E+07, 4.1645E+07,
     + 4.2464E+07, 4.3298E+07, 4.4147E+07, 4.5011E+07, 4.5891E+07,
     +  4.6786E+07, 4.7698E+07, 4.8625E+07, 4.9569E+07, 5.0530E+07,
     + 5.1508E+07, 5.2503E+07, 5.3515E+07, 5.4546E+07, 5.5594E+07,
     +  5.6661E+07, 5.7747E+07, 5.8851E+07, 5.9975E+07, 6.1119E+07,
     + 6.2282E+07, 6.3466E+07, 6.4670E+07, 6.5895E+07, 6.7141E+07,
     +  6.8409E+07, 6.9699E+07, 7.1011E+07, 7.2345E+07, 7.3703E+07,
     + 7.5083E+07, 7.6487E+07, 7.7916E+07, 7.9368E+07, 8.0845E+07,
     +  8.2348E+07, 8.3876E+07, 8.5430E+07, 8.7010E+07, 8.8617E+07,
     + 9.0250E+07, 9.1912E+07, 9.3601E+07, 9.5319E+07, 9.7065E+07,
     +  9.8841E+07, 1.0065E+08, 1.0248E+08, 1.0435E+08, 1.0625E+08,
     + 1.0817E+08, 1.1013E+08, 1.1213E+08, 1.1415E+08, 1.1621E+08,
     +  1.1831E+08, 1.2043E+08, 1.2260E+08, 1.2479E+08, 1.2703E+08,
     + 1.2930E+08, 1.3161E+08, 1.3395E+08, 1.3634E+08, 1.3876E+08,
     +  1.4122E+08, 1.4372E+08, 1.4626E+08, 1.4885E+08, 1.5147E+08,
     + 1.5414E+08, 1.5685E+08, 1.5960E+08, 1.6240E+08, 1.6524E+08,
     +  1.6813E+08, 1.7106E+08, 1.7404E+08, 1.7707E+08, 1.8014E+08,
     + 1.8327E+08, 1.8644E+08, 1.8967E+08, 1.9294E+08, 1.9627E+08,
     +  1.9965E+08, 2.0308E+08, 2.0656E+08, 2.1010E+08, 2.1370E+08,
     + 2.1735E+08, 2.2106E+08, 2.2483E+08, 2.2865E+08, 2.3254E+08,
     +  2.3648E+08, 2.4049E+08, 2.4456E+08, 2.4869E+08, 2.5288E+08,
     + 2.5714E+08, 2.6147E+08, 2.6586E+08, 2.7031E+08, 2.7484E+08,
     +  2.7944E+08, 2.8410E+08, 2.8884E+08, 2.9365E+08, 2.9853E+08,
     + 3.0348E+08, 3.0851E+08, 3.1362E+08, 3.1881E+08, 3.2407E+08,
     +  3.2941E+08, 3.3483E+08, 3.4033E+08, 3.4592E+08, 3.5159E+08,
     + 3.5734E+08, 3.6318E+08, 3.6911E+08, 3.7512E+08, 3.8123E+08,
     +  3.8742E+08, 3.9371E+08, 4.0009E+08, 4.0657E+08, 4.1314E+08,
     + 4.1980E+08, 4.2657E+08, 4.3343E+08, 4.4040E+08, 4.4746E+08,
     +  4.5463E+08, 4.6191E+08, 4.6929E+08, 4.7678E+08, 4.8438E+08,
     + 4.9209E+08, 4.9991E+08, 5.0784E+08, 5.1589E+08, 5.2405E+08,
     +  5.3233E+08, 5.4074E+08, 5.4926E+08, 5.5790E+08, 5.6667E+08,
     + 5.7557E+08, 5.8459E+08, 5.9374E+08, 6.0302E+08, 6.1243E+08,
     +  6.2198E+08, 6.3167E+08, 6.4149E+08, 6.5145E+08, 6.6155E+08,
     + 6.7179E+08, 6.8218E+08, 6.9271E+08, 7.0340E+08, 7.1423E+08,
     +  7.2521E+08, 7.3635E+08, 7.4765E+08, 7.5910E+08, 7.7071E+08,
     + 7.8249E+08, 7.9443E+08, 8.0653E+08, 8.1880E+08, 8.3125E+08,
     +  8.4386E+08, 8.5665E+08, 8.6962E+08, 8.8276E+08, 8.9609E+08,
     + 9.0960E+08, 9.2329E+08, 9.3717E+08, 9.5125E+08, 9.6551E+08,
     +  9.7997E+08, 9.9463E+08, 1.0095E+09, 1.0245E+09, 1.0398E+09,
     + 1.0553E+09, 1.0710E+09, 1.0869E+09, 1.1030E+09, 1.1193E+09,
     +  1.1358E+09, 1.1526E+09, 1.1696E+09, 1.1868E+09, 1.2043E+09,
     + 1.2220E+09, 1.2399E+09, 1.2580E+09, 1.2764E+09, 1.2951E+09,
     +  1.3140E+09, 1.3331E+09, 1.3526E+09, 1.3722E+09, 1.3921E+09,
     + 1.4123E+09, 1.4328E+09, 1.4535E+09, 1.4745E+09, 1.4958E+09,
     +  1.5173E+09, 1.5391E+09, 1.5613E+09, 1.5837E+09, 1.6064E+09,
     + 1.6294E+09, 1.6526E+09, 1.6762E+09, 1.7002E+09, 1.7244E+09,
     +  1.7489E+09, 1.7737E+09, 1.7989E+09, 1.8244E+09, 1.8502E+09,
     + 1.8764E+09, 1.9029E+09, 1.9297E+09, 1.9569E+09, 1.9844E+09,
     +  2.0123E+09, 2.0405E+09, 2.0691E+09, 2.0981E+09, 2.1274E+09,
     + 2.1571E+09, 2.1872E+09, 2.2177E+09, 2.2485E+09, 2.2797E+09,
     +  2.3114E+09, 2.3434E+09, 2.3759E+09, 2.4087E+09, 2.4420E+09,
     + 2.4756E+09, 2.5097E+09, 2.5443E+09, 2.5792E+09, 2.6146E+09,
     +  2.6505E+09, 2.6867E+09, 2.7235E+09, 2.7607E+09, 2.7983E+09,
     + 2.8364E+09, 2.8750E+09, 2.9141E+09, 2.9537E+09, 2.9937E+09,
     +  3.0343E+09, 3.0753E+09, 3.1168E+09, 3.1589E+09, 3.2015E+09,
     + 3.2446E+09, 3.2882E+09, 3.3323E+09, 3.3770E+09, 3.4223E+09,
     +  3.4681E+09, 3.5144E+09, 3.5613E+09, 3.6088E+09, 3.6569E+09,
     + 3.7055E+09, 3.7547E+09, 3.8045E+09, 3.8550E+09, 3.9060E+09,
     +  3.9576E+09, 4.0099E+09, 4.0628E+09, 4.1163E+09, 4.1705E+09,
     + 4.2253E+09, 4.2808E+09, 4.3369E+09, 4.3937E+09, 4.4512E+09,
     +  4.5094E+09, 4.5682E+09, 4.6278E+09, 4.6881E+09, 4.7490E+09,
     + 4.8107E+09, 4.8732E+09, 4.9363E+09, 5.0002E+09, 5.0649E+09,
     +  5.1303E+09, 5.1965E+09, 5.2635E+09, 5.3312E+09, 5.3998E+09,
     + 5.4691E+09, 5.5393E+09, 5.6102E+09, 5.6820E+09, 5.7547E+09,
     +  5.8281E+09, 5.9025E+09, 5.9777E+09, 6.0537E+09, 6.1307E+09,
     + 6.2085E+09, 6.2873E+09, 6.3669E+09, 6.4475E+09, 6.5289E+09,
     +  6.6114E+09, 6.6947E+09, 6.7790E+09, 6.8643E+09, 6.9506E+09,
     + 7.0378E+09, 7.1261E+09, 7.2153E+09, 7.3056E+09, 7.3969E+09,
     +  7.4892E+09, 7.5826E+09, 7.6770E+09, 7.7725E+09, 7.8691E+09,
     + 7.9668E+09, 8.0656E+09, 8.1655E+09, 8.2665E+09, 8.3686E+09,
     +  8.4719E+09, 8.5764E+09, 8.6820E+09, 8.7888E+09, 8.8968E+09,
     + 9.0061E+09, 9.1165E+09, 9.2281E+09, 9.3410E+09, 9.4552E+09,
     +  9.5706E+09, 9.6873E+09, 9.8053E+09, 9.9246E+09, 1.0045E+10,
     + 1.0167E+10, 1.0291E+10, 1.0415E+10, 1.0541E+10, 1.0669E+10,
     +  1.0797E+10, 1.0928E+10, 1.1059E+10, 1.1192E+10, 1.1327E+10,
     + 1.1463E+10, 1.1601E+10, 1.1740E+10, 1.1880E+10, 1.2022E+10,
     +  1.2166E+10, 1.2311E+10, 1.2458E+10, 1.2606E+10, 1.2756E+10,
     + 1.2907E+10, 1.3061E+10, 1.3215E+10, 1.3372E+10, 1.3530E+10,
     +  1.3690E+10, 1.3852E+10, 1.4015E+10, 1.4180E+10, 1.4347E+10,
     + 1.4516E+10, 1.4686E+10, 1.4858E+10, 1.5032E+10, 1.5208E+10,
     +  1.5386E+10, 1.5566E+10, 1.5747E+10, 1.5931E+10, 1.6116E+10,
     + 1.6304E+10, 1.6493E+10, 1.6685E+10, 1.6878E+10, 1.7074E+10,
     +  1.7271E+10, 1.7471E+10, 1.7672E+10, 1.7876E+10, 1.8082E+10,
     + 1.8290E+10, 1.8501E+10, 1.8713E+10, 1.8928E+10, 1.9145E+10,
     +  1.9364E+10, 1.9585E+10, 1.9809E+10, 2.0035E+10, 2.0264E+10,
     + 2.0494E+10, 2.0728E+10, 2.0963E+10, 2.1201E+10, 2.1442E+10,
     +  2.1685E+10, 2.1930E+10, 2.2178E+10, 2.2429E+10, 2.2682E+10,
     + 2.2937E+10, 2.3196E+10, 2.3457E+10, 2.3720E+10, 2.3987E+10,
     +  2.4256E+10, 2.4527E+10, 2.4802E+10, 2.5079E+10, 2.5359E+10,
     + 2.5642E+10, 2.5928E+10, 2.6217E+10, 2.6509E+10, 2.6803E+10,
     +  2.7101E+10, 2.7401E+10, 2.7705E+10, 2.8012E+10, 2.8321E+10,
     + 2.8634E+10, 2.8950E+10, 2.9269E+10, 2.9592E+10, 2.9917E+10,
     +  3.0246E+10, 3.0578E+10, 3.0914E+10, 3.1253E+10, 3.1595E+10,
     + 3.1940E+10, 3.2289E+10, 3.2642E+10, 3.2998E+10, 3.3357E+10,
     +  3.3720E+10, 3.4087E+10, 3.4457E+10, 3.4831E+10, 3.5208E+10,
     + 3.5590E+10, 3.5975E+10, 3.6363E+10, 3.6756E+10, 3.7152E+10,
     +  3.7553E+10, 3.7957E+10, 3.8365E+10, 3.8777E+10, 3.9194E+10,
     + 3.9614E+10, 4.0038E+10, 4.0467E+10, 4.0899E+10, 4.1336E+10,
     +  4.1777E+10, 4.2223E+10, 4.2672E+10, 4.3126E+10, 4.3585E+10,
     + 4.4048E+10, 4.4515E+10, 4.4987E+10, 4.5463E+10, 4.5944E+10,
     +  4.6429E+10, 4.6920E+10, 4.7415E+10, 4.7914E+10, 4.8419E+10,
     + 4.8928E+10, 4.9442E+10, 4.9961E+10, 5.0485E+10, 5.1014E+10,
     +  5.1548E+10, 5.2087E+10, 5.2631E+10, 5.3180E+10, 5.3735E+10,
     + 5.4294E+10, 5.4859E+10, 5.5430E+10, 5.6005E+10, 5.6587E+10,
     +  5.7173E+10, 5.7766E+10, 5.8363E+10, 5.8967E+10, 5.9576E+10,
     + 6.0190E+10, 6.0811E+10, 6.1437E+10, 6.2069E+10, 6.2707E+10,
     +  6.3351E+10, 6.4001E+10, 6.4657E+10, 6.5320E+10, 6.5988E+10,
     + 6.6663E+10, 6.7343E+10, 6.8030E+10, 6.8724E+10, 6.9424E+10,
     +  7.0130E+10, 7.0843E+10, 7.1562E+10, 7.2288E+10, 7.3021E+10,
     + 7.3761E+10, 7.4507E+10, 7.5260E+10, 7.6020E+10, 7.6787E+10,
     +  7.7561E+10, 7.8342E+10, 7.9131E+10, 7.9926E+10, 8.0729E+10,
     + 8.1539E+10, 8.2356E+10, 8.3181E+10, 8.4014E+10, 8.4853E+10,
     +  8.5701E+10, 8.6556E+10, 8.7419E+10, 8.8290E+10, 8.9169E+10,
     + 9.0055E+10, 9.0950E+10, 9.1852E+10, 9.2763E+10, 9.3682E+10,
     +  9.4609E+10, 9.5545E+10, 9.6489E+10, 9.7441E+10, 9.8402E+10,
     + 9.9372E+10, 1.0035E+11, 1.0134E+11, 1.0233E+11, 1.0334E+11,
     +  1.0435E+11, 1.0537E+11, 1.0640E+11, 1.0745E+11, 1.0850E+11,
     + 1.0956E+11, 1.1062E+11, 1.1170E+11, 1.1279E+11, 1.1389E+11,
     +  1.1499E+11, 1.1611E+11, 1.1724E+11, 1.1837E+11, 1.1952E+11,
     + 1.2068E+11, 1.2184E+11, 1.2302E+11, 1.2421E+11, 1.2541E+11,
     +  1.2662E+11, 1.2783E+11, 1.2906E+11, 1.3030E+11, 1.3156E+11,
     + 1.3282E+11, 1.3409E+11, 1.3537E+11, 1.3667E+11, 1.3798E+11,
     +  1.3929E+11, 1.4062E+11, 1.4196E+11, 1.4332E+11, 1.4468E+11,
     + 1.4605E+11, 1.4744E+11, 1.4884E+11, 1.5025E+11, 1.5168E+11,
     +  1.5311E+11, 1.5456E+11, 1.5602E+11, 1.5750E+11, 1.5898E+11,
     + 1.6048E+11, 1.6199E+11, 1.6352E+11, 1.6505E+11, 1.6660E+11,
     +  1.6817E+11, 1.6974E+11, 1.7133E+11, 1.7294E+11, 1.7456E+11,
     + 1.7619E+11, 1.7783E+11, 1.7949E+11, 1.8116E+11, 1.8285E+11,
     +  1.8455E+11, 1.8627E+11, 1.8800E+11, 1.8974E+11, 1.9150E+11,
     + 1.9328E+11, 1.9507E+11, 1.9687E+11, 1.9869E+11, 2.0052E+11,
     +  2.0237E+11, 2.0424E+11, 2.0612E+11, 2.0802E+11, 2.0993E+11,
     + 2.1186E+11, 2.1380E+11, 2.1576E+11, 2.1774E+11, 2.1974E+11,
     +  2.2175E+11, 2.2377E+11, 2.2582E+11, 2.2788E+11, 2.2995E+11,
     + 2.3205E+11, 2.3416E+11, 2.3629E+11, 2.3844E+11, 2.4060E+11,
     +  2.4279E+11, 2.4499E+11, 2.4721E+11, 2.4944E+11, 2.5170E+11,
     + 2.5397E+11, 2.5626E+11, 2.5857E+11, 2.6091E+11, 2.6325E+11,
     +  2.6562E+11, 2.6801E+11, 2.7042E+11, 2.7284E+11, 2.7529E+11,
     + 2.7776E+11, 2.8024E+11, 2.8275E+11, 2.8528E+11, 2.8782E+11,
     +  2.9039E+11, 2.9298E+11, 2.9559E+11, 2.9822E+11, 3.0087E+11,
     + 3.0355E+11, 3.0624E+11, 3.0896E+11, 3.1170E+11, 3.1446E+11,
     +  3.1724E+11, 3.2005E+11, 3.2287E+11, 3.2573E+11, 3.2860E+11,
     + 3.3149E+11, 3.3441E+11, 3.3736E+11, 3.4032E+11, 3.4331E+11,
     +  3.4633E+11, 3.4936E+11, 3.5242E+11, 3.5551E+11, 3.5862E+11,
     + 3.6176E+11, 3.6492E+11, 3.6810E+11, 3.7131E+11, 3.7455E+11,
     +  3.7781E+11, 3.8110E+11, 3.8441E+11, 3.8775E+11, 3.9111E+11,
     + 3.9451E+11, 3.9792E+11, 4.0137E+11, 4.0484E+11, 4.0834E+11,
     +  4.1187E+11, 4.1542E+11, 4.1901E+11, 4.2262E+11, 4.2626E+11,
     + 4.2992E+11, 4.3362E+11, 4.3734E+11, 4.4110E+11, 4.4488E+11,
     +  4.4869E+11, 4.5254E+11, 4.5641E+11, 4.6031E+11, 4.6424E+11,
     + 4.6820E+11, 4.7220E+11, 4.7622E+11, 4.8028E+11, 4.8436E+11,
     +  4.8848E+11, 4.9263E+11, 4.9681E+11, 5.0103E+11, 5.0527E+11,
     + 5.0955E+11, 5.1386E+11, 5.1821E+11, 5.2259E+11, 5.2700E+11,
     +  5.3145E+11, 5.3593E+11, 5.4044E+11, 5.4499E+11, 5.4957E+11,
     + 5.5419E+11, 5.5884E+11, 5.6353E+11, 5.6825E+11, 5.7302E+11,
     +  5.7781E+11, 5.8264E+11, 5.8751E+11, 5.9242E+11, 5.9736E+11,
     + 6.0234E+11, 6.0736E+11, 6.1242E+11, 6.1751E+11, 6.2265E+11,
     +  6.2782E+11, 6.3303E+11, 6.3828E+11, 6.4357E+11, 6.4890E+11,
     + 6.5427E+11, 6.5968E+11, 6.6513E+11, 6.7062E+11, 6.7615E+11,
     +  6.8172E+11, 6.8734E+11, 6.9300E+11, 6.9870E+11, 7.0444E+11,
     + 7.1022E+11, 7.1605E+11, 7.2192E+11, 7.2783E+11, 7.3379E+11,
     +  7.3980E+11, 7.4584E+11, 7.5194E+11, 7.5807E+11, 7.6426E+11,
     + 7.7048E+11, 7.7676E+11, 7.8308E+11, 7.8945E+11, 7.9586E+11,
     +  8.0232E+11, 8.0883E+11, 8.1539E+11, 8.2199E+11, 8.2865E+11,
     + 8.3535E+11, 8.4210E+11, 8.4890E+11, 8.5575E+11, 8.6265E+11,
     +  8.6961E+11, 8.7661E+11, 8.8366E+11, 8.9077E+11, 8.9792E+11,
     + 9.0513E+11, 9.1239E+11, 9.1971E+11, 9.2708E+11, 9.3450E+11,
     +  9.4197E+11, 9.4950E+11, 9.5708E+11, 9.6472E+11, 9.7242E+11,
     + 9.8016E+11, 9.8797E+11, 9.9583E+11, 1.0037E+12, 1.0117E+12,
     +  1.0198E+12, 1.0278E+12, 1.0360E+12, 1.0442E+12, 1.0525E+12,
     + 1.0608E+12, 1.0692E+12, 1.0776E+12, 1.0861E+12, 1.0947E+12,
     +  1.1033E+12, 1.1120E+12, 1.1208E+12, 1.1296E+12, 1.1385E+12,
     + 1.1474E+12, 1.1564E+12, 1.1655E+12, 1.1746E+12, 1.1838E+12,
     +  1.1931E+12, 1.2024E+12, 1.2118E+12, 1.2213E+12, 1.2308E+12,
     + 1.2404E+12, 1.2500E+12, 1.2598E+12, 1.2696E+12, 1.2794E+12,
     +  1.2894E+12, 1.2994E+12, 1.3095E+12, 1.3196E+12, 1.3298E+12,
     + 1.3401E+12, 1.3505E+12, 1.3609E+12, 1.3715E+12, 1.3820E+12,
     +  1.3927E+12, 1.4034E+12, 1.4142E+12, 1.4251E+12, 1.4361E+12,
     + 1.4471E+12, 1.4582E+12, 1.4694E+12, 1.4807E+12, 1.4920E+12,
     +  1.5035E+12, 1.5150E+12, 1.5266E+12, 1.5382E+12, 1.5500E+12,
     + 1.5618E+12, 1.5737E+12, 1.5857E+12, 1.5978E+12, 1.6099E+12,
     +  1.6222E+12, 1.6345E+12, 1.6469E+12, 1.6594E+12, 1.6720E+12,
     + 1.6847E+12, 1.6974E+12, 1.7103E+12, 1.7232E+12, 1.7362E+12,
     +  1.7493E+12, 1.7625E+12, 1.7758E+12, 1.7892E+12, 1.8027E+12,
     + 1.8162E+12, 1.8299E+12, 1.8436E+12, 1.8575E+12, 1.8714E+12,
     +  1.8854E+12, 1.8996E+12, 1.9138E+12, 1.9281E+12, 1.9425E+12,
     + 1.9570E+12, 1.9717E+12, 1.9864E+12, 2.0012E+12, 2.0161E+12,
     +  2.0311E+12, 2.0462E+12, 2.0614E+12, 2.0767E+12, 2.0922E+12,
     + 2.1077E+12, 2.1233E+12, 2.1391E+12, 2.1549E+12, 2.1708E+12,
     +  2.1869E+12, 2.2031E+12, 2.2193E+12, 2.2357E+12, 2.2522E+12,
     + 2.2688E+12, 2.2855E+12, 2.3023E+12, 2.3193E+12, 2.3363E+12,
     +  2.3535E+12, 2.3707E+12, 2.3881E+12, 2.4056E+12, 2.4233E+12,
     + 2.4410E+12, 2.4588E+12, 2.4768E+12, 2.4949E+12, 2.5131E+12,
     +  2.5315E+12, 2.5499E+12, 2.5685E+12, 2.5872E+12, 2.6060E+12,
     + 2.6250E+12, 2.6440E+12, 2.6632E+12, 2.6825E+12, 2.7020E+12,
     +  2.7216E+12, 2.7413E+12, 2.7611E+12, 2.7811E+12, 2.8012E+12,
     + 2.8214E+12, 2.8418E+12, 2.8623E+12, 2.8829E+12, 2.9037E+12,
     +  2.9246E+12, 2.9456E+12, 2.9668E+12, 2.9881E+12, 3.0095E+12,
     + 3.0311E+12, 3.0528E+12, 3.0747E+12, 3.0967E+12, 3.1189E+12,
     +  3.1412E+12, 3.1636E+12, 3.1862E+12, 3.2089E+12, 3.2318E+12,
     + 3.2549E+12, 3.2780E+12, 3.3014E+12, 3.3248E+12, 3.3485E+12,
     +  3.3723E+12, 3.3962E+12, 3.4203E+12, 3.4445E+12, 3.4689E+12,
     + 3.4935E+12, 3.5182E+12, 3.5431E+12, 3.5681E+12, 3.5933E+12,
     +  3.6186E+12, 3.6442E+12, 3.6698E+12, 3.6957E+12, 3.7217E+12,
     + 3.7479E+12, 3.7742E+12, 3.8007E+12, 3.8274E+12, 3.8542E+12,
     +  3.8812E+12, 3.9084E+12, 3.9358E+12, 3.9633E+12, 3.9910E+12,
     + 4.0189E+12, 4.0470E+12, 4.0752E+12, 4.1037E+12, 4.1322E+12,
     +  4.1610E+12, 4.1900E+12, 4.2191E+12, 4.2485E+12, 4.2780E+12,
     + 4.3077E+12, 4.3376E+12, 4.3676E+12, 4.3979E+12, 4.4283E+12,
     +  4.4590E+12, 4.4898E+12, 4.5208E+12, 4.5521E+12, 4.5835E+12,
     + 4.6151E+12, 4.6469E+12, 4.6789E+12, 4.7111E+12, 4.7435E+12,
     +  4.7762E+12, 4.8090E+12, 4.8420E+12, 4.8752E+12, 4.9087E+12,
     + 4.9423E+12, 4.9761E+12, 5.0102E+12, 5.0445E+12, 5.0790E+12,
     +  5.1137E+12, 5.1486E+12, 5.1837E+12, 5.2190E+12, 5.2546E+12,
     + 5.2904E+12, 5.3264E+12, 5.3626E+12, 5.3991E+12, 5.4358E+12,
     +  5.4727E+12, 5.5098E+12, 5.5471E+12, 5.5847E+12, 5.6225E+12,
     + 5.6606E+12, 5.6989E+12, 5.7374E+12, 5.7761E+12, 5.8151E+12,
     +  5.8544E+12, 5.8938E+12, 5.9335E+12, 5.9735E+12, 6.0137E+12,
     + 6.0541E+12, 6.0948E+12, 6.1358E+12, 6.1769E+12, 6.2184E+12,
     +  6.2601E+12, 6.3020E+12, 6.3442E+12, 6.3866E+12, 6.4294E+12,
     + 6.4723E+12, 6.5156E+12, 6.5590E+12, 6.6028E+12, 6.6468E+12,
     +  6.6911E+12, 6.7356E+12, 6.7805E+12, 6.8256E+12, 6.8709E+12,
     + 6.9166E+12, 6.9625E+12, 7.0087E+12, 7.0551E+12, 7.1019E+12,
     +  7.1489E+12, 7.1962E+12, 7.2438E+12, 7.2916E+12, 7.3398E+12,
     + 7.3883E+12, 7.4370E+12, 7.4860E+12, 7.5354E+12, 7.5850E+12,
     +  7.6349E+12, 7.6851E+12, 7.7356E+12, 7.7864E+12, 7.8375E+12,
     + 7.8890E+12, 7.9407E+12, 7.9927E+12, 8.0451E+12, 8.0977E+12,
     +  8.1507E+12, 8.2040E+12, 8.2575E+12, 8.3115E+12, 8.3657E+12,
     + 8.4202E+12, 8.4751E+12, 8.5303E+12, 8.5858E+12, 8.6417E+12,
     +  8.6979E+12, 8.7544E+12, 8.8112E+12, 8.8684E+12, 8.9259E+12,
     + 8.9838E+12, 9.0420E+12, 9.1005E+12, 9.1594E+12, 9.2186E+12,
     +  9.2782E+12, 9.3381E+12, 9.3984E+12, 9.4590E+12, 9.5200E+12,
     + 9.5813E+12, 9.6430E+12, 9.7050E+12, 9.7674E+12, 9.8302E+12,
     +  9.8934E+12, 9.9569E+12, 1.0021E+13, 1.0085E+13, 1.0150E+13,
     + 1.0215E+13, 1.0280E+13, 1.0346E+13, 1.0412E+13, 1.0478E+13,
     +  1.0545E+13, 1.0613E+13, 1.0680E+13, 1.0748E+13, 1.0817E+13,
     + 1.0886E+13, 1.0955E+13, 1.1025E+13, 1.1095E+13, 1.1165E+13,
     +  1.1236E+13, 1.1307E+13, 1.1379E+13, 1.1451E+13, 1.1524E+13,
     + 1.1597E+13, 1.1670E+13, 1.1744E+13, 1.1818E+13, 1.1892E+13,
     +  1.1967E+13, 1.2043E+13, 1.2119E+13, 1.2195E+13, 1.2272E+13,
     + 1.2349E+13, 1.2427E+13, 1.2505E+13, 1.2583E+13, 1.2662E+13,
     +  1.2742E+13, 1.2822E+13, 1.2902E+13, 1.2983E+13, 1.3064E+13,
     + 1.3146E+13, 1.3228E+13, 1.3311E+13, 1.3394E+13, 1.3477E+13,
     +  1.3561E+13, 1.3646E+13, 1.3731E+13, 1.3816E+13, 1.3902E+13,
     + 1.3989E+13, 1.4076E+13, 1.4163E+13, 1.4251E+13, 1.4340E+13,
     +  1.4429E+13, 1.4518E+13, 1.4608E+13, 1.4698E+13, 1.4789E+13,
     + 1.4881E+13, 1.4973E+13, 1.5065E+13, 1.5158E+13, 1.5252E+13,
     +  1.5346E+13, 1.5440E+13, 1.5535E+13, 1.5631E+13, 1.5727E+13,
     + 1.5824E+13, 1.5921E+13, 1.6019E+13, 1.6117E+13, 1.6216E+13,
     +  1.6315E+13, 1.6415E+13, 1.6516E+13, 1.6617E+13, 1.6718E+13,
     + 1.6821E+13, 1.6923E+13, 1.7027E+13, 1.7131E+13, 1.7235E+13,
     +  1.7340E+13, 1.7446E+13, 1.7552E+13, 1.7659E+13, 1.7766E+13,
     + 1.7874E+13, 1.7983E+13, 1.8092E+13, 1.8202E+13, 1.8312E+13,
     +  1.8423E+13, 1.8535E+13, 1.8647E+13, 1.8760E+13, 1.8873E+13,
     + 1.8987E+13, 1.9102E+13, 1.9217E+13, 1.9333E+13, 1.9450E+13,
     +  1.9567E+13, 1.9685E+13, 1.9803E+13, 1.9922E+13, 2.0042E+13,
     + 2.0163E+13, 2.0284E+13, 2.0406E+13, 2.0528E+13, 2.0651E+13,
     +  2.0775E+13, 2.0899E+13, 2.1024E+13, 2.1150E+13, 2.1277E+13,
     + 2.1404E+13, 2.1532E+13, 2.1660E+13, 2.1789E+13, 2.1919E+13,
     +  2.2050E+13, 2.2181E+13, 2.2313E+13, 2.2446E+13, 2.2580E+13,
     + 2.2714E+13, 2.2849E+13, 2.2984E+13, 2.3121E+13, 2.3258E+13,
     +  2.3396E+13, 2.3534E+13, 2.3674E+13, 2.3814E+13, 2.3954E+13,
     + 2.4096E+13, 2.4238E+13, 2.4382E+13, 2.4525E+13, 2.4670E+13,
     +  2.4815E+13, 2.4962E+13, 2.5109E+13, 2.5256E+13, 2.5405E+13,
     + 2.5554E+13, 2.5704E+13, 2.5855E+13, 2.6007E+13, 2.6160E+13,
     +  2.6313E+13, 2.6467E+13, 2.6622E+13, 2.6778E+13, 2.6934E+13,
     + 2.7092E+13, 2.7250E+13, 2.7409E+13, 2.7569E+13, 2.7730E+13,
     +  2.7892E+13, 2.8054E+13, 2.8218E+13, 2.8382E+13, 2.8547E+13,
     + 2.8713E+13, 2.8880E+13, 2.9048E+13, 2.9216E+13, 2.9386E+13,
     +  2.9556E+13, 2.9728E+13, 2.9900E+13, 3.0073E+13, 3.0247E+13,
     + 3.0422E+13, 3.0598E+13, 3.0774E+13, 3.0952E+13, 3.1131E+13,
     +  3.1310E+13, 3.1491E+13, 3.1672E+13, 3.1855E+13, 3.2038E+13,
     + 3.2222E+13, 3.2407E+13, 3.2594E+13, 3.2781E+13, 3.2969E+13,
     +  3.3158E+13, 3.3348E+13, 3.3539E+13, 3.3731E+13, 3.3924E+13,
     + 3.4119E+13, 3.4314E+13, 3.4510E+13, 3.4707E+13, 3.4905E+13,
     +  3.5104E+13, 3.5304E+13, 3.5506E+13, 3.5708E+13, 3.5911E+13,
     + 3.6116E+13, 3.6321E+13, 3.6528E+13, 3.6735E+13, 3.6944E+13,
     +  3.7153E+13, 3.7364E+13, 3.7576E+13, 3.7789E+13, 3.8003E+13,
     + 3.8218E+13, 3.8434E+13, 3.8652E+13, 3.8870E+13, 3.9090E+13,
     +  3.9311E+13, 3.9532E+13, 3.9755E+13, 3.9979E+13, 4.0205E+13,
     + 4.0431E+13, 4.0659E+13, 4.0887E+13, 4.1117E+13, 4.1348E+13,
     +  4.1581E+13, 4.1814E+13, 4.2049E+13, 4.2284E+13, 4.2521E+13,
     + 4.2760E+13, 4.2999E+13, 4.3240E+13, 4.3481E+13, 4.3725E+13,
     +  4.3969E+13, 4.4214E+13, 4.4461E+13, 4.4709E+13, 4.4958E+13,
     + 4.5209E+13, 4.5461E+13, 4.5714E+13, 4.5968E+13, 4.6224E+13,
     +  4.6481E+13, 4.6739E+13, 4.6999E+13, 4.7259E+13, 4.7522E+13,
     + 4.7785E+13, 4.8050E+13, 4.8316E+13, 4.8583E+13, 4.8852E+13,
     +  4.9122E+13, 4.9394E+13, 4.9667E+13, 4.9941E+13, 5.0216E+13,
     + 5.0493E+13, 5.0772E+13, 5.1051E+13, 5.1333E+13, 5.1615E+13,
     +  5.1899E+13, 5.2184E+13, 5.2471E+13, 5.2759E+13, 5.3049E+13,
     + 5.3340E+13, 5.3633E+13, 5.3927E+13, 5.4222E+13, 5.4519E+13,
     +  5.4818E+13, 5.5117E+13, 5.5419E+13, 5.5722E+13, 5.6026E+13,
     + 5.6332E+13, 5.6639E+13, 5.6948E+13, 5.7259E+13, 5.7570E+13,
     +  5.7884E+13, 5.8199E+13, 5.8516E+13, 5.8834E+13, 5.9154E+13,
     + 5.9475E+13, 5.9798E+13, 6.0122E+13, 6.0448E+13, 6.0776E+13,
     +  6.1105E+13, 6.1436E+13, 6.1769E+13, 6.2103E+13, 6.2439E+13,
     + 6.2776E+13, 6.3115E+13, 6.3456E+13, 6.3798E+13, 6.4142E+13,
     +  6.4488E+13, 6.4836E+13, 6.5185E+13, 6.5536E+13, 6.5888E+13,
     + 6.6242E+13, 6.6598E+13, 6.6956E+13, 6.7316E+13, 6.7677E+13,
     +  6.8040E+13, 6.8405E+13, 6.8771E+13, 6.9140E+13, 6.9510E+13,
     + 6.9882E+13, 7.0255E+13, 7.0631E+13, 7.1008E+13, 7.1387E+13,
     +  7.1768E+13, 7.2151E+13, 7.2536E+13, 7.2922E+13, 7.3311E+13,
     + 7.3701E+13, 7.4093E+13, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_NOp   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 1),QofT( 1,NT),Q(NT)
      data Tmx(1)/3500./
      data xgj/ 3./
c...      NO+
c...        --        46
      data (QofT( 1,J),J=1,2408)/ 3.0295E+00, 3.5185E+00, 4.3870E+00,
     +  5.3638E+00, 6.3744E+00, 7.3994E+00, 8.4319E+00, 9.4689E+00,
     + 1.0509E+01, 1.1550E+01, 1.2594E+01, 1.3638E+01, 1.4683E+01,
     +  1.5729E+01, 1.6775E+01, 1.7821E+01, 1.8868E+01, 1.9915E+01,
     + 2.0962E+01, 2.2010E+01, 2.4105E+01, 2.6201E+01, 2.8297E+01,
     +  3.0394E+01, 3.2490E+01, 3.4587E+01, 3.6685E+01, 3.8782E+01,
     + 4.0879E+01, 4.2977E+01, 4.5074E+01, 4.7172E+01, 4.9270E+01,
     +  5.1368E+01, 5.3465E+01, 5.5563E+01, 5.7661E+01, 5.9759E+01,
     + 6.1857E+01, 6.3956E+01, 6.6054E+01, 6.8152E+01, 7.0250E+01,
     +  7.2348E+01, 7.4447E+01, 7.6545E+01, 7.8643E+01, 8.0742E+01,
     + 8.2840E+01, 8.4938E+01, 8.7037E+01, 8.9135E+01, 9.1234E+01,
     +  9.3332E+01, 9.5431E+01, 9.7530E+01, 9.9628E+01, 1.0173E+02,
     + 1.0383E+02, 1.0592E+02, 1.0802E+02, 1.1012E+02, 1.1222E+02,
     +  1.1432E+02, 1.1642E+02, 1.1852E+02, 1.2062E+02, 1.2271E+02,
     + 1.2481E+02, 1.2691E+02, 1.2901E+02, 1.3111E+02, 1.3321E+02,
     +  1.3531E+02, 1.3741E+02, 1.3951E+02, 1.4160E+02, 1.4370E+02,
     + 1.4580E+02, 1.4790E+02, 1.5000E+02, 1.5210E+02, 1.5420E+02,
     +  1.5630E+02, 1.5840E+02, 1.6050E+02, 1.6260E+02, 1.6469E+02,
     + 1.6679E+02, 1.6889E+02, 1.7099E+02, 1.7309E+02, 1.7519E+02,
     +  1.7729E+02, 1.7939E+02, 1.8149E+02, 1.8359E+02, 1.8569E+02,
     + 1.8779E+02, 1.8989E+02, 1.9199E+02, 1.9409E+02, 1.9618E+02,
     +  1.9828E+02, 2.0038E+02, 2.0248E+02, 2.0458E+02, 2.0668E+02,
     + 2.0878E+02, 2.1088E+02, 2.1298E+02, 2.1508E+02, 2.1718E+02,
     +  2.1928E+02, 2.2138E+02, 2.2348E+02, 2.2558E+02, 2.2768E+02,
     + 2.2978E+02, 2.3188E+02, 2.3398E+02, 2.3608E+02, 2.3818E+02,
     +  2.4028E+02, 2.4238E+02, 2.4448E+02, 2.4658E+02, 2.4868E+02,
     + 2.5078E+02, 2.5288E+02, 2.5498E+02, 2.5708E+02, 2.5918E+02,
     +  2.6128E+02, 2.6338E+02, 2.6548E+02, 2.6758E+02, 2.6968E+02,
     + 2.7178E+02, 2.7388E+02, 2.7598E+02, 2.7808E+02, 2.8018E+02,
     +  2.8228E+02, 2.8438E+02, 2.8648E+02, 2.8858E+02, 2.9068E+02,
     + 2.9278E+02, 2.9488E+02, 2.9698E+02, 2.9908E+02, 3.0118E+02,
     +  3.0328E+02, 3.0538E+02, 3.0748E+02, 3.0958E+02, 3.1169E+02,
     + 3.1379E+02, 3.1589E+02, 3.1799E+02, 3.2009E+02, 3.2219E+02,
     +  3.2429E+02, 3.2639E+02, 3.2849E+02, 3.3059E+02, 3.3269E+02,
     + 3.3479E+02, 3.3690E+02, 3.3900E+02, 3.4110E+02, 3.4320E+02,
     +  3.4530E+02, 3.4741E+02, 3.4951E+02, 3.5161E+02, 3.5371E+02,
     + 3.5581E+02, 3.5791E+02, 3.6001E+02, 3.6212E+02, 3.6422E+02,
     +  3.6632E+02, 3.6843E+02, 3.7053E+02, 3.7263E+02, 3.7473E+02,
     + 3.7683E+02, 3.7894E+02, 3.8104E+02, 3.8314E+02, 3.8525E+02,
     +  3.8735E+02, 3.8945E+02, 3.9156E+02, 3.9366E+02, 3.9576E+02,
     + 3.9787E+02, 3.9997E+02, 4.0208E+02, 4.0418E+02, 4.0628E+02,
     +  4.0839E+02, 4.1049E+02, 4.1260E+02, 4.1470E+02, 4.1681E+02,
     + 4.1891E+02, 4.2102E+02, 4.2313E+02, 4.2523E+02, 4.2734E+02,
     +  4.2944E+02, 4.3155E+02, 4.3366E+02, 4.3576E+02, 4.3787E+02,
     + 4.3998E+02, 4.4209E+02, 4.4419E+02, 4.4630E+02, 4.4841E+02,
     +  4.5052E+02, 4.5262E+02, 4.5474E+02, 4.5684E+02, 4.5896E+02,
     + 4.6106E+02, 4.6317E+02, 4.6529E+02, 4.6739E+02, 4.6951E+02,
     +  4.7162E+02, 4.7373E+02, 4.7584E+02, 4.7796E+02, 4.8007E+02,
     + 4.8218E+02, 4.8429E+02, 4.8641E+02, 4.8852E+02, 4.9063E+02,
     +  4.9275E+02, 4.9486E+02, 4.9698E+02, 4.9909E+02, 5.0121E+02,
     + 5.0332E+02, 5.0544E+02, 5.0756E+02, 5.0968E+02, 5.1179E+02,
     +  5.1391E+02, 5.1603E+02, 5.1815E+02, 5.2027E+02, 5.2239E+02,
     + 5.2451E+02, 5.2663E+02, 5.2875E+02, 5.3088E+02, 5.3300E+02,
     +  5.3512E+02, 5.3725E+02, 5.3937E+02, 5.4150E+02, 5.4362E+02,
     + 5.4575E+02, 5.4787E+02, 5.5000E+02, 5.5213E+02, 5.5426E+02,
     +  5.5638E+02, 5.5851E+02, 5.6064E+02, 5.6277E+02, 5.6490E+02,
     + 5.6703E+02, 5.6917E+02, 5.7130E+02, 5.7343E+02, 5.7556E+02,
     +  5.7770E+02, 5.7983E+02, 5.8197E+02, 5.8411E+02, 5.8624E+02,
     + 5.8838E+02, 5.9052E+02, 5.9266E+02, 5.9480E+02, 5.9694E+02,
     +  5.9908E+02, 6.0122E+02, 6.0336E+02, 6.0551E+02, 6.0765E+02,
     + 6.0980E+02, 6.1194E+02, 6.1409E+02, 6.1624E+02, 6.1838E+02,
     +  6.2053E+02, 6.2268E+02, 6.2483E+02, 6.2698E+02, 6.2913E+02,
     + 6.3128E+02, 6.3344E+02, 6.3559E+02, 6.3775E+02, 6.3991E+02,
     +  6.4206E+02, 6.4422E+02, 6.4637E+02, 6.4854E+02, 6.5069E+02,
     + 6.5286E+02, 6.5502E+02, 6.5718E+02, 6.5935E+02, 6.6151E+02,
     +  6.6368E+02, 6.6584E+02, 6.6801E+02, 6.7018E+02, 6.7235E+02,
     + 6.7452E+02, 6.7669E+02, 6.7886E+02, 6.8104E+02, 6.8321E+02,
     +  6.8539E+02, 6.8756E+02, 6.8974E+02, 6.9192E+02, 6.9410E+02,
     + 6.9628E+02, 6.9846E+02, 7.0064E+02, 7.0283E+02, 7.0501E+02,
     +  7.0720E+02, 7.0938E+02, 7.1157E+02, 7.1376E+02, 7.1595E+02,
     + 7.1814E+02, 7.2033E+02, 7.2253E+02, 7.2472E+02, 7.2692E+02,
     +  7.2912E+02, 7.3131E+02, 7.3351E+02, 7.3571E+02, 7.3791E+02,
     + 7.4011E+02, 7.4232E+02, 7.4452E+02, 7.4673E+02, 7.4894E+02,
     +  7.5115E+02, 7.5335E+02, 7.5556E+02, 7.5778E+02, 7.5999E+02,
     + 7.6220E+02, 7.6443E+02, 7.6664E+02, 7.6886E+02, 7.7108E+02,
     +  7.7330E+02, 7.7552E+02, 7.7775E+02, 7.7997E+02, 7.8220E+02,
     + 7.8442E+02, 7.8665E+02, 7.8888E+02, 7.9112E+02, 7.9334E+02,
     +  7.9557E+02, 7.9781E+02, 8.0005E+02, 8.0229E+02, 8.0453E+02,
     + 8.0677E+02, 8.0901E+02, 8.1125E+02, 8.1350E+02, 8.1575E+02,
     +  8.1799E+02, 8.2024E+02, 8.2249E+02, 8.2474E+02, 8.2700E+02,
     + 8.2925E+02, 8.3151E+02, 8.3376E+02, 8.3602E+02, 8.3829E+02,
     +  8.4054E+02, 8.4281E+02, 8.4507E+02, 8.4734E+02, 8.4961E+02,
     + 8.5188E+02, 8.5415E+02, 8.5642E+02, 8.5869E+02, 8.6096E+02,
     +  8.6325E+02, 8.6552E+02, 8.6780E+02, 8.7008E+02, 8.7236E+02,
     + 8.7465E+02, 8.7693E+02, 8.7922E+02, 8.8151E+02, 8.8380E+02,
     +  8.8609E+02, 8.8839E+02, 8.9068E+02, 8.9298E+02, 8.9527E+02,
     + 8.9757E+02, 8.9987E+02, 9.0218E+02, 9.0448E+02, 9.0678E+02,
     +  9.0909E+02, 9.1140E+02, 9.1371E+02, 9.1603E+02, 9.1834E+02,
     + 9.2066E+02, 9.2298E+02, 9.2530E+02, 9.2761E+02, 9.2994E+02,
     +  9.3226E+02, 9.3458E+02, 9.3691E+02, 9.3924E+02, 9.4157E+02,
     + 9.4390E+02, 9.4623E+02, 9.4857E+02, 9.5091E+02, 9.5325E+02,
     +  9.5559E+02, 9.5793E+02, 9.6027E+02, 9.6262E+02, 9.6497E+02,
     + 9.6731E+02, 9.6966E+02, 9.7202E+02, 9.7437E+02, 9.7672E+02,
     +  9.7908E+02, 9.8144E+02, 9.8380E+02, 9.8616E+02, 9.8853E+02,
     + 9.9090E+02, 9.9326E+02, 9.9564E+02, 9.9801E+02, 1.0004E+03,
     +  1.0028E+03, 1.0051E+03, 1.0075E+03, 1.0099E+03, 1.0123E+03,
     + 1.0147E+03, 1.0170E+03, 1.0194E+03, 1.0218E+03, 1.0242E+03,
     +  1.0266E+03, 1.0290E+03, 1.0314E+03, 1.0338E+03, 1.0362E+03,
     + 1.0386E+03, 1.0410E+03, 1.0434E+03, 1.0458E+03, 1.0482E+03,
     +  1.0506E+03, 1.0530E+03, 1.0555E+03, 1.0579E+03, 1.0603E+03,
     + 1.0627E+03, 1.0651E+03, 1.0676E+03, 1.0700E+03, 1.0724E+03,
     +  1.0749E+03, 1.0773E+03, 1.0797E+03, 1.0822E+03, 1.0846E+03,
     + 1.0870E+03, 1.0895E+03, 1.0919E+03, 1.0944E+03, 1.0968E+03,
     +  1.0993E+03, 1.1017E+03, 1.1042E+03, 1.1066E+03, 1.1091E+03,
     + 1.1115E+03, 1.1140E+03, 1.1165E+03, 1.1189E+03, 1.1214E+03,
     +  1.1239E+03, 1.1263E+03, 1.1288E+03, 1.1313E+03, 1.1337E+03,
     + 1.1362E+03, 1.1387E+03, 1.1412E+03, 1.1437E+03, 1.1462E+03,
     +  1.1486E+03, 1.1511E+03, 1.1536E+03, 1.1561E+03, 1.1586E+03,
     + 1.1611E+03, 1.1636E+03, 1.1661E+03, 1.1686E+03, 1.1711E+03,
     +  1.1736E+03, 1.1761E+03, 1.1787E+03, 1.1812E+03, 1.1837E+03,
     + 1.1862E+03, 1.1887E+03, 1.1912E+03, 1.1938E+03, 1.1963E+03,
     +  1.1988E+03, 1.2014E+03, 1.2039E+03, 1.2064E+03, 1.2090E+03,
     + 1.2115E+03, 1.2140E+03, 1.2166E+03, 1.2191E+03, 1.2217E+03,
     +  1.2242E+03, 1.2268E+03, 1.2293E+03, 1.2319E+03, 1.2344E+03,
     + 1.2370E+03, 1.2395E+03, 1.2421E+03, 1.2447E+03, 1.2473E+03,
     +  1.2498E+03, 1.2524E+03, 1.2550E+03, 1.2575E+03, 1.2601E+03,
     + 1.2627E+03, 1.2653E+03, 1.2679E+03, 1.2704E+03, 1.2730E+03,
     +  1.2756E+03, 1.2782E+03, 1.2808E+03, 1.2834E+03, 1.2860E+03,
     + 1.2886E+03, 1.2912E+03, 1.2938E+03, 1.2964E+03, 1.2990E+03,
     +  1.3016E+03, 1.3042E+03, 1.3069E+03, 1.3095E+03, 1.3121E+03,
     + 1.3147E+03, 1.3174E+03, 1.3200E+03, 1.3226E+03, 1.3252E+03,
     +  1.3279E+03, 1.3305E+03, 1.3331E+03, 1.3358E+03, 1.3384E+03,
     + 1.3411E+03, 1.3437E+03, 1.3464E+03, 1.3490E+03, 1.3517E+03,
     +  1.3543E+03, 1.3570E+03, 1.3596E+03, 1.3623E+03, 1.3649E+03,
     + 1.3676E+03, 1.3703E+03, 1.3729E+03, 1.3756E+03, 1.3783E+03,
     +  1.3810E+03, 1.3836E+03, 1.3863E+03, 1.3890E+03, 1.3917E+03,
     + 1.3944E+03, 1.3971E+03, 1.3998E+03, 1.4024E+03, 1.4051E+03,
     +  1.4078E+03, 1.4105E+03, 1.4132E+03, 1.4159E+03, 1.4187E+03,
     + 1.4214E+03, 1.4241E+03, 1.4268E+03, 1.4295E+03, 1.4322E+03,
     +  1.4349E+03, 1.4376E+03, 1.4404E+03, 1.4431E+03, 1.4458E+03,
     + 1.4486E+03, 1.4513E+03, 1.4540E+03, 1.4568E+03, 1.4595E+03,
     +  1.4623E+03, 1.4650E+03, 1.4677E+03, 1.4705E+03, 1.4732E+03,
     + 1.4760E+03, 1.4787E+03, 1.4815E+03, 1.4843E+03, 1.4870E+03,
     +  1.4898E+03, 1.4926E+03, 1.4953E+03, 1.4981E+03, 1.5009E+03,
     + 1.5036E+03, 1.5064E+03, 1.5092E+03, 1.5120E+03, 1.5148E+03,
     +  1.5175E+03, 1.5203E+03, 1.5231E+03, 1.5259E+03, 1.5287E+03,
     + 1.5315E+03, 1.5343E+03, 1.5371E+03, 1.5399E+03, 1.5427E+03,
     +  1.5455E+03, 1.5483E+03, 1.5511E+03, 1.5540E+03, 1.5568E+03,
     + 1.5596E+03, 1.5624E+03, 1.5653E+03, 1.5681E+03, 1.5709E+03,
     +  1.5737E+03, 1.5766E+03, 1.5794E+03, 1.5823E+03, 1.5851E+03,
     + 1.5879E+03, 1.5908E+03, 1.5936E+03, 1.5965E+03, 1.5993E+03,
     +  1.6022E+03, 1.6050E+03, 1.6079E+03, 1.6107E+03, 1.6136E+03,
     + 1.6165E+03, 1.6193E+03, 1.6222E+03, 1.6251E+03, 1.6280E+03,
     +  1.6308E+03, 1.6337E+03, 1.6366E+03, 1.6395E+03, 1.6424E+03,
     + 1.6453E+03, 1.6482E+03, 1.6510E+03, 1.6539E+03, 1.6568E+03,
     +  1.6597E+03, 1.6626E+03, 1.6655E+03, 1.6685E+03, 1.6714E+03,
     + 1.6743E+03, 1.6772E+03, 1.6801E+03, 1.6830E+03, 1.6859E+03,
     +  1.6889E+03, 1.6918E+03, 1.6947E+03, 1.6977E+03, 1.7006E+03,
     + 1.7035E+03, 1.7065E+03, 1.7094E+03, 1.7123E+03, 1.7153E+03,
     +  1.7182E+03, 1.7212E+03, 1.7241E+03, 1.7271E+03, 1.7300E+03,
     + 1.7330E+03, 1.7360E+03, 1.7389E+03, 1.7419E+03, 1.7448E+03,
     +  1.7478E+03, 1.7508E+03, 1.7538E+03, 1.7567E+03, 1.7597E+03,
     + 1.7627E+03, 1.7657E+03, 1.7687E+03, 1.7717E+03, 1.7746E+03,
     +  1.7776E+03, 1.7806E+03, 1.7836E+03, 1.7866E+03, 1.7896E+03,
     + 1.7926E+03, 1.7957E+03, 1.7987E+03, 1.8017E+03, 1.8047E+03,
     +  1.8077E+03, 1.8107E+03, 1.8137E+03, 1.8168E+03, 1.8198E+03,
     + 1.8228E+03, 1.8259E+03, 1.8289E+03, 1.8319E+03, 1.8350E+03,
     +  1.8380E+03, 1.8410E+03, 1.8441E+03, 1.8471E+03, 1.8502E+03,
     + 1.8532E+03, 1.8563E+03, 1.8593E+03, 1.8624E+03, 1.8655E+03,
     +  1.8685E+03, 1.8716E+03, 1.8747E+03, 1.8777E+03, 1.8808E+03,
     + 1.8839E+03, 1.8870E+03, 1.8900E+03, 1.8931E+03, 1.8962E+03,
     +  1.8993E+03, 1.9024E+03, 1.9055E+03, 1.9086E+03, 1.9117E+03,
     + 1.9148E+03, 1.9179E+03, 1.9210E+03, 1.9241E+03, 1.9272E+03,
     +  1.9303E+03, 1.9334E+03, 1.9365E+03, 1.9396E+03, 1.9428E+03,
     + 1.9459E+03, 1.9490E+03, 1.9521E+03, 1.9553E+03, 1.9584E+03,
     +  1.9615E+03, 1.9647E+03, 1.9678E+03, 1.9710E+03, 1.9741E+03,
     + 1.9772E+03, 1.9804E+03, 1.9835E+03, 1.9867E+03, 1.9899E+03,
     +  1.9930E+03, 1.9962E+03, 1.9993E+03, 2.0025E+03, 2.0057E+03,
     + 2.0088E+03, 2.0120E+03, 2.0152E+03, 2.0184E+03, 2.0215E+03,
     +  2.0247E+03, 2.0279E+03, 2.0311E+03, 2.0343E+03, 2.0375E+03,
     + 2.0407E+03, 2.0439E+03, 2.0471E+03, 2.0503E+03, 2.0535E+03,
     +  2.0567E+03, 2.0599E+03, 2.0631E+03, 2.0663E+03, 2.0695E+03,
     + 2.0727E+03, 2.0760E+03, 2.0792E+03, 2.0824E+03, 2.0857E+03,
     +  2.0889E+03, 2.0921E+03, 2.0954E+03, 2.0986E+03, 2.1018E+03,
     + 2.1051E+03, 2.1083E+03, 2.1116E+03, 2.1148E+03, 2.1181E+03,
     +  2.1213E+03, 2.1246E+03, 2.1278E+03, 2.1311E+03, 2.1344E+03,
     + 2.1376E+03, 2.1409E+03, 2.1442E+03, 2.1475E+03, 2.1507E+03,
     +  2.1540E+03, 2.1573E+03, 2.1606E+03, 2.1638E+03, 2.1671E+03,
     + 2.1704E+03, 2.1737E+03, 2.1770E+03, 2.1803E+03, 2.1836E+03,
     +  2.1869E+03, 2.1902E+03, 2.1935E+03, 2.1969E+03, 2.2002E+03,
     + 2.2035E+03, 2.2068E+03, 2.2101E+03, 2.2134E+03, 2.2168E+03,
     +  2.2201E+03, 2.2234E+03, 2.2267E+03, 2.2301E+03, 2.2334E+03,
     + 2.2368E+03, 2.2401E+03, 2.2434E+03, 2.2468E+03, 2.2501E+03,
     +  2.2535E+03, 2.2568E+03, 2.2602E+03, 2.2636E+03, 2.2669E+03,
     + 2.2703E+03, 2.2736E+03, 2.2770E+03, 2.2804E+03, 2.2838E+03,
     +  2.2871E+03, 2.2905E+03, 2.2939E+03, 2.2973E+03, 2.3007E+03,
     + 2.3040E+03, 2.3075E+03, 2.3108E+03, 2.3142E+03, 2.3176E+03,
     +  2.3210E+03, 2.3244E+03, 2.3278E+03, 2.3312E+03, 2.3347E+03,
     + 2.3381E+03, 2.3415E+03, 2.3449E+03, 2.3483E+03, 2.3517E+03,
     +  2.3552E+03, 2.3586E+03, 2.3620E+03, 2.3654E+03, 2.3689E+03,
     + 2.3723E+03, 2.3758E+03, 2.3792E+03, 2.3826E+03, 2.3861E+03,
     +  2.3895E+03, 2.3930E+03, 2.3965E+03, 2.3999E+03, 2.4034E+03,
     + 2.4068E+03, 2.4103E+03, 2.4138E+03, 2.4172E+03, 2.4207E+03,
     +  2.4242E+03, 2.4276E+03, 2.4311E+03, 2.4346E+03, 2.4381E+03,
     + 2.4416E+03, 2.4451E+03, 2.4486E+03, 2.4520E+03, 2.4555E+03,
     +  2.4590E+03, 2.4625E+03, 2.4660E+03, 2.4695E+03, 2.4731E+03,
     + 2.4766E+03, 2.4801E+03, 2.4836E+03, 2.4871E+03, 2.4906E+03,
     +  2.4941E+03, 2.4977E+03, 2.5012E+03, 2.5047E+03, 2.5083E+03,
     + 2.5118E+03, 2.5153E+03, 2.5189E+03, 2.5224E+03, 2.5259E+03,
     +  2.5295E+03, 2.5331E+03, 2.5366E+03, 2.5402E+03, 2.5437E+03,
     + 2.5473E+03, 2.5508E+03, 2.5544E+03, 2.5580E+03, 2.5615E+03,
     +  2.5651E+03, 2.5687E+03, 2.5723E+03, 2.5758E+03, 2.5794E+03,
     + 2.5830E+03, 2.5866E+03, 2.5902E+03, 2.5938E+03, 2.5973E+03,
     +  2.6010E+03, 2.6046E+03, 2.6082E+03, 2.6118E+03, 2.6154E+03,
     + 2.6190E+03, 2.6226E+03, 2.6262E+03, 2.6298E+03, 2.6334E+03,
     +  2.6370E+03, 2.6407E+03, 2.6443E+03, 2.6479E+03, 2.6516E+03,
     + 2.6552E+03, 2.6588E+03, 2.6625E+03, 2.6661E+03, 2.6697E+03,
     +  2.6734E+03, 2.6770E+03, 2.6807E+03, 2.6843E+03, 2.6880E+03,
     + 2.6917E+03, 2.6953E+03, 2.6990E+03, 2.7026E+03, 2.7063E+03,
     +  2.7100E+03, 2.7137E+03, 2.7173E+03, 2.7210E+03, 2.7247E+03,
     + 2.7284E+03, 2.7321E+03, 2.7357E+03, 2.7394E+03, 2.7431E+03,
     +  2.7468E+03, 2.7505E+03, 2.7542E+03, 2.7579E+03, 2.7616E+03,
     + 2.7653E+03, 2.7690E+03, 2.7728E+03, 2.7765E+03, 2.7802E+03,
     +  2.7839E+03, 2.7876E+03, 2.7913E+03, 2.7951E+03, 2.7988E+03,
     + 2.8025E+03, 2.8063E+03, 2.8100E+03, 2.8137E+03, 2.8175E+03,
     +  2.8212E+03, 2.8250E+03, 2.8287E+03, 2.8325E+03, 2.8362E+03,
     + 2.8400E+03, 2.8437E+03, 2.8475E+03, 2.8513E+03, 2.8550E+03,
     +  2.8588E+03, 2.8626E+03, 2.8663E+03, 2.8701E+03, 2.8739E+03,
     + 2.8777E+03, 2.8815E+03, 2.8852E+03, 2.8890E+03, 2.8928E+03,
     +  2.8966E+03, 2.9004E+03, 2.9042E+03, 2.9080E+03, 2.9118E+03,
     + 2.9156E+03, 2.9194E+03, 2.9232E+03, 2.9271E+03, 2.9309E+03,
     +  2.9347E+03, 2.9385E+03, 2.9424E+03, 2.9462E+03, 2.9500E+03,
     + 2.9538E+03, 2.9577E+03, 2.9615E+03, 2.9653E+03, 2.9692E+03,
     +  2.9730E+03, 2.9769E+03, 2.9807E+03, 2.9846E+03, 2.9884E+03,
     + 2.9923E+03, 2.9961E+03, 3.0000E+03, 3.0039E+03, 3.0077E+03,
     +  3.0116E+03, 3.0154E+03, 3.0193E+03, 3.0232E+03, 3.0271E+03,
     + 3.0310E+03, 3.0349E+03, 3.0387E+03, 3.0426E+03, 3.0465E+03,
     +  3.0504E+03, 3.0543E+03, 3.0582E+03, 3.0621E+03, 3.0660E+03,
     + 3.0699E+03, 3.0738E+03, 3.0777E+03, 3.0816E+03, 3.0855E+03,
     +  3.0895E+03, 3.0934E+03, 3.0973E+03, 3.1013E+03, 3.1052E+03,
     + 3.1091E+03, 3.1130E+03, 3.1170E+03, 3.1209E+03, 3.1248E+03,
     +  3.1288E+03, 3.1327E+03, 3.1367E+03, 3.1406E+03, 3.1446E+03,
     + 3.1485E+03, 3.1525E+03, 3.1565E+03, 3.1604E+03, 3.1644E+03,
     +  3.1684E+03, 3.1723E+03, 3.1763E+03, 3.1803E+03, 3.1842E+03,
     + 3.1882E+03, 3.1922E+03, 3.1962E+03, 3.2002E+03, 3.2042E+03,
     +  3.2082E+03, 3.2122E+03, 3.2162E+03, 3.2202E+03, 3.2242E+03,
     + 3.2282E+03, 3.2322E+03, 3.2362E+03, 3.2402E+03, 3.2442E+03,
     +  3.2482E+03, 3.2522E+03, 3.2563E+03, 3.2603E+03, 3.2643E+03,
     + 3.2684E+03, 3.2724E+03, 3.2764E+03, 3.2805E+03, 3.2845E+03,
     +  3.2885E+03, 3.2926E+03, 3.2966E+03, 3.3007E+03, 3.3047E+03,
     + 3.3088E+03, 3.3128E+03, 3.3169E+03, 3.3210E+03, 3.3250E+03,
     +  3.3291E+03, 3.3332E+03, 3.3372E+03, 3.3413E+03, 3.3454E+03,
     + 3.3495E+03, 3.3536E+03, 3.3576E+03, 3.3617E+03, 3.3658E+03,
     +  3.3699E+03, 3.3740E+03, 3.3781E+03, 3.3822E+03, 3.3863E+03,
     + 3.3904E+03, 3.3945E+03, 3.3986E+03, 3.4027E+03, 3.4069E+03,
     +  3.4110E+03, 3.4151E+03, 3.4192E+03, 3.4234E+03, 3.4275E+03,
     + 3.4316E+03, 3.4357E+03, 3.4399E+03, 3.4440E+03, 3.4481E+03,
     +  3.4523E+03, 3.4564E+03, 3.4606E+03, 3.4647E+03, 3.4689E+03,
     + 3.4730E+03, 3.4772E+03, 3.4813E+03, 3.4855E+03, 3.4897E+03,
     +  3.4938E+03, 3.4980E+03, 3.5022E+03, 3.5064E+03, 3.5106E+03,
     + 3.5147E+03, 3.5189E+03, 3.5231E+03, 3.5273E+03, 3.5315E+03,
     +  3.5356E+03, 3.5398E+03, 3.5440E+03, 3.5482E+03, 3.5525E+03,
     + 3.5567E+03, 3.5609E+03, 3.5651E+03, 3.5693E+03, 3.5735E+03,
     +  3.5777E+03, 3.5819E+03, 3.5862E+03, 3.5904E+03, 3.5946E+03,
     + 3.5988E+03, 3.6031E+03, 3.6073E+03, 3.6116E+03, 3.6158E+03,
     +  3.6200E+03, 3.6243E+03, 3.6285E+03, 3.6328E+03, 3.6370E+03,
     + 3.6413E+03, 3.6455E+03, 3.6498E+03, 3.6541E+03, 3.6583E+03,
     +  3.6626E+03, 3.6669E+03, 3.6711E+03, 3.6754E+03, 3.6797E+03,
     + 3.6840E+03, 3.6883E+03, 3.6925E+03, 3.6968E+03, 3.7011E+03,
     +  3.7054E+03, 3.7097E+03, 3.7140E+03, 3.7183E+03, 3.7226E+03,
     + 3.7269E+03, 3.7312E+03, 3.7355E+03, 3.7399E+03, 3.7442E+03,
     +  3.7485E+03, 3.7528E+03, 3.7571E+03, 3.7615E+03, 3.7658E+03,
     + 3.7701E+03, 3.7744E+03, 3.7788E+03, 3.7831E+03, 3.7874E+03,
     +  3.7918E+03, 3.7961E+03, 3.8005E+03, 3.8048E+03, 3.8092E+03,
     + 3.8136E+03, 3.8179E+03, 3.8223E+03, 3.8266E+03, 3.8310E+03,
     +  3.8354E+03, 3.8397E+03, 3.8441E+03, 3.8485E+03, 3.8529E+03,
     + 3.8572E+03, 3.8616E+03, 3.8660E+03, 3.8704E+03, 3.8748E+03,
     +  3.8792E+03, 3.8836E+03, 3.8880E+03, 3.8924E+03, 3.8968E+03,
     + 3.9012E+03, 3.9056E+03, 3.9100E+03, 3.9144E+03, 3.9188E+03,
     +  3.9233E+03, 3.9277E+03, 3.9321E+03, 3.9365E+03, 3.9409E+03,
     + 3.9454E+03, 3.9498E+03, 3.9542E+03, 3.9587E+03, 3.9631E+03,
     +  3.9676E+03, 3.9720E+03, 3.9765E+03, 3.9809E+03, 3.9854E+03,
     + 3.9898E+03, 3.9943E+03, 3.9988E+03, 4.0032E+03, 4.0077E+03,
     +  4.0121E+03, 4.0166E+03, 4.0211E+03, 4.0256E+03, 4.0300E+03,
     + 4.0345E+03, 4.0390E+03, 4.0435E+03, 4.0480E+03, 4.0525E+03,
     +  4.0570E+03, 4.0615E+03, 4.0660E+03, 4.0705E+03, 4.0750E+03,
     + 4.0795E+03, 4.0840E+03, 4.0885E+03, 4.0930E+03, 4.0975E+03,
     +  4.1020E+03, 4.1066E+03, 4.1111E+03, 4.1156E+03, 4.1201E+03,
     + 4.1247E+03, 4.1292E+03, 4.1337E+03, 4.1383E+03, 4.1428E+03,
     +  4.1474E+03, 4.1519E+03, 4.1565E+03, 4.1610E+03, 4.1656E+03,
     + 4.1701E+03, 4.1747E+03, 4.1792E+03, 4.1838E+03, 4.1884E+03,
     +  4.1929E+03, 4.1975E+03, 4.2021E+03, 4.2067E+03, 4.2112E+03,
     + 4.2158E+03, 4.2204E+03, 4.2250E+03, 4.2296E+03, 4.2342E+03,
     +  4.2388E+03, 4.2434E+03, 4.2480E+03, 4.2526E+03, 4.2572E+03,
     + 4.2618E+03, 4.2664E+03, 4.2710E+03, 4.2756E+03, 4.2803E+03,
     +  4.2849E+03, 4.2895E+03, 4.2941E+03, 4.2987E+03, 4.3034E+03,
     + 4.3080E+03, 4.3126E+03, 4.3173E+03, 4.3219E+03, 4.3266E+03,
     +  4.3312E+03, 4.3358E+03, 4.3405E+03, 4.3451E+03, 4.3498E+03,
     + 4.3545E+03, 4.3591E+03, 4.3638E+03, 4.3685E+03, 4.3731E+03,
     +  4.3778E+03, 4.3824E+03, 4.3871E+03, 4.3918E+03, 4.3965E+03,
     + 4.4012E+03, 4.4059E+03, 4.4105E+03, 4.4152E+03, 4.4199E+03,
     +  4.4246E+03, 4.4293E+03, 4.4340E+03, 4.4387E+03, 4.4434E+03,
     + 4.4481E+03, 4.4528E+03, 4.4576E+03, 4.4623E+03, 4.4670E+03,
     +  4.4717E+03, 4.4764E+03, 4.4812E+03, 4.4859E+03, 4.4906E+03,
     + 4.4954E+03, 4.5001E+03, 4.5049E+03, 4.5096E+03, 4.5143E+03,
     +  4.5191E+03, 4.5238E+03, 4.5286E+03, 4.5333E+03, 4.5381E+03,
     + 4.5428E+03, 4.5476E+03, 4.5523E+03, 4.5571E+03, 4.5619E+03,
     +  4.5667E+03, 4.5714E+03, 4.5762E+03, 4.5810E+03, 4.5858E+03,
     + 4.5905E+03, 4.5953E+03, 4.6001E+03, 4.6049E+03, 4.6097E+03,
     +  4.6145E+03, 4.6193E+03, 4.6241E+03, 4.6289E+03, 4.6337E+03,
     + 4.6385E+03, 4.6433E+03, 4.6481E+03, 4.6530E+03, 4.6578E+03,
     +  4.6626E+03, 4.6674E+03, 4.6723E+03, 4.6771E+03, 4.6819E+03,
     + 4.6868E+03, 4.6916E+03, 4.6964E+03, 4.7013E+03, 4.7061E+03,
     +  4.7110E+03, 4.7158E+03, 4.7207E+03, 4.7255E+03, 4.7304E+03,
     + 4.7352E+03, 4.7401E+03, 4.7450E+03, 4.7498E+03, 4.7547E+03,
     +  4.7596E+03, 4.7645E+03, 4.7693E+03, 4.7742E+03, 4.7791E+03,
     + 4.7840E+03, 4.7889E+03, 4.7937E+03, 4.7987E+03, 4.8035E+03,
     +  4.8084E+03, 4.8134E+03, 4.8182E+03, 4.8232E+03, 4.8280E+03,
     + 4.8330E+03, 4.8379E+03, 4.8428E+03, 4.8477E+03, 4.8526E+03,
     +  4.8575E+03, 4.8625E+03, 4.8674E+03, 4.8723E+03, 4.8773E+03,
     + 4.8822E+03, 4.8871E+03, 4.8921E+03, 4.8970E+03, 4.9020E+03,
     +  4.9069E+03, 4.9119E+03, 4.9168E+03, 4.9218E+03, 4.9267E+03,
     + 4.9317E+03, 4.9366E+03, 4.9416E+03, 4.9466E+03, 4.9516E+03,
     +  4.9566E+03, 4.9615E+03, 4.9665E+03, 4.9715E+03, 4.9765E+03,
     + 4.9814E+03, 4.9864E+03, 4.9914E+03, 4.9964E+03, 5.0014E+03,
     +  5.0064E+03, 5.0114E+03, 5.0164E+03, 5.0214E+03, 5.0264E+03,
     + 5.0315E+03, 5.0365E+03, 5.0415E+03, 5.0465E+03, 5.0515E+03,
     +  5.0565E+03, 5.0616E+03, 5.0666E+03, 5.0716E+03, 5.0766E+03,
     + 5.0817E+03, 5.0867E+03, 5.0918E+03, 5.0968E+03, 5.1019E+03,
     +  5.1069E+03, 5.1120E+03, 5.1170E+03, 5.1221E+03, 5.1271E+03,
     + 5.1322E+03, 5.1372E+03, 5.1423E+03, 5.1474E+03, 5.1525E+03,
     +  5.1576E+03, 5.1626E+03, 5.1677E+03, 5.1728E+03, 5.1779E+03,
     + 5.1830E+03, 5.1880E+03, 5.1931E+03, 5.1982E+03, 5.2033E+03,
     +  5.2084E+03, 5.2135E+03, 5.2186E+03, 5.2237E+03, 5.2288E+03,
     + 5.2340E+03, 5.2391E+03, 5.2442E+03, 5.2493E+03, 5.2544E+03,
     +  5.2596E+03, 5.2647E+03, 5.2698E+03, 5.2750E+03, 5.2801E+03,
     + 5.2852E+03, 5.2904E+03, 5.2955E+03, 5.3007E+03, 5.3058E+03,
     +  5.3110E+03, 5.3161E+03, 5.3213E+03, 5.3264E+03, 5.3316E+03,
     + 5.3367E+03, 5.3419E+03, 5.3471E+03, 5.3523E+03, 5.3574E+03,
     +  5.3626E+03, 5.3678E+03, 5.3730E+03, 5.3782E+03, 5.3833E+03,
     + 5.3885E+03, 5.3937E+03, 5.3989E+03, 5.4041E+03, 5.4093E+03,
     +  5.4145E+03, 5.4197E+03, 5.4249E+03, 5.4301E+03, 5.4353E+03,
     + 5.4405E+03, 5.4458E+03, 5.4510E+03, 5.4562E+03, 5.4614E+03,
     +  5.4666E+03, 5.4719E+03, 5.4771E+03, 5.4824E+03, 5.4876E+03,
     + 5.4928E+03, 5.4981E+03, 5.5033E+03, 5.5086E+03, 5.5138E+03,
     +  5.5191E+03, 5.5243E+03, 5.5296E+03, 5.5349E+03, 5.5401E+03,
     + 5.5454E+03, 5.5506E+03, 5.5559E+03, 5.5612E+03, 5.5665E+03,
     +  5.5717E+03, 5.5770E+03, 5.5823E+03, 5.5876E+03, 5.5928E+03,
     + 5.5982E+03, 5.6035E+03, 5.6087E+03, 5.6140E+03, 5.6193E+03,
     +  5.6246E+03, 5.6300E+03, 5.6353E+03, 5.6406E+03, 5.6459E+03,
     + 5.6512E+03, 5.6565E+03, 5.6618E+03, 5.6672E+03, 5.6725E+03,
     +  5.6778E+03, 5.6831E+03, 5.6885E+03, 5.6938E+03, 5.6992E+03,
     + 5.7045E+03, 5.7098E+03, 5.7152E+03, 5.7205E+03, 5.7259E+03,
     +  5.7312E+03, 5.7366E+03, 5.7419E+03, 5.7473E+03, 5.7527E+03,
     + 5.7581E+03, 5.7634E+03, 5.7688E+03, 5.7742E+03, 5.7795E+03,
     +  5.7849E+03, 5.7903E+03, 5.7957E+03, 5.8011E+03, 5.8064E+03,
     + 5.8118E+03, 5.8172E+03, 5.8226E+03, 5.8280E+03, 5.8334E+03,
     +  5.8388E+03, 5.8442E+03, 5.8497E+03, 5.8550E+03, 5.8605E+03,
     + 5.8659E+03, 5.8713E+03, 5.8767E+03, 5.8822E+03, 5.8876E+03,
     +  5.8930E+03, 5.8984E+03, 5.9039E+03, 5.9093E+03, 5.9148E+03,
     + 5.9202E+03, 5.9257E+03, 5.9311E+03, 5.9365E+03, 5.9420E+03,
     +  5.9475E+03, 5.9529E+03, 5.9584E+03, 5.9638E+03, 5.9693E+03,
     + 5.9748E+03, 5.9803E+03, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_HOBr  (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 2),QofT( 2,NT),Q(NT)
      data (Tmx(I),I=1,2)/3500.,3500./
      data xgj/ 8.,8./
c...     HOBr
c...        --       169
      data (QofT( 1,J),J=1,2408)/ 5.7851E+00, 1.5995E+01, 2.8730E+01,
     +  4.3671E+01, 6.0548E+01, 7.9163E+01, 9.9369E+01, 1.2105E+02,
     + 1.4411E+02, 1.6848E+02, 1.9408E+02, 2.2087E+02, 2.4878E+02,
     +  2.7779E+02, 3.0784E+02, 3.3890E+02, 3.7094E+02, 4.0394E+02,
     + 4.3786E+02, 4.7268E+02, 5.4493E+02, 6.2054E+02, 6.9935E+02,
     +  7.8125E+02, 8.6613E+02, 9.5387E+02, 1.0444E+03, 1.1376E+03,
     + 1.2335E+03, 1.3319E+03, 1.4328E+03, 1.5362E+03, 1.6419E+03,
     +  1.7499E+03, 1.8602E+03, 1.9728E+03, 2.0875E+03, 2.2044E+03,
     + 2.3234E+03, 2.4445E+03, 2.5676E+03, 2.6927E+03, 2.8198E+03,
     +  2.9488E+03, 3.0798E+03, 3.2126E+03, 3.3474E+03, 3.4839E+03,
     + 3.6223E+03, 3.7624E+03, 3.9044E+03, 4.0481E+03, 4.1935E+03,
     +  4.3407E+03, 4.4896E+03, 4.6401E+03, 4.7923E+03, 4.9461E+03,
     + 5.1016E+03, 5.2587E+03, 5.4175E+03, 5.5778E+03, 5.7397E+03,
     +  5.9032E+03, 6.0683E+03, 6.2349E+03, 6.4031E+03, 6.5728E+03,
     + 6.7441E+03, 6.9169E+03, 7.0912E+03, 7.2670E+03, 7.4443E+03,
     +  7.6232E+03, 7.8036E+03, 7.9855E+03, 8.1688E+03, 8.3536E+03,
     + 8.5400E+03, 8.7279E+03, 8.9172E+03, 9.1080E+03, 9.3003E+03,
     +  9.4942E+03, 9.6895E+03, 9.8863E+03, 1.0085E+04, 1.0284E+04,
     + 1.0486E+04, 1.0689E+04, 1.0893E+04, 1.1099E+04, 1.1306E+04,
     +  1.1515E+04, 1.1725E+04, 1.1937E+04, 1.2151E+04, 1.2366E+04,
     + 1.2582E+04, 1.2800E+04, 1.3020E+04, 1.3241E+04, 1.3464E+04,
     +  1.3688E+04, 1.3914E+04, 1.4141E+04, 1.4371E+04, 1.4601E+04,
     + 1.4833E+04, 1.5067E+04, 1.5303E+04, 1.5540E+04, 1.5779E+04,
     +  1.6019E+04, 1.6261E+04, 1.6504E+04, 1.6750E+04, 1.6997E+04,
     + 1.7245E+04, 1.7496E+04, 1.7747E+04, 1.8001E+04, 1.8257E+04,
     +  1.8514E+04, 1.8772E+04, 1.9033E+04, 1.9295E+04, 1.9559E+04,
     + 1.9825E+04, 2.0093E+04, 2.0362E+04, 2.0633E+04, 2.0906E+04,
     +  2.1181E+04, 2.1457E+04, 2.1736E+04, 2.2016E+04, 2.2298E+04,
     + 2.2582E+04, 2.2867E+04, 2.3155E+04, 2.3445E+04, 2.3736E+04,
     +  2.4029E+04, 2.4324E+04, 2.4621E+04, 2.4920E+04, 2.5221E+04,
     + 2.5524E+04, 2.5829E+04, 2.6136E+04, 2.6445E+04, 2.6756E+04,
     +  2.7068E+04, 2.7383E+04, 2.7700E+04, 2.8019E+04, 2.8339E+04,
     + 2.8662E+04, 2.8987E+04, 2.9314E+04, 2.9643E+04, 2.9975E+04,
     +  3.0308E+04, 3.0643E+04, 3.0981E+04, 3.1320E+04, 3.1662E+04,
     + 3.2006E+04, 3.2352E+04, 3.2700E+04, 3.3051E+04, 3.3404E+04,
     +  3.3758E+04, 3.4115E+04, 3.4475E+04, 3.4836E+04, 3.5200E+04,
     + 3.5566E+04, 3.5935E+04, 3.6305E+04, 3.6678E+04, 3.7053E+04,
     +  3.7430E+04, 3.7810E+04, 3.8192E+04, 3.8577E+04, 3.8964E+04,
     + 3.9353E+04, 3.9745E+04, 4.0138E+04, 4.0535E+04, 4.0934E+04,
     +  4.1335E+04, 4.1738E+04, 4.2144E+04, 4.2553E+04, 4.2964E+04,
     + 4.3377E+04, 4.3793E+04, 4.4212E+04, 4.4633E+04, 4.5056E+04,
     +  4.5482E+04, 4.5910E+04, 4.6342E+04, 4.6775E+04, 4.7212E+04,
     + 4.7650E+04, 4.8092E+04, 4.8536E+04, 4.8983E+04, 4.9432E+04,
     +  4.9884E+04, 5.0338E+04, 5.0796E+04, 5.1255E+04, 5.1718E+04,
     + 5.2183E+04, 5.2652E+04, 5.3122E+04, 5.3596E+04, 5.4072E+04,
     +  5.4551E+04, 5.5032E+04, 5.5517E+04, 5.6004E+04, 5.6495E+04,
     + 5.6988E+04, 5.7483E+04, 5.7982E+04, 5.8483E+04, 5.8987E+04,
     +  5.9495E+04, 6.0005E+04, 6.0518E+04, 6.1033E+04, 6.1552E+04,
     + 6.2074E+04, 6.2599E+04, 6.3127E+04, 6.3657E+04, 6.4191E+04,
     +  6.4727E+04, 6.5267E+04, 6.5809E+04, 6.6355E+04, 6.6904E+04,
     + 6.7455E+04, 6.8010E+04, 6.8567E+04, 6.9129E+04, 6.9692E+04,
     +  7.0260E+04, 7.0830E+04, 7.1403E+04, 7.1980E+04, 7.2559E+04,
     + 7.3142E+04, 7.3728E+04, 7.4317E+04, 7.4910E+04, 7.5505E+04,
     +  7.6104E+04, 7.6706E+04, 7.7311E+04, 7.7920E+04, 7.8532E+04,
     + 7.9147E+04, 7.9765E+04, 8.0387E+04, 8.1012E+04, 8.1640E+04,
     +  8.2272E+04, 8.2907E+04, 8.3546E+04, 8.4187E+04, 8.4833E+04,
     + 8.5481E+04, 8.6133E+04, 8.6789E+04, 8.7448E+04, 8.8110E+04,
     +  8.8776E+04, 8.9445E+04, 9.0118E+04, 9.0795E+04, 9.1475E+04,
     + 9.2158E+04, 9.2845E+04, 9.3535E+04, 9.4229E+04, 9.4927E+04,
     +  9.5628E+04, 9.6333E+04, 9.7042E+04, 9.7753E+04, 9.8469E+04,
     + 9.9189E+04, 9.9912E+04, 1.0064E+05, 1.0137E+05, 1.0210E+05,
     +  1.0284E+05, 1.0358E+05, 1.0433E+05, 1.0508E+05, 1.0583E+05,
     + 1.0659E+05, 1.0735E+05, 1.0811E+05, 1.0888E+05, 1.0965E+05,
     +  1.1043E+05, 1.1121E+05, 1.1199E+05, 1.1278E+05, 1.1357E+05,
     + 1.1436E+05, 1.1516E+05, 1.1597E+05, 1.1677E+05, 1.1758E+05,
     +  1.1840E+05, 1.1922E+05, 1.2004E+05, 1.2087E+05, 1.2170E+05,
     + 1.2253E+05, 1.2337E+05, 1.2422E+05, 1.2506E+05, 1.2591E+05,
     +  1.2677E+05, 1.2763E+05, 1.2849E+05, 1.2936E+05, 1.3023E+05,
     + 1.3111E+05, 1.3199E+05, 1.3287E+05, 1.3376E+05, 1.3465E+05,
     +  1.3555E+05, 1.3645E+05, 1.3736E+05, 1.3827E+05, 1.3918E+05,
     + 1.4010E+05, 1.4102E+05, 1.4195E+05, 1.4288E+05, 1.4382E+05,
     +  1.4476E+05, 1.4570E+05, 1.4665E+05, 1.4761E+05, 1.4857E+05,
     + 1.4953E+05, 1.5050E+05, 1.5147E+05, 1.5244E+05, 1.5342E+05,
     +  1.5441E+05, 1.5540E+05, 1.5639E+05, 1.5739E+05, 1.5839E+05,
     + 1.5940E+05, 1.6041E+05, 1.6143E+05, 1.6245E+05, 1.6348E+05,
     +  1.6451E+05, 1.6554E+05, 1.6658E+05, 1.6763E+05, 1.6868E+05,
     + 1.6973E+05, 1.7079E+05, 1.7186E+05, 1.7292E+05, 1.7400E+05,
     +  1.7508E+05, 1.7616E+05, 1.7725E+05, 1.7834E+05, 1.7944E+05,
     + 1.8054E+05, 1.8165E+05, 1.8276E+05, 1.8388E+05, 1.8500E+05,
     +  1.8613E+05, 1.8726E+05, 1.8839E+05, 1.8954E+05, 1.9068E+05,
     + 1.9184E+05, 1.9299E+05, 1.9415E+05, 1.9532E+05, 1.9649E+05,
     +  1.9767E+05, 1.9885E+05, 2.0004E+05, 2.0123E+05, 2.0243E+05,
     + 2.0363E+05, 2.0484E+05, 2.0606E+05, 2.0727E+05, 2.0850E+05,
     +  2.0973E+05, 2.1096E+05, 2.1220E+05, 2.1345E+05, 2.1470E+05,
     + 2.1595E+05, 2.1721E+05, 2.1848E+05, 2.1975E+05, 2.2103E+05,
     +  2.2231E+05, 2.2360E+05, 2.2489E+05, 2.2619E+05, 2.2749E+05,
     + 2.2880E+05, 2.3012E+05, 2.3144E+05, 2.3277E+05, 2.3410E+05,
     +  2.3544E+05, 2.3678E+05, 2.3813E+05, 2.3948E+05, 2.4084E+05,
     + 2.4221E+05, 2.4358E+05, 2.4495E+05, 2.4634E+05, 2.4772E+05,
     +  2.4912E+05, 2.5052E+05, 2.5192E+05, 2.5333E+05, 2.5475E+05,
     + 2.5617E+05, 2.5760E+05, 2.5903E+05, 2.6047E+05, 2.6192E+05,
     +  2.6337E+05, 2.6483E+05, 2.6629E+05, 2.6776E+05, 2.6924E+05,
     + 2.7072E+05, 2.7221E+05, 2.7370E+05, 2.7520E+05, 2.7671E+05,
     +  2.7822E+05, 2.7974E+05, 2.8126E+05, 2.8279E+05, 2.8432E+05,
     + 2.8587E+05, 2.8741E+05, 2.8897E+05, 2.9053E+05, 2.9210E+05,
     +  2.9367E+05, 2.9525E+05, 2.9683E+05, 2.9842E+05, 3.0002E+05,
     + 3.0163E+05, 3.0324E+05, 3.0485E+05, 3.0648E+05, 3.0811E+05,
     +  3.0974E+05, 3.1139E+05, 3.1303E+05, 3.1469E+05, 3.1635E+05,
     + 3.1802E+05, 3.1969E+05, 3.2138E+05, 3.2306E+05, 3.2476E+05,
     +  3.2646E+05, 3.2817E+05, 3.2988E+05, 3.3160E+05, 3.3333E+05,
     + 3.3506E+05, 3.3680E+05, 3.3855E+05, 3.4030E+05, 3.4206E+05,
     +  3.4383E+05, 3.4560E+05, 3.4738E+05, 3.4917E+05, 3.5097E+05,
     + 3.5277E+05, 3.5458E+05, 3.5639E+05, 3.5821E+05, 3.6004E+05,
     +  3.6188E+05, 3.6372E+05, 3.6557E+05, 3.6743E+05, 3.6929E+05,
     + 3.7116E+05, 3.7304E+05, 3.7492E+05, 3.7682E+05, 3.7871E+05,
     +  3.8062E+05, 3.8253E+05, 3.8445E+05, 3.8638E+05, 3.8831E+05,
     + 3.9026E+05, 3.9221E+05, 3.9416E+05, 3.9612E+05, 3.9810E+05,
     +  4.0007E+05, 4.0206E+05, 4.0405E+05, 4.0605E+05, 4.0806E+05,
     + 4.1007E+05, 4.1210E+05, 4.1413E+05, 4.1616E+05, 4.1821E+05,
     +  4.2026E+05, 4.2232E+05, 4.2438E+05, 4.2646E+05, 4.2854E+05,
     + 4.3063E+05, 4.3273E+05, 4.3483E+05, 4.3694E+05, 4.3906E+05,
     +  4.4119E+05, 4.4333E+05, 4.4547E+05, 4.4762E+05, 4.4978E+05,
     + 4.5195E+05, 4.5412E+05, 4.5630E+05, 4.5849E+05, 4.6069E+05,
     +  4.6290E+05, 4.6511E+05, 4.6733E+05, 4.6956E+05, 4.7180E+05,
     + 4.7404E+05, 4.7630E+05, 4.7856E+05, 4.8083E+05, 4.8310E+05,
     +  4.8539E+05, 4.8768E+05, 4.8998E+05, 4.9229E+05, 4.9461E+05,
     + 4.9694E+05, 4.9927E+05, 5.0161E+05, 5.0396E+05, 5.0632E+05,
     +  5.0869E+05, 5.1106E+05, 5.1345E+05, 5.1584E+05, 5.1824E+05,
     + 5.2065E+05, 5.2307E+05, 5.2549E+05, 5.2793E+05, 5.3037E+05,
     +  5.3282E+05, 5.3528E+05, 5.3775E+05, 5.4022E+05, 5.4271E+05,
     + 5.4520E+05, 5.4770E+05, 5.5022E+05, 5.5273E+05, 5.5526E+05,
     +  5.5780E+05, 5.6035E+05, 5.6290E+05, 5.6546E+05, 5.6804E+05,
     + 5.7062E+05, 5.7320E+05, 5.7580E+05, 5.7841E+05, 5.8103E+05,
     +  5.8365E+05, 5.8628E+05, 5.8893E+05, 5.9158E+05, 5.9424E+05,
     + 5.9691E+05, 5.9959E+05, 6.0228E+05, 6.0497E+05, 6.0768E+05,
     +  6.1039E+05, 6.1312E+05, 6.1585E+05, 6.1859E+05, 6.2135E+05,
     + 6.2411E+05, 6.2688E+05, 6.2966E+05, 6.3245E+05, 6.3524E+05,
     +  6.3805E+05, 6.4087E+05, 6.4369E+05, 6.4653E+05, 6.4938E+05,
     + 6.5223E+05, 6.5509E+05, 6.5797E+05, 6.6085E+05, 6.6374E+05,
     +  6.6664E+05, 6.6955E+05, 6.7248E+05, 6.7541E+05, 6.7835E+05,
     + 6.8130E+05, 6.8426E+05, 6.8723E+05, 6.9021E+05, 6.9320E+05,
     +  6.9619E+05, 6.9920E+05, 7.0222E+05, 7.0525E+05, 7.0829E+05,
     + 7.1134E+05, 7.1439E+05, 7.1746E+05, 7.2054E+05, 7.2363E+05,
     +  7.2673E+05, 7.2984E+05, 7.3295E+05, 7.3608E+05, 7.3922E+05,
     + 7.4237E+05, 7.4553E+05, 7.4869E+05, 7.5187E+05, 7.5506E+05,
     +  7.5826E+05, 7.6147E+05, 7.6469E+05, 7.6792E+05, 7.7116E+05,
     + 7.7441E+05, 7.7768E+05, 7.8095E+05, 7.8423E+05, 7.8752E+05,
     +  7.9083E+05, 7.9414E+05, 7.9746E+05, 8.0080E+05, 8.0414E+05,
     + 8.0750E+05, 8.1087E+05, 8.1424E+05, 8.1763E+05, 8.2103E+05,
     +  8.2444E+05, 8.2786E+05, 8.3129E+05, 8.3473E+05, 8.3819E+05,
     + 8.4165E+05, 8.4512E+05, 8.4861E+05, 8.5210E+05, 8.5561E+05,
     +  8.5913E+05, 8.6266E+05, 8.6620E+05, 8.6975E+05, 8.7331E+05,
     + 8.7688E+05, 8.8047E+05, 8.8406E+05, 8.8767E+05, 8.9129E+05,
     +  8.9492E+05, 8.9856E+05, 9.0221E+05, 9.0587E+05, 9.0955E+05,
     + 9.1323E+05, 9.1693E+05, 9.2064E+05, 9.2436E+05, 9.2809E+05,
     +  9.3183E+05, 9.3558E+05, 9.3935E+05, 9.4313E+05, 9.4691E+05,
     + 9.5071E+05, 9.5453E+05, 9.5835E+05, 9.6219E+05, 9.6603E+05,
     +  9.6989E+05, 9.7376E+05, 9.7764E+05, 9.8154E+05, 9.8545E+05,
     + 9.8936E+05, 9.9329E+05, 9.9723E+05, 1.0012E+06, 1.0052E+06,
     +  1.0091E+06, 1.0131E+06, 1.0171E+06, 1.0211E+06, 1.0252E+06,
     + 1.0292E+06, 1.0332E+06, 1.0373E+06, 1.0414E+06, 1.0455E+06,
     +  1.0496E+06, 1.0537E+06, 1.0578E+06, 1.0619E+06, 1.0661E+06,
     + 1.0702E+06, 1.0744E+06, 1.0786E+06, 1.0828E+06, 1.0870E+06,
     +  1.0912E+06, 1.0955E+06, 1.0997E+06, 1.1040E+06, 1.1082E+06,
     + 1.1125E+06, 1.1168E+06, 1.1211E+06, 1.1255E+06, 1.1298E+06,
     +  1.1341E+06, 1.1385E+06, 1.1429E+06, 1.1473E+06, 1.1517E+06,
     + 1.1561E+06, 1.1605E+06, 1.1650E+06, 1.1694E+06, 1.1739E+06,
     +  1.1783E+06, 1.1828E+06, 1.1873E+06, 1.1919E+06, 1.1964E+06,
     + 1.2009E+06, 1.2055E+06, 1.2101E+06, 1.2146E+06, 1.2192E+06,
     +  1.2239E+06, 1.2285E+06, 1.2331E+06, 1.2378E+06, 1.2424E+06,
     + 1.2471E+06, 1.2518E+06, 1.2565E+06, 1.2612E+06, 1.2659E+06,
     +  1.2707E+06, 1.2755E+06, 1.2802E+06, 1.2850E+06, 1.2898E+06,
     + 1.2946E+06, 1.2995E+06, 1.3043E+06, 1.3091E+06, 1.3140E+06,
     +  1.3189E+06, 1.3238E+06, 1.3287E+06, 1.3336E+06, 1.3386E+06,
     + 1.3435E+06, 1.3485E+06, 1.3535E+06, 1.3585E+06, 1.3635E+06,
     +  1.3685E+06, 1.3735E+06, 1.3786E+06, 1.3836E+06, 1.3887E+06,
     + 1.3938E+06, 1.3989E+06, 1.4041E+06, 1.4092E+06, 1.4143E+06,
     +  1.4195E+06, 1.4247E+06, 1.4299E+06, 1.4351E+06, 1.4403E+06,
     + 1.4456E+06, 1.4508E+06, 1.4561E+06, 1.4614E+06, 1.4667E+06,
     +  1.4720E+06, 1.4773E+06, 1.4826E+06, 1.4880E+06, 1.4934E+06,
     + 1.4988E+06, 1.5042E+06, 1.5096E+06, 1.5150E+06, 1.5205E+06,
     +  1.5259E+06, 1.5314E+06, 1.5369E+06, 1.5424E+06, 1.5479E+06,
     + 1.5535E+06, 1.5590E+06, 1.5646E+06, 1.5702E+06, 1.5758E+06,
     +  1.5814E+06, 1.5870E+06, 1.5926E+06, 1.5983E+06, 1.6040E+06,
     + 1.6097E+06, 1.6154E+06, 1.6211E+06, 1.6268E+06, 1.6326E+06,
     +  1.6384E+06, 1.6442E+06, 1.6500E+06, 1.6558E+06, 1.6616E+06,
     + 1.6675E+06, 1.6733E+06, 1.6792E+06, 1.6851E+06, 1.6910E+06,
     +  1.6969E+06, 1.7029E+06, 1.7088E+06, 1.7148E+06, 1.7208E+06,
     + 1.7268E+06, 1.7328E+06, 1.7389E+06, 1.7449E+06, 1.7510E+06,
     +  1.7571E+06, 1.7632E+06, 1.7693E+06, 1.7755E+06, 1.7816E+06,
     + 1.7878E+06, 1.7940E+06, 1.8002E+06, 1.8064E+06, 1.8126E+06,
     +  1.8189E+06, 1.8252E+06, 1.8315E+06, 1.8378E+06, 1.8441E+06,
     + 1.8504E+06, 1.8568E+06, 1.8631E+06, 1.8695E+06, 1.8759E+06,
     +  1.8824E+06, 1.8888E+06, 1.8952E+06, 1.9017E+06, 1.9082E+06,
     + 1.9147E+06, 1.9212E+06, 1.9278E+06, 1.9343E+06, 1.9409E+06,
     +  1.9475E+06, 1.9541E+06, 1.9607E+06, 1.9674E+06, 1.9741E+06,
     + 1.9807E+06, 1.9874E+06, 1.9941E+06, 2.0009E+06, 2.0076E+06,
     +  2.0144E+06, 2.0212E+06, 2.0280E+06, 2.0348E+06, 2.0416E+06,
     + 2.0485E+06, 2.0554E+06, 2.0623E+06, 2.0692E+06, 2.0761E+06,
     +  2.0830E+06, 2.0900E+06, 2.0970E+06, 2.1040E+06, 2.1110E+06,
     + 2.1180E+06, 2.1251E+06, 2.1322E+06, 2.1393E+06, 2.1464E+06,
     +  2.1535E+06, 2.1606E+06, 2.1678E+06, 2.1750E+06, 2.1822E+06,
     + 2.1894E+06, 2.1966E+06, 2.2039E+06, 2.2112E+06, 2.2185E+06,
     +  2.2258E+06, 2.2331E+06, 2.2404E+06, 2.2478E+06, 2.2552E+06,
     + 2.2626E+06, 2.2700E+06, 2.2775E+06, 2.2849E+06, 2.2924E+06,
     +  2.2999E+06, 2.3074E+06, 2.3150E+06, 2.3225E+06, 2.3301E+06,
     + 2.3377E+06, 2.3453E+06, 2.3529E+06, 2.3606E+06, 2.3682E+06,
     +  2.3759E+06, 2.3836E+06, 2.3914E+06, 2.3991E+06, 2.4069E+06,
     + 2.4147E+06, 2.4225E+06, 2.4303E+06, 2.4382E+06, 2.4460E+06,
     +  2.4539E+06, 2.4618E+06, 2.4697E+06, 2.4777E+06, 2.4856E+06,
     + 2.4936E+06, 2.5016E+06, 2.5097E+06, 2.5177E+06, 2.5258E+06,
     +  2.5338E+06, 2.5419E+06, 2.5501E+06, 2.5582E+06, 2.5664E+06,
     + 2.5746E+06, 2.5828E+06, 2.5910E+06, 2.5992E+06, 2.6075E+06,
     +  2.6158E+06, 2.6241E+06, 2.6324E+06, 2.6408E+06, 2.6491E+06,
     + 2.6575E+06, 2.6659E+06, 2.6743E+06, 2.6828E+06, 2.6913E+06,
     +  2.6998E+06, 2.7083E+06, 2.7168E+06, 2.7254E+06, 2.7339E+06,
     + 2.7425E+06, 2.7511E+06, 2.7598E+06, 2.7684E+06, 2.7771E+06,
     +  2.7858E+06, 2.7945E+06, 2.8033E+06, 2.8120E+06, 2.8208E+06,
     + 2.8296E+06, 2.8385E+06, 2.8473E+06, 2.8562E+06, 2.8651E+06,
     +  2.8740E+06, 2.8829E+06, 2.8919E+06, 2.9009E+06, 2.9099E+06,
     + 2.9189E+06, 2.9279E+06, 2.9370E+06, 2.9461E+06, 2.9552E+06,
     +  2.9643E+06, 2.9735E+06, 2.9826E+06, 2.9918E+06, 3.0010E+06,
     + 3.0103E+06, 3.0196E+06, 3.0288E+06, 3.0381E+06, 3.0475E+06,
     +  3.0568E+06, 3.0662E+06, 3.0756E+06, 3.0850E+06, 3.0944E+06,
     + 3.1039E+06, 3.1134E+06, 3.1229E+06, 3.1324E+06, 3.1420E+06,
     +  3.1516E+06, 3.1612E+06, 3.1708E+06, 3.1804E+06, 3.1901E+06,
     + 3.1998E+06, 3.2095E+06, 3.2192E+06, 3.2290E+06, 3.2388E+06,
     +  3.2486E+06, 3.2584E+06, 3.2682E+06, 3.2781E+06, 3.2880E+06,
     + 3.2979E+06, 3.3079E+06, 3.3178E+06, 3.3278E+06, 3.3378E+06,
     +  3.3479E+06, 3.3579E+06, 3.3680E+06, 3.3781E+06, 3.3883E+06,
     + 3.3984E+06, 3.4086E+06, 3.4188E+06, 3.4290E+06, 3.4393E+06,
     +  3.4495E+06, 3.4598E+06, 3.4701E+06, 3.4805E+06, 3.4909E+06,
     + 3.5013E+06, 3.5117E+06, 3.5221E+06, 3.5326E+06, 3.5431E+06,
     +  3.5536E+06, 3.5641E+06, 3.5747E+06, 3.5853E+06, 3.5959E+06,
     + 3.6065E+06, 3.6172E+06, 3.6278E+06, 3.6385E+06, 3.6493E+06,
     +  3.6600E+06, 3.6708E+06, 3.6816E+06, 3.6925E+06, 3.7033E+06,
     + 3.7142E+06, 3.7251E+06, 3.7360E+06, 3.7470E+06, 3.7580E+06,
     +  3.7690E+06, 3.7800E+06, 3.7911E+06, 3.8021E+06, 3.8132E+06,
     + 3.8244E+06, 3.8355E+06, 3.8467E+06, 3.8579E+06, 3.8691E+06,
     +  3.8804E+06, 3.8917E+06, 3.9030E+06, 3.9143E+06, 3.9257E+06,
     + 3.9371E+06, 3.9485E+06, 3.9599E+06, 3.9714E+06, 3.9829E+06,
     +  3.9944E+06, 4.0059E+06, 4.0175E+06, 4.0291E+06, 4.0407E+06,
     + 4.0524E+06, 4.0640E+06, 4.0757E+06, 4.0874E+06, 4.0992E+06,
     +  4.1110E+06, 4.1228E+06, 4.1346E+06, 4.1465E+06, 4.1583E+06,
     + 4.1702E+06, 4.1822E+06, 4.1941E+06, 4.2061E+06, 4.2181E+06,
     +  4.2302E+06, 4.2422E+06, 4.2543E+06, 4.2665E+06, 4.2786E+06,
     + 4.2908E+06, 4.3030E+06, 4.3152E+06, 4.3275E+06, 4.3398E+06,
     +  4.3521E+06, 4.3644E+06, 4.3768E+06, 4.3892E+06, 4.4016E+06,
     + 4.4140E+06, 4.4265E+06, 4.4390E+06, 4.4515E+06, 4.4641E+06,
     +  4.4767E+06, 4.4893E+06, 4.5019E+06, 4.5146E+06, 4.5273E+06,
     + 4.5400E+06, 4.5528E+06, 4.5655E+06, 4.5783E+06, 4.5912E+06,
     +  4.6040E+06, 4.6169E+06, 4.6299E+06, 4.6428E+06, 4.6558E+06,
     + 4.6688E+06, 4.6818E+06, 4.6949E+06, 4.7080E+06, 4.7211E+06,
     +  4.7342E+06, 4.7474E+06, 4.7606E+06, 4.7738E+06, 4.7871E+06,
     + 4.8004E+06, 4.8137E+06, 4.8270E+06, 4.8404E+06, 4.8538E+06,
     +  4.8673E+06, 4.8807E+06, 4.8942E+06, 4.9077E+06, 4.9213E+06,
     + 4.9349E+06, 4.9485E+06, 4.9621E+06, 4.9758E+06, 4.9895E+06,
     +  5.0032E+06, 5.0169E+06, 5.0307E+06, 5.0445E+06, 5.0584E+06,
     + 5.0723E+06, 5.0862E+06, 5.1001E+06, 5.1140E+06, 5.1280E+06,
     +  5.1421E+06, 5.1561E+06, 5.1702E+06, 5.1843E+06, 5.1984E+06,
     + 5.2126E+06, 5.2268E+06, 5.2410E+06, 5.2553E+06, 5.2696E+06,
     +  5.2839E+06, 5.2983E+06, 5.3127E+06, 5.3271E+06, 5.3415E+06,
     + 5.3560E+06, 5.3705E+06, 5.3850E+06, 5.3996E+06, 5.4142E+06,
     +  5.4288E+06, 5.4435E+06, 5.4582E+06, 5.4729E+06, 5.4876E+06,
     + 5.5024E+06, 5.5172E+06, 5.5321E+06, 5.5470E+06, 5.5619E+06,
     +  5.5768E+06, 5.5918E+06, 5.6068E+06, 5.6218E+06, 5.6369E+06,
     + 5.6520E+06, 5.6671E+06, 5.6822E+06, 5.6974E+06, 5.7127E+06,
     +  5.7279E+06, 5.7432E+06, 5.7585E+06, 5.7739E+06, 5.7892E+06,
     + 5.8047E+06, 5.8201E+06, 5.8356E+06, 5.8511E+06, 5.8666E+06,
     +  5.8822E+06, 5.8978E+06, 5.9134E+06, 5.9291E+06, 5.9448E+06,
     + 5.9606E+06, 5.9763E+06, 5.9921E+06, 6.0080E+06, 6.0238E+06,
     +  6.0397E+06, 6.0557E+06, 6.0716E+06, 6.0876E+06, 6.1036E+06,
     + 6.1197E+06, 6.1358E+06, 6.1519E+06, 6.1681E+06, 6.1843E+06,
     +  6.2005E+06, 6.2168E+06, 6.2331E+06, 6.2494E+06, 6.2658E+06,
     + 6.2822E+06, 6.2986E+06, 6.3151E+06, 6.3315E+06, 6.3481E+06,
     +  6.3646E+06, 6.3812E+06, 6.3979E+06, 6.4145E+06, 6.4312E+06,
     + 6.4480E+06, 6.4647E+06, 6.4815E+06, 6.4984E+06, 6.5152E+06,
     +  6.5321E+06, 6.5491E+06, 6.5660E+06, 6.5830E+06, 6.6001E+06,
     + 6.6172E+06, 6.6343E+06, 6.6514E+06, 6.6686E+06, 6.6858E+06,
     +  6.7031E+06, 6.7203E+06, 6.7377E+06, 6.7550E+06, 6.7724E+06,
     + 6.7898E+06, 6.8073E+06, 6.8248E+06, 6.8423E+06, 6.8599E+06,
     +  6.8775E+06, 6.8951E+06, 6.9127E+06, 6.9305E+06, 6.9482E+06,
     + 6.9660E+06, 6.9838E+06, 7.0016E+06, 7.0195E+06, 7.0374E+06,
     +  7.0554E+06, 7.0734E+06, 7.0914E+06, 7.1094E+06, 7.1275E+06,
     + 7.1457E+06, 7.1638E+06, 7.1820E+06, 7.2003E+06, 7.2186E+06,
     +  7.2369E+06, 7.2552E+06, 7.2736E+06, 7.2920E+06, 7.3105E+06,
     + 7.3290E+06, 7.3475E+06, 7.3661E+06, 7.3847E+06, 7.4033E+06,
     +  7.4220E+06, 7.4407E+06, 7.4595E+06, 7.4783E+06, 7.4971E+06,
     + 7.5160E+06, 7.5349E+06, 7.5538E+06, 7.5728E+06, 7.5918E+06,
     +  7.6108E+06, 7.6299E+06, 7.6490E+06, 7.6682E+06, 7.6874E+06,
     + 7.7066E+06, 7.7259E+06, 7.7452E+06, 7.7646E+06, 7.7839E+06,
     +  7.8034E+06, 7.8228E+06, 7.8423E+06, 7.8619E+06, 7.8815E+06,
     + 7.9011E+06, 7.9207E+06, 7.9404E+06, 7.9601E+06, 7.9799E+06,
     +  7.9997E+06, 8.0196E+06, 8.0394E+06, 8.0594E+06, 8.0793E+06,
     + 8.0993E+06, 8.1194E+06, 8.1394E+06, 8.1595E+06, 8.1797E+06,
     +  8.1999E+06, 8.2201E+06, 8.2404E+06, 8.2607E+06, 8.2811E+06,
     + 8.3014E+06, 8.3219E+06, 8.3423E+06, 8.3629E+06, 8.3834E+06,
     +  8.4040E+06, 8.4246E+06, 8.4453E+06, 8.4660E+06, 8.4867E+06,
     + 8.5075E+06, 8.5283E+06, 8.5492E+06, 8.5701E+06, 8.5910E+06,
     +  8.6120E+06, 8.6331E+06, 8.6541E+06, 8.6752E+06, 8.6964E+06,
     + 8.7175E+06, 8.7388E+06, 8.7600E+06, 8.7813E+06, 8.8027E+06,
     +  8.8241E+06, 8.8455E+06, 8.8670E+06, 8.8885E+06, 8.9100E+06,
     + 8.9316E+06, 8.9532E+06, 8.9749E+06, 8.9966E+06, 9.0184E+06,
     +  9.0402E+06, 9.0620E+06, 9.0839E+06, 9.1058E+06, 9.1278E+06,
     + 9.1498E+06, 9.1718E+06, 9.1939E+06, 9.2160E+06, 9.2382E+06,
     +  9.2604E+06, 9.2827E+06, 9.3050E+06, 9.3273E+06, 9.3497E+06,
     + 9.3721E+06, 9.3946E+06, 9.4171E+06, 9.4396E+06, 9.4622E+06,
     +  9.4848E+06, 9.5075E+06, 9.5302E+06, 9.5530E+06, 9.5758E+06,
     + 9.5986E+06, 9.6215E+06, 9.6444E+06, 9.6674E+06, 9.6904E+06,
     +  9.7135E+06, 9.7366E+06, 9.7597E+06, 9.7829E+06, 9.8061E+06,
     + 9.8294E+06, 9.8527E+06, 9.8761E+06, 9.8995E+06, 9.9229E+06,
     +  9.9464E+06, 9.9700E+06, 9.9935E+06, 1.0017E+07, 1.0041E+07,
     + 1.0065E+07, 1.0088E+07, 1.0112E+07, 1.0136E+07, 1.0160E+07,
     +  1.0184E+07, 1.0208E+07, 1.0232E+07, 1.0256E+07, 1.0280E+07,
     + 1.0304E+07, 1.0328E+07, 1.0352E+07, 1.0377E+07, 1.0401E+07,
     +  1.0425E+07, 1.0450E+07, 1.0474E+07, 1.0499E+07, 1.0523E+07,
     + 1.0548E+07, 1.0573E+07, 1.0597E+07, 1.0622E+07, 1.0647E+07,
     +  1.0672E+07, 1.0696E+07, 1.0721E+07, 1.0746E+07, 1.0771E+07,
     + 1.0796E+07, 1.0821E+07, 1.0847E+07, 1.0872E+07, 1.0897E+07,
     +  1.0922E+07, 1.0948E+07, 1.0973E+07, 1.0998E+07, 1.1024E+07,
     + 1.1049E+07, 1.1075E+07, 1.1101E+07, 1.1126E+07, 1.1152E+07,
     +  1.1178E+07, 1.1203E+07, 1.1229E+07, 1.1255E+07, 1.1281E+07,
     + 1.1307E+07, 1.1333E+07, 1.1359E+07, 1.1385E+07, 1.1411E+07,
     +  1.1438E+07, 1.1464E+07, 1.1490E+07, 1.1516E+07, 1.1543E+07,
     + 1.1569E+07, 1.1596E+07, 1.1622E+07, 1.1649E+07, 1.1676E+07,
     +  1.1702E+07, 1.1729E+07, 1.1756E+07, 1.1783E+07, 1.1809E+07,
     + 1.1836E+07, 1.1863E+07, 1.1890E+07, 1.1917E+07, 1.1945E+07,
     +  1.1972E+07, 1.1999E+07, 1.2026E+07, 1.2053E+07, 1.2081E+07,
     + 1.2108E+07, 1.2136E+07, 1.2163E+07, 1.2191E+07, 1.2218E+07,
     +  1.2246E+07, 1.2274E+07, 1.2301E+07, 1.2329E+07, 1.2357E+07,
     + 1.2385E+07, 1.2413E+07, 1.2441E+07, 1.2469E+07, 1.2497E+07,
     +  1.2525E+07, 1.2553E+07, 1.2582E+07, 1.2610E+07, 1.2638E+07,
     + 1.2667E+07, 1.2695E+07, 1.2724E+07, 1.2752E+07, 1.2781E+07,
     +  1.2809E+07, 1.2838E+07, 1.2867E+07, 1.2896E+07, 1.2924E+07,
     + 1.2953E+07, 1.2982E+07, 1.3011E+07, 1.3040E+07, 1.3069E+07,
     +  1.3098E+07, 1.3128E+07, 1.3157E+07, 1.3186E+07, 1.3216E+07,
     + 1.3245E+07, 1.3274E+07, 1.3304E+07, 1.3334E+07, 1.3363E+07,
     +  1.3393E+07, 1.3422E+07, 1.3452E+07, 1.3482E+07, 1.3512E+07,
     + 1.3542E+07, 1.3572E+07, 1.3602E+07, 1.3632E+07, 1.3662E+07,
     +  1.3692E+07, 1.3722E+07, 1.3753E+07, 1.3783E+07, 1.3813E+07,
     + 1.3844E+07, 1.3874E+07, 1.3905E+07, 1.3936E+07, 1.3966E+07,
     +  1.3997E+07, 1.4028E+07, 1.4058E+07, 1.4089E+07, 1.4120E+07,
     + 1.4151E+07, 1.4182E+07, 1.4213E+07, 1.4244E+07, 1.4276E+07,
     +  1.4307E+07, 1.4338E+07, 1.4369E+07, 1.4401E+07, 1.4432E+07,
     + 1.4464E+07, 1.4495E+07, 1.4527E+07, 1.4559E+07, 1.4590E+07,
     +  1.4622E+07, 1.4654E+07, 1.4686E+07, 1.4718E+07, 1.4750E+07,
     + 1.4782E+07, 1.4814E+07, 1.4846E+07, 1.4878E+07, 1.4911E+07,
     +  1.4943E+07, 1.4975E+07, 1.5008E+07, 1.5040E+07, 1.5073E+07,
     + 1.5105E+07, 1.5138E+07, 1.5171E+07, 1.5203E+07, 1.5236E+07,
     +  1.5269E+07, 1.5302E+07, 1.5335E+07, 1.5368E+07, 1.5401E+07,
     + 1.5434E+07, 1.5467E+07, 1.5501E+07, 1.5534E+07, 1.5567E+07,
     +  1.5601E+07, 1.5634E+07, 1.5668E+07, 1.5701E+07, 1.5735E+07,
     + 1.5769E+07, 1.5803E+07, 1.5836E+07, 1.5870E+07, 1.5904E+07,
     +  1.5938E+07, 1.5972E+07, 1.6006E+07, 1.6041E+07, 1.6075E+07,
     + 1.6109E+07, 1.6143E+07, 1.6178E+07, 1.6212E+07, 1.6247E+07,
     +  1.6281E+07, 1.6316E+07, 1.6351E+07, 1.6385E+07, 1.6420E+07,
     + 1.6455E+07, 1.6490E+07, 1.6525E+07, 1.6560E+07, 1.6595E+07,
     +  1.6630E+07, 1.6665E+07, 1.6701E+07, 1.6736E+07, 1.6771E+07,
     + 1.6807E+07, 1.6842E+07, 1.6878E+07, 1.6914E+07, 1.6949E+07,
     +  1.6985E+07, 1.7021E+07, 1.7057E+07, 1.7093E+07, 1.7129E+07,
     + 1.7165E+07, 1.7201E+07, 1.7237E+07, 1.7273E+07, 1.7309E+07,
     +  1.7346E+07, 1.7382E+07, 1.7418E+07, 1.7455E+07, 1.7492E+07,
     + 1.7528E+07, 1.7565E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       161
      data (QofT( 2,J),J=1,2408)/ 5.7400E+00, 1.5922E+01, 2.8609E+01,
     +  4.3493E+01, 6.0304E+01, 7.8846E+01, 9.8974E+01, 1.2057E+02,
     + 1.4354E+02, 1.6782E+02, 1.9332E+02, 2.2000E+02, 2.4781E+02,
     +  2.7670E+02, 3.0663E+02, 3.3757E+02, 3.6949E+02, 4.0236E+02,
     + 4.3615E+02, 4.7083E+02, 5.4280E+02, 6.1812E+02, 6.9663E+02,
     +  7.7821E+02, 8.6276E+02, 9.5016E+02, 1.0403E+03, 1.1332E+03,
     + 1.2287E+03, 1.3267E+03, 1.4272E+03, 1.5302E+03, 1.6355E+03,
     +  1.7431E+03, 1.8530E+03, 1.9651E+03, 2.0794E+03, 2.1958E+03,
     + 2.3144E+03, 2.4350E+03, 2.5576E+03, 2.6822E+03, 2.8088E+03,
     +  2.9374E+03, 3.0678E+03, 3.2002E+03, 3.3344E+03, 3.4704E+03,
     + 3.6082E+03, 3.7479E+03, 3.8893E+03, 4.0324E+03, 4.1773E+03,
     +  4.3239E+03, 4.4722E+03, 4.6221E+03, 4.7737E+03, 4.9270E+03,
     + 5.0819E+03, 5.2384E+03, 5.3965E+03, 5.5562E+03, 5.7175E+03,
     +  5.8804E+03, 6.0448E+03, 6.2108E+03, 6.3784E+03, 6.5474E+03,
     + 6.7180E+03, 6.8901E+03, 7.0638E+03, 7.2389E+03, 7.4156E+03,
     +  7.5937E+03, 7.7734E+03, 7.9546E+03, 8.1373E+03, 8.3214E+03,
     + 8.5070E+03, 8.6942E+03, 8.8828E+03, 9.0730E+03, 9.2646E+03,
     +  9.4577E+03, 9.6522E+03, 9.8483E+03, 1.0046E+04, 1.0245E+04,
     + 1.0446E+04, 1.0648E+04, 1.0851E+04, 1.1056E+04, 1.1263E+04,
     +  1.1471E+04, 1.1681E+04, 1.1892E+04, 1.2104E+04, 1.2319E+04,
     + 1.2534E+04, 1.2752E+04, 1.2970E+04, 1.3191E+04, 1.3413E+04,
     +  1.3636E+04, 1.3861E+04, 1.4088E+04, 1.4316E+04, 1.4546E+04,
     + 1.4777E+04, 1.5010E+04, 1.5245E+04, 1.5481E+04, 1.5719E+04,
     +  1.5959E+04, 1.6200E+04, 1.6442E+04, 1.6687E+04, 1.6933E+04,
     + 1.7181E+04, 1.7430E+04, 1.7681E+04, 1.7934E+04, 1.8188E+04,
     +  1.8445E+04, 1.8702E+04, 1.8962E+04, 1.9224E+04, 1.9487E+04,
     + 1.9752E+04, 2.0018E+04, 2.0287E+04, 2.0557E+04, 2.0829E+04,
     +  2.1103E+04, 2.1378E+04, 2.1656E+04, 2.1935E+04, 2.2216E+04,
     + 2.2499E+04, 2.2784E+04, 2.3070E+04, 2.3359E+04, 2.3649E+04,
     +  2.3942E+04, 2.4236E+04, 2.4532E+04, 2.4830E+04, 2.5130E+04,
     + 2.5432E+04, 2.5736E+04, 2.6041E+04, 2.6349E+04, 2.6659E+04,
     +  2.6971E+04, 2.7285E+04, 2.7600E+04, 2.7918E+04, 2.8238E+04,
     + 2.8560E+04, 2.8884E+04, 2.9210E+04, 2.9538E+04, 2.9868E+04,
     +  3.0200E+04, 3.0534E+04, 3.0871E+04, 3.1209E+04, 3.1550E+04,
     + 3.1893E+04, 3.2238E+04, 3.2585E+04, 3.2935E+04, 3.3286E+04,
     +  3.3640E+04, 3.3996E+04, 3.4354E+04, 3.4714E+04, 3.5077E+04,
     + 3.5442E+04, 3.5809E+04, 3.6178E+04, 3.6550E+04, 3.6924E+04,
     +  3.7300E+04, 3.7679E+04, 3.8060E+04, 3.8443E+04, 3.8829E+04,
     + 3.9217E+04, 3.9607E+04, 4.0000E+04, 4.0395E+04, 4.0792E+04,
     +  4.1193E+04, 4.1595E+04, 4.2000E+04, 4.2407E+04, 4.2817E+04,
     + 4.3229E+04, 4.3644E+04, 4.4061E+04, 4.4480E+04, 4.4903E+04,
     +  4.5327E+04, 4.5755E+04, 4.6184E+04, 4.6617E+04, 4.7052E+04,
     + 4.7489E+04, 4.7929E+04, 4.8372E+04, 4.8817E+04, 4.9265E+04,
     +  4.9716E+04, 5.0169E+04, 5.0625E+04, 5.1083E+04, 5.1545E+04,
     + 5.2009E+04, 5.2475E+04, 5.2945E+04, 5.3417E+04, 5.3892E+04,
     +  5.4369E+04, 5.4849E+04, 5.5333E+04, 5.5819E+04, 5.6308E+04,
     + 5.6799E+04, 5.7293E+04, 5.7790E+04, 5.8290E+04, 5.8793E+04,
     +  5.9299E+04, 5.9807E+04, 6.0319E+04, 6.0833E+04, 6.1351E+04,
     + 6.1871E+04, 6.2394E+04, 6.2920E+04, 6.3450E+04, 6.3981E+04,
     +  6.4516E+04, 6.5054E+04, 6.5596E+04, 6.6139E+04, 6.6686E+04,
     + 6.7236E+04, 6.7789E+04, 6.8346E+04, 6.8905E+04, 6.9467E+04,
     +  7.0033E+04, 7.0601E+04, 7.1173E+04, 7.1748E+04, 7.2326E+04,
     + 7.2907E+04, 7.3492E+04, 7.4079E+04, 7.4670E+04, 7.5264E+04,
     +  7.5861E+04, 7.6461E+04, 7.7065E+04, 7.7672E+04, 7.8282E+04,
     + 7.8895E+04, 7.9512E+04, 8.0132E+04, 8.0755E+04, 8.1381E+04,
     +  8.2012E+04, 8.2645E+04, 8.3281E+04, 8.3921E+04, 8.4565E+04,
     + 8.5212E+04, 8.5862E+04, 8.6516E+04, 8.7173E+04, 8.7833E+04,
     +  8.8498E+04, 8.9165E+04, 8.9836E+04, 9.0510E+04, 9.1188E+04,
     + 9.1870E+04, 9.2555E+04, 9.3243E+04, 9.3936E+04, 9.4632E+04,
     +  9.5331E+04, 9.6034E+04, 9.6741E+04, 9.7450E+04, 9.8164E+04,
     + 9.8881E+04, 9.9602E+04, 1.0033E+05, 1.0106E+05, 1.0179E+05,
     +  1.0252E+05, 1.0326E+05, 1.0401E+05, 1.0475E+05, 1.0550E+05,
     + 1.0626E+05, 1.0702E+05, 1.0778E+05, 1.0855E+05, 1.0932E+05,
     +  1.1009E+05, 1.1087E+05, 1.1165E+05, 1.1243E+05, 1.1322E+05,
     + 1.1402E+05, 1.1481E+05, 1.1561E+05, 1.1642E+05, 1.1723E+05,
     +  1.1804E+05, 1.1886E+05, 1.1968E+05, 1.2050E+05, 1.2133E+05,
     + 1.2216E+05, 1.2300E+05, 1.2384E+05, 1.2469E+05, 1.2553E+05,
     +  1.2639E+05, 1.2724E+05, 1.2811E+05, 1.2897E+05, 1.2984E+05,
     + 1.3071E+05, 1.3159E+05, 1.3247E+05, 1.3336E+05, 1.3425E+05,
     +  1.3515E+05, 1.3605E+05, 1.3695E+05, 1.3786E+05, 1.3877E+05,
     + 1.3968E+05, 1.4060E+05, 1.4153E+05, 1.4246E+05, 1.4339E+05,
     +  1.4433E+05, 1.4527E+05, 1.4622E+05, 1.4717E+05, 1.4813E+05,
     + 1.4909E+05, 1.5005E+05, 1.5102E+05, 1.5199E+05, 1.5297E+05,
     +  1.5395E+05, 1.5494E+05, 1.5593E+05, 1.5693E+05, 1.5793E+05,
     + 1.5893E+05, 1.5994E+05, 1.6096E+05, 1.6198E+05, 1.6300E+05,
     +  1.6403E+05, 1.6506E+05, 1.6610E+05, 1.6714E+05, 1.6819E+05,
     + 1.6924E+05, 1.7029E+05, 1.7135E+05, 1.7242E+05, 1.7349E+05,
     +  1.7457E+05, 1.7565E+05, 1.7673E+05, 1.7782E+05, 1.7892E+05,
     + 1.8002E+05, 1.8112E+05, 1.8223E+05, 1.8334E+05, 1.8446E+05,
     +  1.8559E+05, 1.8672E+05, 1.8785E+05, 1.8899E+05, 1.9013E+05,
     + 1.9128E+05, 1.9244E+05, 1.9360E+05, 1.9476E+05, 1.9593E+05,
     +  1.9710E+05, 1.9828E+05, 1.9947E+05, 2.0066E+05, 2.0185E+05,
     + 2.0305E+05, 2.0426E+05, 2.0547E+05, 2.0668E+05, 2.0790E+05,
     +  2.0913E+05, 2.1036E+05, 2.1160E+05, 2.1284E+05, 2.1409E+05,
     + 2.1534E+05, 2.1660E+05, 2.1786E+05, 2.1913E+05, 2.2040E+05,
     +  2.2168E+05, 2.2296E+05, 2.2425E+05, 2.2555E+05, 2.2685E+05,
     + 2.2816E+05, 2.2947E+05, 2.3079E+05, 2.3211E+05, 2.3344E+05,
     +  2.3477E+05, 2.3611E+05, 2.3746E+05, 2.3881E+05, 2.4016E+05,
     + 2.4152E+05, 2.4289E+05, 2.4426E+05, 2.4564E+05, 2.4703E+05,
     +  2.4842E+05, 2.4981E+05, 2.5122E+05, 2.5262E+05, 2.5404E+05,
     + 2.5546E+05, 2.5688E+05, 2.5831E+05, 2.5975E+05, 2.6119E+05,
     +  2.6264E+05, 2.6409E+05, 2.6555E+05, 2.6702E+05, 2.6849E+05,
     + 2.6997E+05, 2.7145E+05, 2.7294E+05, 2.7444E+05, 2.7594E+05,
     +  2.7744E+05, 2.7896E+05, 2.8048E+05, 2.8200E+05, 2.8354E+05,
     + 2.8507E+05, 2.8662E+05, 2.8817E+05, 2.8973E+05, 2.9129E+05,
     +  2.9286E+05, 2.9443E+05, 2.9601E+05, 2.9760E+05, 2.9920E+05,
     + 3.0080E+05, 3.0240E+05, 3.0401E+05, 3.0563E+05, 3.0726E+05,
     +  3.0889E+05, 3.1053E+05, 3.1217E+05, 3.1382E+05, 3.1548E+05,
     + 3.1715E+05, 3.1882E+05, 3.2049E+05, 3.2218E+05, 3.2387E+05,
     +  3.2556E+05, 3.2727E+05, 3.2898E+05, 3.3069E+05, 3.3242E+05,
     + 3.3414E+05, 3.3588E+05, 3.3762E+05, 3.3937E+05, 3.4113E+05,
     +  3.4289E+05, 3.4466E+05, 3.4644E+05, 3.4822E+05, 3.5001E+05,
     + 3.5181E+05, 3.5361E+05, 3.5542E+05, 3.5724E+05, 3.5907E+05,
     +  3.6090E+05, 3.6274E+05, 3.6458E+05, 3.6643E+05, 3.6829E+05,
     + 3.7016E+05, 3.7203E+05, 3.7391E+05, 3.7580E+05, 3.7769E+05,
     +  3.7959E+05, 3.8150E+05, 3.8342E+05, 3.8534E+05, 3.8727E+05,
     + 3.8920E+05, 3.9115E+05, 3.9310E+05, 3.9506E+05, 3.9702E+05,
     +  3.9900E+05, 4.0098E+05, 4.0297E+05, 4.0496E+05, 4.0696E+05,
     + 4.0897E+05, 4.1099E+05, 4.1301E+05, 4.1505E+05, 4.1709E+05,
     +  4.1913E+05, 4.2119E+05, 4.2325E+05, 4.2532E+05, 4.2740E+05,
     + 4.2948E+05, 4.3157E+05, 4.3367E+05, 4.3578E+05, 4.3789E+05,
     +  4.4002E+05, 4.4214E+05, 4.4428E+05, 4.4643E+05, 4.4858E+05,
     + 4.5074E+05, 4.5291E+05, 4.5509E+05, 4.5727E+05, 4.5946E+05,
     +  4.6167E+05, 4.6387E+05, 4.6609E+05, 4.6831E+05, 4.7055E+05,
     + 4.7278E+05, 4.7503E+05, 4.7729E+05, 4.7955E+05, 4.8182E+05,
     +  4.8410E+05, 4.8639E+05, 4.8869E+05, 4.9099E+05, 4.9330E+05,
     + 4.9562E+05, 4.9795E+05, 5.0029E+05, 5.0263E+05, 5.0499E+05,
     +  5.0735E+05, 5.0972E+05, 5.1210E+05, 5.1448E+05, 5.1688E+05,
     + 5.1928E+05, 5.2169E+05, 5.2411E+05, 5.2654E+05, 5.2897E+05,
     +  5.3142E+05, 5.3387E+05, 5.3633E+05, 5.3881E+05, 5.4128E+05,
     + 5.4377E+05, 5.4627E+05, 5.4877E+05, 5.5129E+05, 5.5381E+05,
     +  5.5634E+05, 5.5888E+05, 5.6143E+05, 5.6398E+05, 5.6655E+05,
     + 5.6912E+05, 5.7171E+05, 5.7430E+05, 5.7690E+05, 5.7951E+05,
     +  5.8213E+05, 5.8475E+05, 5.8739E+05, 5.9004E+05, 5.9269E+05,
     + 5.9535E+05, 5.9803E+05, 6.0071E+05, 6.0340E+05, 6.0610E+05,
     +  6.0881E+05, 6.1152E+05, 6.1425E+05, 6.1699E+05, 6.1973E+05,
     + 6.2249E+05, 6.2525E+05, 6.2802E+05, 6.3080E+05, 6.3360E+05,
     +  6.3640E+05, 6.3921E+05, 6.4202E+05, 6.4486E+05, 6.4769E+05,
     + 6.5054E+05, 6.5340E+05, 6.5626E+05, 6.5914E+05, 6.6203E+05,
     +  6.6492E+05, 6.6782E+05, 6.7074E+05, 6.7366E+05, 6.7660E+05,
     + 6.7954E+05, 6.8249E+05, 6.8546E+05, 6.8843E+05, 6.9141E+05,
     +  6.9440E+05, 6.9740E+05, 7.0042E+05, 7.0344E+05, 7.0647E+05,
     + 7.0951E+05, 7.1256E+05, 7.1562E+05, 7.1869E+05, 7.2177E+05,
     +  7.2486E+05, 7.2796E+05, 7.3107E+05, 7.3419E+05, 7.3732E+05,
     + 7.4046E+05, 7.4361E+05, 7.4678E+05, 7.4995E+05, 7.5313E+05,
     +  7.5632E+05, 7.5952E+05, 7.6273E+05, 7.6596E+05, 7.6919E+05,
     + 7.7243E+05, 7.7569E+05, 7.7895E+05, 7.8222E+05, 7.8551E+05,
     +  7.8881E+05, 7.9211E+05, 7.9543E+05, 7.9876E+05, 8.0209E+05,
     + 8.0544E+05, 8.0880E+05, 8.1217E+05, 8.1555E+05, 8.1894E+05,
     +  8.2234E+05, 8.2575E+05, 8.2918E+05, 8.3261E+05, 8.3605E+05,
     + 8.3951E+05, 8.4297E+05, 8.4645E+05, 8.4994E+05, 8.5344E+05,
     +  8.5695E+05, 8.6047E+05, 8.6400E+05, 8.6754E+05, 8.7110E+05,
     + 8.7466E+05, 8.7824E+05, 8.8182E+05, 8.8542E+05, 8.8903E+05,
     +  8.9265E+05, 8.9628E+05, 8.9993E+05, 9.0358E+05, 9.0724E+05,
     + 9.1092E+05, 9.1461E+05, 9.1831E+05, 9.2202E+05, 9.2574E+05,
     +  9.2947E+05, 9.3322E+05, 9.3698E+05, 9.4075E+05, 9.4453E+05,
     + 9.4832E+05, 9.5212E+05, 9.5594E+05, 9.5976E+05, 9.6360E+05,
     +  9.6745E+05, 9.7131E+05, 9.7518E+05, 9.7907E+05, 9.8296E+05,
     + 9.8687E+05, 9.9079E+05, 9.9472E+05, 9.9867E+05, 1.0026E+06,
     +  1.0066E+06, 1.0106E+06, 1.0146E+06, 1.0186E+06, 1.0226E+06,
     + 1.0266E+06, 1.0307E+06, 1.0347E+06, 1.0388E+06, 1.0428E+06,
     +  1.0469E+06, 1.0510E+06, 1.0552E+06, 1.0593E+06, 1.0634E+06,
     + 1.0676E+06, 1.0717E+06, 1.0759E+06, 1.0801E+06, 1.0843E+06,
     +  1.0885E+06, 1.0927E+06, 1.0970E+06, 1.1012E+06, 1.1055E+06,
     + 1.1098E+06, 1.1140E+06, 1.1183E+06, 1.1227E+06, 1.1270E+06,
     +  1.1313E+06, 1.1357E+06, 1.1400E+06, 1.1444E+06, 1.1488E+06,
     + 1.1532E+06, 1.1576E+06, 1.1621E+06, 1.1665E+06, 1.1709E+06,
     +  1.1754E+06, 1.1799E+06, 1.1844E+06, 1.1889E+06, 1.1934E+06,
     + 1.1979E+06, 1.2025E+06, 1.2071E+06, 1.2116E+06, 1.2162E+06,
     +  1.2208E+06, 1.2254E+06, 1.2301E+06, 1.2347E+06, 1.2393E+06,
     + 1.2440E+06, 1.2487E+06, 1.2534E+06, 1.2581E+06, 1.2628E+06,
     +  1.2675E+06, 1.2723E+06, 1.2771E+06, 1.2818E+06, 1.2866E+06,
     + 1.2914E+06, 1.2962E+06, 1.3011E+06, 1.3059E+06, 1.3108E+06,
     +  1.3156E+06, 1.3205E+06, 1.3254E+06, 1.3303E+06, 1.3353E+06,
     + 1.3402E+06, 1.3452E+06, 1.3501E+06, 1.3551E+06, 1.3601E+06,
     +  1.3651E+06, 1.3701E+06, 1.3752E+06, 1.3802E+06, 1.3853E+06,
     + 1.3904E+06, 1.3955E+06, 1.4006E+06, 1.4057E+06, 1.4109E+06,
     +  1.4160E+06, 1.4212E+06, 1.4264E+06, 1.4316E+06, 1.4368E+06,
     + 1.4420E+06, 1.4472E+06, 1.4525E+06, 1.4578E+06, 1.4631E+06,
     +  1.4684E+06, 1.4737E+06, 1.4790E+06, 1.4844E+06, 1.4897E+06,
     + 1.4951E+06, 1.5005E+06, 1.5059E+06, 1.5113E+06, 1.5167E+06,
     +  1.5222E+06, 1.5276E+06, 1.5331E+06, 1.5386E+06, 1.5441E+06,
     + 1.5497E+06, 1.5552E+06, 1.5608E+06, 1.5663E+06, 1.5719E+06,
     +  1.5775E+06, 1.5831E+06, 1.5888E+06, 1.5944E+06, 1.6001E+06,
     + 1.6057E+06, 1.6114E+06, 1.6171E+06, 1.6229E+06, 1.6286E+06,
     +  1.6344E+06, 1.6401E+06, 1.6459E+06, 1.6517E+06, 1.6576E+06,
     + 1.6634E+06, 1.6692E+06, 1.6751E+06, 1.6810E+06, 1.6869E+06,
     +  1.6928E+06, 1.6987E+06, 1.7047E+06, 1.7106E+06, 1.7166E+06,
     + 1.7226E+06, 1.7286E+06, 1.7346E+06, 1.7407E+06, 1.7467E+06,
     +  1.7528E+06, 1.7589E+06, 1.7650E+06, 1.7711E+06, 1.7773E+06,
     + 1.7834E+06, 1.7896E+06, 1.7958E+06, 1.8020E+06, 1.8082E+06,
     +  1.8145E+06, 1.8207E+06, 1.8270E+06, 1.8333E+06, 1.8396E+06,
     + 1.8459E+06, 1.8523E+06, 1.8586E+06, 1.8650E+06, 1.8714E+06,
     +  1.8778E+06, 1.8842E+06, 1.8907E+06, 1.8971E+06, 1.9036E+06,
     + 1.9101E+06, 1.9166E+06, 1.9231E+06, 1.9297E+06, 1.9362E+06,
     +  1.9428E+06, 1.9494E+06, 1.9560E+06, 1.9626E+06, 1.9693E+06,
     + 1.9760E+06, 1.9826E+06, 1.9893E+06, 1.9961E+06, 2.0028E+06,
     +  2.0095E+06, 2.0163E+06, 2.0231E+06, 2.0299E+06, 2.0367E+06,
     + 2.0436E+06, 2.0504E+06, 2.0573E+06, 2.0642E+06, 2.0711E+06,
     +  2.0780E+06, 2.0850E+06, 2.0919E+06, 2.0989E+06, 2.1059E+06,
     + 2.1129E+06, 2.1200E+06, 2.1270E+06, 2.1341E+06, 2.1412E+06,
     +  2.1483E+06, 2.1554E+06, 2.1626E+06, 2.1698E+06, 2.1769E+06,
     + 2.1841E+06, 2.1914E+06, 2.1986E+06, 2.2058E+06, 2.2131E+06,
     +  2.2204E+06, 2.2277E+06, 2.2351E+06, 2.2424E+06, 2.2498E+06,
     + 2.2572E+06, 2.2646E+06, 2.2720E+06, 2.2794E+06, 2.2869E+06,
     +  2.2944E+06, 2.3019E+06, 2.3094E+06, 2.3169E+06, 2.3245E+06,
     + 2.3321E+06, 2.3397E+06, 2.3473E+06, 2.3549E+06, 2.3626E+06,
     +  2.3703E+06, 2.3779E+06, 2.3857E+06, 2.3934E+06, 2.4011E+06,
     + 2.4089E+06, 2.4167E+06, 2.4245E+06, 2.4323E+06, 2.4402E+06,
     +  2.4480E+06, 2.4559E+06, 2.4638E+06, 2.4718E+06, 2.4797E+06,
     + 2.4877E+06, 2.4957E+06, 2.5037E+06, 2.5117E+06, 2.5197E+06,
     +  2.5278E+06, 2.5359E+06, 2.5440E+06, 2.5521E+06, 2.5603E+06,
     + 2.5684E+06, 2.5766E+06, 2.5848E+06, 2.5930E+06, 2.6013E+06,
     +  2.6095E+06, 2.6178E+06, 2.6261E+06, 2.6345E+06, 2.6428E+06,
     + 2.6512E+06, 2.6596E+06, 2.6680E+06, 2.6764E+06, 2.6849E+06,
     +  2.6933E+06, 2.7018E+06, 2.7103E+06, 2.7189E+06, 2.7274E+06,
     + 2.7360E+06, 2.7446E+06, 2.7532E+06, 2.7619E+06, 2.7705E+06,
     +  2.7792E+06, 2.7879E+06, 2.7966E+06, 2.8054E+06, 2.8141E+06,
     + 2.8229E+06, 2.8317E+06, 2.8406E+06, 2.8494E+06, 2.8583E+06,
     +  2.8672E+06, 2.8761E+06, 2.8850E+06, 2.8940E+06, 2.9030E+06,
     + 2.9120E+06, 2.9210E+06, 2.9300E+06, 2.9391E+06, 2.9482E+06,
     +  2.9573E+06, 2.9664E+06, 2.9756E+06, 2.9847E+06, 2.9939E+06,
     + 3.0032E+06, 3.0124E+06, 3.0217E+06, 3.0310E+06, 3.0403E+06,
     +  3.0496E+06, 3.0589E+06, 3.0683E+06, 3.0777E+06, 3.0871E+06,
     + 3.0966E+06, 3.1060E+06, 3.1155E+06, 3.1250E+06, 3.1346E+06,
     +  3.1441E+06, 3.1537E+06, 3.1633E+06, 3.1729E+06, 3.1826E+06,
     + 3.1922E+06, 3.2019E+06, 3.2116E+06, 3.2214E+06, 3.2311E+06,
     +  3.2409E+06, 3.2507E+06, 3.2605E+06, 3.2704E+06, 3.2803E+06,
     + 3.2902E+06, 3.3001E+06, 3.3100E+06, 3.3200E+06, 3.3300E+06,
     +  3.3400E+06, 3.3500E+06, 3.3601E+06, 3.3702E+06, 3.3803E+06,
     + 3.3904E+06, 3.4006E+06, 3.4107E+06, 3.4209E+06, 3.4312E+06,
     +  3.4414E+06, 3.4517E+06, 3.4620E+06, 3.4723E+06, 3.4826E+06,
     + 3.4930E+06, 3.5034E+06, 3.5138E+06, 3.5243E+06, 3.5347E+06,
     +  3.5452E+06, 3.5557E+06, 3.5663E+06, 3.5768E+06, 3.5874E+06,
     + 3.5980E+06, 3.6087E+06, 3.6193E+06, 3.6300E+06, 3.6407E+06,
     +  3.6514E+06, 3.6622E+06, 3.6730E+06, 3.6838E+06, 3.6946E+06,
     + 3.7055E+06, 3.7164E+06, 3.7273E+06, 3.7382E+06, 3.7492E+06,
     +  3.7601E+06, 3.7711E+06, 3.7822E+06, 3.7932E+06, 3.8043E+06,
     + 3.8154E+06, 3.8265E+06, 3.8377E+06, 3.8489E+06, 3.8601E+06,
     +  3.8713E+06, 3.8826E+06, 3.8939E+06, 3.9052E+06, 3.9165E+06,
     + 3.9279E+06, 3.9392E+06, 3.9507E+06, 3.9621E+06, 3.9736E+06,
     +  3.9851E+06, 3.9966E+06, 4.0081E+06, 4.0197E+06, 4.0313E+06,
     + 4.0429E+06, 4.0545E+06, 4.0662E+06, 4.0779E+06, 4.0896E+06,
     +  4.1014E+06, 4.1131E+06, 4.1249E+06, 4.1368E+06, 4.1486E+06,
     + 4.1605E+06, 4.1724E+06, 4.1844E+06, 4.1963E+06, 4.2083E+06,
     +  4.2203E+06, 4.2324E+06, 4.2444E+06, 4.2565E+06, 4.2686E+06,
     + 4.2808E+06, 4.2930E+06, 4.3052E+06, 4.3174E+06, 4.3296E+06,
     +  4.3419E+06, 4.3542E+06, 4.3666E+06, 4.3789E+06, 4.3913E+06,
     + 4.4038E+06, 4.4162E+06, 4.4287E+06, 4.4412E+06, 4.4537E+06,
     +  4.4663E+06, 4.4788E+06, 4.4915E+06, 4.5041E+06, 4.5168E+06,
     + 4.5295E+06, 4.5422E+06, 4.5549E+06, 4.5677E+06, 4.5805E+06,
     +  4.5933E+06, 4.6062E+06, 4.6191E+06, 4.6320E+06, 4.6450E+06,
     + 4.6579E+06, 4.6709E+06, 4.6840E+06, 4.6970E+06, 4.7101E+06,
     +  4.7232E+06, 4.7364E+06, 4.7496E+06, 4.7628E+06, 4.7760E+06,
     + 4.7892E+06, 4.8025E+06, 4.8158E+06, 4.8292E+06, 4.8426E+06,
     +  4.8560E+06, 4.8694E+06, 4.8829E+06, 4.8964E+06, 4.9099E+06,
     + 4.9234E+06, 4.9370E+06, 4.9506E+06, 4.9642E+06, 4.9779E+06,
     +  4.9916E+06, 5.0053E+06, 5.0191E+06, 5.0329E+06, 5.0467E+06,
     + 5.0605E+06, 5.0744E+06, 5.0883E+06, 5.1022E+06, 5.1162E+06,
     +  5.1302E+06, 5.1442E+06, 5.1582E+06, 5.1723E+06, 5.1864E+06,
     + 5.2006E+06, 5.2147E+06, 5.2289E+06, 5.2432E+06, 5.2574E+06,
     +  5.2717E+06, 5.2860E+06, 5.3004E+06, 5.3148E+06, 5.3292E+06,
     + 5.3436E+06, 5.3581E+06, 5.3726E+06, 5.3871E+06, 5.4017E+06,
     +  5.4163E+06, 5.4309E+06, 5.4456E+06, 5.4603E+06, 5.4750E+06,
     + 5.4897E+06, 5.5045E+06, 5.5193E+06, 5.5342E+06, 5.5490E+06,
     +  5.5639E+06, 5.5789E+06, 5.5938E+06, 5.6088E+06, 5.6239E+06,
     + 5.6389E+06, 5.6540E+06, 5.6692E+06, 5.6843E+06, 5.6995E+06,
     +  5.7147E+06, 5.7300E+06, 5.7453E+06, 5.7606E+06, 5.7759E+06,
     + 5.7913E+06, 5.8067E+06, 5.8222E+06, 5.8376E+06, 5.8531E+06,
     +  5.8687E+06, 5.8842E+06, 5.8998E+06, 5.9155E+06, 5.9311E+06,
     + 5.9468E+06, 5.9626E+06, 5.9783E+06, 5.9941E+06, 6.0100E+06,
     +  6.0258E+06, 6.0417E+06, 6.0577E+06, 6.0736E+06, 6.0896E+06,
     + 6.1056E+06, 6.1217E+06, 6.1378E+06, 6.1539E+06, 6.1701E+06,
     +  6.1863E+06, 6.2025E+06, 6.2188E+06, 6.2351E+06, 6.2514E+06,
     + 6.2677E+06, 6.2841E+06, 6.3006E+06, 6.3170E+06, 6.3335E+06,
     +  6.3500E+06, 6.3666E+06, 6.3832E+06, 6.3998E+06, 6.4165E+06,
     + 6.4332E+06, 6.4499E+06, 6.4667E+06, 6.4835E+06, 6.5003E+06,
     +  6.5172E+06, 6.5341E+06, 6.5510E+06, 6.5680E+06, 6.5850E+06,
     + 6.6020E+06, 6.6191E+06, 6.6362E+06, 6.6533E+06, 6.6705E+06,
     +  6.6877E+06, 6.7049E+06, 6.7222E+06, 6.7395E+06, 6.7569E+06,
     + 6.7743E+06, 6.7917E+06, 6.8091E+06, 6.8266E+06, 6.8442E+06,
     +  6.8617E+06, 6.8793E+06, 6.8969E+06, 6.9146E+06, 6.9323E+06,
     + 6.9500E+06, 6.9678E+06, 6.9856E+06, 7.0034E+06, 7.0213E+06,
     +  7.0392E+06, 7.0572E+06, 7.0752E+06, 7.0932E+06, 7.1112E+06,
     + 7.1293E+06, 7.1475E+06, 7.1656E+06, 7.1838E+06, 7.2021E+06,
     +  7.2203E+06, 7.2387E+06, 7.2570E+06, 7.2754E+06, 7.2938E+06,
     + 7.3123E+06, 7.3307E+06, 7.3493E+06, 7.3678E+06, 7.3864E+06,
     +  7.4051E+06, 7.4237E+06, 7.4425E+06, 7.4612E+06, 7.4800E+06,
     + 7.4988E+06, 7.5177E+06, 7.5366E+06, 7.5555E+06, 7.5745E+06,
     +  7.5935E+06, 7.6125E+06, 7.6316E+06, 7.6507E+06, 7.6699E+06,
     + 7.6891E+06, 7.7083E+06, 7.7276E+06, 7.7469E+06, 7.7662E+06,
     +  7.7856E+06, 7.8050E+06, 7.8245E+06, 7.8440E+06, 7.8635E+06,
     + 7.8831E+06, 7.9027E+06, 7.9223E+06, 7.9420E+06, 7.9617E+06,
     +  7.9815E+06, 8.0013E+06, 8.0211E+06, 8.0410E+06, 8.0609E+06,
     + 8.0809E+06, 8.1009E+06, 8.1209E+06, 8.1410E+06, 8.1611E+06,
     +  8.1813E+06, 8.2014E+06, 8.2217E+06, 8.2419E+06, 8.2622E+06,
     + 8.2826E+06, 8.3030E+06, 8.3234E+06, 8.3438E+06, 8.3643E+06,
     +  8.3849E+06, 8.4055E+06, 8.4261E+06, 8.4467E+06, 8.4674E+06,
     + 8.4882E+06, 8.5090E+06, 8.5298E+06, 8.5506E+06, 8.5715E+06,
     +  8.5925E+06, 8.6134E+06, 8.6345E+06, 8.6555E+06, 8.6766E+06,
     + 8.6978E+06, 8.7189E+06, 8.7402E+06, 8.7614E+06, 8.7827E+06,
     +  8.8040E+06, 8.8254E+06, 8.8469E+06, 8.8683E+06, 8.8898E+06,
     + 8.9114E+06, 8.9329E+06, 8.9546E+06, 8.9762E+06, 8.9979E+06,
     +  9.0197E+06, 9.0415E+06, 9.0633E+06, 9.0852E+06, 9.1071E+06,
     + 9.1290E+06, 9.1510E+06, 9.1731E+06, 9.1952E+06, 9.2173E+06,
     +  9.2394E+06, 9.2616E+06, 9.2839E+06, 9.3062E+06, 9.3285E+06,
     + 9.3509E+06, 9.3733E+06, 9.3958E+06, 9.4182E+06, 9.4408E+06,
     +  9.4634E+06, 9.4860E+06, 9.5087E+06, 9.5314E+06, 9.5541E+06,
     + 9.5769E+06, 9.5997E+06, 9.6226E+06, 9.6455E+06, 9.6685E+06,
     +  9.6915E+06, 9.7146E+06, 9.7377E+06, 9.7608E+06, 9.7840E+06,
     + 9.8072E+06, 9.8305E+06, 9.8538E+06, 9.8771E+06, 9.9005E+06,
     +  9.9239E+06, 9.9474E+06, 9.9710E+06, 9.9945E+06, 1.0018E+07,
     + 1.0042E+07, 1.0065E+07, 1.0089E+07, 1.0113E+07, 1.0137E+07,
     +  1.0161E+07, 1.0185E+07, 1.0209E+07, 1.0233E+07, 1.0257E+07,
     + 1.0281E+07, 1.0305E+07, 1.0329E+07, 1.0353E+07, 1.0378E+07,
     +  1.0402E+07, 1.0426E+07, 1.0451E+07, 1.0475E+07, 1.0500E+07,
     + 1.0524E+07, 1.0549E+07, 1.0573E+07, 1.0598E+07, 1.0623E+07,
     +  1.0648E+07, 1.0672E+07, 1.0697E+07, 1.0722E+07, 1.0747E+07,
     + 1.0772E+07, 1.0797E+07, 1.0822E+07, 1.0847E+07, 1.0872E+07,
     +  1.0898E+07, 1.0923E+07, 1.0948E+07, 1.0974E+07, 1.0999E+07,
     + 1.1025E+07, 1.1050E+07, 1.1076E+07, 1.1101E+07, 1.1127E+07,
     +  1.1152E+07, 1.1178E+07, 1.1204E+07, 1.1230E+07, 1.1256E+07,
     + 1.1282E+07, 1.1308E+07, 1.1334E+07, 1.1360E+07, 1.1386E+07,
     +  1.1412E+07, 1.1438E+07, 1.1464E+07, 1.1491E+07, 1.1517E+07,
     + 1.1543E+07, 1.1570E+07, 1.1596E+07, 1.1623E+07, 1.1649E+07,
     +  1.1676E+07, 1.1703E+07, 1.1729E+07, 1.1756E+07, 1.1783E+07,
     + 1.1810E+07, 1.1837E+07, 1.1864E+07, 1.1891E+07, 1.1918E+07,
     +  1.1945E+07, 1.1972E+07, 1.1999E+07, 1.2026E+07, 1.2054E+07,
     + 1.2081E+07, 1.2108E+07, 1.2136E+07, 1.2163E+07, 1.2191E+07,
     +  1.2218E+07, 1.2246E+07, 1.2274E+07, 1.2302E+07, 1.2329E+07,
     + 1.2357E+07, 1.2385E+07, 1.2413E+07, 1.2441E+07, 1.2469E+07,
     +  1.2497E+07, 1.2525E+07, 1.2553E+07, 1.2582E+07, 1.2610E+07,
     + 1.2638E+07, 1.2667E+07, 1.2695E+07, 1.2723E+07, 1.2752E+07,
     +  1.2781E+07, 1.2809E+07, 1.2838E+07, 1.2867E+07, 1.2895E+07,
     + 1.2924E+07, 1.2953E+07, 1.2982E+07, 1.3011E+07, 1.3040E+07,
     +  1.3069E+07, 1.3098E+07, 1.3127E+07, 1.3157E+07, 1.3186E+07,
     + 1.3215E+07, 1.3245E+07, 1.3274E+07, 1.3304E+07, 1.3333E+07,
     +  1.3363E+07, 1.3392E+07, 1.3422E+07, 1.3452E+07, 1.3482E+07,
     + 1.3512E+07, 1.3541E+07, 1.3571E+07, 1.3601E+07, 1.3631E+07,
     +  1.3662E+07, 1.3692E+07, 1.3722E+07, 1.3752E+07, 1.3783E+07,
     + 1.3813E+07, 1.3843E+07, 1.3874E+07, 1.3904E+07, 1.3935E+07,
     +  1.3966E+07, 1.3996E+07, 1.4027E+07, 1.4058E+07, 1.4089E+07,
     + 1.4120E+07, 1.4150E+07, 1.4181E+07, 1.4213E+07, 1.4244E+07,
     +  1.4275E+07, 1.4306E+07, 1.4337E+07, 1.4369E+07, 1.4400E+07,
     + 1.4431E+07, 1.4463E+07, 1.4495E+07, 1.4526E+07, 1.4558E+07,
     +  1.4589E+07, 1.4621E+07, 1.4653E+07, 1.4685E+07, 1.4717E+07,
     + 1.4749E+07, 1.4781E+07, 1.4813E+07, 1.4845E+07, 1.4877E+07,
     +  1.4909E+07, 1.4942E+07, 1.4974E+07, 1.5007E+07, 1.5039E+07,
     + 1.5072E+07, 1.5104E+07, 1.5137E+07, 1.5169E+07, 1.5202E+07,
     +  1.5235E+07, 1.5268E+07, 1.5301E+07, 1.5334E+07, 1.5367E+07,
     + 1.5400E+07, 1.5433E+07, 1.5466E+07, 1.5499E+07, 1.5533E+07,
     +  1.5566E+07, 1.5600E+07, 1.5633E+07, 1.5666E+07, 1.5700E+07,
     + 1.5734E+07, 1.5767E+07, 1.5801E+07, 1.5835E+07, 1.5869E+07,
     +  1.5903E+07, 1.5937E+07, 1.5971E+07, 1.6005E+07, 1.6039E+07,
     + 1.6073E+07, 1.6108E+07, 1.6142E+07, 1.6176E+07, 1.6211E+07,
     +  1.6245E+07, 1.6280E+07, 1.6314E+07, 1.6349E+07, 1.6384E+07,
     + 1.6418E+07, 1.6453E+07, 1.6488E+07, 1.6523E+07, 1.6558E+07,
     +  1.6593E+07, 1.6628E+07, 1.6664E+07, 1.6699E+07, 1.6734E+07,
     + 1.6770E+07, 1.6805E+07, 1.6840E+07, 1.6876E+07, 1.6912E+07,
     +  1.6947E+07, 1.6983E+07, 1.7019E+07, 1.7055E+07, 1.7090E+07,
     + 1.7126E+07, 1.7162E+07, 1.7199E+07, 1.7235E+07, 1.7271E+07,
     +  1.7307E+07, 1.7343E+07, 1.7380E+07, 1.7416E+07, 1.7453E+07,
     + 1.7489E+07, 1.7526E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_C2H4  (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 2),QofT( 2,NT),Q(NT)
      data (Tmx(I),I=1,2)/3500.,3500./
      data xgj/ 1.,2./
c...     C2H4
c...        --       221
      data (QofT( 1,J),J=1,2408)/ 2.9640E+00, 7.0926E+00, 1.2184E+01,
     +  1.8126E+01, 2.4816E+01, 3.2177E+01, 4.0154E+01, 4.8702E+01,
     + 5.7786E+01, 6.7375E+01, 7.7445E+01, 8.7972E+01, 9.8939E+01,
     +  1.1033E+02, 1.2212E+02, 1.3431E+02, 1.4689E+02, 1.5983E+02,
     + 1.7313E+02, 1.8678E+02, 2.1510E+02, 2.4473E+02, 2.7560E+02,
     +  3.0768E+02, 3.4092E+02, 3.7527E+02, 4.1071E+02, 4.4720E+02,
     + 4.8472E+02, 5.2323E+02, 5.6272E+02, 6.0315E+02, 6.4451E+02,
     +  6.8677E+02, 7.2993E+02, 7.7395E+02, 8.1883E+02, 8.6454E+02,
     + 9.1108E+02, 9.5843E+02, 1.0066E+03, 1.0555E+03, 1.1052E+03,
     +  1.1557E+03, 1.2069E+03, 1.2588E+03, 1.3115E+03, 1.3649E+03,
     + 1.4189E+03, 1.4737E+03, 1.5292E+03, 1.5854E+03, 1.6422E+03,
     +  1.6997E+03, 1.7579E+03, 1.8167E+03, 1.8762E+03, 1.9363E+03,
     + 1.9971E+03, 2.0584E+03, 2.1204E+03, 2.1830E+03, 2.2462E+03,
     +  2.3100E+03, 2.3744E+03, 2.4395E+03, 2.5051E+03, 2.5712E+03,
     + 2.6380E+03, 2.7053E+03, 2.7732E+03, 2.8417E+03, 2.9108E+03,
     +  2.9804E+03, 3.0506E+03, 3.1213E+03, 3.1926E+03, 3.2645E+03,
     + 3.3369E+03, 3.4098E+03, 3.4833E+03, 3.5574E+03, 3.6320E+03,
     +  3.7071E+03, 3.7828E+03, 3.8590E+03, 3.9358E+03, 4.0131E+03,
     + 4.0910E+03, 4.1694E+03, 4.2483E+03, 4.3279E+03, 4.4079E+03,
     +  4.4885E+03, 4.5697E+03, 4.6514E+03, 4.7337E+03, 4.8165E+03,
     + 4.9000E+03, 4.9839E+03, 5.0684E+03, 5.1536E+03, 5.2392E+03,
     +  5.3255E+03, 5.4124E+03, 5.4998E+03, 5.5879E+03, 5.6765E+03,
     + 5.7658E+03, 5.8556E+03, 5.9461E+03, 6.0372E+03, 6.1289E+03,
     +  6.2212E+03, 6.3143E+03, 6.4079E+03, 6.5022E+03, 6.5972E+03,
     + 6.6928E+03, 6.7891E+03, 6.8860E+03, 6.9837E+03, 7.0821E+03,
     +  7.1812E+03, 7.2810E+03, 7.3815E+03, 7.4828E+03, 7.5847E+03,
     + 7.6875E+03, 7.7910E+03, 7.8953E+03, 8.0004E+03, 8.1063E+03,
     +  8.2129E+03, 8.3204E+03, 8.4286E+03, 8.5378E+03, 8.6477E+03,
     + 8.7585E+03, 8.8702E+03, 8.9827E+03, 9.0961E+03, 9.2105E+03,
     +  9.3257E+03, 9.4419E+03, 9.5590E+03, 9.6770E+03, 9.7960E+03,
     + 9.9160E+03, 1.0037E+04, 1.0159E+04, 1.0282E+04, 1.0406E+04,
     +  1.0531E+04, 1.0657E+04, 1.0784E+04, 1.0912E+04, 1.1042E+04,
     + 1.1172E+04, 1.1303E+04, 1.1436E+04, 1.1570E+04, 1.1705E+04,
     +  1.1841E+04, 1.1978E+04, 1.2117E+04, 1.2257E+04, 1.2398E+04,
     + 1.2540E+04, 1.2683E+04, 1.2828E+04, 1.2974E+04, 1.3122E+04,
     +  1.3270E+04, 1.3421E+04, 1.3572E+04, 1.3725E+04, 1.3879E+04,
     + 1.4035E+04, 1.4192E+04, 1.4350E+04, 1.4510E+04, 1.4672E+04,
     +  1.4835E+04, 1.4999E+04, 1.5166E+04, 1.5333E+04, 1.5502E+04,
     + 1.5673E+04, 1.5846E+04, 1.6020E+04, 1.6195E+04, 1.6373E+04,
     +  1.6552E+04, 1.6733E+04, 1.6915E+04, 1.7099E+04, 1.7285E+04,
     + 1.7473E+04, 1.7663E+04, 1.7854E+04, 1.8048E+04, 1.8243E+04,
     +  1.8440E+04, 1.8639E+04, 1.8840E+04, 1.9043E+04, 1.9248E+04,
     + 1.9455E+04, 1.9664E+04, 1.9875E+04, 2.0088E+04, 2.0303E+04,
     +  2.0521E+04, 2.0740E+04, 2.0962E+04, 2.1186E+04, 2.1412E+04,
     + 2.1640E+04, 2.1871E+04, 2.2104E+04, 2.2339E+04, 2.2576E+04,
     +  2.2816E+04, 2.3059E+04, 2.3303E+04, 2.3551E+04, 2.3800E+04,
     + 2.4052E+04, 2.4307E+04, 2.4564E+04, 2.4824E+04, 2.5087E+04,
     +  2.5352E+04, 2.5620E+04, 2.5890E+04, 2.6163E+04, 2.6439E+04,
     + 2.6718E+04, 2.7000E+04, 2.7284E+04, 2.7571E+04, 2.7862E+04,
     +  2.8155E+04, 2.8451E+04, 2.8750E+04, 2.9052E+04, 2.9358E+04,
     + 2.9666E+04, 2.9977E+04, 3.0292E+04, 3.0610E+04, 3.0931E+04,
     +  3.1255E+04, 3.1583E+04, 3.1914E+04, 3.2248E+04, 3.2586E+04,
     + 3.2927E+04, 3.3272E+04, 3.3620E+04, 3.3971E+04, 3.4327E+04,
     +  3.4686E+04, 3.5048E+04, 3.5415E+04, 3.5785E+04, 3.6158E+04,
     + 3.6536E+04, 3.6917E+04, 3.7303E+04, 3.7692E+04, 3.8085E+04,
     +  3.8483E+04, 3.8884E+04, 3.9289E+04, 3.9699E+04, 4.0113E+04,
     + 4.0531E+04, 4.0953E+04, 4.1379E+04, 4.1810E+04, 4.2245E+04,
     +  4.2685E+04, 4.3130E+04, 4.3578E+04, 4.4032E+04, 4.4490E+04,
     + 4.4952E+04, 4.5420E+04, 4.5892E+04, 4.6369E+04, 4.6850E+04,
     +  4.7337E+04, 4.7829E+04, 4.8326E+04, 4.8827E+04, 4.9334E+04,
     + 4.9846E+04, 5.0364E+04, 5.0886E+04, 5.1414E+04, 5.1947E+04,
     +  5.2486E+04, 5.3030E+04, 5.3580E+04, 5.4135E+04, 5.4696E+04,
     + 5.5263E+04, 5.5835E+04, 5.6414E+04, 5.6998E+04, 5.7588E+04,
     +  5.8184E+04, 5.8786E+04, 5.9395E+04, 6.0009E+04, 6.0630E+04,
     + 6.1257E+04, 6.1890E+04, 6.2530E+04, 6.3176E+04, 6.3829E+04,
     +  6.4489E+04, 6.5155E+04, 6.5828E+04, 6.6508E+04, 6.7194E+04,
     + 6.7888E+04, 6.8589E+04, 6.9297E+04, 7.0011E+04, 7.0734E+04,
     +  7.1463E+04, 7.2200E+04, 7.2944E+04, 7.3696E+04, 7.4455E+04,
     + 7.5222E+04, 7.5997E+04, 7.6780E+04, 7.7571E+04, 7.8369E+04,
     +  7.9176E+04, 7.9991E+04, 8.0814E+04, 8.1645E+04, 8.2485E+04,
     + 8.3333E+04, 8.4189E+04, 8.5055E+04, 8.5929E+04, 8.6812E+04,
     +  8.7703E+04, 8.8604E+04, 8.9514E+04, 9.0432E+04, 9.1360E+04,
     + 9.2297E+04, 9.3244E+04, 9.4201E+04, 9.5167E+04, 9.6142E+04,
     +  9.7127E+04, 9.8123E+04, 9.9128E+04, 1.0014E+05, 1.0117E+05,
     + 1.0220E+05, 1.0325E+05, 1.0431E+05, 1.0537E+05, 1.0645E+05,
     +  1.0754E+05, 1.0864E+05, 1.0975E+05, 1.1087E+05, 1.1200E+05,
     + 1.1315E+05, 1.1430E+05, 1.1547E+05, 1.1665E+05, 1.1784E+05,
     +  1.1904E+05, 1.2025E+05, 1.2148E+05, 1.2272E+05, 1.2397E+05,
     + 1.2523E+05, 1.2651E+05, 1.2780E+05, 1.2910E+05, 1.3041E+05,
     +  1.3174E+05, 1.3308E+05, 1.3443E+05, 1.3580E+05, 1.3718E+05,
     + 1.3857E+05, 1.3998E+05, 1.4140E+05, 1.4284E+05, 1.4429E+05,
     +  1.4575E+05, 1.4723E+05, 1.4872E+05, 1.5023E+05, 1.5175E+05,
     + 1.5329E+05, 1.5484E+05, 1.5641E+05, 1.5799E+05, 1.5959E+05,
     +  1.6121E+05, 1.6284E+05, 1.6449E+05, 1.6615E+05, 1.6783E+05,
     + 1.6952E+05, 1.7124E+05, 1.7297E+05, 1.7471E+05, 1.7647E+05,
     +  1.7826E+05, 1.8005E+05, 1.8187E+05, 1.8370E+05, 1.8555E+05,
     + 1.8742E+05, 1.8931E+05, 1.9122E+05, 1.9314E+05, 1.9508E+05,
     +  1.9705E+05, 1.9903E+05, 2.0103E+05, 2.0305E+05, 2.0509E+05,
     + 2.0715E+05, 2.0923E+05, 2.1133E+05, 2.1345E+05, 2.1559E+05,
     +  2.1775E+05, 2.1993E+05, 2.2214E+05, 2.2436E+05, 2.2661E+05,
     + 2.2888E+05, 2.3117E+05, 2.3348E+05, 2.3582E+05, 2.3817E+05,
     +  2.4055E+05, 2.4296E+05, 2.4539E+05, 2.4784E+05, 2.5031E+05,
     + 2.5281E+05, 2.5533E+05, 2.5787E+05, 2.6045E+05, 2.6304E+05,
     +  2.6566E+05, 2.6831E+05, 2.7098E+05, 2.7367E+05, 2.7640E+05,
     + 2.7914E+05, 2.8192E+05, 2.8472E+05, 2.8755E+05, 2.9041E+05,
     +  2.9329E+05, 2.9620E+05, 2.9914E+05, 3.0210E+05, 3.0510E+05,
     + 3.0812E+05, 3.1117E+05, 3.1426E+05, 3.1737E+05, 3.2051E+05,
     +  3.2368E+05, 3.2688E+05, 3.3011E+05, 3.3337E+05, 3.3666E+05,
     + 3.3999E+05, 3.4334E+05, 3.4673E+05, 3.5015E+05, 3.5360E+05,
     +  3.5709E+05, 3.6061E+05, 3.6416E+05, 3.6775E+05, 3.7136E+05,
     + 3.7502E+05, 3.7871E+05, 3.8243E+05, 3.8619E+05, 3.8998E+05,
     +  3.9381E+05, 3.9768E+05, 4.0158E+05, 4.0552E+05, 4.0949E+05,
     + 4.1351E+05, 4.1756E+05, 4.2165E+05, 4.2578E+05, 4.2994E+05,
     +  4.3415E+05, 4.3840E+05, 4.4268E+05, 4.4701E+05, 4.5137E+05,
     + 4.5578E+05, 4.6023E+05, 4.6472E+05, 4.6925E+05, 4.7383E+05,
     +  4.7844E+05, 4.8311E+05, 4.8781E+05, 4.9256E+05, 4.9735E+05,
     + 5.0219E+05, 5.0707E+05, 5.1200E+05, 5.1697E+05, 5.2199E+05,
     +  5.2706E+05, 5.3218E+05, 5.3734E+05, 5.4255E+05, 5.4780E+05,
     + 5.5311E+05, 5.5847E+05, 5.6387E+05, 5.6933E+05, 5.7484E+05,
     +  5.8040E+05, 5.8600E+05, 5.9167E+05, 5.9738E+05, 6.0315E+05,
     + 6.0897E+05, 6.1484E+05, 6.2077E+05, 6.2675E+05, 6.3279E+05,
     +  6.3888E+05, 6.4503E+05, 6.5124E+05, 6.5750E+05, 6.6382E+05,
     + 6.7020E+05, 6.7664E+05, 6.8313E+05, 6.8969E+05, 6.9631E+05,
     +  7.0299E+05, 7.0973E+05, 7.1653E+05, 7.2339E+05, 7.3031E+05,
     + 7.3730E+05, 7.4436E+05, 7.5147E+05, 7.5866E+05, 7.6591E+05,
     +  7.7322E+05, 7.8060E+05, 7.8805E+05, 7.9557E+05, 8.0315E+05,
     + 8.1081E+05, 8.1853E+05, 8.2632E+05, 8.3419E+05, 8.4212E+05,
     +  8.5013E+05, 8.5821E+05, 8.6637E+05, 8.7460E+05, 8.8290E+05,
     + 8.9128E+05, 8.9973E+05, 9.0826E+05, 9.1687E+05, 9.2556E+05,
     +  9.3432E+05, 9.4317E+05, 9.5209E+05, 9.6110E+05, 9.7018E+05,
     + 9.7935E+05, 9.8860E+05, 9.9793E+05, 1.0073E+06, 1.0168E+06,
     +  1.0264E+06, 1.0361E+06, 1.0459E+06, 1.0557E+06, 1.0657E+06,
     + 1.0757E+06, 1.0858E+06, 1.0960E+06, 1.1063E+06, 1.1167E+06,
     +  1.1272E+06, 1.1378E+06, 1.1484E+06, 1.1592E+06, 1.1701E+06,
     + 1.1810E+06, 1.1921E+06, 1.2032E+06, 1.2145E+06, 1.2258E+06,
     +  1.2373E+06, 1.2489E+06, 1.2605E+06, 1.2723E+06, 1.2842E+06,
     + 1.2961E+06, 1.3082E+06, 1.3204E+06, 1.3327E+06, 1.3451E+06,
     +  1.3576E+06, 1.3703E+06, 1.3830E+06, 1.3958E+06, 1.4088E+06,
     + 1.4219E+06, 1.4351E+06, 1.4484E+06, 1.4618E+06, 1.4754E+06,
     +  1.4890E+06, 1.5028E+06, 1.5167E+06, 1.5308E+06, 1.5449E+06,
     + 1.5592E+06, 1.5736E+06, 1.5881E+06, 1.6028E+06, 1.6176E+06,
     +  1.6325E+06, 1.6476E+06, 1.6627E+06, 1.6780E+06, 1.6935E+06,
     + 1.7091E+06, 1.7248E+06, 1.7406E+06, 1.7566E+06, 1.7728E+06,
     +  1.7891E+06, 1.8055E+06, 1.8220E+06, 1.8387E+06, 1.8556E+06,
     + 1.8726E+06, 1.8897E+06, 1.9070E+06, 1.9245E+06, 1.9421E+06,
     +  1.9598E+06, 1.9777E+06, 1.9958E+06, 2.0140E+06, 2.0323E+06,
     + 2.0509E+06, 2.0696E+06, 2.0884E+06, 2.1074E+06, 2.1266E+06,
     +  2.1460E+06, 2.1655E+06, 2.1851E+06, 2.2050E+06, 2.2250E+06,
     + 2.2452E+06, 2.2656E+06, 2.2861E+06, 2.3068E+06, 2.3277E+06,
     +  2.3488E+06, 2.3701E+06, 2.3915E+06, 2.4131E+06, 2.4349E+06,
     + 2.4569E+06, 2.4791E+06, 2.5015E+06, 2.5241E+06, 2.5468E+06,
     +  2.5698E+06, 2.5929E+06, 2.6163E+06, 2.6398E+06, 2.6636E+06,
     + 2.6875E+06, 2.7117E+06, 2.7360E+06, 2.7606E+06, 2.7854E+06,
     +  2.8104E+06, 2.8356E+06, 2.8610E+06, 2.8866E+06, 2.9125E+06,
     + 2.9385E+06, 2.9648E+06, 2.9913E+06, 3.0181E+06, 3.0450E+06,
     +  3.0722E+06, 3.0996E+06, 3.1273E+06, 3.1552E+06, 3.1833E+06,
     + 3.2116E+06, 3.2402E+06, 3.2691E+06, 3.2982E+06, 3.3275E+06,
     +  3.3570E+06, 3.3869E+06, 3.4169E+06, 3.4473E+06, 3.4778E+06,
     + 3.5087E+06, 3.5398E+06, 3.5711E+06, 3.6027E+06, 3.6346E+06,
     +  3.6668E+06, 3.6992E+06, 3.7319E+06, 3.7648E+06, 3.7981E+06,
     + 3.8316E+06, 3.8654E+06, 3.8995E+06, 3.9338E+06, 3.9685E+06,
     +  4.0034E+06, 4.0386E+06, 4.0742E+06, 4.1100E+06, 4.1461E+06,
     + 4.1825E+06, 4.2192E+06, 4.2562E+06, 4.2936E+06, 4.3312E+06,
     +  4.3692E+06, 4.4074E+06, 4.4460E+06, 4.4849E+06, 4.5241E+06,
     + 4.5637E+06, 4.6035E+06, 4.6437E+06, 4.6843E+06, 4.7251E+06,
     +  4.7663E+06, 4.8079E+06, 4.8498E+06, 4.8920E+06, 4.9346E+06,
     + 4.9775E+06, 5.0208E+06, 5.0644E+06, 5.1084E+06, 5.1528E+06,
     +  5.1975E+06, 5.2426E+06, 5.2880E+06, 5.3339E+06, 5.3801E+06,
     + 5.4266E+06, 5.4736E+06, 5.5210E+06, 5.5687E+06, 5.6168E+06,
     +  5.6653E+06, 5.7142E+06, 5.7635E+06, 5.8132E+06, 5.8634E+06,
     + 5.9139E+06, 5.9648E+06, 6.0162E+06, 6.0679E+06, 6.1201E+06,
     +  6.1727E+06, 6.2258E+06, 6.2792E+06, 6.3331E+06, 6.3875E+06,
     + 6.4422E+06, 6.4975E+06, 6.5531E+06, 6.6092E+06, 6.6658E+06,
     +  6.7228E+06, 6.7803E+06, 6.8383E+06, 6.8967E+06, 6.9556E+06,
     + 7.0149E+06, 7.0748E+06, 7.1351E+06, 7.1959E+06, 7.2572E+06,
     +  7.3190E+06, 7.3813E+06, 7.4440E+06, 7.5073E+06, 7.5711E+06,
     + 7.6354E+06, 7.7002E+06, 7.7656E+06, 7.8314E+06, 7.8978E+06,
     +  7.9647E+06, 8.0322E+06, 8.1002E+06, 8.1687E+06, 8.2378E+06,
     + 8.3074E+06, 8.3776E+06, 8.4484E+06, 8.5197E+06, 8.5916E+06,
     +  8.6640E+06, 8.7370E+06, 8.8106E+06, 8.8848E+06, 8.9596E+06,
     + 9.0350E+06, 9.1109E+06, 9.1875E+06, 9.2647E+06, 9.3424E+06,
     +  9.4208E+06, 9.4999E+06, 9.5795E+06, 9.6598E+06, 9.7407E+06,
     + 9.8222E+06, 9.9044E+06, 9.9873E+06, 1.0071E+07, 1.0155E+07,
     +  1.0240E+07, 1.0325E+07, 1.0411E+07, 1.0498E+07, 1.0586E+07,
     + 1.0674E+07, 1.0763E+07, 1.0852E+07, 1.0943E+07, 1.1034E+07,
     +  1.1125E+07, 1.1218E+07, 1.1311E+07, 1.1405E+07, 1.1499E+07,
     + 1.1595E+07, 1.1691E+07, 1.1788E+07, 1.1885E+07, 1.1984E+07,
     +  1.2083E+07, 1.2183E+07, 1.2283E+07, 1.2385E+07, 1.2487E+07,
     + 1.2590E+07, 1.2694E+07, 1.2798E+07, 1.2904E+07, 1.3010E+07,
     +  1.3117E+07, 1.3225E+07, 1.3334E+07, 1.3443E+07, 1.3554E+07,
     + 1.3665E+07, 1.3777E+07, 1.3890E+07, 1.4004E+07, 1.4119E+07,
     +  1.4235E+07, 1.4351E+07, 1.4469E+07, 1.4587E+07, 1.4706E+07,
     + 1.4827E+07, 1.4948E+07, 1.5070E+07, 1.5193E+07, 1.5317E+07,
     +  1.5441E+07, 1.5567E+07, 1.5694E+07, 1.5822E+07, 1.5951E+07,
     + 1.6080E+07, 1.6211E+07, 1.6343E+07, 1.6475E+07, 1.6609E+07,
     +  1.6744E+07, 1.6880E+07, 1.7017E+07, 1.7154E+07, 1.7293E+07,
     + 1.7433E+07, 1.7574E+07, 1.7716E+07, 1.7860E+07, 1.8004E+07,
     +  1.8149E+07, 1.8296E+07, 1.8443E+07, 1.8592E+07, 1.8742E+07,
     + 1.8893E+07, 1.9045E+07, 1.9198E+07, 1.9353E+07, 1.9508E+07,
     +  1.9665E+07, 1.9823E+07, 1.9982E+07, 2.0142E+07, 2.0304E+07,
     + 2.0467E+07, 2.0631E+07, 2.0796E+07, 2.0962E+07, 2.1130E+07,
     +  2.1299E+07, 2.1469E+07, 2.1641E+07, 2.1813E+07, 2.1988E+07,
     + 2.2163E+07, 2.2340E+07, 2.2518E+07, 2.2697E+07, 2.2878E+07,
     +  2.3060E+07, 2.3243E+07, 2.3428E+07, 2.3614E+07, 2.3802E+07,
     + 2.3991E+07, 2.4181E+07, 2.4373E+07, 2.4566E+07, 2.4761E+07,
     +  2.4957E+07, 2.5155E+07, 2.5354E+07, 2.5554E+07, 2.5756E+07,
     + 2.5960E+07, 2.6165E+07, 2.6371E+07, 2.6579E+07, 2.6789E+07,
     +  2.7000E+07, 2.7213E+07, 2.7427E+07, 2.7643E+07, 2.7860E+07,
     + 2.8080E+07, 2.8300E+07, 2.8523E+07, 2.8747E+07, 2.8972E+07,
     +  2.9199E+07, 2.9428E+07, 2.9659E+07, 2.9891E+07, 3.0125E+07,
     + 3.0361E+07, 3.0599E+07, 3.0838E+07, 3.1079E+07, 3.1322E+07,
     +  3.1566E+07, 3.1812E+07, 3.2061E+07, 3.2310E+07, 3.2562E+07,
     + 3.2816E+07, 3.3071E+07, 3.3329E+07, 3.3588E+07, 3.3849E+07,
     +  3.4112E+07, 3.4377E+07, 3.4644E+07, 3.4913E+07, 3.5183E+07,
     + 3.5456E+07, 3.5731E+07, 3.6007E+07, 3.6286E+07, 3.6567E+07,
     +  3.6850E+07, 3.7134E+07, 3.7421E+07, 3.7710E+07, 3.8001E+07,
     + 3.8295E+07, 3.8590E+07, 3.8887E+07, 3.9187E+07, 3.9489E+07,
     +  3.9792E+07, 4.0099E+07, 4.0407E+07, 4.0717E+07, 4.1030E+07,
     + 4.1345E+07, 4.1662E+07, 4.1982E+07, 4.2304E+07, 4.2628E+07,
     +  4.2955E+07, 4.3283E+07, 4.3615E+07, 4.3948E+07, 4.4284E+07,
     + 4.4623E+07, 4.4964E+07, 4.5307E+07, 4.5652E+07, 4.6001E+07,
     +  4.6351E+07, 4.6705E+07, 4.7060E+07, 4.7418E+07, 4.7779E+07,
     + 4.8143E+07, 4.8509E+07, 4.8877E+07, 4.9248E+07, 4.9622E+07,
     +  4.9999E+07, 5.0378E+07, 5.0760E+07, 5.1144E+07, 5.1532E+07,
     + 5.1922E+07, 5.2315E+07, 5.2710E+07, 5.3109E+07, 5.3510E+07,
     +  5.3914E+07, 5.4321E+07, 5.4731E+07, 5.5143E+07, 5.5559E+07,
     + 5.5977E+07, 5.6399E+07, 5.6823E+07, 5.7251E+07, 5.7681E+07,
     +  5.8115E+07, 5.8552E+07, 5.8991E+07, 5.9434E+07, 5.9880E+07,
     + 6.0329E+07, 6.0781E+07, 6.1236E+07, 6.1695E+07, 6.2156E+07,
     +  6.2621E+07, 6.3090E+07, 6.3561E+07, 6.4036E+07, 6.4514E+07,
     + 6.4995E+07, 6.5480E+07, 6.5968E+07, 6.6460E+07, 6.6955E+07,
     +  6.7454E+07, 6.7956E+07, 6.8461E+07, 6.8970E+07, 6.9483E+07,
     + 6.9999E+07, 7.0518E+07, 7.1042E+07, 7.1569E+07, 7.2099E+07,
     +  7.2634E+07, 7.3172E+07, 7.3713E+07, 7.4259E+07, 7.4808E+07,
     + 7.5361E+07, 7.5918E+07, 7.6479E+07, 7.7043E+07, 7.7612E+07,
     +  7.8184E+07, 7.8761E+07, 7.9341E+07, 7.9925E+07, 8.0514E+07,
     + 8.1106E+07, 8.1703E+07, 8.2304E+07, 8.2908E+07, 8.3517E+07,
     +  8.4130E+07, 8.4748E+07, 8.5369E+07, 8.5995E+07, 8.6625E+07,
     + 8.7260E+07, 8.7899E+07, 8.8542E+07, 8.9189E+07, 8.9841E+07,
     +  9.0498E+07, 9.1159E+07, 9.1824E+07, 9.2494E+07, 9.3169E+07,
     + 9.3848E+07, 9.4532E+07, 9.5221E+07, 9.5914E+07, 9.6612E+07,
     +  9.7315E+07, 9.8022E+07, 9.8734E+07, 9.9451E+07, 1.0017E+08,
     + 1.0090E+08, 1.0163E+08, 1.0237E+08, 1.0311E+08, 1.0386E+08,
     +  1.0461E+08, 1.0537E+08, 1.0613E+08, 1.0690E+08, 1.0767E+08,
     + 1.0845E+08, 1.0923E+08, 1.1002E+08, 1.1081E+08, 1.1161E+08,
     +  1.1241E+08, 1.1322E+08, 1.1404E+08, 1.1486E+08, 1.1569E+08,
     + 1.1652E+08, 1.1735E+08, 1.1820E+08, 1.1905E+08, 1.1990E+08,
     +  1.2076E+08, 1.2163E+08, 1.2250E+08, 1.2338E+08, 1.2426E+08,
     + 1.2515E+08, 1.2604E+08, 1.2694E+08, 1.2785E+08, 1.2876E+08,
     +  1.2968E+08, 1.3061E+08, 1.3154E+08, 1.3248E+08, 1.3342E+08,
     + 1.3437E+08, 1.3533E+08, 1.3629E+08, 1.3726E+08, 1.3824E+08,
     +  1.3922E+08, 1.4021E+08, 1.4121E+08, 1.4221E+08, 1.4322E+08,
     + 1.4423E+08, 1.4525E+08, 1.4628E+08, 1.4732E+08, 1.4836E+08,
     +  1.4941E+08, 1.5047E+08, 1.5153E+08, 1.5260E+08, 1.5368E+08,
     + 1.5477E+08, 1.5586E+08, 1.5696E+08, 1.5806E+08, 1.5918E+08,
     +  1.6030E+08, 1.6143E+08, 1.6256E+08, 1.6370E+08, 1.6486E+08,
     + 1.6601E+08, 1.6718E+08, 1.6835E+08, 1.6953E+08, 1.7072E+08,
     +  1.7192E+08, 1.7312E+08, 1.7434E+08, 1.7556E+08, 1.7679E+08,
     + 1.7802E+08, 1.7927E+08, 1.8052E+08, 1.8178E+08, 1.8305E+08,
     +  1.8433E+08, 1.8561E+08, 1.8691E+08, 1.8821E+08, 1.8952E+08,
     + 1.9084E+08, 1.9217E+08, 1.9351E+08, 1.9485E+08, 1.9621E+08,
     +  1.9757E+08, 1.9894E+08, 2.0032E+08, 2.0171E+08, 2.0311E+08,
     + 2.0452E+08, 2.0593E+08, 2.0736E+08, 2.0880E+08, 2.1024E+08,
     +  2.1169E+08, 2.1316E+08, 2.1463E+08, 2.1611E+08, 2.1760E+08,
     + 2.1911E+08, 2.2062E+08, 2.2214E+08, 2.2367E+08, 2.2521E+08,
     +  2.2676E+08, 2.2832E+08, 2.2989E+08, 2.3147E+08, 2.3306E+08,
     + 2.3466E+08, 2.3627E+08, 2.3789E+08, 2.3953E+08, 2.4117E+08,
     +  2.4282E+08, 2.4448E+08, 2.4616E+08, 2.4784E+08, 2.4954E+08,
     + 2.5125E+08, 2.5296E+08, 2.5469E+08, 2.5643E+08, 2.5818E+08,
     +  2.5994E+08, 2.6171E+08, 2.6350E+08, 2.6529E+08, 2.6710E+08,
     + 2.6892E+08, 2.7075E+08, 2.7259E+08, 2.7444E+08, 2.7631E+08,
     +  2.7819E+08, 2.8007E+08, 2.8197E+08, 2.8389E+08, 2.8581E+08,
     + 2.8775E+08, 2.8970E+08, 2.9166E+08, 2.9363E+08, 2.9562E+08,
     +  2.9762E+08, 2.9963E+08, 3.0166E+08, 3.0369E+08, 3.0574E+08,
     + 3.0781E+08, 3.0988E+08, 3.1197E+08, 3.1407E+08, 3.1619E+08,
     +  3.1832E+08, 3.2046E+08, 3.2261E+08, 3.2478E+08, 3.2696E+08,
     + 3.2916E+08, 3.3137E+08, 3.3360E+08, 3.3583E+08, 3.3808E+08,
     +  3.4035E+08, 3.4263E+08, 3.4493E+08, 3.4723E+08, 3.4956E+08,
     + 3.5189E+08, 3.5425E+08, 3.5661E+08, 3.5900E+08, 3.6139E+08,
     +  3.6380E+08, 3.6623E+08, 3.6867E+08, 3.7113E+08, 3.7360E+08,
     + 3.7609E+08, 3.7859E+08, 3.8111E+08, 3.8364E+08, 3.8619E+08,
     +  3.8876E+08, 3.9134E+08, 3.9394E+08, 3.9655E+08, 3.9918E+08,
     + 4.0183E+08, 4.0449E+08, 4.0717E+08, 4.0986E+08, 4.1258E+08,
     +  4.1530E+08, 4.1805E+08, 4.2081E+08, 4.2359E+08, 4.2639E+08,
     + 4.2920E+08, 4.3203E+08, 4.3488E+08, 4.3775E+08, 4.4063E+08,
     +  4.4354E+08, 4.4645E+08, 4.4939E+08, 4.5235E+08, 4.5532E+08,
     + 4.5831E+08, 4.6132E+08, 4.6435E+08, 4.6740E+08, 4.7046E+08,
     +  4.7355E+08, 4.7665E+08, 4.7978E+08, 4.8292E+08, 4.8608E+08,
     + 4.8926E+08, 4.9246E+08, 4.9568E+08, 4.9891E+08, 5.0217E+08,
     +  5.0545E+08, 5.0875E+08, 5.1207E+08, 5.1541E+08, 5.1876E+08,
     + 5.2214E+08, 5.2554E+08, 5.2896E+08, 5.3241E+08, 5.3587E+08,
     +  5.3935E+08, 5.4285E+08, 5.4638E+08, 5.4993E+08, 5.5349E+08,
     + 5.5708E+08, 5.6070E+08, 5.6433E+08, 5.6798E+08, 5.7166E+08,
     +  5.7536E+08, 5.7908E+08, 5.8283E+08, 5.8659E+08, 5.9038E+08,
     + 5.9420E+08, 5.9803E+08, 6.0189E+08, 6.0577E+08, 6.0968E+08,
     +  6.1361E+08, 6.1756E+08, 6.2153E+08, 6.2553E+08, 6.2956E+08,
     + 6.3361E+08, 6.3768E+08, 6.4177E+08, 6.4590E+08, 6.5004E+08,
     +  6.5421E+08, 6.5841E+08, 6.6263E+08, 6.6688E+08, 6.7115E+08,
     + 6.7544E+08, 6.7977E+08, 6.8411E+08, 6.8849E+08, 6.9289E+08,
     +  6.9732E+08, 7.0177E+08, 7.0625E+08, 7.1075E+08, 7.1529E+08,
     + 7.1985E+08, 7.2443E+08, 7.2905E+08, 7.3369E+08, 7.3836E+08,
     +  7.4306E+08, 7.4778E+08, 7.5253E+08, 7.5731E+08, 7.6212E+08,
     + 7.6696E+08, 7.7183E+08, 7.7672E+08, 7.8165E+08, 7.8660E+08,
     +  7.9158E+08, 7.9660E+08, 8.0164E+08, 8.0671E+08, 8.1181E+08,
     + 8.1694E+08, 8.2210E+08, 8.2729E+08, 8.3252E+08, 8.3777E+08,
     +  8.4305E+08, 8.4837E+08, 8.5372E+08, 8.5909E+08, 8.6450E+08,
     + 8.6994E+08, 8.7542E+08, 8.8092E+08, 8.8646E+08, 8.9203E+08,
     +  8.9763E+08, 9.0327E+08, 9.0894E+08, 9.1464E+08, 9.2037E+08,
     + 9.2614E+08, 9.3194E+08, 9.3778E+08, 9.4365E+08, 9.4955E+08,
     +  9.5549E+08, 9.6147E+08, 9.6747E+08, 9.7352E+08, 9.7960E+08,
     + 9.8571E+08, 9.9186E+08, 9.9804E+08, 1.0043E+09, 1.0105E+09,
     +  1.0168E+09, 1.0231E+09, 1.0295E+09, 1.0359E+09, 1.0424E+09,
     + 1.0488E+09, 1.0554E+09, 1.0619E+09, 1.0685E+09, 1.0751E+09,
     +  1.0818E+09, 1.0885E+09, 1.0952E+09, 1.1020E+09, 1.1089E+09,
     + 1.1157E+09, 1.1226E+09, 1.1296E+09, 1.1365E+09, 1.1436E+09,
     +  1.1506E+09, 1.1577E+09, 1.1649E+09, 1.1721E+09, 1.1793E+09,
     + 1.1866E+09, 1.1939E+09, 1.2012E+09, 1.2086E+09, 1.2160E+09,
     +  1.2235E+09, 1.2310E+09, 1.2386E+09, 1.2462E+09, 1.2539E+09,
     + 1.2616E+09, 1.2693E+09, 1.2771E+09, 1.2849E+09, 1.2928E+09,
     +  1.3007E+09, 1.3087E+09, 1.3167E+09, 1.3247E+09, 1.3328E+09,
     + 1.3410E+09, 1.3492E+09, 1.3574E+09, 1.3657E+09, 1.3740E+09,
     +  1.3824E+09, 1.3908E+09, 1.3993E+09, 1.4078E+09, 1.4164E+09,
     + 1.4250E+09, 1.4337E+09, 1.4424E+09, 1.4512E+09, 1.4600E+09,
     +  1.4689E+09, 1.4778E+09, 1.4868E+09, 1.4958E+09, 1.5049E+09,
     + 1.5140E+09, 1.5232E+09, 1.5324E+09, 1.5417E+09, 1.5510E+09,
     +  1.5604E+09, 1.5698E+09, 1.5793E+09, 1.5888E+09, 1.5984E+09,
     + 1.6081E+09, 1.6178E+09, 1.6275E+09, 1.6374E+09, 1.6472E+09,
     +  1.6572E+09, 1.6671E+09, 1.6772E+09, 1.6873E+09, 1.6974E+09,
     + 1.7076E+09, 1.7179E+09, 1.7282E+09, 1.7386E+09, 1.7490E+09,
     +  1.7595E+09, 1.7701E+09, 1.7807E+09, 1.7913E+09, 1.8021E+09,
     + 1.8129E+09, 1.8237E+09, 1.8346E+09, 1.8456E+09, 1.8566E+09,
     +  1.8677E+09, 1.8789E+09, 1.8901E+09, 1.9014E+09, 1.9127E+09,
     + 1.9241E+09, 1.9356E+09, 1.9471E+09, 1.9587E+09, 1.9704E+09,
     +  1.9821E+09, 1.9939E+09, 2.0058E+09, 2.0177E+09, 2.0297E+09,
     + 2.0417E+09, 2.0538E+09, 2.0660E+09, 2.0783E+09, 2.0906E+09,
     +  2.1030E+09, 2.1155E+09, 2.1280E+09, 2.1406E+09, 2.1533E+09,
     + 2.1660E+09, 2.1788E+09, 2.1917E+09, 2.2046E+09, 2.2177E+09,
     +  2.2307E+09, 2.2439E+09, 2.2571E+09, 2.2704E+09, 2.2838E+09,
     + 2.2973E+09, 2.3108E+09, 2.3244E+09, 2.3381E+09, 2.3518E+09,
     +  2.3657E+09, 2.3796E+09, 2.3936E+09, 2.4076E+09, 2.4217E+09,
     + 2.4360E+09, 2.4502E+09, 2.4646E+09, 2.4791E+09, 2.4936E+09,
     +  2.5082E+09, 2.5229E+09, 2.5376E+09, 2.5525E+09, 2.5674E+09,
     + 2.5824E+09, 2.5975E+09, 2.6127E+09, 2.6279E+09, 2.6432E+09,
     +  2.6587E+09, 2.6742E+09, 2.6897E+09, 2.7054E+09, 2.7212E+09,
     + 2.7370E+09, 2.7529E+09, 2.7689E+09, 2.7850E+09, 2.8012E+09,
     +  2.8175E+09, 2.8339E+09, 2.8503E+09, 2.8669E+09, 2.8835E+09,
     + 2.9002E+09, 2.9170E+09, 2.9339E+09, 2.9509E+09, 2.9680E+09,
     +  2.9852E+09, 3.0024E+09, 3.0198E+09, 3.0372E+09, 3.0548E+09,
     + 3.0724E+09, 3.0902E+09, 3.1080E+09, 3.1259E+09, 3.1439E+09,
     +  3.1621E+09, 3.1803E+09, 3.1986E+09, 3.2170E+09, 3.2355E+09,
     + 3.2541E+09, 3.2728E+09, 3.2917E+09, 3.3106E+09, 3.3296E+09,
     +  3.3487E+09, 3.3679E+09, 3.3872E+09, 3.4067E+09, 3.4262E+09,
     + 3.4458E+09, 3.4655E+09, 3.4854E+09, 3.5053E+09, 3.5254E+09,
     +  3.5455E+09, 3.5658E+09, 3.5862E+09, 3.6067E+09, 3.6273E+09,
     + 3.6480E+09, 3.6688E+09, 3.6897E+09, 3.7107E+09, 3.7319E+09,
     +  3.7531E+09, 3.7745E+09, 3.7960E+09, 3.8176E+09, 3.8393E+09,
     + 3.8611E+09, 3.8831E+09, 3.9051E+09, 3.9273E+09, 3.9496E+09,
     +  3.9720E+09, 3.9945E+09, 4.0172E+09, 4.0399E+09, 4.0628E+09,
     + 4.0858E+09, 4.1089E+09, 4.1322E+09, 4.1556E+09, 4.1791E+09,
     +  4.2027E+09, 4.2264E+09, 4.2503E+09, 4.2743E+09, 4.2984E+09,
     + 4.3226E+09, 4.3470E+09, 4.3715E+09, 4.3961E+09, 4.4209E+09,
     +  4.4458E+09, 4.4708E+09, 4.4960E+09, 4.5212E+09, 4.5467E+09,
     + 4.5722E+09, 4.5979E+09, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       231
      data (QofT( 2,J),J=1,2408)/ 1.2046E+01, 2.8908E+01, 4.9729E+01,
     +  7.4038E+01, 1.0141E+02, 1.3153E+02, 1.6417E+02, 1.9915E+02,
     + 2.3632E+02, 2.7556E+02, 3.1677E+02, 3.5986E+02, 4.0474E+02,
     +  4.5135E+02, 4.9963E+02, 5.4952E+02, 6.0098E+02, 6.5395E+02,
     + 7.0839E+02, 7.6426E+02, 8.8018E+02, 1.0014E+03, 1.1278E+03,
     +  1.2591E+03, 1.3951E+03, 1.5358E+03, 1.6808E+03, 1.8302E+03,
     + 1.9837E+03, 2.1414E+03, 2.3030E+03, 2.4685E+03, 2.6378E+03,
     +  2.8108E+03, 2.9874E+03, 3.1676E+03, 3.3513E+03, 3.5385E+03,
     + 3.7289E+03, 3.9228E+03, 4.1198E+03, 4.3201E+03, 4.5235E+03,
     +  4.7300E+03, 4.9396E+03, 5.1522E+03, 5.3678E+03, 5.5864E+03,
     + 5.8078E+03, 6.0321E+03, 6.2592E+03, 6.4891E+03, 6.7218E+03,
     +  6.9572E+03, 7.1953E+03, 7.4361E+03, 7.6795E+03, 7.9256E+03,
     + 8.1742E+03, 8.4254E+03, 8.6791E+03, 8.9354E+03, 9.1941E+03,
     +  9.4552E+03, 9.7190E+03, 9.9851E+03, 1.0254E+04, 1.0525E+04,
     + 1.0798E+04, 1.1073E+04, 1.1351E+04, 1.1632E+04, 1.1914E+04,
     +  1.2199E+04, 1.2487E+04, 1.2776E+04, 1.3068E+04, 1.3362E+04,
     + 1.3659E+04, 1.3957E+04, 1.4258E+04, 1.4561E+04, 1.4866E+04,
     +  1.5174E+04, 1.5484E+04, 1.5796E+04, 1.6110E+04, 1.6427E+04,
     + 1.6746E+04, 1.7066E+04, 1.7390E+04, 1.7715E+04, 1.8043E+04,
     +  1.8373E+04, 1.8705E+04, 1.9040E+04, 1.9376E+04, 1.9715E+04,
     + 2.0057E+04, 2.0401E+04, 2.0747E+04, 2.1095E+04, 2.1446E+04,
     +  2.1799E+04, 2.2155E+04, 2.2512E+04, 2.2873E+04, 2.3236E+04,
     + 2.3601E+04, 2.3969E+04, 2.4339E+04, 2.4712E+04, 2.5087E+04,
     +  2.5465E+04, 2.5846E+04, 2.6229E+04, 2.6616E+04, 2.7004E+04,
     + 2.7396E+04, 2.7790E+04, 2.8187E+04, 2.8587E+04, 2.8989E+04,
     +  2.9395E+04, 2.9803E+04, 3.0215E+04, 3.0630E+04, 3.1047E+04,
     + 3.1468E+04, 3.1891E+04, 3.2318E+04, 3.2748E+04, 3.3182E+04,
     +  3.3618E+04, 3.4058E+04, 3.4501E+04, 3.4948E+04, 3.5398E+04,
     + 3.5851E+04, 3.6309E+04, 3.6769E+04, 3.7234E+04, 3.7702E+04,
     +  3.8173E+04, 3.8649E+04, 3.9128E+04, 3.9611E+04, 4.0098E+04,
     + 4.0590E+04, 4.1085E+04, 4.1584E+04, 4.2087E+04, 4.2595E+04,
     +  4.3107E+04, 4.3623E+04, 4.4143E+04, 4.4668E+04, 4.5197E+04,
     + 4.5731E+04, 4.6269E+04, 4.6812E+04, 4.7360E+04, 4.7912E+04,
     +  4.8470E+04, 4.9032E+04, 4.9599E+04, 5.0171E+04, 5.0748E+04,
     + 5.1330E+04, 5.1918E+04, 5.2511E+04, 5.3109E+04, 5.3712E+04,
     +  5.4321E+04, 5.4935E+04, 5.5555E+04, 5.6181E+04, 5.6812E+04,
     + 5.7450E+04, 5.8093E+04, 5.8741E+04, 5.9397E+04, 6.0058E+04,
     +  6.0725E+04, 6.1398E+04, 6.2078E+04, 6.2764E+04, 6.3457E+04,
     + 6.4156E+04, 6.4862E+04, 6.5574E+04, 6.6294E+04, 6.7019E+04,
     +  6.7753E+04, 6.8493E+04, 6.9240E+04, 6.9994E+04, 7.0756E+04,
     + 7.1525E+04, 7.2301E+04, 7.3085E+04, 7.3876E+04, 7.4676E+04,
     +  7.5482E+04, 7.6297E+04, 7.7120E+04, 7.7950E+04, 7.8789E+04,
     + 7.9637E+04, 8.0492E+04, 8.1356E+04, 8.2228E+04, 8.3109E+04,
     +  8.3999E+04, 8.4897E+04, 8.5805E+04, 8.6721E+04, 8.7647E+04,
     + 8.8581E+04, 8.9525E+04, 9.0478E+04, 9.1441E+04, 9.2414E+04,
     +  9.3396E+04, 9.4388E+04, 9.5389E+04, 9.6401E+04, 9.7423E+04,
     + 9.8456E+04, 9.9498E+04, 1.0055E+05, 1.0162E+05, 1.0269E+05,
     +  1.0378E+05, 1.0487E+05, 1.0598E+05, 1.0710E+05, 1.0823E+05,
     + 1.0937E+05, 1.1052E+05, 1.1169E+05, 1.1286E+05, 1.1405E+05,
     +  1.1525E+05, 1.1646E+05, 1.1769E+05, 1.1892E+05, 1.2017E+05,
     + 1.2143E+05, 1.2271E+05, 1.2400E+05, 1.2530E+05, 1.2661E+05,
     +  1.2794E+05, 1.2928E+05, 1.3064E+05, 1.3200E+05, 1.3339E+05,
     + 1.3478E+05, 1.3619E+05, 1.3762E+05, 1.3906E+05, 1.4051E+05,
     +  1.4198E+05, 1.4347E+05, 1.4497E+05, 1.4648E+05, 1.4801E+05,
     + 1.4956E+05, 1.5112E+05, 1.5270E+05, 1.5429E+05, 1.5590E+05,
     +  1.5752E+05, 1.5917E+05, 1.6083E+05, 1.6250E+05, 1.6420E+05,
     + 1.6591E+05, 1.6764E+05, 1.6938E+05, 1.7115E+05, 1.7293E+05,
     +  1.7473E+05, 1.7655E+05, 1.7838E+05, 1.8024E+05, 1.8211E+05,
     + 1.8401E+05, 1.8592E+05, 1.8785E+05, 1.8981E+05, 1.9178E+05,
     +  1.9377E+05, 1.9578E+05, 1.9782E+05, 1.9987E+05, 2.0195E+05,
     + 2.0404E+05, 2.0616E+05, 2.0830E+05, 2.1046E+05, 2.1264E+05,
     +  2.1485E+05, 2.1707E+05, 2.1933E+05, 2.2160E+05, 2.2389E+05,
     + 2.2621E+05, 2.2856E+05, 2.3092E+05, 2.3332E+05, 2.3573E+05,
     +  2.3817E+05, 2.4064E+05, 2.4313E+05, 2.4564E+05, 2.4818E+05,
     + 2.5075E+05, 2.5334E+05, 2.5596E+05, 2.5861E+05, 2.6128E+05,
     +  2.6398E+05, 2.6671E+05, 2.6946E+05, 2.7224E+05, 2.7505E+05,
     + 2.7789E+05, 2.8076E+05, 2.8366E+05, 2.8659E+05, 2.8954E+05,
     +  2.9253E+05, 2.9554E+05, 2.9859E+05, 3.0167E+05, 3.0478E+05,
     + 3.0792E+05, 3.1109E+05, 3.1429E+05, 3.1753E+05, 3.2080E+05,
     +  3.2410E+05, 3.2743E+05, 3.3080E+05, 3.3421E+05, 3.3764E+05,
     + 3.4112E+05, 3.4462E+05, 3.4816E+05, 3.5174E+05, 3.5536E+05,
     +  3.5901E+05, 3.6269E+05, 3.6642E+05, 3.7018E+05, 3.7398E+05,
     + 3.7781E+05, 3.8169E+05, 3.8560E+05, 3.8956E+05, 3.9355E+05,
     +  3.9758E+05, 4.0166E+05, 4.0577E+05, 4.0993E+05, 4.1413E+05,
     + 4.1836E+05, 4.2265E+05, 4.2697E+05, 4.3134E+05, 4.3575E+05,
     +  4.4021E+05, 4.4471E+05, 4.4925E+05, 4.5384E+05, 4.5848E+05,
     + 4.6316E+05, 4.6789E+05, 4.7267E+05, 4.7749E+05, 4.8236E+05,
     +  4.8728E+05, 4.9225E+05, 4.9727E+05, 5.0234E+05, 5.0746E+05,
     + 5.1263E+05, 5.1785E+05, 5.2312E+05, 5.2845E+05, 5.3382E+05,
     +  5.3926E+05, 5.4474E+05, 5.5028E+05, 5.5588E+05, 5.6152E+05,
     + 5.6723E+05, 5.7299E+05, 5.7881E+05, 5.8469E+05, 5.9062E+05,
     +  5.9661E+05, 6.0267E+05, 6.0878E+05, 6.1495E+05, 6.2118E+05,
     + 6.2748E+05, 6.3384E+05, 6.4026E+05, 6.4674E+05, 6.5328E+05,
     +  6.5989E+05, 6.6657E+05, 6.7331E+05, 6.8012E+05, 6.8699E+05,
     + 6.9393E+05, 7.0094E+05, 7.0802E+05, 7.1517E+05, 7.2239E+05,
     +  7.2968E+05, 7.3703E+05, 7.4447E+05, 7.5197E+05, 7.5955E+05,
     + 7.6720E+05, 7.7493E+05, 7.8273E+05, 7.9061E+05, 7.9856E+05,
     +  8.0659E+05, 8.1470E+05, 8.2290E+05, 8.3116E+05, 8.3951E+05,
     + 8.4795E+05, 8.5646E+05, 8.6506E+05, 8.7374E+05, 8.8250E+05,
     +  8.9135E+05, 9.0028E+05, 9.0931E+05, 9.1841E+05, 9.2761E+05,
     + 9.3690E+05, 9.4627E+05, 9.5574E+05, 9.6530E+05, 9.7495E+05,
     +  9.8469E+05, 9.9453E+05, 1.0045E+06, 1.0145E+06, 1.0246E+06,
     + 1.0348E+06, 1.0452E+06, 1.0556E+06, 1.0661E+06, 1.0767E+06,
     +  1.0875E+06, 1.0983E+06, 1.1092E+06, 1.1203E+06, 1.1314E+06,
     + 1.1427E+06, 1.1540E+06, 1.1655E+06, 1.1771E+06, 1.1888E+06,
     +  1.2006E+06, 1.2125E+06, 1.2245E+06, 1.2366E+06, 1.2489E+06,
     + 1.2613E+06, 1.2738E+06, 1.2864E+06, 1.2991E+06, 1.3120E+06,
     +  1.3249E+06, 1.3381E+06, 1.3513E+06, 1.3646E+06, 1.3781E+06,
     + 1.3917E+06, 1.4055E+06, 1.4193E+06, 1.4333E+06, 1.4475E+06,
     +  1.4617E+06, 1.4761E+06, 1.4907E+06, 1.5053E+06, 1.5202E+06,
     + 1.5351E+06, 1.5502E+06, 1.5654E+06, 1.5808E+06, 1.5964E+06,
     +  1.6120E+06, 1.6279E+06, 1.6438E+06, 1.6600E+06, 1.6762E+06,
     + 1.6927E+06, 1.7093E+06, 1.7260E+06, 1.7429E+06, 1.7599E+06,
     +  1.7772E+06, 1.7945E+06, 1.8121E+06, 1.8298E+06, 1.8477E+06,
     + 1.8657E+06, 1.8839E+06, 1.9023E+06, 1.9209E+06, 1.9396E+06,
     +  1.9585E+06, 1.9776E+06, 1.9968E+06, 2.0163E+06, 2.0359E+06,
     + 2.0557E+06, 2.0757E+06, 2.0958E+06, 2.1162E+06, 2.1367E+06,
     +  2.1575E+06, 2.1784E+06, 2.1996E+06, 2.2209E+06, 2.2424E+06,
     + 2.2641E+06, 2.2861E+06, 2.3082E+06, 2.3305E+06, 2.3531E+06,
     +  2.3758E+06, 2.3988E+06, 2.4219E+06, 2.4453E+06, 2.4689E+06,
     + 2.4928E+06, 2.5168E+06, 2.5411E+06, 2.5656E+06, 2.5903E+06,
     +  2.6152E+06, 2.6404E+06, 2.6658E+06, 2.6914E+06, 2.7173E+06,
     + 2.7434E+06, 2.7698E+06, 2.7964E+06, 2.8232E+06, 2.8503E+06,
     +  2.8776E+06, 2.9052E+06, 2.9331E+06, 2.9611E+06, 2.9895E+06,
     + 3.0181E+06, 3.0470E+06, 3.0761E+06, 3.1055E+06, 3.1352E+06,
     +  3.1651E+06, 3.1953E+06, 3.2258E+06, 3.2566E+06, 3.2876E+06,
     + 3.3190E+06, 3.3506E+06, 3.3825E+06, 3.4147E+06, 3.4472E+06,
     +  3.4800E+06, 3.5130E+06, 3.5464E+06, 3.5801E+06, 3.6141E+06,
     + 3.6484E+06, 3.6830E+06, 3.7179E+06, 3.7532E+06, 3.7887E+06,
     +  3.8246E+06, 3.8608E+06, 3.8973E+06, 3.9342E+06, 3.9714E+06,
     + 4.0089E+06, 4.0468E+06, 4.0850E+06, 4.1235E+06, 4.1624E+06,
     +  4.2017E+06, 4.2413E+06, 4.2812E+06, 4.3215E+06, 4.3622E+06,
     + 4.4032E+06, 4.4447E+06, 4.4864E+06, 4.5286E+06, 4.5711E+06,
     +  4.6140E+06, 4.6573E+06, 4.7010E+06, 4.7451E+06, 4.7896E+06,
     + 4.8344E+06, 4.8797E+06, 4.9253E+06, 4.9714E+06, 5.0179E+06,
     +  5.0648E+06, 5.1121E+06, 5.1599E+06, 5.2080E+06, 5.2566E+06,
     + 5.3056E+06, 5.3551E+06, 5.4050E+06, 5.4553E+06, 5.5061E+06,
     +  5.5574E+06, 5.6090E+06, 5.6612E+06, 5.7138E+06, 5.7669E+06,
     + 5.8204E+06, 5.8744E+06, 5.9289E+06, 5.9839E+06, 6.0394E+06,
     +  6.0953E+06, 6.1517E+06, 6.2087E+06, 6.2661E+06, 6.3241E+06,
     + 6.3825E+06, 6.4415E+06, 6.5010E+06, 6.5610E+06, 6.6215E+06,
     +  6.6826E+06, 6.7442E+06, 6.8063E+06, 6.8690E+06, 6.9322E+06,
     + 6.9960E+06, 7.0603E+06, 7.1252E+06, 7.1907E+06, 7.2568E+06,
     +  7.3234E+06, 7.3906E+06, 7.4584E+06, 7.5267E+06, 7.5957E+06,
     + 7.6653E+06, 7.7355E+06, 7.8063E+06, 7.8777E+06, 7.9497E+06,
     +  8.0223E+06, 8.0956E+06, 8.1695E+06, 8.2441E+06, 8.3193E+06,
     + 8.3951E+06, 8.4716E+06, 8.5488E+06, 8.6266E+06, 8.7051E+06,
     +  8.7843E+06, 8.8642E+06, 8.9448E+06, 9.0260E+06, 9.1080E+06,
     + 9.1906E+06, 9.2740E+06, 9.3581E+06, 9.4429E+06, 9.5284E+06,
     +  9.6147E+06, 9.7017E+06, 9.7895E+06, 9.8780E+06, 9.9673E+06,
     + 1.0057E+07, 1.0148E+07, 1.0240E+07, 1.0332E+07, 1.0425E+07,
     +  1.0519E+07, 1.0614E+07, 1.0710E+07, 1.0806E+07, 1.0903E+07,
     + 1.1001E+07, 1.1100E+07, 1.1200E+07, 1.1300E+07, 1.1402E+07,
     +  1.1504E+07, 1.1607E+07, 1.1711E+07, 1.1816E+07, 1.1922E+07,
     + 1.2029E+07, 1.2136E+07, 1.2245E+07, 1.2354E+07, 1.2465E+07,
     +  1.2576E+07, 1.2688E+07, 1.2801E+07, 1.2915E+07, 1.3031E+07,
     + 1.3147E+07, 1.3264E+07, 1.3382E+07, 1.3501E+07, 1.3621E+07,
     +  1.3742E+07, 1.3864E+07, 1.3987E+07, 1.4111E+07, 1.4236E+07,
     + 1.4363E+07, 1.4490E+07, 1.4618E+07, 1.4748E+07, 1.4878E+07,
     +  1.5010E+07, 1.5142E+07, 1.5276E+07, 1.5411E+07, 1.5547E+07,
     + 1.5684E+07, 1.5823E+07, 1.5962E+07, 1.6103E+07, 1.6245E+07,
     +  1.6388E+07, 1.6532E+07, 1.6677E+07, 1.6824E+07, 1.6972E+07,
     + 1.7121E+07, 1.7271E+07, 1.7423E+07, 1.7575E+07, 1.7730E+07,
     +  1.7885E+07, 1.8042E+07, 1.8199E+07, 1.8359E+07, 1.8519E+07,
     + 1.8681E+07, 1.8844E+07, 1.9009E+07, 1.9175E+07, 1.9342E+07,
     +  1.9511E+07, 1.9681E+07, 1.9852E+07, 2.0025E+07, 2.0199E+07,
     + 2.0375E+07, 2.0552E+07, 2.0731E+07, 2.0911E+07, 2.1093E+07,
     +  2.1276E+07, 2.1460E+07, 2.1646E+07, 2.1834E+07, 2.2023E+07,
     + 2.2214E+07, 2.2406E+07, 2.2600E+07, 2.2795E+07, 2.2992E+07,
     +  2.3191E+07, 2.3391E+07, 2.3593E+07, 2.3796E+07, 2.4001E+07,
     + 2.4208E+07, 2.4417E+07, 2.4627E+07, 2.4839E+07, 2.5052E+07,
     +  2.5268E+07, 2.5485E+07, 2.5704E+07, 2.5924E+07, 2.6147E+07,
     + 2.6371E+07, 2.6597E+07, 2.6825E+07, 2.7055E+07, 2.7286E+07,
     +  2.7520E+07, 2.7755E+07, 2.7992E+07, 2.8231E+07, 2.8472E+07,
     + 2.8715E+07, 2.8960E+07, 2.9207E+07, 2.9456E+07, 2.9707E+07,
     +  2.9960E+07, 3.0215E+07, 3.0472E+07, 3.0731E+07, 3.0992E+07,
     + 3.1255E+07, 3.1520E+07, 3.1788E+07, 3.2058E+07, 3.2329E+07,
     +  3.2603E+07, 3.2879E+07, 3.3158E+07, 3.3438E+07, 3.3721E+07,
     + 3.4006E+07, 3.4293E+07, 3.4583E+07, 3.4875E+07, 3.5169E+07,
     +  3.5466E+07, 3.5765E+07, 3.6066E+07, 3.6369E+07, 3.6676E+07,
     + 3.6984E+07, 3.7295E+07, 3.7608E+07, 3.7924E+07, 3.8243E+07,
     +  3.8564E+07, 3.8887E+07, 3.9213E+07, 3.9542E+07, 3.9873E+07,
     + 4.0207E+07, 4.0543E+07, 4.0882E+07, 4.1224E+07, 4.1568E+07,
     +  4.1916E+07, 4.2266E+07, 4.2618E+07, 4.2974E+07, 4.3332E+07,
     + 4.3693E+07, 4.4057E+07, 4.4423E+07, 4.4793E+07, 4.5165E+07,
     +  4.5541E+07, 4.5919E+07, 4.6300E+07, 4.6685E+07, 4.7072E+07,
     + 4.7462E+07, 4.7855E+07, 4.8252E+07, 4.8651E+07, 4.9054E+07,
     +  4.9460E+07, 4.9869E+07, 5.0281E+07, 5.0696E+07, 5.1114E+07,
     + 5.1536E+07, 5.1961E+07, 5.2390E+07, 5.2821E+07, 5.3256E+07,
     +  5.3695E+07, 5.4136E+07, 5.4582E+07, 5.5030E+07, 5.5482E+07,
     + 5.5938E+07, 5.6397E+07, 5.6860E+07, 5.7326E+07, 5.7796E+07,
     +  5.8269E+07, 5.8746E+07, 5.9227E+07, 5.9712E+07, 6.0200E+07,
     + 6.0692E+07, 6.1188E+07, 6.1687E+07, 6.2190E+07, 6.2698E+07,
     +  6.3209E+07, 6.3724E+07, 6.4243E+07, 6.4766E+07, 6.5293E+07,
     + 6.5824E+07, 6.6359E+07, 6.6898E+07, 6.7441E+07, 6.7989E+07,
     +  6.8540E+07, 6.9096E+07, 6.9656E+07, 7.0221E+07, 7.0789E+07,
     + 7.1362E+07, 7.1939E+07, 7.2521E+07, 7.3107E+07, 7.3698E+07,
     +  7.4293E+07, 7.4892E+07, 7.5497E+07, 7.6105E+07, 7.6719E+07,
     + 7.7336E+07, 7.7959E+07, 7.8586E+07, 7.9219E+07, 7.9855E+07,
     +  8.0497E+07, 8.1144E+07, 8.1795E+07, 8.2451E+07, 8.3112E+07,
     + 8.3779E+07, 8.4450E+07, 8.5126E+07, 8.5808E+07, 8.6494E+07,
     +  8.7186E+07, 8.7883E+07, 8.8585E+07, 8.9292E+07, 9.0005E+07,
     + 9.0723E+07, 9.1446E+07, 9.2175E+07, 9.2909E+07, 9.3649E+07,
     +  9.4394E+07, 9.5145E+07, 9.5902E+07, 9.6664E+07, 9.7432E+07,
     + 9.8205E+07, 9.8985E+07, 9.9770E+07, 1.0056E+08, 1.0136E+08,
     +  1.0216E+08, 1.0297E+08, 1.0378E+08, 1.0460E+08, 1.0543E+08,
     + 1.0626E+08, 1.0710E+08, 1.0795E+08, 1.0880E+08, 1.0966E+08,
     +  1.1052E+08, 1.1139E+08, 1.1227E+08, 1.1316E+08, 1.1405E+08,
     + 1.1494E+08, 1.1585E+08, 1.1676E+08, 1.1767E+08, 1.1860E+08,
     +  1.1953E+08, 1.2046E+08, 1.2141E+08, 1.2236E+08, 1.2332E+08,
     + 1.2428E+08, 1.2525E+08, 1.2623E+08, 1.2722E+08, 1.2821E+08,
     +  1.2921E+08, 1.3022E+08, 1.3124E+08, 1.3226E+08, 1.3329E+08,
     + 1.3433E+08, 1.3538E+08, 1.3643E+08, 1.3749E+08, 1.3856E+08,
     +  1.3964E+08, 1.4072E+08, 1.4181E+08, 1.4291E+08, 1.4402E+08,
     + 1.4514E+08, 1.4626E+08, 1.4739E+08, 1.4854E+08, 1.4968E+08,
     +  1.5084E+08, 1.5201E+08, 1.5318E+08, 1.5437E+08, 1.5556E+08,
     + 1.5676E+08, 1.5797E+08, 1.5918E+08, 1.6041E+08, 1.6164E+08,
     +  1.6289E+08, 1.6414E+08, 1.6540E+08, 1.6667E+08, 1.6795E+08,
     + 1.6924E+08, 1.7054E+08, 1.7185E+08, 1.7317E+08, 1.7450E+08,
     +  1.7583E+08, 1.7718E+08, 1.7853E+08, 1.7990E+08, 1.8128E+08,
     + 1.8266E+08, 1.8406E+08, 1.8546E+08, 1.8688E+08, 1.8830E+08,
     +  1.8974E+08, 1.9118E+08, 1.9264E+08, 1.9411E+08, 1.9558E+08,
     + 1.9707E+08, 1.9857E+08, 2.0008E+08, 2.0160E+08, 2.0313E+08,
     +  2.0467E+08, 2.0622E+08, 2.0778E+08, 2.0936E+08, 2.1094E+08,
     + 2.1254E+08, 2.1415E+08, 2.1577E+08, 2.1740E+08, 2.1904E+08,
     +  2.2069E+08, 2.2236E+08, 2.2404E+08, 2.2573E+08, 2.2743E+08,
     + 2.2914E+08, 2.3087E+08, 2.3260E+08, 2.3435E+08, 2.3612E+08,
     +  2.3789E+08, 2.3968E+08, 2.4148E+08, 2.4329E+08, 2.4511E+08,
     + 2.4695E+08, 2.4880E+08, 2.5067E+08, 2.5254E+08, 2.5443E+08,
     +  2.5634E+08, 2.5825E+08, 2.6018E+08, 2.6213E+08, 2.6408E+08,
     + 2.6606E+08, 2.6804E+08, 2.7004E+08, 2.7205E+08, 2.7408E+08,
     +  2.7612E+08, 2.7817E+08, 2.8024E+08, 2.8233E+08, 2.8442E+08,
     + 2.8654E+08, 2.8866E+08, 2.9081E+08, 2.9296E+08, 2.9513E+08,
     +  2.9732E+08, 2.9952E+08, 3.0174E+08, 3.0397E+08, 3.0622E+08,
     + 3.0849E+08, 3.1077E+08, 3.1306E+08, 3.1537E+08, 3.1770E+08,
     +  3.2004E+08, 3.2240E+08, 3.2478E+08, 3.2717E+08, 3.2958E+08,
     + 3.3200E+08, 3.3445E+08, 3.3691E+08, 3.3938E+08, 3.4187E+08,
     +  3.4438E+08, 3.4691E+08, 3.4946E+08, 3.5202E+08, 3.5460E+08,
     + 3.5719E+08, 3.5981E+08, 3.6244E+08, 3.6509E+08, 3.6776E+08,
     +  3.7045E+08, 3.7315E+08, 3.7588E+08, 3.7862E+08, 3.8138E+08,
     + 3.8416E+08, 3.8696E+08, 3.8978E+08, 3.9262E+08, 3.9548E+08,
     +  3.9835E+08, 4.0125E+08, 4.0416E+08, 4.0710E+08, 4.1005E+08,
     + 4.1303E+08, 4.1603E+08, 4.1904E+08, 4.2208E+08, 4.2514E+08,
     +  4.2821E+08, 4.3131E+08, 4.3443E+08, 4.3757E+08, 4.4073E+08,
     + 4.4392E+08, 4.4712E+08, 4.5035E+08, 4.5360E+08, 4.5687E+08,
     +  4.6016E+08, 4.6347E+08, 4.6681E+08, 4.7017E+08, 4.7355E+08,
     + 4.7696E+08, 4.8038E+08, 4.8383E+08, 4.8731E+08, 4.9081E+08,
     +  4.9433E+08, 4.9787E+08, 5.0144E+08, 5.0503E+08, 5.0865E+08,
     + 5.1229E+08, 5.1595E+08, 5.1964E+08, 5.2335E+08, 5.2709E+08,
     +  5.3085E+08, 5.3464E+08, 5.3846E+08, 5.4230E+08, 5.4616E+08,
     + 5.5005E+08, 5.5397E+08, 5.5791E+08, 5.6188E+08, 5.6587E+08,
     +  5.6990E+08, 5.7394E+08, 5.7802E+08, 5.8212E+08, 5.8625E+08,
     + 5.9041E+08, 5.9459E+08, 5.9881E+08, 6.0305E+08, 6.0731E+08,
     +  6.1161E+08, 6.1594E+08, 6.2029E+08, 6.2467E+08, 6.2908E+08,
     + 6.3352E+08, 6.3799E+08, 6.4249E+08, 6.4702E+08, 6.5158E+08,
     +  6.5617E+08, 6.6079E+08, 6.6544E+08, 6.7012E+08, 6.7483E+08,
     + 6.7957E+08, 6.8434E+08, 6.8914E+08, 6.9398E+08, 6.9885E+08,
     +  7.0375E+08, 7.0868E+08, 7.1364E+08, 7.1864E+08, 7.2367E+08,
     + 7.2873E+08, 7.3382E+08, 7.3895E+08, 7.4411E+08, 7.4931E+08,
     +  7.5454E+08, 7.5980E+08, 7.6510E+08, 7.7043E+08, 7.7579E+08,
     + 7.8120E+08, 7.8663E+08, 7.9210E+08, 7.9761E+08, 8.0316E+08,
     +  8.0874E+08, 8.1435E+08, 8.2000E+08, 8.2569E+08, 8.3142E+08,
     + 8.3718E+08, 8.4298E+08, 8.4882E+08, 8.5469E+08, 8.6061E+08,
     +  8.6656E+08, 8.7255E+08, 8.7858E+08, 8.8464E+08, 8.9075E+08,
     + 8.9690E+08, 9.0308E+08, 9.0931E+08, 9.1557E+08, 9.2188E+08,
     +  9.2823E+08, 9.3461E+08, 9.4104E+08, 9.4751E+08, 9.5402E+08,
     + 9.6058E+08, 9.6717E+08, 9.7381E+08, 9.8049E+08, 9.8721E+08,
     +  9.9398E+08, 1.0008E+09, 1.0076E+09, 1.0145E+09, 1.0215E+09,
     + 1.0285E+09, 1.0355E+09, 1.0426E+09, 1.0497E+09, 1.0568E+09,
     +  1.0641E+09, 1.0713E+09, 1.0786E+09, 1.0860E+09, 1.0934E+09,
     + 1.1008E+09, 1.1083E+09, 1.1158E+09, 1.1234E+09, 1.1311E+09,
     +  1.1387E+09, 1.1465E+09, 1.1543E+09, 1.1621E+09, 1.1700E+09,
     + 1.1779E+09, 1.1859E+09, 1.1939E+09, 1.2020E+09, 1.2101E+09,
     +  1.2183E+09, 1.2265E+09, 1.2348E+09, 1.2432E+09, 1.2515E+09,
     + 1.2600E+09, 1.2685E+09, 1.2770E+09, 1.2856E+09, 1.2943E+09,
     +  1.3030E+09, 1.3118E+09, 1.3206E+09, 1.3295E+09, 1.3384E+09,
     + 1.3474E+09, 1.3565E+09, 1.3656E+09, 1.3747E+09, 1.3839E+09,
     +  1.3932E+09, 1.4025E+09, 1.4119E+09, 1.4214E+09, 1.4309E+09,
     + 1.4405E+09, 1.4501E+09, 1.4598E+09, 1.4695E+09, 1.4793E+09,
     +  1.4892E+09, 1.4991E+09, 1.5091E+09, 1.5192E+09, 1.5293E+09,
     + 1.5395E+09, 1.5497E+09, 1.5601E+09, 1.5704E+09, 1.5809E+09,
     +  1.5914E+09, 1.6019E+09, 1.6126E+09, 1.6233E+09, 1.6340E+09,
     + 1.6449E+09, 1.6558E+09, 1.6667E+09, 1.6778E+09, 1.6889E+09,
     +  1.7000E+09, 1.7113E+09, 1.7226E+09, 1.7340E+09, 1.7454E+09,
     + 1.7569E+09, 1.7685E+09, 1.7802E+09, 1.7919E+09, 1.8037E+09,
     +  1.8156E+09, 1.8275E+09, 1.8396E+09, 1.8517E+09, 1.8638E+09,
     + 1.8761E+09, 1.8884E+09, 1.9008E+09, 1.9133E+09, 1.9258E+09,
     +  1.9385E+09, 1.9512E+09, 1.9639E+09, 1.9768E+09, 1.9897E+09,
     + 2.0028E+09, 2.0159E+09, 2.0290E+09, 2.0423E+09, 2.0556E+09,
     +  2.0690E+09, 2.0825E+09, 2.0961E+09, 2.1098E+09, 2.1235E+09,
     + 2.1374E+09, 2.1513E+09, 2.1653E+09, 2.1794E+09, 2.1935E+09,
     +  2.2078E+09, 2.2221E+09, 2.2366E+09, 2.2511E+09, 2.2657E+09,
     + 2.2804E+09, 2.2952E+09, 2.3101E+09, 2.3250E+09, 2.3401E+09,
     +  2.3552E+09, 2.3705E+09, 2.3858E+09, 2.4012E+09, 2.4167E+09,
     + 2.4323E+09, 2.4480E+09, 2.4638E+09, 2.4797E+09, 2.4957E+09,
     +  2.5118E+09, 2.5279E+09, 2.5442E+09, 2.5606E+09, 2.5771E+09,
     + 2.5936E+09, 2.6103E+09, 2.6271E+09, 2.6439E+09, 2.6609E+09,
     +  2.6780E+09, 2.6952E+09, 2.7124E+09, 2.7298E+09, 2.7473E+09,
     + 2.7649E+09, 2.7826E+09, 2.8004E+09, 2.8183E+09, 2.8363E+09,
     +  2.8544E+09, 2.8727E+09, 2.8910E+09, 2.9094E+09, 2.9280E+09,
     + 2.9467E+09, 2.9654E+09, 2.9843E+09, 3.0033E+09, 3.0224E+09,
     +  3.0417E+09, 3.0610E+09, 3.0805E+09, 3.1000E+09, 3.1197E+09,
     + 3.1395E+09, 3.1594E+09, 3.1795E+09, 3.1996E+09, 3.2199E+09,
     +  3.2403E+09, 3.2608E+09, 3.2815E+09, 3.3022E+09, 3.3231E+09,
     + 3.3441E+09, 3.3652E+09, 3.3865E+09, 3.4079E+09, 3.4294E+09,
     +  3.4510E+09, 3.4728E+09, 3.4946E+09, 3.5167E+09, 3.5388E+09,
     + 3.5611E+09, 3.5835E+09, 3.6060E+09, 3.6287E+09, 3.6515E+09,
     +  3.6744E+09, 3.6975E+09, 3.7207E+09, 3.7440E+09, 3.7675E+09,
     + 3.7911E+09, 3.8149E+09, 3.8388E+09, 3.8628E+09, 3.8870E+09,
     +  3.9113E+09, 3.9357E+09, 3.9603E+09, 3.9850E+09, 4.0099E+09,
     + 4.0350E+09, 4.0601E+09, 4.0854E+09, 4.1109E+09, 4.1365E+09,
     +  4.1623E+09, 4.1882E+09, 4.2143E+09, 4.2405E+09, 4.2668E+09,
     + 4.2934E+09, 4.3200E+09, 4.3469E+09, 4.3738E+09, 4.4010E+09,
     +  4.4283E+09, 4.4557E+09, 4.4833E+09, 4.5111E+09, 4.5390E+09,
     + 4.5671E+09, 4.5954E+09, 4.6238E+09, 4.6524E+09, 4.6811E+09,
     +  4.7100E+09, 4.7391E+09, 4.7684E+09, 4.7978E+09, 4.8273E+09,
     + 4.8571E+09, 4.8870E+09, 4.9171E+09, 4.9474E+09, 4.9778E+09,
     +  5.0084E+09, 5.0392E+09, 5.0702E+09, 5.1013E+09, 5.1327E+09,
     + 5.1642E+09, 5.1958E+09, 5.2277E+09, 5.2598E+09, 5.2920E+09,
     +  5.3244E+09, 5.3570E+09, 5.3898E+09, 5.4227E+09, 5.4559E+09,
     + 5.4892E+09, 5.5228E+09, 5.5565E+09, 5.5904E+09, 5.6245E+09,
     +  5.6589E+09, 5.6934E+09, 5.7280E+09, 5.7629E+09, 5.7980E+09,
     + 5.8333E+09, 5.8688E+09, 5.9045E+09, 5.9404E+09, 5.9765E+09,
     +  6.0128E+09, 6.0493E+09, 6.0860E+09, 6.1230E+09, 6.1601E+09,
     + 6.1974E+09, 6.2350E+09, 6.2727E+09, 6.3107E+09, 6.3489E+09,
     +  6.3873E+09, 6.4259E+09, 6.4648E+09, 6.5038E+09, 6.5431E+09,
     + 6.5826E+09, 6.6223E+09, 6.6623E+09, 6.7025E+09, 6.7429E+09,
     +  6.7835E+09, 6.8243E+09, 6.8654E+09, 6.9067E+09, 6.9483E+09,
     + 6.9900E+09, 7.0320E+09, 7.0743E+09, 7.1168E+09, 7.1595E+09,
     +  7.2024E+09, 7.2456E+09, 7.2891E+09, 7.3328E+09, 7.3767E+09,
     + 7.4209E+09, 7.4653E+09, 7.5099E+09, 7.5548E+09, 7.6000E+09,
     +  7.6454E+09, 7.6911E+09, 7.7370E+09, 7.7832E+09, 7.8296E+09,
     + 7.8763E+09, 7.9233E+09, 7.9705E+09, 8.0179E+09, 8.0657E+09,
     +  8.1137E+09, 8.1620E+09, 8.2105E+09, 8.2593E+09, 8.3084E+09,
     + 8.3577E+09, 8.4073E+09, 8.4572E+09, 8.5074E+09, 8.5578E+09,
     +  8.6086E+09, 8.6596E+09, 8.7108E+09, 8.7624E+09, 8.8143E+09,
     + 8.8664E+09, 8.9188E+09, 8.9715E+09, 9.0246E+09, 9.0779E+09,
     +  9.1314E+09, 9.1853E+09, 9.2395E+09, 9.2940E+09, 9.3488E+09,
     + 9.4038E+09, 9.4592E+09, 9.5149E+09, 9.5709E+09, 9.6272E+09,
     +  9.6838E+09, 9.7407E+09, 9.7979E+09, 9.8555E+09, 9.9133E+09,
     + 9.9715E+09, 1.0030E+10, 1.0089E+10, 1.0148E+10, 1.0207E+10,
     +  1.0267E+10, 1.0327E+10, 1.0388E+10, 1.0448E+10, 1.0510E+10,
     + 1.0571E+10, 1.0633E+10, 1.0695E+10, 1.0757E+10, 1.0820E+10,
     +  1.0883E+10, 1.0947E+10, 1.1010E+10, 1.1074E+10, 1.1139E+10,
     + 1.1204E+10, 1.1269E+10, 1.1335E+10, 1.1400E+10, 1.1467E+10,
     +  1.1533E+10, 1.1600E+10, 1.1668E+10, 1.1735E+10, 1.1803E+10,
     + 1.1872E+10, 1.1941E+10, 1.2010E+10, 1.2079E+10, 1.2149E+10,
     +  1.2220E+10, 1.2290E+10, 1.2361E+10, 1.2433E+10, 1.2505E+10,
     + 1.2577E+10, 1.2649E+10, 1.2722E+10, 1.2796E+10, 1.2870E+10,
     +  1.2944E+10, 1.3018E+10, 1.3093E+10, 1.3169E+10, 1.3244E+10,
     + 1.3321E+10, 1.3397E+10, 1.3474E+10, 1.3552E+10, 1.3629E+10,
     +  1.3708E+10, 1.3786E+10, 1.3865E+10, 1.3945E+10, 1.4025E+10,
     + 1.4105E+10, 1.4186E+10, 1.4267E+10, 1.4349E+10, 1.4431E+10,
     +  1.4514E+10, 1.4596E+10, 1.4680E+10, 1.4764E+10, 1.4848E+10,
     + 1.4933E+10, 1.5018E+10, 1.5104E+10, 1.5190E+10, 1.5276E+10,
     +  1.5363E+10, 1.5451E+10, 1.5539E+10, 1.5627E+10, 1.5716E+10,
     + 1.5805E+10, 1.5895E+10, 1.5985E+10, 1.6076E+10, 1.6167E+10,
     +  1.6259E+10, 1.6351E+10, 1.6444E+10, 1.6537E+10, 1.6631E+10,
     + 1.6725E+10, 1.6820E+10, 1.6915E+10, 1.7011E+10, 1.7107E+10,
     +  1.7203E+10, 1.7301E+10, 1.7398E+10, 1.7497E+10, 1.7595E+10,
     + 1.7695E+10, 1.7794E+10, 1.7895E+10, 1.7995E+10, 1.8097E+10,
     +  1.8199E+10, 1.8301E+10, 1.8404E+10, 1.8508E+10, 1.8612E+10,
     + 1.8716E+10, 1.8821E+10, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_CH3OH (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 1),QofT( 1,NT),Q(NT)
      data Tmx(1)/3500./
      data xgj/ 1./
c...    CH3OH
c...        --      2161
      data (QofT( 1,J),J=1,2408)/ 1.4411E+01, 3.3618E+01, 5.7992E+01,
     +  8.6539E+01, 1.1871E+02, 1.5412E+02, 1.9250E+02, 2.3365E+02,
     + 2.7737E+02, 3.2353E+02, 3.7200E+02, 4.2269E+02, 4.7549E+02,
     +  5.3033E+02, 5.8713E+02, 6.4582E+02, 7.0636E+02, 7.6868E+02,
     + 8.3273E+02, 8.9848E+02, 1.0349E+03, 1.1775E+03, 1.3262E+03,
     +  1.4807E+03, 1.6408E+03, 1.8063E+03, 1.9770E+03, 2.1528E+03,
     + 2.3335E+03, 2.5191E+03, 2.7094E+03, 2.9043E+03, 3.1038E+03,
     +  3.3077E+03, 3.5160E+03, 3.7287E+03, 3.9458E+03, 4.1670E+03,
     + 4.3926E+03, 4.6224E+03, 4.8564E+03, 5.0946E+03, 5.3371E+03,
     +  5.5837E+03, 5.8346E+03, 6.0898E+03, 6.3492E+03, 6.6129E+03,
     + 6.8809E+03, 7.1533E+03, 7.4300E+03, 7.7111E+03, 7.9966E+03,
     +  8.2866E+03, 8.5811E+03, 8.8801E+03, 9.1837E+03, 9.4918E+03,
     + 9.8046E+03, 1.0122E+04, 1.0444E+04, 1.0771E+04, 1.1103E+04,
     +  1.1440E+04, 1.1781E+04, 1.2127E+04, 1.2479E+04, 1.2835E+04,
     + 1.3196E+04, 1.3562E+04, 1.3934E+04, 1.4310E+04, 1.4692E+04,
     +  1.5079E+04, 1.5471E+04, 1.5868E+04, 1.6271E+04, 1.6679E+04,
     + 1.7093E+04, 1.7512E+04, 1.7936E+04, 1.8366E+04, 1.8801E+04,
     +  1.9242E+04, 1.9689E+04, 2.0141E+04, 2.0599E+04, 2.1062E+04,
     + 2.1531E+04, 2.2007E+04, 2.2488E+04, 2.2974E+04, 2.3467E+04,
     +  2.3966E+04, 2.4470E+04, 2.4981E+04, 2.5498E+04, 2.6021E+04,
     + 2.6550E+04, 2.7085E+04, 2.7626E+04, 2.8174E+04, 2.8728E+04,
     +  2.9289E+04, 2.9855E+04, 3.0429E+04, 3.1008E+04, 3.1595E+04,
     + 3.2188E+04, 3.2787E+04, 3.3393E+04, 3.4006E+04, 3.4626E+04,
     +  3.5253E+04, 3.5886E+04, 3.6526E+04, 3.7174E+04, 3.7828E+04,
     + 3.8490E+04, 3.9159E+04, 3.9835E+04, 4.0518E+04, 4.1209E+04,
     +  4.1907E+04, 4.2612E+04, 4.3326E+04, 4.4046E+04, 4.4775E+04,
     + 4.5511E+04, 4.6255E+04, 4.7006E+04, 4.7766E+04, 4.8534E+04,
     +  4.9310E+04, 5.0094E+04, 5.0886E+04, 5.1687E+04, 5.2496E+04,
     + 5.3313E+04, 5.4139E+04, 5.4974E+04, 5.5817E+04, 5.6669E+04,
     +  5.7530E+04, 5.8400E+04, 5.9279E+04, 6.0168E+04, 6.1065E+04,
     + 6.1972E+04, 6.2888E+04, 6.3814E+04, 6.4750E+04, 6.5695E+04,
     +  6.6650E+04, 6.7615E+04, 6.8590E+04, 6.9575E+04, 7.0570E+04,
     + 7.1576E+04, 7.2592E+04, 7.3618E+04, 7.4656E+04, 7.5704E+04,
     +  7.6763E+04, 7.7833E+04, 7.8914E+04, 8.0006E+04, 8.1110E+04,
     + 8.2225E+04, 8.3352E+04, 8.4490E+04, 8.5641E+04, 8.6803E+04,
     +  8.7978E+04, 8.9164E+04, 9.0363E+04, 9.1575E+04, 9.2799E+04,
     + 9.4036E+04, 9.5286E+04, 9.6549E+04, 9.7826E+04, 9.9115E+04,
     +  1.0042E+05, 1.0174E+05, 1.0307E+05, 1.0441E+05, 1.0577E+05,
     + 1.0714E+05, 1.0853E+05, 1.0993E+05, 1.1135E+05, 1.1278E+05,
     +  1.1422E+05, 1.1569E+05, 1.1716E+05, 1.1865E+05, 1.2016E+05,
     + 1.2169E+05, 1.2323E+05, 1.2478E+05, 1.2636E+05, 1.2795E+05,
     +  1.2955E+05, 1.3117E+05, 1.3281E+05, 1.3447E+05, 1.3615E+05,
     + 1.3784E+05, 1.3955E+05, 1.4128E+05, 1.4303E+05, 1.4479E+05,
     +  1.4657E+05, 1.4838E+05, 1.5020E+05, 1.5204E+05, 1.5390E+05,
     + 1.5578E+05, 1.5768E+05, 1.5960E+05, 1.6154E+05, 1.6351E+05,
     +  1.6549E+05, 1.6749E+05, 1.6952E+05, 1.7156E+05, 1.7363E+05,
     + 1.7572E+05, 1.7783E+05, 1.7997E+05, 1.8212E+05, 1.8430E+05,
     +  1.8651E+05, 1.8873E+05, 1.9098E+05, 1.9326E+05, 1.9556E+05,
     + 1.9788E+05, 2.0023E+05, 2.0260E+05, 2.0500E+05, 2.0742E+05,
     +  2.0987E+05, 2.1234E+05, 2.1484E+05, 2.1737E+05, 2.1993E+05,
     + 2.2251E+05, 2.2512E+05, 2.2775E+05, 2.3042E+05, 2.3311E+05,
     +  2.3583E+05, 2.3859E+05, 2.4137E+05, 2.4418E+05, 2.4701E+05,
     + 2.4988E+05, 2.5278E+05, 2.5572E+05, 2.5868E+05, 2.6167E+05,
     +  2.6470E+05, 2.6776E+05, 2.7085E+05, 2.7397E+05, 2.7712E+05,
     + 2.8031E+05, 2.8354E+05, 2.8679E+05, 2.9009E+05, 2.9341E+05,
     +  2.9678E+05, 3.0017E+05, 3.0361E+05, 3.0708E+05, 3.1059E+05,
     + 3.1413E+05, 3.1771E+05, 3.2133E+05, 3.2499E+05, 3.2869E+05,
     +  3.3242E+05, 3.3620E+05, 3.4002E+05, 3.4387E+05, 3.4777E+05,
     + 3.5171E+05, 3.5569E+05, 3.5971E+05, 3.6377E+05, 3.6788E+05,
     +  3.7203E+05, 3.7622E+05, 3.8046E+05, 3.8475E+05, 3.8908E+05,
     + 3.9345E+05, 3.9787E+05, 4.0234E+05, 4.0685E+05, 4.1141E+05,
     +  4.1602E+05, 4.2068E+05, 4.2539E+05, 4.3014E+05, 4.3495E+05,
     + 4.3981E+05, 4.4471E+05, 4.4967E+05, 4.5469E+05, 4.5975E+05,
     +  4.6487E+05, 4.7004E+05, 4.7526E+05, 4.8054E+05, 4.8588E+05,
     + 4.9127E+05, 4.9672E+05, 5.0222E+05, 5.0779E+05, 5.1341E+05,
     +  5.1909E+05, 5.2482E+05, 5.3062E+05, 5.3648E+05, 5.4240E+05,
     + 5.4839E+05, 5.5443E+05, 5.6054E+05, 5.6671E+05, 5.7294E+05,
     +  5.7925E+05, 5.8561E+05, 5.9204E+05, 5.9854E+05, 6.0511E+05,
     + 6.1175E+05, 6.1845E+05, 6.2522E+05, 6.3207E+05, 6.3898E+05,
     +  6.4597E+05, 6.5303E+05, 6.6016E+05, 6.6737E+05, 6.7465E+05,
     + 6.8201E+05, 6.8944E+05, 6.9695E+05, 7.0454E+05, 7.1220E+05,
     +  7.1995E+05, 7.2777E+05, 7.3568E+05, 7.4366E+05, 7.5173E+05,
     + 7.5989E+05, 7.6812E+05, 7.7644E+05, 7.8485E+05, 7.9335E+05,
     +  8.0193E+05, 8.1060E+05, 8.1935E+05, 8.2820E+05, 8.3714E+05,
     + 8.4617E+05, 8.5530E+05, 8.6451E+05, 8.7382E+05, 8.8323E+05,
     +  8.9273E+05, 9.0234E+05, 9.1203E+05, 9.2183E+05, 9.3173E+05,
     + 9.4173E+05, 9.5183E+05, 9.6203E+05, 9.7234E+05, 9.8276E+05,
     +  9.9328E+05, 1.0039E+06, 1.0146E+06, 1.0255E+06, 1.0364E+06,
     + 1.0475E+06, 1.0587E+06, 1.0700E+06, 1.0814E+06, 1.0929E+06,
     +  1.1045E+06, 1.1163E+06, 1.1282E+06, 1.1402E+06, 1.1523E+06,
     + 1.1645E+06, 1.1769E+06, 1.1894E+06, 1.2020E+06, 1.2147E+06,
     +  1.2276E+06, 1.2406E+06, 1.2537E+06, 1.2670E+06, 1.2804E+06,
     + 1.2939E+06, 1.3076E+06, 1.3214E+06, 1.3354E+06, 1.3495E+06,
     +  1.3637E+06, 1.3781E+06, 1.3926E+06, 1.4072E+06, 1.4220E+06,
     + 1.4370E+06, 1.4521E+06, 1.4674E+06, 1.4828E+06, 1.4983E+06,
     +  1.5140E+06, 1.5299E+06, 1.5460E+06, 1.5622E+06, 1.5785E+06,
     + 1.5950E+06, 1.6117E+06, 1.6286E+06, 1.6456E+06, 1.6628E+06,
     +  1.6801E+06, 1.6977E+06, 1.7154E+06, 1.7332E+06, 1.7513E+06,
     + 1.7695E+06, 1.7880E+06, 1.8066E+06, 1.8253E+06, 1.8443E+06,
     +  1.8635E+06, 1.8828E+06, 1.9024E+06, 1.9221E+06, 1.9420E+06,
     + 1.9621E+06, 1.9825E+06, 2.0030E+06, 2.0237E+06, 2.0446E+06,
     +  2.0658E+06, 2.0871E+06, 2.1087E+06, 2.1304E+06, 2.1524E+06,
     + 2.1746E+06, 2.1970E+06, 2.2197E+06, 2.2425E+06, 2.2656E+06,
     +  2.2889E+06, 2.3124E+06, 2.3362E+06, 2.3602E+06, 2.3844E+06,
     + 2.4089E+06, 2.4336E+06, 2.4585E+06, 2.4837E+06, 2.5092E+06,
     +  2.5349E+06, 2.5608E+06, 2.5870E+06, 2.6134E+06, 2.6401E+06,
     + 2.6671E+06, 2.6943E+06, 2.7218E+06, 2.7495E+06, 2.7775E+06,
     +  2.8058E+06, 2.8344E+06, 2.8632E+06, 2.8924E+06, 2.9218E+06,
     + 2.9514E+06, 2.9814E+06, 3.0117E+06, 3.0422E+06, 3.0731E+06,
     +  3.1042E+06, 3.1357E+06, 3.1674E+06, 3.1995E+06, 3.2319E+06,
     + 3.2645E+06, 3.2975E+06, 3.3308E+06, 3.3644E+06, 3.3984E+06,
     +  3.4327E+06, 3.4673E+06, 3.5022E+06, 3.5375E+06, 3.5731E+06,
     + 3.6090E+06, 3.6453E+06, 3.6820E+06, 3.7190E+06, 3.7563E+06,
     +  3.7940E+06, 3.8320E+06, 3.8705E+06, 3.9093E+06, 3.9484E+06,
     + 3.9879E+06, 4.0278E+06, 4.0681E+06, 4.1088E+06, 4.1498E+06,
     +  4.1913E+06, 4.2331E+06, 4.2753E+06, 4.3180E+06, 4.3610E+06,
     + 4.4045E+06, 4.4483E+06, 4.4926E+06, 4.5373E+06, 4.5824E+06,
     +  4.6279E+06, 4.6739E+06, 4.7203E+06, 4.7671E+06, 4.8144E+06,
     + 4.8621E+06, 4.9103E+06, 4.9589E+06, 5.0080E+06, 5.0575E+06,
     +  5.1075E+06, 5.1580E+06, 5.2089E+06, 5.2604E+06, 5.3123E+06,
     + 5.3647E+06, 5.4176E+06, 5.4709E+06, 5.5248E+06, 5.5792E+06,
     +  5.6341E+06, 5.6895E+06, 5.7454E+06, 5.8019E+06, 5.8588E+06,
     + 5.9163E+06, 5.9743E+06, 6.0329E+06, 6.0920E+06, 6.1517E+06,
     +  6.2119E+06, 6.2727E+06, 6.3341E+06, 6.3960E+06, 6.4585E+06,
     + 6.5215E+06, 6.5852E+06, 6.6494E+06, 6.7142E+06, 6.7797E+06,
     +  6.8457E+06, 6.9124E+06, 6.9796E+06, 7.0475E+06, 7.1160E+06,
     + 7.1852E+06, 7.2549E+06, 7.3254E+06, 7.3964E+06, 7.4681E+06,
     +  7.5405E+06, 7.6136E+06, 7.6873E+06, 7.7617E+06, 7.8367E+06,
     + 7.9125E+06, 7.9890E+06, 8.0661E+06, 8.1440E+06, 8.2226E+06,
     +  8.3019E+06, 8.3819E+06, 8.4626E+06, 8.5441E+06, 8.6263E+06,
     + 8.7093E+06, 8.7930E+06, 8.8775E+06, 8.9628E+06, 9.0488E+06,
     +  9.1357E+06, 9.2233E+06, 9.3117E+06, 9.4009E+06, 9.4909E+06,
     + 9.5817E+06, 9.6734E+06, 9.7659E+06, 9.8592E+06, 9.9534E+06,
     +  1.0048E+07, 1.0144E+07, 1.0241E+07, 1.0339E+07, 1.0437E+07,
     + 1.0537E+07, 1.0637E+07, 1.0738E+07, 1.0840E+07, 1.0943E+07,
     +  1.1047E+07, 1.1152E+07, 1.1258E+07, 1.1365E+07, 1.1472E+07,
     + 1.1581E+07, 1.1691E+07, 1.1801E+07, 1.1913E+07, 1.2026E+07,
     +  1.2139E+07, 1.2254E+07, 1.2370E+07, 1.2486E+07, 1.2604E+07,
     + 1.2723E+07, 1.2843E+07, 1.2964E+07, 1.3086E+07, 1.3209E+07,
     +  1.3333E+07, 1.3458E+07, 1.3585E+07, 1.3712E+07, 1.3841E+07,
     + 1.3971E+07, 1.4102E+07, 1.4234E+07, 1.4367E+07, 1.4502E+07,
     +  1.4637E+07, 1.4774E+07, 1.4912E+07, 1.5052E+07, 1.5192E+07,
     + 1.5334E+07, 1.5477E+07, 1.5621E+07, 1.5767E+07, 1.5914E+07,
     +  1.6062E+07, 1.6211E+07, 1.6362E+07, 1.6514E+07, 1.6667E+07,
     + 1.6822E+07, 1.6978E+07, 1.7136E+07, 1.7295E+07, 1.7455E+07,
     +  1.7616E+07, 1.7779E+07, 1.7944E+07, 1.8110E+07, 1.8277E+07,
     + 1.8446E+07, 1.8616E+07, 1.8788E+07, 1.8961E+07, 1.9136E+07,
     +  1.9313E+07, 1.9490E+07, 1.9670E+07, 1.9851E+07, 2.0033E+07,
     + 2.0217E+07, 2.0403E+07, 2.0590E+07, 2.0779E+07, 2.0970E+07,
     +  2.1162E+07, 2.1356E+07, 2.1552E+07, 2.1749E+07, 2.1948E+07,
     + 2.2149E+07, 2.2351E+07, 2.2555E+07, 2.2761E+07, 2.2969E+07,
     +  2.3178E+07, 2.3390E+07, 2.3603E+07, 2.3818E+07, 2.4035E+07,
     + 2.4253E+07, 2.4474E+07, 2.4696E+07, 2.4920E+07, 2.5147E+07,
     +  2.5375E+07, 2.5605E+07, 2.5837E+07, 2.6071E+07, 2.6307E+07,
     + 2.6545E+07, 2.6786E+07, 2.7028E+07, 2.7272E+07, 2.7518E+07,
     +  2.7767E+07, 2.8017E+07, 2.8270E+07, 2.8525E+07, 2.8782E+07,
     + 2.9041E+07, 2.9303E+07, 2.9566E+07, 2.9832E+07, 3.0100E+07,
     +  3.0370E+07, 3.0643E+07, 3.0918E+07, 3.1195E+07, 3.1475E+07,
     + 3.1757E+07, 3.2041E+07, 3.2328E+07, 3.2617E+07, 3.2909E+07,
     +  3.3203E+07, 3.3500E+07, 3.3799E+07, 3.4100E+07, 3.4404E+07,
     + 3.4711E+07, 3.5020E+07, 3.5332E+07, 3.5647E+07, 3.5964E+07,
     +  3.6284E+07, 3.6606E+07, 3.6931E+07, 3.7259E+07, 3.7590E+07,
     + 3.7923E+07, 3.8259E+07, 3.8598E+07, 3.8940E+07, 3.9285E+07,
     +  3.9632E+07, 3.9982E+07, 4.0336E+07, 4.0692E+07, 4.1051E+07,
     + 4.1413E+07, 4.1779E+07, 4.2147E+07, 4.2518E+07, 4.2893E+07,
     +  4.3270E+07, 4.3651E+07, 4.4035E+07, 4.4422E+07, 4.4812E+07,
     + 4.5205E+07, 4.5602E+07, 4.6002E+07, 4.6405E+07, 4.6811E+07,
     +  4.7221E+07, 4.7635E+07, 4.8051E+07, 4.8471E+07, 4.8895E+07,
     + 4.9322E+07, 4.9753E+07, 5.0187E+07, 5.0624E+07, 5.1066E+07,
     +  5.1511E+07, 5.1959E+07, 5.2411E+07, 5.2867E+07, 5.3327E+07,
     + 5.3790E+07, 5.4257E+07, 5.4728E+07, 5.5203E+07, 5.5682E+07,
     +  5.6165E+07, 5.6651E+07, 5.7142E+07, 5.7636E+07, 5.8135E+07,
     + 5.8638E+07, 5.9144E+07, 5.9655E+07, 6.0170E+07, 6.0689E+07,
     +  6.1213E+07, 6.1740E+07, 6.2272E+07, 6.2808E+07, 6.3349E+07,
     + 6.3894E+07, 6.4443E+07, 6.4997E+07, 6.5555E+07, 6.6118E+07,
     +  6.6685E+07, 6.7257E+07, 6.7834E+07, 6.8415E+07, 6.9001E+07,
     + 6.9591E+07, 7.0186E+07, 7.0787E+07, 7.1391E+07, 7.2001E+07,
     +  7.2616E+07, 7.3235E+07, 7.3860E+07, 7.4490E+07, 7.5124E+07,
     + 7.5764E+07, 7.6409E+07, 7.7059E+07, 7.7714E+07, 7.8374E+07,
     +  7.9040E+07, 7.9711E+07, 8.0387E+07, 8.1069E+07, 8.1756E+07,
     + 8.2449E+07, 8.3147E+07, 8.3851E+07, 8.4560E+07, 8.5275E+07,
     +  8.5996E+07, 8.6722E+07, 8.7454E+07, 8.8192E+07, 8.8936E+07,
     + 8.9686E+07, 9.0441E+07, 9.1203E+07, 9.1970E+07, 9.2744E+07,
     +  9.3524E+07, 9.4310E+07, 9.5102E+07, 9.5901E+07, 9.6705E+07,
     + 9.7516E+07, 9.8334E+07, 9.9158E+07, 9.9988E+07, 1.0082E+08,
     +  1.0167E+08, 1.0252E+08, 1.0338E+08, 1.0424E+08, 1.0511E+08,
     + 1.0599E+08, 1.0687E+08, 1.0776E+08, 1.0866E+08, 1.0956E+08,
     +  1.1048E+08, 1.1139E+08, 1.1232E+08, 1.1325E+08, 1.1420E+08,
     + 1.1514E+08, 1.1610E+08, 1.1706E+08, 1.1803E+08, 1.1901E+08,
     +  1.2000E+08, 1.2099E+08, 1.2199E+08, 1.2300E+08, 1.2402E+08,
     + 1.2504E+08, 1.2607E+08, 1.2711E+08, 1.2816E+08, 1.2922E+08,
     +  1.3028E+08, 1.3136E+08, 1.3244E+08, 1.3353E+08, 1.3463E+08,
     + 1.3573E+08, 1.3685E+08, 1.3797E+08, 1.3910E+08, 1.4024E+08,
     +  1.4139E+08, 1.4255E+08, 1.4372E+08, 1.4490E+08, 1.4608E+08,
     + 1.4728E+08, 1.4848E+08, 1.4970E+08, 1.5092E+08, 1.5215E+08,
     +  1.5339E+08, 1.5464E+08, 1.5590E+08, 1.5717E+08, 1.5845E+08,
     + 1.5974E+08, 1.6104E+08, 1.6235E+08, 1.6367E+08, 1.6500E+08,
     +  1.6634E+08, 1.6769E+08, 1.6905E+08, 1.7042E+08, 1.7180E+08,
     + 1.7319E+08, 1.7459E+08, 1.7600E+08, 1.7743E+08, 1.7886E+08,
     +  1.8030E+08, 1.8176E+08, 1.8323E+08, 1.8470E+08, 1.8619E+08,
     + 1.8769E+08, 1.8920E+08, 1.9073E+08, 1.9226E+08, 1.9381E+08,
     +  1.9537E+08, 1.9694E+08, 1.9852E+08, 2.0011E+08, 2.0171E+08,
     + 2.0333E+08, 2.0496E+08, 2.0660E+08, 2.0826E+08, 2.0992E+08,
     +  2.1160E+08, 2.1329E+08, 2.1500E+08, 2.1671E+08, 2.1844E+08,
     + 2.2019E+08, 2.2194E+08, 2.2371E+08, 2.2549E+08, 2.2729E+08,
     +  2.2910E+08, 2.3092E+08, 2.3275E+08, 2.3460E+08, 2.3647E+08,
     + 2.3834E+08, 2.4023E+08, 2.4214E+08, 2.4406E+08, 2.4599E+08,
     +  2.4794E+08, 2.4990E+08, 2.5188E+08, 2.5387E+08, 2.5588E+08,
     + 2.5790E+08, 2.5993E+08, 2.6199E+08, 2.6405E+08, 2.6613E+08,
     +  2.6823E+08, 2.7034E+08, 2.7247E+08, 2.7461E+08, 2.7677E+08,
     + 2.7895E+08, 2.8114E+08, 2.8335E+08, 2.8557E+08, 2.8781E+08,
     +  2.9007E+08, 2.9234E+08, 2.9463E+08, 2.9694E+08, 2.9926E+08,
     + 3.0160E+08, 3.0396E+08, 3.0633E+08, 3.0873E+08, 3.1114E+08,
     +  3.1356E+08, 3.1601E+08, 3.1847E+08, 3.2095E+08, 3.2345E+08,
     + 3.2597E+08, 3.2851E+08, 3.3106E+08, 3.3363E+08, 3.3622E+08,
     +  3.3884E+08, 3.4146E+08, 3.4411E+08, 3.4678E+08, 3.4947E+08,
     + 3.5217E+08, 3.5490E+08, 3.5765E+08, 3.6041E+08, 3.6320E+08,
     +  3.6600E+08, 3.6883E+08, 3.7168E+08, 3.7455E+08, 3.7743E+08,
     + 3.8034E+08, 3.8327E+08, 3.8622E+08, 3.8920E+08, 3.9219E+08,
     +  3.9520E+08, 3.9824E+08, 4.0130E+08, 4.0438E+08, 4.0748E+08,
     + 4.1061E+08, 4.1376E+08, 4.1693E+08, 4.2012E+08, 4.2334E+08,
     +  4.2657E+08, 4.2984E+08, 4.3312E+08, 4.3643E+08, 4.3976E+08,
     + 4.4312E+08, 4.4650E+08, 4.4991E+08, 4.5333E+08, 4.5679E+08,
     +  4.6027E+08, 4.6377E+08, 4.6730E+08, 4.7085E+08, 4.7443E+08,
     + 4.7803E+08, 4.8166E+08, 4.8531E+08, 4.8899E+08, 4.9270E+08,
     +  4.9643E+08, 5.0019E+08, 5.0398E+08, 5.0779E+08, 5.1163E+08,
     + 5.1550E+08, 5.1940E+08, 5.2332E+08, 5.2727E+08, 5.3125E+08,
     +  5.3525E+08, 5.3929E+08, 5.4335E+08, 5.4744E+08, 5.5156E+08,
     + 5.5571E+08, 5.5989E+08, 5.6409E+08, 5.6833E+08, 5.7260E+08,
     +  5.7690E+08, 5.8122E+08, 5.8558E+08, 5.8997E+08, 5.9439E+08,
     + 5.9884E+08, 6.0332E+08, 6.0783E+08, 6.1237E+08, 6.1695E+08,
     +  6.2156E+08, 6.2620E+08, 6.3087E+08, 6.3557E+08, 6.4031E+08,
     + 6.4508E+08, 6.4989E+08, 6.5472E+08, 6.5959E+08, 6.6450E+08,
     +  6.6944E+08, 6.7441E+08, 6.7942E+08, 6.8446E+08, 6.8954E+08,
     + 6.9465E+08, 6.9980E+08, 7.0498E+08, 7.1020E+08, 7.1546E+08,
     +  7.2075E+08, 7.2608E+08, 7.3145E+08, 7.3685E+08, 7.4229E+08,
     + 7.4777E+08, 7.5328E+08, 7.5884E+08, 7.6443E+08, 7.7006E+08,
     +  7.7573E+08, 7.8143E+08, 7.8718E+08, 7.9297E+08, 7.9880E+08,
     + 8.0466E+08, 8.1057E+08, 8.1652E+08, 8.2251E+08, 8.2853E+08,
     +  8.3461E+08, 8.4072E+08, 8.4687E+08, 8.5307E+08, 8.5931E+08,
     + 8.6559E+08, 8.7191E+08, 8.7828E+08, 8.8469E+08, 8.9114E+08,
     +  8.9764E+08, 9.0419E+08, 9.1077E+08, 9.1740E+08, 9.2408E+08,
     + 9.3080E+08, 9.3757E+08, 9.4439E+08, 9.5125E+08, 9.5816E+08,
     +  9.6511E+08, 9.7211E+08, 9.7916E+08, 9.8626E+08, 9.9340E+08,
     + 1.0006E+09, 1.0078E+09, 1.0151E+09, 1.0225E+09, 1.0299E+09,
     +  1.0373E+09, 1.0448E+09, 1.0523E+09, 1.0599E+09, 1.0676E+09,
     + 1.0753E+09, 1.0830E+09, 1.0908E+09, 1.0986E+09, 1.1065E+09,
     +  1.1145E+09, 1.1225E+09, 1.1306E+09, 1.1387E+09, 1.1469E+09,
     + 1.1551E+09, 1.1634E+09, 1.1717E+09, 1.1801E+09, 1.1886E+09,
     +  1.1971E+09, 1.2056E+09, 1.2142E+09, 1.2229E+09, 1.2317E+09,
     + 1.2404E+09, 1.2493E+09, 1.2582E+09, 1.2672E+09, 1.2762E+09,
     +  1.2853E+09, 1.2945E+09, 1.3037E+09, 1.3129E+09, 1.3223E+09,
     + 1.3317E+09, 1.3411E+09, 1.3506E+09, 1.3602E+09, 1.3699E+09,
     +  1.3796E+09, 1.3894E+09, 1.3992E+09, 1.4091E+09, 1.4191E+09,
     + 1.4291E+09, 1.4392E+09, 1.4494E+09, 1.4596E+09, 1.4699E+09,
     +  1.4803E+09, 1.4908E+09, 1.5013E+09, 1.5119E+09, 1.5225E+09,
     + 1.5332E+09, 1.5440E+09, 1.5549E+09, 1.5658E+09, 1.5768E+09,
     +  1.5879E+09, 1.5990E+09, 1.6103E+09, 1.6215E+09, 1.6329E+09,
     + 1.6444E+09, 1.6559E+09, 1.6675E+09, 1.6791E+09, 1.6909E+09,
     +  1.7027E+09, 1.7146E+09, 1.7266E+09, 1.7386E+09, 1.7508E+09,
     + 1.7630E+09, 1.7753E+09, 1.7877E+09, 1.8001E+09, 1.8126E+09,
     +  1.8253E+09, 1.8380E+09, 1.8507E+09, 1.8636E+09, 1.8766E+09,
     + 1.8896E+09, 1.9027E+09, 1.9159E+09, 1.9292E+09, 1.9426E+09,
     +  1.9560E+09, 1.9696E+09, 1.9832E+09, 1.9969E+09, 2.0107E+09,
     + 2.0246E+09, 2.0386E+09, 2.0527E+09, 2.0669E+09, 2.0811E+09,
     +  2.0955E+09, 2.1099E+09, 2.1245E+09, 2.1391E+09, 2.1538E+09,
     + 2.1686E+09, 2.1835E+09, 2.1986E+09, 2.2137E+09, 2.2289E+09,
     +  2.2442E+09, 2.2596E+09, 2.2751E+09, 2.2907E+09, 2.3064E+09,
     + 2.3222E+09, 2.3381E+09, 2.3541E+09, 2.3702E+09, 2.3864E+09,
     +  2.4027E+09, 2.4191E+09, 2.4356E+09, 2.4522E+09, 2.4689E+09,
     + 2.4858E+09, 2.5027E+09, 2.5198E+09, 2.5369E+09, 2.5542E+09,
     +  2.5716E+09, 2.5891E+09, 2.6067E+09, 2.6244E+09, 2.6422E+09,
     + 2.6601E+09, 2.6782E+09, 2.6963E+09, 2.7146E+09, 2.7330E+09,
     +  2.7515E+09, 2.7701E+09, 2.7889E+09, 2.8077E+09, 2.8267E+09,
     + 2.8458E+09, 2.8650E+09, 2.8844E+09, 2.9039E+09, 2.9234E+09,
     +  2.9431E+09, 2.9630E+09, 2.9829E+09, 3.0030E+09, 3.0232E+09,
     + 3.0436E+09, 3.0640E+09, 3.0846E+09, 3.1053E+09, 3.1262E+09,
     +  3.1472E+09, 3.1683E+09, 3.1895E+09, 3.2109E+09, 3.2324E+09,
     + 3.2541E+09, 3.2759E+09, 3.2978E+09, 3.3198E+09, 3.3420E+09,
     +  3.3644E+09, 3.3868E+09, 3.4094E+09, 3.4322E+09, 3.4551E+09,
     + 3.4781E+09, 3.5013E+09, 3.5246E+09, 3.5481E+09, 3.5717E+09,
     +  3.5955E+09, 3.6194E+09, 3.6434E+09, 3.6676E+09, 3.6920E+09,
     + 3.7165E+09, 3.7411E+09, 3.7659E+09, 3.7909E+09, 3.8160E+09,
     +  3.8413E+09, 3.8667E+09, 3.8923E+09, 3.9181E+09, 3.9440E+09,
     + 3.9700E+09, 3.9962E+09, 4.0226E+09, 4.0492E+09, 4.0759E+09,
     +  4.1028E+09, 4.1298E+09, 4.1570E+09, 4.1844E+09, 4.2119E+09,
     + 4.2396E+09, 4.2675E+09, 4.2956E+09, 4.3238E+09, 4.3522E+09,
     +  4.3807E+09, 4.4095E+09, 4.4384E+09, 4.4675E+09, 4.4968E+09,
     + 4.5262E+09, 4.5559E+09, 4.5857E+09, 4.6157E+09, 4.6458E+09,
     +  4.6762E+09, 4.7068E+09, 4.7375E+09, 4.7684E+09, 4.7995E+09,
     + 4.8308E+09, 4.8623E+09, 4.8940E+09, 4.9259E+09, 4.9579E+09,
     +  4.9902E+09, 5.0226E+09, 5.0553E+09, 5.0882E+09, 5.1212E+09,
     + 5.1545E+09, 5.1879E+09, 5.2216E+09, 5.2554E+09, 5.2895E+09,
     +  5.3237E+09, 5.3582E+09, 5.3929E+09, 5.4278E+09, 5.4629E+09,
     + 5.4982E+09, 5.5337E+09, 5.5695E+09, 5.6054E+09, 5.6416E+09,
     +  5.6780E+09, 5.7146E+09, 5.7514E+09, 5.7885E+09, 5.8258E+09,
     + 5.8633E+09, 5.9010E+09, 5.9389E+09, 5.9771E+09, 6.0155E+09,
     +  6.0541E+09, 6.0930E+09, 6.1321E+09, 6.1714E+09, 6.2110E+09,
     + 6.2508E+09, 6.2908E+09, 6.3311E+09, 6.3716E+09, 6.4124E+09,
     +  6.4534E+09, 6.4947E+09, 6.5362E+09, 6.5779E+09, 6.6199E+09,
     + 6.6621E+09, 6.7046E+09, 6.7474E+09, 6.7904E+09, 6.8336E+09,
     +  6.8771E+09, 6.9209E+09, 6.9649E+09, 7.0092E+09, 7.0538E+09,
     + 7.0986E+09, 7.1437E+09, 7.1890E+09, 7.2346E+09, 7.2805E+09,
     +  7.3267E+09, 7.3731E+09, 7.4198E+09, 7.4668E+09, 7.5140E+09,
     + 7.5616E+09, 7.6094E+09, 7.6575E+09, 7.7059E+09, 7.7545E+09,
     +  7.8035E+09, 7.8527E+09, 7.9023E+09, 7.9521E+09, 8.0022E+09,
     + 8.0526E+09, 8.1033E+09, 8.1543E+09, 8.2056E+09, 8.2572E+09,
     +  8.3091E+09, 8.3613E+09, 8.4138E+09, 8.4666E+09, 8.5198E+09,
     + 8.5732E+09, 8.6270E+09, 8.6810E+09, 8.7354E+09, 8.7901E+09,
     +  8.8451E+09, 8.9005E+09, 8.9561E+09, 9.0121E+09, 9.0684E+09,
     + 9.1251E+09, 9.1820E+09, 9.2393E+09, 9.2970E+09, 9.3549E+09,
     +  9.4132E+09, 9.4719E+09, 9.5309E+09, 9.5902E+09, 9.6499E+09,
     + 9.7099E+09, 9.7702E+09, 9.8310E+09, 9.8920E+09, 9.9534E+09,
     +  1.0015E+10, 1.0077E+10, 1.0140E+10, 1.0203E+10, 1.0266E+10,
     + 1.0329E+10, 1.0393E+10, 1.0458E+10, 1.0522E+10, 1.0587E+10,
     +  1.0653E+10, 1.0719E+10, 1.0785E+10, 1.0851E+10, 1.0918E+10,
     + 1.0986E+10, 1.1053E+10, 1.1122E+10, 1.1190E+10, 1.1259E+10,
     +  1.1328E+10, 1.1398E+10, 1.1468E+10, 1.1539E+10, 1.1609E+10,
     + 1.1681E+10, 1.1752E+10, 1.1825E+10, 1.1897E+10, 1.1970E+10,
     +  1.2043E+10, 1.2117E+10, 1.2191E+10, 1.2266E+10, 1.2341E+10,
     + 1.2417E+10, 1.2492E+10, 1.2569E+10, 1.2646E+10, 1.2723E+10,
     +  1.2800E+10, 1.2878E+10, 1.2957E+10, 1.3036E+10, 1.3115E+10,
     + 1.3195E+10, 1.3276E+10, 1.3356E+10, 1.3438E+10, 1.3519E+10,
     +  1.3602E+10, 1.3684E+10, 1.3767E+10, 1.3851E+10, 1.3935E+10,
     + 1.4019E+10, 1.4104E+10, 1.4190E+10, 1.4276E+10, 1.4362E+10,
     +  1.4449E+10, 1.4537E+10, 1.4625E+10, 1.4713E+10, 1.4802E+10,
     + 1.4891E+10, 1.4981E+10, 1.5072E+10, 1.5162E+10, 1.5254E+10,
     +  1.5346E+10, 1.5438E+10, 1.5531E+10, 1.5625E+10, 1.5719E+10,
     + 1.5813E+10, 1.5908E+10, 1.6004E+10, 1.6100E+10, 1.6197E+10,
     +  1.6294E+10, 1.6392E+10, 1.6490E+10, 1.6589E+10, 1.6688E+10,
     + 1.6788E+10, 1.6889E+10, 1.6990E+10, 1.7092E+10, 1.7194E+10,
     +  1.7297E+10, 1.7400E+10, 1.7504E+10, 1.7609E+10, 1.7714E+10,
     + 1.7819E+10, 1.7926E+10, 1.8032E+10, 1.8140E+10, 1.8248E+10,
     +  1.8357E+10, 1.8466E+10, 1.8576E+10, 1.8686E+10, 1.8797E+10,
     + 1.8909E+10, 1.9021E+10, 1.9134E+10, 1.9248E+10, 1.9362E+10,
     +  1.9477E+10, 1.9592E+10, 1.9708E+10, 1.9825E+10, 1.9942E+10,
     + 2.0060E+10, 2.0179E+10, 2.0298E+10, 2.0418E+10, 2.0539E+10,
     +  2.0660E+10, 2.0782E+10, 2.0905E+10, 2.1028E+10, 2.1152E+10,
     + 2.1277E+10, 2.1402E+10, 2.1528E+10, 2.1655E+10, 2.1782E+10,
     +  2.1910E+10, 2.2039E+10, 2.2169E+10, 2.2299E+10, 2.2430E+10,
     + 2.2562E+10, 2.2694E+10, 2.2827E+10, 2.2961E+10, 2.3095E+10,
     +  2.3231E+10, 2.3367E+10, 2.3504E+10, 2.3641E+10, 2.3779E+10,
     + 2.3918E+10, 2.4058E+10, 2.4199E+10, 2.4340E+10, 2.4482E+10,
     +  2.4625E+10, 2.4768E+10, 2.4913E+10, 2.5058E+10, 2.5204E+10,
     + 2.5351E+10, 2.5498E+10, 2.5647E+10, 2.5796E+10, 2.5946E+10,
     +  2.6097E+10, 2.6248E+10, 2.6401E+10, 2.6554E+10, 2.6708E+10,
     + 2.6863E+10, 2.7019E+10, 2.7175E+10, 2.7333E+10, 2.7491E+10,
     +  2.7650E+10, 2.7810E+10, 2.7971E+10, 2.8132E+10, 2.8295E+10,
     + 2.8458E+10, 2.8623E+10, 2.8788E+10, 2.8954E+10, 2.9121E+10,
     +  2.9289E+10, 2.9458E+10, 2.9627E+10, 2.9798E+10, 2.9970E+10,
     + 3.0142E+10, 3.0315E+10, 3.0490E+10, 3.0665E+10, 3.0841E+10,
     +  3.1018E+10, 3.1196E+10, 3.1375E+10, 3.1555E+10, 3.1736E+10,
     + 3.1918E+10, 3.2101E+10, 3.2285E+10, 3.2469E+10, 3.2655E+10,
     +  3.2842E+10, 3.3030E+10, 3.3218E+10, 3.3408E+10, 3.3599E+10,
     + 3.3791E+10, 3.3984E+10, 3.4177E+10, 3.4372E+10, 3.4568E+10,
     +  3.4765E+10, 3.4963E+10, 3.5162E+10, 3.5362E+10, 3.5563E+10,
     + 3.5766E+10, 3.5969E+10, 3.6173E+10, 3.6379E+10, 3.6585E+10,
     +  3.6793E+10, 3.7001E+10, 3.7211E+10, 3.7422E+10, 3.7634E+10,
     + 3.7847E+10, 3.8062E+10, 3.8277E+10, 3.8494E+10, 3.8711E+10,
     +  3.8930E+10, 3.9150E+10, 3.9371E+10, 3.9593E+10, 3.9817E+10,
     + 4.0041E+10, 4.0267E+10, 4.0494E+10, 4.0722E+10, 4.0952E+10,
     +  4.1182E+10, 4.1414E+10, 4.1647E+10, 4.1881E+10, 4.2117E+10,
     + 4.2354E+10, 4.2591E+10, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_CH3Br (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 2),QofT( 2,NT),Q(NT)
      data (Tmx(I),I=1,2)/3500.,3500./
      data xgj/ 4.,4./
c...    CH3Br
c...        --       219
      data (QofT( 1,J),J=1,2408)/ 3.0049E+01, 7.0283E+01, 1.1601E+02,
     +  1.6607E+02, 2.2005E+02, 2.7767E+02, 3.3871E+02, 4.0295E+02,
     + 4.7024E+02, 5.4047E+02, 6.1357E+02, 6.8949E+02, 7.6818E+02,
     +  8.4964E+02, 9.3383E+02, 1.0207E+03, 1.1104E+03, 1.2026E+03,
     + 1.2976E+03, 1.3951E+03, 1.5979E+03, 1.8107E+03, 2.0332E+03,
     +  2.2651E+03, 2.5061E+03, 2.7557E+03, 3.0138E+03, 3.2801E+03,
     + 3.5542E+03, 3.8359E+03, 4.1250E+03, 4.4213E+03, 4.7246E+03,
     +  5.0347E+03, 5.3514E+03, 5.6747E+03, 6.0043E+03, 6.3401E+03,
     + 6.6820E+03, 7.0299E+03, 7.3837E+03, 7.7432E+03, 8.1085E+03,
     +  8.4793E+03, 8.8557E+03, 9.2374E+03, 9.6246E+03, 1.0017E+04,
     + 1.0415E+04, 1.0817E+04, 1.1225E+04, 1.1638E+04, 1.2056E+04,
     +  1.2479E+04, 1.2907E+04, 1.3339E+04, 1.3776E+04, 1.4218E+04,
     + 1.4665E+04, 1.5116E+04, 1.5572E+04, 1.6033E+04, 1.6498E+04,
     +  1.6968E+04, 1.7442E+04, 1.7920E+04, 1.8403E+04, 1.8891E+04,
     + 1.9383E+04, 1.9879E+04, 2.0380E+04, 2.0885E+04, 2.1395E+04,
     +  2.1908E+04, 2.2427E+04, 2.2949E+04, 2.3476E+04, 2.4007E+04,
     + 2.4543E+04, 2.5083E+04, 2.5627E+04, 2.6176E+04, 2.6729E+04,
     +  2.7286E+04, 2.7848E+04, 2.8414E+04, 2.8984E+04, 2.9559E+04,
     + 3.0139E+04, 3.0723E+04, 3.1311E+04, 3.1904E+04, 3.2501E+04,
     +  3.3103E+04, 3.3710E+04, 3.4321E+04, 3.4937E+04, 3.5557E+04,
     + 3.6182E+04, 3.6812E+04, 3.7447E+04, 3.8086E+04, 3.8730E+04,
     +  3.9379E+04, 4.0033E+04, 4.0692E+04, 4.1356E+04, 4.2025E+04,
     + 4.2699E+04, 4.3378E+04, 4.4062E+04, 4.4752E+04, 4.5446E+04,
     +  4.6146E+04, 4.6851E+04, 4.7562E+04, 4.8278E+04, 4.9000E+04,
     + 4.9727E+04, 5.0459E+04, 5.1198E+04, 5.1942E+04, 5.2691E+04,
     +  5.3447E+04, 5.4208E+04, 5.4976E+04, 5.5749E+04, 5.6528E+04,
     + 5.7314E+04, 5.8105E+04, 5.8903E+04, 5.9707E+04, 6.0518E+04,
     +  6.1335E+04, 6.2158E+04, 6.2988E+04, 6.3825E+04, 6.4668E+04,
     + 6.5518E+04, 6.6375E+04, 6.7239E+04, 6.8110E+04, 6.8987E+04,
     +  6.9872E+04, 7.0764E+04, 7.1664E+04, 7.2570E+04, 7.3485E+04,
     + 7.4406E+04, 7.5336E+04, 7.6272E+04, 7.7217E+04, 7.8170E+04,
     +  7.9130E+04, 8.0098E+04, 8.1075E+04, 8.2059E+04, 8.3052E+04,
     + 8.4053E+04, 8.5063E+04, 8.6081E+04, 8.7107E+04, 8.8142E+04,
     +  8.9186E+04, 9.0239E+04, 9.1301E+04, 9.2372E+04, 9.3452E+04,
     + 9.4541E+04, 9.5639E+04, 9.6747E+04, 9.7865E+04, 9.8992E+04,
     +  1.0013E+05, 1.0127E+05, 1.0243E+05, 1.0360E+05, 1.0477E+05,
     + 1.0596E+05, 1.0716E+05, 1.0836E+05, 1.0958E+05, 1.1081E+05,
     +  1.1205E+05, 1.1330E+05, 1.1456E+05, 1.1583E+05, 1.1712E+05,
     + 1.1841E+05, 1.1972E+05, 1.2103E+05, 1.2236E+05, 1.2370E+05,
     +  1.2505E+05, 1.2642E+05, 1.2779E+05, 1.2918E+05, 1.3058E+05,
     + 1.3200E+05, 1.3342E+05, 1.3486E+05, 1.3631E+05, 1.3777E+05,
     +  1.3925E+05, 1.4074E+05, 1.4224E+05, 1.4376E+05, 1.4529E+05,
     + 1.4683E+05, 1.4839E+05, 1.4996E+05, 1.5155E+05, 1.5314E+05,
     +  1.5476E+05, 1.5639E+05, 1.5803E+05, 1.5969E+05, 1.6136E+05,
     + 1.6304E+05, 1.6475E+05, 1.6646E+05, 1.6820E+05, 1.6994E+05,
     +  1.7171E+05, 1.7349E+05, 1.7528E+05, 1.7709E+05, 1.7892E+05,
     + 1.8077E+05, 1.8263E+05, 1.8450E+05, 1.8640E+05, 1.8831E+05,
     +  1.9024E+05, 1.9218E+05, 1.9415E+05, 1.9613E+05, 1.9812E+05,
     + 2.0014E+05, 2.0218E+05, 2.0423E+05, 2.0630E+05, 2.0839E+05,
     +  2.1050E+05, 2.1262E+05, 2.1477E+05, 2.1694E+05, 2.1912E+05,
     + 2.2133E+05, 2.2355E+05, 2.2579E+05, 2.2806E+05, 2.3034E+05,
     +  2.3265E+05, 2.3497E+05, 2.3732E+05, 2.3969E+05, 2.4208E+05,
     + 2.4449E+05, 2.4692E+05, 2.4937E+05, 2.5185E+05, 2.5434E+05,
     +  2.5686E+05, 2.5940E+05, 2.6197E+05, 2.6456E+05, 2.6717E+05,
     + 2.6980E+05, 2.7246E+05, 2.7514E+05, 2.7785E+05, 2.8057E+05,
     +  2.8333E+05, 2.8611E+05, 2.8891E+05, 2.9174E+05, 2.9459E+05,
     + 2.9747E+05, 3.0037E+05, 3.0330E+05, 3.0626E+05, 3.0924E+05,
     +  3.1225E+05, 3.1528E+05, 3.1834E+05, 3.2143E+05, 3.2455E+05,
     + 3.2769E+05, 3.3087E+05, 3.3407E+05, 3.3729E+05, 3.4055E+05,
     +  3.4384E+05, 3.4715E+05, 3.5049E+05, 3.5387E+05, 3.5727E+05,
     + 3.6070E+05, 3.6417E+05, 3.6766E+05, 3.7119E+05, 3.7474E+05,
     +  3.7833E+05, 3.8195E+05, 3.8560E+05, 3.8928E+05, 3.9299E+05,
     + 3.9674E+05, 4.0052E+05, 4.0433E+05, 4.0818E+05, 4.1206E+05,
     +  4.1598E+05, 4.1992E+05, 4.2391E+05, 4.2792E+05, 4.3198E+05,
     + 4.3607E+05, 4.4019E+05, 4.4435E+05, 4.4855E+05, 4.5278E+05,
     +  4.5705E+05, 4.6135E+05, 4.6570E+05, 4.7008E+05, 4.7450E+05,
     + 4.7896E+05, 4.8346E+05, 4.8799E+05, 4.9257E+05, 4.9718E+05,
     +  5.0184E+05, 5.0653E+05, 5.1127E+05, 5.1604E+05, 5.2086E+05,
     + 5.2572E+05, 5.3062E+05, 5.3557E+05, 5.4055E+05, 5.4558E+05,
     +  5.5065E+05, 5.5577E+05, 5.6093E+05, 5.6613E+05, 5.7138E+05,
     + 5.7668E+05, 5.8202E+05, 5.8740E+05, 5.9283E+05, 5.9831E+05,
     +  6.0384E+05, 6.0941E+05, 6.1503E+05, 6.2070E+05, 6.2641E+05,
     + 6.3218E+05, 6.3799E+05, 6.4385E+05, 6.4977E+05, 6.5573E+05,
     +  6.6175E+05, 6.6781E+05, 6.7393E+05, 6.8010E+05, 6.8632E+05,
     + 6.9259E+05, 6.9892E+05, 7.0530E+05, 7.1174E+05, 7.1823E+05,
     +  7.2477E+05, 7.3137E+05, 7.3803E+05, 7.4474E+05, 7.5151E+05,
     + 7.5833E+05, 7.6522E+05, 7.7216E+05, 7.7916E+05, 7.8622E+05,
     +  7.9333E+05, 8.0051E+05, 8.0775E+05, 8.1505E+05, 8.2240E+05,
     + 8.2983E+05, 8.3731E+05, 8.4485E+05, 8.5246E+05, 8.6014E+05,
     +  8.6787E+05, 8.7567E+05, 8.8354E+05, 8.9147E+05, 8.9947E+05,
     + 9.0753E+05, 9.1567E+05, 9.2387E+05, 9.3213E+05, 9.4047E+05,
     +  9.4887E+05, 9.5735E+05, 9.6590E+05, 9.7451E+05, 9.8320E+05,
     + 9.9196E+05, 1.0008E+06, 1.0097E+06, 1.0187E+06, 1.0277E+06,
     +  1.0369E+06, 1.0461E+06, 1.0553E+06, 1.0647E+06, 1.0741E+06,
     + 1.0836E+06, 1.0932E+06, 1.1029E+06, 1.1126E+06, 1.1225E+06,
     +  1.1324E+06, 1.1423E+06, 1.1524E+06, 1.1626E+06, 1.1728E+06,
     + 1.1831E+06, 1.1935E+06, 1.2040E+06, 1.2146E+06, 1.2252E+06,
     +  1.2360E+06, 1.2468E+06, 1.2577E+06, 1.2688E+06, 1.2799E+06,
     + 1.2910E+06, 1.3023E+06, 1.3137E+06, 1.3252E+06, 1.3367E+06,
     +  1.3484E+06, 1.3601E+06, 1.3720E+06, 1.3839E+06, 1.3959E+06,
     + 1.4081E+06, 1.4203E+06, 1.4326E+06, 1.4450E+06, 1.4576E+06,
     +  1.4702E+06, 1.4829E+06, 1.4957E+06, 1.5087E+06, 1.5217E+06,
     + 1.5348E+06, 1.5481E+06, 1.5614E+06, 1.5749E+06, 1.5885E+06,
     +  1.6021E+06, 1.6159E+06, 1.6298E+06, 1.6438E+06, 1.6580E+06,
     + 1.6722E+06, 1.6865E+06, 1.7010E+06, 1.7156E+06, 1.7302E+06,
     +  1.7451E+06, 1.7600E+06, 1.7750E+06, 1.7902E+06, 1.8055E+06,
     + 1.8209E+06, 1.8364E+06, 1.8520E+06, 1.8678E+06, 1.8837E+06,
     +  1.8997E+06, 1.9159E+06, 1.9321E+06, 1.9485E+06, 1.9651E+06,
     + 1.9817E+06, 1.9985E+06, 2.0154E+06, 2.0325E+06, 2.0497E+06,
     +  2.0670E+06, 2.0845E+06, 2.1021E+06, 2.1198E+06, 2.1377E+06,
     + 2.1557E+06, 2.1739E+06, 2.1922E+06, 2.2106E+06, 2.2292E+06,
     +  2.2479E+06, 2.2668E+06, 2.2858E+06, 2.3050E+06, 2.3243E+06,
     + 2.3438E+06, 2.3634E+06, 2.3832E+06, 2.4031E+06, 2.4232E+06,
     +  2.4434E+06, 2.4638E+06, 2.4844E+06, 2.5051E+06, 2.5259E+06,
     + 2.5470E+06, 2.5682E+06, 2.5895E+06, 2.6110E+06, 2.6327E+06,
     +  2.6546E+06, 2.6766E+06, 2.6988E+06, 2.7212E+06, 2.7437E+06,
     + 2.7664E+06, 2.7893E+06, 2.8123E+06, 2.8356E+06, 2.8590E+06,
     +  2.8826E+06, 2.9064E+06, 2.9303E+06, 2.9544E+06, 2.9788E+06,
     + 3.0033E+06, 3.0280E+06, 3.0528E+06, 3.0779E+06, 3.1032E+06,
     +  3.1286E+06, 3.1543E+06, 3.1801E+06, 3.2061E+06, 3.2324E+06,
     + 3.2588E+06, 3.2854E+06, 3.3123E+06, 3.3393E+06, 3.3666E+06,
     +  3.3940E+06, 3.4217E+06, 3.4495E+06, 3.4776E+06, 3.5059E+06,
     + 3.5344E+06, 3.5631E+06, 3.5920E+06, 3.6212E+06, 3.6505E+06,
     +  3.6801E+06, 3.7099E+06, 3.7400E+06, 3.7702E+06, 3.8007E+06,
     + 3.8314E+06, 3.8624E+06, 3.8935E+06, 3.9249E+06, 3.9566E+06,
     +  3.9885E+06, 4.0206E+06, 4.0529E+06, 4.0855E+06, 4.1184E+06,
     + 4.1514E+06, 4.1848E+06, 4.2183E+06, 4.2522E+06, 4.2862E+06,
     +  4.3206E+06, 4.3552E+06, 4.3900E+06, 4.4251E+06, 4.4604E+06,
     + 4.4961E+06, 4.5319E+06, 4.5681E+06, 4.6045E+06, 4.6412E+06,
     +  4.6781E+06, 4.7154E+06, 4.7529E+06, 4.7906E+06, 4.8287E+06,
     + 4.8670E+06, 4.9056E+06, 4.9445E+06, 4.9837E+06, 5.0232E+06,
     +  5.0629E+06, 5.1030E+06, 5.1433E+06, 5.1839E+06, 5.2249E+06,
     + 5.2661E+06, 5.3076E+06, 5.3495E+06, 5.3916E+06, 5.4340E+06,
     +  5.4768E+06, 5.5199E+06, 5.5632E+06, 5.6069E+06, 5.6510E+06,
     + 5.6953E+06, 5.7399E+06, 5.7849E+06, 5.8302E+06, 5.8758E+06,
     +  5.9218E+06, 5.9681E+06, 6.0147E+06, 6.0617E+06, 6.1090E+06,
     + 6.1566E+06, 6.2046E+06, 6.2529E+06, 6.3016E+06, 6.3507E+06,
     +  6.4000E+06, 6.4498E+06, 6.4999E+06, 6.5503E+06, 6.6011E+06,
     + 6.6523E+06, 6.7039E+06, 6.7558E+06, 6.8081E+06, 6.8607E+06,
     +  6.9138E+06, 6.9672E+06, 7.0210E+06, 7.0752E+06, 7.1297E+06,
     + 7.1847E+06, 7.2400E+06, 7.2958E+06, 7.3519E+06, 7.4084E+06,
     +  7.4654E+06, 7.5227E+06, 7.5805E+06, 7.6386E+06, 7.6972E+06,
     + 7.7562E+06, 7.8156E+06, 7.8754E+06, 7.9356E+06, 7.9963E+06,
     +  8.0574E+06, 8.1189E+06, 8.1809E+06, 8.2433E+06, 8.3061E+06,
     + 8.3694E+06, 8.4331E+06, 8.4973E+06, 8.5619E+06, 8.6270E+06,
     +  8.6925E+06, 8.7585E+06, 8.8250E+06, 8.8919E+06, 8.9593E+06,
     + 9.0271E+06, 9.0955E+06, 9.1643E+06, 9.2336E+06, 9.3033E+06,
     +  9.3736E+06, 9.4444E+06, 9.5156E+06, 9.5873E+06, 9.6596E+06,
     + 9.7323E+06, 9.8056E+06, 9.8793E+06, 9.9536E+06, 1.0028E+07,
     +  1.0104E+07, 1.0179E+07, 1.0256E+07, 1.0333E+07, 1.0410E+07,
     + 1.0488E+07, 1.0567E+07, 1.0646E+07, 1.0725E+07, 1.0805E+07,
     +  1.0886E+07, 1.0967E+07, 1.1049E+07, 1.1131E+07, 1.1214E+07,
     + 1.1298E+07, 1.1382E+07, 1.1466E+07, 1.1551E+07, 1.1637E+07,
     +  1.1724E+07, 1.1811E+07, 1.1898E+07, 1.1986E+07, 1.2075E+07,
     + 1.2164E+07, 1.2254E+07, 1.2345E+07, 1.2436E+07, 1.2528E+07,
     +  1.2620E+07, 1.2714E+07, 1.2807E+07, 1.2902E+07, 1.2997E+07,
     + 1.3092E+07, 1.3189E+07, 1.3285E+07, 1.3383E+07, 1.3481E+07,
     +  1.3580E+07, 1.3680E+07, 1.3780E+07, 1.3881E+07, 1.3983E+07,
     + 1.4085E+07, 1.4188E+07, 1.4292E+07, 1.4396E+07, 1.4501E+07,
     +  1.4607E+07, 1.4713E+07, 1.4821E+07, 1.4929E+07, 1.5037E+07,
     + 1.5147E+07, 1.5257E+07, 1.5368E+07, 1.5479E+07, 1.5591E+07,
     +  1.5705E+07, 1.5818E+07, 1.5933E+07, 1.6048E+07, 1.6165E+07,
     + 1.6281E+07, 1.6399E+07, 1.6518E+07, 1.6637E+07, 1.6757E+07,
     +  1.6878E+07, 1.7000E+07, 1.7122E+07, 1.7245E+07, 1.7369E+07,
     + 1.7494E+07, 1.7620E+07, 1.7747E+07, 1.7874E+07, 1.8002E+07,
     +  1.8131E+07, 1.8261E+07, 1.8392E+07, 1.8524E+07, 1.8656E+07,
     + 1.8790E+07, 1.8924E+07, 1.9059E+07, 1.9195E+07, 1.9332E+07,
     +  1.9470E+07, 1.9609E+07, 1.9749E+07, 1.9889E+07, 2.0031E+07,
     + 2.0173E+07, 2.0317E+07, 2.0461E+07, 2.0606E+07, 2.0752E+07,
     +  2.0899E+07, 2.1048E+07, 2.1197E+07, 2.1347E+07, 2.1498E+07,
     + 2.1650E+07, 2.1803E+07, 2.1957E+07, 2.2112E+07, 2.2268E+07,
     +  2.2425E+07, 2.2583E+07, 2.2742E+07, 2.2902E+07, 2.3063E+07,
     + 2.3225E+07, 2.3388E+07, 2.3553E+07, 2.3718E+07, 2.3884E+07,
     +  2.4052E+07, 2.4220E+07, 2.4390E+07, 2.4561E+07, 2.4732E+07,
     + 2.4905E+07, 2.5079E+07, 2.5255E+07, 2.5431E+07, 2.5608E+07,
     +  2.5787E+07, 2.5966E+07, 2.6147E+07, 2.6329E+07, 2.6512E+07,
     + 2.6697E+07, 2.6882E+07, 2.7069E+07, 2.7257E+07, 2.7446E+07,
     +  2.7636E+07, 2.7828E+07, 2.8021E+07, 2.8214E+07, 2.8410E+07,
     + 2.8606E+07, 2.8804E+07, 2.9003E+07, 2.9203E+07, 2.9404E+07,
     +  2.9607E+07, 2.9811E+07, 3.0016E+07, 3.0223E+07, 3.0431E+07,
     + 3.0640E+07, 3.0851E+07, 3.1063E+07, 3.1276E+07, 3.1490E+07,
     +  3.1706E+07, 3.1923E+07, 3.2142E+07, 3.2362E+07, 3.2584E+07,
     + 3.2806E+07, 3.3030E+07, 3.3256E+07, 3.3483E+07, 3.3712E+07,
     +  3.3941E+07, 3.4173E+07, 3.4405E+07, 3.4640E+07, 3.4875E+07,
     + 3.5112E+07, 3.5351E+07, 3.5591E+07, 3.5833E+07, 3.6076E+07,
     +  3.6320E+07, 3.6567E+07, 3.6814E+07, 3.7064E+07, 3.7314E+07,
     + 3.7567E+07, 3.7821E+07, 3.8076E+07, 3.8333E+07, 3.8592E+07,
     +  3.8852E+07, 3.9114E+07, 3.9377E+07, 3.9642E+07, 3.9909E+07,
     + 4.0177E+07, 4.0448E+07, 4.0719E+07, 4.0993E+07, 4.1268E+07,
     +  4.1544E+07, 4.1823E+07, 4.2103E+07, 4.2385E+07, 4.2669E+07,
     + 4.2954E+07, 4.3241E+07, 4.3530E+07, 4.3821E+07, 4.4113E+07,
     +  4.4407E+07, 4.4703E+07, 4.5001E+07, 4.5301E+07, 4.5602E+07,
     + 4.5906E+07, 4.6211E+07, 4.6518E+07, 4.6827E+07, 4.7138E+07,
     +  4.7450E+07, 4.7765E+07, 4.8081E+07, 4.8400E+07, 4.8720E+07,
     + 4.9043E+07, 4.9367E+07, 4.9693E+07, 5.0021E+07, 5.0351E+07,
     +  5.0684E+07, 5.1018E+07, 5.1354E+07, 5.1692E+07, 5.2033E+07,
     + 5.2375E+07, 5.2719E+07, 5.3066E+07, 5.3414E+07, 5.3765E+07,
     +  5.4118E+07, 5.4473E+07, 5.4830E+07, 5.5189E+07, 5.5550E+07,
     + 5.5914E+07, 5.6280E+07, 5.6648E+07, 5.7018E+07, 5.7390E+07,
     +  5.7764E+07, 5.8141E+07, 5.8520E+07, 5.8902E+07, 5.9285E+07,
     + 5.9671E+07, 6.0059E+07, 6.0450E+07, 6.0843E+07, 6.1238E+07,
     +  6.1635E+07, 6.2035E+07, 6.2437E+07, 6.2842E+07, 6.3249E+07,
     + 6.3658E+07, 6.4070E+07, 6.4485E+07, 6.4901E+07, 6.5321E+07,
     +  6.5742E+07, 6.6167E+07, 6.6593E+07, 6.7023E+07, 6.7454E+07,
     + 6.7889E+07, 6.8326E+07, 6.8765E+07, 6.9207E+07, 6.9652E+07,
     +  7.0099E+07, 7.0549E+07, 7.1002E+07, 7.1457E+07, 7.1915E+07,
     + 7.2375E+07, 7.2839E+07, 7.3305E+07, 7.3774E+07, 7.4245E+07,
     +  7.4719E+07, 7.5196E+07, 7.5676E+07, 7.6159E+07, 7.6644E+07,
     + 7.7132E+07, 7.7624E+07, 7.8118E+07, 7.8614E+07, 7.9114E+07,
     +  7.9617E+07, 8.0122E+07, 8.0631E+07, 8.1142E+07, 8.1657E+07,
     + 8.2174E+07, 8.2695E+07, 8.3218E+07, 8.3745E+07, 8.4274E+07,
     +  8.4807E+07, 8.5343E+07, 8.5881E+07, 8.6423E+07, 8.6968E+07,
     + 8.7517E+07, 8.8068E+07, 8.8622E+07, 8.9180E+07, 8.9741E+07,
     +  9.0305E+07, 9.0873E+07, 9.1443E+07, 9.2017E+07, 9.2595E+07,
     + 9.3175E+07, 9.3759E+07, 9.4346E+07, 9.4937E+07, 9.5531E+07,
     +  9.6129E+07, 9.6729E+07, 9.7334E+07, 9.7941E+07, 9.8553E+07,
     + 9.9167E+07, 9.9785E+07, 1.0041E+08, 1.0103E+08, 1.0166E+08,
     +  1.0229E+08, 1.0293E+08, 1.0357E+08, 1.0421E+08, 1.0486E+08,
     + 1.0551E+08, 1.0616E+08, 1.0682E+08, 1.0748E+08, 1.0815E+08,
     +  1.0882E+08, 1.0949E+08, 1.1017E+08, 1.1085E+08, 1.1153E+08,
     + 1.1222E+08, 1.1292E+08, 1.1361E+08, 1.1431E+08, 1.1502E+08,
     +  1.1572E+08, 1.1644E+08, 1.1715E+08, 1.1787E+08, 1.1860E+08,
     + 1.1932E+08, 1.2006E+08, 1.2079E+08, 1.2153E+08, 1.2228E+08,
     +  1.2303E+08, 1.2378E+08, 1.2454E+08, 1.2530E+08, 1.2606E+08,
     + 1.2683E+08, 1.2761E+08, 1.2839E+08, 1.2917E+08, 1.2996E+08,
     +  1.3075E+08, 1.3155E+08, 1.3235E+08, 1.3315E+08, 1.3396E+08,
     + 1.3478E+08, 1.3559E+08, 1.3642E+08, 1.3724E+08, 1.3808E+08,
     +  1.3891E+08, 1.3976E+08, 1.4060E+08, 1.4145E+08, 1.4231E+08,
     + 1.4317E+08, 1.4403E+08, 1.4490E+08, 1.4578E+08, 1.4666E+08,
     +  1.4754E+08, 1.4843E+08, 1.4933E+08, 1.5022E+08, 1.5113E+08,
     + 1.5204E+08, 1.5295E+08, 1.5387E+08, 1.5479E+08, 1.5572E+08,
     +  1.5666E+08, 1.5760E+08, 1.5854E+08, 1.5949E+08, 1.6045E+08,
     + 1.6141E+08, 1.6237E+08, 1.6334E+08, 1.6432E+08, 1.6530E+08,
     +  1.6629E+08, 1.6728E+08, 1.6828E+08, 1.6928E+08, 1.7029E+08,
     + 1.7130E+08, 1.7232E+08, 1.7334E+08, 1.7438E+08, 1.7541E+08,
     +  1.7645E+08, 1.7750E+08, 1.7855E+08, 1.7961E+08, 1.8068E+08,
     + 1.8175E+08, 1.8282E+08, 1.8390E+08, 1.8499E+08, 1.8609E+08,
     +  1.8718E+08, 1.8829E+08, 1.8940E+08, 1.9052E+08, 1.9164E+08,
     + 1.9277E+08, 1.9391E+08, 1.9505E+08, 1.9620E+08, 1.9735E+08,
     +  1.9851E+08, 1.9968E+08, 2.0085E+08, 2.0203E+08, 2.0321E+08,
     + 2.0441E+08, 2.0560E+08, 2.0681E+08, 2.0802E+08, 2.0924E+08,
     +  2.1046E+08, 2.1169E+08, 2.1293E+08, 2.1417E+08, 2.1542E+08,
     + 2.1668E+08, 2.1794E+08, 2.1921E+08, 2.2049E+08, 2.2177E+08,
     +  2.2306E+08, 2.2436E+08, 2.2567E+08, 2.2698E+08, 2.2829E+08,
     + 2.2962E+08, 2.3095E+08, 2.3229E+08, 2.3364E+08, 2.3499E+08,
     +  2.3635E+08, 2.3772E+08, 2.3910E+08, 2.4048E+08, 2.4187E+08,
     + 2.4326E+08, 2.4467E+08, 2.4608E+08, 2.4750E+08, 2.4893E+08,
     +  2.5036E+08, 2.5180E+08, 2.5325E+08, 2.5471E+08, 2.5617E+08,
     + 2.5765E+08, 2.5913E+08, 2.6062E+08, 2.6211E+08, 2.6361E+08,
     +  2.6513E+08, 2.6665E+08, 2.6817E+08, 2.6971E+08, 2.7125E+08,
     + 2.7280E+08, 2.7436E+08, 2.7593E+08, 2.7751E+08, 2.7909E+08,
     +  2.8068E+08, 2.8228E+08, 2.8389E+08, 2.8551E+08, 2.8713E+08,
     + 2.8877E+08, 2.9041E+08, 2.9206E+08, 2.9372E+08, 2.9539E+08,
     +  2.9707E+08, 2.9875E+08, 3.0045E+08, 3.0215E+08, 3.0386E+08,
     + 3.0558E+08, 3.0731E+08, 3.0905E+08, 3.1080E+08, 3.1255E+08,
     +  3.1432E+08, 3.1610E+08, 3.1788E+08, 3.1967E+08, 3.2147E+08,
     + 3.2329E+08, 3.2511E+08, 3.2694E+08, 3.2878E+08, 3.3063E+08,
     +  3.3249E+08, 3.3435E+08, 3.3623E+08, 3.3812E+08, 3.4002E+08,
     + 3.4192E+08, 3.4384E+08, 3.4577E+08, 3.4770E+08, 3.4965E+08,
     +  3.5160E+08, 3.5357E+08, 3.5555E+08, 3.5753E+08, 3.5953E+08,
     + 3.6154E+08, 3.6355E+08, 3.6558E+08, 3.6762E+08, 3.6966E+08,
     +  3.7172E+08, 3.7379E+08, 3.7587E+08, 3.7796E+08, 3.8006E+08,
     + 3.8217E+08, 3.8429E+08, 3.8642E+08, 3.8857E+08, 3.9072E+08,
     +  3.9288E+08, 3.9506E+08, 3.9725E+08, 3.9944E+08, 4.0165E+08,
     + 4.0387E+08, 4.0610E+08, 4.0835E+08, 4.1060E+08, 4.1287E+08,
     +  4.1514E+08, 4.1743E+08, 4.1973E+08, 4.2204E+08, 4.2436E+08,
     + 4.2670E+08, 4.2904E+08, 4.3140E+08, 4.3377E+08, 4.3615E+08,
     +  4.3855E+08, 4.4095E+08, 4.4337E+08, 4.4580E+08, 4.4824E+08,
     + 4.5069E+08, 4.5316E+08, 4.5564E+08, 4.5813E+08, 4.6063E+08,
     +  4.6315E+08, 4.6567E+08, 4.6822E+08, 4.7077E+08, 4.7333E+08,
     + 4.7591E+08, 4.7851E+08, 4.8111E+08, 4.8373E+08, 4.8636E+08,
     +  4.8900E+08, 4.9166E+08, 4.9433E+08, 4.9701E+08, 4.9971E+08,
     + 5.0242E+08, 5.0514E+08, 5.0788E+08, 5.1063E+08, 5.1339E+08,
     +  5.1617E+08, 5.1896E+08, 5.2176E+08, 5.2458E+08, 5.2741E+08,
     + 5.3026E+08, 5.3312E+08, 5.3599E+08, 5.3888E+08, 5.4178E+08,
     +  5.4470E+08, 5.4763E+08, 5.5058E+08, 5.5354E+08, 5.5651E+08,
     + 5.5950E+08, 5.6251E+08, 5.6552E+08, 5.6856E+08, 5.7161E+08,
     +  5.7467E+08, 5.7775E+08, 5.8084E+08, 5.8395E+08, 5.8707E+08,
     + 5.9021E+08, 5.9336E+08, 5.9653E+08, 5.9972E+08, 6.0292E+08,
     +  6.0613E+08, 6.0936E+08, 6.1261E+08, 6.1587E+08, 6.1915E+08,
     + 6.2245E+08, 6.2576E+08, 6.2908E+08, 6.3243E+08, 6.3579E+08,
     +  6.3916E+08, 6.4255E+08, 6.4596E+08, 6.4938E+08, 6.5283E+08,
     + 6.5628E+08, 6.5976E+08, 6.6325E+08, 6.6676E+08, 6.7028E+08,
     +  6.7382E+08, 6.7738E+08, 6.8096E+08, 6.8455E+08, 6.8816E+08,
     + 6.9179E+08, 6.9543E+08, 6.9910E+08, 7.0278E+08, 7.0647E+08,
     +  7.1019E+08, 7.1392E+08, 7.1767E+08, 7.2144E+08, 7.2523E+08,
     + 7.2904E+08, 7.3286E+08, 7.3670E+08, 7.4056E+08, 7.4444E+08,
     +  7.4834E+08, 7.5225E+08, 7.5619E+08, 7.6014E+08, 7.6411E+08,
     + 7.6811E+08, 7.7212E+08, 7.7615E+08, 7.8019E+08, 7.8426E+08,
     +  7.8835E+08, 7.9245E+08, 7.9658E+08, 8.0073E+08, 8.0489E+08,
     + 8.0908E+08, 8.1328E+08, 8.1750E+08, 8.2175E+08, 8.2601E+08,
     +  8.3030E+08, 8.3460E+08, 8.3893E+08, 8.4327E+08, 8.4764E+08,
     + 8.5203E+08, 8.5643E+08, 8.6086E+08, 8.6531E+08, 8.6978E+08,
     +  8.7427E+08, 8.7879E+08, 8.8332E+08, 8.8787E+08, 8.9245E+08,
     + 8.9705E+08, 9.0167E+08, 9.0631E+08, 9.1097E+08, 9.1566E+08,
     +  9.2036E+08, 9.2509E+08, 9.2984E+08, 9.3461E+08, 9.3941E+08,
     + 9.4423E+08, 9.4907E+08, 9.5393E+08, 9.5882E+08, 9.6372E+08,
     +  9.6866E+08, 9.7361E+08, 9.7859E+08, 9.8359E+08, 9.8861E+08,
     + 9.9366E+08, 9.9873E+08, 1.0038E+09, 1.0089E+09, 1.0141E+09,
     +  1.0192E+09, 1.0244E+09, 1.0297E+09, 1.0349E+09, 1.0401E+09,
     + 1.0454E+09, 1.0507E+09, 1.0561E+09, 1.0614E+09, 1.0668E+09,
     +  1.0722E+09, 1.0777E+09, 1.0831E+09, 1.0886E+09, 1.0941E+09,
     + 1.0997E+09, 1.1052E+09, 1.1108E+09, 1.1164E+09, 1.1221E+09,
     +  1.1277E+09, 1.1334E+09, 1.1391E+09, 1.1449E+09, 1.1506E+09,
     + 1.1564E+09, 1.1623E+09, 1.1681E+09, 1.1740E+09, 1.1799E+09,
     +  1.1858E+09, 1.1918E+09, 1.1978E+09, 1.2038E+09, 1.2098E+09,
     + 1.2159E+09, 1.2220E+09, 1.2281E+09, 1.2342E+09, 1.2404E+09,
     +  1.2466E+09, 1.2529E+09, 1.2591E+09, 1.2654E+09, 1.2717E+09,
     + 1.2781E+09, 1.2845E+09, 1.2909E+09, 1.2973E+09, 1.3038E+09,
     +  1.3103E+09, 1.3168E+09, 1.3233E+09, 1.3299E+09, 1.3365E+09,
     + 1.3432E+09, 1.3498E+09, 1.3565E+09, 1.3633E+09, 1.3700E+09,
     +  1.3768E+09, 1.3837E+09, 1.3905E+09, 1.3974E+09, 1.4043E+09,
     + 1.4113E+09, 1.4183E+09, 1.4253E+09, 1.4323E+09, 1.4394E+09,
     +  1.4465E+09, 1.4536E+09, 1.4608E+09, 1.4680E+09, 1.4752E+09,
     + 1.4825E+09, 1.4898E+09, 1.4971E+09, 1.5045E+09, 1.5119E+09,
     +  1.5193E+09, 1.5268E+09, 1.5343E+09, 1.5418E+09, 1.5494E+09,
     + 1.5570E+09, 1.5646E+09, 1.5723E+09, 1.5800E+09, 1.5877E+09,
     +  1.5955E+09, 1.6033E+09, 1.6111E+09, 1.6190E+09, 1.6269E+09,
     + 1.6349E+09, 1.6429E+09, 1.6509E+09, 1.6589E+09, 1.6670E+09,
     +  1.6751E+09, 1.6833E+09, 1.6915E+09, 1.6997E+09, 1.7080E+09,
     + 1.7163E+09, 1.7246E+09, 1.7330E+09, 1.7414E+09, 1.7499E+09,
     +  1.7584E+09, 1.7669E+09, 1.7755E+09, 1.7841E+09, 1.7927E+09,
     + 1.8014E+09, 1.8101E+09, 1.8189E+09, 1.8277E+09, 1.8365E+09,
     +  1.8454E+09, 1.8543E+09, 1.8632E+09, 1.8722E+09, 1.8812E+09,
     + 1.8903E+09, 1.8994E+09, 1.9086E+09, 1.9178E+09, 1.9270E+09,
     +  1.9362E+09, 1.9456E+09, 1.9549E+09, 1.9643E+09, 1.9737E+09,
     + 1.9832E+09, 1.9927E+09, 2.0023E+09, 2.0119E+09, 2.0215E+09,
     +  2.0312E+09, 2.0409E+09, 2.0507E+09, 2.0605E+09, 2.0703E+09,
     + 2.0802E+09, 2.0902E+09, 2.1002E+09, 2.1102E+09, 2.1203E+09,
     +  2.1304E+09, 2.1405E+09, 2.1507E+09, 2.1610E+09, 2.1712E+09,
     + 2.1816E+09, 2.1920E+09, 2.2024E+09, 2.2128E+09, 2.2234E+09,
     +  2.2339E+09, 2.2445E+09, 2.2552E+09, 2.2659E+09, 2.2766E+09,
     + 2.2874E+09, 2.2982E+09, 2.3091E+09, 2.3200E+09, 2.3310E+09,
     +  2.3420E+09, 2.3531E+09, 2.3642E+09, 2.3754E+09, 2.3866E+09,
     + 2.3979E+09, 2.4092E+09, 2.4205E+09, 2.4319E+09, 2.4434E+09,
     +  2.4549E+09, 2.4665E+09, 2.4781E+09, 2.4897E+09, 2.5014E+09,
     + 2.5132E+09, 2.5250E+09, 2.5368E+09, 2.5487E+09, 2.5607E+09,
     +  2.5727E+09, 2.5847E+09, 2.5969E+09, 2.6090E+09, 2.6212E+09,
     + 2.6335E+09, 2.6458E+09, 2.6582E+09, 2.6706E+09, 2.6831E+09,
     +  2.6956E+09, 2.7082E+09, 2.7208E+09, 2.7335E+09, 2.7462E+09,
     + 2.7590E+09, 2.7719E+09, 2.7848E+09, 2.7978E+09, 2.8108E+09,
     +  2.8238E+09, 2.8370E+09, 2.8501E+09, 2.8634E+09, 2.8767E+09,
     + 2.8900E+09, 2.9034E+09, 2.9169E+09, 2.9304E+09, 2.9440E+09,
     +  2.9576E+09, 2.9713E+09, 2.9850E+09, 2.9988E+09, 3.0127E+09,
     + 3.0266E+09, 3.0406E+09, 3.0546E+09, 3.0687E+09, 3.0829E+09,
     +  3.0971E+09, 3.1114E+09, 3.1257E+09, 3.1401E+09, 3.1546E+09,
     + 3.1691E+09, 3.1837E+09, 3.1983E+09, 3.2130E+09, 3.2278E+09,
     +  3.2426E+09, 3.2575E+09, 3.2724E+09, 3.2874E+09, 3.3025E+09,
     + 3.3176E+09, 3.3328E+09, 3.3481E+09, 3.3634E+09, 3.3788E+09,
     +  3.3942E+09, 3.4098E+09, 3.4253E+09, 3.4410E+09, 3.4567E+09,
     + 3.4725E+09, 3.4883E+09, 3.5042E+09, 3.5202E+09, 3.5362E+09,
     +  3.5523E+09, 3.5685E+09, 3.5847E+09, 3.6010E+09, 3.6174E+09,
     + 3.6339E+09, 3.6504E+09, 3.6669E+09, 3.6836E+09, 3.7003E+09,
     +  3.7171E+09, 3.7339E+09, 3.7509E+09, 3.7679E+09, 3.7849E+09,
     + 3.8021E+09, 3.8193E+09, 3.8365E+09, 3.8539E+09, 3.8713E+09,
     +  3.8888E+09, 3.9063E+09, 3.9240E+09, 3.9417E+09, 3.9594E+09,
     + 3.9773E+09, 3.9952E+09, 4.0132E+09, 4.0313E+09, 4.0494E+09,
     +  4.0676E+09, 4.0859E+09, 4.1043E+09, 4.1228E+09, 4.1413E+09,
     + 4.1599E+09, 4.1785E+09, 4.1973E+09, 4.2161E+09, 4.2350E+09,
     +  4.2540E+09, 4.2730E+09, 4.2922E+09, 4.3114E+09, 4.3307E+09,
     + 4.3500E+09, 4.3695E+09, 4.3890E+09, 4.4086E+09, 4.4283E+09,
     +  4.4480E+09, 4.4679E+09, 4.4878E+09, 4.5078E+09, 4.5279E+09,
     + 4.5480E+09, 4.5683E+09, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       211
      data (QofT( 2,J),J=1,2408)/ 3.0146E+01, 7.0531E+01, 1.1643E+02,
     +  1.6668E+02, 2.2086E+02, 2.7871E+02, 3.3997E+02, 4.0446E+02,
     + 4.7200E+02, 5.4250E+02, 6.1588E+02, 6.9208E+02, 7.7108E+02,
     +  8.5285E+02, 9.3736E+02, 1.0246E+03, 1.1146E+03, 1.2072E+03,
     + 1.3025E+03, 1.4004E+03, 1.6039E+03, 1.8176E+03, 2.0409E+03,
     +  2.2737E+03, 2.5156E+03, 2.7662E+03, 3.0253E+03, 3.2925E+03,
     + 3.5677E+03, 3.8505E+03, 4.1407E+03, 4.4381E+03, 4.7426E+03,
     +  5.0538E+03, 5.3718E+03, 5.6963E+03, 6.0271E+03, 6.3642E+03,
     + 6.7074E+03, 7.0566E+03, 7.4118E+03, 7.7727E+03, 8.1393E+03,
     +  8.5116E+03, 8.8894E+03, 9.2726E+03, 9.6612E+03, 1.0055E+04,
     + 1.0454E+04, 1.0859E+04, 1.1268E+04, 1.1683E+04, 1.2102E+04,
     +  1.2526E+04, 1.2956E+04, 1.3390E+04, 1.3829E+04, 1.4272E+04,
     + 1.4721E+04, 1.5174E+04, 1.5632E+04, 1.6094E+04, 1.6561E+04,
     +  1.7032E+04, 1.7508E+04, 1.7989E+04, 1.8474E+04, 1.8963E+04,
     + 1.9457E+04, 1.9955E+04, 2.0458E+04, 2.0965E+04, 2.1476E+04,
     +  2.1992E+04, 2.2512E+04, 2.3037E+04, 2.3566E+04, 2.4099E+04,
     + 2.4637E+04, 2.5179E+04, 2.5725E+04, 2.6276E+04, 2.6831E+04,
     +  2.7391E+04, 2.7955E+04, 2.8523E+04, 2.9096E+04, 2.9673E+04,
     + 3.0255E+04, 3.0841E+04, 3.1432E+04, 3.2027E+04, 3.2627E+04,
     +  3.3231E+04, 3.3840E+04, 3.4454E+04, 3.5072E+04, 3.5695E+04,
     + 3.6323E+04, 3.6955E+04, 3.7592E+04, 3.8234E+04, 3.8881E+04,
     +  3.9533E+04, 4.0190E+04, 4.0851E+04, 4.1518E+04, 4.2190E+04,
     + 4.2866E+04, 4.3548E+04, 4.4235E+04, 4.4928E+04, 4.5625E+04,
     +  4.6328E+04, 4.7036E+04, 4.7750E+04, 4.8469E+04, 4.9193E+04,
     + 4.9923E+04, 5.0659E+04, 5.1401E+04, 5.2148E+04, 5.2901E+04,
     +  5.3659E+04, 5.4424E+04, 5.5195E+04, 5.5971E+04, 5.6754E+04,
     + 5.7543E+04, 5.8338E+04, 5.9139E+04, 5.9947E+04, 6.0761E+04,
     +  6.1581E+04, 6.2408E+04, 6.3242E+04, 6.4082E+04, 6.4929E+04,
     + 6.5783E+04, 6.6644E+04, 6.7511E+04, 6.8386E+04, 6.9268E+04,
     +  7.0156E+04, 7.1053E+04, 7.1956E+04, 7.2867E+04, 7.3785E+04,
     + 7.4711E+04, 7.5644E+04, 7.6585E+04, 7.7534E+04, 7.8491E+04,
     +  7.9455E+04, 8.0428E+04, 8.1409E+04, 8.2398E+04, 8.3395E+04,
     + 8.4401E+04, 8.5415E+04, 8.6438E+04, 8.7469E+04, 8.8509E+04,
     +  8.9558E+04, 9.0615E+04, 9.1682E+04, 9.2757E+04, 9.3842E+04,
     + 9.4937E+04, 9.6040E+04, 9.7153E+04, 9.8275E+04, 9.9408E+04,
     +  1.0055E+05, 1.0170E+05, 1.0286E+05, 1.0403E+05, 1.0522E+05,
     + 1.0641E+05, 1.0761E+05, 1.0882E+05, 1.1005E+05, 1.1128E+05,
     +  1.1253E+05, 1.1378E+05, 1.1505E+05, 1.1633E+05, 1.1762E+05,
     + 1.1892E+05, 1.2023E+05, 1.2155E+05, 1.2289E+05, 1.2423E+05,
     +  1.2559E+05, 1.2696E+05, 1.2834E+05, 1.2974E+05, 1.3115E+05,
     + 1.3256E+05, 1.3400E+05, 1.3544E+05, 1.3690E+05, 1.3837E+05,
     +  1.3985E+05, 1.4135E+05, 1.4286E+05, 1.4438E+05, 1.4592E+05,
     + 1.4747E+05, 1.4904E+05, 1.5062E+05, 1.5221E+05, 1.5381E+05,
     +  1.5544E+05, 1.5707E+05, 1.5872E+05, 1.6039E+05, 1.6207E+05,
     + 1.6376E+05, 1.6547E+05, 1.6720E+05, 1.6894E+05, 1.7069E+05,
     +  1.7247E+05, 1.7425E+05, 1.7606E+05, 1.7788E+05, 1.7972E+05,
     + 1.8157E+05, 1.8344E+05, 1.8532E+05, 1.8723E+05, 1.8915E+05,
     +  1.9109E+05, 1.9304E+05, 1.9501E+05, 1.9700E+05, 1.9901E+05,
     + 2.0104E+05, 2.0308E+05, 2.0514E+05, 2.0722E+05, 2.0932E+05,
     +  2.1144E+05, 2.1358E+05, 2.1574E+05, 2.1791E+05, 2.2011E+05,
     + 2.2232E+05, 2.2456E+05, 2.2681E+05, 2.2909E+05, 2.3138E+05,
     +  2.3370E+05, 2.3604E+05, 2.3840E+05, 2.4077E+05, 2.4318E+05,
     + 2.4560E+05, 2.4804E+05, 2.5051E+05, 2.5299E+05, 2.5550E+05,
     +  2.5803E+05, 2.6059E+05, 2.6317E+05, 2.6577E+05, 2.6839E+05,
     + 2.7104E+05, 2.7371E+05, 2.7640E+05, 2.7912E+05, 2.8186E+05,
     +  2.8463E+05, 2.8742E+05, 2.9024E+05, 2.9308E+05, 2.9595E+05,
     + 2.9884E+05, 3.0176E+05, 3.0470E+05, 3.0767E+05, 3.1067E+05,
     +  3.1369E+05, 3.1674E+05, 3.1982E+05, 3.2292E+05, 3.2606E+05,
     + 3.2922E+05, 3.3240E+05, 3.3562E+05, 3.3886E+05, 3.4214E+05,
     +  3.4544E+05, 3.4877E+05, 3.5213E+05, 3.5552E+05, 3.5894E+05,
     + 3.6239E+05, 3.6587E+05, 3.6938E+05, 3.7292E+05, 3.7650E+05,
     +  3.8010E+05, 3.8374E+05, 3.8741E+05, 3.9111E+05, 3.9484E+05,
     + 3.9861E+05, 4.0241E+05, 4.0624E+05, 4.1010E+05, 4.1400E+05,
     +  4.1794E+05, 4.2191E+05, 4.2591E+05, 4.2995E+05, 4.3402E+05,
     + 4.3813E+05, 4.4227E+05, 4.4645E+05, 4.5067E+05, 4.5492E+05,
     +  4.5921E+05, 4.6354E+05, 4.6791E+05, 4.7231E+05, 4.7675E+05,
     + 4.8124E+05, 4.8575E+05, 4.9031E+05, 4.9491E+05, 4.9955E+05,
     +  5.0423E+05, 5.0895E+05, 5.1371E+05, 5.1851E+05, 5.2335E+05,
     + 5.2823E+05, 5.3316E+05, 5.3813E+05, 5.4314E+05, 5.4819E+05,
     +  5.5329E+05, 5.5843E+05, 5.6362E+05, 5.6885E+05, 5.7413E+05,
     + 5.7945E+05, 5.8481E+05, 5.9023E+05, 5.9569E+05, 6.0119E+05,
     +  6.0674E+05, 6.1234E+05, 6.1799E+05, 6.2369E+05, 6.2943E+05,
     + 6.3523E+05, 6.4107E+05, 6.4697E+05, 6.5291E+05, 6.5890E+05,
     +  6.6495E+05, 6.7104E+05, 6.7719E+05, 6.8339E+05, 6.8965E+05,
     + 6.9595E+05, 7.0231E+05, 7.0873E+05, 7.1519E+05, 7.2172E+05,
     +  7.2830E+05, 7.3493E+05, 7.4162E+05, 7.4836E+05, 7.5517E+05,
     + 7.6203E+05, 7.6895E+05, 7.7592E+05, 7.8296E+05, 7.9005E+05,
     +  7.9720E+05, 8.0442E+05, 8.1169E+05, 8.1903E+05, 8.2643E+05,
     + 8.3388E+05, 8.4141E+05, 8.4899E+05, 8.5664E+05, 8.6435E+05,
     +  8.7213E+05, 8.7997E+05, 8.8787E+05, 8.9585E+05, 9.0389E+05,
     + 9.1199E+05, 9.2016E+05, 9.2841E+05, 9.3672E+05, 9.4510E+05,
     +  9.5354E+05, 9.6206E+05, 9.7065E+05, 9.7931E+05, 9.8804E+05,
     + 9.9685E+05, 1.0057E+06, 1.0147E+06, 1.0237E+06, 1.0328E+06,
     +  1.0420E+06, 1.0512E+06, 1.0606E+06, 1.0700E+06, 1.0794E+06,
     + 1.0890E+06, 1.0986E+06, 1.1083E+06, 1.1181E+06, 1.1280E+06,
     +  1.1380E+06, 1.1480E+06, 1.1581E+06, 1.1683E+06, 1.1786E+06,
     + 1.1890E+06, 1.1995E+06, 1.2100E+06, 1.2206E+06, 1.2313E+06,
     +  1.2421E+06, 1.2530E+06, 1.2640E+06, 1.2751E+06, 1.2862E+06,
     + 1.2975E+06, 1.3088E+06, 1.3202E+06, 1.3318E+06, 1.3434E+06,
     +  1.3551E+06, 1.3669E+06, 1.3788E+06, 1.3908E+06, 1.4029E+06,
     + 1.4151E+06, 1.4274E+06, 1.4398E+06, 1.4523E+06, 1.4649E+06,
     +  1.4775E+06, 1.4903E+06, 1.5032E+06, 1.5162E+06, 1.5293E+06,
     + 1.5425E+06, 1.5559E+06, 1.5693E+06, 1.5828E+06, 1.5965E+06,
     +  1.6102E+06, 1.6241E+06, 1.6380E+06, 1.6521E+06, 1.6663E+06,
     + 1.6806E+06, 1.6950E+06, 1.7095E+06, 1.7242E+06, 1.7390E+06,
     +  1.7538E+06, 1.7688E+06, 1.7840E+06, 1.7992E+06, 1.8146E+06,
     + 1.8300E+06, 1.8456E+06, 1.8614E+06, 1.8772E+06, 1.8932E+06,
     +  1.9093E+06, 1.9255E+06, 1.9419E+06, 1.9584E+06, 1.9750E+06,
     + 1.9918E+06, 2.0086E+06, 2.0256E+06, 2.0428E+06, 2.0601E+06,
     +  2.0775E+06, 2.0950E+06, 2.1127E+06, 2.1306E+06, 2.1485E+06,
     + 2.1666E+06, 2.1849E+06, 2.2033E+06, 2.2218E+06, 2.2405E+06,
     +  2.2593E+06, 2.2783E+06, 2.2974E+06, 2.3167E+06, 2.3361E+06,
     + 2.3557E+06, 2.3754E+06, 2.3953E+06, 2.4153E+06, 2.4355E+06,
     +  2.4559E+06, 2.4764E+06, 2.4970E+06, 2.5178E+06, 2.5388E+06,
     + 2.5600E+06, 2.5813E+06, 2.6027E+06, 2.6244E+06, 2.6462E+06,
     +  2.6681E+06, 2.6903E+06, 2.7126E+06, 2.7351E+06, 2.7577E+06,
     + 2.7806E+06, 2.8036E+06, 2.8267E+06, 2.8501E+06, 2.8736E+06,
     +  2.8973E+06, 2.9212E+06, 2.9453E+06, 2.9696E+06, 2.9940E+06,
     + 3.0187E+06, 3.0435E+06, 3.0685E+06, 3.0937E+06, 3.1191E+06,
     +  3.1447E+06, 3.1705E+06, 3.1964E+06, 3.2226E+06, 3.2490E+06,
     + 3.2756E+06, 3.3023E+06, 3.3293E+06, 3.3565E+06, 3.3839E+06,
     +  3.4115E+06, 3.4393E+06, 3.4673E+06, 3.4955E+06, 3.5239E+06,
     + 3.5526E+06, 3.5815E+06, 3.6105E+06, 3.6398E+06, 3.6694E+06,
     +  3.6991E+06, 3.7291E+06, 3.7593E+06, 3.7897E+06, 3.8203E+06,
     + 3.8512E+06, 3.8823E+06, 3.9136E+06, 3.9452E+06, 3.9770E+06,
     +  4.0091E+06, 4.0414E+06, 4.0739E+06, 4.1066E+06, 4.1397E+06,
     + 4.1729E+06, 4.2064E+06, 4.2402E+06, 4.2742E+06, 4.3084E+06,
     +  4.3429E+06, 4.3777E+06, 4.4127E+06, 4.4480E+06, 4.4836E+06,
     + 4.5194E+06, 4.5554E+06, 4.5918E+06, 4.6284E+06, 4.6652E+06,
     +  4.7024E+06, 4.7398E+06, 4.7775E+06, 4.8155E+06, 4.8537E+06,
     + 4.8923E+06, 4.9311E+06, 4.9702E+06, 5.0096E+06, 5.0493E+06,
     +  5.0892E+06, 5.1295E+06, 5.1700E+06, 5.2109E+06, 5.2520E+06,
     + 5.2935E+06, 5.3352E+06, 5.3773E+06, 5.4197E+06, 5.4623E+06,
     +  5.5053E+06, 5.5486E+06, 5.5922E+06, 5.6362E+06, 5.6804E+06,
     + 5.7250E+06, 5.7699E+06, 5.8151E+06, 5.8606E+06, 5.9065E+06,
     +  5.9527E+06, 5.9992E+06, 6.0461E+06, 6.0933E+06, 6.1409E+06,
     + 6.1888E+06, 6.2370E+06, 6.2856E+06, 6.3345E+06, 6.3838E+06,
     +  6.4335E+06, 6.4835E+06, 6.5338E+06, 6.5846E+06, 6.6357E+06,
     + 6.6871E+06, 6.7389E+06, 6.7911E+06, 6.8437E+06, 6.8966E+06,
     +  6.9500E+06, 7.0037E+06, 7.0577E+06, 7.1122E+06, 7.1671E+06,
     + 7.2223E+06, 7.2780E+06, 7.3340E+06, 7.3904E+06, 7.4473E+06,
     +  7.5045E+06, 7.5622E+06, 7.6202E+06, 7.6787E+06, 7.7376E+06,
     + 7.7969E+06, 7.8566E+06, 7.9167E+06, 7.9773E+06, 8.0383E+06,
     +  8.0997E+06, 8.1616E+06, 8.2239E+06, 8.2866E+06, 8.3498E+06,
     + 8.4134E+06, 8.4775E+06, 8.5420E+06, 8.6069E+06, 8.6724E+06,
     +  8.7383E+06, 8.8046E+06, 8.8714E+06, 8.9387E+06, 9.0064E+06,
     + 9.0747E+06, 9.1434E+06, 9.2125E+06, 9.2822E+06, 9.3524E+06,
     +  9.4230E+06, 9.4941E+06, 9.5658E+06, 9.6379E+06, 9.7105E+06,
     + 9.7836E+06, 9.8573E+06, 9.9314E+06, 1.0006E+07, 1.0081E+07,
     +  1.0157E+07, 1.0233E+07, 1.0310E+07, 1.0387E+07, 1.0465E+07,
     + 1.0543E+07, 1.0622E+07, 1.0702E+07, 1.0782E+07, 1.0862E+07,
     +  1.0943E+07, 1.1025E+07, 1.1107E+07, 1.1190E+07, 1.1273E+07,
     + 1.1357E+07, 1.1442E+07, 1.1527E+07, 1.1613E+07, 1.1699E+07,
     +  1.1786E+07, 1.1873E+07, 1.1961E+07, 1.2050E+07, 1.2139E+07,
     + 1.2229E+07, 1.2319E+07, 1.2410E+07, 1.2502E+07, 1.2594E+07,
     +  1.2687E+07, 1.2781E+07, 1.2875E+07, 1.2970E+07, 1.3066E+07,
     + 1.3162E+07, 1.3258E+07, 1.3356E+07, 1.3454E+07, 1.3553E+07,
     +  1.3652E+07, 1.3752E+07, 1.3853E+07, 1.3955E+07, 1.4057E+07,
     + 1.4160E+07, 1.4263E+07, 1.4368E+07, 1.4472E+07, 1.4578E+07,
     +  1.4684E+07, 1.4792E+07, 1.4899E+07, 1.5008E+07, 1.5117E+07,
     + 1.5227E+07, 1.5338E+07, 1.5449E+07, 1.5561E+07, 1.5674E+07,
     +  1.5788E+07, 1.5903E+07, 1.6018E+07, 1.6134E+07, 1.6251E+07,
     + 1.6368E+07, 1.6487E+07, 1.6606E+07, 1.6726E+07, 1.6846E+07,
     +  1.6968E+07, 1.7090E+07, 1.7213E+07, 1.7337E+07, 1.7462E+07,
     + 1.7588E+07, 1.7714E+07, 1.7841E+07, 1.7969E+07, 1.8098E+07,
     +  1.8228E+07, 1.8359E+07, 1.8490E+07, 1.8623E+07, 1.8756E+07,
     + 1.8890E+07, 1.9025E+07, 1.9161E+07, 1.9298E+07, 1.9436E+07,
     +  1.9574E+07, 1.9714E+07, 1.9854E+07, 1.9995E+07, 2.0138E+07,
     + 2.0281E+07, 2.0425E+07, 2.0570E+07, 2.0716E+07, 2.0863E+07,
     +  2.1011E+07, 2.1160E+07, 2.1310E+07, 2.1461E+07, 2.1613E+07,
     + 2.1766E+07, 2.1919E+07, 2.2074E+07, 2.2230E+07, 2.2387E+07,
     +  2.2545E+07, 2.2704E+07, 2.2864E+07, 2.3024E+07, 2.3186E+07,
     + 2.3350E+07, 2.3514E+07, 2.3679E+07, 2.3845E+07, 2.4012E+07,
     +  2.4181E+07, 2.4350E+07, 2.4521E+07, 2.4692E+07, 2.4865E+07,
     + 2.5039E+07, 2.5214E+07, 2.5390E+07, 2.5567E+07, 2.5746E+07,
     +  2.5925E+07, 2.6106E+07, 2.6288E+07, 2.6471E+07, 2.6655E+07,
     + 2.6840E+07, 2.7027E+07, 2.7214E+07, 2.7403E+07, 2.7593E+07,
     +  2.7785E+07, 2.7977E+07, 2.8171E+07, 2.8366E+07, 2.8562E+07,
     + 2.8760E+07, 2.8959E+07, 2.9159E+07, 2.9360E+07, 2.9562E+07,
     +  2.9766E+07, 2.9971E+07, 3.0178E+07, 3.0385E+07, 3.0595E+07,
     + 3.0805E+07, 3.1017E+07, 3.1230E+07, 3.1444E+07, 3.1660E+07,
     +  3.1877E+07, 3.2095E+07, 3.2315E+07, 3.2536E+07, 3.2759E+07,
     + 3.2983E+07, 3.3208E+07, 3.3435E+07, 3.3663E+07, 3.3893E+07,
     +  3.4124E+07, 3.4357E+07, 3.4591E+07, 3.4826E+07, 3.5063E+07,
     + 3.5302E+07, 3.5542E+07, 3.5783E+07, 3.6026E+07, 3.6270E+07,
     +  3.6516E+07, 3.6764E+07, 3.7013E+07, 3.7263E+07, 3.7516E+07,
     + 3.7769E+07, 3.8025E+07, 3.8281E+07, 3.8540E+07, 3.8800E+07,
     +  3.9062E+07, 3.9325E+07, 3.9590E+07, 3.9856E+07, 4.0125E+07,
     + 4.0394E+07, 4.0666E+07, 4.0939E+07, 4.1214E+07, 4.1491E+07,
     +  4.1769E+07, 4.2049E+07, 4.2331E+07, 4.2614E+07, 4.2899E+07,
     + 4.3186E+07, 4.3475E+07, 4.3765E+07, 4.4058E+07, 4.4352E+07,
     +  4.4647E+07, 4.4945E+07, 4.5245E+07, 4.5546E+07, 4.5849E+07,
     + 4.6154E+07, 4.6461E+07, 4.6770E+07, 4.7080E+07, 4.7393E+07,
     +  4.7707E+07, 4.8023E+07, 4.8342E+07, 4.8662E+07, 4.8984E+07,
     + 4.9308E+07, 4.9634E+07, 4.9962E+07, 5.0292E+07, 5.0624E+07,
     +  5.0958E+07, 5.1294E+07, 5.1632E+07, 5.1972E+07, 5.2315E+07,
     + 5.2659E+07, 5.3005E+07, 5.3353E+07, 5.3704E+07, 5.4057E+07,
     +  5.4411E+07, 5.4768E+07, 5.5127E+07, 5.5488E+07, 5.5852E+07,
     + 5.6217E+07, 5.6585E+07, 5.6955E+07, 5.7327E+07, 5.7701E+07,
     +  5.8078E+07, 5.8457E+07, 5.8838E+07, 5.9221E+07, 5.9607E+07,
     + 5.9995E+07, 6.0385E+07, 6.0778E+07, 6.1173E+07, 6.1570E+07,
     +  6.1970E+07, 6.2372E+07, 6.2776E+07, 6.3183E+07, 6.3593E+07,
     + 6.4004E+07, 6.4418E+07, 6.4835E+07, 6.5254E+07, 6.5676E+07,
     +  6.6100E+07, 6.6526E+07, 6.6955E+07, 6.7387E+07, 6.7821E+07,
     + 6.8258E+07, 6.8697E+07, 6.9139E+07, 6.9584E+07, 7.0031E+07,
     +  7.0481E+07, 7.0933E+07, 7.1388E+07, 7.1846E+07, 7.2306E+07,
     + 7.2769E+07, 7.3235E+07, 7.3704E+07, 7.4175E+07, 7.4649E+07,
     +  7.5126E+07, 7.5606E+07, 7.6088E+07, 7.6573E+07, 7.7062E+07,
     + 7.7553E+07, 7.8046E+07, 7.8543E+07, 7.9043E+07, 7.9545E+07,
     +  8.0051E+07, 8.0559E+07, 8.1070E+07, 8.1585E+07, 8.2102E+07,
     + 8.2622E+07, 8.3146E+07, 8.3672E+07, 8.4201E+07, 8.4734E+07,
     +  8.5269E+07, 8.5808E+07, 8.6350E+07, 8.6895E+07, 8.7443E+07,
     + 8.7994E+07, 8.8548E+07, 8.9106E+07, 8.9667E+07, 9.0231E+07,
     +  9.0798E+07, 9.1369E+07, 9.1942E+07, 9.2520E+07, 9.3100E+07,
     + 9.3684E+07, 9.4271E+07, 9.4861E+07, 9.5455E+07, 9.6053E+07,
     +  9.6653E+07, 9.7258E+07, 9.7865E+07, 9.8476E+07, 9.9091E+07,
     + 9.9709E+07, 1.0033E+08, 1.0096E+08, 1.0158E+08, 1.0222E+08,
     +  1.0285E+08, 1.0349E+08, 1.0414E+08, 1.0478E+08, 1.0543E+08,
     + 1.0609E+08, 1.0674E+08, 1.0741E+08, 1.0807E+08, 1.0874E+08,
     +  1.0941E+08, 1.1009E+08, 1.1077E+08, 1.1146E+08, 1.1214E+08,
     + 1.1284E+08, 1.1353E+08, 1.1423E+08, 1.1494E+08, 1.1564E+08,
     +  1.1636E+08, 1.1707E+08, 1.1779E+08, 1.1852E+08, 1.1924E+08,
     + 1.1998E+08, 1.2071E+08, 1.2145E+08, 1.2220E+08, 1.2295E+08,
     +  1.2370E+08, 1.2446E+08, 1.2522E+08, 1.2598E+08, 1.2675E+08,
     + 1.2753E+08, 1.2831E+08, 1.2909E+08, 1.2988E+08, 1.3067E+08,
     +  1.3147E+08, 1.3227E+08, 1.3307E+08, 1.3388E+08, 1.3470E+08,
     + 1.3551E+08, 1.3634E+08, 1.3716E+08, 1.3800E+08, 1.3883E+08,
     +  1.3968E+08, 1.4052E+08, 1.4137E+08, 1.4223E+08, 1.4309E+08,
     + 1.4395E+08, 1.4482E+08, 1.4570E+08, 1.4658E+08, 1.4746E+08,
     +  1.4835E+08, 1.4925E+08, 1.5014E+08, 1.5105E+08, 1.5196E+08,
     + 1.5287E+08, 1.5379E+08, 1.5471E+08, 1.5564E+08, 1.5658E+08,
     +  1.5752E+08, 1.5846E+08, 1.5941E+08, 1.6037E+08, 1.6133E+08,
     + 1.6229E+08, 1.6326E+08, 1.6424E+08, 1.6522E+08, 1.6621E+08,
     +  1.6720E+08, 1.6820E+08, 1.6920E+08, 1.7021E+08, 1.7122E+08,
     + 1.7224E+08, 1.7327E+08, 1.7430E+08, 1.7533E+08, 1.7638E+08,
     +  1.7742E+08, 1.7848E+08, 1.7953E+08, 1.8060E+08, 1.8167E+08,
     + 1.8275E+08, 1.8383E+08, 1.8492E+08, 1.8601E+08, 1.8711E+08,
     +  1.8821E+08, 1.8933E+08, 1.9044E+08, 1.9157E+08, 1.9270E+08,
     + 1.9383E+08, 1.9497E+08, 1.9612E+08, 1.9728E+08, 1.9844E+08,
     +  1.9960E+08, 2.0078E+08, 2.0196E+08, 2.0314E+08, 2.0433E+08,
     + 2.0553E+08, 2.0674E+08, 2.0795E+08, 2.0916E+08, 2.1039E+08,
     +  2.1162E+08, 2.1286E+08, 2.1410E+08, 2.1535E+08, 2.1661E+08,
     + 2.1787E+08, 2.1914E+08, 2.2042E+08, 2.2170E+08, 2.2299E+08,
     +  2.2429E+08, 2.2560E+08, 2.2691E+08, 2.2823E+08, 2.2955E+08,
     + 2.3089E+08, 2.3222E+08, 2.3357E+08, 2.3493E+08, 2.3629E+08,
     +  2.3766E+08, 2.3903E+08, 2.4041E+08, 2.4180E+08, 2.4320E+08,
     + 2.4461E+08, 2.4602E+08, 2.4744E+08, 2.4887E+08, 2.5030E+08,
     +  2.5174E+08, 2.5319E+08, 2.5465E+08, 2.5612E+08, 2.5759E+08,
     + 2.5907E+08, 2.6056E+08, 2.6205E+08, 2.6356E+08, 2.6507E+08,
     +  2.6659E+08, 2.6812E+08, 2.6965E+08, 2.7120E+08, 2.7275E+08,
     + 2.7431E+08, 2.7588E+08, 2.7745E+08, 2.7904E+08, 2.8063E+08,
     +  2.8223E+08, 2.8384E+08, 2.8546E+08, 2.8708E+08, 2.8872E+08,
     + 2.9036E+08, 2.9201E+08, 2.9367E+08, 2.9534E+08, 2.9702E+08,
     +  2.9871E+08, 3.0040E+08, 3.0211E+08, 3.0382E+08, 3.0554E+08,
     + 3.0727E+08, 3.0901E+08, 3.1076E+08, 3.1252E+08, 3.1428E+08,
     +  3.1606E+08, 3.1784E+08, 3.1964E+08, 3.2144E+08, 3.2325E+08,
     + 3.2507E+08, 3.2690E+08, 3.2875E+08, 3.3060E+08, 3.3245E+08,
     +  3.3432E+08, 3.3620E+08, 3.3809E+08, 3.3999E+08, 3.4190E+08,
     + 3.4381E+08, 3.4574E+08, 3.4768E+08, 3.4963E+08, 3.5158E+08,
     +  3.5355E+08, 3.5553E+08, 3.5751E+08, 3.5951E+08, 3.6152E+08,
     + 3.6354E+08, 3.6556E+08, 3.6760E+08, 3.6965E+08, 3.7171E+08,
     +  3.7378E+08, 3.7586E+08, 3.7795E+08, 3.8005E+08, 3.8216E+08,
     + 3.8429E+08, 3.8642E+08, 3.8856E+08, 3.9072E+08, 3.9288E+08,
     +  3.9506E+08, 3.9725E+08, 3.9945E+08, 4.0166E+08, 4.0388E+08,
     + 4.0611E+08, 4.0835E+08, 4.1061E+08, 4.1288E+08, 4.1515E+08,
     +  4.1744E+08, 4.1974E+08, 4.2206E+08, 4.2438E+08, 4.2671E+08,
     + 4.2906E+08, 4.3142E+08, 4.3379E+08, 4.3617E+08, 4.3857E+08,
     +  4.4098E+08, 4.4339E+08, 4.4583E+08, 4.4827E+08, 4.5072E+08,
     + 4.5319E+08, 4.5567E+08, 4.5816E+08, 4.6067E+08, 4.6319E+08,
     +  4.6571E+08, 4.6826E+08, 4.7081E+08, 4.7338E+08, 4.7596E+08,
     + 4.7855E+08, 4.8116E+08, 4.8378E+08, 4.8641E+08, 4.8906E+08,
     +  4.9171E+08, 4.9439E+08, 4.9707E+08, 4.9977E+08, 5.0248E+08,
     + 5.0520E+08, 5.0794E+08, 5.1069E+08, 5.1346E+08, 5.1624E+08,
     +  5.1903E+08, 5.2184E+08, 5.2466E+08, 5.2749E+08, 5.3034E+08,
     + 5.3320E+08, 5.3608E+08, 5.3897E+08, 5.4187E+08, 5.4479E+08,
     +  5.4773E+08, 5.5067E+08, 5.5364E+08, 5.5661E+08, 5.5960E+08,
     + 5.6261E+08, 5.6563E+08, 5.6867E+08, 5.7172E+08, 5.7478E+08,
     +  5.7786E+08, 5.8096E+08, 5.8407E+08, 5.8719E+08, 5.9033E+08,
     + 5.9349E+08, 5.9666E+08, 5.9985E+08, 6.0305E+08, 6.0627E+08,
     +  6.0950E+08, 6.1275E+08, 6.1602E+08, 6.1930E+08, 6.2259E+08,
     + 6.2591E+08, 6.2924E+08, 6.3258E+08, 6.3594E+08, 6.3932E+08,
     +  6.4271E+08, 6.4612E+08, 6.4955E+08, 6.5300E+08, 6.5646E+08,
     + 6.5993E+08, 6.6343E+08, 6.6694E+08, 6.7046E+08, 6.7401E+08,
     +  6.7757E+08, 6.8115E+08, 6.8474E+08, 6.8836E+08, 6.9199E+08,
     + 6.9564E+08, 6.9930E+08, 7.0299E+08, 7.0669E+08, 7.1041E+08,
     +  7.1414E+08, 7.1790E+08, 7.2167E+08, 7.2546E+08, 7.2927E+08,
     + 7.3309E+08, 7.3694E+08, 7.4080E+08, 7.4469E+08, 7.4859E+08,
     +  7.5250E+08, 7.5644E+08, 7.6040E+08, 7.6437E+08, 7.6837E+08,
     + 7.7238E+08, 7.7642E+08, 7.8047E+08, 7.8454E+08, 7.8863E+08,
     +  7.9274E+08, 7.9687E+08, 8.0102E+08, 8.0518E+08, 8.0937E+08,
     + 8.1358E+08, 8.1781E+08, 8.2206E+08, 8.2633E+08, 8.3061E+08,
     +  8.3492E+08, 8.3925E+08, 8.4360E+08, 8.4797E+08, 8.5236E+08,
     + 8.5677E+08, 8.6121E+08, 8.6566E+08, 8.7013E+08, 8.7463E+08,
     +  8.7915E+08, 8.8368E+08, 8.8824E+08, 8.9282E+08, 8.9742E+08,
     + 9.0205E+08, 9.0669E+08, 9.1136E+08, 9.1605E+08, 9.2076E+08,
     +  9.2549E+08, 9.3025E+08, 9.3503E+08, 9.3983E+08, 9.4465E+08,
     + 9.4949E+08, 9.5436E+08, 9.5925E+08, 9.6416E+08, 9.6910E+08,
     +  9.7406E+08, 9.7904E+08, 9.8404E+08, 9.8907E+08, 9.9413E+08,
     + 9.9920E+08, 1.0043E+09, 1.0094E+09, 1.0146E+09, 1.0197E+09,
     +  1.0249E+09, 1.0302E+09, 1.0354E+09, 1.0407E+09, 1.0460E+09,
     + 1.0513E+09, 1.0566E+09, 1.0620E+09, 1.0674E+09, 1.0728E+09,
     +  1.0782E+09, 1.0837E+09, 1.0892E+09, 1.0947E+09, 1.1002E+09,
     + 1.1058E+09, 1.1114E+09, 1.1170E+09, 1.1227E+09, 1.1283E+09,
     +  1.1340E+09, 1.1397E+09, 1.1455E+09, 1.1513E+09, 1.1571E+09,
     + 1.1629E+09, 1.1688E+09, 1.1746E+09, 1.1805E+09, 1.1865E+09,
     +  1.1924E+09, 1.1984E+09, 1.2044E+09, 1.2105E+09, 1.2166E+09,
     + 1.2227E+09, 1.2288E+09, 1.2350E+09, 1.2411E+09, 1.2473E+09,
     +  1.2536E+09, 1.2599E+09, 1.2662E+09, 1.2725E+09, 1.2788E+09,
     + 1.2852E+09, 1.2916E+09, 1.2981E+09, 1.3046E+09, 1.3111E+09,
     +  1.3176E+09, 1.3241E+09, 1.3307E+09, 1.3374E+09, 1.3440E+09,
     + 1.3507E+09, 1.3574E+09, 1.3641E+09, 1.3709E+09, 1.3777E+09,
     +  1.3845E+09, 1.3914E+09, 1.3983E+09, 1.4052E+09, 1.4122E+09,
     + 1.4192E+09, 1.4262E+09, 1.4332E+09, 1.4403E+09, 1.4474E+09,
     +  1.4546E+09, 1.4618E+09, 1.4690E+09, 1.4762E+09, 1.4835E+09,
     + 1.4908E+09, 1.4981E+09, 1.5055E+09, 1.5129E+09, 1.5204E+09,
     +  1.5278E+09, 1.5353E+09, 1.5429E+09, 1.5505E+09, 1.5581E+09,
     + 1.5657E+09, 1.5734E+09, 1.5811E+09, 1.5888E+09, 1.5966E+09,
     +  1.6044E+09, 1.6123E+09, 1.6202E+09, 1.6281E+09, 1.6360E+09,
     + 1.6440E+09, 1.6521E+09, 1.6601E+09, 1.6682E+09, 1.6763E+09,
     +  1.6845E+09, 1.6927E+09, 1.7010E+09, 1.7092E+09, 1.7176E+09,
     + 1.7259E+09, 1.7343E+09, 1.7427E+09, 1.7512E+09, 1.7597E+09,
     +  1.7682E+09, 1.7768E+09, 1.7854E+09, 1.7941E+09, 1.8028E+09,
     + 1.8115E+09, 1.8202E+09, 1.8291E+09, 1.8379E+09, 1.8468E+09,
     +  1.8557E+09, 1.8647E+09, 1.8737E+09, 1.8827E+09, 1.8918E+09,
     + 1.9009E+09, 1.9101E+09, 1.9193E+09, 1.9285E+09, 1.9378E+09,
     +  1.9471E+09, 1.9565E+09, 1.9659E+09, 1.9753E+09, 1.9848E+09,
     + 1.9943E+09, 2.0039E+09, 2.0135E+09, 2.0232E+09, 2.0328E+09,
     +  2.0426E+09, 2.0524E+09, 2.0622E+09, 2.0720E+09, 2.0820E+09,
     + 2.0919E+09, 2.1019E+09, 2.1119E+09, 2.1220E+09, 2.1321E+09,
     +  2.1423E+09, 2.1525E+09, 2.1628E+09, 2.1731E+09, 2.1834E+09,
     + 2.1938E+09, 2.2043E+09, 2.2147E+09, 2.2253E+09, 2.2358E+09,
     +  2.2464E+09, 2.2571E+09, 2.2678E+09, 2.2786E+09, 2.2894E+09,
     + 2.3002E+09, 2.3111E+09, 2.3221E+09, 2.3331E+09, 2.3441E+09,
     +  2.3552E+09, 2.3663E+09, 2.3775E+09, 2.3887E+09, 2.4000E+09,
     + 2.4113E+09, 2.4227E+09, 2.4341E+09, 2.4456E+09, 2.4571E+09,
     +  2.4687E+09, 2.4803E+09, 2.4920E+09, 2.5037E+09, 2.5155E+09,
     + 2.5273E+09, 2.5391E+09, 2.5511E+09, 2.5630E+09, 2.5751E+09,
     +  2.5871E+09, 2.5993E+09, 2.6114E+09, 2.6237E+09, 2.6359E+09,
     + 2.6483E+09, 2.6607E+09, 2.6731E+09, 2.6856E+09, 2.6981E+09,
     +  2.7107E+09, 2.7234E+09, 2.7361E+09, 2.7489E+09, 2.7617E+09,
     + 2.7745E+09, 2.7875E+09, 2.8004E+09, 2.8135E+09, 2.8266E+09,
     +  2.8397E+09, 2.8529E+09, 2.8662E+09, 2.8795E+09, 2.8928E+09,
     + 2.9063E+09, 2.9197E+09, 2.9333E+09, 2.9469E+09, 2.9605E+09,
     +  2.9742E+09, 2.9880E+09, 3.0018E+09, 3.0157E+09, 3.0296E+09,
     + 3.0436E+09, 3.0577E+09, 3.0718E+09, 3.0860E+09, 3.1002E+09,
     +  3.1145E+09, 3.1289E+09, 3.1433E+09, 3.1578E+09, 3.1723E+09,
     + 3.1869E+09, 3.2016E+09, 3.2163E+09, 3.2311E+09, 3.2459E+09,
     +  3.2608E+09, 3.2758E+09, 3.2908E+09, 3.3059E+09, 3.3211E+09,
     + 3.3363E+09, 3.3516E+09, 3.3669E+09, 3.3823E+09, 3.3978E+09,
     +  3.4133E+09, 3.4289E+09, 3.4446E+09, 3.4603E+09, 3.4761E+09,
     + 3.4920E+09, 3.5079E+09, 3.5239E+09, 3.5400E+09, 3.5561E+09,
     +  3.5723E+09, 3.5886E+09, 3.6049E+09, 3.6213E+09, 3.6378E+09,
     + 3.6543E+09, 3.6709E+09, 3.6876E+09, 3.7043E+09, 3.7211E+09,
     +  3.7380E+09, 3.7550E+09, 3.7720E+09, 3.7891E+09, 3.8062E+09,
     + 3.8235E+09, 3.8408E+09, 3.8581E+09, 3.8756E+09, 3.8931E+09,
     +  3.9107E+09, 3.9283E+09, 3.9461E+09, 3.9639E+09, 3.9817E+09,
     + 3.9997E+09, 4.0177E+09, 4.0358E+09, 4.0540E+09, 4.0722E+09,
     +  4.0905E+09, 4.1089E+09, 4.1274E+09, 4.1460E+09, 4.1646E+09,
     + 4.1833E+09, 4.2021E+09, 4.2209E+09, 4.2398E+09, 4.2588E+09,
     +  4.2779E+09, 4.2971E+09, 4.3163E+09, 4.3356E+09, 4.3550E+09,
     + 4.3745E+09, 4.3941E+09, 4.4137E+09, 4.4334E+09, 4.4532E+09,
     +  4.4731E+09, 4.4930E+09, 4.5131E+09, 4.5332E+09, 4.5534E+09,
     + 4.5737E+09, 4.5940E+09, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_CH3CN (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 4),QofT( 4,NT),Q(NT)
      data (Tmx(I),I=1,4)/3500.,3500.,3500.,3500./ 
      data xgj/ 3.,6.,6.,12./
c...    CH3CN
c...        --      2124
      data (QofT( 1,J),J=1,2408)/ 2.3151E+01, 5.4367E+01, 8.9856E+01,
     +  1.2869E+02, 1.7055E+02, 2.1522E+02, 2.6253E+02, 3.1230E+02,
     + 3.6443E+02, 4.1883E+02, 4.7543E+02, 5.3420E+02, 5.9510E+02,
     +  6.5812E+02, 7.2324E+02, 7.9045E+02, 8.5973E+02, 9.3106E+02,
     + 1.0044E+03, 1.0798E+03, 1.2365E+03, 1.4009E+03, 1.5728E+03,
     +  1.7520E+03, 1.9381E+03, 2.1310E+03, 2.3305E+03, 2.5362E+03,
     + 2.7480E+03, 2.9657E+03, 3.1891E+03, 3.4181E+03, 3.6526E+03,
     +  3.8923E+03, 4.1372E+03, 4.3872E+03, 4.6421E+03, 4.9020E+03,
     + 5.1666E+03, 5.4361E+03, 5.7102E+03, 5.9890E+03, 6.2725E+03,
     +  6.5605E+03, 6.8532E+03, 7.1504E+03, 7.4523E+03, 7.7587E+03,
     + 8.0698E+03, 8.3855E+03, 8.7059E+03, 9.0309E+03, 9.3608E+03,
     +  9.6954E+03, 1.0035E+04, 1.0379E+04, 1.0728E+04, 1.1083E+04,
     + 1.1442E+04, 1.1806E+04, 1.2176E+04, 1.2551E+04, 1.2931E+04,
     +  1.3317E+04, 1.3708E+04, 1.4105E+04, 1.4507E+04, 1.4916E+04,
     + 1.5330E+04, 1.5750E+04, 1.6176E+04, 1.6608E+04, 1.7046E+04,
     +  1.7490E+04, 1.7941E+04, 1.8398E+04, 1.8862E+04, 1.9333E+04,
     + 1.9810E+04, 2.0294E+04, 2.0785E+04, 2.1283E+04, 2.1788E+04,
     +  2.2300E+04, 2.2820E+04, 2.3347E+04, 2.3881E+04, 2.4424E+04,
     + 2.4973E+04, 2.5531E+04, 2.6097E+04, 2.6670E+04, 2.7252E+04,
     +  2.7842E+04, 2.8441E+04, 2.9048E+04, 2.9663E+04, 3.0287E+04,
     + 3.0920E+04, 3.1562E+04, 3.2212E+04, 3.2872E+04, 3.3542E+04,
     +  3.4220E+04, 3.4908E+04, 3.5606E+04, 3.6313E+04, 3.7030E+04,
     + 3.7757E+04, 3.8495E+04, 3.9242E+04, 4.0000E+04, 4.0768E+04,
     +  4.1547E+04, 4.2337E+04, 4.3138E+04, 4.3949E+04, 4.4772E+04,
     + 4.5606E+04, 4.6452E+04, 4.7309E+04, 4.8178E+04, 4.9058E+04,
     +  4.9951E+04, 5.0856E+04, 5.1773E+04, 5.2703E+04, 5.3646E+04,
     + 5.4601E+04, 5.5569E+04, 5.6550E+04, 5.7545E+04, 5.8553E+04,
     +  5.9575E+04, 6.0611E+04, 6.1660E+04, 6.2724E+04, 6.3802E+04,
     + 6.4894E+04, 6.6002E+04, 6.7124E+04, 6.8261E+04, 6.9414E+04,
     +  7.0582E+04, 7.1765E+04, 7.2965E+04, 7.4181E+04, 7.5412E+04,
     + 7.6661E+04, 7.7926E+04, 7.9208E+04, 8.0507E+04, 8.1823E+04,
     +  8.3157E+04, 8.4508E+04, 8.5878E+04, 8.7266E+04, 8.8672E+04,
     + 9.0097E+04, 9.1541E+04, 9.3004E+04, 9.4486E+04, 9.5989E+04,
     +  9.7511E+04, 9.9053E+04, 1.0062E+05, 1.0220E+05, 1.0380E+05,
     + 1.0543E+05, 1.0707E+05, 1.0874E+05, 1.1043E+05, 1.1215E+05,
     +  1.1388E+05, 1.1564E+05, 1.1742E+05, 1.1923E+05, 1.2105E+05,
     + 1.2291E+05, 1.2478E+05, 1.2668E+05, 1.2861E+05, 1.3056E+05,
     +  1.3254E+05, 1.3454E+05, 1.3657E+05, 1.3863E+05, 1.4071E+05,
     + 1.4282E+05, 1.4496E+05, 1.4712E+05, 1.4932E+05, 1.5154E+05,
     +  1.5379E+05, 1.5607E+05, 1.5838E+05, 1.6072E+05, 1.6309E+05,
     + 1.6549E+05, 1.6792E+05, 1.7039E+05, 1.7288E+05, 1.7541E+05,
     +  1.7797E+05, 1.8057E+05, 1.8320E+05, 1.8586E+05, 1.8855E+05,
     + 1.9129E+05, 1.9405E+05, 1.9685E+05, 1.9969E+05, 2.0257E+05,
     +  2.0548E+05, 2.0843E+05, 2.1142E+05, 2.1444E+05, 2.1751E+05,
     + 2.2061E+05, 2.2376E+05, 2.2694E+05, 2.3016E+05, 2.3343E+05,
     +  2.3674E+05, 2.4009E+05, 2.4348E+05, 2.4692E+05, 2.5040E+05,
     + 2.5393E+05, 2.5750E+05, 2.6111E+05, 2.6478E+05, 2.6848E+05,
     +  2.7224E+05, 2.7604E+05, 2.7990E+05, 2.8380E+05, 2.8775E+05,
     + 2.9175E+05, 2.9580E+05, 2.9990E+05, 3.0406E+05, 3.0827E+05,
     +  3.1253E+05, 3.1684E+05, 3.2121E+05, 3.2564E+05, 3.3012E+05,
     + 3.3465E+05, 3.3925E+05, 3.4390E+05, 3.4861E+05, 3.5338E+05,
     +  3.5821E+05, 3.6310E+05, 3.6805E+05, 3.7307E+05, 3.7814E+05,
     + 3.8328E+05, 3.8849E+05, 3.9376E+05, 3.9910E+05, 4.0450E+05,
     +  4.0997E+05, 4.1551E+05, 4.2112E+05, 4.2679E+05, 4.3254E+05,
     + 4.3836E+05, 4.4426E+05, 4.5022E+05, 4.5626E+05, 4.6238E+05,
     +  4.6857E+05, 4.7484E+05, 4.8118E+05, 4.8761E+05, 4.9411E+05,
     + 5.0069E+05, 5.0736E+05, 5.1411E+05, 5.2094E+05, 5.2785E+05,
     +  5.3486E+05, 5.4194E+05, 5.4912E+05, 5.5638E+05, 5.6373E+05,
     + 5.7117E+05, 5.7871E+05, 5.8633E+05, 5.9405E+05, 6.0187E+05,
     +  6.0978E+05, 6.1779E+05, 6.2589E+05, 6.3409E+05, 6.4240E+05,
     + 6.5080E+05, 6.5931E+05, 6.6793E+05, 6.7664E+05, 6.8546E+05,
     +  6.9440E+05, 7.0343E+05, 7.1258E+05, 7.2184E+05, 7.3121E+05,
     + 7.4070E+05, 7.5030E+05, 7.6002E+05, 7.6985E+05, 7.7981E+05,
     +  7.8988E+05, 8.0008E+05, 8.1039E+05, 8.2084E+05, 8.3141E+05,
     + 8.4210E+05, 8.5293E+05, 8.6388E+05, 8.7497E+05, 8.8619E+05,
     +  8.9754E+05, 9.0903E+05, 9.2066E+05, 9.3243E+05, 9.4434E+05,
     + 9.5639E+05, 9.6858E+05, 9.8092E+05, 9.9341E+05, 1.0060E+06,
     +  1.0188E+06, 1.0318E+06, 1.0449E+06, 1.0581E+06, 1.0715E+06,
     + 1.0851E+06, 1.0988E+06, 1.1127E+06, 1.1268E+06, 1.1410E+06,
     +  1.1554E+06, 1.1699E+06, 1.1847E+06, 1.1996E+06, 1.2147E+06,
     + 1.2299E+06, 1.2454E+06, 1.2610E+06, 1.2768E+06, 1.2928E+06,
     +  1.3090E+06, 1.3253E+06, 1.3419E+06, 1.3586E+06, 1.3756E+06,
     + 1.3927E+06, 1.4101E+06, 1.4276E+06, 1.4454E+06, 1.4634E+06,
     +  1.4816E+06, 1.4999E+06, 1.5185E+06, 1.5374E+06, 1.5564E+06,
     + 1.5757E+06, 1.5952E+06, 1.6149E+06, 1.6348E+06, 1.6550E+06,
     +  1.6754E+06, 1.6960E+06, 1.7169E+06, 1.7380E+06, 1.7594E+06,
     + 1.7810E+06, 1.8029E+06, 1.8250E+06, 1.8474E+06, 1.8700E+06,
     +  1.8929E+06, 1.9160E+06, 1.9395E+06, 1.9632E+06, 1.9871E+06,
     + 2.0114E+06, 2.0359E+06, 2.0607E+06, 2.0857E+06, 2.1111E+06,
     +  2.1368E+06, 2.1627E+06, 2.1890E+06, 2.2155E+06, 2.2424E+06,
     + 2.2695E+06, 2.2970E+06, 2.3248E+06, 2.3528E+06, 2.3813E+06,
     +  2.4100E+06, 2.4391E+06, 2.4684E+06, 2.4982E+06, 2.5282E+06,
     + 2.5586E+06, 2.5894E+06, 2.6205E+06, 2.6519E+06, 2.6837E+06,
     +  2.7159E+06, 2.7484E+06, 2.7813E+06, 2.8145E+06, 2.8482E+06,
     + 2.8822E+06, 2.9166E+06, 2.9513E+06, 2.9865E+06, 3.0221E+06,
     +  3.0580E+06, 3.0944E+06, 3.1312E+06, 3.1684E+06, 3.2060E+06,
     + 3.2440E+06, 3.2824E+06, 3.3213E+06, 3.3606E+06, 3.4003E+06,
     +  3.4405E+06, 3.4811E+06, 3.5222E+06, 3.5638E+06, 3.6058E+06,
     + 3.6482E+06, 3.6912E+06, 3.7346E+06, 3.7785E+06, 3.8229E+06,
     +  3.8677E+06, 3.9131E+06, 3.9590E+06, 4.0053E+06, 4.0522E+06,
     + 4.0996E+06, 4.1475E+06, 4.1960E+06, 4.2450E+06, 4.2945E+06,
     +  4.3446E+06, 4.3952E+06, 4.4463E+06, 4.4981E+06, 4.5504E+06,
     + 4.6032E+06, 4.6567E+06, 4.7107E+06, 4.7653E+06, 4.8205E+06,
     +  4.8764E+06, 4.9328E+06, 4.9898E+06, 5.0475E+06, 5.1058E+06,
     + 5.1647E+06, 5.2243E+06, 5.2845E+06, 5.3454E+06, 5.4069E+06,
     +  5.4691E+06, 5.5319E+06, 5.5955E+06, 5.6597E+06, 5.7246E+06,
     + 5.7903E+06, 5.8566E+06, 5.9236E+06, 5.9914E+06, 6.0599E+06,
     +  6.1292E+06, 6.1991E+06, 6.2699E+06, 6.3414E+06, 6.4136E+06,
     + 6.4867E+06, 6.5605E+06, 6.6351E+06, 6.7105E+06, 6.7867E+06,
     +  6.8638E+06, 6.9416E+06, 7.0203E+06, 7.0998E+06, 7.1802E+06,
     + 7.2614E+06, 7.3435E+06, 7.4265E+06, 7.5103E+06, 7.5951E+06,
     +  7.6807E+06, 7.7672E+06, 7.8547E+06, 7.9431E+06, 8.0324E+06,
     + 8.1227E+06, 8.2139E+06, 8.3061E+06, 8.3993E+06, 8.4934E+06,
     +  8.5886E+06, 8.6847E+06, 8.7819E+06, 8.8801E+06, 8.9793E+06,
     + 9.0795E+06, 9.1808E+06, 9.2832E+06, 9.3867E+06, 9.4912E+06,
     +  9.5968E+06, 9.7036E+06, 9.8114E+06, 9.9204E+06, 1.0031E+07,
     + 1.0142E+07, 1.0254E+07, 1.0368E+07, 1.0483E+07, 1.0599E+07,
     +  1.0716E+07, 1.0834E+07, 1.0954E+07, 1.1075E+07, 1.1197E+07,
     + 1.1320E+07, 1.1445E+07, 1.1571E+07, 1.1698E+07, 1.1827E+07,
     +  1.1957E+07, 1.2088E+07, 1.2221E+07, 1.2354E+07, 1.2490E+07,
     + 1.2627E+07, 1.2765E+07, 1.2904E+07, 1.3045E+07, 1.3188E+07,
     +  1.3332E+07, 1.3477E+07, 1.3624E+07, 1.3773E+07, 1.3923E+07,
     + 1.4074E+07, 1.4227E+07, 1.4382E+07, 1.4538E+07, 1.4696E+07,
     +  1.4855E+07, 1.5016E+07, 1.5179E+07, 1.5343E+07, 1.5509E+07,
     + 1.5676E+07, 1.5846E+07, 1.6017E+07, 1.6190E+07, 1.6364E+07,
     +  1.6541E+07, 1.6719E+07, 1.6899E+07, 1.7080E+07, 1.7264E+07,
     + 1.7449E+07, 1.7637E+07, 1.7826E+07, 1.8017E+07, 1.8210E+07,
     +  1.8405E+07, 1.8602E+07, 1.8801E+07, 1.9002E+07, 1.9205E+07,
     + 1.9410E+07, 1.9617E+07, 1.9826E+07, 2.0037E+07, 2.0250E+07,
     +  2.0466E+07, 2.0683E+07, 2.0903E+07, 2.1125E+07, 2.1349E+07,
     + 2.1576E+07, 2.1804E+07, 2.2035E+07, 2.2268E+07, 2.2504E+07,
     +  2.2742E+07, 2.2982E+07, 2.3225E+07, 2.3470E+07, 2.3717E+07,
     + 2.3967E+07, 2.4220E+07, 2.4475E+07, 2.4732E+07, 2.4992E+07,
     +  2.5255E+07, 2.5520E+07, 2.5788E+07, 2.6058E+07, 2.6331E+07,
     + 2.6607E+07, 2.6885E+07, 2.7166E+07, 2.7450E+07, 2.7737E+07,
     +  2.8027E+07, 2.8319E+07, 2.8614E+07, 2.8912E+07, 2.9213E+07,
     + 2.9517E+07, 2.9824E+07, 3.0134E+07, 3.0447E+07, 3.0763E+07,
     +  3.1083E+07, 3.1405E+07, 3.1730E+07, 3.2059E+07, 3.2390E+07,
     + 3.2725E+07, 3.3064E+07, 3.3405E+07, 3.3750E+07, 3.4098E+07,
     +  3.4450E+07, 3.4805E+07, 3.5163E+07, 3.5525E+07, 3.5890E+07,
     + 3.6259E+07, 3.6631E+07, 3.7007E+07, 3.7387E+07, 3.7770E+07,
     +  3.8157E+07, 3.8548E+07, 3.8942E+07, 3.9340E+07, 3.9742E+07,
     + 4.0148E+07, 4.0558E+07, 4.0972E+07, 4.1389E+07, 4.1811E+07,
     +  4.2237E+07, 4.2667E+07, 4.3100E+07, 4.3538E+07, 4.3981E+07,
     + 4.4427E+07, 4.4878E+07, 4.5333E+07, 4.5792E+07, 4.6255E+07,
     +  4.6724E+07, 4.7196E+07, 4.7673E+07, 4.8154E+07, 4.8640E+07,
     + 4.9131E+07, 4.9626E+07, 5.0126E+07, 5.0631E+07, 5.1141E+07,
     +  5.1655E+07, 5.2174E+07, 5.2698E+07, 5.3227E+07, 5.3761E+07,
     + 5.4300E+07, 5.4844E+07, 5.5393E+07, 5.5947E+07, 5.6507E+07,
     +  5.7071E+07, 5.7642E+07, 5.8217E+07, 5.8798E+07, 5.9384E+07,
     + 5.9976E+07, 6.0573E+07, 6.1176E+07, 6.1784E+07, 6.2398E+07,
     +  6.3018E+07, 6.3643E+07, 6.4275E+07, 6.4912E+07, 6.5555E+07,
     + 6.6204E+07, 6.6860E+07, 6.7521E+07, 6.8188E+07, 6.8862E+07,
     +  6.9542E+07, 7.0228E+07, 7.0920E+07, 7.1619E+07, 7.2325E+07,
     + 7.3036E+07, 7.3755E+07, 7.4480E+07, 7.5212E+07, 7.5950E+07,
     +  7.6695E+07, 7.7447E+07, 7.8206E+07, 7.8972E+07, 7.9745E+07,
     + 8.0525E+07, 8.1313E+07, 8.2107E+07, 8.2909E+07, 8.3718E+07,
     +  8.4535E+07, 8.5358E+07, 8.6190E+07, 8.7029E+07, 8.7876E+07,
     + 8.8730E+07, 8.9592E+07, 9.0462E+07, 9.1340E+07, 9.2226E+07,
     +  9.3120E+07, 9.4023E+07, 9.4933E+07, 9.5851E+07, 9.6778E+07,
     + 9.7714E+07, 9.8658E+07, 9.9610E+07, 1.0057E+08, 1.0154E+08,
     +  1.0252E+08, 1.0351E+08, 1.0450E+08, 1.0551E+08, 1.0652E+08,
     + 1.0754E+08, 1.0858E+08, 1.0962E+08, 1.1067E+08, 1.1173E+08,
     +  1.1280E+08, 1.1388E+08, 1.1497E+08, 1.1607E+08, 1.1718E+08,
     + 1.1830E+08, 1.1943E+08, 1.2056E+08, 1.2171E+08, 1.2287E+08,
     +  1.2404E+08, 1.2522E+08, 1.2641E+08, 1.2762E+08, 1.2883E+08,
     + 1.3005E+08, 1.3128E+08, 1.3253E+08, 1.3378E+08, 1.3505E+08,
     +  1.3633E+08, 1.3762E+08, 1.3892E+08, 1.4023E+08, 1.4155E+08,
     + 1.4289E+08, 1.4424E+08, 1.4560E+08, 1.4697E+08, 1.4835E+08,
     +  1.4975E+08, 1.5116E+08, 1.5258E+08, 1.5401E+08, 1.5545E+08,
     + 1.5691E+08, 1.5838E+08, 1.5987E+08, 1.6136E+08, 1.6287E+08,
     +  1.6440E+08, 1.6593E+08, 1.6748E+08, 1.6904E+08, 1.7062E+08,
     + 1.7221E+08, 1.7382E+08, 1.7544E+08, 1.7707E+08, 1.7872E+08,
     +  1.8038E+08, 1.8205E+08, 1.8374E+08, 1.8545E+08, 1.8717E+08,
     + 1.8890E+08, 1.9065E+08, 1.9242E+08, 1.9420E+08, 1.9599E+08,
     +  1.9780E+08, 1.9963E+08, 2.0147E+08, 2.0333E+08, 2.0521E+08,
     + 2.0710E+08, 2.0901E+08, 2.1093E+08, 2.1287E+08, 2.1483E+08,
     +  2.1680E+08, 2.1879E+08, 2.2080E+08, 2.2282E+08, 2.2486E+08,
     + 2.2692E+08, 2.2900E+08, 2.3110E+08, 2.3321E+08, 2.3534E+08,
     +  2.3749E+08, 2.3966E+08, 2.4184E+08, 2.4405E+08, 2.4627E+08,
     + 2.4852E+08, 2.5078E+08, 2.5306E+08, 2.5536E+08, 2.5768E+08,
     +  2.6002E+08, 2.6238E+08, 2.6476E+08, 2.6716E+08, 2.6958E+08,
     + 2.7202E+08, 2.7448E+08, 2.7696E+08, 2.7946E+08, 2.8199E+08,
     +  2.8453E+08, 2.8710E+08, 2.8969E+08, 2.9230E+08, 2.9493E+08,
     + 2.9759E+08, 3.0026E+08, 3.0296E+08, 3.0569E+08, 3.0843E+08,
     +  3.1120E+08, 3.1399E+08, 3.1681E+08, 3.1964E+08, 3.2251E+08,
     + 3.2539E+08, 3.2830E+08, 3.3124E+08, 3.3420E+08, 3.3718E+08,
     +  3.4019E+08, 3.4322E+08, 3.4628E+08, 3.4937E+08, 3.5248E+08,
     + 3.5561E+08, 3.5877E+08, 3.6196E+08, 3.6518E+08, 3.6842E+08,
     +  3.7169E+08, 3.7498E+08, 3.7831E+08, 3.8166E+08, 3.8503E+08,
     + 3.8844E+08, 3.9188E+08, 3.9534E+08, 3.9883E+08, 4.0235E+08,
     +  4.0590E+08, 4.0948E+08, 4.1308E+08, 4.1672E+08, 4.2039E+08,
     + 4.2409E+08, 4.2781E+08, 4.3157E+08, 4.3536E+08, 4.3918E+08,
     +  4.4304E+08, 4.4692E+08, 4.5083E+08, 4.5478E+08, 4.5876E+08,
     + 4.6277E+08, 4.6682E+08, 4.7090E+08, 4.7501E+08, 4.7915E+08,
     +  4.8333E+08, 4.8754E+08, 4.9179E+08, 4.9607E+08, 5.0039E+08,
     + 5.0474E+08, 5.0913E+08, 5.1355E+08, 5.1801E+08, 5.2250E+08,
     +  5.2703E+08, 5.3160E+08, 5.3620E+08, 5.4084E+08, 5.4552E+08,
     + 5.5024E+08, 5.5500E+08, 5.5979E+08, 5.6462E+08, 5.6949E+08,
     +  5.7440E+08, 5.7935E+08, 5.8434E+08, 5.8937E+08, 5.9444E+08,
     + 5.9955E+08, 6.0470E+08, 6.0990E+08, 6.1513E+08, 6.2041E+08,
     +  6.2573E+08, 6.3109E+08, 6.3650E+08, 6.4194E+08, 6.4744E+08,
     + 6.5297E+08, 6.5855E+08, 6.6417E+08, 6.6984E+08, 6.7556E+08,
     +  6.8132E+08, 6.8712E+08, 6.9297E+08, 6.9887E+08, 7.0482E+08,
     + 7.1081E+08, 7.1685E+08, 7.2294E+08, 7.2907E+08, 7.3526E+08,
     +  7.4149E+08, 7.4777E+08, 7.5411E+08, 7.6049E+08, 7.6692E+08,
     + 7.7341E+08, 7.7994E+08, 7.8653E+08, 7.9317E+08, 7.9986E+08,
     +  8.0660E+08, 8.1340E+08, 8.2025E+08, 8.2715E+08, 8.3411E+08,
     + 8.4112E+08, 8.4819E+08, 8.5531E+08, 8.6249E+08, 8.6973E+08,
     +  8.7702E+08, 8.8437E+08, 8.9178E+08, 8.9924E+08, 9.0676E+08,
     + 9.1434E+08, 9.2199E+08, 9.2969E+08, 9.3745E+08, 9.4527E+08,
     +  9.5315E+08, 9.6109E+08, 9.6910E+08, 9.7716E+08, 9.8529E+08,
     + 9.9348E+08, 1.0017E+09, 1.0101E+09, 1.0184E+09, 1.0269E+09,
     +  1.0354E+09, 1.0440E+09, 1.0526E+09, 1.0614E+09, 1.0701E+09,
     + 1.0790E+09, 1.0879E+09, 1.0969E+09, 1.1059E+09, 1.1151E+09,
     +  1.1243E+09, 1.1335E+09, 1.1429E+09, 1.1523E+09, 1.1618E+09,
     + 1.1713E+09, 1.1809E+09, 1.1906E+09, 1.2004E+09, 1.2103E+09,
     +  1.2202E+09, 1.2302E+09, 1.2403E+09, 1.2504E+09, 1.2607E+09,
     + 1.2710E+09, 1.2814E+09, 1.2918E+09, 1.3024E+09, 1.3130E+09,
     +  1.3237E+09, 1.3345E+09, 1.3454E+09, 1.3563E+09, 1.3674E+09,
     + 1.3785E+09, 1.3897E+09, 1.4010E+09, 1.4124E+09, 1.4238E+09,
     +  1.4354E+09, 1.4470E+09, 1.4588E+09, 1.4706E+09, 1.4825E+09,
     + 1.4945E+09, 1.5066E+09, 1.5187E+09, 1.5310E+09, 1.5434E+09,
     +  1.5558E+09, 1.5684E+09, 1.5810E+09, 1.5937E+09, 1.6066E+09,
     + 1.6195E+09, 1.6325E+09, 1.6457E+09, 1.6589E+09, 1.6722E+09,
     +  1.6856E+09, 1.6991E+09, 1.7128E+09, 1.7265E+09, 1.7403E+09,
     + 1.7542E+09, 1.7683E+09, 1.7824E+09, 1.7967E+09, 1.8110E+09,
     +  1.8255E+09, 1.8400E+09, 1.8547E+09, 1.8695E+09, 1.8844E+09,
     + 1.8994E+09, 1.9145E+09, 1.9297E+09, 1.9451E+09, 1.9605E+09,
     +  1.9761E+09, 1.9918E+09, 2.0076E+09, 2.0235E+09, 2.0395E+09,
     + 2.0557E+09, 2.0719E+09, 2.0883E+09, 2.1048E+09, 2.1215E+09,
     +  2.1382E+09, 2.1551E+09, 2.1721E+09, 2.1892E+09, 2.2065E+09,
     + 2.2239E+09, 2.2414E+09, 2.2590E+09, 2.2768E+09, 2.2947E+09,
     +  2.3127E+09, 2.3308E+09, 2.3491E+09, 2.3676E+09, 2.3861E+09,
     + 2.4048E+09, 2.4236E+09, 2.4426E+09, 2.4617E+09, 2.4810E+09,
     +  2.5004E+09, 2.5199E+09, 2.5396E+09, 2.5594E+09, 2.5793E+09,
     + 2.5994E+09, 2.6197E+09, 2.6401E+09, 2.6606E+09, 2.6813E+09,
     +  2.7022E+09, 2.7232E+09, 2.7443E+09, 2.7656E+09, 2.7871E+09,
     + 2.8087E+09, 2.8304E+09, 2.8524E+09, 2.8744E+09, 2.8967E+09,
     +  2.9191E+09, 2.9416E+09, 2.9644E+09, 2.9872E+09, 3.0103E+09,
     + 3.0335E+09, 3.0569E+09, 3.0805E+09, 3.1042E+09, 3.1281E+09,
     +  3.1521E+09, 3.1764E+09, 3.2008E+09, 3.2254E+09, 3.2501E+09,
     + 3.2751E+09, 3.3002E+09, 3.3255E+09, 3.3509E+09, 3.3766E+09,
     +  3.4024E+09, 3.4285E+09, 3.4547E+09, 3.4811E+09, 3.5077E+09,
     + 3.5344E+09, 3.5614E+09, 3.5885E+09, 3.6159E+09, 3.6434E+09,
     +  3.6712E+09, 3.6991E+09, 3.7272E+09, 3.7556E+09, 3.7841E+09,
     + 3.8128E+09, 3.8418E+09, 3.8709E+09, 3.9003E+09, 3.9298E+09,
     +  3.9596E+09, 3.9896E+09, 4.0198E+09, 4.0502E+09, 4.0808E+09,
     + 4.1116E+09, 4.1426E+09, 4.1739E+09, 4.2054E+09, 4.2371E+09,
     +  4.2690E+09, 4.3012E+09, 4.3335E+09, 4.3661E+09, 4.3990E+09,
     + 4.4320E+09, 4.4653E+09, 4.4989E+09, 4.5326E+09, 4.5666E+09,
     +  4.6009E+09, 4.6353E+09, 4.6700E+09, 4.7050E+09, 4.7402E+09,
     + 4.7756E+09, 4.8113E+09, 4.8473E+09, 4.8835E+09, 4.9199E+09,
     +  4.9566E+09, 4.9935E+09, 5.0308E+09, 5.0682E+09, 5.1059E+09,
     + 5.1439E+09, 5.1822E+09, 5.2207E+09, 5.2595E+09, 5.2985E+09,
     +  5.3378E+09, 5.3774E+09, 5.4173E+09, 5.4574E+09, 5.4978E+09,
     + 5.5385E+09, 5.5795E+09, 5.6208E+09, 5.6623E+09, 5.7041E+09,
     +  5.7463E+09, 5.7887E+09, 5.8314E+09, 5.8743E+09, 5.9176E+09,
     + 5.9612E+09, 6.0051E+09, 6.0493E+09, 6.0938E+09, 6.1385E+09,
     +  6.1836E+09, 6.2290E+09, 6.2747E+09, 6.3208E+09, 6.3671E+09,
     + 6.4138E+09, 6.4607E+09, 6.5080E+09, 6.5557E+09, 6.6036E+09,
     +  6.6519E+09, 6.7005E+09, 6.7494E+09, 6.7986E+09, 6.8482E+09,
     + 6.8982E+09, 6.9484E+09, 6.9990E+09, 7.0500E+09, 7.1013E+09,
     +  7.1529E+09, 7.2049E+09, 7.2573E+09, 7.3100E+09, 7.3631E+09,
     + 7.4165E+09, 7.4702E+09, 7.5244E+09, 7.5789E+09, 7.6338E+09,
     +  7.6890E+09, 7.7446E+09, 7.8006E+09, 7.8570E+09, 7.9137E+09,
     + 7.9709E+09, 8.0284E+09, 8.0863E+09, 8.1445E+09, 8.2032E+09,
     +  8.2623E+09, 8.3218E+09, 8.3816E+09, 8.4419E+09, 8.5026E+09,
     + 8.5636E+09, 8.6251E+09, 8.6870E+09, 8.7493E+09, 8.8120E+09,
     +  8.8752E+09, 8.9387E+09, 9.0027E+09, 9.0671E+09, 9.1320E+09,
     + 9.1973E+09, 9.2630E+09, 9.3291E+09, 9.3957E+09, 9.4627E+09,
     +  9.5302E+09, 9.5981E+09, 9.6665E+09, 9.7353E+09, 9.8046E+09,
     + 9.8743E+09, 9.9445E+09, 1.0015E+10, 1.0086E+10, 1.0158E+10,
     +  1.0230E+10, 1.0303E+10, 1.0376E+10, 1.0449E+10, 1.0523E+10,
     + 1.0598E+10, 1.0673E+10, 1.0748E+10, 1.0824E+10, 1.0900E+10,
     +  1.0977E+10, 1.1055E+10, 1.1133E+10, 1.1211E+10, 1.1290E+10,
     + 1.1370E+10, 1.1450E+10, 1.1530E+10, 1.1612E+10, 1.1693E+10,
     +  1.1775E+10, 1.1858E+10, 1.1941E+10, 1.2025E+10, 1.2109E+10,
     + 1.2194E+10, 1.2280E+10, 1.2366E+10, 1.2452E+10, 1.2539E+10,
     +  1.2627E+10, 1.2715E+10, 1.2804E+10, 1.2894E+10, 1.2984E+10,
     + 1.3074E+10, 1.3165E+10, 1.3257E+10, 1.3349E+10, 1.3442E+10,
     +  1.3536E+10, 1.3630E+10, 1.3725E+10, 1.3820E+10, 1.3916E+10,
     + 1.4013E+10, 1.4110E+10, 1.4208E+10, 1.4306E+10, 1.4405E+10,
     +  1.4505E+10, 1.4606E+10, 1.4707E+10, 1.4808E+10, 1.4911E+10,
     + 1.5014E+10, 1.5117E+10, 1.5222E+10, 1.5327E+10, 1.5432E+10,
     +  1.5539E+10, 1.5646E+10, 1.5754E+10, 1.5862E+10, 1.5971E+10,
     + 1.6081E+10, 1.6191E+10, 1.6303E+10, 1.6415E+10, 1.6527E+10,
     +  1.6641E+10, 1.6755E+10, 1.6870E+10, 1.6985E+10, 1.7101E+10,
     + 1.7218E+10, 1.7336E+10, 1.7455E+10, 1.7574E+10, 1.7694E+10,
     +  1.7815E+10, 1.7936E+10, 1.8059E+10, 1.8182E+10, 1.8306E+10,
     + 1.8430E+10, 1.8556E+10, 1.8682E+10, 1.8809E+10, 1.8937E+10,
     +  1.9066E+10, 1.9195E+10, 1.9326E+10, 1.9457E+10, 1.9589E+10,
     + 1.9721E+10, 1.9855E+10, 1.9989E+10, 2.0125E+10, 2.0261E+10,
     +  2.0398E+10, 2.0536E+10, 2.0675E+10, 2.0814E+10, 2.0955E+10,
     + 2.1096E+10, 2.1238E+10, 2.1382E+10, 2.1526E+10, 2.1671E+10,
     +  2.1816E+10, 2.1963E+10, 2.2111E+10, 2.2260E+10, 2.2409E+10,
     + 2.2560E+10, 2.2711E+10, 2.2863E+10, 2.3017E+10, 2.3171E+10,
     +  2.3326E+10, 2.3482E+10, 2.3640E+10, 2.3798E+10, 2.3957E+10,
     + 2.4117E+10, 2.4278E+10, 2.4440E+10, 2.4604E+10, 2.4768E+10,
     +  2.4933E+10, 2.5099E+10, 2.5266E+10, 2.5435E+10, 2.5604E+10,
     + 2.5774E+10, 2.5946E+10, 2.6118E+10, 2.6292E+10, 2.6466E+10,
     +  2.6642E+10, 2.6819E+10, 2.6996E+10, 2.7175E+10, 2.7355E+10,
     + 2.7537E+10, 2.7719E+10, 2.7902E+10, 2.8087E+10, 2.8272E+10,
     +  2.8459E+10, 2.8647E+10, 2.8836E+10, 2.9027E+10, 2.9218E+10,
     + 2.9411E+10, 2.9604E+10, 2.9799E+10, 2.9996E+10, 3.0193E+10,
     +  3.0392E+10, 3.0591E+10, 3.0792E+10, 3.0995E+10, 3.1198E+10,
     + 3.1403E+10, 3.1609E+10, 3.1816E+10, 3.2024E+10, 3.2234E+10,
     +  3.2445E+10, 3.2658E+10, 3.2871E+10, 3.3086E+10, 3.3302E+10,
     + 3.3520E+10, 3.3739E+10, 3.3959E+10, 3.4180E+10, 3.4403E+10,
     +  3.4627E+10, 3.4853E+10, 3.5080E+10, 3.5308E+10, 3.5538E+10,
     + 3.5769E+10, 3.6001E+10, 3.6235E+10, 3.6471E+10, 3.6707E+10,
     +  3.6945E+10, 3.7185E+10, 3.7426E+10, 3.7668E+10, 3.7912E+10,
     + 3.8158E+10, 3.8405E+10, 3.8653E+10, 3.8903E+10, 3.9154E+10,
     +  3.9407E+10, 3.9661E+10, 3.9917E+10, 4.0175E+10, 4.0434E+10,
     + 4.0694E+10, 4.0956E+10, 4.1220E+10, 4.1485E+10, 4.1752E+10,
     +  4.2021E+10, 4.2291E+10, 4.2562E+10, 4.2836E+10, 4.3110E+10,
     + 4.3387E+10, 4.3665E+10, 4.3945E+10, 4.4227E+10, 4.4510E+10,
     +  4.4795E+10, 4.5081E+10, 4.5369E+10, 4.5659E+10, 4.5951E+10,
     + 4.6245E+10, 4.6540E+10, 4.6837E+10, 4.7135E+10, 4.7436E+10,
     +  4.7738E+10, 4.8042E+10, 4.8348E+10, 4.8656E+10, 4.8965E+10,
     + 4.9276E+10, 4.9589E+10, 4.9904E+10, 5.0221E+10, 5.0540E+10,
     +  5.0860E+10, 5.1183E+10, 5.1507E+10, 5.1834E+10, 5.2162E+10,
     + 5.2492E+10, 5.2824E+10, 5.3158E+10, 5.3494E+10, 5.3832E+10,
     +  5.4172E+10, 5.4514E+10, 5.4858E+10, 5.5204E+10, 5.5552E+10,
     + 5.5902E+10, 5.6254E+10, 5.6608E+10, 5.6964E+10, 5.7323E+10,
     +  5.7683E+10, 5.8046E+10, 5.8410E+10, 5.8777E+10, 5.9146E+10,
     + 5.9517E+10, 5.9890E+10, 6.0265E+10, 6.0643E+10, 6.1023E+10,
     +  6.1405E+10, 6.1789E+10, 6.2175E+10, 6.2564E+10, 6.2955E+10,
     + 6.3348E+10, 6.3744E+10, 6.4141E+10, 6.4542E+10, 6.4944E+10,
     +  6.5349E+10, 6.5756E+10, 6.6165E+10, 6.6577E+10, 6.6991E+10,
     + 6.7408E+10, 6.7827E+10, 6.8248E+10, 6.8672E+10, 6.9098E+10,
     +  6.9527E+10, 6.9958E+10, 7.0392E+10, 7.0828E+10, 7.1267E+10,
     + 7.1708E+10, 7.2152E+10, 7.2598E+10, 7.3047E+10, 7.3499E+10,
     +  7.3953E+10, 7.4409E+10, 7.4869E+10, 7.5330E+10, 7.5795E+10,
     + 7.6262E+10, 7.6732E+10, 7.7205E+10, 7.7680E+10, 7.8158E+10,
     +  7.8639E+10, 7.9122E+10, 7.9609E+10, 8.0098E+10, 8.0590E+10,
     + 8.1084E+10, 8.1582E+10, 8.2082E+10, 8.2585E+10, 8.3091E+10,
     +  8.3600E+10, 8.4112E+10, 8.4626E+10, 8.5144E+10, 8.5665E+10,
     + 8.6188E+10, 8.6715E+10, 8.7244E+10, 8.7777E+10, 8.8312E+10,
     +  8.8851E+10, 8.9392E+10, 8.9937E+10, 9.0485E+10, 9.1036E+10,
     + 9.1589E+10, 9.2147E+10, 9.2707E+10, 9.3270E+10, 9.3837E+10,
     +  9.4407E+10, 9.4980E+10, 9.5556E+10, 9.6135E+10, 9.6718E+10,
     + 9.7304E+10, 9.7893E+10, 9.8486E+10, 9.9082E+10, 9.9681E+10,
     +  1.0028E+11, 1.0089E+11, 1.0150E+11, 1.0211E+11, 1.0273E+11,
     + 1.0335E+11, 1.0397E+11, 1.0460E+11, 1.0523E+11, 1.0586E+11,
     +  1.0650E+11, 1.0714E+11, 1.0779E+11, 1.0843E+11, 1.0909E+11,
     + 1.0974E+11, 1.1040E+11, 1.1106E+11, 1.1173E+11, 1.1240E+11,
     +  1.1307E+11, 1.1375E+11, 1.1443E+11, 1.1512E+11, 1.1581E+11,
     + 1.1650E+11, 1.1719E+11, 1.1790E+11, 1.1860E+11, 1.1931E+11,
     +  1.2002E+11, 1.2074E+11, 1.2146E+11, 1.2218E+11, 1.2291E+11,
     + 1.2364E+11, 1.2438E+11, 1.2512E+11, 1.2586E+11, 1.2661E+11,
     +  1.2736E+11, 1.2812E+11, 1.2888E+11, 1.2965E+11, 1.3042E+11,
     + 1.3119E+11, 1.3197E+11, 1.3275E+11, 1.3354E+11, 1.3433E+11,
     +  1.3512E+11, 1.3592E+11, 1.3672E+11, 1.3753E+11, 1.3835E+11,
     + 1.3916E+11, 1.3998E+11, 1.4081E+11, 1.4164E+11, 1.4248E+11,
     +  1.4332E+11, 1.4416E+11, 1.4501E+11, 1.4586E+11, 1.4672E+11,
     + 1.4759E+11, 1.4845E+11, 1.4933E+11, 1.5020E+11, 1.5109E+11,
     +  1.5197E+11, 1.5286E+11, 1.5376E+11, 1.5466E+11, 1.5557E+11,
     + 1.5648E+11, 1.5740E+11, 1.5832E+11, 1.5924E+11, 1.6018E+11,
     +  1.6111E+11, 1.6205E+11, 1.6300E+11, 1.6395E+11, 1.6491E+11,
     + 1.6587E+11, 1.6684E+11, 1.6781E+11, 1.6879E+11, 1.6977E+11,
     +  1.7076E+11, 1.7175E+11, 1.7275E+11, 1.7376E+11, 1.7477E+11,
     + 1.7578E+11, 1.7680E+11, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --      3124
      data (QofT( 2,J),J=1,2408)/ 4.7565E+01, 1.1191E+02, 1.8507E+02,
     +  2.6514E+02, 3.5147E+02, 4.4362E+02, 5.4120E+02, 6.4389E+02,
     + 7.5145E+02, 8.6369E+02, 9.8049E+02, 1.1018E+03, 1.2275E+03,
     +  1.3575E+03, 1.4920E+03, 1.6307E+03, 1.7737E+03, 1.9210E+03,
     + 2.0725E+03, 2.2281E+03, 2.5517E+03, 2.8912E+03, 3.2462E+03,
     +  3.6161E+03, 4.0005E+03, 4.3988E+03, 4.8106E+03, 5.2354E+03,
     + 5.6728E+03, 6.1223E+03, 6.5837E+03, 7.0565E+03, 7.5406E+03,
     +  8.0356E+03, 8.5412E+03, 9.0573E+03, 9.5837E+03, 1.0120E+04,
     + 1.0667E+04, 1.1223E+04, 1.1789E+04, 1.2365E+04, 1.2950E+04,
     +  1.3545E+04, 1.4149E+04, 1.4763E+04, 1.5386E+04, 1.6019E+04,
     + 1.6661E+04, 1.7313E+04, 1.7975E+04, 1.8646E+04, 1.9327E+04,
     +  2.0018E+04, 2.0719E+04, 2.1430E+04, 2.2152E+04, 2.2883E+04,
     + 2.3626E+04, 2.4378E+04, 2.5142E+04, 2.5917E+04, 2.6702E+04,
     +  2.7499E+04, 2.8307E+04, 2.9127E+04, 2.9958E+04, 3.0802E+04,
     + 3.1657E+04, 3.2525E+04, 3.3405E+04, 3.4297E+04, 3.5203E+04,
     +  3.6122E+04, 3.7053E+04, 3.7998E+04, 3.8957E+04, 3.9929E+04,
     + 4.0916E+04, 4.1916E+04, 4.2931E+04, 4.3960E+04, 4.5005E+04,
     +  4.6064E+04, 4.7138E+04, 4.8228E+04, 4.9333E+04, 5.0454E+04,
     + 5.1591E+04, 5.2744E+04, 5.3914E+04, 5.5100E+04, 5.6304E+04,
     +  5.7524E+04, 5.8762E+04, 6.0017E+04, 6.1290E+04, 6.2581E+04,
     + 6.3890E+04, 6.5218E+04, 6.6565E+04, 6.7930E+04, 6.9315E+04,
     +  7.0719E+04, 7.2143E+04, 7.3587E+04, 7.5051E+04, 7.6536E+04,
     + 7.8041E+04, 7.9567E+04, 8.1115E+04, 8.2684E+04, 8.4275E+04,
     +  8.5888E+04, 8.7523E+04, 8.9181E+04, 9.0862E+04, 9.2567E+04,
     + 9.4294E+04, 9.6046E+04, 9.7822E+04, 9.9622E+04, 1.0145E+05,
     +  1.0330E+05, 1.0517E+05, 1.0707E+05, 1.0900E+05, 1.1095E+05,
     + 1.1293E+05, 1.1494E+05, 1.1698E+05, 1.1904E+05, 1.2113E+05,
     +  1.2325E+05, 1.2540E+05, 1.2757E+05, 1.2978E+05, 1.3202E+05,
     + 1.3428E+05, 1.3658E+05, 1.3891E+05, 1.4127E+05, 1.4366E+05,
     +  1.4609E+05, 1.4854E+05, 1.5103E+05, 1.5356E+05, 1.5612E+05,
     + 1.5871E+05, 1.6133E+05, 1.6400E+05, 1.6669E+05, 1.6943E+05,
     +  1.7220E+05, 1.7501E+05, 1.7785E+05, 1.8074E+05, 1.8366E+05,
     + 1.8662E+05, 1.8962E+05, 1.9266E+05, 1.9574E+05, 1.9887E+05,
     +  2.0203E+05, 2.0524E+05, 2.0849E+05, 2.1178E+05, 2.1512E+05,
     + 2.1850E+05, 2.2192E+05, 2.2539E+05, 2.2891E+05, 2.3247E+05,
     +  2.3608E+05, 2.3974E+05, 2.4345E+05, 2.4721E+05, 2.5101E+05,
     + 2.5487E+05, 2.5878E+05, 2.6274E+05, 2.6675E+05, 2.7081E+05,
     +  2.7493E+05, 2.7910E+05, 2.8333E+05, 2.8761E+05, 2.9195E+05,
     + 2.9634E+05, 3.0080E+05, 3.0531E+05, 3.0988E+05, 3.1451E+05,
     +  3.1920E+05, 3.2396E+05, 3.2877E+05, 3.3365E+05, 3.3859E+05,
     + 3.4360E+05, 3.4867E+05, 3.5381E+05, 3.5902E+05, 3.6429E+05,
     +  3.6963E+05, 3.7505E+05, 3.8053E+05, 3.8608E+05, 3.9171E+05,
     + 3.9741E+05, 4.0318E+05, 4.0903E+05, 4.1496E+05, 4.2096E+05,
     +  4.2704E+05, 4.3319E+05, 4.3943E+05, 4.4575E+05, 4.5215E+05,
     + 4.5863E+05, 4.6520E+05, 4.7185E+05, 4.7859E+05, 4.8541E+05,
     +  4.9232E+05, 4.9932E+05, 5.0642E+05, 5.1360E+05, 5.2087E+05,
     + 5.2824E+05, 5.3570E+05, 5.4326E+05, 5.5092E+05, 5.5867E+05,
     +  5.6652E+05, 5.7447E+05, 5.8253E+05, 5.9069E+05, 5.9895E+05,
     + 6.0732E+05, 6.1579E+05, 6.2437E+05, 6.3307E+05, 6.4187E+05,
     +  6.5078E+05, 6.5981E+05, 6.6895E+05, 6.7821E+05, 6.8759E+05,
     + 6.9708E+05, 7.0670E+05, 7.1644E+05, 7.2630E+05, 7.3628E+05,
     +  7.4639E+05, 7.5663E+05, 7.6700E+05, 7.7750E+05, 7.8813E+05,
     + 7.9890E+05, 8.0980E+05, 8.2084E+05, 8.3202E+05, 8.4334E+05,
     +  8.5480E+05, 8.6640E+05, 8.7815E+05, 8.9005E+05, 9.0209E+05,
     + 9.1429E+05, 9.2664E+05, 9.3914E+05, 9.5180E+05, 9.6462E+05,
     +  9.7760E+05, 9.9074E+05, 1.0040E+06, 1.0175E+06, 1.0312E+06,
     + 1.0450E+06, 1.0589E+06, 1.0731E+06, 1.0874E+06, 1.1019E+06,
     +  1.1166E+06, 1.1315E+06, 1.1465E+06, 1.1618E+06, 1.1772E+06,
     + 1.1928E+06, 1.2086E+06, 1.2246E+06, 1.2408E+06, 1.2572E+06,
     +  1.2738E+06, 1.2906E+06, 1.3076E+06, 1.3249E+06, 1.3423E+06,
     + 1.3599E+06, 1.3778E+06, 1.3959E+06, 1.4142E+06, 1.4327E+06,
     +  1.4515E+06, 1.4705E+06, 1.4897E+06, 1.5091E+06, 1.5288E+06,
     + 1.5487E+06, 1.5689E+06, 1.5893E+06, 1.6100E+06, 1.6309E+06,
     +  1.6521E+06, 1.6735E+06, 1.6952E+06, 1.7171E+06, 1.7393E+06,
     + 1.7618E+06, 1.7846E+06, 1.8076E+06, 1.8309E+06, 1.8545E+06,
     +  1.8784E+06, 1.9025E+06, 1.9270E+06, 1.9517E+06, 1.9768E+06,
     + 2.0021E+06, 2.0277E+06, 2.0537E+06, 2.0800E+06, 2.1065E+06,
     +  2.1334E+06, 2.1607E+06, 2.1882E+06, 2.2161E+06, 2.2443E+06,
     + 2.2728E+06, 2.3017E+06, 2.3309E+06, 2.3605E+06, 2.3905E+06,
     +  2.4207E+06, 2.4514E+06, 2.4824E+06, 2.5138E+06, 2.5455E+06,
     + 2.5776E+06, 2.6101E+06, 2.6430E+06, 2.6763E+06, 2.7100E+06,
     +  2.7441E+06, 2.7785E+06, 2.8134E+06, 2.8487E+06, 2.8844E+06,
     + 2.9205E+06, 2.9571E+06, 2.9941E+06, 3.0315E+06, 3.0693E+06,
     +  3.1076E+06, 3.1464E+06, 3.1856E+06, 3.2252E+06, 3.2653E+06,
     + 3.3059E+06, 3.3470E+06, 3.3885E+06, 3.4306E+06, 3.4731E+06,
     +  3.5161E+06, 3.5596E+06, 3.6036E+06, 3.6482E+06, 3.6932E+06,
     + 3.7388E+06, 3.7849E+06, 3.8315E+06, 3.8787E+06, 3.9264E+06,
     +  3.9747E+06, 4.0235E+06, 4.0729E+06, 4.1229E+06, 4.1734E+06,
     + 4.2245E+06, 4.2762E+06, 4.3285E+06, 4.3814E+06, 4.4349E+06,
     +  4.4891E+06, 4.5438E+06, 4.5992E+06, 4.6552E+06, 4.7119E+06,
     + 4.7692E+06, 4.8271E+06, 4.8858E+06, 4.9450E+06, 5.0050E+06,
     +  5.0657E+06, 5.1270E+06, 5.1891E+06, 5.2518E+06, 5.3153E+06,
     + 5.3794E+06, 5.4444E+06, 5.5100E+06, 5.5764E+06, 5.6435E+06,
     +  5.7115E+06, 5.7801E+06, 5.8496E+06, 5.9198E+06, 5.9908E+06,
     + 6.0627E+06, 6.1353E+06, 6.2088E+06, 6.2831E+06, 6.3582E+06,
     +  6.4342E+06, 6.5110E+06, 6.5887E+06, 6.6672E+06, 6.7467E+06,
     + 6.8270E+06, 6.9082E+06, 6.9904E+06, 7.0735E+06, 7.1574E+06,
     +  7.2424E+06, 7.3282E+06, 7.4151E+06, 7.5029E+06, 7.5917E+06,
     + 7.6814E+06, 7.7722E+06, 7.8640E+06, 7.9568E+06, 8.0506E+06,
     +  8.1455E+06, 8.2414E+06, 8.3384E+06, 8.4364E+06, 8.5356E+06,
     + 8.6358E+06, 8.7372E+06, 8.8396E+06, 8.9432E+06, 9.0480E+06,
     +  9.1539E+06, 9.2609E+06, 9.3692E+06, 9.4786E+06, 9.5892E+06,
     + 9.7011E+06, 9.8142E+06, 9.9285E+06, 1.0044E+07, 1.0161E+07,
     +  1.0279E+07, 1.0398E+07, 1.0519E+07, 1.0641E+07, 1.0765E+07,
     + 1.0889E+07, 1.1015E+07, 1.1143E+07, 1.1272E+07, 1.1402E+07,
     +  1.1533E+07, 1.1667E+07, 1.1801E+07, 1.1937E+07, 1.2074E+07,
     + 1.2213E+07, 1.2354E+07, 1.2496E+07, 1.2639E+07, 1.2784E+07,
     +  1.2931E+07, 1.3079E+07, 1.3229E+07, 1.3380E+07, 1.3534E+07,
     + 1.3688E+07, 1.3845E+07, 1.4003E+07, 1.4162E+07, 1.4324E+07,
     +  1.4487E+07, 1.4652E+07, 1.4819E+07, 1.4987E+07, 1.5157E+07,
     + 1.5330E+07, 1.5503E+07, 1.5679E+07, 1.5857E+07, 1.6037E+07,
     +  1.6218E+07, 1.6401E+07, 1.6587E+07, 1.6774E+07, 1.6963E+07,
     + 1.7155E+07, 1.7348E+07, 1.7544E+07, 1.7741E+07, 1.7941E+07,
     +  1.8142E+07, 1.8346E+07, 1.8552E+07, 1.8760E+07, 1.8971E+07,
     + 1.9183E+07, 1.9398E+07, 1.9615E+07, 1.9835E+07, 2.0056E+07,
     +  2.0280E+07, 2.0507E+07, 2.0735E+07, 2.0966E+07, 2.1200E+07,
     + 2.1436E+07, 2.1675E+07, 2.1915E+07, 2.2159E+07, 2.2405E+07,
     +  2.2654E+07, 2.2905E+07, 2.3159E+07, 2.3415E+07, 2.3674E+07,
     + 2.3936E+07, 2.4201E+07, 2.4468E+07, 2.4738E+07, 2.5011E+07,
     +  2.5287E+07, 2.5565E+07, 2.5847E+07, 2.6131E+07, 2.6418E+07,
     + 2.6709E+07, 2.7002E+07, 2.7298E+07, 2.7598E+07, 2.7900E+07,
     +  2.8206E+07, 2.8514E+07, 2.8826E+07, 2.9141E+07, 2.9460E+07,
     + 2.9781E+07, 3.0106E+07, 3.0434E+07, 3.0766E+07, 3.1101E+07,
     +  3.1439E+07, 3.1781E+07, 3.2127E+07, 3.2476E+07, 3.2828E+07,
     + 3.3184E+07, 3.3544E+07, 3.3907E+07, 3.4274E+07, 3.4645E+07,
     +  3.5020E+07, 3.5398E+07, 3.5780E+07, 3.6166E+07, 3.6556E+07,
     + 3.6950E+07, 3.7348E+07, 3.7750E+07, 3.8156E+07, 3.8566E+07,
     +  3.8981E+07, 3.9399E+07, 3.9822E+07, 4.0249E+07, 4.0680E+07,
     + 4.1116E+07, 4.1556E+07, 4.2000E+07, 4.2449E+07, 4.2902E+07,
     +  4.3360E+07, 4.3823E+07, 4.4290E+07, 4.4762E+07, 4.5238E+07,
     + 4.5720E+07, 4.6206E+07, 4.6697E+07, 4.7193E+07, 4.7693E+07,
     +  4.8199E+07, 4.8710E+07, 4.9226E+07, 4.9747E+07, 5.0273E+07,
     + 5.0805E+07, 5.1342E+07, 5.1884E+07, 5.2431E+07, 5.2984E+07,
     +  5.3543E+07, 5.4106E+07, 5.4676E+07, 5.5251E+07, 5.5832E+07,
     + 5.6418E+07, 5.7011E+07, 5.7609E+07, 5.8213E+07, 5.8823E+07,
     +  5.9439E+07, 6.0061E+07, 6.0689E+07, 6.1323E+07, 6.1964E+07,
     + 6.2611E+07, 6.3264E+07, 6.3924E+07, 6.4590E+07, 6.5262E+07,
     +  6.5941E+07, 6.6627E+07, 6.7319E+07, 6.8019E+07, 6.8725E+07,
     + 6.9438E+07, 7.0157E+07, 7.0884E+07, 7.1618E+07, 7.2359E+07,
     +  7.3107E+07, 7.3863E+07, 7.4626E+07, 7.5396E+07, 7.6174E+07,
     + 7.6959E+07, 7.7752E+07, 7.8552E+07, 7.9360E+07, 8.0176E+07,
     +  8.1000E+07, 8.1832E+07, 8.2672E+07, 8.3520E+07, 8.4376E+07,
     + 8.5240E+07, 8.6113E+07, 8.6994E+07, 8.7884E+07, 8.8782E+07,
     +  8.9688E+07, 9.0604E+07, 9.1528E+07, 9.2461E+07, 9.3402E+07,
     + 9.4353E+07, 9.5313E+07, 9.6282E+07, 9.7261E+07, 9.8248E+07,
     +  9.9245E+07, 1.0025E+08, 1.0127E+08, 1.0229E+08, 1.0333E+08,
     + 1.0437E+08, 1.0543E+08, 1.0650E+08, 1.0757E+08, 1.0866E+08,
     +  1.0975E+08, 1.1086E+08, 1.1197E+08, 1.1310E+08, 1.1424E+08,
     + 1.1539E+08, 1.1655E+08, 1.1772E+08, 1.1890E+08, 1.2009E+08,
     +  1.2130E+08, 1.2251E+08, 1.2374E+08, 1.2498E+08, 1.2623E+08,
     + 1.2749E+08, 1.2876E+08, 1.3004E+08, 1.3134E+08, 1.3265E+08,
     +  1.3397E+08, 1.3531E+08, 1.3665E+08, 1.3801E+08, 1.3938E+08,
     + 1.4077E+08, 1.4216E+08, 1.4357E+08, 1.4500E+08, 1.4643E+08,
     +  1.4788E+08, 1.4935E+08, 1.5082E+08, 1.5231E+08, 1.5382E+08,
     + 1.5534E+08, 1.5687E+08, 1.5841E+08, 1.5997E+08, 1.6155E+08,
     +  1.6314E+08, 1.6474E+08, 1.6636E+08, 1.6800E+08, 1.6965E+08,
     + 1.7131E+08, 1.7299E+08, 1.7468E+08, 1.7639E+08, 1.7812E+08,
     +  1.7986E+08, 1.8162E+08, 1.8339E+08, 1.8519E+08, 1.8699E+08,
     + 1.8882E+08, 1.9065E+08, 1.9251E+08, 1.9439E+08, 1.9628E+08,
     +  1.9818E+08, 2.0011E+08, 2.0205E+08, 2.0401E+08, 2.0599E+08,
     + 2.0799E+08, 2.1000E+08, 2.1203E+08, 2.1408E+08, 2.1615E+08,
     +  2.1824E+08, 2.2035E+08, 2.2248E+08, 2.2462E+08, 2.2679E+08,
     + 2.2897E+08, 2.3118E+08, 2.3340E+08, 2.3564E+08, 2.3791E+08,
     +  2.4019E+08, 2.4250E+08, 2.4482E+08, 2.4717E+08, 2.4954E+08,
     + 2.5193E+08, 2.5434E+08, 2.5677E+08, 2.5922E+08, 2.6170E+08,
     +  2.6420E+08, 2.6672E+08, 2.6926E+08, 2.7183E+08, 2.7441E+08,
     + 2.7703E+08, 2.7966E+08, 2.8232E+08, 2.8500E+08, 2.8771E+08,
     +  2.9044E+08, 2.9319E+08, 2.9597E+08, 2.9877E+08, 3.0160E+08,
     + 3.0445E+08, 3.0733E+08, 3.1024E+08, 3.1316E+08, 3.1612E+08,
     +  3.1910E+08, 3.2211E+08, 3.2514E+08, 3.2820E+08, 3.3129E+08,
     + 3.3441E+08, 3.3755E+08, 3.4072E+08, 3.4392E+08, 3.4714E+08,
     +  3.5040E+08, 3.5368E+08, 3.5699E+08, 3.6034E+08, 3.6371E+08,
     + 3.6711E+08, 3.7053E+08, 3.7399E+08, 3.7748E+08, 3.8100E+08,
     +  3.8455E+08, 3.8814E+08, 3.9175E+08, 3.9539E+08, 3.9907E+08,
     + 4.0278E+08, 4.0652E+08, 4.1029E+08, 4.1410E+08, 4.1793E+08,
     +  4.2181E+08, 4.2571E+08, 4.2965E+08, 4.3362E+08, 4.3763E+08,
     + 4.4168E+08, 4.4575E+08, 4.4986E+08, 4.5401E+08, 4.5820E+08,
     +  4.6242E+08, 4.6667E+08, 4.7097E+08, 4.7530E+08, 4.7966E+08,
     + 4.8407E+08, 4.8851E+08, 4.9299E+08, 4.9751E+08, 5.0207E+08,
     +  5.0666E+08, 5.1130E+08, 5.1598E+08, 5.2069E+08, 5.2545E+08,
     + 5.3024E+08, 5.3508E+08, 5.3996E+08, 5.4488E+08, 5.4984E+08,
     +  5.5485E+08, 5.5990E+08, 5.6499E+08, 5.7012E+08, 5.7530E+08,
     + 5.8052E+08, 5.8578E+08, 5.9109E+08, 5.9645E+08, 6.0185E+08,
     +  6.0729E+08, 6.1279E+08, 6.1832E+08, 6.2391E+08, 6.2954E+08,
     + 6.3522E+08, 6.4095E+08, 6.4673E+08, 6.5255E+08, 6.5843E+08,
     +  6.6435E+08, 6.7032E+08, 6.7635E+08, 6.8242E+08, 6.8854E+08,
     + 6.9472E+08, 7.0095E+08, 7.0723E+08, 7.1356E+08, 7.1995E+08,
     +  7.2639E+08, 7.3288E+08, 7.3943E+08, 7.4603E+08, 7.5269E+08,
     + 7.5940E+08, 7.6617E+08, 7.7299E+08, 7.7988E+08, 7.8682E+08,
     +  7.9381E+08, 8.0087E+08, 8.0798E+08, 8.1515E+08, 8.2239E+08,
     + 8.2968E+08, 8.3703E+08, 8.4444E+08, 8.5192E+08, 8.5945E+08,
     +  8.6705E+08, 8.7471E+08, 8.8244E+08, 8.9023E+08, 8.9808E+08,
     + 9.0600E+08, 9.1398E+08, 9.2203E+08, 9.3014E+08, 9.3833E+08,
     +  9.4657E+08, 9.5489E+08, 9.6328E+08, 9.7173E+08, 9.8025E+08,
     + 9.8884E+08, 9.9751E+08, 1.0062E+09, 1.0150E+09, 1.0239E+09,
     +  1.0329E+09, 1.0419E+09, 1.0510E+09, 1.0602E+09, 1.0694E+09,
     + 1.0787E+09, 1.0881E+09, 1.0976E+09, 1.1071E+09, 1.1168E+09,
     +  1.1265E+09, 1.1363E+09, 1.1461E+09, 1.1561E+09, 1.1661E+09,
     + 1.1762E+09, 1.1864E+09, 1.1967E+09, 1.2070E+09, 1.2174E+09,
     +  1.2280E+09, 1.2386E+09, 1.2493E+09, 1.2600E+09, 1.2709E+09,
     + 1.2819E+09, 1.2929E+09, 1.3040E+09, 1.3152E+09, 1.3266E+09,
     +  1.3380E+09, 1.3494E+09, 1.3610E+09, 1.3727E+09, 1.3845E+09,
     + 1.3963E+09, 1.4083E+09, 1.4203E+09, 1.4325E+09, 1.4447E+09,
     +  1.4571E+09, 1.4695E+09, 1.4821E+09, 1.4947E+09, 1.5075E+09,
     + 1.5203E+09, 1.5332E+09, 1.5463E+09, 1.5595E+09, 1.5727E+09,
     +  1.5861E+09, 1.5995E+09, 1.6131E+09, 1.6268E+09, 1.6406E+09,
     + 1.6545E+09, 1.6685E+09, 1.6826E+09, 1.6969E+09, 1.7112E+09,
     +  1.7257E+09, 1.7402E+09, 1.7549E+09, 1.7697E+09, 1.7846E+09,
     + 1.7997E+09, 1.8148E+09, 1.8301E+09, 1.8455E+09, 1.8610E+09,
     +  1.8767E+09, 1.8924E+09, 1.9083E+09, 1.9243E+09, 1.9404E+09,
     + 1.9567E+09, 1.9731E+09, 1.9896E+09, 2.0062E+09, 2.0230E+09,
     +  2.0399E+09, 2.0569E+09, 2.0741E+09, 2.0914E+09, 2.1089E+09,
     + 2.1264E+09, 2.1441E+09, 2.1620E+09, 2.1800E+09, 2.1981E+09,
     +  2.2164E+09, 2.2348E+09, 2.2533E+09, 2.2720E+09, 2.2908E+09,
     + 2.3098E+09, 2.3290E+09, 2.3482E+09, 2.3677E+09, 2.3872E+09,
     +  2.4070E+09, 2.4268E+09, 2.4469E+09, 2.4671E+09, 2.4874E+09,
     + 2.5079E+09, 2.5286E+09, 2.5494E+09, 2.5704E+09, 2.5915E+09,
     +  2.6128E+09, 2.6343E+09, 2.6559E+09, 2.6777E+09, 2.6996E+09,
     + 2.7218E+09, 2.7441E+09, 2.7665E+09, 2.7892E+09, 2.8120E+09,
     +  2.8350E+09, 2.8581E+09, 2.8814E+09, 2.9050E+09, 2.9287E+09,
     + 2.9525E+09, 2.9766E+09, 3.0008E+09, 3.0252E+09, 3.0498E+09,
     +  3.0746E+09, 3.0996E+09, 3.1248E+09, 3.1502E+09, 3.1757E+09,
     + 3.2015E+09, 3.2274E+09, 3.2535E+09, 3.2799E+09, 3.3064E+09,
     +  3.3331E+09, 3.3601E+09, 3.3872E+09, 3.4145E+09, 3.4421E+09,
     + 3.4699E+09, 3.4978E+09, 3.5260E+09, 3.5544E+09, 3.5830E+09,
     +  3.6118E+09, 3.6408E+09, 3.6701E+09, 3.6995E+09, 3.7292E+09,
     + 3.7591E+09, 3.7893E+09, 3.8196E+09, 3.8502E+09, 3.8810E+09,
     +  3.9121E+09, 3.9433E+09, 3.9748E+09, 4.0066E+09, 4.0386E+09,
     + 4.0708E+09, 4.1032E+09, 4.1359E+09, 4.1689E+09, 4.2021E+09,
     +  4.2355E+09, 4.2692E+09, 4.3031E+09, 4.3373E+09, 4.3717E+09,
     + 4.4064E+09, 4.4413E+09, 4.4765E+09, 4.5120E+09, 4.5477E+09,
     +  4.5837E+09, 4.6200E+09, 4.6565E+09, 4.6933E+09, 4.7303E+09,
     + 4.7677E+09, 4.8053E+09, 4.8432E+09, 4.8813E+09, 4.9198E+09,
     +  4.9585E+09, 4.9975E+09, 5.0368E+09, 5.0764E+09, 5.1163E+09,
     + 5.1564E+09, 5.1969E+09, 5.2376E+09, 5.2787E+09, 5.3200E+09,
     +  5.3617E+09, 5.4037E+09, 5.4459E+09, 5.4885E+09, 5.5314E+09,
     + 5.5746E+09, 5.6181E+09, 5.6619E+09, 5.7061E+09, 5.7505E+09,
     +  5.7953E+09, 5.8404E+09, 5.8859E+09, 5.9316E+09, 5.9777E+09,
     + 6.0242E+09, 6.0709E+09, 6.1181E+09, 6.1655E+09, 6.2133E+09,
     +  6.2614E+09, 6.3099E+09, 6.3588E+09, 6.4080E+09, 6.4575E+09,
     + 6.5074E+09, 6.5577E+09, 6.6083E+09, 6.6593E+09, 6.7106E+09,
     +  6.7623E+09, 6.8144E+09, 6.8669E+09, 6.9197E+09, 6.9730E+09,
     + 7.0266E+09, 7.0805E+09, 7.1349E+09, 7.1897E+09, 7.2448E+09,
     +  7.3004E+09, 7.3563E+09, 7.4127E+09, 7.4694E+09, 7.5266E+09,
     + 7.5841E+09, 7.6421E+09, 7.7005E+09, 7.7593E+09, 7.8185E+09,
     +  7.8781E+09, 7.9382E+09, 7.9987E+09, 8.0596E+09, 8.1209E+09,
     + 8.1827E+09, 8.2449E+09, 8.3076E+09, 8.3707E+09, 8.4342E+09,
     +  8.4982E+09, 8.5627E+09, 8.6276E+09, 8.6930E+09, 8.7588E+09,
     + 8.8251E+09, 8.8918E+09, 8.9591E+09, 9.0268E+09, 9.0949E+09,
     +  9.1636E+09, 9.2327E+09, 9.3024E+09, 9.3725E+09, 9.4431E+09,
     + 9.5142E+09, 9.5858E+09, 9.6579E+09, 9.7305E+09, 9.8036E+09,
     +  9.8773E+09, 9.9514E+09, 1.0026E+10, 1.0101E+10, 1.0177E+10,
     + 1.0253E+10, 1.0330E+10, 1.0407E+10, 1.0485E+10, 1.0563E+10,
     +  1.0642E+10, 1.0722E+10, 1.0802E+10, 1.0883E+10, 1.0964E+10,
     + 1.1045E+10, 1.1128E+10, 1.1211E+10, 1.1294E+10, 1.1378E+10,
     +  1.1463E+10, 1.1548E+10, 1.1633E+10, 1.1720E+10, 1.1807E+10,
     + 1.1894E+10, 1.1982E+10, 1.2071E+10, 1.2161E+10, 1.2251E+10,
     +  1.2341E+10, 1.2432E+10, 1.2524E+10, 1.2617E+10, 1.2710E+10,
     + 1.2804E+10, 1.2898E+10, 1.2993E+10, 1.3089E+10, 1.3185E+10,
     +  1.3282E+10, 1.3380E+10, 1.3478E+10, 1.3577E+10, 1.3677E+10,
     + 1.3778E+10, 1.3879E+10, 1.3980E+10, 1.4083E+10, 1.4186E+10,
     +  1.4290E+10, 1.4394E+10, 1.4500E+10, 1.4606E+10, 1.4713E+10,
     + 1.4820E+10, 1.4928E+10, 1.5037E+10, 1.5147E+10, 1.5257E+10,
     +  1.5368E+10, 1.5480E+10, 1.5593E+10, 1.5706E+10, 1.5821E+10,
     + 1.5936E+10, 1.6051E+10, 1.6168E+10, 1.6285E+10, 1.6403E+10,
     +  1.6522E+10, 1.6642E+10, 1.6762E+10, 1.6884E+10, 1.7006E+10,
     + 1.7129E+10, 1.7253E+10, 1.7377E+10, 1.7503E+10, 1.7629E+10,
     +  1.7756E+10, 1.7884E+10, 1.8013E+10, 1.8143E+10, 1.8274E+10,
     + 1.8405E+10, 1.8537E+10, 1.8671E+10, 1.8805E+10, 1.8940E+10,
     +  1.9076E+10, 1.9213E+10, 1.9350E+10, 1.9489E+10, 1.9629E+10,
     + 1.9769E+10, 1.9911E+10, 2.0053E+10, 2.0197E+10, 2.0341E+10,
     +  2.0486E+10, 2.0632E+10, 2.0780E+10, 2.0928E+10, 2.1077E+10,
     + 2.1227E+10, 2.1378E+10, 2.1530E+10, 2.1684E+10, 2.1838E+10,
     +  2.1993E+10, 2.2149E+10, 2.2307E+10, 2.2465E+10, 2.2624E+10,
     + 2.2785E+10, 2.2946E+10, 2.3109E+10, 2.3272E+10, 2.3437E+10,
     +  2.3603E+10, 2.3770E+10, 2.3937E+10, 2.4107E+10, 2.4277E+10,
     + 2.4448E+10, 2.4620E+10, 2.4794E+10, 2.4969E+10, 2.5144E+10,
     +  2.5321E+10, 2.5500E+10, 2.5679E+10, 2.5859E+10, 2.6041E+10,
     + 2.6224E+10, 2.6408E+10, 2.6593E+10, 2.6780E+10, 2.6967E+10,
     +  2.7156E+10, 2.7346E+10, 2.7538E+10, 2.7730E+10, 2.7924E+10,
     + 2.8119E+10, 2.8316E+10, 2.8513E+10, 2.8712E+10, 2.8912E+10,
     +  2.9114E+10, 2.9317E+10, 2.9521E+10, 2.9726E+10, 2.9933E+10,
     + 3.0141E+10, 3.0351E+10, 3.0562E+10, 3.0774E+10, 3.0987E+10,
     +  3.1202E+10, 3.1419E+10, 3.1637E+10, 3.1856E+10, 3.2076E+10,
     + 3.2298E+10, 3.2522E+10, 3.2746E+10, 3.2973E+10, 3.3201E+10,
     +  3.3430E+10, 3.3660E+10, 3.3893E+10, 3.4126E+10, 3.4361E+10,
     + 3.4598E+10, 3.4836E+10, 3.5076E+10, 3.5317E+10, 3.5560E+10,
     +  3.5804E+10, 3.6050E+10, 3.6298E+10, 3.6547E+10, 3.6797E+10,
     + 3.7049E+10, 3.7303E+10, 3.7559E+10, 3.7816E+10, 3.8075E+10,
     +  3.8335E+10, 3.8597E+10, 3.8861E+10, 3.9126E+10, 3.9393E+10,
     + 3.9662E+10, 3.9932E+10, 4.0204E+10, 4.0478E+10, 4.0754E+10,
     +  4.1031E+10, 4.1310E+10, 4.1591E+10, 4.1874E+10, 4.2158E+10,
     + 4.2445E+10, 4.2733E+10, 4.3023E+10, 4.3314E+10, 4.3608E+10,
     +  4.3903E+10, 4.4201E+10, 4.4500E+10, 4.4801E+10, 4.5104E+10,
     + 4.5408E+10, 4.5715E+10, 4.6024E+10, 4.6334E+10, 4.6647E+10,
     +  4.6961E+10, 4.7278E+10, 4.7596E+10, 4.7917E+10, 4.8239E+10,
     + 4.8564E+10, 4.8890E+10, 4.9219E+10, 4.9549E+10, 4.9882E+10,
     +  5.0217E+10, 5.0554E+10, 5.0893E+10, 5.1234E+10, 5.1577E+10,
     + 5.1922E+10, 5.2269E+10, 5.2619E+10, 5.2971E+10, 5.3325E+10,
     +  5.3681E+10, 5.4039E+10, 5.4400E+10, 5.4763E+10, 5.5128E+10,
     + 5.5495E+10, 5.5865E+10, 5.6237E+10, 5.6611E+10, 5.6988E+10,
     +  5.7367E+10, 5.7748E+10, 5.8131E+10, 5.8517E+10, 5.8906E+10,
     + 5.9296E+10, 5.9690E+10, 6.0085E+10, 6.0483E+10, 6.0884E+10,
     +  6.1286E+10, 6.1692E+10, 6.2100E+10, 6.2510E+10, 6.2923E+10,
     + 6.3338E+10, 6.3756E+10, 6.4177E+10, 6.4600E+10, 6.5026E+10,
     +  6.5454E+10, 6.5885E+10, 6.6319E+10, 6.6755E+10, 6.7194E+10,
     + 6.7636E+10, 6.8080E+10, 6.8527E+10, 6.8977E+10, 6.9429E+10,
     +  6.9884E+10, 7.0342E+10, 7.0803E+10, 7.1267E+10, 7.1733E+10,
     + 7.2203E+10, 7.2675E+10, 7.3150E+10, 7.3628E+10, 7.4109E+10,
     +  7.4592E+10, 7.5079E+10, 7.5568E+10, 7.6061E+10, 7.6557E+10,
     + 7.7055E+10, 7.7557E+10, 7.8061E+10, 7.8569E+10, 7.9080E+10,
     +  7.9594E+10, 8.0111E+10, 8.0631E+10, 8.1154E+10, 8.1680E+10,
     + 8.2210E+10, 8.2742E+10, 8.3278E+10, 8.3817E+10, 8.4360E+10,
     +  8.4905E+10, 8.5454E+10, 8.6007E+10, 8.6562E+10, 8.7121E+10,
     + 8.7683E+10, 8.8249E+10, 8.8818E+10, 8.9390E+10, 8.9966E+10,
     +  9.0545E+10, 9.1128E+10, 9.1714E+10, 9.2304E+10, 9.2897E+10,
     + 9.3494E+10, 9.4094E+10, 9.4698E+10, 9.5306E+10, 9.5917E+10,
     +  9.6532E+10, 9.7151E+10, 9.7773E+10, 9.8399E+10, 9.9028E+10,
     + 9.9662E+10, 1.0030E+11, 1.0094E+11, 1.0158E+11, 1.0223E+11,
     +  1.0289E+11, 1.0354E+11, 1.0420E+11, 1.0487E+11, 1.0553E+11,
     + 1.0621E+11, 1.0688E+11, 1.0756E+11, 1.0825E+11, 1.0893E+11,
     +  1.0963E+11, 1.1032E+11, 1.1102E+11, 1.1173E+11, 1.1243E+11,
     + 1.1315E+11, 1.1386E+11, 1.1459E+11, 1.1531E+11, 1.1604E+11,
     +  1.1677E+11, 1.1751E+11, 1.1825E+11, 1.1900E+11, 1.1975E+11,
     + 1.2051E+11, 1.2127E+11, 1.2203E+11, 1.2280E+11, 1.2358E+11,
     +  1.2435E+11, 1.2514E+11, 1.2592E+11, 1.2672E+11, 1.2751E+11,
     + 1.2831E+11, 1.2912E+11, 1.2993E+11, 1.3075E+11, 1.3157E+11,
     +  1.3239E+11, 1.3322E+11, 1.3405E+11, 1.3489E+11, 1.3574E+11,
     + 1.3659E+11, 1.3744E+11, 1.3830E+11, 1.3916E+11, 1.4003E+11,
     +  1.4091E+11, 1.4179E+11, 1.4267E+11, 1.4356E+11, 1.4445E+11,
     + 1.4535E+11, 1.4626E+11, 1.4717E+11, 1.4808E+11, 1.4900E+11,
     +  1.4993E+11, 1.5086E+11, 1.5180E+11, 1.5274E+11, 1.5369E+11,
     + 1.5464E+11, 1.5560E+11, 1.5656E+11, 1.5753E+11, 1.5851E+11,
     +  1.5949E+11, 1.6047E+11, 1.6147E+11, 1.6246E+11, 1.6347E+11,
     + 1.6448E+11, 1.6549E+11, 1.6651E+11, 1.6754E+11, 1.6857E+11,
     +  1.6961E+11, 1.7065E+11, 1.7170E+11, 1.7276E+11, 1.7382E+11,
     + 1.7489E+11, 1.7597E+11, 1.7705E+11, 1.7813E+11, 1.7923E+11,
     +  1.8033E+11, 1.8143E+11, 1.8254E+11, 1.8366E+11, 1.8479E+11,
     + 1.8592E+11, 1.8705E+11, 1.8820E+11, 1.8935E+11, 1.9051E+11,
     +  1.9167E+11, 1.9284E+11, 1.9402E+11, 1.9520E+11, 1.9639E+11,
     + 1.9759E+11, 1.9879E+11, 2.0000E+11, 2.0122E+11, 2.0244E+11,
     +  2.0367E+11, 2.0491E+11, 2.0615E+11, 2.0741E+11, 2.0867E+11,
     + 2.0993E+11, 2.1121E+11, 2.1249E+11, 2.1377E+11, 2.1507E+11,
     +  2.1637E+11, 2.1768E+11, 2.1900E+11, 2.2032E+11, 2.2165E+11,
     + 2.2299E+11, 2.2434E+11, 2.2569E+11, 2.2706E+11, 2.2842E+11,
     +  2.2980E+11, 2.3119E+11, 2.3258E+11, 2.3398E+11, 2.3539E+11,
     + 2.3680E+11, 2.3823E+11, 2.3966E+11, 2.4110E+11, 2.4255E+11,
     +  2.4400E+11, 2.4547E+11, 2.4694E+11, 2.4842E+11, 2.4991E+11,
     + 2.5141E+11, 2.5291E+11, 2.5443E+11, 2.5595E+11, 2.5748E+11,
     +  2.5902E+11, 2.6057E+11, 2.6212E+11, 2.6369E+11, 2.6526E+11,
     + 2.6685E+11, 2.6844E+11, 2.7004E+11, 2.7165E+11, 2.7326E+11,
     +  2.7489E+11, 2.7653E+11, 2.7817E+11, 2.7982E+11, 2.8149E+11,
     + 2.8316E+11, 2.8484E+11, 2.8653E+11, 2.8823E+11, 2.8994E+11,
     +  2.9166E+11, 2.9339E+11, 2.9513E+11, 2.9687E+11, 2.9863E+11,
     + 3.0040E+11, 3.0217E+11, 3.0396E+11, 3.0576E+11, 3.0756E+11,
     +  3.0938E+11, 3.1120E+11, 3.1304E+11, 3.1488E+11, 3.1674E+11,
     + 3.1860E+11, 3.2048E+11, 3.2237E+11, 3.2426E+11, 3.2617E+11,
     +  3.2809E+11, 3.3002E+11, 3.3196E+11, 3.3390E+11, 3.3586E+11,
     + 3.3783E+11, 3.3982E+11, 3.4181E+11, 3.4381E+11, 3.4582E+11,
     +  3.4785E+11, 3.4988E+11, 3.5193E+11, 3.5399E+11, 3.5606E+11,
     + 3.5814E+11, 3.6023E+11, 3.6233E+11, 3.6445E+11, 3.6657E+11,
     +  3.6871E+11, 3.7086E+11, 3.7302E+11, 3.7519E+11, 3.7737E+11,
     + 3.7957E+11, 3.8178E+11, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --      2134
      data (QofT( 3,J),J=1,2408)/ 4.6408E+01, 1.0896E+02, 1.8006E+02,
     +  2.5788E+02, 3.4178E+02, 4.3132E+02, 5.2615E+02, 6.2593E+02,
     + 7.3045E+02, 8.3951E+02, 9.5301E+02, 1.0708E+03, 1.1930E+03,
     +  1.3194E+03, 1.4500E+03, 1.5848E+03, 1.7238E+03, 1.8669E+03,
     + 2.0141E+03, 2.1653E+03, 2.4797E+03, 2.8096E+03, 3.1545E+03,
     +  3.5140E+03, 3.8875E+03, 4.2745E+03, 4.6746E+03, 5.0873E+03,
     + 5.5123E+03, 5.9491E+03, 6.3974E+03, 6.8568E+03, 7.3272E+03,
     +  7.8081E+03, 8.2995E+03, 8.8011E+03, 9.3126E+03, 9.8340E+03,
     + 1.0365E+04, 1.0906E+04, 1.1456E+04, 1.2016E+04, 1.2585E+04,
     +  1.3163E+04, 1.3751E+04, 1.4348E+04, 1.4954E+04, 1.5570E+04,
     + 1.6195E+04, 1.6829E+04, 1.7473E+04, 1.8127E+04, 1.8791E+04,
     +  1.9464E+04, 2.0147E+04, 2.0840E+04, 2.1543E+04, 2.2257E+04,
     + 2.2981E+04, 2.3716E+04, 2.4461E+04, 2.5217E+04, 2.5984E+04,
     +  2.6763E+04, 2.7552E+04, 2.8354E+04, 2.9167E+04, 2.9991E+04,
     + 3.0828E+04, 3.1677E+04, 3.2539E+04, 3.3413E+04, 3.4300E+04,
     +  3.5200E+04, 3.6113E+04, 3.7039E+04, 3.7979E+04, 3.8933E+04,
     + 3.9901E+04, 4.0883E+04, 4.1879E+04, 4.2889E+04, 4.3915E+04,
     +  4.4955E+04, 4.6011E+04, 4.7082E+04, 4.8168E+04, 4.9270E+04,
     + 5.0389E+04, 5.1523E+04, 5.2674E+04, 5.3841E+04, 5.5026E+04,
     +  5.6227E+04, 5.7446E+04, 5.8682E+04, 5.9936E+04, 6.1208E+04,
     + 6.2498E+04, 6.3807E+04, 6.5134E+04, 6.6481E+04, 6.7846E+04,
     +  6.9231E+04, 7.0636E+04, 7.2060E+04, 7.3505E+04, 7.4970E+04,
     + 7.6455E+04, 7.7962E+04, 7.9490E+04, 8.1039E+04, 8.2610E+04,
     +  8.4203E+04, 8.5818E+04, 8.7456E+04, 8.9117E+04, 9.0801E+04,
     + 9.2508E+04, 9.4239E+04, 9.5994E+04, 9.7774E+04, 9.9578E+04,
     +  1.0141E+05, 1.0326E+05, 1.0514E+05, 1.0705E+05, 1.0898E+05,
     + 1.1094E+05, 1.1292E+05, 1.1494E+05, 1.1698E+05, 1.1905E+05,
     +  1.2114E+05, 1.2327E+05, 1.2542E+05, 1.2761E+05, 1.2982E+05,
     + 1.3206E+05, 1.3434E+05, 1.3664E+05, 1.3898E+05, 1.4135E+05,
     +  1.4375E+05, 1.4618E+05, 1.4864E+05, 1.5114E+05, 1.5368E+05,
     + 1.5624E+05, 1.5884E+05, 1.6148E+05, 1.6415E+05, 1.6686E+05,
     +  1.6961E+05, 1.7239E+05, 1.7521E+05, 1.7806E+05, 1.8096E+05,
     + 1.8389E+05, 1.8686E+05, 1.8988E+05, 1.9293E+05, 1.9602E+05,
     +  1.9916E+05, 2.0234E+05, 2.0556E+05, 2.0882E+05, 2.1212E+05,
     + 2.1547E+05, 2.1887E+05, 2.2231E+05, 2.2579E+05, 2.2933E+05,
     +  2.3290E+05, 2.3653E+05, 2.4020E+05, 2.4393E+05, 2.4770E+05,
     + 2.5152E+05, 2.5539E+05, 2.5932E+05, 2.6329E+05, 2.6732E+05,
     +  2.7140E+05, 2.7554E+05, 2.7973E+05, 2.8397E+05, 2.8827E+05,
     + 2.9263E+05, 2.9705E+05, 3.0152E+05, 3.0605E+05, 3.1064E+05,
     +  3.1529E+05, 3.2001E+05, 3.2478E+05, 3.2962E+05, 3.3452E+05,
     + 3.3948E+05, 3.4451E+05, 3.4960E+05, 3.5476E+05, 3.5999E+05,
     +  3.6529E+05, 3.7066E+05, 3.7609E+05, 3.8160E+05, 3.8718E+05,
     + 3.9283E+05, 3.9855E+05, 4.0435E+05, 4.1022E+05, 4.1617E+05,
     +  4.2220E+05, 4.2831E+05, 4.3449E+05, 4.4075E+05, 4.4710E+05,
     + 4.5353E+05, 4.6004E+05, 4.6663E+05, 4.7331E+05, 4.8008E+05,
     +  4.8693E+05, 4.9387E+05, 5.0090E+05, 5.0803E+05, 5.1524E+05,
     + 5.2254E+05, 5.2994E+05, 5.3744E+05, 5.4503E+05, 5.5272E+05,
     +  5.6050E+05, 5.6839E+05, 5.7637E+05, 5.8446E+05, 5.9266E+05,
     + 6.0095E+05, 6.0936E+05, 6.1787E+05, 6.2648E+05, 6.3521E+05,
     +  6.4405E+05, 6.5300E+05, 6.6207E+05, 6.7125E+05, 6.8055E+05,
     + 6.8996E+05, 6.9950E+05, 7.0915E+05, 7.1893E+05, 7.2883E+05,
     +  7.3886E+05, 7.4901E+05, 7.5929E+05, 7.6971E+05, 7.8025E+05,
     + 7.9092E+05, 8.0173E+05, 8.1268E+05, 8.2376E+05, 8.3499E+05,
     +  8.4635E+05, 8.5786E+05, 8.6951E+05, 8.8131E+05, 8.9325E+05,
     + 9.0535E+05, 9.1760E+05, 9.2999E+05, 9.4255E+05, 9.5526E+05,
     +  9.6813E+05, 9.8116E+05, 9.9435E+05, 1.0077E+06, 1.0212E+06,
     + 1.0349E+06, 1.0488E+06, 1.0628E+06, 1.0770E+06, 1.0914E+06,
     +  1.1060E+06, 1.1207E+06, 1.1356E+06, 1.1508E+06, 1.1660E+06,
     + 1.1815E+06, 1.1972E+06, 1.2131E+06, 1.2291E+06, 1.2454E+06,
     +  1.2619E+06, 1.2785E+06, 1.2954E+06, 1.3125E+06, 1.3298E+06,
     + 1.3473E+06, 1.3650E+06, 1.3829E+06, 1.4011E+06, 1.4195E+06,
     +  1.4381E+06, 1.4569E+06, 1.4759E+06, 1.4952E+06, 1.5147E+06,
     + 1.5345E+06, 1.5545E+06, 1.5748E+06, 1.5952E+06, 1.6160E+06,
     +  1.6370E+06, 1.6582E+06, 1.6797E+06, 1.7015E+06, 1.7235E+06,
     + 1.7458E+06, 1.7684E+06, 1.7912E+06, 1.8143E+06, 1.8377E+06,
     +  1.8614E+06, 1.8854E+06, 1.9096E+06, 1.9342E+06, 1.9590E+06,
     + 1.9841E+06, 2.0096E+06, 2.0353E+06, 2.0613E+06, 2.0877E+06,
     +  2.1144E+06, 2.1414E+06, 2.1687E+06, 2.1963E+06, 2.2243E+06,
     + 2.2526E+06, 2.2813E+06, 2.3103E+06, 2.3396E+06, 2.3693E+06,
     +  2.3993E+06, 2.4297E+06, 2.4605E+06, 2.4916E+06, 2.5231E+06,
     + 2.5549E+06, 2.5872E+06, 2.6198E+06, 2.6528E+06, 2.6862E+06,
     +  2.7200E+06, 2.7542E+06, 2.7888E+06, 2.8238E+06, 2.8592E+06,
     + 2.8950E+06, 2.9313E+06, 2.9679E+06, 3.0051E+06, 3.0426E+06,
     +  3.0806E+06, 3.1190E+06, 3.1579E+06, 3.1972E+06, 3.2370E+06,
     + 3.2773E+06, 3.3180E+06, 3.3592E+06, 3.4009E+06, 3.4431E+06,
     +  3.4857E+06, 3.5289E+06, 3.5725E+06, 3.6167E+06, 3.6614E+06,
     + 3.7066E+06, 3.7523E+06, 3.7986E+06, 3.8454E+06, 3.8927E+06,
     +  3.9406E+06, 3.9890E+06, 4.0380E+06, 4.0876E+06, 4.1377E+06,
     + 4.1884E+06, 4.2397E+06, 4.2916E+06, 4.3440E+06, 4.3971E+06,
     +  4.4508E+06, 4.5051E+06, 4.5600E+06, 4.6156E+06, 4.6718E+06,
     + 4.7286E+06, 4.7861E+06, 4.8443E+06, 4.9031E+06, 4.9626E+06,
     +  5.0227E+06, 5.0836E+06, 5.1451E+06, 5.2074E+06, 5.2703E+06,
     + 5.3340E+06, 5.3984E+06, 5.4635E+06, 5.5294E+06, 5.5960E+06,
     +  5.6633E+06, 5.7314E+06, 5.8003E+06, 5.8700E+06, 5.9405E+06,
     + 6.0117E+06, 6.0838E+06, 6.1567E+06, 6.2303E+06, 6.3049E+06,
     +  6.3802E+06, 6.4564E+06, 6.5335E+06, 6.6114E+06, 6.6902E+06,
     + 6.7699E+06, 6.8505E+06, 6.9320E+06, 7.0144E+06, 7.0977E+06,
     +  7.1820E+06, 7.2672E+06, 7.3533E+06, 7.4404E+06, 7.5285E+06,
     + 7.6175E+06, 7.7076E+06, 7.7986E+06, 7.8907E+06, 7.9837E+06,
     +  8.0779E+06, 8.1730E+06, 8.2692E+06, 8.3665E+06, 8.4649E+06,
     + 8.5643E+06, 8.6648E+06, 8.7665E+06, 8.8693E+06, 8.9732E+06,
     +  9.0782E+06, 9.1844E+06, 9.2918E+06, 9.4004E+06, 9.5101E+06,
     + 9.6211E+06, 9.7333E+06, 9.8467E+06, 9.9613E+06, 1.0077E+07,
     +  1.0194E+07, 1.0313E+07, 1.0433E+07, 1.0554E+07, 1.0676E+07,
     + 1.0800E+07, 1.0925E+07, 1.1051E+07, 1.1179E+07, 1.1308E+07,
     +  1.1439E+07, 1.1571E+07, 1.1704E+07, 1.1839E+07, 1.1976E+07,
     + 1.2114E+07, 1.2253E+07, 1.2394E+07, 1.2536E+07, 1.2680E+07,
     +  1.2826E+07, 1.2973E+07, 1.3121E+07, 1.3271E+07, 1.3423E+07,
     + 1.3577E+07, 1.3732E+07, 1.3889E+07, 1.4047E+07, 1.4207E+07,
     +  1.4369E+07, 1.4533E+07, 1.4698E+07, 1.4865E+07, 1.5034E+07,
     + 1.5205E+07, 1.5378E+07, 1.5552E+07, 1.5728E+07, 1.5907E+07,
     +  1.6087E+07, 1.6269E+07, 1.6453E+07, 1.6638E+07, 1.6826E+07,
     + 1.7016E+07, 1.7208E+07, 1.7402E+07, 1.7598E+07, 1.7796E+07,
     +  1.7996E+07, 1.8198E+07, 1.8403E+07, 1.8609E+07, 1.8818E+07,
     + 1.9029E+07, 1.9242E+07, 1.9457E+07, 1.9675E+07, 1.9895E+07,
     +  2.0117E+07, 2.0342E+07, 2.0569E+07, 2.0798E+07, 2.1030E+07,
     + 2.1264E+07, 2.1501E+07, 2.1740E+07, 2.1981E+07, 2.2225E+07,
     +  2.2472E+07, 2.2721E+07, 2.2973E+07, 2.3228E+07, 2.3485E+07,
     + 2.3744E+07, 2.4007E+07, 2.4272E+07, 2.4540E+07, 2.4811E+07,
     +  2.5084E+07, 2.5361E+07, 2.5640E+07, 2.5922E+07, 2.6207E+07,
     + 2.6495E+07, 2.6786E+07, 2.7080E+07, 2.7378E+07, 2.7678E+07,
     +  2.7981E+07, 2.8287E+07, 2.8597E+07, 2.8909E+07, 2.9225E+07,
     + 2.9544E+07, 2.9867E+07, 3.0192E+07, 3.0521E+07, 3.0854E+07,
     +  3.1190E+07, 3.1529E+07, 3.1872E+07, 3.2218E+07, 3.2568E+07,
     + 3.2921E+07, 3.3278E+07, 3.3638E+07, 3.4003E+07, 3.4371E+07,
     +  3.4742E+07, 3.5118E+07, 3.5497E+07, 3.5880E+07, 3.6267E+07,
     + 3.6658E+07, 3.7053E+07, 3.7452E+07, 3.7855E+07, 3.8262E+07,
     +  3.8673E+07, 3.9088E+07, 3.9508E+07, 3.9931E+07, 4.0359E+07,
     + 4.0791E+07, 4.1228E+07, 4.1669E+07, 4.2114E+07, 4.2564E+07,
     +  4.3019E+07, 4.3478E+07, 4.3941E+07, 4.4410E+07, 4.4882E+07,
     + 4.5360E+07, 4.5842E+07, 4.6330E+07, 4.6822E+07, 4.7319E+07,
     +  4.7821E+07, 4.8328E+07, 4.8840E+07, 4.9357E+07, 4.9879E+07,
     + 5.0407E+07, 5.0939E+07, 5.1477E+07, 5.2020E+07, 5.2569E+07,
     +  5.3123E+07, 5.3683E+07, 5.4248E+07, 5.4819E+07, 5.5395E+07,
     + 5.5977E+07, 5.6565E+07, 5.7159E+07, 5.7758E+07, 5.8363E+07,
     +  5.8975E+07, 5.9592E+07, 6.0215E+07, 6.0845E+07, 6.1481E+07,
     + 6.2122E+07, 6.2771E+07, 6.3425E+07, 6.4086E+07, 6.4754E+07,
     +  6.5428E+07, 6.6108E+07, 6.6795E+07, 6.7489E+07, 6.8190E+07,
     + 6.8897E+07, 6.9612E+07, 7.0333E+07, 7.1061E+07, 7.1797E+07,
     +  7.2539E+07, 7.3289E+07, 7.4046E+07, 7.4811E+07, 7.5582E+07,
     + 7.6362E+07, 7.7149E+07, 7.7943E+07, 7.8745E+07, 7.9555E+07,
     +  8.0373E+07, 8.1198E+07, 8.2032E+07, 8.2873E+07, 8.3723E+07,
     + 8.4581E+07, 8.5447E+07, 8.6321E+07, 8.7204E+07, 8.8095E+07,
     +  8.8995E+07, 8.9903E+07, 9.0820E+07, 9.1746E+07, 9.2681E+07,
     + 9.3625E+07, 9.4577E+07, 9.5539E+07, 9.6510E+07, 9.7490E+07,
     +  9.8480E+07, 9.9479E+07, 1.0049E+08, 1.0151E+08, 1.0253E+08,
     + 1.0357E+08, 1.0462E+08, 1.0568E+08, 1.0674E+08, 1.0782E+08,
     +  1.0891E+08, 1.1001E+08, 1.1111E+08, 1.1223E+08, 1.1336E+08,
     + 1.1450E+08, 1.1565E+08, 1.1681E+08, 1.1799E+08, 1.1917E+08,
     +  1.2036E+08, 1.2157E+08, 1.2279E+08, 1.2402E+08, 1.2526E+08,
     + 1.2651E+08, 1.2777E+08, 1.2905E+08, 1.3033E+08, 1.3163E+08,
     +  1.3295E+08, 1.3427E+08, 1.3561E+08, 1.3695E+08, 1.3832E+08,
     + 1.3969E+08, 1.4108E+08, 1.4248E+08, 1.4389E+08, 1.4531E+08,
     +  1.4675E+08, 1.4820E+08, 1.4967E+08, 1.5115E+08, 1.5264E+08,
     + 1.5415E+08, 1.5567E+08, 1.5720E+08, 1.5875E+08, 1.6032E+08,
     +  1.6189E+08, 1.6349E+08, 1.6509E+08, 1.6672E+08, 1.6835E+08,
     + 1.7000E+08, 1.7167E+08, 1.7335E+08, 1.7505E+08, 1.7676E+08,
     +  1.7849E+08, 1.8024E+08, 1.8200E+08, 1.8378E+08, 1.8557E+08,
     + 1.8738E+08, 1.8921E+08, 1.9105E+08, 1.9291E+08, 1.9478E+08,
     +  1.9668E+08, 1.9859E+08, 2.0052E+08, 2.0246E+08, 2.0443E+08,
     + 2.0641E+08, 2.0841E+08, 2.1042E+08, 2.1246E+08, 2.1451E+08,
     +  2.1659E+08, 2.1868E+08, 2.2079E+08, 2.2292E+08, 2.2507E+08,
     + 2.2724E+08, 2.2942E+08, 2.3163E+08, 2.3386E+08, 2.3611E+08,
     +  2.3837E+08, 2.4066E+08, 2.4297E+08, 2.4530E+08, 2.4765E+08,
     + 2.5002E+08, 2.5241E+08, 2.5483E+08, 2.5726E+08, 2.5972E+08,
     +  2.6220E+08, 2.6470E+08, 2.6723E+08, 2.6977E+08, 2.7234E+08,
     + 2.7494E+08, 2.7755E+08, 2.8019E+08, 2.8285E+08, 2.8554E+08,
     +  2.8825E+08, 2.9098E+08, 2.9374E+08, 2.9652E+08, 2.9933E+08,
     + 3.0216E+08, 3.0502E+08, 3.0790E+08, 3.1081E+08, 3.1374E+08,
     +  3.1670E+08, 3.1969E+08, 3.2270E+08, 3.2574E+08, 3.2880E+08,
     + 3.3189E+08, 3.3501E+08, 3.3816E+08, 3.4133E+08, 3.4454E+08,
     +  3.4777E+08, 3.5103E+08, 3.5431E+08, 3.5763E+08, 3.6097E+08,
     + 3.6435E+08, 3.6775E+08, 3.7119E+08, 3.7465E+08, 3.7815E+08,
     +  3.8167E+08, 3.8523E+08, 3.8881E+08, 3.9243E+08, 3.9608E+08,
     + 3.9976E+08, 4.0347E+08, 4.0722E+08, 4.1100E+08, 4.1481E+08,
     +  4.1865E+08, 4.2253E+08, 4.2644E+08, 4.3038E+08, 4.3436E+08,
     + 4.3837E+08, 4.4242E+08, 4.4650E+08, 4.5062E+08, 4.5477E+08,
     +  4.5896E+08, 4.6319E+08, 4.6745E+08, 4.7174E+08, 4.7608E+08,
     + 4.8045E+08, 4.8486E+08, 4.8931E+08, 4.9380E+08, 4.9832E+08,
     +  5.0288E+08, 5.0748E+08, 5.1213E+08, 5.1681E+08, 5.2153E+08,
     + 5.2629E+08, 5.3109E+08, 5.3594E+08, 5.4082E+08, 5.4575E+08,
     +  5.5071E+08, 5.5572E+08, 5.6078E+08, 5.6587E+08, 5.7101E+08,
     + 5.7620E+08, 5.8142E+08, 5.8669E+08, 5.9201E+08, 5.9737E+08,
     +  6.0278E+08, 6.0823E+08, 6.1373E+08, 6.1927E+08, 6.2486E+08,
     + 6.3050E+08, 6.3619E+08, 6.4192E+08, 6.4770E+08, 6.5354E+08,
     +  6.5942E+08, 6.6535E+08, 6.7132E+08, 6.7735E+08, 6.8343E+08,
     + 6.8957E+08, 6.9575E+08, 7.0198E+08, 7.0827E+08, 7.1461E+08,
     +  7.2100E+08, 7.2745E+08, 7.3395E+08, 7.4050E+08, 7.4711E+08,
     + 7.5377E+08, 7.6049E+08, 7.6727E+08, 7.7410E+08, 7.8099E+08,
     +  7.8793E+08, 7.9494E+08, 8.0200E+08, 8.0912E+08, 8.1630E+08,
     + 8.2354E+08, 8.3084E+08, 8.3820E+08, 8.4562E+08, 8.5310E+08,
     +  8.6064E+08, 8.6825E+08, 8.7592E+08, 8.8365E+08, 8.9145E+08,
     + 8.9931E+08, 9.0723E+08, 9.1522E+08, 9.2328E+08, 9.3140E+08,
     +  9.3959E+08, 9.4784E+08, 9.5617E+08, 9.6456E+08, 9.7302E+08,
     + 9.8155E+08, 9.9015E+08, 9.9882E+08, 1.0076E+09, 1.0164E+09,
     +  1.0253E+09, 1.0342E+09, 1.0432E+09, 1.0524E+09, 1.0615E+09,
     + 1.0708E+09, 1.0801E+09, 1.0895E+09, 1.0990E+09, 1.1086E+09,
     +  1.1182E+09, 1.1279E+09, 1.1377E+09, 1.1476E+09, 1.1575E+09,
     + 1.1676E+09, 1.1777E+09, 1.1879E+09, 1.1981E+09, 1.2085E+09,
     +  1.2189E+09, 1.2295E+09, 1.2401E+09, 1.2508E+09, 1.2616E+09,
     + 1.2724E+09, 1.2834E+09, 1.2945E+09, 1.3056E+09, 1.3168E+09,
     +  1.3281E+09, 1.3395E+09, 1.3510E+09, 1.3626E+09, 1.3743E+09,
     + 1.3861E+09, 1.3980E+09, 1.4099E+09, 1.4220E+09, 1.4341E+09,
     +  1.4464E+09, 1.4588E+09, 1.4712E+09, 1.4838E+09, 1.4964E+09,
     + 1.5092E+09, 1.5220E+09, 1.5350E+09, 1.5480E+09, 1.5612E+09,
     +  1.5745E+09, 1.5878E+09, 1.6013E+09, 1.6149E+09, 1.6286E+09,
     + 1.6424E+09, 1.6563E+09, 1.6703E+09, 1.6844E+09, 1.6987E+09,
     +  1.7130E+09, 1.7275E+09, 1.7421E+09, 1.7568E+09, 1.7716E+09,
     + 1.7865E+09, 1.8016E+09, 1.8167E+09, 1.8320E+09, 1.8474E+09,
     +  1.8629E+09, 1.8786E+09, 1.8944E+09, 1.9102E+09, 1.9263E+09,
     + 1.9424E+09, 1.9587E+09, 1.9751E+09, 1.9916E+09, 2.0082E+09,
     +  2.0250E+09, 2.0419E+09, 2.0590E+09, 2.0762E+09, 2.0935E+09,
     + 2.1109E+09, 2.1285E+09, 2.1462E+09, 2.1641E+09, 2.1821E+09,
     +  2.2002E+09, 2.2185E+09, 2.2369E+09, 2.2555E+09, 2.2742E+09,
     + 2.2930E+09, 2.3120E+09, 2.3311E+09, 2.3504E+09, 2.3699E+09,
     +  2.3894E+09, 2.4092E+09, 2.4291E+09, 2.4491E+09, 2.4693E+09,
     + 2.4897E+09, 2.5102E+09, 2.5308E+09, 2.5517E+09, 2.5726E+09,
     +  2.5938E+09, 2.6151E+09, 2.6366E+09, 2.6582E+09, 2.6800E+09,
     + 2.7020E+09, 2.7241E+09, 2.7464E+09, 2.7689E+09, 2.7915E+09,
     +  2.8144E+09, 2.8374E+09, 2.8605E+09, 2.8839E+09, 2.9074E+09,
     + 2.9311E+09, 2.9550E+09, 2.9790E+09, 3.0033E+09, 3.0277E+09,
     +  3.0523E+09, 3.0771E+09, 3.1021E+09, 3.1273E+09, 3.1527E+09,
     + 3.1782E+09, 3.2040E+09, 3.2299E+09, 3.2561E+09, 3.2824E+09,
     +  3.3090E+09, 3.3357E+09, 3.3627E+09, 3.3898E+09, 3.4172E+09,
     + 3.4447E+09, 3.4725E+09, 3.5005E+09, 3.5286E+09, 3.5570E+09,
     +  3.5856E+09, 3.6145E+09, 3.6435E+09, 3.6728E+09, 3.7022E+09,
     + 3.7319E+09, 3.7618E+09, 3.7920E+09, 3.8224E+09, 3.8529E+09,
     +  3.8838E+09, 3.9148E+09, 3.9461E+09, 3.9776E+09, 4.0094E+09,
     + 4.0414E+09, 4.0736E+09, 4.1060E+09, 4.1387E+09, 4.1717E+09,
     +  4.2049E+09, 4.2383E+09, 4.2720E+09, 4.3060E+09, 4.3401E+09,
     + 4.3746E+09, 4.4093E+09, 4.4442E+09, 4.4794E+09, 4.5149E+09,
     +  4.5506E+09, 4.5866E+09, 4.6229E+09, 4.6594E+09, 4.6962E+09,
     + 4.7333E+09, 4.7706E+09, 4.8082E+09, 4.8461E+09, 4.8843E+09,
     +  4.9228E+09, 4.9615E+09, 5.0005E+09, 5.0398E+09, 5.0794E+09,
     + 5.1193E+09, 5.1594E+09, 5.1999E+09, 5.2407E+09, 5.2817E+09,
     +  5.3231E+09, 5.3647E+09, 5.4067E+09, 5.4490E+09, 5.4916E+09,
     + 5.5344E+09, 5.5776E+09, 5.6212E+09, 5.6650E+09, 5.7091E+09,
     +  5.7536E+09, 5.7984E+09, 5.8435E+09, 5.8890E+09, 5.9347E+09,
     + 5.9809E+09, 6.0273E+09, 6.0741E+09, 6.1212E+09, 6.1686E+09,
     +  6.2164E+09, 6.2646E+09, 6.3131E+09, 6.3619E+09, 6.4111E+09,
     + 6.4606E+09, 6.5105E+09, 6.5608E+09, 6.6114E+09, 6.6624E+09,
     +  6.7138E+09, 6.7655E+09, 6.8176E+09, 6.8701E+09, 6.9229E+09,
     + 6.9761E+09, 7.0297E+09, 7.0837E+09, 7.1381E+09, 7.1928E+09,
     +  7.2480E+09, 7.3035E+09, 7.3595E+09, 7.4158E+09, 7.4726E+09,
     + 7.5297E+09, 7.5873E+09, 7.6452E+09, 7.7036E+09, 7.7624E+09,
     +  7.8216E+09, 7.8813E+09, 7.9413E+09, 8.0018E+09, 8.0627E+09,
     + 8.1241E+09, 8.1858E+09, 8.2481E+09, 8.3107E+09, 8.3738E+09,
     +  8.4374E+09, 8.5014E+09, 8.5658E+09, 8.6307E+09, 8.6961E+09,
     + 8.7619E+09, 8.8282E+09, 8.8949E+09, 8.9621E+09, 9.0298E+09,
     +  9.0980E+09, 9.1667E+09, 9.2358E+09, 9.3054E+09, 9.3755E+09,
     + 9.4461E+09, 9.5172E+09, 9.5888E+09, 9.6609E+09, 9.7335E+09,
     +  9.8066E+09, 9.8802E+09, 9.9544E+09, 1.0029E+10, 1.0104E+10,
     + 1.0180E+10, 1.0256E+10, 1.0333E+10, 1.0410E+10, 1.0488E+10,
     +  1.0566E+10, 1.0645E+10, 1.0725E+10, 1.0805E+10, 1.0885E+10,
     + 1.0967E+10, 1.1048E+10, 1.1130E+10, 1.1213E+10, 1.1297E+10,
     +  1.1381E+10, 1.1465E+10, 1.1550E+10, 1.1636E+10, 1.1723E+10,
     + 1.1809E+10, 1.1897E+10, 1.1985E+10, 1.2074E+10, 1.2163E+10,
     +  1.2253E+10, 1.2344E+10, 1.2435E+10, 1.2527E+10, 1.2619E+10,
     + 1.2712E+10, 1.2806E+10, 1.2901E+10, 1.2996E+10, 1.3091E+10,
     +  1.3188E+10, 1.3285E+10, 1.3382E+10, 1.3481E+10, 1.3580E+10,
     + 1.3679E+10, 1.3780E+10, 1.3881E+10, 1.3983E+10, 1.4085E+10,
     +  1.4188E+10, 1.4292E+10, 1.4397E+10, 1.4502E+10, 1.4608E+10,
     + 1.4714E+10, 1.4822E+10, 1.4930E+10, 1.5039E+10, 1.5149E+10,
     +  1.5259E+10, 1.5370E+10, 1.5482E+10, 1.5595E+10, 1.5708E+10,
     + 1.5822E+10, 1.5937E+10, 1.6053E+10, 1.6169E+10, 1.6287E+10,
     +  1.6405E+10, 1.6524E+10, 1.6643E+10, 1.6764E+10, 1.6885E+10,
     + 1.7007E+10, 1.7130E+10, 1.7254E+10, 1.7378E+10, 1.7504E+10,
     +  1.7630E+10, 1.7757E+10, 1.7885E+10, 1.8014E+10, 1.8144E+10,
     + 1.8274E+10, 1.8406E+10, 1.8538E+10, 1.8671E+10, 1.8805E+10,
     +  1.8940E+10, 1.9076E+10, 1.9213E+10, 1.9351E+10, 1.9489E+10,
     + 1.9629E+10, 1.9769E+10, 1.9911E+10, 2.0053E+10, 2.0197E+10,
     +  2.0341E+10, 2.0486E+10, 2.0632E+10, 2.0779E+10, 2.0927E+10,
     + 2.1077E+10, 2.1227E+10, 2.1378E+10, 2.1530E+10, 2.1683E+10,
     +  2.1837E+10, 2.1992E+10, 2.2148E+10, 2.2306E+10, 2.2464E+10,
     + 2.2623E+10, 2.2783E+10, 2.2945E+10, 2.3107E+10, 2.3271E+10,
     +  2.3435E+10, 2.3601E+10, 2.3768E+10, 2.3936E+10, 2.4105E+10,
     + 2.4275E+10, 2.4446E+10, 2.4618E+10, 2.4792E+10, 2.4966E+10,
     +  2.5142E+10, 2.5319E+10, 2.5497E+10, 2.5676E+10, 2.5857E+10,
     + 2.6038E+10, 2.6221E+10, 2.6405E+10, 2.6590E+10, 2.6776E+10,
     +  2.6964E+10, 2.7153E+10, 2.7343E+10, 2.7534E+10, 2.7727E+10,
     + 2.7920E+10, 2.8115E+10, 2.8312E+10, 2.8509E+10, 2.8708E+10,
     +  2.8908E+10, 2.9109E+10, 2.9312E+10, 2.9516E+10, 2.9722E+10,
     + 2.9928E+10, 3.0136E+10, 3.0346E+10, 3.0556E+10, 3.0769E+10,
     +  3.0982E+10, 3.1197E+10, 3.1413E+10, 3.1631E+10, 3.1850E+10,
     + 3.2070E+10, 3.2292E+10, 3.2515E+10, 3.2740E+10, 3.2966E+10,
     +  3.3194E+10, 3.3423E+10, 3.3653E+10, 3.3885E+10, 3.4119E+10,
     + 3.4354E+10, 3.4590E+10, 3.4828E+10, 3.5068E+10, 3.5309E+10,
     +  3.5552E+10, 3.5796E+10, 3.6041E+10, 3.6289E+10, 3.6538E+10,
     + 3.6788E+10, 3.7040E+10, 3.7294E+10, 3.7549E+10, 3.7806E+10,
     +  3.8065E+10, 3.8325E+10, 3.8587E+10, 3.8850E+10, 3.9115E+10,
     + 3.9382E+10, 3.9651E+10, 3.9921E+10, 4.0193E+10, 4.0467E+10,
     +  4.0742E+10, 4.1019E+10, 4.1298E+10, 4.1579E+10, 4.1861E+10,
     + 4.2146E+10, 4.2432E+10, 4.2720E+10, 4.3009E+10, 4.3301E+10,
     +  4.3594E+10, 4.3889E+10, 4.4186E+10, 4.4485E+10, 4.4786E+10,
     + 4.5089E+10, 4.5393E+10, 4.5700E+10, 4.6008E+10, 4.6319E+10,
     +  4.6631E+10, 4.6945E+10, 4.7261E+10, 4.7580E+10, 4.7900E+10,
     + 4.8222E+10, 4.8546E+10, 4.8872E+10, 4.9201E+10, 4.9531E+10,
     +  4.9863E+10, 5.0198E+10, 5.0535E+10, 5.0873E+10, 5.1214E+10,
     + 5.1557E+10, 5.1902E+10, 5.2249E+10, 5.2598E+10, 5.2950E+10,
     +  5.3304E+10, 5.3659E+10, 5.4018E+10, 5.4378E+10, 5.4740E+10,
     + 5.5105E+10, 5.5472E+10, 5.5842E+10, 5.6213E+10, 5.6587E+10,
     +  5.6963E+10, 5.7342E+10, 5.7723E+10, 5.8106E+10, 5.8492E+10,
     + 5.8880E+10, 5.9270E+10, 5.9663E+10, 6.0058E+10, 6.0456E+10,
     +  6.0856E+10, 6.1259E+10, 6.1664E+10, 6.2071E+10, 6.2481E+10,
     + 6.2894E+10, 6.3309E+10, 6.3726E+10, 6.4147E+10, 6.4569E+10,
     +  6.4995E+10, 6.5423E+10, 6.5853E+10, 6.6287E+10, 6.6722E+10,
     + 6.7161E+10, 6.7602E+10, 6.8046E+10, 6.8493E+10, 6.8942E+10,
     +  6.9394E+10, 6.9849E+10, 7.0307E+10, 7.0767E+10, 7.1230E+10,
     + 7.1696E+10, 7.2165E+10, 7.2637E+10, 7.3111E+10, 7.3589E+10,
     +  7.4069E+10, 7.4552E+10, 7.5039E+10, 7.5528E+10, 7.6020E+10,
     + 7.6515E+10, 7.7013E+10, 7.7514E+10, 7.8018E+10, 7.8526E+10,
     +  7.9036E+10, 7.9549E+10, 8.0066E+10, 8.0585E+10, 8.1108E+10,
     + 8.1634E+10, 8.2163E+10, 8.2695E+10, 8.3230E+10, 8.3769E+10,
     +  8.4311E+10, 8.4856E+10, 8.5404E+10, 8.5956E+10, 8.6511E+10,
     + 8.7069E+10, 8.7631E+10, 8.8196E+10, 8.8764E+10, 8.9336E+10,
     +  8.9911E+10, 9.0490E+10, 9.1072E+10, 9.1658E+10, 9.2247E+10,
     + 9.2840E+10, 9.3436E+10, 9.4036E+10, 9.4639E+10, 9.5246E+10,
     +  9.5857E+10, 9.6471E+10, 9.7089E+10, 9.7710E+10, 9.8335E+10,
     + 9.8964E+10, 9.9597E+10, 1.0023E+11, 1.0087E+11, 1.0152E+11,
     +  1.0217E+11, 1.0282E+11, 1.0347E+11, 1.0413E+11, 1.0480E+11,
     + 1.0546E+11, 1.0613E+11, 1.0681E+11, 1.0749E+11, 1.0817E+11,
     +  1.0886E+11, 1.0955E+11, 1.1025E+11, 1.1095E+11, 1.1165E+11,
     + 1.1236E+11, 1.1307E+11, 1.1378E+11, 1.1450E+11, 1.1523E+11,
     +  1.1596E+11, 1.1669E+11, 1.1743E+11, 1.1817E+11, 1.1892E+11,
     + 1.1967E+11, 1.2042E+11, 1.2118E+11, 1.2194E+11, 1.2271E+11,
     +  1.2349E+11, 1.2426E+11, 1.2504E+11, 1.2583E+11, 1.2662E+11,
     + 1.2742E+11, 1.2822E+11, 1.2902E+11, 1.2983E+11, 1.3065E+11,
     +  1.3147E+11, 1.3229E+11, 1.3312E+11, 1.3395E+11, 1.3479E+11,
     + 1.3563E+11, 1.3648E+11, 1.3733E+11, 1.3819E+11, 1.3906E+11,
     +  1.3992E+11, 1.4080E+11, 1.4167E+11, 1.4256E+11, 1.4345E+11,
     + 1.4434E+11, 1.4524E+11, 1.4614E+11, 1.4705E+11, 1.4796E+11,
     +  1.4888E+11, 1.4981E+11, 1.5074E+11, 1.5167E+11, 1.5262E+11,
     + 1.5356E+11, 1.5451E+11, 1.5547E+11, 1.5643E+11, 1.5740E+11,
     +  1.5838E+11, 1.5936E+11, 1.6034E+11, 1.6133E+11, 1.6233E+11,
     + 1.6333E+11, 1.6434E+11, 1.6535E+11, 1.6637E+11, 1.6740E+11,
     +  1.6843E+11, 1.6946E+11, 1.7051E+11, 1.7156E+11, 1.7261E+11,
     + 1.7367E+11, 1.7474E+11, 1.7581E+11, 1.7689E+11, 1.7798E+11,
     +  1.7907E+11, 1.8017E+11, 1.8127E+11, 1.8238E+11, 1.8350E+11,
     + 1.8462E+11, 1.8575E+11, 1.8689E+11, 1.8803E+11, 1.8918E+11,
     +  1.9033E+11, 1.9150E+11, 1.9266E+11, 1.9384E+11, 1.9502E+11,
     + 1.9621E+11, 1.9741E+11, 1.9861E+11, 1.9982E+11, 2.0103E+11,
     +  2.0225E+11, 2.0348E+11, 2.0472E+11, 2.0596E+11, 2.0721E+11,
     + 2.0847E+11, 2.0974E+11, 2.1101E+11, 2.1229E+11, 2.1357E+11,
     +  2.1487E+11, 2.1617E+11, 2.1747E+11, 2.1879E+11, 2.2011E+11,
     + 2.2144E+11, 2.2278E+11, 2.2412E+11, 2.2548E+11, 2.2684E+11,
     +  2.2820E+11, 2.2958E+11, 2.3096E+11, 2.3235E+11, 2.3375E+11,
     + 2.3516E+11, 2.3657E+11, 2.3799E+11, 2.3942E+11, 2.4086E+11,
     +  2.4231E+11, 2.4376E+11, 2.4523E+11, 2.4670E+11, 2.4817E+11,
     + 2.4966E+11, 2.5116E+11, 2.5266E+11, 2.5417E+11, 2.5569E+11,
     +  2.5722E+11, 2.5876E+11, 2.6030E+11, 2.6186E+11, 2.6342E+11,
     + 2.6499E+11, 2.6657E+11, 2.6816E+11, 2.6976E+11, 2.7137E+11,
     +  2.7298E+11, 2.7461E+11, 2.7624E+11, 2.7788E+11, 2.7953E+11,
     + 2.8119E+11, 2.8286E+11, 2.8454E+11, 2.8623E+11, 2.8793E+11,
     +  2.8964E+11, 2.9135E+11, 2.9308E+11, 2.9481E+11, 2.9656E+11,
     + 2.9831E+11, 3.0008E+11, 3.0185E+11, 3.0363E+11, 3.0543E+11,
     +  3.0723E+11, 3.0904E+11, 3.1087E+11, 3.1270E+11, 3.1454E+11,
     + 3.1639E+11, 3.1826E+11, 3.2013E+11, 3.2201E+11, 3.2391E+11,
     +  3.2581E+11, 3.2773E+11, 3.2965E+11, 3.3159E+11, 3.3353E+11,
     + 3.3549E+11, 3.3746E+11, 3.3944E+11, 3.4143E+11, 3.4343E+11,
     +  3.4544E+11, 3.4746E+11, 3.4949E+11, 3.5153E+11, 3.5359E+11,
     + 3.5565E+11, 3.5773E+11, 3.5982E+11, 3.6192E+11, 3.6403E+11,
     +  3.6615E+11, 3.6829E+11, 3.7043E+11, 3.7259E+11, 3.7476E+11,
     + 3.7694E+11, 3.7913E+11, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 3,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --      3134
      data (QofT( 4,J),J=1,2408)/ 9.5399E+01, 2.2450E+02, 3.7131E+02,
     +  5.3198E+02, 7.0521E+02, 8.9011E+02, 1.0859E+03, 1.2920E+03,
     + 1.5078E+03, 1.7330E+03, 1.9674E+03, 2.2107E+03, 2.4630E+03,
     +  2.7240E+03, 2.9937E+03, 3.2721E+03, 3.5591E+03, 3.8546E+03,
     + 4.1586E+03, 4.4709E+03, 5.1201E+03, 5.8014E+03, 6.5137E+03,
     +  7.2561E+03, 8.0274E+03, 8.8267E+03, 9.6530E+03, 1.0505E+04,
     + 1.1383E+04, 1.2285E+04, 1.3211E+04, 1.4160E+04, 1.5131E+04,
     +  1.6124E+04, 1.7139E+04, 1.8175E+04, 1.9231E+04, 2.0308E+04,
     + 2.1405E+04, 2.2522E+04, 2.3658E+04, 2.4814E+04, 2.5989E+04,
     +  2.7183E+04, 2.8397E+04, 2.9630E+04, 3.0882E+04, 3.2153E+04,
     + 3.3444E+04, 3.4755E+04, 3.6085E+04, 3.7435E+04, 3.8805E+04,
     +  4.0195E+04, 4.1606E+04, 4.3037E+04, 4.4490E+04, 4.5963E+04,
     + 4.7459E+04, 4.8976E+04, 5.0515E+04, 5.2076E+04, 5.3660E+04,
     +  5.5268E+04, 5.6899E+04, 5.8553E+04, 6.0232E+04, 6.1936E+04,
     + 6.3664E+04, 6.5417E+04, 6.7196E+04, 6.9002E+04, 7.0833E+04,
     +  7.2691E+04, 7.4577E+04, 7.6490E+04, 7.8431E+04, 8.0401E+04,
     + 8.2399E+04, 8.4427E+04, 8.6484E+04, 8.8571E+04, 9.0689E+04,
     +  9.2837E+04, 9.5017E+04, 9.7229E+04, 9.9473E+04, 1.0175E+05,
     + 1.0406E+05, 1.0640E+05, 1.0878E+05, 1.1119E+05, 1.1364E+05,
     +  1.1612E+05, 1.1864E+05, 1.2119E+05, 1.2378E+05, 1.2641E+05,
     + 1.2907E+05, 1.3178E+05, 1.3452E+05, 1.3730E+05, 1.4012E+05,
     +  1.4298E+05, 1.4588E+05, 1.4883E+05, 1.5181E+05, 1.5484E+05,
     + 1.5791E+05, 1.6102E+05, 1.6418E+05, 1.6738E+05, 1.7063E+05,
     +  1.7392E+05, 1.7726E+05, 1.8064E+05, 1.8408E+05, 1.8756E+05,
     + 1.9109E+05, 1.9467E+05, 1.9829E+05, 2.0197E+05, 2.0570E+05,
     +  2.0949E+05, 2.1332E+05, 2.1721E+05, 2.2115E+05, 2.2515E+05,
     + 2.2920E+05, 2.3330E+05, 2.3747E+05, 2.4169E+05, 2.4597E+05,
     +  2.5031E+05, 2.5470E+05, 2.5916E+05, 2.6368E+05, 2.6826E+05,
     + 2.7290E+05, 2.7761E+05, 2.8238E+05, 2.8722E+05, 2.9212E+05,
     +  2.9709E+05, 3.0212E+05, 3.0723E+05, 3.1240E+05, 3.1765E+05,
     + 3.2296E+05, 3.2835E+05, 3.3381E+05, 3.3935E+05, 3.4495E+05,
     +  3.5064E+05, 3.5640E+05, 3.6224E+05, 3.6816E+05, 3.7416E+05,
     + 3.8024E+05, 3.8640E+05, 3.9264E+05, 3.9897E+05, 4.0538E+05,
     +  4.1188E+05, 4.1847E+05, 4.2514E+05, 4.3190E+05, 4.3876E+05,
     + 4.4570E+05, 4.5274E+05, 4.5988E+05, 4.6710E+05, 4.7443E+05,
     +  4.8185E+05, 4.8937E+05, 4.9699E+05, 5.0472E+05, 5.1254E+05,
     + 5.2047E+05, 5.2851E+05, 5.3665E+05, 5.4490E+05, 5.5326E+05,
     +  5.6173E+05, 5.7032E+05, 5.7901E+05, 5.8783E+05, 5.9675E+05,
     + 6.0580E+05, 6.1497E+05, 6.2426E+05, 6.3367E+05, 6.4320E+05,
     +  6.5286E+05, 6.6265E+05, 6.7256E+05, 6.8261E+05, 6.9279E+05,
     + 7.0310E+05, 7.1355E+05, 7.2414E+05, 7.3486E+05, 7.4573E+05,
     +  7.5673E+05, 7.6788E+05, 7.7918E+05, 7.9063E+05, 8.0222E+05,
     + 8.1397E+05, 8.2587E+05, 8.3793E+05, 8.5014E+05, 8.6251E+05,
     +  8.7505E+05, 8.8774E+05, 9.0060E+05, 9.1363E+05, 9.2683E+05,
     + 9.4020E+05, 9.5375E+05, 9.6747E+05, 9.8136E+05, 9.9544E+05,
     +  1.0097E+06, 1.0241E+06, 1.0388E+06, 1.0536E+06, 1.0686E+06,
     + 1.0838E+06, 1.0992E+06, 1.1148E+06, 1.1306E+06, 1.1466E+06,
     +  1.1628E+06, 1.1793E+06, 1.1959E+06, 1.2127E+06, 1.2298E+06,
     + 1.2471E+06, 1.2646E+06, 1.2823E+06, 1.3002E+06, 1.3184E+06,
     +  1.3368E+06, 1.3555E+06, 1.3744E+06, 1.3935E+06, 1.4129E+06,
     + 1.4325E+06, 1.4524E+06, 1.4725E+06, 1.4929E+06, 1.5135E+06,
     +  1.5344E+06, 1.5556E+06, 1.5770E+06, 1.5987E+06, 1.6207E+06,
     + 1.6429E+06, 1.6655E+06, 1.6883E+06, 1.7114E+06, 1.7348E+06,
     +  1.7585E+06, 1.7825E+06, 1.8068E+06, 1.8314E+06, 1.8563E+06,
     + 1.8816E+06, 1.9071E+06, 1.9330E+06, 1.9592E+06, 1.9857E+06,
     +  2.0126E+06, 2.0397E+06, 2.0673E+06, 2.0952E+06, 2.1234E+06,
     + 2.1520E+06, 2.1809E+06, 2.2102E+06, 2.2398E+06, 2.2699E+06,
     +  2.3003E+06, 2.3310E+06, 2.3622E+06, 2.3938E+06, 2.4257E+06,
     + 2.4580E+06, 2.4908E+06, 2.5239E+06, 2.5575E+06, 2.5914E+06,
     +  2.6258E+06, 2.6606E+06, 2.6959E+06, 2.7316E+06, 2.7677E+06,
     + 2.8042E+06, 2.8412E+06, 2.8787E+06, 2.9166E+06, 2.9550E+06,
     +  2.9939E+06, 3.0332E+06, 3.0731E+06, 3.1134E+06, 3.1542E+06,
     + 3.1955E+06, 3.2373E+06, 3.2796E+06, 3.3224E+06, 3.3658E+06,
     +  3.4097E+06, 3.4541E+06, 3.4991E+06, 3.5446E+06, 3.5907E+06,
     + 3.6373E+06, 3.6845E+06, 3.7322E+06, 3.7806E+06, 3.8295E+06,
     +  3.8790E+06, 3.9291E+06, 3.9799E+06, 4.0312E+06, 4.0831E+06,
     + 4.1357E+06, 4.1889E+06, 4.2428E+06, 4.2973E+06, 4.3525E+06,
     +  4.4083E+06, 4.4648E+06, 4.5219E+06, 4.5798E+06, 4.6384E+06,
     + 4.6976E+06, 4.7576E+06, 4.8183E+06, 4.8797E+06, 4.9418E+06,
     +  5.0047E+06, 5.0683E+06, 5.1327E+06, 5.1978E+06, 5.2638E+06,
     + 5.3305E+06, 5.3980E+06, 5.4663E+06, 5.5354E+06, 5.6054E+06,
     +  5.6761E+06, 5.7477E+06, 5.8202E+06, 5.8935E+06, 5.9677E+06,
     + 6.0427E+06, 6.1187E+06, 6.1955E+06, 6.2733E+06, 6.3519E+06,
     +  6.4315E+06, 6.5120E+06, 6.5935E+06, 6.6759E+06, 6.7593E+06,
     + 6.8437E+06, 6.9290E+06, 7.0154E+06, 7.1027E+06, 7.1911E+06,
     +  7.2806E+06, 7.3710E+06, 7.4625E+06, 7.5551E+06, 7.6488E+06,
     + 7.7436E+06, 7.8394E+06, 7.9364E+06, 8.0345E+06, 8.1337E+06,
     +  8.2341E+06, 8.3357E+06, 8.4384E+06, 8.5424E+06, 8.6475E+06,
     + 8.7538E+06, 8.8614E+06, 8.9702E+06, 9.0803E+06, 9.1917E+06,
     +  9.3043E+06, 9.4182E+06, 9.5334E+06, 9.6500E+06, 9.7679E+06,
     + 9.8872E+06, 1.0008E+07, 1.0130E+07, 1.0253E+07, 1.0378E+07,
     +  1.0504E+07, 1.0632E+07, 1.0761E+07, 1.0892E+07, 1.1024E+07,
     + 1.1157E+07, 1.1293E+07, 1.1429E+07, 1.1568E+07, 1.1707E+07,
     +  1.1849E+07, 1.1992E+07, 1.2136E+07, 1.2283E+07, 1.2431E+07,
     + 1.2580E+07, 1.2732E+07, 1.2885E+07, 1.3039E+07, 1.3196E+07,
     +  1.3354E+07, 1.3514E+07, 1.3676E+07, 1.3840E+07, 1.4005E+07,
     + 1.4173E+07, 1.4342E+07, 1.4513E+07, 1.4686E+07, 1.4861E+07,
     +  1.5038E+07, 1.5217E+07, 1.5398E+07, 1.5581E+07, 1.5766E+07,
     + 1.5953E+07, 1.6142E+07, 1.6334E+07, 1.6527E+07, 1.6723E+07,
     +  1.6920E+07, 1.7120E+07, 1.7323E+07, 1.7527E+07, 1.7734E+07,
     + 1.7943E+07, 1.8154E+07, 1.8368E+07, 1.8584E+07, 1.8802E+07,
     +  1.9023E+07, 1.9246E+07, 1.9472E+07, 1.9700E+07, 1.9931E+07,
     + 2.0164E+07, 2.0400E+07, 2.0639E+07, 2.0880E+07, 2.1124E+07,
     +  2.1370E+07, 2.1619E+07, 2.1871E+07, 2.2126E+07, 2.2383E+07,
     + 2.2643E+07, 2.2906E+07, 2.3172E+07, 2.3441E+07, 2.3713E+07,
     +  2.3987E+07, 2.4265E+07, 2.4546E+07, 2.4830E+07, 2.5117E+07,
     + 2.5407E+07, 2.5700E+07, 2.5996E+07, 2.6296E+07, 2.6599E+07,
     +  2.6905E+07, 2.7214E+07, 2.7527E+07, 2.7843E+07, 2.8163E+07,
     + 2.8486E+07, 2.8812E+07, 2.9142E+07, 2.9476E+07, 2.9813E+07,
     +  3.0154E+07, 3.0498E+07, 3.0846E+07, 3.1198E+07, 3.1554E+07,
     + 3.1913E+07, 3.2277E+07, 3.2644E+07, 3.3015E+07, 3.3390E+07,
     +  3.3769E+07, 3.4152E+07, 3.4540E+07, 3.4931E+07, 3.5327E+07,
     + 3.5726E+07, 3.6130E+07, 3.6539E+07, 3.6951E+07, 3.7368E+07,
     +  3.7790E+07, 3.8216E+07, 3.8646E+07, 3.9081E+07, 3.9521E+07,
     + 3.9965E+07, 4.0414E+07, 4.0868E+07, 4.1326E+07, 4.1790E+07,
     +  4.2258E+07, 4.2731E+07, 4.3209E+07, 4.3693E+07, 4.4181E+07,
     + 4.4674E+07, 4.5173E+07, 4.5677E+07, 4.6186E+07, 4.6700E+07,
     +  4.7220E+07, 4.7745E+07, 4.8276E+07, 4.8812E+07, 4.9354E+07,
     + 4.9902E+07, 5.0455E+07, 5.1014E+07, 5.1579E+07, 5.2150E+07,
     +  5.2726E+07, 5.3309E+07, 5.3898E+07, 5.4493E+07, 5.5094E+07,
     + 5.5701E+07, 5.6314E+07, 5.6934E+07, 5.7560E+07, 5.8193E+07,
     +  5.8833E+07, 5.9478E+07, 6.0131E+07, 6.0790E+07, 6.1456E+07,
     + 6.2129E+07, 6.2809E+07, 6.3496E+07, 6.4190E+07, 6.4891E+07,
     +  6.5599E+07, 6.6315E+07, 6.7038E+07, 6.7768E+07, 6.8506E+07,
     + 6.9251E+07, 7.0004E+07, 7.0765E+07, 7.1533E+07, 7.2309E+07,
     +  7.3093E+07, 7.3885E+07, 7.4686E+07, 7.5494E+07, 7.6310E+07,
     + 7.7135E+07, 7.7968E+07, 7.8810E+07, 7.9660E+07, 8.0519E+07,
     +  8.1387E+07, 8.2263E+07, 8.3148E+07, 8.4042E+07, 8.4945E+07,
     + 8.5858E+07, 8.6779E+07, 8.7710E+07, 8.8650E+07, 8.9600E+07,
     +  9.0559E+07, 9.1528E+07, 9.2506E+07, 9.3494E+07, 9.4493E+07,
     + 9.5501E+07, 9.6519E+07, 9.7548E+07, 9.8587E+07, 9.9636E+07,
     +  1.0070E+08, 1.0177E+08, 1.0285E+08, 1.0394E+08, 1.0504E+08,
     + 1.0616E+08, 1.0728E+08, 1.0842E+08, 1.0956E+08, 1.1072E+08,
     +  1.1189E+08, 1.1307E+08, 1.1427E+08, 1.1547E+08, 1.1669E+08,
     + 1.1792E+08, 1.1916E+08, 1.2041E+08, 1.2168E+08, 1.2296E+08,
     +  1.2425E+08, 1.2555E+08, 1.2687E+08, 1.2820E+08, 1.2954E+08,
     + 1.3090E+08, 1.3227E+08, 1.3365E+08, 1.3505E+08, 1.3646E+08,
     +  1.3788E+08, 1.3932E+08, 1.4077E+08, 1.4224E+08, 1.4372E+08,
     + 1.4521E+08, 1.4672E+08, 1.4825E+08, 1.4979E+08, 1.5134E+08,
     +  1.5291E+08, 1.5450E+08, 1.5610E+08, 1.5771E+08, 1.5934E+08,
     + 1.6099E+08, 1.6265E+08, 1.6433E+08, 1.6603E+08, 1.6774E+08,
     +  1.6947E+08, 1.7121E+08, 1.7297E+08, 1.7475E+08, 1.7655E+08,
     + 1.7836E+08, 1.8019E+08, 1.8204E+08, 1.8391E+08, 1.8579E+08,
     +  1.8769E+08, 1.8961E+08, 1.9155E+08, 1.9351E+08, 1.9549E+08,
     + 1.9748E+08, 1.9950E+08, 2.0153E+08, 2.0358E+08, 2.0566E+08,
     +  2.0775E+08, 2.0986E+08, 2.1200E+08, 2.1415E+08, 2.1632E+08,
     + 2.1852E+08, 2.2073E+08, 2.2297E+08, 2.2522E+08, 2.2750E+08,
     +  2.2980E+08, 2.3213E+08, 2.3447E+08, 2.3684E+08, 2.3923E+08,
     + 2.4164E+08, 2.4407E+08, 2.4653E+08, 2.4901E+08, 2.5151E+08,
     +  2.5404E+08, 2.5659E+08, 2.5917E+08, 2.6177E+08, 2.6439E+08,
     + 2.6704E+08, 2.6971E+08, 2.7241E+08, 2.7514E+08, 2.7788E+08,
     +  2.8066E+08, 2.8346E+08, 2.8629E+08, 2.8914E+08, 2.9202E+08,
     + 2.9493E+08, 2.9786E+08, 3.0083E+08, 3.0382E+08, 3.0683E+08,
     +  3.0988E+08, 3.1295E+08, 3.1605E+08, 3.1919E+08, 3.2235E+08,
     + 3.2554E+08, 3.2875E+08, 3.3200E+08, 3.3528E+08, 3.3859E+08,
     +  3.4193E+08, 3.4530E+08, 3.4870E+08, 3.5214E+08, 3.5560E+08,
     + 3.5910E+08, 3.6263E+08, 3.6619E+08, 3.6979E+08, 3.7341E+08,
     +  3.7707E+08, 3.8077E+08, 3.8450E+08, 3.8826E+08, 3.9206E+08,
     + 3.9589E+08, 3.9975E+08, 4.0366E+08, 4.0760E+08, 4.1157E+08,
     +  4.1558E+08, 4.1963E+08, 4.2371E+08, 4.2783E+08, 4.3199E+08,
     + 4.3619E+08, 4.4042E+08, 4.4469E+08, 4.4901E+08, 4.5336E+08,
     +  4.5775E+08, 4.6218E+08, 4.6665E+08, 4.7116E+08, 4.7571E+08,
     + 4.8030E+08, 4.8494E+08, 4.8962E+08, 4.9433E+08, 4.9910E+08,
     +  5.0390E+08, 5.0875E+08, 5.1364E+08, 5.1857E+08, 5.2355E+08,
     + 5.2858E+08, 5.3365E+08, 5.3876E+08, 5.4392E+08, 5.4913E+08,
     +  5.5438E+08, 5.5969E+08, 5.6503E+08, 5.7043E+08, 5.7587E+08,
     + 5.8137E+08, 5.8691E+08, 5.9250E+08, 5.9814E+08, 6.0383E+08,
     +  6.0958E+08, 6.1537E+08, 6.2122E+08, 6.2711E+08, 6.3306E+08,
     + 6.3907E+08, 6.4512E+08, 6.5123E+08, 6.5739E+08, 6.6361E+08,
     +  6.6988E+08, 6.7621E+08, 6.8260E+08, 6.8904E+08, 6.9554E+08,
     + 7.0209E+08, 7.0870E+08, 7.1538E+08, 7.2211E+08, 7.2890E+08,
     +  7.3574E+08, 7.4265E+08, 7.4962E+08, 7.5665E+08, 7.6375E+08,
     + 7.7090E+08, 7.7812E+08, 7.8540E+08, 7.9275E+08, 8.0015E+08,
     +  8.0763E+08, 8.1517E+08, 8.2277E+08, 8.3044E+08, 8.3818E+08,
     + 8.4599E+08, 8.5386E+08, 8.6180E+08, 8.6981E+08, 8.7790E+08,
     +  8.8605E+08, 8.9427E+08, 9.0256E+08, 9.1093E+08, 9.1937E+08,
     + 9.2788E+08, 9.3646E+08, 9.4512E+08, 9.5385E+08, 9.6266E+08,
     +  9.7155E+08, 9.8051E+08, 9.8955E+08, 9.9867E+08, 1.0079E+09,
     + 1.0171E+09, 1.0265E+09, 1.0359E+09, 1.0454E+09, 1.0550E+09,
     +  1.0647E+09, 1.0745E+09, 1.0843E+09, 1.0943E+09, 1.1043E+09,
     + 1.1144E+09, 1.1246E+09, 1.1348E+09, 1.1452E+09, 1.1557E+09,
     +  1.1662E+09, 1.1768E+09, 1.1876E+09, 1.1984E+09, 1.2093E+09,
     + 1.2203E+09, 1.2314E+09, 1.2426E+09, 1.2538E+09, 1.2652E+09,
     +  1.2767E+09, 1.2883E+09, 1.2999E+09, 1.3117E+09, 1.3236E+09,
     + 1.3355E+09, 1.3476E+09, 1.3598E+09, 1.3720E+09, 1.3844E+09,
     +  1.3969E+09, 1.4095E+09, 1.4222E+09, 1.4350E+09, 1.4479E+09,
     + 1.4609E+09, 1.4740E+09, 1.4872E+09, 1.5006E+09, 1.5141E+09,
     +  1.5276E+09, 1.5413E+09, 1.5551E+09, 1.5690E+09, 1.5830E+09,
     + 1.5972E+09, 1.6115E+09, 1.6258E+09, 1.6404E+09, 1.6550E+09,
     +  1.6697E+09, 1.6846E+09, 1.6996E+09, 1.7147E+09, 1.7300E+09,
     + 1.7453E+09, 1.7608E+09, 1.7764E+09, 1.7922E+09, 1.8081E+09,
     +  1.8241E+09, 1.8403E+09, 1.8565E+09, 1.8730E+09, 1.8895E+09,
     + 1.9062E+09, 1.9230E+09, 1.9400E+09, 1.9571E+09, 1.9744E+09,
     +  1.9918E+09, 2.0093E+09, 2.0270E+09, 2.0448E+09, 2.0628E+09,
     + 2.0809E+09, 2.0991E+09, 2.1176E+09, 2.1361E+09, 2.1549E+09,
     +  2.1737E+09, 2.1928E+09, 2.2119E+09, 2.2313E+09, 2.2508E+09,
     + 2.2704E+09, 2.2902E+09, 2.3102E+09, 2.3304E+09, 2.3507E+09,
     +  2.3711E+09, 2.3918E+09, 2.4126E+09, 2.4335E+09, 2.4547E+09,
     + 2.4760E+09, 2.4975E+09, 2.5192E+09, 2.5410E+09, 2.5630E+09,
     +  2.5852E+09, 2.6076E+09, 2.6301E+09, 2.6529E+09, 2.6758E+09,
     + 2.6989E+09, 2.7222E+09, 2.7456E+09, 2.7693E+09, 2.7932E+09,
     +  2.8172E+09, 2.8415E+09, 2.8659E+09, 2.8905E+09, 2.9153E+09,
     + 2.9404E+09, 2.9656E+09, 2.9910E+09, 3.0167E+09, 3.0425E+09,
     +  3.0686E+09, 3.0948E+09, 3.1213E+09, 3.1480E+09, 3.1748E+09,
     + 3.2019E+09, 3.2293E+09, 3.2568E+09, 3.2846E+09, 3.3125E+09,
     +  3.3407E+09, 3.3692E+09, 3.3978E+09, 3.4267E+09, 3.4558E+09,
     + 3.4851E+09, 3.5147E+09, 3.5445E+09, 3.5745E+09, 3.6048E+09,
     +  3.6353E+09, 3.6661E+09, 3.6971E+09, 3.7283E+09, 3.7598E+09,
     + 3.7915E+09, 3.8235E+09, 3.8558E+09, 3.8883E+09, 3.9210E+09,
     +  3.9540E+09, 3.9873E+09, 4.0208E+09, 4.0546E+09, 4.0887E+09,
     + 4.1230E+09, 4.1576E+09, 4.1924E+09, 4.2276E+09, 4.2630E+09,
     +  4.2987E+09, 4.3346E+09, 4.3709E+09, 4.4074E+09, 4.4442E+09,
     + 4.4813E+09, 4.5187E+09, 4.5564E+09, 4.5944E+09, 4.6326E+09,
     +  4.6712E+09, 4.7101E+09, 4.7492E+09, 4.7887E+09, 4.8285E+09,
     + 4.8686E+09, 4.9090E+09, 4.9497E+09, 4.9907E+09, 5.0320E+09,
     +  5.0737E+09, 5.1157E+09, 5.1580E+09, 5.2006E+09, 5.2436E+09,
     + 5.2869E+09, 5.3305E+09, 5.3744E+09, 5.4187E+09, 5.4634E+09,
     +  5.5084E+09, 5.5537E+09, 5.5994E+09, 5.6454E+09, 5.6918E+09,
     + 5.7385E+09, 5.7856E+09, 5.8331E+09, 5.8809E+09, 5.9291E+09,
     +  5.9776E+09, 6.0265E+09, 6.0758E+09, 6.1255E+09, 6.1755E+09,
     + 6.2260E+09, 6.2768E+09, 6.3280E+09, 6.3796E+09, 6.4315E+09,
     +  6.4839E+09, 6.5367E+09, 6.5899E+09, 6.6435E+09, 6.6974E+09,
     + 6.7518E+09, 6.8066E+09, 6.8619E+09, 6.9175E+09, 6.9736E+09,
     +  7.0300E+09, 7.0869E+09, 7.1443E+09, 7.2021E+09, 7.2603E+09,
     + 7.3189E+09, 7.3780E+09, 7.4375E+09, 7.4975E+09, 7.5579E+09,
     +  7.6188E+09, 7.6802E+09, 7.7420E+09, 7.8042E+09, 7.8670E+09,
     + 7.9302E+09, 7.9938E+09, 8.0580E+09, 8.1226E+09, 8.1877E+09,
     +  8.2534E+09, 8.3194E+09, 8.3860E+09, 8.4531E+09, 8.5207E+09,
     + 8.5888E+09, 8.6574E+09, 8.7265E+09, 8.7961E+09, 8.8662E+09,
     +  8.9369E+09, 9.0081E+09, 9.0798E+09, 9.1520E+09, 9.2248E+09,
     + 9.2981E+09, 9.3720E+09, 9.4464E+09, 9.5214E+09, 9.5969E+09,
     +  9.6730E+09, 9.7496E+09, 9.8268E+09, 9.9046E+09, 9.9829E+09,
     + 1.0062E+10, 1.0141E+10, 1.0221E+10, 1.0302E+10, 1.0383E+10,
     +  1.0465E+10, 1.0548E+10, 1.0631E+10, 1.0714E+10, 1.0799E+10,
     + 1.0884E+10, 1.0969E+10, 1.1055E+10, 1.1142E+10, 1.1230E+10,
     +  1.1318E+10, 1.1406E+10, 1.1496E+10, 1.1586E+10, 1.1676E+10,
     + 1.1768E+10, 1.1860E+10, 1.1952E+10, 1.2046E+10, 1.2140E+10,
     +  1.2235E+10, 1.2330E+10, 1.2426E+10, 1.2523E+10, 1.2620E+10,
     + 1.2719E+10, 1.2817E+10, 1.2917E+10, 1.3017E+10, 1.3118E+10,
     +  1.3220E+10, 1.3323E+10, 1.3426E+10, 1.3530E+10, 1.3635E+10,
     + 1.3740E+10, 1.3847E+10, 1.3954E+10, 1.4062E+10, 1.4170E+10,
     +  1.4280E+10, 1.4390E+10, 1.4501E+10, 1.4613E+10, 1.4725E+10,
     + 1.4838E+10, 1.4953E+10, 1.5068E+10, 1.5184E+10, 1.5300E+10,
     +  1.5418E+10, 1.5536E+10, 1.5655E+10, 1.5775E+10, 1.5896E+10,
     + 1.6018E+10, 1.6140E+10, 1.6264E+10, 1.6388E+10, 1.6514E+10,
     +  1.6640E+10, 1.6767E+10, 1.6895E+10, 1.7024E+10, 1.7153E+10,
     + 1.7284E+10, 1.7416E+10, 1.7548E+10, 1.7682E+10, 1.7816E+10,
     +  1.7952E+10, 1.8088E+10, 1.8225E+10, 1.8364E+10, 1.8503E+10,
     + 1.8643E+10, 1.8785E+10, 1.8927E+10, 1.9070E+10, 1.9214E+10,
     +  1.9360E+10, 1.9506E+10, 1.9653E+10, 1.9802E+10, 1.9951E+10,
     + 2.0101E+10, 2.0253E+10, 2.0405E+10, 2.0559E+10, 2.0714E+10,
     +  2.0870E+10, 2.1027E+10, 2.1185E+10, 2.1344E+10, 2.1504E+10,
     + 2.1665E+10, 2.1828E+10, 2.1991E+10, 2.2156E+10, 2.2322E+10,
     +  2.2489E+10, 2.2657E+10, 2.2827E+10, 2.2997E+10, 2.3169E+10,
     + 2.3342E+10, 2.3516E+10, 2.3691E+10, 2.3868E+10, 2.4046E+10,
     +  2.4225E+10, 2.4405E+10, 2.4587E+10, 2.4769E+10, 2.4953E+10,
     + 2.5139E+10, 2.5325E+10, 2.5513E+10, 2.5702E+10, 2.5893E+10,
     +  2.6085E+10, 2.6278E+10, 2.6472E+10, 2.6668E+10, 2.6865E+10,
     + 2.7064E+10, 2.7264E+10, 2.7465E+10, 2.7667E+10, 2.7871E+10,
     +  2.8077E+10, 2.8284E+10, 2.8492E+10, 2.8702E+10, 2.8913E+10,
     + 2.9125E+10, 2.9339E+10, 2.9555E+10, 2.9772E+10, 2.9990E+10,
     +  3.0210E+10, 3.0432E+10, 3.0654E+10, 3.0879E+10, 3.1105E+10,
     + 3.1332E+10, 3.1562E+10, 3.1792E+10, 3.2024E+10, 3.2258E+10,
     +  3.2494E+10, 3.2730E+10, 3.2969E+10, 3.3209E+10, 3.3451E+10,
     + 3.3695E+10, 3.3940E+10, 3.4186E+10, 3.4435E+10, 3.4685E+10,
     +  3.4937E+10, 3.5190E+10, 3.5446E+10, 3.5703E+10, 3.5961E+10,
     + 3.6222E+10, 3.6484E+10, 3.6748E+10, 3.7013E+10, 3.7281E+10,
     +  3.7550E+10, 3.7821E+10, 3.8094E+10, 3.8369E+10, 3.8646E+10,
     + 3.8924E+10, 3.9205E+10, 3.9487E+10, 3.9771E+10, 4.0057E+10,
     +  4.0345E+10, 4.0635E+10, 4.0927E+10, 4.1220E+10, 4.1516E+10,
     + 4.1814E+10, 4.2113E+10, 4.2415E+10, 4.2719E+10, 4.3024E+10,
     +  4.3332E+10, 4.3642E+10, 4.3954E+10, 4.4268E+10, 4.4584E+10,
     + 4.4902E+10, 4.5222E+10, 4.5544E+10, 4.5869E+10, 4.6196E+10,
     +  4.6524E+10, 4.6855E+10, 4.7189E+10, 4.7524E+10, 4.7862E+10,
     + 4.8201E+10, 4.8543E+10, 4.8888E+10, 4.9234E+10, 4.9583E+10,
     +  4.9934E+10, 5.0288E+10, 5.0644E+10, 5.1002E+10, 5.1362E+10,
     + 5.1725E+10, 5.2091E+10, 5.2458E+10, 5.2828E+10, 5.3201E+10,
     +  5.3576E+10, 5.3953E+10, 5.4333E+10, 5.4716E+10, 5.5101E+10,
     + 5.5488E+10, 5.5878E+10, 5.6270E+10, 5.6666E+10, 5.7063E+10,
     +  5.7463E+10, 5.7866E+10, 5.8272E+10, 5.8680E+10, 5.9091E+10,
     + 5.9504E+10, 5.9920E+10, 6.0339E+10, 6.0761E+10, 6.1185E+10,
     +  6.1612E+10, 6.2042E+10, 6.2475E+10, 6.2910E+10, 6.3348E+10,
     + 6.3789E+10, 6.4233E+10, 6.4680E+10, 6.5130E+10, 6.5583E+10,
     +  6.6038E+10, 6.6497E+10, 6.6958E+10, 6.7423E+10, 6.7890E+10,
     + 6.8361E+10, 6.8834E+10, 6.9311E+10, 6.9790E+10, 7.0273E+10,
     +  7.0759E+10, 7.1248E+10, 7.1740E+10, 7.2235E+10, 7.2734E+10,
     + 7.3235E+10, 7.3740E+10, 7.4248E+10, 7.4759E+10, 7.5274E+10,
     +  7.5792E+10, 7.6313E+10, 7.6838E+10, 7.7366E+10, 7.7897E+10,
     + 7.8432E+10, 7.8970E+10, 7.9511E+10, 8.0056E+10, 8.0605E+10,
     +  8.1156E+10, 8.1712E+10, 8.2271E+10, 8.2833E+10, 8.3400E+10,
     + 8.3969E+10, 8.4543E+10, 8.5120E+10, 8.5700E+10, 8.6285E+10,
     +  8.6873E+10, 8.7464E+10, 8.8060E+10, 8.8659E+10, 8.9262E+10,
     + 8.9869E+10, 9.0480E+10, 9.1094E+10, 9.1713E+10, 9.2335E+10,
     +  9.2962E+10, 9.3592E+10, 9.4226E+10, 9.4864E+10, 9.5507E+10,
     + 9.6153E+10, 9.6803E+10, 9.7458E+10, 9.8116E+10, 9.8779E+10,
     +  9.9446E+10, 1.0012E+11, 1.0079E+11, 1.0147E+11, 1.0216E+11,
     + 1.0284E+11, 1.0354E+11, 1.0423E+11, 1.0493E+11, 1.0564E+11,
     +  1.0635E+11, 1.0706E+11, 1.0778E+11, 1.0850E+11, 1.0923E+11,
     + 1.0996E+11, 1.1070E+11, 1.1144E+11, 1.1219E+11, 1.1294E+11,
     +  1.1370E+11, 1.1446E+11, 1.1522E+11, 1.1599E+11, 1.1676E+11,
     + 1.1754E+11, 1.1833E+11, 1.1912E+11, 1.1991E+11, 1.2071E+11,
     +  1.2151E+11, 1.2232E+11, 1.2313E+11, 1.2395E+11, 1.2478E+11,
     + 1.2560E+11, 1.2644E+11, 1.2728E+11, 1.2812E+11, 1.2897E+11,
     +  1.2983E+11, 1.3069E+11, 1.3155E+11, 1.3242E+11, 1.3330E+11,
     + 1.3418E+11, 1.3506E+11, 1.3596E+11, 1.3685E+11, 1.3776E+11,
     +  1.3867E+11, 1.3958E+11, 1.4050E+11, 1.4143E+11, 1.4236E+11,
     + 1.4329E+11, 1.4424E+11, 1.4518E+11, 1.4614E+11, 1.4710E+11,
     +  1.4806E+11, 1.4904E+11, 1.5001E+11, 1.5100E+11, 1.5199E+11,
     + 1.5298E+11, 1.5398E+11, 1.5499E+11, 1.5600E+11, 1.5702E+11,
     +  1.5805E+11, 1.5908E+11, 1.6012E+11, 1.6117E+11, 1.6222E+11,
     + 1.6328E+11, 1.6434E+11, 1.6541E+11, 1.6649E+11, 1.6757E+11,
     +  1.6866E+11, 1.6976E+11, 1.7086E+11, 1.7197E+11, 1.7309E+11,
     + 1.7421E+11, 1.7534E+11, 1.7648E+11, 1.7762E+11, 1.7877E+11,
     +  1.7993E+11, 1.8110E+11, 1.8227E+11, 1.8345E+11, 1.8463E+11,
     + 1.8583E+11, 1.8703E+11, 1.8823E+11, 1.8945E+11, 1.9067E+11,
     +  1.9190E+11, 1.9314E+11, 1.9438E+11, 1.9563E+11, 1.9689E+11,
     + 1.9816E+11, 1.9943E+11, 2.0071E+11, 2.0200E+11, 2.0330E+11,
     +  2.0460E+11, 2.0592E+11, 2.0724E+11, 2.0856E+11, 2.0990E+11,
     + 2.1124E+11, 2.1260E+11, 2.1396E+11, 2.1532E+11, 2.1670E+11,
     +  2.1809E+11, 2.1948E+11, 2.2088E+11, 2.2229E+11, 2.2371E+11,
     + 2.2513E+11, 2.2657E+11, 2.2801E+11, 2.2946E+11, 2.3092E+11,
     +  2.3239E+11, 2.3387E+11, 2.3535E+11, 2.3685E+11, 2.3835E+11,
     + 2.3987E+11, 2.4139E+11, 2.4292E+11, 2.4446E+11, 2.4601E+11,
     +  2.4756E+11, 2.4913E+11, 2.5071E+11, 2.5229E+11, 2.5389E+11,
     + 2.5549E+11, 2.5710E+11, 2.5873E+11, 2.6036E+11, 2.6200E+11,
     +  2.6365E+11, 2.6532E+11, 2.6699E+11, 2.6867E+11, 2.7036E+11,
     + 2.7206E+11, 2.7377E+11, 2.7549E+11, 2.7722E+11, 2.7896E+11,
     +  2.8071E+11, 2.8247E+11, 2.8424E+11, 2.8603E+11, 2.8782E+11,
     + 2.8962E+11, 2.9143E+11, 2.9326E+11, 2.9509E+11, 2.9694E+11,
     +  2.9879E+11, 3.0066E+11, 3.0253E+11, 3.0442E+11, 3.0632E+11,
     + 3.0823E+11, 3.1015E+11, 3.1208E+11, 3.1403E+11, 3.1598E+11,
     +  3.1795E+11, 3.1993E+11, 3.2191E+11, 3.2391E+11, 3.2593E+11,
     + 3.2795E+11, 3.2998E+11, 3.3203E+11, 3.3409E+11, 3.3616E+11,
     +  3.3824E+11, 3.4034E+11, 3.4244E+11, 3.4456E+11, 3.4669E+11,
     + 3.4883E+11, 3.5099E+11, 3.5316E+11, 3.5534E+11, 3.5753E+11,
     +  3.5973E+11, 3.6195E+11, 3.6418E+11, 3.6642E+11, 3.6868E+11,
     + 3.7095E+11, 3.7323E+11, 3.7553E+11, 3.7783E+11, 3.8015E+11,
     +  3.8249E+11, 3.8484E+11, 3.8720E+11, 3.8957E+11, 3.9196E+11,
     + 3.9436E+11, 3.9678E+11, 3.9921E+11, 4.0165E+11, 4.0411E+11,
     +  4.0658E+11, 4.0906E+11, 4.1156E+11, 4.1407E+11, 4.1660E+11,
     + 4.1914E+11, 4.2170E+11, 4.2427E+11, 4.2685E+11, 4.2945E+11,
     +  4.3207E+11, 4.3470E+11, 4.3734E+11, 4.4000E+11, 4.4267E+11,
     + 4.4536E+11, 4.4807E+11, 4.5079E+11, 4.5352E+11, 4.5627E+11,
     +  4.5904E+11, 4.6182E+11, 4.6462E+11, 4.6743E+11, 4.7026E+11,
     + 4.7310E+11, 4.7596E+11, 4.7884E+11, 4.8173E+11, 4.8464E+11,
     +  4.8757E+11, 4.9051E+11, 4.9347E+11, 4.9644E+11, 4.9943E+11,
     + 5.0244E+11, 5.0547E+11, 5.0851E+11, 5.1157E+11, 5.1464E+11,
     +  5.1774E+11, 5.2085E+11, 5.2397E+11, 5.2712E+11, 5.3028E+11,
     + 5.3346E+11, 5.3666E+11, 5.3988E+11, 5.4311E+11, 5.4636E+11,
     +  5.4963E+11, 5.5292E+11, 5.5623E+11, 5.5955E+11, 5.6289E+11,
     + 5.6626E+11, 5.6964E+11, 5.7304E+11, 5.7645E+11, 5.7989E+11,
     +  5.8335E+11, 5.8682E+11, 5.9032E+11, 5.9383E+11, 5.9736E+11,
     + 6.0091E+11, 6.0449E+11, 6.0808E+11, 6.1169E+11, 6.1532E+11,
     +  6.1897E+11, 6.2265E+11, 6.2634E+11, 6.3005E+11, 6.3378E+11,
     + 6.3754E+11, 6.4131E+11, 6.4510E+11, 6.4892E+11, 6.5276E+11,
     +  6.5661E+11, 6.6049E+11, 6.6439E+11, 6.6831E+11, 6.7226E+11,
     + 6.7622E+11, 6.8021E+11, 6.8422E+11, 6.8825E+11, 6.9230E+11,
     +  6.9637E+11, 7.0047E+11, 7.0459E+11, 7.0873E+11, 7.1289E+11,
     + 7.1708E+11, 7.2129E+11, 7.2552E+11, 7.2978E+11, 7.3406E+11,
     +  7.3836E+11, 7.4268E+11, 7.4703E+11, 7.5141E+11, 7.5580E+11,
     + 7.6022E+11, 7.6467E+11, 7.6913E+11, 7.7363E+11, 7.7814E+11,
     +  7.8269E+11, 7.8725E+11, 7.9184E+11, 7.9646E+11, 8.0110E+11,
     + 8.0577E+11, 8.1046E+11, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 4,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_CF4   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 1),QofT( 1,NT),Q(NT)
      data Tmx(1)/0./
      data xgj/ 0./
c...      CF4
c...        --        29
      data (QofT( 1,J),J=1,2408)/ 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_C4H2  (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 1),QofT( 1,NT),Q(NT)
      data Tmx(1)/3500./
      data xgj/ 1./
c...     C4H2
c...        --      2211
      data (QofT( 1,J),J=1,2408)/ 1.0191E+01, 1.9671E+01, 2.9162E+01,
     +  3.8655E+01, 4.8148E+01, 5.7643E+01, 6.7137E+01, 7.6632E+01,
     + 8.6127E+01, 9.5622E+01, 1.0512E+02, 1.1461E+02, 1.2411E+02,
     +  1.3360E+02, 1.4310E+02, 1.5259E+02, 1.6209E+02, 1.7158E+02,
     + 1.8108E+02, 1.9058E+02, 2.0957E+02, 2.2856E+02, 2.4755E+02,
     +  2.6655E+02, 2.8555E+02, 3.0455E+02, 3.2357E+02, 3.4261E+02,
     + 3.6167E+02, 3.8077E+02, 3.9991E+02, 4.1910E+02, 4.3836E+02,
     +  4.5770E+02, 4.7714E+02, 4.9669E+02, 5.1636E+02, 5.3617E+02,
     + 5.5614E+02, 5.7628E+02, 5.9661E+02, 6.1714E+02, 6.3789E+02,
     +  6.5887E+02, 6.8010E+02, 7.0160E+02, 7.2337E+02, 7.4544E+02,
     + 7.6781E+02, 7.9051E+02, 8.1354E+02, 8.3691E+02, 8.6065E+02,
     +  8.8477E+02, 9.0928E+02, 9.3419E+02, 9.5952E+02, 9.8528E+02,
     + 1.0115E+03, 1.0382E+03, 1.0653E+03, 1.0929E+03, 1.1211E+03,
     +  1.1497E+03, 1.1789E+03, 1.2086E+03, 1.2389E+03, 1.2698E+03,
     + 1.3013E+03, 1.3333E+03, 1.3661E+03, 1.3994E+03, 1.4334E+03,
     +  1.4681E+03, 1.5035E+03, 1.5396E+03, 1.5765E+03, 1.6141E+03,
     + 1.6524E+03, 1.6916E+03, 1.7316E+03, 1.7724E+03, 1.8141E+03,
     +  1.8567E+03, 1.9001E+03, 1.9445E+03, 1.9898E+03, 2.0361E+03,
     + 2.0834E+03, 2.1317E+03, 2.1811E+03, 2.2315E+03, 2.2830E+03,
     +  2.3356E+03, 2.3894E+03, 2.4444E+03, 2.5006E+03, 2.5580E+03,
     + 2.6167E+03, 2.6766E+03, 2.7379E+03, 2.8006E+03, 2.8646E+03,
     +  2.9301E+03, 2.9970E+03, 3.0654E+03, 3.1354E+03, 3.2069E+03,
     + 3.2800E+03, 3.3547E+03, 3.4311E+03, 3.5092E+03, 3.5891E+03,
     +  3.6708E+03, 3.7543E+03, 3.8396E+03, 3.9269E+03, 4.0161E+03,
     + 4.1074E+03, 4.2007E+03, 4.2961E+03, 4.3936E+03, 4.4933E+03,
     +  4.5953E+03, 4.6995E+03, 4.8061E+03, 4.9151E+03, 5.0265E+03,
     + 5.1405E+03, 5.2570E+03, 5.3761E+03, 5.4979E+03, 5.6224E+03,
     +  5.7496E+03, 5.8798E+03, 6.0128E+03, 6.1489E+03, 6.2879E+03,
     + 6.4301E+03, 6.5754E+03, 6.7240E+03, 6.8759E+03, 7.0312E+03,
     +  7.1899E+03, 7.3522E+03, 7.5180E+03, 7.6876E+03, 7.8609E+03,
     + 8.0380E+03, 8.2191E+03, 8.4041E+03, 8.5933E+03, 8.7866E+03,
     +  8.9842E+03, 9.1861E+03, 9.3925E+03, 9.6034E+03, 9.8190E+03,
     + 1.0039E+04, 1.0264E+04, 1.0494E+04, 1.0729E+04, 1.0970E+04,
     +  1.1215E+04, 1.1466E+04, 1.1722E+04, 1.1984E+04, 1.2251E+04,
     + 1.2525E+04, 1.2804E+04, 1.3089E+04, 1.3380E+04, 1.3678E+04,
     +  1.3982E+04, 1.4293E+04, 1.4610E+04, 1.4934E+04, 1.5266E+04,
     + 1.5604E+04, 1.5949E+04, 1.6302E+04, 1.6662E+04, 1.7030E+04,
     +  1.7406E+04, 1.7790E+04, 1.8182E+04, 1.8582E+04, 1.8991E+04,
     + 1.9408E+04, 1.9834E+04, 2.0270E+04, 2.0714E+04, 2.1168E+04,
     +  2.1631E+04, 2.2104E+04, 2.2587E+04, 2.3079E+04, 2.3583E+04,
     + 2.4096E+04, 2.4621E+04, 2.5156E+04, 2.5702E+04, 2.6260E+04,
     +  2.6829E+04, 2.7410E+04, 2.8003E+04, 2.8609E+04, 2.9226E+04,
     + 2.9857E+04, 3.0500E+04, 3.1157E+04, 3.1827E+04, 3.2510E+04,
     +  3.3208E+04, 3.3920E+04, 3.4646E+04, 3.5387E+04, 3.6143E+04,
     + 3.6915E+04, 3.7702E+04, 3.8505E+04, 3.9324E+04, 4.0159E+04,
     +  4.1012E+04, 4.1881E+04, 4.2768E+04, 4.3673E+04, 4.4596E+04,
     + 4.5537E+04, 4.6498E+04, 4.7477E+04, 4.8475E+04, 4.9494E+04,
     +  5.0533E+04, 5.1592E+04, 5.2672E+04, 5.3774E+04, 5.4897E+04,
     + 5.6042E+04, 5.7210E+04, 5.8401E+04, 5.9615E+04, 6.0853E+04,
     +  6.2115E+04, 6.3402E+04, 6.4714E+04, 6.6051E+04, 6.7415E+04,
     + 6.8804E+04, 7.0221E+04, 7.1665E+04, 7.3137E+04, 7.4638E+04,
     +  7.6167E+04, 7.7726E+04, 7.9314E+04, 8.0933E+04, 8.2583E+04,
     + 8.4265E+04, 8.5979E+04, 8.7725E+04, 8.9505E+04, 9.1318E+04,
     +  9.3166E+04, 9.5049E+04, 9.6968E+04, 9.8923E+04, 1.0091E+05,
     + 1.0294E+05, 1.0501E+05, 1.0712E+05, 1.0926E+05, 1.1145E+05,
     +  1.1368E+05, 1.1595E+05, 1.1826E+05, 1.2061E+05, 1.2301E+05,
     + 1.2545E+05, 1.2794E+05, 1.3047E+05, 1.3305E+05, 1.3568E+05,
     +  1.3836E+05, 1.4108E+05, 1.4386E+05, 1.4669E+05, 1.4957E+05,
     + 1.5250E+05, 1.5548E+05, 1.5852E+05, 1.6162E+05, 1.6477E+05,
     +  1.6798E+05, 1.7125E+05, 1.7457E+05, 1.7796E+05, 1.8141E+05,
     + 1.8492E+05, 1.8849E+05, 1.9213E+05, 1.9583E+05, 1.9960E+05,
     +  2.0344E+05, 2.0735E+05, 2.1132E+05, 2.1537E+05, 2.1949E+05,
     + 2.2368E+05, 2.2795E+05, 2.3229E+05, 2.3671E+05, 2.4121E+05,
     +  2.4579E+05, 2.5045E+05, 2.5519E+05, 2.6001E+05, 2.6492E+05,
     + 2.6991E+05, 2.7500E+05, 2.8017E+05, 2.8543E+05, 2.9079E+05,
     +  2.9623E+05, 3.0178E+05, 3.0742E+05, 3.1315E+05, 3.1899E+05,
     + 3.2493E+05, 3.3097E+05, 3.3711E+05, 3.4336E+05, 3.4972E+05,
     +  3.5619E+05, 3.6277E+05, 3.6946E+05, 3.7627E+05, 3.8319E+05,
     + 3.9023E+05, 3.9739E+05, 4.0468E+05, 4.1208E+05, 4.1962E+05,
     +  4.2728E+05, 4.3507E+05, 4.4299E+05, 4.5105E+05, 4.5924E+05,
     + 4.6758E+05, 4.7605E+05, 4.8466E+05, 4.9342E+05, 5.0233E+05,
     +  5.1138E+05, 5.2059E+05, 5.2995E+05, 5.3947E+05, 5.4914E+05,
     + 5.5898E+05, 5.6898E+05, 5.7914E+05, 5.8948E+05, 5.9999E+05,
     +  6.1066E+05, 6.2152E+05, 6.3256E+05, 6.4377E+05, 6.5517E+05,
     + 6.6676E+05, 6.7854E+05, 6.9051E+05, 7.0268E+05, 7.1505E+05,
     +  7.2762E+05, 7.4039E+05, 7.5337E+05, 7.6656E+05, 7.7997E+05,
     + 7.9360E+05, 8.0744E+05, 8.2151E+05, 8.3581E+05, 8.5033E+05,
     +  8.6509E+05, 8.8009E+05, 8.9533E+05, 9.1082E+05, 9.2655E+05,
     + 9.4254E+05, 9.5878E+05, 9.7528E+05, 9.9204E+05, 1.0091E+06,
     +  1.0264E+06, 1.0440E+06, 1.0618E+06, 1.0799E+06, 1.0984E+06,
     + 1.1171E+06, 1.1361E+06, 1.1554E+06, 1.1750E+06, 1.1950E+06,
     +  1.2152E+06, 1.2358E+06, 1.2566E+06, 1.2779E+06, 1.2994E+06,
     + 1.3213E+06, 1.3435E+06, 1.3660E+06, 1.3890E+06, 1.4122E+06,
     +  1.4359E+06, 1.4599E+06, 1.4842E+06, 1.5090E+06, 1.5341E+06,
     + 1.5596E+06, 1.5855E+06, 1.6119E+06, 1.6386E+06, 1.6657E+06,
     +  1.6932E+06, 1.7212E+06, 1.7496E+06, 1.7784E+06, 1.8077E+06,
     + 1.8374E+06, 1.8676E+06, 1.8982E+06, 1.9293E+06, 1.9608E+06,
     +  1.9929E+06, 2.0254E+06, 2.0584E+06, 2.0919E+06, 2.1260E+06,
     + 2.1605E+06, 2.1956E+06, 2.2311E+06, 2.2673E+06, 2.3039E+06,
     +  2.3411E+06, 2.3789E+06, 2.4172E+06, 2.4561E+06, 2.4956E+06,
     + 2.5357E+06, 2.5763E+06, 2.6176E+06, 2.6595E+06, 2.7020E+06,
     +  2.7451E+06, 2.7889E+06, 2.8333E+06, 2.8784E+06, 2.9241E+06,
     + 2.9706E+06, 3.0176E+06, 3.0654E+06, 3.1139E+06, 3.1631E+06,
     +  3.2130E+06, 3.2637E+06, 3.3151E+06, 3.3672E+06, 3.4201E+06,
     + 3.4738E+06, 3.5283E+06, 3.5835E+06, 3.6395E+06, 3.6964E+06,
     +  3.7541E+06, 3.8126E+06, 3.8719E+06, 3.9322E+06, 3.9932E+06,
     + 4.0552E+06, 4.1181E+06, 4.1818E+06, 4.2465E+06, 4.3121E+06,
     +  4.3786E+06, 4.4461E+06, 4.5146E+06, 4.5840E+06, 4.6544E+06,
     + 4.7259E+06, 4.7983E+06, 4.8718E+06, 4.9463E+06, 5.0218E+06,
     +  5.0985E+06, 5.1762E+06, 5.2550E+06, 5.3349E+06, 5.4160E+06,
     + 5.4982E+06, 5.5816E+06, 5.6661E+06, 5.7518E+06, 5.8387E+06,
     +  5.9269E+06, 6.0162E+06, 6.1068E+06, 6.1987E+06, 6.2919E+06,
     + 6.3864E+06, 6.4821E+06, 6.5792E+06, 6.6777E+06, 6.7775E+06,
     +  6.8787E+06, 6.9813E+06, 7.0853E+06, 7.1908E+06, 7.2977E+06,
     + 7.4061E+06, 7.5160E+06, 7.6274E+06, 7.7403E+06, 7.8548E+06,
     +  7.9708E+06, 8.0885E+06, 8.2077E+06, 8.3286E+06, 8.4511E+06,
     + 8.5753E+06, 8.7012E+06, 8.8288E+06, 8.9581E+06, 9.0892E+06,
     +  9.2221E+06, 9.3567E+06, 9.4932E+06, 9.6316E+06, 9.7718E+06,
     + 9.9139E+06, 1.0058E+07, 1.0204E+07, 1.0352E+07, 1.0502E+07,
     +  1.0654E+07, 1.0808E+07, 1.0964E+07, 1.1122E+07, 1.1282E+07,
     + 1.1444E+07, 1.1609E+07, 1.1776E+07, 1.1944E+07, 1.2116E+07,
     +  1.2289E+07, 1.2465E+07, 1.2643E+07, 1.2823E+07, 1.3006E+07,
     + 1.3192E+07, 1.3379E+07, 1.3569E+07, 1.3762E+07, 1.3957E+07,
     +  1.4155E+07, 1.4356E+07, 1.4559E+07, 1.4764E+07, 1.4973E+07,
     + 1.5184E+07, 1.5398E+07, 1.5614E+07, 1.5834E+07, 1.6056E+07,
     +  1.6281E+07, 1.6510E+07, 1.6741E+07, 1.6975E+07, 1.7212E+07,
     + 1.7452E+07, 1.7696E+07, 1.7942E+07, 1.8192E+07, 1.8445E+07,
     +  1.8701E+07, 1.8961E+07, 1.9224E+07, 1.9490E+07, 1.9759E+07,
     + 2.0032E+07, 2.0309E+07, 2.0589E+07, 2.0873E+07, 2.1160E+07,
     +  2.1451E+07, 2.1746E+07, 2.2044E+07, 2.2347E+07, 2.2653E+07,
     + 2.2963E+07, 2.3277E+07, 2.3595E+07, 2.3917E+07, 2.4243E+07,
     +  2.4573E+07, 2.4907E+07, 2.5246E+07, 2.5589E+07, 2.5936E+07,
     + 2.6287E+07, 2.6643E+07, 2.7004E+07, 2.7369E+07, 2.7738E+07,
     +  2.8112E+07, 2.8491E+07, 2.8874E+07, 2.9263E+07, 2.9656E+07,
     + 3.0054E+07, 3.0457E+07, 3.0865E+07, 3.1278E+07, 3.1696E+07,
     +  3.2120E+07, 3.2549E+07, 3.2983E+07, 3.3422E+07, 3.3867E+07,
     + 3.4317E+07, 3.4773E+07, 3.5234E+07, 3.5701E+07, 3.6174E+07,
     +  3.6653E+07, 3.7138E+07, 3.7628E+07, 3.8125E+07, 3.8627E+07,
     + 3.9136E+07, 3.9651E+07, 4.0172E+07, 4.0700E+07, 4.1234E+07,
     +  4.1774E+07, 4.2321E+07, 4.2875E+07, 4.3436E+07, 4.4003E+07,
     + 4.4577E+07, 4.5158E+07, 4.5746E+07, 4.6342E+07, 4.6944E+07,
     +  4.7554E+07, 4.8171E+07, 4.8795E+07, 4.9427E+07, 5.0067E+07,
     + 5.0714E+07, 5.1369E+07, 5.2032E+07, 5.2703E+07, 5.3382E+07,
     +  5.4069E+07, 5.4764E+07, 5.5467E+07, 5.6179E+07, 5.6900E+07,
     + 5.7628E+07, 5.8366E+07, 5.9112E+07, 5.9867E+07, 6.0632E+07,
     +  6.1405E+07, 6.2187E+07, 6.2979E+07, 6.3780E+07, 6.4590E+07,
     + 6.5410E+07, 6.6240E+07, 6.7079E+07, 6.7928E+07, 6.8787E+07,
     +  6.9657E+07, 7.0536E+07, 7.1426E+07, 7.2326E+07, 7.3237E+07,
     + 7.4158E+07, 7.5091E+07, 7.6034E+07, 7.6988E+07, 7.7953E+07,
     +  7.8929E+07, 7.9917E+07, 8.0916E+07, 8.1927E+07, 8.2949E+07,
     + 8.3984E+07, 8.5030E+07, 8.6088E+07, 8.7159E+07, 8.8242E+07,
     +  8.9338E+07, 9.0446E+07, 9.1566E+07, 9.2700E+07, 9.3847E+07,
     + 9.5007E+07, 9.6180E+07, 9.7367E+07, 9.8567E+07, 9.9781E+07,
     +  1.0101E+08, 1.0225E+08, 1.0351E+08, 1.0478E+08, 1.0606E+08,
     + 1.0736E+08, 1.0868E+08, 1.1001E+08, 1.1135E+08, 1.1271E+08,
     +  1.1409E+08, 1.1548E+08, 1.1688E+08, 1.1830E+08, 1.1974E+08,
     + 1.2120E+08, 1.2267E+08, 1.2416E+08, 1.2566E+08, 1.2718E+08,
     +  1.2872E+08, 1.3027E+08, 1.3185E+08, 1.3344E+08, 1.3505E+08,
     + 1.3667E+08, 1.3832E+08, 1.3998E+08, 1.4166E+08, 1.4336E+08,
     +  1.4508E+08, 1.4682E+08, 1.4858E+08, 1.5036E+08, 1.5215E+08,
     + 1.5397E+08, 1.5581E+08, 1.5766E+08, 1.5954E+08, 1.6144E+08,
     +  1.6336E+08, 1.6530E+08, 1.6726E+08, 1.6925E+08, 1.7125E+08,
     + 1.7328E+08, 1.7533E+08, 1.7740E+08, 1.7950E+08, 1.8162E+08,
     +  1.8376E+08, 1.8593E+08, 1.8811E+08, 1.9033E+08, 1.9256E+08,
     + 1.9483E+08, 1.9711E+08, 1.9942E+08, 2.0176E+08, 2.0412E+08,
     +  2.0651E+08, 2.0892E+08, 2.1136E+08, 2.1382E+08, 2.1632E+08,
     + 2.1884E+08, 2.2138E+08, 2.2396E+08, 2.2656E+08, 2.2919E+08,
     +  2.3185E+08, 2.3453E+08, 2.3725E+08, 2.3999E+08, 2.4277E+08,
     + 2.4557E+08, 2.4841E+08, 2.5127E+08, 2.5417E+08, 2.5709E+08,
     +  2.6005E+08, 2.6304E+08, 2.6606E+08, 2.6912E+08, 2.7220E+08,
     + 2.7532E+08, 2.7847E+08, 2.8166E+08, 2.8488E+08, 2.8813E+08,
     +  2.9142E+08, 2.9474E+08, 2.9810E+08, 3.0149E+08, 3.0492E+08,
     + 3.0839E+08, 3.1189E+08, 3.1543E+08, 3.1901E+08, 3.2262E+08,
     +  3.2627E+08, 3.2996E+08, 3.3369E+08, 3.3746E+08, 3.4127E+08,
     + 3.4512E+08, 3.4901E+08, 3.5294E+08, 3.5691E+08, 3.6092E+08,
     +  3.6497E+08, 3.6907E+08, 3.7320E+08, 3.7739E+08, 3.8161E+08,
     + 3.8588E+08, 3.9019E+08, 3.9455E+08, 3.9895E+08, 4.0340E+08,
     +  4.0790E+08, 4.1244E+08, 4.1703E+08, 4.2166E+08, 4.2635E+08,
     + 4.3108E+08, 4.3586E+08, 4.4069E+08, 4.4557E+08, 4.5050E+08,
     +  4.5548E+08, 4.6051E+08, 4.6559E+08, 4.7073E+08, 4.7592E+08,
     + 4.8116E+08, 4.8645E+08, 4.9180E+08, 4.9720E+08, 5.0266E+08,
     +  5.0818E+08, 5.1375E+08, 5.1937E+08, 5.2506E+08, 5.3080E+08,
     + 5.3660E+08, 5.4246E+08, 5.4838E+08, 5.5436E+08, 5.6040E+08,
     +  5.6650E+08, 5.7266E+08, 5.7888E+08, 5.8517E+08, 5.9152E+08,
     + 5.9793E+08, 6.0441E+08, 6.1096E+08, 6.1757E+08, 6.2424E+08,
     +  6.3099E+08, 6.3780E+08, 6.4468E+08, 6.5163E+08, 6.5864E+08,
     + 6.6573E+08, 6.7289E+08, 6.8012E+08, 6.8742E+08, 6.9480E+08,
     +  7.0225E+08, 7.0977E+08, 7.1737E+08, 7.2504E+08, 7.3279E+08,
     + 7.4062E+08, 7.4852E+08, 7.5651E+08, 7.6457E+08, 7.7271E+08,
     +  7.8093E+08, 7.8924E+08, 7.9762E+08, 8.0609E+08, 8.1464E+08,
     + 8.2328E+08, 8.3200E+08, 8.4080E+08, 8.4970E+08, 8.5868E+08,
     +  8.6775E+08, 8.7690E+08, 8.8615E+08, 8.9549E+08, 9.0492E+08,
     + 9.1444E+08, 9.2406E+08, 9.3377E+08, 9.4357E+08, 9.5347E+08,
     +  9.6346E+08, 9.7356E+08, 9.8375E+08, 9.9404E+08, 1.0044E+09,
     + 1.0149E+09, 1.0255E+09, 1.0362E+09, 1.0470E+09, 1.0579E+09,
     +  1.0689E+09, 1.0800E+09, 1.0913E+09, 1.1026E+09, 1.1140E+09,
     + 1.1256E+09, 1.1372E+09, 1.1490E+09, 1.1609E+09, 1.1729E+09,
     +  1.1850E+09, 1.1973E+09, 1.2096E+09, 1.2221E+09, 1.2347E+09,
     + 1.2474E+09, 1.2602E+09, 1.2732E+09, 1.2862E+09, 1.2994E+09,
     +  1.3128E+09, 1.3262E+09, 1.3398E+09, 1.3535E+09, 1.3674E+09,
     + 1.3813E+09, 1.3954E+09, 1.4097E+09, 1.4240E+09, 1.4386E+09,
     +  1.4532E+09, 1.4680E+09, 1.4829E+09, 1.4980E+09, 1.5132E+09,
     + 1.5286E+09, 1.5441E+09, 1.5597E+09, 1.5755E+09, 1.5914E+09,
     +  1.6075E+09, 1.6238E+09, 1.6401E+09, 1.6567E+09, 1.6734E+09,
     + 1.6903E+09, 1.7073E+09, 1.7244E+09, 1.7418E+09, 1.7593E+09,
     +  1.7769E+09, 1.7948E+09, 1.8128E+09, 1.8309E+09, 1.8492E+09,
     + 1.8677E+09, 1.8864E+09, 1.9053E+09, 1.9243E+09, 1.9435E+09,
     +  1.9628E+09, 1.9824E+09, 2.0021E+09, 2.0221E+09, 2.0422E+09,
     + 2.0624E+09, 2.0829E+09, 2.1036E+09, 2.1244E+09, 2.1455E+09,
     +  2.1667E+09, 2.1881E+09, 2.2098E+09, 2.2316E+09, 2.2536E+09,
     + 2.2759E+09, 2.2983E+09, 2.3209E+09, 2.3438E+09, 2.3668E+09,
     +  2.3901E+09, 2.4136E+09, 2.4373E+09, 2.4612E+09, 2.4853E+09,
     + 2.5097E+09, 2.5342E+09, 2.5590E+09, 2.5840E+09, 2.6093E+09,
     +  2.6348E+09, 2.6605E+09, 2.6864E+09, 2.7126E+09, 2.7390E+09,
     + 2.7656E+09, 2.7925E+09, 2.8197E+09, 2.8470E+09, 2.8747E+09,
     +  2.9025E+09, 2.9307E+09, 2.9590E+09, 2.9877E+09, 3.0166E+09,
     + 3.0457E+09, 3.0751E+09, 3.1048E+09, 3.1347E+09, 3.1649E+09,
     +  3.1954E+09, 3.2262E+09, 3.2572E+09, 3.2885E+09, 3.3201E+09,
     + 3.3520E+09, 3.3841E+09, 3.4165E+09, 3.4493E+09, 3.4823E+09,
     +  3.5156E+09, 3.5492E+09, 3.5831E+09, 3.6173E+09, 3.6518E+09,
     + 3.6866E+09, 3.7218E+09, 3.7572E+09, 3.7929E+09, 3.8290E+09,
     +  3.8654E+09, 3.9021E+09, 3.9391E+09, 3.9765E+09, 4.0142E+09,
     + 4.0522E+09, 4.0905E+09, 4.1292E+09, 4.1682E+09, 4.2076E+09,
     +  4.2473E+09, 4.2874E+09, 4.3278E+09, 4.3686E+09, 4.4097E+09,
     + 4.4512E+09, 4.4930E+09, 4.5352E+09, 4.5778E+09, 4.6208E+09,
     +  4.6641E+09, 4.7078E+09, 4.7519E+09, 4.7963E+09, 4.8412E+09,
     + 4.8864E+09, 4.9321E+09, 4.9781E+09, 5.0245E+09, 5.0714E+09,
     +  5.1186E+09, 5.1663E+09, 5.2143E+09, 5.2628E+09, 5.3117E+09,
     + 5.3610E+09, 5.4107E+09, 5.4609E+09, 5.5115E+09, 5.5626E+09,
     +  5.6140E+09, 5.6660E+09, 5.7183E+09, 5.7711E+09, 5.8244E+09,
     + 5.8781E+09, 5.9323E+09, 5.9870E+09, 6.0421E+09, 6.0977E+09,
     +  6.1537E+09, 6.2103E+09, 6.2673E+09, 6.3248E+09, 6.3828E+09,
     + 6.4413E+09, 6.5003E+09, 6.5598E+09, 6.6198E+09, 6.6803E+09,
     +  6.7414E+09, 6.8029E+09, 6.8650E+09, 6.9276E+09, 6.9907E+09,
     + 7.0543E+09, 7.1185E+09, 7.1833E+09, 7.2486E+09, 7.3144E+09,
     +  7.3808E+09, 7.4477E+09, 7.5153E+09, 7.5834E+09, 7.6520E+09,
     + 7.7212E+09, 7.7911E+09, 7.8615E+09, 7.9325E+09, 8.0041E+09,
     +  8.0762E+09, 8.1490E+09, 8.2225E+09, 8.2965E+09, 8.3711E+09,
     + 8.4464E+09, 8.5223E+09, 8.5988E+09, 8.6760E+09, 8.7538E+09,
     +  8.8322E+09, 8.9113E+09, 8.9911E+09, 9.0715E+09, 9.1526E+09,
     + 9.2344E+09, 9.3168E+09, 9.4000E+09, 9.4838E+09, 9.5683E+09,
     +  9.6535E+09, 9.7395E+09, 9.8261E+09, 9.9135E+09, 1.0002E+10,
     + 1.0090E+10, 1.0180E+10, 1.0270E+10, 1.0361E+10, 1.0453E+10,
     +  1.0545E+10, 1.0639E+10, 1.0733E+10, 1.0828E+10, 1.0923E+10,
     + 1.1020E+10, 1.1117E+10, 1.1215E+10, 1.1313E+10, 1.1413E+10,
     +  1.1513E+10, 1.1615E+10, 1.1717E+10, 1.1820E+10, 1.1923E+10,
     + 1.2028E+10, 1.2133E+10, 1.2239E+10, 1.2347E+10, 1.2455E+10,
     +  1.2563E+10, 1.2673E+10, 1.2784E+10, 1.2895E+10, 1.3008E+10,
     + 1.3121E+10, 1.3236E+10, 1.3351E+10, 1.3467E+10, 1.3584E+10,
     +  1.3702E+10, 1.3821E+10, 1.3941E+10, 1.4062E+10, 1.4184E+10,
     + 1.4307E+10, 1.4430E+10, 1.4555E+10, 1.4681E+10, 1.4808E+10,
     +  1.4936E+10, 1.5065E+10, 1.5195E+10, 1.5326E+10, 1.5458E+10,
     + 1.5591E+10, 1.5725E+10, 1.5860E+10, 1.5996E+10, 1.6134E+10,
     +  1.6272E+10, 1.6412E+10, 1.6553E+10, 1.6694E+10, 1.6837E+10,
     + 1.6981E+10, 1.7127E+10, 1.7273E+10, 1.7421E+10, 1.7569E+10,
     +  1.7719E+10, 1.7870E+10, 1.8023E+10, 1.8176E+10, 1.8331E+10,
     + 1.8487E+10, 1.8644E+10, 1.8802E+10, 1.8962E+10, 1.9123E+10,
     +  1.9285E+10, 1.9449E+10, 1.9613E+10, 1.9779E+10, 1.9947E+10,
     + 2.0115E+10, 2.0285E+10, 2.0457E+10, 2.0629E+10, 2.0803E+10,
     +  2.0979E+10, 2.1156E+10, 2.1334E+10, 2.1513E+10, 2.1694E+10,
     + 2.1877E+10, 2.2060E+10, 2.2246E+10, 2.2432E+10, 2.2620E+10,
     +  2.2810E+10, 2.3001E+10, 2.3193E+10, 2.3388E+10, 2.3583E+10,
     + 2.3780E+10, 2.3979E+10, 2.4179E+10, 2.4381E+10, 2.4584E+10,
     +  2.4789E+10, 2.4995E+10, 2.5203E+10, 2.5413E+10, 2.5624E+10,
     + 2.5837E+10, 2.6051E+10, 2.6268E+10, 2.6485E+10, 2.6705E+10,
     +  2.6926E+10, 2.7149E+10, 2.7374E+10, 2.7600E+10, 2.7828E+10,
     + 2.8058E+10, 2.8289E+10, 2.8523E+10, 2.8758E+10, 2.8995E+10,
     +  2.9234E+10, 2.9474E+10, 2.9717E+10, 2.9961E+10, 3.0207E+10,
     + 3.0455E+10, 3.0705E+10, 3.0957E+10, 3.1211E+10, 3.1467E+10,
     +  3.1724E+10, 3.1984E+10, 3.2245E+10, 3.2509E+10, 3.2775E+10,
     + 3.3042E+10, 3.3312E+10, 3.3583E+10, 3.3857E+10, 3.4133E+10,
     +  3.4411E+10, 3.4691E+10, 3.4973E+10, 3.5257E+10, 3.5543E+10,
     + 3.5832E+10, 3.6123E+10, 3.6416E+10, 3.6711E+10, 3.7008E+10,
     +  3.7307E+10, 3.7609E+10, 3.7913E+10, 3.8220E+10, 3.8528E+10,
     + 3.8839E+10, 3.9153E+10, 3.9468E+10, 3.9786E+10, 4.0107E+10,
     +  4.0430E+10, 4.0755E+10, 4.1082E+10, 4.1413E+10, 4.1745E+10,
     + 4.2080E+10, 4.2418E+10, 4.2758E+10, 4.3100E+10, 4.3445E+10,
     +  4.3793E+10, 4.4143E+10, 4.4496E+10, 4.4852E+10, 4.5210E+10,
     + 4.5571E+10, 4.5934E+10, 4.6301E+10, 4.6669E+10, 4.7041E+10,
     +  4.7415E+10, 4.7793E+10, 4.8172E+10, 4.8555E+10, 4.8941E+10,
     + 4.9329E+10, 4.9720E+10, 5.0115E+10, 5.0512E+10, 5.0912E+10,
     +  5.1315E+10, 5.1720E+10, 5.2129E+10, 5.2541E+10, 5.2956E+10,
     + 5.3374E+10, 5.3795E+10, 5.4219E+10, 5.4646E+10, 5.5077E+10,
     +  5.5510E+10, 5.5947E+10, 5.6386E+10, 5.6829E+10, 5.7276E+10,
     + 5.7725E+10, 5.8178E+10, 5.8634E+10, 5.9094E+10, 5.9556E+10,
     +  6.0022E+10, 6.0492E+10, 6.0965E+10, 6.1441E+10, 6.1921E+10,
     + 6.2404E+10, 6.2891E+10, 6.3382E+10, 6.3876E+10, 6.4373E+10,
     +  6.4874E+10, 6.5379E+10, 6.5887E+10, 6.6399E+10, 6.6915E+10,
     + 6.7434E+10, 6.7957E+10, 6.8484E+10, 6.9015E+10, 6.9550E+10,
     +  7.0088E+10, 7.0630E+10, 7.1177E+10, 7.1727E+10, 7.2281E+10,
     + 7.2839E+10, 7.3401E+10, 7.3967E+10, 7.4537E+10, 7.5111E+10,
     +  7.5689E+10, 7.6272E+10, 7.6859E+10, 7.7449E+10, 7.8044E+10,
     + 7.8644E+10, 7.9247E+10, 7.9855E+10, 8.0467E+10, 8.1084E+10,
     +  8.1705E+10, 8.2330E+10, 8.2960E+10, 8.3594E+10, 8.4233E+10,
     + 8.4876E+10, 8.5524E+10, 8.6176E+10, 8.6834E+10, 8.7495E+10,
     +  8.8162E+10, 8.8833E+10, 8.9509E+10, 9.0189E+10, 9.0875E+10,
     + 9.1565E+10, 9.2260E+10, 9.2961E+10, 9.3666E+10, 9.4376E+10,
     +  9.5091E+10, 9.5811E+10, 9.6536E+10, 9.7266E+10, 9.8001E+10,
     + 9.8742E+10, 9.9487E+10, 1.0024E+11, 1.0099E+11, 1.0176E+11,
     +  1.0252E+11, 1.0330E+11, 1.0407E+11, 1.0486E+11, 1.0564E+11,
     + 1.0644E+11, 1.0724E+11, 1.0804E+11, 1.0885E+11, 1.0967E+11,
     +  1.1049E+11, 1.1132E+11, 1.1215E+11, 1.1299E+11, 1.1384E+11,
     + 1.1469E+11, 1.1555E+11, 1.1641E+11, 1.1728E+11, 1.1815E+11,
     +  1.1903E+11, 1.1992E+11, 1.2081E+11, 1.2171E+11, 1.2262E+11,
     + 1.2353E+11, 1.2445E+11, 1.2537E+11, 1.2631E+11, 1.2724E+11,
     +  1.2819E+11, 1.2914E+11, 1.3009E+11, 1.3106E+11, 1.3203E+11,
     + 1.3300E+11, 1.3399E+11, 1.3498E+11, 1.3597E+11, 1.3698E+11,
     +  1.3799E+11, 1.3901E+11, 1.4003E+11, 1.4106E+11, 1.4210E+11,
     + 1.4315E+11, 1.4420E+11, 1.4526E+11, 1.4633E+11, 1.4740E+11,
     +  1.4849E+11, 1.4958E+11, 1.5067E+11, 1.5178E+11, 1.5289E+11,
     + 1.5401E+11, 1.5513E+11, 1.5627E+11, 1.5741E+11, 1.5856E+11,
     +  1.5972E+11, 1.6089E+11, 1.6206E+11, 1.6324E+11, 1.6443E+11,
     + 1.6563E+11, 1.6683E+11, 1.6805E+11, 1.6927E+11, 1.7050E+11,
     +  1.7174E+11, 1.7299E+11, 1.7424E+11, 1.7551E+11, 1.7678E+11,
     + 1.7806E+11, 1.7935E+11, 1.8065E+11, 1.8196E+11, 1.8327E+11,
     +  1.8460E+11, 1.8593E+11, 1.8727E+11, 1.8863E+11, 1.8999E+11,
     + 1.9136E+11, 1.9273E+11, 1.9412E+11, 1.9552E+11, 1.9693E+11,
     +  1.9834E+11, 1.9977E+11, 2.0120E+11, 2.0265E+11, 2.0410E+11,
     + 2.0557E+11, 2.0704E+11, 2.0853E+11, 2.1002E+11, 2.1152E+11,
     +  2.1304E+11, 2.1456E+11, 2.1609E+11, 2.1764E+11, 2.1919E+11,
     + 2.2076E+11, 2.2233E+11, 2.2392E+11, 2.2551E+11, 2.2712E+11,
     +  2.2874E+11, 2.3036E+11, 2.3200E+11, 2.3365E+11, 2.3531E+11,
     + 2.3698E+11, 2.3867E+11, 2.4036E+11, 2.4206E+11, 2.4378E+11,
     +  2.4551E+11, 2.4724E+11, 2.4899E+11, 2.5076E+11, 2.5253E+11,
     + 2.5431E+11, 2.5611E+11, 2.5792E+11, 2.5974E+11, 2.6157E+11,
     +  2.6341E+11, 2.6527E+11, 2.6714E+11, 2.6902E+11, 2.7091E+11,
     + 2.7281E+11, 2.7473E+11, 2.7666E+11, 2.7860E+11, 2.8056E+11,
     +  2.8253E+11, 2.8451E+11, 2.8650E+11, 2.8851E+11, 2.9053E+11,
     + 2.9256E+11, 2.9461E+11, 2.9666E+11, 2.9874E+11, 3.0082E+11,
     +  3.0292E+11, 3.0504E+11, 3.0716E+11, 3.0930E+11, 3.1146E+11,
     + 3.1363E+11, 3.1581E+11, 3.1801E+11, 3.2022E+11, 3.2244E+11,
     +  3.2468E+11, 3.2694E+11, 3.2920E+11, 3.3149E+11, 3.3378E+11,
     + 3.3610E+11, 3.3842E+11, 3.4077E+11, 3.4312E+11, 3.4550E+11,
     +  3.4789E+11, 3.5029E+11, 3.5271E+11, 3.5514E+11, 3.5759E+11,
     + 3.6006E+11, 3.6254E+11, 3.6503E+11, 3.6755E+11, 3.7008E+11,
     +  3.7262E+11, 3.7518E+11, 3.7776E+11, 3.8036E+11, 3.8297E+11,
     + 3.8560E+11, 3.8824E+11, 3.9090E+11, 3.9358E+11, 3.9627E+11,
     +  3.9899E+11, 4.0172E+11, 4.0446E+11, 4.0723E+11, 4.1001E+11,
     + 4.1281E+11, 4.1563E+11, 4.1846E+11, 4.2131E+11, 4.2419E+11,
     +  4.2708E+11, 4.2998E+11, 4.3291E+11, 4.3585E+11, 4.3882E+11,
     + 4.4180E+11, 4.4480E+11, 4.4782E+11, 4.5086E+11, 4.5392E+11,
     +  4.5699E+11, 4.6009E+11, 4.6321E+11, 4.6634E+11, 4.6950E+11,
     + 4.7267E+11, 4.7587E+11, 4.7908E+11, 4.8232E+11, 4.8557E+11,
     +  4.8885E+11, 4.9215E+11, 4.9546E+11, 4.9880E+11, 5.0216E+11,
     + 5.0554E+11, 5.0894E+11, 5.1236E+11, 5.1581E+11, 5.1927E+11,
     +  5.2276E+11, 5.2627E+11, 5.2980E+11, 5.3335E+11, 5.3693E+11,
     + 5.4052E+11, 5.4414E+11, 5.4778E+11, 5.5145E+11, 5.5514E+11,
     +  5.5885E+11, 5.6258E+11, 5.6634E+11, 5.7012E+11, 5.7392E+11,
     + 5.7774E+11, 5.8160E+11, 5.8547E+11, 5.8937E+11, 5.9329E+11,
     +  5.9724E+11, 6.0121E+11, 6.0520E+11, 6.0922E+11, 6.1327E+11,
     + 6.1734E+11, 6.2143E+11, 6.2555E+11, 6.2970E+11, 6.3387E+11,
     +  6.3807E+11, 6.4229E+11, 6.4654E+11, 6.5081E+11, 6.5511E+11,
     + 6.5944E+11, 6.6379E+11, 6.6817E+11, 6.7258E+11, 6.7701E+11,
     +  6.8148E+11, 6.8597E+11, 6.9048E+11, 6.9503E+11, 6.9960E+11,
     + 7.0420E+11, 7.0882E+11, 7.1348E+11, 7.1816E+11, 7.2288E+11,
     +  7.2762E+11, 7.3239E+11, 7.3719E+11, 7.4202E+11, 7.4688E+11,
     + 7.5176E+11, 7.5668E+11, 7.6163E+11, 7.6661E+11, 7.7161E+11,
     +  7.7665E+11, 7.8172E+11, 7.8682E+11, 7.9195E+11, 7.9711E+11,
     + 8.0230E+11, 8.0753E+11, 8.1278E+11, 8.1807E+11, 8.2339E+11,
     +  8.2874E+11, 8.3412E+11, 8.3954E+11, 8.4499E+11, 8.5047E+11,
     + 8.5599E+11, 8.6153E+11, 8.6711E+11, 8.7273E+11, 8.7838E+11,
     +  8.8406E+11, 8.8978E+11, 8.9553E+11, 9.0131E+11, 9.0714E+11,
     + 9.1299E+11, 9.1888E+11, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_HC3N  (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 6),QofT( 6,NT),Q(NT)
      data (Tmx(I),I=1,6)/3500.,3500.,3500.,3500.,3500.,3500./
      data xgj/ 6.,12.,12.,12.,4.,9./
c...     HC3N
c...        --     12224
      data (QofT( 1,J),J=1,2408)/ 2.9577E+01, 5.7016E+01, 8.4488E+01,
     +  1.1197E+02, 1.3945E+02, 1.6694E+02, 1.9442E+02, 2.2191E+02,
     + 2.4940E+02, 2.7689E+02, 3.0438E+02, 3.3187E+02, 3.5937E+02,
     +  3.8686E+02, 4.1435E+02, 4.4185E+02, 4.6934E+02, 4.9684E+02,
     + 5.2433E+02, 5.5183E+02, 6.0682E+02, 6.6182E+02, 7.1683E+02,
     +  7.7184E+02, 8.2687E+02, 8.8192E+02, 9.3700E+02, 9.9214E+02,
     + 1.0473E+03, 1.1026E+03, 1.1580E+03, 1.2136E+03, 1.2694E+03,
     +  1.3254E+03, 1.3816E+03, 1.4382E+03, 1.4951E+03, 1.5523E+03,
     + 1.6101E+03, 1.6683E+03, 1.7270E+03, 1.7863E+03, 1.8462E+03,
     +  1.9068E+03, 1.9681E+03, 2.0301E+03, 2.0928E+03, 2.1564E+03,
     + 2.2209E+03, 2.2863E+03, 2.3526E+03, 2.4198E+03, 2.4881E+03,
     +  2.5574E+03, 2.6279E+03, 2.6994E+03, 2.7721E+03, 2.8460E+03,
     + 2.9211E+03, 2.9975E+03, 3.0751E+03, 3.1541E+03, 3.2345E+03,
     +  3.3163E+03, 3.3995E+03, 3.4842E+03, 3.5704E+03, 3.6582E+03,
     + 3.7476E+03, 3.8386E+03, 3.9312E+03, 4.0256E+03, 4.1217E+03,
     +  4.2196E+03, 4.3193E+03, 4.4209E+03, 4.5244E+03, 4.6298E+03,
     + 4.7373E+03, 4.8467E+03, 4.9583E+03, 5.0720E+03, 5.1878E+03,
     +  5.3059E+03, 5.4263E+03, 5.5489E+03, 5.6740E+03, 5.8014E+03,
     + 5.9313E+03, 6.0637E+03, 6.1987E+03, 6.3363E+03, 6.4765E+03,
     +  6.6195E+03, 6.7653E+03, 6.9140E+03, 7.0655E+03, 7.2200E+03,
     + 7.3775E+03, 7.5380E+03, 7.7017E+03, 7.8686E+03, 8.0388E+03,
     +  8.2123E+03, 8.3891E+03, 8.5695E+03, 8.7533E+03, 8.9408E+03,
     + 9.1319E+03, 9.3268E+03, 9.5255E+03, 9.7280E+03, 9.9345E+03,
     +  1.0145E+04, 1.0360E+04, 1.0578E+04, 1.0802E+04, 1.1029E+04,
     + 1.1261E+04, 1.1497E+04, 1.1738E+04, 1.1984E+04, 1.2234E+04,
     +  1.2489E+04, 1.2750E+04, 1.3015E+04, 1.3285E+04, 1.3561E+04,
     + 1.3842E+04, 1.4128E+04, 1.4420E+04, 1.4717E+04, 1.5020E+04,
     +  1.5329E+04, 1.5644E+04, 1.5965E+04, 1.6292E+04, 1.6626E+04,
     + 1.6965E+04, 1.7311E+04, 1.7664E+04, 1.8023E+04, 1.8390E+04,
     +  1.8763E+04, 1.9143E+04, 1.9530E+04, 1.9925E+04, 2.0327E+04,
     + 2.0737E+04, 2.1154E+04, 2.1580E+04, 2.2013E+04, 2.2454E+04,
     +  2.2904E+04, 2.3361E+04, 2.3828E+04, 2.4303E+04, 2.4787E+04,
     + 2.5280E+04, 2.5782E+04, 2.6293E+04, 2.6814E+04, 2.7344E+04,
     +  2.7884E+04, 2.8434E+04, 2.8994E+04, 2.9564E+04, 3.0145E+04,
     + 3.0736E+04, 3.1338E+04, 3.1951E+04, 3.2575E+04, 3.3211E+04,
     +  3.3857E+04, 3.4516E+04, 3.5186E+04, 3.5869E+04, 3.6564E+04,
     + 3.7271E+04, 3.7991E+04, 3.8723E+04, 3.9469E+04, 4.0228E+04,
     +  4.1001E+04, 4.1787E+04, 4.2587E+04, 4.3402E+04, 4.4230E+04,
     + 4.5074E+04, 4.5932E+04, 4.6805E+04, 4.7694E+04, 4.8598E+04,
     +  4.9518E+04, 5.0454E+04, 5.1406E+04, 5.2375E+04, 5.3361E+04,
     + 5.4363E+04, 5.5383E+04, 5.6421E+04, 5.7477E+04, 5.8550E+04,
     +  5.9642E+04, 6.0753E+04, 6.1883E+04, 6.3033E+04, 6.4201E+04,
     + 6.5390E+04, 6.6599E+04, 6.7829E+04, 6.9079E+04, 7.0350E+04,
     +  7.1643E+04, 7.2958E+04, 7.4295E+04, 7.5654E+04, 7.7036E+04,
     + 7.8441E+04, 7.9870E+04, 8.1322E+04, 8.2799E+04, 8.4300E+04,
     +  8.5827E+04, 8.7378E+04, 8.8955E+04, 9.0558E+04, 9.2188E+04,
     + 9.3844E+04, 9.5528E+04, 9.7239E+04, 9.8978E+04, 1.0075E+05,
     +  1.0254E+05, 1.0437E+05, 1.0622E+05, 1.0811E+05, 1.1003E+05,
     + 1.1197E+05, 1.1395E+05, 1.1596E+05, 1.1800E+05, 1.2008E+05,
     +  1.2219E+05, 1.2433E+05, 1.2651E+05, 1.2872E+05, 1.3097E+05,
     + 1.3325E+05, 1.3557E+05, 1.3792E+05, 1.4032E+05, 1.4275E+05,
     +  1.4522E+05, 1.4773E+05, 1.5027E+05, 1.5286E+05, 1.5549E+05,
     + 1.5816E+05, 1.6087E+05, 1.6362E+05, 1.6642E+05, 1.6926E+05,
     +  1.7214E+05, 1.7507E+05, 1.7804E+05, 1.8106E+05, 1.8413E+05,
     + 1.8724E+05, 1.9040E+05, 1.9361E+05, 1.9687E+05, 2.0018E+05,
     +  2.0354E+05, 2.0695E+05, 2.1041E+05, 2.1393E+05, 2.1750E+05,
     + 2.2112E+05, 2.2479E+05, 2.2853E+05, 2.3232E+05, 2.3616E+05,
     +  2.4007E+05, 2.4403E+05, 2.4805E+05, 2.5213E+05, 2.5628E+05,
     + 2.6048E+05, 2.6475E+05, 2.6908E+05, 2.7347E+05, 2.7793E+05,
     +  2.8246E+05, 2.8705E+05, 2.9171E+05, 2.9644E+05, 3.0124E+05,
     + 3.0611E+05, 3.1105E+05, 3.1607E+05, 3.2115E+05, 3.2631E+05,
     +  3.3155E+05, 3.3686E+05, 3.4225E+05, 3.4772E+05, 3.5327E+05,
     + 3.5889E+05, 3.6460E+05, 3.7039E+05, 3.7627E+05, 3.8222E+05,
     +  3.8827E+05, 3.9440E+05, 4.0062E+05, 4.0693E+05, 4.1332E+05,
     + 4.1981E+05, 4.2639E+05, 4.3307E+05, 4.3983E+05, 4.4670E+05,
     +  4.5366E+05, 4.6072E+05, 4.6788E+05, 4.7514E+05, 4.8250E+05,
     + 4.8997E+05, 4.9754E+05, 5.0522E+05, 5.1300E+05, 5.2089E+05,
     +  5.2890E+05, 5.3701E+05, 5.4524E+05, 5.5358E+05, 5.6203E+05,
     + 5.7061E+05, 5.7930E+05, 5.8811E+05, 5.9704E+05, 6.0610E+05,
     +  6.1528E+05, 6.2459E+05, 6.3402E+05, 6.4358E+05, 6.5328E+05,
     + 6.6310E+05, 6.7306E+05, 6.8316E+05, 6.9339E+05, 7.0376E+05,
     +  7.1428E+05, 7.2493E+05, 7.3573E+05, 7.4667E+05, 7.5776E+05,
     + 7.6900E+05, 7.8039E+05, 7.9194E+05, 8.0364E+05, 8.1549E+05,
     +  8.2751E+05, 8.3968E+05, 8.5202E+05, 8.6452E+05, 8.7718E+05,
     + 8.9002E+05, 9.0302E+05, 9.1620E+05, 9.2955E+05, 9.4308E+05,
     +  9.5678E+05, 9.7067E+05, 9.8474E+05, 9.9899E+05, 1.0134E+06,
     + 1.0281E+06, 1.0429E+06, 1.0579E+06, 1.0731E+06, 1.0885E+06,
     +  1.1041E+06, 1.1199E+06, 1.1359E+06, 1.1522E+06, 1.1686E+06,
     + 1.1852E+06, 1.2021E+06, 1.2192E+06, 1.2365E+06, 1.2540E+06,
     +  1.2717E+06, 1.2897E+06, 1.3079E+06, 1.3263E+06, 1.3450E+06,
     + 1.3639E+06, 1.3831E+06, 1.4025E+06, 1.4221E+06, 1.4420E+06,
     +  1.4621E+06, 1.4825E+06, 1.5032E+06, 1.5241E+06, 1.5452E+06,
     + 1.5667E+06, 1.5884E+06, 1.6104E+06, 1.6326E+06, 1.6552E+06,
     +  1.6780E+06, 1.7011E+06, 1.7245E+06, 1.7482E+06, 1.7721E+06,
     + 1.7964E+06, 1.8210E+06, 1.8459E+06, 1.8711E+06, 1.8966E+06,
     +  1.9224E+06, 1.9485E+06, 1.9750E+06, 2.0018E+06, 2.0289E+06,
     + 2.0563E+06, 2.0841E+06, 2.1122E+06, 2.1407E+06, 2.1695E+06,
     +  2.1987E+06, 2.2282E+06, 2.2581E+06, 2.2883E+06, 2.3189E+06,
     + 2.3499E+06, 2.3812E+06, 2.4130E+06, 2.4451E+06, 2.4776E+06,
     +  2.5105E+06, 2.5438E+06, 2.5775E+06, 2.6116E+06, 2.6461E+06,
     + 2.6810E+06, 2.7163E+06, 2.7521E+06, 2.7883E+06, 2.8249E+06,
     +  2.8619E+06, 2.8994E+06, 2.9374E+06, 2.9757E+06, 3.0146E+06,
     + 3.0539E+06, 3.0936E+06, 3.1339E+06, 3.1746E+06, 3.2158E+06,
     +  3.2574E+06, 3.2996E+06, 3.3423E+06, 3.3854E+06, 3.4291E+06,
     + 3.4732E+06, 3.5179E+06, 3.5631E+06, 3.6089E+06, 3.6551E+06,
     +  3.7019E+06, 3.7493E+06, 3.7972E+06, 3.8456E+06, 3.8947E+06,
     + 3.9442E+06, 3.9944E+06, 4.0451E+06, 4.0964E+06, 4.1483E+06,
     +  4.2008E+06, 4.2539E+06, 4.3077E+06, 4.3620E+06, 4.4169E+06,
     + 4.4725E+06, 4.5287E+06, 4.5856E+06, 4.6431E+06, 4.7012E+06,
     +  4.7600E+06, 4.8195E+06, 4.8797E+06, 4.9405E+06, 5.0020E+06,
     + 5.0642E+06, 5.1272E+06, 5.1908E+06, 5.2551E+06, 5.3202E+06,
     +  5.3860E+06, 5.4525E+06, 5.5198E+06, 5.5878E+06, 5.6566E+06,
     + 5.7262E+06, 5.7965E+06, 5.8677E+06, 5.9396E+06, 6.0123E+06,
     +  6.0858E+06, 6.1602E+06, 6.2353E+06, 6.3113E+06, 6.3881E+06,
     + 6.4658E+06, 6.5444E+06, 6.6238E+06, 6.7041E+06, 6.7852E+06,
     +  6.8673E+06, 6.9502E+06, 7.0341E+06, 7.1189E+06, 7.2046E+06,
     + 7.2912E+06, 7.3788E+06, 7.4674E+06, 7.5569E+06, 7.6474E+06,
     +  7.7388E+06, 7.8313E+06, 7.9247E+06, 8.0192E+06, 8.1147E+06,
     + 8.2112E+06, 8.3088E+06, 8.4074E+06, 8.5071E+06, 8.6079E+06,
     +  8.7097E+06, 8.8126E+06, 8.9167E+06, 9.0218E+06, 9.1281E+06,
     + 9.2355E+06, 9.3441E+06, 9.4538E+06, 9.5647E+06, 9.6768E+06,
     +  9.7900E+06, 9.9045E+06, 1.0020E+07, 1.0137E+07, 1.0255E+07,
     + 1.0375E+07, 1.0495E+07, 1.0617E+07, 1.0740E+07, 1.0865E+07,
     +  1.0991E+07, 1.1118E+07, 1.1246E+07, 1.1376E+07, 1.1507E+07,
     + 1.1640E+07, 1.1774E+07, 1.1909E+07, 1.2046E+07, 1.2184E+07,
     +  1.2324E+07, 1.2465E+07, 1.2607E+07, 1.2752E+07, 1.2897E+07,
     + 1.3044E+07, 1.3193E+07, 1.3343E+07, 1.3494E+07, 1.3648E+07,
     +  1.3802E+07, 1.3959E+07, 1.4117E+07, 1.4276E+07, 1.4438E+07,
     + 1.4600E+07, 1.4765E+07, 1.4931E+07, 1.5099E+07, 1.5269E+07,
     +  1.5440E+07, 1.5613E+07, 1.5788E+07, 1.5965E+07, 1.6144E+07,
     + 1.6324E+07, 1.6506E+07, 1.6690E+07, 1.6876E+07, 1.7063E+07,
     +  1.7253E+07, 1.7445E+07, 1.7638E+07, 1.7833E+07, 1.8031E+07,
     + 1.8230E+07, 1.8432E+07, 1.8635E+07, 1.8840E+07, 1.9048E+07,
     +  1.9257E+07, 1.9469E+07, 1.9683E+07, 1.9899E+07, 2.0117E+07,
     + 2.0337E+07, 2.0559E+07, 2.0784E+07, 2.1011E+07, 2.1240E+07,
     +  2.1471E+07, 2.1705E+07, 2.1941E+07, 2.2179E+07, 2.2420E+07,
     + 2.2663E+07, 2.2908E+07, 2.3156E+07, 2.3406E+07, 2.3659E+07,
     +  2.3914E+07, 2.4172E+07, 2.4432E+07, 2.4695E+07, 2.4960E+07,
     + 2.5228E+07, 2.5498E+07, 2.5772E+07, 2.6047E+07, 2.6326E+07,
     +  2.6607E+07, 2.6891E+07, 2.7178E+07, 2.7467E+07, 2.7759E+07,
     + 2.8054E+07, 2.8352E+07, 2.8653E+07, 2.8957E+07, 2.9264E+07,
     +  2.9573E+07, 2.9886E+07, 3.0201E+07, 3.0520E+07, 3.0842E+07,
     + 3.1166E+07, 3.1494E+07, 3.1825E+07, 3.2159E+07, 3.2497E+07,
     +  3.2837E+07, 3.3181E+07, 3.3528E+07, 3.3878E+07, 3.4232E+07,
     + 3.4589E+07, 3.4949E+07, 3.5313E+07, 3.5680E+07, 3.6051E+07,
     +  3.6425E+07, 3.6803E+07, 3.7185E+07, 3.7569E+07, 3.7958E+07,
     + 3.8350E+07, 3.8746E+07, 3.9146E+07, 3.9549E+07, 3.9956E+07,
     +  4.0367E+07, 4.0782E+07, 4.1201E+07, 4.1623E+07, 4.2050E+07,
     + 4.2480E+07, 4.2914E+07, 4.3353E+07, 4.3796E+07, 4.4242E+07,
     +  4.4693E+07, 4.5148E+07, 4.5607E+07, 4.6071E+07, 4.6538E+07,
     + 4.7010E+07, 4.7487E+07, 4.7968E+07, 4.8453E+07, 4.8943E+07,
     +  4.9437E+07, 4.9935E+07, 5.0439E+07, 5.0947E+07, 5.1459E+07,
     + 5.1976E+07, 5.2498E+07, 5.3025E+07, 5.3557E+07, 5.4093E+07,
     +  5.4634E+07, 5.5180E+07, 5.5732E+07, 5.6288E+07, 5.6849E+07,
     + 5.7415E+07, 5.7987E+07, 5.8563E+07, 5.9145E+07, 5.9732E+07,
     +  6.0324E+07, 6.0922E+07, 6.1525E+07, 6.2134E+07, 6.2748E+07,
     + 6.3367E+07, 6.3992E+07, 6.4623E+07, 6.5259E+07, 6.5901E+07,
     +  6.6549E+07, 6.7202E+07, 6.7862E+07, 6.8527E+07, 6.9198E+07,
     + 6.9875E+07, 7.0558E+07, 7.1248E+07, 7.1943E+07, 7.2645E+07,
     +  7.3352E+07, 7.4066E+07, 7.4787E+07, 7.5513E+07, 7.6247E+07,
     + 7.6986E+07, 7.7732E+07, 7.8485E+07, 7.9244E+07, 8.0010E+07,
     +  8.0783E+07, 8.1563E+07, 8.2349E+07, 8.3142E+07, 8.3943E+07,
     + 8.4750E+07, 8.5564E+07, 8.6385E+07, 8.7214E+07, 8.8050E+07,
     +  8.8893E+07, 8.9743E+07, 9.0601E+07, 9.1466E+07, 9.2339E+07,
     + 9.3219E+07, 9.4107E+07, 9.5003E+07, 9.5907E+07, 9.6818E+07,
     +  9.7737E+07, 9.8664E+07, 9.9599E+07, 1.0054E+08, 1.0149E+08,
     + 1.0245E+08, 1.0342E+08, 1.0440E+08, 1.0538E+08, 1.0637E+08,
     +  1.0737E+08, 1.0838E+08, 1.0940E+08, 1.1043E+08, 1.1147E+08,
     + 1.1251E+08, 1.1356E+08, 1.1463E+08, 1.1570E+08, 1.1678E+08,
     +  1.1787E+08, 1.1897E+08, 1.2008E+08, 1.2120E+08, 1.2232E+08,
     + 1.2346E+08, 1.2461E+08, 1.2576E+08, 1.2693E+08, 1.2811E+08,
     +  1.2929E+08, 1.3049E+08, 1.3169E+08, 1.3291E+08, 1.3414E+08,
     + 1.3537E+08, 1.3662E+08, 1.3788E+08, 1.3914E+08, 1.4042E+08,
     +  1.4171E+08, 1.4301E+08, 1.4432E+08, 1.4564E+08, 1.4698E+08,
     + 1.4832E+08, 1.4967E+08, 1.5104E+08, 1.5242E+08, 1.5381E+08,
     +  1.5521E+08, 1.5662E+08, 1.5804E+08, 1.5948E+08, 1.6093E+08,
     + 1.6239E+08, 1.6386E+08, 1.6534E+08, 1.6684E+08, 1.6834E+08,
     +  1.6986E+08, 1.7140E+08, 1.7294E+08, 1.7450E+08, 1.7607E+08,
     + 1.7766E+08, 1.7925E+08, 1.8086E+08, 1.8249E+08, 1.8412E+08,
     +  1.8577E+08, 1.8743E+08, 1.8911E+08, 1.9080E+08, 1.9250E+08,
     + 1.9422E+08, 1.9595E+08, 1.9770E+08, 1.9946E+08, 2.0123E+08,
     +  2.0302E+08, 2.0483E+08, 2.0664E+08, 2.0847E+08, 2.1032E+08,
     + 2.1218E+08, 2.1406E+08, 2.1595E+08, 2.1786E+08, 2.1978E+08,
     +  2.2172E+08, 2.2367E+08, 2.2564E+08, 2.2763E+08, 2.2963E+08,
     + 2.3164E+08, 2.3368E+08, 2.3573E+08, 2.3779E+08, 2.3987E+08,
     +  2.4197E+08, 2.4409E+08, 2.4622E+08, 2.4837E+08, 2.5053E+08,
     + 2.5272E+08, 2.5492E+08, 2.5714E+08, 2.5937E+08, 2.6162E+08,
     +  2.6390E+08, 2.6618E+08, 2.6849E+08, 2.7082E+08, 2.7316E+08,
     + 2.7552E+08, 2.7790E+08, 2.8030E+08, 2.8272E+08, 2.8516E+08,
     +  2.8761E+08, 2.9009E+08, 2.9258E+08, 2.9510E+08, 2.9763E+08,
     + 3.0018E+08, 3.0276E+08, 3.0535E+08, 3.0797E+08, 3.1060E+08,
     +  3.1325E+08, 3.1593E+08, 3.1863E+08, 3.2134E+08, 3.2408E+08,
     + 3.2684E+08, 3.2962E+08, 3.3242E+08, 3.3525E+08, 3.3809E+08,
     +  3.4096E+08, 3.4385E+08, 3.4676E+08, 3.4970E+08, 3.5265E+08,
     + 3.5563E+08, 3.5864E+08, 3.6166E+08, 3.6471E+08, 3.6779E+08,
     +  3.7088E+08, 3.7400E+08, 3.7715E+08, 3.8031E+08, 3.8350E+08,
     + 3.8672E+08, 3.8996E+08, 3.9323E+08, 3.9652E+08, 3.9983E+08,
     +  4.0317E+08, 4.0654E+08, 4.0993E+08, 4.1335E+08, 4.1679E+08,
     + 4.2026E+08, 4.2376E+08, 4.2728E+08, 4.3083E+08, 4.3440E+08,
     +  4.3801E+08, 4.4164E+08, 4.4529E+08, 4.4898E+08, 4.5269E+08,
     + 4.5643E+08, 4.6020E+08, 4.6400E+08, 4.6782E+08, 4.7168E+08,
     +  4.7556E+08, 4.7947E+08, 4.8341E+08, 4.8738E+08, 4.9138E+08,
     + 4.9541E+08, 4.9947E+08, 5.0356E+08, 5.0769E+08, 5.1184E+08,
     +  5.1602E+08, 5.2023E+08, 5.2448E+08, 5.2876E+08, 5.3306E+08,
     + 5.3740E+08, 5.4178E+08, 5.4618E+08, 5.5062E+08, 5.5509E+08,
     +  5.5959E+08, 5.6413E+08, 5.6870E+08, 5.7330E+08, 5.7794E+08,
     + 5.8261E+08, 5.8732E+08, 5.9206E+08, 5.9683E+08, 6.0164E+08,
     +  6.0649E+08, 6.1137E+08, 6.1629E+08, 6.2124E+08, 6.2623E+08,
     + 6.3125E+08, 6.3632E+08, 6.4142E+08, 6.4655E+08, 6.5173E+08,
     +  6.5694E+08, 6.6219E+08, 6.6747E+08, 6.7280E+08, 6.7816E+08,
     + 6.8357E+08, 6.8901E+08, 6.9449E+08, 7.0001E+08, 7.0558E+08,
     +  7.1118E+08, 7.1682E+08, 7.2250E+08, 7.2823E+08, 7.3399E+08,
     + 7.3980E+08, 7.4565E+08, 7.5154E+08, 7.5747E+08, 7.6345E+08,
     +  7.6947E+08, 7.7553E+08, 7.8163E+08, 7.8778E+08, 7.9398E+08,
     + 8.0021E+08, 8.0649E+08, 8.1282E+08, 8.1919E+08, 8.2561E+08,
     +  8.3207E+08, 8.3858E+08, 8.4514E+08, 8.5174E+08, 8.5839E+08,
     + 8.6508E+08, 8.7183E+08, 8.7862E+08, 8.8546E+08, 8.9235E+08,
     +  8.9928E+08, 9.0627E+08, 9.1331E+08, 9.2039E+08, 9.2753E+08,
     + 9.3471E+08, 9.4195E+08, 9.4923E+08, 9.5657E+08, 9.6396E+08,
     +  9.7140E+08, 9.7890E+08, 9.8644E+08, 9.9404E+08, 1.0017E+09,
     + 1.0094E+09, 1.0172E+09, 1.0250E+09, 1.0328E+09, 1.0408E+09,
     +  1.0488E+09, 1.0568E+09, 1.0649E+09, 1.0730E+09, 1.0812E+09,
     + 1.0895E+09, 1.0978E+09, 1.1062E+09, 1.1146E+09, 1.1231E+09,
     +  1.1317E+09, 1.1403E+09, 1.1490E+09, 1.1577E+09, 1.1665E+09,
     + 1.1753E+09, 1.1842E+09, 1.1932E+09, 1.2023E+09, 1.2113E+09,
     +  1.2205E+09, 1.2297E+09, 1.2390E+09, 1.2484E+09, 1.2578E+09,
     + 1.2673E+09, 1.2768E+09, 1.2864E+09, 1.2961E+09, 1.3059E+09,
     +  1.3157E+09, 1.3255E+09, 1.3355E+09, 1.3455E+09, 1.3556E+09,
     + 1.3657E+09, 1.3759E+09, 1.3862E+09, 1.3966E+09, 1.4070E+09,
     +  1.4175E+09, 1.4281E+09, 1.4387E+09, 1.4494E+09, 1.4602E+09,
     + 1.4711E+09, 1.4820E+09, 1.4930E+09, 1.5041E+09, 1.5153E+09,
     +  1.5265E+09, 1.5378E+09, 1.5492E+09, 1.5607E+09, 1.5722E+09,
     + 1.5838E+09, 1.5955E+09, 1.6073E+09, 1.6191E+09, 1.6311E+09,
     +  1.6431E+09, 1.6552E+09, 1.6673E+09, 1.6796E+09, 1.6919E+09,
     + 1.7043E+09, 1.7168E+09, 1.7294E+09, 1.7421E+09, 1.7548E+09,
     +  1.7677E+09, 1.7806E+09, 1.7936E+09, 1.8067E+09, 1.8199E+09,
     + 1.8332E+09, 1.8465E+09, 1.8600E+09, 1.8735E+09, 1.8871E+09,
     +  1.9008E+09, 1.9146E+09, 1.9285E+09, 1.9425E+09, 1.9566E+09,
     + 1.9708E+09, 1.9851E+09, 1.9994E+09, 2.0139E+09, 2.0284E+09,
     +  2.0431E+09, 2.0578E+09, 2.0727E+09, 2.0876E+09, 2.1026E+09,
     + 2.1178E+09, 2.1330E+09, 2.1483E+09, 2.1638E+09, 2.1793E+09,
     +  2.1949E+09, 2.2107E+09, 2.2265E+09, 2.2425E+09, 2.2585E+09,
     + 2.2747E+09, 2.2909E+09, 2.3073E+09, 2.3237E+09, 2.3403E+09,
     +  2.3570E+09, 2.3738E+09, 2.3907E+09, 2.4077E+09, 2.4248E+09,
     + 2.4421E+09, 2.4594E+09, 2.4768E+09, 2.4944E+09, 2.5121E+09,
     +  2.5299E+09, 2.5478E+09, 2.5658E+09, 2.5840E+09, 2.6022E+09,
     + 2.6206E+09, 2.6391E+09, 2.6577E+09, 2.6764E+09, 2.6953E+09,
     +  2.7142E+09, 2.7333E+09, 2.7526E+09, 2.7719E+09, 2.7914E+09,
     + 2.8109E+09, 2.8307E+09, 2.8505E+09, 2.8705E+09, 2.8905E+09,
     +  2.9108E+09, 2.9311E+09, 2.9516E+09, 2.9722E+09, 2.9929E+09,
     + 3.0138E+09, 3.0348E+09, 3.0559E+09, 3.0772E+09, 3.0986E+09,
     +  3.1201E+09, 3.1418E+09, 3.1636E+09, 3.1856E+09, 3.2077E+09,
     + 3.2299E+09, 3.2523E+09, 3.2748E+09, 3.2974E+09, 3.3202E+09,
     +  3.3432E+09, 3.3662E+09, 3.3895E+09, 3.4128E+09, 3.4363E+09,
     + 3.4600E+09, 3.4838E+09, 3.5078E+09, 3.5319E+09, 3.5562E+09,
     +  3.5806E+09, 3.6051E+09, 3.6299E+09, 3.6547E+09, 3.6798E+09,
     + 3.7050E+09, 3.7303E+09, 3.7558E+09, 3.7815E+09, 3.8073E+09,
     +  3.8333E+09, 3.8594E+09, 3.8857E+09, 3.9122E+09, 3.9388E+09,
     + 3.9656E+09, 3.9926E+09, 4.0197E+09, 4.0470E+09, 4.0744E+09,
     +  4.1021E+09, 4.1299E+09, 4.1579E+09, 4.1860E+09, 4.2143E+09,
     + 4.2428E+09, 4.2715E+09, 4.3003E+09, 4.3294E+09, 4.3586E+09,
     +  4.3879E+09, 4.4175E+09, 4.4473E+09, 4.4772E+09, 4.5073E+09,
     + 4.5376E+09, 4.5681E+09, 4.5987E+09, 4.6296E+09, 4.6606E+09,
     +  4.6918E+09, 4.7233E+09, 4.7549E+09, 4.7867E+09, 4.8187E+09,
     + 4.8509E+09, 4.8832E+09, 4.9158E+09, 4.9486E+09, 4.9816E+09,
     +  5.0148E+09, 5.0482E+09, 5.0817E+09, 5.1155E+09, 5.1495E+09,
     + 5.1837E+09, 5.2181E+09, 5.2527E+09, 5.2875E+09, 5.3226E+09,
     +  5.3578E+09, 5.3933E+09, 5.4289E+09, 5.4648E+09, 5.5009E+09,
     + 5.5372E+09, 5.5738E+09, 5.6105E+09, 5.6475E+09, 5.6847E+09,
     +  5.7221E+09, 5.7597E+09, 5.7976E+09, 5.8357E+09, 5.8740E+09,
     + 5.9126E+09, 5.9513E+09, 5.9904E+09, 6.0296E+09, 6.0691E+09,
     +  6.1088E+09, 6.1487E+09, 6.1889E+09, 6.2294E+09, 6.2700E+09,
     + 6.3109E+09, 6.3521E+09, 6.3935E+09, 6.4351E+09, 6.4770E+09,
     +  6.5191E+09, 6.5615E+09, 6.6041E+09, 6.6470E+09, 6.6902E+09,
     + 6.7336E+09, 6.7772E+09, 6.8211E+09, 6.8653E+09, 6.9097E+09,
     +  6.9544E+09, 6.9994E+09, 7.0446E+09, 7.0901E+09, 7.1358E+09,
     + 7.1818E+09, 7.2281E+09, 7.2747E+09, 7.3215E+09, 7.3686E+09,
     +  7.4160E+09, 7.4636E+09, 7.5116E+09, 7.5598E+09, 7.6083E+09,
     + 7.6571E+09, 7.7061E+09, 7.7555E+09, 7.8051E+09, 7.8551E+09,
     +  7.9053E+09, 7.9558E+09, 8.0066E+09, 8.0577E+09, 8.1091E+09,
     + 8.1608E+09, 8.2128E+09, 8.2651E+09, 8.3177E+09, 8.3706E+09,
     +  8.4238E+09, 8.4773E+09, 8.5311E+09, 8.5853E+09, 8.6397E+09,
     + 8.6945E+09, 8.7496E+09, 8.8050E+09, 8.8607E+09, 8.9167E+09,
     +  8.9731E+09, 9.0298E+09, 9.0868E+09, 9.1441E+09, 9.2018E+09,
     + 9.2598E+09, 9.3181E+09, 9.3768E+09, 9.4358E+09, 9.4951E+09,
     +  9.5548E+09, 9.6148E+09, 9.6752E+09, 9.7359E+09, 9.7970E+09,
     + 9.8584E+09, 9.9201E+09, 9.9822E+09, 1.0045E+10, 1.0107E+10,
     +  1.0171E+10, 1.0234E+10, 1.0298E+10, 1.0362E+10, 1.0427E+10,
     + 1.0492E+10, 1.0557E+10, 1.0623E+10, 1.0689E+10, 1.0756E+10,
     +  1.0822E+10, 1.0890E+10, 1.0957E+10, 1.1025E+10, 1.1094E+10,
     + 1.1162E+10, 1.1231E+10, 1.1301E+10, 1.1371E+10, 1.1441E+10,
     +  1.1512E+10, 1.1583E+10, 1.1655E+10, 1.1727E+10, 1.1799E+10,
     + 1.1872E+10, 1.1945E+10, 1.2018E+10, 1.2092E+10, 1.2167E+10,
     +  1.2241E+10, 1.2317E+10, 1.2392E+10, 1.2468E+10, 1.2545E+10,
     + 1.2622E+10, 1.2699E+10, 1.2777E+10, 1.2855E+10, 1.2933E+10,
     +  1.3012E+10, 1.3092E+10, 1.3172E+10, 1.3252E+10, 1.3333E+10,
     + 1.3414E+10, 1.3496E+10, 1.3578E+10, 1.3661E+10, 1.3744E+10,
     +  1.3828E+10, 1.3912E+10, 1.3996E+10, 1.4081E+10, 1.4166E+10,
     + 1.4252E+10, 1.4339E+10, 1.4426E+10, 1.4513E+10, 1.4601E+10,
     +  1.4689E+10, 1.4778E+10, 1.4867E+10, 1.4957E+10, 1.5047E+10,
     + 1.5138E+10, 1.5229E+10, 1.5321E+10, 1.5413E+10, 1.5506E+10,
     +  1.5599E+10, 1.5693E+10, 1.5787E+10, 1.5882E+10, 1.5977E+10,
     + 1.6073E+10, 1.6169E+10, 1.6266E+10, 1.6363E+10, 1.6461E+10,
     +  1.6560E+10, 1.6659E+10, 1.6758E+10, 1.6858E+10, 1.6959E+10,
     + 1.7060E+10, 1.7162E+10, 1.7264E+10, 1.7367E+10, 1.7470E+10,
     +  1.7574E+10, 1.7679E+10, 1.7784E+10, 1.7889E+10, 1.7996E+10,
     + 1.8102E+10, 1.8210E+10, 1.8318E+10, 1.8426E+10, 1.8535E+10,
     +  1.8645E+10, 1.8755E+10, 1.8866E+10, 1.8978E+10, 1.9090E+10,
     + 1.9202E+10, 1.9316E+10, 1.9430E+10, 1.9544E+10, 1.9659E+10,
     +  1.9775E+10, 1.9891E+10, 2.0008E+10, 2.0126E+10, 2.0244E+10,
     + 2.0363E+10, 2.0482E+10, 2.0602E+10, 2.0723E+10, 2.0845E+10,
     +  2.0967E+10, 2.1089E+10, 2.1213E+10, 2.1337E+10, 2.1461E+10,
     + 2.1587E+10, 2.1713E+10, 2.1839E+10, 2.1967E+10, 2.2095E+10,
     +  2.2223E+10, 2.2353E+10, 2.2483E+10, 2.2614E+10, 2.2745E+10,
     + 2.2877E+10, 2.3010E+10, 2.3144E+10, 2.3278E+10, 2.3413E+10,
     +  2.3548E+10, 2.3685E+10, 2.3822E+10, 2.3960E+10, 2.4098E+10,
     + 2.4238E+10, 2.4378E+10, 2.4518E+10, 2.4660E+10, 2.4802E+10,
     +  2.4945E+10, 2.5089E+10, 2.5233E+10, 2.5378E+10, 2.5524E+10,
     + 2.5671E+10, 2.5819E+10, 2.5967E+10, 2.6116E+10, 2.6266E+10,
     +  2.6417E+10, 2.6568E+10, 2.6720E+10, 2.6873E+10, 2.7027E+10,
     + 2.7182E+10, 2.7337E+10, 2.7493E+10, 2.7650E+10, 2.7808E+10,
     +  2.7967E+10, 2.8126E+10, 2.8286E+10, 2.8448E+10, 2.8609E+10,
     + 2.8772E+10, 2.8936E+10, 2.9100E+10, 2.9266E+10, 2.9432E+10,
     +  2.9599E+10, 2.9767E+10, 2.9936E+10, 3.0105E+10, 3.0276E+10,
     + 3.0447E+10, 3.0620E+10, 3.0793E+10, 3.0967E+10, 3.1142E+10,
     +  3.1318E+10, 3.1495E+10, 3.1672E+10, 3.1851E+10, 3.2030E+10,
     + 3.2211E+10, 3.2392E+10, 3.2574E+10, 3.2758E+10, 3.2942E+10,
     +  3.3127E+10, 3.3313E+10, 3.3500E+10, 3.3688E+10, 3.3877E+10,
     + 3.4067E+10, 3.4257E+10, 3.4449E+10, 3.4642E+10, 3.4836E+10,
     +  3.5030E+10, 3.5226E+10, 3.5423E+10, 3.5621E+10, 3.5819E+10,
     + 3.6019E+10, 3.6220E+10, 3.6422E+10, 3.6625E+10, 3.6828E+10,
     +  3.7033E+10, 3.7239E+10, 3.7446E+10, 3.7654E+10, 3.7863E+10,
     + 3.8073E+10, 3.8284E+10, 3.8497E+10, 3.8710E+10, 3.8924E+10,
     +  3.9140E+10, 3.9356E+10, 3.9574E+10, 3.9792E+10, 4.0012E+10,
     + 4.0233E+10, 4.0455E+10, 4.0678E+10, 4.0903E+10, 4.1128E+10,
     +  4.1354E+10, 4.1582E+10, 4.1811E+10, 4.2041E+10, 4.2272E+10,
     + 4.2504E+10, 4.2737E+10, 4.2972E+10, 4.3208E+10, 4.3445E+10,
     +  4.3683E+10, 4.3922E+10, 4.4163E+10, 4.4404E+10, 4.4647E+10,
     + 4.4891E+10, 4.5136E+10, 4.5383E+10, 4.5631E+10, 4.5880E+10,
     +  4.6130E+10, 4.6381E+10, 4.6634E+10, 4.6888E+10, 4.7143E+10,
     + 4.7400E+10, 4.7657E+10, 4.7916E+10, 4.8177E+10, 4.8438E+10,
     +  4.8701E+10, 4.8965E+10, 4.9231E+10, 4.9497E+10, 4.9765E+10,
     + 5.0035E+10, 5.0306E+10, 5.0578E+10, 5.0851E+10, 5.1126E+10,
     +  5.1402E+10, 5.1679E+10, 5.1958E+10, 5.2238E+10, 5.2520E+10,
     + 5.2803E+10, 5.3087E+10, 5.3373E+10, 5.3660E+10, 5.3948E+10,
     +  5.4238E+10, 5.4530E+10, 5.4822E+10, 5.5117E+10, 5.5412E+10,
     + 5.5709E+10, 5.6008E+10, 5.6308E+10, 5.6609E+10, 5.6912E+10,
     +  5.7217E+10, 5.7522E+10, 5.7830E+10, 5.8139E+10, 5.8449E+10,
     + 5.8761E+10, 5.9074E+10, 5.9389E+10, 5.9706E+10, 6.0023E+10,
     +  6.0343E+10, 6.0664E+10, 6.0987E+10, 6.1311E+10, 6.1636E+10,
     + 6.1964E+10, 6.2293E+10, 6.2623E+10, 6.2955E+10, 6.3289E+10,
     +  6.3624E+10, 6.3961E+10, 6.4299E+10, 6.4639E+10, 6.4981E+10,
     + 6.5325E+10, 6.5670E+10, 6.6016E+10, 6.6365E+10, 6.6715E+10,
     +  6.7066E+10, 6.7420E+10, 6.7775E+10, 6.8132E+10, 6.8490E+10,
     + 6.8851E+10, 6.9213E+10, 6.9576E+10, 6.9942E+10, 7.0309E+10,
     +  7.0678E+10, 7.1048E+10, 7.1421E+10, 7.1795E+10, 7.2171E+10,
     + 7.2549E+10, 7.2928E+10, 7.3310E+10, 7.3693E+10, 7.4078E+10,
     +  7.4465E+10, 7.4854E+10, 7.5244E+10, 7.5637E+10, 7.6031E+10,
     + 7.6427E+10, 7.6825E+10, 7.7225E+10, 7.7627E+10, 7.8030E+10,
     +  7.8436E+10, 7.8844E+10, 7.9253E+10, 7.9664E+10, 8.0078E+10,
     + 8.0493E+10, 8.0910E+10, 8.1329E+10, 8.1751E+10, 8.2174E+10,
     +  8.2599E+10, 8.3026E+10, 8.3455E+10, 8.3886E+10, 8.4320E+10,
     + 8.4755E+10, 8.5192E+10, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --     12225
      data (QofT( 2,J),J=1,2408)/ 2.0265E+01, 3.9108E+01, 5.7971E+01,
     +  7.6840E+01, 9.5712E+01, 1.1458E+02, 1.3346E+02, 1.5233E+02,
     + 1.7121E+02, 1.9008E+02, 2.0896E+02, 2.2784E+02, 2.4671E+02,
     +  2.6559E+02, 2.8447E+02, 3.0335E+02, 3.2223E+02, 3.4111E+02,
     + 3.5999E+02, 3.7887E+02, 4.1663E+02, 4.5439E+02, 4.9217E+02,
     +  5.2994E+02, 5.6773E+02, 6.0552E+02, 6.4335E+02, 6.8121E+02,
     + 7.1911E+02, 7.5708E+02, 7.9513E+02, 8.3329E+02, 8.7158E+02,
     +  9.1002E+02, 9.4865E+02, 9.8748E+02, 1.0266E+03, 1.0659E+03,
     + 1.1055E+03, 1.1455E+03, 1.1858E+03, 1.2266E+03, 1.2677E+03,
     +  1.3093E+03, 1.3514E+03, 1.3940E+03, 1.4371E+03, 1.4808E+03,
     + 1.5250E+03, 1.5699E+03, 1.6154E+03, 1.6616E+03, 1.7085E+03,
     +  1.7561E+03, 1.8045E+03, 1.8536E+03, 1.9036E+03, 1.9543E+03,
     + 2.0059E+03, 2.0583E+03, 2.1117E+03, 2.1660E+03, 2.2212E+03,
     +  2.2773E+03, 2.3345E+03, 2.3927E+03, 2.4519E+03, 2.5122E+03,
     + 2.5736E+03, 2.6361E+03, 2.6997E+03, 2.7645E+03, 2.8305E+03,
     +  2.8978E+03, 2.9663E+03, 3.0361E+03, 3.1072E+03, 3.1796E+03,
     + 3.2534E+03, 3.3286E+03, 3.4053E+03, 3.4834E+03, 3.5629E+03,
     +  3.6441E+03, 3.7268E+03, 3.8110E+03, 3.8969E+03, 3.9845E+03,
     + 4.0737E+03, 4.1647E+03, 4.2574E+03, 4.3520E+03, 4.4483E+03,
     +  4.5466E+03, 4.6467E+03, 4.7489E+03, 4.8530E+03, 4.9591E+03,
     + 5.0674E+03, 5.1777E+03, 5.2902E+03, 5.4048E+03, 5.5218E+03,
     +  5.6410E+03, 5.7625E+03, 5.8865E+03, 6.0128E+03, 6.1416E+03,
     + 6.2730E+03, 6.4069E+03, 6.5434E+03, 6.6826E+03, 6.8246E+03,
     +  6.9693E+03, 7.1168E+03, 7.2671E+03, 7.4205E+03, 7.5768E+03,
     + 7.7362E+03, 7.8986E+03, 8.0642E+03, 8.2331E+03, 8.4052E+03,
     +  8.5806E+03, 8.7595E+03, 8.9418E+03, 9.1277E+03, 9.3171E+03,
     + 9.5102E+03, 9.7070E+03, 9.9076E+03, 1.0112E+04, 1.0321E+04,
     +  1.0533E+04, 1.0749E+04, 1.0970E+04, 1.1195E+04, 1.1424E+04,
     + 1.1657E+04, 1.1895E+04, 1.2138E+04, 1.2385E+04, 1.2637E+04,
     +  1.2893E+04, 1.3155E+04, 1.3421E+04, 1.3693E+04, 1.3969E+04,
     + 1.4251E+04, 1.4538E+04, 1.4830E+04, 1.5128E+04, 1.5432E+04,
     +  1.5741E+04, 1.6056E+04, 1.6377E+04, 1.6703E+04, 1.7036E+04,
     + 1.7375E+04, 1.7720E+04, 1.8072E+04, 1.8430E+04, 1.8795E+04,
     +  1.9166E+04, 1.9545E+04, 1.9930E+04, 2.0322E+04, 2.0722E+04,
     + 2.1128E+04, 2.1542E+04, 2.1964E+04, 2.2393E+04, 2.2831E+04,
     +  2.3276E+04, 2.3729E+04, 2.4190E+04, 2.4659E+04, 2.5138E+04,
     + 2.5624E+04, 2.6119E+04, 2.6624E+04, 2.7137E+04, 2.7659E+04,
     +  2.8191E+04, 2.8732E+04, 2.9282E+04, 2.9843E+04, 3.0413E+04,
     + 3.0994E+04, 3.1584E+04, 3.2185E+04, 3.2797E+04, 3.3419E+04,
     +  3.4052E+04, 3.4696E+04, 3.5352E+04, 3.6019E+04, 3.6697E+04,
     + 3.7387E+04, 3.8090E+04, 3.8804E+04, 3.9531E+04, 4.0270E+04,
     +  4.1022E+04, 4.1786E+04, 4.2564E+04, 4.3356E+04, 4.4160E+04,
     + 4.4979E+04, 4.5811E+04, 4.6658E+04, 4.7519E+04, 4.8394E+04,
     +  4.9284E+04, 5.0190E+04, 5.1110E+04, 5.2046E+04, 5.2998E+04,
     + 5.3966E+04, 5.4950E+04, 5.5950E+04, 5.6967E+04, 5.8001E+04,
     +  5.9053E+04, 6.0121E+04, 6.1207E+04, 6.2312E+04, 6.3434E+04,
     + 6.4575E+04, 6.5735E+04, 6.6914E+04, 6.8112E+04, 6.9330E+04,
     +  7.0568E+04, 7.1825E+04, 7.3104E+04, 7.4403E+04, 7.5723E+04,
     + 7.7065E+04, 7.8428E+04, 7.9814E+04, 8.1221E+04, 8.2652E+04,
     +  8.4105E+04, 8.5582E+04, 8.7082E+04, 8.8606E+04, 9.0155E+04,
     + 9.1728E+04, 9.3327E+04, 9.4950E+04, 9.6600E+04, 9.8275E+04,
     +  9.9978E+04, 1.0171E+05, 1.0346E+05, 1.0525E+05, 1.0706E+05,
     + 1.0890E+05, 1.1077E+05, 1.1267E+05, 1.1459E+05, 1.1655E+05,
     +  1.1854E+05, 1.2056E+05, 1.2261E+05, 1.2469E+05, 1.2681E+05,
     + 1.2895E+05, 1.3113E+05, 1.3335E+05, 1.3560E+05, 1.3788E+05,
     +  1.4019E+05, 1.4255E+05, 1.4494E+05, 1.4736E+05, 1.4982E+05,
     + 1.5232E+05, 1.5486E+05, 1.5743E+05, 1.6005E+05, 1.6270E+05,
     +  1.6539E+05, 1.6813E+05, 1.7090E+05, 1.7372E+05, 1.7658E+05,
     + 1.7948E+05, 1.8243E+05, 1.8541E+05, 1.8845E+05, 1.9153E+05,
     +  1.9465E+05, 1.9782E+05, 2.0104E+05, 2.0430E+05, 2.0762E+05,
     + 2.1098E+05, 2.1439E+05, 2.1785E+05, 2.2136E+05, 2.2492E+05,
     +  2.2854E+05, 2.3221E+05, 2.3593E+05, 2.3970E+05, 2.4353E+05,
     + 2.4742E+05, 2.5136E+05, 2.5536E+05, 2.5942E+05, 2.6353E+05,
     +  2.6771E+05, 2.7194E+05, 2.7624E+05, 2.8059E+05, 2.8501E+05,
     + 2.8949E+05, 2.9404E+05, 2.9865E+05, 3.0333E+05, 3.0807E+05,
     +  3.1288E+05, 3.1776E+05, 3.2271E+05, 3.2772E+05, 3.3281E+05,
     + 3.3797E+05, 3.4320E+05, 3.4851E+05, 3.5389E+05, 3.5934E+05,
     +  3.6487E+05, 3.7048E+05, 3.7617E+05, 3.8193E+05, 3.8778E+05,
     + 3.9371E+05, 3.9972E+05, 4.0581E+05, 4.1199E+05, 4.1825E+05,
     +  4.2459E+05, 4.3103E+05, 4.3755E+05, 4.4417E+05, 4.5087E+05,
     + 4.5767E+05, 4.6455E+05, 4.7154E+05, 4.7861E+05, 4.8579E+05,
     +  4.9306E+05, 5.0043E+05, 5.0790E+05, 5.1547E+05, 5.2314E+05,
     + 5.3092E+05, 5.3880E+05, 5.4679E+05, 5.5488E+05, 5.6309E+05,
     +  5.7140E+05, 5.7983E+05, 5.8836E+05, 5.9702E+05, 6.0578E+05,
     + 6.1466E+05, 6.2367E+05, 6.3279E+05, 6.4203E+05, 6.5139E+05,
     +  6.6088E+05, 6.7049E+05, 6.8023E+05, 6.9010E+05, 7.0010E+05,
     + 7.1023E+05, 7.2049E+05, 7.3089E+05, 7.4142E+05, 7.5209E+05,
     +  7.6290E+05, 7.7385E+05, 7.8494E+05, 7.9618E+05, 8.0756E+05,
     + 8.1909E+05, 8.3077E+05, 8.4260E+05, 8.5459E+05, 8.6673E+05,
     +  8.7902E+05, 8.9147E+05, 9.0408E+05, 9.1686E+05, 9.2980E+05,
     + 9.4290E+05, 9.5617E+05, 9.6961E+05, 9.8322E+05, 9.9700E+05,
     +  1.0110E+06, 1.0251E+06, 1.0394E+06, 1.0539E+06, 1.0686E+06,
     + 1.0835E+06, 1.0985E+06, 1.1138E+06, 1.1292E+06, 1.1448E+06,
     +  1.1606E+06, 1.1767E+06, 1.1929E+06, 1.2093E+06, 1.2260E+06,
     + 1.2428E+06, 1.2598E+06, 1.2771E+06, 1.2946E+06, 1.3123E+06,
     +  1.3302E+06, 1.3483E+06, 1.3667E+06, 1.3853E+06, 1.4041E+06,
     + 1.4231E+06, 1.4424E+06, 1.4619E+06, 1.4817E+06, 1.5017E+06,
     +  1.5219E+06, 1.5424E+06, 1.5632E+06, 1.5842E+06, 1.6054E+06,
     + 1.6269E+06, 1.6487E+06, 1.6707E+06, 1.6930E+06, 1.7156E+06,
     +  1.7384E+06, 1.7616E+06, 1.7850E+06, 1.8086E+06, 1.8326E+06,
     + 1.8569E+06, 1.8814E+06, 1.9063E+06, 1.9314E+06, 1.9568E+06,
     +  1.9826E+06, 2.0086E+06, 2.0350E+06, 2.0617E+06, 2.0887E+06,
     + 2.1160E+06, 2.1436E+06, 2.1716E+06, 2.1999E+06, 2.2285E+06,
     +  2.2575E+06, 2.2868E+06, 2.3165E+06, 2.3465E+06, 2.3768E+06,
     + 2.4075E+06, 2.4386E+06, 2.4701E+06, 2.5019E+06, 2.5341E+06,
     +  2.5666E+06, 2.5995E+06, 2.6329E+06, 2.6666E+06, 2.7007E+06,
     + 2.7352E+06, 2.7701E+06, 2.8054E+06, 2.8411E+06, 2.8772E+06,
     +  2.9137E+06, 2.9507E+06, 2.9881E+06, 3.0259E+06, 3.0641E+06,
     + 3.1028E+06, 3.1420E+06, 3.1815E+06, 3.2216E+06, 3.2621E+06,
     +  3.3030E+06, 3.3444E+06, 3.3863E+06, 3.4287E+06, 3.4715E+06,
     + 3.5149E+06, 3.5587E+06, 3.6030E+06, 3.6478E+06, 3.6932E+06,
     +  3.7390E+06, 3.7854E+06, 3.8322E+06, 3.8797E+06, 3.9276E+06,
     + 3.9761E+06, 4.0251E+06, 4.0747E+06, 4.1248E+06, 4.1755E+06,
     +  4.2267E+06, 4.2786E+06, 4.3310E+06, 4.3839E+06, 4.4375E+06,
     + 4.4917E+06, 4.5465E+06, 4.6018E+06, 4.6578E+06, 4.7144E+06,
     +  4.7717E+06, 4.8295E+06, 4.8880E+06, 4.9472E+06, 5.0070E+06,
     + 5.0674E+06, 5.1285E+06, 5.1903E+06, 5.2528E+06, 5.3159E+06,
     +  5.3798E+06, 5.4443E+06, 5.5095E+06, 5.5755E+06, 5.6421E+06,
     + 5.7095E+06, 5.7776E+06, 5.8465E+06, 5.9161E+06, 5.9864E+06,
     +  6.0576E+06, 6.1294E+06, 6.2021E+06, 6.2756E+06, 6.3498E+06,
     + 6.4248E+06, 6.5007E+06, 6.5773E+06, 6.6548E+06, 6.7331E+06,
     +  6.8122E+06, 6.8922E+06, 6.9730E+06, 7.0547E+06, 7.1373E+06,
     + 7.2208E+06, 7.3051E+06, 7.3903E+06, 7.4765E+06, 7.5635E+06,
     +  7.6515E+06, 7.7404E+06, 7.8302E+06, 7.9210E+06, 8.0128E+06,
     + 8.1055E+06, 8.1992E+06, 8.2939E+06, 8.3895E+06, 8.4862E+06,
     +  8.5839E+06, 8.6826E+06, 8.7823E+06, 8.8831E+06, 8.9849E+06,
     + 9.0878E+06, 9.1918E+06, 9.2969E+06, 9.4030E+06, 9.5102E+06,
     +  9.6186E+06, 9.7281E+06, 9.8387E+06, 9.9504E+06, 1.0063E+07,
     + 1.0177E+07, 1.0293E+07, 1.0409E+07, 1.0527E+07, 1.0646E+07,
     +  1.0766E+07, 1.0887E+07, 1.1010E+07, 1.1133E+07, 1.1258E+07,
     + 1.1385E+07, 1.1512E+07, 1.1641E+07, 1.1772E+07, 1.1903E+07,
     +  1.2036E+07, 1.2170E+07, 1.2306E+07, 1.2443E+07, 1.2581E+07,
     + 1.2721E+07, 1.2862E+07, 1.3005E+07, 1.3149E+07, 1.3294E+07,
     +  1.3441E+07, 1.3590E+07, 1.3740E+07, 1.3891E+07, 1.4044E+07,
     + 1.4199E+07, 1.4355E+07, 1.4512E+07, 1.4672E+07, 1.4832E+07,
     +  1.4995E+07, 1.5159E+07, 1.5325E+07, 1.5492E+07, 1.5661E+07,
     + 1.5831E+07, 1.6004E+07, 1.6178E+07, 1.6353E+07, 1.6531E+07,
     +  1.6710E+07, 1.6891E+07, 1.7074E+07, 1.7259E+07, 1.7445E+07,
     + 1.7633E+07, 1.7823E+07, 1.8015E+07, 1.8209E+07, 1.8405E+07,
     +  1.8602E+07, 1.8802E+07, 1.9004E+07, 1.9207E+07, 1.9413E+07,
     + 1.9620E+07, 1.9830E+07, 2.0041E+07, 2.0255E+07, 2.0470E+07,
     +  2.0688E+07, 2.0908E+07, 2.1130E+07, 2.1354E+07, 2.1580E+07,
     + 2.1809E+07, 2.2040E+07, 2.2273E+07, 2.2508E+07, 2.2745E+07,
     +  2.2985E+07, 2.3227E+07, 2.3471E+07, 2.3718E+07, 2.3967E+07,
     + 2.4218E+07, 2.4472E+07, 2.4728E+07, 2.4987E+07, 2.5248E+07,
     +  2.5512E+07, 2.5778E+07, 2.6046E+07, 2.6318E+07, 2.6591E+07,
     + 2.6868E+07, 2.7147E+07, 2.7428E+07, 2.7713E+07, 2.8000E+07,
     +  2.8289E+07, 2.8582E+07, 2.8877E+07, 2.9175E+07, 2.9475E+07,
     + 2.9779E+07, 3.0085E+07, 3.0395E+07, 3.0707E+07, 3.1022E+07,
     +  3.1340E+07, 3.1661E+07, 3.1985E+07, 3.2312E+07, 3.2642E+07,
     + 3.2975E+07, 3.3311E+07, 3.3651E+07, 3.3993E+07, 3.4339E+07,
     +  3.4688E+07, 3.5040E+07, 3.5395E+07, 3.5754E+07, 3.6116E+07,
     + 3.6481E+07, 3.6850E+07, 3.7222E+07, 3.7597E+07, 3.7976E+07,
     +  3.8359E+07, 3.8745E+07, 3.9134E+07, 3.9527E+07, 3.9924E+07,
     + 4.0324E+07, 4.0728E+07, 4.1136E+07, 4.1547E+07, 4.1962E+07,
     +  4.2381E+07, 4.2803E+07, 4.3230E+07, 4.3660E+07, 4.4095E+07,
     + 4.4533E+07, 4.4975E+07, 4.5421E+07, 4.5871E+07, 4.6326E+07,
     +  4.6784E+07, 4.7246E+07, 4.7713E+07, 4.8184E+07, 4.8659E+07,
     + 4.9138E+07, 4.9622E+07, 5.0110E+07, 5.0602E+07, 5.1099E+07,
     +  5.1600E+07, 5.2106E+07, 5.2616E+07, 5.3131E+07, 5.3651E+07,
     + 5.4175E+07, 5.4703E+07, 5.5237E+07, 5.5775E+07, 5.6318E+07,
     +  5.6865E+07, 5.7418E+07, 5.7975E+07, 5.8538E+07, 5.9105E+07,
     + 5.9678E+07, 6.0255E+07, 6.0837E+07, 6.1425E+07, 6.2018E+07,
     +  6.2616E+07, 6.3219E+07, 6.3828E+07, 6.4442E+07, 6.5061E+07,
     + 6.5686E+07, 6.6316E+07, 6.6952E+07, 6.7593E+07, 6.8240E+07,
     +  6.8893E+07, 6.9551E+07, 7.0215E+07, 7.0885E+07, 7.1560E+07,
     + 7.2242E+07, 7.2929E+07, 7.3622E+07, 7.4321E+07, 7.5027E+07,
     +  7.5738E+07, 7.6456E+07, 7.7180E+07, 7.7910E+07, 7.8646E+07,
     + 7.9389E+07, 8.0138E+07, 8.0894E+07, 8.1656E+07, 8.2424E+07,
     +  8.3199E+07, 8.3981E+07, 8.4770E+07, 8.5565E+07, 8.6367E+07,
     + 8.7176E+07, 8.7992E+07, 8.8815E+07, 8.9645E+07, 9.0481E+07,
     +  9.1326E+07, 9.2177E+07, 9.3035E+07, 9.3901E+07, 9.4774E+07,
     + 9.5654E+07, 9.6542E+07, 9.7438E+07, 9.8341E+07, 9.9252E+07,
     +  1.0017E+08, 1.0110E+08, 1.0203E+08, 1.0297E+08, 1.0392E+08,
     + 1.0488E+08, 1.0584E+08, 1.0682E+08, 1.0780E+08, 1.0879E+08,
     +  1.0979E+08, 1.1080E+08, 1.1181E+08, 1.1284E+08, 1.1387E+08,
     + 1.1491E+08, 1.1596E+08, 1.1702E+08, 1.1808E+08, 1.1916E+08,
     +  1.2025E+08, 1.2134E+08, 1.2244E+08, 1.2355E+08, 1.2468E+08,
     + 1.2581E+08, 1.2695E+08, 1.2810E+08, 1.2926E+08, 1.3042E+08,
     +  1.3160E+08, 1.3279E+08, 1.3399E+08, 1.3520E+08, 1.3641E+08,
     + 1.3764E+08, 1.3888E+08, 1.4013E+08, 1.4138E+08, 1.4265E+08,
     +  1.4393E+08, 1.4522E+08, 1.4652E+08, 1.4783E+08, 1.4915E+08,
     + 1.5048E+08, 1.5183E+08, 1.5318E+08, 1.5454E+08, 1.5592E+08,
     +  1.5731E+08, 1.5870E+08, 1.6011E+08, 1.6153E+08, 1.6297E+08,
     + 1.6441E+08, 1.6587E+08, 1.6733E+08, 1.6881E+08, 1.7030E+08,
     +  1.7181E+08, 1.7332E+08, 1.7485E+08, 1.7639E+08, 1.7794E+08,
     + 1.7951E+08, 1.8108E+08, 1.8267E+08, 1.8427E+08, 1.8589E+08,
     +  1.8752E+08, 1.8916E+08, 1.9081E+08, 1.9248E+08, 1.9416E+08,
     + 1.9586E+08, 1.9756E+08, 1.9929E+08, 2.0102E+08, 2.0277E+08,
     +  2.0453E+08, 2.0631E+08, 2.0810E+08, 2.0990E+08, 2.1172E+08,
     + 2.1356E+08, 2.1540E+08, 2.1727E+08, 2.1914E+08, 2.2104E+08,
     +  2.2294E+08, 2.2486E+08, 2.2680E+08, 2.2875E+08, 2.3072E+08,
     + 2.3271E+08, 2.3470E+08, 2.3672E+08, 2.3875E+08, 2.4079E+08,
     +  2.4286E+08, 2.4493E+08, 2.4703E+08, 2.4914E+08, 2.5127E+08,
     + 2.5341E+08, 2.5557E+08, 2.5775E+08, 2.5994E+08, 2.6215E+08,
     +  2.6438E+08, 2.6663E+08, 2.6889E+08, 2.7117E+08, 2.7347E+08,
     + 2.7579E+08, 2.7812E+08, 2.8047E+08, 2.8284E+08, 2.8523E+08,
     +  2.8764E+08, 2.9006E+08, 2.9251E+08, 2.9497E+08, 2.9745E+08,
     + 2.9995E+08, 3.0247E+08, 3.0501E+08, 3.0757E+08, 3.1015E+08,
     +  3.1275E+08, 3.1536E+08, 3.1800E+08, 3.2066E+08, 3.2334E+08,
     + 3.2604E+08, 3.2876E+08, 3.3150E+08, 3.3426E+08, 3.3704E+08,
     +  3.3984E+08, 3.4267E+08, 3.4551E+08, 3.4838E+08, 3.5127E+08,
     + 3.5418E+08, 3.5711E+08, 3.6007E+08, 3.6305E+08, 3.6605E+08,
     +  3.6907E+08, 3.7211E+08, 3.7518E+08, 3.7827E+08, 3.8139E+08,
     + 3.8453E+08, 3.8769E+08, 3.9087E+08, 3.9408E+08, 3.9732E+08,
     +  4.0057E+08, 4.0386E+08, 4.0716E+08, 4.1049E+08, 4.1385E+08,
     + 4.1723E+08, 4.2064E+08, 4.2407E+08, 4.2753E+08, 4.3101E+08,
     +  4.3452E+08, 4.3805E+08, 4.4161E+08, 4.4520E+08, 4.4882E+08,
     + 4.5246E+08, 4.5613E+08, 4.5982E+08, 4.6354E+08, 4.6729E+08,
     +  4.7107E+08, 4.7488E+08, 4.7871E+08, 4.8257E+08, 4.8646E+08,
     + 4.9038E+08, 4.9433E+08, 4.9831E+08, 5.0231E+08, 5.0635E+08,
     +  5.1041E+08, 5.1451E+08, 5.1863E+08, 5.2279E+08, 5.2698E+08,
     + 5.3119E+08, 5.3544E+08, 5.3972E+08, 5.4403E+08, 5.4837E+08,
     +  5.5274E+08, 5.5714E+08, 5.6158E+08, 5.6605E+08, 5.7055E+08,
     + 5.7508E+08, 5.7965E+08, 5.8425E+08, 5.8888E+08, 5.9355E+08,
     +  5.9825E+08, 6.0298E+08, 6.0775E+08, 6.1255E+08, 6.1739E+08,
     + 6.2226E+08, 6.2717E+08, 6.3211E+08, 6.3709E+08, 6.4211E+08,
     +  6.4716E+08, 6.5224E+08, 6.5737E+08, 6.6253E+08, 6.6772E+08,
     + 6.7296E+08, 6.7823E+08, 6.8354E+08, 6.8888E+08, 6.9427E+08,
     +  6.9969E+08, 7.0516E+08, 7.1066E+08, 7.1620E+08, 7.2178E+08,
     + 7.2740E+08, 7.3306E+08, 7.3876E+08, 7.4450E+08, 7.5028E+08,
     +  7.5610E+08, 7.6197E+08, 7.6787E+08, 7.7382E+08, 7.7981E+08,
     + 7.8584E+08, 7.9191E+08, 7.9803E+08, 8.0419E+08, 8.1039E+08,
     +  8.1664E+08, 8.2293E+08, 8.2927E+08, 8.3565E+08, 8.4207E+08,
     + 8.4854E+08, 8.5506E+08, 8.6162E+08, 8.6823E+08, 8.7488E+08,
     +  8.8158E+08, 8.8832E+08, 8.9512E+08, 9.0196E+08, 9.0885E+08,
     + 9.1579E+08, 9.2277E+08, 9.2981E+08, 9.3689E+08, 9.4402E+08,
     +  9.5120E+08, 9.5844E+08, 9.6572E+08, 9.7305E+08, 9.8044E+08,
     + 9.8787E+08, 9.9536E+08, 1.0029E+09, 1.0105E+09, 1.0181E+09,
     +  1.0258E+09, 1.0336E+09, 1.0414E+09, 1.0492E+09, 1.0571E+09,
     + 1.0651E+09, 1.0731E+09, 1.0812E+09, 1.0893E+09, 1.0975E+09,
     +  1.1058E+09, 1.1141E+09, 1.1224E+09, 1.1308E+09, 1.1393E+09,
     + 1.1478E+09, 1.1564E+09, 1.1651E+09, 1.1738E+09, 1.1825E+09,
     +  1.1914E+09, 1.2002E+09, 1.2092E+09, 1.2182E+09, 1.2273E+09,
     + 1.2364E+09, 1.2456E+09, 1.2548E+09, 1.2641E+09, 1.2735E+09,
     +  1.2830E+09, 1.2925E+09, 1.3020E+09, 1.3117E+09, 1.3214E+09,
     + 1.3311E+09, 1.3410E+09, 1.3509E+09, 1.3608E+09, 1.3709E+09,
     +  1.3810E+09, 1.3911E+09, 1.4014E+09, 1.4117E+09, 1.4221E+09,
     + 1.4325E+09, 1.4430E+09, 1.4536E+09, 1.4643E+09, 1.4750E+09,
     +  1.4858E+09, 1.4967E+09, 1.5076E+09, 1.5186E+09, 1.5297E+09,
     + 1.5409E+09, 1.5521E+09, 1.5634E+09, 1.5748E+09, 1.5863E+09,
     +  1.5978E+09, 1.6094E+09, 1.6211E+09, 1.6329E+09, 1.6448E+09,
     + 1.6567E+09, 1.6687E+09, 1.6808E+09, 1.6930E+09, 1.7052E+09,
     +  1.7175E+09, 1.7299E+09, 1.7424E+09, 1.7550E+09, 1.7677E+09,
     + 1.7804E+09, 1.7932E+09, 1.8062E+09, 1.8192E+09, 1.8322E+09,
     +  1.8454E+09, 1.8587E+09, 1.8720E+09, 1.8854E+09, 1.8989E+09,
     + 1.9125E+09, 1.9262E+09, 1.9400E+09, 1.9539E+09, 1.9679E+09,
     +  1.9819E+09, 1.9961E+09, 2.0103E+09, 2.0246E+09, 2.0391E+09,
     + 2.0536E+09, 2.0682E+09, 2.0829E+09, 2.0977E+09, 2.1126E+09,
     +  2.1276E+09, 2.1427E+09, 2.1579E+09, 2.1732E+09, 2.1886E+09,
     + 2.2041E+09, 2.2197E+09, 2.2354E+09, 2.2512E+09, 2.2671E+09,
     +  2.2831E+09, 2.2992E+09, 2.3154E+09, 2.3317E+09, 2.3481E+09,
     + 2.3646E+09, 2.3812E+09, 2.3980E+09, 2.4148E+09, 2.4318E+09,
     +  2.4488E+09, 2.4660E+09, 2.4833E+09, 2.5006E+09, 2.5181E+09,
     + 2.5358E+09, 2.5535E+09, 2.5713E+09, 2.5893E+09, 2.6073E+09,
     +  2.6255E+09, 2.6438E+09, 2.6622E+09, 2.6808E+09, 2.6994E+09,
     + 2.7182E+09, 2.7371E+09, 2.7561E+09, 2.7752E+09, 2.7944E+09,
     +  2.8138E+09, 2.8333E+09, 2.8529E+09, 2.8727E+09, 2.8925E+09,
     + 2.9125E+09, 2.9326E+09, 2.9529E+09, 2.9732E+09, 2.9937E+09,
     +  3.0144E+09, 3.0351E+09, 3.0560E+09, 3.0770E+09, 3.0982E+09,
     + 3.1195E+09, 3.1409E+09, 3.1625E+09, 3.1842E+09, 3.2060E+09,
     +  3.2279E+09, 3.2500E+09, 3.2723E+09, 3.2947E+09, 3.3172E+09,
     + 3.3398E+09, 3.3626E+09, 3.3856E+09, 3.4087E+09, 3.4319E+09,
     +  3.4553E+09, 3.4788E+09, 3.5025E+09, 3.5263E+09, 3.5502E+09,
     + 3.5744E+09, 3.5986E+09, 3.6230E+09, 3.6476E+09, 3.6723E+09,
     +  3.6972E+09, 3.7222E+09, 3.7474E+09, 3.7727E+09, 3.7982E+09,
     + 3.8238E+09, 3.8497E+09, 3.8756E+09, 3.9017E+09, 3.9280E+09,
     +  3.9545E+09, 3.9811E+09, 4.0079E+09, 4.0348E+09, 4.0619E+09,
     + 4.0892E+09, 4.1166E+09, 4.1442E+09, 4.1720E+09, 4.2000E+09,
     +  4.2281E+09, 4.2564E+09, 4.2848E+09, 4.3135E+09, 4.3423E+09,
     + 4.3713E+09, 4.4004E+09, 4.4298E+09, 4.4593E+09, 4.4890E+09,
     +  4.5189E+09, 4.5489E+09, 4.5792E+09, 4.6096E+09, 4.6402E+09,
     + 4.6711E+09, 4.7020E+09, 4.7332E+09, 4.7646E+09, 4.7961E+09,
     +  4.8279E+09, 4.8598E+09, 4.8920E+09, 4.9243E+09, 4.9568E+09,
     + 4.9895E+09, 5.0225E+09, 5.0556E+09, 5.0889E+09, 5.1224E+09,
     +  5.1561E+09, 5.1901E+09, 5.2242E+09, 5.2585E+09, 5.2931E+09,
     + 5.3278E+09, 5.3628E+09, 5.3979E+09, 5.4333E+09, 5.4689E+09,
     +  5.5047E+09, 5.5407E+09, 5.5769E+09, 5.6134E+09, 5.6501E+09,
     + 5.6869E+09, 5.7240E+09, 5.7614E+09, 5.7989E+09, 5.8367E+09,
     +  5.8747E+09, 5.9129E+09, 5.9514E+09, 5.9900E+09, 6.0290E+09,
     + 6.0681E+09, 6.1075E+09, 6.1471E+09, 6.1869E+09, 6.2270E+09,
     +  6.2673E+09, 6.3079E+09, 6.3487E+09, 6.3897E+09, 6.4310E+09,
     + 6.4725E+09, 6.5143E+09, 6.5563E+09, 6.5986E+09, 6.6411E+09,
     +  6.6838E+09, 6.7269E+09, 6.7701E+09, 6.8136E+09, 6.8574E+09,
     + 6.9015E+09, 6.9458E+09, 6.9903E+09, 7.0351E+09, 7.0802E+09,
     +  7.1256E+09, 7.1712E+09, 7.2170E+09, 7.2632E+09, 7.3096E+09,
     + 7.3563E+09, 7.4033E+09, 7.4505E+09, 7.4980E+09, 7.5458E+09,
     +  7.5939E+09, 7.6422E+09, 7.6908E+09, 7.7397E+09, 7.7889E+09,
     + 7.8384E+09, 7.8882E+09, 7.9383E+09, 7.9886E+09, 8.0392E+09,
     +  8.0902E+09, 8.1414E+09, 8.1929E+09, 8.2448E+09, 8.2969E+09,
     + 8.3493E+09, 8.4020E+09, 8.4551E+09, 8.5084E+09, 8.5621E+09,
     +  8.6160E+09, 8.6703E+09, 8.7249E+09, 8.7798E+09, 8.8350E+09,
     + 8.8905E+09, 8.9464E+09, 9.0026E+09, 9.0591E+09, 9.1159E+09,
     +  9.1730E+09, 9.2305E+09, 9.2883E+09, 9.3464E+09, 9.4049E+09,
     + 9.4637E+09, 9.5228E+09, 9.5823E+09, 9.6421E+09, 9.7023E+09,
     +  9.7628E+09, 9.8236E+09, 9.8848E+09, 9.9464E+09, 1.0008E+10,
     + 1.0071E+10, 1.0133E+10, 1.0196E+10, 1.0259E+10, 1.0323E+10,
     +  1.0387E+10, 1.0452E+10, 1.0516E+10, 1.0581E+10, 1.0647E+10,
     + 1.0713E+10, 1.0779E+10, 1.0846E+10, 1.0913E+10, 1.0980E+10,
     +  1.1048E+10, 1.1116E+10, 1.1184E+10, 1.1253E+10, 1.1323E+10,
     + 1.1392E+10, 1.1462E+10, 1.1533E+10, 1.1604E+10, 1.1675E+10,
     +  1.1747E+10, 1.1819E+10, 1.1891E+10, 1.1964E+10, 1.2038E+10,
     + 1.2111E+10, 1.2185E+10, 1.2260E+10, 1.2335E+10, 1.2410E+10,
     +  1.2486E+10, 1.2562E+10, 1.2639E+10, 1.2716E+10, 1.2793E+10,
     + 1.2871E+10, 1.2950E+10, 1.3029E+10, 1.3108E+10, 1.3188E+10,
     +  1.3268E+10, 1.3348E+10, 1.3429E+10, 1.3511E+10, 1.3593E+10,
     + 1.3675E+10, 1.3758E+10, 1.3841E+10, 1.3925E+10, 1.4009E+10,
     +  1.4094E+10, 1.4179E+10, 1.4265E+10, 1.4351E+10, 1.4437E+10,
     + 1.4524E+10, 1.4612E+10, 1.4700E+10, 1.4788E+10, 1.4877E+10,
     +  1.4967E+10, 1.5057E+10, 1.5147E+10, 1.5238E+10, 1.5330E+10,
     + 1.5422E+10, 1.5514E+10, 1.5607E+10, 1.5701E+10, 1.5795E+10,
     +  1.5889E+10, 1.5984E+10, 1.6080E+10, 1.6176E+10, 1.6273E+10,
     + 1.6370E+10, 1.6468E+10, 1.6566E+10, 1.6664E+10, 1.6764E+10,
     +  1.6864E+10, 1.6964E+10, 1.7065E+10, 1.7166E+10, 1.7268E+10,
     + 1.7371E+10, 1.7474E+10, 1.7578E+10, 1.7682E+10, 1.7787E+10,
     +  1.7892E+10, 1.7998E+10, 1.8105E+10, 1.8212E+10, 1.8320E+10,
     + 1.8428E+10, 1.8537E+10, 1.8646E+10, 1.8756E+10, 1.8867E+10,
     +  1.8978E+10, 1.9090E+10, 1.9202E+10, 1.9315E+10, 1.9429E+10,
     + 1.9543E+10, 1.9658E+10, 1.9774E+10, 1.9890E+10, 2.0007E+10,
     +  2.0124E+10, 2.0242E+10, 2.0361E+10, 2.0480E+10, 2.0600E+10,
     + 2.0720E+10, 2.0842E+10, 2.0964E+10, 2.1086E+10, 2.1209E+10,
     +  2.1333E+10, 2.1458E+10, 2.1583E+10, 2.1709E+10, 2.1835E+10,
     + 2.1962E+10, 2.2090E+10, 2.2219E+10, 2.2348E+10, 2.2478E+10,
     +  2.2608E+10, 2.2740E+10, 2.2872E+10, 2.3004E+10, 2.3138E+10,
     + 2.3272E+10, 2.3407E+10, 2.3542E+10, 2.3678E+10, 2.3815E+10,
     +  2.3953E+10, 2.4091E+10, 2.4231E+10, 2.4371E+10, 2.4511E+10,
     + 2.4653E+10, 2.4795E+10, 2.4938E+10, 2.5081E+10, 2.5226E+10,
     +  2.5371E+10, 2.5517E+10, 2.5663E+10, 2.5811E+10, 2.5959E+10,
     + 2.6108E+10, 2.6258E+10, 2.6408E+10, 2.6560E+10, 2.6712E+10,
     +  2.6865E+10, 2.7019E+10, 2.7173E+10, 2.7329E+10, 2.7485E+10,
     + 2.7642E+10, 2.7800E+10, 2.7958E+10, 2.8118E+10, 2.8278E+10,
     +  2.8439E+10, 2.8601E+10, 2.8764E+10, 2.8928E+10, 2.9093E+10,
     + 2.9258E+10, 2.9424E+10, 2.9591E+10, 2.9759E+10, 2.9928E+10,
     +  3.0098E+10, 3.0269E+10, 3.0440E+10, 3.0613E+10, 3.0786E+10,
     + 3.0960E+10, 3.1135E+10, 3.1311E+10, 3.1488E+10, 3.1666E+10,
     +  3.1845E+10, 3.2025E+10, 3.2205E+10, 3.2387E+10, 3.2569E+10,
     + 3.2753E+10, 3.2937E+10, 3.3123E+10, 3.3309E+10, 3.3496E+10,
     +  3.3684E+10, 3.3874E+10, 3.4064E+10, 3.4255E+10, 3.4447E+10,
     + 3.4640E+10, 3.4834E+10, 3.5029E+10, 3.5225E+10, 3.5423E+10,
     +  3.5621E+10, 3.5820E+10, 3.6020E+10, 3.6221E+10, 3.6423E+10,
     + 3.6627E+10, 3.6831E+10, 3.7036E+10, 3.7243E+10, 3.7450E+10,
     +  3.7659E+10, 3.7868E+10, 3.8079E+10, 3.8290E+10, 3.8503E+10,
     + 3.8717E+10, 3.8932E+10, 3.9148E+10, 3.9365E+10, 3.9584E+10,
     +  3.9803E+10, 4.0023E+10, 4.0245E+10, 4.0468E+10, 4.0691E+10,
     + 4.0916E+10, 4.1143E+10, 4.1370E+10, 4.1598E+10, 4.1828E+10,
     +  4.2059E+10, 4.2290E+10, 4.2524E+10, 4.2758E+10, 4.2993E+10,
     + 4.3230E+10, 4.3468E+10, 4.3707E+10, 4.3947E+10, 4.4188E+10,
     +  4.4431E+10, 4.4675E+10, 4.4920E+10, 4.5166E+10, 4.5414E+10,
     + 4.5663E+10, 4.5913E+10, 4.6164E+10, 4.6416E+10, 4.6670E+10,
     +  4.6925E+10, 4.7182E+10, 4.7439E+10, 4.7698E+10, 4.7959E+10,
     + 4.8220E+10, 4.8483E+10, 4.8747E+10, 4.9013E+10, 4.9280E+10,
     +  4.9548E+10, 4.9817E+10, 5.0088E+10, 5.0360E+10, 5.0634E+10,
     + 5.0909E+10, 5.1185E+10, 5.1463E+10, 5.1742E+10, 5.2022E+10,
     +  5.2304E+10, 5.2587E+10, 5.2872E+10, 5.3158E+10, 5.3445E+10,
     + 5.3734E+10, 5.4024E+10, 5.4316E+10, 5.4609E+10, 5.4904E+10,
     +  5.5200E+10, 5.5497E+10, 5.5796E+10, 5.6097E+10, 5.6399E+10,
     + 5.6702E+10, 5.7007E+10, 5.7314E+10, 5.7622E+10, 5.7931E+10,
     +  5.8242E+10, 5.8555E+10, 5.8869E+10, 5.9185E+10, 5.9502E+10,
     + 5.9821E+10, 6.0141E+10, 6.0463E+10, 6.0786E+10, 6.1111E+10,
     +  6.1438E+10, 6.1766E+10, 6.2096E+10, 6.2427E+10, 6.2761E+10,
     + 6.3095E+10, 6.3432E+10, 6.3770E+10, 6.4109E+10, 6.4451E+10,
     +  6.4794E+10, 6.5138E+10, 6.5484E+10, 6.5832E+10, 6.6182E+10,
     + 6.6534E+10, 6.6887E+10, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --     12234
      data (QofT( 3,J),J=1,2408)/ 5.9381E+01, 1.1449E+02, 1.6966E+02,
     +  2.2485E+02, 2.8005E+02, 3.3525E+02, 3.9045E+02, 4.4565E+02,
     + 5.0086E+02, 5.5607E+02, 6.1128E+02, 6.6649E+02, 7.2171E+02,
     +  7.7692E+02, 8.3214E+02, 8.8736E+02, 9.4258E+02, 9.9780E+02,
     + 1.0530E+03, 1.1082E+03, 1.2187E+03, 1.3292E+03, 1.4396E+03,
     +  1.5501E+03, 1.6606E+03, 1.7712E+03, 1.8818E+03, 1.9926E+03,
     + 2.1035E+03, 2.2145E+03, 2.3258E+03, 2.4374E+03, 2.5494E+03,
     +  2.6619E+03, 2.7749E+03, 2.8885E+03, 3.0028E+03, 3.1178E+03,
     + 3.2338E+03, 3.3507E+03, 3.4687E+03, 3.5878E+03, 3.7082E+03,
     +  3.8299E+03, 3.9530E+03, 4.0775E+03, 4.2037E+03, 4.3314E+03,
     + 4.4610E+03, 4.5923E+03, 4.7255E+03, 4.8606E+03, 4.9978E+03,
     +  5.1371E+03, 5.2786E+03, 5.4224E+03, 5.5685E+03, 5.7169E+03,
     + 5.8678E+03, 6.0213E+03, 6.1774E+03, 6.3362E+03, 6.4977E+03,
     +  6.6621E+03, 6.8293E+03, 6.9995E+03, 7.1729E+03, 7.3492E+03,
     + 7.5289E+03, 7.7118E+03, 7.8980E+03, 8.0877E+03, 8.2809E+03,
     +  8.4777E+03, 8.6781E+03, 8.8823E+03, 9.0904E+03, 9.3024E+03,
     + 9.5184E+03, 9.7385E+03, 9.9628E+03, 1.0191E+04, 1.0424E+04,
     +  1.0662E+04, 1.0904E+04, 1.1150E+04, 1.1402E+04, 1.1658E+04,
     + 1.1919E+04, 1.2185E+04, 1.2457E+04, 1.2734E+04, 1.3016E+04,
     +  1.3303E+04, 1.3597E+04, 1.3896E+04, 1.4200E+04, 1.4511E+04,
     + 1.4828E+04, 1.5151E+04, 1.5480E+04, 1.5816E+04, 1.6158E+04,
     +  1.6507E+04, 1.6863E+04, 1.7226E+04, 1.7596E+04, 1.7973E+04,
     + 1.8357E+04, 1.8750E+04, 1.9149E+04, 1.9557E+04, 1.9972E+04,
     +  2.0396E+04, 2.0828E+04, 2.1268E+04, 2.1717E+04, 2.2175E+04,
     + 2.2642E+04, 2.3118E+04, 2.3602E+04, 2.4097E+04, 2.4601E+04,
     +  2.5115E+04, 2.5639E+04, 2.6173E+04, 2.6717E+04, 2.7272E+04,
     + 2.7837E+04, 2.8414E+04, 2.9001E+04, 2.9600E+04, 3.0211E+04,
     +  3.0833E+04, 3.1467E+04, 3.2113E+04, 3.2772E+04, 3.3443E+04,
     + 3.4127E+04, 3.4824E+04, 3.5535E+04, 3.6259E+04, 3.6996E+04,
     +  3.7748E+04, 3.8514E+04, 3.9294E+04, 4.0090E+04, 4.0900E+04,
     + 4.1725E+04, 4.2566E+04, 4.3423E+04, 4.4296E+04, 4.5185E+04,
     +  4.6091E+04, 4.7014E+04, 4.7954E+04, 4.8912E+04, 4.9887E+04,
     + 5.0880E+04, 5.1892E+04, 5.2923E+04, 5.3973E+04, 5.5041E+04,
     +  5.6130E+04, 5.7239E+04, 5.8368E+04, 5.9518E+04, 6.0689E+04,
     + 6.1881E+04, 6.3095E+04, 6.4331E+04, 6.5590E+04, 6.6871E+04,
     +  6.8176E+04, 6.9504E+04, 7.0857E+04, 7.2233E+04, 7.3635E+04,
     + 7.5062E+04, 7.6514E+04, 7.7992E+04, 7.9497E+04, 8.1028E+04,
     +  8.2587E+04, 8.4174E+04, 8.5789E+04, 8.7432E+04, 8.9105E+04,
     + 9.0807E+04, 9.2539E+04, 9.4302E+04, 9.6096E+04, 9.7921E+04,
     +  9.9778E+04, 1.0167E+05, 1.0359E+05, 1.0555E+05, 1.0754E+05,
     + 1.0956E+05, 1.1162E+05, 1.1372E+05, 1.1585E+05, 1.1802E+05,
     +  1.2022E+05, 1.2247E+05, 1.2475E+05, 1.2707E+05, 1.2943E+05,
     + 1.3184E+05, 1.3428E+05, 1.3676E+05, 1.3929E+05, 1.4186E+05,
     +  1.4447E+05, 1.4713E+05, 1.4983E+05, 1.5258E+05, 1.5537E+05,
     + 1.5821E+05, 1.6110E+05, 1.6404E+05, 1.6702E+05, 1.7006E+05,
     +  1.7314E+05, 1.7628E+05, 1.7947E+05, 1.8271E+05, 1.8601E+05,
     + 1.8936E+05, 1.9276E+05, 1.9622E+05, 1.9974E+05, 2.0332E+05,
     +  2.0695E+05, 2.1065E+05, 2.1440E+05, 2.1822E+05, 2.2209E+05,
     + 2.2604E+05, 2.3004E+05, 2.3411E+05, 2.3824E+05, 2.4245E+05,
     +  2.4671E+05, 2.5105E+05, 2.5546E+05, 2.5994E+05, 2.6449E+05,
     + 2.6911E+05, 2.7381E+05, 2.7858E+05, 2.8342E+05, 2.8835E+05,
     +  2.9335E+05, 2.9843E+05, 3.0359E+05, 3.0883E+05, 3.1416E+05,
     + 3.1957E+05, 3.2506E+05, 3.3064E+05, 3.3631E+05, 3.4206E+05,
     +  3.4791E+05, 3.5384E+05, 3.5987E+05, 3.6599E+05, 3.7221E+05,
     + 3.7852E+05, 3.8493E+05, 3.9144E+05, 3.9805E+05, 4.0476E+05,
     +  4.1157E+05, 4.1849E+05, 4.2552E+05, 4.3265E+05, 4.3989E+05,
     + 4.4724E+05, 4.5470E+05, 4.6227E+05, 4.6996E+05, 4.7777E+05,
     +  4.8569E+05, 4.9373E+05, 5.0190E+05, 5.1019E+05, 5.1860E+05,
     + 5.2713E+05, 5.3580E+05, 5.4459E+05, 5.5352E+05, 5.6258E+05,
     +  5.7177E+05, 5.8110E+05, 5.9057E+05, 6.0018E+05, 6.0993E+05,
     + 6.1982E+05, 6.2987E+05, 6.4005E+05, 6.5039E+05, 6.6088E+05,
     +  6.7152E+05, 6.8232E+05, 6.9328E+05, 7.0440E+05, 7.1567E+05,
     + 7.2712E+05, 7.3872E+05, 7.5050E+05, 7.6245E+05, 7.7457E+05,
     +  7.8686E+05, 7.9933E+05, 8.1198E+05, 8.2482E+05, 8.3784E+05,
     + 8.5104E+05, 8.6443E+05, 8.7802E+05, 8.9179E+05, 9.0577E+05,
     +  9.1994E+05, 9.3431E+05, 9.4889E+05, 9.6368E+05, 9.7867E+05,
     + 9.9388E+05, 1.0093E+06, 1.0249E+06, 1.0408E+06, 1.0569E+06,
     +  1.0732E+06, 1.0897E+06, 1.1065E+06, 1.1235E+06, 1.1407E+06,
     + 1.1582E+06, 1.1759E+06, 1.1939E+06, 1.2121E+06, 1.2305E+06,
     +  1.2493E+06, 1.2682E+06, 1.2875E+06, 1.3070E+06, 1.3268E+06,
     + 1.3468E+06, 1.3671E+06, 1.3877E+06, 1.4086E+06, 1.4298E+06,
     +  1.4512E+06, 1.4730E+06, 1.4950E+06, 1.5173E+06, 1.5400E+06,
     + 1.5629E+06, 1.5862E+06, 1.6098E+06, 1.6337E+06, 1.6579E+06,
     +  1.6824E+06, 1.7073E+06, 1.7325E+06, 1.7580E+06, 1.7839E+06,
     + 1.8101E+06, 1.8367E+06, 1.8637E+06, 1.8909E+06, 1.9186E+06,
     +  1.9466E+06, 1.9750E+06, 2.0038E+06, 2.0329E+06, 2.0625E+06,
     + 2.0924E+06, 2.1227E+06, 2.1534E+06, 2.1845E+06, 2.2161E+06,
     +  2.2480E+06, 2.2803E+06, 2.3131E+06, 2.3463E+06, 2.3800E+06,
     + 2.4140E+06, 2.4486E+06, 2.4835E+06, 2.5190E+06, 2.5548E+06,
     +  2.5912E+06, 2.6280E+06, 2.6653E+06, 2.7030E+06, 2.7413E+06,
     + 2.7800E+06, 2.8193E+06, 2.8590E+06, 2.8993E+06, 2.9401E+06,
     +  2.9813E+06, 3.0232E+06, 3.0655E+06, 3.1084E+06, 3.1518E+06,
     + 3.1958E+06, 3.2403E+06, 3.2854E+06, 3.3311E+06, 3.3774E+06,
     +  3.4242E+06, 3.4716E+06, 3.5196E+06, 3.5683E+06, 3.6175E+06,
     + 3.6673E+06, 3.7178E+06, 3.7689E+06, 3.8207E+06, 3.8730E+06,
     +  3.9261E+06, 3.9798E+06, 4.0341E+06, 4.0892E+06, 4.1449E+06,
     + 4.2013E+06, 4.2584E+06, 4.3162E+06, 4.3747E+06, 4.4340E+06,
     +  4.4940E+06, 4.5547E+06, 4.6161E+06, 4.6783E+06, 4.7413E+06,
     + 4.8051E+06, 4.8696E+06, 4.9349E+06, 5.0010E+06, 5.0679E+06,
     +  5.1356E+06, 5.2042E+06, 5.2735E+06, 5.3437E+06, 5.4148E+06,
     + 5.4867E+06, 5.5595E+06, 5.6332E+06, 5.7077E+06, 5.7832E+06,
     +  5.8595E+06, 5.9368E+06, 6.0150E+06, 6.0941E+06, 6.1742E+06,
     + 6.2553E+06, 6.3373E+06, 6.4202E+06, 6.5042E+06, 6.5892E+06,
     +  6.6752E+06, 6.7621E+06, 6.8502E+06, 6.9392E+06, 7.0294E+06,
     + 7.1205E+06, 7.2128E+06, 7.3061E+06, 7.4006E+06, 7.4961E+06,
     +  7.5928E+06, 7.6906E+06, 7.7896E+06, 7.8897E+06, 7.9910E+06,
     + 8.0934E+06, 8.1971E+06, 8.3020E+06, 8.4080E+06, 8.5154E+06,
     +  8.6239E+06, 8.7337E+06, 8.8448E+06, 8.9572E+06, 9.0709E+06,
     + 9.1859E+06, 9.3022E+06, 9.4198E+06, 9.5388E+06, 9.6592E+06,
     +  9.7810E+06, 9.9041E+06, 1.0029E+07, 1.0155E+07, 1.0282E+07,
     + 1.0411E+07, 1.0541E+07, 1.0673E+07, 1.0806E+07, 1.0941E+07,
     +  1.1078E+07, 1.1216E+07, 1.1355E+07, 1.1496E+07, 1.1639E+07,
     + 1.1783E+07, 1.1929E+07, 1.2077E+07, 1.2226E+07, 1.2377E+07,
     +  1.2529E+07, 1.2684E+07, 1.2840E+07, 1.2997E+07, 1.3157E+07,
     + 1.3318E+07, 1.3481E+07, 1.3646E+07, 1.3813E+07, 1.3982E+07,
     +  1.4152E+07, 1.4324E+07, 1.4499E+07, 1.4675E+07, 1.4853E+07,
     + 1.5033E+07, 1.5215E+07, 1.5400E+07, 1.5586E+07, 1.5774E+07,
     +  1.5964E+07, 1.6157E+07, 1.6351E+07, 1.6548E+07, 1.6746E+07,
     + 1.6947E+07, 1.7150E+07, 1.7356E+07, 1.7563E+07, 1.7773E+07,
     +  1.7985E+07, 1.8200E+07, 1.8417E+07, 1.8636E+07, 1.8857E+07,
     + 1.9081E+07, 1.9307E+07, 1.9536E+07, 1.9767E+07, 2.0001E+07,
     +  2.0237E+07, 2.0476E+07, 2.0717E+07, 2.0961E+07, 2.1208E+07,
     + 2.1457E+07, 2.1709E+07, 2.1964E+07, 2.2221E+07, 2.2481E+07,
     +  2.2744E+07, 2.3009E+07, 2.3278E+07, 2.3549E+07, 2.3823E+07,
     + 2.4100E+07, 2.4380E+07, 2.4663E+07, 2.4949E+07, 2.5238E+07,
     +  2.5530E+07, 2.5825E+07, 2.6123E+07, 2.6424E+07, 2.6729E+07,
     + 2.7036E+07, 2.7347E+07, 2.7662E+07, 2.7979E+07, 2.8300E+07,
     +  2.8624E+07, 2.8951E+07, 2.9282E+07, 2.9617E+07, 2.9955E+07,
     + 3.0296E+07, 3.0641E+07, 3.0989E+07, 3.1341E+07, 3.1697E+07,
     +  3.2057E+07, 3.2420E+07, 3.2787E+07, 3.3157E+07, 3.3532E+07,
     + 3.3910E+07, 3.4292E+07, 3.4679E+07, 3.5069E+07, 3.5463E+07,
     +  3.5861E+07, 3.6263E+07, 3.6670E+07, 3.7080E+07, 3.7495E+07,
     + 3.7914E+07, 3.8337E+07, 3.8764E+07, 3.9196E+07, 3.9633E+07,
     +  4.0073E+07, 4.0518E+07, 4.0968E+07, 4.1422E+07, 4.1881E+07,
     + 4.2344E+07, 4.2812E+07, 4.3285E+07, 4.3763E+07, 4.4245E+07,
     +  4.4732E+07, 4.5225E+07, 4.5722E+07, 4.6224E+07, 4.6731E+07,
     + 4.7243E+07, 4.7760E+07, 4.8283E+07, 4.8810E+07, 4.9343E+07,
     +  4.9881E+07, 5.0425E+07, 5.0974E+07, 5.1528E+07, 5.2088E+07,
     + 5.2653E+07, 5.3225E+07, 5.3801E+07, 5.4384E+07, 5.4972E+07,
     +  5.5566E+07, 5.6166E+07, 5.6771E+07, 5.7383E+07, 5.8001E+07,
     + 5.8624E+07, 5.9254E+07, 5.9890E+07, 6.0533E+07, 6.1181E+07,
     +  6.1836E+07, 6.2497E+07, 6.3165E+07, 6.3839E+07, 6.4520E+07,
     + 6.5208E+07, 6.5902E+07, 6.6603E+07, 6.7310E+07, 6.8025E+07,
     +  6.8746E+07, 6.9475E+07, 7.0210E+07, 7.0953E+07, 7.1703E+07,
     + 7.2460E+07, 7.3224E+07, 7.3995E+07, 7.4775E+07, 7.5561E+07,
     +  7.6355E+07, 7.7157E+07, 7.7966E+07, 7.8783E+07, 7.9608E+07,
     + 8.0441E+07, 8.1282E+07, 8.2131E+07, 8.2988E+07, 8.3853E+07,
     +  8.4726E+07, 8.5608E+07, 8.6498E+07, 8.7396E+07, 8.8303E+07,
     + 8.9219E+07, 9.0143E+07, 9.1076E+07, 9.2018E+07, 9.2968E+07,
     +  9.3928E+07, 9.4897E+07, 9.5874E+07, 9.6861E+07, 9.7858E+07,
     + 9.8863E+07, 9.9878E+07, 1.0090E+08, 1.0194E+08, 1.0298E+08,
     +  1.0403E+08, 1.0510E+08, 1.0617E+08, 1.0725E+08, 1.0835E+08,
     + 1.0945E+08, 1.1057E+08, 1.1169E+08, 1.1282E+08, 1.1397E+08,
     +  1.1513E+08, 1.1629E+08, 1.1747E+08, 1.1866E+08, 1.1986E+08,
     + 1.2107E+08, 1.2229E+08, 1.2352E+08, 1.2477E+08, 1.2602E+08,
     +  1.2729E+08, 1.2857E+08, 1.2986E+08, 1.3116E+08, 1.3247E+08,
     + 1.3380E+08, 1.3514E+08, 1.3649E+08, 1.3785E+08, 1.3923E+08,
     +  1.4061E+08, 1.4202E+08, 1.4343E+08, 1.4485E+08, 1.4629E+08,
     + 1.4775E+08, 1.4921E+08, 1.5069E+08, 1.5218E+08, 1.5369E+08,
     +  1.5521E+08, 1.5674E+08, 1.5829E+08, 1.5985E+08, 1.6142E+08,
     + 1.6301E+08, 1.6461E+08, 1.6623E+08, 1.6786E+08, 1.6951E+08,
     +  1.7117E+08, 1.7285E+08, 1.7454E+08, 1.7624E+08, 1.7796E+08,
     + 1.7970E+08, 1.8145E+08, 1.8322E+08, 1.8501E+08, 1.8681E+08,
     +  1.8862E+08, 1.9045E+08, 1.9230E+08, 1.9417E+08, 1.9605E+08,
     + 1.9794E+08, 1.9986E+08, 2.0179E+08, 2.0374E+08, 2.0570E+08,
     +  2.0769E+08, 2.0969E+08, 2.1170E+08, 2.1374E+08, 2.1579E+08,
     + 2.1787E+08, 2.1996E+08, 2.2206E+08, 2.2419E+08, 2.2634E+08,
     +  2.2850E+08, 2.3068E+08, 2.3288E+08, 2.3511E+08, 2.3735E+08,
     + 2.3961E+08, 2.4189E+08, 2.4419E+08, 2.4651E+08, 2.4885E+08,
     +  2.5121E+08, 2.5359E+08, 2.5599E+08, 2.5841E+08, 2.6085E+08,
     + 2.6332E+08, 2.6580E+08, 2.6831E+08, 2.7084E+08, 2.7339E+08,
     +  2.7596E+08, 2.7855E+08, 2.8117E+08, 2.8381E+08, 2.8647E+08,
     + 2.8915E+08, 2.9186E+08, 2.9459E+08, 2.9735E+08, 3.0012E+08,
     +  3.0292E+08, 3.0575E+08, 3.0860E+08, 3.1147E+08, 3.1437E+08,
     + 3.1729E+08, 3.2024E+08, 3.2321E+08, 3.2621E+08, 3.2923E+08,
     +  3.3228E+08, 3.3536E+08, 3.3846E+08, 3.4158E+08, 3.4474E+08,
     + 3.4792E+08, 3.5112E+08, 3.5436E+08, 3.5762E+08, 3.6091E+08,
     +  3.6422E+08, 3.6757E+08, 3.7094E+08, 3.7434E+08, 3.7777E+08,
     + 3.8123E+08, 3.8471E+08, 3.8823E+08, 3.9177E+08, 3.9535E+08,
     +  3.9895E+08, 4.0259E+08, 4.0625E+08, 4.0995E+08, 4.1367E+08,
     + 4.1743E+08, 4.2122E+08, 4.2504E+08, 4.2889E+08, 4.3278E+08,
     +  4.3669E+08, 4.4064E+08, 4.4462E+08, 4.4863E+08, 4.5268E+08,
     + 4.5676E+08, 4.6087E+08, 4.6502E+08, 4.6920E+08, 4.7342E+08,
     +  4.7767E+08, 4.8196E+08, 4.8628E+08, 4.9063E+08, 4.9503E+08,
     + 4.9945E+08, 5.0392E+08, 5.0842E+08, 5.1296E+08, 5.1753E+08,
     +  5.2214E+08, 5.2679E+08, 5.3148E+08, 5.3621E+08, 5.4097E+08,
     + 5.4577E+08, 5.5062E+08, 5.5550E+08, 5.6042E+08, 5.6538E+08,
     +  5.7038E+08, 5.7542E+08, 5.8050E+08, 5.8563E+08, 5.9079E+08,
     + 5.9600E+08, 6.0125E+08, 6.0654E+08, 6.1187E+08, 6.1725E+08,
     +  6.2267E+08, 6.2813E+08, 6.3364E+08, 6.3919E+08, 6.4478E+08,
     + 6.5042E+08, 6.5611E+08, 6.6184E+08, 6.6762E+08, 6.7344E+08,
     +  6.7931E+08, 6.8523E+08, 6.9119E+08, 6.9720E+08, 7.0326E+08,
     + 7.0937E+08, 7.1552E+08, 7.2173E+08, 7.2798E+08, 7.3428E+08,
     +  7.4064E+08, 7.4704E+08, 7.5350E+08, 7.6000E+08, 7.6656E+08,
     + 7.7317E+08, 7.7983E+08, 7.8654E+08, 7.9331E+08, 8.0013E+08,
     +  8.0700E+08, 8.1393E+08, 8.2091E+08, 8.2794E+08, 8.3504E+08,
     + 8.4218E+08, 8.4939E+08, 8.5665E+08, 8.6396E+08, 8.7134E+08,
     +  8.7877E+08, 8.8626E+08, 8.9381E+08, 9.0141E+08, 9.0908E+08,
     + 9.1681E+08, 9.2459E+08, 9.3244E+08, 9.4035E+08, 9.4831E+08,
     +  9.5634E+08, 9.6444E+08, 9.7259E+08, 9.8081E+08, 9.8909E+08,
     + 9.9744E+08, 1.0059E+09, 1.0143E+09, 1.0229E+09, 1.0315E+09,
     +  1.0402E+09, 1.0489E+09, 1.0577E+09, 1.0666E+09, 1.0755E+09,
     + 1.0845E+09, 1.0936E+09, 1.1028E+09, 1.1120E+09, 1.1213E+09,
     +  1.1306E+09, 1.1401E+09, 1.1496E+09, 1.1592E+09, 1.1688E+09,
     + 1.1785E+09, 1.1883E+09, 1.1982E+09, 1.2082E+09, 1.2182E+09,
     +  1.2283E+09, 1.2385E+09, 1.2487E+09, 1.2591E+09, 1.2695E+09,
     + 1.2800E+09, 1.2905E+09, 1.3012E+09, 1.3119E+09, 1.3227E+09,
     +  1.3336E+09, 1.3446E+09, 1.3557E+09, 1.3668E+09, 1.3780E+09,
     + 1.3893E+09, 1.4007E+09, 1.4122E+09, 1.4238E+09, 1.4354E+09,
     +  1.4472E+09, 1.4590E+09, 1.4709E+09, 1.4829E+09, 1.4950E+09,
     + 1.5072E+09, 1.5195E+09, 1.5319E+09, 1.5443E+09, 1.5569E+09,
     +  1.5695E+09, 1.5823E+09, 1.5951E+09, 1.6080E+09, 1.6211E+09,
     + 1.6342E+09, 1.6474E+09, 1.6607E+09, 1.6742E+09, 1.6877E+09,
     +  1.7013E+09, 1.7150E+09, 1.7288E+09, 1.7428E+09, 1.7568E+09,
     + 1.7709E+09, 1.7851E+09, 1.7995E+09, 1.8139E+09, 1.8285E+09,
     +  1.8431E+09, 1.8579E+09, 1.8728E+09, 1.8878E+09, 1.9029E+09,
     + 1.9181E+09, 1.9334E+09, 1.9488E+09, 1.9643E+09, 1.9800E+09,
     +  1.9958E+09, 2.0116E+09, 2.0276E+09, 2.0438E+09, 2.0600E+09,
     + 2.0763E+09, 2.0928E+09, 2.1094E+09, 2.1261E+09, 2.1429E+09,
     +  2.1599E+09, 2.1770E+09, 2.1942E+09, 2.2115E+09, 2.2289E+09,
     + 2.2465E+09, 2.2642E+09, 2.2821E+09, 2.3000E+09, 2.3181E+09,
     +  2.3363E+09, 2.3547E+09, 2.3732E+09, 2.3918E+09, 2.4105E+09,
     + 2.4294E+09, 2.4484E+09, 2.4676E+09, 2.4869E+09, 2.5063E+09,
     +  2.5259E+09, 2.5456E+09, 2.5654E+09, 2.5854E+09, 2.6056E+09,
     + 2.6259E+09, 2.6463E+09, 2.6669E+09, 2.6876E+09, 2.7084E+09,
     +  2.7295E+09, 2.7506E+09, 2.7719E+09, 2.7934E+09, 2.8150E+09,
     + 2.8368E+09, 2.8587E+09, 2.8808E+09, 2.9030E+09, 2.9254E+09,
     +  2.9480E+09, 2.9707E+09, 2.9936E+09, 3.0166E+09, 3.0398E+09,
     + 3.0632E+09, 3.0867E+09, 3.1104E+09, 3.1343E+09, 3.1583E+09,
     +  3.1825E+09, 3.2068E+09, 3.2314E+09, 3.2561E+09, 3.2810E+09,
     + 3.3060E+09, 3.3313E+09, 3.3567E+09, 3.3823E+09, 3.4080E+09,
     +  3.4340E+09, 3.4601E+09, 3.4864E+09, 3.5129E+09, 3.5396E+09,
     + 3.5664E+09, 3.5935E+09, 3.6207E+09, 3.6481E+09, 3.6758E+09,
     +  3.7036E+09, 3.7316E+09, 3.7598E+09, 3.7882E+09, 3.8168E+09,
     + 3.8455E+09, 3.8745E+09, 3.9037E+09, 3.9331E+09, 3.9627E+09,
     +  3.9925E+09, 4.0225E+09, 4.0527E+09, 4.0831E+09, 4.1137E+09,
     + 4.1446E+09, 4.1756E+09, 4.2069E+09, 4.2384E+09, 4.2700E+09,
     +  4.3020E+09, 4.3341E+09, 4.3664E+09, 4.3990E+09, 4.4318E+09,
     + 4.4648E+09, 4.4981E+09, 4.5315E+09, 4.5652E+09, 4.5991E+09,
     +  4.6333E+09, 4.6677E+09, 4.7023E+09, 4.7372E+09, 4.7723E+09,
     + 4.8076E+09, 4.8432E+09, 4.8790E+09, 4.9151E+09, 4.9514E+09,
     +  4.9879E+09, 5.0247E+09, 5.0618E+09, 5.0991E+09, 5.1366E+09,
     + 5.1745E+09, 5.2125E+09, 5.2508E+09, 5.2894E+09, 5.3282E+09,
     +  5.3673E+09, 5.4067E+09, 5.4463E+09, 5.4862E+09, 5.5264E+09,
     + 5.5668E+09, 5.6075E+09, 5.6485E+09, 5.6897E+09, 5.7313E+09,
     +  5.7731E+09, 5.8152E+09, 5.8575E+09, 5.9002E+09, 5.9431E+09,
     + 5.9863E+09, 6.0299E+09, 6.0737E+09, 6.1177E+09, 6.1621E+09,
     +  6.2068E+09, 6.2518E+09, 6.2971E+09, 6.3426E+09, 6.3885E+09,
     + 6.4347E+09, 6.4812E+09, 6.5280E+09, 6.5751E+09, 6.6225E+09,
     +  6.6703E+09, 6.7183E+09, 6.7667E+09, 6.8154E+09, 6.8644E+09,
     + 6.9138E+09, 6.9634E+09, 7.0134E+09, 7.0637E+09, 7.1144E+09,
     +  7.1654E+09, 7.2167E+09, 7.2683E+09, 7.3203E+09, 7.3727E+09,
     + 7.4253E+09, 7.4784E+09, 7.5317E+09, 7.5855E+09, 7.6395E+09,
     +  7.6940E+09, 7.7488E+09, 7.8039E+09, 7.8594E+09, 7.9153E+09,
     + 7.9715E+09, 8.0281E+09, 8.0851E+09, 8.1424E+09, 8.2001E+09,
     +  8.2582E+09, 8.3167E+09, 8.3755E+09, 8.4347E+09, 8.4943E+09,
     + 8.5543E+09, 8.6147E+09, 8.6755E+09, 8.7367E+09, 8.7982E+09,
     +  8.8602E+09, 8.9226E+09, 8.9853E+09, 9.0485E+09, 9.1121E+09,
     + 9.1761E+09, 9.2405E+09, 9.3053E+09, 9.3706E+09, 9.4362E+09,
     +  9.5023E+09, 9.5688E+09, 9.6357E+09, 9.7031E+09, 9.7709E+09,
     + 9.8391E+09, 9.9078E+09, 9.9769E+09, 1.0046E+10, 1.0116E+10,
     +  1.0187E+10, 1.0258E+10, 1.0329E+10, 1.0401E+10, 1.0473E+10,
     + 1.0546E+10, 1.0619E+10, 1.0693E+10, 1.0767E+10, 1.0842E+10,
     +  1.0917E+10, 1.0992E+10, 1.1068E+10, 1.1145E+10, 1.1222E+10,
     + 1.1299E+10, 1.1377E+10, 1.1456E+10, 1.1535E+10, 1.1614E+10,
     +  1.1694E+10, 1.1774E+10, 1.1855E+10, 1.1937E+10, 1.2019E+10,
     + 1.2101E+10, 1.2185E+10, 1.2268E+10, 1.2352E+10, 1.2437E+10,
     +  1.2522E+10, 1.2608E+10, 1.2694E+10, 1.2781E+10, 1.2868E+10,
     + 1.2956E+10, 1.3044E+10, 1.3133E+10, 1.3223E+10, 1.3313E+10,
     +  1.3404E+10, 1.3495E+10, 1.3587E+10, 1.3679E+10, 1.3772E+10,
     + 1.3865E+10, 1.3960E+10, 1.4054E+10, 1.4150E+10, 1.4245E+10,
     +  1.4342E+10, 1.4439E+10, 1.4537E+10, 1.4635E+10, 1.4734E+10,
     + 1.4833E+10, 1.4933E+10, 1.5034E+10, 1.5136E+10, 1.5238E+10,
     +  1.5340E+10, 1.5443E+10, 1.5547E+10, 1.5652E+10, 1.5757E+10,
     + 1.5863E+10, 1.5970E+10, 1.6077E+10, 1.6184E+10, 1.6293E+10,
     +  1.6402E+10, 1.6512E+10, 1.6622E+10, 1.6734E+10, 1.6845E+10,
     + 1.6958E+10, 1.7071E+10, 1.7185E+10, 1.7300E+10, 1.7415E+10,
     +  1.7531E+10, 1.7648E+10, 1.7765E+10, 1.7884E+10, 1.8003E+10,
     + 1.8122E+10, 1.8243E+10, 1.8364E+10, 1.8485E+10, 1.8608E+10,
     +  1.8731E+10, 1.8855E+10, 1.8980E+10, 1.9106E+10, 1.9232E+10,
     + 1.9359E+10, 1.9487E+10, 1.9616E+10, 1.9745E+10, 1.9876E+10,
     +  2.0007E+10, 2.0138E+10, 2.0271E+10, 2.0405E+10, 2.0539E+10,
     + 2.0674E+10, 2.0810E+10, 2.0946E+10, 2.1084E+10, 2.1222E+10,
     +  2.1361E+10, 2.1501E+10, 2.1642E+10, 2.1784E+10, 2.1927E+10,
     + 2.2070E+10, 2.2214E+10, 2.2359E+10, 2.2505E+10, 2.2652E+10,
     +  2.2800E+10, 2.2949E+10, 2.3098E+10, 2.3249E+10, 2.3400E+10,
     + 2.3552E+10, 2.3705E+10, 2.3859E+10, 2.4014E+10, 2.4170E+10,
     +  2.4327E+10, 2.4485E+10, 2.4644E+10, 2.4803E+10, 2.4964E+10,
     + 2.5126E+10, 2.5288E+10, 2.5452E+10, 2.5616E+10, 2.5782E+10,
     +  2.5948E+10, 2.6116E+10, 2.6284E+10, 2.6453E+10, 2.6624E+10,
     + 2.6795E+10, 2.6968E+10, 2.7141E+10, 2.7316E+10, 2.7491E+10,
     +  2.7668E+10, 2.7845E+10, 2.8024E+10, 2.8204E+10, 2.8385E+10,
     + 2.8566E+10, 2.8749E+10, 2.8933E+10, 2.9118E+10, 2.9305E+10,
     +  2.9492E+10, 2.9680E+10, 2.9870E+10, 3.0060E+10, 3.0252E+10,
     + 3.0445E+10, 3.0639E+10, 3.0834E+10, 3.1030E+10, 3.1227E+10,
     +  3.1426E+10, 3.1625E+10, 3.1826E+10, 3.2028E+10, 3.2232E+10,
     + 3.2436E+10, 3.2642E+10, 3.2848E+10, 3.3056E+10, 3.3265E+10,
     +  3.3476E+10, 3.3688E+10, 3.3900E+10, 3.4114E+10, 3.4330E+10,
     + 3.4546E+10, 3.4764E+10, 3.4983E+10, 3.5204E+10, 3.5425E+10,
     +  3.5648E+10, 3.5873E+10, 3.6098E+10, 3.6325E+10, 3.6553E+10,
     + 3.6783E+10, 3.7013E+10, 3.7246E+10, 3.7479E+10, 3.7714E+10,
     +  3.7950E+10, 3.8188E+10, 3.8426E+10, 3.8667E+10, 3.8908E+10,
     + 3.9151E+10, 3.9396E+10, 3.9642E+10, 3.9889E+10, 4.0138E+10,
     +  4.0388E+10, 4.0639E+10, 4.0892E+10, 4.1147E+10, 4.1402E+10,
     + 4.1660E+10, 4.1919E+10, 4.2179E+10, 4.2441E+10, 4.2704E+10,
     +  4.2969E+10, 4.3235E+10, 4.3503E+10, 4.3772E+10, 4.4043E+10,
     + 4.4315E+10, 4.4589E+10, 4.4865E+10, 4.5142E+10, 4.5421E+10,
     +  4.5701E+10, 4.5983E+10, 4.6266E+10, 4.6551E+10, 4.6838E+10,
     + 4.7126E+10, 4.7416E+10, 4.7708E+10, 4.8001E+10, 4.8296E+10,
     +  4.8592E+10, 4.8890E+10, 4.9190E+10, 4.9492E+10, 4.9795E+10,
     + 5.0100E+10, 5.0407E+10, 5.0715E+10, 5.1025E+10, 5.1337E+10,
     +  5.1651E+10, 5.1966E+10, 5.2283E+10, 5.2602E+10, 5.2923E+10,
     + 5.3246E+10, 5.3570E+10, 5.3896E+10, 5.4224E+10, 5.4554E+10,
     +  5.4886E+10, 5.5220E+10, 5.5555E+10, 5.5892E+10, 5.6231E+10,
     + 5.6573E+10, 5.6916E+10, 5.7260E+10, 5.7607E+10, 5.7956E+10,
     +  5.8307E+10, 5.8659E+10, 5.9014E+10, 5.9371E+10, 5.9729E+10,
     + 6.0090E+10, 6.0452E+10, 6.0817E+10, 6.1184E+10, 6.1552E+10,
     +  6.1923E+10, 6.2296E+10, 6.2670E+10, 6.3047E+10, 6.3426E+10,
     + 6.3807E+10, 6.4191E+10, 6.4576E+10, 6.4963E+10, 6.5353E+10,
     +  6.5745E+10, 6.6138E+10, 6.6534E+10, 6.6933E+10, 6.7333E+10,
     + 6.7736E+10, 6.8141E+10, 6.8548E+10, 6.8957E+10, 6.9369E+10,
     +  6.9782E+10, 7.0198E+10, 7.0617E+10, 7.1037E+10, 7.1460E+10,
     + 7.1886E+10, 7.2313E+10, 7.2743E+10, 7.3176E+10, 7.3610E+10,
     +  7.4047E+10, 7.4487E+10, 7.4929E+10, 7.5373E+10, 7.5820E+10,
     + 7.6269E+10, 7.6720E+10, 7.7174E+10, 7.7631E+10, 7.8090E+10,
     +  7.8551E+10, 7.9015E+10, 7.9482E+10, 7.9951E+10, 8.0423E+10,
     + 8.0897E+10, 8.1374E+10, 8.1853E+10, 8.2335E+10, 8.2819E+10,
     +  8.3307E+10, 8.3796E+10, 8.4289E+10, 8.4784E+10, 8.5282E+10,
     + 8.5782E+10, 8.6286E+10, 8.6791E+10, 8.7300E+10, 8.7812E+10,
     +  8.8326E+10, 8.8843E+10, 8.9362E+10, 8.9885E+10, 9.0410E+10,
     + 9.0938E+10, 9.1469E+10, 9.2003E+10, 9.2540E+10, 9.3079E+10,
     +  9.3622E+10, 9.4167E+10, 9.4716E+10, 9.5267E+10, 9.5821E+10,
     + 9.6378E+10, 9.6939E+10, 9.7502E+10, 9.8068E+10, 9.8637E+10,
     +  9.9210E+10, 9.9785E+10, 1.0036E+11, 1.0094E+11, 1.0153E+11,
     + 1.0212E+11, 1.0271E+11, 1.0330E+11, 1.0390E+11, 1.0450E+11,
     +  1.0510E+11, 1.0571E+11, 1.0632E+11, 1.0693E+11, 1.0755E+11,
     + 1.0817E+11, 1.0879E+11, 1.0942E+11, 1.1005E+11, 1.1068E+11,
     +  1.1132E+11, 1.1196E+11, 1.1260E+11, 1.1325E+11, 1.1390E+11,
     + 1.1455E+11, 1.1521E+11, 1.1587E+11, 1.1653E+11, 1.1720E+11,
     +  1.1787E+11, 1.1854E+11, 1.1922E+11, 1.1990E+11, 1.2059E+11,
     + 1.2128E+11, 1.2197E+11, 1.2266E+11, 1.2336E+11, 1.2407E+11,
     +  1.2477E+11, 1.2548E+11, 1.2620E+11, 1.2692E+11, 1.2764E+11,
     + 1.2836E+11, 1.2909E+11, 1.2983E+11, 1.3056E+11, 1.3131E+11,
     +  1.3205E+11, 1.3280E+11, 1.3355E+11, 1.3431E+11, 1.3507E+11,
     + 1.3583E+11, 1.3660E+11, 1.3737E+11, 1.3815E+11, 1.3893E+11,
     +  1.3972E+11, 1.4050E+11, 1.4130E+11, 1.4209E+11, 1.4290E+11,
     + 1.4370E+11, 1.4451E+11, 1.4532E+11, 1.4614E+11, 1.4696E+11,
     +  1.4779E+11, 1.4862E+11, 1.4946E+11, 1.5029E+11, 1.5114E+11,
     + 1.5199E+11, 1.5284E+11, 1.5370E+11, 1.5456E+11, 1.5542E+11,
     +  1.5629E+11, 1.5717E+11, 1.5805E+11, 1.5893E+11, 1.5982E+11,
     + 1.6071E+11, 1.6161E+11, 1.6251E+11, 1.6342E+11, 1.6433E+11,
     +  1.6524E+11, 1.6616E+11, 1.6709E+11, 1.6802E+11, 1.6895E+11,
     + 1.6989E+11, 1.7084E+11, 1.7179E+11, 1.7274E+11, 1.7370E+11,
     +  1.7466E+11, 1.7563E+11, 1.7661E+11, 1.7759E+11, 1.7857E+11,
     + 1.7956E+11, 1.8055E+11, 1.8155E+11, 1.8256E+11, 1.8356E+11,
     +  1.8458E+11, 1.8560E+11, 1.8662E+11, 1.8765E+11, 1.8869E+11,
     + 1.8973E+11, 1.9077E+11, 1.9182E+11, 1.9288E+11, 1.9394E+11,
     +  1.9501E+11, 1.9608E+11, 1.9716E+11, 1.9824E+11, 1.9933E+11,
     + 2.0043E+11, 2.0153E+11, 2.0263E+11, 2.0374E+11, 2.0486E+11,
     +  2.0598E+11, 2.0711E+11, 2.0824E+11, 2.0938E+11, 2.1053E+11,
     + 2.1168E+11, 2.1284E+11, 2.1400E+11, 2.1517E+11, 2.1634E+11,
     +  2.1752E+11, 2.1871E+11, 2.1990E+11, 2.2110E+11, 2.2230E+11,
     + 2.2351E+11, 2.2473E+11, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 3,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --     12324
      data (QofT( 4,J),J=1,2408)/ 5.9387E+01, 1.1450E+02, 1.6968E+02,
     +  2.2487E+02, 2.8007E+02, 3.3528E+02, 3.9049E+02, 4.4570E+02,
     + 5.0091E+02, 5.5612E+02, 6.1134E+02, 6.6655E+02, 7.2177E+02,
     +  7.7699E+02, 8.3221E+02, 8.8744E+02, 9.4266E+02, 9.9788E+02,
     + 1.0531E+03, 1.1083E+03, 1.2188E+03, 1.3293E+03, 1.4397E+03,
     +  1.5502E+03, 1.6608E+03, 1.7713E+03, 1.8819E+03, 1.9927E+03,
     + 2.1035E+03, 2.2146E+03, 2.3259E+03, 2.4375E+03, 2.5495E+03,
     +  2.6619E+03, 2.7749E+03, 2.8884E+03, 3.0027E+03, 3.1178E+03,
     + 3.2337E+03, 3.3506E+03, 3.4685E+03, 3.5876E+03, 3.7079E+03,
     +  3.8295E+03, 3.9526E+03, 4.0771E+03, 4.2031E+03, 4.3309E+03,
     + 4.4603E+03, 4.5916E+03, 4.7247E+03, 4.8598E+03, 4.9968E+03,
     +  5.1361E+03, 5.2775E+03, 5.4211E+03, 5.5670E+03, 5.7154E+03,
     + 5.8662E+03, 6.0195E+03, 6.1754E+03, 6.3341E+03, 6.4954E+03,
     +  6.6596E+03, 6.8267E+03, 6.9967E+03, 7.1699E+03, 7.3460E+03,
     + 7.5254E+03, 7.7081E+03, 7.8941E+03, 8.0835E+03, 8.2764E+03,
     +  8.4730E+03, 8.6731E+03, 8.8770E+03, 9.0848E+03, 9.2964E+03,
     + 9.5121E+03, 9.7318E+03, 9.9557E+03, 1.0184E+04, 1.0416E+04,
     +  1.0653E+04, 1.0895E+04, 1.1141E+04, 1.1392E+04, 1.1648E+04,
     + 1.1908E+04, 1.2174E+04, 1.2445E+04, 1.2721E+04, 1.3003E+04,
     +  1.3290E+04, 1.3582E+04, 1.3880E+04, 1.4184E+04, 1.4495E+04,
     + 1.4811E+04, 1.5133E+04, 1.5461E+04, 1.5796E+04, 1.6137E+04,
     +  1.6485E+04, 1.6840E+04, 1.7202E+04, 1.7571E+04, 1.7947E+04,
     + 1.8331E+04, 1.8721E+04, 1.9120E+04, 1.9526E+04, 1.9941E+04,
     +  2.0363E+04, 2.0793E+04, 2.1232E+04, 2.1680E+04, 2.2136E+04,
     + 2.2601E+04, 2.3075E+04, 2.3558E+04, 2.4051E+04, 2.4553E+04,
     +  2.5065E+04, 2.5587E+04, 2.6119E+04, 2.6661E+04, 2.7214E+04,
     + 2.7777E+04, 2.8351E+04, 2.8936E+04, 2.9533E+04, 3.0141E+04,
     +  3.0760E+04, 3.1392E+04, 3.2035E+04, 3.2691E+04, 3.3359E+04,
     + 3.4040E+04, 3.4734E+04, 3.5441E+04, 3.6162E+04, 3.6896E+04,
     +  3.7644E+04, 3.8406E+04, 3.9182E+04, 3.9974E+04, 4.0780E+04,
     + 4.1601E+04, 4.2438E+04, 4.3290E+04, 4.4158E+04, 4.5042E+04,
     +  4.5943E+04, 4.6861E+04, 4.7796E+04, 4.8748E+04, 4.9718E+04,
     + 5.0705E+04, 5.1711E+04, 5.2736E+04, 5.3779E+04, 5.4842E+04,
     +  5.5924E+04, 5.7026E+04, 5.8148E+04, 5.9290E+04, 6.0453E+04,
     + 6.1638E+04, 6.2844E+04, 6.4072E+04, 6.5322E+04, 6.6595E+04,
     +  6.7891E+04, 6.9210E+04, 7.0553E+04, 7.1920E+04, 7.3311E+04,
     + 7.4727E+04, 7.6169E+04, 7.7636E+04, 7.9130E+04, 8.0650E+04,
     +  8.2197E+04, 8.3772E+04, 8.5374E+04, 8.7005E+04, 8.8664E+04,
     + 9.0353E+04, 9.2071E+04, 9.3819E+04, 9.5598E+04, 9.7408E+04,
     +  9.9250E+04, 1.0112E+05, 1.0303E+05, 1.0497E+05, 1.0694E+05,
     + 1.0895E+05, 1.1099E+05, 1.1307E+05, 1.1518E+05, 1.1733E+05,
     +  1.1952E+05, 1.2174E+05, 1.2400E+05, 1.2630E+05, 1.2864E+05,
     + 1.3102E+05, 1.3344E+05, 1.3590E+05, 1.3840E+05, 1.4094E+05,
     +  1.4353E+05, 1.4616E+05, 1.4883E+05, 1.5155E+05, 1.5432E+05,
     + 1.5713E+05, 1.5998E+05, 1.6289E+05, 1.6584E+05, 1.6885E+05,
     +  1.7190E+05, 1.7500E+05, 1.7815E+05, 1.8136E+05, 1.8462E+05,
     + 1.8793E+05, 1.9130E+05, 1.9472E+05, 1.9820E+05, 2.0173E+05,
     +  2.0532E+05, 2.0897E+05, 2.1268E+05, 2.1645E+05, 2.2028E+05,
     + 2.2417E+05, 2.2813E+05, 2.3214E+05, 2.3623E+05, 2.4037E+05,
     +  2.4459E+05, 2.4887E+05, 2.5322E+05, 2.5764E+05, 2.6213E+05,
     + 2.6669E+05, 2.7132E+05, 2.7603E+05, 2.8081E+05, 2.8567E+05,
     +  2.9060E+05, 2.9561E+05, 3.0070E+05, 3.0587E+05, 3.1112E+05,
     + 3.1645E+05, 3.2186E+05, 3.2736E+05, 3.3294E+05, 3.3861E+05,
     +  3.4437E+05, 3.5022E+05, 3.5616E+05, 3.6219E+05, 3.6831E+05,
     + 3.7452E+05, 3.8083E+05, 3.8724E+05, 3.9374E+05, 4.0035E+05,
     +  4.0705E+05, 4.1386E+05, 4.2077E+05, 4.2779E+05, 4.3491E+05,
     + 4.4214E+05, 4.4947E+05, 4.5692E+05, 4.6448E+05, 4.7216E+05,
     +  4.7994E+05, 4.8785E+05, 4.9587E+05, 5.0402E+05, 5.1228E+05,
     + 5.2067E+05, 5.2918E+05, 5.3782E+05, 5.4658E+05, 5.5548E+05,
     +  5.6450E+05, 5.7366E+05, 5.8296E+05, 5.9239E+05, 6.0195E+05,
     + 6.1166E+05, 6.2151E+05, 6.3151E+05, 6.4165E+05, 6.5193E+05,
     +  6.6237E+05, 6.7296E+05, 6.8370E+05, 6.9460E+05, 7.0565E+05,
     + 7.1687E+05, 7.2824E+05, 7.3978E+05, 7.5148E+05, 7.6335E+05,
     +  7.7540E+05, 7.8761E+05, 8.0000E+05, 8.1256E+05, 8.2530E+05,
     + 8.3823E+05, 8.5133E+05, 8.6462E+05, 8.7810E+05, 8.9177E+05,
     +  9.0564E+05, 9.1969E+05, 9.3395E+05, 9.4841E+05, 9.6306E+05,
     + 9.7793E+05, 9.9300E+05, 1.0083E+06, 1.0238E+06, 1.0395E+06,
     +  1.0554E+06, 1.0716E+06, 1.0879E+06, 1.1045E+06, 1.1214E+06,
     + 1.1384E+06, 1.1557E+06, 1.1732E+06, 1.1910E+06, 1.2090E+06,
     +  1.2273E+06, 1.2458E+06, 1.2646E+06, 1.2836E+06, 1.3029E+06,
     + 1.3224E+06, 1.3422E+06, 1.3623E+06, 1.3826E+06, 1.4032E+06,
     +  1.4241E+06, 1.4453E+06, 1.4668E+06, 1.4885E+06, 1.5106E+06,
     + 1.5329E+06, 1.5556E+06, 1.5785E+06, 1.6018E+06, 1.6253E+06,
     +  1.6492E+06, 1.6734E+06, 1.6979E+06, 1.7227E+06, 1.7479E+06,
     + 1.7734E+06, 1.7992E+06, 1.8254E+06, 1.8519E+06, 1.8788E+06,
     +  1.9060E+06, 1.9335E+06, 1.9615E+06, 1.9898E+06, 2.0184E+06,
     + 2.0475E+06, 2.0769E+06, 2.1067E+06, 2.1369E+06, 2.1675E+06,
     +  2.1984E+06, 2.2298E+06, 2.2616E+06, 2.2938E+06, 2.3264E+06,
     + 2.3594E+06, 2.3929E+06, 2.4267E+06, 2.4611E+06, 2.4958E+06,
     +  2.5310E+06, 2.5666E+06, 2.6027E+06, 2.6393E+06, 2.6763E+06,
     + 2.7138E+06, 2.7518E+06, 2.7902E+06, 2.8291E+06, 2.8686E+06,
     +  2.9085E+06, 2.9489E+06, 2.9898E+06, 3.0313E+06, 3.0732E+06,
     + 3.1157E+06, 3.1587E+06, 3.2023E+06, 3.2464E+06, 3.2910E+06,
     +  3.3362E+06, 3.3820E+06, 3.4283E+06, 3.4752E+06, 3.5227E+06,
     + 3.5708E+06, 3.6195E+06, 3.6687E+06, 3.7186E+06, 3.7691E+06,
     +  3.8202E+06, 3.8719E+06, 3.9243E+06, 3.9773E+06, 4.0310E+06,
     + 4.0853E+06, 4.1402E+06, 4.1959E+06, 4.2522E+06, 4.3092E+06,
     +  4.3669E+06, 4.4253E+06, 4.4844E+06, 4.5442E+06, 4.6047E+06,
     + 4.6660E+06, 4.7280E+06, 4.7907E+06, 4.8542E+06, 4.9185E+06,
     +  4.9835E+06, 5.0493E+06, 5.1159E+06, 5.1833E+06, 5.2515E+06,
     + 5.3205E+06, 5.3903E+06, 5.4610E+06, 5.5325E+06, 5.6048E+06,
     +  5.6780E+06, 5.7521E+06, 5.8270E+06, 5.9028E+06, 5.9795E+06,
     + 6.0571E+06, 6.1357E+06, 6.2151E+06, 6.2955E+06, 6.3768E+06,
     +  6.4591E+06, 6.5423E+06, 6.6265E+06, 6.7117E+06, 6.7978E+06,
     + 6.8850E+06, 6.9732E+06, 7.0624E+06, 7.1526E+06, 7.2439E+06,
     +  7.3362E+06, 7.4296E+06, 7.5241E+06, 7.6196E+06, 7.7163E+06,
     + 7.8141E+06, 7.9129E+06, 8.0130E+06, 8.1141E+06, 8.2164E+06,
     +  8.3199E+06, 8.4246E+06, 8.5305E+06, 8.6375E+06, 8.7458E+06,
     + 8.8553E+06, 8.9661E+06, 9.0781E+06, 9.1913E+06, 9.3059E+06,
     +  9.4217E+06, 9.5389E+06, 9.6573E+06, 9.7771E+06, 9.8983E+06,
     + 1.0021E+07, 1.0145E+07, 1.0270E+07, 1.0397E+07, 1.0525E+07,
     +  1.0654E+07, 1.0785E+07, 1.0917E+07, 1.1051E+07, 1.1187E+07,
     + 1.1324E+07, 1.1462E+07, 1.1602E+07, 1.1743E+07, 1.1886E+07,
     +  1.2031E+07, 1.2177E+07, 1.2325E+07, 1.2474E+07, 1.2625E+07,
     + 1.2778E+07, 1.2932E+07, 1.3088E+07, 1.3246E+07, 1.3406E+07,
     +  1.3567E+07, 1.3730E+07, 1.3895E+07, 1.4061E+07, 1.4230E+07,
     + 1.4400E+07, 1.4572E+07, 1.4746E+07, 1.4921E+07, 1.5099E+07,
     +  1.5279E+07, 1.5460E+07, 1.5644E+07, 1.5829E+07, 1.6016E+07,
     + 1.6206E+07, 1.6397E+07, 1.6591E+07, 1.6786E+07, 1.6984E+07,
     +  1.7184E+07, 1.7386E+07, 1.7590E+07, 1.7796E+07, 1.8004E+07,
     + 1.8215E+07, 1.8428E+07, 1.8643E+07, 1.8860E+07, 1.9080E+07,
     +  1.9302E+07, 1.9526E+07, 1.9753E+07, 1.9982E+07, 2.0214E+07,
     + 2.0448E+07, 2.0684E+07, 2.0923E+07, 2.1164E+07, 2.1408E+07,
     +  2.1654E+07, 2.1903E+07, 2.2155E+07, 2.2409E+07, 2.2666E+07,
     + 2.2925E+07, 2.3187E+07, 2.3452E+07, 2.3720E+07, 2.3990E+07,
     +  2.4263E+07, 2.4539E+07, 2.4818E+07, 2.5100E+07, 2.5385E+07,
     + 2.5672E+07, 2.5963E+07, 2.6256E+07, 2.6553E+07, 2.6852E+07,
     +  2.7155E+07, 2.7460E+07, 2.7769E+07, 2.8081E+07, 2.8396E+07,
     + 2.8714E+07, 2.9036E+07, 2.9360E+07, 2.9689E+07, 3.0020E+07,
     +  3.0355E+07, 3.0693E+07, 3.1034E+07, 3.1379E+07, 3.1728E+07,
     + 3.2080E+07, 3.2435E+07, 3.2794E+07, 3.3157E+07, 3.3523E+07,
     +  3.3893E+07, 3.4267E+07, 3.4644E+07, 3.5025E+07, 3.5410E+07,
     + 3.5799E+07, 3.6191E+07, 3.6588E+07, 3.6988E+07, 3.7393E+07,
     +  3.7801E+07, 3.8214E+07, 3.8630E+07, 3.9051E+07, 3.9476E+07,
     + 3.9905E+07, 4.0338E+07, 4.0775E+07, 4.1217E+07, 4.1663E+07,
     +  4.2114E+07, 4.2569E+07, 4.3028E+07, 4.3492E+07, 4.3960E+07,
     + 4.4433E+07, 4.4911E+07, 4.5393E+07, 4.5880E+07, 4.6372E+07,
     +  4.6868E+07, 4.7369E+07, 4.7875E+07, 4.8386E+07, 4.8902E+07,
     + 4.9423E+07, 4.9949E+07, 5.0480E+07, 5.1016E+07, 5.1557E+07,
     +  5.2104E+07, 5.2656E+07, 5.3213E+07, 5.3775E+07, 5.4343E+07,
     + 5.4916E+07, 5.5494E+07, 5.6079E+07, 5.6668E+07, 5.7264E+07,
     +  5.7865E+07, 5.8471E+07, 5.9084E+07, 5.9702E+07, 6.0326E+07,
     + 6.0957E+07, 6.1593E+07, 6.2235E+07, 6.2883E+07, 6.3537E+07,
     +  6.4197E+07, 6.4864E+07, 6.5537E+07, 6.6216E+07, 6.6902E+07,
     + 6.7594E+07, 6.8292E+07, 6.8997E+07, 6.9709E+07, 7.0427E+07,
     +  7.1152E+07, 7.1884E+07, 7.2623E+07, 7.3368E+07, 7.4120E+07,
     + 7.4880E+07, 7.5646E+07, 7.6420E+07, 7.7200E+07, 7.7988E+07,
     +  7.8784E+07, 7.9586E+07, 8.0396E+07, 8.1213E+07, 8.2038E+07,
     + 8.2871E+07, 8.3711E+07, 8.4559E+07, 8.5415E+07, 8.6278E+07,
     +  8.7150E+07, 8.8029E+07, 8.8917E+07, 8.9812E+07, 9.0716E+07,
     + 9.1628E+07, 9.2548E+07, 9.3477E+07, 9.4414E+07, 9.5360E+07,
     +  9.6314E+07, 9.7277E+07, 9.8248E+07, 9.9229E+07, 1.0022E+08,
     + 1.0122E+08, 1.0222E+08, 1.0324E+08, 1.0426E+08, 1.0530E+08,
     +  1.0634E+08, 1.0740E+08, 1.0846E+08, 1.0953E+08, 1.1061E+08,
     + 1.1171E+08, 1.1281E+08, 1.1392E+08, 1.1504E+08, 1.1617E+08,
     +  1.1731E+08, 1.1846E+08, 1.1962E+08, 1.2080E+08, 1.2198E+08,
     + 1.2317E+08, 1.2437E+08, 1.2559E+08, 1.2681E+08, 1.2805E+08,
     +  1.2929E+08, 1.3055E+08, 1.3182E+08, 1.3310E+08, 1.3439E+08,
     + 1.3569E+08, 1.3701E+08, 1.3833E+08, 1.3967E+08, 1.4102E+08,
     +  1.4238E+08, 1.4375E+08, 1.4513E+08, 1.4653E+08, 1.4794E+08,
     + 1.4936E+08, 1.5079E+08, 1.5224E+08, 1.5370E+08, 1.5517E+08,
     +  1.5665E+08, 1.5815E+08, 1.5965E+08, 1.6118E+08, 1.6271E+08,
     + 1.6426E+08, 1.6582E+08, 1.6740E+08, 1.6899E+08, 1.7059E+08,
     +  1.7221E+08, 1.7384E+08, 1.7548E+08, 1.7714E+08, 1.7881E+08,
     + 1.8050E+08, 1.8220E+08, 1.8392E+08, 1.8565E+08, 1.8739E+08,
     +  1.8915E+08, 1.9093E+08, 1.9272E+08, 1.9452E+08, 1.9634E+08,
     + 1.9818E+08, 2.0003E+08, 2.0190E+08, 2.0378E+08, 2.0568E+08,
     +  2.0760E+08, 2.0953E+08, 2.1147E+08, 2.1344E+08, 2.1542E+08,
     + 2.1741E+08, 2.1943E+08, 2.2146E+08, 2.2351E+08, 2.2557E+08,
     +  2.2765E+08, 2.2975E+08, 2.3187E+08, 2.3401E+08, 2.3616E+08,
     + 2.3833E+08, 2.4052E+08, 2.4272E+08, 2.4495E+08, 2.4719E+08,
     +  2.4945E+08, 2.5173E+08, 2.5403E+08, 2.5635E+08, 2.5869E+08,
     + 2.6105E+08, 2.6343E+08, 2.6582E+08, 2.6824E+08, 2.7067E+08,
     +  2.7313E+08, 2.7561E+08, 2.7810E+08, 2.8062E+08, 2.8316E+08,
     + 2.8572E+08, 2.8829E+08, 2.9089E+08, 2.9352E+08, 2.9616E+08,
     +  2.9882E+08, 3.0151E+08, 3.0422E+08, 3.0695E+08, 3.0970E+08,
     + 3.1248E+08, 3.1527E+08, 3.1809E+08, 3.2094E+08, 3.2380E+08,
     +  3.2669E+08, 3.2960E+08, 3.3254E+08, 3.3550E+08, 3.3848E+08,
     + 3.4149E+08, 3.4452E+08, 3.4758E+08, 3.5066E+08, 3.5376E+08,
     +  3.5689E+08, 3.6005E+08, 3.6323E+08, 3.6643E+08, 3.6966E+08,
     + 3.7292E+08, 3.7620E+08, 3.7951E+08, 3.8285E+08, 3.8621E+08,
     +  3.8960E+08, 3.9302E+08, 3.9646E+08, 3.9993E+08, 4.0343E+08,
     + 4.0695E+08, 4.1051E+08, 4.1409E+08, 4.1770E+08, 4.2134E+08,
     +  4.2500E+08, 4.2870E+08, 4.3242E+08, 4.3618E+08, 4.3996E+08,
     + 4.4377E+08, 4.4762E+08, 4.5149E+08, 4.5539E+08, 4.5933E+08,
     +  4.6329E+08, 4.6729E+08, 4.7132E+08, 4.7538E+08, 4.7947E+08,
     + 4.8359E+08, 4.8774E+08, 4.9193E+08, 4.9615E+08, 5.0040E+08,
     +  5.0469E+08, 5.0900E+08, 5.1336E+08, 5.1774E+08, 5.2216E+08,
     + 5.2661E+08, 5.3110E+08, 5.3562E+08, 5.4018E+08, 5.4477E+08,
     +  5.4940E+08, 5.5406E+08, 5.5876E+08, 5.6350E+08, 5.6827E+08,
     + 5.7308E+08, 5.7792E+08, 5.8280E+08, 5.8772E+08, 5.9268E+08,
     +  5.9767E+08, 6.0271E+08, 6.0778E+08, 6.1289E+08, 6.1803E+08,
     + 6.2322E+08, 6.2845E+08, 6.3371E+08, 6.3902E+08, 6.4437E+08,
     +  6.4976E+08, 6.5518E+08, 6.6065E+08, 6.6616E+08, 6.7171E+08,
     + 6.7731E+08, 6.8294E+08, 6.8862E+08, 6.9434E+08, 7.0011E+08,
     +  7.0591E+08, 7.1177E+08, 7.1766E+08, 7.2360E+08, 7.2958E+08,
     + 7.3561E+08, 7.4168E+08, 7.4780E+08, 7.5397E+08, 7.6018E+08,
     +  7.6643E+08, 7.7274E+08, 7.7909E+08, 7.8548E+08, 7.9193E+08,
     + 7.9842E+08, 8.0496E+08, 8.1155E+08, 8.1819E+08, 8.2488E+08,
     +  8.3161E+08, 8.3840E+08, 8.4524E+08, 8.5212E+08, 8.5906E+08,
     + 8.6605E+08, 8.7309E+08, 8.8018E+08, 8.8733E+08, 8.9452E+08,
     +  9.0177E+08, 9.0907E+08, 9.1643E+08, 9.2384E+08, 9.3130E+08,
     + 9.3882E+08, 9.4640E+08, 9.5402E+08, 9.6171E+08, 9.6945E+08,
     +  9.7725E+08, 9.8510E+08, 9.9301E+08, 1.0010E+09, 1.0090E+09,
     + 1.0171E+09, 1.0252E+09, 1.0334E+09, 1.0417E+09, 1.0500E+09,
     +  1.0584E+09, 1.0668E+09, 1.0753E+09, 1.0839E+09, 1.0925E+09,
     + 1.1012E+09, 1.1100E+09, 1.1188E+09, 1.1276E+09, 1.1366E+09,
     +  1.1456E+09, 1.1547E+09, 1.1638E+09, 1.1730E+09, 1.1823E+09,
     + 1.1916E+09, 1.2010E+09, 1.2104E+09, 1.2200E+09, 1.2296E+09,
     +  1.2392E+09, 1.2490E+09, 1.2588E+09, 1.2687E+09, 1.2786E+09,
     + 1.2886E+09, 1.2987E+09, 1.3089E+09, 1.3191E+09, 1.3294E+09,
     +  1.3398E+09, 1.3502E+09, 1.3607E+09, 1.3713E+09, 1.3820E+09,
     + 1.3927E+09, 1.4036E+09, 1.4145E+09, 1.4254E+09, 1.4365E+09,
     +  1.4476E+09, 1.4588E+09, 1.4701E+09, 1.4814E+09, 1.4929E+09,
     + 1.5044E+09, 1.5160E+09, 1.5277E+09, 1.5395E+09, 1.5513E+09,
     +  1.5632E+09, 1.5752E+09, 1.5873E+09, 1.5995E+09, 1.6118E+09,
     + 1.6241E+09, 1.6366E+09, 1.6491E+09, 1.6617E+09, 1.6744E+09,
     +  1.6872E+09, 1.7000E+09, 1.7130E+09, 1.7260E+09, 1.7392E+09,
     + 1.7524E+09, 1.7657E+09, 1.7791E+09, 1.7926E+09, 1.8062E+09,
     +  1.8199E+09, 1.8337E+09, 1.8476E+09, 1.8615E+09, 1.8756E+09,
     + 1.8898E+09, 1.9040E+09, 1.9184E+09, 1.9328E+09, 1.9474E+09,
     +  1.9620E+09, 1.9768E+09, 1.9916E+09, 2.0066E+09, 2.0216E+09,
     + 2.0368E+09, 2.0520E+09, 2.0674E+09, 2.0829E+09, 2.0984E+09,
     +  2.1141E+09, 2.1299E+09, 2.1458E+09, 2.1618E+09, 2.1779E+09,
     + 2.1941E+09, 2.2104E+09, 2.2268E+09, 2.2433E+09, 2.2600E+09,
     +  2.2768E+09, 2.2936E+09, 2.3106E+09, 2.3277E+09, 2.3449E+09,
     + 2.3622E+09, 2.3797E+09, 2.3973E+09, 2.4149E+09, 2.4327E+09,
     +  2.4506E+09, 2.4687E+09, 2.4868E+09, 2.5051E+09, 2.5235E+09,
     + 2.5420E+09, 2.5606E+09, 2.5794E+09, 2.5983E+09, 2.6173E+09,
     +  2.6364E+09, 2.6557E+09, 2.6751E+09, 2.6946E+09, 2.7143E+09,
     + 2.7340E+09, 2.7539E+09, 2.7740E+09, 2.7941E+09, 2.8144E+09,
     +  2.8349E+09, 2.8554E+09, 2.8761E+09, 2.8970E+09, 2.9180E+09,
     + 2.9391E+09, 2.9603E+09, 2.9817E+09, 3.0032E+09, 3.0249E+09,
     +  3.0467E+09, 3.0687E+09, 3.0907E+09, 3.1130E+09, 3.1354E+09,
     + 3.1579E+09, 3.1806E+09, 3.2034E+09, 3.2263E+09, 3.2495E+09,
     +  3.2727E+09, 3.2961E+09, 3.3197E+09, 3.3434E+09, 3.3673E+09,
     + 3.3913E+09, 3.4155E+09, 3.4398E+09, 3.4643E+09, 3.4890E+09,
     +  3.5138E+09, 3.5388E+09, 3.5639E+09, 3.5892E+09, 3.6146E+09,
     + 3.6402E+09, 3.6660E+09, 3.6920E+09, 3.7181E+09, 3.7443E+09,
     +  3.7708E+09, 3.7974E+09, 3.8242E+09, 3.8511E+09, 3.8782E+09,
     + 3.9055E+09, 3.9330E+09, 3.9606E+09, 3.9885E+09, 4.0165E+09,
     +  4.0446E+09, 4.0730E+09, 4.1015E+09, 4.1302E+09, 4.1591E+09,
     + 4.1882E+09, 4.2174E+09, 4.2469E+09, 4.2765E+09, 4.3063E+09,
     +  4.3363E+09, 4.3665E+09, 4.3969E+09, 4.4274E+09, 4.4582E+09,
     + 4.4891E+09, 4.5203E+09, 4.5516E+09, 4.5831E+09, 4.6149E+09,
     +  4.6468E+09, 4.6789E+09, 4.7113E+09, 4.7438E+09, 4.7765E+09,
     + 4.8095E+09, 4.8426E+09, 4.8760E+09, 4.9095E+09, 4.9433E+09,
     +  4.9773E+09, 5.0114E+09, 5.0458E+09, 5.0804E+09, 5.1153E+09,
     + 5.1503E+09, 5.1856E+09, 5.2210E+09, 5.2567E+09, 5.2926E+09,
     +  5.3288E+09, 5.3651E+09, 5.4017E+09, 5.4385E+09, 5.4755E+09,
     + 5.5128E+09, 5.5503E+09, 5.5880E+09, 5.6259E+09, 5.6641E+09,
     +  5.7025E+09, 5.7412E+09, 5.7800E+09, 5.8192E+09, 5.8585E+09,
     + 5.8981E+09, 5.9380E+09, 5.9780E+09, 6.0184E+09, 6.0589E+09,
     +  6.0997E+09, 6.1408E+09, 6.1821E+09, 6.2237E+09, 6.2655E+09,
     + 6.3076E+09, 6.3499E+09, 6.3925E+09, 6.4353E+09, 6.4784E+09,
     +  6.5218E+09, 6.5654E+09, 6.6093E+09, 6.6534E+09, 6.6978E+09,
     + 6.7425E+09, 6.7874E+09, 6.8327E+09, 6.8781E+09, 6.9239E+09,
     +  6.9699E+09, 7.0162E+09, 7.0628E+09, 7.1097E+09, 7.1568E+09,
     + 7.2043E+09, 7.2520E+09, 7.3000E+09, 7.3483E+09, 7.3968E+09,
     +  7.4457E+09, 7.4948E+09, 7.5443E+09, 7.5940E+09, 7.6440E+09,
     + 7.6944E+09, 7.7450E+09, 7.7959E+09, 7.8471E+09, 7.8987E+09,
     +  7.9505E+09, 8.0026E+09, 8.0551E+09, 8.1079E+09, 8.1609E+09,
     + 8.2143E+09, 8.2680E+09, 8.3220E+09, 8.3763E+09, 8.4310E+09,
     +  8.4860E+09, 8.5413E+09, 8.5969E+09, 8.6528E+09, 8.7091E+09,
     + 8.7657E+09, 8.8226E+09, 8.8799E+09, 8.9375E+09, 8.9954E+09,
     +  9.0537E+09, 9.1123E+09, 9.1713E+09, 9.2306E+09, 9.2902E+09,
     + 9.3502E+09, 9.4106E+09, 9.4712E+09, 9.5323E+09, 9.5937E+09,
     +  9.6554E+09, 9.7175E+09, 9.7800E+09, 9.8428E+09, 9.9060E+09,
     + 9.9696E+09, 1.0034E+10, 1.0098E+10, 1.0162E+10, 1.0228E+10,
     +  1.0293E+10, 1.0359E+10, 1.0425E+10, 1.0491E+10, 1.0558E+10,
     + 1.0626E+10, 1.0693E+10, 1.0761E+10, 1.0830E+10, 1.0899E+10,
     +  1.0968E+10, 1.1038E+10, 1.1108E+10, 1.1178E+10, 1.1249E+10,
     + 1.1320E+10, 1.1392E+10, 1.1464E+10, 1.1537E+10, 1.1610E+10,
     +  1.1683E+10, 1.1757E+10, 1.1831E+10, 1.1905E+10, 1.1980E+10,
     + 1.2056E+10, 1.2132E+10, 1.2208E+10, 1.2285E+10, 1.2362E+10,
     +  1.2439E+10, 1.2517E+10, 1.2596E+10, 1.2675E+10, 1.2754E+10,
     + 1.2834E+10, 1.2914E+10, 1.2995E+10, 1.3076E+10, 1.3158E+10,
     +  1.3240E+10, 1.3322E+10, 1.3405E+10, 1.3489E+10, 1.3572E+10,
     + 1.3657E+10, 1.3742E+10, 1.3827E+10, 1.3913E+10, 1.3999E+10,
     +  1.4086E+10, 1.4173E+10, 1.4261E+10, 1.4349E+10, 1.4438E+10,
     + 1.4527E+10, 1.4617E+10, 1.4707E+10, 1.4798E+10, 1.4889E+10,
     +  1.4980E+10, 1.5073E+10, 1.5165E+10, 1.5259E+10, 1.5352E+10,
     + 1.5447E+10, 1.5541E+10, 1.5637E+10, 1.5733E+10, 1.5829E+10,
     +  1.5926E+10, 1.6023E+10, 1.6121E+10, 1.6220E+10, 1.6319E+10,
     + 1.6418E+10, 1.6518E+10, 1.6619E+10, 1.6720E+10, 1.6822E+10,
     +  1.6924E+10, 1.7027E+10, 1.7131E+10, 1.7235E+10, 1.7339E+10,
     + 1.7445E+10, 1.7550E+10, 1.7657E+10, 1.7764E+10, 1.7871E+10,
     +  1.7979E+10, 1.8088E+10, 1.8197E+10, 1.8307E+10, 1.8417E+10,
     + 1.8528E+10, 1.8640E+10, 1.8752E+10, 1.8865E+10, 1.8978E+10,
     +  1.9092E+10, 1.9207E+10, 1.9322E+10, 1.9438E+10, 1.9555E+10,
     + 1.9672E+10, 1.9790E+10, 1.9908E+10, 2.0027E+10, 2.0147E+10,
     +  2.0267E+10, 2.0388E+10, 2.0510E+10, 2.0632E+10, 2.0755E+10,
     + 2.0879E+10, 2.1003E+10, 2.1128E+10, 2.1253E+10, 2.1380E+10,
     +  2.1507E+10, 2.1634E+10, 2.1762E+10, 2.1891E+10, 2.2021E+10,
     + 2.2151E+10, 2.2282E+10, 2.2414E+10, 2.2547E+10, 2.2680E+10,
     +  2.2814E+10, 2.2948E+10, 2.3083E+10, 2.3219E+10, 2.3356E+10,
     + 2.3493E+10, 2.3632E+10, 2.3770E+10, 2.3910E+10, 2.4050E+10,
     +  2.4191E+10, 2.4333E+10, 2.4476E+10, 2.4619E+10, 2.4763E+10,
     + 2.4908E+10, 2.5054E+10, 2.5200E+10, 2.5347E+10, 2.5495E+10,
     +  2.5644E+10, 2.5793E+10, 2.5943E+10, 2.6094E+10, 2.6246E+10,
     + 2.6399E+10, 2.6552E+10, 2.6706E+10, 2.6861E+10, 2.7017E+10,
     +  2.7173E+10, 2.7331E+10, 2.7489E+10, 2.7648E+10, 2.7808E+10,
     + 2.7969E+10, 2.8130E+10, 2.8293E+10, 2.8456E+10, 2.8620E+10,
     +  2.8785E+10, 2.8950E+10, 2.9117E+10, 2.9285E+10, 2.9453E+10,
     + 2.9622E+10, 2.9792E+10, 2.9963E+10, 3.0135E+10, 3.0308E+10,
     +  3.0481E+10, 3.0656E+10, 3.0831E+10, 3.1007E+10, 3.1184E+10,
     + 3.1362E+10, 3.1541E+10, 3.1721E+10, 3.1902E+10, 3.2084E+10,
     +  3.2267E+10, 3.2450E+10, 3.2635E+10, 3.2820E+10, 3.3007E+10,
     + 3.3194E+10, 3.3382E+10, 3.3572E+10, 3.3762E+10, 3.3953E+10,
     +  3.4145E+10, 3.4338E+10, 3.4533E+10, 3.4728E+10, 3.4924E+10,
     + 3.5121E+10, 3.5319E+10, 3.5518E+10, 3.5718E+10, 3.5919E+10,
     +  3.6121E+10, 3.6325E+10, 3.6529E+10, 3.6734E+10, 3.6940E+10,
     + 3.7147E+10, 3.7356E+10, 3.7565E+10, 3.7775E+10, 3.7987E+10,
     +  3.8199E+10, 3.8413E+10, 3.8628E+10, 3.8843E+10, 3.9060E+10,
     + 3.9278E+10, 3.9497E+10, 3.9717E+10, 3.9938E+10, 4.0161E+10,
     +  4.0384E+10, 4.0608E+10, 4.0834E+10, 4.1061E+10, 4.1288E+10,
     + 4.1517E+10, 4.1748E+10, 4.1979E+10, 4.2211E+10, 4.2445E+10,
     +  4.2679E+10, 4.2915E+10, 4.3152E+10, 4.3391E+10, 4.3630E+10,
     + 4.3870E+10, 4.4112E+10, 4.4355E+10, 4.4599E+10, 4.4845E+10,
     +  4.5091E+10, 4.5339E+10, 4.5588E+10, 4.5838E+10, 4.6090E+10,
     + 4.6342E+10, 4.6596E+10, 4.6851E+10, 4.7108E+10, 4.7365E+10,
     +  4.7624E+10, 4.7884E+10, 4.8146E+10, 4.8408E+10, 4.8672E+10,
     + 4.8938E+10, 4.9204E+10, 4.9472E+10, 4.9741E+10, 5.0012E+10,
     +  5.0284E+10, 5.0557E+10, 5.0831E+10, 5.1107E+10, 5.1384E+10,
     + 5.1662E+10, 5.1942E+10, 5.2223E+10, 5.2506E+10, 5.2790E+10,
     +  5.3075E+10, 5.3362E+10, 5.3650E+10, 5.3939E+10, 5.4230E+10,
     + 5.4522E+10, 5.4816E+10, 5.5111E+10, 5.5407E+10, 5.5705E+10,
     +  5.6004E+10, 5.6305E+10, 5.6607E+10, 5.6911E+10, 5.7216E+10,
     + 5.7522E+10, 5.7830E+10, 5.8140E+10, 5.8451E+10, 5.8763E+10,
     +  5.9077E+10, 5.9393E+10, 5.9709E+10, 6.0028E+10, 6.0348E+10,
     + 6.0669E+10, 6.0992E+10, 6.1317E+10, 6.1643E+10, 6.1971E+10,
     +  6.2300E+10, 6.2630E+10, 6.2963E+10, 6.3297E+10, 6.3632E+10,
     + 6.3969E+10, 6.4308E+10, 6.4648E+10, 6.4990E+10, 6.5333E+10,
     +  6.5678E+10, 6.6025E+10, 6.6373E+10, 6.6723E+10, 6.7075E+10,
     + 6.7428E+10, 6.7783E+10, 6.8140E+10, 6.8498E+10, 6.8858E+10,
     +  6.9220E+10, 6.9583E+10, 6.9948E+10, 7.0315E+10, 7.0683E+10,
     + 7.1054E+10, 7.1425E+10, 7.1799E+10, 7.2174E+10, 7.2552E+10,
     +  7.2931E+10, 7.3311E+10, 7.3694E+10, 7.4078E+10, 7.4464E+10,
     + 7.4852E+10, 7.5241E+10, 7.5633E+10, 7.6026E+10, 7.6421E+10,
     +  7.6818E+10, 7.7216E+10, 7.7617E+10, 7.8019E+10, 7.8424E+10,
     + 7.8830E+10, 7.9238E+10, 7.9648E+10, 8.0059E+10, 8.0473E+10,
     +  8.0889E+10, 8.1306E+10, 8.1726E+10, 8.2147E+10, 8.2570E+10,
     + 8.2995E+10, 8.3423E+10, 8.3852E+10, 8.4283E+10, 8.4716E+10,
     +  8.5151E+10, 8.5588E+10, 8.6027E+10, 8.6468E+10, 8.6911E+10,
     + 8.7356E+10, 8.7803E+10, 8.8252E+10, 8.8704E+10, 8.9157E+10,
     +  8.9612E+10, 9.0069E+10, 9.0529E+10, 9.0990E+10, 9.1454E+10,
     + 9.1920E+10, 9.2388E+10, 9.2858E+10, 9.3330E+10, 9.3804E+10,
     +  9.4280E+10, 9.4759E+10, 9.5240E+10, 9.5722E+10, 9.6207E+10,
     + 9.6695E+10, 9.7184E+10, 9.7676E+10, 9.8170E+10, 9.8666E+10,
     +  9.9164E+10, 9.9664E+10, 1.0017E+11, 1.0067E+11, 1.0118E+11,
     + 1.0169E+11, 1.0220E+11, 1.0272E+11, 1.0323E+11, 1.0375E+11,
     +  1.0427E+11, 1.0479E+11, 1.0532E+11, 1.0585E+11, 1.0638E+11,
     + 1.0691E+11, 1.0745E+11, 1.0798E+11, 1.0852E+11, 1.0907E+11,
     +  1.0961E+11, 1.1016E+11, 1.1071E+11, 1.1126E+11, 1.1181E+11,
     + 1.1237E+11, 1.1293E+11, 1.1349E+11, 1.1406E+11, 1.1462E+11,
     +  1.1519E+11, 1.1577E+11, 1.1634E+11, 1.1692E+11, 1.1750E+11,
     + 1.1808E+11, 1.1866E+11, 1.1925E+11, 1.1984E+11, 1.2043E+11,
     +  1.2103E+11, 1.2162E+11, 1.2222E+11, 1.2283E+11, 1.2343E+11,
     + 1.2404E+11, 1.2465E+11, 1.2527E+11, 1.2588E+11, 1.2650E+11,
     +  1.2712E+11, 1.2775E+11, 1.2837E+11, 1.2900E+11, 1.2963E+11,
     + 1.3027E+11, 1.3091E+11, 1.3155E+11, 1.3219E+11, 1.3284E+11,
     +  1.3349E+11, 1.3414E+11, 1.3479E+11, 1.3545E+11, 1.3611E+11,
     + 1.3677E+11, 1.3744E+11, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 4,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --     13224
      data (QofT( 5,J),J=1,2408)/ 6.0901E+01, 1.1754E+02, 1.7423E+02,
     +  2.3095E+02, 2.8767E+02, 3.4439E+02, 4.0112E+02, 4.5785E+02,
     + 5.1459E+02, 5.7132E+02, 6.2806E+02, 6.8480E+02, 7.4154E+02,
     +  7.9828E+02, 8.5502E+02, 9.1177E+02, 9.6852E+02, 1.0253E+03,
     + 1.0820E+03, 1.1388E+03, 1.2523E+03, 1.3658E+03, 1.4793E+03,
     +  1.5929E+03, 1.7065E+03, 1.8201E+03, 1.9338E+03, 2.0477E+03,
     + 2.1616E+03, 2.2758E+03, 2.3902E+03, 2.5050E+03, 2.6201E+03,
     +  2.7357E+03, 2.8518E+03, 2.9686E+03, 3.0861E+03, 3.2045E+03,
     + 3.3237E+03, 3.4439E+03, 3.5653E+03, 3.6878E+03, 3.8115E+03,
     +  3.9367E+03, 4.0632E+03, 4.1914E+03, 4.3211E+03, 4.4525E+03,
     + 4.5858E+03, 4.7209E+03, 4.8579E+03, 4.9969E+03, 5.1381E+03,
     +  5.2814E+03, 5.4270E+03, 5.5749E+03, 5.7253E+03, 5.8780E+03,
     + 6.0333E+03, 6.1913E+03, 6.3520E+03, 6.5154E+03, 6.6816E+03,
     +  6.8509E+03, 7.0230E+03, 7.1983E+03, 7.3768E+03, 7.5584E+03,
     + 7.7433E+03, 7.9317E+03, 8.1234E+03, 8.3188E+03, 8.5178E+03,
     +  8.7205E+03, 8.9270E+03, 9.1373E+03, 9.3517E+03, 9.5700E+03,
     + 9.7926E+03, 1.0019E+04, 1.0250E+04, 1.0486E+04, 1.0726E+04,
     +  1.0971E+04, 1.1220E+04, 1.1474E+04, 1.1734E+04, 1.1998E+04,
     + 1.2267E+04, 1.2542E+04, 1.2822E+04, 1.3107E+04, 1.3398E+04,
     +  1.3694E+04, 1.3997E+04, 1.4305E+04, 1.4620E+04, 1.4940E+04,
     + 1.5267E+04, 1.5600E+04, 1.5940E+04, 1.6286E+04, 1.6640E+04,
     +  1.7000E+04, 1.7367E+04, 1.7742E+04, 1.8123E+04, 1.8513E+04,
     + 1.8910E+04, 1.9315E+04, 1.9728E+04, 2.0148E+04, 2.0578E+04,
     +  2.1015E+04, 2.1461E+04, 2.1916E+04, 2.2380E+04, 2.2853E+04,
     + 2.3335E+04, 2.3827E+04, 2.4328E+04, 2.4839E+04, 2.5360E+04,
     +  2.5891E+04, 2.6432E+04, 2.6985E+04, 2.7547E+04, 2.8121E+04,
     + 2.8706E+04, 2.9302E+04, 2.9910E+04, 3.0529E+04, 3.1161E+04,
     +  3.1804E+04, 3.2460E+04, 3.3129E+04, 3.3811E+04, 3.4505E+04,
     + 3.5213E+04, 3.5935E+04, 3.6670E+04, 3.7420E+04, 3.8184E+04,
     +  3.8962E+04, 3.9755E+04, 4.0564E+04, 4.1387E+04, 4.2226E+04,
     + 4.3082E+04, 4.3953E+04, 4.4841E+04, 4.5745E+04, 4.6667E+04,
     +  4.7606E+04, 4.8563E+04, 4.9537E+04, 5.0530E+04, 5.1541E+04,
     + 5.2571E+04, 5.3621E+04, 5.4690E+04, 5.5779E+04, 5.6888E+04,
     +  5.8017E+04, 5.9168E+04, 6.0340E+04, 6.1533E+04, 6.2748E+04,
     + 6.3986E+04, 6.5247E+04, 6.6530E+04, 6.7837E+04, 6.9168E+04,
     +  7.0523E+04, 7.1903E+04, 7.3308E+04, 7.4738E+04, 7.6195E+04,
     + 7.7677E+04, 7.9187E+04, 8.0723E+04, 8.2288E+04, 8.3880E+04,
     +  8.5501E+04, 8.7151E+04, 8.8831E+04, 9.0541E+04, 9.2281E+04,
     + 9.4052E+04, 9.5854E+04, 9.7689E+04, 9.9556E+04, 1.0146E+05,
     +  1.0339E+05, 1.0536E+05, 1.0736E+05, 1.0940E+05, 1.1147E+05,
     + 1.1358E+05, 1.1572E+05, 1.1791E+05, 1.2013E+05, 1.2239E+05,
     +  1.2469E+05, 1.2703E+05, 1.2941E+05, 1.3183E+05, 1.3429E+05,
     + 1.3680E+05, 1.3935E+05, 1.4194E+05, 1.4457E+05, 1.4726E+05,
     +  1.4998E+05, 1.5276E+05, 1.5558E+05, 1.5845E+05, 1.6136E+05,
     + 1.6433E+05, 1.6735E+05, 1.7041E+05, 1.7353E+05, 1.7671E+05,
     +  1.7993E+05, 1.8321E+05, 1.8655E+05, 1.8993E+05, 1.9338E+05,
     + 1.9689E+05, 2.0045E+05, 2.0407E+05, 2.0775E+05, 2.1149E+05,
     +  2.1530E+05, 2.1916E+05, 2.2309E+05, 2.2709E+05, 2.3115E+05,
     + 2.3528E+05, 2.3947E+05, 2.4373E+05, 2.4807E+05, 2.5247E+05,
     +  2.5695E+05, 2.6149E+05, 2.6611E+05, 2.7081E+05, 2.7558E+05,
     + 2.8043E+05, 2.8536E+05, 2.9036E+05, 2.9545E+05, 3.0062E+05,
     +  3.0587E+05, 3.1120E+05, 3.1662E+05, 3.2213E+05, 3.2772E+05,
     + 3.3340E+05, 3.3918E+05, 3.4504E+05, 3.5100E+05, 3.5705E+05,
     +  3.6319E+05, 3.6943E+05, 3.7577E+05, 3.8221E+05, 3.8875E+05,
     + 3.9540E+05, 4.0214E+05, 4.0899E+05, 4.1595E+05, 4.2302E+05,
     +  4.3019E+05, 4.3748E+05, 4.4488E+05, 4.5239E+05, 4.6002E+05,
     + 4.6777E+05, 4.7563E+05, 4.8362E+05, 4.9173E+05, 4.9996E+05,
     +  5.0832E+05, 5.1681E+05, 5.2542E+05, 5.3417E+05, 5.4305E+05,
     + 5.5206E+05, 5.6121E+05, 5.7050E+05, 5.7993E+05, 5.8950E+05,
     +  5.9921E+05, 6.0908E+05, 6.1908E+05, 6.2924E+05, 6.3956E+05,
     + 6.5002E+05, 6.6064E+05, 6.7142E+05, 6.8236E+05, 6.9346E+05,
     +  7.0473E+05, 7.1616E+05, 7.2777E+05, 7.3954E+05, 7.5149E+05,
     + 7.6361E+05, 7.7591E+05, 7.8840E+05, 8.0106E+05, 8.1391E+05,
     +  8.2695E+05, 8.4018E+05, 8.5361E+05, 8.6722E+05, 8.8104E+05,
     + 8.9506E+05, 9.0928E+05, 9.2370E+05, 9.3833E+05, 9.5318E+05,
     +  9.6824E+05, 9.8352E+05, 9.9901E+05, 1.0147E+06, 1.0307E+06,
     + 1.0468E+06, 1.0632E+06, 1.0799E+06, 1.0968E+06, 1.1139E+06,
     +  1.1312E+06, 1.1488E+06, 1.1667E+06, 1.1848E+06, 1.2032E+06,
     + 1.2218E+06, 1.2407E+06, 1.2598E+06, 1.2792E+06, 1.2989E+06,
     +  1.3189E+06, 1.3392E+06, 1.3597E+06, 1.3805E+06, 1.4016E+06,
     + 1.4230E+06, 1.4447E+06, 1.4667E+06, 1.4890E+06, 1.5117E+06,
     +  1.5346E+06, 1.5578E+06, 1.5814E+06, 1.6053E+06, 1.6295E+06,
     + 1.6541E+06, 1.6790E+06, 1.7042E+06, 1.7298E+06, 1.7557E+06,
     +  1.7820E+06, 1.8086E+06, 1.8357E+06, 1.8630E+06, 1.8908E+06,
     + 1.9189E+06, 1.9474E+06, 1.9763E+06, 2.0056E+06, 2.0353E+06,
     +  2.0653E+06, 2.0958E+06, 2.1267E+06, 2.1580E+06, 2.1897E+06,
     + 2.2219E+06, 2.2545E+06, 2.2875E+06, 2.3210E+06, 2.3549E+06,
     +  2.3892E+06, 2.4240E+06, 2.4593E+06, 2.4951E+06, 2.5313E+06,
     + 2.5680E+06, 2.6052E+06, 2.6428E+06, 2.6810E+06, 2.7197E+06,
     +  2.7589E+06, 2.7986E+06, 2.8388E+06, 2.8795E+06, 2.9208E+06,
     + 2.9626E+06, 3.0050E+06, 3.0479E+06, 3.0914E+06, 3.1354E+06,
     +  3.1801E+06, 3.2253E+06, 3.2710E+06, 3.3174E+06, 3.3644E+06,
     + 3.4120E+06, 3.4602E+06, 3.5090E+06, 3.5584E+06, 3.6085E+06,
     +  3.6592E+06, 3.7106E+06, 3.7626E+06, 3.8153E+06, 3.8687E+06,
     + 3.9228E+06, 3.9775E+06, 4.0329E+06, 4.0891E+06, 4.1459E+06,
     +  4.2035E+06, 4.2618E+06, 4.3209E+06, 4.3807E+06, 4.4412E+06,
     + 4.5025E+06, 4.5646E+06, 4.6275E+06, 4.6912E+06, 4.7556E+06,
     +  4.8209E+06, 4.8870E+06, 4.9539E+06, 5.0216E+06, 5.0902E+06,
     + 5.1597E+06, 5.2300E+06, 5.3012E+06, 5.3733E+06, 5.4462E+06,
     +  5.5201E+06, 5.5949E+06, 5.6706E+06, 5.7473E+06, 5.8249E+06,
     + 5.9034E+06, 5.9830E+06, 6.0635E+06, 6.1450E+06, 6.2275E+06,
     +  6.3110E+06, 6.3955E+06, 6.4811E+06, 6.5677E+06, 6.6554E+06,
     + 6.7441E+06, 6.8339E+06, 6.9248E+06, 7.0168E+06, 7.1100E+06,
     +  7.2042E+06, 7.2996E+06, 7.3962E+06, 7.4939E+06, 7.5928E+06,
     + 7.6930E+06, 7.7943E+06, 7.8968E+06, 8.0006E+06, 8.1056E+06,
     +  8.2118E+06, 8.3194E+06, 8.4282E+06, 8.5383E+06, 8.6498E+06,
     + 8.7626E+06, 8.8767E+06, 8.9922E+06, 9.1090E+06, 9.2273E+06,
     +  9.3469E+06, 9.4680E+06, 9.5905E+06, 9.7144E+06, 9.8399E+06,
     + 9.9668E+06, 1.0095E+07, 1.0225E+07, 1.0356E+07, 1.0489E+07,
     +  1.0624E+07, 1.0760E+07, 1.0898E+07, 1.1037E+07, 1.1178E+07,
     + 1.1321E+07, 1.1465E+07, 1.1611E+07, 1.1759E+07, 1.1908E+07,
     +  1.2059E+07, 1.2212E+07, 1.2367E+07, 1.2523E+07, 1.2681E+07,
     + 1.2841E+07, 1.3003E+07, 1.3167E+07, 1.3333E+07, 1.3500E+07,
     +  1.3670E+07, 1.3841E+07, 1.4015E+07, 1.4190E+07, 1.4368E+07,
     + 1.4547E+07, 1.4729E+07, 1.4912E+07, 1.5098E+07, 1.5286E+07,
     +  1.5476E+07, 1.5668E+07, 1.5862E+07, 1.6059E+07, 1.6257E+07,
     + 1.6458E+07, 1.6662E+07, 1.6867E+07, 1.7075E+07, 1.7285E+07,
     +  1.7498E+07, 1.7713E+07, 1.7930E+07, 1.8150E+07, 1.8372E+07,
     + 1.8597E+07, 1.8825E+07, 1.9055E+07, 1.9287E+07, 1.9522E+07,
     +  1.9760E+07, 2.0000E+07, 2.0243E+07, 2.0489E+07, 2.0737E+07,
     + 2.0988E+07, 2.1242E+07, 2.1499E+07, 2.1759E+07, 2.2022E+07,
     +  2.2287E+07, 2.2555E+07, 2.2827E+07, 2.3101E+07, 2.3378E+07,
     + 2.3659E+07, 2.3942E+07, 2.4229E+07, 2.4519E+07, 2.4812E+07,
     +  2.5108E+07, 2.5407E+07, 2.5710E+07, 2.6016E+07, 2.6325E+07,
     + 2.6638E+07, 2.6954E+07, 2.7273E+07, 2.7596E+07, 2.7923E+07,
     +  2.8253E+07, 2.8586E+07, 2.8924E+07, 2.9265E+07, 2.9609E+07,
     + 2.9957E+07, 3.0309E+07, 3.0665E+07, 3.1025E+07, 3.1389E+07,
     +  3.1756E+07, 3.2127E+07, 3.2503E+07, 3.2882E+07, 3.3266E+07,
     + 3.3653E+07, 3.4045E+07, 3.4441E+07, 3.4841E+07, 3.5245E+07,
     +  3.5654E+07, 3.6067E+07, 3.6485E+07, 3.6906E+07, 3.7333E+07,
     + 3.7764E+07, 3.8199E+07, 3.8639E+07, 3.9084E+07, 3.9533E+07,
     +  3.9987E+07, 4.0446E+07, 4.0910E+07, 4.1379E+07, 4.1852E+07,
     + 4.2331E+07, 4.2814E+07, 4.3303E+07, 4.3797E+07, 4.4296E+07,
     +  4.4800E+07, 4.5309E+07, 4.5824E+07, 4.6344E+07, 4.6870E+07,
     + 4.7401E+07, 4.7937E+07, 4.8479E+07, 4.9027E+07, 4.9581E+07,
     +  5.0140E+07, 5.0705E+07, 5.1275E+07, 5.1852E+07, 5.2435E+07,
     + 5.3023E+07, 5.3618E+07, 5.4219E+07, 5.4826E+07, 5.5439E+07,
     +  5.6059E+07, 5.6685E+07, 5.7317E+07, 5.7956E+07, 5.8602E+07,
     + 5.9254E+07, 5.9912E+07, 6.0578E+07, 6.1250E+07, 6.1929E+07,
     +  6.2615E+07, 6.3308E+07, 6.4008E+07, 6.4715E+07, 6.5429E+07,
     + 6.6151E+07, 6.6880E+07, 6.7616E+07, 6.8360E+07, 6.9111E+07,
     +  6.9870E+07, 7.0636E+07, 7.1410E+07, 7.2192E+07, 7.2982E+07,
     + 7.3780E+07, 7.4586E+07, 7.5399E+07, 7.6222E+07, 7.7052E+07,
     +  7.7890E+07, 7.8737E+07, 7.9593E+07, 8.0457E+07, 8.1330E+07,
     + 8.2211E+07, 8.3101E+07, 8.4000E+07, 8.4908E+07, 8.5825E+07,
     +  8.6751E+07, 8.7686E+07, 8.8631E+07, 8.9585E+07, 9.0548E+07,
     + 9.1521E+07, 9.2503E+07, 9.3495E+07, 9.4497E+07, 9.5509E+07,
     +  9.6531E+07, 9.7563E+07, 9.8605E+07, 9.9657E+07, 1.0072E+08,
     + 1.0179E+08, 1.0288E+08, 1.0397E+08, 1.0507E+08, 1.0619E+08,
     +  1.0732E+08, 1.0845E+08, 1.0960E+08, 1.1076E+08, 1.1193E+08,
     + 1.1312E+08, 1.1431E+08, 1.1552E+08, 1.1673E+08, 1.1796E+08,
     +  1.1920E+08, 1.2046E+08, 1.2172E+08, 1.2300E+08, 1.2429E+08,
     + 1.2559E+08, 1.2690E+08, 1.2823E+08, 1.2957E+08, 1.3093E+08,
     +  1.3229E+08, 1.3367E+08, 1.3506E+08, 1.3647E+08, 1.3789E+08,
     + 1.3932E+08, 1.4077E+08, 1.4223E+08, 1.4370E+08, 1.4519E+08,
     +  1.4669E+08, 1.4821E+08, 1.4974E+08, 1.5129E+08, 1.5285E+08,
     + 1.5442E+08, 1.5601E+08, 1.5762E+08, 1.5924E+08, 1.6087E+08,
     +  1.6252E+08, 1.6419E+08, 1.6587E+08, 1.6757E+08, 1.6929E+08,
     + 1.7102E+08, 1.7277E+08, 1.7453E+08, 1.7631E+08, 1.7811E+08,
     +  1.7992E+08, 1.8175E+08, 1.8360E+08, 1.8546E+08, 1.8735E+08,
     + 1.8925E+08, 1.9116E+08, 1.9310E+08, 1.9505E+08, 1.9703E+08,
     +  1.9902E+08, 2.0103E+08, 2.0305E+08, 2.0510E+08, 2.0716E+08,
     + 2.0925E+08, 2.1135E+08, 2.1348E+08, 2.1562E+08, 2.1778E+08,
     +  2.1997E+08, 2.2217E+08, 2.2439E+08, 2.2664E+08, 2.2890E+08,
     + 2.3119E+08, 2.3349E+08, 2.3582E+08, 2.3817E+08, 2.4054E+08,
     +  2.4293E+08, 2.4535E+08, 2.4778E+08, 2.5024E+08, 2.5272E+08,
     + 2.5523E+08, 2.5775E+08, 2.6030E+08, 2.6288E+08, 2.6547E+08,
     +  2.6809E+08, 2.7074E+08, 2.7341E+08, 2.7610E+08, 2.7881E+08,
     + 2.8156E+08, 2.8432E+08, 2.8711E+08, 2.8993E+08, 2.9277E+08,
     +  2.9564E+08, 2.9853E+08, 3.0145E+08, 3.0440E+08, 3.0737E+08,
     + 3.1037E+08, 3.1340E+08, 3.1645E+08, 3.1953E+08, 3.2264E+08,
     +  3.2578E+08, 3.2894E+08, 3.3213E+08, 3.3535E+08, 3.3860E+08,
     + 3.4188E+08, 3.4519E+08, 3.4853E+08, 3.5190E+08, 3.5530E+08,
     +  3.5872E+08, 3.6218E+08, 3.6567E+08, 3.6919E+08, 3.7274E+08,
     + 3.7633E+08, 3.7994E+08, 3.8359E+08, 3.8727E+08, 3.9098E+08,
     +  3.9472E+08, 3.9850E+08, 4.0231E+08, 4.0616E+08, 4.1004E+08,
     + 4.1395E+08, 4.1790E+08, 4.2188E+08, 4.2589E+08, 4.2995E+08,
     +  4.3403E+08, 4.3816E+08, 4.4232E+08, 4.4651E+08, 4.5075E+08,
     + 4.5502E+08, 4.5932E+08, 4.6367E+08, 4.6805E+08, 4.7247E+08,
     +  4.7693E+08, 4.8143E+08, 4.8597E+08, 4.9054E+08, 4.9516E+08,
     + 4.9982E+08, 5.0451E+08, 5.0925E+08, 5.1403E+08, 5.1885E+08,
     +  5.2371E+08, 5.2862E+08, 5.3356E+08, 5.3855E+08, 5.4358E+08,
     + 5.4866E+08, 5.5378E+08, 5.5894E+08, 5.6415E+08, 5.6940E+08,
     +  5.7470E+08, 5.8004E+08, 5.8543E+08, 5.9087E+08, 5.9635E+08,
     + 6.0188E+08, 6.0746E+08, 6.1308E+08, 6.1875E+08, 6.2447E+08,
     +  6.3024E+08, 6.3606E+08, 6.4193E+08, 6.4784E+08, 6.5381E+08,
     + 6.5983E+08, 6.6590E+08, 6.7202E+08, 6.7820E+08, 6.8442E+08,
     +  6.9070E+08, 6.9703E+08, 7.0342E+08, 7.0986E+08, 7.1635E+08,
     + 7.2290E+08, 7.2950E+08, 7.3616E+08, 7.4288E+08, 7.4965E+08,
     +  7.5648E+08, 7.6337E+08, 7.7031E+08, 7.7732E+08, 7.8438E+08,
     + 7.9150E+08, 7.9868E+08, 8.0592E+08, 8.1322E+08, 8.2059E+08,
     +  8.2801E+08, 8.3550E+08, 8.4305E+08, 8.5066E+08, 8.5833E+08,
     + 8.6607E+08, 8.7388E+08, 8.8174E+08, 8.8968E+08, 8.9768E+08,
     +  9.0574E+08, 9.1388E+08, 9.2208E+08, 9.3035E+08, 9.3868E+08,
     + 9.4709E+08, 9.5557E+08, 9.6411E+08, 9.7273E+08, 9.8142E+08,
     +  9.9018E+08, 9.9901E+08, 1.0079E+09, 1.0169E+09, 1.0259E+09,
     + 1.0351E+09, 1.0443E+09, 1.0535E+09, 1.0629E+09, 1.0723E+09,
     +  1.0818E+09, 1.0914E+09, 1.1011E+09, 1.1108E+09, 1.1206E+09,
     + 1.1305E+09, 1.1405E+09, 1.1506E+09, 1.1607E+09, 1.1710E+09,
     +  1.1813E+09, 1.1917E+09, 1.2022E+09, 1.2127E+09, 1.2234E+09,
     + 1.2341E+09, 1.2449E+09, 1.2558E+09, 1.2668E+09, 1.2779E+09,
     +  1.2891E+09, 1.3004E+09, 1.3117E+09, 1.3232E+09, 1.3347E+09,
     + 1.3464E+09, 1.3581E+09, 1.3699E+09, 1.3819E+09, 1.3939E+09,
     +  1.4060E+09, 1.4182E+09, 1.4305E+09, 1.4429E+09, 1.4554E+09,
     + 1.4680E+09, 1.4807E+09, 1.4936E+09, 1.5065E+09, 1.5195E+09,
     +  1.5326E+09, 1.5458E+09, 1.5592E+09, 1.5726E+09, 1.5861E+09,
     + 1.5998E+09, 1.6136E+09, 1.6274E+09, 1.6414E+09, 1.6555E+09,
     +  1.6697E+09, 1.6840E+09, 1.6984E+09, 1.7130E+09, 1.7276E+09,
     + 1.7424E+09, 1.7573E+09, 1.7723E+09, 1.7874E+09, 1.8027E+09,
     +  1.8180E+09, 1.8335E+09, 1.8491E+09, 1.8649E+09, 1.8807E+09,
     + 1.8967E+09, 1.9128E+09, 1.9291E+09, 1.9454E+09, 1.9619E+09,
     +  1.9785E+09, 1.9953E+09, 2.0121E+09, 2.0292E+09, 2.0463E+09,
     + 2.0636E+09, 2.0810E+09, 2.0985E+09, 2.1162E+09, 2.1340E+09,
     +  2.1520E+09, 2.1701E+09, 2.1883E+09, 2.2067E+09, 2.2253E+09,
     + 2.2439E+09, 2.2627E+09, 2.2817E+09, 2.3008E+09, 2.3201E+09,
     +  2.3395E+09, 2.3590E+09, 2.3787E+09, 2.3986E+09, 2.4186E+09,
     + 2.4388E+09, 2.4591E+09, 2.4796E+09, 2.5002E+09, 2.5210E+09,
     +  2.5420E+09, 2.5631E+09, 2.5844E+09, 2.6058E+09, 2.6274E+09,
     + 2.6492E+09, 2.6711E+09, 2.6932E+09, 2.7155E+09, 2.7380E+09,
     +  2.7606E+09, 2.7834E+09, 2.8063E+09, 2.8295E+09, 2.8528E+09,
     + 2.8763E+09, 2.9000E+09, 2.9238E+09, 2.9478E+09, 2.9721E+09,
     +  2.9965E+09, 3.0211E+09, 3.0458E+09, 3.0708E+09, 3.0959E+09,
     + 3.1213E+09, 3.1468E+09, 3.1725E+09, 3.1985E+09, 3.2246E+09,
     +  3.2509E+09, 3.2774E+09, 3.3041E+09, 3.3310E+09, 3.3582E+09,
     + 3.3855E+09, 3.4130E+09, 3.4407E+09, 3.4687E+09, 3.4968E+09,
     +  3.5252E+09, 3.5538E+09, 3.5826E+09, 3.6116E+09, 3.6408E+09,
     + 3.6703E+09, 3.6999E+09, 3.7298E+09, 3.7599E+09, 3.7903E+09,
     +  3.8208E+09, 3.8516E+09, 3.8826E+09, 3.9139E+09, 3.9454E+09,
     + 3.9771E+09, 4.0090E+09, 4.0412E+09, 4.0737E+09, 4.1063E+09,
     +  4.1392E+09, 4.1724E+09, 4.2058E+09, 4.2395E+09, 4.2734E+09,
     + 4.3075E+09, 4.3419E+09, 4.3766E+09, 4.4115E+09, 4.4467E+09,
     +  4.4821E+09, 4.5178E+09, 4.5537E+09, 4.5899E+09, 4.6264E+09,
     + 4.6632E+09, 4.7002E+09, 4.7375E+09, 4.7751E+09, 4.8129E+09,
     +  4.8511E+09, 4.8895E+09, 4.9282E+09, 4.9671E+09, 5.0064E+09,
     + 5.0459E+09, 5.0857E+09, 5.1259E+09, 5.1663E+09, 5.2070E+09,
     +  5.2480E+09, 5.2893E+09, 5.3309E+09, 5.3728E+09, 5.4150E+09,
     + 5.4576E+09, 5.5004E+09, 5.5435E+09, 5.5870E+09, 5.6308E+09,
     +  5.6748E+09, 5.7193E+09, 5.7640E+09, 5.8090E+09, 5.8544E+09,
     + 5.9001E+09, 5.9462E+09, 5.9925E+09, 6.0392E+09, 6.0863E+09,
     +  6.1337E+09, 6.1814E+09, 6.2294E+09, 6.2779E+09, 6.3266E+09,
     + 6.3757E+09, 6.4252E+09, 6.4750E+09, 6.5252E+09, 6.5757E+09,
     +  6.6266E+09, 6.6779E+09, 6.7295E+09, 6.7815E+09, 6.8339E+09,
     + 6.8866E+09, 6.9398E+09, 6.9933E+09, 7.0472E+09, 7.1014E+09,
     +  7.1561E+09, 7.2111E+09, 7.2666E+09, 7.3224E+09, 7.3786E+09,
     + 7.4353E+09, 7.4923E+09, 7.5497E+09, 7.6076E+09, 7.6658E+09,
     +  7.7245E+09, 7.7836E+09, 7.8431E+09, 7.9030E+09, 7.9634E+09,
     + 8.0242E+09, 8.0854E+09, 8.1470E+09, 8.2091E+09, 8.2716E+09,
     +  8.3345E+09, 8.3979E+09, 8.4618E+09, 8.5261E+09, 8.5908E+09,
     + 8.6560E+09, 8.7217E+09, 8.7878E+09, 8.8544E+09, 8.9214E+09,
     +  8.9889E+09, 9.0569E+09, 9.1254E+09, 9.1944E+09, 9.2638E+09,
     + 9.3337E+09, 9.4041E+09, 9.4750E+09, 9.5464E+09, 9.6183E+09,
     +  9.6907E+09, 9.7636E+09, 9.8370E+09, 9.9109E+09, 9.9853E+09,
     + 1.0060E+10, 1.0136E+10, 1.0212E+10, 1.0288E+10, 1.0365E+10,
     +  1.0443E+10, 1.0521E+10, 1.0600E+10, 1.0679E+10, 1.0759E+10,
     + 1.0839E+10, 1.0920E+10, 1.1001E+10, 1.1083E+10, 1.1166E+10,
     +  1.1249E+10, 1.1333E+10, 1.1417E+10, 1.1502E+10, 1.1587E+10,
     + 1.1673E+10, 1.1760E+10, 1.1847E+10, 1.1935E+10, 1.2023E+10,
     +  1.2112E+10, 1.2202E+10, 1.2292E+10, 1.2383E+10, 1.2475E+10,
     + 1.2567E+10, 1.2659E+10, 1.2753E+10, 1.2847E+10, 1.2941E+10,
     +  1.3037E+10, 1.3133E+10, 1.3229E+10, 1.3327E+10, 1.3425E+10,
     + 1.3523E+10, 1.3622E+10, 1.3722E+10, 1.3823E+10, 1.3924E+10,
     +  1.4026E+10, 1.4129E+10, 1.4232E+10, 1.4336E+10, 1.4441E+10,
     + 1.4547E+10, 1.4653E+10, 1.4760E+10, 1.4868E+10, 1.4976E+10,
     +  1.5085E+10, 1.5195E+10, 1.5305E+10, 1.5417E+10, 1.5529E+10,
     + 1.5642E+10, 1.5755E+10, 1.5870E+10, 1.5985E+10, 1.6101E+10,
     +  1.6217E+10, 1.6335E+10, 1.6453E+10, 1.6572E+10, 1.6692E+10,
     + 1.6813E+10, 1.6934E+10, 1.7057E+10, 1.7180E+10, 1.7304E+10,
     +  1.7428E+10, 1.7554E+10, 1.7680E+10, 1.7808E+10, 1.7936E+10,
     + 1.8065E+10, 1.8194E+10, 1.8325E+10, 1.8457E+10, 1.8589E+10,
     +  1.8722E+10, 1.8857E+10, 1.8992E+10, 1.9128E+10, 1.9265E+10,
     + 1.9402E+10, 1.9541E+10, 1.9681E+10, 1.9821E+10, 1.9963E+10,
     +  2.0105E+10, 2.0249E+10, 2.0393E+10, 2.0538E+10, 2.0684E+10,
     + 2.0832E+10, 2.0980E+10, 2.1129E+10, 2.1279E+10, 2.1430E+10,
     +  2.1582E+10, 2.1735E+10, 2.1890E+10, 2.2045E+10, 2.2201E+10,
     + 2.2358E+10, 2.2516E+10, 2.2675E+10, 2.2836E+10, 2.2997E+10,
     +  2.3159E+10, 2.3323E+10, 2.3487E+10, 2.3653E+10, 2.3820E+10,
     + 2.3987E+10, 2.4156E+10, 2.4326E+10, 2.4497E+10, 2.4669E+10,
     +  2.4843E+10, 2.5017E+10, 2.5193E+10, 2.5370E+10, 2.5547E+10,
     + 2.5726E+10, 2.5907E+10, 2.6088E+10, 2.6270E+10, 2.6454E+10,
     +  2.6639E+10, 2.6825E+10, 2.7012E+10, 2.7201E+10, 2.7391E+10,
     + 2.7582E+10, 2.7774E+10, 2.7967E+10, 2.8162E+10, 2.8358E+10,
     +  2.8555E+10, 2.8754E+10, 2.8953E+10, 2.9154E+10, 2.9357E+10,
     + 2.9560E+10, 2.9765E+10, 2.9972E+10, 3.0179E+10, 3.0388E+10,
     +  3.0598E+10, 3.0810E+10, 3.1023E+10, 3.1237E+10, 3.1453E+10,
     + 3.1670E+10, 3.1889E+10, 3.2109E+10, 3.2330E+10, 3.2553E+10,
     +  3.2777E+10, 3.3002E+10, 3.3229E+10, 3.3458E+10, 3.3688E+10,
     + 3.3919E+10, 3.4152E+10, 3.4386E+10, 3.4622E+10, 3.4859E+10,
     +  3.5098E+10, 3.5338E+10, 3.5580E+10, 3.5823E+10, 3.6068E+10,
     + 3.6315E+10, 3.6563E+10, 3.6813E+10, 3.7064E+10, 3.7317E+10,
     +  3.7571E+10, 3.7827E+10, 3.8085E+10, 3.8344E+10, 3.8605E+10,
     + 3.8867E+10, 3.9131E+10, 3.9397E+10, 3.9665E+10, 3.9934E+10,
     +  4.0205E+10, 4.0477E+10, 4.0752E+10, 4.1028E+10, 4.1305E+10,
     + 4.1585E+10, 4.1866E+10, 4.2149E+10, 4.2434E+10, 4.2721E+10,
     +  4.3009E+10, 4.3299E+10, 4.3591E+10, 4.3885E+10, 4.4181E+10,
     + 4.4478E+10, 4.4777E+10, 4.5079E+10, 4.5382E+10, 4.5687E+10,
     +  4.5994E+10, 4.6303E+10, 4.6613E+10, 4.6926E+10, 4.7241E+10,
     + 4.7557E+10, 4.7876E+10, 4.8196E+10, 4.8519E+10, 4.8843E+10,
     +  4.9170E+10, 4.9498E+10, 4.9829E+10, 5.0161E+10, 5.0496E+10,
     + 5.0833E+10, 5.1172E+10, 5.1513E+10, 5.1856E+10, 5.2201E+10,
     +  5.2548E+10, 5.2898E+10, 5.3249E+10, 5.3603E+10, 5.3959E+10,
     + 5.4317E+10, 5.4677E+10, 5.5040E+10, 5.5405E+10, 5.5772E+10,
     +  5.6141E+10, 5.6512E+10, 5.6886E+10, 5.7262E+10, 5.7641E+10,
     + 5.8021E+10, 5.8404E+10, 5.8790E+10, 5.9178E+10, 5.9568E+10,
     +  5.9960E+10, 6.0355E+10, 6.0753E+10, 6.1152E+10, 6.1555E+10,
     + 6.1959E+10, 6.2366E+10, 6.2776E+10, 6.3188E+10, 6.3603E+10,
     +  6.4020E+10, 6.4439E+10, 6.4862E+10, 6.5286E+10, 6.5714E+10,
     + 6.6144E+10, 6.6576E+10, 6.7011E+10, 6.7449E+10, 6.7890E+10,
     +  6.8333E+10, 6.8779E+10, 6.9227E+10, 6.9679E+10, 7.0133E+10,
     + 7.0589E+10, 7.1049E+10, 7.1511E+10, 7.1976E+10, 7.2444E+10,
     +  7.2914E+10, 7.3388E+10, 7.3864E+10, 7.4344E+10, 7.4826E+10,
     + 7.5311E+10, 7.5799E+10, 7.6289E+10, 7.6783E+10, 7.7280E+10,
     +  7.7780E+10, 7.8282E+10, 7.8788E+10, 7.9297E+10, 7.9809E+10,
     + 8.0324E+10, 8.0842E+10, 8.1363E+10, 8.1887E+10, 8.2414E+10,
     +  8.2945E+10, 8.3478E+10, 8.4015E+10, 8.4555E+10, 8.5098E+10,
     + 8.5645E+10, 8.6195E+10, 8.6748E+10, 8.7304E+10, 8.7864E+10,
     +  8.8426E+10, 8.8993E+10, 8.9562E+10, 9.0135E+10, 9.0712E+10,
     + 9.1292E+10, 9.1875E+10, 9.2462E+10, 9.3052E+10, 9.3646E+10,
     +  9.4243E+10, 9.4844E+10, 9.5448E+10, 9.6056E+10, 9.6667E+10,
     + 9.7282E+10, 9.7901E+10, 9.8523E+10, 9.9149E+10, 9.9779E+10,
     +  1.0041E+11, 1.0105E+11, 1.0169E+11, 1.0234E+11, 1.0298E+11,
     + 1.0364E+11, 1.0429E+11, 1.0495E+11, 1.0562E+11, 1.0628E+11,
     +  1.0696E+11, 1.0763E+11, 1.0831E+11, 1.0899E+11, 1.0968E+11,
     + 1.1037E+11, 1.1107E+11, 1.1177E+11, 1.1247E+11, 1.1318E+11,
     +  1.1389E+11, 1.1461E+11, 1.1533E+11, 1.1605E+11, 1.1678E+11,
     + 1.1751E+11, 1.1825E+11, 1.1899E+11, 1.1974E+11, 1.2049E+11,
     +  1.2124E+11, 1.2200E+11, 1.2277E+11, 1.2353E+11, 1.2431E+11,
     + 1.2508E+11, 1.2586E+11, 1.2665E+11, 1.2744E+11, 1.2824E+11,
     +  1.2903E+11, 1.2984E+11, 1.3065E+11, 1.3146E+11, 1.3228E+11,
     + 1.3310E+11, 1.3393E+11, 1.3476E+11, 1.3560E+11, 1.3644E+11,
     +  1.3729E+11, 1.3814E+11, 1.3900E+11, 1.3986E+11, 1.4072E+11,
     + 1.4160E+11, 1.4247E+11, 1.4335E+11, 1.4424E+11, 1.4513E+11,
     +  1.4603E+11, 1.4693E+11, 1.4784E+11, 1.4875E+11, 1.4967E+11,
     + 1.5059E+11, 1.5152E+11, 1.5245E+11, 1.5339E+11, 1.5433E+11,
     +  1.5528E+11, 1.5624E+11, 1.5720E+11, 1.5816E+11, 1.5914E+11,
     + 1.6011E+11, 1.6110E+11, 1.6208E+11, 1.6308E+11, 1.6408E+11,
     +  1.6508E+11, 1.6609E+11, 1.6711E+11, 1.6813E+11, 1.6916E+11,
     + 1.7019E+11, 1.7123E+11, 1.7228E+11, 1.7333E+11, 1.7439E+11,
     +  1.7545E+11, 1.7652E+11, 1.7759E+11, 1.7868E+11, 1.7976E+11,
     + 1.8086E+11, 1.8196E+11, 1.8306E+11, 1.8417E+11, 1.8529E+11,
     +  1.8642E+11, 1.8755E+11, 1.8869E+11, 1.8983E+11, 1.9098E+11,
     + 1.9214E+11, 1.9330E+11, 1.9447E+11, 1.9565E+11, 1.9683E+11,
     +  1.9802E+11, 1.9922E+11, 2.0042E+11, 2.0163E+11, 2.0285E+11,
     + 2.0407E+11, 2.0530E+11, 2.0654E+11, 2.0778E+11, 2.0903E+11,
     +  2.1029E+11, 2.1156E+11, 2.1283E+11, 2.1411E+11, 2.1540E+11,
     + 2.1669E+11, 2.1799E+11, 2.1930E+11, 2.2061E+11, 2.2194E+11,
     +  2.2327E+11, 2.2460E+11, 2.2595E+11, 2.2730E+11, 2.2866E+11,
     + 2.3003E+11, 2.3140E+11, 2.3279E+11, 2.3418E+11, 2.3557E+11,
     +  2.3698E+11, 2.3839E+11, 2.3981E+11, 2.4124E+11, 2.4268E+11,
     + 2.4413E+11, 2.4558E+11, 2.4704E+11, 2.4851E+11, 2.4999E+11,
     +  2.5147E+11, 2.5297E+11, 2.5447E+11, 2.5598E+11, 2.5750E+11,
     + 2.5902E+11, 2.6056E+11, 2.6210E+11, 2.6365E+11, 2.6521E+11,
     +  2.6678E+11, 2.6836E+11, 2.6995E+11, 2.7154E+11, 2.7315E+11,
     + 2.7476E+11, 2.7638E+11, 2.7801E+11, 2.7965E+11, 2.8130E+11,
     +  2.8295E+11, 2.8462E+11, 2.8630E+11, 2.8798E+11, 2.8967E+11,
     + 2.9138E+11, 2.9309E+11, 2.9481E+11, 2.9654E+11, 2.9828E+11,
     +  3.0003E+11, 3.0179E+11, 3.0356E+11, 3.0534E+11, 3.0713E+11,
     + 3.0893E+11, 3.1073E+11, 3.1255E+11, 3.1438E+11, 3.1622E+11,
     +  3.1806E+11, 3.1992E+11, 3.2179E+11, 3.2367E+11, 3.2555E+11,
     + 3.2745E+11, 3.2936E+11, 3.3128E+11, 3.3321E+11, 3.3515E+11,
     +  3.3710E+11, 3.3906E+11, 3.4103E+11, 3.4301E+11, 3.4500E+11,
     + 3.4701E+11, 3.4902E+11, 3.5104E+11, 3.5308E+11, 3.5513E+11,
     +  3.5718E+11, 3.5925E+11, 3.6133E+11, 3.6342E+11, 3.6552E+11,
     + 3.6764E+11, 3.6976E+11, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 5,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --     22224
      data (QofT( 6,J),J=1,2408)/ 4.7553E+01, 9.1916E+01, 1.3632E+02,
     +  1.8074E+02, 2.2517E+02, 2.6959E+02, 3.1402E+02, 3.5845E+02,
     + 4.0289E+02, 4.4732E+02, 4.9175E+02, 5.3619E+02, 5.8063E+02,
     +  6.2507E+02, 6.6951E+02, 7.1395E+02, 7.5839E+02, 8.0283E+02,
     + 8.4727E+02, 8.9172E+02, 9.8061E+02, 1.0695E+03, 1.1584E+03,
     +  1.2474E+03, 1.3363E+03, 1.4253E+03, 1.5144E+03, 1.6036E+03,
     + 1.6930E+03, 1.7825E+03, 1.8723E+03, 1.9624E+03, 2.0529E+03,
     +  2.1438E+03, 2.2352E+03, 2.3273E+03, 2.4200E+03, 2.5135E+03,
     + 2.6078E+03, 2.7029E+03, 2.7991E+03, 2.8964E+03, 2.9948E+03,
     +  3.0943E+03, 3.1952E+03, 3.2975E+03, 3.4011E+03, 3.5063E+03,
     + 3.6130E+03, 3.7214E+03, 3.8315E+03, 3.9433E+03, 4.0570E+03,
     +  4.1726E+03, 4.2902E+03, 4.4098E+03, 4.5315E+03, 4.6554E+03,
     + 4.7815E+03, 4.9100E+03, 5.0408E+03, 5.1741E+03, 5.3099E+03,
     +  5.4483E+03, 5.5894E+03, 5.7333E+03, 5.8799E+03, 6.0294E+03,
     + 6.1819E+03, 6.3375E+03, 6.4961E+03, 6.6580E+03, 6.8232E+03,
     +  6.9917E+03, 7.1637E+03, 7.3391E+03, 7.5183E+03, 7.7011E+03,
     + 7.8878E+03, 8.0783E+03, 8.2728E+03, 8.4715E+03, 8.6742E+03,
     +  8.8813E+03, 9.0927E+03, 9.3086E+03, 9.5292E+03, 9.7544E+03,
     + 9.9843E+03, 1.0219E+04, 1.0459E+04, 1.0704E+04, 1.0954E+04,
     +  1.1210E+04, 1.1471E+04, 1.1738E+04, 1.2010E+04, 1.2288E+04,
     + 1.2572E+04, 1.2863E+04, 1.3159E+04, 1.3462E+04, 1.3771E+04,
     +  1.4087E+04, 1.4410E+04, 1.4740E+04, 1.5076E+04, 1.5420E+04,
     + 1.5771E+04, 1.6130E+04, 1.6497E+04, 1.6871E+04, 1.7253E+04,
     +  1.7643E+04, 1.8042E+04, 1.8449E+04, 1.8865E+04, 1.9289E+04,
     + 1.9723E+04, 2.0166E+04, 2.0618E+04, 2.1080E+04, 2.1551E+04,
     +  2.2032E+04, 2.2524E+04, 2.3026E+04, 2.3538E+04, 2.4061E+04,
     + 2.4595E+04, 2.5140E+04, 2.5697E+04, 2.6265E+04, 2.6845E+04,
     +  2.7437E+04, 2.8041E+04, 2.8658E+04, 2.9288E+04, 2.9930E+04,
     + 3.0586E+04, 3.1255E+04, 3.1938E+04, 3.2635E+04, 3.3346E+04,
     +  3.4072E+04, 3.4812E+04, 3.5568E+04, 3.6339E+04, 3.7126E+04,
     + 3.7928E+04, 3.8747E+04, 3.9582E+04, 4.0434E+04, 4.1304E+04,
     +  4.2190E+04, 4.3095E+04, 4.4017E+04, 4.4958E+04, 4.5918E+04,
     + 4.6896E+04, 4.7895E+04, 4.8912E+04, 4.9950E+04, 5.1009E+04,
     +  5.2088E+04, 5.3189E+04, 5.4311E+04, 5.5455E+04, 5.6622E+04,
     + 5.7811E+04, 5.9023E+04, 6.0260E+04, 6.1519E+04, 6.2804E+04,
     +  6.4113E+04, 6.5447E+04, 6.6807E+04, 6.8194E+04, 6.9607E+04,
     + 7.1046E+04, 7.2514E+04, 7.4009E+04, 7.5533E+04, 7.7086E+04,
     +  7.8668E+04, 8.0280E+04, 8.1922E+04, 8.3596E+04, 8.5301E+04,
     + 8.7038E+04, 8.8807E+04, 9.0610E+04, 9.2446E+04, 9.4317E+04,
     +  9.6222E+04, 9.8162E+04, 1.0014E+05, 1.0215E+05, 1.0420E+05,
     + 1.0629E+05, 1.0842E+05, 1.1058E+05, 1.1279E+05, 1.1503E+05,
     +  1.1732E+05, 1.1965E+05, 1.2202E+05, 1.2443E+05, 1.2689E+05,
     + 1.2939E+05, 1.3193E+05, 1.3453E+05, 1.3716E+05, 1.3985E+05,
     +  1.4258E+05, 1.4537E+05, 1.4820E+05, 1.5108E+05, 1.5402E+05,
     + 1.5700E+05, 1.6004E+05, 1.6314E+05, 1.6628E+05, 1.6949E+05,
     +  1.7274E+05, 1.7606E+05, 1.7944E+05, 1.8287E+05, 1.8636E+05,
     + 1.8991E+05, 1.9353E+05, 1.9721E+05, 2.0095E+05, 2.0476E+05,
     +  2.0863E+05, 2.1257E+05, 2.1657E+05, 2.2065E+05, 2.2479E+05,
     + 2.2901E+05, 2.3330E+05, 2.3766E+05, 2.4209E+05, 2.4660E+05,
     +  2.5119E+05, 2.5585E+05, 2.6060E+05, 2.6542E+05, 2.7032E+05,
     + 2.7531E+05, 2.8038E+05, 2.8553E+05, 2.9077E+05, 2.9610E+05,
     +  3.0152E+05, 3.0702E+05, 3.1262E+05, 3.1831E+05, 3.2409E+05,
     + 3.2997E+05, 3.3595E+05, 3.4202E+05, 3.4820E+05, 3.5447E+05,
     +  3.6085E+05, 3.6733E+05, 3.7392E+05, 3.8062E+05, 3.8742E+05,
     + 3.9434E+05, 4.0136E+05, 4.0850E+05, 4.1575E+05, 4.2313E+05,
     +  4.3062E+05, 4.3823E+05, 4.4596E+05, 4.5381E+05, 4.6179E+05,
     + 4.6990E+05, 4.7814E+05, 4.8651E+05, 4.9501E+05, 5.0364E+05,
     +  5.1241E+05, 5.2133E+05, 5.3038E+05, 5.3957E+05, 5.4891E+05,
     + 5.5839E+05, 5.6802E+05, 5.7780E+05, 5.8774E+05, 5.9783E+05,
     +  6.0808E+05, 6.1848E+05, 6.2905E+05, 6.3978E+05, 6.5068E+05,
     + 6.6174E+05, 6.7298E+05, 6.8438E+05, 6.9597E+05, 7.0773E+05,
     +  7.1967E+05, 7.3179E+05, 7.4410E+05, 7.5659E+05, 7.6928E+05,
     + 7.8216E+05, 7.9523E+05, 8.0850E+05, 8.2197E+05, 8.3564E+05,
     +  8.4952E+05, 8.6361E+05, 8.7791E+05, 8.9243E+05, 9.0716E+05,
     + 9.2211E+05, 9.3729E+05, 9.5269E+05, 9.6832E+05, 9.8419E+05,
     +  1.0003E+06, 1.0166E+06, 1.0332E+06, 1.0500E+06, 1.0671E+06,
     + 1.0844E+06, 1.1020E+06, 1.1198E+06, 1.1379E+06, 1.1563E+06,
     +  1.1749E+06, 1.1938E+06, 1.2130E+06, 1.2324E+06, 1.2522E+06,
     + 1.2722E+06, 1.2925E+06, 1.3131E+06, 1.3340E+06, 1.3552E+06,
     +  1.3768E+06, 1.3986E+06, 1.4207E+06, 1.4432E+06, 1.4659E+06,
     + 1.4890E+06, 1.5125E+06, 1.5362E+06, 1.5603E+06, 1.5848E+06,
     +  1.6095E+06, 1.6347E+06, 1.6602E+06, 1.6860E+06, 1.7123E+06,
     + 1.7388E+06, 1.7658E+06, 1.7931E+06, 1.8209E+06, 1.8490E+06,
     +  1.8775E+06, 1.9064E+06, 1.9357E+06, 1.9654E+06, 1.9955E+06,
     + 2.0261E+06, 2.0571E+06, 2.0885E+06, 2.1203E+06, 2.1526E+06,
     +  2.1853E+06, 2.2184E+06, 2.2521E+06, 2.2862E+06, 2.3207E+06,
     + 2.3557E+06, 2.3912E+06, 2.4272E+06, 2.4637E+06, 2.5007E+06,
     +  2.5382E+06, 2.5762E+06, 2.6147E+06, 2.6537E+06, 2.6932E+06,
     + 2.7333E+06, 2.7739E+06, 2.8151E+06, 2.8569E+06, 2.8991E+06,
     +  2.9420E+06, 2.9854E+06, 3.0294E+06, 3.0740E+06, 3.1192E+06,
     + 3.1650E+06, 3.2114E+06, 3.2585E+06, 3.3061E+06, 3.3544E+06,
     +  3.4033E+06, 3.4529E+06, 3.5031E+06, 3.5540E+06, 3.6055E+06,
     + 3.6578E+06, 3.7107E+06, 3.7643E+06, 3.8186E+06, 3.8736E+06,
     +  3.9293E+06, 3.9858E+06, 4.0430E+06, 4.1010E+06, 4.1597E+06,
     + 4.2191E+06, 4.2794E+06, 4.3404E+06, 4.4022E+06, 4.4648E+06,
     +  4.5282E+06, 4.5924E+06, 4.6574E+06, 4.7233E+06, 4.7901E+06,
     + 4.8576E+06, 4.9261E+06, 4.9954E+06, 5.0656E+06, 5.1367E+06,
     +  5.2087E+06, 5.2817E+06, 5.3555E+06, 5.4303E+06, 5.5060E+06,
     + 5.5827E+06, 5.6604E+06, 5.7390E+06, 5.8187E+06, 5.8993E+06,
     +  5.9810E+06, 6.0636E+06, 6.1473E+06, 6.2321E+06, 6.3179E+06,
     + 6.4048E+06, 6.4928E+06, 6.5819E+06, 6.6720E+06, 6.7634E+06,
     +  6.8558E+06, 6.9494E+06, 7.0441E+06, 7.1400E+06, 7.2371E+06,
     + 7.3354E+06, 7.4349E+06, 7.5357E+06, 7.6376E+06, 7.7409E+06,
     +  7.8454E+06, 7.9511E+06, 8.0582E+06, 8.1666E+06, 8.2763E+06,
     + 8.3873E+06, 8.4997E+06, 8.6135E+06, 8.7286E+06, 8.8451E+06,
     +  8.9631E+06, 9.0824E+06, 9.2033E+06, 9.3255E+06, 9.4493E+06,
     + 9.5745E+06, 9.7013E+06, 9.8296E+06, 9.9594E+06, 1.0091E+07,
     +  1.0224E+07, 1.0358E+07, 1.0494E+07, 1.0632E+07, 1.0772E+07,
     + 1.0913E+07, 1.1055E+07, 1.1200E+07, 1.1346E+07, 1.1494E+07,
     +  1.1643E+07, 1.1795E+07, 1.1948E+07, 1.2103E+07, 1.2260E+07,
     + 1.2419E+07, 1.2579E+07, 1.2742E+07, 1.2906E+07, 1.3072E+07,
     +  1.3240E+07, 1.3410E+07, 1.3583E+07, 1.3757E+07, 1.3933E+07,
     + 1.4111E+07, 1.4291E+07, 1.4474E+07, 1.4658E+07, 1.4845E+07,
     +  1.5034E+07, 1.5225E+07, 1.5418E+07, 1.5613E+07, 1.5811E+07,
     + 1.6011E+07, 1.6213E+07, 1.6418E+07, 1.6625E+07, 1.6834E+07,
     +  1.7046E+07, 1.7260E+07, 1.7476E+07, 1.7695E+07, 1.7917E+07,
     + 1.8141E+07, 1.8367E+07, 1.8597E+07, 1.8828E+07, 1.9063E+07,
     +  1.9300E+07, 1.9539E+07, 1.9782E+07, 2.0027E+07, 2.0275E+07,
     + 2.0525E+07, 2.0779E+07, 2.1035E+07, 2.1294E+07, 2.1556E+07,
     +  2.1821E+07, 2.2089E+07, 2.2360E+07, 2.2634E+07, 2.2911E+07,
     + 2.3192E+07, 2.3475E+07, 2.3761E+07, 2.4051E+07, 2.4344E+07,
     +  2.4640E+07, 2.4939E+07, 2.5242E+07, 2.5547E+07, 2.5857E+07,
     + 2.6169E+07, 2.6486E+07, 2.6805E+07, 2.7128E+07, 2.7455E+07,
     +  2.7785E+07, 2.8119E+07, 2.8457E+07, 2.8798E+07, 2.9143E+07,
     + 2.9491E+07, 2.9844E+07, 3.0200E+07, 3.0560E+07, 3.0924E+07,
     +  3.1292E+07, 3.1664E+07, 3.2040E+07, 3.2420E+07, 3.2804E+07,
     + 3.3193E+07, 3.3585E+07, 3.3982E+07, 3.4383E+07, 3.4788E+07,
     +  3.5198E+07, 3.5612E+07, 3.6030E+07, 3.6453E+07, 3.6880E+07,
     + 3.7312E+07, 3.7749E+07, 3.8190E+07, 3.8636E+07, 3.9087E+07,
     +  3.9542E+07, 4.0002E+07, 4.0467E+07, 4.0937E+07, 4.1412E+07,
     + 4.1892E+07, 4.2377E+07, 4.2867E+07, 4.3363E+07, 4.3863E+07,
     +  4.4369E+07, 4.4880E+07, 4.5397E+07, 4.5918E+07, 4.6446E+07,
     + 4.6978E+07, 4.7517E+07, 4.8061E+07, 4.8610E+07, 4.9166E+07,
     +  4.9727E+07, 5.0294E+07, 5.0867E+07, 5.1445E+07, 5.2030E+07,
     + 5.2621E+07, 5.3218E+07, 5.3821E+07, 5.4430E+07, 5.5046E+07,
     +  5.5667E+07, 5.6296E+07, 5.6930E+07, 5.7572E+07, 5.8219E+07,
     + 5.8874E+07, 5.9535E+07, 6.0203E+07, 6.0877E+07, 6.1559E+07,
     +  6.2247E+07, 6.2943E+07, 6.3646E+07, 6.4355E+07, 6.5072E+07,
     + 6.5796E+07, 6.6528E+07, 6.7267E+07, 6.8013E+07, 6.8767E+07,
     +  6.9528E+07, 7.0298E+07, 7.1074E+07, 7.1859E+07, 7.2652E+07,
     + 7.3452E+07, 7.4261E+07, 7.5078E+07, 7.5902E+07, 7.6736E+07,
     +  7.7577E+07, 7.8427E+07, 7.9285E+07, 8.0152E+07, 8.1027E+07,
     + 8.1911E+07, 8.2804E+07, 8.3706E+07, 8.4617E+07, 8.5537E+07,
     +  8.6465E+07, 8.7403E+07, 8.8351E+07, 8.9307E+07, 9.0273E+07,
     + 9.1249E+07, 9.2234E+07, 9.3229E+07, 9.4233E+07, 9.5248E+07,
     +  9.6272E+07, 9.7307E+07, 9.8351E+07, 9.9406E+07, 1.0047E+08,
     + 1.0155E+08, 1.0263E+08, 1.0373E+08, 1.0484E+08, 1.0595E+08,
     +  1.0708E+08, 1.0822E+08, 1.0937E+08, 1.1053E+08, 1.1171E+08,
     + 1.1289E+08, 1.1409E+08, 1.1529E+08, 1.1651E+08, 1.1774E+08,
     +  1.1899E+08, 1.2024E+08, 1.2151E+08, 1.2279E+08, 1.2408E+08,
     + 1.2538E+08, 1.2670E+08, 1.2802E+08, 1.2936E+08, 1.3072E+08,
     +  1.3208E+08, 1.3346E+08, 1.3486E+08, 1.3626E+08, 1.3768E+08,
     + 1.3911E+08, 1.4056E+08, 1.4202E+08, 1.4349E+08, 1.4498E+08,
     +  1.4648E+08, 1.4800E+08, 1.4953E+08, 1.5107E+08, 1.5263E+08,
     + 1.5421E+08, 1.5579E+08, 1.5740E+08, 1.5902E+08, 1.6065E+08,
     +  1.6230E+08, 1.6396E+08, 1.6564E+08, 1.6734E+08, 1.6905E+08,
     + 1.7078E+08, 1.7252E+08, 1.7428E+08, 1.7605E+08, 1.7785E+08,
     +  1.7965E+08, 1.8148E+08, 1.8332E+08, 1.8518E+08, 1.8706E+08,
     + 1.8895E+08, 1.9086E+08, 1.9279E+08, 1.9474E+08, 1.9670E+08,
     +  1.9869E+08, 2.0069E+08, 2.0271E+08, 2.0474E+08, 2.0680E+08,
     + 2.0887E+08, 2.1097E+08, 2.1308E+08, 2.1521E+08, 2.1736E+08,
     +  2.1954E+08, 2.2173E+08, 2.2394E+08, 2.2617E+08, 2.2842E+08,
     + 2.3069E+08, 2.3298E+08, 2.3530E+08, 2.3763E+08, 2.3999E+08,
     +  2.4236E+08, 2.4476E+08, 2.4718E+08, 2.4962E+08, 2.5208E+08,
     + 2.5457E+08, 2.5707E+08, 2.5960E+08, 2.6216E+08, 2.6473E+08,
     +  2.6733E+08, 2.6995E+08, 2.7260E+08, 2.7526E+08, 2.7796E+08,
     + 2.8067E+08, 2.8341E+08, 2.8618E+08, 2.8897E+08, 2.9178E+08,
     +  2.9462E+08, 2.9748E+08, 3.0037E+08, 3.0329E+08, 3.0623E+08,
     + 3.0920E+08, 3.1219E+08, 3.1521E+08, 3.1825E+08, 3.2133E+08,
     +  3.2443E+08, 3.2755E+08, 3.3071E+08, 3.3389E+08, 3.3710E+08,
     + 3.4034E+08, 3.4360E+08, 3.4690E+08, 3.5022E+08, 3.5357E+08,
     +  3.5695E+08, 3.6037E+08, 3.6381E+08, 3.6728E+08, 3.7078E+08,
     + 3.7431E+08, 3.7787E+08, 3.8146E+08, 3.8509E+08, 3.8874E+08,
     +  3.9243E+08, 3.9615E+08, 3.9990E+08, 4.0368E+08, 4.0750E+08,
     + 4.1135E+08, 4.1523E+08, 4.1914E+08, 4.2309E+08, 4.2707E+08,
     +  4.3109E+08, 4.3514E+08, 4.3922E+08, 4.4334E+08, 4.4750E+08,
     + 4.5169E+08, 4.5592E+08, 4.6018E+08, 4.6448E+08, 4.6881E+08,
     +  4.7319E+08, 4.7760E+08, 4.8204E+08, 4.8653E+08, 4.9105E+08,
     + 4.9561E+08, 5.0021E+08, 5.0485E+08, 5.0953E+08, 5.1424E+08,
     +  5.1900E+08, 5.2380E+08, 5.2863E+08, 5.3351E+08, 5.3843E+08,
     + 5.4339E+08, 5.4839E+08, 5.5343E+08, 5.5852E+08, 5.6365E+08,
     +  5.6882E+08, 5.7403E+08, 5.7929E+08, 5.8459E+08, 5.8994E+08,
     + 5.9533E+08, 6.0076E+08, 6.0624E+08, 6.1177E+08, 6.1734E+08,
     +  6.2296E+08, 6.2862E+08, 6.3433E+08, 6.4009E+08, 6.4590E+08,
     + 6.5175E+08, 6.5765E+08, 6.6360E+08, 6.6960E+08, 6.7565E+08,
     +  6.8175E+08, 6.8790E+08, 6.9410E+08, 7.0035E+08, 7.0665E+08,
     + 7.1301E+08, 7.1941E+08, 7.2587E+08, 7.3238E+08, 7.3895E+08,
     +  7.4556E+08, 7.5223E+08, 7.5896E+08, 7.6574E+08, 7.7258E+08,
     + 7.7947E+08, 7.8641E+08, 7.9342E+08, 8.0048E+08, 8.0759E+08,
     +  8.1477E+08, 8.2200E+08, 8.2929E+08, 8.3664E+08, 8.4405E+08,
     + 8.5152E+08, 8.5905E+08, 8.6664E+08, 8.7429E+08, 8.8201E+08,
     +  8.8978E+08, 8.9762E+08, 9.0552E+08, 9.1348E+08, 9.2151E+08,
     + 9.2960E+08, 9.3776E+08, 9.4598E+08, 9.5426E+08, 9.6262E+08,
     +  9.7104E+08, 9.7952E+08, 9.8808E+08, 9.9670E+08, 1.0054E+09,
     + 1.0141E+09, 1.0230E+09, 1.0319E+09, 1.0408E+09, 1.0499E+09,
     +  1.0590E+09, 1.0682E+09, 1.0774E+09, 1.0868E+09, 1.0962E+09,
     + 1.1056E+09, 1.1152E+09, 1.1248E+09, 1.1345E+09, 1.1443E+09,
     +  1.1542E+09, 1.1641E+09, 1.1741E+09, 1.1842E+09, 1.1944E+09,
     + 1.2046E+09, 1.2149E+09, 1.2253E+09, 1.2358E+09, 1.2464E+09,
     +  1.2570E+09, 1.2678E+09, 1.2786E+09, 1.2895E+09, 1.3005E+09,
     + 1.3116E+09, 1.3227E+09, 1.3340E+09, 1.3453E+09, 1.3567E+09,
     +  1.3682E+09, 1.3798E+09, 1.3915E+09, 1.4033E+09, 1.4152E+09,
     + 1.4271E+09, 1.4392E+09, 1.4513E+09, 1.4635E+09, 1.4759E+09,
     +  1.4883E+09, 1.5008E+09, 1.5134E+09, 1.5262E+09, 1.5390E+09,
     + 1.5519E+09, 1.5649E+09, 1.5780E+09, 1.5912E+09, 1.6045E+09,
     +  1.6179E+09, 1.6314E+09, 1.6450E+09, 1.6587E+09, 1.6726E+09,
     + 1.6865E+09, 1.7005E+09, 1.7147E+09, 1.7289E+09, 1.7432E+09,
     +  1.7577E+09, 1.7723E+09, 1.7870E+09, 1.8017E+09, 1.8166E+09,
     + 1.8317E+09, 1.8468E+09, 1.8620E+09, 1.8774E+09, 1.8928E+09,
     +  1.9084E+09, 1.9241E+09, 1.9400E+09, 1.9559E+09, 1.9719E+09,
     + 1.9881E+09, 2.0044E+09, 2.0208E+09, 2.0374E+09, 2.0541E+09,
     +  2.0708E+09, 2.0878E+09, 2.1048E+09, 2.1220E+09, 2.1393E+09,
     + 2.1567E+09, 2.1742E+09, 2.1919E+09, 2.2097E+09, 2.2277E+09,
     +  2.2458E+09, 2.2640E+09, 2.2823E+09, 2.3008E+09, 2.3194E+09,
     + 2.3382E+09, 2.3571E+09, 2.3761E+09, 2.3953E+09, 2.4146E+09,
     +  2.4340E+09, 2.4536E+09, 2.4734E+09, 2.4932E+09, 2.5133E+09,
     + 2.5334E+09, 2.5538E+09, 2.5742E+09, 2.5949E+09, 2.6156E+09,
     +  2.6366E+09, 2.6576E+09, 2.6789E+09, 2.7002E+09, 2.7218E+09,
     + 2.7435E+09, 2.7653E+09, 2.7873E+09, 2.8095E+09, 2.8319E+09,
     +  2.8543E+09, 2.8770E+09, 2.8998E+09, 2.9228E+09, 2.9460E+09,
     + 2.9693E+09, 2.9928E+09, 3.0164E+09, 3.0402E+09, 3.0642E+09,
     +  3.0884E+09, 3.1127E+09, 3.1373E+09, 3.1619E+09, 3.1868E+09,
     + 3.2119E+09, 3.2371E+09, 3.2625E+09, 3.2881E+09, 3.3139E+09,
     +  3.3398E+09, 3.3659E+09, 3.3923E+09, 3.4188E+09, 3.4455E+09,
     + 3.4724E+09, 3.4995E+09, 3.5267E+09, 3.5542E+09, 3.5819E+09,
     +  3.6097E+09, 3.6378E+09, 3.6660E+09, 3.6945E+09, 3.7232E+09,
     + 3.7520E+09, 3.7811E+09, 3.8103E+09, 3.8398E+09, 3.8695E+09,
     +  3.8994E+09, 3.9295E+09, 3.9598E+09, 3.9903E+09, 4.0211E+09,
     + 4.0520E+09, 4.0832E+09, 4.1146E+09, 4.1462E+09, 4.1780E+09,
     +  4.2101E+09, 4.2423E+09, 4.2748E+09, 4.3076E+09, 4.3405E+09,
     + 4.3737E+09, 4.4071E+09, 4.4408E+09, 4.4747E+09, 4.5088E+09,
     +  4.5431E+09, 4.5777E+09, 4.6126E+09, 4.6476E+09, 4.6830E+09,
     + 4.7185E+09, 4.7543E+09, 4.7904E+09, 4.8267E+09, 4.8632E+09,
     +  4.9000E+09, 4.9371E+09, 4.9744E+09, 5.0120E+09, 5.0498E+09,
     + 5.0879E+09, 5.1263E+09, 5.1649E+09, 5.2038E+09, 5.2429E+09,
     +  5.2823E+09, 5.3220E+09, 5.3620E+09, 5.4022E+09, 5.4427E+09,
     + 5.4835E+09, 5.5245E+09, 5.5659E+09, 5.6075E+09, 5.6494E+09,
     +  5.6916E+09, 5.7340E+09, 5.7768E+09, 5.8199E+09, 5.8632E+09,
     + 5.9068E+09, 5.9508E+09, 5.9950E+09, 6.0395E+09, 6.0844E+09,
     +  6.1295E+09, 6.1749E+09, 6.2207E+09, 6.2667E+09, 6.3131E+09,
     + 6.3598E+09, 6.4068E+09, 6.4541E+09, 6.5017E+09, 6.5497E+09,
     +  6.5979E+09, 6.6465E+09, 6.6954E+09, 6.7447E+09, 6.7942E+09,
     + 6.8442E+09, 6.8944E+09, 6.9450E+09, 6.9959E+09, 7.0471E+09,
     +  7.0987E+09, 7.1507E+09, 7.2030E+09, 7.2556E+09, 7.3086E+09,
     + 7.3619E+09, 7.4156E+09, 7.4696E+09, 7.5240E+09, 7.5788E+09,
     +  7.6339E+09, 7.6894E+09, 7.7453E+09, 7.8015E+09, 7.8581E+09,
     + 7.9151E+09, 7.9724E+09, 8.0302E+09, 8.0883E+09, 8.1468E+09,
     +  8.2057E+09, 8.2649E+09, 8.3246E+09, 8.3846E+09, 8.4451E+09,
     + 8.5059E+09, 8.5671E+09, 8.6288E+09, 8.6908E+09, 8.7533E+09,
     +  8.8161E+09, 8.8794E+09, 8.9431E+09, 9.0072E+09, 9.0717E+09,
     + 9.1366E+09, 9.2020E+09, 9.2677E+09, 9.3340E+09, 9.4006E+09,
     +  9.4677E+09, 9.5352E+09, 9.6031E+09, 9.6715E+09, 9.7403E+09,
     + 9.8096E+09, 9.8794E+09, 9.9495E+09, 1.0020E+10, 1.0091E+10,
     +  1.0163E+10, 1.0235E+10, 1.0307E+10, 1.0380E+10, 1.0454E+10,
     + 1.0528E+10, 1.0602E+10, 1.0677E+10, 1.0752E+10, 1.0828E+10,
     +  1.0904E+10, 1.0981E+10, 1.1058E+10, 1.1136E+10, 1.1214E+10,
     + 1.1293E+10, 1.1372E+10, 1.1452E+10, 1.1532E+10, 1.1613E+10,
     +  1.1694E+10, 1.1776E+10, 1.1859E+10, 1.1941E+10, 1.2025E+10,
     + 1.2109E+10, 1.2193E+10, 1.2278E+10, 1.2364E+10, 1.2450E+10,
     +  1.2536E+10, 1.2624E+10, 1.2711E+10, 1.2800E+10, 1.2888E+10,
     + 1.2978E+10, 1.3068E+10, 1.3158E+10, 1.3249E+10, 1.3341E+10,
     +  1.3433E+10, 1.3526E+10, 1.3620E+10, 1.3714E+10, 1.3808E+10,
     + 1.3903E+10, 1.3999E+10, 1.4096E+10, 1.4193E+10, 1.4290E+10,
     +  1.4388E+10, 1.4487E+10, 1.4587E+10, 1.4687E+10, 1.4787E+10,
     + 1.4889E+10, 1.4991E+10, 1.5093E+10, 1.5197E+10, 1.5300E+10,
     +  1.5405E+10, 1.5510E+10, 1.5616E+10, 1.5722E+10, 1.5830E+10,
     + 1.5937E+10, 1.6046E+10, 1.6155E+10, 1.6265E+10, 1.6375E+10,
     +  1.6486E+10, 1.6598E+10, 1.6711E+10, 1.6824E+10, 1.6938E+10,
     + 1.7053E+10, 1.7168E+10, 1.7284E+10, 1.7401E+10, 1.7518E+10,
     +  1.7637E+10, 1.7756E+10, 1.7875E+10, 1.7996E+10, 1.8117E+10,
     + 1.8239E+10, 1.8361E+10, 1.8485E+10, 1.8609E+10, 1.8734E+10,
     +  1.8859E+10, 1.8986E+10, 1.9113E+10, 1.9241E+10, 1.9370E+10,
     + 1.9499E+10, 1.9630E+10, 1.9761E+10, 1.9893E+10, 2.0026E+10,
     +  2.0159E+10, 2.0293E+10, 2.0429E+10, 2.0565E+10, 2.0701E+10,
     + 2.0839E+10, 2.0977E+10, 2.1117E+10, 2.1257E+10, 2.1398E+10,
     +  2.1540E+10, 2.1682E+10, 2.1826E+10, 2.1970E+10, 2.2116E+10,
     + 2.2262E+10, 2.2409E+10, 2.2557E+10, 2.2706E+10, 2.2855E+10,
     +  2.3006E+10, 2.3157E+10, 2.3310E+10, 2.3463E+10, 2.3617E+10,
     + 2.3773E+10, 2.3929E+10, 2.4086E+10, 2.4244E+10, 2.4403E+10,
     +  2.4563E+10, 2.4723E+10, 2.4885E+10, 2.5048E+10, 2.5212E+10,
     + 2.5376E+10, 2.5542E+10, 2.5709E+10, 2.5876E+10, 2.6045E+10,
     +  2.6214E+10, 2.6385E+10, 2.6557E+10, 2.6729E+10, 2.6903E+10,
     + 2.7078E+10, 2.7254E+10, 2.7430E+10, 2.7608E+10, 2.7787E+10,
     +  2.7967E+10, 2.8148E+10, 2.8330E+10, 2.8513E+10, 2.8697E+10,
     + 2.8883E+10, 2.9069E+10, 2.9256E+10, 2.9445E+10, 2.9635E+10,
     +  2.9826E+10, 3.0017E+10, 3.0210E+10, 3.0405E+10, 3.0600E+10,
     + 3.0796E+10, 3.0994E+10, 3.1193E+10, 3.1393E+10, 3.1594E+10,
     +  3.1796E+10, 3.1999E+10, 3.2204E+10, 3.2410E+10, 3.2617E+10,
     + 3.2825E+10, 3.3034E+10, 3.3245E+10, 3.3457E+10, 3.3670E+10,
     +  3.3884E+10, 3.4100E+10, 3.4316E+10, 3.4535E+10, 3.4754E+10,
     + 3.4974E+10, 3.5196E+10, 3.5419E+10, 3.5644E+10, 3.5870E+10,
     +  3.6097E+10, 3.6325E+10, 3.6554E+10, 3.6785E+10, 3.7018E+10,
     + 3.7251E+10, 3.7486E+10, 3.7723E+10, 3.7960E+10, 3.8199E+10,
     +  3.8440E+10, 3.8681E+10, 3.8925E+10, 3.9169E+10, 3.9415E+10,
     + 3.9663E+10, 3.9911E+10, 4.0161E+10, 4.0413E+10, 4.0666E+10,
     +  4.0921E+10, 4.1177E+10, 4.1434E+10, 4.1693E+10, 4.1953E+10,
     + 4.2215E+10, 4.2478E+10, 4.2743E+10, 4.3009E+10, 4.3277E+10,
     +  4.3546E+10, 4.3817E+10, 4.4090E+10, 4.4364E+10, 4.4639E+10,
     + 4.4916E+10, 4.5195E+10, 4.5475E+10, 4.5757E+10, 4.6040E+10,
     +  4.6325E+10, 4.6611E+10, 4.6900E+10, 4.7189E+10, 4.7481E+10,
     + 4.7774E+10, 4.8069E+10, 4.8365E+10, 4.8663E+10, 4.8963E+10,
     +  4.9264E+10, 4.9567E+10, 4.9872E+10, 5.0178E+10, 5.0486E+10,
     + 5.0796E+10, 5.1108E+10, 5.1421E+10, 5.1736E+10, 5.2053E+10,
     +  5.2372E+10, 5.2692E+10, 5.3014E+10, 5.3338E+10, 5.3664E+10,
     + 5.3992E+10, 5.4321E+10, 5.4652E+10, 5.4985E+10, 5.5320E+10,
     +  5.5657E+10, 5.5996E+10, 5.6336E+10, 5.6679E+10, 5.7023E+10,
     + 5.7369E+10, 5.7717E+10, 5.8067E+10, 5.8419E+10, 5.8773E+10,
     +  5.9129E+10, 5.9487E+10, 5.9847E+10, 6.0208E+10, 6.0572E+10,
     + 6.0938E+10, 6.1306E+10, 6.1675E+10, 6.2047E+10, 6.2421E+10,
     +  6.2797E+10, 6.3175E+10, 6.3555E+10, 6.3937E+10, 6.4321E+10,
     + 6.4708E+10, 6.5096E+10, 6.5487E+10, 6.5879E+10, 6.6274E+10,
     +  6.6671E+10, 6.7070E+10, 6.7472E+10, 6.7875E+10, 6.8281E+10,
     + 6.8689E+10, 6.9099E+10, 6.9511E+10, 6.9926E+10, 7.0342E+10,
     +  7.0762E+10, 7.1183E+10, 7.1607E+10, 7.2033E+10, 7.2461E+10,
     + 7.2891E+10, 7.3324E+10, 7.3760E+10, 7.4197E+10, 7.4637E+10,
     +  7.5079E+10, 7.5524E+10, 7.5971E+10, 7.6421E+10, 7.6873E+10,
     + 7.7327E+10, 7.7784E+10, 7.8243E+10, 7.8705E+10, 7.9169E+10,
     +  7.9636E+10, 8.0105E+10, 8.0577E+10, 8.1051E+10, 8.1528E+10,
     + 8.2007E+10, 8.2489E+10, 8.2974E+10, 8.3461E+10, 8.3950E+10,
     +  8.4443E+10, 8.4938E+10, 8.5435E+10, 8.5935E+10, 8.6438E+10,
     + 8.6944E+10, 8.7452E+10, 8.7963E+10, 8.8477E+10, 8.8993E+10,
     +  8.9512E+10, 9.0034E+10, 9.0559E+10, 9.1086E+10, 9.1616E+10,
     + 9.2150E+10, 9.2685E+10, 9.3224E+10, 9.3766E+10, 9.4310E+10,
     +  9.4857E+10, 9.5408E+10, 9.5961E+10, 9.6517E+10, 9.7076E+10,
     + 9.7638E+10, 9.8202E+10, 9.8770E+10, 9.9341E+10, 9.9915E+10,
     +  1.0049E+11, 1.0107E+11, 1.0165E+11, 1.0224E+11, 1.0283E+11,
     + 1.0342E+11, 1.0402E+11, 1.0462E+11, 1.0522E+11, 1.0582E+11,
     +  1.0643E+11, 1.0704E+11, 1.0765E+11, 1.0827E+11, 1.0889E+11,
     + 1.0952E+11, 1.1014E+11, 1.1077E+11, 1.1141E+11, 1.1204E+11,
     +  1.1268E+11, 1.1333E+11, 1.1397E+11, 1.1462E+11, 1.1528E+11,
     + 1.1594E+11, 1.1660E+11, 1.1726E+11, 1.1793E+11, 1.1860E+11,
     +  1.1927E+11, 1.1995E+11, 1.2063E+11, 1.2132E+11, 1.2200E+11,
     + 1.2270E+11, 1.2339E+11, 1.2409E+11, 1.2479E+11, 1.2550E+11,
     +  1.2621E+11, 1.2692E+11, 1.2764E+11, 1.2836E+11, 1.2909E+11,
     + 1.2981E+11, 1.3055E+11, 1.3128E+11, 1.3202E+11, 1.3276E+11,
     +  1.3351E+11, 1.3426E+11, 1.3502E+11, 1.3578E+11, 1.3654E+11,
     + 1.3731E+11, 1.3808E+11, 1.3885E+11, 1.3963E+11, 1.4041E+11,
     +  1.4120E+11, 1.4199E+11, 1.4278E+11, 1.4358E+11, 1.4438E+11,
     + 1.4519E+11, 1.4600E+11, 1.4681E+11, 1.4763E+11, 1.4846E+11,
     +  1.4928E+11, 1.5011E+11, 1.5095E+11, 1.5179E+11, 1.5263E+11,
     + 1.5348E+11, 1.5433E+11, 1.5519E+11, 1.5605E+11, 1.5692E+11,
     +  1.5779E+11, 1.5866E+11, 1.5954E+11, 1.6042E+11, 1.6131E+11,
     + 1.6220E+11, 1.6310E+11, 1.6400E+11, 1.6491E+11, 1.6582E+11,
     +  1.6673E+11, 1.6765E+11, 1.6858E+11, 1.6950E+11, 1.7044E+11,
     + 1.7137E+11, 1.7232E+11, 1.7326E+11, 1.7422E+11, 1.7517E+11,
     +  1.7614E+11, 1.7710E+11, 1.7807E+11, 1.7905E+11, 1.8003E+11,
     + 1.8102E+11, 1.8201E+11, 1.8300E+11, 1.8400E+11, 1.8501E+11,
     +  1.8602E+11, 1.8703E+11, 1.8806E+11, 1.8908E+11, 1.9011E+11,
     + 1.9115E+11, 1.9219E+11, 1.9324E+11, 1.9429E+11, 1.9534E+11,
     +  1.9641E+11, 1.9747E+11, 1.9854E+11, 1.9962E+11, 2.0071E+11,
     + 2.0179E+11, 2.0289E+11, 2.0399E+11, 2.0509E+11, 2.0620E+11,
     +  2.0732E+11, 2.0844E+11, 2.0956E+11, 2.1070E+11, 2.1183E+11,
     + 2.1298E+11, 2.1413E+11, 2.1528E+11, 2.1644E+11, 2.1761E+11,
     +  2.1878E+11, 2.1995E+11, 2.2114E+11, 2.2233E+11, 2.2352E+11,
     + 2.2472E+11, 2.2593E+11, 2.2714E+11, 2.2836E+11, 2.2958E+11,
     +  2.3081E+11, 2.3205E+11, 2.3329E+11, 2.3454E+11, 2.3579E+11,
     + 2.3705E+11, 2.3832E+11, 2.3959E+11, 2.4087E+11, 2.4215E+11,
     +  2.4345E+11, 2.4474E+11, 2.4605E+11, 2.4736E+11, 2.4867E+11,
     + 2.5000E+11, 2.5133E+11, 2.5266E+11, 2.5400E+11, 2.5535E+11,
     +  2.5671E+11, 2.5807E+11, 2.5944E+11, 2.6081E+11, 2.6219E+11,
     + 2.6358E+11, 2.6498E+11, 2.6638E+11, 2.6779E+11, 2.6920E+11,
     +  2.7062E+11, 2.7205E+11, 2.7349E+11, 2.7493E+11, 2.7638E+11,
     + 2.7783E+11, 2.7930E+11, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 6,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_H2    (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 2),QofT( 2,NT),Q(NT)
      data (Tmx(I),I=1,2)/6000.,6000./
      data xgj/ 1.,6./
c...       H2
c...        --        11
      data (QofT( 1,J),J=1,2408)/ 1.0000E+00, 1.0000E+00, 1.0000E+00,
     +  1.0000E+00, 1.0000E+00, 1.0000E+00, 1.0000E+00, 1.0000E+00,
     + 1.0000E+00, 1.0000E+00, 1.0000E+00, 1.0000E+00, 1.0000E+00,
     +  1.0000E+00, 1.0001E+00, 1.0002E+00, 1.0004E+00, 1.0007E+00,
     + 1.0011E+00, 1.0018E+00, 1.0039E+00, 1.0074E+00, 1.0128E+00,
     +  1.0204E+00, 1.0306E+00, 1.0437E+00, 1.0598E+00, 1.0790E+00,
     + 1.1014E+00, 1.1269E+00, 1.1554E+00, 1.1869E+00, 1.2213E+00,
     +  1.2582E+00, 1.2977E+00, 1.3395E+00, 1.3834E+00, 1.4293E+00,
     + 1.4769E+00, 1.5262E+00, 1.5769E+00, 1.6290E+00, 1.6822E+00,
     +  1.7364E+00, 1.7916E+00, 1.8475E+00, 1.9041E+00, 1.9613E+00,
     + 2.0190E+00, 2.0771E+00, 2.1356E+00, 2.1943E+00, 2.2533E+00,
     +  2.3124E+00, 2.3716E+00, 2.4309E+00, 2.4901E+00, 2.5494E+00,
     + 2.6086E+00, 2.6677E+00, 2.7267E+00, 2.7856E+00, 2.8444E+00,
     +  2.9029E+00, 2.9613E+00, 3.0195E+00, 3.0774E+00, 3.1352E+00,
     + 3.1927E+00, 3.2500E+00, 3.3070E+00, 3.3638E+00, 3.4203E+00,
     +  3.4766E+00, 3.5327E+00, 3.5885E+00, 3.6441E+00, 3.6994E+00,
     + 3.7544E+00, 3.8093E+00, 3.8638E+00, 3.9182E+00, 3.9723E+00,
     +  4.0262E+00, 4.0798E+00, 4.1332E+00, 4.1864E+00, 4.2394E+00,
     + 4.2922E+00, 4.3448E+00, 4.3972E+00, 4.4494E+00, 4.5014E+00,
     +  4.5533E+00, 4.6049E+00, 4.6564E+00, 4.7077E+00, 4.7588E+00,
     + 4.8098E+00, 4.8607E+00, 4.9113E+00, 4.9619E+00, 5.0123E+00,
     +  5.0626E+00, 5.1127E+00, 5.1627E+00, 5.2126E+00, 5.2624E+00,
     + 5.3121E+00, 5.3617E+00, 5.4111E+00, 5.4605E+00, 5.5098E+00,
     +  5.5589E+00, 5.6080E+00, 5.6570E+00, 5.7060E+00, 5.7548E+00,
     + 5.8036E+00, 5.8523E+00, 5.9009E+00, 5.9494E+00, 5.9979E+00,
     +  6.0464E+00, 6.0948E+00, 6.1431E+00, 6.1913E+00, 6.2396E+00,
     + 6.2877E+00, 6.3359E+00, 6.3839E+00, 6.4320E+00, 6.4800E+00,
     +  6.5279E+00, 6.5758E+00, 6.6237E+00, 6.6716E+00, 6.7194E+00,
     + 6.7672E+00, 6.8149E+00, 6.8627E+00, 6.9104E+00, 6.9581E+00,
     +  7.0057E+00, 7.0534E+00, 7.1010E+00, 7.1486E+00, 7.1962E+00,
     + 7.2437E+00, 7.2913E+00, 7.3388E+00, 7.3863E+00, 7.4338E+00,
     +  7.4813E+00, 7.5288E+00, 7.5763E+00, 7.6237E+00, 7.6712E+00,
     + 7.7186E+00, 7.7660E+00, 7.8135E+00, 7.8609E+00, 7.9083E+00,
     +  7.9557E+00, 8.0031E+00, 8.0505E+00, 8.0978E+00, 8.1452E+00,
     + 8.1926E+00, 8.2400E+00, 8.2873E+00, 8.3347E+00, 8.3821E+00,
     +  8.4294E+00, 8.4768E+00, 8.5241E+00, 8.5715E+00, 8.6188E+00,
     + 8.6662E+00, 8.7136E+00, 8.7609E+00, 8.8083E+00, 8.8556E+00,
     +  8.9030E+00, 8.9503E+00, 8.9977E+00, 9.0450E+00, 9.0924E+00,
     + 9.1398E+00, 9.1871E+00, 9.2345E+00, 9.2818E+00, 9.3292E+00,
     +  9.3766E+00, 9.4240E+00, 9.4713E+00, 9.5187E+00, 9.5661E+00,
     + 9.6135E+00, 9.6609E+00, 9.7082E+00, 9.7556E+00, 9.8030E+00,
     +  9.8504E+00, 9.8978E+00, 9.9452E+00, 9.9926E+00, 1.0040E+01,
     + 1.0087E+01, 1.0135E+01, 1.0182E+01, 1.0230E+01, 1.0277E+01,
     +  1.0325E+01, 1.0372E+01, 1.0419E+01, 1.0467E+01, 1.0514E+01,
     + 1.0562E+01, 1.0609E+01, 1.0657E+01, 1.0704E+01, 1.0752E+01,
     +  1.0799E+01, 1.0847E+01, 1.0894E+01, 1.0942E+01, 1.0989E+01,
     + 1.1036E+01, 1.1084E+01, 1.1131E+01, 1.1179E+01, 1.1226E+01,
     +  1.1274E+01, 1.1322E+01, 1.1369E+01, 1.1417E+01, 1.1464E+01,
     + 1.1512E+01, 1.1559E+01, 1.1607E+01, 1.1654E+01, 1.1702E+01,
     +  1.1749E+01, 1.1797E+01, 1.1844E+01, 1.1892E+01, 1.1940E+01,
     + 1.1987E+01, 1.2035E+01, 1.2082E+01, 1.2130E+01, 1.2177E+01,
     +  1.2225E+01, 1.2273E+01, 1.2320E+01, 1.2368E+01, 1.2415E+01,
     + 1.2463E+01, 1.2511E+01, 1.2558E+01, 1.2606E+01, 1.2654E+01,
     +  1.2701E+01, 1.2749E+01, 1.2797E+01, 1.2844E+01, 1.2892E+01,
     + 1.2939E+01, 1.2987E+01, 1.3035E+01, 1.3082E+01, 1.3130E+01,
     +  1.3178E+01, 1.3225E+01, 1.3273E+01, 1.3321E+01, 1.3369E+01,
     + 1.3416E+01, 1.3464E+01, 1.3512E+01, 1.3559E+01, 1.3607E+01,
     +  1.3655E+01, 1.3703E+01, 1.3750E+01, 1.3798E+01, 1.3846E+01,
     + 1.3893E+01, 1.3941E+01, 1.3989E+01, 1.4037E+01, 1.4085E+01,
     +  1.4132E+01, 1.4180E+01, 1.4228E+01, 1.4276E+01, 1.4323E+01,
     + 1.4371E+01, 1.4419E+01, 1.4467E+01, 1.4515E+01, 1.4562E+01,
     +  1.4610E+01, 1.4658E+01, 1.4706E+01, 1.4754E+01, 1.4801E+01,
     + 1.4849E+01, 1.4897E+01, 1.4945E+01, 1.4993E+01, 1.5041E+01,
     +  1.5088E+01, 1.5136E+01, 1.5184E+01, 1.5232E+01, 1.5280E+01,
     + 1.5328E+01, 1.5376E+01, 1.5424E+01, 1.5471E+01, 1.5519E+01,
     +  1.5567E+01, 1.5615E+01, 1.5663E+01, 1.5711E+01, 1.5759E+01,
     + 1.5807E+01, 1.5855E+01, 1.5903E+01, 1.5951E+01, 1.5999E+01,
     +  1.6046E+01, 1.6094E+01, 1.6142E+01, 1.6190E+01, 1.6238E+01,
     + 1.6286E+01, 1.6334E+01, 1.6382E+01, 1.6430E+01, 1.6478E+01,
     +  1.6526E+01, 1.6574E+01, 1.6622E+01, 1.6670E+01, 1.6718E+01,
     + 1.6766E+01, 1.6814E+01, 1.6862E+01, 1.6910E+01, 1.6958E+01,
     +  1.7006E+01, 1.7054E+01, 1.7102E+01, 1.7150E+01, 1.7199E+01,
     + 1.7247E+01, 1.7295E+01, 1.7343E+01, 1.7391E+01, 1.7439E+01,
     +  1.7487E+01, 1.7535E+01, 1.7583E+01, 1.7631E+01, 1.7679E+01,
     + 1.7728E+01, 1.7776E+01, 1.7824E+01, 1.7872E+01, 1.7920E+01,
     +  1.7968E+01, 1.8016E+01, 1.8065E+01, 1.8113E+01, 1.8161E+01,
     + 1.8209E+01, 1.8257E+01, 1.8305E+01, 1.8354E+01, 1.8402E+01,
     +  1.8450E+01, 1.8498E+01, 1.8546E+01, 1.8595E+01, 1.8643E+01,
     + 1.8691E+01, 1.8739E+01, 1.8788E+01, 1.8836E+01, 1.8884E+01,
     +  1.8932E+01, 1.8981E+01, 1.9029E+01, 1.9077E+01, 1.9125E+01,
     + 1.9174E+01, 1.9222E+01, 1.9270E+01, 1.9319E+01, 1.9367E+01,
     +  1.9415E+01, 1.9463E+01, 1.9512E+01, 1.9560E+01, 1.9608E+01,
     + 1.9657E+01, 1.9705E+01, 1.9754E+01, 1.9802E+01, 1.9850E+01,
     +  1.9899E+01, 1.9947E+01, 1.9995E+01, 2.0044E+01, 2.0092E+01,
     + 2.0141E+01, 2.0189E+01, 2.0238E+01, 2.0286E+01, 2.0334E+01,
     +  2.0383E+01, 2.0431E+01, 2.0480E+01, 2.0528E+01, 2.0577E+01,
     + 2.0625E+01, 2.0674E+01, 2.0722E+01, 2.0771E+01, 2.0819E+01,
     +  2.0868E+01, 2.0916E+01, 2.0965E+01, 2.1013E+01, 2.1062E+01,
     + 2.1110E+01, 2.1159E+01, 2.1208E+01, 2.1256E+01, 2.1305E+01,
     +  2.1353E+01, 2.1402E+01, 2.1450E+01, 2.1499E+01, 2.1548E+01,
     + 2.1596E+01, 2.1645E+01, 2.1694E+01, 2.1742E+01, 2.1791E+01,
     +  2.1840E+01, 2.1888E+01, 2.1937E+01, 2.1986E+01, 2.2034E+01,
     + 2.2083E+01, 2.2132E+01, 2.2180E+01, 2.2229E+01, 2.2278E+01,
     +  2.2327E+01, 2.2375E+01, 2.2424E+01, 2.2473E+01, 2.2522E+01,
     + 2.2570E+01, 2.2619E+01, 2.2668E+01, 2.2717E+01, 2.2766E+01,
     +  2.2815E+01, 2.2863E+01, 2.2912E+01, 2.2961E+01, 2.3010E+01,
     + 2.3059E+01, 2.3108E+01, 2.3157E+01, 2.3205E+01, 2.3254E+01,
     +  2.3303E+01, 2.3352E+01, 2.3401E+01, 2.3450E+01, 2.3499E+01,
     + 2.3548E+01, 2.3597E+01, 2.3646E+01, 2.3695E+01, 2.3744E+01,
     +  2.3793E+01, 2.3842E+01, 2.3891E+01, 2.3940E+01, 2.3989E+01,
     + 2.4038E+01, 2.4087E+01, 2.4136E+01, 2.4185E+01, 2.4234E+01,
     +  2.4284E+01, 2.4333E+01, 2.4382E+01, 2.4431E+01, 2.4480E+01,
     + 2.4529E+01, 2.4578E+01, 2.4628E+01, 2.4677E+01, 2.4726E+01,
     +  2.4775E+01, 2.4824E+01, 2.4874E+01, 2.4923E+01, 2.4972E+01,
     + 2.5021E+01, 2.5071E+01, 2.5120E+01, 2.5169E+01, 2.5219E+01,
     +  2.5268E+01, 2.5317E+01, 2.5366E+01, 2.5416E+01, 2.5465E+01,
     + 2.5515E+01, 2.5564E+01, 2.5613E+01, 2.5663E+01, 2.5712E+01,
     +  2.5762E+01, 2.5811E+01, 2.5860E+01, 2.5910E+01, 2.5959E+01,
     + 2.6009E+01, 2.6058E+01, 2.6108E+01, 2.6157E+01, 2.6207E+01,
     +  2.6256E+01, 2.6306E+01, 2.6355E+01, 2.6405E+01, 2.6455E+01,
     + 2.6504E+01, 2.6554E+01, 2.6603E+01, 2.6653E+01, 2.6703E+01,
     +  2.6752E+01, 2.6802E+01, 2.6852E+01, 2.6901E+01, 2.6951E+01,
     + 2.7001E+01, 2.7050E+01, 2.7100E+01, 2.7150E+01, 2.7200E+01,
     +  2.7249E+01, 2.7299E+01, 2.7349E+01, 2.7399E+01, 2.7449E+01,
     + 2.7498E+01, 2.7548E+01, 2.7598E+01, 2.7648E+01, 2.7698E+01,
     +  2.7748E+01, 2.7798E+01, 2.7848E+01, 2.7898E+01, 2.7948E+01,
     + 2.7998E+01, 2.8048E+01, 2.8098E+01, 2.8148E+01, 2.8198E+01,
     +  2.8248E+01, 2.8298E+01, 2.8348E+01, 2.8398E+01, 2.8448E+01,
     + 2.8498E+01, 2.8548E+01, 2.8598E+01, 2.8648E+01, 2.8698E+01,
     +  2.8749E+01, 2.8799E+01, 2.8849E+01, 2.8899E+01, 2.8949E+01,
     + 2.9000E+01, 2.9050E+01, 2.9100E+01, 2.9150E+01, 2.9201E+01,
     +  2.9251E+01, 2.9301E+01, 2.9352E+01, 2.9402E+01, 2.9452E+01,
     + 2.9503E+01, 2.9553E+01, 2.9604E+01, 2.9654E+01, 2.9704E+01,
     +  2.9755E+01, 2.9805E+01, 2.9856E+01, 2.9906E+01, 2.9957E+01,
     + 3.0007E+01, 3.0058E+01, 3.0109E+01, 3.0159E+01, 3.0210E+01,
     +  3.0260E+01, 3.0311E+01, 3.0362E+01, 3.0412E+01, 3.0463E+01,
     + 3.0514E+01, 3.0564E+01, 3.0615E+01, 3.0666E+01, 3.0716E+01,
     +  3.0767E+01, 3.0818E+01, 3.0869E+01, 3.0920E+01, 3.0970E+01,
     + 3.1021E+01, 3.1072E+01, 3.1123E+01, 3.1174E+01, 3.1225E+01,
     +  3.1276E+01, 3.1327E+01, 3.1378E+01, 3.1429E+01, 3.1480E+01,
     + 3.1531E+01, 3.1582E+01, 3.1633E+01, 3.1684E+01, 3.1735E+01,
     +  3.1786E+01, 3.1837E+01, 3.1888E+01, 3.1939E+01, 3.1990E+01,
     + 3.2041E+01, 3.2093E+01, 3.2144E+01, 3.2195E+01, 3.2246E+01,
     +  3.2298E+01, 3.2349E+01, 3.2400E+01, 3.2451E+01, 3.2503E+01,
     + 3.2554E+01, 3.2605E+01, 3.2657E+01, 3.2708E+01, 3.2760E+01,
     +  3.2811E+01, 3.2862E+01, 3.2914E+01, 3.2965E+01, 3.3017E+01,
     + 3.3068E+01, 3.3120E+01, 3.3171E+01, 3.3223E+01, 3.3275E+01,
     +  3.3326E+01, 3.3378E+01, 3.3430E+01, 3.3481E+01, 3.3533E+01,
     + 3.3585E+01, 3.3636E+01, 3.3688E+01, 3.3740E+01, 3.3792E+01,
     +  3.3843E+01, 3.3895E+01, 3.3947E+01, 3.3999E+01, 3.4051E+01,
     + 3.4103E+01, 3.4154E+01, 3.4206E+01, 3.4258E+01, 3.4310E+01,
     +  3.4362E+01, 3.4414E+01, 3.4466E+01, 3.4518E+01, 3.4570E+01,
     + 3.4622E+01, 3.4674E+01, 3.4727E+01, 3.4779E+01, 3.4831E+01,
     +  3.4883E+01, 3.4935E+01, 3.4987E+01, 3.5040E+01, 3.5092E+01,
     + 3.5144E+01, 3.5196E+01, 3.5249E+01, 3.5301E+01, 3.5353E+01,
     +  3.5406E+01, 3.5458E+01, 3.5511E+01, 3.5563E+01, 3.5615E+01,
     + 3.5668E+01, 3.5720E+01, 3.5773E+01, 3.5825E+01, 3.5878E+01,
     +  3.5931E+01, 3.5983E+01, 3.6036E+01, 3.6088E+01, 3.6141E+01,
     + 3.6194E+01, 3.6246E+01, 3.6299E+01, 3.6352E+01, 3.6405E+01,
     +  3.6457E+01, 3.6510E+01, 3.6563E+01, 3.6616E+01, 3.6669E+01,
     + 3.6721E+01, 3.6774E+01, 3.6827E+01, 3.6880E+01, 3.6933E+01,
     +  3.6986E+01, 3.7039E+01, 3.7092E+01, 3.7145E+01, 3.7198E+01,
     + 3.7251E+01, 3.7305E+01, 3.7358E+01, 3.7411E+01, 3.7464E+01,
     +  3.7517E+01, 3.7570E+01, 3.7624E+01, 3.7677E+01, 3.7730E+01,
     + 3.7783E+01, 3.7837E+01, 3.7890E+01, 3.7944E+01, 3.7997E+01,
     +  3.8050E+01, 3.8104E+01, 3.8157E+01, 3.8211E+01, 3.8264E+01,
     + 3.8318E+01, 3.8371E+01, 3.8425E+01, 3.8478E+01, 3.8532E+01,
     +  3.8586E+01, 3.8639E+01, 3.8693E+01, 3.8747E+01, 3.8800E+01,
     + 3.8854E+01, 3.8908E+01, 3.8962E+01, 3.9016E+01, 3.9069E+01,
     +  3.9123E+01, 3.9177E+01, 3.9231E+01, 3.9285E+01, 3.9339E+01,
     + 3.9393E+01, 3.9447E+01, 3.9501E+01, 3.9555E+01, 3.9609E+01,
     +  3.9663E+01, 3.9717E+01, 3.9771E+01, 3.9826E+01, 3.9880E+01,
     + 3.9934E+01, 3.9988E+01, 4.0043E+01, 4.0097E+01, 4.0151E+01,
     +  4.0205E+01, 4.0260E+01, 4.0314E+01, 4.0369E+01, 4.0423E+01,
     + 4.0477E+01, 4.0532E+01, 4.0586E+01, 4.0641E+01, 4.0695E+01,
     +  4.0750E+01, 4.0805E+01, 4.0859E+01, 4.0914E+01, 4.0969E+01,
     + 4.1023E+01, 4.1078E+01, 4.1133E+01, 4.1187E+01, 4.1242E+01,
     +  4.1297E+01, 4.1352E+01, 4.1407E+01, 4.1462E+01, 4.1517E+01,
     + 4.1572E+01, 4.1626E+01, 4.1681E+01, 4.1736E+01, 4.1792E+01,
     +  4.1847E+01, 4.1902E+01, 4.1957E+01, 4.2012E+01, 4.2067E+01,
     + 4.2122E+01, 4.2177E+01, 4.2233E+01, 4.2288E+01, 4.2343E+01,
     +  4.2399E+01, 4.2454E+01, 4.2509E+01, 4.2565E+01, 4.2620E+01,
     + 4.2675E+01, 4.2731E+01, 4.2786E+01, 4.2842E+01, 4.2897E+01,
     +  4.2953E+01, 4.3009E+01, 4.3064E+01, 4.3120E+01, 4.3176E+01,
     + 4.3231E+01, 4.3287E+01, 4.3343E+01, 4.3399E+01, 4.3454E+01,
     +  4.3510E+01, 4.3566E+01, 4.3622E+01, 4.3678E+01, 4.3734E+01,
     + 4.3790E+01, 4.3846E+01, 4.3902E+01, 4.3958E+01, 4.4014E+01,
     +  4.4070E+01, 4.4126E+01, 4.4182E+01, 4.4238E+01, 4.4294E+01,
     + 4.4351E+01, 4.4407E+01, 4.4463E+01, 4.4520E+01, 4.4576E+01,
     +  4.4632E+01, 4.4689E+01, 4.4745E+01, 4.4801E+01, 4.4858E+01,
     + 4.4914E+01, 4.4971E+01, 4.5027E+01, 4.5084E+01, 4.5141E+01,
     +  4.5197E+01, 4.5254E+01, 4.5311E+01, 4.5367E+01, 4.5424E+01,
     + 4.5481E+01, 4.5538E+01, 4.5594E+01, 4.5651E+01, 4.5708E+01,
     +  4.5765E+01, 4.5822E+01, 4.5879E+01, 4.5936E+01, 4.5993E+01,
     + 4.6050E+01, 4.6107E+01, 4.6164E+01, 4.6221E+01, 4.6278E+01,
     +  4.6335E+01, 4.6393E+01, 4.6450E+01, 4.6507E+01, 4.6564E+01,
     + 4.6622E+01, 4.6679E+01, 4.6737E+01, 4.6794E+01, 4.6851E+01,
     +  4.6909E+01, 4.6966E+01, 4.7024E+01, 4.7081E+01, 4.7139E+01,
     + 4.7197E+01, 4.7254E+01, 4.7312E+01, 4.7369E+01, 4.7427E+01,
     +  4.7485E+01, 4.7543E+01, 4.7601E+01, 4.7658E+01, 4.7716E+01,
     + 4.7774E+01, 4.7832E+01, 4.7890E+01, 4.7948E+01, 4.8006E+01,
     +  4.8064E+01, 4.8122E+01, 4.8180E+01, 4.8238E+01, 4.8296E+01,
     + 4.8355E+01, 4.8413E+01, 4.8471E+01, 4.8529E+01, 4.8588E+01,
     +  4.8646E+01, 4.8704E+01, 4.8763E+01, 4.8821E+01, 4.8879E+01,
     + 4.8938E+01, 4.8996E+01, 4.9055E+01, 4.9113E+01, 4.9172E+01,
     +  4.9231E+01, 4.9289E+01, 4.9348E+01, 4.9407E+01, 4.9465E+01,
     + 4.9524E+01, 4.9583E+01, 4.9642E+01, 4.9701E+01, 4.9760E+01,
     +  4.9818E+01, 4.9877E+01, 4.9936E+01, 4.9995E+01, 5.0054E+01,
     + 5.0113E+01, 5.0173E+01, 5.0232E+01, 5.0291E+01, 5.0350E+01,
     +  5.0409E+01, 5.0468E+01, 5.0528E+01, 5.0587E+01, 5.0646E+01,
     + 5.0706E+01, 5.0765E+01, 5.0825E+01, 5.0884E+01, 5.0944E+01,
     +  5.1003E+01, 5.1063E+01, 5.1122E+01, 5.1182E+01, 5.1241E+01,
     + 5.1301E+01, 5.1361E+01, 5.1421E+01, 5.1480E+01, 5.1540E+01,
     +  5.1600E+01, 5.1660E+01, 5.1720E+01, 5.1780E+01, 5.1840E+01,
     + 5.1900E+01, 5.1960E+01, 5.2020E+01, 5.2080E+01, 5.2140E+01,
     +  5.2200E+01, 5.2260E+01, 5.2320E+01, 5.2381E+01, 5.2441E+01,
     + 5.2501E+01, 5.2561E+01, 5.2622E+01, 5.2682E+01, 5.2743E+01,
     +  5.2803E+01, 5.2863E+01, 5.2924E+01, 5.2985E+01, 5.3045E+01,
     + 5.3106E+01, 5.3166E+01, 5.3227E+01, 5.3288E+01, 5.3348E+01,
     +  5.3409E+01, 5.3470E+01, 5.3531E+01, 5.3592E+01, 5.3653E+01,
     + 5.3714E+01, 5.3775E+01, 5.3836E+01, 5.3897E+01, 5.3958E+01,
     +  5.4019E+01, 5.4080E+01, 5.4141E+01, 5.4202E+01, 5.4263E+01,
     + 5.4325E+01, 5.4386E+01, 5.4447E+01, 5.4509E+01, 5.4570E+01,
     +  5.4631E+01, 5.4693E+01, 5.4754E+01, 5.4816E+01, 5.4877E+01,
     + 5.4939E+01, 5.5001E+01, 5.5062E+01, 5.5124E+01, 5.5186E+01,
     +  5.5247E+01, 5.5309E+01, 5.5371E+01, 5.5433E+01, 5.5495E+01,
     + 5.5556E+01, 5.5618E+01, 5.5680E+01, 5.5742E+01, 5.5804E+01,
     +  5.5866E+01, 5.5929E+01, 5.5991E+01, 5.6053E+01, 5.6115E+01,
     + 5.6177E+01, 5.6240E+01, 5.6302E+01, 5.6364E+01, 5.6427E+01,
     +  5.6489E+01, 5.6551E+01, 5.6614E+01, 5.6676E+01, 5.6739E+01,
     + 5.6801E+01, 5.6864E+01, 5.6927E+01, 5.6989E+01, 5.7052E+01,
     +  5.7115E+01, 5.7177E+01, 5.7240E+01, 5.7303E+01, 5.7366E+01,
     + 5.7429E+01, 5.7492E+01, 5.7555E+01, 5.7618E+01, 5.7681E+01,
     +  5.7744E+01, 5.7807E+01, 5.7870E+01, 5.7933E+01, 5.7997E+01,
     + 5.8060E+01, 5.8123E+01, 5.8186E+01, 5.8250E+01, 5.8313E+01,
     +  5.8376E+01, 5.8440E+01, 5.8503E+01, 5.8567E+01, 5.8630E+01,
     + 5.8694E+01, 5.8758E+01, 5.8821E+01, 5.8885E+01, 5.8949E+01,
     +  5.9012E+01, 5.9076E+01, 5.9140E+01, 5.9204E+01, 5.9268E+01,
     + 5.9332E+01, 5.9396E+01, 5.9460E+01, 5.9524E+01, 5.9588E+01,
     +  5.9652E+01, 5.9716E+01, 5.9780E+01, 5.9844E+01, 5.9909E+01,
     + 5.9973E+01, 6.0037E+01, 6.0101E+01, 6.0166E+01, 6.0230E+01,
     +  6.0295E+01, 6.0359E+01, 6.0424E+01, 6.0488E+01, 6.0553E+01,
     + 6.0617E+01, 6.0682E+01, 6.0747E+01, 6.0812E+01, 6.0876E+01,
     +  6.0941E+01, 6.1006E+01, 6.1071E+01, 6.1136E+01, 6.1201E+01,
     + 6.1266E+01, 6.1331E+01, 6.1396E+01, 6.1461E+01, 6.1526E+01,
     +  6.1591E+01, 6.1656E+01, 6.1721E+01, 6.1787E+01, 6.1852E+01,
     + 6.1917E+01, 6.1983E+01, 6.2048E+01, 6.2113E+01, 6.2179E+01,
     +  6.2244E+01, 6.2310E+01, 6.2376E+01, 6.2441E+01, 6.2507E+01,
     + 6.2572E+01, 6.2638E+01, 6.2704E+01, 6.2770E+01, 6.2836E+01,
     +  6.2901E+01, 6.2967E+01, 6.3033E+01, 6.3099E+01, 6.3165E+01,
     + 6.3231E+01, 6.3297E+01, 6.3363E+01, 6.3430E+01, 6.3496E+01,
     +  6.3562E+01, 6.3628E+01, 6.3695E+01, 6.3761E+01, 6.3827E+01,
     + 6.3894E+01, 6.3960E+01, 6.4027E+01, 6.4093E+01, 6.4160E+01,
     +  6.4226E+01, 6.4293E+01, 6.4360E+01, 6.4426E+01, 6.4493E+01,
     + 6.4560E+01, 6.4627E+01, 6.4693E+01, 6.4760E+01, 6.4827E+01,
     +  6.4894E+01, 6.4961E+01, 6.5028E+01, 6.5095E+01, 6.5162E+01,
     + 6.5229E+01, 6.5297E+01, 6.5364E+01, 6.5431E+01, 6.5498E+01,
     +  6.5566E+01, 6.5633E+01, 6.5700E+01, 6.5768E+01, 6.5835E+01,
     + 6.5903E+01, 6.5970E+01, 6.6038E+01, 6.6106E+01, 6.6173E+01,
     +  6.6241E+01, 6.6309E+01, 6.6376E+01, 6.6444E+01, 6.6512E+01,
     + 6.6580E+01, 6.6648E+01, 6.6716E+01, 6.6784E+01, 6.6852E+01,
     +  6.6920E+01, 6.6988E+01, 6.7056E+01, 6.7124E+01, 6.7192E+01,
     + 6.7261E+01, 6.7329E+01, 6.7397E+01, 6.7466E+01, 6.7534E+01,
     +  6.7602E+01, 6.7671E+01, 6.7739E+01, 6.7808E+01, 6.7877E+01,
     + 6.7945E+01, 6.8014E+01, 6.8083E+01, 6.8151E+01, 6.8220E+01,
     +  6.8289E+01, 6.8358E+01, 6.8427E+01, 6.8496E+01, 6.8565E+01,
     + 6.8634E+01, 6.8703E+01, 6.8772E+01, 6.8841E+01, 6.8910E+01,
     +  6.8979E+01, 6.9048E+01, 6.9118E+01, 6.9187E+01, 6.9256E+01,
     + 6.9326E+01, 6.9395E+01, 6.9465E+01, 6.9534E+01, 6.9604E+01,
     +  6.9673E+01, 6.9743E+01, 6.9812E+01, 6.9882E+01, 6.9952E+01,
     + 7.0022E+01, 7.0091E+01, 7.0161E+01, 7.0231E+01, 7.0301E+01,
     +  7.0371E+01, 7.0441E+01, 7.0511E+01, 7.0581E+01, 7.0651E+01,
     + 7.0721E+01, 7.0792E+01, 7.0862E+01, 7.0932E+01, 7.1002E+01,
     +  7.1073E+01, 7.1143E+01, 7.1214E+01, 7.1284E+01, 7.1355E+01,
     + 7.1425E+01, 7.1496E+01, 7.1566E+01, 7.1637E+01, 7.1708E+01,
     +  7.1778E+01, 7.1849E+01, 7.1920E+01, 7.1991E+01, 7.2062E+01,
     + 7.2133E+01, 7.2204E+01, 7.2275E+01, 7.2346E+01, 7.2417E+01,
     +  7.2488E+01, 7.2559E+01, 7.2630E+01, 7.2701E+01, 7.2773E+01,
     + 7.2844E+01, 7.2915E+01, 7.2987E+01, 7.3058E+01, 7.3130E+01,
     +  7.3201E+01, 7.3273E+01, 7.3344E+01, 7.3416E+01, 7.3488E+01,
     + 7.3559E+01, 7.3631E+01, 7.3703E+01, 7.3775E+01, 7.3847E+01,
     +  7.3919E+01, 7.3991E+01, 7.4063E+01, 7.4135E+01, 7.4207E+01,
     + 7.4279E+01, 7.4351E+01, 7.4423E+01, 7.4495E+01, 7.4568E+01,
     +  7.4640E+01, 7.4712E+01, 7.4785E+01, 7.4857E+01, 7.4930E+01,
     + 7.5002E+01, 7.5075E+01, 7.5147E+01, 7.5220E+01, 7.5292E+01,
     +  7.5365E+01, 7.5438E+01, 7.5511E+01, 7.5584E+01, 7.5656E+01,
     + 7.5729E+01, 7.5802E+01, 7.5875E+01, 7.5948E+01, 7.6021E+01,
     +  7.6094E+01, 7.6168E+01, 7.6241E+01, 7.6314E+01, 7.6387E+01,
     + 7.6461E+01, 7.6534E+01, 7.6607E+01, 7.6681E+01, 7.6754E+01,
     +  7.6828E+01, 7.6901E+01, 7.6975E+01, 7.7049E+01, 7.7122E+01,
     + 7.7196E+01, 7.7270E+01, 7.7343E+01, 7.7417E+01, 7.7491E+01,
     +  7.7565E+01, 7.7639E+01, 7.7713E+01, 7.7787E+01, 7.7861E+01,
     + 7.7935E+01, 7.8009E+01, 7.8084E+01, 7.8158E+01, 7.8232E+01,
     +  7.8307E+01, 7.8381E+01, 7.8455E+01, 7.8530E+01, 7.8604E+01,
     + 7.8679E+01, 7.8753E+01, 7.8828E+01, 7.8903E+01, 7.8977E+01,
     +  7.9052E+01, 7.9127E+01, 7.9202E+01, 7.9276E+01, 7.9351E+01,
     + 7.9426E+01, 7.9501E+01, 7.9576E+01, 7.9651E+01, 7.9726E+01,
     +  7.9802E+01, 7.9877E+01, 7.9952E+01, 8.0027E+01, 8.0103E+01,
     + 8.0178E+01, 8.0253E+01, 8.0329E+01, 8.0404E+01, 8.0480E+01,
     +  8.0555E+01, 8.0631E+01, 8.0706E+01, 8.0782E+01, 8.0858E+01,
     + 8.0934E+01, 8.1009E+01, 8.1085E+01, 8.1161E+01, 8.1237E+01,
     +  8.1313E+01, 8.1389E+01, 8.1465E+01, 8.1541E+01, 8.1617E+01,
     + 8.1694E+01, 8.1770E+01, 8.1846E+01, 8.1922E+01, 8.1999E+01,
     +  8.2075E+01, 8.2151E+01, 8.2228E+01, 8.2304E+01, 8.2381E+01,
     + 8.2458E+01, 8.2534E+01, 8.2611E+01, 8.2688E+01, 8.2764E+01,
     +  8.2841E+01, 8.2918E+01, 8.2995E+01, 8.3072E+01, 8.3149E+01,
     + 8.3226E+01, 8.3303E+01, 8.3380E+01, 8.3457E+01, 8.3534E+01,
     +  8.3612E+01, 8.3689E+01, 8.3766E+01, 8.3844E+01, 8.3921E+01,
     + 8.3998E+01, 8.4076E+01, 8.4153E+01, 8.4231E+01, 8.4309E+01,
     +  8.4386E+01, 8.4464E+01, 8.4542E+01, 8.4619E+01, 8.4697E+01,
     + 8.4775E+01, 8.4853E+01, 8.4931E+01, 8.5009E+01, 8.5087E+01,
     +  8.5165E+01, 8.5243E+01, 8.5321E+01, 8.5399E+01, 8.5478E+01,
     + 8.5556E+01, 8.5634E+01, 8.5713E+01, 8.5791E+01, 8.5870E+01,
     +  8.5948E+01, 8.6027E+01, 8.6105E+01, 8.6184E+01, 8.6262E+01,
     + 8.6341E+01, 8.6420E+01, 8.6499E+01, 8.6578E+01, 8.6656E+01,
     +  8.6735E+01, 8.6814E+01, 8.6893E+01, 8.6972E+01, 8.7051E+01,
     + 8.7131E+01, 8.7210E+01, 8.7289E+01, 8.7368E+01, 8.7448E+01,
     +  8.7527E+01, 8.7606E+01, 8.7686E+01, 8.7765E+01, 8.7845E+01,
     + 8.7924E+01, 8.8004E+01, 8.8084E+01, 8.8163E+01, 8.8243E+01,
     +  8.8323E+01, 8.8403E+01, 8.8483E+01, 8.8562E+01, 8.8642E+01,
     + 8.8722E+01, 8.8802E+01, 8.8883E+01, 8.8963E+01, 8.9043E+01,
     +  8.9123E+01, 8.9203E+01, 8.9284E+01, 8.9364E+01, 8.9444E+01,
     + 8.9525E+01, 8.9605E+01, 8.9686E+01, 8.9766E+01, 8.9847E+01,
     +  8.9928E+01, 9.0008E+01, 9.0089E+01, 9.0170E+01, 9.0251E+01,
     + 9.0332E+01, 9.0413E+01, 9.0493E+01, 9.0574E+01, 9.0656E+01,
     +  9.0737E+01, 9.0818E+01, 9.0899E+01, 9.0980E+01, 9.1061E+01,
     + 9.1143E+01, 9.1224E+01, 9.1305E+01, 9.1387E+01, 9.1468E+01,
     +  9.1550E+01, 9.1631E+01, 9.1713E+01, 9.1795E+01, 9.1876E+01,
     + 9.1958E+01, 9.2040E+01, 9.2122E+01, 9.2204E+01, 9.2286E+01,
     +  9.2368E+01, 9.2450E+01, 9.2532E+01, 9.2614E+01, 9.2696E+01,
     + 9.2778E+01, 9.2860E+01, 9.2943E+01, 9.3025E+01, 9.3107E+01,
     +  9.3190E+01, 9.3272E+01, 9.3355E+01, 9.3437E+01, 9.3520E+01,
     + 9.3602E+01, 9.3685E+01, 9.3768E+01, 9.3851E+01, 9.3933E+01,
     +  9.4016E+01, 9.4099E+01, 9.4182E+01, 9.4265E+01, 9.4348E+01,
     + 9.4431E+01, 9.4514E+01, 9.4597E+01, 9.4681E+01, 9.4764E+01,
     +  9.4847E+01, 9.4931E+01, 9.5014E+01, 9.5097E+01, 9.5181E+01,
     + 9.5264E+01, 9.5348E+01, 9.5432E+01, 9.5515E+01, 9.5599E+01,
     +  9.5683E+01, 9.5766E+01, 9.5850E+01, 9.5934E+01, 9.6018E+01,
     + 9.6102E+01, 9.6186E+01, 9.6270E+01, 9.6354E+01, 9.6438E+01,
     +  9.6523E+01, 9.6607E+01, 9.6691E+01, 9.6775E+01, 9.6860E+01,
     + 9.6944E+01, 9.7029E+01, 9.7113E+01, 9.7198E+01, 9.7282E+01,
     +  9.7367E+01, 9.7452E+01, 9.7536E+01, 9.7621E+01, 9.7706E+01,
     + 9.7791E+01, 9.7876E+01, 9.7961E+01, 9.8046E+01, 9.8131E+01,
     +  9.8216E+01, 9.8301E+01, 9.8386E+01, 9.8471E+01, 9.8557E+01,
     + 9.8642E+01, 9.8727E+01, 9.8813E+01, 9.8898E+01, 9.8984E+01,
     +  9.9069E+01, 9.9155E+01, 9.9241E+01, 9.9326E+01, 9.9412E+01,
     + 9.9498E+01, 9.9584E+01, 9.9669E+01, 9.9755E+01, 9.9841E+01,
     +  9.9927E+01, 1.0001E+02, 1.0010E+02, 1.0019E+02, 1.0027E+02,
     + 1.0036E+02, 1.0044E+02, 1.0053E+02, 1.0062E+02, 1.0070E+02,
     +  1.0079E+02, 1.0088E+02, 1.0096E+02, 1.0105E+02, 1.0114E+02,
     + 1.0122E+02, 1.0131E+02, 1.0140E+02, 1.0148E+02, 1.0157E+02,
     +  1.0166E+02, 1.0174E+02, 1.0183E+02, 1.0192E+02, 1.0201E+02,
     + 1.0209E+02, 1.0218E+02, 1.0227E+02, 1.0235E+02, 1.0244E+02,
     +  1.0253E+02, 1.0262E+02, 1.0270E+02, 1.0279E+02, 1.0288E+02,
     + 1.0297E+02, 1.0305E+02, 1.0314E+02, 1.0323E+02, 1.0332E+02,
     +  1.0341E+02, 1.0349E+02, 1.0358E+02, 1.0367E+02, 1.0376E+02,
     + 1.0384E+02, 1.0393E+02, 1.0402E+02, 1.0411E+02, 1.0420E+02,
     +  1.0429E+02, 1.0437E+02, 1.0446E+02, 1.0455E+02, 1.0464E+02,
     + 1.0473E+02, 1.0482E+02, 1.0491E+02, 1.0499E+02, 1.0508E+02,
     +  1.0517E+02, 1.0526E+02, 1.0535E+02, 1.0544E+02, 1.0553E+02,
     + 1.0562E+02, 1.0571E+02, 1.0579E+02, 1.0588E+02, 1.0597E+02,
     +  1.0606E+02, 1.0615E+02, 1.0624E+02, 1.0633E+02, 1.0642E+02,
     + 1.0651E+02, 1.0660E+02, 1.0669E+02, 1.0678E+02, 1.0687E+02,
     +  1.0696E+02, 1.0705E+02, 1.0714E+02, 1.0723E+02, 1.0732E+02,
     + 1.0741E+02, 1.0750E+02, 1.0759E+02, 1.0768E+02, 1.0777E+02,
     +  1.0786E+02, 1.0795E+02, 1.0804E+02, 1.0813E+02, 1.0822E+02,
     + 1.0831E+02, 1.0840E+02, 1.0849E+02, 1.0858E+02, 1.0867E+02,
     +  1.0876E+02, 1.0885E+02, 1.0894E+02, 1.0903E+02, 1.0912E+02,
     + 1.0921E+02, 1.0931E+02, 1.0940E+02, 1.0949E+02, 1.0958E+02,
     +  1.0967E+02, 1.0976E+02, 1.0985E+02, 1.0994E+02, 1.1003E+02,
     + 1.1013E+02, 1.1022E+02, 1.1031E+02, 1.1040E+02, 1.1049E+02,
     +  1.1058E+02, 1.1068E+02, 1.1077E+02, 1.1086E+02, 1.1095E+02,
     + 1.1104E+02, 1.1113E+02, 1.1123E+02, 1.1132E+02, 1.1141E+02,
     +  1.1150E+02, 1.1159E+02, 1.1169E+02, 1.1178E+02, 1.1187E+02,
     + 1.1196E+02, 1.1206E+02, 1.1215E+02, 1.1224E+02, 1.1233E+02,
     +  1.1243E+02, 1.1252E+02, 1.1261E+02, 1.1270E+02, 1.1280E+02,
     + 1.1289E+02, 1.1298E+02, 1.1307E+02, 1.1317E+02, 1.1326E+02,
     +  1.1335E+02, 1.1345E+02, 1.1354E+02, 1.1363E+02, 1.1373E+02,
     + 1.1382E+02, 1.1391E+02, 1.1401E+02, 1.1410E+02, 1.1419E+02,
     +  1.1429E+02, 1.1438E+02, 1.1447E+02, 1.1457E+02, 1.1466E+02,
     + 1.1475E+02, 1.1485E+02, 1.1494E+02, 1.1504E+02, 1.1513E+02,
     +  1.1522E+02, 1.1532E+02, 1.1541E+02, 1.1551E+02, 1.1560E+02,
     + 1.1569E+02, 1.1579E+02, 1.1588E+02, 1.1598E+02, 1.1607E+02,
     +  1.1617E+02, 1.1626E+02, 1.1636E+02, 1.1645E+02, 1.1654E+02,
     + 1.1664E+02, 1.1673E+02, 1.1683E+02, 1.1692E+02, 1.1702E+02,
     +  1.1711E+02, 1.1721E+02, 1.1730E+02, 1.1740E+02, 1.1749E+02,
     + 1.1759E+02, 1.1768E+02, 1.1778E+02, 1.1788E+02, 1.1797E+02,
     +  1.1807E+02, 1.1816E+02, 1.1826E+02, 1.1835E+02, 1.1845E+02,
     + 1.1854E+02, 1.1864E+02, 1.1874E+02, 1.1883E+02, 1.1893E+02,
     +  1.1902E+02, 1.1912E+02, 1.1922E+02, 1.1931E+02, 1.1941E+02,
     + 1.1950E+02, 1.1960E+02, 1.1970E+02, 1.1979E+02, 1.1989E+02,
     +  1.1999E+02, 1.2008E+02, 1.2018E+02, 1.2027E+02, 1.2037E+02,
     + 1.2047E+02, 1.2056E+02, 1.2066E+02, 1.2076E+02, 1.2086E+02,
     +  1.2095E+02, 1.2105E+02, 1.2115E+02, 1.2124E+02, 1.2134E+02,
     + 1.2144E+02, 1.2153E+02, 1.2163E+02, 1.2173E+02, 1.2183E+02,
     +  1.2192E+02, 1.2202E+02, 1.2212E+02, 1.2222E+02, 1.2231E+02,
     + 1.2241E+02, 1.2251E+02, 1.2261E+02, 1.2270E+02, 1.2280E+02,
     +  1.2290E+02, 1.2300E+02, 1.2310E+02, 1.2319E+02, 1.2329E+02,
     + 1.2339E+02, 1.2349E+02, 1.2359E+02, 1.2369E+02, 1.2378E+02,
     +  1.2388E+02, 1.2398E+02, 1.2408E+02, 1.2418E+02, 1.2428E+02,
     + 1.2438E+02, 1.2447E+02, 1.2457E+02, 1.2467E+02, 1.2477E+02,
     +  1.2487E+02, 1.2497E+02, 1.2507E+02, 1.2517E+02, 1.2527E+02,
     + 1.2536E+02, 1.2546E+02, 1.2556E+02, 1.2566E+02, 1.2576E+02,
     +  1.2586E+02, 1.2596E+02, 1.2606E+02, 1.2616E+02, 1.2626E+02,
     + 1.2636E+02, 1.2646E+02, 1.2656E+02, 1.2666E+02, 1.2676E+02,
     +  1.2686E+02, 1.2696E+02, 1.2706E+02, 1.2716E+02, 1.2726E+02,
     + 1.2736E+02, 1.2746E+02, 1.2756E+02, 1.2766E+02, 1.2776E+02,
     +  1.2786E+02, 1.2796E+02, 1.2806E+02, 1.2816E+02, 1.2826E+02,
     + 1.2836E+02, 1.2846E+02, 1.2856E+02, 1.2866E+02, 1.2876E+02,
     +  1.2887E+02, 1.2897E+02, 1.2907E+02, 1.2917E+02, 1.2927E+02,
     + 1.2937E+02, 1.2947E+02, 1.2957E+02, 1.2967E+02, 1.2978E+02,
     +  1.2988E+02, 1.2998E+02, 1.3008E+02, 1.3018E+02, 1.3028E+02,
     + 1.3039E+02, 1.3049E+02, 1.3059E+02, 1.3069E+02, 1.3079E+02,
     +  1.3089E+02, 1.3100E+02, 1.3110E+02, 1.3120E+02, 1.3130E+02,
     + 1.3140E+02, 1.3151E+02, 1.3161E+02, 1.3171E+02, 1.3181E+02,
     +  1.3192E+02, 1.3202E+02, 1.3212E+02, 1.3222E+02, 1.3233E+02,
     + 1.3243E+02, 1.3253E+02, 1.3263E+02, 1.3274E+02, 1.3284E+02,
     +  1.3294E+02, 1.3305E+02, 1.3315E+02, 1.3325E+02, 1.3336E+02,
     + 1.3346E+02, 1.3356E+02, 1.3366E+02, 1.3377E+02, 1.3387E+02,
     +  1.3397E+02, 1.3408E+02, 1.3418E+02, 1.3429E+02, 1.3439E+02,
     + 1.3449E+02, 1.3460E+02, 1.3470E+02, 1.3480E+02, 1.3491E+02,
     +  1.3501E+02, 1.3512E+02, 1.3522E+02, 1.3532E+02, 1.3543E+02,
     + 1.3553E+02, 1.3564E+02, 1.3574E+02, 1.3585E+02, 1.3595E+02,
     +  1.3605E+02, 1.3616E+02, 1.3626E+02, 1.3637E+02, 1.3647E+02,
     + 1.3658E+02, 1.3668E+02, 1.3679E+02, 1.3689E+02, 1.3700E+02,
     +  1.3710E+02, 1.3721E+02, 1.3731E+02, 1.3742E+02, 1.3752E+02,
     + 1.3763E+02, 1.3773E+02, 1.3784E+02, 1.3794E+02, 1.3805E+02,
     +  1.3815E+02, 1.3826E+02, 1.3837E+02, 1.3847E+02, 1.3858E+02,
     + 1.3868E+02, 1.3879E+02, 1.3889E+02, 1.3900E+02, 1.3911E+02,
     +  1.3921E+02, 1.3932E+02, 1.3942E+02, 1.3953E+02, 1.3964E+02,
     + 1.3974E+02, 1.3985E+02, 1.3996E+02, 1.4006E+02, 1.4017E+02,
     +  1.4028E+02, 1.4038E+02, 1.4049E+02, 1.4059E+02, 1.4070E+02,
     + 1.4081E+02, 1.4092E+02, 1.4102E+02, 1.4113E+02, 1.4124E+02,
     +  1.4134E+02, 1.4145E+02, 1.4156E+02, 1.4166E+02, 1.4177E+02,
     + 1.4188E+02, 1.4199E+02, 1.4209E+02, 1.4220E+02, 1.4231E+02,
     +  1.4242E+02, 1.4252E+02, 1.4263E+02, 1.4274E+02, 1.4285E+02,
     + 1.4296E+02, 1.4306E+02, 1.4317E+02, 1.4328E+02, 1.4339E+02,
     +  1.4350E+02, 1.4360E+02, 1.4371E+02, 1.4382E+02, 1.4393E+02,
     + 1.4404E+02, 1.4414E+02, 1.4425E+02, 1.4436E+02, 1.4447E+02,
     +  1.4458E+02, 1.4469E+02, 1.4480E+02, 1.4491E+02, 1.4501E+02,
     + 1.4512E+02, 1.4523E+02, 1.4534E+02, 1.4545E+02, 1.4556E+02,
     +  1.4567E+02, 1.4578E+02, 1.4589E+02, 1.4600E+02, 1.4611E+02,
     + 1.4621E+02, 1.4632E+02, 1.4643E+02, 1.4654E+02, 1.4665E+02,
     +  1.4676E+02, 1.4687E+02, 1.4698E+02, 1.4709E+02, 1.4720E+02,
     + 1.4731E+02, 1.4742E+02, 1.4753E+02, 1.4764E+02, 1.4775E+02,
     +  1.4786E+02, 1.4797E+02, 1.4808E+02, 1.4819E+02, 1.4830E+02,
     + 1.4841E+02, 1.4852E+02, 1.4863E+02, 1.4875E+02, 1.4886E+02,
     +  1.4897E+02, 1.4908E+02, 1.4919E+02, 1.4930E+02, 1.4941E+02,
     + 1.4952E+02, 1.4963E+02, 1.4974E+02, 1.4985E+02, 1.4997E+02,
     +  1.5008E+02, 1.5019E+02, 1.5030E+02, 1.5041E+02, 1.5052E+02,
     + 1.5063E+02, 1.5075E+02, 1.5086E+02, 1.5097E+02, 1.5108E+02,
     +  1.5119E+02, 1.5130E+02, 1.5142E+02, 1.5153E+02, 1.5164E+02,
     + 1.5175E+02, 1.5186E+02, 1.5198E+02, 1.5209E+02, 1.5220E+02,
     +  1.5231E+02, 1.5243E+02, 1.5254E+02, 1.5265E+02, 1.5276E+02,
     + 1.5288E+02, 1.5299E+02, 1.5310E+02, 1.5321E+02, 1.5333E+02,
     +  1.5344E+02, 1.5355E+02, 1.5367E+02, 1.5378E+02, 1.5389E+02,
     + 1.5400E+02, 1.5412E+02, 1.5423E+02, 1.5434E+02, 1.5446E+02,
     +  1.5457E+02, 1.5468E+02, 1.5480E+02, 1.5491E+02, 1.5503E+02,
     + 1.5514E+02, 1.5525E+02, 1.5537E+02, 1.5548E+02, 1.5559E+02,
     +  1.5571E+02, 1.5582E+02, 1.5594E+02, 1.5605E+02, 1.5616E+02,
     + 1.5628E+02, 1.5639E+02, 1.5651E+02, 1.5662E+02, 1.5674E+02,
     +  1.5685E+02, 1.5697E+02, 1.5708E+02, 1.5719E+02, 1.5731E+02,
     + 1.5742E+02, 1.5754E+02, 1.5765E+02, 1.5777E+02, 1.5788E+02,
     +  1.5800E+02, 1.5811E+02, 1.5823E+02, 1.5834E+02, 1.5846E+02,
     + 1.5857E+02, 1.5869E+02, 1.5881E+02, 1.5892E+02, 1.5904E+02,
     +  1.5915E+02, 1.5927E+02, 1.5938E+02, 1.5950E+02, 1.5961E+02,
     + 1.5973E+02, 1.5985E+02, 1.5996E+02, 1.6008E+02, 1.6019E+02,
     +  1.6031E+02, 1.6043E+02, 1.6054E+02, 1.6066E+02, 1.6078E+02,
     + 1.6089E+02, 1.6101E+02, 1.6112E+02, 1.6124E+02, 1.6136E+02,
     +  1.6147E+02, 1.6159E+02, 1.6171E+02, 1.6183E+02, 1.6194E+02,
     + 1.6206E+02, 1.6218E+02, 1.6229E+02, 1.6241E+02, 1.6253E+02,
     +  1.6264E+02, 1.6276E+02, 1.6288E+02, 1.6300E+02, 1.6311E+02,
     + 1.6323E+02, 1.6335E+02, 1.6347E+02, 1.6358E+02, 1.6370E+02,
     +  1.6382E+02, 1.6394E+02, 1.6406E+02, 1.6417E+02, 1.6429E+02,
     + 1.6441E+02, 1.6453E+02, 1.6465E+02, 1.6476E+02, 1.6488E+02,
     +  1.6500E+02, 1.6512E+02, 1.6524E+02, 1.6536E+02, 1.6547E+02,
     + 1.6559E+02, 1.6571E+02, 1.6583E+02, 1.6595E+02, 1.6607E+02,
     +  1.6619E+02, 1.6630E+02, 1.6642E+02, 1.6654E+02, 1.6666E+02,
     + 1.6678E+02, 1.6690E+02, 1.6702E+02, 1.6714E+02, 1.6726E+02,
     +  1.6738E+02, 1.6750E+02, 1.6762E+02, 1.6774E+02, 1.6786E+02,
     + 1.6798E+02, 1.6810E+02, 1.6822E+02, 1.6833E+02, 1.6845E+02,
     +  1.6857E+02, 1.6869E+02, 1.6881E+02, 1.6894E+02, 1.6906E+02,
     + 1.6918E+02, 1.6930E+02, 1.6942E+02, 1.6954E+02, 1.6966E+02,
     +  1.6978E+02, 1.6990E+02, 1.7002E+02, 1.7014E+02, 1.7026E+02,
     + 1.7038E+02, 1.7050E+02, 1.7062E+02, 1.7074E+02, 1.7086E+02,
     +  1.7099E+02, 1.7111E+02, 1.7123E+02, 1.7135E+02, 1.7147E+02,
     + 1.7159E+02, 1.7171E+02, 1.7183E+02, 1.7196E+02, 1.7208E+02,
     +  1.7220E+02, 1.7232E+02, 1.7244E+02, 1.7256E+02, 1.7269E+02,
     + 1.7281E+02, 1.7293E+02, 1.7305E+02, 1.7317E+02, 1.7330E+02,
     +  1.7342E+02, 1.7354E+02, 1.7366E+02, 1.7379E+02, 1.7391E+02,
     + 1.7403E+02, 1.7415E+02, 1.7428E+02, 1.7440E+02, 1.7452E+02,
     +  1.7464E+02, 1.7477E+02, 1.7489E+02, 1.7501E+02, 1.7514E+02/
      data (QofT( 1,J),J=2409,4510)/ 1.7526E+02, 1.7538E+02,
     +  1.7550E+02, 1.7563E+02, 1.7575E+02, 1.7587E+02, 1.7600E+02,
     +  1.7612E+02,
     + 1.7624E+02, 1.7637E+02, 1.7649E+02, 1.7662E+02, 1.7674E+02,
     +  1.7686E+02, 1.7699E+02, 1.7711E+02, 1.7723E+02, 1.7736E+02,
     + 1.7748E+02, 1.7761E+02, 1.7773E+02, 1.7786E+02, 1.7798E+02,
     +  1.7810E+02, 1.7823E+02, 1.7835E+02, 1.7848E+02, 1.7860E+02,
     + 1.7873E+02, 1.7885E+02, 1.7898E+02, 1.7910E+02, 1.7923E+02,
     +  1.7935E+02, 1.7948E+02, 1.7960E+02, 1.7973E+02, 1.7985E+02,
     + 1.7998E+02, 1.8010E+02, 1.8023E+02, 1.8035E+02, 1.8048E+02,
     +  1.8060E+02, 1.8073E+02, 1.8085E+02, 1.8098E+02, 1.8110E+02,
     + 1.8123E+02, 1.8136E+02, 1.8148E+02, 1.8161E+02, 1.8173E+02,
     +  1.8186E+02, 1.8199E+02, 1.8211E+02, 1.8224E+02, 1.8236E+02,
     + 1.8249E+02, 1.8262E+02, 1.8274E+02, 1.8287E+02, 1.8300E+02,
     +  1.8312E+02, 1.8325E+02, 1.8338E+02, 1.8350E+02, 1.8363E+02,
     + 1.8376E+02, 1.8388E+02, 1.8401E+02, 1.8414E+02, 1.8427E+02,
     +  1.8439E+02, 1.8452E+02, 1.8465E+02, 1.8477E+02, 1.8490E+02,
     + 1.8503E+02, 1.8516E+02, 1.8528E+02, 1.8541E+02, 1.8554E+02,
     +  1.8567E+02, 1.8579E+02, 1.8592E+02, 1.8605E+02, 1.8618E+02,
     + 1.8631E+02, 1.8643E+02, 1.8656E+02, 1.8669E+02, 1.8682E+02,
     +  1.8695E+02, 1.8708E+02, 1.8720E+02, 1.8733E+02, 1.8746E+02,
     + 1.8759E+02, 1.8772E+02, 1.8785E+02, 1.8798E+02, 1.8810E+02,
     +  1.8823E+02, 1.8836E+02, 1.8849E+02, 1.8862E+02, 1.8875E+02,
     + 1.8888E+02, 1.8901E+02, 1.8914E+02, 1.8927E+02, 1.8940E+02,
     +  1.8953E+02, 1.8966E+02, 1.8978E+02, 1.8991E+02, 1.9004E+02,
     + 1.9017E+02, 1.9030E+02, 1.9043E+02, 1.9056E+02, 1.9069E+02,
     +  1.9082E+02, 1.9095E+02, 1.9108E+02, 1.9121E+02, 1.9134E+02,
     + 1.9147E+02, 1.9161E+02, 1.9174E+02, 1.9187E+02, 1.9200E+02,
     +  1.9213E+02, 1.9226E+02, 1.9239E+02, 1.9252E+02, 1.9265E+02,
     + 1.9278E+02, 1.9291E+02, 1.9304E+02, 1.9317E+02, 1.9331E+02,
     +  1.9344E+02, 1.9357E+02, 1.9370E+02, 1.9383E+02, 1.9396E+02,
     + 1.9409E+02, 1.9422E+02, 1.9436E+02, 1.9449E+02, 1.9462E+02,
     +  1.9475E+02, 1.9488E+02, 1.9502E+02, 1.9515E+02, 1.9528E+02,
     + 1.9541E+02, 1.9554E+02, 1.9568E+02, 1.9581E+02, 1.9594E+02,
     +  1.9607E+02, 1.9620E+02, 1.9634E+02, 1.9647E+02, 1.9660E+02,
     + 1.9674E+02, 1.9687E+02, 1.9700E+02, 1.9713E+02, 1.9727E+02,
     +  1.9740E+02, 1.9753E+02, 1.9767E+02, 1.9780E+02, 1.9793E+02,
     + 1.9806E+02, 1.9820E+02, 1.9833E+02, 1.9846E+02, 1.9860E+02,
     +  1.9873E+02, 1.9887E+02, 1.9900E+02, 1.9913E+02, 1.9927E+02,
     + 1.9940E+02, 1.9953E+02, 1.9967E+02, 1.9980E+02, 1.9994E+02,
     +  2.0007E+02, 2.0020E+02, 2.0034E+02, 2.0047E+02, 2.0061E+02,
     + 2.0074E+02, 2.0088E+02, 2.0101E+02, 2.0114E+02, 2.0128E+02,
     +  2.0141E+02, 2.0155E+02, 2.0168E+02, 2.0182E+02, 2.0195E+02,
     + 2.0209E+02, 2.0222E+02, 2.0236E+02, 2.0249E+02, 2.0263E+02,
     +  2.0276E+02, 2.0290E+02, 2.0304E+02, 2.0317E+02, 2.0331E+02,
     + 2.0344E+02, 2.0358E+02, 2.0371E+02, 2.0385E+02, 2.0398E+02,
     +  2.0412E+02, 2.0426E+02, 2.0439E+02, 2.0453E+02, 2.0466E+02,
     + 2.0480E+02, 2.0494E+02, 2.0507E+02, 2.0521E+02, 2.0535E+02,
     +  2.0548E+02, 2.0562E+02, 2.0576E+02, 2.0589E+02, 2.0603E+02,
     + 2.0617E+02, 2.0630E+02, 2.0644E+02, 2.0658E+02, 2.0671E+02,
     +  2.0685E+02, 2.0699E+02, 2.0712E+02, 2.0726E+02, 2.0740E+02,
     + 2.0754E+02, 2.0767E+02, 2.0781E+02, 2.0795E+02, 2.0809E+02,
     +  2.0822E+02, 2.0836E+02, 2.0850E+02, 2.0864E+02, 2.0877E+02,
     + 2.0891E+02, 2.0905E+02, 2.0919E+02, 2.0933E+02, 2.0947E+02,
     +  2.0960E+02, 2.0974E+02, 2.0988E+02, 2.1002E+02, 2.1016E+02,
     + 2.1030E+02, 2.1043E+02, 2.1057E+02, 2.1071E+02, 2.1085E+02,
     +  2.1099E+02, 2.1113E+02, 2.1127E+02, 2.1141E+02, 2.1154E+02,
     + 2.1168E+02, 2.1182E+02, 2.1196E+02, 2.1210E+02, 2.1224E+02,
     +  2.1238E+02, 2.1252E+02, 2.1266E+02, 2.1280E+02, 2.1294E+02,
     + 2.1308E+02, 2.1322E+02, 2.1336E+02, 2.1350E+02, 2.1364E+02,
     +  2.1378E+02, 2.1392E+02, 2.1406E+02, 2.1420E+02, 2.1434E+02,
     + 2.1448E+02, 2.1462E+02, 2.1476E+02, 2.1490E+02, 2.1504E+02,
     +  2.1518E+02, 2.1532E+02, 2.1546E+02, 2.1560E+02, 2.1574E+02,
     + 2.1588E+02, 2.1603E+02, 2.1617E+02, 2.1631E+02, 2.1645E+02,
     +  2.1659E+02, 2.1673E+02, 2.1687E+02, 2.1701E+02, 2.1716E+02,
     + 2.1730E+02, 2.1744E+02, 2.1758E+02, 2.1772E+02, 2.1786E+02,
     +  2.1801E+02, 2.1815E+02, 2.1829E+02, 2.1843E+02, 2.1857E+02,
     + 2.1872E+02, 2.1886E+02, 2.1900E+02, 2.1914E+02, 2.1928E+02,
     +  2.1943E+02, 2.1957E+02, 2.1971E+02, 2.1985E+02, 2.2000E+02,
     + 2.2014E+02, 2.2028E+02, 2.2043E+02, 2.2057E+02, 2.2071E+02,
     +  2.2085E+02, 2.2100E+02, 2.2114E+02, 2.2128E+02, 2.2143E+02,
     + 2.2157E+02, 2.2171E+02, 2.2186E+02, 2.2200E+02, 2.2214E+02,
     +  2.2229E+02, 2.2243E+02, 2.2257E+02, 2.2272E+02, 2.2286E+02,
     + 2.2301E+02, 2.2315E+02, 2.2329E+02, 2.2344E+02, 2.2358E+02,
     +  2.2373E+02, 2.2387E+02, 2.2402E+02, 2.2416E+02, 2.2430E+02,
     + 2.2445E+02, 2.2459E+02, 2.2474E+02, 2.2488E+02, 2.2503E+02,
     +  2.2517E+02, 2.2532E+02, 2.2546E+02, 2.2561E+02, 2.2575E+02,
     + 2.2590E+02, 2.2604E+02, 2.2619E+02, 2.2633E+02, 2.2648E+02,
     +  2.2662E+02, 2.2677E+02, 2.2691E+02, 2.2706E+02, 2.2721E+02,
     + 2.2735E+02, 2.2750E+02, 2.2764E+02, 2.2779E+02, 2.2794E+02,
     +  2.2808E+02, 2.2823E+02, 2.2837E+02, 2.2852E+02, 2.2867E+02,
     + 2.2881E+02, 2.2896E+02, 2.2910E+02, 2.2925E+02, 2.2940E+02,
     +  2.2954E+02, 2.2969E+02, 2.2984E+02, 2.2998E+02, 2.3013E+02,
     + 2.3028E+02, 2.3043E+02, 2.3057E+02, 2.3072E+02, 2.3087E+02,
     +  2.3101E+02, 2.3116E+02, 2.3131E+02, 2.3146E+02, 2.3160E+02,
     + 2.3175E+02, 2.3190E+02, 2.3205E+02, 2.3219E+02, 2.3234E+02,
     +  2.3249E+02, 2.3264E+02, 2.3279E+02, 2.3293E+02, 2.3308E+02,
     + 2.3323E+02, 2.3338E+02, 2.3353E+02, 2.3367E+02, 2.3382E+02,
     +  2.3397E+02, 2.3412E+02, 2.3427E+02, 2.3442E+02, 2.3457E+02,
     + 2.3471E+02, 2.3486E+02, 2.3501E+02, 2.3516E+02, 2.3531E+02,
     +  2.3546E+02, 2.3561E+02, 2.3576E+02, 2.3591E+02, 2.3606E+02,
     + 2.3621E+02, 2.3636E+02, 2.3650E+02, 2.3665E+02, 2.3680E+02,
     +  2.3695E+02, 2.3710E+02, 2.3725E+02, 2.3740E+02, 2.3755E+02,
     + 2.3770E+02, 2.3785E+02, 2.3800E+02, 2.3815E+02, 2.3830E+02,
     +  2.3845E+02, 2.3860E+02, 2.3875E+02, 2.3890E+02, 2.3906E+02,
     + 2.3921E+02, 2.3936E+02, 2.3951E+02, 2.3966E+02, 2.3981E+02,
     +  2.3996E+02, 2.4011E+02, 2.4026E+02, 2.4041E+02, 2.4056E+02,
     + 2.4072E+02, 2.4087E+02, 2.4102E+02, 2.4117E+02, 2.4132E+02,
     +  2.4147E+02, 2.4162E+02, 2.4178E+02, 2.4193E+02, 2.4208E+02,
     + 2.4223E+02, 2.4238E+02, 2.4253E+02, 2.4269E+02, 2.4284E+02,
     +  2.4299E+02, 2.4314E+02, 2.4330E+02, 2.4345E+02, 2.4360E+02,
     + 2.4375E+02, 2.4390E+02, 2.4406E+02, 2.4421E+02, 2.4436E+02,
     +  2.4452E+02, 2.4467E+02, 2.4482E+02, 2.4497E+02, 2.4513E+02,
     + 2.4528E+02, 2.4543E+02, 2.4559E+02, 2.4574E+02, 2.4589E+02,
     +  2.4605E+02, 2.4620E+02, 2.4635E+02, 2.4651E+02, 2.4666E+02,
     + 2.4681E+02, 2.4697E+02, 2.4712E+02, 2.4728E+02, 2.4743E+02,
     +  2.4758E+02, 2.4774E+02, 2.4789E+02, 2.4805E+02, 2.4820E+02,
     + 2.4835E+02, 2.4851E+02, 2.4866E+02, 2.4882E+02, 2.4897E+02,
     +  2.4913E+02, 2.4928E+02, 2.4944E+02, 2.4959E+02, 2.4975E+02,
     + 2.4990E+02, 2.5005E+02, 2.5021E+02, 2.5036E+02, 2.5052E+02,
     +  2.5068E+02, 2.5083E+02, 2.5099E+02, 2.5114E+02, 2.5130E+02,
     + 2.5145E+02, 2.5161E+02, 2.5176E+02, 2.5192E+02, 2.5207E+02,
     +  2.5223E+02, 2.5239E+02, 2.5254E+02, 2.5270E+02, 2.5285E+02,
     + 2.5301E+02, 2.5317E+02, 2.5332E+02, 2.5348E+02, 2.5364E+02,
     +  2.5379E+02, 2.5395E+02, 2.5411E+02, 2.5426E+02, 2.5442E+02,
     + 2.5458E+02, 2.5473E+02, 2.5489E+02, 2.5505E+02, 2.5520E+02,
     +  2.5536E+02, 2.5552E+02, 2.5567E+02, 2.5583E+02, 2.5599E+02,
     + 2.5615E+02, 2.5630E+02, 2.5646E+02, 2.5662E+02, 2.5678E+02,
     +  2.5693E+02, 2.5709E+02, 2.5725E+02, 2.5741E+02, 2.5757E+02,
     + 2.5772E+02, 2.5788E+02, 2.5804E+02, 2.5820E+02, 2.5836E+02,
     +  2.5851E+02, 2.5867E+02, 2.5883E+02, 2.5899E+02, 2.5915E+02,
     + 2.5931E+02, 2.5947E+02, 2.5962E+02, 2.5978E+02, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        12
      data (QofT( 2,J),J=1,2408)/ 6.0000E+00, 6.0000E+00, 6.0000E+00,
     +  6.0000E+00, 6.0000E+00, 6.0000E+00, 6.0000E+00, 6.0000E+00,
     + 6.0000E+00, 6.0000E+00, 6.0002E+00, 6.0004E+00, 6.0009E+00,
     +  6.0019E+00, 6.0035E+00, 6.0059E+00, 6.0095E+00, 6.0144E+00,
     + 6.0209E+00, 6.0293E+00, 6.0526E+00, 6.0855E+00, 6.1291E+00,
     +  6.1837E+00, 6.2494E+00, 6.3260E+00, 6.4129E+00, 6.5095E+00,
     + 6.6151E+00, 6.7288E+00, 6.8500E+00, 6.9779E+00, 7.1117E+00,
     +  7.2509E+00, 7.3948E+00, 7.5429E+00, 7.6947E+00, 7.8497E+00,
     + 8.0077E+00, 8.1683E+00, 8.3311E+00, 8.4960E+00, 8.6626E+00,
     +  8.8309E+00, 9.0006E+00, 9.1716E+00, 9.3437E+00, 9.5169E+00,
     + 9.6910E+00, 9.8660E+00, 1.0042E+01, 1.0218E+01, 1.0395E+01,
     +  1.0573E+01, 1.0751E+01, 1.0930E+01, 1.1109E+01, 1.1289E+01,
     + 1.1469E+01, 1.1649E+01, 1.1830E+01, 1.2011E+01, 1.2192E+01,
     +  1.2373E+01, 1.2555E+01, 1.2737E+01, 1.2920E+01, 1.3102E+01,
     + 1.3285E+01, 1.3468E+01, 1.3651E+01, 1.3834E+01, 1.4018E+01,
     +  1.4201E+01, 1.4385E+01, 1.4569E+01, 1.4753E+01, 1.4937E+01,
     + 1.5122E+01, 1.5306E+01, 1.5491E+01, 1.5675E+01, 1.5860E+01,
     +  1.6045E+01, 1.6230E+01, 1.6415E+01, 1.6600E+01, 1.6785E+01,
     + 1.6971E+01, 1.7156E+01, 1.7341E+01, 1.7527E+01, 1.7713E+01,
     +  1.7898E+01, 1.8084E+01, 1.8270E+01, 1.8456E+01, 1.8642E+01,
     + 1.8828E+01, 1.9014E+01, 1.9200E+01, 1.9386E+01, 1.9572E+01,
     +  1.9759E+01, 1.9945E+01, 2.0131E+01, 2.0318E+01, 2.0504E+01,
     + 2.0691E+01, 2.0878E+01, 2.1064E+01, 2.1251E+01, 2.1438E+01,
     +  2.1624E+01, 2.1811E+01, 2.1998E+01, 2.2185E+01, 2.2372E+01,
     + 2.2559E+01, 2.2746E+01, 2.2933E+01, 2.3120E+01, 2.3307E+01,
     +  2.3494E+01, 2.3681E+01, 2.3868E+01, 2.4056E+01, 2.4243E+01,
     + 2.4430E+01, 2.4618E+01, 2.4805E+01, 2.4992E+01, 2.5180E+01,
     +  2.5367E+01, 2.5555E+01, 2.5742E+01, 2.5930E+01, 2.6117E+01,
     + 2.6305E+01, 2.6493E+01, 2.6680E+01, 2.6868E+01, 2.7056E+01,
     +  2.7244E+01, 2.7431E+01, 2.7619E+01, 2.7807E+01, 2.7995E+01,
     + 2.8183E+01, 2.8370E+01, 2.8558E+01, 2.8746E+01, 2.8934E+01,
     +  2.9122E+01, 2.9310E+01, 2.9498E+01, 2.9686E+01, 2.9874E+01,
     + 3.0063E+01, 3.0251E+01, 3.0439E+01, 3.0627E+01, 3.0815E+01,
     +  3.1003E+01, 3.1192E+01, 3.1380E+01, 3.1568E+01, 3.1756E+01,
     + 3.1945E+01, 3.2133E+01, 3.2321E+01, 3.2510E+01, 3.2698E+01,
     +  3.2887E+01, 3.3075E+01, 3.3264E+01, 3.3452E+01, 3.3641E+01,
     + 3.3829E+01, 3.4018E+01, 3.4206E+01, 3.4395E+01, 3.4583E+01,
     +  3.4772E+01, 3.4961E+01, 3.5149E+01, 3.5338E+01, 3.5527E+01,
     + 3.5715E+01, 3.5904E+01, 3.6093E+01, 3.6282E+01, 3.6470E+01,
     +  3.6659E+01, 3.6848E+01, 3.7037E+01, 3.7226E+01, 3.7415E+01,
     + 3.7604E+01, 3.7792E+01, 3.7981E+01, 3.8170E+01, 3.8359E+01,
     +  3.8548E+01, 3.8737E+01, 3.8926E+01, 3.9115E+01, 3.9304E+01,
     + 3.9494E+01, 3.9683E+01, 3.9872E+01, 4.0061E+01, 4.0250E+01,
     +  4.0439E+01, 4.0628E+01, 4.0818E+01, 4.1007E+01, 4.1196E+01,
     + 4.1385E+01, 4.1575E+01, 4.1764E+01, 4.1953E+01, 4.2142E+01,
     +  4.2332E+01, 4.2521E+01, 4.2711E+01, 4.2900E+01, 4.3089E+01,
     + 4.3279E+01, 4.3468E+01, 4.3658E+01, 4.3847E+01, 4.4037E+01,
     +  4.4226E+01, 4.4416E+01, 4.4605E+01, 4.4795E+01, 4.4984E+01,
     + 4.5174E+01, 4.5364E+01, 4.5553E+01, 4.5743E+01, 4.5933E+01,
     +  4.6122E+01, 4.6312E+01, 4.6502E+01, 4.6691E+01, 4.6881E+01,
     + 4.7071E+01, 4.7261E+01, 4.7450E+01, 4.7640E+01, 4.7830E+01,
     +  4.8020E+01, 4.8210E+01, 4.8400E+01, 4.8590E+01, 4.8780E+01,
     + 4.8970E+01, 4.9160E+01, 4.9350E+01, 4.9540E+01, 4.9730E+01,
     +  4.9920E+01, 5.0110E+01, 5.0300E+01, 5.0490E+01, 5.0680E+01,
     + 5.0870E+01, 5.1060E+01, 5.1250E+01, 5.1441E+01, 5.1631E+01,
     +  5.1821E+01, 5.2011E+01, 5.2202E+01, 5.2392E+01, 5.2582E+01,
     + 5.2772E+01, 5.2963E+01, 5.3153E+01, 5.3344E+01, 5.3534E+01,
     +  5.3724E+01, 5.3915E+01, 5.4105E+01, 5.4296E+01, 5.4486E+01,
     + 5.4677E+01, 5.4867E+01, 5.5058E+01, 5.5248E+01, 5.5439E+01,
     +  5.5630E+01, 5.5820E+01, 5.6011E+01, 5.6202E+01, 5.6392E+01,
     + 5.6583E+01, 5.6774E+01, 5.6965E+01, 5.7155E+01, 5.7346E+01,
     +  5.7537E+01, 5.7728E+01, 5.7919E+01, 5.8110E+01, 5.8301E+01,
     + 5.8492E+01, 5.8683E+01, 5.8874E+01, 5.9065E+01, 5.9256E+01,
     +  5.9447E+01, 5.9638E+01, 5.9829E+01, 6.0020E+01, 6.0211E+01,
     + 6.0403E+01, 6.0594E+01, 6.0785E+01, 6.0976E+01, 6.1168E+01,
     +  6.1359E+01, 6.1550E+01, 6.1742E+01, 6.1933E+01, 6.2125E+01,
     + 6.2316E+01, 6.2507E+01, 6.2699E+01, 6.2890E+01, 6.3082E+01,
     +  6.3274E+01, 6.3465E+01, 6.3657E+01, 6.3849E+01, 6.4040E+01,
     + 6.4232E+01, 6.4424E+01, 6.4616E+01, 6.4807E+01, 6.4999E+01,
     +  6.5191E+01, 6.5383E+01, 6.5575E+01, 6.5767E+01, 6.5959E+01,
     + 6.6151E+01, 6.6343E+01, 6.6535E+01, 6.6727E+01, 6.6919E+01,
     +  6.7111E+01, 6.7304E+01, 6.7496E+01, 6.7688E+01, 6.7880E+01,
     + 6.8073E+01, 6.8265E+01, 6.8458E+01, 6.8650E+01, 6.8842E+01,
     +  6.9035E+01, 6.9227E+01, 6.9420E+01, 6.9613E+01, 6.9805E+01,
     + 6.9998E+01, 7.0191E+01, 7.0383E+01, 7.0576E+01, 7.0769E+01,
     +  7.0962E+01, 7.1155E+01, 7.1348E+01, 7.1541E+01, 7.1734E+01,
     + 7.1927E+01, 7.2120E+01, 7.2313E+01, 7.2506E+01, 7.2699E+01,
     +  7.2892E+01, 7.3086E+01, 7.3279E+01, 7.3472E+01, 7.3666E+01,
     + 7.3859E+01, 7.4053E+01, 7.4246E+01, 7.4440E+01, 7.4633E+01,
     +  7.4827E+01, 7.5021E+01, 7.5214E+01, 7.5408E+01, 7.5602E+01,
     + 7.5796E+01, 7.5990E+01, 7.6184E+01, 7.6378E+01, 7.6572E+01,
     +  7.6766E+01, 7.6960E+01, 7.7154E+01, 7.7348E+01, 7.7542E+01,
     + 7.7737E+01, 7.7931E+01, 7.8125E+01, 7.8320E+01, 7.8514E+01,
     +  7.8709E+01, 7.8903E+01, 7.9098E+01, 7.9293E+01, 7.9487E+01,
     + 7.9682E+01, 7.9877E+01, 8.0072E+01, 8.0267E+01, 8.0462E+01,
     +  8.0657E+01, 8.0852E+01, 8.1047E+01, 8.1242E+01, 8.1437E+01,
     + 8.1632E+01, 8.1828E+01, 8.2023E+01, 8.2218E+01, 8.2414E+01,
     +  8.2609E+01, 8.2805E+01, 8.3001E+01, 8.3196E+01, 8.3392E+01,
     + 8.3588E+01, 8.3784E+01, 8.3980E+01, 8.4175E+01, 8.4371E+01,
     +  8.4568E+01, 8.4764E+01, 8.4960E+01, 8.5156E+01, 8.5352E+01,
     + 8.5549E+01, 8.5745E+01, 8.5942E+01, 8.6138E+01, 8.6335E+01,
     +  8.6531E+01, 8.6728E+01, 8.6925E+01, 8.7122E+01, 8.7318E+01,
     + 8.7515E+01, 8.7712E+01, 8.7909E+01, 8.8107E+01, 8.8304E+01,
     +  8.8501E+01, 8.8698E+01, 8.8896E+01, 8.9093E+01, 8.9291E+01,
     + 8.9488E+01, 8.9686E+01, 8.9883E+01, 9.0081E+01, 9.0279E+01,
     +  9.0477E+01, 9.0675E+01, 9.0873E+01, 9.1071E+01, 9.1269E+01,
     + 9.1467E+01, 9.1665E+01, 9.1864E+01, 9.2062E+01, 9.2261E+01,
     +  9.2459E+01, 9.2658E+01, 9.2856E+01, 9.3055E+01, 9.3254E+01,
     + 9.3453E+01, 9.3652E+01, 9.3851E+01, 9.4050E+01, 9.4249E+01,
     +  9.4448E+01, 9.4648E+01, 9.4847E+01, 9.5046E+01, 9.5246E+01,
     + 9.5445E+01, 9.5645E+01, 9.5845E+01, 9.6045E+01, 9.6245E+01,
     +  9.6445E+01, 9.6645E+01, 9.6845E+01, 9.7045E+01, 9.7245E+01,
     + 9.7445E+01, 9.7646E+01, 9.7846E+01, 9.8047E+01, 9.8247E+01,
     +  9.8448E+01, 9.8649E+01, 9.8850E+01, 9.9051E+01, 9.9252E+01,
     + 9.9453E+01, 9.9654E+01, 9.9855E+01, 1.0006E+02, 1.0026E+02,
     +  1.0046E+02, 1.0066E+02, 1.0086E+02, 1.0106E+02, 1.0127E+02,
     + 1.0147E+02, 1.0167E+02, 1.0187E+02, 1.0207E+02, 1.0228E+02,
     +  1.0248E+02, 1.0268E+02, 1.0288E+02, 1.0309E+02, 1.0329E+02,
     + 1.0349E+02, 1.0369E+02, 1.0390E+02, 1.0410E+02, 1.0430E+02,
     +  1.0451E+02, 1.0471E+02, 1.0491E+02, 1.0512E+02, 1.0532E+02,
     + 1.0552E+02, 1.0573E+02, 1.0593E+02, 1.0613E+02, 1.0634E+02,
     +  1.0654E+02, 1.0675E+02, 1.0695E+02, 1.0715E+02, 1.0736E+02,
     + 1.0756E+02, 1.0777E+02, 1.0797E+02, 1.0818E+02, 1.0838E+02,
     +  1.0859E+02, 1.0879E+02, 1.0900E+02, 1.0920E+02, 1.0941E+02,
     + 1.0961E+02, 1.0982E+02, 1.1002E+02, 1.1023E+02, 1.1044E+02,
     +  1.1064E+02, 1.1085E+02, 1.1105E+02, 1.1126E+02, 1.1147E+02,
     + 1.1167E+02, 1.1188E+02, 1.1208E+02, 1.1229E+02, 1.1250E+02,
     +  1.1270E+02, 1.1291E+02, 1.1312E+02, 1.1333E+02, 1.1353E+02,
     + 1.1374E+02, 1.1395E+02, 1.1415E+02, 1.1436E+02, 1.1457E+02,
     +  1.1478E+02, 1.1499E+02, 1.1519E+02, 1.1540E+02, 1.1561E+02,
     + 1.1582E+02, 1.1603E+02, 1.1623E+02, 1.1644E+02, 1.1665E+02,
     +  1.1686E+02, 1.1707E+02, 1.1728E+02, 1.1749E+02, 1.1770E+02,
     + 1.1791E+02, 1.1811E+02, 1.1832E+02, 1.1853E+02, 1.1874E+02,
     +  1.1895E+02, 1.1916E+02, 1.1937E+02, 1.1958E+02, 1.1979E+02,
     + 1.2000E+02, 1.2021E+02, 1.2042E+02, 1.2064E+02, 1.2085E+02,
     +  1.2106E+02, 1.2127E+02, 1.2148E+02, 1.2169E+02, 1.2190E+02,
     + 1.2211E+02, 1.2232E+02, 1.2254E+02, 1.2275E+02, 1.2296E+02,
     +  1.2317E+02, 1.2338E+02, 1.2359E+02, 1.2381E+02, 1.2402E+02,
     + 1.2423E+02, 1.2444E+02, 1.2466E+02, 1.2487E+02, 1.2508E+02,
     +  1.2530E+02, 1.2551E+02, 1.2572E+02, 1.2594E+02, 1.2615E+02,
     + 1.2636E+02, 1.2658E+02, 1.2679E+02, 1.2700E+02, 1.2722E+02,
     +  1.2743E+02, 1.2765E+02, 1.2786E+02, 1.2807E+02, 1.2829E+02,
     + 1.2850E+02, 1.2872E+02, 1.2893E+02, 1.2915E+02, 1.2936E+02,
     +  1.2958E+02, 1.2979E+02, 1.3001E+02, 1.3022E+02, 1.3044E+02,
     + 1.3066E+02, 1.3087E+02, 1.3109E+02, 1.3130E+02, 1.3152E+02,
     +  1.3174E+02, 1.3195E+02, 1.3217E+02, 1.3239E+02, 1.3260E+02,
     + 1.3282E+02, 1.3304E+02, 1.3326E+02, 1.3347E+02, 1.3369E+02,
     +  1.3391E+02, 1.3413E+02, 1.3434E+02, 1.3456E+02, 1.3478E+02,
     + 1.3500E+02, 1.3522E+02, 1.3543E+02, 1.3565E+02, 1.3587E+02,
     +  1.3609E+02, 1.3631E+02, 1.3653E+02, 1.3675E+02, 1.3697E+02,
     + 1.3719E+02, 1.3741E+02, 1.3762E+02, 1.3784E+02, 1.3806E+02,
     +  1.3828E+02, 1.3850E+02, 1.3872E+02, 1.3895E+02, 1.3917E+02,
     + 1.3939E+02, 1.3961E+02, 1.3983E+02, 1.4005E+02, 1.4027E+02,
     +  1.4049E+02, 1.4071E+02, 1.4093E+02, 1.4116E+02, 1.4138E+02,
     + 1.4160E+02, 1.4182E+02, 1.4204E+02, 1.4227E+02, 1.4249E+02,
     +  1.4271E+02, 1.4293E+02, 1.4316E+02, 1.4338E+02, 1.4360E+02,
     + 1.4382E+02, 1.4405E+02, 1.4427E+02, 1.4450E+02, 1.4472E+02,
     +  1.4494E+02, 1.4517E+02, 1.4539E+02, 1.4561E+02, 1.4584E+02,
     + 1.4606E+02, 1.4629E+02, 1.4651E+02, 1.4674E+02, 1.4696E+02,
     +  1.4719E+02, 1.4741E+02, 1.4764E+02, 1.4786E+02, 1.4809E+02,
     + 1.4831E+02, 1.4854E+02, 1.4877E+02, 1.4899E+02, 1.4922E+02,
     +  1.4945E+02, 1.4967E+02, 1.4990E+02, 1.5013E+02, 1.5035E+02,
     + 1.5058E+02, 1.5081E+02, 1.5103E+02, 1.5126E+02, 1.5149E+02,
     +  1.5172E+02, 1.5194E+02, 1.5217E+02, 1.5240E+02, 1.5263E+02,
     + 1.5286E+02, 1.5309E+02, 1.5331E+02, 1.5354E+02, 1.5377E+02,
     +  1.5400E+02, 1.5423E+02, 1.5446E+02, 1.5469E+02, 1.5492E+02,
     + 1.5515E+02, 1.5538E+02, 1.5561E+02, 1.5584E+02, 1.5607E+02,
     +  1.5630E+02, 1.5653E+02, 1.5676E+02, 1.5699E+02, 1.5722E+02,
     + 1.5745E+02, 1.5769E+02, 1.5792E+02, 1.5815E+02, 1.5838E+02,
     +  1.5861E+02, 1.5884E+02, 1.5908E+02, 1.5931E+02, 1.5954E+02,
     + 1.5977E+02, 1.6001E+02, 1.6024E+02, 1.6047E+02, 1.6070E+02,
     +  1.6094E+02, 1.6117E+02, 1.6140E+02, 1.6164E+02, 1.6187E+02,
     + 1.6211E+02, 1.6234E+02, 1.6257E+02, 1.6281E+02, 1.6304E+02,
     +  1.6328E+02, 1.6351E+02, 1.6375E+02, 1.6398E+02, 1.6422E+02,
     + 1.6445E+02, 1.6469E+02, 1.6492E+02, 1.6516E+02, 1.6540E+02,
     +  1.6563E+02, 1.6587E+02, 1.6610E+02, 1.6634E+02, 1.6658E+02,
     + 1.6681E+02, 1.6705E+02, 1.6729E+02, 1.6752E+02, 1.6776E+02,
     +  1.6800E+02, 1.6824E+02, 1.6847E+02, 1.6871E+02, 1.6895E+02,
     + 1.6919E+02, 1.6943E+02, 1.6967E+02, 1.6990E+02, 1.7014E+02,
     +  1.7038E+02, 1.7062E+02, 1.7086E+02, 1.7110E+02, 1.7134E+02,
     + 1.7158E+02, 1.7182E+02, 1.7206E+02, 1.7230E+02, 1.7254E+02,
     +  1.7278E+02, 1.7302E+02, 1.7326E+02, 1.7350E+02, 1.7374E+02,
     + 1.7398E+02, 1.7423E+02, 1.7447E+02, 1.7471E+02, 1.7495E+02,
     +  1.7519E+02, 1.7543E+02, 1.7568E+02, 1.7592E+02, 1.7616E+02,
     + 1.7640E+02, 1.7665E+02, 1.7689E+02, 1.7713E+02, 1.7738E+02,
     +  1.7762E+02, 1.7786E+02, 1.7811E+02, 1.7835E+02, 1.7859E+02,
     + 1.7884E+02, 1.7908E+02, 1.7933E+02, 1.7957E+02, 1.7982E+02,
     +  1.8006E+02, 1.8031E+02, 1.8055E+02, 1.8080E+02, 1.8104E+02,
     + 1.8129E+02, 1.8153E+02, 1.8178E+02, 1.8203E+02, 1.8227E+02,
     +  1.8252E+02, 1.8276E+02, 1.8301E+02, 1.8326E+02, 1.8351E+02,
     + 1.8375E+02, 1.8400E+02, 1.8425E+02, 1.8449E+02, 1.8474E+02,
     +  1.8499E+02, 1.8524E+02, 1.8549E+02, 1.8574E+02, 1.8598E+02,
     + 1.8623E+02, 1.8648E+02, 1.8673E+02, 1.8698E+02, 1.8723E+02,
     +  1.8748E+02, 1.8773E+02, 1.8798E+02, 1.8823E+02, 1.8848E+02,
     + 1.8873E+02, 1.8898E+02, 1.8923E+02, 1.8948E+02, 1.8973E+02,
     +  1.8998E+02, 1.9023E+02, 1.9048E+02, 1.9074E+02, 1.9099E+02,
     + 1.9124E+02, 1.9149E+02, 1.9174E+02, 1.9199E+02, 1.9225E+02,
     +  1.9250E+02, 1.9275E+02, 1.9301E+02, 1.9326E+02, 1.9351E+02,
     + 1.9377E+02, 1.9402E+02, 1.9427E+02, 1.9453E+02, 1.9478E+02,
     +  1.9503E+02, 1.9529E+02, 1.9554E+02, 1.9580E+02, 1.9605E+02,
     + 1.9631E+02, 1.9656E+02, 1.9682E+02, 1.9707E+02, 1.9733E+02,
     +  1.9758E+02, 1.9784E+02, 1.9810E+02, 1.9835E+02, 1.9861E+02,
     + 1.9887E+02, 1.9912E+02, 1.9938E+02, 1.9964E+02, 1.9989E+02,
     +  2.0015E+02, 2.0041E+02, 2.0067E+02, 2.0092E+02, 2.0118E+02,
     + 2.0144E+02, 2.0170E+02, 2.0196E+02, 2.0222E+02, 2.0247E+02,
     +  2.0273E+02, 2.0299E+02, 2.0325E+02, 2.0351E+02, 2.0377E+02,
     + 2.0403E+02, 2.0429E+02, 2.0455E+02, 2.0481E+02, 2.0507E+02,
     +  2.0533E+02, 2.0559E+02, 2.0585E+02, 2.0611E+02, 2.0638E+02,
     + 2.0664E+02, 2.0690E+02, 2.0716E+02, 2.0742E+02, 2.0768E+02,
     +  2.0795E+02, 2.0821E+02, 2.0847E+02, 2.0873E+02, 2.0900E+02,
     + 2.0926E+02, 2.0952E+02, 2.0979E+02, 2.1005E+02, 2.1031E+02,
     +  2.1058E+02, 2.1084E+02, 2.1110E+02, 2.1137E+02, 2.1163E+02,
     + 2.1190E+02, 2.1216E+02, 2.1243E+02, 2.1269E+02, 2.1296E+02,
     +  2.1322E+02, 2.1349E+02, 2.1376E+02, 2.1402E+02, 2.1429E+02,
     + 2.1455E+02, 2.1482E+02, 2.1509E+02, 2.1535E+02, 2.1562E+02,
     +  2.1589E+02, 2.1615E+02, 2.1642E+02, 2.1669E+02, 2.1696E+02,
     + 2.1723E+02, 2.1749E+02, 2.1776E+02, 2.1803E+02, 2.1830E+02,
     +  2.1857E+02, 2.1884E+02, 2.1911E+02, 2.1938E+02, 2.1964E+02,
     + 2.1991E+02, 2.2018E+02, 2.2045E+02, 2.2072E+02, 2.2099E+02,
     +  2.2127E+02, 2.2154E+02, 2.2181E+02, 2.2208E+02, 2.2235E+02,
     + 2.2262E+02, 2.2289E+02, 2.2316E+02, 2.2343E+02, 2.2371E+02,
     +  2.2398E+02, 2.2425E+02, 2.2452E+02, 2.2480E+02, 2.2507E+02,
     + 2.2534E+02, 2.2562E+02, 2.2589E+02, 2.2616E+02, 2.2644E+02,
     +  2.2671E+02, 2.2698E+02, 2.2726E+02, 2.2753E+02, 2.2781E+02,
     + 2.2808E+02, 2.2836E+02, 2.2863E+02, 2.2891E+02, 2.2918E+02,
     +  2.2946E+02, 2.2973E+02, 2.3001E+02, 2.3028E+02, 2.3056E+02,
     + 2.3084E+02, 2.3111E+02, 2.3139E+02, 2.3167E+02, 2.3194E+02,
     +  2.3222E+02, 2.3250E+02, 2.3278E+02, 2.3305E+02, 2.3333E+02,
     + 2.3361E+02, 2.3389E+02, 2.3417E+02, 2.3445E+02, 2.3472E+02,
     +  2.3500E+02, 2.3528E+02, 2.3556E+02, 2.3584E+02, 2.3612E+02,
     + 2.3640E+02, 2.3668E+02, 2.3696E+02, 2.3724E+02, 2.3752E+02,
     +  2.3780E+02, 2.3808E+02, 2.3836E+02, 2.3865E+02, 2.3893E+02,
     + 2.3921E+02, 2.3949E+02, 2.3977E+02, 2.4005E+02, 2.4034E+02,
     +  2.4062E+02, 2.4090E+02, 2.4118E+02, 2.4147E+02, 2.4175E+02,
     + 2.4203E+02, 2.4232E+02, 2.4260E+02, 2.4288E+02, 2.4317E+02,
     +  2.4345E+02, 2.4374E+02, 2.4402E+02, 2.4431E+02, 2.4459E+02,
     + 2.4488E+02, 2.4516E+02, 2.4545E+02, 2.4573E+02, 2.4602E+02,
     +  2.4630E+02, 2.4659E+02, 2.4688E+02, 2.4716E+02, 2.4745E+02,
     + 2.4774E+02, 2.4802E+02, 2.4831E+02, 2.4860E+02, 2.4888E+02,
     +  2.4917E+02, 2.4946E+02, 2.4975E+02, 2.5004E+02, 2.5033E+02,
     + 2.5061E+02, 2.5090E+02, 2.5119E+02, 2.5148E+02, 2.5177E+02,
     +  2.5206E+02, 2.5235E+02, 2.5264E+02, 2.5293E+02, 2.5322E+02,
     + 2.5351E+02, 2.5380E+02, 2.5409E+02, 2.5438E+02, 2.5467E+02,
     +  2.5496E+02, 2.5525E+02, 2.5555E+02, 2.5584E+02, 2.5613E+02,
     + 2.5642E+02, 2.5671E+02, 2.5701E+02, 2.5730E+02, 2.5759E+02,
     +  2.5788E+02, 2.5818E+02, 2.5847E+02, 2.5876E+02, 2.5906E+02,
     + 2.5935E+02, 2.5965E+02, 2.5994E+02, 2.6023E+02, 2.6053E+02,
     +  2.6082E+02, 2.6112E+02, 2.6141E+02, 2.6171E+02, 2.6201E+02,
     + 2.6230E+02, 2.6260E+02, 2.6289E+02, 2.6319E+02, 2.6349E+02,
     +  2.6378E+02, 2.6408E+02, 2.6438E+02, 2.6467E+02, 2.6497E+02,
     + 2.6527E+02, 2.6556E+02, 2.6586E+02, 2.6616E+02, 2.6646E+02,
     +  2.6676E+02, 2.6706E+02, 2.6735E+02, 2.6765E+02, 2.6795E+02,
     + 2.6825E+02, 2.6855E+02, 2.6885E+02, 2.6915E+02, 2.6945E+02,
     +  2.6975E+02, 2.7005E+02, 2.7035E+02, 2.7065E+02, 2.7095E+02,
     + 2.7125E+02, 2.7156E+02, 2.7186E+02, 2.7216E+02, 2.7246E+02,
     +  2.7276E+02, 2.7306E+02, 2.7337E+02, 2.7367E+02, 2.7397E+02,
     + 2.7428E+02, 2.7458E+02, 2.7488E+02, 2.7518E+02, 2.7549E+02,
     +  2.7579E+02, 2.7610E+02, 2.7640E+02, 2.7670E+02, 2.7701E+02,
     + 2.7731E+02, 2.7762E+02, 2.7792E+02, 2.7823E+02, 2.7853E+02,
     +  2.7884E+02, 2.7915E+02, 2.7945E+02, 2.7976E+02, 2.8006E+02,
     + 2.8037E+02, 2.8068E+02, 2.8099E+02, 2.8129E+02, 2.8160E+02,
     +  2.8191E+02, 2.8221E+02, 2.8252E+02, 2.8283E+02, 2.8314E+02,
     + 2.8345E+02, 2.8376E+02, 2.8406E+02, 2.8437E+02, 2.8468E+02,
     +  2.8499E+02, 2.8530E+02, 2.8561E+02, 2.8592E+02, 2.8623E+02,
     + 2.8654E+02, 2.8685E+02, 2.8716E+02, 2.8747E+02, 2.8778E+02,
     +  2.8810E+02, 2.8841E+02, 2.8872E+02, 2.8903E+02, 2.8934E+02,
     + 2.8965E+02, 2.8997E+02, 2.9028E+02, 2.9059E+02, 2.9090E+02,
     +  2.9122E+02, 2.9153E+02, 2.9184E+02, 2.9216E+02, 2.9247E+02,
     + 2.9279E+02, 2.9310E+02, 2.9341E+02, 2.9373E+02, 2.9404E+02,
     +  2.9436E+02, 2.9467E+02, 2.9499E+02, 2.9530E+02, 2.9562E+02,
     + 2.9594E+02, 2.9625E+02, 2.9657E+02, 2.9688E+02, 2.9720E+02,
     +  2.9752E+02, 2.9783E+02, 2.9815E+02, 2.9847E+02, 2.9879E+02,
     + 2.9910E+02, 2.9942E+02, 2.9974E+02, 3.0006E+02, 3.0038E+02,
     +  3.0070E+02, 3.0101E+02, 3.0133E+02, 3.0165E+02, 3.0197E+02,
     + 3.0229E+02, 3.0261E+02, 3.0293E+02, 3.0325E+02, 3.0357E+02,
     +  3.0389E+02, 3.0421E+02, 3.0453E+02, 3.0486E+02, 3.0518E+02,
     + 3.0550E+02, 3.0582E+02, 3.0614E+02, 3.0646E+02, 3.0679E+02,
     +  3.0711E+02, 3.0743E+02, 3.0775E+02, 3.0808E+02, 3.0840E+02,
     + 3.0872E+02, 3.0905E+02, 3.0937E+02, 3.0969E+02, 3.1002E+02,
     +  3.1034E+02, 3.1067E+02, 3.1099E+02, 3.1132E+02, 3.1164E+02,
     + 3.1197E+02, 3.1229E+02, 3.1262E+02, 3.1294E+02, 3.1327E+02,
     +  3.1360E+02, 3.1392E+02, 3.1425E+02, 3.1458E+02, 3.1490E+02,
     + 3.1523E+02, 3.1556E+02, 3.1589E+02, 3.1621E+02, 3.1654E+02,
     +  3.1687E+02, 3.1720E+02, 3.1753E+02, 3.1786E+02, 3.1818E+02,
     + 3.1851E+02, 3.1884E+02, 3.1917E+02, 3.1950E+02, 3.1983E+02,
     +  3.2016E+02, 3.2049E+02, 3.2082E+02, 3.2115E+02, 3.2148E+02,
     + 3.2181E+02, 3.2215E+02, 3.2248E+02, 3.2281E+02, 3.2314E+02,
     +  3.2347E+02, 3.2380E+02, 3.2414E+02, 3.2447E+02, 3.2480E+02,
     + 3.2514E+02, 3.2547E+02, 3.2580E+02, 3.2614E+02, 3.2647E+02,
     +  3.2680E+02, 3.2714E+02, 3.2747E+02, 3.2781E+02, 3.2814E+02,
     + 3.2848E+02, 3.2881E+02, 3.2915E+02, 3.2948E+02, 3.2982E+02,
     +  3.3015E+02, 3.3049E+02, 3.3083E+02, 3.3116E+02, 3.3150E+02,
     + 3.3184E+02, 3.3217E+02, 3.3251E+02, 3.3285E+02, 3.3318E+02,
     +  3.3352E+02, 3.3386E+02, 3.3420E+02, 3.3454E+02, 3.3488E+02,
     + 3.3521E+02, 3.3555E+02, 3.3589E+02, 3.3623E+02, 3.3657E+02,
     +  3.3691E+02, 3.3725E+02, 3.3759E+02, 3.3793E+02, 3.3827E+02,
     + 3.3861E+02, 3.3895E+02, 3.3930E+02, 3.3964E+02, 3.3998E+02,
     +  3.4032E+02, 3.4066E+02, 3.4100E+02, 3.4135E+02, 3.4169E+02,
     + 3.4203E+02, 3.4237E+02, 3.4272E+02, 3.4306E+02, 3.4340E+02,
     +  3.4375E+02, 3.4409E+02, 3.4444E+02, 3.4478E+02, 3.4512E+02,
     + 3.4547E+02, 3.4581E+02, 3.4616E+02, 3.4650E+02, 3.4685E+02,
     +  3.4720E+02, 3.4754E+02, 3.4789E+02, 3.4823E+02, 3.4858E+02,
     + 3.4893E+02, 3.4927E+02, 3.4962E+02, 3.4997E+02, 3.5031E+02,
     +  3.5066E+02, 3.5101E+02, 3.5136E+02, 3.5171E+02, 3.5206E+02,
     + 3.5240E+02, 3.5275E+02, 3.5310E+02, 3.5345E+02, 3.5380E+02,
     +  3.5415E+02, 3.5450E+02, 3.5485E+02, 3.5520E+02, 3.5555E+02,
     + 3.5590E+02, 3.5625E+02, 3.5660E+02, 3.5695E+02, 3.5731E+02,
     +  3.5766E+02, 3.5801E+02, 3.5836E+02, 3.5871E+02, 3.5907E+02,
     + 3.5942E+02, 3.5977E+02, 3.6012E+02, 3.6048E+02, 3.6083E+02,
     +  3.6118E+02, 3.6154E+02, 3.6189E+02, 3.6225E+02, 3.6260E+02,
     + 3.6295E+02, 3.6331E+02, 3.6366E+02, 3.6402E+02, 3.6438E+02,
     +  3.6473E+02, 3.6509E+02, 3.6544E+02, 3.6580E+02, 3.6616E+02,
     + 3.6651E+02, 3.6687E+02, 3.6723E+02, 3.6758E+02, 3.6794E+02,
     +  3.6830E+02, 3.6866E+02, 3.6901E+02, 3.6937E+02, 3.6973E+02,
     + 3.7009E+02, 3.7045E+02, 3.7081E+02, 3.7117E+02, 3.7153E+02,
     +  3.7188E+02, 3.7224E+02, 3.7260E+02, 3.7296E+02, 3.7333E+02,
     + 3.7369E+02, 3.7405E+02, 3.7441E+02, 3.7477E+02, 3.7513E+02,
     +  3.7549E+02, 3.7585E+02, 3.7622E+02, 3.7658E+02, 3.7694E+02,
     + 3.7730E+02, 3.7767E+02, 3.7803E+02, 3.7839E+02, 3.7876E+02,
     +  3.7912E+02, 3.7948E+02, 3.7985E+02, 3.8021E+02, 3.8058E+02,
     + 3.8094E+02, 3.8131E+02, 3.8167E+02, 3.8204E+02, 3.8240E+02,
     +  3.8277E+02, 3.8313E+02, 3.8350E+02, 3.8387E+02, 3.8423E+02,
     + 3.8460E+02, 3.8497E+02, 3.8533E+02, 3.8570E+02, 3.8607E+02,
     +  3.8644E+02, 3.8680E+02, 3.8717E+02, 3.8754E+02, 3.8791E+02,
     + 3.8828E+02, 3.8865E+02, 3.8902E+02, 3.8939E+02, 3.8975E+02,
     +  3.9012E+02, 3.9049E+02, 3.9086E+02, 3.9124E+02, 3.9161E+02,
     + 3.9198E+02, 3.9235E+02, 3.9272E+02, 3.9309E+02, 3.9346E+02,
     +  3.9383E+02, 3.9421E+02, 3.9458E+02, 3.9495E+02, 3.9532E+02,
     + 3.9570E+02, 3.9607E+02, 3.9644E+02, 3.9682E+02, 3.9719E+02,
     +  3.9756E+02, 3.9794E+02, 3.9831E+02, 3.9869E+02, 3.9906E+02,
     + 3.9944E+02, 3.9981E+02, 4.0019E+02, 4.0056E+02, 4.0094E+02,
     +  4.0131E+02, 4.0169E+02, 4.0207E+02, 4.0244E+02, 4.0282E+02,
     + 4.0320E+02, 4.0357E+02, 4.0395E+02, 4.0433E+02, 4.0471E+02,
     +  4.0508E+02, 4.0546E+02, 4.0584E+02, 4.0622E+02, 4.0660E+02,
     + 4.0698E+02, 4.0736E+02, 4.0774E+02, 4.0812E+02, 4.0850E+02,
     +  4.0888E+02, 4.0926E+02, 4.0964E+02, 4.1002E+02, 4.1040E+02,
     + 4.1078E+02, 4.1116E+02, 4.1154E+02, 4.1193E+02, 4.1231E+02,
     +  4.1269E+02, 4.1307E+02, 4.1345E+02, 4.1384E+02, 4.1422E+02,
     + 4.1460E+02, 4.1499E+02, 4.1537E+02, 4.1575E+02, 4.1614E+02,
     +  4.1652E+02, 4.1691E+02, 4.1729E+02, 4.1768E+02, 4.1806E+02,
     + 4.1845E+02, 4.1883E+02, 4.1922E+02, 4.1960E+02, 4.1999E+02,
     +  4.2038E+02, 4.2076E+02, 4.2115E+02, 4.2154E+02, 4.2192E+02,
     + 4.2231E+02, 4.2270E+02, 4.2309E+02, 4.2348E+02, 4.2386E+02,
     +  4.2425E+02, 4.2464E+02, 4.2503E+02, 4.2542E+02, 4.2581E+02,
     + 4.2620E+02, 4.2659E+02, 4.2698E+02, 4.2737E+02, 4.2776E+02,
     +  4.2815E+02, 4.2854E+02, 4.2893E+02, 4.2932E+02, 4.2971E+02,
     + 4.3011E+02, 4.3050E+02, 4.3089E+02, 4.3128E+02, 4.3167E+02,
     +  4.3207E+02, 4.3246E+02, 4.3285E+02, 4.3325E+02, 4.3364E+02,
     + 4.3403E+02, 4.3443E+02, 4.3482E+02, 4.3522E+02, 4.3561E+02,
     +  4.3601E+02, 4.3640E+02, 4.3680E+02, 4.3719E+02, 4.3759E+02,
     + 4.3798E+02, 4.3838E+02, 4.3878E+02, 4.3917E+02, 4.3957E+02,
     +  4.3997E+02, 4.4036E+02, 4.4076E+02, 4.4116E+02, 4.4156E+02,
     + 4.4195E+02, 4.4235E+02, 4.4275E+02, 4.4315E+02, 4.4355E+02,
     +  4.4395E+02, 4.4435E+02, 4.4475E+02, 4.4515E+02, 4.4555E+02,
     + 4.4595E+02, 4.4635E+02, 4.4675E+02, 4.4715E+02, 4.4755E+02,
     +  4.4795E+02, 4.4835E+02, 4.4875E+02, 4.4916E+02, 4.4956E+02,
     + 4.4996E+02, 4.5036E+02, 4.5076E+02, 4.5117E+02, 4.5157E+02,
     +  4.5197E+02, 4.5238E+02, 4.5278E+02, 4.5319E+02, 4.5359E+02,
     + 4.5399E+02, 4.5440E+02, 4.5480E+02, 4.5521E+02, 4.5561E+02,
     +  4.5602E+02, 4.5643E+02, 4.5683E+02, 4.5724E+02, 4.5764E+02,
     + 4.5805E+02, 4.5846E+02, 4.5886E+02, 4.5927E+02, 4.5968E+02,
     +  4.6009E+02, 4.6050E+02, 4.6090E+02, 4.6131E+02, 4.6172E+02,
     + 4.6213E+02, 4.6254E+02, 4.6295E+02, 4.6336E+02, 4.6377E+02,
     +  4.6418E+02, 4.6459E+02, 4.6500E+02, 4.6541E+02, 4.6582E+02,
     + 4.6623E+02, 4.6664E+02, 4.6705E+02, 4.6746E+02, 4.6788E+02,
     +  4.6829E+02, 4.6870E+02, 4.6911E+02, 4.6952E+02, 4.6994E+02,
     + 4.7035E+02, 4.7076E+02, 4.7118E+02, 4.7159E+02, 4.7201E+02,
     +  4.7242E+02, 4.7283E+02, 4.7325E+02, 4.7366E+02, 4.7408E+02,
     + 4.7449E+02, 4.7491E+02, 4.7533E+02, 4.7574E+02, 4.7616E+02,
     +  4.7657E+02, 4.7699E+02, 4.7741E+02, 4.7782E+02, 4.7824E+02,
     + 4.7866E+02, 4.7908E+02, 4.7949E+02, 4.7991E+02, 4.8033E+02,
     +  4.8075E+02, 4.8117E+02, 4.8159E+02, 4.8201E+02, 4.8243E+02,
     + 4.8285E+02, 4.8327E+02, 4.8369E+02, 4.8411E+02, 4.8453E+02,
     +  4.8495E+02, 4.8537E+02, 4.8579E+02, 4.8621E+02, 4.8663E+02,
     + 4.8706E+02, 4.8748E+02, 4.8790E+02, 4.8832E+02, 4.8875E+02,
     +  4.8917E+02, 4.8959E+02, 4.9002E+02, 4.9044E+02, 4.9086E+02,
     + 4.9129E+02, 4.9171E+02, 4.9214E+02, 4.9256E+02, 4.9299E+02,
     +  4.9341E+02, 4.9384E+02, 4.9426E+02, 4.9469E+02, 4.9511E+02,
     + 4.9554E+02, 4.9597E+02, 4.9639E+02, 4.9682E+02, 4.9725E+02,
     +  4.9768E+02, 4.9810E+02, 4.9853E+02, 4.9896E+02, 4.9939E+02,
     + 4.9982E+02, 5.0024E+02, 5.0067E+02, 5.0110E+02, 5.0153E+02,
     +  5.0196E+02, 5.0239E+02, 5.0282E+02, 5.0325E+02, 5.0368E+02,
     + 5.0411E+02, 5.0454E+02, 5.0498E+02, 5.0541E+02, 5.0584E+02,
     +  5.0627E+02, 5.0670E+02, 5.0714E+02, 5.0757E+02, 5.0800E+02,
     + 5.0843E+02, 5.0887E+02, 5.0930E+02, 5.0973E+02, 5.1017E+02,
     +  5.1060E+02, 5.1104E+02, 5.1147E+02, 5.1191E+02, 5.1234E+02,
     + 5.1278E+02, 5.1321E+02, 5.1365E+02, 5.1408E+02, 5.1452E+02,
     +  5.1496E+02, 5.1539E+02, 5.1583E+02, 5.1627E+02, 5.1670E+02,
     + 5.1714E+02, 5.1758E+02, 5.1802E+02, 5.1845E+02, 5.1889E+02,
     +  5.1933E+02, 5.1977E+02, 5.2021E+02, 5.2065E+02, 5.2109E+02,
     + 5.2153E+02, 5.2197E+02, 5.2241E+02, 5.2285E+02, 5.2329E+02,
     +  5.2373E+02, 5.2417E+02, 5.2461E+02, 5.2505E+02, 5.2550E+02,
     + 5.2594E+02, 5.2638E+02, 5.2682E+02, 5.2727E+02, 5.2771E+02,
     +  5.2815E+02, 5.2860E+02, 5.2904E+02, 5.2948E+02, 5.2993E+02,
     + 5.3037E+02, 5.3082E+02, 5.3126E+02, 5.3171E+02, 5.3215E+02,
     +  5.3260E+02, 5.3304E+02, 5.3349E+02, 5.3393E+02, 5.3438E+02,
     + 5.3483E+02, 5.3527E+02, 5.3572E+02, 5.3617E+02, 5.3661E+02,
     +  5.3706E+02, 5.3751E+02, 5.3796E+02, 5.3841E+02, 5.3886E+02,
     + 5.3930E+02, 5.3975E+02, 5.4020E+02, 5.4065E+02, 5.4110E+02,
     +  5.4155E+02, 5.4200E+02, 5.4245E+02, 5.4290E+02, 5.4335E+02,
     + 5.4381E+02, 5.4426E+02, 5.4471E+02, 5.4516E+02, 5.4561E+02,
     +  5.4606E+02, 5.4652E+02, 5.4697E+02, 5.4742E+02, 5.4788E+02,
     + 5.4833E+02, 5.4878E+02, 5.4924E+02, 5.4969E+02, 5.5015E+02,
     +  5.5060E+02, 5.5105E+02, 5.5151E+02, 5.5197E+02, 5.5242E+02,
     + 5.5288E+02, 5.5333E+02, 5.5379E+02, 5.5424E+02, 5.5470E+02,
     +  5.5516E+02, 5.5562E+02, 5.5607E+02, 5.5653E+02, 5.5699E+02,
     + 5.5745E+02, 5.5790E+02, 5.5836E+02, 5.5882E+02, 5.5928E+02,
     +  5.5974E+02, 5.6020E+02, 5.6066E+02, 5.6112E+02, 5.6158E+02,
     + 5.6204E+02, 5.6250E+02, 5.6296E+02, 5.6342E+02, 5.6388E+02,
     +  5.6434E+02, 5.6481E+02, 5.6527E+02, 5.6573E+02, 5.6619E+02,
     + 5.6666E+02, 5.6712E+02, 5.6758E+02, 5.6804E+02, 5.6851E+02,
     +  5.6897E+02, 5.6944E+02, 5.6990E+02, 5.7037E+02, 5.7083E+02,
     + 5.7129E+02, 5.7176E+02, 5.7223E+02, 5.7269E+02, 5.7316E+02,
     +  5.7362E+02, 5.7409E+02, 5.7456E+02, 5.7502E+02, 5.7549E+02,
     + 5.7596E+02, 5.7643E+02, 5.7689E+02, 5.7736E+02, 5.7783E+02,
     +  5.7830E+02, 5.7877E+02, 5.7924E+02, 5.7970E+02, 5.8017E+02,
     + 5.8064E+02, 5.8111E+02, 5.8158E+02, 5.8205E+02, 5.8252E+02,
     +  5.8300E+02, 5.8347E+02, 5.8394E+02, 5.8441E+02, 5.8488E+02,
     + 5.8535E+02, 5.8583E+02, 5.8630E+02, 5.8677E+02, 5.8724E+02,
     +  5.8772E+02, 5.8819E+02, 5.8866E+02, 5.8914E+02, 5.8961E+02,
     + 5.9009E+02, 5.9056E+02, 5.9104E+02, 5.9151E+02, 5.9199E+02,
     +  5.9246E+02, 5.9294E+02, 5.9341E+02, 5.9389E+02, 5.9437E+02,
     + 5.9484E+02, 5.9532E+02, 5.9580E+02, 5.9627E+02, 5.9675E+02,
     +  5.9723E+02, 5.9771E+02, 5.9819E+02, 5.9867E+02, 5.9914E+02,
     + 5.9962E+02, 6.0010E+02, 6.0058E+02, 6.0106E+02, 6.0154E+02,
     +  6.0202E+02, 6.0250E+02, 6.0298E+02, 6.0346E+02, 6.0395E+02,
     + 6.0443E+02, 6.0491E+02, 6.0539E+02, 6.0587E+02, 6.0636E+02,
     +  6.0684E+02, 6.0732E+02, 6.0780E+02, 6.0829E+02, 6.0877E+02,
     + 6.0926E+02, 6.0974E+02, 6.1022E+02, 6.1071E+02, 6.1119E+02,
     +  6.1168E+02, 6.1216E+02, 6.1265E+02, 6.1313E+02, 6.1362E+02,
     + 6.1411E+02, 6.1459E+02, 6.1508E+02, 6.1557E+02, 6.1605E+02,
     +  6.1654E+02, 6.1703E+02, 6.1752E+02, 6.1801E+02, 6.1849E+02,
     + 6.1898E+02, 6.1947E+02, 6.1996E+02, 6.2045E+02, 6.2094E+02,
     +  6.2143E+02, 6.2192E+02, 6.2241E+02, 6.2290E+02, 6.2339E+02,
     + 6.2388E+02, 6.2437E+02, 6.2486E+02, 6.2536E+02, 6.2585E+02,
     +  6.2634E+02, 6.2683E+02, 6.2733E+02, 6.2782E+02, 6.2831E+02,
     + 6.2880E+02, 6.2930E+02, 6.2979E+02, 6.3029E+02, 6.3078E+02,
     +  6.3128E+02, 6.3177E+02, 6.3227E+02, 6.3276E+02, 6.3326E+02,
     + 6.3375E+02, 6.3425E+02, 6.3474E+02, 6.3524E+02, 6.3574E+02,
     +  6.3624E+02, 6.3673E+02, 6.3723E+02, 6.3773E+02, 6.3823E+02,
     + 6.3872E+02, 6.3922E+02, 6.3972E+02, 6.4022E+02, 6.4072E+02,
     +  6.4122E+02, 6.4172E+02, 6.4222E+02, 6.4272E+02, 6.4322E+02,
     + 6.4372E+02, 6.4422E+02, 6.4472E+02, 6.4522E+02, 6.4573E+02,
     +  6.4623E+02, 6.4673E+02, 6.4723E+02, 6.4773E+02, 6.4824E+02,
     + 6.4874E+02, 6.4924E+02, 6.4975E+02, 6.5025E+02, 6.5076E+02,
     +  6.5126E+02, 6.5176E+02, 6.5227E+02, 6.5277E+02, 6.5328E+02,
     + 6.5379E+02, 6.5429E+02, 6.5480E+02, 6.5530E+02, 6.5581E+02,
     +  6.5632E+02, 6.5682E+02, 6.5733E+02, 6.5784E+02, 6.5835E+02,
     + 6.5885E+02, 6.5936E+02, 6.5987E+02, 6.6038E+02, 6.6089E+02,
     +  6.6140E+02, 6.6191E+02, 6.6242E+02, 6.6293E+02, 6.6344E+02,
     + 6.6395E+02, 6.6446E+02, 6.6497E+02, 6.6548E+02, 6.6599E+02,
     +  6.6650E+02, 6.6702E+02, 6.6753E+02, 6.6804E+02, 6.6855E+02,
     + 6.6907E+02, 6.6958E+02, 6.7009E+02, 6.7061E+02, 6.7112E+02,
     +  6.7163E+02, 6.7215E+02, 6.7266E+02, 6.7318E+02, 6.7369E+02,
     + 6.7421E+02, 6.7472E+02, 6.7524E+02, 6.7576E+02, 6.7627E+02,
     +  6.7679E+02, 6.7731E+02, 6.7782E+02, 6.7834E+02, 6.7886E+02,
     + 6.7938E+02, 6.7989E+02, 6.8041E+02, 6.8093E+02, 6.8145E+02,
     +  6.8197E+02, 6.8249E+02, 6.8301E+02, 6.8353E+02, 6.8405E+02,
     + 6.8457E+02, 6.8509E+02, 6.8561E+02, 6.8613E+02, 6.8665E+02,
     +  6.8717E+02, 6.8770E+02, 6.8822E+02, 6.8874E+02, 6.8926E+02,
     + 6.8979E+02, 6.9031E+02, 6.9083E+02, 6.9136E+02, 6.9188E+02,
     +  6.9240E+02, 6.9293E+02, 6.9345E+02, 6.9398E+02, 6.9450E+02,
     + 6.9503E+02, 6.9555E+02, 6.9608E+02, 6.9660E+02, 6.9713E+02,
     +  6.9766E+02, 6.9818E+02, 6.9871E+02, 6.9924E+02, 6.9977E+02,
     + 7.0029E+02, 7.0082E+02, 7.0135E+02, 7.0188E+02, 7.0241E+02,
     +  7.0294E+02, 7.0347E+02, 7.0400E+02, 7.0453E+02, 7.0506E+02,
     + 7.0559E+02, 7.0612E+02, 7.0665E+02, 7.0718E+02, 7.0771E+02,
     +  7.0824E+02, 7.0877E+02, 7.0931E+02, 7.0984E+02, 7.1037E+02,
     + 7.1090E+02, 7.1144E+02, 7.1197E+02, 7.1250E+02, 7.1304E+02,
     +  7.1357E+02, 7.1411E+02, 7.1464E+02, 7.1518E+02, 7.1571E+02,
     + 7.1625E+02, 7.1678E+02, 7.1732E+02, 7.1785E+02, 7.1839E+02,
     +  7.1893E+02, 7.1946E+02, 7.2000E+02, 7.2054E+02, 7.2108E+02,
     + 7.2161E+02, 7.2215E+02, 7.2269E+02, 7.2323E+02, 7.2377E+02,
     +  7.2431E+02, 7.2485E+02, 7.2539E+02, 7.2593E+02, 7.2647E+02,
     + 7.2701E+02, 7.2755E+02, 7.2809E+02, 7.2863E+02, 7.2917E+02,
     +  7.2971E+02, 7.3025E+02, 7.3080E+02, 7.3134E+02, 7.3188E+02,
     + 7.3242E+02, 7.3297E+02, 7.3351E+02, 7.3405E+02, 7.3460E+02,
     +  7.3514E+02, 7.3569E+02, 7.3623E+02, 7.3678E+02, 7.3732E+02,
     + 7.3787E+02, 7.3841E+02, 7.3896E+02, 7.3951E+02, 7.4005E+02,
     +  7.4060E+02, 7.4115E+02, 7.4169E+02, 7.4224E+02, 7.4279E+02,
     + 7.4334E+02, 7.4388E+02, 7.4443E+02, 7.4498E+02, 7.4553E+02,
     +  7.4608E+02, 7.4663E+02, 7.4718E+02, 7.4773E+02, 7.4828E+02,
     + 7.4883E+02, 7.4938E+02, 7.4993E+02, 7.5048E+02, 7.5104E+02,
     +  7.5159E+02, 7.5214E+02, 7.5269E+02, 7.5324E+02, 7.5380E+02,
     + 7.5435E+02, 7.5490E+02, 7.5546E+02, 7.5601E+02, 7.5657E+02,
     +  7.5712E+02, 7.5768E+02, 7.5823E+02, 7.5879E+02, 7.5934E+02,
     + 7.5990E+02, 7.6045E+02, 7.6101E+02, 7.6157E+02, 7.6212E+02,
     +  7.6268E+02, 7.6324E+02, 7.6379E+02, 7.6435E+02, 7.6491E+02,
     + 7.6547E+02, 7.6603E+02, 7.6659E+02, 7.6714E+02, 7.6770E+02,
     +  7.6826E+02, 7.6882E+02, 7.6938E+02, 7.6994E+02, 7.7050E+02,
     + 7.7107E+02, 7.7163E+02, 7.7219E+02, 7.7275E+02, 7.7331E+02,
     +  7.7387E+02, 7.7444E+02, 7.7500E+02, 7.7556E+02, 7.7613E+02/
      data (QofT( 2,J),J=2409,4510)/ 7.7669E+02, 7.7725E+02,
     +  7.7782E+02, 7.7838E+02, 7.7895E+02, 7.7951E+02, 7.8007E+02,
     +  7.8064E+02,
     + 7.8121E+02, 7.8177E+02, 7.8234E+02, 7.8290E+02, 7.8347E+02,
     +  7.8404E+02, 7.8460E+02, 7.8517E+02, 7.8574E+02, 7.8631E+02,
     + 7.8688E+02, 7.8744E+02, 7.8801E+02, 7.8858E+02, 7.8915E+02,
     +  7.8972E+02, 7.9029E+02, 7.9086E+02, 7.9143E+02, 7.9200E+02,
     + 7.9257E+02, 7.9314E+02, 7.9371E+02, 7.9429E+02, 7.9486E+02,
     +  7.9543E+02, 7.9600E+02, 7.9657E+02, 7.9715E+02, 7.9772E+02,
     + 7.9829E+02, 7.9887E+02, 7.9944E+02, 8.0002E+02, 8.0059E+02,
     +  8.0116E+02, 8.0174E+02, 8.0231E+02, 8.0289E+02, 8.0347E+02,
     + 8.0404E+02, 8.0462E+02, 8.0519E+02, 8.0577E+02, 8.0635E+02,
     +  8.0693E+02, 8.0750E+02, 8.0808E+02, 8.0866E+02, 8.0924E+02,
     + 8.0982E+02, 8.1040E+02, 8.1097E+02, 8.1155E+02, 8.1213E+02,
     +  8.1271E+02, 8.1329E+02, 8.1387E+02, 8.1445E+02, 8.1504E+02,
     + 8.1562E+02, 8.1620E+02, 8.1678E+02, 8.1736E+02, 8.1794E+02,
     +  8.1853E+02, 8.1911E+02, 8.1969E+02, 8.2028E+02, 8.2086E+02,
     + 8.2144E+02, 8.2203E+02, 8.2261E+02, 8.2320E+02, 8.2378E+02,
     +  8.2437E+02, 8.2495E+02, 8.2554E+02, 8.2613E+02, 8.2671E+02,
     + 8.2730E+02, 8.2788E+02, 8.2847E+02, 8.2906E+02, 8.2965E+02,
     +  8.3023E+02, 8.3082E+02, 8.3141E+02, 8.3200E+02, 8.3259E+02,
     + 8.3318E+02, 8.3377E+02, 8.3436E+02, 8.3495E+02, 8.3554E+02,
     +  8.3613E+02, 8.3672E+02, 8.3731E+02, 8.3790E+02, 8.3849E+02,
     + 8.3908E+02, 8.3968E+02, 8.4027E+02, 8.4086E+02, 8.4145E+02,
     +  8.4205E+02, 8.4264E+02, 8.4323E+02, 8.4383E+02, 8.4442E+02,
     + 8.4502E+02, 8.4561E+02, 8.4621E+02, 8.4680E+02, 8.4740E+02,
     +  8.4799E+02, 8.4859E+02, 8.4919E+02, 8.4978E+02, 8.5038E+02,
     + 8.5098E+02, 8.5157E+02, 8.5217E+02, 8.5277E+02, 8.5337E+02,
     +  8.5397E+02, 8.5457E+02, 8.5517E+02, 8.5576E+02, 8.5636E+02,
     + 8.5696E+02, 8.5756E+02, 8.5816E+02, 8.5876E+02, 8.5937E+02,
     +  8.5997E+02, 8.6057E+02, 8.6117E+02, 8.6177E+02, 8.6237E+02,
     + 8.6298E+02, 8.6358E+02, 8.6418E+02, 8.6479E+02, 8.6539E+02,
     +  8.6599E+02, 8.6660E+02, 8.6720E+02, 8.6781E+02, 8.6841E+02,
     + 8.6902E+02, 8.6962E+02, 8.7023E+02, 8.7083E+02, 8.7144E+02,
     +  8.7205E+02, 8.7265E+02, 8.7326E+02, 8.7387E+02, 8.7448E+02,
     + 8.7508E+02, 8.7569E+02, 8.7630E+02, 8.7691E+02, 8.7752E+02,
     +  8.7813E+02, 8.7874E+02, 8.7935E+02, 8.7996E+02, 8.8057E+02,
     + 8.8118E+02, 8.8179E+02, 8.8240E+02, 8.8301E+02, 8.8362E+02,
     +  8.8424E+02, 8.8485E+02, 8.8546E+02, 8.8607E+02, 8.8669E+02,
     + 8.8730E+02, 8.8791E+02, 8.8853E+02, 8.8914E+02, 8.8976E+02,
     +  8.9037E+02, 8.9098E+02, 8.9160E+02, 8.9222E+02, 8.9283E+02,
     + 8.9345E+02, 8.9406E+02, 8.9468E+02, 8.9530E+02, 8.9591E+02,
     +  8.9653E+02, 8.9715E+02, 8.9777E+02, 8.9839E+02, 8.9900E+02,
     + 8.9962E+02, 9.0024E+02, 9.0086E+02, 9.0148E+02, 9.0210E+02,
     +  9.0272E+02, 9.0334E+02, 9.0396E+02, 9.0458E+02, 9.0520E+02,
     + 9.0582E+02, 9.0645E+02, 9.0707E+02, 9.0769E+02, 9.0831E+02,
     +  9.0894E+02, 9.0956E+02, 9.1018E+02, 9.1081E+02, 9.1143E+02,
     + 9.1205E+02, 9.1268E+02, 9.1330E+02, 9.1393E+02, 9.1455E+02,
     +  9.1518E+02, 9.1580E+02, 9.1643E+02, 9.1706E+02, 9.1768E+02,
     + 9.1831E+02, 9.1894E+02, 9.1957E+02, 9.2019E+02, 9.2082E+02,
     +  9.2145E+02, 9.2208E+02, 9.2271E+02, 9.2334E+02, 9.2397E+02,
     + 9.2460E+02, 9.2523E+02, 9.2586E+02, 9.2649E+02, 9.2712E+02,
     +  9.2775E+02, 9.2838E+02, 9.2901E+02, 9.2964E+02, 9.3027E+02,
     + 9.3091E+02, 9.3154E+02, 9.3217E+02, 9.3281E+02, 9.3344E+02,
     +  9.3407E+02, 9.3471E+02, 9.3534E+02, 9.3598E+02, 9.3661E+02,
     + 9.3725E+02, 9.3788E+02, 9.3852E+02, 9.3915E+02, 9.3979E+02,
     +  9.4043E+02, 9.4106E+02, 9.4170E+02, 9.4234E+02, 9.4297E+02,
     + 9.4361E+02, 9.4425E+02, 9.4489E+02, 9.4553E+02, 9.4617E+02,
     +  9.4681E+02, 9.4745E+02, 9.4809E+02, 9.4873E+02, 9.4937E+02,
     + 9.5001E+02, 9.5065E+02, 9.5129E+02, 9.5193E+02, 9.5257E+02,
     +  9.5321E+02, 9.5386E+02, 9.5450E+02, 9.5514E+02, 9.5578E+02,
     + 9.5643E+02, 9.5707E+02, 9.5772E+02, 9.5836E+02, 9.5900E+02,
     +  9.5965E+02, 9.6029E+02, 9.6094E+02, 9.6159E+02, 9.6223E+02,
     + 9.6288E+02, 9.6352E+02, 9.6417E+02, 9.6482E+02, 9.6547E+02,
     +  9.6611E+02, 9.6676E+02, 9.6741E+02, 9.6806E+02, 9.6871E+02,
     + 9.6935E+02, 9.7000E+02, 9.7065E+02, 9.7130E+02, 9.7195E+02,
     +  9.7260E+02, 9.7325E+02, 9.7391E+02, 9.7456E+02, 9.7521E+02,
     + 9.7586E+02, 9.7651E+02, 9.7716E+02, 9.7782E+02, 9.7847E+02,
     +  9.7912E+02, 9.7978E+02, 9.8043E+02, 9.8108E+02, 9.8174E+02,
     + 9.8239E+02, 9.8305E+02, 9.8370E+02, 9.8436E+02, 9.8501E+02,
     +  9.8567E+02, 9.8633E+02, 9.8698E+02, 9.8764E+02, 9.8830E+02,
     + 9.8895E+02, 9.8961E+02, 9.9027E+02, 9.9093E+02, 9.9159E+02,
     +  9.9224E+02, 9.9290E+02, 9.9356E+02, 9.9422E+02, 9.9488E+02,
     + 9.9554E+02, 9.9620E+02, 9.9686E+02, 9.9752E+02, 9.9818E+02,
     +  9.9885E+02, 9.9951E+02, 1.0002E+03, 1.0008E+03, 1.0015E+03,
     + 1.0022E+03, 1.0028E+03, 1.0035E+03, 1.0041E+03, 1.0048E+03,
     +  1.0055E+03, 1.0061E+03, 1.0068E+03, 1.0075E+03, 1.0081E+03,
     + 1.0088E+03, 1.0095E+03, 1.0101E+03, 1.0108E+03, 1.0115E+03,
     +  1.0121E+03, 1.0128E+03, 1.0135E+03, 1.0141E+03, 1.0148E+03,
     + 1.0155E+03, 1.0161E+03, 1.0168E+03, 1.0175E+03, 1.0182E+03,
     +  1.0188E+03, 1.0195E+03, 1.0202E+03, 1.0208E+03, 1.0215E+03,
     + 1.0222E+03, 1.0228E+03, 1.0235E+03, 1.0242E+03, 1.0249E+03,
     +  1.0255E+03, 1.0262E+03, 1.0269E+03, 1.0276E+03, 1.0282E+03,
     + 1.0289E+03, 1.0296E+03, 1.0303E+03, 1.0309E+03, 1.0316E+03,
     +  1.0323E+03, 1.0330E+03, 1.0336E+03, 1.0343E+03, 1.0350E+03,
     + 1.0357E+03, 1.0363E+03, 1.0370E+03, 1.0377E+03, 1.0384E+03,
     +  1.0391E+03, 1.0397E+03, 1.0404E+03, 1.0411E+03, 1.0418E+03,
     + 1.0424E+03, 1.0431E+03, 1.0438E+03, 1.0445E+03, 1.0452E+03,
     +  1.0459E+03, 1.0465E+03, 1.0472E+03, 1.0479E+03, 1.0486E+03,
     + 1.0493E+03, 1.0499E+03, 1.0506E+03, 1.0513E+03, 1.0520E+03,
     +  1.0527E+03, 1.0534E+03, 1.0540E+03, 1.0547E+03, 1.0554E+03,
     + 1.0561E+03, 1.0568E+03, 1.0575E+03, 1.0582E+03, 1.0588E+03,
     +  1.0595E+03, 1.0602E+03, 1.0609E+03, 1.0616E+03, 1.0623E+03,
     + 1.0630E+03, 1.0637E+03, 1.0644E+03, 1.0650E+03, 1.0657E+03,
     +  1.0664E+03, 1.0671E+03, 1.0678E+03, 1.0685E+03, 1.0692E+03,
     + 1.0699E+03, 1.0706E+03, 1.0713E+03, 1.0719E+03, 1.0726E+03,
     +  1.0733E+03, 1.0740E+03, 1.0747E+03, 1.0754E+03, 1.0761E+03,
     + 1.0768E+03, 1.0775E+03, 1.0782E+03, 1.0789E+03, 1.0796E+03,
     +  1.0803E+03, 1.0810E+03, 1.0817E+03, 1.0824E+03, 1.0831E+03,
     + 1.0838E+03, 1.0845E+03, 1.0852E+03, 1.0858E+03, 1.0865E+03,
     +  1.0872E+03, 1.0879E+03, 1.0886E+03, 1.0893E+03, 1.0900E+03,
     + 1.0907E+03, 1.0914E+03, 1.0921E+03, 1.0928E+03, 1.0935E+03,
     +  1.0942E+03, 1.0949E+03, 1.0956E+03, 1.0963E+03, 1.0971E+03,
     + 1.0978E+03, 1.0985E+03, 1.0992E+03, 1.0999E+03, 1.1006E+03,
     +  1.1013E+03, 1.1020E+03, 1.1027E+03, 1.1034E+03, 1.1041E+03,
     + 1.1048E+03, 1.1055E+03, 1.1062E+03, 1.1069E+03, 1.1076E+03,
     +  1.1083E+03, 1.1090E+03, 1.1097E+03, 1.1104E+03, 1.1112E+03,
     + 1.1119E+03, 1.1126E+03, 1.1133E+03, 1.1140E+03, 1.1147E+03,
     +  1.1154E+03, 1.1161E+03, 1.1168E+03, 1.1175E+03, 1.1183E+03,
     + 1.1190E+03, 1.1197E+03, 1.1204E+03, 1.1211E+03, 1.1218E+03,
     +  1.1225E+03, 1.1232E+03, 1.1240E+03, 1.1247E+03, 1.1254E+03,
     + 1.1261E+03, 1.1268E+03, 1.1275E+03, 1.1282E+03, 1.1290E+03,
     +  1.1297E+03, 1.1304E+03, 1.1311E+03, 1.1318E+03, 1.1325E+03,
     + 1.1333E+03, 1.1340E+03, 1.1347E+03, 1.1354E+03, 1.1361E+03,
     +  1.1368E+03, 1.1376E+03, 1.1383E+03, 1.1390E+03, 1.1397E+03,
     + 1.1404E+03, 1.1412E+03, 1.1419E+03, 1.1426E+03, 1.1433E+03,
     +  1.1440E+03, 1.1448E+03, 1.1455E+03, 1.1462E+03, 1.1469E+03,
     + 1.1477E+03, 1.1484E+03, 1.1491E+03, 1.1498E+03, 1.1505E+03,
     +  1.1513E+03, 1.1520E+03, 1.1527E+03, 1.1534E+03, 1.1542E+03,
     + 1.1549E+03, 1.1556E+03, 1.1563E+03, 1.1571E+03, 1.1578E+03,
     +  1.1585E+03, 1.1593E+03, 1.1600E+03, 1.1607E+03, 1.1614E+03,
     + 1.1622E+03, 1.1629E+03, 1.1636E+03, 1.1644E+03, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_CS    (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 4),QofT( 4,NT),Q(NT)
      data (Tmx(I),I=1,4)/3500.,3500.,3500.,3500./ 
      data xgj/ 1.,1.,2.,4./
c...       CS
c...        --        22
      data (QofT( 1,J),J=1,2408)/ 1.2906E+00, 2.0792E+00, 2.9139E+00,
     +  3.7571E+00, 4.6036E+00, 5.4516E+00, 6.3004E+00, 7.1498E+00,
     + 7.9995E+00, 8.8496E+00, 9.6997E+00, 1.0550E+01, 1.1401E+01,
     +  1.2251E+01, 1.3102E+01, 1.3952E+01, 1.4803E+01, 1.5654E+01,
     + 1.6505E+01, 1.7356E+01, 1.9058E+01, 2.0760E+01, 2.2462E+01,
     +  2.4165E+01, 2.5867E+01, 2.7570E+01, 2.9272E+01, 3.0975E+01,
     + 3.2678E+01, 3.4381E+01, 3.6084E+01, 3.7787E+01, 3.9491E+01,
     +  4.1194E+01, 4.2898E+01, 4.4601E+01, 4.6305E+01, 4.8009E+01,
     + 4.9713E+01, 5.1416E+01, 5.3120E+01, 5.4825E+01, 5.6529E+01,
     +  5.8233E+01, 5.9938E+01, 6.1642E+01, 6.3347E+01, 6.5051E+01,
     + 6.6756E+01, 6.8461E+01, 7.0166E+01, 7.1871E+01, 7.3576E+01,
     +  7.5281E+01, 7.6986E+01, 7.8692E+01, 8.0397E+01, 8.2103E+01,
     + 8.3808E+01, 8.5514E+01, 8.7220E+01, 8.8926E+01, 9.0632E+01,
     +  9.2338E+01, 9.4044E+01, 9.5750E+01, 9.7456E+01, 9.9163E+01,
     + 1.0087E+02, 1.0258E+02, 1.0428E+02, 1.0599E+02, 1.0770E+02,
     +  1.0940E+02, 1.1111E+02, 1.1282E+02, 1.1452E+02, 1.1623E+02,
     + 1.1794E+02, 1.1965E+02, 1.2135E+02, 1.2306E+02, 1.2477E+02,
     +  1.2648E+02, 1.2819E+02, 1.2989E+02, 1.3160E+02, 1.3331E+02,
     + 1.3502E+02, 1.3673E+02, 1.3844E+02, 1.4014E+02, 1.4185E+02,
     +  1.4356E+02, 1.4527E+02, 1.4698E+02, 1.4869E+02, 1.5040E+02,
     + 1.5211E+02, 1.5382E+02, 1.5553E+02, 1.5724E+02, 1.5895E+02,
     +  1.6066E+02, 1.6237E+02, 1.6408E+02, 1.6579E+02, 1.6751E+02,
     + 1.6922E+02, 1.7093E+02, 1.7264E+02, 1.7435E+02, 1.7607E+02,
     +  1.7778E+02, 1.7949E+02, 1.8121E+02, 1.8292E+02, 1.8463E+02,
     + 1.8635E+02, 1.8806E+02, 1.8978E+02, 1.9149E+02, 1.9321E+02,
     +  1.9492E+02, 1.9664E+02, 1.9836E+02, 2.0008E+02, 2.0179E+02,
     + 2.0351E+02, 2.0523E+02, 2.0695E+02, 2.0867E+02, 2.1039E+02,
     +  2.1211E+02, 2.1383E+02, 2.1555E+02, 2.1728E+02, 2.1900E+02,
     + 2.2072E+02, 2.2245E+02, 2.2417E+02, 2.2590E+02, 2.2763E+02,
     +  2.2935E+02, 2.3108E+02, 2.3281E+02, 2.3454E+02, 2.3627E+02,
     + 2.3800E+02, 2.3973E+02, 2.4147E+02, 2.4320E+02, 2.4493E+02,
     +  2.4667E+02, 2.4841E+02, 2.5014E+02, 2.5188E+02, 2.5362E+02,
     + 2.5536E+02, 2.5710E+02, 2.5884E+02, 2.6059E+02, 2.6233E+02,
     +  2.6408E+02, 2.6582E+02, 2.6757E+02, 2.6932E+02, 2.7107E+02,
     + 2.7282E+02, 2.7458E+02, 2.7633E+02, 2.7808E+02, 2.7984E+02,
     +  2.8160E+02, 2.8336E+02, 2.8512E+02, 2.8688E+02, 2.8864E+02,
     + 2.9041E+02, 2.9217E+02, 2.9394E+02, 2.9571E+02, 2.9748E+02,
     +  2.9925E+02, 3.0103E+02, 3.0280E+02, 3.0458E+02, 3.0636E+02,
     + 3.0814E+02, 3.0992E+02, 3.1170E+02, 3.1349E+02, 3.1527E+02,
     +  3.1706E+02, 3.1885E+02, 3.2064E+02, 3.2244E+02, 3.2423E+02,
     + 3.2603E+02, 3.2783E+02, 3.2963E+02, 3.3143E+02, 3.3323E+02,
     +  3.3504E+02, 3.3685E+02, 3.3866E+02, 3.4047E+02, 3.4228E+02,
     + 3.4410E+02, 3.4592E+02, 3.4774E+02, 3.4956E+02, 3.5138E+02,
     +  3.5321E+02, 3.5504E+02, 3.5687E+02, 3.5870E+02, 3.6054E+02,
     + 3.6237E+02, 3.6421E+02, 3.6605E+02, 3.6790E+02, 3.6974E+02,
     +  3.7159E+02, 3.7344E+02, 3.7529E+02, 3.7715E+02, 3.7900E+02,
     + 3.8086E+02, 3.8272E+02, 3.8459E+02, 3.8645E+02, 3.8832E+02,
     +  3.9019E+02, 3.9207E+02, 3.9394E+02, 3.9582E+02, 3.9770E+02,
     + 3.9958E+02, 4.0147E+02, 4.0336E+02, 4.0525E+02, 4.0714E+02,
     +  4.0903E+02, 4.1093E+02, 4.1283E+02, 4.1474E+02, 4.1664E+02,
     + 4.1855E+02, 4.2046E+02, 4.2237E+02, 4.2429E+02, 4.2621E+02,
     +  4.2813E+02, 4.3005E+02, 4.3198E+02, 4.3391E+02, 4.3584E+02,
     + 4.3777E+02, 4.3971E+02, 4.4165E+02, 4.4359E+02, 4.4554E+02,
     +  4.4749E+02, 4.4944E+02, 4.5139E+02, 4.5335E+02, 4.5531E+02,
     + 4.5727E+02, 4.5923E+02, 4.6120E+02, 4.6317E+02, 4.6515E+02,
     +  4.6712E+02, 4.6910E+02, 4.7108E+02, 4.7307E+02, 4.7505E+02,
     + 4.7705E+02, 4.7904E+02, 4.8104E+02, 4.8303E+02, 4.8504E+02,
     +  4.8704E+02, 4.8905E+02, 4.9106E+02, 4.9308E+02, 4.9509E+02,
     + 4.9711E+02, 4.9914E+02, 5.0116E+02, 5.0319E+02, 5.0522E+02,
     +  5.0726E+02, 5.0930E+02, 5.1134E+02, 5.1338E+02, 5.1543E+02,
     + 5.1748E+02, 5.1953E+02, 5.2159E+02, 5.2365E+02, 5.2571E+02,
     +  5.2778E+02, 5.2985E+02, 5.3192E+02, 5.3399E+02, 5.3607E+02,
     + 5.3815E+02, 5.4024E+02, 5.4233E+02, 5.4442E+02, 5.4651E+02,
     +  5.4861E+02, 5.5071E+02, 5.5281E+02, 5.5492E+02, 5.5703E+02,
     + 5.5914E+02, 5.6126E+02, 5.6338E+02, 5.6550E+02, 5.6763E+02,
     +  5.6976E+02, 5.7189E+02, 5.7403E+02, 5.7616E+02, 5.7831E+02,
     + 5.8045E+02, 5.8260E+02, 5.8475E+02, 5.8691E+02, 5.8907E+02,
     +  5.9123E+02, 5.9340E+02, 5.9556E+02, 5.9774E+02, 5.9991E+02,
     + 6.0209E+02, 6.0427E+02, 6.0646E+02, 6.0865E+02, 6.1084E+02,
     +  6.1303E+02, 6.1523E+02, 6.1743E+02, 6.1964E+02, 6.2185E+02,
     + 6.2406E+02, 6.2627E+02, 6.2849E+02, 6.3072E+02, 6.3294E+02,
     +  6.3517E+02, 6.3740E+02, 6.3964E+02, 6.4188E+02, 6.4412E+02,
     + 6.4637E+02, 6.4862E+02, 6.5087E+02, 6.5313E+02, 6.5538E+02,
     +  6.5765E+02, 6.5991E+02, 6.6218E+02, 6.6446E+02, 6.6674E+02,
     + 6.6902E+02, 6.7130E+02, 6.7359E+02, 6.7588E+02, 6.7817E+02,
     +  6.8047E+02, 6.8277E+02, 6.8508E+02, 6.8738E+02, 6.8969E+02,
     + 6.9201E+02, 6.9433E+02, 6.9665E+02, 6.9898E+02, 7.0131E+02,
     +  7.0364E+02, 7.0597E+02, 7.0831E+02, 7.1066E+02, 7.1300E+02,
     + 7.1535E+02, 7.1771E+02, 7.2007E+02, 7.2243E+02, 7.2479E+02,
     +  7.2716E+02, 7.2953E+02, 7.3190E+02, 7.3428E+02, 7.3667E+02,
     + 7.3905E+02, 7.4144E+02, 7.4383E+02, 7.4623E+02, 7.4863E+02,
     +  7.5103E+02, 7.5344E+02, 7.5585E+02, 7.5826E+02, 7.6068E+02,
     + 7.6310E+02, 7.6553E+02, 7.6796E+02, 7.7039E+02, 7.7282E+02,
     +  7.7526E+02, 7.7771E+02, 7.8015E+02, 7.8260E+02, 7.8506E+02,
     + 7.8751E+02, 7.8998E+02, 7.9244E+02, 7.9491E+02, 7.9738E+02,
     +  7.9985E+02, 8.0233E+02, 8.0482E+02, 8.0730E+02, 8.0979E+02,
     + 8.1229E+02, 8.1478E+02, 8.1728E+02, 8.1979E+02, 8.2230E+02,
     +  8.2481E+02, 8.2732E+02, 8.2984E+02, 8.3237E+02, 8.3489E+02,
     + 8.3742E+02, 8.3996E+02, 8.4249E+02, 8.4503E+02, 8.4758E+02,
     +  8.5013E+02, 8.5268E+02, 8.5523E+02, 8.5779E+02, 8.6036E+02,
     + 8.6292E+02, 8.6549E+02, 8.6807E+02, 8.7064E+02, 8.7323E+02,
     +  8.7581E+02, 8.7840E+02, 8.8099E+02, 8.8359E+02, 8.8619E+02,
     + 8.8879E+02, 8.9140E+02, 8.9401E+02, 8.9662E+02, 8.9924E+02,
     +  9.0186E+02, 9.0449E+02, 9.0712E+02, 9.0975E+02, 9.1239E+02,
     + 9.1503E+02, 9.1767E+02, 9.2032E+02, 9.2297E+02, 9.2563E+02,
     +  9.2829E+02, 9.3095E+02, 9.3362E+02, 9.3629E+02, 9.3896E+02,
     + 9.4164E+02, 9.4432E+02, 9.4701E+02, 9.4969E+02, 9.5239E+02,
     +  9.5508E+02, 9.5778E+02, 9.6049E+02, 9.6319E+02, 9.6591E+02,
     + 9.6862E+02, 9.7134E+02, 9.7406E+02, 9.7679E+02, 9.7952E+02,
     +  9.8225E+02, 9.8499E+02, 9.8773E+02, 9.9048E+02, 9.9322E+02,
     + 9.9598E+02, 9.9873E+02, 1.0015E+03, 1.0043E+03, 1.0070E+03,
     +  1.0098E+03, 1.0126E+03, 1.0153E+03, 1.0181E+03, 1.0209E+03,
     + 1.0237E+03, 1.0265E+03, 1.0293E+03, 1.0321E+03, 1.0349E+03,
     +  1.0377E+03, 1.0405E+03, 1.0433E+03, 1.0462E+03, 1.0490E+03,
     + 1.0518E+03, 1.0546E+03, 1.0575E+03, 1.0603E+03, 1.0632E+03,
     +  1.0660E+03, 1.0689E+03, 1.0717E+03, 1.0746E+03, 1.0774E+03,
     + 1.0803E+03, 1.0832E+03, 1.0860E+03, 1.0889E+03, 1.0918E+03,
     +  1.0947E+03, 1.0976E+03, 1.1005E+03, 1.1034E+03, 1.1063E+03,
     + 1.1092E+03, 1.1121E+03, 1.1150E+03, 1.1179E+03, 1.1208E+03,
     +  1.1237E+03, 1.1267E+03, 1.1296E+03, 1.1325E+03, 1.1355E+03,
     + 1.1384E+03, 1.1414E+03, 1.1443E+03, 1.1473E+03, 1.1502E+03,
     +  1.1532E+03, 1.1561E+03, 1.1591E+03, 1.1621E+03, 1.1651E+03,
     + 1.1680E+03, 1.1710E+03, 1.1740E+03, 1.1770E+03, 1.1800E+03,
     +  1.1830E+03, 1.1860E+03, 1.1890E+03, 1.1920E+03, 1.1950E+03,
     + 1.1981E+03, 1.2011E+03, 1.2041E+03, 1.2071E+03, 1.2102E+03,
     +  1.2132E+03, 1.2162E+03, 1.2193E+03, 1.2223E+03, 1.2254E+03,
     + 1.2284E+03, 1.2315E+03, 1.2346E+03, 1.2376E+03, 1.2407E+03,
     +  1.2438E+03, 1.2469E+03, 1.2499E+03, 1.2530E+03, 1.2561E+03,
     + 1.2592E+03, 1.2623E+03, 1.2654E+03, 1.2685E+03, 1.2716E+03,
     +  1.2748E+03, 1.2779E+03, 1.2810E+03, 1.2841E+03, 1.2872E+03,
     + 1.2904E+03, 1.2935E+03, 1.2967E+03, 1.2998E+03, 1.3030E+03,
     +  1.3061E+03, 1.3093E+03, 1.3124E+03, 1.3156E+03, 1.3188E+03,
     + 1.3219E+03, 1.3251E+03, 1.3283E+03, 1.3315E+03, 1.3347E+03,
     +  1.3378E+03, 1.3410E+03, 1.3442E+03, 1.3474E+03, 1.3507E+03,
     + 1.3539E+03, 1.3571E+03, 1.3603E+03, 1.3635E+03, 1.3667E+03,
     +  1.3700E+03, 1.3732E+03, 1.3764E+03, 1.3797E+03, 1.3829E+03,
     + 1.3862E+03, 1.3894E+03, 1.3927E+03, 1.3960E+03, 1.3992E+03,
     +  1.4025E+03, 1.4058E+03, 1.4090E+03, 1.4123E+03, 1.4156E+03,
     + 1.4189E+03, 1.4222E+03, 1.4255E+03, 1.4288E+03, 1.4321E+03,
     +  1.4354E+03, 1.4387E+03, 1.4420E+03, 1.4453E+03, 1.4487E+03,
     + 1.4520E+03, 1.4553E+03, 1.4586E+03, 1.4620E+03, 1.4653E+03,
     +  1.4687E+03, 1.4720E+03, 1.4754E+03, 1.4787E+03, 1.4821E+03,
     + 1.4855E+03, 1.4888E+03, 1.4922E+03, 1.4956E+03, 1.4990E+03,
     +  1.5024E+03, 1.5057E+03, 1.5091E+03, 1.5125E+03, 1.5159E+03,
     + 1.5193E+03, 1.5228E+03, 1.5262E+03, 1.5296E+03, 1.5330E+03,
     +  1.5364E+03, 1.5399E+03, 1.5433E+03, 1.5467E+03, 1.5502E+03,
     + 1.5536E+03, 1.5571E+03, 1.5605E+03, 1.5640E+03, 1.5674E+03,
     +  1.5709E+03, 1.5744E+03, 1.5778E+03, 1.5813E+03, 1.5848E+03,
     + 1.5883E+03, 1.5917E+03, 1.5952E+03, 1.5987E+03, 1.6022E+03,
     +  1.6057E+03, 1.6092E+03, 1.6128E+03, 1.6163E+03, 1.6198E+03,
     + 1.6233E+03, 1.6268E+03, 1.6304E+03, 1.6339E+03, 1.6374E+03,
     +  1.6410E+03, 1.6445E+03, 1.6481E+03, 1.6516E+03, 1.6552E+03,
     + 1.6587E+03, 1.6623E+03, 1.6659E+03, 1.6695E+03, 1.6730E+03,
     +  1.6766E+03, 1.6802E+03, 1.6838E+03, 1.6874E+03, 1.6910E+03,
     + 1.6946E+03, 1.6982E+03, 1.7018E+03, 1.7054E+03, 1.7090E+03,
     +  1.7126E+03, 1.7163E+03, 1.7199E+03, 1.7235E+03, 1.7272E+03,
     + 1.7308E+03, 1.7344E+03, 1.7381E+03, 1.7417E+03, 1.7454E+03,
     +  1.7491E+03, 1.7527E+03, 1.7564E+03, 1.7601E+03, 1.7637E+03,
     + 1.7674E+03, 1.7711E+03, 1.7748E+03, 1.7785E+03, 1.7822E+03,
     +  1.7859E+03, 1.7896E+03, 1.7933E+03, 1.7970E+03, 1.8007E+03,
     + 1.8044E+03, 1.8081E+03, 1.8119E+03, 1.8156E+03, 1.8193E+03,
     +  1.8231E+03, 1.8268E+03, 1.8306E+03, 1.8343E+03, 1.8381E+03,
     + 1.8418E+03, 1.8456E+03, 1.8494E+03, 1.8531E+03, 1.8569E+03,
     +  1.8607E+03, 1.8645E+03, 1.8682E+03, 1.8720E+03, 1.8758E+03,
     + 1.8796E+03, 1.8834E+03, 1.8872E+03, 1.8910E+03, 1.8949E+03,
     +  1.8987E+03, 1.9025E+03, 1.9063E+03, 1.9101E+03, 1.9140E+03,
     + 1.9178E+03, 1.9217E+03, 1.9255E+03, 1.9294E+03, 1.9332E+03,
     +  1.9371E+03, 1.9409E+03, 1.9448E+03, 1.9487E+03, 1.9525E+03,
     + 1.9564E+03, 1.9603E+03, 1.9642E+03, 1.9681E+03, 1.9720E+03,
     +  1.9759E+03, 1.9798E+03, 1.9837E+03, 1.9876E+03, 1.9915E+03,
     + 1.9954E+03, 1.9993E+03, 2.0032E+03, 2.0072E+03, 2.0111E+03,
     +  2.0150E+03, 2.0190E+03, 2.0229E+03, 2.0269E+03, 2.0308E+03,
     + 2.0348E+03, 2.0387E+03, 2.0427E+03, 2.0467E+03, 2.0506E+03,
     +  2.0546E+03, 2.0586E+03, 2.0626E+03, 2.0666E+03, 2.0706E+03,
     + 2.0746E+03, 2.0786E+03, 2.0826E+03, 2.0866E+03, 2.0906E+03,
     +  2.0946E+03, 2.0986E+03, 2.1026E+03, 2.1067E+03, 2.1107E+03,
     + 2.1147E+03, 2.1188E+03, 2.1228E+03, 2.1269E+03, 2.1309E+03,
     +  2.1350E+03, 2.1390E+03, 2.1431E+03, 2.1472E+03, 2.1512E+03,
     + 2.1553E+03, 2.1594E+03, 2.1635E+03, 2.1675E+03, 2.1716E+03,
     +  2.1757E+03, 2.1798E+03, 2.1839E+03, 2.1880E+03, 2.1922E+03,
     + 2.1963E+03, 2.2004E+03, 2.2045E+03, 2.2086E+03, 2.2128E+03,
     +  2.2169E+03, 2.2210E+03, 2.2252E+03, 2.2293E+03, 2.2335E+03,
     + 2.2376E+03, 2.2418E+03, 2.2460E+03, 2.2501E+03, 2.2543E+03,
     +  2.2585E+03, 2.2626E+03, 2.2668E+03, 2.2710E+03, 2.2752E+03,
     + 2.2794E+03, 2.2836E+03, 2.2878E+03, 2.2920E+03, 2.2962E+03,
     +  2.3004E+03, 2.3047E+03, 2.3089E+03, 2.3131E+03, 2.3173E+03,
     + 2.3216E+03, 2.3258E+03, 2.3301E+03, 2.3343E+03, 2.3385E+03,
     +  2.3428E+03, 2.3471E+03, 2.3513E+03, 2.3556E+03, 2.3599E+03,
     + 2.3641E+03, 2.3684E+03, 2.3727E+03, 2.3770E+03, 2.3813E+03,
     +  2.3856E+03, 2.3899E+03, 2.3942E+03, 2.3985E+03, 2.4028E+03,
     + 2.4071E+03, 2.4114E+03, 2.4157E+03, 2.4201E+03, 2.4244E+03,
     +  2.4287E+03, 2.4331E+03, 2.4374E+03, 2.4418E+03, 2.4461E+03,
     + 2.4505E+03, 2.4548E+03, 2.4592E+03, 2.4636E+03, 2.4679E+03,
     +  2.4723E+03, 2.4767E+03, 2.4811E+03, 2.4854E+03, 2.4898E+03,
     + 2.4942E+03, 2.4986E+03, 2.5030E+03, 2.5074E+03, 2.5118E+03,
     +  2.5163E+03, 2.5207E+03, 2.5251E+03, 2.5295E+03, 2.5340E+03,
     + 2.5384E+03, 2.5428E+03, 2.5473E+03, 2.5517E+03, 2.5562E+03,
     +  2.5606E+03, 2.5651E+03, 2.5696E+03, 2.5740E+03, 2.5785E+03,
     + 2.5830E+03, 2.5874E+03, 2.5919E+03, 2.5964E+03, 2.6009E+03,
     +  2.6054E+03, 2.6099E+03, 2.6144E+03, 2.6189E+03, 2.6234E+03,
     + 2.6279E+03, 2.6325E+03, 2.6370E+03, 2.6415E+03, 2.6460E+03,
     +  2.6506E+03, 2.6551E+03, 2.6597E+03, 2.6642E+03, 2.6688E+03,
     + 2.6733E+03, 2.6779E+03, 2.6824E+03, 2.6870E+03, 2.6916E+03,
     +  2.6961E+03, 2.7007E+03, 2.7053E+03, 2.7099E+03, 2.7145E+03,
     + 2.7191E+03, 2.7237E+03, 2.7283E+03, 2.7329E+03, 2.7375E+03,
     +  2.7421E+03, 2.7467E+03, 2.7514E+03, 2.7560E+03, 2.7606E+03,
     + 2.7653E+03, 2.7699E+03, 2.7745E+03, 2.7792E+03, 2.7838E+03,
     +  2.7885E+03, 2.7932E+03, 2.7978E+03, 2.8025E+03, 2.8072E+03,
     + 2.8118E+03, 2.8165E+03, 2.8212E+03, 2.8259E+03, 2.8306E+03,
     +  2.8353E+03, 2.8400E+03, 2.8447E+03, 2.8494E+03, 2.8541E+03,
     + 2.8588E+03, 2.8635E+03, 2.8683E+03, 2.8730E+03, 2.8777E+03,
     +  2.8825E+03, 2.8872E+03, 2.8919E+03, 2.8967E+03, 2.9014E+03,
     + 2.9062E+03, 2.9110E+03, 2.9157E+03, 2.9205E+03, 2.9253E+03,
     +  2.9300E+03, 2.9348E+03, 2.9396E+03, 2.9444E+03, 2.9492E+03,
     + 2.9540E+03, 2.9588E+03, 2.9636E+03, 2.9684E+03, 2.9732E+03,
     +  2.9780E+03, 2.9829E+03, 2.9877E+03, 2.9925E+03, 2.9973E+03,
     + 3.0022E+03, 3.0070E+03, 3.0119E+03, 3.0167E+03, 3.0216E+03,
     +  3.0264E+03, 3.0313E+03, 3.0362E+03, 3.0410E+03, 3.0459E+03,
     + 3.0508E+03, 3.0557E+03, 3.0606E+03, 3.0654E+03, 3.0703E+03,
     +  3.0752E+03, 3.0801E+03, 3.0850E+03, 3.0900E+03, 3.0949E+03,
     + 3.0998E+03, 3.1047E+03, 3.1096E+03, 3.1146E+03, 3.1195E+03,
     +  3.1244E+03, 3.1294E+03, 3.1343E+03, 3.1393E+03, 3.1442E+03,
     + 3.1492E+03, 3.1542E+03, 3.1591E+03, 3.1641E+03, 3.1691E+03,
     +  3.1741E+03, 3.1790E+03, 3.1840E+03, 3.1890E+03, 3.1940E+03,
     + 3.1990E+03, 3.2040E+03, 3.2090E+03, 3.2140E+03, 3.2190E+03,
     +  3.2241E+03, 3.2291E+03, 3.2341E+03, 3.2392E+03, 3.2442E+03,
     + 3.2492E+03, 3.2543E+03, 3.2593E+03, 3.2644E+03, 3.2694E+03,
     +  3.2745E+03, 3.2796E+03, 3.2846E+03, 3.2897E+03, 3.2948E+03,
     + 3.2999E+03, 3.3049E+03, 3.3100E+03, 3.3151E+03, 3.3202E+03,
     +  3.3253E+03, 3.3304E+03, 3.3355E+03, 3.3406E+03, 3.3458E+03,
     + 3.3509E+03, 3.3560E+03, 3.3611E+03, 3.3663E+03, 3.3714E+03,
     +  3.3766E+03, 3.3817E+03, 3.3869E+03, 3.3920E+03, 3.3972E+03,
     + 3.4023E+03, 3.4075E+03, 3.4127E+03, 3.4178E+03, 3.4230E+03,
     +  3.4282E+03, 3.4334E+03, 3.4386E+03, 3.4438E+03, 3.4490E+03,
     + 3.4542E+03, 3.4594E+03, 3.4646E+03, 3.4698E+03, 3.4750E+03,
     +  3.4802E+03, 3.4855E+03, 3.4907E+03, 3.4959E+03, 3.5012E+03,
     + 3.5064E+03, 3.5117E+03, 3.5169E+03, 3.5222E+03, 3.5274E+03,
     +  3.5327E+03, 3.5380E+03, 3.5432E+03, 3.5485E+03, 3.5538E+03,
     + 3.5591E+03, 3.5644E+03, 3.5697E+03, 3.5750E+03, 3.5803E+03,
     +  3.5856E+03, 3.5909E+03, 3.5962E+03, 3.6015E+03, 3.6068E+03,
     + 3.6122E+03, 3.6175E+03, 3.6228E+03, 3.6282E+03, 3.6335E+03,
     +  3.6388E+03, 3.6442E+03, 3.6495E+03, 3.6549E+03, 3.6603E+03,
     + 3.6656E+03, 3.6710E+03, 3.6764E+03, 3.6818E+03, 3.6871E+03,
     +  3.6925E+03, 3.6979E+03, 3.7033E+03, 3.7087E+03, 3.7141E+03,
     + 3.7195E+03, 3.7249E+03, 3.7303E+03, 3.7358E+03, 3.7412E+03,
     +  3.7466E+03, 3.7520E+03, 3.7575E+03, 3.7629E+03, 3.7684E+03,
     + 3.7738E+03, 3.7793E+03, 3.7847E+03, 3.7902E+03, 3.7956E+03,
     +  3.8011E+03, 3.8066E+03, 3.8121E+03, 3.8175E+03, 3.8230E+03,
     + 3.8285E+03, 3.8340E+03, 3.8395E+03, 3.8450E+03, 3.8505E+03,
     +  3.8560E+03, 3.8615E+03, 3.8671E+03, 3.8726E+03, 3.8781E+03,
     + 3.8836E+03, 3.8892E+03, 3.8947E+03, 3.9003E+03, 3.9058E+03,
     +  3.9113E+03, 3.9169E+03, 3.9225E+03, 3.9280E+03, 3.9336E+03,
     + 3.9392E+03, 3.9447E+03, 3.9503E+03, 3.9559E+03, 3.9615E+03,
     +  3.9671E+03, 3.9727E+03, 3.9783E+03, 3.9839E+03, 3.9895E+03,
     + 3.9951E+03, 4.0007E+03, 4.0063E+03, 4.0120E+03, 4.0176E+03,
     +  4.0232E+03, 4.0289E+03, 4.0345E+03, 4.0401E+03, 4.0458E+03,
     + 4.0514E+03, 4.0571E+03, 4.0628E+03, 4.0684E+03, 4.0741E+03,
     +  4.0798E+03, 4.0854E+03, 4.0911E+03, 4.0968E+03, 4.1025E+03,
     + 4.1082E+03, 4.1139E+03, 4.1196E+03, 4.1253E+03, 4.1310E+03,
     +  4.1367E+03, 4.1424E+03, 4.1482E+03, 4.1539E+03, 4.1596E+03,
     + 4.1654E+03, 4.1711E+03, 4.1768E+03, 4.1826E+03, 4.1883E+03,
     +  4.1941E+03, 4.1999E+03, 4.2056E+03, 4.2114E+03, 4.2172E+03,
     + 4.2229E+03, 4.2287E+03, 4.2345E+03, 4.2403E+03, 4.2461E+03,
     +  4.2519E+03, 4.2577E+03, 4.2635E+03, 4.2693E+03, 4.2751E+03,
     + 4.2809E+03, 4.2868E+03, 4.2926E+03, 4.2984E+03, 4.3042E+03,
     +  4.3101E+03, 4.3159E+03, 4.3218E+03, 4.3276E+03, 4.3335E+03,
     + 4.3393E+03, 4.3452E+03, 4.3511E+03, 4.3569E+03, 4.3628E+03,
     +  4.3687E+03, 4.3746E+03, 4.3805E+03, 4.3864E+03, 4.3923E+03,
     + 4.3982E+03, 4.4041E+03, 4.4100E+03, 4.4159E+03, 4.4218E+03,
     +  4.4277E+03, 4.4336E+03, 4.4396E+03, 4.4455E+03, 4.4514E+03,
     + 4.4574E+03, 4.4633E+03, 4.4693E+03, 4.4752E+03, 4.4812E+03,
     +  4.4871E+03, 4.4931E+03, 4.4991E+03, 4.5051E+03, 4.5110E+03,
     + 4.5170E+03, 4.5230E+03, 4.5290E+03, 4.5350E+03, 4.5410E+03,
     +  4.5470E+03, 4.5530E+03, 4.5590E+03, 4.5650E+03, 4.5711E+03,
     + 4.5771E+03, 4.5831E+03, 4.5891E+03, 4.5952E+03, 4.6012E+03,
     +  4.6073E+03, 4.6133E+03, 4.6194E+03, 4.6254E+03, 4.6315E+03,
     + 4.6375E+03, 4.6436E+03, 4.6497E+03, 4.6558E+03, 4.6618E+03,
     +  4.6679E+03, 4.6740E+03, 4.6801E+03, 4.6862E+03, 4.6923E+03,
     + 4.6984E+03, 4.7045E+03, 4.7107E+03, 4.7168E+03, 4.7229E+03,
     +  4.7290E+03, 4.7352E+03, 4.7413E+03, 4.7474E+03, 4.7536E+03,
     + 4.7597E+03, 4.7659E+03, 4.7720E+03, 4.7782E+03, 4.7844E+03,
     +  4.7905E+03, 4.7967E+03, 4.8029E+03, 4.8091E+03, 4.8152E+03,
     + 4.8214E+03, 4.8276E+03, 4.8338E+03, 4.8400E+03, 4.8462E+03,
     +  4.8524E+03, 4.8587E+03, 4.8649E+03, 4.8711E+03, 4.8773E+03,
     + 4.8836E+03, 4.8898E+03, 4.8960E+03, 4.9023E+03, 4.9085E+03,
     +  4.9148E+03, 4.9210E+03, 4.9273E+03, 4.9336E+03, 4.9398E+03,
     + 4.9461E+03, 4.9524E+03, 4.9587E+03, 4.9649E+03, 4.9712E+03,
     +  4.9775E+03, 4.9838E+03, 4.9901E+03, 4.9964E+03, 5.0027E+03,
     + 5.0091E+03, 5.0154E+03, 5.0217E+03, 5.0280E+03, 5.0344E+03,
     +  5.0407E+03, 5.0470E+03, 5.0534E+03, 5.0597E+03, 5.0661E+03,
     + 5.0724E+03, 5.0788E+03, 5.0851E+03, 5.0915E+03, 5.0979E+03,
     +  5.1043E+03, 5.1106E+03, 5.1170E+03, 5.1234E+03, 5.1298E+03,
     + 5.1362E+03, 5.1426E+03, 5.1490E+03, 5.1554E+03, 5.1618E+03,
     +  5.1683E+03, 5.1747E+03, 5.1811E+03, 5.1875E+03, 5.1940E+03,
     + 5.2004E+03, 5.2069E+03, 5.2133E+03, 5.2198E+03, 5.2262E+03,
     +  5.2327E+03, 5.2391E+03, 5.2456E+03, 5.2521E+03, 5.2586E+03,
     + 5.2650E+03, 5.2715E+03, 5.2780E+03, 5.2845E+03, 5.2910E+03,
     +  5.2975E+03, 5.3040E+03, 5.3105E+03, 5.3170E+03, 5.3236E+03,
     + 5.3301E+03, 5.3366E+03, 5.3431E+03, 5.3497E+03, 5.3562E+03,
     +  5.3627E+03, 5.3693E+03, 5.3758E+03, 5.3824E+03, 5.3890E+03,
     + 5.3955E+03, 5.4021E+03, 5.4087E+03, 5.4152E+03, 5.4218E+03,
     +  5.4284E+03, 5.4350E+03, 5.4416E+03, 5.4482E+03, 5.4548E+03,
     + 5.4614E+03, 5.4680E+03, 5.4746E+03, 5.4812E+03, 5.4879E+03,
     +  5.4945E+03, 5.5011E+03, 5.5078E+03, 5.5144E+03, 5.5210E+03,
     + 5.5277E+03, 5.5343E+03, 5.5410E+03, 5.5477E+03, 5.5543E+03,
     +  5.5610E+03, 5.5677E+03, 5.5743E+03, 5.5810E+03, 5.5877E+03,
     + 5.5944E+03, 5.6011E+03, 5.6078E+03, 5.6145E+03, 5.6212E+03,
     +  5.6279E+03, 5.6346E+03, 5.6413E+03, 5.6481E+03, 5.6548E+03,
     + 5.6615E+03, 5.6683E+03, 5.6750E+03, 5.6817E+03, 5.6885E+03,
     +  5.6952E+03, 5.7020E+03, 5.7088E+03, 5.7155E+03, 5.7223E+03,
     + 5.7291E+03, 5.7358E+03, 5.7426E+03, 5.7494E+03, 5.7562E+03,
     +  5.7630E+03, 5.7698E+03, 5.7766E+03, 5.7834E+03, 5.7902E+03,
     + 5.7970E+03, 5.8039E+03, 5.8107E+03, 5.8175E+03, 5.8243E+03,
     +  5.8312E+03, 5.8380E+03, 5.8449E+03, 5.8517E+03, 5.8586E+03,
     + 5.8654E+03, 5.8723E+03, 5.8791E+03, 5.8860E+03, 5.8929E+03,
     +  5.8998E+03, 5.9067E+03, 5.9135E+03, 5.9204E+03, 5.9273E+03,
     + 5.9342E+03, 5.9411E+03, 5.9480E+03, 5.9549E+03, 5.9619E+03,
     +  5.9688E+03, 5.9757E+03, 5.9826E+03, 5.9896E+03, 5.9965E+03,
     + 6.0034E+03, 6.0104E+03, 6.0173E+03, 6.0243E+03, 6.0313E+03,
     +  6.0382E+03, 6.0452E+03, 6.0522E+03, 6.0591E+03, 6.0661E+03,
     + 6.0731E+03, 6.0801E+03, 6.0871E+03, 6.0941E+03, 6.1011E+03,
     +  6.1081E+03, 6.1151E+03, 6.1221E+03, 6.1291E+03, 6.1361E+03,
     + 6.1432E+03, 6.1502E+03, 6.1572E+03, 6.1643E+03, 6.1713E+03,
     +  6.1783E+03, 6.1854E+03, 6.1924E+03, 6.1995E+03, 6.2066E+03,
     + 6.2136E+03, 6.2207E+03, 6.2278E+03, 6.2349E+03, 6.2419E+03,
     +  6.2490E+03, 6.2561E+03, 6.2632E+03, 6.2703E+03, 6.2774E+03,
     + 6.2845E+03, 6.2917E+03, 6.2988E+03, 6.3059E+03, 6.3130E+03,
     +  6.3202E+03, 6.3273E+03, 6.3344E+03, 6.3416E+03, 6.3487E+03,
     + 6.3559E+03, 6.3630E+03, 6.3702E+03, 6.3774E+03, 6.3845E+03,
     +  6.3917E+03, 6.3989E+03, 6.4061E+03, 6.4132E+03, 6.4204E+03,
     + 6.4276E+03, 6.4348E+03, 6.4420E+03, 6.4492E+03, 6.4564E+03,
     +  6.4637E+03, 6.4709E+03, 6.4781E+03, 6.4853E+03, 6.4926E+03,
     + 6.4998E+03, 6.5070E+03, 6.5143E+03, 6.5215E+03, 6.5288E+03,
     +  6.5360E+03, 6.5433E+03, 6.5506E+03, 6.5578E+03, 6.5651E+03,
     + 6.5724E+03, 6.5797E+03, 6.5870E+03, 6.5942E+03, 6.6015E+03,
     +  6.6088E+03, 6.6161E+03, 6.6235E+03, 6.6308E+03, 6.6381E+03,
     + 6.6454E+03, 6.6527E+03, 6.6601E+03, 6.6674E+03, 6.6747E+03,
     +  6.6821E+03, 6.6894E+03, 6.6968E+03, 6.7041E+03, 6.7115E+03,
     + 6.7188E+03, 6.7262E+03, 6.7336E+03, 6.7410E+03, 6.7483E+03,
     +  6.7557E+03, 6.7631E+03, 6.7705E+03, 6.7779E+03, 6.7853E+03,
     + 6.7927E+03, 6.8001E+03, 6.8075E+03, 6.8149E+03, 6.8224E+03,
     +  6.8298E+03, 6.8372E+03, 6.8447E+03, 6.8521E+03, 6.8595E+03,
     + 6.8670E+03, 6.8744E+03, 6.8819E+03, 6.8894E+03, 6.8968E+03,
     +  6.9043E+03, 6.9118E+03, 6.9192E+03, 6.9267E+03, 6.9342E+03,
     + 6.9417E+03, 6.9492E+03, 6.9567E+03, 6.9642E+03, 6.9717E+03,
     +  6.9792E+03, 6.9867E+03, 6.9942E+03, 7.0018E+03, 7.0093E+03,
     + 7.0168E+03, 7.0244E+03, 7.0319E+03, 7.0395E+03, 7.0470E+03,
     +  7.0546E+03, 7.0621E+03, 7.0697E+03, 7.0772E+03, 7.0848E+03,
     + 7.0924E+03, 7.1000E+03, 7.1076E+03, 7.1151E+03, 7.1227E+03,
     +  7.1303E+03, 7.1379E+03, 7.1455E+03, 7.1531E+03, 7.1608E+03,
     + 7.1684E+03, 7.1760E+03, 7.1836E+03, 7.1912E+03, 7.1989E+03,
     +  7.2065E+03, 7.2142E+03, 7.2218E+03, 7.2295E+03, 7.2371E+03,
     + 7.2448E+03, 7.2524E+03, 7.2601E+03, 7.2678E+03, 7.2755E+03,
     +  7.2831E+03, 7.2908E+03, 7.2985E+03, 7.3062E+03, 7.3139E+03,
     + 7.3216E+03, 7.3293E+03, 7.3370E+03, 7.3447E+03, 7.3525E+03,
     +  7.3602E+03, 7.3679E+03, 7.3756E+03, 7.3834E+03, 7.3911E+03,
     + 7.3989E+03, 7.4066E+03, 7.4144E+03, 7.4221E+03, 7.4299E+03,
     +  7.4377E+03, 7.4454E+03, 7.4532E+03, 7.4610E+03, 7.4688E+03,
     + 7.4766E+03, 7.4843E+03, 7.4921E+03, 7.4999E+03, 7.5078E+03,
     +  7.5156E+03, 7.5234E+03, 7.5312E+03, 7.5390E+03, 7.5468E+03,
     + 7.5547E+03, 7.5625E+03, 7.5703E+03, 7.5782E+03, 7.5860E+03,
     +  7.5939E+03, 7.6017E+03, 7.6096E+03, 7.6175E+03, 7.6253E+03,
     + 7.6332E+03, 7.6411E+03, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        24
      data (QofT( 2,J),J=1,2408)/ 1.3019E+00, 2.1061E+00, 2.9549E+00,
     +  3.8121E+00, 4.6725E+00, 5.5343E+00, 6.3971E+00, 7.2603E+00,
     + 8.1239E+00, 8.9878E+00, 9.8518E+00, 1.0716E+01, 1.1580E+01,
     +  1.2445E+01, 1.3309E+01, 1.4174E+01, 1.5038E+01, 1.5903E+01,
     + 1.6768E+01, 1.7633E+01, 1.9362E+01, 2.1092E+01, 2.2822E+01,
     +  2.4552E+01, 2.6282E+01, 2.8013E+01, 2.9743E+01, 3.1473E+01,
     + 3.3204E+01, 3.4935E+01, 3.6666E+01, 3.8396E+01, 4.0127E+01,
     +  4.1859E+01, 4.3590E+01, 4.5321E+01, 4.7052E+01, 4.8784E+01,
     + 5.0515E+01, 5.2247E+01, 5.3979E+01, 5.5711E+01, 5.7443E+01,
     +  5.9175E+01, 6.0907E+01, 6.2639E+01, 6.4371E+01, 6.6104E+01,
     + 6.7836E+01, 6.9569E+01, 7.1301E+01, 7.3034E+01, 7.4767E+01,
     +  7.6500E+01, 7.8233E+01, 7.9966E+01, 8.1699E+01, 8.3432E+01,
     + 8.5166E+01, 8.6899E+01, 8.8633E+01, 9.0366E+01, 9.2100E+01,
     +  9.3834E+01, 9.5568E+01, 9.7302E+01, 9.9036E+01, 1.0077E+02,
     + 1.0250E+02, 1.0424E+02, 1.0597E+02, 1.0771E+02, 1.0944E+02,
     +  1.1118E+02, 1.1291E+02, 1.1465E+02, 1.1638E+02, 1.1812E+02,
     + 1.1985E+02, 1.2159E+02, 1.2332E+02, 1.2506E+02, 1.2679E+02,
     +  1.2853E+02, 1.3027E+02, 1.3200E+02, 1.3374E+02, 1.3547E+02,
     + 1.3721E+02, 1.3895E+02, 1.4068E+02, 1.4242E+02, 1.4416E+02,
     +  1.4589E+02, 1.4763E+02, 1.4937E+02, 1.5110E+02, 1.5284E+02,
     + 1.5458E+02, 1.5632E+02, 1.5805E+02, 1.5979E+02, 1.6153E+02,
     +  1.6327E+02, 1.6501E+02, 1.6675E+02, 1.6849E+02, 1.7023E+02,
     + 1.7196E+02, 1.7370E+02, 1.7544E+02, 1.7719E+02, 1.7893E+02,
     +  1.8067E+02, 1.8241E+02, 1.8415E+02, 1.8589E+02, 1.8763E+02,
     + 1.8938E+02, 1.9112E+02, 1.9286E+02, 1.9461E+02, 1.9635E+02,
     +  1.9809E+02, 1.9984E+02, 2.0158E+02, 2.0333E+02, 2.0508E+02,
     + 2.0682E+02, 2.0857E+02, 2.1032E+02, 2.1207E+02, 2.1381E+02,
     +  2.1556E+02, 2.1731E+02, 2.1906E+02, 2.2081E+02, 2.2257E+02,
     + 2.2432E+02, 2.2607E+02, 2.2783E+02, 2.2958E+02, 2.3134E+02,
     +  2.3309E+02, 2.3485E+02, 2.3661E+02, 2.3836E+02, 2.4012E+02,
     + 2.4188E+02, 2.4364E+02, 2.4541E+02, 2.4717E+02, 2.4893E+02,
     +  2.5070E+02, 2.5246E+02, 2.5423E+02, 2.5600E+02, 2.5777E+02,
     + 2.5954E+02, 2.6131E+02, 2.6308E+02, 2.6485E+02, 2.6663E+02,
     +  2.6840E+02, 2.7018E+02, 2.7195E+02, 2.7373E+02, 2.7551E+02,
     + 2.7730E+02, 2.7908E+02, 2.8086E+02, 2.8265E+02, 2.8443E+02,
     +  2.8622E+02, 2.8801E+02, 2.8980E+02, 2.9159E+02, 2.9339E+02,
     + 2.9518E+02, 2.9698E+02, 2.9878E+02, 3.0058E+02, 3.0238E+02,
     +  3.0418E+02, 3.0599E+02, 3.0779E+02, 3.0960E+02, 3.1141E+02,
     + 3.1322E+02, 3.1503E+02, 3.1685E+02, 3.1866E+02, 3.2048E+02,
     +  3.2230E+02, 3.2412E+02, 3.2595E+02, 3.2777E+02, 3.2960E+02,
     + 3.3143E+02, 3.3326E+02, 3.3509E+02, 3.3693E+02, 3.3876E+02,
     +  3.4060E+02, 3.4244E+02, 3.4428E+02, 3.4613E+02, 3.4798E+02,
     + 3.4982E+02, 3.5167E+02, 3.5353E+02, 3.5538E+02, 3.5724E+02,
     +  3.5910E+02, 3.6096E+02, 3.6282E+02, 3.6469E+02, 3.6656E+02,
     + 3.6843E+02, 3.7030E+02, 3.7217E+02, 3.7405E+02, 3.7593E+02,
     +  3.7781E+02, 3.7970E+02, 3.8158E+02, 3.8347E+02, 3.8536E+02,
     + 3.8725E+02, 3.8915E+02, 3.9105E+02, 3.9295E+02, 3.9485E+02,
     +  3.9676E+02, 3.9866E+02, 4.0058E+02, 4.0249E+02, 4.0440E+02,
     + 4.0632E+02, 4.0824E+02, 4.1016E+02, 4.1209E+02, 4.1402E+02,
     +  4.1595E+02, 4.1788E+02, 4.1982E+02, 4.2176E+02, 4.2370E+02,
     + 4.2564E+02, 4.2759E+02, 4.2954E+02, 4.3149E+02, 4.3345E+02,
     +  4.3540E+02, 4.3736E+02, 4.3933E+02, 4.4129E+02, 4.4326E+02,
     + 4.4523E+02, 4.4721E+02, 4.4918E+02, 4.5116E+02, 4.5314E+02,
     +  4.5513E+02, 4.5712E+02, 4.5911E+02, 4.6110E+02, 4.6310E+02,
     + 4.6510E+02, 4.6710E+02, 4.6911E+02, 4.7112E+02, 4.7313E+02,
     +  4.7514E+02, 4.7716E+02, 4.7918E+02, 4.8120E+02, 4.8323E+02,
     + 4.8526E+02, 4.8729E+02, 4.8932E+02, 4.9136E+02, 4.9340E+02,
     +  4.9545E+02, 4.9749E+02, 4.9954E+02, 5.0160E+02, 5.0365E+02,
     + 5.0571E+02, 5.0778E+02, 5.0984E+02, 5.1191E+02, 5.1398E+02,
     +  5.1606E+02, 5.1814E+02, 5.2022E+02, 5.2230E+02, 5.2439E+02,
     + 5.2648E+02, 5.2857E+02, 5.3067E+02, 5.3277E+02, 5.3488E+02,
     +  5.3698E+02, 5.3909E+02, 5.4120E+02, 5.4332E+02, 5.4544E+02,
     + 5.4756E+02, 5.4969E+02, 5.5182E+02, 5.5395E+02, 5.5609E+02,
     +  5.5823E+02, 5.6037E+02, 5.6251E+02, 5.6466E+02, 5.6681E+02,
     + 5.6897E+02, 5.7113E+02, 5.7329E+02, 5.7546E+02, 5.7763E+02,
     +  5.7980E+02, 5.8197E+02, 5.8415E+02, 5.8633E+02, 5.8852E+02,
     + 5.9071E+02, 5.9290E+02, 5.9510E+02, 5.9729E+02, 5.9950E+02,
     +  6.0170E+02, 6.0391E+02, 6.0612E+02, 6.0834E+02, 6.1056E+02,
     + 6.1278E+02, 6.1501E+02, 6.1724E+02, 6.1947E+02, 6.2171E+02,
     +  6.2395E+02, 6.2619E+02, 6.2844E+02, 6.3069E+02, 6.3294E+02,
     + 6.3520E+02, 6.3746E+02, 6.3972E+02, 6.4199E+02, 6.4426E+02,
     +  6.4654E+02, 6.4881E+02, 6.5109E+02, 6.5338E+02, 6.5567E+02,
     + 6.5796E+02, 6.6026E+02, 6.6255E+02, 6.6486E+02, 6.6716E+02,
     +  6.6947E+02, 6.7179E+02, 6.7410E+02, 6.7642E+02, 6.7875E+02,
     + 6.8107E+02, 6.8340E+02, 6.8574E+02, 6.8808E+02, 6.9042E+02,
     +  6.9276E+02, 6.9511E+02, 6.9746E+02, 6.9982E+02, 7.0218E+02,
     + 7.0454E+02, 7.0691E+02, 7.0928E+02, 7.1165E+02, 7.1403E+02,
     +  7.1641E+02, 7.1880E+02, 7.2118E+02, 7.2358E+02, 7.2597E+02,
     + 7.2837E+02, 7.3077E+02, 7.3318E+02, 7.3559E+02, 7.3800E+02,
     +  7.4042E+02, 7.4284E+02, 7.4526E+02, 7.4769E+02, 7.5012E+02,
     + 7.5256E+02, 7.5500E+02, 7.5744E+02, 7.5989E+02, 7.6234E+02,
     +  7.6479E+02, 7.6725E+02, 7.6971E+02, 7.7217E+02, 7.7464E+02,
     + 7.7712E+02, 7.7959E+02, 7.8207E+02, 7.8455E+02, 7.8704E+02,
     +  7.8953E+02, 7.9202E+02, 7.9452E+02, 7.9702E+02, 7.9953E+02,
     + 8.0204E+02, 8.0455E+02, 8.0707E+02, 8.0959E+02, 8.1211E+02,
     +  8.1464E+02, 8.1717E+02, 8.1970E+02, 8.2224E+02, 8.2479E+02,
     + 8.2733E+02, 8.2988E+02, 8.3244E+02, 8.3499E+02, 8.3755E+02,
     +  8.4012E+02, 8.4269E+02, 8.4526E+02, 8.4784E+02, 8.5042E+02,
     + 8.5300E+02, 8.5559E+02, 8.5818E+02, 8.6077E+02, 8.6337E+02,
     +  8.6597E+02, 8.6858E+02, 8.7119E+02, 8.7380E+02, 8.7642E+02,
     + 8.7904E+02, 8.8167E+02, 8.8429E+02, 8.8693E+02, 8.8956E+02,
     +  8.9220E+02, 8.9485E+02, 8.9749E+02, 9.0015E+02, 9.0280E+02,
     + 9.0546E+02, 9.0812E+02, 9.1079E+02, 9.1346E+02, 9.1613E+02,
     +  9.1881E+02, 9.2149E+02, 9.2418E+02, 9.2687E+02, 9.2956E+02,
     + 9.3226E+02, 9.3496E+02, 9.3766E+02, 9.4037E+02, 9.4308E+02,
     +  9.4580E+02, 9.4852E+02, 9.5124E+02, 9.5397E+02, 9.5670E+02,
     + 9.5944E+02, 9.6217E+02, 9.6492E+02, 9.6766E+02, 9.7041E+02,
     +  9.7317E+02, 9.7593E+02, 9.7869E+02, 9.8145E+02, 9.8422E+02,
     + 9.8700E+02, 9.8977E+02, 9.9255E+02, 9.9534E+02, 9.9813E+02,
     +  1.0009E+03, 1.0037E+03, 1.0065E+03, 1.0093E+03, 1.0121E+03,
     + 1.0149E+03, 1.0178E+03, 1.0206E+03, 1.0234E+03, 1.0262E+03,
     +  1.0291E+03, 1.0319E+03, 1.0347E+03, 1.0376E+03, 1.0404E+03,
     + 1.0433E+03, 1.0461E+03, 1.0490E+03, 1.0518E+03, 1.0547E+03,
     +  1.0576E+03, 1.0605E+03, 1.0633E+03, 1.0662E+03, 1.0691E+03,
     + 1.0720E+03, 1.0749E+03, 1.0778E+03, 1.0807E+03, 1.0836E+03,
     +  1.0865E+03, 1.0894E+03, 1.0923E+03, 1.0952E+03, 1.0982E+03,
     + 1.1011E+03, 1.1040E+03, 1.1070E+03, 1.1099E+03, 1.1128E+03,
     +  1.1158E+03, 1.1187E+03, 1.1217E+03, 1.1247E+03, 1.1276E+03,
     + 1.1306E+03, 1.1336E+03, 1.1365E+03, 1.1395E+03, 1.1425E+03,
     +  1.1455E+03, 1.1485E+03, 1.1515E+03, 1.1545E+03, 1.1575E+03,
     + 1.1605E+03, 1.1635E+03, 1.1665E+03, 1.1695E+03, 1.1725E+03,
     +  1.1756E+03, 1.1786E+03, 1.1816E+03, 1.1847E+03, 1.1877E+03,
     + 1.1907E+03, 1.1938E+03, 1.1968E+03, 1.1999E+03, 1.2030E+03,
     +  1.2060E+03, 1.2091E+03, 1.2122E+03, 1.2152E+03, 1.2183E+03,
     + 1.2214E+03, 1.2245E+03, 1.2276E+03, 1.2307E+03, 1.2338E+03,
     +  1.2369E+03, 1.2400E+03, 1.2431E+03, 1.2462E+03, 1.2493E+03,
     + 1.2525E+03, 1.2556E+03, 1.2587E+03, 1.2619E+03, 1.2650E+03,
     +  1.2681E+03, 1.2713E+03, 1.2744E+03, 1.2776E+03, 1.2807E+03,
     + 1.2839E+03, 1.2871E+03, 1.2902E+03, 1.2934E+03, 1.2966E+03,
     +  1.2998E+03, 1.3030E+03, 1.3062E+03, 1.3094E+03, 1.3126E+03,
     + 1.3158E+03, 1.3190E+03, 1.3222E+03, 1.3254E+03, 1.3286E+03,
     +  1.3318E+03, 1.3351E+03, 1.3383E+03, 1.3415E+03, 1.3448E+03,
     + 1.3480E+03, 1.3512E+03, 1.3545E+03, 1.3577E+03, 1.3610E+03,
     +  1.3643E+03, 1.3675E+03, 1.3708E+03, 1.3741E+03, 1.3773E+03,
     + 1.3806E+03, 1.3839E+03, 1.3872E+03, 1.3905E+03, 1.3938E+03,
     +  1.3971E+03, 1.4004E+03, 1.4037E+03, 1.4070E+03, 1.4103E+03,
     + 1.4137E+03, 1.4170E+03, 1.4203E+03, 1.4237E+03, 1.4270E+03,
     +  1.4303E+03, 1.4337E+03, 1.4370E+03, 1.4404E+03, 1.4437E+03,
     + 1.4471E+03, 1.4505E+03, 1.4538E+03, 1.4572E+03, 1.4606E+03,
     +  1.4640E+03, 1.4673E+03, 1.4707E+03, 1.4741E+03, 1.4775E+03,
     + 1.4809E+03, 1.4843E+03, 1.4877E+03, 1.4911E+03, 1.4946E+03,
     +  1.4980E+03, 1.5014E+03, 1.5048E+03, 1.5083E+03, 1.5117E+03,
     + 1.5151E+03, 1.5186E+03, 1.5220E+03, 1.5255E+03, 1.5290E+03,
     +  1.5324E+03, 1.5359E+03, 1.5393E+03, 1.5428E+03, 1.5463E+03,
     + 1.5498E+03, 1.5533E+03, 1.5567E+03, 1.5602E+03, 1.5637E+03,
     +  1.5672E+03, 1.5707E+03, 1.5742E+03, 1.5778E+03, 1.5813E+03,
     + 1.5848E+03, 1.5883E+03, 1.5919E+03, 1.5954E+03, 1.5989E+03,
     +  1.6025E+03, 1.6060E+03, 1.6096E+03, 1.6131E+03, 1.6167E+03,
     + 1.6202E+03, 1.6238E+03, 1.6274E+03, 1.6309E+03, 1.6345E+03,
     +  1.6381E+03, 1.6417E+03, 1.6453E+03, 1.6489E+03, 1.6525E+03,
     + 1.6561E+03, 1.6597E+03, 1.6633E+03, 1.6669E+03, 1.6705E+03,
     +  1.6741E+03, 1.6777E+03, 1.6814E+03, 1.6850E+03, 1.6886E+03,
     + 1.6923E+03, 1.6959E+03, 1.6996E+03, 1.7032E+03, 1.7069E+03,
     +  1.7105E+03, 1.7142E+03, 1.7179E+03, 1.7216E+03, 1.7252E+03,
     + 1.7289E+03, 1.7326E+03, 1.7363E+03, 1.7400E+03, 1.7437E+03,
     +  1.7474E+03, 1.7511E+03, 1.7548E+03, 1.7585E+03, 1.7622E+03,
     + 1.7659E+03, 1.7697E+03, 1.7734E+03, 1.7771E+03, 1.7809E+03,
     +  1.7846E+03, 1.7884E+03, 1.7921E+03, 1.7959E+03, 1.7996E+03,
     + 1.8034E+03, 1.8071E+03, 1.8109E+03, 1.8147E+03, 1.8185E+03,
     +  1.8222E+03, 1.8260E+03, 1.8298E+03, 1.8336E+03, 1.8374E+03,
     + 1.8412E+03, 1.8450E+03, 1.8488E+03, 1.8526E+03, 1.8565E+03,
     +  1.8603E+03, 1.8641E+03, 1.8679E+03, 1.8718E+03, 1.8756E+03,
     + 1.8795E+03, 1.8833E+03, 1.8872E+03, 1.8910E+03, 1.8949E+03,
     +  1.8987E+03, 1.9026E+03, 1.9065E+03, 1.9103E+03, 1.9142E+03,
     + 1.9181E+03, 1.9220E+03, 1.9259E+03, 1.9298E+03, 1.9337E+03,
     +  1.9376E+03, 1.9415E+03, 1.9454E+03, 1.9493E+03, 1.9532E+03,
     + 1.9572E+03, 1.9611E+03, 1.9650E+03, 1.9689E+03, 1.9729E+03,
     +  1.9768E+03, 1.9808E+03, 1.9847E+03, 1.9887E+03, 1.9926E+03,
     + 1.9966E+03, 2.0006E+03, 2.0045E+03, 2.0085E+03, 2.0125E+03,
     +  2.0165E+03, 2.0205E+03, 2.0245E+03, 2.0285E+03, 2.0325E+03,
     + 2.0365E+03, 2.0405E+03, 2.0445E+03, 2.0485E+03, 2.0525E+03,
     +  2.0566E+03, 2.0606E+03, 2.0646E+03, 2.0687E+03, 2.0727E+03,
     + 2.0767E+03, 2.0808E+03, 2.0848E+03, 2.0889E+03, 2.0930E+03,
     +  2.0970E+03, 2.1011E+03, 2.1052E+03, 2.1092E+03, 2.1133E+03,
     + 2.1174E+03, 2.1215E+03, 2.1256E+03, 2.1297E+03, 2.1338E+03,
     +  2.1379E+03, 2.1420E+03, 2.1461E+03, 2.1502E+03, 2.1544E+03,
     + 2.1585E+03, 2.1626E+03, 2.1668E+03, 2.1709E+03, 2.1750E+03,
     +  2.1792E+03, 2.1833E+03, 2.1875E+03, 2.1917E+03, 2.1958E+03,
     + 2.2000E+03, 2.2042E+03, 2.2083E+03, 2.2125E+03, 2.2167E+03,
     +  2.2209E+03, 2.2251E+03, 2.2293E+03, 2.2335E+03, 2.2377E+03,
     + 2.2419E+03, 2.2461E+03, 2.2503E+03, 2.2545E+03, 2.2588E+03,
     +  2.2630E+03, 2.2672E+03, 2.2714E+03, 2.2757E+03, 2.2799E+03,
     + 2.2842E+03, 2.2884E+03, 2.2927E+03, 2.2970E+03, 2.3012E+03,
     +  2.3055E+03, 2.3098E+03, 2.3140E+03, 2.3183E+03, 2.3226E+03,
     + 2.3269E+03, 2.3312E+03, 2.3355E+03, 2.3398E+03, 2.3441E+03,
     +  2.3484E+03, 2.3527E+03, 2.3570E+03, 2.3614E+03, 2.3657E+03,
     + 2.3700E+03, 2.3744E+03, 2.3787E+03, 2.3830E+03, 2.3874E+03,
     +  2.3917E+03, 2.3961E+03, 2.4004E+03, 2.4048E+03, 2.4092E+03,
     + 2.4135E+03, 2.4179E+03, 2.4223E+03, 2.4267E+03, 2.4311E+03,
     +  2.4355E+03, 2.4399E+03, 2.4443E+03, 2.4487E+03, 2.4531E+03,
     + 2.4575E+03, 2.4619E+03, 2.4663E+03, 2.4707E+03, 2.4752E+03,
     +  2.4796E+03, 2.4840E+03, 2.4885E+03, 2.4929E+03, 2.4974E+03,
     + 2.5018E+03, 2.5063E+03, 2.5108E+03, 2.5152E+03, 2.5197E+03,
     +  2.5242E+03, 2.5286E+03, 2.5331E+03, 2.5376E+03, 2.5421E+03,
     + 2.5466E+03, 2.5511E+03, 2.5556E+03, 2.5601E+03, 2.5646E+03,
     +  2.5691E+03, 2.5737E+03, 2.5782E+03, 2.5827E+03, 2.5872E+03,
     + 2.5918E+03, 2.5963E+03, 2.6009E+03, 2.6054E+03, 2.6100E+03,
     +  2.6145E+03, 2.6191E+03, 2.6236E+03, 2.6282E+03, 2.6328E+03,
     + 2.6373E+03, 2.6419E+03, 2.6465E+03, 2.6511E+03, 2.6557E+03,
     +  2.6603E+03, 2.6649E+03, 2.6695E+03, 2.6741E+03, 2.6787E+03,
     + 2.6833E+03, 2.6880E+03, 2.6926E+03, 2.6972E+03, 2.7019E+03,
     +  2.7065E+03, 2.7111E+03, 2.7158E+03, 2.7204E+03, 2.7251E+03,
     + 2.7297E+03, 2.7344E+03, 2.7391E+03, 2.7437E+03, 2.7484E+03,
     +  2.7531E+03, 2.7578E+03, 2.7625E+03, 2.7672E+03, 2.7719E+03,
     + 2.7766E+03, 2.7813E+03, 2.7860E+03, 2.7907E+03, 2.7954E+03,
     +  2.8001E+03, 2.8049E+03, 2.8096E+03, 2.8143E+03, 2.8191E+03,
     + 2.8238E+03, 2.8285E+03, 2.8333E+03, 2.8380E+03, 2.8428E+03,
     +  2.8476E+03, 2.8523E+03, 2.8571E+03, 2.8619E+03, 2.8667E+03,
     + 2.8714E+03, 2.8762E+03, 2.8810E+03, 2.8858E+03, 2.8906E+03,
     +  2.8954E+03, 2.9002E+03, 2.9050E+03, 2.9098E+03, 2.9147E+03,
     + 2.9195E+03, 2.9243E+03, 2.9292E+03, 2.9340E+03, 2.9388E+03,
     +  2.9437E+03, 2.9485E+03, 2.9534E+03, 2.9582E+03, 2.9631E+03,
     + 2.9680E+03, 2.9728E+03, 2.9777E+03, 2.9826E+03, 2.9875E+03,
     +  2.9923E+03, 2.9972E+03, 3.0021E+03, 3.0070E+03, 3.0119E+03,
     + 3.0168E+03, 3.0218E+03, 3.0267E+03, 3.0316E+03, 3.0365E+03,
     +  3.0414E+03, 3.0464E+03, 3.0513E+03, 3.0562E+03, 3.0612E+03,
     + 3.0661E+03, 3.0711E+03, 3.0760E+03, 3.0810E+03, 3.0860E+03,
     +  3.0909E+03, 3.0959E+03, 3.1009E+03, 3.1059E+03, 3.1109E+03,
     + 3.1158E+03, 3.1208E+03, 3.1258E+03, 3.1308E+03, 3.1358E+03,
     +  3.1409E+03, 3.1459E+03, 3.1509E+03, 3.1559E+03, 3.1609E+03,
     + 3.1660E+03, 3.1710E+03, 3.1760E+03, 3.1811E+03, 3.1861E+03,
     +  3.1912E+03, 3.1962E+03, 3.2013E+03, 3.2064E+03, 3.2114E+03,
     + 3.2165E+03, 3.2216E+03, 3.2267E+03, 3.2318E+03, 3.2368E+03,
     +  3.2419E+03, 3.2470E+03, 3.2521E+03, 3.2572E+03, 3.2624E+03,
     + 3.2675E+03, 3.2726E+03, 3.2777E+03, 3.2828E+03, 3.2880E+03,
     +  3.2931E+03, 3.2982E+03, 3.3034E+03, 3.3085E+03, 3.3137E+03,
     + 3.3188E+03, 3.3240E+03, 3.3292E+03, 3.3343E+03, 3.3395E+03,
     +  3.3447E+03, 3.3499E+03, 3.3551E+03, 3.3602E+03, 3.3654E+03,
     + 3.3706E+03, 3.3758E+03, 3.3810E+03, 3.3862E+03, 3.3915E+03,
     +  3.3967E+03, 3.4019E+03, 3.4071E+03, 3.4124E+03, 3.4176E+03,
     + 3.4228E+03, 3.4281E+03, 3.4333E+03, 3.4386E+03, 3.4438E+03,
     +  3.4491E+03, 3.4544E+03, 3.4596E+03, 3.4649E+03, 3.4702E+03,
     + 3.4755E+03, 3.4807E+03, 3.4860E+03, 3.4913E+03, 3.4966E+03,
     +  3.5019E+03, 3.5072E+03, 3.5125E+03, 3.5179E+03, 3.5232E+03,
     + 3.5285E+03, 3.5338E+03, 3.5392E+03, 3.5445E+03, 3.5498E+03,
     +  3.5552E+03, 3.5605E+03, 3.5659E+03, 3.5712E+03, 3.5766E+03,
     + 3.5819E+03, 3.5873E+03, 3.5927E+03, 3.5981E+03, 3.6034E+03,
     +  3.6088E+03, 3.6142E+03, 3.6196E+03, 3.6250E+03, 3.6304E+03,
     + 3.6358E+03, 3.6412E+03, 3.6466E+03, 3.6521E+03, 3.6575E+03,
     +  3.6629E+03, 3.6683E+03, 3.6738E+03, 3.6792E+03, 3.6847E+03,
     + 3.6901E+03, 3.6956E+03, 3.7010E+03, 3.7065E+03, 3.7119E+03,
     +  3.7174E+03, 3.7229E+03, 3.7284E+03, 3.7338E+03, 3.7393E+03,
     + 3.7448E+03, 3.7503E+03, 3.7558E+03, 3.7613E+03, 3.7668E+03,
     +  3.7723E+03, 3.7778E+03, 3.7834E+03, 3.7889E+03, 3.7944E+03,
     + 3.8000E+03, 3.8055E+03, 3.8110E+03, 3.8166E+03, 3.8221E+03,
     +  3.8277E+03, 3.8332E+03, 3.8388E+03, 3.8444E+03, 3.8499E+03,
     + 3.8555E+03, 3.8611E+03, 3.8667E+03, 3.8722E+03, 3.8778E+03,
     +  3.8834E+03, 3.8890E+03, 3.8946E+03, 3.9002E+03, 3.9058E+03,
     + 3.9115E+03, 3.9171E+03, 3.9227E+03, 3.9283E+03, 3.9340E+03,
     +  3.9396E+03, 3.9452E+03, 3.9509E+03, 3.9565E+03, 3.9622E+03,
     + 3.9679E+03, 3.9735E+03, 3.9792E+03, 3.9848E+03, 3.9905E+03,
     +  3.9962E+03, 4.0019E+03, 4.0076E+03, 4.0133E+03, 4.0190E+03,
     + 4.0247E+03, 4.0304E+03, 4.0361E+03, 4.0418E+03, 4.0475E+03,
     +  4.0532E+03, 4.0589E+03, 4.0647E+03, 4.0704E+03, 4.0761E+03,
     + 4.0819E+03, 4.0876E+03, 4.0934E+03, 4.0991E+03, 4.1049E+03,
     +  4.1107E+03, 4.1164E+03, 4.1222E+03, 4.1280E+03, 4.1337E+03,
     + 4.1395E+03, 4.1453E+03, 4.1511E+03, 4.1569E+03, 4.1627E+03,
     +  4.1685E+03, 4.1743E+03, 4.1801E+03, 4.1860E+03, 4.1918E+03,
     + 4.1976E+03, 4.2034E+03, 4.2093E+03, 4.2151E+03, 4.2209E+03,
     +  4.2268E+03, 4.2326E+03, 4.2385E+03, 4.2444E+03, 4.2502E+03,
     + 4.2561E+03, 4.2620E+03, 4.2678E+03, 4.2737E+03, 4.2796E+03,
     +  4.2855E+03, 4.2914E+03, 4.2973E+03, 4.3032E+03, 4.3091E+03,
     + 4.3150E+03, 4.3209E+03, 4.3268E+03, 4.3328E+03, 4.3387E+03,
     +  4.3446E+03, 4.3505E+03, 4.3565E+03, 4.3624E+03, 4.3684E+03,
     + 4.3743E+03, 4.3803E+03, 4.3862E+03, 4.3922E+03, 4.3982E+03,
     +  4.4042E+03, 4.4101E+03, 4.4161E+03, 4.4221E+03, 4.4281E+03,
     + 4.4341E+03, 4.4401E+03, 4.4461E+03, 4.4521E+03, 4.4581E+03,
     +  4.4641E+03, 4.4701E+03, 4.4762E+03, 4.4822E+03, 4.4882E+03,
     + 4.4943E+03, 4.5003E+03, 4.5063E+03, 4.5124E+03, 4.5184E+03,
     +  4.5245E+03, 4.5306E+03, 4.5366E+03, 4.5427E+03, 4.5488E+03,
     + 4.5549E+03, 4.5609E+03, 4.5670E+03, 4.5731E+03, 4.5792E+03,
     +  4.5853E+03, 4.5914E+03, 4.5975E+03, 4.6036E+03, 4.6098E+03,
     + 4.6159E+03, 4.6220E+03, 4.6281E+03, 4.6343E+03, 4.6404E+03,
     +  4.6466E+03, 4.6527E+03, 4.6589E+03, 4.6650E+03, 4.6712E+03,
     + 4.6773E+03, 4.6835E+03, 4.6897E+03, 4.6958E+03, 4.7020E+03,
     +  4.7082E+03, 4.7144E+03, 4.7206E+03, 4.7268E+03, 4.7330E+03,
     + 4.7392E+03, 4.7454E+03, 4.7516E+03, 4.7578E+03, 4.7641E+03,
     +  4.7703E+03, 4.7765E+03, 4.7828E+03, 4.7890E+03, 4.7952E+03,
     + 4.8015E+03, 4.8077E+03, 4.8140E+03, 4.8203E+03, 4.8265E+03,
     +  4.8328E+03, 4.8391E+03, 4.8454E+03, 4.8516E+03, 4.8579E+03,
     + 4.8642E+03, 4.8705E+03, 4.8768E+03, 4.8831E+03, 4.8894E+03,
     +  4.8957E+03, 4.9020E+03, 4.9084E+03, 4.9147E+03, 4.9210E+03,
     + 4.9274E+03, 4.9337E+03, 4.9400E+03, 4.9464E+03, 4.9527E+03,
     +  4.9591E+03, 4.9654E+03, 4.9718E+03, 4.9782E+03, 4.9845E+03,
     + 4.9909E+03, 4.9973E+03, 5.0037E+03, 5.0101E+03, 5.0165E+03,
     +  5.0229E+03, 5.0293E+03, 5.0357E+03, 5.0421E+03, 5.0485E+03,
     + 5.0549E+03, 5.0613E+03, 5.0678E+03, 5.0742E+03, 5.0806E+03,
     +  5.0871E+03, 5.0935E+03, 5.1000E+03, 5.1064E+03, 5.1129E+03,
     + 5.1193E+03, 5.1258E+03, 5.1323E+03, 5.1387E+03, 5.1452E+03,
     +  5.1517E+03, 5.1582E+03, 5.1647E+03, 5.1712E+03, 5.1777E+03,
     + 5.1842E+03, 5.1907E+03, 5.1972E+03, 5.2037E+03, 5.2102E+03,
     +  5.2168E+03, 5.2233E+03, 5.2298E+03, 5.2364E+03, 5.2429E+03,
     + 5.2494E+03, 5.2560E+03, 5.2626E+03, 5.2691E+03, 5.2757E+03,
     +  5.2822E+03, 5.2888E+03, 5.2954E+03, 5.3020E+03, 5.3086E+03,
     + 5.3151E+03, 5.3217E+03, 5.3283E+03, 5.3349E+03, 5.3415E+03,
     +  5.3482E+03, 5.3548E+03, 5.3614E+03, 5.3680E+03, 5.3746E+03,
     + 5.3813E+03, 5.3879E+03, 5.3945E+03, 5.4012E+03, 5.4078E+03,
     +  5.4145E+03, 5.4211E+03, 5.4278E+03, 5.4345E+03, 5.4411E+03,
     + 5.4478E+03, 5.4545E+03, 5.4612E+03, 5.4679E+03, 5.4746E+03,
     +  5.4813E+03, 5.4880E+03, 5.4947E+03, 5.5014E+03, 5.5081E+03,
     + 5.5148E+03, 5.5215E+03, 5.5282E+03, 5.5350E+03, 5.5417E+03,
     +  5.5484E+03, 5.5552E+03, 5.5619E+03, 5.5687E+03, 5.5754E+03,
     + 5.5822E+03, 5.5890E+03, 5.5957E+03, 5.6025E+03, 5.6093E+03,
     +  5.6161E+03, 5.6229E+03, 5.6296E+03, 5.6364E+03, 5.6432E+03,
     + 5.6500E+03, 5.6568E+03, 5.6637E+03, 5.6705E+03, 5.6773E+03,
     +  5.6841E+03, 5.6909E+03, 5.6978E+03, 5.7046E+03, 5.7115E+03,
     + 5.7183E+03, 5.7251E+03, 5.7320E+03, 5.7389E+03, 5.7457E+03,
     +  5.7526E+03, 5.7595E+03, 5.7663E+03, 5.7732E+03, 5.7801E+03,
     + 5.7870E+03, 5.7939E+03, 5.8008E+03, 5.8077E+03, 5.8146E+03,
     +  5.8215E+03, 5.8284E+03, 5.8353E+03, 5.8423E+03, 5.8492E+03,
     + 5.8561E+03, 5.8630E+03, 5.8700E+03, 5.8769E+03, 5.8839E+03,
     +  5.8908E+03, 5.8978E+03, 5.9047E+03, 5.9117E+03, 5.9187E+03,
     + 5.9257E+03, 5.9326E+03, 5.9396E+03, 5.9466E+03, 5.9536E+03,
     +  5.9606E+03, 5.9676E+03, 5.9746E+03, 5.9816E+03, 5.9886E+03,
     + 5.9956E+03, 6.0027E+03, 6.0097E+03, 6.0167E+03, 6.0237E+03,
     +  6.0308E+03, 6.0378E+03, 6.0449E+03, 6.0519E+03, 6.0590E+03,
     + 6.0660E+03, 6.0731E+03, 6.0802E+03, 6.0873E+03, 6.0943E+03,
     +  6.1014E+03, 6.1085E+03, 6.1156E+03, 6.1227E+03, 6.1298E+03,
     + 6.1369E+03, 6.1440E+03, 6.1511E+03, 6.1582E+03, 6.1653E+03,
     +  6.1725E+03, 6.1796E+03, 6.1867E+03, 6.1939E+03, 6.2010E+03,
     + 6.2082E+03, 6.2153E+03, 6.2225E+03, 6.2296E+03, 6.2368E+03,
     +  6.2439E+03, 6.2511E+03, 6.2583E+03, 6.2655E+03, 6.2727E+03,
     + 6.2799E+03, 6.2870E+03, 6.2942E+03, 6.3014E+03, 6.3087E+03,
     +  6.3159E+03, 6.3231E+03, 6.3303E+03, 6.3375E+03, 6.3448E+03,
     + 6.3520E+03, 6.3592E+03, 6.3665E+03, 6.3737E+03, 6.3810E+03,
     +  6.3882E+03, 6.3955E+03, 6.4027E+03, 6.4100E+03, 6.4173E+03,
     + 6.4245E+03, 6.4318E+03, 6.4391E+03, 6.4464E+03, 6.4537E+03,
     +  6.4610E+03, 6.4683E+03, 6.4756E+03, 6.4829E+03, 6.4902E+03,
     + 6.4975E+03, 6.5049E+03, 6.5122E+03, 6.5195E+03, 6.5269E+03,
     +  6.5342E+03, 6.5415E+03, 6.5489E+03, 6.5562E+03, 6.5636E+03,
     + 6.5710E+03, 6.5783E+03, 6.5857E+03, 6.5931E+03, 6.6005E+03,
     +  6.6078E+03, 6.6152E+03, 6.6226E+03, 6.6300E+03, 6.6374E+03,
     + 6.6448E+03, 6.6522E+03, 6.6596E+03, 6.6671E+03, 6.6745E+03,
     +  6.6819E+03, 6.6893E+03, 6.6968E+03, 6.7042E+03, 6.7117E+03,
     + 6.7191E+03, 6.7266E+03, 6.7340E+03, 6.7415E+03, 6.7490E+03,
     +  6.7564E+03, 6.7639E+03, 6.7714E+03, 6.7789E+03, 6.7863E+03,
     + 6.7938E+03, 6.8013E+03, 6.8088E+03, 6.8163E+03, 6.8239E+03,
     +  6.8314E+03, 6.8389E+03, 6.8464E+03, 6.8539E+03, 6.8615E+03,
     + 6.8690E+03, 6.8765E+03, 6.8841E+03, 6.8916E+03, 6.8992E+03,
     +  6.9067E+03, 6.9143E+03, 6.9219E+03, 6.9294E+03, 6.9370E+03,
     + 6.9446E+03, 6.9522E+03, 6.9598E+03, 6.9673E+03, 6.9749E+03,
     +  6.9825E+03, 6.9902E+03, 6.9978E+03, 7.0054E+03, 7.0130E+03,
     + 7.0206E+03, 7.0282E+03, 7.0359E+03, 7.0435E+03, 7.0511E+03,
     +  7.0588E+03, 7.0664E+03, 7.0741E+03, 7.0817E+03, 7.0894E+03,
     + 7.0971E+03, 7.1047E+03, 7.1124E+03, 7.1201E+03, 7.1278E+03,
     +  7.1355E+03, 7.1432E+03, 7.1509E+03, 7.1586E+03, 7.1663E+03,
     + 7.1740E+03, 7.1817E+03, 7.1894E+03, 7.1971E+03, 7.2048E+03,
     +  7.2126E+03, 7.2203E+03, 7.2280E+03, 7.2358E+03, 7.2435E+03,
     + 7.2513E+03, 7.2590E+03, 7.2668E+03, 7.2746E+03, 7.2823E+03,
     +  7.2901E+03, 7.2979E+03, 7.3057E+03, 7.3135E+03, 7.3213E+03,
     + 7.3291E+03, 7.3369E+03, 7.3447E+03, 7.3525E+03, 7.3603E+03,
     +  7.3681E+03, 7.3759E+03, 7.3837E+03, 7.3916E+03, 7.3994E+03,
     + 7.4073E+03, 7.4151E+03, 7.4229E+03, 7.4308E+03, 7.4387E+03,
     +  7.4465E+03, 7.4544E+03, 7.4623E+03, 7.4701E+03, 7.4780E+03,
     + 7.4859E+03, 7.4938E+03, 7.5017E+03, 7.5096E+03, 7.5175E+03,
     +  7.5254E+03, 7.5333E+03, 7.5412E+03, 7.5491E+03, 7.5570E+03,
     + 7.5650E+03, 7.5729E+03, 7.5808E+03, 7.5888E+03, 7.5967E+03,
     +  7.6047E+03, 7.6126E+03, 7.6206E+03, 7.6285E+03, 7.6365E+03,
     + 7.6445E+03, 7.6524E+03, 7.6604E+03, 7.6684E+03, 7.6764E+03,
     +  7.6844E+03, 7.6924E+03, 7.7004E+03, 7.7084E+03, 7.7164E+03,
     + 7.7244E+03, 7.7324E+03, 7.7404E+03, 7.7485E+03, 7.7565E+03,
     +  7.7645E+03, 7.7726E+03, 7.7806E+03, 7.7887E+03, 7.7967E+03,
     + 7.8048E+03, 7.8128E+03, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        32
      data (QofT( 3,J),J=1,2408)/ 2.6650E+00, 4.3546E+00, 6.1271E+00,
     +  7.9155E+00, 9.7099E+00, 1.1507E+01, 1.3306E+01, 1.5106E+01,
     + 1.6906E+01, 1.8708E+01, 2.0509E+01, 2.2311E+01, 2.4113E+01,
     +  2.5915E+01, 2.7717E+01, 2.9519E+01, 3.1322E+01, 3.3125E+01,
     + 3.4927E+01, 3.6730E+01, 4.0336E+01, 4.3942E+01, 4.7549E+01,
     +  5.1156E+01, 5.4763E+01, 5.8370E+01, 6.1977E+01, 6.5585E+01,
     + 6.9193E+01, 7.2801E+01, 7.6409E+01, 8.0018E+01, 8.3627E+01,
     +  8.7235E+01, 9.0845E+01, 9.4454E+01, 9.8063E+01, 1.0167E+02,
     + 1.0528E+02, 1.0889E+02, 1.1250E+02, 1.1611E+02, 1.1972E+02,
     +  1.2333E+02, 1.2695E+02, 1.3056E+02, 1.3417E+02, 1.3778E+02,
     + 1.4139E+02, 1.4500E+02, 1.4862E+02, 1.5223E+02, 1.5584E+02,
     +  1.5945E+02, 1.6307E+02, 1.6668E+02, 1.7029E+02, 1.7391E+02,
     + 1.7752E+02, 1.8113E+02, 1.8475E+02, 1.8836E+02, 1.9198E+02,
     +  1.9559E+02, 1.9920E+02, 2.0282E+02, 2.0643E+02, 2.1005E+02,
     + 2.1367E+02, 2.1728E+02, 2.2090E+02, 2.2451E+02, 2.2813E+02,
     +  2.3175E+02, 2.3536E+02, 2.3898E+02, 2.4260E+02, 2.4621E+02,
     + 2.4983E+02, 2.5345E+02, 2.5707E+02, 2.6068E+02, 2.6430E+02,
     +  2.6792E+02, 2.7154E+02, 2.7516E+02, 2.7878E+02, 2.8240E+02,
     + 2.8602E+02, 2.8964E+02, 2.9326E+02, 2.9688E+02, 3.0050E+02,
     +  3.0412E+02, 3.0774E+02, 3.1136E+02, 3.1499E+02, 3.1861E+02,
     + 3.2223E+02, 3.2585E+02, 3.2948E+02, 3.3310E+02, 3.3672E+02,
     +  3.4035E+02, 3.4397E+02, 3.4760E+02, 3.5123E+02, 3.5485E+02,
     + 3.5848E+02, 3.6211E+02, 3.6574E+02, 3.6937E+02, 3.7300E+02,
     +  3.7663E+02, 3.8026E+02, 3.8389E+02, 3.8752E+02, 3.9115E+02,
     + 3.9479E+02, 3.9842E+02, 4.0206E+02, 4.0569E+02, 4.0933E+02,
     +  4.1297E+02, 4.1661E+02, 4.2025E+02, 4.2389E+02, 4.2753E+02,
     + 4.3118E+02, 4.3482E+02, 4.3847E+02, 4.4211E+02, 4.4576E+02,
     +  4.4941E+02, 4.5306E+02, 4.5672E+02, 4.6037E+02, 4.6402E+02,
     + 4.6768E+02, 4.7134E+02, 4.7500E+02, 4.7866E+02, 4.8232E+02,
     +  4.8599E+02, 4.8966E+02, 4.9332E+02, 4.9699E+02, 5.0067E+02,
     + 5.0434E+02, 5.0802E+02, 5.1169E+02, 5.1537E+02, 5.1906E+02,
     +  5.2274E+02, 5.2643E+02, 5.3012E+02, 5.3381E+02, 5.3750E+02,
     + 5.4120E+02, 5.4489E+02, 5.4860E+02, 5.5230E+02, 5.5600E+02,
     +  5.5971E+02, 5.6342E+02, 5.6714E+02, 5.7085E+02, 5.7457E+02,
     + 5.7830E+02, 5.8202E+02, 5.8575E+02, 5.8948E+02, 5.9321E+02,
     +  5.9695E+02, 6.0069E+02, 6.0443E+02, 6.0818E+02, 6.1193E+02,
     + 6.1568E+02, 6.1944E+02, 6.2320E+02, 6.2696E+02, 6.3073E+02,
     +  6.3450E+02, 6.3827E+02, 6.4205E+02, 6.4583E+02, 6.4961E+02,
     + 6.5340E+02, 6.5719E+02, 6.6099E+02, 6.6479E+02, 6.6859E+02,
     +  6.7240E+02, 6.7621E+02, 6.8003E+02, 6.8385E+02, 6.8767E+02,
     + 6.9150E+02, 6.9533E+02, 6.9917E+02, 7.0301E+02, 7.0685E+02,
     +  7.1070E+02, 7.1455E+02, 7.1841E+02, 7.2227E+02, 7.2614E+02,
     + 7.3001E+02, 7.3389E+02, 7.3777E+02, 7.4165E+02, 7.4554E+02,
     +  7.4944E+02, 7.5334E+02, 7.5724E+02, 7.6115E+02, 7.6506E+02,
     + 7.6898E+02, 7.7290E+02, 7.7683E+02, 7.8077E+02, 7.8470E+02,
     +  7.8865E+02, 7.9260E+02, 7.9655E+02, 8.0051E+02, 8.0447E+02,
     + 8.0844E+02, 8.1241E+02, 8.1639E+02, 8.2038E+02, 8.2437E+02,
     +  8.2836E+02, 8.3237E+02, 8.3637E+02, 8.4038E+02, 8.4440E+02,
     + 8.4842E+02, 8.5245E+02, 8.5649E+02, 8.6053E+02, 8.6457E+02,
     +  8.6862E+02, 8.7268E+02, 8.7674E+02, 8.8081E+02, 8.8488E+02,
     + 8.8896E+02, 8.9305E+02, 8.9714E+02, 9.0124E+02, 9.0534E+02,
     +  9.0945E+02, 9.1356E+02, 9.1768E+02, 9.2181E+02, 9.2594E+02,
     + 9.3008E+02, 9.3423E+02, 9.3838E+02, 9.4253E+02, 9.4670E+02,
     +  9.5087E+02, 9.5504E+02, 9.5922E+02, 9.6341E+02, 9.6761E+02,
     + 9.7181E+02, 9.7601E+02, 9.8023E+02, 9.8445E+02, 9.8867E+02,
     +  9.9290E+02, 9.9714E+02, 1.0014E+03, 1.0056E+03, 1.0099E+03,
     + 1.0142E+03, 1.0184E+03, 1.0227E+03, 1.0270E+03, 1.0313E+03,
     +  1.0356E+03, 1.0399E+03, 1.0442E+03, 1.0485E+03, 1.0528E+03,
     + 1.0572E+03, 1.0615E+03, 1.0658E+03, 1.0702E+03, 1.0745E+03,
     +  1.0789E+03, 1.0833E+03, 1.0877E+03, 1.0920E+03, 1.0964E+03,
     + 1.1008E+03, 1.1052E+03, 1.1096E+03, 1.1141E+03, 1.1185E+03,
     +  1.1229E+03, 1.1274E+03, 1.1318E+03, 1.1363E+03, 1.1407E+03,
     + 1.1452E+03, 1.1497E+03, 1.1541E+03, 1.1586E+03, 1.1631E+03,
     +  1.1676E+03, 1.1721E+03, 1.1766E+03, 1.1812E+03, 1.1857E+03,
     + 1.1902E+03, 1.1948E+03, 1.1993E+03, 1.2039E+03, 1.2084E+03,
     +  1.2130E+03, 1.2176E+03, 1.2222E+03, 1.2268E+03, 1.2314E+03,
     + 1.2360E+03, 1.2406E+03, 1.2452E+03, 1.2499E+03, 1.2545E+03,
     +  1.2591E+03, 1.2638E+03, 1.2685E+03, 1.2731E+03, 1.2778E+03,
     + 1.2825E+03, 1.2872E+03, 1.2919E+03, 1.2966E+03, 1.3013E+03,
     +  1.3060E+03, 1.3107E+03, 1.3154E+03, 1.3202E+03, 1.3249E+03,
     + 1.3297E+03, 1.3345E+03, 1.3392E+03, 1.3440E+03, 1.3488E+03,
     +  1.3536E+03, 1.3584E+03, 1.3632E+03, 1.3680E+03, 1.3728E+03,
     + 1.3776E+03, 1.3825E+03, 1.3873E+03, 1.3922E+03, 1.3970E+03,
     +  1.4019E+03, 1.4068E+03, 1.4117E+03, 1.4166E+03, 1.4215E+03,
     + 1.4264E+03, 1.4313E+03, 1.4362E+03, 1.4411E+03, 1.4461E+03,
     +  1.4510E+03, 1.4560E+03, 1.4609E+03, 1.4659E+03, 1.4709E+03,
     + 1.4758E+03, 1.4808E+03, 1.4858E+03, 1.4908E+03, 1.4958E+03,
     +  1.5009E+03, 1.5059E+03, 1.5109E+03, 1.5160E+03, 1.5210E+03,
     + 1.5261E+03, 1.5311E+03, 1.5362E+03, 1.5413E+03, 1.5464E+03,
     +  1.5515E+03, 1.5566E+03, 1.5617E+03, 1.5668E+03, 1.5720E+03,
     + 1.5771E+03, 1.5822E+03, 1.5874E+03, 1.5926E+03, 1.5977E+03,
     +  1.6029E+03, 1.6081E+03, 1.6133E+03, 1.6185E+03, 1.6237E+03,
     + 1.6289E+03, 1.6341E+03, 1.6394E+03, 1.6446E+03, 1.6498E+03,
     +  1.6551E+03, 1.6604E+03, 1.6656E+03, 1.6709E+03, 1.6762E+03,
     + 1.6815E+03, 1.6868E+03, 1.6921E+03, 1.6974E+03, 1.7027E+03,
     +  1.7081E+03, 1.7134E+03, 1.7188E+03, 1.7241E+03, 1.7295E+03,
     + 1.7349E+03, 1.7402E+03, 1.7456E+03, 1.7510E+03, 1.7564E+03,
     +  1.7619E+03, 1.7673E+03, 1.7727E+03, 1.7781E+03, 1.7836E+03,
     + 1.7890E+03, 1.7945E+03, 1.8000E+03, 1.8055E+03, 1.8109E+03,
     +  1.8164E+03, 1.8219E+03, 1.8274E+03, 1.8330E+03, 1.8385E+03,
     + 1.8440E+03, 1.8496E+03, 1.8551E+03, 1.8607E+03, 1.8662E+03,
     +  1.8718E+03, 1.8774E+03, 1.8830E+03, 1.8886E+03, 1.8942E+03,
     + 1.8998E+03, 1.9054E+03, 1.9111E+03, 1.9167E+03, 1.9223E+03,
     +  1.9280E+03, 1.9337E+03, 1.9393E+03, 1.9450E+03, 1.9507E+03,
     + 1.9564E+03, 1.9621E+03, 1.9678E+03, 1.9735E+03, 1.9793E+03,
     +  1.9850E+03, 1.9907E+03, 1.9965E+03, 2.0023E+03, 2.0080E+03,
     + 2.0138E+03, 2.0196E+03, 2.0254E+03, 2.0312E+03, 2.0370E+03,
     +  2.0428E+03, 2.0486E+03, 2.0545E+03, 2.0603E+03, 2.0662E+03,
     + 2.0720E+03, 2.0779E+03, 2.0838E+03, 2.0896E+03, 2.0955E+03,
     +  2.1014E+03, 2.1073E+03, 2.1133E+03, 2.1192E+03, 2.1251E+03,
     + 2.1311E+03, 2.1370E+03, 2.1430E+03, 2.1489E+03, 2.1549E+03,
     +  2.1609E+03, 2.1669E+03, 2.1729E+03, 2.1789E+03, 2.1849E+03,
     + 2.1909E+03, 2.1970E+03, 2.2030E+03, 2.2090E+03, 2.2151E+03,
     +  2.2212E+03, 2.2272E+03, 2.2333E+03, 2.2394E+03, 2.2455E+03,
     + 2.2516E+03, 2.2577E+03, 2.2638E+03, 2.2700E+03, 2.2761E+03,
     +  2.2823E+03, 2.2884E+03, 2.2946E+03, 2.3007E+03, 2.3069E+03,
     + 2.3131E+03, 2.3193E+03, 2.3255E+03, 2.3317E+03, 2.3380E+03,
     +  2.3442E+03, 2.3504E+03, 2.3567E+03, 2.3629E+03, 2.3692E+03,
     + 2.3755E+03, 2.3817E+03, 2.3880E+03, 2.3943E+03, 2.4006E+03,
     +  2.4069E+03, 2.4133E+03, 2.4196E+03, 2.4259E+03, 2.4323E+03,
     + 2.4386E+03, 2.4450E+03, 2.4514E+03, 2.4577E+03, 2.4641E+03,
     +  2.4705E+03, 2.4769E+03, 2.4833E+03, 2.4898E+03, 2.4962E+03,
     + 2.5026E+03, 2.5091E+03, 2.5155E+03, 2.5220E+03, 2.5285E+03,
     +  2.5349E+03, 2.5414E+03, 2.5479E+03, 2.5544E+03, 2.5609E+03,
     + 2.5675E+03, 2.5740E+03, 2.5805E+03, 2.5871E+03, 2.5936E+03,
     +  2.6002E+03, 2.6068E+03, 2.6134E+03, 2.6199E+03, 2.6265E+03,
     + 2.6331E+03, 2.6398E+03, 2.6464E+03, 2.6530E+03, 2.6597E+03,
     +  2.6663E+03, 2.6730E+03, 2.6796E+03, 2.6863E+03, 2.6930E+03,
     + 2.6997E+03, 2.7064E+03, 2.7131E+03, 2.7198E+03, 2.7265E+03,
     +  2.7332E+03, 2.7400E+03, 2.7467E+03, 2.7535E+03, 2.7602E+03,
     + 2.7670E+03, 2.7738E+03, 2.7806E+03, 2.7874E+03, 2.7942E+03,
     +  2.8010E+03, 2.8078E+03, 2.8147E+03, 2.8215E+03, 2.8284E+03,
     + 2.8352E+03, 2.8421E+03, 2.8490E+03, 2.8558E+03, 2.8627E+03,
     +  2.8696E+03, 2.8765E+03, 2.8835E+03, 2.8904E+03, 2.8973E+03,
     + 2.9043E+03, 2.9112E+03, 2.9182E+03, 2.9251E+03, 2.9321E+03,
     +  2.9391E+03, 2.9461E+03, 2.9531E+03, 2.9601E+03, 2.9671E+03,
     + 2.9741E+03, 2.9812E+03, 2.9882E+03, 2.9953E+03, 3.0023E+03,
     +  3.0094E+03, 3.0165E+03, 3.0236E+03, 3.0307E+03, 3.0378E+03,
     + 3.0449E+03, 3.0520E+03, 3.0591E+03, 3.0663E+03, 3.0734E+03,
     +  3.0806E+03, 3.0877E+03, 3.0949E+03, 3.1021E+03, 3.1093E+03,
     + 3.1164E+03, 3.1237E+03, 3.1309E+03, 3.1381E+03, 3.1453E+03,
     +  3.1526E+03, 3.1598E+03, 3.1671E+03, 3.1743E+03, 3.1816E+03,
     + 3.1889E+03, 3.1962E+03, 3.2035E+03, 3.2108E+03, 3.2181E+03,
     +  3.2254E+03, 3.2327E+03, 3.2401E+03, 3.2474E+03, 3.2548E+03,
     + 3.2622E+03, 3.2695E+03, 3.2769E+03, 3.2843E+03, 3.2917E+03,
     +  3.2991E+03, 3.3065E+03, 3.3140E+03, 3.3214E+03, 3.3288E+03,
     + 3.3363E+03, 3.3437E+03, 3.3512E+03, 3.3587E+03, 3.3662E+03,
     +  3.3737E+03, 3.3812E+03, 3.3887E+03, 3.3962E+03, 3.4037E+03,
     + 3.4113E+03, 3.4188E+03, 3.4264E+03, 3.4339E+03, 3.4415E+03,
     +  3.4491E+03, 3.4567E+03, 3.4643E+03, 3.4719E+03, 3.4795E+03,
     + 3.4871E+03, 3.4947E+03, 3.5024E+03, 3.5100E+03, 3.5177E+03,
     +  3.5253E+03, 3.5330E+03, 3.5407E+03, 3.5484E+03, 3.5561E+03,
     + 3.5638E+03, 3.5715E+03, 3.5792E+03, 3.5870E+03, 3.5947E+03,
     +  3.6024E+03, 3.6102E+03, 3.6180E+03, 3.6258E+03, 3.6335E+03,
     + 3.6413E+03, 3.6491E+03, 3.6569E+03, 3.6648E+03, 3.6726E+03,
     +  3.6804E+03, 3.6883E+03, 3.6961E+03, 3.7040E+03, 3.7119E+03,
     + 3.7197E+03, 3.7276E+03, 3.7355E+03, 3.7434E+03, 3.7513E+03,
     +  3.7593E+03, 3.7672E+03, 3.7751E+03, 3.7831E+03, 3.7910E+03,
     + 3.7990E+03, 3.8070E+03, 3.8149E+03, 3.8229E+03, 3.8309E+03,
     +  3.8389E+03, 3.8470E+03, 3.8550E+03, 3.8630E+03, 3.8711E+03,
     + 3.8791E+03, 3.8872E+03, 3.8952E+03, 3.9033E+03, 3.9114E+03,
     +  3.9195E+03, 3.9276E+03, 3.9357E+03, 3.9438E+03, 3.9519E+03,
     + 3.9601E+03, 3.9682E+03, 3.9764E+03, 3.9845E+03, 3.9927E+03,
     +  4.0009E+03, 4.0091E+03, 4.0173E+03, 4.0255E+03, 4.0337E+03,
     + 4.0419E+03, 4.0502E+03, 4.0584E+03, 4.0666E+03, 4.0749E+03,
     +  4.0832E+03, 4.0914E+03, 4.0997E+03, 4.1080E+03, 4.1163E+03,
     + 4.1246E+03, 4.1329E+03, 4.1413E+03, 4.1496E+03, 4.1579E+03,
     +  4.1663E+03, 4.1747E+03, 4.1830E+03, 4.1914E+03, 4.1998E+03,
     + 4.2082E+03, 4.2166E+03, 4.2250E+03, 4.2334E+03, 4.2419E+03,
     +  4.2503E+03, 4.2587E+03, 4.2672E+03, 4.2757E+03, 4.2841E+03,
     + 4.2926E+03, 4.3011E+03, 4.3096E+03, 4.3181E+03, 4.3266E+03,
     +  4.3351E+03, 4.3437E+03, 4.3522E+03, 4.3608E+03, 4.3693E+03,
     + 4.3779E+03, 4.3865E+03, 4.3951E+03, 4.4037E+03, 4.4123E+03,
     +  4.4209E+03, 4.4295E+03, 4.4381E+03, 4.4468E+03, 4.4554E+03,
     + 4.4641E+03, 4.4727E+03, 4.4814E+03, 4.4901E+03, 4.4988E+03,
     +  4.5075E+03, 4.5162E+03, 4.5249E+03, 4.5336E+03, 4.5423E+03,
     + 4.5511E+03, 4.5598E+03, 4.5686E+03, 4.5774E+03, 4.5861E+03,
     +  4.5949E+03, 4.6037E+03, 4.6125E+03, 4.6213E+03, 4.6301E+03,
     + 4.6390E+03, 4.6478E+03, 4.6566E+03, 4.6655E+03, 4.6744E+03,
     +  4.6832E+03, 4.6921E+03, 4.7010E+03, 4.7099E+03, 4.7188E+03,
     + 4.7277E+03, 4.7366E+03, 4.7456E+03, 4.7545E+03, 4.7635E+03,
     +  4.7724E+03, 4.7814E+03, 4.7904E+03, 4.7994E+03, 4.8083E+03,
     + 4.8173E+03, 4.8264E+03, 4.8354E+03, 4.8444E+03, 4.8534E+03,
     +  4.8625E+03, 4.8715E+03, 4.8806E+03, 4.8897E+03, 4.8988E+03,
     + 4.9078E+03, 4.9169E+03, 4.9260E+03, 4.9352E+03, 4.9443E+03,
     +  4.9534E+03, 4.9626E+03, 4.9717E+03, 4.9809E+03, 4.9900E+03,
     + 4.9992E+03, 5.0084E+03, 5.0176E+03, 5.0268E+03, 5.0360E+03,
     +  5.0452E+03, 5.0544E+03, 5.0637E+03, 5.0729E+03, 5.0822E+03,
     + 5.0914E+03, 5.1007E+03, 5.1100E+03, 5.1193E+03, 5.1286E+03,
     +  5.1379E+03, 5.1472E+03, 5.1565E+03, 5.1659E+03, 5.1752E+03,
     + 5.1845E+03, 5.1939E+03, 5.2033E+03, 5.2126E+03, 5.2220E+03,
     +  5.2314E+03, 5.2408E+03, 5.2502E+03, 5.2597E+03, 5.2691E+03,
     + 5.2785E+03, 5.2880E+03, 5.2974E+03, 5.3069E+03, 5.3164E+03,
     +  5.3258E+03, 5.3353E+03, 5.3448E+03, 5.3543E+03, 5.3638E+03,
     + 5.3734E+03, 5.3829E+03, 5.3924E+03, 5.4020E+03, 5.4116E+03,
     +  5.4211E+03, 5.4307E+03, 5.4403E+03, 5.4499E+03, 5.4595E+03,
     + 5.4691E+03, 5.4787E+03, 5.4883E+03, 5.4980E+03, 5.5076E+03,
     +  5.5173E+03, 5.5270E+03, 5.5366E+03, 5.5463E+03, 5.5560E+03,
     + 5.5657E+03, 5.5754E+03, 5.5851E+03, 5.5949E+03, 5.6046E+03,
     +  5.6143E+03, 5.6241E+03, 5.6338E+03, 5.6436E+03, 5.6534E+03,
     + 5.6632E+03, 5.6730E+03, 5.6828E+03, 5.6926E+03, 5.7024E+03,
     +  5.7122E+03, 5.7221E+03, 5.7319E+03, 5.7418E+03, 5.7517E+03,
     + 5.7615E+03, 5.7714E+03, 5.7813E+03, 5.7912E+03, 5.8011E+03,
     +  5.8110E+03, 5.8210E+03, 5.8309E+03, 5.8408E+03, 5.8508E+03,
     + 5.8608E+03, 5.8707E+03, 5.8807E+03, 5.8907E+03, 5.9007E+03,
     +  5.9107E+03, 5.9207E+03, 5.9307E+03, 5.9408E+03, 5.9508E+03,
     + 5.9609E+03, 5.9709E+03, 5.9810E+03, 5.9911E+03, 6.0012E+03,
     +  6.0112E+03, 6.0213E+03, 6.0315E+03, 6.0416E+03, 6.0517E+03,
     + 6.0618E+03, 6.0720E+03, 6.0821E+03, 6.0923E+03, 6.1025E+03,
     +  6.1127E+03, 6.1229E+03, 6.1331E+03, 6.1433E+03, 6.1535E+03,
     + 6.1637E+03, 6.1739E+03, 6.1842E+03, 6.1944E+03, 6.2047E+03,
     +  6.2150E+03, 6.2253E+03, 6.2355E+03, 6.2458E+03, 6.2561E+03,
     + 6.2665E+03, 6.2768E+03, 6.2871E+03, 6.2974E+03, 6.3078E+03,
     +  6.3182E+03, 6.3285E+03, 6.3389E+03, 6.3493E+03, 6.3597E+03,
     + 6.3701E+03, 6.3805E+03, 6.3909E+03, 6.4013E+03, 6.4118E+03,
     +  6.4222E+03, 6.4327E+03, 6.4431E+03, 6.4536E+03, 6.4641E+03,
     + 6.4746E+03, 6.4851E+03, 6.4956E+03, 6.5061E+03, 6.5166E+03,
     +  6.5272E+03, 6.5377E+03, 6.5483E+03, 6.5588E+03, 6.5694E+03,
     + 6.5800E+03, 6.5906E+03, 6.6012E+03, 6.6118E+03, 6.6224E+03,
     +  6.6330E+03, 6.6436E+03, 6.6543E+03, 6.6649E+03, 6.6756E+03,
     + 6.6862E+03, 6.6969E+03, 6.7076E+03, 6.7183E+03, 6.7290E+03,
     +  6.7397E+03, 6.7504E+03, 6.7612E+03, 6.7719E+03, 6.7826E+03,
     + 6.7934E+03, 6.8042E+03, 6.8149E+03, 6.8257E+03, 6.8365E+03,
     +  6.8473E+03, 6.8581E+03, 6.8689E+03, 6.8798E+03, 6.8906E+03,
     + 6.9014E+03, 6.9123E+03, 6.9231E+03, 6.9340E+03, 6.9449E+03,
     +  6.9558E+03, 6.9667E+03, 6.9776E+03, 6.9885E+03, 6.9994E+03,
     + 7.0104E+03, 7.0213E+03, 7.0322E+03, 7.0432E+03, 7.0542E+03,
     +  7.0651E+03, 7.0761E+03, 7.0871E+03, 7.0981E+03, 7.1091E+03,
     + 7.1202E+03, 7.1312E+03, 7.1422E+03, 7.1533E+03, 7.1643E+03,
     +  7.1754E+03, 7.1865E+03, 7.1976E+03, 7.2086E+03, 7.2197E+03,
     + 7.2309E+03, 7.2420E+03, 7.2531E+03, 7.2642E+03, 7.2754E+03,
     +  7.2865E+03, 7.2977E+03, 7.3089E+03, 7.3200E+03, 7.3312E+03,
     + 7.3424E+03, 7.3536E+03, 7.3649E+03, 7.3761E+03, 7.3873E+03,
     +  7.3986E+03, 7.4098E+03, 7.4211E+03, 7.4323E+03, 7.4436E+03,
     + 7.4549E+03, 7.4662E+03, 7.4775E+03, 7.4888E+03, 7.5001E+03,
     +  7.5115E+03, 7.5228E+03, 7.5342E+03, 7.5455E+03, 7.5569E+03,
     + 7.5683E+03, 7.5796E+03, 7.5910E+03, 7.6024E+03, 7.6138E+03,
     +  7.6253E+03, 7.6367E+03, 7.6481E+03, 7.6596E+03, 7.6710E+03,
     + 7.6825E+03, 7.6940E+03, 7.7055E+03, 7.7169E+03, 7.7284E+03,
     +  7.7400E+03, 7.7515E+03, 7.7630E+03, 7.7745E+03, 7.7861E+03,
     + 7.7976E+03, 7.8092E+03, 7.8208E+03, 7.8323E+03, 7.8439E+03,
     +  7.8555E+03, 7.8671E+03, 7.8788E+03, 7.8904E+03, 7.9020E+03,
     + 7.9137E+03, 7.9253E+03, 7.9370E+03, 7.9486E+03, 7.9603E+03,
     +  7.9720E+03, 7.9837E+03, 7.9954E+03, 8.0071E+03, 8.0188E+03,
     + 8.0306E+03, 8.0423E+03, 8.0541E+03, 8.0658E+03, 8.0776E+03,
     +  8.0894E+03, 8.1012E+03, 8.1129E+03, 8.1248E+03, 8.1366E+03,
     + 8.1484E+03, 8.1602E+03, 8.1721E+03, 8.1839E+03, 8.1958E+03,
     +  8.2076E+03, 8.2195E+03, 8.2314E+03, 8.2433E+03, 8.2552E+03,
     + 8.2671E+03, 8.2790E+03, 8.2909E+03, 8.3029E+03, 8.3148E+03,
     +  8.3268E+03, 8.3387E+03, 8.3507E+03, 8.3627E+03, 8.3747E+03,
     + 8.3867E+03, 8.3987E+03, 8.4107E+03, 8.4227E+03, 8.4348E+03,
     +  8.4468E+03, 8.4589E+03, 8.4709E+03, 8.4830E+03, 8.4951E+03,
     + 8.5072E+03, 8.5193E+03, 8.5314E+03, 8.5435E+03, 8.5556E+03,
     +  8.5677E+03, 8.5799E+03, 8.5920E+03, 8.6042E+03, 8.6164E+03,
     + 8.6285E+03, 8.6407E+03, 8.6529E+03, 8.6651E+03, 8.6773E+03,
     +  8.6896E+03, 8.7018E+03, 8.7140E+03, 8.7263E+03, 8.7385E+03,
     + 8.7508E+03, 8.7631E+03, 8.7754E+03, 8.7877E+03, 8.8000E+03,
     +  8.8123E+03, 8.8246E+03, 8.8369E+03, 8.8493E+03, 8.8616E+03,
     + 8.8740E+03, 8.8864E+03, 8.8987E+03, 8.9111E+03, 8.9235E+03,
     +  8.9359E+03, 8.9483E+03, 8.9607E+03, 8.9732E+03, 8.9856E+03,
     + 8.9981E+03, 9.0105E+03, 9.0230E+03, 9.0354E+03, 9.0479E+03,
     +  9.0604E+03, 9.0729E+03, 9.0854E+03, 9.0980E+03, 9.1105E+03,
     + 9.1230E+03, 9.1356E+03, 9.1481E+03, 9.1607E+03, 9.1733E+03,
     +  9.1858E+03, 9.1984E+03, 9.2110E+03, 9.2236E+03, 9.2362E+03,
     + 9.2489E+03, 9.2615E+03, 9.2742E+03, 9.2868E+03, 9.2995E+03,
     +  9.3121E+03, 9.3248E+03, 9.3375E+03, 9.3502E+03, 9.3629E+03,
     + 9.3756E+03, 9.3884E+03, 9.4011E+03, 9.4138E+03, 9.4266E+03,
     +  9.4394E+03, 9.4521E+03, 9.4649E+03, 9.4777E+03, 9.4905E+03,
     + 9.5033E+03, 9.5161E+03, 9.5289E+03, 9.5418E+03, 9.5546E+03,
     +  9.5675E+03, 9.5803E+03, 9.5932E+03, 9.6061E+03, 9.6190E+03,
     + 9.6319E+03, 9.6448E+03, 9.6577E+03, 9.6706E+03, 9.6835E+03,
     +  9.6965E+03, 9.7094E+03, 9.7224E+03, 9.7353E+03, 9.7483E+03,
     + 9.7613E+03, 9.7743E+03, 9.7873E+03, 9.8003E+03, 9.8133E+03,
     +  9.8264E+03, 9.8394E+03, 9.8525E+03, 9.8655E+03, 9.8786E+03,
     + 9.8917E+03, 9.9048E+03, 9.9178E+03, 9.9310E+03, 9.9441E+03,
     +  9.9572E+03, 9.9703E+03, 9.9835E+03, 9.9966E+03, 1.0010E+04,
     + 1.0023E+04, 1.0036E+04, 1.0049E+04, 1.0062E+04, 1.0076E+04,
     +  1.0089E+04, 1.0102E+04, 1.0115E+04, 1.0129E+04, 1.0142E+04,
     + 1.0155E+04, 1.0168E+04, 1.0182E+04, 1.0195E+04, 1.0208E+04,
     +  1.0222E+04, 1.0235E+04, 1.0248E+04, 1.0261E+04, 1.0275E+04,
     + 1.0288E+04, 1.0302E+04, 1.0315E+04, 1.0328E+04, 1.0342E+04,
     +  1.0355E+04, 1.0368E+04, 1.0382E+04, 1.0395E+04, 1.0409E+04,
     + 1.0422E+04, 1.0436E+04, 1.0449E+04, 1.0462E+04, 1.0476E+04,
     +  1.0489E+04, 1.0503E+04, 1.0516E+04, 1.0530E+04, 1.0543E+04,
     + 1.0557E+04, 1.0571E+04, 1.0584E+04, 1.0598E+04, 1.0611E+04,
     +  1.0625E+04, 1.0638E+04, 1.0652E+04, 1.0666E+04, 1.0679E+04,
     + 1.0693E+04, 1.0706E+04, 1.0720E+04, 1.0734E+04, 1.0747E+04,
     +  1.0761E+04, 1.0775E+04, 1.0788E+04, 1.0802E+04, 1.0816E+04,
     + 1.0829E+04, 1.0843E+04, 1.0857E+04, 1.0871E+04, 1.0884E+04,
     +  1.0898E+04, 1.0912E+04, 1.0926E+04, 1.0939E+04, 1.0953E+04,
     + 1.0967E+04, 1.0981E+04, 1.0995E+04, 1.1009E+04, 1.1022E+04,
     +  1.1036E+04, 1.1050E+04, 1.1064E+04, 1.1078E+04, 1.1092E+04,
     + 1.1106E+04, 1.1119E+04, 1.1133E+04, 1.1147E+04, 1.1161E+04,
     +  1.1175E+04, 1.1189E+04, 1.1203E+04, 1.1217E+04, 1.1231E+04,
     + 1.1245E+04, 1.1259E+04, 1.1273E+04, 1.1287E+04, 1.1301E+04,
     +  1.1315E+04, 1.1329E+04, 1.1343E+04, 1.1357E+04, 1.1371E+04,
     + 1.1385E+04, 1.1399E+04, 1.1413E+04, 1.1428E+04, 1.1442E+04,
     +  1.1456E+04, 1.1470E+04, 1.1484E+04, 1.1498E+04, 1.1512E+04,
     + 1.1526E+04, 1.1541E+04, 1.1555E+04, 1.1569E+04, 1.1583E+04,
     +  1.1597E+04, 1.1612E+04, 1.1626E+04, 1.1640E+04, 1.1654E+04,
     + 1.1669E+04, 1.1683E+04, 1.1697E+04, 1.1711E+04, 1.1726E+04,
     +  1.1740E+04, 1.1754E+04, 1.1769E+04, 1.1783E+04, 1.1797E+04,
     + 1.1812E+04, 1.1826E+04, 1.1840E+04, 1.1855E+04, 1.1869E+04,
     +  1.1884E+04, 1.1898E+04, 1.1912E+04, 1.1927E+04, 1.1941E+04,
     + 1.1956E+04, 1.1970E+04, 1.1985E+04, 1.1999E+04, 1.2013E+04,
     +  1.2028E+04, 1.2042E+04, 1.2057E+04, 1.2071E+04, 1.2086E+04,
     + 1.2100E+04, 1.2115E+04, 1.2130E+04, 1.2144E+04, 1.2159E+04,
     +  1.2173E+04, 1.2188E+04, 1.2202E+04, 1.2217E+04, 1.2232E+04,
     + 1.2246E+04, 1.2261E+04, 1.2276E+04, 1.2290E+04, 1.2305E+04,
     +  1.2319E+04, 1.2334E+04, 1.2349E+04, 1.2364E+04, 1.2378E+04,
     + 1.2393E+04, 1.2408E+04, 1.2422E+04, 1.2437E+04, 1.2452E+04,
     +  1.2467E+04, 1.2481E+04, 1.2496E+04, 1.2511E+04, 1.2526E+04,
     + 1.2541E+04, 1.2555E+04, 1.2570E+04, 1.2585E+04, 1.2600E+04,
     +  1.2615E+04, 1.2630E+04, 1.2644E+04, 1.2659E+04, 1.2674E+04,
     + 1.2689E+04, 1.2704E+04, 1.2719E+04, 1.2734E+04, 1.2749E+04,
     +  1.2764E+04, 1.2779E+04, 1.2794E+04, 1.2809E+04, 1.2823E+04,
     + 1.2838E+04, 1.2853E+04, 1.2868E+04, 1.2883E+04, 1.2899E+04,
     +  1.2914E+04, 1.2929E+04, 1.2944E+04, 1.2959E+04, 1.2974E+04,
     + 1.2989E+04, 1.3004E+04, 1.3019E+04, 1.3034E+04, 1.3049E+04,
     +  1.3064E+04, 1.3079E+04, 1.3095E+04, 1.3110E+04, 1.3125E+04,
     + 1.3140E+04, 1.3155E+04, 1.3170E+04, 1.3186E+04, 1.3201E+04,
     +  1.3216E+04, 1.3231E+04, 1.3247E+04, 1.3262E+04, 1.3277E+04,
     + 1.3292E+04, 1.3308E+04, 1.3323E+04, 1.3338E+04, 1.3353E+04,
     +  1.3369E+04, 1.3384E+04, 1.3399E+04, 1.3415E+04, 1.3430E+04,
     + 1.3445E+04, 1.3461E+04, 1.3476E+04, 1.3491E+04, 1.3507E+04,
     +  1.3522E+04, 1.3538E+04, 1.3553E+04, 1.3568E+04, 1.3584E+04,
     + 1.3599E+04, 1.3615E+04, 1.3630E+04, 1.3646E+04, 1.3661E+04,
     +  1.3677E+04, 1.3692E+04, 1.3708E+04, 1.3723E+04, 1.3739E+04,
     + 1.3754E+04, 1.3770E+04, 1.3785E+04, 1.3801E+04, 1.3817E+04,
     +  1.3832E+04, 1.3848E+04, 1.3863E+04, 1.3879E+04, 1.3895E+04,
     + 1.3910E+04, 1.3926E+04, 1.3941E+04, 1.3957E+04, 1.3973E+04,
     +  1.3988E+04, 1.4004E+04, 1.4020E+04, 1.4035E+04, 1.4051E+04,
     + 1.4067E+04, 1.4083E+04, 1.4098E+04, 1.4114E+04, 1.4130E+04,
     +  1.4146E+04, 1.4161E+04, 1.4177E+04, 1.4193E+04, 1.4209E+04,
     + 1.4225E+04, 1.4240E+04, 1.4256E+04, 1.4272E+04, 1.4288E+04,
     +  1.4304E+04, 1.4320E+04, 1.4336E+04, 1.4351E+04, 1.4367E+04,
     + 1.4383E+04, 1.4399E+04, 1.4415E+04, 1.4431E+04, 1.4447E+04,
     +  1.4463E+04, 1.4479E+04, 1.4495E+04, 1.4511E+04, 1.4527E+04,
     + 1.4543E+04, 1.4559E+04, 1.4575E+04, 1.4591E+04, 1.4607E+04,
     +  1.4623E+04, 1.4639E+04, 1.4655E+04, 1.4671E+04, 1.4687E+04,
     + 1.4703E+04, 1.4719E+04, 1.4735E+04, 1.4752E+04, 1.4768E+04,
     +  1.4784E+04, 1.4800E+04, 1.4816E+04, 1.4832E+04, 1.4848E+04,
     + 1.4865E+04, 1.4881E+04, 1.4897E+04, 1.4913E+04, 1.4929E+04,
     +  1.4946E+04, 1.4962E+04, 1.4978E+04, 1.4994E+04, 1.5011E+04,
     + 1.5027E+04, 1.5043E+04, 1.5059E+04, 1.5076E+04, 1.5092E+04,
     +  1.5108E+04, 1.5125E+04, 1.5141E+04, 1.5157E+04, 1.5174E+04,
     + 1.5190E+04, 1.5206E+04, 1.5223E+04, 1.5239E+04, 1.5256E+04,
     +  1.5272E+04, 1.5289E+04, 1.5305E+04, 1.5321E+04, 1.5338E+04,
     + 1.5354E+04, 1.5371E+04, 1.5387E+04, 1.5404E+04, 1.5420E+04,
     +  1.5437E+04, 1.5453E+04, 1.5470E+04, 1.5486E+04, 1.5503E+04,
     + 1.5519E+04, 1.5536E+04, 1.5552E+04, 1.5569E+04, 1.5586E+04,
     +  1.5602E+04, 1.5619E+04, 1.5635E+04, 1.5652E+04, 1.5669E+04,
     + 1.5685E+04, 1.5702E+04, 1.5719E+04, 1.5735E+04, 1.5752E+04,
     +  1.5769E+04, 1.5785E+04, 1.5802E+04, 1.5819E+04, 1.5836E+04,
     + 1.5852E+04, 1.5869E+04, 1.5886E+04, 1.5903E+04, 1.5919E+04,
     +  1.5936E+04, 1.5953E+04, 1.5970E+04, 1.5987E+04, 1.6003E+04,
     + 1.6020E+04, 1.6037E+04, 1.6054E+04, 1.6071E+04, 1.6088E+04,
     +  1.6104E+04, 1.6121E+04, 1.6138E+04, 1.6155E+04, 1.6172E+04,
     + 1.6189E+04, 1.6206E+04, 1.6223E+04, 1.6240E+04, 1.6257E+04,
     +  1.6274E+04, 1.6291E+04, 1.6308E+04, 1.6325E+04, 1.6342E+04,
     + 1.6359E+04, 1.6376E+04, 1.6393E+04, 1.6410E+04, 1.6427E+04,
     +  1.6444E+04, 1.6461E+04, 1.6478E+04, 1.6495E+04, 1.6512E+04,
     + 1.6529E+04, 1.6546E+04, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 3,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        23
      data (QofT( 4,J),J=1,2408)/ 5.1854E+00, 8.3719E+00, 1.1739E+01,
     +  1.5141E+01, 1.8555E+01, 2.1976E+01, 2.5400E+01, 2.8826E+01,
     + 3.2253E+01, 3.5681E+01, 3.9110E+01, 4.2540E+01, 4.5970E+01,
     +  4.9401E+01, 5.2832E+01, 5.6263E+01, 5.9695E+01, 6.3126E+01,
     + 6.6558E+01, 6.9990E+01, 7.6855E+01, 8.3720E+01, 9.0586E+01,
     +  9.7452E+01, 1.0432E+02, 1.1119E+02, 1.1805E+02, 1.2492E+02,
     + 1.3179E+02, 1.3866E+02, 1.4553E+02, 1.5240E+02, 1.5927E+02,
     +  1.6614E+02, 1.7301E+02, 1.7988E+02, 1.8675E+02, 1.9362E+02,
     + 2.0049E+02, 2.0737E+02, 2.1424E+02, 2.2111E+02, 2.2799E+02,
     +  2.3486E+02, 2.4174E+02, 2.4861E+02, 2.5549E+02, 2.6236E+02,
     + 2.6924E+02, 2.7611E+02, 2.8299E+02, 2.8987E+02, 2.9674E+02,
     +  3.0362E+02, 3.1050E+02, 3.1738E+02, 3.2426E+02, 3.3113E+02,
     + 3.3801E+02, 3.4489E+02, 3.5177E+02, 3.5865E+02, 3.6553E+02,
     +  3.7242E+02, 3.7930E+02, 3.8618E+02, 3.9306E+02, 3.9994E+02,
     + 4.0683E+02, 4.1371E+02, 4.2059E+02, 4.2748E+02, 4.3436E+02,
     +  4.4125E+02, 4.4813E+02, 4.5502E+02, 4.6190E+02, 4.6879E+02,
     + 4.7567E+02, 4.8256E+02, 4.8945E+02, 4.9634E+02, 5.0322E+02,
     +  5.1011E+02, 5.1700E+02, 5.2389E+02, 5.3078E+02, 5.3767E+02,
     + 5.4456E+02, 5.5145E+02, 5.5834E+02, 5.6524E+02, 5.7213E+02,
     +  5.7902E+02, 5.8592E+02, 5.9281E+02, 5.9971E+02, 6.0660E+02,
     + 6.1350E+02, 6.2039E+02, 6.2729E+02, 6.3419E+02, 6.4109E+02,
     +  6.4799E+02, 6.5489E+02, 6.6179E+02, 6.6869E+02, 6.7559E+02,
     + 6.8250E+02, 6.8940E+02, 6.9631E+02, 7.0321E+02, 7.1012E+02,
     +  7.1703E+02, 7.2394E+02, 7.3085E+02, 7.3776E+02, 7.4468E+02,
     + 7.5159E+02, 7.5851E+02, 7.6543E+02, 7.7235E+02, 7.7927E+02,
     +  7.8619E+02, 7.9311E+02, 8.0004E+02, 8.0697E+02, 8.1390E+02,
     + 8.2083E+02, 8.2776E+02, 8.3470E+02, 8.4163E+02, 8.4857E+02,
     +  8.5551E+02, 8.6246E+02, 8.6941E+02, 8.7635E+02, 8.8331E+02,
     + 8.9026E+02, 8.9722E+02, 9.0417E+02, 9.1114E+02, 9.1810E+02,
     +  9.2507E+02, 9.3204E+02, 9.3901E+02, 9.4599E+02, 9.5297E+02,
     + 9.5996E+02, 9.6694E+02, 9.7393E+02, 9.8093E+02, 9.8792E+02,
     +  9.9493E+02, 1.0019E+03, 1.0089E+03, 1.0160E+03, 1.0230E+03,
     + 1.0300E+03, 1.0370E+03, 1.0440E+03, 1.0511E+03, 1.0581E+03,
     +  1.0652E+03, 1.0722E+03, 1.0793E+03, 1.0863E+03, 1.0934E+03,
     + 1.1004E+03, 1.1075E+03, 1.1146E+03, 1.1217E+03, 1.1288E+03,
     +  1.1359E+03, 1.1430E+03, 1.1501E+03, 1.1572E+03, 1.1643E+03,
     + 1.1714E+03, 1.1785E+03, 1.1857E+03, 1.1928E+03, 1.2000E+03,
     +  1.2071E+03, 1.2143E+03, 1.2214E+03, 1.2286E+03, 1.2358E+03,
     + 1.2430E+03, 1.2501E+03, 1.2573E+03, 1.2645E+03, 1.2718E+03,
     +  1.2790E+03, 1.2862E+03, 1.2934E+03, 1.3007E+03, 1.3079E+03,
     + 1.3152E+03, 1.3224E+03, 1.3297E+03, 1.3370E+03, 1.3443E+03,
     +  1.3515E+03, 1.3588E+03, 1.3662E+03, 1.3735E+03, 1.3808E+03,
     + 1.3881E+03, 1.3955E+03, 1.4028E+03, 1.4102E+03, 1.4175E+03,
     +  1.4249E+03, 1.4323E+03, 1.4397E+03, 1.4471E+03, 1.4545E+03,
     + 1.4619E+03, 1.4693E+03, 1.4767E+03, 1.4842E+03, 1.4916E+03,
     +  1.4991E+03, 1.5066E+03, 1.5140E+03, 1.5215E+03, 1.5290E+03,
     + 1.5365E+03, 1.5441E+03, 1.5516E+03, 1.5591E+03, 1.5667E+03,
     +  1.5742E+03, 1.5818E+03, 1.5893E+03, 1.5969E+03, 1.6045E+03,
     + 1.6121E+03, 1.6197E+03, 1.6274E+03, 1.6350E+03, 1.6426E+03,
     +  1.6503E+03, 1.6580E+03, 1.6656E+03, 1.6733E+03, 1.6810E+03,
     + 1.6887E+03, 1.6964E+03, 1.7042E+03, 1.7119E+03, 1.7196E+03,
     +  1.7274E+03, 1.7352E+03, 1.7430E+03, 1.7507E+03, 1.7585E+03,
     + 1.7664E+03, 1.7742E+03, 1.7820E+03, 1.7899E+03, 1.7977E+03,
     +  1.8056E+03, 1.8135E+03, 1.8214E+03, 1.8293E+03, 1.8372E+03,
     + 1.8451E+03, 1.8530E+03, 1.8610E+03, 1.8689E+03, 1.8769E+03,
     +  1.8849E+03, 1.8929E+03, 1.9009E+03, 1.9089E+03, 1.9169E+03,
     + 1.9250E+03, 1.9330E+03, 1.9411E+03, 1.9492E+03, 1.9573E+03,
     +  1.9654E+03, 1.9735E+03, 1.9816E+03, 1.9897E+03, 1.9979E+03,
     + 2.0061E+03, 2.0142E+03, 2.0224E+03, 2.0306E+03, 2.0388E+03,
     +  2.0470E+03, 2.0553E+03, 2.0635E+03, 2.0718E+03, 2.0801E+03,
     + 2.0883E+03, 2.0966E+03, 2.1049E+03, 2.1133E+03, 2.1216E+03,
     +  2.1299E+03, 2.1383E+03, 2.1467E+03, 2.1551E+03, 2.1635E+03,
     + 2.1719E+03, 2.1803E+03, 2.1887E+03, 2.1972E+03, 2.2056E+03,
     +  2.2141E+03, 2.2226E+03, 2.2311E+03, 2.2396E+03, 2.2481E+03,
     + 2.2567E+03, 2.2652E+03, 2.2738E+03, 2.2824E+03, 2.2910E+03,
     +  2.2996E+03, 2.3082E+03, 2.3168E+03, 2.3255E+03, 2.3341E+03,
     + 2.3428E+03, 2.3515E+03, 2.3602E+03, 2.3689E+03, 2.3776E+03,
     +  2.3863E+03, 2.3951E+03, 2.4039E+03, 2.4126E+03, 2.4214E+03,
     + 2.4302E+03, 2.4391E+03, 2.4479E+03, 2.4567E+03, 2.4656E+03,
     +  2.4745E+03, 2.4833E+03, 2.4922E+03, 2.5012E+03, 2.5101E+03,
     + 2.5190E+03, 2.5280E+03, 2.5369E+03, 2.5459E+03, 2.5549E+03,
     +  2.5639E+03, 2.5730E+03, 2.5820E+03, 2.5910E+03, 2.6001E+03,
     + 2.6092E+03, 2.6183E+03, 2.6274E+03, 2.6365E+03, 2.6456E+03,
     +  2.6548E+03, 2.6639E+03, 2.6731E+03, 2.6823E+03, 2.6915E+03,
     + 2.7007E+03, 2.7100E+03, 2.7192E+03, 2.7285E+03, 2.7377E+03,
     +  2.7470E+03, 2.7563E+03, 2.7656E+03, 2.7750E+03, 2.7843E+03,
     + 2.7937E+03, 2.8031E+03, 2.8124E+03, 2.8218E+03, 2.8313E+03,
     +  2.8407E+03, 2.8501E+03, 2.8596E+03, 2.8691E+03, 2.8785E+03,
     + 2.8880E+03, 2.8976E+03, 2.9071E+03, 2.9166E+03, 2.9262E+03,
     +  2.9358E+03, 2.9453E+03, 2.9550E+03, 2.9646E+03, 2.9742E+03,
     + 2.9838E+03, 2.9935E+03, 3.0032E+03, 3.0129E+03, 3.0226E+03,
     +  3.0323E+03, 3.0420E+03, 3.0518E+03, 3.0615E+03, 3.0713E+03,
     + 3.0811E+03, 3.0909E+03, 3.1007E+03, 3.1105E+03, 3.1204E+03,
     +  3.1302E+03, 3.1401E+03, 3.1500E+03, 3.1599E+03, 3.1698E+03,
     + 3.1798E+03, 3.1897E+03, 3.1997E+03, 3.2097E+03, 3.2197E+03,
     +  3.2297E+03, 3.2397E+03, 3.2497E+03, 3.2598E+03, 3.2698E+03,
     + 3.2799E+03, 3.2900E+03, 3.3001E+03, 3.3103E+03, 3.3204E+03,
     +  3.3306E+03, 3.3407E+03, 3.3509E+03, 3.3611E+03, 3.3713E+03,
     + 3.3816E+03, 3.3918E+03, 3.4021E+03, 3.4123E+03, 3.4226E+03,
     +  3.4329E+03, 3.4432E+03, 3.4536E+03, 3.4639E+03, 3.4743E+03,
     + 3.4847E+03, 3.4951E+03, 3.5055E+03, 3.5159E+03, 3.5263E+03,
     +  3.5368E+03, 3.5472E+03, 3.5577E+03, 3.5682E+03, 3.5787E+03,
     + 3.5893E+03, 3.5998E+03, 3.6104E+03, 3.6209E+03, 3.6315E+03,
     +  3.6421E+03, 3.6527E+03, 3.6634E+03, 3.6740E+03, 3.6847E+03,
     + 3.6954E+03, 3.7060E+03, 3.7168E+03, 3.7275E+03, 3.7382E+03,
     +  3.7490E+03, 3.7597E+03, 3.7705E+03, 3.7813E+03, 3.7921E+03,
     + 3.8030E+03, 3.8138E+03, 3.8247E+03, 3.8355E+03, 3.8464E+03,
     +  3.8573E+03, 3.8682E+03, 3.8792E+03, 3.8901E+03, 3.9011E+03,
     + 3.9121E+03, 3.9231E+03, 3.9341E+03, 3.9451E+03, 3.9561E+03,
     +  3.9672E+03, 3.9783E+03, 3.9893E+03, 4.0004E+03, 4.0116E+03,
     + 4.0227E+03, 4.0338E+03, 4.0450E+03, 4.0562E+03, 4.0674E+03,
     +  4.0786E+03, 4.0898E+03, 4.1010E+03, 4.1123E+03, 4.1236E+03,
     + 4.1348E+03, 4.1461E+03, 4.1575E+03, 4.1688E+03, 4.1801E+03,
     +  4.1915E+03, 4.2029E+03, 4.2143E+03, 4.2257E+03, 4.2371E+03,
     + 4.2485E+03, 4.2600E+03, 4.2714E+03, 4.2829E+03, 4.2944E+03,
     +  4.3059E+03, 4.3175E+03, 4.3290E+03, 4.3406E+03, 4.3521E+03,
     + 4.3637E+03, 4.3753E+03, 4.3869E+03, 4.3986E+03, 4.4102E+03,
     +  4.4219E+03, 4.4336E+03, 4.4453E+03, 4.4570E+03, 4.4687E+03,
     + 4.4805E+03, 4.4922E+03, 4.5040E+03, 4.5158E+03, 4.5276E+03,
     +  4.5394E+03, 4.5512E+03, 4.5631E+03, 4.5750E+03, 4.5868E+03,
     + 4.5987E+03, 4.6107E+03, 4.6226E+03, 4.6345E+03, 4.6465E+03,
     +  4.6585E+03, 4.6705E+03, 4.6825E+03, 4.6945E+03, 4.7065E+03,
     + 4.7186E+03, 4.7306E+03, 4.7427E+03, 4.7548E+03, 4.7669E+03,
     +  4.7791E+03, 4.7912E+03, 4.8034E+03, 4.8156E+03, 4.8277E+03,
     + 4.8400E+03, 4.8522E+03, 4.8644E+03, 4.8767E+03, 4.8889E+03,
     +  4.9012E+03, 4.9135E+03, 4.9258E+03, 4.9382E+03, 4.9505E+03,
     + 4.9629E+03, 4.9753E+03, 4.9877E+03, 5.0001E+03, 5.0125E+03,
     +  5.0249E+03, 5.0374E+03, 5.0499E+03, 5.0623E+03, 5.0748E+03,
     + 5.0874E+03, 5.0999E+03, 5.1124E+03, 5.1250E+03, 5.1376E+03,
     +  5.1502E+03, 5.1628E+03, 5.1754E+03, 5.1881E+03, 5.2007E+03,
     + 5.2134E+03, 5.2261E+03, 5.2388E+03, 5.2515E+03, 5.2643E+03,
     +  5.2770E+03, 5.2898E+03, 5.3026E+03, 5.3154E+03, 5.3282E+03,
     + 5.3410E+03, 5.3539E+03, 5.3667E+03, 5.3796E+03, 5.3925E+03,
     +  5.4054E+03, 5.4183E+03, 5.4313E+03, 5.4442E+03, 5.4572E+03,
     + 5.4702E+03, 5.4832E+03, 5.4962E+03, 5.5092E+03, 5.5223E+03,
     +  5.5354E+03, 5.5484E+03, 5.5615E+03, 5.5746E+03, 5.5878E+03,
     + 5.6009E+03, 5.6141E+03, 5.6272E+03, 5.6404E+03, 5.6536E+03,
     +  5.6669E+03, 5.6801E+03, 5.6934E+03, 5.7066E+03, 5.7199E+03,
     + 5.7332E+03, 5.7465E+03, 5.7599E+03, 5.7732E+03, 5.7866E+03,
     +  5.8000E+03, 5.8134E+03, 5.8268E+03, 5.8402E+03, 5.8536E+03,
     + 5.8671E+03, 5.8806E+03, 5.8941E+03, 5.9076E+03, 5.9211E+03,
     +  5.9346E+03, 5.9482E+03, 5.9618E+03, 5.9753E+03, 5.9889E+03,
     + 6.0026E+03, 6.0162E+03, 6.0298E+03, 6.0435E+03, 6.0572E+03,
     +  6.0709E+03, 6.0846E+03, 6.0983E+03, 6.1121E+03, 6.1258E+03,
     + 6.1396E+03, 6.1534E+03, 6.1672E+03, 6.1810E+03, 6.1948E+03,
     +  6.2087E+03, 6.2226E+03, 6.2364E+03, 6.2503E+03, 6.2643E+03,
     + 6.2782E+03, 6.2921E+03, 6.3061E+03, 6.3201E+03, 6.3341E+03,
     +  6.3481E+03, 6.3621E+03, 6.3762E+03, 6.3902E+03, 6.4043E+03,
     + 6.4184E+03, 6.4325E+03, 6.4466E+03, 6.4608E+03, 6.4749E+03,
     +  6.4891E+03, 6.5033E+03, 6.5175E+03, 6.5317E+03, 6.5459E+03,
     + 6.5602E+03, 6.5744E+03, 6.5887E+03, 6.6030E+03, 6.6173E+03,
     +  6.6317E+03, 6.6460E+03, 6.6604E+03, 6.6747E+03, 6.6891E+03,
     + 6.7035E+03, 6.7180E+03, 6.7324E+03, 6.7469E+03, 6.7613E+03,
     +  6.7758E+03, 6.7903E+03, 6.8048E+03, 6.8194E+03, 6.8339E+03,
     + 6.8485E+03, 6.8631E+03, 6.8777E+03, 6.8923E+03, 6.9069E+03,
     +  6.9216E+03, 6.9362E+03, 6.9509E+03, 6.9656E+03, 6.9803E+03,
     + 6.9950E+03, 7.0098E+03, 7.0245E+03, 7.0393E+03, 7.0541E+03,
     +  7.0689E+03, 7.0837E+03, 7.0986E+03, 7.1134E+03, 7.1283E+03,
     + 7.1432E+03, 7.1581E+03, 7.1730E+03, 7.1879E+03, 7.2029E+03,
     +  7.2178E+03, 7.2328E+03, 7.2478E+03, 7.2628E+03, 7.2779E+03,
     + 7.2929E+03, 7.3080E+03, 7.3230E+03, 7.3381E+03, 7.3532E+03,
     +  7.3684E+03, 7.3835E+03, 7.3987E+03, 7.4138E+03, 7.4290E+03,
     + 7.4442E+03, 7.4594E+03, 7.4747E+03, 7.4899E+03, 7.5052E+03,
     +  7.5205E+03, 7.5358E+03, 7.5511E+03, 7.5664E+03, 7.5818E+03,
     + 7.5972E+03, 7.6125E+03, 7.6279E+03, 7.6433E+03, 7.6588E+03,
     +  7.6742E+03, 7.6897E+03, 7.7052E+03, 7.7206E+03, 7.7362E+03,
     + 7.7517E+03, 7.7672E+03, 7.7828E+03, 7.7983E+03, 7.8139E+03,
     +  7.8295E+03, 7.8452E+03, 7.8608E+03, 7.8765E+03, 7.8921E+03,
     + 7.9078E+03, 7.9235E+03, 7.9392E+03, 7.9550E+03, 7.9707E+03,
     +  7.9865E+03, 8.0023E+03, 8.0180E+03, 8.0339E+03, 8.0497E+03,
     + 8.0655E+03, 8.0814E+03, 8.0973E+03, 8.1132E+03, 8.1291E+03,
     +  8.1450E+03, 8.1609E+03, 8.1769E+03, 8.1929E+03, 8.2089E+03,
     + 8.2249E+03, 8.2409E+03, 8.2569E+03, 8.2730E+03, 8.2891E+03,
     +  8.3051E+03, 8.3212E+03, 8.3374E+03, 8.3535E+03, 8.3696E+03,
     + 8.3858E+03, 8.4020E+03, 8.4182E+03, 8.4344E+03, 8.4506E+03,
     +  8.4669E+03, 8.4831E+03, 8.4994E+03, 8.5157E+03, 8.5320E+03,
     + 8.5484E+03, 8.5647E+03, 8.5811E+03, 8.5974E+03, 8.6138E+03,
     +  8.6302E+03, 8.6467E+03, 8.6631E+03, 8.6796E+03, 8.6960E+03,
     + 8.7125E+03, 8.7290E+03, 8.7455E+03, 8.7621E+03, 8.7786E+03,
     +  8.7952E+03, 8.8118E+03, 8.8284E+03, 8.8450E+03, 8.8616E+03,
     + 8.8783E+03, 8.8950E+03, 8.9116E+03, 8.9283E+03, 8.9451E+03,
     +  8.9618E+03, 8.9785E+03, 8.9953E+03, 9.0121E+03, 9.0289E+03,
     + 9.0457E+03, 9.0625E+03, 9.0793E+03, 9.0962E+03, 9.1131E+03,
     +  9.1300E+03, 9.1469E+03, 9.1638E+03, 9.1807E+03, 9.1977E+03,
     + 9.2147E+03, 9.2317E+03, 9.2487E+03, 9.2657E+03, 9.2827E+03,
     +  9.2998E+03, 9.3169E+03, 9.3339E+03, 9.3510E+03, 9.3682E+03,
     + 9.3853E+03, 9.4024E+03, 9.4196E+03, 9.4368E+03, 9.4540E+03,
     +  9.4712E+03, 9.4884E+03, 9.5057E+03, 9.5229E+03, 9.5402E+03,
     + 9.5575E+03, 9.5748E+03, 9.5922E+03, 9.6095E+03, 9.6269E+03,
     +  9.6442E+03, 9.6616E+03, 9.6791E+03, 9.6965E+03, 9.7139E+03,
     + 9.7314E+03, 9.7489E+03, 9.7663E+03, 9.7838E+03, 9.8014E+03,
     +  9.8189E+03, 9.8365E+03, 9.8540E+03, 9.8716E+03, 9.8892E+03,
     + 9.9068E+03, 9.9245E+03, 9.9421E+03, 9.9598E+03, 9.9775E+03,
     +  9.9952E+03, 1.0013E+04, 1.0031E+04, 1.0048E+04, 1.0066E+04,
     + 1.0084E+04, 1.0102E+04, 1.0120E+04, 1.0137E+04, 1.0155E+04,
     +  1.0173E+04, 1.0191E+04, 1.0209E+04, 1.0227E+04, 1.0245E+04,
     + 1.0263E+04, 1.0281E+04, 1.0299E+04, 1.0317E+04, 1.0335E+04,
     +  1.0353E+04, 1.0371E+04, 1.0389E+04, 1.0407E+04, 1.0425E+04,
     + 1.0443E+04, 1.0461E+04, 1.0479E+04, 1.0497E+04, 1.0516E+04,
     +  1.0534E+04, 1.0552E+04, 1.0570E+04, 1.0588E+04, 1.0607E+04,
     + 1.0625E+04, 1.0643E+04, 1.0662E+04, 1.0680E+04, 1.0698E+04,
     +  1.0717E+04, 1.0735E+04, 1.0753E+04, 1.0772E+04, 1.0790E+04,
     + 1.0809E+04, 1.0827E+04, 1.0845E+04, 1.0864E+04, 1.0882E+04,
     +  1.0901E+04, 1.0920E+04, 1.0938E+04, 1.0957E+04, 1.0975E+04,
     + 1.0994E+04, 1.1012E+04, 1.1031E+04, 1.1050E+04, 1.1068E+04,
     +  1.1087E+04, 1.1106E+04, 1.1124E+04, 1.1143E+04, 1.1162E+04,
     + 1.1181E+04, 1.1199E+04, 1.1218E+04, 1.1237E+04, 1.1256E+04,
     +  1.1275E+04, 1.1294E+04, 1.1312E+04, 1.1331E+04, 1.1350E+04,
     + 1.1369E+04, 1.1388E+04, 1.1407E+04, 1.1426E+04, 1.1445E+04,
     +  1.1464E+04, 1.1483E+04, 1.1502E+04, 1.1521E+04, 1.1540E+04,
     + 1.1559E+04, 1.1578E+04, 1.1597E+04, 1.1617E+04, 1.1636E+04,
     +  1.1655E+04, 1.1674E+04, 1.1693E+04, 1.1713E+04, 1.1732E+04,
     + 1.1751E+04, 1.1770E+04, 1.1790E+04, 1.1809E+04, 1.1828E+04,
     +  1.1847E+04, 1.1867E+04, 1.1886E+04, 1.1906E+04, 1.1925E+04,
     + 1.1944E+04, 1.1964E+04, 1.1983E+04, 1.2003E+04, 1.2022E+04,
     +  1.2042E+04, 1.2061E+04, 1.2081E+04, 1.2100E+04, 1.2120E+04,
     + 1.2139E+04, 1.2159E+04, 1.2179E+04, 1.2198E+04, 1.2218E+04,
     +  1.2238E+04, 1.2257E+04, 1.2277E+04, 1.2297E+04, 1.2316E+04,
     + 1.2336E+04, 1.2356E+04, 1.2376E+04, 1.2395E+04, 1.2415E+04,
     +  1.2435E+04, 1.2455E+04, 1.2475E+04, 1.2495E+04, 1.2514E+04,
     + 1.2534E+04, 1.2554E+04, 1.2574E+04, 1.2594E+04, 1.2614E+04,
     +  1.2634E+04, 1.2654E+04, 1.2674E+04, 1.2694E+04, 1.2714E+04,
     + 1.2734E+04, 1.2754E+04, 1.2775E+04, 1.2795E+04, 1.2815E+04,
     +  1.2835E+04, 1.2855E+04, 1.2875E+04, 1.2895E+04, 1.2916E+04,
     + 1.2936E+04, 1.2956E+04, 1.2976E+04, 1.2997E+04, 1.3017E+04,
     +  1.3037E+04, 1.3058E+04, 1.3078E+04, 1.3098E+04, 1.3119E+04,
     + 1.3139E+04, 1.3160E+04, 1.3180E+04, 1.3200E+04, 1.3221E+04,
     +  1.3241E+04, 1.3262E+04, 1.3282E+04, 1.3303E+04, 1.3323E+04,
     + 1.3344E+04, 1.3365E+04, 1.3385E+04, 1.3406E+04, 1.3426E+04,
     +  1.3447E+04, 1.3468E+04, 1.3488E+04, 1.3509E+04, 1.3530E+04,
     + 1.3551E+04, 1.3571E+04, 1.3592E+04, 1.3613E+04, 1.3634E+04,
     +  1.3654E+04, 1.3675E+04, 1.3696E+04, 1.3717E+04, 1.3738E+04,
     + 1.3759E+04, 1.3780E+04, 1.3801E+04, 1.3821E+04, 1.3842E+04,
     +  1.3863E+04, 1.3884E+04, 1.3905E+04, 1.3926E+04, 1.3947E+04,
     + 1.3969E+04, 1.3990E+04, 1.4011E+04, 1.4032E+04, 1.4053E+04,
     +  1.4074E+04, 1.4095E+04, 1.4116E+04, 1.4138E+04, 1.4159E+04,
     + 1.4180E+04, 1.4201E+04, 1.4222E+04, 1.4244E+04, 1.4265E+04,
     +  1.4286E+04, 1.4308E+04, 1.4329E+04, 1.4350E+04, 1.4372E+04,
     + 1.4393E+04, 1.4415E+04, 1.4436E+04, 1.4457E+04, 1.4479E+04,
     +  1.4500E+04, 1.4522E+04, 1.4543E+04, 1.4565E+04, 1.4586E+04,
     + 1.4608E+04, 1.4629E+04, 1.4651E+04, 1.4673E+04, 1.4694E+04,
     +  1.4716E+04, 1.4738E+04, 1.4759E+04, 1.4781E+04, 1.4803E+04,
     + 1.4824E+04, 1.4846E+04, 1.4868E+04, 1.4890E+04, 1.4911E+04,
     +  1.4933E+04, 1.4955E+04, 1.4977E+04, 1.4999E+04, 1.5021E+04,
     + 1.5042E+04, 1.5064E+04, 1.5086E+04, 1.5108E+04, 1.5130E+04,
     +  1.5152E+04, 1.5174E+04, 1.5196E+04, 1.5218E+04, 1.5240E+04,
     + 1.5262E+04, 1.5284E+04, 1.5306E+04, 1.5328E+04, 1.5350E+04,
     +  1.5373E+04, 1.5395E+04, 1.5417E+04, 1.5439E+04, 1.5461E+04,
     + 1.5484E+04, 1.5506E+04, 1.5528E+04, 1.5550E+04, 1.5573E+04,
     +  1.5595E+04, 1.5617E+04, 1.5639E+04, 1.5662E+04, 1.5684E+04,
     + 1.5707E+04, 1.5729E+04, 1.5751E+04, 1.5774E+04, 1.5796E+04,
     +  1.5819E+04, 1.5841E+04, 1.5864E+04, 1.5886E+04, 1.5909E+04,
     + 1.5931E+04, 1.5954E+04, 1.5976E+04, 1.5999E+04, 1.6022E+04,
     +  1.6044E+04, 1.6067E+04, 1.6090E+04, 1.6112E+04, 1.6135E+04,
     + 1.6158E+04, 1.6180E+04, 1.6203E+04, 1.6226E+04, 1.6249E+04,
     +  1.6271E+04, 1.6294E+04, 1.6317E+04, 1.6340E+04, 1.6363E+04,
     + 1.6386E+04, 1.6409E+04, 1.6432E+04, 1.6454E+04, 1.6477E+04,
     +  1.6500E+04, 1.6523E+04, 1.6546E+04, 1.6569E+04, 1.6592E+04,
     + 1.6615E+04, 1.6638E+04, 1.6662E+04, 1.6685E+04, 1.6708E+04,
     +  1.6731E+04, 1.6754E+04, 1.6777E+04, 1.6800E+04, 1.6824E+04,
     + 1.6847E+04, 1.6870E+04, 1.6893E+04, 1.6917E+04, 1.6940E+04,
     +  1.6963E+04, 1.6986E+04, 1.7010E+04, 1.7033E+04, 1.7056E+04,
     + 1.7080E+04, 1.7103E+04, 1.7127E+04, 1.7150E+04, 1.7174E+04,
     +  1.7197E+04, 1.7220E+04, 1.7244E+04, 1.7267E+04, 1.7291E+04,
     + 1.7315E+04, 1.7338E+04, 1.7362E+04, 1.7385E+04, 1.7409E+04,
     +  1.7432E+04, 1.7456E+04, 1.7480E+04, 1.7503E+04, 1.7527E+04,
     + 1.7551E+04, 1.7575E+04, 1.7598E+04, 1.7622E+04, 1.7646E+04,
     +  1.7670E+04, 1.7694E+04, 1.7717E+04, 1.7741E+04, 1.7765E+04,
     + 1.7789E+04, 1.7813E+04, 1.7837E+04, 1.7861E+04, 1.7885E+04,
     +  1.7909E+04, 1.7933E+04, 1.7957E+04, 1.7981E+04, 1.8005E+04,
     + 1.8029E+04, 1.8053E+04, 1.8077E+04, 1.8101E+04, 1.8125E+04,
     +  1.8149E+04, 1.8173E+04, 1.8197E+04, 1.8222E+04, 1.8246E+04,
     + 1.8270E+04, 1.8294E+04, 1.8319E+04, 1.8343E+04, 1.8367E+04,
     +  1.8391E+04, 1.8416E+04, 1.8440E+04, 1.8464E+04, 1.8489E+04,
     + 1.8513E+04, 1.8537E+04, 1.8562E+04, 1.8586E+04, 1.8611E+04,
     +  1.8635E+04, 1.8660E+04, 1.8684E+04, 1.8709E+04, 1.8733E+04,
     + 1.8758E+04, 1.8782E+04, 1.8807E+04, 1.8832E+04, 1.8856E+04,
     +  1.8881E+04, 1.8905E+04, 1.8930E+04, 1.8955E+04, 1.8980E+04,
     + 1.9004E+04, 1.9029E+04, 1.9054E+04, 1.9078E+04, 1.9103E+04,
     +  1.9128E+04, 1.9153E+04, 1.9178E+04, 1.9203E+04, 1.9227E+04,
     + 1.9252E+04, 1.9277E+04, 1.9302E+04, 1.9327E+04, 1.9352E+04,
     +  1.9377E+04, 1.9402E+04, 1.9427E+04, 1.9452E+04, 1.9477E+04,
     + 1.9502E+04, 1.9527E+04, 1.9552E+04, 1.9577E+04, 1.9602E+04,
     +  1.9628E+04, 1.9653E+04, 1.9678E+04, 1.9703E+04, 1.9728E+04,
     + 1.9754E+04, 1.9779E+04, 1.9804E+04, 1.9829E+04, 1.9855E+04,
     +  1.9880E+04, 1.9905E+04, 1.9931E+04, 1.9956E+04, 1.9981E+04,
     + 2.0007E+04, 2.0032E+04, 2.0057E+04, 2.0083E+04, 2.0108E+04,
     +  2.0134E+04, 2.0159E+04, 2.0185E+04, 2.0210E+04, 2.0236E+04,
     + 2.0261E+04, 2.0287E+04, 2.0313E+04, 2.0338E+04, 2.0364E+04,
     +  2.0389E+04, 2.0415E+04, 2.0441E+04, 2.0467E+04, 2.0492E+04,
     + 2.0518E+04, 2.0544E+04, 2.0569E+04, 2.0595E+04, 2.0621E+04,
     +  2.0647E+04, 2.0673E+04, 2.0698E+04, 2.0724E+04, 2.0750E+04,
     + 2.0776E+04, 2.0802E+04, 2.0828E+04, 2.0854E+04, 2.0880E+04,
     +  2.0906E+04, 2.0932E+04, 2.0958E+04, 2.0984E+04, 2.1010E+04,
     + 2.1036E+04, 2.1062E+04, 2.1088E+04, 2.1114E+04, 2.1140E+04,
     +  2.1167E+04, 2.1193E+04, 2.1219E+04, 2.1245E+04, 2.1271E+04,
     + 2.1298E+04, 2.1324E+04, 2.1350E+04, 2.1376E+04, 2.1403E+04,
     +  2.1429E+04, 2.1455E+04, 2.1482E+04, 2.1508E+04, 2.1534E+04,
     + 2.1561E+04, 2.1587E+04, 2.1614E+04, 2.1640E+04, 2.1667E+04,
     +  2.1693E+04, 2.1720E+04, 2.1746E+04, 2.1773E+04, 2.1799E+04,
     + 2.1826E+04, 2.1852E+04, 2.1879E+04, 2.1906E+04, 2.1932E+04,
     +  2.1959E+04, 2.1985E+04, 2.2012E+04, 2.2039E+04, 2.2066E+04,
     + 2.2092E+04, 2.2119E+04, 2.2146E+04, 2.2173E+04, 2.2199E+04,
     +  2.2226E+04, 2.2253E+04, 2.2280E+04, 2.2307E+04, 2.2334E+04,
     + 2.2361E+04, 2.2388E+04, 2.2414E+04, 2.2441E+04, 2.2468E+04,
     +  2.2495E+04, 2.2522E+04, 2.2549E+04, 2.2576E+04, 2.2604E+04,
     + 2.2631E+04, 2.2658E+04, 2.2685E+04, 2.2712E+04, 2.2739E+04,
     +  2.2766E+04, 2.2793E+04, 2.2821E+04, 2.2848E+04, 2.2875E+04,
     + 2.2902E+04, 2.2930E+04, 2.2957E+04, 2.2984E+04, 2.3011E+04,
     +  2.3039E+04, 2.3066E+04, 2.3094E+04, 2.3121E+04, 2.3148E+04,
     + 2.3176E+04, 2.3203E+04, 2.3231E+04, 2.3258E+04, 2.3286E+04,
     +  2.3313E+04, 2.3341E+04, 2.3368E+04, 2.3396E+04, 2.3423E+04,
     + 2.3451E+04, 2.3478E+04, 2.3506E+04, 2.3534E+04, 2.3561E+04,
     +  2.3589E+04, 2.3617E+04, 2.3644E+04, 2.3672E+04, 2.3700E+04,
     + 2.3728E+04, 2.3755E+04, 2.3783E+04, 2.3811E+04, 2.3839E+04,
     +  2.3867E+04, 2.3894E+04, 2.3922E+04, 2.3950E+04, 2.3978E+04,
     + 2.4006E+04, 2.4034E+04, 2.4062E+04, 2.4090E+04, 2.4118E+04,
     +  2.4146E+04, 2.4174E+04, 2.4202E+04, 2.4230E+04, 2.4258E+04,
     + 2.4286E+04, 2.4314E+04, 2.4343E+04, 2.4371E+04, 2.4399E+04,
     +  2.4427E+04, 2.4455E+04, 2.4483E+04, 2.4512E+04, 2.4540E+04,
     + 2.4568E+04, 2.4596E+04, 2.4625E+04, 2.4653E+04, 2.4681E+04,
     +  2.4710E+04, 2.4738E+04, 2.4767E+04, 2.4795E+04, 2.4823E+04,
     + 2.4852E+04, 2.4880E+04, 2.4909E+04, 2.4937E+04, 2.4966E+04,
     +  2.4994E+04, 2.5023E+04, 2.5051E+04, 2.5080E+04, 2.5108E+04,
     + 2.5137E+04, 2.5166E+04, 2.5194E+04, 2.5223E+04, 2.5252E+04,
     +  2.5280E+04, 2.5309E+04, 2.5338E+04, 2.5367E+04, 2.5395E+04,
     + 2.5424E+04, 2.5453E+04, 2.5482E+04, 2.5511E+04, 2.5539E+04,
     +  2.5568E+04, 2.5597E+04, 2.5626E+04, 2.5655E+04, 2.5684E+04,
     + 2.5713E+04, 2.5742E+04, 2.5771E+04, 2.5800E+04, 2.5829E+04,
     +  2.5858E+04, 2.5887E+04, 2.5916E+04, 2.5945E+04, 2.5974E+04,
     + 2.6003E+04, 2.6032E+04, 2.6061E+04, 2.6091E+04, 2.6120E+04,
     +  2.6149E+04, 2.6178E+04, 2.6208E+04, 2.6237E+04, 2.6266E+04,
     + 2.6295E+04, 2.6325E+04, 2.6354E+04, 2.6383E+04, 2.6413E+04,
     +  2.6442E+04, 2.6471E+04, 2.6501E+04, 2.6530E+04, 2.6560E+04,
     + 2.6589E+04, 2.6619E+04, 2.6648E+04, 2.6678E+04, 2.6707E+04,
     +  2.6737E+04, 2.6766E+04, 2.6796E+04, 2.6825E+04, 2.6855E+04,
     + 2.6885E+04, 2.6914E+04, 2.6944E+04, 2.6974E+04, 2.7003E+04,
     +  2.7033E+04, 2.7063E+04, 2.7093E+04, 2.7122E+04, 2.7152E+04,
     + 2.7182E+04, 2.7212E+04, 2.7242E+04, 2.7271E+04, 2.7301E+04,
     +  2.7331E+04, 2.7361E+04, 2.7391E+04, 2.7421E+04, 2.7451E+04,
     + 2.7481E+04, 2.7511E+04, 2.7541E+04, 2.7571E+04, 2.7601E+04,
     +  2.7631E+04, 2.7661E+04, 2.7691E+04, 2.7721E+04, 2.7751E+04,
     + 2.7782E+04, 2.7812E+04, 2.7842E+04, 2.7872E+04, 2.7902E+04,
     +  2.7933E+04, 2.7963E+04, 2.7993E+04, 2.8023E+04, 2.8054E+04,
     + 2.8084E+04, 2.8114E+04, 2.8145E+04, 2.8175E+04, 2.8205E+04,
     +  2.8236E+04, 2.8266E+04, 2.8297E+04, 2.8327E+04, 2.8358E+04,
     + 2.8388E+04, 2.8419E+04, 2.8449E+04, 2.8480E+04, 2.8510E+04,
     +  2.8541E+04, 2.8571E+04, 2.8602E+04, 2.8633E+04, 2.8663E+04,
     + 2.8694E+04, 2.8725E+04, 2.8755E+04, 2.8786E+04, 2.8817E+04,
     +  2.8848E+04, 2.8878E+04, 2.8909E+04, 2.8940E+04, 2.8971E+04,
     + 2.9002E+04, 2.9032E+04, 2.9063E+04, 2.9094E+04, 2.9125E+04,
     +  2.9156E+04, 2.9187E+04, 2.9218E+04, 2.9249E+04, 2.9280E+04,
     + 2.9311E+04, 2.9342E+04, 2.9373E+04, 2.9404E+04, 2.9435E+04,
     +  2.9466E+04, 2.9497E+04, 2.9528E+04, 2.9560E+04, 2.9591E+04,
     + 2.9622E+04, 2.9653E+04, 2.9684E+04, 2.9716E+04, 2.9747E+04,
     +  2.9778E+04, 2.9809E+04, 2.9841E+04, 2.9872E+04, 2.9903E+04,
     + 2.9935E+04, 2.9966E+04, 2.9997E+04, 3.0029E+04, 3.0060E+04,
     +  3.0092E+04, 3.0123E+04, 3.0155E+04, 3.0186E+04, 3.0218E+04,
     + 3.0249E+04, 3.0281E+04, 3.0312E+04, 3.0344E+04, 3.0375E+04,
     +  3.0407E+04, 3.0439E+04, 3.0470E+04, 3.0502E+04, 3.0534E+04,
     + 3.0565E+04, 3.0597E+04, 3.0629E+04, 3.0660E+04, 3.0692E+04,
     +  3.0724E+04, 3.0756E+04, 3.0788E+04, 3.0819E+04, 3.0851E+04,
     + 3.0883E+04, 3.0915E+04, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 4,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_SO3   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 1),QofT( 1,NT),Q(NT)
      data Tmx(1)/3500./
      data xgj/ 1./
c...      SO3
c...        --        26
      data (QofT( 1,J),J=1,2408)/ 1.4143E+00, 3.4773E+00, 6.2912E+00,
     +  9.6174E+00, 1.3384E+01, 1.7544E+01, 2.2064E+01, 2.6917E+01,
     + 3.2081E+01, 3.7539E+01, 4.3275E+01, 4.9278E+01, 5.5534E+01,
     +  6.2035E+01, 6.8772E+01, 7.5737E+01, 8.2922E+01, 9.0320E+01,
     + 9.7927E+01, 1.0574E+02, 1.2194E+02, 1.3890E+02, 1.5658E+02,
     +  1.7495E+02, 1.9399E+02, 2.1367E+02, 2.3398E+02, 2.5489E+02,
     + 2.7639E+02, 2.9847E+02, 3.2110E+02, 3.4428E+02, 3.6799E+02,
     +  3.9222E+02, 4.1697E+02, 4.4221E+02, 4.6794E+02, 4.9415E+02,
     + 5.2084E+02, 5.4799E+02, 5.7560E+02, 6.0366E+02, 6.3217E+02,
     +  6.6111E+02, 6.9048E+02, 7.2029E+02, 7.5051E+02, 7.8116E+02,
     + 8.1222E+02, 8.4369E+02, 8.7557E+02, 9.0786E+02, 9.4055E+02,
     +  9.7365E+02, 1.0072E+03, 1.0411E+03, 1.0754E+03, 1.1101E+03,
     + 1.1452E+03, 1.1807E+03, 1.2166E+03, 1.2529E+03, 1.2897E+03,
     +  1.3268E+03, 1.3644E+03, 1.4024E+03, 1.4408E+03, 1.4797E+03,
     + 1.5190E+03, 1.5587E+03, 1.5988E+03, 1.6394E+03, 1.6805E+03,
     +  1.7220E+03, 1.7640E+03, 1.8065E+03, 1.8494E+03, 1.8928E+03,
     + 1.9368E+03, 1.9812E+03, 2.0261E+03, 2.0715E+03, 2.1175E+03,
     +  2.1640E+03, 2.2110E+03, 2.2586E+03, 2.3068E+03, 2.3555E+03,
     + 2.4047E+03, 2.4546E+03, 2.5050E+03, 2.5561E+03, 2.6077E+03,
     +  2.6600E+03, 2.7129E+03, 2.7664E+03, 2.8206E+03, 2.8754E+03,
     + 2.9309E+03, 2.9871E+03, 3.0439E+03, 3.1015E+03, 3.1598E+03,
     +  3.2187E+03, 3.2784E+03, 3.3388E+03, 3.4000E+03, 3.4619E+03,
     + 3.5246E+03, 3.5881E+03, 3.6524E+03, 3.7174E+03, 3.7833E+03,
     +  3.8500E+03, 3.9175E+03, 3.9859E+03, 4.0551E+03, 4.1252E+03,
     + 4.1962E+03, 4.2680E+03, 4.3408E+03, 4.4145E+03, 4.4891E+03,
     +  4.5646E+03, 4.6411E+03, 4.7185E+03, 4.7969E+03, 4.8763E+03,
     + 4.9567E+03, 5.0381E+03, 5.1206E+03, 5.2041E+03, 5.2886E+03,
     +  5.3742E+03, 5.4608E+03, 5.5486E+03, 5.6374E+03, 5.7274E+03,
     + 5.8185E+03, 5.9107E+03, 6.0041E+03, 6.0987E+03, 6.1945E+03,
     +  6.2914E+03, 6.3896E+03, 6.4890E+03, 6.5897E+03, 6.6916E+03,
     + 6.7948E+03, 6.8992E+03, 7.0050E+03, 7.1121E+03, 7.2205E+03,
     +  7.3303E+03, 7.4415E+03, 7.5540E+03, 7.6679E+03, 7.7833E+03,
     + 7.9001E+03, 8.0183E+03, 8.1380E+03, 8.2592E+03, 8.3819E+03,
     +  8.5061E+03, 8.6318E+03, 8.7591E+03, 8.8879E+03, 9.0184E+03,
     + 9.1504E+03, 9.2841E+03, 9.4194E+03, 9.5564E+03, 9.6951E+03,
     +  9.8354E+03, 9.9775E+03, 1.0121E+04, 1.0267E+04, 1.0414E+04,
     + 1.0563E+04, 1.0714E+04, 1.0867E+04, 1.1022E+04, 1.1178E+04,
     +  1.1337E+04, 1.1497E+04, 1.1659E+04, 1.1824E+04, 1.1990E+04,
     + 1.2158E+04, 1.2328E+04, 1.2501E+04, 1.2675E+04, 1.2851E+04,
     +  1.3030E+04, 1.3211E+04, 1.3394E+04, 1.3579E+04, 1.3766E+04,
     + 1.3956E+04, 1.4147E+04, 1.4341E+04, 1.4538E+04, 1.4737E+04,
     +  1.4938E+04, 1.5141E+04, 1.5347E+04, 1.5555E+04, 1.5766E+04,
     + 1.5979E+04, 1.6195E+04, 1.6413E+04, 1.6634E+04, 1.6857E+04,
     +  1.7084E+04, 1.7312E+04, 1.7544E+04, 1.7778E+04, 1.8014E+04,
     + 1.8254E+04, 1.8496E+04, 1.8742E+04, 1.8990E+04, 1.9241E+04,
     +  1.9494E+04, 1.9751E+04, 2.0011E+04, 2.0273E+04, 2.0539E+04,
     + 2.0808E+04, 2.1080E+04, 2.1355E+04, 2.1633E+04, 2.1914E+04,
     +  2.2199E+04, 2.2487E+04, 2.2778E+04, 2.3072E+04, 2.3370E+04,
     + 2.3671E+04, 2.3976E+04, 2.4284E+04, 2.4595E+04, 2.4910E+04,
     +  2.5229E+04, 2.5551E+04, 2.5877E+04, 2.6206E+04, 2.6539E+04,
     + 2.6876E+04, 2.7217E+04, 2.7561E+04, 2.7910E+04, 2.8262E+04,
     +  2.8618E+04, 2.8978E+04, 2.9342E+04, 2.9711E+04, 3.0083E+04,
     + 3.0459E+04, 3.0840E+04, 3.1225E+04, 3.1614E+04, 3.2007E+04,
     +  3.2405E+04, 3.2807E+04, 3.3213E+04, 3.3624E+04, 3.4039E+04,
     + 3.4459E+04, 3.4884E+04, 3.5313E+04, 3.5747E+04, 3.6185E+04,
     +  3.6628E+04, 3.7077E+04, 3.7529E+04, 3.7987E+04, 3.8450E+04,
     + 3.8918E+04, 3.9391E+04, 3.9869E+04, 4.0352E+04, 4.0840E+04,
     +  4.1334E+04, 4.1833E+04, 4.2337E+04, 4.2846E+04, 4.3361E+04,
     + 4.3882E+04, 4.4408E+04, 4.4939E+04, 4.5477E+04, 4.6019E+04,
     +  4.6568E+04, 4.7123E+04, 4.7683E+04, 4.8249E+04, 4.8821E+04,
     + 4.9400E+04, 4.9984E+04, 5.0574E+04, 5.1171E+04, 5.1774E+04,
     +  5.2383E+04, 5.2998E+04, 5.3620E+04, 5.4249E+04, 5.4884E+04,
     + 5.5525E+04, 5.6173E+04, 5.6828E+04, 5.7490E+04, 5.8158E+04,
     +  5.8834E+04, 5.9516E+04, 6.0205E+04, 6.0902E+04, 6.1605E+04,
     + 6.2316E+04, 6.3034E+04, 6.3760E+04, 6.4492E+04, 6.5233E+04,
     +  6.5980E+04, 6.6736E+04, 6.7499E+04, 6.8270E+04, 6.9048E+04,
     + 6.9835E+04, 7.0629E+04, 7.1431E+04, 7.2242E+04, 7.3060E+04,
     +  7.3887E+04, 7.4722E+04, 7.5400E+04, 7.6248E+04, 7.7105E+04,
     + 7.7971E+04, 7.8845E+04, 7.9728E+04, 8.0620E+04, 8.1521E+04,
     +  8.2430E+04, 8.3349E+04, 8.4276E+04, 8.5213E+04, 8.6159E+04,
     + 8.7114E+04, 8.8078E+04, 8.9052E+04, 9.0036E+04, 9.1029E+04,
     +  9.2032E+04, 9.3045E+04, 9.4067E+04, 9.5100E+04, 9.6142E+04,
     + 9.7195E+04, 9.8258E+04, 9.9331E+04, 1.0041E+05, 1.0151E+05,
     +  1.0261E+05, 1.0373E+05, 1.0485E+05, 1.0599E+05, 1.0714E+05,
     + 1.0830E+05, 1.0947E+05, 1.1065E+05, 1.1184E+05, 1.1304E+05,
     +  1.1426E+05, 1.1548E+05, 1.1672E+05, 1.1797E+05, 1.1923E+05,
     + 1.2051E+05, 1.2179E+05, 1.2309E+05, 1.2440E+05, 1.2572E+05,
     +  1.2706E+05, 1.2841E+05, 1.2977E+05, 1.3114E+05, 1.3252E+05,
     + 1.3392E+05, 1.3533E+05, 1.3676E+05, 1.3820E+05, 1.3965E+05,
     +  1.4111E+05, 1.4259E+05, 1.4408E+05, 1.4559E+05, 1.4711E+05,
     + 1.4864E+05, 1.5019E+05, 1.5175E+05, 1.5333E+05, 1.5492E+05,
     +  1.5652E+05, 1.5814E+05, 1.5978E+05, 1.6143E+05, 1.6309E+05,
     + 1.6477E+05, 1.6647E+05, 1.6818E+05, 1.6990E+05, 1.7165E+05,
     +  1.7340E+05, 1.7518E+05, 1.7697E+05, 1.7877E+05, 1.8059E+05,
     + 1.8243E+05, 1.8429E+05, 1.8616E+05, 1.8805E+05, 1.8995E+05,
     +  1.9187E+05, 1.9381E+05, 1.9577E+05, 1.9774E+05, 1.9973E+05,
     + 2.0174E+05, 2.0377E+05, 2.0581E+05, 2.0788E+05, 2.0996E+05,
     +  2.1206E+05, 2.1418E+05, 2.1631E+05, 2.1847E+05, 2.2064E+05,
     + 2.2284E+05, 2.2505E+05, 2.2728E+05, 2.2953E+05, 2.3180E+05,
     +  2.3409E+05, 2.3640E+05, 2.3874E+05, 2.4109E+05, 2.4346E+05,
     + 2.4585E+05, 2.4826E+05, 2.5070E+05, 2.5315E+05, 2.5563E+05,
     +  2.5812E+05, 2.6064E+05, 2.6318E+05, 2.6575E+05, 2.6833E+05,
     + 2.7094E+05, 2.7356E+05, 2.7621E+05, 2.7889E+05, 2.8158E+05,
     +  2.8430E+05, 2.8704E+05, 2.8981E+05, 2.9260E+05, 2.9541E+05,
     + 2.9825E+05, 3.0111E+05, 3.0399E+05, 3.0690E+05, 3.0983E+05,
     +  3.1279E+05, 3.1577E+05, 3.1878E+05, 3.2181E+05, 3.2487E+05,
     + 3.2795E+05, 3.3106E+05, 3.3420E+05, 3.3736E+05, 3.4055E+05,
     +  3.4376E+05, 3.4700E+05, 3.5027E+05, 3.5356E+05, 3.5688E+05,
     + 3.6023E+05, 3.6361E+05, 3.6701E+05, 3.7045E+05, 3.7391E+05,
     +  3.7740E+05, 3.8091E+05, 3.8446E+05, 3.8803E+05, 3.9164E+05,
     + 3.9527E+05, 3.9894E+05, 4.0263E+05, 4.0635E+05, 4.1011E+05,
     +  4.1389E+05, 4.1770E+05, 4.2155E+05, 4.2542E+05, 4.2933E+05,
     + 4.3327E+05, 4.3724E+05, 4.4124E+05, 4.4528E+05, 4.4935E+05,
     +  4.5345E+05, 4.5758E+05, 4.6174E+05, 4.6594E+05, 4.7017E+05,
     + 4.7444E+05, 4.7874E+05, 4.8307E+05, 4.8744E+05, 4.9184E+05,
     +  4.9628E+05, 5.0075E+05, 5.0526E+05, 5.0981E+05, 5.1439E+05,
     + 5.1900E+05, 5.2365E+05, 5.2834E+05, 5.3307E+05, 5.3783E+05,
     +  5.4263E+05, 5.4746E+05, 5.5234E+05, 5.5725E+05, 5.6220E+05,
     + 5.6719E+05, 5.7222E+05, 5.7728E+05, 5.8239E+05, 5.8754E+05,
     +  5.9272E+05, 5.9795E+05, 6.0321E+05, 6.0852E+05, 6.1387E+05,
     + 6.1925E+05, 6.2468E+05, 6.3016E+05, 6.3567E+05, 6.4122E+05,
     +  6.4682E+05, 6.5246E+05, 6.5815E+05, 6.6387E+05, 6.6964E+05,
     + 6.7546E+05, 6.8132E+05, 6.8722E+05, 6.9317E+05, 6.9916E+05,
     +  7.0520E+05, 7.1128E+05, 7.1741E+05, 7.2358E+05, 7.2981E+05,
     + 7.3607E+05, 7.4239E+05, 7.4875E+05, 7.5516E+05, 7.6162E+05,
     +  7.6813E+05, 7.7468E+05, 7.8129E+05, 7.8794E+05, 7.9464E+05,
     + 8.0140E+05, 8.0820E+05, 8.1505E+05, 8.2195E+05, 8.2891E+05,
     +  8.3592E+05, 8.4297E+05, 8.5008E+05, 8.5724E+05, 8.6446E+05,
     + 8.7173E+05, 8.7905E+05, 8.8642E+05, 8.9385E+05, 9.0133E+05,
     +  9.0887E+05, 9.1646E+05, 9.2411E+05, 9.3182E+05, 9.3958E+05,
     + 9.4739E+05, 9.5526E+05, 9.6319E+05, 9.7118E+05, 9.7923E+05,
     +  9.8733E+05, 9.9549E+05, 1.0037E+06, 1.0120E+06, 1.0203E+06,
     + 1.0287E+06, 1.0372E+06, 1.0457E+06, 1.0543E+06, 1.0629E+06,
     +  1.0716E+06, 1.0804E+06, 1.0892E+06, 1.0981E+06, 1.1071E+06,
     + 1.1161E+06, 1.1252E+06, 1.1343E+06, 1.1435E+06, 1.1528E+06,
     +  1.1621E+06, 1.1716E+06, 1.1810E+06, 1.1906E+06, 1.2002E+06,
     + 1.2099E+06, 1.2196E+06, 1.2294E+06, 1.2393E+06, 1.2492E+06,
     +  1.2593E+06, 1.2693E+06, 1.2795E+06, 1.2897E+06, 1.3000E+06,
     + 1.3104E+06, 1.3208E+06, 1.3314E+06, 1.3420E+06, 1.3526E+06,
     +  1.3634E+06, 1.3742E+06, 1.3850E+06, 1.3960E+06, 1.4070E+06,
     + 1.4182E+06, 1.4293E+06, 1.4406E+06, 1.4520E+06, 1.4634E+06,
     +  1.4749E+06, 1.4864E+06, 1.4981E+06, 1.5098E+06, 1.5216E+06,
     + 1.5335E+06, 1.5455E+06, 1.5576E+06, 1.5697E+06, 1.5819E+06,
     +  1.5942E+06, 1.6066E+06, 1.6191E+06, 1.6316E+06, 1.6443E+06,
     + 1.6570E+06, 1.6698E+06, 1.6827E+06, 1.6957E+06, 1.7088E+06,
     +  1.7219E+06, 1.7352E+06, 1.7485E+06, 1.7619E+06, 1.7754E+06,
     + 1.7890E+06, 1.8027E+06, 1.8165E+06, 1.8304E+06, 1.8443E+06,
     +  1.8584E+06, 1.8726E+06, 1.8868E+06, 1.9011E+06, 1.9156E+06,
     + 1.9301E+06, 1.9447E+06, 1.9594E+06, 1.9743E+06, 1.9892E+06,
     +  2.0042E+06, 2.0193E+06, 2.0345E+06, 2.0498E+06, 2.0652E+06,
     + 2.0807E+06, 2.0963E+06, 2.1120E+06, 2.1279E+06, 2.1438E+06,
     +  2.1598E+06, 2.1759E+06, 2.1921E+06, 2.2085E+06, 2.2249E+06,
     + 2.2415E+06, 2.2581E+06, 2.2749E+06, 2.2917E+06, 2.3087E+06,
     +  2.3258E+06, 2.3430E+06, 2.3603E+06, 2.3777E+06, 2.3952E+06,
     + 2.4129E+06, 2.4306E+06, 2.4485E+06, 2.4664E+06, 2.4845E+06,
     +  2.5027E+06, 2.5211E+06, 2.5395E+06, 2.5580E+06, 2.5767E+06,
     + 2.5955E+06, 2.6144E+06, 2.6334E+06, 2.6526E+06, 2.6718E+06,
     +  2.6912E+06, 2.7107E+06, 2.7304E+06, 2.7501E+06, 2.7700E+06,
     + 2.7900E+06, 2.8101E+06, 2.8304E+06, 2.8508E+06, 2.8713E+06,
     +  2.8919E+06, 2.9127E+06, 2.9336E+06, 2.9546E+06, 2.9757E+06,
     + 2.9970E+06, 3.0184E+06, 3.0400E+06, 3.0617E+06, 3.0835E+06,
     +  3.1054E+06, 3.1275E+06, 3.1497E+06, 3.1721E+06, 3.1946E+06,
     + 3.2172E+06, 3.2400E+06, 3.2629E+06, 3.2859E+06, 3.3091E+06,
     +  3.3325E+06, 3.3559E+06, 3.3796E+06, 3.4033E+06, 3.4272E+06,
     + 3.4513E+06, 3.4755E+06, 3.4998E+06, 3.5243E+06, 3.5489E+06,
     +  3.5737E+06, 3.5987E+06, 3.6237E+06, 3.6490E+06, 3.6744E+06,
     + 3.6999E+06, 3.7256E+06, 3.7514E+06, 3.7775E+06, 3.8036E+06,
     +  3.8299E+06, 3.8564E+06, 3.8830E+06, 3.9098E+06, 3.9368E+06,
     + 3.9639E+06, 3.9912E+06, 4.0186E+06, 4.0462E+06, 4.0739E+06,
     +  4.1019E+06, 4.1300E+06, 4.1582E+06, 4.1866E+06, 4.2152E+06,
     + 4.2440E+06, 4.2729E+06, 4.3020E+06, 4.3313E+06, 4.3607E+06,
     +  4.3903E+06, 4.4201E+06, 4.4501E+06, 4.4802E+06, 4.5105E+06,
     + 4.5410E+06, 4.5717E+06, 4.6025E+06, 4.6336E+06, 4.6648E+06,
     +  4.6961E+06, 4.7277E+06, 4.7595E+06, 4.7914E+06, 4.8235E+06,
     + 4.8558E+06, 4.8883E+06, 4.9210E+06, 4.9539E+06, 4.9869E+06,
     +  5.0202E+06, 5.0536E+06, 5.0873E+06, 5.1211E+06, 5.1551E+06,
     + 5.1893E+06, 5.2237E+06, 5.2583E+06, 5.2931E+06, 5.3281E+06,
     +  5.3633E+06, 5.3987E+06, 5.4343E+06, 5.4701E+06, 5.5062E+06,
     + 5.5424E+06, 5.5788E+06, 5.6154E+06, 5.6522E+06, 5.6893E+06,
     +  5.7265E+06, 5.7640E+06, 5.8017E+06, 5.8395E+06, 5.8776E+06,
     + 5.9159E+06, 5.9545E+06, 5.9932E+06, 6.0322E+06, 6.0713E+06,
     +  6.1107E+06, 6.1503E+06, 6.1902E+06, 6.2302E+06, 6.2705E+06,
     + 6.3110E+06, 6.3518E+06, 6.3927E+06, 6.4339E+06, 6.4753E+06,
     +  6.5170E+06, 6.5588E+06, 6.6009E+06, 6.6433E+06, 6.6858E+06,
     + 6.7286E+06, 6.7717E+06, 6.8150E+06, 6.8585E+06, 6.9022E+06,
     +  6.9462E+06, 6.9905E+06, 7.0349E+06, 7.0797E+06, 7.1246E+06,
     + 7.1698E+06, 7.2153E+06, 7.2610E+06, 7.3070E+06, 7.3532E+06,
     +  7.3996E+06, 7.4463E+06, 7.4933E+06, 7.5405E+06, 7.5880E+06,
     + 7.6357E+06, 7.6837E+06, 7.7319E+06, 7.7805E+06, 7.8292E+06,
     +  7.8783E+06, 7.9276E+06, 7.9771E+06, 8.0269E+06, 8.0770E+06,
     + 8.1274E+06, 8.1780E+06, 8.2289E+06, 8.2801E+06, 8.3316E+06,
     +  8.3833E+06, 8.4353E+06, 8.4876E+06, 8.5401E+06, 8.5930E+06,
     + 8.6461E+06, 8.6995E+06, 8.7532E+06, 8.8071E+06, 8.8614E+06,
     +  8.9159E+06, 8.9708E+06, 9.0259E+06, 9.0813E+06, 9.1370E+06,
     + 9.1930E+06, 9.2493E+06, 9.3059E+06, 9.3628E+06, 9.4199E+06,
     +  9.4774E+06, 9.5352E+06, 9.5933E+06, 9.6517E+06, 9.7104E+06,
     + 9.7694E+06, 9.8287E+06, 9.8883E+06, 9.9482E+06, 1.0008E+07,
     +  1.0069E+07, 1.0130E+07, 1.0191E+07, 1.0253E+07, 1.0314E+07,
     + 1.0377E+07, 1.0439E+07, 1.0502E+07, 1.0565E+07, 1.0628E+07,
     +  1.0692E+07, 1.0756E+07, 1.0821E+07, 1.0885E+07, 1.0950E+07,
     + 1.1016E+07, 1.1082E+07, 1.1148E+07, 1.1214E+07, 1.1281E+07,
     +  1.1348E+07, 1.1415E+07, 1.1483E+07, 1.1551E+07, 1.1620E+07,
     + 1.1689E+07, 1.1758E+07, 1.1827E+07, 1.1897E+07, 1.1968E+07,
     +  1.2038E+07, 1.2109E+07, 1.2180E+07, 1.2252E+07, 1.2324E+07,
     + 1.2396E+07, 1.2469E+07, 1.2542E+07, 1.2616E+07, 1.2690E+07,
     +  1.2764E+07, 1.2839E+07, 1.2914E+07, 1.2989E+07, 1.3065E+07,
     + 1.3141E+07, 1.3217E+07, 1.3294E+07, 1.3371E+07, 1.3449E+07,
     +  1.3527E+07, 1.3605E+07, 1.3684E+07, 1.3763E+07, 1.3843E+07,
     + 1.3923E+07, 1.4003E+07, 1.4084E+07, 1.4165E+07, 1.4247E+07,
     +  1.4329E+07, 1.4411E+07, 1.4494E+07, 1.4577E+07, 1.4661E+07,
     + 1.4745E+07, 1.4829E+07, 1.4914E+07, 1.4999E+07, 1.5085E+07,
     +  1.5171E+07, 1.5258E+07, 1.5345E+07, 1.5432E+07, 1.5520E+07,
     + 1.5608E+07, 1.5697E+07, 1.5786E+07, 1.5875E+07, 1.5965E+07,
     +  1.6055E+07, 1.6146E+07, 1.6238E+07, 1.6329E+07, 1.6421E+07,
     + 1.6514E+07, 1.6607E+07, 1.6700E+07, 1.6794E+07, 1.6889E+07,
     +  1.6984E+07, 1.7079E+07, 1.7175E+07, 1.7271E+07, 1.7368E+07,
     + 1.7465E+07, 1.7562E+07, 1.7660E+07, 1.7759E+07, 1.7858E+07,
     +  1.7957E+07, 1.8057E+07, 1.8158E+07, 1.8259E+07, 1.8360E+07,
     + 1.8462E+07, 1.8564E+07, 1.8667E+07, 1.8770E+07, 1.8874E+07,
     +  1.8978E+07, 1.9083E+07, 1.9189E+07, 1.9294E+07, 1.9401E+07,
     + 1.9507E+07, 1.9615E+07, 1.9722E+07, 1.9831E+07, 1.9940E+07,
     +  2.0049E+07, 2.0159E+07, 2.0269E+07, 2.0380E+07, 2.0491E+07,
     + 2.0603E+07, 2.0716E+07, 2.0829E+07, 2.0942E+07, 2.1056E+07,
     +  2.1171E+07, 2.1286E+07, 2.1401E+07, 2.1517E+07, 2.1634E+07,
     + 2.1751E+07, 2.1869E+07, 2.1987E+07, 2.2106E+07, 2.2225E+07,
     +  2.2345E+07, 2.2466E+07, 2.2587E+07, 2.2709E+07, 2.2831E+07,
     + 2.2953E+07, 2.3077E+07, 2.3201E+07, 2.3325E+07, 2.3450E+07,
     +  2.3576E+07, 2.3702E+07, 2.3828E+07, 2.3956E+07, 2.4084E+07,
     + 2.4212E+07, 2.4341E+07, 2.4471E+07, 2.4601E+07, 2.4732E+07,
     +  2.4863E+07, 2.4995E+07, 2.5128E+07, 2.5261E+07, 2.5395E+07,
     + 2.5529E+07, 2.5664E+07, 2.5800E+07, 2.5936E+07, 2.6073E+07,
     +  2.6211E+07, 2.6349E+07, 2.6487E+07, 2.6627E+07, 2.6767E+07,
     + 2.6907E+07, 2.7048E+07, 2.7190E+07, 2.7333E+07, 2.7476E+07,
     +  2.7620E+07, 2.7764E+07, 2.7909E+07, 2.8055E+07, 2.8201E+07,
     + 2.8348E+07, 2.8496E+07, 2.8644E+07, 2.8793E+07, 2.8943E+07,
     +  2.9093E+07, 2.9244E+07, 2.9396E+07, 2.9548E+07, 2.9701E+07,
     + 2.9855E+07, 3.0009E+07, 3.0164E+07, 3.0320E+07, 3.0476E+07,
     +  3.0633E+07, 3.0791E+07, 3.0949E+07, 3.1108E+07, 3.1268E+07,
     + 3.1429E+07, 3.1590E+07, 3.1752E+07, 3.1915E+07, 3.2078E+07,
     +  3.2242E+07, 3.2407E+07, 3.2572E+07, 3.2739E+07, 3.2906E+07,
     + 3.3073E+07, 3.3242E+07, 3.3411E+07, 3.3581E+07, 3.3751E+07,
     +  3.3923E+07, 3.4095E+07, 3.4268E+07, 3.4441E+07, 3.4616E+07,
     + 3.4791E+07, 3.4966E+07, 3.5143E+07, 3.5320E+07, 3.5499E+07,
     +  3.5677E+07, 3.5857E+07, 3.6038E+07, 3.6219E+07, 3.6401E+07,
     + 3.6583E+07, 3.6767E+07, 3.6951E+07, 3.7136E+07, 3.7322E+07,
     +  3.7509E+07, 3.7697E+07, 3.7885E+07, 3.8074E+07, 3.8264E+07,
     + 3.8455E+07, 3.8646E+07, 3.8838E+07, 3.9032E+07, 3.9225E+07,
     +  3.9420E+07, 3.9616E+07, 3.9812E+07, 4.0010E+07, 4.0208E+07,
     + 4.0407E+07, 4.0606E+07, 4.0807E+07, 4.1008E+07, 4.1211E+07,
     +  4.1414E+07, 4.1618E+07, 4.1823E+07, 4.2028E+07, 4.2235E+07,
     + 4.2442E+07, 4.2651E+07, 4.2860E+07, 4.3070E+07, 4.3281E+07,
     +  4.3493E+07, 4.3705E+07, 4.3919E+07, 4.4133E+07, 4.4349E+07,
     + 4.4565E+07, 4.4782E+07, 4.5000E+07, 4.5219E+07, 4.5439E+07,
     +  4.5660E+07, 4.5882E+07, 4.6104E+07, 4.6328E+07, 4.6552E+07,
     + 4.6777E+07, 4.7004E+07, 4.7231E+07, 4.7459E+07, 4.7688E+07,
     +  4.7918E+07, 4.8149E+07, 4.8381E+07, 4.8614E+07, 4.8848E+07,
     + 4.9083E+07, 4.9319E+07, 4.9555E+07, 4.9793E+07, 5.0032E+07,
     +  5.0271E+07, 5.0512E+07, 5.0753E+07, 5.0996E+07, 5.1240E+07,
     + 5.1484E+07, 5.1730E+07, 5.1976E+07, 5.2224E+07, 5.2472E+07,
     +  5.2722E+07, 5.2972E+07, 5.3224E+07, 5.3476E+07, 5.3730E+07,
     + 5.3985E+07, 5.4240E+07, 5.4497E+07, 5.4755E+07, 5.5013E+07,
     +  5.5273E+07, 5.5534E+07, 5.5796E+07, 5.6059E+07, 5.6323E+07,
     + 5.6588E+07, 5.6854E+07, 5.7121E+07, 5.7389E+07, 5.7659E+07,
     +  5.7929E+07, 5.8201E+07, 5.8473E+07, 5.8747E+07, 5.9021E+07,
     + 5.9297E+07, 5.9574E+07, 5.9852E+07, 6.0131E+07, 6.0411E+07,
     +  6.0693E+07, 6.0975E+07, 6.1259E+07, 6.1543E+07, 6.1829E+07,
     + 6.2116E+07, 6.2404E+07, 6.2693E+07, 6.2984E+07, 6.3275E+07,
     +  6.3568E+07, 6.3862E+07, 6.4157E+07, 6.4453E+07, 6.4750E+07,
     + 6.5048E+07, 6.5348E+07, 6.5649E+07, 6.5950E+07, 6.6254E+07,
     +  6.6558E+07, 6.6863E+07, 6.7170E+07, 6.7478E+07, 6.7787E+07,
     + 6.8097E+07, 6.8409E+07, 6.8721E+07, 6.9035E+07, 6.9350E+07,
     +  6.9667E+07, 6.9984E+07, 7.0303E+07, 7.0623E+07, 7.0944E+07,
     + 7.1267E+07, 7.1591E+07, 7.1916E+07, 7.2242E+07, 7.2569E+07,
     +  7.2898E+07, 7.3228E+07, 7.3559E+07, 7.3892E+07, 7.4226E+07,
     + 7.4561E+07, 7.4897E+07, 7.5235E+07, 7.5574E+07, 7.5914E+07,
     +  7.6256E+07, 7.6599E+07, 7.6943E+07, 7.7289E+07, 7.7635E+07,
     + 7.7984E+07, 7.8333E+07, 7.8684E+07, 7.9036E+07, 7.9390E+07,
     +  7.9744E+07, 8.0101E+07, 8.0458E+07, 8.0817E+07, 8.1177E+07,
     + 8.1539E+07, 8.1902E+07, 8.2266E+07, 8.2632E+07, 8.2999E+07,
     +  8.3367E+07, 8.3737E+07, 8.4109E+07, 8.4481E+07, 8.4855E+07,
     + 8.5231E+07, 8.5608E+07, 8.5986E+07, 8.6366E+07, 8.6747E+07,
     +  8.7129E+07, 8.7513E+07, 8.7899E+07, 8.8286E+07, 8.8674E+07,
     + 8.9064E+07, 8.9455E+07, 8.9848E+07, 9.0242E+07, 9.0638E+07,
     +  9.1035E+07, 9.1433E+07, 9.1833E+07, 9.2235E+07, 9.2638E+07,
     + 9.3042E+07, 9.3448E+07, 9.3856E+07, 9.4265E+07, 9.4675E+07,
     +  9.5087E+07, 9.5501E+07, 9.5916E+07, 9.6333E+07, 9.6751E+07,
     + 9.7170E+07, 9.7592E+07, 9.8015E+07, 9.8439E+07, 9.8865E+07,
     +  9.9292E+07, 9.9721E+07, 1.0015E+08, 1.0058E+08, 1.0102E+08,
     + 1.0145E+08, 1.0189E+08, 1.0233E+08, 1.0277E+08, 1.0321E+08,
     +  1.0365E+08, 1.0410E+08, 1.0455E+08, 1.0499E+08, 1.0544E+08,
     + 1.0590E+08, 1.0635E+08, 1.0680E+08, 1.0726E+08, 1.0772E+08,
     +  1.0818E+08, 1.0864E+08, 1.0910E+08, 1.0957E+08, 1.1003E+08,
     + 1.1050E+08, 1.1097E+08, 1.1144E+08, 1.1192E+08, 1.1239E+08,
     +  1.1287E+08, 1.1334E+08, 1.1382E+08, 1.1431E+08, 1.1479E+08,
     + 1.1528E+08, 1.1576E+08, 1.1625E+08, 1.1674E+08, 1.1723E+08,
     +  1.1773E+08, 1.1822E+08, 1.1872E+08, 1.1922E+08, 1.1972E+08,
     + 1.2022E+08, 1.2073E+08, 1.2123E+08, 1.2174E+08, 1.2225E+08,
     +  1.2276E+08, 1.2328E+08, 1.2379E+08, 1.2431E+08, 1.2483E+08,
     + 1.2535E+08, 1.2587E+08, 1.2640E+08, 1.2693E+08, 1.2745E+08,
     +  1.2798E+08, 1.2852E+08, 1.2905E+08, 1.2959E+08, 1.3012E+08,
     + 1.3066E+08, 1.3120E+08, 1.3175E+08, 1.3229E+08, 1.3284E+08,
     +  1.3339E+08, 1.3394E+08, 1.3449E+08, 1.3505E+08, 1.3561E+08,
     + 1.3616E+08, 1.3672E+08, 1.3729E+08, 1.3785E+08, 1.3842E+08,
     +  1.3899E+08, 1.3956E+08, 1.4013E+08, 1.4071E+08, 1.4128E+08,
     + 1.4186E+08, 1.4244E+08, 1.4302E+08, 1.4361E+08, 1.4419E+08,
     +  1.4478E+08, 1.4537E+08, 1.4597E+08, 1.4656E+08, 1.4716E+08,
     + 1.4776E+08, 1.4836E+08, 1.4896E+08, 1.4957E+08, 1.5017E+08,
     +  1.5078E+08, 1.5139E+08, 1.5201E+08, 1.5262E+08, 1.5324E+08,
     + 1.5386E+08, 1.5448E+08, 1.5511E+08, 1.5573E+08, 1.5636E+08,
     +  1.5699E+08, 1.5762E+08, 1.5826E+08, 1.5890E+08, 1.5953E+08,
     + 1.6018E+08, 1.6082E+08, 1.6147E+08, 1.6211E+08, 1.6276E+08,
     +  1.6342E+08, 1.6407E+08, 1.6473E+08, 1.6539E+08, 1.6605E+08,
     + 1.6671E+08, 1.6738E+08, 1.6804E+08, 1.6871E+08, 1.6939E+08,
     +  1.7006E+08, 1.7074E+08, 1.7142E+08, 1.7210E+08, 1.7278E+08,
     + 1.7347E+08, 1.7416E+08, 1.7485E+08, 1.7554E+08, 1.7624E+08,
     +  1.7693E+08, 1.7763E+08, 1.7834E+08, 1.7904E+08, 1.7975E+08,
     + 1.8046E+08, 1.8117E+08, 1.8188E+08, 1.8260E+08, 1.8332E+08,
     +  1.8404E+08, 1.8476E+08, 1.8549E+08, 1.8622E+08, 1.8695E+08,
     + 1.8768E+08, 1.8842E+08, 1.8916E+08, 1.8990E+08, 1.9064E+08,
     +  1.9139E+08, 1.9213E+08, 1.9288E+08, 1.9364E+08, 1.9439E+08,
     + 1.9515E+08, 1.9591E+08, 1.9667E+08, 1.9744E+08, 1.9821E+08,
     +  1.9898E+08, 1.9975E+08, 2.0053E+08, 2.0131E+08, 2.0209E+08,
     + 2.0287E+08, 2.0366E+08, 2.0444E+08, 2.0524E+08, 2.0603E+08,
     +  2.0683E+08, 2.0762E+08, 2.0843E+08, 2.0923E+08, 2.1004E+08,
     + 2.1085E+08, 2.1166E+08, 2.1247E+08, 2.1329E+08, 2.1411E+08,
     +  2.1493E+08, 2.1576E+08, 2.1659E+08, 2.1742E+08, 2.1825E+08,
     + 2.1909E+08, 2.1992E+08, 2.2077E+08, 2.2161E+08, 2.2246E+08,
     +  2.2331E+08, 2.2416E+08, 2.2501E+08, 2.2587E+08, 2.2673E+08,
     + 2.2760E+08, 2.2846E+08, 2.2933E+08, 2.3020E+08, 2.3108E+08,
     +  2.3195E+08, 2.3283E+08, 2.3372E+08, 2.3460E+08, 2.3549E+08,
     + 2.3638E+08, 2.3728E+08, 2.3818E+08, 2.3908E+08, 2.3998E+08,
     +  2.4088E+08, 2.4179E+08, 2.4270E+08, 2.4362E+08, 2.4454E+08,
     + 2.4546E+08, 2.4638E+08, 2.4731E+08, 2.4824E+08, 2.4917E+08,
     +  2.5010E+08, 2.5104E+08, 2.5198E+08, 2.5293E+08, 2.5387E+08,
     + 2.5482E+08, 2.5578E+08, 2.5673E+08, 2.5769E+08, 2.5865E+08,
     +  2.5962E+08, 2.6059E+08, 2.6156E+08, 2.6253E+08, 2.6351E+08,
     + 2.6449E+08, 2.6547E+08, 2.6646E+08, 2.6745E+08, 2.6844E+08,
     +  2.6944E+08, 2.7044E+08, 2.7144E+08, 2.7245E+08, 2.7345E+08,
     + 2.7447E+08, 2.7548E+08, 2.7650E+08, 2.7752E+08, 2.7854E+08,
     +  2.7957E+08, 2.8060E+08, 2.8164E+08, 2.8267E+08, 2.8371E+08,
     + 2.8476E+08, 2.8580E+08, 2.8685E+08, 2.8791E+08, 2.8896E+08,
     +  2.9002E+08, 2.9109E+08, 2.9215E+08, 2.9322E+08, 2.9430E+08,
     + 2.9537E+08, 2.9645E+08, 2.9753E+08, 2.9862E+08, 2.9971E+08,
     +  3.0080E+08, 3.0190E+08, 3.0300E+08, 3.0410E+08, 3.0521E+08,
     + 3.0632E+08, 3.0743E+08, 3.0855E+08, 3.0967E+08, 3.1079E+08,
     +  3.1192E+08, 3.1305E+08, 3.1418E+08, 3.1532E+08, 3.1646E+08,
     + 3.1761E+08, 3.1876E+08, 3.1991E+08, 3.2106E+08, 3.2222E+08,
     +  3.2338E+08, 3.2455E+08, 3.2572E+08, 3.2689E+08, 3.2807E+08,
     + 3.2925E+08, 3.3043E+08, 3.3162E+08, 3.3281E+08, 3.3400E+08,
     +  3.3520E+08, 3.3640E+08, 3.3761E+08, 3.3881E+08, 3.4003E+08,
     + 3.4124E+08, 3.4246E+08, 3.4369E+08, 3.4491E+08, 3.4614E+08,
     +  3.4738E+08, 3.4862E+08, 3.4986E+08, 3.5110E+08, 3.5235E+08,
     + 3.5361E+08, 3.5486E+08, 3.5612E+08, 3.5739E+08, 3.5866E+08,
     +  3.5993E+08, 3.6120E+08, 3.6248E+08, 3.6377E+08, 3.6505E+08,
     + 3.6635E+08, 3.6764E+08, 3.6894E+08, 3.7024E+08, 3.7155E+08,
     +  3.7286E+08, 3.7417E+08, 3.7549E+08, 3.7681E+08, 3.7814E+08,
     + 3.7947E+08, 3.8080E+08, 3.8214E+08, 3.8348E+08, 3.8483E+08,
     +  3.8618E+08, 3.8753E+08, 3.8889E+08, 3.9025E+08, 3.9162E+08,
     + 3.9299E+08, 3.9436E+08, 3.9574E+08, 3.9712E+08, 3.9851E+08,
     +  3.9990E+08, 4.0130E+08, 4.0269E+08, 4.0410E+08, 4.0550E+08,
     + 4.0691E+08, 4.0833E+08, 4.0975E+08, 4.1117E+08, 4.1260E+08,
     +  4.1403E+08, 4.1547E+08, 4.1691E+08, 4.1835E+08, 4.1980E+08,
     + 4.2125E+08, 4.2271E+08, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_C2N2  (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 2),QofT( 2,NT),Q(NT)
      data (Tmx(I),I=1,2)/3500.,3500./
      data xgj/ 1.,1./
c...     C2N2
c...        --      4224
      data (QofT( 1,J),J=1,2408)/ 2.1479E+01, 4.1351E+01, 6.1247E+01,
     +  8.1150E+01, 1.0105E+02, 1.2096E+02, 1.4087E+02, 1.6077E+02,
     + 1.8068E+02, 2.0059E+02, 2.2050E+02, 2.4041E+02, 2.6032E+02,
     +  2.8022E+02, 3.0013E+02, 3.2004E+02, 3.3995E+02, 3.5986E+02,
     + 3.7977E+02, 3.9968E+02, 4.3950E+02, 4.7932E+02, 5.1913E+02,
     +  5.5896E+02, 5.9879E+02, 6.3863E+02, 6.7848E+02, 7.1836E+02,
     + 7.5827E+02, 7.9823E+02, 8.3825E+02, 8.7835E+02, 9.1856E+02,
     +  9.5889E+02, 9.9937E+02, 1.0400E+03, 1.0809E+03, 1.1220E+03,
     + 1.1633E+03, 1.2049E+03, 1.2469E+03, 1.2892E+03, 1.3318E+03,
     +  1.3749E+03, 1.4184E+03, 1.4623E+03, 1.5068E+03, 1.5517E+03,
     + 1.5972E+03, 1.6433E+03, 1.6900E+03, 1.7372E+03, 1.7852E+03,
     +  1.8338E+03, 1.8831E+03, 1.9331E+03, 1.9839E+03, 2.0354E+03,
     + 2.0877E+03, 2.1408E+03, 2.1948E+03, 2.2496E+03, 2.3054E+03,
     +  2.3620E+03, 2.4195E+03, 2.4781E+03, 2.5376E+03, 2.5980E+03,
     + 2.6595E+03, 2.7221E+03, 2.7857E+03, 2.8504E+03, 2.9162E+03,
     +  2.9832E+03, 3.0513E+03, 3.1206E+03, 3.1911E+03, 3.2629E+03,
     + 3.3359E+03, 3.4101E+03, 3.4857E+03, 3.5626E+03, 3.6409E+03,
     +  3.7205E+03, 3.8015E+03, 3.8840E+03, 3.9679E+03, 4.0533E+03,
     + 4.1402E+03, 4.2286E+03, 4.3186E+03, 4.4102E+03, 4.5034E+03,
     +  4.5982E+03, 4.6947E+03, 4.7929E+03, 4.8928E+03, 4.9944E+03,
     + 5.0979E+03, 5.2031E+03, 5.3102E+03, 5.4191E+03, 5.5300E+03,
     +  5.6427E+03, 5.7575E+03, 5.8742E+03, 5.9930E+03, 6.1138E+03,
     + 6.2366E+03, 6.3617E+03, 6.4888E+03, 6.6181E+03, 6.7497E+03,
     +  6.8835E+03, 7.0196E+03, 7.1580E+03, 7.2987E+03, 7.4418E+03,
     + 7.5874E+03, 7.7355E+03, 7.8860E+03, 8.0390E+03, 8.1947E+03,
     +  8.3529E+03, 8.5138E+03, 8.6774E+03, 8.8437E+03, 9.0127E+03,
     + 9.1846E+03, 9.3593E+03, 9.5369E+03, 9.7174E+03, 9.9009E+03,
     +  1.0087E+04, 1.0277E+04, 1.0470E+04, 1.0665E+04, 1.0864E+04,
     + 1.1066E+04, 1.1272E+04, 1.1481E+04, 1.1693E+04, 1.1908E+04,
     +  1.2127E+04, 1.2350E+04, 1.2576E+04, 1.2805E+04, 1.3039E+04,
     + 1.3275E+04, 1.3516E+04, 1.3761E+04, 1.4009E+04, 1.4261E+04,
     +  1.4517E+04, 1.4777E+04, 1.5042E+04, 1.5310E+04, 1.5582E+04,
     + 1.5859E+04, 1.6140E+04, 1.6425E+04, 1.6715E+04, 1.7009E+04,
     +  1.7308E+04, 1.7611E+04, 1.7918E+04, 1.8231E+04, 1.8548E+04,
     + 1.8870E+04, 1.9196E+04, 1.9528E+04, 1.9865E+04, 2.0206E+04,
     +  2.0553E+04, 2.0905E+04, 2.1262E+04, 2.1624E+04, 2.1992E+04,
     + 2.2365E+04, 2.2744E+04, 2.3128E+04, 2.3518E+04, 2.3913E+04,
     +  2.4314E+04, 2.4721E+04, 2.5134E+04, 2.5553E+04, 2.5978E+04,
     + 2.6409E+04, 2.6847E+04, 2.7290E+04, 2.7740E+04, 2.8196E+04,
     +  2.8659E+04, 2.9128E+04, 2.9604E+04, 3.0087E+04, 3.0577E+04,
     + 3.1073E+04, 3.1576E+04, 3.2087E+04, 3.2604E+04, 3.3129E+04,
     +  3.3661E+04, 3.4200E+04, 3.4747E+04, 3.5302E+04, 3.5864E+04,
     + 3.6433E+04, 3.7011E+04, 3.7596E+04, 3.8190E+04, 3.8791E+04,
     +  3.9401E+04, 4.0019E+04, 4.0645E+04, 4.1280E+04, 4.1924E+04,
     + 4.2576E+04, 4.3236E+04, 4.3906E+04, 4.4584E+04, 4.5272E+04,
     +  4.5969E+04, 4.6675E+04, 4.7390E+04, 4.8115E+04, 4.8849E+04,
     + 4.9593E+04, 5.0347E+04, 5.1110E+04, 5.1884E+04, 5.2667E+04,
     +  5.3461E+04, 5.4265E+04, 5.5080E+04, 5.5905E+04, 5.6741E+04,
     + 5.7587E+04, 5.8445E+04, 5.9313E+04, 6.0193E+04, 6.1084E+04,
     +  6.1986E+04, 6.2899E+04, 6.3825E+04, 6.4761E+04, 6.5710E+04,
     + 6.6671E+04, 6.7644E+04, 6.8629E+04, 6.9626E+04, 7.0636E+04,
     +  7.1659E+04, 7.2694E+04, 7.3742E+04, 7.4803E+04, 7.5878E+04,
     + 7.6965E+04, 7.8066E+04, 7.9181E+04, 8.0309E+04, 8.1451E+04,
     +  8.2607E+04, 8.3778E+04, 8.4962E+04, 8.6161E+04, 8.7375E+04,
     + 8.8603E+04, 8.9846E+04, 9.1104E+04, 9.2378E+04, 9.3666E+04,
     +  9.4970E+04, 9.6290E+04, 9.7626E+04, 9.8977E+04, 1.0035E+05,
     + 1.0173E+05, 1.0313E+05, 1.0455E+05, 1.0598E+05, 1.0743E+05,
     +  1.0890E+05, 1.1038E+05, 1.1189E+05, 1.1341E+05, 1.1494E+05,
     + 1.1650E+05, 1.1807E+05, 1.1967E+05, 1.2128E+05, 1.2291E+05,
     +  1.2456E+05, 1.2622E+05, 1.2791E+05, 1.2962E+05, 1.3134E+05,
     + 1.3309E+05, 1.3485E+05, 1.3664E+05, 1.3845E+05, 1.4027E+05,
     +  1.4212E+05, 1.4399E+05, 1.4588E+05, 1.4779E+05, 1.4972E+05,
     + 1.5168E+05, 1.5365E+05, 1.5565E+05, 1.5767E+05, 1.5972E+05,
     +  1.6178E+05, 1.6387E+05, 1.6599E+05, 1.6812E+05, 1.7028E+05,
     + 1.7247E+05, 1.7467E+05, 1.7691E+05, 1.7916E+05, 1.8145E+05,
     +  1.8375E+05, 1.8609E+05, 1.8845E+05, 1.9083E+05, 1.9324E+05,
     + 1.9568E+05, 1.9814E+05, 2.0063E+05, 2.0314E+05, 2.0569E+05,
     +  2.0826E+05, 2.1086E+05, 2.1349E+05, 2.1614E+05, 2.1883E+05,
     + 2.2154E+05, 2.2428E+05, 2.2705E+05, 2.2985E+05, 2.3268E+05,
     +  2.3554E+05, 2.3843E+05, 2.4136E+05, 2.4431E+05, 2.4729E+05,
     + 2.5031E+05, 2.5335E+05, 2.5643E+05, 2.5955E+05, 2.6269E+05,
     +  2.6587E+05, 2.6908E+05, 2.7232E+05, 2.7560E+05, 2.7891E+05,
     + 2.8226E+05, 2.8564E+05, 2.8905E+05, 2.9250E+05, 2.9599E+05,
     +  2.9951E+05, 3.0307E+05, 3.0667E+05, 3.1030E+05, 3.1397E+05,
     + 3.1768E+05, 3.2143E+05, 3.2521E+05, 3.2903E+05, 3.3289E+05,
     +  3.3679E+05, 3.4073E+05, 3.4471E+05, 3.4873E+05, 3.5279E+05,
     + 3.5689E+05, 3.6104E+05, 3.6522E+05, 3.6945E+05, 3.7372E+05,
     +  3.7803E+05, 3.8238E+05, 3.8678E+05, 3.9122E+05, 3.9571E+05,
     + 4.0024E+05, 4.0481E+05, 4.0943E+05, 4.1410E+05, 4.1881E+05,
     +  4.2357E+05, 4.2838E+05, 4.3323E+05, 4.3813E+05, 4.4308E+05,
     + 4.4808E+05, 4.5313E+05, 4.5822E+05, 4.6337E+05, 4.6856E+05,
     +  4.7381E+05, 4.7910E+05, 4.8445E+05, 4.8985E+05, 4.9530E+05,
     + 5.0081E+05, 5.0637E+05, 5.1198E+05, 5.1764E+05, 5.2336E+05,
     +  5.2914E+05, 5.3497E+05, 5.4085E+05, 5.4679E+05, 5.5279E+05,
     + 5.5885E+05, 5.6496E+05, 5.7113E+05, 5.7736E+05, 5.8365E+05,
     +  5.9000E+05, 5.9641E+05, 6.0287E+05, 6.0940E+05, 6.1599E+05,
     + 6.2265E+05, 6.2936E+05, 6.3614E+05, 6.4298E+05, 6.4988E+05,
     +  6.5685E+05, 6.6389E+05, 6.7099E+05, 6.7815E+05, 6.8539E+05,
     + 6.9268E+05, 7.0005E+05, 7.0749E+05, 7.1499E+05, 7.2256E+05,
     +  7.3020E+05, 7.3792E+05, 7.4570E+05, 7.5355E+05, 7.6148E+05,
     + 7.6948E+05, 7.7755E+05, 7.8570E+05, 7.9392E+05, 8.0221E+05,
     +  8.1058E+05, 8.1903E+05, 8.2755E+05, 8.3615E+05, 8.4483E+05,
     + 8.5358E+05, 8.6242E+05, 8.7133E+05, 8.8032E+05, 8.8940E+05,
     +  8.9856E+05, 9.0779E+05, 9.1711E+05, 9.2652E+05, 9.3601E+05,
     + 9.4558E+05, 9.5524E+05, 9.6498E+05, 9.7481E+05, 9.8473E+05,
     +  9.9473E+05, 1.0048E+06, 1.0150E+06, 1.0253E+06, 1.0356E+06,
     + 1.0461E+06, 1.0566E+06, 1.0673E+06, 1.0780E+06, 1.0888E+06,
     +  1.0998E+06, 1.1108E+06, 1.1219E+06, 1.1331E+06, 1.1444E+06,
     + 1.1558E+06, 1.1673E+06, 1.1789E+06, 1.1906E+06, 1.2024E+06,
     +  1.2143E+06, 1.2263E+06, 1.2384E+06, 1.2506E+06, 1.2630E+06,
     + 1.2754E+06, 1.2879E+06, 1.3006E+06, 1.3133E+06, 1.3262E+06,
     +  1.3391E+06, 1.3522E+06, 1.3654E+06, 1.3787E+06, 1.3921E+06,
     + 1.4056E+06, 1.4193E+06, 1.4330E+06, 1.4469E+06, 1.4609E+06,
     +  1.4750E+06, 1.4892E+06, 1.5036E+06, 1.5181E+06, 1.5326E+06,
     + 1.5474E+06, 1.5622E+06, 1.5772E+06, 1.5922E+06, 1.6074E+06,
     +  1.6228E+06, 1.6382E+06, 1.6538E+06, 1.6695E+06, 1.6854E+06,
     + 1.7014E+06, 1.7175E+06, 1.7337E+06, 1.7501E+06, 1.7666E+06,
     +  1.7833E+06, 1.8001E+06, 1.8170E+06, 1.8341E+06, 1.8513E+06,
     + 1.8686E+06, 1.8861E+06, 1.9037E+06, 1.9215E+06, 1.9394E+06,
     +  1.9575E+06, 1.9757E+06, 1.9941E+06, 2.0126E+06, 2.0312E+06,
     + 2.0500E+06, 2.0690E+06, 2.0881E+06, 2.1074E+06, 2.1268E+06,
     +  2.1464E+06, 2.1661E+06, 2.1860E+06, 2.2061E+06, 2.2263E+06,
     + 2.2467E+06, 2.2672E+06, 2.2879E+06, 2.3088E+06, 2.3298E+06,
     +  2.3510E+06, 2.3724E+06, 2.3939E+06, 2.4156E+06, 2.4375E+06,
     + 2.4595E+06, 2.4818E+06, 2.5042E+06, 2.5268E+06, 2.5495E+06,
     +  2.5725E+06, 2.5956E+06, 2.6189E+06, 2.6423E+06, 2.6660E+06,
     + 2.6899E+06, 2.7139E+06, 2.7381E+06, 2.7625E+06, 2.7871E+06,
     +  2.8119E+06, 2.8369E+06, 2.8621E+06, 2.8875E+06, 2.9130E+06,
     + 2.9388E+06, 2.9648E+06, 2.9909E+06, 3.0173E+06, 3.0439E+06,
     +  3.0707E+06, 3.0976E+06, 3.1248E+06, 3.1522E+06, 3.1798E+06,
     + 3.2077E+06, 3.2357E+06, 3.2639E+06, 3.2924E+06, 3.3211E+06,
     +  3.3500E+06, 3.3791E+06, 3.4084E+06, 3.4380E+06, 3.4678E+06,
     + 3.4978E+06, 3.5280E+06, 3.5585E+06, 3.5892E+06, 3.6201E+06,
     +  3.6513E+06, 3.6827E+06, 3.7143E+06, 3.7462E+06, 3.7783E+06,
     + 3.8106E+06, 3.8432E+06, 3.8760E+06, 3.9091E+06, 3.9424E+06,
     +  3.9760E+06, 4.0098E+06, 4.0439E+06, 4.0782E+06, 4.1128E+06,
     + 4.1476E+06, 4.1827E+06, 4.2181E+06, 4.2537E+06, 4.2896E+06,
     +  4.3257E+06, 4.3621E+06, 4.3988E+06, 4.4357E+06, 4.4729E+06,
     + 4.5104E+06, 4.5482E+06, 4.5862E+06, 4.6246E+06, 4.6631E+06,
     +  4.7020E+06, 4.7412E+06, 4.7806E+06, 4.8204E+06, 4.8604E+06,
     + 4.9007E+06, 4.9413E+06, 4.9822E+06, 5.0234E+06, 5.0649E+06,
     +  5.1067E+06, 5.1488E+06, 5.1912E+06, 5.2339E+06, 5.2769E+06,
     + 5.3202E+06, 5.3638E+06, 5.4078E+06, 5.4521E+06, 5.4966E+06,
     +  5.5415E+06, 5.5867E+06, 5.6323E+06, 5.6781E+06, 5.7243E+06,
     + 5.7708E+06, 5.8177E+06, 5.8649E+06, 5.9124E+06, 5.9602E+06,
     +  6.0084E+06, 6.0569E+06, 6.1058E+06, 6.1550E+06, 6.2046E+06,
     + 6.2545E+06, 6.3048E+06, 6.3554E+06, 6.4064E+06, 6.4577E+06,
     +  6.5094E+06, 6.5614E+06, 6.6138E+06, 6.6666E+06, 6.7198E+06,
     + 6.7733E+06, 6.8272E+06, 6.8815E+06, 6.9361E+06, 6.9911E+06,
     +  7.0465E+06, 7.1023E+06, 7.1585E+06, 7.2151E+06, 7.2720E+06,
     + 7.3294E+06, 7.3871E+06, 7.4453E+06, 7.5038E+06, 7.5628E+06,
     +  7.6221E+06, 7.6819E+06, 7.7420E+06, 7.8026E+06, 7.8636E+06,
     + 7.9250E+06, 7.9869E+06, 8.0491E+06, 8.1118E+06, 8.1749E+06,
     +  8.2384E+06, 8.3024E+06, 8.3668E+06, 8.4316E+06, 8.4969E+06,
     + 8.5626E+06, 8.6288E+06, 8.6954E+06, 8.7625E+06, 8.8300E+06,
     +  8.8980E+06, 8.9664E+06, 9.0353E+06, 9.1046E+06, 9.1745E+06,
     + 9.2447E+06, 9.3155E+06, 9.3867E+06, 9.4584E+06, 9.5306E+06,
     +  9.6033E+06, 9.6765E+06, 9.7501E+06, 9.8242E+06, 9.8988E+06,
     + 9.9740E+06, 1.0050E+07, 1.0126E+07, 1.0202E+07, 1.0279E+07,
     +  1.0357E+07, 1.0435E+07, 1.0514E+07, 1.0593E+07, 1.0673E+07,
     + 1.0753E+07, 1.0834E+07, 1.0915E+07, 1.0997E+07, 1.1079E+07,
     +  1.1162E+07, 1.1246E+07, 1.1330E+07, 1.1414E+07, 1.1499E+07,
     + 1.1585E+07, 1.1671E+07, 1.1758E+07, 1.1845E+07, 1.1933E+07,
     +  1.2022E+07, 1.2111E+07, 1.2200E+07, 1.2291E+07, 1.2381E+07,
     + 1.2473E+07, 1.2565E+07, 1.2657E+07, 1.2750E+07, 1.2844E+07,
     +  1.2939E+07, 1.3034E+07, 1.3129E+07, 1.3225E+07, 1.3322E+07,
     + 1.3420E+07, 1.3518E+07, 1.3616E+07, 1.3716E+07, 1.3816E+07,
     +  1.3916E+07, 1.4017E+07, 1.4119E+07, 1.4222E+07, 1.4325E+07,
     + 1.4429E+07, 1.4533E+07, 1.4638E+07, 1.4744E+07, 1.4851E+07,
     +  1.4958E+07, 1.5066E+07, 1.5174E+07, 1.5283E+07, 1.5393E+07,
     + 1.5504E+07, 1.5615E+07, 1.5727E+07, 1.5840E+07, 1.5953E+07,
     +  1.6067E+07, 1.6182E+07, 1.6297E+07, 1.6414E+07, 1.6531E+07,
     + 1.6648E+07, 1.6767E+07, 1.6886E+07, 1.7006E+07, 1.7126E+07,
     +  1.7248E+07, 1.7370E+07, 1.7493E+07, 1.7617E+07, 1.7741E+07,
     + 1.7866E+07, 1.7992E+07, 1.8119E+07, 1.8247E+07, 1.8375E+07,
     +  1.8504E+07, 1.8634E+07, 1.8765E+07, 1.8896E+07, 1.9028E+07,
     + 1.9162E+07, 1.9296E+07, 1.9430E+07, 1.9566E+07, 1.9702E+07,
     +  1.9840E+07, 1.9978E+07, 2.0117E+07, 2.0256E+07, 2.0397E+07,
     + 2.0538E+07, 2.0681E+07, 2.0824E+07, 2.0968E+07, 2.1113E+07,
     +  2.1259E+07, 2.1406E+07, 2.1553E+07, 2.1702E+07, 2.1851E+07,
     + 2.2001E+07, 2.2152E+07, 2.2305E+07, 2.2458E+07, 2.2611E+07,
     +  2.2766E+07, 2.2922E+07, 2.3079E+07, 2.3236E+07, 2.3395E+07,
     + 2.3555E+07, 2.3715E+07, 2.3877E+07, 2.4039E+07, 2.4202E+07,
     +  2.4367E+07, 2.4532E+07, 2.4698E+07, 2.4866E+07, 2.5034E+07,
     + 2.5203E+07, 2.5373E+07, 2.5545E+07, 2.5717E+07, 2.5890E+07,
     +  2.6065E+07, 2.6240E+07, 2.6417E+07, 2.6594E+07, 2.6772E+07,
     + 2.6952E+07, 2.7133E+07, 2.7314E+07, 2.7497E+07, 2.7681E+07,
     +  2.7866E+07, 2.8052E+07, 2.8239E+07, 2.8427E+07, 2.8616E+07,
     + 2.8806E+07, 2.8998E+07, 2.9190E+07, 2.9384E+07, 2.9579E+07,
     +  2.9775E+07, 2.9972E+07, 3.0170E+07, 3.0369E+07, 3.0570E+07,
     + 3.0771E+07, 3.0974E+07, 3.1178E+07, 3.1383E+07, 3.1590E+07,
     +  3.1797E+07, 3.2006E+07, 3.2216E+07, 3.2427E+07, 3.2640E+07,
     + 3.2853E+07, 3.3068E+07, 3.3284E+07, 3.3501E+07, 3.3720E+07,
     +  3.3939E+07, 3.4160E+07, 3.4383E+07, 3.4606E+07, 3.4831E+07,
     + 3.5057E+07, 3.5284E+07, 3.5513E+07, 3.5743E+07, 3.5974E+07,
     +  3.6207E+07, 3.6441E+07, 3.6676E+07, 3.6913E+07, 3.7150E+07,
     + 3.7390E+07, 3.7630E+07, 3.7872E+07, 3.8115E+07, 3.8360E+07,
     +  3.8606E+07, 3.8854E+07, 3.9102E+07, 3.9353E+07, 3.9604E+07,
     + 3.9857E+07, 4.0112E+07, 4.0367E+07, 4.0625E+07, 4.0883E+07,
     +  4.1144E+07, 4.1405E+07, 4.1668E+07, 4.1933E+07, 4.2199E+07,
     + 4.2466E+07, 4.2735E+07, 4.3006E+07, 4.3278E+07, 4.3551E+07,
     +  4.3826E+07, 4.4102E+07, 4.4380E+07, 4.4660E+07, 4.4941E+07,
     + 4.5224E+07, 4.5508E+07, 4.5794E+07, 4.6081E+07, 4.6370E+07,
     +  4.6660E+07, 4.6952E+07, 4.7246E+07, 4.7541E+07, 4.7838E+07,
     + 4.8137E+07, 4.8437E+07, 4.8739E+07, 4.9042E+07, 4.9347E+07,
     +  4.9654E+07, 4.9963E+07, 5.0273E+07, 5.0585E+07, 5.0898E+07,
     + 5.1213E+07, 5.1530E+07, 5.1849E+07, 5.2169E+07, 5.2491E+07,
     +  5.2815E+07, 5.3141E+07, 5.3468E+07, 5.3797E+07, 5.4128E+07,
     + 5.4461E+07, 5.4795E+07, 5.5131E+07, 5.5469E+07, 5.5809E+07,
     +  5.6151E+07, 5.6495E+07, 5.6840E+07, 5.7187E+07, 5.7536E+07,
     + 5.7887E+07, 5.8240E+07, 5.8595E+07, 5.8951E+07, 5.9310E+07,
     +  5.9670E+07, 6.0032E+07, 6.0397E+07, 6.0763E+07, 6.1131E+07,
     + 6.1501E+07, 6.1873E+07, 6.2247E+07, 6.2623E+07, 6.3001E+07,
     +  6.3381E+07, 6.3763E+07, 6.4147E+07, 6.4533E+07, 6.4921E+07,
     + 6.5311E+07, 6.5703E+07, 6.6098E+07, 6.6494E+07, 6.6892E+07,
     +  6.7293E+07, 6.7695E+07, 6.8100E+07, 6.8507E+07, 6.8916E+07,
     + 6.9327E+07, 6.9740E+07, 7.0155E+07, 7.0573E+07, 7.0992E+07,
     +  7.1414E+07, 7.1838E+07, 7.2265E+07, 7.2693E+07, 7.3124E+07,
     + 7.3557E+07, 7.3992E+07, 7.4430E+07, 7.4869E+07, 7.5311E+07,
     +  7.5756E+07, 7.6202E+07, 7.6651E+07, 7.7102E+07, 7.7556E+07,
     + 7.8012E+07, 7.8470E+07, 7.8930E+07, 7.9393E+07, 7.9859E+07,
     +  8.0326E+07, 8.0797E+07, 8.1269E+07, 8.1744E+07, 8.2221E+07,
     + 8.2701E+07, 8.3183E+07, 8.3668E+07, 8.4155E+07, 8.4645E+07,
     +  8.5137E+07, 8.5632E+07, 8.6129E+07, 8.6629E+07, 8.7131E+07,
     + 8.7636E+07, 8.8143E+07, 8.8653E+07, 8.9166E+07, 8.9681E+07,
     +  9.0199E+07, 9.0719E+07, 9.1242E+07, 9.1768E+07, 9.2296E+07,
     + 9.2827E+07, 9.3360E+07, 9.3897E+07, 9.4436E+07, 9.4977E+07,
     +  9.5522E+07, 9.6069E+07, 9.6619E+07, 9.7172E+07, 9.7727E+07,
     + 9.8285E+07, 9.8846E+07, 9.9410E+07, 9.9977E+07, 1.0055E+08,
     +  1.0112E+08, 1.0169E+08, 1.0227E+08, 1.0285E+08, 1.0344E+08,
     + 1.0402E+08, 1.0461E+08, 1.0521E+08, 1.0580E+08, 1.0640E+08,
     +  1.0700E+08, 1.0760E+08, 1.0821E+08, 1.0882E+08, 1.0944E+08,
     + 1.1005E+08, 1.1067E+08, 1.1129E+08, 1.1192E+08, 1.1255E+08,
     +  1.1318E+08, 1.1382E+08, 1.1445E+08, 1.1509E+08, 1.1574E+08,
     + 1.1639E+08, 1.1704E+08, 1.1769E+08, 1.1835E+08, 1.1901E+08,
     +  1.1967E+08, 1.2034E+08, 1.2101E+08, 1.2168E+08, 1.2236E+08,
     + 1.2304E+08, 1.2372E+08, 1.2441E+08, 1.2510E+08, 1.2579E+08,
     +  1.2648E+08, 1.2718E+08, 1.2789E+08, 1.2859E+08, 1.2930E+08,
     + 1.3002E+08, 1.3073E+08, 1.3145E+08, 1.3218E+08, 1.3291E+08,
     +  1.3364E+08, 1.3437E+08, 1.3511E+08, 1.3585E+08, 1.3660E+08,
     + 1.3734E+08, 1.3810E+08, 1.3885E+08, 1.3961E+08, 1.4037E+08,
     +  1.4114E+08, 1.4191E+08, 1.4268E+08, 1.4346E+08, 1.4424E+08,
     + 1.4503E+08, 1.4582E+08, 1.4661E+08, 1.4741E+08, 1.4821E+08,
     +  1.4901E+08, 1.4982E+08, 1.5063E+08, 1.5145E+08, 1.5227E+08,
     + 1.5309E+08, 1.5392E+08, 1.5475E+08, 1.5558E+08, 1.5642E+08,
     +  1.5726E+08, 1.5811E+08, 1.5896E+08, 1.5982E+08, 1.6068E+08,
     + 1.6154E+08, 1.6241E+08, 1.6328E+08, 1.6415E+08, 1.6503E+08,
     +  1.6591E+08, 1.6680E+08, 1.6769E+08, 1.6859E+08, 1.6949E+08,
     + 1.7039E+08, 1.7130E+08, 1.7222E+08, 1.7313E+08, 1.7405E+08,
     +  1.7498E+08, 1.7591E+08, 1.7684E+08, 1.7778E+08, 1.7872E+08,
     + 1.7967E+08, 1.8062E+08, 1.8158E+08, 1.8254E+08, 1.8351E+08,
     +  1.8447E+08, 1.8545E+08, 1.8643E+08, 1.8741E+08, 1.8840E+08,
     + 1.8939E+08, 1.9039E+08, 1.9139E+08, 1.9239E+08, 1.9340E+08,
     +  1.9442E+08, 1.9544E+08, 1.9646E+08, 1.9749E+08, 1.9853E+08,
     + 1.9957E+08, 2.0061E+08, 2.0166E+08, 2.0271E+08, 2.0377E+08,
     +  2.0483E+08, 2.0590E+08, 2.0697E+08, 2.0805E+08, 2.0913E+08,
     + 2.1022E+08, 2.1131E+08, 2.1241E+08, 2.1351E+08, 2.1462E+08,
     +  2.1573E+08, 2.1684E+08, 2.1797E+08, 2.1909E+08, 2.2023E+08,
     + 2.2136E+08, 2.2251E+08, 2.2365E+08, 2.2481E+08, 2.2597E+08,
     +  2.2713E+08, 2.2830E+08, 2.2947E+08, 2.3065E+08, 2.3183E+08,
     + 2.3302E+08, 2.3422E+08, 2.3542E+08, 2.3663E+08, 2.3784E+08,
     +  2.3905E+08, 2.4028E+08, 2.4150E+08, 2.4274E+08, 2.4398E+08,
     + 2.4522E+08, 2.4647E+08, 2.4773E+08, 2.4899E+08, 2.5025E+08,
     +  2.5153E+08, 2.5280E+08, 2.5409E+08, 2.5538E+08, 2.5667E+08,
     + 2.5797E+08, 2.5928E+08, 2.6059E+08, 2.6191E+08, 2.6323E+08,
     +  2.6456E+08, 2.6590E+08, 2.6724E+08, 2.6859E+08, 2.6994E+08,
     + 2.7130E+08, 2.7267E+08, 2.7404E+08, 2.7542E+08, 2.7680E+08,
     +  2.7819E+08, 2.7959E+08, 2.8099E+08, 2.8240E+08, 2.8381E+08,
     + 2.8523E+08, 2.8666E+08, 2.8809E+08, 2.8953E+08, 2.9098E+08,
     +  2.9243E+08, 2.9389E+08, 2.9535E+08, 2.9682E+08, 2.9830E+08,
     + 2.9978E+08, 3.0127E+08, 3.0277E+08, 3.0427E+08, 3.0578E+08,
     +  3.0730E+08, 3.0882E+08, 3.1035E+08, 3.1189E+08, 3.1343E+08,
     + 3.1498E+08, 3.1654E+08, 3.1810E+08, 3.1967E+08, 3.2125E+08,
     +  3.2283E+08, 3.2442E+08, 3.2602E+08, 3.2762E+08, 3.2923E+08,
     + 3.3085E+08, 3.3247E+08, 3.3411E+08, 3.3575E+08, 3.3739E+08,
     +  3.3904E+08, 3.4070E+08, 3.4237E+08, 3.4405E+08, 3.4573E+08,
     + 3.4742E+08, 3.4911E+08, 3.5081E+08, 3.5252E+08, 3.5424E+08,
     +  3.5597E+08, 3.5770E+08, 3.5944E+08, 3.6119E+08, 3.6294E+08,
     + 3.6470E+08, 3.6647E+08, 3.6825E+08, 3.7003E+08, 3.7183E+08,
     +  3.7363E+08, 3.7543E+08, 3.7725E+08, 3.7907E+08, 3.8090E+08,
     + 3.8274E+08, 3.8459E+08, 3.8644E+08, 3.8830E+08, 3.9017E+08,
     +  3.9205E+08, 3.9393E+08, 3.9583E+08, 3.9773E+08, 3.9964E+08,
     + 4.0155E+08, 4.0348E+08, 4.0541E+08, 4.0735E+08, 4.0930E+08,
     +  4.1126E+08, 4.1323E+08, 4.1520E+08, 4.1718E+08, 4.1918E+08,
     + 4.2117E+08, 4.2318E+08, 4.2520E+08, 4.2722E+08, 4.2925E+08,
     +  4.3130E+08, 4.3335E+08, 4.3540E+08, 4.3747E+08, 4.3955E+08,
     + 4.4163E+08, 4.4372E+08, 4.4582E+08, 4.4793E+08, 4.5005E+08,
     +  4.5218E+08, 4.5432E+08, 4.5646E+08, 4.5861E+08, 4.6078E+08,
     + 4.6295E+08, 4.6513E+08, 4.6732E+08, 4.6952E+08, 4.7173E+08,
     +  4.7394E+08, 4.7617E+08, 4.7841E+08, 4.8065E+08, 4.8290E+08,
     + 4.8517E+08, 4.8744E+08, 4.8972E+08, 4.9201E+08, 4.9431E+08,
     +  4.9662E+08, 4.9894E+08, 5.0127E+08, 5.0361E+08, 5.0595E+08,
     + 5.0831E+08, 5.1068E+08, 5.1306E+08, 5.1544E+08, 5.1784E+08,
     +  5.2024E+08, 5.2266E+08, 5.2508E+08, 5.2752E+08, 5.2996E+08,
     + 5.3242E+08, 5.3488E+08, 5.3736E+08, 5.3984E+08, 5.4234E+08,
     +  5.4484E+08, 5.4736E+08, 5.4988E+08, 5.5242E+08, 5.5497E+08,
     + 5.5752E+08, 5.6009E+08, 5.6267E+08, 5.6525E+08, 5.6785E+08,
     +  5.7046E+08, 5.7308E+08, 5.7571E+08, 5.7835E+08, 5.8100E+08,
     + 5.8366E+08, 5.8633E+08, 5.8901E+08, 5.9170E+08, 5.9441E+08,
     +  5.9712E+08, 5.9985E+08, 6.0258E+08, 6.0533E+08, 6.0809E+08,
     + 6.1086E+08, 6.1364E+08, 6.1643E+08, 6.1923E+08, 6.2205E+08,
     +  6.2487E+08, 6.2771E+08, 6.3055E+08, 6.3341E+08, 6.3628E+08,
     + 6.3916E+08, 6.4206E+08, 6.4496E+08, 6.4788E+08, 6.5080E+08,
     +  6.5374E+08, 6.5669E+08, 6.5966E+08, 6.6263E+08, 6.6562E+08,
     + 6.6861E+08, 6.7162E+08, 6.7464E+08, 6.7768E+08, 6.8072E+08,
     +  6.8378E+08, 6.8685E+08, 6.8993E+08, 6.9302E+08, 6.9613E+08,
     + 6.9924E+08, 7.0237E+08, 7.0552E+08, 7.0867E+08, 7.1184E+08,
     +  7.1502E+08, 7.1821E+08, 7.2141E+08, 7.2463E+08, 7.2786E+08,
     + 7.3110E+08, 7.3435E+08, 7.3762E+08, 7.4090E+08, 7.4419E+08,
     +  7.4750E+08, 7.5082E+08, 7.5415E+08, 7.5749E+08, 7.6085E+08,
     + 7.6422E+08, 7.6760E+08, 7.7100E+08, 7.7441E+08, 7.7783E+08,
     +  7.8126E+08, 7.8471E+08, 7.8817E+08, 7.9165E+08, 7.9514E+08,
     + 7.9864E+08, 8.0216E+08, 8.0569E+08, 8.0923E+08, 8.1279E+08,
     +  8.1636E+08, 8.1994E+08, 8.2354E+08, 8.2715E+08, 8.3078E+08,
     + 8.3442E+08, 8.3807E+08, 8.4174E+08, 8.4542E+08, 8.4912E+08,
     +  8.5283E+08, 8.5655E+08, 8.6029E+08, 8.6404E+08, 8.6781E+08,
     + 8.7159E+08, 8.7539E+08, 8.7920E+08, 8.8302E+08, 8.8686E+08,
     +  8.9072E+08, 8.9459E+08, 8.9847E+08, 9.0237E+08, 9.0628E+08,
     + 9.1021E+08, 9.1415E+08, 9.1811E+08, 9.2208E+08, 9.2607E+08,
     +  9.3007E+08, 9.3409E+08, 9.3812E+08, 9.4217E+08, 9.4624E+08,
     + 9.5032E+08, 9.5441E+08, 9.5852E+08, 9.6265E+08, 9.6679E+08,
     +  9.7094E+08, 9.7512E+08, 9.7930E+08, 9.8351E+08, 9.8773E+08,
     + 9.9196E+08, 9.9621E+08, 1.0005E+09, 1.0048E+09, 1.0091E+09,
     +  1.0134E+09, 1.0177E+09, 1.0221E+09, 1.0264E+09, 1.0308E+09,
     + 1.0352E+09, 1.0396E+09, 1.0440E+09, 1.0485E+09, 1.0530E+09,
     +  1.0574E+09, 1.0619E+09, 1.0664E+09, 1.0710E+09, 1.0755E+09,
     + 1.0801E+09, 1.0847E+09, 1.0893E+09, 1.0939E+09, 1.0985E+09,
     +  1.1032E+09, 1.1078E+09, 1.1125E+09, 1.1172E+09, 1.1219E+09,
     + 1.1267E+09, 1.1314E+09, 1.1362E+09, 1.1410E+09, 1.1458E+09,
     +  1.1506E+09, 1.1554E+09, 1.1603E+09, 1.1652E+09, 1.1701E+09,
     + 1.1750E+09, 1.1799E+09, 1.1849E+09, 1.1898E+09, 1.1948E+09,
     +  1.1998E+09, 1.2048E+09, 1.2099E+09, 1.2149E+09, 1.2200E+09,
     + 1.2251E+09, 1.2302E+09, 1.2354E+09, 1.2405E+09, 1.2457E+09,
     +  1.2509E+09, 1.2561E+09, 1.2613E+09, 1.2666E+09, 1.2718E+09,
     + 1.2771E+09, 1.2824E+09, 1.2877E+09, 1.2931E+09, 1.2984E+09,
     +  1.3038E+09, 1.3092E+09, 1.3147E+09, 1.3201E+09, 1.3256E+09,
     + 1.3310E+09, 1.3365E+09, 1.3421E+09, 1.3476E+09, 1.3532E+09,
     +  1.3587E+09, 1.3643E+09, 1.3700E+09, 1.3756E+09, 1.3813E+09,
     + 1.3869E+09, 1.3926E+09, 1.3984E+09, 1.4041E+09, 1.4099E+09,
     +  1.4157E+09, 1.4215E+09, 1.4273E+09, 1.4331E+09, 1.4390E+09,
     + 1.4449E+09, 1.4508E+09, 1.4567E+09, 1.4627E+09, 1.4687E+09,
     +  1.4746E+09, 1.4807E+09, 1.4867E+09, 1.4928E+09, 1.4988E+09,
     + 1.5049E+09, 1.5111E+09, 1.5172E+09, 1.5234E+09, 1.5296E+09,
     +  1.5358E+09, 1.5420E+09, 1.5483E+09, 1.5546E+09, 1.5609E+09,
     + 1.5672E+09, 1.5735E+09, 1.5799E+09, 1.5863E+09, 1.5927E+09,
     +  1.5991E+09, 1.6056E+09, 1.6121E+09, 1.6186E+09, 1.6251E+09,
     + 1.6317E+09, 1.6382E+09, 1.6448E+09, 1.6515E+09, 1.6581E+09,
     +  1.6648E+09, 1.6715E+09, 1.6782E+09, 1.6849E+09, 1.6917E+09,
     + 1.6985E+09, 1.7053E+09, 1.7121E+09, 1.7190E+09, 1.7258E+09,
     +  1.7327E+09, 1.7397E+09, 1.7466E+09, 1.7536E+09, 1.7606E+09,
     + 1.7676E+09, 1.7747E+09, 1.7818E+09, 1.7889E+09, 1.7960E+09,
     +  1.8031E+09, 1.8103E+09, 1.8175E+09, 1.8247E+09, 1.8320E+09,
     + 1.8393E+09, 1.8466E+09, 1.8539E+09, 1.8613E+09, 1.8686E+09,
     +  1.8760E+09, 1.8835E+09, 1.8909E+09, 1.8984E+09, 1.9059E+09,
     + 1.9134E+09, 1.9210E+09, 1.9286E+09, 1.9362E+09, 1.9438E+09,
     +  1.9515E+09, 1.9592E+09, 1.9669E+09, 1.9746E+09, 1.9824E+09,
     + 1.9902E+09, 1.9980E+09, 2.0059E+09, 2.0138E+09, 2.0217E+09,
     +  2.0296E+09, 2.0376E+09, 2.0455E+09, 2.0536E+09, 2.0616E+09,
     + 2.0697E+09, 2.0778E+09, 2.0859E+09, 2.0940E+09, 2.1022E+09,
     +  2.1104E+09, 2.1187E+09, 2.1269E+09, 2.1352E+09, 2.1435E+09,
     + 2.1519E+09, 2.1603E+09, 2.1687E+09, 2.1771E+09, 2.1856E+09,
     +  2.1941E+09, 2.2026E+09, 2.2111E+09, 2.2197E+09, 2.2283E+09,
     + 2.2369E+09, 2.2456E+09, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --      5225
      data (QofT( 2,J),J=1,2408)/ 1.0106E+01, 1.9501E+01, 2.8906E+01,
     +  3.8313E+01, 4.7722E+01, 5.7131E+01, 6.6540E+01, 7.5949E+01,
     + 8.5359E+01, 9.4769E+01, 1.0418E+02, 1.1359E+02, 1.2300E+02,
     +  1.3241E+02, 1.4182E+02, 1.5123E+02, 1.6064E+02, 1.7005E+02,
     + 1.7946E+02, 1.8887E+02, 2.0769E+02, 2.2651E+02, 2.4533E+02,
     +  2.6416E+02, 2.8298E+02, 3.0181E+02, 3.2065E+02, 3.3950E+02,
     + 3.5836E+02, 3.7725E+02, 3.9617E+02, 4.1512E+02, 4.3413E+02,
     +  4.5319E+02, 4.7232E+02, 4.9154E+02, 5.1085E+02, 5.3026E+02,
     + 5.4980E+02, 5.6948E+02, 5.8930E+02, 6.0929E+02, 6.2945E+02,
     +  6.4981E+02, 6.7037E+02, 6.9115E+02, 7.1215E+02, 7.3340E+02,
     + 7.5491E+02, 7.7668E+02, 7.9873E+02, 8.2108E+02, 8.4374E+02,
     +  8.6671E+02, 8.9001E+02, 9.1365E+02, 9.3765E+02, 9.6200E+02,
     + 9.8672E+02, 1.0118E+03, 1.0374E+03, 1.0633E+03, 1.0896E+03,
     +  1.1164E+03, 1.1436E+03, 1.1712E+03, 1.1994E+03, 1.2279E+03,
     + 1.2570E+03, 1.2866E+03, 1.3166E+03, 1.3472E+03, 1.3783E+03,
     +  1.4100E+03, 1.4422E+03, 1.4749E+03, 1.5083E+03, 1.5422E+03,
     + 1.5767E+03, 1.6118E+03, 1.6475E+03, 1.6839E+03, 1.7208E+03,
     +  1.7585E+03, 1.7968E+03, 1.8358E+03, 1.8754E+03, 1.9158E+03,
     + 1.9568E+03, 1.9987E+03, 2.0412E+03, 2.0845E+03, 2.1285E+03,
     +  2.1733E+03, 2.2189E+03, 2.2653E+03, 2.3126E+03, 2.3606E+03,
     + 2.4095E+03, 2.4592E+03, 2.5098E+03, 2.5613E+03, 2.6137E+03,
     +  2.6670E+03, 2.7213E+03, 2.7764E+03, 2.8326E+03, 2.8897E+03,
     + 2.9477E+03, 3.0068E+03, 3.0669E+03, 3.1281E+03, 3.1902E+03,
     +  3.2535E+03, 3.3178E+03, 3.3832E+03, 3.4497E+03, 3.5174E+03,
     + 3.5862E+03, 3.6562E+03, 3.7273E+03, 3.7997E+03, 3.8732E+03,
     +  3.9480E+03, 4.0240E+03, 4.1014E+03, 4.1800E+03, 4.2599E+03,
     + 4.3411E+03, 4.4237E+03, 4.5076E+03, 4.5929E+03, 4.6796E+03,
     +  4.7678E+03, 4.8574E+03, 4.9484E+03, 5.0410E+03, 5.1350E+03,
     + 5.2305E+03, 5.3276E+03, 5.4263E+03, 5.5266E+03, 5.6284E+03,
     +  5.7319E+03, 5.8371E+03, 5.9439E+03, 6.0524E+03, 6.1627E+03,
     + 6.2747E+03, 6.3884E+03, 6.5040E+03, 6.6214E+03, 6.7406E+03,
     +  6.8616E+03, 6.9846E+03, 7.1095E+03, 7.2363E+03, 7.3650E+03,
     + 7.4958E+03, 7.6286E+03, 7.7634E+03, 7.9003E+03, 8.0393E+03,
     +  8.1804E+03, 8.3237E+03, 8.4691E+03, 8.6168E+03, 8.7666E+03,
     + 8.9188E+03, 9.0732E+03, 9.2299E+03, 9.3890E+03, 9.5505E+03,
     +  9.7144E+03, 9.8807E+03, 1.0049E+04, 1.0221E+04, 1.0395E+04,
     + 1.0571E+04, 1.0750E+04, 1.0931E+04, 1.1116E+04, 1.1303E+04,
     +  1.1492E+04, 1.1685E+04, 1.1880E+04, 1.2078E+04, 1.2279E+04,
     + 1.2482E+04, 1.2689E+04, 1.2899E+04, 1.3111E+04, 1.3327E+04,
     +  1.3546E+04, 1.3768E+04, 1.3993E+04, 1.4221E+04, 1.4452E+04,
     + 1.4687E+04, 1.4925E+04, 1.5166E+04, 1.5410E+04, 1.5658E+04,
     +  1.5910E+04, 1.6165E+04, 1.6423E+04, 1.6685E+04, 1.6951E+04,
     + 1.7220E+04, 1.7493E+04, 1.7770E+04, 1.8050E+04, 1.8335E+04,
     +  1.8623E+04, 1.8915E+04, 1.9211E+04, 1.9511E+04, 1.9815E+04,
     + 2.0123E+04, 2.0436E+04, 2.0752E+04, 2.1073E+04, 2.1398E+04,
     +  2.1727E+04, 2.2061E+04, 2.2399E+04, 2.2741E+04, 2.3088E+04,
     + 2.3440E+04, 2.3796E+04, 2.4157E+04, 2.4523E+04, 2.4893E+04,
     +  2.5269E+04, 2.5649E+04, 2.6034E+04, 2.6424E+04, 2.6819E+04,
     + 2.7219E+04, 2.7624E+04, 2.8034E+04, 2.8450E+04, 2.8871E+04,
     +  2.9298E+04, 2.9729E+04, 3.0167E+04, 3.0610E+04, 3.1058E+04,
     + 3.1512E+04, 3.1972E+04, 3.2437E+04, 3.2909E+04, 3.3386E+04,
     +  3.3870E+04, 3.4359E+04, 3.4854E+04, 3.5356E+04, 3.5864E+04,
     + 3.6378E+04, 3.6898E+04, 3.7425E+04, 3.7958E+04, 3.8498E+04,
     +  3.9044E+04, 3.9598E+04, 4.0157E+04, 4.0724E+04, 4.1298E+04,
     + 4.1878E+04, 4.2466E+04, 4.3060E+04, 4.3662E+04, 4.4271E+04,
     +  4.4888E+04, 4.5512E+04, 4.6143E+04, 4.6782E+04, 4.7428E+04,
     + 4.8082E+04, 4.8744E+04, 4.9414E+04, 5.0091E+04, 5.0777E+04,
     +  5.1471E+04, 5.2173E+04, 5.2883E+04, 5.3601E+04, 5.4328E+04,
     + 5.5064E+04, 5.5808E+04, 5.6560E+04, 5.7322E+04, 5.8092E+04,
     +  5.8871E+04, 5.9659E+04, 6.0457E+04, 6.1263E+04, 6.2079E+04,
     + 6.2904E+04, 6.3738E+04, 6.4582E+04, 6.5436E+04, 6.6299E+04,
     +  6.7173E+04, 6.8056E+04, 6.8949E+04, 6.9852E+04, 7.0766E+04,
     + 7.1690E+04, 7.2624E+04, 7.3568E+04, 7.4524E+04, 7.5490E+04,
     +  7.6467E+04, 7.7454E+04, 7.8453E+04, 7.9463E+04, 8.0484E+04,
     + 8.1516E+04, 8.2560E+04, 8.3615E+04, 8.4682E+04, 8.5761E+04,
     +  8.6851E+04, 8.7954E+04, 8.9068E+04, 9.0195E+04, 9.1334E+04,
     + 9.2486E+04, 9.3650E+04, 9.4827E+04, 9.6016E+04, 9.7218E+04,
     +  9.8434E+04, 9.9662E+04, 1.0090E+05, 1.0216E+05, 1.0343E+05,
     + 1.0471E+05, 1.0601E+05, 1.0732E+05, 1.0864E+05, 1.0998E+05,
     +  1.1133E+05, 1.1270E+05, 1.1408E+05, 1.1547E+05, 1.1688E+05,
     + 1.1831E+05, 1.1975E+05, 1.2120E+05, 1.2267E+05, 1.2416E+05,
     +  1.2566E+05, 1.2718E+05, 1.2871E+05, 1.3026E+05, 1.3183E+05,
     + 1.3341E+05, 1.3501E+05, 1.3662E+05, 1.3825E+05, 1.3990E+05,
     +  1.4157E+05, 1.4325E+05, 1.4495E+05, 1.4666E+05, 1.4840E+05,
     + 1.5015E+05, 1.5192E+05, 1.5371E+05, 1.5552E+05, 1.5734E+05,
     +  1.5918E+05, 1.6105E+05, 1.6293E+05, 1.6483E+05, 1.6675E+05,
     + 1.6868E+05, 1.7064E+05, 1.7262E+05, 1.7462E+05, 1.7664E+05,
     +  1.7867E+05, 1.8073E+05, 1.8281E+05, 1.8491E+05, 1.8703E+05,
     + 1.8917E+05, 1.9133E+05, 1.9352E+05, 1.9572E+05, 1.9795E+05,
     +  2.0020E+05, 2.0247E+05, 2.0477E+05, 2.0708E+05, 2.0942E+05,
     + 2.1178E+05, 2.1417E+05, 2.1658E+05, 2.1901E+05, 2.2146E+05,
     +  2.2394E+05, 2.2645E+05, 2.2897E+05, 2.3153E+05, 2.3410E+05,
     + 2.3671E+05, 2.3933E+05, 2.4198E+05, 2.4466E+05, 2.4737E+05,
     +  2.5010E+05, 2.5285E+05, 2.5563E+05, 2.5844E+05, 2.6128E+05,
     + 2.6414E+05, 2.6703E+05, 2.6994E+05, 2.7289E+05, 2.7586E+05,
     +  2.7886E+05, 2.8189E+05, 2.8495E+05, 2.8803E+05, 2.9115E+05,
     + 2.9429E+05, 2.9747E+05, 3.0067E+05, 3.0390E+05, 3.0717E+05,
     +  3.1046E+05, 3.1378E+05, 3.1714E+05, 3.2053E+05, 3.2395E+05,
     + 3.2740E+05, 3.3088E+05, 3.3439E+05, 3.3794E+05, 3.4152E+05,
     +  3.4513E+05, 3.4877E+05, 3.5245E+05, 3.5616E+05, 3.5991E+05,
     + 3.6369E+05, 3.6751E+05, 3.7136E+05, 3.7524E+05, 3.7916E+05,
     +  3.8312E+05, 3.8711E+05, 3.9114E+05, 3.9520E+05, 3.9930E+05,
     + 4.0344E+05, 4.0762E+05, 4.1183E+05, 4.1608E+05, 4.2037E+05,
     +  4.2470E+05, 4.2906E+05, 4.3347E+05, 4.3791E+05, 4.4240E+05,
     + 4.4692E+05, 4.5149E+05, 4.5609E+05, 4.6074E+05, 4.6543E+05,
     +  4.7015E+05, 4.7493E+05, 4.7974E+05, 4.8459E+05, 4.8949E+05,
     + 4.9443E+05, 4.9942E+05, 5.0444E+05, 5.0952E+05, 5.1463E+05,
     +  5.1979E+05, 5.2500E+05, 5.3025E+05, 5.3555E+05, 5.4089E+05,
     + 5.4628E+05, 5.5172E+05, 5.5720E+05, 5.6273E+05, 5.6831E+05,
     +  5.7394E+05, 5.7961E+05, 5.8534E+05, 5.9111E+05, 5.9694E+05,
     + 6.0281E+05, 6.0873E+05, 6.1471E+05, 6.2073E+05, 6.2681E+05,
     +  6.3294E+05, 6.3912E+05, 6.4535E+05, 6.5164E+05, 6.5798E+05,
     + 6.6437E+05, 6.7082E+05, 6.7732E+05, 6.8388E+05, 6.9049E+05,
     +  6.9716E+05, 7.0388E+05, 7.1067E+05, 7.1750E+05, 7.2440E+05,
     + 7.3135E+05, 7.3836E+05, 7.4543E+05, 7.5256E+05, 7.5975E+05,
     +  7.6700E+05, 7.7430E+05, 7.8167E+05, 7.8910E+05, 7.9659E+05,
     + 8.0415E+05, 8.1176E+05, 8.1944E+05, 8.2718E+05, 8.3499E+05,
     +  8.4286E+05, 8.5080E+05, 8.5880E+05, 8.6686E+05, 8.7499E+05,
     + 8.8319E+05, 8.9146E+05, 8.9979E+05, 9.0819E+05, 9.1666E+05,
     +  9.2520E+05, 9.3380E+05, 9.4248E+05, 9.5123E+05, 9.6005E+05,
     + 9.6894E+05, 9.7790E+05, 9.8693E+05, 9.9604E+05, 1.0052E+06,
     +  1.0145E+06, 1.0238E+06, 1.0332E+06, 1.0427E+06, 1.0522E+06,
     + 1.0619E+06, 1.0716E+06, 1.0814E+06, 1.0912E+06, 1.1012E+06,
     +  1.1112E+06, 1.1213E+06, 1.1315E+06, 1.1417E+06, 1.1521E+06,
     + 1.1625E+06, 1.1730E+06, 1.1836E+06, 1.1943E+06, 1.2050E+06,
     +  1.2159E+06, 1.2268E+06, 1.2378E+06, 1.2489E+06, 1.2601E+06,
     + 1.2713E+06, 1.2827E+06, 1.2942E+06, 1.3057E+06, 1.3173E+06,
     +  1.3290E+06, 1.3408E+06, 1.3527E+06, 1.3647E+06, 1.3768E+06,
     + 1.3890E+06, 1.4013E+06, 1.4136E+06, 1.4261E+06, 1.4387E+06,
     +  1.4513E+06, 1.4641E+06, 1.4769E+06, 1.4899E+06, 1.5029E+06,
     + 1.5161E+06, 1.5293E+06, 1.5427E+06, 1.5561E+06, 1.5697E+06,
     +  1.5833E+06, 1.5971E+06, 1.6110E+06, 1.6249E+06, 1.6390E+06,
     + 1.6532E+06, 1.6675E+06, 1.6819E+06, 1.6964E+06, 1.7110E+06,
     +  1.7257E+06, 1.7406E+06, 1.7555E+06, 1.7706E+06, 1.7858E+06,
     + 1.8011E+06, 1.8165E+06, 1.8320E+06, 1.8476E+06, 1.8634E+06,
     +  1.8792E+06, 1.8952E+06, 1.9113E+06, 1.9275E+06, 1.9439E+06,
     + 1.9603E+06, 1.9769E+06, 1.9936E+06, 2.0105E+06, 2.0274E+06,
     +  2.0445E+06, 2.0617E+06, 2.0790E+06, 2.0965E+06, 2.1141E+06,
     + 2.1318E+06, 2.1497E+06, 2.1676E+06, 2.1858E+06, 2.2040E+06,
     +  2.2224E+06, 2.2409E+06, 2.2595E+06, 2.2783E+06, 2.2972E+06,
     + 2.3163E+06, 2.3355E+06, 2.3548E+06, 2.3743E+06, 2.3939E+06,
     +  2.4136E+06, 2.4335E+06, 2.4536E+06, 2.4737E+06, 2.4941E+06,
     + 2.5146E+06, 2.5352E+06, 2.5559E+06, 2.5769E+06, 2.5979E+06,
     +  2.6191E+06, 2.6405E+06, 2.6620E+06, 2.6837E+06, 2.7055E+06,
     + 2.7275E+06, 2.7497E+06, 2.7720E+06, 2.7944E+06, 2.8170E+06,
     +  2.8398E+06, 2.8628E+06, 2.8858E+06, 2.9091E+06, 2.9325E+06,
     + 2.9561E+06, 2.9799E+06, 3.0038E+06, 3.0279E+06, 3.0522E+06,
     +  3.0766E+06, 3.1012E+06, 3.1260E+06, 3.1509E+06, 3.1760E+06,
     + 3.2013E+06, 3.2268E+06, 3.2524E+06, 3.2783E+06, 3.3043E+06,
     +  3.3305E+06, 3.3568E+06, 3.3834E+06, 3.4101E+06, 3.4370E+06,
     + 3.4642E+06, 3.4914E+06, 3.5189E+06, 3.5466E+06, 3.5745E+06,
     +  3.6025E+06, 3.6308E+06, 3.6592E+06, 3.6878E+06, 3.7167E+06,
     + 3.7457E+06, 3.7749E+06, 3.8043E+06, 3.8340E+06, 3.8638E+06,
     +  3.8938E+06, 3.9240E+06, 3.9545E+06, 3.9851E+06, 4.0160E+06,
     + 4.0470E+06, 4.0783E+06, 4.1098E+06, 4.1415E+06, 4.1734E+06,
     +  4.2055E+06, 4.2379E+06, 4.2704E+06, 4.3032E+06, 4.3362E+06,
     + 4.3694E+06, 4.4029E+06, 4.4365E+06, 4.4704E+06, 4.5045E+06,
     +  4.5389E+06, 4.5735E+06, 4.6083E+06, 4.6433E+06, 4.6786E+06,
     + 4.7141E+06, 4.7498E+06, 4.7858E+06, 4.8220E+06, 4.8585E+06,
     +  4.8952E+06, 4.9321E+06, 4.9693E+06, 5.0067E+06, 5.0444E+06,
     + 5.0823E+06, 5.1205E+06, 5.1589E+06, 5.1976E+06, 5.2365E+06,
     +  5.2757E+06, 5.3151E+06, 5.3548E+06, 5.3948E+06, 5.4350E+06,
     + 5.4755E+06, 5.5163E+06, 5.5573E+06, 5.5985E+06, 5.6401E+06,
     +  5.6819E+06, 5.7240E+06, 5.7664E+06, 5.8090E+06, 5.8519E+06,
     + 5.8951E+06, 5.9386E+06, 5.9823E+06, 6.0263E+06, 6.0706E+06,
     +  6.1152E+06, 6.1601E+06, 6.2053E+06, 6.2508E+06, 6.2965E+06,
     + 6.3426E+06, 6.3889E+06, 6.4356E+06, 6.4825E+06, 6.5297E+06,
     +  6.5773E+06, 6.6251E+06, 6.6733E+06, 6.7217E+06, 6.7705E+06,
     + 6.8195E+06, 6.8689E+06, 6.9186E+06, 6.9686E+06, 7.0189E+06,
     +  7.0696E+06, 7.1206E+06, 7.1718E+06, 7.2234E+06, 7.2754E+06,
     + 7.3276E+06, 7.3802E+06, 7.4331E+06, 7.4864E+06, 7.5400E+06,
     +  7.5939E+06, 7.6482E+06, 7.7028E+06, 7.7577E+06, 7.8130E+06,
     + 7.8686E+06, 7.9246E+06, 7.9809E+06, 8.0376E+06, 8.0946E+06,
     +  8.1520E+06, 8.2097E+06, 8.2678E+06, 8.3263E+06, 8.3851E+06,
     + 8.4442E+06, 8.5038E+06, 8.5637E+06, 8.6240E+06, 8.6846E+06,
     +  8.7457E+06, 8.8071E+06, 8.8688E+06, 8.9310E+06, 8.9935E+06,
     + 9.0565E+06, 9.1198E+06, 9.1835E+06, 9.2475E+06, 9.3120E+06,
     +  9.3769E+06, 9.4422E+06, 9.5078E+06, 9.5739E+06, 9.6403E+06,
     + 9.7072E+06, 9.7745E+06, 9.8422E+06, 9.9103E+06, 9.9788E+06,
     +  1.0048E+07, 1.0117E+07, 1.0187E+07, 1.0257E+07, 1.0328E+07,
     + 1.0399E+07, 1.0470E+07, 1.0542E+07, 1.0614E+07, 1.0687E+07,
     +  1.0760E+07, 1.0834E+07, 1.0908E+07, 1.0982E+07, 1.1057E+07,
     + 1.1133E+07, 1.1209E+07, 1.1285E+07, 1.1362E+07, 1.1439E+07,
     +  1.1517E+07, 1.1595E+07, 1.1673E+07, 1.1752E+07, 1.1832E+07,
     + 1.1912E+07, 1.1992E+07, 1.2073E+07, 1.2155E+07, 1.2237E+07,
     +  1.2319E+07, 1.2402E+07, 1.2485E+07, 1.2569E+07, 1.2654E+07,
     + 1.2738E+07, 1.2824E+07, 1.2910E+07, 1.2996E+07, 1.3083E+07,
     +  1.3170E+07, 1.3258E+07, 1.3347E+07, 1.3435E+07, 1.3525E+07,
     + 1.3615E+07, 1.3705E+07, 1.3796E+07, 1.3888E+07, 1.3980E+07,
     +  1.4073E+07, 1.4166E+07, 1.4259E+07, 1.4354E+07, 1.4448E+07,
     + 1.4544E+07, 1.4639E+07, 1.4736E+07, 1.4833E+07, 1.4930E+07,
     +  1.5028E+07, 1.5127E+07, 1.5226E+07, 1.5326E+07, 1.5427E+07,
     + 1.5527E+07, 1.5629E+07, 1.5731E+07, 1.5834E+07, 1.5937E+07,
     +  1.6041E+07, 1.6145E+07, 1.6250E+07, 1.6356E+07, 1.6462E+07,
     + 1.6569E+07, 1.6677E+07, 1.6785E+07, 1.6893E+07, 1.7003E+07,
     +  1.7113E+07, 1.7223E+07, 1.7334E+07, 1.7446E+07, 1.7559E+07,
     + 1.7672E+07, 1.7785E+07, 1.7900E+07, 1.8015E+07, 1.8130E+07,
     +  1.8247E+07, 1.8363E+07, 1.8481E+07, 1.8599E+07, 1.8718E+07,
     + 1.8838E+07, 1.8958E+07, 1.9079E+07, 1.9201E+07, 1.9323E+07,
     +  1.9446E+07, 1.9569E+07, 1.9694E+07, 1.9819E+07, 1.9944E+07,
     + 2.0071E+07, 2.0198E+07, 2.0326E+07, 2.0454E+07, 2.0584E+07,
     +  2.0714E+07, 2.0844E+07, 2.0976E+07, 2.1108E+07, 2.1241E+07,
     + 2.1374E+07, 2.1508E+07, 2.1644E+07, 2.1779E+07, 2.1916E+07,
     +  2.2053E+07, 2.2191E+07, 2.2330E+07, 2.2470E+07, 2.2610E+07,
     + 2.2751E+07, 2.2893E+07, 2.3036E+07, 2.3179E+07, 2.3323E+07,
     +  2.3468E+07, 2.3614E+07, 2.3760E+07, 2.3908E+07, 2.4056E+07,
     + 2.4205E+07, 2.4355E+07, 2.4505E+07, 2.4657E+07, 2.4809E+07,
     +  2.4962E+07, 2.5116E+07, 2.5271E+07, 2.5426E+07, 2.5583E+07,
     + 2.5740E+07, 2.5898E+07, 2.6057E+07, 2.6217E+07, 2.6377E+07,
     +  2.6539E+07, 2.6701E+07, 2.6864E+07, 2.7028E+07, 2.7193E+07,
     + 2.7359E+07, 2.7526E+07, 2.7694E+07, 2.7862E+07, 2.8032E+07,
     +  2.8202E+07, 2.8373E+07, 2.8545E+07, 2.8718E+07, 2.8892E+07,
     + 2.9067E+07, 2.9243E+07, 2.9420E+07, 2.9598E+07, 2.9776E+07,
     +  2.9956E+07, 3.0136E+07, 3.0318E+07, 3.0500E+07, 3.0684E+07,
     + 3.0868E+07, 3.1053E+07, 3.1240E+07, 3.1427E+07, 3.1615E+07,
     +  3.1805E+07, 3.1995E+07, 3.2186E+07, 3.2378E+07, 3.2572E+07,
     + 3.2766E+07, 3.2961E+07, 3.3157E+07, 3.3355E+07, 3.3553E+07,
     +  3.3753E+07, 3.3953E+07, 3.4154E+07, 3.4357E+07, 3.4561E+07,
     + 3.4765E+07, 3.4971E+07, 3.5178E+07, 3.5385E+07, 3.5594E+07,
     +  3.5804E+07, 3.6015E+07, 3.6228E+07, 3.6441E+07, 3.6655E+07,
     + 3.6871E+07, 3.7087E+07, 3.7305E+07, 3.7524E+07, 3.7744E+07,
     +  3.7965E+07, 3.8187E+07, 3.8410E+07, 3.8635E+07, 3.8860E+07,
     + 3.9087E+07, 3.9315E+07, 3.9544E+07, 3.9774E+07, 4.0006E+07,
     +  4.0238E+07, 4.0472E+07, 4.0707E+07, 4.0943E+07, 4.1181E+07,
     + 4.1419E+07, 4.1659E+07, 4.1900E+07, 4.2142E+07, 4.2386E+07,
     +  4.2630E+07, 4.2876E+07, 4.3124E+07, 4.3372E+07, 4.3622E+07,
     + 4.3873E+07, 4.4125E+07, 4.4378E+07, 4.4633E+07, 4.4889E+07,
     +  4.5146E+07, 4.5405E+07, 4.5665E+07, 4.5926E+07, 4.6189E+07,
     + 4.6452E+07, 4.6718E+07, 4.6984E+07, 4.7252E+07, 4.7521E+07,
     +  4.7791E+07, 4.8063E+07, 4.8336E+07, 4.8611E+07, 4.8887E+07,
     + 4.9164E+07, 4.9443E+07, 4.9723E+07, 5.0004E+07, 5.0287E+07,
     +  5.0571E+07, 5.0857E+07, 5.1144E+07, 5.1433E+07, 5.1722E+07,
     + 5.2014E+07, 5.2307E+07, 5.2601E+07, 5.2896E+07, 5.3193E+07,
     +  5.3492E+07, 5.3792E+07, 5.4094E+07, 5.4397E+07, 5.4701E+07,
     + 5.5007E+07, 5.5315E+07, 5.5623E+07, 5.5934E+07, 5.6246E+07,
     +  5.6560E+07, 5.6875E+07, 5.7191E+07, 5.7509E+07, 5.7829E+07,
     + 5.8150E+07, 5.8473E+07, 5.8797E+07, 5.9123E+07, 5.9451E+07,
     +  5.9780E+07, 6.0111E+07, 6.0443E+07, 6.0777E+07, 6.1113E+07,
     + 6.1450E+07, 6.1789E+07, 6.2129E+07, 6.2471E+07, 6.2815E+07,
     +  6.3160E+07, 6.3507E+07, 6.3856E+07, 6.4206E+07, 6.4558E+07,
     + 6.4912E+07, 6.5268E+07, 6.5625E+07, 6.5984E+07, 6.6344E+07,
     +  6.6707E+07, 6.7071E+07, 6.7437E+07, 6.7804E+07, 6.8173E+07,
     + 6.8544E+07, 6.8917E+07, 6.9292E+07, 6.9668E+07, 7.0047E+07,
     +  7.0426E+07, 7.0808E+07, 7.1192E+07, 7.1577E+07, 7.1965E+07,
     + 7.2354E+07, 7.2745E+07, 7.3137E+07, 7.3532E+07, 7.3928E+07,
     +  7.4327E+07, 7.4727E+07, 7.5129E+07, 7.5533E+07, 7.5939E+07,
     + 7.6347E+07, 7.6757E+07, 7.7169E+07, 7.7582E+07, 7.7998E+07,
     +  7.8415E+07, 7.8835E+07, 7.9256E+07, 7.9680E+07, 8.0105E+07,
     + 8.0532E+07, 8.0962E+07, 8.1393E+07, 8.1827E+07, 8.2262E+07,
     +  8.2699E+07, 8.3139E+07, 8.3580E+07, 8.4024E+07, 8.4470E+07,
     + 8.4917E+07, 8.5367E+07, 8.5819E+07, 8.6273E+07, 8.6729E+07,
     +  8.7187E+07, 8.7648E+07, 8.8110E+07, 8.8575E+07, 8.9041E+07,
     + 8.9510E+07, 8.9981E+07, 9.0455E+07, 9.0930E+07, 9.1408E+07,
     +  9.1887E+07, 9.2369E+07, 9.2853E+07, 9.3340E+07, 9.3829E+07,
     + 9.4319E+07, 9.4813E+07, 9.5308E+07, 9.5806E+07, 9.6306E+07,
     +  9.6808E+07, 9.7312E+07, 9.7819E+07, 9.8328E+07, 9.8840E+07,
     + 9.9354E+07, 9.9870E+07, 1.0039E+08, 1.0091E+08, 1.0143E+08,
     +  1.0196E+08, 1.0249E+08, 1.0302E+08, 1.0355E+08, 1.0408E+08,
     + 1.0462E+08, 1.0516E+08, 1.0570E+08, 1.0625E+08, 1.0680E+08,
     +  1.0735E+08, 1.0790E+08, 1.0845E+08, 1.0901E+08, 1.0957E+08,
     + 1.1013E+08, 1.1070E+08, 1.1127E+08, 1.1184E+08, 1.1241E+08,
     +  1.1298E+08, 1.1356E+08, 1.1414E+08, 1.1472E+08, 1.1531E+08,
     + 1.1590E+08, 1.1649E+08, 1.1708E+08, 1.1768E+08, 1.1828E+08,
     +  1.1888E+08, 1.1948E+08, 1.2009E+08, 1.2070E+08, 1.2131E+08,
     + 1.2192E+08, 1.2254E+08, 1.2316E+08, 1.2378E+08, 1.2441E+08,
     +  1.2504E+08, 1.2567E+08, 1.2630E+08, 1.2694E+08, 1.2758E+08,
     + 1.2822E+08, 1.2887E+08, 1.2952E+08, 1.3017E+08, 1.3082E+08,
     +  1.3148E+08, 1.3214E+08, 1.3280E+08, 1.3347E+08, 1.3413E+08,
     + 1.3481E+08, 1.3548E+08, 1.3616E+08, 1.3684E+08, 1.3752E+08,
     +  1.3821E+08, 1.3890E+08, 1.3959E+08, 1.4028E+08, 1.4098E+08,
     + 1.4168E+08, 1.4239E+08, 1.4310E+08, 1.4381E+08, 1.4452E+08,
     +  1.4524E+08, 1.4596E+08, 1.4668E+08, 1.4741E+08, 1.4813E+08,
     + 1.4887E+08, 1.4960E+08, 1.5034E+08, 1.5108E+08, 1.5183E+08,
     +  1.5258E+08, 1.5333E+08, 1.5408E+08, 1.5484E+08, 1.5560E+08,
     + 1.5637E+08, 1.5713E+08, 1.5791E+08, 1.5868E+08, 1.5946E+08,
     +  1.6024E+08, 1.6102E+08, 1.6181E+08, 1.6260E+08, 1.6340E+08,
     + 1.6420E+08, 1.6500E+08, 1.6580E+08, 1.6661E+08, 1.6742E+08,
     +  1.6824E+08, 1.6906E+08, 1.6988E+08, 1.7070E+08, 1.7153E+08,
     + 1.7237E+08, 1.7320E+08, 1.7404E+08, 1.7489E+08, 1.7573E+08,
     +  1.7658E+08, 1.7744E+08, 1.7830E+08, 1.7916E+08, 1.8002E+08,
     + 1.8089E+08, 1.8176E+08, 1.8264E+08, 1.8352E+08, 1.8440E+08,
     +  1.8529E+08, 1.8618E+08, 1.8708E+08, 1.8797E+08, 1.8888E+08,
     + 1.8978E+08, 1.9069E+08, 1.9161E+08, 1.9252E+08, 1.9345E+08,
     +  1.9437E+08, 1.9530E+08, 1.9623E+08, 1.9717E+08, 1.9811E+08,
     + 1.9906E+08, 2.0000E+08, 2.0096E+08, 2.0191E+08, 2.0287E+08,
     +  2.0384E+08, 2.0481E+08, 2.0578E+08, 2.0676E+08, 2.0774E+08,
     + 2.0872E+08, 2.0971E+08, 2.1070E+08, 2.1170E+08, 2.1270E+08,
     +  2.1371E+08, 2.1472E+08, 2.1573E+08, 2.1675E+08, 2.1777E+08,
     + 2.1880E+08, 2.1983E+08, 2.2086E+08, 2.2190E+08, 2.2295E+08,
     +  2.2399E+08, 2.2505E+08, 2.2610E+08, 2.2716E+08, 2.2823E+08,
     + 2.2930E+08, 2.3037E+08, 2.3145E+08, 2.3253E+08, 2.3362E+08,
     +  2.3471E+08, 2.3581E+08, 2.3691E+08, 2.3801E+08, 2.3912E+08,
     + 2.4024E+08, 2.4136E+08, 2.4248E+08, 2.4361E+08, 2.4474E+08,
     +  2.4588E+08, 2.4702E+08, 2.4816E+08, 2.4932E+08, 2.5047E+08,
     + 2.5163E+08, 2.5280E+08, 2.5397E+08, 2.5514E+08, 2.5632E+08,
     +  2.5750E+08, 2.5869E+08, 2.5989E+08, 2.6108E+08, 2.6229E+08,
     + 2.6349E+08, 2.6471E+08, 2.6593E+08, 2.6715E+08, 2.6838E+08,
     +  2.6961E+08, 2.7085E+08, 2.7209E+08, 2.7334E+08, 2.7459E+08,
     + 2.7585E+08, 2.7711E+08, 2.7838E+08, 2.7965E+08, 2.8093E+08,
     +  2.8221E+08, 2.8350E+08, 2.8479E+08, 2.8609E+08, 2.8739E+08,
     + 2.8870E+08, 2.9002E+08, 2.9134E+08, 2.9266E+08, 2.9399E+08,
     +  2.9532E+08, 2.9667E+08, 2.9801E+08, 2.9936E+08, 3.0072E+08,
     + 3.0208E+08, 3.0345E+08, 3.0482E+08, 3.0620E+08, 3.0758E+08,
     +  3.0897E+08, 3.1036E+08, 3.1176E+08, 3.1317E+08, 3.1458E+08,
     + 3.1600E+08, 3.1742E+08, 3.1885E+08, 3.2028E+08, 3.2172E+08,
     +  3.2317E+08, 3.2462E+08, 3.2607E+08, 3.2753E+08, 3.2900E+08,
     + 3.3047E+08, 3.3195E+08, 3.3344E+08, 3.3493E+08, 3.3643E+08,
     +  3.3793E+08, 3.3944E+08, 3.4095E+08, 3.4247E+08, 3.4400E+08,
     + 3.4553E+08, 3.4707E+08, 3.4861E+08, 3.5016E+08, 3.5172E+08,
     +  3.5328E+08, 3.5485E+08, 3.5642E+08, 3.5800E+08, 3.5959E+08,
     + 3.6118E+08, 3.6278E+08, 3.6438E+08, 3.6600E+08, 3.6761E+08,
     +  3.6924E+08, 3.7087E+08, 3.7250E+08, 3.7415E+08, 3.7580E+08,
     + 3.7745E+08, 3.7911E+08, 3.8078E+08, 3.8245E+08, 3.8414E+08,
     +  3.8582E+08, 3.8752E+08, 3.8922E+08, 3.9093E+08, 3.9264E+08,
     + 3.9436E+08, 3.9609E+08, 3.9782E+08, 3.9956E+08, 4.0131E+08,
     +  4.0306E+08, 4.0482E+08, 4.0659E+08, 4.0836E+08, 4.1014E+08,
     + 4.1193E+08, 4.1372E+08, 4.1552E+08, 4.1733E+08, 4.1914E+08,
     +  4.2097E+08, 4.2279E+08, 4.2463E+08, 4.2647E+08, 4.2832E+08,
     + 4.3018E+08, 4.3204E+08, 4.3391E+08, 4.3579E+08, 4.3767E+08,
     +  4.3957E+08, 4.4146E+08, 4.4337E+08, 4.4528E+08, 4.4720E+08,
     + 4.4913E+08, 4.5107E+08, 4.5301E+08, 4.5496E+08, 4.5692E+08,
     +  4.5888E+08, 4.6085E+08, 4.6283E+08, 4.6482E+08, 4.6681E+08,
     + 4.6882E+08, 4.7082E+08, 4.7284E+08, 4.7487E+08, 4.7690E+08,
     +  4.7894E+08, 4.8098E+08, 4.8304E+08, 4.8510E+08, 4.8717E+08,
     + 4.8925E+08, 4.9134E+08, 4.9343E+08, 4.9553E+08, 4.9764E+08,
     +  4.9976E+08, 5.0188E+08, 5.0401E+08, 5.0615E+08, 5.0830E+08,
     + 5.1046E+08, 5.1262E+08, 5.1480E+08, 5.1698E+08, 5.1917E+08,
     +  5.2136E+08, 5.2357E+08, 5.2578E+08, 5.2800E+08, 5.3023E+08,
     + 5.3247E+08, 5.3472E+08, 5.3697E+08, 5.3924E+08, 5.4151E+08,
     +  5.4379E+08, 5.4607E+08, 5.4837E+08, 5.5068E+08, 5.5299E+08,
     + 5.5531E+08, 5.5764E+08, 5.5998E+08, 5.6233E+08, 5.6468E+08,
     +  5.6705E+08, 5.6942E+08, 5.7181E+08, 5.7420E+08, 5.7660E+08,
     + 5.7900E+08, 5.8142E+08, 5.8385E+08, 5.8628E+08, 5.8873E+08,
     +  5.9118E+08, 5.9364E+08, 5.9611E+08, 5.9859E+08, 6.0108E+08,
     + 6.0358E+08, 6.0609E+08, 6.0860E+08, 6.1113E+08, 6.1366E+08,
     +  6.1620E+08, 6.1876E+08, 6.2132E+08, 6.2389E+08, 6.2647E+08,
     + 6.2906E+08, 6.3166E+08, 6.3427E+08, 6.3689E+08, 6.3952E+08,
     +  6.4215E+08, 6.4480E+08, 6.4746E+08, 6.5012E+08, 6.5280E+08,
     + 6.5548E+08, 6.5818E+08, 6.6088E+08, 6.6360E+08, 6.6632E+08,
     +  6.6905E+08, 6.7180E+08, 6.7455E+08, 6.7731E+08, 6.8009E+08,
     + 6.8287E+08, 6.8566E+08, 6.8847E+08, 6.9128E+08, 6.9410E+08,
     +  6.9694E+08, 6.9978E+08, 7.0263E+08, 7.0550E+08, 7.0837E+08,
     + 7.1126E+08, 7.1415E+08, 7.1705E+08, 7.1997E+08, 7.2290E+08,
     +  7.2583E+08, 7.2878E+08, 7.3173E+08, 7.3470E+08, 7.3768E+08,
     + 7.4067E+08, 7.4367E+08, 7.4668E+08, 7.4970E+08, 7.5273E+08,
     +  7.5577E+08, 7.5882E+08, 7.6189E+08, 7.6496E+08, 7.6805E+08,
     + 7.7114E+08, 7.7425E+08, 7.7737E+08, 7.8050E+08, 7.8364E+08,
     +  7.8679E+08, 7.8995E+08, 7.9312E+08, 7.9631E+08, 7.9950E+08,
     + 8.0271E+08, 8.0593E+08, 8.0916E+08, 8.1240E+08, 8.1565E+08,
     +  8.1891E+08, 8.2219E+08, 8.2547E+08, 8.2877E+08, 8.3208E+08,
     + 8.3540E+08, 8.3874E+08, 8.4208E+08, 8.4544E+08, 8.4880E+08,
     +  8.5218E+08, 8.5557E+08, 8.5898E+08, 8.6239E+08, 8.6582E+08,
     + 8.6926E+08, 8.7271E+08, 8.7617E+08, 8.7964E+08, 8.8313E+08,
     +  8.8663E+08, 8.9014E+08, 8.9366E+08, 8.9720E+08, 9.0075E+08,
     + 9.0431E+08, 9.0788E+08, 9.1146E+08, 9.1506E+08, 9.1867E+08,
     +  9.2229E+08, 9.2593E+08, 9.2957E+08, 9.3323E+08, 9.3690E+08,
     + 9.4059E+08, 9.4429E+08, 9.4800E+08, 9.5172E+08, 9.5545E+08,
     +  9.5920E+08, 9.6296E+08, 9.6674E+08, 9.7053E+08, 9.7433E+08,
     + 9.7814E+08, 9.8196E+08, 9.8580E+08, 9.8966E+08, 9.9352E+08,
     +  9.9740E+08, 1.0013E+09, 1.0052E+09, 1.0091E+09, 1.0131E+09,
     + 1.0170E+09, 1.0210E+09, 1.0249E+09, 1.0289E+09, 1.0329E+09,
     +  1.0369E+09, 1.0410E+09, 1.0450E+09, 1.0490E+09, 1.0531E+09,
     + 1.0572E+09, 1.0613E+09, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_COCl2 (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 2),QofT( 2,NT),Q(NT)
      data (Tmx(I),I=1,2)/5000.,5000./
      data xgj/ 1.,1./
c...    COCl2
c...        --      2655
      data (QofT( 1,J),J=1,2408)/ 1.7215E+02, 4.7742E+02, 8.7140E+02,
     +  1.3373E+03, 1.8653E+03, 2.4489E+03, 3.0832E+03, 3.7644E+03,
     + 4.4895E+03, 5.2561E+03, 6.0618E+03, 6.9051E+03, 7.7842E+03,
     +  8.6978E+03, 9.6445E+03, 1.0623E+04, 1.1633E+04, 1.2673E+04,
     + 1.3743E+04, 1.4841E+04, 1.7119E+04, 1.9504E+04, 2.1990E+04,
     +  2.4574E+04, 2.7252E+04, 3.0021E+04, 3.2878E+04, 3.5821E+04,
     + 3.8847E+04, 4.1954E+04, 4.5140E+04, 4.8404E+04, 5.1744E+04,
     +  5.5158E+04, 5.8647E+04, 6.2208E+04, 6.5840E+04, 6.9544E+04,
     + 7.3318E+04, 7.7162E+04, 8.1077E+04, 8.5061E+04, 8.9114E+04,
     +  9.3237E+04, 9.7430E+04, 1.0169E+05, 1.0603E+05, 1.1043E+05,
     + 1.1491E+05, 1.1946E+05, 1.2408E+05, 1.2878E+05, 1.3355E+05,
     +  1.3839E+05, 1.4332E+05, 1.4832E+05, 1.5340E+05, 1.5856E+05,
     + 1.6380E+05, 1.6912E+05, 1.7453E+05, 1.8002E+05, 1.8560E+05,
     +  1.9127E+05, 1.9703E+05, 2.0288E+05, 2.0883E+05, 2.1487E+05,
     + 2.2101E+05, 2.2724E+05, 2.3358E+05, 2.4002E+05, 2.4656E+05,
     +  2.5321E+05, 2.5997E+05, 2.6684E+05, 2.7382E+05, 2.8092E+05,
     + 2.8813E+05, 2.9547E+05, 3.0292E+05, 3.1050E+05, 3.1820E+05,
     +  3.2604E+05, 3.3400E+05, 3.4210E+05, 3.5033E+05, 3.5870E+05,
     + 3.6721E+05, 3.7587E+05, 3.8467E+05, 3.9362E+05, 4.0272E+05,
     +  4.1198E+05, 4.2139E+05, 4.3096E+05, 4.4070E+05, 4.5059E+05,
     + 4.6066E+05, 4.7090E+05, 4.8131E+05, 4.9190E+05, 5.0267E+05,
     +  5.1363E+05, 5.2477E+05, 5.3610E+05, 5.4762E+05, 5.5934E+05,
     + 5.7125E+05, 5.8337E+05, 5.9570E+05, 6.0824E+05, 6.2098E+05,
     +  6.3395E+05, 6.4714E+05, 6.6054E+05, 6.7418E+05, 6.8805E+05,
     + 7.0215E+05, 7.1649E+05, 7.3107E+05, 7.4590E+05, 7.6097E+05,
     +  7.7631E+05, 7.9190E+05, 8.0775E+05, 8.2386E+05, 8.4025E+05,
     + 8.5691E+05, 8.7385E+05, 8.9107E+05, 9.0858E+05, 9.2638E+05,
     +  9.4448E+05, 9.6288E+05, 9.8158E+05, 1.0006E+06, 1.0199E+06,
     + 1.0396E+06, 1.0595E+06, 1.0798E+06, 1.1005E+06, 1.1214E+06,
     +  1.1427E+06, 1.1644E+06, 1.1864E+06, 1.2088E+06, 1.2315E+06,
     + 1.2546E+06, 1.2781E+06, 1.3019E+06, 1.3262E+06, 1.3508E+06,
     +  1.3758E+06, 1.4013E+06, 1.4271E+06, 1.4534E+06, 1.4800E+06,
     + 1.5071E+06, 1.5347E+06, 1.5626E+06, 1.5910E+06, 1.6199E+06,
     +  1.6492E+06, 1.6790E+06, 1.7092E+06, 1.7399E+06, 1.7711E+06,
     + 1.8028E+06, 1.8350E+06, 1.8676E+06, 1.9008E+06, 1.9345E+06,
     +  1.9687E+06, 2.0035E+06, 2.0388E+06, 2.0746E+06, 2.1110E+06,
     + 2.1479E+06, 2.1854E+06, 2.2235E+06, 2.2621E+06, 2.3014E+06,
     +  2.3412E+06, 2.3817E+06, 2.4227E+06, 2.4644E+06, 2.5067E+06,
     + 2.5496E+06, 2.5932E+06, 2.6374E+06, 2.6823E+06, 2.7278E+06,
     +  2.7741E+06, 2.8210E+06, 2.8686E+06, 2.9169E+06, 2.9659E+06,
     + 3.0157E+06, 3.0662E+06, 3.1174E+06, 3.1694E+06, 3.2221E+06,
     +  3.2756E+06, 3.3299E+06, 3.3849E+06, 3.4408E+06, 3.4975E+06,
     + 3.5550E+06, 3.6133E+06, 3.6724E+06, 3.7324E+06, 3.7933E+06,
     +  3.8550E+06, 3.9176E+06, 3.9811E+06, 4.0455E+06, 4.1109E+06,
     + 4.1771E+06, 4.2443E+06, 4.3124E+06, 4.3815E+06, 4.4515E+06,
     +  4.5226E+06, 4.5946E+06, 4.6676E+06, 4.7417E+06, 4.8167E+06,
     + 4.8928E+06, 4.9700E+06, 5.0482E+06, 5.1276E+06, 5.2080E+06,
     +  5.2895E+06, 5.3721E+06, 5.4559E+06, 5.5408E+06, 5.6268E+06,
     + 5.7140E+06, 5.8025E+06, 5.8921E+06, 5.9829E+06, 6.0749E+06,
     +  6.1682E+06, 6.2628E+06, 6.3586E+06, 6.4557E+06, 6.5541E+06,
     + 6.6538E+06, 6.7548E+06, 6.8572E+06, 6.9609E+06, 7.0661E+06,
     +  7.1726E+06, 7.2805E+06, 7.3898E+06, 7.5006E+06, 7.6128E+06,
     + 7.7265E+06, 7.8416E+06, 7.9583E+06, 8.0765E+06, 8.1962E+06,
     +  8.3175E+06, 8.4404E+06, 8.5648E+06, 8.6909E+06, 8.8186E+06,
     + 8.9479E+06, 9.0788E+06, 9.2115E+06, 9.3458E+06, 9.4819E+06,
     +  9.6197E+06, 9.7592E+06, 9.9006E+06, 1.0044E+07, 1.0189E+07,
     + 1.0335E+07, 1.0484E+07, 1.0634E+07, 1.0787E+07, 1.0941E+07,
     +  1.1097E+07, 1.1255E+07, 1.1415E+07, 1.1578E+07, 1.1742E+07,
     + 1.1908E+07, 1.2076E+07, 1.2246E+07, 1.2419E+07, 1.2593E+07,
     +  1.2770E+07, 1.2949E+07, 1.3130E+07, 1.3313E+07, 1.3498E+07,
     + 1.3686E+07, 1.3876E+07, 1.4068E+07, 1.4263E+07, 1.4460E+07,
     +  1.4659E+07, 1.4861E+07, 1.5065E+07, 1.5272E+07, 1.5481E+07,
     + 1.5693E+07, 1.5907E+07, 1.6124E+07, 1.6343E+07, 1.6565E+07,
     +  1.6789E+07, 1.7017E+07, 1.7247E+07, 1.7479E+07, 1.7714E+07,
     + 1.7953E+07, 1.8193E+07, 1.8437E+07, 1.8684E+07, 1.8933E+07,
     +  1.9186E+07, 1.9441E+07, 1.9699E+07, 1.9960E+07, 2.0225E+07,
     + 2.0492E+07, 2.0763E+07, 2.1036E+07, 2.1313E+07, 2.1593E+07,
     +  2.1876E+07, 2.2162E+07, 2.2452E+07, 2.2745E+07, 2.3041E+07,
     + 2.3341E+07, 2.3644E+07, 2.3950E+07, 2.4260E+07, 2.4574E+07,
     +  2.4891E+07, 2.5211E+07, 2.5535E+07, 2.5863E+07, 2.6195E+07,
     + 2.6530E+07, 2.6869E+07, 2.7212E+07, 2.7558E+07, 2.7909E+07,
     +  2.8263E+07, 2.8621E+07, 2.8983E+07, 2.9349E+07, 2.9720E+07,
     + 3.0094E+07, 3.0472E+07, 3.0855E+07, 3.1242E+07, 3.1633E+07,
     +  3.2028E+07, 3.2428E+07, 3.2831E+07, 3.3240E+07, 3.3653E+07,
     + 3.4070E+07, 3.4492E+07, 3.4918E+07, 3.5349E+07, 3.5784E+07,
     +  3.6225E+07, 3.6669E+07, 3.7119E+07, 3.7574E+07, 3.8033E+07,
     + 3.8497E+07, 3.8967E+07, 3.9441E+07, 3.9920E+07, 4.0405E+07,
     +  4.0894E+07, 4.1389E+07, 4.1889E+07, 4.2394E+07, 4.2904E+07,
     + 4.3420E+07, 4.3941E+07, 4.4468E+07, 4.5000E+07, 4.5538E+07,
     +  4.6081E+07, 4.6630E+07, 4.7185E+07, 4.7746E+07, 4.8312E+07,
     + 4.8884E+07, 4.9462E+07, 5.0046E+07, 5.0637E+07, 5.1233E+07,
     +  5.1835E+07, 5.2443E+07, 5.3058E+07, 5.3679E+07, 5.4306E+07,
     + 5.4940E+07, 5.5580E+07, 5.6227E+07, 5.6880E+07, 5.7540E+07,
     +  5.8206E+07, 5.8880E+07, 5.9560E+07, 6.0247E+07, 6.0940E+07,
     + 6.1641E+07, 6.2349E+07, 6.3064E+07, 6.3786E+07, 6.4515E+07,
     +  6.5252E+07, 6.5996E+07, 6.6747E+07, 6.7506E+07, 6.8272E+07,
     + 6.9046E+07, 6.9827E+07, 7.0616E+07, 7.1413E+07, 7.2218E+07,
     +  7.3031E+07, 7.3852E+07, 7.4680E+07, 7.5517E+07, 7.6362E+07,
     + 7.7216E+07, 7.8077E+07, 7.8947E+07, 7.9826E+07, 8.0713E+07,
     +  8.1608E+07, 8.2512E+07, 8.3425E+07, 8.4347E+07, 8.5278E+07,
     + 8.6217E+07, 8.7166E+07, 8.8124E+07, 8.9091E+07, 9.0067E+07,
     +  9.1052E+07, 9.2047E+07, 9.3051E+07, 9.4065E+07, 9.5089E+07,
     + 9.6122E+07, 9.7165E+07, 9.8218E+07, 9.9281E+07, 1.0035E+08,
     +  1.0144E+08, 1.0253E+08, 1.0363E+08, 1.0475E+08, 1.0587E+08,
     + 1.0701E+08, 1.0815E+08, 1.0931E+08, 1.1047E+08, 1.1165E+08,
     +  1.1284E+08, 1.1404E+08, 1.1525E+08, 1.1647E+08, 1.1771E+08,
     + 1.1895E+08, 1.2021E+08, 1.2147E+08, 1.2275E+08, 1.2405E+08,
     +  1.2535E+08, 1.2666E+08, 1.2799E+08, 1.2933E+08, 1.3068E+08,
     + 1.3204E+08, 1.3342E+08, 1.3481E+08, 1.3621E+08, 1.3762E+08,
     +  1.3905E+08, 1.4049E+08, 1.4194E+08, 1.4341E+08, 1.4488E+08,
     + 1.4638E+08, 1.4788E+08, 1.4940E+08, 1.5093E+08, 1.5248E+08,
     +  1.5404E+08, 1.5561E+08, 1.5720E+08, 1.5880E+08, 1.6042E+08,
     + 1.6205E+08, 1.6369E+08, 1.6535E+08, 1.6703E+08, 1.6872E+08,
     +  1.7042E+08, 1.7214E+08, 1.7387E+08, 1.7562E+08, 1.7739E+08,
     + 1.7916E+08, 1.8096E+08, 1.8277E+08, 1.8460E+08, 1.8644E+08,
     +  1.8830E+08, 1.9017E+08, 1.9207E+08, 1.9397E+08, 1.9590E+08,
     + 1.9784E+08, 1.9980E+08, 2.0177E+08, 2.0376E+08, 2.0577E+08,
     +  2.0780E+08, 2.0984E+08, 2.1190E+08, 2.1398E+08, 2.1607E+08,
     + 2.1819E+08, 2.2032E+08, 2.2247E+08, 2.2464E+08, 2.2682E+08,
     +  2.2903E+08, 2.3125E+08, 2.3350E+08, 2.3576E+08, 2.3804E+08,
     + 2.4034E+08, 2.4266E+08, 2.4500E+08, 2.4735E+08, 2.4973E+08,
     +  2.5213E+08, 2.5455E+08, 2.5699E+08, 2.5945E+08, 2.6192E+08,
     + 2.6442E+08, 2.6695E+08, 2.6949E+08, 2.7205E+08, 2.7463E+08,
     +  2.7724E+08, 2.7987E+08, 2.8251E+08, 2.8518E+08, 2.8788E+08,
     + 2.9059E+08, 2.9333E+08, 2.9609E+08, 2.9887E+08, 3.0167E+08,
     +  3.0450E+08, 3.0735E+08, 3.1022E+08, 3.1312E+08, 3.1604E+08,
     + 3.1899E+08, 3.2195E+08, 3.2495E+08, 3.2796E+08, 3.3100E+08,
     +  3.3407E+08, 3.3716E+08, 3.4028E+08, 3.4342E+08, 3.4658E+08,
     + 3.4977E+08, 3.5299E+08, 3.5623E+08, 3.5950E+08, 3.6279E+08,
     +  3.6611E+08, 3.6946E+08, 3.7283E+08, 3.7623E+08, 3.7966E+08,
     + 3.8312E+08, 3.8660E+08, 3.9011E+08, 3.9364E+08, 3.9721E+08,
     +  4.0080E+08, 4.0442E+08, 4.0807E+08, 4.1175E+08, 4.1546E+08,
     + 4.1920E+08, 4.2296E+08, 4.2676E+08, 4.3058E+08, 4.3444E+08,
     +  4.3832E+08, 4.4224E+08, 4.4618E+08, 4.5016E+08, 4.5417E+08,
     + 4.5821E+08, 4.6228E+08, 4.6638E+08, 4.7051E+08, 4.7467E+08,
     +  4.7887E+08, 4.8310E+08, 4.8736E+08, 4.9166E+08, 4.9598E+08,
     + 5.0034E+08, 5.0474E+08, 5.0916E+08, 5.1362E+08, 5.1812E+08,
     +  5.2265E+08, 5.2721E+08, 5.3181E+08, 5.3644E+08, 5.4111E+08,
     + 5.4582E+08, 5.5056E+08, 5.5533E+08, 5.6014E+08, 5.6499E+08,
     +  5.6987E+08, 5.7480E+08, 5.7975E+08, 5.8475E+08, 5.8978E+08,
     + 5.9485E+08, 5.9996E+08, 6.0511E+08, 6.1029E+08, 6.1551E+08,
     +  6.2078E+08, 6.2608E+08, 6.3142E+08, 6.3680E+08, 6.4222E+08,
     + 6.4768E+08, 6.5318E+08, 6.5872E+08, 6.6431E+08, 6.6993E+08,
     +  6.7560E+08, 6.8130E+08, 6.8705E+08, 6.9284E+08, 6.9868E+08,
     + 7.0455E+08, 7.1047E+08, 7.1643E+08, 7.2244E+08, 7.2849E+08,
     +  7.3459E+08, 7.4072E+08, 7.4691E+08, 7.5314E+08, 7.5941E+08,
     + 7.6573E+08, 7.7209E+08, 7.7850E+08, 7.8496E+08, 7.9146E+08,
     +  7.9801E+08, 8.0461E+08, 8.1126E+08, 8.1795E+08, 8.2469E+08,
     + 8.3148E+08, 8.3832E+08, 8.4520E+08, 8.5214E+08, 8.5913E+08,
     +  8.6616E+08, 8.7325E+08, 8.8038E+08, 8.8757E+08, 8.9481E+08,
     + 9.0210E+08, 9.0944E+08, 9.1683E+08, 9.2428E+08, 9.3177E+08,
     +  9.3932E+08, 9.4693E+08, 9.5459E+08, 9.6230E+08, 9.7006E+08,
     + 9.7788E+08, 9.8576E+08, 9.9369E+08, 1.0017E+09, 1.0097E+09,
     +  1.0178E+09, 1.0260E+09, 1.0342E+09, 1.0424E+09, 1.0508E+09,
     + 1.0592E+09, 1.0676E+09, 1.0761E+09, 1.0847E+09, 1.0933E+09,
     +  1.1020E+09, 1.1107E+09, 1.1195E+09, 1.1283E+09, 1.1373E+09,
     + 1.1462E+09, 1.1553E+09, 1.1644E+09, 1.1736E+09, 1.1828E+09,
     +  1.1921E+09, 1.2014E+09, 1.2109E+09, 1.2204E+09, 1.2299E+09,
     + 1.2395E+09, 1.2492E+09, 1.2589E+09, 1.2688E+09, 1.2786E+09,
     +  1.2886E+09, 1.2986E+09, 1.3087E+09, 1.3188E+09, 1.3290E+09,
     + 1.3393E+09, 1.3497E+09, 1.3601E+09, 1.3706E+09, 1.3812E+09,
     +  1.3918E+09, 1.4025E+09, 1.4133E+09, 1.4241E+09, 1.4351E+09,
     + 1.4461E+09, 1.4571E+09, 1.4683E+09, 1.4795E+09, 1.4908E+09,
     +  1.5022E+09, 1.5136E+09, 1.5251E+09, 1.5367E+09, 1.5484E+09,
     + 1.5601E+09, 1.5720E+09, 1.5839E+09, 1.5958E+09, 1.6079E+09,
     +  1.6200E+09, 1.6323E+09, 1.6446E+09, 1.6569E+09, 1.6694E+09,
     + 1.6820E+09, 1.6946E+09, 1.7073E+09, 1.7201E+09, 1.7330E+09,
     +  1.7459E+09, 1.7590E+09, 1.7721E+09, 1.7853E+09, 1.7986E+09,
     + 1.8120E+09, 1.8255E+09, 1.8390E+09, 1.8527E+09, 1.8664E+09,
     +  1.8802E+09, 1.8941E+09, 1.9081E+09, 1.9222E+09, 1.9364E+09,
     + 1.9507E+09, 1.9650E+09, 1.9795E+09, 1.9940E+09, 2.0087E+09,
     +  2.0234E+09, 2.0383E+09, 2.0532E+09, 2.0682E+09, 2.0833E+09,
     + 2.0985E+09, 2.1138E+09, 2.1292E+09, 2.1447E+09, 2.1603E+09,
     +  2.1760E+09, 2.1918E+09, 2.2077E+09, 2.2237E+09, 2.2398E+09,
     + 2.2560E+09, 2.2723E+09, 2.2887E+09, 2.3052E+09, 2.3219E+09,
     +  2.3386E+09, 2.3554E+09, 2.3723E+09, 2.3893E+09, 2.4065E+09,
     + 2.4237E+09, 2.4411E+09, 2.4585E+09, 2.4761E+09, 2.4938E+09,
     +  2.5116E+09, 2.5295E+09, 2.5475E+09, 2.5656E+09, 2.5838E+09,
     + 2.6022E+09, 2.6206E+09, 2.6392E+09, 2.6579E+09, 2.6767E+09,
     +  2.6956E+09, 2.7147E+09, 2.7338E+09, 2.7531E+09, 2.7725E+09,
     + 2.7920E+09, 2.8116E+09, 2.8313E+09, 2.8512E+09, 2.8712E+09,
     +  2.8913E+09, 2.9115E+09, 2.9319E+09, 2.9524E+09, 2.9730E+09,
     + 2.9937E+09, 3.0145E+09, 3.0355E+09, 3.0566E+09, 3.0779E+09,
     +  3.0992E+09, 3.1207E+09, 3.1423E+09, 3.1641E+09, 3.1860E+09,
     + 3.2080E+09, 3.2301E+09, 3.2524E+09, 3.2748E+09, 3.2974E+09,
     +  3.3200E+09, 3.3429E+09, 3.3658E+09, 3.3889E+09, 3.4121E+09,
     + 3.4355E+09, 3.4590E+09, 3.4827E+09, 3.5064E+09, 3.5304E+09,
     +  3.5544E+09, 3.5786E+09, 3.6030E+09, 3.6275E+09, 3.6522E+09,
     + 3.6769E+09, 3.7019E+09, 3.7270E+09, 3.7522E+09, 3.7776E+09,
     +  3.8031E+09, 3.8288E+09, 3.8546E+09, 3.8806E+09, 3.9067E+09,
     + 3.9330E+09, 3.9595E+09, 3.9861E+09, 4.0128E+09, 4.0397E+09,
     +  4.0668E+09, 4.0940E+09, 4.1214E+09, 4.1489E+09, 4.1766E+09,
     + 4.2045E+09, 4.2325E+09, 4.2607E+09, 4.2890E+09, 4.3175E+09,
     +  4.3462E+09, 4.3751E+09, 4.4041E+09, 4.4332E+09, 4.4626E+09,
     + 4.4921E+09, 4.5218E+09, 4.5516E+09, 4.5817E+09, 4.6119E+09,
     +  4.6422E+09, 4.6728E+09, 4.7035E+09, 4.7344E+09, 4.7655E+09,
     + 4.7967E+09, 4.8281E+09, 4.8597E+09, 4.8915E+09, 4.9235E+09,
     +  4.9556E+09, 4.9880E+09, 5.0205E+09, 5.0532E+09, 5.0861E+09,
     + 5.1191E+09, 5.1524E+09, 5.1858E+09, 5.2194E+09, 5.2533E+09,
     +  5.2873E+09, 5.3215E+09, 5.3559E+09, 5.3905E+09, 5.4253E+09,
     + 5.4602E+09, 5.4954E+09, 5.5308E+09, 5.5664E+09, 5.6021E+09,
     +  5.6381E+09, 5.6743E+09, 5.7106E+09, 5.7472E+09, 5.7840E+09,
     + 5.8210E+09, 5.8582E+09, 5.8955E+09, 5.9332E+09, 5.9710E+09,
     +  6.0090E+09, 6.0472E+09, 6.0857E+09, 6.1243E+09, 6.1632E+09,
     + 6.2023E+09, 6.2416E+09, 6.2811E+09, 6.3208E+09, 6.3608E+09,
     +  6.4009E+09, 6.4413E+09, 6.4819E+09, 6.5228E+09, 6.5638E+09,
     + 6.6051E+09, 6.6466E+09, 6.6884E+09, 6.7303E+09, 6.7725E+09,
     +  6.8149E+09, 6.8576E+09, 6.9005E+09, 6.9436E+09, 6.9869E+09,
     + 7.0305E+09, 7.0744E+09, 7.1184E+09, 7.1627E+09, 7.2073E+09,
     +  7.2520E+09, 7.2971E+09, 7.3423E+09, 7.3878E+09, 7.4336E+09,
     + 7.4796E+09, 7.5259E+09, 7.5724E+09, 7.6191E+09, 7.6661E+09,
     +  7.7134E+09, 7.7609E+09, 7.8086E+09, 7.8566E+09, 7.9049E+09,
     + 7.9534E+09, 8.0022E+09, 8.0513E+09, 8.1006E+09, 8.1502E+09,
     +  8.2000E+09, 8.2501E+09, 8.3005E+09, 8.3511E+09, 8.4020E+09,
     + 8.4532E+09, 8.5046E+09, 8.5563E+09, 8.6083E+09, 8.6606E+09,
     +  8.7131E+09, 8.7659E+09, 8.8190E+09, 8.8724E+09, 8.9261E+09,
     + 8.9800E+09, 9.0342E+09, 9.0887E+09, 9.1435E+09, 9.1986E+09,
     +  9.2540E+09, 9.3096E+09, 9.3656E+09, 9.4218E+09, 9.4784E+09,
     + 9.5352E+09, 9.5923E+09, 9.6497E+09, 9.7075E+09, 9.7655E+09,
     +  9.8238E+09, 9.8824E+09, 9.9414E+09, 1.0001E+10, 1.0060E+10,
     + 1.0120E+10, 1.0180E+10, 1.0241E+10, 1.0301E+10, 1.0363E+10,
     +  1.0424E+10, 1.0486E+10, 1.0548E+10, 1.0610E+10, 1.0673E+10,
     + 1.0736E+10, 1.0799E+10, 1.0863E+10, 1.0927E+10, 1.0991E+10,
     +  1.1056E+10, 1.1121E+10, 1.1186E+10, 1.1252E+10, 1.1318E+10,
     + 1.1384E+10, 1.1451E+10, 1.1517E+10, 1.1585E+10, 1.1652E+10,
     +  1.1720E+10, 1.1789E+10, 1.1858E+10, 1.1927E+10, 1.1996E+10,
     + 1.2066E+10, 1.2136E+10, 1.2206E+10, 1.2277E+10, 1.2348E+10,
     +  1.2420E+10, 1.2492E+10, 1.2564E+10, 1.2637E+10, 1.2710E+10,
     + 1.2783E+10, 1.2857E+10, 1.2931E+10, 1.3005E+10, 1.3080E+10,
     +  1.3155E+10, 1.3231E+10, 1.3307E+10, 1.3383E+10, 1.3460E+10,
     + 1.3537E+10, 1.3614E+10, 1.3692E+10, 1.3770E+10, 1.3849E+10,
     +  1.3928E+10, 1.4008E+10, 1.4087E+10, 1.4168E+10, 1.4248E+10,
     + 1.4329E+10, 1.4411E+10, 1.4492E+10, 1.4575E+10, 1.4657E+10,
     +  1.4740E+10, 1.4824E+10, 1.4908E+10, 1.4992E+10, 1.5077E+10,
     + 1.5162E+10, 1.5247E+10, 1.5333E+10, 1.5419E+10, 1.5506E+10,
     +  1.5593E+10, 1.5681E+10, 1.5769E+10, 1.5858E+10, 1.5946E+10,
     + 1.6036E+10, 1.6126E+10, 1.6216E+10, 1.6306E+10, 1.6398E+10,
     +  1.6489E+10, 1.6581E+10, 1.6673E+10, 1.6766E+10, 1.6860E+10,
     + 1.6953E+10, 1.7048E+10, 1.7142E+10, 1.7237E+10, 1.7333E+10,
     +  1.7429E+10, 1.7526E+10, 1.7623E+10, 1.7720E+10, 1.7818E+10,
     + 1.7916E+10, 1.8015E+10, 1.8115E+10, 1.8214E+10, 1.8315E+10,
     +  1.8416E+10, 1.8517E+10, 1.8619E+10, 1.8721E+10, 1.8824E+10,
     + 1.8927E+10, 1.9030E+10, 1.9135E+10, 1.9239E+10, 1.9344E+10,
     +  1.9450E+10, 1.9556E+10, 1.9663E+10, 1.9770E+10, 1.9878E+10,
     + 1.9986E+10, 2.0095E+10, 2.0204E+10, 2.0314E+10, 2.0424E+10,
     +  2.0535E+10, 2.0646E+10, 2.0758E+10, 2.0871E+10, 2.0983E+10,
     + 2.1097E+10, 2.1211E+10, 2.1325E+10, 2.1440E+10, 2.1556E+10,
     +  2.1672E+10, 2.1789E+10, 2.1906E+10, 2.2024E+10, 2.2142E+10,
     + 2.2261E+10, 2.2380E+10, 2.2500E+10, 2.2621E+10, 2.2742E+10,
     +  2.2864E+10, 2.2986E+10, 2.3109E+10, 2.3232E+10, 2.3356E+10,
     + 2.3480E+10, 2.3605E+10, 2.3731E+10, 2.3857E+10, 2.3984E+10,
     +  2.4112E+10, 2.4240E+10, 2.4368E+10, 2.4497E+10, 2.4627E+10,
     + 2.4757E+10, 2.4888E+10, 2.5020E+10, 2.5152E+10, 2.5285E+10,
     +  2.5418E+10, 2.5552E+10, 2.5687E+10, 2.5822E+10, 2.5958E+10,
     + 2.6095E+10, 2.6232E+10, 2.6370E+10, 2.6508E+10, 2.6647E+10,
     +  2.6787E+10, 2.6927E+10, 2.7068E+10, 2.7209E+10, 2.7351E+10,
     + 2.7494E+10, 2.7638E+10, 2.7782E+10, 2.7927E+10, 2.8072E+10,
     +  2.8218E+10, 2.8365E+10, 2.8512E+10, 2.8660E+10, 2.8809E+10,
     + 2.8959E+10, 2.9109E+10, 2.9260E+10, 2.9411E+10, 2.9563E+10,
     +  2.9716E+10, 2.9869E+10, 3.0024E+10, 3.0179E+10, 3.0334E+10,
     + 3.0490E+10, 3.0647E+10, 3.0805E+10, 3.0964E+10, 3.1123E+10,
     +  3.1282E+10, 3.1443E+10, 3.1604E+10, 3.1766E+10, 3.1929E+10,
     + 3.2092E+10, 3.2256E+10, 3.2421E+10, 3.2587E+10, 3.2753E+10,
     +  3.2920E+10, 3.3088E+10, 3.3257E+10, 3.3426E+10, 3.3596E+10,
     + 3.3767E+10, 3.3938E+10, 3.4111E+10, 3.4284E+10, 3.4458E+10,
     +  3.4632E+10, 3.4808E+10, 3.4984E+10, 3.5161E+10, 3.5338E+10,
     + 3.5517E+10, 3.5696E+10, 3.5876E+10, 3.6057E+10, 3.6239E+10,
     +  3.6421E+10, 3.6604E+10, 3.6788E+10, 3.6973E+10, 3.7159E+10,
     + 3.7345E+10, 3.7532E+10, 3.7720E+10, 3.7909E+10, 3.8099E+10,
     +  3.8290E+10, 3.8481E+10, 3.8673E+10, 3.8866E+10, 3.9060E+10,
     + 3.9255E+10, 3.9450E+10, 3.9647E+10, 3.9844E+10, 4.0042E+10,
     +  4.0241E+10, 4.0441E+10, 4.0642E+10, 4.0843E+10, 4.1046E+10,
     + 4.1249E+10, 4.1453E+10, 4.1658E+10, 4.1864E+10, 4.2071E+10,
     +  4.2279E+10, 4.2487E+10, 4.2697E+10, 4.2907E+10, 4.3118E+10,
     + 4.3331E+10, 4.3544E+10, 4.3758E+10, 4.3973E+10, 4.4188E+10,
     +  4.4405E+10, 4.4623E+10, 4.4841E+10, 4.5061E+10, 4.5281E+10,
     + 4.5503E+10, 4.5725E+10, 4.5949E+10, 4.6173E+10, 4.6398E+10,
     +  4.6624E+10, 4.6851E+10, 4.7079E+10, 4.7308E+10, 4.7538E+10,
     + 4.7769E+10, 4.8001E+10, 4.8234E+10, 4.8468E+10, 4.8703E+10,
     +  4.8939E+10, 4.9176E+10, 4.9414E+10, 4.9653E+10, 4.9893E+10,
     + 5.0134E+10, 5.0375E+10, 5.0618E+10, 5.0862E+10, 5.1107E+10,
     +  5.1353E+10, 5.1600E+10, 5.1848E+10, 5.2097E+10, 5.2348E+10,
     + 5.2599E+10, 5.2851E+10, 5.3104E+10, 5.3359E+10, 5.3614E+10,
     +  5.3870E+10, 5.4128E+10, 5.4386E+10, 5.4646E+10, 5.4907E+10,
     + 5.5169E+10, 5.5432E+10, 5.5696E+10, 5.5961E+10, 5.6227E+10,
     +  5.6494E+10, 5.6763E+10, 5.7032E+10, 5.7303E+10, 5.7574E+10,
     + 5.7847E+10, 5.8121E+10, 5.8396E+10, 5.8673E+10, 5.8950E+10,
     +  5.9228E+10, 5.9508E+10, 5.9789E+10, 6.0071E+10, 6.0354E+10,
     + 6.0638E+10, 6.0924E+10, 6.1210E+10, 6.1498E+10, 6.1787E+10,
     +  6.2077E+10, 6.2368E+10, 6.2661E+10, 6.2955E+10, 6.3249E+10,
     + 6.3545E+10, 6.3843E+10, 6.4141E+10, 6.4441E+10, 6.4742E+10,
     +  6.5044E+10, 6.5347E+10, 6.5652E+10, 6.5958E+10, 6.6265E+10,
     + 6.6573E+10, 6.6883E+10, 6.7194E+10, 6.7506E+10, 6.7819E+10,
     +  6.8134E+10, 6.8450E+10, 6.8767E+10, 6.9085E+10, 6.9405E+10,
     + 6.9726E+10, 7.0048E+10, 7.0372E+10, 7.0697E+10, 7.1023E+10,
     +  7.1350E+10, 7.1679E+10, 7.2009E+10, 7.2341E+10, 7.2673E+10,
     + 7.3007E+10, 7.3343E+10, 7.3680E+10, 7.4018E+10, 7.4357E+10,
     +  7.4698E+10, 7.5040E+10, 7.5384E+10, 7.5729E+10, 7.6075E+10,
     + 7.6423E+10, 7.6772E+10, 7.7122E+10, 7.7474E+10, 7.7827E+10,
     +  7.8182E+10, 7.8538E+10, 7.8895E+10, 7.9254E+10, 7.9614E+10,
     + 7.9976E+10, 8.0339E+10, 8.0703E+10, 8.1069E+10, 8.1437E+10,
     +  8.1806E+10, 8.2176E+10, 8.2548E+10, 8.2921E+10, 8.3296E+10,
     + 8.3672E+10, 8.4050E+10, 8.4429E+10, 8.4809E+10, 8.5191E+10,
     +  8.5575E+10, 8.5960E+10, 8.6347E+10, 8.6735E+10, 8.7125E+10,
     + 8.7516E+10, 8.7908E+10, 8.8303E+10, 8.8698E+10, 8.9096E+10,
     +  8.9495E+10, 8.9895E+10, 9.0297E+10, 9.0700E+10, 9.1105E+10,
     + 9.1512E+10, 9.1920E+10, 9.2330E+10, 9.2742E+10, 9.3155E+10,
     +  9.3569E+10, 9.3985E+10, 9.4403E+10, 9.4823E+10, 9.5244E+10,
     + 9.5666E+10, 9.6091E+10, 9.6517E+10, 9.6944E+10, 9.7373E+10,
     +  9.7804E+10, 9.8237E+10, 9.8671E+10, 9.9107E+10, 9.9545E+10,
     + 9.9984E+10, 1.0042E+11, 1.0087E+11, 1.0131E+11, 1.0176E+11,
     +  1.0221E+11, 1.0265E+11, 1.0311E+11, 1.0356E+11, 1.0401E+11,
     + 1.0447E+11, 1.0493E+11, 1.0539E+11, 1.0585E+11, 1.0631E+11,
     +  1.0678E+11, 1.0724E+11, 1.0771E+11, 1.0818E+11, 1.0866E+11,
     + 1.0913E+11, 1.0961E+11, 1.1008E+11, 1.1056E+11, 1.1104E+11,
     +  1.1153E+11, 1.1201E+11, 1.1250E+11, 1.1299E+11, 1.1348E+11,
     + 1.1397E+11, 1.1446E+11, 1.1496E+11, 1.1546E+11, 1.1596E+11,
     +  1.1646E+11, 1.1696E+11, 1.1747E+11, 1.1797E+11, 1.1848E+11,
     + 1.1899E+11, 1.1951E+11, 1.2002E+11, 1.2054E+11, 1.2106E+11,
     +  1.2158E+11, 1.2210E+11, 1.2263E+11, 1.2315E+11, 1.2368E+11,
     + 1.2421E+11, 1.2474E+11, 1.2528E+11, 1.2581E+11, 1.2635E+11,
     +  1.2689E+11, 1.2744E+11, 1.2798E+11, 1.2853E+11, 1.2908E+11,
     + 1.2963E+11, 1.3018E+11, 1.3073E+11, 1.3129E+11, 1.3185E+11,
     +  1.3241E+11, 1.3297E+11, 1.3354E+11, 1.3410E+11, 1.3467E+11,
     + 1.3524E+11, 1.3582E+11, 1.3639E+11, 1.3697E+11, 1.3755E+11,
     +  1.3813E+11, 1.3872E+11, 1.3930E+11, 1.3989E+11, 1.4048E+11,
     + 1.4107E+11, 1.4167E+11, 1.4227E+11, 1.4287E+11, 1.4347E+11,
     +  1.4407E+11, 1.4468E+11, 1.4528E+11, 1.4589E+11, 1.4651E+11,
     + 1.4712E+11, 1.4774E+11, 1.4836E+11, 1.4898E+11, 1.4960E+11,
     +  1.5023E+11, 1.5086E+11, 1.5149E+11, 1.5212E+11, 1.5275E+11,
     + 1.5339E+11, 1.5403E+11, 1.5467E+11, 1.5532E+11, 1.5596E+11,
     +  1.5661E+11, 1.5727E+11, 1.5792E+11, 1.5858E+11, 1.5923E+11,
     + 1.5989E+11, 1.6056E+11, 1.6122E+11, 1.6189E+11, 1.6256E+11,
     +  1.6323E+11, 1.6391E+11, 1.6459E+11, 1.6527E+11, 1.6595E+11,
     + 1.6664E+11, 1.6732E+11, 1.6801E+11, 1.6871E+11, 1.6940E+11,
     +  1.7010E+11, 1.7080E+11, 1.7150E+11, 1.7220E+11, 1.7291E+11,
     + 1.7362E+11, 1.7433E+11, 1.7505E+11, 1.7577E+11, 1.7649E+11,
     +  1.7721E+11, 1.7793E+11, 1.7866E+11, 1.7939E+11, 1.8013E+11,
     + 1.8086E+11, 1.8160E+11, 1.8234E+11, 1.8308E+11, 1.8383E+11,
     +  1.8458E+11, 1.8533E+11, 1.8608E+11, 1.8684E+11, 1.8760E+11,
     + 1.8836E+11, 1.8913E+11, 1.8990E+11, 1.9067E+11, 1.9144E+11,
     +  1.9221E+11, 1.9299E+11, 1.9377E+11, 1.9456E+11, 1.9534E+11,
     + 1.9613E+11, 1.9693E+11, 1.9772E+11, 1.9852E+11, 1.9932E+11,
     +  2.0012E+11, 2.0093E+11, 2.0174E+11, 2.0255E+11, 2.0336E+11,
     + 2.0418E+11, 2.0500E+11, 2.0582E+11, 2.0665E+11, 2.0748E+11,
     +  2.0831E+11, 2.0915E+11, 2.0998E+11, 2.1082E+11, 2.1167E+11,
     + 2.1251E+11, 2.1336E+11, 2.1422E+11, 2.1507E+11, 2.1593E+11,
     +  2.1679E+11, 2.1766E+11, 2.1852E+11, 2.1939E+11, 2.2027E+11,
     + 2.2114E+11, 2.2202E+11, 2.2290E+11, 2.2379E+11, 2.2468E+11,
     +  2.2557E+11, 2.2646E+11, 2.2736E+11, 2.2826E+11, 2.2917E+11,
     + 2.3007E+11, 2.3098E+11, 2.3190E+11, 2.3281E+11, 2.3373E+11,
     +  2.3465E+11, 2.3558E+11, 2.3651E+11, 2.3744E+11, 2.3838E+11,
     + 2.3932E+11, 2.4026E+11, 2.4120E+11, 2.4215E+11, 2.4310E+11,
     +  2.4406E+11, 2.4502E+11, 2.4598E+11, 2.4694E+11, 2.4791E+11,
     + 2.4888E+11, 2.4985E+11, 2.5083E+11, 2.5181E+11, 2.5280E+11,
     +  2.5379E+11, 2.5478E+11, 2.5577E+11, 2.5677E+11, 2.5777E+11,
     + 2.5877E+11, 2.5978E+11, 2.6079E+11, 2.6181E+11, 2.6283E+11,
     +  2.6385E+11, 2.6487E+11, 2.6590E+11, 2.6694E+11, 2.6797E+11,
     + 2.6901E+11, 2.7005E+11, 2.7110E+11, 2.7215E+11, 2.7320E+11,
     +  2.7426E+11, 2.7532E+11, 2.7638E+11, 2.7745E+11, 2.7852E+11,
     + 2.7960E+11, 2.8067E+11, 2.8176E+11, 2.8284E+11, 2.8393E+11,
     +  2.8502E+11, 2.8612E+11, 2.8722E+11, 2.8832E+11, 2.8943E+11,
     + 2.9054E+11, 2.9166E+11, 2.9278E+11, 2.9390E+11, 2.9502E+11,
     +  2.9615E+11, 2.9729E+11, 2.9842E+11, 2.9957E+11, 3.0071E+11,
     + 3.0186E+11, 3.0301E+11, 3.0417E+11, 3.0533E+11, 3.0649E+11,
     +  3.0766E+11, 3.0883E+11, 3.1001E+11, 3.1119E+11, 3.1237E+11,
     + 3.1356E+11, 3.1475E+11, 3.1594E+11, 3.1714E+11, 3.1835E+11,
     +  3.1955E+11, 3.2077E+11, 3.2198E+11, 3.2320E+11, 3.2442E+11,
     + 3.2565E+11, 3.2688E+11, 3.2812E+11, 3.2936E+11, 3.3060E+11,
     +  3.3185E+11, 3.3310E+11, 3.3435E+11, 3.3561E+11, 3.3688E+11,
     + 3.3814E+11, 3.3942E+11, 3.4069E+11, 3.4197E+11, 3.4326E+11,
     +  3.4455E+11, 3.4584E+11, 3.4714E+11, 3.4844E+11, 3.4974E+11,
     + 3.5105E+11, 3.5237E+11, 3.5369E+11, 3.5501E+11, 3.5634E+11,
     +  3.5767E+11, 3.5900E+11, 3.6034E+11, 3.6169E+11, 3.6304E+11,
     + 3.6439E+11, 3.6575E+11, 3.6711E+11, 3.6848E+11, 3.6985E+11,
     +  3.7122E+11, 3.7260E+11, 3.7399E+11, 3.7537E+11, 3.7677E+11,
     + 3.7817E+11, 3.7957E+11, 3.8097E+11, 3.8239E+11, 3.8380E+11,
     +  3.8522E+11, 3.8665E+11, 3.8808E+11, 3.8951E+11, 3.9095E+11,
     + 3.9239E+11, 3.9384E+11, 3.9529E+11, 3.9675E+11, 3.9821E+11,
     +  3.9968E+11, 4.0115E+11, 4.0262E+11, 4.0410E+11, 4.0559E+11,
     + 4.0708E+11, 4.0857E+11, 4.1007E+11, 4.1158E+11, 4.1309E+11,
     +  4.1460E+11, 4.1612E+11, 4.1764E+11, 4.1917E+11, 4.2071E+11,
     + 4.2224E+11, 4.2379E+11, 4.2534E+11, 4.2689E+11, 4.2845E+11,
     +  4.3001E+11, 4.3158E+11, 4.3315E+11, 4.3473E+11, 4.3631E+11,
     + 4.3790E+11, 4.3949E+11, 4.4109E+11, 4.4269E+11, 4.4430E+11,
     +  4.4591E+11, 4.4753E+11, 4.4915E+11, 4.5078E+11, 4.5242E+11,
     + 4.5406E+11, 4.5570E+11, 4.5735E+11, 4.5900E+11, 4.6066E+11,
     +  4.6233E+11, 4.6400E+11, 4.6567E+11, 4.6735E+11, 4.6904E+11,
     + 4.7073E+11, 4.7242E+11, 4.7413E+11, 4.7583E+11, 4.7754E+11,
     +  4.7926E+11, 4.8098E+11, 4.8271E+11, 4.8445E+11, 4.8619E+11,
     + 4.8793E+11, 4.8968E+11, 4.9144E+11, 4.9320E+11, 4.9496E+11,
     +  4.9673E+11, 4.9851E+11, 5.0030E+11, 5.0208E+11, 5.0388E+11,
     + 5.0568E+11, 5.0748E+11, 5.0929E+11, 5.1111E+11, 5.1293E+11,
     +  5.1476E+11, 5.1659E+11, 5.1843E+11, 5.2028E+11, 5.2213E+11,
     + 5.2398E+11, 5.2585E+11, 5.2771E+11, 5.2959E+11, 5.3147E+11,
     +  5.3335E+11, 5.3524E+11, 5.3714E+11, 5.3904E+11, 5.4095E+11,
     + 5.4286E+11, 5.4478E+11, 5.4671E+11, 5.4864E+11, 5.5058E+11,
     +  5.5252E+11, 5.5447E+11, 5.5643E+11, 5.5839E+11, 5.6036E+11,
     + 5.6233E+11, 5.6431E+11, 5.6630E+11, 5.6829E+11, 5.7029E+11,
     +  5.7229E+11, 5.7430E+11, 5.7632E+11, 5.7834E+11, 5.8037E+11,
     + 5.8240E+11, 5.8445E+11, 5.8649E+11, 5.8855E+11, 5.9061E+11,
     +  5.9267E+11, 5.9475E+11, 5.9682E+11, 5.9891E+11, 6.0100E+11,
     + 6.0310E+11, 6.0520E+11, 6.0731E+11, 6.0943E+11, 6.1155E+11,
     +  6.1368E+11, 6.1582E+11, 6.1796E+11, 6.2011E+11, 6.2227E+11,
     + 6.2443E+11, 6.2660E+11, 6.2877E+11, 6.3095E+11, 6.3314E+11,
     +  6.3534E+11, 6.3754E+11, 6.3975E+11, 6.4196E+11, 6.4418E+11,
     + 6.4641E+11, 6.4865E+11, 6.5089E+11, 6.5314E+11, 6.5539E+11,
     +  6.5766E+11, 6.5992E+11, 6.6220E+11, 6.6448E+11, 6.6677E+11,
     + 6.6907E+11, 6.7137E+11, 6.7368E+11, 6.7600E+11, 6.7832E+11,
     +  6.8065E+11, 6.8299E+11, 6.8533E+11, 6.8768E+11, 6.9004E+11,
     + 6.9241E+11, 6.9478E+11, 6.9716E+11, 6.9955E+11, 7.0194E+11,
     +  7.0434E+11, 7.0675E+11, 7.0917E+11, 7.1159E+11, 7.1402E+11,
     + 7.1646E+11, 7.1890E+11, 7.2135E+11, 7.2381E+11, 7.2628E+11,
     +  7.2875E+11, 7.3123E+11, 7.3372E+11, 7.3621E+11, 7.3872E+11,
     + 7.4123E+11, 7.4374E+11, 7.4627E+11, 7.4880E+11, 7.5134E+11,
     +  7.5389E+11, 7.5644E+11, 7.5901E+11, 7.6158E+11, 7.6415E+11,
     + 7.6674E+11, 7.6933E+11, 7.7193E+11, 7.7454E+11, 7.7715E+11,
     +  7.7978E+11, 7.8241E+11, 7.8505E+11, 7.8769E+11, 7.9035E+11,
     + 7.9301E+11, 7.9568E+11, 7.9836E+11, 8.0104E+11, 8.0374E+11,
     +  8.0644E+11, 8.0915E+11, 8.1186E+11, 8.1459E+11, 8.1732E+11,
     + 8.2006E+11, 8.2281E+11, 8.2557E+11, 8.2833E+11, 8.3111E+11,
     +  8.3389E+11, 8.3668E+11, 8.3947E+11, 8.4228E+11, 8.4509E+11,
     + 8.4791E+11, 8.5074E+11, 8.5358E+11, 8.5643E+11, 8.5928E+11,
     +  8.6215E+11, 8.6502E+11, 8.6790E+11, 8.7079E+11, 8.7368E+11,
     + 8.7659E+11, 8.7950E+11, 8.8242E+11, 8.8535E+11, 8.8829E+11,
     +  8.9123E+11, 8.9419E+11, 8.9715E+11, 9.0013E+11, 9.0311E+11,
     + 9.0610E+11, 9.0909E+11, 9.1210E+11, 9.1512E+11, 9.1814E+11,
     +  9.2117E+11, 9.2421E+11, 9.2726E+11, 9.3032E+11, 9.3339E+11,
     + 9.3647E+11, 9.3955E+11, 9.4265E+11, 9.4575E+11, 9.4886E+11,
     +  9.5198E+11, 9.5511E+11, 9.5825E+11, 9.6140E+11, 9.6456E+11,
     + 9.6772E+11, 9.7090E+11, 9.7408E+11, 9.7727E+11, 9.8047E+11,
     +  9.8369E+11, 9.8691E+11, 9.9014E+11, 9.9337E+11, 9.9662E+11,
     + 9.9988E+11, 1.0031E+12, 1.0064E+12, 1.0097E+12, 1.0130E+12,
     +  1.0163E+12, 1.0196E+12, 1.0229E+12, 1.0263E+12, 1.0296E+12,
     + 1.0330E+12, 1.0363E+12, 1.0397E+12, 1.0431E+12, 1.0465E+12,
     +  1.0499E+12, 1.0533E+12, 1.0567E+12, 1.0601E+12, 1.0635E+12,
     + 1.0670E+12, 1.0704E+12, 1.0739E+12, 1.0774E+12, 1.0809E+12,
     +  1.0844E+12, 1.0879E+12, 1.0914E+12, 1.0949E+12, 1.0984E+12,
     + 1.1020E+12, 1.1055E+12, 1.1091E+12, 1.1127E+12, 1.1163E+12,
     +  1.1199E+12, 1.1235E+12, 1.1271E+12, 1.1307E+12, 1.1343E+12,
     + 1.1380E+12, 1.1416E+12, 1.1453E+12, 1.1490E+12, 1.1527E+12,
     +  1.1564E+12, 1.1601E+12, 1.1638E+12, 1.1675E+12, 1.1713E+12,
     + 1.1750E+12, 1.1788E+12, 1.1825E+12, 1.1863E+12, 1.1901E+12,
     +  1.1939E+12, 1.1977E+12, 1.2015E+12, 1.2054E+12, 1.2092E+12,
     + 1.2131E+12, 1.2169E+12, 1.2208E+12, 1.2247E+12, 1.2286E+12,
     +  1.2325E+12, 1.2364E+12, 1.2403E+12, 1.2443E+12, 1.2482E+12,
     + 1.2522E+12, 1.2562E+12, 1.2601E+12, 1.2641E+12, 1.2681E+12,
     +  1.2722E+12, 1.2762E+12, 1.2802E+12, 1.2843E+12, 1.2883E+12,
     + 1.2924E+12, 1.2965E+12, 1.3006E+12, 1.3047E+12, 1.3088E+12,
     +  1.3129E+12, 1.3171E+12, 1.3212E+12, 1.3254E+12, 1.3296E+12,
     + 1.3338E+12, 1.3380E+12, 1.3422E+12, 1.3464E+12, 1.3506E+12,
     +  1.3549E+12, 1.3591E+12, 1.3634E+12, 1.3677E+12, 1.3719E+12,
     + 1.3762E+12, 1.3806E+12, 1.3849E+12, 1.3892E+12, 1.3936E+12,
     +  1.3979E+12, 1.4023E+12, 1.4067E+12, 1.4111E+12, 1.4155E+12,
     + 1.4199E+12, 1.4243E+12, 1.4288E+12, 1.4332E+12, 1.4377E+12,
     +  1.4422E+12, 1.4467E+12, 1.4512E+12, 1.4557E+12, 1.4602E+12,
     + 1.4648E+12, 1.4693E+12, 1.4739E+12, 1.4785E+12, 1.4830E+12,
     +  1.4876E+12, 1.4923E+12, 1.4969E+12, 1.5015E+12, 1.5062E+12,
     + 1.5108E+12, 1.5155E+12, 1.5202E+12, 1.5249E+12, 1.5296E+12,
     +  1.5344E+12, 1.5391E+12, 1.5439E+12, 1.5486E+12, 1.5534E+12,
     + 1.5582E+12, 1.5630E+12, 1.5678E+12, 1.5726E+12, 1.5775E+12,
     +  1.5823E+12, 1.5872E+12, 1.5921E+12, 1.5970E+12, 1.6019E+12,
     + 1.6068E+12, 1.6118E+12, 1.6167E+12, 1.6217E+12, 1.6266E+12,
     +  1.6316E+12, 1.6366E+12, 1.6416E+12, 1.6467E+12, 1.6517E+12,
     + 1.6568E+12, 1.6618E+12, 1.6669E+12, 1.6720E+12, 1.6771E+12,
     +  1.6822E+12, 1.6874E+12, 1.6925E+12, 1.6977E+12, 1.7029E+12,
     + 1.7081E+12, 1.7133E+12, 1.7185E+12, 1.7237E+12, 1.7290E+12,
     +  1.7342E+12, 1.7395E+12, 1.7448E+12, 1.7501E+12, 1.7554E+12,
     + 1.7607E+12, 1.7661E+12, 1.7714E+12, 1.7768E+12, 1.7822E+12,
     +  1.7876E+12, 1.7930E+12, 1.7984E+12, 1.8039E+12, 1.8093E+12,
     + 1.8148E+12, 1.8203E+12, 1.8258E+12, 1.8313E+12, 1.8368E+12,
     +  1.8424E+12, 1.8479E+12, 1.8535E+12, 1.8591E+12, 1.8647E+12,
     + 1.8703E+12, 1.8759E+12, 1.8816E+12, 1.8872E+12, 1.8929E+12,
     +  1.8986E+12, 1.9043E+12, 1.9100E+12, 1.9158E+12, 1.9215E+12,
     + 1.9273E+12, 1.9331E+12, 1.9389E+12, 1.9447E+12, 1.9505E+12,
     +  1.9563E+12, 1.9622E+12, 1.9681E+12, 1.9740E+12, 1.9799E+12,
     + 1.9858E+12, 1.9917E+12, 1.9977E+12, 2.0036E+12, 2.0096E+12,
     +  2.0156E+12, 2.0216E+12, 2.0276E+12, 2.0337E+12, 2.0397E+12,
     + 2.0458E+12, 2.0519E+12, 2.0580E+12, 2.0641E+12, 2.0703E+12,
     +  2.0764E+12, 2.0826E+12, 2.0888E+12, 2.0950E+12, 2.1012E+12,
     + 2.1074E+12, 2.1137E+12, 2.1199E+12, 2.1262E+12, 2.1325E+12,
     +  2.1388E+12, 2.1451E+12, 2.1515E+12, 2.1579E+12, 2.1642E+12,
     + 2.1706E+12, 2.1770E+12, 2.1835E+12, 2.1899E+12, 2.1964E+12,
     +  2.2028E+12, 2.2093E+12, 2.2158E+12, 2.2224E+12, 2.2289E+12,
     + 2.2355E+12, 2.2421E+12, 2.2486E+12, 2.2553E+12, 2.2619E+12,
     +  2.2685E+12, 2.2752E+12, 2.2819E+12, 2.2886E+12, 2.2953E+12,
     + 2.3020E+12, 2.3088E+12, 2.3155E+12, 2.3223E+12, 2.3291E+12,
     +  2.3359E+12, 2.3428E+12, 2.3496E+12, 2.3565E+12, 2.3634E+12,
     + 2.3703E+12, 2.3772E+12, 2.3841E+12, 2.3911E+12, 2.3980E+12,
     +  2.4050E+12, 2.4120E+12, 2.4191E+12, 2.4261E+12, 2.4332E+12,
     + 2.4403E+12, 2.4474E+12, 2.4545E+12, 2.4616E+12, 2.4688E+12,
     +  2.4759E+12, 2.4831E+12, 2.4903E+12, 2.4976E+12, 2.5048E+12/
      data (QofT( 1,J),J=2409,4510)/ 2.5121E+12, 2.5193E+12,
     +  2.5266E+12, 2.5340E+12, 2.5413E+12, 2.5486E+12, 2.5560E+12,
     +  2.5634E+12,
     + 2.5708E+12, 2.5782E+12, 2.5857E+12, 2.5932E+12, 2.6006E+12,
     +  2.6081E+12, 2.6157E+12, 2.6232E+12, 2.6308E+12, 2.6383E+12,
     + 2.6459E+12, 2.6536E+12, 2.6612E+12, 2.6688E+12, 2.6765E+12,
     +  2.6842E+12, 2.6919E+12, 2.6996E+12, 2.7074E+12, 2.7152E+12,
     + 2.7230E+12, 2.7308E+12, 2.7386E+12, 2.7464E+12, 2.7543E+12,
     +  2.7622E+12, 2.7701E+12, 2.7780E+12, 2.7860E+12, 2.7939E+12,
     + 2.8019E+12, 2.8099E+12, 2.8180E+12, 2.8260E+12, 2.8341E+12,
     +  2.8422E+12, 2.8503E+12, 2.8584E+12, 2.8665E+12, 2.8747E+12,
     + 2.8829E+12, 2.8911E+12, 2.8993E+12, 2.9076E+12, 2.9158E+12,
     +  2.9241E+12, 2.9324E+12, 2.9408E+12, 2.9491E+12, 2.9575E+12,
     + 2.9659E+12, 2.9743E+12, 2.9827E+12, 2.9912E+12, 2.9996E+12,
     +  3.0081E+12, 3.0166E+12, 3.0252E+12, 3.0337E+12, 3.0423E+12,
     + 3.0509E+12, 3.0595E+12, 3.0682E+12, 3.0768E+12, 3.0855E+12,
     +  3.0942E+12, 3.1030E+12, 3.1117E+12, 3.1205E+12, 3.1293E+12,
     + 3.1381E+12, 3.1469E+12, 3.1558E+12, 3.1647E+12, 3.1736E+12,
     +  3.1825E+12, 3.1914E+12, 3.2004E+12, 3.2094E+12, 3.2184E+12,
     + 3.2274E+12, 3.2365E+12, 3.2455E+12, 3.2546E+12, 3.2638E+12,
     +  3.2729E+12, 3.2821E+12, 3.2912E+12, 3.3005E+12, 3.3097E+12,
     + 3.3189E+12, 3.3282E+12, 3.3375E+12, 3.3468E+12, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --      2657
      data (QofT( 2,J),J=1,2408)/ 3.5368E+02, 9.8127E+02, 1.7913E+03,
     +  2.7492E+03, 3.8349E+03, 5.0348E+03, 6.3390E+03, 7.7397E+03,
     + 9.2306E+03, 1.0807E+04, 1.2464E+04, 1.4197E+04, 1.6005E+04,
     +  1.7883E+04, 1.9830E+04, 2.1843E+04, 2.3919E+04, 2.6058E+04,
     + 2.8257E+04, 3.0514E+04, 3.5199E+04, 4.0103E+04, 4.5215E+04,
     +  5.0528E+04, 5.6034E+04, 6.1728E+04, 6.7603E+04, 7.3653E+04,
     + 7.9875E+04, 8.6264E+04, 9.2816E+04, 9.9527E+04, 1.0639E+05,
     +  1.1342E+05, 1.2059E+05, 1.2791E+05, 1.3538E+05, 1.4300E+05,
     + 1.5076E+05, 1.5866E+05, 1.6671E+05, 1.7490E+05, 1.8324E+05,
     +  1.9171E+05, 2.0034E+05, 2.0910E+05, 2.1801E+05, 2.2707E+05,
     + 2.3628E+05, 2.4563E+05, 2.5513E+05, 2.6479E+05, 2.7460E+05,
     +  2.8456E+05, 2.9468E+05, 3.0497E+05, 3.1541E+05, 3.2602E+05,
     + 3.3680E+05, 3.4775E+05, 3.5887E+05, 3.7016E+05, 3.8164E+05,
     +  3.9329E+05, 4.0514E+05, 4.1717E+05, 4.2939E+05, 4.4181E+05,
     + 4.5443E+05, 4.6725E+05, 4.8028E+05, 4.9352E+05, 5.0698E+05,
     +  5.2065E+05, 5.3455E+05, 5.4868E+05, 5.6303E+05, 5.7763E+05,
     + 5.9246E+05, 6.0754E+05, 6.2287E+05, 6.3845E+05, 6.5430E+05,
     +  6.7040E+05, 6.8678E+05, 7.0343E+05, 7.2036E+05, 7.3757E+05,
     + 7.5507E+05, 7.7287E+05, 7.9097E+05, 8.0937E+05, 8.2808E+05,
     +  8.4712E+05, 8.6647E+05, 8.8615E+05, 9.0617E+05, 9.2652E+05,
     + 9.4722E+05, 9.6828E+05, 9.8969E+05, 1.0115E+06, 1.0336E+06,
     +  1.0561E+06, 1.0790E+06, 1.1023E+06, 1.1260E+06, 1.1501E+06,
     + 1.1746E+06, 1.1996E+06, 1.2249E+06, 1.2507E+06, 1.2769E+06,
     +  1.3036E+06, 1.3307E+06, 1.3582E+06, 1.3863E+06, 1.4148E+06,
     + 1.4438E+06, 1.4733E+06, 1.5033E+06, 1.5338E+06, 1.5648E+06,
     +  1.5963E+06, 1.6284E+06, 1.6609E+06, 1.6941E+06, 1.7278E+06,
     + 1.7621E+06, 1.7969E+06, 1.8323E+06, 1.8683E+06, 1.9049E+06,
     +  1.9421E+06, 1.9800E+06, 2.0184E+06, 2.0575E+06, 2.0973E+06,
     + 2.1377E+06, 2.1787E+06, 2.2205E+06, 2.2629E+06, 2.3060E+06,
     +  2.3498E+06, 2.3944E+06, 2.4396E+06, 2.4857E+06, 2.5324E+06,
     + 2.5799E+06, 2.6282E+06, 2.6773E+06, 2.7271E+06, 2.7778E+06,
     +  2.8292E+06, 2.8815E+06, 2.9347E+06, 2.9887E+06, 3.0435E+06,
     + 3.0992E+06, 3.1558E+06, 3.2134E+06, 3.2718E+06, 3.3311E+06,
     +  3.3914E+06, 3.4526E+06, 3.5148E+06, 3.5780E+06, 3.6421E+06,
     + 3.7073E+06, 3.7735E+06, 3.8407E+06, 3.9089E+06, 3.9782E+06,
     +  4.0486E+06, 4.1201E+06, 4.1926E+06, 4.2663E+06, 4.3412E+06,
     + 4.4171E+06, 4.4942E+06, 4.5725E+06, 4.6520E+06, 4.7327E+06,
     +  4.8147E+06, 4.8978E+06, 4.9823E+06, 5.0680E+06, 5.1550E+06,
     + 5.2433E+06, 5.3329E+06, 5.4238E+06, 5.5162E+06, 5.6099E+06,
     +  5.7049E+06, 5.8014E+06, 5.8994E+06, 5.9988E+06, 6.0996E+06,
     + 6.2019E+06, 6.3058E+06, 6.4111E+06, 6.5180E+06, 6.6265E+06,
     +  6.7365E+06, 6.8481E+06, 6.9614E+06, 7.0763E+06, 7.1929E+06,
     + 7.3111E+06, 7.4311E+06, 7.5527E+06, 7.6761E+06, 7.8013E+06,
     +  7.9283E+06, 8.0571E+06, 8.1877E+06, 8.3202E+06, 8.4545E+06,
     + 8.5908E+06, 8.7289E+06, 8.8691E+06, 9.0112E+06, 9.1552E+06,
     +  9.3013E+06, 9.4495E+06, 9.5997E+06, 9.7520E+06, 9.9064E+06,
     + 1.0063E+07, 1.0222E+07, 1.0383E+07, 1.0546E+07, 1.0711E+07,
     +  1.0879E+07, 1.1049E+07, 1.1221E+07, 1.1396E+07, 1.1573E+07,
     + 1.1752E+07, 1.1934E+07, 1.2118E+07, 1.2305E+07, 1.2495E+07,
     +  1.2686E+07, 1.2881E+07, 1.3078E+07, 1.3278E+07, 1.3480E+07,
     + 1.3685E+07, 1.3893E+07, 1.4104E+07, 1.4317E+07, 1.4533E+07,
     +  1.4752E+07, 1.4974E+07, 1.5199E+07, 1.5427E+07, 1.5658E+07,
     + 1.5892E+07, 1.6129E+07, 1.6369E+07, 1.6612E+07, 1.6858E+07,
     +  1.7108E+07, 1.7360E+07, 1.7616E+07, 1.7876E+07, 1.8138E+07,
     + 1.8404E+07, 1.8674E+07, 1.8947E+07, 1.9223E+07, 1.9503E+07,
     +  1.9787E+07, 2.0074E+07, 2.0364E+07, 2.0659E+07, 2.0957E+07,
     + 2.1259E+07, 2.1564E+07, 2.1874E+07, 2.2187E+07, 2.2505E+07,
     +  2.2826E+07, 2.3151E+07, 2.3481E+07, 2.3814E+07, 2.4152E+07,
     + 2.4494E+07, 2.4840E+07, 2.5190E+07, 2.5545E+07, 2.5904E+07,
     +  2.6267E+07, 2.6635E+07, 2.7007E+07, 2.7384E+07, 2.7766E+07,
     + 2.8152E+07, 2.8543E+07, 2.8939E+07, 2.9339E+07, 2.9744E+07,
     +  3.0154E+07, 3.0570E+07, 3.0990E+07, 3.1415E+07, 3.1845E+07,
     + 3.2280E+07, 3.2721E+07, 3.3167E+07, 3.3618E+07, 3.4075E+07,
     +  3.4537E+07, 3.5004E+07, 3.5477E+07, 3.5955E+07, 3.6440E+07,
     + 3.6930E+07, 3.7425E+07, 3.7927E+07, 3.8434E+07, 3.8947E+07,
     +  3.9466E+07, 3.9992E+07, 4.0523E+07, 4.1061E+07, 4.1604E+07,
     + 4.2154E+07, 4.2711E+07, 4.3274E+07, 4.3843E+07, 4.4419E+07,
     +  4.5001E+07, 4.5591E+07, 4.6186E+07, 4.6789E+07, 4.7399E+07,
     + 4.8015E+07, 4.8639E+07, 4.9269E+07, 4.9907E+07, 5.0552E+07,
     +  5.1204E+07, 5.1864E+07, 5.2531E+07, 5.3205E+07, 5.3887E+07,
     + 5.4577E+07, 5.5274E+07, 5.5979E+07, 5.6692E+07, 5.7413E+07,
     +  5.8142E+07, 5.8879E+07, 5.9624E+07, 6.0378E+07, 6.1140E+07,
     + 6.1910E+07, 6.2688E+07, 6.3475E+07, 6.4271E+07, 6.5076E+07,
     +  6.5889E+07, 6.6711E+07, 6.7542E+07, 6.8382E+07, 6.9232E+07,
     + 7.0090E+07, 7.0958E+07, 7.1835E+07, 7.2722E+07, 7.3618E+07,
     +  7.4524E+07, 7.5439E+07, 7.6364E+07, 7.7300E+07, 7.8245E+07,
     + 7.9200E+07, 8.0166E+07, 8.1142E+07, 8.2128E+07, 8.3124E+07,
     +  8.4132E+07, 8.5149E+07, 8.6178E+07, 8.7217E+07, 8.8268E+07,
     + 8.9329E+07, 9.0402E+07, 9.1485E+07, 9.2581E+07, 9.3687E+07,
     +  9.4805E+07, 9.5935E+07, 9.7076E+07, 9.8230E+07, 9.9395E+07,
     + 1.0057E+08, 1.0176E+08, 1.0296E+08, 1.0418E+08, 1.0540E+08,
     +  1.0664E+08, 1.0790E+08, 1.0916E+08, 1.1044E+08, 1.1173E+08,
     + 1.1303E+08, 1.1435E+08, 1.1568E+08, 1.1703E+08, 1.1838E+08,
     +  1.1975E+08, 1.2114E+08, 1.2254E+08, 1.2395E+08, 1.2538E+08,
     + 1.2682E+08, 1.2828E+08, 1.2975E+08, 1.3124E+08, 1.3274E+08,
     +  1.3425E+08, 1.3578E+08, 1.3733E+08, 1.3889E+08, 1.4047E+08,
     + 1.4206E+08, 1.4367E+08, 1.4529E+08, 1.4693E+08, 1.4859E+08,
     +  1.5026E+08, 1.5195E+08, 1.5365E+08, 1.5538E+08, 1.5711E+08,
     + 1.5887E+08, 1.6064E+08, 1.6243E+08, 1.6424E+08, 1.6607E+08,
     +  1.6791E+08, 1.6977E+08, 1.7165E+08, 1.7355E+08, 1.7546E+08,
     + 1.7739E+08, 1.7935E+08, 1.8132E+08, 1.8331E+08, 1.8531E+08,
     +  1.8734E+08, 1.8939E+08, 1.9146E+08, 1.9354E+08, 1.9565E+08,
     + 1.9778E+08, 1.9992E+08, 2.0209E+08, 2.0428E+08, 2.0648E+08,
     +  2.0871E+08, 2.1096E+08, 2.1323E+08, 2.1552E+08, 2.1784E+08,
     + 2.2017E+08, 2.2253E+08, 2.2491E+08, 2.2731E+08, 2.2973E+08,
     +  2.3218E+08, 2.3465E+08, 2.3714E+08, 2.3965E+08, 2.4219E+08,
     + 2.4475E+08, 2.4734E+08, 2.4995E+08, 2.5258E+08, 2.5524E+08,
     +  2.5792E+08, 2.6062E+08, 2.6336E+08, 2.6611E+08, 2.6889E+08,
     + 2.7170E+08, 2.7453E+08, 2.7739E+08, 2.8027E+08, 2.8318E+08,
     +  2.8611E+08, 2.8908E+08, 2.9206E+08, 2.9508E+08, 2.9812E+08,
     + 3.0119E+08, 3.0429E+08, 3.0742E+08, 3.1057E+08, 3.1375E+08,
     +  3.1696E+08, 3.2020E+08, 3.2347E+08, 3.2676E+08, 3.3009E+08,
     + 3.3344E+08, 3.3683E+08, 3.4024E+08, 3.4369E+08, 3.4717E+08,
     +  3.5067E+08, 3.5421E+08, 3.5778E+08, 3.6138E+08, 3.6501E+08,
     + 3.6867E+08, 3.7237E+08, 3.7609E+08, 3.7985E+08, 3.8364E+08,
     +  3.8747E+08, 3.9133E+08, 3.9522E+08, 3.9915E+08, 4.0311E+08,
     + 4.0710E+08, 4.1113E+08, 4.1519E+08, 4.1929E+08, 4.2342E+08,
     +  4.2759E+08, 4.3180E+08, 4.3604E+08, 4.4031E+08, 4.4463E+08,
     + 4.4898E+08, 4.5336E+08, 4.5779E+08, 4.6225E+08, 4.6675E+08,
     +  4.7129E+08, 4.7587E+08, 4.8048E+08, 4.8514E+08, 4.8983E+08,
     + 4.9456E+08, 4.9934E+08, 5.0415E+08, 5.0900E+08, 5.1390E+08,
     +  5.1883E+08, 5.2381E+08, 5.2883E+08, 5.3389E+08, 5.3899E+08,
     + 5.4414E+08, 5.4932E+08, 5.5456E+08, 5.5983E+08, 5.6515E+08,
     +  5.7051E+08, 5.7591E+08, 5.8137E+08, 5.8686E+08, 5.9240E+08,
     + 5.9799E+08, 6.0362E+08, 6.0930E+08, 6.1502E+08, 6.2080E+08,
     +  6.2661E+08, 6.3248E+08, 6.3840E+08, 6.4436E+08, 6.5037E+08,
     + 6.5643E+08, 6.6254E+08, 6.6870E+08, 6.7490E+08, 6.8116E+08,
     +  6.8747E+08, 6.9383E+08, 7.0024E+08, 7.0671E+08, 7.1322E+08,
     + 7.1979E+08, 7.2641E+08, 7.3308E+08, 7.3981E+08, 7.4659E+08,
     +  7.5342E+08, 7.6031E+08, 7.6725E+08, 7.7425E+08, 7.8131E+08,
     + 7.8842E+08, 7.9558E+08, 8.0280E+08, 8.1009E+08, 8.1742E+08,
     +  8.2482E+08, 8.3227E+08, 8.3978E+08, 8.4736E+08, 8.5499E+08,
     + 8.6268E+08, 8.7043E+08, 8.7824E+08, 8.8611E+08, 8.9405E+08,
     +  9.0204E+08, 9.1010E+08, 9.1822E+08, 9.2641E+08, 9.3465E+08,
     + 9.4297E+08, 9.5134E+08, 9.5978E+08, 9.6829E+08, 9.7686E+08,
     +  9.8550E+08, 9.9420E+08, 1.0030E+09, 1.0118E+09, 1.0207E+09,
     + 1.0297E+09, 1.0387E+09, 1.0478E+09, 1.0570E+09, 1.0663E+09,
     +  1.0756E+09, 1.0850E+09, 1.0945E+09, 1.1040E+09, 1.1136E+09,
     + 1.1233E+09, 1.1330E+09, 1.1429E+09, 1.1528E+09, 1.1627E+09,
     +  1.1728E+09, 1.1829E+09, 1.1931E+09, 1.2034E+09, 1.2138E+09,
     + 1.2242E+09, 1.2347E+09, 1.2453E+09, 1.2560E+09, 1.2667E+09,
     +  1.2776E+09, 1.2885E+09, 1.2995E+09, 1.3105E+09, 1.3217E+09,
     + 1.3329E+09, 1.3443E+09, 1.3557E+09, 1.3672E+09, 1.3787E+09,
     +  1.3904E+09, 1.4021E+09, 1.4140E+09, 1.4259E+09, 1.4379E+09,
     + 1.4500E+09, 1.4622E+09, 1.4745E+09, 1.4868E+09, 1.4993E+09,
     +  1.5118E+09, 1.5245E+09, 1.5372E+09, 1.5500E+09, 1.5629E+09,
     + 1.5759E+09, 1.5890E+09, 1.6022E+09, 1.6155E+09, 1.6289E+09,
     +  1.6424E+09, 1.6560E+09, 1.6696E+09, 1.6834E+09, 1.6973E+09,
     + 1.7113E+09, 1.7253E+09, 1.7395E+09, 1.7538E+09, 1.7682E+09,
     +  1.7826E+09, 1.7972E+09, 1.8119E+09, 1.8267E+09, 1.8416E+09,
     + 1.8566E+09, 1.8717E+09, 1.8869E+09, 1.9023E+09, 1.9177E+09,
     +  1.9332E+09, 1.9489E+09, 1.9646E+09, 1.9805E+09, 1.9965E+09,
     + 2.0126E+09, 2.0288E+09, 2.0451E+09, 2.0616E+09, 2.0781E+09,
     +  2.0948E+09, 2.1116E+09, 2.1285E+09, 2.1455E+09, 2.1626E+09,
     + 2.1799E+09, 2.1973E+09, 2.2148E+09, 2.2324E+09, 2.2501E+09,
     +  2.2680E+09, 2.2860E+09, 2.3041E+09, 2.3223E+09, 2.3407E+09,
     + 2.3592E+09, 2.3778E+09, 2.3965E+09, 2.4154E+09, 2.4344E+09,
     +  2.4535E+09, 2.4728E+09, 2.4922E+09, 2.5117E+09, 2.5313E+09,
     + 2.5511E+09, 2.5711E+09, 2.5911E+09, 2.6113E+09, 2.6317E+09,
     +  2.6521E+09, 2.6727E+09, 2.6935E+09, 2.7144E+09, 2.7354E+09,
     + 2.7566E+09, 2.7779E+09, 2.7994E+09, 2.8210E+09, 2.8427E+09,
     +  2.8646E+09, 2.8866E+09, 2.9088E+09, 2.9312E+09, 2.9536E+09,
     + 2.9763E+09, 2.9991E+09, 3.0220E+09, 3.0451E+09, 3.0683E+09,
     +  3.0917E+09, 3.1153E+09, 3.1390E+09, 3.1629E+09, 3.1869E+09,
     + 3.2111E+09, 3.2354E+09, 3.2599E+09, 3.2846E+09, 3.3094E+09,
     +  3.3344E+09, 3.3596E+09, 3.3849E+09, 3.4104E+09, 3.4360E+09,
     + 3.4619E+09, 3.4878E+09, 3.5140E+09, 3.5403E+09, 3.5668E+09,
     +  3.5935E+09, 3.6204E+09, 3.6474E+09, 3.6746E+09, 3.7020E+09,
     + 3.7295E+09, 3.7572E+09, 3.7851E+09, 3.8132E+09, 3.8415E+09,
     +  3.8700E+09, 3.8986E+09, 3.9274E+09, 3.9564E+09, 3.9856E+09,
     + 4.0150E+09, 4.0446E+09, 4.0743E+09, 4.1043E+09, 4.1344E+09,
     +  4.1647E+09, 4.1953E+09, 4.2260E+09, 4.2569E+09, 4.2880E+09,
     + 4.3193E+09, 4.3508E+09, 4.3826E+09, 4.4145E+09, 4.4466E+09,
     +  4.4789E+09, 4.5114E+09, 4.5441E+09, 4.5771E+09, 4.6102E+09,
     + 4.6436E+09, 4.6771E+09, 4.7109E+09, 4.7449E+09, 4.7791E+09,
     +  4.8135E+09, 4.8481E+09, 4.8829E+09, 4.9180E+09, 4.9533E+09,
     + 4.9888E+09, 5.0245E+09, 5.0604E+09, 5.0966E+09, 5.1330E+09,
     +  5.1696E+09, 5.2065E+09, 5.2435E+09, 5.2808E+09, 5.3184E+09,
     + 5.3561E+09, 5.3941E+09, 5.4324E+09, 5.4708E+09, 5.5095E+09,
     +  5.5485E+09, 5.5877E+09, 5.6271E+09, 5.6667E+09, 5.7066E+09,
     + 5.7468E+09, 5.7872E+09, 5.8278E+09, 5.8687E+09, 5.9099E+09,
     +  5.9513E+09, 5.9929E+09, 6.0348E+09, 6.0770E+09, 6.1194E+09,
     + 6.1620E+09, 6.2050E+09, 6.2482E+09, 6.2916E+09, 6.3353E+09,
     +  6.3793E+09, 6.4235E+09, 6.4680E+09, 6.5128E+09, 6.5579E+09,
     + 6.6032E+09, 6.6488E+09, 6.6946E+09, 6.7408E+09, 6.7872E+09,
     +  6.8339E+09, 6.8808E+09, 6.9281E+09, 6.9756E+09, 7.0234E+09,
     + 7.0715E+09, 7.1199E+09, 7.1686E+09, 7.2176E+09, 7.2668E+09,
     +  7.3164E+09, 7.3662E+09, 7.4164E+09, 7.4668E+09, 7.5175E+09,
     + 7.5686E+09, 7.6199E+09, 7.6715E+09, 7.7235E+09, 7.7757E+09,
     +  7.8283E+09, 7.8811E+09, 7.9343E+09, 7.9878E+09, 8.0416E+09,
     + 8.0957E+09, 8.1501E+09, 8.2049E+09, 8.2600E+09, 8.3154E+09,
     +  8.3711E+09, 8.4271E+09, 8.4835E+09, 8.5402E+09, 8.5972E+09,
     + 8.6545E+09, 8.7122E+09, 8.7702E+09, 8.8286E+09, 8.8873E+09,
     +  8.9463E+09, 9.0057E+09, 9.0654E+09, 9.1255E+09, 9.1859E+09,
     + 9.2466E+09, 9.3077E+09, 9.3692E+09, 9.4310E+09, 9.4932E+09,
     +  9.5557E+09, 9.6186E+09, 9.6818E+09, 9.7454E+09, 9.8094E+09,
     + 9.8737E+09, 9.9384E+09, 1.0003E+10, 1.0069E+10, 1.0135E+10,
     +  1.0201E+10, 1.0267E+10, 1.0334E+10, 1.0402E+10, 1.0469E+10,
     + 1.0537E+10, 1.0606E+10, 1.0675E+10, 1.0744E+10, 1.0814E+10,
     +  1.0884E+10, 1.0954E+10, 1.1025E+10, 1.1096E+10, 1.1168E+10,
     + 1.1240E+10, 1.1312E+10, 1.1385E+10, 1.1458E+10, 1.1532E+10,
     +  1.1606E+10, 1.1680E+10, 1.1755E+10, 1.1830E+10, 1.1906E+10,
     + 1.1982E+10, 1.2059E+10, 1.2136E+10, 1.2213E+10, 1.2291E+10,
     +  1.2369E+10, 1.2448E+10, 1.2527E+10, 1.2607E+10, 1.2687E+10,
     + 1.2767E+10, 1.2848E+10, 1.2929E+10, 1.3011E+10, 1.3094E+10,
     +  1.3176E+10, 1.3259E+10, 1.3343E+10, 1.3427E+10, 1.3512E+10,
     + 1.3597E+10, 1.3682E+10, 1.3768E+10, 1.3854E+10, 1.3941E+10,
     +  1.4028E+10, 1.4116E+10, 1.4205E+10, 1.4293E+10, 1.4383E+10,
     + 1.4472E+10, 1.4563E+10, 1.4653E+10, 1.4744E+10, 1.4836E+10,
     +  1.4928E+10, 1.5021E+10, 1.5114E+10, 1.5208E+10, 1.5302E+10,
     + 1.5397E+10, 1.5492E+10, 1.5588E+10, 1.5684E+10, 1.5781E+10,
     +  1.5878E+10, 1.5976E+10, 1.6074E+10, 1.6173E+10, 1.6272E+10,
     + 1.6372E+10, 1.6473E+10, 1.6574E+10, 1.6675E+10, 1.6777E+10,
     +  1.6880E+10, 1.6983E+10, 1.7087E+10, 1.7191E+10, 1.7296E+10,
     + 1.7401E+10, 1.7507E+10, 1.7613E+10, 1.7721E+10, 1.7828E+10,
     +  1.7936E+10, 1.8045E+10, 1.8154E+10, 1.8264E+10, 1.8375E+10,
     + 1.8486E+10, 1.8597E+10, 1.8710E+10, 1.8822E+10, 1.8936E+10,
     +  1.9050E+10, 1.9164E+10, 1.9279E+10, 1.9395E+10, 1.9512E+10,
     + 1.9629E+10, 1.9746E+10, 1.9864E+10, 1.9983E+10, 2.0103E+10,
     +  2.0223E+10, 2.0344E+10, 2.0465E+10, 2.0587E+10, 2.0709E+10,
     + 2.0833E+10, 2.0956E+10, 2.1081E+10, 2.1206E+10, 2.1332E+10,
     +  2.1458E+10, 2.1585E+10, 2.1713E+10, 2.1842E+10, 2.1971E+10,
     + 2.2100E+10, 2.2231E+10, 2.2362E+10, 2.2493E+10, 2.2626E+10,
     +  2.2759E+10, 2.2893E+10, 2.3027E+10, 2.3162E+10, 2.3298E+10,
     + 2.3435E+10, 2.3572E+10, 2.3710E+10, 2.3848E+10, 2.3988E+10,
     +  2.4128E+10, 2.4268E+10, 2.4410E+10, 2.4552E+10, 2.4695E+10,
     + 2.4838E+10, 2.4983E+10, 2.5128E+10, 2.5274E+10, 2.5420E+10,
     +  2.5567E+10, 2.5715E+10, 2.5864E+10, 2.6014E+10, 2.6164E+10,
     + 2.6315E+10, 2.6467E+10, 2.6619E+10, 2.6772E+10, 2.6926E+10,
     +  2.7081E+10, 2.7237E+10, 2.7393E+10, 2.7550E+10, 2.7708E+10,
     + 2.7867E+10, 2.8027E+10, 2.8187E+10, 2.8348E+10, 2.8510E+10,
     +  2.8673E+10, 2.8836E+10, 2.9000E+10, 2.9166E+10, 2.9332E+10,
     + 2.9498E+10, 2.9666E+10, 2.9834E+10, 3.0004E+10, 3.0174E+10,
     +  3.0345E+10, 3.0516E+10, 3.0689E+10, 3.0863E+10, 3.1037E+10,
     + 3.1212E+10, 3.1388E+10, 3.1565E+10, 3.1743E+10, 3.1921E+10,
     +  3.2101E+10, 3.2281E+10, 3.2463E+10, 3.2645E+10, 3.2828E+10,
     + 3.3012E+10, 3.3197E+10, 3.3382E+10, 3.3569E+10, 3.3756E+10,
     +  3.3945E+10, 3.4134E+10, 3.4325E+10, 3.4516E+10, 3.4708E+10,
     + 3.4901E+10, 3.5095E+10, 3.5290E+10, 3.5486E+10, 3.5683E+10,
     +  3.5880E+10, 3.6079E+10, 3.6279E+10, 3.6479E+10, 3.6681E+10,
     + 3.6884E+10, 3.7087E+10, 3.7292E+10, 3.7497E+10, 3.7704E+10,
     +  3.7911E+10, 3.8120E+10, 3.8329E+10, 3.8540E+10, 3.8751E+10,
     + 3.8964E+10, 3.9177E+10, 3.9392E+10, 3.9607E+10, 3.9824E+10,
     +  4.0041E+10, 4.0260E+10, 4.0480E+10, 4.0700E+10, 4.0922E+10,
     + 4.1145E+10, 4.1369E+10, 4.1594E+10, 4.1820E+10, 4.2047E+10,
     +  4.2275E+10, 4.2504E+10, 4.2734E+10, 4.2966E+10, 4.3198E+10,
     + 4.3432E+10, 4.3666E+10, 4.3902E+10, 4.4139E+10, 4.4377E+10,
     +  4.4616E+10, 4.4856E+10, 4.5097E+10, 4.5340E+10, 4.5583E+10,
     + 4.5828E+10, 4.6074E+10, 4.6321E+10, 4.6569E+10, 4.6818E+10,
     +  4.7069E+10, 4.7320E+10, 4.7573E+10, 4.7827E+10, 4.8082E+10,
     + 4.8339E+10, 4.8596E+10, 4.8855E+10, 4.9115E+10, 4.9376E+10,
     +  4.9638E+10, 4.9902E+10, 5.0167E+10, 5.0433E+10, 5.0700E+10,
     + 5.0968E+10, 5.1238E+10, 5.1509E+10, 5.1781E+10, 5.2054E+10,
     +  5.2329E+10, 5.2605E+10, 5.2882E+10, 5.3160E+10, 5.3440E+10,
     + 5.3721E+10, 5.4003E+10, 5.4287E+10, 5.4572E+10, 5.4858E+10,
     +  5.5146E+10, 5.5434E+10, 5.5724E+10, 5.6016E+10, 5.6309E+10,
     + 5.6603E+10, 5.6898E+10, 5.7195E+10, 5.7493E+10, 5.7792E+10,
     +  5.8093E+10, 5.8395E+10, 5.8699E+10, 5.9004E+10, 5.9310E+10,
     + 5.9618E+10, 5.9927E+10, 6.0237E+10, 6.0549E+10, 6.0862E+10,
     +  6.1177E+10, 6.1493E+10, 6.1810E+10, 6.2129E+10, 6.2450E+10,
     + 6.2771E+10, 6.3095E+10, 6.3419E+10, 6.3745E+10, 6.4073E+10,
     +  6.4402E+10, 6.4733E+10, 6.5065E+10, 6.5398E+10, 6.5733E+10,
     + 6.6069E+10, 6.6407E+10, 6.6747E+10, 6.7088E+10, 6.7430E+10,
     +  6.7774E+10, 6.8120E+10, 6.8467E+10, 6.8815E+10, 6.9165E+10,
     + 6.9517E+10, 6.9870E+10, 7.0225E+10, 7.0581E+10, 7.0939E+10,
     +  7.1299E+10, 7.1660E+10, 7.2022E+10, 7.2387E+10, 7.2752E+10,
     + 7.3120E+10, 7.3489E+10, 7.3860E+10, 7.4232E+10, 7.4606E+10,
     +  7.4982E+10, 7.5359E+10, 7.5738E+10, 7.6118E+10, 7.6500E+10,
     + 7.6884E+10, 7.7270E+10, 7.7657E+10, 7.8046E+10, 7.8437E+10,
     +  7.8829E+10, 7.9223E+10, 7.9619E+10, 8.0016E+10, 8.0415E+10,
     + 8.0816E+10, 8.1219E+10, 8.1623E+10, 8.2029E+10, 8.2437E+10,
     +  8.2847E+10, 8.3258E+10, 8.3671E+10, 8.4086E+10, 8.4503E+10,
     + 8.4922E+10, 8.5342E+10, 8.5764E+10, 8.6188E+10, 8.6614E+10,
     +  8.7042E+10, 8.7471E+10, 8.7903E+10, 8.8336E+10, 8.8771E+10,
     + 8.9208E+10, 8.9646E+10, 9.0087E+10, 9.0530E+10, 9.0974E+10,
     +  9.1420E+10, 9.1868E+10, 9.2319E+10, 9.2771E+10, 9.3225E+10,
     + 9.3680E+10, 9.4138E+10, 9.4598E+10, 9.5060E+10, 9.5523E+10,
     +  9.5989E+10, 9.6457E+10, 9.6926E+10, 9.7398E+10, 9.7871E+10,
     + 9.8347E+10, 9.8825E+10, 9.9304E+10, 9.9786E+10, 1.0027E+11,
     +  1.0076E+11, 1.0124E+11, 1.0173E+11, 1.0222E+11, 1.0272E+11,
     + 1.0321E+11, 1.0371E+11, 1.0421E+11, 1.0471E+11, 1.0522E+11,
     +  1.0573E+11, 1.0623E+11, 1.0675E+11, 1.0726E+11, 1.0777E+11,
     + 1.0829E+11, 1.0881E+11, 1.0933E+11, 1.0985E+11, 1.1038E+11,
     +  1.1091E+11, 1.1144E+11, 1.1197E+11, 1.1251E+11, 1.1304E+11,
     + 1.1358E+11, 1.1412E+11, 1.1467E+11, 1.1521E+11, 1.1576E+11,
     +  1.1631E+11, 1.1686E+11, 1.1742E+11, 1.1798E+11, 1.1853E+11,
     + 1.1910E+11, 1.1966E+11, 1.2023E+11, 1.2080E+11, 1.2137E+11,
     +  1.2194E+11, 1.2252E+11, 1.2309E+11, 1.2367E+11, 1.2426E+11,
     + 1.2484E+11, 1.2543E+11, 1.2602E+11, 1.2661E+11, 1.2721E+11,
     +  1.2781E+11, 1.2841E+11, 1.2901E+11, 1.2961E+11, 1.3022E+11,
     + 1.3083E+11, 1.3144E+11, 1.3206E+11, 1.3267E+11, 1.3329E+11,
     +  1.3391E+11, 1.3454E+11, 1.3517E+11, 1.3580E+11, 1.3643E+11,
     + 1.3706E+11, 1.3770E+11, 1.3834E+11, 1.3898E+11, 1.3963E+11,
     +  1.4028E+11, 1.4093E+11, 1.4158E+11, 1.4223E+11, 1.4289E+11,
     + 1.4355E+11, 1.4422E+11, 1.4488E+11, 1.4555E+11, 1.4622E+11,
     +  1.4690E+11, 1.4758E+11, 1.4826E+11, 1.4894E+11, 1.4962E+11,
     + 1.5031E+11, 1.5100E+11, 1.5169E+11, 1.5239E+11, 1.5309E+11,
     +  1.5379E+11, 1.5450E+11, 1.5520E+11, 1.5591E+11, 1.5663E+11,
     + 1.5734E+11, 1.5806E+11, 1.5878E+11, 1.5951E+11, 1.6023E+11,
     +  1.6096E+11, 1.6170E+11, 1.6243E+11, 1.6317E+11, 1.6391E+11,
     + 1.6466E+11, 1.6541E+11, 1.6616E+11, 1.6691E+11, 1.6767E+11,
     +  1.6843E+11, 1.6919E+11, 1.6995E+11, 1.7072E+11, 1.7149E+11,
     + 1.7227E+11, 1.7305E+11, 1.7383E+11, 1.7461E+11, 1.7540E+11,
     +  1.7619E+11, 1.7698E+11, 1.7778E+11, 1.7858E+11, 1.7938E+11,
     + 1.8018E+11, 1.8099E+11, 1.8180E+11, 1.8262E+11, 1.8344E+11,
     +  1.8426E+11, 1.8508E+11, 1.8591E+11, 1.8674E+11, 1.8757E+11,
     + 1.8841E+11, 1.8925E+11, 1.9010E+11, 1.9094E+11, 1.9179E+11,
     +  1.9265E+11, 1.9350E+11, 1.9436E+11, 1.9523E+11, 1.9609E+11,
     + 1.9696E+11, 1.9784E+11, 1.9872E+11, 1.9960E+11, 2.0048E+11,
     +  2.0137E+11, 2.0226E+11, 2.0315E+11, 2.0405E+11, 2.0495E+11,
     + 2.0585E+11, 2.0676E+11, 2.0767E+11, 2.0859E+11, 2.0951E+11,
     +  2.1043E+11, 2.1135E+11, 2.1228E+11, 2.1322E+11, 2.1415E+11,
     + 2.1509E+11, 2.1603E+11, 2.1698E+11, 2.1793E+11, 2.1888E+11,
     +  2.1984E+11, 2.2080E+11, 2.2177E+11, 2.2274E+11, 2.2371E+11,
     + 2.2469E+11, 2.2566E+11, 2.2665E+11, 2.2764E+11, 2.2863E+11,
     +  2.2962E+11, 2.3062E+11, 2.3162E+11, 2.3263E+11, 2.3364E+11,
     + 2.3465E+11, 2.3567E+11, 2.3669E+11, 2.3771E+11, 2.3874E+11,
     +  2.3978E+11, 2.4081E+11, 2.4185E+11, 2.4290E+11, 2.4394E+11,
     + 2.4500E+11, 2.4605E+11, 2.4711E+11, 2.4818E+11, 2.4925E+11,
     +  2.5032E+11, 2.5139E+11, 2.5247E+11, 2.5356E+11, 2.5465E+11,
     + 2.5574E+11, 2.5684E+11, 2.5794E+11, 2.5904E+11, 2.6015E+11,
     +  2.6126E+11, 2.6238E+11, 2.6350E+11, 2.6462E+11, 2.6575E+11,
     + 2.6689E+11, 2.6803E+11, 2.6917E+11, 2.7031E+11, 2.7146E+11,
     +  2.7262E+11, 2.7378E+11, 2.7494E+11, 2.7611E+11, 2.7728E+11,
     + 2.7846E+11, 2.7964E+11, 2.8082E+11, 2.8201E+11, 2.8321E+11,
     +  2.8440E+11, 2.8561E+11, 2.8681E+11, 2.8802E+11, 2.8924E+11,
     + 2.9046E+11, 2.9168E+11, 2.9291E+11, 2.9415E+11, 2.9539E+11,
     +  2.9663E+11, 2.9788E+11, 2.9913E+11, 3.0038E+11, 3.0165E+11,
     + 3.0291E+11, 3.0418E+11, 3.0546E+11, 3.0674E+11, 3.0802E+11,
     +  3.0931E+11, 3.1060E+11, 3.1190E+11, 3.1320E+11, 3.1451E+11,
     + 3.1582E+11, 3.1714E+11, 3.1846E+11, 3.1979E+11, 3.2112E+11,
     +  3.2246E+11, 3.2380E+11, 3.2514E+11, 3.2650E+11, 3.2785E+11,
     + 3.2921E+11, 3.3058E+11, 3.3195E+11, 3.3332E+11, 3.3470E+11,
     +  3.3609E+11, 3.3748E+11, 3.3887E+11, 3.4028E+11, 3.4168E+11,
     + 3.4309E+11, 3.4451E+11, 3.4593E+11, 3.4735E+11, 3.4878E+11,
     +  3.5022E+11, 3.5166E+11, 3.5311E+11, 3.5456E+11, 3.5602E+11,
     + 3.5748E+11, 3.5894E+11, 3.6042E+11, 3.6189E+11, 3.6338E+11,
     +  3.6486E+11, 3.6636E+11, 3.6786E+11, 3.6936E+11, 3.7087E+11,
     + 3.7238E+11, 3.7390E+11, 3.7543E+11, 3.7696E+11, 3.7850E+11,
     +  3.8004E+11, 3.8159E+11, 3.8314E+11, 3.8470E+11, 3.8626E+11,
     + 3.8783E+11, 3.8940E+11, 3.9098E+11, 3.9257E+11, 3.9416E+11,
     +  3.9576E+11, 3.9736E+11, 3.9897E+11, 4.0058E+11, 4.0220E+11,
     + 4.0383E+11, 4.0546E+11, 4.0710E+11, 4.0874E+11, 4.1039E+11,
     +  4.1204E+11, 4.1370E+11, 4.1537E+11, 4.1704E+11, 4.1872E+11,
     + 4.2040E+11, 4.2209E+11, 4.2378E+11, 4.2549E+11, 4.2719E+11,
     +  4.2890E+11, 4.3062E+11, 4.3235E+11, 4.3408E+11, 4.3582E+11,
     + 4.3756E+11, 4.3931E+11, 4.4106E+11, 4.4282E+11, 4.4459E+11,
     +  4.4636E+11, 4.4814E+11, 4.4993E+11, 4.5172E+11, 4.5352E+11,
     + 4.5532E+11, 4.5713E+11, 4.5895E+11, 4.6077E+11, 4.6260E+11,
     +  4.6444E+11, 4.6628E+11, 4.6813E+11, 4.6998E+11, 4.7185E+11,
     + 4.7371E+11, 4.7559E+11, 4.7747E+11, 4.7935E+11, 4.8125E+11,
     +  4.8315E+11, 4.8505E+11, 4.8697E+11, 4.8889E+11, 4.9081E+11,
     + 4.9275E+11, 4.9468E+11, 4.9663E+11, 4.9858E+11, 5.0054E+11,
     +  5.0251E+11, 5.0448E+11, 5.0646E+11, 5.0845E+11, 5.1044E+11,
     + 5.1244E+11, 5.1445E+11, 5.1646E+11, 5.1848E+11, 5.2051E+11,
     +  5.2254E+11, 5.2458E+11, 5.2663E+11, 5.2868E+11, 5.3074E+11,
     + 5.3281E+11, 5.3489E+11, 5.3697E+11, 5.3906E+11, 5.4116E+11,
     +  5.4326E+11, 5.4537E+11, 5.4749E+11, 5.4962E+11, 5.5175E+11,
     + 5.5389E+11, 5.5604E+11, 5.5819E+11, 5.6035E+11, 5.6252E+11,
     +  5.6470E+11, 5.6688E+11, 5.6907E+11, 5.7127E+11, 5.7347E+11,
     + 5.7569E+11, 5.7791E+11, 5.8013E+11, 5.8237E+11, 5.8461E+11,
     +  5.8686E+11, 5.8912E+11, 5.9138E+11, 5.9366E+11, 5.9594E+11,
     + 5.9823E+11, 6.0052E+11, 6.0282E+11, 6.0514E+11, 6.0745E+11,
     +  6.0978E+11, 6.1211E+11, 6.1446E+11, 6.1681E+11, 6.1916E+11,
     + 6.2153E+11, 6.2390E+11, 6.2628E+11, 6.2867E+11, 6.3107E+11,
     +  6.3347E+11, 6.3589E+11, 6.3831E+11, 6.4074E+11, 6.4317E+11,
     + 6.4562E+11, 6.4807E+11, 6.5053E+11, 6.5300E+11, 6.5548E+11,
     +  6.5796E+11, 6.6046E+11, 6.6296E+11, 6.6547E+11, 6.6799E+11,
     + 6.7051E+11, 6.7305E+11, 6.7559E+11, 6.7814E+11, 6.8070E+11,
     +  6.8327E+11, 6.8585E+11, 6.8844E+11, 6.9103E+11, 6.9363E+11,
     + 6.9624E+11, 6.9886E+11, 7.0149E+11, 7.0413E+11, 7.0677E+11,
     +  7.0943E+11, 7.1209E+11, 7.1476E+11, 7.1744E+11, 7.2013E+11,
     + 7.2283E+11, 7.2553E+11, 7.2825E+11, 7.3097E+11, 7.3370E+11,
     +  7.3644E+11, 7.3919E+11, 7.4195E+11, 7.4472E+11, 7.4750E+11,
     + 7.5029E+11, 7.5308E+11, 7.5589E+11, 7.5870E+11, 7.6152E+11,
     +  7.6435E+11, 7.6719E+11, 7.7004E+11, 7.7290E+11, 7.7577E+11,
     + 7.7865E+11, 7.8154E+11, 7.8443E+11, 7.8734E+11, 7.9025E+11,
     +  7.9318E+11, 7.9611E+11, 7.9906E+11, 8.0201E+11, 8.0497E+11,
     + 8.0794E+11, 8.1092E+11, 8.1391E+11, 8.1692E+11, 8.1993E+11,
     +  8.2294E+11, 8.2597E+11, 8.2901E+11, 8.3206E+11, 8.3512E+11,
     + 8.3819E+11, 8.4127E+11, 8.4435E+11, 8.4745E+11, 8.5056E+11,
     +  8.5368E+11, 8.5681E+11, 8.5994E+11, 8.6309E+11, 8.6625E+11,
     + 8.6942E+11, 8.7259E+11, 8.7578E+11, 8.7898E+11, 8.8219E+11,
     +  8.8540E+11, 8.8863E+11, 8.9187E+11, 8.9512E+11, 8.9838E+11,
     + 9.0165E+11, 9.0493E+11, 9.0822E+11, 9.1152E+11, 9.1483E+11,
     +  9.1815E+11, 9.2148E+11, 9.2483E+11, 9.2818E+11, 9.3154E+11,
     + 9.3492E+11, 9.3830E+11, 9.4170E+11, 9.4510E+11, 9.4852E+11,
     +  9.5195E+11, 9.5539E+11, 9.5884E+11, 9.6230E+11, 9.6577E+11,
     + 9.6925E+11, 9.7274E+11, 9.7624E+11, 9.7976E+11, 9.8329E+11,
     +  9.8682E+11, 9.9037E+11, 9.9393E+11, 9.9750E+11, 1.0011E+12,
     + 1.0047E+12, 1.0083E+12, 1.0119E+12, 1.0155E+12, 1.0192E+12,
     +  1.0228E+12, 1.0265E+12, 1.0301E+12, 1.0338E+12, 1.0375E+12,
     + 1.0412E+12, 1.0449E+12, 1.0487E+12, 1.0524E+12, 1.0562E+12,
     +  1.0599E+12, 1.0637E+12, 1.0675E+12, 1.0713E+12, 1.0751E+12,
     + 1.0789E+12, 1.0827E+12, 1.0866E+12, 1.0904E+12, 1.0943E+12,
     +  1.0982E+12, 1.1021E+12, 1.1060E+12, 1.1099E+12, 1.1138E+12,
     + 1.1178E+12, 1.1217E+12, 1.1257E+12, 1.1297E+12, 1.1337E+12,
     +  1.1377E+12, 1.1417E+12, 1.1457E+12, 1.1498E+12, 1.1538E+12,
     + 1.1579E+12, 1.1620E+12, 1.1660E+12, 1.1701E+12, 1.1743E+12,
     +  1.1784E+12, 1.1825E+12, 1.1867E+12, 1.1908E+12, 1.1950E+12,
     + 1.1992E+12, 1.2034E+12, 1.2076E+12, 1.2119E+12, 1.2161E+12,
     +  1.2204E+12, 1.2246E+12, 1.2289E+12, 1.2332E+12, 1.2375E+12,
     + 1.2418E+12, 1.2462E+12, 1.2505E+12, 1.2549E+12, 1.2592E+12,
     +  1.2636E+12, 1.2680E+12, 1.2724E+12, 1.2769E+12, 1.2813E+12,
     + 1.2857E+12, 1.2902E+12, 1.2947E+12, 1.2992E+12, 1.3037E+12,
     +  1.3082E+12, 1.3127E+12, 1.3173E+12, 1.3218E+12, 1.3264E+12,
     + 1.3310E+12, 1.3356E+12, 1.3402E+12, 1.3449E+12, 1.3495E+12,
     +  1.3542E+12, 1.3588E+12, 1.3635E+12, 1.3682E+12, 1.3729E+12,
     + 1.3777E+12, 1.3824E+12, 1.3872E+12, 1.3919E+12, 1.3967E+12,
     +  1.4015E+12, 1.4063E+12, 1.4112E+12, 1.4160E+12, 1.4209E+12,
     + 1.4257E+12, 1.4306E+12, 1.4355E+12, 1.4404E+12, 1.4454E+12,
     +  1.4503E+12, 1.4553E+12, 1.4602E+12, 1.4652E+12, 1.4702E+12,
     + 1.4752E+12, 1.4803E+12, 1.4853E+12, 1.4904E+12, 1.4955E+12,
     +  1.5006E+12, 1.5057E+12, 1.5108E+12, 1.5159E+12, 1.5211E+12,
     + 1.5262E+12, 1.5314E+12, 1.5366E+12, 1.5418E+12, 1.5471E+12,
     +  1.5523E+12, 1.5576E+12, 1.5629E+12, 1.5682E+12, 1.5735E+12,
     + 1.5788E+12, 1.5841E+12, 1.5895E+12, 1.5948E+12, 1.6002E+12,
     +  1.6056E+12, 1.6111E+12, 1.6165E+12, 1.6219E+12, 1.6274E+12,
     + 1.6329E+12, 1.6384E+12, 1.6439E+12, 1.6494E+12, 1.6550E+12,
     +  1.6605E+12, 1.6661E+12, 1.6717E+12, 1.6773E+12, 1.6829E+12,
     + 1.6886E+12, 1.6942E+12, 1.6999E+12, 1.7056E+12, 1.7113E+12,
     +  1.7171E+12, 1.7228E+12, 1.7286E+12, 1.7343E+12, 1.7401E+12,
     + 1.7459E+12, 1.7518E+12, 1.7576E+12, 1.7635E+12, 1.7693E+12,
     +  1.7752E+12, 1.7812E+12, 1.7871E+12, 1.7930E+12, 1.7990E+12,
     + 1.8050E+12, 1.8110E+12, 1.8170E+12, 1.8230E+12, 1.8291E+12,
     +  1.8351E+12, 1.8412E+12, 1.8473E+12, 1.8535E+12, 1.8596E+12,
     + 1.8657E+12, 1.8719E+12, 1.8781E+12, 1.8843E+12, 1.8905E+12,
     +  1.8968E+12, 1.9031E+12, 1.9093E+12, 1.9156E+12, 1.9220E+12,
     + 1.9283E+12, 1.9346E+12, 1.9410E+12, 1.9474E+12, 1.9538E+12,
     +  1.9602E+12, 1.9667E+12, 1.9731E+12, 1.9796E+12, 1.9861E+12,
     + 1.9926E+12, 1.9992E+12, 2.0057E+12, 2.0123E+12, 2.0189E+12,
     +  2.0255E+12, 2.0321E+12, 2.0388E+12, 2.0455E+12, 2.0522E+12,
     + 2.0589E+12, 2.0656E+12, 2.0723E+12, 2.0791E+12, 2.0859E+12,
     +  2.0927E+12, 2.0995E+12, 2.1063E+12, 2.1132E+12, 2.1201E+12,
     + 2.1270E+12, 2.1339E+12, 2.1408E+12, 2.1478E+12, 2.1548E+12,
     +  2.1618E+12, 2.1688E+12, 2.1758E+12, 2.1829E+12, 2.1900E+12,
     + 2.1971E+12, 2.2042E+12, 2.2113E+12, 2.2185E+12, 2.2256E+12,
     +  2.2328E+12, 2.2401E+12, 2.2473E+12, 2.2546E+12, 2.2618E+12,
     + 2.2691E+12, 2.2765E+12, 2.2838E+12, 2.2912E+12, 2.2985E+12,
     +  2.3059E+12, 2.3134E+12, 2.3208E+12, 2.3283E+12, 2.3358E+12,
     + 2.3433E+12, 2.3508E+12, 2.3583E+12, 2.3659E+12, 2.3735E+12,
     +  2.3811E+12, 2.3887E+12, 2.3964E+12, 2.4041E+12, 2.4118E+12,
     + 2.4195E+12, 2.4272E+12, 2.4350E+12, 2.4428E+12, 2.4506E+12,
     +  2.4584E+12, 2.4662E+12, 2.4741E+12, 2.4820E+12, 2.4899E+12,
     + 2.4979E+12, 2.5058E+12, 2.5138E+12, 2.5218E+12, 2.5298E+12,
     +  2.5379E+12, 2.5459E+12, 2.5540E+12, 2.5621E+12, 2.5703E+12,
     + 2.5784E+12, 2.5866E+12, 2.5948E+12, 2.6030E+12, 2.6113E+12,
     +  2.6196E+12, 2.6279E+12, 2.6362E+12, 2.6445E+12, 2.6529E+12,
     + 2.6613E+12, 2.6697E+12, 2.6781E+12, 2.6865E+12, 2.6950E+12,
     +  2.7035E+12, 2.7121E+12, 2.7206E+12, 2.7292E+12, 2.7378E+12,
     + 2.7464E+12, 2.7550E+12, 2.7637E+12, 2.7724E+12, 2.7811E+12,
     +  2.7898E+12, 2.7986E+12, 2.8074E+12, 2.8162E+12, 2.8250E+12,
     + 2.8339E+12, 2.8428E+12, 2.8517E+12, 2.8606E+12, 2.8695E+12,
     +  2.8785E+12, 2.8875E+12, 2.8965E+12, 2.9056E+12, 2.9147E+12,
     + 2.9238E+12, 2.9329E+12, 2.9420E+12, 2.9512E+12, 2.9604E+12,
     +  2.9696E+12, 2.9789E+12, 2.9882E+12, 2.9975E+12, 3.0068E+12,
     + 3.0161E+12, 3.0255E+12, 3.0349E+12, 3.0444E+12, 3.0538E+12,
     +  3.0633E+12, 3.0728E+12, 3.0823E+12, 3.0919E+12, 3.1014E+12,
     + 3.1111E+12, 3.1207E+12, 3.1303E+12, 3.1400E+12, 3.1497E+12,
     +  3.1595E+12, 3.1692E+12, 3.1790E+12, 3.1888E+12, 3.1987E+12,
     + 3.2085E+12, 3.2184E+12, 3.2284E+12, 3.2383E+12, 3.2483E+12,
     +  3.2583E+12, 3.2683E+12, 3.2784E+12, 3.2884E+12, 3.2986E+12,
     + 3.3087E+12, 3.3188E+12, 3.3290E+12, 3.3393E+12, 3.3495E+12,
     +  3.3598E+12, 3.3701E+12, 3.3804E+12, 3.3907E+12, 3.4011E+12,
     + 3.4115E+12, 3.4220E+12, 3.4324E+12, 3.4429E+12, 3.4534E+12,
     +  3.4640E+12, 3.4746E+12, 3.4852E+12, 3.4958E+12, 3.5065E+12,
     + 3.5172E+12, 3.5279E+12, 3.5386E+12, 3.5494E+12, 3.5602E+12,
     +  3.5710E+12, 3.5819E+12, 3.5928E+12, 3.6037E+12, 3.6146E+12,
     + 3.6256E+12, 3.6366E+12, 3.6476E+12, 3.6587E+12, 3.6698E+12,
     +  3.6809E+12, 3.6921E+12, 3.7032E+12, 3.7145E+12, 3.7257E+12,
     + 3.7370E+12, 3.7483E+12, 3.7596E+12, 3.7709E+12, 3.7823E+12,
     +  3.7937E+12, 3.8052E+12, 3.8167E+12, 3.8282E+12, 3.8397E+12,
     + 3.8513E+12, 3.8629E+12, 3.8745E+12, 3.8862E+12, 3.8978E+12,
     +  3.9096E+12, 3.9213E+12, 3.9331E+12, 3.9449E+12, 3.9567E+12,
     + 3.9686E+12, 3.9805E+12, 3.9925E+12, 4.0044E+12, 4.0164E+12,
     +  4.0284E+12, 4.0405E+12, 4.0526E+12, 4.0647E+12, 4.0769E+12,
     + 4.0891E+12, 4.1013E+12, 4.1135E+12, 4.1258E+12, 4.1381E+12,
     +  4.1505E+12, 4.1628E+12, 4.1753E+12, 4.1877E+12, 4.2002E+12,
     + 4.2127E+12, 4.2252E+12, 4.2378E+12, 4.2504E+12, 4.2630E+12,
     +  4.2757E+12, 4.2884E+12, 4.3011E+12, 4.3139E+12, 4.3267E+12,
     + 4.3395E+12, 4.3524E+12, 4.3653E+12, 4.3782E+12, 4.3912E+12,
     +  4.4042E+12, 4.4172E+12, 4.4303E+12, 4.4434E+12, 4.4565E+12,
     + 4.4697E+12, 4.4829E+12, 4.4961E+12, 4.5094E+12, 4.5227E+12,
     +  4.5360E+12, 4.5494E+12, 4.5628E+12, 4.5763E+12, 4.5897E+12,
     + 4.6032E+12, 4.6168E+12, 4.6304E+12, 4.6440E+12, 4.6576E+12,
     +  4.6713E+12, 4.6850E+12, 4.6988E+12, 4.7126E+12, 4.7264E+12,
     + 4.7403E+12, 4.7541E+12, 4.7681E+12, 4.7820E+12, 4.7960E+12,
     +  4.8101E+12, 4.8242E+12, 4.8383E+12, 4.8524E+12, 4.8666E+12,
     + 4.8808E+12, 4.8950E+12, 4.9093E+12, 4.9237E+12, 4.9380E+12,
     +  4.9524E+12, 4.9669E+12, 4.9813E+12, 4.9958E+12, 5.0104E+12,
     + 5.0250E+12, 5.0396E+12, 5.0542E+12, 5.0689E+12, 5.0836E+12,
     +  5.0984E+12, 5.1132E+12, 5.1281E+12, 5.1429E+12, 5.1578E+12/
      data (QofT( 2,J),J=2409,4510)/ 5.1728E+12, 5.1878E+12,
     +  5.2028E+12, 5.2179E+12, 5.2330E+12, 5.2481E+12, 5.2633E+12,
     +  5.2785E+12,
     + 5.2938E+12, 5.3091E+12, 5.3244E+12, 5.3398E+12, 5.3552E+12,
     +  5.3706E+12, 5.3861E+12, 5.4017E+12, 5.4172E+12, 5.4328E+12,
     + 5.4485E+12, 5.4642E+12, 5.4799E+12, 5.4956E+12, 5.5115E+12,
     +  5.5273E+12, 5.5432E+12, 5.5591E+12, 5.5751E+12, 5.5911E+12,
     + 5.6071E+12, 5.6232E+12, 5.6393E+12, 5.6555E+12, 5.6717E+12,
     +  5.6879E+12, 5.7042E+12, 5.7205E+12, 5.7369E+12, 5.7533E+12,
     + 5.7697E+12, 5.7862E+12, 5.8027E+12, 5.8193E+12, 5.8359E+12,
     +  5.8525E+12, 5.8692E+12, 5.8860E+12, 5.9027E+12, 5.9196E+12,
     + 5.9364E+12, 5.9533E+12, 5.9703E+12, 5.9872E+12, 6.0043E+12,
     +  6.0213E+12, 6.0384E+12, 6.0556E+12, 6.0728E+12, 6.0900E+12,
     + 6.1073E+12, 6.1246E+12, 6.1420E+12, 6.1594E+12, 6.1768E+12,
     +  6.1943E+12, 6.2119E+12, 6.2294E+12, 6.2471E+12, 6.2647E+12,
     + 6.2824E+12, 6.3002E+12, 6.3180E+12, 6.3358E+12, 6.3537E+12,
     +  6.3716E+12, 6.3896E+12, 6.4076E+12, 6.4257E+12, 6.4438E+12,
     + 6.4619E+12, 6.4801E+12, 6.4984E+12, 6.5167E+12, 6.5350E+12,
     +  6.5534E+12, 6.5718E+12, 6.5902E+12, 6.6087E+12, 6.6273E+12,
     + 6.6459E+12, 6.6645E+12, 6.6832E+12, 6.7020E+12, 6.7207E+12,
     +  6.7396E+12, 6.7584E+12, 6.7774E+12, 6.7963E+12, 6.8153E+12,
     + 6.8344E+12, 6.8535E+12, 6.8726E+12, 6.8918E+12, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_SO    (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 3),QofT( 3,NT),Q(NT)
      data (Tmx(I),I=1,3)/3500.,3500.,3500./
      data xgj/ 1.,1.,1./
c...       SO
c...        --        26
      data (QofT( 1,J),J=1,2408)/ 1.7696E+00, 3.0757E+00, 4.4215E+00,
     +  5.8771E+00, 7.4844E+00, 9.2481E+00, 1.1155E+01, 1.3188E+01,
     + 1.5327E+01, 1.7558E+01, 1.9865E+01, 2.2237E+01, 2.4666E+01,
     +  2.7142E+01, 2.9659E+01, 3.2212E+01, 3.4796E+01, 3.7408E+01,
     + 4.0043E+01, 4.2700E+01, 4.8068E+01, 5.3496E+01, 5.8974E+01,
     +  6.4492E+01, 7.0043E+01, 7.5622E+01, 8.1225E+01, 8.6849E+01,
     + 9.2489E+01, 9.8146E+01, 1.0382E+02, 1.0950E+02, 1.1519E+02,
     +  1.2089E+02, 1.2660E+02, 1.3231E+02, 1.3803E+02, 1.4376E+02,
     + 1.4950E+02, 1.5524E+02, 1.6098E+02, 1.6673E+02, 1.7248E+02,
     +  1.7823E+02, 1.8399E+02, 1.8975E+02, 1.9551E+02, 2.0128E+02,
     + 2.0704E+02, 2.1281E+02, 2.1859E+02, 2.2436E+02, 2.3014E+02,
     +  2.3591E+02, 2.4169E+02, 2.4747E+02, 2.5325E+02, 2.5904E+02,
     + 2.6482E+02, 2.7061E+02, 2.7639E+02, 2.8218E+02, 2.8797E+02,
     +  2.9376E+02, 2.9955E+02, 3.0534E+02, 3.1113E+02, 3.1693E+02,
     + 3.2272E+02, 3.2851E+02, 3.3431E+02, 3.4011E+02, 3.4590E+02,
     +  3.5170E+02, 3.5750E+02, 3.6329E+02, 3.6909E+02, 3.7490E+02,
     + 3.8070E+02, 3.8650E+02, 3.9230E+02, 3.9810E+02, 4.0390E+02,
     +  4.0970E+02, 4.1551E+02, 4.2131E+02, 4.2711E+02, 4.3292E+02,
     + 4.3872E+02, 4.4453E+02, 4.5034E+02, 4.5615E+02, 4.6195E+02,
     +  4.6776E+02, 4.7357E+02, 4.7938E+02, 4.8519E+02, 4.9100E+02,
     + 4.9682E+02, 5.0263E+02, 5.0844E+02, 5.1426E+02, 5.2007E+02,
     +  5.2589E+02, 5.3171E+02, 5.3753E+02, 5.4335E+02, 5.4917E+02,
     + 5.5499E+02, 5.6081E+02, 5.6664E+02, 5.7246E+02, 5.7829E+02,
     +  5.8412E+02, 5.8995E+02, 5.9578E+02, 6.0162E+02, 6.0745E+02,
     + 6.1329E+02, 6.1913E+02, 6.2497E+02, 6.3081E+02, 6.3665E+02,
     +  6.4250E+02, 6.4834E+02, 6.5419E+02, 6.6005E+02, 6.6590E+02,
     + 6.7176E+02, 6.7762E+02, 6.8348E+02, 6.8935E+02, 6.9522E+02,
     +  7.0109E+02, 7.0696E+02, 7.1284E+02, 7.1872E+02, 7.2460E+02,
     + 7.3048E+02, 7.3637E+02, 7.4227E+02, 7.4817E+02, 7.5406E+02,
     +  7.5996E+02, 7.6588E+02, 7.7178E+02, 7.7770E+02, 7.8362E+02,
     + 7.8954E+02, 7.9546E+02, 8.0140E+02, 8.0733E+02, 8.1327E+02,
     +  8.1921E+02, 8.2516E+02, 8.3112E+02, 8.3707E+02, 8.4304E+02,
     + 8.4900E+02, 8.5498E+02, 8.6096E+02, 8.6694E+02, 8.7292E+02,
     +  8.7892E+02, 8.8492E+02, 8.9092E+02, 8.9693E+02, 9.0295E+02,
     + 9.0897E+02, 9.1500E+02, 9.2102E+02, 9.2706E+02, 9.3311E+02,
     +  9.3915E+02, 9.4522E+02, 9.5128E+02, 9.5735E+02, 9.6342E+02,
     + 9.6951E+02, 9.7559E+02, 9.8169E+02, 9.8780E+02, 9.9390E+02,
     +  1.0000E+03, 1.0061E+03, 1.0123E+03, 1.0184E+03, 1.0245E+03,
     + 1.0307E+03, 1.0369E+03, 1.0430E+03, 1.0492E+03, 1.0554E+03,
     +  1.0616E+03, 1.0678E+03, 1.0740E+03, 1.0802E+03, 1.0864E+03,
     + 1.0926E+03, 1.0989E+03, 1.1051E+03, 1.1114E+03, 1.1176E+03,
     +  1.1239E+03, 1.1302E+03, 1.1365E+03, 1.1427E+03, 1.1491E+03,
     + 1.1554E+03, 1.1617E+03, 1.1680E+03, 1.1743E+03, 1.1807E+03,
     +  1.1870E+03, 1.1934E+03, 1.1998E+03, 1.2062E+03, 1.2125E+03,
     + 1.2190E+03, 1.2254E+03, 1.2318E+03, 1.2382E+03, 1.2446E+03,
     +  1.2511E+03, 1.2576E+03, 1.2640E+03, 1.2705E+03, 1.2770E+03,
     + 1.2835E+03, 1.2900E+03, 1.2965E+03, 1.3030E+03, 1.3095E+03,
     +  1.3161E+03, 1.3226E+03, 1.3292E+03, 1.3358E+03, 1.3424E+03,
     + 1.3490E+03, 1.3556E+03, 1.3622E+03, 1.3688E+03, 1.3754E+03,
     +  1.3821E+03, 1.3887E+03, 1.3954E+03, 1.4021E+03, 1.4088E+03,
     + 1.4155E+03, 1.4222E+03, 1.4289E+03, 1.4356E+03, 1.4424E+03,
     +  1.4491E+03, 1.4559E+03, 1.4626E+03, 1.4694E+03, 1.4762E+03,
     + 1.4830E+03, 1.4899E+03, 1.4967E+03, 1.5035E+03, 1.5104E+03,
     +  1.5172E+03, 1.5241E+03, 1.5310E+03, 1.5379E+03, 1.5448E+03,
     + 1.5517E+03, 1.5586E+03, 1.5656E+03, 1.5725E+03, 1.5795E+03,
     +  1.5864E+03, 1.5934E+03, 1.6004E+03, 1.6074E+03, 1.6145E+03,
     + 1.6215E+03, 1.6285E+03, 1.6356E+03, 1.6427E+03, 1.6497E+03,
     +  1.6568E+03, 1.6639E+03, 1.6710E+03, 1.6782E+03, 1.6853E+03,
     + 1.6925E+03, 1.6996E+03, 1.7068E+03, 1.7140E+03, 1.7212E+03,
     +  1.7284E+03, 1.7356E+03, 1.7428E+03, 1.7501E+03, 1.7573E+03,
     + 1.7646E+03, 1.7719E+03, 1.7792E+03, 1.7865E+03, 1.7938E+03,
     +  1.8011E+03, 1.8085E+03, 1.8158E+03, 1.8232E+03, 1.8306E+03,
     + 1.8379E+03, 1.8454E+03, 1.8528E+03, 1.8602E+03, 1.8676E+03,
     +  1.8751E+03, 1.8826E+03, 1.8900E+03, 1.8975E+03, 1.9050E+03,
     + 1.9125E+03, 1.9201E+03, 1.9276E+03, 1.9352E+03, 1.9427E+03,
     +  1.9503E+03, 1.9579E+03, 1.9655E+03, 1.9731E+03, 1.9807E+03,
     + 1.9884E+03, 1.9960E+03, 2.0037E+03, 2.0114E+03, 2.0191E+03,
     +  2.0268E+03, 2.0345E+03, 2.0422E+03, 2.0500E+03, 2.0577E+03,
     + 2.0655E+03, 2.0733E+03, 2.0811E+03, 2.0889E+03, 2.0967E+03,
     +  2.1045E+03, 2.1124E+03, 2.1202E+03, 2.1281E+03, 2.1360E+03,
     + 2.1439E+03, 2.1518E+03, 2.1597E+03, 2.1676E+03, 2.1756E+03,
     +  2.1836E+03, 2.1915E+03, 2.1995E+03, 2.2075E+03, 2.2155E+03,
     + 2.2236E+03, 2.2316E+03, 2.2397E+03, 2.2477E+03, 2.2558E+03,
     +  2.2639E+03, 2.2720E+03, 2.2801E+03, 2.2883E+03, 2.2964E+03,
     + 2.3046E+03, 2.3127E+03, 2.3209E+03, 2.3291E+03, 2.3373E+03,
     +  2.3456E+03, 2.3538E+03, 2.3620E+03, 2.3703E+03, 2.3786E+03,
     + 2.3869E+03, 2.3952E+03, 2.4035E+03, 2.4118E+03, 2.4202E+03,
     +  2.4285E+03, 2.4369E+03, 2.4453E+03, 2.4537E+03, 2.4621E+03,
     + 2.4705E+03, 2.4790E+03, 2.4874E+03, 2.4959E+03, 2.5044E+03,
     +  2.5129E+03, 2.5214E+03, 2.5299E+03, 2.5384E+03, 2.5470E+03,
     + 2.5555E+03, 2.5641E+03, 2.5727E+03, 2.5813E+03, 2.5899E+03,
     +  2.5985E+03, 2.6072E+03, 2.6158E+03, 2.6245E+03, 2.6332E+03,
     + 2.6419E+03, 2.6506E+03, 2.6593E+03, 2.6680E+03, 2.6768E+03,
     +  2.6855E+03, 2.6943E+03, 2.7031E+03, 2.7119E+03, 2.7207E+03,
     + 2.7295E+03, 2.7384E+03, 2.7472E+03, 2.7561E+03, 2.7650E+03,
     +  2.7739E+03, 2.7828E+03, 2.7917E+03, 2.8007E+03, 2.8096E+03,
     + 2.8186E+03, 2.8276E+03, 2.8366E+03, 2.8456E+03, 2.8546E+03,
     +  2.8636E+03, 2.8727E+03, 2.8817E+03, 2.8908E+03, 2.8999E+03,
     + 2.9090E+03, 2.9181E+03, 2.9272E+03, 2.9364E+03, 2.9455E+03,
     +  2.9547E+03, 2.9639E+03, 2.9731E+03, 2.9823E+03, 2.9916E+03,
     + 3.0008E+03, 3.0100E+03, 3.0193E+03, 3.0286E+03, 3.0379E+03,
     +  3.0472E+03, 3.0565E+03, 3.0658E+03, 3.0752E+03, 3.0846E+03,
     + 3.0940E+03, 3.1033E+03, 3.1127E+03, 3.1222E+03, 3.1316E+03,
     +  3.1410E+03, 3.1505E+03, 3.1600E+03, 3.1695E+03, 3.1790E+03,
     + 3.1885E+03, 3.1980E+03, 3.2076E+03, 3.2171E+03, 3.2267E+03,
     +  3.2363E+03, 3.2459E+03, 3.2555E+03, 3.2651E+03, 3.2748E+03,
     + 3.2844E+03, 3.2941E+03, 3.3038E+03, 3.3135E+03, 3.3232E+03,
     +  3.3329E+03, 3.3426E+03, 3.3524E+03, 3.3622E+03, 3.3719E+03,
     + 3.3817E+03, 3.3915E+03, 3.4014E+03, 3.4112E+03, 3.4210E+03,
     +  3.4309E+03, 3.4408E+03, 3.4507E+03, 3.4606E+03, 3.4705E+03,
     + 3.4804E+03, 3.4904E+03, 3.5004E+03, 3.5103E+03, 3.5203E+03,
     +  3.5303E+03, 3.5404E+03, 3.5504E+03, 3.5604E+03, 3.5705E+03,
     + 3.5805E+03, 3.5906E+03, 3.6008E+03, 3.6109E+03, 3.6210E+03,
     +  3.6311E+03, 3.6413E+03, 3.6515E+03, 3.6616E+03, 3.6718E+03,
     + 3.6821E+03, 3.6923E+03, 3.7025E+03, 3.7128E+03, 3.7231E+03,
     +  3.7333E+03, 3.7436E+03, 3.7539E+03, 3.7643E+03, 3.7746E+03,
     + 3.7850E+03, 3.7953E+03, 3.8057E+03, 3.8161E+03, 3.8265E+03,
     +  3.8370E+03, 3.8474E+03, 3.8578E+03, 3.8683E+03, 3.8788E+03,
     + 3.8893E+03, 3.8998E+03, 3.9103E+03, 3.9209E+03, 3.9314E+03,
     +  3.9420E+03, 3.9526E+03, 3.9631E+03, 3.9738E+03, 3.9844E+03,
     + 3.9950E+03, 4.0057E+03, 4.0163E+03, 4.0270E+03, 4.0377E+03,
     +  4.0484E+03, 4.0591E+03, 4.0698E+03, 4.0806E+03, 4.0914E+03,
     + 4.1021E+03, 4.1129E+03, 4.1237E+03, 4.1345E+03, 4.1454E+03,
     +  4.1562E+03, 4.1671E+03, 4.1780E+03, 4.1889E+03, 4.1997E+03,
     + 4.2107E+03, 4.2216E+03, 4.2326E+03, 4.2435E+03, 4.2545E+03,
     +  4.2655E+03, 4.2765E+03, 4.2875E+03, 4.2985E+03, 4.3096E+03,
     + 4.3206E+03, 4.3317E+03, 4.3428E+03, 4.3539E+03, 4.3650E+03,
     +  4.3762E+03, 4.3873E+03, 4.3984E+03, 4.4096E+03, 4.4208E+03,
     + 4.4320E+03, 4.4432E+03, 4.4545E+03, 4.4657E+03, 4.4769E+03,
     +  4.4882E+03, 4.4995E+03, 4.5108E+03, 4.5221E+03, 4.5334E+03,
     + 4.5448E+03, 4.5561E+03, 4.5675E+03, 4.5789E+03, 4.5903E+03,
     +  4.6017E+03, 4.6131E+03, 4.6246E+03, 4.6360E+03, 4.6475E+03,
     + 4.6590E+03, 4.6705E+03, 4.6820E+03, 4.6935E+03, 4.7051E+03,
     +  4.7166E+03, 4.7282E+03, 4.7398E+03, 4.7514E+03, 4.7630E+03,
     + 4.7746E+03, 4.7862E+03, 4.7979E+03, 4.8096E+03, 4.8212E+03,
     +  4.8330E+03, 4.8447E+03, 4.8564E+03, 4.8681E+03, 4.8799E+03,
     + 4.8917E+03, 4.9034E+03, 4.9152E+03, 4.9271E+03, 4.9389E+03,
     +  4.9507E+03, 4.9626E+03, 4.9744E+03, 4.9863E+03, 4.9982E+03,
     + 5.0101E+03, 5.0221E+03, 5.0340E+03, 5.0459E+03, 5.0579E+03,
     +  5.0699E+03, 5.0819E+03, 5.0939E+03, 5.1059E+03, 5.1179E+03,
     + 5.1300E+03, 5.1421E+03, 5.1542E+03, 5.1662E+03, 5.1784E+03,
     +  5.1905E+03, 5.2026E+03, 5.2148E+03, 5.2269E+03, 5.2391E+03,
     + 5.2513E+03, 5.2636E+03, 5.2758E+03, 5.2880E+03, 5.3002E+03,
     +  5.3125E+03, 5.3248E+03, 5.3371E+03, 5.3494E+03, 5.3617E+03,
     + 5.3741E+03, 5.3864E+03, 5.3988E+03, 5.4112E+03, 5.4236E+03,
     +  5.4360E+03, 5.4484E+03, 5.4608E+03, 5.4733E+03, 5.4858E+03,
     + 5.4983E+03, 5.5108E+03, 5.5233E+03, 5.5358E+03, 5.5483E+03,
     +  5.5609E+03, 5.5734E+03, 5.5860E+03, 5.5986E+03, 5.6112E+03,
     + 5.6238E+03, 5.6365E+03, 5.6491E+03, 5.6618E+03, 5.6745E+03,
     +  5.6872E+03, 5.6999E+03, 5.7126E+03, 5.7253E+03, 5.7381E+03,
     + 5.7508E+03, 5.7636E+03, 5.7764E+03, 5.7893E+03, 5.8021E+03,
     +  5.8149E+03, 5.8277E+03, 5.8406E+03, 5.8535E+03, 5.8664E+03,
     + 5.8793E+03, 5.8923E+03, 5.9052E+03, 5.9181E+03, 5.9311E+03,
     +  5.9441E+03, 5.9571E+03, 5.9701E+03, 5.9831E+03, 5.9962E+03,
     + 6.0092E+03, 6.0223E+03, 6.0354E+03, 6.0484E+03, 6.0616E+03,
     +  6.0747E+03, 6.0878E+03, 6.1010E+03, 6.1141E+03, 6.1273E+03,
     + 6.1405E+03, 6.1538E+03, 6.1669E+03, 6.1802E+03, 6.1935E+03,
     +  6.2067E+03, 6.2200E+03, 6.2333E+03, 6.2466E+03, 6.2599E+03,
     + 6.2733E+03, 6.2866E+03, 6.3000E+03, 6.3134E+03, 6.3268E+03,
     +  6.3402E+03, 6.3536E+03, 6.3670E+03, 6.3805E+03, 6.3940E+03,
     + 6.4075E+03, 6.4209E+03, 6.4345E+03, 6.4480E+03, 6.4615E+03,
     +  6.4751E+03, 6.4887E+03, 6.5022E+03, 6.5158E+03, 6.5294E+03,
     + 6.5431E+03, 6.5567E+03, 6.5703E+03, 6.5840E+03, 6.5977E+03,
     +  6.6114E+03, 6.6251E+03, 6.6388E+03, 6.6526E+03, 6.6663E+03,
     + 6.6801E+03, 6.6939E+03, 6.7077E+03, 6.7215E+03, 6.7353E+03,
     +  6.7492E+03, 6.7630E+03, 6.7769E+03, 6.7908E+03, 6.8047E+03,
     + 6.8186E+03, 6.8325E+03, 6.8465E+03, 6.8604E+03, 6.8744E+03,
     +  6.8884E+03, 6.9024E+03, 6.9164E+03, 6.9304E+03, 6.9445E+03,
     + 6.9585E+03, 6.9726E+03, 6.9867E+03, 7.0008E+03, 7.0149E+03,
     +  7.0290E+03, 7.0432E+03, 7.0573E+03, 7.0715E+03, 7.0857E+03,
     + 7.0999E+03, 7.1141E+03, 7.1283E+03, 7.1426E+03, 7.1568E+03,
     +  7.1711E+03, 7.1854E+03, 7.1997E+03, 7.2140E+03, 7.2283E+03,
     + 7.2427E+03, 7.2570E+03, 7.2714E+03, 7.2858E+03, 7.3002E+03,
     +  7.3146E+03, 7.3290E+03, 7.3435E+03, 7.3580E+03, 7.3724E+03,
     + 7.3869E+03, 7.4014E+03, 7.4159E+03, 7.4305E+03, 7.4450E+03,
     +  7.4596E+03, 7.4741E+03, 7.4887E+03, 7.5033E+03, 7.5179E+03,
     + 7.5326E+03, 7.5472E+03, 7.5619E+03, 7.5766E+03, 7.5912E+03,
     +  7.6060E+03, 7.6207E+03, 7.6354E+03, 7.6502E+03, 7.6649E+03,
     + 7.6797E+03, 7.6945E+03, 7.7093E+03, 7.7241E+03, 7.7389E+03,
     +  7.7538E+03, 7.7686E+03, 7.7835E+03, 7.7984E+03, 7.8133E+03,
     + 7.8282E+03, 7.8432E+03, 7.8581E+03, 7.8731E+03, 7.8880E+03,
     +  7.9030E+03, 7.9180E+03, 7.9330E+03, 7.9481E+03, 7.9632E+03,
     + 7.9782E+03, 7.9933E+03, 8.0084E+03, 8.0235E+03, 8.0386E+03,
     +  8.0537E+03, 8.0689E+03, 8.0841E+03, 8.0992E+03, 8.1144E+03,
     + 8.1297E+03, 8.1449E+03, 8.1601E+03, 8.1754E+03, 8.1906E+03,
     +  8.2059E+03, 8.2212E+03, 8.2365E+03, 8.2518E+03, 8.2672E+03,
     + 8.2825E+03, 8.2979E+03, 8.3132E+03, 8.3287E+03, 8.3440E+03,
     +  8.3595E+03, 8.3749E+03, 8.3904E+03, 8.4058E+03, 8.4213E+03,
     + 8.4368E+03, 8.4523E+03, 8.4679E+03, 8.4834E+03, 8.4989E+03,
     +  8.5145E+03, 8.5301E+03, 8.5457E+03, 8.5613E+03, 8.5769E+03,
     + 8.5926E+03, 8.6082E+03, 8.6239E+03, 8.6396E+03, 8.6553E+03,
     +  8.6710E+03, 8.6867E+03, 8.7024E+03, 8.7182E+03, 8.7340E+03,
     + 8.7498E+03, 8.7655E+03, 8.7814E+03, 8.7972E+03, 8.8130E+03,
     +  8.8289E+03, 8.8448E+03, 8.8607E+03, 8.8766E+03, 8.8925E+03,
     + 8.9084E+03, 8.9244E+03, 8.9403E+03, 8.9563E+03, 8.9722E+03,
     +  8.9882E+03, 9.0043E+03, 9.0203E+03, 9.0363E+03, 9.0524E+03,
     + 9.0685E+03, 9.0845E+03, 9.1007E+03, 9.1168E+03, 9.1329E+03,
     +  9.1490E+03, 9.1652E+03, 9.1814E+03, 9.1976E+03, 9.2138E+03,
     + 9.2300E+03, 9.2462E+03, 9.2625E+03, 9.2787E+03, 9.2950E+03,
     +  9.3113E+03, 9.3276E+03, 9.3439E+03, 9.3602E+03, 9.3766E+03,
     + 9.3930E+03, 9.4093E+03, 9.4257E+03, 9.4421E+03, 9.4585E+03,
     +  9.4750E+03, 9.4914E+03, 9.5079E+03, 9.5243E+03, 9.5408E+03,
     + 9.5574E+03, 9.5739E+03, 9.5904E+03, 9.6070E+03, 9.6235E+03,
     +  9.6401E+03, 9.6567E+03, 9.6733E+03, 9.6899E+03, 9.7065E+03,
     + 9.7232E+03, 9.7399E+03, 9.7565E+03, 9.7732E+03, 9.7899E+03,
     +  9.8067E+03, 9.8234E+03, 9.8401E+03, 9.8569E+03, 9.8737E+03,
     + 9.8905E+03, 9.9073E+03, 9.9241E+03, 9.9410E+03, 9.9578E+03,
     +  9.9747E+03, 9.9915E+03, 1.0008E+04, 1.0025E+04, 1.0042E+04,
     + 1.0059E+04, 1.0076E+04, 1.0093E+04, 1.0110E+04, 1.0127E+04,
     +  1.0144E+04, 1.0161E+04, 1.0178E+04, 1.0195E+04, 1.0212E+04,
     + 1.0229E+04, 1.0247E+04, 1.0264E+04, 1.0281E+04, 1.0298E+04,
     +  1.0315E+04, 1.0332E+04, 1.0349E+04, 1.0367E+04, 1.0384E+04,
     + 1.0401E+04, 1.0418E+04, 1.0436E+04, 1.0453E+04, 1.0470E+04,
     +  1.0487E+04, 1.0505E+04, 1.0522E+04, 1.0539E+04, 1.0557E+04,
     + 1.0574E+04, 1.0592E+04, 1.0609E+04, 1.0626E+04, 1.0644E+04,
     +  1.0661E+04, 1.0679E+04, 1.0696E+04, 1.0714E+04, 1.0731E+04,
     + 1.0749E+04, 1.0766E+04, 1.0784E+04, 1.0801E+04, 1.0819E+04,
     +  1.0836E+04, 1.0854E+04, 1.0872E+04, 1.0889E+04, 1.0907E+04,
     + 1.0925E+04, 1.0942E+04, 1.0960E+04, 1.0978E+04, 1.0995E+04,
     +  1.1013E+04, 1.1031E+04, 1.1049E+04, 1.1066E+04, 1.1084E+04,
     + 1.1102E+04, 1.1120E+04, 1.1138E+04, 1.1156E+04, 1.1173E+04,
     +  1.1191E+04, 1.1209E+04, 1.1227E+04, 1.1245E+04, 1.1263E+04,
     + 1.1281E+04, 1.1299E+04, 1.1317E+04, 1.1335E+04, 1.1353E+04,
     +  1.1371E+04, 1.1389E+04, 1.1407E+04, 1.1425E+04, 1.1443E+04,
     + 1.1461E+04, 1.1479E+04, 1.1497E+04, 1.1516E+04, 1.1534E+04,
     +  1.1552E+04, 1.1570E+04, 1.1588E+04, 1.1606E+04, 1.1625E+04,
     + 1.1643E+04, 1.1661E+04, 1.1679E+04, 1.1698E+04, 1.1716E+04,
     +  1.1734E+04, 1.1753E+04, 1.1771E+04, 1.1789E+04, 1.1808E+04,
     + 1.1826E+04, 1.1844E+04, 1.1863E+04, 1.1881E+04, 1.1900E+04,
     +  1.1918E+04, 1.1937E+04, 1.1955E+04, 1.1974E+04, 1.1992E+04,
     + 1.2011E+04, 1.2029E+04, 1.2048E+04, 1.2066E+04, 1.2085E+04,
     +  1.2104E+04, 1.2122E+04, 1.2141E+04, 1.2159E+04, 1.2178E+04,
     + 1.2197E+04, 1.2216E+04, 1.2234E+04, 1.2253E+04, 1.2272E+04,
     +  1.2290E+04, 1.2309E+04, 1.2328E+04, 1.2347E+04, 1.2366E+04,
     + 1.2384E+04, 1.2403E+04, 1.2422E+04, 1.2441E+04, 1.2460E+04,
     +  1.2479E+04, 1.2498E+04, 1.2516E+04, 1.2535E+04, 1.2554E+04,
     + 1.2573E+04, 1.2592E+04, 1.2611E+04, 1.2630E+04, 1.2649E+04,
     +  1.2668E+04, 1.2687E+04, 1.2706E+04, 1.2726E+04, 1.2745E+04,
     + 1.2764E+04, 1.2783E+04, 1.2802E+04, 1.2821E+04, 1.2840E+04,
     +  1.2859E+04, 1.2879E+04, 1.2898E+04, 1.2917E+04, 1.2936E+04,
     + 1.2956E+04, 1.2975E+04, 1.2994E+04, 1.3013E+04, 1.3033E+04,
     +  1.3052E+04, 1.3071E+04, 1.3091E+04, 1.3110E+04, 1.3130E+04,
     + 1.3149E+04, 1.3168E+04, 1.3188E+04, 1.3207E+04, 1.3227E+04,
     +  1.3246E+04, 1.3266E+04, 1.3285E+04, 1.3305E+04, 1.3324E+04,
     + 1.3344E+04, 1.3363E+04, 1.3383E+04, 1.3402E+04, 1.3422E+04,
     +  1.3442E+04, 1.3461E+04, 1.3481E+04, 1.3501E+04, 1.3520E+04,
     + 1.3540E+04, 1.3560E+04, 1.3579E+04, 1.3599E+04, 1.3619E+04,
     +  1.3639E+04, 1.3658E+04, 1.3678E+04, 1.3698E+04, 1.3718E+04,
     + 1.3738E+04, 1.3757E+04, 1.3777E+04, 1.3797E+04, 1.3817E+04,
     +  1.3837E+04, 1.3857E+04, 1.3877E+04, 1.3897E+04, 1.3917E+04,
     + 1.3937E+04, 1.3957E+04, 1.3977E+04, 1.3997E+04, 1.4017E+04,
     +  1.4037E+04, 1.4057E+04, 1.4077E+04, 1.4097E+04, 1.4117E+04,
     + 1.4137E+04, 1.4157E+04, 1.4177E+04, 1.4198E+04, 1.4218E+04,
     +  1.4238E+04, 1.4258E+04, 1.4278E+04, 1.4299E+04, 1.4319E+04,
     + 1.4339E+04, 1.4359E+04, 1.4380E+04, 1.4400E+04, 1.4420E+04,
     +  1.4441E+04, 1.4461E+04, 1.4481E+04, 1.4502E+04, 1.4522E+04,
     + 1.4543E+04, 1.4563E+04, 1.4583E+04, 1.4604E+04, 1.4624E+04,
     +  1.4645E+04, 1.4665E+04, 1.4686E+04, 1.4706E+04, 1.4727E+04,
     + 1.4747E+04, 1.4768E+04, 1.4789E+04, 1.4809E+04, 1.4830E+04,
     +  1.4850E+04, 1.4871E+04, 1.4892E+04, 1.4912E+04, 1.4933E+04,
     + 1.4954E+04, 1.4975E+04, 1.4995E+04, 1.5016E+04, 1.5037E+04,
     +  1.5058E+04, 1.5078E+04, 1.5099E+04, 1.5120E+04, 1.5141E+04,
     + 1.5162E+04, 1.5182E+04, 1.5203E+04, 1.5224E+04, 1.5245E+04,
     +  1.5266E+04, 1.5287E+04, 1.5308E+04, 1.5329E+04, 1.5350E+04,
     + 1.5371E+04, 1.5392E+04, 1.5413E+04, 1.5434E+04, 1.5455E+04,
     +  1.5476E+04, 1.5497E+04, 1.5518E+04, 1.5539E+04, 1.5560E+04,
     + 1.5581E+04, 1.5603E+04, 1.5624E+04, 1.5645E+04, 1.5666E+04,
     +  1.5687E+04, 1.5709E+04, 1.5730E+04, 1.5751E+04, 1.5772E+04,
     + 1.5794E+04, 1.5815E+04, 1.5836E+04, 1.5858E+04, 1.5879E+04,
     +  1.5900E+04, 1.5922E+04, 1.5943E+04, 1.5964E+04, 1.5986E+04,
     + 1.6007E+04, 1.6029E+04, 1.6050E+04, 1.6071E+04, 1.6093E+04,
     +  1.6114E+04, 1.6136E+04, 1.6157E+04, 1.6179E+04, 1.6200E+04,
     + 1.6222E+04, 1.6244E+04, 1.6265E+04, 1.6287E+04, 1.6308E+04,
     +  1.6330E+04, 1.6352E+04, 1.6373E+04, 1.6395E+04, 1.6417E+04,
     + 1.6439E+04, 1.6460E+04, 1.6482E+04, 1.6504E+04, 1.6526E+04,
     +  1.6547E+04, 1.6569E+04, 1.6591E+04, 1.6613E+04, 1.6635E+04,
     + 1.6656E+04, 1.6678E+04, 1.6700E+04, 1.6722E+04, 1.6744E+04,
     +  1.6766E+04, 1.6788E+04, 1.6810E+04, 1.6832E+04, 1.6854E+04,
     + 1.6876E+04, 1.6898E+04, 1.6920E+04, 1.6942E+04, 1.6964E+04,
     +  1.6986E+04, 1.7008E+04, 1.7030E+04, 1.7052E+04, 1.7074E+04,
     + 1.7096E+04, 1.7119E+04, 1.7141E+04, 1.7163E+04, 1.7185E+04,
     +  1.7207E+04, 1.7230E+04, 1.7252E+04, 1.7274E+04, 1.7296E+04,
     + 1.7319E+04, 1.7341E+04, 1.7363E+04, 1.7386E+04, 1.7408E+04,
     +  1.7430E+04, 1.7453E+04, 1.7475E+04, 1.7497E+04, 1.7520E+04,
     + 1.7542E+04, 1.7565E+04, 1.7587E+04, 1.7610E+04, 1.7632E+04,
     +  1.7655E+04, 1.7677E+04, 1.7700E+04, 1.7722E+04, 1.7745E+04,
     + 1.7767E+04, 1.7790E+04, 1.7813E+04, 1.7835E+04, 1.7858E+04,
     +  1.7880E+04, 1.7903E+04, 1.7926E+04, 1.7948E+04, 1.7971E+04,
     + 1.7994E+04, 1.8017E+04, 1.8039E+04, 1.8062E+04, 1.8085E+04,
     +  1.8108E+04, 1.8130E+04, 1.8153E+04, 1.8176E+04, 1.8199E+04,
     + 1.8222E+04, 1.8245E+04, 1.8268E+04, 1.8290E+04, 1.8313E+04,
     +  1.8336E+04, 1.8359E+04, 1.8382E+04, 1.8405E+04, 1.8428E+04,
     + 1.8451E+04, 1.8474E+04, 1.8497E+04, 1.8520E+04, 1.8543E+04,
     +  1.8566E+04, 1.8589E+04, 1.8613E+04, 1.8636E+04, 1.8659E+04,
     + 1.8682E+04, 1.8705E+04, 1.8728E+04, 1.8752E+04, 1.8775E+04,
     +  1.8798E+04, 1.8821E+04, 1.8844E+04, 1.8868E+04, 1.8891E+04,
     + 1.8914E+04, 1.8938E+04, 1.8961E+04, 1.8984E+04, 1.9008E+04,
     +  1.9031E+04, 1.9054E+04, 1.9078E+04, 1.9101E+04, 1.9125E+04,
     + 1.9148E+04, 1.9171E+04, 1.9195E+04, 1.9218E+04, 1.9242E+04,
     +  1.9265E+04, 1.9289E+04, 1.9312E+04, 1.9336E+04, 1.9360E+04,
     + 1.9383E+04, 1.9407E+04, 1.9430E+04, 1.9454E+04, 1.9478E+04,
     +  1.9501E+04, 1.9525E+04, 1.9549E+04, 1.9572E+04, 1.9596E+04,
     + 1.9620E+04, 1.9643E+04, 1.9667E+04, 1.9691E+04, 1.9715E+04,
     +  1.9738E+04, 1.9762E+04, 1.9786E+04, 1.9810E+04, 1.9834E+04,
     + 1.9858E+04, 1.9882E+04, 1.9905E+04, 1.9929E+04, 1.9953E+04,
     +  1.9977E+04, 2.0001E+04, 2.0025E+04, 2.0049E+04, 2.0073E+04,
     + 2.0097E+04, 2.0121E+04, 2.0145E+04, 2.0169E+04, 2.0193E+04,
     +  2.0217E+04, 2.0241E+04, 2.0266E+04, 2.0290E+04, 2.0314E+04,
     + 2.0338E+04, 2.0362E+04, 2.0386E+04, 2.0410E+04, 2.0435E+04,
     +  2.0459E+04, 2.0483E+04, 2.0507E+04, 2.0532E+04, 2.0556E+04,
     + 2.0580E+04, 2.0605E+04, 2.0629E+04, 2.0653E+04, 2.0677E+04,
     +  2.0702E+04, 2.0726E+04, 2.0751E+04, 2.0775E+04, 2.0799E+04,
     + 2.0824E+04, 2.0848E+04, 2.0873E+04, 2.0897E+04, 2.0922E+04,
     +  2.0946E+04, 2.0971E+04, 2.0995E+04, 2.1020E+04, 2.1044E+04,
     + 2.1069E+04, 2.1094E+04, 2.1118E+04, 2.1143E+04, 2.1167E+04,
     +  2.1192E+04, 2.1217E+04, 2.1242E+04, 2.1266E+04, 2.1291E+04,
     + 2.1316E+04, 2.1340E+04, 2.1365E+04, 2.1390E+04, 2.1415E+04,
     +  2.1439E+04, 2.1464E+04, 2.1489E+04, 2.1514E+04, 2.1539E+04,
     + 2.1564E+04, 2.1588E+04, 2.1613E+04, 2.1638E+04, 2.1663E+04,
     +  2.1688E+04, 2.1713E+04, 2.1738E+04, 2.1763E+04, 2.1788E+04,
     + 2.1813E+04, 2.1838E+04, 2.1863E+04, 2.1888E+04, 2.1913E+04,
     +  2.1938E+04, 2.1963E+04, 2.1988E+04, 2.2014E+04, 2.2039E+04,
     + 2.2064E+04, 2.2089E+04, 2.2114E+04, 2.2139E+04, 2.2165E+04,
     +  2.2190E+04, 2.2215E+04, 2.2240E+04, 2.2265E+04, 2.2291E+04,
     + 2.2316E+04, 2.2341E+04, 2.2367E+04, 2.2392E+04, 2.2417E+04,
     +  2.2443E+04, 2.2468E+04, 2.2494E+04, 2.2519E+04, 2.2544E+04,
     + 2.2570E+04, 2.2595E+04, 2.2621E+04, 2.2646E+04, 2.2672E+04,
     +  2.2697E+04, 2.2723E+04, 2.2748E+04, 2.2774E+04, 2.2799E+04,
     + 2.2825E+04, 2.2850E+04, 2.2876E+04, 2.2902E+04, 2.2927E+04,
     +  2.2953E+04, 2.2979E+04, 2.3004E+04, 2.3030E+04, 2.3056E+04,
     + 2.3081E+04, 2.3107E+04, 2.3133E+04, 2.3159E+04, 2.3184E+04,
     +  2.3210E+04, 2.3236E+04, 2.3262E+04, 2.3288E+04, 2.3313E+04,
     + 2.3339E+04, 2.3365E+04, 2.3391E+04, 2.3417E+04, 2.3443E+04,
     +  2.3469E+04, 2.3495E+04, 2.3521E+04, 2.3547E+04, 2.3573E+04,
     + 2.3599E+04, 2.3625E+04, 2.3651E+04, 2.3677E+04, 2.3703E+04,
     +  2.3729E+04, 2.3755E+04, 2.3781E+04, 2.3807E+04, 2.3833E+04,
     + 2.3859E+04, 2.3886E+04, 2.3912E+04, 2.3938E+04, 2.3964E+04,
     +  2.3990E+04, 2.4017E+04, 2.4043E+04, 2.4069E+04, 2.4095E+04,
     + 2.4122E+04, 2.4148E+04, 2.4174E+04, 2.4200E+04, 2.4227E+04,
     +  2.4253E+04, 2.4280E+04, 2.4306E+04, 2.4332E+04, 2.4359E+04,
     + 2.4385E+04, 2.4412E+04, 2.4438E+04, 2.4464E+04, 2.4491E+04,
     +  2.4518E+04, 2.4544E+04, 2.4571E+04, 2.4597E+04, 2.4624E+04,
     + 2.4650E+04, 2.4677E+04, 2.4703E+04, 2.4730E+04, 2.4757E+04,
     +  2.4783E+04, 2.4810E+04, 2.4837E+04, 2.4863E+04, 2.4890E+04,
     + 2.4917E+04, 2.4943E+04, 2.4970E+04, 2.4997E+04, 2.5023E+04,
     +  2.5050E+04, 2.5077E+04, 2.5104E+04, 2.5131E+04, 2.5158E+04,
     + 2.5184E+04, 2.5211E+04, 2.5238E+04, 2.5265E+04, 2.5292E+04,
     +  2.5319E+04, 2.5346E+04, 2.5373E+04, 2.5400E+04, 2.5427E+04,
     + 2.5454E+04, 2.5481E+04, 2.5508E+04, 2.5535E+04, 2.5562E+04,
     +  2.5589E+04, 2.5616E+04, 2.5643E+04, 2.5670E+04, 2.5697E+04,
     + 2.5724E+04, 2.5751E+04, 2.5778E+04, 2.5806E+04, 2.5833E+04,
     +  2.5860E+04, 2.5887E+04, 2.5914E+04, 2.5942E+04, 2.5969E+04,
     + 2.5996E+04, 2.6023E+04, 2.6051E+04, 2.6078E+04, 2.6105E+04,
     +  2.6133E+04, 2.6160E+04, 2.6187E+04, 2.6215E+04, 2.6242E+04,
     + 2.6270E+04, 2.6297E+04, 2.6324E+04, 2.6352E+04, 2.6379E+04,
     +  2.6407E+04, 2.6434E+04, 2.6462E+04, 2.6489E+04, 2.6517E+04,
     + 2.6544E+04, 2.6572E+04, 2.6600E+04, 2.6627E+04, 2.6655E+04,
     +  2.6682E+04, 2.6710E+04, 2.6738E+04, 2.6765E+04, 2.6793E+04,
     + 2.6821E+04, 2.6848E+04, 2.6876E+04, 2.6904E+04, 2.6932E+04,
     +  2.6959E+04, 2.6987E+04, 2.7015E+04, 2.7043E+04, 2.7070E+04,
     + 2.7098E+04, 2.7126E+04, 2.7154E+04, 2.7182E+04, 2.7210E+04,
     +  2.7238E+04, 2.7265E+04, 2.7293E+04, 2.7321E+04, 2.7349E+04,
     + 2.7377E+04, 2.7405E+04, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        46
      data (QofT( 2,J),J=1,2408)/ 1.7842E+00, 3.1078E+00, 4.4724E+00,
     +  5.9493E+00, 7.5813E+00, 9.3734E+00, 1.1312E+01, 1.3380E+01,
     + 1.5557E+01, 1.7827E+01, 2.0176E+01, 2.2592E+01, 2.5065E+01,
     +  2.7587E+01, 3.0152E+01, 3.2753E+01, 3.5387E+01, 3.8048E+01,
     + 4.0734E+01, 4.3442E+01, 4.8913E+01, 5.4447E+01, 6.0031E+01,
     +  6.5656E+01, 7.1316E+01, 7.7005E+01, 8.2718E+01, 8.8452E+01,
     + 9.4204E+01, 9.9972E+01, 1.0575E+02, 1.1155E+02, 1.1735E+02,
     +  1.2316E+02, 1.2898E+02, 1.3481E+02, 1.4065E+02, 1.4649E+02,
     + 1.5234E+02, 1.5819E+02, 1.6405E+02, 1.6991E+02, 1.7577E+02,
     +  1.8164E+02, 1.8752E+02, 1.9339E+02, 1.9927E+02, 2.0515E+02,
     + 2.1103E+02, 2.1691E+02, 2.2280E+02, 2.2869E+02, 2.3458E+02,
     +  2.4047E+02, 2.4637E+02, 2.5226E+02, 2.5816E+02, 2.6406E+02,
     + 2.6996E+02, 2.7586E+02, 2.8176E+02, 2.8766E+02, 2.9357E+02,
     +  2.9947E+02, 3.0538E+02, 3.1128E+02, 3.1719E+02, 3.2310E+02,
     + 3.2901E+02, 3.3492E+02, 3.4083E+02, 3.4674E+02, 3.5265E+02,
     +  3.5856E+02, 3.6448E+02, 3.7039E+02, 3.7631E+02, 3.8222E+02,
     + 3.8814E+02, 3.9405E+02, 3.9997E+02, 4.0589E+02, 4.1181E+02,
     +  4.1772E+02, 4.2364E+02, 4.2956E+02, 4.3548E+02, 4.4140E+02,
     + 4.4732E+02, 4.5325E+02, 4.5917E+02, 4.6509E+02, 4.7102E+02,
     +  4.7694E+02, 4.8287E+02, 4.8879E+02, 4.9472E+02, 5.0065E+02,
     + 5.0658E+02, 5.1250E+02, 5.1844E+02, 5.2437E+02, 5.3030E+02,
     +  5.3623E+02, 5.4217E+02, 5.4810E+02, 5.5404E+02, 5.5998E+02,
     + 5.6592E+02, 5.7186E+02, 5.7780E+02, 5.8374E+02, 5.8969E+02,
     +  5.9563E+02, 6.0158E+02, 6.0753E+02, 6.1348E+02, 6.1943E+02,
     + 6.2539E+02, 6.3135E+02, 6.3730E+02, 6.4327E+02, 6.4923E+02,
     +  6.5519E+02, 6.6116E+02, 6.6713E+02, 6.7310E+02, 6.7908E+02,
     + 6.8505E+02, 6.9103E+02, 6.9701E+02, 7.0300E+02, 7.0899E+02,
     +  7.1498E+02, 7.2098E+02, 7.2697E+02, 7.3297E+02, 7.3898E+02,
     + 7.4499E+02, 7.5099E+02, 7.5701E+02, 7.6302E+02, 7.6905E+02,
     +  7.7507E+02, 7.8110E+02, 7.8714E+02, 7.9318E+02, 7.9922E+02,
     + 8.0527E+02, 8.1131E+02, 8.1737E+02, 8.2342E+02, 8.2949E+02,
     +  8.3556E+02, 8.4163E+02, 8.4771E+02, 8.5380E+02, 8.5988E+02,
     + 8.6598E+02, 8.7208E+02, 8.7818E+02, 8.8429E+02, 8.9041E+02,
     +  8.9653E+02, 9.0266E+02, 9.0878E+02, 9.1492E+02, 9.2107E+02,
     + 9.2722E+02, 9.3338E+02, 9.3953E+02, 9.4570E+02, 9.5188E+02,
     +  9.5806E+02, 9.6425E+02, 9.7044E+02, 9.7664E+02, 9.8285E+02,
     + 9.8907E+02, 9.9529E+02, 1.0015E+03, 1.0077E+03, 1.0140E+03,
     +  1.0202E+03, 1.0265E+03, 1.0328E+03, 1.0390E+03, 1.0453E+03,
     + 1.0516E+03, 1.0579E+03, 1.0642E+03, 1.0705E+03, 1.0768E+03,
     +  1.0832E+03, 1.0895E+03, 1.0958E+03, 1.1022E+03, 1.1085E+03,
     + 1.1149E+03, 1.1213E+03, 1.1277E+03, 1.1341E+03, 1.1405E+03,
     +  1.1469E+03, 1.1533E+03, 1.1597E+03, 1.1661E+03, 1.1726E+03,
     + 1.1790E+03, 1.1855E+03, 1.1920E+03, 1.1985E+03, 1.2050E+03,
     +  1.2115E+03, 1.2180E+03, 1.2245E+03, 1.2310E+03, 1.2375E+03,
     + 1.2441E+03, 1.2506E+03, 1.2572E+03, 1.2638E+03, 1.2704E+03,
     +  1.2770E+03, 1.2836E+03, 1.2902E+03, 1.2968E+03, 1.3034E+03,
     + 1.3101E+03, 1.3167E+03, 1.3234E+03, 1.3301E+03, 1.3368E+03,
     +  1.3435E+03, 1.3502E+03, 1.3569E+03, 1.3636E+03, 1.3703E+03,
     + 1.3771E+03, 1.3839E+03, 1.3906E+03, 1.3974E+03, 1.4042E+03,
     +  1.4110E+03, 1.4178E+03, 1.4246E+03, 1.4315E+03, 1.4383E+03,
     + 1.4452E+03, 1.4520E+03, 1.4589E+03, 1.4658E+03, 1.4727E+03,
     +  1.4796E+03, 1.4865E+03, 1.4935E+03, 1.5004E+03, 1.5074E+03,
     + 1.5143E+03, 1.5213E+03, 1.5283E+03, 1.5353E+03, 1.5423E+03,
     +  1.5494E+03, 1.5564E+03, 1.5635E+03, 1.5705E+03, 1.5776E+03,
     + 1.5847E+03, 1.5918E+03, 1.5989E+03, 1.6060E+03, 1.6131E+03,
     +  1.6203E+03, 1.6274E+03, 1.6346E+03, 1.6418E+03, 1.6490E+03,
     + 1.6562E+03, 1.6634E+03, 1.6706E+03, 1.6778E+03, 1.6851E+03,
     +  1.6924E+03, 1.6996E+03, 1.7069E+03, 1.7142E+03, 1.7215E+03,
     + 1.7289E+03, 1.7362E+03, 1.7435E+03, 1.7509E+03, 1.7583E+03,
     +  1.7657E+03, 1.7731E+03, 1.7805E+03, 1.7879E+03, 1.7953E+03,
     + 1.8028E+03, 1.8102E+03, 1.8177E+03, 1.8252E+03, 1.8327E+03,
     +  1.8402E+03, 1.8477E+03, 1.8553E+03, 1.8628E+03, 1.8704E+03,
     + 1.8779E+03, 1.8855E+03, 1.8931E+03, 1.9008E+03, 1.9084E+03,
     +  1.9160E+03, 1.9237E+03, 1.9313E+03, 1.9390E+03, 1.9467E+03,
     + 1.9544E+03, 1.9621E+03, 1.9699E+03, 1.9776E+03, 1.9853E+03,
     +  1.9931E+03, 2.0009E+03, 2.0087E+03, 2.0165E+03, 2.0243E+03,
     + 2.0321E+03, 2.0400E+03, 2.0479E+03, 2.0557E+03, 2.0636E+03,
     +  2.0715E+03, 2.0794E+03, 2.0873E+03, 2.0953E+03, 2.1032E+03,
     + 2.1112E+03, 2.1192E+03, 2.1272E+03, 2.1352E+03, 2.1432E+03,
     +  2.1512E+03, 2.1593E+03, 2.1673E+03, 2.1754E+03, 2.1835E+03,
     + 2.1916E+03, 2.1997E+03, 2.2078E+03, 2.2160E+03, 2.2241E+03,
     +  2.2323E+03, 2.2405E+03, 2.2486E+03, 2.2569E+03, 2.2651E+03,
     + 2.2733E+03, 2.2816E+03, 2.2898E+03, 2.2981E+03, 2.3064E+03,
     +  2.3147E+03, 2.3230E+03, 2.3313E+03, 2.3397E+03, 2.3480E+03,
     + 2.3564E+03, 2.3648E+03, 2.3732E+03, 2.3816E+03, 2.3900E+03,
     +  2.3984E+03, 2.4069E+03, 2.4154E+03, 2.4238E+03, 2.4323E+03,
     + 2.4408E+03, 2.4493E+03, 2.4579E+03, 2.4664E+03, 2.4750E+03,
     +  2.4836E+03, 2.4921E+03, 2.5007E+03, 2.5094E+03, 2.5180E+03,
     + 2.5266E+03, 2.5353E+03, 2.5440E+03, 2.5527E+03, 2.5613E+03,
     +  2.5701E+03, 2.5788E+03, 2.5875E+03, 2.5963E+03, 2.6050E+03,
     + 2.6138E+03, 2.6226E+03, 2.6314E+03, 2.6403E+03, 2.6491E+03,
     +  2.6579E+03, 2.6668E+03, 2.6757E+03, 2.6846E+03, 2.6935E+03,
     + 2.7024E+03, 2.7113E+03, 2.7203E+03, 2.7293E+03, 2.7382E+03,
     +  2.7472E+03, 2.7562E+03, 2.7652E+03, 2.7743E+03, 2.7833E+03,
     + 2.7924E+03, 2.8015E+03, 2.8106E+03, 2.8197E+03, 2.8288E+03,
     +  2.8379E+03, 2.8470E+03, 2.8562E+03, 2.8654E+03, 2.8746E+03,
     + 2.8838E+03, 2.8930E+03, 2.9022E+03, 2.9114E+03, 2.9207E+03,
     +  2.9300E+03, 2.9393E+03, 2.9486E+03, 2.9579E+03, 2.9672E+03,
     + 2.9766E+03, 2.9859E+03, 2.9953E+03, 3.0047E+03, 3.0141E+03,
     +  3.0235E+03, 3.0329E+03, 3.0423E+03, 3.0518E+03, 3.0613E+03,
     + 3.0708E+03, 3.0803E+03, 3.0898E+03, 3.0993E+03, 3.1088E+03,
     +  3.1184E+03, 3.1280E+03, 3.1376E+03, 3.1472E+03, 3.1568E+03,
     + 3.1664E+03, 3.1760E+03, 3.1857E+03, 3.1954E+03, 3.2051E+03,
     +  3.2147E+03, 3.2245E+03, 3.2342E+03, 3.2439E+03, 3.2537E+03,
     + 3.2634E+03, 3.2732E+03, 3.2830E+03, 3.2929E+03, 3.3027E+03,
     +  3.3125E+03, 3.3224E+03, 3.3322E+03, 3.3421E+03, 3.3520E+03,
     + 3.3619E+03, 3.3719E+03, 3.3818E+03, 3.3918E+03, 3.4017E+03,
     +  3.4117E+03, 3.4217E+03, 3.4317E+03, 3.4417E+03, 3.4518E+03,
     + 3.4618E+03, 3.4719E+03, 3.4820E+03, 3.4921E+03, 3.5022E+03,
     +  3.5123E+03, 3.5225E+03, 3.5326E+03, 3.5428E+03, 3.5530E+03,
     + 3.5632E+03, 3.5734E+03, 3.5836E+03, 3.5939E+03, 3.6041E+03,
     +  3.6144E+03, 3.6247E+03, 3.6350E+03, 3.6453E+03, 3.6556E+03,
     + 3.6660E+03, 3.6763E+03, 3.6867E+03, 3.6971E+03, 3.7075E+03,
     +  3.7179E+03, 3.7283E+03, 3.7388E+03, 3.7492E+03, 3.7597E+03,
     + 3.7702E+03, 3.7807E+03, 3.7912E+03, 3.8018E+03, 3.8123E+03,
     +  3.8229E+03, 3.8334E+03, 3.8440E+03, 3.8546E+03, 3.8652E+03,
     + 3.8759E+03, 3.8865E+03, 3.8972E+03, 3.9079E+03, 3.9186E+03,
     +  3.9293E+03, 3.9400E+03, 3.9507E+03, 3.9615E+03, 3.9722E+03,
     + 3.9830E+03, 3.9938E+03, 4.0046E+03, 4.0154E+03, 4.0263E+03,
     +  4.0371E+03, 4.0480E+03, 4.0589E+03, 4.0698E+03, 4.0807E+03,
     + 4.0916E+03, 4.1025E+03, 4.1135E+03, 4.1244E+03, 4.1354E+03,
     +  4.1464E+03, 4.1574E+03, 4.1685E+03, 4.1795E+03, 4.1905E+03,
     + 4.2016E+03, 4.2127E+03, 4.2238E+03, 4.2349E+03, 4.2460E+03,
     +  4.2572E+03, 4.2683E+03, 4.2795E+03, 4.2907E+03, 4.3019E+03,
     + 4.3131E+03, 4.3243E+03, 4.3356E+03, 4.3468E+03, 4.3581E+03,
     +  4.3694E+03, 4.3807E+03, 4.3920E+03, 4.4033E+03, 4.4147E+03,
     + 4.4260E+03, 4.4374E+03, 4.4488E+03, 4.4602E+03, 4.4716E+03,
     +  4.4831E+03, 4.4945E+03, 4.5059E+03, 4.5174E+03, 4.5289E+03,
     + 4.5404E+03, 4.5519E+03, 4.5635E+03, 4.5750E+03, 4.5866E+03,
     +  4.5982E+03, 4.6098E+03, 4.6214E+03, 4.6330E+03, 4.6447E+03,
     + 4.6563E+03, 4.6680E+03, 4.6796E+03, 4.6913E+03, 4.7030E+03,
     +  4.7148E+03, 4.7265E+03, 4.7382E+03, 4.7500E+03, 4.7618E+03,
     + 4.7736E+03, 4.7854E+03, 4.7972E+03, 4.8091E+03, 4.8209E+03,
     +  4.8328E+03, 4.8447E+03, 4.8566E+03, 4.8685E+03, 4.8805E+03,
     + 4.8924E+03, 4.9043E+03, 4.9163E+03, 4.9283E+03, 4.9403E+03,
     +  4.9523E+03, 4.9644E+03, 4.9764E+03, 4.9885E+03, 5.0005E+03,
     + 5.0126E+03, 5.0247E+03, 5.0368E+03, 5.0490E+03, 5.0611E+03,
     +  5.0733E+03, 5.0855E+03, 5.0977E+03, 5.1099E+03, 5.1221E+03,
     + 5.1343E+03, 5.1466E+03, 5.1589E+03, 5.1711E+03, 5.1834E+03,
     +  5.1958E+03, 5.2081E+03, 5.2204E+03, 5.2327E+03, 5.2451E+03,
     + 5.2575E+03, 5.2699E+03, 5.2823E+03, 5.2948E+03, 5.3072E+03,
     +  5.3196E+03, 5.3321E+03, 5.3446E+03, 5.3571E+03, 5.3696E+03,
     + 5.3822E+03, 5.3947E+03, 5.4073E+03, 5.4199E+03, 5.4324E+03,
     +  5.4450E+03, 5.4577E+03, 5.4703E+03, 5.4829E+03, 5.4956E+03,
     + 5.5083E+03, 5.5210E+03, 5.5337E+03, 5.5464E+03, 5.5591E+03,
     +  5.5719E+03, 5.5847E+03, 5.5974E+03, 5.6102E+03, 5.6230E+03,
     + 5.6359E+03, 5.6487E+03, 5.6615E+03, 5.6744E+03, 5.6873E+03,
     +  5.7002E+03, 5.7131E+03, 5.7260E+03, 5.7390E+03, 5.7520E+03,
     + 5.7649E+03, 5.7779E+03, 5.7909E+03, 5.8039E+03, 5.8169E+03,
     +  5.8300E+03, 5.8431E+03, 5.8561E+03, 5.8692E+03, 5.8823E+03,
     + 5.8955E+03, 5.9086E+03, 5.9217E+03, 5.9349E+03, 5.9481E+03,
     +  5.9612E+03, 5.9745E+03, 5.9877E+03, 6.0009E+03, 6.0142E+03,
     + 6.0274E+03, 6.0407E+03, 6.0540E+03, 6.0673E+03, 6.0807E+03,
     +  6.0940E+03, 6.1074E+03, 6.1207E+03, 6.1341E+03, 6.1475E+03,
     + 6.1609E+03, 6.1743E+03, 6.1878E+03, 6.2013E+03, 6.2147E+03,
     +  6.2282E+03, 6.2417E+03, 6.2552E+03, 6.2687E+03, 6.2823E+03,
     + 6.2959E+03, 6.3094E+03, 6.3231E+03, 6.3367E+03, 6.3503E+03,
     +  6.3639E+03, 6.3775E+03, 6.3912E+03, 6.4049E+03, 6.4186E+03,
     + 6.4323E+03, 6.4460E+03, 6.4597E+03, 6.4735E+03, 6.4873E+03,
     +  6.5011E+03, 6.5148E+03, 6.5287E+03, 6.5425E+03, 6.5563E+03,
     + 6.5702E+03, 6.5840E+03, 6.5979E+03, 6.6119E+03, 6.6258E+03,
     +  6.6397E+03, 6.6536E+03, 6.6676E+03, 6.6816E+03, 6.6956E+03,
     + 6.7096E+03, 6.7236E+03, 6.7376E+03, 6.7517E+03, 6.7657E+03,
     +  6.7798E+03, 6.7939E+03, 6.8080E+03, 6.8221E+03, 6.8363E+03,
     + 6.8504E+03, 6.8646E+03, 6.8788E+03, 6.8930E+03, 6.9072E+03,
     +  6.9214E+03, 6.9357E+03, 6.9499E+03, 6.9642E+03, 6.9785E+03,
     + 6.9928E+03, 7.0071E+03, 7.0214E+03, 7.0357E+03, 7.0501E+03,
     +  7.0645E+03, 7.0789E+03, 7.0933E+03, 7.1077E+03, 7.1221E+03,
     + 7.1366E+03, 7.1510E+03, 7.1655E+03, 7.1800E+03, 7.1945E+03,
     +  7.2090E+03, 7.2236E+03, 7.2381E+03, 7.2527E+03, 7.2673E+03,
     + 7.2819E+03, 7.2965E+03, 7.3111E+03, 7.3257E+03, 7.3404E+03,
     +  7.3551E+03, 7.3698E+03, 7.3845E+03, 7.3992E+03, 7.4139E+03,
     + 7.4287E+03, 7.4434E+03, 7.4582E+03, 7.4729E+03, 7.4878E+03,
     +  7.5026E+03, 7.5174E+03, 7.5323E+03, 7.5471E+03, 7.5620E+03,
     + 7.5769E+03, 7.5918E+03, 7.6067E+03, 7.6217E+03, 7.6366E+03,
     +  7.6516E+03, 7.6666E+03, 7.6816E+03, 7.6965E+03, 7.7116E+03,
     + 7.7266E+03, 7.7417E+03, 7.7567E+03, 7.7718E+03, 7.7869E+03,
     +  7.8021E+03, 7.8172E+03, 7.8323E+03, 7.8475E+03, 7.8626E+03,
     + 7.8778E+03, 7.8930E+03, 7.9083E+03, 7.9235E+03, 7.9387E+03,
     +  7.9540E+03, 7.9693E+03, 7.9846E+03, 7.9999E+03, 8.0152E+03,
     + 8.0306E+03, 8.0459E+03, 8.0613E+03, 8.0766E+03, 8.0920E+03,
     +  8.1075E+03, 8.1229E+03, 8.1383E+03, 8.1538E+03, 8.1692E+03,
     + 8.1847E+03, 8.2002E+03, 8.2157E+03, 8.2313E+03, 8.2468E+03,
     +  8.2623E+03, 8.2779E+03, 8.2935E+03, 8.3091E+03, 8.3247E+03,
     + 8.3404E+03, 8.3561E+03, 8.3717E+03, 8.3874E+03, 8.4030E+03,
     +  8.4188E+03, 8.4345E+03, 8.4502E+03, 8.4660E+03, 8.4818E+03,
     + 8.4976E+03, 8.5133E+03, 8.5291E+03, 8.5450E+03, 8.5608E+03,
     +  8.5767E+03, 8.5926E+03, 8.6084E+03, 8.6243E+03, 8.6402E+03,
     + 8.6562E+03, 8.6721E+03, 8.6881E+03, 8.7040E+03, 8.7200E+03,
     +  8.7360E+03, 8.7520E+03, 8.7681E+03, 8.7841E+03, 8.8002E+03,
     + 8.8163E+03, 8.8324E+03, 8.8485E+03, 8.8646E+03, 8.8807E+03,
     +  8.8969E+03, 8.9131E+03, 8.9293E+03, 8.9454E+03, 8.9616E+03,
     + 8.9779E+03, 8.9941E+03, 9.0104E+03, 9.0267E+03, 9.0429E+03,
     +  9.0592E+03, 9.0756E+03, 9.0919E+03, 9.1083E+03, 9.1246E+03,
     + 9.1410E+03, 9.1574E+03, 9.1738E+03, 9.1902E+03, 9.2066E+03,
     +  9.2231E+03, 9.2396E+03, 9.2560E+03, 9.2725E+03, 9.2890E+03,
     + 9.3055E+03, 9.3221E+03, 9.3386E+03, 9.3552E+03, 9.3718E+03,
     +  9.3884E+03, 9.4050E+03, 9.4216E+03, 9.4383E+03, 9.4549E+03,
     + 9.4716E+03, 9.4883E+03, 9.5050E+03, 9.5217E+03, 9.5384E+03,
     +  9.5552E+03, 9.5719E+03, 9.5887E+03, 9.6055E+03, 9.6223E+03,
     + 9.6392E+03, 9.6560E+03, 9.6728E+03, 9.6897E+03, 9.7066E+03,
     +  9.7235E+03, 9.7404E+03, 9.7573E+03, 9.7743E+03, 9.7912E+03,
     + 9.8082E+03, 9.8252E+03, 9.8422E+03, 9.8592E+03, 9.8762E+03,
     +  9.8933E+03, 9.9103E+03, 9.9274E+03, 9.9445E+03, 9.9616E+03,
     + 9.9787E+03, 9.9958E+03, 1.0013E+04, 1.0030E+04, 1.0047E+04,
     +  1.0065E+04, 1.0082E+04, 1.0099E+04, 1.0116E+04, 1.0133E+04,
     + 1.0151E+04, 1.0168E+04, 1.0185E+04, 1.0203E+04, 1.0220E+04,
     +  1.0237E+04, 1.0255E+04, 1.0272E+04, 1.0289E+04, 1.0307E+04,
     + 1.0324E+04, 1.0342E+04, 1.0359E+04, 1.0377E+04, 1.0394E+04,
     +  1.0412E+04, 1.0429E+04, 1.0447E+04, 1.0464E+04, 1.0482E+04,
     + 1.0499E+04, 1.0517E+04, 1.0534E+04, 1.0552E+04, 1.0570E+04,
     +  1.0587E+04, 1.0605E+04, 1.0623E+04, 1.0640E+04, 1.0658E+04,
     + 1.0676E+04, 1.0693E+04, 1.0711E+04, 1.0729E+04, 1.0747E+04,
     +  1.0764E+04, 1.0782E+04, 1.0800E+04, 1.0818E+04, 1.0836E+04,
     + 1.0854E+04, 1.0872E+04, 1.0889E+04, 1.0907E+04, 1.0925E+04,
     +  1.0943E+04, 1.0961E+04, 1.0979E+04, 1.0997E+04, 1.1015E+04,
     + 1.1033E+04, 1.1051E+04, 1.1069E+04, 1.1087E+04, 1.1105E+04,
     +  1.1123E+04, 1.1141E+04, 1.1160E+04, 1.1178E+04, 1.1196E+04,
     + 1.1214E+04, 1.1232E+04, 1.1250E+04, 1.1269E+04, 1.1287E+04,
     +  1.1305E+04, 1.1323E+04, 1.1342E+04, 1.1360E+04, 1.1378E+04,
     + 1.1396E+04, 1.1415E+04, 1.1433E+04, 1.1451E+04, 1.1470E+04,
     +  1.1488E+04, 1.1507E+04, 1.1525E+04, 1.1543E+04, 1.1562E+04,
     + 1.1580E+04, 1.1599E+04, 1.1617E+04, 1.1636E+04, 1.1654E+04,
     +  1.1673E+04, 1.1691E+04, 1.1710E+04, 1.1729E+04, 1.1747E+04,
     + 1.1766E+04, 1.1784E+04, 1.1803E+04, 1.1822E+04, 1.1840E+04,
     +  1.1859E+04, 1.1878E+04, 1.1896E+04, 1.1915E+04, 1.1934E+04,
     + 1.1953E+04, 1.1971E+04, 1.1990E+04, 1.2009E+04, 1.2028E+04,
     +  1.2047E+04, 1.2065E+04, 1.2084E+04, 1.2103E+04, 1.2122E+04,
     + 1.2141E+04, 1.2160E+04, 1.2179E+04, 1.2198E+04, 1.2217E+04,
     +  1.2236E+04, 1.2255E+04, 1.2274E+04, 1.2293E+04, 1.2312E+04,
     + 1.2331E+04, 1.2350E+04, 1.2369E+04, 1.2388E+04, 1.2407E+04,
     +  1.2426E+04, 1.2446E+04, 1.2465E+04, 1.2484E+04, 1.2503E+04,
     + 1.2522E+04, 1.2541E+04, 1.2561E+04, 1.2580E+04, 1.2599E+04,
     +  1.2618E+04, 1.2638E+04, 1.2657E+04, 1.2676E+04, 1.2696E+04,
     + 1.2715E+04, 1.2734E+04, 1.2754E+04, 1.2773E+04, 1.2793E+04,
     +  1.2812E+04, 1.2831E+04, 1.2851E+04, 1.2870E+04, 1.2890E+04,
     + 1.2909E+04, 1.2929E+04, 1.2948E+04, 1.2968E+04, 1.2988E+04,
     +  1.3007E+04, 1.3027E+04, 1.3046E+04, 1.3066E+04, 1.3086E+04,
     + 1.3105E+04, 1.3125E+04, 1.3145E+04, 1.3164E+04, 1.3184E+04,
     +  1.3204E+04, 1.3223E+04, 1.3243E+04, 1.3263E+04, 1.3283E+04,
     + 1.3303E+04, 1.3322E+04, 1.3342E+04, 1.3362E+04, 1.3382E+04,
     +  1.3402E+04, 1.3422E+04, 1.3442E+04, 1.3461E+04, 1.3481E+04,
     + 1.3501E+04, 1.3521E+04, 1.3541E+04, 1.3561E+04, 1.3581E+04,
     +  1.3601E+04, 1.3621E+04, 1.3641E+04, 1.3661E+04, 1.3682E+04,
     + 1.3702E+04, 1.3722E+04, 1.3742E+04, 1.3762E+04, 1.3782E+04,
     +  1.3802E+04, 1.3822E+04, 1.3843E+04, 1.3863E+04, 1.3883E+04,
     + 1.3903E+04, 1.3924E+04, 1.3944E+04, 1.3964E+04, 1.3984E+04,
     +  1.4005E+04, 1.4025E+04, 1.4046E+04, 1.4066E+04, 1.4086E+04,
     + 1.4107E+04, 1.4127E+04, 1.4148E+04, 1.4168E+04, 1.4188E+04,
     +  1.4209E+04, 1.4229E+04, 1.4250E+04, 1.4270E+04, 1.4291E+04,
     + 1.4311E+04, 1.4332E+04, 1.4353E+04, 1.4373E+04, 1.4394E+04,
     +  1.4414E+04, 1.4435E+04, 1.4456E+04, 1.4476E+04, 1.4497E+04,
     + 1.4518E+04, 1.4538E+04, 1.4559E+04, 1.4580E+04, 1.4601E+04,
     +  1.4621E+04, 1.4642E+04, 1.4663E+04, 1.4684E+04, 1.4705E+04,
     + 1.4725E+04, 1.4746E+04, 1.4767E+04, 1.4788E+04, 1.4809E+04,
     +  1.4830E+04, 1.4851E+04, 1.4872E+04, 1.4893E+04, 1.4914E+04,
     + 1.4935E+04, 1.4956E+04, 1.4977E+04, 1.4998E+04, 1.5019E+04,
     +  1.5040E+04, 1.5061E+04, 1.5082E+04, 1.5103E+04, 1.5124E+04,
     + 1.5145E+04, 1.5166E+04, 1.5188E+04, 1.5209E+04, 1.5230E+04,
     +  1.5251E+04, 1.5272E+04, 1.5294E+04, 1.5315E+04, 1.5336E+04,
     + 1.5357E+04, 1.5379E+04, 1.5400E+04, 1.5421E+04, 1.5443E+04,
     +  1.5464E+04, 1.5485E+04, 1.5507E+04, 1.5528E+04, 1.5550E+04,
     + 1.5571E+04, 1.5593E+04, 1.5614E+04, 1.5636E+04, 1.5657E+04,
     +  1.5679E+04, 1.5700E+04, 1.5722E+04, 1.5743E+04, 1.5765E+04,
     + 1.5786E+04, 1.5808E+04, 1.5830E+04, 1.5851E+04, 1.5873E+04,
     +  1.5894E+04, 1.5916E+04, 1.5938E+04, 1.5960E+04, 1.5981E+04,
     + 1.6003E+04, 1.6025E+04, 1.6046E+04, 1.6068E+04, 1.6090E+04,
     +  1.6112E+04, 1.6134E+04, 1.6156E+04, 1.6177E+04, 1.6199E+04,
     + 1.6221E+04, 1.6243E+04, 1.6265E+04, 1.6287E+04, 1.6309E+04,
     +  1.6331E+04, 1.6353E+04, 1.6375E+04, 1.6397E+04, 1.6419E+04,
     + 1.6441E+04, 1.6463E+04, 1.6485E+04, 1.6507E+04, 1.6529E+04,
     +  1.6551E+04, 1.6573E+04, 1.6595E+04, 1.6617E+04, 1.6640E+04,
     + 1.6662E+04, 1.6684E+04, 1.6706E+04, 1.6728E+04, 1.6751E+04,
     +  1.6773E+04, 1.6795E+04, 1.6817E+04, 1.6840E+04, 1.6862E+04,
     + 1.6884E+04, 1.6907E+04, 1.6929E+04, 1.6951E+04, 1.6974E+04,
     +  1.6996E+04, 1.7019E+04, 1.7041E+04, 1.7064E+04, 1.7086E+04,
     + 1.7108E+04, 1.7131E+04, 1.7154E+04, 1.7176E+04, 1.7199E+04,
     +  1.7221E+04, 1.7244E+04, 1.7266E+04, 1.7289E+04, 1.7311E+04,
     + 1.7334E+04, 1.7357E+04, 1.7379E+04, 1.7402E+04, 1.7425E+04,
     +  1.7447E+04, 1.7470E+04, 1.7493E+04, 1.7516E+04, 1.7538E+04,
     + 1.7561E+04, 1.7584E+04, 1.7607E+04, 1.7629E+04, 1.7652E+04,
     +  1.7675E+04, 1.7698E+04, 1.7721E+04, 1.7744E+04, 1.7767E+04,
     + 1.7790E+04, 1.7812E+04, 1.7835E+04, 1.7858E+04, 1.7881E+04,
     +  1.7904E+04, 1.7927E+04, 1.7950E+04, 1.7973E+04, 1.7997E+04,
     + 1.8020E+04, 1.8043E+04, 1.8066E+04, 1.8089E+04, 1.8112E+04,
     +  1.8135E+04, 1.8158E+04, 1.8181E+04, 1.8205E+04, 1.8228E+04,
     + 1.8251E+04, 1.8274E+04, 1.8298E+04, 1.8321E+04, 1.8344E+04,
     +  1.8367E+04, 1.8391E+04, 1.8414E+04, 1.8437E+04, 1.8461E+04,
     + 1.8484E+04, 1.8507E+04, 1.8531E+04, 1.8554E+04, 1.8578E+04,
     +  1.8601E+04, 1.8625E+04, 1.8648E+04, 1.8671E+04, 1.8695E+04,
     + 1.8718E+04, 1.8742E+04, 1.8765E+04, 1.8789E+04, 1.8813E+04,
     +  1.8836E+04, 1.8860E+04, 1.8883E+04, 1.8907E+04, 1.8931E+04,
     + 1.8954E+04, 1.8978E+04, 1.9002E+04, 1.9025E+04, 1.9049E+04,
     +  1.9073E+04, 1.9097E+04, 1.9120E+04, 1.9144E+04, 1.9168E+04,
     + 1.9192E+04, 1.9216E+04, 1.9239E+04, 1.9263E+04, 1.9287E+04,
     +  1.9311E+04, 1.9335E+04, 1.9359E+04, 1.9383E+04, 1.9407E+04,
     + 1.9431E+04, 1.9455E+04, 1.9479E+04, 1.9503E+04, 1.9527E+04,
     +  1.9551E+04, 1.9575E+04, 1.9599E+04, 1.9623E+04, 1.9647E+04,
     + 1.9671E+04, 1.9695E+04, 1.9719E+04, 1.9743E+04, 1.9767E+04,
     +  1.9792E+04, 1.9816E+04, 1.9840E+04, 1.9864E+04, 1.9888E+04,
     + 1.9913E+04, 1.9937E+04, 1.9961E+04, 1.9986E+04, 2.0010E+04,
     +  2.0034E+04, 2.0058E+04, 2.0083E+04, 2.0107E+04, 2.0132E+04,
     + 2.0156E+04, 2.0180E+04, 2.0205E+04, 2.0229E+04, 2.0254E+04,
     +  2.0278E+04, 2.0303E+04, 2.0327E+04, 2.0352E+04, 2.0376E+04,
     + 2.0401E+04, 2.0425E+04, 2.0450E+04, 2.0474E+04, 2.0499E+04,
     +  2.0524E+04, 2.0548E+04, 2.0573E+04, 2.0598E+04, 2.0622E+04,
     + 2.0647E+04, 2.0672E+04, 2.0696E+04, 2.0721E+04, 2.0746E+04,
     +  2.0771E+04, 2.0795E+04, 2.0820E+04, 2.0845E+04, 2.0870E+04,
     + 2.0895E+04, 2.0919E+04, 2.0944E+04, 2.0969E+04, 2.0994E+04,
     +  2.1019E+04, 2.1044E+04, 2.1069E+04, 2.1094E+04, 2.1119E+04,
     + 2.1144E+04, 2.1169E+04, 2.1194E+04, 2.1219E+04, 2.1244E+04,
     +  2.1269E+04, 2.1294E+04, 2.1319E+04, 2.1344E+04, 2.1369E+04,
     + 2.1394E+04, 2.1420E+04, 2.1445E+04, 2.1470E+04, 2.1495E+04,
     +  2.1520E+04, 2.1545E+04, 2.1571E+04, 2.1596E+04, 2.1621E+04,
     + 2.1647E+04, 2.1672E+04, 2.1697E+04, 2.1722E+04, 2.1748E+04,
     +  2.1773E+04, 2.1798E+04, 2.1824E+04, 2.1849E+04, 2.1875E+04,
     + 2.1900E+04, 2.1926E+04, 2.1951E+04, 2.1976E+04, 2.2002E+04,
     +  2.2027E+04, 2.2053E+04, 2.2078E+04, 2.2104E+04, 2.2129E+04,
     + 2.2155E+04, 2.2181E+04, 2.2206E+04, 2.2232E+04, 2.2258E+04,
     +  2.2283E+04, 2.2309E+04, 2.2334E+04, 2.2360E+04, 2.2386E+04,
     + 2.2412E+04, 2.2437E+04, 2.2463E+04, 2.2489E+04, 2.2515E+04,
     +  2.2540E+04, 2.2566E+04, 2.2592E+04, 2.2618E+04, 2.2644E+04,
     + 2.2669E+04, 2.2695E+04, 2.2721E+04, 2.2747E+04, 2.2773E+04,
     +  2.2799E+04, 2.2825E+04, 2.2851E+04, 2.2877E+04, 2.2903E+04,
     + 2.2929E+04, 2.2955E+04, 2.2981E+04, 2.3007E+04, 2.3033E+04,
     +  2.3059E+04, 2.3085E+04, 2.3111E+04, 2.3137E+04, 2.3164E+04,
     + 2.3190E+04, 2.3216E+04, 2.3242E+04, 2.3268E+04, 2.3294E+04,
     +  2.3321E+04, 2.3347E+04, 2.3373E+04, 2.3399E+04, 2.3426E+04,
     + 2.3452E+04, 2.3478E+04, 2.3505E+04, 2.3531E+04, 2.3557E+04,
     +  2.3584E+04, 2.3610E+04, 2.3637E+04, 2.3663E+04, 2.3689E+04,
     + 2.3716E+04, 2.3742E+04, 2.3769E+04, 2.3795E+04, 2.3822E+04,
     +  2.3848E+04, 2.3875E+04, 2.3901E+04, 2.3928E+04, 2.3954E+04,
     + 2.3981E+04, 2.4008E+04, 2.4034E+04, 2.4061E+04, 2.4088E+04,
     +  2.4114E+04, 2.4141E+04, 2.4168E+04, 2.4194E+04, 2.4221E+04,
     + 2.4248E+04, 2.4274E+04, 2.4301E+04, 2.4328E+04, 2.4355E+04,
     +  2.4382E+04, 2.4408E+04, 2.4435E+04, 2.4462E+04, 2.4489E+04,
     + 2.4516E+04, 2.4543E+04, 2.4570E+04, 2.4597E+04, 2.4623E+04,
     +  2.4650E+04, 2.4677E+04, 2.4704E+04, 2.4731E+04, 2.4758E+04,
     + 2.4785E+04, 2.4812E+04, 2.4840E+04, 2.4867E+04, 2.4894E+04,
     +  2.4921E+04, 2.4948E+04, 2.4975E+04, 2.5002E+04, 2.5029E+04,
     + 2.5057E+04, 2.5084E+04, 2.5111E+04, 2.5138E+04, 2.5165E+04,
     +  2.5193E+04, 2.5220E+04, 2.5247E+04, 2.5274E+04, 2.5302E+04,
     + 2.5329E+04, 2.5356E+04, 2.5384E+04, 2.5411E+04, 2.5438E+04,
     +  2.5466E+04, 2.5493E+04, 2.5521E+04, 2.5548E+04, 2.5575E+04,
     + 2.5603E+04, 2.5630E+04, 2.5658E+04, 2.5685E+04, 2.5713E+04,
     +  2.5740E+04, 2.5768E+04, 2.5795E+04, 2.5823E+04, 2.5851E+04,
     + 2.5878E+04, 2.5906E+04, 2.5933E+04, 2.5961E+04, 2.5989E+04,
     +  2.6016E+04, 2.6044E+04, 2.6072E+04, 2.6100E+04, 2.6127E+04,
     + 2.6155E+04, 2.6183E+04, 2.6211E+04, 2.6238E+04, 2.6266E+04,
     +  2.6294E+04, 2.6322E+04, 2.6350E+04, 2.6378E+04, 2.6405E+04,
     + 2.6433E+04, 2.6461E+04, 2.6489E+04, 2.6517E+04, 2.6545E+04,
     +  2.6573E+04, 2.6601E+04, 2.6629E+04, 2.6657E+04, 2.6685E+04,
     + 2.6713E+04, 2.6741E+04, 2.6769E+04, 2.6797E+04, 2.6825E+04,
     +  2.6853E+04, 2.6881E+04, 2.6910E+04, 2.6938E+04, 2.6966E+04,
     + 2.6994E+04, 2.7022E+04, 2.7050E+04, 2.7079E+04, 2.7107E+04,
     +  2.7135E+04, 2.7163E+04, 2.7192E+04, 2.7220E+04, 2.7248E+04,
     + 2.7277E+04, 2.7305E+04, 2.7333E+04, 2.7362E+04, 2.7390E+04,
     +  2.7418E+04, 2.7447E+04, 2.7475E+04, 2.7504E+04, 2.7532E+04,
     + 2.7561E+04, 2.7589E+04, 2.7617E+04, 2.7646E+04, 2.7675E+04,
     +  2.7703E+04, 2.7732E+04, 2.7760E+04, 2.7789E+04, 2.7817E+04,
     + 2.7846E+04, 2.7875E+04, 2.7903E+04, 2.7932E+04, 2.7961E+04,
     +  2.7989E+04, 2.8018E+04, 2.8047E+04, 2.8075E+04, 2.8104E+04,
     + 2.8133E+04, 2.8162E+04, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --        28
      data (QofT( 3,J),J=1,2408)/ 1.8294E+00, 3.2061E+00, 4.6275E+00,
     +  6.1687E+00, 7.8752E+00, 9.7526E+00, 1.1787E+01, 1.3959E+01,
     + 1.6249E+01, 1.8639E+01, 2.1114E+01, 2.3661E+01, 2.6270E+01,
     +  2.8931E+01, 3.1638E+01, 3.4385E+01, 3.7166E+01, 3.9977E+01,
     + 4.2815E+01, 4.5677E+01, 5.1460E+01, 5.7311E+01, 6.3216E+01,
     +  6.9165E+01, 7.5152E+01, 8.1170E+01, 8.7215E+01, 9.3282E+01,
     + 9.9368E+01, 1.0547E+02, 1.1159E+02, 1.1772E+02, 1.2386E+02,
     +  1.3002E+02, 1.3618E+02, 1.4235E+02, 1.4853E+02, 1.5471E+02,
     + 1.6090E+02, 1.6709E+02, 1.7329E+02, 1.7950E+02, 1.8571E+02,
     +  1.9192E+02, 1.9814E+02, 2.0436E+02, 2.1058E+02, 2.1680E+02,
     + 2.2303E+02, 2.2926E+02, 2.3549E+02, 2.4173E+02, 2.4796E+02,
     +  2.5420E+02, 2.6044E+02, 2.6668E+02, 2.7293E+02, 2.7917E+02,
     + 2.8542E+02, 2.9167E+02, 2.9791E+02, 3.0416E+02, 3.1041E+02,
     +  3.1667E+02, 3.2292E+02, 3.2917E+02, 3.3543E+02, 3.4168E+02,
     + 3.4794E+02, 3.5420E+02, 3.6045E+02, 3.6671E+02, 3.7297E+02,
     +  3.7923E+02, 3.8550E+02, 3.9176E+02, 3.9802E+02, 4.0428E+02,
     + 4.1055E+02, 4.1681E+02, 4.2308E+02, 4.2934E+02, 4.3561E+02,
     +  4.4187E+02, 4.4814E+02, 4.5441E+02, 4.6068E+02, 4.6695E+02,
     + 4.7322E+02, 4.7949E+02, 4.8576E+02, 4.9204E+02, 4.9831E+02,
     +  5.0458E+02, 5.1086E+02, 5.1713E+02, 5.2341E+02, 5.2969E+02,
     + 5.3597E+02, 5.4225E+02, 5.4853E+02, 5.5481E+02, 5.6110E+02,
     +  5.6738E+02, 5.7367E+02, 5.7995E+02, 5.8625E+02, 5.9253E+02,
     + 5.9883E+02, 6.0512E+02, 6.1142E+02, 6.1771E+02, 6.2401E+02,
     +  6.3031E+02, 6.3661E+02, 6.4292E+02, 6.4922E+02, 6.5554E+02,
     + 6.6184E+02, 6.6816E+02, 6.7448E+02, 6.8079E+02, 6.8712E+02,
     +  6.9344E+02, 6.9977E+02, 7.0609E+02, 7.1243E+02, 7.1876E+02,
     + 7.2510E+02, 7.3144E+02, 7.3779E+02, 7.4414E+02, 7.5048E+02,
     +  7.5684E+02, 7.6320E+02, 7.6956E+02, 7.7593E+02, 7.8230E+02,
     + 7.8868E+02, 7.9505E+02, 8.0143E+02, 8.0782E+02, 8.1421E+02,
     +  8.2061E+02, 8.2701E+02, 8.3341E+02, 8.3982E+02, 8.4624E+02,
     + 8.5266E+02, 8.5908E+02, 8.6552E+02, 8.7195E+02, 8.7840E+02,
     +  8.8484E+02, 8.9129E+02, 8.9775E+02, 9.0422E+02, 9.1068E+02,
     + 9.1716E+02, 9.2364E+02, 9.3013E+02, 9.3662E+02, 9.4313E+02,
     +  9.4963E+02, 9.5615E+02, 9.6267E+02, 9.6920E+02, 9.7573E+02,
     + 9.8227E+02, 9.8882E+02, 9.9538E+02, 1.0019E+03, 1.0085E+03,
     +  1.0151E+03, 1.0217E+03, 1.0283E+03, 1.0349E+03, 1.0415E+03,
     + 1.0481E+03, 1.0547E+03, 1.0613E+03, 1.0680E+03, 1.0746E+03,
     +  1.0813E+03, 1.0879E+03, 1.0946E+03, 1.1013E+03, 1.1080E+03,
     + 1.1147E+03, 1.1214E+03, 1.1281E+03, 1.1348E+03, 1.1416E+03,
     +  1.1483E+03, 1.1551E+03, 1.1618E+03, 1.1686E+03, 1.1754E+03,
     + 1.1822E+03, 1.1890E+03, 1.1958E+03, 1.2026E+03, 1.2094E+03,
     +  1.2163E+03, 1.2231E+03, 1.2300E+03, 1.2368E+03, 1.2437E+03,
     + 1.2506E+03, 1.2575E+03, 1.2644E+03, 1.2713E+03, 1.2783E+03,
     +  1.2852E+03, 1.2921E+03, 1.2991E+03, 1.3061E+03, 1.3131E+03,
     + 1.3201E+03, 1.3271E+03, 1.3341E+03, 1.3411E+03, 1.3481E+03,
     +  1.3552E+03, 1.3622E+03, 1.3693E+03, 1.3764E+03, 1.3835E+03,
     + 1.3906E+03, 1.3977E+03, 1.4048E+03, 1.4120E+03, 1.4191E+03,
     +  1.4263E+03, 1.4334E+03, 1.4406E+03, 1.4478E+03, 1.4550E+03,
     + 1.4623E+03, 1.4695E+03, 1.4767E+03, 1.4840E+03, 1.4912E+03,
     +  1.4985E+03, 1.5058E+03, 1.5131E+03, 1.5204E+03, 1.5278E+03,
     + 1.5351E+03, 1.5425E+03, 1.5498E+03, 1.5572E+03, 1.5646E+03,
     +  1.5720E+03, 1.5794E+03, 1.5868E+03, 1.5943E+03, 1.6017E+03,
     + 1.6092E+03, 1.6167E+03, 1.6241E+03, 1.6316E+03, 1.6392E+03,
     +  1.6467E+03, 1.6542E+03, 1.6618E+03, 1.6694E+03, 1.6769E+03,
     + 1.6845E+03, 1.6921E+03, 1.6998E+03, 1.7074E+03, 1.7150E+03,
     +  1.7227E+03, 1.7304E+03, 1.7380E+03, 1.7457E+03, 1.7534E+03,
     + 1.7612E+03, 1.7689E+03, 1.7767E+03, 1.7844E+03, 1.7922E+03,
     +  1.8000E+03, 1.8078E+03, 1.8156E+03, 1.8234E+03, 1.8313E+03,
     + 1.8391E+03, 1.8470E+03, 1.8549E+03, 1.8628E+03, 1.8707E+03,
     +  1.8786E+03, 1.8866E+03, 1.8945E+03, 1.9025E+03, 1.9105E+03,
     + 1.9185E+03, 1.9265E+03, 1.9345E+03, 1.9426E+03, 1.9506E+03,
     +  1.9587E+03, 1.9668E+03, 1.9749E+03, 1.9830E+03, 1.9911E+03,
     + 1.9992E+03, 2.0074E+03, 2.0155E+03, 2.0237E+03, 2.0319E+03,
     +  2.0401E+03, 2.0483E+03, 2.0566E+03, 2.0648E+03, 2.0731E+03,
     + 2.0813E+03, 2.0896E+03, 2.0979E+03, 2.1063E+03, 2.1146E+03,
     +  2.1229E+03, 2.1313E+03, 2.1397E+03, 2.1481E+03, 2.1565E+03,
     + 2.1649E+03, 2.1733E+03, 2.1818E+03, 2.1903E+03, 2.1987E+03,
     +  2.2072E+03, 2.2157E+03, 2.2243E+03, 2.2328E+03, 2.2413E+03,
     + 2.2499E+03, 2.2585E+03, 2.2671E+03, 2.2757E+03, 2.2843E+03,
     +  2.2929E+03, 2.3016E+03, 2.3103E+03, 2.3190E+03, 2.3276E+03,
     + 2.3364E+03, 2.3451E+03, 2.3538E+03, 2.3626E+03, 2.3714E+03,
     +  2.3801E+03, 2.3890E+03, 2.3978E+03, 2.4066E+03, 2.4154E+03,
     + 2.4243E+03, 2.4332E+03, 2.4421E+03, 2.4510E+03, 2.4599E+03,
     +  2.4688E+03, 2.4778E+03, 2.4867E+03, 2.4957E+03, 2.5047E+03,
     + 2.5137E+03, 2.5227E+03, 2.5318E+03, 2.5408E+03, 2.5499E+03,
     +  2.5590E+03, 2.5681E+03, 2.5772E+03, 2.5863E+03, 2.5955E+03,
     + 2.6046E+03, 2.6138E+03, 2.6230E+03, 2.6322E+03, 2.6414E+03,
     +  2.6506E+03, 2.6599E+03, 2.6692E+03, 2.6784E+03, 2.6877E+03,
     + 2.6970E+03, 2.7064E+03, 2.7157E+03, 2.7251E+03, 2.7344E+03,
     +  2.7438E+03, 2.7532E+03, 2.7626E+03, 2.7721E+03, 2.7815E+03,
     + 2.7910E+03, 2.8004E+03, 2.8099E+03, 2.8194E+03, 2.8290E+03,
     +  2.8385E+03, 2.8480E+03, 2.8576E+03, 2.8672E+03, 2.8768E+03,
     + 2.8864E+03, 2.8960E+03, 2.9057E+03, 2.9153E+03, 2.9250E+03,
     +  2.9347E+03, 2.9444E+03, 2.9541E+03, 2.9639E+03, 2.9736E+03,
     + 2.9834E+03, 2.9932E+03, 3.0030E+03, 3.0128E+03, 3.0226E+03,
     +  3.0325E+03, 3.0423E+03, 3.0522E+03, 3.0621E+03, 3.0720E+03,
     + 3.0819E+03, 3.0919E+03, 3.1018E+03, 3.1118E+03, 3.1217E+03,
     +  3.1317E+03, 3.1418E+03, 3.1518E+03, 3.1618E+03, 3.1719E+03,
     + 3.1820E+03, 3.1921E+03, 3.2022E+03, 3.2123E+03, 3.2224E+03,
     +  3.2326E+03, 3.2427E+03, 3.2529E+03, 3.2631E+03, 3.2733E+03,
     + 3.2836E+03, 3.2938E+03, 3.3041E+03, 3.3143E+03, 3.3247E+03,
     +  3.3350E+03, 3.3453E+03, 3.3556E+03, 3.3660E+03, 3.3763E+03,
     + 3.3867E+03, 3.3971E+03, 3.4075E+03, 3.4180E+03, 3.4284E+03,
     +  3.4389E+03, 3.4494E+03, 3.4599E+03, 3.4704E+03, 3.4809E+03,
     + 3.4914E+03, 3.5020E+03, 3.5126E+03, 3.5232E+03, 3.5338E+03,
     +  3.5444E+03, 3.5550E+03, 3.5657E+03, 3.5763E+03, 3.5870E+03,
     + 3.5977E+03, 3.6084E+03, 3.6192E+03, 3.6299E+03, 3.6407E+03,
     +  3.6514E+03, 3.6622E+03, 3.6730E+03, 3.6839E+03, 3.6947E+03,
     + 3.7056E+03, 3.7164E+03, 3.7273E+03, 3.7382E+03, 3.7491E+03,
     +  3.7601E+03, 3.7710E+03, 3.7820E+03, 3.7930E+03, 3.8039E+03,
     + 3.8150E+03, 3.8260E+03, 3.8370E+03, 3.8481E+03, 3.8592E+03,
     +  3.8703E+03, 3.8814E+03, 3.8925E+03, 3.9036E+03, 3.9148E+03,
     + 3.9260E+03, 3.9371E+03, 3.9483E+03, 3.9596E+03, 3.9708E+03,
     +  3.9820E+03, 3.9933E+03, 4.0046E+03, 4.0159E+03, 4.0272E+03,
     + 4.0385E+03, 4.0498E+03, 4.0612E+03, 4.0726E+03, 4.0840E+03,
     +  4.0954E+03, 4.1068E+03, 4.1182E+03, 4.1297E+03, 4.1411E+03,
     + 4.1527E+03, 4.1642E+03, 4.1757E+03, 4.1872E+03, 4.1987E+03,
     +  4.2103E+03, 4.2219E+03, 4.2335E+03, 4.2451E+03, 4.2567E+03,
     + 4.2684E+03, 4.2800E+03, 4.2917E+03, 4.3034E+03, 4.3151E+03,
     +  4.3268E+03, 4.3385E+03, 4.3503E+03, 4.3621E+03, 4.3739E+03,
     + 4.3856E+03, 4.3975E+03, 4.4093E+03, 4.4211E+03, 4.4330E+03,
     +  4.4449E+03, 4.4568E+03, 4.4687E+03, 4.4806E+03, 4.4926E+03,
     + 4.5045E+03, 4.5165E+03, 4.5285E+03, 4.5405E+03, 4.5525E+03,
     +  4.5646E+03, 4.5766E+03, 4.5887E+03, 4.6008E+03, 4.6129E+03,
     + 4.6250E+03, 4.6371E+03, 4.6493E+03, 4.6614E+03, 4.6736E+03,
     +  4.6858E+03, 4.6980E+03, 4.7103E+03, 4.7225E+03, 4.7348E+03,
     + 4.7471E+03, 4.7594E+03, 4.7716E+03, 4.7840E+03, 4.7963E+03,
     +  4.8087E+03, 4.8211E+03, 4.8335E+03, 4.8459E+03, 4.8583E+03,
     + 4.8707E+03, 4.8831E+03, 4.8956E+03, 4.9081E+03, 4.9206E+03,
     +  4.9331E+03, 4.9457E+03, 4.9582E+03, 4.9708E+03, 4.9833E+03,
     + 4.9960E+03, 5.0085E+03, 5.0212E+03, 5.0338E+03, 5.0465E+03,
     +  5.0592E+03, 5.0719E+03, 5.0846E+03, 5.0973E+03, 5.1100E+03,
     + 5.1228E+03, 5.1356E+03, 5.1483E+03, 5.1611E+03, 5.1739E+03,
     +  5.1868E+03, 5.1996E+03, 5.2125E+03, 5.2254E+03, 5.2383E+03,
     + 5.2512E+03, 5.2642E+03, 5.2771E+03, 5.2901E+03, 5.3031E+03,
     +  5.3160E+03, 5.3290E+03, 5.3421E+03, 5.3551E+03, 5.3682E+03,
     + 5.3812E+03, 5.3943E+03, 5.4074E+03, 5.4205E+03, 5.4337E+03,
     +  5.4469E+03, 5.4600E+03, 5.4732E+03, 5.4864E+03, 5.4996E+03,
     + 5.5129E+03, 5.5261E+03, 5.5394E+03, 5.5526E+03, 5.5660E+03,
     +  5.5793E+03, 5.5926E+03, 5.6059E+03, 5.6193E+03, 5.6327E+03,
     + 5.6461E+03, 5.6595E+03, 5.6729E+03, 5.6864E+03, 5.6998E+03,
     +  5.7133E+03, 5.7268E+03, 5.7403E+03, 5.7538E+03, 5.7674E+03,
     + 5.7809E+03, 5.7945E+03, 5.8081E+03, 5.8217E+03, 5.8353E+03,
     +  5.8489E+03, 5.8626E+03, 5.8762E+03, 5.8899E+03, 5.9036E+03,
     + 5.9173E+03, 5.9311E+03, 5.9448E+03, 5.9586E+03, 5.9724E+03,
     +  5.9862E+03, 6.0000E+03, 6.0138E+03, 6.0276E+03, 6.0415E+03,
     + 6.0554E+03, 6.0693E+03, 6.0831E+03, 6.0971E+03, 6.1110E+03,
     +  6.1250E+03, 6.1390E+03, 6.1529E+03, 6.1670E+03, 6.1810E+03,
     + 6.1950E+03, 6.2090E+03, 6.2231E+03, 6.2372E+03, 6.2513E+03,
     +  6.2654E+03, 6.2796E+03, 6.2937E+03, 6.3079E+03, 6.3220E+03,
     + 6.3362E+03, 6.3505E+03, 6.3647E+03, 6.3789E+03, 6.3932E+03,
     +  6.4075E+03, 6.4218E+03, 6.4361E+03, 6.4504E+03, 6.4647E+03,
     + 6.4791E+03, 6.4935E+03, 6.5078E+03, 6.5223E+03, 6.5367E+03,
     +  6.5511E+03, 6.5656E+03, 6.5800E+03, 6.5945E+03, 6.6090E+03,
     + 6.6235E+03, 6.6380E+03, 6.6526E+03, 6.6672E+03, 6.6817E+03,
     +  6.6964E+03, 6.7110E+03, 6.7256E+03, 6.7402E+03, 6.7549E+03,
     + 6.7696E+03, 6.7843E+03, 6.7990E+03, 6.8137E+03, 6.8285E+03,
     +  6.8432E+03, 6.8580E+03, 6.8728E+03, 6.8876E+03, 6.9024E+03,
     + 6.9173E+03, 6.9321E+03, 6.9470E+03, 6.9619E+03, 6.9768E+03,
     +  6.9917E+03, 7.0066E+03, 7.0216E+03, 7.0365E+03, 7.0515E+03,
     + 7.0665E+03, 7.0815E+03, 7.0966E+03, 7.1116E+03, 7.1267E+03,
     +  7.1418E+03, 7.1569E+03, 7.1720E+03, 7.1871E+03, 7.2022E+03,
     + 7.2174E+03, 7.2326E+03, 7.2478E+03, 7.2630E+03, 7.2782E+03,
     +  7.2934E+03, 7.3087E+03, 7.3240E+03, 7.3393E+03, 7.3546E+03,
     + 7.3699E+03, 7.3852E+03, 7.4006E+03, 7.4159E+03, 7.4314E+03,
     +  7.4468E+03, 7.4622E+03, 7.4776E+03, 7.4931E+03, 7.5085E+03,
     + 7.5240E+03, 7.5395E+03, 7.5550E+03, 7.5705E+03, 7.5861E+03,
     +  7.6017E+03, 7.6173E+03, 7.6328E+03, 7.6485E+03, 7.6641E+03,
     + 7.6797E+03, 7.6954E+03, 7.7111E+03, 7.7268E+03, 7.7424E+03,
     +  7.7582E+03, 7.7739E+03, 7.7897E+03, 7.8054E+03, 7.8212E+03,
     + 7.8371E+03, 7.8529E+03, 7.8687E+03, 7.8845E+03, 7.9004E+03,
     +  7.9163E+03, 7.9322E+03, 7.9481E+03, 7.9641E+03, 7.9800E+03,
     + 7.9960E+03, 8.0119E+03, 8.0280E+03, 8.0440E+03, 8.0600E+03,
     +  8.0761E+03, 8.0921E+03, 8.1082E+03, 8.1243E+03, 8.1404E+03,
     + 8.1565E+03, 8.1727E+03, 8.1889E+03, 8.2050E+03, 8.2212E+03,
     +  8.2374E+03, 8.2536E+03, 8.2699E+03, 8.2861E+03, 8.3024E+03,
     + 8.3187E+03, 8.3350E+03, 8.3513E+03, 8.3677E+03, 8.3840E+03,
     +  8.4004E+03, 8.4168E+03, 8.4332E+03, 8.4496E+03, 8.4660E+03,
     + 8.4825E+03, 8.4989E+03, 8.5154E+03, 8.5319E+03, 8.5485E+03,
     +  8.5650E+03, 8.5815E+03, 8.5981E+03, 8.6147E+03, 8.6313E+03,
     + 8.6479E+03, 8.6645E+03, 8.6812E+03, 8.6978E+03, 8.7145E+03,
     +  8.7312E+03, 8.7479E+03, 8.7646E+03, 8.7813E+03, 8.7981E+03,
     + 8.8149E+03, 8.8317E+03, 8.8485E+03, 8.8653E+03, 8.8821E+03,
     +  8.8990E+03, 8.9159E+03, 8.9328E+03, 8.9497E+03, 8.9666E+03,
     + 8.9835E+03, 9.0005E+03, 9.0175E+03, 9.0344E+03, 9.0514E+03,
     +  9.0684E+03, 9.0855E+03, 9.1025E+03, 9.1196E+03, 9.1367E+03,
     + 9.1538E+03, 9.1709E+03, 9.1880E+03, 9.2052E+03, 9.2223E+03,
     +  9.2395E+03, 9.2567E+03, 9.2739E+03, 9.2911E+03, 9.3084E+03,
     + 9.3256E+03, 9.3429E+03, 9.3602E+03, 9.3775E+03, 9.3948E+03,
     +  9.4121E+03, 9.4295E+03, 9.4469E+03, 9.4643E+03, 9.4817E+03,
     + 9.4991E+03, 9.5165E+03, 9.5340E+03, 9.5515E+03, 9.5689E+03,
     +  9.5864E+03, 9.6040E+03, 9.6215E+03, 9.6390E+03, 9.6566E+03,
     + 9.6742E+03, 9.6918E+03, 9.7094E+03, 9.7270E+03, 9.7447E+03,
     +  9.7624E+03, 9.7800E+03, 9.7977E+03, 9.8154E+03, 9.8331E+03,
     + 9.8509E+03, 9.8687E+03, 9.8864E+03, 9.9042E+03, 9.9221E+03,
     +  9.9399E+03, 9.9577E+03, 9.9755E+03, 9.9935E+03, 1.0011E+04,
     + 1.0029E+04, 1.0047E+04, 1.0065E+04, 1.0083E+04, 1.0101E+04,
     +  1.0119E+04, 1.0137E+04, 1.0155E+04, 1.0173E+04, 1.0191E+04,
     + 1.0209E+04, 1.0227E+04, 1.0245E+04, 1.0264E+04, 1.0282E+04,
     +  1.0300E+04, 1.0318E+04, 1.0336E+04, 1.0354E+04, 1.0372E+04,
     + 1.0391E+04, 1.0409E+04, 1.0427E+04, 1.0446E+04, 1.0464E+04,
     +  1.0482E+04, 1.0500E+04, 1.0519E+04, 1.0537E+04, 1.0556E+04,
     + 1.0574E+04, 1.0592E+04, 1.0611E+04, 1.0629E+04, 1.0648E+04,
     +  1.0666E+04, 1.0685E+04, 1.0703E+04, 1.0722E+04, 1.0740E+04,
     + 1.0759E+04, 1.0777E+04, 1.0796E+04, 1.0815E+04, 1.0833E+04,
     +  1.0852E+04, 1.0870E+04, 1.0889E+04, 1.0908E+04, 1.0926E+04,
     + 1.0945E+04, 1.0964E+04, 1.0983E+04, 1.1001E+04, 1.1020E+04,
     +  1.1039E+04, 1.1058E+04, 1.1077E+04, 1.1095E+04, 1.1114E+04,
     + 1.1133E+04, 1.1152E+04, 1.1171E+04, 1.1190E+04, 1.1209E+04,
     +  1.1228E+04, 1.1247E+04, 1.1266E+04, 1.1285E+04, 1.1304E+04,
     + 1.1323E+04, 1.1342E+04, 1.1361E+04, 1.1380E+04, 1.1399E+04,
     +  1.1418E+04, 1.1437E+04, 1.1457E+04, 1.1476E+04, 1.1495E+04,
     + 1.1514E+04, 1.1533E+04, 1.1553E+04, 1.1572E+04, 1.1591E+04,
     +  1.1610E+04, 1.1630E+04, 1.1649E+04, 1.1668E+04, 1.1688E+04,
     + 1.1707E+04, 1.1726E+04, 1.1746E+04, 1.1765E+04, 1.1785E+04,
     +  1.1804E+04, 1.1824E+04, 1.1843E+04, 1.1863E+04, 1.1882E+04,
     + 1.1902E+04, 1.1921E+04, 1.1941E+04, 1.1960E+04, 1.1980E+04,
     +  1.1999E+04, 1.2019E+04, 1.2039E+04, 1.2058E+04, 1.2078E+04,
     + 1.2098E+04, 1.2117E+04, 1.2137E+04, 1.2157E+04, 1.2177E+04,
     +  1.2196E+04, 1.2216E+04, 1.2236E+04, 1.2256E+04, 1.2276E+04,
     + 1.2295E+04, 1.2315E+04, 1.2335E+04, 1.2355E+04, 1.2375E+04,
     +  1.2395E+04, 1.2415E+04, 1.2435E+04, 1.2455E+04, 1.2475E+04,
     + 1.2495E+04, 1.2515E+04, 1.2535E+04, 1.2555E+04, 1.2575E+04,
     +  1.2595E+04, 1.2615E+04, 1.2635E+04, 1.2656E+04, 1.2676E+04,
     + 1.2696E+04, 1.2716E+04, 1.2736E+04, 1.2756E+04, 1.2777E+04,
     +  1.2797E+04, 1.2817E+04, 1.2838E+04, 1.2858E+04, 1.2878E+04,
     + 1.2898E+04, 1.2919E+04, 1.2939E+04, 1.2960E+04, 1.2980E+04,
     +  1.3000E+04, 1.3021E+04, 1.3041E+04, 1.3062E+04, 1.3082E+04,
     + 1.3103E+04, 1.3123E+04, 1.3144E+04, 1.3164E+04, 1.3185E+04,
     +  1.3205E+04, 1.3226E+04, 1.3247E+04, 1.3267E+04, 1.3288E+04,
     + 1.3309E+04, 1.3329E+04, 1.3350E+04, 1.3371E+04, 1.3391E+04,
     +  1.3412E+04, 1.3433E+04, 1.3454E+04, 1.3474E+04, 1.3495E+04,
     + 1.3516E+04, 1.3537E+04, 1.3558E+04, 1.3579E+04, 1.3599E+04,
     +  1.3620E+04, 1.3641E+04, 1.3662E+04, 1.3683E+04, 1.3704E+04,
     + 1.3725E+04, 1.3746E+04, 1.3767E+04, 1.3788E+04, 1.3809E+04,
     +  1.3830E+04, 1.3851E+04, 1.3872E+04, 1.3894E+04, 1.3915E+04,
     + 1.3936E+04, 1.3957E+04, 1.3978E+04, 1.3999E+04, 1.4021E+04,
     +  1.4042E+04, 1.4063E+04, 1.4084E+04, 1.4106E+04, 1.4127E+04,
     + 1.4148E+04, 1.4170E+04, 1.4191E+04, 1.4212E+04, 1.4234E+04,
     +  1.4255E+04, 1.4276E+04, 1.4298E+04, 1.4319E+04, 1.4341E+04,
     + 1.4362E+04, 1.4384E+04, 1.4405E+04, 1.4427E+04, 1.4448E+04,
     +  1.4470E+04, 1.4491E+04, 1.4513E+04, 1.4534E+04, 1.4556E+04,
     + 1.4578E+04, 1.4599E+04, 1.4621E+04, 1.4643E+04, 1.4664E+04,
     +  1.4686E+04, 1.4708E+04, 1.4730E+04, 1.4751E+04, 1.4773E+04,
     + 1.4795E+04, 1.4817E+04, 1.4838E+04, 1.4860E+04, 1.4882E+04,
     +  1.4904E+04, 1.4926E+04, 1.4948E+04, 1.4970E+04, 1.4992E+04,
     + 1.5014E+04, 1.5036E+04, 1.5058E+04, 1.5080E+04, 1.5102E+04,
     +  1.5124E+04, 1.5146E+04, 1.5168E+04, 1.5190E+04, 1.5212E+04,
     + 1.5234E+04, 1.5256E+04, 1.5278E+04, 1.5301E+04, 1.5323E+04,
     +  1.5345E+04, 1.5367E+04, 1.5389E+04, 1.5412E+04, 1.5434E+04,
     + 1.5456E+04, 1.5478E+04, 1.5501E+04, 1.5523E+04, 1.5545E+04,
     +  1.5568E+04, 1.5590E+04, 1.5613E+04, 1.5635E+04, 1.5657E+04,
     + 1.5680E+04, 1.5702E+04, 1.5725E+04, 1.5747E+04, 1.5770E+04,
     +  1.5792E+04, 1.5815E+04, 1.5837E+04, 1.5860E+04, 1.5882E+04,
     + 1.5905E+04, 1.5928E+04, 1.5950E+04, 1.5973E+04, 1.5996E+04,
     +  1.6018E+04, 1.6041E+04, 1.6064E+04, 1.6086E+04, 1.6109E+04,
     + 1.6132E+04, 1.6155E+04, 1.6177E+04, 1.6200E+04, 1.6223E+04,
     +  1.6246E+04, 1.6269E+04, 1.6292E+04, 1.6315E+04, 1.6337E+04,
     + 1.6360E+04, 1.6383E+04, 1.6406E+04, 1.6429E+04, 1.6452E+04,
     +  1.6475E+04, 1.6498E+04, 1.6521E+04, 1.6544E+04, 1.6567E+04,
     + 1.6591E+04, 1.6614E+04, 1.6637E+04, 1.6660E+04, 1.6683E+04,
     +  1.6706E+04, 1.6729E+04, 1.6753E+04, 1.6776E+04, 1.6799E+04,
     + 1.6822E+04, 1.6845E+04, 1.6869E+04, 1.6892E+04, 1.6915E+04,
     +  1.6939E+04, 1.6962E+04, 1.6985E+04, 1.7009E+04, 1.7032E+04,
     + 1.7056E+04, 1.7079E+04, 1.7102E+04, 1.7126E+04, 1.7149E+04,
     +  1.7173E+04, 1.7196E+04, 1.7220E+04, 1.7243E+04, 1.7267E+04,
     + 1.7291E+04, 1.7314E+04, 1.7338E+04, 1.7361E+04, 1.7385E+04,
     +  1.7409E+04, 1.7432E+04, 1.7456E+04, 1.7480E+04, 1.7503E+04,
     + 1.7527E+04, 1.7551E+04, 1.7575E+04, 1.7598E+04, 1.7622E+04,
     +  1.7646E+04, 1.7670E+04, 1.7694E+04, 1.7717E+04, 1.7741E+04,
     + 1.7765E+04, 1.7789E+04, 1.7813E+04, 1.7837E+04, 1.7861E+04,
     +  1.7885E+04, 1.7909E+04, 1.7933E+04, 1.7957E+04, 1.7981E+04,
     + 1.8005E+04, 1.8029E+04, 1.8053E+04, 1.8077E+04, 1.8101E+04,
     +  1.8125E+04, 1.8150E+04, 1.8174E+04, 1.8198E+04, 1.8222E+04,
     + 1.8246E+04, 1.8271E+04, 1.8295E+04, 1.8319E+04, 1.8343E+04,
     +  1.8368E+04, 1.8392E+04, 1.8416E+04, 1.8441E+04, 1.8465E+04,
     + 1.8489E+04, 1.8514E+04, 1.8538E+04, 1.8563E+04, 1.8587E+04,
     +  1.8611E+04, 1.8636E+04, 1.8660E+04, 1.8685E+04, 1.8709E+04,
     + 1.8734E+04, 1.8758E+04, 1.8783E+04, 1.8808E+04, 1.8832E+04,
     +  1.8857E+04, 1.8881E+04, 1.8906E+04, 1.8931E+04, 1.8955E+04,
     + 1.8980E+04, 1.9005E+04, 1.9030E+04, 1.9054E+04, 1.9079E+04,
     +  1.9104E+04, 1.9129E+04, 1.9153E+04, 1.9178E+04, 1.9203E+04,
     + 1.9228E+04, 1.9253E+04, 1.9278E+04, 1.9303E+04, 1.9328E+04,
     +  1.9352E+04, 1.9377E+04, 1.9402E+04, 1.9427E+04, 1.9452E+04,
     + 1.9477E+04, 1.9502E+04, 1.9527E+04, 1.9552E+04, 1.9578E+04,
     +  1.9603E+04, 1.9628E+04, 1.9653E+04, 1.9678E+04, 1.9703E+04,
     + 1.9728E+04, 1.9754E+04, 1.9779E+04, 1.9804E+04, 1.9829E+04,
     +  1.9854E+04, 1.9880E+04, 1.9905E+04, 1.9930E+04, 1.9956E+04,
     + 1.9981E+04, 2.0006E+04, 2.0032E+04, 2.0057E+04, 2.0082E+04,
     +  2.0108E+04, 2.0133E+04, 2.0159E+04, 2.0184E+04, 2.0210E+04,
     + 2.0235E+04, 2.0261E+04, 2.0286E+04, 2.0312E+04, 2.0337E+04,
     +  2.0363E+04, 2.0388E+04, 2.0414E+04, 2.0440E+04, 2.0465E+04,
     + 2.0491E+04, 2.0517E+04, 2.0542E+04, 2.0568E+04, 2.0594E+04,
     +  2.0619E+04, 2.0645E+04, 2.0671E+04, 2.0697E+04, 2.0723E+04,
     + 2.0748E+04, 2.0774E+04, 2.0800E+04, 2.0826E+04, 2.0852E+04,
     +  2.0878E+04, 2.0904E+04, 2.0930E+04, 2.0955E+04, 2.0981E+04,
     + 2.1007E+04, 2.1033E+04, 2.1059E+04, 2.1085E+04, 2.1111E+04,
     +  2.1137E+04, 2.1164E+04, 2.1190E+04, 2.1216E+04, 2.1242E+04,
     + 2.1268E+04, 2.1294E+04, 2.1320E+04, 2.1346E+04, 2.1373E+04,
     +  2.1399E+04, 2.1425E+04, 2.1451E+04, 2.1478E+04, 2.1504E+04,
     + 2.1530E+04, 2.1556E+04, 2.1583E+04, 2.1609E+04, 2.1635E+04,
     +  2.1662E+04, 2.1688E+04, 2.1715E+04, 2.1741E+04, 2.1767E+04,
     + 2.1794E+04, 2.1820E+04, 2.1847E+04, 2.1873E+04, 2.1900E+04,
     +  2.1926E+04, 2.1953E+04, 2.1979E+04, 2.2006E+04, 2.2033E+04,
     + 2.2059E+04, 2.2086E+04, 2.2113E+04, 2.2139E+04, 2.2166E+04,
     +  2.2193E+04, 2.2219E+04, 2.2246E+04, 2.2273E+04, 2.2299E+04,
     + 2.2326E+04, 2.2353E+04, 2.2380E+04, 2.2407E+04, 2.2433E+04,
     +  2.2460E+04, 2.2487E+04, 2.2514E+04, 2.2541E+04, 2.2568E+04,
     + 2.2595E+04, 2.2622E+04, 2.2649E+04, 2.2676E+04, 2.2703E+04,
     +  2.2730E+04, 2.2757E+04, 2.2784E+04, 2.2811E+04, 2.2838E+04,
     + 2.2865E+04, 2.2892E+04, 2.2919E+04, 2.2946E+04, 2.2973E+04,
     +  2.3000E+04, 2.3028E+04, 2.3055E+04, 2.3082E+04, 2.3109E+04,
     + 2.3137E+04, 2.3164E+04, 2.3191E+04, 2.3218E+04, 2.3246E+04,
     +  2.3273E+04, 2.3300E+04, 2.3328E+04, 2.3355E+04, 2.3382E+04,
     + 2.3410E+04, 2.3437E+04, 2.3465E+04, 2.3492E+04, 2.3520E+04,
     +  2.3547E+04, 2.3575E+04, 2.3602E+04, 2.3630E+04, 2.3657E+04,
     + 2.3685E+04, 2.3712E+04, 2.3740E+04, 2.3767E+04, 2.3795E+04,
     +  2.3823E+04, 2.3850E+04, 2.3878E+04, 2.3906E+04, 2.3933E+04,
     + 2.3961E+04, 2.3989E+04, 2.4017E+04, 2.4044E+04, 2.4072E+04,
     +  2.4100E+04, 2.4128E+04, 2.4156E+04, 2.4183E+04, 2.4211E+04,
     + 2.4239E+04, 2.4267E+04, 2.4295E+04, 2.4323E+04, 2.4351E+04,
     +  2.4379E+04, 2.4407E+04, 2.4435E+04, 2.4463E+04, 2.4491E+04,
     + 2.4519E+04, 2.4547E+04, 2.4575E+04, 2.4603E+04, 2.4631E+04,
     +  2.4659E+04, 2.4687E+04, 2.4715E+04, 2.4744E+04, 2.4772E+04,
     + 2.4800E+04, 2.4828E+04, 2.4856E+04, 2.4885E+04, 2.4913E+04,
     +  2.4941E+04, 2.4969E+04, 2.4998E+04, 2.5026E+04, 2.5054E+04,
     + 2.5083E+04, 2.5111E+04, 2.5139E+04, 2.5168E+04, 2.5196E+04,
     +  2.5225E+04, 2.5253E+04, 2.5282E+04, 2.5310E+04, 2.5338E+04,
     + 2.5367E+04, 2.5396E+04, 2.5424E+04, 2.5453E+04, 2.5481E+04,
     +  2.5510E+04, 2.5538E+04, 2.5567E+04, 2.5596E+04, 2.5624E+04,
     + 2.5653E+04, 2.5682E+04, 2.5710E+04, 2.5739E+04, 2.5768E+04,
     +  2.5796E+04, 2.5825E+04, 2.5854E+04, 2.5883E+04, 2.5911E+04,
     + 2.5940E+04, 2.5969E+04, 2.5998E+04, 2.6027E+04, 2.6056E+04,
     +  2.6085E+04, 2.6114E+04, 2.6142E+04, 2.6171E+04, 2.6200E+04,
     + 2.6229E+04, 2.6258E+04, 2.6287E+04, 2.6316E+04, 2.6345E+04,
     +  2.6374E+04, 2.6403E+04, 2.6433E+04, 2.6462E+04, 2.6491E+04,
     + 2.6520E+04, 2.6549E+04, 2.6578E+04, 2.6607E+04, 2.6637E+04,
     +  2.6666E+04, 2.6695E+04, 2.6724E+04, 2.6753E+04, 2.6783E+04,
     + 2.6812E+04, 2.6841E+04, 2.6871E+04, 2.6900E+04, 2.6929E+04,
     +  2.6959E+04, 2.6988E+04, 2.7017E+04, 2.7047E+04, 2.7076E+04,
     + 2.7106E+04, 2.7135E+04, 2.7165E+04, 2.7194E+04, 2.7224E+04,
     +  2.7253E+04, 2.7283E+04, 2.7312E+04, 2.7342E+04, 2.7371E+04,
     + 2.7401E+04, 2.7430E+04, 2.7460E+04, 2.7490E+04, 2.7519E+04,
     +  2.7549E+04, 2.7579E+04, 2.7608E+04, 2.7638E+04, 2.7668E+04,
     + 2.7698E+04, 2.7727E+04, 2.7757E+04, 2.7787E+04, 2.7817E+04,
     +  2.7847E+04, 2.7876E+04, 2.7906E+04, 2.7936E+04, 2.7966E+04,
     + 2.7996E+04, 2.8026E+04, 2.8056E+04, 2.8086E+04, 2.8116E+04,
     +  2.8146E+04, 2.8176E+04, 2.8206E+04, 2.8236E+04, 2.8266E+04,
     + 2.8296E+04, 2.8326E+04, 2.8356E+04, 2.8386E+04, 2.8416E+04,
     +  2.8446E+04, 2.8476E+04, 2.8507E+04, 2.8537E+04, 2.8567E+04,
     + 2.8597E+04, 2.8627E+04, 2.8658E+04, 2.8688E+04, 2.8718E+04,
     +  2.8748E+04, 2.8779E+04, 2.8809E+04, 2.8839E+04, 2.8870E+04,
     + 2.8900E+04, 2.8931E+04, 2.8961E+04, 2.8991E+04, 2.9022E+04,
     +  2.9052E+04, 2.9083E+04, 2.9113E+04, 2.9144E+04, 2.9174E+04,
     + 2.9205E+04, 2.9235E+04, 2.9266E+04, 2.9296E+04, 2.9327E+04,
     +  2.9358E+04, 2.9388E+04, 2.9419E+04, 2.9449E+04, 2.9480E+04,
     + 2.9511E+04, 2.9541E+04, 2.9572E+04, 2.9603E+04, 2.9634E+04,
     +  2.9664E+04, 2.9695E+04, 2.9726E+04, 2.9757E+04, 2.9788E+04,
     + 2.9818E+04, 2.9849E+04, 2.9880E+04, 2.9911E+04, 2.9942E+04,
     +  2.9973E+04, 3.0004E+04, 3.0034E+04, 3.0065E+04, 3.0096E+04,
     + 3.0127E+04, 3.0158E+04, 3.0189E+04, 3.0220E+04, 3.0251E+04,
     +  3.0283E+04, 3.0314E+04, 3.0345E+04, 3.0376E+04, 3.0407E+04,
     + 3.0438E+04, 3.0469E+04, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 3,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_C3H4  (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 1),QofT( 1,NT),Q(NT)
      data Tmx(1)/3500./ 
      data xgj/ 2./
c...     C3H4
c...        --      1221
      data (QofT( 1,J),J=1,2408)/ 2.2424E+01, 4.2711E+01, 6.6063E+01,
     +  9.2748E+01, 1.2223E+02, 1.5401E+02, 1.8781E+02, 2.2341E+02,
     + 2.6072E+02, 2.9965E+02, 3.4016E+02, 3.8222E+02, 4.2580E+02,
     +  4.7089E+02, 5.1748E+02, 5.6557E+02, 6.1513E+02, 6.6615E+02,
     + 7.1863E+02, 7.7255E+02, 8.8462E+02, 1.0022E+03, 1.1252E+03,
     +  1.2533E+03, 1.3864E+03, 1.5244E+03, 1.6670E+03, 1.8142E+03,
     + 1.9657E+03, 2.1214E+03, 2.2813E+03, 2.4451E+03, 2.6129E+03,
     +  2.7844E+03, 2.9597E+03, 3.1387E+03, 3.3212E+03, 3.5074E+03,
     + 3.6970E+03, 3.8902E+03, 4.0868E+03, 4.2869E+03, 4.4904E+03,
     +  4.6974E+03, 4.9079E+03, 5.1219E+03, 5.3394E+03, 5.5604E+03,
     + 5.7849E+03, 6.0131E+03, 6.2449E+03, 6.4804E+03, 6.7196E+03,
     +  6.9627E+03, 7.2096E+03, 7.4605E+03, 7.7153E+03, 7.9741E+03,
     + 8.2371E+03, 8.5043E+03, 8.7757E+03, 9.0514E+03, 9.3315E+03,
     +  9.6163E+03, 9.9055E+03, 1.0199E+04, 1.0498E+04, 1.0802E+04,
     + 1.1110E+04, 1.1424E+04, 1.1742E+04, 1.2066E+04, 1.2395E+04,
     +  1.2730E+04, 1.3070E+04, 1.3416E+04, 1.3767E+04, 1.4124E+04,
     + 1.4487E+04, 1.4857E+04, 1.5232E+04, 1.5614E+04, 1.6002E+04,
     +  1.6396E+04, 1.6797E+04, 1.7205E+04, 1.7620E+04, 1.8041E+04,
     + 1.8470E+04, 1.8906E+04, 1.9350E+04, 1.9801E+04, 2.0259E+04,
     +  2.0725E+04, 2.1199E+04, 2.1681E+04, 2.2172E+04, 2.2670E+04,
     + 2.3178E+04, 2.3693E+04, 2.4218E+04, 2.4751E+04, 2.5294E+04,
     +  2.5845E+04, 2.6407E+04, 2.6977E+04, 2.7558E+04, 2.8148E+04,
     + 2.8748E+04, 2.9359E+04, 2.9980E+04, 3.0611E+04, 3.1253E+04,
     +  3.1907E+04, 3.2571E+04, 3.3247E+04, 3.3934E+04, 3.4633E+04,
     + 3.5344E+04, 3.6067E+04, 3.6802E+04, 3.7550E+04, 3.8311E+04,
     +  3.9084E+04, 3.9871E+04, 4.0671E+04, 4.1485E+04, 4.2313E+04,
     + 4.3155E+04, 4.4011E+04, 4.4881E+04, 4.5767E+04, 4.6668E+04,
     +  4.7584E+04, 4.8515E+04, 4.9462E+04, 5.0426E+04, 5.1405E+04,
     + 5.2402E+04, 5.3415E+04, 5.4446E+04, 5.5494E+04, 5.6560E+04,
     +  5.7643E+04, 5.8746E+04, 5.9867E+04, 6.1007E+04, 6.2166E+04,
     + 6.3345E+04, 6.4543E+04, 6.5763E+04, 6.7002E+04, 6.8263E+04,
     +  6.9545E+04, 7.0849E+04, 7.2176E+04, 7.3525E+04, 7.4897E+04,
     + 7.6292E+04, 7.7711E+04, 7.9154E+04, 8.0621E+04, 8.2113E+04,
     +  8.3630E+04, 8.5173E+04, 8.6741E+04, 8.8336E+04, 8.9958E+04,
     + 9.1607E+04, 9.3284E+04, 9.4989E+04, 9.6723E+04, 9.8486E+04,
     +  1.0028E+05, 1.0210E+05, 1.0395E+05, 1.0584E+05, 1.0775E+05,
     + 1.0970E+05, 1.1168E+05, 1.1370E+05, 1.1574E+05, 1.1782E+05,
     +  1.1994E+05, 1.2209E+05, 1.2428E+05, 1.2650E+05, 1.2877E+05,
     + 1.3106E+05, 1.3340E+05, 1.3578E+05, 1.3819E+05, 1.4065E+05,
     +  1.4314E+05, 1.4568E+05, 1.4826E+05, 1.5088E+05, 1.5355E+05,
     + 1.5626E+05, 1.5901E+05, 1.6181E+05, 1.6466E+05, 1.6755E+05,
     +  1.7050E+05, 1.7349E+05, 1.7652E+05, 1.7961E+05, 1.8275E+05,
     + 1.8595E+05, 1.8919E+05, 1.9249E+05, 1.9584E+05, 1.9925E+05,
     +  2.0271E+05, 2.0623E+05, 2.0981E+05, 2.1344E+05, 2.1714E+05,
     + 2.2089E+05, 2.2471E+05, 2.2859E+05, 2.3253E+05, 2.3654E+05,
     +  2.4061E+05, 2.4475E+05, 2.4895E+05, 2.5323E+05, 2.5757E+05,
     + 2.6198E+05, 2.6647E+05, 2.7103E+05, 2.7566E+05, 2.8037E+05,
     +  2.8515E+05, 2.9001E+05, 2.9495E+05, 2.9997E+05, 3.0507E+05,
     + 3.1025E+05, 3.1552E+05, 3.2087E+05, 3.2630E+05, 3.3183E+05,
     +  3.3744E+05, 3.4314E+05, 3.4894E+05, 3.5482E+05, 3.6081E+05,
     + 3.6688E+05, 3.7306E+05, 3.7933E+05, 3.8570E+05, 3.9218E+05,
     +  3.9876E+05, 4.0544E+05, 4.1223E+05, 4.1913E+05, 4.2613E+05,
     + 4.3325E+05, 4.4048E+05, 4.4783E+05, 4.5529E+05, 4.6288E+05,
     +  4.7058E+05, 4.7840E+05, 4.8635E+05, 4.9442E+05, 5.0262E+05,
     + 5.1095E+05, 5.1941E+05, 5.2800E+05, 5.3673E+05, 5.4560E+05,
     +  5.5461E+05, 5.6375E+05, 5.7304E+05, 5.8248E+05, 5.9207E+05,
     + 6.0180E+05, 6.1169E+05, 6.2173E+05, 6.3193E+05, 6.4229E+05,
     +  6.5281E+05, 6.6349E+05, 6.7434E+05, 6.8536E+05, 6.9655E+05,
     + 7.0791E+05, 7.1946E+05, 7.3118E+05, 7.4308E+05, 7.5516E+05,
     +  7.6744E+05, 7.7990E+05, 7.9256E+05, 8.0541E+05, 8.1846E+05,
     + 8.3171E+05, 8.4517E+05, 8.5883E+05, 8.7271E+05, 8.8679E+05,
     +  9.0110E+05, 9.1562E+05, 9.3037E+05, 9.4534E+05, 9.6054E+05,
     + 9.7598E+05, 9.9165E+05, 1.0076E+06, 1.0237E+06, 1.0401E+06,
     +  1.0568E+06, 1.0737E+06, 1.0908E+06, 1.1083E+06, 1.1259E+06,
     + 1.1439E+06, 1.1621E+06, 1.1806E+06, 1.1994E+06, 1.2185E+06,
     +  1.2379E+06, 1.2575E+06, 1.2775E+06, 1.2977E+06, 1.3183E+06,
     + 1.3392E+06, 1.3603E+06, 1.3818E+06, 1.4037E+06, 1.4258E+06,
     +  1.4483E+06, 1.4711E+06, 1.4943E+06, 1.5178E+06, 1.5417E+06,
     + 1.5659E+06, 1.5905E+06, 1.6155E+06, 1.6408E+06, 1.6665E+06,
     +  1.6926E+06, 1.7191E+06, 1.7460E+06, 1.7732E+06, 1.8009E+06,
     + 1.8290E+06, 1.8575E+06, 1.8865E+06, 1.9158E+06, 1.9456E+06,
     +  1.9759E+06, 2.0066E+06, 2.0377E+06, 2.0693E+06, 2.1014E+06,
     + 2.1339E+06, 2.1669E+06, 2.2004E+06, 2.2344E+06, 2.2689E+06,
     +  2.3040E+06, 2.3395E+06, 2.3755E+06, 2.4121E+06, 2.4492E+06,
     + 2.4869E+06, 2.5251E+06, 2.5638E+06, 2.6032E+06, 2.6431E+06,
     +  2.6836E+06, 2.7246E+06, 2.7663E+06, 2.8086E+06, 2.8515E+06,
     + 2.8950E+06, 2.9392E+06, 2.9840E+06, 3.0294E+06, 3.0755E+06,
     +  3.1223E+06, 3.1697E+06, 3.2179E+06, 3.2667E+06, 3.3163E+06,
     + 3.3665E+06, 3.4175E+06, 3.4692E+06, 3.5217E+06, 3.5749E+06,
     +  3.6289E+06, 3.6836E+06, 3.7391E+06, 3.7955E+06, 3.8526E+06,
     + 3.9106E+06, 3.9694E+06, 4.0290E+06, 4.0895E+06, 4.1509E+06,
     +  4.2131E+06, 4.2762E+06, 4.3402E+06, 4.4052E+06, 4.4710E+06,
     + 4.5378E+06, 4.6056E+06, 4.6743E+06, 4.7439E+06, 4.8146E+06,
     +  4.8863E+06, 4.9589E+06, 5.0327E+06, 5.1074E+06, 5.1832E+06,
     + 5.2601E+06, 5.3381E+06, 5.4171E+06, 5.4973E+06, 5.5786E+06,
     +  5.6611E+06, 5.7447E+06, 5.8295E+06, 5.9155E+06, 6.0027E+06,
     + 6.0911E+06, 6.1807E+06, 6.2716E+06, 6.3638E+06, 6.4573E+06,
     +  6.5520E+06, 6.6481E+06, 6.7455E+06, 6.8443E+06, 6.9445E+06,
     + 7.0461E+06, 7.1490E+06, 7.2534E+06, 7.3593E+06, 7.4666E+06,
     +  7.5754E+06, 7.6858E+06, 7.7976E+06, 7.9110E+06, 8.0260E+06,
     + 8.1426E+06, 8.2607E+06, 8.3805E+06, 8.5020E+06, 8.6251E+06,
     +  8.7499E+06, 8.8765E+06, 9.0047E+06, 9.1348E+06, 9.2666E+06,
     + 9.4003E+06, 9.5357E+06, 9.6731E+06, 9.8123E+06, 9.9534E+06,
     +  1.0096E+07, 1.0241E+07, 1.0388E+07, 1.0537E+07, 1.0688E+07,
     + 1.0841E+07, 1.0997E+07, 1.1154E+07, 1.1313E+07, 1.1475E+07,
     +  1.1639E+07, 1.1805E+07, 1.1973E+07, 1.2143E+07, 1.2316E+07,
     + 1.2491E+07, 1.2669E+07, 1.2849E+07, 1.3031E+07, 1.3216E+07,
     +  1.3403E+07, 1.3593E+07, 1.3786E+07, 1.3981E+07, 1.4178E+07,
     + 1.4378E+07, 1.4581E+07, 1.4787E+07, 1.4996E+07, 1.5207E+07,
     +  1.5421E+07, 1.5638E+07, 1.5858E+07, 1.6080E+07, 1.6306E+07,
     + 1.6535E+07, 1.6767E+07, 1.7001E+07, 1.7239E+07, 1.7481E+07,
     +  1.7725E+07, 1.7973E+07, 1.8224E+07, 1.8478E+07, 1.8735E+07,
     + 1.8996E+07, 1.9261E+07, 1.9529E+07, 1.9800E+07, 2.0076E+07,
     +  2.0354E+07, 2.0637E+07, 2.0923E+07, 2.1213E+07, 2.1506E+07,
     + 2.1804E+07, 2.2106E+07, 2.2411E+07, 2.2721E+07, 2.3034E+07,
     +  2.3352E+07, 2.3674E+07, 2.4000E+07, 2.4330E+07, 2.4665E+07,
     + 2.5003E+07, 2.5347E+07, 2.5695E+07, 2.6047E+07, 2.6404E+07,
     +  2.6766E+07, 2.7132E+07, 2.7503E+07, 2.7879E+07, 2.8260E+07,
     + 2.8646E+07, 2.9037E+07, 2.9433E+07, 2.9834E+07, 3.0240E+07,
     +  3.0651E+07, 3.1068E+07, 3.1490E+07, 3.1918E+07, 3.2351E+07,
     + 3.2789E+07, 3.3234E+07, 3.3684E+07, 3.4140E+07, 3.4601E+07,
     +  3.5069E+07, 3.5542E+07, 3.6022E+07, 3.6508E+07, 3.7000E+07,
     + 3.7498E+07, 3.8003E+07, 3.8514E+07, 3.9032E+07, 3.9556E+07,
     +  4.0087E+07, 4.0625E+07, 4.1169E+07, 4.1721E+07, 4.2280E+07,
     + 4.2845E+07, 4.3418E+07, 4.3998E+07, 4.4586E+07, 4.5181E+07,
     +  4.5783E+07, 4.6393E+07, 4.7011E+07, 4.7637E+07, 4.8270E+07,
     + 4.8912E+07, 4.9561E+07, 5.0219E+07, 5.0885E+07, 5.1560E+07,
     +  5.2242E+07, 5.2934E+07, 5.3634E+07, 5.4343E+07, 5.5061E+07,
     + 5.5788E+07, 5.6524E+07, 5.7269E+07, 5.8024E+07, 5.8788E+07,
     +  5.9562E+07, 6.0345E+07, 6.1138E+07, 6.1941E+07, 6.2754E+07,
     + 6.3577E+07, 6.4410E+07, 6.5254E+07, 6.6108E+07, 6.6972E+07,
     +  6.7848E+07, 6.8734E+07, 6.9632E+07, 7.0540E+07, 7.1460E+07,
     + 7.2391E+07, 7.3333E+07, 7.4287E+07, 7.5253E+07, 7.6231E+07,
     +  7.7221E+07, 7.8223E+07, 7.9238E+07, 8.0265E+07, 8.1305E+07,
     + 8.2357E+07, 8.3422E+07, 8.4501E+07, 8.5592E+07, 8.6697E+07,
     +  8.7816E+07, 8.8948E+07, 9.0094E+07, 9.1254E+07, 9.2428E+07,
     + 9.3617E+07, 9.4820E+07, 9.6038E+07, 9.7271E+07, 9.8518E+07,
     +  9.9781E+07, 1.0106E+08, 1.0235E+08, 1.0366E+08, 1.0499E+08,
     + 1.0633E+08, 1.0769E+08, 1.0906E+08, 1.1045E+08, 1.1186E+08,
     +  1.1328E+08, 1.1472E+08, 1.1618E+08, 1.1766E+08, 1.1915E+08,
     + 1.2067E+08, 1.2220E+08, 1.2375E+08, 1.2531E+08, 1.2690E+08,
     +  1.2851E+08, 1.3013E+08, 1.3178E+08, 1.3344E+08, 1.3512E+08,
     + 1.3683E+08, 1.3855E+08, 1.4030E+08, 1.4206E+08, 1.4385E+08,
     +  1.4566E+08, 1.4749E+08, 1.4934E+08, 1.5122E+08, 1.5311E+08,
     + 1.5503E+08, 1.5697E+08, 1.5894E+08, 1.6092E+08, 1.6293E+08,
     +  1.6497E+08, 1.6703E+08, 1.6911E+08, 1.7122E+08, 1.7335E+08,
     + 1.7551E+08, 1.7770E+08, 1.7991E+08, 1.8214E+08, 1.8440E+08,
     +  1.8669E+08, 1.8901E+08, 1.9135E+08, 1.9372E+08, 1.9612E+08,
     + 1.9855E+08, 2.0100E+08, 2.0348E+08, 2.0600E+08, 2.0854E+08,
     +  2.1111E+08, 2.1371E+08, 2.1634E+08, 2.1901E+08, 2.2170E+08,
     + 2.2443E+08, 2.2718E+08, 2.2997E+08, 2.3279E+08, 2.3565E+08,
     +  2.3854E+08, 2.4146E+08, 2.4441E+08, 2.4740E+08, 2.5043E+08,
     + 2.5348E+08, 2.5658E+08, 2.5971E+08, 2.6288E+08, 2.6608E+08,
     +  2.6932E+08, 2.7260E+08, 2.7591E+08, 2.7926E+08, 2.8266E+08,
     + 2.8609E+08, 2.8956E+08, 2.9307E+08, 2.9662E+08, 3.0021E+08,
     +  3.0384E+08, 3.0752E+08, 3.1123E+08, 3.1499E+08, 3.1879E+08,
     + 3.2264E+08, 3.2653E+08, 3.3046E+08, 3.3444E+08, 3.3847E+08,
     +  3.4254E+08, 3.4665E+08, 3.5082E+08, 3.5503E+08, 3.5928E+08,
     + 3.6359E+08, 3.6795E+08, 3.7235E+08, 3.7681E+08, 3.8131E+08,
     +  3.8587E+08, 3.9048E+08, 3.9514E+08, 3.9985E+08, 4.0462E+08,
     + 4.0944E+08, 4.1431E+08, 4.1924E+08, 4.2422E+08, 4.2927E+08,
     +  4.3436E+08, 4.3952E+08, 4.4473E+08, 4.5000E+08, 4.5533E+08,
     + 4.6072E+08, 4.6617E+08, 4.7169E+08, 4.7726E+08, 4.8290E+08,
     +  4.8860E+08, 4.9436E+08, 5.0019E+08, 5.0608E+08, 5.1204E+08,
     + 5.1806E+08, 5.2415E+08, 5.3031E+08, 5.3654E+08, 5.4284E+08,
     +  5.4920E+08, 5.5564E+08, 5.6215E+08, 5.6873E+08, 5.7539E+08,
     + 5.8211E+08, 5.8892E+08, 5.9579E+08, 6.0275E+08, 6.0978E+08,
     +  6.1689E+08, 6.2407E+08, 6.3134E+08, 6.3869E+08, 6.4611E+08,
     + 6.5362E+08, 6.6121E+08, 6.6889E+08, 6.7665E+08, 6.8449E+08,
     +  6.9242E+08, 7.0044E+08, 7.0855E+08, 7.1674E+08, 7.2502E+08,
     + 7.3340E+08, 7.4186E+08, 7.5042E+08, 7.5907E+08, 7.6782E+08,
     +  7.7666E+08, 7.8560E+08, 7.9464E+08, 8.0377E+08, 8.1300E+08,
     + 8.2234E+08, 8.3177E+08, 8.4131E+08, 8.5095E+08, 8.6070E+08,
     +  8.7055E+08, 8.8051E+08, 8.9057E+08, 9.0075E+08, 9.1103E+08,
     + 9.2143E+08, 9.3193E+08, 9.4256E+08, 9.5329E+08, 9.6414E+08,
     +  9.7511E+08, 9.8620E+08, 9.9740E+08, 1.0087E+09, 1.0202E+09,
     + 1.0317E+09, 1.0434E+09, 1.0553E+09, 1.0672E+09, 1.0793E+09,
     +  1.0915E+09, 1.1038E+09, 1.1163E+09, 1.1289E+09, 1.1416E+09,
     + 1.1545E+09, 1.1675E+09, 1.1807E+09, 1.1939E+09, 1.2074E+09,
     +  1.2209E+09, 1.2347E+09, 1.2485E+09, 1.2625E+09, 1.2767E+09,
     + 1.2910E+09, 1.3054E+09, 1.3201E+09, 1.3348E+09, 1.3498E+09,
     +  1.3648E+09, 1.3801E+09, 1.3955E+09, 1.4110E+09, 1.4268E+09,
     + 1.4427E+09, 1.4587E+09, 1.4750E+09, 1.4914E+09, 1.5079E+09,
     +  1.5247E+09, 1.5416E+09, 1.5587E+09, 1.5760E+09, 1.5934E+09,
     + 1.6111E+09, 1.6289E+09, 1.6469E+09, 1.6651E+09, 1.6835E+09,
     +  1.7021E+09, 1.7209E+09, 1.7399E+09, 1.7591E+09, 1.7784E+09,
     + 1.7980E+09, 1.8178E+09, 1.8378E+09, 1.8580E+09, 1.8784E+09,
     +  1.8990E+09, 1.9198E+09, 1.9409E+09, 1.9621E+09, 1.9836E+09,
     + 2.0053E+09, 2.0273E+09, 2.0494E+09, 2.0718E+09, 2.0944E+09,
     +  2.1173E+09, 2.1404E+09, 2.1637E+09, 2.1873E+09, 2.2111E+09,
     + 2.2351E+09, 2.2594E+09, 2.2840E+09, 2.3088E+09, 2.3339E+09,
     +  2.3592E+09, 2.3848E+09, 2.4106E+09, 2.4367E+09, 2.4631E+09,
     + 2.4897E+09, 2.5166E+09, 2.5438E+09, 2.5713E+09, 2.5990E+09,
     +  2.6270E+09, 2.6554E+09, 2.6840E+09, 2.7129E+09, 2.7420E+09,
     + 2.7715E+09, 2.8013E+09, 2.8314E+09, 2.8618E+09, 2.8925E+09,
     +  2.9235E+09, 2.9548E+09, 2.9865E+09, 3.0184E+09, 3.0507E+09,
     + 3.0834E+09, 3.1163E+09, 3.1496E+09, 3.1832E+09, 3.2171E+09,
     +  3.2514E+09, 3.2861E+09, 3.3211E+09, 3.3564E+09, 3.3921E+09,
     + 3.4281E+09, 3.4646E+09, 3.5013E+09, 3.5385E+09, 3.5760E+09,
     +  3.6139E+09, 3.6522E+09, 3.6909E+09, 3.7299E+09, 3.7694E+09,
     + 3.8092E+09, 3.8494E+09, 3.8901E+09, 3.9311E+09, 3.9725E+09,
     +  4.0144E+09, 4.0567E+09, 4.0994E+09, 4.1425E+09, 4.1861E+09,
     + 4.2301E+09, 4.2745E+09, 4.3193E+09, 4.3646E+09, 4.4104E+09,
     +  4.4566E+09, 4.5033E+09, 4.5504E+09, 4.5980E+09, 4.6461E+09,
     + 4.6946E+09, 4.7436E+09, 4.7931E+09, 4.8431E+09, 4.8936E+09,
     +  4.9446E+09, 4.9961E+09, 5.0481E+09, 5.1006E+09, 5.1536E+09,
     + 5.2071E+09, 5.2612E+09, 5.3158E+09, 5.3709E+09, 5.4266E+09,
     +  5.4828E+09, 5.5396E+09, 5.5969E+09, 5.6548E+09, 5.7132E+09,
     + 5.7722E+09, 5.8318E+09, 5.8920E+09, 5.9527E+09, 6.0141E+09,
     +  6.0760E+09, 6.1386E+09, 6.2018E+09, 6.2655E+09, 6.3299E+09,
     + 6.3949E+09, 6.4606E+09, 6.5269E+09, 6.5938E+09, 6.6614E+09,
     +  6.7296E+09, 6.7985E+09, 6.8681E+09, 6.9383E+09, 7.0092E+09,
     + 7.0808E+09, 7.1531E+09, 7.2261E+09, 7.2997E+09, 7.3741E+09,
     +  7.4492E+09, 7.5251E+09, 7.6016E+09, 7.6789E+09, 7.7570E+09,
     + 7.8358E+09, 7.9153E+09, 7.9956E+09, 8.0767E+09, 8.1585E+09,
     +  8.2412E+09, 8.3246E+09, 8.4088E+09, 8.4939E+09, 8.5797E+09,
     + 8.6664E+09, 8.7538E+09, 8.8422E+09, 8.9313E+09, 9.0213E+09,
     +  9.1122E+09, 9.2039E+09, 9.2965E+09, 9.3900E+09, 9.4844E+09,
     + 9.5797E+09, 9.6758E+09, 9.7729E+09, 9.8709E+09, 9.9699E+09,
     +  1.0070E+10, 1.0171E+10, 1.0272E+10, 1.0375E+10, 1.0479E+10,
     + 1.0583E+10, 1.0689E+10, 1.0796E+10, 1.0903E+10, 1.1012E+10,
     +  1.1122E+10, 1.1233E+10, 1.1344E+10, 1.1457E+10, 1.1571E+10,
     + 1.1686E+10, 1.1802E+10, 1.1919E+10, 1.2037E+10, 1.2157E+10,
     +  1.2277E+10, 1.2399E+10, 1.2521E+10, 1.2645E+10, 1.2770E+10,
     + 1.2896E+10, 1.3024E+10, 1.3152E+10, 1.3282E+10, 1.3413E+10,
     +  1.3545E+10, 1.3679E+10, 1.3813E+10, 1.3949E+10, 1.4086E+10,
     + 1.4225E+10, 1.4364E+10, 1.4505E+10, 1.4648E+10, 1.4791E+10,
     +  1.4936E+10, 1.5082E+10, 1.5230E+10, 1.5379E+10, 1.5529E+10,
     + 1.5681E+10, 1.5834E+10, 1.5989E+10, 1.6145E+10, 1.6302E+10,
     +  1.6461E+10, 1.6622E+10, 1.6784E+10, 1.6947E+10, 1.7112E+10,
     + 1.7278E+10, 1.7446E+10, 1.7615E+10, 1.7786E+10, 1.7959E+10,
     +  1.8133E+10, 1.8309E+10, 1.8486E+10, 1.8665E+10, 1.8845E+10,
     + 1.9028E+10, 1.9211E+10, 1.9397E+10, 1.9584E+10, 1.9773E+10,
     +  1.9964E+10, 2.0156E+10, 2.0350E+10, 2.0546E+10, 2.0744E+10,
     + 2.0943E+10, 2.1145E+10, 2.1348E+10, 2.1553E+10, 2.1760E+10,
     +  2.1968E+10, 2.2179E+10, 2.2391E+10, 2.2606E+10, 2.2822E+10,
     + 2.3040E+10, 2.3261E+10, 2.3483E+10, 2.3707E+10, 2.3933E+10,
     +  2.4162E+10, 2.4392E+10, 2.4625E+10, 2.4859E+10, 2.5096E+10,
     + 2.5335E+10, 2.5576E+10, 2.5819E+10, 2.6064E+10, 2.6311E+10,
     +  2.6561E+10, 2.6813E+10, 2.7067E+10, 2.7323E+10, 2.7582E+10,
     + 2.7843E+10, 2.8107E+10, 2.8372E+10, 2.8640E+10, 2.8911E+10,
     +  2.9184E+10, 2.9459E+10, 2.9737E+10, 3.0017E+10, 3.0300E+10,
     + 3.0585E+10, 3.0873E+10, 3.1163E+10, 3.1456E+10, 3.1752E+10,
     +  3.2050E+10, 3.2351E+10, 3.2654E+10, 3.2960E+10, 3.3269E+10,
     + 3.3581E+10, 3.3895E+10, 3.4212E+10, 3.4532E+10, 3.4855E+10,
     +  3.5180E+10, 3.5509E+10, 3.5840E+10, 3.6174E+10, 3.6511E+10,
     + 3.6852E+10, 3.7195E+10, 3.7541E+10, 3.7890E+10, 3.8242E+10,
     +  3.8598E+10, 3.8956E+10, 3.9318E+10, 3.9682E+10, 4.0050E+10,
     + 4.0422E+10, 4.0796E+10, 4.1174E+10, 4.1555E+10, 4.1939E+10,
     +  4.2327E+10, 4.2718E+10, 4.3112E+10, 4.3510E+10, 4.3911E+10,
     + 4.4316E+10, 4.4725E+10, 4.5137E+10, 4.5552E+10, 4.5971E+10,
     +  4.6394E+10, 4.6820E+10, 4.7250E+10, 4.7684E+10, 4.8122E+10,
     + 4.8563E+10, 4.9008E+10, 4.9457E+10, 4.9910E+10, 5.0367E+10,
     +  5.0828E+10, 5.1293E+10, 5.1762E+10, 5.2234E+10, 5.2711E+10,
     + 5.3192E+10, 5.3677E+10, 5.4167E+10, 5.4660E+10, 5.5158E+10,
     +  5.5660E+10, 5.6166E+10, 5.6677E+10, 5.7192E+10, 5.7712E+10,
     + 5.8236E+10, 5.8764E+10, 5.9297E+10, 5.9834E+10, 6.0376E+10,
     +  6.0923E+10, 6.1475E+10, 6.2031E+10, 6.2592E+10, 6.3157E+10,
     + 6.3728E+10, 6.4303E+10, 6.4883E+10, 6.5468E+10, 6.6058E+10,
     +  6.6654E+10, 6.7254E+10, 6.7859E+10, 6.8469E+10, 6.9085E+10,
     + 6.9706E+10, 7.0332E+10, 7.0963E+10, 7.1600E+10, 7.2242E+10,
     +  7.2890E+10, 7.3543E+10, 7.4201E+10, 7.4865E+10, 7.5535E+10,
     + 7.6211E+10, 7.6892E+10, 7.7578E+10, 7.8271E+10, 7.8969E+10,
     +  7.9674E+10, 8.0384E+10, 8.1100E+10, 8.1822E+10, 8.2551E+10,
     + 8.3285E+10, 8.4025E+10, 8.4772E+10, 8.5525E+10, 8.6285E+10,
     +  8.7050E+10, 8.7822E+10, 8.8601E+10, 8.9386E+10, 9.0177E+10,
     + 9.0976E+10, 9.1780E+10, 9.2592E+10, 9.3410E+10, 9.4236E+10,
     +  9.5068E+10, 9.5907E+10, 9.6753E+10, 9.7606E+10, 9.8466E+10,
     + 9.9333E+10, 1.0021E+11, 1.0109E+11, 1.0198E+11, 1.0287E+11,
     +  1.0378E+11, 1.0469E+11, 1.0561E+11, 1.0653E+11, 1.0747E+11,
     + 1.0841E+11, 1.0936E+11, 1.1032E+11, 1.1128E+11, 1.1226E+11,
     +  1.1324E+11, 1.1423E+11, 1.1522E+11, 1.1623E+11, 1.1724E+11,
     + 1.1827E+11, 1.1930E+11, 1.2034E+11, 1.2138E+11, 1.2244E+11,
     +  1.2351E+11, 1.2458E+11, 1.2566E+11, 1.2675E+11, 1.2785E+11,
     + 1.2896E+11, 1.3008E+11, 1.3121E+11, 1.3235E+11, 1.3349E+11,
     +  1.3465E+11, 1.3581E+11, 1.3699E+11, 1.3817E+11, 1.3937E+11,
     + 1.4057E+11, 1.4178E+11, 1.4301E+11, 1.4424E+11, 1.4548E+11,
     +  1.4673E+11, 1.4800E+11, 1.4927E+11, 1.5056E+11, 1.5185E+11,
     + 1.5315E+11, 1.5447E+11, 1.5580E+11, 1.5713E+11, 1.5848E+11,
     +  1.5984E+11, 1.6121E+11, 1.6259E+11, 1.6398E+11, 1.6538E+11,
     + 1.6680E+11, 1.6822E+11, 1.6966E+11, 1.7111E+11, 1.7257E+11,
     +  1.7404E+11, 1.7552E+11, 1.7702E+11, 1.7853E+11, 1.8005E+11,
     + 1.8158E+11, 1.8312E+11, 1.8468E+11, 1.8625E+11, 1.8783E+11,
     +  1.8943E+11, 1.9103E+11, 1.9265E+11, 1.9429E+11, 1.9593E+11,
     + 1.9759E+11, 1.9926E+11, 2.0095E+11, 2.0265E+11, 2.0436E+11,
     +  2.0609E+11, 2.0783E+11, 2.0958E+11, 2.1135E+11, 2.1313E+11,
     + 2.1493E+11, 2.1674E+11, 2.1856E+11, 2.2040E+11, 2.2226E+11,
     +  2.2413E+11, 2.2601E+11, 2.2791E+11, 2.2982E+11, 2.3175E+11,
     + 2.3369E+11, 2.3565E+11, 2.3763E+11, 2.3962E+11, 2.4162E+11,
     +  2.4364E+11, 2.4568E+11, 2.4774E+11, 2.4981E+11, 2.5189E+11,
     + 2.5399E+11, 2.5611E+11, 2.5825E+11, 2.6040E+11, 2.6257E+11,
     +  2.6476E+11, 2.6696E+11, 2.6918E+11, 2.7142E+11, 2.7368E+11,
     + 2.7595E+11, 2.7824E+11, 2.8055E+11, 2.8288E+11, 2.8522E+11,
     +  2.8759E+11, 2.8997E+11, 2.9237E+11, 2.9479E+11, 2.9723E+11,
     + 2.9969E+11, 3.0216E+11, 3.0466E+11, 3.0718E+11, 3.0971E+11,
     +  3.1227E+11, 3.1484E+11, 3.1744E+11, 3.2005E+11, 3.2268E+11,
     + 3.2534E+11, 3.2802E+11, 3.3071E+11, 3.3343E+11, 3.3617E+11,
     +  3.3893E+11, 3.4171E+11, 3.4451E+11, 3.4734E+11, 3.5018E+11,
     + 3.5305E+11, 3.5594E+11, 3.5886E+11, 3.6179E+11, 3.6475E+11,
     +  3.6773E+11, 3.7073E+11, 3.7376E+11, 3.7681E+11, 3.7988E+11,
     + 3.8298E+11, 3.8610E+11, 3.8925E+11, 3.9241E+11, 3.9561E+11,
     +  3.9883E+11, 4.0207E+11, 4.0533E+11, 4.0863E+11, 4.1194E+11,
     + 4.1529E+11, 4.1866E+11, 4.2205E+11, 4.2547E+11, 4.2892E+11,
     +  4.3239E+11, 4.3589E+11, 4.3941E+11, 4.4296E+11, 4.4654E+11,
     + 4.5015E+11, 4.5379E+11, 4.5745E+11, 4.6114E+11, 4.6485E+11,
     +  4.6860E+11, 4.7237E+11, 4.7618E+11, 4.8001E+11, 4.8387E+11,
     + 4.8776E+11, 4.9168E+11, 4.9563E+11, 4.9961E+11, 5.0362E+11,
     +  5.0766E+11, 5.1173E+11, 5.1583E+11, 5.1996E+11, 5.2412E+11,
     + 5.2832E+11, 5.3254E+11, 5.3680E+11, 5.4109E+11, 5.4541E+11,
     +  5.4977E+11, 5.5416E+11, 5.5858E+11, 5.6303E+11, 5.6752E+11,
     + 5.7204E+11, 5.7659E+11, 5.8118E+11, 5.8581E+11, 5.9046E+11,
     +  5.9516E+11, 5.9989E+11, 6.0465E+11, 6.0945E+11, 6.1428E+11,
     + 6.1915E+11, 6.2406E+11, 6.2901E+11, 6.3399E+11, 6.3900E+11,
     +  6.4406E+11, 6.4915E+11, 6.5428E+11, 6.5945E+11, 6.6466E+11,
     + 6.6991E+11, 6.7519E+11, 6.8052E+11, 6.8588E+11, 6.9129E+11,
     +  6.9673E+11, 7.0222E+11, 7.0774E+11, 7.1331E+11, 7.1892E+11,
     + 7.2457E+11, 7.3026E+11, 7.3599E+11, 7.4177E+11, 7.4758E+11,
     +  7.5345E+11, 7.5935E+11, 7.6530E+11, 7.7129E+11, 7.7733E+11,
     + 7.8341E+11, 7.8953E+11, 7.9570E+11, 8.0192E+11, 8.0818E+11,
     +  8.1449E+11, 8.2084E+11, 8.2724E+11, 8.3369E+11, 8.4019E+11,
     + 8.4673E+11, 8.5332E+11, 8.5996E+11, 8.6665E+11, 8.7339E+11,
     +  8.8017E+11, 8.8701E+11, 8.9389E+11, 9.0083E+11, 9.0782E+11,
     + 9.1486E+11, 9.2194E+11, 9.2909E+11, 9.3628E+11, 9.4352E+11,
     +  9.5082E+11, 9.5817E+11, 9.6558E+11, 9.7304E+11, 9.8055E+11,
     + 9.8812E+11, 9.9574E+11, 1.0034E+12, 1.0112E+12, 1.0189E+12,
     +  1.0268E+12, 1.0347E+12, 1.0427E+12, 1.0507E+12, 1.0588E+12,
     + 1.0669E+12, 1.0751E+12, 1.0833E+12, 1.0917E+12, 1.1000E+12,
     +  1.1085E+12, 1.1170E+12, 1.1255E+12, 1.1341E+12, 1.1428E+12,
     + 1.1515E+12, 1.1604E+12, 1.1692E+12, 1.1782E+12, 1.1872E+12,
     +  1.1962E+12, 1.2053E+12, 1.2145E+12, 1.2238E+12, 1.2331E+12,
     + 1.2425E+12, 1.2520E+12, 1.2615E+12, 1.2711E+12, 1.2808E+12,
     +  1.2905E+12, 1.3003E+12, 1.3102E+12, 1.3201E+12, 1.3301E+12,
     + 1.3402E+12, 1.3504E+12, 1.3606E+12, 1.3709E+12, 1.3813E+12,
     +  1.3917E+12, 1.4023E+12, 1.4129E+12, 1.4235E+12, 1.4343E+12,
     + 1.4451E+12, 1.4560E+12, 1.4670E+12, 1.4781E+12, 1.4892E+12,
     +  1.5004E+12, 1.5117E+12, 1.5231E+12, 1.5346E+12, 1.5461E+12,
     + 1.5577E+12, 1.5694E+12, 1.5812E+12, 1.5931E+12, 1.6050E+12,
     +  1.6171E+12, 1.6292E+12, 1.6414E+12, 1.6537E+12, 1.6661E+12,
     + 1.6785E+12, 1.6911E+12, 1.7038E+12, 1.7165E+12, 1.7293E+12,
     +  1.7422E+12, 1.7552E+12, 1.7683E+12, 1.7815E+12, 1.7948E+12,
     + 1.8082E+12, 1.8216E+12, 1.8352E+12, 1.8489E+12, 1.8626E+12,
     +  1.8765E+12, 1.8904E+12, 1.9045E+12, 1.9186E+12, 1.9328E+12,
     + 1.9472E+12, 1.9616E+12, 1.9762E+12, 1.9908E+12, 2.0056E+12,
     +  2.0204E+12, 2.0354E+12, 2.0504E+12, 2.0656E+12, 2.0809E+12,
     + 2.0962E+12, 2.1117E+12, 2.1273E+12, 2.1430E+12, 2.1588E+12,
     +  2.1747E+12, 2.1907E+12, 2.2069E+12, 2.2231E+12, 2.2395E+12,
     + 2.2560E+12, 2.2725E+12, 2.2893E+12, 2.3061E+12, 2.3230E+12,
     +  2.3401E+12, 2.3572E+12, 2.3745E+12, 2.3919E+12, 2.4095E+12,
     + 2.4271E+12, 2.4449E+12, 2.4628E+12, 2.4808E+12, 2.4989E+12,
     +  2.5172E+12, 2.5356E+12, 2.5541E+12, 2.5728E+12, 2.5915E+12,
     + 2.6104E+12, 2.6295E+12, 2.6486E+12, 2.6679E+12, 2.6874E+12,
     +  2.7069E+12, 2.7266E+12, 2.7464E+12, 2.7664E+12, 2.7865E+12,
     + 2.8068E+12, 2.8271E+12, 2.8477E+12, 2.8683E+12, 2.8891E+12,
     +  2.9101E+12, 2.9311E+12, 2.9524E+12, 2.9737E+12, 2.9953E+12,
     + 3.0169E+12, 3.0387E+12, 3.0607E+12, 3.0828E+12, 3.1051E+12,
     +  3.1275E+12, 3.1500E+12, 3.1728E+12, 3.1956E+12, 3.2187E+12,
     + 3.2418E+12, 3.2652E+12, 3.2887E+12, 3.3123E+12, 3.3362E+12,
     +  3.3601E+12, 3.3843E+12, 3.4086E+12, 3.4330E+12, 3.4577E+12,
     + 3.4825E+12, 3.5075E+12, 3.5326E+12, 3.5579E+12, 3.5834E+12,
     +  3.6090E+12, 3.6348E+12, 3.6608E+12, 3.6870E+12, 3.7134E+12,
     + 3.7399E+12, 3.7666E+12, 3.7935E+12, 3.8205E+12, 3.8478E+12,
     +  3.8752E+12, 3.9028E+12, 3.9306E+12, 3.9586E+12, 3.9868E+12,
     + 4.0151E+12, 4.0437E+12, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_CH3   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 1),QofT( 1,NT),Q(NT)
      data Tmx(1)/3500./
      data xgj/ 1./
c...      CH3
c...        --      2111
      data (QofT( 1,J),J=1,2408)/ 2.0000E-06, 4.1600E-04, 1.3727E-02,
     +  8.1811E-02, 2.4371E-01, 5.1095E-01, 8.7493E-01, 1.3195E+00,
     + 1.8282E+00, 2.3873E+00, 2.9864E+00, 3.6178E+00, 4.2761E+00,
     +  4.9575E+00, 5.6596E+00, 6.3807E+00, 7.1198E+00, 7.8763E+00,
     + 8.6497E+00, 9.4398E+00, 1.1070E+01, 1.2767E+01, 1.4531E+01,
     +  1.6362E+01, 1.8259E+01, 2.0222E+01, 2.2250E+01, 2.4343E+01,
     + 2.6499E+01, 2.8717E+01, 3.0996E+01, 3.3335E+01, 3.5733E+01,
     +  3.8188E+01, 4.0699E+01, 4.3265E+01, 4.5886E+01, 4.8560E+01,
     + 5.1286E+01, 5.4063E+01, 5.6891E+01, 5.9769E+01, 6.2695E+01,
     +  6.5670E+01, 6.8693E+01, 7.1762E+01, 7.4877E+01, 7.8039E+01,
     + 8.1245E+01, 8.4496E+01, 8.7790E+01, 9.1129E+01, 9.4510E+01,
     +  9.7934E+01, 1.0140E+02, 1.0491E+02, 1.0846E+02, 1.1205E+02,
     + 1.1568E+02, 1.1935E+02, 1.2306E+02, 1.2681E+02, 1.3060E+02,
     +  1.3443E+02, 1.3830E+02, 1.4221E+02, 1.4615E+02, 1.5014E+02,
     + 1.5416E+02, 1.5822E+02, 1.6232E+02, 1.6646E+02, 1.7063E+02,
     +  1.7484E+02, 1.7909E+02, 1.8338E+02, 1.8770E+02, 1.9206E+02,
     + 1.9646E+02, 2.0089E+02, 2.0537E+02, 2.0988E+02, 2.1442E+02,
     +  2.1901E+02, 2.2363E+02, 2.2829E+02, 2.3298E+02, 2.3772E+02,
     + 2.4249E+02, 2.4730E+02, 2.5214E+02, 2.5702E+02, 2.6194E+02,
     +  2.6690E+02, 2.7190E+02, 2.7693E+02, 2.8200E+02, 2.8711E+02,
     + 2.9226E+02, 2.9745E+02, 3.0267E+02, 3.0793E+02, 3.1324E+02,
     +  3.1858E+02, 3.2395E+02, 3.2937E+02, 3.3483E+02, 3.4033E+02,
     + 3.4586E+02, 3.5144E+02, 3.5705E+02, 3.6270E+02, 3.6840E+02,
     +  3.7413E+02, 3.7991E+02, 3.8572E+02, 3.9158E+02, 3.9747E+02,
     + 4.0341E+02, 4.0939E+02, 4.1541E+02, 4.2147E+02, 4.2757E+02,
     +  4.3372E+02, 4.3990E+02, 4.4613E+02, 4.5240E+02, 4.5871E+02,
     + 4.6506E+02, 4.7146E+02, 4.7790E+02, 4.8438E+02, 4.9091E+02,
     +  4.9748E+02, 5.0409E+02, 5.1075E+02, 5.1745E+02, 5.2420E+02,
     + 5.3099E+02, 5.3782E+02, 5.4470E+02, 5.5162E+02, 5.5859E+02,
     +  5.6560E+02, 5.7266E+02, 5.7977E+02, 5.8692E+02, 5.9412E+02,
     + 6.0136E+02, 6.0865E+02, 6.1598E+02, 6.2337E+02, 6.3079E+02,
     +  6.3827E+02, 6.4579E+02, 6.5337E+02, 6.6099E+02, 6.6865E+02,
     + 6.7637E+02, 6.8413E+02, 6.9194E+02, 6.9980E+02, 7.0771E+02,
     +  7.1567E+02, 7.2368E+02, 7.3174E+02, 7.3984E+02, 7.4800E+02,
     + 7.5621E+02, 7.6447E+02, 7.7278E+02, 7.8114E+02, 7.8955E+02,
     +  7.9801E+02, 8.0652E+02, 8.1508E+02, 8.2370E+02, 8.3237E+02,
     + 8.4109E+02, 8.4986E+02, 8.5869E+02, 8.6757E+02, 8.7650E+02,
     +  8.8548E+02, 8.9452E+02, 9.0361E+02, 9.1276E+02, 9.2196E+02,
     + 9.3121E+02, 9.4052E+02, 9.4989E+02, 9.5931E+02, 9.6878E+02,
     +  9.7831E+02, 9.8790E+02, 9.9754E+02, 1.0072E+03, 1.0170E+03,
     + 1.0268E+03, 1.0367E+03, 1.0466E+03, 1.0566E+03, 1.0666E+03,
     +  1.0767E+03, 1.0869E+03, 1.0971E+03, 1.1074E+03, 1.1177E+03,
     + 1.1281E+03, 1.1385E+03, 1.1490E+03, 1.1596E+03, 1.1702E+03,
     +  1.1809E+03, 1.1917E+03, 1.2025E+03, 1.2133E+03, 1.2243E+03,
     + 1.2353E+03, 1.2463E+03, 1.2574E+03, 1.2686E+03, 1.2798E+03,
     +  1.2911E+03, 1.3025E+03, 1.3139E+03, 1.3254E+03, 1.3370E+03,
     + 1.3486E+03, 1.3603E+03, 1.3720E+03, 1.3838E+03, 1.3957E+03,
     +  1.4076E+03, 1.4197E+03, 1.4317E+03, 1.4439E+03, 1.4561E+03,
     + 1.4683E+03, 1.4807E+03, 1.4931E+03, 1.5055E+03, 1.5181E+03,
     +  1.5307E+03, 1.5433E+03, 1.5561E+03, 1.5689E+03, 1.5818E+03,
     + 1.5947E+03, 1.6077E+03, 1.6208E+03, 1.6339E+03, 1.6472E+03,
     +  1.6604E+03, 1.6738E+03, 1.6872E+03, 1.7007E+03, 1.7143E+03,
     + 1.7280E+03, 1.7417E+03, 1.7555E+03, 1.7693E+03, 1.7832E+03,
     +  1.7973E+03, 1.8113E+03, 1.8255E+03, 1.8397E+03, 1.8540E+03,
     + 1.8684E+03, 1.8828E+03, 1.8973E+03, 1.9119E+03, 1.9266E+03,
     +  1.9414E+03, 1.9562E+03, 1.9711E+03, 1.9860E+03, 2.0011E+03,
     + 2.0162E+03, 2.0314E+03, 2.0467E+03, 2.0621E+03, 2.0775E+03,
     +  2.0930E+03, 2.1086E+03, 2.1243E+03, 2.1400E+03, 2.1559E+03,
     + 2.1718E+03, 2.1878E+03, 2.2038E+03, 2.2200E+03, 2.2362E+03,
     +  2.2525E+03, 2.2689E+03, 2.2854E+03, 2.3020E+03, 2.3186E+03,
     + 2.3353E+03, 2.3521E+03, 2.3690E+03, 2.3860E+03, 2.4031E+03,
     +  2.4202E+03, 2.4374E+03, 2.4547E+03, 2.4721E+03, 2.4896E+03,
     + 2.5072E+03, 2.5248E+03, 2.5426E+03, 2.5604E+03, 2.5783E+03,
     +  2.5963E+03, 2.6144E+03, 2.6326E+03, 2.6508E+03, 2.6692E+03,
     + 2.6876E+03, 2.7061E+03, 2.7248E+03, 2.7435E+03, 2.7623E+03,
     +  2.7811E+03, 2.8001E+03, 2.8192E+03, 2.8384E+03, 2.8576E+03,
     + 2.8770E+03, 2.8964E+03, 2.9159E+03, 2.9355E+03, 2.9553E+03,
     +  2.9751E+03, 2.9950E+03, 3.0150E+03, 3.0351E+03, 3.0553E+03,
     + 3.0755E+03, 3.0959E+03, 3.1164E+03, 3.1370E+03, 3.1576E+03,
     +  3.1784E+03, 3.1993E+03, 3.2202E+03, 3.2413E+03, 3.2624E+03,
     + 3.2837E+03, 3.3050E+03, 3.3265E+03, 3.3480E+03, 3.3697E+03,
     +  3.3915E+03, 3.4133E+03, 3.4353E+03, 3.4573E+03, 3.4795E+03,
     + 3.5017E+03, 3.5241E+03, 3.5466E+03, 3.5691E+03, 3.5918E+03,
     +  3.6146E+03, 3.6375E+03, 3.6605E+03, 3.6836E+03, 3.7068E+03,
     + 3.7301E+03, 3.7535E+03, 3.7770E+03, 3.8006E+03, 3.8244E+03,
     +  3.8482E+03, 3.8721E+03, 3.8962E+03, 3.9204E+03, 3.9446E+03,
     + 3.9690E+03, 3.9935E+03, 4.0181E+03, 4.0428E+03, 4.0677E+03,
     +  4.0926E+03, 4.1177E+03, 4.1428E+03, 4.1681E+03, 4.1935E+03,
     + 4.2190E+03, 4.2446E+03, 4.2703E+03, 4.2962E+03, 4.3222E+03,
     +  4.3482E+03, 4.3744E+03, 4.4007E+03, 4.4272E+03, 4.4537E+03,
     + 4.4804E+03, 4.5071E+03, 4.5340E+03, 4.5611E+03, 4.5882E+03,
     +  4.6154E+03, 4.6428E+03, 4.6703E+03, 4.6979E+03, 4.7257E+03,
     + 4.7535E+03, 4.7815E+03, 4.8096E+03, 4.8378E+03, 4.8662E+03,
     +  4.8947E+03, 4.9233E+03, 4.9520E+03, 4.9808E+03, 5.0098E+03,
     + 5.0389E+03, 5.0681E+03, 5.0975E+03, 5.1270E+03, 5.1566E+03,
     +  5.1863E+03, 5.2162E+03, 5.2461E+03, 5.2763E+03, 5.3065E+03,
     + 5.3369E+03, 5.3674E+03, 5.3981E+03, 5.4288E+03, 5.4597E+03,
     +  5.4908E+03, 5.5219E+03, 5.5532E+03, 5.5847E+03, 5.6163E+03,
     + 5.6480E+03, 5.6798E+03, 5.7118E+03, 5.7439E+03, 5.7762E+03,
     +  5.8085E+03, 5.8411E+03, 5.8737E+03, 5.9065E+03, 5.9395E+03,
     + 5.9725E+03, 6.0058E+03, 6.0391E+03, 6.0726E+03, 6.1063E+03,
     +  6.1400E+03, 6.1740E+03, 6.2080E+03, 6.2422E+03, 6.2766E+03,
     + 6.3111E+03, 6.3457E+03, 6.3805E+03, 6.4155E+03, 6.4505E+03,
     +  6.4858E+03, 6.5211E+03, 6.5566E+03, 6.5923E+03, 6.6281E+03,
     + 6.6641E+03, 6.7002E+03, 6.7365E+03, 6.7729E+03, 6.8094E+03,
     +  6.8461E+03, 6.8830E+03, 6.9200E+03, 6.9572E+03, 6.9945E+03,
     + 7.0320E+03, 7.0696E+03, 7.1074E+03, 7.1454E+03, 7.1835E+03,
     +  7.2217E+03, 7.2601E+03, 7.2987E+03, 7.3374E+03, 7.3763E+03,
     + 7.4153E+03, 7.4546E+03, 7.4939E+03, 7.5334E+03, 7.5731E+03,
     +  7.6130E+03, 7.6530E+03, 7.6931E+03, 7.7335E+03, 7.7740E+03,
     + 7.8146E+03, 7.8555E+03, 7.8964E+03, 7.9376E+03, 7.9789E+03,
     +  8.0204E+03, 8.0621E+03, 8.1039E+03, 8.1459E+03, 8.1880E+03,
     + 8.2304E+03, 8.2729E+03, 8.3155E+03, 8.3584E+03, 8.4014E+03,
     +  8.4446E+03, 8.4879E+03, 8.5315E+03, 8.5752E+03, 8.6190E+03,
     + 8.6631E+03, 8.7073E+03, 8.7517E+03, 8.7963E+03, 8.8411E+03,
     +  8.8860E+03, 8.9311E+03, 8.9764E+03, 9.0219E+03, 9.0675E+03,
     + 9.1134E+03, 9.1594E+03, 9.2056E+03, 9.2520E+03, 9.2985E+03,
     +  9.3453E+03, 9.3922E+03, 9.4393E+03, 9.4866E+03, 9.5341E+03,
     + 9.5818E+03, 9.6296E+03, 9.6777E+03, 9.7259E+03, 9.7743E+03,
     +  9.8229E+03, 9.8717E+03, 9.9207E+03, 9.9699E+03, 1.0019E+04,
     + 1.0069E+04, 1.0119E+04, 1.0169E+04, 1.0219E+04, 1.0269E+04,
     +  1.0320E+04, 1.0370E+04, 1.0421E+04, 1.0472E+04, 1.0524E+04,
     + 1.0575E+04, 1.0627E+04, 1.0679E+04, 1.0731E+04, 1.0783E+04,
     +  1.0836E+04, 1.0889E+04, 1.0941E+04, 1.0995E+04, 1.1048E+04,
     + 1.1101E+04, 1.1155E+04, 1.1209E+04, 1.1263E+04, 1.1318E+04,
     +  1.1372E+04, 1.1427E+04, 1.1482E+04, 1.1537E+04, 1.1593E+04,
     + 1.1648E+04, 1.1704E+04, 1.1760E+04, 1.1816E+04, 1.1873E+04,
     +  1.1930E+04, 1.1986E+04, 1.2044E+04, 1.2101E+04, 1.2158E+04,
     + 1.2216E+04, 1.2274E+04, 1.2332E+04, 1.2391E+04, 1.2449E+04,
     +  1.2508E+04, 1.2567E+04, 1.2627E+04, 1.2686E+04, 1.2746E+04,
     + 1.2806E+04, 1.2866E+04, 1.2926E+04, 1.2987E+04, 1.3048E+04,
     +  1.3109E+04, 1.3170E+04, 1.3232E+04, 1.3294E+04, 1.3356E+04,
     + 1.3418E+04, 1.3481E+04, 1.3543E+04, 1.3606E+04, 1.3669E+04,
     +  1.3733E+04, 1.3796E+04, 1.3860E+04, 1.3924E+04, 1.3989E+04,
     + 1.4053E+04, 1.4118E+04, 1.4183E+04, 1.4249E+04, 1.4314E+04,
     +  1.4380E+04, 1.4446E+04, 1.4512E+04, 1.4579E+04, 1.4646E+04,
     + 1.4713E+04, 1.4780E+04, 1.4847E+04, 1.4915E+04, 1.4983E+04,
     +  1.5051E+04, 1.5120E+04, 1.5188E+04, 1.5257E+04, 1.5327E+04,
     + 1.5396E+04, 1.5466E+04, 1.5536E+04, 1.5606E+04, 1.5677E+04,
     +  1.5748E+04, 1.5819E+04, 1.5890E+04, 1.5961E+04, 1.6033E+04,
     + 1.6105E+04, 1.6178E+04, 1.6250E+04, 1.6323E+04, 1.6396E+04,
     +  1.6469E+04, 1.6543E+04, 1.6617E+04, 1.6691E+04, 1.6766E+04,
     + 1.6840E+04, 1.6915E+04, 1.6990E+04, 1.7066E+04, 1.7142E+04,
     +  1.7218E+04, 1.7294E+04, 1.7371E+04, 1.7448E+04, 1.7525E+04,
     + 1.7602E+04, 1.7680E+04, 1.7758E+04, 1.7836E+04, 1.7914E+04,
     +  1.7993E+04, 1.8072E+04, 1.8152E+04, 1.8231E+04, 1.8311E+04,
     + 1.8391E+04, 1.8472E+04, 1.8553E+04, 1.8634E+04, 1.8715E+04,
     +  1.8797E+04, 1.8879E+04, 1.8961E+04, 1.9043E+04, 1.9126E+04,
     + 1.9209E+04, 1.9293E+04, 1.9376E+04, 1.9460E+04, 1.9544E+04,
     +  1.9629E+04, 1.9714E+04, 1.9799E+04, 1.9884E+04, 1.9970E+04,
     + 2.0056E+04, 2.0142E+04, 2.0229E+04, 2.0316E+04, 2.0403E+04,
     +  2.0491E+04, 2.0579E+04, 2.0667E+04, 2.0755E+04, 2.0844E+04,
     + 2.0933E+04, 2.1022E+04, 2.1112E+04, 2.1202E+04, 2.1292E+04,
     +  2.1383E+04, 2.1474E+04, 2.1565E+04, 2.1657E+04, 2.1749E+04,
     + 2.1841E+04, 2.1933E+04, 2.2026E+04, 2.2119E+04, 2.2213E+04,
     +  2.2306E+04, 2.2401E+04, 2.2495E+04, 2.2590E+04, 2.2685E+04,
     + 2.2780E+04, 2.2876E+04, 2.2972E+04, 2.3068E+04, 2.3165E+04,
     +  2.3262E+04, 2.3359E+04, 2.3457E+04, 2.3555E+04, 2.3654E+04,
     + 2.3752E+04, 2.3851E+04, 2.3951E+04, 2.4050E+04, 2.4150E+04,
     +  2.4251E+04, 2.4352E+04, 2.4453E+04, 2.4554E+04, 2.4656E+04,
     + 2.4758E+04, 2.4860E+04, 2.4963E+04, 2.5066E+04, 2.5170E+04,
     +  2.5273E+04, 2.5378E+04, 2.5482E+04, 2.5587E+04, 2.5692E+04,
     + 2.5798E+04, 2.5904E+04, 2.6010E+04, 2.6117E+04, 2.6224E+04,
     +  2.6331E+04, 2.6439E+04, 2.6547E+04, 2.6655E+04, 2.6764E+04,
     + 2.6873E+04, 2.6983E+04, 2.7092E+04, 2.7203E+04, 2.7313E+04,
     +  2.7424E+04, 2.7536E+04, 2.7647E+04, 2.7759E+04, 2.7872E+04,
     + 2.7985E+04, 2.8098E+04, 2.8211E+04, 2.8325E+04, 2.8440E+04,
     +  2.8554E+04, 2.8669E+04, 2.8785E+04, 2.8901E+04, 2.9017E+04,
     + 2.9134E+04, 2.9250E+04, 2.9368E+04, 2.9486E+04, 2.9604E+04,
     +  2.9722E+04, 2.9841E+04, 2.9960E+04, 3.0080E+04, 3.0200E+04,
     + 3.0321E+04, 3.0441E+04, 3.0563E+04, 3.0684E+04, 3.0806E+04,
     +  3.0929E+04, 3.1052E+04, 3.1175E+04, 3.1299E+04, 3.1423E+04,
     + 3.1547E+04, 3.1672E+04, 3.1797E+04, 3.1923E+04, 3.2049E+04,
     +  3.2175E+04, 3.2302E+04, 3.2429E+04, 3.2557E+04, 3.2685E+04,
     + 3.2814E+04, 3.2943E+04, 3.3072E+04, 3.3202E+04, 3.3332E+04,
     +  3.3463E+04, 3.3594E+04, 3.3725E+04, 3.3857E+04, 3.3989E+04,
     + 3.4122E+04, 3.4255E+04, 3.4389E+04, 3.4523E+04, 3.4657E+04,
     +  3.4792E+04, 3.4927E+04, 3.5063E+04, 3.5199E+04, 3.5335E+04,
     + 3.5472E+04, 3.5610E+04, 3.5748E+04, 3.5886E+04, 3.6025E+04,
     +  3.6164E+04, 3.6304E+04, 3.6444E+04, 3.6584E+04, 3.6725E+04,
     + 3.6866E+04, 3.7008E+04, 3.7151E+04, 3.7293E+04, 3.7437E+04,
     +  3.7580E+04, 3.7724E+04, 3.7869E+04, 3.8014E+04, 3.8159E+04,
     + 3.8305E+04, 3.8452E+04, 3.8598E+04, 3.8746E+04, 3.8893E+04,
     +  3.9042E+04, 3.9190E+04, 3.9340E+04, 3.9489E+04, 3.9639E+04,
     + 3.9790E+04, 3.9941E+04, 4.0092E+04, 4.0244E+04, 4.0397E+04,
     +  4.0550E+04, 4.0703E+04, 4.0857E+04, 4.1011E+04, 4.1166E+04,
     + 4.1321E+04, 4.1477E+04, 4.1633E+04, 4.1790E+04, 4.1947E+04,
     +  4.2105E+04, 4.2263E+04, 4.2422E+04, 4.2581E+04, 4.2741E+04,
     + 4.2901E+04, 4.3061E+04, 4.3223E+04, 4.3384E+04, 4.3546E+04,
     +  4.3709E+04, 4.3872E+04, 4.4036E+04, 4.4200E+04, 4.4365E+04,
     + 4.4530E+04, 4.4695E+04, 4.4862E+04, 4.5028E+04, 4.5195E+04,
     +  4.5363E+04, 4.5531E+04, 4.5700E+04, 4.5869E+04, 4.6039E+04,
     + 4.6209E+04, 4.6380E+04, 4.6551E+04, 4.6723E+04, 4.6895E+04,
     +  4.7068E+04, 4.7242E+04, 4.7416E+04, 4.7590E+04, 4.7765E+04,
     + 4.7940E+04, 4.8116E+04, 4.8293E+04, 4.8470E+04, 4.8648E+04,
     +  4.8826E+04, 4.9005E+04, 4.9184E+04, 4.9364E+04, 4.9544E+04,
     + 4.9725E+04, 4.9906E+04, 5.0088E+04, 5.0271E+04, 5.0454E+04,
     +  5.0637E+04, 5.0821E+04, 5.1006E+04, 5.1191E+04, 5.1377E+04,
     + 5.1563E+04, 5.1750E+04, 5.1938E+04, 5.2126E+04, 5.2314E+04,
     +  5.2504E+04, 5.2693E+04, 5.2884E+04, 5.3074E+04, 5.3266E+04,
     + 5.3458E+04, 5.3650E+04, 5.3843E+04, 5.4037E+04, 5.4231E+04,
     +  5.4426E+04, 5.4622E+04, 5.4818E+04, 5.5014E+04, 5.5211E+04,
     + 5.5409E+04, 5.5607E+04, 5.5806E+04, 5.6006E+04, 5.6206E+04,
     +  5.6407E+04, 5.6608E+04, 5.6810E+04, 5.7012E+04, 5.7215E+04,
     + 5.7419E+04, 5.7623E+04, 5.7828E+04, 5.8033E+04, 5.8239E+04,
     +  5.8446E+04, 5.8653E+04, 5.8861E+04, 5.9070E+04, 5.9279E+04,
     + 5.9488E+04, 5.9699E+04, 5.9910E+04, 6.0121E+04, 6.0333E+04,
     +  6.0546E+04, 6.0759E+04, 6.0973E+04, 6.1188E+04, 6.1403E+04,
     + 6.1619E+04, 6.1835E+04, 6.2053E+04, 6.2270E+04, 6.2489E+04,
     +  6.2708E+04, 6.2927E+04, 6.3148E+04, 6.3369E+04, 6.3590E+04,
     + 6.3812E+04, 6.4035E+04, 6.4259E+04, 6.4483E+04, 6.4707E+04,
     +  6.4933E+04, 6.5159E+04, 6.5386E+04, 6.5613E+04, 6.5841E+04,
     + 6.6070E+04, 6.6299E+04, 6.6529E+04, 6.6759E+04, 6.6991E+04,
     +  6.7223E+04, 6.7455E+04, 6.7689E+04, 6.7923E+04, 6.8157E+04,
     + 6.8392E+04, 6.8628E+04, 6.8865E+04, 6.9102E+04, 6.9340E+04,
     +  6.9579E+04, 6.9818E+04, 7.0058E+04, 7.0299E+04, 7.0540E+04,
     + 7.0782E+04, 7.1025E+04, 7.1268E+04, 7.1513E+04, 7.1757E+04,
     +  7.2003E+04, 7.2249E+04, 7.2496E+04, 7.2744E+04, 7.2992E+04,
     + 7.3241E+04, 7.3491E+04, 7.3741E+04, 7.3992E+04, 7.4244E+04,
     +  7.4496E+04, 7.4749E+04, 7.5003E+04, 7.5258E+04, 7.5513E+04,
     + 7.5769E+04, 7.6026E+04, 7.6284E+04, 7.6542E+04, 7.6801E+04,
     +  7.7061E+04, 7.7321E+04, 7.7582E+04, 7.7844E+04, 7.8107E+04,
     + 7.8370E+04, 7.8634E+04, 7.8899E+04, 7.9164E+04, 7.9430E+04,
     +  7.9697E+04, 7.9965E+04, 8.0234E+04, 8.0503E+04, 8.0773E+04,
     + 8.1044E+04, 8.1315E+04, 8.1587E+04, 8.1860E+04, 8.2134E+04,
     +  8.2408E+04, 8.2684E+04, 8.2960E+04, 8.3236E+04, 8.3514E+04,
     + 8.3792E+04, 8.4071E+04, 8.4351E+04, 8.4632E+04, 8.4913E+04,
     +  8.5195E+04, 8.5478E+04, 8.5762E+04, 8.6046E+04, 8.6332E+04,
     + 8.6618E+04, 8.6904E+04, 8.7192E+04, 8.7480E+04, 8.7770E+04,
     +  8.8060E+04, 8.8350E+04, 8.8642E+04, 8.8934E+04, 8.9227E+04,
     + 8.9521E+04, 8.9816E+04, 9.0112E+04, 9.0408E+04, 9.0705E+04,
     +  9.1003E+04, 9.1302E+04, 9.1602E+04, 9.1902E+04, 9.2203E+04,
     + 9.2505E+04, 9.2808E+04, 9.3112E+04, 9.3416E+04, 9.3722E+04,
     +  9.4028E+04, 9.4335E+04, 9.4642E+04, 9.4951E+04, 9.5260E+04,
     + 9.5571E+04, 9.5882E+04, 9.6194E+04, 9.6507E+04, 9.6820E+04,
     +  9.7135E+04, 9.7450E+04, 9.7766E+04, 9.8083E+04, 9.8401E+04,
     + 9.8720E+04, 9.9040E+04, 9.9360E+04, 9.9681E+04, 1.0000E+05,
     +  1.0033E+05, 1.0065E+05, 1.0098E+05, 1.0130E+05, 1.0163E+05,
     + 1.0195E+05, 1.0228E+05, 1.0261E+05, 1.0294E+05, 1.0327E+05,
     +  1.0360E+05, 1.0394E+05, 1.0427E+05, 1.0461E+05, 1.0494E+05,
     + 1.0528E+05, 1.0561E+05, 1.0595E+05, 1.0629E+05, 1.0663E+05,
     +  1.0697E+05, 1.0731E+05, 1.0765E+05, 1.0800E+05, 1.0834E+05,
     + 1.0869E+05, 1.0903E+05, 1.0938E+05, 1.0973E+05, 1.1008E+05,
     +  1.1043E+05, 1.1078E+05, 1.1113E+05, 1.1148E+05, 1.1183E+05,
     + 1.1219E+05, 1.1254E+05, 1.1290E+05, 1.1326E+05, 1.1362E+05,
     +  1.1397E+05, 1.1433E+05, 1.1470E+05, 1.1506E+05, 1.1542E+05,
     + 1.1578E+05, 1.1615E+05, 1.1651E+05, 1.1688E+05, 1.1725E+05,
     +  1.1762E+05, 1.1799E+05, 1.1836E+05, 1.1873E+05, 1.1910E+05,
     + 1.1947E+05, 1.1985E+05, 1.2022E+05, 1.2060E+05, 1.2098E+05,
     +  1.2135E+05, 1.2173E+05, 1.2211E+05, 1.2250E+05, 1.2288E+05,
     + 1.2326E+05, 1.2364E+05, 1.2403E+05, 1.2442E+05, 1.2480E+05,
     +  1.2519E+05, 1.2558E+05, 1.2597E+05, 1.2636E+05, 1.2675E+05,
     + 1.2715E+05, 1.2754E+05, 1.2793E+05, 1.2833E+05, 1.2873E+05,
     +  1.2913E+05, 1.2953E+05, 1.2993E+05, 1.3033E+05, 1.3073E+05,
     + 1.3113E+05, 1.3154E+05, 1.3194E+05, 1.3235E+05, 1.3276E+05,
     +  1.3316E+05, 1.3357E+05, 1.3398E+05, 1.3439E+05, 1.3481E+05,
     + 1.3522E+05, 1.3564E+05, 1.3605E+05, 1.3647E+05, 1.3689E+05,
     +  1.3730E+05, 1.3772E+05, 1.3815E+05, 1.3857E+05, 1.3899E+05,
     + 1.3942E+05, 1.3984E+05, 1.4027E+05, 1.4069E+05, 1.4112E+05,
     +  1.4155E+05, 1.4198E+05, 1.4241E+05, 1.4285E+05, 1.4328E+05,
     + 1.4372E+05, 1.4415E+05, 1.4459E+05, 1.4503E+05, 1.4547E+05,
     +  1.4591E+05, 1.4635E+05, 1.4679E+05, 1.4724E+05, 1.4768E+05,
     + 1.4813E+05, 1.4857E+05, 1.4902E+05, 1.4947E+05, 1.4992E+05,
     +  1.5037E+05, 1.5083E+05, 1.5128E+05, 1.5174E+05, 1.5219E+05,
     + 1.5265E+05, 1.5311E+05, 1.5357E+05, 1.5403E+05, 1.5449E+05,
     +  1.5495E+05, 1.5542E+05, 1.5588E+05, 1.5635E+05, 1.5682E+05,
     + 1.5729E+05, 1.5776E+05, 1.5823E+05, 1.5870E+05, 1.5917E+05,
     +  1.5965E+05, 1.6012E+05, 1.6060E+05, 1.6108E+05, 1.6156E+05,
     + 1.6204E+05, 1.6252E+05, 1.6300E+05, 1.6349E+05, 1.6397E+05,
     +  1.6446E+05, 1.6495E+05, 1.6543E+05, 1.6592E+05, 1.6642E+05,
     + 1.6691E+05, 1.6740E+05, 1.6790E+05, 1.6839E+05, 1.6889E+05,
     +  1.6939E+05, 1.6989E+05, 1.7039E+05, 1.7089E+05, 1.7139E+05,
     + 1.7190E+05, 1.7240E+05, 1.7291E+05, 1.7342E+05, 1.7393E+05,
     +  1.7444E+05, 1.7495E+05, 1.7547E+05, 1.7598E+05, 1.7650E+05,
     + 1.7701E+05, 1.7753E+05, 1.7805E+05, 1.7857E+05, 1.7909E+05,
     +  1.7962E+05, 1.8014E+05, 1.8067E+05, 1.8119E+05, 1.8172E+05,
     + 1.8225E+05, 1.8278E+05, 1.8332E+05, 1.8385E+05, 1.8438E+05,
     +  1.8492E+05, 1.8546E+05, 1.8600E+05, 1.8654E+05, 1.8708E+05,
     + 1.8762E+05, 1.8816E+05, 1.8871E+05, 1.8925E+05, 1.8980E+05,
     +  1.9035E+05, 1.9090E+05, 1.9145E+05, 1.9201E+05, 1.9256E+05,
     + 1.9312E+05, 1.9367E+05, 1.9423E+05, 1.9479E+05, 1.9535E+05,
     +  1.9592E+05, 1.9648E+05, 1.9704E+05, 1.9761E+05, 1.9818E+05,
     + 1.9875E+05, 1.9932E+05, 1.9989E+05, 2.0046E+05, 2.0104E+05,
     +  2.0161E+05, 2.0219E+05, 2.0277E+05, 2.0335E+05, 2.0393E+05,
     + 2.0451E+05, 2.0510E+05, 2.0568E+05, 2.0627E+05, 2.0686E+05,
     +  2.0745E+05, 2.0804E+05, 2.0863E+05, 2.0922E+05, 2.0982E+05,
     + 2.1042E+05, 2.1101E+05, 2.1161E+05, 2.1221E+05, 2.1282E+05,
     +  2.1342E+05, 2.1402E+05, 2.1463E+05, 2.1524E+05, 2.1585E+05,
     + 2.1646E+05, 2.1707E+05, 2.1768E+05, 2.1830E+05, 2.1892E+05,
     +  2.1953E+05, 2.2015E+05, 2.2077E+05, 2.2140E+05, 2.2202E+05,
     + 2.2264E+05, 2.2327E+05, 2.2390E+05, 2.2453E+05, 2.2516E+05,
     +  2.2579E+05, 2.2643E+05, 2.2706E+05, 2.2770E+05, 2.2834E+05,
     + 2.2898E+05, 2.2962E+05, 2.3026E+05, 2.3091E+05, 2.3155E+05,
     +  2.3220E+05, 2.3285E+05, 2.3350E+05, 2.3415E+05, 2.3480E+05,
     + 2.3546E+05, 2.3611E+05, 2.3677E+05, 2.3743E+05, 2.3809E+05,
     +  2.3875E+05, 2.3942E+05, 2.4008E+05, 2.4075E+05, 2.4142E+05,
     + 2.4209E+05, 2.4276E+05, 2.4343E+05, 2.4411E+05, 2.4478E+05,
     +  2.4546E+05, 2.4614E+05, 2.4682E+05, 2.4750E+05, 2.4819E+05,
     + 2.4887E+05, 2.4956E+05, 2.5025E+05, 2.5094E+05, 2.5163E+05,
     +  2.5232E+05, 2.5302E+05, 2.5372E+05, 2.5441E+05, 2.5511E+05,
     + 2.5582E+05, 2.5652E+05, 2.5722E+05, 2.5793E+05, 2.5864E+05,
     +  2.5935E+05, 2.6006E+05, 2.6077E+05, 2.6148E+05, 2.6220E+05,
     + 2.6292E+05, 2.6364E+05, 2.6436E+05, 2.6508E+05, 2.6580E+05,
     +  2.6653E+05, 2.6726E+05, 2.6798E+05, 2.6871E+05, 2.6945E+05,
     + 2.7018E+05, 2.7092E+05, 2.7165E+05, 2.7239E+05, 2.7313E+05,
     +  2.7387E+05, 2.7462E+05, 2.7536E+05, 2.7611E+05, 2.7686E+05,
     + 2.7761E+05, 2.7836E+05, 2.7911E+05, 2.7987E+05, 2.8063E+05,
     +  2.8139E+05, 2.8215E+05, 2.8291E+05, 2.8367E+05, 2.8444E+05,
     + 2.8521E+05, 2.8597E+05, 2.8675E+05, 2.8752E+05, 2.8829E+05,
     +  2.8907E+05, 2.8985E+05, 2.9063E+05, 2.9141E+05, 2.9219E+05,
     + 2.9297E+05, 2.9376E+05, 2.9455E+05, 2.9534E+05, 2.9613E+05,
     +  2.9692E+05, 2.9772E+05, 2.9851E+05, 2.9931E+05, 3.0011E+05,
     + 3.0092E+05, 3.0172E+05, 3.0253E+05, 3.0333E+05, 3.0414E+05,
     +  3.0495E+05, 3.0577E+05, 3.0658E+05, 3.0740E+05, 3.0821E+05,
     + 3.0903E+05, 3.0986E+05, 3.1068E+05, 3.1151E+05, 3.1233E+05,
     +  3.1316E+05, 3.1399E+05, 3.1483E+05, 3.1566E+05, 3.1650E+05,
     + 3.1733E+05, 3.1817E+05, 3.1902E+05, 3.1986E+05, 3.2071E+05,
     +  3.2155E+05, 3.2240E+05, 3.2325E+05, 3.2411E+05, 3.2496E+05,
     + 3.2582E+05, 3.2668E+05, 3.2754E+05, 3.2840E+05, 3.2926E+05,
     +  3.3013E+05, 3.3100E+05, 3.3187E+05, 3.3274E+05, 3.3361E+05,
     + 3.3449E+05, 3.3536E+05, 3.3624E+05, 3.3712E+05, 3.3801E+05,
     +  3.3889E+05, 3.3978E+05, 3.4067E+05, 3.4156E+05, 3.4245E+05,
     + 3.4335E+05, 3.4424E+05, 3.4514E+05, 3.4604E+05, 3.4694E+05,
     +  3.4785E+05, 3.4875E+05, 3.4966E+05, 3.5057E+05, 3.5148E+05,
     + 3.5240E+05, 3.5331E+05, 3.5423E+05, 3.5515E+05, 3.5607E+05,
     +  3.5700E+05, 3.5792E+05, 3.5885E+05, 3.5978E+05, 3.6071E+05,
     + 3.6165E+05, 3.6258E+05, 3.6352E+05, 3.6446E+05, 3.6540E+05,
     +  3.6634E+05, 3.6729E+05, 3.6824E+05, 3.6919E+05, 3.7014E+05,
     + 3.7109E+05, 3.7205E+05, 3.7301E+05, 3.7397E+05, 3.7493E+05,
     +  3.7589E+05, 3.7686E+05, 3.7783E+05, 3.7880E+05, 3.7977E+05,
     + 3.8074E+05, 3.8172E+05, 3.8270E+05, 3.8368E+05, 3.8466E+05,
     +  3.8565E+05, 3.8663E+05, 3.8762E+05, 3.8861E+05, 3.8961E+05,
     + 3.9060E+05, 3.9160E+05, 3.9260E+05, 3.9360E+05, 3.9460E+05,
     +  3.9561E+05, 3.9661E+05, 3.9762E+05, 3.9864E+05, 3.9965E+05,
     + 4.0067E+05, 4.0168E+05, 4.0271E+05, 4.0373E+05, 4.0475E+05,
     +  4.0578E+05, 4.0681E+05, 4.0784E+05, 4.0887E+05, 4.0991E+05,
     + 4.1095E+05, 4.1199E+05, 4.1303E+05, 4.1407E+05, 4.1512E+05,
     +  4.1617E+05, 4.1722E+05, 4.1827E+05, 4.1933E+05, 4.2038E+05,
     + 4.2144E+05, 4.2250E+05, 4.2357E+05, 4.2463E+05, 4.2570E+05,
     +  4.2677E+05, 4.2785E+05, 4.2892E+05, 4.3000E+05, 4.3108E+05,
     + 4.3216E+05, 4.3324E+05, 4.3433E+05, 4.3542E+05, 4.3651E+05,
     +  4.3760E+05, 4.3870E+05, 4.3979E+05, 4.4089E+05, 4.4199E+05,
     + 4.4310E+05, 4.4420E+05, 4.4531E+05, 4.4642E+05, 4.4754E+05,
     +  4.4865E+05, 4.4977E+05, 4.5089E+05, 4.5201E+05, 4.5314E+05,
     + 4.5427E+05, 4.5539E+05, 4.5653E+05, 4.5766E+05, 4.5880E+05,
     +  4.5994E+05, 4.6108E+05, 4.6222E+05, 4.6337E+05, 4.6451E+05,
     + 4.6566E+05, 4.6682E+05, 4.6797E+05, 4.6913E+05, 4.7029E+05,
     +  4.7145E+05, 4.7262E+05, 4.7378E+05, 4.7495E+05, 4.7612E+05,
     + 4.7730E+05, 4.7847E+05, 4.7965E+05, 4.8083E+05, 4.8202E+05,
     +  4.8320E+05, 4.8439E+05, 4.8558E+05, 4.8677E+05, 4.8797E+05,
     + 4.8917E+05, 4.9037E+05, 4.9157E+05, 4.9278E+05, 4.9398E+05,
     +  4.9519E+05, 4.9641E+05, 4.9762E+05, 4.9884E+05, 5.0006E+05,
     + 5.0128E+05, 5.0251E+05, 5.0373E+05, 5.0496E+05, 5.0620E+05,
     +  5.0743E+05, 5.0867E+05, 5.0991E+05, 5.1115E+05, 5.1239E+05,
     + 5.1364E+05, 5.1489E+05, 5.1614E+05, 5.1740E+05, 5.1865E+05,
     +  5.1991E+05, 5.2118E+05, 5.2244E+05, 5.2371E+05, 5.2498E+05,
     + 5.2625E+05, 5.2753E+05, 5.2880E+05, 5.3008E+05, 5.3137E+05,
     +  5.3265E+05, 5.3394E+05, 5.3523E+05, 5.3652E+05, 5.3782E+05,
     + 5.3911E+05, 5.4041E+05, 5.4172E+05, 5.4302E+05, 5.4433E+05,
     +  5.4564E+05, 5.4695E+05, 5.4827E+05, 5.4959E+05, 5.5091E+05,
     + 5.5223E+05, 5.5356E+05, 5.5489E+05, 5.5622E+05, 5.5756E+05,
     +  5.5889E+05, 5.6023E+05, 5.6157E+05, 5.6292E+05, 5.6427E+05,
     + 5.6562E+05, 5.6697E+05, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
c
c     *****************
      Subroutine QT_CS2   (                       
     I T,       ! temperature in K 
     I iso,       ! isotopologue code (HITRAN INDEX)
     O gsi,       ! state independent nuclear degeneracy factor
     O QT)       ! Total Internal Partition Function
 
      implicit DOUBLE PRECISION (a-h,o-z)
      parameter (NT=4510)
      INCLUDE 'SPECIES_2016.CMN'
      COMMON/Temperatures/tdat(NT)
      dimension Tmx(Max_ISO)
      
      dimension xgj( 4),QofT( 4,NT),Q(NT)
      data (Tmx(I),I=1,4)/3500.,3500.,3500.,3500./
      data xgj/ 1.,1.,4.,2./
c...      CS2
c...        --       222
      data (QofT( 1,J),J=1,2408)/ 3.3572E+00, 6.5397E+00, 9.7241E+00,
     +  1.2909E+01, 1.6094E+01, 1.9280E+01, 2.2465E+01, 2.5651E+01,
     + 2.8836E+01, 3.2022E+01, 3.5208E+01, 3.8394E+01, 4.1580E+01,
     +  4.4766E+01, 4.7952E+01, 5.1138E+01, 5.4325E+01, 5.7511E+01,
     + 6.0697E+01, 6.3884E+01, 7.0257E+01, 7.6631E+01, 8.3005E+01,
     +  8.9379E+01, 9.5753E+01, 1.0213E+02, 1.0850E+02, 1.1488E+02,
     + 1.2125E+02, 1.2763E+02, 1.3401E+02, 1.4038E+02, 1.4676E+02,
     +  1.5314E+02, 1.5952E+02, 1.6590E+02, 1.7228E+02, 1.7866E+02,
     + 1.8505E+02, 1.9143E+02, 1.9782E+02, 2.0422E+02, 2.1062E+02,
     +  2.1702E+02, 2.2343E+02, 2.2985E+02, 2.3628E+02, 2.4271E+02,
     + 2.4916E+02, 2.5562E+02, 2.6209E+02, 2.6858E+02, 2.7509E+02,
     +  2.8161E+02, 2.8815E+02, 2.9471E+02, 3.0130E+02, 3.0791E+02,
     + 3.1455E+02, 3.2121E+02, 3.2791E+02, 3.3463E+02, 3.4139E+02,
     +  3.4818E+02, 3.5501E+02, 3.6188E+02, 3.6878E+02, 3.7573E+02,
     + 3.8272E+02, 3.8976E+02, 3.9684E+02, 4.0398E+02, 4.1116E+02,
     +  4.1839E+02, 4.2568E+02, 4.3302E+02, 4.4042E+02, 4.4787E+02,
     + 4.5539E+02, 4.6297E+02, 4.7061E+02, 4.7831E+02, 4.8608E+02,
     +  4.9392E+02, 5.0182E+02, 5.0980E+02, 5.1785E+02, 5.2597E+02,
     + 5.3416E+02, 5.4243E+02, 5.5078E+02, 5.5921E+02, 5.6771E+02,
     +  5.7630E+02, 5.8497E+02, 5.9373E+02, 6.0257E+02, 6.1150E+02,
     + 6.2051E+02, 6.2962E+02, 6.3881E+02, 6.4810E+02, 6.5748E+02,
     +  6.6696E+02, 6.7653E+02, 6.8620E+02, 6.9596E+02, 7.0583E+02,
     + 7.1579E+02, 7.2586E+02, 7.3603E+02, 7.4631E+02, 7.5669E+02,
     +  7.6718E+02, 7.7777E+02, 7.8848E+02, 7.9930E+02, 8.1022E+02,
     + 8.2126E+02, 8.3242E+02, 8.4369E+02, 8.5507E+02, 8.6657E+02,
     +  8.7820E+02, 8.8994E+02, 9.0180E+02, 9.1379E+02, 9.2590E+02,
     + 9.3813E+02, 9.5049E+02, 9.6298E+02, 9.7559E+02, 9.8834E+02,
     +  1.0012E+03, 1.0142E+03, 1.0274E+03, 1.0406E+03, 1.0540E+03,
     + 1.0676E+03, 1.0813E+03, 1.0951E+03, 1.1091E+03, 1.1232E+03,
     +  1.1374E+03, 1.1518E+03, 1.1663E+03, 1.1810E+03, 1.1959E+03,
     + 1.2108E+03, 1.2260E+03, 1.2413E+03, 1.2567E+03, 1.2723E+03,
     +  1.2880E+03, 1.3039E+03, 1.3200E+03, 1.3362E+03, 1.3526E+03,
     + 1.3691E+03, 1.3858E+03, 1.4027E+03, 1.4197E+03, 1.4369E+03,
     +  1.4543E+03, 1.4718E+03, 1.4895E+03, 1.5074E+03, 1.5254E+03,
     + 1.5437E+03, 1.5621E+03, 1.5806E+03, 1.5994E+03, 1.6183E+03,
     +  1.6375E+03, 1.6568E+03, 1.6762E+03, 1.6959E+03, 1.7158E+03,
     + 1.7358E+03, 1.7561E+03, 1.7765E+03, 1.7971E+03, 1.8179E+03,
     +  1.8389E+03, 1.8601E+03, 1.8815E+03, 1.9031E+03, 1.9249E+03,
     + 1.9469E+03, 1.9691E+03, 1.9916E+03, 2.0142E+03, 2.0370E+03,
     +  2.0601E+03, 2.0833E+03, 2.1068E+03, 2.1304E+03, 2.1543E+03,
     + 2.1784E+03, 2.2028E+03, 2.2273E+03, 2.2521E+03, 2.2771E+03,
     +  2.3023E+03, 2.3278E+03, 2.3534E+03, 2.3793E+03, 2.4055E+03,
     + 2.4319E+03, 2.4585E+03, 2.4853E+03, 2.5124E+03, 2.5397E+03,
     +  2.5673E+03, 2.5951E+03, 2.6231E+03, 2.6514E+03, 2.6799E+03,
     + 2.7087E+03, 2.7378E+03, 2.7671E+03, 2.7966E+03, 2.8264E+03,
     +  2.8565E+03, 2.8868E+03, 2.9174E+03, 2.9482E+03, 2.9793E+03,
     + 3.0107E+03, 3.0424E+03, 3.0743E+03, 3.1065E+03, 3.1389E+03,
     +  3.1716E+03, 3.2047E+03, 3.2379E+03, 3.2715E+03, 3.3054E+03,
     + 3.3395E+03, 3.3739E+03, 3.4087E+03, 3.4437E+03, 3.4789E+03,
     +  3.5145E+03, 3.5504E+03, 3.5866E+03, 3.6231E+03, 3.6599E+03,
     + 3.6969E+03, 3.7343E+03, 3.7720E+03, 3.8100E+03, 3.8484E+03,
     +  3.8870E+03, 3.9259E+03, 3.9652E+03, 4.0048E+03, 4.0447E+03,
     + 4.0849E+03, 4.1254E+03, 4.1663E+03, 4.2075E+03, 4.2490E+03,
     +  4.2909E+03, 4.3331E+03, 4.3756E+03, 4.4185E+03, 4.4617E+03,
     + 4.5053E+03, 4.5492E+03, 4.5935E+03, 4.6381E+03, 4.6830E+03,
     +  4.7283E+03, 4.7740E+03, 4.8200E+03, 4.8664E+03, 4.9131E+03,
     + 4.9602E+03, 5.0077E+03, 5.0556E+03, 5.1038E+03, 5.1524E+03,
     +  5.2013E+03, 5.2507E+03, 5.3004E+03, 5.3505E+03, 5.4010E+03,
     + 5.4519E+03, 5.5031E+03, 5.5548E+03, 5.6068E+03, 5.6593E+03,
     +  5.7121E+03, 5.7653E+03, 5.8190E+03, 5.8730E+03, 5.9275E+03,
     + 5.9824E+03, 6.0376E+03, 6.0933E+03, 6.1494E+03, 6.2060E+03,
     +  6.2629E+03, 6.3203E+03, 6.3781E+03, 6.4363E+03, 6.4950E+03,
     + 6.5541E+03, 6.6136E+03, 6.6736E+03, 6.7340E+03, 6.7948E+03,
     +  6.8561E+03, 6.9179E+03, 6.9801E+03, 7.0427E+03, 7.1058E+03,
     + 7.1694E+03, 7.2334E+03, 7.2979E+03, 7.3629E+03, 7.4283E+03,
     +  7.4942E+03, 7.5606E+03, 7.6275E+03, 7.6948E+03, 7.7626E+03,
     + 7.8309E+03, 7.8997E+03, 7.9690E+03, 8.0387E+03, 8.1090E+03,
     +  8.1798E+03, 8.2510E+03, 8.3228E+03, 8.3951E+03, 8.4678E+03,
     + 8.5411E+03, 8.6149E+03, 8.6893E+03, 8.7641E+03, 8.8395E+03,
     +  8.9154E+03, 8.9918E+03, 9.0687E+03, 9.1462E+03, 9.2243E+03,
     + 9.3028E+03, 9.3819E+03, 9.4616E+03, 9.5418E+03, 9.6225E+03,
     +  9.7038E+03, 9.7857E+03, 9.8681E+03, 9.9511E+03, 1.0035E+04,
     + 1.0119E+04, 1.0203E+04, 1.0289E+04, 1.0375E+04, 1.0461E+04,
     +  1.0548E+04, 1.0636E+04, 1.0724E+04, 1.0813E+04, 1.0902E+04,
     + 1.0992E+04, 1.1082E+04, 1.1174E+04, 1.1265E+04, 1.1358E+04,
     +  1.1451E+04, 1.1544E+04, 1.1639E+04, 1.1734E+04, 1.1829E+04,
     + 1.1925E+04, 1.2022E+04, 1.2119E+04, 1.2217E+04, 1.2316E+04,
     +  1.2415E+04, 1.2515E+04, 1.2616E+04, 1.2717E+04, 1.2819E+04,
     + 1.2922E+04, 1.3025E+04, 1.3129E+04, 1.3234E+04, 1.3339E+04,
     +  1.3445E+04, 1.3552E+04, 1.3659E+04, 1.3767E+04, 1.3876E+04,
     + 1.3985E+04, 1.4095E+04, 1.4206E+04, 1.4317E+04, 1.4430E+04,
     +  1.4543E+04, 1.4656E+04, 1.4771E+04, 1.4886E+04, 1.5002E+04,
     + 1.5118E+04, 1.5235E+04, 1.5353E+04, 1.5472E+04, 1.5592E+04,
     +  1.5712E+04, 1.5833E+04, 1.5955E+04, 1.6077E+04, 1.6201E+04,
     + 1.6325E+04, 1.6449E+04, 1.6575E+04, 1.6701E+04, 1.6828E+04,
     +  1.6956E+04, 1.7085E+04, 1.7215E+04, 1.7345E+04, 1.7476E+04,
     + 1.7608E+04, 1.7741E+04, 1.7874E+04, 1.8008E+04, 1.8144E+04,
     +  1.8280E+04, 1.8416E+04, 1.8554E+04, 1.8692E+04, 1.8832E+04,
     + 1.8972E+04, 1.9113E+04, 1.9255E+04, 1.9397E+04, 1.9541E+04,
     +  1.9685E+04, 1.9831E+04, 1.9977E+04, 2.0124E+04, 2.0272E+04,
     + 2.0420E+04, 2.0570E+04, 2.0720E+04, 2.0872E+04, 2.1024E+04,
     +  2.1177E+04, 2.1331E+04, 2.1486E+04, 2.1642E+04, 2.1799E+04,
     + 2.1957E+04, 2.2116E+04, 2.2275E+04, 2.2436E+04, 2.2597E+04,
     +  2.2760E+04, 2.2923E+04, 2.3087E+04, 2.3253E+04, 2.3419E+04,
     + 2.3586E+04, 2.3754E+04, 2.3923E+04, 2.4094E+04, 2.4265E+04,
     +  2.4437E+04, 2.4610E+04, 2.4784E+04, 2.4959E+04, 2.5135E+04,
     + 2.5312E+04, 2.5490E+04, 2.5669E+04, 2.5849E+04, 2.6030E+04,
     +  2.6212E+04, 2.6395E+04, 2.6580E+04, 2.6765E+04, 2.6951E+04,
     + 2.7138E+04, 2.7327E+04, 2.7516E+04, 2.7707E+04, 2.7898E+04,
     +  2.8091E+04, 2.8285E+04, 2.8479E+04, 2.8675E+04, 2.8872E+04,
     + 2.9070E+04, 2.9269E+04, 2.9470E+04, 2.9671E+04, 2.9873E+04,
     +  3.0077E+04, 3.0282E+04, 3.0488E+04, 3.0695E+04, 3.0903E+04,
     + 3.1112E+04, 3.1322E+04, 3.1534E+04, 3.1747E+04, 3.1960E+04,
     +  3.2175E+04, 3.2392E+04, 3.2609E+04, 3.2828E+04, 3.3047E+04,
     + 3.3268E+04, 3.3490E+04, 3.3714E+04, 3.3938E+04, 3.4164E+04,
     +  3.4391E+04, 3.4619E+04, 3.4848E+04, 3.5079E+04, 3.5311E+04,
     + 3.5544E+04, 3.5778E+04, 3.6014E+04, 3.6250E+04, 3.6489E+04,
     +  3.6728E+04, 3.6968E+04, 3.7210E+04, 3.7454E+04, 3.7698E+04,
     + 3.7944E+04, 3.8191E+04, 3.8439E+04, 3.8689E+04, 3.8940E+04,
     +  3.9192E+04, 3.9445E+04, 3.9700E+04, 3.9957E+04, 4.0214E+04,
     + 4.0473E+04, 4.0733E+04, 4.0995E+04, 4.1258E+04, 4.1522E+04,
     +  4.1788E+04, 4.2055E+04, 4.2323E+04, 4.2593E+04, 4.2864E+04,
     + 4.3137E+04, 4.3411E+04, 4.3687E+04, 4.3963E+04, 4.4242E+04,
     +  4.4521E+04, 4.4802E+04, 4.5085E+04, 4.5369E+04, 4.5654E+04,
     + 4.5941E+04, 4.6230E+04, 4.6519E+04, 4.6811E+04, 4.7103E+04,
     +  4.7397E+04, 4.7693E+04, 4.7990E+04, 4.8289E+04, 4.8589E+04,
     + 4.8891E+04, 4.9194E+04, 4.9499E+04, 4.9805E+04, 5.0113E+04,
     +  5.0422E+04, 5.0733E+04, 5.1045E+04, 5.1359E+04, 5.1675E+04,
     + 5.1992E+04, 5.2310E+04, 5.2631E+04, 5.2952E+04, 5.3276E+04,
     +  5.3601E+04, 5.3927E+04, 5.4256E+04, 5.4585E+04, 5.4917E+04,
     + 5.5250E+04, 5.5584E+04, 5.5921E+04, 5.6259E+04, 5.6598E+04,
     +  5.6939E+04, 5.7282E+04, 5.7627E+04, 5.7973E+04, 5.8321E+04,
     + 5.8671E+04, 5.9022E+04, 5.9375E+04, 5.9730E+04, 6.0086E+04,
     +  6.0444E+04, 6.0804E+04, 6.1166E+04, 6.1529E+04, 6.1894E+04,
     + 6.2261E+04, 6.2629E+04, 6.3000E+04, 6.3372E+04, 6.3746E+04,
     +  6.4121E+04, 6.4499E+04, 6.4878E+04, 6.5259E+04, 6.5642E+04,
     + 6.6026E+04, 6.6413E+04, 6.6801E+04, 6.7191E+04, 6.7583E+04,
     +  6.7977E+04, 6.8373E+04, 6.8770E+04, 6.9170E+04, 6.9571E+04,
     + 6.9974E+04, 7.0379E+04, 7.0786E+04, 7.1195E+04, 7.1605E+04,
     +  7.2018E+04, 7.2432E+04, 7.2849E+04, 7.3267E+04, 7.3688E+04,
     + 7.4110E+04, 7.4534E+04, 7.4960E+04, 7.5388E+04, 7.5818E+04,
     +  7.6251E+04, 7.6685E+04, 7.7121E+04, 7.7559E+04, 7.7999E+04,
     + 7.8441E+04, 7.8885E+04, 7.9331E+04, 7.9779E+04, 8.0230E+04,
     +  8.0682E+04, 8.1136E+04, 8.1593E+04, 8.2051E+04, 8.2512E+04,
     + 8.2974E+04, 8.3439E+04, 8.3906E+04, 8.4375E+04, 8.4846E+04,
     +  8.5319E+04, 8.5794E+04, 8.6271E+04, 8.6751E+04, 8.7233E+04,
     + 8.7717E+04, 8.8203E+04, 8.8691E+04, 8.9181E+04, 8.9674E+04,
     +  9.0168E+04, 9.0665E+04, 9.1165E+04, 9.1666E+04, 9.2169E+04,
     + 9.2675E+04, 9.3183E+04, 9.3694E+04, 9.4206E+04, 9.4721E+04,
     +  9.5238E+04, 9.5758E+04, 9.6279E+04, 9.6803E+04, 9.7329E+04,
     + 9.7858E+04, 9.8389E+04, 9.8922E+04, 9.9458E+04, 9.9995E+04,
     +  1.0054E+05, 1.0108E+05, 1.0162E+05, 1.0217E+05, 1.0272E+05,
     + 1.0327E+05, 1.0383E+05, 1.0438E+05, 1.0494E+05, 1.0550E+05,
     +  1.0607E+05, 1.0663E+05, 1.0720E+05, 1.0778E+05, 1.0835E+05,
     + 1.0893E+05, 1.0950E+05, 1.1009E+05, 1.1067E+05, 1.1126E+05,
     +  1.1184E+05, 1.1244E+05, 1.1303E+05, 1.1363E+05, 1.1422E+05,
     + 1.1483E+05, 1.1543E+05, 1.1604E+05, 1.1664E+05, 1.1726E+05,
     +  1.1787E+05, 1.1849E+05, 1.1911E+05, 1.1973E+05, 1.2035E+05,
     + 1.2098E+05, 1.2161E+05, 1.2224E+05, 1.2288E+05, 1.2352E+05,
     +  1.2416E+05, 1.2480E+05, 1.2545E+05, 1.2609E+05, 1.2675E+05,
     + 1.2740E+05, 1.2806E+05, 1.2872E+05, 1.2938E+05, 1.3004E+05,
     +  1.3071E+05, 1.3138E+05, 1.3205E+05, 1.3273E+05, 1.3341E+05,
     + 1.3409E+05, 1.3478E+05, 1.3546E+05, 1.3615E+05, 1.3685E+05,
     +  1.3754E+05, 1.3824E+05, 1.3894E+05, 1.3964E+05, 1.4035E+05,
     + 1.4106E+05, 1.4178E+05, 1.4249E+05, 1.4321E+05, 1.4393E+05,
     +  1.4466E+05, 1.4538E+05, 1.4611E+05, 1.4685E+05, 1.4758E+05,
     + 1.4832E+05, 1.4907E+05, 1.4981E+05, 1.5056E+05, 1.5131E+05,
     +  1.5206E+05, 1.5282E+05, 1.5358E+05, 1.5435E+05, 1.5511E+05,
     + 1.5588E+05, 1.5665E+05, 1.5743E+05, 1.5821E+05, 1.5899E+05,
     +  1.5978E+05, 1.6056E+05, 1.6136E+05, 1.6215E+05, 1.6295E+05,
     + 1.6375E+05, 1.6455E+05, 1.6536E+05, 1.6617E+05, 1.6698E+05,
     +  1.6780E+05, 1.6862E+05, 1.6944E+05, 1.7027E+05, 1.7110E+05,
     + 1.7193E+05, 1.7277E+05, 1.7361E+05, 1.7445E+05, 1.7530E+05,
     +  1.7615E+05, 1.7700E+05, 1.7785E+05, 1.7871E+05, 1.7958E+05,
     + 1.8044E+05, 1.8131E+05, 1.8218E+05, 1.8306E+05, 1.8394E+05,
     +  1.8482E+05, 1.8571E+05, 1.8660E+05, 1.8749E+05, 1.8839E+05,
     + 1.8929E+05, 1.9019E+05, 1.9110E+05, 1.9201E+05, 1.9292E+05,
     +  1.9384E+05, 1.9476E+05, 1.9569E+05, 1.9662E+05, 1.9755E+05,
     + 1.9848E+05, 1.9942E+05, 2.0036E+05, 2.0131E+05, 2.0226E+05,
     +  2.0321E+05, 2.0417E+05, 2.0513E+05, 2.0609E+05, 2.0706E+05,
     + 2.0803E+05, 2.0901E+05, 2.0999E+05, 2.1097E+05, 2.1195E+05,
     +  2.1294E+05, 2.1394E+05, 2.1493E+05, 2.1593E+05, 2.1694E+05,
     + 2.1795E+05, 2.1896E+05, 2.1998E+05, 2.2100E+05, 2.2202E+05,
     +  2.2305E+05, 2.2408E+05, 2.2511E+05, 2.2615E+05, 2.2720E+05,
     + 2.2824E+05, 2.2929E+05, 2.3035E+05, 2.3140E+05, 2.3247E+05,
     +  2.3353E+05, 2.3460E+05, 2.3568E+05, 2.3676E+05, 2.3784E+05,
     + 2.3892E+05, 2.4001E+05, 2.4111E+05, 2.4221E+05, 2.4331E+05,
     +  2.4441E+05, 2.4552E+05, 2.4664E+05, 2.4776E+05, 2.4888E+05,
     + 2.5000E+05, 2.5114E+05, 2.5227E+05, 2.5341E+05, 2.5455E+05,
     +  2.5570E+05, 2.5685E+05, 2.5800E+05, 2.5916E+05, 2.6033E+05,
     + 2.6150E+05, 2.6267E+05, 2.6384E+05, 2.6503E+05, 2.6621E+05,
     +  2.6740E+05, 2.6859E+05, 2.6979E+05, 2.7099E+05, 2.7220E+05,
     + 2.7341E+05, 2.7462E+05, 2.7584E+05, 2.7707E+05, 2.7830E+05,
     +  2.7953E+05, 2.8076E+05, 2.8201E+05, 2.8325E+05, 2.8450E+05,
     + 2.8576E+05, 2.8701E+05, 2.8828E+05, 2.8955E+05, 2.9082E+05,
     +  2.9210E+05, 2.9338E+05, 2.9466E+05, 2.9595E+05, 2.9725E+05,
     + 2.9855E+05, 2.9985E+05, 3.0116E+05, 3.0247E+05, 3.0379E+05,
     +  3.0511E+05, 3.0644E+05, 3.0777E+05, 3.0911E+05, 3.1045E+05,
     + 3.1180E+05, 3.1315E+05, 3.1450E+05, 3.1586E+05, 3.1723E+05,
     +  3.1860E+05, 3.1997E+05, 3.2135E+05, 3.2273E+05, 3.2412E+05,
     + 3.2551E+05, 3.2691E+05, 3.2832E+05, 3.2972E+05, 3.3114E+05,
     +  3.3255E+05, 3.3398E+05, 3.3540E+05, 3.3684E+05, 3.3827E+05,
     + 3.3972E+05, 3.4116E+05, 3.4261E+05, 3.4407E+05, 3.4553E+05,
     +  3.4700E+05, 3.4847E+05, 3.4995E+05, 3.5143E+05, 3.5292E+05,
     + 3.5441E+05, 3.5591E+05, 3.5741E+05, 3.5892E+05, 3.6043E+05,
     +  3.6195E+05, 3.6347E+05, 3.6500E+05, 3.6653E+05, 3.6807E+05,
     + 3.6961E+05, 3.7116E+05, 3.7272E+05, 3.7428E+05, 3.7584E+05,
     +  3.7741E+05, 3.7898E+05, 3.8056E+05, 3.8215E+05, 3.8374E+05,
     + 3.8534E+05, 3.8694E+05, 3.8855E+05, 3.9016E+05, 3.9178E+05,
     +  3.9340E+05, 3.9503E+05, 3.9666E+05, 3.9830E+05, 3.9995E+05,
     + 4.0160E+05, 4.0325E+05, 4.0491E+05, 4.0658E+05, 4.0825E+05,
     +  4.0993E+05, 4.1161E+05, 4.1330E+05, 4.1500E+05, 4.1670E+05,
     + 4.1840E+05, 4.2011E+05, 4.2183E+05, 4.2355E+05, 4.2528E+05,
     +  4.2701E+05, 4.2875E+05, 4.3050E+05, 4.3225E+05, 4.3401E+05,
     + 4.3577E+05, 4.3754E+05, 4.3931E+05, 4.4109E+05, 4.4288E+05,
     +  4.4467E+05, 4.4646E+05, 4.4827E+05, 4.5008E+05, 4.5189E+05,
     + 4.5371E+05, 4.5554E+05, 4.5737E+05, 4.5921E+05, 4.6105E+05,
     +  4.6290E+05, 4.6476E+05, 4.6662E+05, 4.6849E+05, 4.7036E+05,
     + 4.7225E+05, 4.7413E+05, 4.7602E+05, 4.7792E+05, 4.7983E+05,
     +  4.8174E+05, 4.8365E+05, 4.8558E+05, 4.8751E+05, 4.8944E+05,
     + 4.9138E+05, 4.9333E+05, 4.9528E+05, 4.9724E+05, 4.9921E+05,
     +  5.0118E+05, 5.0316E+05, 5.0515E+05, 5.0714E+05, 5.0914E+05,
     + 5.1114E+05, 5.1315E+05, 5.1517E+05, 5.1719E+05, 5.1922E+05,
     +  5.2126E+05, 5.2330E+05, 5.2535E+05, 5.2740E+05, 5.2946E+05,
     + 5.3153E+05, 5.3361E+05, 5.3569E+05, 5.3778E+05, 5.3987E+05,
     +  5.4197E+05, 5.4408E+05, 5.4619E+05, 5.4831E+05, 5.5044E+05,
     + 5.5257E+05, 5.5472E+05, 5.5686E+05, 5.5902E+05, 5.6118E+05,
     +  5.6334E+05, 5.6552E+05, 5.6770E+05, 5.6989E+05, 5.7208E+05,
     + 5.7428E+05, 5.7649E+05, 5.7870E+05, 5.8093E+05, 5.8315E+05,
     +  5.8539E+05, 5.8763E+05, 5.8988E+05, 5.9214E+05, 5.9440E+05,
     + 5.9667E+05, 5.9895E+05, 6.0123E+05, 6.0352E+05, 6.0582E+05,
     +  6.0812E+05, 6.1044E+05, 6.1276E+05, 6.1508E+05, 6.1742E+05,
     + 6.1976E+05, 6.2210E+05, 6.2446E+05, 6.2682E+05, 6.2919E+05,
     +  6.3157E+05, 6.3395E+05, 6.3634E+05, 6.3874E+05, 6.4114E+05,
     + 6.4356E+05, 6.4597E+05, 6.4840E+05, 6.5084E+05, 6.5328E+05,
     +  6.5573E+05, 6.5818E+05, 6.6065E+05, 6.6312E+05, 6.6560E+05,
     + 6.6808E+05, 6.7058E+05, 6.7308E+05, 6.7559E+05, 6.7810E+05,
     +  6.8063E+05, 6.8316E+05, 6.8570E+05, 6.8824E+05, 6.9080E+05,
     + 6.9336E+05, 6.9593E+05, 6.9851E+05, 7.0109E+05, 7.0368E+05,
     +  7.0628E+05, 7.0889E+05, 7.1151E+05, 7.1413E+05, 7.1676E+05,
     + 7.1940E+05, 7.2205E+05, 7.2470E+05, 7.2736E+05, 7.3003E+05,
     +  7.3271E+05, 7.3540E+05, 7.3809E+05, 7.4079E+05, 7.4350E+05,
     + 7.4622E+05, 7.4895E+05, 7.5168E+05, 7.5442E+05, 7.5717E+05,
     +  7.5993E+05, 7.6270E+05, 7.6547E+05, 7.6825E+05, 7.7104E+05,
     + 7.7384E+05, 7.7665E+05, 7.7946E+05, 7.8229E+05, 7.8512E+05,
     +  7.8796E+05, 7.9081E+05, 7.9366E+05, 7.9653E+05, 7.9940E+05,
     + 8.0228E+05, 8.0517E+05, 8.0807E+05, 8.1097E+05, 8.1389E+05,
     +  8.1681E+05, 8.1974E+05, 8.2268E+05, 8.2563E+05, 8.2859E+05,
     + 8.3155E+05, 8.3453E+05, 8.3751E+05, 8.4050E+05, 8.4350E+05,
     +  8.4651E+05, 8.4952E+05, 8.5255E+05, 8.5558E+05, 8.5863E+05,
     + 8.6168E+05, 8.6474E+05, 8.6781E+05, 8.7089E+05, 8.7397E+05,
     +  8.7707E+05, 8.8017E+05, 8.8328E+05, 8.8641E+05, 8.8954E+05,
     + 8.9268E+05, 8.9582E+05, 8.9898E+05, 9.0215E+05, 9.0532E+05,
     +  9.0851E+05, 9.1170E+05, 9.1490E+05, 9.1812E+05, 9.2134E+05,
     + 9.2457E+05, 9.2780E+05, 9.3105E+05, 9.3431E+05, 9.3757E+05,
     +  9.4085E+05, 9.4413E+05, 9.4743E+05, 9.5073E+05, 9.5404E+05,
     + 9.5736E+05, 9.6070E+05, 9.6404E+05, 9.6738E+05, 9.7074E+05,
     +  9.7411E+05, 9.7749E+05, 9.8088E+05, 9.8427E+05, 9.8768E+05,
     + 9.9109E+05, 9.9452E+05, 9.9795E+05, 1.0014E+06, 1.0048E+06,
     +  1.0083E+06, 1.0118E+06, 1.0153E+06, 1.0188E+06, 1.0223E+06,
     + 1.0258E+06, 1.0293E+06, 1.0328E+06, 1.0364E+06, 1.0399E+06,
     +  1.0435E+06, 1.0470E+06, 1.0506E+06, 1.0542E+06, 1.0578E+06,
     + 1.0614E+06, 1.0650E+06, 1.0687E+06, 1.0723E+06, 1.0760E+06,
     +  1.0796E+06, 1.0833E+06, 1.0870E+06, 1.0907E+06, 1.0944E+06,
     + 1.0981E+06, 1.1018E+06, 1.1055E+06, 1.1092E+06, 1.1130E+06,
     +  1.1168E+06, 1.1205E+06, 1.1243E+06, 1.1281E+06, 1.1319E+06,
     + 1.1357E+06, 1.1395E+06, 1.1434E+06, 1.1472E+06, 1.1511E+06,
     +  1.1549E+06, 1.1588E+06, 1.1627E+06, 1.1666E+06, 1.1705E+06,
     + 1.1744E+06, 1.1783E+06, 1.1823E+06, 1.1862E+06, 1.1902E+06,
     +  1.1941E+06, 1.1981E+06, 1.2021E+06, 1.2061E+06, 1.2101E+06,
     + 1.2141E+06, 1.2182E+06, 1.2222E+06, 1.2263E+06, 1.2303E+06,
     +  1.2344E+06, 1.2385E+06, 1.2426E+06, 1.2467E+06, 1.2508E+06,
     + 1.2549E+06, 1.2591E+06, 1.2632E+06, 1.2674E+06, 1.2716E+06,
     +  1.2757E+06, 1.2799E+06, 1.2841E+06, 1.2884E+06, 1.2926E+06,
     + 1.2968E+06, 1.3011E+06, 1.3054E+06, 1.3096E+06, 1.3139E+06,
     +  1.3182E+06, 1.3225E+06, 1.3268E+06, 1.3312E+06, 1.3355E+06,
     + 1.3399E+06, 1.3442E+06, 1.3486E+06, 1.3530E+06, 1.3574E+06,
     +  1.3618E+06, 1.3662E+06, 1.3707E+06, 1.3751E+06, 1.3796E+06,
     + 1.3840E+06, 1.3885E+06, 1.3930E+06, 1.3975E+06, 1.4020E+06,
     +  1.4065E+06, 1.4111E+06, 1.4156E+06, 1.4202E+06, 1.4248E+06,
     + 1.4294E+06, 1.4340E+06, 1.4386E+06, 1.4432E+06, 1.4478E+06,
     +  1.4525E+06, 1.4571E+06, 1.4618E+06, 1.4665E+06, 1.4712E+06,
     + 1.4759E+06, 1.4806E+06, 1.4853E+06, 1.4901E+06, 1.4948E+06,
     +  1.4996E+06, 1.5044E+06, 1.5091E+06, 1.5139E+06, 1.5188E+06,
     + 1.5236E+06, 1.5284E+06, 1.5333E+06, 1.5381E+06, 1.5430E+06,
     +  1.5479E+06, 1.5528E+06, 1.5577E+06, 1.5627E+06, 1.5676E+06,
     + 1.5726E+06, 1.5775E+06, 1.5825E+06, 1.5875E+06, 1.5925E+06,
     +  1.5975E+06, 1.6025E+06, 1.6076E+06, 1.6126E+06, 1.6177E+06,
     + 1.6228E+06, 1.6279E+06, 1.6330E+06, 1.6381E+06, 1.6432E+06,
     +  1.6484E+06, 1.6535E+06, 1.6587E+06, 1.6639E+06, 1.6691E+06,
     + 1.6743E+06, 1.6795E+06, 1.6847E+06, 1.6900E+06, 1.6952E+06,
     +  1.7005E+06, 1.7058E+06, 1.7111E+06, 1.7164E+06, 1.7217E+06,
     + 1.7271E+06, 1.7324E+06, 1.7378E+06, 1.7432E+06, 1.7486E+06,
     +  1.7540E+06, 1.7594E+06, 1.7648E+06, 1.7703E+06, 1.7757E+06,
     + 1.7812E+06, 1.7867E+06, 1.7922E+06, 1.7977E+06, 1.8032E+06,
     +  1.8088E+06, 1.8143E+06, 1.8199E+06, 1.8255E+06, 1.8311E+06,
     + 1.8367E+06, 1.8423E+06, 1.8480E+06, 1.8536E+06, 1.8593E+06,
     +  1.8650E+06, 1.8707E+06, 1.8764E+06, 1.8821E+06, 1.8878E+06,
     + 1.8936E+06, 1.8993E+06, 1.9051E+06, 1.9109E+06, 1.9167E+06,
     +  1.9225E+06, 1.9284E+06, 1.9342E+06, 1.9401E+06, 1.9460E+06,
     + 1.9519E+06, 1.9578E+06, 1.9637E+06, 1.9696E+06, 1.9756E+06,
     +  1.9815E+06, 1.9875E+06, 1.9935E+06, 1.9995E+06, 2.0055E+06,
     + 2.0116E+06, 2.0176E+06, 2.0237E+06, 2.0298E+06, 2.0359E+06,
     +  2.0420E+06, 2.0481E+06, 2.0543E+06, 2.0604E+06, 2.0666E+06,
     + 2.0728E+06, 2.0790E+06, 2.0852E+06, 2.0914E+06, 2.0976E+06,
     +  2.1039E+06, 2.1102E+06, 2.1165E+06, 2.1228E+06, 2.1291E+06,
     + 2.1354E+06, 2.1418E+06, 2.1481E+06, 2.1545E+06, 2.1609E+06,
     +  2.1673E+06, 2.1738E+06, 2.1802E+06, 2.1867E+06, 2.1931E+06,
     + 2.1996E+06, 2.2061E+06, 2.2126E+06, 2.2192E+06, 2.2257E+06,
     +  2.2323E+06, 2.2389E+06, 2.2455E+06, 2.2521E+06, 2.2587E+06,
     + 2.2653E+06, 2.2720E+06, 2.2787E+06, 2.2854E+06, 2.2921E+06,
     +  2.2988E+06, 2.3055E+06, 2.3123E+06, 2.3190E+06, 2.3258E+06,
     + 2.3326E+06, 2.3395E+06, 2.3463E+06, 2.3531E+06, 2.3600E+06,
     +  2.3669E+06, 2.3738E+06, 2.3807E+06, 2.3876E+06, 2.3946E+06,
     + 2.4015E+06, 2.4085E+06, 2.4155E+06, 2.4225E+06, 2.4295E+06,
     +  2.4366E+06, 2.4436E+06, 2.4507E+06, 2.4578E+06, 2.4649E+06,
     + 2.4721E+06, 2.4792E+06, 2.4864E+06, 2.4935E+06, 2.5007E+06,
     +  2.5079E+06, 2.5152E+06, 2.5224E+06, 2.5297E+06, 2.5369E+06,
     + 2.5442E+06, 2.5515E+06, 2.5589E+06, 2.5662E+06, 2.5736E+06,
     +  2.5810E+06, 2.5883E+06, 2.5958E+06, 2.6032E+06, 2.6106E+06,
     + 2.6181E+06, 2.6256E+06, 2.6331E+06, 2.6406E+06, 2.6481E+06,
     +  2.6557E+06, 2.6632E+06, 2.6708E+06, 2.6784E+06, 2.6860E+06,
     + 2.6937E+06, 2.7013E+06, 2.7090E+06, 2.7167E+06, 2.7244E+06,
     +  2.7321E+06, 2.7399E+06, 2.7476E+06, 2.7554E+06, 2.7632E+06,
     + 2.7710E+06, 2.7788E+06, 2.7867E+06, 2.7945E+06, 2.8024E+06,
     +  2.8103E+06, 2.8183E+06, 2.8262E+06, 2.8341E+06, 2.8421E+06,
     + 2.8501E+06, 2.8581E+06, 2.8661E+06, 2.8742E+06, 2.8823E+06,
     +  2.8903E+06, 2.8984E+06, 2.9066E+06, 2.9147E+06, 2.9228E+06,
     + 2.9310E+06, 2.9392E+06, 2.9474E+06, 2.9556E+06, 2.9639E+06,
     +  2.9722E+06, 2.9804E+06, 2.9887E+06, 2.9971E+06, 3.0054E+06,
     + 3.0138E+06, 3.0221E+06, 3.0305E+06, 3.0389E+06, 3.0474E+06,
     +  3.0558E+06, 3.0643E+06, 3.0728E+06, 3.0813E+06, 3.0898E+06,
     + 3.0984E+06, 3.1069E+06, 3.1155E+06, 3.1241E+06, 3.1328E+06,
     +  3.1414E+06, 3.1501E+06, 3.1587E+06, 3.1674E+06, 3.1762E+06,
     + 3.1849E+06, 3.1936E+06, 3.2024E+06, 3.2112E+06, 3.2200E+06,
     +  3.2289E+06, 3.2377E+06, 3.2466E+06, 3.2555E+06, 3.2644E+06,
     + 3.2733E+06, 3.2823E+06, 3.2913E+06, 3.3003E+06, 3.3093E+06,
     +  3.3183E+06, 3.3274E+06, 3.3364E+06, 3.3455E+06, 3.3546E+06,
     + 3.3638E+06, 3.3729E+06, 3.3821E+06, 3.3913E+06, 3.4005E+06,
     +  3.4097E+06, 3.4190E+06, 3.4282E+06, 3.4375E+06, 3.4468E+06,
     + 3.4562E+06, 3.4655E+06, 3.4749E+06, 3.4843E+06, 3.4937E+06,
     +  3.5031E+06, 3.5126E+06, 3.5221E+06, 3.5316E+06, 3.5411E+06,
     + 3.5506E+06, 3.5602E+06, 3.5697E+06, 3.5793E+06, 3.5890E+06,
     +  3.5986E+06, 3.6083E+06, 3.6179E+06, 3.6276E+06, 3.6374E+06,
     + 3.6471E+06, 3.6569E+06, 3.6667E+06, 3.6765E+06, 3.6863E+06,
     +  3.6962E+06, 3.7060E+06, 3.7159E+06, 3.7258E+06, 3.7358E+06,
     + 3.7457E+06, 3.7557E+06, 3.7657E+06, 3.7757E+06, 3.7858E+06,
     +  3.7958E+06, 3.8059E+06, 3.8160E+06, 3.8261E+06, 3.8363E+06,
     + 3.8465E+06, 3.8567E+06, 3.8669E+06, 3.8771E+06, 3.8874E+06,
     +  3.8976E+06, 3.9079E+06, 3.9183E+06, 3.9286E+06, 3.9390E+06,
     + 3.9494E+06, 3.9598E+06, 3.9702E+06, 3.9807E+06, 3.9912E+06,
     +  4.0017E+06, 4.0122E+06, 4.0227E+06, 4.0333E+06, 4.0439E+06,
     + 4.0545E+06, 4.0651E+06, 4.0758E+06, 4.0865E+06, 4.0972E+06,
     +  4.1079E+06, 4.1186E+06, 4.1294E+06, 4.1402E+06, 4.1510E+06,
     + 4.1618E+06, 4.1727E+06, 4.1836E+06, 4.1945E+06, 4.2054E+06,
     +  4.2164E+06, 4.2273E+06, 4.2383E+06, 4.2494E+06, 4.2604E+06,
     + 4.2715E+06, 4.2826E+06, 4.2937E+06, 4.3048E+06, 4.3160E+06,
     +  4.3271E+06, 4.3383E+06, 4.3496E+06, 4.3608E+06, 4.3721E+06,
     + 4.3834E+06, 4.3947E+06, 4.4061E+06, 4.4174E+06, 4.4288E+06,
     +  4.4402E+06, 4.4517E+06, 4.4631E+06, 4.4746E+06, 4.4861E+06,
     + 4.4977E+06, 4.5092E+06, 4.5208E+06, 4.5324E+06, 4.5441E+06,
     +  4.5557E+06, 4.5674E+06, 4.5791E+06, 4.5908E+06, 4.6026E+06,
     + 4.6144E+06, 4.6262E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 1,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       224
      data (QofT( 2,J),J=1,2408)/ 6.9085E+00, 1.3468E+01, 2.0031E+01,
     +  2.6595E+01, 3.3160E+01, 3.9725E+01, 4.6291E+01, 5.2856E+01,
     + 5.9422E+01, 6.5988E+01, 7.2554E+01, 7.9120E+01, 8.5687E+01,
     +  9.2253E+01, 9.8820E+01, 1.0539E+02, 1.1195E+02, 1.1852E+02,
     + 1.2509E+02, 1.3166E+02, 1.4479E+02, 1.5793E+02, 1.7106E+02,
     +  1.8420E+02, 1.9734E+02, 2.1048E+02, 2.2362E+02, 2.3676E+02,
     + 2.4990E+02, 2.6304E+02, 2.7618E+02, 2.8932E+02, 3.0247E+02,
     +  3.1561E+02, 3.2876E+02, 3.4191E+02, 3.5506E+02, 3.6821E+02,
     + 3.8137E+02, 3.9454E+02, 4.0771E+02, 4.2089E+02, 4.3408E+02,
     +  4.4728E+02, 4.6049E+02, 4.7372E+02, 4.8697E+02, 5.0024E+02,
     + 5.1353E+02, 5.2685E+02, 5.4019E+02, 5.5357E+02, 5.6698E+02,
     +  5.8042E+02, 5.9391E+02, 6.0744E+02, 6.2102E+02, 6.3465E+02,
     + 6.4834E+02, 6.6208E+02, 6.7589E+02, 6.8976E+02, 7.0370E+02,
     +  7.1771E+02, 7.3180E+02, 7.4596E+02, 7.6021E+02, 7.7455E+02,
     + 7.8898E+02, 8.0350E+02, 8.1811E+02, 8.3283E+02, 8.4766E+02,
     +  8.6259E+02, 8.7763E+02, 8.9278E+02, 9.0806E+02, 9.2345E+02,
     + 9.3897E+02, 9.5462E+02, 9.7040E+02, 9.8631E+02, 1.0024E+03,
     +  1.0185E+03, 1.0349E+03, 1.0514E+03, 1.0680E+03, 1.0848E+03,
     + 1.1017E+03, 1.1188E+03, 1.1361E+03, 1.1535E+03, 1.1711E+03,
     +  1.1888E+03, 1.2067E+03, 1.2249E+03, 1.2431E+03, 1.2616E+03,
     + 1.2802E+03, 1.2991E+03, 1.3181E+03, 1.3373E+03, 1.3567E+03,
     +  1.3763E+03, 1.3961E+03, 1.4161E+03, 1.4364E+03, 1.4568E+03,
     + 1.4774E+03, 1.4982E+03, 1.5193E+03, 1.5406E+03, 1.5621E+03,
     +  1.5838E+03, 1.6057E+03, 1.6279E+03, 1.6503E+03, 1.6729E+03,
     + 1.6958E+03, 1.7189E+03, 1.7423E+03, 1.7659E+03, 1.7897E+03,
     +  1.8138E+03, 1.8381E+03, 1.8627E+03, 1.8875E+03, 1.9126E+03,
     + 1.9380E+03, 1.9636E+03, 1.9895E+03, 2.0157E+03, 2.0421E+03,
     +  2.0688E+03, 2.0958E+03, 2.1230E+03, 2.1506E+03, 2.1784E+03,
     + 2.2065E+03, 2.2349E+03, 2.2635E+03, 2.2925E+03, 2.3218E+03,
     +  2.3513E+03, 2.3812E+03, 2.4114E+03, 2.4419E+03, 2.4726E+03,
     + 2.5037E+03, 2.5351E+03, 2.5669E+03, 2.5989E+03, 2.6313E+03,
     +  2.6640E+03, 2.6970E+03, 2.7303E+03, 2.7640E+03, 2.7980E+03,
     + 2.8324E+03, 2.8670E+03, 2.9021E+03, 2.9375E+03, 2.9732E+03,
     +  3.0092E+03, 3.0457E+03, 3.0825E+03, 3.1196E+03, 3.1571E+03,
     + 3.1950E+03, 3.2332E+03, 3.2718E+03, 3.3108E+03, 3.3502E+03,
     +  3.3899E+03, 3.4300E+03, 3.4705E+03, 3.5114E+03, 3.5527E+03,
     + 3.5944E+03, 3.6365E+03, 3.6789E+03, 3.7218E+03, 3.7651E+03,
     +  3.8088E+03, 3.8529E+03, 3.8974E+03, 3.9423E+03, 3.9876E+03,
     + 4.0334E+03, 4.0796E+03, 4.1262E+03, 4.1733E+03, 4.2208E+03,
     +  4.2687E+03, 4.3171E+03, 4.3659E+03, 4.4152E+03, 4.4649E+03,
     + 4.5151E+03, 4.5657E+03, 4.6168E+03, 4.6684E+03, 4.7204E+03,
     +  4.7729E+03, 4.8259E+03, 4.8793E+03, 4.9333E+03, 4.9877E+03,
     + 5.0426E+03, 5.0980E+03, 5.1539E+03, 5.2103E+03, 5.2671E+03,
     +  5.3245E+03, 5.3824E+03, 5.4408E+03, 5.4998E+03, 5.5592E+03,
     + 5.6192E+03, 5.6797E+03, 5.7407E+03, 5.8022E+03, 5.8643E+03,
     +  5.9270E+03, 5.9901E+03, 6.0539E+03, 6.1181E+03, 6.1829E+03,
     + 6.2483E+03, 6.3143E+03, 6.3808E+03, 6.4479E+03, 6.5155E+03,
     +  6.5837E+03, 6.6525E+03, 6.7219E+03, 6.7919E+03, 6.8625E+03,
     + 6.9336E+03, 7.0054E+03, 7.0778E+03, 7.1508E+03, 7.2243E+03,
     +  7.2986E+03, 7.3734E+03, 7.4488E+03, 7.5249E+03, 7.6016E+03,
     + 7.6789E+03, 7.7569E+03, 7.8355E+03, 7.9148E+03, 7.9947E+03,
     +  8.0753E+03, 8.1565E+03, 8.2384E+03, 8.3210E+03, 8.4042E+03,
     + 8.4881E+03, 8.5727E+03, 8.6580E+03, 8.7440E+03, 8.8307E+03,
     +  8.9180E+03, 9.0061E+03, 9.0948E+03, 9.1843E+03, 9.2745E+03,
     + 9.3654E+03, 9.4571E+03, 9.5494E+03, 9.6425E+03, 9.7364E+03,
     +  9.8309E+03, 9.9263E+03, 1.0022E+04, 1.0119E+04, 1.0217E+04,
     + 1.0315E+04, 1.0414E+04, 1.0514E+04, 1.0615E+04, 1.0716E+04,
     +  1.0818E+04, 1.0922E+04, 1.1025E+04, 1.1130E+04, 1.1235E+04,
     + 1.1342E+04, 1.1449E+04, 1.1557E+04, 1.1665E+04, 1.1775E+04,
     +  1.1885E+04, 1.1996E+04, 1.2108E+04, 1.2221E+04, 1.2335E+04,
     + 1.2450E+04, 1.2565E+04, 1.2682E+04, 1.2799E+04, 1.2917E+04,
     +  1.3036E+04, 1.3156E+04, 1.3277E+04, 1.3398E+04, 1.3521E+04,
     + 1.3644E+04, 1.3769E+04, 1.3894E+04, 1.4020E+04, 1.4148E+04,
     +  1.4276E+04, 1.4405E+04, 1.4535E+04, 1.4666E+04, 1.4798E+04,
     + 1.4931E+04, 1.5064E+04, 1.5199E+04, 1.5335E+04, 1.5472E+04,
     +  1.5610E+04, 1.5748E+04, 1.5888E+04, 1.6029E+04, 1.6171E+04,
     + 1.6314E+04, 1.6458E+04, 1.6602E+04, 1.6748E+04, 1.6895E+04,
     +  1.7043E+04, 1.7192E+04, 1.7342E+04, 1.7494E+04, 1.7646E+04,
     + 1.7799E+04, 1.7954E+04, 1.8109E+04, 1.8266E+04, 1.8423E+04,
     +  1.8582E+04, 1.8742E+04, 1.8903E+04, 1.9065E+04, 1.9228E+04,
     + 1.9393E+04, 1.9558E+04, 1.9725E+04, 1.9893E+04, 2.0062E+04,
     +  2.0232E+04, 2.0403E+04, 2.0576E+04, 2.0749E+04, 2.0924E+04,
     + 2.1100E+04, 2.1277E+04, 2.1456E+04, 2.1636E+04, 2.1816E+04,
     +  2.1999E+04, 2.2182E+04, 2.2366E+04, 2.2552E+04, 2.2739E+04,
     + 2.2928E+04, 2.3117E+04, 2.3308E+04, 2.3500E+04, 2.3694E+04,
     +  2.3889E+04, 2.4085E+04, 2.4282E+04, 2.4481E+04, 2.4681E+04,
     + 2.4882E+04, 2.5085E+04, 2.5288E+04, 2.5494E+04, 2.5700E+04,
     +  2.5908E+04, 2.6118E+04, 2.6329E+04, 2.6541E+04, 2.6754E+04,
     + 2.6969E+04, 2.7185E+04, 2.7403E+04, 2.7622E+04, 2.7843E+04,
     +  2.8065E+04, 2.8288E+04, 2.8513E+04, 2.8739E+04, 2.8967E+04,
     + 2.9196E+04, 2.9427E+04, 2.9659E+04, 2.9893E+04, 3.0128E+04,
     +  3.0364E+04, 3.0603E+04, 3.0842E+04, 3.1083E+04, 3.1326E+04,
     + 3.1570E+04, 3.1816E+04, 3.2063E+04, 3.2312E+04, 3.2563E+04,
     +  3.2815E+04, 3.3068E+04, 3.3324E+04, 3.3580E+04, 3.3839E+04,
     + 3.4099E+04, 3.4360E+04, 3.4624E+04, 3.4888E+04, 3.5155E+04,
     +  3.5423E+04, 3.5693E+04, 3.5964E+04, 3.6238E+04, 3.6512E+04,
     + 3.6789E+04, 3.7067E+04, 3.7347E+04, 3.7629E+04, 3.7912E+04,
     +  3.8197E+04, 3.8484E+04, 3.8772E+04, 3.9063E+04, 3.9355E+04,
     + 3.9649E+04, 3.9944E+04, 4.0242E+04, 4.0541E+04, 4.0842E+04,
     +  4.1145E+04, 4.1449E+04, 4.1756E+04, 4.2064E+04, 4.2374E+04,
     + 4.2686E+04, 4.3000E+04, 4.3316E+04, 4.3633E+04, 4.3953E+04,
     +  4.4274E+04, 4.4597E+04, 4.4922E+04, 4.5249E+04, 4.5578E+04,
     + 4.5909E+04, 4.6242E+04, 4.6577E+04, 4.6914E+04, 4.7253E+04,
     +  4.7593E+04, 4.7936E+04, 4.8281E+04, 4.8628E+04, 4.8976E+04,
     + 4.9327E+04, 4.9680E+04, 5.0035E+04, 5.0392E+04, 5.0751E+04,
     +  5.1112E+04, 5.1475E+04, 5.1840E+04, 5.2207E+04, 5.2577E+04,
     + 5.2948E+04, 5.3322E+04, 5.3698E+04, 5.4075E+04, 5.4456E+04,
     +  5.4838E+04, 5.5222E+04, 5.5609E+04, 5.5997E+04, 5.6388E+04,
     + 5.6782E+04, 5.7177E+04, 5.7575E+04, 5.7974E+04, 5.8376E+04,
     +  5.8781E+04, 5.9187E+04, 5.9596E+04, 6.0007E+04, 6.0421E+04,
     + 6.0837E+04, 6.1255E+04, 6.1675E+04, 6.2098E+04, 6.2523E+04,
     +  6.2950E+04, 6.3380E+04, 6.3812E+04, 6.4247E+04, 6.4684E+04,
     + 6.5123E+04, 6.5565E+04, 6.6009E+04, 6.6456E+04, 6.6905E+04,
     +  6.7356E+04, 6.7810E+04, 6.8267E+04, 6.8726E+04, 6.9187E+04,
     + 6.9651E+04, 7.0117E+04, 7.0586E+04, 7.1058E+04, 7.1532E+04,
     +  7.2008E+04, 7.2488E+04, 7.2969E+04, 7.3454E+04, 7.3941E+04,
     + 7.4430E+04, 7.4922E+04, 7.5417E+04, 7.5914E+04, 7.6414E+04,
     +  7.6917E+04, 7.7423E+04, 7.7931E+04, 7.8441E+04, 7.8955E+04,
     + 7.9471E+04, 7.9990E+04, 8.0512E+04, 8.1036E+04, 8.1563E+04,
     +  8.2093E+04, 8.2626E+04, 8.3161E+04, 8.3700E+04, 8.4241E+04,
     + 8.4785E+04, 8.5332E+04, 8.5881E+04, 8.6434E+04, 8.6989E+04,
     +  8.7548E+04, 8.8109E+04, 8.8673E+04, 8.9240E+04, 8.9810E+04,
     + 9.0382E+04, 9.0958E+04, 9.1537E+04, 9.2119E+04, 9.2704E+04,
     +  9.3291E+04, 9.3882E+04, 9.4476E+04, 9.5072E+04, 9.5672E+04,
     + 9.6275E+04, 9.6881E+04, 9.7490E+04, 9.8102E+04, 9.8718E+04,
     +  9.9336E+04, 9.9957E+04, 1.0058E+05, 1.0121E+05, 1.0184E+05,
     + 1.0247E+05, 1.0311E+05, 1.0375E+05, 1.0440E+05, 1.0504E+05,
     +  1.0569E+05, 1.0635E+05, 1.0700E+05, 1.0766E+05, 1.0833E+05,
     + 1.0899E+05, 1.0966E+05, 1.1034E+05, 1.1101E+05, 1.1169E+05,
     +  1.1238E+05, 1.1306E+05, 1.1375E+05, 1.1445E+05, 1.1514E+05,
     + 1.1584E+05, 1.1655E+05, 1.1725E+05, 1.1796E+05, 1.1868E+05,
     +  1.1940E+05, 1.2012E+05, 1.2084E+05, 1.2157E+05, 1.2230E+05,
     + 1.2304E+05, 1.2377E+05, 1.2452E+05, 1.2526E+05, 1.2601E+05,
     +  1.2677E+05, 1.2752E+05, 1.2828E+05, 1.2905E+05, 1.2981E+05,
     + 1.3059E+05, 1.3136E+05, 1.3214E+05, 1.3292E+05, 1.3371E+05,
     +  1.3450E+05, 1.3529E+05, 1.3609E+05, 1.3689E+05, 1.3770E+05,
     + 1.3851E+05, 1.3932E+05, 1.4013E+05, 1.4096E+05, 1.4178E+05,
     +  1.4261E+05, 1.4344E+05, 1.4428E+05, 1.4512E+05, 1.4596E+05,
     + 1.4681E+05, 1.4766E+05, 1.4852E+05, 1.4938E+05, 1.5024E+05,
     +  1.5111E+05, 1.5198E+05, 1.5286E+05, 1.5374E+05, 1.5462E+05,
     + 1.5551E+05, 1.5640E+05, 1.5730E+05, 1.5820E+05, 1.5910E+05,
     +  1.6001E+05, 1.6093E+05, 1.6184E+05, 1.6277E+05, 1.6369E+05,
     + 1.6462E+05, 1.6556E+05, 1.6650E+05, 1.6744E+05, 1.6839E+05,
     +  1.6934E+05, 1.7029E+05, 1.7125E+05, 1.7222E+05, 1.7319E+05,
     + 1.7416E+05, 1.7514E+05, 1.7612E+05, 1.7711E+05, 1.7810E+05,
     +  1.7909E+05, 1.8009E+05, 1.8110E+05, 1.8211E+05, 1.8312E+05,
     + 1.8414E+05, 1.8516E+05, 1.8619E+05, 1.8722E+05, 1.8826E+05,
     +  1.8930E+05, 1.9035E+05, 1.9140E+05, 1.9245E+05, 1.9351E+05,
     + 1.9458E+05, 1.9565E+05, 1.9672E+05, 1.9780E+05, 1.9888E+05,
     +  1.9997E+05, 2.0107E+05, 2.0216E+05, 2.0327E+05, 2.0437E+05,
     + 2.0549E+05, 2.0660E+05, 2.0773E+05, 2.0885E+05, 2.0999E+05,
     +  2.1112E+05, 2.1227E+05, 2.1341E+05, 2.1456E+05, 2.1572E+05,
     + 2.1688E+05, 2.1805E+05, 2.1922E+05, 2.2040E+05, 2.2158E+05,
     +  2.2277E+05, 2.2396E+05, 2.2516E+05, 2.2636E+05, 2.2757E+05,
     + 2.2879E+05, 2.3000E+05, 2.3123E+05, 2.3246E+05, 2.3369E+05,
     +  2.3493E+05, 2.3618E+05, 2.3743E+05, 2.3868E+05, 2.3994E+05,
     + 2.4121E+05, 2.4248E+05, 2.4376E+05, 2.4504E+05, 2.4633E+05,
     +  2.4762E+05, 2.4892E+05, 2.5023E+05, 2.5154E+05, 2.5285E+05,
     + 2.5417E+05, 2.5550E+05, 2.5683E+05, 2.5817E+05, 2.5951E+05,
     +  2.6086E+05, 2.6222E+05, 2.6358E+05, 2.6494E+05, 2.6631E+05,
     + 2.6769E+05, 2.6907E+05, 2.7046E+05, 2.7186E+05, 2.7326E+05,
     +  2.7466E+05, 2.7608E+05, 2.7749E+05, 2.7892E+05, 2.8035E+05,
     + 2.8178E+05, 2.8322E+05, 2.8467E+05, 2.8612E+05, 2.8758E+05,
     +  2.8905E+05, 2.9052E+05, 2.9200E+05, 2.9348E+05, 2.9497E+05,
     + 2.9647E+05, 2.9797E+05, 2.9948E+05, 3.0099E+05, 3.0251E+05,
     +  3.0404E+05, 3.0557E+05, 3.0711E+05, 3.0865E+05, 3.1021E+05,
     + 3.1176E+05, 3.1333E+05, 3.1490E+05, 3.1647E+05, 3.1806E+05,
     +  3.1965E+05, 3.2124E+05, 3.2284E+05, 3.2445E+05, 3.2607E+05,
     + 3.2769E+05, 3.2932E+05, 3.3095E+05, 3.3259E+05, 3.3424E+05,
     +  3.3589E+05, 3.3755E+05, 3.3922E+05, 3.4090E+05, 3.4258E+05,
     + 3.4426E+05, 3.4596E+05, 3.4766E+05, 3.4937E+05, 3.5108E+05,
     +  3.5280E+05, 3.5453E+05, 3.5626E+05, 3.5800E+05, 3.5975E+05,
     + 3.6151E+05, 3.6327E+05, 3.6504E+05, 3.6681E+05, 3.6860E+05,
     +  3.7039E+05, 3.7218E+05, 3.7399E+05, 3.7580E+05, 3.7762E+05,
     + 3.7944E+05, 3.8127E+05, 3.8311E+05, 3.8496E+05, 3.8681E+05,
     +  3.8867E+05, 3.9054E+05, 3.9242E+05, 3.9430E+05, 3.9619E+05,
     + 3.9809E+05, 3.9999E+05, 4.0190E+05, 4.0382E+05, 4.0575E+05,
     +  4.0768E+05, 4.0962E+05, 4.1157E+05, 4.1353E+05, 4.1549E+05,
     + 4.1746E+05, 4.1944E+05, 4.2143E+05, 4.2342E+05, 4.2542E+05,
     +  4.2743E+05, 4.2945E+05, 4.3147E+05, 4.3350E+05, 4.3554E+05,
     + 4.3759E+05, 4.3965E+05, 4.4171E+05, 4.4378E+05, 4.4586E+05,
     +  4.4794E+05, 4.5004E+05, 4.5214E+05, 4.5425E+05, 4.5637E+05,
     + 4.5849E+05, 4.6063E+05, 4.6277E+05, 4.6492E+05, 4.6708E+05,
     +  4.6924E+05, 4.7142E+05, 4.7360E+05, 4.7579E+05, 4.7799E+05,
     + 4.8020E+05, 4.8241E+05, 4.8463E+05, 4.8687E+05, 4.8910E+05,
     +  4.9135E+05, 4.9361E+05, 4.9587E+05, 4.9815E+05, 5.0043E+05,
     + 5.0272E+05, 5.0502E+05, 5.0732E+05, 5.0964E+05, 5.1196E+05,
     +  5.1429E+05, 5.1663E+05, 5.1898E+05, 5.2134E+05, 5.2371E+05,
     + 5.2608E+05, 5.2847E+05, 5.3086E+05, 5.3326E+05, 5.3567E+05,
     +  5.3809E+05, 5.4051E+05, 5.4295E+05, 5.4540E+05, 5.4785E+05,
     + 5.5031E+05, 5.5278E+05, 5.5527E+05, 5.5775E+05, 5.6025E+05,
     +  5.6276E+05, 5.6528E+05, 5.6780E+05, 5.7034E+05, 5.7288E+05,
     + 5.7544E+05, 5.7800E+05, 5.8057E+05, 5.8315E+05, 5.8574E+05,
     +  5.8834E+05, 5.9095E+05, 5.9356E+05, 5.9619E+05, 5.9883E+05,
     + 6.0147E+05, 6.0413E+05, 6.0679E+05, 6.0947E+05, 6.1215E+05,
     +  6.1484E+05, 6.1755E+05, 6.2026E+05, 6.2298E+05, 6.2571E+05,
     + 6.2845E+05, 6.3120E+05, 6.3396E+05, 6.3673E+05, 6.3951E+05,
     +  6.4230E+05, 6.4510E+05, 6.4791E+05, 6.5073E+05, 6.5356E+05,
     + 6.5640E+05, 6.5925E+05, 6.6211E+05, 6.6497E+05, 6.6785E+05,
     +  6.7074E+05, 6.7364E+05, 6.7655E+05, 6.7947E+05, 6.8240E+05,
     + 6.8534E+05, 6.8829E+05, 6.9125E+05, 6.9422E+05, 6.9720E+05,
     +  7.0019E+05, 7.0319E+05, 7.0620E+05, 7.0922E+05, 7.1225E+05,
     + 7.1529E+05, 7.1835E+05, 7.2141E+05, 7.2448E+05, 7.2757E+05,
     +  7.3066E+05, 7.3377E+05, 7.3688E+05, 7.4001E+05, 7.4315E+05,
     + 7.4629E+05, 7.4945E+05, 7.5262E+05, 7.5580E+05, 7.5899E+05,
     +  7.6220E+05, 7.6541E+05, 7.6863E+05, 7.7187E+05, 7.7511E+05,
     + 7.7837E+05, 7.8164E+05, 7.8492E+05, 7.8821E+05, 7.9151E+05,
     +  7.9482E+05, 7.9814E+05, 8.0148E+05, 8.0482E+05, 8.0818E+05,
     + 8.1155E+05, 8.1493E+05, 8.1832E+05, 8.2172E+05, 8.2513E+05,
     +  8.2856E+05, 8.3199E+05, 8.3544E+05, 8.3890E+05, 8.4237E+05,
     + 8.4585E+05, 8.4935E+05, 8.5285E+05, 8.5637E+05, 8.5990E+05,
     +  8.6344E+05, 8.6699E+05, 8.7055E+05, 8.7413E+05, 8.7772E+05,
     + 8.8131E+05, 8.8493E+05, 8.8855E+05, 8.9218E+05, 8.9583E+05,
     +  8.9949E+05, 9.0316E+05, 9.0684E+05, 9.1054E+05, 9.1425E+05,
     + 9.1797E+05, 9.2170E+05, 9.2544E+05, 9.2920E+05, 9.3297E+05,
     +  9.3675E+05, 9.4054E+05, 9.4434E+05, 9.4816E+05, 9.5199E+05,
     + 9.5583E+05, 9.5969E+05, 9.6356E+05, 9.6744E+05, 9.7133E+05,
     +  9.7523E+05, 9.7915E+05, 9.8308E+05, 9.8702E+05, 9.9098E+05,
     + 9.9495E+05, 9.9893E+05, 1.0029E+06, 1.0069E+06, 1.0110E+06,
     +  1.0150E+06, 1.0190E+06, 1.0231E+06, 1.0272E+06, 1.0312E+06,
     + 1.0353E+06, 1.0395E+06, 1.0436E+06, 1.0477E+06, 1.0519E+06,
     +  1.0560E+06, 1.0602E+06, 1.0644E+06, 1.0686E+06, 1.0728E+06,
     + 1.0770E+06, 1.0813E+06, 1.0855E+06, 1.0898E+06, 1.0941E+06,
     +  1.0984E+06, 1.1027E+06, 1.1070E+06, 1.1114E+06, 1.1157E+06,
     + 1.1201E+06, 1.1245E+06, 1.1289E+06, 1.1333E+06, 1.1377E+06,
     +  1.1421E+06, 1.1466E+06, 1.1510E+06, 1.1555E+06, 1.1600E+06,
     + 1.1645E+06, 1.1690E+06, 1.1736E+06, 1.1781E+06, 1.1827E+06,
     +  1.1872E+06, 1.1918E+06, 1.1964E+06, 1.2011E+06, 1.2057E+06,
     + 1.2103E+06, 1.2150E+06, 1.2197E+06, 1.2244E+06, 1.2291E+06,
     +  1.2338E+06, 1.2385E+06, 1.2433E+06, 1.2480E+06, 1.2528E+06,
     + 1.2576E+06, 1.2624E+06, 1.2672E+06, 1.2721E+06, 1.2769E+06,
     +  1.2818E+06, 1.2867E+06, 1.2916E+06, 1.2965E+06, 1.3014E+06,
     + 1.3064E+06, 1.3113E+06, 1.3163E+06, 1.3213E+06, 1.3263E+06,
     +  1.3313E+06, 1.3363E+06, 1.3414E+06, 1.3464E+06, 1.3515E+06,
     + 1.3566E+06, 1.3617E+06, 1.3668E+06, 1.3720E+06, 1.3771E+06,
     +  1.3823E+06, 1.3875E+06, 1.3927E+06, 1.3979E+06, 1.4031E+06,
     + 1.4084E+06, 1.4137E+06, 1.4189E+06, 1.4242E+06, 1.4296E+06,
     +  1.4349E+06, 1.4402E+06, 1.4456E+06, 1.4510E+06, 1.4564E+06,
     + 1.4618E+06, 1.4672E+06, 1.4726E+06, 1.4781E+06, 1.4836E+06,
     +  1.4891E+06, 1.4946E+06, 1.5001E+06, 1.5056E+06, 1.5112E+06,
     + 1.5168E+06, 1.5224E+06, 1.5280E+06, 1.5336E+06, 1.5392E+06,
     +  1.5449E+06, 1.5505E+06, 1.5562E+06, 1.5619E+06, 1.5677E+06,
     + 1.5734E+06, 1.5792E+06, 1.5849E+06, 1.5907E+06, 1.5965E+06,
     +  1.6024E+06, 1.6082E+06, 1.6141E+06, 1.6199E+06, 1.6258E+06,
     + 1.6317E+06, 1.6377E+06, 1.6436E+06, 1.6496E+06, 1.6556E+06,
     +  1.6616E+06, 1.6676E+06, 1.6736E+06, 1.6797E+06, 1.6857E+06,
     + 1.6918E+06, 1.6979E+06, 1.7040E+06, 1.7102E+06, 1.7163E+06,
     +  1.7225E+06, 1.7287E+06, 1.7349E+06, 1.7411E+06, 1.7474E+06,
     + 1.7536E+06, 1.7599E+06, 1.7662E+06, 1.7725E+06, 1.7789E+06,
     +  1.7852E+06, 1.7916E+06, 1.7980E+06, 1.8044E+06, 1.8108E+06,
     + 1.8173E+06, 1.8237E+06, 1.8302E+06, 1.8367E+06, 1.8432E+06,
     +  1.8498E+06, 1.8563E+06, 1.8629E+06, 1.8695E+06, 1.8761E+06,
     + 1.8827E+06, 1.8894E+06, 1.8961E+06, 1.9028E+06, 1.9095E+06,
     +  1.9162E+06, 1.9229E+06, 1.9297E+06, 1.9365E+06, 1.9433E+06,
     + 1.9501E+06, 1.9569E+06, 1.9638E+06, 1.9707E+06, 1.9776E+06,
     +  1.9845E+06, 1.9914E+06, 1.9984E+06, 2.0054E+06, 2.0124E+06,
     + 2.0194E+06, 2.0264E+06, 2.0335E+06, 2.0406E+06, 2.0477E+06,
     +  2.0548E+06, 2.0619E+06, 2.0691E+06, 2.0762E+06, 2.0834E+06,
     + 2.0907E+06, 2.0979E+06, 2.1051E+06, 2.1124E+06, 2.1197E+06,
     +  2.1270E+06, 2.1344E+06, 2.1417E+06, 2.1491E+06, 2.1565E+06,
     + 2.1639E+06, 2.1714E+06, 2.1788E+06, 2.1863E+06, 2.1938E+06,
     +  2.2013E+06, 2.2089E+06, 2.2164E+06, 2.2240E+06, 2.2316E+06,
     + 2.2392E+06, 2.2469E+06, 2.2546E+06, 2.2622E+06, 2.2700E+06,
     +  2.2777E+06, 2.2854E+06, 2.2932E+06, 2.3010E+06, 2.3088E+06,
     + 2.3167E+06, 2.3245E+06, 2.3324E+06, 2.3403E+06, 2.3482E+06,
     +  2.3562E+06, 2.3641E+06, 2.3721E+06, 2.3801E+06, 2.3882E+06,
     + 2.3962E+06, 2.4043E+06, 2.4124E+06, 2.4205E+06, 2.4286E+06,
     +  2.4368E+06, 2.4450E+06, 2.4532E+06, 2.4614E+06, 2.4697E+06,
     + 2.4779E+06, 2.4862E+06, 2.4946E+06, 2.5029E+06, 2.5113E+06,
     +  2.5196E+06, 2.5280E+06, 2.5365E+06, 2.5449E+06, 2.5534E+06,
     + 2.5619E+06, 2.5704E+06, 2.5790E+06, 2.5875E+06, 2.5961E+06,
     +  2.6047E+06, 2.6134E+06, 2.6220E+06, 2.6307E+06, 2.6394E+06,
     + 2.6481E+06, 2.6569E+06, 2.6657E+06, 2.6745E+06, 2.6833E+06,
     +  2.6921E+06, 2.7010E+06, 2.7099E+06, 2.7188E+06, 2.7277E+06,
     + 2.7367E+06, 2.7457E+06, 2.7547E+06, 2.7637E+06, 2.7728E+06,
     +  2.7819E+06, 2.7910E+06, 2.8001E+06, 2.8093E+06, 2.8184E+06,
     + 2.8276E+06, 2.8369E+06, 2.8461E+06, 2.8554E+06, 2.8647E+06,
     +  2.8740E+06, 2.8833E+06, 2.8927E+06, 2.9021E+06, 2.9115E+06,
     + 2.9210E+06, 2.9304E+06, 2.9399E+06, 2.9494E+06, 2.9590E+06,
     +  2.9686E+06, 2.9782E+06, 2.9878E+06, 2.9974E+06, 3.0071E+06,
     + 3.0168E+06, 3.0265E+06, 3.0362E+06, 3.0460E+06, 3.0558E+06,
     +  3.0656E+06, 3.0755E+06, 3.0853E+06, 3.0952E+06, 3.1051E+06,
     + 3.1151E+06, 3.1251E+06, 3.1351E+06, 3.1451E+06, 3.1551E+06,
     +  3.1652E+06, 3.1753E+06, 3.1854E+06, 3.1956E+06, 3.2058E+06,
     + 3.2160E+06, 3.2262E+06, 3.2364E+06, 3.2467E+06, 3.2570E+06,
     +  3.2674E+06, 3.2777E+06, 3.2881E+06, 3.2985E+06, 3.3090E+06,
     + 3.3194E+06, 3.3299E+06, 3.3405E+06, 3.3510E+06, 3.3616E+06,
     +  3.3722E+06, 3.3828E+06, 3.3935E+06, 3.4042E+06, 3.4149E+06,
     + 3.4256E+06, 3.4364E+06, 3.4472E+06, 3.4580E+06, 3.4688E+06,
     +  3.4797E+06, 3.4906E+06, 3.5015E+06, 3.5125E+06, 3.5235E+06,
     + 3.5345E+06, 3.5455E+06, 3.5566E+06, 3.5677E+06, 3.5788E+06,
     +  3.5899E+06, 3.6011E+06, 3.6123E+06, 3.6235E+06, 3.6348E+06,
     + 3.6461E+06, 3.6574E+06, 3.6688E+06, 3.6801E+06, 3.6915E+06,
     +  3.7030E+06, 3.7144E+06, 3.7259E+06, 3.7374E+06, 3.7490E+06,
     + 3.7605E+06, 3.7721E+06, 3.7838E+06, 3.7954E+06, 3.8071E+06,
     +  3.8188E+06, 3.8306E+06, 3.8424E+06, 3.8542E+06, 3.8660E+06,
     + 3.8779E+06, 3.8898E+06, 3.9017E+06, 3.9136E+06, 3.9256E+06,
     +  3.9376E+06, 3.9497E+06, 3.9617E+06, 3.9738E+06, 3.9860E+06,
     + 3.9981E+06, 4.0103E+06, 4.0225E+06, 4.0348E+06, 4.0470E+06,
     +  4.0593E+06, 4.0717E+06, 4.0841E+06, 4.0965E+06, 4.1089E+06,
     + 4.1213E+06, 4.1338E+06, 4.1463E+06, 4.1589E+06, 4.1715E+06,
     +  4.1841E+06, 4.1967E+06, 4.2094E+06, 4.2221E+06, 4.2348E+06,
     + 4.2476E+06, 4.2604E+06, 4.2732E+06, 4.2861E+06, 4.2990E+06,
     +  4.3119E+06, 4.3248E+06, 4.3378E+06, 4.3508E+06, 4.3639E+06,
     + 4.3770E+06, 4.3901E+06, 4.4032E+06, 4.4164E+06, 4.4296E+06,
     +  4.4428E+06, 4.4561E+06, 4.4694E+06, 4.4827E+06, 4.4961E+06,
     + 4.5095E+06, 4.5229E+06, 4.5364E+06, 4.5499E+06, 4.5634E+06,
     +  4.5769E+06, 4.5905E+06, 4.6041E+06, 4.6178E+06, 4.6315E+06,
     + 4.6452E+06, 4.6589E+06, 4.6727E+06, 4.6865E+06, 4.7004E+06,
     +  4.7143E+06, 4.7282E+06, 4.7421E+06, 4.7561E+06, 4.7701E+06,
     + 4.7842E+06, 4.7983E+06, 4.8124E+06, 4.8265E+06, 4.8407E+06,
     +  4.8549E+06, 4.8692E+06, 4.8834E+06, 4.8978E+06, 4.9121E+06,
     + 4.9265E+06, 4.9409E+06, 4.9554E+06, 4.9698E+06, 4.9844E+06,
     +  4.9989E+06, 5.0135E+06, 5.0281E+06, 5.0428E+06, 5.0575E+06,
     + 5.0722E+06, 5.0869E+06, 5.1017E+06, 5.1166E+06, 5.1314E+06,
     +  5.1463E+06, 5.1613E+06, 5.1762E+06, 5.1912E+06, 5.2063E+06,
     + 5.2213E+06, 5.2364E+06, 5.2516E+06, 5.2668E+06, 5.2820E+06,
     +  5.2972E+06, 5.3125E+06, 5.3278E+06, 5.3432E+06, 5.3586E+06,
     + 5.3740E+06, 5.3894E+06, 5.4049E+06, 5.4205E+06, 5.4360E+06,
     +  5.4516E+06, 5.4673E+06, 5.4830E+06, 5.4987E+06, 5.5144E+06,
     + 5.5302E+06, 5.5460E+06, 5.5619E+06, 5.5778E+06, 5.5937E+06,
     +  5.6097E+06, 5.6257E+06, 5.6417E+06, 5.6578E+06, 5.6739E+06,
     + 5.6901E+06, 5.7062E+06, 5.7225E+06, 5.7387E+06, 5.7550E+06,
     +  5.7714E+06, 5.7877E+06, 5.8041E+06, 5.8206E+06, 5.8371E+06,
     + 5.8536E+06, 5.8702E+06, 5.8868E+06, 5.9034E+06, 5.9201E+06,
     +  5.9368E+06, 5.9535E+06, 5.9703E+06, 5.9871E+06, 6.0040E+06,
     + 6.0209E+06, 6.0378E+06, 6.0548E+06, 6.0718E+06, 6.0889E+06,
     +  6.1060E+06, 6.1231E+06, 6.1403E+06, 6.1575E+06, 6.1748E+06,
     + 6.1920E+06, 6.2094E+06, 6.2267E+06, 6.2441E+06, 6.2616E+06,
     +  6.2791E+06, 6.2966E+06, 6.3141E+06, 6.3318E+06, 6.3494E+06,
     + 6.3671E+06, 6.3848E+06, 6.4025E+06, 6.4203E+06, 6.4382E+06,
     +  6.4561E+06, 6.4740E+06, 6.4919E+06, 6.5099E+06, 6.5280E+06,
     + 6.5461E+06, 6.5642E+06, 6.5823E+06, 6.6005E+06, 6.6188E+06,
     +  6.6371E+06, 6.6554E+06, 6.6737E+06, 6.6921E+06, 6.7106E+06,
     + 6.7291E+06, 6.7476E+06, 6.7662E+06, 6.7848E+06, 6.8034E+06,
     +  6.8221E+06, 6.8408E+06, 6.8596E+06, 6.8784E+06, 6.8973E+06,
     + 6.9162E+06, 6.9351E+06, 6.9541E+06, 6.9731E+06, 6.9922E+06,
     +  7.0113E+06, 7.0304E+06, 7.0496E+06, 7.0688E+06, 7.0881E+06,
     + 7.1074E+06, 7.1268E+06, 7.1462E+06, 7.1656E+06, 7.1851E+06,
     +  7.2046E+06, 7.2242E+06, 7.2438E+06, 7.2635E+06, 7.2832E+06,
     + 7.3029E+06, 7.3227E+06, 7.3425E+06, 7.3624E+06, 7.3823E+06,
     +  7.4023E+06, 7.4223E+06, 7.4423E+06, 7.4624E+06, 7.4825E+06,
     + 7.5027E+06, 7.5229E+06, 7.5432E+06, 7.5635E+06, 7.5839E+06,
     +  7.6043E+06, 7.6247E+06, 7.6452E+06, 7.6657E+06, 7.6863E+06,
     + 7.7069E+06, 7.7276E+06, 7.7483E+06, 7.7690E+06, 7.7898E+06,
     +  7.8106E+06, 7.8315E+06, 7.8525E+06, 7.8734E+06, 7.8945E+06,
     + 7.9155E+06, 7.9366E+06, 7.9578E+06, 7.9790E+06, 8.0002E+06,
     +  8.0215E+06, 8.0429E+06, 8.0642E+06, 8.0857E+06, 8.1071E+06,
     + 8.1287E+06, 8.1502E+06, 8.1718E+06, 8.1935E+06, 8.2152E+06,
     +  8.2369E+06, 8.2587E+06, 8.2806E+06, 8.3025E+06, 8.3244E+06,
     + 8.3464E+06, 8.3684E+06, 8.3905E+06, 8.4126E+06, 8.4348E+06,
     +  8.4570E+06, 8.4793E+06, 8.5016E+06, 8.5239E+06, 8.5463E+06,
     + 8.5688E+06, 8.5913E+06, 8.6138E+06, 8.6364E+06, 8.6591E+06,
     +  8.6818E+06, 8.7045E+06, 8.7273E+06, 8.7501E+06, 8.7730E+06,
     + 8.7959E+06, 8.8189E+06, 8.8419E+06, 8.8650E+06, 8.8881E+06,
     +  8.9113E+06, 8.9345E+06, 8.9578E+06, 8.9811E+06, 9.0045E+06,
     + 9.0279E+06, 9.0513E+06, 9.0749E+06, 9.0984E+06, 9.1220E+06,
     +  9.1457E+06, 9.1694E+06, 9.1932E+06, 9.2170E+06, 9.2408E+06,
     + 9.2647E+06, 9.2887E+06, 9.3127E+06, 9.3368E+06, 9.3609E+06,
     +  9.3850E+06, 9.4092E+06, 9.4335E+06, 9.4578E+06, 9.4822E+06,
     + 9.5066E+06, 9.5310E+06, 9.5555E+06, 9.5801E+06, 9.6047E+06,
     +  9.6294E+06, 9.6541E+06, 9.6788E+06, 9.7037E+06, 9.7285E+06,
     + 9.7534E+06, 9.7784E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 2,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       223
      data (QofT( 3,J),J=1,2408)/ 2.7248E+01, 5.3098E+01, 7.8964E+01,
     +  1.0483E+02, 1.3071E+02, 1.5658E+02, 1.8245E+02, 2.0833E+02,
     + 2.3420E+02, 2.6008E+02, 2.8596E+02, 3.1184E+02, 3.3772E+02,
     +  3.6359E+02, 3.8947E+02, 4.1535E+02, 4.4124E+02, 4.6712E+02,
     + 4.9300E+02, 5.1888E+02, 5.7065E+02, 6.2242E+02, 6.7419E+02,
     +  7.2597E+02, 7.7774E+02, 8.2952E+02, 8.8130E+02, 9.3309E+02,
     + 9.8487E+02, 1.0367E+03, 1.0885E+03, 1.1403E+03, 1.1921E+03,
     +  1.2439E+03, 1.2957E+03, 1.3475E+03, 1.3993E+03, 1.4512E+03,
     + 1.5030E+03, 1.5549E+03, 1.6068E+03, 1.6588E+03, 1.7107E+03,
     +  1.7628E+03, 1.8148E+03, 1.8670E+03, 1.9192E+03, 1.9715E+03,
     + 2.0238E+03, 2.0763E+03, 2.1289E+03, 2.1816E+03, 2.2344E+03,
     +  2.2874E+03, 2.3406E+03, 2.3939E+03, 2.4474E+03, 2.5011E+03,
     + 2.5550E+03, 2.6092E+03, 2.6636E+03, 2.7182E+03, 2.7731E+03,
     +  2.8283E+03, 2.8838E+03, 2.9396E+03, 2.9958E+03, 3.0522E+03,
     + 3.1091E+03, 3.1662E+03, 3.2238E+03, 3.2818E+03, 3.3402E+03,
     +  3.3990E+03, 3.4582E+03, 3.5179E+03, 3.5780E+03, 3.6386E+03,
     + 3.6997E+03, 3.7613E+03, 3.8235E+03, 3.8861E+03, 3.9493E+03,
     +  4.0130E+03, 4.0773E+03, 4.1422E+03, 4.2076E+03, 4.2737E+03,
     + 4.3403E+03, 4.4076E+03, 4.4755E+03, 4.5440E+03, 4.6132E+03,
     +  4.6831E+03, 4.7537E+03, 4.8249E+03, 4.8968E+03, 4.9695E+03,
     + 5.0428E+03, 5.1169E+03, 5.1917E+03, 5.2673E+03, 5.3437E+03,
     +  5.4208E+03, 5.4987E+03, 5.5774E+03, 5.6568E+03, 5.7371E+03,
     + 5.8183E+03, 5.9002E+03, 5.9830E+03, 6.0667E+03, 6.1512E+03,
     +  6.2366E+03, 6.3229E+03, 6.4100E+03, 6.4981E+03, 6.5871E+03,
     + 6.6770E+03, 6.7678E+03, 6.8596E+03, 6.9523E+03, 7.0460E+03,
     +  7.1406E+03, 7.2363E+03, 7.3329E+03, 7.4305E+03, 7.5292E+03,
     + 7.6288E+03, 7.7295E+03, 7.8312E+03, 7.9340E+03, 8.0378E+03,
     +  8.1427E+03, 8.2487E+03, 8.3558E+03, 8.4639E+03, 8.5732E+03,
     + 8.6836E+03, 8.7951E+03, 8.9078E+03, 9.0216E+03, 9.1365E+03,
     +  9.2526E+03, 9.3699E+03, 9.4884E+03, 9.6081E+03, 9.7290E+03,
     + 9.8511E+03, 9.9744E+03, 1.0099E+04, 1.0225E+04, 1.0352E+04,
     +  1.0480E+04, 1.0610E+04, 1.0741E+04, 1.0873E+04, 1.1007E+04,
     + 1.1141E+04, 1.1278E+04, 1.1415E+04, 1.1554E+04, 1.1694E+04,
     +  1.1836E+04, 1.1979E+04, 1.2123E+04, 1.2269E+04, 1.2416E+04,
     + 1.2565E+04, 1.2715E+04, 1.2866E+04, 1.3019E+04, 1.3174E+04,
     +  1.3330E+04, 1.3487E+04, 1.3646E+04, 1.3807E+04, 1.3969E+04,
     + 1.4132E+04, 1.4297E+04, 1.4464E+04, 1.4632E+04, 1.4802E+04,
     +  1.4973E+04, 1.5146E+04, 1.5321E+04, 1.5497E+04, 1.5675E+04,
     + 1.5855E+04, 1.6036E+04, 1.6219E+04, 1.6403E+04, 1.6590E+04,
     +  1.6778E+04, 1.6967E+04, 1.7159E+04, 1.7352E+04, 1.7547E+04,
     + 1.7744E+04, 1.7943E+04, 1.8143E+04, 1.8345E+04, 1.8549E+04,
     +  1.8755E+04, 1.8963E+04, 1.9172E+04, 1.9384E+04, 1.9597E+04,
     + 1.9812E+04, 2.0030E+04, 2.0249E+04, 2.0470E+04, 2.0693E+04,
     +  2.0918E+04, 2.1145E+04, 2.1374E+04, 2.1605E+04, 2.1838E+04,
     + 2.2073E+04, 2.2310E+04, 2.2549E+04, 2.2791E+04, 2.3034E+04,
     +  2.3279E+04, 2.3527E+04, 2.3777E+04, 2.4029E+04, 2.4283E+04,
     + 2.4539E+04, 2.4798E+04, 2.5058E+04, 2.5321E+04, 2.5586E+04,
     +  2.5854E+04, 2.6123E+04, 2.6395E+04, 2.6669E+04, 2.6946E+04,
     + 2.7225E+04, 2.7506E+04, 2.7790E+04, 2.8075E+04, 2.8364E+04,
     +  2.8655E+04, 2.8948E+04, 2.9243E+04, 2.9541E+04, 2.9842E+04,
     + 3.0145E+04, 3.0450E+04, 3.0758E+04, 3.1069E+04, 3.1382E+04,
     +  3.1698E+04, 3.2016E+04, 3.2337E+04, 3.2660E+04, 3.2986E+04,
     + 3.3315E+04, 3.3646E+04, 3.3980E+04, 3.4317E+04, 3.4656E+04,
     +  3.4998E+04, 3.5343E+04, 3.5691E+04, 3.6041E+04, 3.6395E+04,
     + 3.6751E+04, 3.7110E+04, 3.7471E+04, 3.7836E+04, 3.8203E+04,
     +  3.8574E+04, 3.8947E+04, 3.9323E+04, 3.9702E+04, 4.0084E+04,
     + 4.0470E+04, 4.0858E+04, 4.1249E+04, 4.1643E+04, 4.2040E+04,
     +  4.2440E+04, 4.2844E+04, 4.3250E+04, 4.3660E+04, 4.4073E+04,
     + 4.4489E+04, 4.4908E+04, 4.5330E+04, 4.5756E+04, 4.6185E+04,
     +  4.6617E+04, 4.7052E+04, 4.7491E+04, 4.7933E+04, 4.8378E+04,
     + 4.8827E+04, 4.9279E+04, 4.9734E+04, 5.0193E+04, 5.0655E+04,
     +  5.1121E+04, 5.1590E+04, 5.2063E+04, 5.2539E+04, 5.3019E+04,
     + 5.3502E+04, 5.3989E+04, 5.4480E+04, 5.4974E+04, 5.5471E+04,
     +  5.5973E+04, 5.6478E+04, 5.6987E+04, 5.7499E+04, 5.8015E+04,
     + 5.8535E+04, 5.9059E+04, 5.9587E+04, 6.0118E+04, 6.0654E+04,
     +  6.1193E+04, 6.1736E+04, 6.2283E+04, 6.2833E+04, 6.3388E+04,
     + 6.3947E+04, 6.4510E+04, 6.5077E+04, 6.5647E+04, 6.6222E+04,
     +  6.6801E+04, 6.7384E+04, 6.7972E+04, 6.8563E+04, 6.9158E+04,
     + 6.9758E+04, 7.0362E+04, 7.0970E+04, 7.1583E+04, 7.2199E+04,
     +  7.2820E+04, 7.3446E+04, 7.4076E+04, 7.4710E+04, 7.5348E+04,
     + 7.5991E+04, 7.6639E+04, 7.7290E+04, 7.7947E+04, 7.8608E+04,
     +  7.9273E+04, 7.9943E+04, 8.0618E+04, 8.1297E+04, 8.1981E+04,
     + 8.2669E+04, 8.3362E+04, 8.4060E+04, 8.4763E+04, 8.5470E+04,
     +  8.6182E+04, 8.6899E+04, 8.7621E+04, 8.8348E+04, 8.9080E+04,
     + 8.9816E+04, 9.0558E+04, 9.1304E+04, 9.2055E+04, 9.2812E+04,
     +  9.3573E+04, 9.4340E+04, 9.5111E+04, 9.5888E+04, 9.6670E+04,
     + 9.7457E+04, 9.8249E+04, 9.9046E+04, 9.9849E+04, 1.0066E+05,
     +  1.0147E+05, 1.0229E+05, 1.0311E+05, 1.0394E+05, 1.0478E+05,
     + 1.0562E+05, 1.0646E+05, 1.0731E+05, 1.0817E+05, 1.0903E+05,
     +  1.0990E+05, 1.1077E+05, 1.1165E+05, 1.1254E+05, 1.1343E+05,
     + 1.1432E+05, 1.1522E+05, 1.1613E+05, 1.1704E+05, 1.1796E+05,
     +  1.1889E+05, 1.1982E+05, 1.2076E+05, 1.2170E+05, 1.2265E+05,
     + 1.2360E+05, 1.2456E+05, 1.2553E+05, 1.2650E+05, 1.2748E+05,
     +  1.2847E+05, 1.2946E+05, 1.3045E+05, 1.3146E+05, 1.3247E+05,
     + 1.3348E+05, 1.3451E+05, 1.3553E+05, 1.3657E+05, 1.3761E+05,
     +  1.3866E+05, 1.3971E+05, 1.4077E+05, 1.4184E+05, 1.4291E+05,
     + 1.4400E+05, 1.4508E+05, 1.4618E+05, 1.4728E+05, 1.4838E+05,
     +  1.4950E+05, 1.5062E+05, 1.5175E+05, 1.5288E+05, 1.5402E+05,
     + 1.5517E+05, 1.5632E+05, 1.5749E+05, 1.5865E+05, 1.5983E+05,
     +  1.6101E+05, 1.6220E+05, 1.6340E+05, 1.6461E+05, 1.6582E+05,
     + 1.6704E+05, 1.6826E+05, 1.6949E+05, 1.7074E+05, 1.7198E+05,
     +  1.7324E+05, 1.7450E+05, 1.7577E+05, 1.7705E+05, 1.7833E+05,
     + 1.7963E+05, 1.8093E+05, 1.8224E+05, 1.8355E+05, 1.8487E+05,
     +  1.8621E+05, 1.8754E+05, 1.8889E+05, 1.9024E+05, 1.9161E+05,
     + 1.9298E+05, 1.9436E+05, 1.9574E+05, 1.9714E+05, 1.9854E+05,
     +  1.9995E+05, 2.0137E+05, 2.0279E+05, 2.0423E+05, 2.0567E+05,
     + 2.0712E+05, 2.0858E+05, 2.1005E+05, 2.1152E+05, 2.1301E+05,
     +  2.1450E+05, 2.1600E+05, 2.1751E+05, 2.1903E+05, 2.2056E+05,
     + 2.2209E+05, 2.2363E+05, 2.2519E+05, 2.2675E+05, 2.2832E+05,
     +  2.2990E+05, 2.3149E+05, 2.3308E+05, 2.3469E+05, 2.3630E+05,
     + 2.3793E+05, 2.3956E+05, 2.4120E+05, 2.4285E+05, 2.4451E+05,
     +  2.4618E+05, 2.4786E+05, 2.4954E+05, 2.5124E+05, 2.5295E+05,
     + 2.5466E+05, 2.5639E+05, 2.5812E+05, 2.5987E+05, 2.6162E+05,
     +  2.6338E+05, 2.6515E+05, 2.6694E+05, 2.6873E+05, 2.7053E+05,
     + 2.7234E+05, 2.7416E+05, 2.7599E+05, 2.7783E+05, 2.7968E+05,
     +  2.8154E+05, 2.8341E+05, 2.8529E+05, 2.8719E+05, 2.8909E+05,
     + 2.9100E+05, 2.9292E+05, 2.9485E+05, 2.9679E+05, 2.9874E+05,
     +  3.0071E+05, 3.0268E+05, 3.0466E+05, 3.0666E+05, 3.0866E+05,
     + 3.1068E+05, 3.1270E+05, 3.1474E+05, 3.1679E+05, 3.1884E+05,
     +  3.2091E+05, 3.2299E+05, 3.2508E+05, 3.2718E+05, 3.2929E+05,
     + 3.3142E+05, 3.3355E+05, 3.3570E+05, 3.3785E+05, 3.4002E+05,
     +  3.4220E+05, 3.4439E+05, 3.4659E+05, 3.4881E+05, 3.5103E+05,
     + 3.5327E+05, 3.5551E+05, 3.5777E+05, 3.6004E+05, 3.6232E+05,
     +  3.6462E+05, 3.6692E+05, 3.6924E+05, 3.7157E+05, 3.7391E+05,
     + 3.7626E+05, 3.7863E+05, 3.8100E+05, 3.8339E+05, 3.8579E+05,
     +  3.8821E+05, 3.9063E+05, 3.9307E+05, 3.9552E+05, 3.9798E+05,
     + 4.0046E+05, 4.0294E+05, 4.0544E+05, 4.0795E+05, 4.1048E+05,
     +  4.1302E+05, 4.1557E+05, 4.1813E+05, 4.2070E+05, 4.2329E+05,
     + 4.2589E+05, 4.2851E+05, 4.3113E+05, 4.3377E+05, 4.3642E+05,
     +  4.3909E+05, 4.4177E+05, 4.4446E+05, 4.4717E+05, 4.4989E+05,
     + 4.5262E+05, 4.5536E+05, 4.5812E+05, 4.6089E+05, 4.6368E+05,
     +  4.6648E+05, 4.6929E+05, 4.7212E+05, 4.7496E+05, 4.7781E+05,
     + 4.8068E+05, 4.8357E+05, 4.8646E+05, 4.8937E+05, 4.9230E+05,
     +  4.9523E+05, 4.9819E+05, 5.0115E+05, 5.0413E+05, 5.0713E+05,
     + 5.1014E+05, 5.1316E+05, 5.1620E+05, 5.1925E+05, 5.2232E+05,
     +  5.2540E+05, 5.2850E+05, 5.3161E+05, 5.3474E+05, 5.3788E+05,
     + 5.4103E+05, 5.4420E+05, 5.4739E+05, 5.5059E+05, 5.5381E+05,
     +  5.5704E+05, 5.6029E+05, 5.6355E+05, 5.6682E+05, 5.7012E+05,
     + 5.7342E+05, 5.7675E+05, 5.8009E+05, 5.8344E+05, 5.8681E+05,
     +  5.9020E+05, 5.9360E+05, 5.9702E+05, 6.0045E+05, 6.0390E+05,
     + 6.0736E+05, 6.1084E+05, 6.1434E+05, 6.1785E+05, 6.2138E+05,
     +  6.2493E+05, 6.2849E+05, 6.3207E+05, 6.3567E+05, 6.3928E+05,
     + 6.4291E+05, 6.4655E+05, 6.5021E+05, 6.5389E+05, 6.5758E+05,
     +  6.6130E+05, 6.6502E+05, 6.6877E+05, 6.7253E+05, 6.7631E+05,
     + 6.8011E+05, 6.8392E+05, 6.8775E+05, 6.9160E+05, 6.9547E+05,
     +  6.9935E+05, 7.0325E+05, 7.0717E+05, 7.1111E+05, 7.1506E+05,
     + 7.1903E+05, 7.2302E+05, 7.2703E+05, 7.3105E+05, 7.3509E+05,
     +  7.3915E+05, 7.4323E+05, 7.4733E+05, 7.5145E+05, 7.5558E+05,
     + 7.5973E+05, 7.6390E+05, 7.6809E+05, 7.7230E+05, 7.7652E+05,
     +  7.8077E+05, 7.8503E+05, 7.8931E+05, 7.9361E+05, 7.9793E+05,
     + 8.0227E+05, 8.0663E+05, 8.1101E+05, 8.1540E+05, 8.1982E+05,
     +  8.2425E+05, 8.2870E+05, 8.3318E+05, 8.3767E+05, 8.4218E+05,
     + 8.4671E+05, 8.5126E+05, 8.5584E+05, 8.6043E+05, 8.6504E+05,
     +  8.6967E+05, 8.7432E+05, 8.7899E+05, 8.8368E+05, 8.8839E+05,
     + 8.9312E+05, 8.9787E+05, 9.0264E+05, 9.0744E+05, 9.1225E+05,
     +  9.1708E+05, 9.2194E+05, 9.2681E+05, 9.3171E+05, 9.3662E+05,
     + 9.4156E+05, 9.4652E+05, 9.5150E+05, 9.5650E+05, 9.6152E+05,
     +  9.6656E+05, 9.7162E+05, 9.7671E+05, 9.8182E+05, 9.8694E+05,
     + 9.9209E+05, 9.9727E+05, 1.0025E+06, 1.0077E+06, 1.0129E+06,
     +  1.0182E+06, 1.0235E+06, 1.0288E+06, 1.0341E+06, 1.0394E+06,
     + 1.0448E+06, 1.0502E+06, 1.0556E+06, 1.0610E+06, 1.0665E+06,
     +  1.0720E+06, 1.0775E+06, 1.0830E+06, 1.0886E+06, 1.0941E+06,
     + 1.0997E+06, 1.1054E+06, 1.1110E+06, 1.1167E+06, 1.1223E+06,
     +  1.1281E+06, 1.1338E+06, 1.1396E+06, 1.1453E+06, 1.1511E+06,
     + 1.1570E+06, 1.1628E+06, 1.1687E+06, 1.1746E+06, 1.1805E+06,
     +  1.1865E+06, 1.1925E+06, 1.1985E+06, 1.2045E+06, 1.2105E+06,
     + 1.2166E+06, 1.2227E+06, 1.2288E+06, 1.2350E+06, 1.2411E+06,
     +  1.2473E+06, 1.2535E+06, 1.2598E+06, 1.2661E+06, 1.2724E+06,
     + 1.2787E+06, 1.2850E+06, 1.2914E+06, 1.2978E+06, 1.3042E+06,
     +  1.3107E+06, 1.3171E+06, 1.3236E+06, 1.3301E+06, 1.3367E+06,
     + 1.3433E+06, 1.3499E+06, 1.3565E+06, 1.3632E+06, 1.3698E+06,
     +  1.3765E+06, 1.3833E+06, 1.3900E+06, 1.3968E+06, 1.4036E+06,
     + 1.4105E+06, 1.4174E+06, 1.4242E+06, 1.4312E+06, 1.4381E+06,
     +  1.4451E+06, 1.4521E+06, 1.4591E+06, 1.4662E+06, 1.4733E+06,
     + 1.4804E+06, 1.4875E+06, 1.4947E+06, 1.5019E+06, 1.5091E+06,
     +  1.5164E+06, 1.5236E+06, 1.5309E+06, 1.5383E+06, 1.5456E+06,
     + 1.5530E+06, 1.5605E+06, 1.5679E+06, 1.5754E+06, 1.5829E+06,
     +  1.5904E+06, 1.5980E+06, 1.6056E+06, 1.6132E+06, 1.6209E+06,
     + 1.6285E+06, 1.6363E+06, 1.6440E+06, 1.6518E+06, 1.6596E+06,
     +  1.6674E+06, 1.6752E+06, 1.6831E+06, 1.6911E+06, 1.6990E+06,
     + 1.7070E+06, 1.7150E+06, 1.7230E+06, 1.7311E+06, 1.7392E+06,
     +  1.7473E+06, 1.7555E+06, 1.7637E+06, 1.7719E+06, 1.7802E+06,
     + 1.7884E+06, 1.7968E+06, 1.8051E+06, 1.8135E+06, 1.8219E+06,
     +  1.8303E+06, 1.8388E+06, 1.8473E+06, 1.8558E+06, 1.8644E+06,
     + 1.8730E+06, 1.8816E+06, 1.8903E+06, 1.8990E+06, 1.9077E+06,
     +  1.9165E+06, 1.9253E+06, 1.9341E+06, 1.9429E+06, 1.9518E+06,
     + 1.9607E+06, 1.9697E+06, 1.9787E+06, 1.9877E+06, 1.9968E+06,
     +  2.0058E+06, 2.0150E+06, 2.0241E+06, 2.0333E+06, 2.0425E+06,
     + 2.0518E+06, 2.0611E+06, 2.0704E+06, 2.0797E+06, 2.0891E+06,
     +  2.0986E+06, 2.1080E+06, 2.1175E+06, 2.1270E+06, 2.1366E+06,
     + 2.1462E+06, 2.1558E+06, 2.1655E+06, 2.1752E+06, 2.1849E+06,
     +  2.1947E+06, 2.2045E+06, 2.2143E+06, 2.2242E+06, 2.2341E+06,
     + 2.2441E+06, 2.2540E+06, 2.2641E+06, 2.2741E+06, 2.2842E+06,
     +  2.2943E+06, 2.3045E+06, 2.3147E+06, 2.3249E+06, 2.3352E+06,
     + 2.3455E+06, 2.3558E+06, 2.3662E+06, 2.3766E+06, 2.3871E+06,
     +  2.3976E+06, 2.4081E+06, 2.4187E+06, 2.4293E+06, 2.4399E+06,
     + 2.4506E+06, 2.4613E+06, 2.4721E+06, 2.4829E+06, 2.4937E+06,
     +  2.5046E+06, 2.5155E+06, 2.5264E+06, 2.5374E+06, 2.5484E+06,
     + 2.5595E+06, 2.5706E+06, 2.5817E+06, 2.5929E+06, 2.6041E+06,
     +  2.6153E+06, 2.6266E+06, 2.6380E+06, 2.6493E+06, 2.6607E+06,
     + 2.6722E+06, 2.6837E+06, 2.6952E+06, 2.7068E+06, 2.7184E+06,
     +  2.7300E+06, 2.7417E+06, 2.7534E+06, 2.7652E+06, 2.7770E+06,
     + 2.7889E+06, 2.8008E+06, 2.8127E+06, 2.8247E+06, 2.8367E+06,
     +  2.8487E+06, 2.8608E+06, 2.8730E+06, 2.8851E+06, 2.8974E+06,
     + 2.9096E+06, 2.9219E+06, 2.9343E+06, 2.9467E+06, 2.9591E+06,
     +  2.9716E+06, 2.9841E+06, 2.9966E+06, 3.0092E+06, 3.0219E+06,
     + 3.0346E+06, 3.0473E+06, 3.0601E+06, 3.0729E+06, 3.0857E+06,
     +  3.0986E+06, 3.1116E+06, 3.1246E+06, 3.1376E+06, 3.1507E+06,
     + 3.1638E+06, 3.1769E+06, 3.1901E+06, 3.2034E+06, 3.2167E+06,
     +  3.2300E+06, 3.2434E+06, 3.2568E+06, 3.2703E+06, 3.2838E+06,
     + 3.2974E+06, 3.3110E+06, 3.3246E+06, 3.3383E+06, 3.3521E+06,
     +  3.3659E+06, 3.3797E+06, 3.3936E+06, 3.4075E+06, 3.4215E+06,
     + 3.4355E+06, 3.4496E+06, 3.4637E+06, 3.4778E+06, 3.4920E+06,
     +  3.5063E+06, 3.5206E+06, 3.5349E+06, 3.5493E+06, 3.5638E+06,
     + 3.5782E+06, 3.5928E+06, 3.6074E+06, 3.6220E+06, 3.6367E+06,
     +  3.6514E+06, 3.6661E+06, 3.6810E+06, 3.6958E+06, 3.7107E+06,
     + 3.7257E+06, 3.7407E+06, 3.7558E+06, 3.7709E+06, 3.7861E+06,
     +  3.8013E+06, 3.8165E+06, 3.8318E+06, 3.8472E+06, 3.8626E+06,
     + 3.8780E+06, 3.8935E+06, 3.9091E+06, 3.9247E+06, 3.9404E+06,
     +  3.9561E+06, 3.9718E+06, 3.9876E+06, 4.0035E+06, 4.0194E+06,
     + 4.0353E+06, 4.0513E+06, 4.0674E+06, 4.0835E+06, 4.0997E+06,
     +  4.1159E+06, 4.1322E+06, 4.1485E+06, 4.1648E+06, 4.1813E+06,
     + 4.1977E+06, 4.2143E+06, 4.2308E+06, 4.2475E+06, 4.2642E+06,
     +  4.2809E+06, 4.2977E+06, 4.3145E+06, 4.3314E+06, 4.3484E+06,
     + 4.3654E+06, 4.3824E+06, 4.3995E+06, 4.4167E+06, 4.4339E+06,
     +  4.4512E+06, 4.4685E+06, 4.4859E+06, 4.5033E+06, 4.5208E+06,
     + 4.5383E+06, 4.5559E+06, 4.5736E+06, 4.5913E+06, 4.6090E+06,
     +  4.6268E+06, 4.6447E+06, 4.6626E+06, 4.6806E+06, 4.6987E+06,
     + 4.7168E+06, 4.7349E+06, 4.7531E+06, 4.7714E+06, 4.7897E+06,
     +  4.8081E+06, 4.8265E+06, 4.8450E+06, 4.8635E+06, 4.8821E+06,
     + 4.9008E+06, 4.9195E+06, 4.9383E+06, 4.9571E+06, 4.9760E+06,
     +  4.9950E+06, 5.0140E+06, 5.0330E+06, 5.0522E+06, 5.0713E+06,
     + 5.0906E+06, 5.1099E+06, 5.1292E+06, 5.1487E+06, 5.1681E+06,
     +  5.1877E+06, 5.2073E+06, 5.2269E+06, 5.2466E+06, 5.2664E+06,
     + 5.2862E+06, 5.3061E+06, 5.3261E+06, 5.3461E+06, 5.3662E+06,
     +  5.3863E+06, 5.4065E+06, 5.4268E+06, 5.4471E+06, 5.4675E+06,
     + 5.4879E+06, 5.5084E+06, 5.5290E+06, 5.5496E+06, 5.5703E+06,
     +  5.5910E+06, 5.6118E+06, 5.6327E+06, 5.6536E+06, 5.6746E+06,
     + 5.6957E+06, 5.7168E+06, 5.7380E+06, 5.7593E+06, 5.7806E+06,
     +  5.8019E+06, 5.8234E+06, 5.8449E+06, 5.8665E+06, 5.8881E+06,
     + 5.9098E+06, 5.9315E+06, 5.9534E+06, 5.9753E+06, 5.9972E+06,
     +  6.0192E+06, 6.0413E+06, 6.0635E+06, 6.0857E+06, 6.1080E+06,
     + 6.1303E+06, 6.1527E+06, 6.1752E+06, 6.1977E+06, 6.2203E+06,
     +  6.2430E+06, 6.2658E+06, 6.2886E+06, 6.3114E+06, 6.3344E+06,
     + 6.3574E+06, 6.3805E+06, 6.4036E+06, 6.4268E+06, 6.4501E+06,
     +  6.4734E+06, 6.4969E+06, 6.5203E+06, 6.5439E+06, 6.5675E+06,
     + 6.5912E+06, 6.6150E+06, 6.6388E+06, 6.6627E+06, 6.6866E+06,
     +  6.7107E+06, 6.7348E+06, 6.7590E+06, 6.7832E+06, 6.8075E+06,
     + 6.8319E+06, 6.8563E+06, 6.8809E+06, 6.9055E+06, 6.9301E+06,
     +  6.9549E+06, 6.9797E+06, 7.0045E+06, 7.0295E+06, 7.0545E+06,
     + 7.0796E+06, 7.1048E+06, 7.1300E+06, 7.1553E+06, 7.1807E+06,
     +  7.2061E+06, 7.2317E+06, 7.2573E+06, 7.2829E+06, 7.3087E+06,
     + 7.3345E+06, 7.3604E+06, 7.3863E+06, 7.4124E+06, 7.4385E+06,
     +  7.4647E+06, 7.4909E+06, 7.5173E+06, 7.5437E+06, 7.5701E+06,
     + 7.5967E+06, 7.6233E+06, 7.6500E+06, 7.6768E+06, 7.7037E+06,
     +  7.7306E+06, 7.7576E+06, 7.7847E+06, 7.8119E+06, 7.8391E+06,
     + 7.8664E+06, 7.8938E+06, 7.9213E+06, 7.9488E+06, 7.9764E+06,
     +  8.0041E+06, 8.0319E+06, 8.0597E+06, 8.0877E+06, 8.1157E+06,
     + 8.1438E+06, 8.1719E+06, 8.2002E+06, 8.2285E+06, 8.2569E+06,
     +  8.2854E+06, 8.3139E+06, 8.3426E+06, 8.3713E+06, 8.4001E+06,
     + 8.4289E+06, 8.4579E+06, 8.4869E+06, 8.5160E+06, 8.5452E+06,
     +  8.5745E+06, 8.6039E+06, 8.6333E+06, 8.6628E+06, 8.6924E+06,
     + 8.7221E+06, 8.7518E+06, 8.7817E+06, 8.8116E+06, 8.8416E+06,
     +  8.8717E+06, 8.9019E+06, 8.9321E+06, 8.9625E+06, 8.9929E+06,
     + 9.0234E+06, 9.0540E+06, 9.0846E+06, 9.1154E+06, 9.1462E+06,
     +  9.1771E+06, 9.2082E+06, 9.2392E+06, 9.2704E+06, 9.3017E+06,
     + 9.3330E+06, 9.3644E+06, 9.3959E+06, 9.4275E+06, 9.4592E+06,
     +  9.4910E+06, 9.5228E+06, 9.5548E+06, 9.5868E+06, 9.6189E+06,
     + 9.6511E+06, 9.6834E+06, 9.7158E+06, 9.7482E+06, 9.7808E+06,
     +  9.8134E+06, 9.8461E+06, 9.8789E+06, 9.9118E+06, 9.9448E+06,
     + 9.9779E+06, 1.0011E+07, 1.0044E+07, 1.0078E+07, 1.0111E+07,
     +  1.0145E+07, 1.0178E+07, 1.0212E+07, 1.0246E+07, 1.0280E+07,
     + 1.0314E+07, 1.0348E+07, 1.0382E+07, 1.0416E+07, 1.0450E+07,
     +  1.0485E+07, 1.0519E+07, 1.0554E+07, 1.0588E+07, 1.0623E+07,
     + 1.0658E+07, 1.0693E+07, 1.0728E+07, 1.0763E+07, 1.0799E+07,
     +  1.0834E+07, 1.0869E+07, 1.0905E+07, 1.0940E+07, 1.0976E+07,
     + 1.1012E+07, 1.1048E+07, 1.1084E+07, 1.1120E+07, 1.1156E+07,
     +  1.1192E+07, 1.1229E+07, 1.1265E+07, 1.1302E+07, 1.1338E+07,
     + 1.1375E+07, 1.1412E+07, 1.1449E+07, 1.1486E+07, 1.1523E+07,
     +  1.1560E+07, 1.1598E+07, 1.1635E+07, 1.1673E+07, 1.1710E+07,
     + 1.1748E+07, 1.1786E+07, 1.1824E+07, 1.1862E+07, 1.1900E+07,
     +  1.1938E+07, 1.1976E+07, 1.2015E+07, 1.2053E+07, 1.2092E+07,
     + 1.2131E+07, 1.2169E+07, 1.2208E+07, 1.2247E+07, 1.2286E+07,
     +  1.2326E+07, 1.2365E+07, 1.2404E+07, 1.2444E+07, 1.2483E+07,
     + 1.2523E+07, 1.2563E+07, 1.2603E+07, 1.2643E+07, 1.2683E+07,
     +  1.2723E+07, 1.2764E+07, 1.2804E+07, 1.2845E+07, 1.2885E+07,
     + 1.2926E+07, 1.2967E+07, 1.3008E+07, 1.3049E+07, 1.3090E+07,
     +  1.3131E+07, 1.3172E+07, 1.3214E+07, 1.3256E+07, 1.3297E+07,
     + 1.3339E+07, 1.3381E+07, 1.3423E+07, 1.3465E+07, 1.3507E+07,
     +  1.3549E+07, 1.3592E+07, 1.3634E+07, 1.3677E+07, 1.3720E+07,
     + 1.3763E+07, 1.3806E+07, 1.3849E+07, 1.3892E+07, 1.3935E+07,
     +  1.3978E+07, 1.4022E+07, 1.4065E+07, 1.4109E+07, 1.4153E+07,
     + 1.4197E+07, 1.4241E+07, 1.4285E+07, 1.4329E+07, 1.4374E+07,
     +  1.4418E+07, 1.4463E+07, 1.4507E+07, 1.4552E+07, 1.4597E+07,
     + 1.4642E+07, 1.4687E+07, 1.4733E+07, 1.4778E+07, 1.4823E+07,
     +  1.4869E+07, 1.4915E+07, 1.4961E+07, 1.5007E+07, 1.5053E+07,
     + 1.5099E+07, 1.5145E+07, 1.5191E+07, 1.5238E+07, 1.5284E+07,
     +  1.5331E+07, 1.5378E+07, 1.5425E+07, 1.5472E+07, 1.5519E+07,
     + 1.5567E+07, 1.5614E+07, 1.5662E+07, 1.5709E+07, 1.5757E+07,
     +  1.5805E+07, 1.5853E+07, 1.5901E+07, 1.5949E+07, 1.5998E+07,
     + 1.6046E+07, 1.6095E+07, 1.6143E+07, 1.6192E+07, 1.6241E+07,
     +  1.6290E+07, 1.6339E+07, 1.6389E+07, 1.6438E+07, 1.6488E+07,
     + 1.6537E+07, 1.6587E+07, 1.6637E+07, 1.6687E+07, 1.6737E+07,
     +  1.6787E+07, 1.6838E+07, 1.6888E+07, 1.6939E+07, 1.6990E+07,
     + 1.7041E+07, 1.7092E+07, 1.7143E+07, 1.7194E+07, 1.7245E+07,
     +  1.7297E+07, 1.7348E+07, 1.7400E+07, 1.7452E+07, 1.7504E+07,
     + 1.7556E+07, 1.7608E+07, 1.7661E+07, 1.7713E+07, 1.7766E+07,
     +  1.7819E+07, 1.7871E+07, 1.7924E+07, 1.7978E+07, 1.8031E+07,
     + 1.8084E+07, 1.8138E+07, 1.8191E+07, 1.8245E+07, 1.8299E+07,
     +  1.8353E+07, 1.8407E+07, 1.8461E+07, 1.8516E+07, 1.8570E+07,
     + 1.8625E+07, 1.8680E+07, 1.8735E+07, 1.8790E+07, 1.8845E+07,
     +  1.8900E+07, 1.8956E+07, 1.9011E+07, 1.9067E+07, 1.9123E+07,
     + 1.9179E+07, 1.9235E+07, 1.9291E+07, 1.9347E+07, 1.9404E+07,
     +  1.9460E+07, 1.9517E+07, 1.9574E+07, 1.9631E+07, 1.9688E+07,
     + 1.9745E+07, 1.9803E+07, 1.9860E+07, 1.9918E+07, 1.9976E+07,
     +  2.0034E+07, 2.0092E+07, 2.0150E+07, 2.0208E+07, 2.0267E+07,
     + 2.0326E+07, 2.0384E+07, 2.0443E+07, 2.0502E+07, 2.0562E+07,
     +  2.0621E+07, 2.0680E+07, 2.0740E+07, 2.0800E+07, 2.0859E+07,
     + 2.0919E+07, 2.0980E+07, 2.1040E+07, 2.1100E+07, 2.1161E+07,
     +  2.1222E+07, 2.1282E+07, 2.1343E+07, 2.1405E+07, 2.1466E+07,
     + 2.1527E+07, 2.1589E+07, 2.1650E+07, 2.1712E+07, 2.1774E+07,
     +  2.1836E+07, 2.1899E+07, 2.1961E+07, 2.2024E+07, 2.2086E+07,
     + 2.2149E+07, 2.2212E+07, 2.2275E+07, 2.2338E+07, 2.2402E+07,
     +  2.2465E+07, 2.2529E+07, 2.2593E+07, 2.2657E+07, 2.2721E+07,
     + 2.2785E+07, 2.2850E+07, 2.2914E+07, 2.2979E+07, 2.3044E+07,
     +  2.3109E+07, 2.3174E+07, 2.3239E+07, 2.3305E+07, 2.3370E+07,
     + 2.3436E+07, 2.3502E+07, 2.3568E+07, 2.3634E+07, 2.3701E+07,
     +  2.3767E+07, 2.3834E+07, 2.3901E+07, 2.3967E+07, 2.4035E+07,
     + 2.4102E+07, 2.4169E+07, 2.4237E+07, 2.4304E+07, 2.4372E+07,
     +  2.4440E+07, 2.4508E+07, 2.4577E+07, 2.4645E+07, 2.4714E+07,
     + 2.4783E+07, 2.4852E+07, 2.4921E+07, 2.4990E+07, 2.5059E+07,
     +  2.5129E+07, 2.5199E+07, 2.5268E+07, 2.5338E+07, 2.5409E+07,
     + 2.5479E+07, 2.5549E+07, 2.5620E+07, 2.5691E+07, 2.5762E+07,
     +  2.5833E+07, 2.5904E+07, 2.5976E+07, 2.6047E+07, 2.6119E+07,
     + 2.6191E+07, 2.6263E+07, 2.6335E+07, 2.6407E+07, 2.6480E+07,
     +  2.6553E+07, 2.6626E+07, 2.6699E+07, 2.6772E+07, 2.6845E+07,
     + 2.6919E+07, 2.6992E+07, 2.7066E+07, 2.7140E+07, 2.7214E+07,
     +  2.7288E+07, 2.7363E+07, 2.7438E+07, 2.7512E+07, 2.7587E+07,
     + 2.7663E+07, 2.7738E+07, 2.7813E+07, 2.7889E+07, 2.7965E+07,
     +  2.8041E+07, 2.8117E+07, 2.8193E+07, 2.8270E+07, 2.8346E+07,
     + 2.8423E+07, 2.8500E+07, 2.8577E+07, 2.8654E+07, 2.8732E+07,
     +  2.8809E+07, 2.8887E+07, 2.8965E+07, 2.9043E+07, 2.9122E+07,
     + 2.9200E+07, 2.9279E+07, 2.9358E+07, 2.9437E+07, 2.9516E+07,
     +  2.9595E+07, 2.9675E+07, 2.9754E+07, 2.9834E+07, 2.9914E+07,
     + 2.9994E+07, 3.0075E+07, 3.0155E+07, 3.0236E+07, 3.0317E+07,
     +  3.0398E+07, 3.0479E+07, 3.0560E+07, 3.0642E+07, 3.0724E+07,
     + 3.0806E+07, 3.0888E+07, 3.0970E+07, 3.1053E+07, 3.1135E+07,
     +  3.1218E+07, 3.1301E+07, 3.1384E+07, 3.1467E+07, 3.1551E+07,
     + 3.1635E+07, 3.1719E+07, 3.1803E+07, 3.1887E+07, 3.1971E+07,
     +  3.2056E+07, 3.2141E+07, 3.2226E+07, 3.2311E+07, 3.2396E+07,
     + 3.2482E+07, 3.2567E+07, 3.2653E+07, 3.2739E+07, 3.2825E+07,
     +  3.2912E+07, 3.2998E+07, 3.3085E+07, 3.3172E+07, 3.3259E+07,
     + 3.3347E+07, 3.3434E+07, 3.3522E+07, 3.3610E+07, 3.3698E+07,
     +  3.3786E+07, 3.3874E+07, 3.3963E+07, 3.4052E+07, 3.4141E+07,
     + 3.4230E+07, 3.4319E+07, 3.4409E+07, 3.4499E+07, 3.4589E+07,
     +  3.4679E+07, 3.4769E+07, 3.4859E+07, 3.4950E+07, 3.5041E+07,
     + 3.5132E+07, 3.5223E+07, 3.5315E+07, 3.5406E+07, 3.5498E+07,
     +  3.5590E+07, 3.5682E+07, 3.5775E+07, 3.5868E+07, 3.5960E+07,
     + 3.6053E+07, 3.6146E+07, 3.6240E+07, 3.6333E+07, 3.6427E+07,
     +  3.6521E+07, 3.6615E+07, 3.6710E+07, 3.6804E+07, 3.6899E+07,
     + 3.6994E+07, 3.7089E+07, 3.7184E+07, 3.7280E+07, 3.7375E+07,
     +  3.7471E+07, 3.7568E+07, 3.7664E+07, 3.7760E+07, 3.7857E+07,
     + 3.7954E+07, 3.8051E+07, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 3,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
c...        --       232
      data (QofT( 4,J),J=1,2408)/ 6.7134E+00, 1.3077E+01, 1.9445E+01,
     +  2.5814E+01, 3.2183E+01, 3.8553E+01, 4.4923E+01, 5.1293E+01,
     + 5.7663E+01, 6.4033E+01, 7.0404E+01, 7.6775E+01, 8.3145E+01,
     +  8.9516E+01, 9.5888E+01, 1.0226E+02, 1.0863E+02, 1.1500E+02,
     + 1.2137E+02, 1.2775E+02, 1.4049E+02, 1.5324E+02, 1.6598E+02,
     +  1.7873E+02, 1.9147E+02, 2.0422E+02, 2.1697E+02, 2.2972E+02,
     + 2.4247E+02, 2.5522E+02, 2.6797E+02, 2.8072E+02, 2.9347E+02,
     +  3.0622E+02, 3.1898E+02, 3.3174E+02, 3.4450E+02, 3.5727E+02,
     + 3.7004E+02, 3.8281E+02, 3.9560E+02, 4.0839E+02, 4.2119E+02,
     +  4.3400E+02, 4.4683E+02, 4.5968E+02, 4.7254E+02, 4.8543E+02,
     + 4.9834E+02, 5.1128E+02, 5.2425E+02, 5.3725E+02, 5.5028E+02,
     +  5.6336E+02, 5.7648E+02, 5.8964E+02, 6.0285E+02, 6.1612E+02,
     + 6.2944E+02, 6.4283E+02, 6.5627E+02, 6.6979E+02, 6.8337E+02,
     +  6.9703E+02, 7.1076E+02, 7.2458E+02, 7.3848E+02, 7.5247E+02,
     + 7.6655E+02, 7.8073E+02, 7.9501E+02, 8.0938E+02, 8.2387E+02,
     +  8.3846E+02, 8.5316E+02, 8.6798E+02, 8.8292E+02, 8.9798E+02,
     + 9.1316E+02, 9.2848E+02, 9.4392E+02, 9.5950E+02, 9.7521E+02,
     +  9.9107E+02, 1.0071E+03, 1.0232E+03, 1.0395E+03, 1.0560E+03,
     + 1.0726E+03, 1.0893E+03, 1.1062E+03, 1.1233E+03, 1.1406E+03,
     +  1.1580E+03, 1.1756E+03, 1.1934E+03, 1.2113E+03, 1.2294E+03,
     + 1.2477E+03, 1.2662E+03, 1.2849E+03, 1.3038E+03, 1.3228E+03,
     +  1.3421E+03, 1.3616E+03, 1.3812E+03, 1.4011E+03, 1.4211E+03,
     + 1.4414E+03, 1.4619E+03, 1.4826E+03, 1.5035E+03, 1.5246E+03,
     +  1.5460E+03, 1.5675E+03, 1.5893E+03, 1.6114E+03, 1.6336E+03,
     + 1.6561E+03, 1.6788E+03, 1.7018E+03, 1.7250E+03, 1.7484E+03,
     +  1.7721E+03, 1.7960E+03, 1.8202E+03, 1.8447E+03, 1.8693E+03,
     + 1.8943E+03, 1.9195E+03, 1.9449E+03, 1.9707E+03, 1.9967E+03,
     +  2.0229E+03, 2.0494E+03, 2.0762E+03, 2.1033E+03, 2.1307E+03,
     + 2.1583E+03, 2.1862E+03, 2.2144E+03, 2.2429E+03, 2.2717E+03,
     +  2.3008E+03, 2.3301E+03, 2.3598E+03, 2.3897E+03, 2.4200E+03,
     + 2.4506E+03, 2.4814E+03, 2.5126E+03, 2.5441E+03, 2.5759E+03,
     +  2.6080E+03, 2.6405E+03, 2.6732E+03, 2.7063E+03, 2.7397E+03,
     + 2.7734E+03, 2.8075E+03, 2.8419E+03, 2.8766E+03, 2.9117E+03,
     +  2.9471E+03, 2.9829E+03, 3.0190E+03, 3.0554E+03, 3.0922E+03,
     + 3.1294E+03, 3.1669E+03, 3.2047E+03, 3.2430E+03, 3.2816E+03,
     +  3.3205E+03, 3.3598E+03, 3.3995E+03, 3.4396E+03, 3.4800E+03,
     + 3.5209E+03, 3.5621E+03, 3.6037E+03, 3.6456E+03, 3.6880E+03,
     +  3.7307E+03, 3.7739E+03, 3.8174E+03, 3.8614E+03, 3.9057E+03,
     + 3.9505E+03, 3.9956E+03, 4.0412E+03, 4.0872E+03, 4.1336E+03,
     +  4.1804E+03, 4.2277E+03, 4.2753E+03, 4.3234E+03, 4.3719E+03,
     + 4.4209E+03, 4.4703E+03, 4.5201E+03, 4.5704E+03, 4.6211E+03,
     +  4.6722E+03, 4.7238E+03, 4.7759E+03, 4.8284E+03, 4.8814E+03,
     + 4.9348E+03, 4.9887E+03, 5.0430E+03, 5.0979E+03, 5.1532E+03,
     +  5.2089E+03, 5.2652E+03, 5.3219E+03, 5.3791E+03, 5.4368E+03,
     + 5.4950E+03, 5.5537E+03, 5.6129E+03, 5.6726E+03, 5.7328E+03,
     +  5.7934E+03, 5.8546E+03, 5.9163E+03, 5.9786E+03, 6.0413E+03,
     + 6.1045E+03, 6.1683E+03, 6.2326E+03, 6.2975E+03, 6.3628E+03,
     +  6.4287E+03, 6.4952E+03, 6.5621E+03, 6.6297E+03, 6.6977E+03,
     + 6.7664E+03, 6.8356E+03, 6.9053E+03, 6.9756E+03, 7.0465E+03,
     +  7.1179E+03, 7.1899E+03, 7.2625E+03, 7.3356E+03, 7.4094E+03,
     + 7.4837E+03, 7.5586E+03, 7.6341E+03, 7.7102E+03, 7.7869E+03,
     +  7.8642E+03, 7.9420E+03, 8.0205E+03, 8.0997E+03, 8.1794E+03,
     + 8.2597E+03, 8.3407E+03, 8.4223E+03, 8.5045E+03, 8.5874E+03,
     +  8.6708E+03, 8.7550E+03, 8.8397E+03, 8.9251E+03, 9.0112E+03,
     + 9.0979E+03, 9.1853E+03, 9.2733E+03, 9.3620E+03, 9.4514E+03,
     +  9.5414E+03, 9.6321E+03, 9.7235E+03, 9.8156E+03, 9.9083E+03,
     + 1.0002E+04, 1.0096E+04, 1.0191E+04, 1.0286E+04, 1.0382E+04,
     +  1.0479E+04, 1.0577E+04, 1.0675E+04, 1.0775E+04, 1.0874E+04,
     + 1.0975E+04, 1.1076E+04, 1.1178E+04, 1.1281E+04, 1.1384E+04,
     +  1.1489E+04, 1.1594E+04, 1.1699E+04, 1.1806E+04, 1.1913E+04,
     + 1.2021E+04, 1.2130E+04, 1.2240E+04, 1.2350E+04, 1.2461E+04,
     +  1.2573E+04, 1.2686E+04, 1.2799E+04, 1.2913E+04, 1.3029E+04,
     + 1.3144E+04, 1.3261E+04, 1.3379E+04, 1.3497E+04, 1.3616E+04,
     +  1.3736E+04, 1.3857E+04, 1.3979E+04, 1.4101E+04, 1.4225E+04,
     + 1.4349E+04, 1.4474E+04, 1.4600E+04, 1.4727E+04, 1.4854E+04,
     +  1.4983E+04, 1.5112E+04, 1.5242E+04, 1.5373E+04, 1.5506E+04,
     + 1.5638E+04, 1.5772E+04, 1.5907E+04, 1.6043E+04, 1.6179E+04,
     +  1.6317E+04, 1.6455E+04, 1.6594E+04, 1.6734E+04, 1.6876E+04,
     + 1.7018E+04, 1.7161E+04, 1.7305E+04, 1.7450E+04, 1.7596E+04,
     +  1.7742E+04, 1.7890E+04, 1.8039E+04, 1.8189E+04, 1.8339E+04,
     + 1.8491E+04, 1.8644E+04, 1.8798E+04, 1.8952E+04, 1.9108E+04,
     +  1.9265E+04, 1.9422E+04, 1.9581E+04, 1.9741E+04, 1.9902E+04,
     + 2.0063E+04, 2.0226E+04, 2.0390E+04, 2.0555E+04, 2.0721E+04,
     +  2.0888E+04, 2.1056E+04, 2.1225E+04, 2.1396E+04, 2.1567E+04,
     + 2.1739E+04, 2.1913E+04, 2.2087E+04, 2.2263E+04, 2.2440E+04,
     +  2.2618E+04, 2.2797E+04, 2.2977E+04, 2.3158E+04, 2.3340E+04,
     + 2.3524E+04, 2.3708E+04, 2.3894E+04, 2.4081E+04, 2.4269E+04,
     +  2.4458E+04, 2.4649E+04, 2.4840E+04, 2.5033E+04, 2.5227E+04,
     + 2.5422E+04, 2.5618E+04, 2.5816E+04, 2.6014E+04, 2.6214E+04,
     +  2.6415E+04, 2.6617E+04, 2.6821E+04, 2.7026E+04, 2.7232E+04,
     + 2.7439E+04, 2.7647E+04, 2.7857E+04, 2.8068E+04, 2.8280E+04,
     +  2.8494E+04, 2.8709E+04, 2.8925E+04, 2.9142E+04, 2.9361E+04,
     + 2.9581E+04, 2.9802E+04, 3.0024E+04, 3.0248E+04, 3.0473E+04,
     +  3.0700E+04, 3.0928E+04, 3.1157E+04, 3.1387E+04, 3.1619E+04,
     + 3.1852E+04, 3.2087E+04, 3.2323E+04, 3.2560E+04, 3.2799E+04,
     +  3.3039E+04, 3.3280E+04, 3.3523E+04, 3.3767E+04, 3.4013E+04,
     + 3.4260E+04, 3.4509E+04, 3.4759E+04, 3.5010E+04, 3.5263E+04,
     +  3.5517E+04, 3.5773E+04, 3.6030E+04, 3.6288E+04, 3.6548E+04,
     + 3.6810E+04, 3.7073E+04, 3.7337E+04, 3.7603E+04, 3.7871E+04,
     +  3.8140E+04, 3.8410E+04, 3.8682E+04, 3.8956E+04, 3.9231E+04,
     + 3.9508E+04, 3.9786E+04, 4.0065E+04, 4.0347E+04, 4.0629E+04,
     +  4.0914E+04, 4.1200E+04, 4.1487E+04, 4.1777E+04, 4.2067E+04,
     + 4.2360E+04, 4.2654E+04, 4.2949E+04, 4.3246E+04, 4.3545E+04,
     +  4.3846E+04, 4.4148E+04, 4.4452E+04, 4.4757E+04, 4.5064E+04,
     + 4.5373E+04, 4.5683E+04, 4.5995E+04, 4.6309E+04, 4.6625E+04,
     +  4.6942E+04, 4.7261E+04, 4.7581E+04, 4.7904E+04, 4.8228E+04,
     + 4.8554E+04, 4.8881E+04, 4.9210E+04, 4.9541E+04, 4.9874E+04,
     +  5.0209E+04, 5.0545E+04, 5.0883E+04, 5.1223E+04, 5.1565E+04,
     + 5.1909E+04, 5.2254E+04, 5.2601E+04, 5.2950E+04, 5.3301E+04,
     +  5.3654E+04, 5.4009E+04, 5.4365E+04, 5.4724E+04, 5.5084E+04,
     + 5.5446E+04, 5.5810E+04, 5.6176E+04, 5.6543E+04, 5.6913E+04,
     +  5.7285E+04, 5.7658E+04, 5.8034E+04, 5.8411E+04, 5.8790E+04,
     + 5.9172E+04, 5.9555E+04, 5.9940E+04, 6.0327E+04, 6.0717E+04,
     +  6.1108E+04, 6.1501E+04, 6.1896E+04, 6.2293E+04, 6.2693E+04,
     + 6.3094E+04, 6.3497E+04, 6.3903E+04, 6.4310E+04, 6.4720E+04,
     +  6.5131E+04, 6.5545E+04, 6.5961E+04, 6.6378E+04, 6.6798E+04,
     + 6.7220E+04, 6.7645E+04, 6.8071E+04, 6.8499E+04, 6.8930E+04,
     +  6.9363E+04, 6.9798E+04, 7.0235E+04, 7.0674E+04, 7.1115E+04,
     + 7.1559E+04, 7.2005E+04, 7.2453E+04, 7.2903E+04, 7.3355E+04,
     +  7.3810E+04, 7.4267E+04, 7.4726E+04, 7.5188E+04, 7.5651E+04,
     + 7.6117E+04, 7.6586E+04, 7.7056E+04, 7.7529E+04, 7.8004E+04,
     +  7.8482E+04, 7.8962E+04, 7.9444E+04, 7.9928E+04, 8.0415E+04,
     + 8.0904E+04, 8.1396E+04, 8.1890E+04, 8.2386E+04, 8.2885E+04,
     +  8.3386E+04, 8.3890E+04, 8.4396E+04, 8.4904E+04, 8.5415E+04,
     + 8.5928E+04, 8.6444E+04, 8.6962E+04, 8.7483E+04, 8.8006E+04,
     +  8.8532E+04, 8.9060E+04, 8.9590E+04, 9.0124E+04, 9.0659E+04,
     + 9.1198E+04, 9.1738E+04, 9.2282E+04, 9.2828E+04, 9.3376E+04,
     +  9.3927E+04, 9.4481E+04, 9.5037E+04, 9.5596E+04, 9.6157E+04,
     + 9.6721E+04, 9.7288E+04, 9.7858E+04, 9.8430E+04, 9.9004E+04,
     +  9.9582E+04, 1.0016E+05, 1.0074E+05, 1.0133E+05, 1.0192E+05,
     + 1.0251E+05, 1.0310E+05, 1.0370E+05, 1.0430E+05, 1.0490E+05,
     +  1.0550E+05, 1.0611E+05, 1.0672E+05, 1.0734E+05, 1.0795E+05,
     + 1.0857E+05, 1.0919E+05, 1.0982E+05, 1.1044E+05, 1.1107E+05,
     +  1.1171E+05, 1.1234E+05, 1.1298E+05, 1.1362E+05, 1.1427E+05,
     + 1.1491E+05, 1.1556E+05, 1.1622E+05, 1.1687E+05, 1.1753E+05,
     +  1.1819E+05, 1.1886E+05, 1.1953E+05, 1.2020E+05, 1.2087E+05,
     + 1.2155E+05, 1.2223E+05, 1.2291E+05, 1.2360E+05, 1.2429E+05,
     +  1.2498E+05, 1.2568E+05, 1.2637E+05, 1.2708E+05, 1.2778E+05,
     + 1.2849E+05, 1.2920E+05, 1.2991E+05, 1.3063E+05, 1.3135E+05,
     +  1.3207E+05, 1.3280E+05, 1.3353E+05, 1.3426E+05, 1.3500E+05,
     + 1.3574E+05, 1.3648E+05, 1.3723E+05, 1.3798E+05, 1.3873E+05,
     +  1.3949E+05, 1.4024E+05, 1.4101E+05, 1.4177E+05, 1.4254E+05,
     + 1.4331E+05, 1.4409E+05, 1.4487E+05, 1.4565E+05, 1.4644E+05,
     +  1.4723E+05, 1.4802E+05, 1.4881E+05, 1.4961E+05, 1.5042E+05,
     + 1.5122E+05, 1.5203E+05, 1.5285E+05, 1.5366E+05, 1.5448E+05,
     +  1.5531E+05, 1.5613E+05, 1.5696E+05, 1.5780E+05, 1.5864E+05,
     + 1.5948E+05, 1.6032E+05, 1.6117E+05, 1.6202E+05, 1.6288E+05,
     +  1.6374E+05, 1.6460E+05, 1.6547E+05, 1.6634E+05, 1.6721E+05,
     + 1.6809E+05, 1.6897E+05, 1.6985E+05, 1.7074E+05, 1.7163E+05,
     +  1.7253E+05, 1.7343E+05, 1.7433E+05, 1.7524E+05, 1.7615E+05,
     + 1.7707E+05, 1.7798E+05, 1.7891E+05, 1.7983E+05, 1.8076E+05,
     +  1.8170E+05, 1.8263E+05, 1.8358E+05, 1.8452E+05, 1.8547E+05,
     + 1.8642E+05, 1.8738E+05, 1.8834E+05, 1.8931E+05, 1.9028E+05,
     +  1.9125E+05, 1.9223E+05, 1.9321E+05, 1.9419E+05, 1.9518E+05,
     + 1.9617E+05, 1.9717E+05, 1.9817E+05, 1.9918E+05, 2.0019E+05,
     +  2.0120E+05, 2.0222E+05, 2.0324E+05, 2.0426E+05, 2.0529E+05,
     + 2.0633E+05, 2.0736E+05, 2.0841E+05, 2.0945E+05, 2.1050E+05,
     +  2.1156E+05, 2.1262E+05, 2.1368E+05, 2.1475E+05, 2.1582E+05,
     + 2.1690E+05, 2.1798E+05, 2.1906E+05, 2.2015E+05, 2.2124E+05,
     +  2.2234E+05, 2.2344E+05, 2.2455E+05, 2.2566E+05, 2.2678E+05,
     + 2.2789E+05, 2.2902E+05, 2.3015E+05, 2.3128E+05, 2.3242E+05,
     +  2.3356E+05, 2.3471E+05, 2.3586E+05, 2.3701E+05, 2.3817E+05,
     + 2.3934E+05, 2.4050E+05, 2.4168E+05, 2.4286E+05, 2.4404E+05,
     +  2.4523E+05, 2.4642E+05, 2.4761E+05, 2.4882E+05, 2.5002E+05,
     + 2.5123E+05, 2.5245E+05, 2.5367E+05, 2.5489E+05, 2.5612E+05,
     +  2.5736E+05, 2.5859E+05, 2.5984E+05, 2.6109E+05, 2.6234E+05,
     + 2.6360E+05, 2.6486E+05, 2.6613E+05, 2.6740E+05, 2.6868E+05,
     +  2.6996E+05, 2.7125E+05, 2.7254E+05, 2.7384E+05, 2.7514E+05,
     + 2.7645E+05, 2.7776E+05, 2.7907E+05, 2.8040E+05, 2.8172E+05,
     +  2.8305E+05, 2.8439E+05, 2.8573E+05, 2.8708E+05, 2.8843E+05,
     + 2.8979E+05, 2.9115E+05, 2.9252E+05, 2.9389E+05, 2.9527E+05,
     +  2.9665E+05, 2.9804E+05, 2.9944E+05, 3.0083E+05, 3.0224E+05,
     + 3.0365E+05, 3.0506E+05, 3.0648E+05, 3.0791E+05, 3.0934E+05,
     +  3.1077E+05, 3.1221E+05, 3.1366E+05, 3.1511E+05, 3.1657E+05,
     + 3.1803E+05, 3.1950E+05, 3.2097E+05, 3.2245E+05, 3.2393E+05,
     +  3.2542E+05, 3.2692E+05, 3.2842E+05, 3.2992E+05, 3.3143E+05,
     + 3.3295E+05, 3.3447E+05, 3.3600E+05, 3.3754E+05, 3.3908E+05,
     +  3.4062E+05, 3.4217E+05, 3.4373E+05, 3.4529E+05, 3.4686E+05,
     + 3.4843E+05, 3.5001E+05, 3.5160E+05, 3.5319E+05, 3.5478E+05,
     +  3.5638E+05, 3.5799E+05, 3.5961E+05, 3.6123E+05, 3.6285E+05,
     + 3.6448E+05, 3.6612E+05, 3.6776E+05, 3.6941E+05, 3.7107E+05,
     +  3.7273E+05, 3.7439E+05, 3.7607E+05, 3.7774E+05, 3.7943E+05,
     + 3.8112E+05, 3.8282E+05, 3.8452E+05, 3.8623E+05, 3.8794E+05,
     +  3.8966E+05, 3.9139E+05, 3.9312E+05, 3.9486E+05, 3.9661E+05,
     + 3.9836E+05, 4.0012E+05, 4.0188E+05, 4.0365E+05, 4.0543E+05,
     +  4.0721E+05, 4.0900E+05, 4.1080E+05, 4.1260E+05, 4.1441E+05,
     + 4.1622E+05, 4.1804E+05, 4.1987E+05, 4.2171E+05, 4.2355E+05,
     +  4.2539E+05, 4.2724E+05, 4.2910E+05, 4.3097E+05, 4.3284E+05,
     + 4.3472E+05, 4.3661E+05, 4.3850E+05, 4.4040E+05, 4.4230E+05,
     +  4.4421E+05, 4.4613E+05, 4.4806E+05, 4.4999E+05, 4.5193E+05,
     + 4.5387E+05, 4.5582E+05, 4.5778E+05, 4.5975E+05, 4.6172E+05,
     +  4.6370E+05, 4.6568E+05, 4.6768E+05, 4.6968E+05, 4.7168E+05,
     + 4.7370E+05, 4.7572E+05, 4.7774E+05, 4.7978E+05, 4.8182E+05,
     +  4.8386E+05, 4.8592E+05, 4.8798E+05, 4.9005E+05, 4.9212E+05,
     + 4.9421E+05, 4.9630E+05, 4.9839E+05, 5.0050E+05, 5.0261E+05,
     +  5.0473E+05, 5.0685E+05, 5.0898E+05, 5.1112E+05, 5.1327E+05,
     + 5.1542E+05, 5.1759E+05, 5.1975E+05, 5.2193E+05, 5.2411E+05,
     +  5.2630E+05, 5.2850E+05, 5.3071E+05, 5.3292E+05, 5.3514E+05,
     + 5.3737E+05, 5.3960E+05, 5.4184E+05, 5.4409E+05, 5.4635E+05,
     +  5.4861E+05, 5.5089E+05, 5.5317E+05, 5.5545E+05, 5.5775E+05,
     + 5.6005E+05, 5.6236E+05, 5.6468E+05, 5.6700E+05, 5.6934E+05,
     +  5.7168E+05, 5.7403E+05, 5.7638E+05, 5.7875E+05, 5.8112E+05,
     + 5.8350E+05, 5.8588E+05, 5.8828E+05, 5.9068E+05, 5.9309E+05,
     +  5.9551E+05, 5.9794E+05, 6.0037E+05, 6.0282E+05, 6.0527E+05,
     + 6.0773E+05, 6.1019E+05, 6.1267E+05, 6.1515E+05, 6.1764E+05,
     +  6.2014E+05, 6.2265E+05, 6.2516E+05, 6.2768E+05, 6.3022E+05,
     + 6.3275E+05, 6.3530E+05, 6.3786E+05, 6.4042E+05, 6.4299E+05,
     +  6.4558E+05, 6.4816E+05, 6.5076E+05, 6.5337E+05, 6.5598E+05,
     + 6.5860E+05, 6.6123E+05, 6.6387E+05, 6.6652E+05, 6.6918E+05,
     +  6.7184E+05, 6.7452E+05, 6.7720E+05, 6.7989E+05, 6.8259E+05,
     + 6.8529E+05, 6.8801E+05, 6.9073E+05, 6.9347E+05, 6.9621E+05,
     +  6.9896E+05, 7.0172E+05, 7.0449E+05, 7.0726E+05, 7.1005E+05,
     + 7.1285E+05, 7.1565E+05, 7.1846E+05, 7.2128E+05, 7.2411E+05,
     +  7.2695E+05, 7.2980E+05, 7.3265E+05, 7.3552E+05, 7.3839E+05,
     + 7.4128E+05, 7.4417E+05, 7.4707E+05, 7.4998E+05, 7.5290E+05,
     +  7.5583E+05, 7.5877E+05, 7.6172E+05, 7.6467E+05, 7.6764E+05,
     + 7.7061E+05, 7.7360E+05, 7.7659E+05, 7.7959E+05, 7.8261E+05,
     +  7.8563E+05, 7.8866E+05, 7.9170E+05, 7.9475E+05, 7.9781E+05,
     + 8.0088E+05, 8.0395E+05, 8.0704E+05, 8.1014E+05, 8.1324E+05,
     +  8.1636E+05, 8.1949E+05, 8.2262E+05, 8.2576E+05, 8.2892E+05,
     + 8.3208E+05, 8.3526E+05, 8.3844E+05, 8.4163E+05, 8.4484E+05,
     +  8.4805E+05, 8.5127E+05, 8.5450E+05, 8.5775E+05, 8.6100E+05,
     + 8.6426E+05, 8.6753E+05, 8.7081E+05, 8.7410E+05, 8.7741E+05,
     +  8.8072E+05, 8.8404E+05, 8.8737E+05, 8.9071E+05, 8.9406E+05,
     + 8.9743E+05, 9.0080E+05, 9.0418E+05, 9.0757E+05, 9.1098E+05,
     +  9.1439E+05, 9.1781E+05, 9.2125E+05, 9.2469E+05, 9.2814E+05,
     + 9.3161E+05, 9.3508E+05, 9.3857E+05, 9.4206E+05, 9.4557E+05,
     +  9.4909E+05, 9.5261E+05, 9.5615E+05, 9.5970E+05, 9.6326E+05,
     + 9.6683E+05, 9.7041E+05, 9.7400E+05, 9.7760E+05, 9.8121E+05,
     +  9.8483E+05, 9.8847E+05, 9.9211E+05, 9.9577E+05, 9.9943E+05,
     + 1.0031E+06, 1.0068E+06, 1.0105E+06, 1.0142E+06, 1.0179E+06,
     +  1.0217E+06, 1.0254E+06, 1.0291E+06, 1.0329E+06, 1.0367E+06,
     + 1.0405E+06, 1.0443E+06, 1.0481E+06, 1.0519E+06, 1.0557E+06,
     +  1.0596E+06, 1.0634E+06, 1.0673E+06, 1.0712E+06, 1.0751E+06,
     + 1.0789E+06, 1.0829E+06, 1.0868E+06, 1.0907E+06, 1.0947E+06,
     +  1.0986E+06, 1.1026E+06, 1.1066E+06, 1.1105E+06, 1.1145E+06,
     + 1.1186E+06, 1.1226E+06, 1.1266E+06, 1.1307E+06, 1.1347E+06,
     +  1.1388E+06, 1.1429E+06, 1.1470E+06, 1.1511E+06, 1.1552E+06,
     + 1.1593E+06, 1.1635E+06, 1.1676E+06, 1.1718E+06, 1.1760E+06,
     +  1.1802E+06, 1.1844E+06, 1.1886E+06, 1.1928E+06, 1.1970E+06,
     + 1.2013E+06, 1.2055E+06, 1.2098E+06, 1.2141E+06, 1.2184E+06,
     +  1.2227E+06, 1.2270E+06, 1.2314E+06, 1.2357E+06, 1.2401E+06,
     + 1.2444E+06, 1.2488E+06, 1.2532E+06, 1.2576E+06, 1.2621E+06,
     +  1.2665E+06, 1.2709E+06, 1.2754E+06, 1.2799E+06, 1.2843E+06,
     + 1.2888E+06, 1.2934E+06, 1.2979E+06, 1.3024E+06, 1.3070E+06,
     +  1.3115E+06, 1.3161E+06, 1.3207E+06, 1.3253E+06, 1.3299E+06,
     + 1.3345E+06, 1.3391E+06, 1.3438E+06, 1.3485E+06, 1.3531E+06,
     +  1.3578E+06, 1.3625E+06, 1.3672E+06, 1.3720E+06, 1.3767E+06,
     + 1.3815E+06, 1.3862E+06, 1.3910E+06, 1.3958E+06, 1.4006E+06,
     +  1.4054E+06, 1.4103E+06, 1.4151E+06, 1.4200E+06, 1.4248E+06,
     + 1.4297E+06, 1.4346E+06, 1.4395E+06, 1.4445E+06, 1.4494E+06,
     +  1.4544E+06, 1.4593E+06, 1.4643E+06, 1.4693E+06, 1.4743E+06,
     + 1.4793E+06, 1.4844E+06, 1.4894E+06, 1.4945E+06, 1.4996E+06,
     +  1.5047E+06, 1.5098E+06, 1.5149E+06, 1.5200E+06, 1.5252E+06,
     + 1.5303E+06, 1.5355E+06, 1.5407E+06, 1.5459E+06, 1.5511E+06,
     +  1.5564E+06, 1.5616E+06, 1.5669E+06, 1.5721E+06, 1.5774E+06,
     + 1.5827E+06, 1.5880E+06, 1.5934E+06, 1.5987E+06, 1.6041E+06,
     +  1.6095E+06, 1.6149E+06, 1.6203E+06, 1.6257E+06, 1.6311E+06,
     + 1.6366E+06, 1.6420E+06, 1.6475E+06, 1.6530E+06, 1.6585E+06,
     +  1.6640E+06, 1.6695E+06, 1.6751E+06, 1.6807E+06, 1.6862E+06,
     + 1.6918E+06, 1.6974E+06, 1.7031E+06, 1.7087E+06, 1.7144E+06,
     +  1.7200E+06, 1.7257E+06, 1.7314E+06, 1.7371E+06, 1.7429E+06,
     + 1.7486E+06, 1.7544E+06, 1.7602E+06, 1.7659E+06, 1.7717E+06,
     +  1.7776E+06, 1.7834E+06, 1.7893E+06, 1.7951E+06, 1.8010E+06,
     + 1.8069E+06, 1.8128E+06, 1.8188E+06, 1.8247E+06, 1.8307E+06,
     +  1.8366E+06, 1.8426E+06, 1.8486E+06, 1.8547E+06, 1.8607E+06,
     + 1.8668E+06, 1.8728E+06, 1.8789E+06, 1.8850E+06, 1.8911E+06,
     +  1.8973E+06, 1.9034E+06, 1.9096E+06, 1.9158E+06, 1.9220E+06,
     + 1.9282E+06, 1.9344E+06, 1.9407E+06, 1.9469E+06, 1.9532E+06,
     +  1.9595E+06, 1.9658E+06, 1.9721E+06, 1.9785E+06, 1.9848E+06,
     + 1.9912E+06, 1.9976E+06, 2.0040E+06, 2.0105E+06, 2.0169E+06,
     +  2.0234E+06, 2.0298E+06, 2.0363E+06, 2.0428E+06, 2.0494E+06,
     + 2.0559E+06, 2.0625E+06, 2.0691E+06, 2.0756E+06, 2.0823E+06,
     +  2.0889E+06, 2.0955E+06, 2.1022E+06, 2.1089E+06, 2.1156E+06,
     + 2.1223E+06, 2.1290E+06, 2.1358E+06, 2.1425E+06, 2.1493E+06,
     +  2.1561E+06, 2.1629E+06, 2.1697E+06, 2.1766E+06, 2.1835E+06,
     + 2.1904E+06, 2.1973E+06, 2.2042E+06, 2.2111E+06, 2.2181E+06,
     +  2.2250E+06, 2.2320E+06, 2.2390E+06, 2.2461E+06, 2.2531E+06,
     + 2.2602E+06, 2.2673E+06, 2.2743E+06, 2.2815E+06, 2.2886E+06,
     +  2.2957E+06, 2.3029E+06, 2.3101E+06, 2.3173E+06, 2.3245E+06,
     + 2.3318E+06, 2.3390E+06, 2.3463E+06, 2.3536E+06, 2.3609E+06,
     +  2.3682E+06, 2.3756E+06, 2.3830E+06, 2.3903E+06, 2.3978E+06,
     + 2.4052E+06, 2.4126E+06, 2.4201E+06, 2.4276E+06, 2.4351E+06,
     +  2.4426E+06, 2.4501E+06, 2.4577E+06, 2.4652E+06, 2.4728E+06,
     + 2.4804E+06, 2.4881E+06, 2.4957E+06, 2.5034E+06, 2.5110E+06,
     +  2.5188E+06, 2.5265E+06, 2.5342E+06, 2.5420E+06, 2.5498E+06,
     + 2.5576E+06, 2.5654E+06, 2.5732E+06, 2.5811E+06, 2.5889E+06,
     +  2.5968E+06, 2.6048E+06, 2.6127E+06, 2.6206E+06, 2.6286E+06,
     + 2.6366E+06, 2.6446E+06, 2.6526E+06, 2.6607E+06, 2.6688E+06,
     +  2.6769E+06, 2.6850E+06, 2.6931E+06, 2.7012E+06, 2.7094E+06,
     + 2.7176E+06, 2.7258E+06, 2.7340E+06, 2.7423E+06, 2.7506E+06,
     +  2.7588E+06, 2.7672E+06, 2.7755E+06, 2.7838E+06, 2.7922E+06,
     + 2.8006E+06, 2.8090E+06, 2.8174E+06, 2.8259E+06, 2.8343E+06,
     +  2.8428E+06, 2.8514E+06, 2.8599E+06, 2.8684E+06, 2.8770E+06,
     + 2.8856E+06, 2.8942E+06, 2.9028E+06, 2.9115E+06, 2.9202E+06,
     +  2.9289E+06, 2.9376E+06, 2.9463E+06, 2.9551E+06, 2.9639E+06,
     + 2.9727E+06, 2.9815E+06, 2.9903E+06, 2.9992E+06, 3.0081E+06,
     +  3.0170E+06, 3.0259E+06, 3.0349E+06, 3.0438E+06, 3.0528E+06,
     + 3.0618E+06, 3.0709E+06, 3.0799E+06, 3.0890E+06, 3.0981E+06,
     +  3.1072E+06, 3.1164E+06, 3.1255E+06, 3.1347E+06, 3.1439E+06,
     + 3.1531E+06, 3.1624E+06, 3.1717E+06, 3.1810E+06, 3.1903E+06,
     +  3.1996E+06, 3.2090E+06, 3.2184E+06, 3.2278E+06, 3.2372E+06,
     + 3.2466E+06, 3.2561E+06, 3.2656E+06, 3.2751E+06, 3.2846E+06,
     +  3.2942E+06, 3.3038E+06, 3.3134E+06, 3.3230E+06, 3.3327E+06,
     + 3.3423E+06, 3.3520E+06, 3.3617E+06, 3.3715E+06, 3.3812E+06,
     +  3.3910E+06, 3.4008E+06, 3.4107E+06, 3.4205E+06, 3.4304E+06,
     + 3.4403E+06, 3.4502E+06, 3.4601E+06, 3.4701E+06, 3.4801E+06,
     +  3.4901E+06, 3.5002E+06, 3.5102E+06, 3.5203E+06, 3.5304E+06,
     + 3.5405E+06, 3.5507E+06, 3.5609E+06, 3.5711E+06, 3.5813E+06,
     +  3.5915E+06, 3.6018E+06, 3.6121E+06, 3.6224E+06, 3.6327E+06,
     + 3.6431E+06, 3.6535E+06, 3.6639E+06, 3.6743E+06, 3.6848E+06,
     +  3.6953E+06, 3.7058E+06, 3.7163E+06, 3.7269E+06, 3.7375E+06,
     + 3.7481E+06, 3.7587E+06, 3.7693E+06, 3.7800E+06, 3.7907E+06,
     +  3.8014E+06, 3.8122E+06, 3.8230E+06, 3.8338E+06, 3.8446E+06,
     + 3.8554E+06, 3.8663E+06, 3.8772E+06, 3.8881E+06, 3.8991E+06,
     +  3.9100E+06, 3.9210E+06, 3.9321E+06, 3.9431E+06, 3.9542E+06,
     + 3.9653E+06, 3.9764E+06, 3.9875E+06, 3.9987E+06, 4.0099E+06,
     +  4.0211E+06, 4.0324E+06, 4.0436E+06, 4.0549E+06, 4.0662E+06,
     + 4.0776E+06, 4.0890E+06, 4.1004E+06, 4.1118E+06, 4.1232E+06,
     +  4.1347E+06, 4.1462E+06, 4.1577E+06, 4.1693E+06, 4.1809E+06,
     + 4.1925E+06, 4.2041E+06, 4.2157E+06, 4.2274E+06, 4.2391E+06,
     +  4.2509E+06, 4.2626E+06, 4.2744E+06, 4.2862E+06, 4.2981E+06,
     + 4.3099E+06, 4.3218E+06, 4.3337E+06, 4.3457E+06, 4.3576E+06,
     +  4.3696E+06, 4.3817E+06, 4.3937E+06, 4.4058E+06, 4.4179E+06,
     + 4.4300E+06, 4.4422E+06, 4.4544E+06, 4.4666E+06, 4.4788E+06,
     +  4.4911E+06, 4.5034E+06, 4.5157E+06, 4.5280E+06, 4.5404E+06,
     + 4.5528E+06, 4.5652E+06, 4.5777E+06, 4.5901E+06, 4.6026E+06,
     +  4.6152E+06, 4.6277E+06, 4.6403E+06, 4.6530E+06, 4.6656E+06,
     + 4.6783E+06, 4.6910E+06, 4.7037E+06, 4.7165E+06, 4.7292E+06,
     +  4.7420E+06, 4.7549E+06, 4.7678E+06, 4.7806E+06, 4.7936E+06,
     + 4.8065E+06, 4.8195E+06, 4.8325E+06, 4.8455E+06, 4.8586E+06,
     +  4.8717E+06, 4.8848E+06, 4.8980E+06, 4.9111E+06, 4.9244E+06,
     + 4.9376E+06, 4.9508E+06, 4.9641E+06, 4.9775E+06, 4.9908E+06,
     +  5.0042E+06, 5.0176E+06, 5.0310E+06, 5.0445E+06, 5.0580E+06,
     + 5.0715E+06, 5.0851E+06, 5.0986E+06, 5.1122E+06, 5.1259E+06,
     +  5.1396E+06, 5.1532E+06, 5.1670E+06, 5.1807E+06, 5.1945E+06,
     + 5.2083E+06, 5.2222E+06, 5.2360E+06, 5.2500E+06, 5.2639E+06,
     +  5.2778E+06, 5.2918E+06, 5.3059E+06, 5.3199E+06, 5.3340E+06,
     + 5.3481E+06, 5.3622E+06, 5.3764E+06, 5.3906E+06, 5.4049E+06,
     +  5.4191E+06, 5.4334E+06, 5.4477E+06, 5.4621E+06, 5.4765E+06,
     + 5.4909E+06, 5.5053E+06, 5.5198E+06, 5.5343E+06, 5.5488E+06,
     +  5.5634E+06, 5.5780E+06, 5.5926E+06, 5.6073E+06, 5.6220E+06,
     + 5.6367E+06, 5.6515E+06, 5.6662E+06, 5.6811E+06, 5.6959E+06,
     +  5.7108E+06, 5.7257E+06, 5.7406E+06, 5.7556E+06, 5.7706E+06,
     + 5.7856E+06, 5.8007E+06, 5.8158E+06, 5.8309E+06, 5.8461E+06,
     +  5.8613E+06, 5.8765E+06, 5.8917E+06, 5.9070E+06, 5.9223E+06,
     + 5.9377E+06, 5.9531E+06, 5.9685E+06, 5.9839E+06, 5.9994E+06,
     +  6.0149E+06, 6.0305E+06, 6.0460E+06, 6.0617E+06, 6.0773E+06,
     + 6.0930E+06, 6.1087E+06, 6.1244E+06, 6.1402E+06, 6.1560E+06,
     +  6.1718E+06, 6.1877E+06, 6.2036E+06, 6.2195E+06, 6.2355E+06,
     + 6.2515E+06, 6.2675E+06, 6.2836E+06, 6.2997E+06, 6.3158E+06,
     +  6.3320E+06, 6.3482E+06, 6.3644E+06, 6.3807E+06, 6.3970E+06,
     + 6.4133E+06, 6.4297E+06, 6.4461E+06, 6.4625E+06, 6.4790E+06,
     +  6.4954E+06, 6.5120E+06, 6.5285E+06, 6.5451E+06, 6.5618E+06,
     + 6.5785E+06, 6.5952E+06, 6.6119E+06, 6.6287E+06, 6.6455E+06,
     +  6.6623E+06, 6.6792E+06, 6.6961E+06, 6.7130E+06, 6.7300E+06,
     + 6.7470E+06, 6.7641E+06, 6.7811E+06, 6.7983E+06, 6.8154E+06,
     +  6.8326E+06, 6.8498E+06, 6.8671E+06, 6.8844E+06, 6.9017E+06,
     + 6.9190E+06, 6.9364E+06, 6.9539E+06, 6.9713E+06, 6.9888E+06,
     +  7.0064E+06, 7.0239E+06, 7.0415E+06, 7.0592E+06, 7.0769E+06,
     + 7.0946E+06, 7.1123E+06, 7.1301E+06, 7.1479E+06, 7.1658E+06,
     +  7.1837E+06, 7.2016E+06, 7.2196E+06, 7.2376E+06, 7.2556E+06,
     + 7.2737E+06, 7.2918E+06, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
      data (QofT( 4,J),J=2409,4510)/ 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     +  0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00,
     + 0.0000E+00, 0.0000E+00, 0.0000E+00, 0.0000E+00/
  
      eps=0.01
c
      TmI = Tmx(iso)
      ndat = 20+(TmI-20)/2
      gsi = xgj(iso)
c
      do I=1,ndat
        Q(I)=QofT(iso,I)
      enddo
 
c
c...value depends on temperature range
      if(T.lt.1. .OR. T.gt.TmI) then
        Qt = -1.
        write(*,'(a)') '  OUT OF TEMPERATURE RANGE'
        go to 99
      endif
  
C...check if T is one of the temperatures in our list
        do I = 1,ndat
          if(T.eq.Tdat(I))then
            Qt = Q(I)
            go to 99
          endif
        enddo
c
        do I=1,ndat
          if(T.gt.Tdat(I))then
            DT = Tdat(I+1) - Tdat(I)
            Qt = Q(I) + (Q(I+1)-Q(I))*(T - Tdat(I))/DT
          endif
        enddo
c
c      
   99 return
      end
